/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 6789
#define LARGE_STATE_COUNT 4111
#define SYMBOL_COUNT 585
#define ALIAS_COUNT 3
#define TOKEN_COUNT 286
#define EXTERNAL_TOKEN_COUNT 0
#define FIELD_COUNT 48
#define MAX_ALIAS_SEQUENCE_LENGTH 16
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 161
#define SUPERTYPE_COUNT 8

enum ts_symbol_identifiers {
  sym_identifier = 1,
  aux_sym_preproc_include_token1 = 2,
  aux_sym_preproc_include_token2 = 3,
  aux_sym_preproc_include_token3 = 4,
  aux_sym_preproc_def_token1 = 5,
  anon_sym_LPAREN = 6,
  anon_sym_DOT_DOT_DOT = 7,
  anon_sym_COMMA = 8,
  anon_sym_RPAREN = 9,
  aux_sym_preproc_if_token1 = 10,
  anon_sym_LF = 11,
  aux_sym_preproc_if_token2 = 12,
  aux_sym_preproc_ifdef_token1 = 13,
  aux_sym_preproc_ifdef_token2 = 14,
  aux_sym_preproc_else_token1 = 15,
  aux_sym_preproc_elif_token1 = 16,
  aux_sym_preproc_elifdef_token1 = 17,
  aux_sym_preproc_elifdef_token2 = 18,
  sym_preproc_arg = 19,
  sym_preproc_directive = 20,
  anon_sym_LPAREN2 = 21,
  anon_sym_defined = 22,
  anon_sym_BANG = 23,
  anon_sym_TILDE = 24,
  anon_sym_DASH = 25,
  anon_sym_PLUS = 26,
  anon_sym_STAR = 27,
  anon_sym_SLASH = 28,
  anon_sym_PERCENT = 29,
  anon_sym_PIPE_PIPE = 30,
  anon_sym_AMP_AMP = 31,
  anon_sym_PIPE = 32,
  anon_sym_CARET = 33,
  anon_sym_AMP = 34,
  anon_sym_EQ_EQ = 35,
  anon_sym_BANG_EQ = 36,
  anon_sym_GT = 37,
  anon_sym_GT_EQ = 38,
  anon_sym_LT_EQ = 39,
  anon_sym_LT = 40,
  anon_sym_LT_LT = 41,
  anon_sym_GT_GT = 42,
  anon_sym_SEMI = 43,
  anon_sym___extension__ = 44,
  anon_sym_typedef = 45,
  anon_sym_extern = 46,
  anon_sym___attribute__ = 47,
  anon_sym___attribute = 48,
  anon_sym_noreturn = 49,
  anon_sym_nothrow = 50,
  anon_sym_COLON_COLON = 51,
  anon_sym_LBRACK = 52,
  anon_sym_RBRACK = 53,
  anon_sym___declspec = 54,
  anon_sym___based = 55,
  anon_sym___cdecl = 56,
  anon_sym___clrcall = 57,
  anon_sym___stdcall = 58,
  anon_sym___fastcall = 59,
  anon_sym___thiscall = 60,
  anon_sym___vectorcall = 61,
  sym_ms_restrict_modifier = 62,
  sym_ms_unsigned_ptr_modifier = 63,
  sym_ms_signed_ptr_modifier = 64,
  anon_sym__unaligned = 65,
  anon_sym___unaligned = 66,
  anon_sym_LBRACE = 67,
  anon_sym_RBRACE = 68,
  anon_sym_signed = 69,
  anon_sym_unsigned = 70,
  anon_sym_long = 71,
  anon_sym_short = 72,
  anon_sym_static = 73,
  anon_sym_EQ = 74,
  anon_sym_ATautoreleasepool = 75,
  anon_sym_auto = 76,
  anon_sym_register = 77,
  anon_sym_inline = 78,
  anon_sym___inline = 79,
  anon_sym___inline__ = 80,
  anon_sym___forceinline = 81,
  anon_sym_thread_local = 82,
  anon_sym___thread = 83,
  anon_sym_CG_EXTERN = 84,
  anon_sym_CG_INLINE = 85,
  anon_sym_FOUNDATION_EXPORT = 86,
  anon_sym_FOUNDATION_EXTERN = 87,
  anon_sym_FOUNDATION_STATIC_INLINE = 88,
  anon_sym_IBOutlet = 89,
  anon_sym_IBInspectable = 90,
  anon_sym_IB_DESIGNABLE = 91,
  anon_sym_NS_INLINE = 92,
  anon_sym_NS_VALID_UNTIL_END_OF_SCOPE = 93,
  anon_sym_OBJC_EXPORT = 94,
  anon_sym_OBJC_ROOT_CLASS = 95,
  anon_sym_UIKIT_EXTERN = 96,
  anon_sym_const = 97,
  anon_sym_constexpr = 98,
  anon_sym_volatile = 99,
  anon_sym_restrict = 100,
  anon_sym___restrict__ = 101,
  anon_sym__Atomic = 102,
  anon_sym__Noreturn = 103,
  anon_sym__Nonnull = 104,
  anon_sym_nullable = 105,
  anon_sym__Complex = 106,
  anon_sym__Nullable = 107,
  anon_sym__Nullable_result = 108,
  anon_sym__Null_unspecified = 109,
  anon_sym___autoreleasing = 110,
  anon_sym___block = 111,
  anon_sym___bridge = 112,
  anon_sym___bridge_retained = 113,
  anon_sym___bridge_transfer = 114,
  anon_sym___complex = 115,
  anon_sym___const = 116,
  anon_sym___imag = 117,
  anon_sym___kindof = 118,
  anon_sym___nonnull = 119,
  anon_sym___nullable = 120,
  anon_sym___ptrauth_objc_class_ro = 121,
  anon_sym___ptrauth_objc_isa_pointer = 122,
  anon_sym___ptrauth_objc_super_pointer = 123,
  anon_sym___real = 124,
  anon_sym___strong = 125,
  anon_sym___unsafe_unretained = 126,
  anon_sym___unused = 127,
  anon_sym___weak = 128,
  anon_sym_alignas = 129,
  anon_sym__Alignas = 130,
  sym_primitive_type = 131,
  anon_sym_enum = 132,
  anon_sym_COLON = 133,
  anon_sym_struct = 134,
  anon_sym_union = 135,
  anon_sym_if = 136,
  anon_sym_else = 137,
  anon_sym_switch = 138,
  anon_sym_case = 139,
  anon_sym_default = 140,
  anon_sym_while = 141,
  anon_sym_do = 142,
  anon_sym_for = 143,
  anon_sym_in = 144,
  anon_sym_return = 145,
  anon_sym_break = 146,
  anon_sym_continue = 147,
  anon_sym_goto = 148,
  anon_sym_QMARK = 149,
  anon_sym_STAR_EQ = 150,
  anon_sym_SLASH_EQ = 151,
  anon_sym_PERCENT_EQ = 152,
  anon_sym_PLUS_EQ = 153,
  anon_sym_DASH_EQ = 154,
  anon_sym_LT_LT_EQ = 155,
  anon_sym_GT_GT_EQ = 156,
  anon_sym_AMP_EQ = 157,
  anon_sym_CARET_EQ = 158,
  anon_sym_PIPE_EQ = 159,
  anon_sym_DASH_DASH = 160,
  anon_sym_PLUS_PLUS = 161,
  anon_sym_sizeof = 162,
  anon_sym___alignof__ = 163,
  anon_sym___alignof = 164,
  anon_sym__alignof = 165,
  anon_sym_alignof = 166,
  anon_sym__Alignof = 167,
  anon_sym_offsetof = 168,
  anon_sym__Generic = 169,
  anon_sym_asm = 170,
  anon_sym___asm__ = 171,
  anon_sym___asm = 172,
  anon_sym___volatile__ = 173,
  anon_sym_LT2 = 174,
  anon_sym_DOT = 175,
  anon_sym_DASH_GT = 176,
  sym_number_literal = 177,
  anon_sym_L_SQUOTE = 178,
  anon_sym_u_SQUOTE = 179,
  anon_sym_U_SQUOTE = 180,
  anon_sym_u8_SQUOTE = 181,
  anon_sym_SQUOTE = 182,
  aux_sym_char_literal_token1 = 183,
  anon_sym_AT = 184,
  anon_sym_DQUOTE = 185,
  anon_sym_L_DQUOTE = 186,
  anon_sym_u_DQUOTE = 187,
  anon_sym_U_DQUOTE = 188,
  anon_sym_u8_DQUOTE = 189,
  aux_sym_string_literal_token1 = 190,
  sym_escape_sequence = 191,
  sym_system_lib_string = 192,
  sym_true = 193,
  sym_false = 194,
  anon_sym_NULL = 195,
  anon_sym_nullptr = 196,
  sym_comment = 197,
  anon_sym_objc_bridge_related = 198,
  anon_sym___typeof__ = 199,
  anon_sym___typeof = 200,
  anon_sym_typeof = 201,
  anon_sym_availability = 202,
  sym_version_number = 203,
  anon_sym_ios = 204,
  anon_sym_tvos = 205,
  anon_sym_macos = 206,
  anon_sym_macosx = 207,
  anon_sym_watchos = 208,
  anon_sym_ATimport = 209,
  aux_sym_preproc_undef_token1 = 210,
  anon_sym_POUND = 211,
  anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE = 212,
  anon_sym_NS_ROOT_CLASS = 213,
  anon_sym_NS_UNAVAILABLE = 214,
  anon_sym_NS_REQUIRES_NIL_TERMINATION = 215,
  anon_sym_CF_RETURNS_RETAINED = 216,
  anon_sym_CF_RETURNS_NOT_RETAINED = 217,
  anon_sym_DEPRECATED_ATTRIBUTE = 218,
  anon_sym_UI_APPEARANCE_SELECTOR = 219,
  anon_sym_UNAVAILABLE_ATTRIBUTE = 220,
  anon_sym_CF_FORMAT_FUNCTION = 221,
  anon_sym_NS_AVAILABLE = 222,
  anon_sym___IOS_AVAILABLE = 223,
  anon_sym_NS_AVAILABLE_IOS = 224,
  anon_sym_API_AVAILABLE = 225,
  anon_sym_API_UNAVAILABLE = 226,
  anon_sym_API_DEPRECATED = 227,
  anon_sym_NS_ENUM_AVAILABLE_IOS = 228,
  anon_sym_NS_DEPRECATED_IOS = 229,
  anon_sym_NS_ENUM_DEPRECATED_IOS = 230,
  anon_sym_NS_FORMAT_FUNCTION = 231,
  anon_sym_DEPRECATED_MSG_ATTRIBUTE = 232,
  anon_sym___deprecated_msg = 233,
  anon_sym___deprecated_enum_msg = 234,
  anon_sym_NS_SWIFT_NAME = 235,
  anon_sym_NS_SWIFT_UNAVAILABLE = 236,
  anon_sym_NS_EXTENSION_UNAVAILABLE_IOS = 237,
  anon_sym_NS_CLASS_AVAILABLE_IOS = 238,
  anon_sym_NS_CLASS_DEPRECATED_IOS = 239,
  anon_sym___OSX_AVAILABLE_STARTING = 240,
  anon_sym_ATprotocol = 241,
  anon_sym_class = 242,
  anon_sym_ATend = 243,
  anon_sym_ATinterface = 244,
  anon_sym_ATimplementation = 245,
  anon_sym___covariant = 246,
  anon_sym___contravariant = 247,
  anon_sym_ATprivate = 248,
  anon_sym_ATprotected = 249,
  anon_sym_ATpackage = 250,
  anon_sym_ATpublic = 251,
  anon_sym_ATcompatibility_alias = 252,
  anon_sym_AToptional = 253,
  anon_sym_ATrequired = 254,
  anon_sym_ATsynthesize = 255,
  anon_sym_ATdynamic = 256,
  anon_sym_LPARENclass_RPAREN = 257,
  anon_sym_ATproperty = 258,
  anon_sym_ATtry = 259,
  anon_sym___try = 260,
  anon_sym_ATcatch = 261,
  anon_sym___catch = 262,
  anon_sym_ATfinally = 263,
  anon_sym___finally = 264,
  anon_sym_ATthrow = 265,
  anon_sym_ATselector = 266,
  aux_sym_selector_expression_token1 = 267,
  anon_sym_ATavailable = 268,
  anon_sym___builtin_available = 269,
  anon_sym_va_arg = 270,
  aux_sym_ms_asm_block_token1 = 271,
  anon_sym_ATencode = 272,
  anon_sym_ATsynchronized = 273,
  anon_sym_BOOL = 274,
  anon_sym_IMP = 275,
  anon_sym_SEL = 276,
  anon_sym_Class = 277,
  anon_sym_id = 278,
  anon_sym_ATdefs = 279,
  anon_sym_out = 280,
  anon_sym_inout = 281,
  anon_sym_bycopy = 282,
  anon_sym_byref = 283,
  anon_sym_oneway = 284,
  anon_sym_COLON2 = 285,
  sym_translation_unit = 286,
  sym__top_level_item = 287,
  sym__block_item = 288,
  sym_preproc_include = 289,
  sym_preproc_def = 290,
  sym_preproc_function_def = 291,
  sym_preproc_params = 292,
  sym_preproc_call = 293,
  sym_preproc_if = 294,
  sym_preproc_ifdef = 295,
  sym_preproc_else = 296,
  sym_preproc_elif = 297,
  sym_preproc_elifdef = 298,
  sym_preproc_if_in_field_declaration_list = 299,
  sym_preproc_ifdef_in_field_declaration_list = 300,
  sym_preproc_else_in_field_declaration_list = 301,
  sym_preproc_elif_in_field_declaration_list = 302,
  sym_preproc_elifdef_in_field_declaration_list = 303,
  sym__preproc_expression = 304,
  sym_preproc_parenthesized_expression = 305,
  sym_preproc_defined = 306,
  sym_preproc_unary_expression = 307,
  sym_preproc_call_expression = 308,
  sym_preproc_argument_list = 309,
  sym_preproc_binary_expression = 310,
  sym_function_definition = 311,
  sym__old_style_function_definition = 312,
  sym_declaration = 313,
  sym_type_definition = 314,
  sym__type_definition_type = 315,
  sym__type_definition_declarators = 316,
  sym__declaration_modifiers = 317,
  sym__declaration_specifiers = 318,
  sym_linkage_specification = 319,
  sym_attribute_specifier = 320,
  sym_attribute = 321,
  sym_attribute_declaration = 322,
  sym_ms_declspec_modifier = 323,
  sym_ms_based_modifier = 324,
  sym_ms_call_modifier = 325,
  sym_ms_unaligned_ptr_modifier = 326,
  sym_ms_pointer_modifier = 327,
  sym_declaration_list = 328,
  sym__declarator = 329,
  sym__field_declarator = 330,
  sym__type_declarator = 331,
  sym__abstract_declarator = 332,
  sym_parenthesized_declarator = 333,
  sym_parenthesized_field_declarator = 334,
  sym_parenthesized_type_declarator = 335,
  sym_abstract_parenthesized_declarator = 336,
  sym_attributed_field_declarator = 337,
  sym_pointer_declarator = 338,
  sym_pointer_field_declarator = 339,
  sym_pointer_type_declarator = 340,
  sym_abstract_pointer_declarator = 341,
  sym_function_declarator = 342,
  sym_function_field_declarator = 343,
  sym_function_type_declarator = 344,
  sym_abstract_function_declarator = 345,
  sym__old_style_function_declarator = 346,
  sym_array_declarator = 347,
  sym_array_field_declarator = 348,
  sym_array_type_declarator = 349,
  sym_abstract_array_declarator = 350,
  sym_init_declarator = 351,
  sym_compound_statement = 352,
  sym_storage_class_specifier = 353,
  sym_type_qualifier = 354,
  sym_alignas_qualifier = 355,
  sym_type_specifier = 356,
  sym_sized_type_specifier = 357,
  sym_enum_specifier = 358,
  sym_enumerator_list = 359,
  sym_struct_specifier = 360,
  sym_union_specifier = 361,
  sym_field_declaration_list = 362,
  sym__field_declaration_list_item = 363,
  sym_field_declaration = 364,
  sym_bitfield_clause = 365,
  sym_enumerator = 366,
  sym_variadic_parameter = 367,
  sym_parameter_list = 368,
  sym__old_style_parameter_list = 369,
  sym_parameter_declaration = 370,
  sym_attributed_statement = 371,
  sym_statement = 372,
  sym__top_level_statement = 373,
  sym_labeled_statement = 374,
  sym__top_level_expression_statement = 375,
  sym_expression_statement = 376,
  sym_if_statement = 377,
  sym_else_clause = 378,
  sym_switch_statement = 379,
  sym_case_statement = 380,
  sym_while_statement = 381,
  sym_do_statement = 382,
  sym_for_statement = 383,
  sym__for_statement_body = 384,
  sym_return_statement = 385,
  sym_break_statement = 386,
  sym_continue_statement = 387,
  sym_goto_statement = 388,
  sym_expression = 389,
  sym__string = 390,
  sym_comma_expression = 391,
  sym_conditional_expression = 392,
  sym_assignment_expression = 393,
  sym_pointer_expression = 394,
  sym_unary_expression = 395,
  sym_binary_expression = 396,
  sym_update_expression = 397,
  sym_cast_expression = 398,
  sym_type_descriptor = 399,
  sym_sizeof_expression = 400,
  sym_alignof_expression = 401,
  sym_offsetof_expression = 402,
  sym_generic_expression = 403,
  sym_subscript_expression = 404,
  sym_call_expression = 405,
  sym_gnu_asm_expression = 406,
  sym_gnu_asm_qualifier = 407,
  sym_gnu_asm_output_operand_list = 408,
  sym_gnu_asm_output_operand = 409,
  sym_gnu_asm_input_operand_list = 410,
  sym_gnu_asm_input_operand = 411,
  sym_gnu_asm_clobber_list = 412,
  sym_gnu_asm_goto_list = 413,
  sym_extension_expression = 414,
  sym_argument_list = 415,
  sym_field_expression = 416,
  sym_compound_literal_expression = 417,
  sym_parenthesized_expression = 418,
  sym_initializer_list = 419,
  sym_initializer_pair = 420,
  sym_subscript_designator = 421,
  sym_subscript_range_designator = 422,
  sym_field_designator = 423,
  sym_char_literal = 424,
  sym_concatenated_string = 425,
  sym_string_literal = 426,
  sym_null = 427,
  sym__empty_declaration = 428,
  sym_macro_type_specifier = 429,
  sym_objc_bridge = 430,
  sym_typeof_specifier = 431,
  sym_availability = 432,
  sym_version = 433,
  sym_platform = 434,
  sym_module_import = 435,
  sym_preproc_if_in_implementation_definition = 436,
  sym_preproc_else_in_implementation_definition = 437,
  sym_preproc_elif_in_implementation_definition = 438,
  sym_preproc_if_in_interface_declaration = 439,
  sym_preproc_else_in_interface_declaration = 440,
  sym_preproc_elif_in_interface_declaration = 441,
  sym_preproc_ifdef_in_enumerator = 442,
  sym_preproc_else_in_enumerator = 443,
  sym_preproc_elif_in_enumerator = 444,
  sym_preproc_undef = 445,
  sym_preproc_linemarker = 446,
  sym_availability_attribute_specifier = 447,
  sym_protocol_forward_declaration = 448,
  sym_class_declaration = 449,
  sym_class_interface = 450,
  sym__class_interface_header = 451,
  sym__type_params = 452,
  sym__class_interface_inheritance = 453,
  sym_class_implementation = 454,
  sym__class_implementation_header = 455,
  sym__class_implementation_inheritance = 456,
  sym_protocol_reference_list = 457,
  sym_parameterized_arguments = 458,
  sym_generic_arguments = 459,
  sym_instance_variables = 460,
  sym_instance_variable = 461,
  sym_visibility_specification = 462,
  sym_protocol_declaration = 463,
  sym_compatibility_alias_declaration = 464,
  sym_qualified_protocol_interface_declaration = 465,
  sym_implementation_definition = 466,
  sym_property_implementation = 467,
  sym_method_definition = 468,
  sym_method_type = 469,
  sym_keyword_declarator = 470,
  sym_property_declaration = 471,
  sym_property_attributes_declaration = 472,
  sym_property_attribute = 473,
  sym_method_declaration = 474,
  sym_method_parameter = 475,
  sym_c_method_parameter = 476,
  sym_struct_declaration = 477,
  sym_atomic_declaration = 478,
  sym_specifier_qualifier = 479,
  sym_struct_declarator = 480,
  sym_try_statement = 481,
  sym_catch_clause = 482,
  sym_finally_clause = 483,
  sym_throw_statement = 484,
  sym_selector_expression = 485,
  sym_available_expression = 486,
  sym_range_expression = 487,
  sym_block_literal = 488,
  sym_message_expression = 489,
  sym_va_arg_expression = 490,
  sym_ms_asm_block = 491,
  sym_encode_expression = 492,
  sym_synchronized_statement = 493,
  sym_block_pointer_declarator = 494,
  sym_block_pointer_field_declarator = 495,
  sym_block_pointer_type_declarator = 496,
  sym_abstract_block_pointer_declarator = 497,
  sym_generic_specifier = 498,
  sym_typedefed_specifier = 499,
  sym_array_type_specifier = 500,
  sym_atdef_field = 501,
  sym_protocol_qualifier = 502,
  sym_type_name = 503,
  sym_at_expression = 504,
  sym_dictionary_literal = 505,
  sym_dictionary_pair = 506,
  sym_array_literal = 507,
  sym_method_identifier = 508,
  aux_sym_translation_unit_repeat1 = 509,
  aux_sym_preproc_params_repeat1 = 510,
  aux_sym_preproc_if_repeat1 = 511,
  aux_sym_preproc_if_in_field_declaration_list_repeat1 = 512,
  aux_sym_preproc_argument_list_repeat1 = 513,
  aux_sym__old_style_function_definition_repeat1 = 514,
  aux_sym_declaration_repeat1 = 515,
  aux_sym__type_definition_type_repeat1 = 516,
  aux_sym__type_definition_declarators_repeat1 = 517,
  aux_sym__declaration_specifiers_repeat1 = 518,
  aux_sym_attribute_specifier_repeat1 = 519,
  aux_sym_attribute_repeat1 = 520,
  aux_sym_attribute_declaration_repeat1 = 521,
  aux_sym_declaration_list_repeat1 = 522,
  aux_sym_attributed_field_declarator_repeat1 = 523,
  aux_sym_pointer_declarator_repeat1 = 524,
  aux_sym_function_declarator_repeat1 = 525,
  aux_sym_function_field_declarator_repeat1 = 526,
  aux_sym_array_declarator_repeat1 = 527,
  aux_sym_sized_type_specifier_repeat1 = 528,
  aux_sym_enumerator_list_repeat1 = 529,
  aux_sym_field_declaration_repeat1 = 530,
  aux_sym_parameter_list_repeat1 = 531,
  aux_sym__old_style_parameter_list_repeat1 = 532,
  aux_sym_case_statement_repeat1 = 533,
  aux_sym_generic_expression_repeat1 = 534,
  aux_sym_gnu_asm_expression_repeat1 = 535,
  aux_sym_gnu_asm_output_operand_list_repeat1 = 536,
  aux_sym_gnu_asm_input_operand_list_repeat1 = 537,
  aux_sym_gnu_asm_clobber_list_repeat1 = 538,
  aux_sym_gnu_asm_goto_list_repeat1 = 539,
  aux_sym_argument_list_repeat1 = 540,
  aux_sym_argument_list_repeat2 = 541,
  aux_sym_initializer_list_repeat1 = 542,
  aux_sym_initializer_pair_repeat1 = 543,
  aux_sym_char_literal_repeat1 = 544,
  aux_sym_concatenated_string_repeat1 = 545,
  aux_sym_string_literal_repeat1 = 546,
  aux_sym_availability_repeat1 = 547,
  aux_sym_version_repeat1 = 548,
  aux_sym_module_import_repeat1 = 549,
  aux_sym_preproc_if_in_implementation_definition_repeat1 = 550,
  aux_sym_preproc_if_in_interface_declaration_repeat1 = 551,
  aux_sym_preproc_ifdef_in_enumerator_repeat1 = 552,
  aux_sym_availability_attribute_specifier_repeat1 = 553,
  aux_sym_protocol_forward_declaration_repeat1 = 554,
  aux_sym_class_declaration_repeat1 = 555,
  aux_sym_class_interface_repeat1 = 556,
  aux_sym_class_implementation_repeat1 = 557,
  aux_sym_parameterized_arguments_repeat1 = 558,
  aux_sym_parameterized_arguments_repeat2 = 559,
  aux_sym_generic_arguments_repeat1 = 560,
  aux_sym_instance_variables_repeat1 = 561,
  aux_sym_protocol_declaration_repeat1 = 562,
  aux_sym_property_implementation_repeat1 = 563,
  aux_sym_method_definition_repeat1 = 564,
  aux_sym_method_definition_repeat2 = 565,
  aux_sym_method_type_repeat1 = 566,
  aux_sym_keyword_selector_repeat1 = 567,
  aux_sym_property_attributes_declaration_repeat1 = 568,
  aux_sym_method_declaration_repeat1 = 569,
  aux_sym_method_declaration_repeat2 = 570,
  aux_sym_c_method_parameter_repeat1 = 571,
  aux_sym_struct_declaration_repeat1 = 572,
  aux_sym_struct_declaration_repeat2 = 573,
  aux_sym_try_statement_repeat1 = 574,
  aux_sym_selector_expression_repeat1 = 575,
  aux_sym_selector_expression_repeat2 = 576,
  aux_sym_available_expression_repeat1 = 577,
  aux_sym_message_expression_repeat1 = 578,
  aux_sym_message_expression_repeat2 = 579,
  aux_sym_generic_specifier_repeat1 = 580,
  aux_sym_type_name_repeat1 = 581,
  aux_sym_dictionary_literal_repeat1 = 582,
  aux_sym_method_identifier_repeat1 = 583,
  aux_sym_method_identifier_repeat2 = 584,
  alias_sym_field_identifier = 585,
  alias_sym_statement_identifier = 586,
  alias_sym_type_identifier = 587,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [aux_sym_preproc_include_token1] = "#include",
  [aux_sym_preproc_include_token2] = "#import",
  [aux_sym_preproc_include_token3] = "preproc_include_token3",
  [aux_sym_preproc_def_token1] = "#define",
  [anon_sym_LPAREN] = "(",
  [anon_sym_DOT_DOT_DOT] = "...",
  [anon_sym_COMMA] = ",",
  [anon_sym_RPAREN] = ")",
  [aux_sym_preproc_if_token1] = "#if",
  [anon_sym_LF] = "\n",
  [aux_sym_preproc_if_token2] = "#endif",
  [aux_sym_preproc_ifdef_token1] = "#ifdef",
  [aux_sym_preproc_ifdef_token2] = "#ifndef",
  [aux_sym_preproc_else_token1] = "#else",
  [aux_sym_preproc_elif_token1] = "#elif",
  [aux_sym_preproc_elifdef_token1] = "#elifdef",
  [aux_sym_preproc_elifdef_token2] = "#elifndef",
  [sym_preproc_arg] = "preproc_arg",
  [sym_preproc_directive] = "preproc_directive",
  [anon_sym_LPAREN2] = "(",
  [anon_sym_defined] = "defined",
  [anon_sym_BANG] = "!",
  [anon_sym_TILDE] = "~",
  [anon_sym_DASH] = "-",
  [anon_sym_PLUS] = "+",
  [anon_sym_STAR] = "*",
  [anon_sym_SLASH] = "/",
  [anon_sym_PERCENT] = "%",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_PIPE] = "|",
  [anon_sym_CARET] = "^",
  [anon_sym_AMP] = "&",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_GT] = ">",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_LT] = "<",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_SEMI] = ";",
  [anon_sym___extension__] = "__extension__",
  [anon_sym_typedef] = "typedef",
  [anon_sym_extern] = "extern",
  [anon_sym___attribute__] = "__attribute__",
  [anon_sym___attribute] = "__attribute",
  [anon_sym_noreturn] = "noreturn",
  [anon_sym_nothrow] = "nothrow",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym___declspec] = "__declspec",
  [anon_sym___based] = "__based",
  [anon_sym___cdecl] = "__cdecl",
  [anon_sym___clrcall] = "__clrcall",
  [anon_sym___stdcall] = "__stdcall",
  [anon_sym___fastcall] = "__fastcall",
  [anon_sym___thiscall] = "__thiscall",
  [anon_sym___vectorcall] = "__vectorcall",
  [sym_ms_restrict_modifier] = "ms_restrict_modifier",
  [sym_ms_unsigned_ptr_modifier] = "ms_unsigned_ptr_modifier",
  [sym_ms_signed_ptr_modifier] = "ms_signed_ptr_modifier",
  [anon_sym__unaligned] = "_unaligned",
  [anon_sym___unaligned] = "__unaligned",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_signed] = "signed",
  [anon_sym_unsigned] = "unsigned",
  [anon_sym_long] = "long",
  [anon_sym_short] = "short",
  [anon_sym_static] = "static",
  [anon_sym_EQ] = "=",
  [anon_sym_ATautoreleasepool] = "@autoreleasepool",
  [anon_sym_auto] = "auto",
  [anon_sym_register] = "register",
  [anon_sym_inline] = "inline",
  [anon_sym___inline] = "__inline",
  [anon_sym___inline__] = "__inline__",
  [anon_sym___forceinline] = "__forceinline",
  [anon_sym_thread_local] = "thread_local",
  [anon_sym___thread] = "__thread",
  [anon_sym_CG_EXTERN] = "CG_EXTERN",
  [anon_sym_CG_INLINE] = "CG_INLINE",
  [anon_sym_FOUNDATION_EXPORT] = "FOUNDATION_EXPORT",
  [anon_sym_FOUNDATION_EXTERN] = "FOUNDATION_EXTERN",
  [anon_sym_FOUNDATION_STATIC_INLINE] = "FOUNDATION_STATIC_INLINE",
  [anon_sym_IBOutlet] = "IBOutlet",
  [anon_sym_IBInspectable] = "IBInspectable",
  [anon_sym_IB_DESIGNABLE] = "IB_DESIGNABLE",
  [anon_sym_NS_INLINE] = "NS_INLINE",
  [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = "NS_VALID_UNTIL_END_OF_SCOPE",
  [anon_sym_OBJC_EXPORT] = "OBJC_EXPORT",
  [anon_sym_OBJC_ROOT_CLASS] = "OBJC_ROOT_CLASS",
  [anon_sym_UIKIT_EXTERN] = "UIKIT_EXTERN",
  [anon_sym_const] = "const",
  [anon_sym_constexpr] = "constexpr",
  [anon_sym_volatile] = "volatile",
  [anon_sym_restrict] = "restrict",
  [anon_sym___restrict__] = "__restrict__",
  [anon_sym__Atomic] = "_Atomic",
  [anon_sym__Noreturn] = "_Noreturn",
  [anon_sym__Nonnull] = "_Nonnull",
  [anon_sym_nullable] = "nullable",
  [anon_sym__Complex] = "_Complex",
  [anon_sym__Nullable] = "_Nullable",
  [anon_sym__Nullable_result] = "_Nullable_result",
  [anon_sym__Null_unspecified] = "_Null_unspecified",
  [anon_sym___autoreleasing] = "__autoreleasing",
  [anon_sym___block] = "__block",
  [anon_sym___bridge] = "__bridge",
  [anon_sym___bridge_retained] = "__bridge_retained",
  [anon_sym___bridge_transfer] = "__bridge_transfer",
  [anon_sym___complex] = "__complex",
  [anon_sym___const] = "__const",
  [anon_sym___imag] = "__imag",
  [anon_sym___kindof] = "__kindof",
  [anon_sym___nonnull] = "__nonnull",
  [anon_sym___nullable] = "__nullable",
  [anon_sym___ptrauth_objc_class_ro] = "__ptrauth_objc_class_ro",
  [anon_sym___ptrauth_objc_isa_pointer] = "__ptrauth_objc_isa_pointer",
  [anon_sym___ptrauth_objc_super_pointer] = "__ptrauth_objc_super_pointer",
  [anon_sym___real] = "__real",
  [anon_sym___strong] = "__strong",
  [anon_sym___unsafe_unretained] = "__unsafe_unretained",
  [anon_sym___unused] = "__unused",
  [anon_sym___weak] = "__weak",
  [anon_sym_alignas] = "alignas",
  [anon_sym__Alignas] = "_Alignas",
  [sym_primitive_type] = "primitive_type",
  [anon_sym_enum] = "enum",
  [anon_sym_COLON] = ":",
  [anon_sym_struct] = "struct",
  [anon_sym_union] = "union",
  [anon_sym_if] = "if",
  [anon_sym_else] = "else",
  [anon_sym_switch] = "switch",
  [anon_sym_case] = "case",
  [anon_sym_default] = "default",
  [anon_sym_while] = "while",
  [anon_sym_do] = "do",
  [anon_sym_for] = "for",
  [anon_sym_in] = "in",
  [anon_sym_return] = "return",
  [anon_sym_break] = "break",
  [anon_sym_continue] = "continue",
  [anon_sym_goto] = "goto",
  [anon_sym_QMARK] = "\?",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_sizeof] = "sizeof",
  [anon_sym___alignof__] = "__alignof__",
  [anon_sym___alignof] = "__alignof",
  [anon_sym__alignof] = "_alignof",
  [anon_sym_alignof] = "alignof",
  [anon_sym__Alignof] = "_Alignof",
  [anon_sym_offsetof] = "offsetof",
  [anon_sym__Generic] = "_Generic",
  [anon_sym_asm] = "asm",
  [anon_sym___asm__] = "__asm__",
  [anon_sym___asm] = "__asm",
  [anon_sym___volatile__] = "__volatile__",
  [anon_sym_LT2] = "<",
  [anon_sym_DOT] = ".",
  [anon_sym_DASH_GT] = "->",
  [sym_number_literal] = "number_literal",
  [anon_sym_L_SQUOTE] = "L'",
  [anon_sym_u_SQUOTE] = "u'",
  [anon_sym_U_SQUOTE] = "U'",
  [anon_sym_u8_SQUOTE] = "u8'",
  [anon_sym_SQUOTE] = "'",
  [aux_sym_char_literal_token1] = "character",
  [anon_sym_AT] = "@",
  [anon_sym_DQUOTE] = "\"",
  [anon_sym_L_DQUOTE] = "L\"",
  [anon_sym_u_DQUOTE] = "u\"",
  [anon_sym_U_DQUOTE] = "U\"",
  [anon_sym_u8_DQUOTE] = "u8\"",
  [aux_sym_string_literal_token1] = "string_content",
  [sym_escape_sequence] = "escape_sequence",
  [sym_system_lib_string] = "system_lib_string",
  [sym_true] = "true",
  [sym_false] = "false",
  [anon_sym_NULL] = "NULL",
  [anon_sym_nullptr] = "nullptr",
  [sym_comment] = "comment",
  [anon_sym_objc_bridge_related] = "objc_bridge_related",
  [anon_sym___typeof__] = "__typeof__",
  [anon_sym___typeof] = "__typeof",
  [anon_sym_typeof] = "typeof",
  [anon_sym_availability] = "availability",
  [sym_version_number] = "version_number",
  [anon_sym_ios] = "ios",
  [anon_sym_tvos] = "tvos",
  [anon_sym_macos] = "macos",
  [anon_sym_macosx] = "macosx",
  [anon_sym_watchos] = "watchos",
  [anon_sym_ATimport] = "@import",
  [aux_sym_preproc_undef_token1] = "#undef",
  [anon_sym_POUND] = "#",
  [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = "NS_AUTOMATED_REFCOUNT_UNAVAILABLE",
  [anon_sym_NS_ROOT_CLASS] = "NS_ROOT_CLASS",
  [anon_sym_NS_UNAVAILABLE] = "NS_UNAVAILABLE",
  [anon_sym_NS_REQUIRES_NIL_TERMINATION] = "NS_REQUIRES_NIL_TERMINATION",
  [anon_sym_CF_RETURNS_RETAINED] = "CF_RETURNS_RETAINED",
  [anon_sym_CF_RETURNS_NOT_RETAINED] = "CF_RETURNS_NOT_RETAINED",
  [anon_sym_DEPRECATED_ATTRIBUTE] = "DEPRECATED_ATTRIBUTE",
  [anon_sym_UI_APPEARANCE_SELECTOR] = "UI_APPEARANCE_SELECTOR",
  [anon_sym_UNAVAILABLE_ATTRIBUTE] = "UNAVAILABLE_ATTRIBUTE",
  [anon_sym_CF_FORMAT_FUNCTION] = "CF_FORMAT_FUNCTION",
  [anon_sym_NS_AVAILABLE] = "NS_AVAILABLE",
  [anon_sym___IOS_AVAILABLE] = "__IOS_AVAILABLE",
  [anon_sym_NS_AVAILABLE_IOS] = "NS_AVAILABLE_IOS",
  [anon_sym_API_AVAILABLE] = "API_AVAILABLE",
  [anon_sym_API_UNAVAILABLE] = "API_UNAVAILABLE",
  [anon_sym_API_DEPRECATED] = "API_DEPRECATED",
  [anon_sym_NS_ENUM_AVAILABLE_IOS] = "NS_ENUM_AVAILABLE_IOS",
  [anon_sym_NS_DEPRECATED_IOS] = "NS_DEPRECATED_IOS",
  [anon_sym_NS_ENUM_DEPRECATED_IOS] = "NS_ENUM_DEPRECATED_IOS",
  [anon_sym_NS_FORMAT_FUNCTION] = "NS_FORMAT_FUNCTION",
  [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = "DEPRECATED_MSG_ATTRIBUTE",
  [anon_sym___deprecated_msg] = "__deprecated_msg",
  [anon_sym___deprecated_enum_msg] = "__deprecated_enum_msg",
  [anon_sym_NS_SWIFT_NAME] = "NS_SWIFT_NAME",
  [anon_sym_NS_SWIFT_UNAVAILABLE] = "NS_SWIFT_UNAVAILABLE",
  [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = "NS_EXTENSION_UNAVAILABLE_IOS",
  [anon_sym_NS_CLASS_AVAILABLE_IOS] = "NS_CLASS_AVAILABLE_IOS",
  [anon_sym_NS_CLASS_DEPRECATED_IOS] = "NS_CLASS_DEPRECATED_IOS",
  [anon_sym___OSX_AVAILABLE_STARTING] = "__OSX_AVAILABLE_STARTING",
  [anon_sym_ATprotocol] = "@protocol",
  [anon_sym_class] = "class",
  [anon_sym_ATend] = "@end",
  [anon_sym_ATinterface] = "@interface",
  [anon_sym_ATimplementation] = "@implementation",
  [anon_sym___covariant] = "__covariant",
  [anon_sym___contravariant] = "__contravariant",
  [anon_sym_ATprivate] = "@private",
  [anon_sym_ATprotected] = "@protected",
  [anon_sym_ATpackage] = "@package",
  [anon_sym_ATpublic] = "@public",
  [anon_sym_ATcompatibility_alias] = "@compatibility_alias",
  [anon_sym_AToptional] = "@optional",
  [anon_sym_ATrequired] = "@required",
  [anon_sym_ATsynthesize] = "@synthesize",
  [anon_sym_ATdynamic] = "@dynamic",
  [anon_sym_LPARENclass_RPAREN] = "(class)",
  [anon_sym_ATproperty] = "@property",
  [anon_sym_ATtry] = "@try",
  [anon_sym___try] = "__try",
  [anon_sym_ATcatch] = "@catch",
  [anon_sym___catch] = "__catch",
  [anon_sym_ATfinally] = "@finally",
  [anon_sym___finally] = "__finally",
  [anon_sym_ATthrow] = "@throw",
  [anon_sym_ATselector] = "@selector",
  [aux_sym_selector_expression_token1] = "selector_expression_token1",
  [anon_sym_ATavailable] = "@available",
  [anon_sym___builtin_available] = "__builtin_available",
  [anon_sym_va_arg] = "va_arg",
  [aux_sym_ms_asm_block_token1] = "ms_asm_block_token1",
  [anon_sym_ATencode] = "@encode",
  [anon_sym_ATsynchronized] = "@synchronized",
  [anon_sym_BOOL] = "BOOL",
  [anon_sym_IMP] = "IMP",
  [anon_sym_SEL] = "SEL",
  [anon_sym_Class] = "Class",
  [anon_sym_id] = "id",
  [anon_sym_ATdefs] = "@defs",
  [anon_sym_out] = "out",
  [anon_sym_inout] = "inout",
  [anon_sym_bycopy] = "bycopy",
  [anon_sym_byref] = "byref",
  [anon_sym_oneway] = "oneway",
  [anon_sym_COLON2] = ":",
  [sym_translation_unit] = "translation_unit",
  [sym__top_level_item] = "_top_level_item",
  [sym__block_item] = "_block_item",
  [sym_preproc_include] = "preproc_include",
  [sym_preproc_def] = "preproc_def",
  [sym_preproc_function_def] = "preproc_function_def",
  [sym_preproc_params] = "preproc_params",
  [sym_preproc_call] = "preproc_call",
  [sym_preproc_if] = "preproc_if",
  [sym_preproc_ifdef] = "preproc_ifdef",
  [sym_preproc_else] = "preproc_else",
  [sym_preproc_elif] = "preproc_elif",
  [sym_preproc_elifdef] = "preproc_elifdef",
  [sym_preproc_if_in_field_declaration_list] = "preproc_if",
  [sym_preproc_ifdef_in_field_declaration_list] = "preproc_ifdef",
  [sym_preproc_else_in_field_declaration_list] = "preproc_else",
  [sym_preproc_elif_in_field_declaration_list] = "preproc_elif",
  [sym_preproc_elifdef_in_field_declaration_list] = "preproc_elifdef",
  [sym__preproc_expression] = "_preproc_expression",
  [sym_preproc_parenthesized_expression] = "parenthesized_expression",
  [sym_preproc_defined] = "preproc_defined",
  [sym_preproc_unary_expression] = "unary_expression",
  [sym_preproc_call_expression] = "call_expression",
  [sym_preproc_argument_list] = "argument_list",
  [sym_preproc_binary_expression] = "binary_expression",
  [sym_function_definition] = "function_definition",
  [sym__old_style_function_definition] = "function_definition",
  [sym_declaration] = "declaration",
  [sym_type_definition] = "type_definition",
  [sym__type_definition_type] = "_type_definition_type",
  [sym__type_definition_declarators] = "_type_definition_declarators",
  [sym__declaration_modifiers] = "_declaration_modifiers",
  [sym__declaration_specifiers] = "_declaration_specifiers",
  [sym_linkage_specification] = "linkage_specification",
  [sym_attribute_specifier] = "attribute_specifier",
  [sym_attribute] = "attribute",
  [sym_attribute_declaration] = "attribute_declaration",
  [sym_ms_declspec_modifier] = "ms_declspec_modifier",
  [sym_ms_based_modifier] = "ms_based_modifier",
  [sym_ms_call_modifier] = "ms_call_modifier",
  [sym_ms_unaligned_ptr_modifier] = "ms_unaligned_ptr_modifier",
  [sym_ms_pointer_modifier] = "ms_pointer_modifier",
  [sym_declaration_list] = "declaration_list",
  [sym__declarator] = "_declarator",
  [sym__field_declarator] = "_field_declarator",
  [sym__type_declarator] = "_type_declarator",
  [sym__abstract_declarator] = "_abstract_declarator",
  [sym_parenthesized_declarator] = "parenthesized_declarator",
  [sym_parenthesized_field_declarator] = "parenthesized_declarator",
  [sym_parenthesized_type_declarator] = "parenthesized_declarator",
  [sym_abstract_parenthesized_declarator] = "abstract_parenthesized_declarator",
  [sym_attributed_field_declarator] = "attributed_declarator",
  [sym_pointer_declarator] = "pointer_declarator",
  [sym_pointer_field_declarator] = "pointer_declarator",
  [sym_pointer_type_declarator] = "pointer_declarator",
  [sym_abstract_pointer_declarator] = "abstract_pointer_declarator",
  [sym_function_declarator] = "function_declarator",
  [sym_function_field_declarator] = "function_declarator",
  [sym_function_type_declarator] = "function_declarator",
  [sym_abstract_function_declarator] = "abstract_function_declarator",
  [sym__old_style_function_declarator] = "function_declarator",
  [sym_array_declarator] = "array_declarator",
  [sym_array_field_declarator] = "array_declarator",
  [sym_array_type_declarator] = "array_declarator",
  [sym_abstract_array_declarator] = "abstract_array_declarator",
  [sym_init_declarator] = "init_declarator",
  [sym_compound_statement] = "compound_statement",
  [sym_storage_class_specifier] = "storage_class_specifier",
  [sym_type_qualifier] = "type_qualifier",
  [sym_alignas_qualifier] = "alignas_qualifier",
  [sym_type_specifier] = "type_specifier",
  [sym_sized_type_specifier] = "sized_type_specifier",
  [sym_enum_specifier] = "enum_specifier",
  [sym_enumerator_list] = "enumerator_list",
  [sym_struct_specifier] = "struct_specifier",
  [sym_union_specifier] = "union_specifier",
  [sym_field_declaration_list] = "field_declaration_list",
  [sym__field_declaration_list_item] = "_field_declaration_list_item",
  [sym_field_declaration] = "field_declaration",
  [sym_bitfield_clause] = "bitfield_clause",
  [sym_enumerator] = "enumerator",
  [sym_variadic_parameter] = "variadic_parameter",
  [sym_parameter_list] = "parameter_list",
  [sym__old_style_parameter_list] = "parameter_list",
  [sym_parameter_declaration] = "parameter_declaration",
  [sym_attributed_statement] = "attributed_statement",
  [sym_statement] = "statement",
  [sym__top_level_statement] = "_top_level_statement",
  [sym_labeled_statement] = "labeled_statement",
  [sym__top_level_expression_statement] = "expression_statement",
  [sym_expression_statement] = "expression_statement",
  [sym_if_statement] = "if_statement",
  [sym_else_clause] = "else_clause",
  [sym_switch_statement] = "switch_statement",
  [sym_case_statement] = "case_statement",
  [sym_while_statement] = "while_statement",
  [sym_do_statement] = "do_statement",
  [sym_for_statement] = "for_statement",
  [sym__for_statement_body] = "_for_statement_body",
  [sym_return_statement] = "return_statement",
  [sym_break_statement] = "break_statement",
  [sym_continue_statement] = "continue_statement",
  [sym_goto_statement] = "goto_statement",
  [sym_expression] = "expression",
  [sym__string] = "_string",
  [sym_comma_expression] = "comma_expression",
  [sym_conditional_expression] = "conditional_expression",
  [sym_assignment_expression] = "assignment_expression",
  [sym_pointer_expression] = "pointer_expression",
  [sym_unary_expression] = "unary_expression",
  [sym_binary_expression] = "binary_expression",
  [sym_update_expression] = "update_expression",
  [sym_cast_expression] = "cast_expression",
  [sym_type_descriptor] = "type_descriptor",
  [sym_sizeof_expression] = "sizeof_expression",
  [sym_alignof_expression] = "alignof_expression",
  [sym_offsetof_expression] = "offsetof_expression",
  [sym_generic_expression] = "generic_expression",
  [sym_subscript_expression] = "subscript_expression",
  [sym_call_expression] = "call_expression",
  [sym_gnu_asm_expression] = "gnu_asm_expression",
  [sym_gnu_asm_qualifier] = "gnu_asm_qualifier",
  [sym_gnu_asm_output_operand_list] = "gnu_asm_output_operand_list",
  [sym_gnu_asm_output_operand] = "gnu_asm_output_operand",
  [sym_gnu_asm_input_operand_list] = "gnu_asm_input_operand_list",
  [sym_gnu_asm_input_operand] = "gnu_asm_input_operand",
  [sym_gnu_asm_clobber_list] = "gnu_asm_clobber_list",
  [sym_gnu_asm_goto_list] = "gnu_asm_goto_list",
  [sym_extension_expression] = "extension_expression",
  [sym_argument_list] = "argument_list",
  [sym_field_expression] = "field_expression",
  [sym_compound_literal_expression] = "compound_literal_expression",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym_initializer_list] = "initializer_list",
  [sym_initializer_pair] = "initializer_pair",
  [sym_subscript_designator] = "subscript_designator",
  [sym_subscript_range_designator] = "subscript_range_designator",
  [sym_field_designator] = "field_designator",
  [sym_char_literal] = "char_literal",
  [sym_concatenated_string] = "concatenated_string",
  [sym_string_literal] = "string_literal",
  [sym_null] = "null",
  [sym__empty_declaration] = "_empty_declaration",
  [sym_macro_type_specifier] = "macro_type_specifier",
  [sym_objc_bridge] = "objc_bridge",
  [sym_typeof_specifier] = "typeof_specifier",
  [sym_availability] = "availability",
  [sym_version] = "version",
  [sym_platform] = "platform",
  [sym_module_import] = "module_import",
  [sym_preproc_if_in_implementation_definition] = "preproc_if",
  [sym_preproc_else_in_implementation_definition] = "preproc_else",
  [sym_preproc_elif_in_implementation_definition] = "preproc_elif",
  [sym_preproc_if_in_interface_declaration] = "preproc_if",
  [sym_preproc_else_in_interface_declaration] = "preproc_else",
  [sym_preproc_elif_in_interface_declaration] = "preproc_elif",
  [sym_preproc_ifdef_in_enumerator] = "preproc_ifdef",
  [sym_preproc_else_in_enumerator] = "preproc_else",
  [sym_preproc_elif_in_enumerator] = "preproc_elif",
  [sym_preproc_undef] = "preproc_undef",
  [sym_preproc_linemarker] = "preproc_linemarker",
  [sym_availability_attribute_specifier] = "availability_attribute_specifier",
  [sym_protocol_forward_declaration] = "protocol_forward_declaration",
  [sym_class_declaration] = "class_declaration",
  [sym_class_interface] = "class_interface",
  [sym__class_interface_header] = "_class_interface_header",
  [sym__type_params] = "_type_params",
  [sym__class_interface_inheritance] = "_class_interface_inheritance",
  [sym_class_implementation] = "class_implementation",
  [sym__class_implementation_header] = "_class_implementation_header",
  [sym__class_implementation_inheritance] = "_class_implementation_inheritance",
  [sym_protocol_reference_list] = "protocol_reference_list",
  [sym_parameterized_arguments] = "parameterized_arguments",
  [sym_generic_arguments] = "generic_arguments",
  [sym_instance_variables] = "instance_variables",
  [sym_instance_variable] = "instance_variable",
  [sym_visibility_specification] = "visibility_specification",
  [sym_protocol_declaration] = "protocol_declaration",
  [sym_compatibility_alias_declaration] = "compatibility_alias_declaration",
  [sym_qualified_protocol_interface_declaration] = "qualified_protocol_interface_declaration",
  [sym_implementation_definition] = "implementation_definition",
  [sym_property_implementation] = "property_implementation",
  [sym_method_definition] = "method_definition",
  [sym_method_type] = "method_type",
  [sym_keyword_declarator] = "keyword_declarator",
  [sym_property_declaration] = "property_declaration",
  [sym_property_attributes_declaration] = "property_attributes_declaration",
  [sym_property_attribute] = "property_attribute",
  [sym_method_declaration] = "method_declaration",
  [sym_method_parameter] = "method_parameter",
  [sym_c_method_parameter] = "method_parameter",
  [sym_struct_declaration] = "struct_declaration",
  [sym_atomic_declaration] = "atomic_declaration",
  [sym_specifier_qualifier] = "specifier_qualifier",
  [sym_struct_declarator] = "struct_declarator",
  [sym_try_statement] = "try_statement",
  [sym_catch_clause] = "catch_clause",
  [sym_finally_clause] = "finally_clause",
  [sym_throw_statement] = "throw_statement",
  [sym_selector_expression] = "selector_expression",
  [sym_available_expression] = "available_expression",
  [sym_range_expression] = "range_expression",
  [sym_block_literal] = "block_literal",
  [sym_message_expression] = "message_expression",
  [sym_va_arg_expression] = "va_arg_expression",
  [sym_ms_asm_block] = "ms_asm_block",
  [sym_encode_expression] = "encode_expression",
  [sym_synchronized_statement] = "synchronized_statement",
  [sym_block_pointer_declarator] = "block_pointer_declarator",
  [sym_block_pointer_field_declarator] = "block_pointer_declarator",
  [sym_block_pointer_type_declarator] = "block_pointer_declarator",
  [sym_abstract_block_pointer_declarator] = "abstract_block_pointer_declarator",
  [sym_generic_specifier] = "generic_specifier",
  [sym_typedefed_specifier] = "typedefed_specifier",
  [sym_array_type_specifier] = "array_type_specifier",
  [sym_atdef_field] = "atdef_field",
  [sym_protocol_qualifier] = "protocol_qualifier",
  [sym_type_name] = "type_name",
  [sym_at_expression] = "at_expression",
  [sym_dictionary_literal] = "dictionary_literal",
  [sym_dictionary_pair] = "dictionary_pair",
  [sym_array_literal] = "array_literal",
  [sym_method_identifier] = "method_identifier",
  [aux_sym_translation_unit_repeat1] = "translation_unit_repeat1",
  [aux_sym_preproc_params_repeat1] = "preproc_params_repeat1",
  [aux_sym_preproc_if_repeat1] = "preproc_if_repeat1",
  [aux_sym_preproc_if_in_field_declaration_list_repeat1] = "preproc_if_in_field_declaration_list_repeat1",
  [aux_sym_preproc_argument_list_repeat1] = "preproc_argument_list_repeat1",
  [aux_sym__old_style_function_definition_repeat1] = "_old_style_function_definition_repeat1",
  [aux_sym_declaration_repeat1] = "declaration_repeat1",
  [aux_sym__type_definition_type_repeat1] = "_type_definition_type_repeat1",
  [aux_sym__type_definition_declarators_repeat1] = "_type_definition_declarators_repeat1",
  [aux_sym__declaration_specifiers_repeat1] = "_declaration_specifiers_repeat1",
  [aux_sym_attribute_specifier_repeat1] = "attribute_specifier_repeat1",
  [aux_sym_attribute_repeat1] = "attribute_repeat1",
  [aux_sym_attribute_declaration_repeat1] = "attribute_declaration_repeat1",
  [aux_sym_declaration_list_repeat1] = "declaration_list_repeat1",
  [aux_sym_attributed_field_declarator_repeat1] = "attributed_field_declarator_repeat1",
  [aux_sym_pointer_declarator_repeat1] = "pointer_declarator_repeat1",
  [aux_sym_function_declarator_repeat1] = "function_declarator_repeat1",
  [aux_sym_function_field_declarator_repeat1] = "function_field_declarator_repeat1",
  [aux_sym_array_declarator_repeat1] = "array_declarator_repeat1",
  [aux_sym_sized_type_specifier_repeat1] = "sized_type_specifier_repeat1",
  [aux_sym_enumerator_list_repeat1] = "enumerator_list_repeat1",
  [aux_sym_field_declaration_repeat1] = "field_declaration_repeat1",
  [aux_sym_parameter_list_repeat1] = "parameter_list_repeat1",
  [aux_sym__old_style_parameter_list_repeat1] = "_old_style_parameter_list_repeat1",
  [aux_sym_case_statement_repeat1] = "case_statement_repeat1",
  [aux_sym_generic_expression_repeat1] = "generic_expression_repeat1",
  [aux_sym_gnu_asm_expression_repeat1] = "gnu_asm_expression_repeat1",
  [aux_sym_gnu_asm_output_operand_list_repeat1] = "gnu_asm_output_operand_list_repeat1",
  [aux_sym_gnu_asm_input_operand_list_repeat1] = "gnu_asm_input_operand_list_repeat1",
  [aux_sym_gnu_asm_clobber_list_repeat1] = "gnu_asm_clobber_list_repeat1",
  [aux_sym_gnu_asm_goto_list_repeat1] = "gnu_asm_goto_list_repeat1",
  [aux_sym_argument_list_repeat1] = "argument_list_repeat1",
  [aux_sym_argument_list_repeat2] = "argument_list_repeat2",
  [aux_sym_initializer_list_repeat1] = "initializer_list_repeat1",
  [aux_sym_initializer_pair_repeat1] = "initializer_pair_repeat1",
  [aux_sym_char_literal_repeat1] = "char_literal_repeat1",
  [aux_sym_concatenated_string_repeat1] = "concatenated_string_repeat1",
  [aux_sym_string_literal_repeat1] = "string_literal_repeat1",
  [aux_sym_availability_repeat1] = "availability_repeat1",
  [aux_sym_version_repeat1] = "version_repeat1",
  [aux_sym_module_import_repeat1] = "module_import_repeat1",
  [aux_sym_preproc_if_in_implementation_definition_repeat1] = "preproc_if_in_implementation_definition_repeat1",
  [aux_sym_preproc_if_in_interface_declaration_repeat1] = "preproc_if_in_interface_declaration_repeat1",
  [aux_sym_preproc_ifdef_in_enumerator_repeat1] = "preproc_ifdef_in_enumerator_repeat1",
  [aux_sym_availability_attribute_specifier_repeat1] = "availability_attribute_specifier_repeat1",
  [aux_sym_protocol_forward_declaration_repeat1] = "protocol_forward_declaration_repeat1",
  [aux_sym_class_declaration_repeat1] = "class_declaration_repeat1",
  [aux_sym_class_interface_repeat1] = "class_interface_repeat1",
  [aux_sym_class_implementation_repeat1] = "class_implementation_repeat1",
  [aux_sym_parameterized_arguments_repeat1] = "parameterized_arguments_repeat1",
  [aux_sym_parameterized_arguments_repeat2] = "parameterized_arguments_repeat2",
  [aux_sym_generic_arguments_repeat1] = "generic_arguments_repeat1",
  [aux_sym_instance_variables_repeat1] = "instance_variables_repeat1",
  [aux_sym_protocol_declaration_repeat1] = "protocol_declaration_repeat1",
  [aux_sym_property_implementation_repeat1] = "property_implementation_repeat1",
  [aux_sym_method_definition_repeat1] = "method_definition_repeat1",
  [aux_sym_method_definition_repeat2] = "method_definition_repeat2",
  [aux_sym_method_type_repeat1] = "method_type_repeat1",
  [aux_sym_keyword_selector_repeat1] = "keyword_selector_repeat1",
  [aux_sym_property_attributes_declaration_repeat1] = "property_attributes_declaration_repeat1",
  [aux_sym_method_declaration_repeat1] = "method_declaration_repeat1",
  [aux_sym_method_declaration_repeat2] = "method_declaration_repeat2",
  [aux_sym_c_method_parameter_repeat1] = "c_method_parameter_repeat1",
  [aux_sym_struct_declaration_repeat1] = "struct_declaration_repeat1",
  [aux_sym_struct_declaration_repeat2] = "struct_declaration_repeat2",
  [aux_sym_try_statement_repeat1] = "try_statement_repeat1",
  [aux_sym_selector_expression_repeat1] = "selector_expression_repeat1",
  [aux_sym_selector_expression_repeat2] = "selector_expression_repeat2",
  [aux_sym_available_expression_repeat1] = "available_expression_repeat1",
  [aux_sym_message_expression_repeat1] = "message_expression_repeat1",
  [aux_sym_message_expression_repeat2] = "message_expression_repeat2",
  [aux_sym_generic_specifier_repeat1] = "generic_specifier_repeat1",
  [aux_sym_type_name_repeat1] = "type_name_repeat1",
  [aux_sym_dictionary_literal_repeat1] = "dictionary_literal_repeat1",
  [aux_sym_method_identifier_repeat1] = "method_identifier_repeat1",
  [aux_sym_method_identifier_repeat2] = "method_identifier_repeat2",
  [alias_sym_field_identifier] = "field_identifier",
  [alias_sym_statement_identifier] = "statement_identifier",
  [alias_sym_type_identifier] = "type_identifier",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [aux_sym_preproc_include_token1] = aux_sym_preproc_include_token1,
  [aux_sym_preproc_include_token2] = aux_sym_preproc_include_token2,
  [aux_sym_preproc_include_token3] = aux_sym_preproc_include_token3,
  [aux_sym_preproc_def_token1] = aux_sym_preproc_def_token1,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_DOT_DOT_DOT] = anon_sym_DOT_DOT_DOT,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [aux_sym_preproc_if_token1] = aux_sym_preproc_if_token1,
  [anon_sym_LF] = anon_sym_LF,
  [aux_sym_preproc_if_token2] = aux_sym_preproc_if_token2,
  [aux_sym_preproc_ifdef_token1] = aux_sym_preproc_ifdef_token1,
  [aux_sym_preproc_ifdef_token2] = aux_sym_preproc_ifdef_token2,
  [aux_sym_preproc_else_token1] = aux_sym_preproc_else_token1,
  [aux_sym_preproc_elif_token1] = aux_sym_preproc_elif_token1,
  [aux_sym_preproc_elifdef_token1] = aux_sym_preproc_elifdef_token1,
  [aux_sym_preproc_elifdef_token2] = aux_sym_preproc_elifdef_token2,
  [sym_preproc_arg] = sym_preproc_arg,
  [sym_preproc_directive] = sym_preproc_directive,
  [anon_sym_LPAREN2] = anon_sym_LPAREN,
  [anon_sym_defined] = anon_sym_defined,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym___extension__] = anon_sym___extension__,
  [anon_sym_typedef] = anon_sym_typedef,
  [anon_sym_extern] = anon_sym_extern,
  [anon_sym___attribute__] = anon_sym___attribute__,
  [anon_sym___attribute] = anon_sym___attribute,
  [anon_sym_noreturn] = anon_sym_noreturn,
  [anon_sym_nothrow] = anon_sym_nothrow,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym___declspec] = anon_sym___declspec,
  [anon_sym___based] = anon_sym___based,
  [anon_sym___cdecl] = anon_sym___cdecl,
  [anon_sym___clrcall] = anon_sym___clrcall,
  [anon_sym___stdcall] = anon_sym___stdcall,
  [anon_sym___fastcall] = anon_sym___fastcall,
  [anon_sym___thiscall] = anon_sym___thiscall,
  [anon_sym___vectorcall] = anon_sym___vectorcall,
  [sym_ms_restrict_modifier] = sym_ms_restrict_modifier,
  [sym_ms_unsigned_ptr_modifier] = sym_ms_unsigned_ptr_modifier,
  [sym_ms_signed_ptr_modifier] = sym_ms_signed_ptr_modifier,
  [anon_sym__unaligned] = anon_sym__unaligned,
  [anon_sym___unaligned] = anon_sym___unaligned,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_signed] = anon_sym_signed,
  [anon_sym_unsigned] = anon_sym_unsigned,
  [anon_sym_long] = anon_sym_long,
  [anon_sym_short] = anon_sym_short,
  [anon_sym_static] = anon_sym_static,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_ATautoreleasepool] = anon_sym_ATautoreleasepool,
  [anon_sym_auto] = anon_sym_auto,
  [anon_sym_register] = anon_sym_register,
  [anon_sym_inline] = anon_sym_inline,
  [anon_sym___inline] = anon_sym___inline,
  [anon_sym___inline__] = anon_sym___inline__,
  [anon_sym___forceinline] = anon_sym___forceinline,
  [anon_sym_thread_local] = anon_sym_thread_local,
  [anon_sym___thread] = anon_sym___thread,
  [anon_sym_CG_EXTERN] = anon_sym_CG_EXTERN,
  [anon_sym_CG_INLINE] = anon_sym_CG_INLINE,
  [anon_sym_FOUNDATION_EXPORT] = anon_sym_FOUNDATION_EXPORT,
  [anon_sym_FOUNDATION_EXTERN] = anon_sym_FOUNDATION_EXTERN,
  [anon_sym_FOUNDATION_STATIC_INLINE] = anon_sym_FOUNDATION_STATIC_INLINE,
  [anon_sym_IBOutlet] = anon_sym_IBOutlet,
  [anon_sym_IBInspectable] = anon_sym_IBInspectable,
  [anon_sym_IB_DESIGNABLE] = anon_sym_IB_DESIGNABLE,
  [anon_sym_NS_INLINE] = anon_sym_NS_INLINE,
  [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = anon_sym_NS_VALID_UNTIL_END_OF_SCOPE,
  [anon_sym_OBJC_EXPORT] = anon_sym_OBJC_EXPORT,
  [anon_sym_OBJC_ROOT_CLASS] = anon_sym_OBJC_ROOT_CLASS,
  [anon_sym_UIKIT_EXTERN] = anon_sym_UIKIT_EXTERN,
  [anon_sym_const] = anon_sym_const,
  [anon_sym_constexpr] = anon_sym_constexpr,
  [anon_sym_volatile] = anon_sym_volatile,
  [anon_sym_restrict] = anon_sym_restrict,
  [anon_sym___restrict__] = anon_sym___restrict__,
  [anon_sym__Atomic] = anon_sym__Atomic,
  [anon_sym__Noreturn] = anon_sym__Noreturn,
  [anon_sym__Nonnull] = anon_sym__Nonnull,
  [anon_sym_nullable] = anon_sym_nullable,
  [anon_sym__Complex] = anon_sym__Complex,
  [anon_sym__Nullable] = anon_sym__Nullable,
  [anon_sym__Nullable_result] = anon_sym__Nullable_result,
  [anon_sym__Null_unspecified] = anon_sym__Null_unspecified,
  [anon_sym___autoreleasing] = anon_sym___autoreleasing,
  [anon_sym___block] = anon_sym___block,
  [anon_sym___bridge] = anon_sym___bridge,
  [anon_sym___bridge_retained] = anon_sym___bridge_retained,
  [anon_sym___bridge_transfer] = anon_sym___bridge_transfer,
  [anon_sym___complex] = anon_sym___complex,
  [anon_sym___const] = anon_sym___const,
  [anon_sym___imag] = anon_sym___imag,
  [anon_sym___kindof] = anon_sym___kindof,
  [anon_sym___nonnull] = anon_sym___nonnull,
  [anon_sym___nullable] = anon_sym___nullable,
  [anon_sym___ptrauth_objc_class_ro] = anon_sym___ptrauth_objc_class_ro,
  [anon_sym___ptrauth_objc_isa_pointer] = anon_sym___ptrauth_objc_isa_pointer,
  [anon_sym___ptrauth_objc_super_pointer] = anon_sym___ptrauth_objc_super_pointer,
  [anon_sym___real] = anon_sym___real,
  [anon_sym___strong] = anon_sym___strong,
  [anon_sym___unsafe_unretained] = anon_sym___unsafe_unretained,
  [anon_sym___unused] = anon_sym___unused,
  [anon_sym___weak] = anon_sym___weak,
  [anon_sym_alignas] = anon_sym_alignas,
  [anon_sym__Alignas] = anon_sym__Alignas,
  [sym_primitive_type] = sym_primitive_type,
  [anon_sym_enum] = anon_sym_enum,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_struct] = anon_sym_struct,
  [anon_sym_union] = anon_sym_union,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_switch] = anon_sym_switch,
  [anon_sym_case] = anon_sym_case,
  [anon_sym_default] = anon_sym_default,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_return] = anon_sym_return,
  [anon_sym_break] = anon_sym_break,
  [anon_sym_continue] = anon_sym_continue,
  [anon_sym_goto] = anon_sym_goto,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_sizeof] = anon_sym_sizeof,
  [anon_sym___alignof__] = anon_sym___alignof__,
  [anon_sym___alignof] = anon_sym___alignof,
  [anon_sym__alignof] = anon_sym__alignof,
  [anon_sym_alignof] = anon_sym_alignof,
  [anon_sym__Alignof] = anon_sym__Alignof,
  [anon_sym_offsetof] = anon_sym_offsetof,
  [anon_sym__Generic] = anon_sym__Generic,
  [anon_sym_asm] = anon_sym_asm,
  [anon_sym___asm__] = anon_sym___asm__,
  [anon_sym___asm] = anon_sym___asm,
  [anon_sym___volatile__] = anon_sym___volatile__,
  [anon_sym_LT2] = anon_sym_LT,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [sym_number_literal] = sym_number_literal,
  [anon_sym_L_SQUOTE] = anon_sym_L_SQUOTE,
  [anon_sym_u_SQUOTE] = anon_sym_u_SQUOTE,
  [anon_sym_U_SQUOTE] = anon_sym_U_SQUOTE,
  [anon_sym_u8_SQUOTE] = anon_sym_u8_SQUOTE,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [aux_sym_char_literal_token1] = aux_sym_char_literal_token1,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [anon_sym_L_DQUOTE] = anon_sym_L_DQUOTE,
  [anon_sym_u_DQUOTE] = anon_sym_u_DQUOTE,
  [anon_sym_U_DQUOTE] = anon_sym_U_DQUOTE,
  [anon_sym_u8_DQUOTE] = anon_sym_u8_DQUOTE,
  [aux_sym_string_literal_token1] = aux_sym_string_literal_token1,
  [sym_escape_sequence] = sym_escape_sequence,
  [sym_system_lib_string] = sym_system_lib_string,
  [sym_true] = sym_true,
  [sym_false] = sym_false,
  [anon_sym_NULL] = anon_sym_NULL,
  [anon_sym_nullptr] = anon_sym_nullptr,
  [sym_comment] = sym_comment,
  [anon_sym_objc_bridge_related] = anon_sym_objc_bridge_related,
  [anon_sym___typeof__] = anon_sym___typeof__,
  [anon_sym___typeof] = anon_sym___typeof,
  [anon_sym_typeof] = anon_sym_typeof,
  [anon_sym_availability] = anon_sym_availability,
  [sym_version_number] = sym_version_number,
  [anon_sym_ios] = anon_sym_ios,
  [anon_sym_tvos] = anon_sym_tvos,
  [anon_sym_macos] = anon_sym_macos,
  [anon_sym_macosx] = anon_sym_macosx,
  [anon_sym_watchos] = anon_sym_watchos,
  [anon_sym_ATimport] = anon_sym_ATimport,
  [aux_sym_preproc_undef_token1] = aux_sym_preproc_undef_token1,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE,
  [anon_sym_NS_ROOT_CLASS] = anon_sym_NS_ROOT_CLASS,
  [anon_sym_NS_UNAVAILABLE] = anon_sym_NS_UNAVAILABLE,
  [anon_sym_NS_REQUIRES_NIL_TERMINATION] = anon_sym_NS_REQUIRES_NIL_TERMINATION,
  [anon_sym_CF_RETURNS_RETAINED] = anon_sym_CF_RETURNS_RETAINED,
  [anon_sym_CF_RETURNS_NOT_RETAINED] = anon_sym_CF_RETURNS_NOT_RETAINED,
  [anon_sym_DEPRECATED_ATTRIBUTE] = anon_sym_DEPRECATED_ATTRIBUTE,
  [anon_sym_UI_APPEARANCE_SELECTOR] = anon_sym_UI_APPEARANCE_SELECTOR,
  [anon_sym_UNAVAILABLE_ATTRIBUTE] = anon_sym_UNAVAILABLE_ATTRIBUTE,
  [anon_sym_CF_FORMAT_FUNCTION] = anon_sym_CF_FORMAT_FUNCTION,
  [anon_sym_NS_AVAILABLE] = anon_sym_NS_AVAILABLE,
  [anon_sym___IOS_AVAILABLE] = anon_sym___IOS_AVAILABLE,
  [anon_sym_NS_AVAILABLE_IOS] = anon_sym_NS_AVAILABLE_IOS,
  [anon_sym_API_AVAILABLE] = anon_sym_API_AVAILABLE,
  [anon_sym_API_UNAVAILABLE] = anon_sym_API_UNAVAILABLE,
  [anon_sym_API_DEPRECATED] = anon_sym_API_DEPRECATED,
  [anon_sym_NS_ENUM_AVAILABLE_IOS] = anon_sym_NS_ENUM_AVAILABLE_IOS,
  [anon_sym_NS_DEPRECATED_IOS] = anon_sym_NS_DEPRECATED_IOS,
  [anon_sym_NS_ENUM_DEPRECATED_IOS] = anon_sym_NS_ENUM_DEPRECATED_IOS,
  [anon_sym_NS_FORMAT_FUNCTION] = anon_sym_NS_FORMAT_FUNCTION,
  [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = anon_sym_DEPRECATED_MSG_ATTRIBUTE,
  [anon_sym___deprecated_msg] = anon_sym___deprecated_msg,
  [anon_sym___deprecated_enum_msg] = anon_sym___deprecated_enum_msg,
  [anon_sym_NS_SWIFT_NAME] = anon_sym_NS_SWIFT_NAME,
  [anon_sym_NS_SWIFT_UNAVAILABLE] = anon_sym_NS_SWIFT_UNAVAILABLE,
  [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = anon_sym_NS_EXTENSION_UNAVAILABLE_IOS,
  [anon_sym_NS_CLASS_AVAILABLE_IOS] = anon_sym_NS_CLASS_AVAILABLE_IOS,
  [anon_sym_NS_CLASS_DEPRECATED_IOS] = anon_sym_NS_CLASS_DEPRECATED_IOS,
  [anon_sym___OSX_AVAILABLE_STARTING] = anon_sym___OSX_AVAILABLE_STARTING,
  [anon_sym_ATprotocol] = anon_sym_ATprotocol,
  [anon_sym_class] = anon_sym_class,
  [anon_sym_ATend] = anon_sym_ATend,
  [anon_sym_ATinterface] = anon_sym_ATinterface,
  [anon_sym_ATimplementation] = anon_sym_ATimplementation,
  [anon_sym___covariant] = anon_sym___covariant,
  [anon_sym___contravariant] = anon_sym___contravariant,
  [anon_sym_ATprivate] = anon_sym_ATprivate,
  [anon_sym_ATprotected] = anon_sym_ATprotected,
  [anon_sym_ATpackage] = anon_sym_ATpackage,
  [anon_sym_ATpublic] = anon_sym_ATpublic,
  [anon_sym_ATcompatibility_alias] = anon_sym_ATcompatibility_alias,
  [anon_sym_AToptional] = anon_sym_AToptional,
  [anon_sym_ATrequired] = anon_sym_ATrequired,
  [anon_sym_ATsynthesize] = anon_sym_ATsynthesize,
  [anon_sym_ATdynamic] = anon_sym_ATdynamic,
  [anon_sym_LPARENclass_RPAREN] = anon_sym_LPARENclass_RPAREN,
  [anon_sym_ATproperty] = anon_sym_ATproperty,
  [anon_sym_ATtry] = anon_sym_ATtry,
  [anon_sym___try] = anon_sym___try,
  [anon_sym_ATcatch] = anon_sym_ATcatch,
  [anon_sym___catch] = anon_sym___catch,
  [anon_sym_ATfinally] = anon_sym_ATfinally,
  [anon_sym___finally] = anon_sym___finally,
  [anon_sym_ATthrow] = anon_sym_ATthrow,
  [anon_sym_ATselector] = anon_sym_ATselector,
  [aux_sym_selector_expression_token1] = aux_sym_selector_expression_token1,
  [anon_sym_ATavailable] = anon_sym_ATavailable,
  [anon_sym___builtin_available] = anon_sym___builtin_available,
  [anon_sym_va_arg] = anon_sym_va_arg,
  [aux_sym_ms_asm_block_token1] = aux_sym_ms_asm_block_token1,
  [anon_sym_ATencode] = anon_sym_ATencode,
  [anon_sym_ATsynchronized] = anon_sym_ATsynchronized,
  [anon_sym_BOOL] = anon_sym_BOOL,
  [anon_sym_IMP] = anon_sym_IMP,
  [anon_sym_SEL] = anon_sym_SEL,
  [anon_sym_Class] = anon_sym_Class,
  [anon_sym_id] = anon_sym_id,
  [anon_sym_ATdefs] = anon_sym_ATdefs,
  [anon_sym_out] = anon_sym_out,
  [anon_sym_inout] = anon_sym_inout,
  [anon_sym_bycopy] = anon_sym_bycopy,
  [anon_sym_byref] = anon_sym_byref,
  [anon_sym_oneway] = anon_sym_oneway,
  [anon_sym_COLON2] = anon_sym_COLON,
  [sym_translation_unit] = sym_translation_unit,
  [sym__top_level_item] = sym__top_level_item,
  [sym__block_item] = sym__block_item,
  [sym_preproc_include] = sym_preproc_include,
  [sym_preproc_def] = sym_preproc_def,
  [sym_preproc_function_def] = sym_preproc_function_def,
  [sym_preproc_params] = sym_preproc_params,
  [sym_preproc_call] = sym_preproc_call,
  [sym_preproc_if] = sym_preproc_if,
  [sym_preproc_ifdef] = sym_preproc_ifdef,
  [sym_preproc_else] = sym_preproc_else,
  [sym_preproc_elif] = sym_preproc_elif,
  [sym_preproc_elifdef] = sym_preproc_elifdef,
  [sym_preproc_if_in_field_declaration_list] = sym_preproc_if,
  [sym_preproc_ifdef_in_field_declaration_list] = sym_preproc_ifdef,
  [sym_preproc_else_in_field_declaration_list] = sym_preproc_else,
  [sym_preproc_elif_in_field_declaration_list] = sym_preproc_elif,
  [sym_preproc_elifdef_in_field_declaration_list] = sym_preproc_elifdef,
  [sym__preproc_expression] = sym__preproc_expression,
  [sym_preproc_parenthesized_expression] = sym_parenthesized_expression,
  [sym_preproc_defined] = sym_preproc_defined,
  [sym_preproc_unary_expression] = sym_unary_expression,
  [sym_preproc_call_expression] = sym_call_expression,
  [sym_preproc_argument_list] = sym_argument_list,
  [sym_preproc_binary_expression] = sym_binary_expression,
  [sym_function_definition] = sym_function_definition,
  [sym__old_style_function_definition] = sym_function_definition,
  [sym_declaration] = sym_declaration,
  [sym_type_definition] = sym_type_definition,
  [sym__type_definition_type] = sym__type_definition_type,
  [sym__type_definition_declarators] = sym__type_definition_declarators,
  [sym__declaration_modifiers] = sym__declaration_modifiers,
  [sym__declaration_specifiers] = sym__declaration_specifiers,
  [sym_linkage_specification] = sym_linkage_specification,
  [sym_attribute_specifier] = sym_attribute_specifier,
  [sym_attribute] = sym_attribute,
  [sym_attribute_declaration] = sym_attribute_declaration,
  [sym_ms_declspec_modifier] = sym_ms_declspec_modifier,
  [sym_ms_based_modifier] = sym_ms_based_modifier,
  [sym_ms_call_modifier] = sym_ms_call_modifier,
  [sym_ms_unaligned_ptr_modifier] = sym_ms_unaligned_ptr_modifier,
  [sym_ms_pointer_modifier] = sym_ms_pointer_modifier,
  [sym_declaration_list] = sym_declaration_list,
  [sym__declarator] = sym__declarator,
  [sym__field_declarator] = sym__field_declarator,
  [sym__type_declarator] = sym__type_declarator,
  [sym__abstract_declarator] = sym__abstract_declarator,
  [sym_parenthesized_declarator] = sym_parenthesized_declarator,
  [sym_parenthesized_field_declarator] = sym_parenthesized_declarator,
  [sym_parenthesized_type_declarator] = sym_parenthesized_declarator,
  [sym_abstract_parenthesized_declarator] = sym_abstract_parenthesized_declarator,
  [sym_attributed_field_declarator] = sym_attributed_field_declarator,
  [sym_pointer_declarator] = sym_pointer_declarator,
  [sym_pointer_field_declarator] = sym_pointer_declarator,
  [sym_pointer_type_declarator] = sym_pointer_declarator,
  [sym_abstract_pointer_declarator] = sym_abstract_pointer_declarator,
  [sym_function_declarator] = sym_function_declarator,
  [sym_function_field_declarator] = sym_function_declarator,
  [sym_function_type_declarator] = sym_function_declarator,
  [sym_abstract_function_declarator] = sym_abstract_function_declarator,
  [sym__old_style_function_declarator] = sym_function_declarator,
  [sym_array_declarator] = sym_array_declarator,
  [sym_array_field_declarator] = sym_array_declarator,
  [sym_array_type_declarator] = sym_array_declarator,
  [sym_abstract_array_declarator] = sym_abstract_array_declarator,
  [sym_init_declarator] = sym_init_declarator,
  [sym_compound_statement] = sym_compound_statement,
  [sym_storage_class_specifier] = sym_storage_class_specifier,
  [sym_type_qualifier] = sym_type_qualifier,
  [sym_alignas_qualifier] = sym_alignas_qualifier,
  [sym_type_specifier] = sym_type_specifier,
  [sym_sized_type_specifier] = sym_sized_type_specifier,
  [sym_enum_specifier] = sym_enum_specifier,
  [sym_enumerator_list] = sym_enumerator_list,
  [sym_struct_specifier] = sym_struct_specifier,
  [sym_union_specifier] = sym_union_specifier,
  [sym_field_declaration_list] = sym_field_declaration_list,
  [sym__field_declaration_list_item] = sym__field_declaration_list_item,
  [sym_field_declaration] = sym_field_declaration,
  [sym_bitfield_clause] = sym_bitfield_clause,
  [sym_enumerator] = sym_enumerator,
  [sym_variadic_parameter] = sym_variadic_parameter,
  [sym_parameter_list] = sym_parameter_list,
  [sym__old_style_parameter_list] = sym_parameter_list,
  [sym_parameter_declaration] = sym_parameter_declaration,
  [sym_attributed_statement] = sym_attributed_statement,
  [sym_statement] = sym_statement,
  [sym__top_level_statement] = sym__top_level_statement,
  [sym_labeled_statement] = sym_labeled_statement,
  [sym__top_level_expression_statement] = sym_expression_statement,
  [sym_expression_statement] = sym_expression_statement,
  [sym_if_statement] = sym_if_statement,
  [sym_else_clause] = sym_else_clause,
  [sym_switch_statement] = sym_switch_statement,
  [sym_case_statement] = sym_case_statement,
  [sym_while_statement] = sym_while_statement,
  [sym_do_statement] = sym_do_statement,
  [sym_for_statement] = sym_for_statement,
  [sym__for_statement_body] = sym__for_statement_body,
  [sym_return_statement] = sym_return_statement,
  [sym_break_statement] = sym_break_statement,
  [sym_continue_statement] = sym_continue_statement,
  [sym_goto_statement] = sym_goto_statement,
  [sym_expression] = sym_expression,
  [sym__string] = sym__string,
  [sym_comma_expression] = sym_comma_expression,
  [sym_conditional_expression] = sym_conditional_expression,
  [sym_assignment_expression] = sym_assignment_expression,
  [sym_pointer_expression] = sym_pointer_expression,
  [sym_unary_expression] = sym_unary_expression,
  [sym_binary_expression] = sym_binary_expression,
  [sym_update_expression] = sym_update_expression,
  [sym_cast_expression] = sym_cast_expression,
  [sym_type_descriptor] = sym_type_descriptor,
  [sym_sizeof_expression] = sym_sizeof_expression,
  [sym_alignof_expression] = sym_alignof_expression,
  [sym_offsetof_expression] = sym_offsetof_expression,
  [sym_generic_expression] = sym_generic_expression,
  [sym_subscript_expression] = sym_subscript_expression,
  [sym_call_expression] = sym_call_expression,
  [sym_gnu_asm_expression] = sym_gnu_asm_expression,
  [sym_gnu_asm_qualifier] = sym_gnu_asm_qualifier,
  [sym_gnu_asm_output_operand_list] = sym_gnu_asm_output_operand_list,
  [sym_gnu_asm_output_operand] = sym_gnu_asm_output_operand,
  [sym_gnu_asm_input_operand_list] = sym_gnu_asm_input_operand_list,
  [sym_gnu_asm_input_operand] = sym_gnu_asm_input_operand,
  [sym_gnu_asm_clobber_list] = sym_gnu_asm_clobber_list,
  [sym_gnu_asm_goto_list] = sym_gnu_asm_goto_list,
  [sym_extension_expression] = sym_extension_expression,
  [sym_argument_list] = sym_argument_list,
  [sym_field_expression] = sym_field_expression,
  [sym_compound_literal_expression] = sym_compound_literal_expression,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym_initializer_list] = sym_initializer_list,
  [sym_initializer_pair] = sym_initializer_pair,
  [sym_subscript_designator] = sym_subscript_designator,
  [sym_subscript_range_designator] = sym_subscript_range_designator,
  [sym_field_designator] = sym_field_designator,
  [sym_char_literal] = sym_char_literal,
  [sym_concatenated_string] = sym_concatenated_string,
  [sym_string_literal] = sym_string_literal,
  [sym_null] = sym_null,
  [sym__empty_declaration] = sym__empty_declaration,
  [sym_macro_type_specifier] = sym_macro_type_specifier,
  [sym_objc_bridge] = sym_objc_bridge,
  [sym_typeof_specifier] = sym_typeof_specifier,
  [sym_availability] = sym_availability,
  [sym_version] = sym_version,
  [sym_platform] = sym_platform,
  [sym_module_import] = sym_module_import,
  [sym_preproc_if_in_implementation_definition] = sym_preproc_if,
  [sym_preproc_else_in_implementation_definition] = sym_preproc_else,
  [sym_preproc_elif_in_implementation_definition] = sym_preproc_elif,
  [sym_preproc_if_in_interface_declaration] = sym_preproc_if,
  [sym_preproc_else_in_interface_declaration] = sym_preproc_else,
  [sym_preproc_elif_in_interface_declaration] = sym_preproc_elif,
  [sym_preproc_ifdef_in_enumerator] = sym_preproc_ifdef,
  [sym_preproc_else_in_enumerator] = sym_preproc_else,
  [sym_preproc_elif_in_enumerator] = sym_preproc_elif,
  [sym_preproc_undef] = sym_preproc_undef,
  [sym_preproc_linemarker] = sym_preproc_linemarker,
  [sym_availability_attribute_specifier] = sym_availability_attribute_specifier,
  [sym_protocol_forward_declaration] = sym_protocol_forward_declaration,
  [sym_class_declaration] = sym_class_declaration,
  [sym_class_interface] = sym_class_interface,
  [sym__class_interface_header] = sym__class_interface_header,
  [sym__type_params] = sym__type_params,
  [sym__class_interface_inheritance] = sym__class_interface_inheritance,
  [sym_class_implementation] = sym_class_implementation,
  [sym__class_implementation_header] = sym__class_implementation_header,
  [sym__class_implementation_inheritance] = sym__class_implementation_inheritance,
  [sym_protocol_reference_list] = sym_protocol_reference_list,
  [sym_parameterized_arguments] = sym_parameterized_arguments,
  [sym_generic_arguments] = sym_generic_arguments,
  [sym_instance_variables] = sym_instance_variables,
  [sym_instance_variable] = sym_instance_variable,
  [sym_visibility_specification] = sym_visibility_specification,
  [sym_protocol_declaration] = sym_protocol_declaration,
  [sym_compatibility_alias_declaration] = sym_compatibility_alias_declaration,
  [sym_qualified_protocol_interface_declaration] = sym_qualified_protocol_interface_declaration,
  [sym_implementation_definition] = sym_implementation_definition,
  [sym_property_implementation] = sym_property_implementation,
  [sym_method_definition] = sym_method_definition,
  [sym_method_type] = sym_method_type,
  [sym_keyword_declarator] = sym_keyword_declarator,
  [sym_property_declaration] = sym_property_declaration,
  [sym_property_attributes_declaration] = sym_property_attributes_declaration,
  [sym_property_attribute] = sym_property_attribute,
  [sym_method_declaration] = sym_method_declaration,
  [sym_method_parameter] = sym_method_parameter,
  [sym_c_method_parameter] = sym_method_parameter,
  [sym_struct_declaration] = sym_struct_declaration,
  [sym_atomic_declaration] = sym_atomic_declaration,
  [sym_specifier_qualifier] = sym_specifier_qualifier,
  [sym_struct_declarator] = sym_struct_declarator,
  [sym_try_statement] = sym_try_statement,
  [sym_catch_clause] = sym_catch_clause,
  [sym_finally_clause] = sym_finally_clause,
  [sym_throw_statement] = sym_throw_statement,
  [sym_selector_expression] = sym_selector_expression,
  [sym_available_expression] = sym_available_expression,
  [sym_range_expression] = sym_range_expression,
  [sym_block_literal] = sym_block_literal,
  [sym_message_expression] = sym_message_expression,
  [sym_va_arg_expression] = sym_va_arg_expression,
  [sym_ms_asm_block] = sym_ms_asm_block,
  [sym_encode_expression] = sym_encode_expression,
  [sym_synchronized_statement] = sym_synchronized_statement,
  [sym_block_pointer_declarator] = sym_block_pointer_declarator,
  [sym_block_pointer_field_declarator] = sym_block_pointer_declarator,
  [sym_block_pointer_type_declarator] = sym_block_pointer_declarator,
  [sym_abstract_block_pointer_declarator] = sym_abstract_block_pointer_declarator,
  [sym_generic_specifier] = sym_generic_specifier,
  [sym_typedefed_specifier] = sym_typedefed_specifier,
  [sym_array_type_specifier] = sym_array_type_specifier,
  [sym_atdef_field] = sym_atdef_field,
  [sym_protocol_qualifier] = sym_protocol_qualifier,
  [sym_type_name] = sym_type_name,
  [sym_at_expression] = sym_at_expression,
  [sym_dictionary_literal] = sym_dictionary_literal,
  [sym_dictionary_pair] = sym_dictionary_pair,
  [sym_array_literal] = sym_array_literal,
  [sym_method_identifier] = sym_method_identifier,
  [aux_sym_translation_unit_repeat1] = aux_sym_translation_unit_repeat1,
  [aux_sym_preproc_params_repeat1] = aux_sym_preproc_params_repeat1,
  [aux_sym_preproc_if_repeat1] = aux_sym_preproc_if_repeat1,
  [aux_sym_preproc_if_in_field_declaration_list_repeat1] = aux_sym_preproc_if_in_field_declaration_list_repeat1,
  [aux_sym_preproc_argument_list_repeat1] = aux_sym_preproc_argument_list_repeat1,
  [aux_sym__old_style_function_definition_repeat1] = aux_sym__old_style_function_definition_repeat1,
  [aux_sym_declaration_repeat1] = aux_sym_declaration_repeat1,
  [aux_sym__type_definition_type_repeat1] = aux_sym__type_definition_type_repeat1,
  [aux_sym__type_definition_declarators_repeat1] = aux_sym__type_definition_declarators_repeat1,
  [aux_sym__declaration_specifiers_repeat1] = aux_sym__declaration_specifiers_repeat1,
  [aux_sym_attribute_specifier_repeat1] = aux_sym_attribute_specifier_repeat1,
  [aux_sym_attribute_repeat1] = aux_sym_attribute_repeat1,
  [aux_sym_attribute_declaration_repeat1] = aux_sym_attribute_declaration_repeat1,
  [aux_sym_declaration_list_repeat1] = aux_sym_declaration_list_repeat1,
  [aux_sym_attributed_field_declarator_repeat1] = aux_sym_attributed_field_declarator_repeat1,
  [aux_sym_pointer_declarator_repeat1] = aux_sym_pointer_declarator_repeat1,
  [aux_sym_function_declarator_repeat1] = aux_sym_function_declarator_repeat1,
  [aux_sym_function_field_declarator_repeat1] = aux_sym_function_field_declarator_repeat1,
  [aux_sym_array_declarator_repeat1] = aux_sym_array_declarator_repeat1,
  [aux_sym_sized_type_specifier_repeat1] = aux_sym_sized_type_specifier_repeat1,
  [aux_sym_enumerator_list_repeat1] = aux_sym_enumerator_list_repeat1,
  [aux_sym_field_declaration_repeat1] = aux_sym_field_declaration_repeat1,
  [aux_sym_parameter_list_repeat1] = aux_sym_parameter_list_repeat1,
  [aux_sym__old_style_parameter_list_repeat1] = aux_sym__old_style_parameter_list_repeat1,
  [aux_sym_case_statement_repeat1] = aux_sym_case_statement_repeat1,
  [aux_sym_generic_expression_repeat1] = aux_sym_generic_expression_repeat1,
  [aux_sym_gnu_asm_expression_repeat1] = aux_sym_gnu_asm_expression_repeat1,
  [aux_sym_gnu_asm_output_operand_list_repeat1] = aux_sym_gnu_asm_output_operand_list_repeat1,
  [aux_sym_gnu_asm_input_operand_list_repeat1] = aux_sym_gnu_asm_input_operand_list_repeat1,
  [aux_sym_gnu_asm_clobber_list_repeat1] = aux_sym_gnu_asm_clobber_list_repeat1,
  [aux_sym_gnu_asm_goto_list_repeat1] = aux_sym_gnu_asm_goto_list_repeat1,
  [aux_sym_argument_list_repeat1] = aux_sym_argument_list_repeat1,
  [aux_sym_argument_list_repeat2] = aux_sym_argument_list_repeat2,
  [aux_sym_initializer_list_repeat1] = aux_sym_initializer_list_repeat1,
  [aux_sym_initializer_pair_repeat1] = aux_sym_initializer_pair_repeat1,
  [aux_sym_char_literal_repeat1] = aux_sym_char_literal_repeat1,
  [aux_sym_concatenated_string_repeat1] = aux_sym_concatenated_string_repeat1,
  [aux_sym_string_literal_repeat1] = aux_sym_string_literal_repeat1,
  [aux_sym_availability_repeat1] = aux_sym_availability_repeat1,
  [aux_sym_version_repeat1] = aux_sym_version_repeat1,
  [aux_sym_module_import_repeat1] = aux_sym_module_import_repeat1,
  [aux_sym_preproc_if_in_implementation_definition_repeat1] = aux_sym_preproc_if_in_implementation_definition_repeat1,
  [aux_sym_preproc_if_in_interface_declaration_repeat1] = aux_sym_preproc_if_in_interface_declaration_repeat1,
  [aux_sym_preproc_ifdef_in_enumerator_repeat1] = aux_sym_preproc_ifdef_in_enumerator_repeat1,
  [aux_sym_availability_attribute_specifier_repeat1] = aux_sym_availability_attribute_specifier_repeat1,
  [aux_sym_protocol_forward_declaration_repeat1] = aux_sym_protocol_forward_declaration_repeat1,
  [aux_sym_class_declaration_repeat1] = aux_sym_class_declaration_repeat1,
  [aux_sym_class_interface_repeat1] = aux_sym_class_interface_repeat1,
  [aux_sym_class_implementation_repeat1] = aux_sym_class_implementation_repeat1,
  [aux_sym_parameterized_arguments_repeat1] = aux_sym_parameterized_arguments_repeat1,
  [aux_sym_parameterized_arguments_repeat2] = aux_sym_parameterized_arguments_repeat2,
  [aux_sym_generic_arguments_repeat1] = aux_sym_generic_arguments_repeat1,
  [aux_sym_instance_variables_repeat1] = aux_sym_instance_variables_repeat1,
  [aux_sym_protocol_declaration_repeat1] = aux_sym_protocol_declaration_repeat1,
  [aux_sym_property_implementation_repeat1] = aux_sym_property_implementation_repeat1,
  [aux_sym_method_definition_repeat1] = aux_sym_method_definition_repeat1,
  [aux_sym_method_definition_repeat2] = aux_sym_method_definition_repeat2,
  [aux_sym_method_type_repeat1] = aux_sym_method_type_repeat1,
  [aux_sym_keyword_selector_repeat1] = aux_sym_keyword_selector_repeat1,
  [aux_sym_property_attributes_declaration_repeat1] = aux_sym_property_attributes_declaration_repeat1,
  [aux_sym_method_declaration_repeat1] = aux_sym_method_declaration_repeat1,
  [aux_sym_method_declaration_repeat2] = aux_sym_method_declaration_repeat2,
  [aux_sym_c_method_parameter_repeat1] = aux_sym_c_method_parameter_repeat1,
  [aux_sym_struct_declaration_repeat1] = aux_sym_struct_declaration_repeat1,
  [aux_sym_struct_declaration_repeat2] = aux_sym_struct_declaration_repeat2,
  [aux_sym_try_statement_repeat1] = aux_sym_try_statement_repeat1,
  [aux_sym_selector_expression_repeat1] = aux_sym_selector_expression_repeat1,
  [aux_sym_selector_expression_repeat2] = aux_sym_selector_expression_repeat2,
  [aux_sym_available_expression_repeat1] = aux_sym_available_expression_repeat1,
  [aux_sym_message_expression_repeat1] = aux_sym_message_expression_repeat1,
  [aux_sym_message_expression_repeat2] = aux_sym_message_expression_repeat2,
  [aux_sym_generic_specifier_repeat1] = aux_sym_generic_specifier_repeat1,
  [aux_sym_type_name_repeat1] = aux_sym_type_name_repeat1,
  [aux_sym_dictionary_literal_repeat1] = aux_sym_dictionary_literal_repeat1,
  [aux_sym_method_identifier_repeat1] = aux_sym_method_identifier_repeat1,
  [aux_sym_method_identifier_repeat2] = aux_sym_method_identifier_repeat2,
  [alias_sym_field_identifier] = alias_sym_field_identifier,
  [alias_sym_statement_identifier] = alias_sym_statement_identifier,
  [alias_sym_type_identifier] = alias_sym_type_identifier,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_preproc_include_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_include_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_include_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_def_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_if_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LF] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_if_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_ifdef_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_ifdef_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_else_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_elif_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_elifdef_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_elifdef_token2] = {
    .visible = true,
    .named = false,
  },
  [sym_preproc_arg] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_directive] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LPAREN2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_defined] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___extension__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_typedef] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_extern] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___attribute__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___attribute] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_noreturn] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nothrow] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___declspec] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___based] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___cdecl] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___clrcall] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___stdcall] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___fastcall] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___thiscall] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___vectorcall] = {
    .visible = true,
    .named = false,
  },
  [sym_ms_restrict_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_unsigned_ptr_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_signed_ptr_modifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym__unaligned] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___unaligned] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_signed] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsigned] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_long] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_short] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_static] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATautoreleasepool] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_auto] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_register] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_inline] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___inline] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___inline__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___forceinline] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_thread_local] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___thread] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CG_EXTERN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CG_INLINE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_FOUNDATION_EXPORT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_FOUNDATION_EXTERN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_FOUNDATION_STATIC_INLINE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_IBOutlet] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_IBInspectable] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_IB_DESIGNABLE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_INLINE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_OBJC_EXPORT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_OBJC_ROOT_CLASS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_UIKIT_EXTERN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_const] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_constexpr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_volatile] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_restrict] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___restrict__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Atomic] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Noreturn] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Nonnull] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nullable] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Complex] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Nullable] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Nullable_result] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Null_unspecified] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___autoreleasing] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___block] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___bridge] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___bridge_retained] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___bridge_transfer] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___complex] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___const] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___imag] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___kindof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___nonnull] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___nullable] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___ptrauth_objc_class_ro] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___ptrauth_objc_isa_pointer] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___ptrauth_objc_super_pointer] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___real] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___strong] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___unsafe_unretained] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___unused] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___weak] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_alignas] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Alignas] = {
    .visible = true,
    .named = false,
  },
  [sym_primitive_type] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_enum] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_struct] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_union] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_switch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_case] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_default] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_return] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_break] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_continue] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_goto] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sizeof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___alignof__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___alignof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__alignof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_alignof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Alignof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_offsetof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Generic] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_asm] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___asm__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___asm] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___volatile__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [sym_number_literal] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_L_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_U_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u8_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_char_literal_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_L_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_U_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u8_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_literal_token1] = {
    .visible = true,
    .named = true,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_system_lib_string] = {
    .visible = true,
    .named = true,
  },
  [sym_true] = {
    .visible = true,
    .named = true,
  },
  [sym_false] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_NULL] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nullptr] = {
    .visible = true,
    .named = false,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_objc_bridge_related] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___typeof__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___typeof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_typeof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_availability] = {
    .visible = true,
    .named = false,
  },
  [sym_version_number] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_ios] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_tvos] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_macos] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_macosx] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_watchos] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATimport] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_undef_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_ROOT_CLASS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_UNAVAILABLE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_REQUIRES_NIL_TERMINATION] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CF_RETURNS_RETAINED] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CF_RETURNS_NOT_RETAINED] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DEPRECATED_ATTRIBUTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_UI_APPEARANCE_SELECTOR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_UNAVAILABLE_ATTRIBUTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CF_FORMAT_FUNCTION] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_AVAILABLE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___IOS_AVAILABLE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_AVAILABLE_IOS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_API_AVAILABLE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_API_UNAVAILABLE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_API_DEPRECATED] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_ENUM_AVAILABLE_IOS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_DEPRECATED_IOS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_ENUM_DEPRECATED_IOS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_FORMAT_FUNCTION] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___deprecated_msg] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___deprecated_enum_msg] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_SWIFT_NAME] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_SWIFT_UNAVAILABLE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_CLASS_AVAILABLE_IOS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NS_CLASS_DEPRECATED_IOS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___OSX_AVAILABLE_STARTING] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATprotocol] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_class] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATend] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATinterface] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATimplementation] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___covariant] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___contravariant] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATprivate] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATprotected] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATpackage] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATpublic] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATcompatibility_alias] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AToptional] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATrequired] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATsynthesize] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATdynamic] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPARENclass_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATproperty] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATtry] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___try] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATcatch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___catch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATfinally] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___finally] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATthrow] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATselector] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_selector_expression_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_ATavailable] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___builtin_available] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_va_arg] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_ms_asm_block_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_ATencode] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATsynchronized] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BOOL] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_IMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEL] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_Class] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_id] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATdefs] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_out] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_inout] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_bycopy] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_byref] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_oneway] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON2] = {
    .visible = true,
    .named = false,
  },
  [sym_translation_unit] = {
    .visible = true,
    .named = true,
  },
  [sym__top_level_item] = {
    .visible = false,
    .named = true,
  },
  [sym__block_item] = {
    .visible = false,
    .named = true,
  },
  [sym_preproc_include] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_def] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_function_def] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_params] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_call] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_ifdef] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elif] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elifdef] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_in_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_ifdef_in_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else_in_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elif_in_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elifdef_in_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym__preproc_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_preproc_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_defined] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym__old_style_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym__type_definition_type] = {
    .visible = false,
    .named = true,
  },
  [sym__type_definition_declarators] = {
    .visible = false,
    .named = true,
  },
  [sym__declaration_modifiers] = {
    .visible = false,
    .named = true,
  },
  [sym__declaration_specifiers] = {
    .visible = false,
    .named = true,
  },
  [sym_linkage_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_declspec_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_based_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_call_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_unaligned_ptr_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_pointer_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym__declarator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__field_declarator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__type_declarator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__abstract_declarator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_parenthesized_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_type_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_parenthesized_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_attributed_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_pointer_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_pointer_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_pointer_type_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_pointer_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_function_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_function_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_function_type_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_function_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym__old_style_function_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_array_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_array_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_array_type_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_array_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_init_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_compound_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_storage_class_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_type_qualifier] = {
    .visible = true,
    .named = true,
  },
  [sym_alignas_qualifier] = {
    .visible = true,
    .named = true,
  },
  [sym_type_specifier] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_sized_type_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_enumerator_list] = {
    .visible = true,
    .named = true,
  },
  [sym_struct_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_union_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym__field_declaration_list_item] = {
    .visible = false,
    .named = true,
  },
  [sym_field_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_bitfield_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_enumerator] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter_list] = {
    .visible = true,
    .named = true,
  },
  [sym__old_style_parameter_list] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_attributed_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_statement] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__top_level_statement] = {
    .visible = false,
    .named = true,
  },
  [sym_labeled_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__top_level_expression_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_case_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_while_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_do_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__for_statement_body] = {
    .visible = false,
    .named = true,
  },
  [sym_return_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_break_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_continue_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_goto_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__string] = {
    .visible = false,
    .named = true,
  },
  [sym_comma_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_pointer_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_update_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_cast_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_type_descriptor] = {
    .visible = true,
    .named = true,
  },
  [sym_sizeof_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_alignof_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_offsetof_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_subscript_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_qualifier] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_output_operand_list] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_output_operand] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_input_operand_list] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_input_operand] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_clobber_list] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_goto_list] = {
    .visible = true,
    .named = true,
  },
  [sym_extension_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym_field_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_compound_literal_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_initializer_list] = {
    .visible = true,
    .named = true,
  },
  [sym_initializer_pair] = {
    .visible = true,
    .named = true,
  },
  [sym_subscript_designator] = {
    .visible = true,
    .named = true,
  },
  [sym_subscript_range_designator] = {
    .visible = true,
    .named = true,
  },
  [sym_field_designator] = {
    .visible = true,
    .named = true,
  },
  [sym_char_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_concatenated_string] = {
    .visible = true,
    .named = true,
  },
  [sym_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_null] = {
    .visible = true,
    .named = true,
  },
  [sym__empty_declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_macro_type_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_objc_bridge] = {
    .visible = true,
    .named = true,
  },
  [sym_typeof_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_availability] = {
    .visible = true,
    .named = true,
  },
  [sym_version] = {
    .visible = true,
    .named = true,
  },
  [sym_platform] = {
    .visible = true,
    .named = true,
  },
  [sym_module_import] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_in_implementation_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else_in_implementation_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elif_in_implementation_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_in_interface_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else_in_interface_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elif_in_interface_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_ifdef_in_enumerator] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else_in_enumerator] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elif_in_enumerator] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_undef] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_linemarker] = {
    .visible = true,
    .named = true,
  },
  [sym_availability_attribute_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_protocol_forward_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_class_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_class_interface] = {
    .visible = true,
    .named = true,
  },
  [sym__class_interface_header] = {
    .visible = false,
    .named = true,
  },
  [sym__type_params] = {
    .visible = false,
    .named = true,
  },
  [sym__class_interface_inheritance] = {
    .visible = false,
    .named = true,
  },
  [sym_class_implementation] = {
    .visible = true,
    .named = true,
  },
  [sym__class_implementation_header] = {
    .visible = false,
    .named = true,
  },
  [sym__class_implementation_inheritance] = {
    .visible = false,
    .named = true,
  },
  [sym_protocol_reference_list] = {
    .visible = true,
    .named = true,
  },
  [sym_parameterized_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variables] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_visibility_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_protocol_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_compatibility_alias_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_qualified_protocol_interface_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_implementation_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_property_implementation] = {
    .visible = true,
    .named = true,
  },
  [sym_method_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_method_type] = {
    .visible = true,
    .named = true,
  },
  [sym_keyword_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_property_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_property_attributes_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_property_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_method_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_method_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_c_method_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_struct_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_atomic_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_specifier_qualifier] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_struct_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_try_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_catch_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_finally_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_throw_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_selector_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_available_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_range_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_block_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_message_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_va_arg_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_asm_block] = {
    .visible = true,
    .named = true,
  },
  [sym_encode_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_synchronized_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_block_pointer_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_block_pointer_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_block_pointer_type_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_block_pointer_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_typedefed_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_array_type_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_atdef_field] = {
    .visible = true,
    .named = true,
  },
  [sym_protocol_qualifier] = {
    .visible = true,
    .named = true,
  },
  [sym_type_name] = {
    .visible = true,
    .named = true,
  },
  [sym_at_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_dictionary_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_dictionary_pair] = {
    .visible = true,
    .named = true,
  },
  [sym_array_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_method_identifier] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_translation_unit_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_params_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_in_field_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_argument_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__old_style_function_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__type_definition_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__type_definition_declarators_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__declaration_specifiers_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_specifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attributed_field_declarator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pointer_declarator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_declarator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_field_declarator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_array_declarator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_sized_type_specifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enumerator_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_field_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_parameter_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__old_style_parameter_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_case_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_generic_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_gnu_asm_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_gnu_asm_output_operand_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_gnu_asm_input_operand_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_gnu_asm_clobber_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_gnu_asm_goto_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_argument_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_argument_list_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_initializer_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_initializer_pair_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_char_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_concatenated_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_availability_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_version_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_import_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_in_implementation_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_in_interface_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_ifdef_in_enumerator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_availability_attribute_specifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_protocol_forward_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_interface_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_implementation_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_parameterized_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_parameterized_arguments_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_generic_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_instance_variables_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_protocol_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_property_implementation_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_method_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_method_definition_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_method_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_keyword_selector_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_property_attributes_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_method_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_method_declaration_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_c_method_parameter_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_struct_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_struct_declaration_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_try_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_selector_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_selector_expression_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_available_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_message_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_message_expression_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_generic_specifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_dictionary_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_method_identifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_method_identifier_repeat2] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_field_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_statement_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_type_identifier] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alias = 1,
  field_alternative = 2,
  field_argument = 3,
  field_arguments = 4,
  field_assembly_code = 5,
  field_base = 6,
  field_body = 7,
  field_category = 8,
  field_class = 9,
  field_clobbers = 10,
  field_column = 11,
  field_condition = 12,
  field_consequence = 13,
  field_constraint = 14,
  field_declarator = 15,
  field_designator = 16,
  field_directive = 17,
  field_end = 18,
  field_field = 19,
  field_filename = 20,
  field_function = 21,
  field_goto_labels = 22,
  field_index = 23,
  field_initializer = 24,
  field_input_operands = 25,
  field_label = 26,
  field_left = 27,
  field_member = 28,
  field_method = 29,
  field_name = 30,
  field_operand = 31,
  field_operator = 32,
  field_output_operands = 33,
  field_parameters = 34,
  field_path = 35,
  field_prefix = 36,
  field_receiver = 37,
  field_register = 38,
  field_right = 39,
  field_row = 40,
  field_size = 41,
  field_start = 42,
  field_superclass = 43,
  field_symbol = 44,
  field_type = 45,
  field_underlying_type = 46,
  field_update = 47,
  field_value = 48,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_alternative] = "alternative",
  [field_argument] = "argument",
  [field_arguments] = "arguments",
  [field_assembly_code] = "assembly_code",
  [field_base] = "base",
  [field_body] = "body",
  [field_category] = "category",
  [field_class] = "class",
  [field_clobbers] = "clobbers",
  [field_column] = "column",
  [field_condition] = "condition",
  [field_consequence] = "consequence",
  [field_constraint] = "constraint",
  [field_declarator] = "declarator",
  [field_designator] = "designator",
  [field_directive] = "directive",
  [field_end] = "end",
  [field_field] = "field",
  [field_filename] = "filename",
  [field_function] = "function",
  [field_goto_labels] = "goto_labels",
  [field_index] = "index",
  [field_initializer] = "initializer",
  [field_input_operands] = "input_operands",
  [field_label] = "label",
  [field_left] = "left",
  [field_member] = "member",
  [field_method] = "method",
  [field_name] = "name",
  [field_operand] = "operand",
  [field_operator] = "operator",
  [field_output_operands] = "output_operands",
  [field_parameters] = "parameters",
  [field_path] = "path",
  [field_prefix] = "prefix",
  [field_receiver] = "receiver",
  [field_register] = "register",
  [field_right] = "right",
  [field_row] = "row",
  [field_size] = "size",
  [field_start] = "start",
  [field_superclass] = "superclass",
  [field_symbol] = "symbol",
  [field_type] = "type",
  [field_underlying_type] = "underlying_type",
  [field_update] = "update",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [3] = {.index = 0, .length = 3},
  [4] = {.index = 3, .length = 1},
  [5] = {.index = 4, .length = 1},
  [6] = {.index = 5, .length = 2},
  [7] = {.index = 7, .length = 1},
  [8] = {.index = 8, .length = 1},
  [9] = {.index = 9, .length = 1},
  [10] = {.index = 3, .length = 1},
  [11] = {.index = 10, .length = 2},
  [12] = {.index = 12, .length = 2},
  [13] = {.index = 14, .length = 1},
  [14] = {.index = 14, .length = 1},
  [15] = {.index = 15, .length = 2},
  [16] = {.index = 8, .length = 1},
  [17] = {.index = 17, .length = 2},
  [18] = {.index = 19, .length = 2},
  [19] = {.index = 21, .length = 1},
  [20] = {.index = 22, .length = 1},
  [21] = {.index = 23, .length = 1},
  [22] = {.index = 24, .length = 2},
  [24] = {.index = 26, .length = 1},
  [25] = {.index = 27, .length = 2},
  [26] = {.index = 29, .length = 1},
  [27] = {.index = 30, .length = 1},
  [28] = {.index = 31, .length = 2},
  [29] = {.index = 33, .length = 1},
  [30] = {.index = 34, .length = 1},
  [31] = {.index = 35, .length = 2},
  [32] = {.index = 37, .length = 2},
  [33] = {.index = 39, .length = 1},
  [34] = {.index = 40, .length = 3},
  [35] = {.index = 43, .length = 1},
  [36] = {.index = 44, .length = 1},
  [37] = {.index = 45, .length = 2},
  [38] = {.index = 47, .length = 2},
  [39] = {.index = 49, .length = 3},
  [40] = {.index = 52, .length = 3},
  [41] = {.index = 55, .length = 1},
  [42] = {.index = 56, .length = 2},
  [43] = {.index = 58, .length = 1},
  [44] = {.index = 58, .length = 1},
  [45] = {.index = 59, .length = 2},
  [46] = {.index = 61, .length = 2},
  [47] = {.index = 63, .length = 1},
  [48] = {.index = 64, .length = 2},
  [50] = {.index = 66, .length = 2},
  [51] = {.index = 68, .length = 2},
  [52] = {.index = 70, .length = 2},
  [53] = {.index = 72, .length = 2},
  [54] = {.index = 74, .length = 1},
  [55] = {.index = 75, .length = 2},
  [56] = {.index = 77, .length = 2},
  [57] = {.index = 79, .length = 2},
  [58] = {.index = 81, .length = 2},
  [59] = {.index = 83, .length = 2},
  [60] = {.index = 85, .length = 2},
  [61] = {.index = 85, .length = 2},
  [62] = {.index = 87, .length = 2},
  [63] = {.index = 89, .length = 1},
  [64] = {.index = 90, .length = 1},
  [66] = {.index = 91, .length = 3},
  [67] = {.index = 94, .length = 1},
  [68] = {.index = 95, .length = 1},
  [69] = {.index = 96, .length = 2},
  [70] = {.index = 98, .length = 2},
  [71] = {.index = 100, .length = 1},
  [72] = {.index = 101, .length = 2},
  [73] = {.index = 103, .length = 3},
  [74] = {.index = 106, .length = 3},
  [75] = {.index = 109, .length = 2},
  [76] = {.index = 111, .length = 5},
  [77] = {.index = 116, .length = 3},
  [78] = {.index = 119, .length = 2},
  [79] = {.index = 121, .length = 2},
  [80] = {.index = 123, .length = 1},
  [82] = {.index = 124, .length = 2},
  [83] = {.index = 126, .length = 3},
  [84] = {.index = 129, .length = 2},
  [85] = {.index = 131, .length = 2},
  [87] = {.index = 133, .length = 1},
  [88] = {.index = 134, .length = 2},
  [89] = {.index = 136, .length = 1},
  [90] = {.index = 137, .length = 2},
  [91] = {.index = 139, .length = 2},
  [92] = {.index = 141, .length = 2},
  [93] = {.index = 143, .length = 2},
  [94] = {.index = 145, .length = 3},
  [95] = {.index = 145, .length = 3},
  [96] = {.index = 148, .length = 2},
  [97] = {.index = 150, .length = 2},
  [98] = {.index = 150, .length = 2},
  [99] = {.index = 152, .length = 2},
  [100] = {.index = 154, .length = 2},
  [101] = {.index = 156, .length = 1},
  [102] = {.index = 157, .length = 2},
  [103] = {.index = 159, .length = 2},
  [104] = {.index = 161, .length = 4},
  [105] = {.index = 165, .length = 1},
  [106] = {.index = 166, .length = 2},
  [107] = {.index = 168, .length = 1},
  [108] = {.index = 169, .length = 2},
  [109] = {.index = 171, .length = 2},
  [110] = {.index = 173, .length = 2},
  [111] = {.index = 175, .length = 4},
  [112] = {.index = 179, .length = 5},
  [113] = {.index = 184, .length = 3},
  [115] = {.index = 187, .length = 2},
  [116] = {.index = 189, .length = 1},
  [118] = {.index = 190, .length = 1},
  [119] = {.index = 191, .length = 2},
  [120] = {.index = 193, .length = 2},
  [121] = {.index = 195, .length = 3},
  [122] = {.index = 195, .length = 3},
  [123] = {.index = 198, .length = 2},
  [124] = {.index = 200, .length = 2},
  [125] = {.index = 200, .length = 2},
  [126] = {.index = 202, .length = 2},
  [127] = {.index = 204, .length = 3},
  [128] = {.index = 207, .length = 2},
  [129] = {.index = 209, .length = 2},
  [130] = {.index = 211, .length = 2},
  [131] = {.index = 213, .length = 2},
  [132] = {.index = 215, .length = 3},
  [133] = {.index = 218, .length = 2},
  [134] = {.index = 220, .length = 3},
  [136] = {.index = 223, .length = 2},
  [137] = {.index = 223, .length = 2},
  [138] = {.index = 225, .length = 2},
  [139] = {.index = 227, .length = 3},
  [140] = {.index = 227, .length = 3},
  [141] = {.index = 230, .length = 3},
  [142] = {.index = 233, .length = 3},
  [143] = {.index = 236, .length = 2},
  [144] = {.index = 238, .length = 1},
  [145] = {.index = 239, .length = 4},
  [146] = {.index = 243, .length = 3},
  [147] = {.index = 246, .length = 2},
  [148] = {.index = 248, .length = 1},
  [149] = {.index = 249, .length = 2},
  [150] = {.index = 251, .length = 2},
  [151] = {.index = 39, .length = 1},
  [152] = {.index = 253, .length = 5},
  [153] = {.index = 258, .length = 4},
  [155] = {.index = 262, .length = 2},
  [156] = {.index = 264, .length = 2},
  [157] = {.index = 266, .length = 2},
  [158] = {.index = 268, .length = 5},
  [159] = {.index = 273, .length = 2},
  [160] = {.index = 275, .length = 3},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_body, 0, .inherited = true},
    {field_declarator, 0, .inherited = true},
    {field_type, 0, .inherited = true},
  [3] =
    {field_type, 0},
  [4] =
    {field_directive, 0},
  [5] =
    {field_argument, 1},
    {field_operator, 0},
  [7] =
    {field_value, 1},
  [8] =
    {field_name, 1},
  [9] =
    {field_body, 1},
  [10] =
    {field_argument, 0},
    {field_operator, 1},
  [12] =
    {field_arguments, 1},
    {field_function, 0},
  [14] =
    {field_type, 1},
  [15] =
    {field_directive, 0},
    {field_path, 1},
  [17] =
    {field_argument, 1},
    {field_directive, 0},
  [19] =
    {field_declarator, 1},
    {field_type, 0},
  [21] =
    {field_parameters, 0},
  [22] =
    {field_type, 0, .inherited = true},
  [23] =
    {field_declarator, 1},
  [24] =
    {field_declarator, 0},
    {field_parameters, 1},
  [26] =
    {field_declarator, 0},
  [27] =
    {field_body, 2},
    {field_value, 1},
  [29] =
    {field_name, 0},
  [30] =
    {field_method, 0},
  [31] =
    {field_body, 2},
    {field_name, 1},
  [33] =
    {field_name, 2},
  [34] =
    {field_body, 2},
  [35] =
    {field_condition, 1},
    {field_consequence, 2},
  [37] =
    {field_body, 2},
    {field_condition, 1},
  [39] =
    {field_label, 1},
  [40] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [43] =
    {field_label, 0},
  [44] =
    {field_path, 1},
  [45] =
    {field_alias, 2},
    {field_class, 1},
  [47] =
    {field_declarator, 1},
    {field_type, 0, .inherited = true},
  [49] =
    {field_body, 2},
    {field_declarator, 1},
    {field_type, 0, .inherited = true},
  [52] =
    {field_argument, 0},
    {field_field, 2},
    {field_operator, 1},
  [55] =
    {field_superclass, 1},
  [56] =
    {field_category, 1, .inherited = true},
    {field_superclass, 1, .inherited = true},
  [58] =
    {field_type, 2},
  [59] =
    {field_name, 1},
    {field_value, 2},
  [61] =
    {field_name, 1},
    {field_parameters, 2},
  [63] =
    {field_condition, 1},
  [64] =
    {field_alternative, 2},
    {field_name, 1},
  [66] =
    {field_declarator, 2},
    {field_type, 0},
  [68] =
    {field_left, 0},
    {field_right, 2},
  [70] =
    {field_type, 1},
    {field_value, 3},
  [72] =
    {field_declarator, 2},
    {field_type, 1},
  [74] =
    {field_declarator, 2},
  [75] =
    {field_declarator, 2, .inherited = true},
    {field_type, 1, .inherited = true},
  [77] =
    {field_declarator, 0},
    {field_declarator, 1, .inherited = true},
  [79] =
    {field_method, 2, .inherited = true},
    {field_receiver, 1},
  [81] =
    {field_method, 0, .inherited = true},
    {field_method, 1, .inherited = true},
  [83] =
    {field_base, 2},
    {field_body, 3},
  [85] =
    {field_name, 1},
    {field_underlying_type, 3},
  [87] =
    {field_body, 3},
    {field_name, 2},
  [89] =
    {field_name, 3},
  [90] =
    {field_body, 3},
  [91] =
    {field_alternative, 3},
    {field_condition, 1},
    {field_consequence, 2},
  [94] =
    {field_initializer, 0},
  [95] =
    {field_assembly_code, 2},
  [96] =
    {field_name, 0},
    {field_type, 2},
  [98] =
    {field_path, 1},
    {field_path, 2},
  [100] =
    {field_filename, 2},
  [101] =
    {field_declarator, 0},
    {field_value, 2},
  [103] =
    {field_declarator, 1},
    {field_declarator, 2},
    {field_type, 0, .inherited = true},
  [106] =
    {field_declarator, 1},
    {field_declarator, 2, .inherited = true},
    {field_type, 0, .inherited = true},
  [109] =
    {field_declarator, 0, .inherited = true},
    {field_declarator, 1, .inherited = true},
  [111] =
    {field_body, 3},
    {field_declarator, 1},
    {field_declarator, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_type, 0, .inherited = true},
  [116] =
    {field_body, 3},
    {field_declarator, 2},
    {field_type, 1, .inherited = true},
  [119] =
    {field_argument, 0},
    {field_index, 2},
  [121] =
    {field_alternative, 3},
    {field_condition, 0},
  [123] =
    {field_category, 1},
  [124] =
    {field_category, 2, .inherited = true},
    {field_superclass, 2, .inherited = true},
  [126] =
    {field_name, 1},
    {field_parameters, 2},
    {field_value, 3},
  [129] =
    {field_alternative, 3},
    {field_condition, 1},
  [131] =
    {field_alternative, 3},
    {field_name, 1},
  [133] =
    {field_size, 1},
  [134] =
    {field_declarator, 3},
    {field_type, 1},
  [136] =
    {field_declarator, 3},
  [137] =
    {field_declarator, 0},
    {field_size, 2},
  [139] =
    {field_declarator, 3, .inherited = true},
    {field_type, 2, .inherited = true},
  [141] =
    {field_name, 2},
    {field_prefix, 0},
  [143] =
    {field_name, 0},
    {field_value, 2},
  [145] =
    {field_body, 4},
    {field_name, 1},
    {field_underlying_type, 3},
  [148] =
    {field_base, 3},
    {field_body, 4},
  [150] =
    {field_name, 2},
    {field_underlying_type, 4},
  [152] =
    {field_body, 4},
    {field_name, 3},
  [154] =
    {field_body, 1},
    {field_condition, 3},
  [156] =
    {field_update, 2},
  [157] =
    {field_initializer, 0},
    {field_update, 2},
  [159] =
    {field_condition, 1},
    {field_initializer, 0},
  [161] =
    {field_body, 4},
    {field_condition, 2, .inherited = true},
    {field_initializer, 2, .inherited = true},
    {field_update, 2, .inherited = true},
  [165] =
    {field_operand, 1},
  [166] =
    {field_assembly_code, 2},
    {field_output_operands, 3},
  [168] =
    {field_assembly_code, 3},
  [169] =
    {field_filename, 2},
    {field_row, 3},
  [171] =
    {field_declarator, 1},
    {field_declarator, 2},
  [173] =
    {field_declarator, 0},
    {field_value, 3},
  [175] =
    {field_declarator, 1},
    {field_declarator, 2},
    {field_declarator, 3, .inherited = true},
    {field_type, 0, .inherited = true},
  [179] =
    {field_body, 4},
    {field_declarator, 2},
    {field_declarator, 2, .inherited = true},
    {field_parameters, 2, .inherited = true},
    {field_type, 1, .inherited = true},
  [184] =
    {field_alternative, 4},
    {field_condition, 0},
    {field_consequence, 2},
  [187] =
    {field_alternative, 4},
    {field_condition, 1},
  [189] =
    {field_size, 2},
  [190] =
    {field_declarator, 4},
  [191] =
    {field_declarator, 0},
    {field_size, 3},
  [193] =
    {field_declarator, 4, .inherited = true},
    {field_type, 3, .inherited = true},
  [195] =
    {field_body, 5},
    {field_name, 2},
    {field_underlying_type, 4},
  [198] =
    {field_base, 4},
    {field_body, 5},
  [200] =
    {field_name, 3},
    {field_underlying_type, 5},
  [202] =
    {field_condition, 1},
    {field_update, 3},
  [204] =
    {field_condition, 1},
    {field_initializer, 0},
    {field_update, 3},
  [207] =
    {field_initializer, 0},
    {field_update, 3},
  [209] =
    {field_condition, 2},
    {field_initializer, 0},
  [211] =
    {field_member, 4},
    {field_type, 2},
  [213] =
    {field_operand, 1},
    {field_operand, 2, .inherited = true},
  [215] =
    {field_assembly_code, 2},
    {field_input_operands, 4},
    {field_output_operands, 3},
  [218] =
    {field_assembly_code, 3},
    {field_output_operands, 4},
  [220] =
    {field_column, 4},
    {field_filename, 2},
    {field_row, 3},
  [223] =
    {field_designator, 0},
    {field_value, 2},
  [225] =
    {field_declarator, 5, .inherited = true},
    {field_type, 4, .inherited = true},
  [227] =
    {field_body, 6},
    {field_name, 3},
    {field_underlying_type, 5},
  [230] =
    {field_declarator, 1},
    {field_declarator, 3, .inherited = true},
    {field_type, 0, .inherited = true},
  [233] =
    {field_condition, 2},
    {field_initializer, 0},
    {field_update, 4},
  [236] =
    {field_operand, 0, .inherited = true},
    {field_operand, 1, .inherited = true},
  [238] =
    {field_register, 1},
  [239] =
    {field_assembly_code, 2},
    {field_clobbers, 5},
    {field_input_operands, 4},
    {field_output_operands, 3},
  [243] =
    {field_assembly_code, 3},
    {field_input_operands, 5},
    {field_output_operands, 4},
  [246] =
    {field_declarator, 6, .inherited = true},
    {field_type, 5, .inherited = true},
  [248] =
    {field_type, 2, .inherited = true},
  [249] =
    {field_constraint, 0},
    {field_value, 2},
  [251] =
    {field_register, 1},
    {field_register, 2, .inherited = true},
  [253] =
    {field_assembly_code, 2},
    {field_clobbers, 5},
    {field_goto_labels, 6},
    {field_input_operands, 4},
    {field_output_operands, 3},
  [258] =
    {field_assembly_code, 3},
    {field_clobbers, 6},
    {field_input_operands, 5},
    {field_output_operands, 4},
  [262] =
    {field_end, 3},
    {field_start, 1},
  [264] =
    {field_register, 0, .inherited = true},
    {field_register, 1, .inherited = true},
  [266] =
    {field_label, 1},
    {field_label, 2, .inherited = true},
  [268] =
    {field_assembly_code, 3},
    {field_clobbers, 6},
    {field_goto_labels, 7},
    {field_input_operands, 5},
    {field_output_operands, 4},
  [273] =
    {field_label, 0, .inherited = true},
    {field_label, 1, .inherited = true},
  [275] =
    {field_constraint, 3},
    {field_symbol, 1},
    {field_value, 5},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = sym_identifier,
  },
  [2] = {
    [0] = alias_sym_type_identifier,
  },
  [8] = {
    [1] = alias_sym_type_identifier,
  },
  [10] = {
    [0] = alias_sym_type_identifier,
  },
  [14] = {
    [1] = alias_sym_type_identifier,
  },
  [23] = {
    [0] = sym_primitive_type,
  },
  [28] = {
    [1] = alias_sym_type_identifier,
  },
  [29] = {
    [2] = alias_sym_type_identifier,
  },
  [33] = {
    [1] = alias_sym_statement_identifier,
  },
  [35] = {
    [0] = alias_sym_statement_identifier,
  },
  [40] = {
    [2] = alias_sym_field_identifier,
  },
  [44] = {
    [2] = alias_sym_type_identifier,
  },
  [49] = {
    [1] = alias_sym_type_identifier,
  },
  [59] = {
    [2] = alias_sym_type_identifier,
  },
  [60] = {
    [1] = alias_sym_type_identifier,
  },
  [61] = {
    [1] = alias_sym_type_identifier,
    [3] = alias_sym_type_identifier,
  },
  [62] = {
    [2] = alias_sym_type_identifier,
  },
  [63] = {
    [3] = alias_sym_type_identifier,
  },
  [65] = {
    [0] = alias_sym_field_identifier,
  },
  [81] = {
    [1] = sym_identifier,
  },
  [86] = {
    [2] = alias_sym_type_identifier,
  },
  [94] = {
    [1] = alias_sym_type_identifier,
  },
  [95] = {
    [1] = alias_sym_type_identifier,
    [3] = alias_sym_type_identifier,
  },
  [96] = {
    [3] = alias_sym_type_identifier,
  },
  [97] = {
    [2] = alias_sym_type_identifier,
  },
  [98] = {
    [2] = alias_sym_type_identifier,
    [4] = alias_sym_type_identifier,
  },
  [99] = {
    [3] = alias_sym_type_identifier,
  },
  [114] = {
    [2] = sym_identifier,
  },
  [117] = {
    [1] = alias_sym_field_identifier,
  },
  [121] = {
    [2] = alias_sym_type_identifier,
  },
  [122] = {
    [2] = alias_sym_type_identifier,
    [4] = alias_sym_type_identifier,
  },
  [123] = {
    [4] = alias_sym_type_identifier,
  },
  [124] = {
    [3] = alias_sym_type_identifier,
  },
  [125] = {
    [3] = alias_sym_type_identifier,
    [5] = alias_sym_type_identifier,
  },
  [130] = {
    [4] = alias_sym_field_identifier,
  },
  [135] = {
    [3] = sym_identifier,
  },
  [136] = {
    [0] = alias_sym_field_identifier,
  },
  [139] = {
    [3] = alias_sym_type_identifier,
  },
  [140] = {
    [3] = alias_sym_type_identifier,
    [5] = alias_sym_type_identifier,
  },
  [154] = {
    [4] = alias_sym_field_identifier,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 2,
  [4] = 4,
  [5] = 2,
  [6] = 4,
  [7] = 4,
  [8] = 4,
  [9] = 2,
  [10] = 4,
  [11] = 2,
  [12] = 4,
  [13] = 2,
  [14] = 4,
  [15] = 2,
  [16] = 4,
  [17] = 2,
  [18] = 4,
  [19] = 2,
  [20] = 20,
  [21] = 21,
  [22] = 22,
  [23] = 20,
  [24] = 24,
  [25] = 21,
  [26] = 26,
  [27] = 20,
  [28] = 21,
  [29] = 20,
  [30] = 21,
  [31] = 21,
  [32] = 20,
  [33] = 33,
  [34] = 21,
  [35] = 35,
  [36] = 20,
  [37] = 33,
  [38] = 33,
  [39] = 39,
  [40] = 40,
  [41] = 41,
  [42] = 42,
  [43] = 43,
  [44] = 44,
  [45] = 44,
  [46] = 46,
  [47] = 47,
  [48] = 48,
  [49] = 48,
  [50] = 41,
  [51] = 47,
  [52] = 41,
  [53] = 43,
  [54] = 44,
  [55] = 46,
  [56] = 47,
  [57] = 48,
  [58] = 41,
  [59] = 43,
  [60] = 44,
  [61] = 46,
  [62] = 47,
  [63] = 48,
  [64] = 41,
  [65] = 43,
  [66] = 44,
  [67] = 46,
  [68] = 47,
  [69] = 48,
  [70] = 41,
  [71] = 44,
  [72] = 47,
  [73] = 48,
  [74] = 41,
  [75] = 44,
  [76] = 47,
  [77] = 46,
  [78] = 41,
  [79] = 47,
  [80] = 44,
  [81] = 47,
  [82] = 43,
  [83] = 48,
  [84] = 41,
  [85] = 48,
  [86] = 44,
  [87] = 47,
  [88] = 48,
  [89] = 41,
  [90] = 44,
  [91] = 47,
  [92] = 48,
  [93] = 41,
  [94] = 44,
  [95] = 47,
  [96] = 48,
  [97] = 41,
  [98] = 44,
  [99] = 47,
  [100] = 48,
  [101] = 41,
  [102] = 44,
  [103] = 47,
  [104] = 48,
  [105] = 41,
  [106] = 44,
  [107] = 47,
  [108] = 48,
  [109] = 41,
  [110] = 44,
  [111] = 47,
  [112] = 48,
  [113] = 41,
  [114] = 44,
  [115] = 47,
  [116] = 48,
  [117] = 44,
  [118] = 47,
  [119] = 48,
  [120] = 41,
  [121] = 44,
  [122] = 47,
  [123] = 48,
  [124] = 41,
  [125] = 44,
  [126] = 41,
  [127] = 48,
  [128] = 128,
  [129] = 129,
  [130] = 130,
  [131] = 131,
  [132] = 132,
  [133] = 133,
  [134] = 134,
  [135] = 132,
  [136] = 134,
  [137] = 131,
  [138] = 133,
  [139] = 130,
  [140] = 132,
  [141] = 131,
  [142] = 133,
  [143] = 130,
  [144] = 134,
  [145] = 132,
  [146] = 132,
  [147] = 130,
  [148] = 134,
  [149] = 133,
  [150] = 131,
  [151] = 131,
  [152] = 133,
  [153] = 130,
  [154] = 134,
  [155] = 155,
  [156] = 134,
  [157] = 155,
  [158] = 155,
  [159] = 155,
  [160] = 131,
  [161] = 155,
  [162] = 155,
  [163] = 132,
  [164] = 133,
  [165] = 155,
  [166] = 155,
  [167] = 130,
  [168] = 168,
  [169] = 169,
  [170] = 169,
  [171] = 169,
  [172] = 169,
  [173] = 169,
  [174] = 169,
  [175] = 169,
  [176] = 169,
  [177] = 168,
  [178] = 168,
  [179] = 179,
  [180] = 180,
  [181] = 168,
  [182] = 179,
  [183] = 183,
  [184] = 184,
  [185] = 180,
  [186] = 186,
  [187] = 187,
  [188] = 188,
  [189] = 179,
  [190] = 180,
  [191] = 188,
  [192] = 187,
  [193] = 183,
  [194] = 186,
  [195] = 184,
  [196] = 179,
  [197] = 197,
  [198] = 179,
  [199] = 199,
  [200] = 200,
  [201] = 201,
  [202] = 186,
  [203] = 203,
  [204] = 204,
  [205] = 205,
  [206] = 206,
  [207] = 207,
  [208] = 208,
  [209] = 209,
  [210] = 210,
  [211] = 187,
  [212] = 212,
  [213] = 188,
  [214] = 214,
  [215] = 215,
  [216] = 216,
  [217] = 217,
  [218] = 218,
  [219] = 219,
  [220] = 220,
  [221] = 180,
  [222] = 222,
  [223] = 223,
  [224] = 183,
  [225] = 184,
  [226] = 186,
  [227] = 227,
  [228] = 228,
  [229] = 229,
  [230] = 230,
  [231] = 231,
  [232] = 232,
  [233] = 233,
  [234] = 234,
  [235] = 235,
  [236] = 236,
  [237] = 237,
  [238] = 238,
  [239] = 239,
  [240] = 240,
  [241] = 241,
  [242] = 242,
  [243] = 243,
  [244] = 183,
  [245] = 245,
  [246] = 246,
  [247] = 247,
  [248] = 248,
  [249] = 249,
  [250] = 184,
  [251] = 251,
  [252] = 252,
  [253] = 253,
  [254] = 254,
  [255] = 255,
  [256] = 256,
  [257] = 257,
  [258] = 258,
  [259] = 259,
  [260] = 260,
  [261] = 261,
  [262] = 262,
  [263] = 263,
  [264] = 264,
  [265] = 265,
  [266] = 266,
  [267] = 267,
  [268] = 268,
  [269] = 269,
  [270] = 270,
  [271] = 271,
  [272] = 272,
  [273] = 273,
  [274] = 274,
  [275] = 180,
  [276] = 276,
  [277] = 277,
  [278] = 278,
  [279] = 279,
  [280] = 280,
  [281] = 281,
  [282] = 282,
  [283] = 283,
  [284] = 284,
  [285] = 285,
  [286] = 286,
  [287] = 197,
  [288] = 288,
  [289] = 289,
  [290] = 290,
  [291] = 291,
  [292] = 292,
  [293] = 293,
  [294] = 294,
  [295] = 295,
  [296] = 296,
  [297] = 297,
  [298] = 298,
  [299] = 299,
  [300] = 300,
  [301] = 301,
  [302] = 302,
  [303] = 303,
  [304] = 304,
  [305] = 305,
  [306] = 306,
  [307] = 307,
  [308] = 308,
  [309] = 309,
  [310] = 310,
  [311] = 311,
  [312] = 312,
  [313] = 313,
  [314] = 314,
  [315] = 315,
  [316] = 316,
  [317] = 317,
  [318] = 318,
  [319] = 319,
  [320] = 320,
  [321] = 321,
  [322] = 322,
  [323] = 323,
  [324] = 324,
  [325] = 325,
  [326] = 326,
  [327] = 327,
  [328] = 328,
  [329] = 329,
  [330] = 330,
  [331] = 331,
  [332] = 332,
  [333] = 333,
  [334] = 334,
  [335] = 335,
  [336] = 336,
  [337] = 337,
  [338] = 338,
  [339] = 339,
  [340] = 340,
  [341] = 341,
  [342] = 342,
  [343] = 343,
  [344] = 344,
  [345] = 345,
  [346] = 346,
  [347] = 347,
  [348] = 348,
  [349] = 349,
  [350] = 350,
  [351] = 351,
  [352] = 352,
  [353] = 353,
  [354] = 354,
  [355] = 355,
  [356] = 356,
  [357] = 357,
  [358] = 358,
  [359] = 359,
  [360] = 360,
  [361] = 229,
  [362] = 268,
  [363] = 228,
  [364] = 242,
  [365] = 251,
  [366] = 254,
  [367] = 269,
  [368] = 270,
  [369] = 257,
  [370] = 271,
  [371] = 260,
  [372] = 263,
  [373] = 183,
  [374] = 245,
  [375] = 252,
  [376] = 203,
  [377] = 200,
  [378] = 205,
  [379] = 206,
  [380] = 184,
  [381] = 186,
  [382] = 188,
  [383] = 230,
  [384] = 272,
  [385] = 183,
  [386] = 184,
  [387] = 186,
  [388] = 234,
  [389] = 236,
  [390] = 237,
  [391] = 238,
  [392] = 276,
  [393] = 240,
  [394] = 273,
  [395] = 243,
  [396] = 274,
  [397] = 219,
  [398] = 207,
  [399] = 209,
  [400] = 210,
  [401] = 217,
  [402] = 227,
  [403] = 199,
  [404] = 218,
  [405] = 223,
  [406] = 183,
  [407] = 184,
  [408] = 258,
  [409] = 188,
  [410] = 246,
  [411] = 247,
  [412] = 248,
  [413] = 249,
  [414] = 253,
  [415] = 239,
  [416] = 255,
  [417] = 186,
  [418] = 220,
  [419] = 204,
  [420] = 256,
  [421] = 208,
  [422] = 212,
  [423] = 214,
  [424] = 233,
  [425] = 215,
  [426] = 259,
  [427] = 187,
  [428] = 261,
  [429] = 262,
  [430] = 216,
  [431] = 264,
  [432] = 235,
  [433] = 241,
  [434] = 265,
  [435] = 266,
  [436] = 267,
  [437] = 201,
  [438] = 187,
  [439] = 232,
  [440] = 222,
  [441] = 231,
  [442] = 308,
  [443] = 314,
  [444] = 315,
  [445] = 283,
  [446] = 284,
  [447] = 329,
  [448] = 330,
  [449] = 316,
  [450] = 317,
  [451] = 318,
  [452] = 335,
  [453] = 336,
  [454] = 337,
  [455] = 338,
  [456] = 339,
  [457] = 348,
  [458] = 340,
  [459] = 341,
  [460] = 342,
  [461] = 343,
  [462] = 179,
  [463] = 344,
  [464] = 345,
  [465] = 346,
  [466] = 347,
  [467] = 360,
  [468] = 307,
  [469] = 297,
  [470] = 349,
  [471] = 299,
  [472] = 350,
  [473] = 351,
  [474] = 300,
  [475] = 352,
  [476] = 319,
  [477] = 309,
  [478] = 353,
  [479] = 354,
  [480] = 355,
  [481] = 356,
  [482] = 277,
  [483] = 313,
  [484] = 197,
  [485] = 278,
  [486] = 292,
  [487] = 293,
  [488] = 302,
  [489] = 282,
  [490] = 359,
  [491] = 279,
  [492] = 285,
  [493] = 294,
  [494] = 320,
  [495] = 358,
  [496] = 321,
  [497] = 322,
  [498] = 323,
  [499] = 324,
  [500] = 325,
  [501] = 326,
  [502] = 288,
  [503] = 281,
  [504] = 357,
  [505] = 327,
  [506] = 328,
  [507] = 303,
  [508] = 304,
  [509] = 311,
  [510] = 290,
  [511] = 312,
  [512] = 305,
  [513] = 331,
  [514] = 295,
  [515] = 298,
  [516] = 306,
  [517] = 280,
  [518] = 286,
  [519] = 289,
  [520] = 310,
  [521] = 332,
  [522] = 333,
  [523] = 334,
  [524] = 291,
  [525] = 296,
  [526] = 301,
  [527] = 241,
  [528] = 251,
  [529] = 254,
  [530] = 269,
  [531] = 259,
  [532] = 270,
  [533] = 261,
  [534] = 262,
  [535] = 232,
  [536] = 264,
  [537] = 265,
  [538] = 266,
  [539] = 239,
  [540] = 186,
  [541] = 258,
  [542] = 267,
  [543] = 230,
  [544] = 204,
  [545] = 268,
  [546] = 234,
  [547] = 271,
  [548] = 236,
  [549] = 237,
  [550] = 245,
  [551] = 252,
  [552] = 203,
  [553] = 200,
  [554] = 205,
  [555] = 206,
  [556] = 207,
  [557] = 209,
  [558] = 210,
  [559] = 217,
  [560] = 227,
  [561] = 199,
  [562] = 229,
  [563] = 231,
  [564] = 233,
  [565] = 235,
  [566] = 184,
  [567] = 242,
  [568] = 208,
  [569] = 276,
  [570] = 240,
  [571] = 201,
  [572] = 180,
  [573] = 243,
  [574] = 212,
  [575] = 214,
  [576] = 246,
  [577] = 247,
  [578] = 248,
  [579] = 249,
  [580] = 215,
  [581] = 218,
  [582] = 219,
  [583] = 220,
  [584] = 183,
  [585] = 216,
  [586] = 253,
  [587] = 222,
  [588] = 255,
  [589] = 256,
  [590] = 223,
  [591] = 257,
  [592] = 260,
  [593] = 263,
  [594] = 228,
  [595] = 168,
  [596] = 272,
  [597] = 273,
  [598] = 274,
  [599] = 238,
  [600] = 297,
  [601] = 320,
  [602] = 321,
  [603] = 322,
  [604] = 323,
  [605] = 324,
  [606] = 325,
  [607] = 326,
  [608] = 327,
  [609] = 328,
  [610] = 331,
  [611] = 332,
  [612] = 333,
  [613] = 334,
  [614] = 335,
  [615] = 336,
  [616] = 337,
  [617] = 338,
  [618] = 339,
  [619] = 340,
  [620] = 341,
  [621] = 342,
  [622] = 343,
  [623] = 344,
  [624] = 345,
  [625] = 346,
  [626] = 347,
  [627] = 349,
  [628] = 350,
  [629] = 351,
  [630] = 352,
  [631] = 353,
  [632] = 354,
  [633] = 355,
  [634] = 356,
  [635] = 277,
  [636] = 278,
  [637] = 282,
  [638] = 359,
  [639] = 279,
  [640] = 197,
  [641] = 358,
  [642] = 281,
  [643] = 357,
  [644] = 319,
  [645] = 280,
  [646] = 286,
  [647] = 289,
  [648] = 310,
  [649] = 283,
  [650] = 284,
  [651] = 329,
  [652] = 330,
  [653] = 348,
  [654] = 299,
  [655] = 300,
  [656] = 197,
  [657] = 313,
  [658] = 293,
  [659] = 301,
  [660] = 285,
  [661] = 288,
  [662] = 290,
  [663] = 291,
  [664] = 292,
  [665] = 294,
  [666] = 295,
  [667] = 296,
  [668] = 298,
  [669] = 302,
  [670] = 303,
  [671] = 304,
  [672] = 305,
  [673] = 360,
  [674] = 307,
  [675] = 308,
  [676] = 309,
  [677] = 311,
  [678] = 312,
  [679] = 314,
  [680] = 315,
  [681] = 316,
  [682] = 317,
  [683] = 318,
  [684] = 306,
  [685] = 253,
  [686] = 258,
  [687] = 239,
  [688] = 218,
  [689] = 183,
  [690] = 232,
  [691] = 248,
  [692] = 219,
  [693] = 220,
  [694] = 184,
  [695] = 201,
  [696] = 256,
  [697] = 215,
  [698] = 236,
  [699] = 237,
  [700] = 186,
  [701] = 204,
  [702] = 208,
  [703] = 212,
  [704] = 257,
  [705] = 214,
  [706] = 215,
  [707] = 216,
  [708] = 230,
  [709] = 230,
  [710] = 260,
  [711] = 263,
  [712] = 238,
  [713] = 234,
  [714] = 236,
  [715] = 237,
  [716] = 238,
  [717] = 218,
  [718] = 219,
  [719] = 220,
  [720] = 276,
  [721] = 240,
  [722] = 243,
  [723] = 187,
  [724] = 246,
  [725] = 247,
  [726] = 248,
  [727] = 249,
  [728] = 264,
  [729] = 216,
  [730] = 265,
  [731] = 266,
  [732] = 222,
  [733] = 223,
  [734] = 228,
  [735] = 242,
  [736] = 269,
  [737] = 251,
  [738] = 254,
  [739] = 253,
  [740] = 255,
  [741] = 256,
  [742] = 212,
  [743] = 257,
  [744] = 260,
  [745] = 263,
  [746] = 269,
  [747] = 270,
  [748] = 270,
  [749] = 272,
  [750] = 273,
  [751] = 271,
  [752] = 183,
  [753] = 184,
  [754] = 186,
  [755] = 259,
  [756] = 255,
  [757] = 261,
  [758] = 262,
  [759] = 264,
  [760] = 265,
  [761] = 266,
  [762] = 274,
  [763] = 188,
  [764] = 245,
  [765] = 249,
  [766] = 203,
  [767] = 200,
  [768] = 205,
  [769] = 206,
  [770] = 207,
  [771] = 209,
  [772] = 210,
  [773] = 217,
  [774] = 227,
  [775] = 199,
  [776] = 229,
  [777] = 231,
  [778] = 233,
  [779] = 214,
  [780] = 235,
  [781] = 241,
  [782] = 183,
  [783] = 276,
  [784] = 240,
  [785] = 267,
  [786] = 259,
  [787] = 261,
  [788] = 262,
  [789] = 222,
  [790] = 184,
  [791] = 268,
  [792] = 271,
  [793] = 223,
  [794] = 245,
  [795] = 252,
  [796] = 203,
  [797] = 200,
  [798] = 205,
  [799] = 206,
  [800] = 207,
  [801] = 209,
  [802] = 210,
  [803] = 217,
  [804] = 227,
  [805] = 199,
  [806] = 229,
  [807] = 231,
  [808] = 233,
  [809] = 235,
  [810] = 241,
  [811] = 267,
  [812] = 228,
  [813] = 242,
  [814] = 251,
  [815] = 268,
  [816] = 272,
  [817] = 186,
  [818] = 273,
  [819] = 254,
  [820] = 258,
  [821] = 239,
  [822] = 274,
  [823] = 201,
  [824] = 234,
  [825] = 232,
  [826] = 168,
  [827] = 243,
  [828] = 246,
  [829] = 247,
  [830] = 204,
  [831] = 208,
  [832] = 252,
  [833] = 309,
  [834] = 360,
  [835] = 326,
  [836] = 359,
  [837] = 327,
  [838] = 300,
  [839] = 311,
  [840] = 328,
  [841] = 284,
  [842] = 279,
  [843] = 329,
  [844] = 330,
  [845] = 285,
  [846] = 305,
  [847] = 307,
  [848] = 308,
  [849] = 294,
  [850] = 331,
  [851] = 313,
  [852] = 314,
  [853] = 315,
  [854] = 316,
  [855] = 317,
  [856] = 318,
  [857] = 297,
  [858] = 302,
  [859] = 286,
  [860] = 289,
  [861] = 280,
  [862] = 332,
  [863] = 333,
  [864] = 334,
  [865] = 319,
  [866] = 295,
  [867] = 296,
  [868] = 281,
  [869] = 310,
  [870] = 306,
  [871] = 293,
  [872] = 335,
  [873] = 336,
  [874] = 337,
  [875] = 338,
  [876] = 301,
  [877] = 339,
  [878] = 283,
  [879] = 340,
  [880] = 320,
  [881] = 341,
  [882] = 342,
  [883] = 321,
  [884] = 343,
  [885] = 344,
  [886] = 345,
  [887] = 346,
  [888] = 347,
  [889] = 348,
  [890] = 322,
  [891] = 323,
  [892] = 324,
  [893] = 350,
  [894] = 351,
  [895] = 357,
  [896] = 298,
  [897] = 352,
  [898] = 303,
  [899] = 353,
  [900] = 354,
  [901] = 358,
  [902] = 325,
  [903] = 355,
  [904] = 356,
  [905] = 277,
  [906] = 299,
  [907] = 278,
  [908] = 304,
  [909] = 184,
  [910] = 183,
  [911] = 911,
  [912] = 911,
  [913] = 186,
  [914] = 911,
  [915] = 911,
  [916] = 911,
  [917] = 197,
  [918] = 223,
  [919] = 919,
  [920] = 269,
  [921] = 270,
  [922] = 236,
  [923] = 264,
  [924] = 265,
  [925] = 248,
  [926] = 261,
  [927] = 271,
  [928] = 237,
  [929] = 259,
  [930] = 245,
  [931] = 252,
  [932] = 203,
  [933] = 200,
  [934] = 205,
  [935] = 206,
  [936] = 267,
  [937] = 207,
  [938] = 209,
  [939] = 210,
  [940] = 266,
  [941] = 217,
  [942] = 238,
  [943] = 222,
  [944] = 268,
  [945] = 228,
  [946] = 234,
  [947] = 256,
  [948] = 246,
  [949] = 247,
  [950] = 239,
  [951] = 276,
  [952] = 199,
  [953] = 229,
  [954] = 231,
  [955] = 232,
  [956] = 240,
  [957] = 255,
  [958] = 233,
  [959] = 235,
  [960] = 241,
  [961] = 204,
  [962] = 230,
  [963] = 208,
  [964] = 253,
  [965] = 212,
  [966] = 249,
  [967] = 258,
  [968] = 214,
  [969] = 215,
  [970] = 216,
  [971] = 262,
  [972] = 242,
  [973] = 251,
  [974] = 254,
  [975] = 243,
  [976] = 227,
  [977] = 289,
  [978] = 978,
  [979] = 332,
  [980] = 333,
  [981] = 334,
  [982] = 982,
  [983] = 983,
  [984] = 335,
  [985] = 336,
  [986] = 337,
  [987] = 338,
  [988] = 339,
  [989] = 340,
  [990] = 341,
  [991] = 342,
  [992] = 911,
  [993] = 344,
  [994] = 345,
  [995] = 346,
  [996] = 347,
  [997] = 348,
  [998] = 257,
  [999] = 260,
  [1000] = 263,
  [1001] = 350,
  [1002] = 351,
  [1003] = 352,
  [1004] = 294,
  [1005] = 353,
  [1006] = 354,
  [1007] = 355,
  [1008] = 356,
  [1009] = 277,
  [1010] = 278,
  [1011] = 272,
  [1012] = 273,
  [1013] = 302,
  [1014] = 306,
  [1015] = 359,
  [1016] = 279,
  [1017] = 274,
  [1018] = 303,
  [1019] = 304,
  [1020] = 305,
  [1021] = 295,
  [1022] = 280,
  [1023] = 360,
  [1024] = 307,
  [1025] = 911,
  [1026] = 308,
  [1027] = 309,
  [1028] = 286,
  [1029] = 1029,
  [1030] = 310,
  [1031] = 293,
  [1032] = 301,
  [1033] = 201,
  [1034] = 311,
  [1035] = 296,
  [1036] = 358,
  [1037] = 297,
  [1038] = 314,
  [1039] = 315,
  [1040] = 316,
  [1041] = 317,
  [1042] = 318,
  [1043] = 281,
  [1044] = 313,
  [1045] = 319,
  [1046] = 1046,
  [1047] = 357,
  [1048] = 283,
  [1049] = 320,
  [1050] = 321,
  [1051] = 322,
  [1052] = 323,
  [1053] = 324,
  [1054] = 325,
  [1055] = 326,
  [1056] = 327,
  [1057] = 328,
  [1058] = 284,
  [1059] = 329,
  [1060] = 330,
  [1061] = 285,
  [1062] = 218,
  [1063] = 219,
  [1064] = 220,
  [1065] = 331,
  [1066] = 298,
  [1067] = 911,
  [1068] = 911,
  [1069] = 911,
  [1070] = 911,
  [1071] = 343,
  [1072] = 1072,
  [1073] = 919,
  [1074] = 1074,
  [1075] = 1075,
  [1076] = 179,
  [1077] = 180,
  [1078] = 188,
  [1079] = 183,
  [1080] = 184,
  [1081] = 186,
  [1082] = 187,
  [1083] = 1083,
  [1084] = 1084,
  [1085] = 197,
  [1086] = 1084,
  [1087] = 1084,
  [1088] = 1084,
  [1089] = 1084,
  [1090] = 200,
  [1091] = 203,
  [1092] = 238,
  [1093] = 218,
  [1094] = 219,
  [1095] = 220,
  [1096] = 276,
  [1097] = 257,
  [1098] = 260,
  [1099] = 263,
  [1100] = 272,
  [1101] = 273,
  [1102] = 274,
  [1103] = 183,
  [1104] = 184,
  [1105] = 258,
  [1106] = 239,
  [1107] = 186,
  [1108] = 204,
  [1109] = 208,
  [1110] = 212,
  [1111] = 214,
  [1112] = 215,
  [1113] = 216,
  [1114] = 222,
  [1115] = 223,
  [1116] = 228,
  [1117] = 242,
  [1118] = 251,
  [1119] = 254,
  [1120] = 269,
  [1121] = 270,
  [1122] = 240,
  [1123] = 1083,
  [1124] = 206,
  [1125] = 207,
  [1126] = 209,
  [1127] = 210,
  [1128] = 217,
  [1129] = 227,
  [1130] = 199,
  [1131] = 243,
  [1132] = 246,
  [1133] = 247,
  [1134] = 237,
  [1135] = 249,
  [1136] = 229,
  [1137] = 205,
  [1138] = 231,
  [1139] = 233,
  [1140] = 253,
  [1141] = 235,
  [1142] = 255,
  [1143] = 1075,
  [1144] = 256,
  [1145] = 241,
  [1146] = 1074,
  [1147] = 259,
  [1148] = 261,
  [1149] = 262,
  [1150] = 264,
  [1151] = 265,
  [1152] = 266,
  [1153] = 267,
  [1154] = 230,
  [1155] = 234,
  [1156] = 268,
  [1157] = 271,
  [1158] = 201,
  [1159] = 232,
  [1160] = 236,
  [1161] = 245,
  [1162] = 252,
  [1163] = 248,
  [1164] = 1164,
  [1165] = 1165,
  [1166] = 1166,
  [1167] = 1165,
  [1168] = 1166,
  [1169] = 1164,
  [1170] = 1166,
  [1171] = 1166,
  [1172] = 1165,
  [1173] = 1165,
  [1174] = 1164,
  [1175] = 1164,
  [1176] = 1164,
  [1177] = 1166,
  [1178] = 1165,
  [1179] = 1179,
  [1180] = 1179,
  [1181] = 1179,
  [1182] = 1179,
  [1183] = 1179,
  [1184] = 1184,
  [1185] = 1185,
  [1186] = 1184,
  [1187] = 1185,
  [1188] = 1184,
  [1189] = 1189,
  [1190] = 1190,
  [1191] = 1190,
  [1192] = 1184,
  [1193] = 1193,
  [1194] = 1184,
  [1195] = 1190,
  [1196] = 1185,
  [1197] = 1197,
  [1198] = 1198,
  [1199] = 1199,
  [1200] = 1200,
  [1201] = 1200,
  [1202] = 1200,
  [1203] = 1199,
  [1204] = 1197,
  [1205] = 1197,
  [1206] = 1200,
  [1207] = 1199,
  [1208] = 1199,
  [1209] = 1199,
  [1210] = 1197,
  [1211] = 1200,
  [1212] = 1197,
  [1213] = 1213,
  [1214] = 1213,
  [1215] = 1215,
  [1216] = 1216,
  [1217] = 1217,
  [1218] = 1218,
  [1219] = 1219,
  [1220] = 1220,
  [1221] = 1221,
  [1222] = 1219,
  [1223] = 1223,
  [1224] = 1213,
  [1225] = 1216,
  [1226] = 1223,
  [1227] = 1215,
  [1228] = 1228,
  [1229] = 1215,
  [1230] = 1218,
  [1231] = 1218,
  [1232] = 1218,
  [1233] = 1223,
  [1234] = 1221,
  [1235] = 1219,
  [1236] = 1218,
  [1237] = 1220,
  [1238] = 1216,
  [1239] = 1223,
  [1240] = 1215,
  [1241] = 1221,
  [1242] = 1216,
  [1243] = 1219,
  [1244] = 1217,
  [1245] = 1220,
  [1246] = 1220,
  [1247] = 1221,
  [1248] = 1221,
  [1249] = 1219,
  [1250] = 1215,
  [1251] = 1251,
  [1252] = 1223,
  [1253] = 1216,
  [1254] = 1217,
  [1255] = 1220,
  [1256] = 1256,
  [1257] = 1257,
  [1258] = 1258,
  [1259] = 1259,
  [1260] = 1260,
  [1261] = 1261,
  [1262] = 1262,
  [1263] = 1262,
  [1264] = 1258,
  [1265] = 1258,
  [1266] = 1266,
  [1267] = 1198,
  [1268] = 1268,
  [1269] = 1262,
  [1270] = 1257,
  [1271] = 1271,
  [1272] = 1257,
  [1273] = 1273,
  [1274] = 1256,
  [1275] = 1268,
  [1276] = 1276,
  [1277] = 1277,
  [1278] = 1260,
  [1279] = 1261,
  [1280] = 1266,
  [1281] = 1276,
  [1282] = 1257,
  [1283] = 1262,
  [1284] = 1256,
  [1285] = 1268,
  [1286] = 1276,
  [1287] = 1262,
  [1288] = 1258,
  [1289] = 1289,
  [1290] = 1256,
  [1291] = 1277,
  [1292] = 1260,
  [1293] = 1261,
  [1294] = 1277,
  [1295] = 1260,
  [1296] = 1273,
  [1297] = 1261,
  [1298] = 1266,
  [1299] = 1266,
  [1300] = 1276,
  [1301] = 1257,
  [1302] = 1273,
  [1303] = 1256,
  [1304] = 1268,
  [1305] = 1276,
  [1306] = 1273,
  [1307] = 1266,
  [1308] = 1277,
  [1309] = 1258,
  [1310] = 1268,
  [1311] = 1277,
  [1312] = 1260,
  [1313] = 1261,
  [1314] = 1273,
  [1315] = 1315,
  [1316] = 1316,
  [1317] = 1317,
  [1318] = 1316,
  [1319] = 1315,
  [1320] = 1315,
  [1321] = 1321,
  [1322] = 1322,
  [1323] = 1321,
  [1324] = 1321,
  [1325] = 1316,
  [1326] = 1326,
  [1327] = 1315,
  [1328] = 1316,
  [1329] = 1321,
  [1330] = 1315,
  [1331] = 1321,
  [1332] = 1316,
  [1333] = 1333,
  [1334] = 1334,
  [1335] = 1335,
  [1336] = 1336,
  [1337] = 1337,
  [1338] = 1338,
  [1339] = 1339,
  [1340] = 1340,
  [1341] = 1341,
  [1342] = 1342,
  [1343] = 1343,
  [1344] = 1344,
  [1345] = 1345,
  [1346] = 1334,
  [1347] = 1335,
  [1348] = 1348,
  [1349] = 1349,
  [1350] = 1350,
  [1351] = 1351,
  [1352] = 1348,
  [1353] = 1349,
  [1354] = 1354,
  [1355] = 1337,
  [1356] = 1351,
  [1357] = 1354,
  [1358] = 1337,
  [1359] = 1359,
  [1360] = 1360,
  [1361] = 1336,
  [1362] = 1344,
  [1363] = 1338,
  [1364] = 1339,
  [1365] = 1340,
  [1366] = 1341,
  [1367] = 1367,
  [1368] = 1342,
  [1369] = 1343,
  [1370] = 1345,
  [1371] = 1334,
  [1372] = 1335,
  [1373] = 1348,
  [1374] = 1349,
  [1375] = 1359,
  [1376] = 1360,
  [1377] = 1351,
  [1378] = 1336,
  [1379] = 1344,
  [1380] = 1322,
  [1381] = 1354,
  [1382] = 1337,
  [1383] = 1383,
  [1384] = 1338,
  [1385] = 1339,
  [1386] = 1340,
  [1387] = 1341,
  [1388] = 1367,
  [1389] = 1359,
  [1390] = 1360,
  [1391] = 1342,
  [1392] = 1336,
  [1393] = 1344,
  [1394] = 1343,
  [1395] = 1359,
  [1396] = 1360,
  [1397] = 1338,
  [1398] = 1339,
  [1399] = 1340,
  [1400] = 1341,
  [1401] = 1367,
  [1402] = 1342,
  [1403] = 1343,
  [1404] = 1345,
  [1405] = 1334,
  [1406] = 1335,
  [1407] = 1348,
  [1408] = 1345,
  [1409] = 1334,
  [1410] = 1335,
  [1411] = 1348,
  [1412] = 1349,
  [1413] = 1349,
  [1414] = 1351,
  [1415] = 1351,
  [1416] = 1338,
  [1417] = 1339,
  [1418] = 1340,
  [1419] = 1341,
  [1420] = 1367,
  [1421] = 1342,
  [1422] = 1343,
  [1423] = 1345,
  [1424] = 1354,
  [1425] = 1337,
  [1426] = 1354,
  [1427] = 1359,
  [1428] = 1360,
  [1429] = 1336,
  [1430] = 1344,
  [1431] = 1367,
  [1432] = 1432,
  [1433] = 1433,
  [1434] = 1434,
  [1435] = 1435,
  [1436] = 1436,
  [1437] = 1437,
  [1438] = 1438,
  [1439] = 1439,
  [1440] = 1440,
  [1441] = 1441,
  [1442] = 1442,
  [1443] = 1443,
  [1444] = 1444,
  [1445] = 1445,
  [1446] = 1445,
  [1447] = 1447,
  [1448] = 1448,
  [1449] = 1449,
  [1450] = 1450,
  [1451] = 183,
  [1452] = 1452,
  [1453] = 1453,
  [1454] = 1454,
  [1455] = 1455,
  [1456] = 1456,
  [1457] = 1457,
  [1458] = 1458,
  [1459] = 1459,
  [1460] = 1460,
  [1461] = 1461,
  [1462] = 1462,
  [1463] = 1463,
  [1464] = 1464,
  [1465] = 1465,
  [1466] = 1466,
  [1467] = 1467,
  [1468] = 1468,
  [1469] = 1469,
  [1470] = 1470,
  [1471] = 1471,
  [1472] = 1472,
  [1473] = 1473,
  [1474] = 1474,
  [1475] = 1475,
  [1476] = 1476,
  [1477] = 1477,
  [1478] = 1478,
  [1479] = 1479,
  [1480] = 1480,
  [1481] = 1481,
  [1482] = 1482,
  [1483] = 1483,
  [1484] = 1484,
  [1485] = 1485,
  [1486] = 1486,
  [1487] = 1487,
  [1488] = 1488,
  [1489] = 1489,
  [1490] = 1490,
  [1491] = 1491,
  [1492] = 1492,
  [1493] = 1493,
  [1494] = 1494,
  [1495] = 1495,
  [1496] = 184,
  [1497] = 1497,
  [1498] = 1498,
  [1499] = 1499,
  [1500] = 1500,
  [1501] = 1501,
  [1502] = 1502,
  [1503] = 1460,
  [1504] = 1504,
  [1505] = 186,
  [1506] = 1506,
  [1507] = 1507,
  [1508] = 1508,
  [1509] = 1509,
  [1510] = 1510,
  [1511] = 1511,
  [1512] = 1512,
  [1513] = 1513,
  [1514] = 1514,
  [1515] = 1514,
  [1516] = 1516,
  [1517] = 1517,
  [1518] = 1518,
  [1519] = 1517,
  [1520] = 1520,
  [1521] = 1521,
  [1522] = 1517,
  [1523] = 1523,
  [1524] = 1518,
  [1525] = 1520,
  [1526] = 1518,
  [1527] = 1514,
  [1528] = 1520,
  [1529] = 1529,
  [1530] = 1530,
  [1531] = 1530,
  [1532] = 1529,
  [1533] = 1529,
  [1534] = 1529,
  [1535] = 1530,
  [1536] = 1529,
  [1537] = 1529,
  [1538] = 1529,
  [1539] = 1530,
  [1540] = 1529,
  [1541] = 1530,
  [1542] = 1529,
  [1543] = 1530,
  [1544] = 1530,
  [1545] = 1530,
  [1546] = 1530,
  [1547] = 1529,
  [1548] = 1529,
  [1549] = 1530,
  [1550] = 1530,
  [1551] = 1530,
  [1552] = 1529,
  [1553] = 1553,
  [1554] = 1554,
  [1555] = 1555,
  [1556] = 1556,
  [1557] = 1557,
  [1558] = 1558,
  [1559] = 1559,
  [1560] = 1560,
  [1561] = 1561,
  [1562] = 1562,
  [1563] = 1563,
  [1564] = 1564,
  [1565] = 1565,
  [1566] = 1566,
  [1567] = 1567,
  [1568] = 183,
  [1569] = 184,
  [1570] = 1570,
  [1571] = 186,
  [1572] = 1454,
  [1573] = 1573,
  [1574] = 1574,
  [1575] = 1574,
  [1576] = 1576,
  [1577] = 1577,
  [1578] = 1577,
  [1579] = 1579,
  [1580] = 1579,
  [1581] = 1581,
  [1582] = 1579,
  [1583] = 1579,
  [1584] = 1584,
  [1585] = 1585,
  [1586] = 1579,
  [1587] = 1579,
  [1588] = 1584,
  [1589] = 1585,
  [1590] = 1590,
  [1591] = 1590,
  [1592] = 1554,
  [1593] = 1432,
  [1594] = 1579,
  [1595] = 1585,
  [1596] = 1579,
  [1597] = 1579,
  [1598] = 1585,
  [1599] = 1433,
  [1600] = 1585,
  [1601] = 1434,
  [1602] = 1577,
  [1603] = 1581,
  [1604] = 1585,
  [1605] = 1577,
  [1606] = 1579,
  [1607] = 1585,
  [1608] = 1579,
  [1609] = 1585,
  [1610] = 1579,
  [1611] = 1554,
  [1612] = 1612,
  [1613] = 1577,
  [1614] = 1614,
  [1615] = 1436,
  [1616] = 1616,
  [1617] = 1437,
  [1618] = 1432,
  [1619] = 1444,
  [1620] = 1435,
  [1621] = 1433,
  [1622] = 1438,
  [1623] = 1441,
  [1624] = 1440,
  [1625] = 1442,
  [1626] = 1439,
  [1627] = 1434,
  [1628] = 1479,
  [1629] = 1449,
  [1630] = 1473,
  [1631] = 1472,
  [1632] = 1474,
  [1633] = 1469,
  [1634] = 1461,
  [1635] = 1462,
  [1636] = 1480,
  [1637] = 1482,
  [1638] = 1512,
  [1639] = 1475,
  [1640] = 1483,
  [1641] = 1465,
  [1642] = 1448,
  [1643] = 1457,
  [1644] = 1458,
  [1645] = 1511,
  [1646] = 1491,
  [1647] = 1493,
  [1648] = 1487,
  [1649] = 1488,
  [1650] = 1477,
  [1651] = 1490,
  [1652] = 1501,
  [1653] = 1507,
  [1654] = 1485,
  [1655] = 1486,
  [1656] = 1470,
  [1657] = 1468,
  [1658] = 1506,
  [1659] = 1502,
  [1660] = 1492,
  [1661] = 1471,
  [1662] = 1499,
  [1663] = 1450,
  [1664] = 1504,
  [1665] = 1464,
  [1666] = 1478,
  [1667] = 1445,
  [1668] = 1500,
  [1669] = 1506,
  [1670] = 1468,
  [1671] = 1478,
  [1672] = 1504,
  [1673] = 1507,
  [1674] = 1674,
  [1675] = 1675,
  [1676] = 1676,
  [1677] = 1677,
  [1678] = 1678,
  [1679] = 1679,
  [1680] = 1680,
  [1681] = 1681,
  [1682] = 1682,
  [1683] = 1683,
  [1684] = 1675,
  [1685] = 1674,
  [1686] = 1686,
  [1687] = 1687,
  [1688] = 1676,
  [1689] = 1677,
  [1690] = 1683,
  [1691] = 1675,
  [1692] = 1674,
  [1693] = 1676,
  [1694] = 1677,
  [1695] = 1678,
  [1696] = 1679,
  [1697] = 1680,
  [1698] = 1681,
  [1699] = 1682,
  [1700] = 1440,
  [1701] = 1678,
  [1702] = 1679,
  [1703] = 1686,
  [1704] = 1686,
  [1705] = 1687,
  [1706] = 1680,
  [1707] = 1681,
  [1708] = 1682,
  [1709] = 1687,
  [1710] = 1440,
  [1711] = 1683,
  [1712] = 1712,
  [1713] = 1712,
  [1714] = 1712,
  [1715] = 1715,
  [1716] = 1715,
  [1717] = 1717,
  [1718] = 1718,
  [1719] = 1717,
  [1720] = 1715,
  [1721] = 1712,
  [1722] = 1718,
  [1723] = 1718,
  [1724] = 1712,
  [1725] = 1717,
  [1726] = 1726,
  [1727] = 1445,
  [1728] = 1728,
  [1729] = 1729,
  [1730] = 1730,
  [1731] = 1731,
  [1732] = 1732,
  [1733] = 1733,
  [1734] = 1555,
  [1735] = 1735,
  [1736] = 1736,
  [1737] = 1737,
  [1738] = 1732,
  [1739] = 1737,
  [1740] = 1729,
  [1741] = 1557,
  [1742] = 1732,
  [1743] = 1558,
  [1744] = 1730,
  [1745] = 1560,
  [1746] = 1570,
  [1747] = 1563,
  [1748] = 1564,
  [1749] = 1565,
  [1750] = 1567,
  [1751] = 1559,
  [1752] = 1562,
  [1753] = 1737,
  [1754] = 1731,
  [1755] = 1755,
  [1756] = 1755,
  [1757] = 1731,
  [1758] = 1566,
  [1759] = 1556,
  [1760] = 1755,
  [1761] = 1729,
  [1762] = 1762,
  [1763] = 1561,
  [1764] = 1730,
  [1765] = 1765,
  [1766] = 1553,
  [1767] = 1767,
  [1768] = 1768,
  [1769] = 1769,
  [1770] = 1769,
  [1771] = 1771,
  [1772] = 1772,
  [1773] = 1773,
  [1774] = 1771,
  [1775] = 1481,
  [1776] = 1453,
  [1777] = 1508,
  [1778] = 1510,
  [1779] = 1779,
  [1780] = 1455,
  [1781] = 1781,
  [1782] = 1782,
  [1783] = 1767,
  [1784] = 1494,
  [1785] = 1495,
  [1786] = 1786,
  [1787] = 1781,
  [1788] = 1786,
  [1789] = 1789,
  [1790] = 1790,
  [1791] = 1779,
  [1792] = 1771,
  [1793] = 1793,
  [1794] = 1786,
  [1795] = 1452,
  [1796] = 1772,
  [1797] = 1773,
  [1798] = 1772,
  [1799] = 1773,
  [1800] = 1456,
  [1801] = 1466,
  [1802] = 1782,
  [1803] = 1767,
  [1804] = 1782,
  [1805] = 1805,
  [1806] = 1789,
  [1807] = 1790,
  [1808] = 1793,
  [1809] = 1809,
  [1810] = 1768,
  [1811] = 1779,
  [1812] = 1781,
  [1813] = 1789,
  [1814] = 1790,
  [1815] = 1793,
  [1816] = 1809,
  [1817] = 1768,
  [1818] = 1809,
  [1819] = 1769,
  [1820] = 1820,
  [1821] = 1820,
  [1822] = 1822,
  [1823] = 1822,
  [1824] = 1822,
  [1825] = 1445,
  [1826] = 1826,
  [1827] = 1827,
  [1828] = 1828,
  [1829] = 1829,
  [1830] = 1830,
  [1831] = 1831,
  [1832] = 1832,
  [1833] = 1826,
  [1834] = 1834,
  [1835] = 1828,
  [1836] = 1836,
  [1837] = 1837,
  [1838] = 1831,
  [1839] = 1839,
  [1840] = 1836,
  [1841] = 1841,
  [1842] = 1837,
  [1843] = 1839,
  [1844] = 1841,
  [1845] = 1845,
  [1846] = 1845,
  [1847] = 1847,
  [1848] = 1848,
  [1849] = 1848,
  [1850] = 1834,
  [1851] = 1851,
  [1852] = 1832,
  [1853] = 1853,
  [1854] = 1847,
  [1855] = 1853,
  [1856] = 1851,
  [1857] = 1832,
  [1858] = 1827,
  [1859] = 1829,
  [1860] = 1834,
  [1861] = 1828,
  [1862] = 1827,
  [1863] = 1831,
  [1864] = 1851,
  [1865] = 1826,
  [1866] = 1829,
  [1867] = 1836,
  [1868] = 1837,
  [1869] = 1839,
  [1870] = 1841,
  [1871] = 1845,
  [1872] = 1848,
  [1873] = 1830,
  [1874] = 1847,
  [1875] = 1853,
  [1876] = 1830,
  [1877] = 1877,
  [1878] = 1878,
  [1879] = 1877,
  [1880] = 1878,
  [1881] = 1881,
  [1882] = 1882,
  [1883] = 1877,
  [1884] = 1877,
  [1885] = 1885,
  [1886] = 1886,
  [1887] = 1887,
  [1888] = 1888,
  [1889] = 1889,
  [1890] = 1878,
  [1891] = 1878,
  [1892] = 1892,
  [1893] = 1877,
  [1894] = 1878,
  [1895] = 1895,
  [1896] = 1896,
  [1897] = 1897,
  [1898] = 1898,
  [1899] = 1899,
  [1900] = 1558,
  [1901] = 1553,
  [1902] = 1560,
  [1903] = 1570,
  [1904] = 1563,
  [1905] = 1564,
  [1906] = 1565,
  [1907] = 1567,
  [1908] = 1559,
  [1909] = 1562,
  [1910] = 1566,
  [1911] = 1556,
  [1912] = 1555,
  [1913] = 1561,
  [1914] = 1432,
  [1915] = 1433,
  [1916] = 1434,
  [1917] = 1557,
  [1918] = 1454,
  [1919] = 1919,
  [1920] = 1920,
  [1921] = 1921,
  [1922] = 258,
  [1923] = 239,
  [1924] = 1919,
  [1925] = 1925,
  [1926] = 1926,
  [1927] = 1920,
  [1928] = 1494,
  [1929] = 1495,
  [1930] = 1930,
  [1931] = 1930,
  [1932] = 1932,
  [1933] = 1926,
  [1934] = 1934,
  [1935] = 1935,
  [1936] = 1466,
  [1937] = 1937,
  [1938] = 1921,
  [1939] = 1921,
  [1940] = 1882,
  [1941] = 1941,
  [1942] = 1508,
  [1943] = 204,
  [1944] = 208,
  [1945] = 212,
  [1946] = 214,
  [1947] = 215,
  [1948] = 216,
  [1949] = 1510,
  [1950] = 1935,
  [1951] = 1935,
  [1952] = 1952,
  [1953] = 1925,
  [1954] = 1954,
  [1955] = 1932,
  [1956] = 1956,
  [1957] = 1919,
  [1958] = 222,
  [1959] = 223,
  [1960] = 228,
  [1961] = 242,
  [1962] = 251,
  [1963] = 254,
  [1964] = 1920,
  [1965] = 1925,
  [1966] = 1926,
  [1967] = 1453,
  [1968] = 1455,
  [1969] = 1956,
  [1970] = 1970,
  [1971] = 1930,
  [1972] = 1937,
  [1973] = 269,
  [1974] = 1956,
  [1975] = 1937,
  [1976] = 1452,
  [1977] = 1456,
  [1978] = 270,
  [1979] = 1954,
  [1980] = 1932,
  [1981] = 1481,
  [1982] = 1954,
  [1983] = 1983,
  [1984] = 1984,
  [1985] = 1985,
  [1986] = 1986,
  [1987] = 1987,
  [1988] = 1988,
  [1989] = 1989,
  [1990] = 358,
  [1991] = 1991,
  [1992] = 1992,
  [1993] = 1993,
  [1994] = 1994,
  [1995] = 1995,
  [1996] = 1996,
  [1997] = 1997,
  [1998] = 1998,
  [1999] = 1999,
  [2000] = 2000,
  [2001] = 1984,
  [2002] = 1985,
  [2003] = 1986,
  [2004] = 1987,
  [2005] = 1988,
  [2006] = 1989,
  [2007] = 1991,
  [2008] = 1468,
  [2009] = 2009,
  [2010] = 1995,
  [2011] = 1994,
  [2012] = 1995,
  [2013] = 1996,
  [2014] = 2014,
  [2015] = 1994,
  [2016] = 1995,
  [2017] = 1994,
  [2018] = 1999,
  [2019] = 2019,
  [2020] = 1983,
  [2021] = 1984,
  [2022] = 1999,
  [2023] = 2000,
  [2024] = 1507,
  [2025] = 1989,
  [2026] = 1986,
  [2027] = 1991,
  [2028] = 281,
  [2029] = 1478,
  [2030] = 357,
  [2031] = 2014,
  [2032] = 2000,
  [2033] = 1506,
  [2034] = 2034,
  [2035] = 1985,
  [2036] = 1997,
  [2037] = 2014,
  [2038] = 2019,
  [2039] = 1983,
  [2040] = 2019,
  [2041] = 1983,
  [2042] = 2019,
  [2043] = 1983,
  [2044] = 1983,
  [2045] = 1983,
  [2046] = 1988,
  [2047] = 1987,
  [2048] = 1994,
  [2049] = 1998,
  [2050] = 2019,
  [2051] = 1995,
  [2052] = 2052,
  [2053] = 1996,
  [2054] = 1997,
  [2055] = 1998,
  [2056] = 1504,
  [2057] = 286,
  [2058] = 1556,
  [2059] = 312,
  [2060] = 183,
  [2061] = 184,
  [2062] = 186,
  [2063] = 1882,
  [2064] = 2064,
  [2065] = 2065,
  [2066] = 2066,
  [2067] = 2067,
  [2068] = 282,
  [2069] = 1440,
  [2070] = 2066,
  [2071] = 1952,
  [2072] = 1440,
  [2073] = 1952,
  [2074] = 2074,
  [2075] = 349,
  [2076] = 2076,
  [2077] = 2077,
  [2078] = 1561,
  [2079] = 2079,
  [2080] = 2080,
  [2081] = 183,
  [2082] = 184,
  [2083] = 186,
  [2084] = 2080,
  [2085] = 919,
  [2086] = 1557,
  [2087] = 1558,
  [2088] = 1553,
  [2089] = 1560,
  [2090] = 1570,
  [2091] = 1563,
  [2092] = 1564,
  [2093] = 1565,
  [2094] = 1567,
  [2095] = 1559,
  [2096] = 1562,
  [2097] = 1566,
  [2098] = 1555,
  [2099] = 1072,
  [2100] = 2100,
  [2101] = 2101,
  [2102] = 2100,
  [2103] = 269,
  [2104] = 2104,
  [2105] = 2105,
  [2106] = 2106,
  [2107] = 2107,
  [2108] = 2108,
  [2109] = 2109,
  [2110] = 2110,
  [2111] = 2111,
  [2112] = 2112,
  [2113] = 270,
  [2114] = 358,
  [2115] = 281,
  [2116] = 313,
  [2117] = 313,
  [2118] = 357,
  [2119] = 286,
  [2120] = 2120,
  [2121] = 2121,
  [2122] = 306,
  [2123] = 280,
  [2124] = 293,
  [2125] = 301,
  [2126] = 2067,
  [2127] = 2127,
  [2128] = 293,
  [2129] = 301,
  [2130] = 285,
  [2131] = 201,
  [2132] = 2132,
  [2133] = 2133,
  [2134] = 251,
  [2135] = 2067,
  [2136] = 272,
  [2137] = 2105,
  [2138] = 2138,
  [2139] = 273,
  [2140] = 2140,
  [2141] = 283,
  [2142] = 283,
  [2143] = 285,
  [2144] = 218,
  [2145] = 254,
  [2146] = 220,
  [2147] = 201,
  [2148] = 218,
  [2149] = 219,
  [2150] = 220,
  [2151] = 274,
  [2152] = 257,
  [2153] = 260,
  [2154] = 263,
  [2155] = 272,
  [2156] = 273,
  [2157] = 274,
  [2158] = 258,
  [2159] = 239,
  [2160] = 2120,
  [2161] = 2127,
  [2162] = 2162,
  [2163] = 2163,
  [2164] = 204,
  [2165] = 208,
  [2166] = 2120,
  [2167] = 2167,
  [2168] = 2127,
  [2169] = 212,
  [2170] = 2120,
  [2171] = 2127,
  [2172] = 2133,
  [2173] = 2120,
  [2174] = 2127,
  [2175] = 2127,
  [2176] = 214,
  [2177] = 2127,
  [2178] = 215,
  [2179] = 2127,
  [2180] = 2180,
  [2181] = 1454,
  [2182] = 216,
  [2183] = 2183,
  [2184] = 2184,
  [2185] = 2138,
  [2186] = 2186,
  [2187] = 2167,
  [2188] = 306,
  [2189] = 2189,
  [2190] = 280,
  [2191] = 2183,
  [2192] = 2180,
  [2193] = 242,
  [2194] = 222,
  [2195] = 223,
  [2196] = 257,
  [2197] = 260,
  [2198] = 263,
  [2199] = 228,
  [2200] = 2121,
  [2201] = 219,
  [2202] = 289,
  [2203] = 2203,
  [2204] = 289,
  [2205] = 2205,
  [2206] = 2066,
  [2207] = 2207,
  [2208] = 358,
  [2209] = 281,
  [2210] = 284,
  [2211] = 357,
  [2212] = 2212,
  [2213] = 291,
  [2214] = 2214,
  [2215] = 186,
  [2216] = 286,
  [2217] = 292,
  [2218] = 2218,
  [2219] = 2067,
  [2220] = 2220,
  [2221] = 2221,
  [2222] = 2222,
  [2223] = 2223,
  [2224] = 2224,
  [2225] = 2225,
  [2226] = 2226,
  [2227] = 2227,
  [2228] = 2228,
  [2229] = 2229,
  [2230] = 2230,
  [2231] = 183,
  [2232] = 2230,
  [2233] = 2233,
  [2234] = 290,
  [2235] = 2235,
  [2236] = 2236,
  [2237] = 288,
  [2238] = 184,
  [2239] = 2239,
  [2240] = 2240,
  [2241] = 2241,
  [2242] = 284,
  [2243] = 219,
  [2244] = 2244,
  [2245] = 2245,
  [2246] = 2246,
  [2247] = 263,
  [2248] = 2248,
  [2249] = 2249,
  [2250] = 2111,
  [2251] = 2251,
  [2252] = 2140,
  [2253] = 257,
  [2254] = 272,
  [2255] = 273,
  [2256] = 2112,
  [2257] = 2184,
  [2258] = 306,
  [2259] = 280,
  [2260] = 2101,
  [2261] = 274,
  [2262] = 2140,
  [2263] = 283,
  [2264] = 258,
  [2265] = 239,
  [2266] = 204,
  [2267] = 2245,
  [2268] = 2163,
  [2269] = 2269,
  [2270] = 2121,
  [2271] = 2104,
  [2272] = 2272,
  [2273] = 208,
  [2274] = 2274,
  [2275] = 2275,
  [2276] = 2276,
  [2277] = 2106,
  [2278] = 212,
  [2279] = 214,
  [2280] = 2180,
  [2281] = 2184,
  [2282] = 2282,
  [2283] = 2107,
  [2284] = 215,
  [2285] = 313,
  [2286] = 2286,
  [2287] = 216,
  [2288] = 2249,
  [2289] = 2108,
  [2290] = 2183,
  [2291] = 2274,
  [2292] = 289,
  [2293] = 293,
  [2294] = 301,
  [2295] = 2110,
  [2296] = 2100,
  [2297] = 2186,
  [2298] = 284,
  [2299] = 2109,
  [2300] = 285,
  [2301] = 2110,
  [2302] = 222,
  [2303] = 2303,
  [2304] = 2269,
  [2305] = 223,
  [2306] = 228,
  [2307] = 2307,
  [2308] = 2104,
  [2309] = 2186,
  [2310] = 2105,
  [2311] = 201,
  [2312] = 2106,
  [2313] = 2133,
  [2314] = 2167,
  [2315] = 2107,
  [2316] = 242,
  [2317] = 2272,
  [2318] = 2318,
  [2319] = 2282,
  [2320] = 2286,
  [2321] = 2138,
  [2322] = 2322,
  [2323] = 2275,
  [2324] = 251,
  [2325] = 254,
  [2326] = 269,
  [2327] = 2248,
  [2328] = 2111,
  [2329] = 270,
  [2330] = 2330,
  [2331] = 260,
  [2332] = 2112,
  [2333] = 2333,
  [2334] = 2163,
  [2335] = 2276,
  [2336] = 2336,
  [2337] = 2303,
  [2338] = 2338,
  [2339] = 2108,
  [2340] = 2340,
  [2341] = 218,
  [2342] = 2109,
  [2343] = 220,
  [2344] = 2330,
  [2345] = 2101,
  [2346] = 2203,
  [2347] = 2203,
  [2348] = 2348,
  [2349] = 2349,
  [2350] = 2228,
  [2351] = 2224,
  [2352] = 2220,
  [2353] = 2353,
  [2354] = 2354,
  [2355] = 288,
  [2356] = 2356,
  [2357] = 2357,
  [2358] = 1074,
  [2359] = 2229,
  [2360] = 2241,
  [2361] = 2240,
  [2362] = 2218,
  [2363] = 2236,
  [2364] = 2223,
  [2365] = 2239,
  [2366] = 2226,
  [2367] = 2205,
  [2368] = 2368,
  [2369] = 2222,
  [2370] = 2225,
  [2371] = 2235,
  [2372] = 2372,
  [2373] = 290,
  [2374] = 2227,
  [2375] = 291,
  [2376] = 2226,
  [2377] = 2227,
  [2378] = 2220,
  [2379] = 2218,
  [2380] = 290,
  [2381] = 2381,
  [2382] = 291,
  [2383] = 292,
  [2384] = 2229,
  [2385] = 2228,
  [2386] = 2235,
  [2387] = 2240,
  [2388] = 2221,
  [2389] = 2236,
  [2390] = 2222,
  [2391] = 2239,
  [2392] = 2221,
  [2393] = 2233,
  [2394] = 2233,
  [2395] = 2205,
  [2396] = 2241,
  [2397] = 288,
  [2398] = 292,
  [2399] = 2225,
  [2400] = 2224,
  [2401] = 1075,
  [2402] = 2223,
  [2403] = 2245,
  [2404] = 2286,
  [2405] = 2405,
  [2406] = 2406,
  [2407] = 2407,
  [2408] = 2275,
  [2409] = 2249,
  [2410] = 2410,
  [2411] = 2318,
  [2412] = 2269,
  [2413] = 2407,
  [2414] = 2410,
  [2415] = 2330,
  [2416] = 2405,
  [2417] = 2303,
  [2418] = 2406,
  [2419] = 2419,
  [2420] = 2276,
  [2421] = 2272,
  [2422] = 2274,
  [2423] = 2248,
  [2424] = 2282,
  [2425] = 2425,
  [2426] = 2426,
  [2427] = 2427,
  [2428] = 2428,
  [2429] = 2429,
  [2430] = 2426,
  [2431] = 2431,
  [2432] = 2428,
  [2433] = 2426,
  [2434] = 2431,
  [2435] = 2435,
  [2436] = 2429,
  [2437] = 2426,
  [2438] = 2429,
  [2439] = 2431,
  [2440] = 2426,
  [2441] = 2431,
  [2442] = 2427,
  [2443] = 2428,
  [2444] = 2427,
  [2445] = 2435,
  [2446] = 2429,
  [2447] = 2427,
  [2448] = 2431,
  [2449] = 2426,
  [2450] = 2428,
  [2451] = 2429,
  [2452] = 2425,
  [2453] = 2431,
  [2454] = 2428,
  [2455] = 2431,
  [2456] = 2429,
  [2457] = 2428,
  [2458] = 2425,
  [2459] = 2435,
  [2460] = 2427,
  [2461] = 2427,
  [2462] = 2431,
  [2463] = 2426,
  [2464] = 2428,
  [2465] = 2435,
  [2466] = 2435,
  [2467] = 2425,
  [2468] = 2428,
  [2469] = 2425,
  [2470] = 2470,
  [2471] = 2426,
  [2472] = 2429,
  [2473] = 2435,
  [2474] = 2425,
  [2475] = 2427,
  [2476] = 2425,
  [2477] = 2427,
  [2478] = 2435,
  [2479] = 2435,
  [2480] = 2429,
  [2481] = 2481,
  [2482] = 2482,
  [2483] = 2483,
  [2484] = 1558,
  [2485] = 1553,
  [2486] = 1560,
  [2487] = 2487,
  [2488] = 1570,
  [2489] = 1563,
  [2490] = 1564,
  [2491] = 1565,
  [2492] = 2492,
  [2493] = 1567,
  [2494] = 1559,
  [2495] = 1562,
  [2496] = 1566,
  [2497] = 1556,
  [2498] = 2498,
  [2499] = 2499,
  [2500] = 2500,
  [2501] = 2501,
  [2502] = 2502,
  [2503] = 2503,
  [2504] = 2504,
  [2505] = 1555,
  [2506] = 1561,
  [2507] = 2507,
  [2508] = 2508,
  [2509] = 2509,
  [2510] = 2510,
  [2511] = 2511,
  [2512] = 2512,
  [2513] = 1557,
  [2514] = 2514,
  [2515] = 2515,
  [2516] = 2516,
  [2517] = 2516,
  [2518] = 2518,
  [2519] = 2519,
  [2520] = 2520,
  [2521] = 2521,
  [2522] = 2522,
  [2523] = 2523,
  [2524] = 2521,
  [2525] = 2525,
  [2526] = 2526,
  [2527] = 2527,
  [2528] = 2528,
  [2529] = 2529,
  [2530] = 2516,
  [2531] = 2529,
  [2532] = 2523,
  [2533] = 2515,
  [2534] = 2514,
  [2535] = 2535,
  [2536] = 2536,
  [2537] = 2528,
  [2538] = 2529,
  [2539] = 2514,
  [2540] = 2519,
  [2541] = 2498,
  [2542] = 168,
  [2543] = 2543,
  [2544] = 2544,
  [2545] = 2545,
  [2546] = 2546,
  [2547] = 2547,
  [2548] = 2548,
  [2549] = 2546,
  [2550] = 2550,
  [2551] = 2551,
  [2552] = 2552,
  [2553] = 2550,
  [2554] = 2551,
  [2555] = 2545,
  [2556] = 2548,
  [2557] = 2557,
  [2558] = 2547,
  [2559] = 2557,
  [2560] = 2546,
  [2561] = 2547,
  [2562] = 2562,
  [2563] = 2546,
  [2564] = 2547,
  [2565] = 2545,
  [2566] = 2514,
  [2567] = 2546,
  [2568] = 2547,
  [2569] = 2569,
  [2570] = 2546,
  [2571] = 2547,
  [2572] = 2550,
  [2573] = 2551,
  [2574] = 2546,
  [2575] = 2547,
  [2576] = 2546,
  [2577] = 2547,
  [2578] = 2548,
  [2579] = 2550,
  [2580] = 2551,
  [2581] = 2548,
  [2582] = 2545,
  [2583] = 2550,
  [2584] = 2551,
  [2585] = 2548,
  [2586] = 2545,
  [2587] = 1728,
  [2588] = 2588,
  [2589] = 2589,
  [2590] = 2590,
  [2591] = 2591,
  [2592] = 2592,
  [2593] = 2593,
  [2594] = 2594,
  [2595] = 2595,
  [2596] = 2596,
  [2597] = 2597,
  [2598] = 2598,
  [2599] = 2599,
  [2600] = 2600,
  [2601] = 2601,
  [2602] = 2602,
  [2603] = 2603,
  [2604] = 358,
  [2605] = 281,
  [2606] = 357,
  [2607] = 286,
  [2608] = 289,
  [2609] = 1733,
  [2610] = 2610,
  [2611] = 2611,
  [2612] = 2612,
  [2613] = 2613,
  [2614] = 2614,
  [2615] = 2615,
  [2616] = 2616,
  [2617] = 2617,
  [2618] = 284,
  [2619] = 2619,
  [2620] = 2620,
  [2621] = 2621,
  [2622] = 2622,
  [2623] = 2623,
  [2624] = 2624,
  [2625] = 2625,
  [2626] = 2626,
  [2627] = 2623,
  [2628] = 2628,
  [2629] = 2629,
  [2630] = 1557,
  [2631] = 1558,
  [2632] = 1553,
  [2633] = 1560,
  [2634] = 1570,
  [2635] = 1563,
  [2636] = 1564,
  [2637] = 1565,
  [2638] = 1567,
  [2639] = 1559,
  [2640] = 1562,
  [2641] = 1566,
  [2642] = 1556,
  [2643] = 2643,
  [2644] = 2644,
  [2645] = 2645,
  [2646] = 2646,
  [2647] = 2647,
  [2648] = 1555,
  [2649] = 1561,
  [2650] = 2650,
  [2651] = 2651,
  [2652] = 2652,
  [2653] = 2653,
  [2654] = 2654,
  [2655] = 2655,
  [2656] = 2655,
  [2657] = 2655,
  [2658] = 919,
  [2659] = 1072,
  [2660] = 2655,
  [2661] = 2655,
  [2662] = 2662,
  [2663] = 2663,
  [2664] = 2664,
  [2665] = 2665,
  [2666] = 168,
  [2667] = 2667,
  [2668] = 2668,
  [2669] = 2669,
  [2670] = 2664,
  [2671] = 2671,
  [2672] = 2672,
  [2673] = 2673,
  [2674] = 2674,
  [2675] = 2675,
  [2676] = 2665,
  [2677] = 2672,
  [2678] = 2678,
  [2679] = 2679,
  [2680] = 1506,
  [2681] = 2681,
  [2682] = 2682,
  [2683] = 168,
  [2684] = 2684,
  [2685] = 2601,
  [2686] = 2626,
  [2687] = 2687,
  [2688] = 2595,
  [2689] = 2650,
  [2690] = 2651,
  [2691] = 2652,
  [2692] = 2653,
  [2693] = 2643,
  [2694] = 2596,
  [2695] = 2597,
  [2696] = 2599,
  [2697] = 2600,
  [2698] = 2601,
  [2699] = 2629,
  [2700] = 2644,
  [2701] = 2645,
  [2702] = 2646,
  [2703] = 2647,
  [2704] = 2588,
  [2705] = 2589,
  [2706] = 2590,
  [2707] = 2591,
  [2708] = 2708,
  [2709] = 2603,
  [2710] = 2625,
  [2711] = 2626,
  [2712] = 2594,
  [2713] = 2595,
  [2714] = 2596,
  [2715] = 2597,
  [2716] = 2599,
  [2717] = 2600,
  [2718] = 2602,
  [2719] = 2602,
  [2720] = 2610,
  [2721] = 2611,
  [2722] = 2612,
  [2723] = 2613,
  [2724] = 2614,
  [2725] = 2615,
  [2726] = 2654,
  [2727] = 2617,
  [2728] = 2619,
  [2729] = 2620,
  [2730] = 2621,
  [2731] = 2622,
  [2732] = 2624,
  [2733] = 2629,
  [2734] = 2644,
  [2735] = 2645,
  [2736] = 2646,
  [2737] = 2647,
  [2738] = 2588,
  [2739] = 2589,
  [2740] = 2590,
  [2741] = 2591,
  [2742] = 2592,
  [2743] = 2593,
  [2744] = 2744,
  [2745] = 2592,
  [2746] = 2593,
  [2747] = 1507,
  [2748] = 2748,
  [2749] = 1468,
  [2750] = 2750,
  [2751] = 2628,
  [2752] = 2603,
  [2753] = 2753,
  [2754] = 2754,
  [2755] = 2650,
  [2756] = 2651,
  [2757] = 2652,
  [2758] = 2653,
  [2759] = 2643,
  [2760] = 2628,
  [2761] = 2625,
  [2762] = 2610,
  [2763] = 2611,
  [2764] = 2612,
  [2765] = 2613,
  [2766] = 2614,
  [2767] = 2615,
  [2768] = 2654,
  [2769] = 2617,
  [2770] = 2619,
  [2771] = 2620,
  [2772] = 2621,
  [2773] = 2622,
  [2774] = 2624,
  [2775] = 2594,
  [2776] = 2776,
  [2777] = 2669,
  [2778] = 2778,
  [2779] = 2779,
  [2780] = 2780,
  [2781] = 2674,
  [2782] = 2782,
  [2783] = 2783,
  [2784] = 2784,
  [2785] = 2785,
  [2786] = 2786,
  [2787] = 2787,
  [2788] = 2788,
  [2789] = 2789,
  [2790] = 2790,
  [2791] = 2791,
  [2792] = 2792,
  [2793] = 2787,
  [2794] = 2792,
  [2795] = 2795,
  [2796] = 2796,
  [2797] = 2792,
  [2798] = 2788,
  [2799] = 2792,
  [2800] = 2800,
  [2801] = 2792,
  [2802] = 2802,
  [2803] = 2803,
  [2804] = 2804,
  [2805] = 2805,
  [2806] = 2806,
  [2807] = 2807,
  [2808] = 2808,
  [2809] = 2809,
  [2810] = 2804,
  [2811] = 2783,
  [2812] = 2812,
  [2813] = 2813,
  [2814] = 2814,
  [2815] = 2813,
  [2816] = 2812,
  [2817] = 2812,
  [2818] = 2814,
  [2819] = 2812,
  [2820] = 2820,
  [2821] = 2820,
  [2822] = 2822,
  [2823] = 2782,
  [2824] = 2813,
  [2825] = 2812,
  [2826] = 2826,
  [2827] = 2820,
  [2828] = 2820,
  [2829] = 2814,
  [2830] = 2822,
  [2831] = 2831,
  [2832] = 2812,
  [2833] = 2833,
  [2834] = 2834,
  [2835] = 2812,
  [2836] = 2836,
  [2837] = 2814,
  [2838] = 2812,
  [2839] = 2839,
  [2840] = 2836,
  [2841] = 2813,
  [2842] = 2822,
  [2843] = 2820,
  [2844] = 2814,
  [2845] = 2826,
  [2846] = 2813,
  [2847] = 2836,
  [2848] = 2836,
  [2849] = 2836,
  [2850] = 2826,
  [2851] = 2851,
  [2852] = 2851,
  [2853] = 2851,
  [2854] = 2854,
  [2855] = 2855,
  [2856] = 2856,
  [2857] = 2857,
  [2858] = 168,
  [2859] = 2859,
  [2860] = 2855,
  [2861] = 2836,
  [2862] = 2862,
  [2863] = 2836,
  [2864] = 1993,
  [2865] = 2855,
  [2866] = 2866,
  [2867] = 2855,
  [2868] = 2855,
  [2869] = 2855,
  [2870] = 2836,
  [2871] = 2855,
  [2872] = 2855,
  [2873] = 2873,
  [2874] = 2836,
  [2875] = 2875,
  [2876] = 2876,
  [2877] = 2877,
  [2878] = 2878,
  [2879] = 2879,
  [2880] = 2880,
  [2881] = 2875,
  [2882] = 2875,
  [2883] = 2879,
  [2884] = 2884,
  [2885] = 2885,
  [2886] = 2877,
  [2887] = 2887,
  [2888] = 2875,
  [2889] = 2875,
  [2890] = 2875,
  [2891] = 2879,
  [2892] = 2878,
  [2893] = 2875,
  [2894] = 2887,
  [2895] = 2878,
  [2896] = 2877,
  [2897] = 2897,
  [2898] = 2836,
  [2899] = 2875,
  [2900] = 2900,
  [2901] = 2887,
  [2902] = 2902,
  [2903] = 2903,
  [2904] = 2904,
  [2905] = 2905,
  [2906] = 2906,
  [2907] = 2907,
  [2908] = 2908,
  [2909] = 2909,
  [2910] = 2910,
  [2911] = 2911,
  [2912] = 2912,
  [2913] = 2913,
  [2914] = 2914,
  [2915] = 2915,
  [2916] = 2481,
  [2917] = 2917,
  [2918] = 2918,
  [2919] = 2919,
  [2920] = 2482,
  [2921] = 2921,
  [2922] = 2922,
  [2923] = 2913,
  [2924] = 2924,
  [2925] = 2925,
  [2926] = 2926,
  [2927] = 2927,
  [2928] = 2928,
  [2929] = 2929,
  [2930] = 2930,
  [2931] = 2931,
  [2932] = 2932,
  [2933] = 2933,
  [2934] = 2934,
  [2935] = 2935,
  [2936] = 2936,
  [2937] = 2937,
  [2938] = 2938,
  [2939] = 2939,
  [2940] = 2940,
  [2941] = 2941,
  [2942] = 2942,
  [2943] = 2943,
  [2944] = 2944,
  [2945] = 2945,
  [2946] = 2946,
  [2947] = 2947,
  [2948] = 2948,
  [2949] = 2949,
  [2950] = 2950,
  [2951] = 2951,
  [2952] = 2952,
  [2953] = 2953,
  [2954] = 168,
  [2955] = 2955,
  [2956] = 2956,
  [2957] = 2957,
  [2958] = 2958,
  [2959] = 2933,
  [2960] = 2960,
  [2961] = 2961,
  [2962] = 2962,
  [2963] = 2963,
  [2964] = 2964,
  [2965] = 2965,
  [2966] = 2966,
  [2967] = 2967,
  [2968] = 2968,
  [2969] = 2969,
  [2970] = 2970,
  [2971] = 2971,
  [2972] = 2972,
  [2973] = 2973,
  [2974] = 2974,
  [2975] = 2975,
  [2976] = 2976,
  [2977] = 2977,
  [2978] = 2978,
  [2979] = 2979,
  [2980] = 2980,
  [2981] = 2981,
  [2982] = 2982,
  [2983] = 2983,
  [2984] = 2984,
  [2985] = 2985,
  [2986] = 2986,
  [2987] = 2987,
  [2988] = 2988,
  [2989] = 2989,
  [2990] = 2990,
  [2991] = 2991,
  [2992] = 2992,
  [2993] = 2993,
  [2994] = 2994,
  [2995] = 2995,
  [2996] = 2996,
  [2997] = 2997,
  [2998] = 2933,
  [2999] = 2999,
  [3000] = 3000,
  [3001] = 3001,
  [3002] = 3002,
  [3003] = 3003,
  [3004] = 3004,
  [3005] = 2933,
  [3006] = 2933,
  [3007] = 2552,
  [3008] = 3008,
  [3009] = 168,
  [3010] = 3010,
  [3011] = 2470,
  [3012] = 3008,
  [3013] = 3013,
  [3014] = 3014,
  [3015] = 3015,
  [3016] = 3016,
  [3017] = 3017,
  [3018] = 3015,
  [3019] = 3017,
  [3020] = 2682,
  [3021] = 2678,
  [3022] = 3022,
  [3023] = 3023,
  [3024] = 3016,
  [3025] = 3025,
  [3026] = 3026,
  [3027] = 3027,
  [3028] = 2681,
  [3029] = 2679,
  [3030] = 3017,
  [3031] = 3013,
  [3032] = 3032,
  [3033] = 3033,
  [3034] = 3034,
  [3035] = 3016,
  [3036] = 3027,
  [3037] = 3013,
  [3038] = 3038,
  [3039] = 3039,
  [3040] = 3040,
  [3041] = 3027,
  [3042] = 3015,
  [3043] = 3023,
  [3044] = 3023,
  [3045] = 3045,
  [3046] = 3046,
  [3047] = 3047,
  [3048] = 2667,
  [3049] = 2748,
  [3050] = 2750,
  [3051] = 2785,
  [3052] = 2671,
  [3053] = 2753,
  [3054] = 2754,
  [3055] = 2708,
  [3056] = 2779,
  [3057] = 3057,
  [3058] = 3058,
  [3059] = 2786,
  [3060] = 2902,
  [3061] = 3061,
  [3062] = 3058,
  [3063] = 3045,
  [3064] = 2902,
  [3065] = 2778,
  [3066] = 2662,
  [3067] = 2776,
  [3068] = 2687,
  [3069] = 3061,
  [3070] = 2784,
  [3071] = 2780,
  [3072] = 3057,
  [3073] = 3073,
  [3074] = 3074,
  [3075] = 3075,
  [3076] = 3076,
  [3077] = 3077,
  [3078] = 3078,
  [3079] = 3079,
  [3080] = 3080,
  [3081] = 3081,
  [3082] = 3082,
  [3083] = 3076,
  [3084] = 3075,
  [3085] = 3076,
  [3086] = 3077,
  [3087] = 3078,
  [3088] = 3079,
  [3089] = 3081,
  [3090] = 3082,
  [3091] = 3075,
  [3092] = 3076,
  [3093] = 3077,
  [3094] = 3078,
  [3095] = 3079,
  [3096] = 3081,
  [3097] = 3082,
  [3098] = 3075,
  [3099] = 3076,
  [3100] = 3077,
  [3101] = 3078,
  [3102] = 3079,
  [3103] = 3081,
  [3104] = 3082,
  [3105] = 3081,
  [3106] = 3075,
  [3107] = 3076,
  [3108] = 3077,
  [3109] = 3078,
  [3110] = 3079,
  [3111] = 3081,
  [3112] = 3082,
  [3113] = 3077,
  [3114] = 3078,
  [3115] = 3079,
  [3116] = 3082,
  [3117] = 3077,
  [3118] = 3076,
  [3119] = 3119,
  [3120] = 3077,
  [3121] = 3073,
  [3122] = 3122,
  [3123] = 3076,
  [3124] = 3081,
  [3125] = 3125,
  [3126] = 3082,
  [3127] = 3122,
  [3128] = 3128,
  [3129] = 3080,
  [3130] = 3073,
  [3131] = 3131,
  [3132] = 3081,
  [3133] = 3080,
  [3134] = 3134,
  [3135] = 3135,
  [3136] = 3136,
  [3137] = 3137,
  [3138] = 3078,
  [3139] = 3079,
  [3140] = 3125,
  [3141] = 3082,
  [3142] = 3074,
  [3143] = 3078,
  [3144] = 3144,
  [3145] = 3075,
  [3146] = 3075,
  [3147] = 3147,
  [3148] = 3148,
  [3149] = 3137,
  [3150] = 3075,
  [3151] = 3151,
  [3152] = 3119,
  [3153] = 3075,
  [3154] = 3081,
  [3155] = 3155,
  [3156] = 3075,
  [3157] = 3157,
  [3158] = 2552,
  [3159] = 3159,
  [3160] = 3076,
  [3161] = 3077,
  [3162] = 3076,
  [3163] = 3077,
  [3164] = 3079,
  [3165] = 3078,
  [3166] = 3079,
  [3167] = 3167,
  [3168] = 3076,
  [3169] = 3078,
  [3170] = 3074,
  [3171] = 3079,
  [3172] = 3077,
  [3173] = 3078,
  [3174] = 3079,
  [3175] = 3175,
  [3176] = 3176,
  [3177] = 3075,
  [3178] = 3178,
  [3179] = 3131,
  [3180] = 3176,
  [3181] = 3181,
  [3182] = 3081,
  [3183] = 3082,
  [3184] = 3122,
  [3185] = 3073,
  [3186] = 3081,
  [3187] = 3137,
  [3188] = 3125,
  [3189] = 3189,
  [3190] = 3175,
  [3191] = 3131,
  [3192] = 3192,
  [3193] = 3119,
  [3194] = 3119,
  [3195] = 3148,
  [3196] = 3148,
  [3197] = 3082,
  [3198] = 3198,
  [3199] = 3176,
  [3200] = 3082,
  [3201] = 3175,
  [3202] = 2678,
  [3203] = 2681,
  [3204] = 2679,
  [3205] = 3205,
  [3206] = 3206,
  [3207] = 1765,
  [3208] = 2682,
  [3209] = 2748,
  [3210] = 2750,
  [3211] = 2687,
  [3212] = 2708,
  [3213] = 1765,
  [3214] = 3205,
  [3215] = 2753,
  [3216] = 2754,
  [3217] = 3217,
  [3218] = 312,
  [3219] = 349,
  [3220] = 282,
  [3221] = 3221,
  [3222] = 2663,
  [3223] = 3223,
  [3224] = 3224,
  [3225] = 3225,
  [3226] = 3226,
  [3227] = 3227,
  [3228] = 3228,
  [3229] = 3229,
  [3230] = 3228,
  [3231] = 3231,
  [3232] = 3232,
  [3233] = 2684,
  [3234] = 3234,
  [3235] = 3231,
  [3236] = 3236,
  [3237] = 3231,
  [3238] = 3238,
  [3239] = 3239,
  [3240] = 3240,
  [3241] = 3241,
  [3242] = 3242,
  [3243] = 3243,
  [3244] = 3244,
  [3245] = 3245,
  [3246] = 3246,
  [3247] = 3247,
  [3248] = 3248,
  [3249] = 3249,
  [3250] = 3250,
  [3251] = 3251,
  [3252] = 3252,
  [3253] = 3253,
  [3254] = 3254,
  [3255] = 3248,
  [3256] = 3256,
  [3257] = 3257,
  [3258] = 3258,
  [3259] = 3259,
  [3260] = 3260,
  [3261] = 3261,
  [3262] = 3262,
  [3263] = 3263,
  [3264] = 3264,
  [3265] = 2807,
  [3266] = 3266,
  [3267] = 2806,
  [3268] = 3268,
  [3269] = 3269,
  [3270] = 2791,
  [3271] = 3271,
  [3272] = 3269,
  [3273] = 3273,
  [3274] = 2789,
  [3275] = 3275,
  [3276] = 3276,
  [3277] = 3277,
  [3278] = 2809,
  [3279] = 3279,
  [3280] = 3280,
  [3281] = 3281,
  [3282] = 3280,
  [3283] = 3283,
  [3284] = 3281,
  [3285] = 3285,
  [3286] = 3286,
  [3287] = 3287,
  [3288] = 3283,
  [3289] = 3289,
  [3290] = 3290,
  [3291] = 3291,
  [3292] = 3287,
  [3293] = 3293,
  [3294] = 3294,
  [3295] = 3295,
  [3296] = 3296,
  [3297] = 3297,
  [3298] = 3297,
  [3299] = 3297,
  [3300] = 3297,
  [3301] = 3297,
  [3302] = 3297,
  [3303] = 3297,
  [3304] = 3297,
  [3305] = 3305,
  [3306] = 3306,
  [3307] = 3307,
  [3308] = 3308,
  [3309] = 3309,
  [3310] = 3310,
  [3311] = 3311,
  [3312] = 3312,
  [3313] = 3313,
  [3314] = 3314,
  [3315] = 3315,
  [3316] = 3316,
  [3317] = 3314,
  [3318] = 3311,
  [3319] = 3319,
  [3320] = 3320,
  [3321] = 3319,
  [3322] = 3322,
  [3323] = 1072,
  [3324] = 3324,
  [3325] = 3325,
  [3326] = 3324,
  [3327] = 3324,
  [3328] = 3324,
  [3329] = 3329,
  [3330] = 3330,
  [3331] = 919,
  [3332] = 3332,
  [3333] = 3332,
  [3334] = 3334,
  [3335] = 3335,
  [3336] = 3336,
  [3337] = 3332,
  [3338] = 3338,
  [3339] = 3332,
  [3340] = 3332,
  [3341] = 3341,
  [3342] = 3335,
  [3343] = 3343,
  [3344] = 3338,
  [3345] = 3345,
  [3346] = 3346,
  [3347] = 3335,
  [3348] = 3348,
  [3349] = 3349,
  [3350] = 3350,
  [3351] = 3332,
  [3352] = 3335,
  [3353] = 3338,
  [3354] = 3354,
  [3355] = 3336,
  [3356] = 3335,
  [3357] = 3338,
  [3358] = 3332,
  [3359] = 3354,
  [3360] = 3332,
  [3361] = 3354,
  [3362] = 3362,
  [3363] = 3363,
  [3364] = 3332,
  [3365] = 3335,
  [3366] = 3366,
  [3367] = 3367,
  [3368] = 3350,
  [3369] = 3335,
  [3370] = 3338,
  [3371] = 3338,
  [3372] = 3332,
  [3373] = 3335,
  [3374] = 3335,
  [3375] = 3348,
  [3376] = 3332,
  [3377] = 3335,
  [3378] = 3335,
  [3379] = 3379,
  [3380] = 3335,
  [3381] = 3381,
  [3382] = 3346,
  [3383] = 3381,
  [3384] = 3338,
  [3385] = 3354,
  [3386] = 3332,
  [3387] = 3354,
  [3388] = 3363,
  [3389] = 3363,
  [3390] = 3390,
  [3391] = 3391,
  [3392] = 3363,
  [3393] = 3349,
  [3394] = 3394,
  [3395] = 3395,
  [3396] = 3396,
  [3397] = 3397,
  [3398] = 3398,
  [3399] = 3399,
  [3400] = 3398,
  [3401] = 3401,
  [3402] = 2487,
  [3403] = 3403,
  [3404] = 3404,
  [3405] = 3405,
  [3406] = 3406,
  [3407] = 3399,
  [3408] = 3408,
  [3409] = 3409,
  [3410] = 3410,
  [3411] = 3411,
  [3412] = 3412,
  [3413] = 3397,
  [3414] = 3404,
  [3415] = 3399,
  [3416] = 3416,
  [3417] = 3399,
  [3418] = 3418,
  [3419] = 3406,
  [3420] = 3398,
  [3421] = 3398,
  [3422] = 3398,
  [3423] = 3399,
  [3424] = 3399,
  [3425] = 3411,
  [3426] = 3398,
  [3427] = 3427,
  [3428] = 3428,
  [3429] = 3429,
  [3430] = 3399,
  [3431] = 3398,
  [3432] = 3398,
  [3433] = 3405,
  [3434] = 3398,
  [3435] = 3406,
  [3436] = 3436,
  [3437] = 3403,
  [3438] = 3428,
  [3439] = 3439,
  [3440] = 3398,
  [3441] = 3404,
  [3442] = 3442,
  [3443] = 3403,
  [3444] = 3398,
  [3445] = 3445,
  [3446] = 3403,
  [3447] = 3398,
  [3448] = 3448,
  [3449] = 3449,
  [3450] = 3450,
  [3451] = 3451,
  [3452] = 3452,
  [3453] = 3448,
  [3454] = 3454,
  [3455] = 3455,
  [3456] = 3456,
  [3457] = 3448,
  [3458] = 3458,
  [3459] = 3459,
  [3460] = 3458,
  [3461] = 3461,
  [3462] = 3462,
  [3463] = 3449,
  [3464] = 3464,
  [3465] = 3464,
  [3466] = 3466,
  [3467] = 3467,
  [3468] = 3454,
  [3469] = 3455,
  [3470] = 3470,
  [3471] = 3471,
  [3472] = 3456,
  [3473] = 3470,
  [3474] = 3474,
  [3475] = 3448,
  [3476] = 3461,
  [3477] = 3462,
  [3478] = 3449,
  [3479] = 3479,
  [3480] = 3464,
  [3481] = 3461,
  [3482] = 3482,
  [3483] = 3462,
  [3484] = 3484,
  [3485] = 3456,
  [3486] = 3459,
  [3487] = 3458,
  [3488] = 3466,
  [3489] = 3467,
  [3490] = 3490,
  [3491] = 3491,
  [3492] = 3454,
  [3493] = 3449,
  [3494] = 3455,
  [3495] = 3450,
  [3496] = 3466,
  [3497] = 3497,
  [3498] = 3456,
  [3499] = 3499,
  [3500] = 3459,
  [3501] = 3501,
  [3502] = 3501,
  [3503] = 3451,
  [3504] = 3458,
  [3505] = 3450,
  [3506] = 3464,
  [3507] = 3448,
  [3508] = 3461,
  [3509] = 3462,
  [3510] = 3449,
  [3511] = 3464,
  [3512] = 3466,
  [3513] = 3467,
  [3514] = 3454,
  [3515] = 3466,
  [3516] = 3455,
  [3517] = 3450,
  [3518] = 3467,
  [3519] = 3454,
  [3520] = 3470,
  [3521] = 3452,
  [3522] = 3471,
  [3523] = 3479,
  [3524] = 3524,
  [3525] = 3450,
  [3526] = 3451,
  [3527] = 3474,
  [3528] = 3528,
  [3529] = 3456,
  [3530] = 3530,
  [3531] = 3490,
  [3532] = 3532,
  [3533] = 3456,
  [3534] = 3458,
  [3535] = 3499,
  [3536] = 3501,
  [3537] = 3524,
  [3538] = 3451,
  [3539] = 3459,
  [3540] = 3458,
  [3541] = 3528,
  [3542] = 3530,
  [3543] = 3470,
  [3544] = 3501,
  [3545] = 3524,
  [3546] = 3451,
  [3547] = 3459,
  [3548] = 3458,
  [3549] = 3524,
  [3550] = 3455,
  [3551] = 3524,
  [3552] = 3458,
  [3553] = 3471,
  [3554] = 3471,
  [3555] = 3501,
  [3556] = 3451,
  [3557] = 3499,
  [3558] = 3558,
  [3559] = 3448,
  [3560] = 3461,
  [3561] = 3462,
  [3562] = 3449,
  [3563] = 3464,
  [3564] = 3466,
  [3565] = 3467,
  [3566] = 3454,
  [3567] = 3455,
  [3568] = 3470,
  [3569] = 3471,
  [3570] = 3558,
  [3571] = 3467,
  [3572] = 3479,
  [3573] = 3573,
  [3574] = 3450,
  [3575] = 3528,
  [3576] = 3474,
  [3577] = 3452,
  [3578] = 3530,
  [3579] = 3573,
  [3580] = 3479,
  [3581] = 3474,
  [3582] = 3558,
  [3583] = 3499,
  [3584] = 3458,
  [3585] = 3501,
  [3586] = 3451,
  [3587] = 3558,
  [3588] = 3448,
  [3589] = 3461,
  [3590] = 3462,
  [3591] = 3449,
  [3592] = 3464,
  [3593] = 3466,
  [3594] = 3467,
  [3595] = 3454,
  [3596] = 3455,
  [3597] = 3470,
  [3598] = 3471,
  [3599] = 3479,
  [3600] = 3461,
  [3601] = 3462,
  [3602] = 3449,
  [3603] = 3464,
  [3604] = 3450,
  [3605] = 3466,
  [3606] = 3474,
  [3607] = 3491,
  [3608] = 3501,
  [3609] = 3451,
  [3610] = 3448,
  [3611] = 3461,
  [3612] = 3462,
  [3613] = 3449,
  [3614] = 3464,
  [3615] = 3466,
  [3616] = 3467,
  [3617] = 3454,
  [3618] = 3455,
  [3619] = 3470,
  [3620] = 3467,
  [3621] = 3479,
  [3622] = 3454,
  [3623] = 3474,
  [3624] = 3530,
  [3625] = 3455,
  [3626] = 3470,
  [3627] = 3450,
  [3628] = 3470,
  [3629] = 3474,
  [3630] = 3456,
  [3631] = 3459,
  [3632] = 3458,
  [3633] = 3471,
  [3634] = 3456,
  [3635] = 3459,
  [3636] = 3471,
  [3637] = 2356,
  [3638] = 3501,
  [3639] = 3501,
  [3640] = 3451,
  [3641] = 3558,
  [3642] = 3448,
  [3643] = 3461,
  [3644] = 3462,
  [3645] = 3449,
  [3646] = 3464,
  [3647] = 3466,
  [3648] = 3467,
  [3649] = 3454,
  [3650] = 3455,
  [3651] = 3470,
  [3652] = 3471,
  [3653] = 3653,
  [3654] = 3479,
  [3655] = 3573,
  [3656] = 3451,
  [3657] = 3450,
  [3658] = 3528,
  [3659] = 3474,
  [3660] = 3530,
  [3661] = 3661,
  [3662] = 3479,
  [3663] = 3532,
  [3664] = 3664,
  [3665] = 3573,
  [3666] = 3528,
  [3667] = 3530,
  [3668] = 3474,
  [3669] = 3573,
  [3670] = 3528,
  [3671] = 3530,
  [3672] = 2354,
  [3673] = 3573,
  [3674] = 3528,
  [3675] = 3530,
  [3676] = 3573,
  [3677] = 3573,
  [3678] = 3456,
  [3679] = 3499,
  [3680] = 3680,
  [3681] = 3459,
  [3682] = 3458,
  [3683] = 3680,
  [3684] = 3684,
  [3685] = 3528,
  [3686] = 3501,
  [3687] = 3501,
  [3688] = 3451,
  [3689] = 3448,
  [3690] = 3461,
  [3691] = 3462,
  [3692] = 3449,
  [3693] = 3464,
  [3694] = 3466,
  [3695] = 3467,
  [3696] = 3454,
  [3697] = 3455,
  [3698] = 3470,
  [3699] = 3471,
  [3700] = 3450,
  [3701] = 3479,
  [3702] = 3702,
  [3703] = 3450,
  [3704] = 3479,
  [3705] = 3474,
  [3706] = 3456,
  [3707] = 3459,
  [3708] = 3458,
  [3709] = 3524,
  [3710] = 3710,
  [3711] = 3456,
  [3712] = 3499,
  [3713] = 3479,
  [3714] = 3459,
  [3715] = 3715,
  [3716] = 3501,
  [3717] = 3451,
  [3718] = 3448,
  [3719] = 3461,
  [3720] = 3462,
  [3721] = 3449,
  [3722] = 3464,
  [3723] = 3466,
  [3724] = 3467,
  [3725] = 3454,
  [3726] = 3455,
  [3727] = 3470,
  [3728] = 3471,
  [3729] = 3729,
  [3730] = 3479,
  [3731] = 3731,
  [3732] = 3474,
  [3733] = 3450,
  [3734] = 3474,
  [3735] = 3456,
  [3736] = 3459,
  [3737] = 3737,
  [3738] = 3558,
  [3739] = 3739,
  [3740] = 3740,
  [3741] = 3448,
  [3742] = 3459,
  [3743] = 3461,
  [3744] = 3462,
  [3745] = 3739,
  [3746] = 3471,
  [3747] = 3747,
  [3748] = 3747,
  [3749] = 3749,
  [3750] = 3749,
  [3751] = 3749,
  [3752] = 3747,
  [3753] = 3753,
  [3754] = 3754,
  [3755] = 3754,
  [3756] = 3753,
  [3757] = 3753,
  [3758] = 3754,
  [3759] = 1765,
  [3760] = 349,
  [3761] = 3761,
  [3762] = 312,
  [3763] = 282,
  [3764] = 2552,
  [3765] = 3765,
  [3766] = 3766,
  [3767] = 3765,
  [3768] = 3766,
  [3769] = 3769,
  [3770] = 3769,
  [3771] = 2862,
  [3772] = 2856,
  [3773] = 3766,
  [3774] = 2662,
  [3775] = 2667,
  [3776] = 3765,
  [3777] = 2671,
  [3778] = 3769,
  [3779] = 2487,
  [3780] = 2885,
  [3781] = 2873,
  [3782] = 2876,
  [3783] = 2880,
  [3784] = 2897,
  [3785] = 2900,
  [3786] = 2909,
  [3787] = 2914,
  [3788] = 2938,
  [3789] = 2969,
  [3790] = 2973,
  [3791] = 2470,
  [3792] = 2912,
  [3793] = 2917,
  [3794] = 2927,
  [3795] = 2922,
  [3796] = 2926,
  [3797] = 2921,
  [3798] = 2924,
  [3799] = 2999,
  [3800] = 3058,
  [3801] = 2918,
  [3802] = 2919,
  [3803] = 2679,
  [3804] = 2925,
  [3805] = 2911,
  [3806] = 2682,
  [3807] = 2678,
  [3808] = 2905,
  [3809] = 2966,
  [3810] = 2967,
  [3811] = 3811,
  [3812] = 2961,
  [3813] = 2908,
  [3814] = 2982,
  [3815] = 2985,
  [3816] = 2992,
  [3817] = 2993,
  [3818] = 2995,
  [3819] = 3045,
  [3820] = 3061,
  [3821] = 2929,
  [3822] = 2910,
  [3823] = 2681,
  [3824] = 2981,
  [3825] = 3825,
  [3826] = 3826,
  [3827] = 3827,
  [3828] = 3828,
  [3829] = 2708,
  [3830] = 1762,
  [3831] = 2748,
  [3832] = 2687,
  [3833] = 3828,
  [3834] = 3032,
  [3835] = 3034,
  [3836] = 2354,
  [3837] = 3825,
  [3838] = 2937,
  [3839] = 2940,
  [3840] = 2943,
  [3841] = 2945,
  [3842] = 2947,
  [3843] = 2356,
  [3844] = 3147,
  [3845] = 1765,
  [3846] = 2951,
  [3847] = 2952,
  [3848] = 2930,
  [3849] = 1736,
  [3850] = 3151,
  [3851] = 3157,
  [3852] = 2949,
  [3853] = 3002,
  [3854] = 2942,
  [3855] = 2946,
  [3856] = 2948,
  [3857] = 2957,
  [3858] = 3826,
  [3859] = 3003,
  [3860] = 3004,
  [3861] = 2931,
  [3862] = 2972,
  [3863] = 2750,
  [3864] = 2932,
  [3865] = 2936,
  [3866] = 2753,
  [3867] = 2934,
  [3868] = 3178,
  [3869] = 2939,
  [3870] = 2971,
  [3871] = 3025,
  [3872] = 3026,
  [3873] = 2953,
  [3874] = 2955,
  [3875] = 3875,
  [3876] = 2941,
  [3877] = 2968,
  [3878] = 2935,
  [3879] = 2944,
  [3880] = 3880,
  [3881] = 3881,
  [3882] = 3192,
  [3883] = 3167,
  [3884] = 2974,
  [3885] = 3181,
  [3886] = 3825,
  [3887] = 3826,
  [3888] = 3189,
  [3889] = 2956,
  [3890] = 3198,
  [3891] = 3136,
  [3892] = 2960,
  [3893] = 3828,
  [3894] = 2963,
  [3895] = 3875,
  [3896] = 3880,
  [3897] = 3881,
  [3898] = 3825,
  [3899] = 3826,
  [3900] = 3828,
  [3901] = 2950,
  [3902] = 2754,
  [3903] = 3875,
  [3904] = 2958,
  [3905] = 3880,
  [3906] = 3881,
  [3907] = 3825,
  [3908] = 3826,
  [3909] = 3128,
  [3910] = 3828,
  [3911] = 3875,
  [3912] = 3880,
  [3913] = 3881,
  [3914] = 3825,
  [3915] = 3826,
  [3916] = 3828,
  [3917] = 3134,
  [3918] = 3875,
  [3919] = 3880,
  [3920] = 3881,
  [3921] = 3825,
  [3922] = 3826,
  [3923] = 3828,
  [3924] = 3875,
  [3925] = 3880,
  [3926] = 3881,
  [3927] = 3825,
  [3928] = 3826,
  [3929] = 3828,
  [3930] = 3875,
  [3931] = 3880,
  [3932] = 3881,
  [3933] = 3825,
  [3934] = 3826,
  [3935] = 3828,
  [3936] = 3875,
  [3937] = 3880,
  [3938] = 3881,
  [3939] = 3825,
  [3940] = 3826,
  [3941] = 3828,
  [3942] = 2965,
  [3943] = 3943,
  [3944] = 3944,
  [3945] = 1882,
  [3946] = 2928,
  [3947] = 2964,
  [3948] = 3880,
  [3949] = 3881,
  [3950] = 2996,
  [3951] = 3039,
  [3952] = 3040,
  [3953] = 3135,
  [3954] = 3875,
  [3955] = 2975,
  [3956] = 2976,
  [3957] = 2977,
  [3958] = 3144,
  [3959] = 2980,
  [3960] = 2979,
  [3961] = 3159,
  [3962] = 3881,
  [3963] = 2983,
  [3964] = 2984,
  [3965] = 2978,
  [3966] = 2986,
  [3967] = 2987,
  [3968] = 2994,
  [3969] = 2988,
  [3970] = 2989,
  [3971] = 2990,
  [3972] = 2997,
  [3973] = 2991,
  [3974] = 3974,
  [3975] = 3880,
  [3976] = 2915,
  [3977] = 3875,
  [3978] = 2786,
  [3979] = 1495,
  [3980] = 2778,
  [3981] = 1481,
  [3982] = 2776,
  [3983] = 1453,
  [3984] = 1455,
  [3985] = 3045,
  [3986] = 2779,
  [3987] = 2784,
  [3988] = 1508,
  [3989] = 1510,
  [3990] = 3990,
  [3991] = 1452,
  [3992] = 2780,
  [3993] = 1456,
  [3994] = 3058,
  [3995] = 1466,
  [3996] = 3990,
  [3997] = 2785,
  [3998] = 1494,
  [3999] = 3061,
  [4000] = 4000,
  [4001] = 4000,
  [4002] = 4002,
  [4003] = 3025,
  [4004] = 4004,
  [4005] = 4002,
  [4006] = 4004,
  [4007] = 4004,
  [4008] = 4004,
  [4009] = 4002,
  [4010] = 4004,
  [4011] = 4011,
  [4012] = 3026,
  [4013] = 4002,
  [4014] = 4011,
  [4015] = 3032,
  [4016] = 4000,
  [4017] = 4004,
  [4018] = 4002,
  [4019] = 3039,
  [4020] = 4000,
  [4021] = 4011,
  [4022] = 4000,
  [4023] = 4002,
  [4024] = 4004,
  [4025] = 4004,
  [4026] = 4011,
  [4027] = 4011,
  [4028] = 3034,
  [4029] = 4000,
  [4030] = 4004,
  [4031] = 4011,
  [4032] = 4002,
  [4033] = 4002,
  [4034] = 3040,
  [4035] = 4002,
  [4036] = 4011,
  [4037] = 4004,
  [4038] = 4011,
  [4039] = 4011,
  [4040] = 4000,
  [4041] = 4000,
  [4042] = 4000,
  [4043] = 4000,
  [4044] = 4002,
  [4045] = 4011,
  [4046] = 3032,
  [4047] = 4047,
  [4048] = 329,
  [4049] = 330,
  [4050] = 2275,
  [4051] = 2330,
  [4052] = 2274,
  [4053] = 3025,
  [4054] = 4054,
  [4055] = 3026,
  [4056] = 310,
  [4057] = 313,
  [4058] = 293,
  [4059] = 301,
  [4060] = 4060,
  [4061] = 348,
  [4062] = 3039,
  [4063] = 285,
  [4064] = 3034,
  [4065] = 3058,
  [4066] = 4066,
  [4067] = 3061,
  [4068] = 3040,
  [4069] = 3045,
  [4070] = 2272,
  [4071] = 1074,
  [4072] = 1075,
  [4073] = 1498,
  [4074] = 3136,
  [4075] = 3135,
  [4076] = 4076,
  [4077] = 4077,
  [4078] = 3167,
  [4079] = 3181,
  [4080] = 3144,
  [4081] = 4081,
  [4082] = 4081,
  [4083] = 4076,
  [4084] = 4084,
  [4085] = 3147,
  [4086] = 4077,
  [4087] = 2902,
  [4088] = 3189,
  [4089] = 4089,
  [4090] = 3159,
  [4091] = 3178,
  [4092] = 4092,
  [4093] = 4081,
  [4094] = 4089,
  [4095] = 3192,
  [4096] = 4092,
  [4097] = 4081,
  [4098] = 3151,
  [4099] = 3198,
  [4100] = 3157,
  [4101] = 3128,
  [4102] = 3134,
  [4103] = 4103,
  [4104] = 4103,
  [4105] = 1074,
  [4106] = 1075,
  [4107] = 4107,
  [4108] = 4107,
  [4109] = 4107,
  [4110] = 4107,
  [4111] = 4111,
  [4112] = 4112,
  [4113] = 4113,
  [4114] = 4112,
  [4115] = 4115,
  [4116] = 4111,
  [4117] = 4117,
  [4118] = 4112,
  [4119] = 4111,
  [4120] = 4112,
  [4121] = 4111,
  [4122] = 4122,
  [4123] = 4122,
  [4124] = 4122,
  [4125] = 4122,
  [4126] = 4122,
  [4127] = 4122,
  [4128] = 4122,
  [4129] = 4129,
  [4130] = 4129,
  [4131] = 4131,
  [4132] = 3025,
  [4133] = 3058,
  [4134] = 3026,
  [4135] = 3032,
  [4136] = 3034,
  [4137] = 4131,
  [4138] = 3045,
  [4139] = 3061,
  [4140] = 3039,
  [4141] = 3040,
  [4142] = 4142,
  [4143] = 3192,
  [4144] = 3189,
  [4145] = 3167,
  [4146] = 3157,
  [4147] = 3181,
  [4148] = 3144,
  [4149] = 2354,
  [4150] = 2356,
  [4151] = 3136,
  [4152] = 3128,
  [4153] = 3151,
  [4154] = 3134,
  [4155] = 3135,
  [4156] = 3147,
  [4157] = 3198,
  [4158] = 4158,
  [4159] = 4158,
  [4160] = 4142,
  [4161] = 3178,
  [4162] = 3159,
  [4163] = 4163,
  [4164] = 2856,
  [4165] = 2862,
  [4166] = 4166,
  [4167] = 4167,
  [4168] = 4168,
  [4169] = 270,
  [4170] = 251,
  [4171] = 254,
  [4172] = 4172,
  [4173] = 212,
  [4174] = 214,
  [4175] = 258,
  [4176] = 223,
  [4177] = 215,
  [4178] = 216,
  [4179] = 4179,
  [4180] = 4180,
  [4181] = 4181,
  [4182] = 4182,
  [4183] = 222,
  [4184] = 4180,
  [4185] = 2862,
  [4186] = 4180,
  [4187] = 4180,
  [4188] = 228,
  [4189] = 4189,
  [4190] = 239,
  [4191] = 4180,
  [4192] = 269,
  [4193] = 4172,
  [4194] = 204,
  [4195] = 4195,
  [4196] = 2856,
  [4197] = 242,
  [4198] = 208,
  [4199] = 4182,
  [4200] = 2902,
  [4201] = 4201,
  [4202] = 4202,
  [4203] = 4203,
  [4204] = 4204,
  [4205] = 4202,
  [4206] = 4201,
  [4207] = 4207,
  [4208] = 4203,
  [4209] = 4207,
  [4210] = 4201,
  [4211] = 4202,
  [4212] = 4202,
  [4213] = 4202,
  [4214] = 4207,
  [4215] = 4215,
  [4216] = 4203,
  [4217] = 4204,
  [4218] = 4201,
  [4219] = 4204,
  [4220] = 1445,
  [4221] = 4201,
  [4222] = 4204,
  [4223] = 4201,
  [4224] = 4207,
  [4225] = 4207,
  [4226] = 4204,
  [4227] = 4203,
  [4228] = 4202,
  [4229] = 4201,
  [4230] = 4203,
  [4231] = 4202,
  [4232] = 4203,
  [4233] = 4202,
  [4234] = 4203,
  [4235] = 4203,
  [4236] = 4201,
  [4237] = 2914,
  [4238] = 2993,
  [4239] = 2999,
  [4240] = 4240,
  [4241] = 2992,
  [4242] = 4242,
  [4243] = 4243,
  [4244] = 2995,
  [4245] = 2914,
  [4246] = 4246,
  [4247] = 4247,
  [4248] = 2969,
  [4249] = 4249,
  [4250] = 4250,
  [4251] = 2938,
  [4252] = 4252,
  [4253] = 4253,
  [4254] = 4254,
  [4255] = 4255,
  [4256] = 2961,
  [4257] = 2966,
  [4258] = 4258,
  [4259] = 4259,
  [4260] = 2973,
  [4261] = 2967,
  [4262] = 2982,
  [4263] = 2985,
  [4264] = 2856,
  [4265] = 1445,
  [4266] = 4266,
  [4267] = 4267,
  [4268] = 4268,
  [4269] = 2862,
  [4270] = 4270,
  [4271] = 4271,
  [4272] = 4272,
  [4273] = 4273,
  [4274] = 4274,
  [4275] = 4272,
  [4276] = 4276,
  [4277] = 4272,
  [4278] = 4272,
  [4279] = 4272,
  [4280] = 4272,
  [4281] = 4272,
  [4282] = 4282,
  [4283] = 4272,
  [4284] = 3045,
  [4285] = 3061,
  [4286] = 3058,
  [4287] = 4266,
  [4288] = 1445,
  [4289] = 4270,
  [4290] = 4271,
  [4291] = 4268,
  [4292] = 1564,
  [4293] = 1561,
  [4294] = 4273,
  [4295] = 1562,
  [4296] = 2914,
  [4297] = 1566,
  [4298] = 1567,
  [4299] = 1557,
  [4300] = 1555,
  [4301] = 1560,
  [4302] = 1556,
  [4303] = 1559,
  [4304] = 1565,
  [4305] = 1558,
  [4306] = 4274,
  [4307] = 1553,
  [4308] = 3224,
  [4309] = 1570,
  [4310] = 1563,
  [4311] = 3223,
  [4312] = 1454,
  [4313] = 4313,
  [4314] = 4314,
  [4315] = 1566,
  [4316] = 1567,
  [4317] = 1560,
  [4318] = 1557,
  [4319] = 1570,
  [4320] = 1555,
  [4321] = 1561,
  [4322] = 1559,
  [4323] = 1564,
  [4324] = 1558,
  [4325] = 1565,
  [4326] = 1556,
  [4327] = 1562,
  [4328] = 1553,
  [4329] = 1563,
  [4330] = 4330,
  [4331] = 4330,
  [4332] = 4330,
  [4333] = 4330,
  [4334] = 4330,
  [4335] = 4330,
  [4336] = 4330,
  [4337] = 4330,
  [4338] = 4330,
  [4339] = 4330,
  [4340] = 4330,
  [4341] = 1454,
  [4342] = 4330,
  [4343] = 1555,
  [4344] = 1438,
  [4345] = 1558,
  [4346] = 1561,
  [4347] = 1553,
  [4348] = 1557,
  [4349] = 1560,
  [4350] = 1570,
  [4351] = 1563,
  [4352] = 1564,
  [4353] = 1565,
  [4354] = 1567,
  [4355] = 1559,
  [4356] = 1562,
  [4357] = 1566,
  [4358] = 1556,
  [4359] = 1454,
  [4360] = 4360,
  [4361] = 4361,
  [4362] = 4362,
  [4363] = 1562,
  [4364] = 1566,
  [4365] = 1558,
  [4366] = 1553,
  [4367] = 1560,
  [4368] = 1570,
  [4369] = 1563,
  [4370] = 1564,
  [4371] = 1565,
  [4372] = 1567,
  [4373] = 1559,
  [4374] = 1555,
  [4375] = 1557,
  [4376] = 1561,
  [4377] = 1556,
  [4378] = 4378,
  [4379] = 4379,
  [4380] = 4378,
  [4381] = 4381,
  [4382] = 1567,
  [4383] = 1556,
  [4384] = 1557,
  [4385] = 1558,
  [4386] = 1553,
  [4387] = 1560,
  [4388] = 1570,
  [4389] = 1563,
  [4390] = 1564,
  [4391] = 1565,
  [4392] = 1559,
  [4393] = 1562,
  [4394] = 1566,
  [4395] = 1555,
  [4396] = 1561,
  [4397] = 4397,
  [4398] = 4398,
  [4399] = 4399,
  [4400] = 4400,
  [4401] = 4401,
  [4402] = 4402,
  [4403] = 4403,
  [4404] = 4403,
  [4405] = 4405,
  [4406] = 4400,
  [4407] = 4407,
  [4408] = 4408,
  [4409] = 4400,
  [4410] = 4400,
  [4411] = 4401,
  [4412] = 4400,
  [4413] = 4400,
  [4414] = 4400,
  [4415] = 4415,
  [4416] = 4400,
  [4417] = 1570,
  [4418] = 4418,
  [4419] = 4418,
  [4420] = 4420,
  [4421] = 1557,
  [4422] = 1558,
  [4423] = 1553,
  [4424] = 1560,
  [4425] = 4418,
  [4426] = 1563,
  [4427] = 1564,
  [4428] = 1565,
  [4429] = 1567,
  [4430] = 1559,
  [4431] = 1562,
  [4432] = 4420,
  [4433] = 1566,
  [4434] = 4397,
  [4435] = 1556,
  [4436] = 4436,
  [4437] = 4437,
  [4438] = 4438,
  [4439] = 4439,
  [4440] = 4438,
  [4441] = 4441,
  [4442] = 4442,
  [4443] = 4443,
  [4444] = 4444,
  [4445] = 4445,
  [4446] = 4418,
  [4447] = 4418,
  [4448] = 4448,
  [4449] = 4420,
  [4450] = 4420,
  [4451] = 4438,
  [4452] = 4452,
  [4453] = 4453,
  [4454] = 4454,
  [4455] = 4438,
  [4456] = 4420,
  [4457] = 4398,
  [4458] = 1555,
  [4459] = 1561,
  [4460] = 4438,
  [4461] = 4453,
  [4462] = 4462,
  [4463] = 4463,
  [4464] = 4418,
  [4465] = 4442,
  [4466] = 4420,
  [4467] = 4438,
  [4468] = 4418,
  [4469] = 4420,
  [4470] = 1438,
  [4471] = 4438,
  [4472] = 4472,
  [4473] = 4473,
  [4474] = 4474,
  [4475] = 4438,
  [4476] = 4476,
  [4477] = 4438,
  [4478] = 4438,
  [4479] = 4479,
  [4480] = 4480,
  [4481] = 4481,
  [4482] = 4438,
  [4483] = 4397,
  [4484] = 4438,
  [4485] = 4485,
  [4486] = 4486,
  [4487] = 4487,
  [4488] = 4488,
  [4489] = 4489,
  [4490] = 4462,
  [4491] = 4491,
  [4492] = 4492,
  [4493] = 4493,
  [4494] = 4494,
  [4495] = 4495,
  [4496] = 4491,
  [4497] = 4493,
  [4498] = 4498,
  [4499] = 4499,
  [4500] = 4500,
  [4501] = 4493,
  [4502] = 4488,
  [4503] = 4503,
  [4504] = 4504,
  [4505] = 4505,
  [4506] = 4506,
  [4507] = 4507,
  [4508] = 4508,
  [4509] = 4503,
  [4510] = 4504,
  [4511] = 4503,
  [4512] = 4504,
  [4513] = 4485,
  [4514] = 4503,
  [4515] = 4504,
  [4516] = 4503,
  [4517] = 4504,
  [4518] = 4518,
  [4519] = 4519,
  [4520] = 4520,
  [4521] = 4491,
  [4522] = 4493,
  [4523] = 4523,
  [4524] = 4486,
  [4525] = 4519,
  [4526] = 4503,
  [4527] = 4527,
  [4528] = 4528,
  [4529] = 4504,
  [4530] = 4530,
  [4531] = 4531,
  [4532] = 4532,
  [4533] = 4493,
  [4534] = 4488,
  [4535] = 4535,
  [4536] = 4489,
  [4537] = 4537,
  [4538] = 4535,
  [4539] = 4491,
  [4540] = 4485,
  [4541] = 4504,
  [4542] = 4542,
  [4543] = 4543,
  [4544] = 4492,
  [4545] = 4493,
  [4546] = 4519,
  [4547] = 2499,
  [4548] = 4488,
  [4549] = 4549,
  [4550] = 4491,
  [4551] = 4542,
  [4552] = 4503,
  [4553] = 4553,
  [4554] = 4493,
  [4555] = 4492,
  [4556] = 4488,
  [4557] = 4503,
  [4558] = 4558,
  [4559] = 4491,
  [4560] = 4492,
  [4561] = 4504,
  [4562] = 4549,
  [4563] = 4563,
  [4564] = 4564,
  [4565] = 4565,
  [4566] = 4566,
  [4567] = 4567,
  [4568] = 4568,
  [4569] = 4567,
  [4570] = 4567,
  [4571] = 4571,
  [4572] = 4572,
  [4573] = 4573,
  [4574] = 4574,
  [4575] = 4575,
  [4576] = 4576,
  [4577] = 4577,
  [4578] = 4578,
  [4579] = 4579,
  [4580] = 4580,
  [4581] = 4581,
  [4582] = 4572,
  [4583] = 4579,
  [4584] = 4572,
  [4585] = 4585,
  [4586] = 4586,
  [4587] = 4572,
  [4588] = 4588,
  [4589] = 4572,
  [4590] = 4590,
  [4591] = 4591,
  [4592] = 4592,
  [4593] = 4593,
  [4594] = 4594,
  [4595] = 4595,
  [4596] = 4596,
  [4597] = 4597,
  [4598] = 4573,
  [4599] = 4595,
  [4600] = 4574,
  [4601] = 4575,
  [4602] = 4571,
  [4603] = 4603,
  [4604] = 4603,
  [4605] = 4576,
  [4606] = 4579,
  [4607] = 4577,
  [4608] = 4578,
  [4609] = 4585,
  [4610] = 4580,
  [4611] = 4597,
  [4612] = 4612,
  [4613] = 4568,
  [4614] = 4572,
  [4615] = 4591,
  [4616] = 4573,
  [4617] = 4578,
  [4618] = 4603,
  [4619] = 4575,
  [4620] = 4592,
  [4621] = 4593,
  [4622] = 4594,
  [4623] = 1469,
  [4624] = 4624,
  [4625] = 4625,
  [4626] = 4626,
  [4627] = 4627,
  [4628] = 4628,
  [4629] = 4629,
  [4630] = 4630,
  [4631] = 4631,
  [4632] = 4632,
  [4633] = 4633,
  [4634] = 4628,
  [4635] = 4635,
  [4636] = 4636,
  [4637] = 4637,
  [4638] = 4638,
  [4639] = 4639,
  [4640] = 4640,
  [4641] = 4641,
  [4642] = 4642,
  [4643] = 4643,
  [4644] = 4644,
  [4645] = 4645,
  [4646] = 4642,
  [4647] = 4647,
  [4648] = 4648,
  [4649] = 4649,
  [4650] = 4650,
  [4651] = 4651,
  [4652] = 4643,
  [4653] = 4644,
  [4654] = 4645,
  [4655] = 4650,
  [4656] = 4627,
  [4657] = 4649,
  [4658] = 4649,
  [4659] = 4648,
  [4660] = 4624,
  [4661] = 4661,
  [4662] = 4662,
  [4663] = 4625,
  [4664] = 4664,
  [4665] = 4629,
  [4666] = 4649,
  [4667] = 4631,
  [4668] = 4626,
  [4669] = 4651,
  [4670] = 4632,
  [4671] = 4649,
  [4672] = 4639,
  [4673] = 4633,
  [4674] = 1469,
  [4675] = 4675,
  [4676] = 4636,
  [4677] = 4637,
  [4678] = 4662,
  [4679] = 4679,
  [4680] = 4649,
  [4681] = 4662,
  [4682] = 4650,
  [4683] = 4641,
  [4684] = 4635,
  [4685] = 4648,
  [4686] = 4640,
  [4687] = 4687,
  [4688] = 4688,
  [4689] = 4689,
  [4690] = 4687,
  [4691] = 4691,
  [4692] = 4692,
  [4693] = 4693,
  [4694] = 4694,
  [4695] = 4695,
  [4696] = 4696,
  [4697] = 4697,
  [4698] = 4691,
  [4699] = 4689,
  [4700] = 4687,
  [4701] = 4691,
  [4702] = 4692,
  [4703] = 4693,
  [4704] = 4694,
  [4705] = 4695,
  [4706] = 4696,
  [4707] = 4697,
  [4708] = 4688,
  [4709] = 4689,
  [4710] = 4687,
  [4711] = 4691,
  [4712] = 4692,
  [4713] = 4693,
  [4714] = 4694,
  [4715] = 4695,
  [4716] = 4696,
  [4717] = 4697,
  [4718] = 4688,
  [4719] = 4689,
  [4720] = 4687,
  [4721] = 4691,
  [4722] = 4693,
  [4723] = 4696,
  [4724] = 4694,
  [4725] = 4695,
  [4726] = 4697,
  [4727] = 4692,
  [4728] = 4688,
  [4729] = 4691,
  [4730] = 4692,
  [4731] = 4689,
  [4732] = 4696,
  [4733] = 4697,
  [4734] = 4693,
  [4735] = 4694,
  [4736] = 4695,
  [4737] = 4688,
  [4738] = 4689,
  [4739] = 4687,
  [4740] = 4691,
  [4741] = 4692,
  [4742] = 4693,
  [4743] = 4694,
  [4744] = 4695,
  [4745] = 4696,
  [4746] = 4697,
  [4747] = 4696,
  [4748] = 4688,
  [4749] = 4697,
  [4750] = 4688,
  [4751] = 4689,
  [4752] = 4687,
  [4753] = 4689,
  [4754] = 4691,
  [4755] = 4687,
  [4756] = 4691,
  [4757] = 4692,
  [4758] = 4693,
  [4759] = 4692,
  [4760] = 4693,
  [4761] = 4694,
  [4762] = 4695,
  [4763] = 4696,
  [4764] = 4697,
  [4765] = 4688,
  [4766] = 4692,
  [4767] = 4689,
  [4768] = 4687,
  [4769] = 4693,
  [4770] = 4691,
  [4771] = 4694,
  [4772] = 4695,
  [4773] = 4692,
  [4774] = 4693,
  [4775] = 4694,
  [4776] = 4694,
  [4777] = 4695,
  [4778] = 4695,
  [4779] = 4696,
  [4780] = 4697,
  [4781] = 4688,
  [4782] = 4696,
  [4783] = 4697,
  [4784] = 4689,
  [4785] = 4687,
  [4786] = 4688,
  [4787] = 4787,
  [4788] = 4788,
  [4789] = 4789,
  [4790] = 4790,
  [4791] = 4790,
  [4792] = 4792,
  [4793] = 4790,
  [4794] = 4788,
  [4795] = 4795,
  [4796] = 4796,
  [4797] = 4795,
  [4798] = 4796,
  [4799] = 4799,
  [4800] = 4795,
  [4801] = 4801,
  [4802] = 4802,
  [4803] = 4796,
  [4804] = 2552,
  [4805] = 4799,
  [4806] = 4796,
  [4807] = 4796,
  [4808] = 4801,
  [4809] = 4802,
  [4810] = 4810,
  [4811] = 4810,
  [4812] = 4810,
  [4813] = 4810,
  [4814] = 4810,
  [4815] = 4815,
  [4816] = 4816,
  [4817] = 4816,
  [4818] = 4810,
  [4819] = 4816,
  [4820] = 4810,
  [4821] = 4821,
  [4822] = 4810,
  [4823] = 4823,
  [4824] = 4824,
  [4825] = 4825,
  [4826] = 4826,
  [4827] = 4823,
  [4828] = 4823,
  [4829] = 4829,
  [4830] = 4830,
  [4831] = 4831,
  [4832] = 4825,
  [4833] = 4826,
  [4834] = 4834,
  [4835] = 4835,
  [4836] = 4836,
  [4837] = 4837,
  [4838] = 4838,
  [4839] = 4839,
  [4840] = 4840,
  [4841] = 4837,
  [4842] = 4842,
  [4843] = 4843,
  [4844] = 2902,
  [4845] = 4845,
  [4846] = 4846,
  [4847] = 4847,
  [4848] = 4848,
  [4849] = 4849,
  [4850] = 4843,
  [4851] = 4851,
  [4852] = 4843,
  [4853] = 4837,
  [4854] = 4854,
  [4855] = 4846,
  [4856] = 4856,
  [4857] = 4837,
  [4858] = 4843,
  [4859] = 4837,
  [4860] = 4860,
  [4861] = 4861,
  [4862] = 4846,
  [4863] = 4863,
  [4864] = 4864,
  [4865] = 4865,
  [4866] = 4866,
  [4867] = 4867,
  [4868] = 4846,
  [4869] = 4869,
  [4870] = 4870,
  [4871] = 4871,
  [4872] = 4871,
  [4873] = 4873,
  [4874] = 4873,
  [4875] = 4871,
  [4876] = 4871,
  [4877] = 4873,
  [4878] = 2679,
  [4879] = 4871,
  [4880] = 2682,
  [4881] = 4881,
  [4882] = 4882,
  [4883] = 2681,
  [4884] = 4871,
  [4885] = 4871,
  [4886] = 2678,
  [4887] = 179,
  [4888] = 4888,
  [4889] = 4889,
  [4890] = 4890,
  [4891] = 4891,
  [4892] = 4891,
  [4893] = 4893,
  [4894] = 4894,
  [4895] = 4895,
  [4896] = 4896,
  [4897] = 4897,
  [4898] = 4893,
  [4899] = 4899,
  [4900] = 4900,
  [4901] = 4901,
  [4902] = 4902,
  [4903] = 4903,
  [4904] = 4904,
  [4905] = 4905,
  [4906] = 4888,
  [4907] = 4896,
  [4908] = 4905,
  [4909] = 4909,
  [4910] = 4910,
  [4911] = 4911,
  [4912] = 4912,
  [4913] = 4913,
  [4914] = 4914,
  [4915] = 4893,
  [4916] = 4913,
  [4917] = 4917,
  [4918] = 4918,
  [4919] = 4909,
  [4920] = 4920,
  [4921] = 4921,
  [4922] = 4893,
  [4923] = 4923,
  [4924] = 4893,
  [4925] = 4925,
  [4926] = 4893,
  [4927] = 4914,
  [4928] = 4928,
  [4929] = 4891,
  [4930] = 4893,
  [4931] = 4912,
  [4932] = 4890,
  [4933] = 4895,
  [4934] = 4900,
  [4935] = 4901,
  [4936] = 4903,
  [4937] = 4913,
  [4938] = 180,
  [4939] = 4920,
  [4940] = 4895,
  [4941] = 4900,
  [4942] = 4901,
  [4943] = 4903,
  [4944] = 4911,
  [4945] = 4897,
  [4946] = 4946,
  [4947] = 4912,
  [4948] = 4923,
  [4949] = 4949,
  [4950] = 4950,
  [4951] = 4951,
  [4952] = 4952,
  [4953] = 4953,
  [4954] = 4950,
  [4955] = 4955,
  [4956] = 4956,
  [4957] = 4953,
  [4958] = 4953,
  [4959] = 4959,
  [4960] = 4956,
  [4961] = 4952,
  [4962] = 4959,
  [4963] = 4963,
  [4964] = 4959,
  [4965] = 4965,
  [4966] = 4966,
  [4967] = 4953,
  [4968] = 4952,
  [4969] = 4965,
  [4970] = 4950,
  [4971] = 4959,
  [4972] = 4963,
  [4973] = 4951,
  [4974] = 4974,
  [4975] = 4951,
  [4976] = 4976,
  [4977] = 4953,
  [4978] = 4978,
  [4979] = 4979,
  [4980] = 4980,
  [4981] = 4951,
  [4982] = 4953,
  [4983] = 4953,
  [4984] = 4965,
  [4985] = 4985,
  [4986] = 4986,
  [4987] = 4956,
  [4988] = 4978,
  [4989] = 4979,
  [4990] = 4990,
  [4991] = 4991,
  [4992] = 4952,
  [4993] = 4993,
  [4994] = 4953,
  [4995] = 4995,
  [4996] = 4959,
  [4997] = 4963,
  [4998] = 4963,
  [4999] = 4990,
  [5000] = 4991,
  [5001] = 4950,
  [5002] = 4980,
  [5003] = 4950,
  [5004] = 5004,
  [5005] = 5005,
  [5006] = 5005,
  [5007] = 5007,
  [5008] = 5008,
  [5009] = 5009,
  [5010] = 5010,
  [5011] = 5011,
  [5012] = 2753,
  [5013] = 2754,
  [5014] = 5014,
  [5015] = 5010,
  [5016] = 5011,
  [5017] = 5004,
  [5018] = 5004,
  [5019] = 5005,
  [5020] = 4966,
  [5021] = 5021,
  [5022] = 5022,
  [5023] = 5022,
  [5024] = 2748,
  [5025] = 5025,
  [5026] = 5026,
  [5027] = 5025,
  [5028] = 4867,
  [5029] = 5005,
  [5030] = 5030,
  [5031] = 5031,
  [5032] = 2750,
  [5033] = 4974,
  [5034] = 5034,
  [5035] = 5035,
  [5036] = 5036,
  [5037] = 5022,
  [5038] = 5038,
  [5039] = 5031,
  [5040] = 5022,
  [5041] = 5005,
  [5042] = 5026,
  [5043] = 5043,
  [5044] = 5009,
  [5045] = 5038,
  [5046] = 5025,
  [5047] = 5005,
  [5048] = 5048,
  [5049] = 5031,
  [5050] = 4910,
  [5051] = 5051,
  [5052] = 5052,
  [5053] = 5022,
  [5054] = 5007,
  [5055] = 5010,
  [5056] = 5056,
  [5057] = 5009,
  [5058] = 2687,
  [5059] = 5059,
  [5060] = 5056,
  [5061] = 5038,
  [5062] = 5062,
  [5063] = 4851,
  [5064] = 5064,
  [5065] = 4854,
  [5066] = 5066,
  [5067] = 5059,
  [5068] = 5022,
  [5069] = 5069,
  [5070] = 5070,
  [5071] = 5034,
  [5072] = 5062,
  [5073] = 5073,
  [5074] = 5074,
  [5075] = 5056,
  [5076] = 5008,
  [5077] = 5070,
  [5078] = 5078,
  [5079] = 5066,
  [5080] = 5005,
  [5081] = 5011,
  [5082] = 4865,
  [5083] = 5007,
  [5084] = 5034,
  [5085] = 2708,
  [5086] = 5036,
  [5087] = 5062,
  [5088] = 4866,
  [5089] = 4845,
  [5090] = 5005,
  [5091] = 5022,
  [5092] = 5064,
  [5093] = 5014,
  [5094] = 5064,
  [5095] = 5095,
  [5096] = 2356,
  [5097] = 5097,
  [5098] = 2381,
  [5099] = 5095,
  [5100] = 2348,
  [5101] = 349,
  [5102] = 1075,
  [5103] = 5103,
  [5104] = 5104,
  [5105] = 2372,
  [5106] = 5106,
  [5107] = 282,
  [5108] = 5108,
  [5109] = 2357,
  [5110] = 5110,
  [5111] = 5110,
  [5112] = 5112,
  [5113] = 5113,
  [5114] = 5103,
  [5115] = 5115,
  [5116] = 5115,
  [5117] = 5117,
  [5118] = 5118,
  [5119] = 5103,
  [5120] = 5095,
  [5121] = 5121,
  [5122] = 5103,
  [5123] = 5115,
  [5124] = 5124,
  [5125] = 5125,
  [5126] = 5126,
  [5127] = 5127,
  [5128] = 5106,
  [5129] = 5110,
  [5130] = 5103,
  [5131] = 5106,
  [5132] = 5132,
  [5133] = 312,
  [5134] = 5110,
  [5135] = 2354,
  [5136] = 5106,
  [5137] = 5125,
  [5138] = 5138,
  [5139] = 5139,
  [5140] = 5138,
  [5141] = 5115,
  [5142] = 5142,
  [5143] = 5143,
  [5144] = 5144,
  [5145] = 5145,
  [5146] = 5146,
  [5147] = 1074,
  [5148] = 5148,
  [5149] = 5095,
  [5150] = 5106,
  [5151] = 5151,
  [5152] = 2340,
  [5153] = 5115,
  [5154] = 5154,
  [5155] = 5095,
  [5156] = 5156,
  [5157] = 2353,
  [5158] = 5110,
  [5159] = 5159,
  [5160] = 5144,
  [5161] = 2349,
  [5162] = 2368,
  [5163] = 5163,
  [5164] = 3260,
  [5165] = 5165,
  [5166] = 5166,
  [5167] = 5167,
  [5168] = 5168,
  [5169] = 5169,
  [5170] = 5170,
  [5171] = 5171,
  [5172] = 5172,
  [5173] = 5173,
  [5174] = 5174,
  [5175] = 5175,
  [5176] = 5163,
  [5177] = 5177,
  [5178] = 5178,
  [5179] = 5179,
  [5180] = 5168,
  [5181] = 5181,
  [5182] = 5182,
  [5183] = 5183,
  [5184] = 5184,
  [5185] = 5185,
  [5186] = 5186,
  [5187] = 3246,
  [5188] = 5188,
  [5189] = 5189,
  [5190] = 5165,
  [5191] = 5191,
  [5192] = 5163,
  [5193] = 5166,
  [5194] = 5194,
  [5195] = 5195,
  [5196] = 5166,
  [5197] = 5173,
  [5198] = 5198,
  [5199] = 5199,
  [5200] = 5200,
  [5201] = 5201,
  [5202] = 5202,
  [5203] = 3263,
  [5204] = 5171,
  [5205] = 3244,
  [5206] = 5172,
  [5207] = 5168,
  [5208] = 5208,
  [5209] = 5175,
  [5210] = 5210,
  [5211] = 5165,
  [5212] = 5185,
  [5213] = 5191,
  [5214] = 5199,
  [5215] = 5185,
  [5216] = 3245,
  [5217] = 5168,
  [5218] = 5177,
  [5219] = 5175,
  [5220] = 5220,
  [5221] = 5185,
  [5222] = 5222,
  [5223] = 5165,
  [5224] = 5224,
  [5225] = 5225,
  [5226] = 5074,
  [5227] = 5227,
  [5228] = 5167,
  [5229] = 5229,
  [5230] = 5165,
  [5231] = 5199,
  [5232] = 5232,
  [5233] = 5233,
  [5234] = 5234,
  [5235] = 5235,
  [5236] = 5222,
  [5237] = 5202,
  [5238] = 5174,
  [5239] = 5191,
  [5240] = 5168,
  [5241] = 5191,
  [5242] = 5165,
  [5243] = 5243,
  [5244] = 5220,
  [5245] = 5163,
  [5246] = 3250,
  [5247] = 5247,
  [5248] = 5166,
  [5249] = 5199,
  [5250] = 5168,
  [5251] = 5179,
  [5252] = 5202,
  [5253] = 5165,
  [5254] = 5177,
  [5255] = 5181,
  [5256] = 5256,
  [5257] = 5179,
  [5258] = 5195,
  [5259] = 5175,
  [5260] = 5175,
  [5261] = 5185,
  [5262] = 5168,
  [5263] = 5191,
  [5264] = 5179,
  [5265] = 5265,
  [5266] = 5266,
  [5267] = 5225,
  [5268] = 5165,
  [5269] = 5181,
  [5270] = 5198,
  [5271] = 5271,
  [5272] = 5272,
  [5273] = 5273,
  [5274] = 5199,
  [5275] = 5185,
  [5276] = 5276,
  [5277] = 5277,
  [5278] = 5177,
  [5279] = 5202,
  [5280] = 5198,
  [5281] = 5199,
  [5282] = 5195,
  [5283] = 5185,
  [5284] = 5166,
  [5285] = 5175,
  [5286] = 5210,
  [5287] = 5173,
  [5288] = 5288,
  [5289] = 5289,
  [5290] = 5199,
  [5291] = 5291,
  [5292] = 5199,
  [5293] = 5293,
  [5294] = 5294,
  [5295] = 5181,
  [5296] = 5179,
  [5297] = 5179,
  [5298] = 5166,
  [5299] = 5191,
  [5300] = 5300,
  [5301] = 5210,
  [5302] = 5225,
  [5303] = 5167,
  [5304] = 5174,
  [5305] = 5163,
  [5306] = 5220,
  [5307] = 5172,
  [5308] = 5199,
  [5309] = 5163,
  [5310] = 5310,
  [5311] = 5222,
  [5312] = 5179,
  [5313] = 5195,
  [5314] = 5314,
  [5315] = 5315,
  [5316] = 5316,
  [5317] = 5317,
  [5318] = 5318,
  [5319] = 5319,
  [5320] = 5320,
  [5321] = 5321,
  [5322] = 5322,
  [5323] = 5323,
  [5324] = 5324,
  [5325] = 5325,
  [5326] = 5314,
  [5327] = 5316,
  [5328] = 5318,
  [5329] = 5319,
  [5330] = 5325,
  [5331] = 5331,
  [5332] = 5332,
  [5333] = 5333,
  [5334] = 5334,
  [5335] = 5335,
  [5336] = 5336,
  [5337] = 5337,
  [5338] = 5338,
  [5339] = 5339,
  [5340] = 5340,
  [5341] = 5341,
  [5342] = 5342,
  [5343] = 5315,
  [5344] = 5344,
  [5345] = 5345,
  [5346] = 5314,
  [5347] = 5325,
  [5348] = 5333,
  [5349] = 5335,
  [5350] = 5350,
  [5351] = 5351,
  [5352] = 5352,
  [5353] = 5353,
  [5354] = 5314,
  [5355] = 5325,
  [5356] = 5333,
  [5357] = 5333,
  [5358] = 5320,
  [5359] = 5359,
  [5360] = 5340,
  [5361] = 5361,
  [5362] = 5362,
  [5363] = 5363,
  [5364] = 5364,
  [5365] = 5365,
  [5366] = 5366,
  [5367] = 5367,
  [5368] = 5331,
  [5369] = 5332,
  [5370] = 5370,
  [5371] = 5333,
  [5372] = 5372,
  [5373] = 5370,
  [5374] = 5316,
  [5375] = 5318,
  [5376] = 5319,
  [5377] = 5335,
  [5378] = 5378,
  [5379] = 5379,
  [5380] = 5380,
  [5381] = 5381,
  [5382] = 5321,
  [5383] = 5383,
  [5384] = 5384,
  [5385] = 5385,
  [5386] = 5386,
  [5387] = 5387,
  [5388] = 5388,
  [5389] = 5389,
  [5390] = 5381,
  [5391] = 5391,
  [5392] = 5322,
  [5393] = 5315,
  [5394] = 5394,
  [5395] = 5395,
  [5396] = 5334,
  [5397] = 5335,
  [5398] = 5398,
  [5399] = 5399,
  [5400] = 5320,
  [5401] = 5370,
  [5402] = 5402,
  [5403] = 5403,
  [5404] = 5404,
  [5405] = 5405,
  [5406] = 5406,
  [5407] = 5407,
  [5408] = 5408,
  [5409] = 5409,
  [5410] = 5410,
  [5411] = 5411,
  [5412] = 5412,
  [5413] = 5413,
  [5414] = 5414,
  [5415] = 5415,
  [5416] = 5416,
  [5417] = 5417,
  [5418] = 5418,
  [5419] = 5419,
  [5420] = 5420,
  [5421] = 5421,
  [5422] = 5422,
  [5423] = 5423,
  [5424] = 5424,
  [5425] = 5335,
  [5426] = 5334,
  [5427] = 5427,
  [5428] = 5428,
  [5429] = 5429,
  [5430] = 5362,
  [5431] = 5337,
  [5432] = 5432,
  [5433] = 5433,
  [5434] = 5434,
  [5435] = 5384,
  [5436] = 5381,
  [5437] = 5367,
  [5438] = 5420,
  [5439] = 5439,
  [5440] = 5379,
  [5441] = 5324,
  [5442] = 5389,
  [5443] = 5394,
  [5444] = 5398,
  [5445] = 5409,
  [5446] = 5446,
  [5447] = 5447,
  [5448] = 5405,
  [5449] = 5406,
  [5450] = 5450,
  [5451] = 5352,
  [5452] = 5452,
  [5453] = 5416,
  [5454] = 5417,
  [5455] = 5447,
  [5456] = 5409,
  [5457] = 5457,
  [5458] = 5458,
  [5459] = 5459,
  [5460] = 5460,
  [5461] = 5384,
  [5462] = 5462,
  [5463] = 5381,
  [5464] = 5464,
  [5465] = 5465,
  [5466] = 5315,
  [5467] = 5398,
  [5468] = 5398,
  [5469] = 5323,
  [5470] = 5405,
  [5471] = 5406,
  [5472] = 5402,
  [5473] = 5416,
  [5474] = 5417,
  [5475] = 5475,
  [5476] = 5476,
  [5477] = 5452,
  [5478] = 5381,
  [5479] = 5479,
  [5480] = 5480,
  [5481] = 5481,
  [5482] = 5331,
  [5483] = 5483,
  [5484] = 5398,
  [5485] = 5386,
  [5486] = 5486,
  [5487] = 5487,
  [5488] = 5405,
  [5489] = 5406,
  [5490] = 5388,
  [5491] = 5416,
  [5492] = 5417,
  [5493] = 5352,
  [5494] = 5381,
  [5495] = 5495,
  [5496] = 5398,
  [5497] = 5332,
  [5498] = 5323,
  [5499] = 5405,
  [5500] = 5406,
  [5501] = 5452,
  [5502] = 5416,
  [5503] = 5417,
  [5504] = 5481,
  [5505] = 5381,
  [5506] = 5405,
  [5507] = 5507,
  [5508] = 5508,
  [5509] = 5418,
  [5510] = 5416,
  [5511] = 5417,
  [5512] = 5512,
  [5513] = 5513,
  [5514] = 5514,
  [5515] = 5515,
  [5516] = 5352,
  [5517] = 5324,
  [5518] = 5417,
  [5519] = 5519,
  [5520] = 5520,
  [5521] = 5314,
  [5522] = 5522,
  [5523] = 5523,
  [5524] = 5403,
  [5525] = 5525,
  [5526] = 5418,
  [5527] = 5407,
  [5528] = 5434,
  [5529] = 5439,
  [5530] = 5381,
  [5531] = 5531,
  [5532] = 5532,
  [5533] = 5344,
  [5534] = 5534,
  [5535] = 5535,
  [5536] = 5536,
  [5537] = 5315,
  [5538] = 5538,
  [5539] = 5316,
  [5540] = 5318,
  [5541] = 5319,
  [5542] = 5542,
  [5543] = 5543,
  [5544] = 5321,
  [5545] = 5545,
  [5546] = 5325,
  [5547] = 5402,
  [5548] = 5341,
  [5549] = 5427,
  [5550] = 5342,
  [5551] = 5151,
  [5552] = 5344,
  [5553] = 5553,
  [5554] = 5554,
  [5555] = 5324,
  [5556] = 5337,
  [5557] = 5481,
  [5558] = 5534,
  [5559] = 5559,
  [5560] = 5560,
  [5561] = 5331,
  [5562] = 5367,
  [5563] = 5332,
  [5564] = 5564,
  [5565] = 5379,
  [5566] = 5333,
  [5567] = 5447,
  [5568] = 5389,
  [5569] = 5386,
  [5570] = 5388,
  [5571] = 5571,
  [5572] = 5334,
  [5573] = 5394,
  [5574] = 5416,
  [5575] = 5575,
  [5576] = 5576,
  [5577] = 5559,
  [5578] = 5578,
  [5579] = 5335,
  [5580] = 5580,
  [5581] = 5320,
  [5582] = 5411,
  [5583] = 5433,
  [5584] = 197,
  [5585] = 5585,
  [5586] = 5323,
  [5587] = 5395,
  [5588] = 5414,
  [5589] = 5589,
  [5590] = 5452,
  [5591] = 5591,
  [5592] = 5515,
  [5593] = 5593,
  [5594] = 5481,
  [5595] = 5352,
  [5596] = 5591,
  [5597] = 5597,
  [5598] = 5598,
  [5599] = 5598,
  [5600] = 5600,
  [5601] = 5601,
  [5602] = 5315,
  [5603] = 5603,
  [5604] = 5604,
  [5605] = 5402,
  [5606] = 5571,
  [5607] = 5325,
  [5608] = 5417,
  [5609] = 5536,
  [5610] = 5384,
  [5611] = 5611,
  [5612] = 5402,
  [5613] = 5324,
  [5614] = 5391,
  [5615] = 5615,
  [5616] = 5314,
  [5617] = 5617,
  [5618] = 5618,
  [5619] = 5597,
  [5620] = 5620,
  [5621] = 5621,
  [5622] = 5315,
  [5623] = 5578,
  [5624] = 5624,
  [5625] = 5427,
  [5626] = 5452,
  [5627] = 5481,
  [5628] = 5628,
  [5629] = 5314,
  [5630] = 5403,
  [5631] = 5407,
  [5632] = 5434,
  [5633] = 5439,
  [5634] = 5634,
  [5635] = 5403,
  [5636] = 5407,
  [5637] = 5637,
  [5638] = 5415,
  [5639] = 5423,
  [5640] = 5434,
  [5641] = 5439,
  [5642] = 5446,
  [5643] = 5487,
  [5644] = 5535,
  [5645] = 5538,
  [5646] = 5542,
  [5647] = 5647,
  [5648] = 5324,
  [5649] = 5649,
  [5650] = 5334,
  [5651] = 5651,
  [5652] = 5316,
  [5653] = 5314,
  [5654] = 5318,
  [5655] = 5319,
  [5656] = 5656,
  [5657] = 5434,
  [5658] = 5439,
  [5659] = 5321,
  [5660] = 5322,
  [5661] = 5457,
  [5662] = 5662,
  [5663] = 5325,
  [5664] = 5664,
  [5665] = 5405,
  [5666] = 5666,
  [5667] = 5667,
  [5668] = 5668,
  [5669] = 5669,
  [5670] = 5316,
  [5671] = 5318,
  [5672] = 5319,
  [5673] = 5673,
  [5674] = 5321,
  [5675] = 5325,
  [5676] = 5341,
  [5677] = 5342,
  [5678] = 5344,
  [5679] = 5344,
  [5680] = 5385,
  [5681] = 5601,
  [5682] = 5682,
  [5683] = 5683,
  [5684] = 5684,
  [5685] = 5331,
  [5686] = 5332,
  [5687] = 5597,
  [5688] = 5333,
  [5689] = 5362,
  [5690] = 5334,
  [5691] = 5691,
  [5692] = 5341,
  [5693] = 5331,
  [5694] = 5332,
  [5695] = 5342,
  [5696] = 5621,
  [5697] = 5691,
  [5698] = 5698,
  [5699] = 5335,
  [5700] = 5333,
  [5701] = 5701,
  [5702] = 5406,
  [5703] = 5151,
  [5704] = 5386,
  [5705] = 5388,
  [5706] = 5335,
  [5707] = 5508,
  [5708] = 5512,
  [5709] = 5698,
  [5710] = 5391,
  [5711] = 5662,
  [5712] = 5712,
  [5713] = 5713,
  [5714] = 5714,
  [5715] = 5715,
  [5716] = 5716,
  [5717] = 5712,
  [5718] = 5718,
  [5719] = 5719,
  [5720] = 5720,
  [5721] = 5721,
  [5722] = 5721,
  [5723] = 5723,
  [5724] = 5724,
  [5725] = 3232,
  [5726] = 5714,
  [5727] = 5727,
  [5728] = 5728,
  [5729] = 5712,
  [5730] = 5730,
  [5731] = 5731,
  [5732] = 5718,
  [5733] = 5719,
  [5734] = 5734,
  [5735] = 5727,
  [5736] = 5714,
  [5737] = 5737,
  [5738] = 5719,
  [5739] = 5739,
  [5740] = 5740,
  [5741] = 5728,
  [5742] = 5730,
  [5743] = 5719,
  [5744] = 5744,
  [5745] = 5745,
  [5746] = 5746,
  [5747] = 5737,
  [5748] = 5727,
  [5749] = 5749,
  [5750] = 5750,
  [5751] = 5720,
  [5752] = 5724,
  [5753] = 5753,
  [5754] = 5714,
  [5755] = 5727,
  [5756] = 5756,
  [5757] = 5749,
  [5758] = 5758,
  [5759] = 5759,
  [5760] = 5731,
  [5761] = 5761,
  [5762] = 5762,
  [5763] = 5763,
  [5764] = 5724,
  [5765] = 5765,
  [5766] = 5739,
  [5767] = 5767,
  [5768] = 5768,
  [5769] = 5769,
  [5770] = 5716,
  [5771] = 5771,
  [5772] = 5758,
  [5773] = 5773,
  [5774] = 5774,
  [5775] = 5720,
  [5776] = 5740,
  [5777] = 5718,
  [5778] = 5768,
  [5779] = 5779,
  [5780] = 5780,
  [5781] = 5781,
  [5782] = 5721,
  [5783] = 5730,
  [5784] = 5784,
  [5785] = 5785,
  [5786] = 5739,
  [5787] = 5787,
  [5788] = 5769,
  [5789] = 5773,
  [5790] = 1507,
  [5791] = 5727,
  [5792] = 5739,
  [5793] = 5793,
  [5794] = 5794,
  [5795] = 5795,
  [5796] = 5796,
  [5797] = 5769,
  [5798] = 5773,
  [5799] = 5799,
  [5800] = 5800,
  [5801] = 5731,
  [5802] = 5802,
  [5803] = 5716,
  [5804] = 5795,
  [5805] = 5720,
  [5806] = 5806,
  [5807] = 5807,
  [5808] = 5721,
  [5809] = 5809,
  [5810] = 5795,
  [5811] = 5811,
  [5812] = 5724,
  [5813] = 5813,
  [5814] = 5814,
  [5815] = 5815,
  [5816] = 5716,
  [5817] = 5817,
  [5818] = 5724,
  [5819] = 5720,
  [5820] = 5718,
  [5821] = 5724,
  [5822] = 5719,
  [5823] = 1468,
  [5824] = 5824,
  [5825] = 5773,
  [5826] = 5826,
  [5827] = 5827,
  [5828] = 5712,
  [5829] = 5749,
  [5830] = 5718,
  [5831] = 5731,
  [5832] = 5795,
  [5833] = 5833,
  [5834] = 5728,
  [5835] = 5737,
  [5836] = 5737,
  [5837] = 5837,
  [5838] = 5739,
  [5839] = 5727,
  [5840] = 5716,
  [5841] = 5841,
  [5842] = 5837,
  [5843] = 1506,
  [5844] = 5716,
  [5845] = 5845,
  [5846] = 5740,
  [5847] = 5758,
  [5848] = 5848,
  [5849] = 5740,
  [5850] = 5768,
  [5851] = 5768,
  [5852] = 5852,
  [5853] = 5719,
  [5854] = 5730,
  [5855] = 5855,
  [5856] = 5856,
  [5857] = 5768,
  [5858] = 5795,
  [5859] = 5734,
  [5860] = 5769,
  [5861] = 5758,
  [5862] = 5862,
  [5863] = 5721,
  [5864] = 5814,
  [5865] = 5865,
  [5866] = 5716,
  [5867] = 5867,
  [5868] = 5720,
  [5869] = 5721,
  [5870] = 5728,
  [5871] = 5871,
  [5872] = 5872,
  [5873] = 5730,
  [5874] = 5724,
  [5875] = 5721,
  [5876] = 5876,
  [5877] = 5773,
  [5878] = 5728,
  [5879] = 5739,
  [5880] = 5880,
  [5881] = 5881,
  [5882] = 5882,
  [5883] = 5739,
  [5884] = 5728,
  [5885] = 5719,
  [5886] = 5716,
  [5887] = 5887,
  [5888] = 5888,
  [5889] = 5727,
  [5890] = 5890,
  [5891] = 5719,
  [5892] = 5892,
  [5893] = 5720,
  [5894] = 5739,
  [5895] = 5719,
  [5896] = 5721,
  [5897] = 5897,
  [5898] = 5898,
  [5899] = 5739,
  [5900] = 5900,
  [5901] = 5901,
  [5902] = 5902,
  [5903] = 5903,
  [5904] = 5904,
  [5905] = 5905,
  [5906] = 5906,
  [5907] = 5907,
  [5908] = 5908,
  [5909] = 5909,
  [5910] = 5910,
  [5911] = 5911,
  [5912] = 5912,
  [5913] = 5913,
  [5914] = 5914,
  [5915] = 5915,
  [5916] = 5916,
  [5917] = 5917,
  [5918] = 5918,
  [5919] = 5919,
  [5920] = 5920,
  [5921] = 5921,
  [5922] = 5922,
  [5923] = 5923,
  [5924] = 5924,
  [5925] = 5925,
  [5926] = 5926,
  [5927] = 5927,
  [5928] = 5928,
  [5929] = 5929,
  [5930] = 5930,
  [5931] = 5931,
  [5932] = 5932,
  [5933] = 5933,
  [5934] = 5934,
  [5935] = 5935,
  [5936] = 5936,
  [5937] = 5937,
  [5938] = 5938,
  [5939] = 5939,
  [5940] = 5940,
  [5941] = 5941,
  [5942] = 5923,
  [5943] = 5901,
  [5944] = 5944,
  [5945] = 5945,
  [5946] = 5946,
  [5947] = 5947,
  [5948] = 5948,
  [5949] = 5949,
  [5950] = 5950,
  [5951] = 5951,
  [5952] = 5952,
  [5953] = 5953,
  [5954] = 5954,
  [5955] = 5955,
  [5956] = 5956,
  [5957] = 5957,
  [5958] = 5958,
  [5959] = 5909,
  [5960] = 5960,
  [5961] = 5961,
  [5962] = 5960,
  [5963] = 5963,
  [5964] = 5937,
  [5965] = 5965,
  [5966] = 5966,
  [5967] = 5967,
  [5968] = 5968,
  [5969] = 5969,
  [5970] = 5902,
  [5971] = 5971,
  [5972] = 5908,
  [5973] = 5973,
  [5974] = 5974,
  [5975] = 5975,
  [5976] = 5912,
  [5977] = 5922,
  [5978] = 5978,
  [5979] = 5979,
  [5980] = 5971,
  [5981] = 5904,
  [5982] = 5978,
  [5983] = 5906,
  [5984] = 5984,
  [5985] = 5985,
  [5986] = 5916,
  [5987] = 5917,
  [5988] = 5988,
  [5989] = 5924,
  [5990] = 5920,
  [5991] = 5991,
  [5992] = 5992,
  [5993] = 5993,
  [5994] = 5994,
  [5995] = 5925,
  [5996] = 5927,
  [5997] = 5928,
  [5998] = 5929,
  [5999] = 5930,
  [6000] = 6000,
  [6001] = 5932,
  [6002] = 5905,
  [6003] = 6003,
  [6004] = 5984,
  [6005] = 6005,
  [6006] = 6006,
  [6007] = 5941,
  [6008] = 6008,
  [6009] = 6009,
  [6010] = 5946,
  [6011] = 6011,
  [6012] = 5935,
  [6013] = 5949,
  [6014] = 5903,
  [6015] = 5951,
  [6016] = 5937,
  [6017] = 5939,
  [6018] = 6018,
  [6019] = 6019,
  [6020] = 5956,
  [6021] = 5945,
  [6022] = 5926,
  [6023] = 5960,
  [6024] = 5925,
  [6025] = 5909,
  [6026] = 5965,
  [6027] = 5923,
  [6028] = 5963,
  [6029] = 5966,
  [6030] = 6030,
  [6031] = 5902,
  [6032] = 6032,
  [6033] = 5926,
  [6034] = 5974,
  [6035] = 5908,
  [6036] = 5935,
  [6037] = 5912,
  [6038] = 5922,
  [6039] = 5988,
  [6040] = 5954,
  [6041] = 5904,
  [6042] = 5939,
  [6043] = 6043,
  [6044] = 5916,
  [6045] = 5917,
  [6046] = 5920,
  [6047] = 5971,
  [6048] = 5978,
  [6049] = 5984,
  [6050] = 5988,
  [6051] = 5925,
  [6052] = 5928,
  [6053] = 5932,
  [6054] = 5932,
  [6055] = 5991,
  [6056] = 5992,
  [6057] = 5994,
  [6058] = 5946,
  [6059] = 5905,
  [6060] = 5949,
  [6061] = 5991,
  [6062] = 5901,
  [6063] = 5969,
  [6064] = 5935,
  [6065] = 5992,
  [6066] = 5956,
  [6067] = 6067,
  [6068] = 5939,
  [6069] = 5960,
  [6070] = 5992,
  [6071] = 5965,
  [6072] = 5937,
  [6073] = 5994,
  [6074] = 5945,
  [6075] = 6075,
  [6076] = 5945,
  [6077] = 6077,
  [6078] = 6000,
  [6079] = 5974,
  [6080] = 6080,
  [6081] = 5902,
  [6082] = 5994,
  [6083] = 5908,
  [6084] = 5912,
  [6085] = 5922,
  [6086] = 5904,
  [6087] = 6087,
  [6088] = 5916,
  [6089] = 5917,
  [6090] = 5920,
  [6091] = 5988,
  [6092] = 5925,
  [6093] = 5928,
  [6094] = 5932,
  [6095] = 6043,
  [6096] = 5991,
  [6097] = 6097,
  [6098] = 5946,
  [6099] = 6099,
  [6100] = 6100,
  [6101] = 6101,
  [6102] = 5939,
  [6103] = 5956,
  [6104] = 5918,
  [6105] = 6105,
  [6106] = 6005,
  [6107] = 6107,
  [6108] = 6108,
  [6109] = 5945,
  [6110] = 5929,
  [6111] = 6077,
  [6112] = 6112,
  [6113] = 5926,
  [6114] = 5954,
  [6115] = 6115,
  [6116] = 6116,
  [6117] = 5915,
  [6118] = 5904,
  [6119] = 5916,
  [6120] = 5917,
  [6121] = 5920,
  [6122] = 5928,
  [6123] = 6107,
  [6124] = 5946,
  [6125] = 6125,
  [6126] = 5988,
  [6127] = 5974,
  [6128] = 6128,
  [6129] = 5905,
  [6130] = 5992,
  [6131] = 6131,
  [6132] = 5993,
  [6133] = 6133,
  [6134] = 5958,
  [6135] = 6135,
  [6136] = 5950,
  [6137] = 5916,
  [6138] = 5917,
  [6139] = 5920,
  [6140] = 5958,
  [6141] = 5946,
  [6142] = 5937,
  [6143] = 6143,
  [6144] = 5927,
  [6145] = 6145,
  [6146] = 6146,
  [6147] = 5953,
  [6148] = 6148,
  [6149] = 4627,
  [6150] = 5955,
  [6151] = 5920,
  [6152] = 5909,
  [6153] = 5946,
  [6154] = 5963,
  [6155] = 5966,
  [6156] = 6156,
  [6157] = 6157,
  [6158] = 6158,
  [6159] = 5939,
  [6160] = 6160,
  [6161] = 5924,
  [6162] = 5993,
  [6163] = 5920,
  [6164] = 5946,
  [6165] = 5920,
  [6166] = 5920,
  [6167] = 5920,
  [6168] = 5920,
  [6169] = 5920,
  [6170] = 5920,
  [6171] = 5920,
  [6172] = 5920,
  [6173] = 6173,
  [6174] = 6116,
  [6175] = 5915,
  [6176] = 6176,
  [6177] = 6177,
  [6178] = 6178,
  [6179] = 6100,
  [6180] = 6180,
  [6181] = 6181,
  [6182] = 6182,
  [6183] = 5921,
  [6184] = 6087,
  [6185] = 6185,
  [6186] = 6186,
  [6187] = 6187,
  [6188] = 6188,
  [6189] = 6189,
  [6190] = 6190,
  [6191] = 5903,
  [6192] = 6192,
  [6193] = 5916,
  [6194] = 5917,
  [6195] = 5961,
  [6196] = 6196,
  [6197] = 5920,
  [6198] = 6101,
  [6199] = 6101,
  [6200] = 5918,
  [6201] = 6201,
  [6202] = 5941,
  [6203] = 6203,
  [6204] = 5961,
  [6205] = 6080,
  [6206] = 5915,
  [6207] = 5902,
  [6208] = 6208,
  [6209] = 6209,
  [6210] = 5925,
  [6211] = 5908,
  [6212] = 5915,
  [6213] = 5912,
  [6214] = 5922,
  [6215] = 6215,
  [6216] = 5927,
  [6217] = 6217,
  [6218] = 6218,
  [6219] = 5928,
  [6220] = 5929,
  [6221] = 5930,
  [6222] = 6222,
  [6223] = 6223,
  [6224] = 5932,
  [6225] = 5930,
  [6226] = 6226,
  [6227] = 5993,
  [6228] = 6228,
  [6229] = 6196,
  [6230] = 5921,
  [6231] = 6231,
  [6232] = 5958,
  [6233] = 6233,
  [6234] = 6032,
  [6235] = 5961,
  [6236] = 6236,
  [6237] = 6237,
  [6238] = 5921,
  [6239] = 5903,
  [6240] = 5945,
  [6241] = 6241,
  [6242] = 6242,
  [6243] = 5971,
  [6244] = 5978,
  [6245] = 5918,
  [6246] = 6237,
  [6247] = 6143,
  [6248] = 5950,
  [6249] = 5971,
  [6250] = 5958,
  [6251] = 5978,
  [6252] = 5984,
  [6253] = 6253,
  [6254] = 6009,
  [6255] = 6018,
  [6256] = 5971,
  [6257] = 6222,
  [6258] = 5961,
  [6259] = 6259,
  [6260] = 5978,
  [6261] = 5975,
  [6262] = 6262,
  [6263] = 5938,
  [6264] = 5975,
  [6265] = 6265,
  [6266] = 5979,
  [6267] = 5984,
  [6268] = 6268,
  [6269] = 6269,
  [6270] = 6270,
  [6271] = 6271,
  [6272] = 5984,
  [6273] = 5904,
  [6274] = 6143,
  [6275] = 5941,
  [6276] = 5988,
  [6277] = 6188,
  [6278] = 5991,
  [6279] = 6279,
  [6280] = 6280,
  [6281] = 4635,
  [6282] = 6282,
  [6283] = 6283,
  [6284] = 6284,
  [6285] = 4625,
  [6286] = 5923,
  [6287] = 6287,
  [6288] = 5926,
  [6289] = 6289,
  [6290] = 5988,
  [6291] = 6291,
  [6292] = 6292,
  [6293] = 5935,
  [6294] = 5953,
  [6295] = 5991,
  [6296] = 5946,
  [6297] = 6297,
  [6298] = 6298,
  [6299] = 6299,
  [6300] = 6148,
  [6301] = 5954,
  [6302] = 6302,
  [6303] = 5923,
  [6304] = 6304,
  [6305] = 5992,
  [6306] = 6306,
  [6307] = 5926,
  [6308] = 6189,
  [6309] = 5994,
  [6310] = 6077,
  [6311] = 6311,
  [6312] = 5951,
  [6313] = 6313,
  [6314] = 5991,
  [6315] = 6146,
  [6316] = 5958,
  [6317] = 5949,
  [6318] = 6262,
  [6319] = 6319,
  [6320] = 6320,
  [6321] = 5951,
  [6322] = 6322,
  [6323] = 6009,
  [6324] = 6324,
  [6325] = 5961,
  [6326] = 5938,
  [6327] = 5928,
  [6328] = 5905,
  [6329] = 5935,
  [6330] = 6330,
  [6331] = 5954,
  [6332] = 5905,
  [6333] = 5903,
  [6334] = 6334,
  [6335] = 6335,
  [6336] = 5937,
  [6337] = 6185,
  [6338] = 5919,
  [6339] = 5901,
  [6340] = 6340,
  [6341] = 6196,
  [6342] = 6005,
  [6343] = 6343,
  [6344] = 6173,
  [6345] = 6345,
  [6346] = 6201,
  [6347] = 5902,
  [6348] = 5939,
  [6349] = 5956,
  [6350] = 5921,
  [6351] = 5913,
  [6352] = 5903,
  [6353] = 6018,
  [6354] = 5905,
  [6355] = 6107,
  [6356] = 5960,
  [6357] = 5937,
  [6358] = 6358,
  [6359] = 5921,
  [6360] = 5957,
  [6361] = 5903,
  [6362] = 6362,
  [6363] = 6363,
  [6364] = 5969,
  [6365] = 5950,
  [6366] = 5924,
  [6367] = 6107,
  [6368] = 6000,
  [6369] = 5945,
  [6370] = 5993,
  [6371] = 5960,
  [6372] = 5908,
  [6373] = 6373,
  [6374] = 6270,
  [6375] = 6176,
  [6376] = 6178,
  [6377] = 6100,
  [6378] = 5953,
  [6379] = 5940,
  [6380] = 6185,
  [6381] = 6363,
  [6382] = 6382,
  [6383] = 5913,
  [6384] = 6148,
  [6385] = 5940,
  [6386] = 6107,
  [6387] = 6340,
  [6388] = 5906,
  [6389] = 6043,
  [6390] = 6182,
  [6391] = 5967,
  [6392] = 6080,
  [6393] = 6393,
  [6394] = 5955,
  [6395] = 5919,
  [6396] = 6396,
  [6397] = 5915,
  [6398] = 5918,
  [6399] = 5939,
  [6400] = 5921,
  [6401] = 6396,
  [6402] = 5903,
  [6403] = 5945,
  [6404] = 6005,
  [6405] = 6262,
  [6406] = 6406,
  [6407] = 5915,
  [6408] = 5912,
  [6409] = 5955,
  [6410] = 6410,
  [6411] = 6411,
  [6412] = 5965,
  [6413] = 6413,
  [6414] = 5963,
  [6415] = 6415,
  [6416] = 6173,
  [6417] = 5919,
  [6418] = 6032,
  [6419] = 5958,
  [6420] = 5961,
  [6421] = 5909,
  [6422] = 5922,
  [6423] = 6423,
  [6424] = 5909,
  [6425] = 6107,
  [6426] = 5963,
  [6427] = 5975,
  [6428] = 6097,
  [6429] = 5975,
  [6430] = 5966,
  [6431] = 6158,
  [6432] = 5963,
  [6433] = 5979,
  [6434] = 5966,
  [6435] = 6435,
  [6436] = 6436,
  [6437] = 6009,
  [6438] = 5953,
  [6439] = 6000,
  [6440] = 5973,
  [6441] = 6176,
  [6442] = 6157,
  [6443] = 6324,
  [6444] = 5939,
  [6445] = 6445,
  [6446] = 6176,
  [6447] = 6178,
  [6448] = 6271,
  [6449] = 6075,
  [6450] = 6185,
  [6451] = 6178,
  [6452] = 6410,
  [6453] = 6335,
  [6454] = 5955,
  [6455] = 5975,
  [6456] = 6345,
  [6457] = 6158,
  [6458] = 6182,
  [6459] = 6080,
  [6460] = 5902,
  [6461] = 5923,
  [6462] = 6279,
  [6463] = 5953,
  [6464] = 6298,
  [6465] = 6148,
  [6466] = 5908,
  [6467] = 5955,
  [6468] = 5974,
  [6469] = 5926,
  [6470] = 5912,
  [6471] = 5935,
  [6472] = 5922,
  [6473] = 5967,
  [6474] = 6043,
  [6475] = 5954,
  [6476] = 5979,
  [6477] = 6176,
  [6478] = 6178,
  [6479] = 5966,
  [6480] = 6185,
  [6481] = 5971,
  [6482] = 6105,
  [6483] = 6145,
  [6484] = 5978,
  [6485] = 5901,
  [6486] = 5920,
  [6487] = 6080,
  [6488] = 6306,
  [6489] = 5984,
  [6490] = 6009,
  [6491] = 5937,
  [6492] = 6148,
  [6493] = 5988,
  [6494] = 5909,
  [6495] = 6311,
  [6496] = 6496,
  [6497] = 6101,
  [6498] = 5991,
  [6499] = 6499,
  [6500] = 5963,
  [6501] = 6501,
  [6502] = 5966,
  [6503] = 6176,
  [6504] = 5907,
  [6505] = 6157,
  [6506] = 6158,
  [6507] = 6507,
  [6508] = 6080,
  [6509] = 5939,
  [6510] = 5992,
  [6511] = 6182,
  [6512] = 5916,
  [6513] = 6513,
  [6514] = 5945,
  [6515] = 5904,
  [6516] = 5994,
  [6517] = 5949,
  [6518] = 6201,
  [6519] = 6145,
  [6520] = 6077,
  [6521] = 5906,
  [6522] = 6018,
  [6523] = 6176,
  [6524] = 5934,
  [6525] = 6107,
  [6526] = 6182,
  [6527] = 5905,
  [6528] = 6080,
  [6529] = 5955,
  [6530] = 5909,
  [6531] = 6107,
  [6532] = 6077,
  [6533] = 5963,
  [6534] = 6534,
  [6535] = 5966,
  [6536] = 5918,
  [6537] = 6345,
  [6538] = 6538,
  [6539] = 6157,
  [6540] = 5975,
  [6541] = 5979,
  [6542] = 6324,
  [6543] = 6543,
  [6544] = 6335,
  [6545] = 6545,
  [6546] = 6158,
  [6547] = 5902,
  [6548] = 6548,
  [6549] = 5908,
  [6550] = 5912,
  [6551] = 5922,
  [6552] = 6228,
  [6553] = 6553,
  [6554] = 5902,
  [6555] = 5923,
  [6556] = 6556,
  [6557] = 5916,
  [6558] = 5945,
  [6559] = 5917,
  [6560] = 6560,
  [6561] = 6561,
  [6562] = 6125,
  [6563] = 5908,
  [6564] = 6292,
  [6565] = 6236,
  [6566] = 5912,
  [6567] = 5953,
  [6568] = 6148,
  [6569] = 6187,
  [6570] = 6570,
  [6571] = 5955,
  [6572] = 5909,
  [6573] = 6215,
  [6574] = 5920,
  [6575] = 5963,
  [6576] = 6100,
  [6577] = 5935,
  [6578] = 6343,
  [6579] = 5926,
  [6580] = 5966,
  [6581] = 5921,
  [6582] = 5922,
  [6583] = 5903,
  [6584] = 6157,
  [6585] = 6393,
  [6586] = 6158,
  [6587] = 6587,
  [6588] = 5975,
  [6589] = 5935,
  [6590] = 6590,
  [6591] = 5925,
  [6592] = 6209,
  [6593] = 5937,
  [6594] = 5927,
  [6595] = 5928,
  [6596] = 6596,
  [6597] = 5937,
  [6598] = 6182,
  [6599] = 6107,
  [6600] = 5929,
  [6601] = 5930,
  [6602] = 5939,
  [6603] = 5932,
  [6604] = 6112,
  [6605] = 5950,
  [6606] = 5973,
  [6607] = 4626,
  [6608] = 6608,
  [6609] = 5924,
  [6610] = 6330,
  [6611] = 5902,
  [6612] = 6612,
  [6613] = 5917,
  [6614] = 5908,
  [6615] = 5912,
  [6616] = 5971,
  [6617] = 5978,
  [6618] = 5984,
  [6619] = 5922,
  [6620] = 6334,
  [6621] = 6009,
  [6622] = 6018,
  [6623] = 6186,
  [6624] = 5938,
  [6625] = 6436,
  [6626] = 5957,
  [6627] = 5945,
  [6628] = 6358,
  [6629] = 6345,
  [6630] = 5941,
  [6631] = 5904,
  [6632] = 5953,
  [6633] = 5956,
  [6634] = 5923,
  [6635] = 5955,
  [6636] = 6636,
  [6637] = 5946,
  [6638] = 6373,
  [6639] = 6306,
  [6640] = 5909,
  [6641] = 5963,
  [6642] = 5966,
  [6643] = 6643,
  [6644] = 5907,
  [6645] = 6587,
  [6646] = 5934,
  [6647] = 5949,
  [6648] = 6125,
  [6649] = 5950,
  [6650] = 5971,
  [6651] = 5921,
  [6652] = 5978,
  [6653] = 5903,
  [6654] = 6187,
  [6655] = 5984,
  [6656] = 6101,
  [6657] = 6009,
  [6658] = 6215,
  [6659] = 5937,
  [6660] = 6018,
  [6661] = 6107,
  [6662] = 5994,
  [6663] = 6663,
  [6664] = 5951,
  [6665] = 6133,
  [6666] = 5926,
  [6667] = 5902,
  [6668] = 6100,
  [6669] = 6143,
  [6670] = 6182,
  [6671] = 5956,
  [6672] = 5921,
  [6673] = 6143,
  [6674] = 5903,
  [6675] = 5957,
  [6676] = 6393,
  [6677] = 5988,
  [6678] = 6587,
  [6679] = 6363,
  [6680] = 6209,
  [6681] = 6411,
  [6682] = 5988,
  [6683] = 5991,
  [6684] = 5991,
  [6685] = 5965,
  [6686] = 6107,
  [6687] = 5908,
  [6688] = 5960,
  [6689] = 6146,
  [6690] = 6181,
  [6691] = 5912,
  [6692] = 6270,
  [6693] = 6693,
  [6694] = 6313,
  [6695] = 5923,
  [6696] = 6279,
  [6697] = 6157,
  [6698] = 5992,
  [6699] = 6699,
  [6700] = 5969,
  [6701] = 5994,
  [6702] = 6177,
  [6703] = 6077,
  [6704] = 6330,
  [6705] = 5922,
  [6706] = 6706,
  [6707] = 6663,
  [6708] = 5960,
  [6709] = 6268,
  [6710] = 6710,
  [6711] = 5905,
  [6712] = 6712,
  [6713] = 5924,
  [6714] = 6714,
  [6715] = 5969,
  [6716] = 5993,
  [6717] = 6160,
  [6718] = 6158,
  [6719] = 5965,
  [6720] = 5946,
  [6721] = 6721,
  [6722] = 5935,
  [6723] = 5921,
  [6724] = 6724,
  [6725] = 5903,
  [6726] = 6608,
  [6727] = 5905,
  [6728] = 6107,
  [6729] = 5918,
  [6730] = 5992,
  [6731] = 5994,
  [6732] = 5971,
  [6733] = 6100,
  [6734] = 6077,
  [6735] = 6393,
  [6736] = 6587,
  [6737] = 6223,
  [6738] = 5973,
  [6739] = 6739,
  [6740] = 6097,
  [6741] = 6209,
  [6742] = 6105,
  [6743] = 6100,
  [6744] = 5978,
  [6745] = 6393,
  [6746] = 6587,
  [6747] = 6747,
  [6748] = 6112,
  [6749] = 5969,
  [6750] = 6075,
  [6751] = 6100,
  [6752] = 6133,
  [6753] = 6393,
  [6754] = 6587,
  [6755] = 5924,
  [6756] = 6160,
  [6757] = 5993,
  [6758] = 6043,
  [6759] = 6223,
  [6760] = 6393,
  [6761] = 6587,
  [6762] = 6313,
  [6763] = 6393,
  [6764] = 6587,
  [6765] = 6765,
  [6766] = 6393,
  [6767] = 5984,
  [6768] = 6393,
  [6769] = 5954,
  [6770] = 6233,
  [6771] = 6299,
  [6772] = 5974,
  [6773] = 6322,
  [6774] = 5918,
  [6775] = 5988,
  [6776] = 6279,
  [6777] = 6237,
  [6778] = 5991,
  [6779] = 5918,
  [6780] = 6262,
  [6781] = 5921,
  [6782] = 6271,
  [6783] = 5992,
  [6784] = 6259,
  [6785] = 5994,
  [6786] = 5937,
  [6787] = 6242,
  [6788] = 5953,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym__abstract_declarator,
  sym__declarator,
  sym__field_declarator,
  sym__type_declarator,
  sym_expression,
  sym_specifier_qualifier,
  sym_statement,
  sym_type_specifier,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym__abstract_declarator] = {.index = 0, .length = 5},
  [sym__declarator] = {.index = 5, .length = 6},
  [sym__field_declarator] = {.index = 11, .length = 18},
  [sym__type_declarator] = {.index = 29, .length = 18},
  [sym_expression] = {.index = 47, .length = 36},
  [sym_specifier_qualifier] = {.index = 83, .length = 3},
  [sym_statement] = {.index = 86, .length = 18},
  [sym_type_specifier] = {.index = 104, .length = 11},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    sym_abstract_array_declarator,
    sym_abstract_block_pointer_declarator,
    sym_abstract_function_declarator,
    sym_abstract_parenthesized_declarator,
    sym_abstract_pointer_declarator,
  [5] =
    sym_array_declarator,
    sym_block_pointer_declarator,
    sym_function_declarator,
    sym_identifier,
    sym_parenthesized_declarator,
    sym_pointer_declarator,
  [11] =
    alias_sym_field_identifier,
    sym__old_style_function_declarator,
    sym_array_declarator,
    sym_array_field_declarator,
    sym_array_type_declarator,
    sym_attributed_field_declarator,
    sym_block_pointer_declarator,
    sym_block_pointer_field_declarator,
    sym_block_pointer_type_declarator,
    sym_function_declarator,
    sym_function_field_declarator,
    sym_function_type_declarator,
    sym_parenthesized_declarator,
    sym_parenthesized_field_declarator,
    sym_parenthesized_type_declarator,
    sym_pointer_declarator,
    sym_pointer_field_declarator,
    sym_pointer_type_declarator,
  [29] =
    alias_sym_type_identifier,
    sym__old_style_function_declarator,
    sym_array_declarator,
    sym_array_field_declarator,
    sym_array_type_declarator,
    sym_block_pointer_declarator,
    sym_block_pointer_field_declarator,
    sym_block_pointer_type_declarator,
    sym_function_declarator,
    sym_function_field_declarator,
    sym_function_type_declarator,
    sym_parenthesized_declarator,
    sym_parenthesized_field_declarator,
    sym_parenthesized_type_declarator,
    sym_pointer_declarator,
    sym_pointer_field_declarator,
    sym_pointer_type_declarator,
    sym_primitive_type,
  [47] =
    sym_alignof_expression,
    sym_array_literal,
    sym_assignment_expression,
    sym_at_expression,
    sym_available_expression,
    sym_binary_expression,
    sym_block_literal,
    sym_call_expression,
    sym_cast_expression,
    sym_char_literal,
    sym_compound_literal_expression,
    sym_concatenated_string,
    sym_conditional_expression,
    sym_dictionary_literal,
    sym_encode_expression,
    sym_extension_expression,
    sym_false,
    sym_field_expression,
    sym_generic_expression,
    sym_gnu_asm_expression,
    sym_identifier,
    sym_message_expression,
    sym_null,
    sym_number_literal,
    sym_offsetof_expression,
    sym_parenthesized_expression,
    sym_pointer_expression,
    sym_range_expression,
    sym_selector_expression,
    sym_sizeof_expression,
    sym_string_literal,
    sym_subscript_expression,
    sym_true,
    sym_unary_expression,
    sym_update_expression,
    sym_va_arg_expression,
  [83] =
    sym_protocol_qualifier,
    sym_type_qualifier,
    sym_type_specifier,
  [86] =
    sym_attributed_statement,
    sym_break_statement,
    sym_case_statement,
    sym_compound_statement,
    sym_continue_statement,
    sym_do_statement,
    sym_expression_statement,
    sym_for_statement,
    sym_goto_statement,
    sym_if_statement,
    sym_labeled_statement,
    sym_ms_asm_block,
    sym_return_statement,
    sym_switch_statement,
    sym_synchronized_statement,
    sym_throw_statement,
    sym_try_statement,
    sym_while_statement,
  [104] =
    alias_sym_type_identifier,
    sym_array_type_specifier,
    sym_enum_specifier,
    sym_generic_specifier,
    sym_macro_type_specifier,
    sym_primitive_type,
    sym_sized_type_specifier,
    sym_struct_specifier,
    sym_typedefed_specifier,
    sym_typeof_specifier,
    sym_union_specifier,
};

static const TSCharacterRange sym_identifier_character_set_1[] = {
  {'$', '$'}, {'A', 'Z'}, {'\\', '\\'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xba, 0xba},
  {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x370, 0x374},
  {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5},
  {0x3f7, 0x481}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x620, 0x64a},
  {0x66e, 0x66f}, {0x671, 0x6d3}, {0x6d5, 0x6d5}, {0x6e5, 0x6e6}, {0x6ee, 0x6ef}, {0x6fa, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x710},
  {0x712, 0x72f}, {0x74d, 0x7a5}, {0x7b1, 0x7b1}, {0x7ca, 0x7ea}, {0x7f4, 0x7f5}, {0x7fa, 0x7fa}, {0x800, 0x815}, {0x81a, 0x81a},
  {0x824, 0x824}, {0x828, 0x828}, {0x840, 0x858}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x8a0, 0x8c9}, {0x904, 0x939},
  {0x93d, 0x93d}, {0x950, 0x950}, {0x958, 0x961}, {0x971, 0x980}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0},
  {0x9b2, 0x9b2}, {0x9b6, 0x9b9}, {0x9bd, 0x9bd}, {0x9ce, 0x9ce}, {0x9dc, 0x9dd}, {0x9df, 0x9e1}, {0x9f0, 0x9f1}, {0x9fc, 0x9fc},
  {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39}, {0xa59, 0xa5c},
  {0xa5e, 0xa5e}, {0xa72, 0xa74}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9},
  {0xabd, 0xabd}, {0xad0, 0xad0}, {0xae0, 0xae1}, {0xaf9, 0xaf9}, {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30},
  {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3d, 0xb3d}, {0xb5c, 0xb5d}, {0xb5f, 0xb61}, {0xb71, 0xb71}, {0xb83, 0xb83}, {0xb85, 0xb8a},
  {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9},
  {0xbd0, 0xbd0}, {0xc05, 0xc0c}, {0xc0e, 0xc10}, {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3d, 0xc3d}, {0xc58, 0xc5a}, {0xc5d, 0xc5d},
  {0xc60, 0xc61}, {0xc80, 0xc80}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9}, {0xcbd, 0xcbd},
  {0xcdd, 0xcde}, {0xce0, 0xce1}, {0xcf1, 0xcf2}, {0xd04, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd3a}, {0xd3d, 0xd3d}, {0xd4e, 0xd4e},
  {0xd54, 0xd56}, {0xd5f, 0xd61}, {0xd7a, 0xd7f}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6},
  {0xe01, 0xe30}, {0xe32, 0xe32}, {0xe40, 0xe46}, {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5},
  {0xea7, 0xeb0}, {0xeb2, 0xeb2}, {0xebd, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf40, 0xf47},
  {0xf49, 0xf6c}, {0xf88, 0xf8c}, {0x1000, 0x102a}, {0x103f, 0x103f}, {0x1050, 0x1055}, {0x105a, 0x105d}, {0x1061, 0x1061}, {0x1065, 0x1066},
  {0x106e, 0x1070}, {0x1075, 0x1081}, {0x108e, 0x108e}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248},
  {0x124a, 0x124d}, {0x1250, 0x1256}, {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5},
  {0x12b8, 0x12be}, {0x12c0, 0x12c0}, {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x1380, 0x138f},
  {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1711},
  {0x171f, 0x1731}, {0x1740, 0x1751}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1780, 0x17b3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dc}, {0x1820, 0x1878},
  {0x1880, 0x18a8}, {0x18aa, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1950, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9},
  {0x1a00, 0x1a16}, {0x1a20, 0x1a54}, {0x1aa7, 0x1aa7}, {0x1b05, 0x1b33}, {0x1b45, 0x1b4c}, {0x1b83, 0x1ba0}, {0x1bae, 0x1baf}, {0x1bba, 0x1be5},
  {0x1c00, 0x1c23}, {0x1c4d, 0x1c4f}, {0x1c5a, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1ce9, 0x1cec}, {0x1cee, 0x1cf3},
  {0x1cf5, 0x1cf6}, {0x1cfa, 0x1cfa}, {0x1d00, 0x1dbf}, {0x1e00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57},
  {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4},
  {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x2071, 0x2071}, {0x207f, 0x207f},
  {0x2090, 0x209c}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126},
  {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f}, {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cee},
  {0x2cf2, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d80, 0x2d96}, {0x2da0, 0x2da6},
  {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x3005, 0x3007},
  {0x3021, 0x3029}, {0x3031, 0x3035}, {0x3038, 0x303c}, {0x3041, 0x3096}, {0x309d, 0x309f}, {0x30a1, 0x30fa}, {0x30fc, 0x30ff}, {0x3105, 0x312f},
  {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff}, {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa61f},
  {0xa62a, 0xa62b}, {0xa640, 0xa66e}, {0xa67f, 0xa69d}, {0xa6a0, 0xa6ef}, {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1},
  {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa801}, {0xa803, 0xa805}, {0xa807, 0xa80a}, {0xa80c, 0xa822}, {0xa840, 0xa873}, {0xa882, 0xa8b3},
  {0xa8f2, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa8fe}, {0xa90a, 0xa925}, {0xa930, 0xa946}, {0xa960, 0xa97c}, {0xa984, 0xa9b2}, {0xa9cf, 0xa9cf},
  {0xa9e0, 0xa9e4}, {0xa9e6, 0xa9ef}, {0xa9fa, 0xa9fe}, {0xaa00, 0xaa28}, {0xaa40, 0xaa42}, {0xaa44, 0xaa4b}, {0xaa60, 0xaa76}, {0xaa7a, 0xaa7a},
  {0xaa7e, 0xaaaf}, {0xaab1, 0xaab1}, {0xaab5, 0xaab6}, {0xaab9, 0xaabd}, {0xaac0, 0xaac0}, {0xaac2, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaea},
  {0xaaf2, 0xaaf4}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69},
  {0xab70, 0xabe2}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17},
  {0xfb1d, 0xfb1d}, {0xfb1f, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1},
  {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77},
  {0xfe79, 0xfe79}, {0xfe7b, 0xfe7b}, {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff21, 0xff3a}, {0xff41, 0xff5a}, {0xff66, 0xff9d}, {0xffa0, 0xffbe},
  {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d},
  {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a},
  {0x10350, 0x10375}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb},
  {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1},
  {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0},
  {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876},
  {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a00},
  {0x10a10, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae4}, {0x10b00, 0x10b35},
  {0x10b40, 0x10b55}, {0x10b60, 0x10b72}, {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d23}, {0x10d4a, 0x10d65},
  {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10f00, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f45}, {0x10f70, 0x10f81},
  {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11003, 0x11037}, {0x11071, 0x11072}, {0x11075, 0x11075}, {0x11083, 0x110af}, {0x110d0, 0x110e8}, {0x11103, 0x11126},
  {0x11144, 0x11144}, {0x11147, 0x11147}, {0x11150, 0x11172}, {0x11176, 0x11176}, {0x11183, 0x111b2}, {0x111c1, 0x111c4}, {0x111da, 0x111da}, {0x111dc, 0x111dc},
  {0x11200, 0x11211}, {0x11213, 0x1122b}, {0x1123f, 0x11240}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8},
  {0x112b0, 0x112de}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133d, 0x1133d},
  {0x11350, 0x11350}, {0x1135d, 0x11361}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113b7}, {0x113d1, 0x113d1},
  {0x113d3, 0x113d3}, {0x11400, 0x11434}, {0x11447, 0x1144a}, {0x1145f, 0x11461}, {0x11480, 0x114af}, {0x114c4, 0x114c5}, {0x114c7, 0x114c7}, {0x11580, 0x115ae},
  {0x115d8, 0x115db}, {0x11600, 0x1162f}, {0x11644, 0x11644}, {0x11680, 0x116aa}, {0x116b8, 0x116b8}, {0x11700, 0x1171a}, {0x11740, 0x11746}, {0x11800, 0x1182b},
  {0x118a0, 0x118df}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x1192f}, {0x1193f, 0x1193f}, {0x11941, 0x11941},
  {0x119a0, 0x119a7}, {0x119aa, 0x119d0}, {0x119e1, 0x119e1}, {0x119e3, 0x119e3}, {0x11a00, 0x11a00}, {0x11a0b, 0x11a32}, {0x11a3a, 0x11a3a}, {0x11a50, 0x11a50},
  {0x11a5c, 0x11a89}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c2e}, {0x11c40, 0x11c40}, {0x11c72, 0x11c8f},
  {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d30}, {0x11d46, 0x11d46}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d89}, {0x11d98, 0x11d98},
  {0x11ee0, 0x11ef2}, {0x11f02, 0x11f02}, {0x11f04, 0x11f10}, {0x11f12, 0x11f33}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543},
  {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13441, 0x13446}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x1611d}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e},
  {0x16a70, 0x16abe}, {0x16ad0, 0x16aed}, {0x16b00, 0x16b2f}, {0x16b40, 0x16b43}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16e40, 0x16e7f},
  {0x16f00, 0x16f4a}, {0x16f50, 0x16f50}, {0x16f93, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe3}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08},
  {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167},
  {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f},
  {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a},
  {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5},
  {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788},
  {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e030, 0x1e06d}, {0x1e100, 0x1e12c}, {0x1e137, 0x1e13d},
  {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ad}, {0x1e2c0, 0x1e2eb}, {0x1e4d0, 0x1e4eb}, {0x1e5d0, 0x1e5ed}, {0x1e5f0, 0x1e5f0}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb},
  {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e900, 0x1e943}, {0x1e94b, 0x1e94b}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22},
  {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47},
  {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b},
  {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c},
  {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739},
  {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
};

static const TSCharacterRange sym_identifier_character_set_2[] = {
  {'$', '$'}, {'0', '9'}, {'A', 'Z'}, {'\\', '\\'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5},
  {0xb7, 0xb7}, {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec},
  {0x2ee, 0x2ee}, {0x300, 0x374}, {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1},
  {0x3a3, 0x3f5}, {0x3f7, 0x481}, {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd},
  {0x5bf, 0x5bf}, {0x5c1, 0x5c2}, {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669},
  {0x66e, 0x6d3}, {0x6d5, 0x6dc}, {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5},
  {0x7fa, 0x7fa}, {0x7fd, 0x7fd}, {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1},
  {0x8e3, 0x963}, {0x966, 0x96f}, {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2},
  {0x9b6, 0x9b9}, {0x9bc, 0x9c4}, {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1},
  {0x9fc, 0x9fc}, {0x9fe, 0x9fe}, {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33},
  {0xa35, 0xa36}, {0xa38, 0xa39}, {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c},
  {0xa5e, 0xa5e}, {0xa66, 0xa75}, {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3},
  {0xab5, 0xab9}, {0xabc, 0xac5}, {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff},
  {0xb01, 0xb03}, {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44},
  {0xb47, 0xb48}, {0xb4b, 0xb4d}, {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83},
  {0xb85, 0xb8a}, {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa},
  {0xbae, 0xbb9}, {0xbbe, 0xbc2}, {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c},
  {0xc0e, 0xc10}, {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a},
  {0xc5d, 0xc5d}, {0xc60, 0xc63}, {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3},
  {0xcb5, 0xcb9}, {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef},
  {0xcf1, 0xcf3}, {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63},
  {0xd66, 0xd6f}, {0xd7a, 0xd7f}, {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6},
  {0xdca, 0xdca}, {0xdcf, 0xdd4}, {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e},
  {0xe50, 0xe59}, {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4},
  {0xec6, 0xec6}, {0xec8, 0xece}, {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35},
  {0xf37, 0xf37}, {0xf39, 0xf39}, {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6},
  {0x1000, 0x1049}, {0x1050, 0x109d}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d},
  {0x1250, 0x1256}, {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be},
  {0x12c0, 0x12c0}, {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371},
  {0x1380, 0x138f}, {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8},
  {0x1700, 0x1715}, {0x171f, 0x1734}, {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7},
  {0x17dc, 0x17dd}, {0x17e0, 0x17e9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e},
  {0x1920, 0x192b}, {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b},
  {0x1a20, 0x1a5e}, {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c},
  {0x1b50, 0x1b59}, {0x1b6b, 0x1b73}, {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba},
  {0x1cbd, 0x1cbf}, {0x1cd0, 0x1cd2}, {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57},
  {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4},
  {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040},
  {0x2054, 0x2054}, {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102},
  {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139},
  {0x213c, 0x213f}, {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27},
  {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe},
  {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035},
  {0x3038, 0x303c}, {0x3041, 0x3096}, {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf},
  {0x31f0, 0x31ff}, {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d},
  {0xa67f, 0xa6f1}, {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827},
  {0xa82c, 0xa82c}, {0xa840, 0xa873}, {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953},
  {0xa960, 0xa97c}, {0xa980, 0xa9c0}, {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76},
  {0xaa7a, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26},
  {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6},
  {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c},
  {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7},
  {0xfdf0, 0xfdf9}, {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f}, {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77},
  {0xfe79, 0xfe79}, {0xfe7b, 0xfe7b}, {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f}, {0xff41, 0xff5a},
  {0xff65, 0xffbe}, {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a},
  {0x1003c, 0x1003d}, {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0},
  {0x102e0, 0x102e0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5},
  {0x10400, 0x1049d}, {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a},
  {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736},
  {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835},
  {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915},
  {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35},
  {0x10a38, 0x10a3a}, {0x10a3f, 0x10a3f}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55},
  {0x10b60, 0x10b72}, {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65},
  {0x10d69, 0x10d6d}, {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c}, {0x10f27, 0x10f27},
  {0x10f30, 0x10f50}, {0x10f70, 0x10f85}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2},
  {0x110d0, 0x110e8}, {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4},
  {0x111c9, 0x111cc}, {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288},
  {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310},
  {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350},
  {0x11357, 0x11357}, {0x1135d, 0x11363}, {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5},
  {0x113b7, 0x113c0}, {0x113c2, 0x113c2}, {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459},
  {0x1145e, 0x11461}, {0x11480, 0x114c5}, {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640},
  {0x11644, 0x11644}, {0x11650, 0x11659}, {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x11739},
  {0x11740, 0x11746}, {0x11800, 0x1183a}, {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935},
  {0x11937, 0x11938}, {0x1193b, 0x11943}, {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e},
  {0x11a47, 0x11a47}, {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36},
  {0x11c38, 0x11c40}, {0x11c50, 0x11c59}, {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36},
  {0x11d3a, 0x11d3a}, {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91},
  {0x11d93, 0x11d98}, {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0},
  {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646},
  {0x16100, 0x16139}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4},
  {0x16b00, 0x16b36}, {0x16b40, 0x16b43}, {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f},
  {0x16f00, 0x16f4a}, {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5},
  {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155},
  {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9},
  {0x1cf00, 0x1cf2d}, {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244},
  {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb},
  {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544},
  {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734},
  {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36},
  {0x1da3b, 0x1da6c}, {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006},
  {0x1e008, 0x1e018}, {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d},
  {0x1e140, 0x1e149}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb},
  {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f},
  {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42},
  {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59},
  {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77},
  {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1fbf0, 0x1fbf9},
  {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a},
  {0x31350, 0x323af}, {0xe0100, 0xe01ef},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(420);
      ADVANCE_MAP(
        '!', 509,
        '"', 624,
        '#', 677,
        '%', 529,
        '&', 538,
        '\'', 604,
        '(', 425,
        ')', 428,
        '*', 525,
        '+', 520,
        ',', 427,
        '-', 512,
        '.', 578,
        '/', 527,
        '0', 584,
        ':', 731,
        ';', 553,
        '<', 576,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 610,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 2,
        ']', 556,
        '^', 535,
        'u', 647,
        '{', 557,
        '|', 532,
        '}', 558,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(415);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(585);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 1:
      if (lookahead == '\n') SKIP(93);
      END_STATE();
    case 2:
      if (lookahead == '\n') SKIP(93);
      if (lookahead == '\r') SKIP(1);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 3:
      if (lookahead == '\n') SKIP(101);
      END_STATE();
    case 4:
      if (lookahead == '\n') SKIP(101);
      if (lookahead == '\r') SKIP(3);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 5:
      if (lookahead == '\n') SKIP(99);
      END_STATE();
    case 6:
      if (lookahead == '\n') SKIP(99);
      if (lookahead == '\r') SKIP(5);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 7:
      if (lookahead == '\n') SKIP(108);
      END_STATE();
    case 8:
      if (lookahead == '\n') SKIP(108);
      if (lookahead == '\r') SKIP(7);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 9:
      if (lookahead == '\n') SKIP(107);
      END_STATE();
    case 10:
      if (lookahead == '\n') SKIP(107);
      if (lookahead == '\r') SKIP(9);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 11:
      if (lookahead == '\n') SKIP(102);
      END_STATE();
    case 12:
      if (lookahead == '\n') SKIP(102);
      if (lookahead == '\r') SKIP(11);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 13:
      if (lookahead == '\n') SKIP(96);
      END_STATE();
    case 14:
      if (lookahead == '\n') SKIP(96);
      if (lookahead == '\r') SKIP(13);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 15:
      if (lookahead == '\n') SKIP(111);
      END_STATE();
    case 16:
      if (lookahead == '\n') SKIP(111);
      if (lookahead == '\r') SKIP(15);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 17:
      if (lookahead == '\n') SKIP(115);
      END_STATE();
    case 18:
      if (lookahead == '\n') SKIP(115);
      if (lookahead == '\r') SKIP(17);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 19:
      if (lookahead == '\n') SKIP(110);
      END_STATE();
    case 20:
      if (lookahead == '\n') SKIP(110);
      if (lookahead == '\r') SKIP(19);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 21:
      if (lookahead == '\n') SKIP(114);
      END_STATE();
    case 22:
      if (lookahead == '\n') SKIP(114);
      if (lookahead == '\r') SKIP(21);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 23:
      if (lookahead == '\n') SKIP(132);
      END_STATE();
    case 24:
      if (lookahead == '\n') SKIP(132);
      if (lookahead == '\r') SKIP(23);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 25:
      if (lookahead == '\n') SKIP(127);
      END_STATE();
    case 26:
      if (lookahead == '\n') SKIP(127);
      if (lookahead == '\r') SKIP(25);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 27:
      if (lookahead == '\n') SKIP(133);
      END_STATE();
    case 28:
      if (lookahead == '\n') SKIP(133);
      if (lookahead == '\r') SKIP(27);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 29:
      if (lookahead == '\n') SKIP(120);
      END_STATE();
    case 30:
      if (lookahead == '\n') SKIP(120);
      if (lookahead == '\r') SKIP(29);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 31:
      if (lookahead == '\n') SKIP(124);
      END_STATE();
    case 32:
      if (lookahead == '\n') SKIP(124);
      if (lookahead == '\r') SKIP(31);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 33:
      if (lookahead == '\n') SKIP(112);
      END_STATE();
    case 34:
      if (lookahead == '\n') SKIP(112);
      if (lookahead == '\r') SKIP(33);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 35:
      if (lookahead == '\n') SKIP(117);
      END_STATE();
    case 36:
      if (lookahead == '\n') SKIP(117);
      if (lookahead == '\r') SKIP(35);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 37:
      if (lookahead == '\n') SKIP(123);
      END_STATE();
    case 38:
      if (lookahead == '\n') SKIP(123);
      if (lookahead == '\r') SKIP(37);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 39:
      if (lookahead == '\n') SKIP(134);
      END_STATE();
    case 40:
      if (lookahead == '\n') SKIP(134);
      if (lookahead == '\r') SKIP(39);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 41:
      if (lookahead == '\n') SKIP(130);
      END_STATE();
    case 42:
      if (lookahead == '\n') SKIP(130);
      if (lookahead == '\r') SKIP(41);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 43:
      if (lookahead == '\n') SKIP(122);
      END_STATE();
    case 44:
      if (lookahead == '\n') SKIP(122);
      if (lookahead == '\r') SKIP(43);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 45:
      if (lookahead == '\n') SKIP(121);
      END_STATE();
    case 46:
      if (lookahead == '\n') SKIP(121);
      if (lookahead == '\r') SKIP(45);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 47:
      if (lookahead == '\n') SKIP(116);
      END_STATE();
    case 48:
      if (lookahead == '\n') SKIP(116);
      if (lookahead == '\r') SKIP(47);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 49:
      if (lookahead == '\n') SKIP(51);
      END_STATE();
    case 50:
      if (lookahead == '\n') SKIP(51);
      if (lookahead == '\r') SKIP(49);
      END_STATE();
    case 51:
      ADVANCE_MAP(
        '\n', 431,
        '!', 151,
        '%', 528,
        '&', 537,
        '(', 506,
        '*', 524,
        '+', 518,
        '-', 511,
        '/', 526,
        '<', 548,
        '=', 152,
        '>', 543,
      );
      if (lookahead == '\\') SKIP(50);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '|') ADVANCE(533);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(51);
      END_STATE();
    case 52:
      if (lookahead == '\n') SKIP(129);
      END_STATE();
    case 53:
      if (lookahead == '\n') SKIP(129);
      if (lookahead == '\r') SKIP(52);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 54:
      if (lookahead == '\n') SKIP(131);
      if (lookahead == '"') ADVANCE(624);
      if (lookahead == '/') ADVANCE(629);
      if (lookahead == '\\') ADVANCE(55);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) ADVANCE(632);
      if (lookahead != 0) ADVANCE(633);
      END_STATE();
    case 55:
      if (lookahead == '\n') ADVANCE(635);
      if (lookahead == '\r') ADVANCE(634);
      if (lookahead == 'U') ADVANCE(413);
      if (lookahead == 'u') ADVANCE(405);
      if (lookahead == 'x') ADVANCE(399);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(637);
      if (lookahead != 0) ADVANCE(634);
      END_STATE();
    case 56:
      if (lookahead == '\n') ADVANCE(423);
      if (lookahead == '\r') ADVANCE(60);
      if (lookahead == '(') ADVANCE(425);
      if (lookahead == '/') ADVANCE(459);
      if (lookahead == '\\') ADVANCE(453);
      if (lookahead == 0xa0) ADVANCE(455);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(149);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 57:
      if (lookahead == '\n') ADVANCE(423);
      if (lookahead == '\r') ADVANCE(60);
      if (lookahead == '/') ADVANCE(459);
      if (lookahead == '\\') ADVANCE(453);
      if (lookahead == 0xa0) ADVANCE(455);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(149);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 58:
      if (lookahead == '\n') ADVANCE(423);
      if (lookahead == '\r') ADVANCE(59);
      if (lookahead == '(') ADVANCE(506);
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '/') ADVANCE(139);
      if (lookahead == '0') ADVANCE(586);
      if (lookahead == '\\') SKIP(63);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(143);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(136);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 59:
      if (lookahead == '\n') ADVANCE(423);
      if (lookahead == '(') ADVANCE(506);
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '/') ADVANCE(139);
      if (lookahead == '0') ADVANCE(586);
      if (lookahead == '\\') SKIP(63);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(143);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(136);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 60:
      if (lookahead == '\n') ADVANCE(423);
      if (lookahead == '/') ADVANCE(459);
      if (lookahead == '\\') ADVANCE(453);
      if (lookahead == 0xa0) ADVANCE(455);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(149);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 61:
      if (lookahead == '\n') SKIP(135);
      if (lookahead == '\'') ADVANCE(604);
      if (lookahead == '/') ADVANCE(607);
      if (lookahead == '\\') ADVANCE(606);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) ADVANCE(608);
      if (lookahead != 0) ADVANCE(605);
      END_STATE();
    case 62:
      if (lookahead == '\n') SKIP(136);
      END_STATE();
    case 63:
      if (lookahead == '\n') SKIP(136);
      if (lookahead == '\r') SKIP(62);
      END_STATE();
    case 64:
      if (lookahead == '\n') SKIP(137);
      END_STATE();
    case 65:
      if (lookahead == '\n') SKIP(137);
      if (lookahead == '\r') SKIP(64);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 66:
      if (lookahead == '\n') SKIP(94);
      END_STATE();
    case 67:
      if (lookahead == '\n') SKIP(94);
      if (lookahead == '\r') SKIP(66);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 68:
      if (lookahead == '\n') SKIP(105);
      END_STATE();
    case 69:
      if (lookahead == '\n') SKIP(105);
      if (lookahead == '\r') SKIP(68);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 70:
      if (lookahead == '\n') SKIP(103);
      END_STATE();
    case 71:
      if (lookahead == '\n') SKIP(103);
      if (lookahead == '\r') SKIP(70);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 72:
      if (lookahead == '\n') SKIP(97);
      END_STATE();
    case 73:
      if (lookahead == '\n') SKIP(97);
      if (lookahead == '\r') SKIP(72);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 74:
      if (lookahead == '\n') SKIP(113);
      END_STATE();
    case 75:
      if (lookahead == '\n') SKIP(113);
      if (lookahead == '\r') SKIP(74);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 76:
      if (lookahead == '\n') SKIP(125);
      END_STATE();
    case 77:
      if (lookahead == '\n') SKIP(125);
      if (lookahead == '\r') SKIP(76);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 78:
      if (lookahead == '\n') SKIP(118);
      END_STATE();
    case 79:
      if (lookahead == '\n') SKIP(118);
      if (lookahead == '\r') SKIP(78);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 80:
      if (lookahead == '\n') SKIP(95);
      END_STATE();
    case 81:
      if (lookahead == '\n') SKIP(95);
      if (lookahead == '\r') SKIP(80);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 82:
      if (lookahead == '\n') SKIP(100);
      END_STATE();
    case 83:
      if (lookahead == '\n') SKIP(100);
      if (lookahead == '\r') SKIP(82);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 84:
      if (lookahead == '\n') SKIP(128);
      END_STATE();
    case 85:
      if (lookahead == '\n') SKIP(128);
      if (lookahead == '\r') SKIP(84);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 86:
      if (lookahead == '\n') SKIP(106);
      END_STATE();
    case 87:
      if (lookahead == '\n') SKIP(106);
      if (lookahead == '\r') SKIP(86);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 88:
      if (lookahead == '\n') SKIP(104);
      END_STATE();
    case 89:
      if (lookahead == '\n') SKIP(104);
      if (lookahead == '\r') SKIP(88);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 90:
      if (lookahead == '\n') SKIP(109);
      END_STATE();
    case 91:
      if (lookahead == '\n') SKIP(109);
      if (lookahead == '\r') SKIP(90);
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 92:
      if (lookahead == '\r') ADVANCE(671);
      if (lookahead == '\\') ADVANCE(661);
      if (lookahead != 0) ADVANCE(670);
      END_STATE();
    case 93:
      ADVANCE_MAP(
        '!', 509,
        '"', 624,
        '#', 677,
        '%', 529,
        '&', 538,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 525,
        '+', 520,
        ',', 427,
        '-', 512,
        '.', 578,
        '/', 527,
        '0', 584,
        ':', 562,
        ';', 553,
        '<', 547,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 610,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 2,
        ']', 556,
        '^', 535,
        'u', 647,
        '{', 557,
        '|', 532,
        '}', 558,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(93);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(585);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 94:
      ADVANCE_MAP(
        '!', 509,
        '"', 624,
        '#', 270,
        '%', 529,
        '&', 538,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 525,
        '+', 520,
        ',', 427,
        '-', 512,
        '.', 578,
        '/', 527,
        '0', 586,
        ':', 562,
        ';', 553,
        '<', 547,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 621,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 67,
        ']', 556,
        '^', 535,
        'u', 647,
        '{', 557,
        '|', 532,
        '}', 558,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(94);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 95:
      ADVANCE_MAP(
        '!', 509,
        '"', 624,
        '#', 270,
        '%', 528,
        '&', 537,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 521,
        ',', 427,
        '-', 513,
        '.', 578,
        '/', 526,
        '0', 586,
        ':', 562,
        ';', 553,
        '<', 548,
        '=', 152,
        '>', 543,
        '?', 563,
        '@', 621,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 81,
        ']', 556,
        '^', 534,
        'u', 647,
        '{', 557,
        '|', 533,
        '}', 558,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(95);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 96:
      ADVANCE_MAP(
        '!', 509,
        '"', 624,
        '%', 529,
        '&', 538,
        '\'', 604,
        '(', 506,
        '*', 525,
        '+', 520,
        ',', 427,
        '-', 512,
        '.', 578,
        '/', 527,
        '0', 586,
        ';', 553,
        '<', 547,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 619,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 14,
        '^', 535,
        'u', 647,
        '{', 557,
        '|', 532,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(96);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 97:
      ADVANCE_MAP(
        '!', 509,
        '"', 624,
        '%', 528,
        '&', 537,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        ',', 427,
        '-', 513,
        '.', 578,
        '/', 526,
        '0', 586,
        ';', 553,
        '<', 548,
        '=', 152,
        '>', 543,
        '?', 563,
        '@', 619,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 73,
        '^', 534,
        'u', 647,
        '{', 557,
        '|', 533,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(97);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 98:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 677,
        '&', 536,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ':', 731,
        ';', 553,
        '@', 614,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 6,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(99);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 99:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 677,
        '&', 536,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 614,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 6,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(99);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 100:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 677,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 610,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 83,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(100);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 101:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 680,
        '&', 536,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 521,
        ',', 427,
        '-', 514,
        '.', 147,
        '/', 139,
        '0', 586,
        ':', 562,
        ';', 553,
        '<', 546,
        '>', 541,
        '@', 616,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 4,
        ']', 556,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(101);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 102:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 680,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 615,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 12,
        '^', 534,
        'u', 647,
        '{', 557,
        '}', 558,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(102);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 103:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 680,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 611,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 71,
        '^', 534,
        'u', 647,
        '{', 557,
        '}', 558,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(103);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 104:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 680,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 612,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 89,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(104);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 105:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 239,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 618,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 69,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(105);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 106:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 678,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 610,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 87,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(106);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 107:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 678,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 614,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 10,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(107);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 108:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 679,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        ',', 427,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ':', 150,
        ';', 553,
        '@', 614,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 8,
        ']', 556,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(108);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 109:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 679,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 610,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 91,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(109);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 110:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '&', 536,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 521,
        ',', 427,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ':', 562,
        ';', 553,
        '>', 541,
        '@', 617,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 20,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(110);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 111:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '&', 536,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 521,
        ',', 427,
        '-', 514,
        '.', 579,
        '/', 139,
        '0', 586,
        ':', 562,
        ';', 553,
        '@', 621,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 16,
        ']', 556,
        '^', 534,
        'u', 647,
        '{', 557,
        '}', 558,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(111);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 112:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '&', 536,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        '<', 546,
        '@', 619,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 34,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(112);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 113:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 584,
        '@', 621,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 75,
        '^', 534,
        'u', 647,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(113);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(585);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 114:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 613,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 22,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(114);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 115:
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        '@', 620,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 18,
        '^', 534,
        'u', 647,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(115);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 116:
      ADVANCE_MAP(
        '!', 508,
        '\'', 604,
        '(', 506,
        ')', 428,
        '+', 523,
        '-', 517,
        '.', 391,
        '/', 139,
        '0', 586,
        '<', 153,
        'L', 651,
        'U', 652,
        '\\', 48,
        'u', 653,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(116);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 117:
      ADVANCE_MAP(
        '!', 151,
        '"', 624,
        '#', 270,
        '%', 529,
        '&', 538,
        '(', 506,
        ')', 428,
        '*', 525,
        '+', 522,
        ',', 427,
        '-', 515,
        '.', 577,
        '/', 527,
        ':', 562,
        ';', 553,
        '<', 547,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 609,
        'L', 644,
        'U', 646,
        '[', 555,
        '\\', 36,
        ']', 556,
        '^', 535,
        'u', 648,
        '|', 532,
        '}', 558,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(117);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 118:
      ADVANCE_MAP(
        '!', 151,
        '"', 624,
        '#', 270,
        '%', 528,
        '&', 537,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 519,
        ',', 427,
        '-', 516,
        '.', 577,
        '/', 526,
        ':', 562,
        ';', 553,
        '<', 548,
        '=', 152,
        '>', 543,
        '?', 563,
        '@', 609,
        'L', 644,
        'U', 646,
        '[', 555,
        '\\', 79,
        ']', 556,
        '^', 534,
        'u', 648,
        '|', 533,
        '}', 558,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(118);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 119:
      ADVANCE_MAP(
        '!', 151,
        '"', 624,
        '%', 529,
        '&', 538,
        '(', 506,
        ')', 428,
        '*', 525,
        '+', 522,
        ',', 427,
        '-', 515,
        '.', 577,
        '/', 527,
        ':', 731,
        '<', 576,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 609,
        'L', 644,
        'U', 646,
        '[', 555,
        '\\', 46,
        '^', 535,
        'u', 648,
        '|', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(121);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(675);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 120:
      ADVANCE_MAP(
        '!', 151,
        '"', 624,
        '%', 529,
        '&', 538,
        '(', 506,
        ')', 428,
        '*', 525,
        '+', 522,
        ',', 427,
        '-', 515,
        '.', 577,
        '/', 527,
        ';', 553,
        '<', 547,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 622,
        'L', 644,
        'U', 646,
        '[', 555,
        '\\', 30,
        '^', 535,
        'u', 648,
        '{', 557,
        '|', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(120);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 121:
      ADVANCE_MAP(
        '!', 151,
        '"', 624,
        '%', 529,
        '&', 538,
        '(', 506,
        ')', 428,
        '*', 525,
        '+', 522,
        ',', 427,
        '-', 515,
        '.', 577,
        '/', 527,
        '<', 547,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 609,
        'L', 644,
        'U', 646,
        '[', 555,
        '\\', 46,
        '^', 535,
        'u', 648,
        '|', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(121);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(675);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 122:
      ADVANCE_MAP(
        '!', 151,
        '"', 624,
        '%', 529,
        '&', 538,
        '(', 506,
        '*', 525,
        '+', 522,
        ',', 427,
        '-', 515,
        '.', 577,
        '/', 527,
        ':', 150,
        '<', 547,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 609,
        'L', 644,
        'U', 646,
        '[', 555,
        '\\', 44,
        ']', 556,
        '^', 535,
        'u', 648,
        '|', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(122);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 123:
      ADVANCE_MAP(
        '!', 151,
        '"', 624,
        '%', 528,
        '&', 537,
        '(', 506,
        '*', 524,
        '+', 519,
        ',', 427,
        '-', 516,
        '.', 577,
        '/', 526,
        ';', 553,
        '<', 548,
        '=', 152,
        '>', 543,
        '?', 563,
        '@', 622,
        'L', 644,
        'U', 646,
        '[', 555,
        '\\', 38,
        '^', 534,
        'u', 648,
        '{', 557,
        '|', 533,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(123);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 124:
      ADVANCE_MAP(
        '!', 151,
        '#', 223,
        '%', 529,
        '&', 538,
        '(', 506,
        ')', 428,
        '*', 525,
        '+', 522,
        ',', 427,
        '-', 515,
        '.', 577,
        '/', 527,
        ':', 562,
        ';', 553,
        '<', 547,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 162,
        '[', 555,
        '\\', 32,
        ']', 556,
        '^', 535,
        '{', 557,
        '|', 532,
        '}', 558,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(124);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 125:
      ADVANCE_MAP(
        '!', 151,
        '#', 196,
        '%', 528,
        '&', 537,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 519,
        ',', 427,
        '-', 516,
        '.', 577,
        '/', 526,
        ':', 562,
        ';', 553,
        '<', 548,
        '=', 152,
        '>', 543,
        '?', 563,
        '@', 160,
        '[', 555,
        '\\', 77,
        ']', 556,
        '^', 534,
        '{', 557,
        '|', 533,
        '}', 558,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(125);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 126:
      ADVANCE_MAP(
        '!', 151,
        '#', 200,
        '%', 528,
        '&', 537,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 518,
        ',', 427,
        '-', 511,
        '/', 526,
        ':', 731,
        ';', 553,
        '<', 548,
        '=', 152,
        '>', 543,
        '@', 203,
        '[', 555,
        '\\', 26,
        '^', 534,
        '|', 533,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(127);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 127:
      ADVANCE_MAP(
        '!', 151,
        '#', 200,
        '%', 528,
        '&', 537,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 518,
        ',', 427,
        '-', 511,
        '/', 526,
        ';', 553,
        '<', 548,
        '=', 152,
        '>', 543,
        '@', 203,
        '[', 555,
        '\\', 26,
        '^', 534,
        '|', 533,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(127);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 128:
      ADVANCE_MAP(
        '!', 151,
        '#', 247,
        '%', 528,
        '&', 537,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 519,
        ',', 427,
        '-', 516,
        '.', 577,
        '/', 526,
        ':', 562,
        ';', 553,
        '<', 548,
        '=', 560,
        '>', 543,
        '?', 563,
        '@', 159,
        '[', 555,
        '\\', 85,
        ']', 556,
        '^', 534,
        '{', 557,
        '|', 533,
        '}', 558,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(128);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 129:
      ADVANCE_MAP(
        '"', 624,
        '/', 139,
        '<', 153,
        '@', 609,
        'L', 644,
        'U', 646,
        '\\', 53,
        'u', 648,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(129);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 130:
      ADVANCE_MAP(
        '"', 624,
        '/', 139,
        '@', 623,
        'L', 644,
        'U', 646,
        '[', 555,
        '\\', 42,
        'u', 648,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(130);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 131:
      if (lookahead == '"') ADVANCE(624);
      if (lookahead == '/') ADVANCE(139);
      if (lookahead == '\\') ADVANCE(55);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(131);
      END_STATE();
    case 132:
      ADVANCE_MAP(
        '#', 202,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 518,
        ',', 427,
        '-', 511,
        '.', 146,
        '/', 139,
        ':', 562,
        ';', 553,
        '<', 546,
        '=', 559,
        '>', 541,
        '@', 158,
        '[', 555,
        '\\', 24,
        ']', 556,
        '^', 534,
        '{', 557,
        '}', 558,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(132);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(675);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 133:
      ADVANCE_MAP(
        '#', 198,
        '(', 506,
        '*', 524,
        '+', 518,
        '-', 511,
        '/', 139,
        ';', 553,
        '@', 208,
        '[', 555,
        '\\', 28,
        '^', 534,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(133);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 134:
      ADVANCE_MAP(
        '#', 273,
        '(', 506,
        ')', 428,
        '*', 524,
        ',', 427,
        '.', 391,
        '/', 139,
        '0', 586,
        ':', 562,
        ';', 553,
        '<', 546,
        '=', 559,
        '>', 541,
        '@', 161,
        '[', 555,
        '\\', 40,
        '^', 534,
        '{', 557,
        '}', 558,
        '+', 143,
        '-', 143,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(134);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 135:
      if (lookahead == '\'') ADVANCE(604);
      if (lookahead == '/') ADVANCE(139);
      if (lookahead == '\\') ADVANCE(55);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(135);
      END_STATE();
    case 136:
      if (lookahead == '(') ADVANCE(506);
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '/') ADVANCE(139);
      if (lookahead == '0') ADVANCE(586);
      if (lookahead == '\\') SKIP(63);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(143);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(136);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 137:
      if (lookahead == '(') ADVANCE(192);
      if (lookahead == '/') ADVANCE(139);
      if (lookahead == '\\') ADVANCE(65);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(137);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 138:
      if (lookahead == ')') ADVANCE(694);
      END_STATE();
    case 139:
      if (lookahead == '*') ADVANCE(141);
      if (lookahead == '/') ADVANCE(670);
      END_STATE();
    case 140:
      if (lookahead == '*') ADVANCE(140);
      if (lookahead == '/') ADVANCE(657);
      if (lookahead != 0) ADVANCE(141);
      END_STATE();
    case 141:
      if (lookahead == '*') ADVANCE(140);
      if (lookahead != 0) ADVANCE(141);
      END_STATE();
    case 142:
      if (lookahead == '*') ADVANCE(140);
      if (lookahead != 0) ADVANCE(449);
      END_STATE();
    case 143:
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '0') ADVANCE(586);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 144:
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '0') ADVANCE(582);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(583);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(595);
      END_STATE();
    case 145:
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '0') ADVANCE(587);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 146:
      if (lookahead == '.') ADVANCE(148);
      END_STATE();
    case 147:
      if (lookahead == '.') ADVANCE(148);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(581);
      END_STATE();
    case 148:
      if (lookahead == '.') ADVANCE(426);
      END_STATE();
    case 149:
      if (lookahead == '/') ADVANCE(459);
      if (lookahead == '\\') ADVANCE(453);
      if (lookahead == 0xa0) ADVANCE(455);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(149);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 150:
      if (lookahead == ':') ADVANCE(554);
      END_STATE();
    case 151:
      if (lookahead == '=') ADVANCE(540);
      END_STATE();
    case 152:
      if (lookahead == '=') ADVANCE(539);
      END_STATE();
    case 153:
      if (lookahead == '>') ADVANCE(641);
      if (lookahead == '\\') ADVANCE(154);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(153);
      END_STATE();
    case 154:
      if (lookahead == '>') ADVANCE(642);
      if (lookahead == '\\') ADVANCE(154);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(153);
      END_STATE();
    case 155:
      if (lookahead == 'U') ADVANCE(412);
      if (lookahead == 'u') ADVANCE(404);
      END_STATE();
    case 156:
      if (lookahead == '_') ADVANCE(177);
      END_STATE();
    case 157:
      if (lookahead == 'a') ADVANCE(182);
      END_STATE();
    case 158:
      ADVANCE_MAP(
        'a', 376,
        'c', 165,
        'd', 238,
        'e', 313,
        'f', 282,
        'i', 305,
        'o', 340,
        'p', 351,
        'r', 222,
        's', 387,
      );
      END_STATE();
    case 159:
      if (lookahead == 'a') ADVANCE(376);
      if (lookahead == 'c') ADVANCE(165);
      if (lookahead == 'f') ADVANCE(282);
      END_STATE();
    case 160:
      if (lookahead == 'a') ADVANCE(376);
      if (lookahead == 'd') ADVANCE(237);
      if (lookahead == 'i') ADVANCE(305);
      if (lookahead == 'p') ADVANCE(343);
      END_STATE();
    case 161:
      if (lookahead == 'a') ADVANCE(376);
      if (lookahead == 'i') ADVANCE(305);
      if (lookahead == 'p') ADVANCE(176);
      END_STATE();
    case 162:
      if (lookahead == 'a') ADVANCE(376);
      if (lookahead == 'p') ADVANCE(175);
      END_STATE();
    case 163:
      if (lookahead == 'a') ADVANCE(259);
      END_STATE();
    case 164:
      if (lookahead == 'a') ADVANCE(265);
      END_STATE();
    case 165:
      if (lookahead == 'a') ADVANCE(362);
      END_STATE();
    case 166:
      if (lookahead == 'a') ADVANCE(362);
      if (lookahead == 'o') ADVANCE(303);
      END_STATE();
    case 167:
      if (lookahead == 'a') ADVANCE(304);
      END_STATE();
    case 168:
      if (lookahead == 'a') ADVANCE(359);
      END_STATE();
    case 169:
      if (lookahead == 'a') ADVANCE(293);
      END_STATE();
    case 170:
      if (lookahead == 'a') ADVANCE(354);
      END_STATE();
    case 171:
      if (lookahead == 'a') ADVANCE(366);
      END_STATE();
    case 172:
      if (lookahead == 'a') ADVANCE(191);
      END_STATE();
    case 173:
      if (lookahead == 'a') ADVANCE(358);
      END_STATE();
    case 174:
      if (lookahead == 'a') ADVANCE(369);
      END_STATE();
    case 175:
      if (lookahead == 'a') ADVANCE(185);
      if (lookahead == 'r') ADVANCE(269);
      if (lookahead == 'u') ADVANCE(180);
      END_STATE();
    case 176:
      if (lookahead == 'a') ADVANCE(185);
      if (lookahead == 'r') ADVANCE(268);
      if (lookahead == 'u') ADVANCE(180);
      END_STATE();
    case 177:
      if (lookahead == 'a') ADVANCE(296);
      END_STATE();
    case 178:
      if (lookahead == 'a') ADVANCE(287);
      END_STATE();
    case 179:
      if (lookahead == 'a') ADVANCE(372);
      END_STATE();
    case 180:
      if (lookahead == 'b') ADVANCE(301);
      END_STATE();
    case 181:
      if (lookahead == 'b') ADVANCE(272);
      END_STATE();
    case 182:
      if (lookahead == 'b') ADVANCE(298);
      END_STATE();
    case 183:
      if (lookahead == 'c') ADVANCE(260);
      END_STATE();
    case 184:
      if (lookahead == 'c') ADVANCE(693);
      END_STATE();
    case 185:
      if (lookahead == 'c') ADVANCE(284);
      END_STATE();
    case 186:
      if (lookahead == 'c') ADVANCE(688);
      END_STATE();
    case 187:
      if (lookahead == 'c') ADVANCE(263);
      END_STATE();
    case 188:
      if (lookahead == 'c') ADVANCE(263);
      if (lookahead == 't') ADVANCE(261);
      END_STATE();
    case 189:
      if (lookahead == 'c') ADVANCE(320);
      END_STATE();
    case 190:
      if (lookahead == 'c') ADVANCE(326);
      END_STATE();
    case 191:
      if (lookahead == 'c') ADVANCE(220);
      END_STATE();
    case 192:
      if (lookahead == 'c') ADVANCE(299);
      END_STATE();
    case 193:
      if (lookahead == 'c') ADVANCE(373);
      END_STATE();
    case 194:
      if (lookahead == 'c') ADVANCE(370);
      END_STATE();
    case 195:
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'e') ADVANCE(494);
      if (lookahead == 'i') ADVANCE(480);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(195);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 196:
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'e') ADVANCE(494);
      if (lookahead == 'i') ADVANCE(481);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(196);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 197:
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'e') ADVANCE(497);
      if (lookahead == 'i') ADVANCE(480);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(197);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 198:
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'e') ADVANCE(497);
      if (lookahead == 'i') ADVANCE(481);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(198);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 199:
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'e') ADVANCE(496);
      if (lookahead == 'i') ADVANCE(480);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(199);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 200:
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'e') ADVANCE(496);
      if (lookahead == 'i') ADVANCE(481);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(200);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 201:
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'i') ADVANCE(480);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(201);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 202:
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'i') ADVANCE(481);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(202);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 203:
      if (lookahead == 'd') ADVANCE(386);
      if (lookahead == 'p') ADVANCE(352);
      if (lookahead == 's') ADVANCE(387);
      END_STATE();
    case 204:
      if (lookahead == 'd') ADVANCE(729);
      END_STATE();
    case 205:
      if (lookahead == 'd') ADVANCE(682);
      END_STATE();
    case 206:
      if (lookahead == 'd') ADVANCE(691);
      END_STATE();
    case 207:
      if (lookahead == 'd') ADVANCE(686);
      END_STATE();
    case 208:
      if (lookahead == 'd') ADVANCE(238);
      if (lookahead == 'p') ADVANCE(352);
      if (lookahead == 's') ADVANCE(387);
      END_STATE();
    case 209:
      if (lookahead == 'd') ADVANCE(217);
      END_STATE();
    case 210:
      if (lookahead == 'd') ADVANCE(266);
      END_STATE();
    case 211:
      if (lookahead == 'd') ADVANCE(233);
      if (lookahead == 'n') ADVANCE(212);
      END_STATE();
    case 212:
      if (lookahead == 'd') ADVANCE(235);
      END_STATE();
    case 213:
      if (lookahead == 'd') ADVANCE(244);
      END_STATE();
    case 214:
      if (lookahead == 'e') ADVANCE(289);
      END_STATE();
    case 215:
      if (lookahead == 'e') ADVANCE(289);
      if (lookahead == 'y') ADVANCE(308);
      END_STATE();
    case 216:
      if (lookahead == 'e') ADVANCE(289);
      if (lookahead == 'y') ADVANCE(315);
      END_STATE();
    case 217:
      if (lookahead == 'e') ADVANCE(728);
      END_STATE();
    case 218:
      if (lookahead == 'e') ADVANCE(357);
      END_STATE();
    case 219:
      if (lookahead == 'e') ADVANCE(719);
      END_STATE();
    case 220:
      if (lookahead == 'e') ADVANCE(683);
      END_STATE();
    case 221:
      if (lookahead == 'e') ADVANCE(692);
      END_STATE();
    case 222:
      if (lookahead == 'e') ADVANCE(342);
      END_STATE();
    case 223:
      if (lookahead == 'e') ADVANCE(288);
      if (lookahead == 'i') ADVANCE(252);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(223);
      END_STATE();
    case 224:
      if (lookahead == 'e') ADVANCE(438);
      END_STATE();
    case 225:
      if (lookahead == 'e') ADVANCE(687);
      END_STATE();
    case 226:
      if (lookahead == 'e') ADVANCE(685);
      END_STATE();
    case 227:
      if (lookahead == 'e') ADVANCE(204);
      END_STATE();
    case 228:
      if (lookahead == 'e') ADVANCE(306);
      END_STATE();
    case 229:
      if (lookahead == 'e') ADVANCE(194);
      END_STATE();
    case 230:
      if (lookahead == 'e') ADVANCE(336);
      END_STATE();
    case 231:
      if (lookahead == 'e') ADVANCE(314);
      END_STATE();
    case 232:
      if (lookahead == 'e') ADVANCE(344);
      END_STATE();
    case 233:
      if (lookahead == 'e') ADVANCE(250);
      END_STATE();
    case 234:
      if (lookahead == 'e') ADVANCE(206);
      END_STATE();
    case 235:
      if (lookahead == 'e') ADVANCE(251);
      END_STATE();
    case 236:
      if (lookahead == 'e') ADVANCE(207);
      END_STATE();
    case 237:
      if (lookahead == 'e') ADVANCE(257);
      END_STATE();
    case 238:
      if (lookahead == 'e') ADVANCE(257);
      if (lookahead == 'y') ADVANCE(310);
      END_STATE();
    case 239:
      if (lookahead == 'e') ADVANCE(312);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(239);
      END_STATE();
    case 240:
      if (lookahead == 'e') ADVANCE(168);
      END_STATE();
    case 241:
      if (lookahead == 'e') ADVANCE(193);
      END_STATE();
    case 242:
      if (lookahead == 'e') ADVANCE(193);
      if (lookahead == 'o') ADVANCE(190);
      END_STATE();
    case 243:
      if (lookahead == 'e') ADVANCE(254);
      END_STATE();
    case 244:
      if (lookahead == 'e') ADVANCE(255);
      END_STATE();
    case 245:
      if (lookahead == 'e') ADVANCE(297);
      END_STATE();
    case 246:
      if (lookahead == 'e') ADVANCE(350);
      END_STATE();
    case 247:
      if (lookahead == 'e') ADVANCE(300);
      if (lookahead == 'i') ADVANCE(249);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(247);
      END_STATE();
    case 248:
      if (lookahead == 'f') ADVANCE(432);
      END_STATE();
    case 249:
      if (lookahead == 'f') ADVANCE(211);
      END_STATE();
    case 250:
      if (lookahead == 'f') ADVANCE(434);
      END_STATE();
    case 251:
      if (lookahead == 'f') ADVANCE(436);
      END_STATE();
    case 252:
      if (lookahead == 'f') ADVANCE(429);
      END_STATE();
    case 253:
      if (lookahead == 'f') ADVANCE(441);
      END_STATE();
    case 254:
      if (lookahead == 'f') ADVANCE(444);
      END_STATE();
    case 255:
      if (lookahead == 'f') ADVANCE(446);
      END_STATE();
    case 256:
      if (lookahead == 'f') ADVANCE(440);
      END_STATE();
    case 257:
      if (lookahead == 'f') ADVANCE(355);
      END_STATE();
    case 258:
      if (lookahead == 'f') ADVANCE(172);
      END_STATE();
    case 259:
      if (lookahead == 'g') ADVANCE(225);
      END_STATE();
    case 260:
      if (lookahead == 'h') ADVANCE(697);
      END_STATE();
    case 261:
      if (lookahead == 'h') ADVANCE(218);
      END_STATE();
    case 262:
      if (lookahead == 'h') ADVANCE(346);
      if (lookahead == 'r') ADVANCE(382);
      END_STATE();
    case 263:
      if (lookahead == 'h') ADVANCE(349);
      END_STATE();
    case 264:
      if (lookahead == 'i') ADVANCE(388);
      END_STATE();
    case 265:
      if (lookahead == 'i') ADVANCE(294);
      END_STATE();
    case 266:
      if (lookahead == 'i') ADVANCE(248);
      END_STATE();
    case 267:
      if (lookahead == 'i') ADVANCE(181);
      END_STATE();
    case 268:
      if (lookahead == 'i') ADVANCE(380);
      if (lookahead == 'o') ADVANCE(338);
      END_STATE();
    case 269:
      if (lookahead == 'i') ADVANCE(380);
      if (lookahead == 'o') ADVANCE(374);
      END_STATE();
    case 270:
      if (lookahead == 'i') ADVANCE(249);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(270);
      END_STATE();
    case 271:
      if (lookahead == 'i') ADVANCE(184);
      END_STATE();
    case 272:
      if (lookahead == 'i') ADVANCE(295);
      END_STATE();
    case 273:
      if (lookahead == 'i') ADVANCE(252);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(273);
      END_STATE();
    case 274:
      if (lookahead == 'i') ADVANCE(365);
      END_STATE();
    case 275:
      if (lookahead == 'i') ADVANCE(253);
      if (lookahead == 's') ADVANCE(224);
      END_STATE();
    case 276:
      if (lookahead == 'i') ADVANCE(170);
      END_STATE();
    case 277:
      if (lookahead == 'i') ADVANCE(186);
      END_STATE();
    case 278:
      if (lookahead == 'i') ADVANCE(327);
      END_STATE();
    case 279:
      if (lookahead == 'i') ADVANCE(256);
      if (lookahead == 's') ADVANCE(224);
      END_STATE();
    case 280:
      if (lookahead == 'i') ADVANCE(333);
      END_STATE();
    case 281:
      if (lookahead == 'i') ADVANCE(389);
      END_STATE();
    case 282:
      if (lookahead == 'i') ADVANCE(311);
      END_STATE();
    case 283:
      if (lookahead == 'i') ADVANCE(353);
      END_STATE();
    case 284:
      if (lookahead == 'k') ADVANCE(163);
      END_STATE();
    case 285:
      if (lookahead == 'l') ADVANCE(681);
      END_STATE();
    case 286:
      if (lookahead == 'l') ADVANCE(561);
      END_STATE();
    case 287:
      if (lookahead == 'l') ADVANCE(690);
      END_STATE();
    case 288:
      if (lookahead == 'l') ADVANCE(275);
      if (lookahead == 'n') ADVANCE(210);
      END_STATE();
    case 289:
      if (lookahead == 'l') ADVANCE(229);
      END_STATE();
    case 290:
      if (lookahead == 'l') ADVANCE(383);
      END_STATE();
    case 291:
      if (lookahead == 'l') ADVANCE(228);
      END_STATE();
    case 292:
      if (lookahead == 'l') ADVANCE(228);
      if (lookahead == 'o') ADVANCE(348);
      END_STATE();
    case 293:
      if (lookahead == 'l') ADVANCE(290);
      END_STATE();
    case 294:
      if (lookahead == 'l') ADVANCE(157);
      END_STATE();
    case 295:
      if (lookahead == 'l') ADVANCE(274);
      END_STATE();
    case 296:
      if (lookahead == 'l') ADVANCE(276);
      END_STATE();
    case 297:
      if (lookahead == 'l') ADVANCE(240);
      END_STATE();
    case 298:
      if (lookahead == 'l') ADVANCE(219);
      END_STATE();
    case 299:
      if (lookahead == 'l') ADVANCE(173);
      END_STATE();
    case 300:
      if (lookahead == 'l') ADVANCE(279);
      if (lookahead == 'n') ADVANCE(210);
      END_STATE();
    case 301:
      if (lookahead == 'l') ADVANCE(277);
      END_STATE();
    case 302:
      if (lookahead == 'm') ADVANCE(334);
      if (lookahead == 'n') ADVANCE(360);
      END_STATE();
    case 303:
      if (lookahead == 'm') ADVANCE(335);
      END_STATE();
    case 304:
      if (lookahead == 'm') ADVANCE(271);
      END_STATE();
    case 305:
      if (lookahead == 'm') ADVANCE(341);
      if (lookahead == 'n') ADVANCE(360);
      END_STATE();
    case 306:
      if (lookahead == 'm') ADVANCE(231);
      END_STATE();
    case 307:
      if (lookahead == 'n') ADVANCE(189);
      END_STATE();
    case 308:
      if (lookahead == 'n') ADVANCE(188);
      END_STATE();
    case 309:
      if (lookahead == 'n') ADVANCE(684);
      END_STATE();
    case 310:
      if (lookahead == 'n') ADVANCE(167);
      END_STATE();
    case 311:
      if (lookahead == 'n') ADVANCE(169);
      END_STATE();
    case 312:
      if (lookahead == 'n') ADVANCE(210);
      END_STATE();
    case 313:
      if (lookahead == 'n') ADVANCE(205);
      END_STATE();
    case 314:
      if (lookahead == 'n') ADVANCE(371);
      END_STATE();
    case 315:
      if (lookahead == 'n') ADVANCE(187);
      END_STATE();
    case 316:
      if (lookahead == 'n') ADVANCE(364);
      END_STATE();
    case 317:
      if (lookahead == 'n') ADVANCE(281);
      END_STATE();
    case 318:
      if (lookahead == 'n') ADVANCE(178);
      END_STATE();
    case 319:
      if (lookahead == 'o') ADVANCE(303);
      END_STATE();
    case 320:
      if (lookahead == 'o') ADVANCE(209);
      END_STATE();
    case 321:
      if (lookahead == 'o') ADVANCE(381);
      END_STATE();
    case 322:
      if (lookahead == 'o') ADVANCE(339);
      END_STATE();
    case 323:
      if (lookahead == 'o') ADVANCE(347);
      END_STATE();
    case 324:
      if (lookahead == 'o') ADVANCE(317);
      END_STATE();
    case 325:
      if (lookahead == 'o') ADVANCE(345);
      END_STATE();
    case 326:
      if (lookahead == 'o') ADVANCE(285);
      END_STATE();
    case 327:
      if (lookahead == 'o') ADVANCE(309);
      END_STATE();
    case 328:
      if (lookahead == 'o') ADVANCE(337);
      END_STATE();
    case 329:
      if (lookahead == 'o') ADVANCE(286);
      END_STATE();
    case 330:
      if (lookahead == 'o') ADVANCE(329);
      END_STATE();
    case 331:
      if (lookahead == 'o') ADVANCE(368);
      END_STATE();
    case 332:
      if (lookahead == 'o') ADVANCE(190);
      END_STATE();
    case 333:
      if (lookahead == 'o') ADVANCE(318);
      END_STATE();
    case 334:
      if (lookahead == 'p') ADVANCE(292);
      END_STATE();
    case 335:
      if (lookahead == 'p') ADVANCE(171);
      END_STATE();
    case 336:
      if (lookahead == 'p') ADVANCE(330);
      END_STATE();
    case 337:
      if (lookahead == 'p') ADVANCE(246);
      END_STATE();
    case 338:
      if (lookahead == 'p') ADVANCE(246);
      if (lookahead == 't') ADVANCE(242);
      END_STATE();
    case 339:
      if (lookahead == 'p') ADVANCE(246);
      if (lookahead == 't') ADVANCE(332);
      END_STATE();
    case 340:
      if (lookahead == 'p') ADVANCE(375);
      END_STATE();
    case 341:
      if (lookahead == 'p') ADVANCE(291);
      END_STATE();
    case 342:
      if (lookahead == 'q') ADVANCE(378);
      END_STATE();
    case 343:
      if (lookahead == 'r') ADVANCE(331);
      END_STATE();
    case 344:
      if (lookahead == 'r') ADVANCE(258);
      END_STATE();
    case 345:
      if (lookahead == 'r') ADVANCE(700);
      END_STATE();
    case 346:
      if (lookahead == 'r') ADVANCE(321);
      END_STATE();
    case 347:
      if (lookahead == 'r') ADVANCE(245);
      END_STATE();
    case 348:
      if (lookahead == 'r') ADVANCE(361);
      END_STATE();
    case 349:
      if (lookahead == 'r') ADVANCE(324);
      END_STATE();
    case 350:
      if (lookahead == 'r') ADVANCE(367);
      END_STATE();
    case 351:
      if (lookahead == 'r') ADVANCE(322);
      END_STATE();
    case 352:
      if (lookahead == 'r') ADVANCE(328);
      END_STATE();
    case 353:
      if (lookahead == 'r') ADVANCE(234);
      END_STATE();
    case 354:
      if (lookahead == 's') ADVANCE(689);
      END_STATE();
    case 355:
      if (lookahead == 's') ADVANCE(730);
      END_STATE();
    case 356:
      if (lookahead == 's') ADVANCE(138);
      END_STATE();
    case 357:
      if (lookahead == 's') ADVANCE(264);
      END_STATE();
    case 358:
      if (lookahead == 's') ADVANCE(356);
      END_STATE();
    case 359:
      if (lookahead == 's') ADVANCE(230);
      END_STATE();
    case 360:
      if (lookahead == 't') ADVANCE(232);
      END_STATE();
    case 361:
      if (lookahead == 't') ADVANCE(676);
      END_STATE();
    case 362:
      if (lookahead == 't') ADVANCE(183);
      END_STATE();
    case 363:
      if (lookahead == 't') ADVANCE(323);
      END_STATE();
    case 364:
      if (lookahead == 't') ADVANCE(261);
      END_STATE();
    case 365:
      if (lookahead == 't') ADVANCE(384);
      END_STATE();
    case 366:
      if (lookahead == 't') ADVANCE(267);
      END_STATE();
    case 367:
      if (lookahead == 't') ADVANCE(385);
      END_STATE();
    case 368:
      if (lookahead == 't') ADVANCE(332);
      END_STATE();
    case 369:
      if (lookahead == 't') ADVANCE(278);
      END_STATE();
    case 370:
      if (lookahead == 't') ADVANCE(325);
      END_STATE();
    case 371:
      if (lookahead == 't') ADVANCE(174);
      END_STATE();
    case 372:
      if (lookahead == 't') ADVANCE(226);
      END_STATE();
    case 373:
      if (lookahead == 't') ADVANCE(236);
      END_STATE();
    case 374:
      if (lookahead == 't') ADVANCE(241);
      END_STATE();
    case 375:
      if (lookahead == 't') ADVANCE(280);
      END_STATE();
    case 376:
      if (lookahead == 'u') ADVANCE(363);
      END_STATE();
    case 377:
      if (lookahead == 'u') ADVANCE(363);
      if (lookahead == 'v') ADVANCE(164);
      END_STATE();
    case 378:
      if (lookahead == 'u') ADVANCE(283);
      END_STATE();
    case 379:
      if (lookahead == 'v') ADVANCE(164);
      END_STATE();
    case 380:
      if (lookahead == 'v') ADVANCE(179);
      END_STATE();
    case 381:
      if (lookahead == 'w') ADVANCE(699);
      END_STATE();
    case 382:
      if (lookahead == 'y') ADVANCE(696);
      END_STATE();
    case 383:
      if (lookahead == 'y') ADVANCE(698);
      END_STATE();
    case 384:
      if (lookahead == 'y') ADVANCE(156);
      END_STATE();
    case 385:
      if (lookahead == 'y') ADVANCE(695);
      END_STATE();
    case 386:
      if (lookahead == 'y') ADVANCE(310);
      END_STATE();
    case 387:
      if (lookahead == 'y') ADVANCE(316);
      END_STATE();
    case 388:
      if (lookahead == 'z') ADVANCE(221);
      END_STATE();
    case 389:
      if (lookahead == 'z') ADVANCE(227);
      END_STATE();
    case 390:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 391:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(581);
      END_STATE();
    case 392:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(675);
      END_STATE();
    case 393:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(583);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(595);
      END_STATE();
    case 394:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(655);
      END_STATE();
    case 395:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(592);
      END_STATE();
    case 396:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(595);
      END_STATE();
    case 397:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(589);
      END_STATE();
    case 398:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(634);
      END_STATE();
    case 399:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(640);
      END_STATE();
    case 400:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(394);
      END_STATE();
    case 401:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(398);
      END_STATE();
    case 402:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(400);
      END_STATE();
    case 403:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(401);
      END_STATE();
    case 404:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(402);
      END_STATE();
    case 405:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(403);
      END_STATE();
    case 406:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(404);
      END_STATE();
    case 407:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(405);
      END_STATE();
    case 408:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(406);
      END_STATE();
    case 409:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(407);
      END_STATE();
    case 410:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(408);
      END_STATE();
    case 411:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(409);
      END_STATE();
    case 412:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(410);
      END_STATE();
    case 413:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(411);
      END_STATE();
    case 414:
      if (lookahead != 0 &&
          lookahead != '*') ADVANCE(462);
      END_STATE();
    case 415:
      if (eof) ADVANCE(420);
      ADVANCE_MAP(
        '!', 509,
        '"', 624,
        '#', 677,
        '%', 529,
        '&', 538,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 525,
        '+', 520,
        ',', 427,
        '-', 512,
        '.', 578,
        '/', 527,
        '0', 584,
        ':', 562,
        ';', 553,
        '<', 547,
        '=', 560,
        '>', 542,
        '?', 563,
        '@', 610,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 2,
        ']', 556,
        '^', 535,
        'u', 647,
        '{', 557,
        '|', 532,
        '}', 558,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(415);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(585);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 416:
      if (eof) ADVANCE(420);
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 680,
        '&', 536,
        '\'', 604,
        '(', 506,
        ')', 428,
        '*', 524,
        '+', 521,
        ',', 427,
        '-', 514,
        '.', 147,
        '/', 139,
        '0', 586,
        ':', 562,
        ';', 553,
        '<', 546,
        '>', 541,
        '@', 616,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 4,
        ']', 556,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(416);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 417:
      if (eof) ADVANCE(420);
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 680,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 615,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 12,
        '^', 534,
        'u', 647,
        '{', 557,
        '}', 558,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(417);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 418:
      if (eof) ADVANCE(420);
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 680,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 611,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 71,
        '^', 534,
        'u', 647,
        '{', 557,
        '}', 558,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(418);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 419:
      if (eof) ADVANCE(420);
      ADVANCE_MAP(
        '!', 508,
        '"', 624,
        '#', 680,
        '&', 536,
        '\'', 604,
        '(', 506,
        '*', 524,
        '+', 521,
        '-', 514,
        '.', 391,
        '/', 139,
        '0', 586,
        ';', 553,
        '@', 612,
        'L', 643,
        'U', 645,
        '[', 555,
        '\\', 89,
        '^', 534,
        'u', 647,
        '{', 557,
        '~', 510,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(419);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(655);
      END_STATE();
    case 420:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 421:
      ACCEPT_TOKEN(aux_sym_preproc_include_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 422:
      ACCEPT_TOKEN(aux_sym_preproc_include_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 423:
      ACCEPT_TOKEN(aux_sym_preproc_include_token3);
      END_STATE();
    case 424:
      ACCEPT_TOKEN(aux_sym_preproc_def_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 425:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 426:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_DOT);
      END_STATE();
    case 427:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 428:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 429:
      ACCEPT_TOKEN(aux_sym_preproc_if_token1);
      if (lookahead == 'd') ADVANCE(233);
      if (lookahead == 'n') ADVANCE(212);
      END_STATE();
    case 430:
      ACCEPT_TOKEN(aux_sym_preproc_if_token1);
      if (lookahead == 'd') ADVANCE(474);
      if (lookahead == 'n') ADVANCE(468);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 431:
      ACCEPT_TOKEN(anon_sym_LF);
      if (lookahead == '\n') ADVANCE(431);
      END_STATE();
    case 432:
      ACCEPT_TOKEN(aux_sym_preproc_if_token2);
      END_STATE();
    case 433:
      ACCEPT_TOKEN(aux_sym_preproc_if_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 434:
      ACCEPT_TOKEN(aux_sym_preproc_ifdef_token1);
      END_STATE();
    case 435:
      ACCEPT_TOKEN(aux_sym_preproc_ifdef_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 436:
      ACCEPT_TOKEN(aux_sym_preproc_ifdef_token2);
      END_STATE();
    case 437:
      ACCEPT_TOKEN(aux_sym_preproc_ifdef_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 438:
      ACCEPT_TOKEN(aux_sym_preproc_else_token1);
      END_STATE();
    case 439:
      ACCEPT_TOKEN(aux_sym_preproc_else_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 440:
      ACCEPT_TOKEN(aux_sym_preproc_elif_token1);
      END_STATE();
    case 441:
      ACCEPT_TOKEN(aux_sym_preproc_elif_token1);
      if (lookahead == 'd') ADVANCE(243);
      if (lookahead == 'n') ADVANCE(213);
      END_STATE();
    case 442:
      ACCEPT_TOKEN(aux_sym_preproc_elif_token1);
      if (lookahead == 'd') ADVANCE(477);
      if (lookahead == 'n') ADVANCE(469);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 443:
      ACCEPT_TOKEN(aux_sym_preproc_elif_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 444:
      ACCEPT_TOKEN(aux_sym_preproc_elifdef_token1);
      END_STATE();
    case 445:
      ACCEPT_TOKEN(aux_sym_preproc_elifdef_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 446:
      ACCEPT_TOKEN(aux_sym_preproc_elifdef_token2);
      END_STATE();
    case 447:
      ACCEPT_TOKEN(aux_sym_preproc_elifdef_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 448:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') ADVANCE(141);
      if (lookahead == '*') ADVANCE(448);
      if (lookahead == '/') ADVANCE(657);
      if (lookahead == '\\') ADVANCE(458);
      if (lookahead != 0) ADVANCE(449);
      END_STATE();
    case 449:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') ADVANCE(141);
      if (lookahead == '*') ADVANCE(448);
      if (lookahead == '/') ADVANCE(142);
      if (lookahead == '\\') ADVANCE(458);
      if (lookahead != 0) ADVANCE(449);
      END_STATE();
    case 450:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') ADVANCE(670);
      if (lookahead == '\r') ADVANCE(658);
      if (lookahead == '/') ADVANCE(666);
      if (lookahead == '\\') ADVANCE(663);
      if (lookahead != 0) ADVANCE(667);
      END_STATE();
    case 451:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') ADVANCE(455);
      if (lookahead == '\r') ADVANCE(452);
      if (lookahead == '/') ADVANCE(414);
      if (lookahead == '\\') ADVANCE(456);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 452:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') ADVANCE(455);
      if (lookahead == '/') ADVANCE(414);
      if (lookahead == '\\') ADVANCE(456);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 453:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') SKIP(149);
      if (lookahead == '\r') ADVANCE(454);
      if (lookahead == '/') ADVANCE(414);
      if (lookahead == '\\') ADVANCE(456);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 454:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') SKIP(149);
      if (lookahead == '/') ADVANCE(414);
      if (lookahead == '\\') ADVANCE(456);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 455:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') SKIP(149);
      if (lookahead == '/') ADVANCE(460);
      if (lookahead == '\\') ADVANCE(451);
      if (lookahead == 0xa0) ADVANCE(455);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(455);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 456:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\r') ADVANCE(463);
      if (lookahead == '/') ADVANCE(414);
      if (lookahead == '\\') ADVANCE(456);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 457:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\r') ADVANCE(668);
      if (lookahead == '/') ADVANCE(666);
      if (lookahead == '\\') ADVANCE(663);
      if (lookahead != 0) ADVANCE(667);
      END_STATE();
    case 458:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\r') ADVANCE(461);
      if (lookahead == '*') ADVANCE(448);
      if (lookahead == '/') ADVANCE(142);
      if (lookahead == '\\') ADVANCE(458);
      if (lookahead != 0) ADVANCE(449);
      END_STATE();
    case 459:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '*') ADVANCE(449);
      if (lookahead == '/') ADVANCE(666);
      if (lookahead == '\\') ADVANCE(456);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(462);
      END_STATE();
    case 460:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '*') ADVANCE(449);
      if (lookahead == '/') ADVANCE(669);
      if (lookahead == '\\') ADVANCE(456);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 461:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '*') ADVANCE(448);
      if (lookahead == '/') ADVANCE(142);
      if (lookahead == '\\') ADVANCE(458);
      if (lookahead != 0) ADVANCE(449);
      END_STATE();
    case 462:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '/') ADVANCE(414);
      if (lookahead == '\\') ADVANCE(456);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(462);
      END_STATE();
    case 463:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '/') ADVANCE(414);
      if (lookahead == '\\') ADVANCE(456);
      if (lookahead != 0) ADVANCE(462);
      END_STATE();
    case 464:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'c') ADVANCE(495);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 465:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'd') ADVANCE(492);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 466:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'd') ADVANCE(473);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 467:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'd') ADVANCE(475);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 468:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'd') ADVANCE(476);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 469:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'd') ADVANCE(478);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 470:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(482);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 471:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(439);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 472:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(424);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 473:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(421);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 474:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(485);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 475:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(479);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 476:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(486);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 477:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(487);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 478:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(488);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 479:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(505);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 480:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(430);
      if (lookahead == 'm') ADVANCE(501);
      if (lookahead == 'n') ADVANCE(464);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 481:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(430);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 482:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(490);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 483:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(442);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 484:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(433);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 485:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(435);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 486:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(437);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 487:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(445);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 488:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(447);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 489:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(443);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 490:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'i') ADVANCE(499);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 491:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'i') ADVANCE(483);
      if (lookahead == 's') ADVANCE(471);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 492:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'i') ADVANCE(484);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 493:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'i') ADVANCE(489);
      if (lookahead == 's') ADVANCE(471);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 494:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'l') ADVANCE(491);
      if (lookahead == 'n') ADVANCE(465);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 495:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'l') ADVANCE(504);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 496:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'l') ADVANCE(493);
      if (lookahead == 'n') ADVANCE(465);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 497:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'n') ADVANCE(465);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 498:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'n') ADVANCE(467);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 499:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'n') ADVANCE(472);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 500:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'o') ADVANCE(502);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 501:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'p') ADVANCE(500);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 502:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'r') ADVANCE(503);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 503:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 't') ADVANCE(422);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 504:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'u') ADVANCE(466);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 505:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 506:
      ACCEPT_TOKEN(anon_sym_LPAREN2);
      END_STATE();
    case 507:
      ACCEPT_TOKEN(anon_sym_LPAREN2);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 508:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 509:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(540);
      END_STATE();
    case 510:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 511:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 512:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(574);
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '0') ADVANCE(586);
      if (lookahead == '=') ADVANCE(568);
      if (lookahead == '>') ADVANCE(580);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 513:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(574);
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '0') ADVANCE(586);
      if (lookahead == '>') ADVANCE(580);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 514:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(574);
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '0') ADVANCE(586);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 515:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(574);
      if (lookahead == '=') ADVANCE(568);
      if (lookahead == '>') ADVANCE(580);
      END_STATE();
    case 516:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(574);
      if (lookahead == '>') ADVANCE(580);
      END_STATE();
    case 517:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '0') ADVANCE(586);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 518:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 519:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(575);
      END_STATE();
    case 520:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(575);
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '0') ADVANCE(586);
      if (lookahead == '=') ADVANCE(567);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 521:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(575);
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '0') ADVANCE(586);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 522:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(575);
      if (lookahead == '=') ADVANCE(567);
      END_STATE();
    case 523:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '.') ADVANCE(391);
      if (lookahead == '0') ADVANCE(586);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 524:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 525:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '=') ADVANCE(564);
      END_STATE();
    case 526:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(141);
      if (lookahead == '/') ADVANCE(670);
      END_STATE();
    case 527:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(141);
      if (lookahead == '/') ADVANCE(670);
      if (lookahead == '=') ADVANCE(565);
      END_STATE();
    case 528:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 529:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(566);
      END_STATE();
    case 530:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 531:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 532:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(573);
      if (lookahead == '|') ADVANCE(530);
      END_STATE();
    case 533:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(530);
      END_STATE();
    case 534:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 535:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(572);
      END_STATE();
    case 536:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 537:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(531);
      END_STATE();
    case 538:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(531);
      if (lookahead == '=') ADVANCE(571);
      END_STATE();
    case 539:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      END_STATE();
    case 540:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 541:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 542:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(544);
      if (lookahead == '>') ADVANCE(552);
      END_STATE();
    case 543:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(544);
      if (lookahead == '>') ADVANCE(551);
      END_STATE();
    case 544:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 545:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 546:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 547:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(550);
      if (lookahead == '=') ADVANCE(545);
      END_STATE();
    case 548:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(549);
      if (lookahead == '=') ADVANCE(545);
      END_STATE();
    case 549:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 550:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(569);
      END_STATE();
    case 551:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 552:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(570);
      END_STATE();
    case 553:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 554:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 555:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 556:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 557:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 558:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 559:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 560:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(539);
      END_STATE();
    case 561:
      ACCEPT_TOKEN(anon_sym_ATautoreleasepool);
      END_STATE();
    case 562:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 563:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 564:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 565:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 566:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 567:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 568:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 569:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 570:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 571:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 572:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 573:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 574:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 575:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 576:
      ACCEPT_TOKEN(anon_sym_LT2);
      if (lookahead == '<') ADVANCE(550);
      if (lookahead == '=') ADVANCE(545);
      END_STATE();
    case 577:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(148);
      END_STATE();
    case 578:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(148);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(581);
      END_STATE();
    case 579:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(581);
      END_STATE();
    case 580:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 581:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 391,
        'E', 596,
        'P', 596,
        'e', 596,
        'p', 596,
        'F', 599,
        'L', 599,
        'U', 599,
        'f', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(581);
      END_STATE();
    case 582:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 393,
        '.', 598,
        'b', 594,
        'x', 396,
        'E', 593,
        'e', 593,
        'F', 595,
        'f', 595,
        'P', 596,
        'p', 596,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('A' <= lookahead && lookahead <= 'D') ||
          ('a' <= lookahead && lookahead <= 'd')) ADVANCE(595);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(583);
      END_STATE();
    case 583:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 393,
        '.', 598,
        'E', 593,
        'e', 593,
        'F', 595,
        'f', 595,
        'P', 596,
        'p', 596,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('A' <= lookahead && lookahead <= 'D') ||
          ('a' <= lookahead && lookahead <= 'd')) ADVANCE(595);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(583);
      END_STATE();
    case 584:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 390,
        '.', 597,
        '_', 392,
        'b', 145,
        'x', 144,
        'E', 596,
        'P', 596,
        'e', 596,
        'p', 596,
        'F', 599,
        'L', 599,
        'U', 599,
        'f', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(585);
      END_STATE();
    case 585:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 390,
        '.', 597,
        '_', 392,
        'E', 596,
        'P', 596,
        'e', 596,
        'p', 596,
        'F', 599,
        'L', 599,
        'U', 599,
        'f', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(585);
      END_STATE();
    case 586:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 390,
        '.', 598,
        'b', 145,
        'x', 144,
        'E', 596,
        'P', 596,
        'e', 596,
        'p', 596,
        'F', 599,
        'L', 599,
        'U', 599,
        'f', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 587:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 390,
        '.', 598,
        'b', 390,
        'x', 396,
        'E', 596,
        'P', 596,
        'e', 596,
        'p', 596,
        'F', 599,
        'L', 599,
        'U', 599,
        'f', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 588:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 390,
        '.', 598,
        'E', 596,
        'P', 596,
        'e', 596,
        'p', 596,
        'F', 599,
        'L', 599,
        'U', 599,
        'f', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(588);
      END_STATE();
    case 589:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 397,
        'F', 589,
        'f', 589,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'E') ||
          ('a' <= lookahead && lookahead <= 'e')) ADVANCE(589);
      END_STATE();
    case 590:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 395,
        '+', 397,
        '-', 397,
        'E', 590,
        'e', 590,
        'F', 592,
        'f', 592,
        'P', 596,
        'p', 596,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'D') ||
          ('a' <= lookahead && lookahead <= 'd')) ADVANCE(592);
      END_STATE();
    case 591:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 395,
        '.', 392,
        '_', 392,
        'E', 590,
        'e', 590,
        'F', 592,
        'f', 592,
        'P', 596,
        'p', 596,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('A' <= lookahead && lookahead <= 'D') ||
          ('a' <= lookahead && lookahead <= 'd')) ADVANCE(592);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(591);
      END_STATE();
    case 592:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 395,
        'E', 590,
        'e', 590,
        'F', 592,
        'f', 592,
        'P', 596,
        'p', 596,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'D') ||
          ('a' <= lookahead && lookahead <= 'd')) ADVANCE(592);
      END_STATE();
    case 593:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 396,
        '.', 598,
        '+', 397,
        '-', 397,
        'E', 593,
        'e', 593,
        'F', 595,
        'f', 595,
        'P', 596,
        'p', 596,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'D') ||
          ('a' <= lookahead && lookahead <= 'd')) ADVANCE(595);
      END_STATE();
    case 594:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 396,
        '.', 598,
        'E', 593,
        'e', 593,
        'F', 595,
        'f', 595,
        'P', 596,
        'p', 596,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('A' <= lookahead && lookahead <= 'D') ||
          ('a' <= lookahead && lookahead <= 'd')) ADVANCE(595);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(583);
      END_STATE();
    case 595:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 396,
        '.', 598,
        'E', 593,
        'e', 593,
        'F', 595,
        'f', 595,
        'P', 596,
        'p', 596,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'D') ||
          ('a' <= lookahead && lookahead <= 'd')) ADVANCE(595);
      END_STATE();
    case 596:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '+', 397,
        '-', 397,
        'F', 589,
        'f', 589,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'E') ||
          ('a' <= lookahead && lookahead <= 'e')) ADVANCE(589);
      END_STATE();
    case 597:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        'E', 590,
        'e', 590,
        'F', 592,
        'f', 592,
        'P', 596,
        'p', 596,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('A' <= lookahead && lookahead <= 'D') ||
          ('a' <= lookahead && lookahead <= 'd')) ADVANCE(592);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(591);
      END_STATE();
    case 598:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        'E', 590,
        'e', 590,
        'F', 592,
        'f', 592,
        'P', 596,
        'p', 596,
        'L', 599,
        'U', 599,
        'i', 599,
        'l', 599,
        'u', 599,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'D') ||
          ('a' <= lookahead && lookahead <= 'd')) ADVANCE(592);
      END_STATE();
    case 599:
      ACCEPT_TOKEN(sym_number_literal);
      if (lookahead == 'F' ||
          lookahead == 'L' ||
          lookahead == 'U' ||
          lookahead == 'f' ||
          lookahead == 'i' ||
          lookahead == 'l' ||
          lookahead == 'u') ADVANCE(599);
      END_STATE();
    case 600:
      ACCEPT_TOKEN(anon_sym_L_SQUOTE);
      END_STATE();
    case 601:
      ACCEPT_TOKEN(anon_sym_u_SQUOTE);
      END_STATE();
    case 602:
      ACCEPT_TOKEN(anon_sym_U_SQUOTE);
      END_STATE();
    case 603:
      ACCEPT_TOKEN(anon_sym_u8_SQUOTE);
      END_STATE();
    case 604:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 605:
      ACCEPT_TOKEN(aux_sym_char_literal_token1);
      END_STATE();
    case 606:
      ACCEPT_TOKEN(aux_sym_char_literal_token1);
      if (lookahead == '\n') ADVANCE(635);
      if (lookahead == '\r') ADVANCE(634);
      if (lookahead == 'U') ADVANCE(413);
      if (lookahead == 'u') ADVANCE(405);
      if (lookahead == 'x') ADVANCE(399);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(637);
      if (lookahead != 0) ADVANCE(634);
      END_STATE();
    case 607:
      ACCEPT_TOKEN(aux_sym_char_literal_token1);
      if (lookahead == '*') ADVANCE(141);
      if (lookahead == '/') ADVANCE(670);
      END_STATE();
    case 608:
      ACCEPT_TOKEN(aux_sym_char_literal_token1);
      if (lookahead == '\\') ADVANCE(55);
      END_STATE();
    case 609:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 610:
      ACCEPT_TOKEN(anon_sym_AT);
      ADVANCE_MAP(
        'a', 377,
        'c', 166,
        'd', 386,
        'e', 307,
        'f', 282,
        'i', 302,
        'p', 343,
        's', 215,
        't', 262,
      );
      END_STATE();
    case 611:
      ACCEPT_TOKEN(anon_sym_AT);
      ADVANCE_MAP(
        'a', 377,
        'c', 166,
        'e', 307,
        'f', 282,
        'i', 302,
        'p', 343,
        's', 216,
        't', 262,
      );
      END_STATE();
    case 612:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'a') ADVANCE(377);
      if (lookahead == 'c') ADVANCE(166);
      if (lookahead == 'e') ADVANCE(307);
      if (lookahead == 'f') ADVANCE(282);
      if (lookahead == 'i') ADVANCE(302);
      if (lookahead == 'p') ADVANCE(343);
      if (lookahead == 's') ADVANCE(214);
      END_STATE();
    case 613:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'a') ADVANCE(377);
      if (lookahead == 'c') ADVANCE(165);
      if (lookahead == 'e') ADVANCE(307);
      if (lookahead == 'f') ADVANCE(282);
      if (lookahead == 's') ADVANCE(216);
      if (lookahead == 't') ADVANCE(262);
      END_STATE();
    case 614:
      ACCEPT_TOKEN(anon_sym_AT);
      ADVANCE_MAP(
        'a', 377,
        'c', 319,
        'd', 386,
        'e', 307,
        'i', 302,
        'p', 343,
        's', 215,
        't', 262,
      );
      END_STATE();
    case 615:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'a') ADVANCE(377);
      if (lookahead == 'c') ADVANCE(319);
      if (lookahead == 'e') ADVANCE(307);
      if (lookahead == 'i') ADVANCE(302);
      if (lookahead == 'p') ADVANCE(343);
      if (lookahead == 's') ADVANCE(216);
      if (lookahead == 't') ADVANCE(262);
      END_STATE();
    case 616:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'a') ADVANCE(377);
      if (lookahead == 'c') ADVANCE(319);
      if (lookahead == 'e') ADVANCE(307);
      if (lookahead == 'i') ADVANCE(302);
      if (lookahead == 'p') ADVANCE(343);
      if (lookahead == 's') ADVANCE(214);
      END_STATE();
    case 617:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'a') ADVANCE(377);
      if (lookahead == 'e') ADVANCE(307);
      if (lookahead == 'i') ADVANCE(305);
      if (lookahead == 'p') ADVANCE(343);
      if (lookahead == 's') ADVANCE(216);
      if (lookahead == 't') ADVANCE(262);
      END_STATE();
    case 618:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'a') ADVANCE(377);
      if (lookahead == 'e') ADVANCE(307);
      if (lookahead == 's') ADVANCE(216);
      if (lookahead == 't') ADVANCE(262);
      END_STATE();
    case 619:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'a') ADVANCE(377);
      if (lookahead == 'e') ADVANCE(307);
      if (lookahead == 's') ADVANCE(214);
      END_STATE();
    case 620:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'a') ADVANCE(379);
      if (lookahead == 'e') ADVANCE(307);
      if (lookahead == 'i') ADVANCE(305);
      if (lookahead == 'p') ADVANCE(343);
      if (lookahead == 's') ADVANCE(214);
      END_STATE();
    case 621:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'a') ADVANCE(379);
      if (lookahead == 'e') ADVANCE(307);
      if (lookahead == 's') ADVANCE(214);
      END_STATE();
    case 622:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'a') ADVANCE(376);
      END_STATE();
    case 623:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'i') ADVANCE(305);
      if (lookahead == 'p') ADVANCE(343);
      END_STATE();
    case 624:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 625:
      ACCEPT_TOKEN(anon_sym_L_DQUOTE);
      END_STATE();
    case 626:
      ACCEPT_TOKEN(anon_sym_u_DQUOTE);
      END_STATE();
    case 627:
      ACCEPT_TOKEN(anon_sym_U_DQUOTE);
      END_STATE();
    case 628:
      ACCEPT_TOKEN(anon_sym_u8_DQUOTE);
      END_STATE();
    case 629:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      if (lookahead == '*') ADVANCE(631);
      if (lookahead == '/') ADVANCE(633);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(633);
      END_STATE();
    case 630:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      if (lookahead == '*') ADVANCE(630);
      if (lookahead == '/') ADVANCE(633);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(631);
      END_STATE();
    case 631:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      if (lookahead == '*') ADVANCE(630);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(631);
      END_STATE();
    case 632:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      if (lookahead == '/') ADVANCE(629);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) ADVANCE(632);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(633);
      END_STATE();
    case 633:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(633);
      END_STATE();
    case 634:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 635:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (lookahead == '\\') ADVANCE(55);
      END_STATE();
    case 636:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(634);
      END_STATE();
    case 637:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(636);
      END_STATE();
    case 638:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(634);
      END_STATE();
    case 639:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(638);
      END_STATE();
    case 640:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(639);
      END_STATE();
    case 641:
      ACCEPT_TOKEN(sym_system_lib_string);
      END_STATE();
    case 642:
      ACCEPT_TOKEN(sym_system_lib_string);
      if (lookahead == '>') ADVANCE(641);
      if (lookahead == '\\') ADVANCE(154);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(153);
      END_STATE();
    case 643:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(625);
      if (lookahead == '\'') ADVANCE(600);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 644:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(625);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 645:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(627);
      if (lookahead == '\'') ADVANCE(602);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 646:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(627);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 647:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(626);
      if (lookahead == '\'') ADVANCE(601);
      if (lookahead == '8') ADVANCE(649);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 648:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(626);
      if (lookahead == '8') ADVANCE(650);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 649:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(628);
      if (lookahead == '\'') ADVANCE(603);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 650:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(628);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 651:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\'') ADVANCE(600);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 652:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\'') ADVANCE(602);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 653:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\'') ADVANCE(601);
      if (lookahead == '8') ADVANCE(654);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 654:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\'') ADVANCE(603);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 655:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(155);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(655);
      END_STATE();
    case 656:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(709);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(656);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 657:
      ACCEPT_TOKEN(sym_comment);
      END_STATE();
    case 658:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\n') ADVANCE(670);
      if (lookahead == '/') ADVANCE(666);
      if (lookahead == '\\') ADVANCE(457);
      if (lookahead != 0) ADVANCE(667);
      END_STATE();
    case 659:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\n') ADVANCE(718);
      if (lookahead == ')') ADVANCE(670);
      if (lookahead == '\\') ADVANCE(703);
      if (lookahead != 0) ADVANCE(659);
      END_STATE();
    case 660:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\n') ADVANCE(727);
      if (lookahead == '\\') ADVANCE(722);
      if (lookahead == '}') ADVANCE(670);
      if (lookahead != 0) ADVANCE(660);
      END_STATE();
    case 661:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\r') ADVANCE(671);
      if (lookahead == '\\') ADVANCE(661);
      if (lookahead != 0) ADVANCE(670);
      END_STATE();
    case 662:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\r') ADVANCE(665);
      if (lookahead == ')') ADVANCE(670);
      if (lookahead == '\\') ADVANCE(662);
      if (lookahead != 0) ADVANCE(659);
      END_STATE();
    case 663:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\r') ADVANCE(668);
      if (lookahead == '/') ADVANCE(666);
      if (lookahead == '\\') ADVANCE(663);
      if (lookahead != 0) ADVANCE(667);
      END_STATE();
    case 664:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\r') ADVANCE(672);
      if (lookahead == '\\') ADVANCE(664);
      if (lookahead == '}') ADVANCE(670);
      if (lookahead != 0) ADVANCE(660);
      END_STATE();
    case 665:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == ')') ADVANCE(670);
      if (lookahead == '\\') ADVANCE(703);
      if (lookahead != 0) ADVANCE(659);
      END_STATE();
    case 666:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '*') ADVANCE(670);
      if (lookahead == '\\') ADVANCE(450);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(667);
      END_STATE();
    case 667:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '/') ADVANCE(666);
      if (lookahead == '\\') ADVANCE(457);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(667);
      END_STATE();
    case 668:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '/') ADVANCE(666);
      if (lookahead == '\\') ADVANCE(457);
      if (lookahead != 0) ADVANCE(667);
      END_STATE();
    case 669:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '/') ADVANCE(669);
      if (lookahead == '\\') ADVANCE(457);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(667);
      END_STATE();
    case 670:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\\') ADVANCE(92);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(670);
      END_STATE();
    case 671:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\\') ADVANCE(92);
      if (lookahead != 0) ADVANCE(670);
      END_STATE();
    case 672:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\\') ADVANCE(722);
      if (lookahead == '}') ADVANCE(670);
      if (lookahead != 0) ADVANCE(660);
      END_STATE();
    case 673:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 674:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '}') ADVANCE(727);
      END_STATE();
    case 675:
      ACCEPT_TOKEN(sym_version_number);
      if (lookahead == '.' ||
          lookahead == '_') ADVANCE(392);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(675);
      END_STATE();
    case 676:
      ACCEPT_TOKEN(anon_sym_ATimport);
      END_STATE();
    case 677:
      ACCEPT_TOKEN(anon_sym_POUND);
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'e') ADVANCE(494);
      if (lookahead == 'i') ADVANCE(480);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(195);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 678:
      ACCEPT_TOKEN(anon_sym_POUND);
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'e') ADVANCE(497);
      if (lookahead == 'i') ADVANCE(480);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(197);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 679:
      ACCEPT_TOKEN(anon_sym_POUND);
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'e') ADVANCE(496);
      if (lookahead == 'i') ADVANCE(480);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(199);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 680:
      ACCEPT_TOKEN(anon_sym_POUND);
      if (lookahead == 'd') ADVANCE(470);
      if (lookahead == 'i') ADVANCE(480);
      if (lookahead == 'u') ADVANCE(498);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(201);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(505);
      END_STATE();
    case 681:
      ACCEPT_TOKEN(anon_sym_ATprotocol);
      END_STATE();
    case 682:
      ACCEPT_TOKEN(anon_sym_ATend);
      END_STATE();
    case 683:
      ACCEPT_TOKEN(anon_sym_ATinterface);
      END_STATE();
    case 684:
      ACCEPT_TOKEN(anon_sym_ATimplementation);
      END_STATE();
    case 685:
      ACCEPT_TOKEN(anon_sym_ATprivate);
      END_STATE();
    case 686:
      ACCEPT_TOKEN(anon_sym_ATprotected);
      END_STATE();
    case 687:
      ACCEPT_TOKEN(anon_sym_ATpackage);
      END_STATE();
    case 688:
      ACCEPT_TOKEN(anon_sym_ATpublic);
      END_STATE();
    case 689:
      ACCEPT_TOKEN(anon_sym_ATcompatibility_alias);
      END_STATE();
    case 690:
      ACCEPT_TOKEN(anon_sym_AToptional);
      END_STATE();
    case 691:
      ACCEPT_TOKEN(anon_sym_ATrequired);
      END_STATE();
    case 692:
      ACCEPT_TOKEN(anon_sym_ATsynthesize);
      END_STATE();
    case 693:
      ACCEPT_TOKEN(anon_sym_ATdynamic);
      END_STATE();
    case 694:
      ACCEPT_TOKEN(anon_sym_LPARENclass_RPAREN);
      END_STATE();
    case 695:
      ACCEPT_TOKEN(anon_sym_ATproperty);
      END_STATE();
    case 696:
      ACCEPT_TOKEN(anon_sym_ATtry);
      END_STATE();
    case 697:
      ACCEPT_TOKEN(anon_sym_ATcatch);
      END_STATE();
    case 698:
      ACCEPT_TOKEN(anon_sym_ATfinally);
      END_STATE();
    case 699:
      ACCEPT_TOKEN(anon_sym_ATthrow);
      END_STATE();
    case 700:
      ACCEPT_TOKEN(anon_sym_ATselector);
      END_STATE();
    case 701:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (lookahead == '\n') ADVANCE(705);
      if (lookahead == '\r') ADVANCE(702);
      if (lookahead == 'U') ADVANCE(717);
      if (lookahead == 'u') ADVANCE(713);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 702:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (lookahead == '\n') ADVANCE(705);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 703:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (lookahead == '\r') ADVANCE(665);
      if (lookahead == ')') ADVANCE(670);
      if (lookahead == '\\') ADVANCE(662);
      if (lookahead != 0) ADVANCE(659);
      END_STATE();
    case 704:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (lookahead == '(') ADVANCE(507);
      if (lookahead == '/') ADVANCE(708);
      if (lookahead == ':') ADVANCE(732);
      if (lookahead == '\\') ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) ADVANCE(705);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(656);
      if (lookahead != 0 &&
          lookahead != '(' &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 705:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (lookahead == '(') ADVANCE(507);
      if (lookahead == '/') ADVANCE(708);
      if (lookahead == '\\') ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) ADVANCE(705);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(656);
      if (lookahead != 0 &&
          lookahead != '(' &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 706:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (lookahead == ')') ADVANCE(141);
      if (lookahead == '*') ADVANCE(706);
      if (lookahead == '/') ADVANCE(673);
      if (lookahead != 0) ADVANCE(707);
      END_STATE();
    case 707:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (lookahead == ')') ADVANCE(141);
      if (lookahead == '*') ADVANCE(706);
      if (lookahead != 0) ADVANCE(707);
      END_STATE();
    case 708:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (lookahead == '*') ADVANCE(707);
      if (lookahead == '/') ADVANCE(659);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != '*') ADVANCE(718);
      END_STATE();
    case 709:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (lookahead == 'U') ADVANCE(717);
      if (lookahead == 'u') ADVANCE(713);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 710:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(656);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 711:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(710);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 712:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(711);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 713:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(712);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 714:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(713);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 715:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(714);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 716:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(715);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 717:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(716);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 718:
      ACCEPT_TOKEN(aux_sym_selector_expression_token1);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    case 719:
      ACCEPT_TOKEN(anon_sym_ATavailable);
      END_STATE();
    case 720:
      ACCEPT_TOKEN(aux_sym_ms_asm_block_token1);
      if (lookahead == '\n') ADVANCE(726);
      if (lookahead == '\r') ADVANCE(721);
      if (lookahead != 0 &&
          lookahead != '}') ADVANCE(727);
      END_STATE();
    case 721:
      ACCEPT_TOKEN(aux_sym_ms_asm_block_token1);
      if (lookahead == '\n') ADVANCE(726);
      if (lookahead != 0 &&
          lookahead != '}') ADVANCE(727);
      END_STATE();
    case 722:
      ACCEPT_TOKEN(aux_sym_ms_asm_block_token1);
      if (lookahead == '\r') ADVANCE(672);
      if (lookahead == '\\') ADVANCE(664);
      if (lookahead == '}') ADVANCE(670);
      if (lookahead != 0) ADVANCE(660);
      END_STATE();
    case 723:
      ACCEPT_TOKEN(aux_sym_ms_asm_block_token1);
      if (lookahead == '*') ADVANCE(725);
      if (lookahead == '/') ADVANCE(660);
      if (lookahead != 0 &&
          lookahead != '}') ADVANCE(727);
      END_STATE();
    case 724:
      ACCEPT_TOKEN(aux_sym_ms_asm_block_token1);
      if (lookahead == '*') ADVANCE(724);
      if (lookahead == '/') ADVANCE(674);
      if (lookahead == '}') ADVANCE(141);
      if (lookahead != 0) ADVANCE(725);
      END_STATE();
    case 725:
      ACCEPT_TOKEN(aux_sym_ms_asm_block_token1);
      if (lookahead == '*') ADVANCE(724);
      if (lookahead == '}') ADVANCE(141);
      if (lookahead != 0) ADVANCE(725);
      END_STATE();
    case 726:
      ACCEPT_TOKEN(aux_sym_ms_asm_block_token1);
      if (lookahead == '/') ADVANCE(723);
      if (lookahead == '\\') ADVANCE(720);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) ADVANCE(726);
      if (lookahead != 0 &&
          lookahead != '}') ADVANCE(727);
      END_STATE();
    case 727:
      ACCEPT_TOKEN(aux_sym_ms_asm_block_token1);
      if (lookahead != 0 &&
          lookahead != '}') ADVANCE(727);
      END_STATE();
    case 728:
      ACCEPT_TOKEN(anon_sym_ATencode);
      END_STATE();
    case 729:
      ACCEPT_TOKEN(anon_sym_ATsynchronized);
      END_STATE();
    case 730:
      ACCEPT_TOKEN(anon_sym_ATdefs);
      END_STATE();
    case 731:
      ACCEPT_TOKEN(anon_sym_COLON2);
      END_STATE();
    case 732:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(718);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        'A', 1,
        'B', 2,
        'C', 3,
        'D', 4,
        'F', 5,
        'I', 6,
        'N', 7,
        'O', 8,
        'S', 9,
        'T', 10,
        'U', 11,
      );
      if (lookahead == '\\') SKIP(12);
      if (lookahead == '_') ADVANCE(13);
      if (lookahead == 'a') ADVANCE(14);
      if (lookahead == 'b') ADVANCE(15);
      if (lookahead == 'c') ADVANCE(16);
      if (lookahead == 'd') ADVANCE(17);
      if (lookahead == 'e') ADVANCE(18);
      if (lookahead == 'f') ADVANCE(19);
      if (lookahead == 'g') ADVANCE(20);
      if (lookahead == 'i') ADVANCE(21);
      if (lookahead == 'l') ADVANCE(22);
      if (lookahead == 'm') ADVANCE(23);
      if (lookahead == 'n') ADVANCE(24);
      if (lookahead == 'o') ADVANCE(25);
      if (lookahead == 'p') ADVANCE(26);
      if (lookahead == 'r') ADVANCE(27);
      if (lookahead == 's') ADVANCE(28);
      if (lookahead == 't') ADVANCE(29);
      if (lookahead == 'u') ADVANCE(30);
      if (lookahead == 'v') ADVANCE(31);
      if (lookahead == 'w') ADVANCE(32);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(0);
      END_STATE();
    case 1:
      if (lookahead == 'P') ADVANCE(33);
      END_STATE();
    case 2:
      if (lookahead == 'O') ADVANCE(34);
      END_STATE();
    case 3:
      if (lookahead == 'F') ADVANCE(35);
      if (lookahead == 'G') ADVANCE(36);
      if (lookahead == 'l') ADVANCE(37);
      END_STATE();
    case 4:
      if (lookahead == 'E') ADVANCE(38);
      END_STATE();
    case 5:
      if (lookahead == 'A') ADVANCE(39);
      if (lookahead == 'O') ADVANCE(40);
      END_STATE();
    case 6:
      if (lookahead == 'B') ADVANCE(41);
      if (lookahead == 'M') ADVANCE(42);
      END_STATE();
    case 7:
      if (lookahead == 'S') ADVANCE(43);
      if (lookahead == 'U') ADVANCE(44);
      END_STATE();
    case 8:
      if (lookahead == 'B') ADVANCE(45);
      END_STATE();
    case 9:
      if (lookahead == 'E') ADVANCE(46);
      END_STATE();
    case 10:
      if (lookahead == 'R') ADVANCE(47);
      END_STATE();
    case 11:
      if (lookahead == 'I') ADVANCE(48);
      if (lookahead == 'N') ADVANCE(49);
      END_STATE();
    case 12:
      if (lookahead == '\n') SKIP(0);
      if (lookahead == '\r') SKIP(50);
      END_STATE();
    case 13:
      if (lookahead == 'A') ADVANCE(51);
      if (lookahead == 'C') ADVANCE(52);
      if (lookahead == 'G') ADVANCE(53);
      if (lookahead == 'N') ADVANCE(54);
      if (lookahead == '_') ADVANCE(55);
      if (lookahead == 'a') ADVANCE(56);
      if (lookahead == 'u') ADVANCE(57);
      END_STATE();
    case 14:
      if (lookahead == 'l') ADVANCE(58);
      if (lookahead == 's') ADVANCE(59);
      if (lookahead == 'u') ADVANCE(60);
      if (lookahead == 'v') ADVANCE(61);
      END_STATE();
    case 15:
      if (lookahead == 'o') ADVANCE(62);
      if (lookahead == 'r') ADVANCE(63);
      if (lookahead == 'y') ADVANCE(64);
      END_STATE();
    case 16:
      if (lookahead == 'a') ADVANCE(65);
      if (lookahead == 'h') ADVANCE(66);
      if (lookahead == 'l') ADVANCE(67);
      if (lookahead == 'o') ADVANCE(68);
      END_STATE();
    case 17:
      if (lookahead == 'e') ADVANCE(69);
      if (lookahead == 'o') ADVANCE(70);
      END_STATE();
    case 18:
      if (lookahead == 'l') ADVANCE(71);
      if (lookahead == 'n') ADVANCE(72);
      if (lookahead == 'x') ADVANCE(73);
      END_STATE();
    case 19:
      if (lookahead == 'a') ADVANCE(74);
      if (lookahead == 'l') ADVANCE(75);
      if (lookahead == 'o') ADVANCE(76);
      END_STATE();
    case 20:
      if (lookahead == 'o') ADVANCE(77);
      END_STATE();
    case 21:
      if (lookahead == 'd') ADVANCE(78);
      if (lookahead == 'f') ADVANCE(79);
      if (lookahead == 'n') ADVANCE(80);
      if (lookahead == 'o') ADVANCE(81);
      END_STATE();
    case 22:
      if (lookahead == 'o') ADVANCE(82);
      END_STATE();
    case 23:
      if (lookahead == 'a') ADVANCE(83);
      END_STATE();
    case 24:
      if (lookahead == 'o') ADVANCE(84);
      if (lookahead == 'u') ADVANCE(85);
      END_STATE();
    case 25:
      if (lookahead == 'b') ADVANCE(86);
      if (lookahead == 'f') ADVANCE(87);
      if (lookahead == 'n') ADVANCE(88);
      if (lookahead == 'u') ADVANCE(89);
      END_STATE();
    case 26:
      if (lookahead == 't') ADVANCE(90);
      END_STATE();
    case 27:
      if (lookahead == 'e') ADVANCE(91);
      END_STATE();
    case 28:
      if (lookahead == 'h') ADVANCE(92);
      if (lookahead == 'i') ADVANCE(93);
      if (lookahead == 's') ADVANCE(94);
      if (lookahead == 't') ADVANCE(95);
      if (lookahead == 'w') ADVANCE(96);
      END_STATE();
    case 29:
      if (lookahead == 'h') ADVANCE(97);
      if (lookahead == 'r') ADVANCE(98);
      if (lookahead == 'v') ADVANCE(99);
      if (lookahead == 'y') ADVANCE(100);
      END_STATE();
    case 30:
      if (lookahead == 'i') ADVANCE(101);
      if (lookahead == 'n') ADVANCE(102);
      END_STATE();
    case 31:
      if (lookahead == 'a') ADVANCE(103);
      if (lookahead == 'o') ADVANCE(104);
      END_STATE();
    case 32:
      if (lookahead == 'a') ADVANCE(105);
      if (lookahead == 'h') ADVANCE(106);
      END_STATE();
    case 33:
      if (lookahead == 'I') ADVANCE(107);
      END_STATE();
    case 34:
      if (lookahead == 'O') ADVANCE(108);
      END_STATE();
    case 35:
      if (lookahead == '_') ADVANCE(109);
      END_STATE();
    case 36:
      if (lookahead == '_') ADVANCE(110);
      END_STATE();
    case 37:
      if (lookahead == 'a') ADVANCE(111);
      END_STATE();
    case 38:
      if (lookahead == 'P') ADVANCE(112);
      END_STATE();
    case 39:
      if (lookahead == 'L') ADVANCE(113);
      END_STATE();
    case 40:
      if (lookahead == 'U') ADVANCE(114);
      END_STATE();
    case 41:
      if (lookahead == 'I') ADVANCE(115);
      if (lookahead == 'O') ADVANCE(116);
      if (lookahead == '_') ADVANCE(117);
      END_STATE();
    case 42:
      if (lookahead == 'P') ADVANCE(118);
      END_STATE();
    case 43:
      if (lookahead == '_') ADVANCE(119);
      END_STATE();
    case 44:
      if (lookahead == 'L') ADVANCE(120);
      END_STATE();
    case 45:
      if (lookahead == 'J') ADVANCE(121);
      END_STATE();
    case 46:
      if (lookahead == 'L') ADVANCE(122);
      END_STATE();
    case 47:
      if (lookahead == 'U') ADVANCE(123);
      END_STATE();
    case 48:
      if (lookahead == 'K') ADVANCE(124);
      if (lookahead == '_') ADVANCE(125);
      END_STATE();
    case 49:
      if (lookahead == 'A') ADVANCE(126);
      END_STATE();
    case 50:
      if (lookahead == '\n') SKIP(0);
      END_STATE();
    case 51:
      if (lookahead == 'l') ADVANCE(127);
      if (lookahead == 't') ADVANCE(128);
      END_STATE();
    case 52:
      if (lookahead == 'o') ADVANCE(129);
      END_STATE();
    case 53:
      if (lookahead == 'e') ADVANCE(130);
      END_STATE();
    case 54:
      if (lookahead == 'o') ADVANCE(131);
      if (lookahead == 'u') ADVANCE(132);
      END_STATE();
    case 55:
      ADVANCE_MAP(
        'I', 133,
        'O', 134,
        'a', 135,
        'b', 136,
        'c', 137,
        'd', 138,
        'e', 139,
        'f', 140,
        'i', 141,
        'k', 142,
        'n', 143,
        'p', 144,
        'r', 145,
        's', 146,
        't', 147,
        'u', 148,
        'v', 149,
        'w', 150,
      );
      END_STATE();
    case 56:
      if (lookahead == 'l') ADVANCE(151);
      END_STATE();
    case 57:
      if (lookahead == 'n') ADVANCE(152);
      END_STATE();
    case 58:
      if (lookahead == 'i') ADVANCE(153);
      END_STATE();
    case 59:
      if (lookahead == 'm') ADVANCE(154);
      END_STATE();
    case 60:
      if (lookahead == 't') ADVANCE(155);
      END_STATE();
    case 61:
      if (lookahead == 'a') ADVANCE(156);
      END_STATE();
    case 62:
      if (lookahead == 'o') ADVANCE(157);
      END_STATE();
    case 63:
      if (lookahead == 'e') ADVANCE(158);
      END_STATE();
    case 64:
      if (lookahead == 'c') ADVANCE(159);
      if (lookahead == 'r') ADVANCE(160);
      END_STATE();
    case 65:
      if (lookahead == 's') ADVANCE(161);
      END_STATE();
    case 66:
      if (lookahead == 'a') ADVANCE(162);
      END_STATE();
    case 67:
      if (lookahead == 'a') ADVANCE(163);
      END_STATE();
    case 68:
      if (lookahead == 'n') ADVANCE(164);
      END_STATE();
    case 69:
      if (lookahead == 'f') ADVANCE(165);
      END_STATE();
    case 70:
      ACCEPT_TOKEN(anon_sym_do);
      if (lookahead == 'u') ADVANCE(166);
      END_STATE();
    case 71:
      if (lookahead == 's') ADVANCE(167);
      END_STATE();
    case 72:
      if (lookahead == 'u') ADVANCE(168);
      END_STATE();
    case 73:
      if (lookahead == 't') ADVANCE(169);
      END_STATE();
    case 74:
      if (lookahead == 'l') ADVANCE(170);
      END_STATE();
    case 75:
      if (lookahead == 'o') ADVANCE(171);
      END_STATE();
    case 76:
      if (lookahead == 'r') ADVANCE(172);
      END_STATE();
    case 77:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 78:
      ACCEPT_TOKEN(anon_sym_id);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 80:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == 'l') ADVANCE(174);
      if (lookahead == 'o') ADVANCE(175);
      if (lookahead == 't') ADVANCE(176);
      END_STATE();
    case 81:
      if (lookahead == 's') ADVANCE(177);
      END_STATE();
    case 82:
      if (lookahead == 'n') ADVANCE(178);
      END_STATE();
    case 83:
      if (lookahead == 'c') ADVANCE(179);
      if (lookahead == 'x') ADVANCE(180);
      END_STATE();
    case 84:
      if (lookahead == 'r') ADVANCE(181);
      if (lookahead == 't') ADVANCE(182);
      END_STATE();
    case 85:
      if (lookahead == 'l') ADVANCE(183);
      END_STATE();
    case 86:
      if (lookahead == 'j') ADVANCE(184);
      END_STATE();
    case 87:
      if (lookahead == 'f') ADVANCE(185);
      END_STATE();
    case 88:
      if (lookahead == 'e') ADVANCE(186);
      END_STATE();
    case 89:
      if (lookahead == 't') ADVANCE(187);
      END_STATE();
    case 90:
      if (lookahead == 'r') ADVANCE(188);
      END_STATE();
    case 91:
      if (lookahead == 'g') ADVANCE(189);
      if (lookahead == 's') ADVANCE(190);
      if (lookahead == 't') ADVANCE(191);
      END_STATE();
    case 92:
      if (lookahead == 'o') ADVANCE(192);
      END_STATE();
    case 93:
      if (lookahead == 'g') ADVANCE(193);
      if (lookahead == 'z') ADVANCE(194);
      END_STATE();
    case 94:
      if (lookahead == 'i') ADVANCE(195);
      END_STATE();
    case 95:
      if (lookahead == 'a') ADVANCE(196);
      if (lookahead == 'r') ADVANCE(197);
      END_STATE();
    case 96:
      if (lookahead == 'i') ADVANCE(198);
      END_STATE();
    case 97:
      if (lookahead == 'r') ADVANCE(199);
      END_STATE();
    case 98:
      if (lookahead == 'u') ADVANCE(200);
      END_STATE();
    case 99:
      if (lookahead == 'o') ADVANCE(201);
      END_STATE();
    case 100:
      if (lookahead == 'p') ADVANCE(202);
      END_STATE();
    case 101:
      if (lookahead == 'n') ADVANCE(203);
      END_STATE();
    case 102:
      if (lookahead == 'i') ADVANCE(204);
      if (lookahead == 's') ADVANCE(205);
      END_STATE();
    case 103:
      if (lookahead == '_') ADVANCE(206);
      END_STATE();
    case 104:
      if (lookahead == 'i') ADVANCE(207);
      if (lookahead == 'l') ADVANCE(208);
      END_STATE();
    case 105:
      if (lookahead == 't') ADVANCE(209);
      END_STATE();
    case 106:
      if (lookahead == 'i') ADVANCE(210);
      END_STATE();
    case 107:
      if (lookahead == '_') ADVANCE(211);
      END_STATE();
    case 108:
      if (lookahead == 'L') ADVANCE(212);
      END_STATE();
    case 109:
      if (lookahead == 'F') ADVANCE(213);
      if (lookahead == 'R') ADVANCE(214);
      END_STATE();
    case 110:
      if (lookahead == 'E') ADVANCE(215);
      if (lookahead == 'I') ADVANCE(216);
      END_STATE();
    case 111:
      if (lookahead == 's') ADVANCE(217);
      END_STATE();
    case 112:
      if (lookahead == 'R') ADVANCE(218);
      END_STATE();
    case 113:
      if (lookahead == 'S') ADVANCE(219);
      END_STATE();
    case 114:
      if (lookahead == 'N') ADVANCE(220);
      END_STATE();
    case 115:
      if (lookahead == 'n') ADVANCE(221);
      END_STATE();
    case 116:
      if (lookahead == 'u') ADVANCE(222);
      END_STATE();
    case 117:
      if (lookahead == 'D') ADVANCE(223);
      END_STATE();
    case 118:
      ACCEPT_TOKEN(anon_sym_IMP);
      END_STATE();
    case 119:
      ADVANCE_MAP(
        'A', 224,
        'C', 225,
        'D', 226,
        'E', 227,
        'F', 228,
        'I', 229,
        'R', 230,
        'S', 231,
        'U', 232,
        'V', 233,
      );
      END_STATE();
    case 120:
      if (lookahead == 'L') ADVANCE(234);
      END_STATE();
    case 121:
      if (lookahead == 'C') ADVANCE(235);
      END_STATE();
    case 122:
      ACCEPT_TOKEN(anon_sym_SEL);
      END_STATE();
    case 123:
      if (lookahead == 'E') ADVANCE(236);
      END_STATE();
    case 124:
      if (lookahead == 'I') ADVANCE(237);
      END_STATE();
    case 125:
      if (lookahead == 'A') ADVANCE(238);
      END_STATE();
    case 126:
      if (lookahead == 'V') ADVANCE(239);
      END_STATE();
    case 127:
      if (lookahead == 'i') ADVANCE(240);
      END_STATE();
    case 128:
      if (lookahead == 'o') ADVANCE(241);
      END_STATE();
    case 129:
      if (lookahead == 'm') ADVANCE(242);
      END_STATE();
    case 130:
      if (lookahead == 'n') ADVANCE(243);
      END_STATE();
    case 131:
      if (lookahead == 'n') ADVANCE(244);
      if (lookahead == 'r') ADVANCE(245);
      END_STATE();
    case 132:
      if (lookahead == 'l') ADVANCE(246);
      END_STATE();
    case 133:
      if (lookahead == 'O') ADVANCE(247);
      END_STATE();
    case 134:
      if (lookahead == 'S') ADVANCE(248);
      END_STATE();
    case 135:
      if (lookahead == 'l') ADVANCE(249);
      if (lookahead == 's') ADVANCE(250);
      if (lookahead == 't') ADVANCE(251);
      if (lookahead == 'u') ADVANCE(252);
      END_STATE();
    case 136:
      if (lookahead == 'a') ADVANCE(253);
      if (lookahead == 'l') ADVANCE(254);
      if (lookahead == 'r') ADVANCE(255);
      if (lookahead == 'u') ADVANCE(256);
      END_STATE();
    case 137:
      if (lookahead == 'a') ADVANCE(257);
      if (lookahead == 'd') ADVANCE(258);
      if (lookahead == 'l') ADVANCE(259);
      if (lookahead == 'o') ADVANCE(260);
      END_STATE();
    case 138:
      if (lookahead == 'e') ADVANCE(261);
      END_STATE();
    case 139:
      if (lookahead == 'x') ADVANCE(262);
      END_STATE();
    case 140:
      if (lookahead == 'a') ADVANCE(263);
      if (lookahead == 'i') ADVANCE(264);
      if (lookahead == 'o') ADVANCE(265);
      END_STATE();
    case 141:
      if (lookahead == 'm') ADVANCE(266);
      if (lookahead == 'n') ADVANCE(267);
      END_STATE();
    case 142:
      if (lookahead == 'i') ADVANCE(268);
      END_STATE();
    case 143:
      if (lookahead == 'o') ADVANCE(269);
      if (lookahead == 'u') ADVANCE(270);
      END_STATE();
    case 144:
      if (lookahead == 't') ADVANCE(271);
      END_STATE();
    case 145:
      if (lookahead == 'e') ADVANCE(272);
      END_STATE();
    case 146:
      if (lookahead == 'p') ADVANCE(273);
      if (lookahead == 't') ADVANCE(274);
      END_STATE();
    case 147:
      if (lookahead == 'h') ADVANCE(275);
      if (lookahead == 'r') ADVANCE(276);
      if (lookahead == 'y') ADVANCE(277);
      END_STATE();
    case 148:
      if (lookahead == 'n') ADVANCE(278);
      if (lookahead == 'p') ADVANCE(279);
      END_STATE();
    case 149:
      if (lookahead == 'e') ADVANCE(280);
      if (lookahead == 'o') ADVANCE(281);
      END_STATE();
    case 150:
      if (lookahead == 'e') ADVANCE(282);
      END_STATE();
    case 151:
      if (lookahead == 'i') ADVANCE(283);
      END_STATE();
    case 152:
      if (lookahead == 'a') ADVANCE(284);
      END_STATE();
    case 153:
      if (lookahead == 'g') ADVANCE(285);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(anon_sym_asm);
      END_STATE();
    case 155:
      if (lookahead == 'o') ADVANCE(286);
      END_STATE();
    case 156:
      if (lookahead == 'i') ADVANCE(287);
      END_STATE();
    case 157:
      if (lookahead == 'l') ADVANCE(288);
      END_STATE();
    case 158:
      if (lookahead == 'a') ADVANCE(289);
      END_STATE();
    case 159:
      if (lookahead == 'o') ADVANCE(290);
      END_STATE();
    case 160:
      if (lookahead == 'e') ADVANCE(291);
      END_STATE();
    case 161:
      if (lookahead == 'e') ADVANCE(292);
      END_STATE();
    case 162:
      if (lookahead == 'r') ADVANCE(293);
      END_STATE();
    case 163:
      if (lookahead == 's') ADVANCE(294);
      END_STATE();
    case 164:
      if (lookahead == 's') ADVANCE(295);
      if (lookahead == 't') ADVANCE(296);
      END_STATE();
    case 165:
      if (lookahead == 'a') ADVANCE(297);
      if (lookahead == 'i') ADVANCE(298);
      END_STATE();
    case 166:
      if (lookahead == 'b') ADVANCE(299);
      END_STATE();
    case 167:
      if (lookahead == 'e') ADVANCE(300);
      END_STATE();
    case 168:
      if (lookahead == 'm') ADVANCE(301);
      END_STATE();
    case 169:
      if (lookahead == 'e') ADVANCE(302);
      END_STATE();
    case 170:
      if (lookahead == 's') ADVANCE(303);
      END_STATE();
    case 171:
      if (lookahead == 'a') ADVANCE(304);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 173:
      if (lookahead == 'o') ADVANCE(305);
      END_STATE();
    case 174:
      if (lookahead == 'i') ADVANCE(306);
      END_STATE();
    case 175:
      if (lookahead == 'u') ADVANCE(307);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(sym_primitive_type);
      if (lookahead == '1') ADVANCE(308);
      if (lookahead == '3') ADVANCE(309);
      if (lookahead == '6') ADVANCE(310);
      if (lookahead == '8') ADVANCE(311);
      if (lookahead == 'p') ADVANCE(312);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_ios);
      END_STATE();
    case 178:
      if (lookahead == 'g') ADVANCE(313);
      END_STATE();
    case 179:
      if (lookahead == 'o') ADVANCE(314);
      END_STATE();
    case 180:
      if (lookahead == '_') ADVANCE(315);
      END_STATE();
    case 181:
      if (lookahead == 'e') ADVANCE(316);
      END_STATE();
    case 182:
      if (lookahead == 'h') ADVANCE(317);
      END_STATE();
    case 183:
      if (lookahead == 'l') ADVANCE(318);
      END_STATE();
    case 184:
      if (lookahead == 'c') ADVANCE(319);
      END_STATE();
    case 185:
      if (lookahead == 's') ADVANCE(320);
      END_STATE();
    case 186:
      if (lookahead == 'w') ADVANCE(321);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_out);
      END_STATE();
    case 188:
      if (lookahead == 'd') ADVANCE(322);
      END_STATE();
    case 189:
      if (lookahead == 'i') ADVANCE(323);
      END_STATE();
    case 190:
      if (lookahead == 't') ADVANCE(324);
      END_STATE();
    case 191:
      if (lookahead == 'u') ADVANCE(325);
      END_STATE();
    case 192:
      if (lookahead == 'r') ADVANCE(326);
      END_STATE();
    case 193:
      if (lookahead == 'n') ADVANCE(327);
      END_STATE();
    case 194:
      if (lookahead == 'e') ADVANCE(328);
      END_STATE();
    case 195:
      if (lookahead == 'z') ADVANCE(329);
      END_STATE();
    case 196:
      if (lookahead == 't') ADVANCE(330);
      END_STATE();
    case 197:
      if (lookahead == 'u') ADVANCE(331);
      END_STATE();
    case 198:
      if (lookahead == 't') ADVANCE(332);
      END_STATE();
    case 199:
      if (lookahead == 'e') ADVANCE(333);
      END_STATE();
    case 200:
      if (lookahead == 'e') ADVANCE(236);
      END_STATE();
    case 201:
      if (lookahead == 's') ADVANCE(334);
      END_STATE();
    case 202:
      if (lookahead == 'e') ADVANCE(335);
      END_STATE();
    case 203:
      if (lookahead == 't') ADVANCE(336);
      END_STATE();
    case 204:
      if (lookahead == 'o') ADVANCE(337);
      END_STATE();
    case 205:
      if (lookahead == 'i') ADVANCE(338);
      END_STATE();
    case 206:
      if (lookahead == 'a') ADVANCE(339);
      END_STATE();
    case 207:
      if (lookahead == 'd') ADVANCE(288);
      END_STATE();
    case 208:
      if (lookahead == 'a') ADVANCE(340);
      END_STATE();
    case 209:
      if (lookahead == 'c') ADVANCE(341);
      END_STATE();
    case 210:
      if (lookahead == 'l') ADVANCE(342);
      END_STATE();
    case 211:
      if (lookahead == 'A') ADVANCE(343);
      if (lookahead == 'D') ADVANCE(344);
      if (lookahead == 'U') ADVANCE(345);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_BOOL);
      END_STATE();
    case 213:
      if (lookahead == 'O') ADVANCE(346);
      END_STATE();
    case 214:
      if (lookahead == 'E') ADVANCE(347);
      END_STATE();
    case 215:
      if (lookahead == 'X') ADVANCE(348);
      END_STATE();
    case 216:
      if (lookahead == 'N') ADVANCE(349);
      END_STATE();
    case 217:
      if (lookahead == 's') ADVANCE(350);
      END_STATE();
    case 218:
      if (lookahead == 'E') ADVANCE(351);
      END_STATE();
    case 219:
      if (lookahead == 'E') ADVANCE(352);
      END_STATE();
    case 220:
      if (lookahead == 'D') ADVANCE(353);
      END_STATE();
    case 221:
      if (lookahead == 's') ADVANCE(354);
      END_STATE();
    case 222:
      if (lookahead == 't') ADVANCE(355);
      END_STATE();
    case 223:
      if (lookahead == 'E') ADVANCE(356);
      END_STATE();
    case 224:
      if (lookahead == 'U') ADVANCE(357);
      if (lookahead == 'V') ADVANCE(358);
      END_STATE();
    case 225:
      if (lookahead == 'L') ADVANCE(359);
      END_STATE();
    case 226:
      if (lookahead == 'E') ADVANCE(360);
      END_STATE();
    case 227:
      if (lookahead == 'N') ADVANCE(361);
      if (lookahead == 'X') ADVANCE(362);
      END_STATE();
    case 228:
      if (lookahead == 'O') ADVANCE(363);
      END_STATE();
    case 229:
      if (lookahead == 'N') ADVANCE(364);
      END_STATE();
    case 230:
      if (lookahead == 'E') ADVANCE(365);
      if (lookahead == 'O') ADVANCE(366);
      END_STATE();
    case 231:
      if (lookahead == 'W') ADVANCE(367);
      END_STATE();
    case 232:
      if (lookahead == 'N') ADVANCE(368);
      END_STATE();
    case 233:
      if (lookahead == 'A') ADVANCE(369);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_NULL);
      END_STATE();
    case 235:
      if (lookahead == '_') ADVANCE(370);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(sym_true);
      END_STATE();
    case 237:
      if (lookahead == 'T') ADVANCE(371);
      END_STATE();
    case 238:
      if (lookahead == 'P') ADVANCE(372);
      END_STATE();
    case 239:
      if (lookahead == 'A') ADVANCE(373);
      END_STATE();
    case 240:
      if (lookahead == 'g') ADVANCE(374);
      END_STATE();
    case 241:
      if (lookahead == 'm') ADVANCE(375);
      END_STATE();
    case 242:
      if (lookahead == 'p') ADVANCE(376);
      END_STATE();
    case 243:
      if (lookahead == 'e') ADVANCE(377);
      END_STATE();
    case 244:
      if (lookahead == 'n') ADVANCE(378);
      END_STATE();
    case 245:
      if (lookahead == 'e') ADVANCE(379);
      END_STATE();
    case 246:
      if (lookahead == 'l') ADVANCE(380);
      END_STATE();
    case 247:
      if (lookahead == 'S') ADVANCE(381);
      END_STATE();
    case 248:
      if (lookahead == 'X') ADVANCE(382);
      END_STATE();
    case 249:
      if (lookahead == 'i') ADVANCE(383);
      END_STATE();
    case 250:
      if (lookahead == 'm') ADVANCE(384);
      END_STATE();
    case 251:
      if (lookahead == 't') ADVANCE(385);
      END_STATE();
    case 252:
      if (lookahead == 't') ADVANCE(386);
      END_STATE();
    case 253:
      if (lookahead == 's') ADVANCE(387);
      END_STATE();
    case 254:
      if (lookahead == 'o') ADVANCE(388);
      END_STATE();
    case 255:
      if (lookahead == 'i') ADVANCE(389);
      END_STATE();
    case 256:
      if (lookahead == 'i') ADVANCE(390);
      END_STATE();
    case 257:
      if (lookahead == 't') ADVANCE(391);
      END_STATE();
    case 258:
      if (lookahead == 'e') ADVANCE(392);
      END_STATE();
    case 259:
      if (lookahead == 'r') ADVANCE(393);
      END_STATE();
    case 260:
      if (lookahead == 'm') ADVANCE(394);
      if (lookahead == 'n') ADVANCE(395);
      if (lookahead == 'v') ADVANCE(396);
      END_STATE();
    case 261:
      if (lookahead == 'c') ADVANCE(397);
      if (lookahead == 'p') ADVANCE(398);
      END_STATE();
    case 262:
      if (lookahead == 't') ADVANCE(399);
      END_STATE();
    case 263:
      if (lookahead == 's') ADVANCE(400);
      END_STATE();
    case 264:
      if (lookahead == 'n') ADVANCE(401);
      END_STATE();
    case 265:
      if (lookahead == 'r') ADVANCE(402);
      END_STATE();
    case 266:
      if (lookahead == 'a') ADVANCE(403);
      END_STATE();
    case 267:
      if (lookahead == 'l') ADVANCE(404);
      END_STATE();
    case 268:
      if (lookahead == 'n') ADVANCE(405);
      END_STATE();
    case 269:
      if (lookahead == 'n') ADVANCE(406);
      END_STATE();
    case 270:
      if (lookahead == 'l') ADVANCE(407);
      END_STATE();
    case 271:
      if (lookahead == 'r') ADVANCE(408);
      END_STATE();
    case 272:
      if (lookahead == 'a') ADVANCE(409);
      if (lookahead == 's') ADVANCE(410);
      END_STATE();
    case 273:
      if (lookahead == 't') ADVANCE(411);
      END_STATE();
    case 274:
      if (lookahead == 'd') ADVANCE(412);
      if (lookahead == 'r') ADVANCE(413);
      END_STATE();
    case 275:
      if (lookahead == 'i') ADVANCE(414);
      if (lookahead == 'r') ADVANCE(415);
      END_STATE();
    case 276:
      if (lookahead == 'y') ADVANCE(416);
      END_STATE();
    case 277:
      if (lookahead == 'p') ADVANCE(417);
      END_STATE();
    case 278:
      if (lookahead == 'a') ADVANCE(418);
      if (lookahead == 's') ADVANCE(419);
      if (lookahead == 'u') ADVANCE(420);
      END_STATE();
    case 279:
      if (lookahead == 't') ADVANCE(421);
      END_STATE();
    case 280:
      if (lookahead == 'c') ADVANCE(422);
      END_STATE();
    case 281:
      if (lookahead == 'l') ADVANCE(423);
      END_STATE();
    case 282:
      if (lookahead == 'a') ADVANCE(424);
      END_STATE();
    case 283:
      if (lookahead == 'g') ADVANCE(425);
      END_STATE();
    case 284:
      if (lookahead == 'l') ADVANCE(426);
      END_STATE();
    case 285:
      if (lookahead == 'n') ADVANCE(427);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(anon_sym_auto);
      END_STATE();
    case 287:
      if (lookahead == 'l') ADVANCE(428);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(sym_primitive_type);
      END_STATE();
    case 289:
      if (lookahead == 'k') ADVANCE(429);
      END_STATE();
    case 290:
      if (lookahead == 'p') ADVANCE(430);
      END_STATE();
    case 291:
      if (lookahead == 'f') ADVANCE(431);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(anon_sym_case);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(sym_primitive_type);
      if (lookahead == '1') ADVANCE(432);
      if (lookahead == '3') ADVANCE(433);
      if (lookahead == '6') ADVANCE(434);
      if (lookahead == '8') ADVANCE(435);
      if (lookahead == 'p') ADVANCE(436);
      END_STATE();
    case 294:
      if (lookahead == 's') ADVANCE(437);
      END_STATE();
    case 295:
      if (lookahead == 't') ADVANCE(438);
      END_STATE();
    case 296:
      if (lookahead == 'i') ADVANCE(439);
      END_STATE();
    case 297:
      if (lookahead == 'u') ADVANCE(440);
      END_STATE();
    case 298:
      if (lookahead == 'n') ADVANCE(441);
      END_STATE();
    case 299:
      if (lookahead == 'l') ADVANCE(442);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(anon_sym_enum);
      END_STATE();
    case 302:
      if (lookahead == 'r') ADVANCE(443);
      END_STATE();
    case 303:
      if (lookahead == 'e') ADVANCE(352);
      END_STATE();
    case 304:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(anon_sym_goto);
      END_STATE();
    case 306:
      if (lookahead == 'n') ADVANCE(444);
      END_STATE();
    case 307:
      if (lookahead == 't') ADVANCE(445);
      END_STATE();
    case 308:
      if (lookahead == '6') ADVANCE(446);
      END_STATE();
    case 309:
      if (lookahead == '2') ADVANCE(447);
      END_STATE();
    case 310:
      if (lookahead == '4') ADVANCE(448);
      END_STATE();
    case 311:
      if (lookahead == '_') ADVANCE(449);
      END_STATE();
    case 312:
      if (lookahead == 't') ADVANCE(450);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(anon_sym_long);
      END_STATE();
    case 314:
      if (lookahead == 's') ADVANCE(451);
      END_STATE();
    case 315:
      if (lookahead == 'a') ADVANCE(452);
      END_STATE();
    case 316:
      if (lookahead == 't') ADVANCE(453);
      END_STATE();
    case 317:
      if (lookahead == 'r') ADVANCE(454);
      END_STATE();
    case 318:
      if (lookahead == 'a') ADVANCE(455);
      if (lookahead == 'p') ADVANCE(456);
      END_STATE();
    case 319:
      if (lookahead == '_') ADVANCE(457);
      END_STATE();
    case 320:
      if (lookahead == 'e') ADVANCE(458);
      END_STATE();
    case 321:
      if (lookahead == 'a') ADVANCE(459);
      END_STATE();
    case 322:
      if (lookahead == 'i') ADVANCE(460);
      END_STATE();
    case 323:
      if (lookahead == 's') ADVANCE(461);
      END_STATE();
    case 324:
      if (lookahead == 'r') ADVANCE(462);
      END_STATE();
    case 325:
      if (lookahead == 'r') ADVANCE(463);
      END_STATE();
    case 326:
      if (lookahead == 't') ADVANCE(464);
      END_STATE();
    case 327:
      if (lookahead == 'e') ADVANCE(465);
      END_STATE();
    case 328:
      if (lookahead == '_') ADVANCE(466);
      if (lookahead == 'o') ADVANCE(467);
      END_STATE();
    case 329:
      if (lookahead == 'e') ADVANCE(468);
      END_STATE();
    case 330:
      if (lookahead == 'i') ADVANCE(469);
      END_STATE();
    case 331:
      if (lookahead == 'c') ADVANCE(470);
      END_STATE();
    case 332:
      if (lookahead == 'c') ADVANCE(471);
      END_STATE();
    case 333:
      if (lookahead == 'a') ADVANCE(472);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(anon_sym_tvos);
      END_STATE();
    case 335:
      if (lookahead == 'd') ADVANCE(473);
      if (lookahead == 'o') ADVANCE(474);
      END_STATE();
    case 336:
      if (lookahead == '1') ADVANCE(475);
      if (lookahead == '3') ADVANCE(476);
      if (lookahead == '6') ADVANCE(477);
      if (lookahead == '8') ADVANCE(478);
      if (lookahead == 'p') ADVANCE(479);
      END_STATE();
    case 337:
      if (lookahead == 'n') ADVANCE(480);
      END_STATE();
    case 338:
      if (lookahead == 'g') ADVANCE(481);
      END_STATE();
    case 339:
      if (lookahead == 'r') ADVANCE(482);
      END_STATE();
    case 340:
      if (lookahead == 't') ADVANCE(483);
      END_STATE();
    case 341:
      if (lookahead == 'h') ADVANCE(484);
      END_STATE();
    case 342:
      if (lookahead == 'e') ADVANCE(485);
      END_STATE();
    case 343:
      if (lookahead == 'V') ADVANCE(486);
      END_STATE();
    case 344:
      if (lookahead == 'E') ADVANCE(487);
      END_STATE();
    case 345:
      if (lookahead == 'N') ADVANCE(488);
      END_STATE();
    case 346:
      if (lookahead == 'R') ADVANCE(489);
      END_STATE();
    case 347:
      if (lookahead == 'T') ADVANCE(490);
      END_STATE();
    case 348:
      if (lookahead == 'T') ADVANCE(491);
      END_STATE();
    case 349:
      if (lookahead == 'L') ADVANCE(492);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(anon_sym_Class);
      END_STATE();
    case 351:
      if (lookahead == 'C') ADVANCE(493);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(sym_false);
      END_STATE();
    case 353:
      if (lookahead == 'A') ADVANCE(494);
      END_STATE();
    case 354:
      if (lookahead == 'p') ADVANCE(495);
      END_STATE();
    case 355:
      if (lookahead == 'l') ADVANCE(496);
      END_STATE();
    case 356:
      if (lookahead == 'S') ADVANCE(497);
      END_STATE();
    case 357:
      if (lookahead == 'T') ADVANCE(498);
      END_STATE();
    case 358:
      if (lookahead == 'A') ADVANCE(499);
      END_STATE();
    case 359:
      if (lookahead == 'A') ADVANCE(500);
      END_STATE();
    case 360:
      if (lookahead == 'P') ADVANCE(501);
      END_STATE();
    case 361:
      if (lookahead == 'U') ADVANCE(502);
      END_STATE();
    case 362:
      if (lookahead == 'T') ADVANCE(503);
      END_STATE();
    case 363:
      if (lookahead == 'R') ADVANCE(504);
      END_STATE();
    case 364:
      if (lookahead == 'L') ADVANCE(505);
      END_STATE();
    case 365:
      if (lookahead == 'Q') ADVANCE(506);
      END_STATE();
    case 366:
      if (lookahead == 'O') ADVANCE(507);
      END_STATE();
    case 367:
      if (lookahead == 'I') ADVANCE(508);
      END_STATE();
    case 368:
      if (lookahead == 'A') ADVANCE(509);
      END_STATE();
    case 369:
      if (lookahead == 'L') ADVANCE(510);
      END_STATE();
    case 370:
      if (lookahead == 'E') ADVANCE(511);
      if (lookahead == 'R') ADVANCE(512);
      END_STATE();
    case 371:
      if (lookahead == '_') ADVANCE(513);
      END_STATE();
    case 372:
      if (lookahead == 'P') ADVANCE(514);
      END_STATE();
    case 373:
      if (lookahead == 'I') ADVANCE(515);
      END_STATE();
    case 374:
      if (lookahead == 'n') ADVANCE(516);
      END_STATE();
    case 375:
      if (lookahead == 'i') ADVANCE(517);
      END_STATE();
    case 376:
      if (lookahead == 'l') ADVANCE(518);
      END_STATE();
    case 377:
      if (lookahead == 'r') ADVANCE(519);
      END_STATE();
    case 378:
      if (lookahead == 'u') ADVANCE(520);
      END_STATE();
    case 379:
      if (lookahead == 't') ADVANCE(521);
      END_STATE();
    case 380:
      if (lookahead == '_') ADVANCE(522);
      if (lookahead == 'a') ADVANCE(523);
      END_STATE();
    case 381:
      if (lookahead == '_') ADVANCE(524);
      END_STATE();
    case 382:
      if (lookahead == '_') ADVANCE(525);
      END_STATE();
    case 383:
      if (lookahead == 'g') ADVANCE(526);
      END_STATE();
    case 384:
      ACCEPT_TOKEN(anon_sym___asm);
      if (lookahead == '_') ADVANCE(527);
      END_STATE();
    case 385:
      if (lookahead == 'r') ADVANCE(528);
      END_STATE();
    case 386:
      if (lookahead == 'o') ADVANCE(529);
      END_STATE();
    case 387:
      if (lookahead == 'e') ADVANCE(530);
      END_STATE();
    case 388:
      if (lookahead == 'c') ADVANCE(531);
      END_STATE();
    case 389:
      if (lookahead == 'd') ADVANCE(532);
      END_STATE();
    case 390:
      if (lookahead == 'l') ADVANCE(533);
      END_STATE();
    case 391:
      if (lookahead == 'c') ADVANCE(534);
      END_STATE();
    case 392:
      if (lookahead == 'c') ADVANCE(535);
      END_STATE();
    case 393:
      if (lookahead == 'c') ADVANCE(536);
      END_STATE();
    case 394:
      if (lookahead == 'p') ADVANCE(537);
      END_STATE();
    case 395:
      if (lookahead == 's') ADVANCE(538);
      if (lookahead == 't') ADVANCE(539);
      END_STATE();
    case 396:
      if (lookahead == 'a') ADVANCE(540);
      END_STATE();
    case 397:
      if (lookahead == 'l') ADVANCE(541);
      END_STATE();
    case 398:
      if (lookahead == 'r') ADVANCE(542);
      END_STATE();
    case 399:
      if (lookahead == 'e') ADVANCE(543);
      END_STATE();
    case 400:
      if (lookahead == 't') ADVANCE(544);
      END_STATE();
    case 401:
      if (lookahead == 'a') ADVANCE(545);
      END_STATE();
    case 402:
      if (lookahead == 'c') ADVANCE(546);
      END_STATE();
    case 403:
      if (lookahead == 'g') ADVANCE(547);
      END_STATE();
    case 404:
      if (lookahead == 'i') ADVANCE(548);
      END_STATE();
    case 405:
      if (lookahead == 'd') ADVANCE(549);
      END_STATE();
    case 406:
      if (lookahead == 'n') ADVANCE(550);
      END_STATE();
    case 407:
      if (lookahead == 'l') ADVANCE(551);
      END_STATE();
    case 408:
      if (lookahead == 'a') ADVANCE(552);
      END_STATE();
    case 409:
      if (lookahead == 'l') ADVANCE(553);
      END_STATE();
    case 410:
      if (lookahead == 't') ADVANCE(554);
      END_STATE();
    case 411:
      if (lookahead == 'r') ADVANCE(555);
      END_STATE();
    case 412:
      if (lookahead == 'c') ADVANCE(556);
      END_STATE();
    case 413:
      if (lookahead == 'o') ADVANCE(557);
      END_STATE();
    case 414:
      if (lookahead == 's') ADVANCE(558);
      END_STATE();
    case 415:
      if (lookahead == 'e') ADVANCE(559);
      END_STATE();
    case 416:
      ACCEPT_TOKEN(anon_sym___try);
      END_STATE();
    case 417:
      if (lookahead == 'e') ADVANCE(560);
      END_STATE();
    case 418:
      if (lookahead == 'l') ADVANCE(561);
      END_STATE();
    case 419:
      if (lookahead == 'a') ADVANCE(562);
      END_STATE();
    case 420:
      if (lookahead == 's') ADVANCE(563);
      END_STATE();
    case 421:
      if (lookahead == 'r') ADVANCE(564);
      END_STATE();
    case 422:
      if (lookahead == 't') ADVANCE(565);
      END_STATE();
    case 423:
      if (lookahead == 'a') ADVANCE(566);
      END_STATE();
    case 424:
      if (lookahead == 'k') ADVANCE(567);
      END_STATE();
    case 425:
      if (lookahead == 'n') ADVANCE(568);
      END_STATE();
    case 426:
      if (lookahead == 'i') ADVANCE(569);
      END_STATE();
    case 427:
      if (lookahead == 'a') ADVANCE(570);
      if (lookahead == 'o') ADVANCE(571);
      END_STATE();
    case 428:
      if (lookahead == 'a') ADVANCE(572);
      END_STATE();
    case 429:
      ACCEPT_TOKEN(anon_sym_break);
      END_STATE();
    case 430:
      if (lookahead == 'y') ADVANCE(573);
      END_STATE();
    case 431:
      ACCEPT_TOKEN(anon_sym_byref);
      END_STATE();
    case 432:
      if (lookahead == '6') ADVANCE(574);
      END_STATE();
    case 433:
      if (lookahead == '2') ADVANCE(575);
      END_STATE();
    case 434:
      if (lookahead == '4') ADVANCE(576);
      END_STATE();
    case 435:
      if (lookahead == '_') ADVANCE(577);
      END_STATE();
    case 436:
      if (lookahead == 't') ADVANCE(578);
      END_STATE();
    case 437:
      ACCEPT_TOKEN(anon_sym_class);
      END_STATE();
    case 438:
      ACCEPT_TOKEN(anon_sym_const);
      if (lookahead == 'e') ADVANCE(579);
      END_STATE();
    case 439:
      if (lookahead == 'n') ADVANCE(580);
      END_STATE();
    case 440:
      if (lookahead == 'l') ADVANCE(581);
      END_STATE();
    case 441:
      if (lookahead == 'e') ADVANCE(582);
      END_STATE();
    case 442:
      if (lookahead == 'e') ADVANCE(288);
      END_STATE();
    case 443:
      if (lookahead == 'n') ADVANCE(583);
      END_STATE();
    case 444:
      if (lookahead == 'e') ADVANCE(584);
      END_STATE();
    case 445:
      ACCEPT_TOKEN(anon_sym_inout);
      END_STATE();
    case 446:
      if (lookahead == '_') ADVANCE(585);
      END_STATE();
    case 447:
      if (lookahead == '_') ADVANCE(586);
      END_STATE();
    case 448:
      if (lookahead == '_') ADVANCE(587);
      END_STATE();
    case 449:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 450:
      if (lookahead == 'r') ADVANCE(588);
      END_STATE();
    case 451:
      ACCEPT_TOKEN(anon_sym_macos);
      if (lookahead == 'x') ADVANCE(589);
      END_STATE();
    case 452:
      if (lookahead == 'l') ADVANCE(590);
      END_STATE();
    case 453:
      if (lookahead == 'u') ADVANCE(591);
      END_STATE();
    case 454:
      if (lookahead == 'o') ADVANCE(592);
      END_STATE();
    case 455:
      if (lookahead == 'b') ADVANCE(593);
      END_STATE();
    case 456:
      if (lookahead == 't') ADVANCE(594);
      END_STATE();
    case 457:
      if (lookahead == 'b') ADVANCE(595);
      END_STATE();
    case 458:
      if (lookahead == 't') ADVANCE(596);
      END_STATE();
    case 459:
      if (lookahead == 'y') ADVANCE(597);
      END_STATE();
    case 460:
      if (lookahead == 'f') ADVANCE(598);
      END_STATE();
    case 461:
      if (lookahead == 't') ADVANCE(599);
      END_STATE();
    case 462:
      if (lookahead == 'i') ADVANCE(600);
      END_STATE();
    case 463:
      if (lookahead == 'n') ADVANCE(601);
      END_STATE();
    case 464:
      ACCEPT_TOKEN(anon_sym_short);
      END_STATE();
    case 465:
      if (lookahead == 'd') ADVANCE(602);
      END_STATE();
    case 466:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 467:
      if (lookahead == 'f') ADVANCE(603);
      END_STATE();
    case 468:
      if (lookahead == '_') ADVANCE(604);
      END_STATE();
    case 469:
      if (lookahead == 'c') ADVANCE(605);
      END_STATE();
    case 470:
      if (lookahead == 't') ADVANCE(606);
      END_STATE();
    case 471:
      if (lookahead == 'h') ADVANCE(607);
      END_STATE();
    case 472:
      if (lookahead == 'd') ADVANCE(608);
      END_STATE();
    case 473:
      if (lookahead == 'e') ADVANCE(609);
      END_STATE();
    case 474:
      if (lookahead == 'f') ADVANCE(610);
      END_STATE();
    case 475:
      if (lookahead == '6') ADVANCE(611);
      END_STATE();
    case 476:
      if (lookahead == '2') ADVANCE(612);
      END_STATE();
    case 477:
      if (lookahead == '4') ADVANCE(613);
      END_STATE();
    case 478:
      if (lookahead == '_') ADVANCE(614);
      END_STATE();
    case 479:
      if (lookahead == 't') ADVANCE(615);
      END_STATE();
    case 480:
      ACCEPT_TOKEN(anon_sym_union);
      END_STATE();
    case 481:
      if (lookahead == 'n') ADVANCE(616);
      END_STATE();
    case 482:
      if (lookahead == 'g') ADVANCE(617);
      END_STATE();
    case 483:
      if (lookahead == 'i') ADVANCE(618);
      END_STATE();
    case 484:
      if (lookahead == 'o') ADVANCE(619);
      END_STATE();
    case 485:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 486:
      if (lookahead == 'A') ADVANCE(620);
      END_STATE();
    case 487:
      if (lookahead == 'P') ADVANCE(621);
      END_STATE();
    case 488:
      if (lookahead == 'A') ADVANCE(622);
      END_STATE();
    case 489:
      if (lookahead == 'M') ADVANCE(623);
      END_STATE();
    case 490:
      if (lookahead == 'U') ADVANCE(624);
      END_STATE();
    case 491:
      if (lookahead == 'E') ADVANCE(625);
      END_STATE();
    case 492:
      if (lookahead == 'I') ADVANCE(626);
      END_STATE();
    case 493:
      if (lookahead == 'A') ADVANCE(627);
      END_STATE();
    case 494:
      if (lookahead == 'T') ADVANCE(628);
      END_STATE();
    case 495:
      if (lookahead == 'e') ADVANCE(629);
      END_STATE();
    case 496:
      if (lookahead == 'e') ADVANCE(630);
      END_STATE();
    case 497:
      if (lookahead == 'I') ADVANCE(631);
      END_STATE();
    case 498:
      if (lookahead == 'O') ADVANCE(632);
      END_STATE();
    case 499:
      if (lookahead == 'I') ADVANCE(633);
      END_STATE();
    case 500:
      if (lookahead == 'S') ADVANCE(634);
      END_STATE();
    case 501:
      if (lookahead == 'R') ADVANCE(635);
      END_STATE();
    case 502:
      if (lookahead == 'M') ADVANCE(636);
      END_STATE();
    case 503:
      if (lookahead == 'E') ADVANCE(637);
      END_STATE();
    case 504:
      if (lookahead == 'M') ADVANCE(638);
      END_STATE();
    case 505:
      if (lookahead == 'I') ADVANCE(639);
      END_STATE();
    case 506:
      if (lookahead == 'U') ADVANCE(640);
      END_STATE();
    case 507:
      if (lookahead == 'T') ADVANCE(641);
      END_STATE();
    case 508:
      if (lookahead == 'F') ADVANCE(642);
      END_STATE();
    case 509:
      if (lookahead == 'V') ADVANCE(643);
      END_STATE();
    case 510:
      if (lookahead == 'I') ADVANCE(644);
      END_STATE();
    case 511:
      if (lookahead == 'X') ADVANCE(645);
      END_STATE();
    case 512:
      if (lookahead == 'O') ADVANCE(646);
      END_STATE();
    case 513:
      if (lookahead == 'E') ADVANCE(647);
      END_STATE();
    case 514:
      if (lookahead == 'E') ADVANCE(648);
      END_STATE();
    case 515:
      if (lookahead == 'L') ADVANCE(649);
      END_STATE();
    case 516:
      if (lookahead == 'a') ADVANCE(650);
      if (lookahead == 'o') ADVANCE(651);
      END_STATE();
    case 517:
      if (lookahead == 'c') ADVANCE(652);
      END_STATE();
    case 518:
      if (lookahead == 'e') ADVANCE(653);
      END_STATE();
    case 519:
      if (lookahead == 'i') ADVANCE(654);
      END_STATE();
    case 520:
      if (lookahead == 'l') ADVANCE(655);
      END_STATE();
    case 521:
      if (lookahead == 'u') ADVANCE(656);
      END_STATE();
    case 522:
      if (lookahead == 'u') ADVANCE(657);
      END_STATE();
    case 523:
      if (lookahead == 'b') ADVANCE(658);
      END_STATE();
    case 524:
      if (lookahead == 'A') ADVANCE(659);
      END_STATE();
    case 525:
      if (lookahead == 'A') ADVANCE(660);
      END_STATE();
    case 526:
      if (lookahead == 'n') ADVANCE(661);
      END_STATE();
    case 527:
      if (lookahead == '_') ADVANCE(662);
      END_STATE();
    case 528:
      if (lookahead == 'i') ADVANCE(663);
      END_STATE();
    case 529:
      if (lookahead == 'r') ADVANCE(664);
      END_STATE();
    case 530:
      if (lookahead == 'd') ADVANCE(665);
      END_STATE();
    case 531:
      if (lookahead == 'k') ADVANCE(666);
      END_STATE();
    case 532:
      if (lookahead == 'g') ADVANCE(667);
      END_STATE();
    case 533:
      if (lookahead == 't') ADVANCE(668);
      END_STATE();
    case 534:
      if (lookahead == 'h') ADVANCE(669);
      END_STATE();
    case 535:
      if (lookahead == 'l') ADVANCE(670);
      END_STATE();
    case 536:
      if (lookahead == 'a') ADVANCE(671);
      END_STATE();
    case 537:
      if (lookahead == 'l') ADVANCE(672);
      END_STATE();
    case 538:
      if (lookahead == 't') ADVANCE(673);
      END_STATE();
    case 539:
      if (lookahead == 'r') ADVANCE(674);
      END_STATE();
    case 540:
      if (lookahead == 'r') ADVANCE(675);
      END_STATE();
    case 541:
      if (lookahead == 's') ADVANCE(676);
      END_STATE();
    case 542:
      if (lookahead == 'e') ADVANCE(677);
      END_STATE();
    case 543:
      if (lookahead == 'n') ADVANCE(678);
      END_STATE();
    case 544:
      if (lookahead == 'c') ADVANCE(679);
      END_STATE();
    case 545:
      if (lookahead == 'l') ADVANCE(680);
      END_STATE();
    case 546:
      if (lookahead == 'e') ADVANCE(681);
      END_STATE();
    case 547:
      ACCEPT_TOKEN(anon_sym___imag);
      END_STATE();
    case 548:
      if (lookahead == 'n') ADVANCE(682);
      END_STATE();
    case 549:
      if (lookahead == 'o') ADVANCE(683);
      END_STATE();
    case 550:
      if (lookahead == 'u') ADVANCE(684);
      END_STATE();
    case 551:
      if (lookahead == 'a') ADVANCE(685);
      END_STATE();
    case 552:
      if (lookahead == 'u') ADVANCE(686);
      END_STATE();
    case 553:
      ACCEPT_TOKEN(anon_sym___real);
      END_STATE();
    case 554:
      if (lookahead == 'r') ADVANCE(687);
      END_STATE();
    case 555:
      ACCEPT_TOKEN(sym_ms_signed_ptr_modifier);
      END_STATE();
    case 556:
      if (lookahead == 'a') ADVANCE(688);
      END_STATE();
    case 557:
      if (lookahead == 'n') ADVANCE(689);
      END_STATE();
    case 558:
      if (lookahead == 'c') ADVANCE(690);
      END_STATE();
    case 559:
      if (lookahead == 'a') ADVANCE(691);
      END_STATE();
    case 560:
      if (lookahead == 'o') ADVANCE(692);
      END_STATE();
    case 561:
      if (lookahead == 'i') ADVANCE(693);
      END_STATE();
    case 562:
      if (lookahead == 'f') ADVANCE(694);
      END_STATE();
    case 563:
      if (lookahead == 'e') ADVANCE(695);
      END_STATE();
    case 564:
      ACCEPT_TOKEN(sym_ms_unsigned_ptr_modifier);
      END_STATE();
    case 565:
      if (lookahead == 'o') ADVANCE(696);
      END_STATE();
    case 566:
      if (lookahead == 't') ADVANCE(697);
      END_STATE();
    case 567:
      ACCEPT_TOKEN(anon_sym___weak);
      END_STATE();
    case 568:
      if (lookahead == 'o') ADVANCE(698);
      END_STATE();
    case 569:
      if (lookahead == 'g') ADVANCE(699);
      END_STATE();
    case 570:
      if (lookahead == 's') ADVANCE(700);
      END_STATE();
    case 571:
      if (lookahead == 'f') ADVANCE(701);
      END_STATE();
    case 572:
      if (lookahead == 'b') ADVANCE(702);
      END_STATE();
    case 573:
      ACCEPT_TOKEN(anon_sym_bycopy);
      END_STATE();
    case 574:
      if (lookahead == '_') ADVANCE(703);
      END_STATE();
    case 575:
      if (lookahead == '_') ADVANCE(704);
      END_STATE();
    case 576:
      if (lookahead == '_') ADVANCE(705);
      END_STATE();
    case 577:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 578:
      if (lookahead == 'r') ADVANCE(706);
      END_STATE();
    case 579:
      if (lookahead == 'x') ADVANCE(707);
      END_STATE();
    case 580:
      if (lookahead == 'u') ADVANCE(708);
      END_STATE();
    case 581:
      if (lookahead == 't') ADVANCE(709);
      END_STATE();
    case 582:
      if (lookahead == 'd') ADVANCE(710);
      END_STATE();
    case 583:
      ACCEPT_TOKEN(anon_sym_extern);
      END_STATE();
    case 584:
      ACCEPT_TOKEN(anon_sym_inline);
      END_STATE();
    case 585:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 586:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 587:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 588:
      if (lookahead == '_') ADVANCE(711);
      END_STATE();
    case 589:
      ACCEPT_TOKEN(anon_sym_macosx);
      END_STATE();
    case 590:
      if (lookahead == 'i') ADVANCE(712);
      END_STATE();
    case 591:
      if (lookahead == 'r') ADVANCE(713);
      END_STATE();
    case 592:
      if (lookahead == 'w') ADVANCE(714);
      END_STATE();
    case 593:
      if (lookahead == 'l') ADVANCE(715);
      END_STATE();
    case 594:
      if (lookahead == 'r') ADVANCE(716);
      END_STATE();
    case 595:
      if (lookahead == 'r') ADVANCE(717);
      END_STATE();
    case 596:
      if (lookahead == 'o') ADVANCE(718);
      END_STATE();
    case 597:
      ACCEPT_TOKEN(anon_sym_oneway);
      END_STATE();
    case 598:
      if (lookahead == 'f') ADVANCE(719);
      END_STATE();
    case 599:
      if (lookahead == 'e') ADVANCE(720);
      END_STATE();
    case 600:
      if (lookahead == 'c') ADVANCE(721);
      END_STATE();
    case 601:
      ACCEPT_TOKEN(anon_sym_return);
      END_STATE();
    case 602:
      ACCEPT_TOKEN(anon_sym_signed);
      END_STATE();
    case 603:
      ACCEPT_TOKEN(anon_sym_sizeof);
      END_STATE();
    case 604:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 605:
      ACCEPT_TOKEN(anon_sym_static);
      END_STATE();
    case 606:
      ACCEPT_TOKEN(anon_sym_struct);
      END_STATE();
    case 607:
      ACCEPT_TOKEN(anon_sym_switch);
      END_STATE();
    case 608:
      if (lookahead == '_') ADVANCE(722);
      END_STATE();
    case 609:
      if (lookahead == 'f') ADVANCE(723);
      END_STATE();
    case 610:
      ACCEPT_TOKEN(anon_sym_typeof);
      END_STATE();
    case 611:
      if (lookahead == '_') ADVANCE(724);
      END_STATE();
    case 612:
      if (lookahead == '_') ADVANCE(725);
      END_STATE();
    case 613:
      if (lookahead == '_') ADVANCE(726);
      END_STATE();
    case 614:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 615:
      if (lookahead == 'r') ADVANCE(727);
      END_STATE();
    case 616:
      if (lookahead == 'e') ADVANCE(728);
      END_STATE();
    case 617:
      ACCEPT_TOKEN(anon_sym_va_arg);
      END_STATE();
    case 618:
      if (lookahead == 'l') ADVANCE(729);
      END_STATE();
    case 619:
      if (lookahead == 's') ADVANCE(730);
      END_STATE();
    case 620:
      if (lookahead == 'I') ADVANCE(731);
      END_STATE();
    case 621:
      if (lookahead == 'R') ADVANCE(732);
      END_STATE();
    case 622:
      if (lookahead == 'V') ADVANCE(733);
      END_STATE();
    case 623:
      if (lookahead == 'A') ADVANCE(734);
      END_STATE();
    case 624:
      if (lookahead == 'R') ADVANCE(735);
      END_STATE();
    case 625:
      if (lookahead == 'R') ADVANCE(736);
      END_STATE();
    case 626:
      if (lookahead == 'N') ADVANCE(737);
      END_STATE();
    case 627:
      if (lookahead == 'T') ADVANCE(738);
      END_STATE();
    case 628:
      if (lookahead == 'I') ADVANCE(739);
      END_STATE();
    case 629:
      if (lookahead == 'c') ADVANCE(740);
      END_STATE();
    case 630:
      if (lookahead == 't') ADVANCE(741);
      END_STATE();
    case 631:
      if (lookahead == 'G') ADVANCE(742);
      END_STATE();
    case 632:
      if (lookahead == 'M') ADVANCE(743);
      END_STATE();
    case 633:
      if (lookahead == 'L') ADVANCE(744);
      END_STATE();
    case 634:
      if (lookahead == 'S') ADVANCE(745);
      END_STATE();
    case 635:
      if (lookahead == 'E') ADVANCE(746);
      END_STATE();
    case 636:
      if (lookahead == '_') ADVANCE(747);
      END_STATE();
    case 637:
      if (lookahead == 'N') ADVANCE(748);
      END_STATE();
    case 638:
      if (lookahead == 'A') ADVANCE(749);
      END_STATE();
    case 639:
      if (lookahead == 'N') ADVANCE(750);
      END_STATE();
    case 640:
      if (lookahead == 'I') ADVANCE(751);
      END_STATE();
    case 641:
      if (lookahead == '_') ADVANCE(752);
      END_STATE();
    case 642:
      if (lookahead == 'T') ADVANCE(753);
      END_STATE();
    case 643:
      if (lookahead == 'A') ADVANCE(754);
      END_STATE();
    case 644:
      if (lookahead == 'D') ADVANCE(755);
      END_STATE();
    case 645:
      if (lookahead == 'P') ADVANCE(756);
      END_STATE();
    case 646:
      if (lookahead == 'O') ADVANCE(757);
      END_STATE();
    case 647:
      if (lookahead == 'X') ADVANCE(758);
      END_STATE();
    case 648:
      if (lookahead == 'A') ADVANCE(759);
      END_STATE();
    case 649:
      if (lookahead == 'A') ADVANCE(760);
      END_STATE();
    case 650:
      if (lookahead == 's') ADVANCE(761);
      END_STATE();
    case 651:
      if (lookahead == 'f') ADVANCE(762);
      END_STATE();
    case 652:
      ACCEPT_TOKEN(anon_sym__Atomic);
      END_STATE();
    case 653:
      if (lookahead == 'x') ADVANCE(763);
      END_STATE();
    case 654:
      if (lookahead == 'c') ADVANCE(764);
      END_STATE();
    case 655:
      if (lookahead == 'l') ADVANCE(765);
      END_STATE();
    case 656:
      if (lookahead == 'r') ADVANCE(766);
      END_STATE();
    case 657:
      if (lookahead == 'n') ADVANCE(767);
      END_STATE();
    case 658:
      if (lookahead == 'l') ADVANCE(768);
      END_STATE();
    case 659:
      if (lookahead == 'V') ADVANCE(769);
      END_STATE();
    case 660:
      if (lookahead == 'V') ADVANCE(770);
      END_STATE();
    case 661:
      if (lookahead == 'o') ADVANCE(771);
      END_STATE();
    case 662:
      ACCEPT_TOKEN(anon_sym___asm__);
      END_STATE();
    case 663:
      if (lookahead == 'b') ADVANCE(772);
      END_STATE();
    case 664:
      if (lookahead == 'e') ADVANCE(773);
      END_STATE();
    case 665:
      ACCEPT_TOKEN(anon_sym___based);
      END_STATE();
    case 666:
      ACCEPT_TOKEN(anon_sym___block);
      END_STATE();
    case 667:
      if (lookahead == 'e') ADVANCE(774);
      END_STATE();
    case 668:
      if (lookahead == 'i') ADVANCE(775);
      END_STATE();
    case 669:
      ACCEPT_TOKEN(anon_sym___catch);
      END_STATE();
    case 670:
      ACCEPT_TOKEN(anon_sym___cdecl);
      END_STATE();
    case 671:
      if (lookahead == 'l') ADVANCE(776);
      END_STATE();
    case 672:
      if (lookahead == 'e') ADVANCE(777);
      END_STATE();
    case 673:
      ACCEPT_TOKEN(anon_sym___const);
      END_STATE();
    case 674:
      if (lookahead == 'a') ADVANCE(778);
      END_STATE();
    case 675:
      if (lookahead == 'i') ADVANCE(779);
      END_STATE();
    case 676:
      if (lookahead == 'p') ADVANCE(780);
      END_STATE();
    case 677:
      if (lookahead == 'c') ADVANCE(781);
      END_STATE();
    case 678:
      if (lookahead == 's') ADVANCE(782);
      END_STATE();
    case 679:
      if (lookahead == 'a') ADVANCE(783);
      END_STATE();
    case 680:
      if (lookahead == 'l') ADVANCE(784);
      END_STATE();
    case 681:
      if (lookahead == 'i') ADVANCE(785);
      END_STATE();
    case 682:
      if (lookahead == 'e') ADVANCE(786);
      END_STATE();
    case 683:
      if (lookahead == 'f') ADVANCE(787);
      END_STATE();
    case 684:
      if (lookahead == 'l') ADVANCE(788);
      END_STATE();
    case 685:
      if (lookahead == 'b') ADVANCE(789);
      END_STATE();
    case 686:
      if (lookahead == 't') ADVANCE(790);
      END_STATE();
    case 687:
      if (lookahead == 'i') ADVANCE(791);
      END_STATE();
    case 688:
      if (lookahead == 'l') ADVANCE(792);
      END_STATE();
    case 689:
      if (lookahead == 'g') ADVANCE(793);
      END_STATE();
    case 690:
      if (lookahead == 'a') ADVANCE(794);
      END_STATE();
    case 691:
      if (lookahead == 'd') ADVANCE(795);
      END_STATE();
    case 692:
      if (lookahead == 'f') ADVANCE(796);
      END_STATE();
    case 693:
      if (lookahead == 'g') ADVANCE(797);
      END_STATE();
    case 694:
      if (lookahead == 'e') ADVANCE(798);
      END_STATE();
    case 695:
      if (lookahead == 'd') ADVANCE(799);
      END_STATE();
    case 696:
      if (lookahead == 'r') ADVANCE(800);
      END_STATE();
    case 697:
      if (lookahead == 'i') ADVANCE(801);
      END_STATE();
    case 698:
      if (lookahead == 'f') ADVANCE(802);
      END_STATE();
    case 699:
      if (lookahead == 'n') ADVANCE(803);
      END_STATE();
    case 700:
      ACCEPT_TOKEN(anon_sym_alignas);
      END_STATE();
    case 701:
      ACCEPT_TOKEN(anon_sym_alignof);
      END_STATE();
    case 702:
      if (lookahead == 'i') ADVANCE(804);
      END_STATE();
    case 703:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 704:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 705:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 706:
      if (lookahead == '_') ADVANCE(805);
      END_STATE();
    case 707:
      if (lookahead == 'p') ADVANCE(806);
      END_STATE();
    case 708:
      if (lookahead == 'e') ADVANCE(807);
      END_STATE();
    case 709:
      ACCEPT_TOKEN(anon_sym_default);
      END_STATE();
    case 710:
      ACCEPT_TOKEN(anon_sym_defined);
      END_STATE();
    case 711:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 712:
      if (lookahead == 'g') ADVANCE(808);
      END_STATE();
    case 713:
      if (lookahead == 'n') ADVANCE(809);
      END_STATE();
    case 714:
      ACCEPT_TOKEN(anon_sym_nothrow);
      END_STATE();
    case 715:
      if (lookahead == 'e') ADVANCE(810);
      END_STATE();
    case 716:
      ACCEPT_TOKEN(anon_sym_nullptr);
      if (lookahead == '_') ADVANCE(811);
      END_STATE();
    case 717:
      if (lookahead == 'i') ADVANCE(812);
      END_STATE();
    case 718:
      if (lookahead == 'f') ADVANCE(813);
      END_STATE();
    case 719:
      if (lookahead == '_') ADVANCE(814);
      END_STATE();
    case 720:
      if (lookahead == 'r') ADVANCE(815);
      END_STATE();
    case 721:
      if (lookahead == 't') ADVANCE(816);
      END_STATE();
    case 722:
      if (lookahead == 'l') ADVANCE(817);
      END_STATE();
    case 723:
      ACCEPT_TOKEN(anon_sym_typedef);
      END_STATE();
    case 724:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 725:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 726:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 727:
      if (lookahead == '_') ADVANCE(818);
      END_STATE();
    case 728:
      if (lookahead == 'd') ADVANCE(819);
      END_STATE();
    case 729:
      if (lookahead == 'e') ADVANCE(820);
      END_STATE();
    case 730:
      ACCEPT_TOKEN(anon_sym_watchos);
      END_STATE();
    case 731:
      if (lookahead == 'L') ADVANCE(821);
      END_STATE();
    case 732:
      if (lookahead == 'E') ADVANCE(822);
      END_STATE();
    case 733:
      if (lookahead == 'A') ADVANCE(823);
      END_STATE();
    case 734:
      if (lookahead == 'T') ADVANCE(824);
      END_STATE();
    case 735:
      if (lookahead == 'N') ADVANCE(825);
      END_STATE();
    case 736:
      if (lookahead == 'N') ADVANCE(826);
      END_STATE();
    case 737:
      if (lookahead == 'E') ADVANCE(827);
      END_STATE();
    case 738:
      if (lookahead == 'E') ADVANCE(828);
      END_STATE();
    case 739:
      if (lookahead == 'O') ADVANCE(829);
      END_STATE();
    case 740:
      if (lookahead == 't') ADVANCE(830);
      END_STATE();
    case 741:
      ACCEPT_TOKEN(anon_sym_IBOutlet);
      END_STATE();
    case 742:
      if (lookahead == 'N') ADVANCE(831);
      END_STATE();
    case 743:
      if (lookahead == 'A') ADVANCE(832);
      END_STATE();
    case 744:
      if (lookahead == 'A') ADVANCE(833);
      END_STATE();
    case 745:
      if (lookahead == '_') ADVANCE(834);
      END_STATE();
    case 746:
      if (lookahead == 'C') ADVANCE(835);
      END_STATE();
    case 747:
      if (lookahead == 'A') ADVANCE(836);
      if (lookahead == 'D') ADVANCE(837);
      END_STATE();
    case 748:
      if (lookahead == 'S') ADVANCE(838);
      END_STATE();
    case 749:
      if (lookahead == 'T') ADVANCE(839);
      END_STATE();
    case 750:
      if (lookahead == 'E') ADVANCE(840);
      END_STATE();
    case 751:
      if (lookahead == 'R') ADVANCE(841);
      END_STATE();
    case 752:
      if (lookahead == 'C') ADVANCE(842);
      END_STATE();
    case 753:
      if (lookahead == '_') ADVANCE(843);
      END_STATE();
    case 754:
      if (lookahead == 'I') ADVANCE(844);
      END_STATE();
    case 755:
      if (lookahead == '_') ADVANCE(845);
      END_STATE();
    case 756:
      if (lookahead == 'O') ADVANCE(846);
      END_STATE();
    case 757:
      if (lookahead == 'T') ADVANCE(847);
      END_STATE();
    case 758:
      if (lookahead == 'T') ADVANCE(848);
      END_STATE();
    case 759:
      if (lookahead == 'R') ADVANCE(849);
      END_STATE();
    case 760:
      if (lookahead == 'B') ADVANCE(850);
      END_STATE();
    case 761:
      ACCEPT_TOKEN(anon_sym__Alignas);
      END_STATE();
    case 762:
      ACCEPT_TOKEN(anon_sym__Alignof);
      END_STATE();
    case 763:
      ACCEPT_TOKEN(anon_sym__Complex);
      END_STATE();
    case 764:
      ACCEPT_TOKEN(anon_sym__Generic);
      END_STATE();
    case 765:
      ACCEPT_TOKEN(anon_sym__Nonnull);
      END_STATE();
    case 766:
      if (lookahead == 'n') ADVANCE(851);
      END_STATE();
    case 767:
      if (lookahead == 's') ADVANCE(852);
      END_STATE();
    case 768:
      if (lookahead == 'e') ADVANCE(853);
      END_STATE();
    case 769:
      if (lookahead == 'A') ADVANCE(854);
      END_STATE();
    case 770:
      if (lookahead == 'A') ADVANCE(855);
      END_STATE();
    case 771:
      if (lookahead == 'f') ADVANCE(856);
      END_STATE();
    case 772:
      if (lookahead == 'u') ADVANCE(857);
      END_STATE();
    case 773:
      if (lookahead == 'l') ADVANCE(858);
      END_STATE();
    case 774:
      ACCEPT_TOKEN(anon_sym___bridge);
      if (lookahead == '_') ADVANCE(859);
      END_STATE();
    case 775:
      if (lookahead == 'n') ADVANCE(860);
      END_STATE();
    case 776:
      if (lookahead == 'l') ADVANCE(861);
      END_STATE();
    case 777:
      if (lookahead == 'x') ADVANCE(862);
      END_STATE();
    case 778:
      if (lookahead == 'v') ADVANCE(863);
      END_STATE();
    case 779:
      if (lookahead == 'a') ADVANCE(864);
      END_STATE();
    case 780:
      if (lookahead == 'e') ADVANCE(865);
      END_STATE();
    case 781:
      if (lookahead == 'a') ADVANCE(866);
      END_STATE();
    case 782:
      if (lookahead == 'i') ADVANCE(867);
      END_STATE();
    case 783:
      if (lookahead == 'l') ADVANCE(868);
      END_STATE();
    case 784:
      if (lookahead == 'y') ADVANCE(869);
      END_STATE();
    case 785:
      if (lookahead == 'n') ADVANCE(870);
      END_STATE();
    case 786:
      ACCEPT_TOKEN(anon_sym___inline);
      if (lookahead == '_') ADVANCE(871);
      END_STATE();
    case 787:
      ACCEPT_TOKEN(anon_sym___kindof);
      END_STATE();
    case 788:
      if (lookahead == 'l') ADVANCE(872);
      END_STATE();
    case 789:
      if (lookahead == 'l') ADVANCE(873);
      END_STATE();
    case 790:
      if (lookahead == 'h') ADVANCE(874);
      END_STATE();
    case 791:
      if (lookahead == 'c') ADVANCE(875);
      END_STATE();
    case 792:
      if (lookahead == 'l') ADVANCE(876);
      END_STATE();
    case 793:
      ACCEPT_TOKEN(anon_sym___strong);
      END_STATE();
    case 794:
      if (lookahead == 'l') ADVANCE(877);
      END_STATE();
    case 795:
      ACCEPT_TOKEN(anon_sym___thread);
      END_STATE();
    case 796:
      ACCEPT_TOKEN(anon_sym___typeof);
      if (lookahead == '_') ADVANCE(878);
      END_STATE();
    case 797:
      if (lookahead == 'n') ADVANCE(879);
      END_STATE();
    case 798:
      if (lookahead == '_') ADVANCE(880);
      END_STATE();
    case 799:
      ACCEPT_TOKEN(anon_sym___unused);
      END_STATE();
    case 800:
      if (lookahead == 'c') ADVANCE(881);
      END_STATE();
    case 801:
      if (lookahead == 'l') ADVANCE(882);
      END_STATE();
    case 802:
      ACCEPT_TOKEN(anon_sym__alignof);
      END_STATE();
    case 803:
      if (lookahead == 'e') ADVANCE(883);
      END_STATE();
    case 804:
      if (lookahead == 'l') ADVANCE(884);
      END_STATE();
    case 805:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 806:
      if (lookahead == 'r') ADVANCE(885);
      END_STATE();
    case 807:
      ACCEPT_TOKEN(anon_sym_continue);
      END_STATE();
    case 808:
      if (lookahead == 'n') ADVANCE(886);
      END_STATE();
    case 809:
      ACCEPT_TOKEN(anon_sym_noreturn);
      END_STATE();
    case 810:
      ACCEPT_TOKEN(anon_sym_nullable);
      END_STATE();
    case 811:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 812:
      if (lookahead == 'd') ADVANCE(887);
      END_STATE();
    case 813:
      ACCEPT_TOKEN(anon_sym_offsetof);
      END_STATE();
    case 814:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 815:
      ACCEPT_TOKEN(anon_sym_register);
      END_STATE();
    case 816:
      ACCEPT_TOKEN(anon_sym_restrict);
      END_STATE();
    case 817:
      if (lookahead == 'o') ADVANCE(888);
      END_STATE();
    case 818:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 819:
      ACCEPT_TOKEN(anon_sym_unsigned);
      END_STATE();
    case 820:
      ACCEPT_TOKEN(anon_sym_volatile);
      END_STATE();
    case 821:
      if (lookahead == 'A') ADVANCE(889);
      END_STATE();
    case 822:
      if (lookahead == 'C') ADVANCE(890);
      END_STATE();
    case 823:
      if (lookahead == 'I') ADVANCE(891);
      END_STATE();
    case 824:
      if (lookahead == '_') ADVANCE(892);
      END_STATE();
    case 825:
      if (lookahead == 'S') ADVANCE(893);
      END_STATE();
    case 826:
      ACCEPT_TOKEN(anon_sym_CG_EXTERN);
      END_STATE();
    case 827:
      ACCEPT_TOKEN(anon_sym_CG_INLINE);
      END_STATE();
    case 828:
      if (lookahead == 'D') ADVANCE(894);
      END_STATE();
    case 829:
      if (lookahead == 'N') ADVANCE(895);
      END_STATE();
    case 830:
      if (lookahead == 'a') ADVANCE(896);
      END_STATE();
    case 831:
      if (lookahead == 'A') ADVANCE(897);
      END_STATE();
    case 832:
      if (lookahead == 'T') ADVANCE(898);
      END_STATE();
    case 833:
      if (lookahead == 'B') ADVANCE(899);
      END_STATE();
    case 834:
      if (lookahead == 'A') ADVANCE(900);
      if (lookahead == 'D') ADVANCE(901);
      END_STATE();
    case 835:
      if (lookahead == 'A') ADVANCE(902);
      END_STATE();
    case 836:
      if (lookahead == 'V') ADVANCE(903);
      END_STATE();
    case 837:
      if (lookahead == 'E') ADVANCE(904);
      END_STATE();
    case 838:
      if (lookahead == 'I') ADVANCE(905);
      END_STATE();
    case 839:
      if (lookahead == '_') ADVANCE(906);
      END_STATE();
    case 840:
      ACCEPT_TOKEN(anon_sym_NS_INLINE);
      END_STATE();
    case 841:
      if (lookahead == 'E') ADVANCE(907);
      END_STATE();
    case 842:
      if (lookahead == 'L') ADVANCE(908);
      END_STATE();
    case 843:
      if (lookahead == 'N') ADVANCE(909);
      if (lookahead == 'U') ADVANCE(910);
      END_STATE();
    case 844:
      if (lookahead == 'L') ADVANCE(911);
      END_STATE();
    case 845:
      if (lookahead == 'U') ADVANCE(912);
      END_STATE();
    case 846:
      if (lookahead == 'R') ADVANCE(913);
      END_STATE();
    case 847:
      if (lookahead == '_') ADVANCE(914);
      END_STATE();
    case 848:
      if (lookahead == 'E') ADVANCE(915);
      END_STATE();
    case 849:
      if (lookahead == 'A') ADVANCE(916);
      END_STATE();
    case 850:
      if (lookahead == 'L') ADVANCE(917);
      END_STATE();
    case 851:
      ACCEPT_TOKEN(anon_sym__Noreturn);
      END_STATE();
    case 852:
      if (lookahead == 'p') ADVANCE(918);
      END_STATE();
    case 853:
      ACCEPT_TOKEN(anon_sym__Nullable);
      if (lookahead == '_') ADVANCE(919);
      END_STATE();
    case 854:
      if (lookahead == 'I') ADVANCE(920);
      END_STATE();
    case 855:
      if (lookahead == 'I') ADVANCE(921);
      END_STATE();
    case 856:
      ACCEPT_TOKEN(anon_sym___alignof);
      if (lookahead == '_') ADVANCE(922);
      END_STATE();
    case 857:
      if (lookahead == 't') ADVANCE(923);
      END_STATE();
    case 858:
      if (lookahead == 'e') ADVANCE(924);
      END_STATE();
    case 859:
      if (lookahead == 'r') ADVANCE(925);
      if (lookahead == 't') ADVANCE(926);
      END_STATE();
    case 860:
      if (lookahead == '_') ADVANCE(927);
      END_STATE();
    case 861:
      ACCEPT_TOKEN(anon_sym___clrcall);
      END_STATE();
    case 862:
      ACCEPT_TOKEN(anon_sym___complex);
      END_STATE();
    case 863:
      if (lookahead == 'a') ADVANCE(928);
      END_STATE();
    case 864:
      if (lookahead == 'n') ADVANCE(929);
      END_STATE();
    case 865:
      if (lookahead == 'c') ADVANCE(930);
      END_STATE();
    case 866:
      if (lookahead == 't') ADVANCE(931);
      END_STATE();
    case 867:
      if (lookahead == 'o') ADVANCE(932);
      END_STATE();
    case 868:
      if (lookahead == 'l') ADVANCE(933);
      END_STATE();
    case 869:
      ACCEPT_TOKEN(anon_sym___finally);
      END_STATE();
    case 870:
      if (lookahead == 'l') ADVANCE(934);
      END_STATE();
    case 871:
      if (lookahead == '_') ADVANCE(935);
      END_STATE();
    case 872:
      ACCEPT_TOKEN(anon_sym___nonnull);
      END_STATE();
    case 873:
      if (lookahead == 'e') ADVANCE(936);
      END_STATE();
    case 874:
      if (lookahead == '_') ADVANCE(937);
      END_STATE();
    case 875:
      if (lookahead == 't') ADVANCE(938);
      END_STATE();
    case 876:
      ACCEPT_TOKEN(anon_sym___stdcall);
      END_STATE();
    case 877:
      if (lookahead == 'l') ADVANCE(939);
      END_STATE();
    case 878:
      if (lookahead == '_') ADVANCE(940);
      END_STATE();
    case 879:
      if (lookahead == 'e') ADVANCE(941);
      END_STATE();
    case 880:
      if (lookahead == 'u') ADVANCE(942);
      END_STATE();
    case 881:
      if (lookahead == 'a') ADVANCE(943);
      END_STATE();
    case 882:
      if (lookahead == 'e') ADVANCE(944);
      END_STATE();
    case 883:
      if (lookahead == 'd') ADVANCE(945);
      END_STATE();
    case 884:
      if (lookahead == 'i') ADVANCE(946);
      END_STATE();
    case 885:
      ACCEPT_TOKEN(anon_sym_constexpr);
      END_STATE();
    case 886:
      if (lookahead == '_') ADVANCE(947);
      END_STATE();
    case 887:
      if (lookahead == 'g') ADVANCE(948);
      END_STATE();
    case 888:
      if (lookahead == 'c') ADVANCE(949);
      END_STATE();
    case 889:
      if (lookahead == 'B') ADVANCE(950);
      END_STATE();
    case 890:
      if (lookahead == 'A') ADVANCE(951);
      END_STATE();
    case 891:
      if (lookahead == 'L') ADVANCE(952);
      END_STATE();
    case 892:
      if (lookahead == 'F') ADVANCE(953);
      END_STATE();
    case 893:
      if (lookahead == '_') ADVANCE(954);
      END_STATE();
    case 894:
      if (lookahead == '_') ADVANCE(955);
      END_STATE();
    case 895:
      if (lookahead == '_') ADVANCE(956);
      END_STATE();
    case 896:
      if (lookahead == 'b') ADVANCE(957);
      END_STATE();
    case 897:
      if (lookahead == 'B') ADVANCE(958);
      END_STATE();
    case 898:
      if (lookahead == 'E') ADVANCE(959);
      END_STATE();
    case 899:
      if (lookahead == 'L') ADVANCE(960);
      END_STATE();
    case 900:
      if (lookahead == 'V') ADVANCE(961);
      END_STATE();
    case 901:
      if (lookahead == 'E') ADVANCE(962);
      END_STATE();
    case 902:
      if (lookahead == 'T') ADVANCE(963);
      END_STATE();
    case 903:
      if (lookahead == 'A') ADVANCE(964);
      END_STATE();
    case 904:
      if (lookahead == 'P') ADVANCE(965);
      END_STATE();
    case 905:
      if (lookahead == 'O') ADVANCE(966);
      END_STATE();
    case 906:
      if (lookahead == 'F') ADVANCE(967);
      END_STATE();
    case 907:
      if (lookahead == 'S') ADVANCE(968);
      END_STATE();
    case 908:
      if (lookahead == 'A') ADVANCE(969);
      END_STATE();
    case 909:
      if (lookahead == 'A') ADVANCE(970);
      END_STATE();
    case 910:
      if (lookahead == 'N') ADVANCE(971);
      END_STATE();
    case 911:
      if (lookahead == 'A') ADVANCE(972);
      END_STATE();
    case 912:
      if (lookahead == 'N') ADVANCE(973);
      END_STATE();
    case 913:
      if (lookahead == 'T') ADVANCE(974);
      END_STATE();
    case 914:
      if (lookahead == 'C') ADVANCE(975);
      END_STATE();
    case 915:
      if (lookahead == 'R') ADVANCE(976);
      END_STATE();
    case 916:
      if (lookahead == 'N') ADVANCE(977);
      END_STATE();
    case 917:
      if (lookahead == 'E') ADVANCE(978);
      END_STATE();
    case 918:
      if (lookahead == 'e') ADVANCE(979);
      END_STATE();
    case 919:
      if (lookahead == 'r') ADVANCE(980);
      END_STATE();
    case 920:
      if (lookahead == 'L') ADVANCE(981);
      END_STATE();
    case 921:
      if (lookahead == 'L') ADVANCE(982);
      END_STATE();
    case 922:
      if (lookahead == '_') ADVANCE(983);
      END_STATE();
    case 923:
      if (lookahead == 'e') ADVANCE(984);
      END_STATE();
    case 924:
      if (lookahead == 'a') ADVANCE(985);
      END_STATE();
    case 925:
      if (lookahead == 'e') ADVANCE(986);
      END_STATE();
    case 926:
      if (lookahead == 'r') ADVANCE(987);
      END_STATE();
    case 927:
      if (lookahead == 'a') ADVANCE(988);
      END_STATE();
    case 928:
      if (lookahead == 'r') ADVANCE(989);
      END_STATE();
    case 929:
      if (lookahead == 't') ADVANCE(990);
      END_STATE();
    case 930:
      ACCEPT_TOKEN(anon_sym___declspec);
      END_STATE();
    case 931:
      if (lookahead == 'e') ADVANCE(991);
      END_STATE();
    case 932:
      if (lookahead == 'n') ADVANCE(992);
      END_STATE();
    case 933:
      ACCEPT_TOKEN(anon_sym___fastcall);
      END_STATE();
    case 934:
      if (lookahead == 'i') ADVANCE(993);
      END_STATE();
    case 935:
      ACCEPT_TOKEN(anon_sym___inline__);
      END_STATE();
    case 936:
      ACCEPT_TOKEN(anon_sym___nullable);
      END_STATE();
    case 937:
      if (lookahead == 'o') ADVANCE(994);
      END_STATE();
    case 938:
      ACCEPT_TOKEN(sym_ms_restrict_modifier);
      if (lookahead == '_') ADVANCE(995);
      END_STATE();
    case 939:
      ACCEPT_TOKEN(anon_sym___thiscall);
      END_STATE();
    case 940:
      ACCEPT_TOKEN(anon_sym___typeof__);
      END_STATE();
    case 941:
      if (lookahead == 'd') ADVANCE(996);
      END_STATE();
    case 942:
      if (lookahead == 'n') ADVANCE(997);
      END_STATE();
    case 943:
      if (lookahead == 'l') ADVANCE(998);
      END_STATE();
    case 944:
      if (lookahead == '_') ADVANCE(999);
      END_STATE();
    case 945:
      ACCEPT_TOKEN(anon_sym__unaligned);
      END_STATE();
    case 946:
      if (lookahead == 't') ADVANCE(1000);
      END_STATE();
    case 947:
      if (lookahead == 't') ADVANCE(288);
      END_STATE();
    case 948:
      if (lookahead == 'e') ADVANCE(1001);
      END_STATE();
    case 949:
      if (lookahead == 'a') ADVANCE(1002);
      END_STATE();
    case 950:
      if (lookahead == 'L') ADVANCE(1003);
      END_STATE();
    case 951:
      if (lookahead == 'T') ADVANCE(1004);
      END_STATE();
    case 952:
      if (lookahead == 'A') ADVANCE(1005);
      END_STATE();
    case 953:
      if (lookahead == 'U') ADVANCE(1006);
      END_STATE();
    case 954:
      if (lookahead == 'N') ADVANCE(1007);
      if (lookahead == 'R') ADVANCE(1008);
      END_STATE();
    case 955:
      if (lookahead == 'A') ADVANCE(1009);
      if (lookahead == 'M') ADVANCE(1010);
      END_STATE();
    case 956:
      if (lookahead == 'E') ADVANCE(1011);
      if (lookahead == 'S') ADVANCE(1012);
      END_STATE();
    case 957:
      if (lookahead == 'l') ADVANCE(1013);
      END_STATE();
    case 958:
      if (lookahead == 'L') ADVANCE(1014);
      END_STATE();
    case 959:
      if (lookahead == 'D') ADVANCE(1015);
      END_STATE();
    case 960:
      if (lookahead == 'E') ADVANCE(1016);
      END_STATE();
    case 961:
      if (lookahead == 'A') ADVANCE(1017);
      END_STATE();
    case 962:
      if (lookahead == 'P') ADVANCE(1018);
      END_STATE();
    case 963:
      if (lookahead == 'E') ADVANCE(1019);
      END_STATE();
    case 964:
      if (lookahead == 'I') ADVANCE(1020);
      END_STATE();
    case 965:
      if (lookahead == 'R') ADVANCE(1021);
      END_STATE();
    case 966:
      if (lookahead == 'N') ADVANCE(1022);
      END_STATE();
    case 967:
      if (lookahead == 'U') ADVANCE(1023);
      END_STATE();
    case 968:
      if (lookahead == '_') ADVANCE(1024);
      END_STATE();
    case 969:
      if (lookahead == 'S') ADVANCE(1025);
      END_STATE();
    case 970:
      if (lookahead == 'M') ADVANCE(1026);
      END_STATE();
    case 971:
      if (lookahead == 'A') ADVANCE(1027);
      END_STATE();
    case 972:
      if (lookahead == 'B') ADVANCE(1028);
      END_STATE();
    case 973:
      if (lookahead == 'T') ADVANCE(1029);
      END_STATE();
    case 974:
      ACCEPT_TOKEN(anon_sym_OBJC_EXPORT);
      END_STATE();
    case 975:
      if (lookahead == 'L') ADVANCE(1030);
      END_STATE();
    case 976:
      if (lookahead == 'N') ADVANCE(1031);
      END_STATE();
    case 977:
      if (lookahead == 'C') ADVANCE(1032);
      END_STATE();
    case 978:
      if (lookahead == '_') ADVANCE(1033);
      END_STATE();
    case 979:
      if (lookahead == 'c') ADVANCE(1034);
      END_STATE();
    case 980:
      if (lookahead == 'e') ADVANCE(1035);
      END_STATE();
    case 981:
      if (lookahead == 'A') ADVANCE(1036);
      END_STATE();
    case 982:
      if (lookahead == 'A') ADVANCE(1037);
      END_STATE();
    case 983:
      ACCEPT_TOKEN(anon_sym___alignof__);
      END_STATE();
    case 984:
      ACCEPT_TOKEN(anon_sym___attribute);
      if (lookahead == '_') ADVANCE(1038);
      END_STATE();
    case 985:
      if (lookahead == 's') ADVANCE(1039);
      END_STATE();
    case 986:
      if (lookahead == 't') ADVANCE(1040);
      END_STATE();
    case 987:
      if (lookahead == 'a') ADVANCE(1041);
      END_STATE();
    case 988:
      if (lookahead == 'v') ADVANCE(1042);
      END_STATE();
    case 989:
      if (lookahead == 'i') ADVANCE(1043);
      END_STATE();
    case 990:
      ACCEPT_TOKEN(anon_sym___covariant);
      END_STATE();
    case 991:
      if (lookahead == 'd') ADVANCE(1044);
      END_STATE();
    case 992:
      if (lookahead == '_') ADVANCE(1045);
      END_STATE();
    case 993:
      if (lookahead == 'n') ADVANCE(1046);
      END_STATE();
    case 994:
      if (lookahead == 'b') ADVANCE(1047);
      END_STATE();
    case 995:
      if (lookahead == '_') ADVANCE(1048);
      END_STATE();
    case 996:
      ACCEPT_TOKEN(anon_sym___unaligned);
      END_STATE();
    case 997:
      if (lookahead == 'r') ADVANCE(1049);
      END_STATE();
    case 998:
      if (lookahead == 'l') ADVANCE(1050);
      END_STATE();
    case 999:
      if (lookahead == '_') ADVANCE(1051);
      END_STATE();
    case 1000:
      if (lookahead == 'y') ADVANCE(1052);
      END_STATE();
    case 1001:
      if (lookahead == '_') ADVANCE(1053);
      END_STATE();
    case 1002:
      if (lookahead == 'l') ADVANCE(1054);
      END_STATE();
    case 1003:
      if (lookahead == 'E') ADVANCE(1055);
      END_STATE();
    case 1004:
      if (lookahead == 'E') ADVANCE(1056);
      END_STATE();
    case 1005:
      if (lookahead == 'B') ADVANCE(1057);
      END_STATE();
    case 1006:
      if (lookahead == 'N') ADVANCE(1058);
      END_STATE();
    case 1007:
      if (lookahead == 'O') ADVANCE(1059);
      END_STATE();
    case 1008:
      if (lookahead == 'E') ADVANCE(1060);
      END_STATE();
    case 1009:
      if (lookahead == 'T') ADVANCE(1061);
      END_STATE();
    case 1010:
      if (lookahead == 'S') ADVANCE(1062);
      END_STATE();
    case 1011:
      if (lookahead == 'X') ADVANCE(1063);
      END_STATE();
    case 1012:
      if (lookahead == 'T') ADVANCE(1064);
      END_STATE();
    case 1013:
      if (lookahead == 'e') ADVANCE(1065);
      END_STATE();
    case 1014:
      if (lookahead == 'E') ADVANCE(1066);
      END_STATE();
    case 1015:
      if (lookahead == '_') ADVANCE(1067);
      END_STATE();
    case 1016:
      ACCEPT_TOKEN(anon_sym_NS_AVAILABLE);
      if (lookahead == '_') ADVANCE(1068);
      END_STATE();
    case 1017:
      if (lookahead == 'I') ADVANCE(1069);
      END_STATE();
    case 1018:
      if (lookahead == 'R') ADVANCE(1070);
      END_STATE();
    case 1019:
      if (lookahead == 'D') ADVANCE(1071);
      END_STATE();
    case 1020:
      if (lookahead == 'L') ADVANCE(1072);
      END_STATE();
    case 1021:
      if (lookahead == 'E') ADVANCE(1073);
      END_STATE();
    case 1022:
      if (lookahead == '_') ADVANCE(1074);
      END_STATE();
    case 1023:
      if (lookahead == 'N') ADVANCE(1075);
      END_STATE();
    case 1024:
      if (lookahead == 'N') ADVANCE(1076);
      END_STATE();
    case 1025:
      if (lookahead == 'S') ADVANCE(1077);
      END_STATE();
    case 1026:
      if (lookahead == 'E') ADVANCE(1078);
      END_STATE();
    case 1027:
      if (lookahead == 'V') ADVANCE(1079);
      END_STATE();
    case 1028:
      if (lookahead == 'L') ADVANCE(1080);
      END_STATE();
    case 1029:
      if (lookahead == 'I') ADVANCE(1081);
      END_STATE();
    case 1030:
      if (lookahead == 'A') ADVANCE(1082);
      END_STATE();
    case 1031:
      ACCEPT_TOKEN(anon_sym_UIKIT_EXTERN);
      END_STATE();
    case 1032:
      if (lookahead == 'E') ADVANCE(1083);
      END_STATE();
    case 1033:
      if (lookahead == 'A') ADVANCE(1084);
      END_STATE();
    case 1034:
      if (lookahead == 'i') ADVANCE(1085);
      END_STATE();
    case 1035:
      if (lookahead == 's') ADVANCE(1086);
      END_STATE();
    case 1036:
      if (lookahead == 'B') ADVANCE(1087);
      END_STATE();
    case 1037:
      if (lookahead == 'B') ADVANCE(1088);
      END_STATE();
    case 1038:
      if (lookahead == '_') ADVANCE(1089);
      END_STATE();
    case 1039:
      if (lookahead == 'i') ADVANCE(1090);
      END_STATE();
    case 1040:
      if (lookahead == 'a') ADVANCE(1091);
      END_STATE();
    case 1041:
      if (lookahead == 'n') ADVANCE(1092);
      END_STATE();
    case 1042:
      if (lookahead == 'a') ADVANCE(1093);
      END_STATE();
    case 1043:
      if (lookahead == 'a') ADVANCE(1094);
      END_STATE();
    case 1044:
      if (lookahead == '_') ADVANCE(1095);
      END_STATE();
    case 1045:
      if (lookahead == '_') ADVANCE(1096);
      END_STATE();
    case 1046:
      if (lookahead == 'e') ADVANCE(1097);
      END_STATE();
    case 1047:
      if (lookahead == 'j') ADVANCE(1098);
      END_STATE();
    case 1048:
      ACCEPT_TOKEN(anon_sym___restrict__);
      END_STATE();
    case 1049:
      if (lookahead == 'e') ADVANCE(1099);
      END_STATE();
    case 1050:
      ACCEPT_TOKEN(anon_sym___vectorcall);
      END_STATE();
    case 1051:
      ACCEPT_TOKEN(anon_sym___volatile__);
      END_STATE();
    case 1052:
      ACCEPT_TOKEN(anon_sym_availability);
      END_STATE();
    case 1053:
      if (lookahead == 'r') ADVANCE(1100);
      END_STATE();
    case 1054:
      ACCEPT_TOKEN(anon_sym_thread_local);
      END_STATE();
    case 1055:
      ACCEPT_TOKEN(anon_sym_API_AVAILABLE);
      END_STATE();
    case 1056:
      if (lookahead == 'D') ADVANCE(1101);
      END_STATE();
    case 1057:
      if (lookahead == 'L') ADVANCE(1102);
      END_STATE();
    case 1058:
      if (lookahead == 'C') ADVANCE(1103);
      END_STATE();
    case 1059:
      if (lookahead == 'T') ADVANCE(1104);
      END_STATE();
    case 1060:
      if (lookahead == 'T') ADVANCE(1105);
      END_STATE();
    case 1061:
      if (lookahead == 'T') ADVANCE(1106);
      END_STATE();
    case 1062:
      if (lookahead == 'G') ADVANCE(1107);
      END_STATE();
    case 1063:
      if (lookahead == 'P') ADVANCE(1108);
      if (lookahead == 'T') ADVANCE(1109);
      END_STATE();
    case 1064:
      if (lookahead == 'A') ADVANCE(1110);
      END_STATE();
    case 1065:
      ACCEPT_TOKEN(anon_sym_IBInspectable);
      END_STATE();
    case 1066:
      ACCEPT_TOKEN(anon_sym_IB_DESIGNABLE);
      END_STATE();
    case 1067:
      if (lookahead == 'R') ADVANCE(1111);
      END_STATE();
    case 1068:
      if (lookahead == 'I') ADVANCE(1112);
      END_STATE();
    case 1069:
      if (lookahead == 'L') ADVANCE(1113);
      END_STATE();
    case 1070:
      if (lookahead == 'E') ADVANCE(1114);
      END_STATE();
    case 1071:
      if (lookahead == '_') ADVANCE(1115);
      END_STATE();
    case 1072:
      if (lookahead == 'A') ADVANCE(1116);
      END_STATE();
    case 1073:
      if (lookahead == 'C') ADVANCE(1117);
      END_STATE();
    case 1074:
      if (lookahead == 'U') ADVANCE(1118);
      END_STATE();
    case 1075:
      if (lookahead == 'C') ADVANCE(1119);
      END_STATE();
    case 1076:
      if (lookahead == 'I') ADVANCE(1120);
      END_STATE();
    case 1077:
      ACCEPT_TOKEN(anon_sym_NS_ROOT_CLASS);
      END_STATE();
    case 1078:
      ACCEPT_TOKEN(anon_sym_NS_SWIFT_NAME);
      END_STATE();
    case 1079:
      if (lookahead == 'A') ADVANCE(1121);
      END_STATE();
    case 1080:
      if (lookahead == 'E') ADVANCE(1122);
      END_STATE();
    case 1081:
      if (lookahead == 'L') ADVANCE(1123);
      END_STATE();
    case 1082:
      if (lookahead == 'S') ADVANCE(1124);
      END_STATE();
    case 1083:
      if (lookahead == '_') ADVANCE(1125);
      END_STATE();
    case 1084:
      if (lookahead == 'T') ADVANCE(1126);
      END_STATE();
    case 1085:
      if (lookahead == 'f') ADVANCE(1127);
      END_STATE();
    case 1086:
      if (lookahead == 'u') ADVANCE(1128);
      END_STATE();
    case 1087:
      if (lookahead == 'L') ADVANCE(1129);
      END_STATE();
    case 1088:
      if (lookahead == 'L') ADVANCE(1130);
      END_STATE();
    case 1089:
      ACCEPT_TOKEN(anon_sym___attribute__);
      END_STATE();
    case 1090:
      if (lookahead == 'n') ADVANCE(1131);
      END_STATE();
    case 1091:
      if (lookahead == 'i') ADVANCE(1132);
      END_STATE();
    case 1092:
      if (lookahead == 's') ADVANCE(1133);
      END_STATE();
    case 1093:
      if (lookahead == 'i') ADVANCE(1134);
      END_STATE();
    case 1094:
      if (lookahead == 'n') ADVANCE(1135);
      END_STATE();
    case 1095:
      if (lookahead == 'e') ADVANCE(1136);
      if (lookahead == 'm') ADVANCE(1137);
      END_STATE();
    case 1096:
      ACCEPT_TOKEN(anon_sym___extension__);
      END_STATE();
    case 1097:
      ACCEPT_TOKEN(anon_sym___forceinline);
      END_STATE();
    case 1098:
      if (lookahead == 'c') ADVANCE(1138);
      END_STATE();
    case 1099:
      if (lookahead == 't') ADVANCE(1139);
      END_STATE();
    case 1100:
      if (lookahead == 'e') ADVANCE(1140);
      END_STATE();
    case 1101:
      ACCEPT_TOKEN(anon_sym_API_DEPRECATED);
      END_STATE();
    case 1102:
      if (lookahead == 'E') ADVANCE(1141);
      END_STATE();
    case 1103:
      if (lookahead == 'T') ADVANCE(1142);
      END_STATE();
    case 1104:
      if (lookahead == '_') ADVANCE(1143);
      END_STATE();
    case 1105:
      if (lookahead == 'A') ADVANCE(1144);
      END_STATE();
    case 1106:
      if (lookahead == 'R') ADVANCE(1145);
      END_STATE();
    case 1107:
      if (lookahead == '_') ADVANCE(1146);
      END_STATE();
    case 1108:
      if (lookahead == 'O') ADVANCE(1147);
      END_STATE();
    case 1109:
      if (lookahead == 'E') ADVANCE(1148);
      END_STATE();
    case 1110:
      if (lookahead == 'T') ADVANCE(1149);
      END_STATE();
    case 1111:
      if (lookahead == 'E') ADVANCE(1150);
      END_STATE();
    case 1112:
      if (lookahead == 'O') ADVANCE(1151);
      END_STATE();
    case 1113:
      if (lookahead == 'A') ADVANCE(1152);
      END_STATE();
    case 1114:
      if (lookahead == 'C') ADVANCE(1153);
      END_STATE();
    case 1115:
      if (lookahead == 'I') ADVANCE(1154);
      END_STATE();
    case 1116:
      if (lookahead == 'B') ADVANCE(1155);
      END_STATE();
    case 1117:
      if (lookahead == 'A') ADVANCE(1156);
      END_STATE();
    case 1118:
      if (lookahead == 'N') ADVANCE(1157);
      END_STATE();
    case 1119:
      if (lookahead == 'T') ADVANCE(1158);
      END_STATE();
    case 1120:
      if (lookahead == 'L') ADVANCE(1159);
      END_STATE();
    case 1121:
      if (lookahead == 'I') ADVANCE(1160);
      END_STATE();
    case 1122:
      ACCEPT_TOKEN(anon_sym_NS_UNAVAILABLE);
      END_STATE();
    case 1123:
      if (lookahead == '_') ADVANCE(1161);
      END_STATE();
    case 1124:
      if (lookahead == 'S') ADVANCE(1162);
      END_STATE();
    case 1125:
      if (lookahead == 'S') ADVANCE(1163);
      END_STATE();
    case 1126:
      if (lookahead == 'T') ADVANCE(1164);
      END_STATE();
    case 1127:
      if (lookahead == 'i') ADVANCE(1165);
      END_STATE();
    case 1128:
      if (lookahead == 'l') ADVANCE(1166);
      END_STATE();
    case 1129:
      if (lookahead == 'E') ADVANCE(1167);
      END_STATE();
    case 1130:
      if (lookahead == 'E') ADVANCE(1168);
      END_STATE();
    case 1131:
      if (lookahead == 'g') ADVANCE(1169);
      END_STATE();
    case 1132:
      if (lookahead == 'n') ADVANCE(1170);
      END_STATE();
    case 1133:
      if (lookahead == 'f') ADVANCE(1171);
      END_STATE();
    case 1134:
      if (lookahead == 'l') ADVANCE(1172);
      END_STATE();
    case 1135:
      if (lookahead == 't') ADVANCE(1173);
      END_STATE();
    case 1136:
      if (lookahead == 'n') ADVANCE(1174);
      END_STATE();
    case 1137:
      if (lookahead == 's') ADVANCE(1175);
      END_STATE();
    case 1138:
      if (lookahead == '_') ADVANCE(1176);
      END_STATE();
    case 1139:
      if (lookahead == 'a') ADVANCE(1177);
      END_STATE();
    case 1140:
      if (lookahead == 'l') ADVANCE(1178);
      END_STATE();
    case 1141:
      ACCEPT_TOKEN(anon_sym_API_UNAVAILABLE);
      END_STATE();
    case 1142:
      if (lookahead == 'I') ADVANCE(1179);
      END_STATE();
    case 1143:
      if (lookahead == 'R') ADVANCE(1180);
      END_STATE();
    case 1144:
      if (lookahead == 'I') ADVANCE(1181);
      END_STATE();
    case 1145:
      if (lookahead == 'I') ADVANCE(1182);
      END_STATE();
    case 1146:
      if (lookahead == 'A') ADVANCE(1183);
      END_STATE();
    case 1147:
      if (lookahead == 'R') ADVANCE(1184);
      END_STATE();
    case 1148:
      if (lookahead == 'R') ADVANCE(1185);
      END_STATE();
    case 1149:
      if (lookahead == 'I') ADVANCE(1186);
      END_STATE();
    case 1150:
      if (lookahead == 'F') ADVANCE(1187);
      END_STATE();
    case 1151:
      if (lookahead == 'S') ADVANCE(1188);
      END_STATE();
    case 1152:
      if (lookahead == 'B') ADVANCE(1189);
      END_STATE();
    case 1153:
      if (lookahead == 'A') ADVANCE(1190);
      END_STATE();
    case 1154:
      if (lookahead == 'O') ADVANCE(1191);
      END_STATE();
    case 1155:
      if (lookahead == 'L') ADVANCE(1192);
      END_STATE();
    case 1156:
      if (lookahead == 'T') ADVANCE(1193);
      END_STATE();
    case 1157:
      if (lookahead == 'A') ADVANCE(1194);
      END_STATE();
    case 1158:
      if (lookahead == 'I') ADVANCE(1195);
      END_STATE();
    case 1159:
      if (lookahead == '_') ADVANCE(1196);
      END_STATE();
    case 1160:
      if (lookahead == 'L') ADVANCE(1197);
      END_STATE();
    case 1161:
      if (lookahead == 'E') ADVANCE(1198);
      END_STATE();
    case 1162:
      ACCEPT_TOKEN(anon_sym_OBJC_ROOT_CLASS);
      END_STATE();
    case 1163:
      if (lookahead == 'E') ADVANCE(1199);
      END_STATE();
    case 1164:
      if (lookahead == 'R') ADVANCE(1200);
      END_STATE();
    case 1165:
      if (lookahead == 'e') ADVANCE(1201);
      END_STATE();
    case 1166:
      if (lookahead == 't') ADVANCE(1202);
      END_STATE();
    case 1167:
      ACCEPT_TOKEN(anon_sym___IOS_AVAILABLE);
      END_STATE();
    case 1168:
      if (lookahead == '_') ADVANCE(1203);
      END_STATE();
    case 1169:
      ACCEPT_TOKEN(anon_sym___autoreleasing);
      END_STATE();
    case 1170:
      if (lookahead == 'e') ADVANCE(1204);
      END_STATE();
    case 1171:
      if (lookahead == 'e') ADVANCE(1205);
      END_STATE();
    case 1172:
      if (lookahead == 'a') ADVANCE(1206);
      END_STATE();
    case 1173:
      ACCEPT_TOKEN(anon_sym___contravariant);
      END_STATE();
    case 1174:
      if (lookahead == 'u') ADVANCE(1207);
      END_STATE();
    case 1175:
      if (lookahead == 'g') ADVANCE(1208);
      END_STATE();
    case 1176:
      if (lookahead == 'c') ADVANCE(1209);
      if (lookahead == 'i') ADVANCE(1210);
      if (lookahead == 's') ADVANCE(1211);
      END_STATE();
    case 1177:
      if (lookahead == 'i') ADVANCE(1212);
      END_STATE();
    case 1178:
      if (lookahead == 'a') ADVANCE(1213);
      END_STATE();
    case 1179:
      if (lookahead == 'O') ADVANCE(1214);
      END_STATE();
    case 1180:
      if (lookahead == 'E') ADVANCE(1215);
      END_STATE();
    case 1181:
      if (lookahead == 'N') ADVANCE(1216);
      END_STATE();
    case 1182:
      if (lookahead == 'B') ADVANCE(1217);
      END_STATE();
    case 1183:
      if (lookahead == 'T') ADVANCE(1218);
      END_STATE();
    case 1184:
      if (lookahead == 'T') ADVANCE(1219);
      END_STATE();
    case 1185:
      if (lookahead == 'N') ADVANCE(1220);
      END_STATE();
    case 1186:
      if (lookahead == 'C') ADVANCE(1221);
      END_STATE();
    case 1187:
      if (lookahead == 'C') ADVANCE(1222);
      END_STATE();
    case 1188:
      ACCEPT_TOKEN(anon_sym_NS_AVAILABLE_IOS);
      END_STATE();
    case 1189:
      if (lookahead == 'L') ADVANCE(1223);
      END_STATE();
    case 1190:
      if (lookahead == 'T') ADVANCE(1224);
      END_STATE();
    case 1191:
      if (lookahead == 'S') ADVANCE(1225);
      END_STATE();
    case 1192:
      if (lookahead == 'E') ADVANCE(1226);
      END_STATE();
    case 1193:
      if (lookahead == 'E') ADVANCE(1227);
      END_STATE();
    case 1194:
      if (lookahead == 'V') ADVANCE(1228);
      END_STATE();
    case 1195:
      if (lookahead == 'O') ADVANCE(1229);
      END_STATE();
    case 1196:
      if (lookahead == 'T') ADVANCE(1230);
      END_STATE();
    case 1197:
      if (lookahead == 'A') ADVANCE(1231);
      END_STATE();
    case 1198:
      if (lookahead == 'N') ADVANCE(1232);
      END_STATE();
    case 1199:
      if (lookahead == 'L') ADVANCE(1233);
      END_STATE();
    case 1200:
      if (lookahead == 'I') ADVANCE(1234);
      END_STATE();
    case 1201:
      if (lookahead == 'd') ADVANCE(1235);
      END_STATE();
    case 1202:
      ACCEPT_TOKEN(anon_sym__Nullable_result);
      END_STATE();
    case 1203:
      if (lookahead == 'S') ADVANCE(1236);
      END_STATE();
    case 1204:
      if (lookahead == 'd') ADVANCE(1237);
      END_STATE();
    case 1205:
      if (lookahead == 'r') ADVANCE(1238);
      END_STATE();
    case 1206:
      if (lookahead == 'b') ADVANCE(1239);
      END_STATE();
    case 1207:
      if (lookahead == 'm') ADVANCE(1240);
      END_STATE();
    case 1208:
      ACCEPT_TOKEN(anon_sym___deprecated_msg);
      END_STATE();
    case 1209:
      if (lookahead == 'l') ADVANCE(1241);
      END_STATE();
    case 1210:
      if (lookahead == 's') ADVANCE(1242);
      END_STATE();
    case 1211:
      if (lookahead == 'u') ADVANCE(1243);
      END_STATE();
    case 1212:
      if (lookahead == 'n') ADVANCE(1244);
      END_STATE();
    case 1213:
      if (lookahead == 't') ADVANCE(1245);
      END_STATE();
    case 1214:
      if (lookahead == 'N') ADVANCE(1246);
      END_STATE();
    case 1215:
      if (lookahead == 'T') ADVANCE(1247);
      END_STATE();
    case 1216:
      if (lookahead == 'E') ADVANCE(1248);
      END_STATE();
    case 1217:
      if (lookahead == 'U') ADVANCE(1249);
      END_STATE();
    case 1218:
      if (lookahead == 'T') ADVANCE(1250);
      END_STATE();
    case 1219:
      ACCEPT_TOKEN(anon_sym_FOUNDATION_EXPORT);
      END_STATE();
    case 1220:
      ACCEPT_TOKEN(anon_sym_FOUNDATION_EXTERN);
      END_STATE();
    case 1221:
      if (lookahead == '_') ADVANCE(1251);
      END_STATE();
    case 1222:
      if (lookahead == 'O') ADVANCE(1252);
      END_STATE();
    case 1223:
      if (lookahead == 'E') ADVANCE(1253);
      END_STATE();
    case 1224:
      if (lookahead == 'E') ADVANCE(1254);
      END_STATE();
    case 1225:
      ACCEPT_TOKEN(anon_sym_NS_DEPRECATED_IOS);
      END_STATE();
    case 1226:
      if (lookahead == '_') ADVANCE(1255);
      END_STATE();
    case 1227:
      if (lookahead == 'D') ADVANCE(1256);
      END_STATE();
    case 1228:
      if (lookahead == 'A') ADVANCE(1257);
      END_STATE();
    case 1229:
      if (lookahead == 'N') ADVANCE(1258);
      END_STATE();
    case 1230:
      if (lookahead == 'E') ADVANCE(1259);
      END_STATE();
    case 1231:
      if (lookahead == 'B') ADVANCE(1260);
      END_STATE();
    case 1232:
      if (lookahead == 'D') ADVANCE(1261);
      END_STATE();
    case 1233:
      if (lookahead == 'E') ADVANCE(1262);
      END_STATE();
    case 1234:
      if (lookahead == 'B') ADVANCE(1263);
      END_STATE();
    case 1235:
      ACCEPT_TOKEN(anon_sym__Null_unspecified);
      END_STATE();
    case 1236:
      if (lookahead == 'T') ADVANCE(1264);
      END_STATE();
    case 1237:
      ACCEPT_TOKEN(anon_sym___bridge_retained);
      END_STATE();
    case 1238:
      ACCEPT_TOKEN(anon_sym___bridge_transfer);
      END_STATE();
    case 1239:
      if (lookahead == 'l') ADVANCE(1265);
      END_STATE();
    case 1240:
      if (lookahead == '_') ADVANCE(1266);
      END_STATE();
    case 1241:
      if (lookahead == 'a') ADVANCE(1267);
      END_STATE();
    case 1242:
      if (lookahead == 'a') ADVANCE(1268);
      END_STATE();
    case 1243:
      if (lookahead == 'p') ADVANCE(1269);
      END_STATE();
    case 1244:
      if (lookahead == 'e') ADVANCE(1270);
      END_STATE();
    case 1245:
      if (lookahead == 'e') ADVANCE(1271);
      END_STATE();
    case 1246:
      ACCEPT_TOKEN(anon_sym_CF_FORMAT_FUNCTION);
      END_STATE();
    case 1247:
      if (lookahead == 'A') ADVANCE(1272);
      END_STATE();
    case 1248:
      if (lookahead == 'D') ADVANCE(1273);
      END_STATE();
    case 1249:
      if (lookahead == 'T') ADVANCE(1274);
      END_STATE();
    case 1250:
      if (lookahead == 'R') ADVANCE(1275);
      END_STATE();
    case 1251:
      if (lookahead == 'I') ADVANCE(1276);
      END_STATE();
    case 1252:
      if (lookahead == 'U') ADVANCE(1277);
      END_STATE();
    case 1253:
      if (lookahead == '_') ADVANCE(1278);
      END_STATE();
    case 1254:
      if (lookahead == 'D') ADVANCE(1279);
      END_STATE();
    case 1255:
      if (lookahead == 'I') ADVANCE(1280);
      END_STATE();
    case 1256:
      if (lookahead == '_') ADVANCE(1281);
      END_STATE();
    case 1257:
      if (lookahead == 'I') ADVANCE(1282);
      END_STATE();
    case 1258:
      ACCEPT_TOKEN(anon_sym_NS_FORMAT_FUNCTION);
      END_STATE();
    case 1259:
      if (lookahead == 'R') ADVANCE(1283);
      END_STATE();
    case 1260:
      if (lookahead == 'L') ADVANCE(1284);
      END_STATE();
    case 1261:
      if (lookahead == '_') ADVANCE(1285);
      END_STATE();
    case 1262:
      if (lookahead == 'C') ADVANCE(1286);
      END_STATE();
    case 1263:
      if (lookahead == 'U') ADVANCE(1287);
      END_STATE();
    case 1264:
      if (lookahead == 'A') ADVANCE(1288);
      END_STATE();
    case 1265:
      if (lookahead == 'e') ADVANCE(1289);
      END_STATE();
    case 1266:
      if (lookahead == 'm') ADVANCE(1290);
      END_STATE();
    case 1267:
      if (lookahead == 's') ADVANCE(1291);
      END_STATE();
    case 1268:
      if (lookahead == '_') ADVANCE(1292);
      END_STATE();
    case 1269:
      if (lookahead == 'e') ADVANCE(1293);
      END_STATE();
    case 1270:
      if (lookahead == 'd') ADVANCE(1294);
      END_STATE();
    case 1271:
      if (lookahead == 'd') ADVANCE(1295);
      END_STATE();
    case 1272:
      if (lookahead == 'I') ADVANCE(1296);
      END_STATE();
    case 1273:
      ACCEPT_TOKEN(anon_sym_CF_RETURNS_RETAINED);
      END_STATE();
    case 1274:
      if (lookahead == 'E') ADVANCE(1297);
      END_STATE();
    case 1275:
      if (lookahead == 'I') ADVANCE(1298);
      END_STATE();
    case 1276:
      if (lookahead == 'N') ADVANCE(1299);
      END_STATE();
    case 1277:
      if (lookahead == 'N') ADVANCE(1300);
      END_STATE();
    case 1278:
      if (lookahead == 'I') ADVANCE(1301);
      END_STATE();
    case 1279:
      if (lookahead == '_') ADVANCE(1302);
      END_STATE();
    case 1280:
      if (lookahead == 'O') ADVANCE(1303);
      END_STATE();
    case 1281:
      if (lookahead == 'I') ADVANCE(1304);
      END_STATE();
    case 1282:
      if (lookahead == 'L') ADVANCE(1305);
      END_STATE();
    case 1283:
      if (lookahead == 'M') ADVANCE(1306);
      END_STATE();
    case 1284:
      if (lookahead == 'E') ADVANCE(1307);
      END_STATE();
    case 1285:
      if (lookahead == 'O') ADVANCE(1308);
      END_STATE();
    case 1286:
      if (lookahead == 'T') ADVANCE(1309);
      END_STATE();
    case 1287:
      if (lookahead == 'T') ADVANCE(1310);
      END_STATE();
    case 1288:
      if (lookahead == 'R') ADVANCE(1311);
      END_STATE();
    case 1289:
      ACCEPT_TOKEN(anon_sym___builtin_available);
      END_STATE();
    case 1290:
      if (lookahead == 's') ADVANCE(1312);
      END_STATE();
    case 1291:
      if (lookahead == 's') ADVANCE(1313);
      END_STATE();
    case 1292:
      if (lookahead == 'p') ADVANCE(1314);
      END_STATE();
    case 1293:
      if (lookahead == 'r') ADVANCE(1315);
      END_STATE();
    case 1294:
      ACCEPT_TOKEN(anon_sym___unsafe_unretained);
      END_STATE();
    case 1295:
      ACCEPT_TOKEN(anon_sym_objc_bridge_related);
      END_STATE();
    case 1296:
      if (lookahead == 'N') ADVANCE(1316);
      END_STATE();
    case 1297:
      ACCEPT_TOKEN(anon_sym_DEPRECATED_ATTRIBUTE);
      END_STATE();
    case 1298:
      if (lookahead == 'B') ADVANCE(1317);
      END_STATE();
    case 1299:
      if (lookahead == 'L') ADVANCE(1318);
      END_STATE();
    case 1300:
      if (lookahead == 'T') ADVANCE(1319);
      END_STATE();
    case 1301:
      if (lookahead == 'O') ADVANCE(1320);
      END_STATE();
    case 1302:
      if (lookahead == 'I') ADVANCE(1321);
      END_STATE();
    case 1303:
      if (lookahead == 'S') ADVANCE(1322);
      END_STATE();
    case 1304:
      if (lookahead == 'O') ADVANCE(1323);
      END_STATE();
    case 1305:
      if (lookahead == 'A') ADVANCE(1324);
      END_STATE();
    case 1306:
      if (lookahead == 'I') ADVANCE(1325);
      END_STATE();
    case 1307:
      ACCEPT_TOKEN(anon_sym_NS_SWIFT_UNAVAILABLE);
      END_STATE();
    case 1308:
      if (lookahead == 'F') ADVANCE(1326);
      END_STATE();
    case 1309:
      if (lookahead == 'O') ADVANCE(1327);
      END_STATE();
    case 1310:
      if (lookahead == 'E') ADVANCE(1328);
      END_STATE();
    case 1311:
      if (lookahead == 'T') ADVANCE(1329);
      END_STATE();
    case 1312:
      if (lookahead == 'g') ADVANCE(1330);
      END_STATE();
    case 1313:
      if (lookahead == '_') ADVANCE(1331);
      END_STATE();
    case 1314:
      if (lookahead == 'o') ADVANCE(1332);
      END_STATE();
    case 1315:
      if (lookahead == '_') ADVANCE(1333);
      END_STATE();
    case 1316:
      if (lookahead == 'E') ADVANCE(1334);
      END_STATE();
    case 1317:
      if (lookahead == 'U') ADVANCE(1335);
      END_STATE();
    case 1318:
      if (lookahead == 'I') ADVANCE(1336);
      END_STATE();
    case 1319:
      if (lookahead == '_') ADVANCE(1337);
      END_STATE();
    case 1320:
      if (lookahead == 'S') ADVANCE(1338);
      END_STATE();
    case 1321:
      if (lookahead == 'O') ADVANCE(1339);
      END_STATE();
    case 1322:
      ACCEPT_TOKEN(anon_sym_NS_ENUM_AVAILABLE_IOS);
      END_STATE();
    case 1323:
      if (lookahead == 'S') ADVANCE(1340);
      END_STATE();
    case 1324:
      if (lookahead == 'B') ADVANCE(1341);
      END_STATE();
    case 1325:
      if (lookahead == 'N') ADVANCE(1342);
      END_STATE();
    case 1326:
      if (lookahead == '_') ADVANCE(1343);
      END_STATE();
    case 1327:
      if (lookahead == 'R') ADVANCE(1344);
      END_STATE();
    case 1328:
      ACCEPT_TOKEN(anon_sym_UNAVAILABLE_ATTRIBUTE);
      END_STATE();
    case 1329:
      if (lookahead == 'I') ADVANCE(1345);
      END_STATE();
    case 1330:
      ACCEPT_TOKEN(anon_sym___deprecated_enum_msg);
      END_STATE();
    case 1331:
      if (lookahead == 'r') ADVANCE(1346);
      END_STATE();
    case 1332:
      if (lookahead == 'i') ADVANCE(1347);
      END_STATE();
    case 1333:
      if (lookahead == 'p') ADVANCE(1348);
      END_STATE();
    case 1334:
      if (lookahead == 'D') ADVANCE(1349);
      END_STATE();
    case 1335:
      if (lookahead == 'T') ADVANCE(1350);
      END_STATE();
    case 1336:
      if (lookahead == 'N') ADVANCE(1351);
      END_STATE();
    case 1337:
      if (lookahead == 'U') ADVANCE(1352);
      END_STATE();
    case 1338:
      ACCEPT_TOKEN(anon_sym_NS_CLASS_AVAILABLE_IOS);
      END_STATE();
    case 1339:
      if (lookahead == 'S') ADVANCE(1353);
      END_STATE();
    case 1340:
      ACCEPT_TOKEN(anon_sym_NS_ENUM_DEPRECATED_IOS);
      END_STATE();
    case 1341:
      if (lookahead == 'L') ADVANCE(1354);
      END_STATE();
    case 1342:
      if (lookahead == 'A') ADVANCE(1355);
      END_STATE();
    case 1343:
      if (lookahead == 'S') ADVANCE(1356);
      END_STATE();
    case 1344:
      ACCEPT_TOKEN(anon_sym_UI_APPEARANCE_SELECTOR);
      END_STATE();
    case 1345:
      if (lookahead == 'N') ADVANCE(1357);
      END_STATE();
    case 1346:
      if (lookahead == 'o') ADVANCE(1358);
      END_STATE();
    case 1347:
      if (lookahead == 'n') ADVANCE(1359);
      END_STATE();
    case 1348:
      if (lookahead == 'o') ADVANCE(1360);
      END_STATE();
    case 1349:
      ACCEPT_TOKEN(anon_sym_CF_RETURNS_NOT_RETAINED);
      END_STATE();
    case 1350:
      if (lookahead == 'E') ADVANCE(1361);
      END_STATE();
    case 1351:
      if (lookahead == 'E') ADVANCE(1362);
      END_STATE();
    case 1352:
      if (lookahead == 'N') ADVANCE(1363);
      END_STATE();
    case 1353:
      ACCEPT_TOKEN(anon_sym_NS_CLASS_DEPRECATED_IOS);
      END_STATE();
    case 1354:
      if (lookahead == 'E') ADVANCE(1364);
      END_STATE();
    case 1355:
      if (lookahead == 'T') ADVANCE(1365);
      END_STATE();
    case 1356:
      if (lookahead == 'C') ADVANCE(1366);
      END_STATE();
    case 1357:
      if (lookahead == 'G') ADVANCE(1367);
      END_STATE();
    case 1358:
      ACCEPT_TOKEN(anon_sym___ptrauth_objc_class_ro);
      END_STATE();
    case 1359:
      if (lookahead == 't') ADVANCE(1368);
      END_STATE();
    case 1360:
      if (lookahead == 'i') ADVANCE(1369);
      END_STATE();
    case 1361:
      ACCEPT_TOKEN(anon_sym_DEPRECATED_MSG_ATTRIBUTE);
      END_STATE();
    case 1362:
      ACCEPT_TOKEN(anon_sym_FOUNDATION_STATIC_INLINE);
      END_STATE();
    case 1363:
      if (lookahead == 'A') ADVANCE(1370);
      END_STATE();
    case 1364:
      if (lookahead == '_') ADVANCE(1371);
      END_STATE();
    case 1365:
      if (lookahead == 'I') ADVANCE(1372);
      END_STATE();
    case 1366:
      if (lookahead == 'O') ADVANCE(1373);
      END_STATE();
    case 1367:
      ACCEPT_TOKEN(anon_sym___OSX_AVAILABLE_STARTING);
      END_STATE();
    case 1368:
      if (lookahead == 'e') ADVANCE(1374);
      END_STATE();
    case 1369:
      if (lookahead == 'n') ADVANCE(1375);
      END_STATE();
    case 1370:
      if (lookahead == 'V') ADVANCE(1376);
      END_STATE();
    case 1371:
      if (lookahead == 'I') ADVANCE(1377);
      END_STATE();
    case 1372:
      if (lookahead == 'O') ADVANCE(1378);
      END_STATE();
    case 1373:
      if (lookahead == 'P') ADVANCE(1379);
      END_STATE();
    case 1374:
      if (lookahead == 'r') ADVANCE(1380);
      END_STATE();
    case 1375:
      if (lookahead == 't') ADVANCE(1381);
      END_STATE();
    case 1376:
      if (lookahead == 'A') ADVANCE(1382);
      END_STATE();
    case 1377:
      if (lookahead == 'O') ADVANCE(1383);
      END_STATE();
    case 1378:
      if (lookahead == 'N') ADVANCE(1384);
      END_STATE();
    case 1379:
      if (lookahead == 'E') ADVANCE(1385);
      END_STATE();
    case 1380:
      ACCEPT_TOKEN(anon_sym___ptrauth_objc_isa_pointer);
      END_STATE();
    case 1381:
      if (lookahead == 'e') ADVANCE(1386);
      END_STATE();
    case 1382:
      if (lookahead == 'I') ADVANCE(1387);
      END_STATE();
    case 1383:
      if (lookahead == 'S') ADVANCE(1388);
      END_STATE();
    case 1384:
      ACCEPT_TOKEN(anon_sym_NS_REQUIRES_NIL_TERMINATION);
      END_STATE();
    case 1385:
      ACCEPT_TOKEN(anon_sym_NS_VALID_UNTIL_END_OF_SCOPE);
      END_STATE();
    case 1386:
      if (lookahead == 'r') ADVANCE(1389);
      END_STATE();
    case 1387:
      if (lookahead == 'L') ADVANCE(1390);
      END_STATE();
    case 1388:
      ACCEPT_TOKEN(anon_sym_NS_EXTENSION_UNAVAILABLE_IOS);
      END_STATE();
    case 1389:
      ACCEPT_TOKEN(anon_sym___ptrauth_objc_super_pointer);
      END_STATE();
    case 1390:
      if (lookahead == 'A') ADVANCE(1391);
      END_STATE();
    case 1391:
      if (lookahead == 'B') ADVANCE(1392);
      END_STATE();
    case 1392:
      if (lookahead == 'L') ADVANCE(1393);
      END_STATE();
    case 1393:
      if (lookahead == 'E') ADVANCE(1394);
      END_STATE();
    case 1394:
      ACCEPT_TOKEN(anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0},
  [1] = {.lex_state = 416},
  [2] = {.lex_state = 98},
  [3] = {.lex_state = 98},
  [4] = {.lex_state = 98},
  [5] = {.lex_state = 98},
  [6] = {.lex_state = 98},
  [7] = {.lex_state = 98},
  [8] = {.lex_state = 98},
  [9] = {.lex_state = 98},
  [10] = {.lex_state = 98},
  [11] = {.lex_state = 98},
  [12] = {.lex_state = 98},
  [13] = {.lex_state = 98},
  [14] = {.lex_state = 98},
  [15] = {.lex_state = 98},
  [16] = {.lex_state = 98},
  [17] = {.lex_state = 98},
  [18] = {.lex_state = 98},
  [19] = {.lex_state = 98},
  [20] = {.lex_state = 108},
  [21] = {.lex_state = 108},
  [22] = {.lex_state = 108},
  [23] = {.lex_state = 108},
  [24] = {.lex_state = 108},
  [25] = {.lex_state = 108},
  [26] = {.lex_state = 108},
  [27] = {.lex_state = 108},
  [28] = {.lex_state = 108},
  [29] = {.lex_state = 108},
  [30] = {.lex_state = 108},
  [31] = {.lex_state = 108},
  [32] = {.lex_state = 108},
  [33] = {.lex_state = 98},
  [34] = {.lex_state = 108},
  [35] = {.lex_state = 108},
  [36] = {.lex_state = 108},
  [37] = {.lex_state = 108},
  [38] = {.lex_state = 107},
  [39] = {.lex_state = 107},
  [40] = {.lex_state = 107},
  [41] = {.lex_state = 417},
  [42] = {.lex_state = 417},
  [43] = {.lex_state = 417},
  [44] = {.lex_state = 417},
  [45] = {.lex_state = 417},
  [46] = {.lex_state = 417},
  [47] = {.lex_state = 417},
  [48] = {.lex_state = 417},
  [49] = {.lex_state = 417},
  [50] = {.lex_state = 417},
  [51] = {.lex_state = 417},
  [52] = {.lex_state = 417},
  [53] = {.lex_state = 417},
  [54] = {.lex_state = 417},
  [55] = {.lex_state = 417},
  [56] = {.lex_state = 417},
  [57] = {.lex_state = 417},
  [58] = {.lex_state = 417},
  [59] = {.lex_state = 417},
  [60] = {.lex_state = 417},
  [61] = {.lex_state = 417},
  [62] = {.lex_state = 417},
  [63] = {.lex_state = 417},
  [64] = {.lex_state = 417},
  [65] = {.lex_state = 417},
  [66] = {.lex_state = 417},
  [67] = {.lex_state = 417},
  [68] = {.lex_state = 417},
  [69] = {.lex_state = 417},
  [70] = {.lex_state = 417},
  [71] = {.lex_state = 417},
  [72] = {.lex_state = 417},
  [73] = {.lex_state = 417},
  [74] = {.lex_state = 417},
  [75] = {.lex_state = 417},
  [76] = {.lex_state = 417},
  [77] = {.lex_state = 417},
  [78] = {.lex_state = 417},
  [79] = {.lex_state = 417},
  [80] = {.lex_state = 417},
  [81] = {.lex_state = 417},
  [82] = {.lex_state = 417},
  [83] = {.lex_state = 417},
  [84] = {.lex_state = 417},
  [85] = {.lex_state = 417},
  [86] = {.lex_state = 417},
  [87] = {.lex_state = 417},
  [88] = {.lex_state = 417},
  [89] = {.lex_state = 417},
  [90] = {.lex_state = 417},
  [91] = {.lex_state = 417},
  [92] = {.lex_state = 417},
  [93] = {.lex_state = 417},
  [94] = {.lex_state = 417},
  [95] = {.lex_state = 417},
  [96] = {.lex_state = 417},
  [97] = {.lex_state = 417},
  [98] = {.lex_state = 417},
  [99] = {.lex_state = 417},
  [100] = {.lex_state = 417},
  [101] = {.lex_state = 417},
  [102] = {.lex_state = 417},
  [103] = {.lex_state = 417},
  [104] = {.lex_state = 417},
  [105] = {.lex_state = 417},
  [106] = {.lex_state = 417},
  [107] = {.lex_state = 417},
  [108] = {.lex_state = 417},
  [109] = {.lex_state = 417},
  [110] = {.lex_state = 417},
  [111] = {.lex_state = 417},
  [112] = {.lex_state = 417},
  [113] = {.lex_state = 417},
  [114] = {.lex_state = 417},
  [115] = {.lex_state = 417},
  [116] = {.lex_state = 417},
  [117] = {.lex_state = 417},
  [118] = {.lex_state = 417},
  [119] = {.lex_state = 417},
  [120] = {.lex_state = 417},
  [121] = {.lex_state = 417},
  [122] = {.lex_state = 417},
  [123] = {.lex_state = 417},
  [124] = {.lex_state = 417},
  [125] = {.lex_state = 417},
  [126] = {.lex_state = 417},
  [127] = {.lex_state = 417},
  [128] = {.lex_state = 416},
  [129] = {.lex_state = 416},
  [130] = {.lex_state = 98},
  [131] = {.lex_state = 98},
  [132] = {.lex_state = 98},
  [133] = {.lex_state = 98},
  [134] = {.lex_state = 98},
  [135] = {.lex_state = 108},
  [136] = {.lex_state = 108},
  [137] = {.lex_state = 108},
  [138] = {.lex_state = 108},
  [139] = {.lex_state = 108},
  [140] = {.lex_state = 107},
  [141] = {.lex_state = 107},
  [142] = {.lex_state = 107},
  [143] = {.lex_state = 107},
  [144] = {.lex_state = 107},
  [145] = {.lex_state = 417},
  [146] = {.lex_state = 417},
  [147] = {.lex_state = 417},
  [148] = {.lex_state = 417},
  [149] = {.lex_state = 417},
  [150] = {.lex_state = 417},
  [151] = {.lex_state = 417},
  [152] = {.lex_state = 417},
  [153] = {.lex_state = 417},
  [154] = {.lex_state = 417},
  [155] = {.lex_state = 105},
  [156] = {.lex_state = 105},
  [157] = {.lex_state = 105},
  [158] = {.lex_state = 105},
  [159] = {.lex_state = 105},
  [160] = {.lex_state = 105},
  [161] = {.lex_state = 105},
  [162] = {.lex_state = 105},
  [163] = {.lex_state = 105},
  [164] = {.lex_state = 105},
  [165] = {.lex_state = 105},
  [166] = {.lex_state = 105},
  [167] = {.lex_state = 105},
  [168] = {.lex_state = 96},
  [169] = {.lex_state = 111},
  [170] = {.lex_state = 111},
  [171] = {.lex_state = 111},
  [172] = {.lex_state = 111},
  [173] = {.lex_state = 111},
  [174] = {.lex_state = 111},
  [175] = {.lex_state = 111},
  [176] = {.lex_state = 111},
  [177] = {.lex_state = 97},
  [178] = {.lex_state = 94},
  [179] = {.lex_state = 100},
  [180] = {.lex_state = 100},
  [181] = {.lex_state = 94},
  [182] = {.lex_state = 109},
  [183] = {.lex_state = 100},
  [184] = {.lex_state = 100},
  [185] = {.lex_state = 109},
  [186] = {.lex_state = 100},
  [187] = {.lex_state = 100},
  [188] = {.lex_state = 100},
  [189] = {.lex_state = 106},
  [190] = {.lex_state = 106},
  [191] = {.lex_state = 109},
  [192] = {.lex_state = 109},
  [193] = {.lex_state = 109},
  [194] = {.lex_state = 109},
  [195] = {.lex_state = 109},
  [196] = {.lex_state = 418},
  [197] = {.lex_state = 98},
  [198] = {.lex_state = 418},
  [199] = {.lex_state = 98},
  [200] = {.lex_state = 98},
  [201] = {.lex_state = 98},
  [202] = {.lex_state = 98},
  [203] = {.lex_state = 98},
  [204] = {.lex_state = 98},
  [205] = {.lex_state = 98},
  [206] = {.lex_state = 98},
  [207] = {.lex_state = 98},
  [208] = {.lex_state = 98},
  [209] = {.lex_state = 98},
  [210] = {.lex_state = 98},
  [211] = {.lex_state = 106},
  [212] = {.lex_state = 98},
  [213] = {.lex_state = 106},
  [214] = {.lex_state = 98},
  [215] = {.lex_state = 98},
  [216] = {.lex_state = 98},
  [217] = {.lex_state = 98},
  [218] = {.lex_state = 98},
  [219] = {.lex_state = 98},
  [220] = {.lex_state = 98},
  [221] = {.lex_state = 418},
  [222] = {.lex_state = 98},
  [223] = {.lex_state = 98},
  [224] = {.lex_state = 106},
  [225] = {.lex_state = 106},
  [226] = {.lex_state = 106},
  [227] = {.lex_state = 98},
  [228] = {.lex_state = 98},
  [229] = {.lex_state = 98},
  [230] = {.lex_state = 98},
  [231] = {.lex_state = 98},
  [232] = {.lex_state = 98},
  [233] = {.lex_state = 98},
  [234] = {.lex_state = 98},
  [235] = {.lex_state = 98},
  [236] = {.lex_state = 98},
  [237] = {.lex_state = 98},
  [238] = {.lex_state = 98},
  [239] = {.lex_state = 98},
  [240] = {.lex_state = 98},
  [241] = {.lex_state = 98},
  [242] = {.lex_state = 98},
  [243] = {.lex_state = 98},
  [244] = {.lex_state = 98},
  [245] = {.lex_state = 98},
  [246] = {.lex_state = 98},
  [247] = {.lex_state = 98},
  [248] = {.lex_state = 98},
  [249] = {.lex_state = 98},
  [250] = {.lex_state = 98},
  [251] = {.lex_state = 98},
  [252] = {.lex_state = 98},
  [253] = {.lex_state = 98},
  [254] = {.lex_state = 98},
  [255] = {.lex_state = 98},
  [256] = {.lex_state = 98},
  [257] = {.lex_state = 98},
  [258] = {.lex_state = 98},
  [259] = {.lex_state = 98},
  [260] = {.lex_state = 98},
  [261] = {.lex_state = 98},
  [262] = {.lex_state = 98},
  [263] = {.lex_state = 98},
  [264] = {.lex_state = 98},
  [265] = {.lex_state = 98},
  [266] = {.lex_state = 98},
  [267] = {.lex_state = 98},
  [268] = {.lex_state = 98},
  [269] = {.lex_state = 98},
  [270] = {.lex_state = 98},
  [271] = {.lex_state = 98},
  [272] = {.lex_state = 98},
  [273] = {.lex_state = 98},
  [274] = {.lex_state = 98},
  [275] = {.lex_state = 418},
  [276] = {.lex_state = 98},
  [277] = {.lex_state = 98},
  [278] = {.lex_state = 98},
  [279] = {.lex_state = 98},
  [280] = {.lex_state = 98},
  [281] = {.lex_state = 98},
  [282] = {.lex_state = 98},
  [283] = {.lex_state = 98},
  [284] = {.lex_state = 98},
  [285] = {.lex_state = 98},
  [286] = {.lex_state = 98},
  [287] = {.lex_state = 108},
  [288] = {.lex_state = 98},
  [289] = {.lex_state = 98},
  [290] = {.lex_state = 98},
  [291] = {.lex_state = 98},
  [292] = {.lex_state = 98},
  [293] = {.lex_state = 98},
  [294] = {.lex_state = 98},
  [295] = {.lex_state = 98},
  [296] = {.lex_state = 98},
  [297] = {.lex_state = 98},
  [298] = {.lex_state = 98},
  [299] = {.lex_state = 98},
  [300] = {.lex_state = 98},
  [301] = {.lex_state = 98},
  [302] = {.lex_state = 98},
  [303] = {.lex_state = 98},
  [304] = {.lex_state = 98},
  [305] = {.lex_state = 98},
  [306] = {.lex_state = 98},
  [307] = {.lex_state = 98},
  [308] = {.lex_state = 98},
  [309] = {.lex_state = 98},
  [310] = {.lex_state = 98},
  [311] = {.lex_state = 98},
  [312] = {.lex_state = 98},
  [313] = {.lex_state = 98},
  [314] = {.lex_state = 98},
  [315] = {.lex_state = 98},
  [316] = {.lex_state = 98},
  [317] = {.lex_state = 98},
  [318] = {.lex_state = 98},
  [319] = {.lex_state = 98},
  [320] = {.lex_state = 98},
  [321] = {.lex_state = 98},
  [322] = {.lex_state = 98},
  [323] = {.lex_state = 98},
  [324] = {.lex_state = 98},
  [325] = {.lex_state = 98},
  [326] = {.lex_state = 98},
  [327] = {.lex_state = 98},
  [328] = {.lex_state = 98},
  [329] = {.lex_state = 98},
  [330] = {.lex_state = 98},
  [331] = {.lex_state = 98},
  [332] = {.lex_state = 98},
  [333] = {.lex_state = 98},
  [334] = {.lex_state = 98},
  [335] = {.lex_state = 98},
  [336] = {.lex_state = 98},
  [337] = {.lex_state = 98},
  [338] = {.lex_state = 98},
  [339] = {.lex_state = 98},
  [340] = {.lex_state = 98},
  [341] = {.lex_state = 98},
  [342] = {.lex_state = 98},
  [343] = {.lex_state = 98},
  [344] = {.lex_state = 98},
  [345] = {.lex_state = 98},
  [346] = {.lex_state = 98},
  [347] = {.lex_state = 98},
  [348] = {.lex_state = 98},
  [349] = {.lex_state = 98},
  [350] = {.lex_state = 98},
  [351] = {.lex_state = 98},
  [352] = {.lex_state = 98},
  [353] = {.lex_state = 98},
  [354] = {.lex_state = 98},
  [355] = {.lex_state = 98},
  [356] = {.lex_state = 98},
  [357] = {.lex_state = 98},
  [358] = {.lex_state = 98},
  [359] = {.lex_state = 98},
  [360] = {.lex_state = 98},
  [361] = {.lex_state = 108},
  [362] = {.lex_state = 108},
  [363] = {.lex_state = 108},
  [364] = {.lex_state = 108},
  [365] = {.lex_state = 108},
  [366] = {.lex_state = 108},
  [367] = {.lex_state = 108},
  [368] = {.lex_state = 108},
  [369] = {.lex_state = 108},
  [370] = {.lex_state = 108},
  [371] = {.lex_state = 108},
  [372] = {.lex_state = 108},
  [373] = {.lex_state = 418},
  [374] = {.lex_state = 108},
  [375] = {.lex_state = 108},
  [376] = {.lex_state = 108},
  [377] = {.lex_state = 108},
  [378] = {.lex_state = 108},
  [379] = {.lex_state = 108},
  [380] = {.lex_state = 418},
  [381] = {.lex_state = 418},
  [382] = {.lex_state = 418},
  [383] = {.lex_state = 108},
  [384] = {.lex_state = 108},
  [385] = {.lex_state = 418},
  [386] = {.lex_state = 418},
  [387] = {.lex_state = 418},
  [388] = {.lex_state = 108},
  [389] = {.lex_state = 108},
  [390] = {.lex_state = 108},
  [391] = {.lex_state = 108},
  [392] = {.lex_state = 108},
  [393] = {.lex_state = 108},
  [394] = {.lex_state = 108},
  [395] = {.lex_state = 108},
  [396] = {.lex_state = 108},
  [397] = {.lex_state = 108},
  [398] = {.lex_state = 108},
  [399] = {.lex_state = 108},
  [400] = {.lex_state = 108},
  [401] = {.lex_state = 108},
  [402] = {.lex_state = 108},
  [403] = {.lex_state = 108},
  [404] = {.lex_state = 108},
  [405] = {.lex_state = 108},
  [406] = {.lex_state = 108},
  [407] = {.lex_state = 108},
  [408] = {.lex_state = 108},
  [409] = {.lex_state = 418},
  [410] = {.lex_state = 108},
  [411] = {.lex_state = 108},
  [412] = {.lex_state = 108},
  [413] = {.lex_state = 108},
  [414] = {.lex_state = 108},
  [415] = {.lex_state = 108},
  [416] = {.lex_state = 108},
  [417] = {.lex_state = 108},
  [418] = {.lex_state = 108},
  [419] = {.lex_state = 108},
  [420] = {.lex_state = 108},
  [421] = {.lex_state = 108},
  [422] = {.lex_state = 108},
  [423] = {.lex_state = 108},
  [424] = {.lex_state = 108},
  [425] = {.lex_state = 108},
  [426] = {.lex_state = 108},
  [427] = {.lex_state = 418},
  [428] = {.lex_state = 108},
  [429] = {.lex_state = 108},
  [430] = {.lex_state = 108},
  [431] = {.lex_state = 108},
  [432] = {.lex_state = 108},
  [433] = {.lex_state = 108},
  [434] = {.lex_state = 108},
  [435] = {.lex_state = 108},
  [436] = {.lex_state = 108},
  [437] = {.lex_state = 108},
  [438] = {.lex_state = 418},
  [439] = {.lex_state = 108},
  [440] = {.lex_state = 108},
  [441] = {.lex_state = 108},
  [442] = {.lex_state = 108},
  [443] = {.lex_state = 108},
  [444] = {.lex_state = 108},
  [445] = {.lex_state = 108},
  [446] = {.lex_state = 108},
  [447] = {.lex_state = 108},
  [448] = {.lex_state = 108},
  [449] = {.lex_state = 108},
  [450] = {.lex_state = 108},
  [451] = {.lex_state = 108},
  [452] = {.lex_state = 108},
  [453] = {.lex_state = 108},
  [454] = {.lex_state = 108},
  [455] = {.lex_state = 108},
  [456] = {.lex_state = 108},
  [457] = {.lex_state = 108},
  [458] = {.lex_state = 108},
  [459] = {.lex_state = 108},
  [460] = {.lex_state = 108},
  [461] = {.lex_state = 108},
  [462] = {.lex_state = 419},
  [463] = {.lex_state = 108},
  [464] = {.lex_state = 108},
  [465] = {.lex_state = 108},
  [466] = {.lex_state = 108},
  [467] = {.lex_state = 108},
  [468] = {.lex_state = 108},
  [469] = {.lex_state = 108},
  [470] = {.lex_state = 108},
  [471] = {.lex_state = 108},
  [472] = {.lex_state = 108},
  [473] = {.lex_state = 108},
  [474] = {.lex_state = 108},
  [475] = {.lex_state = 108},
  [476] = {.lex_state = 108},
  [477] = {.lex_state = 108},
  [478] = {.lex_state = 108},
  [479] = {.lex_state = 108},
  [480] = {.lex_state = 108},
  [481] = {.lex_state = 108},
  [482] = {.lex_state = 108},
  [483] = {.lex_state = 108},
  [484] = {.lex_state = 107},
  [485] = {.lex_state = 108},
  [486] = {.lex_state = 108},
  [487] = {.lex_state = 108},
  [488] = {.lex_state = 108},
  [489] = {.lex_state = 108},
  [490] = {.lex_state = 108},
  [491] = {.lex_state = 108},
  [492] = {.lex_state = 108},
  [493] = {.lex_state = 108},
  [494] = {.lex_state = 108},
  [495] = {.lex_state = 108},
  [496] = {.lex_state = 108},
  [497] = {.lex_state = 108},
  [498] = {.lex_state = 108},
  [499] = {.lex_state = 108},
  [500] = {.lex_state = 108},
  [501] = {.lex_state = 108},
  [502] = {.lex_state = 108},
  [503] = {.lex_state = 108},
  [504] = {.lex_state = 108},
  [505] = {.lex_state = 108},
  [506] = {.lex_state = 108},
  [507] = {.lex_state = 108},
  [508] = {.lex_state = 108},
  [509] = {.lex_state = 108},
  [510] = {.lex_state = 108},
  [511] = {.lex_state = 108},
  [512] = {.lex_state = 108},
  [513] = {.lex_state = 108},
  [514] = {.lex_state = 108},
  [515] = {.lex_state = 108},
  [516] = {.lex_state = 108},
  [517] = {.lex_state = 108},
  [518] = {.lex_state = 108},
  [519] = {.lex_state = 108},
  [520] = {.lex_state = 108},
  [521] = {.lex_state = 108},
  [522] = {.lex_state = 108},
  [523] = {.lex_state = 108},
  [524] = {.lex_state = 108},
  [525] = {.lex_state = 108},
  [526] = {.lex_state = 108},
  [527] = {.lex_state = 107},
  [528] = {.lex_state = 107},
  [529] = {.lex_state = 107},
  [530] = {.lex_state = 107},
  [531] = {.lex_state = 107},
  [532] = {.lex_state = 107},
  [533] = {.lex_state = 107},
  [534] = {.lex_state = 107},
  [535] = {.lex_state = 107},
  [536] = {.lex_state = 107},
  [537] = {.lex_state = 107},
  [538] = {.lex_state = 107},
  [539] = {.lex_state = 107},
  [540] = {.lex_state = 107},
  [541] = {.lex_state = 107},
  [542] = {.lex_state = 107},
  [543] = {.lex_state = 107},
  [544] = {.lex_state = 107},
  [545] = {.lex_state = 107},
  [546] = {.lex_state = 107},
  [547] = {.lex_state = 107},
  [548] = {.lex_state = 107},
  [549] = {.lex_state = 107},
  [550] = {.lex_state = 107},
  [551] = {.lex_state = 107},
  [552] = {.lex_state = 107},
  [553] = {.lex_state = 107},
  [554] = {.lex_state = 107},
  [555] = {.lex_state = 107},
  [556] = {.lex_state = 107},
  [557] = {.lex_state = 107},
  [558] = {.lex_state = 107},
  [559] = {.lex_state = 107},
  [560] = {.lex_state = 107},
  [561] = {.lex_state = 107},
  [562] = {.lex_state = 107},
  [563] = {.lex_state = 107},
  [564] = {.lex_state = 107},
  [565] = {.lex_state = 107},
  [566] = {.lex_state = 107},
  [567] = {.lex_state = 107},
  [568] = {.lex_state = 107},
  [569] = {.lex_state = 107},
  [570] = {.lex_state = 107},
  [571] = {.lex_state = 107},
  [572] = {.lex_state = 419},
  [573] = {.lex_state = 107},
  [574] = {.lex_state = 107},
  [575] = {.lex_state = 107},
  [576] = {.lex_state = 107},
  [577] = {.lex_state = 107},
  [578] = {.lex_state = 107},
  [579] = {.lex_state = 107},
  [580] = {.lex_state = 107},
  [581] = {.lex_state = 107},
  [582] = {.lex_state = 107},
  [583] = {.lex_state = 107},
  [584] = {.lex_state = 107},
  [585] = {.lex_state = 107},
  [586] = {.lex_state = 107},
  [587] = {.lex_state = 107},
  [588] = {.lex_state = 107},
  [589] = {.lex_state = 107},
  [590] = {.lex_state = 107},
  [591] = {.lex_state = 107},
  [592] = {.lex_state = 107},
  [593] = {.lex_state = 107},
  [594] = {.lex_state = 107},
  [595] = {.lex_state = 95},
  [596] = {.lex_state = 107},
  [597] = {.lex_state = 107},
  [598] = {.lex_state = 107},
  [599] = {.lex_state = 107},
  [600] = {.lex_state = 107},
  [601] = {.lex_state = 107},
  [602] = {.lex_state = 107},
  [603] = {.lex_state = 107},
  [604] = {.lex_state = 107},
  [605] = {.lex_state = 107},
  [606] = {.lex_state = 107},
  [607] = {.lex_state = 107},
  [608] = {.lex_state = 107},
  [609] = {.lex_state = 107},
  [610] = {.lex_state = 107},
  [611] = {.lex_state = 107},
  [612] = {.lex_state = 107},
  [613] = {.lex_state = 107},
  [614] = {.lex_state = 107},
  [615] = {.lex_state = 107},
  [616] = {.lex_state = 107},
  [617] = {.lex_state = 107},
  [618] = {.lex_state = 107},
  [619] = {.lex_state = 107},
  [620] = {.lex_state = 107},
  [621] = {.lex_state = 107},
  [622] = {.lex_state = 107},
  [623] = {.lex_state = 107},
  [624] = {.lex_state = 107},
  [625] = {.lex_state = 107},
  [626] = {.lex_state = 107},
  [627] = {.lex_state = 107},
  [628] = {.lex_state = 107},
  [629] = {.lex_state = 107},
  [630] = {.lex_state = 107},
  [631] = {.lex_state = 107},
  [632] = {.lex_state = 107},
  [633] = {.lex_state = 107},
  [634] = {.lex_state = 107},
  [635] = {.lex_state = 107},
  [636] = {.lex_state = 107},
  [637] = {.lex_state = 107},
  [638] = {.lex_state = 107},
  [639] = {.lex_state = 107},
  [640] = {.lex_state = 417},
  [641] = {.lex_state = 107},
  [642] = {.lex_state = 107},
  [643] = {.lex_state = 107},
  [644] = {.lex_state = 107},
  [645] = {.lex_state = 107},
  [646] = {.lex_state = 107},
  [647] = {.lex_state = 107},
  [648] = {.lex_state = 107},
  [649] = {.lex_state = 107},
  [650] = {.lex_state = 107},
  [651] = {.lex_state = 107},
  [652] = {.lex_state = 107},
  [653] = {.lex_state = 107},
  [654] = {.lex_state = 107},
  [655] = {.lex_state = 107},
  [656] = {.lex_state = 417},
  [657] = {.lex_state = 107},
  [658] = {.lex_state = 107},
  [659] = {.lex_state = 107},
  [660] = {.lex_state = 107},
  [661] = {.lex_state = 107},
  [662] = {.lex_state = 107},
  [663] = {.lex_state = 107},
  [664] = {.lex_state = 107},
  [665] = {.lex_state = 107},
  [666] = {.lex_state = 107},
  [667] = {.lex_state = 107},
  [668] = {.lex_state = 107},
  [669] = {.lex_state = 107},
  [670] = {.lex_state = 107},
  [671] = {.lex_state = 107},
  [672] = {.lex_state = 107},
  [673] = {.lex_state = 107},
  [674] = {.lex_state = 107},
  [675] = {.lex_state = 107},
  [676] = {.lex_state = 107},
  [677] = {.lex_state = 107},
  [678] = {.lex_state = 107},
  [679] = {.lex_state = 107},
  [680] = {.lex_state = 107},
  [681] = {.lex_state = 107},
  [682] = {.lex_state = 107},
  [683] = {.lex_state = 107},
  [684] = {.lex_state = 107},
  [685] = {.lex_state = 417},
  [686] = {.lex_state = 417},
  [687] = {.lex_state = 417},
  [688] = {.lex_state = 417},
  [689] = {.lex_state = 417},
  [690] = {.lex_state = 417},
  [691] = {.lex_state = 417},
  [692] = {.lex_state = 417},
  [693] = {.lex_state = 417},
  [694] = {.lex_state = 417},
  [695] = {.lex_state = 417},
  [696] = {.lex_state = 417},
  [697] = {.lex_state = 417},
  [698] = {.lex_state = 417},
  [699] = {.lex_state = 417},
  [700] = {.lex_state = 417},
  [701] = {.lex_state = 417},
  [702] = {.lex_state = 417},
  [703] = {.lex_state = 417},
  [704] = {.lex_state = 417},
  [705] = {.lex_state = 417},
  [706] = {.lex_state = 417},
  [707] = {.lex_state = 417},
  [708] = {.lex_state = 417},
  [709] = {.lex_state = 417},
  [710] = {.lex_state = 417},
  [711] = {.lex_state = 417},
  [712] = {.lex_state = 417},
  [713] = {.lex_state = 417},
  [714] = {.lex_state = 417},
  [715] = {.lex_state = 417},
  [716] = {.lex_state = 417},
  [717] = {.lex_state = 417},
  [718] = {.lex_state = 417},
  [719] = {.lex_state = 417},
  [720] = {.lex_state = 417},
  [721] = {.lex_state = 417},
  [722] = {.lex_state = 417},
  [723] = {.lex_state = 419},
  [724] = {.lex_state = 417},
  [725] = {.lex_state = 417},
  [726] = {.lex_state = 417},
  [727] = {.lex_state = 417},
  [728] = {.lex_state = 417},
  [729] = {.lex_state = 417},
  [730] = {.lex_state = 417},
  [731] = {.lex_state = 417},
  [732] = {.lex_state = 417},
  [733] = {.lex_state = 417},
  [734] = {.lex_state = 417},
  [735] = {.lex_state = 417},
  [736] = {.lex_state = 417},
  [737] = {.lex_state = 417},
  [738] = {.lex_state = 417},
  [739] = {.lex_state = 417},
  [740] = {.lex_state = 417},
  [741] = {.lex_state = 417},
  [742] = {.lex_state = 417},
  [743] = {.lex_state = 417},
  [744] = {.lex_state = 417},
  [745] = {.lex_state = 417},
  [746] = {.lex_state = 417},
  [747] = {.lex_state = 417},
  [748] = {.lex_state = 417},
  [749] = {.lex_state = 417},
  [750] = {.lex_state = 417},
  [751] = {.lex_state = 417},
  [752] = {.lex_state = 419},
  [753] = {.lex_state = 419},
  [754] = {.lex_state = 419},
  [755] = {.lex_state = 417},
  [756] = {.lex_state = 417},
  [757] = {.lex_state = 417},
  [758] = {.lex_state = 417},
  [759] = {.lex_state = 417},
  [760] = {.lex_state = 417},
  [761] = {.lex_state = 417},
  [762] = {.lex_state = 417},
  [763] = {.lex_state = 419},
  [764] = {.lex_state = 417},
  [765] = {.lex_state = 417},
  [766] = {.lex_state = 417},
  [767] = {.lex_state = 417},
  [768] = {.lex_state = 417},
  [769] = {.lex_state = 417},
  [770] = {.lex_state = 417},
  [771] = {.lex_state = 417},
  [772] = {.lex_state = 417},
  [773] = {.lex_state = 417},
  [774] = {.lex_state = 417},
  [775] = {.lex_state = 417},
  [776] = {.lex_state = 417},
  [777] = {.lex_state = 417},
  [778] = {.lex_state = 417},
  [779] = {.lex_state = 417},
  [780] = {.lex_state = 417},
  [781] = {.lex_state = 417},
  [782] = {.lex_state = 417},
  [783] = {.lex_state = 417},
  [784] = {.lex_state = 417},
  [785] = {.lex_state = 417},
  [786] = {.lex_state = 417},
  [787] = {.lex_state = 417},
  [788] = {.lex_state = 417},
  [789] = {.lex_state = 417},
  [790] = {.lex_state = 417},
  [791] = {.lex_state = 417},
  [792] = {.lex_state = 417},
  [793] = {.lex_state = 417},
  [794] = {.lex_state = 417},
  [795] = {.lex_state = 417},
  [796] = {.lex_state = 417},
  [797] = {.lex_state = 417},
  [798] = {.lex_state = 417},
  [799] = {.lex_state = 417},
  [800] = {.lex_state = 417},
  [801] = {.lex_state = 417},
  [802] = {.lex_state = 417},
  [803] = {.lex_state = 417},
  [804] = {.lex_state = 417},
  [805] = {.lex_state = 417},
  [806] = {.lex_state = 417},
  [807] = {.lex_state = 417},
  [808] = {.lex_state = 417},
  [809] = {.lex_state = 417},
  [810] = {.lex_state = 417},
  [811] = {.lex_state = 417},
  [812] = {.lex_state = 417},
  [813] = {.lex_state = 417},
  [814] = {.lex_state = 417},
  [815] = {.lex_state = 417},
  [816] = {.lex_state = 417},
  [817] = {.lex_state = 417},
  [818] = {.lex_state = 417},
  [819] = {.lex_state = 417},
  [820] = {.lex_state = 417},
  [821] = {.lex_state = 417},
  [822] = {.lex_state = 417},
  [823] = {.lex_state = 417},
  [824] = {.lex_state = 417},
  [825] = {.lex_state = 417},
  [826] = {.lex_state = 95},
  [827] = {.lex_state = 417},
  [828] = {.lex_state = 417},
  [829] = {.lex_state = 417},
  [830] = {.lex_state = 417},
  [831] = {.lex_state = 417},
  [832] = {.lex_state = 417},
  [833] = {.lex_state = 417},
  [834] = {.lex_state = 417},
  [835] = {.lex_state = 417},
  [836] = {.lex_state = 417},
  [837] = {.lex_state = 417},
  [838] = {.lex_state = 417},
  [839] = {.lex_state = 417},
  [840] = {.lex_state = 417},
  [841] = {.lex_state = 417},
  [842] = {.lex_state = 417},
  [843] = {.lex_state = 417},
  [844] = {.lex_state = 417},
  [845] = {.lex_state = 417},
  [846] = {.lex_state = 417},
  [847] = {.lex_state = 417},
  [848] = {.lex_state = 417},
  [849] = {.lex_state = 417},
  [850] = {.lex_state = 417},
  [851] = {.lex_state = 417},
  [852] = {.lex_state = 417},
  [853] = {.lex_state = 417},
  [854] = {.lex_state = 417},
  [855] = {.lex_state = 417},
  [856] = {.lex_state = 417},
  [857] = {.lex_state = 417},
  [858] = {.lex_state = 417},
  [859] = {.lex_state = 417},
  [860] = {.lex_state = 417},
  [861] = {.lex_state = 417},
  [862] = {.lex_state = 417},
  [863] = {.lex_state = 417},
  [864] = {.lex_state = 417},
  [865] = {.lex_state = 417},
  [866] = {.lex_state = 417},
  [867] = {.lex_state = 417},
  [868] = {.lex_state = 417},
  [869] = {.lex_state = 417},
  [870] = {.lex_state = 417},
  [871] = {.lex_state = 417},
  [872] = {.lex_state = 417},
  [873] = {.lex_state = 417},
  [874] = {.lex_state = 417},
  [875] = {.lex_state = 417},
  [876] = {.lex_state = 417},
  [877] = {.lex_state = 417},
  [878] = {.lex_state = 417},
  [879] = {.lex_state = 417},
  [880] = {.lex_state = 417},
  [881] = {.lex_state = 417},
  [882] = {.lex_state = 417},
  [883] = {.lex_state = 417},
  [884] = {.lex_state = 417},
  [885] = {.lex_state = 417},
  [886] = {.lex_state = 417},
  [887] = {.lex_state = 417},
  [888] = {.lex_state = 417},
  [889] = {.lex_state = 417},
  [890] = {.lex_state = 417},
  [891] = {.lex_state = 417},
  [892] = {.lex_state = 417},
  [893] = {.lex_state = 417},
  [894] = {.lex_state = 417},
  [895] = {.lex_state = 417},
  [896] = {.lex_state = 417},
  [897] = {.lex_state = 417},
  [898] = {.lex_state = 417},
  [899] = {.lex_state = 417},
  [900] = {.lex_state = 417},
  [901] = {.lex_state = 417},
  [902] = {.lex_state = 417},
  [903] = {.lex_state = 417},
  [904] = {.lex_state = 417},
  [905] = {.lex_state = 417},
  [906] = {.lex_state = 417},
  [907] = {.lex_state = 417},
  [908] = {.lex_state = 417},
  [909] = {.lex_state = 416},
  [910] = {.lex_state = 416},
  [911] = {.lex_state = 115},
  [912] = {.lex_state = 115},
  [913] = {.lex_state = 416},
  [914] = {.lex_state = 115},
  [915] = {.lex_state = 115},
  [916] = {.lex_state = 115},
  [917] = {.lex_state = 416},
  [918] = {.lex_state = 416},
  [919] = {.lex_state = 115},
  [920] = {.lex_state = 416},
  [921] = {.lex_state = 416},
  [922] = {.lex_state = 416},
  [923] = {.lex_state = 416},
  [924] = {.lex_state = 416},
  [925] = {.lex_state = 416},
  [926] = {.lex_state = 416},
  [927] = {.lex_state = 416},
  [928] = {.lex_state = 416},
  [929] = {.lex_state = 416},
  [930] = {.lex_state = 416},
  [931] = {.lex_state = 416},
  [932] = {.lex_state = 416},
  [933] = {.lex_state = 416},
  [934] = {.lex_state = 416},
  [935] = {.lex_state = 416},
  [936] = {.lex_state = 416},
  [937] = {.lex_state = 416},
  [938] = {.lex_state = 416},
  [939] = {.lex_state = 416},
  [940] = {.lex_state = 416},
  [941] = {.lex_state = 416},
  [942] = {.lex_state = 416},
  [943] = {.lex_state = 416},
  [944] = {.lex_state = 416},
  [945] = {.lex_state = 416},
  [946] = {.lex_state = 416},
  [947] = {.lex_state = 416},
  [948] = {.lex_state = 416},
  [949] = {.lex_state = 416},
  [950] = {.lex_state = 416},
  [951] = {.lex_state = 416},
  [952] = {.lex_state = 416},
  [953] = {.lex_state = 416},
  [954] = {.lex_state = 416},
  [955] = {.lex_state = 416},
  [956] = {.lex_state = 416},
  [957] = {.lex_state = 416},
  [958] = {.lex_state = 416},
  [959] = {.lex_state = 416},
  [960] = {.lex_state = 416},
  [961] = {.lex_state = 416},
  [962] = {.lex_state = 416},
  [963] = {.lex_state = 416},
  [964] = {.lex_state = 416},
  [965] = {.lex_state = 416},
  [966] = {.lex_state = 416},
  [967] = {.lex_state = 416},
  [968] = {.lex_state = 416},
  [969] = {.lex_state = 416},
  [970] = {.lex_state = 416},
  [971] = {.lex_state = 416},
  [972] = {.lex_state = 416},
  [973] = {.lex_state = 416},
  [974] = {.lex_state = 416},
  [975] = {.lex_state = 416},
  [976] = {.lex_state = 416},
  [977] = {.lex_state = 416},
  [978] = {.lex_state = 416},
  [979] = {.lex_state = 416},
  [980] = {.lex_state = 416},
  [981] = {.lex_state = 416},
  [982] = {.lex_state = 416},
  [983] = {.lex_state = 416},
  [984] = {.lex_state = 416},
  [985] = {.lex_state = 416},
  [986] = {.lex_state = 416},
  [987] = {.lex_state = 416},
  [988] = {.lex_state = 416},
  [989] = {.lex_state = 416},
  [990] = {.lex_state = 416},
  [991] = {.lex_state = 416},
  [992] = {.lex_state = 111},
  [993] = {.lex_state = 416},
  [994] = {.lex_state = 416},
  [995] = {.lex_state = 416},
  [996] = {.lex_state = 416},
  [997] = {.lex_state = 416},
  [998] = {.lex_state = 416},
  [999] = {.lex_state = 416},
  [1000] = {.lex_state = 416},
  [1001] = {.lex_state = 416},
  [1002] = {.lex_state = 416},
  [1003] = {.lex_state = 416},
  [1004] = {.lex_state = 416},
  [1005] = {.lex_state = 416},
  [1006] = {.lex_state = 416},
  [1007] = {.lex_state = 416},
  [1008] = {.lex_state = 416},
  [1009] = {.lex_state = 416},
  [1010] = {.lex_state = 416},
  [1011] = {.lex_state = 416},
  [1012] = {.lex_state = 416},
  [1013] = {.lex_state = 416},
  [1014] = {.lex_state = 416},
  [1015] = {.lex_state = 416},
  [1016] = {.lex_state = 416},
  [1017] = {.lex_state = 416},
  [1018] = {.lex_state = 416},
  [1019] = {.lex_state = 416},
  [1020] = {.lex_state = 416},
  [1021] = {.lex_state = 416},
  [1022] = {.lex_state = 416},
  [1023] = {.lex_state = 416},
  [1024] = {.lex_state = 416},
  [1025] = {.lex_state = 111},
  [1026] = {.lex_state = 416},
  [1027] = {.lex_state = 416},
  [1028] = {.lex_state = 416},
  [1029] = {.lex_state = 416},
  [1030] = {.lex_state = 416},
  [1031] = {.lex_state = 416},
  [1032] = {.lex_state = 416},
  [1033] = {.lex_state = 416},
  [1034] = {.lex_state = 416},
  [1035] = {.lex_state = 416},
  [1036] = {.lex_state = 416},
  [1037] = {.lex_state = 416},
  [1038] = {.lex_state = 416},
  [1039] = {.lex_state = 416},
  [1040] = {.lex_state = 416},
  [1041] = {.lex_state = 416},
  [1042] = {.lex_state = 416},
  [1043] = {.lex_state = 416},
  [1044] = {.lex_state = 416},
  [1045] = {.lex_state = 416},
  [1046] = {.lex_state = 416},
  [1047] = {.lex_state = 416},
  [1048] = {.lex_state = 416},
  [1049] = {.lex_state = 416},
  [1050] = {.lex_state = 416},
  [1051] = {.lex_state = 416},
  [1052] = {.lex_state = 416},
  [1053] = {.lex_state = 416},
  [1054] = {.lex_state = 416},
  [1055] = {.lex_state = 416},
  [1056] = {.lex_state = 416},
  [1057] = {.lex_state = 416},
  [1058] = {.lex_state = 416},
  [1059] = {.lex_state = 416},
  [1060] = {.lex_state = 416},
  [1061] = {.lex_state = 416},
  [1062] = {.lex_state = 416},
  [1063] = {.lex_state = 416},
  [1064] = {.lex_state = 416},
  [1065] = {.lex_state = 416},
  [1066] = {.lex_state = 416},
  [1067] = {.lex_state = 111},
  [1068] = {.lex_state = 111},
  [1069] = {.lex_state = 111},
  [1070] = {.lex_state = 111},
  [1071] = {.lex_state = 416},
  [1072] = {.lex_state = 111},
  [1073] = {.lex_state = 111},
  [1074] = {.lex_state = 110},
  [1075] = {.lex_state = 110},
  [1076] = {.lex_state = 114},
  [1077] = {.lex_state = 114},
  [1078] = {.lex_state = 114},
  [1079] = {.lex_state = 114},
  [1080] = {.lex_state = 114},
  [1081] = {.lex_state = 114},
  [1082] = {.lex_state = 114},
  [1083] = {.lex_state = 110},
  [1084] = {.lex_state = 132},
  [1085] = {.lex_state = 105},
  [1086] = {.lex_state = 132},
  [1087] = {.lex_state = 132},
  [1088] = {.lex_state = 132},
  [1089] = {.lex_state = 132},
  [1090] = {.lex_state = 105},
  [1091] = {.lex_state = 105},
  [1092] = {.lex_state = 105},
  [1093] = {.lex_state = 105},
  [1094] = {.lex_state = 105},
  [1095] = {.lex_state = 105},
  [1096] = {.lex_state = 105},
  [1097] = {.lex_state = 105},
  [1098] = {.lex_state = 105},
  [1099] = {.lex_state = 105},
  [1100] = {.lex_state = 105},
  [1101] = {.lex_state = 105},
  [1102] = {.lex_state = 105},
  [1103] = {.lex_state = 105},
  [1104] = {.lex_state = 105},
  [1105] = {.lex_state = 105},
  [1106] = {.lex_state = 105},
  [1107] = {.lex_state = 105},
  [1108] = {.lex_state = 105},
  [1109] = {.lex_state = 105},
  [1110] = {.lex_state = 105},
  [1111] = {.lex_state = 105},
  [1112] = {.lex_state = 105},
  [1113] = {.lex_state = 105},
  [1114] = {.lex_state = 105},
  [1115] = {.lex_state = 105},
  [1116] = {.lex_state = 105},
  [1117] = {.lex_state = 105},
  [1118] = {.lex_state = 105},
  [1119] = {.lex_state = 105},
  [1120] = {.lex_state = 105},
  [1121] = {.lex_state = 105},
  [1122] = {.lex_state = 105},
  [1123] = {.lex_state = 105},
  [1124] = {.lex_state = 105},
  [1125] = {.lex_state = 105},
  [1126] = {.lex_state = 105},
  [1127] = {.lex_state = 105},
  [1128] = {.lex_state = 105},
  [1129] = {.lex_state = 105},
  [1130] = {.lex_state = 105},
  [1131] = {.lex_state = 105},
  [1132] = {.lex_state = 105},
  [1133] = {.lex_state = 105},
  [1134] = {.lex_state = 105},
  [1135] = {.lex_state = 105},
  [1136] = {.lex_state = 105},
  [1137] = {.lex_state = 105},
  [1138] = {.lex_state = 105},
  [1139] = {.lex_state = 105},
  [1140] = {.lex_state = 105},
  [1141] = {.lex_state = 105},
  [1142] = {.lex_state = 105},
  [1143] = {.lex_state = 105},
  [1144] = {.lex_state = 105},
  [1145] = {.lex_state = 105},
  [1146] = {.lex_state = 105},
  [1147] = {.lex_state = 105},
  [1148] = {.lex_state = 105},
  [1149] = {.lex_state = 105},
  [1150] = {.lex_state = 105},
  [1151] = {.lex_state = 105},
  [1152] = {.lex_state = 105},
  [1153] = {.lex_state = 105},
  [1154] = {.lex_state = 105},
  [1155] = {.lex_state = 105},
  [1156] = {.lex_state = 105},
  [1157] = {.lex_state = 105},
  [1158] = {.lex_state = 105},
  [1159] = {.lex_state = 105},
  [1160] = {.lex_state = 105},
  [1161] = {.lex_state = 105},
  [1162] = {.lex_state = 105},
  [1163] = {.lex_state = 105},
  [1164] = {.lex_state = 132},
  [1165] = {.lex_state = 132},
  [1166] = {.lex_state = 132},
  [1167] = {.lex_state = 132},
  [1168] = {.lex_state = 132},
  [1169] = {.lex_state = 132},
  [1170] = {.lex_state = 132},
  [1171] = {.lex_state = 132},
  [1172] = {.lex_state = 132},
  [1173] = {.lex_state = 132},
  [1174] = {.lex_state = 132},
  [1175] = {.lex_state = 132},
  [1176] = {.lex_state = 132},
  [1177] = {.lex_state = 132},
  [1178] = {.lex_state = 132},
  [1179] = {.lex_state = 132},
  [1180] = {.lex_state = 132},
  [1181] = {.lex_state = 132},
  [1182] = {.lex_state = 132},
  [1183] = {.lex_state = 132},
  [1184] = {.lex_state = 132},
  [1185] = {.lex_state = 126},
  [1186] = {.lex_state = 132},
  [1187] = {.lex_state = 126},
  [1188] = {.lex_state = 132},
  [1189] = {.lex_state = 126},
  [1190] = {.lex_state = 126},
  [1191] = {.lex_state = 126},
  [1192] = {.lex_state = 132},
  [1193] = {.lex_state = 126},
  [1194] = {.lex_state = 132},
  [1195] = {.lex_state = 126},
  [1196] = {.lex_state = 126},
  [1197] = {.lex_state = 132},
  [1198] = {.lex_state = 126},
  [1199] = {.lex_state = 132},
  [1200] = {.lex_state = 132},
  [1201] = {.lex_state = 132},
  [1202] = {.lex_state = 132},
  [1203] = {.lex_state = 132},
  [1204] = {.lex_state = 132},
  [1205] = {.lex_state = 132},
  [1206] = {.lex_state = 132},
  [1207] = {.lex_state = 132},
  [1208] = {.lex_state = 132},
  [1209] = {.lex_state = 132},
  [1210] = {.lex_state = 132},
  [1211] = {.lex_state = 132},
  [1212] = {.lex_state = 132},
  [1213] = {.lex_state = 126},
  [1214] = {.lex_state = 126},
  [1215] = {.lex_state = 132},
  [1216] = {.lex_state = 132},
  [1217] = {.lex_state = 126},
  [1218] = {.lex_state = 132},
  [1219] = {.lex_state = 132},
  [1220] = {.lex_state = 132},
  [1221] = {.lex_state = 132},
  [1222] = {.lex_state = 132},
  [1223] = {.lex_state = 132},
  [1224] = {.lex_state = 126},
  [1225] = {.lex_state = 132},
  [1226] = {.lex_state = 132},
  [1227] = {.lex_state = 132},
  [1228] = {.lex_state = 126},
  [1229] = {.lex_state = 132},
  [1230] = {.lex_state = 132},
  [1231] = {.lex_state = 132},
  [1232] = {.lex_state = 132},
  [1233] = {.lex_state = 132},
  [1234] = {.lex_state = 132},
  [1235] = {.lex_state = 132},
  [1236] = {.lex_state = 132},
  [1237] = {.lex_state = 132},
  [1238] = {.lex_state = 132},
  [1239] = {.lex_state = 132},
  [1240] = {.lex_state = 132},
  [1241] = {.lex_state = 132},
  [1242] = {.lex_state = 132},
  [1243] = {.lex_state = 132},
  [1244] = {.lex_state = 126},
  [1245] = {.lex_state = 132},
  [1246] = {.lex_state = 132},
  [1247] = {.lex_state = 132},
  [1248] = {.lex_state = 132},
  [1249] = {.lex_state = 132},
  [1250] = {.lex_state = 132},
  [1251] = {.lex_state = 126},
  [1252] = {.lex_state = 132},
  [1253] = {.lex_state = 132},
  [1254] = {.lex_state = 126},
  [1255] = {.lex_state = 132},
  [1256] = {.lex_state = 132},
  [1257] = {.lex_state = 132},
  [1258] = {.lex_state = 132},
  [1259] = {.lex_state = 133},
  [1260] = {.lex_state = 132},
  [1261] = {.lex_state = 132},
  [1262] = {.lex_state = 132},
  [1263] = {.lex_state = 132},
  [1264] = {.lex_state = 132},
  [1265] = {.lex_state = 132},
  [1266] = {.lex_state = 132},
  [1267] = {.lex_state = 133},
  [1268] = {.lex_state = 132},
  [1269] = {.lex_state = 132},
  [1270] = {.lex_state = 132},
  [1271] = {.lex_state = 132},
  [1272] = {.lex_state = 132},
  [1273] = {.lex_state = 132},
  [1274] = {.lex_state = 132},
  [1275] = {.lex_state = 132},
  [1276] = {.lex_state = 132},
  [1277] = {.lex_state = 132},
  [1278] = {.lex_state = 132},
  [1279] = {.lex_state = 132},
  [1280] = {.lex_state = 132},
  [1281] = {.lex_state = 132},
  [1282] = {.lex_state = 132},
  [1283] = {.lex_state = 132},
  [1284] = {.lex_state = 132},
  [1285] = {.lex_state = 132},
  [1286] = {.lex_state = 132},
  [1287] = {.lex_state = 132},
  [1288] = {.lex_state = 132},
  [1289] = {.lex_state = 133},
  [1290] = {.lex_state = 132},
  [1291] = {.lex_state = 132},
  [1292] = {.lex_state = 132},
  [1293] = {.lex_state = 132},
  [1294] = {.lex_state = 132},
  [1295] = {.lex_state = 132},
  [1296] = {.lex_state = 132},
  [1297] = {.lex_state = 132},
  [1298] = {.lex_state = 132},
  [1299] = {.lex_state = 132},
  [1300] = {.lex_state = 132},
  [1301] = {.lex_state = 132},
  [1302] = {.lex_state = 132},
  [1303] = {.lex_state = 132},
  [1304] = {.lex_state = 132},
  [1305] = {.lex_state = 132},
  [1306] = {.lex_state = 132},
  [1307] = {.lex_state = 132},
  [1308] = {.lex_state = 132},
  [1309] = {.lex_state = 132},
  [1310] = {.lex_state = 132},
  [1311] = {.lex_state = 132},
  [1312] = {.lex_state = 132},
  [1313] = {.lex_state = 132},
  [1314] = {.lex_state = 132},
  [1315] = {.lex_state = 132},
  [1316] = {.lex_state = 132},
  [1317] = {.lex_state = 132},
  [1318] = {.lex_state = 132},
  [1319] = {.lex_state = 132},
  [1320] = {.lex_state = 132},
  [1321] = {.lex_state = 132},
  [1322] = {.lex_state = 126},
  [1323] = {.lex_state = 132},
  [1324] = {.lex_state = 132},
  [1325] = {.lex_state = 132},
  [1326] = {.lex_state = 132},
  [1327] = {.lex_state = 132},
  [1328] = {.lex_state = 132},
  [1329] = {.lex_state = 132},
  [1330] = {.lex_state = 132},
  [1331] = {.lex_state = 132},
  [1332] = {.lex_state = 132},
  [1333] = {.lex_state = 132},
  [1334] = {.lex_state = 132},
  [1335] = {.lex_state = 132},
  [1336] = {.lex_state = 132},
  [1337] = {.lex_state = 132},
  [1338] = {.lex_state = 132},
  [1339] = {.lex_state = 132},
  [1340] = {.lex_state = 132},
  [1341] = {.lex_state = 132},
  [1342] = {.lex_state = 132},
  [1343] = {.lex_state = 132},
  [1344] = {.lex_state = 132},
  [1345] = {.lex_state = 132},
  [1346] = {.lex_state = 132},
  [1347] = {.lex_state = 132},
  [1348] = {.lex_state = 132},
  [1349] = {.lex_state = 132},
  [1350] = {.lex_state = 133},
  [1351] = {.lex_state = 132},
  [1352] = {.lex_state = 132},
  [1353] = {.lex_state = 132},
  [1354] = {.lex_state = 132},
  [1355] = {.lex_state = 132},
  [1356] = {.lex_state = 132},
  [1357] = {.lex_state = 132},
  [1358] = {.lex_state = 132},
  [1359] = {.lex_state = 132},
  [1360] = {.lex_state = 132},
  [1361] = {.lex_state = 132},
  [1362] = {.lex_state = 132},
  [1363] = {.lex_state = 132},
  [1364] = {.lex_state = 132},
  [1365] = {.lex_state = 132},
  [1366] = {.lex_state = 132},
  [1367] = {.lex_state = 132},
  [1368] = {.lex_state = 132},
  [1369] = {.lex_state = 132},
  [1370] = {.lex_state = 132},
  [1371] = {.lex_state = 132},
  [1372] = {.lex_state = 132},
  [1373] = {.lex_state = 132},
  [1374] = {.lex_state = 132},
  [1375] = {.lex_state = 132},
  [1376] = {.lex_state = 132},
  [1377] = {.lex_state = 132},
  [1378] = {.lex_state = 132},
  [1379] = {.lex_state = 132},
  [1380] = {.lex_state = 133},
  [1381] = {.lex_state = 132},
  [1382] = {.lex_state = 132},
  [1383] = {.lex_state = 133},
  [1384] = {.lex_state = 132},
  [1385] = {.lex_state = 132},
  [1386] = {.lex_state = 132},
  [1387] = {.lex_state = 132},
  [1388] = {.lex_state = 132},
  [1389] = {.lex_state = 132},
  [1390] = {.lex_state = 132},
  [1391] = {.lex_state = 132},
  [1392] = {.lex_state = 132},
  [1393] = {.lex_state = 132},
  [1394] = {.lex_state = 132},
  [1395] = {.lex_state = 132},
  [1396] = {.lex_state = 132},
  [1397] = {.lex_state = 132},
  [1398] = {.lex_state = 132},
  [1399] = {.lex_state = 132},
  [1400] = {.lex_state = 132},
  [1401] = {.lex_state = 132},
  [1402] = {.lex_state = 132},
  [1403] = {.lex_state = 132},
  [1404] = {.lex_state = 132},
  [1405] = {.lex_state = 132},
  [1406] = {.lex_state = 132},
  [1407] = {.lex_state = 132},
  [1408] = {.lex_state = 132},
  [1409] = {.lex_state = 132},
  [1410] = {.lex_state = 132},
  [1411] = {.lex_state = 132},
  [1412] = {.lex_state = 132},
  [1413] = {.lex_state = 132},
  [1414] = {.lex_state = 132},
  [1415] = {.lex_state = 132},
  [1416] = {.lex_state = 132},
  [1417] = {.lex_state = 132},
  [1418] = {.lex_state = 132},
  [1419] = {.lex_state = 132},
  [1420] = {.lex_state = 132},
  [1421] = {.lex_state = 132},
  [1422] = {.lex_state = 132},
  [1423] = {.lex_state = 132},
  [1424] = {.lex_state = 132},
  [1425] = {.lex_state = 132},
  [1426] = {.lex_state = 132},
  [1427] = {.lex_state = 132},
  [1428] = {.lex_state = 132},
  [1429] = {.lex_state = 132},
  [1430] = {.lex_state = 132},
  [1431] = {.lex_state = 132},
  [1432] = {.lex_state = 120},
  [1433] = {.lex_state = 120},
  [1434] = {.lex_state = 120},
  [1435] = {.lex_state = 124},
  [1436] = {.lex_state = 124},
  [1437] = {.lex_state = 124},
  [1438] = {.lex_state = 124},
  [1439] = {.lex_state = 124},
  [1440] = {.lex_state = 120},
  [1441] = {.lex_state = 124},
  [1442] = {.lex_state = 124},
  [1443] = {.lex_state = 124},
  [1444] = {.lex_state = 124},
  [1445] = {.lex_state = 120},
  [1446] = {.lex_state = 120},
  [1447] = {.lex_state = 124},
  [1448] = {.lex_state = 124},
  [1449] = {.lex_state = 124},
  [1450] = {.lex_state = 124},
  [1451] = {.lex_state = 124},
  [1452] = {.lex_state = 124},
  [1453] = {.lex_state = 124},
  [1454] = {.lex_state = 124},
  [1455] = {.lex_state = 124},
  [1456] = {.lex_state = 124},
  [1457] = {.lex_state = 124},
  [1458] = {.lex_state = 124},
  [1459] = {.lex_state = 124},
  [1460] = {.lex_state = 132},
  [1461] = {.lex_state = 124},
  [1462] = {.lex_state = 124},
  [1463] = {.lex_state = 124},
  [1464] = {.lex_state = 124},
  [1465] = {.lex_state = 124},
  [1466] = {.lex_state = 124},
  [1467] = {.lex_state = 124},
  [1468] = {.lex_state = 120},
  [1469] = {.lex_state = 124},
  [1470] = {.lex_state = 124},
  [1471] = {.lex_state = 124},
  [1472] = {.lex_state = 124},
  [1473] = {.lex_state = 124},
  [1474] = {.lex_state = 124},
  [1475] = {.lex_state = 124},
  [1476] = {.lex_state = 124},
  [1477] = {.lex_state = 124},
  [1478] = {.lex_state = 120},
  [1479] = {.lex_state = 124},
  [1480] = {.lex_state = 124},
  [1481] = {.lex_state = 124},
  [1482] = {.lex_state = 124},
  [1483] = {.lex_state = 124},
  [1484] = {.lex_state = 124},
  [1485] = {.lex_state = 124},
  [1486] = {.lex_state = 124},
  [1487] = {.lex_state = 124},
  [1488] = {.lex_state = 124},
  [1489] = {.lex_state = 124},
  [1490] = {.lex_state = 124},
  [1491] = {.lex_state = 124},
  [1492] = {.lex_state = 124},
  [1493] = {.lex_state = 124},
  [1494] = {.lex_state = 124},
  [1495] = {.lex_state = 124},
  [1496] = {.lex_state = 124},
  [1497] = {.lex_state = 124},
  [1498] = {.lex_state = 124},
  [1499] = {.lex_state = 124},
  [1500] = {.lex_state = 124},
  [1501] = {.lex_state = 124},
  [1502] = {.lex_state = 124},
  [1503] = {.lex_state = 132},
  [1504] = {.lex_state = 120},
  [1505] = {.lex_state = 124},
  [1506] = {.lex_state = 120},
  [1507] = {.lex_state = 120},
  [1508] = {.lex_state = 124},
  [1509] = {.lex_state = 124},
  [1510] = {.lex_state = 124},
  [1511] = {.lex_state = 124},
  [1512] = {.lex_state = 124},
  [1513] = {.lex_state = 125},
  [1514] = {.lex_state = 125},
  [1515] = {.lex_state = 125},
  [1516] = {.lex_state = 125},
  [1517] = {.lex_state = 125},
  [1518] = {.lex_state = 125},
  [1519] = {.lex_state = 125},
  [1520] = {.lex_state = 125},
  [1521] = {.lex_state = 125},
  [1522] = {.lex_state = 125},
  [1523] = {.lex_state = 125},
  [1524] = {.lex_state = 125},
  [1525] = {.lex_state = 125},
  [1526] = {.lex_state = 125},
  [1527] = {.lex_state = 125},
  [1528] = {.lex_state = 125},
  [1529] = {.lex_state = 112},
  [1530] = {.lex_state = 112},
  [1531] = {.lex_state = 112},
  [1532] = {.lex_state = 112},
  [1533] = {.lex_state = 112},
  [1534] = {.lex_state = 112},
  [1535] = {.lex_state = 112},
  [1536] = {.lex_state = 112},
  [1537] = {.lex_state = 112},
  [1538] = {.lex_state = 112},
  [1539] = {.lex_state = 112},
  [1540] = {.lex_state = 112},
  [1541] = {.lex_state = 112},
  [1542] = {.lex_state = 112},
  [1543] = {.lex_state = 112},
  [1544] = {.lex_state = 112},
  [1545] = {.lex_state = 112},
  [1546] = {.lex_state = 112},
  [1547] = {.lex_state = 112},
  [1548] = {.lex_state = 112},
  [1549] = {.lex_state = 112},
  [1550] = {.lex_state = 112},
  [1551] = {.lex_state = 112},
  [1552] = {.lex_state = 112},
  [1553] = {.lex_state = 124},
  [1554] = {.lex_state = 125},
  [1555] = {.lex_state = 124},
  [1556] = {.lex_state = 124},
  [1557] = {.lex_state = 124},
  [1558] = {.lex_state = 124},
  [1559] = {.lex_state = 124},
  [1560] = {.lex_state = 124},
  [1561] = {.lex_state = 124},
  [1562] = {.lex_state = 124},
  [1563] = {.lex_state = 124},
  [1564] = {.lex_state = 124},
  [1565] = {.lex_state = 124},
  [1566] = {.lex_state = 124},
  [1567] = {.lex_state = 124},
  [1568] = {.lex_state = 128},
  [1569] = {.lex_state = 128},
  [1570] = {.lex_state = 124},
  [1571] = {.lex_state = 128},
  [1572] = {.lex_state = 124},
  [1573] = {.lex_state = 132},
  [1574] = {.lex_state = 132},
  [1575] = {.lex_state = 132},
  [1576] = {.lex_state = 132},
  [1577] = {.lex_state = 111},
  [1578] = {.lex_state = 111},
  [1579] = {.lex_state = 117},
  [1580] = {.lex_state = 117},
  [1581] = {.lex_state = 132},
  [1582] = {.lex_state = 117},
  [1583] = {.lex_state = 117},
  [1584] = {.lex_state = 111},
  [1585] = {.lex_state = 117},
  [1586] = {.lex_state = 117},
  [1587] = {.lex_state = 117},
  [1588] = {.lex_state = 111},
  [1589] = {.lex_state = 117},
  [1590] = {.lex_state = 132},
  [1591] = {.lex_state = 132},
  [1592] = {.lex_state = 133},
  [1593] = {.lex_state = 123},
  [1594] = {.lex_state = 117},
  [1595] = {.lex_state = 117},
  [1596] = {.lex_state = 117},
  [1597] = {.lex_state = 117},
  [1598] = {.lex_state = 117},
  [1599] = {.lex_state = 123},
  [1600] = {.lex_state = 117},
  [1601] = {.lex_state = 123},
  [1602] = {.lex_state = 111},
  [1603] = {.lex_state = 132},
  [1604] = {.lex_state = 117},
  [1605] = {.lex_state = 111},
  [1606] = {.lex_state = 117},
  [1607] = {.lex_state = 117},
  [1608] = {.lex_state = 117},
  [1609] = {.lex_state = 117},
  [1610] = {.lex_state = 117},
  [1611] = {.lex_state = 132},
  [1612] = {.lex_state = 133},
  [1613] = {.lex_state = 111},
  [1614] = {.lex_state = 133},
  [1615] = {.lex_state = 128},
  [1616] = {.lex_state = 117},
  [1617] = {.lex_state = 128},
  [1618] = {.lex_state = 117},
  [1619] = {.lex_state = 128},
  [1620] = {.lex_state = 128},
  [1621] = {.lex_state = 117},
  [1622] = {.lex_state = 128},
  [1623] = {.lex_state = 128},
  [1624] = {.lex_state = 123},
  [1625] = {.lex_state = 128},
  [1626] = {.lex_state = 128},
  [1627] = {.lex_state = 117},
  [1628] = {.lex_state = 128},
  [1629] = {.lex_state = 128},
  [1630] = {.lex_state = 128},
  [1631] = {.lex_state = 128},
  [1632] = {.lex_state = 128},
  [1633] = {.lex_state = 128},
  [1634] = {.lex_state = 128},
  [1635] = {.lex_state = 128},
  [1636] = {.lex_state = 128},
  [1637] = {.lex_state = 128},
  [1638] = {.lex_state = 128},
  [1639] = {.lex_state = 128},
  [1640] = {.lex_state = 128},
  [1641] = {.lex_state = 128},
  [1642] = {.lex_state = 128},
  [1643] = {.lex_state = 128},
  [1644] = {.lex_state = 128},
  [1645] = {.lex_state = 128},
  [1646] = {.lex_state = 128},
  [1647] = {.lex_state = 128},
  [1648] = {.lex_state = 128},
  [1649] = {.lex_state = 128},
  [1650] = {.lex_state = 128},
  [1651] = {.lex_state = 128},
  [1652] = {.lex_state = 128},
  [1653] = {.lex_state = 123},
  [1654] = {.lex_state = 128},
  [1655] = {.lex_state = 128},
  [1656] = {.lex_state = 128},
  [1657] = {.lex_state = 123},
  [1658] = {.lex_state = 123},
  [1659] = {.lex_state = 128},
  [1660] = {.lex_state = 128},
  [1661] = {.lex_state = 128},
  [1662] = {.lex_state = 128},
  [1663] = {.lex_state = 128},
  [1664] = {.lex_state = 123},
  [1665] = {.lex_state = 128},
  [1666] = {.lex_state = 123},
  [1667] = {.lex_state = 117},
  [1668] = {.lex_state = 128},
  [1669] = {.lex_state = 117},
  [1670] = {.lex_state = 117},
  [1671] = {.lex_state = 117},
  [1672] = {.lex_state = 117},
  [1673] = {.lex_state = 117},
  [1674] = {.lex_state = 132},
  [1675] = {.lex_state = 132},
  [1676] = {.lex_state = 132},
  [1677] = {.lex_state = 132},
  [1678] = {.lex_state = 132},
  [1679] = {.lex_state = 132},
  [1680] = {.lex_state = 132},
  [1681] = {.lex_state = 132},
  [1682] = {.lex_state = 132},
  [1683] = {.lex_state = 132},
  [1684] = {.lex_state = 132},
  [1685] = {.lex_state = 132},
  [1686] = {.lex_state = 132},
  [1687] = {.lex_state = 132},
  [1688] = {.lex_state = 132},
  [1689] = {.lex_state = 132},
  [1690] = {.lex_state = 132},
  [1691] = {.lex_state = 132},
  [1692] = {.lex_state = 132},
  [1693] = {.lex_state = 132},
  [1694] = {.lex_state = 132},
  [1695] = {.lex_state = 132},
  [1696] = {.lex_state = 132},
  [1697] = {.lex_state = 132},
  [1698] = {.lex_state = 132},
  [1699] = {.lex_state = 132},
  [1700] = {.lex_state = 117},
  [1701] = {.lex_state = 132},
  [1702] = {.lex_state = 132},
  [1703] = {.lex_state = 132},
  [1704] = {.lex_state = 132},
  [1705] = {.lex_state = 132},
  [1706] = {.lex_state = 132},
  [1707] = {.lex_state = 132},
  [1708] = {.lex_state = 132},
  [1709] = {.lex_state = 132},
  [1710] = {.lex_state = 117},
  [1711] = {.lex_state = 132},
  [1712] = {.lex_state = 132},
  [1713] = {.lex_state = 132},
  [1714] = {.lex_state = 132},
  [1715] = {.lex_state = 132},
  [1716] = {.lex_state = 132},
  [1717] = {.lex_state = 132},
  [1718] = {.lex_state = 132},
  [1719] = {.lex_state = 132},
  [1720] = {.lex_state = 132},
  [1721] = {.lex_state = 132},
  [1722] = {.lex_state = 132},
  [1723] = {.lex_state = 132},
  [1724] = {.lex_state = 132},
  [1725] = {.lex_state = 132},
  [1726] = {.lex_state = 132},
  [1727] = {.lex_state = 117},
  [1728] = {.lex_state = 125},
  [1729] = {.lex_state = 132},
  [1730] = {.lex_state = 132},
  [1731] = {.lex_state = 132},
  [1732] = {.lex_state = 132},
  [1733] = {.lex_state = 125},
  [1734] = {.lex_state = 125},
  [1735] = {.lex_state = 132},
  [1736] = {.lex_state = 132},
  [1737] = {.lex_state = 132},
  [1738] = {.lex_state = 132},
  [1739] = {.lex_state = 132},
  [1740] = {.lex_state = 132},
  [1741] = {.lex_state = 125},
  [1742] = {.lex_state = 132},
  [1743] = {.lex_state = 125},
  [1744] = {.lex_state = 132},
  [1745] = {.lex_state = 125},
  [1746] = {.lex_state = 125},
  [1747] = {.lex_state = 125},
  [1748] = {.lex_state = 125},
  [1749] = {.lex_state = 125},
  [1750] = {.lex_state = 125},
  [1751] = {.lex_state = 125},
  [1752] = {.lex_state = 125},
  [1753] = {.lex_state = 132},
  [1754] = {.lex_state = 132},
  [1755] = {.lex_state = 132},
  [1756] = {.lex_state = 132},
  [1757] = {.lex_state = 132},
  [1758] = {.lex_state = 125},
  [1759] = {.lex_state = 125},
  [1760] = {.lex_state = 132},
  [1761] = {.lex_state = 132},
  [1762] = {.lex_state = 132},
  [1763] = {.lex_state = 125},
  [1764] = {.lex_state = 132},
  [1765] = {.lex_state = 125},
  [1766] = {.lex_state = 125},
  [1767] = {.lex_state = 132},
  [1768] = {.lex_state = 132},
  [1769] = {.lex_state = 132},
  [1770] = {.lex_state = 132},
  [1771] = {.lex_state = 132},
  [1772] = {.lex_state = 132},
  [1773] = {.lex_state = 132},
  [1774] = {.lex_state = 132},
  [1775] = {.lex_state = 125},
  [1776] = {.lex_state = 125},
  [1777] = {.lex_state = 125},
  [1778] = {.lex_state = 125},
  [1779] = {.lex_state = 132},
  [1780] = {.lex_state = 125},
  [1781] = {.lex_state = 132},
  [1782] = {.lex_state = 132},
  [1783] = {.lex_state = 132},
  [1784] = {.lex_state = 125},
  [1785] = {.lex_state = 125},
  [1786] = {.lex_state = 132},
  [1787] = {.lex_state = 132},
  [1788] = {.lex_state = 132},
  [1789] = {.lex_state = 132},
  [1790] = {.lex_state = 132},
  [1791] = {.lex_state = 132},
  [1792] = {.lex_state = 132},
  [1793] = {.lex_state = 132},
  [1794] = {.lex_state = 132},
  [1795] = {.lex_state = 125},
  [1796] = {.lex_state = 132},
  [1797] = {.lex_state = 132},
  [1798] = {.lex_state = 132},
  [1799] = {.lex_state = 132},
  [1800] = {.lex_state = 125},
  [1801] = {.lex_state = 125},
  [1802] = {.lex_state = 132},
  [1803] = {.lex_state = 132},
  [1804] = {.lex_state = 132},
  [1805] = {.lex_state = 125},
  [1806] = {.lex_state = 132},
  [1807] = {.lex_state = 132},
  [1808] = {.lex_state = 132},
  [1809] = {.lex_state = 132},
  [1810] = {.lex_state = 132},
  [1811] = {.lex_state = 132},
  [1812] = {.lex_state = 132},
  [1813] = {.lex_state = 132},
  [1814] = {.lex_state = 132},
  [1815] = {.lex_state = 132},
  [1816] = {.lex_state = 132},
  [1817] = {.lex_state = 132},
  [1818] = {.lex_state = 132},
  [1819] = {.lex_state = 132},
  [1820] = {.lex_state = 132},
  [1821] = {.lex_state = 132},
  [1822] = {.lex_state = 132},
  [1823] = {.lex_state = 132},
  [1824] = {.lex_state = 132},
  [1825] = {.lex_state = 117},
  [1826] = {.lex_state = 132},
  [1827] = {.lex_state = 132},
  [1828] = {.lex_state = 132},
  [1829] = {.lex_state = 132},
  [1830] = {.lex_state = 132},
  [1831] = {.lex_state = 132},
  [1832] = {.lex_state = 132},
  [1833] = {.lex_state = 132},
  [1834] = {.lex_state = 132},
  [1835] = {.lex_state = 132},
  [1836] = {.lex_state = 132},
  [1837] = {.lex_state = 132},
  [1838] = {.lex_state = 132},
  [1839] = {.lex_state = 132},
  [1840] = {.lex_state = 132},
  [1841] = {.lex_state = 132},
  [1842] = {.lex_state = 132},
  [1843] = {.lex_state = 132},
  [1844] = {.lex_state = 132},
  [1845] = {.lex_state = 132},
  [1846] = {.lex_state = 132},
  [1847] = {.lex_state = 132},
  [1848] = {.lex_state = 132},
  [1849] = {.lex_state = 132},
  [1850] = {.lex_state = 132},
  [1851] = {.lex_state = 132},
  [1852] = {.lex_state = 132},
  [1853] = {.lex_state = 132},
  [1854] = {.lex_state = 132},
  [1855] = {.lex_state = 132},
  [1856] = {.lex_state = 132},
  [1857] = {.lex_state = 132},
  [1858] = {.lex_state = 132},
  [1859] = {.lex_state = 132},
  [1860] = {.lex_state = 132},
  [1861] = {.lex_state = 132},
  [1862] = {.lex_state = 132},
  [1863] = {.lex_state = 132},
  [1864] = {.lex_state = 132},
  [1865] = {.lex_state = 132},
  [1866] = {.lex_state = 132},
  [1867] = {.lex_state = 132},
  [1868] = {.lex_state = 132},
  [1869] = {.lex_state = 132},
  [1870] = {.lex_state = 132},
  [1871] = {.lex_state = 132},
  [1872] = {.lex_state = 132},
  [1873] = {.lex_state = 132},
  [1874] = {.lex_state = 132},
  [1875] = {.lex_state = 132},
  [1876] = {.lex_state = 132},
  [1877] = {.lex_state = 132},
  [1878] = {.lex_state = 132},
  [1879] = {.lex_state = 132},
  [1880] = {.lex_state = 132},
  [1881] = {.lex_state = 132},
  [1882] = {.lex_state = 132},
  [1883] = {.lex_state = 132},
  [1884] = {.lex_state = 132},
  [1885] = {.lex_state = 132},
  [1886] = {.lex_state = 132},
  [1887] = {.lex_state = 132},
  [1888] = {.lex_state = 132},
  [1889] = {.lex_state = 132},
  [1890] = {.lex_state = 132},
  [1891] = {.lex_state = 132},
  [1892] = {.lex_state = 132},
  [1893] = {.lex_state = 132},
  [1894] = {.lex_state = 132},
  [1895] = {.lex_state = 132},
  [1896] = {.lex_state = 132},
  [1897] = {.lex_state = 132},
  [1898] = {.lex_state = 132},
  [1899] = {.lex_state = 132},
  [1900] = {.lex_state = 124},
  [1901] = {.lex_state = 124},
  [1902] = {.lex_state = 124},
  [1903] = {.lex_state = 124},
  [1904] = {.lex_state = 124},
  [1905] = {.lex_state = 124},
  [1906] = {.lex_state = 124},
  [1907] = {.lex_state = 124},
  [1908] = {.lex_state = 124},
  [1909] = {.lex_state = 124},
  [1910] = {.lex_state = 124},
  [1911] = {.lex_state = 124},
  [1912] = {.lex_state = 124},
  [1913] = {.lex_state = 124},
  [1914] = {.lex_state = 118},
  [1915] = {.lex_state = 118},
  [1916] = {.lex_state = 118},
  [1917] = {.lex_state = 124},
  [1918] = {.lex_state = 124},
  [1919] = {.lex_state = 132},
  [1920] = {.lex_state = 132},
  [1921] = {.lex_state = 132},
  [1922] = {.lex_state = 132},
  [1923] = {.lex_state = 132},
  [1924] = {.lex_state = 132},
  [1925] = {.lex_state = 132},
  [1926] = {.lex_state = 132},
  [1927] = {.lex_state = 132},
  [1928] = {.lex_state = 128},
  [1929] = {.lex_state = 128},
  [1930] = {.lex_state = 132},
  [1931] = {.lex_state = 132},
  [1932] = {.lex_state = 132},
  [1933] = {.lex_state = 132},
  [1934] = {.lex_state = 132},
  [1935] = {.lex_state = 132},
  [1936] = {.lex_state = 128},
  [1937] = {.lex_state = 132},
  [1938] = {.lex_state = 132},
  [1939] = {.lex_state = 132},
  [1940] = {.lex_state = 126},
  [1941] = {.lex_state = 132},
  [1942] = {.lex_state = 128},
  [1943] = {.lex_state = 132},
  [1944] = {.lex_state = 132},
  [1945] = {.lex_state = 132},
  [1946] = {.lex_state = 132},
  [1947] = {.lex_state = 132},
  [1948] = {.lex_state = 132},
  [1949] = {.lex_state = 128},
  [1950] = {.lex_state = 132},
  [1951] = {.lex_state = 132},
  [1952] = {.lex_state = 126},
  [1953] = {.lex_state = 132},
  [1954] = {.lex_state = 132},
  [1955] = {.lex_state = 132},
  [1956] = {.lex_state = 132},
  [1957] = {.lex_state = 132},
  [1958] = {.lex_state = 132},
  [1959] = {.lex_state = 132},
  [1960] = {.lex_state = 132},
  [1961] = {.lex_state = 132},
  [1962] = {.lex_state = 132},
  [1963] = {.lex_state = 132},
  [1964] = {.lex_state = 132},
  [1965] = {.lex_state = 132},
  [1966] = {.lex_state = 132},
  [1967] = {.lex_state = 128},
  [1968] = {.lex_state = 128},
  [1969] = {.lex_state = 132},
  [1970] = {.lex_state = 132},
  [1971] = {.lex_state = 132},
  [1972] = {.lex_state = 132},
  [1973] = {.lex_state = 132},
  [1974] = {.lex_state = 132},
  [1975] = {.lex_state = 132},
  [1976] = {.lex_state = 128},
  [1977] = {.lex_state = 128},
  [1978] = {.lex_state = 132},
  [1979] = {.lex_state = 132},
  [1980] = {.lex_state = 132},
  [1981] = {.lex_state = 128},
  [1982] = {.lex_state = 132},
  [1983] = {.lex_state = 112},
  [1984] = {.lex_state = 132},
  [1985] = {.lex_state = 132},
  [1986] = {.lex_state = 132},
  [1987] = {.lex_state = 132},
  [1988] = {.lex_state = 132},
  [1989] = {.lex_state = 132},
  [1990] = {.lex_state = 132},
  [1991] = {.lex_state = 132},
  [1992] = {.lex_state = 132},
  [1993] = {.lex_state = 132},
  [1994] = {.lex_state = 132},
  [1995] = {.lex_state = 132},
  [1996] = {.lex_state = 132},
  [1997] = {.lex_state = 132},
  [1998] = {.lex_state = 132},
  [1999] = {.lex_state = 132},
  [2000] = {.lex_state = 132},
  [2001] = {.lex_state = 132},
  [2002] = {.lex_state = 132},
  [2003] = {.lex_state = 132},
  [2004] = {.lex_state = 132},
  [2005] = {.lex_state = 132},
  [2006] = {.lex_state = 132},
  [2007] = {.lex_state = 132},
  [2008] = {.lex_state = 118},
  [2009] = {.lex_state = 132},
  [2010] = {.lex_state = 132},
  [2011] = {.lex_state = 132},
  [2012] = {.lex_state = 132},
  [2013] = {.lex_state = 132},
  [2014] = {.lex_state = 132},
  [2015] = {.lex_state = 132},
  [2016] = {.lex_state = 132},
  [2017] = {.lex_state = 132},
  [2018] = {.lex_state = 132},
  [2019] = {.lex_state = 132},
  [2020] = {.lex_state = 112},
  [2021] = {.lex_state = 132},
  [2022] = {.lex_state = 132},
  [2023] = {.lex_state = 132},
  [2024] = {.lex_state = 118},
  [2025] = {.lex_state = 132},
  [2026] = {.lex_state = 132},
  [2027] = {.lex_state = 132},
  [2028] = {.lex_state = 132},
  [2029] = {.lex_state = 118},
  [2030] = {.lex_state = 132},
  [2031] = {.lex_state = 132},
  [2032] = {.lex_state = 132},
  [2033] = {.lex_state = 118},
  [2034] = {.lex_state = 132},
  [2035] = {.lex_state = 132},
  [2036] = {.lex_state = 132},
  [2037] = {.lex_state = 132},
  [2038] = {.lex_state = 132},
  [2039] = {.lex_state = 112},
  [2040] = {.lex_state = 132},
  [2041] = {.lex_state = 112},
  [2042] = {.lex_state = 132},
  [2043] = {.lex_state = 112},
  [2044] = {.lex_state = 112},
  [2045] = {.lex_state = 112},
  [2046] = {.lex_state = 132},
  [2047] = {.lex_state = 132},
  [2048] = {.lex_state = 132},
  [2049] = {.lex_state = 132},
  [2050] = {.lex_state = 132},
  [2051] = {.lex_state = 132},
  [2052] = {.lex_state = 132},
  [2053] = {.lex_state = 132},
  [2054] = {.lex_state = 132},
  [2055] = {.lex_state = 132},
  [2056] = {.lex_state = 118},
  [2057] = {.lex_state = 132},
  [2058] = {.lex_state = 124},
  [2059] = {.lex_state = 134},
  [2060] = {.lex_state = 126},
  [2061] = {.lex_state = 126},
  [2062] = {.lex_state = 126},
  [2063] = {.lex_state = 133},
  [2064] = {.lex_state = 132},
  [2065] = {.lex_state = 132},
  [2066] = {.lex_state = 126},
  [2067] = {.lex_state = 132},
  [2068] = {.lex_state = 134},
  [2069] = {.lex_state = 118},
  [2070] = {.lex_state = 132},
  [2071] = {.lex_state = 132},
  [2072] = {.lex_state = 118},
  [2073] = {.lex_state = 133},
  [2074] = {.lex_state = 132},
  [2075] = {.lex_state = 134},
  [2076] = {.lex_state = 112},
  [2077] = {.lex_state = 132},
  [2078] = {.lex_state = 124},
  [2079] = {.lex_state = 132},
  [2080] = {.lex_state = 132},
  [2081] = {.lex_state = 132},
  [2082] = {.lex_state = 132},
  [2083] = {.lex_state = 132},
  [2084] = {.lex_state = 132},
  [2085] = {.lex_state = 111},
  [2086] = {.lex_state = 124},
  [2087] = {.lex_state = 124},
  [2088] = {.lex_state = 124},
  [2089] = {.lex_state = 124},
  [2090] = {.lex_state = 124},
  [2091] = {.lex_state = 124},
  [2092] = {.lex_state = 124},
  [2093] = {.lex_state = 124},
  [2094] = {.lex_state = 124},
  [2095] = {.lex_state = 124},
  [2096] = {.lex_state = 124},
  [2097] = {.lex_state = 124},
  [2098] = {.lex_state = 124},
  [2099] = {.lex_state = 111},
  [2100] = {.lex_state = 132},
  [2101] = {.lex_state = 126},
  [2102] = {.lex_state = 126},
  [2103] = {.lex_state = 126},
  [2104] = {.lex_state = 126},
  [2105] = {.lex_state = 126},
  [2106] = {.lex_state = 126},
  [2107] = {.lex_state = 126},
  [2108] = {.lex_state = 126},
  [2109] = {.lex_state = 126},
  [2110] = {.lex_state = 126},
  [2111] = {.lex_state = 126},
  [2112] = {.lex_state = 126},
  [2113] = {.lex_state = 126},
  [2114] = {.lex_state = 126},
  [2115] = {.lex_state = 126},
  [2116] = {.lex_state = 126},
  [2117] = {.lex_state = 132},
  [2118] = {.lex_state = 126},
  [2119] = {.lex_state = 126},
  [2120] = {.lex_state = 132},
  [2121] = {.lex_state = 132},
  [2122] = {.lex_state = 132},
  [2123] = {.lex_state = 132},
  [2124] = {.lex_state = 126},
  [2125] = {.lex_state = 126},
  [2126] = {.lex_state = 132},
  [2127] = {.lex_state = 132},
  [2128] = {.lex_state = 132},
  [2129] = {.lex_state = 132},
  [2130] = {.lex_state = 126},
  [2131] = {.lex_state = 132},
  [2132] = {.lex_state = 132},
  [2133] = {.lex_state = 126},
  [2134] = {.lex_state = 126},
  [2135] = {.lex_state = 132},
  [2136] = {.lex_state = 132},
  [2137] = {.lex_state = 132},
  [2138] = {.lex_state = 126},
  [2139] = {.lex_state = 132},
  [2140] = {.lex_state = 126},
  [2141] = {.lex_state = 132},
  [2142] = {.lex_state = 126},
  [2143] = {.lex_state = 132},
  [2144] = {.lex_state = 132},
  [2145] = {.lex_state = 126},
  [2146] = {.lex_state = 132},
  [2147] = {.lex_state = 126},
  [2148] = {.lex_state = 126},
  [2149] = {.lex_state = 126},
  [2150] = {.lex_state = 126},
  [2151] = {.lex_state = 132},
  [2152] = {.lex_state = 126},
  [2153] = {.lex_state = 126},
  [2154] = {.lex_state = 126},
  [2155] = {.lex_state = 126},
  [2156] = {.lex_state = 126},
  [2157] = {.lex_state = 126},
  [2158] = {.lex_state = 126},
  [2159] = {.lex_state = 126},
  [2160] = {.lex_state = 132},
  [2161] = {.lex_state = 132},
  [2162] = {.lex_state = 132},
  [2163] = {.lex_state = 126},
  [2164] = {.lex_state = 126},
  [2165] = {.lex_state = 126},
  [2166] = {.lex_state = 132},
  [2167] = {.lex_state = 132},
  [2168] = {.lex_state = 132},
  [2169] = {.lex_state = 126},
  [2170] = {.lex_state = 132},
  [2171] = {.lex_state = 132},
  [2172] = {.lex_state = 132},
  [2173] = {.lex_state = 132},
  [2174] = {.lex_state = 132},
  [2175] = {.lex_state = 132},
  [2176] = {.lex_state = 126},
  [2177] = {.lex_state = 132},
  [2178] = {.lex_state = 126},
  [2179] = {.lex_state = 132},
  [2180] = {.lex_state = 126},
  [2181] = {.lex_state = 124},
  [2182] = {.lex_state = 126},
  [2183] = {.lex_state = 126},
  [2184] = {.lex_state = 126},
  [2185] = {.lex_state = 132},
  [2186] = {.lex_state = 126},
  [2187] = {.lex_state = 126},
  [2188] = {.lex_state = 126},
  [2189] = {.lex_state = 132},
  [2190] = {.lex_state = 126},
  [2191] = {.lex_state = 132},
  [2192] = {.lex_state = 132},
  [2193] = {.lex_state = 126},
  [2194] = {.lex_state = 126},
  [2195] = {.lex_state = 126},
  [2196] = {.lex_state = 132},
  [2197] = {.lex_state = 132},
  [2198] = {.lex_state = 132},
  [2199] = {.lex_state = 126},
  [2200] = {.lex_state = 126},
  [2201] = {.lex_state = 132},
  [2202] = {.lex_state = 132},
  [2203] = {.lex_state = 126},
  [2204] = {.lex_state = 126},
  [2205] = {.lex_state = 126},
  [2206] = {.lex_state = 133},
  [2207] = {.lex_state = 132},
  [2208] = {.lex_state = 133},
  [2209] = {.lex_state = 133},
  [2210] = {.lex_state = 132},
  [2211] = {.lex_state = 133},
  [2212] = {.lex_state = 132},
  [2213] = {.lex_state = 126},
  [2214] = {.lex_state = 132},
  [2215] = {.lex_state = 133},
  [2216] = {.lex_state = 133},
  [2217] = {.lex_state = 126},
  [2218] = {.lex_state = 126},
  [2219] = {.lex_state = 132},
  [2220] = {.lex_state = 126},
  [2221] = {.lex_state = 126},
  [2222] = {.lex_state = 126},
  [2223] = {.lex_state = 126},
  [2224] = {.lex_state = 126},
  [2225] = {.lex_state = 126},
  [2226] = {.lex_state = 126},
  [2227] = {.lex_state = 126},
  [2228] = {.lex_state = 126},
  [2229] = {.lex_state = 126},
  [2230] = {.lex_state = 132},
  [2231] = {.lex_state = 133},
  [2232] = {.lex_state = 132},
  [2233] = {.lex_state = 126},
  [2234] = {.lex_state = 126},
  [2235] = {.lex_state = 126},
  [2236] = {.lex_state = 126},
  [2237] = {.lex_state = 126},
  [2238] = {.lex_state = 133},
  [2239] = {.lex_state = 126},
  [2240] = {.lex_state = 126},
  [2241] = {.lex_state = 126},
  [2242] = {.lex_state = 126},
  [2243] = {.lex_state = 133},
  [2244] = {.lex_state = 132},
  [2245] = {.lex_state = 132},
  [2246] = {.lex_state = 132},
  [2247] = {.lex_state = 133},
  [2248] = {.lex_state = 126},
  [2249] = {.lex_state = 132},
  [2250] = {.lex_state = 133},
  [2251] = {.lex_state = 132},
  [2252] = {.lex_state = 132},
  [2253] = {.lex_state = 133},
  [2254] = {.lex_state = 133},
  [2255] = {.lex_state = 133},
  [2256] = {.lex_state = 133},
  [2257] = {.lex_state = 132},
  [2258] = {.lex_state = 133},
  [2259] = {.lex_state = 133},
  [2260] = {.lex_state = 132},
  [2261] = {.lex_state = 133},
  [2262] = {.lex_state = 133},
  [2263] = {.lex_state = 133},
  [2264] = {.lex_state = 133},
  [2265] = {.lex_state = 133},
  [2266] = {.lex_state = 133},
  [2267] = {.lex_state = 126},
  [2268] = {.lex_state = 132},
  [2269] = {.lex_state = 126},
  [2270] = {.lex_state = 133},
  [2271] = {.lex_state = 132},
  [2272] = {.lex_state = 126},
  [2273] = {.lex_state = 133},
  [2274] = {.lex_state = 132},
  [2275] = {.lex_state = 132},
  [2276] = {.lex_state = 126},
  [2277] = {.lex_state = 132},
  [2278] = {.lex_state = 133},
  [2279] = {.lex_state = 133},
  [2280] = {.lex_state = 133},
  [2281] = {.lex_state = 133},
  [2282] = {.lex_state = 126},
  [2283] = {.lex_state = 132},
  [2284] = {.lex_state = 133},
  [2285] = {.lex_state = 133},
  [2286] = {.lex_state = 126},
  [2287] = {.lex_state = 133},
  [2288] = {.lex_state = 126},
  [2289] = {.lex_state = 132},
  [2290] = {.lex_state = 133},
  [2291] = {.lex_state = 126},
  [2292] = {.lex_state = 133},
  [2293] = {.lex_state = 133},
  [2294] = {.lex_state = 133},
  [2295] = {.lex_state = 133},
  [2296] = {.lex_state = 133},
  [2297] = {.lex_state = 133},
  [2298] = {.lex_state = 133},
  [2299] = {.lex_state = 132},
  [2300] = {.lex_state = 133},
  [2301] = {.lex_state = 132},
  [2302] = {.lex_state = 133},
  [2303] = {.lex_state = 132},
  [2304] = {.lex_state = 132},
  [2305] = {.lex_state = 133},
  [2306] = {.lex_state = 133},
  [2307] = {.lex_state = 132},
  [2308] = {.lex_state = 133},
  [2309] = {.lex_state = 132},
  [2310] = {.lex_state = 133},
  [2311] = {.lex_state = 133},
  [2312] = {.lex_state = 133},
  [2313] = {.lex_state = 133},
  [2314] = {.lex_state = 133},
  [2315] = {.lex_state = 133},
  [2316] = {.lex_state = 133},
  [2317] = {.lex_state = 132},
  [2318] = {.lex_state = 126},
  [2319] = {.lex_state = 132},
  [2320] = {.lex_state = 132},
  [2321] = {.lex_state = 133},
  [2322] = {.lex_state = 132},
  [2323] = {.lex_state = 126},
  [2324] = {.lex_state = 133},
  [2325] = {.lex_state = 133},
  [2326] = {.lex_state = 133},
  [2327] = {.lex_state = 132},
  [2328] = {.lex_state = 132},
  [2329] = {.lex_state = 133},
  [2330] = {.lex_state = 126},
  [2331] = {.lex_state = 133},
  [2332] = {.lex_state = 132},
  [2333] = {.lex_state = 132},
  [2334] = {.lex_state = 133},
  [2335] = {.lex_state = 132},
  [2336] = {.lex_state = 132},
  [2337] = {.lex_state = 126},
  [2338] = {.lex_state = 132},
  [2339] = {.lex_state = 133},
  [2340] = {.lex_state = 134},
  [2341] = {.lex_state = 133},
  [2342] = {.lex_state = 133},
  [2343] = {.lex_state = 133},
  [2344] = {.lex_state = 132},
  [2345] = {.lex_state = 133},
  [2346] = {.lex_state = 132},
  [2347] = {.lex_state = 133},
  [2348] = {.lex_state = 134},
  [2349] = {.lex_state = 134},
  [2350] = {.lex_state = 133},
  [2351] = {.lex_state = 133},
  [2352] = {.lex_state = 132},
  [2353] = {.lex_state = 134},
  [2354] = {.lex_state = 134},
  [2355] = {.lex_state = 133},
  [2356] = {.lex_state = 134},
  [2357] = {.lex_state = 134},
  [2358] = {.lex_state = 134},
  [2359] = {.lex_state = 132},
  [2360] = {.lex_state = 132},
  [2361] = {.lex_state = 133},
  [2362] = {.lex_state = 132},
  [2363] = {.lex_state = 133},
  [2364] = {.lex_state = 133},
  [2365] = {.lex_state = 132},
  [2366] = {.lex_state = 133},
  [2367] = {.lex_state = 132},
  [2368] = {.lex_state = 134},
  [2369] = {.lex_state = 133},
  [2370] = {.lex_state = 132},
  [2371] = {.lex_state = 132},
  [2372] = {.lex_state = 134},
  [2373] = {.lex_state = 133},
  [2374] = {.lex_state = 132},
  [2375] = {.lex_state = 133},
  [2376] = {.lex_state = 132},
  [2377] = {.lex_state = 133},
  [2378] = {.lex_state = 133},
  [2379] = {.lex_state = 133},
  [2380] = {.lex_state = 132},
  [2381] = {.lex_state = 134},
  [2382] = {.lex_state = 132},
  [2383] = {.lex_state = 133},
  [2384] = {.lex_state = 133},
  [2385] = {.lex_state = 132},
  [2386] = {.lex_state = 133},
  [2387] = {.lex_state = 132},
  [2388] = {.lex_state = 132},
  [2389] = {.lex_state = 132},
  [2390] = {.lex_state = 132},
  [2391] = {.lex_state = 133},
  [2392] = {.lex_state = 133},
  [2393] = {.lex_state = 133},
  [2394] = {.lex_state = 132},
  [2395] = {.lex_state = 133},
  [2396] = {.lex_state = 133},
  [2397] = {.lex_state = 132},
  [2398] = {.lex_state = 132},
  [2399] = {.lex_state = 133},
  [2400] = {.lex_state = 132},
  [2401] = {.lex_state = 134},
  [2402] = {.lex_state = 132},
  [2403] = {.lex_state = 133},
  [2404] = {.lex_state = 133},
  [2405] = {.lex_state = 132},
  [2406] = {.lex_state = 132},
  [2407] = {.lex_state = 132},
  [2408] = {.lex_state = 133},
  [2409] = {.lex_state = 133},
  [2410] = {.lex_state = 132},
  [2411] = {.lex_state = 133},
  [2412] = {.lex_state = 133},
  [2413] = {.lex_state = 132},
  [2414] = {.lex_state = 132},
  [2415] = {.lex_state = 133},
  [2416] = {.lex_state = 132},
  [2417] = {.lex_state = 133},
  [2418] = {.lex_state = 132},
  [2419] = {.lex_state = 112},
  [2420] = {.lex_state = 133},
  [2421] = {.lex_state = 133},
  [2422] = {.lex_state = 133},
  [2423] = {.lex_state = 133},
  [2424] = {.lex_state = 133},
  [2425] = {.lex_state = 105},
  [2426] = {.lex_state = 105},
  [2427] = {.lex_state = 105},
  [2428] = {.lex_state = 105},
  [2429] = {.lex_state = 105},
  [2430] = {.lex_state = 105},
  [2431] = {.lex_state = 105},
  [2432] = {.lex_state = 105},
  [2433] = {.lex_state = 105},
  [2434] = {.lex_state = 105},
  [2435] = {.lex_state = 105},
  [2436] = {.lex_state = 105},
  [2437] = {.lex_state = 105},
  [2438] = {.lex_state = 105},
  [2439] = {.lex_state = 105},
  [2440] = {.lex_state = 105},
  [2441] = {.lex_state = 105},
  [2442] = {.lex_state = 105},
  [2443] = {.lex_state = 105},
  [2444] = {.lex_state = 105},
  [2445] = {.lex_state = 105},
  [2446] = {.lex_state = 105},
  [2447] = {.lex_state = 105},
  [2448] = {.lex_state = 105},
  [2449] = {.lex_state = 105},
  [2450] = {.lex_state = 105},
  [2451] = {.lex_state = 105},
  [2452] = {.lex_state = 105},
  [2453] = {.lex_state = 105},
  [2454] = {.lex_state = 105},
  [2455] = {.lex_state = 105},
  [2456] = {.lex_state = 105},
  [2457] = {.lex_state = 105},
  [2458] = {.lex_state = 105},
  [2459] = {.lex_state = 105},
  [2460] = {.lex_state = 105},
  [2461] = {.lex_state = 105},
  [2462] = {.lex_state = 105},
  [2463] = {.lex_state = 105},
  [2464] = {.lex_state = 105},
  [2465] = {.lex_state = 105},
  [2466] = {.lex_state = 105},
  [2467] = {.lex_state = 105},
  [2468] = {.lex_state = 105},
  [2469] = {.lex_state = 105},
  [2470] = {.lex_state = 132},
  [2471] = {.lex_state = 105},
  [2472] = {.lex_state = 105},
  [2473] = {.lex_state = 105},
  [2474] = {.lex_state = 105},
  [2475] = {.lex_state = 105},
  [2476] = {.lex_state = 105},
  [2477] = {.lex_state = 105},
  [2478] = {.lex_state = 105},
  [2479] = {.lex_state = 105},
  [2480] = {.lex_state = 105},
  [2481] = {.lex_state = 132},
  [2482] = {.lex_state = 132},
  [2483] = {.lex_state = 132},
  [2484] = {.lex_state = 128},
  [2485] = {.lex_state = 128},
  [2486] = {.lex_state = 128},
  [2487] = {.lex_state = 132},
  [2488] = {.lex_state = 128},
  [2489] = {.lex_state = 128},
  [2490] = {.lex_state = 128},
  [2491] = {.lex_state = 128},
  [2492] = {.lex_state = 132},
  [2493] = {.lex_state = 128},
  [2494] = {.lex_state = 128},
  [2495] = {.lex_state = 128},
  [2496] = {.lex_state = 128},
  [2497] = {.lex_state = 128},
  [2498] = {.lex_state = 125},
  [2499] = {.lex_state = 128},
  [2500] = {.lex_state = 132},
  [2501] = {.lex_state = 132},
  [2502] = {.lex_state = 132},
  [2503] = {.lex_state = 132},
  [2504] = {.lex_state = 132},
  [2505] = {.lex_state = 128},
  [2506] = {.lex_state = 128},
  [2507] = {.lex_state = 132},
  [2508] = {.lex_state = 132},
  [2509] = {.lex_state = 132},
  [2510] = {.lex_state = 111},
  [2511] = {.lex_state = 132},
  [2512] = {.lex_state = 132},
  [2513] = {.lex_state = 128},
  [2514] = {.lex_state = 132},
  [2515] = {.lex_state = 132},
  [2516] = {.lex_state = 111},
  [2517] = {.lex_state = 111},
  [2518] = {.lex_state = 111},
  [2519] = {.lex_state = 132},
  [2520] = {.lex_state = 111},
  [2521] = {.lex_state = 132},
  [2522] = {.lex_state = 111},
  [2523] = {.lex_state = 132},
  [2524] = {.lex_state = 132},
  [2525] = {.lex_state = 111},
  [2526] = {.lex_state = 111},
  [2527] = {.lex_state = 111},
  [2528] = {.lex_state = 111},
  [2529] = {.lex_state = 111},
  [2530] = {.lex_state = 111},
  [2531] = {.lex_state = 111},
  [2532] = {.lex_state = 132},
  [2533] = {.lex_state = 132},
  [2534] = {.lex_state = 132},
  [2535] = {.lex_state = 111},
  [2536] = {.lex_state = 111},
  [2537] = {.lex_state = 111},
  [2538] = {.lex_state = 111},
  [2539] = {.lex_state = 132},
  [2540] = {.lex_state = 132},
  [2541] = {.lex_state = 125},
  [2542] = {.lex_state = 94},
  [2543] = {.lex_state = 111},
  [2544] = {.lex_state = 111},
  [2545] = {.lex_state = 132},
  [2546] = {.lex_state = 105},
  [2547] = {.lex_state = 105},
  [2548] = {.lex_state = 132},
  [2549] = {.lex_state = 105},
  [2550] = {.lex_state = 132},
  [2551] = {.lex_state = 132},
  [2552] = {.lex_state = 132},
  [2553] = {.lex_state = 132},
  [2554] = {.lex_state = 132},
  [2555] = {.lex_state = 132},
  [2556] = {.lex_state = 132},
  [2557] = {.lex_state = 111},
  [2558] = {.lex_state = 105},
  [2559] = {.lex_state = 111},
  [2560] = {.lex_state = 105},
  [2561] = {.lex_state = 105},
  [2562] = {.lex_state = 125},
  [2563] = {.lex_state = 105},
  [2564] = {.lex_state = 105},
  [2565] = {.lex_state = 132},
  [2566] = {.lex_state = 132},
  [2567] = {.lex_state = 105},
  [2568] = {.lex_state = 105},
  [2569] = {.lex_state = 111},
  [2570] = {.lex_state = 105},
  [2571] = {.lex_state = 105},
  [2572] = {.lex_state = 132},
  [2573] = {.lex_state = 132},
  [2574] = {.lex_state = 105},
  [2575] = {.lex_state = 105},
  [2576] = {.lex_state = 105},
  [2577] = {.lex_state = 105},
  [2578] = {.lex_state = 132},
  [2579] = {.lex_state = 132},
  [2580] = {.lex_state = 132},
  [2581] = {.lex_state = 132},
  [2582] = {.lex_state = 132},
  [2583] = {.lex_state = 132},
  [2584] = {.lex_state = 132},
  [2585] = {.lex_state = 132},
  [2586] = {.lex_state = 132},
  [2587] = {.lex_state = 125},
  [2588] = {.lex_state = 125},
  [2589] = {.lex_state = 125},
  [2590] = {.lex_state = 125},
  [2591] = {.lex_state = 125},
  [2592] = {.lex_state = 125},
  [2593] = {.lex_state = 125},
  [2594] = {.lex_state = 125},
  [2595] = {.lex_state = 125},
  [2596] = {.lex_state = 125},
  [2597] = {.lex_state = 125},
  [2598] = {.lex_state = 125},
  [2599] = {.lex_state = 125},
  [2600] = {.lex_state = 125},
  [2601] = {.lex_state = 125},
  [2602] = {.lex_state = 125},
  [2603] = {.lex_state = 125},
  [2604] = {.lex_state = 125},
  [2605] = {.lex_state = 125},
  [2606] = {.lex_state = 125},
  [2607] = {.lex_state = 125},
  [2608] = {.lex_state = 125},
  [2609] = {.lex_state = 125},
  [2610] = {.lex_state = 125},
  [2611] = {.lex_state = 125},
  [2612] = {.lex_state = 125},
  [2613] = {.lex_state = 125},
  [2614] = {.lex_state = 125},
  [2615] = {.lex_state = 125},
  [2616] = {.lex_state = 125},
  [2617] = {.lex_state = 125},
  [2618] = {.lex_state = 125},
  [2619] = {.lex_state = 125},
  [2620] = {.lex_state = 125},
  [2621] = {.lex_state = 125},
  [2622] = {.lex_state = 125},
  [2623] = {.lex_state = 111},
  [2624] = {.lex_state = 125},
  [2625] = {.lex_state = 125},
  [2626] = {.lex_state = 125},
  [2627] = {.lex_state = 111},
  [2628] = {.lex_state = 125},
  [2629] = {.lex_state = 125},
  [2630] = {.lex_state = 125},
  [2631] = {.lex_state = 125},
  [2632] = {.lex_state = 125},
  [2633] = {.lex_state = 125},
  [2634] = {.lex_state = 125},
  [2635] = {.lex_state = 125},
  [2636] = {.lex_state = 125},
  [2637] = {.lex_state = 125},
  [2638] = {.lex_state = 125},
  [2639] = {.lex_state = 125},
  [2640] = {.lex_state = 125},
  [2641] = {.lex_state = 125},
  [2642] = {.lex_state = 125},
  [2643] = {.lex_state = 125},
  [2644] = {.lex_state = 125},
  [2645] = {.lex_state = 125},
  [2646] = {.lex_state = 125},
  [2647] = {.lex_state = 125},
  [2648] = {.lex_state = 125},
  [2649] = {.lex_state = 125},
  [2650] = {.lex_state = 125},
  [2651] = {.lex_state = 125},
  [2652] = {.lex_state = 125},
  [2653] = {.lex_state = 125},
  [2654] = {.lex_state = 125},
  [2655] = {.lex_state = 130},
  [2656] = {.lex_state = 130},
  [2657] = {.lex_state = 130},
  [2658] = {.lex_state = 111},
  [2659] = {.lex_state = 111},
  [2660] = {.lex_state = 130},
  [2661] = {.lex_state = 130},
  [2662] = {.lex_state = 132},
  [2663] = {.lex_state = 132},
  [2664] = {.lex_state = 132},
  [2665] = {.lex_state = 132},
  [2666] = {.lex_state = 94},
  [2667] = {.lex_state = 132},
  [2668] = {.lex_state = 132},
  [2669] = {.lex_state = 132},
  [2670] = {.lex_state = 132},
  [2671] = {.lex_state = 132},
  [2672] = {.lex_state = 132},
  [2673] = {.lex_state = 132},
  [2674] = {.lex_state = 132},
  [2675] = {.lex_state = 132},
  [2676] = {.lex_state = 132},
  [2677] = {.lex_state = 132},
  [2678] = {.lex_state = 132},
  [2679] = {.lex_state = 132},
  [2680] = {.lex_state = 132},
  [2681] = {.lex_state = 132},
  [2682] = {.lex_state = 132},
  [2683] = {.lex_state = 94},
  [2684] = {.lex_state = 132},
  [2685] = {.lex_state = 133},
  [2686] = {.lex_state = 132},
  [2687] = {.lex_state = 132},
  [2688] = {.lex_state = 132},
  [2689] = {.lex_state = 133},
  [2690] = {.lex_state = 133},
  [2691] = {.lex_state = 133},
  [2692] = {.lex_state = 133},
  [2693] = {.lex_state = 133},
  [2694] = {.lex_state = 132},
  [2695] = {.lex_state = 132},
  [2696] = {.lex_state = 132},
  [2697] = {.lex_state = 132},
  [2698] = {.lex_state = 132},
  [2699] = {.lex_state = 132},
  [2700] = {.lex_state = 132},
  [2701] = {.lex_state = 132},
  [2702] = {.lex_state = 132},
  [2703] = {.lex_state = 132},
  [2704] = {.lex_state = 132},
  [2705] = {.lex_state = 132},
  [2706] = {.lex_state = 132},
  [2707] = {.lex_state = 132},
  [2708] = {.lex_state = 132},
  [2709] = {.lex_state = 133},
  [2710] = {.lex_state = 133},
  [2711] = {.lex_state = 133},
  [2712] = {.lex_state = 133},
  [2713] = {.lex_state = 133},
  [2714] = {.lex_state = 133},
  [2715] = {.lex_state = 133},
  [2716] = {.lex_state = 133},
  [2717] = {.lex_state = 133},
  [2718] = {.lex_state = 133},
  [2719] = {.lex_state = 132},
  [2720] = {.lex_state = 133},
  [2721] = {.lex_state = 133},
  [2722] = {.lex_state = 133},
  [2723] = {.lex_state = 133},
  [2724] = {.lex_state = 133},
  [2725] = {.lex_state = 133},
  [2726] = {.lex_state = 133},
  [2727] = {.lex_state = 133},
  [2728] = {.lex_state = 133},
  [2729] = {.lex_state = 133},
  [2730] = {.lex_state = 133},
  [2731] = {.lex_state = 133},
  [2732] = {.lex_state = 133},
  [2733] = {.lex_state = 133},
  [2734] = {.lex_state = 133},
  [2735] = {.lex_state = 133},
  [2736] = {.lex_state = 133},
  [2737] = {.lex_state = 133},
  [2738] = {.lex_state = 133},
  [2739] = {.lex_state = 133},
  [2740] = {.lex_state = 133},
  [2741] = {.lex_state = 133},
  [2742] = {.lex_state = 132},
  [2743] = {.lex_state = 132},
  [2744] = {.lex_state = 132},
  [2745] = {.lex_state = 133},
  [2746] = {.lex_state = 133},
  [2747] = {.lex_state = 132},
  [2748] = {.lex_state = 132},
  [2749] = {.lex_state = 132},
  [2750] = {.lex_state = 132},
  [2751] = {.lex_state = 132},
  [2752] = {.lex_state = 132},
  [2753] = {.lex_state = 132},
  [2754] = {.lex_state = 132},
  [2755] = {.lex_state = 132},
  [2756] = {.lex_state = 132},
  [2757] = {.lex_state = 132},
  [2758] = {.lex_state = 132},
  [2759] = {.lex_state = 132},
  [2760] = {.lex_state = 133},
  [2761] = {.lex_state = 132},
  [2762] = {.lex_state = 132},
  [2763] = {.lex_state = 132},
  [2764] = {.lex_state = 132},
  [2765] = {.lex_state = 132},
  [2766] = {.lex_state = 132},
  [2767] = {.lex_state = 132},
  [2768] = {.lex_state = 132},
  [2769] = {.lex_state = 132},
  [2770] = {.lex_state = 132},
  [2771] = {.lex_state = 132},
  [2772] = {.lex_state = 132},
  [2773] = {.lex_state = 132},
  [2774] = {.lex_state = 132},
  [2775] = {.lex_state = 132},
  [2776] = {.lex_state = 132},
  [2777] = {.lex_state = 132},
  [2778] = {.lex_state = 132},
  [2779] = {.lex_state = 132},
  [2780] = {.lex_state = 132},
  [2781] = {.lex_state = 132},
  [2782] = {.lex_state = 132},
  [2783] = {.lex_state = 132},
  [2784] = {.lex_state = 132},
  [2785] = {.lex_state = 132},
  [2786] = {.lex_state = 132},
  [2787] = {.lex_state = 132},
  [2788] = {.lex_state = 132},
  [2789] = {.lex_state = 132},
  [2790] = {.lex_state = 132},
  [2791] = {.lex_state = 132},
  [2792] = {.lex_state = 132},
  [2793] = {.lex_state = 132},
  [2794] = {.lex_state = 132},
  [2795] = {.lex_state = 132},
  [2796] = {.lex_state = 132},
  [2797] = {.lex_state = 132},
  [2798] = {.lex_state = 132},
  [2799] = {.lex_state = 132},
  [2800] = {.lex_state = 132},
  [2801] = {.lex_state = 132},
  [2802] = {.lex_state = 132},
  [2803] = {.lex_state = 132},
  [2804] = {.lex_state = 132},
  [2805] = {.lex_state = 132},
  [2806] = {.lex_state = 132},
  [2807] = {.lex_state = 132},
  [2808] = {.lex_state = 132},
  [2809] = {.lex_state = 132},
  [2810] = {.lex_state = 132},
  [2811] = {.lex_state = 132},
  [2812] = {.lex_state = 132},
  [2813] = {.lex_state = 132},
  [2814] = {.lex_state = 132},
  [2815] = {.lex_state = 132},
  [2816] = {.lex_state = 132},
  [2817] = {.lex_state = 132},
  [2818] = {.lex_state = 132},
  [2819] = {.lex_state = 132},
  [2820] = {.lex_state = 132},
  [2821] = {.lex_state = 132},
  [2822] = {.lex_state = 132},
  [2823] = {.lex_state = 132},
  [2824] = {.lex_state = 132},
  [2825] = {.lex_state = 132},
  [2826] = {.lex_state = 132},
  [2827] = {.lex_state = 132},
  [2828] = {.lex_state = 132},
  [2829] = {.lex_state = 132},
  [2830] = {.lex_state = 132},
  [2831] = {.lex_state = 132},
  [2832] = {.lex_state = 132},
  [2833] = {.lex_state = 132},
  [2834] = {.lex_state = 132},
  [2835] = {.lex_state = 132},
  [2836] = {.lex_state = 132},
  [2837] = {.lex_state = 132},
  [2838] = {.lex_state = 132},
  [2839] = {.lex_state = 132},
  [2840] = {.lex_state = 132},
  [2841] = {.lex_state = 132},
  [2842] = {.lex_state = 132},
  [2843] = {.lex_state = 132},
  [2844] = {.lex_state = 132},
  [2845] = {.lex_state = 132},
  [2846] = {.lex_state = 132},
  [2847] = {.lex_state = 132},
  [2848] = {.lex_state = 132},
  [2849] = {.lex_state = 132},
  [2850] = {.lex_state = 132},
  [2851] = {.lex_state = 132},
  [2852] = {.lex_state = 132},
  [2853] = {.lex_state = 132},
  [2854] = {.lex_state = 132},
  [2855] = {.lex_state = 132},
  [2856] = {.lex_state = 132},
  [2857] = {.lex_state = 132},
  [2858] = {.lex_state = 95},
  [2859] = {.lex_state = 132},
  [2860] = {.lex_state = 132},
  [2861] = {.lex_state = 132},
  [2862] = {.lex_state = 132},
  [2863] = {.lex_state = 132},
  [2864] = {.lex_state = 132},
  [2865] = {.lex_state = 132},
  [2866] = {.lex_state = 132},
  [2867] = {.lex_state = 132},
  [2868] = {.lex_state = 132},
  [2869] = {.lex_state = 132},
  [2870] = {.lex_state = 132},
  [2871] = {.lex_state = 132},
  [2872] = {.lex_state = 132},
  [2873] = {.lex_state = 132},
  [2874] = {.lex_state = 132},
  [2875] = {.lex_state = 132},
  [2876] = {.lex_state = 132},
  [2877] = {.lex_state = 132},
  [2878] = {.lex_state = 132},
  [2879] = {.lex_state = 132},
  [2880] = {.lex_state = 132},
  [2881] = {.lex_state = 132},
  [2882] = {.lex_state = 132},
  [2883] = {.lex_state = 132},
  [2884] = {.lex_state = 132},
  [2885] = {.lex_state = 132},
  [2886] = {.lex_state = 132},
  [2887] = {.lex_state = 132},
  [2888] = {.lex_state = 132},
  [2889] = {.lex_state = 132},
  [2890] = {.lex_state = 132},
  [2891] = {.lex_state = 132},
  [2892] = {.lex_state = 132},
  [2893] = {.lex_state = 132},
  [2894] = {.lex_state = 132},
  [2895] = {.lex_state = 132},
  [2896] = {.lex_state = 132},
  [2897] = {.lex_state = 132},
  [2898] = {.lex_state = 132},
  [2899] = {.lex_state = 132},
  [2900] = {.lex_state = 132},
  [2901] = {.lex_state = 132},
  [2902] = {.lex_state = 132},
  [2903] = {.lex_state = 132},
  [2904] = {.lex_state = 132},
  [2905] = {.lex_state = 132},
  [2906] = {.lex_state = 132},
  [2907] = {.lex_state = 132},
  [2908] = {.lex_state = 132},
  [2909] = {.lex_state = 132},
  [2910] = {.lex_state = 132},
  [2911] = {.lex_state = 132},
  [2912] = {.lex_state = 132},
  [2913] = {.lex_state = 124},
  [2914] = {.lex_state = 132},
  [2915] = {.lex_state = 132},
  [2916] = {.lex_state = 132},
  [2917] = {.lex_state = 132},
  [2918] = {.lex_state = 132},
  [2919] = {.lex_state = 132},
  [2920] = {.lex_state = 132},
  [2921] = {.lex_state = 132},
  [2922] = {.lex_state = 132},
  [2923] = {.lex_state = 124},
  [2924] = {.lex_state = 132},
  [2925] = {.lex_state = 132},
  [2926] = {.lex_state = 132},
  [2927] = {.lex_state = 132},
  [2928] = {.lex_state = 132},
  [2929] = {.lex_state = 132},
  [2930] = {.lex_state = 132},
  [2931] = {.lex_state = 132},
  [2932] = {.lex_state = 132},
  [2933] = {.lex_state = 132},
  [2934] = {.lex_state = 132},
  [2935] = {.lex_state = 132},
  [2936] = {.lex_state = 132},
  [2937] = {.lex_state = 132},
  [2938] = {.lex_state = 132},
  [2939] = {.lex_state = 132},
  [2940] = {.lex_state = 132},
  [2941] = {.lex_state = 132},
  [2942] = {.lex_state = 132},
  [2943] = {.lex_state = 132},
  [2944] = {.lex_state = 132},
  [2945] = {.lex_state = 132},
  [2946] = {.lex_state = 132},
  [2947] = {.lex_state = 132},
  [2948] = {.lex_state = 132},
  [2949] = {.lex_state = 132},
  [2950] = {.lex_state = 132},
  [2951] = {.lex_state = 132},
  [2952] = {.lex_state = 132},
  [2953] = {.lex_state = 132},
  [2954] = {.lex_state = 95},
  [2955] = {.lex_state = 132},
  [2956] = {.lex_state = 132},
  [2957] = {.lex_state = 132},
  [2958] = {.lex_state = 132},
  [2959] = {.lex_state = 132},
  [2960] = {.lex_state = 132},
  [2961] = {.lex_state = 132},
  [2962] = {.lex_state = 132},
  [2963] = {.lex_state = 132},
  [2964] = {.lex_state = 132},
  [2965] = {.lex_state = 132},
  [2966] = {.lex_state = 132},
  [2967] = {.lex_state = 132},
  [2968] = {.lex_state = 132},
  [2969] = {.lex_state = 132},
  [2970] = {.lex_state = 132},
  [2971] = {.lex_state = 132},
  [2972] = {.lex_state = 132},
  [2973] = {.lex_state = 132},
  [2974] = {.lex_state = 132},
  [2975] = {.lex_state = 132},
  [2976] = {.lex_state = 132},
  [2977] = {.lex_state = 132},
  [2978] = {.lex_state = 132},
  [2979] = {.lex_state = 132},
  [2980] = {.lex_state = 132},
  [2981] = {.lex_state = 132},
  [2982] = {.lex_state = 132},
  [2983] = {.lex_state = 132},
  [2984] = {.lex_state = 132},
  [2985] = {.lex_state = 132},
  [2986] = {.lex_state = 132},
  [2987] = {.lex_state = 132},
  [2988] = {.lex_state = 132},
  [2989] = {.lex_state = 132},
  [2990] = {.lex_state = 132},
  [2991] = {.lex_state = 132},
  [2992] = {.lex_state = 132},
  [2993] = {.lex_state = 132},
  [2994] = {.lex_state = 132},
  [2995] = {.lex_state = 132},
  [2996] = {.lex_state = 132},
  [2997] = {.lex_state = 132},
  [2998] = {.lex_state = 132},
  [2999] = {.lex_state = 132},
  [3000] = {.lex_state = 132},
  [3001] = {.lex_state = 132},
  [3002] = {.lex_state = 132},
  [3003] = {.lex_state = 132},
  [3004] = {.lex_state = 132},
  [3005] = {.lex_state = 132},
  [3006] = {.lex_state = 132},
  [3007] = {.lex_state = 132},
  [3008] = {.lex_state = 132},
  [3009] = {.lex_state = 95},
  [3010] = {.lex_state = 132},
  [3011] = {.lex_state = 132},
  [3012] = {.lex_state = 132},
  [3013] = {.lex_state = 132},
  [3014] = {.lex_state = 132},
  [3015] = {.lex_state = 132},
  [3016] = {.lex_state = 132},
  [3017] = {.lex_state = 132},
  [3018] = {.lex_state = 132},
  [3019] = {.lex_state = 132},
  [3020] = {.lex_state = 132},
  [3021] = {.lex_state = 132},
  [3022] = {.lex_state = 124},
  [3023] = {.lex_state = 132},
  [3024] = {.lex_state = 132},
  [3025] = {.lex_state = 132},
  [3026] = {.lex_state = 132},
  [3027] = {.lex_state = 132},
  [3028] = {.lex_state = 132},
  [3029] = {.lex_state = 132},
  [3030] = {.lex_state = 132},
  [3031] = {.lex_state = 132},
  [3032] = {.lex_state = 132},
  [3033] = {.lex_state = 132},
  [3034] = {.lex_state = 132},
  [3035] = {.lex_state = 132},
  [3036] = {.lex_state = 132},
  [3037] = {.lex_state = 132},
  [3038] = {.lex_state = 124},
  [3039] = {.lex_state = 132},
  [3040] = {.lex_state = 132},
  [3041] = {.lex_state = 132},
  [3042] = {.lex_state = 132},
  [3043] = {.lex_state = 132},
  [3044] = {.lex_state = 132},
  [3045] = {.lex_state = 132},
  [3046] = {.lex_state = 132},
  [3047] = {.lex_state = 132},
  [3048] = {.lex_state = 132},
  [3049] = {.lex_state = 132},
  [3050] = {.lex_state = 132},
  [3051] = {.lex_state = 132},
  [3052] = {.lex_state = 132},
  [3053] = {.lex_state = 132},
  [3054] = {.lex_state = 132},
  [3055] = {.lex_state = 132},
  [3056] = {.lex_state = 132},
  [3057] = {.lex_state = 132},
  [3058] = {.lex_state = 132},
  [3059] = {.lex_state = 132},
  [3060] = {.lex_state = 132},
  [3061] = {.lex_state = 132},
  [3062] = {.lex_state = 132},
  [3063] = {.lex_state = 132},
  [3064] = {.lex_state = 132},
  [3065] = {.lex_state = 132},
  [3066] = {.lex_state = 132},
  [3067] = {.lex_state = 132},
  [3068] = {.lex_state = 132},
  [3069] = {.lex_state = 132},
  [3070] = {.lex_state = 132},
  [3071] = {.lex_state = 132},
  [3072] = {.lex_state = 132},
  [3073] = {.lex_state = 132},
  [3074] = {.lex_state = 132},
  [3075] = {.lex_state = 132},
  [3076] = {.lex_state = 132},
  [3077] = {.lex_state = 132},
  [3078] = {.lex_state = 132},
  [3079] = {.lex_state = 132},
  [3080] = {.lex_state = 132},
  [3081] = {.lex_state = 132},
  [3082] = {.lex_state = 132},
  [3083] = {.lex_state = 132},
  [3084] = {.lex_state = 132},
  [3085] = {.lex_state = 132},
  [3086] = {.lex_state = 132},
  [3087] = {.lex_state = 132},
  [3088] = {.lex_state = 132},
  [3089] = {.lex_state = 132},
  [3090] = {.lex_state = 132},
  [3091] = {.lex_state = 132},
  [3092] = {.lex_state = 132},
  [3093] = {.lex_state = 132},
  [3094] = {.lex_state = 132},
  [3095] = {.lex_state = 132},
  [3096] = {.lex_state = 132},
  [3097] = {.lex_state = 132},
  [3098] = {.lex_state = 132},
  [3099] = {.lex_state = 132},
  [3100] = {.lex_state = 132},
  [3101] = {.lex_state = 132},
  [3102] = {.lex_state = 132},
  [3103] = {.lex_state = 132},
  [3104] = {.lex_state = 132},
  [3105] = {.lex_state = 132},
  [3106] = {.lex_state = 132},
  [3107] = {.lex_state = 132},
  [3108] = {.lex_state = 132},
  [3109] = {.lex_state = 132},
  [3110] = {.lex_state = 132},
  [3111] = {.lex_state = 132},
  [3112] = {.lex_state = 132},
  [3113] = {.lex_state = 132},
  [3114] = {.lex_state = 132},
  [3115] = {.lex_state = 132},
  [3116] = {.lex_state = 132},
  [3117] = {.lex_state = 132},
  [3118] = {.lex_state = 132},
  [3119] = {.lex_state = 132},
  [3120] = {.lex_state = 132},
  [3121] = {.lex_state = 132},
  [3122] = {.lex_state = 132},
  [3123] = {.lex_state = 132},
  [3124] = {.lex_state = 132},
  [3125] = {.lex_state = 132},
  [3126] = {.lex_state = 132},
  [3127] = {.lex_state = 132},
  [3128] = {.lex_state = 132},
  [3129] = {.lex_state = 132},
  [3130] = {.lex_state = 132},
  [3131] = {.lex_state = 132},
  [3132] = {.lex_state = 132},
  [3133] = {.lex_state = 132},
  [3134] = {.lex_state = 132},
  [3135] = {.lex_state = 132},
  [3136] = {.lex_state = 132},
  [3137] = {.lex_state = 132},
  [3138] = {.lex_state = 132},
  [3139] = {.lex_state = 132},
  [3140] = {.lex_state = 132},
  [3141] = {.lex_state = 132},
  [3142] = {.lex_state = 132},
  [3143] = {.lex_state = 132},
  [3144] = {.lex_state = 132},
  [3145] = {.lex_state = 132},
  [3146] = {.lex_state = 132},
  [3147] = {.lex_state = 132},
  [3148] = {.lex_state = 132},
  [3149] = {.lex_state = 132},
  [3150] = {.lex_state = 132},
  [3151] = {.lex_state = 132},
  [3152] = {.lex_state = 132},
  [3153] = {.lex_state = 132},
  [3154] = {.lex_state = 132},
  [3155] = {.lex_state = 132},
  [3156] = {.lex_state = 132},
  [3157] = {.lex_state = 132},
  [3158] = {.lex_state = 132},
  [3159] = {.lex_state = 132},
  [3160] = {.lex_state = 132},
  [3161] = {.lex_state = 132},
  [3162] = {.lex_state = 132},
  [3163] = {.lex_state = 132},
  [3164] = {.lex_state = 132},
  [3165] = {.lex_state = 132},
  [3166] = {.lex_state = 132},
  [3167] = {.lex_state = 132},
  [3168] = {.lex_state = 132},
  [3169] = {.lex_state = 132},
  [3170] = {.lex_state = 132},
  [3171] = {.lex_state = 132},
  [3172] = {.lex_state = 132},
  [3173] = {.lex_state = 132},
  [3174] = {.lex_state = 132},
  [3175] = {.lex_state = 132},
  [3176] = {.lex_state = 132},
  [3177] = {.lex_state = 132},
  [3178] = {.lex_state = 132},
  [3179] = {.lex_state = 132},
  [3180] = {.lex_state = 132},
  [3181] = {.lex_state = 132},
  [3182] = {.lex_state = 132},
  [3183] = {.lex_state = 132},
  [3184] = {.lex_state = 132},
  [3185] = {.lex_state = 132},
  [3186] = {.lex_state = 132},
  [3187] = {.lex_state = 132},
  [3188] = {.lex_state = 132},
  [3189] = {.lex_state = 132},
  [3190] = {.lex_state = 132},
  [3191] = {.lex_state = 132},
  [3192] = {.lex_state = 132},
  [3193] = {.lex_state = 132},
  [3194] = {.lex_state = 132},
  [3195] = {.lex_state = 132},
  [3196] = {.lex_state = 132},
  [3197] = {.lex_state = 132},
  [3198] = {.lex_state = 132},
  [3199] = {.lex_state = 132},
  [3200] = {.lex_state = 132},
  [3201] = {.lex_state = 132},
  [3202] = {.lex_state = 132},
  [3203] = {.lex_state = 132},
  [3204] = {.lex_state = 132},
  [3205] = {.lex_state = 132},
  [3206] = {.lex_state = 132},
  [3207] = {.lex_state = 95},
  [3208] = {.lex_state = 132},
  [3209] = {.lex_state = 132},
  [3210] = {.lex_state = 132},
  [3211] = {.lex_state = 132},
  [3212] = {.lex_state = 132},
  [3213] = {.lex_state = 95},
  [3214] = {.lex_state = 132},
  [3215] = {.lex_state = 132},
  [3216] = {.lex_state = 132},
  [3217] = {.lex_state = 132},
  [3218] = {.lex_state = 132},
  [3219] = {.lex_state = 132},
  [3220] = {.lex_state = 132},
  [3221] = {.lex_state = 132},
  [3222] = {.lex_state = 132},
  [3223] = {.lex_state = 132},
  [3224] = {.lex_state = 132},
  [3225] = {.lex_state = 132},
  [3226] = {.lex_state = 132},
  [3227] = {.lex_state = 132},
  [3228] = {.lex_state = 132},
  [3229] = {.lex_state = 132},
  [3230] = {.lex_state = 132},
  [3231] = {.lex_state = 132},
  [3232] = {.lex_state = 124},
  [3233] = {.lex_state = 132},
  [3234] = {.lex_state = 132},
  [3235] = {.lex_state = 132},
  [3236] = {.lex_state = 132},
  [3237] = {.lex_state = 132},
  [3238] = {.lex_state = 132},
  [3239] = {.lex_state = 132},
  [3240] = {.lex_state = 132},
  [3241] = {.lex_state = 132},
  [3242] = {.lex_state = 132},
  [3243] = {.lex_state = 132},
  [3244] = {.lex_state = 132},
  [3245] = {.lex_state = 132},
  [3246] = {.lex_state = 132},
  [3247] = {.lex_state = 132},
  [3248] = {.lex_state = 132},
  [3249] = {.lex_state = 132},
  [3250] = {.lex_state = 132},
  [3251] = {.lex_state = 132},
  [3252] = {.lex_state = 132},
  [3253] = {.lex_state = 132},
  [3254] = {.lex_state = 132},
  [3255] = {.lex_state = 132},
  [3256] = {.lex_state = 132},
  [3257] = {.lex_state = 132},
  [3258] = {.lex_state = 132},
  [3259] = {.lex_state = 132},
  [3260] = {.lex_state = 132},
  [3261] = {.lex_state = 132},
  [3262] = {.lex_state = 132},
  [3263] = {.lex_state = 132},
  [3264] = {.lex_state = 132},
  [3265] = {.lex_state = 132},
  [3266] = {.lex_state = 132},
  [3267] = {.lex_state = 132},
  [3268] = {.lex_state = 132},
  [3269] = {.lex_state = 111},
  [3270] = {.lex_state = 132},
  [3271] = {.lex_state = 124},
  [3272] = {.lex_state = 111},
  [3273] = {.lex_state = 132},
  [3274] = {.lex_state = 132},
  [3275] = {.lex_state = 132},
  [3276] = {.lex_state = 132},
  [3277] = {.lex_state = 124},
  [3278] = {.lex_state = 132},
  [3279] = {.lex_state = 124},
  [3280] = {.lex_state = 111},
  [3281] = {.lex_state = 111},
  [3282] = {.lex_state = 111},
  [3283] = {.lex_state = 132},
  [3284] = {.lex_state = 111},
  [3285] = {.lex_state = 132},
  [3286] = {.lex_state = 132},
  [3287] = {.lex_state = 132},
  [3288] = {.lex_state = 132},
  [3289] = {.lex_state = 132},
  [3290] = {.lex_state = 113},
  [3291] = {.lex_state = 113},
  [3292] = {.lex_state = 132},
  [3293] = {.lex_state = 111},
  [3294] = {.lex_state = 132},
  [3295] = {.lex_state = 132},
  [3296] = {.lex_state = 132},
  [3297] = {.lex_state = 132},
  [3298] = {.lex_state = 132},
  [3299] = {.lex_state = 132},
  [3300] = {.lex_state = 132},
  [3301] = {.lex_state = 132},
  [3302] = {.lex_state = 132},
  [3303] = {.lex_state = 132},
  [3304] = {.lex_state = 132},
  [3305] = {.lex_state = 132},
  [3306] = {.lex_state = 132},
  [3307] = {.lex_state = 132},
  [3308] = {.lex_state = 132},
  [3309] = {.lex_state = 132},
  [3310] = {.lex_state = 132},
  [3311] = {.lex_state = 132},
  [3312] = {.lex_state = 132},
  [3313] = {.lex_state = 132},
  [3314] = {.lex_state = 132},
  [3315] = {.lex_state = 132},
  [3316] = {.lex_state = 132},
  [3317] = {.lex_state = 132},
  [3318] = {.lex_state = 132},
  [3319] = {.lex_state = 112},
  [3320] = {.lex_state = 111},
  [3321] = {.lex_state = 112},
  [3322] = {.lex_state = 111},
  [3323] = {.lex_state = 111},
  [3324] = {.lex_state = 132},
  [3325] = {.lex_state = 111},
  [3326] = {.lex_state = 132},
  [3327] = {.lex_state = 132},
  [3328] = {.lex_state = 132},
  [3329] = {.lex_state = 117},
  [3330] = {.lex_state = 124},
  [3331] = {.lex_state = 111},
  [3332] = {.lex_state = 111},
  [3333] = {.lex_state = 111},
  [3334] = {.lex_state = 111},
  [3335] = {.lex_state = 111},
  [3336] = {.lex_state = 111},
  [3337] = {.lex_state = 111},
  [3338] = {.lex_state = 111},
  [3339] = {.lex_state = 111},
  [3340] = {.lex_state = 111},
  [3341] = {.lex_state = 111},
  [3342] = {.lex_state = 111},
  [3343] = {.lex_state = 111},
  [3344] = {.lex_state = 111},
  [3345] = {.lex_state = 111},
  [3346] = {.lex_state = 111},
  [3347] = {.lex_state = 111},
  [3348] = {.lex_state = 111},
  [3349] = {.lex_state = 111},
  [3350] = {.lex_state = 111},
  [3351] = {.lex_state = 111},
  [3352] = {.lex_state = 111},
  [3353] = {.lex_state = 111},
  [3354] = {.lex_state = 111},
  [3355] = {.lex_state = 111},
  [3356] = {.lex_state = 111},
  [3357] = {.lex_state = 111},
  [3358] = {.lex_state = 111},
  [3359] = {.lex_state = 111},
  [3360] = {.lex_state = 111},
  [3361] = {.lex_state = 111},
  [3362] = {.lex_state = 111},
  [3363] = {.lex_state = 111},
  [3364] = {.lex_state = 111},
  [3365] = {.lex_state = 111},
  [3366] = {.lex_state = 111},
  [3367] = {.lex_state = 111},
  [3368] = {.lex_state = 111},
  [3369] = {.lex_state = 111},
  [3370] = {.lex_state = 111},
  [3371] = {.lex_state = 111},
  [3372] = {.lex_state = 111},
  [3373] = {.lex_state = 111},
  [3374] = {.lex_state = 111},
  [3375] = {.lex_state = 111},
  [3376] = {.lex_state = 111},
  [3377] = {.lex_state = 111},
  [3378] = {.lex_state = 111},
  [3379] = {.lex_state = 111},
  [3380] = {.lex_state = 111},
  [3381] = {.lex_state = 111},
  [3382] = {.lex_state = 111},
  [3383] = {.lex_state = 111},
  [3384] = {.lex_state = 111},
  [3385] = {.lex_state = 111},
  [3386] = {.lex_state = 111},
  [3387] = {.lex_state = 111},
  [3388] = {.lex_state = 111},
  [3389] = {.lex_state = 111},
  [3390] = {.lex_state = 111},
  [3391] = {.lex_state = 111},
  [3392] = {.lex_state = 111},
  [3393] = {.lex_state = 111},
  [3394] = {.lex_state = 111},
  [3395] = {.lex_state = 111},
  [3396] = {.lex_state = 111},
  [3397] = {.lex_state = 111},
  [3398] = {.lex_state = 111},
  [3399] = {.lex_state = 111},
  [3400] = {.lex_state = 111},
  [3401] = {.lex_state = 111},
  [3402] = {.lex_state = 111},
  [3403] = {.lex_state = 111},
  [3404] = {.lex_state = 111},
  [3405] = {.lex_state = 111},
  [3406] = {.lex_state = 111},
  [3407] = {.lex_state = 111},
  [3408] = {.lex_state = 111},
  [3409] = {.lex_state = 111},
  [3410] = {.lex_state = 111},
  [3411] = {.lex_state = 111},
  [3412] = {.lex_state = 111},
  [3413] = {.lex_state = 111},
  [3414] = {.lex_state = 111},
  [3415] = {.lex_state = 111},
  [3416] = {.lex_state = 111},
  [3417] = {.lex_state = 111},
  [3418] = {.lex_state = 111},
  [3419] = {.lex_state = 111},
  [3420] = {.lex_state = 111},
  [3421] = {.lex_state = 111},
  [3422] = {.lex_state = 111},
  [3423] = {.lex_state = 111},
  [3424] = {.lex_state = 111},
  [3425] = {.lex_state = 111},
  [3426] = {.lex_state = 111},
  [3427] = {.lex_state = 111},
  [3428] = {.lex_state = 111},
  [3429] = {.lex_state = 111},
  [3430] = {.lex_state = 111},
  [3431] = {.lex_state = 111},
  [3432] = {.lex_state = 111},
  [3433] = {.lex_state = 111},
  [3434] = {.lex_state = 111},
  [3435] = {.lex_state = 111},
  [3436] = {.lex_state = 111},
  [3437] = {.lex_state = 111},
  [3438] = {.lex_state = 111},
  [3439] = {.lex_state = 111},
  [3440] = {.lex_state = 111},
  [3441] = {.lex_state = 111},
  [3442] = {.lex_state = 111},
  [3443] = {.lex_state = 111},
  [3444] = {.lex_state = 111},
  [3445] = {.lex_state = 111},
  [3446] = {.lex_state = 111},
  [3447] = {.lex_state = 111},
  [3448] = {.lex_state = 111},
  [3449] = {.lex_state = 111},
  [3450] = {.lex_state = 111},
  [3451] = {.lex_state = 111},
  [3452] = {.lex_state = 111},
  [3453] = {.lex_state = 111},
  [3454] = {.lex_state = 111},
  [3455] = {.lex_state = 111},
  [3456] = {.lex_state = 111},
  [3457] = {.lex_state = 111},
  [3458] = {.lex_state = 111},
  [3459] = {.lex_state = 111},
  [3460] = {.lex_state = 111},
  [3461] = {.lex_state = 111},
  [3462] = {.lex_state = 111},
  [3463] = {.lex_state = 111},
  [3464] = {.lex_state = 111},
  [3465] = {.lex_state = 111},
  [3466] = {.lex_state = 111},
  [3467] = {.lex_state = 111},
  [3468] = {.lex_state = 111},
  [3469] = {.lex_state = 111},
  [3470] = {.lex_state = 111},
  [3471] = {.lex_state = 111},
  [3472] = {.lex_state = 111},
  [3473] = {.lex_state = 111},
  [3474] = {.lex_state = 111},
  [3475] = {.lex_state = 111},
  [3476] = {.lex_state = 111},
  [3477] = {.lex_state = 111},
  [3478] = {.lex_state = 111},
  [3479] = {.lex_state = 111},
  [3480] = {.lex_state = 111},
  [3481] = {.lex_state = 111},
  [3482] = {.lex_state = 111},
  [3483] = {.lex_state = 111},
  [3484] = {.lex_state = 111},
  [3485] = {.lex_state = 111},
  [3486] = {.lex_state = 111},
  [3487] = {.lex_state = 111},
  [3488] = {.lex_state = 111},
  [3489] = {.lex_state = 111},
  [3490] = {.lex_state = 111},
  [3491] = {.lex_state = 111},
  [3492] = {.lex_state = 111},
  [3493] = {.lex_state = 111},
  [3494] = {.lex_state = 111},
  [3495] = {.lex_state = 111},
  [3496] = {.lex_state = 111},
  [3497] = {.lex_state = 111},
  [3498] = {.lex_state = 111},
  [3499] = {.lex_state = 111},
  [3500] = {.lex_state = 111},
  [3501] = {.lex_state = 111},
  [3502] = {.lex_state = 111},
  [3503] = {.lex_state = 111},
  [3504] = {.lex_state = 111},
  [3505] = {.lex_state = 111},
  [3506] = {.lex_state = 111},
  [3507] = {.lex_state = 111},
  [3508] = {.lex_state = 111},
  [3509] = {.lex_state = 111},
  [3510] = {.lex_state = 111},
  [3511] = {.lex_state = 111},
  [3512] = {.lex_state = 111},
  [3513] = {.lex_state = 111},
  [3514] = {.lex_state = 111},
  [3515] = {.lex_state = 111},
  [3516] = {.lex_state = 111},
  [3517] = {.lex_state = 111},
  [3518] = {.lex_state = 111},
  [3519] = {.lex_state = 111},
  [3520] = {.lex_state = 111},
  [3521] = {.lex_state = 111},
  [3522] = {.lex_state = 111},
  [3523] = {.lex_state = 111},
  [3524] = {.lex_state = 111},
  [3525] = {.lex_state = 111},
  [3526] = {.lex_state = 111},
  [3527] = {.lex_state = 111},
  [3528] = {.lex_state = 111},
  [3529] = {.lex_state = 111},
  [3530] = {.lex_state = 111},
  [3531] = {.lex_state = 111},
  [3532] = {.lex_state = 111},
  [3533] = {.lex_state = 111},
  [3534] = {.lex_state = 111},
  [3535] = {.lex_state = 111},
  [3536] = {.lex_state = 111},
  [3537] = {.lex_state = 111},
  [3538] = {.lex_state = 111},
  [3539] = {.lex_state = 111},
  [3540] = {.lex_state = 111},
  [3541] = {.lex_state = 111},
  [3542] = {.lex_state = 111},
  [3543] = {.lex_state = 111},
  [3544] = {.lex_state = 111},
  [3545] = {.lex_state = 111},
  [3546] = {.lex_state = 111},
  [3547] = {.lex_state = 111},
  [3548] = {.lex_state = 111},
  [3549] = {.lex_state = 111},
  [3550] = {.lex_state = 111},
  [3551] = {.lex_state = 111},
  [3552] = {.lex_state = 111},
  [3553] = {.lex_state = 111},
  [3554] = {.lex_state = 111},
  [3555] = {.lex_state = 111},
  [3556] = {.lex_state = 111},
  [3557] = {.lex_state = 111},
  [3558] = {.lex_state = 111},
  [3559] = {.lex_state = 111},
  [3560] = {.lex_state = 111},
  [3561] = {.lex_state = 111},
  [3562] = {.lex_state = 111},
  [3563] = {.lex_state = 111},
  [3564] = {.lex_state = 111},
  [3565] = {.lex_state = 111},
  [3566] = {.lex_state = 111},
  [3567] = {.lex_state = 111},
  [3568] = {.lex_state = 111},
  [3569] = {.lex_state = 111},
  [3570] = {.lex_state = 111},
  [3571] = {.lex_state = 111},
  [3572] = {.lex_state = 111},
  [3573] = {.lex_state = 111},
  [3574] = {.lex_state = 111},
  [3575] = {.lex_state = 111},
  [3576] = {.lex_state = 111},
  [3577] = {.lex_state = 111},
  [3578] = {.lex_state = 111},
  [3579] = {.lex_state = 111},
  [3580] = {.lex_state = 111},
  [3581] = {.lex_state = 111},
  [3582] = {.lex_state = 111},
  [3583] = {.lex_state = 111},
  [3584] = {.lex_state = 111},
  [3585] = {.lex_state = 111},
  [3586] = {.lex_state = 111},
  [3587] = {.lex_state = 111},
  [3588] = {.lex_state = 111},
  [3589] = {.lex_state = 111},
  [3590] = {.lex_state = 111},
  [3591] = {.lex_state = 111},
  [3592] = {.lex_state = 111},
  [3593] = {.lex_state = 111},
  [3594] = {.lex_state = 111},
  [3595] = {.lex_state = 111},
  [3596] = {.lex_state = 111},
  [3597] = {.lex_state = 111},
  [3598] = {.lex_state = 111},
  [3599] = {.lex_state = 111},
  [3600] = {.lex_state = 111},
  [3601] = {.lex_state = 111},
  [3602] = {.lex_state = 111},
  [3603] = {.lex_state = 111},
  [3604] = {.lex_state = 111},
  [3605] = {.lex_state = 111},
  [3606] = {.lex_state = 111},
  [3607] = {.lex_state = 111},
  [3608] = {.lex_state = 111},
  [3609] = {.lex_state = 111},
  [3610] = {.lex_state = 111},
  [3611] = {.lex_state = 111},
  [3612] = {.lex_state = 111},
  [3613] = {.lex_state = 111},
  [3614] = {.lex_state = 111},
  [3615] = {.lex_state = 111},
  [3616] = {.lex_state = 111},
  [3617] = {.lex_state = 111},
  [3618] = {.lex_state = 111},
  [3619] = {.lex_state = 111},
  [3620] = {.lex_state = 111},
  [3621] = {.lex_state = 111},
  [3622] = {.lex_state = 111},
  [3623] = {.lex_state = 111},
  [3624] = {.lex_state = 111},
  [3625] = {.lex_state = 111},
  [3626] = {.lex_state = 111},
  [3627] = {.lex_state = 111},
  [3628] = {.lex_state = 111},
  [3629] = {.lex_state = 111},
  [3630] = {.lex_state = 111},
  [3631] = {.lex_state = 111},
  [3632] = {.lex_state = 111},
  [3633] = {.lex_state = 111},
  [3634] = {.lex_state = 111},
  [3635] = {.lex_state = 111},
  [3636] = {.lex_state = 111},
  [3637] = {.lex_state = 111},
  [3638] = {.lex_state = 111},
  [3639] = {.lex_state = 111},
  [3640] = {.lex_state = 111},
  [3641] = {.lex_state = 111},
  [3642] = {.lex_state = 111},
  [3643] = {.lex_state = 111},
  [3644] = {.lex_state = 111},
  [3645] = {.lex_state = 111},
  [3646] = {.lex_state = 111},
  [3647] = {.lex_state = 111},
  [3648] = {.lex_state = 111},
  [3649] = {.lex_state = 111},
  [3650] = {.lex_state = 111},
  [3651] = {.lex_state = 111},
  [3652] = {.lex_state = 111},
  [3653] = {.lex_state = 111},
  [3654] = {.lex_state = 111},
  [3655] = {.lex_state = 111},
  [3656] = {.lex_state = 111},
  [3657] = {.lex_state = 111},
  [3658] = {.lex_state = 111},
  [3659] = {.lex_state = 111},
  [3660] = {.lex_state = 111},
  [3661] = {.lex_state = 111},
  [3662] = {.lex_state = 111},
  [3663] = {.lex_state = 111},
  [3664] = {.lex_state = 111},
  [3665] = {.lex_state = 111},
  [3666] = {.lex_state = 111},
  [3667] = {.lex_state = 111},
  [3668] = {.lex_state = 111},
  [3669] = {.lex_state = 111},
  [3670] = {.lex_state = 111},
  [3671] = {.lex_state = 111},
  [3672] = {.lex_state = 111},
  [3673] = {.lex_state = 111},
  [3674] = {.lex_state = 111},
  [3675] = {.lex_state = 111},
  [3676] = {.lex_state = 111},
  [3677] = {.lex_state = 111},
  [3678] = {.lex_state = 111},
  [3679] = {.lex_state = 111},
  [3680] = {.lex_state = 111},
  [3681] = {.lex_state = 111},
  [3682] = {.lex_state = 111},
  [3683] = {.lex_state = 111},
  [3684] = {.lex_state = 111},
  [3685] = {.lex_state = 111},
  [3686] = {.lex_state = 111},
  [3687] = {.lex_state = 111},
  [3688] = {.lex_state = 111},
  [3689] = {.lex_state = 111},
  [3690] = {.lex_state = 111},
  [3691] = {.lex_state = 111},
  [3692] = {.lex_state = 111},
  [3693] = {.lex_state = 111},
  [3694] = {.lex_state = 111},
  [3695] = {.lex_state = 111},
  [3696] = {.lex_state = 111},
  [3697] = {.lex_state = 111},
  [3698] = {.lex_state = 111},
  [3699] = {.lex_state = 111},
  [3700] = {.lex_state = 111},
  [3701] = {.lex_state = 111},
  [3702] = {.lex_state = 111},
  [3703] = {.lex_state = 111},
  [3704] = {.lex_state = 111},
  [3705] = {.lex_state = 111},
  [3706] = {.lex_state = 111},
  [3707] = {.lex_state = 111},
  [3708] = {.lex_state = 111},
  [3709] = {.lex_state = 111},
  [3710] = {.lex_state = 111},
  [3711] = {.lex_state = 111},
  [3712] = {.lex_state = 111},
  [3713] = {.lex_state = 111},
  [3714] = {.lex_state = 111},
  [3715] = {.lex_state = 111},
  [3716] = {.lex_state = 111},
  [3717] = {.lex_state = 111},
  [3718] = {.lex_state = 111},
  [3719] = {.lex_state = 111},
  [3720] = {.lex_state = 111},
  [3721] = {.lex_state = 111},
  [3722] = {.lex_state = 111},
  [3723] = {.lex_state = 111},
  [3724] = {.lex_state = 111},
  [3725] = {.lex_state = 111},
  [3726] = {.lex_state = 111},
  [3727] = {.lex_state = 111},
  [3728] = {.lex_state = 111},
  [3729] = {.lex_state = 111},
  [3730] = {.lex_state = 111},
  [3731] = {.lex_state = 111},
  [3732] = {.lex_state = 111},
  [3733] = {.lex_state = 111},
  [3734] = {.lex_state = 111},
  [3735] = {.lex_state = 111},
  [3736] = {.lex_state = 111},
  [3737] = {.lex_state = 111},
  [3738] = {.lex_state = 111},
  [3739] = {.lex_state = 111},
  [3740] = {.lex_state = 111},
  [3741] = {.lex_state = 111},
  [3742] = {.lex_state = 111},
  [3743] = {.lex_state = 111},
  [3744] = {.lex_state = 111},
  [3745] = {.lex_state = 111},
  [3746] = {.lex_state = 111},
  [3747] = {.lex_state = 132},
  [3748] = {.lex_state = 132},
  [3749] = {.lex_state = 132},
  [3750] = {.lex_state = 132},
  [3751] = {.lex_state = 132},
  [3752] = {.lex_state = 132},
  [3753] = {.lex_state = 132},
  [3754] = {.lex_state = 132},
  [3755] = {.lex_state = 132},
  [3756] = {.lex_state = 132},
  [3757] = {.lex_state = 132},
  [3758] = {.lex_state = 132},
  [3759] = {.lex_state = 125},
  [3760] = {.lex_state = 134},
  [3761] = {.lex_state = 132},
  [3762] = {.lex_state = 134},
  [3763] = {.lex_state = 134},
  [3764] = {.lex_state = 132},
  [3765] = {.lex_state = 132},
  [3766] = {.lex_state = 132},
  [3767] = {.lex_state = 132},
  [3768] = {.lex_state = 132},
  [3769] = {.lex_state = 132},
  [3770] = {.lex_state = 132},
  [3771] = {.lex_state = 132},
  [3772] = {.lex_state = 132},
  [3773] = {.lex_state = 132},
  [3774] = {.lex_state = 132},
  [3775] = {.lex_state = 132},
  [3776] = {.lex_state = 132},
  [3777] = {.lex_state = 132},
  [3778] = {.lex_state = 132},
  [3779] = {.lex_state = 132},
  [3780] = {.lex_state = 132},
  [3781] = {.lex_state = 132},
  [3782] = {.lex_state = 132},
  [3783] = {.lex_state = 132},
  [3784] = {.lex_state = 132},
  [3785] = {.lex_state = 132},
  [3786] = {.lex_state = 132},
  [3787] = {.lex_state = 132},
  [3788] = {.lex_state = 132},
  [3789] = {.lex_state = 132},
  [3790] = {.lex_state = 132},
  [3791] = {.lex_state = 132},
  [3792] = {.lex_state = 132},
  [3793] = {.lex_state = 132},
  [3794] = {.lex_state = 132},
  [3795] = {.lex_state = 132},
  [3796] = {.lex_state = 132},
  [3797] = {.lex_state = 132},
  [3798] = {.lex_state = 132},
  [3799] = {.lex_state = 132},
  [3800] = {.lex_state = 132},
  [3801] = {.lex_state = 132},
  [3802] = {.lex_state = 132},
  [3803] = {.lex_state = 132},
  [3804] = {.lex_state = 132},
  [3805] = {.lex_state = 132},
  [3806] = {.lex_state = 132},
  [3807] = {.lex_state = 132},
  [3808] = {.lex_state = 132},
  [3809] = {.lex_state = 132},
  [3810] = {.lex_state = 132},
  [3811] = {.lex_state = 132},
  [3812] = {.lex_state = 132},
  [3813] = {.lex_state = 132},
  [3814] = {.lex_state = 132},
  [3815] = {.lex_state = 132},
  [3816] = {.lex_state = 132},
  [3817] = {.lex_state = 132},
  [3818] = {.lex_state = 132},
  [3819] = {.lex_state = 132},
  [3820] = {.lex_state = 132},
  [3821] = {.lex_state = 132},
  [3822] = {.lex_state = 132},
  [3823] = {.lex_state = 132},
  [3824] = {.lex_state = 132},
  [3825] = {.lex_state = 132},
  [3826] = {.lex_state = 132},
  [3827] = {.lex_state = 132},
  [3828] = {.lex_state = 132},
  [3829] = {.lex_state = 132},
  [3830] = {.lex_state = 132},
  [3831] = {.lex_state = 132},
  [3832] = {.lex_state = 132},
  [3833] = {.lex_state = 132},
  [3834] = {.lex_state = 132},
  [3835] = {.lex_state = 132},
  [3836] = {.lex_state = 132},
  [3837] = {.lex_state = 132},
  [3838] = {.lex_state = 132},
  [3839] = {.lex_state = 132},
  [3840] = {.lex_state = 132},
  [3841] = {.lex_state = 132},
  [3842] = {.lex_state = 132},
  [3843] = {.lex_state = 132},
  [3844] = {.lex_state = 132},
  [3845] = {.lex_state = 95},
  [3846] = {.lex_state = 132},
  [3847] = {.lex_state = 132},
  [3848] = {.lex_state = 132},
  [3849] = {.lex_state = 132},
  [3850] = {.lex_state = 132},
  [3851] = {.lex_state = 132},
  [3852] = {.lex_state = 132},
  [3853] = {.lex_state = 132},
  [3854] = {.lex_state = 132},
  [3855] = {.lex_state = 132},
  [3856] = {.lex_state = 132},
  [3857] = {.lex_state = 132},
  [3858] = {.lex_state = 132},
  [3859] = {.lex_state = 132},
  [3860] = {.lex_state = 132},
  [3861] = {.lex_state = 132},
  [3862] = {.lex_state = 132},
  [3863] = {.lex_state = 132},
  [3864] = {.lex_state = 132},
  [3865] = {.lex_state = 132},
  [3866] = {.lex_state = 132},
  [3867] = {.lex_state = 132},
  [3868] = {.lex_state = 132},
  [3869] = {.lex_state = 132},
  [3870] = {.lex_state = 132},
  [3871] = {.lex_state = 132},
  [3872] = {.lex_state = 132},
  [3873] = {.lex_state = 132},
  [3874] = {.lex_state = 132},
  [3875] = {.lex_state = 132},
  [3876] = {.lex_state = 132},
  [3877] = {.lex_state = 132},
  [3878] = {.lex_state = 132},
  [3879] = {.lex_state = 132},
  [3880] = {.lex_state = 132},
  [3881] = {.lex_state = 132},
  [3882] = {.lex_state = 132},
  [3883] = {.lex_state = 132},
  [3884] = {.lex_state = 132},
  [3885] = {.lex_state = 132},
  [3886] = {.lex_state = 132},
  [3887] = {.lex_state = 132},
  [3888] = {.lex_state = 132},
  [3889] = {.lex_state = 132},
  [3890] = {.lex_state = 132},
  [3891] = {.lex_state = 132},
  [3892] = {.lex_state = 132},
  [3893] = {.lex_state = 132},
  [3894] = {.lex_state = 132},
  [3895] = {.lex_state = 132},
  [3896] = {.lex_state = 132},
  [3897] = {.lex_state = 132},
  [3898] = {.lex_state = 132},
  [3899] = {.lex_state = 132},
  [3900] = {.lex_state = 132},
  [3901] = {.lex_state = 132},
  [3902] = {.lex_state = 132},
  [3903] = {.lex_state = 132},
  [3904] = {.lex_state = 132},
  [3905] = {.lex_state = 132},
  [3906] = {.lex_state = 132},
  [3907] = {.lex_state = 132},
  [3908] = {.lex_state = 132},
  [3909] = {.lex_state = 132},
  [3910] = {.lex_state = 132},
  [3911] = {.lex_state = 132},
  [3912] = {.lex_state = 132},
  [3913] = {.lex_state = 132},
  [3914] = {.lex_state = 132},
  [3915] = {.lex_state = 132},
  [3916] = {.lex_state = 132},
  [3917] = {.lex_state = 132},
  [3918] = {.lex_state = 132},
  [3919] = {.lex_state = 132},
  [3920] = {.lex_state = 132},
  [3921] = {.lex_state = 132},
  [3922] = {.lex_state = 132},
  [3923] = {.lex_state = 132},
  [3924] = {.lex_state = 132},
  [3925] = {.lex_state = 132},
  [3926] = {.lex_state = 132},
  [3927] = {.lex_state = 132},
  [3928] = {.lex_state = 132},
  [3929] = {.lex_state = 132},
  [3930] = {.lex_state = 132},
  [3931] = {.lex_state = 132},
  [3932] = {.lex_state = 132},
  [3933] = {.lex_state = 132},
  [3934] = {.lex_state = 132},
  [3935] = {.lex_state = 132},
  [3936] = {.lex_state = 132},
  [3937] = {.lex_state = 132},
  [3938] = {.lex_state = 132},
  [3939] = {.lex_state = 132},
  [3940] = {.lex_state = 132},
  [3941] = {.lex_state = 132},
  [3942] = {.lex_state = 132},
  [3943] = {.lex_state = 132},
  [3944] = {.lex_state = 132},
  [3945] = {.lex_state = 132},
  [3946] = {.lex_state = 132},
  [3947] = {.lex_state = 132},
  [3948] = {.lex_state = 132},
  [3949] = {.lex_state = 132},
  [3950] = {.lex_state = 132},
  [3951] = {.lex_state = 132},
  [3952] = {.lex_state = 132},
  [3953] = {.lex_state = 132},
  [3954] = {.lex_state = 132},
  [3955] = {.lex_state = 132},
  [3956] = {.lex_state = 132},
  [3957] = {.lex_state = 132},
  [3958] = {.lex_state = 132},
  [3959] = {.lex_state = 132},
  [3960] = {.lex_state = 132},
  [3961] = {.lex_state = 132},
  [3962] = {.lex_state = 132},
  [3963] = {.lex_state = 132},
  [3964] = {.lex_state = 132},
  [3965] = {.lex_state = 132},
  [3966] = {.lex_state = 132},
  [3967] = {.lex_state = 132},
  [3968] = {.lex_state = 132},
  [3969] = {.lex_state = 132},
  [3970] = {.lex_state = 132},
  [3971] = {.lex_state = 132},
  [3972] = {.lex_state = 132},
  [3973] = {.lex_state = 132},
  [3974] = {.lex_state = 132},
  [3975] = {.lex_state = 132},
  [3976] = {.lex_state = 132},
  [3977] = {.lex_state = 132},
  [3978] = {.lex_state = 132},
  [3979] = {.lex_state = 132},
  [3980] = {.lex_state = 132},
  [3981] = {.lex_state = 132},
  [3982] = {.lex_state = 132},
  [3983] = {.lex_state = 132},
  [3984] = {.lex_state = 132},
  [3985] = {.lex_state = 132},
  [3986] = {.lex_state = 132},
  [3987] = {.lex_state = 132},
  [3988] = {.lex_state = 132},
  [3989] = {.lex_state = 132},
  [3990] = {.lex_state = 132},
  [3991] = {.lex_state = 132},
  [3992] = {.lex_state = 132},
  [3993] = {.lex_state = 132},
  [3994] = {.lex_state = 132},
  [3995] = {.lex_state = 132},
  [3996] = {.lex_state = 132},
  [3997] = {.lex_state = 132},
  [3998] = {.lex_state = 132},
  [3999] = {.lex_state = 132},
  [4000] = {.lex_state = 132},
  [4001] = {.lex_state = 132},
  [4002] = {.lex_state = 132},
  [4003] = {.lex_state = 132},
  [4004] = {.lex_state = 132},
  [4005] = {.lex_state = 132},
  [4006] = {.lex_state = 132},
  [4007] = {.lex_state = 132},
  [4008] = {.lex_state = 132},
  [4009] = {.lex_state = 132},
  [4010] = {.lex_state = 132},
  [4011] = {.lex_state = 132},
  [4012] = {.lex_state = 132},
  [4013] = {.lex_state = 132},
  [4014] = {.lex_state = 132},
  [4015] = {.lex_state = 132},
  [4016] = {.lex_state = 132},
  [4017] = {.lex_state = 132},
  [4018] = {.lex_state = 132},
  [4019] = {.lex_state = 132},
  [4020] = {.lex_state = 132},
  [4021] = {.lex_state = 132},
  [4022] = {.lex_state = 132},
  [4023] = {.lex_state = 132},
  [4024] = {.lex_state = 132},
  [4025] = {.lex_state = 132},
  [4026] = {.lex_state = 132},
  [4027] = {.lex_state = 132},
  [4028] = {.lex_state = 132},
  [4029] = {.lex_state = 132},
  [4030] = {.lex_state = 132},
  [4031] = {.lex_state = 132},
  [4032] = {.lex_state = 132},
  [4033] = {.lex_state = 132},
  [4034] = {.lex_state = 132},
  [4035] = {.lex_state = 132},
  [4036] = {.lex_state = 132},
  [4037] = {.lex_state = 132},
  [4038] = {.lex_state = 132},
  [4039] = {.lex_state = 132},
  [4040] = {.lex_state = 132},
  [4041] = {.lex_state = 132},
  [4042] = {.lex_state = 132},
  [4043] = {.lex_state = 132},
  [4044] = {.lex_state = 132},
  [4045] = {.lex_state = 132},
  [4046] = {.lex_state = 132},
  [4047] = {.lex_state = 124},
  [4048] = {.lex_state = 124},
  [4049] = {.lex_state = 124},
  [4050] = {.lex_state = 124},
  [4051] = {.lex_state = 124},
  [4052] = {.lex_state = 124},
  [4053] = {.lex_state = 132},
  [4054] = {.lex_state = 124},
  [4055] = {.lex_state = 132},
  [4056] = {.lex_state = 124},
  [4057] = {.lex_state = 124},
  [4058] = {.lex_state = 124},
  [4059] = {.lex_state = 124},
  [4060] = {.lex_state = 124},
  [4061] = {.lex_state = 124},
  [4062] = {.lex_state = 132},
  [4063] = {.lex_state = 124},
  [4064] = {.lex_state = 132},
  [4065] = {.lex_state = 132},
  [4066] = {.lex_state = 132},
  [4067] = {.lex_state = 132},
  [4068] = {.lex_state = 132},
  [4069] = {.lex_state = 132},
  [4070] = {.lex_state = 124},
  [4071] = {.lex_state = 105},
  [4072] = {.lex_state = 105},
  [4073] = {.lex_state = 105},
  [4074] = {.lex_state = 132},
  [4075] = {.lex_state = 132},
  [4076] = {.lex_state = 132},
  [4077] = {.lex_state = 132},
  [4078] = {.lex_state = 132},
  [4079] = {.lex_state = 132},
  [4080] = {.lex_state = 132},
  [4081] = {.lex_state = 132},
  [4082] = {.lex_state = 132},
  [4083] = {.lex_state = 132},
  [4084] = {.lex_state = 132},
  [4085] = {.lex_state = 132},
  [4086] = {.lex_state = 132},
  [4087] = {.lex_state = 132},
  [4088] = {.lex_state = 132},
  [4089] = {.lex_state = 132},
  [4090] = {.lex_state = 132},
  [4091] = {.lex_state = 132},
  [4092] = {.lex_state = 132},
  [4093] = {.lex_state = 132},
  [4094] = {.lex_state = 132},
  [4095] = {.lex_state = 132},
  [4096] = {.lex_state = 132},
  [4097] = {.lex_state = 132},
  [4098] = {.lex_state = 132},
  [4099] = {.lex_state = 132},
  [4100] = {.lex_state = 132},
  [4101] = {.lex_state = 132},
  [4102] = {.lex_state = 132},
  [4103] = {.lex_state = 132},
  [4104] = {.lex_state = 132},
  [4105] = {.lex_state = 124},
  [4106] = {.lex_state = 124},
  [4107] = {.lex_state = 132},
  [4108] = {.lex_state = 132},
  [4109] = {.lex_state = 132},
  [4110] = {.lex_state = 132},
  [4111] = {.lex_state = 132},
  [4112] = {.lex_state = 132},
  [4113] = {.lex_state = 132},
  [4114] = {.lex_state = 132},
  [4115] = {.lex_state = 132},
  [4116] = {.lex_state = 132},
  [4117] = {.lex_state = 132},
  [4118] = {.lex_state = 132},
  [4119] = {.lex_state = 132},
  [4120] = {.lex_state = 132},
  [4121] = {.lex_state = 132},
  [4122] = {.lex_state = 111},
  [4123] = {.lex_state = 111},
  [4124] = {.lex_state = 111},
  [4125] = {.lex_state = 111},
  [4126] = {.lex_state = 111},
  [4127] = {.lex_state = 111},
  [4128] = {.lex_state = 111},
  [4129] = {.lex_state = 132},
  [4130] = {.lex_state = 132},
  [4131] = {.lex_state = 132},
  [4132] = {.lex_state = 132},
  [4133] = {.lex_state = 132},
  [4134] = {.lex_state = 132},
  [4135] = {.lex_state = 132},
  [4136] = {.lex_state = 132},
  [4137] = {.lex_state = 132},
  [4138] = {.lex_state = 132},
  [4139] = {.lex_state = 132},
  [4140] = {.lex_state = 132},
  [4141] = {.lex_state = 132},
  [4142] = {.lex_state = 132},
  [4143] = {.lex_state = 132},
  [4144] = {.lex_state = 132},
  [4145] = {.lex_state = 132},
  [4146] = {.lex_state = 132},
  [4147] = {.lex_state = 132},
  [4148] = {.lex_state = 132},
  [4149] = {.lex_state = 111},
  [4150] = {.lex_state = 111},
  [4151] = {.lex_state = 132},
  [4152] = {.lex_state = 132},
  [4153] = {.lex_state = 132},
  [4154] = {.lex_state = 132},
  [4155] = {.lex_state = 132},
  [4156] = {.lex_state = 132},
  [4157] = {.lex_state = 132},
  [4158] = {.lex_state = 132},
  [4159] = {.lex_state = 132},
  [4160] = {.lex_state = 132},
  [4161] = {.lex_state = 132},
  [4162] = {.lex_state = 132},
  [4163] = {.lex_state = 132},
  [4164] = {.lex_state = 132},
  [4165] = {.lex_state = 132},
  [4166] = {.lex_state = 132},
  [4167] = {.lex_state = 132},
  [4168] = {.lex_state = 132},
  [4169] = {.lex_state = 111},
  [4170] = {.lex_state = 111},
  [4171] = {.lex_state = 111},
  [4172] = {.lex_state = 132},
  [4173] = {.lex_state = 111},
  [4174] = {.lex_state = 111},
  [4175] = {.lex_state = 111},
  [4176] = {.lex_state = 111},
  [4177] = {.lex_state = 111},
  [4178] = {.lex_state = 111},
  [4179] = {.lex_state = 132},
  [4180] = {.lex_state = 132},
  [4181] = {.lex_state = 132},
  [4182] = {.lex_state = 132},
  [4183] = {.lex_state = 111},
  [4184] = {.lex_state = 132},
  [4185] = {.lex_state = 132},
  [4186] = {.lex_state = 132},
  [4187] = {.lex_state = 132},
  [4188] = {.lex_state = 111},
  [4189] = {.lex_state = 132},
  [4190] = {.lex_state = 111},
  [4191] = {.lex_state = 132},
  [4192] = {.lex_state = 111},
  [4193] = {.lex_state = 132},
  [4194] = {.lex_state = 111},
  [4195] = {.lex_state = 132},
  [4196] = {.lex_state = 132},
  [4197] = {.lex_state = 111},
  [4198] = {.lex_state = 111},
  [4199] = {.lex_state = 132},
  [4200] = {.lex_state = 132},
  [4201] = {.lex_state = 132},
  [4202] = {.lex_state = 132},
  [4203] = {.lex_state = 132},
  [4204] = {.lex_state = 132},
  [4205] = {.lex_state = 132},
  [4206] = {.lex_state = 132},
  [4207] = {.lex_state = 132},
  [4208] = {.lex_state = 132},
  [4209] = {.lex_state = 132},
  [4210] = {.lex_state = 132},
  [4211] = {.lex_state = 132},
  [4212] = {.lex_state = 132},
  [4213] = {.lex_state = 132},
  [4214] = {.lex_state = 132},
  [4215] = {.lex_state = 132},
  [4216] = {.lex_state = 132},
  [4217] = {.lex_state = 132},
  [4218] = {.lex_state = 132},
  [4219] = {.lex_state = 132},
  [4220] = {.lex_state = 117},
  [4221] = {.lex_state = 132},
  [4222] = {.lex_state = 132},
  [4223] = {.lex_state = 132},
  [4224] = {.lex_state = 132},
  [4225] = {.lex_state = 132},
  [4226] = {.lex_state = 132},
  [4227] = {.lex_state = 132},
  [4228] = {.lex_state = 132},
  [4229] = {.lex_state = 132},
  [4230] = {.lex_state = 132},
  [4231] = {.lex_state = 132},
  [4232] = {.lex_state = 132},
  [4233] = {.lex_state = 132},
  [4234] = {.lex_state = 132},
  [4235] = {.lex_state = 132},
  [4236] = {.lex_state = 132},
  [4237] = {.lex_state = 132},
  [4238] = {.lex_state = 132},
  [4239] = {.lex_state = 132},
  [4240] = {.lex_state = 132},
  [4241] = {.lex_state = 132},
  [4242] = {.lex_state = 132},
  [4243] = {.lex_state = 132},
  [4244] = {.lex_state = 132},
  [4245] = {.lex_state = 132},
  [4246] = {.lex_state = 132},
  [4247] = {.lex_state = 132},
  [4248] = {.lex_state = 132},
  [4249] = {.lex_state = 132},
  [4250] = {.lex_state = 132},
  [4251] = {.lex_state = 132},
  [4252] = {.lex_state = 132},
  [4253] = {.lex_state = 132},
  [4254] = {.lex_state = 132},
  [4255] = {.lex_state = 122},
  [4256] = {.lex_state = 132},
  [4257] = {.lex_state = 132},
  [4258] = {.lex_state = 132},
  [4259] = {.lex_state = 132},
  [4260] = {.lex_state = 132},
  [4261] = {.lex_state = 132},
  [4262] = {.lex_state = 132},
  [4263] = {.lex_state = 132},
  [4264] = {.lex_state = 132},
  [4265] = {.lex_state = 117},
  [4266] = {.lex_state = 132},
  [4267] = {.lex_state = 117},
  [4268] = {.lex_state = 132},
  [4269] = {.lex_state = 132},
  [4270] = {.lex_state = 132},
  [4271] = {.lex_state = 132},
  [4272] = {.lex_state = 117},
  [4273] = {.lex_state = 132},
  [4274] = {.lex_state = 132},
  [4275] = {.lex_state = 117},
  [4276] = {.lex_state = 119},
  [4277] = {.lex_state = 117},
  [4278] = {.lex_state = 117},
  [4279] = {.lex_state = 117},
  [4280] = {.lex_state = 117},
  [4281] = {.lex_state = 117},
  [4282] = {.lex_state = 117},
  [4283] = {.lex_state = 117},
  [4284] = {.lex_state = 132},
  [4285] = {.lex_state = 132},
  [4286] = {.lex_state = 132},
  [4287] = {.lex_state = 132},
  [4288] = {.lex_state = 117},
  [4289] = {.lex_state = 132},
  [4290] = {.lex_state = 132},
  [4291] = {.lex_state = 132},
  [4292] = {.lex_state = 124},
  [4293] = {.lex_state = 124},
  [4294] = {.lex_state = 132},
  [4295] = {.lex_state = 124},
  [4296] = {.lex_state = 132},
  [4297] = {.lex_state = 124},
  [4298] = {.lex_state = 124},
  [4299] = {.lex_state = 124},
  [4300] = {.lex_state = 124},
  [4301] = {.lex_state = 124},
  [4302] = {.lex_state = 124},
  [4303] = {.lex_state = 124},
  [4304] = {.lex_state = 124},
  [4305] = {.lex_state = 124},
  [4306] = {.lex_state = 132},
  [4307] = {.lex_state = 124},
  [4308] = {.lex_state = 132},
  [4309] = {.lex_state = 124},
  [4310] = {.lex_state = 124},
  [4311] = {.lex_state = 132},
  [4312] = {.lex_state = 124},
  [4313] = {.lex_state = 132},
  [4314] = {.lex_state = 132},
  [4315] = {.lex_state = 124},
  [4316] = {.lex_state = 124},
  [4317] = {.lex_state = 124},
  [4318] = {.lex_state = 124},
  [4319] = {.lex_state = 124},
  [4320] = {.lex_state = 124},
  [4321] = {.lex_state = 124},
  [4322] = {.lex_state = 124},
  [4323] = {.lex_state = 124},
  [4324] = {.lex_state = 124},
  [4325] = {.lex_state = 124},
  [4326] = {.lex_state = 124},
  [4327] = {.lex_state = 124},
  [4328] = {.lex_state = 124},
  [4329] = {.lex_state = 124},
  [4330] = {.lex_state = 132},
  [4331] = {.lex_state = 132},
  [4332] = {.lex_state = 132},
  [4333] = {.lex_state = 132},
  [4334] = {.lex_state = 132},
  [4335] = {.lex_state = 132},
  [4336] = {.lex_state = 132},
  [4337] = {.lex_state = 132},
  [4338] = {.lex_state = 132},
  [4339] = {.lex_state = 132},
  [4340] = {.lex_state = 132},
  [4341] = {.lex_state = 124},
  [4342] = {.lex_state = 132},
  [4343] = {.lex_state = 120},
  [4344] = {.lex_state = 120},
  [4345] = {.lex_state = 120},
  [4346] = {.lex_state = 120},
  [4347] = {.lex_state = 120},
  [4348] = {.lex_state = 120},
  [4349] = {.lex_state = 120},
  [4350] = {.lex_state = 120},
  [4351] = {.lex_state = 120},
  [4352] = {.lex_state = 120},
  [4353] = {.lex_state = 120},
  [4354] = {.lex_state = 120},
  [4355] = {.lex_state = 120},
  [4356] = {.lex_state = 120},
  [4357] = {.lex_state = 120},
  [4358] = {.lex_state = 120},
  [4359] = {.lex_state = 120},
  [4360] = {.lex_state = 120},
  [4361] = {.lex_state = 125},
  [4362] = {.lex_state = 125},
  [4363] = {.lex_state = 125},
  [4364] = {.lex_state = 125},
  [4365] = {.lex_state = 125},
  [4366] = {.lex_state = 125},
  [4367] = {.lex_state = 125},
  [4368] = {.lex_state = 125},
  [4369] = {.lex_state = 125},
  [4370] = {.lex_state = 125},
  [4371] = {.lex_state = 125},
  [4372] = {.lex_state = 125},
  [4373] = {.lex_state = 125},
  [4374] = {.lex_state = 125},
  [4375] = {.lex_state = 125},
  [4376] = {.lex_state = 125},
  [4377] = {.lex_state = 125},
  [4378] = {.lex_state = 125},
  [4379] = {.lex_state = 125},
  [4380] = {.lex_state = 125},
  [4381] = {.lex_state = 125},
  [4382] = {.lex_state = 125},
  [4383] = {.lex_state = 125},
  [4384] = {.lex_state = 125},
  [4385] = {.lex_state = 125},
  [4386] = {.lex_state = 125},
  [4387] = {.lex_state = 125},
  [4388] = {.lex_state = 125},
  [4389] = {.lex_state = 125},
  [4390] = {.lex_state = 125},
  [4391] = {.lex_state = 125},
  [4392] = {.lex_state = 125},
  [4393] = {.lex_state = 125},
  [4394] = {.lex_state = 125},
  [4395] = {.lex_state = 125},
  [4396] = {.lex_state = 125},
  [4397] = {.lex_state = 125},
  [4398] = {.lex_state = 125},
  [4399] = {.lex_state = 125},
  [4400] = {.lex_state = 125},
  [4401] = {.lex_state = 125},
  [4402] = {.lex_state = 125},
  [4403] = {.lex_state = 125},
  [4404] = {.lex_state = 125},
  [4405] = {.lex_state = 125},
  [4406] = {.lex_state = 125},
  [4407] = {.lex_state = 125},
  [4408] = {.lex_state = 125},
  [4409] = {.lex_state = 125},
  [4410] = {.lex_state = 125},
  [4411] = {.lex_state = 125},
  [4412] = {.lex_state = 125},
  [4413] = {.lex_state = 125},
  [4414] = {.lex_state = 125},
  [4415] = {.lex_state = 125},
  [4416] = {.lex_state = 125},
  [4417] = {.lex_state = 125},
  [4418] = {.lex_state = 125},
  [4419] = {.lex_state = 125},
  [4420] = {.lex_state = 125},
  [4421] = {.lex_state = 125},
  [4422] = {.lex_state = 125},
  [4423] = {.lex_state = 125},
  [4424] = {.lex_state = 125},
  [4425] = {.lex_state = 125},
  [4426] = {.lex_state = 125},
  [4427] = {.lex_state = 125},
  [4428] = {.lex_state = 125},
  [4429] = {.lex_state = 125},
  [4430] = {.lex_state = 125},
  [4431] = {.lex_state = 125},
  [4432] = {.lex_state = 125},
  [4433] = {.lex_state = 125},
  [4434] = {.lex_state = 125},
  [4435] = {.lex_state = 125},
  [4436] = {.lex_state = 125},
  [4437] = {.lex_state = 125},
  [4438] = {.lex_state = 125},
  [4439] = {.lex_state = 125},
  [4440] = {.lex_state = 125},
  [4441] = {.lex_state = 125},
  [4442] = {.lex_state = 125},
  [4443] = {.lex_state = 125},
  [4444] = {.lex_state = 125},
  [4445] = {.lex_state = 125},
  [4446] = {.lex_state = 125},
  [4447] = {.lex_state = 125},
  [4448] = {.lex_state = 125},
  [4449] = {.lex_state = 125},
  [4450] = {.lex_state = 125},
  [4451] = {.lex_state = 125},
  [4452] = {.lex_state = 125},
  [4453] = {.lex_state = 125},
  [4454] = {.lex_state = 125},
  [4455] = {.lex_state = 125},
  [4456] = {.lex_state = 125},
  [4457] = {.lex_state = 125},
  [4458] = {.lex_state = 125},
  [4459] = {.lex_state = 125},
  [4460] = {.lex_state = 125},
  [4461] = {.lex_state = 125},
  [4462] = {.lex_state = 125},
  [4463] = {.lex_state = 125},
  [4464] = {.lex_state = 125},
  [4465] = {.lex_state = 125},
  [4466] = {.lex_state = 125},
  [4467] = {.lex_state = 125},
  [4468] = {.lex_state = 125},
  [4469] = {.lex_state = 125},
  [4470] = {.lex_state = 125},
  [4471] = {.lex_state = 125},
  [4472] = {.lex_state = 125},
  [4473] = {.lex_state = 125},
  [4474] = {.lex_state = 125},
  [4475] = {.lex_state = 125},
  [4476] = {.lex_state = 125},
  [4477] = {.lex_state = 125},
  [4478] = {.lex_state = 125},
  [4479] = {.lex_state = 125},
  [4480] = {.lex_state = 125},
  [4481] = {.lex_state = 125},
  [4482] = {.lex_state = 125},
  [4483] = {.lex_state = 125},
  [4484] = {.lex_state = 125},
  [4485] = {.lex_state = 125},
  [4486] = {.lex_state = 125},
  [4487] = {.lex_state = 125},
  [4488] = {.lex_state = 125},
  [4489] = {.lex_state = 125},
  [4490] = {.lex_state = 125},
  [4491] = {.lex_state = 125},
  [4492] = {.lex_state = 132},
  [4493] = {.lex_state = 125},
  [4494] = {.lex_state = 125},
  [4495] = {.lex_state = 125},
  [4496] = {.lex_state = 125},
  [4497] = {.lex_state = 125},
  [4498] = {.lex_state = 125},
  [4499] = {.lex_state = 125},
  [4500] = {.lex_state = 125},
  [4501] = {.lex_state = 125},
  [4502] = {.lex_state = 125},
  [4503] = {.lex_state = 125},
  [4504] = {.lex_state = 125},
  [4505] = {.lex_state = 125},
  [4506] = {.lex_state = 125},
  [4507] = {.lex_state = 125},
  [4508] = {.lex_state = 125},
  [4509] = {.lex_state = 125},
  [4510] = {.lex_state = 125},
  [4511] = {.lex_state = 125},
  [4512] = {.lex_state = 125},
  [4513] = {.lex_state = 125},
  [4514] = {.lex_state = 125},
  [4515] = {.lex_state = 125},
  [4516] = {.lex_state = 125},
  [4517] = {.lex_state = 125},
  [4518] = {.lex_state = 125},
  [4519] = {.lex_state = 125},
  [4520] = {.lex_state = 125},
  [4521] = {.lex_state = 125},
  [4522] = {.lex_state = 125},
  [4523] = {.lex_state = 125},
  [4524] = {.lex_state = 125},
  [4525] = {.lex_state = 125},
  [4526] = {.lex_state = 125},
  [4527] = {.lex_state = 125},
  [4528] = {.lex_state = 125},
  [4529] = {.lex_state = 125},
  [4530] = {.lex_state = 125},
  [4531] = {.lex_state = 125},
  [4532] = {.lex_state = 125},
  [4533] = {.lex_state = 125},
  [4534] = {.lex_state = 125},
  [4535] = {.lex_state = 125},
  [4536] = {.lex_state = 125},
  [4537] = {.lex_state = 125},
  [4538] = {.lex_state = 125},
  [4539] = {.lex_state = 125},
  [4540] = {.lex_state = 125},
  [4541] = {.lex_state = 125},
  [4542] = {.lex_state = 125},
  [4543] = {.lex_state = 125},
  [4544] = {.lex_state = 132},
  [4545] = {.lex_state = 125},
  [4546] = {.lex_state = 125},
  [4547] = {.lex_state = 125},
  [4548] = {.lex_state = 125},
  [4549] = {.lex_state = 125},
  [4550] = {.lex_state = 125},
  [4551] = {.lex_state = 125},
  [4552] = {.lex_state = 125},
  [4553] = {.lex_state = 132},
  [4554] = {.lex_state = 125},
  [4555] = {.lex_state = 132},
  [4556] = {.lex_state = 125},
  [4557] = {.lex_state = 125},
  [4558] = {.lex_state = 132},
  [4559] = {.lex_state = 125},
  [4560] = {.lex_state = 132},
  [4561] = {.lex_state = 125},
  [4562] = {.lex_state = 125},
  [4563] = {.lex_state = 125},
  [4564] = {.lex_state = 125},
  [4565] = {.lex_state = 125},
  [4566] = {.lex_state = 132},
  [4567] = {.lex_state = 116},
  [4568] = {.lex_state = 126},
  [4569] = {.lex_state = 116},
  [4570] = {.lex_state = 116},
  [4571] = {.lex_state = 116},
  [4572] = {.lex_state = 116},
  [4573] = {.lex_state = 116},
  [4574] = {.lex_state = 116},
  [4575] = {.lex_state = 116},
  [4576] = {.lex_state = 116},
  [4577] = {.lex_state = 116},
  [4578] = {.lex_state = 116},
  [4579] = {.lex_state = 126},
  [4580] = {.lex_state = 116},
  [4581] = {.lex_state = 116},
  [4582] = {.lex_state = 116},
  [4583] = {.lex_state = 126},
  [4584] = {.lex_state = 116},
  [4585] = {.lex_state = 116},
  [4586] = {.lex_state = 116},
  [4587] = {.lex_state = 116},
  [4588] = {.lex_state = 116},
  [4589] = {.lex_state = 116},
  [4590] = {.lex_state = 116},
  [4591] = {.lex_state = 116},
  [4592] = {.lex_state = 116},
  [4593] = {.lex_state = 116},
  [4594] = {.lex_state = 116},
  [4595] = {.lex_state = 116},
  [4596] = {.lex_state = 116},
  [4597] = {.lex_state = 116},
  [4598] = {.lex_state = 116},
  [4599] = {.lex_state = 116},
  [4600] = {.lex_state = 116},
  [4601] = {.lex_state = 116},
  [4602] = {.lex_state = 116},
  [4603] = {.lex_state = 132},
  [4604] = {.lex_state = 132},
  [4605] = {.lex_state = 116},
  [4606] = {.lex_state = 126},
  [4607] = {.lex_state = 116},
  [4608] = {.lex_state = 116},
  [4609] = {.lex_state = 116},
  [4610] = {.lex_state = 116},
  [4611] = {.lex_state = 116},
  [4612] = {.lex_state = 116},
  [4613] = {.lex_state = 51},
  [4614] = {.lex_state = 116},
  [4615] = {.lex_state = 116},
  [4616] = {.lex_state = 116},
  [4617] = {.lex_state = 116},
  [4618] = {.lex_state = 132},
  [4619] = {.lex_state = 116},
  [4620] = {.lex_state = 116},
  [4621] = {.lex_state = 116},
  [4622] = {.lex_state = 116},
  [4623] = {.lex_state = 126},
  [4624] = {.lex_state = 126},
  [4625] = {.lex_state = 126},
  [4626] = {.lex_state = 126},
  [4627] = {.lex_state = 126},
  [4628] = {.lex_state = 119},
  [4629] = {.lex_state = 126},
  [4630] = {.lex_state = 126},
  [4631] = {.lex_state = 126},
  [4632] = {.lex_state = 126},
  [4633] = {.lex_state = 126},
  [4634] = {.lex_state = 119},
  [4635] = {.lex_state = 126},
  [4636] = {.lex_state = 126},
  [4637] = {.lex_state = 126},
  [4638] = {.lex_state = 119},
  [4639] = {.lex_state = 126},
  [4640] = {.lex_state = 126},
  [4641] = {.lex_state = 126},
  [4642] = {.lex_state = 126},
  [4643] = {.lex_state = 126},
  [4644] = {.lex_state = 126},
  [4645] = {.lex_state = 126},
  [4646] = {.lex_state = 51},
  [4647] = {.lex_state = 51},
  [4648] = {.lex_state = 51},
  [4649] = {.lex_state = 51},
  [4650] = {.lex_state = 51},
  [4651] = {.lex_state = 126},
  [4652] = {.lex_state = 51},
  [4653] = {.lex_state = 51},
  [4654] = {.lex_state = 51},
  [4655] = {.lex_state = 51},
  [4656] = {.lex_state = 51},
  [4657] = {.lex_state = 51},
  [4658] = {.lex_state = 51},
  [4659] = {.lex_state = 51},
  [4660] = {.lex_state = 51},
  [4661] = {.lex_state = 51},
  [4662] = {.lex_state = 51},
  [4663] = {.lex_state = 51},
  [4664] = {.lex_state = 51},
  [4665] = {.lex_state = 51},
  [4666] = {.lex_state = 51},
  [4667] = {.lex_state = 51},
  [4668] = {.lex_state = 51},
  [4669] = {.lex_state = 126},
  [4670] = {.lex_state = 51},
  [4671] = {.lex_state = 51},
  [4672] = {.lex_state = 51},
  [4673] = {.lex_state = 51},
  [4674] = {.lex_state = 51},
  [4675] = {.lex_state = 51},
  [4676] = {.lex_state = 51},
  [4677] = {.lex_state = 51},
  [4678] = {.lex_state = 51},
  [4679] = {.lex_state = 51},
  [4680] = {.lex_state = 51},
  [4681] = {.lex_state = 51},
  [4682] = {.lex_state = 51},
  [4683] = {.lex_state = 51},
  [4684] = {.lex_state = 51},
  [4685] = {.lex_state = 51},
  [4686] = {.lex_state = 51},
  [4687] = {.lex_state = 132},
  [4688] = {.lex_state = 132},
  [4689] = {.lex_state = 132},
  [4690] = {.lex_state = 132},
  [4691] = {.lex_state = 132},
  [4692] = {.lex_state = 132},
  [4693] = {.lex_state = 132},
  [4694] = {.lex_state = 132},
  [4695] = {.lex_state = 132},
  [4696] = {.lex_state = 132},
  [4697] = {.lex_state = 132},
  [4698] = {.lex_state = 132},
  [4699] = {.lex_state = 132},
  [4700] = {.lex_state = 132},
  [4701] = {.lex_state = 132},
  [4702] = {.lex_state = 132},
  [4703] = {.lex_state = 132},
  [4704] = {.lex_state = 132},
  [4705] = {.lex_state = 132},
  [4706] = {.lex_state = 132},
  [4707] = {.lex_state = 132},
  [4708] = {.lex_state = 132},
  [4709] = {.lex_state = 132},
  [4710] = {.lex_state = 132},
  [4711] = {.lex_state = 132},
  [4712] = {.lex_state = 132},
  [4713] = {.lex_state = 132},
  [4714] = {.lex_state = 132},
  [4715] = {.lex_state = 132},
  [4716] = {.lex_state = 132},
  [4717] = {.lex_state = 132},
  [4718] = {.lex_state = 132},
  [4719] = {.lex_state = 132},
  [4720] = {.lex_state = 132},
  [4721] = {.lex_state = 132},
  [4722] = {.lex_state = 132},
  [4723] = {.lex_state = 132},
  [4724] = {.lex_state = 132},
  [4725] = {.lex_state = 132},
  [4726] = {.lex_state = 132},
  [4727] = {.lex_state = 132},
  [4728] = {.lex_state = 132},
  [4729] = {.lex_state = 132},
  [4730] = {.lex_state = 132},
  [4731] = {.lex_state = 132},
  [4732] = {.lex_state = 132},
  [4733] = {.lex_state = 132},
  [4734] = {.lex_state = 132},
  [4735] = {.lex_state = 132},
  [4736] = {.lex_state = 132},
  [4737] = {.lex_state = 132},
  [4738] = {.lex_state = 132},
  [4739] = {.lex_state = 132},
  [4740] = {.lex_state = 132},
  [4741] = {.lex_state = 132},
  [4742] = {.lex_state = 132},
  [4743] = {.lex_state = 132},
  [4744] = {.lex_state = 132},
  [4745] = {.lex_state = 132},
  [4746] = {.lex_state = 132},
  [4747] = {.lex_state = 132},
  [4748] = {.lex_state = 132},
  [4749] = {.lex_state = 132},
  [4750] = {.lex_state = 132},
  [4751] = {.lex_state = 132},
  [4752] = {.lex_state = 132},
  [4753] = {.lex_state = 132},
  [4754] = {.lex_state = 132},
  [4755] = {.lex_state = 132},
  [4756] = {.lex_state = 132},
  [4757] = {.lex_state = 132},
  [4758] = {.lex_state = 132},
  [4759] = {.lex_state = 132},
  [4760] = {.lex_state = 132},
  [4761] = {.lex_state = 132},
  [4762] = {.lex_state = 132},
  [4763] = {.lex_state = 132},
  [4764] = {.lex_state = 132},
  [4765] = {.lex_state = 132},
  [4766] = {.lex_state = 132},
  [4767] = {.lex_state = 132},
  [4768] = {.lex_state = 132},
  [4769] = {.lex_state = 132},
  [4770] = {.lex_state = 132},
  [4771] = {.lex_state = 132},
  [4772] = {.lex_state = 132},
  [4773] = {.lex_state = 132},
  [4774] = {.lex_state = 132},
  [4775] = {.lex_state = 132},
  [4776] = {.lex_state = 132},
  [4777] = {.lex_state = 132},
  [4778] = {.lex_state = 132},
  [4779] = {.lex_state = 132},
  [4780] = {.lex_state = 132},
  [4781] = {.lex_state = 132},
  [4782] = {.lex_state = 132},
  [4783] = {.lex_state = 132},
  [4784] = {.lex_state = 132},
  [4785] = {.lex_state = 132},
  [4786] = {.lex_state = 132},
  [4787] = {.lex_state = 132},
  [4788] = {.lex_state = 132},
  [4789] = {.lex_state = 132},
  [4790] = {.lex_state = 132},
  [4791] = {.lex_state = 132},
  [4792] = {.lex_state = 132},
  [4793] = {.lex_state = 132},
  [4794] = {.lex_state = 416},
  [4795] = {.lex_state = 132},
  [4796] = {.lex_state = 132},
  [4797] = {.lex_state = 132},
  [4798] = {.lex_state = 132},
  [4799] = {.lex_state = 132},
  [4800] = {.lex_state = 132},
  [4801] = {.lex_state = 132},
  [4802] = {.lex_state = 119},
  [4803] = {.lex_state = 132},
  [4804] = {.lex_state = 132},
  [4805] = {.lex_state = 132},
  [4806] = {.lex_state = 132},
  [4807] = {.lex_state = 132},
  [4808] = {.lex_state = 132},
  [4809] = {.lex_state = 119},
  [4810] = {.lex_state = 132},
  [4811] = {.lex_state = 132},
  [4812] = {.lex_state = 132},
  [4813] = {.lex_state = 132},
  [4814] = {.lex_state = 132},
  [4815] = {.lex_state = 117},
  [4816] = {.lex_state = 132},
  [4817] = {.lex_state = 132},
  [4818] = {.lex_state = 132},
  [4819] = {.lex_state = 132},
  [4820] = {.lex_state = 132},
  [4821] = {.lex_state = 119},
  [4822] = {.lex_state = 132},
  [4823] = {.lex_state = 132},
  [4824] = {.lex_state = 124},
  [4825] = {.lex_state = 117},
  [4826] = {.lex_state = 132},
  [4827] = {.lex_state = 132},
  [4828] = {.lex_state = 132},
  [4829] = {.lex_state = 132},
  [4830] = {.lex_state = 117},
  [4831] = {.lex_state = 124},
  [4832] = {.lex_state = 117},
  [4833] = {.lex_state = 132},
  [4834] = {.lex_state = 117},
  [4835] = {.lex_state = 132},
  [4836] = {.lex_state = 132},
  [4837] = {.lex_state = 129},
  [4838] = {.lex_state = 132},
  [4839] = {.lex_state = 132},
  [4840] = {.lex_state = 132},
  [4841] = {.lex_state = 129},
  [4842] = {.lex_state = 132},
  [4843] = {.lex_state = 117},
  [4844] = {.lex_state = 132},
  [4845] = {.lex_state = 132},
  [4846] = {.lex_state = 117},
  [4847] = {.lex_state = 132},
  [4848] = {.lex_state = 132},
  [4849] = {.lex_state = 117},
  [4850] = {.lex_state = 117},
  [4851] = {.lex_state = 132},
  [4852] = {.lex_state = 117},
  [4853] = {.lex_state = 129},
  [4854] = {.lex_state = 132},
  [4855] = {.lex_state = 117},
  [4856] = {.lex_state = 132},
  [4857] = {.lex_state = 129},
  [4858] = {.lex_state = 117},
  [4859] = {.lex_state = 129},
  [4860] = {.lex_state = 132},
  [4861] = {.lex_state = 132},
  [4862] = {.lex_state = 117},
  [4863] = {.lex_state = 132},
  [4864] = {.lex_state = 132},
  [4865] = {.lex_state = 132},
  [4866] = {.lex_state = 132},
  [4867] = {.lex_state = 132},
  [4868] = {.lex_state = 117},
  [4869] = {.lex_state = 132},
  [4870] = {.lex_state = 117},
  [4871] = {.lex_state = 132},
  [4872] = {.lex_state = 132},
  [4873] = {.lex_state = 132},
  [4874] = {.lex_state = 132},
  [4875] = {.lex_state = 132},
  [4876] = {.lex_state = 132},
  [4877] = {.lex_state = 132},
  [4878] = {.lex_state = 132},
  [4879] = {.lex_state = 132},
  [4880] = {.lex_state = 132},
  [4881] = {.lex_state = 117},
  [4882] = {.lex_state = 117},
  [4883] = {.lex_state = 132},
  [4884] = {.lex_state = 132},
  [4885] = {.lex_state = 132},
  [4886] = {.lex_state = 132},
  [4887] = {.lex_state = 132},
  [4888] = {.lex_state = 124},
  [4889] = {.lex_state = 132},
  [4890] = {.lex_state = 132},
  [4891] = {.lex_state = 132},
  [4892] = {.lex_state = 132},
  [4893] = {.lex_state = 132},
  [4894] = {.lex_state = 128},
  [4895] = {.lex_state = 132},
  [4896] = {.lex_state = 132},
  [4897] = {.lex_state = 132},
  [4898] = {.lex_state = 132},
  [4899] = {.lex_state = 132},
  [4900] = {.lex_state = 132},
  [4901] = {.lex_state = 132},
  [4902] = {.lex_state = 132},
  [4903] = {.lex_state = 132},
  [4904] = {.lex_state = 132},
  [4905] = {.lex_state = 132},
  [4906] = {.lex_state = 124},
  [4907] = {.lex_state = 132},
  [4908] = {.lex_state = 132},
  [4909] = {.lex_state = 132},
  [4910] = {.lex_state = 124},
  [4911] = {.lex_state = 124},
  [4912] = {.lex_state = 132},
  [4913] = {.lex_state = 132},
  [4914] = {.lex_state = 124},
  [4915] = {.lex_state = 132},
  [4916] = {.lex_state = 132},
  [4917] = {.lex_state = 128},
  [4918] = {.lex_state = 132},
  [4919] = {.lex_state = 132},
  [4920] = {.lex_state = 124},
  [4921] = {.lex_state = 124},
  [4922] = {.lex_state = 132},
  [4923] = {.lex_state = 132},
  [4924] = {.lex_state = 132},
  [4925] = {.lex_state = 132},
  [4926] = {.lex_state = 132},
  [4927] = {.lex_state = 124},
  [4928] = {.lex_state = 132},
  [4929] = {.lex_state = 132},
  [4930] = {.lex_state = 132},
  [4931] = {.lex_state = 132},
  [4932] = {.lex_state = 132},
  [4933] = {.lex_state = 132},
  [4934] = {.lex_state = 132},
  [4935] = {.lex_state = 132},
  [4936] = {.lex_state = 132},
  [4937] = {.lex_state = 132},
  [4938] = {.lex_state = 132},
  [4939] = {.lex_state = 124},
  [4940] = {.lex_state = 132},
  [4941] = {.lex_state = 132},
  [4942] = {.lex_state = 132},
  [4943] = {.lex_state = 132},
  [4944] = {.lex_state = 124},
  [4945] = {.lex_state = 132},
  [4946] = {.lex_state = 132},
  [4947] = {.lex_state = 132},
  [4948] = {.lex_state = 132},
  [4949] = {.lex_state = 132},
  [4950] = {.lex_state = 117},
  [4951] = {.lex_state = 704},
  [4952] = {.lex_state = 132},
  [4953] = {.lex_state = 132},
  [4954] = {.lex_state = 117},
  [4955] = {.lex_state = 117},
  [4956] = {.lex_state = 132},
  [4957] = {.lex_state = 132},
  [4958] = {.lex_state = 132},
  [4959] = {.lex_state = 416},
  [4960] = {.lex_state = 132},
  [4961] = {.lex_state = 132},
  [4962] = {.lex_state = 416},
  [4963] = {.lex_state = 132},
  [4964] = {.lex_state = 416},
  [4965] = {.lex_state = 132},
  [4966] = {.lex_state = 132},
  [4967] = {.lex_state = 132},
  [4968] = {.lex_state = 132},
  [4969] = {.lex_state = 132},
  [4970] = {.lex_state = 117},
  [4971] = {.lex_state = 416},
  [4972] = {.lex_state = 132},
  [4973] = {.lex_state = 704},
  [4974] = {.lex_state = 132},
  [4975] = {.lex_state = 704},
  [4976] = {.lex_state = 0},
  [4977] = {.lex_state = 132},
  [4978] = {.lex_state = 416},
  [4979] = {.lex_state = 416},
  [4980] = {.lex_state = 124},
  [4981] = {.lex_state = 704},
  [4982] = {.lex_state = 132},
  [4983] = {.lex_state = 132},
  [4984] = {.lex_state = 132},
  [4985] = {.lex_state = 117},
  [4986] = {.lex_state = 132},
  [4987] = {.lex_state = 132},
  [4988] = {.lex_state = 132},
  [4989] = {.lex_state = 132},
  [4990] = {.lex_state = 132},
  [4991] = {.lex_state = 132},
  [4992] = {.lex_state = 132},
  [4993] = {.lex_state = 117},
  [4994] = {.lex_state = 132},
  [4995] = {.lex_state = 0},
  [4996] = {.lex_state = 416},
  [4997] = {.lex_state = 132},
  [4998] = {.lex_state = 132},
  [4999] = {.lex_state = 416},
  [5000] = {.lex_state = 416},
  [5001] = {.lex_state = 117},
  [5002] = {.lex_state = 124},
  [5003] = {.lex_state = 117},
  [5004] = {.lex_state = 132},
  [5005] = {.lex_state = 416},
  [5006] = {.lex_state = 416},
  [5007] = {.lex_state = 132},
  [5008] = {.lex_state = 132},
  [5009] = {.lex_state = 132},
  [5010] = {.lex_state = 132},
  [5011] = {.lex_state = 132},
  [5012] = {.lex_state = 416},
  [5013] = {.lex_state = 416},
  [5014] = {.lex_state = 132},
  [5015] = {.lex_state = 132},
  [5016] = {.lex_state = 132},
  [5017] = {.lex_state = 132},
  [5018] = {.lex_state = 132},
  [5019] = {.lex_state = 416},
  [5020] = {.lex_state = 416},
  [5021] = {.lex_state = 132},
  [5022] = {.lex_state = 132},
  [5023] = {.lex_state = 132},
  [5024] = {.lex_state = 416},
  [5025] = {.lex_state = 132},
  [5026] = {.lex_state = 124},
  [5027] = {.lex_state = 132},
  [5028] = {.lex_state = 416},
  [5029] = {.lex_state = 416},
  [5030] = {.lex_state = 416},
  [5031] = {.lex_state = 132},
  [5032] = {.lex_state = 416},
  [5033] = {.lex_state = 416},
  [5034] = {.lex_state = 132},
  [5035] = {.lex_state = 126},
  [5036] = {.lex_state = 124},
  [5037] = {.lex_state = 132},
  [5038] = {.lex_state = 132},
  [5039] = {.lex_state = 132},
  [5040] = {.lex_state = 132},
  [5041] = {.lex_state = 416},
  [5042] = {.lex_state = 124},
  [5043] = {.lex_state = 416},
  [5044] = {.lex_state = 132},
  [5045] = {.lex_state = 132},
  [5046] = {.lex_state = 132},
  [5047] = {.lex_state = 416},
  [5048] = {.lex_state = 416},
  [5049] = {.lex_state = 132},
  [5050] = {.lex_state = 128},
  [5051] = {.lex_state = 132},
  [5052] = {.lex_state = 126},
  [5053] = {.lex_state = 132},
  [5054] = {.lex_state = 132},
  [5055] = {.lex_state = 132},
  [5056] = {.lex_state = 132},
  [5057] = {.lex_state = 132},
  [5058] = {.lex_state = 416},
  [5059] = {.lex_state = 132},
  [5060] = {.lex_state = 132},
  [5061] = {.lex_state = 132},
  [5062] = {.lex_state = 132},
  [5063] = {.lex_state = 416},
  [5064] = {.lex_state = 132},
  [5065] = {.lex_state = 416},
  [5066] = {.lex_state = 124},
  [5067] = {.lex_state = 132},
  [5068] = {.lex_state = 132},
  [5069] = {.lex_state = 416},
  [5070] = {.lex_state = 132},
  [5071] = {.lex_state = 132},
  [5072] = {.lex_state = 132},
  [5073] = {.lex_state = 416},
  [5074] = {.lex_state = 124},
  [5075] = {.lex_state = 132},
  [5076] = {.lex_state = 132},
  [5077] = {.lex_state = 132},
  [5078] = {.lex_state = 416},
  [5079] = {.lex_state = 124},
  [5080] = {.lex_state = 416},
  [5081] = {.lex_state = 132},
  [5082] = {.lex_state = 416},
  [5083] = {.lex_state = 132},
  [5084] = {.lex_state = 132},
  [5085] = {.lex_state = 416},
  [5086] = {.lex_state = 124},
  [5087] = {.lex_state = 132},
  [5088] = {.lex_state = 416},
  [5089] = {.lex_state = 416},
  [5090] = {.lex_state = 416},
  [5091] = {.lex_state = 132},
  [5092] = {.lex_state = 132},
  [5093] = {.lex_state = 132},
  [5094] = {.lex_state = 132},
  [5095] = {.lex_state = 416},
  [5096] = {.lex_state = 132},
  [5097] = {.lex_state = 132},
  [5098] = {.lex_state = 132},
  [5099] = {.lex_state = 416},
  [5100] = {.lex_state = 132},
  [5101] = {.lex_state = 132},
  [5102] = {.lex_state = 132},
  [5103] = {.lex_state = 132},
  [5104] = {.lex_state = 124},
  [5105] = {.lex_state = 132},
  [5106] = {.lex_state = 132},
  [5107] = {.lex_state = 132},
  [5108] = {.lex_state = 124},
  [5109] = {.lex_state = 132},
  [5110] = {.lex_state = 132},
  [5111] = {.lex_state = 132},
  [5112] = {.lex_state = 132},
  [5113] = {.lex_state = 126},
  [5114] = {.lex_state = 132},
  [5115] = {.lex_state = 132},
  [5116] = {.lex_state = 132},
  [5117] = {.lex_state = 124},
  [5118] = {.lex_state = 126},
  [5119] = {.lex_state = 132},
  [5120] = {.lex_state = 416},
  [5121] = {.lex_state = 124},
  [5122] = {.lex_state = 132},
  [5123] = {.lex_state = 132},
  [5124] = {.lex_state = 124},
  [5125] = {.lex_state = 132},
  [5126] = {.lex_state = 416},
  [5127] = {.lex_state = 416},
  [5128] = {.lex_state = 132},
  [5129] = {.lex_state = 132},
  [5130] = {.lex_state = 132},
  [5131] = {.lex_state = 132},
  [5132] = {.lex_state = 704},
  [5133] = {.lex_state = 132},
  [5134] = {.lex_state = 132},
  [5135] = {.lex_state = 132},
  [5136] = {.lex_state = 132},
  [5137] = {.lex_state = 132},
  [5138] = {.lex_state = 132},
  [5139] = {.lex_state = 132},
  [5140] = {.lex_state = 132},
  [5141] = {.lex_state = 132},
  [5142] = {.lex_state = 126},
  [5143] = {.lex_state = 132},
  [5144] = {.lex_state = 126},
  [5145] = {.lex_state = 132},
  [5146] = {.lex_state = 132},
  [5147] = {.lex_state = 132},
  [5148] = {.lex_state = 416},
  [5149] = {.lex_state = 416},
  [5150] = {.lex_state = 132},
  [5151] = {.lex_state = 132},
  [5152] = {.lex_state = 132},
  [5153] = {.lex_state = 132},
  [5154] = {.lex_state = 416},
  [5155] = {.lex_state = 416},
  [5156] = {.lex_state = 124},
  [5157] = {.lex_state = 132},
  [5158] = {.lex_state = 132},
  [5159] = {.lex_state = 124},
  [5160] = {.lex_state = 126},
  [5161] = {.lex_state = 132},
  [5162] = {.lex_state = 132},
  [5163] = {.lex_state = 0},
  [5164] = {.lex_state = 416},
  [5165] = {.lex_state = 416},
  [5166] = {.lex_state = 54},
  [5167] = {.lex_state = 132},
  [5168] = {.lex_state = 132},
  [5169] = {.lex_state = 132},
  [5170] = {.lex_state = 0},
  [5171] = {.lex_state = 0},
  [5172] = {.lex_state = 416},
  [5173] = {.lex_state = 416},
  [5174] = {.lex_state = 132},
  [5175] = {.lex_state = 54},
  [5176] = {.lex_state = 0},
  [5177] = {.lex_state = 54},
  [5178] = {.lex_state = 416},
  [5179] = {.lex_state = 54},
  [5180] = {.lex_state = 132},
  [5181] = {.lex_state = 98},
  [5182] = {.lex_state = 416},
  [5183] = {.lex_state = 416},
  [5184] = {.lex_state = 416},
  [5185] = {.lex_state = 132},
  [5186] = {.lex_state = 416},
  [5187] = {.lex_state = 416},
  [5188] = {.lex_state = 0},
  [5189] = {.lex_state = 416},
  [5190] = {.lex_state = 416},
  [5191] = {.lex_state = 54},
  [5192] = {.lex_state = 0},
  [5193] = {.lex_state = 54},
  [5194] = {.lex_state = 416},
  [5195] = {.lex_state = 54},
  [5196] = {.lex_state = 54},
  [5197] = {.lex_state = 416},
  [5198] = {.lex_state = 132},
  [5199] = {.lex_state = 56},
  [5200] = {.lex_state = 416},
  [5201] = {.lex_state = 132},
  [5202] = {.lex_state = 61},
  [5203] = {.lex_state = 416},
  [5204] = {.lex_state = 0},
  [5205] = {.lex_state = 416},
  [5206] = {.lex_state = 416},
  [5207] = {.lex_state = 132},
  [5208] = {.lex_state = 132},
  [5209] = {.lex_state = 54},
  [5210] = {.lex_state = 132},
  [5211] = {.lex_state = 416},
  [5212] = {.lex_state = 132},
  [5213] = {.lex_state = 54},
  [5214] = {.lex_state = 56},
  [5215] = {.lex_state = 132},
  [5216] = {.lex_state = 416},
  [5217] = {.lex_state = 132},
  [5218] = {.lex_state = 54},
  [5219] = {.lex_state = 54},
  [5220] = {.lex_state = 132},
  [5221] = {.lex_state = 132},
  [5222] = {.lex_state = 132},
  [5223] = {.lex_state = 416},
  [5224] = {.lex_state = 108},
  [5225] = {.lex_state = 132},
  [5226] = {.lex_state = 128},
  [5227] = {.lex_state = 416},
  [5228] = {.lex_state = 132},
  [5229] = {.lex_state = 132},
  [5230] = {.lex_state = 416},
  [5231] = {.lex_state = 56},
  [5232] = {.lex_state = 61},
  [5233] = {.lex_state = 54},
  [5234] = {.lex_state = 416},
  [5235] = {.lex_state = 132},
  [5236] = {.lex_state = 132},
  [5237] = {.lex_state = 61},
  [5238] = {.lex_state = 132},
  [5239] = {.lex_state = 54},
  [5240] = {.lex_state = 132},
  [5241] = {.lex_state = 54},
  [5242] = {.lex_state = 416},
  [5243] = {.lex_state = 416},
  [5244] = {.lex_state = 132},
  [5245] = {.lex_state = 0},
  [5246] = {.lex_state = 416},
  [5247] = {.lex_state = 416},
  [5248] = {.lex_state = 54},
  [5249] = {.lex_state = 56},
  [5250] = {.lex_state = 132},
  [5251] = {.lex_state = 54},
  [5252] = {.lex_state = 61},
  [5253] = {.lex_state = 416},
  [5254] = {.lex_state = 54},
  [5255] = {.lex_state = 98},
  [5256] = {.lex_state = 132},
  [5257] = {.lex_state = 54},
  [5258] = {.lex_state = 54},
  [5259] = {.lex_state = 54},
  [5260] = {.lex_state = 54},
  [5261] = {.lex_state = 132},
  [5262] = {.lex_state = 132},
  [5263] = {.lex_state = 54},
  [5264] = {.lex_state = 54},
  [5265] = {.lex_state = 416},
  [5266] = {.lex_state = 124},
  [5267] = {.lex_state = 132},
  [5268] = {.lex_state = 416},
  [5269] = {.lex_state = 98},
  [5270] = {.lex_state = 132},
  [5271] = {.lex_state = 416},
  [5272] = {.lex_state = 416},
  [5273] = {.lex_state = 124},
  [5274] = {.lex_state = 56},
  [5275] = {.lex_state = 132},
  [5276] = {.lex_state = 416},
  [5277] = {.lex_state = 416},
  [5278] = {.lex_state = 54},
  [5279] = {.lex_state = 61},
  [5280] = {.lex_state = 132},
  [5281] = {.lex_state = 56},
  [5282] = {.lex_state = 54},
  [5283] = {.lex_state = 132},
  [5284] = {.lex_state = 54},
  [5285] = {.lex_state = 54},
  [5286] = {.lex_state = 132},
  [5287] = {.lex_state = 416},
  [5288] = {.lex_state = 0},
  [5289] = {.lex_state = 416},
  [5290] = {.lex_state = 56},
  [5291] = {.lex_state = 416},
  [5292] = {.lex_state = 56},
  [5293] = {.lex_state = 0},
  [5294] = {.lex_state = 416},
  [5295] = {.lex_state = 98},
  [5296] = {.lex_state = 54},
  [5297] = {.lex_state = 54},
  [5298] = {.lex_state = 54},
  [5299] = {.lex_state = 54},
  [5300] = {.lex_state = 416},
  [5301] = {.lex_state = 132},
  [5302] = {.lex_state = 132},
  [5303] = {.lex_state = 132},
  [5304] = {.lex_state = 132},
  [5305] = {.lex_state = 0},
  [5306] = {.lex_state = 132},
  [5307] = {.lex_state = 416},
  [5308] = {.lex_state = 56},
  [5309] = {.lex_state = 0},
  [5310] = {.lex_state = 416},
  [5311] = {.lex_state = 132},
  [5312] = {.lex_state = 54},
  [5313] = {.lex_state = 54},
  [5314] = {.lex_state = 0},
  [5315] = {.lex_state = 0},
  [5316] = {.lex_state = 0},
  [5317] = {.lex_state = 0},
  [5318] = {.lex_state = 0},
  [5319] = {.lex_state = 0},
  [5320] = {.lex_state = 416},
  [5321] = {.lex_state = 0},
  [5322] = {.lex_state = 0},
  [5323] = {.lex_state = 416},
  [5324] = {.lex_state = 0},
  [5325] = {.lex_state = 0},
  [5326] = {.lex_state = 0},
  [5327] = {.lex_state = 0},
  [5328] = {.lex_state = 0},
  [5329] = {.lex_state = 0},
  [5330] = {.lex_state = 0},
  [5331] = {.lex_state = 0},
  [5332] = {.lex_state = 0},
  [5333] = {.lex_state = 0},
  [5334] = {.lex_state = 0},
  [5335] = {.lex_state = 0},
  [5336] = {.lex_state = 0},
  [5337] = {.lex_state = 0},
  [5338] = {.lex_state = 0},
  [5339] = {.lex_state = 0},
  [5340] = {.lex_state = 0},
  [5341] = {.lex_state = 416},
  [5342] = {.lex_state = 416},
  [5343] = {.lex_state = 0},
  [5344] = {.lex_state = 0},
  [5345] = {.lex_state = 132},
  [5346] = {.lex_state = 0},
  [5347] = {.lex_state = 0},
  [5348] = {.lex_state = 0},
  [5349] = {.lex_state = 0},
  [5350] = {.lex_state = 0},
  [5351] = {.lex_state = 0},
  [5352] = {.lex_state = 58},
  [5353] = {.lex_state = 416},
  [5354] = {.lex_state = 0},
  [5355] = {.lex_state = 0},
  [5356] = {.lex_state = 0},
  [5357] = {.lex_state = 0},
  [5358] = {.lex_state = 416},
  [5359] = {.lex_state = 0},
  [5360] = {.lex_state = 0},
  [5361] = {.lex_state = 0},
  [5362] = {.lex_state = 0},
  [5363] = {.lex_state = 132},
  [5364] = {.lex_state = 0},
  [5365] = {.lex_state = 0},
  [5366] = {.lex_state = 0},
  [5367] = {.lex_state = 0},
  [5368] = {.lex_state = 0},
  [5369] = {.lex_state = 0},
  [5370] = {.lex_state = 0},
  [5371] = {.lex_state = 0},
  [5372] = {.lex_state = 132},
  [5373] = {.lex_state = 0},
  [5374] = {.lex_state = 0},
  [5375] = {.lex_state = 0},
  [5376] = {.lex_state = 0},
  [5377] = {.lex_state = 0},
  [5378] = {.lex_state = 416},
  [5379] = {.lex_state = 0},
  [5380] = {.lex_state = 416},
  [5381] = {.lex_state = 0},
  [5382] = {.lex_state = 0},
  [5383] = {.lex_state = 416},
  [5384] = {.lex_state = 61},
  [5385] = {.lex_state = 0},
  [5386] = {.lex_state = 416},
  [5387] = {.lex_state = 0},
  [5388] = {.lex_state = 416},
  [5389] = {.lex_state = 0},
  [5390] = {.lex_state = 0},
  [5391] = {.lex_state = 0},
  [5392] = {.lex_state = 0},
  [5393] = {.lex_state = 0},
  [5394] = {.lex_state = 0},
  [5395] = {.lex_state = 0},
  [5396] = {.lex_state = 0},
  [5397] = {.lex_state = 0},
  [5398] = {.lex_state = 0},
  [5399] = {.lex_state = 416},
  [5400] = {.lex_state = 416},
  [5401] = {.lex_state = 0},
  [5402] = {.lex_state = 0},
  [5403] = {.lex_state = 416},
  [5404] = {.lex_state = 416},
  [5405] = {.lex_state = 0},
  [5406] = {.lex_state = 0},
  [5407] = {.lex_state = 416},
  [5408] = {.lex_state = 0},
  [5409] = {.lex_state = 0},
  [5410] = {.lex_state = 416},
  [5411] = {.lex_state = 132},
  [5412] = {.lex_state = 132},
  [5413] = {.lex_state = 416},
  [5414] = {.lex_state = 416},
  [5415] = {.lex_state = 0},
  [5416] = {.lex_state = 0},
  [5417] = {.lex_state = 0},
  [5418] = {.lex_state = 0},
  [5419] = {.lex_state = 416},
  [5420] = {.lex_state = 0},
  [5421] = {.lex_state = 0},
  [5422] = {.lex_state = 416},
  [5423] = {.lex_state = 0},
  [5424] = {.lex_state = 416},
  [5425] = {.lex_state = 0},
  [5426] = {.lex_state = 0},
  [5427] = {.lex_state = 0},
  [5428] = {.lex_state = 0},
  [5429] = {.lex_state = 416},
  [5430] = {.lex_state = 0},
  [5431] = {.lex_state = 0},
  [5432] = {.lex_state = 0},
  [5433] = {.lex_state = 132},
  [5434] = {.lex_state = 0},
  [5435] = {.lex_state = 61},
  [5436] = {.lex_state = 0},
  [5437] = {.lex_state = 0},
  [5438] = {.lex_state = 0},
  [5439] = {.lex_state = 0},
  [5440] = {.lex_state = 0},
  [5441] = {.lex_state = 0},
  [5442] = {.lex_state = 0},
  [5443] = {.lex_state = 0},
  [5444] = {.lex_state = 0},
  [5445] = {.lex_state = 0},
  [5446] = {.lex_state = 416},
  [5447] = {.lex_state = 0},
  [5448] = {.lex_state = 0},
  [5449] = {.lex_state = 0},
  [5450] = {.lex_state = 416},
  [5451] = {.lex_state = 58},
  [5452] = {.lex_state = 0},
  [5453] = {.lex_state = 0},
  [5454] = {.lex_state = 0},
  [5455] = {.lex_state = 0},
  [5456] = {.lex_state = 0},
  [5457] = {.lex_state = 0},
  [5458] = {.lex_state = 0},
  [5459] = {.lex_state = 0},
  [5460] = {.lex_state = 0},
  [5461] = {.lex_state = 61},
  [5462] = {.lex_state = 0},
  [5463] = {.lex_state = 0},
  [5464] = {.lex_state = 416},
  [5465] = {.lex_state = 416},
  [5466] = {.lex_state = 0},
  [5467] = {.lex_state = 0},
  [5468] = {.lex_state = 0},
  [5469] = {.lex_state = 416},
  [5470] = {.lex_state = 0},
  [5471] = {.lex_state = 0},
  [5472] = {.lex_state = 0},
  [5473] = {.lex_state = 0},
  [5474] = {.lex_state = 0},
  [5475] = {.lex_state = 0},
  [5476] = {.lex_state = 0},
  [5477] = {.lex_state = 0},
  [5478] = {.lex_state = 0},
  [5479] = {.lex_state = 0},
  [5480] = {.lex_state = 0},
  [5481] = {.lex_state = 0},
  [5482] = {.lex_state = 0},
  [5483] = {.lex_state = 0},
  [5484] = {.lex_state = 0},
  [5485] = {.lex_state = 416},
  [5486] = {.lex_state = 0},
  [5487] = {.lex_state = 0},
  [5488] = {.lex_state = 0},
  [5489] = {.lex_state = 0},
  [5490] = {.lex_state = 416},
  [5491] = {.lex_state = 0},
  [5492] = {.lex_state = 0},
  [5493] = {.lex_state = 58},
  [5494] = {.lex_state = 0},
  [5495] = {.lex_state = 0},
  [5496] = {.lex_state = 0},
  [5497] = {.lex_state = 0},
  [5498] = {.lex_state = 416},
  [5499] = {.lex_state = 0},
  [5500] = {.lex_state = 0},
  [5501] = {.lex_state = 0},
  [5502] = {.lex_state = 0},
  [5503] = {.lex_state = 0},
  [5504] = {.lex_state = 0},
  [5505] = {.lex_state = 0},
  [5506] = {.lex_state = 0},
  [5507] = {.lex_state = 416},
  [5508] = {.lex_state = 0},
  [5509] = {.lex_state = 0},
  [5510] = {.lex_state = 0},
  [5511] = {.lex_state = 0},
  [5512] = {.lex_state = 0},
  [5513] = {.lex_state = 0},
  [5514] = {.lex_state = 416},
  [5515] = {.lex_state = 0},
  [5516] = {.lex_state = 58},
  [5517] = {.lex_state = 0},
  [5518] = {.lex_state = 0},
  [5519] = {.lex_state = 0},
  [5520] = {.lex_state = 0},
  [5521] = {.lex_state = 0},
  [5522] = {.lex_state = 416},
  [5523] = {.lex_state = 416},
  [5524] = {.lex_state = 416},
  [5525] = {.lex_state = 0},
  [5526] = {.lex_state = 0},
  [5527] = {.lex_state = 416},
  [5528] = {.lex_state = 0},
  [5529] = {.lex_state = 0},
  [5530] = {.lex_state = 0},
  [5531] = {.lex_state = 0},
  [5532] = {.lex_state = 0},
  [5533] = {.lex_state = 0},
  [5534] = {.lex_state = 132},
  [5535] = {.lex_state = 0},
  [5536] = {.lex_state = 132},
  [5537] = {.lex_state = 0},
  [5538] = {.lex_state = 0},
  [5539] = {.lex_state = 0},
  [5540] = {.lex_state = 0},
  [5541] = {.lex_state = 0},
  [5542] = {.lex_state = 416},
  [5543] = {.lex_state = 0},
  [5544] = {.lex_state = 0},
  [5545] = {.lex_state = 0},
  [5546] = {.lex_state = 0},
  [5547] = {.lex_state = 0},
  [5548] = {.lex_state = 416},
  [5549] = {.lex_state = 0},
  [5550] = {.lex_state = 416},
  [5551] = {.lex_state = 0},
  [5552] = {.lex_state = 0},
  [5553] = {.lex_state = 0},
  [5554] = {.lex_state = 416},
  [5555] = {.lex_state = 0},
  [5556] = {.lex_state = 0},
  [5557] = {.lex_state = 0},
  [5558] = {.lex_state = 132},
  [5559] = {.lex_state = 132},
  [5560] = {.lex_state = 416},
  [5561] = {.lex_state = 0},
  [5562] = {.lex_state = 0},
  [5563] = {.lex_state = 0},
  [5564] = {.lex_state = 0},
  [5565] = {.lex_state = 0},
  [5566] = {.lex_state = 0},
  [5567] = {.lex_state = 0},
  [5568] = {.lex_state = 0},
  [5569] = {.lex_state = 416},
  [5570] = {.lex_state = 416},
  [5571] = {.lex_state = 0},
  [5572] = {.lex_state = 0},
  [5573] = {.lex_state = 0},
  [5574] = {.lex_state = 0},
  [5575] = {.lex_state = 0},
  [5576] = {.lex_state = 416},
  [5577] = {.lex_state = 132},
  [5578] = {.lex_state = 132},
  [5579] = {.lex_state = 0},
  [5580] = {.lex_state = 132},
  [5581] = {.lex_state = 416},
  [5582] = {.lex_state = 132},
  [5583] = {.lex_state = 132},
  [5584] = {.lex_state = 132},
  [5585] = {.lex_state = 0},
  [5586] = {.lex_state = 416},
  [5587] = {.lex_state = 0},
  [5588] = {.lex_state = 416},
  [5589] = {.lex_state = 0},
  [5590] = {.lex_state = 0},
  [5591] = {.lex_state = 0},
  [5592] = {.lex_state = 0},
  [5593] = {.lex_state = 0},
  [5594] = {.lex_state = 0},
  [5595] = {.lex_state = 58},
  [5596] = {.lex_state = 0},
  [5597] = {.lex_state = 0},
  [5598] = {.lex_state = 416},
  [5599] = {.lex_state = 416},
  [5600] = {.lex_state = 416},
  [5601] = {.lex_state = 0},
  [5602] = {.lex_state = 0},
  [5603] = {.lex_state = 0},
  [5604] = {.lex_state = 416},
  [5605] = {.lex_state = 0},
  [5606] = {.lex_state = 0},
  [5607] = {.lex_state = 0},
  [5608] = {.lex_state = 0},
  [5609] = {.lex_state = 132},
  [5610] = {.lex_state = 61},
  [5611] = {.lex_state = 416},
  [5612] = {.lex_state = 0},
  [5613] = {.lex_state = 0},
  [5614] = {.lex_state = 0},
  [5615] = {.lex_state = 132},
  [5616] = {.lex_state = 0},
  [5617] = {.lex_state = 0},
  [5618] = {.lex_state = 416},
  [5619] = {.lex_state = 0},
  [5620] = {.lex_state = 0},
  [5621] = {.lex_state = 0},
  [5622] = {.lex_state = 0},
  [5623] = {.lex_state = 132},
  [5624] = {.lex_state = 416},
  [5625] = {.lex_state = 0},
  [5626] = {.lex_state = 0},
  [5627] = {.lex_state = 0},
  [5628] = {.lex_state = 416},
  [5629] = {.lex_state = 0},
  [5630] = {.lex_state = 416},
  [5631] = {.lex_state = 416},
  [5632] = {.lex_state = 0},
  [5633] = {.lex_state = 0},
  [5634] = {.lex_state = 0},
  [5635] = {.lex_state = 416},
  [5636] = {.lex_state = 416},
  [5637] = {.lex_state = 416},
  [5638] = {.lex_state = 0},
  [5639] = {.lex_state = 0},
  [5640] = {.lex_state = 0},
  [5641] = {.lex_state = 0},
  [5642] = {.lex_state = 416},
  [5643] = {.lex_state = 0},
  [5644] = {.lex_state = 0},
  [5645] = {.lex_state = 0},
  [5646] = {.lex_state = 416},
  [5647] = {.lex_state = 0},
  [5648] = {.lex_state = 0},
  [5649] = {.lex_state = 416},
  [5650] = {.lex_state = 0},
  [5651] = {.lex_state = 416},
  [5652] = {.lex_state = 0},
  [5653] = {.lex_state = 0},
  [5654] = {.lex_state = 0},
  [5655] = {.lex_state = 0},
  [5656] = {.lex_state = 416},
  [5657] = {.lex_state = 0},
  [5658] = {.lex_state = 0},
  [5659] = {.lex_state = 0},
  [5660] = {.lex_state = 0},
  [5661] = {.lex_state = 0},
  [5662] = {.lex_state = 0},
  [5663] = {.lex_state = 0},
  [5664] = {.lex_state = 416},
  [5665] = {.lex_state = 0},
  [5666] = {.lex_state = 132},
  [5667] = {.lex_state = 416},
  [5668] = {.lex_state = 416},
  [5669] = {.lex_state = 0},
  [5670] = {.lex_state = 0},
  [5671] = {.lex_state = 0},
  [5672] = {.lex_state = 0},
  [5673] = {.lex_state = 132},
  [5674] = {.lex_state = 0},
  [5675] = {.lex_state = 0},
  [5676] = {.lex_state = 416},
  [5677] = {.lex_state = 416},
  [5678] = {.lex_state = 0},
  [5679] = {.lex_state = 0},
  [5680] = {.lex_state = 0},
  [5681] = {.lex_state = 0},
  [5682] = {.lex_state = 132},
  [5683] = {.lex_state = 0},
  [5684] = {.lex_state = 0},
  [5685] = {.lex_state = 0},
  [5686] = {.lex_state = 0},
  [5687] = {.lex_state = 0},
  [5688] = {.lex_state = 0},
  [5689] = {.lex_state = 0},
  [5690] = {.lex_state = 0},
  [5691] = {.lex_state = 0},
  [5692] = {.lex_state = 416},
  [5693] = {.lex_state = 0},
  [5694] = {.lex_state = 0},
  [5695] = {.lex_state = 416},
  [5696] = {.lex_state = 0},
  [5697] = {.lex_state = 0},
  [5698] = {.lex_state = 0},
  [5699] = {.lex_state = 0},
  [5700] = {.lex_state = 0},
  [5701] = {.lex_state = 0},
  [5702] = {.lex_state = 0},
  [5703] = {.lex_state = 0},
  [5704] = {.lex_state = 416},
  [5705] = {.lex_state = 416},
  [5706] = {.lex_state = 0},
  [5707] = {.lex_state = 0},
  [5708] = {.lex_state = 0},
  [5709] = {.lex_state = 0},
  [5710] = {.lex_state = 0},
  [5711] = {.lex_state = 0},
  [5712] = {.lex_state = 0},
  [5713] = {.lex_state = 0},
  [5714] = {.lex_state = 0},
  [5715] = {.lex_state = 0},
  [5716] = {.lex_state = 416},
  [5717] = {.lex_state = 0},
  [5718] = {.lex_state = 132},
  [5719] = {.lex_state = 57},
  [5720] = {.lex_state = 416},
  [5721] = {.lex_state = 416},
  [5722] = {.lex_state = 416},
  [5723] = {.lex_state = 134},
  [5724] = {.lex_state = 416},
  [5725] = {.lex_state = 0},
  [5726] = {.lex_state = 0},
  [5727] = {.lex_state = 416},
  [5728] = {.lex_state = 137},
  [5729] = {.lex_state = 0},
  [5730] = {.lex_state = 58},
  [5731] = {.lex_state = 416},
  [5732] = {.lex_state = 132},
  [5733] = {.lex_state = 57},
  [5734] = {.lex_state = 132},
  [5735] = {.lex_state = 416},
  [5736] = {.lex_state = 0},
  [5737] = {.lex_state = 132},
  [5738] = {.lex_state = 57},
  [5739] = {.lex_state = 57},
  [5740] = {.lex_state = 0},
  [5741] = {.lex_state = 137},
  [5742] = {.lex_state = 58},
  [5743] = {.lex_state = 57},
  [5744] = {.lex_state = 0},
  [5745] = {.lex_state = 0},
  [5746] = {.lex_state = 98},
  [5747] = {.lex_state = 132},
  [5748] = {.lex_state = 416},
  [5749] = {.lex_state = 57},
  [5750] = {.lex_state = 0},
  [5751] = {.lex_state = 416},
  [5752] = {.lex_state = 416},
  [5753] = {.lex_state = 0},
  [5754] = {.lex_state = 0},
  [5755] = {.lex_state = 416},
  [5756] = {.lex_state = 0},
  [5757] = {.lex_state = 57},
  [5758] = {.lex_state = 0},
  [5759] = {.lex_state = 132},
  [5760] = {.lex_state = 416},
  [5761] = {.lex_state = 0},
  [5762] = {.lex_state = 0},
  [5763] = {.lex_state = 416},
  [5764] = {.lex_state = 416},
  [5765] = {.lex_state = 416},
  [5766] = {.lex_state = 57},
  [5767] = {.lex_state = 0},
  [5768] = {.lex_state = 132},
  [5769] = {.lex_state = 0},
  [5770] = {.lex_state = 416},
  [5771] = {.lex_state = 134},
  [5772] = {.lex_state = 0},
  [5773] = {.lex_state = 132},
  [5774] = {.lex_state = 0},
  [5775] = {.lex_state = 416},
  [5776] = {.lex_state = 0},
  [5777] = {.lex_state = 132},
  [5778] = {.lex_state = 132},
  [5779] = {.lex_state = 0},
  [5780] = {.lex_state = 0},
  [5781] = {.lex_state = 132},
  [5782] = {.lex_state = 416},
  [5783] = {.lex_state = 58},
  [5784] = {.lex_state = 0},
  [5785] = {.lex_state = 416},
  [5786] = {.lex_state = 57},
  [5787] = {.lex_state = 0},
  [5788] = {.lex_state = 0},
  [5789] = {.lex_state = 132},
  [5790] = {.lex_state = 58},
  [5791] = {.lex_state = 416},
  [5792] = {.lex_state = 57},
  [5793] = {.lex_state = 0},
  [5794] = {.lex_state = 0},
  [5795] = {.lex_state = 58},
  [5796] = {.lex_state = 0},
  [5797] = {.lex_state = 0},
  [5798] = {.lex_state = 132},
  [5799] = {.lex_state = 0},
  [5800] = {.lex_state = 0},
  [5801] = {.lex_state = 416},
  [5802] = {.lex_state = 0},
  [5803] = {.lex_state = 416},
  [5804] = {.lex_state = 58},
  [5805] = {.lex_state = 416},
  [5806] = {.lex_state = 416},
  [5807] = {.lex_state = 0},
  [5808] = {.lex_state = 416},
  [5809] = {.lex_state = 0},
  [5810] = {.lex_state = 58},
  [5811] = {.lex_state = 0},
  [5812] = {.lex_state = 416},
  [5813] = {.lex_state = 57},
  [5814] = {.lex_state = 132},
  [5815] = {.lex_state = 132},
  [5816] = {.lex_state = 416},
  [5817] = {.lex_state = 57},
  [5818] = {.lex_state = 416},
  [5819] = {.lex_state = 416},
  [5820] = {.lex_state = 132},
  [5821] = {.lex_state = 416},
  [5822] = {.lex_state = 57},
  [5823] = {.lex_state = 58},
  [5824] = {.lex_state = 0},
  [5825] = {.lex_state = 132},
  [5826] = {.lex_state = 416},
  [5827] = {.lex_state = 416},
  [5828] = {.lex_state = 0},
  [5829] = {.lex_state = 57},
  [5830] = {.lex_state = 132},
  [5831] = {.lex_state = 416},
  [5832] = {.lex_state = 58},
  [5833] = {.lex_state = 0},
  [5834] = {.lex_state = 137},
  [5835] = {.lex_state = 132},
  [5836] = {.lex_state = 132},
  [5837] = {.lex_state = 132},
  [5838] = {.lex_state = 57},
  [5839] = {.lex_state = 416},
  [5840] = {.lex_state = 416},
  [5841] = {.lex_state = 0},
  [5842] = {.lex_state = 132},
  [5843] = {.lex_state = 58},
  [5844] = {.lex_state = 416},
  [5845] = {.lex_state = 132},
  [5846] = {.lex_state = 0},
  [5847] = {.lex_state = 0},
  [5848] = {.lex_state = 0},
  [5849] = {.lex_state = 0},
  [5850] = {.lex_state = 132},
  [5851] = {.lex_state = 132},
  [5852] = {.lex_state = 0},
  [5853] = {.lex_state = 57},
  [5854] = {.lex_state = 58},
  [5855] = {.lex_state = 0},
  [5856] = {.lex_state = 98},
  [5857] = {.lex_state = 132},
  [5858] = {.lex_state = 58},
  [5859] = {.lex_state = 132},
  [5860] = {.lex_state = 0},
  [5861] = {.lex_state = 0},
  [5862] = {.lex_state = 0},
  [5863] = {.lex_state = 416},
  [5864] = {.lex_state = 132},
  [5865] = {.lex_state = 132},
  [5866] = {.lex_state = 416},
  [5867] = {.lex_state = 0},
  [5868] = {.lex_state = 416},
  [5869] = {.lex_state = 416},
  [5870] = {.lex_state = 137},
  [5871] = {.lex_state = 57},
  [5872] = {.lex_state = 416},
  [5873] = {.lex_state = 58},
  [5874] = {.lex_state = 416},
  [5875] = {.lex_state = 416},
  [5876] = {.lex_state = 0},
  [5877] = {.lex_state = 132},
  [5878] = {.lex_state = 137},
  [5879] = {.lex_state = 57},
  [5880] = {.lex_state = 132},
  [5881] = {.lex_state = 132},
  [5882] = {.lex_state = 57},
  [5883] = {.lex_state = 57},
  [5884] = {.lex_state = 137},
  [5885] = {.lex_state = 57},
  [5886] = {.lex_state = 416},
  [5887] = {.lex_state = 0},
  [5888] = {.lex_state = 0},
  [5889] = {.lex_state = 416},
  [5890] = {.lex_state = 132},
  [5891] = {.lex_state = 57},
  [5892] = {.lex_state = 0},
  [5893] = {.lex_state = 416},
  [5894] = {.lex_state = 57},
  [5895] = {.lex_state = 57},
  [5896] = {.lex_state = 416},
  [5897] = {.lex_state = 0},
  [5898] = {.lex_state = 0},
  [5899] = {.lex_state = 57},
  [5900] = {.lex_state = 0},
  [5901] = {.lex_state = 0},
  [5902] = {.lex_state = 0},
  [5903] = {.lex_state = 0},
  [5904] = {.lex_state = 0},
  [5905] = {.lex_state = 0},
  [5906] = {.lex_state = 132},
  [5907] = {.lex_state = 105},
  [5908] = {.lex_state = 0},
  [5909] = {.lex_state = 0},
  [5910] = {.lex_state = 105},
  [5911] = {.lex_state = 105},
  [5912] = {.lex_state = 0},
  [5913] = {.lex_state = 0},
  [5914] = {.lex_state = 416},
  [5915] = {.lex_state = 0},
  [5916] = {.lex_state = 132},
  [5917] = {.lex_state = 132},
  [5918] = {.lex_state = 58},
  [5919] = {.lex_state = 0},
  [5920] = {.lex_state = 0},
  [5921] = {.lex_state = 0},
  [5922] = {.lex_state = 0},
  [5923] = {.lex_state = 58},
  [5924] = {.lex_state = 0},
  [5925] = {.lex_state = 132},
  [5926] = {.lex_state = 105},
  [5927] = {.lex_state = 132},
  [5928] = {.lex_state = 132},
  [5929] = {.lex_state = 132},
  [5930] = {.lex_state = 132},
  [5931] = {.lex_state = 416},
  [5932] = {.lex_state = 0},
  [5933] = {.lex_state = 0},
  [5934] = {.lex_state = 0},
  [5935] = {.lex_state = 0},
  [5936] = {.lex_state = 0},
  [5937] = {.lex_state = 0},
  [5938] = {.lex_state = 132},
  [5939] = {.lex_state = 0},
  [5940] = {.lex_state = 0},
  [5941] = {.lex_state = 132},
  [5942] = {.lex_state = 58},
  [5943] = {.lex_state = 0},
  [5944] = {.lex_state = 416},
  [5945] = {.lex_state = 0},
  [5946] = {.lex_state = 132},
  [5947] = {.lex_state = 0},
  [5948] = {.lex_state = 132},
  [5949] = {.lex_state = 132},
  [5950] = {.lex_state = 105},
  [5951] = {.lex_state = 132},
  [5952] = {.lex_state = 0},
  [5953] = {.lex_state = 58},
  [5954] = {.lex_state = 0},
  [5955] = {.lex_state = 105},
  [5956] = {.lex_state = 726},
  [5957] = {.lex_state = 132},
  [5958] = {.lex_state = 0},
  [5959] = {.lex_state = 0},
  [5960] = {.lex_state = 0},
  [5961] = {.lex_state = 0},
  [5962] = {.lex_state = 0},
  [5963] = {.lex_state = 0},
  [5964] = {.lex_state = 0},
  [5965] = {.lex_state = 132},
  [5966] = {.lex_state = 0},
  [5967] = {.lex_state = 0},
  [5968] = {.lex_state = 0},
  [5969] = {.lex_state = 416},
  [5970] = {.lex_state = 0},
  [5971] = {.lex_state = 0},
  [5972] = {.lex_state = 0},
  [5973] = {.lex_state = 132},
  [5974] = {.lex_state = 132},
  [5975] = {.lex_state = 58},
  [5976] = {.lex_state = 0},
  [5977] = {.lex_state = 0},
  [5978] = {.lex_state = 0},
  [5979] = {.lex_state = 132},
  [5980] = {.lex_state = 0},
  [5981] = {.lex_state = 0},
  [5982] = {.lex_state = 0},
  [5983] = {.lex_state = 132},
  [5984] = {.lex_state = 0},
  [5985] = {.lex_state = 0},
  [5986] = {.lex_state = 132},
  [5987] = {.lex_state = 132},
  [5988] = {.lex_state = 0},
  [5989] = {.lex_state = 0},
  [5990] = {.lex_state = 0},
  [5991] = {.lex_state = 0},
  [5992] = {.lex_state = 0},
  [5993] = {.lex_state = 0},
  [5994] = {.lex_state = 0},
  [5995] = {.lex_state = 132},
  [5996] = {.lex_state = 132},
  [5997] = {.lex_state = 132},
  [5998] = {.lex_state = 132},
  [5999] = {.lex_state = 132},
  [6000] = {.lex_state = 0},
  [6001] = {.lex_state = 0},
  [6002] = {.lex_state = 0},
  [6003] = {.lex_state = 0},
  [6004] = {.lex_state = 0},
  [6005] = {.lex_state = 0},
  [6006] = {.lex_state = 0},
  [6007] = {.lex_state = 132},
  [6008] = {.lex_state = 132},
  [6009] = {.lex_state = 0},
  [6010] = {.lex_state = 132},
  [6011] = {.lex_state = 416},
  [6012] = {.lex_state = 0},
  [6013] = {.lex_state = 132},
  [6014] = {.lex_state = 0},
  [6015] = {.lex_state = 132},
  [6016] = {.lex_state = 0},
  [6017] = {.lex_state = 0},
  [6018] = {.lex_state = 0},
  [6019] = {.lex_state = 0},
  [6020] = {.lex_state = 726},
  [6021] = {.lex_state = 0},
  [6022] = {.lex_state = 105},
  [6023] = {.lex_state = 0},
  [6024] = {.lex_state = 132},
  [6025] = {.lex_state = 0},
  [6026] = {.lex_state = 132},
  [6027] = {.lex_state = 58},
  [6028] = {.lex_state = 0},
  [6029] = {.lex_state = 0},
  [6030] = {.lex_state = 0},
  [6031] = {.lex_state = 0},
  [6032] = {.lex_state = 416},
  [6033] = {.lex_state = 105},
  [6034] = {.lex_state = 132},
  [6035] = {.lex_state = 0},
  [6036] = {.lex_state = 0},
  [6037] = {.lex_state = 0},
  [6038] = {.lex_state = 0},
  [6039] = {.lex_state = 0},
  [6040] = {.lex_state = 0},
  [6041] = {.lex_state = 0},
  [6042] = {.lex_state = 0},
  [6043] = {.lex_state = 58},
  [6044] = {.lex_state = 132},
  [6045] = {.lex_state = 132},
  [6046] = {.lex_state = 0},
  [6047] = {.lex_state = 0},
  [6048] = {.lex_state = 0},
  [6049] = {.lex_state = 0},
  [6050] = {.lex_state = 0},
  [6051] = {.lex_state = 132},
  [6052] = {.lex_state = 132},
  [6053] = {.lex_state = 0},
  [6054] = {.lex_state = 0},
  [6055] = {.lex_state = 0},
  [6056] = {.lex_state = 0},
  [6057] = {.lex_state = 0},
  [6058] = {.lex_state = 132},
  [6059] = {.lex_state = 0},
  [6060] = {.lex_state = 132},
  [6061] = {.lex_state = 0},
  [6062] = {.lex_state = 0},
  [6063] = {.lex_state = 416},
  [6064] = {.lex_state = 0},
  [6065] = {.lex_state = 0},
  [6066] = {.lex_state = 726},
  [6067] = {.lex_state = 105},
  [6068] = {.lex_state = 0},
  [6069] = {.lex_state = 0},
  [6070] = {.lex_state = 0},
  [6071] = {.lex_state = 132},
  [6072] = {.lex_state = 0},
  [6073] = {.lex_state = 0},
  [6074] = {.lex_state = 0},
  [6075] = {.lex_state = 132},
  [6076] = {.lex_state = 0},
  [6077] = {.lex_state = 0},
  [6078] = {.lex_state = 0},
  [6079] = {.lex_state = 132},
  [6080] = {.lex_state = 132},
  [6081] = {.lex_state = 0},
  [6082] = {.lex_state = 0},
  [6083] = {.lex_state = 0},
  [6084] = {.lex_state = 0},
  [6085] = {.lex_state = 0},
  [6086] = {.lex_state = 0},
  [6087] = {.lex_state = 416},
  [6088] = {.lex_state = 132},
  [6089] = {.lex_state = 132},
  [6090] = {.lex_state = 0},
  [6091] = {.lex_state = 0},
  [6092] = {.lex_state = 132},
  [6093] = {.lex_state = 132},
  [6094] = {.lex_state = 0},
  [6095] = {.lex_state = 58},
  [6096] = {.lex_state = 0},
  [6097] = {.lex_state = 105},
  [6098] = {.lex_state = 132},
  [6099] = {.lex_state = 105},
  [6100] = {.lex_state = 416},
  [6101] = {.lex_state = 0},
  [6102] = {.lex_state = 0},
  [6103] = {.lex_state = 726},
  [6104] = {.lex_state = 58},
  [6105] = {.lex_state = 105},
  [6106] = {.lex_state = 0},
  [6107] = {.lex_state = 416},
  [6108] = {.lex_state = 132},
  [6109] = {.lex_state = 0},
  [6110] = {.lex_state = 132},
  [6111] = {.lex_state = 0},
  [6112] = {.lex_state = 0},
  [6113] = {.lex_state = 105},
  [6114] = {.lex_state = 0},
  [6115] = {.lex_state = 132},
  [6116] = {.lex_state = 0},
  [6117] = {.lex_state = 0},
  [6118] = {.lex_state = 0},
  [6119] = {.lex_state = 132},
  [6120] = {.lex_state = 132},
  [6121] = {.lex_state = 0},
  [6122] = {.lex_state = 132},
  [6123] = {.lex_state = 416},
  [6124] = {.lex_state = 132},
  [6125] = {.lex_state = 0},
  [6126] = {.lex_state = 0},
  [6127] = {.lex_state = 132},
  [6128] = {.lex_state = 132},
  [6129] = {.lex_state = 0},
  [6130] = {.lex_state = 0},
  [6131] = {.lex_state = 0},
  [6132] = {.lex_state = 0},
  [6133] = {.lex_state = 0},
  [6134] = {.lex_state = 0},
  [6135] = {.lex_state = 132},
  [6136] = {.lex_state = 105},
  [6137] = {.lex_state = 132},
  [6138] = {.lex_state = 132},
  [6139] = {.lex_state = 0},
  [6140] = {.lex_state = 0},
  [6141] = {.lex_state = 132},
  [6142] = {.lex_state = 0},
  [6143] = {.lex_state = 58},
  [6144] = {.lex_state = 132},
  [6145] = {.lex_state = 0},
  [6146] = {.lex_state = 105},
  [6147] = {.lex_state = 58},
  [6148] = {.lex_state = 105},
  [6149] = {.lex_state = 58},
  [6150] = {.lex_state = 105},
  [6151] = {.lex_state = 0},
  [6152] = {.lex_state = 0},
  [6153] = {.lex_state = 132},
  [6154] = {.lex_state = 0},
  [6155] = {.lex_state = 0},
  [6156] = {.lex_state = 0},
  [6157] = {.lex_state = 0},
  [6158] = {.lex_state = 0},
  [6159] = {.lex_state = 0},
  [6160] = {.lex_state = 0},
  [6161] = {.lex_state = 0},
  [6162] = {.lex_state = 0},
  [6163] = {.lex_state = 0},
  [6164] = {.lex_state = 132},
  [6165] = {.lex_state = 0},
  [6166] = {.lex_state = 0},
  [6167] = {.lex_state = 0},
  [6168] = {.lex_state = 0},
  [6169] = {.lex_state = 0},
  [6170] = {.lex_state = 0},
  [6171] = {.lex_state = 0},
  [6172] = {.lex_state = 0},
  [6173] = {.lex_state = 0},
  [6174] = {.lex_state = 0},
  [6175] = {.lex_state = 0},
  [6176] = {.lex_state = 416},
  [6177] = {.lex_state = 416},
  [6178] = {.lex_state = 416},
  [6179] = {.lex_state = 416},
  [6180] = {.lex_state = 132},
  [6181] = {.lex_state = 416},
  [6182] = {.lex_state = 0},
  [6183] = {.lex_state = 0},
  [6184] = {.lex_state = 416},
  [6185] = {.lex_state = 416},
  [6186] = {.lex_state = 416},
  [6187] = {.lex_state = 0},
  [6188] = {.lex_state = 416},
  [6189] = {.lex_state = 416},
  [6190] = {.lex_state = 105},
  [6191] = {.lex_state = 0},
  [6192] = {.lex_state = 132},
  [6193] = {.lex_state = 132},
  [6194] = {.lex_state = 132},
  [6195] = {.lex_state = 0},
  [6196] = {.lex_state = 105},
  [6197] = {.lex_state = 0},
  [6198] = {.lex_state = 0},
  [6199] = {.lex_state = 0},
  [6200] = {.lex_state = 58},
  [6201] = {.lex_state = 0},
  [6202] = {.lex_state = 132},
  [6203] = {.lex_state = 0},
  [6204] = {.lex_state = 0},
  [6205] = {.lex_state = 132},
  [6206] = {.lex_state = 0},
  [6207] = {.lex_state = 0},
  [6208] = {.lex_state = 132},
  [6209] = {.lex_state = 0},
  [6210] = {.lex_state = 132},
  [6211] = {.lex_state = 0},
  [6212] = {.lex_state = 0},
  [6213] = {.lex_state = 0},
  [6214] = {.lex_state = 0},
  [6215] = {.lex_state = 0},
  [6216] = {.lex_state = 132},
  [6217] = {.lex_state = 0},
  [6218] = {.lex_state = 0},
  [6219] = {.lex_state = 132},
  [6220] = {.lex_state = 132},
  [6221] = {.lex_state = 132},
  [6222] = {.lex_state = 0},
  [6223] = {.lex_state = 0},
  [6224] = {.lex_state = 0},
  [6225] = {.lex_state = 132},
  [6226] = {.lex_state = 105},
  [6227] = {.lex_state = 0},
  [6228] = {.lex_state = 0},
  [6229] = {.lex_state = 105},
  [6230] = {.lex_state = 0},
  [6231] = {.lex_state = 132},
  [6232] = {.lex_state = 0},
  [6233] = {.lex_state = 416},
  [6234] = {.lex_state = 416},
  [6235] = {.lex_state = 0},
  [6236] = {.lex_state = 0},
  [6237] = {.lex_state = 105},
  [6238] = {.lex_state = 0},
  [6239] = {.lex_state = 0},
  [6240] = {.lex_state = 0},
  [6241] = {.lex_state = 105},
  [6242] = {.lex_state = 416},
  [6243] = {.lex_state = 0},
  [6244] = {.lex_state = 0},
  [6245] = {.lex_state = 58},
  [6246] = {.lex_state = 105},
  [6247] = {.lex_state = 58},
  [6248] = {.lex_state = 105},
  [6249] = {.lex_state = 0},
  [6250] = {.lex_state = 0},
  [6251] = {.lex_state = 0},
  [6252] = {.lex_state = 0},
  [6253] = {.lex_state = 0},
  [6254] = {.lex_state = 0},
  [6255] = {.lex_state = 0},
  [6256] = {.lex_state = 0},
  [6257] = {.lex_state = 0},
  [6258] = {.lex_state = 0},
  [6259] = {.lex_state = 132},
  [6260] = {.lex_state = 0},
  [6261] = {.lex_state = 58},
  [6262] = {.lex_state = 0},
  [6263] = {.lex_state = 132},
  [6264] = {.lex_state = 58},
  [6265] = {.lex_state = 105},
  [6266] = {.lex_state = 132},
  [6267] = {.lex_state = 0},
  [6268] = {.lex_state = 416},
  [6269] = {.lex_state = 132},
  [6270] = {.lex_state = 0},
  [6271] = {.lex_state = 105},
  [6272] = {.lex_state = 0},
  [6273] = {.lex_state = 0},
  [6274] = {.lex_state = 58},
  [6275] = {.lex_state = 132},
  [6276] = {.lex_state = 0},
  [6277] = {.lex_state = 416},
  [6278] = {.lex_state = 0},
  [6279] = {.lex_state = 0},
  [6280] = {.lex_state = 132},
  [6281] = {.lex_state = 58},
  [6282] = {.lex_state = 105},
  [6283] = {.lex_state = 0},
  [6284] = {.lex_state = 132},
  [6285] = {.lex_state = 58},
  [6286] = {.lex_state = 58},
  [6287] = {.lex_state = 416},
  [6288] = {.lex_state = 105},
  [6289] = {.lex_state = 0},
  [6290] = {.lex_state = 0},
  [6291] = {.lex_state = 105},
  [6292] = {.lex_state = 132},
  [6293] = {.lex_state = 0},
  [6294] = {.lex_state = 58},
  [6295] = {.lex_state = 0},
  [6296] = {.lex_state = 132},
  [6297] = {.lex_state = 0},
  [6298] = {.lex_state = 0},
  [6299] = {.lex_state = 416},
  [6300] = {.lex_state = 105},
  [6301] = {.lex_state = 0},
  [6302] = {.lex_state = 105},
  [6303] = {.lex_state = 58},
  [6304] = {.lex_state = 0},
  [6305] = {.lex_state = 0},
  [6306] = {.lex_state = 132},
  [6307] = {.lex_state = 105},
  [6308] = {.lex_state = 416},
  [6309] = {.lex_state = 0},
  [6310] = {.lex_state = 0},
  [6311] = {.lex_state = 0},
  [6312] = {.lex_state = 132},
  [6313] = {.lex_state = 0},
  [6314] = {.lex_state = 0},
  [6315] = {.lex_state = 105},
  [6316] = {.lex_state = 0},
  [6317] = {.lex_state = 132},
  [6318] = {.lex_state = 0},
  [6319] = {.lex_state = 132},
  [6320] = {.lex_state = 0},
  [6321] = {.lex_state = 132},
  [6322] = {.lex_state = 132},
  [6323] = {.lex_state = 0},
  [6324] = {.lex_state = 0},
  [6325] = {.lex_state = 0},
  [6326] = {.lex_state = 132},
  [6327] = {.lex_state = 132},
  [6328] = {.lex_state = 0},
  [6329] = {.lex_state = 0},
  [6330] = {.lex_state = 105},
  [6331] = {.lex_state = 0},
  [6332] = {.lex_state = 0},
  [6333] = {.lex_state = 0},
  [6334] = {.lex_state = 0},
  [6335] = {.lex_state = 0},
  [6336] = {.lex_state = 0},
  [6337] = {.lex_state = 416},
  [6338] = {.lex_state = 0},
  [6339] = {.lex_state = 0},
  [6340] = {.lex_state = 0},
  [6341] = {.lex_state = 105},
  [6342] = {.lex_state = 0},
  [6343] = {.lex_state = 0},
  [6344] = {.lex_state = 0},
  [6345] = {.lex_state = 0},
  [6346] = {.lex_state = 0},
  [6347] = {.lex_state = 0},
  [6348] = {.lex_state = 0},
  [6349] = {.lex_state = 726},
  [6350] = {.lex_state = 0},
  [6351] = {.lex_state = 0},
  [6352] = {.lex_state = 0},
  [6353] = {.lex_state = 0},
  [6354] = {.lex_state = 0},
  [6355] = {.lex_state = 416},
  [6356] = {.lex_state = 0},
  [6357] = {.lex_state = 0},
  [6358] = {.lex_state = 0},
  [6359] = {.lex_state = 0},
  [6360] = {.lex_state = 132},
  [6361] = {.lex_state = 0},
  [6362] = {.lex_state = 0},
  [6363] = {.lex_state = 132},
  [6364] = {.lex_state = 416},
  [6365] = {.lex_state = 105},
  [6366] = {.lex_state = 0},
  [6367] = {.lex_state = 416},
  [6368] = {.lex_state = 0},
  [6369] = {.lex_state = 0},
  [6370] = {.lex_state = 0},
  [6371] = {.lex_state = 0},
  [6372] = {.lex_state = 0},
  [6373] = {.lex_state = 0},
  [6374] = {.lex_state = 0},
  [6375] = {.lex_state = 416},
  [6376] = {.lex_state = 416},
  [6377] = {.lex_state = 416},
  [6378] = {.lex_state = 58},
  [6379] = {.lex_state = 0},
  [6380] = {.lex_state = 416},
  [6381] = {.lex_state = 132},
  [6382] = {.lex_state = 105},
  [6383] = {.lex_state = 0},
  [6384] = {.lex_state = 105},
  [6385] = {.lex_state = 0},
  [6386] = {.lex_state = 416},
  [6387] = {.lex_state = 0},
  [6388] = {.lex_state = 132},
  [6389] = {.lex_state = 58},
  [6390] = {.lex_state = 0},
  [6391] = {.lex_state = 0},
  [6392] = {.lex_state = 132},
  [6393] = {.lex_state = 132},
  [6394] = {.lex_state = 105},
  [6395] = {.lex_state = 0},
  [6396] = {.lex_state = 0},
  [6397] = {.lex_state = 0},
  [6398] = {.lex_state = 58},
  [6399] = {.lex_state = 0},
  [6400] = {.lex_state = 0},
  [6401] = {.lex_state = 0},
  [6402] = {.lex_state = 0},
  [6403] = {.lex_state = 0},
  [6404] = {.lex_state = 0},
  [6405] = {.lex_state = 0},
  [6406] = {.lex_state = 132},
  [6407] = {.lex_state = 0},
  [6408] = {.lex_state = 0},
  [6409] = {.lex_state = 105},
  [6410] = {.lex_state = 132},
  [6411] = {.lex_state = 0},
  [6412] = {.lex_state = 132},
  [6413] = {.lex_state = 132},
  [6414] = {.lex_state = 0},
  [6415] = {.lex_state = 132},
  [6416] = {.lex_state = 0},
  [6417] = {.lex_state = 0},
  [6418] = {.lex_state = 416},
  [6419] = {.lex_state = 0},
  [6420] = {.lex_state = 0},
  [6421] = {.lex_state = 0},
  [6422] = {.lex_state = 0},
  [6423] = {.lex_state = 132},
  [6424] = {.lex_state = 0},
  [6425] = {.lex_state = 416},
  [6426] = {.lex_state = 0},
  [6427] = {.lex_state = 58},
  [6428] = {.lex_state = 105},
  [6429] = {.lex_state = 58},
  [6430] = {.lex_state = 0},
  [6431] = {.lex_state = 0},
  [6432] = {.lex_state = 0},
  [6433] = {.lex_state = 132},
  [6434] = {.lex_state = 0},
  [6435] = {.lex_state = 105},
  [6436] = {.lex_state = 0},
  [6437] = {.lex_state = 0},
  [6438] = {.lex_state = 58},
  [6439] = {.lex_state = 0},
  [6440] = {.lex_state = 132},
  [6441] = {.lex_state = 416},
  [6442] = {.lex_state = 0},
  [6443] = {.lex_state = 0},
  [6444] = {.lex_state = 0},
  [6445] = {.lex_state = 132},
  [6446] = {.lex_state = 416},
  [6447] = {.lex_state = 416},
  [6448] = {.lex_state = 105},
  [6449] = {.lex_state = 132},
  [6450] = {.lex_state = 416},
  [6451] = {.lex_state = 416},
  [6452] = {.lex_state = 132},
  [6453] = {.lex_state = 0},
  [6454] = {.lex_state = 105},
  [6455] = {.lex_state = 58},
  [6456] = {.lex_state = 0},
  [6457] = {.lex_state = 0},
  [6458] = {.lex_state = 0},
  [6459] = {.lex_state = 132},
  [6460] = {.lex_state = 0},
  [6461] = {.lex_state = 58},
  [6462] = {.lex_state = 0},
  [6463] = {.lex_state = 58},
  [6464] = {.lex_state = 0},
  [6465] = {.lex_state = 105},
  [6466] = {.lex_state = 0},
  [6467] = {.lex_state = 105},
  [6468] = {.lex_state = 132},
  [6469] = {.lex_state = 105},
  [6470] = {.lex_state = 0},
  [6471] = {.lex_state = 0},
  [6472] = {.lex_state = 0},
  [6473] = {.lex_state = 0},
  [6474] = {.lex_state = 58},
  [6475] = {.lex_state = 0},
  [6476] = {.lex_state = 132},
  [6477] = {.lex_state = 416},
  [6478] = {.lex_state = 416},
  [6479] = {.lex_state = 0},
  [6480] = {.lex_state = 416},
  [6481] = {.lex_state = 0},
  [6482] = {.lex_state = 105},
  [6483] = {.lex_state = 0},
  [6484] = {.lex_state = 0},
  [6485] = {.lex_state = 0},
  [6486] = {.lex_state = 0},
  [6487] = {.lex_state = 132},
  [6488] = {.lex_state = 132},
  [6489] = {.lex_state = 0},
  [6490] = {.lex_state = 0},
  [6491] = {.lex_state = 0},
  [6492] = {.lex_state = 105},
  [6493] = {.lex_state = 0},
  [6494] = {.lex_state = 0},
  [6495] = {.lex_state = 0},
  [6496] = {.lex_state = 132},
  [6497] = {.lex_state = 0},
  [6498] = {.lex_state = 0},
  [6499] = {.lex_state = 132},
  [6500] = {.lex_state = 0},
  [6501] = {.lex_state = 132},
  [6502] = {.lex_state = 0},
  [6503] = {.lex_state = 416},
  [6504] = {.lex_state = 105},
  [6505] = {.lex_state = 0},
  [6506] = {.lex_state = 0},
  [6507] = {.lex_state = 416},
  [6508] = {.lex_state = 132},
  [6509] = {.lex_state = 0},
  [6510] = {.lex_state = 0},
  [6511] = {.lex_state = 0},
  [6512] = {.lex_state = 132},
  [6513] = {.lex_state = 132},
  [6514] = {.lex_state = 0},
  [6515] = {.lex_state = 0},
  [6516] = {.lex_state = 0},
  [6517] = {.lex_state = 132},
  [6518] = {.lex_state = 0},
  [6519] = {.lex_state = 0},
  [6520] = {.lex_state = 0},
  [6521] = {.lex_state = 132},
  [6522] = {.lex_state = 0},
  [6523] = {.lex_state = 416},
  [6524] = {.lex_state = 0},
  [6525] = {.lex_state = 416},
  [6526] = {.lex_state = 0},
  [6527] = {.lex_state = 0},
  [6528] = {.lex_state = 132},
  [6529] = {.lex_state = 105},
  [6530] = {.lex_state = 0},
  [6531] = {.lex_state = 416},
  [6532] = {.lex_state = 0},
  [6533] = {.lex_state = 0},
  [6534] = {.lex_state = 0},
  [6535] = {.lex_state = 0},
  [6536] = {.lex_state = 58},
  [6537] = {.lex_state = 0},
  [6538] = {.lex_state = 0},
  [6539] = {.lex_state = 0},
  [6540] = {.lex_state = 58},
  [6541] = {.lex_state = 132},
  [6542] = {.lex_state = 0},
  [6543] = {.lex_state = 0},
  [6544] = {.lex_state = 0},
  [6545] = {.lex_state = 0},
  [6546] = {.lex_state = 0},
  [6547] = {.lex_state = 0},
  [6548] = {.lex_state = 132},
  [6549] = {.lex_state = 0},
  [6550] = {.lex_state = 0},
  [6551] = {.lex_state = 0},
  [6552] = {.lex_state = 0},
  [6553] = {.lex_state = 0},
  [6554] = {.lex_state = 0},
  [6555] = {.lex_state = 58},
  [6556] = {.lex_state = 132},
  [6557] = {.lex_state = 132},
  [6558] = {.lex_state = 0},
  [6559] = {.lex_state = 132},
  [6560] = {.lex_state = 132},
  [6561] = {.lex_state = 105},
  [6562] = {.lex_state = 0},
  [6563] = {.lex_state = 0},
  [6564] = {.lex_state = 132},
  [6565] = {.lex_state = 0},
  [6566] = {.lex_state = 0},
  [6567] = {.lex_state = 58},
  [6568] = {.lex_state = 105},
  [6569] = {.lex_state = 0},
  [6570] = {.lex_state = 105},
  [6571] = {.lex_state = 105},
  [6572] = {.lex_state = 0},
  [6573] = {.lex_state = 0},
  [6574] = {.lex_state = 0},
  [6575] = {.lex_state = 0},
  [6576] = {.lex_state = 416},
  [6577] = {.lex_state = 0},
  [6578] = {.lex_state = 0},
  [6579] = {.lex_state = 105},
  [6580] = {.lex_state = 0},
  [6581] = {.lex_state = 0},
  [6582] = {.lex_state = 0},
  [6583] = {.lex_state = 0},
  [6584] = {.lex_state = 0},
  [6585] = {.lex_state = 132},
  [6586] = {.lex_state = 0},
  [6587] = {.lex_state = 416},
  [6588] = {.lex_state = 58},
  [6589] = {.lex_state = 0},
  [6590] = {.lex_state = 0},
  [6591] = {.lex_state = 132},
  [6592] = {.lex_state = 0},
  [6593] = {.lex_state = 0},
  [6594] = {.lex_state = 132},
  [6595] = {.lex_state = 132},
  [6596] = {.lex_state = 132},
  [6597] = {.lex_state = 0},
  [6598] = {.lex_state = 0},
  [6599] = {.lex_state = 416},
  [6600] = {.lex_state = 132},
  [6601] = {.lex_state = 132},
  [6602] = {.lex_state = 0},
  [6603] = {.lex_state = 0},
  [6604] = {.lex_state = 0},
  [6605] = {.lex_state = 105},
  [6606] = {.lex_state = 132},
  [6607] = {.lex_state = 58},
  [6608] = {.lex_state = 0},
  [6609] = {.lex_state = 0},
  [6610] = {.lex_state = 105},
  [6611] = {.lex_state = 0},
  [6612] = {.lex_state = 0},
  [6613] = {.lex_state = 132},
  [6614] = {.lex_state = 0},
  [6615] = {.lex_state = 0},
  [6616] = {.lex_state = 0},
  [6617] = {.lex_state = 0},
  [6618] = {.lex_state = 0},
  [6619] = {.lex_state = 0},
  [6620] = {.lex_state = 0},
  [6621] = {.lex_state = 0},
  [6622] = {.lex_state = 0},
  [6623] = {.lex_state = 416},
  [6624] = {.lex_state = 132},
  [6625] = {.lex_state = 0},
  [6626] = {.lex_state = 132},
  [6627] = {.lex_state = 0},
  [6628] = {.lex_state = 0},
  [6629] = {.lex_state = 0},
  [6630] = {.lex_state = 132},
  [6631] = {.lex_state = 0},
  [6632] = {.lex_state = 58},
  [6633] = {.lex_state = 726},
  [6634] = {.lex_state = 58},
  [6635] = {.lex_state = 105},
  [6636] = {.lex_state = 105},
  [6637] = {.lex_state = 132},
  [6638] = {.lex_state = 0},
  [6639] = {.lex_state = 132},
  [6640] = {.lex_state = 0},
  [6641] = {.lex_state = 0},
  [6642] = {.lex_state = 0},
  [6643] = {.lex_state = 132},
  [6644] = {.lex_state = 105},
  [6645] = {.lex_state = 416},
  [6646] = {.lex_state = 0},
  [6647] = {.lex_state = 132},
  [6648] = {.lex_state = 0},
  [6649] = {.lex_state = 105},
  [6650] = {.lex_state = 0},
  [6651] = {.lex_state = 0},
  [6652] = {.lex_state = 0},
  [6653] = {.lex_state = 0},
  [6654] = {.lex_state = 0},
  [6655] = {.lex_state = 0},
  [6656] = {.lex_state = 0},
  [6657] = {.lex_state = 0},
  [6658] = {.lex_state = 0},
  [6659] = {.lex_state = 0},
  [6660] = {.lex_state = 0},
  [6661] = {.lex_state = 416},
  [6662] = {.lex_state = 0},
  [6663] = {.lex_state = 0},
  [6664] = {.lex_state = 132},
  [6665] = {.lex_state = 0},
  [6666] = {.lex_state = 105},
  [6667] = {.lex_state = 0},
  [6668] = {.lex_state = 416},
  [6669] = {.lex_state = 58},
  [6670] = {.lex_state = 0},
  [6671] = {.lex_state = 726},
  [6672] = {.lex_state = 0},
  [6673] = {.lex_state = 58},
  [6674] = {.lex_state = 0},
  [6675] = {.lex_state = 132},
  [6676] = {.lex_state = 132},
  [6677] = {.lex_state = 0},
  [6678] = {.lex_state = 416},
  [6679] = {.lex_state = 132},
  [6680] = {.lex_state = 0},
  [6681] = {.lex_state = 0},
  [6682] = {.lex_state = 0},
  [6683] = {.lex_state = 0},
  [6684] = {.lex_state = 0},
  [6685] = {.lex_state = 132},
  [6686] = {.lex_state = 416},
  [6687] = {.lex_state = 0},
  [6688] = {.lex_state = 0},
  [6689] = {.lex_state = 105},
  [6690] = {.lex_state = 416},
  [6691] = {.lex_state = 0},
  [6692] = {.lex_state = 0},
  [6693] = {.lex_state = 0},
  [6694] = {.lex_state = 0},
  [6695] = {.lex_state = 58},
  [6696] = {.lex_state = 0},
  [6697] = {.lex_state = 0},
  [6698] = {.lex_state = 0},
  [6699] = {.lex_state = 0},
  [6700] = {.lex_state = 416},
  [6701] = {.lex_state = 0},
  [6702] = {.lex_state = 416},
  [6703] = {.lex_state = 0},
  [6704] = {.lex_state = 105},
  [6705] = {.lex_state = 0},
  [6706] = {.lex_state = 0},
  [6707] = {.lex_state = 0},
  [6708] = {.lex_state = 0},
  [6709] = {.lex_state = 416},
  [6710] = {.lex_state = 132},
  [6711] = {.lex_state = 0},
  [6712] = {.lex_state = 132},
  [6713] = {.lex_state = 0},
  [6714] = {.lex_state = 416},
  [6715] = {.lex_state = 416},
  [6716] = {.lex_state = 0},
  [6717] = {.lex_state = 0},
  [6718] = {.lex_state = 0},
  [6719] = {.lex_state = 132},
  [6720] = {.lex_state = 132},
  [6721] = {.lex_state = 132},
  [6722] = {.lex_state = 0},
  [6723] = {.lex_state = 0},
  [6724] = {.lex_state = 132},
  [6725] = {.lex_state = 0},
  [6726] = {.lex_state = 0},
  [6727] = {.lex_state = 0},
  [6728] = {.lex_state = 416},
  [6729] = {.lex_state = 58},
  [6730] = {.lex_state = 0},
  [6731] = {.lex_state = 0},
  [6732] = {.lex_state = 0},
  [6733] = {.lex_state = 416},
  [6734] = {.lex_state = 0},
  [6735] = {.lex_state = 132},
  [6736] = {.lex_state = 416},
  [6737] = {.lex_state = 0},
  [6738] = {.lex_state = 132},
  [6739] = {.lex_state = 0},
  [6740] = {.lex_state = 105},
  [6741] = {.lex_state = 0},
  [6742] = {.lex_state = 105},
  [6743] = {.lex_state = 416},
  [6744] = {.lex_state = 0},
  [6745] = {.lex_state = 132},
  [6746] = {.lex_state = 416},
  [6747] = {.lex_state = 0},
  [6748] = {.lex_state = 0},
  [6749] = {.lex_state = 416},
  [6750] = {.lex_state = 132},
  [6751] = {.lex_state = 416},
  [6752] = {.lex_state = 0},
  [6753] = {.lex_state = 132},
  [6754] = {.lex_state = 416},
  [6755] = {.lex_state = 0},
  [6756] = {.lex_state = 0},
  [6757] = {.lex_state = 0},
  [6758] = {.lex_state = 58},
  [6759] = {.lex_state = 0},
  [6760] = {.lex_state = 132},
  [6761] = {.lex_state = 416},
  [6762] = {.lex_state = 0},
  [6763] = {.lex_state = 132},
  [6764] = {.lex_state = 416},
  [6765] = {.lex_state = 132},
  [6766] = {.lex_state = 132},
  [6767] = {.lex_state = 0},
  [6768] = {.lex_state = 132},
  [6769] = {.lex_state = 0},
  [6770] = {.lex_state = 416},
  [6771] = {.lex_state = 416},
  [6772] = {.lex_state = 132},
  [6773] = {.lex_state = 132},
  [6774] = {.lex_state = 58},
  [6775] = {.lex_state = 0},
  [6776] = {.lex_state = 0},
  [6777] = {.lex_state = 105},
  [6778] = {.lex_state = 0},
  [6779] = {.lex_state = 58},
  [6780] = {.lex_state = 0},
  [6781] = {.lex_state = 0},
  [6782] = {.lex_state = 105},
  [6783] = {.lex_state = 0},
  [6784] = {.lex_state = 132},
  [6785] = {.lex_state = 0},
  [6786] = {.lex_state = 0},
  [6787] = {.lex_state = 416},
  [6788] = {.lex_state = 58},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [aux_sym_preproc_include_token1] = ACTIONS(1),
    [aux_sym_preproc_include_token2] = ACTIONS(1),
    [aux_sym_preproc_def_token1] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [aux_sym_preproc_if_token1] = ACTIONS(1),
    [aux_sym_preproc_if_token2] = ACTIONS(1),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1),
    [aux_sym_preproc_else_token1] = ACTIONS(1),
    [aux_sym_preproc_elif_token1] = ACTIONS(1),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1),
    [sym_preproc_directive] = ACTIONS(1),
    [anon_sym_LPAREN2] = ACTIONS(1),
    [anon_sym_defined] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym___extension__] = ACTIONS(1),
    [anon_sym_typedef] = ACTIONS(1),
    [anon_sym_extern] = ACTIONS(1),
    [anon_sym___attribute__] = ACTIONS(1),
    [anon_sym___attribute] = ACTIONS(1),
    [anon_sym_noreturn] = ACTIONS(1),
    [anon_sym_nothrow] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym___declspec] = ACTIONS(1),
    [anon_sym___based] = ACTIONS(1),
    [anon_sym___cdecl] = ACTIONS(1),
    [anon_sym___clrcall] = ACTIONS(1),
    [anon_sym___stdcall] = ACTIONS(1),
    [anon_sym___fastcall] = ACTIONS(1),
    [anon_sym___thiscall] = ACTIONS(1),
    [anon_sym___vectorcall] = ACTIONS(1),
    [sym_ms_restrict_modifier] = ACTIONS(1),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(1),
    [sym_ms_signed_ptr_modifier] = ACTIONS(1),
    [anon_sym__unaligned] = ACTIONS(1),
    [anon_sym___unaligned] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_signed] = ACTIONS(1),
    [anon_sym_unsigned] = ACTIONS(1),
    [anon_sym_long] = ACTIONS(1),
    [anon_sym_short] = ACTIONS(1),
    [anon_sym_static] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_ATautoreleasepool] = ACTIONS(1),
    [anon_sym_auto] = ACTIONS(1),
    [anon_sym_register] = ACTIONS(1),
    [anon_sym_inline] = ACTIONS(1),
    [anon_sym___inline] = ACTIONS(1),
    [anon_sym___inline__] = ACTIONS(1),
    [anon_sym___forceinline] = ACTIONS(1),
    [anon_sym_thread_local] = ACTIONS(1),
    [anon_sym___thread] = ACTIONS(1),
    [anon_sym_CG_EXTERN] = ACTIONS(1),
    [anon_sym_CG_INLINE] = ACTIONS(1),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(1),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(1),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(1),
    [anon_sym_IBOutlet] = ACTIONS(1),
    [anon_sym_IBInspectable] = ACTIONS(1),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(1),
    [anon_sym_NS_INLINE] = ACTIONS(1),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(1),
    [anon_sym_OBJC_EXPORT] = ACTIONS(1),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(1),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(1),
    [anon_sym_const] = ACTIONS(1),
    [anon_sym_constexpr] = ACTIONS(1),
    [anon_sym_volatile] = ACTIONS(1),
    [anon_sym_restrict] = ACTIONS(1),
    [anon_sym___restrict__] = ACTIONS(1),
    [anon_sym__Atomic] = ACTIONS(1),
    [anon_sym__Noreturn] = ACTIONS(1),
    [anon_sym__Nonnull] = ACTIONS(1),
    [anon_sym_nullable] = ACTIONS(1),
    [anon_sym__Complex] = ACTIONS(1),
    [anon_sym__Nullable] = ACTIONS(1),
    [anon_sym__Nullable_result] = ACTIONS(1),
    [anon_sym__Null_unspecified] = ACTIONS(1),
    [anon_sym___autoreleasing] = ACTIONS(1),
    [anon_sym___block] = ACTIONS(1),
    [anon_sym___bridge] = ACTIONS(1),
    [anon_sym___bridge_retained] = ACTIONS(1),
    [anon_sym___bridge_transfer] = ACTIONS(1),
    [anon_sym___complex] = ACTIONS(1),
    [anon_sym___const] = ACTIONS(1),
    [anon_sym___imag] = ACTIONS(1),
    [anon_sym___kindof] = ACTIONS(1),
    [anon_sym___nonnull] = ACTIONS(1),
    [anon_sym___nullable] = ACTIONS(1),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(1),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(1),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(1),
    [anon_sym___real] = ACTIONS(1),
    [anon_sym___strong] = ACTIONS(1),
    [anon_sym___unsafe_unretained] = ACTIONS(1),
    [anon_sym___unused] = ACTIONS(1),
    [anon_sym___weak] = ACTIONS(1),
    [anon_sym_alignas] = ACTIONS(1),
    [anon_sym__Alignas] = ACTIONS(1),
    [sym_primitive_type] = ACTIONS(1),
    [anon_sym_enum] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_struct] = ACTIONS(1),
    [anon_sym_union] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_switch] = ACTIONS(1),
    [anon_sym_case] = ACTIONS(1),
    [anon_sym_default] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_return] = ACTIONS(1),
    [anon_sym_break] = ACTIONS(1),
    [anon_sym_continue] = ACTIONS(1),
    [anon_sym_goto] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_LT_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_sizeof] = ACTIONS(1),
    [anon_sym___alignof__] = ACTIONS(1),
    [anon_sym___alignof] = ACTIONS(1),
    [anon_sym__alignof] = ACTIONS(1),
    [anon_sym_alignof] = ACTIONS(1),
    [anon_sym__Alignof] = ACTIONS(1),
    [anon_sym_offsetof] = ACTIONS(1),
    [anon_sym__Generic] = ACTIONS(1),
    [anon_sym_asm] = ACTIONS(1),
    [anon_sym___asm__] = ACTIONS(1),
    [anon_sym___asm] = ACTIONS(1),
    [anon_sym___volatile__] = ACTIONS(1),
    [anon_sym_LT2] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [sym_number_literal] = ACTIONS(1),
    [anon_sym_L_SQUOTE] = ACTIONS(1),
    [anon_sym_u_SQUOTE] = ACTIONS(1),
    [anon_sym_U_SQUOTE] = ACTIONS(1),
    [anon_sym_u8_SQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_L_DQUOTE] = ACTIONS(1),
    [anon_sym_u_DQUOTE] = ACTIONS(1),
    [anon_sym_U_DQUOTE] = ACTIONS(1),
    [anon_sym_u8_DQUOTE] = ACTIONS(1),
    [sym_true] = ACTIONS(1),
    [sym_false] = ACTIONS(1),
    [anon_sym_NULL] = ACTIONS(1),
    [anon_sym_nullptr] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
    [anon_sym_objc_bridge_related] = ACTIONS(1),
    [anon_sym___typeof__] = ACTIONS(1),
    [anon_sym___typeof] = ACTIONS(1),
    [anon_sym_typeof] = ACTIONS(1),
    [anon_sym_availability] = ACTIONS(1),
    [sym_version_number] = ACTIONS(1),
    [anon_sym_ios] = ACTIONS(1),
    [anon_sym_tvos] = ACTIONS(1),
    [anon_sym_macos] = ACTIONS(1),
    [anon_sym_macosx] = ACTIONS(1),
    [anon_sym_watchos] = ACTIONS(1),
    [anon_sym_ATimport] = ACTIONS(1),
    [aux_sym_preproc_undef_token1] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(1),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(1),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(1),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(1),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(1),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(1),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(1),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(1),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(1),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(1),
    [anon_sym_NS_AVAILABLE] = ACTIONS(1),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(1),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(1),
    [anon_sym_API_AVAILABLE] = ACTIONS(1),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(1),
    [anon_sym_API_DEPRECATED] = ACTIONS(1),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(1),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(1),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(1),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(1),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(1),
    [anon_sym___deprecated_msg] = ACTIONS(1),
    [anon_sym___deprecated_enum_msg] = ACTIONS(1),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(1),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(1),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(1),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(1),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(1),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(1),
    [anon_sym_ATprotocol] = ACTIONS(1),
    [anon_sym_class] = ACTIONS(1),
    [anon_sym_ATinterface] = ACTIONS(1),
    [anon_sym_ATimplementation] = ACTIONS(1),
    [anon_sym___covariant] = ACTIONS(1),
    [anon_sym___contravariant] = ACTIONS(1),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1),
    [anon_sym_ATsynthesize] = ACTIONS(1),
    [anon_sym_ATdynamic] = ACTIONS(1),
    [anon_sym_ATtry] = ACTIONS(1),
    [anon_sym___try] = ACTIONS(1),
    [anon_sym_ATcatch] = ACTIONS(1),
    [anon_sym___catch] = ACTIONS(1),
    [anon_sym_ATfinally] = ACTIONS(1),
    [anon_sym___finally] = ACTIONS(1),
    [anon_sym_ATthrow] = ACTIONS(1),
    [anon_sym_ATselector] = ACTIONS(1),
    [anon_sym_ATavailable] = ACTIONS(1),
    [anon_sym___builtin_available] = ACTIONS(1),
    [anon_sym_va_arg] = ACTIONS(1),
    [anon_sym_ATencode] = ACTIONS(1),
    [anon_sym_ATsynchronized] = ACTIONS(1),
    [anon_sym_BOOL] = ACTIONS(1),
    [anon_sym_IMP] = ACTIONS(1),
    [anon_sym_SEL] = ACTIONS(1),
    [anon_sym_Class] = ACTIONS(1),
    [anon_sym_id] = ACTIONS(1),
    [anon_sym_out] = ACTIONS(1),
    [anon_sym_inout] = ACTIONS(1),
    [anon_sym_bycopy] = ACTIONS(1),
    [anon_sym_byref] = ACTIONS(1),
    [anon_sym_oneway] = ACTIONS(1),
    [anon_sym_COLON2] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_translation_unit] = STATE(6534),
    [sym__top_level_item] = STATE(129),
    [sym_preproc_include] = STATE(129),
    [sym_preproc_def] = STATE(129),
    [sym_preproc_function_def] = STATE(129),
    [sym_preproc_call] = STATE(129),
    [sym_preproc_if] = STATE(129),
    [sym_preproc_ifdef] = STATE(129),
    [sym_function_definition] = STATE(129),
    [sym__old_style_function_definition] = STATE(983),
    [sym_declaration] = STATE(129),
    [sym_type_definition] = STATE(129),
    [sym__declaration_modifiers] = STATE(2050),
    [sym__declaration_specifiers] = STATE(4180),
    [sym_linkage_specification] = STATE(129),
    [sym_attribute_specifier] = STATE(2050),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2836),
    [sym_ms_call_modifier] = STATE(2120),
    [sym_compound_statement] = STATE(129),
    [sym_storage_class_specifier] = STATE(2050),
    [sym_type_qualifier] = STATE(2050),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2998),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(978),
    [sym__top_level_statement] = STATE(129),
    [sym_labeled_statement] = STATE(129),
    [sym__top_level_expression_statement] = STATE(129),
    [sym_if_statement] = STATE(129),
    [sym_switch_statement] = STATE(129),
    [sym_case_statement] = STATE(129),
    [sym_while_statement] = STATE(129),
    [sym_do_statement] = STATE(129),
    [sym_for_statement] = STATE(129),
    [sym_return_statement] = STATE(129),
    [sym_break_statement] = STATE(129),
    [sym_continue_statement] = STATE(129),
    [sym_goto_statement] = STATE(129),
    [sym_expression] = STATE(4565),
    [sym__string] = STATE(4564),
    [sym_conditional_expression] = STATE(4564),
    [sym_assignment_expression] = STATE(4564),
    [sym_pointer_expression] = STATE(4360),
    [sym_unary_expression] = STATE(4564),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(4564),
    [sym_cast_expression] = STATE(4564),
    [sym_sizeof_expression] = STATE(4564),
    [sym_alignof_expression] = STATE(4564),
    [sym_offsetof_expression] = STATE(4564),
    [sym_generic_expression] = STATE(4564),
    [sym_subscript_expression] = STATE(4360),
    [sym_call_expression] = STATE(4360),
    [sym_gnu_asm_expression] = STATE(4564),
    [sym_extension_expression] = STATE(4564),
    [sym_field_expression] = STATE(4360),
    [sym_compound_literal_expression] = STATE(4564),
    [sym_parenthesized_expression] = STATE(4360),
    [sym_char_literal] = STATE(4564),
    [sym_concatenated_string] = STATE(4564),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(4564),
    [sym__empty_declaration] = STATE(129),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(129),
    [sym_preproc_undef] = STATE(129),
    [sym_preproc_linemarker] = STATE(129),
    [sym_availability_attribute_specifier] = STATE(2050),
    [sym_protocol_forward_declaration] = STATE(129),
    [sym_class_declaration] = STATE(129),
    [sym_class_interface] = STATE(129),
    [sym__class_interface_header] = STATE(1165),
    [sym_class_implementation] = STATE(129),
    [sym__class_implementation_header] = STATE(1088),
    [sym_protocol_declaration] = STATE(129),
    [sym_compatibility_alias_declaration] = STATE(129),
    [sym_selector_expression] = STATE(4564),
    [sym_available_expression] = STATE(4564),
    [sym_range_expression] = STATE(4564),
    [sym_block_literal] = STATE(4564),
    [sym_message_expression] = STATE(4564),
    [sym_va_arg_expression] = STATE(4564),
    [sym_encode_expression] = STATE(4564),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(4564),
    [sym_dictionary_literal] = STATE(4564),
    [sym_array_literal] = STATE(4564),
    [aux_sym_translation_unit_repeat1] = STATE(129),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2050),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2451),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [ts_builtin_sym_end] = ACTIONS(5),
    [sym_identifier] = ACTIONS(7),
    [aux_sym_preproc_include_token1] = ACTIONS(9),
    [aux_sym_preproc_include_token2] = ACTIONS(9),
    [aux_sym_preproc_def_token1] = ACTIONS(11),
    [aux_sym_preproc_if_token1] = ACTIONS(13),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(15),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(15),
    [sym_preproc_directive] = ACTIONS(17),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(29),
    [anon_sym___extension__] = ACTIONS(31),
    [anon_sym_typedef] = ACTIONS(33),
    [anon_sym_extern] = ACTIONS(35),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(55),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(65),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(69),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(77),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(83),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(105),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(109),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(113),
    [sym_false] = ACTIONS(113),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(119),
    [aux_sym_preproc_undef_token1] = ACTIONS(121),
    [anon_sym_POUND] = ACTIONS(123),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(129),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(135),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(149),
  },
  [STATE(2)] = {
    [sym__block_item] = STATE(33),
    [sym_preproc_include] = STATE(33),
    [sym_preproc_def] = STATE(33),
    [sym_preproc_function_def] = STATE(33),
    [sym_preproc_call] = STATE(33),
    [sym_preproc_if] = STATE(33),
    [sym_preproc_ifdef] = STATE(33),
    [sym_preproc_else] = STATE(6394),
    [sym_preproc_elif] = STATE(6394),
    [sym_preproc_elifdef] = STATE(6394),
    [sym_function_definition] = STATE(33),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(33),
    [sym_type_definition] = STATE(33),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(33),
    [sym_attribute_specifier] = STATE(33),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(33),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(33),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(33),
    [sym_preproc_undef] = STATE(33),
    [sym_preproc_linemarker] = STATE(33),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(33),
    [sym_class_declaration] = STATE(33),
    [sym_class_interface] = STATE(33),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(33),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(33),
    [sym_compatibility_alias_declaration] = STATE(33),
    [sym_property_implementation] = STATE(33),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(33),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(159),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(3)] = {
    [sym__block_item] = STATE(33),
    [sym_preproc_include] = STATE(33),
    [sym_preproc_def] = STATE(33),
    [sym_preproc_function_def] = STATE(33),
    [sym_preproc_call] = STATE(33),
    [sym_preproc_if] = STATE(33),
    [sym_preproc_ifdef] = STATE(33),
    [sym_preproc_else] = STATE(6454),
    [sym_preproc_elif] = STATE(6454),
    [sym_preproc_elifdef] = STATE(6454),
    [sym_function_definition] = STATE(33),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(33),
    [sym_type_definition] = STATE(33),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(33),
    [sym_attribute_specifier] = STATE(33),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(33),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(33),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(33),
    [sym_preproc_undef] = STATE(33),
    [sym_preproc_linemarker] = STATE(33),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(33),
    [sym_class_declaration] = STATE(33),
    [sym_class_interface] = STATE(33),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(33),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(33),
    [sym_compatibility_alias_declaration] = STATE(33),
    [sym_property_implementation] = STATE(33),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(33),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(245),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(4)] = {
    [sym__block_item] = STATE(5),
    [sym_preproc_include] = STATE(5),
    [sym_preproc_def] = STATE(5),
    [sym_preproc_function_def] = STATE(5),
    [sym_preproc_call] = STATE(5),
    [sym_preproc_if] = STATE(5),
    [sym_preproc_ifdef] = STATE(5),
    [sym_preproc_else] = STATE(6307),
    [sym_preproc_elif] = STATE(6307),
    [sym_preproc_elifdef] = STATE(6307),
    [sym_function_definition] = STATE(5),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(5),
    [sym_type_definition] = STATE(5),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(5),
    [sym_attribute_specifier] = STATE(5),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(5),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(5),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(5),
    [sym_preproc_undef] = STATE(5),
    [sym_preproc_linemarker] = STATE(5),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(5),
    [sym_class_declaration] = STATE(5),
    [sym_class_interface] = STATE(5),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(5),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(5),
    [sym_compatibility_alias_declaration] = STATE(5),
    [sym_property_implementation] = STATE(5),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(5),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(247),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(5)] = {
    [sym__block_item] = STATE(33),
    [sym_preproc_include] = STATE(33),
    [sym_preproc_def] = STATE(33),
    [sym_preproc_function_def] = STATE(33),
    [sym_preproc_call] = STATE(33),
    [sym_preproc_if] = STATE(33),
    [sym_preproc_ifdef] = STATE(33),
    [sym_preproc_else] = STATE(6467),
    [sym_preproc_elif] = STATE(6467),
    [sym_preproc_elifdef] = STATE(6467),
    [sym_function_definition] = STATE(33),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(33),
    [sym_type_definition] = STATE(33),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(33),
    [sym_attribute_specifier] = STATE(33),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(33),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(33),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(33),
    [sym_preproc_undef] = STATE(33),
    [sym_preproc_linemarker] = STATE(33),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(33),
    [sym_class_declaration] = STATE(33),
    [sym_class_interface] = STATE(33),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(33),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(33),
    [sym_compatibility_alias_declaration] = STATE(33),
    [sym_property_implementation] = STATE(33),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(33),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(249),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(6)] = {
    [sym__block_item] = STATE(2),
    [sym_preproc_include] = STATE(2),
    [sym_preproc_def] = STATE(2),
    [sym_preproc_function_def] = STATE(2),
    [sym_preproc_call] = STATE(2),
    [sym_preproc_if] = STATE(2),
    [sym_preproc_ifdef] = STATE(2),
    [sym_preproc_else] = STATE(6666),
    [sym_preproc_elif] = STATE(6666),
    [sym_preproc_elifdef] = STATE(6666),
    [sym_function_definition] = STATE(2),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(2),
    [sym_type_definition] = STATE(2),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(2),
    [sym_attribute_specifier] = STATE(2),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(2),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(2),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(2),
    [sym_preproc_undef] = STATE(2),
    [sym_preproc_linemarker] = STATE(2),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(2),
    [sym_class_declaration] = STATE(2),
    [sym_class_interface] = STATE(2),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(2),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(2),
    [sym_compatibility_alias_declaration] = STATE(2),
    [sym_property_implementation] = STATE(2),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(2),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(251),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(7)] = {
    [sym__block_item] = STATE(3),
    [sym_preproc_include] = STATE(3),
    [sym_preproc_def] = STATE(3),
    [sym_preproc_function_def] = STATE(3),
    [sym_preproc_call] = STATE(3),
    [sym_preproc_if] = STATE(3),
    [sym_preproc_ifdef] = STATE(3),
    [sym_preproc_else] = STATE(6113),
    [sym_preproc_elif] = STATE(6113),
    [sym_preproc_elifdef] = STATE(6113),
    [sym_function_definition] = STATE(3),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(3),
    [sym_type_definition] = STATE(3),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(3),
    [sym_attribute_specifier] = STATE(3),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(3),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(3),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(3),
    [sym_preproc_undef] = STATE(3),
    [sym_preproc_linemarker] = STATE(3),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(3),
    [sym_class_declaration] = STATE(3),
    [sym_class_interface] = STATE(3),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(3),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(3),
    [sym_compatibility_alias_declaration] = STATE(3),
    [sym_property_implementation] = STATE(3),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(3),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(253),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(8)] = {
    [sym__block_item] = STATE(9),
    [sym_preproc_include] = STATE(9),
    [sym_preproc_def] = STATE(9),
    [sym_preproc_function_def] = STATE(9),
    [sym_preproc_call] = STATE(9),
    [sym_preproc_if] = STATE(9),
    [sym_preproc_ifdef] = STATE(9),
    [sym_preproc_else] = STATE(6033),
    [sym_preproc_elif] = STATE(6033),
    [sym_preproc_elifdef] = STATE(6033),
    [sym_function_definition] = STATE(9),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(9),
    [sym_type_definition] = STATE(9),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(9),
    [sym_attribute_specifier] = STATE(9),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(9),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(9),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(9),
    [sym_preproc_undef] = STATE(9),
    [sym_preproc_linemarker] = STATE(9),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(9),
    [sym_class_declaration] = STATE(9),
    [sym_class_interface] = STATE(9),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(9),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(9),
    [sym_compatibility_alias_declaration] = STATE(9),
    [sym_property_implementation] = STATE(9),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(9),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(255),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(9)] = {
    [sym__block_item] = STATE(33),
    [sym_preproc_include] = STATE(33),
    [sym_preproc_def] = STATE(33),
    [sym_preproc_function_def] = STATE(33),
    [sym_preproc_call] = STATE(33),
    [sym_preproc_if] = STATE(33),
    [sym_preproc_ifdef] = STATE(33),
    [sym_preproc_else] = STATE(6150),
    [sym_preproc_elif] = STATE(6150),
    [sym_preproc_elifdef] = STATE(6150),
    [sym_function_definition] = STATE(33),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(33),
    [sym_type_definition] = STATE(33),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(33),
    [sym_attribute_specifier] = STATE(33),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(33),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(33),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(33),
    [sym_preproc_undef] = STATE(33),
    [sym_preproc_linemarker] = STATE(33),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(33),
    [sym_class_declaration] = STATE(33),
    [sym_class_interface] = STATE(33),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(33),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(33),
    [sym_compatibility_alias_declaration] = STATE(33),
    [sym_property_implementation] = STATE(33),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(33),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(257),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(10)] = {
    [sym__block_item] = STATE(11),
    [sym_preproc_include] = STATE(11),
    [sym_preproc_def] = STATE(11),
    [sym_preproc_function_def] = STATE(11),
    [sym_preproc_call] = STATE(11),
    [sym_preproc_if] = STATE(11),
    [sym_preproc_ifdef] = STATE(11),
    [sym_preproc_else] = STATE(6469),
    [sym_preproc_elif] = STATE(6469),
    [sym_preproc_elifdef] = STATE(6469),
    [sym_function_definition] = STATE(11),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(11),
    [sym_type_definition] = STATE(11),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(11),
    [sym_attribute_specifier] = STATE(11),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(11),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(11),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(11),
    [sym_preproc_undef] = STATE(11),
    [sym_preproc_linemarker] = STATE(11),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(11),
    [sym_class_declaration] = STATE(11),
    [sym_class_interface] = STATE(11),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(11),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(11),
    [sym_compatibility_alias_declaration] = STATE(11),
    [sym_property_implementation] = STATE(11),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(11),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(259),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(11)] = {
    [sym__block_item] = STATE(33),
    [sym_preproc_include] = STATE(33),
    [sym_preproc_def] = STATE(33),
    [sym_preproc_function_def] = STATE(33),
    [sym_preproc_call] = STATE(33),
    [sym_preproc_if] = STATE(33),
    [sym_preproc_ifdef] = STATE(33),
    [sym_preproc_else] = STATE(6571),
    [sym_preproc_elif] = STATE(6571),
    [sym_preproc_elifdef] = STATE(6571),
    [sym_function_definition] = STATE(33),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(33),
    [sym_type_definition] = STATE(33),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(33),
    [sym_attribute_specifier] = STATE(33),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(33),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(33),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(33),
    [sym_preproc_undef] = STATE(33),
    [sym_preproc_linemarker] = STATE(33),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(33),
    [sym_class_declaration] = STATE(33),
    [sym_class_interface] = STATE(33),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(33),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(33),
    [sym_compatibility_alias_declaration] = STATE(33),
    [sym_property_implementation] = STATE(33),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(33),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(261),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(12)] = {
    [sym__block_item] = STATE(13),
    [sym_preproc_include] = STATE(13),
    [sym_preproc_def] = STATE(13),
    [sym_preproc_function_def] = STATE(13),
    [sym_preproc_call] = STATE(13),
    [sym_preproc_if] = STATE(13),
    [sym_preproc_ifdef] = STATE(13),
    [sym_preproc_else] = STATE(6022),
    [sym_preproc_elif] = STATE(6022),
    [sym_preproc_elifdef] = STATE(6022),
    [sym_function_definition] = STATE(13),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(13),
    [sym_type_definition] = STATE(13),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(13),
    [sym_attribute_specifier] = STATE(13),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(13),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(13),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(13),
    [sym_preproc_undef] = STATE(13),
    [sym_preproc_linemarker] = STATE(13),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(13),
    [sym_class_declaration] = STATE(13),
    [sym_class_interface] = STATE(13),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(13),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(13),
    [sym_compatibility_alias_declaration] = STATE(13),
    [sym_property_implementation] = STATE(13),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(13),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(263),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(13)] = {
    [sym__block_item] = STATE(33),
    [sym_preproc_include] = STATE(33),
    [sym_preproc_def] = STATE(33),
    [sym_preproc_function_def] = STATE(33),
    [sym_preproc_call] = STATE(33),
    [sym_preproc_if] = STATE(33),
    [sym_preproc_ifdef] = STATE(33),
    [sym_preproc_else] = STATE(6529),
    [sym_preproc_elif] = STATE(6529),
    [sym_preproc_elifdef] = STATE(6529),
    [sym_function_definition] = STATE(33),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(33),
    [sym_type_definition] = STATE(33),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(33),
    [sym_attribute_specifier] = STATE(33),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(33),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(33),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(33),
    [sym_preproc_undef] = STATE(33),
    [sym_preproc_linemarker] = STATE(33),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(33),
    [sym_class_declaration] = STATE(33),
    [sym_class_interface] = STATE(33),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(33),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(33),
    [sym_compatibility_alias_declaration] = STATE(33),
    [sym_property_implementation] = STATE(33),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(33),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(265),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(14)] = {
    [sym__block_item] = STATE(15),
    [sym_preproc_include] = STATE(15),
    [sym_preproc_def] = STATE(15),
    [sym_preproc_function_def] = STATE(15),
    [sym_preproc_call] = STATE(15),
    [sym_preproc_if] = STATE(15),
    [sym_preproc_ifdef] = STATE(15),
    [sym_preproc_else] = STATE(6288),
    [sym_preproc_elif] = STATE(6288),
    [sym_preproc_elifdef] = STATE(6288),
    [sym_function_definition] = STATE(15),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(15),
    [sym_type_definition] = STATE(15),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(15),
    [sym_attribute_specifier] = STATE(15),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(15),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(15),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(15),
    [sym_preproc_undef] = STATE(15),
    [sym_preproc_linemarker] = STATE(15),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(15),
    [sym_class_declaration] = STATE(15),
    [sym_class_interface] = STATE(15),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(15),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(15),
    [sym_compatibility_alias_declaration] = STATE(15),
    [sym_property_implementation] = STATE(15),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(15),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(267),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(15)] = {
    [sym__block_item] = STATE(33),
    [sym_preproc_include] = STATE(33),
    [sym_preproc_def] = STATE(33),
    [sym_preproc_function_def] = STATE(33),
    [sym_preproc_call] = STATE(33),
    [sym_preproc_if] = STATE(33),
    [sym_preproc_ifdef] = STATE(33),
    [sym_preproc_else] = STATE(6409),
    [sym_preproc_elif] = STATE(6409),
    [sym_preproc_elifdef] = STATE(6409),
    [sym_function_definition] = STATE(33),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(33),
    [sym_type_definition] = STATE(33),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(33),
    [sym_attribute_specifier] = STATE(33),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(33),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(33),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(33),
    [sym_preproc_undef] = STATE(33),
    [sym_preproc_linemarker] = STATE(33),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(33),
    [sym_class_declaration] = STATE(33),
    [sym_class_interface] = STATE(33),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(33),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(33),
    [sym_compatibility_alias_declaration] = STATE(33),
    [sym_property_implementation] = STATE(33),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(33),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(269),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(16)] = {
    [sym__block_item] = STATE(17),
    [sym_preproc_include] = STATE(17),
    [sym_preproc_def] = STATE(17),
    [sym_preproc_function_def] = STATE(17),
    [sym_preproc_call] = STATE(17),
    [sym_preproc_if] = STATE(17),
    [sym_preproc_ifdef] = STATE(17),
    [sym_preproc_else] = STATE(6579),
    [sym_preproc_elif] = STATE(6579),
    [sym_preproc_elifdef] = STATE(6579),
    [sym_function_definition] = STATE(17),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(17),
    [sym_type_definition] = STATE(17),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(17),
    [sym_attribute_specifier] = STATE(17),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(17),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(17),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(17),
    [sym_preproc_undef] = STATE(17),
    [sym_preproc_linemarker] = STATE(17),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(17),
    [sym_class_declaration] = STATE(17),
    [sym_class_interface] = STATE(17),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(17),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(17),
    [sym_compatibility_alias_declaration] = STATE(17),
    [sym_property_implementation] = STATE(17),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(17),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(271),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(17)] = {
    [sym__block_item] = STATE(33),
    [sym_preproc_include] = STATE(33),
    [sym_preproc_def] = STATE(33),
    [sym_preproc_function_def] = STATE(33),
    [sym_preproc_call] = STATE(33),
    [sym_preproc_if] = STATE(33),
    [sym_preproc_ifdef] = STATE(33),
    [sym_preproc_else] = STATE(6635),
    [sym_preproc_elif] = STATE(6635),
    [sym_preproc_elifdef] = STATE(6635),
    [sym_function_definition] = STATE(33),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(33),
    [sym_type_definition] = STATE(33),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(33),
    [sym_attribute_specifier] = STATE(33),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(33),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(33),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(33),
    [sym_preproc_undef] = STATE(33),
    [sym_preproc_linemarker] = STATE(33),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(33),
    [sym_class_declaration] = STATE(33),
    [sym_class_interface] = STATE(33),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(33),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(33),
    [sym_compatibility_alias_declaration] = STATE(33),
    [sym_property_implementation] = STATE(33),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(33),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(273),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(18)] = {
    [sym__block_item] = STATE(19),
    [sym_preproc_include] = STATE(19),
    [sym_preproc_def] = STATE(19),
    [sym_preproc_function_def] = STATE(19),
    [sym_preproc_call] = STATE(19),
    [sym_preproc_if] = STATE(19),
    [sym_preproc_ifdef] = STATE(19),
    [sym_preproc_else] = STATE(5926),
    [sym_preproc_elif] = STATE(5926),
    [sym_preproc_elifdef] = STATE(5926),
    [sym_function_definition] = STATE(19),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(19),
    [sym_type_definition] = STATE(19),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(19),
    [sym_attribute_specifier] = STATE(19),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(19),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(19),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(19),
    [sym_preproc_undef] = STATE(19),
    [sym_preproc_linemarker] = STATE(19),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(19),
    [sym_class_declaration] = STATE(19),
    [sym_class_interface] = STATE(19),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(19),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(19),
    [sym_compatibility_alias_declaration] = STATE(19),
    [sym_property_implementation] = STATE(19),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(19),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(275),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(19)] = {
    [sym__block_item] = STATE(33),
    [sym_preproc_include] = STATE(33),
    [sym_preproc_def] = STATE(33),
    [sym_preproc_function_def] = STATE(33),
    [sym_preproc_call] = STATE(33),
    [sym_preproc_if] = STATE(33),
    [sym_preproc_ifdef] = STATE(33),
    [sym_preproc_else] = STATE(5955),
    [sym_preproc_elif] = STATE(5955),
    [sym_preproc_elifdef] = STATE(5955),
    [sym_function_definition] = STATE(33),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(33),
    [sym_type_definition] = STATE(33),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(33),
    [sym_attribute_specifier] = STATE(33),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(33),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(33),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(33),
    [sym_preproc_undef] = STATE(33),
    [sym_preproc_linemarker] = STATE(33),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(33),
    [sym_class_declaration] = STATE(33),
    [sym_class_interface] = STATE(33),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(33),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(33),
    [sym_compatibility_alias_declaration] = STATE(33),
    [sym_property_implementation] = STATE(33),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(33),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(151),
    [aux_sym_preproc_include_token1] = ACTIONS(153),
    [aux_sym_preproc_include_token2] = ACTIONS(153),
    [aux_sym_preproc_def_token1] = ACTIONS(155),
    [aux_sym_preproc_if_token1] = ACTIONS(157),
    [aux_sym_preproc_if_token2] = ACTIONS(277),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(161),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(161),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(167),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(167),
    [sym_preproc_directive] = ACTIONS(169),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(173),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(177),
    [anon_sym___attribute__] = ACTIONS(179),
    [anon_sym___attribute] = ACTIONS(179),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(221),
    [aux_sym_preproc_undef_token1] = ACTIONS(223),
    [anon_sym_POUND] = ACTIONS(225),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(227),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(229),
    [anon_sym_ATsynthesize] = ACTIONS(231),
    [anon_sym_ATdynamic] = ACTIONS(233),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(20)] = {
    [sym__block_item] = STATE(21),
    [sym_preproc_include] = STATE(21),
    [sym_preproc_def] = STATE(21),
    [sym_preproc_function_def] = STATE(21),
    [sym_preproc_call] = STATE(21),
    [sym_preproc_if] = STATE(21),
    [sym_preproc_ifdef] = STATE(21),
    [sym_preproc_else] = STATE(6384),
    [sym_preproc_elif] = STATE(6384),
    [sym_function_definition] = STATE(21),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(21),
    [sym_type_definition] = STATE(21),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(21),
    [sym_attribute_specifier] = STATE(21),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(21),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(21),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(21),
    [sym_preproc_undef] = STATE(21),
    [sym_preproc_linemarker] = STATE(21),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(21),
    [sym_class_declaration] = STATE(21),
    [sym_class_interface] = STATE(21),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(21),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(21),
    [sym_compatibility_alias_declaration] = STATE(21),
    [sym_property_implementation] = STATE(21),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(21),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(287),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(21)] = {
    [sym__block_item] = STATE(37),
    [sym_preproc_include] = STATE(37),
    [sym_preproc_def] = STATE(37),
    [sym_preproc_function_def] = STATE(37),
    [sym_preproc_call] = STATE(37),
    [sym_preproc_if] = STATE(37),
    [sym_preproc_ifdef] = STATE(37),
    [sym_preproc_else] = STATE(6136),
    [sym_preproc_elif] = STATE(6136),
    [sym_function_definition] = STATE(37),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(37),
    [sym_type_definition] = STATE(37),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(37),
    [sym_attribute_specifier] = STATE(37),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(37),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(37),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(37),
    [sym_preproc_undef] = STATE(37),
    [sym_preproc_linemarker] = STATE(37),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(37),
    [sym_class_declaration] = STATE(37),
    [sym_class_interface] = STATE(37),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(37),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(37),
    [sym_compatibility_alias_declaration] = STATE(37),
    [sym_property_implementation] = STATE(37),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(37),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(355),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(22)] = {
    [sym__block_item] = STATE(26),
    [sym_preproc_include] = STATE(26),
    [sym_preproc_def] = STATE(26),
    [sym_preproc_function_def] = STATE(26),
    [sym_preproc_call] = STATE(26),
    [sym_preproc_if] = STATE(26),
    [sym_preproc_ifdef] = STATE(26),
    [sym_preproc_else] = STATE(6226),
    [sym_preproc_elif] = STATE(6226),
    [sym_function_definition] = STATE(26),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(26),
    [sym_type_definition] = STATE(26),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(26),
    [sym_attribute_specifier] = STATE(26),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(26),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(26),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(26),
    [sym_preproc_undef] = STATE(26),
    [sym_preproc_linemarker] = STATE(26),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(26),
    [sym_class_declaration] = STATE(26),
    [sym_class_interface] = STATE(26),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(26),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(26),
    [sym_compatibility_alias_declaration] = STATE(26),
    [sym_property_implementation] = STATE(26),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(26),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(357),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(23)] = {
    [sym__block_item] = STATE(25),
    [sym_preproc_include] = STATE(25),
    [sym_preproc_def] = STATE(25),
    [sym_preproc_function_def] = STATE(25),
    [sym_preproc_call] = STATE(25),
    [sym_preproc_if] = STATE(25),
    [sym_preproc_ifdef] = STATE(25),
    [sym_preproc_else] = STATE(6148),
    [sym_preproc_elif] = STATE(6148),
    [sym_function_definition] = STATE(25),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(25),
    [sym_type_definition] = STATE(25),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(25),
    [sym_attribute_specifier] = STATE(25),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(25),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(25),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(25),
    [sym_preproc_undef] = STATE(25),
    [sym_preproc_linemarker] = STATE(25),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(25),
    [sym_class_declaration] = STATE(25),
    [sym_class_interface] = STATE(25),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(25),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(25),
    [sym_compatibility_alias_declaration] = STATE(25),
    [sym_property_implementation] = STATE(25),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(25),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(359),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(24)] = {
    [sym__block_item] = STATE(37),
    [sym_preproc_include] = STATE(37),
    [sym_preproc_def] = STATE(37),
    [sym_preproc_function_def] = STATE(37),
    [sym_preproc_call] = STATE(37),
    [sym_preproc_if] = STATE(37),
    [sym_preproc_ifdef] = STATE(37),
    [sym_preproc_else] = STATE(6302),
    [sym_preproc_elif] = STATE(6302),
    [sym_function_definition] = STATE(37),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(37),
    [sym_type_definition] = STATE(37),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(37),
    [sym_attribute_specifier] = STATE(37),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(37),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(37),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(37),
    [sym_preproc_undef] = STATE(37),
    [sym_preproc_linemarker] = STATE(37),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(37),
    [sym_class_declaration] = STATE(37),
    [sym_class_interface] = STATE(37),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(37),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(37),
    [sym_compatibility_alias_declaration] = STATE(37),
    [sym_property_implementation] = STATE(37),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(37),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(361),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(25)] = {
    [sym__block_item] = STATE(37),
    [sym_preproc_include] = STATE(37),
    [sym_preproc_def] = STATE(37),
    [sym_preproc_function_def] = STATE(37),
    [sym_preproc_call] = STATE(37),
    [sym_preproc_if] = STATE(37),
    [sym_preproc_ifdef] = STATE(37),
    [sym_preproc_else] = STATE(6248),
    [sym_preproc_elif] = STATE(6248),
    [sym_function_definition] = STATE(37),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(37),
    [sym_type_definition] = STATE(37),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(37),
    [sym_attribute_specifier] = STATE(37),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(37),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(37),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(37),
    [sym_preproc_undef] = STATE(37),
    [sym_preproc_linemarker] = STATE(37),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(37),
    [sym_class_declaration] = STATE(37),
    [sym_class_interface] = STATE(37),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(37),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(37),
    [sym_compatibility_alias_declaration] = STATE(37),
    [sym_property_implementation] = STATE(37),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(37),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(363),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(26)] = {
    [sym__block_item] = STATE(37),
    [sym_preproc_include] = STATE(37),
    [sym_preproc_def] = STATE(37),
    [sym_preproc_function_def] = STATE(37),
    [sym_preproc_call] = STATE(37),
    [sym_preproc_if] = STATE(37),
    [sym_preproc_ifdef] = STATE(37),
    [sym_preproc_else] = STATE(6570),
    [sym_preproc_elif] = STATE(6570),
    [sym_function_definition] = STATE(37),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(37),
    [sym_type_definition] = STATE(37),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(37),
    [sym_attribute_specifier] = STATE(37),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(37),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(37),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(37),
    [sym_preproc_undef] = STATE(37),
    [sym_preproc_linemarker] = STATE(37),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(37),
    [sym_class_declaration] = STATE(37),
    [sym_class_interface] = STATE(37),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(37),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(37),
    [sym_compatibility_alias_declaration] = STATE(37),
    [sym_property_implementation] = STATE(37),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(37),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(365),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(27)] = {
    [sym__block_item] = STATE(28),
    [sym_preproc_include] = STATE(28),
    [sym_preproc_def] = STATE(28),
    [sym_preproc_function_def] = STATE(28),
    [sym_preproc_call] = STATE(28),
    [sym_preproc_if] = STATE(28),
    [sym_preproc_ifdef] = STATE(28),
    [sym_preproc_else] = STATE(6568),
    [sym_preproc_elif] = STATE(6568),
    [sym_function_definition] = STATE(28),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(28),
    [sym_type_definition] = STATE(28),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(28),
    [sym_attribute_specifier] = STATE(28),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(28),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(28),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(28),
    [sym_preproc_undef] = STATE(28),
    [sym_preproc_linemarker] = STATE(28),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(28),
    [sym_class_declaration] = STATE(28),
    [sym_class_interface] = STATE(28),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(28),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(28),
    [sym_compatibility_alias_declaration] = STATE(28),
    [sym_property_implementation] = STATE(28),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(28),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(367),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(28)] = {
    [sym__block_item] = STATE(37),
    [sym_preproc_include] = STATE(37),
    [sym_preproc_def] = STATE(37),
    [sym_preproc_function_def] = STATE(37),
    [sym_preproc_call] = STATE(37),
    [sym_preproc_if] = STATE(37),
    [sym_preproc_ifdef] = STATE(37),
    [sym_preproc_else] = STATE(6649),
    [sym_preproc_elif] = STATE(6649),
    [sym_function_definition] = STATE(37),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(37),
    [sym_type_definition] = STATE(37),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(37),
    [sym_attribute_specifier] = STATE(37),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(37),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(37),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(37),
    [sym_preproc_undef] = STATE(37),
    [sym_preproc_linemarker] = STATE(37),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(37),
    [sym_class_declaration] = STATE(37),
    [sym_class_interface] = STATE(37),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(37),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(37),
    [sym_compatibility_alias_declaration] = STATE(37),
    [sym_property_implementation] = STATE(37),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(37),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(369),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(29)] = {
    [sym__block_item] = STATE(31),
    [sym_preproc_include] = STATE(31),
    [sym_preproc_def] = STATE(31),
    [sym_preproc_function_def] = STATE(31),
    [sym_preproc_call] = STATE(31),
    [sym_preproc_if] = STATE(31),
    [sym_preproc_ifdef] = STATE(31),
    [sym_preproc_else] = STATE(6492),
    [sym_preproc_elif] = STATE(6492),
    [sym_function_definition] = STATE(31),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(31),
    [sym_type_definition] = STATE(31),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(31),
    [sym_attribute_specifier] = STATE(31),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(31),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(31),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(31),
    [sym_preproc_undef] = STATE(31),
    [sym_preproc_linemarker] = STATE(31),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(31),
    [sym_class_declaration] = STATE(31),
    [sym_class_interface] = STATE(31),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(31),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(31),
    [sym_compatibility_alias_declaration] = STATE(31),
    [sym_property_implementation] = STATE(31),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(31),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(371),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(30)] = {
    [sym__block_item] = STATE(37),
    [sym_preproc_include] = STATE(37),
    [sym_preproc_def] = STATE(37),
    [sym_preproc_function_def] = STATE(37),
    [sym_preproc_call] = STATE(37),
    [sym_preproc_if] = STATE(37),
    [sym_preproc_ifdef] = STATE(37),
    [sym_preproc_else] = STATE(6365),
    [sym_preproc_elif] = STATE(6365),
    [sym_function_definition] = STATE(37),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(37),
    [sym_type_definition] = STATE(37),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(37),
    [sym_attribute_specifier] = STATE(37),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(37),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(37),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(37),
    [sym_preproc_undef] = STATE(37),
    [sym_preproc_linemarker] = STATE(37),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(37),
    [sym_class_declaration] = STATE(37),
    [sym_class_interface] = STATE(37),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(37),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(37),
    [sym_compatibility_alias_declaration] = STATE(37),
    [sym_property_implementation] = STATE(37),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(37),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(373),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(31)] = {
    [sym__block_item] = STATE(37),
    [sym_preproc_include] = STATE(37),
    [sym_preproc_def] = STATE(37),
    [sym_preproc_function_def] = STATE(37),
    [sym_preproc_call] = STATE(37),
    [sym_preproc_if] = STATE(37),
    [sym_preproc_ifdef] = STATE(37),
    [sym_preproc_else] = STATE(5950),
    [sym_preproc_elif] = STATE(5950),
    [sym_function_definition] = STATE(37),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(37),
    [sym_type_definition] = STATE(37),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(37),
    [sym_attribute_specifier] = STATE(37),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(37),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(37),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(37),
    [sym_preproc_undef] = STATE(37),
    [sym_preproc_linemarker] = STATE(37),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(37),
    [sym_class_declaration] = STATE(37),
    [sym_class_interface] = STATE(37),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(37),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(37),
    [sym_compatibility_alias_declaration] = STATE(37),
    [sym_property_implementation] = STATE(37),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(37),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(375),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(32)] = {
    [sym__block_item] = STATE(34),
    [sym_preproc_include] = STATE(34),
    [sym_preproc_def] = STATE(34),
    [sym_preproc_function_def] = STATE(34),
    [sym_preproc_call] = STATE(34),
    [sym_preproc_if] = STATE(34),
    [sym_preproc_ifdef] = STATE(34),
    [sym_preproc_else] = STATE(6465),
    [sym_preproc_elif] = STATE(6465),
    [sym_function_definition] = STATE(34),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(34),
    [sym_type_definition] = STATE(34),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(34),
    [sym_attribute_specifier] = STATE(34),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(34),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(34),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(34),
    [sym_preproc_undef] = STATE(34),
    [sym_preproc_linemarker] = STATE(34),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(34),
    [sym_class_declaration] = STATE(34),
    [sym_class_interface] = STATE(34),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(34),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(34),
    [sym_compatibility_alias_declaration] = STATE(34),
    [sym_property_implementation] = STATE(34),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(34),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(377),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(33)] = {
    [sym__block_item] = STATE(33),
    [sym_preproc_include] = STATE(33),
    [sym_preproc_def] = STATE(33),
    [sym_preproc_function_def] = STATE(33),
    [sym_preproc_call] = STATE(33),
    [sym_preproc_if] = STATE(33),
    [sym_preproc_ifdef] = STATE(33),
    [sym_function_definition] = STATE(33),
    [sym__old_style_function_definition] = STATE(299),
    [sym_declaration] = STATE(33),
    [sym_type_definition] = STATE(33),
    [sym__declaration_modifiers] = STATE(2038),
    [sym__declaration_specifiers] = STATE(4184),
    [sym_linkage_specification] = STATE(33),
    [sym_attribute_specifier] = STATE(33),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_ms_call_modifier] = STATE(2166),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2038),
    [sym_type_qualifier] = STATE(2038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3005),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(300),
    [sym_statement] = STATE(33),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(33),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(33),
    [sym_preproc_undef] = STATE(33),
    [sym_preproc_linemarker] = STATE(33),
    [sym_availability_attribute_specifier] = STATE(2038),
    [sym_protocol_forward_declaration] = STATE(33),
    [sym_class_declaration] = STATE(33),
    [sym_class_interface] = STATE(33),
    [sym__class_interface_header] = STATE(1172),
    [sym_class_implementation] = STATE(33),
    [sym__class_implementation_header] = STATE(1086),
    [sym_protocol_declaration] = STATE(33),
    [sym_compatibility_alias_declaration] = STATE(33),
    [sym_property_implementation] = STATE(33),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(33),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2038),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(379),
    [aux_sym_preproc_include_token1] = ACTIONS(382),
    [aux_sym_preproc_include_token2] = ACTIONS(382),
    [aux_sym_preproc_def_token1] = ACTIONS(385),
    [aux_sym_preproc_if_token1] = ACTIONS(388),
    [aux_sym_preproc_if_token2] = ACTIONS(391),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(393),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(393),
    [aux_sym_preproc_else_token1] = ACTIONS(391),
    [aux_sym_preproc_elif_token1] = ACTIONS(391),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(391),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(391),
    [sym_preproc_directive] = ACTIONS(396),
    [anon_sym_LPAREN2] = ACTIONS(399),
    [anon_sym_BANG] = ACTIONS(402),
    [anon_sym_TILDE] = ACTIONS(402),
    [anon_sym_DASH] = ACTIONS(405),
    [anon_sym_PLUS] = ACTIONS(405),
    [anon_sym_STAR] = ACTIONS(408),
    [anon_sym_CARET] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(408),
    [anon_sym_SEMI] = ACTIONS(414),
    [anon_sym___extension__] = ACTIONS(417),
    [anon_sym_typedef] = ACTIONS(420),
    [anon_sym_extern] = ACTIONS(423),
    [anon_sym___attribute__] = ACTIONS(426),
    [anon_sym___attribute] = ACTIONS(426),
    [anon_sym_noreturn] = ACTIONS(429),
    [anon_sym_LBRACK] = ACTIONS(432),
    [anon_sym___declspec] = ACTIONS(435),
    [anon_sym___cdecl] = ACTIONS(438),
    [anon_sym___clrcall] = ACTIONS(438),
    [anon_sym___stdcall] = ACTIONS(438),
    [anon_sym___fastcall] = ACTIONS(438),
    [anon_sym___thiscall] = ACTIONS(438),
    [anon_sym___vectorcall] = ACTIONS(438),
    [anon_sym_LBRACE] = ACTIONS(441),
    [anon_sym_signed] = ACTIONS(444),
    [anon_sym_unsigned] = ACTIONS(444),
    [anon_sym_long] = ACTIONS(444),
    [anon_sym_short] = ACTIONS(444),
    [anon_sym_static] = ACTIONS(447),
    [anon_sym_ATautoreleasepool] = ACTIONS(450),
    [anon_sym_auto] = ACTIONS(447),
    [anon_sym_register] = ACTIONS(447),
    [anon_sym_inline] = ACTIONS(447),
    [anon_sym___inline] = ACTIONS(447),
    [anon_sym___inline__] = ACTIONS(447),
    [anon_sym___forceinline] = ACTIONS(447),
    [anon_sym_thread_local] = ACTIONS(447),
    [anon_sym___thread] = ACTIONS(447),
    [anon_sym_CG_EXTERN] = ACTIONS(447),
    [anon_sym_CG_INLINE] = ACTIONS(447),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(447),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(447),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(447),
    [anon_sym_IBOutlet] = ACTIONS(447),
    [anon_sym_IBInspectable] = ACTIONS(447),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(447),
    [anon_sym_NS_INLINE] = ACTIONS(447),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(447),
    [anon_sym_OBJC_EXPORT] = ACTIONS(447),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(447),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(447),
    [anon_sym_const] = ACTIONS(453),
    [anon_sym_constexpr] = ACTIONS(429),
    [anon_sym_volatile] = ACTIONS(429),
    [anon_sym_restrict] = ACTIONS(429),
    [anon_sym___restrict__] = ACTIONS(429),
    [anon_sym__Atomic] = ACTIONS(429),
    [anon_sym__Noreturn] = ACTIONS(429),
    [anon_sym__Nonnull] = ACTIONS(429),
    [anon_sym_nullable] = ACTIONS(429),
    [anon_sym__Complex] = ACTIONS(429),
    [anon_sym__Nullable] = ACTIONS(429),
    [anon_sym__Nullable_result] = ACTIONS(429),
    [anon_sym__Null_unspecified] = ACTIONS(429),
    [anon_sym___autoreleasing] = ACTIONS(429),
    [anon_sym___block] = ACTIONS(429),
    [anon_sym___bridge] = ACTIONS(429),
    [anon_sym___bridge_retained] = ACTIONS(429),
    [anon_sym___bridge_transfer] = ACTIONS(429),
    [anon_sym___complex] = ACTIONS(429),
    [anon_sym___const] = ACTIONS(429),
    [anon_sym___imag] = ACTIONS(456),
    [anon_sym___kindof] = ACTIONS(429),
    [anon_sym___nonnull] = ACTIONS(429),
    [anon_sym___nullable] = ACTIONS(429),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(429),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(429),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(429),
    [anon_sym___real] = ACTIONS(456),
    [anon_sym___strong] = ACTIONS(429),
    [anon_sym___unsafe_unretained] = ACTIONS(429),
    [anon_sym___unused] = ACTIONS(429),
    [anon_sym___weak] = ACTIONS(429),
    [anon_sym_alignas] = ACTIONS(459),
    [anon_sym__Alignas] = ACTIONS(459),
    [sym_primitive_type] = ACTIONS(462),
    [anon_sym_enum] = ACTIONS(465),
    [anon_sym_struct] = ACTIONS(468),
    [anon_sym_union] = ACTIONS(471),
    [anon_sym_if] = ACTIONS(474),
    [anon_sym_switch] = ACTIONS(477),
    [anon_sym_case] = ACTIONS(480),
    [anon_sym_default] = ACTIONS(483),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_do] = ACTIONS(489),
    [anon_sym_for] = ACTIONS(492),
    [anon_sym_in] = ACTIONS(495),
    [anon_sym_return] = ACTIONS(498),
    [anon_sym_break] = ACTIONS(501),
    [anon_sym_continue] = ACTIONS(504),
    [anon_sym_goto] = ACTIONS(507),
    [anon_sym_DASH_DASH] = ACTIONS(510),
    [anon_sym_PLUS_PLUS] = ACTIONS(510),
    [anon_sym_sizeof] = ACTIONS(513),
    [anon_sym___alignof__] = ACTIONS(516),
    [anon_sym___alignof] = ACTIONS(516),
    [anon_sym__alignof] = ACTIONS(516),
    [anon_sym_alignof] = ACTIONS(516),
    [anon_sym__Alignof] = ACTIONS(516),
    [anon_sym_offsetof] = ACTIONS(519),
    [anon_sym__Generic] = ACTIONS(522),
    [anon_sym_asm] = ACTIONS(525),
    [anon_sym___asm__] = ACTIONS(525),
    [anon_sym___asm] = ACTIONS(528),
    [sym_number_literal] = ACTIONS(531),
    [anon_sym_L_SQUOTE] = ACTIONS(534),
    [anon_sym_u_SQUOTE] = ACTIONS(534),
    [anon_sym_U_SQUOTE] = ACTIONS(534),
    [anon_sym_u8_SQUOTE] = ACTIONS(534),
    [anon_sym_SQUOTE] = ACTIONS(534),
    [anon_sym_AT] = ACTIONS(537),
    [anon_sym_DQUOTE] = ACTIONS(540),
    [anon_sym_L_DQUOTE] = ACTIONS(540),
    [anon_sym_u_DQUOTE] = ACTIONS(540),
    [anon_sym_U_DQUOTE] = ACTIONS(540),
    [anon_sym_u8_DQUOTE] = ACTIONS(540),
    [sym_true] = ACTIONS(543),
    [sym_false] = ACTIONS(543),
    [anon_sym_NULL] = ACTIONS(546),
    [anon_sym_nullptr] = ACTIONS(546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(549),
    [anon_sym___typeof] = ACTIONS(549),
    [anon_sym_typeof] = ACTIONS(549),
    [anon_sym_ATimport] = ACTIONS(552),
    [aux_sym_preproc_undef_token1] = ACTIONS(555),
    [anon_sym_POUND] = ACTIONS(558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(561),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(561),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(561),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(561),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(561),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(561),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(561),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(561),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(561),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(564),
    [anon_sym_NS_AVAILABLE] = ACTIONS(564),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(564),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_API_AVAILABLE] = ACTIONS(564),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(564),
    [anon_sym_API_DEPRECATED] = ACTIONS(564),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(564),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(564),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(564),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(564),
    [anon_sym___deprecated_msg] = ACTIONS(564),
    [anon_sym___deprecated_enum_msg] = ACTIONS(564),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(564),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(564),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(564),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(564),
    [anon_sym_ATprotocol] = ACTIONS(567),
    [anon_sym_ATinterface] = ACTIONS(570),
    [anon_sym_ATimplementation] = ACTIONS(573),
    [anon_sym_ATcompatibility_alias] = ACTIONS(576),
    [anon_sym_ATsynthesize] = ACTIONS(579),
    [anon_sym_ATdynamic] = ACTIONS(582),
    [anon_sym_ATtry] = ACTIONS(585),
    [anon_sym___try] = ACTIONS(588),
    [anon_sym_ATthrow] = ACTIONS(591),
    [anon_sym_ATselector] = ACTIONS(594),
    [anon_sym_ATavailable] = ACTIONS(597),
    [anon_sym___builtin_available] = ACTIONS(600),
    [anon_sym_va_arg] = ACTIONS(603),
    [anon_sym_ATencode] = ACTIONS(606),
    [anon_sym_ATsynchronized] = ACTIONS(609),
    [anon_sym_BOOL] = ACTIONS(612),
    [anon_sym_IMP] = ACTIONS(612),
    [anon_sym_SEL] = ACTIONS(612),
    [anon_sym_Class] = ACTIONS(612),
    [anon_sym_id] = ACTIONS(615),
  },
  [STATE(34)] = {
    [sym__block_item] = STATE(37),
    [sym_preproc_include] = STATE(37),
    [sym_preproc_def] = STATE(37),
    [sym_preproc_function_def] = STATE(37),
    [sym_preproc_call] = STATE(37),
    [sym_preproc_if] = STATE(37),
    [sym_preproc_ifdef] = STATE(37),
    [sym_preproc_else] = STATE(6605),
    [sym_preproc_elif] = STATE(6605),
    [sym_function_definition] = STATE(37),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(37),
    [sym_type_definition] = STATE(37),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(37),
    [sym_attribute_specifier] = STATE(37),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(37),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(37),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(37),
    [sym_preproc_undef] = STATE(37),
    [sym_preproc_linemarker] = STATE(37),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(37),
    [sym_class_declaration] = STATE(37),
    [sym_class_interface] = STATE(37),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(37),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(37),
    [sym_compatibility_alias_declaration] = STATE(37),
    [sym_property_implementation] = STATE(37),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(37),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(618),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(35)] = {
    [sym__block_item] = STATE(24),
    [sym_preproc_include] = STATE(24),
    [sym_preproc_def] = STATE(24),
    [sym_preproc_function_def] = STATE(24),
    [sym_preproc_call] = STATE(24),
    [sym_preproc_if] = STATE(24),
    [sym_preproc_ifdef] = STATE(24),
    [sym_preproc_else] = STATE(6382),
    [sym_preproc_elif] = STATE(6382),
    [sym_function_definition] = STATE(24),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(24),
    [sym_type_definition] = STATE(24),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(24),
    [sym_attribute_specifier] = STATE(24),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(24),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(24),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(24),
    [sym_preproc_undef] = STATE(24),
    [sym_preproc_linemarker] = STATE(24),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(24),
    [sym_class_declaration] = STATE(24),
    [sym_class_interface] = STATE(24),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(24),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(24),
    [sym_compatibility_alias_declaration] = STATE(24),
    [sym_property_implementation] = STATE(24),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(24),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(620),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(36)] = {
    [sym__block_item] = STATE(30),
    [sym_preproc_include] = STATE(30),
    [sym_preproc_def] = STATE(30),
    [sym_preproc_function_def] = STATE(30),
    [sym_preproc_call] = STATE(30),
    [sym_preproc_if] = STATE(30),
    [sym_preproc_ifdef] = STATE(30),
    [sym_preproc_else] = STATE(6300),
    [sym_preproc_elif] = STATE(6300),
    [sym_function_definition] = STATE(30),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(30),
    [sym_type_definition] = STATE(30),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(30),
    [sym_attribute_specifier] = STATE(30),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(30),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(30),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(30),
    [sym_preproc_undef] = STATE(30),
    [sym_preproc_linemarker] = STATE(30),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(30),
    [sym_class_declaration] = STATE(30),
    [sym_class_interface] = STATE(30),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(30),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(30),
    [sym_compatibility_alias_declaration] = STATE(30),
    [sym_property_implementation] = STATE(30),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(30),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(279),
    [aux_sym_preproc_include_token1] = ACTIONS(281),
    [aux_sym_preproc_include_token2] = ACTIONS(281),
    [aux_sym_preproc_def_token1] = ACTIONS(283),
    [aux_sym_preproc_if_token1] = ACTIONS(285),
    [aux_sym_preproc_if_token2] = ACTIONS(622),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(289),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(289),
    [aux_sym_preproc_else_token1] = ACTIONS(163),
    [aux_sym_preproc_elif_token1] = ACTIONS(165),
    [sym_preproc_directive] = ACTIONS(291),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(295),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(299),
    [anon_sym___attribute__] = ACTIONS(301),
    [anon_sym___attribute] = ACTIONS(301),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(331),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(333),
    [aux_sym_preproc_undef_token1] = ACTIONS(335),
    [anon_sym_POUND] = ACTIONS(337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(339),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(341),
    [anon_sym_ATsynthesize] = ACTIONS(343),
    [anon_sym_ATdynamic] = ACTIONS(345),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(37)] = {
    [sym__block_item] = STATE(37),
    [sym_preproc_include] = STATE(37),
    [sym_preproc_def] = STATE(37),
    [sym_preproc_function_def] = STATE(37),
    [sym_preproc_call] = STATE(37),
    [sym_preproc_if] = STATE(37),
    [sym_preproc_ifdef] = STATE(37),
    [sym_function_definition] = STATE(37),
    [sym__old_style_function_definition] = STATE(471),
    [sym_declaration] = STATE(37),
    [sym_type_definition] = STATE(37),
    [sym__declaration_modifiers] = STATE(2040),
    [sym__declaration_specifiers] = STATE(4187),
    [sym_linkage_specification] = STATE(37),
    [sym_attribute_specifier] = STATE(37),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_ms_call_modifier] = STATE(2170),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2040),
    [sym_type_qualifier] = STATE(2040),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2959),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(474),
    [sym_statement] = STATE(37),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(37),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(37),
    [sym_preproc_undef] = STATE(37),
    [sym_preproc_linemarker] = STATE(37),
    [sym_availability_attribute_specifier] = STATE(2040),
    [sym_protocol_forward_declaration] = STATE(37),
    [sym_class_declaration] = STATE(37),
    [sym_class_interface] = STATE(37),
    [sym__class_interface_header] = STATE(1178),
    [sym_class_implementation] = STATE(37),
    [sym__class_implementation_header] = STATE(1087),
    [sym_protocol_declaration] = STATE(37),
    [sym_compatibility_alias_declaration] = STATE(37),
    [sym_property_implementation] = STATE(37),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(37),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2040),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(624),
    [aux_sym_preproc_include_token1] = ACTIONS(627),
    [aux_sym_preproc_include_token2] = ACTIONS(627),
    [aux_sym_preproc_def_token1] = ACTIONS(630),
    [aux_sym_preproc_if_token1] = ACTIONS(633),
    [aux_sym_preproc_if_token2] = ACTIONS(391),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(636),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(636),
    [aux_sym_preproc_else_token1] = ACTIONS(391),
    [aux_sym_preproc_elif_token1] = ACTIONS(391),
    [sym_preproc_directive] = ACTIONS(639),
    [anon_sym_LPAREN2] = ACTIONS(399),
    [anon_sym_BANG] = ACTIONS(402),
    [anon_sym_TILDE] = ACTIONS(402),
    [anon_sym_DASH] = ACTIONS(405),
    [anon_sym_PLUS] = ACTIONS(405),
    [anon_sym_STAR] = ACTIONS(408),
    [anon_sym_CARET] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(408),
    [anon_sym_SEMI] = ACTIONS(642),
    [anon_sym___extension__] = ACTIONS(645),
    [anon_sym_typedef] = ACTIONS(648),
    [anon_sym_extern] = ACTIONS(651),
    [anon_sym___attribute__] = ACTIONS(654),
    [anon_sym___attribute] = ACTIONS(654),
    [anon_sym_noreturn] = ACTIONS(429),
    [anon_sym_LBRACK] = ACTIONS(432),
    [anon_sym___declspec] = ACTIONS(435),
    [anon_sym___cdecl] = ACTIONS(438),
    [anon_sym___clrcall] = ACTIONS(438),
    [anon_sym___stdcall] = ACTIONS(438),
    [anon_sym___fastcall] = ACTIONS(438),
    [anon_sym___thiscall] = ACTIONS(438),
    [anon_sym___vectorcall] = ACTIONS(438),
    [anon_sym_LBRACE] = ACTIONS(657),
    [anon_sym_signed] = ACTIONS(444),
    [anon_sym_unsigned] = ACTIONS(444),
    [anon_sym_long] = ACTIONS(444),
    [anon_sym_short] = ACTIONS(444),
    [anon_sym_static] = ACTIONS(447),
    [anon_sym_ATautoreleasepool] = ACTIONS(660),
    [anon_sym_auto] = ACTIONS(447),
    [anon_sym_register] = ACTIONS(447),
    [anon_sym_inline] = ACTIONS(447),
    [anon_sym___inline] = ACTIONS(447),
    [anon_sym___inline__] = ACTIONS(447),
    [anon_sym___forceinline] = ACTIONS(447),
    [anon_sym_thread_local] = ACTIONS(447),
    [anon_sym___thread] = ACTIONS(447),
    [anon_sym_CG_EXTERN] = ACTIONS(447),
    [anon_sym_CG_INLINE] = ACTIONS(447),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(447),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(447),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(447),
    [anon_sym_IBOutlet] = ACTIONS(447),
    [anon_sym_IBInspectable] = ACTIONS(447),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(447),
    [anon_sym_NS_INLINE] = ACTIONS(447),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(447),
    [anon_sym_OBJC_EXPORT] = ACTIONS(447),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(447),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(447),
    [anon_sym_const] = ACTIONS(453),
    [anon_sym_constexpr] = ACTIONS(429),
    [anon_sym_volatile] = ACTIONS(429),
    [anon_sym_restrict] = ACTIONS(429),
    [anon_sym___restrict__] = ACTIONS(429),
    [anon_sym__Atomic] = ACTIONS(429),
    [anon_sym__Noreturn] = ACTIONS(429),
    [anon_sym__Nonnull] = ACTIONS(429),
    [anon_sym_nullable] = ACTIONS(429),
    [anon_sym__Complex] = ACTIONS(429),
    [anon_sym__Nullable] = ACTIONS(429),
    [anon_sym__Nullable_result] = ACTIONS(429),
    [anon_sym__Null_unspecified] = ACTIONS(429),
    [anon_sym___autoreleasing] = ACTIONS(429),
    [anon_sym___block] = ACTIONS(429),
    [anon_sym___bridge] = ACTIONS(429),
    [anon_sym___bridge_retained] = ACTIONS(429),
    [anon_sym___bridge_transfer] = ACTIONS(429),
    [anon_sym___complex] = ACTIONS(429),
    [anon_sym___const] = ACTIONS(429),
    [anon_sym___imag] = ACTIONS(456),
    [anon_sym___kindof] = ACTIONS(429),
    [anon_sym___nonnull] = ACTIONS(429),
    [anon_sym___nullable] = ACTIONS(429),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(429),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(429),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(429),
    [anon_sym___real] = ACTIONS(456),
    [anon_sym___strong] = ACTIONS(429),
    [anon_sym___unsafe_unretained] = ACTIONS(429),
    [anon_sym___unused] = ACTIONS(429),
    [anon_sym___weak] = ACTIONS(429),
    [anon_sym_alignas] = ACTIONS(459),
    [anon_sym__Alignas] = ACTIONS(459),
    [sym_primitive_type] = ACTIONS(462),
    [anon_sym_enum] = ACTIONS(465),
    [anon_sym_struct] = ACTIONS(468),
    [anon_sym_union] = ACTIONS(471),
    [anon_sym_if] = ACTIONS(663),
    [anon_sym_switch] = ACTIONS(666),
    [anon_sym_case] = ACTIONS(669),
    [anon_sym_default] = ACTIONS(672),
    [anon_sym_while] = ACTIONS(675),
    [anon_sym_do] = ACTIONS(678),
    [anon_sym_for] = ACTIONS(681),
    [anon_sym_in] = ACTIONS(495),
    [anon_sym_return] = ACTIONS(684),
    [anon_sym_break] = ACTIONS(687),
    [anon_sym_continue] = ACTIONS(690),
    [anon_sym_goto] = ACTIONS(693),
    [anon_sym_DASH_DASH] = ACTIONS(510),
    [anon_sym_PLUS_PLUS] = ACTIONS(510),
    [anon_sym_sizeof] = ACTIONS(513),
    [anon_sym___alignof__] = ACTIONS(516),
    [anon_sym___alignof] = ACTIONS(516),
    [anon_sym__alignof] = ACTIONS(516),
    [anon_sym_alignof] = ACTIONS(516),
    [anon_sym__Alignof] = ACTIONS(516),
    [anon_sym_offsetof] = ACTIONS(519),
    [anon_sym__Generic] = ACTIONS(522),
    [anon_sym_asm] = ACTIONS(525),
    [anon_sym___asm__] = ACTIONS(525),
    [anon_sym___asm] = ACTIONS(696),
    [sym_number_literal] = ACTIONS(531),
    [anon_sym_L_SQUOTE] = ACTIONS(534),
    [anon_sym_u_SQUOTE] = ACTIONS(534),
    [anon_sym_U_SQUOTE] = ACTIONS(534),
    [anon_sym_u8_SQUOTE] = ACTIONS(534),
    [anon_sym_SQUOTE] = ACTIONS(534),
    [anon_sym_AT] = ACTIONS(699),
    [anon_sym_DQUOTE] = ACTIONS(540),
    [anon_sym_L_DQUOTE] = ACTIONS(540),
    [anon_sym_u_DQUOTE] = ACTIONS(540),
    [anon_sym_U_DQUOTE] = ACTIONS(540),
    [anon_sym_u8_DQUOTE] = ACTIONS(540),
    [sym_true] = ACTIONS(543),
    [sym_false] = ACTIONS(543),
    [anon_sym_NULL] = ACTIONS(546),
    [anon_sym_nullptr] = ACTIONS(546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(549),
    [anon_sym___typeof] = ACTIONS(549),
    [anon_sym_typeof] = ACTIONS(549),
    [anon_sym_ATimport] = ACTIONS(702),
    [aux_sym_preproc_undef_token1] = ACTIONS(705),
    [anon_sym_POUND] = ACTIONS(708),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(561),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(561),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(561),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(561),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(561),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(561),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(561),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(561),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(561),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(564),
    [anon_sym_NS_AVAILABLE] = ACTIONS(564),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(564),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_API_AVAILABLE] = ACTIONS(564),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(564),
    [anon_sym_API_DEPRECATED] = ACTIONS(564),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(564),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(564),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(564),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(564),
    [anon_sym___deprecated_msg] = ACTIONS(564),
    [anon_sym___deprecated_enum_msg] = ACTIONS(564),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(564),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(564),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(564),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(564),
    [anon_sym_ATprotocol] = ACTIONS(711),
    [anon_sym_ATinterface] = ACTIONS(570),
    [anon_sym_ATimplementation] = ACTIONS(573),
    [anon_sym_ATcompatibility_alias] = ACTIONS(714),
    [anon_sym_ATsynthesize] = ACTIONS(717),
    [anon_sym_ATdynamic] = ACTIONS(720),
    [anon_sym_ATtry] = ACTIONS(723),
    [anon_sym___try] = ACTIONS(726),
    [anon_sym_ATthrow] = ACTIONS(729),
    [anon_sym_ATselector] = ACTIONS(594),
    [anon_sym_ATavailable] = ACTIONS(597),
    [anon_sym___builtin_available] = ACTIONS(600),
    [anon_sym_va_arg] = ACTIONS(603),
    [anon_sym_ATencode] = ACTIONS(606),
    [anon_sym_ATsynchronized] = ACTIONS(732),
    [anon_sym_BOOL] = ACTIONS(612),
    [anon_sym_IMP] = ACTIONS(612),
    [anon_sym_SEL] = ACTIONS(612),
    [anon_sym_Class] = ACTIONS(612),
    [anon_sym_id] = ACTIONS(615),
  },
  [STATE(38)] = {
    [sym__block_item] = STATE(38),
    [sym_preproc_include] = STATE(38),
    [sym_preproc_def] = STATE(38),
    [sym_preproc_function_def] = STATE(38),
    [sym_preproc_call] = STATE(38),
    [sym_preproc_if] = STATE(38),
    [sym_preproc_ifdef] = STATE(38),
    [sym_function_definition] = STATE(38),
    [sym__old_style_function_definition] = STATE(654),
    [sym_declaration] = STATE(38),
    [sym_type_definition] = STATE(38),
    [sym__declaration_modifiers] = STATE(2042),
    [sym__declaration_specifiers] = STATE(4191),
    [sym_linkage_specification] = STATE(38),
    [sym_attribute_specifier] = STATE(38),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2849),
    [sym_ms_call_modifier] = STATE(2173),
    [sym_compound_statement] = STATE(535),
    [sym_storage_class_specifier] = STATE(2042),
    [sym_type_qualifier] = STATE(2042),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3006),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(655),
    [sym_statement] = STATE(38),
    [sym_labeled_statement] = STATE(535),
    [sym_expression_statement] = STATE(535),
    [sym_if_statement] = STATE(535),
    [sym_switch_statement] = STATE(535),
    [sym_case_statement] = STATE(535),
    [sym_while_statement] = STATE(535),
    [sym_do_statement] = STATE(535),
    [sym_for_statement] = STATE(535),
    [sym_return_statement] = STATE(535),
    [sym_break_statement] = STATE(535),
    [sym_continue_statement] = STATE(535),
    [sym_goto_statement] = STATE(535),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(38),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(38),
    [sym_preproc_undef] = STATE(38),
    [sym_preproc_linemarker] = STATE(38),
    [sym_availability_attribute_specifier] = STATE(2042),
    [sym_protocol_forward_declaration] = STATE(38),
    [sym_class_declaration] = STATE(38),
    [sym_class_interface] = STATE(38),
    [sym__class_interface_header] = STATE(1167),
    [sym_class_implementation] = STATE(38),
    [sym__class_implementation_header] = STATE(1089),
    [sym_protocol_declaration] = STATE(38),
    [sym_compatibility_alias_declaration] = STATE(38),
    [sym_property_implementation] = STATE(38),
    [sym_try_statement] = STATE(535),
    [sym_throw_statement] = STATE(535),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(535),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(535),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(38),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2042),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(735),
    [aux_sym_preproc_include_token1] = ACTIONS(738),
    [aux_sym_preproc_include_token2] = ACTIONS(738),
    [aux_sym_preproc_def_token1] = ACTIONS(741),
    [aux_sym_preproc_if_token1] = ACTIONS(744),
    [aux_sym_preproc_if_token2] = ACTIONS(391),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(747),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(747),
    [sym_preproc_directive] = ACTIONS(750),
    [anon_sym_LPAREN2] = ACTIONS(399),
    [anon_sym_BANG] = ACTIONS(402),
    [anon_sym_TILDE] = ACTIONS(402),
    [anon_sym_DASH] = ACTIONS(405),
    [anon_sym_PLUS] = ACTIONS(405),
    [anon_sym_STAR] = ACTIONS(408),
    [anon_sym_CARET] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(408),
    [anon_sym_SEMI] = ACTIONS(753),
    [anon_sym___extension__] = ACTIONS(756),
    [anon_sym_typedef] = ACTIONS(759),
    [anon_sym_extern] = ACTIONS(762),
    [anon_sym___attribute__] = ACTIONS(765),
    [anon_sym___attribute] = ACTIONS(765),
    [anon_sym_noreturn] = ACTIONS(429),
    [anon_sym_LBRACK] = ACTIONS(432),
    [anon_sym___declspec] = ACTIONS(435),
    [anon_sym___cdecl] = ACTIONS(438),
    [anon_sym___clrcall] = ACTIONS(438),
    [anon_sym___stdcall] = ACTIONS(438),
    [anon_sym___fastcall] = ACTIONS(438),
    [anon_sym___thiscall] = ACTIONS(438),
    [anon_sym___vectorcall] = ACTIONS(438),
    [anon_sym_LBRACE] = ACTIONS(768),
    [anon_sym_signed] = ACTIONS(444),
    [anon_sym_unsigned] = ACTIONS(444),
    [anon_sym_long] = ACTIONS(444),
    [anon_sym_short] = ACTIONS(444),
    [anon_sym_static] = ACTIONS(447),
    [anon_sym_ATautoreleasepool] = ACTIONS(771),
    [anon_sym_auto] = ACTIONS(447),
    [anon_sym_register] = ACTIONS(447),
    [anon_sym_inline] = ACTIONS(447),
    [anon_sym___inline] = ACTIONS(447),
    [anon_sym___inline__] = ACTIONS(447),
    [anon_sym___forceinline] = ACTIONS(447),
    [anon_sym_thread_local] = ACTIONS(447),
    [anon_sym___thread] = ACTIONS(447),
    [anon_sym_CG_EXTERN] = ACTIONS(447),
    [anon_sym_CG_INLINE] = ACTIONS(447),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(447),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(447),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(447),
    [anon_sym_IBOutlet] = ACTIONS(447),
    [anon_sym_IBInspectable] = ACTIONS(447),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(447),
    [anon_sym_NS_INLINE] = ACTIONS(447),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(447),
    [anon_sym_OBJC_EXPORT] = ACTIONS(447),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(447),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(447),
    [anon_sym_const] = ACTIONS(453),
    [anon_sym_constexpr] = ACTIONS(429),
    [anon_sym_volatile] = ACTIONS(429),
    [anon_sym_restrict] = ACTIONS(429),
    [anon_sym___restrict__] = ACTIONS(429),
    [anon_sym__Atomic] = ACTIONS(429),
    [anon_sym__Noreturn] = ACTIONS(429),
    [anon_sym__Nonnull] = ACTIONS(429),
    [anon_sym_nullable] = ACTIONS(429),
    [anon_sym__Complex] = ACTIONS(429),
    [anon_sym__Nullable] = ACTIONS(429),
    [anon_sym__Nullable_result] = ACTIONS(429),
    [anon_sym__Null_unspecified] = ACTIONS(429),
    [anon_sym___autoreleasing] = ACTIONS(429),
    [anon_sym___block] = ACTIONS(429),
    [anon_sym___bridge] = ACTIONS(429),
    [anon_sym___bridge_retained] = ACTIONS(429),
    [anon_sym___bridge_transfer] = ACTIONS(429),
    [anon_sym___complex] = ACTIONS(429),
    [anon_sym___const] = ACTIONS(429),
    [anon_sym___imag] = ACTIONS(456),
    [anon_sym___kindof] = ACTIONS(429),
    [anon_sym___nonnull] = ACTIONS(429),
    [anon_sym___nullable] = ACTIONS(429),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(429),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(429),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(429),
    [anon_sym___real] = ACTIONS(456),
    [anon_sym___strong] = ACTIONS(429),
    [anon_sym___unsafe_unretained] = ACTIONS(429),
    [anon_sym___unused] = ACTIONS(429),
    [anon_sym___weak] = ACTIONS(429),
    [anon_sym_alignas] = ACTIONS(459),
    [anon_sym__Alignas] = ACTIONS(459),
    [sym_primitive_type] = ACTIONS(462),
    [anon_sym_enum] = ACTIONS(465),
    [anon_sym_struct] = ACTIONS(468),
    [anon_sym_union] = ACTIONS(471),
    [anon_sym_if] = ACTIONS(774),
    [anon_sym_switch] = ACTIONS(777),
    [anon_sym_case] = ACTIONS(780),
    [anon_sym_default] = ACTIONS(783),
    [anon_sym_while] = ACTIONS(786),
    [anon_sym_do] = ACTIONS(789),
    [anon_sym_for] = ACTIONS(792),
    [anon_sym_in] = ACTIONS(495),
    [anon_sym_return] = ACTIONS(795),
    [anon_sym_break] = ACTIONS(798),
    [anon_sym_continue] = ACTIONS(801),
    [anon_sym_goto] = ACTIONS(804),
    [anon_sym_DASH_DASH] = ACTIONS(510),
    [anon_sym_PLUS_PLUS] = ACTIONS(510),
    [anon_sym_sizeof] = ACTIONS(513),
    [anon_sym___alignof__] = ACTIONS(516),
    [anon_sym___alignof] = ACTIONS(516),
    [anon_sym__alignof] = ACTIONS(516),
    [anon_sym_alignof] = ACTIONS(516),
    [anon_sym__Alignof] = ACTIONS(516),
    [anon_sym_offsetof] = ACTIONS(519),
    [anon_sym__Generic] = ACTIONS(522),
    [anon_sym_asm] = ACTIONS(525),
    [anon_sym___asm__] = ACTIONS(525),
    [anon_sym___asm] = ACTIONS(807),
    [sym_number_literal] = ACTIONS(531),
    [anon_sym_L_SQUOTE] = ACTIONS(534),
    [anon_sym_u_SQUOTE] = ACTIONS(534),
    [anon_sym_U_SQUOTE] = ACTIONS(534),
    [anon_sym_u8_SQUOTE] = ACTIONS(534),
    [anon_sym_SQUOTE] = ACTIONS(534),
    [anon_sym_AT] = ACTIONS(810),
    [anon_sym_DQUOTE] = ACTIONS(540),
    [anon_sym_L_DQUOTE] = ACTIONS(540),
    [anon_sym_u_DQUOTE] = ACTIONS(540),
    [anon_sym_U_DQUOTE] = ACTIONS(540),
    [anon_sym_u8_DQUOTE] = ACTIONS(540),
    [sym_true] = ACTIONS(543),
    [sym_false] = ACTIONS(543),
    [anon_sym_NULL] = ACTIONS(546),
    [anon_sym_nullptr] = ACTIONS(546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(549),
    [anon_sym___typeof] = ACTIONS(549),
    [anon_sym_typeof] = ACTIONS(549),
    [anon_sym_ATimport] = ACTIONS(813),
    [aux_sym_preproc_undef_token1] = ACTIONS(816),
    [anon_sym_POUND] = ACTIONS(819),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(561),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(561),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(561),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(561),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(561),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(561),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(561),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(561),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(561),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(564),
    [anon_sym_NS_AVAILABLE] = ACTIONS(564),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(564),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_API_AVAILABLE] = ACTIONS(564),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(564),
    [anon_sym_API_DEPRECATED] = ACTIONS(564),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(564),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(564),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(564),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(564),
    [anon_sym___deprecated_msg] = ACTIONS(564),
    [anon_sym___deprecated_enum_msg] = ACTIONS(564),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(564),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(564),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(564),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(564),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(564),
    [anon_sym_ATprotocol] = ACTIONS(822),
    [anon_sym_ATinterface] = ACTIONS(570),
    [anon_sym_ATimplementation] = ACTIONS(573),
    [anon_sym_ATcompatibility_alias] = ACTIONS(825),
    [anon_sym_ATsynthesize] = ACTIONS(828),
    [anon_sym_ATdynamic] = ACTIONS(831),
    [anon_sym_ATtry] = ACTIONS(834),
    [anon_sym___try] = ACTIONS(837),
    [anon_sym_ATthrow] = ACTIONS(840),
    [anon_sym_ATselector] = ACTIONS(594),
    [anon_sym_ATavailable] = ACTIONS(597),
    [anon_sym___builtin_available] = ACTIONS(600),
    [anon_sym_va_arg] = ACTIONS(603),
    [anon_sym_ATencode] = ACTIONS(606),
    [anon_sym_ATsynchronized] = ACTIONS(843),
    [anon_sym_BOOL] = ACTIONS(612),
    [anon_sym_IMP] = ACTIONS(612),
    [anon_sym_SEL] = ACTIONS(612),
    [anon_sym_Class] = ACTIONS(612),
    [anon_sym_id] = ACTIONS(615),
  },
  [STATE(39)] = {
    [sym__block_item] = STATE(40),
    [sym_preproc_include] = STATE(40),
    [sym_preproc_def] = STATE(40),
    [sym_preproc_function_def] = STATE(40),
    [sym_preproc_call] = STATE(40),
    [sym_preproc_if] = STATE(40),
    [sym_preproc_ifdef] = STATE(40),
    [sym_function_definition] = STATE(40),
    [sym__old_style_function_definition] = STATE(654),
    [sym_declaration] = STATE(40),
    [sym_type_definition] = STATE(40),
    [sym__declaration_modifiers] = STATE(2042),
    [sym__declaration_specifiers] = STATE(4191),
    [sym_linkage_specification] = STATE(40),
    [sym_attribute_specifier] = STATE(40),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2849),
    [sym_ms_call_modifier] = STATE(2173),
    [sym_compound_statement] = STATE(535),
    [sym_storage_class_specifier] = STATE(2042),
    [sym_type_qualifier] = STATE(2042),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3006),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(655),
    [sym_statement] = STATE(40),
    [sym_labeled_statement] = STATE(535),
    [sym_expression_statement] = STATE(535),
    [sym_if_statement] = STATE(535),
    [sym_switch_statement] = STATE(535),
    [sym_case_statement] = STATE(535),
    [sym_while_statement] = STATE(535),
    [sym_do_statement] = STATE(535),
    [sym_for_statement] = STATE(535),
    [sym_return_statement] = STATE(535),
    [sym_break_statement] = STATE(535),
    [sym_continue_statement] = STATE(535),
    [sym_goto_statement] = STATE(535),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(40),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(40),
    [sym_preproc_undef] = STATE(40),
    [sym_preproc_linemarker] = STATE(40),
    [sym_availability_attribute_specifier] = STATE(2042),
    [sym_protocol_forward_declaration] = STATE(40),
    [sym_class_declaration] = STATE(40),
    [sym_class_interface] = STATE(40),
    [sym__class_interface_header] = STATE(1167),
    [sym_class_implementation] = STATE(40),
    [sym__class_implementation_header] = STATE(1089),
    [sym_protocol_declaration] = STATE(40),
    [sym_compatibility_alias_declaration] = STATE(40),
    [sym_property_implementation] = STATE(40),
    [sym_try_statement] = STATE(535),
    [sym_throw_statement] = STATE(535),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(535),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(535),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(40),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2042),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(846),
    [aux_sym_preproc_include_token1] = ACTIONS(848),
    [aux_sym_preproc_include_token2] = ACTIONS(848),
    [aux_sym_preproc_def_token1] = ACTIONS(850),
    [aux_sym_preproc_if_token1] = ACTIONS(852),
    [aux_sym_preproc_if_token2] = ACTIONS(854),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(856),
    [sym_preproc_directive] = ACTIONS(858),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(862),
    [anon_sym_typedef] = ACTIONS(864),
    [anon_sym_extern] = ACTIONS(866),
    [anon_sym___attribute__] = ACTIONS(868),
    [anon_sym___attribute] = ACTIONS(868),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(878),
    [anon_sym_default] = ACTIONS(880),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(898),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(900),
    [aux_sym_preproc_undef_token1] = ACTIONS(902),
    [anon_sym_POUND] = ACTIONS(904),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(906),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(908),
    [anon_sym_ATsynthesize] = ACTIONS(910),
    [anon_sym_ATdynamic] = ACTIONS(912),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(40)] = {
    [sym__block_item] = STATE(38),
    [sym_preproc_include] = STATE(38),
    [sym_preproc_def] = STATE(38),
    [sym_preproc_function_def] = STATE(38),
    [sym_preproc_call] = STATE(38),
    [sym_preproc_if] = STATE(38),
    [sym_preproc_ifdef] = STATE(38),
    [sym_function_definition] = STATE(38),
    [sym__old_style_function_definition] = STATE(654),
    [sym_declaration] = STATE(38),
    [sym_type_definition] = STATE(38),
    [sym__declaration_modifiers] = STATE(2042),
    [sym__declaration_specifiers] = STATE(4191),
    [sym_linkage_specification] = STATE(38),
    [sym_attribute_specifier] = STATE(38),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2849),
    [sym_ms_call_modifier] = STATE(2173),
    [sym_compound_statement] = STATE(535),
    [sym_storage_class_specifier] = STATE(2042),
    [sym_type_qualifier] = STATE(2042),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(3006),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(655),
    [sym_statement] = STATE(38),
    [sym_labeled_statement] = STATE(535),
    [sym_expression_statement] = STATE(535),
    [sym_if_statement] = STATE(535),
    [sym_switch_statement] = STATE(535),
    [sym_case_statement] = STATE(535),
    [sym_while_statement] = STATE(535),
    [sym_do_statement] = STATE(535),
    [sym_for_statement] = STATE(535),
    [sym_return_statement] = STATE(535),
    [sym_break_statement] = STATE(535),
    [sym_continue_statement] = STATE(535),
    [sym_goto_statement] = STATE(535),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(38),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(38),
    [sym_preproc_undef] = STATE(38),
    [sym_preproc_linemarker] = STATE(38),
    [sym_availability_attribute_specifier] = STATE(2042),
    [sym_protocol_forward_declaration] = STATE(38),
    [sym_class_declaration] = STATE(38),
    [sym_class_interface] = STATE(38),
    [sym__class_interface_header] = STATE(1167),
    [sym_class_implementation] = STATE(38),
    [sym__class_implementation_header] = STATE(1089),
    [sym_protocol_declaration] = STATE(38),
    [sym_compatibility_alias_declaration] = STATE(38),
    [sym_property_implementation] = STATE(38),
    [sym_try_statement] = STATE(535),
    [sym_throw_statement] = STATE(535),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(535),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(535),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_preproc_if_repeat1] = STATE(38),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2042),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(846),
    [aux_sym_preproc_include_token1] = ACTIONS(848),
    [aux_sym_preproc_include_token2] = ACTIONS(848),
    [aux_sym_preproc_def_token1] = ACTIONS(850),
    [aux_sym_preproc_if_token1] = ACTIONS(852),
    [aux_sym_preproc_if_token2] = ACTIONS(922),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(856),
    [sym_preproc_directive] = ACTIONS(858),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(862),
    [anon_sym_typedef] = ACTIONS(864),
    [anon_sym_extern] = ACTIONS(866),
    [anon_sym___attribute__] = ACTIONS(868),
    [anon_sym___attribute] = ACTIONS(868),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(878),
    [anon_sym_default] = ACTIONS(880),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(898),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(900),
    [aux_sym_preproc_undef_token1] = ACTIONS(902),
    [anon_sym_POUND] = ACTIONS(904),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(906),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(908),
    [anon_sym_ATsynthesize] = ACTIONS(910),
    [anon_sym_ATdynamic] = ACTIONS(912),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(41)] = {
    [sym__block_item] = STATE(117),
    [sym_preproc_include] = STATE(117),
    [sym_preproc_def] = STATE(117),
    [sym_preproc_function_def] = STATE(117),
    [sym_preproc_call] = STATE(117),
    [sym_preproc_if] = STATE(117),
    [sym_preproc_ifdef] = STATE(117),
    [sym_function_definition] = STATE(117),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(117),
    [sym_type_definition] = STATE(117),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(117),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(117),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(117),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(117),
    [sym_preproc_undef] = STATE(117),
    [sym_preproc_linemarker] = STATE(117),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(117),
    [sym_class_declaration] = STATE(117),
    [sym_class_interface] = STATE(117),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(117),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(117),
    [sym_compatibility_alias_declaration] = STATE(117),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(117),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(946),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(42)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(994),
    [aux_sym_preproc_include_token1] = ACTIONS(997),
    [aux_sym_preproc_include_token2] = ACTIONS(997),
    [aux_sym_preproc_def_token1] = ACTIONS(1000),
    [aux_sym_preproc_if_token1] = ACTIONS(1003),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1006),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1006),
    [sym_preproc_directive] = ACTIONS(1009),
    [anon_sym_LPAREN2] = ACTIONS(1012),
    [anon_sym_BANG] = ACTIONS(1015),
    [anon_sym_TILDE] = ACTIONS(1015),
    [anon_sym_DASH] = ACTIONS(1018),
    [anon_sym_PLUS] = ACTIONS(1018),
    [anon_sym_STAR] = ACTIONS(1021),
    [anon_sym_CARET] = ACTIONS(1024),
    [anon_sym_AMP] = ACTIONS(1021),
    [anon_sym_SEMI] = ACTIONS(1027),
    [anon_sym___extension__] = ACTIONS(1030),
    [anon_sym_typedef] = ACTIONS(1033),
    [anon_sym_extern] = ACTIONS(1036),
    [anon_sym___attribute__] = ACTIONS(1039),
    [anon_sym___attribute] = ACTIONS(1039),
    [anon_sym_noreturn] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1045),
    [anon_sym___declspec] = ACTIONS(1048),
    [anon_sym___cdecl] = ACTIONS(1051),
    [anon_sym___clrcall] = ACTIONS(1051),
    [anon_sym___stdcall] = ACTIONS(1051),
    [anon_sym___fastcall] = ACTIONS(1051),
    [anon_sym___thiscall] = ACTIONS(1051),
    [anon_sym___vectorcall] = ACTIONS(1051),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_RBRACE] = ACTIONS(1057),
    [anon_sym_signed] = ACTIONS(1059),
    [anon_sym_unsigned] = ACTIONS(1059),
    [anon_sym_long] = ACTIONS(1059),
    [anon_sym_short] = ACTIONS(1059),
    [anon_sym_static] = ACTIONS(1062),
    [anon_sym_ATautoreleasepool] = ACTIONS(1065),
    [anon_sym_auto] = ACTIONS(1062),
    [anon_sym_register] = ACTIONS(1062),
    [anon_sym_inline] = ACTIONS(1062),
    [anon_sym___inline] = ACTIONS(1062),
    [anon_sym___inline__] = ACTIONS(1062),
    [anon_sym___forceinline] = ACTIONS(1062),
    [anon_sym_thread_local] = ACTIONS(1062),
    [anon_sym___thread] = ACTIONS(1062),
    [anon_sym_CG_EXTERN] = ACTIONS(1062),
    [anon_sym_CG_INLINE] = ACTIONS(1062),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(1062),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(1062),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(1062),
    [anon_sym_IBOutlet] = ACTIONS(1062),
    [anon_sym_IBInspectable] = ACTIONS(1062),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(1062),
    [anon_sym_NS_INLINE] = ACTIONS(1062),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(1062),
    [anon_sym_OBJC_EXPORT] = ACTIONS(1062),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(1062),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(1062),
    [anon_sym_const] = ACTIONS(1068),
    [anon_sym_constexpr] = ACTIONS(1042),
    [anon_sym_volatile] = ACTIONS(1042),
    [anon_sym_restrict] = ACTIONS(1042),
    [anon_sym___restrict__] = ACTIONS(1042),
    [anon_sym__Atomic] = ACTIONS(1042),
    [anon_sym__Noreturn] = ACTIONS(1042),
    [anon_sym__Nonnull] = ACTIONS(1042),
    [anon_sym_nullable] = ACTIONS(1042),
    [anon_sym__Complex] = ACTIONS(1042),
    [anon_sym__Nullable] = ACTIONS(1042),
    [anon_sym__Nullable_result] = ACTIONS(1042),
    [anon_sym__Null_unspecified] = ACTIONS(1042),
    [anon_sym___autoreleasing] = ACTIONS(1042),
    [anon_sym___block] = ACTIONS(1042),
    [anon_sym___bridge] = ACTIONS(1042),
    [anon_sym___bridge_retained] = ACTIONS(1042),
    [anon_sym___bridge_transfer] = ACTIONS(1042),
    [anon_sym___complex] = ACTIONS(1042),
    [anon_sym___const] = ACTIONS(1042),
    [anon_sym___imag] = ACTIONS(1071),
    [anon_sym___kindof] = ACTIONS(1042),
    [anon_sym___nonnull] = ACTIONS(1042),
    [anon_sym___nullable] = ACTIONS(1042),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(1042),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(1042),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(1042),
    [anon_sym___real] = ACTIONS(1071),
    [anon_sym___strong] = ACTIONS(1042),
    [anon_sym___unsafe_unretained] = ACTIONS(1042),
    [anon_sym___unused] = ACTIONS(1042),
    [anon_sym___weak] = ACTIONS(1042),
    [anon_sym_alignas] = ACTIONS(1074),
    [anon_sym__Alignas] = ACTIONS(1074),
    [sym_primitive_type] = ACTIONS(1077),
    [anon_sym_enum] = ACTIONS(1080),
    [anon_sym_struct] = ACTIONS(1083),
    [anon_sym_union] = ACTIONS(1086),
    [anon_sym_if] = ACTIONS(1089),
    [anon_sym_switch] = ACTIONS(1092),
    [anon_sym_case] = ACTIONS(1095),
    [anon_sym_default] = ACTIONS(1098),
    [anon_sym_while] = ACTIONS(1101),
    [anon_sym_do] = ACTIONS(1104),
    [anon_sym_for] = ACTIONS(1107),
    [anon_sym_in] = ACTIONS(1110),
    [anon_sym_return] = ACTIONS(1113),
    [anon_sym_break] = ACTIONS(1116),
    [anon_sym_continue] = ACTIONS(1119),
    [anon_sym_goto] = ACTIONS(1122),
    [anon_sym_DASH_DASH] = ACTIONS(1125),
    [anon_sym_PLUS_PLUS] = ACTIONS(1125),
    [anon_sym_sizeof] = ACTIONS(1128),
    [anon_sym___alignof__] = ACTIONS(1131),
    [anon_sym___alignof] = ACTIONS(1131),
    [anon_sym__alignof] = ACTIONS(1131),
    [anon_sym_alignof] = ACTIONS(1131),
    [anon_sym__Alignof] = ACTIONS(1131),
    [anon_sym_offsetof] = ACTIONS(1134),
    [anon_sym__Generic] = ACTIONS(1137),
    [anon_sym_asm] = ACTIONS(1140),
    [anon_sym___asm__] = ACTIONS(1140),
    [anon_sym___asm] = ACTIONS(1143),
    [sym_number_literal] = ACTIONS(1146),
    [anon_sym_L_SQUOTE] = ACTIONS(1149),
    [anon_sym_u_SQUOTE] = ACTIONS(1149),
    [anon_sym_U_SQUOTE] = ACTIONS(1149),
    [anon_sym_u8_SQUOTE] = ACTIONS(1149),
    [anon_sym_SQUOTE] = ACTIONS(1149),
    [anon_sym_AT] = ACTIONS(1152),
    [anon_sym_DQUOTE] = ACTIONS(1155),
    [anon_sym_L_DQUOTE] = ACTIONS(1155),
    [anon_sym_u_DQUOTE] = ACTIONS(1155),
    [anon_sym_U_DQUOTE] = ACTIONS(1155),
    [anon_sym_u8_DQUOTE] = ACTIONS(1155),
    [sym_true] = ACTIONS(1158),
    [sym_false] = ACTIONS(1158),
    [anon_sym_NULL] = ACTIONS(1161),
    [anon_sym_nullptr] = ACTIONS(1161),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(1164),
    [anon_sym___typeof] = ACTIONS(1164),
    [anon_sym_typeof] = ACTIONS(1164),
    [anon_sym_ATimport] = ACTIONS(1167),
    [aux_sym_preproc_undef_token1] = ACTIONS(1170),
    [anon_sym_POUND] = ACTIONS(1173),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(1176),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(1176),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(1176),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(1176),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(1176),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(1176),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(1176),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(1176),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(1176),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(1179),
    [anon_sym_NS_AVAILABLE] = ACTIONS(1179),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(1179),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(1179),
    [anon_sym_API_AVAILABLE] = ACTIONS(1179),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(1179),
    [anon_sym_API_DEPRECATED] = ACTIONS(1179),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(1179),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(1179),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(1179),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(1179),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(1179),
    [anon_sym___deprecated_msg] = ACTIONS(1179),
    [anon_sym___deprecated_enum_msg] = ACTIONS(1179),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(1179),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(1179),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(1179),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(1179),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(1179),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(1179),
    [anon_sym_ATprotocol] = ACTIONS(1182),
    [anon_sym_ATinterface] = ACTIONS(1185),
    [anon_sym_ATimplementation] = ACTIONS(1188),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1191),
    [anon_sym_ATtry] = ACTIONS(1194),
    [anon_sym___try] = ACTIONS(1197),
    [anon_sym_ATthrow] = ACTIONS(1200),
    [anon_sym_ATselector] = ACTIONS(1203),
    [anon_sym_ATavailable] = ACTIONS(1206),
    [anon_sym___builtin_available] = ACTIONS(1209),
    [anon_sym_va_arg] = ACTIONS(1212),
    [anon_sym_ATencode] = ACTIONS(1215),
    [anon_sym_ATsynchronized] = ACTIONS(1218),
    [anon_sym_BOOL] = ACTIONS(1221),
    [anon_sym_IMP] = ACTIONS(1221),
    [anon_sym_SEL] = ACTIONS(1221),
    [anon_sym_Class] = ACTIONS(1221),
    [anon_sym_id] = ACTIONS(1224),
  },
  [STATE(43)] = {
    [sym__block_item] = STATE(46),
    [sym_preproc_include] = STATE(46),
    [sym_preproc_def] = STATE(46),
    [sym_preproc_function_def] = STATE(46),
    [sym_preproc_call] = STATE(46),
    [sym_preproc_if] = STATE(46),
    [sym_preproc_ifdef] = STATE(46),
    [sym_function_definition] = STATE(46),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(46),
    [sym_type_definition] = STATE(46),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(46),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(46),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(46),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(46),
    [sym_preproc_undef] = STATE(46),
    [sym_preproc_linemarker] = STATE(46),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(46),
    [sym_class_declaration] = STATE(46),
    [sym_class_interface] = STATE(46),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(46),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(46),
    [sym_compatibility_alias_declaration] = STATE(46),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(46),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1227),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(44)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1229),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(45)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1231),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(46)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1233),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(47)] = {
    [sym__block_item] = STATE(49),
    [sym_preproc_include] = STATE(49),
    [sym_preproc_def] = STATE(49),
    [sym_preproc_function_def] = STATE(49),
    [sym_preproc_call] = STATE(49),
    [sym_preproc_if] = STATE(49),
    [sym_preproc_ifdef] = STATE(49),
    [sym_function_definition] = STATE(49),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(49),
    [sym_type_definition] = STATE(49),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(49),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(49),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(49),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(49),
    [sym_preproc_undef] = STATE(49),
    [sym_preproc_linemarker] = STATE(49),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(49),
    [sym_class_declaration] = STATE(49),
    [sym_class_interface] = STATE(49),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(49),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(49),
    [sym_compatibility_alias_declaration] = STATE(49),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(49),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1235),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(48)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1237),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(49)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1239),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(50)] = {
    [sym__block_item] = STATE(54),
    [sym_preproc_include] = STATE(54),
    [sym_preproc_def] = STATE(54),
    [sym_preproc_function_def] = STATE(54),
    [sym_preproc_call] = STATE(54),
    [sym_preproc_if] = STATE(54),
    [sym_preproc_ifdef] = STATE(54),
    [sym_function_definition] = STATE(54),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(54),
    [sym_type_definition] = STATE(54),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(54),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(54),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(54),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(54),
    [sym_preproc_undef] = STATE(54),
    [sym_preproc_linemarker] = STATE(54),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(54),
    [sym_class_declaration] = STATE(54),
    [sym_class_interface] = STATE(54),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(54),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(54),
    [sym_compatibility_alias_declaration] = STATE(54),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(54),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1239),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(51)] = {
    [sym__block_item] = STATE(48),
    [sym_preproc_include] = STATE(48),
    [sym_preproc_def] = STATE(48),
    [sym_preproc_function_def] = STATE(48),
    [sym_preproc_call] = STATE(48),
    [sym_preproc_if] = STATE(48),
    [sym_preproc_ifdef] = STATE(48),
    [sym_function_definition] = STATE(48),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(48),
    [sym_type_definition] = STATE(48),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(48),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(48),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(48),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(48),
    [sym_preproc_undef] = STATE(48),
    [sym_preproc_linemarker] = STATE(48),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(48),
    [sym_class_declaration] = STATE(48),
    [sym_class_interface] = STATE(48),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(48),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(48),
    [sym_compatibility_alias_declaration] = STATE(48),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(48),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1241),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(52)] = {
    [sym__block_item] = STATE(45),
    [sym_preproc_include] = STATE(45),
    [sym_preproc_def] = STATE(45),
    [sym_preproc_function_def] = STATE(45),
    [sym_preproc_call] = STATE(45),
    [sym_preproc_if] = STATE(45),
    [sym_preproc_ifdef] = STATE(45),
    [sym_function_definition] = STATE(45),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(45),
    [sym_type_definition] = STATE(45),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(45),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(45),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(45),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(45),
    [sym_preproc_undef] = STATE(45),
    [sym_preproc_linemarker] = STATE(45),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(45),
    [sym_class_declaration] = STATE(45),
    [sym_class_interface] = STATE(45),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(45),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(45),
    [sym_compatibility_alias_declaration] = STATE(45),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(45),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1237),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(53)] = {
    [sym__block_item] = STATE(55),
    [sym_preproc_include] = STATE(55),
    [sym_preproc_def] = STATE(55),
    [sym_preproc_function_def] = STATE(55),
    [sym_preproc_call] = STATE(55),
    [sym_preproc_if] = STATE(55),
    [sym_preproc_ifdef] = STATE(55),
    [sym_function_definition] = STATE(55),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(55),
    [sym_type_definition] = STATE(55),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(55),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(55),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(55),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(55),
    [sym_preproc_undef] = STATE(55),
    [sym_preproc_linemarker] = STATE(55),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(55),
    [sym_class_declaration] = STATE(55),
    [sym_class_interface] = STATE(55),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(55),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(55),
    [sym_compatibility_alias_declaration] = STATE(55),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(55),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1243),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(54)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1245),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(55)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1247),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(56)] = {
    [sym__block_item] = STATE(57),
    [sym_preproc_include] = STATE(57),
    [sym_preproc_def] = STATE(57),
    [sym_preproc_function_def] = STATE(57),
    [sym_preproc_call] = STATE(57),
    [sym_preproc_if] = STATE(57),
    [sym_preproc_ifdef] = STATE(57),
    [sym_function_definition] = STATE(57),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(57),
    [sym_type_definition] = STATE(57),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(57),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(57),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(57),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(57),
    [sym_preproc_undef] = STATE(57),
    [sym_preproc_linemarker] = STATE(57),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(57),
    [sym_class_declaration] = STATE(57),
    [sym_class_interface] = STATE(57),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(57),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(57),
    [sym_compatibility_alias_declaration] = STATE(57),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(57),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1249),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(57)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1251),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(58)] = {
    [sym__block_item] = STATE(60),
    [sym_preproc_include] = STATE(60),
    [sym_preproc_def] = STATE(60),
    [sym_preproc_function_def] = STATE(60),
    [sym_preproc_call] = STATE(60),
    [sym_preproc_if] = STATE(60),
    [sym_preproc_ifdef] = STATE(60),
    [sym_function_definition] = STATE(60),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(60),
    [sym_type_definition] = STATE(60),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(60),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(60),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(60),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(60),
    [sym_preproc_undef] = STATE(60),
    [sym_preproc_linemarker] = STATE(60),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(60),
    [sym_class_declaration] = STATE(60),
    [sym_class_interface] = STATE(60),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(60),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(60),
    [sym_compatibility_alias_declaration] = STATE(60),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(60),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1251),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(59)] = {
    [sym__block_item] = STATE(61),
    [sym_preproc_include] = STATE(61),
    [sym_preproc_def] = STATE(61),
    [sym_preproc_function_def] = STATE(61),
    [sym_preproc_call] = STATE(61),
    [sym_preproc_if] = STATE(61),
    [sym_preproc_ifdef] = STATE(61),
    [sym_function_definition] = STATE(61),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(61),
    [sym_type_definition] = STATE(61),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(61),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(61),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(61),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(61),
    [sym_preproc_undef] = STATE(61),
    [sym_preproc_linemarker] = STATE(61),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(61),
    [sym_class_declaration] = STATE(61),
    [sym_class_interface] = STATE(61),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(61),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(61),
    [sym_compatibility_alias_declaration] = STATE(61),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(61),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1253),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(60)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1255),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(61)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1257),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(62)] = {
    [sym__block_item] = STATE(63),
    [sym_preproc_include] = STATE(63),
    [sym_preproc_def] = STATE(63),
    [sym_preproc_function_def] = STATE(63),
    [sym_preproc_call] = STATE(63),
    [sym_preproc_if] = STATE(63),
    [sym_preproc_ifdef] = STATE(63),
    [sym_function_definition] = STATE(63),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(63),
    [sym_type_definition] = STATE(63),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(63),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(63),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(63),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(63),
    [sym_preproc_undef] = STATE(63),
    [sym_preproc_linemarker] = STATE(63),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(63),
    [sym_class_declaration] = STATE(63),
    [sym_class_interface] = STATE(63),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(63),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(63),
    [sym_compatibility_alias_declaration] = STATE(63),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(63),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1259),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(63)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1261),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(64)] = {
    [sym__block_item] = STATE(66),
    [sym_preproc_include] = STATE(66),
    [sym_preproc_def] = STATE(66),
    [sym_preproc_function_def] = STATE(66),
    [sym_preproc_call] = STATE(66),
    [sym_preproc_if] = STATE(66),
    [sym_preproc_ifdef] = STATE(66),
    [sym_function_definition] = STATE(66),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(66),
    [sym_type_definition] = STATE(66),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(66),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(66),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(66),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(66),
    [sym_preproc_undef] = STATE(66),
    [sym_preproc_linemarker] = STATE(66),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(66),
    [sym_class_declaration] = STATE(66),
    [sym_class_interface] = STATE(66),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(66),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(66),
    [sym_compatibility_alias_declaration] = STATE(66),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(66),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1261),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(65)] = {
    [sym__block_item] = STATE(67),
    [sym_preproc_include] = STATE(67),
    [sym_preproc_def] = STATE(67),
    [sym_preproc_function_def] = STATE(67),
    [sym_preproc_call] = STATE(67),
    [sym_preproc_if] = STATE(67),
    [sym_preproc_ifdef] = STATE(67),
    [sym_function_definition] = STATE(67),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(67),
    [sym_type_definition] = STATE(67),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(67),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(67),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(67),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(67),
    [sym_preproc_undef] = STATE(67),
    [sym_preproc_linemarker] = STATE(67),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(67),
    [sym_class_declaration] = STATE(67),
    [sym_class_interface] = STATE(67),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(67),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(67),
    [sym_compatibility_alias_declaration] = STATE(67),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(67),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1263),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(66)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1265),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(67)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1267),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(68)] = {
    [sym__block_item] = STATE(69),
    [sym_preproc_include] = STATE(69),
    [sym_preproc_def] = STATE(69),
    [sym_preproc_function_def] = STATE(69),
    [sym_preproc_call] = STATE(69),
    [sym_preproc_if] = STATE(69),
    [sym_preproc_ifdef] = STATE(69),
    [sym_function_definition] = STATE(69),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(69),
    [sym_type_definition] = STATE(69),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(69),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(69),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(69),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(69),
    [sym_preproc_undef] = STATE(69),
    [sym_preproc_linemarker] = STATE(69),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(69),
    [sym_class_declaration] = STATE(69),
    [sym_class_interface] = STATE(69),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(69),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(69),
    [sym_compatibility_alias_declaration] = STATE(69),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(69),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1269),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(69)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1271),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(70)] = {
    [sym__block_item] = STATE(71),
    [sym_preproc_include] = STATE(71),
    [sym_preproc_def] = STATE(71),
    [sym_preproc_function_def] = STATE(71),
    [sym_preproc_call] = STATE(71),
    [sym_preproc_if] = STATE(71),
    [sym_preproc_ifdef] = STATE(71),
    [sym_function_definition] = STATE(71),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(71),
    [sym_type_definition] = STATE(71),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(71),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(71),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(71),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(71),
    [sym_preproc_undef] = STATE(71),
    [sym_preproc_linemarker] = STATE(71),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(71),
    [sym_class_declaration] = STATE(71),
    [sym_class_interface] = STATE(71),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(71),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(71),
    [sym_compatibility_alias_declaration] = STATE(71),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(71),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1271),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(71)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1273),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(72)] = {
    [sym__block_item] = STATE(73),
    [sym_preproc_include] = STATE(73),
    [sym_preproc_def] = STATE(73),
    [sym_preproc_function_def] = STATE(73),
    [sym_preproc_call] = STATE(73),
    [sym_preproc_if] = STATE(73),
    [sym_preproc_ifdef] = STATE(73),
    [sym_function_definition] = STATE(73),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(73),
    [sym_type_definition] = STATE(73),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(73),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(73),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(73),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(73),
    [sym_preproc_undef] = STATE(73),
    [sym_preproc_linemarker] = STATE(73),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(73),
    [sym_class_declaration] = STATE(73),
    [sym_class_interface] = STATE(73),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(73),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(73),
    [sym_compatibility_alias_declaration] = STATE(73),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(73),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1275),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(73)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1277),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(74)] = {
    [sym__block_item] = STATE(75),
    [sym_preproc_include] = STATE(75),
    [sym_preproc_def] = STATE(75),
    [sym_preproc_function_def] = STATE(75),
    [sym_preproc_call] = STATE(75),
    [sym_preproc_if] = STATE(75),
    [sym_preproc_ifdef] = STATE(75),
    [sym_function_definition] = STATE(75),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(75),
    [sym_type_definition] = STATE(75),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(75),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(75),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(75),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(75),
    [sym_preproc_undef] = STATE(75),
    [sym_preproc_linemarker] = STATE(75),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(75),
    [sym_class_declaration] = STATE(75),
    [sym_class_interface] = STATE(75),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(75),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(75),
    [sym_compatibility_alias_declaration] = STATE(75),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(75),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1277),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(75)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1279),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(76)] = {
    [sym__block_item] = STATE(127),
    [sym_preproc_include] = STATE(127),
    [sym_preproc_def] = STATE(127),
    [sym_preproc_function_def] = STATE(127),
    [sym_preproc_call] = STATE(127),
    [sym_preproc_if] = STATE(127),
    [sym_preproc_ifdef] = STATE(127),
    [sym_function_definition] = STATE(127),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(127),
    [sym_type_definition] = STATE(127),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(127),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(127),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(127),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(127),
    [sym_preproc_undef] = STATE(127),
    [sym_preproc_linemarker] = STATE(127),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(127),
    [sym_class_declaration] = STATE(127),
    [sym_class_interface] = STATE(127),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(127),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(127),
    [sym_compatibility_alias_declaration] = STATE(127),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(127),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1281),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(77)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1283),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(78)] = {
    [sym__block_item] = STATE(80),
    [sym_preproc_include] = STATE(80),
    [sym_preproc_def] = STATE(80),
    [sym_preproc_function_def] = STATE(80),
    [sym_preproc_call] = STATE(80),
    [sym_preproc_if] = STATE(80),
    [sym_preproc_ifdef] = STATE(80),
    [sym_function_definition] = STATE(80),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(80),
    [sym_type_definition] = STATE(80),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(80),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(80),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(80),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(80),
    [sym_preproc_undef] = STATE(80),
    [sym_preproc_linemarker] = STATE(80),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(80),
    [sym_class_declaration] = STATE(80),
    [sym_class_interface] = STATE(80),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(80),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(80),
    [sym_compatibility_alias_declaration] = STATE(80),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(80),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1285),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(79)] = {
    [sym__block_item] = STATE(85),
    [sym_preproc_include] = STATE(85),
    [sym_preproc_def] = STATE(85),
    [sym_preproc_function_def] = STATE(85),
    [sym_preproc_call] = STATE(85),
    [sym_preproc_if] = STATE(85),
    [sym_preproc_ifdef] = STATE(85),
    [sym_function_definition] = STATE(85),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(85),
    [sym_type_definition] = STATE(85),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(85),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(85),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(85),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(85),
    [sym_preproc_undef] = STATE(85),
    [sym_preproc_linemarker] = STATE(85),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(85),
    [sym_class_declaration] = STATE(85),
    [sym_class_interface] = STATE(85),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(85),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(85),
    [sym_compatibility_alias_declaration] = STATE(85),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(85),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1287),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(80)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1289),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(81)] = {
    [sym__block_item] = STATE(83),
    [sym_preproc_include] = STATE(83),
    [sym_preproc_def] = STATE(83),
    [sym_preproc_function_def] = STATE(83),
    [sym_preproc_call] = STATE(83),
    [sym_preproc_if] = STATE(83),
    [sym_preproc_ifdef] = STATE(83),
    [sym_function_definition] = STATE(83),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(83),
    [sym_type_definition] = STATE(83),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(83),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(83),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(83),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(83),
    [sym_preproc_undef] = STATE(83),
    [sym_preproc_linemarker] = STATE(83),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(83),
    [sym_class_declaration] = STATE(83),
    [sym_class_interface] = STATE(83),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(83),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(83),
    [sym_compatibility_alias_declaration] = STATE(83),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(83),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1291),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(82)] = {
    [sym__block_item] = STATE(77),
    [sym_preproc_include] = STATE(77),
    [sym_preproc_def] = STATE(77),
    [sym_preproc_function_def] = STATE(77),
    [sym_preproc_call] = STATE(77),
    [sym_preproc_if] = STATE(77),
    [sym_preproc_ifdef] = STATE(77),
    [sym_function_definition] = STATE(77),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(77),
    [sym_type_definition] = STATE(77),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(77),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(77),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(77),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(77),
    [sym_preproc_undef] = STATE(77),
    [sym_preproc_linemarker] = STATE(77),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(77),
    [sym_class_declaration] = STATE(77),
    [sym_class_interface] = STATE(77),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(77),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(77),
    [sym_compatibility_alias_declaration] = STATE(77),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(77),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1293),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(83)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1295),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(84)] = {
    [sym__block_item] = STATE(86),
    [sym_preproc_include] = STATE(86),
    [sym_preproc_def] = STATE(86),
    [sym_preproc_function_def] = STATE(86),
    [sym_preproc_call] = STATE(86),
    [sym_preproc_if] = STATE(86),
    [sym_preproc_ifdef] = STATE(86),
    [sym_function_definition] = STATE(86),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(86),
    [sym_type_definition] = STATE(86),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(86),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(86),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(86),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(86),
    [sym_preproc_undef] = STATE(86),
    [sym_preproc_linemarker] = STATE(86),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(86),
    [sym_class_declaration] = STATE(86),
    [sym_class_interface] = STATE(86),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(86),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(86),
    [sym_compatibility_alias_declaration] = STATE(86),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(86),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1295),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(85)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1297),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(86)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1299),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(87)] = {
    [sym__block_item] = STATE(88),
    [sym_preproc_include] = STATE(88),
    [sym_preproc_def] = STATE(88),
    [sym_preproc_function_def] = STATE(88),
    [sym_preproc_call] = STATE(88),
    [sym_preproc_if] = STATE(88),
    [sym_preproc_ifdef] = STATE(88),
    [sym_function_definition] = STATE(88),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(88),
    [sym_type_definition] = STATE(88),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(88),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(88),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(88),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(88),
    [sym_preproc_undef] = STATE(88),
    [sym_preproc_linemarker] = STATE(88),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(88),
    [sym_class_declaration] = STATE(88),
    [sym_class_interface] = STATE(88),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(88),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(88),
    [sym_compatibility_alias_declaration] = STATE(88),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(88),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1301),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(88)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(89)] = {
    [sym__block_item] = STATE(90),
    [sym_preproc_include] = STATE(90),
    [sym_preproc_def] = STATE(90),
    [sym_preproc_function_def] = STATE(90),
    [sym_preproc_call] = STATE(90),
    [sym_preproc_if] = STATE(90),
    [sym_preproc_ifdef] = STATE(90),
    [sym_function_definition] = STATE(90),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(90),
    [sym_type_definition] = STATE(90),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(90),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(90),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(90),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(90),
    [sym_preproc_undef] = STATE(90),
    [sym_preproc_linemarker] = STATE(90),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(90),
    [sym_class_declaration] = STATE(90),
    [sym_class_interface] = STATE(90),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(90),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(90),
    [sym_compatibility_alias_declaration] = STATE(90),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(90),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(90)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1305),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(91)] = {
    [sym__block_item] = STATE(92),
    [sym_preproc_include] = STATE(92),
    [sym_preproc_def] = STATE(92),
    [sym_preproc_function_def] = STATE(92),
    [sym_preproc_call] = STATE(92),
    [sym_preproc_if] = STATE(92),
    [sym_preproc_ifdef] = STATE(92),
    [sym_function_definition] = STATE(92),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(92),
    [sym_type_definition] = STATE(92),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(92),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(92),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(92),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(92),
    [sym_preproc_undef] = STATE(92),
    [sym_preproc_linemarker] = STATE(92),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(92),
    [sym_class_declaration] = STATE(92),
    [sym_class_interface] = STATE(92),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(92),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(92),
    [sym_compatibility_alias_declaration] = STATE(92),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(92),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1307),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(92)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1309),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(93)] = {
    [sym__block_item] = STATE(94),
    [sym_preproc_include] = STATE(94),
    [sym_preproc_def] = STATE(94),
    [sym_preproc_function_def] = STATE(94),
    [sym_preproc_call] = STATE(94),
    [sym_preproc_if] = STATE(94),
    [sym_preproc_ifdef] = STATE(94),
    [sym_function_definition] = STATE(94),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(94),
    [sym_type_definition] = STATE(94),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(94),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(94),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(94),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(94),
    [sym_preproc_undef] = STATE(94),
    [sym_preproc_linemarker] = STATE(94),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(94),
    [sym_class_declaration] = STATE(94),
    [sym_class_interface] = STATE(94),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(94),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(94),
    [sym_compatibility_alias_declaration] = STATE(94),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(94),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1309),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(94)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1311),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(95)] = {
    [sym__block_item] = STATE(96),
    [sym_preproc_include] = STATE(96),
    [sym_preproc_def] = STATE(96),
    [sym_preproc_function_def] = STATE(96),
    [sym_preproc_call] = STATE(96),
    [sym_preproc_if] = STATE(96),
    [sym_preproc_ifdef] = STATE(96),
    [sym_function_definition] = STATE(96),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(96),
    [sym_type_definition] = STATE(96),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(96),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(96),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(96),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(96),
    [sym_preproc_undef] = STATE(96),
    [sym_preproc_linemarker] = STATE(96),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(96),
    [sym_class_declaration] = STATE(96),
    [sym_class_interface] = STATE(96),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(96),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(96),
    [sym_compatibility_alias_declaration] = STATE(96),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(96),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1313),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(96)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1315),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(97)] = {
    [sym__block_item] = STATE(98),
    [sym_preproc_include] = STATE(98),
    [sym_preproc_def] = STATE(98),
    [sym_preproc_function_def] = STATE(98),
    [sym_preproc_call] = STATE(98),
    [sym_preproc_if] = STATE(98),
    [sym_preproc_ifdef] = STATE(98),
    [sym_function_definition] = STATE(98),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(98),
    [sym_type_definition] = STATE(98),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(98),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(98),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(98),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(98),
    [sym_preproc_undef] = STATE(98),
    [sym_preproc_linemarker] = STATE(98),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(98),
    [sym_class_declaration] = STATE(98),
    [sym_class_interface] = STATE(98),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(98),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(98),
    [sym_compatibility_alias_declaration] = STATE(98),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(98),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1315),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(98)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1317),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(99)] = {
    [sym__block_item] = STATE(100),
    [sym_preproc_include] = STATE(100),
    [sym_preproc_def] = STATE(100),
    [sym_preproc_function_def] = STATE(100),
    [sym_preproc_call] = STATE(100),
    [sym_preproc_if] = STATE(100),
    [sym_preproc_ifdef] = STATE(100),
    [sym_function_definition] = STATE(100),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(100),
    [sym_type_definition] = STATE(100),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(100),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(100),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(100),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(100),
    [sym_preproc_undef] = STATE(100),
    [sym_preproc_linemarker] = STATE(100),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(100),
    [sym_class_declaration] = STATE(100),
    [sym_class_interface] = STATE(100),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(100),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(100),
    [sym_compatibility_alias_declaration] = STATE(100),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(100),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1319),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(100)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1321),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(101)] = {
    [sym__block_item] = STATE(102),
    [sym_preproc_include] = STATE(102),
    [sym_preproc_def] = STATE(102),
    [sym_preproc_function_def] = STATE(102),
    [sym_preproc_call] = STATE(102),
    [sym_preproc_if] = STATE(102),
    [sym_preproc_ifdef] = STATE(102),
    [sym_function_definition] = STATE(102),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(102),
    [sym_type_definition] = STATE(102),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(102),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(102),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(102),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(102),
    [sym_preproc_undef] = STATE(102),
    [sym_preproc_linemarker] = STATE(102),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(102),
    [sym_class_declaration] = STATE(102),
    [sym_class_interface] = STATE(102),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(102),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(102),
    [sym_compatibility_alias_declaration] = STATE(102),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(102),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1321),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(102)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1323),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(103)] = {
    [sym__block_item] = STATE(104),
    [sym_preproc_include] = STATE(104),
    [sym_preproc_def] = STATE(104),
    [sym_preproc_function_def] = STATE(104),
    [sym_preproc_call] = STATE(104),
    [sym_preproc_if] = STATE(104),
    [sym_preproc_ifdef] = STATE(104),
    [sym_function_definition] = STATE(104),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(104),
    [sym_type_definition] = STATE(104),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(104),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(104),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(104),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(104),
    [sym_preproc_undef] = STATE(104),
    [sym_preproc_linemarker] = STATE(104),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(104),
    [sym_class_declaration] = STATE(104),
    [sym_class_interface] = STATE(104),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(104),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(104),
    [sym_compatibility_alias_declaration] = STATE(104),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(104),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1325),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(104)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1327),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(105)] = {
    [sym__block_item] = STATE(106),
    [sym_preproc_include] = STATE(106),
    [sym_preproc_def] = STATE(106),
    [sym_preproc_function_def] = STATE(106),
    [sym_preproc_call] = STATE(106),
    [sym_preproc_if] = STATE(106),
    [sym_preproc_ifdef] = STATE(106),
    [sym_function_definition] = STATE(106),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(106),
    [sym_type_definition] = STATE(106),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(106),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(106),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(106),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(106),
    [sym_preproc_undef] = STATE(106),
    [sym_preproc_linemarker] = STATE(106),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(106),
    [sym_class_declaration] = STATE(106),
    [sym_class_interface] = STATE(106),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(106),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(106),
    [sym_compatibility_alias_declaration] = STATE(106),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(106),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1327),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(106)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1329),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(107)] = {
    [sym__block_item] = STATE(108),
    [sym_preproc_include] = STATE(108),
    [sym_preproc_def] = STATE(108),
    [sym_preproc_function_def] = STATE(108),
    [sym_preproc_call] = STATE(108),
    [sym_preproc_if] = STATE(108),
    [sym_preproc_ifdef] = STATE(108),
    [sym_function_definition] = STATE(108),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(108),
    [sym_type_definition] = STATE(108),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(108),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(108),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(108),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(108),
    [sym_preproc_undef] = STATE(108),
    [sym_preproc_linemarker] = STATE(108),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(108),
    [sym_class_declaration] = STATE(108),
    [sym_class_interface] = STATE(108),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(108),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(108),
    [sym_compatibility_alias_declaration] = STATE(108),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(108),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1331),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(108)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1333),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(109)] = {
    [sym__block_item] = STATE(110),
    [sym_preproc_include] = STATE(110),
    [sym_preproc_def] = STATE(110),
    [sym_preproc_function_def] = STATE(110),
    [sym_preproc_call] = STATE(110),
    [sym_preproc_if] = STATE(110),
    [sym_preproc_ifdef] = STATE(110),
    [sym_function_definition] = STATE(110),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(110),
    [sym_type_definition] = STATE(110),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(110),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(110),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(110),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(110),
    [sym_preproc_undef] = STATE(110),
    [sym_preproc_linemarker] = STATE(110),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(110),
    [sym_class_declaration] = STATE(110),
    [sym_class_interface] = STATE(110),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(110),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(110),
    [sym_compatibility_alias_declaration] = STATE(110),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(110),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1333),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(110)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1335),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(111)] = {
    [sym__block_item] = STATE(112),
    [sym_preproc_include] = STATE(112),
    [sym_preproc_def] = STATE(112),
    [sym_preproc_function_def] = STATE(112),
    [sym_preproc_call] = STATE(112),
    [sym_preproc_if] = STATE(112),
    [sym_preproc_ifdef] = STATE(112),
    [sym_function_definition] = STATE(112),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(112),
    [sym_type_definition] = STATE(112),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(112),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(112),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(112),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(112),
    [sym_preproc_undef] = STATE(112),
    [sym_preproc_linemarker] = STATE(112),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(112),
    [sym_class_declaration] = STATE(112),
    [sym_class_interface] = STATE(112),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(112),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(112),
    [sym_compatibility_alias_declaration] = STATE(112),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(112),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1337),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(112)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1339),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(113)] = {
    [sym__block_item] = STATE(114),
    [sym_preproc_include] = STATE(114),
    [sym_preproc_def] = STATE(114),
    [sym_preproc_function_def] = STATE(114),
    [sym_preproc_call] = STATE(114),
    [sym_preproc_if] = STATE(114),
    [sym_preproc_ifdef] = STATE(114),
    [sym_function_definition] = STATE(114),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(114),
    [sym_type_definition] = STATE(114),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(114),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(114),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(114),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(114),
    [sym_preproc_undef] = STATE(114),
    [sym_preproc_linemarker] = STATE(114),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(114),
    [sym_class_declaration] = STATE(114),
    [sym_class_interface] = STATE(114),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(114),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(114),
    [sym_compatibility_alias_declaration] = STATE(114),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(114),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1339),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(114)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1341),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(115)] = {
    [sym__block_item] = STATE(116),
    [sym_preproc_include] = STATE(116),
    [sym_preproc_def] = STATE(116),
    [sym_preproc_function_def] = STATE(116),
    [sym_preproc_call] = STATE(116),
    [sym_preproc_if] = STATE(116),
    [sym_preproc_ifdef] = STATE(116),
    [sym_function_definition] = STATE(116),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(116),
    [sym_type_definition] = STATE(116),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(116),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(116),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(116),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(116),
    [sym_preproc_undef] = STATE(116),
    [sym_preproc_linemarker] = STATE(116),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(116),
    [sym_class_declaration] = STATE(116),
    [sym_class_interface] = STATE(116),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(116),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(116),
    [sym_compatibility_alias_declaration] = STATE(116),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(116),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1343),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(116)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(946),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(117)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1345),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(118)] = {
    [sym__block_item] = STATE(119),
    [sym_preproc_include] = STATE(119),
    [sym_preproc_def] = STATE(119),
    [sym_preproc_function_def] = STATE(119),
    [sym_preproc_call] = STATE(119),
    [sym_preproc_if] = STATE(119),
    [sym_preproc_ifdef] = STATE(119),
    [sym_function_definition] = STATE(119),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(119),
    [sym_type_definition] = STATE(119),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(119),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(119),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(119),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(119),
    [sym_preproc_undef] = STATE(119),
    [sym_preproc_linemarker] = STATE(119),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(119),
    [sym_class_declaration] = STATE(119),
    [sym_class_interface] = STATE(119),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(119),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(119),
    [sym_compatibility_alias_declaration] = STATE(119),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(119),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1347),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(119)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1349),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(120)] = {
    [sym__block_item] = STATE(121),
    [sym_preproc_include] = STATE(121),
    [sym_preproc_def] = STATE(121),
    [sym_preproc_function_def] = STATE(121),
    [sym_preproc_call] = STATE(121),
    [sym_preproc_if] = STATE(121),
    [sym_preproc_ifdef] = STATE(121),
    [sym_function_definition] = STATE(121),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(121),
    [sym_type_definition] = STATE(121),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(121),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(121),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(121),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(121),
    [sym_preproc_undef] = STATE(121),
    [sym_preproc_linemarker] = STATE(121),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(121),
    [sym_class_declaration] = STATE(121),
    [sym_class_interface] = STATE(121),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(121),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(121),
    [sym_compatibility_alias_declaration] = STATE(121),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(121),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1349),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(121)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1351),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(122)] = {
    [sym__block_item] = STATE(123),
    [sym_preproc_include] = STATE(123),
    [sym_preproc_def] = STATE(123),
    [sym_preproc_function_def] = STATE(123),
    [sym_preproc_call] = STATE(123),
    [sym_preproc_if] = STATE(123),
    [sym_preproc_ifdef] = STATE(123),
    [sym_function_definition] = STATE(123),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(123),
    [sym_type_definition] = STATE(123),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(123),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(123),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(123),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(123),
    [sym_preproc_undef] = STATE(123),
    [sym_preproc_linemarker] = STATE(123),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(123),
    [sym_class_declaration] = STATE(123),
    [sym_class_interface] = STATE(123),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(123),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(123),
    [sym_compatibility_alias_declaration] = STATE(123),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(123),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1353),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(123)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1355),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(124)] = {
    [sym__block_item] = STATE(125),
    [sym_preproc_include] = STATE(125),
    [sym_preproc_def] = STATE(125),
    [sym_preproc_function_def] = STATE(125),
    [sym_preproc_call] = STATE(125),
    [sym_preproc_if] = STATE(125),
    [sym_preproc_ifdef] = STATE(125),
    [sym_function_definition] = STATE(125),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(125),
    [sym_type_definition] = STATE(125),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(125),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(125),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(125),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(125),
    [sym_preproc_undef] = STATE(125),
    [sym_preproc_linemarker] = STATE(125),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(125),
    [sym_class_declaration] = STATE(125),
    [sym_class_interface] = STATE(125),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(125),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(125),
    [sym_compatibility_alias_declaration] = STATE(125),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(125),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1355),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(125)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1357),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(126)] = {
    [sym__block_item] = STATE(44),
    [sym_preproc_include] = STATE(44),
    [sym_preproc_def] = STATE(44),
    [sym_preproc_function_def] = STATE(44),
    [sym_preproc_call] = STATE(44),
    [sym_preproc_if] = STATE(44),
    [sym_preproc_ifdef] = STATE(44),
    [sym_function_definition] = STATE(44),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(44),
    [sym_type_definition] = STATE(44),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(44),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(44),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(44),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(44),
    [sym_preproc_undef] = STATE(44),
    [sym_preproc_linemarker] = STATE(44),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(44),
    [sym_class_declaration] = STATE(44),
    [sym_class_interface] = STATE(44),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(44),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(44),
    [sym_compatibility_alias_declaration] = STATE(44),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(44),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1297),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(127)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(906),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(2019),
    [sym__declaration_specifiers] = STATE(4186),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(2019),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_ms_call_modifier] = STATE(2160),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2019),
    [sym_type_qualifier] = STATE(2019),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2933),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(838),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(42),
    [sym_preproc_undef] = STATE(42),
    [sym_preproc_linemarker] = STATE(42),
    [sym_availability_attribute_specifier] = STATE(2019),
    [sym_protocol_forward_declaration] = STATE(42),
    [sym_class_declaration] = STATE(42),
    [sym_class_interface] = STATE(42),
    [sym__class_interface_header] = STATE(1173),
    [sym_class_implementation] = STATE(42),
    [sym__class_implementation_header] = STATE(1084),
    [sym_protocol_declaration] = STATE(42),
    [sym_compatibility_alias_declaration] = STATE(42),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2019),
    [aux_sym_declaration_list_repeat1] = STATE(42),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(924),
    [aux_sym_preproc_include_token1] = ACTIONS(926),
    [aux_sym_preproc_include_token2] = ACTIONS(926),
    [aux_sym_preproc_def_token1] = ACTIONS(928),
    [aux_sym_preproc_if_token1] = ACTIONS(930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(932),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(932),
    [sym_preproc_directive] = ACTIONS(934),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(938),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(942),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1285),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(976),
    [aux_sym_preproc_undef_token1] = ACTIONS(978),
    [anon_sym_POUND] = ACTIONS(980),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(982),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(984),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(128)] = {
    [sym__top_level_item] = STATE(128),
    [sym_preproc_include] = STATE(128),
    [sym_preproc_def] = STATE(128),
    [sym_preproc_function_def] = STATE(128),
    [sym_preproc_call] = STATE(128),
    [sym_preproc_if] = STATE(128),
    [sym_preproc_ifdef] = STATE(128),
    [sym_function_definition] = STATE(128),
    [sym__old_style_function_definition] = STATE(983),
    [sym_declaration] = STATE(128),
    [sym_type_definition] = STATE(128),
    [sym__declaration_modifiers] = STATE(2050),
    [sym__declaration_specifiers] = STATE(4180),
    [sym_linkage_specification] = STATE(128),
    [sym_attribute_specifier] = STATE(2050),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2836),
    [sym_ms_call_modifier] = STATE(2120),
    [sym_compound_statement] = STATE(128),
    [sym_storage_class_specifier] = STATE(2050),
    [sym_type_qualifier] = STATE(2050),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2998),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(978),
    [sym__top_level_statement] = STATE(128),
    [sym_labeled_statement] = STATE(128),
    [sym__top_level_expression_statement] = STATE(128),
    [sym_if_statement] = STATE(128),
    [sym_switch_statement] = STATE(128),
    [sym_case_statement] = STATE(128),
    [sym_while_statement] = STATE(128),
    [sym_do_statement] = STATE(128),
    [sym_for_statement] = STATE(128),
    [sym_return_statement] = STATE(128),
    [sym_break_statement] = STATE(128),
    [sym_continue_statement] = STATE(128),
    [sym_goto_statement] = STATE(128),
    [sym_expression] = STATE(4565),
    [sym__string] = STATE(4564),
    [sym_conditional_expression] = STATE(4564),
    [sym_assignment_expression] = STATE(4564),
    [sym_pointer_expression] = STATE(4360),
    [sym_unary_expression] = STATE(4564),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(4564),
    [sym_cast_expression] = STATE(4564),
    [sym_sizeof_expression] = STATE(4564),
    [sym_alignof_expression] = STATE(4564),
    [sym_offsetof_expression] = STATE(4564),
    [sym_generic_expression] = STATE(4564),
    [sym_subscript_expression] = STATE(4360),
    [sym_call_expression] = STATE(4360),
    [sym_gnu_asm_expression] = STATE(4564),
    [sym_extension_expression] = STATE(4564),
    [sym_field_expression] = STATE(4360),
    [sym_compound_literal_expression] = STATE(4564),
    [sym_parenthesized_expression] = STATE(4360),
    [sym_char_literal] = STATE(4564),
    [sym_concatenated_string] = STATE(4564),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(4564),
    [sym__empty_declaration] = STATE(128),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(128),
    [sym_preproc_undef] = STATE(128),
    [sym_preproc_linemarker] = STATE(128),
    [sym_availability_attribute_specifier] = STATE(2050),
    [sym_protocol_forward_declaration] = STATE(128),
    [sym_class_declaration] = STATE(128),
    [sym_class_interface] = STATE(128),
    [sym__class_interface_header] = STATE(1165),
    [sym_class_implementation] = STATE(128),
    [sym__class_implementation_header] = STATE(1088),
    [sym_protocol_declaration] = STATE(128),
    [sym_compatibility_alias_declaration] = STATE(128),
    [sym_selector_expression] = STATE(4564),
    [sym_available_expression] = STATE(4564),
    [sym_range_expression] = STATE(4564),
    [sym_block_literal] = STATE(4564),
    [sym_message_expression] = STATE(4564),
    [sym_va_arg_expression] = STATE(4564),
    [sym_encode_expression] = STATE(4564),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(4564),
    [sym_dictionary_literal] = STATE(4564),
    [sym_array_literal] = STATE(4564),
    [aux_sym_translation_unit_repeat1] = STATE(128),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2050),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2451),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [ts_builtin_sym_end] = ACTIONS(1359),
    [sym_identifier] = ACTIONS(1361),
    [aux_sym_preproc_include_token1] = ACTIONS(1364),
    [aux_sym_preproc_include_token2] = ACTIONS(1364),
    [aux_sym_preproc_def_token1] = ACTIONS(1367),
    [aux_sym_preproc_if_token1] = ACTIONS(1370),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1373),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1373),
    [sym_preproc_directive] = ACTIONS(1376),
    [anon_sym_LPAREN2] = ACTIONS(1379),
    [anon_sym_BANG] = ACTIONS(1382),
    [anon_sym_TILDE] = ACTIONS(1382),
    [anon_sym_DASH] = ACTIONS(1385),
    [anon_sym_PLUS] = ACTIONS(1385),
    [anon_sym_STAR] = ACTIONS(1388),
    [anon_sym_CARET] = ACTIONS(1391),
    [anon_sym_AMP] = ACTIONS(1388),
    [anon_sym_SEMI] = ACTIONS(1394),
    [anon_sym___extension__] = ACTIONS(1397),
    [anon_sym_typedef] = ACTIONS(1400),
    [anon_sym_extern] = ACTIONS(1403),
    [anon_sym___attribute__] = ACTIONS(1406),
    [anon_sym___attribute] = ACTIONS(1406),
    [anon_sym_noreturn] = ACTIONS(1409),
    [anon_sym_LBRACK] = ACTIONS(1412),
    [anon_sym___declspec] = ACTIONS(1415),
    [anon_sym___cdecl] = ACTIONS(1418),
    [anon_sym___clrcall] = ACTIONS(1418),
    [anon_sym___stdcall] = ACTIONS(1418),
    [anon_sym___fastcall] = ACTIONS(1418),
    [anon_sym___thiscall] = ACTIONS(1418),
    [anon_sym___vectorcall] = ACTIONS(1418),
    [anon_sym_LBRACE] = ACTIONS(1421),
    [anon_sym_signed] = ACTIONS(1424),
    [anon_sym_unsigned] = ACTIONS(1424),
    [anon_sym_long] = ACTIONS(1424),
    [anon_sym_short] = ACTIONS(1424),
    [anon_sym_static] = ACTIONS(1427),
    [anon_sym_ATautoreleasepool] = ACTIONS(1430),
    [anon_sym_auto] = ACTIONS(1427),
    [anon_sym_register] = ACTIONS(1427),
    [anon_sym_inline] = ACTIONS(1427),
    [anon_sym___inline] = ACTIONS(1427),
    [anon_sym___inline__] = ACTIONS(1427),
    [anon_sym___forceinline] = ACTIONS(1427),
    [anon_sym_thread_local] = ACTIONS(1427),
    [anon_sym___thread] = ACTIONS(1427),
    [anon_sym_CG_EXTERN] = ACTIONS(1427),
    [anon_sym_CG_INLINE] = ACTIONS(1427),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(1427),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(1427),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(1427),
    [anon_sym_IBOutlet] = ACTIONS(1427),
    [anon_sym_IBInspectable] = ACTIONS(1427),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(1427),
    [anon_sym_NS_INLINE] = ACTIONS(1427),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(1427),
    [anon_sym_OBJC_EXPORT] = ACTIONS(1427),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(1427),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(1427),
    [anon_sym_const] = ACTIONS(1433),
    [anon_sym_constexpr] = ACTIONS(1409),
    [anon_sym_volatile] = ACTIONS(1409),
    [anon_sym_restrict] = ACTIONS(1409),
    [anon_sym___restrict__] = ACTIONS(1409),
    [anon_sym__Atomic] = ACTIONS(1409),
    [anon_sym__Noreturn] = ACTIONS(1409),
    [anon_sym__Nonnull] = ACTIONS(1409),
    [anon_sym_nullable] = ACTIONS(1409),
    [anon_sym__Complex] = ACTIONS(1409),
    [anon_sym__Nullable] = ACTIONS(1409),
    [anon_sym__Nullable_result] = ACTIONS(1409),
    [anon_sym__Null_unspecified] = ACTIONS(1409),
    [anon_sym___autoreleasing] = ACTIONS(1409),
    [anon_sym___block] = ACTIONS(1409),
    [anon_sym___bridge] = ACTIONS(1409),
    [anon_sym___bridge_retained] = ACTIONS(1409),
    [anon_sym___bridge_transfer] = ACTIONS(1409),
    [anon_sym___complex] = ACTIONS(1409),
    [anon_sym___const] = ACTIONS(1409),
    [anon_sym___imag] = ACTIONS(1436),
    [anon_sym___kindof] = ACTIONS(1409),
    [anon_sym___nonnull] = ACTIONS(1409),
    [anon_sym___nullable] = ACTIONS(1409),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(1409),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(1409),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(1409),
    [anon_sym___real] = ACTIONS(1436),
    [anon_sym___strong] = ACTIONS(1409),
    [anon_sym___unsafe_unretained] = ACTIONS(1409),
    [anon_sym___unused] = ACTIONS(1409),
    [anon_sym___weak] = ACTIONS(1409),
    [anon_sym_alignas] = ACTIONS(1439),
    [anon_sym__Alignas] = ACTIONS(1439),
    [sym_primitive_type] = ACTIONS(1442),
    [anon_sym_enum] = ACTIONS(1445),
    [anon_sym_struct] = ACTIONS(1448),
    [anon_sym_union] = ACTIONS(1451),
    [anon_sym_if] = ACTIONS(1454),
    [anon_sym_switch] = ACTIONS(1457),
    [anon_sym_case] = ACTIONS(1460),
    [anon_sym_default] = ACTIONS(1463),
    [anon_sym_while] = ACTIONS(1466),
    [anon_sym_do] = ACTIONS(1469),
    [anon_sym_for] = ACTIONS(1472),
    [anon_sym_in] = ACTIONS(1475),
    [anon_sym_return] = ACTIONS(1478),
    [anon_sym_break] = ACTIONS(1481),
    [anon_sym_continue] = ACTIONS(1484),
    [anon_sym_goto] = ACTIONS(1487),
    [anon_sym_DASH_DASH] = ACTIONS(1490),
    [anon_sym_PLUS_PLUS] = ACTIONS(1490),
    [anon_sym_sizeof] = ACTIONS(1493),
    [anon_sym___alignof__] = ACTIONS(1496),
    [anon_sym___alignof] = ACTIONS(1496),
    [anon_sym__alignof] = ACTIONS(1496),
    [anon_sym_alignof] = ACTIONS(1496),
    [anon_sym__Alignof] = ACTIONS(1496),
    [anon_sym_offsetof] = ACTIONS(1499),
    [anon_sym__Generic] = ACTIONS(1502),
    [anon_sym_asm] = ACTIONS(1505),
    [anon_sym___asm__] = ACTIONS(1505),
    [anon_sym___asm] = ACTIONS(1505),
    [sym_number_literal] = ACTIONS(1508),
    [anon_sym_L_SQUOTE] = ACTIONS(1511),
    [anon_sym_u_SQUOTE] = ACTIONS(1511),
    [anon_sym_U_SQUOTE] = ACTIONS(1511),
    [anon_sym_u8_SQUOTE] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1511),
    [anon_sym_AT] = ACTIONS(1514),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_L_DQUOTE] = ACTIONS(1517),
    [anon_sym_u_DQUOTE] = ACTIONS(1517),
    [anon_sym_U_DQUOTE] = ACTIONS(1517),
    [anon_sym_u8_DQUOTE] = ACTIONS(1517),
    [sym_true] = ACTIONS(1520),
    [sym_false] = ACTIONS(1520),
    [anon_sym_NULL] = ACTIONS(1523),
    [anon_sym_nullptr] = ACTIONS(1523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(1526),
    [anon_sym___typeof] = ACTIONS(1526),
    [anon_sym_typeof] = ACTIONS(1526),
    [anon_sym_ATimport] = ACTIONS(1529),
    [aux_sym_preproc_undef_token1] = ACTIONS(1532),
    [anon_sym_POUND] = ACTIONS(1535),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(1538),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(1538),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(1538),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(1538),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(1538),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(1538),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(1538),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(1538),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(1538),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(1541),
    [anon_sym_NS_AVAILABLE] = ACTIONS(1541),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(1541),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(1541),
    [anon_sym_API_AVAILABLE] = ACTIONS(1541),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(1541),
    [anon_sym_API_DEPRECATED] = ACTIONS(1541),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(1541),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(1541),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(1541),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(1541),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(1541),
    [anon_sym___deprecated_msg] = ACTIONS(1541),
    [anon_sym___deprecated_enum_msg] = ACTIONS(1541),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(1541),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(1541),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(1541),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(1541),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(1541),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(1541),
    [anon_sym_ATprotocol] = ACTIONS(1544),
    [anon_sym_ATinterface] = ACTIONS(1547),
    [anon_sym_ATimplementation] = ACTIONS(1550),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1553),
    [anon_sym_ATselector] = ACTIONS(1556),
    [anon_sym_ATavailable] = ACTIONS(1559),
    [anon_sym___builtin_available] = ACTIONS(1562),
    [anon_sym_va_arg] = ACTIONS(1565),
    [anon_sym_ATencode] = ACTIONS(1568),
    [anon_sym_BOOL] = ACTIONS(1571),
    [anon_sym_IMP] = ACTIONS(1571),
    [anon_sym_SEL] = ACTIONS(1571),
    [anon_sym_Class] = ACTIONS(1571),
    [anon_sym_id] = ACTIONS(1574),
  },
  [STATE(129)] = {
    [sym__top_level_item] = STATE(128),
    [sym_preproc_include] = STATE(128),
    [sym_preproc_def] = STATE(128),
    [sym_preproc_function_def] = STATE(128),
    [sym_preproc_call] = STATE(128),
    [sym_preproc_if] = STATE(128),
    [sym_preproc_ifdef] = STATE(128),
    [sym_function_definition] = STATE(128),
    [sym__old_style_function_definition] = STATE(983),
    [sym_declaration] = STATE(128),
    [sym_type_definition] = STATE(128),
    [sym__declaration_modifiers] = STATE(2050),
    [sym__declaration_specifiers] = STATE(4180),
    [sym_linkage_specification] = STATE(128),
    [sym_attribute_specifier] = STATE(2050),
    [sym_attribute_declaration] = STATE(1083),
    [sym_ms_declspec_modifier] = STATE(2836),
    [sym_ms_call_modifier] = STATE(2120),
    [sym_compound_statement] = STATE(128),
    [sym_storage_class_specifier] = STATE(2050),
    [sym_type_qualifier] = STATE(2050),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2998),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(978),
    [sym__top_level_statement] = STATE(128),
    [sym_labeled_statement] = STATE(128),
    [sym__top_level_expression_statement] = STATE(128),
    [sym_if_statement] = STATE(128),
    [sym_switch_statement] = STATE(128),
    [sym_case_statement] = STATE(128),
    [sym_while_statement] = STATE(128),
    [sym_do_statement] = STATE(128),
    [sym_for_statement] = STATE(128),
    [sym_return_statement] = STATE(128),
    [sym_break_statement] = STATE(128),
    [sym_continue_statement] = STATE(128),
    [sym_goto_statement] = STATE(128),
    [sym_expression] = STATE(4565),
    [sym__string] = STATE(4564),
    [sym_conditional_expression] = STATE(4564),
    [sym_assignment_expression] = STATE(4564),
    [sym_pointer_expression] = STATE(4360),
    [sym_unary_expression] = STATE(4564),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(4564),
    [sym_cast_expression] = STATE(4564),
    [sym_sizeof_expression] = STATE(4564),
    [sym_alignof_expression] = STATE(4564),
    [sym_offsetof_expression] = STATE(4564),
    [sym_generic_expression] = STATE(4564),
    [sym_subscript_expression] = STATE(4360),
    [sym_call_expression] = STATE(4360),
    [sym_gnu_asm_expression] = STATE(4564),
    [sym_extension_expression] = STATE(4564),
    [sym_field_expression] = STATE(4360),
    [sym_compound_literal_expression] = STATE(4564),
    [sym_parenthesized_expression] = STATE(4360),
    [sym_char_literal] = STATE(4564),
    [sym_concatenated_string] = STATE(4564),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(4564),
    [sym__empty_declaration] = STATE(128),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_module_import] = STATE(128),
    [sym_preproc_undef] = STATE(128),
    [sym_preproc_linemarker] = STATE(128),
    [sym_availability_attribute_specifier] = STATE(2050),
    [sym_protocol_forward_declaration] = STATE(128),
    [sym_class_declaration] = STATE(128),
    [sym_class_interface] = STATE(128),
    [sym__class_interface_header] = STATE(1165),
    [sym_class_implementation] = STATE(128),
    [sym__class_implementation_header] = STATE(1088),
    [sym_protocol_declaration] = STATE(128),
    [sym_compatibility_alias_declaration] = STATE(128),
    [sym_selector_expression] = STATE(4564),
    [sym_available_expression] = STATE(4564),
    [sym_range_expression] = STATE(4564),
    [sym_block_literal] = STATE(4564),
    [sym_message_expression] = STATE(4564),
    [sym_va_arg_expression] = STATE(4564),
    [sym_encode_expression] = STATE(4564),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(4564),
    [sym_dictionary_literal] = STATE(4564),
    [sym_array_literal] = STATE(4564),
    [aux_sym_translation_unit_repeat1] = STATE(128),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2050),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2451),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [ts_builtin_sym_end] = ACTIONS(1577),
    [sym_identifier] = ACTIONS(7),
    [aux_sym_preproc_include_token1] = ACTIONS(9),
    [aux_sym_preproc_include_token2] = ACTIONS(9),
    [aux_sym_preproc_def_token1] = ACTIONS(11),
    [aux_sym_preproc_if_token1] = ACTIONS(13),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(15),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(15),
    [sym_preproc_directive] = ACTIONS(17),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(29),
    [anon_sym___extension__] = ACTIONS(31),
    [anon_sym_typedef] = ACTIONS(33),
    [anon_sym_extern] = ACTIONS(35),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(41),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(55),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(57),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(57),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(65),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(69),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(77),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(83),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(105),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(109),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(113),
    [sym_false] = ACTIONS(113),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(119),
    [aux_sym_preproc_undef_token1] = ACTIONS(121),
    [anon_sym_POUND] = ACTIONS(123),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(129),
    [anon_sym_ATinterface] = ACTIONS(131),
    [anon_sym_ATimplementation] = ACTIONS(133),
    [anon_sym_ATcompatibility_alias] = ACTIONS(135),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(149),
  },
  [STATE(130)] = {
    [sym_declaration] = STATE(134),
    [sym_type_definition] = STATE(134),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4231),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_compound_statement] = STATE(134),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(134),
    [sym_labeled_statement] = STATE(134),
    [sym_expression_statement] = STATE(134),
    [sym_if_statement] = STATE(134),
    [sym_switch_statement] = STATE(134),
    [sym_while_statement] = STATE(134),
    [sym_do_statement] = STATE(134),
    [sym_for_statement] = STATE(134),
    [sym_return_statement] = STATE(134),
    [sym_break_statement] = STATE(134),
    [sym_continue_statement] = STATE(134),
    [sym_goto_statement] = STATE(134),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(134),
    [sym_throw_statement] = STATE(134),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(134),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(134),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(134),
    [sym_identifier] = ACTIONS(1579),
    [aux_sym_preproc_include_token1] = ACTIONS(1581),
    [aux_sym_preproc_include_token2] = ACTIONS(1581),
    [aux_sym_preproc_def_token1] = ACTIONS(1581),
    [aux_sym_preproc_if_token1] = ACTIONS(1581),
    [aux_sym_preproc_if_token2] = ACTIONS(1581),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1581),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1581),
    [aux_sym_preproc_else_token1] = ACTIONS(1581),
    [aux_sym_preproc_elif_token1] = ACTIONS(1581),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1581),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1581),
    [sym_preproc_directive] = ACTIONS(1581),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(1583),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1581),
    [anon_sym___clrcall] = ACTIONS(1581),
    [anon_sym___stdcall] = ACTIONS(1581),
    [anon_sym___fastcall] = ACTIONS(1581),
    [anon_sym___thiscall] = ACTIONS(1581),
    [anon_sym___vectorcall] = ACTIONS(1581),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_else] = ACTIONS(1581),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(1581),
    [anon_sym_default] = ACTIONS(1581),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1591),
    [aux_sym_preproc_undef_token1] = ACTIONS(1581),
    [anon_sym_POUND] = ACTIONS(1581),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1591),
    [anon_sym_ATinterface] = ACTIONS(1591),
    [anon_sym_ATimplementation] = ACTIONS(1591),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1591),
    [anon_sym_ATsynthesize] = ACTIONS(1591),
    [anon_sym_ATdynamic] = ACTIONS(1591),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(131)] = {
    [sym_declaration] = STATE(133),
    [sym_type_definition] = STATE(133),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4231),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_compound_statement] = STATE(133),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(133),
    [sym_labeled_statement] = STATE(133),
    [sym_expression_statement] = STATE(133),
    [sym_if_statement] = STATE(133),
    [sym_switch_statement] = STATE(133),
    [sym_while_statement] = STATE(133),
    [sym_do_statement] = STATE(133),
    [sym_for_statement] = STATE(133),
    [sym_return_statement] = STATE(133),
    [sym_break_statement] = STATE(133),
    [sym_continue_statement] = STATE(133),
    [sym_goto_statement] = STATE(133),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(133),
    [sym_throw_statement] = STATE(133),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(133),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(133),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(133),
    [sym_identifier] = ACTIONS(1579),
    [aux_sym_preproc_include_token1] = ACTIONS(1595),
    [aux_sym_preproc_include_token2] = ACTIONS(1595),
    [aux_sym_preproc_def_token1] = ACTIONS(1595),
    [aux_sym_preproc_if_token1] = ACTIONS(1595),
    [aux_sym_preproc_if_token2] = ACTIONS(1595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1595),
    [aux_sym_preproc_else_token1] = ACTIONS(1595),
    [aux_sym_preproc_elif_token1] = ACTIONS(1595),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1595),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1595),
    [sym_preproc_directive] = ACTIONS(1595),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(1583),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1595),
    [anon_sym___clrcall] = ACTIONS(1595),
    [anon_sym___stdcall] = ACTIONS(1595),
    [anon_sym___fastcall] = ACTIONS(1595),
    [anon_sym___thiscall] = ACTIONS(1595),
    [anon_sym___vectorcall] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_else] = ACTIONS(1595),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(1595),
    [anon_sym_default] = ACTIONS(1595),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1597),
    [aux_sym_preproc_undef_token1] = ACTIONS(1595),
    [anon_sym_POUND] = ACTIONS(1595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1597),
    [anon_sym_ATinterface] = ACTIONS(1597),
    [anon_sym_ATimplementation] = ACTIONS(1597),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1597),
    [anon_sym_ATsynthesize] = ACTIONS(1597),
    [anon_sym_ATdynamic] = ACTIONS(1597),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(132)] = {
    [sym_declaration] = STATE(130),
    [sym_type_definition] = STATE(130),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4231),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_compound_statement] = STATE(130),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(130),
    [sym_labeled_statement] = STATE(130),
    [sym_expression_statement] = STATE(130),
    [sym_if_statement] = STATE(130),
    [sym_switch_statement] = STATE(130),
    [sym_while_statement] = STATE(130),
    [sym_do_statement] = STATE(130),
    [sym_for_statement] = STATE(130),
    [sym_return_statement] = STATE(130),
    [sym_break_statement] = STATE(130),
    [sym_continue_statement] = STATE(130),
    [sym_goto_statement] = STATE(130),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(130),
    [sym_throw_statement] = STATE(130),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(130),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(130),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(130),
    [sym_identifier] = ACTIONS(1579),
    [aux_sym_preproc_include_token1] = ACTIONS(1599),
    [aux_sym_preproc_include_token2] = ACTIONS(1599),
    [aux_sym_preproc_def_token1] = ACTIONS(1599),
    [aux_sym_preproc_if_token1] = ACTIONS(1599),
    [aux_sym_preproc_if_token2] = ACTIONS(1599),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1599),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1599),
    [aux_sym_preproc_else_token1] = ACTIONS(1599),
    [aux_sym_preproc_elif_token1] = ACTIONS(1599),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1599),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1599),
    [sym_preproc_directive] = ACTIONS(1599),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(1583),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1599),
    [anon_sym___clrcall] = ACTIONS(1599),
    [anon_sym___stdcall] = ACTIONS(1599),
    [anon_sym___fastcall] = ACTIONS(1599),
    [anon_sym___thiscall] = ACTIONS(1599),
    [anon_sym___vectorcall] = ACTIONS(1599),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_else] = ACTIONS(1599),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(1599),
    [anon_sym_default] = ACTIONS(1599),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1601),
    [aux_sym_preproc_undef_token1] = ACTIONS(1599),
    [anon_sym_POUND] = ACTIONS(1599),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1601),
    [anon_sym_ATinterface] = ACTIONS(1601),
    [anon_sym_ATimplementation] = ACTIONS(1601),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1601),
    [anon_sym_ATsynthesize] = ACTIONS(1601),
    [anon_sym_ATdynamic] = ACTIONS(1601),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(133)] = {
    [sym_declaration] = STATE(134),
    [sym_type_definition] = STATE(134),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4231),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_compound_statement] = STATE(134),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(134),
    [sym_labeled_statement] = STATE(134),
    [sym_expression_statement] = STATE(134),
    [sym_if_statement] = STATE(134),
    [sym_switch_statement] = STATE(134),
    [sym_while_statement] = STATE(134),
    [sym_do_statement] = STATE(134),
    [sym_for_statement] = STATE(134),
    [sym_return_statement] = STATE(134),
    [sym_break_statement] = STATE(134),
    [sym_continue_statement] = STATE(134),
    [sym_goto_statement] = STATE(134),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(134),
    [sym_throw_statement] = STATE(134),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(134),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(134),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(134),
    [sym_identifier] = ACTIONS(1579),
    [aux_sym_preproc_include_token1] = ACTIONS(1603),
    [aux_sym_preproc_include_token2] = ACTIONS(1603),
    [aux_sym_preproc_def_token1] = ACTIONS(1603),
    [aux_sym_preproc_if_token1] = ACTIONS(1603),
    [aux_sym_preproc_if_token2] = ACTIONS(1603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1603),
    [aux_sym_preproc_else_token1] = ACTIONS(1603),
    [aux_sym_preproc_elif_token1] = ACTIONS(1603),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1603),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1603),
    [sym_preproc_directive] = ACTIONS(1603),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(1583),
    [anon_sym_typedef] = ACTIONS(175),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1603),
    [anon_sym___clrcall] = ACTIONS(1603),
    [anon_sym___stdcall] = ACTIONS(1603),
    [anon_sym___fastcall] = ACTIONS(1603),
    [anon_sym___thiscall] = ACTIONS(1603),
    [anon_sym___vectorcall] = ACTIONS(1603),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_else] = ACTIONS(1603),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(1603),
    [anon_sym_default] = ACTIONS(1603),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1605),
    [aux_sym_preproc_undef_token1] = ACTIONS(1603),
    [anon_sym_POUND] = ACTIONS(1603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1605),
    [anon_sym_ATinterface] = ACTIONS(1605),
    [anon_sym_ATimplementation] = ACTIONS(1605),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1605),
    [anon_sym_ATsynthesize] = ACTIONS(1605),
    [anon_sym_ATdynamic] = ACTIONS(1605),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(134)] = {
    [sym_declaration] = STATE(134),
    [sym_type_definition] = STATE(134),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4231),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2847),
    [sym_compound_statement] = STATE(134),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(134),
    [sym_labeled_statement] = STATE(134),
    [sym_expression_statement] = STATE(134),
    [sym_if_statement] = STATE(134),
    [sym_switch_statement] = STATE(134),
    [sym_while_statement] = STATE(134),
    [sym_do_statement] = STATE(134),
    [sym_for_statement] = STATE(134),
    [sym_return_statement] = STATE(134),
    [sym_break_statement] = STATE(134),
    [sym_continue_statement] = STATE(134),
    [sym_goto_statement] = STATE(134),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(134),
    [sym_throw_statement] = STATE(134),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(134),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(134),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(134),
    [sym_identifier] = ACTIONS(1607),
    [aux_sym_preproc_include_token1] = ACTIONS(1610),
    [aux_sym_preproc_include_token2] = ACTIONS(1610),
    [aux_sym_preproc_def_token1] = ACTIONS(1610),
    [aux_sym_preproc_if_token1] = ACTIONS(1610),
    [aux_sym_preproc_if_token2] = ACTIONS(1610),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1610),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1610),
    [aux_sym_preproc_else_token1] = ACTIONS(1610),
    [aux_sym_preproc_elif_token1] = ACTIONS(1610),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1610),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1610),
    [sym_preproc_directive] = ACTIONS(1610),
    [anon_sym_LPAREN2] = ACTIONS(1612),
    [anon_sym_BANG] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_DASH] = ACTIONS(1618),
    [anon_sym_PLUS] = ACTIONS(1618),
    [anon_sym_STAR] = ACTIONS(1621),
    [anon_sym_CARET] = ACTIONS(1624),
    [anon_sym_AMP] = ACTIONS(1621),
    [anon_sym_SEMI] = ACTIONS(1627),
    [anon_sym___extension__] = ACTIONS(1630),
    [anon_sym_typedef] = ACTIONS(1633),
    [anon_sym_extern] = ACTIONS(1636),
    [anon_sym___attribute__] = ACTIONS(1639),
    [anon_sym___attribute] = ACTIONS(1639),
    [anon_sym_noreturn] = ACTIONS(1642),
    [anon_sym_LBRACK] = ACTIONS(1645),
    [anon_sym___declspec] = ACTIONS(1648),
    [anon_sym___cdecl] = ACTIONS(1610),
    [anon_sym___clrcall] = ACTIONS(1610),
    [anon_sym___stdcall] = ACTIONS(1610),
    [anon_sym___fastcall] = ACTIONS(1610),
    [anon_sym___thiscall] = ACTIONS(1610),
    [anon_sym___vectorcall] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1651),
    [anon_sym_signed] = ACTIONS(1654),
    [anon_sym_unsigned] = ACTIONS(1654),
    [anon_sym_long] = ACTIONS(1654),
    [anon_sym_short] = ACTIONS(1654),
    [anon_sym_static] = ACTIONS(1636),
    [anon_sym_ATautoreleasepool] = ACTIONS(1657),
    [anon_sym_auto] = ACTIONS(1636),
    [anon_sym_register] = ACTIONS(1636),
    [anon_sym_inline] = ACTIONS(1636),
    [anon_sym___inline] = ACTIONS(1636),
    [anon_sym___inline__] = ACTIONS(1636),
    [anon_sym___forceinline] = ACTIONS(1636),
    [anon_sym_thread_local] = ACTIONS(1636),
    [anon_sym___thread] = ACTIONS(1636),
    [anon_sym_CG_EXTERN] = ACTIONS(1636),
    [anon_sym_CG_INLINE] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(1636),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(1636),
    [anon_sym_IBOutlet] = ACTIONS(1636),
    [anon_sym_IBInspectable] = ACTIONS(1636),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(1636),
    [anon_sym_NS_INLINE] = ACTIONS(1636),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(1636),
    [anon_sym_OBJC_EXPORT] = ACTIONS(1636),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(1636),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(1636),
    [anon_sym_const] = ACTIONS(1660),
    [anon_sym_constexpr] = ACTIONS(1642),
    [anon_sym_volatile] = ACTIONS(1642),
    [anon_sym_restrict] = ACTIONS(1642),
    [anon_sym___restrict__] = ACTIONS(1642),
    [anon_sym__Atomic] = ACTIONS(1642),
    [anon_sym__Noreturn] = ACTIONS(1642),
    [anon_sym__Nonnull] = ACTIONS(1642),
    [anon_sym_nullable] = ACTIONS(1642),
    [anon_sym__Complex] = ACTIONS(1642),
    [anon_sym__Nullable] = ACTIONS(1642),
    [anon_sym__Nullable_result] = ACTIONS(1642),
    [anon_sym__Null_unspecified] = ACTIONS(1642),
    [anon_sym___autoreleasing] = ACTIONS(1642),
    [anon_sym___block] = ACTIONS(1642),
    [anon_sym___bridge] = ACTIONS(1642),
    [anon_sym___bridge_retained] = ACTIONS(1642),
    [anon_sym___bridge_transfer] = ACTIONS(1642),
    [anon_sym___complex] = ACTIONS(1642),
    [anon_sym___const] = ACTIONS(1642),
    [anon_sym___imag] = ACTIONS(1663),
    [anon_sym___kindof] = ACTIONS(1642),
    [anon_sym___nonnull] = ACTIONS(1642),
    [anon_sym___nullable] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(1642),
    [anon_sym___real] = ACTIONS(1663),
    [anon_sym___strong] = ACTIONS(1642),
    [anon_sym___unsafe_unretained] = ACTIONS(1642),
    [anon_sym___unused] = ACTIONS(1642),
    [anon_sym___weak] = ACTIONS(1642),
    [anon_sym_alignas] = ACTIONS(1666),
    [anon_sym__Alignas] = ACTIONS(1666),
    [sym_primitive_type] = ACTIONS(1669),
    [anon_sym_enum] = ACTIONS(1672),
    [anon_sym_struct] = ACTIONS(1675),
    [anon_sym_union] = ACTIONS(1678),
    [anon_sym_if] = ACTIONS(1681),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_switch] = ACTIONS(1684),
    [anon_sym_case] = ACTIONS(1610),
    [anon_sym_default] = ACTIONS(1610),
    [anon_sym_while] = ACTIONS(1687),
    [anon_sym_do] = ACTIONS(1690),
    [anon_sym_for] = ACTIONS(1693),
    [anon_sym_in] = ACTIONS(1696),
    [anon_sym_return] = ACTIONS(1699),
    [anon_sym_break] = ACTIONS(1702),
    [anon_sym_continue] = ACTIONS(1705),
    [anon_sym_goto] = ACTIONS(1708),
    [anon_sym_DASH_DASH] = ACTIONS(1711),
    [anon_sym_PLUS_PLUS] = ACTIONS(1711),
    [anon_sym_sizeof] = ACTIONS(1714),
    [anon_sym___alignof__] = ACTIONS(1717),
    [anon_sym___alignof] = ACTIONS(1717),
    [anon_sym__alignof] = ACTIONS(1717),
    [anon_sym_alignof] = ACTIONS(1717),
    [anon_sym__Alignof] = ACTIONS(1717),
    [anon_sym_offsetof] = ACTIONS(1720),
    [anon_sym__Generic] = ACTIONS(1723),
    [anon_sym_asm] = ACTIONS(1726),
    [anon_sym___asm__] = ACTIONS(1726),
    [anon_sym___asm] = ACTIONS(1729),
    [sym_number_literal] = ACTIONS(1732),
    [anon_sym_L_SQUOTE] = ACTIONS(1735),
    [anon_sym_u_SQUOTE] = ACTIONS(1735),
    [anon_sym_U_SQUOTE] = ACTIONS(1735),
    [anon_sym_u8_SQUOTE] = ACTIONS(1735),
    [anon_sym_SQUOTE] = ACTIONS(1735),
    [anon_sym_AT] = ACTIONS(1738),
    [anon_sym_DQUOTE] = ACTIONS(1741),
    [anon_sym_L_DQUOTE] = ACTIONS(1741),
    [anon_sym_u_DQUOTE] = ACTIONS(1741),
    [anon_sym_U_DQUOTE] = ACTIONS(1741),
    [anon_sym_u8_DQUOTE] = ACTIONS(1741),
    [sym_true] = ACTIONS(1744),
    [sym_false] = ACTIONS(1744),
    [anon_sym_NULL] = ACTIONS(1747),
    [anon_sym_nullptr] = ACTIONS(1747),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(1750),
    [anon_sym___typeof] = ACTIONS(1750),
    [anon_sym_typeof] = ACTIONS(1750),
    [anon_sym_ATimport] = ACTIONS(1753),
    [aux_sym_preproc_undef_token1] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(1755),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(1755),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(1755),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE] = ACTIONS(1758),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_API_AVAILABLE] = ACTIONS(1758),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_API_DEPRECATED] = ACTIONS(1758),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(1758),
    [anon_sym___deprecated_msg] = ACTIONS(1758),
    [anon_sym___deprecated_enum_msg] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(1758),
    [anon_sym_ATprotocol] = ACTIONS(1753),
    [anon_sym_ATinterface] = ACTIONS(1753),
    [anon_sym_ATimplementation] = ACTIONS(1753),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1753),
    [anon_sym_ATsynthesize] = ACTIONS(1753),
    [anon_sym_ATdynamic] = ACTIONS(1753),
    [anon_sym_ATtry] = ACTIONS(1761),
    [anon_sym___try] = ACTIONS(1764),
    [anon_sym_ATthrow] = ACTIONS(1767),
    [anon_sym_ATselector] = ACTIONS(1770),
    [anon_sym_ATavailable] = ACTIONS(1773),
    [anon_sym___builtin_available] = ACTIONS(1776),
    [anon_sym_va_arg] = ACTIONS(1779),
    [anon_sym_ATencode] = ACTIONS(1782),
    [anon_sym_ATsynchronized] = ACTIONS(1785),
    [anon_sym_BOOL] = ACTIONS(1788),
    [anon_sym_IMP] = ACTIONS(1788),
    [anon_sym_SEL] = ACTIONS(1788),
    [anon_sym_Class] = ACTIONS(1788),
    [anon_sym_id] = ACTIONS(1791),
  },
  [STATE(135)] = {
    [sym_declaration] = STATE(139),
    [sym_type_definition] = STATE(139),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4202),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_compound_statement] = STATE(139),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(139),
    [sym_labeled_statement] = STATE(139),
    [sym_expression_statement] = STATE(139),
    [sym_if_statement] = STATE(139),
    [sym_switch_statement] = STATE(139),
    [sym_while_statement] = STATE(139),
    [sym_do_statement] = STATE(139),
    [sym_for_statement] = STATE(139),
    [sym_return_statement] = STATE(139),
    [sym_break_statement] = STATE(139),
    [sym_continue_statement] = STATE(139),
    [sym_goto_statement] = STATE(139),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(139),
    [sym_throw_statement] = STATE(139),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(139),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(139),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(139),
    [sym_identifier] = ACTIONS(1794),
    [aux_sym_preproc_include_token1] = ACTIONS(1599),
    [aux_sym_preproc_include_token2] = ACTIONS(1599),
    [aux_sym_preproc_def_token1] = ACTIONS(1599),
    [aux_sym_preproc_if_token1] = ACTIONS(1599),
    [aux_sym_preproc_if_token2] = ACTIONS(1599),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1599),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1599),
    [aux_sym_preproc_else_token1] = ACTIONS(1599),
    [aux_sym_preproc_elif_token1] = ACTIONS(1599),
    [sym_preproc_directive] = ACTIONS(1599),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(1796),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1599),
    [anon_sym___clrcall] = ACTIONS(1599),
    [anon_sym___stdcall] = ACTIONS(1599),
    [anon_sym___fastcall] = ACTIONS(1599),
    [anon_sym___thiscall] = ACTIONS(1599),
    [anon_sym___vectorcall] = ACTIONS(1599),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_else] = ACTIONS(1599),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(1599),
    [anon_sym_default] = ACTIONS(1599),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1601),
    [aux_sym_preproc_undef_token1] = ACTIONS(1599),
    [anon_sym_POUND] = ACTIONS(1599),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1601),
    [anon_sym_ATinterface] = ACTIONS(1601),
    [anon_sym_ATimplementation] = ACTIONS(1601),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1601),
    [anon_sym_ATsynthesize] = ACTIONS(1601),
    [anon_sym_ATdynamic] = ACTIONS(1601),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(136)] = {
    [sym_declaration] = STATE(136),
    [sym_type_definition] = STATE(136),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4202),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_compound_statement] = STATE(136),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(136),
    [sym_labeled_statement] = STATE(136),
    [sym_expression_statement] = STATE(136),
    [sym_if_statement] = STATE(136),
    [sym_switch_statement] = STATE(136),
    [sym_while_statement] = STATE(136),
    [sym_do_statement] = STATE(136),
    [sym_for_statement] = STATE(136),
    [sym_return_statement] = STATE(136),
    [sym_break_statement] = STATE(136),
    [sym_continue_statement] = STATE(136),
    [sym_goto_statement] = STATE(136),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(136),
    [sym_throw_statement] = STATE(136),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(136),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(136),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(136),
    [sym_identifier] = ACTIONS(1798),
    [aux_sym_preproc_include_token1] = ACTIONS(1610),
    [aux_sym_preproc_include_token2] = ACTIONS(1610),
    [aux_sym_preproc_def_token1] = ACTIONS(1610),
    [aux_sym_preproc_if_token1] = ACTIONS(1610),
    [aux_sym_preproc_if_token2] = ACTIONS(1610),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1610),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1610),
    [aux_sym_preproc_else_token1] = ACTIONS(1610),
    [aux_sym_preproc_elif_token1] = ACTIONS(1610),
    [sym_preproc_directive] = ACTIONS(1610),
    [anon_sym_LPAREN2] = ACTIONS(1612),
    [anon_sym_BANG] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_DASH] = ACTIONS(1618),
    [anon_sym_PLUS] = ACTIONS(1618),
    [anon_sym_STAR] = ACTIONS(1621),
    [anon_sym_CARET] = ACTIONS(1624),
    [anon_sym_AMP] = ACTIONS(1621),
    [anon_sym_SEMI] = ACTIONS(1801),
    [anon_sym___extension__] = ACTIONS(1804),
    [anon_sym_typedef] = ACTIONS(1807),
    [anon_sym_extern] = ACTIONS(1636),
    [anon_sym___attribute__] = ACTIONS(1639),
    [anon_sym___attribute] = ACTIONS(1639),
    [anon_sym_noreturn] = ACTIONS(1642),
    [anon_sym_LBRACK] = ACTIONS(1645),
    [anon_sym___declspec] = ACTIONS(1648),
    [anon_sym___cdecl] = ACTIONS(1610),
    [anon_sym___clrcall] = ACTIONS(1610),
    [anon_sym___stdcall] = ACTIONS(1610),
    [anon_sym___fastcall] = ACTIONS(1610),
    [anon_sym___thiscall] = ACTIONS(1610),
    [anon_sym___vectorcall] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1810),
    [anon_sym_signed] = ACTIONS(1654),
    [anon_sym_unsigned] = ACTIONS(1654),
    [anon_sym_long] = ACTIONS(1654),
    [anon_sym_short] = ACTIONS(1654),
    [anon_sym_static] = ACTIONS(1636),
    [anon_sym_ATautoreleasepool] = ACTIONS(1813),
    [anon_sym_auto] = ACTIONS(1636),
    [anon_sym_register] = ACTIONS(1636),
    [anon_sym_inline] = ACTIONS(1636),
    [anon_sym___inline] = ACTIONS(1636),
    [anon_sym___inline__] = ACTIONS(1636),
    [anon_sym___forceinline] = ACTIONS(1636),
    [anon_sym_thread_local] = ACTIONS(1636),
    [anon_sym___thread] = ACTIONS(1636),
    [anon_sym_CG_EXTERN] = ACTIONS(1636),
    [anon_sym_CG_INLINE] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(1636),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(1636),
    [anon_sym_IBOutlet] = ACTIONS(1636),
    [anon_sym_IBInspectable] = ACTIONS(1636),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(1636),
    [anon_sym_NS_INLINE] = ACTIONS(1636),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(1636),
    [anon_sym_OBJC_EXPORT] = ACTIONS(1636),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(1636),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(1636),
    [anon_sym_const] = ACTIONS(1660),
    [anon_sym_constexpr] = ACTIONS(1642),
    [anon_sym_volatile] = ACTIONS(1642),
    [anon_sym_restrict] = ACTIONS(1642),
    [anon_sym___restrict__] = ACTIONS(1642),
    [anon_sym__Atomic] = ACTIONS(1642),
    [anon_sym__Noreturn] = ACTIONS(1642),
    [anon_sym__Nonnull] = ACTIONS(1642),
    [anon_sym_nullable] = ACTIONS(1642),
    [anon_sym__Complex] = ACTIONS(1642),
    [anon_sym__Nullable] = ACTIONS(1642),
    [anon_sym__Nullable_result] = ACTIONS(1642),
    [anon_sym__Null_unspecified] = ACTIONS(1642),
    [anon_sym___autoreleasing] = ACTIONS(1642),
    [anon_sym___block] = ACTIONS(1642),
    [anon_sym___bridge] = ACTIONS(1642),
    [anon_sym___bridge_retained] = ACTIONS(1642),
    [anon_sym___bridge_transfer] = ACTIONS(1642),
    [anon_sym___complex] = ACTIONS(1642),
    [anon_sym___const] = ACTIONS(1642),
    [anon_sym___imag] = ACTIONS(1663),
    [anon_sym___kindof] = ACTIONS(1642),
    [anon_sym___nonnull] = ACTIONS(1642),
    [anon_sym___nullable] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(1642),
    [anon_sym___real] = ACTIONS(1663),
    [anon_sym___strong] = ACTIONS(1642),
    [anon_sym___unsafe_unretained] = ACTIONS(1642),
    [anon_sym___unused] = ACTIONS(1642),
    [anon_sym___weak] = ACTIONS(1642),
    [anon_sym_alignas] = ACTIONS(1666),
    [anon_sym__Alignas] = ACTIONS(1666),
    [sym_primitive_type] = ACTIONS(1669),
    [anon_sym_enum] = ACTIONS(1672),
    [anon_sym_struct] = ACTIONS(1675),
    [anon_sym_union] = ACTIONS(1678),
    [anon_sym_if] = ACTIONS(1816),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_switch] = ACTIONS(1819),
    [anon_sym_case] = ACTIONS(1610),
    [anon_sym_default] = ACTIONS(1610),
    [anon_sym_while] = ACTIONS(1822),
    [anon_sym_do] = ACTIONS(1825),
    [anon_sym_for] = ACTIONS(1828),
    [anon_sym_in] = ACTIONS(1696),
    [anon_sym_return] = ACTIONS(1831),
    [anon_sym_break] = ACTIONS(1834),
    [anon_sym_continue] = ACTIONS(1837),
    [anon_sym_goto] = ACTIONS(1840),
    [anon_sym_DASH_DASH] = ACTIONS(1711),
    [anon_sym_PLUS_PLUS] = ACTIONS(1711),
    [anon_sym_sizeof] = ACTIONS(1714),
    [anon_sym___alignof__] = ACTIONS(1717),
    [anon_sym___alignof] = ACTIONS(1717),
    [anon_sym__alignof] = ACTIONS(1717),
    [anon_sym_alignof] = ACTIONS(1717),
    [anon_sym__Alignof] = ACTIONS(1717),
    [anon_sym_offsetof] = ACTIONS(1720),
    [anon_sym__Generic] = ACTIONS(1723),
    [anon_sym_asm] = ACTIONS(1726),
    [anon_sym___asm__] = ACTIONS(1726),
    [anon_sym___asm] = ACTIONS(1843),
    [sym_number_literal] = ACTIONS(1732),
    [anon_sym_L_SQUOTE] = ACTIONS(1735),
    [anon_sym_u_SQUOTE] = ACTIONS(1735),
    [anon_sym_U_SQUOTE] = ACTIONS(1735),
    [anon_sym_u8_SQUOTE] = ACTIONS(1735),
    [anon_sym_SQUOTE] = ACTIONS(1735),
    [anon_sym_AT] = ACTIONS(1738),
    [anon_sym_DQUOTE] = ACTIONS(1741),
    [anon_sym_L_DQUOTE] = ACTIONS(1741),
    [anon_sym_u_DQUOTE] = ACTIONS(1741),
    [anon_sym_U_DQUOTE] = ACTIONS(1741),
    [anon_sym_u8_DQUOTE] = ACTIONS(1741),
    [sym_true] = ACTIONS(1744),
    [sym_false] = ACTIONS(1744),
    [anon_sym_NULL] = ACTIONS(1747),
    [anon_sym_nullptr] = ACTIONS(1747),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(1750),
    [anon_sym___typeof] = ACTIONS(1750),
    [anon_sym_typeof] = ACTIONS(1750),
    [anon_sym_ATimport] = ACTIONS(1753),
    [aux_sym_preproc_undef_token1] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(1755),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(1755),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(1755),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE] = ACTIONS(1758),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_API_AVAILABLE] = ACTIONS(1758),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_API_DEPRECATED] = ACTIONS(1758),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(1758),
    [anon_sym___deprecated_msg] = ACTIONS(1758),
    [anon_sym___deprecated_enum_msg] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(1758),
    [anon_sym_ATprotocol] = ACTIONS(1753),
    [anon_sym_ATinterface] = ACTIONS(1753),
    [anon_sym_ATimplementation] = ACTIONS(1753),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1753),
    [anon_sym_ATsynthesize] = ACTIONS(1753),
    [anon_sym_ATdynamic] = ACTIONS(1753),
    [anon_sym_ATtry] = ACTIONS(1846),
    [anon_sym___try] = ACTIONS(1849),
    [anon_sym_ATthrow] = ACTIONS(1852),
    [anon_sym_ATselector] = ACTIONS(1770),
    [anon_sym_ATavailable] = ACTIONS(1773),
    [anon_sym___builtin_available] = ACTIONS(1776),
    [anon_sym_va_arg] = ACTIONS(1779),
    [anon_sym_ATencode] = ACTIONS(1782),
    [anon_sym_ATsynchronized] = ACTIONS(1855),
    [anon_sym_BOOL] = ACTIONS(1788),
    [anon_sym_IMP] = ACTIONS(1788),
    [anon_sym_SEL] = ACTIONS(1788),
    [anon_sym_Class] = ACTIONS(1788),
    [anon_sym_id] = ACTIONS(1791),
  },
  [STATE(137)] = {
    [sym_declaration] = STATE(138),
    [sym_type_definition] = STATE(138),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4202),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_compound_statement] = STATE(138),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(138),
    [sym_labeled_statement] = STATE(138),
    [sym_expression_statement] = STATE(138),
    [sym_if_statement] = STATE(138),
    [sym_switch_statement] = STATE(138),
    [sym_while_statement] = STATE(138),
    [sym_do_statement] = STATE(138),
    [sym_for_statement] = STATE(138),
    [sym_return_statement] = STATE(138),
    [sym_break_statement] = STATE(138),
    [sym_continue_statement] = STATE(138),
    [sym_goto_statement] = STATE(138),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(138),
    [sym_throw_statement] = STATE(138),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(138),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(138),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(138),
    [sym_identifier] = ACTIONS(1794),
    [aux_sym_preproc_include_token1] = ACTIONS(1595),
    [aux_sym_preproc_include_token2] = ACTIONS(1595),
    [aux_sym_preproc_def_token1] = ACTIONS(1595),
    [aux_sym_preproc_if_token1] = ACTIONS(1595),
    [aux_sym_preproc_if_token2] = ACTIONS(1595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1595),
    [aux_sym_preproc_else_token1] = ACTIONS(1595),
    [aux_sym_preproc_elif_token1] = ACTIONS(1595),
    [sym_preproc_directive] = ACTIONS(1595),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(1796),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1595),
    [anon_sym___clrcall] = ACTIONS(1595),
    [anon_sym___stdcall] = ACTIONS(1595),
    [anon_sym___fastcall] = ACTIONS(1595),
    [anon_sym___thiscall] = ACTIONS(1595),
    [anon_sym___vectorcall] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_else] = ACTIONS(1595),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(1595),
    [anon_sym_default] = ACTIONS(1595),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1597),
    [aux_sym_preproc_undef_token1] = ACTIONS(1595),
    [anon_sym_POUND] = ACTIONS(1595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1597),
    [anon_sym_ATinterface] = ACTIONS(1597),
    [anon_sym_ATimplementation] = ACTIONS(1597),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1597),
    [anon_sym_ATsynthesize] = ACTIONS(1597),
    [anon_sym_ATdynamic] = ACTIONS(1597),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(138)] = {
    [sym_declaration] = STATE(136),
    [sym_type_definition] = STATE(136),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4202),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_compound_statement] = STATE(136),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(136),
    [sym_labeled_statement] = STATE(136),
    [sym_expression_statement] = STATE(136),
    [sym_if_statement] = STATE(136),
    [sym_switch_statement] = STATE(136),
    [sym_while_statement] = STATE(136),
    [sym_do_statement] = STATE(136),
    [sym_for_statement] = STATE(136),
    [sym_return_statement] = STATE(136),
    [sym_break_statement] = STATE(136),
    [sym_continue_statement] = STATE(136),
    [sym_goto_statement] = STATE(136),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(136),
    [sym_throw_statement] = STATE(136),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(136),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(136),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(136),
    [sym_identifier] = ACTIONS(1794),
    [aux_sym_preproc_include_token1] = ACTIONS(1603),
    [aux_sym_preproc_include_token2] = ACTIONS(1603),
    [aux_sym_preproc_def_token1] = ACTIONS(1603),
    [aux_sym_preproc_if_token1] = ACTIONS(1603),
    [aux_sym_preproc_if_token2] = ACTIONS(1603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1603),
    [aux_sym_preproc_else_token1] = ACTIONS(1603),
    [aux_sym_preproc_elif_token1] = ACTIONS(1603),
    [sym_preproc_directive] = ACTIONS(1603),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(1796),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1603),
    [anon_sym___clrcall] = ACTIONS(1603),
    [anon_sym___stdcall] = ACTIONS(1603),
    [anon_sym___fastcall] = ACTIONS(1603),
    [anon_sym___thiscall] = ACTIONS(1603),
    [anon_sym___vectorcall] = ACTIONS(1603),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_else] = ACTIONS(1603),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(1603),
    [anon_sym_default] = ACTIONS(1603),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1605),
    [aux_sym_preproc_undef_token1] = ACTIONS(1603),
    [anon_sym_POUND] = ACTIONS(1603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1605),
    [anon_sym_ATinterface] = ACTIONS(1605),
    [anon_sym_ATimplementation] = ACTIONS(1605),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1605),
    [anon_sym_ATsynthesize] = ACTIONS(1605),
    [anon_sym_ATdynamic] = ACTIONS(1605),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(139)] = {
    [sym_declaration] = STATE(136),
    [sym_type_definition] = STATE(136),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4202),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2848),
    [sym_compound_statement] = STATE(136),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(136),
    [sym_labeled_statement] = STATE(136),
    [sym_expression_statement] = STATE(136),
    [sym_if_statement] = STATE(136),
    [sym_switch_statement] = STATE(136),
    [sym_while_statement] = STATE(136),
    [sym_do_statement] = STATE(136),
    [sym_for_statement] = STATE(136),
    [sym_return_statement] = STATE(136),
    [sym_break_statement] = STATE(136),
    [sym_continue_statement] = STATE(136),
    [sym_goto_statement] = STATE(136),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(136),
    [sym_throw_statement] = STATE(136),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(136),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(136),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(136),
    [sym_identifier] = ACTIONS(1794),
    [aux_sym_preproc_include_token1] = ACTIONS(1581),
    [aux_sym_preproc_include_token2] = ACTIONS(1581),
    [aux_sym_preproc_def_token1] = ACTIONS(1581),
    [aux_sym_preproc_if_token1] = ACTIONS(1581),
    [aux_sym_preproc_if_token2] = ACTIONS(1581),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1581),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1581),
    [aux_sym_preproc_else_token1] = ACTIONS(1581),
    [aux_sym_preproc_elif_token1] = ACTIONS(1581),
    [sym_preproc_directive] = ACTIONS(1581),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(1796),
    [anon_sym_typedef] = ACTIONS(297),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1581),
    [anon_sym___clrcall] = ACTIONS(1581),
    [anon_sym___stdcall] = ACTIONS(1581),
    [anon_sym___fastcall] = ACTIONS(1581),
    [anon_sym___thiscall] = ACTIONS(1581),
    [anon_sym___vectorcall] = ACTIONS(1581),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_else] = ACTIONS(1581),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(1581),
    [anon_sym_default] = ACTIONS(1581),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1591),
    [aux_sym_preproc_undef_token1] = ACTIONS(1581),
    [anon_sym_POUND] = ACTIONS(1581),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1591),
    [anon_sym_ATinterface] = ACTIONS(1591),
    [anon_sym_ATimplementation] = ACTIONS(1591),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1591),
    [anon_sym_ATsynthesize] = ACTIONS(1591),
    [anon_sym_ATdynamic] = ACTIONS(1591),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(140)] = {
    [sym_declaration] = STATE(143),
    [sym_type_definition] = STATE(143),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4228),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2849),
    [sym_compound_statement] = STATE(143),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(143),
    [sym_labeled_statement] = STATE(143),
    [sym_expression_statement] = STATE(143),
    [sym_if_statement] = STATE(143),
    [sym_switch_statement] = STATE(143),
    [sym_while_statement] = STATE(143),
    [sym_do_statement] = STATE(143),
    [sym_for_statement] = STATE(143),
    [sym_return_statement] = STATE(143),
    [sym_break_statement] = STATE(143),
    [sym_continue_statement] = STATE(143),
    [sym_goto_statement] = STATE(143),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(143),
    [sym_throw_statement] = STATE(143),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(143),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(143),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(143),
    [sym_identifier] = ACTIONS(1858),
    [aux_sym_preproc_include_token1] = ACTIONS(1599),
    [aux_sym_preproc_include_token2] = ACTIONS(1599),
    [aux_sym_preproc_def_token1] = ACTIONS(1599),
    [aux_sym_preproc_if_token1] = ACTIONS(1599),
    [aux_sym_preproc_if_token2] = ACTIONS(1599),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1599),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1599),
    [sym_preproc_directive] = ACTIONS(1599),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(1860),
    [anon_sym_typedef] = ACTIONS(864),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1599),
    [anon_sym___clrcall] = ACTIONS(1599),
    [anon_sym___stdcall] = ACTIONS(1599),
    [anon_sym___fastcall] = ACTIONS(1599),
    [anon_sym___thiscall] = ACTIONS(1599),
    [anon_sym___vectorcall] = ACTIONS(1599),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_else] = ACTIONS(1599),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(1599),
    [anon_sym_default] = ACTIONS(1599),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1601),
    [aux_sym_preproc_undef_token1] = ACTIONS(1599),
    [anon_sym_POUND] = ACTIONS(1599),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1601),
    [anon_sym_ATinterface] = ACTIONS(1601),
    [anon_sym_ATimplementation] = ACTIONS(1601),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1601),
    [anon_sym_ATsynthesize] = ACTIONS(1601),
    [anon_sym_ATdynamic] = ACTIONS(1601),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(141)] = {
    [sym_declaration] = STATE(142),
    [sym_type_definition] = STATE(142),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4228),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2849),
    [sym_compound_statement] = STATE(142),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(142),
    [sym_labeled_statement] = STATE(142),
    [sym_expression_statement] = STATE(142),
    [sym_if_statement] = STATE(142),
    [sym_switch_statement] = STATE(142),
    [sym_while_statement] = STATE(142),
    [sym_do_statement] = STATE(142),
    [sym_for_statement] = STATE(142),
    [sym_return_statement] = STATE(142),
    [sym_break_statement] = STATE(142),
    [sym_continue_statement] = STATE(142),
    [sym_goto_statement] = STATE(142),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(142),
    [sym_throw_statement] = STATE(142),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(142),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(142),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(142),
    [sym_identifier] = ACTIONS(1858),
    [aux_sym_preproc_include_token1] = ACTIONS(1595),
    [aux_sym_preproc_include_token2] = ACTIONS(1595),
    [aux_sym_preproc_def_token1] = ACTIONS(1595),
    [aux_sym_preproc_if_token1] = ACTIONS(1595),
    [aux_sym_preproc_if_token2] = ACTIONS(1595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1595),
    [sym_preproc_directive] = ACTIONS(1595),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(1860),
    [anon_sym_typedef] = ACTIONS(864),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1595),
    [anon_sym___clrcall] = ACTIONS(1595),
    [anon_sym___stdcall] = ACTIONS(1595),
    [anon_sym___fastcall] = ACTIONS(1595),
    [anon_sym___thiscall] = ACTIONS(1595),
    [anon_sym___vectorcall] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_else] = ACTIONS(1595),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(1595),
    [anon_sym_default] = ACTIONS(1595),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1597),
    [aux_sym_preproc_undef_token1] = ACTIONS(1595),
    [anon_sym_POUND] = ACTIONS(1595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1597),
    [anon_sym_ATinterface] = ACTIONS(1597),
    [anon_sym_ATimplementation] = ACTIONS(1597),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1597),
    [anon_sym_ATsynthesize] = ACTIONS(1597),
    [anon_sym_ATdynamic] = ACTIONS(1597),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(142)] = {
    [sym_declaration] = STATE(144),
    [sym_type_definition] = STATE(144),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4228),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2849),
    [sym_compound_statement] = STATE(144),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(144),
    [sym_labeled_statement] = STATE(144),
    [sym_expression_statement] = STATE(144),
    [sym_if_statement] = STATE(144),
    [sym_switch_statement] = STATE(144),
    [sym_while_statement] = STATE(144),
    [sym_do_statement] = STATE(144),
    [sym_for_statement] = STATE(144),
    [sym_return_statement] = STATE(144),
    [sym_break_statement] = STATE(144),
    [sym_continue_statement] = STATE(144),
    [sym_goto_statement] = STATE(144),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(144),
    [sym_throw_statement] = STATE(144),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(144),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(144),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(144),
    [sym_identifier] = ACTIONS(1858),
    [aux_sym_preproc_include_token1] = ACTIONS(1603),
    [aux_sym_preproc_include_token2] = ACTIONS(1603),
    [aux_sym_preproc_def_token1] = ACTIONS(1603),
    [aux_sym_preproc_if_token1] = ACTIONS(1603),
    [aux_sym_preproc_if_token2] = ACTIONS(1603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1603),
    [sym_preproc_directive] = ACTIONS(1603),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(1860),
    [anon_sym_typedef] = ACTIONS(864),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1603),
    [anon_sym___clrcall] = ACTIONS(1603),
    [anon_sym___stdcall] = ACTIONS(1603),
    [anon_sym___fastcall] = ACTIONS(1603),
    [anon_sym___thiscall] = ACTIONS(1603),
    [anon_sym___vectorcall] = ACTIONS(1603),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_else] = ACTIONS(1603),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(1603),
    [anon_sym_default] = ACTIONS(1603),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1605),
    [aux_sym_preproc_undef_token1] = ACTIONS(1603),
    [anon_sym_POUND] = ACTIONS(1603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1605),
    [anon_sym_ATinterface] = ACTIONS(1605),
    [anon_sym_ATimplementation] = ACTIONS(1605),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1605),
    [anon_sym_ATsynthesize] = ACTIONS(1605),
    [anon_sym_ATdynamic] = ACTIONS(1605),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(143)] = {
    [sym_declaration] = STATE(144),
    [sym_type_definition] = STATE(144),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4228),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2849),
    [sym_compound_statement] = STATE(144),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(144),
    [sym_labeled_statement] = STATE(144),
    [sym_expression_statement] = STATE(144),
    [sym_if_statement] = STATE(144),
    [sym_switch_statement] = STATE(144),
    [sym_while_statement] = STATE(144),
    [sym_do_statement] = STATE(144),
    [sym_for_statement] = STATE(144),
    [sym_return_statement] = STATE(144),
    [sym_break_statement] = STATE(144),
    [sym_continue_statement] = STATE(144),
    [sym_goto_statement] = STATE(144),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(144),
    [sym_throw_statement] = STATE(144),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(144),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(144),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(144),
    [sym_identifier] = ACTIONS(1858),
    [aux_sym_preproc_include_token1] = ACTIONS(1581),
    [aux_sym_preproc_include_token2] = ACTIONS(1581),
    [aux_sym_preproc_def_token1] = ACTIONS(1581),
    [aux_sym_preproc_if_token1] = ACTIONS(1581),
    [aux_sym_preproc_if_token2] = ACTIONS(1581),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1581),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1581),
    [sym_preproc_directive] = ACTIONS(1581),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(1860),
    [anon_sym_typedef] = ACTIONS(864),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1581),
    [anon_sym___clrcall] = ACTIONS(1581),
    [anon_sym___stdcall] = ACTIONS(1581),
    [anon_sym___fastcall] = ACTIONS(1581),
    [anon_sym___thiscall] = ACTIONS(1581),
    [anon_sym___vectorcall] = ACTIONS(1581),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_else] = ACTIONS(1581),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(1581),
    [anon_sym_default] = ACTIONS(1581),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1591),
    [aux_sym_preproc_undef_token1] = ACTIONS(1581),
    [anon_sym_POUND] = ACTIONS(1581),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1591),
    [anon_sym_ATinterface] = ACTIONS(1591),
    [anon_sym_ATimplementation] = ACTIONS(1591),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1591),
    [anon_sym_ATsynthesize] = ACTIONS(1591),
    [anon_sym_ATdynamic] = ACTIONS(1591),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(144)] = {
    [sym_declaration] = STATE(144),
    [sym_type_definition] = STATE(144),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4228),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2849),
    [sym_compound_statement] = STATE(144),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(144),
    [sym_labeled_statement] = STATE(144),
    [sym_expression_statement] = STATE(144),
    [sym_if_statement] = STATE(144),
    [sym_switch_statement] = STATE(144),
    [sym_while_statement] = STATE(144),
    [sym_do_statement] = STATE(144),
    [sym_for_statement] = STATE(144),
    [sym_return_statement] = STATE(144),
    [sym_break_statement] = STATE(144),
    [sym_continue_statement] = STATE(144),
    [sym_goto_statement] = STATE(144),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(144),
    [sym_throw_statement] = STATE(144),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(144),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(144),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(144),
    [sym_identifier] = ACTIONS(1862),
    [aux_sym_preproc_include_token1] = ACTIONS(1610),
    [aux_sym_preproc_include_token2] = ACTIONS(1610),
    [aux_sym_preproc_def_token1] = ACTIONS(1610),
    [aux_sym_preproc_if_token1] = ACTIONS(1610),
    [aux_sym_preproc_if_token2] = ACTIONS(1610),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1610),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1610),
    [sym_preproc_directive] = ACTIONS(1610),
    [anon_sym_LPAREN2] = ACTIONS(1612),
    [anon_sym_BANG] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_DASH] = ACTIONS(1618),
    [anon_sym_PLUS] = ACTIONS(1618),
    [anon_sym_STAR] = ACTIONS(1621),
    [anon_sym_CARET] = ACTIONS(1624),
    [anon_sym_AMP] = ACTIONS(1621),
    [anon_sym_SEMI] = ACTIONS(1865),
    [anon_sym___extension__] = ACTIONS(1868),
    [anon_sym_typedef] = ACTIONS(1871),
    [anon_sym_extern] = ACTIONS(1636),
    [anon_sym___attribute__] = ACTIONS(1639),
    [anon_sym___attribute] = ACTIONS(1639),
    [anon_sym_noreturn] = ACTIONS(1642),
    [anon_sym_LBRACK] = ACTIONS(1645),
    [anon_sym___declspec] = ACTIONS(1648),
    [anon_sym___cdecl] = ACTIONS(1610),
    [anon_sym___clrcall] = ACTIONS(1610),
    [anon_sym___stdcall] = ACTIONS(1610),
    [anon_sym___fastcall] = ACTIONS(1610),
    [anon_sym___thiscall] = ACTIONS(1610),
    [anon_sym___vectorcall] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1874),
    [anon_sym_signed] = ACTIONS(1654),
    [anon_sym_unsigned] = ACTIONS(1654),
    [anon_sym_long] = ACTIONS(1654),
    [anon_sym_short] = ACTIONS(1654),
    [anon_sym_static] = ACTIONS(1636),
    [anon_sym_ATautoreleasepool] = ACTIONS(1877),
    [anon_sym_auto] = ACTIONS(1636),
    [anon_sym_register] = ACTIONS(1636),
    [anon_sym_inline] = ACTIONS(1636),
    [anon_sym___inline] = ACTIONS(1636),
    [anon_sym___inline__] = ACTIONS(1636),
    [anon_sym___forceinline] = ACTIONS(1636),
    [anon_sym_thread_local] = ACTIONS(1636),
    [anon_sym___thread] = ACTIONS(1636),
    [anon_sym_CG_EXTERN] = ACTIONS(1636),
    [anon_sym_CG_INLINE] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(1636),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(1636),
    [anon_sym_IBOutlet] = ACTIONS(1636),
    [anon_sym_IBInspectable] = ACTIONS(1636),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(1636),
    [anon_sym_NS_INLINE] = ACTIONS(1636),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(1636),
    [anon_sym_OBJC_EXPORT] = ACTIONS(1636),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(1636),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(1636),
    [anon_sym_const] = ACTIONS(1660),
    [anon_sym_constexpr] = ACTIONS(1642),
    [anon_sym_volatile] = ACTIONS(1642),
    [anon_sym_restrict] = ACTIONS(1642),
    [anon_sym___restrict__] = ACTIONS(1642),
    [anon_sym__Atomic] = ACTIONS(1642),
    [anon_sym__Noreturn] = ACTIONS(1642),
    [anon_sym__Nonnull] = ACTIONS(1642),
    [anon_sym_nullable] = ACTIONS(1642),
    [anon_sym__Complex] = ACTIONS(1642),
    [anon_sym__Nullable] = ACTIONS(1642),
    [anon_sym__Nullable_result] = ACTIONS(1642),
    [anon_sym__Null_unspecified] = ACTIONS(1642),
    [anon_sym___autoreleasing] = ACTIONS(1642),
    [anon_sym___block] = ACTIONS(1642),
    [anon_sym___bridge] = ACTIONS(1642),
    [anon_sym___bridge_retained] = ACTIONS(1642),
    [anon_sym___bridge_transfer] = ACTIONS(1642),
    [anon_sym___complex] = ACTIONS(1642),
    [anon_sym___const] = ACTIONS(1642),
    [anon_sym___imag] = ACTIONS(1663),
    [anon_sym___kindof] = ACTIONS(1642),
    [anon_sym___nonnull] = ACTIONS(1642),
    [anon_sym___nullable] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(1642),
    [anon_sym___real] = ACTIONS(1663),
    [anon_sym___strong] = ACTIONS(1642),
    [anon_sym___unsafe_unretained] = ACTIONS(1642),
    [anon_sym___unused] = ACTIONS(1642),
    [anon_sym___weak] = ACTIONS(1642),
    [anon_sym_alignas] = ACTIONS(1666),
    [anon_sym__Alignas] = ACTIONS(1666),
    [sym_primitive_type] = ACTIONS(1669),
    [anon_sym_enum] = ACTIONS(1672),
    [anon_sym_struct] = ACTIONS(1675),
    [anon_sym_union] = ACTIONS(1678),
    [anon_sym_if] = ACTIONS(1880),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_switch] = ACTIONS(1883),
    [anon_sym_case] = ACTIONS(1610),
    [anon_sym_default] = ACTIONS(1610),
    [anon_sym_while] = ACTIONS(1886),
    [anon_sym_do] = ACTIONS(1889),
    [anon_sym_for] = ACTIONS(1892),
    [anon_sym_in] = ACTIONS(1696),
    [anon_sym_return] = ACTIONS(1895),
    [anon_sym_break] = ACTIONS(1898),
    [anon_sym_continue] = ACTIONS(1901),
    [anon_sym_goto] = ACTIONS(1904),
    [anon_sym_DASH_DASH] = ACTIONS(1711),
    [anon_sym_PLUS_PLUS] = ACTIONS(1711),
    [anon_sym_sizeof] = ACTIONS(1714),
    [anon_sym___alignof__] = ACTIONS(1717),
    [anon_sym___alignof] = ACTIONS(1717),
    [anon_sym__alignof] = ACTIONS(1717),
    [anon_sym_alignof] = ACTIONS(1717),
    [anon_sym__Alignof] = ACTIONS(1717),
    [anon_sym_offsetof] = ACTIONS(1720),
    [anon_sym__Generic] = ACTIONS(1723),
    [anon_sym_asm] = ACTIONS(1726),
    [anon_sym___asm__] = ACTIONS(1726),
    [anon_sym___asm] = ACTIONS(1907),
    [sym_number_literal] = ACTIONS(1732),
    [anon_sym_L_SQUOTE] = ACTIONS(1735),
    [anon_sym_u_SQUOTE] = ACTIONS(1735),
    [anon_sym_U_SQUOTE] = ACTIONS(1735),
    [anon_sym_u8_SQUOTE] = ACTIONS(1735),
    [anon_sym_SQUOTE] = ACTIONS(1735),
    [anon_sym_AT] = ACTIONS(1738),
    [anon_sym_DQUOTE] = ACTIONS(1741),
    [anon_sym_L_DQUOTE] = ACTIONS(1741),
    [anon_sym_u_DQUOTE] = ACTIONS(1741),
    [anon_sym_U_DQUOTE] = ACTIONS(1741),
    [anon_sym_u8_DQUOTE] = ACTIONS(1741),
    [sym_true] = ACTIONS(1744),
    [sym_false] = ACTIONS(1744),
    [anon_sym_NULL] = ACTIONS(1747),
    [anon_sym_nullptr] = ACTIONS(1747),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(1750),
    [anon_sym___typeof] = ACTIONS(1750),
    [anon_sym_typeof] = ACTIONS(1750),
    [anon_sym_ATimport] = ACTIONS(1753),
    [aux_sym_preproc_undef_token1] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(1755),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(1755),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(1755),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE] = ACTIONS(1758),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_API_AVAILABLE] = ACTIONS(1758),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_API_DEPRECATED] = ACTIONS(1758),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(1758),
    [anon_sym___deprecated_msg] = ACTIONS(1758),
    [anon_sym___deprecated_enum_msg] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(1758),
    [anon_sym_ATprotocol] = ACTIONS(1753),
    [anon_sym_ATinterface] = ACTIONS(1753),
    [anon_sym_ATimplementation] = ACTIONS(1753),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1753),
    [anon_sym_ATsynthesize] = ACTIONS(1753),
    [anon_sym_ATdynamic] = ACTIONS(1753),
    [anon_sym_ATtry] = ACTIONS(1910),
    [anon_sym___try] = ACTIONS(1913),
    [anon_sym_ATthrow] = ACTIONS(1916),
    [anon_sym_ATselector] = ACTIONS(1770),
    [anon_sym_ATavailable] = ACTIONS(1773),
    [anon_sym___builtin_available] = ACTIONS(1776),
    [anon_sym_va_arg] = ACTIONS(1779),
    [anon_sym_ATencode] = ACTIONS(1782),
    [anon_sym_ATsynchronized] = ACTIONS(1919),
    [anon_sym_BOOL] = ACTIONS(1788),
    [anon_sym_IMP] = ACTIONS(1788),
    [anon_sym_SEL] = ACTIONS(1788),
    [anon_sym_Class] = ACTIONS(1788),
    [anon_sym_id] = ACTIONS(1791),
  },
  [STATE(145)] = {
    [sym_declaration] = STATE(153),
    [sym_type_definition] = STATE(153),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4213),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_compound_statement] = STATE(153),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(153),
    [sym_labeled_statement] = STATE(153),
    [sym_expression_statement] = STATE(153),
    [sym_if_statement] = STATE(153),
    [sym_switch_statement] = STATE(153),
    [sym_while_statement] = STATE(153),
    [sym_do_statement] = STATE(153),
    [sym_for_statement] = STATE(153),
    [sym_return_statement] = STATE(153),
    [sym_break_statement] = STATE(153),
    [sym_continue_statement] = STATE(153),
    [sym_goto_statement] = STATE(153),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(153),
    [sym_throw_statement] = STATE(153),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(153),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(153),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(153),
    [sym_identifier] = ACTIONS(1922),
    [aux_sym_preproc_include_token1] = ACTIONS(1599),
    [aux_sym_preproc_include_token2] = ACTIONS(1599),
    [aux_sym_preproc_def_token1] = ACTIONS(1599),
    [aux_sym_preproc_if_token1] = ACTIONS(1599),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1599),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1599),
    [sym_preproc_directive] = ACTIONS(1599),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(1924),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1599),
    [anon_sym___clrcall] = ACTIONS(1599),
    [anon_sym___stdcall] = ACTIONS(1599),
    [anon_sym___fastcall] = ACTIONS(1599),
    [anon_sym___thiscall] = ACTIONS(1599),
    [anon_sym___vectorcall] = ACTIONS(1599),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1601),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_else] = ACTIONS(1599),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(1599),
    [anon_sym_default] = ACTIONS(1599),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1601),
    [aux_sym_preproc_undef_token1] = ACTIONS(1599),
    [anon_sym_POUND] = ACTIONS(1599),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1601),
    [anon_sym_ATinterface] = ACTIONS(1601),
    [anon_sym_ATimplementation] = ACTIONS(1601),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1601),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(146)] = {
    [sym_declaration] = STATE(147),
    [sym_type_definition] = STATE(147),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4212),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2898),
    [sym_compound_statement] = STATE(147),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(147),
    [sym_labeled_statement] = STATE(147),
    [sym_expression_statement] = STATE(147),
    [sym_if_statement] = STATE(147),
    [sym_switch_statement] = STATE(147),
    [sym_while_statement] = STATE(147),
    [sym_do_statement] = STATE(147),
    [sym_for_statement] = STATE(147),
    [sym_return_statement] = STATE(147),
    [sym_break_statement] = STATE(147),
    [sym_continue_statement] = STATE(147),
    [sym_goto_statement] = STATE(147),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(147),
    [sym_throw_statement] = STATE(147),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(147),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(147),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(147),
    [ts_builtin_sym_end] = ACTIONS(1601),
    [sym_identifier] = ACTIONS(1926),
    [aux_sym_preproc_include_token1] = ACTIONS(1599),
    [aux_sym_preproc_include_token2] = ACTIONS(1599),
    [aux_sym_preproc_def_token1] = ACTIONS(1599),
    [aux_sym_preproc_if_token1] = ACTIONS(1599),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1599),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1599),
    [sym_preproc_directive] = ACTIONS(1599),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(1930),
    [anon_sym_typedef] = ACTIONS(1932),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1599),
    [anon_sym___clrcall] = ACTIONS(1599),
    [anon_sym___stdcall] = ACTIONS(1599),
    [anon_sym___fastcall] = ACTIONS(1599),
    [anon_sym___thiscall] = ACTIONS(1599),
    [anon_sym___vectorcall] = ACTIONS(1599),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_else] = ACTIONS(1599),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_case] = ACTIONS(1599),
    [anon_sym_default] = ACTIONS(1599),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1601),
    [aux_sym_preproc_undef_token1] = ACTIONS(1599),
    [anon_sym_POUND] = ACTIONS(1599),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1601),
    [anon_sym_ATinterface] = ACTIONS(1601),
    [anon_sym_ATimplementation] = ACTIONS(1601),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1601),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(147)] = {
    [sym_declaration] = STATE(148),
    [sym_type_definition] = STATE(148),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4212),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2898),
    [sym_compound_statement] = STATE(148),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(148),
    [sym_labeled_statement] = STATE(148),
    [sym_expression_statement] = STATE(148),
    [sym_if_statement] = STATE(148),
    [sym_switch_statement] = STATE(148),
    [sym_while_statement] = STATE(148),
    [sym_do_statement] = STATE(148),
    [sym_for_statement] = STATE(148),
    [sym_return_statement] = STATE(148),
    [sym_break_statement] = STATE(148),
    [sym_continue_statement] = STATE(148),
    [sym_goto_statement] = STATE(148),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(148),
    [sym_throw_statement] = STATE(148),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(148),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(148),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(148),
    [ts_builtin_sym_end] = ACTIONS(1591),
    [sym_identifier] = ACTIONS(1926),
    [aux_sym_preproc_include_token1] = ACTIONS(1581),
    [aux_sym_preproc_include_token2] = ACTIONS(1581),
    [aux_sym_preproc_def_token1] = ACTIONS(1581),
    [aux_sym_preproc_if_token1] = ACTIONS(1581),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1581),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1581),
    [sym_preproc_directive] = ACTIONS(1581),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(1930),
    [anon_sym_typedef] = ACTIONS(1932),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1581),
    [anon_sym___clrcall] = ACTIONS(1581),
    [anon_sym___stdcall] = ACTIONS(1581),
    [anon_sym___fastcall] = ACTIONS(1581),
    [anon_sym___thiscall] = ACTIONS(1581),
    [anon_sym___vectorcall] = ACTIONS(1581),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_else] = ACTIONS(1581),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_case] = ACTIONS(1581),
    [anon_sym_default] = ACTIONS(1581),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1591),
    [aux_sym_preproc_undef_token1] = ACTIONS(1581),
    [anon_sym_POUND] = ACTIONS(1581),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1591),
    [anon_sym_ATinterface] = ACTIONS(1591),
    [anon_sym_ATimplementation] = ACTIONS(1591),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1591),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(148)] = {
    [sym_declaration] = STATE(148),
    [sym_type_definition] = STATE(148),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4212),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2898),
    [sym_compound_statement] = STATE(148),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(148),
    [sym_labeled_statement] = STATE(148),
    [sym_expression_statement] = STATE(148),
    [sym_if_statement] = STATE(148),
    [sym_switch_statement] = STATE(148),
    [sym_while_statement] = STATE(148),
    [sym_do_statement] = STATE(148),
    [sym_for_statement] = STATE(148),
    [sym_return_statement] = STATE(148),
    [sym_break_statement] = STATE(148),
    [sym_continue_statement] = STATE(148),
    [sym_goto_statement] = STATE(148),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(148),
    [sym_throw_statement] = STATE(148),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(148),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(148),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(148),
    [ts_builtin_sym_end] = ACTIONS(1753),
    [sym_identifier] = ACTIONS(1966),
    [aux_sym_preproc_include_token1] = ACTIONS(1610),
    [aux_sym_preproc_include_token2] = ACTIONS(1610),
    [aux_sym_preproc_def_token1] = ACTIONS(1610),
    [aux_sym_preproc_if_token1] = ACTIONS(1610),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1610),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1610),
    [sym_preproc_directive] = ACTIONS(1610),
    [anon_sym_LPAREN2] = ACTIONS(1612),
    [anon_sym_BANG] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_DASH] = ACTIONS(1618),
    [anon_sym_PLUS] = ACTIONS(1618),
    [anon_sym_STAR] = ACTIONS(1621),
    [anon_sym_CARET] = ACTIONS(1624),
    [anon_sym_AMP] = ACTIONS(1621),
    [anon_sym_SEMI] = ACTIONS(1969),
    [anon_sym___extension__] = ACTIONS(1972),
    [anon_sym_typedef] = ACTIONS(1975),
    [anon_sym_extern] = ACTIONS(1636),
    [anon_sym___attribute__] = ACTIONS(1639),
    [anon_sym___attribute] = ACTIONS(1639),
    [anon_sym_noreturn] = ACTIONS(1642),
    [anon_sym_LBRACK] = ACTIONS(1645),
    [anon_sym___declspec] = ACTIONS(1648),
    [anon_sym___cdecl] = ACTIONS(1610),
    [anon_sym___clrcall] = ACTIONS(1610),
    [anon_sym___stdcall] = ACTIONS(1610),
    [anon_sym___fastcall] = ACTIONS(1610),
    [anon_sym___thiscall] = ACTIONS(1610),
    [anon_sym___vectorcall] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1978),
    [anon_sym_signed] = ACTIONS(1654),
    [anon_sym_unsigned] = ACTIONS(1654),
    [anon_sym_long] = ACTIONS(1654),
    [anon_sym_short] = ACTIONS(1654),
    [anon_sym_static] = ACTIONS(1636),
    [anon_sym_ATautoreleasepool] = ACTIONS(1981),
    [anon_sym_auto] = ACTIONS(1636),
    [anon_sym_register] = ACTIONS(1636),
    [anon_sym_inline] = ACTIONS(1636),
    [anon_sym___inline] = ACTIONS(1636),
    [anon_sym___inline__] = ACTIONS(1636),
    [anon_sym___forceinline] = ACTIONS(1636),
    [anon_sym_thread_local] = ACTIONS(1636),
    [anon_sym___thread] = ACTIONS(1636),
    [anon_sym_CG_EXTERN] = ACTIONS(1636),
    [anon_sym_CG_INLINE] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(1636),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(1636),
    [anon_sym_IBOutlet] = ACTIONS(1636),
    [anon_sym_IBInspectable] = ACTIONS(1636),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(1636),
    [anon_sym_NS_INLINE] = ACTIONS(1636),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(1636),
    [anon_sym_OBJC_EXPORT] = ACTIONS(1636),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(1636),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(1636),
    [anon_sym_const] = ACTIONS(1660),
    [anon_sym_constexpr] = ACTIONS(1642),
    [anon_sym_volatile] = ACTIONS(1642),
    [anon_sym_restrict] = ACTIONS(1642),
    [anon_sym___restrict__] = ACTIONS(1642),
    [anon_sym__Atomic] = ACTIONS(1642),
    [anon_sym__Noreturn] = ACTIONS(1642),
    [anon_sym__Nonnull] = ACTIONS(1642),
    [anon_sym_nullable] = ACTIONS(1642),
    [anon_sym__Complex] = ACTIONS(1642),
    [anon_sym__Nullable] = ACTIONS(1642),
    [anon_sym__Nullable_result] = ACTIONS(1642),
    [anon_sym__Null_unspecified] = ACTIONS(1642),
    [anon_sym___autoreleasing] = ACTIONS(1642),
    [anon_sym___block] = ACTIONS(1642),
    [anon_sym___bridge] = ACTIONS(1642),
    [anon_sym___bridge_retained] = ACTIONS(1642),
    [anon_sym___bridge_transfer] = ACTIONS(1642),
    [anon_sym___complex] = ACTIONS(1642),
    [anon_sym___const] = ACTIONS(1642),
    [anon_sym___imag] = ACTIONS(1663),
    [anon_sym___kindof] = ACTIONS(1642),
    [anon_sym___nonnull] = ACTIONS(1642),
    [anon_sym___nullable] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(1642),
    [anon_sym___real] = ACTIONS(1663),
    [anon_sym___strong] = ACTIONS(1642),
    [anon_sym___unsafe_unretained] = ACTIONS(1642),
    [anon_sym___unused] = ACTIONS(1642),
    [anon_sym___weak] = ACTIONS(1642),
    [anon_sym_alignas] = ACTIONS(1666),
    [anon_sym__Alignas] = ACTIONS(1666),
    [sym_primitive_type] = ACTIONS(1669),
    [anon_sym_enum] = ACTIONS(1672),
    [anon_sym_struct] = ACTIONS(1675),
    [anon_sym_union] = ACTIONS(1678),
    [anon_sym_if] = ACTIONS(1984),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_switch] = ACTIONS(1987),
    [anon_sym_case] = ACTIONS(1610),
    [anon_sym_default] = ACTIONS(1610),
    [anon_sym_while] = ACTIONS(1990),
    [anon_sym_do] = ACTIONS(1993),
    [anon_sym_for] = ACTIONS(1996),
    [anon_sym_in] = ACTIONS(1696),
    [anon_sym_return] = ACTIONS(1999),
    [anon_sym_break] = ACTIONS(2002),
    [anon_sym_continue] = ACTIONS(2005),
    [anon_sym_goto] = ACTIONS(2008),
    [anon_sym_DASH_DASH] = ACTIONS(1711),
    [anon_sym_PLUS_PLUS] = ACTIONS(1711),
    [anon_sym_sizeof] = ACTIONS(1714),
    [anon_sym___alignof__] = ACTIONS(1717),
    [anon_sym___alignof] = ACTIONS(1717),
    [anon_sym__alignof] = ACTIONS(1717),
    [anon_sym_alignof] = ACTIONS(1717),
    [anon_sym__Alignof] = ACTIONS(1717),
    [anon_sym_offsetof] = ACTIONS(1720),
    [anon_sym__Generic] = ACTIONS(1723),
    [anon_sym_asm] = ACTIONS(1726),
    [anon_sym___asm__] = ACTIONS(1726),
    [anon_sym___asm] = ACTIONS(2011),
    [sym_number_literal] = ACTIONS(1732),
    [anon_sym_L_SQUOTE] = ACTIONS(1735),
    [anon_sym_u_SQUOTE] = ACTIONS(1735),
    [anon_sym_U_SQUOTE] = ACTIONS(1735),
    [anon_sym_u8_SQUOTE] = ACTIONS(1735),
    [anon_sym_SQUOTE] = ACTIONS(1735),
    [anon_sym_AT] = ACTIONS(1738),
    [anon_sym_DQUOTE] = ACTIONS(1741),
    [anon_sym_L_DQUOTE] = ACTIONS(1741),
    [anon_sym_u_DQUOTE] = ACTIONS(1741),
    [anon_sym_U_DQUOTE] = ACTIONS(1741),
    [anon_sym_u8_DQUOTE] = ACTIONS(1741),
    [sym_true] = ACTIONS(1744),
    [sym_false] = ACTIONS(1744),
    [anon_sym_NULL] = ACTIONS(1747),
    [anon_sym_nullptr] = ACTIONS(1747),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(1750),
    [anon_sym___typeof] = ACTIONS(1750),
    [anon_sym_typeof] = ACTIONS(1750),
    [anon_sym_ATimport] = ACTIONS(1753),
    [aux_sym_preproc_undef_token1] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(1755),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(1755),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(1755),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE] = ACTIONS(1758),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_API_AVAILABLE] = ACTIONS(1758),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_API_DEPRECATED] = ACTIONS(1758),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(1758),
    [anon_sym___deprecated_msg] = ACTIONS(1758),
    [anon_sym___deprecated_enum_msg] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(1758),
    [anon_sym_ATprotocol] = ACTIONS(1753),
    [anon_sym_ATinterface] = ACTIONS(1753),
    [anon_sym_ATimplementation] = ACTIONS(1753),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1753),
    [anon_sym_ATtry] = ACTIONS(2014),
    [anon_sym___try] = ACTIONS(2017),
    [anon_sym_ATthrow] = ACTIONS(2020),
    [anon_sym_ATselector] = ACTIONS(1770),
    [anon_sym_ATavailable] = ACTIONS(1773),
    [anon_sym___builtin_available] = ACTIONS(1776),
    [anon_sym_va_arg] = ACTIONS(1779),
    [anon_sym_ATencode] = ACTIONS(1782),
    [anon_sym_ATsynchronized] = ACTIONS(2023),
    [anon_sym_BOOL] = ACTIONS(1788),
    [anon_sym_IMP] = ACTIONS(1788),
    [anon_sym_SEL] = ACTIONS(1788),
    [anon_sym_Class] = ACTIONS(1788),
    [anon_sym_id] = ACTIONS(1791),
  },
  [STATE(149)] = {
    [sym_declaration] = STATE(154),
    [sym_type_definition] = STATE(154),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4213),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_compound_statement] = STATE(154),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(154),
    [sym_labeled_statement] = STATE(154),
    [sym_expression_statement] = STATE(154),
    [sym_if_statement] = STATE(154),
    [sym_switch_statement] = STATE(154),
    [sym_while_statement] = STATE(154),
    [sym_do_statement] = STATE(154),
    [sym_for_statement] = STATE(154),
    [sym_return_statement] = STATE(154),
    [sym_break_statement] = STATE(154),
    [sym_continue_statement] = STATE(154),
    [sym_goto_statement] = STATE(154),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(154),
    [sym_throw_statement] = STATE(154),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(154),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(154),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(154),
    [sym_identifier] = ACTIONS(1922),
    [aux_sym_preproc_include_token1] = ACTIONS(1603),
    [aux_sym_preproc_include_token2] = ACTIONS(1603),
    [aux_sym_preproc_def_token1] = ACTIONS(1603),
    [aux_sym_preproc_if_token1] = ACTIONS(1603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1603),
    [sym_preproc_directive] = ACTIONS(1603),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(1924),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1603),
    [anon_sym___clrcall] = ACTIONS(1603),
    [anon_sym___stdcall] = ACTIONS(1603),
    [anon_sym___fastcall] = ACTIONS(1603),
    [anon_sym___thiscall] = ACTIONS(1603),
    [anon_sym___vectorcall] = ACTIONS(1603),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1605),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_else] = ACTIONS(1603),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(1603),
    [anon_sym_default] = ACTIONS(1603),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1605),
    [aux_sym_preproc_undef_token1] = ACTIONS(1603),
    [anon_sym_POUND] = ACTIONS(1603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1605),
    [anon_sym_ATinterface] = ACTIONS(1605),
    [anon_sym_ATimplementation] = ACTIONS(1605),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1605),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(150)] = {
    [sym_declaration] = STATE(152),
    [sym_type_definition] = STATE(152),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4212),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2898),
    [sym_compound_statement] = STATE(152),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(152),
    [sym_labeled_statement] = STATE(152),
    [sym_expression_statement] = STATE(152),
    [sym_if_statement] = STATE(152),
    [sym_switch_statement] = STATE(152),
    [sym_while_statement] = STATE(152),
    [sym_do_statement] = STATE(152),
    [sym_for_statement] = STATE(152),
    [sym_return_statement] = STATE(152),
    [sym_break_statement] = STATE(152),
    [sym_continue_statement] = STATE(152),
    [sym_goto_statement] = STATE(152),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(152),
    [sym_throw_statement] = STATE(152),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(152),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(152),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(152),
    [ts_builtin_sym_end] = ACTIONS(1597),
    [sym_identifier] = ACTIONS(1926),
    [aux_sym_preproc_include_token1] = ACTIONS(1595),
    [aux_sym_preproc_include_token2] = ACTIONS(1595),
    [aux_sym_preproc_def_token1] = ACTIONS(1595),
    [aux_sym_preproc_if_token1] = ACTIONS(1595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1595),
    [sym_preproc_directive] = ACTIONS(1595),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(1930),
    [anon_sym_typedef] = ACTIONS(1932),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1595),
    [anon_sym___clrcall] = ACTIONS(1595),
    [anon_sym___stdcall] = ACTIONS(1595),
    [anon_sym___fastcall] = ACTIONS(1595),
    [anon_sym___thiscall] = ACTIONS(1595),
    [anon_sym___vectorcall] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_else] = ACTIONS(1595),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_case] = ACTIONS(1595),
    [anon_sym_default] = ACTIONS(1595),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1597),
    [aux_sym_preproc_undef_token1] = ACTIONS(1595),
    [anon_sym_POUND] = ACTIONS(1595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1597),
    [anon_sym_ATinterface] = ACTIONS(1597),
    [anon_sym_ATimplementation] = ACTIONS(1597),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1597),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(151)] = {
    [sym_declaration] = STATE(149),
    [sym_type_definition] = STATE(149),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4213),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_compound_statement] = STATE(149),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(149),
    [sym_labeled_statement] = STATE(149),
    [sym_expression_statement] = STATE(149),
    [sym_if_statement] = STATE(149),
    [sym_switch_statement] = STATE(149),
    [sym_while_statement] = STATE(149),
    [sym_do_statement] = STATE(149),
    [sym_for_statement] = STATE(149),
    [sym_return_statement] = STATE(149),
    [sym_break_statement] = STATE(149),
    [sym_continue_statement] = STATE(149),
    [sym_goto_statement] = STATE(149),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(149),
    [sym_throw_statement] = STATE(149),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(149),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(149),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(149),
    [sym_identifier] = ACTIONS(1922),
    [aux_sym_preproc_include_token1] = ACTIONS(1595),
    [aux_sym_preproc_include_token2] = ACTIONS(1595),
    [aux_sym_preproc_def_token1] = ACTIONS(1595),
    [aux_sym_preproc_if_token1] = ACTIONS(1595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1595),
    [sym_preproc_directive] = ACTIONS(1595),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(1924),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1595),
    [anon_sym___clrcall] = ACTIONS(1595),
    [anon_sym___stdcall] = ACTIONS(1595),
    [anon_sym___fastcall] = ACTIONS(1595),
    [anon_sym___thiscall] = ACTIONS(1595),
    [anon_sym___vectorcall] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1597),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_else] = ACTIONS(1595),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(1595),
    [anon_sym_default] = ACTIONS(1595),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1597),
    [aux_sym_preproc_undef_token1] = ACTIONS(1595),
    [anon_sym_POUND] = ACTIONS(1595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1597),
    [anon_sym_ATinterface] = ACTIONS(1597),
    [anon_sym_ATimplementation] = ACTIONS(1597),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1597),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(152)] = {
    [sym_declaration] = STATE(148),
    [sym_type_definition] = STATE(148),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4212),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2898),
    [sym_compound_statement] = STATE(148),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(148),
    [sym_labeled_statement] = STATE(148),
    [sym_expression_statement] = STATE(148),
    [sym_if_statement] = STATE(148),
    [sym_switch_statement] = STATE(148),
    [sym_while_statement] = STATE(148),
    [sym_do_statement] = STATE(148),
    [sym_for_statement] = STATE(148),
    [sym_return_statement] = STATE(148),
    [sym_break_statement] = STATE(148),
    [sym_continue_statement] = STATE(148),
    [sym_goto_statement] = STATE(148),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(148),
    [sym_throw_statement] = STATE(148),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(148),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(148),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(148),
    [ts_builtin_sym_end] = ACTIONS(1605),
    [sym_identifier] = ACTIONS(1926),
    [aux_sym_preproc_include_token1] = ACTIONS(1603),
    [aux_sym_preproc_include_token2] = ACTIONS(1603),
    [aux_sym_preproc_def_token1] = ACTIONS(1603),
    [aux_sym_preproc_if_token1] = ACTIONS(1603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1603),
    [sym_preproc_directive] = ACTIONS(1603),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(1930),
    [anon_sym_typedef] = ACTIONS(1932),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1603),
    [anon_sym___clrcall] = ACTIONS(1603),
    [anon_sym___stdcall] = ACTIONS(1603),
    [anon_sym___fastcall] = ACTIONS(1603),
    [anon_sym___thiscall] = ACTIONS(1603),
    [anon_sym___vectorcall] = ACTIONS(1603),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_else] = ACTIONS(1603),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_case] = ACTIONS(1603),
    [anon_sym_default] = ACTIONS(1603),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1605),
    [aux_sym_preproc_undef_token1] = ACTIONS(1603),
    [anon_sym_POUND] = ACTIONS(1603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1605),
    [anon_sym_ATinterface] = ACTIONS(1605),
    [anon_sym_ATimplementation] = ACTIONS(1605),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1605),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(153)] = {
    [sym_declaration] = STATE(154),
    [sym_type_definition] = STATE(154),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4213),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_compound_statement] = STATE(154),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(154),
    [sym_labeled_statement] = STATE(154),
    [sym_expression_statement] = STATE(154),
    [sym_if_statement] = STATE(154),
    [sym_switch_statement] = STATE(154),
    [sym_while_statement] = STATE(154),
    [sym_do_statement] = STATE(154),
    [sym_for_statement] = STATE(154),
    [sym_return_statement] = STATE(154),
    [sym_break_statement] = STATE(154),
    [sym_continue_statement] = STATE(154),
    [sym_goto_statement] = STATE(154),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(154),
    [sym_throw_statement] = STATE(154),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(154),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(154),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(154),
    [sym_identifier] = ACTIONS(1922),
    [aux_sym_preproc_include_token1] = ACTIONS(1581),
    [aux_sym_preproc_include_token2] = ACTIONS(1581),
    [aux_sym_preproc_def_token1] = ACTIONS(1581),
    [aux_sym_preproc_if_token1] = ACTIONS(1581),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1581),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1581),
    [sym_preproc_directive] = ACTIONS(1581),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(1924),
    [anon_sym_typedef] = ACTIONS(940),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(1581),
    [anon_sym___clrcall] = ACTIONS(1581),
    [anon_sym___stdcall] = ACTIONS(1581),
    [anon_sym___fastcall] = ACTIONS(1581),
    [anon_sym___thiscall] = ACTIONS(1581),
    [anon_sym___vectorcall] = ACTIONS(1581),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_RBRACE] = ACTIONS(1591),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_else] = ACTIONS(1581),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(1581),
    [anon_sym_default] = ACTIONS(1581),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATimport] = ACTIONS(1591),
    [aux_sym_preproc_undef_token1] = ACTIONS(1581),
    [anon_sym_POUND] = ACTIONS(1581),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(1591),
    [anon_sym_ATinterface] = ACTIONS(1591),
    [anon_sym_ATimplementation] = ACTIONS(1591),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1591),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(154)] = {
    [sym_declaration] = STATE(154),
    [sym_type_definition] = STATE(154),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4213),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2840),
    [sym_compound_statement] = STATE(154),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(154),
    [sym_labeled_statement] = STATE(154),
    [sym_expression_statement] = STATE(154),
    [sym_if_statement] = STATE(154),
    [sym_switch_statement] = STATE(154),
    [sym_while_statement] = STATE(154),
    [sym_do_statement] = STATE(154),
    [sym_for_statement] = STATE(154),
    [sym_return_statement] = STATE(154),
    [sym_break_statement] = STATE(154),
    [sym_continue_statement] = STATE(154),
    [sym_goto_statement] = STATE(154),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(154),
    [sym_throw_statement] = STATE(154),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(154),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(154),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(154),
    [sym_identifier] = ACTIONS(2026),
    [aux_sym_preproc_include_token1] = ACTIONS(1610),
    [aux_sym_preproc_include_token2] = ACTIONS(1610),
    [aux_sym_preproc_def_token1] = ACTIONS(1610),
    [aux_sym_preproc_if_token1] = ACTIONS(1610),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1610),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1610),
    [sym_preproc_directive] = ACTIONS(1610),
    [anon_sym_LPAREN2] = ACTIONS(1612),
    [anon_sym_BANG] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_DASH] = ACTIONS(1618),
    [anon_sym_PLUS] = ACTIONS(1618),
    [anon_sym_STAR] = ACTIONS(1621),
    [anon_sym_CARET] = ACTIONS(1624),
    [anon_sym_AMP] = ACTIONS(1621),
    [anon_sym_SEMI] = ACTIONS(2029),
    [anon_sym___extension__] = ACTIONS(2032),
    [anon_sym_typedef] = ACTIONS(2035),
    [anon_sym_extern] = ACTIONS(1636),
    [anon_sym___attribute__] = ACTIONS(1639),
    [anon_sym___attribute] = ACTIONS(1639),
    [anon_sym_noreturn] = ACTIONS(1642),
    [anon_sym_LBRACK] = ACTIONS(1645),
    [anon_sym___declspec] = ACTIONS(1648),
    [anon_sym___cdecl] = ACTIONS(1610),
    [anon_sym___clrcall] = ACTIONS(1610),
    [anon_sym___stdcall] = ACTIONS(1610),
    [anon_sym___fastcall] = ACTIONS(1610),
    [anon_sym___thiscall] = ACTIONS(1610),
    [anon_sym___vectorcall] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(2038),
    [anon_sym_RBRACE] = ACTIONS(1753),
    [anon_sym_signed] = ACTIONS(1654),
    [anon_sym_unsigned] = ACTIONS(1654),
    [anon_sym_long] = ACTIONS(1654),
    [anon_sym_short] = ACTIONS(1654),
    [anon_sym_static] = ACTIONS(1636),
    [anon_sym_ATautoreleasepool] = ACTIONS(2041),
    [anon_sym_auto] = ACTIONS(1636),
    [anon_sym_register] = ACTIONS(1636),
    [anon_sym_inline] = ACTIONS(1636),
    [anon_sym___inline] = ACTIONS(1636),
    [anon_sym___inline__] = ACTIONS(1636),
    [anon_sym___forceinline] = ACTIONS(1636),
    [anon_sym_thread_local] = ACTIONS(1636),
    [anon_sym___thread] = ACTIONS(1636),
    [anon_sym_CG_EXTERN] = ACTIONS(1636),
    [anon_sym_CG_INLINE] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(1636),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(1636),
    [anon_sym_IBOutlet] = ACTIONS(1636),
    [anon_sym_IBInspectable] = ACTIONS(1636),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(1636),
    [anon_sym_NS_INLINE] = ACTIONS(1636),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(1636),
    [anon_sym_OBJC_EXPORT] = ACTIONS(1636),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(1636),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(1636),
    [anon_sym_const] = ACTIONS(1660),
    [anon_sym_constexpr] = ACTIONS(1642),
    [anon_sym_volatile] = ACTIONS(1642),
    [anon_sym_restrict] = ACTIONS(1642),
    [anon_sym___restrict__] = ACTIONS(1642),
    [anon_sym__Atomic] = ACTIONS(1642),
    [anon_sym__Noreturn] = ACTIONS(1642),
    [anon_sym__Nonnull] = ACTIONS(1642),
    [anon_sym_nullable] = ACTIONS(1642),
    [anon_sym__Complex] = ACTIONS(1642),
    [anon_sym__Nullable] = ACTIONS(1642),
    [anon_sym__Nullable_result] = ACTIONS(1642),
    [anon_sym__Null_unspecified] = ACTIONS(1642),
    [anon_sym___autoreleasing] = ACTIONS(1642),
    [anon_sym___block] = ACTIONS(1642),
    [anon_sym___bridge] = ACTIONS(1642),
    [anon_sym___bridge_retained] = ACTIONS(1642),
    [anon_sym___bridge_transfer] = ACTIONS(1642),
    [anon_sym___complex] = ACTIONS(1642),
    [anon_sym___const] = ACTIONS(1642),
    [anon_sym___imag] = ACTIONS(1663),
    [anon_sym___kindof] = ACTIONS(1642),
    [anon_sym___nonnull] = ACTIONS(1642),
    [anon_sym___nullable] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(1642),
    [anon_sym___real] = ACTIONS(1663),
    [anon_sym___strong] = ACTIONS(1642),
    [anon_sym___unsafe_unretained] = ACTIONS(1642),
    [anon_sym___unused] = ACTIONS(1642),
    [anon_sym___weak] = ACTIONS(1642),
    [anon_sym_alignas] = ACTIONS(1666),
    [anon_sym__Alignas] = ACTIONS(1666),
    [sym_primitive_type] = ACTIONS(1669),
    [anon_sym_enum] = ACTIONS(1672),
    [anon_sym_struct] = ACTIONS(1675),
    [anon_sym_union] = ACTIONS(1678),
    [anon_sym_if] = ACTIONS(2044),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_switch] = ACTIONS(2047),
    [anon_sym_case] = ACTIONS(1610),
    [anon_sym_default] = ACTIONS(1610),
    [anon_sym_while] = ACTIONS(2050),
    [anon_sym_do] = ACTIONS(2053),
    [anon_sym_for] = ACTIONS(2056),
    [anon_sym_in] = ACTIONS(1696),
    [anon_sym_return] = ACTIONS(2059),
    [anon_sym_break] = ACTIONS(2062),
    [anon_sym_continue] = ACTIONS(2065),
    [anon_sym_goto] = ACTIONS(2068),
    [anon_sym_DASH_DASH] = ACTIONS(1711),
    [anon_sym_PLUS_PLUS] = ACTIONS(1711),
    [anon_sym_sizeof] = ACTIONS(1714),
    [anon_sym___alignof__] = ACTIONS(1717),
    [anon_sym___alignof] = ACTIONS(1717),
    [anon_sym__alignof] = ACTIONS(1717),
    [anon_sym_alignof] = ACTIONS(1717),
    [anon_sym__Alignof] = ACTIONS(1717),
    [anon_sym_offsetof] = ACTIONS(1720),
    [anon_sym__Generic] = ACTIONS(1723),
    [anon_sym_asm] = ACTIONS(1726),
    [anon_sym___asm__] = ACTIONS(1726),
    [anon_sym___asm] = ACTIONS(2071),
    [sym_number_literal] = ACTIONS(1732),
    [anon_sym_L_SQUOTE] = ACTIONS(1735),
    [anon_sym_u_SQUOTE] = ACTIONS(1735),
    [anon_sym_U_SQUOTE] = ACTIONS(1735),
    [anon_sym_u8_SQUOTE] = ACTIONS(1735),
    [anon_sym_SQUOTE] = ACTIONS(1735),
    [anon_sym_AT] = ACTIONS(1738),
    [anon_sym_DQUOTE] = ACTIONS(1741),
    [anon_sym_L_DQUOTE] = ACTIONS(1741),
    [anon_sym_u_DQUOTE] = ACTIONS(1741),
    [anon_sym_U_DQUOTE] = ACTIONS(1741),
    [anon_sym_u8_DQUOTE] = ACTIONS(1741),
    [sym_true] = ACTIONS(1744),
    [sym_false] = ACTIONS(1744),
    [anon_sym_NULL] = ACTIONS(1747),
    [anon_sym_nullptr] = ACTIONS(1747),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(1750),
    [anon_sym___typeof] = ACTIONS(1750),
    [anon_sym_typeof] = ACTIONS(1750),
    [anon_sym_ATimport] = ACTIONS(1753),
    [aux_sym_preproc_undef_token1] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(1755),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(1755),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(1755),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE] = ACTIONS(1758),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_API_AVAILABLE] = ACTIONS(1758),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_API_DEPRECATED] = ACTIONS(1758),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(1758),
    [anon_sym___deprecated_msg] = ACTIONS(1758),
    [anon_sym___deprecated_enum_msg] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(1758),
    [anon_sym_ATprotocol] = ACTIONS(1753),
    [anon_sym_ATinterface] = ACTIONS(1753),
    [anon_sym_ATimplementation] = ACTIONS(1753),
    [anon_sym_ATcompatibility_alias] = ACTIONS(1753),
    [anon_sym_ATtry] = ACTIONS(2074),
    [anon_sym___try] = ACTIONS(2077),
    [anon_sym_ATthrow] = ACTIONS(2080),
    [anon_sym_ATselector] = ACTIONS(1770),
    [anon_sym_ATavailable] = ACTIONS(1773),
    [anon_sym___builtin_available] = ACTIONS(1776),
    [anon_sym_va_arg] = ACTIONS(1779),
    [anon_sym_ATencode] = ACTIONS(1782),
    [anon_sym_ATsynchronized] = ACTIONS(2083),
    [anon_sym_BOOL] = ACTIONS(1788),
    [anon_sym_IMP] = ACTIONS(1788),
    [anon_sym_SEL] = ACTIONS(1788),
    [anon_sym_Class] = ACTIONS(1788),
    [anon_sym_id] = ACTIONS(1791),
  },
  [STATE(155)] = {
    [sym_declaration] = STATE(765),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4213),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(765),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(1922),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(156)] = {
    [sym_declaration] = STATE(156),
    [sym_type_definition] = STATE(156),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4211),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2874),
    [sym_compound_statement] = STATE(156),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(156),
    [sym_labeled_statement] = STATE(156),
    [sym_expression_statement] = STATE(156),
    [sym_if_statement] = STATE(156),
    [sym_switch_statement] = STATE(156),
    [sym_while_statement] = STATE(156),
    [sym_do_statement] = STATE(156),
    [sym_for_statement] = STATE(156),
    [sym_return_statement] = STATE(156),
    [sym_break_statement] = STATE(156),
    [sym_continue_statement] = STATE(156),
    [sym_goto_statement] = STATE(156),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(156),
    [sym_throw_statement] = STATE(156),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(156),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(156),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(156),
    [sym_identifier] = ACTIONS(2088),
    [anon_sym_LPAREN2] = ACTIONS(1612),
    [anon_sym_BANG] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_DASH] = ACTIONS(1618),
    [anon_sym_PLUS] = ACTIONS(1618),
    [anon_sym_STAR] = ACTIONS(1621),
    [anon_sym_CARET] = ACTIONS(1624),
    [anon_sym_AMP] = ACTIONS(1621),
    [anon_sym_SEMI] = ACTIONS(2091),
    [anon_sym___extension__] = ACTIONS(2094),
    [anon_sym_typedef] = ACTIONS(2097),
    [anon_sym_extern] = ACTIONS(1636),
    [anon_sym___attribute__] = ACTIONS(1639),
    [anon_sym___attribute] = ACTIONS(1639),
    [anon_sym_noreturn] = ACTIONS(1642),
    [anon_sym_LBRACK] = ACTIONS(1645),
    [anon_sym___declspec] = ACTIONS(1648),
    [anon_sym_LBRACE] = ACTIONS(2100),
    [anon_sym_signed] = ACTIONS(1654),
    [anon_sym_unsigned] = ACTIONS(1654),
    [anon_sym_long] = ACTIONS(1654),
    [anon_sym_short] = ACTIONS(1654),
    [anon_sym_static] = ACTIONS(1636),
    [anon_sym_ATautoreleasepool] = ACTIONS(2103),
    [anon_sym_auto] = ACTIONS(1636),
    [anon_sym_register] = ACTIONS(1636),
    [anon_sym_inline] = ACTIONS(1636),
    [anon_sym___inline] = ACTIONS(1636),
    [anon_sym___inline__] = ACTIONS(1636),
    [anon_sym___forceinline] = ACTIONS(1636),
    [anon_sym_thread_local] = ACTIONS(1636),
    [anon_sym___thread] = ACTIONS(1636),
    [anon_sym_CG_EXTERN] = ACTIONS(1636),
    [anon_sym_CG_INLINE] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(1636),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(1636),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(1636),
    [anon_sym_IBOutlet] = ACTIONS(1636),
    [anon_sym_IBInspectable] = ACTIONS(1636),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(1636),
    [anon_sym_NS_INLINE] = ACTIONS(1636),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(1636),
    [anon_sym_OBJC_EXPORT] = ACTIONS(1636),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(1636),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(1636),
    [anon_sym_const] = ACTIONS(1660),
    [anon_sym_constexpr] = ACTIONS(1642),
    [anon_sym_volatile] = ACTIONS(1642),
    [anon_sym_restrict] = ACTIONS(1642),
    [anon_sym___restrict__] = ACTIONS(1642),
    [anon_sym__Atomic] = ACTIONS(1642),
    [anon_sym__Noreturn] = ACTIONS(1642),
    [anon_sym__Nonnull] = ACTIONS(1642),
    [anon_sym_nullable] = ACTIONS(1642),
    [anon_sym__Complex] = ACTIONS(1642),
    [anon_sym__Nullable] = ACTIONS(1642),
    [anon_sym__Nullable_result] = ACTIONS(1642),
    [anon_sym__Null_unspecified] = ACTIONS(1642),
    [anon_sym___autoreleasing] = ACTIONS(1642),
    [anon_sym___block] = ACTIONS(1642),
    [anon_sym___bridge] = ACTIONS(1642),
    [anon_sym___bridge_retained] = ACTIONS(1642),
    [anon_sym___bridge_transfer] = ACTIONS(1642),
    [anon_sym___complex] = ACTIONS(1642),
    [anon_sym___const] = ACTIONS(1642),
    [anon_sym___imag] = ACTIONS(1663),
    [anon_sym___kindof] = ACTIONS(1642),
    [anon_sym___nonnull] = ACTIONS(1642),
    [anon_sym___nullable] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(1642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(1642),
    [anon_sym___real] = ACTIONS(1663),
    [anon_sym___strong] = ACTIONS(1642),
    [anon_sym___unsafe_unretained] = ACTIONS(1642),
    [anon_sym___unused] = ACTIONS(1642),
    [anon_sym___weak] = ACTIONS(1642),
    [anon_sym_alignas] = ACTIONS(1666),
    [anon_sym__Alignas] = ACTIONS(1666),
    [sym_primitive_type] = ACTIONS(1669),
    [anon_sym_enum] = ACTIONS(1672),
    [anon_sym_struct] = ACTIONS(1675),
    [anon_sym_union] = ACTIONS(1678),
    [anon_sym_if] = ACTIONS(2106),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_switch] = ACTIONS(2109),
    [anon_sym_while] = ACTIONS(2112),
    [anon_sym_do] = ACTIONS(2115),
    [anon_sym_for] = ACTIONS(2118),
    [anon_sym_in] = ACTIONS(1696),
    [anon_sym_return] = ACTIONS(2121),
    [anon_sym_break] = ACTIONS(2124),
    [anon_sym_continue] = ACTIONS(2127),
    [anon_sym_goto] = ACTIONS(2130),
    [anon_sym_DASH_DASH] = ACTIONS(1711),
    [anon_sym_PLUS_PLUS] = ACTIONS(1711),
    [anon_sym_sizeof] = ACTIONS(1714),
    [anon_sym___alignof__] = ACTIONS(1717),
    [anon_sym___alignof] = ACTIONS(1717),
    [anon_sym__alignof] = ACTIONS(1717),
    [anon_sym_alignof] = ACTIONS(1717),
    [anon_sym__Alignof] = ACTIONS(1717),
    [anon_sym_offsetof] = ACTIONS(1720),
    [anon_sym__Generic] = ACTIONS(1723),
    [anon_sym_asm] = ACTIONS(1726),
    [anon_sym___asm__] = ACTIONS(1726),
    [anon_sym___asm] = ACTIONS(2133),
    [sym_number_literal] = ACTIONS(1732),
    [anon_sym_L_SQUOTE] = ACTIONS(1735),
    [anon_sym_u_SQUOTE] = ACTIONS(1735),
    [anon_sym_U_SQUOTE] = ACTIONS(1735),
    [anon_sym_u8_SQUOTE] = ACTIONS(1735),
    [anon_sym_SQUOTE] = ACTIONS(1735),
    [anon_sym_AT] = ACTIONS(1738),
    [anon_sym_DQUOTE] = ACTIONS(1741),
    [anon_sym_L_DQUOTE] = ACTIONS(1741),
    [anon_sym_u_DQUOTE] = ACTIONS(1741),
    [anon_sym_U_DQUOTE] = ACTIONS(1741),
    [anon_sym_u8_DQUOTE] = ACTIONS(1741),
    [sym_true] = ACTIONS(1744),
    [sym_false] = ACTIONS(1744),
    [anon_sym_NULL] = ACTIONS(1747),
    [anon_sym_nullptr] = ACTIONS(1747),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(1750),
    [anon_sym___typeof] = ACTIONS(1750),
    [anon_sym_typeof] = ACTIONS(1750),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(1755),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(1755),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(1755),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(1755),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(1755),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(1755),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE] = ACTIONS(1758),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_API_AVAILABLE] = ACTIONS(1758),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_API_DEPRECATED] = ACTIONS(1758),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(1758),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(1758),
    [anon_sym___deprecated_msg] = ACTIONS(1758),
    [anon_sym___deprecated_enum_msg] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(1758),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(1758),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(1758),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(1758),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(1758),
    [anon_sym_ATtry] = ACTIONS(2136),
    [anon_sym___try] = ACTIONS(2139),
    [anon_sym_ATthrow] = ACTIONS(2142),
    [anon_sym_ATselector] = ACTIONS(1770),
    [anon_sym_ATavailable] = ACTIONS(1773),
    [anon_sym___builtin_available] = ACTIONS(1776),
    [anon_sym_va_arg] = ACTIONS(1779),
    [anon_sym_ATencode] = ACTIONS(1782),
    [anon_sym_ATsynchronized] = ACTIONS(2145),
    [anon_sym_BOOL] = ACTIONS(1788),
    [anon_sym_IMP] = ACTIONS(1788),
    [anon_sym_SEL] = ACTIONS(1788),
    [anon_sym_Class] = ACTIONS(1788),
    [anon_sym_id] = ACTIONS(1791),
  },
  [STATE(157)] = {
    [sym_declaration] = STATE(966),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4233),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(955),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(955),
    [sym_statement] = STATE(966),
    [sym_labeled_statement] = STATE(955),
    [sym_expression_statement] = STATE(955),
    [sym_if_statement] = STATE(955),
    [sym_switch_statement] = STATE(955),
    [sym_case_statement] = STATE(955),
    [sym_while_statement] = STATE(955),
    [sym_do_statement] = STATE(955),
    [sym_for_statement] = STATE(955),
    [sym_return_statement] = STATE(955),
    [sym_break_statement] = STATE(955),
    [sym_continue_statement] = STATE(955),
    [sym_goto_statement] = STATE(955),
    [sym_expression] = STATE(4419),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6212),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(955),
    [sym_throw_statement] = STATE(955),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(955),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(955),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2451),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2148),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2150),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(69),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(77),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2152),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(2154),
    [anon_sym___try] = ACTIONS(2156),
    [anon_sym_ATthrow] = ACTIONS(2158),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2160),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(158)] = {
    [sym_declaration] = STATE(249),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4231),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(232),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(232),
    [sym_statement] = STATE(249),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(1579),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(159)] = {
    [sym_declaration] = STATE(966),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4233),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(825),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(966),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2162),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(160)] = {
    [sym_declaration] = STATE(164),
    [sym_type_definition] = STATE(164),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4211),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2874),
    [sym_compound_statement] = STATE(164),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(164),
    [sym_labeled_statement] = STATE(164),
    [sym_expression_statement] = STATE(164),
    [sym_if_statement] = STATE(164),
    [sym_switch_statement] = STATE(164),
    [sym_while_statement] = STATE(164),
    [sym_do_statement] = STATE(164),
    [sym_for_statement] = STATE(164),
    [sym_return_statement] = STATE(164),
    [sym_break_statement] = STATE(164),
    [sym_continue_statement] = STATE(164),
    [sym_goto_statement] = STATE(164),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(164),
    [sym_throw_statement] = STATE(164),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(164),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(164),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(164),
    [sym_identifier] = ACTIONS(2180),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2184),
    [anon_sym_typedef] = ACTIONS(2186),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_else] = ACTIONS(1595),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(161)] = {
    [sym_declaration] = STATE(727),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4212),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(690),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(690),
    [sym_statement] = STATE(727),
    [sym_labeled_statement] = STATE(690),
    [sym_expression_statement] = STATE(690),
    [sym_if_statement] = STATE(690),
    [sym_switch_statement] = STATE(690),
    [sym_case_statement] = STATE(690),
    [sym_while_statement] = STATE(690),
    [sym_do_statement] = STATE(690),
    [sym_for_statement] = STATE(690),
    [sym_return_statement] = STATE(690),
    [sym_break_statement] = STATE(690),
    [sym_continue_statement] = STATE(690),
    [sym_goto_statement] = STATE(690),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(690),
    [sym_throw_statement] = STATE(690),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(690),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(690),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(1926),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(162)] = {
    [sym_declaration] = STATE(413),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4202),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(439),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(439),
    [sym_statement] = STATE(413),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(1794),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(163)] = {
    [sym_declaration] = STATE(167),
    [sym_type_definition] = STATE(167),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4211),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2874),
    [sym_compound_statement] = STATE(167),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(167),
    [sym_labeled_statement] = STATE(167),
    [sym_expression_statement] = STATE(167),
    [sym_if_statement] = STATE(167),
    [sym_switch_statement] = STATE(167),
    [sym_while_statement] = STATE(167),
    [sym_do_statement] = STATE(167),
    [sym_for_statement] = STATE(167),
    [sym_return_statement] = STATE(167),
    [sym_break_statement] = STATE(167),
    [sym_continue_statement] = STATE(167),
    [sym_goto_statement] = STATE(167),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(167),
    [sym_throw_statement] = STATE(167),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(167),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(167),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(167),
    [sym_identifier] = ACTIONS(2180),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2184),
    [anon_sym_typedef] = ACTIONS(2186),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_else] = ACTIONS(1599),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(164)] = {
    [sym_declaration] = STATE(156),
    [sym_type_definition] = STATE(156),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4211),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2874),
    [sym_compound_statement] = STATE(156),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(156),
    [sym_labeled_statement] = STATE(156),
    [sym_expression_statement] = STATE(156),
    [sym_if_statement] = STATE(156),
    [sym_switch_statement] = STATE(156),
    [sym_while_statement] = STATE(156),
    [sym_do_statement] = STATE(156),
    [sym_for_statement] = STATE(156),
    [sym_return_statement] = STATE(156),
    [sym_break_statement] = STATE(156),
    [sym_continue_statement] = STATE(156),
    [sym_goto_statement] = STATE(156),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(156),
    [sym_throw_statement] = STATE(156),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(156),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(156),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(156),
    [sym_identifier] = ACTIONS(2180),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2184),
    [anon_sym_typedef] = ACTIONS(2186),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_else] = ACTIONS(1603),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(165)] = {
    [sym_declaration] = STATE(579),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4228),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(535),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(535),
    [sym_statement] = STATE(579),
    [sym_labeled_statement] = STATE(535),
    [sym_expression_statement] = STATE(535),
    [sym_if_statement] = STATE(535),
    [sym_switch_statement] = STATE(535),
    [sym_case_statement] = STATE(535),
    [sym_while_statement] = STATE(535),
    [sym_do_statement] = STATE(535),
    [sym_for_statement] = STATE(535),
    [sym_return_statement] = STATE(535),
    [sym_break_statement] = STATE(535),
    [sym_continue_statement] = STATE(535),
    [sym_goto_statement] = STATE(535),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(535),
    [sym_throw_statement] = STATE(535),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(535),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(535),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(1858),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(878),
    [anon_sym_default] = ACTIONS(880),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(166)] = {
    [sym_declaration] = STATE(1135),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4211),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(1159),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(1159),
    [sym_statement] = STATE(1135),
    [sym_labeled_statement] = STATE(1159),
    [sym_expression_statement] = STATE(1159),
    [sym_if_statement] = STATE(1159),
    [sym_switch_statement] = STATE(1159),
    [sym_case_statement] = STATE(1159),
    [sym_while_statement] = STATE(1159),
    [sym_do_statement] = STATE(1159),
    [sym_for_statement] = STATE(1159),
    [sym_return_statement] = STATE(1159),
    [sym_break_statement] = STATE(1159),
    [sym_continue_statement] = STATE(1159),
    [sym_goto_statement] = STATE(1159),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(1159),
    [sym_throw_statement] = STATE(1159),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(1159),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(1159),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2180),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(167)] = {
    [sym_declaration] = STATE(156),
    [sym_type_definition] = STATE(156),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4211),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(1123),
    [sym_ms_declspec_modifier] = STATE(2874),
    [sym_compound_statement] = STATE(156),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_attributed_statement] = STATE(156),
    [sym_labeled_statement] = STATE(156),
    [sym_expression_statement] = STATE(156),
    [sym_if_statement] = STATE(156),
    [sym_switch_statement] = STATE(156),
    [sym_while_statement] = STATE(156),
    [sym_do_statement] = STATE(156),
    [sym_for_statement] = STATE(156),
    [sym_return_statement] = STATE(156),
    [sym_break_statement] = STATE(156),
    [sym_continue_statement] = STATE(156),
    [sym_goto_statement] = STATE(156),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_try_statement] = STATE(156),
    [sym_throw_statement] = STATE(156),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(156),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(156),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_case_statement_repeat1] = STATE(156),
    [sym_identifier] = ACTIONS(2180),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2184),
    [anon_sym_typedef] = ACTIONS(2186),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(1585),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_else] = ACTIONS(1581),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(168)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2220),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(2224),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2220),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2220),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2220),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2220),
    [anon_sym_GT_GT] = ACTIONS(2220),
    [anon_sym_SEMI] = ACTIONS(2222),
    [anon_sym___extension__] = ACTIONS(2220),
    [anon_sym_extern] = ACTIONS(2220),
    [anon_sym___attribute__] = ACTIONS(2220),
    [anon_sym___attribute] = ACTIONS(2220),
    [anon_sym_noreturn] = ACTIONS(2220),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym___declspec] = ACTIONS(2220),
    [anon_sym_LBRACE] = ACTIONS(2222),
    [anon_sym_signed] = ACTIONS(2220),
    [anon_sym_unsigned] = ACTIONS(2220),
    [anon_sym_long] = ACTIONS(2220),
    [anon_sym_short] = ACTIONS(2220),
    [anon_sym_static] = ACTIONS(2220),
    [anon_sym_EQ] = ACTIONS(2220),
    [anon_sym_ATautoreleasepool] = ACTIONS(2222),
    [anon_sym_auto] = ACTIONS(2220),
    [anon_sym_register] = ACTIONS(2220),
    [anon_sym_inline] = ACTIONS(2220),
    [anon_sym___inline] = ACTIONS(2220),
    [anon_sym___inline__] = ACTIONS(2220),
    [anon_sym___forceinline] = ACTIONS(2220),
    [anon_sym_thread_local] = ACTIONS(2220),
    [anon_sym___thread] = ACTIONS(2220),
    [anon_sym_CG_EXTERN] = ACTIONS(2220),
    [anon_sym_CG_INLINE] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2220),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2220),
    [anon_sym_IBOutlet] = ACTIONS(2220),
    [anon_sym_IBInspectable] = ACTIONS(2220),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2220),
    [anon_sym_NS_INLINE] = ACTIONS(2220),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2220),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2220),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2220),
    [anon_sym_const] = ACTIONS(2220),
    [anon_sym_constexpr] = ACTIONS(2220),
    [anon_sym_volatile] = ACTIONS(2220),
    [anon_sym_restrict] = ACTIONS(2220),
    [anon_sym___restrict__] = ACTIONS(2220),
    [anon_sym__Atomic] = ACTIONS(2220),
    [anon_sym__Noreturn] = ACTIONS(2220),
    [anon_sym__Nonnull] = ACTIONS(2220),
    [anon_sym_nullable] = ACTIONS(2220),
    [anon_sym__Complex] = ACTIONS(2220),
    [anon_sym__Nullable] = ACTIONS(2220),
    [anon_sym__Nullable_result] = ACTIONS(2220),
    [anon_sym__Null_unspecified] = ACTIONS(2220),
    [anon_sym___autoreleasing] = ACTIONS(2220),
    [anon_sym___block] = ACTIONS(2220),
    [anon_sym___bridge] = ACTIONS(2220),
    [anon_sym___bridge_retained] = ACTIONS(2220),
    [anon_sym___bridge_transfer] = ACTIONS(2220),
    [anon_sym___complex] = ACTIONS(2220),
    [anon_sym___const] = ACTIONS(2220),
    [anon_sym___imag] = ACTIONS(2220),
    [anon_sym___kindof] = ACTIONS(2220),
    [anon_sym___nonnull] = ACTIONS(2220),
    [anon_sym___nullable] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2220),
    [anon_sym___real] = ACTIONS(2220),
    [anon_sym___strong] = ACTIONS(2220),
    [anon_sym___unsafe_unretained] = ACTIONS(2220),
    [anon_sym___unused] = ACTIONS(2220),
    [anon_sym___weak] = ACTIONS(2220),
    [anon_sym_alignas] = ACTIONS(2220),
    [anon_sym__Alignas] = ACTIONS(2220),
    [sym_primitive_type] = ACTIONS(2220),
    [anon_sym_enum] = ACTIONS(2220),
    [anon_sym_struct] = ACTIONS(2220),
    [anon_sym_union] = ACTIONS(2220),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_STAR_EQ] = ACTIONS(2222),
    [anon_sym_SLASH_EQ] = ACTIONS(2222),
    [anon_sym_PERCENT_EQ] = ACTIONS(2222),
    [anon_sym_PLUS_EQ] = ACTIONS(2222),
    [anon_sym_DASH_EQ] = ACTIONS(2222),
    [anon_sym_LT_LT_EQ] = ACTIONS(2222),
    [anon_sym_GT_GT_EQ] = ACTIONS(2222),
    [anon_sym_AMP_EQ] = ACTIONS(2222),
    [anon_sym_CARET_EQ] = ACTIONS(2222),
    [anon_sym_PIPE_EQ] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2220),
    [anon_sym___typeof] = ACTIONS(2220),
    [anon_sym_typeof] = ACTIONS(2220),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2220),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2220),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2220),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_API_AVAILABLE] = ACTIONS(2220),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_API_DEPRECATED] = ACTIONS(2220),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym___deprecated_msg] = ACTIONS(2220),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2220),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_BOOL] = ACTIONS(2220),
    [anon_sym_IMP] = ACTIONS(2220),
    [anon_sym_SEL] = ACTIONS(2220),
    [anon_sym_Class] = ACTIONS(2220),
    [anon_sym_id] = ACTIONS(2220),
  },
  [STATE(169)] = {
    [sym_declaration] = STATE(3366),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4227),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__for_statement_body] = STATE(6371),
    [sym_expression] = STATE(4439),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6747),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2262),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2264),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2266),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(170)] = {
    [sym_declaration] = STATE(3366),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4235),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__for_statement_body] = STATE(6069),
    [sym_expression] = STATE(4439),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6747),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2268),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2264),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2266),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(171)] = {
    [sym_declaration] = STATE(3366),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4230),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__for_statement_body] = STATE(6688),
    [sym_expression] = STATE(4439),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6747),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2270),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2264),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2266),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(172)] = {
    [sym_declaration] = STATE(3366),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4216),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__for_statement_body] = STATE(6356),
    [sym_expression] = STATE(4439),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6747),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2272),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2264),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2266),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(173)] = {
    [sym_declaration] = STATE(3366),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4208),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__for_statement_body] = STATE(6708),
    [sym_expression] = STATE(4439),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6747),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2274),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2264),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2266),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(174)] = {
    [sym_declaration] = STATE(3366),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4203),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__for_statement_body] = STATE(5962),
    [sym_expression] = STATE(4439),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6747),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2276),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2264),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2266),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(175)] = {
    [sym_declaration] = STATE(3366),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4232),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__for_statement_body] = STATE(5960),
    [sym_expression] = STATE(4439),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6747),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2278),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2264),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2266),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(176)] = {
    [sym_declaration] = STATE(3366),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4234),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__for_statement_body] = STATE(6023),
    [sym_expression] = STATE(4439),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6747),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2280),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2264),
    [anon_sym___extension__] = ACTIONS(2086),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2266),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(185),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(1587),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(1587),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(1593),
  },
  [STATE(177)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2220),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(2282),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2222),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2222),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2222),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2222),
    [anon_sym_GT_GT] = ACTIONS(2222),
    [anon_sym_SEMI] = ACTIONS(2222),
    [anon_sym___extension__] = ACTIONS(2220),
    [anon_sym_extern] = ACTIONS(2220),
    [anon_sym___attribute__] = ACTIONS(2220),
    [anon_sym___attribute] = ACTIONS(2220),
    [anon_sym_noreturn] = ACTIONS(2220),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym___declspec] = ACTIONS(2220),
    [anon_sym_LBRACE] = ACTIONS(2222),
    [anon_sym_signed] = ACTIONS(2220),
    [anon_sym_unsigned] = ACTIONS(2220),
    [anon_sym_long] = ACTIONS(2220),
    [anon_sym_short] = ACTIONS(2220),
    [anon_sym_static] = ACTIONS(2220),
    [anon_sym_ATautoreleasepool] = ACTIONS(2222),
    [anon_sym_auto] = ACTIONS(2220),
    [anon_sym_register] = ACTIONS(2220),
    [anon_sym_inline] = ACTIONS(2220),
    [anon_sym___inline] = ACTIONS(2220),
    [anon_sym___inline__] = ACTIONS(2220),
    [anon_sym___forceinline] = ACTIONS(2220),
    [anon_sym_thread_local] = ACTIONS(2220),
    [anon_sym___thread] = ACTIONS(2220),
    [anon_sym_CG_EXTERN] = ACTIONS(2220),
    [anon_sym_CG_INLINE] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2220),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2220),
    [anon_sym_IBOutlet] = ACTIONS(2220),
    [anon_sym_IBInspectable] = ACTIONS(2220),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2220),
    [anon_sym_NS_INLINE] = ACTIONS(2220),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2220),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2220),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2220),
    [anon_sym_const] = ACTIONS(2220),
    [anon_sym_constexpr] = ACTIONS(2220),
    [anon_sym_volatile] = ACTIONS(2220),
    [anon_sym_restrict] = ACTIONS(2220),
    [anon_sym___restrict__] = ACTIONS(2220),
    [anon_sym__Atomic] = ACTIONS(2220),
    [anon_sym__Noreturn] = ACTIONS(2220),
    [anon_sym__Nonnull] = ACTIONS(2220),
    [anon_sym_nullable] = ACTIONS(2220),
    [anon_sym__Complex] = ACTIONS(2220),
    [anon_sym__Nullable] = ACTIONS(2220),
    [anon_sym__Nullable_result] = ACTIONS(2220),
    [anon_sym__Null_unspecified] = ACTIONS(2220),
    [anon_sym___autoreleasing] = ACTIONS(2220),
    [anon_sym___block] = ACTIONS(2220),
    [anon_sym___bridge] = ACTIONS(2220),
    [anon_sym___bridge_retained] = ACTIONS(2220),
    [anon_sym___bridge_transfer] = ACTIONS(2220),
    [anon_sym___complex] = ACTIONS(2220),
    [anon_sym___const] = ACTIONS(2220),
    [anon_sym___imag] = ACTIONS(2220),
    [anon_sym___kindof] = ACTIONS(2220),
    [anon_sym___nonnull] = ACTIONS(2220),
    [anon_sym___nullable] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2220),
    [anon_sym___real] = ACTIONS(2220),
    [anon_sym___strong] = ACTIONS(2220),
    [anon_sym___unsafe_unretained] = ACTIONS(2220),
    [anon_sym___unused] = ACTIONS(2220),
    [anon_sym___weak] = ACTIONS(2220),
    [anon_sym_alignas] = ACTIONS(2220),
    [anon_sym__Alignas] = ACTIONS(2220),
    [sym_primitive_type] = ACTIONS(2220),
    [anon_sym_enum] = ACTIONS(2220),
    [anon_sym_struct] = ACTIONS(2220),
    [anon_sym_union] = ACTIONS(2220),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2220),
    [anon_sym___typeof] = ACTIONS(2220),
    [anon_sym_typeof] = ACTIONS(2220),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2220),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2220),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2220),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_API_AVAILABLE] = ACTIONS(2220),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_API_DEPRECATED] = ACTIONS(2220),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym___deprecated_msg] = ACTIONS(2220),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2220),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2220),
    [anon_sym_IMP] = ACTIONS(2220),
    [anon_sym_SEL] = ACTIONS(2220),
    [anon_sym_Class] = ACTIONS(2220),
    [anon_sym_id] = ACTIONS(2220),
  },
  [STATE(178)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2220),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2222),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(2294),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2220),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2220),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2220),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2220),
    [anon_sym_GT_GT] = ACTIONS(2220),
    [anon_sym___extension__] = ACTIONS(2220),
    [anon_sym_extern] = ACTIONS(2220),
    [anon_sym___attribute__] = ACTIONS(2220),
    [anon_sym___attribute] = ACTIONS(2220),
    [anon_sym_noreturn] = ACTIONS(2220),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym___declspec] = ACTIONS(2220),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_RBRACE] = ACTIONS(2222),
    [anon_sym_static] = ACTIONS(2220),
    [anon_sym_EQ] = ACTIONS(2220),
    [anon_sym_auto] = ACTIONS(2220),
    [anon_sym_register] = ACTIONS(2220),
    [anon_sym_inline] = ACTIONS(2220),
    [anon_sym___inline] = ACTIONS(2220),
    [anon_sym___inline__] = ACTIONS(2220),
    [anon_sym___forceinline] = ACTIONS(2220),
    [anon_sym_thread_local] = ACTIONS(2220),
    [anon_sym___thread] = ACTIONS(2220),
    [anon_sym_CG_EXTERN] = ACTIONS(2220),
    [anon_sym_CG_INLINE] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2220),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2220),
    [anon_sym_IBOutlet] = ACTIONS(2220),
    [anon_sym_IBInspectable] = ACTIONS(2220),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2220),
    [anon_sym_NS_INLINE] = ACTIONS(2220),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2220),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2220),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2220),
    [anon_sym_const] = ACTIONS(2220),
    [anon_sym_constexpr] = ACTIONS(2220),
    [anon_sym_volatile] = ACTIONS(2220),
    [anon_sym_restrict] = ACTIONS(2220),
    [anon_sym___restrict__] = ACTIONS(2220),
    [anon_sym__Atomic] = ACTIONS(2220),
    [anon_sym__Noreturn] = ACTIONS(2220),
    [anon_sym__Nonnull] = ACTIONS(2220),
    [anon_sym_nullable] = ACTIONS(2220),
    [anon_sym__Complex] = ACTIONS(2220),
    [anon_sym__Nullable] = ACTIONS(2220),
    [anon_sym__Nullable_result] = ACTIONS(2220),
    [anon_sym__Null_unspecified] = ACTIONS(2220),
    [anon_sym___autoreleasing] = ACTIONS(2220),
    [anon_sym___block] = ACTIONS(2220),
    [anon_sym___bridge] = ACTIONS(2220),
    [anon_sym___bridge_retained] = ACTIONS(2220),
    [anon_sym___bridge_transfer] = ACTIONS(2220),
    [anon_sym___complex] = ACTIONS(2220),
    [anon_sym___const] = ACTIONS(2220),
    [anon_sym___imag] = ACTIONS(2220),
    [anon_sym___kindof] = ACTIONS(2220),
    [anon_sym___nonnull] = ACTIONS(2220),
    [anon_sym___nullable] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2220),
    [anon_sym___real] = ACTIONS(2220),
    [anon_sym___strong] = ACTIONS(2220),
    [anon_sym___unsafe_unretained] = ACTIONS(2220),
    [anon_sym___unused] = ACTIONS(2220),
    [anon_sym___weak] = ACTIONS(2220),
    [anon_sym_alignas] = ACTIONS(2220),
    [anon_sym__Alignas] = ACTIONS(2220),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_STAR_EQ] = ACTIONS(2222),
    [anon_sym_SLASH_EQ] = ACTIONS(2222),
    [anon_sym_PERCENT_EQ] = ACTIONS(2222),
    [anon_sym_PLUS_EQ] = ACTIONS(2222),
    [anon_sym_DASH_EQ] = ACTIONS(2222),
    [anon_sym_LT_LT_EQ] = ACTIONS(2222),
    [anon_sym_GT_GT_EQ] = ACTIONS(2222),
    [anon_sym_AMP_EQ] = ACTIONS(2222),
    [anon_sym_CARET_EQ] = ACTIONS(2222),
    [anon_sym_PIPE_EQ] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2220),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2220),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2220),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_API_AVAILABLE] = ACTIONS(2220),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_API_DEPRECATED] = ACTIONS(2220),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym___deprecated_msg] = ACTIONS(2220),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2220),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(179)] = {
    [sym_catch_clause] = STATE(180),
    [sym_finally_clause] = STATE(262),
    [aux_sym_try_statement_repeat1] = STATE(180),
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token2] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [aux_sym_preproc_else_token1] = ACTIONS(2306),
    [aux_sym_preproc_elif_token1] = ACTIONS(2306),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATsynthesize] = ACTIONS(2308),
    [anon_sym_ATdynamic] = ACTIONS(2308),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATcatch] = ACTIONS(2310),
    [anon_sym___catch] = ACTIONS(2312),
    [anon_sym_ATfinally] = ACTIONS(2314),
    [anon_sym___finally] = ACTIONS(2316),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(180)] = {
    [sym_catch_clause] = STATE(180),
    [aux_sym_try_statement_repeat1] = STATE(180),
    [sym_identifier] = ACTIONS(2318),
    [aux_sym_preproc_include_token1] = ACTIONS(2318),
    [aux_sym_preproc_include_token2] = ACTIONS(2318),
    [aux_sym_preproc_def_token1] = ACTIONS(2318),
    [aux_sym_preproc_if_token1] = ACTIONS(2318),
    [aux_sym_preproc_if_token2] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2318),
    [aux_sym_preproc_else_token1] = ACTIONS(2318),
    [aux_sym_preproc_elif_token1] = ACTIONS(2318),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2318),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2318),
    [sym_preproc_directive] = ACTIONS(2318),
    [anon_sym_LPAREN2] = ACTIONS(2320),
    [anon_sym_BANG] = ACTIONS(2320),
    [anon_sym_TILDE] = ACTIONS(2320),
    [anon_sym_DASH] = ACTIONS(2318),
    [anon_sym_PLUS] = ACTIONS(2318),
    [anon_sym_STAR] = ACTIONS(2320),
    [anon_sym_CARET] = ACTIONS(2320),
    [anon_sym_AMP] = ACTIONS(2320),
    [anon_sym_SEMI] = ACTIONS(2320),
    [anon_sym___extension__] = ACTIONS(2318),
    [anon_sym_typedef] = ACTIONS(2318),
    [anon_sym_extern] = ACTIONS(2318),
    [anon_sym___attribute__] = ACTIONS(2318),
    [anon_sym___attribute] = ACTIONS(2318),
    [anon_sym_noreturn] = ACTIONS(2318),
    [anon_sym_LBRACK] = ACTIONS(2320),
    [anon_sym___declspec] = ACTIONS(2318),
    [anon_sym___cdecl] = ACTIONS(2318),
    [anon_sym___clrcall] = ACTIONS(2318),
    [anon_sym___stdcall] = ACTIONS(2318),
    [anon_sym___fastcall] = ACTIONS(2318),
    [anon_sym___thiscall] = ACTIONS(2318),
    [anon_sym___vectorcall] = ACTIONS(2318),
    [anon_sym_LBRACE] = ACTIONS(2320),
    [anon_sym_signed] = ACTIONS(2318),
    [anon_sym_unsigned] = ACTIONS(2318),
    [anon_sym_long] = ACTIONS(2318),
    [anon_sym_short] = ACTIONS(2318),
    [anon_sym_static] = ACTIONS(2318),
    [anon_sym_ATautoreleasepool] = ACTIONS(2320),
    [anon_sym_auto] = ACTIONS(2318),
    [anon_sym_register] = ACTIONS(2318),
    [anon_sym_inline] = ACTIONS(2318),
    [anon_sym___inline] = ACTIONS(2318),
    [anon_sym___inline__] = ACTIONS(2318),
    [anon_sym___forceinline] = ACTIONS(2318),
    [anon_sym_thread_local] = ACTIONS(2318),
    [anon_sym___thread] = ACTIONS(2318),
    [anon_sym_CG_EXTERN] = ACTIONS(2318),
    [anon_sym_CG_INLINE] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2318),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2318),
    [anon_sym_IBOutlet] = ACTIONS(2318),
    [anon_sym_IBInspectable] = ACTIONS(2318),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2318),
    [anon_sym_NS_INLINE] = ACTIONS(2318),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2318),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2318),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2318),
    [anon_sym_const] = ACTIONS(2318),
    [anon_sym_constexpr] = ACTIONS(2318),
    [anon_sym_volatile] = ACTIONS(2318),
    [anon_sym_restrict] = ACTIONS(2318),
    [anon_sym___restrict__] = ACTIONS(2318),
    [anon_sym__Atomic] = ACTIONS(2318),
    [anon_sym__Noreturn] = ACTIONS(2318),
    [anon_sym__Nonnull] = ACTIONS(2318),
    [anon_sym_nullable] = ACTIONS(2318),
    [anon_sym__Complex] = ACTIONS(2318),
    [anon_sym__Nullable] = ACTIONS(2318),
    [anon_sym__Nullable_result] = ACTIONS(2318),
    [anon_sym__Null_unspecified] = ACTIONS(2318),
    [anon_sym___autoreleasing] = ACTIONS(2318),
    [anon_sym___block] = ACTIONS(2318),
    [anon_sym___bridge] = ACTIONS(2318),
    [anon_sym___bridge_retained] = ACTIONS(2318),
    [anon_sym___bridge_transfer] = ACTIONS(2318),
    [anon_sym___complex] = ACTIONS(2318),
    [anon_sym___const] = ACTIONS(2318),
    [anon_sym___imag] = ACTIONS(2318),
    [anon_sym___kindof] = ACTIONS(2318),
    [anon_sym___nonnull] = ACTIONS(2318),
    [anon_sym___nullable] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2318),
    [anon_sym___real] = ACTIONS(2318),
    [anon_sym___strong] = ACTIONS(2318),
    [anon_sym___unsafe_unretained] = ACTIONS(2318),
    [anon_sym___unused] = ACTIONS(2318),
    [anon_sym___weak] = ACTIONS(2318),
    [anon_sym_alignas] = ACTIONS(2318),
    [anon_sym__Alignas] = ACTIONS(2318),
    [sym_primitive_type] = ACTIONS(2318),
    [anon_sym_enum] = ACTIONS(2318),
    [anon_sym_struct] = ACTIONS(2318),
    [anon_sym_union] = ACTIONS(2318),
    [anon_sym_if] = ACTIONS(2318),
    [anon_sym_else] = ACTIONS(2318),
    [anon_sym_switch] = ACTIONS(2318),
    [anon_sym_case] = ACTIONS(2318),
    [anon_sym_default] = ACTIONS(2318),
    [anon_sym_while] = ACTIONS(2318),
    [anon_sym_do] = ACTIONS(2318),
    [anon_sym_for] = ACTIONS(2318),
    [anon_sym_in] = ACTIONS(2318),
    [anon_sym_return] = ACTIONS(2318),
    [anon_sym_break] = ACTIONS(2318),
    [anon_sym_continue] = ACTIONS(2318),
    [anon_sym_goto] = ACTIONS(2318),
    [anon_sym_DASH_DASH] = ACTIONS(2320),
    [anon_sym_PLUS_PLUS] = ACTIONS(2320),
    [anon_sym_sizeof] = ACTIONS(2318),
    [anon_sym___alignof__] = ACTIONS(2318),
    [anon_sym___alignof] = ACTIONS(2318),
    [anon_sym__alignof] = ACTIONS(2318),
    [anon_sym_alignof] = ACTIONS(2318),
    [anon_sym__Alignof] = ACTIONS(2318),
    [anon_sym_offsetof] = ACTIONS(2318),
    [anon_sym__Generic] = ACTIONS(2318),
    [anon_sym_asm] = ACTIONS(2318),
    [anon_sym___asm__] = ACTIONS(2318),
    [anon_sym___asm] = ACTIONS(2318),
    [sym_number_literal] = ACTIONS(2320),
    [anon_sym_L_SQUOTE] = ACTIONS(2320),
    [anon_sym_u_SQUOTE] = ACTIONS(2320),
    [anon_sym_U_SQUOTE] = ACTIONS(2320),
    [anon_sym_u8_SQUOTE] = ACTIONS(2320),
    [anon_sym_SQUOTE] = ACTIONS(2320),
    [anon_sym_AT] = ACTIONS(2318),
    [anon_sym_DQUOTE] = ACTIONS(2320),
    [anon_sym_L_DQUOTE] = ACTIONS(2320),
    [anon_sym_u_DQUOTE] = ACTIONS(2320),
    [anon_sym_U_DQUOTE] = ACTIONS(2320),
    [anon_sym_u8_DQUOTE] = ACTIONS(2320),
    [sym_true] = ACTIONS(2318),
    [sym_false] = ACTIONS(2318),
    [anon_sym_NULL] = ACTIONS(2318),
    [anon_sym_nullptr] = ACTIONS(2318),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2318),
    [anon_sym___typeof] = ACTIONS(2318),
    [anon_sym_typeof] = ACTIONS(2318),
    [anon_sym_ATimport] = ACTIONS(2320),
    [aux_sym_preproc_undef_token1] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2318),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2318),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2318),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2318),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_API_AVAILABLE] = ACTIONS(2318),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_API_DEPRECATED] = ACTIONS(2318),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym___deprecated_msg] = ACTIONS(2318),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2318),
    [anon_sym_ATprotocol] = ACTIONS(2320),
    [anon_sym_ATinterface] = ACTIONS(2320),
    [anon_sym_ATimplementation] = ACTIONS(2320),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2320),
    [anon_sym_ATsynthesize] = ACTIONS(2320),
    [anon_sym_ATdynamic] = ACTIONS(2320),
    [anon_sym_ATtry] = ACTIONS(2320),
    [anon_sym___try] = ACTIONS(2318),
    [anon_sym_ATcatch] = ACTIONS(2322),
    [anon_sym___catch] = ACTIONS(2325),
    [anon_sym_ATfinally] = ACTIONS(2320),
    [anon_sym___finally] = ACTIONS(2318),
    [anon_sym_ATthrow] = ACTIONS(2320),
    [anon_sym_ATselector] = ACTIONS(2320),
    [anon_sym_ATavailable] = ACTIONS(2320),
    [anon_sym___builtin_available] = ACTIONS(2318),
    [anon_sym_va_arg] = ACTIONS(2318),
    [anon_sym_ATencode] = ACTIONS(2320),
    [anon_sym_ATsynchronized] = ACTIONS(2320),
    [anon_sym_BOOL] = ACTIONS(2318),
    [anon_sym_IMP] = ACTIONS(2318),
    [anon_sym_SEL] = ACTIONS(2318),
    [anon_sym_Class] = ACTIONS(2318),
    [anon_sym_id] = ACTIONS(2318),
  },
  [STATE(181)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(2330),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2220),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2220),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2220),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2220),
    [anon_sym_GT_GT] = ACTIONS(2220),
    [anon_sym_SEMI] = ACTIONS(2222),
    [anon_sym___extension__] = ACTIONS(2220),
    [anon_sym_extern] = ACTIONS(2220),
    [anon_sym___attribute__] = ACTIONS(2220),
    [anon_sym___attribute] = ACTIONS(2220),
    [anon_sym_noreturn] = ACTIONS(2220),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym___declspec] = ACTIONS(2220),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_static] = ACTIONS(2220),
    [anon_sym_EQ] = ACTIONS(2220),
    [anon_sym_auto] = ACTIONS(2220),
    [anon_sym_register] = ACTIONS(2220),
    [anon_sym_inline] = ACTIONS(2220),
    [anon_sym___inline] = ACTIONS(2220),
    [anon_sym___inline__] = ACTIONS(2220),
    [anon_sym___forceinline] = ACTIONS(2220),
    [anon_sym_thread_local] = ACTIONS(2220),
    [anon_sym___thread] = ACTIONS(2220),
    [anon_sym_CG_EXTERN] = ACTIONS(2220),
    [anon_sym_CG_INLINE] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2220),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2220),
    [anon_sym_IBOutlet] = ACTIONS(2220),
    [anon_sym_IBInspectable] = ACTIONS(2220),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2220),
    [anon_sym_NS_INLINE] = ACTIONS(2220),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2220),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2220),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2220),
    [anon_sym_const] = ACTIONS(2220),
    [anon_sym_constexpr] = ACTIONS(2220),
    [anon_sym_volatile] = ACTIONS(2220),
    [anon_sym_restrict] = ACTIONS(2220),
    [anon_sym___restrict__] = ACTIONS(2220),
    [anon_sym__Atomic] = ACTIONS(2220),
    [anon_sym__Noreturn] = ACTIONS(2220),
    [anon_sym__Nonnull] = ACTIONS(2220),
    [anon_sym_nullable] = ACTIONS(2220),
    [anon_sym__Complex] = ACTIONS(2220),
    [anon_sym__Nullable] = ACTIONS(2220),
    [anon_sym__Nullable_result] = ACTIONS(2220),
    [anon_sym__Null_unspecified] = ACTIONS(2220),
    [anon_sym___autoreleasing] = ACTIONS(2220),
    [anon_sym___block] = ACTIONS(2220),
    [anon_sym___bridge] = ACTIONS(2220),
    [anon_sym___bridge_retained] = ACTIONS(2220),
    [anon_sym___bridge_transfer] = ACTIONS(2220),
    [anon_sym___complex] = ACTIONS(2220),
    [anon_sym___const] = ACTIONS(2220),
    [anon_sym___imag] = ACTIONS(2220),
    [anon_sym___kindof] = ACTIONS(2220),
    [anon_sym___nonnull] = ACTIONS(2220),
    [anon_sym___nullable] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2220),
    [anon_sym___real] = ACTIONS(2220),
    [anon_sym___strong] = ACTIONS(2220),
    [anon_sym___unsafe_unretained] = ACTIONS(2220),
    [anon_sym___unused] = ACTIONS(2220),
    [anon_sym___weak] = ACTIONS(2220),
    [anon_sym_alignas] = ACTIONS(2220),
    [anon_sym__Alignas] = ACTIONS(2220),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_STAR_EQ] = ACTIONS(2222),
    [anon_sym_SLASH_EQ] = ACTIONS(2222),
    [anon_sym_PERCENT_EQ] = ACTIONS(2222),
    [anon_sym_PLUS_EQ] = ACTIONS(2222),
    [anon_sym_DASH_EQ] = ACTIONS(2222),
    [anon_sym_LT_LT_EQ] = ACTIONS(2222),
    [anon_sym_GT_GT_EQ] = ACTIONS(2222),
    [anon_sym_AMP_EQ] = ACTIONS(2222),
    [anon_sym_CARET_EQ] = ACTIONS(2222),
    [anon_sym_PIPE_EQ] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2220),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2220),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2220),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_API_AVAILABLE] = ACTIONS(2220),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_API_DEPRECATED] = ACTIONS(2220),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym___deprecated_msg] = ACTIONS(2220),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2220),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(182)] = {
    [sym_catch_clause] = STATE(185),
    [sym_finally_clause] = STATE(429),
    [aux_sym_try_statement_repeat1] = STATE(185),
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token2] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [aux_sym_preproc_else_token1] = ACTIONS(2306),
    [aux_sym_preproc_elif_token1] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATsynthesize] = ACTIONS(2308),
    [anon_sym_ATdynamic] = ACTIONS(2308),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATcatch] = ACTIONS(2338),
    [anon_sym___catch] = ACTIONS(2340),
    [anon_sym_ATfinally] = ACTIONS(2342),
    [anon_sym___finally] = ACTIONS(2344),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(183)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token2] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [aux_sym_preproc_else_token1] = ACTIONS(2346),
    [aux_sym_preproc_elif_token1] = ACTIONS(2346),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATsynthesize] = ACTIONS(2348),
    [anon_sym_ATdynamic] = ACTIONS(2348),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATcatch] = ACTIONS(2348),
    [anon_sym___catch] = ACTIONS(2346),
    [anon_sym_ATfinally] = ACTIONS(2348),
    [anon_sym___finally] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(184)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token2] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [aux_sym_preproc_else_token1] = ACTIONS(2350),
    [aux_sym_preproc_elif_token1] = ACTIONS(2350),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATsynthesize] = ACTIONS(2352),
    [anon_sym_ATdynamic] = ACTIONS(2352),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATcatch] = ACTIONS(2352),
    [anon_sym___catch] = ACTIONS(2350),
    [anon_sym_ATfinally] = ACTIONS(2352),
    [anon_sym___finally] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(185)] = {
    [sym_catch_clause] = STATE(185),
    [aux_sym_try_statement_repeat1] = STATE(185),
    [sym_identifier] = ACTIONS(2318),
    [aux_sym_preproc_include_token1] = ACTIONS(2318),
    [aux_sym_preproc_include_token2] = ACTIONS(2318),
    [aux_sym_preproc_def_token1] = ACTIONS(2318),
    [aux_sym_preproc_if_token1] = ACTIONS(2318),
    [aux_sym_preproc_if_token2] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2318),
    [aux_sym_preproc_else_token1] = ACTIONS(2318),
    [aux_sym_preproc_elif_token1] = ACTIONS(2318),
    [sym_preproc_directive] = ACTIONS(2318),
    [anon_sym_LPAREN2] = ACTIONS(2320),
    [anon_sym_BANG] = ACTIONS(2320),
    [anon_sym_TILDE] = ACTIONS(2320),
    [anon_sym_DASH] = ACTIONS(2318),
    [anon_sym_PLUS] = ACTIONS(2318),
    [anon_sym_STAR] = ACTIONS(2320),
    [anon_sym_CARET] = ACTIONS(2320),
    [anon_sym_AMP] = ACTIONS(2320),
    [anon_sym_SEMI] = ACTIONS(2320),
    [anon_sym___extension__] = ACTIONS(2318),
    [anon_sym_typedef] = ACTIONS(2318),
    [anon_sym_extern] = ACTIONS(2318),
    [anon_sym___attribute__] = ACTIONS(2318),
    [anon_sym___attribute] = ACTIONS(2318),
    [anon_sym_noreturn] = ACTIONS(2318),
    [anon_sym_LBRACK] = ACTIONS(2320),
    [anon_sym___declspec] = ACTIONS(2318),
    [anon_sym___cdecl] = ACTIONS(2318),
    [anon_sym___clrcall] = ACTIONS(2318),
    [anon_sym___stdcall] = ACTIONS(2318),
    [anon_sym___fastcall] = ACTIONS(2318),
    [anon_sym___thiscall] = ACTIONS(2318),
    [anon_sym___vectorcall] = ACTIONS(2318),
    [anon_sym_LBRACE] = ACTIONS(2320),
    [anon_sym_signed] = ACTIONS(2318),
    [anon_sym_unsigned] = ACTIONS(2318),
    [anon_sym_long] = ACTIONS(2318),
    [anon_sym_short] = ACTIONS(2318),
    [anon_sym_static] = ACTIONS(2318),
    [anon_sym_ATautoreleasepool] = ACTIONS(2320),
    [anon_sym_auto] = ACTIONS(2318),
    [anon_sym_register] = ACTIONS(2318),
    [anon_sym_inline] = ACTIONS(2318),
    [anon_sym___inline] = ACTIONS(2318),
    [anon_sym___inline__] = ACTIONS(2318),
    [anon_sym___forceinline] = ACTIONS(2318),
    [anon_sym_thread_local] = ACTIONS(2318),
    [anon_sym___thread] = ACTIONS(2318),
    [anon_sym_CG_EXTERN] = ACTIONS(2318),
    [anon_sym_CG_INLINE] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2318),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2318),
    [anon_sym_IBOutlet] = ACTIONS(2318),
    [anon_sym_IBInspectable] = ACTIONS(2318),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2318),
    [anon_sym_NS_INLINE] = ACTIONS(2318),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2318),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2318),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2318),
    [anon_sym_const] = ACTIONS(2318),
    [anon_sym_constexpr] = ACTIONS(2318),
    [anon_sym_volatile] = ACTIONS(2318),
    [anon_sym_restrict] = ACTIONS(2318),
    [anon_sym___restrict__] = ACTIONS(2318),
    [anon_sym__Atomic] = ACTIONS(2318),
    [anon_sym__Noreturn] = ACTIONS(2318),
    [anon_sym__Nonnull] = ACTIONS(2318),
    [anon_sym_nullable] = ACTIONS(2318),
    [anon_sym__Complex] = ACTIONS(2318),
    [anon_sym__Nullable] = ACTIONS(2318),
    [anon_sym__Nullable_result] = ACTIONS(2318),
    [anon_sym__Null_unspecified] = ACTIONS(2318),
    [anon_sym___autoreleasing] = ACTIONS(2318),
    [anon_sym___block] = ACTIONS(2318),
    [anon_sym___bridge] = ACTIONS(2318),
    [anon_sym___bridge_retained] = ACTIONS(2318),
    [anon_sym___bridge_transfer] = ACTIONS(2318),
    [anon_sym___complex] = ACTIONS(2318),
    [anon_sym___const] = ACTIONS(2318),
    [anon_sym___imag] = ACTIONS(2318),
    [anon_sym___kindof] = ACTIONS(2318),
    [anon_sym___nonnull] = ACTIONS(2318),
    [anon_sym___nullable] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2318),
    [anon_sym___real] = ACTIONS(2318),
    [anon_sym___strong] = ACTIONS(2318),
    [anon_sym___unsafe_unretained] = ACTIONS(2318),
    [anon_sym___unused] = ACTIONS(2318),
    [anon_sym___weak] = ACTIONS(2318),
    [anon_sym_alignas] = ACTIONS(2318),
    [anon_sym__Alignas] = ACTIONS(2318),
    [sym_primitive_type] = ACTIONS(2318),
    [anon_sym_enum] = ACTIONS(2318),
    [anon_sym_struct] = ACTIONS(2318),
    [anon_sym_union] = ACTIONS(2318),
    [anon_sym_if] = ACTIONS(2318),
    [anon_sym_else] = ACTIONS(2318),
    [anon_sym_switch] = ACTIONS(2318),
    [anon_sym_case] = ACTIONS(2318),
    [anon_sym_default] = ACTIONS(2318),
    [anon_sym_while] = ACTIONS(2318),
    [anon_sym_do] = ACTIONS(2318),
    [anon_sym_for] = ACTIONS(2318),
    [anon_sym_in] = ACTIONS(2318),
    [anon_sym_return] = ACTIONS(2318),
    [anon_sym_break] = ACTIONS(2318),
    [anon_sym_continue] = ACTIONS(2318),
    [anon_sym_goto] = ACTIONS(2318),
    [anon_sym_DASH_DASH] = ACTIONS(2320),
    [anon_sym_PLUS_PLUS] = ACTIONS(2320),
    [anon_sym_sizeof] = ACTIONS(2318),
    [anon_sym___alignof__] = ACTIONS(2318),
    [anon_sym___alignof] = ACTIONS(2318),
    [anon_sym__alignof] = ACTIONS(2318),
    [anon_sym_alignof] = ACTIONS(2318),
    [anon_sym__Alignof] = ACTIONS(2318),
    [anon_sym_offsetof] = ACTIONS(2318),
    [anon_sym__Generic] = ACTIONS(2318),
    [anon_sym_asm] = ACTIONS(2318),
    [anon_sym___asm__] = ACTIONS(2318),
    [anon_sym___asm] = ACTIONS(2318),
    [sym_number_literal] = ACTIONS(2320),
    [anon_sym_L_SQUOTE] = ACTIONS(2320),
    [anon_sym_u_SQUOTE] = ACTIONS(2320),
    [anon_sym_U_SQUOTE] = ACTIONS(2320),
    [anon_sym_u8_SQUOTE] = ACTIONS(2320),
    [anon_sym_SQUOTE] = ACTIONS(2320),
    [anon_sym_AT] = ACTIONS(2318),
    [anon_sym_DQUOTE] = ACTIONS(2320),
    [anon_sym_L_DQUOTE] = ACTIONS(2320),
    [anon_sym_u_DQUOTE] = ACTIONS(2320),
    [anon_sym_U_DQUOTE] = ACTIONS(2320),
    [anon_sym_u8_DQUOTE] = ACTIONS(2320),
    [sym_true] = ACTIONS(2318),
    [sym_false] = ACTIONS(2318),
    [anon_sym_NULL] = ACTIONS(2318),
    [anon_sym_nullptr] = ACTIONS(2318),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2318),
    [anon_sym___typeof] = ACTIONS(2318),
    [anon_sym_typeof] = ACTIONS(2318),
    [anon_sym_ATimport] = ACTIONS(2320),
    [aux_sym_preproc_undef_token1] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2318),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2318),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2318),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2318),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_API_AVAILABLE] = ACTIONS(2318),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_API_DEPRECATED] = ACTIONS(2318),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym___deprecated_msg] = ACTIONS(2318),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2318),
    [anon_sym_ATprotocol] = ACTIONS(2320),
    [anon_sym_ATinterface] = ACTIONS(2320),
    [anon_sym_ATimplementation] = ACTIONS(2320),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2320),
    [anon_sym_ATsynthesize] = ACTIONS(2320),
    [anon_sym_ATdynamic] = ACTIONS(2320),
    [anon_sym_ATtry] = ACTIONS(2320),
    [anon_sym___try] = ACTIONS(2318),
    [anon_sym_ATcatch] = ACTIONS(2354),
    [anon_sym___catch] = ACTIONS(2357),
    [anon_sym_ATfinally] = ACTIONS(2320),
    [anon_sym___finally] = ACTIONS(2318),
    [anon_sym_ATthrow] = ACTIONS(2320),
    [anon_sym_ATselector] = ACTIONS(2320),
    [anon_sym_ATavailable] = ACTIONS(2320),
    [anon_sym___builtin_available] = ACTIONS(2318),
    [anon_sym_va_arg] = ACTIONS(2318),
    [anon_sym_ATencode] = ACTIONS(2320),
    [anon_sym_ATsynchronized] = ACTIONS(2320),
    [anon_sym_BOOL] = ACTIONS(2318),
    [anon_sym_IMP] = ACTIONS(2318),
    [anon_sym_SEL] = ACTIONS(2318),
    [anon_sym_Class] = ACTIONS(2318),
    [anon_sym_id] = ACTIONS(2318),
  },
  [STATE(186)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token2] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [aux_sym_preproc_else_token1] = ACTIONS(2360),
    [aux_sym_preproc_elif_token1] = ACTIONS(2360),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATsynthesize] = ACTIONS(2362),
    [anon_sym_ATdynamic] = ACTIONS(2362),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATcatch] = ACTIONS(2362),
    [anon_sym___catch] = ACTIONS(2360),
    [anon_sym_ATfinally] = ACTIONS(2362),
    [anon_sym___finally] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(187)] = {
    [sym_identifier] = ACTIONS(2364),
    [aux_sym_preproc_include_token1] = ACTIONS(2364),
    [aux_sym_preproc_include_token2] = ACTIONS(2364),
    [aux_sym_preproc_def_token1] = ACTIONS(2364),
    [aux_sym_preproc_if_token1] = ACTIONS(2364),
    [aux_sym_preproc_if_token2] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2364),
    [aux_sym_preproc_else_token1] = ACTIONS(2364),
    [aux_sym_preproc_elif_token1] = ACTIONS(2364),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2364),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2364),
    [sym_preproc_directive] = ACTIONS(2364),
    [anon_sym_LPAREN2] = ACTIONS(2366),
    [anon_sym_BANG] = ACTIONS(2366),
    [anon_sym_TILDE] = ACTIONS(2366),
    [anon_sym_DASH] = ACTIONS(2364),
    [anon_sym_PLUS] = ACTIONS(2364),
    [anon_sym_STAR] = ACTIONS(2366),
    [anon_sym_CARET] = ACTIONS(2366),
    [anon_sym_AMP] = ACTIONS(2366),
    [anon_sym_SEMI] = ACTIONS(2366),
    [anon_sym___extension__] = ACTIONS(2364),
    [anon_sym_typedef] = ACTIONS(2364),
    [anon_sym_extern] = ACTIONS(2364),
    [anon_sym___attribute__] = ACTIONS(2364),
    [anon_sym___attribute] = ACTIONS(2364),
    [anon_sym_noreturn] = ACTIONS(2364),
    [anon_sym_LBRACK] = ACTIONS(2366),
    [anon_sym___declspec] = ACTIONS(2364),
    [anon_sym___cdecl] = ACTIONS(2364),
    [anon_sym___clrcall] = ACTIONS(2364),
    [anon_sym___stdcall] = ACTIONS(2364),
    [anon_sym___fastcall] = ACTIONS(2364),
    [anon_sym___thiscall] = ACTIONS(2364),
    [anon_sym___vectorcall] = ACTIONS(2364),
    [anon_sym_LBRACE] = ACTIONS(2366),
    [anon_sym_signed] = ACTIONS(2364),
    [anon_sym_unsigned] = ACTIONS(2364),
    [anon_sym_long] = ACTIONS(2364),
    [anon_sym_short] = ACTIONS(2364),
    [anon_sym_static] = ACTIONS(2364),
    [anon_sym_ATautoreleasepool] = ACTIONS(2366),
    [anon_sym_auto] = ACTIONS(2364),
    [anon_sym_register] = ACTIONS(2364),
    [anon_sym_inline] = ACTIONS(2364),
    [anon_sym___inline] = ACTIONS(2364),
    [anon_sym___inline__] = ACTIONS(2364),
    [anon_sym___forceinline] = ACTIONS(2364),
    [anon_sym_thread_local] = ACTIONS(2364),
    [anon_sym___thread] = ACTIONS(2364),
    [anon_sym_CG_EXTERN] = ACTIONS(2364),
    [anon_sym_CG_INLINE] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2364),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2364),
    [anon_sym_IBOutlet] = ACTIONS(2364),
    [anon_sym_IBInspectable] = ACTIONS(2364),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2364),
    [anon_sym_NS_INLINE] = ACTIONS(2364),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2364),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2364),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2364),
    [anon_sym_const] = ACTIONS(2364),
    [anon_sym_constexpr] = ACTIONS(2364),
    [anon_sym_volatile] = ACTIONS(2364),
    [anon_sym_restrict] = ACTIONS(2364),
    [anon_sym___restrict__] = ACTIONS(2364),
    [anon_sym__Atomic] = ACTIONS(2364),
    [anon_sym__Noreturn] = ACTIONS(2364),
    [anon_sym__Nonnull] = ACTIONS(2364),
    [anon_sym_nullable] = ACTIONS(2364),
    [anon_sym__Complex] = ACTIONS(2364),
    [anon_sym__Nullable] = ACTIONS(2364),
    [anon_sym__Nullable_result] = ACTIONS(2364),
    [anon_sym__Null_unspecified] = ACTIONS(2364),
    [anon_sym___autoreleasing] = ACTIONS(2364),
    [anon_sym___block] = ACTIONS(2364),
    [anon_sym___bridge] = ACTIONS(2364),
    [anon_sym___bridge_retained] = ACTIONS(2364),
    [anon_sym___bridge_transfer] = ACTIONS(2364),
    [anon_sym___complex] = ACTIONS(2364),
    [anon_sym___const] = ACTIONS(2364),
    [anon_sym___imag] = ACTIONS(2364),
    [anon_sym___kindof] = ACTIONS(2364),
    [anon_sym___nonnull] = ACTIONS(2364),
    [anon_sym___nullable] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2364),
    [anon_sym___real] = ACTIONS(2364),
    [anon_sym___strong] = ACTIONS(2364),
    [anon_sym___unsafe_unretained] = ACTIONS(2364),
    [anon_sym___unused] = ACTIONS(2364),
    [anon_sym___weak] = ACTIONS(2364),
    [anon_sym_alignas] = ACTIONS(2364),
    [anon_sym__Alignas] = ACTIONS(2364),
    [sym_primitive_type] = ACTIONS(2364),
    [anon_sym_enum] = ACTIONS(2364),
    [anon_sym_struct] = ACTIONS(2364),
    [anon_sym_union] = ACTIONS(2364),
    [anon_sym_if] = ACTIONS(2364),
    [anon_sym_else] = ACTIONS(2364),
    [anon_sym_switch] = ACTIONS(2364),
    [anon_sym_case] = ACTIONS(2364),
    [anon_sym_default] = ACTIONS(2364),
    [anon_sym_while] = ACTIONS(2364),
    [anon_sym_do] = ACTIONS(2364),
    [anon_sym_for] = ACTIONS(2364),
    [anon_sym_in] = ACTIONS(2364),
    [anon_sym_return] = ACTIONS(2364),
    [anon_sym_break] = ACTIONS(2364),
    [anon_sym_continue] = ACTIONS(2364),
    [anon_sym_goto] = ACTIONS(2364),
    [anon_sym_DASH_DASH] = ACTIONS(2366),
    [anon_sym_PLUS_PLUS] = ACTIONS(2366),
    [anon_sym_sizeof] = ACTIONS(2364),
    [anon_sym___alignof__] = ACTIONS(2364),
    [anon_sym___alignof] = ACTIONS(2364),
    [anon_sym__alignof] = ACTIONS(2364),
    [anon_sym_alignof] = ACTIONS(2364),
    [anon_sym__Alignof] = ACTIONS(2364),
    [anon_sym_offsetof] = ACTIONS(2364),
    [anon_sym__Generic] = ACTIONS(2364),
    [anon_sym_asm] = ACTIONS(2364),
    [anon_sym___asm__] = ACTIONS(2364),
    [anon_sym___asm] = ACTIONS(2364),
    [sym_number_literal] = ACTIONS(2366),
    [anon_sym_L_SQUOTE] = ACTIONS(2366),
    [anon_sym_u_SQUOTE] = ACTIONS(2366),
    [anon_sym_U_SQUOTE] = ACTIONS(2366),
    [anon_sym_u8_SQUOTE] = ACTIONS(2366),
    [anon_sym_SQUOTE] = ACTIONS(2366),
    [anon_sym_AT] = ACTIONS(2364),
    [anon_sym_DQUOTE] = ACTIONS(2366),
    [anon_sym_L_DQUOTE] = ACTIONS(2366),
    [anon_sym_u_DQUOTE] = ACTIONS(2366),
    [anon_sym_U_DQUOTE] = ACTIONS(2366),
    [anon_sym_u8_DQUOTE] = ACTIONS(2366),
    [sym_true] = ACTIONS(2364),
    [sym_false] = ACTIONS(2364),
    [anon_sym_NULL] = ACTIONS(2364),
    [anon_sym_nullptr] = ACTIONS(2364),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2364),
    [anon_sym___typeof] = ACTIONS(2364),
    [anon_sym_typeof] = ACTIONS(2364),
    [anon_sym_ATimport] = ACTIONS(2366),
    [aux_sym_preproc_undef_token1] = ACTIONS(2364),
    [anon_sym_POUND] = ACTIONS(2364),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2364),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2364),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2364),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_API_AVAILABLE] = ACTIONS(2364),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_API_DEPRECATED] = ACTIONS(2364),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym___deprecated_msg] = ACTIONS(2364),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2364),
    [anon_sym_ATprotocol] = ACTIONS(2366),
    [anon_sym_ATinterface] = ACTIONS(2366),
    [anon_sym_ATimplementation] = ACTIONS(2366),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2366),
    [anon_sym_ATsynthesize] = ACTIONS(2366),
    [anon_sym_ATdynamic] = ACTIONS(2366),
    [anon_sym_ATtry] = ACTIONS(2366),
    [anon_sym___try] = ACTIONS(2364),
    [anon_sym_ATcatch] = ACTIONS(2366),
    [anon_sym___catch] = ACTIONS(2364),
    [anon_sym_ATfinally] = ACTIONS(2366),
    [anon_sym___finally] = ACTIONS(2364),
    [anon_sym_ATthrow] = ACTIONS(2366),
    [anon_sym_ATselector] = ACTIONS(2366),
    [anon_sym_ATavailable] = ACTIONS(2366),
    [anon_sym___builtin_available] = ACTIONS(2364),
    [anon_sym_va_arg] = ACTIONS(2364),
    [anon_sym_ATencode] = ACTIONS(2366),
    [anon_sym_ATsynchronized] = ACTIONS(2366),
    [anon_sym_BOOL] = ACTIONS(2364),
    [anon_sym_IMP] = ACTIONS(2364),
    [anon_sym_SEL] = ACTIONS(2364),
    [anon_sym_Class] = ACTIONS(2364),
    [anon_sym_id] = ACTIONS(2364),
  },
  [STATE(188)] = {
    [sym_identifier] = ACTIONS(2368),
    [aux_sym_preproc_include_token1] = ACTIONS(2368),
    [aux_sym_preproc_include_token2] = ACTIONS(2368),
    [aux_sym_preproc_def_token1] = ACTIONS(2368),
    [aux_sym_preproc_if_token1] = ACTIONS(2368),
    [aux_sym_preproc_if_token2] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2368),
    [aux_sym_preproc_else_token1] = ACTIONS(2368),
    [aux_sym_preproc_elif_token1] = ACTIONS(2368),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2368),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2368),
    [sym_preproc_directive] = ACTIONS(2368),
    [anon_sym_LPAREN2] = ACTIONS(2370),
    [anon_sym_BANG] = ACTIONS(2370),
    [anon_sym_TILDE] = ACTIONS(2370),
    [anon_sym_DASH] = ACTIONS(2368),
    [anon_sym_PLUS] = ACTIONS(2368),
    [anon_sym_STAR] = ACTIONS(2370),
    [anon_sym_CARET] = ACTIONS(2370),
    [anon_sym_AMP] = ACTIONS(2370),
    [anon_sym_SEMI] = ACTIONS(2370),
    [anon_sym___extension__] = ACTIONS(2368),
    [anon_sym_typedef] = ACTIONS(2368),
    [anon_sym_extern] = ACTIONS(2368),
    [anon_sym___attribute__] = ACTIONS(2368),
    [anon_sym___attribute] = ACTIONS(2368),
    [anon_sym_noreturn] = ACTIONS(2368),
    [anon_sym_LBRACK] = ACTIONS(2370),
    [anon_sym___declspec] = ACTIONS(2368),
    [anon_sym___cdecl] = ACTIONS(2368),
    [anon_sym___clrcall] = ACTIONS(2368),
    [anon_sym___stdcall] = ACTIONS(2368),
    [anon_sym___fastcall] = ACTIONS(2368),
    [anon_sym___thiscall] = ACTIONS(2368),
    [anon_sym___vectorcall] = ACTIONS(2368),
    [anon_sym_LBRACE] = ACTIONS(2370),
    [anon_sym_signed] = ACTIONS(2368),
    [anon_sym_unsigned] = ACTIONS(2368),
    [anon_sym_long] = ACTIONS(2368),
    [anon_sym_short] = ACTIONS(2368),
    [anon_sym_static] = ACTIONS(2368),
    [anon_sym_ATautoreleasepool] = ACTIONS(2370),
    [anon_sym_auto] = ACTIONS(2368),
    [anon_sym_register] = ACTIONS(2368),
    [anon_sym_inline] = ACTIONS(2368),
    [anon_sym___inline] = ACTIONS(2368),
    [anon_sym___inline__] = ACTIONS(2368),
    [anon_sym___forceinline] = ACTIONS(2368),
    [anon_sym_thread_local] = ACTIONS(2368),
    [anon_sym___thread] = ACTIONS(2368),
    [anon_sym_CG_EXTERN] = ACTIONS(2368),
    [anon_sym_CG_INLINE] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2368),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2368),
    [anon_sym_IBOutlet] = ACTIONS(2368),
    [anon_sym_IBInspectable] = ACTIONS(2368),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2368),
    [anon_sym_NS_INLINE] = ACTIONS(2368),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2368),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2368),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2368),
    [anon_sym_const] = ACTIONS(2368),
    [anon_sym_constexpr] = ACTIONS(2368),
    [anon_sym_volatile] = ACTIONS(2368),
    [anon_sym_restrict] = ACTIONS(2368),
    [anon_sym___restrict__] = ACTIONS(2368),
    [anon_sym__Atomic] = ACTIONS(2368),
    [anon_sym__Noreturn] = ACTIONS(2368),
    [anon_sym__Nonnull] = ACTIONS(2368),
    [anon_sym_nullable] = ACTIONS(2368),
    [anon_sym__Complex] = ACTIONS(2368),
    [anon_sym__Nullable] = ACTIONS(2368),
    [anon_sym__Nullable_result] = ACTIONS(2368),
    [anon_sym__Null_unspecified] = ACTIONS(2368),
    [anon_sym___autoreleasing] = ACTIONS(2368),
    [anon_sym___block] = ACTIONS(2368),
    [anon_sym___bridge] = ACTIONS(2368),
    [anon_sym___bridge_retained] = ACTIONS(2368),
    [anon_sym___bridge_transfer] = ACTIONS(2368),
    [anon_sym___complex] = ACTIONS(2368),
    [anon_sym___const] = ACTIONS(2368),
    [anon_sym___imag] = ACTIONS(2368),
    [anon_sym___kindof] = ACTIONS(2368),
    [anon_sym___nonnull] = ACTIONS(2368),
    [anon_sym___nullable] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2368),
    [anon_sym___real] = ACTIONS(2368),
    [anon_sym___strong] = ACTIONS(2368),
    [anon_sym___unsafe_unretained] = ACTIONS(2368),
    [anon_sym___unused] = ACTIONS(2368),
    [anon_sym___weak] = ACTIONS(2368),
    [anon_sym_alignas] = ACTIONS(2368),
    [anon_sym__Alignas] = ACTIONS(2368),
    [sym_primitive_type] = ACTIONS(2368),
    [anon_sym_enum] = ACTIONS(2368),
    [anon_sym_struct] = ACTIONS(2368),
    [anon_sym_union] = ACTIONS(2368),
    [anon_sym_if] = ACTIONS(2368),
    [anon_sym_else] = ACTIONS(2368),
    [anon_sym_switch] = ACTIONS(2368),
    [anon_sym_case] = ACTIONS(2368),
    [anon_sym_default] = ACTIONS(2368),
    [anon_sym_while] = ACTIONS(2368),
    [anon_sym_do] = ACTIONS(2368),
    [anon_sym_for] = ACTIONS(2368),
    [anon_sym_in] = ACTIONS(2368),
    [anon_sym_return] = ACTIONS(2368),
    [anon_sym_break] = ACTIONS(2368),
    [anon_sym_continue] = ACTIONS(2368),
    [anon_sym_goto] = ACTIONS(2368),
    [anon_sym_DASH_DASH] = ACTIONS(2370),
    [anon_sym_PLUS_PLUS] = ACTIONS(2370),
    [anon_sym_sizeof] = ACTIONS(2368),
    [anon_sym___alignof__] = ACTIONS(2368),
    [anon_sym___alignof] = ACTIONS(2368),
    [anon_sym__alignof] = ACTIONS(2368),
    [anon_sym_alignof] = ACTIONS(2368),
    [anon_sym__Alignof] = ACTIONS(2368),
    [anon_sym_offsetof] = ACTIONS(2368),
    [anon_sym__Generic] = ACTIONS(2368),
    [anon_sym_asm] = ACTIONS(2368),
    [anon_sym___asm__] = ACTIONS(2368),
    [anon_sym___asm] = ACTIONS(2368),
    [sym_number_literal] = ACTIONS(2370),
    [anon_sym_L_SQUOTE] = ACTIONS(2370),
    [anon_sym_u_SQUOTE] = ACTIONS(2370),
    [anon_sym_U_SQUOTE] = ACTIONS(2370),
    [anon_sym_u8_SQUOTE] = ACTIONS(2370),
    [anon_sym_SQUOTE] = ACTIONS(2370),
    [anon_sym_AT] = ACTIONS(2368),
    [anon_sym_DQUOTE] = ACTIONS(2370),
    [anon_sym_L_DQUOTE] = ACTIONS(2370),
    [anon_sym_u_DQUOTE] = ACTIONS(2370),
    [anon_sym_U_DQUOTE] = ACTIONS(2370),
    [anon_sym_u8_DQUOTE] = ACTIONS(2370),
    [sym_true] = ACTIONS(2368),
    [sym_false] = ACTIONS(2368),
    [anon_sym_NULL] = ACTIONS(2368),
    [anon_sym_nullptr] = ACTIONS(2368),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2368),
    [anon_sym___typeof] = ACTIONS(2368),
    [anon_sym_typeof] = ACTIONS(2368),
    [anon_sym_ATimport] = ACTIONS(2370),
    [aux_sym_preproc_undef_token1] = ACTIONS(2368),
    [anon_sym_POUND] = ACTIONS(2368),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2368),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2368),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2368),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_API_AVAILABLE] = ACTIONS(2368),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_API_DEPRECATED] = ACTIONS(2368),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym___deprecated_msg] = ACTIONS(2368),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2368),
    [anon_sym_ATprotocol] = ACTIONS(2370),
    [anon_sym_ATinterface] = ACTIONS(2370),
    [anon_sym_ATimplementation] = ACTIONS(2370),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2370),
    [anon_sym_ATsynthesize] = ACTIONS(2370),
    [anon_sym_ATdynamic] = ACTIONS(2370),
    [anon_sym_ATtry] = ACTIONS(2370),
    [anon_sym___try] = ACTIONS(2368),
    [anon_sym_ATcatch] = ACTIONS(2370),
    [anon_sym___catch] = ACTIONS(2368),
    [anon_sym_ATfinally] = ACTIONS(2370),
    [anon_sym___finally] = ACTIONS(2368),
    [anon_sym_ATthrow] = ACTIONS(2370),
    [anon_sym_ATselector] = ACTIONS(2370),
    [anon_sym_ATavailable] = ACTIONS(2370),
    [anon_sym___builtin_available] = ACTIONS(2368),
    [anon_sym_va_arg] = ACTIONS(2368),
    [anon_sym_ATencode] = ACTIONS(2370),
    [anon_sym_ATsynchronized] = ACTIONS(2370),
    [anon_sym_BOOL] = ACTIONS(2368),
    [anon_sym_IMP] = ACTIONS(2368),
    [anon_sym_SEL] = ACTIONS(2368),
    [anon_sym_Class] = ACTIONS(2368),
    [anon_sym_id] = ACTIONS(2368),
  },
  [STATE(189)] = {
    [sym_catch_clause] = STATE(190),
    [sym_finally_clause] = STATE(534),
    [aux_sym_try_statement_repeat1] = STATE(190),
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token2] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATsynthesize] = ACTIONS(2308),
    [anon_sym_ATdynamic] = ACTIONS(2308),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATcatch] = ACTIONS(2372),
    [anon_sym___catch] = ACTIONS(2374),
    [anon_sym_ATfinally] = ACTIONS(2376),
    [anon_sym___finally] = ACTIONS(2378),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(190)] = {
    [sym_catch_clause] = STATE(190),
    [aux_sym_try_statement_repeat1] = STATE(190),
    [sym_identifier] = ACTIONS(2318),
    [aux_sym_preproc_include_token1] = ACTIONS(2318),
    [aux_sym_preproc_include_token2] = ACTIONS(2318),
    [aux_sym_preproc_def_token1] = ACTIONS(2318),
    [aux_sym_preproc_if_token1] = ACTIONS(2318),
    [aux_sym_preproc_if_token2] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2318),
    [sym_preproc_directive] = ACTIONS(2318),
    [anon_sym_LPAREN2] = ACTIONS(2320),
    [anon_sym_BANG] = ACTIONS(2320),
    [anon_sym_TILDE] = ACTIONS(2320),
    [anon_sym_DASH] = ACTIONS(2318),
    [anon_sym_PLUS] = ACTIONS(2318),
    [anon_sym_STAR] = ACTIONS(2320),
    [anon_sym_CARET] = ACTIONS(2320),
    [anon_sym_AMP] = ACTIONS(2320),
    [anon_sym_SEMI] = ACTIONS(2320),
    [anon_sym___extension__] = ACTIONS(2318),
    [anon_sym_typedef] = ACTIONS(2318),
    [anon_sym_extern] = ACTIONS(2318),
    [anon_sym___attribute__] = ACTIONS(2318),
    [anon_sym___attribute] = ACTIONS(2318),
    [anon_sym_noreturn] = ACTIONS(2318),
    [anon_sym_LBRACK] = ACTIONS(2320),
    [anon_sym___declspec] = ACTIONS(2318),
    [anon_sym___cdecl] = ACTIONS(2318),
    [anon_sym___clrcall] = ACTIONS(2318),
    [anon_sym___stdcall] = ACTIONS(2318),
    [anon_sym___fastcall] = ACTIONS(2318),
    [anon_sym___thiscall] = ACTIONS(2318),
    [anon_sym___vectorcall] = ACTIONS(2318),
    [anon_sym_LBRACE] = ACTIONS(2320),
    [anon_sym_signed] = ACTIONS(2318),
    [anon_sym_unsigned] = ACTIONS(2318),
    [anon_sym_long] = ACTIONS(2318),
    [anon_sym_short] = ACTIONS(2318),
    [anon_sym_static] = ACTIONS(2318),
    [anon_sym_ATautoreleasepool] = ACTIONS(2320),
    [anon_sym_auto] = ACTIONS(2318),
    [anon_sym_register] = ACTIONS(2318),
    [anon_sym_inline] = ACTIONS(2318),
    [anon_sym___inline] = ACTIONS(2318),
    [anon_sym___inline__] = ACTIONS(2318),
    [anon_sym___forceinline] = ACTIONS(2318),
    [anon_sym_thread_local] = ACTIONS(2318),
    [anon_sym___thread] = ACTIONS(2318),
    [anon_sym_CG_EXTERN] = ACTIONS(2318),
    [anon_sym_CG_INLINE] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2318),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2318),
    [anon_sym_IBOutlet] = ACTIONS(2318),
    [anon_sym_IBInspectable] = ACTIONS(2318),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2318),
    [anon_sym_NS_INLINE] = ACTIONS(2318),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2318),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2318),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2318),
    [anon_sym_const] = ACTIONS(2318),
    [anon_sym_constexpr] = ACTIONS(2318),
    [anon_sym_volatile] = ACTIONS(2318),
    [anon_sym_restrict] = ACTIONS(2318),
    [anon_sym___restrict__] = ACTIONS(2318),
    [anon_sym__Atomic] = ACTIONS(2318),
    [anon_sym__Noreturn] = ACTIONS(2318),
    [anon_sym__Nonnull] = ACTIONS(2318),
    [anon_sym_nullable] = ACTIONS(2318),
    [anon_sym__Complex] = ACTIONS(2318),
    [anon_sym__Nullable] = ACTIONS(2318),
    [anon_sym__Nullable_result] = ACTIONS(2318),
    [anon_sym__Null_unspecified] = ACTIONS(2318),
    [anon_sym___autoreleasing] = ACTIONS(2318),
    [anon_sym___block] = ACTIONS(2318),
    [anon_sym___bridge] = ACTIONS(2318),
    [anon_sym___bridge_retained] = ACTIONS(2318),
    [anon_sym___bridge_transfer] = ACTIONS(2318),
    [anon_sym___complex] = ACTIONS(2318),
    [anon_sym___const] = ACTIONS(2318),
    [anon_sym___imag] = ACTIONS(2318),
    [anon_sym___kindof] = ACTIONS(2318),
    [anon_sym___nonnull] = ACTIONS(2318),
    [anon_sym___nullable] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2318),
    [anon_sym___real] = ACTIONS(2318),
    [anon_sym___strong] = ACTIONS(2318),
    [anon_sym___unsafe_unretained] = ACTIONS(2318),
    [anon_sym___unused] = ACTIONS(2318),
    [anon_sym___weak] = ACTIONS(2318),
    [anon_sym_alignas] = ACTIONS(2318),
    [anon_sym__Alignas] = ACTIONS(2318),
    [sym_primitive_type] = ACTIONS(2318),
    [anon_sym_enum] = ACTIONS(2318),
    [anon_sym_struct] = ACTIONS(2318),
    [anon_sym_union] = ACTIONS(2318),
    [anon_sym_if] = ACTIONS(2318),
    [anon_sym_else] = ACTIONS(2318),
    [anon_sym_switch] = ACTIONS(2318),
    [anon_sym_case] = ACTIONS(2318),
    [anon_sym_default] = ACTIONS(2318),
    [anon_sym_while] = ACTIONS(2318),
    [anon_sym_do] = ACTIONS(2318),
    [anon_sym_for] = ACTIONS(2318),
    [anon_sym_in] = ACTIONS(2318),
    [anon_sym_return] = ACTIONS(2318),
    [anon_sym_break] = ACTIONS(2318),
    [anon_sym_continue] = ACTIONS(2318),
    [anon_sym_goto] = ACTIONS(2318),
    [anon_sym_DASH_DASH] = ACTIONS(2320),
    [anon_sym_PLUS_PLUS] = ACTIONS(2320),
    [anon_sym_sizeof] = ACTIONS(2318),
    [anon_sym___alignof__] = ACTIONS(2318),
    [anon_sym___alignof] = ACTIONS(2318),
    [anon_sym__alignof] = ACTIONS(2318),
    [anon_sym_alignof] = ACTIONS(2318),
    [anon_sym__Alignof] = ACTIONS(2318),
    [anon_sym_offsetof] = ACTIONS(2318),
    [anon_sym__Generic] = ACTIONS(2318),
    [anon_sym_asm] = ACTIONS(2318),
    [anon_sym___asm__] = ACTIONS(2318),
    [anon_sym___asm] = ACTIONS(2318),
    [sym_number_literal] = ACTIONS(2320),
    [anon_sym_L_SQUOTE] = ACTIONS(2320),
    [anon_sym_u_SQUOTE] = ACTIONS(2320),
    [anon_sym_U_SQUOTE] = ACTIONS(2320),
    [anon_sym_u8_SQUOTE] = ACTIONS(2320),
    [anon_sym_SQUOTE] = ACTIONS(2320),
    [anon_sym_AT] = ACTIONS(2318),
    [anon_sym_DQUOTE] = ACTIONS(2320),
    [anon_sym_L_DQUOTE] = ACTIONS(2320),
    [anon_sym_u_DQUOTE] = ACTIONS(2320),
    [anon_sym_U_DQUOTE] = ACTIONS(2320),
    [anon_sym_u8_DQUOTE] = ACTIONS(2320),
    [sym_true] = ACTIONS(2318),
    [sym_false] = ACTIONS(2318),
    [anon_sym_NULL] = ACTIONS(2318),
    [anon_sym_nullptr] = ACTIONS(2318),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2318),
    [anon_sym___typeof] = ACTIONS(2318),
    [anon_sym_typeof] = ACTIONS(2318),
    [anon_sym_ATimport] = ACTIONS(2320),
    [aux_sym_preproc_undef_token1] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2318),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2318),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2318),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2318),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_API_AVAILABLE] = ACTIONS(2318),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_API_DEPRECATED] = ACTIONS(2318),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym___deprecated_msg] = ACTIONS(2318),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2318),
    [anon_sym_ATprotocol] = ACTIONS(2320),
    [anon_sym_ATinterface] = ACTIONS(2320),
    [anon_sym_ATimplementation] = ACTIONS(2320),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2320),
    [anon_sym_ATsynthesize] = ACTIONS(2320),
    [anon_sym_ATdynamic] = ACTIONS(2320),
    [anon_sym_ATtry] = ACTIONS(2320),
    [anon_sym___try] = ACTIONS(2318),
    [anon_sym_ATcatch] = ACTIONS(2380),
    [anon_sym___catch] = ACTIONS(2383),
    [anon_sym_ATfinally] = ACTIONS(2320),
    [anon_sym___finally] = ACTIONS(2318),
    [anon_sym_ATthrow] = ACTIONS(2320),
    [anon_sym_ATselector] = ACTIONS(2320),
    [anon_sym_ATavailable] = ACTIONS(2320),
    [anon_sym___builtin_available] = ACTIONS(2318),
    [anon_sym_va_arg] = ACTIONS(2318),
    [anon_sym_ATencode] = ACTIONS(2320),
    [anon_sym_ATsynchronized] = ACTIONS(2320),
    [anon_sym_BOOL] = ACTIONS(2318),
    [anon_sym_IMP] = ACTIONS(2318),
    [anon_sym_SEL] = ACTIONS(2318),
    [anon_sym_Class] = ACTIONS(2318),
    [anon_sym_id] = ACTIONS(2318),
  },
  [STATE(191)] = {
    [sym_identifier] = ACTIONS(2368),
    [aux_sym_preproc_include_token1] = ACTIONS(2368),
    [aux_sym_preproc_include_token2] = ACTIONS(2368),
    [aux_sym_preproc_def_token1] = ACTIONS(2368),
    [aux_sym_preproc_if_token1] = ACTIONS(2368),
    [aux_sym_preproc_if_token2] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2368),
    [aux_sym_preproc_else_token1] = ACTIONS(2368),
    [aux_sym_preproc_elif_token1] = ACTIONS(2368),
    [sym_preproc_directive] = ACTIONS(2368),
    [anon_sym_LPAREN2] = ACTIONS(2370),
    [anon_sym_BANG] = ACTIONS(2370),
    [anon_sym_TILDE] = ACTIONS(2370),
    [anon_sym_DASH] = ACTIONS(2368),
    [anon_sym_PLUS] = ACTIONS(2368),
    [anon_sym_STAR] = ACTIONS(2370),
    [anon_sym_CARET] = ACTIONS(2370),
    [anon_sym_AMP] = ACTIONS(2370),
    [anon_sym_SEMI] = ACTIONS(2370),
    [anon_sym___extension__] = ACTIONS(2368),
    [anon_sym_typedef] = ACTIONS(2368),
    [anon_sym_extern] = ACTIONS(2368),
    [anon_sym___attribute__] = ACTIONS(2368),
    [anon_sym___attribute] = ACTIONS(2368),
    [anon_sym_noreturn] = ACTIONS(2368),
    [anon_sym_LBRACK] = ACTIONS(2370),
    [anon_sym___declspec] = ACTIONS(2368),
    [anon_sym___cdecl] = ACTIONS(2368),
    [anon_sym___clrcall] = ACTIONS(2368),
    [anon_sym___stdcall] = ACTIONS(2368),
    [anon_sym___fastcall] = ACTIONS(2368),
    [anon_sym___thiscall] = ACTIONS(2368),
    [anon_sym___vectorcall] = ACTIONS(2368),
    [anon_sym_LBRACE] = ACTIONS(2370),
    [anon_sym_signed] = ACTIONS(2368),
    [anon_sym_unsigned] = ACTIONS(2368),
    [anon_sym_long] = ACTIONS(2368),
    [anon_sym_short] = ACTIONS(2368),
    [anon_sym_static] = ACTIONS(2368),
    [anon_sym_ATautoreleasepool] = ACTIONS(2370),
    [anon_sym_auto] = ACTIONS(2368),
    [anon_sym_register] = ACTIONS(2368),
    [anon_sym_inline] = ACTIONS(2368),
    [anon_sym___inline] = ACTIONS(2368),
    [anon_sym___inline__] = ACTIONS(2368),
    [anon_sym___forceinline] = ACTIONS(2368),
    [anon_sym_thread_local] = ACTIONS(2368),
    [anon_sym___thread] = ACTIONS(2368),
    [anon_sym_CG_EXTERN] = ACTIONS(2368),
    [anon_sym_CG_INLINE] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2368),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2368),
    [anon_sym_IBOutlet] = ACTIONS(2368),
    [anon_sym_IBInspectable] = ACTIONS(2368),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2368),
    [anon_sym_NS_INLINE] = ACTIONS(2368),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2368),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2368),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2368),
    [anon_sym_const] = ACTIONS(2368),
    [anon_sym_constexpr] = ACTIONS(2368),
    [anon_sym_volatile] = ACTIONS(2368),
    [anon_sym_restrict] = ACTIONS(2368),
    [anon_sym___restrict__] = ACTIONS(2368),
    [anon_sym__Atomic] = ACTIONS(2368),
    [anon_sym__Noreturn] = ACTIONS(2368),
    [anon_sym__Nonnull] = ACTIONS(2368),
    [anon_sym_nullable] = ACTIONS(2368),
    [anon_sym__Complex] = ACTIONS(2368),
    [anon_sym__Nullable] = ACTIONS(2368),
    [anon_sym__Nullable_result] = ACTIONS(2368),
    [anon_sym__Null_unspecified] = ACTIONS(2368),
    [anon_sym___autoreleasing] = ACTIONS(2368),
    [anon_sym___block] = ACTIONS(2368),
    [anon_sym___bridge] = ACTIONS(2368),
    [anon_sym___bridge_retained] = ACTIONS(2368),
    [anon_sym___bridge_transfer] = ACTIONS(2368),
    [anon_sym___complex] = ACTIONS(2368),
    [anon_sym___const] = ACTIONS(2368),
    [anon_sym___imag] = ACTIONS(2368),
    [anon_sym___kindof] = ACTIONS(2368),
    [anon_sym___nonnull] = ACTIONS(2368),
    [anon_sym___nullable] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2368),
    [anon_sym___real] = ACTIONS(2368),
    [anon_sym___strong] = ACTIONS(2368),
    [anon_sym___unsafe_unretained] = ACTIONS(2368),
    [anon_sym___unused] = ACTIONS(2368),
    [anon_sym___weak] = ACTIONS(2368),
    [anon_sym_alignas] = ACTIONS(2368),
    [anon_sym__Alignas] = ACTIONS(2368),
    [sym_primitive_type] = ACTIONS(2368),
    [anon_sym_enum] = ACTIONS(2368),
    [anon_sym_struct] = ACTIONS(2368),
    [anon_sym_union] = ACTIONS(2368),
    [anon_sym_if] = ACTIONS(2368),
    [anon_sym_else] = ACTIONS(2368),
    [anon_sym_switch] = ACTIONS(2368),
    [anon_sym_case] = ACTIONS(2368),
    [anon_sym_default] = ACTIONS(2368),
    [anon_sym_while] = ACTIONS(2368),
    [anon_sym_do] = ACTIONS(2368),
    [anon_sym_for] = ACTIONS(2368),
    [anon_sym_in] = ACTIONS(2368),
    [anon_sym_return] = ACTIONS(2368),
    [anon_sym_break] = ACTIONS(2368),
    [anon_sym_continue] = ACTIONS(2368),
    [anon_sym_goto] = ACTIONS(2368),
    [anon_sym_DASH_DASH] = ACTIONS(2370),
    [anon_sym_PLUS_PLUS] = ACTIONS(2370),
    [anon_sym_sizeof] = ACTIONS(2368),
    [anon_sym___alignof__] = ACTIONS(2368),
    [anon_sym___alignof] = ACTIONS(2368),
    [anon_sym__alignof] = ACTIONS(2368),
    [anon_sym_alignof] = ACTIONS(2368),
    [anon_sym__Alignof] = ACTIONS(2368),
    [anon_sym_offsetof] = ACTIONS(2368),
    [anon_sym__Generic] = ACTIONS(2368),
    [anon_sym_asm] = ACTIONS(2368),
    [anon_sym___asm__] = ACTIONS(2368),
    [anon_sym___asm] = ACTIONS(2368),
    [sym_number_literal] = ACTIONS(2370),
    [anon_sym_L_SQUOTE] = ACTIONS(2370),
    [anon_sym_u_SQUOTE] = ACTIONS(2370),
    [anon_sym_U_SQUOTE] = ACTIONS(2370),
    [anon_sym_u8_SQUOTE] = ACTIONS(2370),
    [anon_sym_SQUOTE] = ACTIONS(2370),
    [anon_sym_AT] = ACTIONS(2368),
    [anon_sym_DQUOTE] = ACTIONS(2370),
    [anon_sym_L_DQUOTE] = ACTIONS(2370),
    [anon_sym_u_DQUOTE] = ACTIONS(2370),
    [anon_sym_U_DQUOTE] = ACTIONS(2370),
    [anon_sym_u8_DQUOTE] = ACTIONS(2370),
    [sym_true] = ACTIONS(2368),
    [sym_false] = ACTIONS(2368),
    [anon_sym_NULL] = ACTIONS(2368),
    [anon_sym_nullptr] = ACTIONS(2368),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2368),
    [anon_sym___typeof] = ACTIONS(2368),
    [anon_sym_typeof] = ACTIONS(2368),
    [anon_sym_ATimport] = ACTIONS(2370),
    [aux_sym_preproc_undef_token1] = ACTIONS(2368),
    [anon_sym_POUND] = ACTIONS(2368),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2368),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2368),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2368),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_API_AVAILABLE] = ACTIONS(2368),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_API_DEPRECATED] = ACTIONS(2368),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym___deprecated_msg] = ACTIONS(2368),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2368),
    [anon_sym_ATprotocol] = ACTIONS(2370),
    [anon_sym_ATinterface] = ACTIONS(2370),
    [anon_sym_ATimplementation] = ACTIONS(2370),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2370),
    [anon_sym_ATsynthesize] = ACTIONS(2370),
    [anon_sym_ATdynamic] = ACTIONS(2370),
    [anon_sym_ATtry] = ACTIONS(2370),
    [anon_sym___try] = ACTIONS(2368),
    [anon_sym_ATcatch] = ACTIONS(2370),
    [anon_sym___catch] = ACTIONS(2368),
    [anon_sym_ATfinally] = ACTIONS(2370),
    [anon_sym___finally] = ACTIONS(2368),
    [anon_sym_ATthrow] = ACTIONS(2370),
    [anon_sym_ATselector] = ACTIONS(2370),
    [anon_sym_ATavailable] = ACTIONS(2370),
    [anon_sym___builtin_available] = ACTIONS(2368),
    [anon_sym_va_arg] = ACTIONS(2368),
    [anon_sym_ATencode] = ACTIONS(2370),
    [anon_sym_ATsynchronized] = ACTIONS(2370),
    [anon_sym_BOOL] = ACTIONS(2368),
    [anon_sym_IMP] = ACTIONS(2368),
    [anon_sym_SEL] = ACTIONS(2368),
    [anon_sym_Class] = ACTIONS(2368),
    [anon_sym_id] = ACTIONS(2368),
  },
  [STATE(192)] = {
    [sym_identifier] = ACTIONS(2364),
    [aux_sym_preproc_include_token1] = ACTIONS(2364),
    [aux_sym_preproc_include_token2] = ACTIONS(2364),
    [aux_sym_preproc_def_token1] = ACTIONS(2364),
    [aux_sym_preproc_if_token1] = ACTIONS(2364),
    [aux_sym_preproc_if_token2] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2364),
    [aux_sym_preproc_else_token1] = ACTIONS(2364),
    [aux_sym_preproc_elif_token1] = ACTIONS(2364),
    [sym_preproc_directive] = ACTIONS(2364),
    [anon_sym_LPAREN2] = ACTIONS(2366),
    [anon_sym_BANG] = ACTIONS(2366),
    [anon_sym_TILDE] = ACTIONS(2366),
    [anon_sym_DASH] = ACTIONS(2364),
    [anon_sym_PLUS] = ACTIONS(2364),
    [anon_sym_STAR] = ACTIONS(2366),
    [anon_sym_CARET] = ACTIONS(2366),
    [anon_sym_AMP] = ACTIONS(2366),
    [anon_sym_SEMI] = ACTIONS(2366),
    [anon_sym___extension__] = ACTIONS(2364),
    [anon_sym_typedef] = ACTIONS(2364),
    [anon_sym_extern] = ACTIONS(2364),
    [anon_sym___attribute__] = ACTIONS(2364),
    [anon_sym___attribute] = ACTIONS(2364),
    [anon_sym_noreturn] = ACTIONS(2364),
    [anon_sym_LBRACK] = ACTIONS(2366),
    [anon_sym___declspec] = ACTIONS(2364),
    [anon_sym___cdecl] = ACTIONS(2364),
    [anon_sym___clrcall] = ACTIONS(2364),
    [anon_sym___stdcall] = ACTIONS(2364),
    [anon_sym___fastcall] = ACTIONS(2364),
    [anon_sym___thiscall] = ACTIONS(2364),
    [anon_sym___vectorcall] = ACTIONS(2364),
    [anon_sym_LBRACE] = ACTIONS(2366),
    [anon_sym_signed] = ACTIONS(2364),
    [anon_sym_unsigned] = ACTIONS(2364),
    [anon_sym_long] = ACTIONS(2364),
    [anon_sym_short] = ACTIONS(2364),
    [anon_sym_static] = ACTIONS(2364),
    [anon_sym_ATautoreleasepool] = ACTIONS(2366),
    [anon_sym_auto] = ACTIONS(2364),
    [anon_sym_register] = ACTIONS(2364),
    [anon_sym_inline] = ACTIONS(2364),
    [anon_sym___inline] = ACTIONS(2364),
    [anon_sym___inline__] = ACTIONS(2364),
    [anon_sym___forceinline] = ACTIONS(2364),
    [anon_sym_thread_local] = ACTIONS(2364),
    [anon_sym___thread] = ACTIONS(2364),
    [anon_sym_CG_EXTERN] = ACTIONS(2364),
    [anon_sym_CG_INLINE] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2364),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2364),
    [anon_sym_IBOutlet] = ACTIONS(2364),
    [anon_sym_IBInspectable] = ACTIONS(2364),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2364),
    [anon_sym_NS_INLINE] = ACTIONS(2364),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2364),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2364),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2364),
    [anon_sym_const] = ACTIONS(2364),
    [anon_sym_constexpr] = ACTIONS(2364),
    [anon_sym_volatile] = ACTIONS(2364),
    [anon_sym_restrict] = ACTIONS(2364),
    [anon_sym___restrict__] = ACTIONS(2364),
    [anon_sym__Atomic] = ACTIONS(2364),
    [anon_sym__Noreturn] = ACTIONS(2364),
    [anon_sym__Nonnull] = ACTIONS(2364),
    [anon_sym_nullable] = ACTIONS(2364),
    [anon_sym__Complex] = ACTIONS(2364),
    [anon_sym__Nullable] = ACTIONS(2364),
    [anon_sym__Nullable_result] = ACTIONS(2364),
    [anon_sym__Null_unspecified] = ACTIONS(2364),
    [anon_sym___autoreleasing] = ACTIONS(2364),
    [anon_sym___block] = ACTIONS(2364),
    [anon_sym___bridge] = ACTIONS(2364),
    [anon_sym___bridge_retained] = ACTIONS(2364),
    [anon_sym___bridge_transfer] = ACTIONS(2364),
    [anon_sym___complex] = ACTIONS(2364),
    [anon_sym___const] = ACTIONS(2364),
    [anon_sym___imag] = ACTIONS(2364),
    [anon_sym___kindof] = ACTIONS(2364),
    [anon_sym___nonnull] = ACTIONS(2364),
    [anon_sym___nullable] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2364),
    [anon_sym___real] = ACTIONS(2364),
    [anon_sym___strong] = ACTIONS(2364),
    [anon_sym___unsafe_unretained] = ACTIONS(2364),
    [anon_sym___unused] = ACTIONS(2364),
    [anon_sym___weak] = ACTIONS(2364),
    [anon_sym_alignas] = ACTIONS(2364),
    [anon_sym__Alignas] = ACTIONS(2364),
    [sym_primitive_type] = ACTIONS(2364),
    [anon_sym_enum] = ACTIONS(2364),
    [anon_sym_struct] = ACTIONS(2364),
    [anon_sym_union] = ACTIONS(2364),
    [anon_sym_if] = ACTIONS(2364),
    [anon_sym_else] = ACTIONS(2364),
    [anon_sym_switch] = ACTIONS(2364),
    [anon_sym_case] = ACTIONS(2364),
    [anon_sym_default] = ACTIONS(2364),
    [anon_sym_while] = ACTIONS(2364),
    [anon_sym_do] = ACTIONS(2364),
    [anon_sym_for] = ACTIONS(2364),
    [anon_sym_in] = ACTIONS(2364),
    [anon_sym_return] = ACTIONS(2364),
    [anon_sym_break] = ACTIONS(2364),
    [anon_sym_continue] = ACTIONS(2364),
    [anon_sym_goto] = ACTIONS(2364),
    [anon_sym_DASH_DASH] = ACTIONS(2366),
    [anon_sym_PLUS_PLUS] = ACTIONS(2366),
    [anon_sym_sizeof] = ACTIONS(2364),
    [anon_sym___alignof__] = ACTIONS(2364),
    [anon_sym___alignof] = ACTIONS(2364),
    [anon_sym__alignof] = ACTIONS(2364),
    [anon_sym_alignof] = ACTIONS(2364),
    [anon_sym__Alignof] = ACTIONS(2364),
    [anon_sym_offsetof] = ACTIONS(2364),
    [anon_sym__Generic] = ACTIONS(2364),
    [anon_sym_asm] = ACTIONS(2364),
    [anon_sym___asm__] = ACTIONS(2364),
    [anon_sym___asm] = ACTIONS(2364),
    [sym_number_literal] = ACTIONS(2366),
    [anon_sym_L_SQUOTE] = ACTIONS(2366),
    [anon_sym_u_SQUOTE] = ACTIONS(2366),
    [anon_sym_U_SQUOTE] = ACTIONS(2366),
    [anon_sym_u8_SQUOTE] = ACTIONS(2366),
    [anon_sym_SQUOTE] = ACTIONS(2366),
    [anon_sym_AT] = ACTIONS(2364),
    [anon_sym_DQUOTE] = ACTIONS(2366),
    [anon_sym_L_DQUOTE] = ACTIONS(2366),
    [anon_sym_u_DQUOTE] = ACTIONS(2366),
    [anon_sym_U_DQUOTE] = ACTIONS(2366),
    [anon_sym_u8_DQUOTE] = ACTIONS(2366),
    [sym_true] = ACTIONS(2364),
    [sym_false] = ACTIONS(2364),
    [anon_sym_NULL] = ACTIONS(2364),
    [anon_sym_nullptr] = ACTIONS(2364),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2364),
    [anon_sym___typeof] = ACTIONS(2364),
    [anon_sym_typeof] = ACTIONS(2364),
    [anon_sym_ATimport] = ACTIONS(2366),
    [aux_sym_preproc_undef_token1] = ACTIONS(2364),
    [anon_sym_POUND] = ACTIONS(2364),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2364),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2364),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2364),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_API_AVAILABLE] = ACTIONS(2364),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_API_DEPRECATED] = ACTIONS(2364),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym___deprecated_msg] = ACTIONS(2364),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2364),
    [anon_sym_ATprotocol] = ACTIONS(2366),
    [anon_sym_ATinterface] = ACTIONS(2366),
    [anon_sym_ATimplementation] = ACTIONS(2366),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2366),
    [anon_sym_ATsynthesize] = ACTIONS(2366),
    [anon_sym_ATdynamic] = ACTIONS(2366),
    [anon_sym_ATtry] = ACTIONS(2366),
    [anon_sym___try] = ACTIONS(2364),
    [anon_sym_ATcatch] = ACTIONS(2366),
    [anon_sym___catch] = ACTIONS(2364),
    [anon_sym_ATfinally] = ACTIONS(2366),
    [anon_sym___finally] = ACTIONS(2364),
    [anon_sym_ATthrow] = ACTIONS(2366),
    [anon_sym_ATselector] = ACTIONS(2366),
    [anon_sym_ATavailable] = ACTIONS(2366),
    [anon_sym___builtin_available] = ACTIONS(2364),
    [anon_sym_va_arg] = ACTIONS(2364),
    [anon_sym_ATencode] = ACTIONS(2366),
    [anon_sym_ATsynchronized] = ACTIONS(2366),
    [anon_sym_BOOL] = ACTIONS(2364),
    [anon_sym_IMP] = ACTIONS(2364),
    [anon_sym_SEL] = ACTIONS(2364),
    [anon_sym_Class] = ACTIONS(2364),
    [anon_sym_id] = ACTIONS(2364),
  },
  [STATE(193)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token2] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [aux_sym_preproc_else_token1] = ACTIONS(2346),
    [aux_sym_preproc_elif_token1] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATsynthesize] = ACTIONS(2348),
    [anon_sym_ATdynamic] = ACTIONS(2348),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATcatch] = ACTIONS(2348),
    [anon_sym___catch] = ACTIONS(2346),
    [anon_sym_ATfinally] = ACTIONS(2348),
    [anon_sym___finally] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(194)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token2] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [aux_sym_preproc_else_token1] = ACTIONS(2360),
    [aux_sym_preproc_elif_token1] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATsynthesize] = ACTIONS(2362),
    [anon_sym_ATdynamic] = ACTIONS(2362),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATcatch] = ACTIONS(2362),
    [anon_sym___catch] = ACTIONS(2360),
    [anon_sym_ATfinally] = ACTIONS(2362),
    [anon_sym___finally] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(195)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token2] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [aux_sym_preproc_else_token1] = ACTIONS(2350),
    [aux_sym_preproc_elif_token1] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATsynthesize] = ACTIONS(2352),
    [anon_sym_ATdynamic] = ACTIONS(2352),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATcatch] = ACTIONS(2352),
    [anon_sym___catch] = ACTIONS(2350),
    [anon_sym_ATfinally] = ACTIONS(2352),
    [anon_sym___finally] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(196)] = {
    [sym_catch_clause] = STATE(221),
    [sym_finally_clause] = STATE(758),
    [aux_sym_try_statement_repeat1] = STATE(221),
    [ts_builtin_sym_end] = ACTIONS(2308),
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATcatch] = ACTIONS(2386),
    [anon_sym___catch] = ACTIONS(2388),
    [anon_sym_ATfinally] = ACTIONS(2390),
    [anon_sym___finally] = ACTIONS(2392),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(197)] = {
    [sym_else_clause] = STATE(256),
    [sym_identifier] = ACTIONS(2394),
    [aux_sym_preproc_include_token1] = ACTIONS(2394),
    [aux_sym_preproc_include_token2] = ACTIONS(2394),
    [aux_sym_preproc_def_token1] = ACTIONS(2394),
    [aux_sym_preproc_if_token1] = ACTIONS(2394),
    [aux_sym_preproc_if_token2] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2394),
    [aux_sym_preproc_else_token1] = ACTIONS(2394),
    [aux_sym_preproc_elif_token1] = ACTIONS(2394),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2394),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2394),
    [sym_preproc_directive] = ACTIONS(2394),
    [anon_sym_LPAREN2] = ACTIONS(2396),
    [anon_sym_BANG] = ACTIONS(2396),
    [anon_sym_TILDE] = ACTIONS(2396),
    [anon_sym_DASH] = ACTIONS(2394),
    [anon_sym_PLUS] = ACTIONS(2394),
    [anon_sym_STAR] = ACTIONS(2396),
    [anon_sym_CARET] = ACTIONS(2396),
    [anon_sym_AMP] = ACTIONS(2396),
    [anon_sym_SEMI] = ACTIONS(2396),
    [anon_sym___extension__] = ACTIONS(2394),
    [anon_sym_typedef] = ACTIONS(2394),
    [anon_sym_extern] = ACTIONS(2394),
    [anon_sym___attribute__] = ACTIONS(2394),
    [anon_sym___attribute] = ACTIONS(2394),
    [anon_sym_noreturn] = ACTIONS(2394),
    [anon_sym_LBRACK] = ACTIONS(2396),
    [anon_sym___declspec] = ACTIONS(2394),
    [anon_sym___cdecl] = ACTIONS(2394),
    [anon_sym___clrcall] = ACTIONS(2394),
    [anon_sym___stdcall] = ACTIONS(2394),
    [anon_sym___fastcall] = ACTIONS(2394),
    [anon_sym___thiscall] = ACTIONS(2394),
    [anon_sym___vectorcall] = ACTIONS(2394),
    [anon_sym_LBRACE] = ACTIONS(2396),
    [anon_sym_signed] = ACTIONS(2394),
    [anon_sym_unsigned] = ACTIONS(2394),
    [anon_sym_long] = ACTIONS(2394),
    [anon_sym_short] = ACTIONS(2394),
    [anon_sym_static] = ACTIONS(2394),
    [anon_sym_ATautoreleasepool] = ACTIONS(2396),
    [anon_sym_auto] = ACTIONS(2394),
    [anon_sym_register] = ACTIONS(2394),
    [anon_sym_inline] = ACTIONS(2394),
    [anon_sym___inline] = ACTIONS(2394),
    [anon_sym___inline__] = ACTIONS(2394),
    [anon_sym___forceinline] = ACTIONS(2394),
    [anon_sym_thread_local] = ACTIONS(2394),
    [anon_sym___thread] = ACTIONS(2394),
    [anon_sym_CG_EXTERN] = ACTIONS(2394),
    [anon_sym_CG_INLINE] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2394),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2394),
    [anon_sym_IBOutlet] = ACTIONS(2394),
    [anon_sym_IBInspectable] = ACTIONS(2394),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2394),
    [anon_sym_NS_INLINE] = ACTIONS(2394),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2394),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2394),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2394),
    [anon_sym_const] = ACTIONS(2394),
    [anon_sym_constexpr] = ACTIONS(2394),
    [anon_sym_volatile] = ACTIONS(2394),
    [anon_sym_restrict] = ACTIONS(2394),
    [anon_sym___restrict__] = ACTIONS(2394),
    [anon_sym__Atomic] = ACTIONS(2394),
    [anon_sym__Noreturn] = ACTIONS(2394),
    [anon_sym__Nonnull] = ACTIONS(2394),
    [anon_sym_nullable] = ACTIONS(2394),
    [anon_sym__Complex] = ACTIONS(2394),
    [anon_sym__Nullable] = ACTIONS(2394),
    [anon_sym__Nullable_result] = ACTIONS(2394),
    [anon_sym__Null_unspecified] = ACTIONS(2394),
    [anon_sym___autoreleasing] = ACTIONS(2394),
    [anon_sym___block] = ACTIONS(2394),
    [anon_sym___bridge] = ACTIONS(2394),
    [anon_sym___bridge_retained] = ACTIONS(2394),
    [anon_sym___bridge_transfer] = ACTIONS(2394),
    [anon_sym___complex] = ACTIONS(2394),
    [anon_sym___const] = ACTIONS(2394),
    [anon_sym___imag] = ACTIONS(2394),
    [anon_sym___kindof] = ACTIONS(2394),
    [anon_sym___nonnull] = ACTIONS(2394),
    [anon_sym___nullable] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2394),
    [anon_sym___real] = ACTIONS(2394),
    [anon_sym___strong] = ACTIONS(2394),
    [anon_sym___unsafe_unretained] = ACTIONS(2394),
    [anon_sym___unused] = ACTIONS(2394),
    [anon_sym___weak] = ACTIONS(2394),
    [anon_sym_alignas] = ACTIONS(2394),
    [anon_sym__Alignas] = ACTIONS(2394),
    [sym_primitive_type] = ACTIONS(2394),
    [anon_sym_enum] = ACTIONS(2394),
    [anon_sym_struct] = ACTIONS(2394),
    [anon_sym_union] = ACTIONS(2394),
    [anon_sym_if] = ACTIONS(2394),
    [anon_sym_else] = ACTIONS(2398),
    [anon_sym_switch] = ACTIONS(2394),
    [anon_sym_case] = ACTIONS(2394),
    [anon_sym_default] = ACTIONS(2394),
    [anon_sym_while] = ACTIONS(2394),
    [anon_sym_do] = ACTIONS(2394),
    [anon_sym_for] = ACTIONS(2394),
    [anon_sym_in] = ACTIONS(2394),
    [anon_sym_return] = ACTIONS(2394),
    [anon_sym_break] = ACTIONS(2394),
    [anon_sym_continue] = ACTIONS(2394),
    [anon_sym_goto] = ACTIONS(2394),
    [anon_sym_DASH_DASH] = ACTIONS(2396),
    [anon_sym_PLUS_PLUS] = ACTIONS(2396),
    [anon_sym_sizeof] = ACTIONS(2394),
    [anon_sym___alignof__] = ACTIONS(2394),
    [anon_sym___alignof] = ACTIONS(2394),
    [anon_sym__alignof] = ACTIONS(2394),
    [anon_sym_alignof] = ACTIONS(2394),
    [anon_sym__Alignof] = ACTIONS(2394),
    [anon_sym_offsetof] = ACTIONS(2394),
    [anon_sym__Generic] = ACTIONS(2394),
    [anon_sym_asm] = ACTIONS(2394),
    [anon_sym___asm__] = ACTIONS(2394),
    [anon_sym___asm] = ACTIONS(2394),
    [sym_number_literal] = ACTIONS(2396),
    [anon_sym_L_SQUOTE] = ACTIONS(2396),
    [anon_sym_u_SQUOTE] = ACTIONS(2396),
    [anon_sym_U_SQUOTE] = ACTIONS(2396),
    [anon_sym_u8_SQUOTE] = ACTIONS(2396),
    [anon_sym_SQUOTE] = ACTIONS(2396),
    [anon_sym_AT] = ACTIONS(2394),
    [anon_sym_DQUOTE] = ACTIONS(2396),
    [anon_sym_L_DQUOTE] = ACTIONS(2396),
    [anon_sym_u_DQUOTE] = ACTIONS(2396),
    [anon_sym_U_DQUOTE] = ACTIONS(2396),
    [anon_sym_u8_DQUOTE] = ACTIONS(2396),
    [sym_true] = ACTIONS(2394),
    [sym_false] = ACTIONS(2394),
    [anon_sym_NULL] = ACTIONS(2394),
    [anon_sym_nullptr] = ACTIONS(2394),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2394),
    [anon_sym___typeof] = ACTIONS(2394),
    [anon_sym_typeof] = ACTIONS(2394),
    [anon_sym_ATimport] = ACTIONS(2396),
    [aux_sym_preproc_undef_token1] = ACTIONS(2394),
    [anon_sym_POUND] = ACTIONS(2394),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2394),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2394),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2394),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_API_AVAILABLE] = ACTIONS(2394),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_API_DEPRECATED] = ACTIONS(2394),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym___deprecated_msg] = ACTIONS(2394),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2394),
    [anon_sym_ATprotocol] = ACTIONS(2396),
    [anon_sym_ATinterface] = ACTIONS(2396),
    [anon_sym_ATimplementation] = ACTIONS(2396),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2396),
    [anon_sym_ATsynthesize] = ACTIONS(2396),
    [anon_sym_ATdynamic] = ACTIONS(2396),
    [anon_sym_ATtry] = ACTIONS(2396),
    [anon_sym___try] = ACTIONS(2394),
    [anon_sym_ATthrow] = ACTIONS(2396),
    [anon_sym_ATselector] = ACTIONS(2396),
    [anon_sym_ATavailable] = ACTIONS(2396),
    [anon_sym___builtin_available] = ACTIONS(2394),
    [anon_sym_va_arg] = ACTIONS(2394),
    [anon_sym_ATencode] = ACTIONS(2396),
    [anon_sym_ATsynchronized] = ACTIONS(2396),
    [anon_sym_BOOL] = ACTIONS(2394),
    [anon_sym_IMP] = ACTIONS(2394),
    [anon_sym_SEL] = ACTIONS(2394),
    [anon_sym_Class] = ACTIONS(2394),
    [anon_sym_id] = ACTIONS(2394),
  },
  [STATE(198)] = {
    [sym_catch_clause] = STATE(275),
    [sym_finally_clause] = STATE(788),
    [aux_sym_try_statement_repeat1] = STATE(275),
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_RBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATcatch] = ACTIONS(2400),
    [anon_sym___catch] = ACTIONS(2402),
    [anon_sym_ATfinally] = ACTIONS(2404),
    [anon_sym___finally] = ACTIONS(2406),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(199)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(200)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(201)] = {
    [sym_identifier] = ACTIONS(2412),
    [aux_sym_preproc_include_token1] = ACTIONS(2412),
    [aux_sym_preproc_include_token2] = ACTIONS(2412),
    [aux_sym_preproc_def_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token2] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2412),
    [aux_sym_preproc_else_token1] = ACTIONS(2412),
    [aux_sym_preproc_elif_token1] = ACTIONS(2412),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2412),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2412),
    [sym_preproc_directive] = ACTIONS(2412),
    [anon_sym_LPAREN2] = ACTIONS(2414),
    [anon_sym_BANG] = ACTIONS(2414),
    [anon_sym_TILDE] = ACTIONS(2414),
    [anon_sym_DASH] = ACTIONS(2412),
    [anon_sym_PLUS] = ACTIONS(2412),
    [anon_sym_STAR] = ACTIONS(2414),
    [anon_sym_CARET] = ACTIONS(2414),
    [anon_sym_AMP] = ACTIONS(2414),
    [anon_sym_SEMI] = ACTIONS(2414),
    [anon_sym___extension__] = ACTIONS(2412),
    [anon_sym_typedef] = ACTIONS(2412),
    [anon_sym_extern] = ACTIONS(2412),
    [anon_sym___attribute__] = ACTIONS(2412),
    [anon_sym___attribute] = ACTIONS(2412),
    [anon_sym_noreturn] = ACTIONS(2412),
    [anon_sym_LBRACK] = ACTIONS(2414),
    [anon_sym___declspec] = ACTIONS(2412),
    [anon_sym___cdecl] = ACTIONS(2412),
    [anon_sym___clrcall] = ACTIONS(2412),
    [anon_sym___stdcall] = ACTIONS(2412),
    [anon_sym___fastcall] = ACTIONS(2412),
    [anon_sym___thiscall] = ACTIONS(2412),
    [anon_sym___vectorcall] = ACTIONS(2412),
    [anon_sym_LBRACE] = ACTIONS(2414),
    [anon_sym_signed] = ACTIONS(2412),
    [anon_sym_unsigned] = ACTIONS(2412),
    [anon_sym_long] = ACTIONS(2412),
    [anon_sym_short] = ACTIONS(2412),
    [anon_sym_static] = ACTIONS(2412),
    [anon_sym_ATautoreleasepool] = ACTIONS(2414),
    [anon_sym_auto] = ACTIONS(2412),
    [anon_sym_register] = ACTIONS(2412),
    [anon_sym_inline] = ACTIONS(2412),
    [anon_sym___inline] = ACTIONS(2412),
    [anon_sym___inline__] = ACTIONS(2412),
    [anon_sym___forceinline] = ACTIONS(2412),
    [anon_sym_thread_local] = ACTIONS(2412),
    [anon_sym___thread] = ACTIONS(2412),
    [anon_sym_CG_EXTERN] = ACTIONS(2412),
    [anon_sym_CG_INLINE] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2412),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2412),
    [anon_sym_IBOutlet] = ACTIONS(2412),
    [anon_sym_IBInspectable] = ACTIONS(2412),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2412),
    [anon_sym_NS_INLINE] = ACTIONS(2412),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2412),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2412),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2412),
    [anon_sym_const] = ACTIONS(2412),
    [anon_sym_constexpr] = ACTIONS(2412),
    [anon_sym_volatile] = ACTIONS(2412),
    [anon_sym_restrict] = ACTIONS(2412),
    [anon_sym___restrict__] = ACTIONS(2412),
    [anon_sym__Atomic] = ACTIONS(2412),
    [anon_sym__Noreturn] = ACTIONS(2412),
    [anon_sym__Nonnull] = ACTIONS(2412),
    [anon_sym_nullable] = ACTIONS(2412),
    [anon_sym__Complex] = ACTIONS(2412),
    [anon_sym__Nullable] = ACTIONS(2412),
    [anon_sym__Nullable_result] = ACTIONS(2412),
    [anon_sym__Null_unspecified] = ACTIONS(2412),
    [anon_sym___autoreleasing] = ACTIONS(2412),
    [anon_sym___block] = ACTIONS(2412),
    [anon_sym___bridge] = ACTIONS(2412),
    [anon_sym___bridge_retained] = ACTIONS(2412),
    [anon_sym___bridge_transfer] = ACTIONS(2412),
    [anon_sym___complex] = ACTIONS(2412),
    [anon_sym___const] = ACTIONS(2412),
    [anon_sym___imag] = ACTIONS(2412),
    [anon_sym___kindof] = ACTIONS(2412),
    [anon_sym___nonnull] = ACTIONS(2412),
    [anon_sym___nullable] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2412),
    [anon_sym___real] = ACTIONS(2412),
    [anon_sym___strong] = ACTIONS(2412),
    [anon_sym___unsafe_unretained] = ACTIONS(2412),
    [anon_sym___unused] = ACTIONS(2412),
    [anon_sym___weak] = ACTIONS(2412),
    [anon_sym_alignas] = ACTIONS(2412),
    [anon_sym__Alignas] = ACTIONS(2412),
    [sym_primitive_type] = ACTIONS(2412),
    [anon_sym_enum] = ACTIONS(2412),
    [anon_sym_struct] = ACTIONS(2412),
    [anon_sym_union] = ACTIONS(2412),
    [anon_sym_if] = ACTIONS(2412),
    [anon_sym_else] = ACTIONS(2412),
    [anon_sym_switch] = ACTIONS(2412),
    [anon_sym_case] = ACTIONS(2412),
    [anon_sym_default] = ACTIONS(2412),
    [anon_sym_while] = ACTIONS(2412),
    [anon_sym_do] = ACTIONS(2412),
    [anon_sym_for] = ACTIONS(2412),
    [anon_sym_in] = ACTIONS(2412),
    [anon_sym_return] = ACTIONS(2412),
    [anon_sym_break] = ACTIONS(2412),
    [anon_sym_continue] = ACTIONS(2412),
    [anon_sym_goto] = ACTIONS(2412),
    [anon_sym_DASH_DASH] = ACTIONS(2414),
    [anon_sym_PLUS_PLUS] = ACTIONS(2414),
    [anon_sym_sizeof] = ACTIONS(2412),
    [anon_sym___alignof__] = ACTIONS(2412),
    [anon_sym___alignof] = ACTIONS(2412),
    [anon_sym__alignof] = ACTIONS(2412),
    [anon_sym_alignof] = ACTIONS(2412),
    [anon_sym__Alignof] = ACTIONS(2412),
    [anon_sym_offsetof] = ACTIONS(2412),
    [anon_sym__Generic] = ACTIONS(2412),
    [anon_sym_asm] = ACTIONS(2412),
    [anon_sym___asm__] = ACTIONS(2412),
    [anon_sym___asm] = ACTIONS(2412),
    [sym_number_literal] = ACTIONS(2414),
    [anon_sym_L_SQUOTE] = ACTIONS(2414),
    [anon_sym_u_SQUOTE] = ACTIONS(2414),
    [anon_sym_U_SQUOTE] = ACTIONS(2414),
    [anon_sym_u8_SQUOTE] = ACTIONS(2414),
    [anon_sym_SQUOTE] = ACTIONS(2414),
    [anon_sym_AT] = ACTIONS(2412),
    [anon_sym_DQUOTE] = ACTIONS(2414),
    [anon_sym_L_DQUOTE] = ACTIONS(2414),
    [anon_sym_u_DQUOTE] = ACTIONS(2414),
    [anon_sym_U_DQUOTE] = ACTIONS(2414),
    [anon_sym_u8_DQUOTE] = ACTIONS(2414),
    [sym_true] = ACTIONS(2412),
    [sym_false] = ACTIONS(2412),
    [anon_sym_NULL] = ACTIONS(2412),
    [anon_sym_nullptr] = ACTIONS(2412),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2412),
    [anon_sym___typeof] = ACTIONS(2412),
    [anon_sym_typeof] = ACTIONS(2412),
    [anon_sym_ATimport] = ACTIONS(2414),
    [aux_sym_preproc_undef_token1] = ACTIONS(2412),
    [anon_sym_POUND] = ACTIONS(2412),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2412),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2412),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2412),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_API_AVAILABLE] = ACTIONS(2412),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_API_DEPRECATED] = ACTIONS(2412),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym___deprecated_msg] = ACTIONS(2412),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2412),
    [anon_sym_ATprotocol] = ACTIONS(2414),
    [anon_sym_ATinterface] = ACTIONS(2414),
    [anon_sym_ATimplementation] = ACTIONS(2414),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2414),
    [anon_sym_ATsynthesize] = ACTIONS(2414),
    [anon_sym_ATdynamic] = ACTIONS(2414),
    [anon_sym_ATtry] = ACTIONS(2414),
    [anon_sym___try] = ACTIONS(2412),
    [anon_sym_ATthrow] = ACTIONS(2414),
    [anon_sym_ATselector] = ACTIONS(2414),
    [anon_sym_ATavailable] = ACTIONS(2414),
    [anon_sym___builtin_available] = ACTIONS(2412),
    [anon_sym_va_arg] = ACTIONS(2412),
    [anon_sym_ATencode] = ACTIONS(2414),
    [anon_sym_ATsynchronized] = ACTIONS(2414),
    [anon_sym_BOOL] = ACTIONS(2412),
    [anon_sym_IMP] = ACTIONS(2412),
    [anon_sym_SEL] = ACTIONS(2412),
    [anon_sym_Class] = ACTIONS(2412),
    [anon_sym_id] = ACTIONS(2412),
  },
  [STATE(202)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token2] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [aux_sym_preproc_else_token1] = ACTIONS(2360),
    [aux_sym_preproc_elif_token1] = ACTIONS(2360),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATsynthesize] = ACTIONS(2362),
    [anon_sym_ATdynamic] = ACTIONS(2362),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(203)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(204)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token2] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [aux_sym_preproc_else_token1] = ACTIONS(2416),
    [aux_sym_preproc_elif_token1] = ACTIONS(2416),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(205)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(206)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(207)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(208)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token2] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [aux_sym_preproc_else_token1] = ACTIONS(2420),
    [aux_sym_preproc_elif_token1] = ACTIONS(2420),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(209)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(210)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(211)] = {
    [sym_identifier] = ACTIONS(2364),
    [aux_sym_preproc_include_token1] = ACTIONS(2364),
    [aux_sym_preproc_include_token2] = ACTIONS(2364),
    [aux_sym_preproc_def_token1] = ACTIONS(2364),
    [aux_sym_preproc_if_token1] = ACTIONS(2364),
    [aux_sym_preproc_if_token2] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2364),
    [sym_preproc_directive] = ACTIONS(2364),
    [anon_sym_LPAREN2] = ACTIONS(2366),
    [anon_sym_BANG] = ACTIONS(2366),
    [anon_sym_TILDE] = ACTIONS(2366),
    [anon_sym_DASH] = ACTIONS(2364),
    [anon_sym_PLUS] = ACTIONS(2364),
    [anon_sym_STAR] = ACTIONS(2366),
    [anon_sym_CARET] = ACTIONS(2366),
    [anon_sym_AMP] = ACTIONS(2366),
    [anon_sym_SEMI] = ACTIONS(2366),
    [anon_sym___extension__] = ACTIONS(2364),
    [anon_sym_typedef] = ACTIONS(2364),
    [anon_sym_extern] = ACTIONS(2364),
    [anon_sym___attribute__] = ACTIONS(2364),
    [anon_sym___attribute] = ACTIONS(2364),
    [anon_sym_noreturn] = ACTIONS(2364),
    [anon_sym_LBRACK] = ACTIONS(2366),
    [anon_sym___declspec] = ACTIONS(2364),
    [anon_sym___cdecl] = ACTIONS(2364),
    [anon_sym___clrcall] = ACTIONS(2364),
    [anon_sym___stdcall] = ACTIONS(2364),
    [anon_sym___fastcall] = ACTIONS(2364),
    [anon_sym___thiscall] = ACTIONS(2364),
    [anon_sym___vectorcall] = ACTIONS(2364),
    [anon_sym_LBRACE] = ACTIONS(2366),
    [anon_sym_signed] = ACTIONS(2364),
    [anon_sym_unsigned] = ACTIONS(2364),
    [anon_sym_long] = ACTIONS(2364),
    [anon_sym_short] = ACTIONS(2364),
    [anon_sym_static] = ACTIONS(2364),
    [anon_sym_ATautoreleasepool] = ACTIONS(2366),
    [anon_sym_auto] = ACTIONS(2364),
    [anon_sym_register] = ACTIONS(2364),
    [anon_sym_inline] = ACTIONS(2364),
    [anon_sym___inline] = ACTIONS(2364),
    [anon_sym___inline__] = ACTIONS(2364),
    [anon_sym___forceinline] = ACTIONS(2364),
    [anon_sym_thread_local] = ACTIONS(2364),
    [anon_sym___thread] = ACTIONS(2364),
    [anon_sym_CG_EXTERN] = ACTIONS(2364),
    [anon_sym_CG_INLINE] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2364),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2364),
    [anon_sym_IBOutlet] = ACTIONS(2364),
    [anon_sym_IBInspectable] = ACTIONS(2364),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2364),
    [anon_sym_NS_INLINE] = ACTIONS(2364),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2364),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2364),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2364),
    [anon_sym_const] = ACTIONS(2364),
    [anon_sym_constexpr] = ACTIONS(2364),
    [anon_sym_volatile] = ACTIONS(2364),
    [anon_sym_restrict] = ACTIONS(2364),
    [anon_sym___restrict__] = ACTIONS(2364),
    [anon_sym__Atomic] = ACTIONS(2364),
    [anon_sym__Noreturn] = ACTIONS(2364),
    [anon_sym__Nonnull] = ACTIONS(2364),
    [anon_sym_nullable] = ACTIONS(2364),
    [anon_sym__Complex] = ACTIONS(2364),
    [anon_sym__Nullable] = ACTIONS(2364),
    [anon_sym__Nullable_result] = ACTIONS(2364),
    [anon_sym__Null_unspecified] = ACTIONS(2364),
    [anon_sym___autoreleasing] = ACTIONS(2364),
    [anon_sym___block] = ACTIONS(2364),
    [anon_sym___bridge] = ACTIONS(2364),
    [anon_sym___bridge_retained] = ACTIONS(2364),
    [anon_sym___bridge_transfer] = ACTIONS(2364),
    [anon_sym___complex] = ACTIONS(2364),
    [anon_sym___const] = ACTIONS(2364),
    [anon_sym___imag] = ACTIONS(2364),
    [anon_sym___kindof] = ACTIONS(2364),
    [anon_sym___nonnull] = ACTIONS(2364),
    [anon_sym___nullable] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2364),
    [anon_sym___real] = ACTIONS(2364),
    [anon_sym___strong] = ACTIONS(2364),
    [anon_sym___unsafe_unretained] = ACTIONS(2364),
    [anon_sym___unused] = ACTIONS(2364),
    [anon_sym___weak] = ACTIONS(2364),
    [anon_sym_alignas] = ACTIONS(2364),
    [anon_sym__Alignas] = ACTIONS(2364),
    [sym_primitive_type] = ACTIONS(2364),
    [anon_sym_enum] = ACTIONS(2364),
    [anon_sym_struct] = ACTIONS(2364),
    [anon_sym_union] = ACTIONS(2364),
    [anon_sym_if] = ACTIONS(2364),
    [anon_sym_else] = ACTIONS(2364),
    [anon_sym_switch] = ACTIONS(2364),
    [anon_sym_case] = ACTIONS(2364),
    [anon_sym_default] = ACTIONS(2364),
    [anon_sym_while] = ACTIONS(2364),
    [anon_sym_do] = ACTIONS(2364),
    [anon_sym_for] = ACTIONS(2364),
    [anon_sym_in] = ACTIONS(2364),
    [anon_sym_return] = ACTIONS(2364),
    [anon_sym_break] = ACTIONS(2364),
    [anon_sym_continue] = ACTIONS(2364),
    [anon_sym_goto] = ACTIONS(2364),
    [anon_sym_DASH_DASH] = ACTIONS(2366),
    [anon_sym_PLUS_PLUS] = ACTIONS(2366),
    [anon_sym_sizeof] = ACTIONS(2364),
    [anon_sym___alignof__] = ACTIONS(2364),
    [anon_sym___alignof] = ACTIONS(2364),
    [anon_sym__alignof] = ACTIONS(2364),
    [anon_sym_alignof] = ACTIONS(2364),
    [anon_sym__Alignof] = ACTIONS(2364),
    [anon_sym_offsetof] = ACTIONS(2364),
    [anon_sym__Generic] = ACTIONS(2364),
    [anon_sym_asm] = ACTIONS(2364),
    [anon_sym___asm__] = ACTIONS(2364),
    [anon_sym___asm] = ACTIONS(2364),
    [sym_number_literal] = ACTIONS(2366),
    [anon_sym_L_SQUOTE] = ACTIONS(2366),
    [anon_sym_u_SQUOTE] = ACTIONS(2366),
    [anon_sym_U_SQUOTE] = ACTIONS(2366),
    [anon_sym_u8_SQUOTE] = ACTIONS(2366),
    [anon_sym_SQUOTE] = ACTIONS(2366),
    [anon_sym_AT] = ACTIONS(2364),
    [anon_sym_DQUOTE] = ACTIONS(2366),
    [anon_sym_L_DQUOTE] = ACTIONS(2366),
    [anon_sym_u_DQUOTE] = ACTIONS(2366),
    [anon_sym_U_DQUOTE] = ACTIONS(2366),
    [anon_sym_u8_DQUOTE] = ACTIONS(2366),
    [sym_true] = ACTIONS(2364),
    [sym_false] = ACTIONS(2364),
    [anon_sym_NULL] = ACTIONS(2364),
    [anon_sym_nullptr] = ACTIONS(2364),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2364),
    [anon_sym___typeof] = ACTIONS(2364),
    [anon_sym_typeof] = ACTIONS(2364),
    [anon_sym_ATimport] = ACTIONS(2366),
    [aux_sym_preproc_undef_token1] = ACTIONS(2364),
    [anon_sym_POUND] = ACTIONS(2364),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2364),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2364),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2364),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_API_AVAILABLE] = ACTIONS(2364),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_API_DEPRECATED] = ACTIONS(2364),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym___deprecated_msg] = ACTIONS(2364),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2364),
    [anon_sym_ATprotocol] = ACTIONS(2366),
    [anon_sym_ATinterface] = ACTIONS(2366),
    [anon_sym_ATimplementation] = ACTIONS(2366),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2366),
    [anon_sym_ATsynthesize] = ACTIONS(2366),
    [anon_sym_ATdynamic] = ACTIONS(2366),
    [anon_sym_ATtry] = ACTIONS(2366),
    [anon_sym___try] = ACTIONS(2364),
    [anon_sym_ATcatch] = ACTIONS(2366),
    [anon_sym___catch] = ACTIONS(2364),
    [anon_sym_ATfinally] = ACTIONS(2366),
    [anon_sym___finally] = ACTIONS(2364),
    [anon_sym_ATthrow] = ACTIONS(2366),
    [anon_sym_ATselector] = ACTIONS(2366),
    [anon_sym_ATavailable] = ACTIONS(2366),
    [anon_sym___builtin_available] = ACTIONS(2364),
    [anon_sym_va_arg] = ACTIONS(2364),
    [anon_sym_ATencode] = ACTIONS(2366),
    [anon_sym_ATsynchronized] = ACTIONS(2366),
    [anon_sym_BOOL] = ACTIONS(2364),
    [anon_sym_IMP] = ACTIONS(2364),
    [anon_sym_SEL] = ACTIONS(2364),
    [anon_sym_Class] = ACTIONS(2364),
    [anon_sym_id] = ACTIONS(2364),
  },
  [STATE(212)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token2] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [aux_sym_preproc_else_token1] = ACTIONS(2424),
    [aux_sym_preproc_elif_token1] = ACTIONS(2424),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(213)] = {
    [sym_identifier] = ACTIONS(2368),
    [aux_sym_preproc_include_token1] = ACTIONS(2368),
    [aux_sym_preproc_include_token2] = ACTIONS(2368),
    [aux_sym_preproc_def_token1] = ACTIONS(2368),
    [aux_sym_preproc_if_token1] = ACTIONS(2368),
    [aux_sym_preproc_if_token2] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2368),
    [sym_preproc_directive] = ACTIONS(2368),
    [anon_sym_LPAREN2] = ACTIONS(2370),
    [anon_sym_BANG] = ACTIONS(2370),
    [anon_sym_TILDE] = ACTIONS(2370),
    [anon_sym_DASH] = ACTIONS(2368),
    [anon_sym_PLUS] = ACTIONS(2368),
    [anon_sym_STAR] = ACTIONS(2370),
    [anon_sym_CARET] = ACTIONS(2370),
    [anon_sym_AMP] = ACTIONS(2370),
    [anon_sym_SEMI] = ACTIONS(2370),
    [anon_sym___extension__] = ACTIONS(2368),
    [anon_sym_typedef] = ACTIONS(2368),
    [anon_sym_extern] = ACTIONS(2368),
    [anon_sym___attribute__] = ACTIONS(2368),
    [anon_sym___attribute] = ACTIONS(2368),
    [anon_sym_noreturn] = ACTIONS(2368),
    [anon_sym_LBRACK] = ACTIONS(2370),
    [anon_sym___declspec] = ACTIONS(2368),
    [anon_sym___cdecl] = ACTIONS(2368),
    [anon_sym___clrcall] = ACTIONS(2368),
    [anon_sym___stdcall] = ACTIONS(2368),
    [anon_sym___fastcall] = ACTIONS(2368),
    [anon_sym___thiscall] = ACTIONS(2368),
    [anon_sym___vectorcall] = ACTIONS(2368),
    [anon_sym_LBRACE] = ACTIONS(2370),
    [anon_sym_signed] = ACTIONS(2368),
    [anon_sym_unsigned] = ACTIONS(2368),
    [anon_sym_long] = ACTIONS(2368),
    [anon_sym_short] = ACTIONS(2368),
    [anon_sym_static] = ACTIONS(2368),
    [anon_sym_ATautoreleasepool] = ACTIONS(2370),
    [anon_sym_auto] = ACTIONS(2368),
    [anon_sym_register] = ACTIONS(2368),
    [anon_sym_inline] = ACTIONS(2368),
    [anon_sym___inline] = ACTIONS(2368),
    [anon_sym___inline__] = ACTIONS(2368),
    [anon_sym___forceinline] = ACTIONS(2368),
    [anon_sym_thread_local] = ACTIONS(2368),
    [anon_sym___thread] = ACTIONS(2368),
    [anon_sym_CG_EXTERN] = ACTIONS(2368),
    [anon_sym_CG_INLINE] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2368),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2368),
    [anon_sym_IBOutlet] = ACTIONS(2368),
    [anon_sym_IBInspectable] = ACTIONS(2368),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2368),
    [anon_sym_NS_INLINE] = ACTIONS(2368),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2368),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2368),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2368),
    [anon_sym_const] = ACTIONS(2368),
    [anon_sym_constexpr] = ACTIONS(2368),
    [anon_sym_volatile] = ACTIONS(2368),
    [anon_sym_restrict] = ACTIONS(2368),
    [anon_sym___restrict__] = ACTIONS(2368),
    [anon_sym__Atomic] = ACTIONS(2368),
    [anon_sym__Noreturn] = ACTIONS(2368),
    [anon_sym__Nonnull] = ACTIONS(2368),
    [anon_sym_nullable] = ACTIONS(2368),
    [anon_sym__Complex] = ACTIONS(2368),
    [anon_sym__Nullable] = ACTIONS(2368),
    [anon_sym__Nullable_result] = ACTIONS(2368),
    [anon_sym__Null_unspecified] = ACTIONS(2368),
    [anon_sym___autoreleasing] = ACTIONS(2368),
    [anon_sym___block] = ACTIONS(2368),
    [anon_sym___bridge] = ACTIONS(2368),
    [anon_sym___bridge_retained] = ACTIONS(2368),
    [anon_sym___bridge_transfer] = ACTIONS(2368),
    [anon_sym___complex] = ACTIONS(2368),
    [anon_sym___const] = ACTIONS(2368),
    [anon_sym___imag] = ACTIONS(2368),
    [anon_sym___kindof] = ACTIONS(2368),
    [anon_sym___nonnull] = ACTIONS(2368),
    [anon_sym___nullable] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2368),
    [anon_sym___real] = ACTIONS(2368),
    [anon_sym___strong] = ACTIONS(2368),
    [anon_sym___unsafe_unretained] = ACTIONS(2368),
    [anon_sym___unused] = ACTIONS(2368),
    [anon_sym___weak] = ACTIONS(2368),
    [anon_sym_alignas] = ACTIONS(2368),
    [anon_sym__Alignas] = ACTIONS(2368),
    [sym_primitive_type] = ACTIONS(2368),
    [anon_sym_enum] = ACTIONS(2368),
    [anon_sym_struct] = ACTIONS(2368),
    [anon_sym_union] = ACTIONS(2368),
    [anon_sym_if] = ACTIONS(2368),
    [anon_sym_else] = ACTIONS(2368),
    [anon_sym_switch] = ACTIONS(2368),
    [anon_sym_case] = ACTIONS(2368),
    [anon_sym_default] = ACTIONS(2368),
    [anon_sym_while] = ACTIONS(2368),
    [anon_sym_do] = ACTIONS(2368),
    [anon_sym_for] = ACTIONS(2368),
    [anon_sym_in] = ACTIONS(2368),
    [anon_sym_return] = ACTIONS(2368),
    [anon_sym_break] = ACTIONS(2368),
    [anon_sym_continue] = ACTIONS(2368),
    [anon_sym_goto] = ACTIONS(2368),
    [anon_sym_DASH_DASH] = ACTIONS(2370),
    [anon_sym_PLUS_PLUS] = ACTIONS(2370),
    [anon_sym_sizeof] = ACTIONS(2368),
    [anon_sym___alignof__] = ACTIONS(2368),
    [anon_sym___alignof] = ACTIONS(2368),
    [anon_sym__alignof] = ACTIONS(2368),
    [anon_sym_alignof] = ACTIONS(2368),
    [anon_sym__Alignof] = ACTIONS(2368),
    [anon_sym_offsetof] = ACTIONS(2368),
    [anon_sym__Generic] = ACTIONS(2368),
    [anon_sym_asm] = ACTIONS(2368),
    [anon_sym___asm__] = ACTIONS(2368),
    [anon_sym___asm] = ACTIONS(2368),
    [sym_number_literal] = ACTIONS(2370),
    [anon_sym_L_SQUOTE] = ACTIONS(2370),
    [anon_sym_u_SQUOTE] = ACTIONS(2370),
    [anon_sym_U_SQUOTE] = ACTIONS(2370),
    [anon_sym_u8_SQUOTE] = ACTIONS(2370),
    [anon_sym_SQUOTE] = ACTIONS(2370),
    [anon_sym_AT] = ACTIONS(2368),
    [anon_sym_DQUOTE] = ACTIONS(2370),
    [anon_sym_L_DQUOTE] = ACTIONS(2370),
    [anon_sym_u_DQUOTE] = ACTIONS(2370),
    [anon_sym_U_DQUOTE] = ACTIONS(2370),
    [anon_sym_u8_DQUOTE] = ACTIONS(2370),
    [sym_true] = ACTIONS(2368),
    [sym_false] = ACTIONS(2368),
    [anon_sym_NULL] = ACTIONS(2368),
    [anon_sym_nullptr] = ACTIONS(2368),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2368),
    [anon_sym___typeof] = ACTIONS(2368),
    [anon_sym_typeof] = ACTIONS(2368),
    [anon_sym_ATimport] = ACTIONS(2370),
    [aux_sym_preproc_undef_token1] = ACTIONS(2368),
    [anon_sym_POUND] = ACTIONS(2368),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2368),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2368),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2368),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_API_AVAILABLE] = ACTIONS(2368),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_API_DEPRECATED] = ACTIONS(2368),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym___deprecated_msg] = ACTIONS(2368),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2368),
    [anon_sym_ATprotocol] = ACTIONS(2370),
    [anon_sym_ATinterface] = ACTIONS(2370),
    [anon_sym_ATimplementation] = ACTIONS(2370),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2370),
    [anon_sym_ATsynthesize] = ACTIONS(2370),
    [anon_sym_ATdynamic] = ACTIONS(2370),
    [anon_sym_ATtry] = ACTIONS(2370),
    [anon_sym___try] = ACTIONS(2368),
    [anon_sym_ATcatch] = ACTIONS(2370),
    [anon_sym___catch] = ACTIONS(2368),
    [anon_sym_ATfinally] = ACTIONS(2370),
    [anon_sym___finally] = ACTIONS(2368),
    [anon_sym_ATthrow] = ACTIONS(2370),
    [anon_sym_ATselector] = ACTIONS(2370),
    [anon_sym_ATavailable] = ACTIONS(2370),
    [anon_sym___builtin_available] = ACTIONS(2368),
    [anon_sym_va_arg] = ACTIONS(2368),
    [anon_sym_ATencode] = ACTIONS(2370),
    [anon_sym_ATsynchronized] = ACTIONS(2370),
    [anon_sym_BOOL] = ACTIONS(2368),
    [anon_sym_IMP] = ACTIONS(2368),
    [anon_sym_SEL] = ACTIONS(2368),
    [anon_sym_Class] = ACTIONS(2368),
    [anon_sym_id] = ACTIONS(2368),
  },
  [STATE(214)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token2] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [aux_sym_preproc_else_token1] = ACTIONS(2416),
    [aux_sym_preproc_elif_token1] = ACTIONS(2416),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(215)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token2] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [aux_sym_preproc_else_token1] = ACTIONS(2424),
    [aux_sym_preproc_elif_token1] = ACTIONS(2424),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(216)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token2] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [aux_sym_preproc_else_token1] = ACTIONS(2420),
    [aux_sym_preproc_elif_token1] = ACTIONS(2420),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(217)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(218)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [aux_sym_preproc_else_token1] = ACTIONS(2428),
    [aux_sym_preproc_elif_token1] = ACTIONS(2428),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(219)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [aux_sym_preproc_else_token1] = ACTIONS(2428),
    [aux_sym_preproc_elif_token1] = ACTIONS(2428),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(220)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [aux_sym_preproc_else_token1] = ACTIONS(2428),
    [aux_sym_preproc_elif_token1] = ACTIONS(2428),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(221)] = {
    [sym_catch_clause] = STATE(221),
    [aux_sym_try_statement_repeat1] = STATE(221),
    [ts_builtin_sym_end] = ACTIONS(2320),
    [sym_identifier] = ACTIONS(2318),
    [aux_sym_preproc_include_token1] = ACTIONS(2318),
    [aux_sym_preproc_include_token2] = ACTIONS(2318),
    [aux_sym_preproc_def_token1] = ACTIONS(2318),
    [aux_sym_preproc_if_token1] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2318),
    [sym_preproc_directive] = ACTIONS(2318),
    [anon_sym_LPAREN2] = ACTIONS(2320),
    [anon_sym_BANG] = ACTIONS(2320),
    [anon_sym_TILDE] = ACTIONS(2320),
    [anon_sym_DASH] = ACTIONS(2318),
    [anon_sym_PLUS] = ACTIONS(2318),
    [anon_sym_STAR] = ACTIONS(2320),
    [anon_sym_CARET] = ACTIONS(2320),
    [anon_sym_AMP] = ACTIONS(2320),
    [anon_sym_SEMI] = ACTIONS(2320),
    [anon_sym___extension__] = ACTIONS(2318),
    [anon_sym_typedef] = ACTIONS(2318),
    [anon_sym_extern] = ACTIONS(2318),
    [anon_sym___attribute__] = ACTIONS(2318),
    [anon_sym___attribute] = ACTIONS(2318),
    [anon_sym_noreturn] = ACTIONS(2318),
    [anon_sym_LBRACK] = ACTIONS(2320),
    [anon_sym___declspec] = ACTIONS(2318),
    [anon_sym___cdecl] = ACTIONS(2318),
    [anon_sym___clrcall] = ACTIONS(2318),
    [anon_sym___stdcall] = ACTIONS(2318),
    [anon_sym___fastcall] = ACTIONS(2318),
    [anon_sym___thiscall] = ACTIONS(2318),
    [anon_sym___vectorcall] = ACTIONS(2318),
    [anon_sym_LBRACE] = ACTIONS(2320),
    [anon_sym_signed] = ACTIONS(2318),
    [anon_sym_unsigned] = ACTIONS(2318),
    [anon_sym_long] = ACTIONS(2318),
    [anon_sym_short] = ACTIONS(2318),
    [anon_sym_static] = ACTIONS(2318),
    [anon_sym_ATautoreleasepool] = ACTIONS(2320),
    [anon_sym_auto] = ACTIONS(2318),
    [anon_sym_register] = ACTIONS(2318),
    [anon_sym_inline] = ACTIONS(2318),
    [anon_sym___inline] = ACTIONS(2318),
    [anon_sym___inline__] = ACTIONS(2318),
    [anon_sym___forceinline] = ACTIONS(2318),
    [anon_sym_thread_local] = ACTIONS(2318),
    [anon_sym___thread] = ACTIONS(2318),
    [anon_sym_CG_EXTERN] = ACTIONS(2318),
    [anon_sym_CG_INLINE] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2318),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2318),
    [anon_sym_IBOutlet] = ACTIONS(2318),
    [anon_sym_IBInspectable] = ACTIONS(2318),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2318),
    [anon_sym_NS_INLINE] = ACTIONS(2318),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2318),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2318),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2318),
    [anon_sym_const] = ACTIONS(2318),
    [anon_sym_constexpr] = ACTIONS(2318),
    [anon_sym_volatile] = ACTIONS(2318),
    [anon_sym_restrict] = ACTIONS(2318),
    [anon_sym___restrict__] = ACTIONS(2318),
    [anon_sym__Atomic] = ACTIONS(2318),
    [anon_sym__Noreturn] = ACTIONS(2318),
    [anon_sym__Nonnull] = ACTIONS(2318),
    [anon_sym_nullable] = ACTIONS(2318),
    [anon_sym__Complex] = ACTIONS(2318),
    [anon_sym__Nullable] = ACTIONS(2318),
    [anon_sym__Nullable_result] = ACTIONS(2318),
    [anon_sym__Null_unspecified] = ACTIONS(2318),
    [anon_sym___autoreleasing] = ACTIONS(2318),
    [anon_sym___block] = ACTIONS(2318),
    [anon_sym___bridge] = ACTIONS(2318),
    [anon_sym___bridge_retained] = ACTIONS(2318),
    [anon_sym___bridge_transfer] = ACTIONS(2318),
    [anon_sym___complex] = ACTIONS(2318),
    [anon_sym___const] = ACTIONS(2318),
    [anon_sym___imag] = ACTIONS(2318),
    [anon_sym___kindof] = ACTIONS(2318),
    [anon_sym___nonnull] = ACTIONS(2318),
    [anon_sym___nullable] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2318),
    [anon_sym___real] = ACTIONS(2318),
    [anon_sym___strong] = ACTIONS(2318),
    [anon_sym___unsafe_unretained] = ACTIONS(2318),
    [anon_sym___unused] = ACTIONS(2318),
    [anon_sym___weak] = ACTIONS(2318),
    [anon_sym_alignas] = ACTIONS(2318),
    [anon_sym__Alignas] = ACTIONS(2318),
    [sym_primitive_type] = ACTIONS(2318),
    [anon_sym_enum] = ACTIONS(2318),
    [anon_sym_struct] = ACTIONS(2318),
    [anon_sym_union] = ACTIONS(2318),
    [anon_sym_if] = ACTIONS(2318),
    [anon_sym_else] = ACTIONS(2318),
    [anon_sym_switch] = ACTIONS(2318),
    [anon_sym_case] = ACTIONS(2318),
    [anon_sym_default] = ACTIONS(2318),
    [anon_sym_while] = ACTIONS(2318),
    [anon_sym_do] = ACTIONS(2318),
    [anon_sym_for] = ACTIONS(2318),
    [anon_sym_in] = ACTIONS(2318),
    [anon_sym_return] = ACTIONS(2318),
    [anon_sym_break] = ACTIONS(2318),
    [anon_sym_continue] = ACTIONS(2318),
    [anon_sym_goto] = ACTIONS(2318),
    [anon_sym_DASH_DASH] = ACTIONS(2320),
    [anon_sym_PLUS_PLUS] = ACTIONS(2320),
    [anon_sym_sizeof] = ACTIONS(2318),
    [anon_sym___alignof__] = ACTIONS(2318),
    [anon_sym___alignof] = ACTIONS(2318),
    [anon_sym__alignof] = ACTIONS(2318),
    [anon_sym_alignof] = ACTIONS(2318),
    [anon_sym__Alignof] = ACTIONS(2318),
    [anon_sym_offsetof] = ACTIONS(2318),
    [anon_sym__Generic] = ACTIONS(2318),
    [anon_sym_asm] = ACTIONS(2318),
    [anon_sym___asm__] = ACTIONS(2318),
    [anon_sym___asm] = ACTIONS(2318),
    [sym_number_literal] = ACTIONS(2320),
    [anon_sym_L_SQUOTE] = ACTIONS(2320),
    [anon_sym_u_SQUOTE] = ACTIONS(2320),
    [anon_sym_U_SQUOTE] = ACTIONS(2320),
    [anon_sym_u8_SQUOTE] = ACTIONS(2320),
    [anon_sym_SQUOTE] = ACTIONS(2320),
    [anon_sym_AT] = ACTIONS(2318),
    [anon_sym_DQUOTE] = ACTIONS(2320),
    [anon_sym_L_DQUOTE] = ACTIONS(2320),
    [anon_sym_u_DQUOTE] = ACTIONS(2320),
    [anon_sym_U_DQUOTE] = ACTIONS(2320),
    [anon_sym_u8_DQUOTE] = ACTIONS(2320),
    [sym_true] = ACTIONS(2318),
    [sym_false] = ACTIONS(2318),
    [anon_sym_NULL] = ACTIONS(2318),
    [anon_sym_nullptr] = ACTIONS(2318),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2318),
    [anon_sym___typeof] = ACTIONS(2318),
    [anon_sym_typeof] = ACTIONS(2318),
    [anon_sym_ATimport] = ACTIONS(2320),
    [aux_sym_preproc_undef_token1] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2318),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2318),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2318),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2318),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_API_AVAILABLE] = ACTIONS(2318),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_API_DEPRECATED] = ACTIONS(2318),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym___deprecated_msg] = ACTIONS(2318),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2318),
    [anon_sym_ATprotocol] = ACTIONS(2320),
    [anon_sym_ATinterface] = ACTIONS(2320),
    [anon_sym_ATimplementation] = ACTIONS(2320),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2320),
    [anon_sym_ATtry] = ACTIONS(2320),
    [anon_sym___try] = ACTIONS(2318),
    [anon_sym_ATcatch] = ACTIONS(2432),
    [anon_sym___catch] = ACTIONS(2435),
    [anon_sym_ATfinally] = ACTIONS(2320),
    [anon_sym___finally] = ACTIONS(2318),
    [anon_sym_ATthrow] = ACTIONS(2320),
    [anon_sym_ATselector] = ACTIONS(2320),
    [anon_sym_ATavailable] = ACTIONS(2320),
    [anon_sym___builtin_available] = ACTIONS(2318),
    [anon_sym_va_arg] = ACTIONS(2318),
    [anon_sym_ATencode] = ACTIONS(2320),
    [anon_sym_ATsynchronized] = ACTIONS(2320),
    [anon_sym_BOOL] = ACTIONS(2318),
    [anon_sym_IMP] = ACTIONS(2318),
    [anon_sym_SEL] = ACTIONS(2318),
    [anon_sym_Class] = ACTIONS(2318),
    [anon_sym_id] = ACTIONS(2318),
  },
  [STATE(222)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token2] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [aux_sym_preproc_else_token1] = ACTIONS(2438),
    [aux_sym_preproc_elif_token1] = ACTIONS(2438),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(223)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token2] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [aux_sym_preproc_else_token1] = ACTIONS(2442),
    [aux_sym_preproc_elif_token1] = ACTIONS(2442),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(224)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token2] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATsynthesize] = ACTIONS(2348),
    [anon_sym_ATdynamic] = ACTIONS(2348),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATcatch] = ACTIONS(2348),
    [anon_sym___catch] = ACTIONS(2346),
    [anon_sym_ATfinally] = ACTIONS(2348),
    [anon_sym___finally] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(225)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token2] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATsynthesize] = ACTIONS(2352),
    [anon_sym_ATdynamic] = ACTIONS(2352),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATcatch] = ACTIONS(2352),
    [anon_sym___catch] = ACTIONS(2350),
    [anon_sym_ATfinally] = ACTIONS(2352),
    [anon_sym___finally] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(226)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token2] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATsynthesize] = ACTIONS(2362),
    [anon_sym_ATdynamic] = ACTIONS(2362),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATcatch] = ACTIONS(2362),
    [anon_sym___catch] = ACTIONS(2360),
    [anon_sym_ATfinally] = ACTIONS(2362),
    [anon_sym___finally] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(227)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(228)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token2] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [aux_sym_preproc_else_token1] = ACTIONS(2446),
    [aux_sym_preproc_elif_token1] = ACTIONS(2446),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(229)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(230)] = {
    [sym_identifier] = ACTIONS(2450),
    [aux_sym_preproc_include_token1] = ACTIONS(2450),
    [aux_sym_preproc_include_token2] = ACTIONS(2450),
    [aux_sym_preproc_def_token1] = ACTIONS(2450),
    [aux_sym_preproc_if_token1] = ACTIONS(2450),
    [aux_sym_preproc_if_token2] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2450),
    [aux_sym_preproc_else_token1] = ACTIONS(2450),
    [aux_sym_preproc_elif_token1] = ACTIONS(2450),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2450),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2450),
    [sym_preproc_directive] = ACTIONS(2450),
    [anon_sym_LPAREN2] = ACTIONS(2452),
    [anon_sym_BANG] = ACTIONS(2452),
    [anon_sym_TILDE] = ACTIONS(2452),
    [anon_sym_DASH] = ACTIONS(2450),
    [anon_sym_PLUS] = ACTIONS(2450),
    [anon_sym_STAR] = ACTIONS(2452),
    [anon_sym_CARET] = ACTIONS(2452),
    [anon_sym_AMP] = ACTIONS(2452),
    [anon_sym_SEMI] = ACTIONS(2452),
    [anon_sym___extension__] = ACTIONS(2450),
    [anon_sym_typedef] = ACTIONS(2450),
    [anon_sym_extern] = ACTIONS(2450),
    [anon_sym___attribute__] = ACTIONS(2450),
    [anon_sym___attribute] = ACTIONS(2450),
    [anon_sym_noreturn] = ACTIONS(2450),
    [anon_sym_LBRACK] = ACTIONS(2452),
    [anon_sym___declspec] = ACTIONS(2450),
    [anon_sym___cdecl] = ACTIONS(2450),
    [anon_sym___clrcall] = ACTIONS(2450),
    [anon_sym___stdcall] = ACTIONS(2450),
    [anon_sym___fastcall] = ACTIONS(2450),
    [anon_sym___thiscall] = ACTIONS(2450),
    [anon_sym___vectorcall] = ACTIONS(2450),
    [anon_sym_LBRACE] = ACTIONS(2452),
    [anon_sym_signed] = ACTIONS(2450),
    [anon_sym_unsigned] = ACTIONS(2450),
    [anon_sym_long] = ACTIONS(2450),
    [anon_sym_short] = ACTIONS(2450),
    [anon_sym_static] = ACTIONS(2450),
    [anon_sym_ATautoreleasepool] = ACTIONS(2452),
    [anon_sym_auto] = ACTIONS(2450),
    [anon_sym_register] = ACTIONS(2450),
    [anon_sym_inline] = ACTIONS(2450),
    [anon_sym___inline] = ACTIONS(2450),
    [anon_sym___inline__] = ACTIONS(2450),
    [anon_sym___forceinline] = ACTIONS(2450),
    [anon_sym_thread_local] = ACTIONS(2450),
    [anon_sym___thread] = ACTIONS(2450),
    [anon_sym_CG_EXTERN] = ACTIONS(2450),
    [anon_sym_CG_INLINE] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2450),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2450),
    [anon_sym_IBOutlet] = ACTIONS(2450),
    [anon_sym_IBInspectable] = ACTIONS(2450),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2450),
    [anon_sym_NS_INLINE] = ACTIONS(2450),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2450),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2450),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2450),
    [anon_sym_const] = ACTIONS(2450),
    [anon_sym_constexpr] = ACTIONS(2450),
    [anon_sym_volatile] = ACTIONS(2450),
    [anon_sym_restrict] = ACTIONS(2450),
    [anon_sym___restrict__] = ACTIONS(2450),
    [anon_sym__Atomic] = ACTIONS(2450),
    [anon_sym__Noreturn] = ACTIONS(2450),
    [anon_sym__Nonnull] = ACTIONS(2450),
    [anon_sym_nullable] = ACTIONS(2450),
    [anon_sym__Complex] = ACTIONS(2450),
    [anon_sym__Nullable] = ACTIONS(2450),
    [anon_sym__Nullable_result] = ACTIONS(2450),
    [anon_sym__Null_unspecified] = ACTIONS(2450),
    [anon_sym___autoreleasing] = ACTIONS(2450),
    [anon_sym___block] = ACTIONS(2450),
    [anon_sym___bridge] = ACTIONS(2450),
    [anon_sym___bridge_retained] = ACTIONS(2450),
    [anon_sym___bridge_transfer] = ACTIONS(2450),
    [anon_sym___complex] = ACTIONS(2450),
    [anon_sym___const] = ACTIONS(2450),
    [anon_sym___imag] = ACTIONS(2450),
    [anon_sym___kindof] = ACTIONS(2450),
    [anon_sym___nonnull] = ACTIONS(2450),
    [anon_sym___nullable] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2450),
    [anon_sym___real] = ACTIONS(2450),
    [anon_sym___strong] = ACTIONS(2450),
    [anon_sym___unsafe_unretained] = ACTIONS(2450),
    [anon_sym___unused] = ACTIONS(2450),
    [anon_sym___weak] = ACTIONS(2450),
    [anon_sym_alignas] = ACTIONS(2450),
    [anon_sym__Alignas] = ACTIONS(2450),
    [sym_primitive_type] = ACTIONS(2450),
    [anon_sym_enum] = ACTIONS(2450),
    [anon_sym_struct] = ACTIONS(2450),
    [anon_sym_union] = ACTIONS(2450),
    [anon_sym_if] = ACTIONS(2450),
    [anon_sym_else] = ACTIONS(2450),
    [anon_sym_switch] = ACTIONS(2450),
    [anon_sym_case] = ACTIONS(2450),
    [anon_sym_default] = ACTIONS(2450),
    [anon_sym_while] = ACTIONS(2450),
    [anon_sym_do] = ACTIONS(2450),
    [anon_sym_for] = ACTIONS(2450),
    [anon_sym_in] = ACTIONS(2450),
    [anon_sym_return] = ACTIONS(2450),
    [anon_sym_break] = ACTIONS(2450),
    [anon_sym_continue] = ACTIONS(2450),
    [anon_sym_goto] = ACTIONS(2450),
    [anon_sym_DASH_DASH] = ACTIONS(2452),
    [anon_sym_PLUS_PLUS] = ACTIONS(2452),
    [anon_sym_sizeof] = ACTIONS(2450),
    [anon_sym___alignof__] = ACTIONS(2450),
    [anon_sym___alignof] = ACTIONS(2450),
    [anon_sym__alignof] = ACTIONS(2450),
    [anon_sym_alignof] = ACTIONS(2450),
    [anon_sym__Alignof] = ACTIONS(2450),
    [anon_sym_offsetof] = ACTIONS(2450),
    [anon_sym__Generic] = ACTIONS(2450),
    [anon_sym_asm] = ACTIONS(2450),
    [anon_sym___asm__] = ACTIONS(2450),
    [anon_sym___asm] = ACTIONS(2450),
    [sym_number_literal] = ACTIONS(2452),
    [anon_sym_L_SQUOTE] = ACTIONS(2452),
    [anon_sym_u_SQUOTE] = ACTIONS(2452),
    [anon_sym_U_SQUOTE] = ACTIONS(2452),
    [anon_sym_u8_SQUOTE] = ACTIONS(2452),
    [anon_sym_SQUOTE] = ACTIONS(2452),
    [anon_sym_AT] = ACTIONS(2450),
    [anon_sym_DQUOTE] = ACTIONS(2452),
    [anon_sym_L_DQUOTE] = ACTIONS(2452),
    [anon_sym_u_DQUOTE] = ACTIONS(2452),
    [anon_sym_U_DQUOTE] = ACTIONS(2452),
    [anon_sym_u8_DQUOTE] = ACTIONS(2452),
    [sym_true] = ACTIONS(2450),
    [sym_false] = ACTIONS(2450),
    [anon_sym_NULL] = ACTIONS(2450),
    [anon_sym_nullptr] = ACTIONS(2450),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2450),
    [anon_sym___typeof] = ACTIONS(2450),
    [anon_sym_typeof] = ACTIONS(2450),
    [anon_sym_ATimport] = ACTIONS(2452),
    [aux_sym_preproc_undef_token1] = ACTIONS(2450),
    [anon_sym_POUND] = ACTIONS(2450),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2450),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2450),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2450),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_API_AVAILABLE] = ACTIONS(2450),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_API_DEPRECATED] = ACTIONS(2450),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym___deprecated_msg] = ACTIONS(2450),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2450),
    [anon_sym_ATprotocol] = ACTIONS(2452),
    [anon_sym_ATinterface] = ACTIONS(2452),
    [anon_sym_ATimplementation] = ACTIONS(2452),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2452),
    [anon_sym_ATsynthesize] = ACTIONS(2452),
    [anon_sym_ATdynamic] = ACTIONS(2452),
    [anon_sym_ATtry] = ACTIONS(2452),
    [anon_sym___try] = ACTIONS(2450),
    [anon_sym_ATthrow] = ACTIONS(2452),
    [anon_sym_ATselector] = ACTIONS(2452),
    [anon_sym_ATavailable] = ACTIONS(2452),
    [anon_sym___builtin_available] = ACTIONS(2450),
    [anon_sym_va_arg] = ACTIONS(2450),
    [anon_sym_ATencode] = ACTIONS(2452),
    [anon_sym_ATsynchronized] = ACTIONS(2452),
    [anon_sym_BOOL] = ACTIONS(2450),
    [anon_sym_IMP] = ACTIONS(2450),
    [anon_sym_SEL] = ACTIONS(2450),
    [anon_sym_Class] = ACTIONS(2450),
    [anon_sym_id] = ACTIONS(2450),
  },
  [STATE(231)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(232)] = {
    [sym_identifier] = ACTIONS(2454),
    [aux_sym_preproc_include_token1] = ACTIONS(2454),
    [aux_sym_preproc_include_token2] = ACTIONS(2454),
    [aux_sym_preproc_def_token1] = ACTIONS(2454),
    [aux_sym_preproc_if_token1] = ACTIONS(2454),
    [aux_sym_preproc_if_token2] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2454),
    [aux_sym_preproc_else_token1] = ACTIONS(2454),
    [aux_sym_preproc_elif_token1] = ACTIONS(2454),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2454),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2454),
    [sym_preproc_directive] = ACTIONS(2454),
    [anon_sym_LPAREN2] = ACTIONS(2456),
    [anon_sym_BANG] = ACTIONS(2456),
    [anon_sym_TILDE] = ACTIONS(2456),
    [anon_sym_DASH] = ACTIONS(2454),
    [anon_sym_PLUS] = ACTIONS(2454),
    [anon_sym_STAR] = ACTIONS(2456),
    [anon_sym_CARET] = ACTIONS(2456),
    [anon_sym_AMP] = ACTIONS(2456),
    [anon_sym_SEMI] = ACTIONS(2456),
    [anon_sym___extension__] = ACTIONS(2454),
    [anon_sym_typedef] = ACTIONS(2454),
    [anon_sym_extern] = ACTIONS(2454),
    [anon_sym___attribute__] = ACTIONS(2454),
    [anon_sym___attribute] = ACTIONS(2454),
    [anon_sym_noreturn] = ACTIONS(2454),
    [anon_sym_LBRACK] = ACTIONS(2456),
    [anon_sym___declspec] = ACTIONS(2454),
    [anon_sym___cdecl] = ACTIONS(2454),
    [anon_sym___clrcall] = ACTIONS(2454),
    [anon_sym___stdcall] = ACTIONS(2454),
    [anon_sym___fastcall] = ACTIONS(2454),
    [anon_sym___thiscall] = ACTIONS(2454),
    [anon_sym___vectorcall] = ACTIONS(2454),
    [anon_sym_LBRACE] = ACTIONS(2456),
    [anon_sym_signed] = ACTIONS(2454),
    [anon_sym_unsigned] = ACTIONS(2454),
    [anon_sym_long] = ACTIONS(2454),
    [anon_sym_short] = ACTIONS(2454),
    [anon_sym_static] = ACTIONS(2454),
    [anon_sym_ATautoreleasepool] = ACTIONS(2456),
    [anon_sym_auto] = ACTIONS(2454),
    [anon_sym_register] = ACTIONS(2454),
    [anon_sym_inline] = ACTIONS(2454),
    [anon_sym___inline] = ACTIONS(2454),
    [anon_sym___inline__] = ACTIONS(2454),
    [anon_sym___forceinline] = ACTIONS(2454),
    [anon_sym_thread_local] = ACTIONS(2454),
    [anon_sym___thread] = ACTIONS(2454),
    [anon_sym_CG_EXTERN] = ACTIONS(2454),
    [anon_sym_CG_INLINE] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2454),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2454),
    [anon_sym_IBOutlet] = ACTIONS(2454),
    [anon_sym_IBInspectable] = ACTIONS(2454),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2454),
    [anon_sym_NS_INLINE] = ACTIONS(2454),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2454),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2454),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2454),
    [anon_sym_const] = ACTIONS(2454),
    [anon_sym_constexpr] = ACTIONS(2454),
    [anon_sym_volatile] = ACTIONS(2454),
    [anon_sym_restrict] = ACTIONS(2454),
    [anon_sym___restrict__] = ACTIONS(2454),
    [anon_sym__Atomic] = ACTIONS(2454),
    [anon_sym__Noreturn] = ACTIONS(2454),
    [anon_sym__Nonnull] = ACTIONS(2454),
    [anon_sym_nullable] = ACTIONS(2454),
    [anon_sym__Complex] = ACTIONS(2454),
    [anon_sym__Nullable] = ACTIONS(2454),
    [anon_sym__Nullable_result] = ACTIONS(2454),
    [anon_sym__Null_unspecified] = ACTIONS(2454),
    [anon_sym___autoreleasing] = ACTIONS(2454),
    [anon_sym___block] = ACTIONS(2454),
    [anon_sym___bridge] = ACTIONS(2454),
    [anon_sym___bridge_retained] = ACTIONS(2454),
    [anon_sym___bridge_transfer] = ACTIONS(2454),
    [anon_sym___complex] = ACTIONS(2454),
    [anon_sym___const] = ACTIONS(2454),
    [anon_sym___imag] = ACTIONS(2454),
    [anon_sym___kindof] = ACTIONS(2454),
    [anon_sym___nonnull] = ACTIONS(2454),
    [anon_sym___nullable] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2454),
    [anon_sym___real] = ACTIONS(2454),
    [anon_sym___strong] = ACTIONS(2454),
    [anon_sym___unsafe_unretained] = ACTIONS(2454),
    [anon_sym___unused] = ACTIONS(2454),
    [anon_sym___weak] = ACTIONS(2454),
    [anon_sym_alignas] = ACTIONS(2454),
    [anon_sym__Alignas] = ACTIONS(2454),
    [sym_primitive_type] = ACTIONS(2454),
    [anon_sym_enum] = ACTIONS(2454),
    [anon_sym_struct] = ACTIONS(2454),
    [anon_sym_union] = ACTIONS(2454),
    [anon_sym_if] = ACTIONS(2454),
    [anon_sym_else] = ACTIONS(2454),
    [anon_sym_switch] = ACTIONS(2454),
    [anon_sym_case] = ACTIONS(2454),
    [anon_sym_default] = ACTIONS(2454),
    [anon_sym_while] = ACTIONS(2454),
    [anon_sym_do] = ACTIONS(2454),
    [anon_sym_for] = ACTIONS(2454),
    [anon_sym_in] = ACTIONS(2454),
    [anon_sym_return] = ACTIONS(2454),
    [anon_sym_break] = ACTIONS(2454),
    [anon_sym_continue] = ACTIONS(2454),
    [anon_sym_goto] = ACTIONS(2454),
    [anon_sym_DASH_DASH] = ACTIONS(2456),
    [anon_sym_PLUS_PLUS] = ACTIONS(2456),
    [anon_sym_sizeof] = ACTIONS(2454),
    [anon_sym___alignof__] = ACTIONS(2454),
    [anon_sym___alignof] = ACTIONS(2454),
    [anon_sym__alignof] = ACTIONS(2454),
    [anon_sym_alignof] = ACTIONS(2454),
    [anon_sym__Alignof] = ACTIONS(2454),
    [anon_sym_offsetof] = ACTIONS(2454),
    [anon_sym__Generic] = ACTIONS(2454),
    [anon_sym_asm] = ACTIONS(2454),
    [anon_sym___asm__] = ACTIONS(2454),
    [anon_sym___asm] = ACTIONS(2454),
    [sym_number_literal] = ACTIONS(2456),
    [anon_sym_L_SQUOTE] = ACTIONS(2456),
    [anon_sym_u_SQUOTE] = ACTIONS(2456),
    [anon_sym_U_SQUOTE] = ACTIONS(2456),
    [anon_sym_u8_SQUOTE] = ACTIONS(2456),
    [anon_sym_SQUOTE] = ACTIONS(2456),
    [anon_sym_AT] = ACTIONS(2454),
    [anon_sym_DQUOTE] = ACTIONS(2456),
    [anon_sym_L_DQUOTE] = ACTIONS(2456),
    [anon_sym_u_DQUOTE] = ACTIONS(2456),
    [anon_sym_U_DQUOTE] = ACTIONS(2456),
    [anon_sym_u8_DQUOTE] = ACTIONS(2456),
    [sym_true] = ACTIONS(2454),
    [sym_false] = ACTIONS(2454),
    [anon_sym_NULL] = ACTIONS(2454),
    [anon_sym_nullptr] = ACTIONS(2454),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2454),
    [anon_sym___typeof] = ACTIONS(2454),
    [anon_sym_typeof] = ACTIONS(2454),
    [anon_sym_ATimport] = ACTIONS(2456),
    [aux_sym_preproc_undef_token1] = ACTIONS(2454),
    [anon_sym_POUND] = ACTIONS(2454),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2454),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2454),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2454),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_API_AVAILABLE] = ACTIONS(2454),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_API_DEPRECATED] = ACTIONS(2454),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym___deprecated_msg] = ACTIONS(2454),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2454),
    [anon_sym_ATprotocol] = ACTIONS(2456),
    [anon_sym_ATinterface] = ACTIONS(2456),
    [anon_sym_ATimplementation] = ACTIONS(2456),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2456),
    [anon_sym_ATsynthesize] = ACTIONS(2456),
    [anon_sym_ATdynamic] = ACTIONS(2456),
    [anon_sym_ATtry] = ACTIONS(2456),
    [anon_sym___try] = ACTIONS(2454),
    [anon_sym_ATthrow] = ACTIONS(2456),
    [anon_sym_ATselector] = ACTIONS(2456),
    [anon_sym_ATavailable] = ACTIONS(2456),
    [anon_sym___builtin_available] = ACTIONS(2454),
    [anon_sym_va_arg] = ACTIONS(2454),
    [anon_sym_ATencode] = ACTIONS(2456),
    [anon_sym_ATsynchronized] = ACTIONS(2456),
    [anon_sym_BOOL] = ACTIONS(2454),
    [anon_sym_IMP] = ACTIONS(2454),
    [anon_sym_SEL] = ACTIONS(2454),
    [anon_sym_Class] = ACTIONS(2454),
    [anon_sym_id] = ACTIONS(2454),
  },
  [STATE(233)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(234)] = {
    [sym_identifier] = ACTIONS(2458),
    [aux_sym_preproc_include_token1] = ACTIONS(2458),
    [aux_sym_preproc_include_token2] = ACTIONS(2458),
    [aux_sym_preproc_def_token1] = ACTIONS(2458),
    [aux_sym_preproc_if_token1] = ACTIONS(2458),
    [aux_sym_preproc_if_token2] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2458),
    [aux_sym_preproc_else_token1] = ACTIONS(2458),
    [aux_sym_preproc_elif_token1] = ACTIONS(2458),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2458),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2458),
    [sym_preproc_directive] = ACTIONS(2458),
    [anon_sym_LPAREN2] = ACTIONS(2460),
    [anon_sym_BANG] = ACTIONS(2460),
    [anon_sym_TILDE] = ACTIONS(2460),
    [anon_sym_DASH] = ACTIONS(2458),
    [anon_sym_PLUS] = ACTIONS(2458),
    [anon_sym_STAR] = ACTIONS(2460),
    [anon_sym_CARET] = ACTIONS(2460),
    [anon_sym_AMP] = ACTIONS(2460),
    [anon_sym_SEMI] = ACTIONS(2460),
    [anon_sym___extension__] = ACTIONS(2458),
    [anon_sym_typedef] = ACTIONS(2458),
    [anon_sym_extern] = ACTIONS(2458),
    [anon_sym___attribute__] = ACTIONS(2458),
    [anon_sym___attribute] = ACTIONS(2458),
    [anon_sym_noreturn] = ACTIONS(2458),
    [anon_sym_LBRACK] = ACTIONS(2460),
    [anon_sym___declspec] = ACTIONS(2458),
    [anon_sym___cdecl] = ACTIONS(2458),
    [anon_sym___clrcall] = ACTIONS(2458),
    [anon_sym___stdcall] = ACTIONS(2458),
    [anon_sym___fastcall] = ACTIONS(2458),
    [anon_sym___thiscall] = ACTIONS(2458),
    [anon_sym___vectorcall] = ACTIONS(2458),
    [anon_sym_LBRACE] = ACTIONS(2460),
    [anon_sym_signed] = ACTIONS(2458),
    [anon_sym_unsigned] = ACTIONS(2458),
    [anon_sym_long] = ACTIONS(2458),
    [anon_sym_short] = ACTIONS(2458),
    [anon_sym_static] = ACTIONS(2458),
    [anon_sym_ATautoreleasepool] = ACTIONS(2460),
    [anon_sym_auto] = ACTIONS(2458),
    [anon_sym_register] = ACTIONS(2458),
    [anon_sym_inline] = ACTIONS(2458),
    [anon_sym___inline] = ACTIONS(2458),
    [anon_sym___inline__] = ACTIONS(2458),
    [anon_sym___forceinline] = ACTIONS(2458),
    [anon_sym_thread_local] = ACTIONS(2458),
    [anon_sym___thread] = ACTIONS(2458),
    [anon_sym_CG_EXTERN] = ACTIONS(2458),
    [anon_sym_CG_INLINE] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2458),
    [anon_sym_IBOutlet] = ACTIONS(2458),
    [anon_sym_IBInspectable] = ACTIONS(2458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2458),
    [anon_sym_NS_INLINE] = ACTIONS(2458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2458),
    [anon_sym_const] = ACTIONS(2458),
    [anon_sym_constexpr] = ACTIONS(2458),
    [anon_sym_volatile] = ACTIONS(2458),
    [anon_sym_restrict] = ACTIONS(2458),
    [anon_sym___restrict__] = ACTIONS(2458),
    [anon_sym__Atomic] = ACTIONS(2458),
    [anon_sym__Noreturn] = ACTIONS(2458),
    [anon_sym__Nonnull] = ACTIONS(2458),
    [anon_sym_nullable] = ACTIONS(2458),
    [anon_sym__Complex] = ACTIONS(2458),
    [anon_sym__Nullable] = ACTIONS(2458),
    [anon_sym__Nullable_result] = ACTIONS(2458),
    [anon_sym__Null_unspecified] = ACTIONS(2458),
    [anon_sym___autoreleasing] = ACTIONS(2458),
    [anon_sym___block] = ACTIONS(2458),
    [anon_sym___bridge] = ACTIONS(2458),
    [anon_sym___bridge_retained] = ACTIONS(2458),
    [anon_sym___bridge_transfer] = ACTIONS(2458),
    [anon_sym___complex] = ACTIONS(2458),
    [anon_sym___const] = ACTIONS(2458),
    [anon_sym___imag] = ACTIONS(2458),
    [anon_sym___kindof] = ACTIONS(2458),
    [anon_sym___nonnull] = ACTIONS(2458),
    [anon_sym___nullable] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2458),
    [anon_sym___real] = ACTIONS(2458),
    [anon_sym___strong] = ACTIONS(2458),
    [anon_sym___unsafe_unretained] = ACTIONS(2458),
    [anon_sym___unused] = ACTIONS(2458),
    [anon_sym___weak] = ACTIONS(2458),
    [anon_sym_alignas] = ACTIONS(2458),
    [anon_sym__Alignas] = ACTIONS(2458),
    [sym_primitive_type] = ACTIONS(2458),
    [anon_sym_enum] = ACTIONS(2458),
    [anon_sym_struct] = ACTIONS(2458),
    [anon_sym_union] = ACTIONS(2458),
    [anon_sym_if] = ACTIONS(2458),
    [anon_sym_else] = ACTIONS(2458),
    [anon_sym_switch] = ACTIONS(2458),
    [anon_sym_case] = ACTIONS(2458),
    [anon_sym_default] = ACTIONS(2458),
    [anon_sym_while] = ACTIONS(2458),
    [anon_sym_do] = ACTIONS(2458),
    [anon_sym_for] = ACTIONS(2458),
    [anon_sym_in] = ACTIONS(2458),
    [anon_sym_return] = ACTIONS(2458),
    [anon_sym_break] = ACTIONS(2458),
    [anon_sym_continue] = ACTIONS(2458),
    [anon_sym_goto] = ACTIONS(2458),
    [anon_sym_DASH_DASH] = ACTIONS(2460),
    [anon_sym_PLUS_PLUS] = ACTIONS(2460),
    [anon_sym_sizeof] = ACTIONS(2458),
    [anon_sym___alignof__] = ACTIONS(2458),
    [anon_sym___alignof] = ACTIONS(2458),
    [anon_sym__alignof] = ACTIONS(2458),
    [anon_sym_alignof] = ACTIONS(2458),
    [anon_sym__Alignof] = ACTIONS(2458),
    [anon_sym_offsetof] = ACTIONS(2458),
    [anon_sym__Generic] = ACTIONS(2458),
    [anon_sym_asm] = ACTIONS(2458),
    [anon_sym___asm__] = ACTIONS(2458),
    [anon_sym___asm] = ACTIONS(2458),
    [sym_number_literal] = ACTIONS(2460),
    [anon_sym_L_SQUOTE] = ACTIONS(2460),
    [anon_sym_u_SQUOTE] = ACTIONS(2460),
    [anon_sym_U_SQUOTE] = ACTIONS(2460),
    [anon_sym_u8_SQUOTE] = ACTIONS(2460),
    [anon_sym_SQUOTE] = ACTIONS(2460),
    [anon_sym_AT] = ACTIONS(2458),
    [anon_sym_DQUOTE] = ACTIONS(2460),
    [anon_sym_L_DQUOTE] = ACTIONS(2460),
    [anon_sym_u_DQUOTE] = ACTIONS(2460),
    [anon_sym_U_DQUOTE] = ACTIONS(2460),
    [anon_sym_u8_DQUOTE] = ACTIONS(2460),
    [sym_true] = ACTIONS(2458),
    [sym_false] = ACTIONS(2458),
    [anon_sym_NULL] = ACTIONS(2458),
    [anon_sym_nullptr] = ACTIONS(2458),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2458),
    [anon_sym___typeof] = ACTIONS(2458),
    [anon_sym_typeof] = ACTIONS(2458),
    [anon_sym_ATimport] = ACTIONS(2460),
    [aux_sym_preproc_undef_token1] = ACTIONS(2458),
    [anon_sym_POUND] = ACTIONS(2458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_API_AVAILABLE] = ACTIONS(2458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_API_DEPRECATED] = ACTIONS(2458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym___deprecated_msg] = ACTIONS(2458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2458),
    [anon_sym_ATprotocol] = ACTIONS(2460),
    [anon_sym_ATinterface] = ACTIONS(2460),
    [anon_sym_ATimplementation] = ACTIONS(2460),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2460),
    [anon_sym_ATsynthesize] = ACTIONS(2460),
    [anon_sym_ATdynamic] = ACTIONS(2460),
    [anon_sym_ATtry] = ACTIONS(2460),
    [anon_sym___try] = ACTIONS(2458),
    [anon_sym_ATthrow] = ACTIONS(2460),
    [anon_sym_ATselector] = ACTIONS(2460),
    [anon_sym_ATavailable] = ACTIONS(2460),
    [anon_sym___builtin_available] = ACTIONS(2458),
    [anon_sym_va_arg] = ACTIONS(2458),
    [anon_sym_ATencode] = ACTIONS(2460),
    [anon_sym_ATsynchronized] = ACTIONS(2460),
    [anon_sym_BOOL] = ACTIONS(2458),
    [anon_sym_IMP] = ACTIONS(2458),
    [anon_sym_SEL] = ACTIONS(2458),
    [anon_sym_Class] = ACTIONS(2458),
    [anon_sym_id] = ACTIONS(2458),
  },
  [STATE(235)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(236)] = {
    [sym_identifier] = ACTIONS(2462),
    [aux_sym_preproc_include_token1] = ACTIONS(2462),
    [aux_sym_preproc_include_token2] = ACTIONS(2462),
    [aux_sym_preproc_def_token1] = ACTIONS(2462),
    [aux_sym_preproc_if_token1] = ACTIONS(2462),
    [aux_sym_preproc_if_token2] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2462),
    [aux_sym_preproc_else_token1] = ACTIONS(2462),
    [aux_sym_preproc_elif_token1] = ACTIONS(2462),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2462),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2462),
    [sym_preproc_directive] = ACTIONS(2462),
    [anon_sym_LPAREN2] = ACTIONS(2464),
    [anon_sym_BANG] = ACTIONS(2464),
    [anon_sym_TILDE] = ACTIONS(2464),
    [anon_sym_DASH] = ACTIONS(2462),
    [anon_sym_PLUS] = ACTIONS(2462),
    [anon_sym_STAR] = ACTIONS(2464),
    [anon_sym_CARET] = ACTIONS(2464),
    [anon_sym_AMP] = ACTIONS(2464),
    [anon_sym_SEMI] = ACTIONS(2464),
    [anon_sym___extension__] = ACTIONS(2462),
    [anon_sym_typedef] = ACTIONS(2462),
    [anon_sym_extern] = ACTIONS(2462),
    [anon_sym___attribute__] = ACTIONS(2462),
    [anon_sym___attribute] = ACTIONS(2462),
    [anon_sym_noreturn] = ACTIONS(2462),
    [anon_sym_LBRACK] = ACTIONS(2464),
    [anon_sym___declspec] = ACTIONS(2462),
    [anon_sym___cdecl] = ACTIONS(2462),
    [anon_sym___clrcall] = ACTIONS(2462),
    [anon_sym___stdcall] = ACTIONS(2462),
    [anon_sym___fastcall] = ACTIONS(2462),
    [anon_sym___thiscall] = ACTIONS(2462),
    [anon_sym___vectorcall] = ACTIONS(2462),
    [anon_sym_LBRACE] = ACTIONS(2464),
    [anon_sym_signed] = ACTIONS(2462),
    [anon_sym_unsigned] = ACTIONS(2462),
    [anon_sym_long] = ACTIONS(2462),
    [anon_sym_short] = ACTIONS(2462),
    [anon_sym_static] = ACTIONS(2462),
    [anon_sym_ATautoreleasepool] = ACTIONS(2464),
    [anon_sym_auto] = ACTIONS(2462),
    [anon_sym_register] = ACTIONS(2462),
    [anon_sym_inline] = ACTIONS(2462),
    [anon_sym___inline] = ACTIONS(2462),
    [anon_sym___inline__] = ACTIONS(2462),
    [anon_sym___forceinline] = ACTIONS(2462),
    [anon_sym_thread_local] = ACTIONS(2462),
    [anon_sym___thread] = ACTIONS(2462),
    [anon_sym_CG_EXTERN] = ACTIONS(2462),
    [anon_sym_CG_INLINE] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2462),
    [anon_sym_IBOutlet] = ACTIONS(2462),
    [anon_sym_IBInspectable] = ACTIONS(2462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2462),
    [anon_sym_NS_INLINE] = ACTIONS(2462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2462),
    [anon_sym_const] = ACTIONS(2462),
    [anon_sym_constexpr] = ACTIONS(2462),
    [anon_sym_volatile] = ACTIONS(2462),
    [anon_sym_restrict] = ACTIONS(2462),
    [anon_sym___restrict__] = ACTIONS(2462),
    [anon_sym__Atomic] = ACTIONS(2462),
    [anon_sym__Noreturn] = ACTIONS(2462),
    [anon_sym__Nonnull] = ACTIONS(2462),
    [anon_sym_nullable] = ACTIONS(2462),
    [anon_sym__Complex] = ACTIONS(2462),
    [anon_sym__Nullable] = ACTIONS(2462),
    [anon_sym__Nullable_result] = ACTIONS(2462),
    [anon_sym__Null_unspecified] = ACTIONS(2462),
    [anon_sym___autoreleasing] = ACTIONS(2462),
    [anon_sym___block] = ACTIONS(2462),
    [anon_sym___bridge] = ACTIONS(2462),
    [anon_sym___bridge_retained] = ACTIONS(2462),
    [anon_sym___bridge_transfer] = ACTIONS(2462),
    [anon_sym___complex] = ACTIONS(2462),
    [anon_sym___const] = ACTIONS(2462),
    [anon_sym___imag] = ACTIONS(2462),
    [anon_sym___kindof] = ACTIONS(2462),
    [anon_sym___nonnull] = ACTIONS(2462),
    [anon_sym___nullable] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2462),
    [anon_sym___real] = ACTIONS(2462),
    [anon_sym___strong] = ACTIONS(2462),
    [anon_sym___unsafe_unretained] = ACTIONS(2462),
    [anon_sym___unused] = ACTIONS(2462),
    [anon_sym___weak] = ACTIONS(2462),
    [anon_sym_alignas] = ACTIONS(2462),
    [anon_sym__Alignas] = ACTIONS(2462),
    [sym_primitive_type] = ACTIONS(2462),
    [anon_sym_enum] = ACTIONS(2462),
    [anon_sym_struct] = ACTIONS(2462),
    [anon_sym_union] = ACTIONS(2462),
    [anon_sym_if] = ACTIONS(2462),
    [anon_sym_else] = ACTIONS(2462),
    [anon_sym_switch] = ACTIONS(2462),
    [anon_sym_case] = ACTIONS(2462),
    [anon_sym_default] = ACTIONS(2462),
    [anon_sym_while] = ACTIONS(2462),
    [anon_sym_do] = ACTIONS(2462),
    [anon_sym_for] = ACTIONS(2462),
    [anon_sym_in] = ACTIONS(2462),
    [anon_sym_return] = ACTIONS(2462),
    [anon_sym_break] = ACTIONS(2462),
    [anon_sym_continue] = ACTIONS(2462),
    [anon_sym_goto] = ACTIONS(2462),
    [anon_sym_DASH_DASH] = ACTIONS(2464),
    [anon_sym_PLUS_PLUS] = ACTIONS(2464),
    [anon_sym_sizeof] = ACTIONS(2462),
    [anon_sym___alignof__] = ACTIONS(2462),
    [anon_sym___alignof] = ACTIONS(2462),
    [anon_sym__alignof] = ACTIONS(2462),
    [anon_sym_alignof] = ACTIONS(2462),
    [anon_sym__Alignof] = ACTIONS(2462),
    [anon_sym_offsetof] = ACTIONS(2462),
    [anon_sym__Generic] = ACTIONS(2462),
    [anon_sym_asm] = ACTIONS(2462),
    [anon_sym___asm__] = ACTIONS(2462),
    [anon_sym___asm] = ACTIONS(2462),
    [sym_number_literal] = ACTIONS(2464),
    [anon_sym_L_SQUOTE] = ACTIONS(2464),
    [anon_sym_u_SQUOTE] = ACTIONS(2464),
    [anon_sym_U_SQUOTE] = ACTIONS(2464),
    [anon_sym_u8_SQUOTE] = ACTIONS(2464),
    [anon_sym_SQUOTE] = ACTIONS(2464),
    [anon_sym_AT] = ACTIONS(2462),
    [anon_sym_DQUOTE] = ACTIONS(2464),
    [anon_sym_L_DQUOTE] = ACTIONS(2464),
    [anon_sym_u_DQUOTE] = ACTIONS(2464),
    [anon_sym_U_DQUOTE] = ACTIONS(2464),
    [anon_sym_u8_DQUOTE] = ACTIONS(2464),
    [sym_true] = ACTIONS(2462),
    [sym_false] = ACTIONS(2462),
    [anon_sym_NULL] = ACTIONS(2462),
    [anon_sym_nullptr] = ACTIONS(2462),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2462),
    [anon_sym___typeof] = ACTIONS(2462),
    [anon_sym_typeof] = ACTIONS(2462),
    [anon_sym_ATimport] = ACTIONS(2464),
    [aux_sym_preproc_undef_token1] = ACTIONS(2462),
    [anon_sym_POUND] = ACTIONS(2462),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2462),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_API_AVAILABLE] = ACTIONS(2462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_API_DEPRECATED] = ACTIONS(2462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym___deprecated_msg] = ACTIONS(2462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2462),
    [anon_sym_ATprotocol] = ACTIONS(2464),
    [anon_sym_ATinterface] = ACTIONS(2464),
    [anon_sym_ATimplementation] = ACTIONS(2464),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2464),
    [anon_sym_ATsynthesize] = ACTIONS(2464),
    [anon_sym_ATdynamic] = ACTIONS(2464),
    [anon_sym_ATtry] = ACTIONS(2464),
    [anon_sym___try] = ACTIONS(2462),
    [anon_sym_ATthrow] = ACTIONS(2464),
    [anon_sym_ATselector] = ACTIONS(2464),
    [anon_sym_ATavailable] = ACTIONS(2464),
    [anon_sym___builtin_available] = ACTIONS(2462),
    [anon_sym_va_arg] = ACTIONS(2462),
    [anon_sym_ATencode] = ACTIONS(2464),
    [anon_sym_ATsynchronized] = ACTIONS(2464),
    [anon_sym_BOOL] = ACTIONS(2462),
    [anon_sym_IMP] = ACTIONS(2462),
    [anon_sym_SEL] = ACTIONS(2462),
    [anon_sym_Class] = ACTIONS(2462),
    [anon_sym_id] = ACTIONS(2462),
  },
  [STATE(237)] = {
    [sym_identifier] = ACTIONS(2466),
    [aux_sym_preproc_include_token1] = ACTIONS(2466),
    [aux_sym_preproc_include_token2] = ACTIONS(2466),
    [aux_sym_preproc_def_token1] = ACTIONS(2466),
    [aux_sym_preproc_if_token1] = ACTIONS(2466),
    [aux_sym_preproc_if_token2] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2466),
    [aux_sym_preproc_else_token1] = ACTIONS(2466),
    [aux_sym_preproc_elif_token1] = ACTIONS(2466),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2466),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2466),
    [sym_preproc_directive] = ACTIONS(2466),
    [anon_sym_LPAREN2] = ACTIONS(2468),
    [anon_sym_BANG] = ACTIONS(2468),
    [anon_sym_TILDE] = ACTIONS(2468),
    [anon_sym_DASH] = ACTIONS(2466),
    [anon_sym_PLUS] = ACTIONS(2466),
    [anon_sym_STAR] = ACTIONS(2468),
    [anon_sym_CARET] = ACTIONS(2468),
    [anon_sym_AMP] = ACTIONS(2468),
    [anon_sym_SEMI] = ACTIONS(2468),
    [anon_sym___extension__] = ACTIONS(2466),
    [anon_sym_typedef] = ACTIONS(2466),
    [anon_sym_extern] = ACTIONS(2466),
    [anon_sym___attribute__] = ACTIONS(2466),
    [anon_sym___attribute] = ACTIONS(2466),
    [anon_sym_noreturn] = ACTIONS(2466),
    [anon_sym_LBRACK] = ACTIONS(2468),
    [anon_sym___declspec] = ACTIONS(2466),
    [anon_sym___cdecl] = ACTIONS(2466),
    [anon_sym___clrcall] = ACTIONS(2466),
    [anon_sym___stdcall] = ACTIONS(2466),
    [anon_sym___fastcall] = ACTIONS(2466),
    [anon_sym___thiscall] = ACTIONS(2466),
    [anon_sym___vectorcall] = ACTIONS(2466),
    [anon_sym_LBRACE] = ACTIONS(2468),
    [anon_sym_signed] = ACTIONS(2466),
    [anon_sym_unsigned] = ACTIONS(2466),
    [anon_sym_long] = ACTIONS(2466),
    [anon_sym_short] = ACTIONS(2466),
    [anon_sym_static] = ACTIONS(2466),
    [anon_sym_ATautoreleasepool] = ACTIONS(2468),
    [anon_sym_auto] = ACTIONS(2466),
    [anon_sym_register] = ACTIONS(2466),
    [anon_sym_inline] = ACTIONS(2466),
    [anon_sym___inline] = ACTIONS(2466),
    [anon_sym___inline__] = ACTIONS(2466),
    [anon_sym___forceinline] = ACTIONS(2466),
    [anon_sym_thread_local] = ACTIONS(2466),
    [anon_sym___thread] = ACTIONS(2466),
    [anon_sym_CG_EXTERN] = ACTIONS(2466),
    [anon_sym_CG_INLINE] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2466),
    [anon_sym_IBOutlet] = ACTIONS(2466),
    [anon_sym_IBInspectable] = ACTIONS(2466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2466),
    [anon_sym_NS_INLINE] = ACTIONS(2466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2466),
    [anon_sym_const] = ACTIONS(2466),
    [anon_sym_constexpr] = ACTIONS(2466),
    [anon_sym_volatile] = ACTIONS(2466),
    [anon_sym_restrict] = ACTIONS(2466),
    [anon_sym___restrict__] = ACTIONS(2466),
    [anon_sym__Atomic] = ACTIONS(2466),
    [anon_sym__Noreturn] = ACTIONS(2466),
    [anon_sym__Nonnull] = ACTIONS(2466),
    [anon_sym_nullable] = ACTIONS(2466),
    [anon_sym__Complex] = ACTIONS(2466),
    [anon_sym__Nullable] = ACTIONS(2466),
    [anon_sym__Nullable_result] = ACTIONS(2466),
    [anon_sym__Null_unspecified] = ACTIONS(2466),
    [anon_sym___autoreleasing] = ACTIONS(2466),
    [anon_sym___block] = ACTIONS(2466),
    [anon_sym___bridge] = ACTIONS(2466),
    [anon_sym___bridge_retained] = ACTIONS(2466),
    [anon_sym___bridge_transfer] = ACTIONS(2466),
    [anon_sym___complex] = ACTIONS(2466),
    [anon_sym___const] = ACTIONS(2466),
    [anon_sym___imag] = ACTIONS(2466),
    [anon_sym___kindof] = ACTIONS(2466),
    [anon_sym___nonnull] = ACTIONS(2466),
    [anon_sym___nullable] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2466),
    [anon_sym___real] = ACTIONS(2466),
    [anon_sym___strong] = ACTIONS(2466),
    [anon_sym___unsafe_unretained] = ACTIONS(2466),
    [anon_sym___unused] = ACTIONS(2466),
    [anon_sym___weak] = ACTIONS(2466),
    [anon_sym_alignas] = ACTIONS(2466),
    [anon_sym__Alignas] = ACTIONS(2466),
    [sym_primitive_type] = ACTIONS(2466),
    [anon_sym_enum] = ACTIONS(2466),
    [anon_sym_struct] = ACTIONS(2466),
    [anon_sym_union] = ACTIONS(2466),
    [anon_sym_if] = ACTIONS(2466),
    [anon_sym_else] = ACTIONS(2466),
    [anon_sym_switch] = ACTIONS(2466),
    [anon_sym_case] = ACTIONS(2466),
    [anon_sym_default] = ACTIONS(2466),
    [anon_sym_while] = ACTIONS(2466),
    [anon_sym_do] = ACTIONS(2466),
    [anon_sym_for] = ACTIONS(2466),
    [anon_sym_in] = ACTIONS(2466),
    [anon_sym_return] = ACTIONS(2466),
    [anon_sym_break] = ACTIONS(2466),
    [anon_sym_continue] = ACTIONS(2466),
    [anon_sym_goto] = ACTIONS(2466),
    [anon_sym_DASH_DASH] = ACTIONS(2468),
    [anon_sym_PLUS_PLUS] = ACTIONS(2468),
    [anon_sym_sizeof] = ACTIONS(2466),
    [anon_sym___alignof__] = ACTIONS(2466),
    [anon_sym___alignof] = ACTIONS(2466),
    [anon_sym__alignof] = ACTIONS(2466),
    [anon_sym_alignof] = ACTIONS(2466),
    [anon_sym__Alignof] = ACTIONS(2466),
    [anon_sym_offsetof] = ACTIONS(2466),
    [anon_sym__Generic] = ACTIONS(2466),
    [anon_sym_asm] = ACTIONS(2466),
    [anon_sym___asm__] = ACTIONS(2466),
    [anon_sym___asm] = ACTIONS(2466),
    [sym_number_literal] = ACTIONS(2468),
    [anon_sym_L_SQUOTE] = ACTIONS(2468),
    [anon_sym_u_SQUOTE] = ACTIONS(2468),
    [anon_sym_U_SQUOTE] = ACTIONS(2468),
    [anon_sym_u8_SQUOTE] = ACTIONS(2468),
    [anon_sym_SQUOTE] = ACTIONS(2468),
    [anon_sym_AT] = ACTIONS(2466),
    [anon_sym_DQUOTE] = ACTIONS(2468),
    [anon_sym_L_DQUOTE] = ACTIONS(2468),
    [anon_sym_u_DQUOTE] = ACTIONS(2468),
    [anon_sym_U_DQUOTE] = ACTIONS(2468),
    [anon_sym_u8_DQUOTE] = ACTIONS(2468),
    [sym_true] = ACTIONS(2466),
    [sym_false] = ACTIONS(2466),
    [anon_sym_NULL] = ACTIONS(2466),
    [anon_sym_nullptr] = ACTIONS(2466),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2466),
    [anon_sym___typeof] = ACTIONS(2466),
    [anon_sym_typeof] = ACTIONS(2466),
    [anon_sym_ATimport] = ACTIONS(2468),
    [aux_sym_preproc_undef_token1] = ACTIONS(2466),
    [anon_sym_POUND] = ACTIONS(2466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_API_AVAILABLE] = ACTIONS(2466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_API_DEPRECATED] = ACTIONS(2466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym___deprecated_msg] = ACTIONS(2466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2466),
    [anon_sym_ATprotocol] = ACTIONS(2468),
    [anon_sym_ATinterface] = ACTIONS(2468),
    [anon_sym_ATimplementation] = ACTIONS(2468),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2468),
    [anon_sym_ATsynthesize] = ACTIONS(2468),
    [anon_sym_ATdynamic] = ACTIONS(2468),
    [anon_sym_ATtry] = ACTIONS(2468),
    [anon_sym___try] = ACTIONS(2466),
    [anon_sym_ATthrow] = ACTIONS(2468),
    [anon_sym_ATselector] = ACTIONS(2468),
    [anon_sym_ATavailable] = ACTIONS(2468),
    [anon_sym___builtin_available] = ACTIONS(2466),
    [anon_sym_va_arg] = ACTIONS(2466),
    [anon_sym_ATencode] = ACTIONS(2468),
    [anon_sym_ATsynchronized] = ACTIONS(2468),
    [anon_sym_BOOL] = ACTIONS(2466),
    [anon_sym_IMP] = ACTIONS(2466),
    [anon_sym_SEL] = ACTIONS(2466),
    [anon_sym_Class] = ACTIONS(2466),
    [anon_sym_id] = ACTIONS(2466),
  },
  [STATE(238)] = {
    [sym_identifier] = ACTIONS(2470),
    [aux_sym_preproc_include_token1] = ACTIONS(2470),
    [aux_sym_preproc_include_token2] = ACTIONS(2470),
    [aux_sym_preproc_def_token1] = ACTIONS(2470),
    [aux_sym_preproc_if_token1] = ACTIONS(2470),
    [aux_sym_preproc_if_token2] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2470),
    [aux_sym_preproc_else_token1] = ACTIONS(2470),
    [aux_sym_preproc_elif_token1] = ACTIONS(2470),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2470),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2470),
    [sym_preproc_directive] = ACTIONS(2470),
    [anon_sym_LPAREN2] = ACTIONS(2472),
    [anon_sym_BANG] = ACTIONS(2472),
    [anon_sym_TILDE] = ACTIONS(2472),
    [anon_sym_DASH] = ACTIONS(2470),
    [anon_sym_PLUS] = ACTIONS(2470),
    [anon_sym_STAR] = ACTIONS(2472),
    [anon_sym_CARET] = ACTIONS(2472),
    [anon_sym_AMP] = ACTIONS(2472),
    [anon_sym_SEMI] = ACTIONS(2472),
    [anon_sym___extension__] = ACTIONS(2470),
    [anon_sym_typedef] = ACTIONS(2470),
    [anon_sym_extern] = ACTIONS(2470),
    [anon_sym___attribute__] = ACTIONS(2470),
    [anon_sym___attribute] = ACTIONS(2470),
    [anon_sym_noreturn] = ACTIONS(2470),
    [anon_sym_LBRACK] = ACTIONS(2472),
    [anon_sym___declspec] = ACTIONS(2470),
    [anon_sym___cdecl] = ACTIONS(2470),
    [anon_sym___clrcall] = ACTIONS(2470),
    [anon_sym___stdcall] = ACTIONS(2470),
    [anon_sym___fastcall] = ACTIONS(2470),
    [anon_sym___thiscall] = ACTIONS(2470),
    [anon_sym___vectorcall] = ACTIONS(2470),
    [anon_sym_LBRACE] = ACTIONS(2472),
    [anon_sym_signed] = ACTIONS(2470),
    [anon_sym_unsigned] = ACTIONS(2470),
    [anon_sym_long] = ACTIONS(2470),
    [anon_sym_short] = ACTIONS(2470),
    [anon_sym_static] = ACTIONS(2470),
    [anon_sym_ATautoreleasepool] = ACTIONS(2472),
    [anon_sym_auto] = ACTIONS(2470),
    [anon_sym_register] = ACTIONS(2470),
    [anon_sym_inline] = ACTIONS(2470),
    [anon_sym___inline] = ACTIONS(2470),
    [anon_sym___inline__] = ACTIONS(2470),
    [anon_sym___forceinline] = ACTIONS(2470),
    [anon_sym_thread_local] = ACTIONS(2470),
    [anon_sym___thread] = ACTIONS(2470),
    [anon_sym_CG_EXTERN] = ACTIONS(2470),
    [anon_sym_CG_INLINE] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2470),
    [anon_sym_IBOutlet] = ACTIONS(2470),
    [anon_sym_IBInspectable] = ACTIONS(2470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2470),
    [anon_sym_NS_INLINE] = ACTIONS(2470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2470),
    [anon_sym_const] = ACTIONS(2470),
    [anon_sym_constexpr] = ACTIONS(2470),
    [anon_sym_volatile] = ACTIONS(2470),
    [anon_sym_restrict] = ACTIONS(2470),
    [anon_sym___restrict__] = ACTIONS(2470),
    [anon_sym__Atomic] = ACTIONS(2470),
    [anon_sym__Noreturn] = ACTIONS(2470),
    [anon_sym__Nonnull] = ACTIONS(2470),
    [anon_sym_nullable] = ACTIONS(2470),
    [anon_sym__Complex] = ACTIONS(2470),
    [anon_sym__Nullable] = ACTIONS(2470),
    [anon_sym__Nullable_result] = ACTIONS(2470),
    [anon_sym__Null_unspecified] = ACTIONS(2470),
    [anon_sym___autoreleasing] = ACTIONS(2470),
    [anon_sym___block] = ACTIONS(2470),
    [anon_sym___bridge] = ACTIONS(2470),
    [anon_sym___bridge_retained] = ACTIONS(2470),
    [anon_sym___bridge_transfer] = ACTIONS(2470),
    [anon_sym___complex] = ACTIONS(2470),
    [anon_sym___const] = ACTIONS(2470),
    [anon_sym___imag] = ACTIONS(2470),
    [anon_sym___kindof] = ACTIONS(2470),
    [anon_sym___nonnull] = ACTIONS(2470),
    [anon_sym___nullable] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2470),
    [anon_sym___real] = ACTIONS(2470),
    [anon_sym___strong] = ACTIONS(2470),
    [anon_sym___unsafe_unretained] = ACTIONS(2470),
    [anon_sym___unused] = ACTIONS(2470),
    [anon_sym___weak] = ACTIONS(2470),
    [anon_sym_alignas] = ACTIONS(2470),
    [anon_sym__Alignas] = ACTIONS(2470),
    [sym_primitive_type] = ACTIONS(2470),
    [anon_sym_enum] = ACTIONS(2470),
    [anon_sym_struct] = ACTIONS(2470),
    [anon_sym_union] = ACTIONS(2470),
    [anon_sym_if] = ACTIONS(2470),
    [anon_sym_else] = ACTIONS(2470),
    [anon_sym_switch] = ACTIONS(2470),
    [anon_sym_case] = ACTIONS(2470),
    [anon_sym_default] = ACTIONS(2470),
    [anon_sym_while] = ACTIONS(2470),
    [anon_sym_do] = ACTIONS(2470),
    [anon_sym_for] = ACTIONS(2470),
    [anon_sym_in] = ACTIONS(2470),
    [anon_sym_return] = ACTIONS(2470),
    [anon_sym_break] = ACTIONS(2470),
    [anon_sym_continue] = ACTIONS(2470),
    [anon_sym_goto] = ACTIONS(2470),
    [anon_sym_DASH_DASH] = ACTIONS(2472),
    [anon_sym_PLUS_PLUS] = ACTIONS(2472),
    [anon_sym_sizeof] = ACTIONS(2470),
    [anon_sym___alignof__] = ACTIONS(2470),
    [anon_sym___alignof] = ACTIONS(2470),
    [anon_sym__alignof] = ACTIONS(2470),
    [anon_sym_alignof] = ACTIONS(2470),
    [anon_sym__Alignof] = ACTIONS(2470),
    [anon_sym_offsetof] = ACTIONS(2470),
    [anon_sym__Generic] = ACTIONS(2470),
    [anon_sym_asm] = ACTIONS(2470),
    [anon_sym___asm__] = ACTIONS(2470),
    [anon_sym___asm] = ACTIONS(2470),
    [sym_number_literal] = ACTIONS(2472),
    [anon_sym_L_SQUOTE] = ACTIONS(2472),
    [anon_sym_u_SQUOTE] = ACTIONS(2472),
    [anon_sym_U_SQUOTE] = ACTIONS(2472),
    [anon_sym_u8_SQUOTE] = ACTIONS(2472),
    [anon_sym_SQUOTE] = ACTIONS(2472),
    [anon_sym_AT] = ACTIONS(2470),
    [anon_sym_DQUOTE] = ACTIONS(2472),
    [anon_sym_L_DQUOTE] = ACTIONS(2472),
    [anon_sym_u_DQUOTE] = ACTIONS(2472),
    [anon_sym_U_DQUOTE] = ACTIONS(2472),
    [anon_sym_u8_DQUOTE] = ACTIONS(2472),
    [sym_true] = ACTIONS(2470),
    [sym_false] = ACTIONS(2470),
    [anon_sym_NULL] = ACTIONS(2470),
    [anon_sym_nullptr] = ACTIONS(2470),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2470),
    [anon_sym___typeof] = ACTIONS(2470),
    [anon_sym_typeof] = ACTIONS(2470),
    [anon_sym_ATimport] = ACTIONS(2472),
    [aux_sym_preproc_undef_token1] = ACTIONS(2470),
    [anon_sym_POUND] = ACTIONS(2470),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_API_AVAILABLE] = ACTIONS(2470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_API_DEPRECATED] = ACTIONS(2470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym___deprecated_msg] = ACTIONS(2470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2470),
    [anon_sym_ATprotocol] = ACTIONS(2472),
    [anon_sym_ATinterface] = ACTIONS(2472),
    [anon_sym_ATimplementation] = ACTIONS(2472),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2472),
    [anon_sym_ATsynthesize] = ACTIONS(2472),
    [anon_sym_ATdynamic] = ACTIONS(2472),
    [anon_sym_ATtry] = ACTIONS(2472),
    [anon_sym___try] = ACTIONS(2470),
    [anon_sym_ATthrow] = ACTIONS(2472),
    [anon_sym_ATselector] = ACTIONS(2472),
    [anon_sym_ATavailable] = ACTIONS(2472),
    [anon_sym___builtin_available] = ACTIONS(2470),
    [anon_sym_va_arg] = ACTIONS(2470),
    [anon_sym_ATencode] = ACTIONS(2472),
    [anon_sym_ATsynchronized] = ACTIONS(2472),
    [anon_sym_BOOL] = ACTIONS(2470),
    [anon_sym_IMP] = ACTIONS(2470),
    [anon_sym_SEL] = ACTIONS(2470),
    [anon_sym_Class] = ACTIONS(2470),
    [anon_sym_id] = ACTIONS(2470),
  },
  [STATE(239)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token2] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [aux_sym_preproc_else_token1] = ACTIONS(2474),
    [aux_sym_preproc_elif_token1] = ACTIONS(2474),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(240)] = {
    [sym_identifier] = ACTIONS(2478),
    [aux_sym_preproc_include_token1] = ACTIONS(2478),
    [aux_sym_preproc_include_token2] = ACTIONS(2478),
    [aux_sym_preproc_def_token1] = ACTIONS(2478),
    [aux_sym_preproc_if_token1] = ACTIONS(2478),
    [aux_sym_preproc_if_token2] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2478),
    [aux_sym_preproc_else_token1] = ACTIONS(2478),
    [aux_sym_preproc_elif_token1] = ACTIONS(2478),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2478),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2478),
    [sym_preproc_directive] = ACTIONS(2478),
    [anon_sym_LPAREN2] = ACTIONS(2480),
    [anon_sym_BANG] = ACTIONS(2480),
    [anon_sym_TILDE] = ACTIONS(2480),
    [anon_sym_DASH] = ACTIONS(2478),
    [anon_sym_PLUS] = ACTIONS(2478),
    [anon_sym_STAR] = ACTIONS(2480),
    [anon_sym_CARET] = ACTIONS(2480),
    [anon_sym_AMP] = ACTIONS(2480),
    [anon_sym_SEMI] = ACTIONS(2480),
    [anon_sym___extension__] = ACTIONS(2478),
    [anon_sym_typedef] = ACTIONS(2478),
    [anon_sym_extern] = ACTIONS(2478),
    [anon_sym___attribute__] = ACTIONS(2478),
    [anon_sym___attribute] = ACTIONS(2478),
    [anon_sym_noreturn] = ACTIONS(2478),
    [anon_sym_LBRACK] = ACTIONS(2480),
    [anon_sym___declspec] = ACTIONS(2478),
    [anon_sym___cdecl] = ACTIONS(2478),
    [anon_sym___clrcall] = ACTIONS(2478),
    [anon_sym___stdcall] = ACTIONS(2478),
    [anon_sym___fastcall] = ACTIONS(2478),
    [anon_sym___thiscall] = ACTIONS(2478),
    [anon_sym___vectorcall] = ACTIONS(2478),
    [anon_sym_LBRACE] = ACTIONS(2480),
    [anon_sym_signed] = ACTIONS(2478),
    [anon_sym_unsigned] = ACTIONS(2478),
    [anon_sym_long] = ACTIONS(2478),
    [anon_sym_short] = ACTIONS(2478),
    [anon_sym_static] = ACTIONS(2478),
    [anon_sym_ATautoreleasepool] = ACTIONS(2480),
    [anon_sym_auto] = ACTIONS(2478),
    [anon_sym_register] = ACTIONS(2478),
    [anon_sym_inline] = ACTIONS(2478),
    [anon_sym___inline] = ACTIONS(2478),
    [anon_sym___inline__] = ACTIONS(2478),
    [anon_sym___forceinline] = ACTIONS(2478),
    [anon_sym_thread_local] = ACTIONS(2478),
    [anon_sym___thread] = ACTIONS(2478),
    [anon_sym_CG_EXTERN] = ACTIONS(2478),
    [anon_sym_CG_INLINE] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2478),
    [anon_sym_IBOutlet] = ACTIONS(2478),
    [anon_sym_IBInspectable] = ACTIONS(2478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2478),
    [anon_sym_NS_INLINE] = ACTIONS(2478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2478),
    [anon_sym_const] = ACTIONS(2478),
    [anon_sym_constexpr] = ACTIONS(2478),
    [anon_sym_volatile] = ACTIONS(2478),
    [anon_sym_restrict] = ACTIONS(2478),
    [anon_sym___restrict__] = ACTIONS(2478),
    [anon_sym__Atomic] = ACTIONS(2478),
    [anon_sym__Noreturn] = ACTIONS(2478),
    [anon_sym__Nonnull] = ACTIONS(2478),
    [anon_sym_nullable] = ACTIONS(2478),
    [anon_sym__Complex] = ACTIONS(2478),
    [anon_sym__Nullable] = ACTIONS(2478),
    [anon_sym__Nullable_result] = ACTIONS(2478),
    [anon_sym__Null_unspecified] = ACTIONS(2478),
    [anon_sym___autoreleasing] = ACTIONS(2478),
    [anon_sym___block] = ACTIONS(2478),
    [anon_sym___bridge] = ACTIONS(2478),
    [anon_sym___bridge_retained] = ACTIONS(2478),
    [anon_sym___bridge_transfer] = ACTIONS(2478),
    [anon_sym___complex] = ACTIONS(2478),
    [anon_sym___const] = ACTIONS(2478),
    [anon_sym___imag] = ACTIONS(2478),
    [anon_sym___kindof] = ACTIONS(2478),
    [anon_sym___nonnull] = ACTIONS(2478),
    [anon_sym___nullable] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2478),
    [anon_sym___real] = ACTIONS(2478),
    [anon_sym___strong] = ACTIONS(2478),
    [anon_sym___unsafe_unretained] = ACTIONS(2478),
    [anon_sym___unused] = ACTIONS(2478),
    [anon_sym___weak] = ACTIONS(2478),
    [anon_sym_alignas] = ACTIONS(2478),
    [anon_sym__Alignas] = ACTIONS(2478),
    [sym_primitive_type] = ACTIONS(2478),
    [anon_sym_enum] = ACTIONS(2478),
    [anon_sym_struct] = ACTIONS(2478),
    [anon_sym_union] = ACTIONS(2478),
    [anon_sym_if] = ACTIONS(2478),
    [anon_sym_else] = ACTIONS(2478),
    [anon_sym_switch] = ACTIONS(2478),
    [anon_sym_case] = ACTIONS(2478),
    [anon_sym_default] = ACTIONS(2478),
    [anon_sym_while] = ACTIONS(2478),
    [anon_sym_do] = ACTIONS(2478),
    [anon_sym_for] = ACTIONS(2478),
    [anon_sym_in] = ACTIONS(2478),
    [anon_sym_return] = ACTIONS(2478),
    [anon_sym_break] = ACTIONS(2478),
    [anon_sym_continue] = ACTIONS(2478),
    [anon_sym_goto] = ACTIONS(2478),
    [anon_sym_DASH_DASH] = ACTIONS(2480),
    [anon_sym_PLUS_PLUS] = ACTIONS(2480),
    [anon_sym_sizeof] = ACTIONS(2478),
    [anon_sym___alignof__] = ACTIONS(2478),
    [anon_sym___alignof] = ACTIONS(2478),
    [anon_sym__alignof] = ACTIONS(2478),
    [anon_sym_alignof] = ACTIONS(2478),
    [anon_sym__Alignof] = ACTIONS(2478),
    [anon_sym_offsetof] = ACTIONS(2478),
    [anon_sym__Generic] = ACTIONS(2478),
    [anon_sym_asm] = ACTIONS(2478),
    [anon_sym___asm__] = ACTIONS(2478),
    [anon_sym___asm] = ACTIONS(2478),
    [sym_number_literal] = ACTIONS(2480),
    [anon_sym_L_SQUOTE] = ACTIONS(2480),
    [anon_sym_u_SQUOTE] = ACTIONS(2480),
    [anon_sym_U_SQUOTE] = ACTIONS(2480),
    [anon_sym_u8_SQUOTE] = ACTIONS(2480),
    [anon_sym_SQUOTE] = ACTIONS(2480),
    [anon_sym_AT] = ACTIONS(2478),
    [anon_sym_DQUOTE] = ACTIONS(2480),
    [anon_sym_L_DQUOTE] = ACTIONS(2480),
    [anon_sym_u_DQUOTE] = ACTIONS(2480),
    [anon_sym_U_DQUOTE] = ACTIONS(2480),
    [anon_sym_u8_DQUOTE] = ACTIONS(2480),
    [sym_true] = ACTIONS(2478),
    [sym_false] = ACTIONS(2478),
    [anon_sym_NULL] = ACTIONS(2478),
    [anon_sym_nullptr] = ACTIONS(2478),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2478),
    [anon_sym___typeof] = ACTIONS(2478),
    [anon_sym_typeof] = ACTIONS(2478),
    [anon_sym_ATimport] = ACTIONS(2480),
    [aux_sym_preproc_undef_token1] = ACTIONS(2478),
    [anon_sym_POUND] = ACTIONS(2478),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_API_AVAILABLE] = ACTIONS(2478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_API_DEPRECATED] = ACTIONS(2478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym___deprecated_msg] = ACTIONS(2478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2478),
    [anon_sym_ATprotocol] = ACTIONS(2480),
    [anon_sym_ATinterface] = ACTIONS(2480),
    [anon_sym_ATimplementation] = ACTIONS(2480),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2480),
    [anon_sym_ATsynthesize] = ACTIONS(2480),
    [anon_sym_ATdynamic] = ACTIONS(2480),
    [anon_sym_ATtry] = ACTIONS(2480),
    [anon_sym___try] = ACTIONS(2478),
    [anon_sym_ATthrow] = ACTIONS(2480),
    [anon_sym_ATselector] = ACTIONS(2480),
    [anon_sym_ATavailable] = ACTIONS(2480),
    [anon_sym___builtin_available] = ACTIONS(2478),
    [anon_sym_va_arg] = ACTIONS(2478),
    [anon_sym_ATencode] = ACTIONS(2480),
    [anon_sym_ATsynchronized] = ACTIONS(2480),
    [anon_sym_BOOL] = ACTIONS(2478),
    [anon_sym_IMP] = ACTIONS(2478),
    [anon_sym_SEL] = ACTIONS(2478),
    [anon_sym_Class] = ACTIONS(2478),
    [anon_sym_id] = ACTIONS(2478),
  },
  [STATE(241)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(242)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token2] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [aux_sym_preproc_else_token1] = ACTIONS(2446),
    [aux_sym_preproc_elif_token1] = ACTIONS(2446),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(243)] = {
    [sym_identifier] = ACTIONS(2482),
    [aux_sym_preproc_include_token1] = ACTIONS(2482),
    [aux_sym_preproc_include_token2] = ACTIONS(2482),
    [aux_sym_preproc_def_token1] = ACTIONS(2482),
    [aux_sym_preproc_if_token1] = ACTIONS(2482),
    [aux_sym_preproc_if_token2] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2482),
    [aux_sym_preproc_else_token1] = ACTIONS(2482),
    [aux_sym_preproc_elif_token1] = ACTIONS(2482),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2482),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2482),
    [sym_preproc_directive] = ACTIONS(2482),
    [anon_sym_LPAREN2] = ACTIONS(2484),
    [anon_sym_BANG] = ACTIONS(2484),
    [anon_sym_TILDE] = ACTIONS(2484),
    [anon_sym_DASH] = ACTIONS(2482),
    [anon_sym_PLUS] = ACTIONS(2482),
    [anon_sym_STAR] = ACTIONS(2484),
    [anon_sym_CARET] = ACTIONS(2484),
    [anon_sym_AMP] = ACTIONS(2484),
    [anon_sym_SEMI] = ACTIONS(2484),
    [anon_sym___extension__] = ACTIONS(2482),
    [anon_sym_typedef] = ACTIONS(2482),
    [anon_sym_extern] = ACTIONS(2482),
    [anon_sym___attribute__] = ACTIONS(2482),
    [anon_sym___attribute] = ACTIONS(2482),
    [anon_sym_noreturn] = ACTIONS(2482),
    [anon_sym_LBRACK] = ACTIONS(2484),
    [anon_sym___declspec] = ACTIONS(2482),
    [anon_sym___cdecl] = ACTIONS(2482),
    [anon_sym___clrcall] = ACTIONS(2482),
    [anon_sym___stdcall] = ACTIONS(2482),
    [anon_sym___fastcall] = ACTIONS(2482),
    [anon_sym___thiscall] = ACTIONS(2482),
    [anon_sym___vectorcall] = ACTIONS(2482),
    [anon_sym_LBRACE] = ACTIONS(2484),
    [anon_sym_signed] = ACTIONS(2482),
    [anon_sym_unsigned] = ACTIONS(2482),
    [anon_sym_long] = ACTIONS(2482),
    [anon_sym_short] = ACTIONS(2482),
    [anon_sym_static] = ACTIONS(2482),
    [anon_sym_ATautoreleasepool] = ACTIONS(2484),
    [anon_sym_auto] = ACTIONS(2482),
    [anon_sym_register] = ACTIONS(2482),
    [anon_sym_inline] = ACTIONS(2482),
    [anon_sym___inline] = ACTIONS(2482),
    [anon_sym___inline__] = ACTIONS(2482),
    [anon_sym___forceinline] = ACTIONS(2482),
    [anon_sym_thread_local] = ACTIONS(2482),
    [anon_sym___thread] = ACTIONS(2482),
    [anon_sym_CG_EXTERN] = ACTIONS(2482),
    [anon_sym_CG_INLINE] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2482),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2482),
    [anon_sym_IBOutlet] = ACTIONS(2482),
    [anon_sym_IBInspectable] = ACTIONS(2482),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2482),
    [anon_sym_NS_INLINE] = ACTIONS(2482),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2482),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2482),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2482),
    [anon_sym_const] = ACTIONS(2482),
    [anon_sym_constexpr] = ACTIONS(2482),
    [anon_sym_volatile] = ACTIONS(2482),
    [anon_sym_restrict] = ACTIONS(2482),
    [anon_sym___restrict__] = ACTIONS(2482),
    [anon_sym__Atomic] = ACTIONS(2482),
    [anon_sym__Noreturn] = ACTIONS(2482),
    [anon_sym__Nonnull] = ACTIONS(2482),
    [anon_sym_nullable] = ACTIONS(2482),
    [anon_sym__Complex] = ACTIONS(2482),
    [anon_sym__Nullable] = ACTIONS(2482),
    [anon_sym__Nullable_result] = ACTIONS(2482),
    [anon_sym__Null_unspecified] = ACTIONS(2482),
    [anon_sym___autoreleasing] = ACTIONS(2482),
    [anon_sym___block] = ACTIONS(2482),
    [anon_sym___bridge] = ACTIONS(2482),
    [anon_sym___bridge_retained] = ACTIONS(2482),
    [anon_sym___bridge_transfer] = ACTIONS(2482),
    [anon_sym___complex] = ACTIONS(2482),
    [anon_sym___const] = ACTIONS(2482),
    [anon_sym___imag] = ACTIONS(2482),
    [anon_sym___kindof] = ACTIONS(2482),
    [anon_sym___nonnull] = ACTIONS(2482),
    [anon_sym___nullable] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2482),
    [anon_sym___real] = ACTIONS(2482),
    [anon_sym___strong] = ACTIONS(2482),
    [anon_sym___unsafe_unretained] = ACTIONS(2482),
    [anon_sym___unused] = ACTIONS(2482),
    [anon_sym___weak] = ACTIONS(2482),
    [anon_sym_alignas] = ACTIONS(2482),
    [anon_sym__Alignas] = ACTIONS(2482),
    [sym_primitive_type] = ACTIONS(2482),
    [anon_sym_enum] = ACTIONS(2482),
    [anon_sym_struct] = ACTIONS(2482),
    [anon_sym_union] = ACTIONS(2482),
    [anon_sym_if] = ACTIONS(2482),
    [anon_sym_else] = ACTIONS(2482),
    [anon_sym_switch] = ACTIONS(2482),
    [anon_sym_case] = ACTIONS(2482),
    [anon_sym_default] = ACTIONS(2482),
    [anon_sym_while] = ACTIONS(2482),
    [anon_sym_do] = ACTIONS(2482),
    [anon_sym_for] = ACTIONS(2482),
    [anon_sym_in] = ACTIONS(2482),
    [anon_sym_return] = ACTIONS(2482),
    [anon_sym_break] = ACTIONS(2482),
    [anon_sym_continue] = ACTIONS(2482),
    [anon_sym_goto] = ACTIONS(2482),
    [anon_sym_DASH_DASH] = ACTIONS(2484),
    [anon_sym_PLUS_PLUS] = ACTIONS(2484),
    [anon_sym_sizeof] = ACTIONS(2482),
    [anon_sym___alignof__] = ACTIONS(2482),
    [anon_sym___alignof] = ACTIONS(2482),
    [anon_sym__alignof] = ACTIONS(2482),
    [anon_sym_alignof] = ACTIONS(2482),
    [anon_sym__Alignof] = ACTIONS(2482),
    [anon_sym_offsetof] = ACTIONS(2482),
    [anon_sym__Generic] = ACTIONS(2482),
    [anon_sym_asm] = ACTIONS(2482),
    [anon_sym___asm__] = ACTIONS(2482),
    [anon_sym___asm] = ACTIONS(2482),
    [sym_number_literal] = ACTIONS(2484),
    [anon_sym_L_SQUOTE] = ACTIONS(2484),
    [anon_sym_u_SQUOTE] = ACTIONS(2484),
    [anon_sym_U_SQUOTE] = ACTIONS(2484),
    [anon_sym_u8_SQUOTE] = ACTIONS(2484),
    [anon_sym_SQUOTE] = ACTIONS(2484),
    [anon_sym_AT] = ACTIONS(2482),
    [anon_sym_DQUOTE] = ACTIONS(2484),
    [anon_sym_L_DQUOTE] = ACTIONS(2484),
    [anon_sym_u_DQUOTE] = ACTIONS(2484),
    [anon_sym_U_DQUOTE] = ACTIONS(2484),
    [anon_sym_u8_DQUOTE] = ACTIONS(2484),
    [sym_true] = ACTIONS(2482),
    [sym_false] = ACTIONS(2482),
    [anon_sym_NULL] = ACTIONS(2482),
    [anon_sym_nullptr] = ACTIONS(2482),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2482),
    [anon_sym___typeof] = ACTIONS(2482),
    [anon_sym_typeof] = ACTIONS(2482),
    [anon_sym_ATimport] = ACTIONS(2484),
    [aux_sym_preproc_undef_token1] = ACTIONS(2482),
    [anon_sym_POUND] = ACTIONS(2482),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2482),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2482),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2482),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_API_AVAILABLE] = ACTIONS(2482),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_API_DEPRECATED] = ACTIONS(2482),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym___deprecated_msg] = ACTIONS(2482),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2482),
    [anon_sym_ATprotocol] = ACTIONS(2484),
    [anon_sym_ATinterface] = ACTIONS(2484),
    [anon_sym_ATimplementation] = ACTIONS(2484),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2484),
    [anon_sym_ATsynthesize] = ACTIONS(2484),
    [anon_sym_ATdynamic] = ACTIONS(2484),
    [anon_sym_ATtry] = ACTIONS(2484),
    [anon_sym___try] = ACTIONS(2482),
    [anon_sym_ATthrow] = ACTIONS(2484),
    [anon_sym_ATselector] = ACTIONS(2484),
    [anon_sym_ATavailable] = ACTIONS(2484),
    [anon_sym___builtin_available] = ACTIONS(2482),
    [anon_sym_va_arg] = ACTIONS(2482),
    [anon_sym_ATencode] = ACTIONS(2484),
    [anon_sym_ATsynchronized] = ACTIONS(2484),
    [anon_sym_BOOL] = ACTIONS(2482),
    [anon_sym_IMP] = ACTIONS(2482),
    [anon_sym_SEL] = ACTIONS(2482),
    [anon_sym_Class] = ACTIONS(2482),
    [anon_sym_id] = ACTIONS(2482),
  },
  [STATE(244)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token2] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [aux_sym_preproc_else_token1] = ACTIONS(2346),
    [aux_sym_preproc_elif_token1] = ACTIONS(2346),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATsynthesize] = ACTIONS(2348),
    [anon_sym_ATdynamic] = ACTIONS(2348),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(245)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(246)] = {
    [sym_identifier] = ACTIONS(2486),
    [aux_sym_preproc_include_token1] = ACTIONS(2486),
    [aux_sym_preproc_include_token2] = ACTIONS(2486),
    [aux_sym_preproc_def_token1] = ACTIONS(2486),
    [aux_sym_preproc_if_token1] = ACTIONS(2486),
    [aux_sym_preproc_if_token2] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2486),
    [aux_sym_preproc_else_token1] = ACTIONS(2486),
    [aux_sym_preproc_elif_token1] = ACTIONS(2486),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2486),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2486),
    [sym_preproc_directive] = ACTIONS(2486),
    [anon_sym_LPAREN2] = ACTIONS(2488),
    [anon_sym_BANG] = ACTIONS(2488),
    [anon_sym_TILDE] = ACTIONS(2488),
    [anon_sym_DASH] = ACTIONS(2486),
    [anon_sym_PLUS] = ACTIONS(2486),
    [anon_sym_STAR] = ACTIONS(2488),
    [anon_sym_CARET] = ACTIONS(2488),
    [anon_sym_AMP] = ACTIONS(2488),
    [anon_sym_SEMI] = ACTIONS(2488),
    [anon_sym___extension__] = ACTIONS(2486),
    [anon_sym_typedef] = ACTIONS(2486),
    [anon_sym_extern] = ACTIONS(2486),
    [anon_sym___attribute__] = ACTIONS(2486),
    [anon_sym___attribute] = ACTIONS(2486),
    [anon_sym_noreturn] = ACTIONS(2486),
    [anon_sym_LBRACK] = ACTIONS(2488),
    [anon_sym___declspec] = ACTIONS(2486),
    [anon_sym___cdecl] = ACTIONS(2486),
    [anon_sym___clrcall] = ACTIONS(2486),
    [anon_sym___stdcall] = ACTIONS(2486),
    [anon_sym___fastcall] = ACTIONS(2486),
    [anon_sym___thiscall] = ACTIONS(2486),
    [anon_sym___vectorcall] = ACTIONS(2486),
    [anon_sym_LBRACE] = ACTIONS(2488),
    [anon_sym_signed] = ACTIONS(2486),
    [anon_sym_unsigned] = ACTIONS(2486),
    [anon_sym_long] = ACTIONS(2486),
    [anon_sym_short] = ACTIONS(2486),
    [anon_sym_static] = ACTIONS(2486),
    [anon_sym_ATautoreleasepool] = ACTIONS(2488),
    [anon_sym_auto] = ACTIONS(2486),
    [anon_sym_register] = ACTIONS(2486),
    [anon_sym_inline] = ACTIONS(2486),
    [anon_sym___inline] = ACTIONS(2486),
    [anon_sym___inline__] = ACTIONS(2486),
    [anon_sym___forceinline] = ACTIONS(2486),
    [anon_sym_thread_local] = ACTIONS(2486),
    [anon_sym___thread] = ACTIONS(2486),
    [anon_sym_CG_EXTERN] = ACTIONS(2486),
    [anon_sym_CG_INLINE] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2486),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2486),
    [anon_sym_IBOutlet] = ACTIONS(2486),
    [anon_sym_IBInspectable] = ACTIONS(2486),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2486),
    [anon_sym_NS_INLINE] = ACTIONS(2486),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2486),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2486),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2486),
    [anon_sym_const] = ACTIONS(2486),
    [anon_sym_constexpr] = ACTIONS(2486),
    [anon_sym_volatile] = ACTIONS(2486),
    [anon_sym_restrict] = ACTIONS(2486),
    [anon_sym___restrict__] = ACTIONS(2486),
    [anon_sym__Atomic] = ACTIONS(2486),
    [anon_sym__Noreturn] = ACTIONS(2486),
    [anon_sym__Nonnull] = ACTIONS(2486),
    [anon_sym_nullable] = ACTIONS(2486),
    [anon_sym__Complex] = ACTIONS(2486),
    [anon_sym__Nullable] = ACTIONS(2486),
    [anon_sym__Nullable_result] = ACTIONS(2486),
    [anon_sym__Null_unspecified] = ACTIONS(2486),
    [anon_sym___autoreleasing] = ACTIONS(2486),
    [anon_sym___block] = ACTIONS(2486),
    [anon_sym___bridge] = ACTIONS(2486),
    [anon_sym___bridge_retained] = ACTIONS(2486),
    [anon_sym___bridge_transfer] = ACTIONS(2486),
    [anon_sym___complex] = ACTIONS(2486),
    [anon_sym___const] = ACTIONS(2486),
    [anon_sym___imag] = ACTIONS(2486),
    [anon_sym___kindof] = ACTIONS(2486),
    [anon_sym___nonnull] = ACTIONS(2486),
    [anon_sym___nullable] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2486),
    [anon_sym___real] = ACTIONS(2486),
    [anon_sym___strong] = ACTIONS(2486),
    [anon_sym___unsafe_unretained] = ACTIONS(2486),
    [anon_sym___unused] = ACTIONS(2486),
    [anon_sym___weak] = ACTIONS(2486),
    [anon_sym_alignas] = ACTIONS(2486),
    [anon_sym__Alignas] = ACTIONS(2486),
    [sym_primitive_type] = ACTIONS(2486),
    [anon_sym_enum] = ACTIONS(2486),
    [anon_sym_struct] = ACTIONS(2486),
    [anon_sym_union] = ACTIONS(2486),
    [anon_sym_if] = ACTIONS(2486),
    [anon_sym_else] = ACTIONS(2486),
    [anon_sym_switch] = ACTIONS(2486),
    [anon_sym_case] = ACTIONS(2486),
    [anon_sym_default] = ACTIONS(2486),
    [anon_sym_while] = ACTIONS(2486),
    [anon_sym_do] = ACTIONS(2486),
    [anon_sym_for] = ACTIONS(2486),
    [anon_sym_in] = ACTIONS(2486),
    [anon_sym_return] = ACTIONS(2486),
    [anon_sym_break] = ACTIONS(2486),
    [anon_sym_continue] = ACTIONS(2486),
    [anon_sym_goto] = ACTIONS(2486),
    [anon_sym_DASH_DASH] = ACTIONS(2488),
    [anon_sym_PLUS_PLUS] = ACTIONS(2488),
    [anon_sym_sizeof] = ACTIONS(2486),
    [anon_sym___alignof__] = ACTIONS(2486),
    [anon_sym___alignof] = ACTIONS(2486),
    [anon_sym__alignof] = ACTIONS(2486),
    [anon_sym_alignof] = ACTIONS(2486),
    [anon_sym__Alignof] = ACTIONS(2486),
    [anon_sym_offsetof] = ACTIONS(2486),
    [anon_sym__Generic] = ACTIONS(2486),
    [anon_sym_asm] = ACTIONS(2486),
    [anon_sym___asm__] = ACTIONS(2486),
    [anon_sym___asm] = ACTIONS(2486),
    [sym_number_literal] = ACTIONS(2488),
    [anon_sym_L_SQUOTE] = ACTIONS(2488),
    [anon_sym_u_SQUOTE] = ACTIONS(2488),
    [anon_sym_U_SQUOTE] = ACTIONS(2488),
    [anon_sym_u8_SQUOTE] = ACTIONS(2488),
    [anon_sym_SQUOTE] = ACTIONS(2488),
    [anon_sym_AT] = ACTIONS(2486),
    [anon_sym_DQUOTE] = ACTIONS(2488),
    [anon_sym_L_DQUOTE] = ACTIONS(2488),
    [anon_sym_u_DQUOTE] = ACTIONS(2488),
    [anon_sym_U_DQUOTE] = ACTIONS(2488),
    [anon_sym_u8_DQUOTE] = ACTIONS(2488),
    [sym_true] = ACTIONS(2486),
    [sym_false] = ACTIONS(2486),
    [anon_sym_NULL] = ACTIONS(2486),
    [anon_sym_nullptr] = ACTIONS(2486),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2486),
    [anon_sym___typeof] = ACTIONS(2486),
    [anon_sym_typeof] = ACTIONS(2486),
    [anon_sym_ATimport] = ACTIONS(2488),
    [aux_sym_preproc_undef_token1] = ACTIONS(2486),
    [anon_sym_POUND] = ACTIONS(2486),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2486),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2486),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2486),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_API_AVAILABLE] = ACTIONS(2486),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_API_DEPRECATED] = ACTIONS(2486),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym___deprecated_msg] = ACTIONS(2486),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2486),
    [anon_sym_ATprotocol] = ACTIONS(2488),
    [anon_sym_ATinterface] = ACTIONS(2488),
    [anon_sym_ATimplementation] = ACTIONS(2488),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2488),
    [anon_sym_ATsynthesize] = ACTIONS(2488),
    [anon_sym_ATdynamic] = ACTIONS(2488),
    [anon_sym_ATtry] = ACTIONS(2488),
    [anon_sym___try] = ACTIONS(2486),
    [anon_sym_ATthrow] = ACTIONS(2488),
    [anon_sym_ATselector] = ACTIONS(2488),
    [anon_sym_ATavailable] = ACTIONS(2488),
    [anon_sym___builtin_available] = ACTIONS(2486),
    [anon_sym_va_arg] = ACTIONS(2486),
    [anon_sym_ATencode] = ACTIONS(2488),
    [anon_sym_ATsynchronized] = ACTIONS(2488),
    [anon_sym_BOOL] = ACTIONS(2486),
    [anon_sym_IMP] = ACTIONS(2486),
    [anon_sym_SEL] = ACTIONS(2486),
    [anon_sym_Class] = ACTIONS(2486),
    [anon_sym_id] = ACTIONS(2486),
  },
  [STATE(247)] = {
    [sym_identifier] = ACTIONS(2490),
    [aux_sym_preproc_include_token1] = ACTIONS(2490),
    [aux_sym_preproc_include_token2] = ACTIONS(2490),
    [aux_sym_preproc_def_token1] = ACTIONS(2490),
    [aux_sym_preproc_if_token1] = ACTIONS(2490),
    [aux_sym_preproc_if_token2] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2490),
    [aux_sym_preproc_else_token1] = ACTIONS(2490),
    [aux_sym_preproc_elif_token1] = ACTIONS(2490),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2490),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2490),
    [sym_preproc_directive] = ACTIONS(2490),
    [anon_sym_LPAREN2] = ACTIONS(2492),
    [anon_sym_BANG] = ACTIONS(2492),
    [anon_sym_TILDE] = ACTIONS(2492),
    [anon_sym_DASH] = ACTIONS(2490),
    [anon_sym_PLUS] = ACTIONS(2490),
    [anon_sym_STAR] = ACTIONS(2492),
    [anon_sym_CARET] = ACTIONS(2492),
    [anon_sym_AMP] = ACTIONS(2492),
    [anon_sym_SEMI] = ACTIONS(2492),
    [anon_sym___extension__] = ACTIONS(2490),
    [anon_sym_typedef] = ACTIONS(2490),
    [anon_sym_extern] = ACTIONS(2490),
    [anon_sym___attribute__] = ACTIONS(2490),
    [anon_sym___attribute] = ACTIONS(2490),
    [anon_sym_noreturn] = ACTIONS(2490),
    [anon_sym_LBRACK] = ACTIONS(2492),
    [anon_sym___declspec] = ACTIONS(2490),
    [anon_sym___cdecl] = ACTIONS(2490),
    [anon_sym___clrcall] = ACTIONS(2490),
    [anon_sym___stdcall] = ACTIONS(2490),
    [anon_sym___fastcall] = ACTIONS(2490),
    [anon_sym___thiscall] = ACTIONS(2490),
    [anon_sym___vectorcall] = ACTIONS(2490),
    [anon_sym_LBRACE] = ACTIONS(2492),
    [anon_sym_signed] = ACTIONS(2490),
    [anon_sym_unsigned] = ACTIONS(2490),
    [anon_sym_long] = ACTIONS(2490),
    [anon_sym_short] = ACTIONS(2490),
    [anon_sym_static] = ACTIONS(2490),
    [anon_sym_ATautoreleasepool] = ACTIONS(2492),
    [anon_sym_auto] = ACTIONS(2490),
    [anon_sym_register] = ACTIONS(2490),
    [anon_sym_inline] = ACTIONS(2490),
    [anon_sym___inline] = ACTIONS(2490),
    [anon_sym___inline__] = ACTIONS(2490),
    [anon_sym___forceinline] = ACTIONS(2490),
    [anon_sym_thread_local] = ACTIONS(2490),
    [anon_sym___thread] = ACTIONS(2490),
    [anon_sym_CG_EXTERN] = ACTIONS(2490),
    [anon_sym_CG_INLINE] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2490),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2490),
    [anon_sym_IBOutlet] = ACTIONS(2490),
    [anon_sym_IBInspectable] = ACTIONS(2490),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2490),
    [anon_sym_NS_INLINE] = ACTIONS(2490),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2490),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2490),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2490),
    [anon_sym_const] = ACTIONS(2490),
    [anon_sym_constexpr] = ACTIONS(2490),
    [anon_sym_volatile] = ACTIONS(2490),
    [anon_sym_restrict] = ACTIONS(2490),
    [anon_sym___restrict__] = ACTIONS(2490),
    [anon_sym__Atomic] = ACTIONS(2490),
    [anon_sym__Noreturn] = ACTIONS(2490),
    [anon_sym__Nonnull] = ACTIONS(2490),
    [anon_sym_nullable] = ACTIONS(2490),
    [anon_sym__Complex] = ACTIONS(2490),
    [anon_sym__Nullable] = ACTIONS(2490),
    [anon_sym__Nullable_result] = ACTIONS(2490),
    [anon_sym__Null_unspecified] = ACTIONS(2490),
    [anon_sym___autoreleasing] = ACTIONS(2490),
    [anon_sym___block] = ACTIONS(2490),
    [anon_sym___bridge] = ACTIONS(2490),
    [anon_sym___bridge_retained] = ACTIONS(2490),
    [anon_sym___bridge_transfer] = ACTIONS(2490),
    [anon_sym___complex] = ACTIONS(2490),
    [anon_sym___const] = ACTIONS(2490),
    [anon_sym___imag] = ACTIONS(2490),
    [anon_sym___kindof] = ACTIONS(2490),
    [anon_sym___nonnull] = ACTIONS(2490),
    [anon_sym___nullable] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2490),
    [anon_sym___real] = ACTIONS(2490),
    [anon_sym___strong] = ACTIONS(2490),
    [anon_sym___unsafe_unretained] = ACTIONS(2490),
    [anon_sym___unused] = ACTIONS(2490),
    [anon_sym___weak] = ACTIONS(2490),
    [anon_sym_alignas] = ACTIONS(2490),
    [anon_sym__Alignas] = ACTIONS(2490),
    [sym_primitive_type] = ACTIONS(2490),
    [anon_sym_enum] = ACTIONS(2490),
    [anon_sym_struct] = ACTIONS(2490),
    [anon_sym_union] = ACTIONS(2490),
    [anon_sym_if] = ACTIONS(2490),
    [anon_sym_else] = ACTIONS(2490),
    [anon_sym_switch] = ACTIONS(2490),
    [anon_sym_case] = ACTIONS(2490),
    [anon_sym_default] = ACTIONS(2490),
    [anon_sym_while] = ACTIONS(2490),
    [anon_sym_do] = ACTIONS(2490),
    [anon_sym_for] = ACTIONS(2490),
    [anon_sym_in] = ACTIONS(2490),
    [anon_sym_return] = ACTIONS(2490),
    [anon_sym_break] = ACTIONS(2490),
    [anon_sym_continue] = ACTIONS(2490),
    [anon_sym_goto] = ACTIONS(2490),
    [anon_sym_DASH_DASH] = ACTIONS(2492),
    [anon_sym_PLUS_PLUS] = ACTIONS(2492),
    [anon_sym_sizeof] = ACTIONS(2490),
    [anon_sym___alignof__] = ACTIONS(2490),
    [anon_sym___alignof] = ACTIONS(2490),
    [anon_sym__alignof] = ACTIONS(2490),
    [anon_sym_alignof] = ACTIONS(2490),
    [anon_sym__Alignof] = ACTIONS(2490),
    [anon_sym_offsetof] = ACTIONS(2490),
    [anon_sym__Generic] = ACTIONS(2490),
    [anon_sym_asm] = ACTIONS(2490),
    [anon_sym___asm__] = ACTIONS(2490),
    [anon_sym___asm] = ACTIONS(2490),
    [sym_number_literal] = ACTIONS(2492),
    [anon_sym_L_SQUOTE] = ACTIONS(2492),
    [anon_sym_u_SQUOTE] = ACTIONS(2492),
    [anon_sym_U_SQUOTE] = ACTIONS(2492),
    [anon_sym_u8_SQUOTE] = ACTIONS(2492),
    [anon_sym_SQUOTE] = ACTIONS(2492),
    [anon_sym_AT] = ACTIONS(2490),
    [anon_sym_DQUOTE] = ACTIONS(2492),
    [anon_sym_L_DQUOTE] = ACTIONS(2492),
    [anon_sym_u_DQUOTE] = ACTIONS(2492),
    [anon_sym_U_DQUOTE] = ACTIONS(2492),
    [anon_sym_u8_DQUOTE] = ACTIONS(2492),
    [sym_true] = ACTIONS(2490),
    [sym_false] = ACTIONS(2490),
    [anon_sym_NULL] = ACTIONS(2490),
    [anon_sym_nullptr] = ACTIONS(2490),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2490),
    [anon_sym___typeof] = ACTIONS(2490),
    [anon_sym_typeof] = ACTIONS(2490),
    [anon_sym_ATimport] = ACTIONS(2492),
    [aux_sym_preproc_undef_token1] = ACTIONS(2490),
    [anon_sym_POUND] = ACTIONS(2490),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2490),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2490),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2490),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_API_AVAILABLE] = ACTIONS(2490),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_API_DEPRECATED] = ACTIONS(2490),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym___deprecated_msg] = ACTIONS(2490),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2490),
    [anon_sym_ATprotocol] = ACTIONS(2492),
    [anon_sym_ATinterface] = ACTIONS(2492),
    [anon_sym_ATimplementation] = ACTIONS(2492),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2492),
    [anon_sym_ATsynthesize] = ACTIONS(2492),
    [anon_sym_ATdynamic] = ACTIONS(2492),
    [anon_sym_ATtry] = ACTIONS(2492),
    [anon_sym___try] = ACTIONS(2490),
    [anon_sym_ATthrow] = ACTIONS(2492),
    [anon_sym_ATselector] = ACTIONS(2492),
    [anon_sym_ATavailable] = ACTIONS(2492),
    [anon_sym___builtin_available] = ACTIONS(2490),
    [anon_sym_va_arg] = ACTIONS(2490),
    [anon_sym_ATencode] = ACTIONS(2492),
    [anon_sym_ATsynchronized] = ACTIONS(2492),
    [anon_sym_BOOL] = ACTIONS(2490),
    [anon_sym_IMP] = ACTIONS(2490),
    [anon_sym_SEL] = ACTIONS(2490),
    [anon_sym_Class] = ACTIONS(2490),
    [anon_sym_id] = ACTIONS(2490),
  },
  [STATE(248)] = {
    [sym_identifier] = ACTIONS(2494),
    [aux_sym_preproc_include_token1] = ACTIONS(2494),
    [aux_sym_preproc_include_token2] = ACTIONS(2494),
    [aux_sym_preproc_def_token1] = ACTIONS(2494),
    [aux_sym_preproc_if_token1] = ACTIONS(2494),
    [aux_sym_preproc_if_token2] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2494),
    [aux_sym_preproc_else_token1] = ACTIONS(2494),
    [aux_sym_preproc_elif_token1] = ACTIONS(2494),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2494),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2494),
    [sym_preproc_directive] = ACTIONS(2494),
    [anon_sym_LPAREN2] = ACTIONS(2496),
    [anon_sym_BANG] = ACTIONS(2496),
    [anon_sym_TILDE] = ACTIONS(2496),
    [anon_sym_DASH] = ACTIONS(2494),
    [anon_sym_PLUS] = ACTIONS(2494),
    [anon_sym_STAR] = ACTIONS(2496),
    [anon_sym_CARET] = ACTIONS(2496),
    [anon_sym_AMP] = ACTIONS(2496),
    [anon_sym_SEMI] = ACTIONS(2496),
    [anon_sym___extension__] = ACTIONS(2494),
    [anon_sym_typedef] = ACTIONS(2494),
    [anon_sym_extern] = ACTIONS(2494),
    [anon_sym___attribute__] = ACTIONS(2494),
    [anon_sym___attribute] = ACTIONS(2494),
    [anon_sym_noreturn] = ACTIONS(2494),
    [anon_sym_LBRACK] = ACTIONS(2496),
    [anon_sym___declspec] = ACTIONS(2494),
    [anon_sym___cdecl] = ACTIONS(2494),
    [anon_sym___clrcall] = ACTIONS(2494),
    [anon_sym___stdcall] = ACTIONS(2494),
    [anon_sym___fastcall] = ACTIONS(2494),
    [anon_sym___thiscall] = ACTIONS(2494),
    [anon_sym___vectorcall] = ACTIONS(2494),
    [anon_sym_LBRACE] = ACTIONS(2496),
    [anon_sym_signed] = ACTIONS(2494),
    [anon_sym_unsigned] = ACTIONS(2494),
    [anon_sym_long] = ACTIONS(2494),
    [anon_sym_short] = ACTIONS(2494),
    [anon_sym_static] = ACTIONS(2494),
    [anon_sym_ATautoreleasepool] = ACTIONS(2496),
    [anon_sym_auto] = ACTIONS(2494),
    [anon_sym_register] = ACTIONS(2494),
    [anon_sym_inline] = ACTIONS(2494),
    [anon_sym___inline] = ACTIONS(2494),
    [anon_sym___inline__] = ACTIONS(2494),
    [anon_sym___forceinline] = ACTIONS(2494),
    [anon_sym_thread_local] = ACTIONS(2494),
    [anon_sym___thread] = ACTIONS(2494),
    [anon_sym_CG_EXTERN] = ACTIONS(2494),
    [anon_sym_CG_INLINE] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2494),
    [anon_sym_IBOutlet] = ACTIONS(2494),
    [anon_sym_IBInspectable] = ACTIONS(2494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2494),
    [anon_sym_NS_INLINE] = ACTIONS(2494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2494),
    [anon_sym_const] = ACTIONS(2494),
    [anon_sym_constexpr] = ACTIONS(2494),
    [anon_sym_volatile] = ACTIONS(2494),
    [anon_sym_restrict] = ACTIONS(2494),
    [anon_sym___restrict__] = ACTIONS(2494),
    [anon_sym__Atomic] = ACTIONS(2494),
    [anon_sym__Noreturn] = ACTIONS(2494),
    [anon_sym__Nonnull] = ACTIONS(2494),
    [anon_sym_nullable] = ACTIONS(2494),
    [anon_sym__Complex] = ACTIONS(2494),
    [anon_sym__Nullable] = ACTIONS(2494),
    [anon_sym__Nullable_result] = ACTIONS(2494),
    [anon_sym__Null_unspecified] = ACTIONS(2494),
    [anon_sym___autoreleasing] = ACTIONS(2494),
    [anon_sym___block] = ACTIONS(2494),
    [anon_sym___bridge] = ACTIONS(2494),
    [anon_sym___bridge_retained] = ACTIONS(2494),
    [anon_sym___bridge_transfer] = ACTIONS(2494),
    [anon_sym___complex] = ACTIONS(2494),
    [anon_sym___const] = ACTIONS(2494),
    [anon_sym___imag] = ACTIONS(2494),
    [anon_sym___kindof] = ACTIONS(2494),
    [anon_sym___nonnull] = ACTIONS(2494),
    [anon_sym___nullable] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2494),
    [anon_sym___real] = ACTIONS(2494),
    [anon_sym___strong] = ACTIONS(2494),
    [anon_sym___unsafe_unretained] = ACTIONS(2494),
    [anon_sym___unused] = ACTIONS(2494),
    [anon_sym___weak] = ACTIONS(2494),
    [anon_sym_alignas] = ACTIONS(2494),
    [anon_sym__Alignas] = ACTIONS(2494),
    [sym_primitive_type] = ACTIONS(2494),
    [anon_sym_enum] = ACTIONS(2494),
    [anon_sym_struct] = ACTIONS(2494),
    [anon_sym_union] = ACTIONS(2494),
    [anon_sym_if] = ACTIONS(2494),
    [anon_sym_else] = ACTIONS(2494),
    [anon_sym_switch] = ACTIONS(2494),
    [anon_sym_case] = ACTIONS(2494),
    [anon_sym_default] = ACTIONS(2494),
    [anon_sym_while] = ACTIONS(2494),
    [anon_sym_do] = ACTIONS(2494),
    [anon_sym_for] = ACTIONS(2494),
    [anon_sym_in] = ACTIONS(2494),
    [anon_sym_return] = ACTIONS(2494),
    [anon_sym_break] = ACTIONS(2494),
    [anon_sym_continue] = ACTIONS(2494),
    [anon_sym_goto] = ACTIONS(2494),
    [anon_sym_DASH_DASH] = ACTIONS(2496),
    [anon_sym_PLUS_PLUS] = ACTIONS(2496),
    [anon_sym_sizeof] = ACTIONS(2494),
    [anon_sym___alignof__] = ACTIONS(2494),
    [anon_sym___alignof] = ACTIONS(2494),
    [anon_sym__alignof] = ACTIONS(2494),
    [anon_sym_alignof] = ACTIONS(2494),
    [anon_sym__Alignof] = ACTIONS(2494),
    [anon_sym_offsetof] = ACTIONS(2494),
    [anon_sym__Generic] = ACTIONS(2494),
    [anon_sym_asm] = ACTIONS(2494),
    [anon_sym___asm__] = ACTIONS(2494),
    [anon_sym___asm] = ACTIONS(2494),
    [sym_number_literal] = ACTIONS(2496),
    [anon_sym_L_SQUOTE] = ACTIONS(2496),
    [anon_sym_u_SQUOTE] = ACTIONS(2496),
    [anon_sym_U_SQUOTE] = ACTIONS(2496),
    [anon_sym_u8_SQUOTE] = ACTIONS(2496),
    [anon_sym_SQUOTE] = ACTIONS(2496),
    [anon_sym_AT] = ACTIONS(2494),
    [anon_sym_DQUOTE] = ACTIONS(2496),
    [anon_sym_L_DQUOTE] = ACTIONS(2496),
    [anon_sym_u_DQUOTE] = ACTIONS(2496),
    [anon_sym_U_DQUOTE] = ACTIONS(2496),
    [anon_sym_u8_DQUOTE] = ACTIONS(2496),
    [sym_true] = ACTIONS(2494),
    [sym_false] = ACTIONS(2494),
    [anon_sym_NULL] = ACTIONS(2494),
    [anon_sym_nullptr] = ACTIONS(2494),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2494),
    [anon_sym___typeof] = ACTIONS(2494),
    [anon_sym_typeof] = ACTIONS(2494),
    [anon_sym_ATimport] = ACTIONS(2496),
    [aux_sym_preproc_undef_token1] = ACTIONS(2494),
    [anon_sym_POUND] = ACTIONS(2494),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_API_AVAILABLE] = ACTIONS(2494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_API_DEPRECATED] = ACTIONS(2494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym___deprecated_msg] = ACTIONS(2494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2494),
    [anon_sym_ATprotocol] = ACTIONS(2496),
    [anon_sym_ATinterface] = ACTIONS(2496),
    [anon_sym_ATimplementation] = ACTIONS(2496),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2496),
    [anon_sym_ATsynthesize] = ACTIONS(2496),
    [anon_sym_ATdynamic] = ACTIONS(2496),
    [anon_sym_ATtry] = ACTIONS(2496),
    [anon_sym___try] = ACTIONS(2494),
    [anon_sym_ATthrow] = ACTIONS(2496),
    [anon_sym_ATselector] = ACTIONS(2496),
    [anon_sym_ATavailable] = ACTIONS(2496),
    [anon_sym___builtin_available] = ACTIONS(2494),
    [anon_sym_va_arg] = ACTIONS(2494),
    [anon_sym_ATencode] = ACTIONS(2496),
    [anon_sym_ATsynchronized] = ACTIONS(2496),
    [anon_sym_BOOL] = ACTIONS(2494),
    [anon_sym_IMP] = ACTIONS(2494),
    [anon_sym_SEL] = ACTIONS(2494),
    [anon_sym_Class] = ACTIONS(2494),
    [anon_sym_id] = ACTIONS(2494),
  },
  [STATE(249)] = {
    [sym_identifier] = ACTIONS(2498),
    [aux_sym_preproc_include_token1] = ACTIONS(2498),
    [aux_sym_preproc_include_token2] = ACTIONS(2498),
    [aux_sym_preproc_def_token1] = ACTIONS(2498),
    [aux_sym_preproc_if_token1] = ACTIONS(2498),
    [aux_sym_preproc_if_token2] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2498),
    [aux_sym_preproc_else_token1] = ACTIONS(2498),
    [aux_sym_preproc_elif_token1] = ACTIONS(2498),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2498),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2498),
    [sym_preproc_directive] = ACTIONS(2498),
    [anon_sym_LPAREN2] = ACTIONS(2500),
    [anon_sym_BANG] = ACTIONS(2500),
    [anon_sym_TILDE] = ACTIONS(2500),
    [anon_sym_DASH] = ACTIONS(2498),
    [anon_sym_PLUS] = ACTIONS(2498),
    [anon_sym_STAR] = ACTIONS(2500),
    [anon_sym_CARET] = ACTIONS(2500),
    [anon_sym_AMP] = ACTIONS(2500),
    [anon_sym_SEMI] = ACTIONS(2500),
    [anon_sym___extension__] = ACTIONS(2498),
    [anon_sym_typedef] = ACTIONS(2498),
    [anon_sym_extern] = ACTIONS(2498),
    [anon_sym___attribute__] = ACTIONS(2498),
    [anon_sym___attribute] = ACTIONS(2498),
    [anon_sym_noreturn] = ACTIONS(2498),
    [anon_sym_LBRACK] = ACTIONS(2500),
    [anon_sym___declspec] = ACTIONS(2498),
    [anon_sym___cdecl] = ACTIONS(2498),
    [anon_sym___clrcall] = ACTIONS(2498),
    [anon_sym___stdcall] = ACTIONS(2498),
    [anon_sym___fastcall] = ACTIONS(2498),
    [anon_sym___thiscall] = ACTIONS(2498),
    [anon_sym___vectorcall] = ACTIONS(2498),
    [anon_sym_LBRACE] = ACTIONS(2500),
    [anon_sym_signed] = ACTIONS(2498),
    [anon_sym_unsigned] = ACTIONS(2498),
    [anon_sym_long] = ACTIONS(2498),
    [anon_sym_short] = ACTIONS(2498),
    [anon_sym_static] = ACTIONS(2498),
    [anon_sym_ATautoreleasepool] = ACTIONS(2500),
    [anon_sym_auto] = ACTIONS(2498),
    [anon_sym_register] = ACTIONS(2498),
    [anon_sym_inline] = ACTIONS(2498),
    [anon_sym___inline] = ACTIONS(2498),
    [anon_sym___inline__] = ACTIONS(2498),
    [anon_sym___forceinline] = ACTIONS(2498),
    [anon_sym_thread_local] = ACTIONS(2498),
    [anon_sym___thread] = ACTIONS(2498),
    [anon_sym_CG_EXTERN] = ACTIONS(2498),
    [anon_sym_CG_INLINE] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2498),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2498),
    [anon_sym_IBOutlet] = ACTIONS(2498),
    [anon_sym_IBInspectable] = ACTIONS(2498),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2498),
    [anon_sym_NS_INLINE] = ACTIONS(2498),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2498),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2498),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2498),
    [anon_sym_const] = ACTIONS(2498),
    [anon_sym_constexpr] = ACTIONS(2498),
    [anon_sym_volatile] = ACTIONS(2498),
    [anon_sym_restrict] = ACTIONS(2498),
    [anon_sym___restrict__] = ACTIONS(2498),
    [anon_sym__Atomic] = ACTIONS(2498),
    [anon_sym__Noreturn] = ACTIONS(2498),
    [anon_sym__Nonnull] = ACTIONS(2498),
    [anon_sym_nullable] = ACTIONS(2498),
    [anon_sym__Complex] = ACTIONS(2498),
    [anon_sym__Nullable] = ACTIONS(2498),
    [anon_sym__Nullable_result] = ACTIONS(2498),
    [anon_sym__Null_unspecified] = ACTIONS(2498),
    [anon_sym___autoreleasing] = ACTIONS(2498),
    [anon_sym___block] = ACTIONS(2498),
    [anon_sym___bridge] = ACTIONS(2498),
    [anon_sym___bridge_retained] = ACTIONS(2498),
    [anon_sym___bridge_transfer] = ACTIONS(2498),
    [anon_sym___complex] = ACTIONS(2498),
    [anon_sym___const] = ACTIONS(2498),
    [anon_sym___imag] = ACTIONS(2498),
    [anon_sym___kindof] = ACTIONS(2498),
    [anon_sym___nonnull] = ACTIONS(2498),
    [anon_sym___nullable] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2498),
    [anon_sym___real] = ACTIONS(2498),
    [anon_sym___strong] = ACTIONS(2498),
    [anon_sym___unsafe_unretained] = ACTIONS(2498),
    [anon_sym___unused] = ACTIONS(2498),
    [anon_sym___weak] = ACTIONS(2498),
    [anon_sym_alignas] = ACTIONS(2498),
    [anon_sym__Alignas] = ACTIONS(2498),
    [sym_primitive_type] = ACTIONS(2498),
    [anon_sym_enum] = ACTIONS(2498),
    [anon_sym_struct] = ACTIONS(2498),
    [anon_sym_union] = ACTIONS(2498),
    [anon_sym_if] = ACTIONS(2498),
    [anon_sym_else] = ACTIONS(2498),
    [anon_sym_switch] = ACTIONS(2498),
    [anon_sym_case] = ACTIONS(2498),
    [anon_sym_default] = ACTIONS(2498),
    [anon_sym_while] = ACTIONS(2498),
    [anon_sym_do] = ACTIONS(2498),
    [anon_sym_for] = ACTIONS(2498),
    [anon_sym_in] = ACTIONS(2498),
    [anon_sym_return] = ACTIONS(2498),
    [anon_sym_break] = ACTIONS(2498),
    [anon_sym_continue] = ACTIONS(2498),
    [anon_sym_goto] = ACTIONS(2498),
    [anon_sym_DASH_DASH] = ACTIONS(2500),
    [anon_sym_PLUS_PLUS] = ACTIONS(2500),
    [anon_sym_sizeof] = ACTIONS(2498),
    [anon_sym___alignof__] = ACTIONS(2498),
    [anon_sym___alignof] = ACTIONS(2498),
    [anon_sym__alignof] = ACTIONS(2498),
    [anon_sym_alignof] = ACTIONS(2498),
    [anon_sym__Alignof] = ACTIONS(2498),
    [anon_sym_offsetof] = ACTIONS(2498),
    [anon_sym__Generic] = ACTIONS(2498),
    [anon_sym_asm] = ACTIONS(2498),
    [anon_sym___asm__] = ACTIONS(2498),
    [anon_sym___asm] = ACTIONS(2498),
    [sym_number_literal] = ACTIONS(2500),
    [anon_sym_L_SQUOTE] = ACTIONS(2500),
    [anon_sym_u_SQUOTE] = ACTIONS(2500),
    [anon_sym_U_SQUOTE] = ACTIONS(2500),
    [anon_sym_u8_SQUOTE] = ACTIONS(2500),
    [anon_sym_SQUOTE] = ACTIONS(2500),
    [anon_sym_AT] = ACTIONS(2498),
    [anon_sym_DQUOTE] = ACTIONS(2500),
    [anon_sym_L_DQUOTE] = ACTIONS(2500),
    [anon_sym_u_DQUOTE] = ACTIONS(2500),
    [anon_sym_U_DQUOTE] = ACTIONS(2500),
    [anon_sym_u8_DQUOTE] = ACTIONS(2500),
    [sym_true] = ACTIONS(2498),
    [sym_false] = ACTIONS(2498),
    [anon_sym_NULL] = ACTIONS(2498),
    [anon_sym_nullptr] = ACTIONS(2498),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2498),
    [anon_sym___typeof] = ACTIONS(2498),
    [anon_sym_typeof] = ACTIONS(2498),
    [anon_sym_ATimport] = ACTIONS(2500),
    [aux_sym_preproc_undef_token1] = ACTIONS(2498),
    [anon_sym_POUND] = ACTIONS(2498),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2498),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2498),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2498),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_API_AVAILABLE] = ACTIONS(2498),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_API_DEPRECATED] = ACTIONS(2498),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym___deprecated_msg] = ACTIONS(2498),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2498),
    [anon_sym_ATprotocol] = ACTIONS(2500),
    [anon_sym_ATinterface] = ACTIONS(2500),
    [anon_sym_ATimplementation] = ACTIONS(2500),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2500),
    [anon_sym_ATsynthesize] = ACTIONS(2500),
    [anon_sym_ATdynamic] = ACTIONS(2500),
    [anon_sym_ATtry] = ACTIONS(2500),
    [anon_sym___try] = ACTIONS(2498),
    [anon_sym_ATthrow] = ACTIONS(2500),
    [anon_sym_ATselector] = ACTIONS(2500),
    [anon_sym_ATavailable] = ACTIONS(2500),
    [anon_sym___builtin_available] = ACTIONS(2498),
    [anon_sym_va_arg] = ACTIONS(2498),
    [anon_sym_ATencode] = ACTIONS(2500),
    [anon_sym_ATsynchronized] = ACTIONS(2500),
    [anon_sym_BOOL] = ACTIONS(2498),
    [anon_sym_IMP] = ACTIONS(2498),
    [anon_sym_SEL] = ACTIONS(2498),
    [anon_sym_Class] = ACTIONS(2498),
    [anon_sym_id] = ACTIONS(2498),
  },
  [STATE(250)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token2] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [aux_sym_preproc_else_token1] = ACTIONS(2350),
    [aux_sym_preproc_elif_token1] = ACTIONS(2350),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATsynthesize] = ACTIONS(2352),
    [anon_sym_ATdynamic] = ACTIONS(2352),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(251)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token2] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [aux_sym_preproc_else_token1] = ACTIONS(2438),
    [aux_sym_preproc_elif_token1] = ACTIONS(2438),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(252)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(253)] = {
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token2] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [aux_sym_preproc_else_token1] = ACTIONS(2306),
    [aux_sym_preproc_elif_token1] = ACTIONS(2306),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATsynthesize] = ACTIONS(2308),
    [anon_sym_ATdynamic] = ACTIONS(2308),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(254)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token2] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [aux_sym_preproc_else_token1] = ACTIONS(2442),
    [aux_sym_preproc_elif_token1] = ACTIONS(2442),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(255)] = {
    [sym_identifier] = ACTIONS(2502),
    [aux_sym_preproc_include_token1] = ACTIONS(2502),
    [aux_sym_preproc_include_token2] = ACTIONS(2502),
    [aux_sym_preproc_def_token1] = ACTIONS(2502),
    [aux_sym_preproc_if_token1] = ACTIONS(2502),
    [aux_sym_preproc_if_token2] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2502),
    [aux_sym_preproc_else_token1] = ACTIONS(2502),
    [aux_sym_preproc_elif_token1] = ACTIONS(2502),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2502),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2502),
    [sym_preproc_directive] = ACTIONS(2502),
    [anon_sym_LPAREN2] = ACTIONS(2504),
    [anon_sym_BANG] = ACTIONS(2504),
    [anon_sym_TILDE] = ACTIONS(2504),
    [anon_sym_DASH] = ACTIONS(2502),
    [anon_sym_PLUS] = ACTIONS(2502),
    [anon_sym_STAR] = ACTIONS(2504),
    [anon_sym_CARET] = ACTIONS(2504),
    [anon_sym_AMP] = ACTIONS(2504),
    [anon_sym_SEMI] = ACTIONS(2504),
    [anon_sym___extension__] = ACTIONS(2502),
    [anon_sym_typedef] = ACTIONS(2502),
    [anon_sym_extern] = ACTIONS(2502),
    [anon_sym___attribute__] = ACTIONS(2502),
    [anon_sym___attribute] = ACTIONS(2502),
    [anon_sym_noreturn] = ACTIONS(2502),
    [anon_sym_LBRACK] = ACTIONS(2504),
    [anon_sym___declspec] = ACTIONS(2502),
    [anon_sym___cdecl] = ACTIONS(2502),
    [anon_sym___clrcall] = ACTIONS(2502),
    [anon_sym___stdcall] = ACTIONS(2502),
    [anon_sym___fastcall] = ACTIONS(2502),
    [anon_sym___thiscall] = ACTIONS(2502),
    [anon_sym___vectorcall] = ACTIONS(2502),
    [anon_sym_LBRACE] = ACTIONS(2504),
    [anon_sym_signed] = ACTIONS(2502),
    [anon_sym_unsigned] = ACTIONS(2502),
    [anon_sym_long] = ACTIONS(2502),
    [anon_sym_short] = ACTIONS(2502),
    [anon_sym_static] = ACTIONS(2502),
    [anon_sym_ATautoreleasepool] = ACTIONS(2504),
    [anon_sym_auto] = ACTIONS(2502),
    [anon_sym_register] = ACTIONS(2502),
    [anon_sym_inline] = ACTIONS(2502),
    [anon_sym___inline] = ACTIONS(2502),
    [anon_sym___inline__] = ACTIONS(2502),
    [anon_sym___forceinline] = ACTIONS(2502),
    [anon_sym_thread_local] = ACTIONS(2502),
    [anon_sym___thread] = ACTIONS(2502),
    [anon_sym_CG_EXTERN] = ACTIONS(2502),
    [anon_sym_CG_INLINE] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2502),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2502),
    [anon_sym_IBOutlet] = ACTIONS(2502),
    [anon_sym_IBInspectable] = ACTIONS(2502),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2502),
    [anon_sym_NS_INLINE] = ACTIONS(2502),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2502),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2502),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2502),
    [anon_sym_const] = ACTIONS(2502),
    [anon_sym_constexpr] = ACTIONS(2502),
    [anon_sym_volatile] = ACTIONS(2502),
    [anon_sym_restrict] = ACTIONS(2502),
    [anon_sym___restrict__] = ACTIONS(2502),
    [anon_sym__Atomic] = ACTIONS(2502),
    [anon_sym__Noreturn] = ACTIONS(2502),
    [anon_sym__Nonnull] = ACTIONS(2502),
    [anon_sym_nullable] = ACTIONS(2502),
    [anon_sym__Complex] = ACTIONS(2502),
    [anon_sym__Nullable] = ACTIONS(2502),
    [anon_sym__Nullable_result] = ACTIONS(2502),
    [anon_sym__Null_unspecified] = ACTIONS(2502),
    [anon_sym___autoreleasing] = ACTIONS(2502),
    [anon_sym___block] = ACTIONS(2502),
    [anon_sym___bridge] = ACTIONS(2502),
    [anon_sym___bridge_retained] = ACTIONS(2502),
    [anon_sym___bridge_transfer] = ACTIONS(2502),
    [anon_sym___complex] = ACTIONS(2502),
    [anon_sym___const] = ACTIONS(2502),
    [anon_sym___imag] = ACTIONS(2502),
    [anon_sym___kindof] = ACTIONS(2502),
    [anon_sym___nonnull] = ACTIONS(2502),
    [anon_sym___nullable] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2502),
    [anon_sym___real] = ACTIONS(2502),
    [anon_sym___strong] = ACTIONS(2502),
    [anon_sym___unsafe_unretained] = ACTIONS(2502),
    [anon_sym___unused] = ACTIONS(2502),
    [anon_sym___weak] = ACTIONS(2502),
    [anon_sym_alignas] = ACTIONS(2502),
    [anon_sym__Alignas] = ACTIONS(2502),
    [sym_primitive_type] = ACTIONS(2502),
    [anon_sym_enum] = ACTIONS(2502),
    [anon_sym_struct] = ACTIONS(2502),
    [anon_sym_union] = ACTIONS(2502),
    [anon_sym_if] = ACTIONS(2502),
    [anon_sym_else] = ACTIONS(2502),
    [anon_sym_switch] = ACTIONS(2502),
    [anon_sym_case] = ACTIONS(2502),
    [anon_sym_default] = ACTIONS(2502),
    [anon_sym_while] = ACTIONS(2502),
    [anon_sym_do] = ACTIONS(2502),
    [anon_sym_for] = ACTIONS(2502),
    [anon_sym_in] = ACTIONS(2502),
    [anon_sym_return] = ACTIONS(2502),
    [anon_sym_break] = ACTIONS(2502),
    [anon_sym_continue] = ACTIONS(2502),
    [anon_sym_goto] = ACTIONS(2502),
    [anon_sym_DASH_DASH] = ACTIONS(2504),
    [anon_sym_PLUS_PLUS] = ACTIONS(2504),
    [anon_sym_sizeof] = ACTIONS(2502),
    [anon_sym___alignof__] = ACTIONS(2502),
    [anon_sym___alignof] = ACTIONS(2502),
    [anon_sym__alignof] = ACTIONS(2502),
    [anon_sym_alignof] = ACTIONS(2502),
    [anon_sym__Alignof] = ACTIONS(2502),
    [anon_sym_offsetof] = ACTIONS(2502),
    [anon_sym__Generic] = ACTIONS(2502),
    [anon_sym_asm] = ACTIONS(2502),
    [anon_sym___asm__] = ACTIONS(2502),
    [anon_sym___asm] = ACTIONS(2502),
    [sym_number_literal] = ACTIONS(2504),
    [anon_sym_L_SQUOTE] = ACTIONS(2504),
    [anon_sym_u_SQUOTE] = ACTIONS(2504),
    [anon_sym_U_SQUOTE] = ACTIONS(2504),
    [anon_sym_u8_SQUOTE] = ACTIONS(2504),
    [anon_sym_SQUOTE] = ACTIONS(2504),
    [anon_sym_AT] = ACTIONS(2502),
    [anon_sym_DQUOTE] = ACTIONS(2504),
    [anon_sym_L_DQUOTE] = ACTIONS(2504),
    [anon_sym_u_DQUOTE] = ACTIONS(2504),
    [anon_sym_U_DQUOTE] = ACTIONS(2504),
    [anon_sym_u8_DQUOTE] = ACTIONS(2504),
    [sym_true] = ACTIONS(2502),
    [sym_false] = ACTIONS(2502),
    [anon_sym_NULL] = ACTIONS(2502),
    [anon_sym_nullptr] = ACTIONS(2502),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2502),
    [anon_sym___typeof] = ACTIONS(2502),
    [anon_sym_typeof] = ACTIONS(2502),
    [anon_sym_ATimport] = ACTIONS(2504),
    [aux_sym_preproc_undef_token1] = ACTIONS(2502),
    [anon_sym_POUND] = ACTIONS(2502),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2502),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2502),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2502),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_API_AVAILABLE] = ACTIONS(2502),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_API_DEPRECATED] = ACTIONS(2502),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym___deprecated_msg] = ACTIONS(2502),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2502),
    [anon_sym_ATprotocol] = ACTIONS(2504),
    [anon_sym_ATinterface] = ACTIONS(2504),
    [anon_sym_ATimplementation] = ACTIONS(2504),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2504),
    [anon_sym_ATsynthesize] = ACTIONS(2504),
    [anon_sym_ATdynamic] = ACTIONS(2504),
    [anon_sym_ATtry] = ACTIONS(2504),
    [anon_sym___try] = ACTIONS(2502),
    [anon_sym_ATthrow] = ACTIONS(2504),
    [anon_sym_ATselector] = ACTIONS(2504),
    [anon_sym_ATavailable] = ACTIONS(2504),
    [anon_sym___builtin_available] = ACTIONS(2502),
    [anon_sym_va_arg] = ACTIONS(2502),
    [anon_sym_ATencode] = ACTIONS(2504),
    [anon_sym_ATsynchronized] = ACTIONS(2504),
    [anon_sym_BOOL] = ACTIONS(2502),
    [anon_sym_IMP] = ACTIONS(2502),
    [anon_sym_SEL] = ACTIONS(2502),
    [anon_sym_Class] = ACTIONS(2502),
    [anon_sym_id] = ACTIONS(2502),
  },
  [STATE(256)] = {
    [sym_identifier] = ACTIONS(2506),
    [aux_sym_preproc_include_token1] = ACTIONS(2506),
    [aux_sym_preproc_include_token2] = ACTIONS(2506),
    [aux_sym_preproc_def_token1] = ACTIONS(2506),
    [aux_sym_preproc_if_token1] = ACTIONS(2506),
    [aux_sym_preproc_if_token2] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2506),
    [aux_sym_preproc_else_token1] = ACTIONS(2506),
    [aux_sym_preproc_elif_token1] = ACTIONS(2506),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2506),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2506),
    [sym_preproc_directive] = ACTIONS(2506),
    [anon_sym_LPAREN2] = ACTIONS(2508),
    [anon_sym_BANG] = ACTIONS(2508),
    [anon_sym_TILDE] = ACTIONS(2508),
    [anon_sym_DASH] = ACTIONS(2506),
    [anon_sym_PLUS] = ACTIONS(2506),
    [anon_sym_STAR] = ACTIONS(2508),
    [anon_sym_CARET] = ACTIONS(2508),
    [anon_sym_AMP] = ACTIONS(2508),
    [anon_sym_SEMI] = ACTIONS(2508),
    [anon_sym___extension__] = ACTIONS(2506),
    [anon_sym_typedef] = ACTIONS(2506),
    [anon_sym_extern] = ACTIONS(2506),
    [anon_sym___attribute__] = ACTIONS(2506),
    [anon_sym___attribute] = ACTIONS(2506),
    [anon_sym_noreturn] = ACTIONS(2506),
    [anon_sym_LBRACK] = ACTIONS(2508),
    [anon_sym___declspec] = ACTIONS(2506),
    [anon_sym___cdecl] = ACTIONS(2506),
    [anon_sym___clrcall] = ACTIONS(2506),
    [anon_sym___stdcall] = ACTIONS(2506),
    [anon_sym___fastcall] = ACTIONS(2506),
    [anon_sym___thiscall] = ACTIONS(2506),
    [anon_sym___vectorcall] = ACTIONS(2506),
    [anon_sym_LBRACE] = ACTIONS(2508),
    [anon_sym_signed] = ACTIONS(2506),
    [anon_sym_unsigned] = ACTIONS(2506),
    [anon_sym_long] = ACTIONS(2506),
    [anon_sym_short] = ACTIONS(2506),
    [anon_sym_static] = ACTIONS(2506),
    [anon_sym_ATautoreleasepool] = ACTIONS(2508),
    [anon_sym_auto] = ACTIONS(2506),
    [anon_sym_register] = ACTIONS(2506),
    [anon_sym_inline] = ACTIONS(2506),
    [anon_sym___inline] = ACTIONS(2506),
    [anon_sym___inline__] = ACTIONS(2506),
    [anon_sym___forceinline] = ACTIONS(2506),
    [anon_sym_thread_local] = ACTIONS(2506),
    [anon_sym___thread] = ACTIONS(2506),
    [anon_sym_CG_EXTERN] = ACTIONS(2506),
    [anon_sym_CG_INLINE] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2506),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2506),
    [anon_sym_IBOutlet] = ACTIONS(2506),
    [anon_sym_IBInspectable] = ACTIONS(2506),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2506),
    [anon_sym_NS_INLINE] = ACTIONS(2506),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2506),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2506),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2506),
    [anon_sym_const] = ACTIONS(2506),
    [anon_sym_constexpr] = ACTIONS(2506),
    [anon_sym_volatile] = ACTIONS(2506),
    [anon_sym_restrict] = ACTIONS(2506),
    [anon_sym___restrict__] = ACTIONS(2506),
    [anon_sym__Atomic] = ACTIONS(2506),
    [anon_sym__Noreturn] = ACTIONS(2506),
    [anon_sym__Nonnull] = ACTIONS(2506),
    [anon_sym_nullable] = ACTIONS(2506),
    [anon_sym__Complex] = ACTIONS(2506),
    [anon_sym__Nullable] = ACTIONS(2506),
    [anon_sym__Nullable_result] = ACTIONS(2506),
    [anon_sym__Null_unspecified] = ACTIONS(2506),
    [anon_sym___autoreleasing] = ACTIONS(2506),
    [anon_sym___block] = ACTIONS(2506),
    [anon_sym___bridge] = ACTIONS(2506),
    [anon_sym___bridge_retained] = ACTIONS(2506),
    [anon_sym___bridge_transfer] = ACTIONS(2506),
    [anon_sym___complex] = ACTIONS(2506),
    [anon_sym___const] = ACTIONS(2506),
    [anon_sym___imag] = ACTIONS(2506),
    [anon_sym___kindof] = ACTIONS(2506),
    [anon_sym___nonnull] = ACTIONS(2506),
    [anon_sym___nullable] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2506),
    [anon_sym___real] = ACTIONS(2506),
    [anon_sym___strong] = ACTIONS(2506),
    [anon_sym___unsafe_unretained] = ACTIONS(2506),
    [anon_sym___unused] = ACTIONS(2506),
    [anon_sym___weak] = ACTIONS(2506),
    [anon_sym_alignas] = ACTIONS(2506),
    [anon_sym__Alignas] = ACTIONS(2506),
    [sym_primitive_type] = ACTIONS(2506),
    [anon_sym_enum] = ACTIONS(2506),
    [anon_sym_struct] = ACTIONS(2506),
    [anon_sym_union] = ACTIONS(2506),
    [anon_sym_if] = ACTIONS(2506),
    [anon_sym_else] = ACTIONS(2506),
    [anon_sym_switch] = ACTIONS(2506),
    [anon_sym_case] = ACTIONS(2506),
    [anon_sym_default] = ACTIONS(2506),
    [anon_sym_while] = ACTIONS(2506),
    [anon_sym_do] = ACTIONS(2506),
    [anon_sym_for] = ACTIONS(2506),
    [anon_sym_in] = ACTIONS(2506),
    [anon_sym_return] = ACTIONS(2506),
    [anon_sym_break] = ACTIONS(2506),
    [anon_sym_continue] = ACTIONS(2506),
    [anon_sym_goto] = ACTIONS(2506),
    [anon_sym_DASH_DASH] = ACTIONS(2508),
    [anon_sym_PLUS_PLUS] = ACTIONS(2508),
    [anon_sym_sizeof] = ACTIONS(2506),
    [anon_sym___alignof__] = ACTIONS(2506),
    [anon_sym___alignof] = ACTIONS(2506),
    [anon_sym__alignof] = ACTIONS(2506),
    [anon_sym_alignof] = ACTIONS(2506),
    [anon_sym__Alignof] = ACTIONS(2506),
    [anon_sym_offsetof] = ACTIONS(2506),
    [anon_sym__Generic] = ACTIONS(2506),
    [anon_sym_asm] = ACTIONS(2506),
    [anon_sym___asm__] = ACTIONS(2506),
    [anon_sym___asm] = ACTIONS(2506),
    [sym_number_literal] = ACTIONS(2508),
    [anon_sym_L_SQUOTE] = ACTIONS(2508),
    [anon_sym_u_SQUOTE] = ACTIONS(2508),
    [anon_sym_U_SQUOTE] = ACTIONS(2508),
    [anon_sym_u8_SQUOTE] = ACTIONS(2508),
    [anon_sym_SQUOTE] = ACTIONS(2508),
    [anon_sym_AT] = ACTIONS(2506),
    [anon_sym_DQUOTE] = ACTIONS(2508),
    [anon_sym_L_DQUOTE] = ACTIONS(2508),
    [anon_sym_u_DQUOTE] = ACTIONS(2508),
    [anon_sym_U_DQUOTE] = ACTIONS(2508),
    [anon_sym_u8_DQUOTE] = ACTIONS(2508),
    [sym_true] = ACTIONS(2506),
    [sym_false] = ACTIONS(2506),
    [anon_sym_NULL] = ACTIONS(2506),
    [anon_sym_nullptr] = ACTIONS(2506),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2506),
    [anon_sym___typeof] = ACTIONS(2506),
    [anon_sym_typeof] = ACTIONS(2506),
    [anon_sym_ATimport] = ACTIONS(2508),
    [aux_sym_preproc_undef_token1] = ACTIONS(2506),
    [anon_sym_POUND] = ACTIONS(2506),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2506),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2506),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2506),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_API_AVAILABLE] = ACTIONS(2506),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_API_DEPRECATED] = ACTIONS(2506),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym___deprecated_msg] = ACTIONS(2506),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2506),
    [anon_sym_ATprotocol] = ACTIONS(2508),
    [anon_sym_ATinterface] = ACTIONS(2508),
    [anon_sym_ATimplementation] = ACTIONS(2508),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2508),
    [anon_sym_ATsynthesize] = ACTIONS(2508),
    [anon_sym_ATdynamic] = ACTIONS(2508),
    [anon_sym_ATtry] = ACTIONS(2508),
    [anon_sym___try] = ACTIONS(2506),
    [anon_sym_ATthrow] = ACTIONS(2508),
    [anon_sym_ATselector] = ACTIONS(2508),
    [anon_sym_ATavailable] = ACTIONS(2508),
    [anon_sym___builtin_available] = ACTIONS(2506),
    [anon_sym_va_arg] = ACTIONS(2506),
    [anon_sym_ATencode] = ACTIONS(2508),
    [anon_sym_ATsynchronized] = ACTIONS(2508),
    [anon_sym_BOOL] = ACTIONS(2506),
    [anon_sym_IMP] = ACTIONS(2506),
    [anon_sym_SEL] = ACTIONS(2506),
    [anon_sym_Class] = ACTIONS(2506),
    [anon_sym_id] = ACTIONS(2506),
  },
  [STATE(257)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [aux_sym_preproc_else_token1] = ACTIONS(2510),
    [aux_sym_preproc_elif_token1] = ACTIONS(2510),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(258)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token2] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [aux_sym_preproc_else_token1] = ACTIONS(2474),
    [aux_sym_preproc_elif_token1] = ACTIONS(2474),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(259)] = {
    [sym_identifier] = ACTIONS(2514),
    [aux_sym_preproc_include_token1] = ACTIONS(2514),
    [aux_sym_preproc_include_token2] = ACTIONS(2514),
    [aux_sym_preproc_def_token1] = ACTIONS(2514),
    [aux_sym_preproc_if_token1] = ACTIONS(2514),
    [aux_sym_preproc_if_token2] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2514),
    [aux_sym_preproc_else_token1] = ACTIONS(2514),
    [aux_sym_preproc_elif_token1] = ACTIONS(2514),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2514),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2514),
    [sym_preproc_directive] = ACTIONS(2514),
    [anon_sym_LPAREN2] = ACTIONS(2516),
    [anon_sym_BANG] = ACTIONS(2516),
    [anon_sym_TILDE] = ACTIONS(2516),
    [anon_sym_DASH] = ACTIONS(2514),
    [anon_sym_PLUS] = ACTIONS(2514),
    [anon_sym_STAR] = ACTIONS(2516),
    [anon_sym_CARET] = ACTIONS(2516),
    [anon_sym_AMP] = ACTIONS(2516),
    [anon_sym_SEMI] = ACTIONS(2516),
    [anon_sym___extension__] = ACTIONS(2514),
    [anon_sym_typedef] = ACTIONS(2514),
    [anon_sym_extern] = ACTIONS(2514),
    [anon_sym___attribute__] = ACTIONS(2514),
    [anon_sym___attribute] = ACTIONS(2514),
    [anon_sym_noreturn] = ACTIONS(2514),
    [anon_sym_LBRACK] = ACTIONS(2516),
    [anon_sym___declspec] = ACTIONS(2514),
    [anon_sym___cdecl] = ACTIONS(2514),
    [anon_sym___clrcall] = ACTIONS(2514),
    [anon_sym___stdcall] = ACTIONS(2514),
    [anon_sym___fastcall] = ACTIONS(2514),
    [anon_sym___thiscall] = ACTIONS(2514),
    [anon_sym___vectorcall] = ACTIONS(2514),
    [anon_sym_LBRACE] = ACTIONS(2516),
    [anon_sym_signed] = ACTIONS(2514),
    [anon_sym_unsigned] = ACTIONS(2514),
    [anon_sym_long] = ACTIONS(2514),
    [anon_sym_short] = ACTIONS(2514),
    [anon_sym_static] = ACTIONS(2514),
    [anon_sym_ATautoreleasepool] = ACTIONS(2516),
    [anon_sym_auto] = ACTIONS(2514),
    [anon_sym_register] = ACTIONS(2514),
    [anon_sym_inline] = ACTIONS(2514),
    [anon_sym___inline] = ACTIONS(2514),
    [anon_sym___inline__] = ACTIONS(2514),
    [anon_sym___forceinline] = ACTIONS(2514),
    [anon_sym_thread_local] = ACTIONS(2514),
    [anon_sym___thread] = ACTIONS(2514),
    [anon_sym_CG_EXTERN] = ACTIONS(2514),
    [anon_sym_CG_INLINE] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2514),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2514),
    [anon_sym_IBOutlet] = ACTIONS(2514),
    [anon_sym_IBInspectable] = ACTIONS(2514),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2514),
    [anon_sym_NS_INLINE] = ACTIONS(2514),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2514),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2514),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2514),
    [anon_sym_const] = ACTIONS(2514),
    [anon_sym_constexpr] = ACTIONS(2514),
    [anon_sym_volatile] = ACTIONS(2514),
    [anon_sym_restrict] = ACTIONS(2514),
    [anon_sym___restrict__] = ACTIONS(2514),
    [anon_sym__Atomic] = ACTIONS(2514),
    [anon_sym__Noreturn] = ACTIONS(2514),
    [anon_sym__Nonnull] = ACTIONS(2514),
    [anon_sym_nullable] = ACTIONS(2514),
    [anon_sym__Complex] = ACTIONS(2514),
    [anon_sym__Nullable] = ACTIONS(2514),
    [anon_sym__Nullable_result] = ACTIONS(2514),
    [anon_sym__Null_unspecified] = ACTIONS(2514),
    [anon_sym___autoreleasing] = ACTIONS(2514),
    [anon_sym___block] = ACTIONS(2514),
    [anon_sym___bridge] = ACTIONS(2514),
    [anon_sym___bridge_retained] = ACTIONS(2514),
    [anon_sym___bridge_transfer] = ACTIONS(2514),
    [anon_sym___complex] = ACTIONS(2514),
    [anon_sym___const] = ACTIONS(2514),
    [anon_sym___imag] = ACTIONS(2514),
    [anon_sym___kindof] = ACTIONS(2514),
    [anon_sym___nonnull] = ACTIONS(2514),
    [anon_sym___nullable] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2514),
    [anon_sym___real] = ACTIONS(2514),
    [anon_sym___strong] = ACTIONS(2514),
    [anon_sym___unsafe_unretained] = ACTIONS(2514),
    [anon_sym___unused] = ACTIONS(2514),
    [anon_sym___weak] = ACTIONS(2514),
    [anon_sym_alignas] = ACTIONS(2514),
    [anon_sym__Alignas] = ACTIONS(2514),
    [sym_primitive_type] = ACTIONS(2514),
    [anon_sym_enum] = ACTIONS(2514),
    [anon_sym_struct] = ACTIONS(2514),
    [anon_sym_union] = ACTIONS(2514),
    [anon_sym_if] = ACTIONS(2514),
    [anon_sym_else] = ACTIONS(2514),
    [anon_sym_switch] = ACTIONS(2514),
    [anon_sym_case] = ACTIONS(2514),
    [anon_sym_default] = ACTIONS(2514),
    [anon_sym_while] = ACTIONS(2514),
    [anon_sym_do] = ACTIONS(2514),
    [anon_sym_for] = ACTIONS(2514),
    [anon_sym_in] = ACTIONS(2514),
    [anon_sym_return] = ACTIONS(2514),
    [anon_sym_break] = ACTIONS(2514),
    [anon_sym_continue] = ACTIONS(2514),
    [anon_sym_goto] = ACTIONS(2514),
    [anon_sym_DASH_DASH] = ACTIONS(2516),
    [anon_sym_PLUS_PLUS] = ACTIONS(2516),
    [anon_sym_sizeof] = ACTIONS(2514),
    [anon_sym___alignof__] = ACTIONS(2514),
    [anon_sym___alignof] = ACTIONS(2514),
    [anon_sym__alignof] = ACTIONS(2514),
    [anon_sym_alignof] = ACTIONS(2514),
    [anon_sym__Alignof] = ACTIONS(2514),
    [anon_sym_offsetof] = ACTIONS(2514),
    [anon_sym__Generic] = ACTIONS(2514),
    [anon_sym_asm] = ACTIONS(2514),
    [anon_sym___asm__] = ACTIONS(2514),
    [anon_sym___asm] = ACTIONS(2514),
    [sym_number_literal] = ACTIONS(2516),
    [anon_sym_L_SQUOTE] = ACTIONS(2516),
    [anon_sym_u_SQUOTE] = ACTIONS(2516),
    [anon_sym_U_SQUOTE] = ACTIONS(2516),
    [anon_sym_u8_SQUOTE] = ACTIONS(2516),
    [anon_sym_SQUOTE] = ACTIONS(2516),
    [anon_sym_AT] = ACTIONS(2514),
    [anon_sym_DQUOTE] = ACTIONS(2516),
    [anon_sym_L_DQUOTE] = ACTIONS(2516),
    [anon_sym_u_DQUOTE] = ACTIONS(2516),
    [anon_sym_U_DQUOTE] = ACTIONS(2516),
    [anon_sym_u8_DQUOTE] = ACTIONS(2516),
    [sym_true] = ACTIONS(2514),
    [sym_false] = ACTIONS(2514),
    [anon_sym_NULL] = ACTIONS(2514),
    [anon_sym_nullptr] = ACTIONS(2514),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2514),
    [anon_sym___typeof] = ACTIONS(2514),
    [anon_sym_typeof] = ACTIONS(2514),
    [anon_sym_ATimport] = ACTIONS(2516),
    [aux_sym_preproc_undef_token1] = ACTIONS(2514),
    [anon_sym_POUND] = ACTIONS(2514),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2514),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2514),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2514),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_API_AVAILABLE] = ACTIONS(2514),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_API_DEPRECATED] = ACTIONS(2514),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym___deprecated_msg] = ACTIONS(2514),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2514),
    [anon_sym_ATprotocol] = ACTIONS(2516),
    [anon_sym_ATinterface] = ACTIONS(2516),
    [anon_sym_ATimplementation] = ACTIONS(2516),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2516),
    [anon_sym_ATsynthesize] = ACTIONS(2516),
    [anon_sym_ATdynamic] = ACTIONS(2516),
    [anon_sym_ATtry] = ACTIONS(2516),
    [anon_sym___try] = ACTIONS(2514),
    [anon_sym_ATthrow] = ACTIONS(2516),
    [anon_sym_ATselector] = ACTIONS(2516),
    [anon_sym_ATavailable] = ACTIONS(2516),
    [anon_sym___builtin_available] = ACTIONS(2514),
    [anon_sym_va_arg] = ACTIONS(2514),
    [anon_sym_ATencode] = ACTIONS(2516),
    [anon_sym_ATsynchronized] = ACTIONS(2516),
    [anon_sym_BOOL] = ACTIONS(2514),
    [anon_sym_IMP] = ACTIONS(2514),
    [anon_sym_SEL] = ACTIONS(2514),
    [anon_sym_Class] = ACTIONS(2514),
    [anon_sym_id] = ACTIONS(2514),
  },
  [STATE(260)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [aux_sym_preproc_else_token1] = ACTIONS(2510),
    [aux_sym_preproc_elif_token1] = ACTIONS(2510),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(261)] = {
    [sym_identifier] = ACTIONS(2518),
    [aux_sym_preproc_include_token1] = ACTIONS(2518),
    [aux_sym_preproc_include_token2] = ACTIONS(2518),
    [aux_sym_preproc_def_token1] = ACTIONS(2518),
    [aux_sym_preproc_if_token1] = ACTIONS(2518),
    [aux_sym_preproc_if_token2] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2518),
    [aux_sym_preproc_else_token1] = ACTIONS(2518),
    [aux_sym_preproc_elif_token1] = ACTIONS(2518),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2518),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2518),
    [sym_preproc_directive] = ACTIONS(2518),
    [anon_sym_LPAREN2] = ACTIONS(2520),
    [anon_sym_BANG] = ACTIONS(2520),
    [anon_sym_TILDE] = ACTIONS(2520),
    [anon_sym_DASH] = ACTIONS(2518),
    [anon_sym_PLUS] = ACTIONS(2518),
    [anon_sym_STAR] = ACTIONS(2520),
    [anon_sym_CARET] = ACTIONS(2520),
    [anon_sym_AMP] = ACTIONS(2520),
    [anon_sym_SEMI] = ACTIONS(2520),
    [anon_sym___extension__] = ACTIONS(2518),
    [anon_sym_typedef] = ACTIONS(2518),
    [anon_sym_extern] = ACTIONS(2518),
    [anon_sym___attribute__] = ACTIONS(2518),
    [anon_sym___attribute] = ACTIONS(2518),
    [anon_sym_noreturn] = ACTIONS(2518),
    [anon_sym_LBRACK] = ACTIONS(2520),
    [anon_sym___declspec] = ACTIONS(2518),
    [anon_sym___cdecl] = ACTIONS(2518),
    [anon_sym___clrcall] = ACTIONS(2518),
    [anon_sym___stdcall] = ACTIONS(2518),
    [anon_sym___fastcall] = ACTIONS(2518),
    [anon_sym___thiscall] = ACTIONS(2518),
    [anon_sym___vectorcall] = ACTIONS(2518),
    [anon_sym_LBRACE] = ACTIONS(2520),
    [anon_sym_signed] = ACTIONS(2518),
    [anon_sym_unsigned] = ACTIONS(2518),
    [anon_sym_long] = ACTIONS(2518),
    [anon_sym_short] = ACTIONS(2518),
    [anon_sym_static] = ACTIONS(2518),
    [anon_sym_ATautoreleasepool] = ACTIONS(2520),
    [anon_sym_auto] = ACTIONS(2518),
    [anon_sym_register] = ACTIONS(2518),
    [anon_sym_inline] = ACTIONS(2518),
    [anon_sym___inline] = ACTIONS(2518),
    [anon_sym___inline__] = ACTIONS(2518),
    [anon_sym___forceinline] = ACTIONS(2518),
    [anon_sym_thread_local] = ACTIONS(2518),
    [anon_sym___thread] = ACTIONS(2518),
    [anon_sym_CG_EXTERN] = ACTIONS(2518),
    [anon_sym_CG_INLINE] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2518),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2518),
    [anon_sym_IBOutlet] = ACTIONS(2518),
    [anon_sym_IBInspectable] = ACTIONS(2518),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2518),
    [anon_sym_NS_INLINE] = ACTIONS(2518),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2518),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2518),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2518),
    [anon_sym_const] = ACTIONS(2518),
    [anon_sym_constexpr] = ACTIONS(2518),
    [anon_sym_volatile] = ACTIONS(2518),
    [anon_sym_restrict] = ACTIONS(2518),
    [anon_sym___restrict__] = ACTIONS(2518),
    [anon_sym__Atomic] = ACTIONS(2518),
    [anon_sym__Noreturn] = ACTIONS(2518),
    [anon_sym__Nonnull] = ACTIONS(2518),
    [anon_sym_nullable] = ACTIONS(2518),
    [anon_sym__Complex] = ACTIONS(2518),
    [anon_sym__Nullable] = ACTIONS(2518),
    [anon_sym__Nullable_result] = ACTIONS(2518),
    [anon_sym__Null_unspecified] = ACTIONS(2518),
    [anon_sym___autoreleasing] = ACTIONS(2518),
    [anon_sym___block] = ACTIONS(2518),
    [anon_sym___bridge] = ACTIONS(2518),
    [anon_sym___bridge_retained] = ACTIONS(2518),
    [anon_sym___bridge_transfer] = ACTIONS(2518),
    [anon_sym___complex] = ACTIONS(2518),
    [anon_sym___const] = ACTIONS(2518),
    [anon_sym___imag] = ACTIONS(2518),
    [anon_sym___kindof] = ACTIONS(2518),
    [anon_sym___nonnull] = ACTIONS(2518),
    [anon_sym___nullable] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2518),
    [anon_sym___real] = ACTIONS(2518),
    [anon_sym___strong] = ACTIONS(2518),
    [anon_sym___unsafe_unretained] = ACTIONS(2518),
    [anon_sym___unused] = ACTIONS(2518),
    [anon_sym___weak] = ACTIONS(2518),
    [anon_sym_alignas] = ACTIONS(2518),
    [anon_sym__Alignas] = ACTIONS(2518),
    [sym_primitive_type] = ACTIONS(2518),
    [anon_sym_enum] = ACTIONS(2518),
    [anon_sym_struct] = ACTIONS(2518),
    [anon_sym_union] = ACTIONS(2518),
    [anon_sym_if] = ACTIONS(2518),
    [anon_sym_else] = ACTIONS(2518),
    [anon_sym_switch] = ACTIONS(2518),
    [anon_sym_case] = ACTIONS(2518),
    [anon_sym_default] = ACTIONS(2518),
    [anon_sym_while] = ACTIONS(2518),
    [anon_sym_do] = ACTIONS(2518),
    [anon_sym_for] = ACTIONS(2518),
    [anon_sym_in] = ACTIONS(2518),
    [anon_sym_return] = ACTIONS(2518),
    [anon_sym_break] = ACTIONS(2518),
    [anon_sym_continue] = ACTIONS(2518),
    [anon_sym_goto] = ACTIONS(2518),
    [anon_sym_DASH_DASH] = ACTIONS(2520),
    [anon_sym_PLUS_PLUS] = ACTIONS(2520),
    [anon_sym_sizeof] = ACTIONS(2518),
    [anon_sym___alignof__] = ACTIONS(2518),
    [anon_sym___alignof] = ACTIONS(2518),
    [anon_sym__alignof] = ACTIONS(2518),
    [anon_sym_alignof] = ACTIONS(2518),
    [anon_sym__Alignof] = ACTIONS(2518),
    [anon_sym_offsetof] = ACTIONS(2518),
    [anon_sym__Generic] = ACTIONS(2518),
    [anon_sym_asm] = ACTIONS(2518),
    [anon_sym___asm__] = ACTIONS(2518),
    [anon_sym___asm] = ACTIONS(2518),
    [sym_number_literal] = ACTIONS(2520),
    [anon_sym_L_SQUOTE] = ACTIONS(2520),
    [anon_sym_u_SQUOTE] = ACTIONS(2520),
    [anon_sym_U_SQUOTE] = ACTIONS(2520),
    [anon_sym_u8_SQUOTE] = ACTIONS(2520),
    [anon_sym_SQUOTE] = ACTIONS(2520),
    [anon_sym_AT] = ACTIONS(2518),
    [anon_sym_DQUOTE] = ACTIONS(2520),
    [anon_sym_L_DQUOTE] = ACTIONS(2520),
    [anon_sym_u_DQUOTE] = ACTIONS(2520),
    [anon_sym_U_DQUOTE] = ACTIONS(2520),
    [anon_sym_u8_DQUOTE] = ACTIONS(2520),
    [sym_true] = ACTIONS(2518),
    [sym_false] = ACTIONS(2518),
    [anon_sym_NULL] = ACTIONS(2518),
    [anon_sym_nullptr] = ACTIONS(2518),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2518),
    [anon_sym___typeof] = ACTIONS(2518),
    [anon_sym_typeof] = ACTIONS(2518),
    [anon_sym_ATimport] = ACTIONS(2520),
    [aux_sym_preproc_undef_token1] = ACTIONS(2518),
    [anon_sym_POUND] = ACTIONS(2518),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2518),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2518),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2518),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_API_AVAILABLE] = ACTIONS(2518),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_API_DEPRECATED] = ACTIONS(2518),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym___deprecated_msg] = ACTIONS(2518),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2518),
    [anon_sym_ATprotocol] = ACTIONS(2520),
    [anon_sym_ATinterface] = ACTIONS(2520),
    [anon_sym_ATimplementation] = ACTIONS(2520),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2520),
    [anon_sym_ATsynthesize] = ACTIONS(2520),
    [anon_sym_ATdynamic] = ACTIONS(2520),
    [anon_sym_ATtry] = ACTIONS(2520),
    [anon_sym___try] = ACTIONS(2518),
    [anon_sym_ATthrow] = ACTIONS(2520),
    [anon_sym_ATselector] = ACTIONS(2520),
    [anon_sym_ATavailable] = ACTIONS(2520),
    [anon_sym___builtin_available] = ACTIONS(2518),
    [anon_sym_va_arg] = ACTIONS(2518),
    [anon_sym_ATencode] = ACTIONS(2520),
    [anon_sym_ATsynchronized] = ACTIONS(2520),
    [anon_sym_BOOL] = ACTIONS(2518),
    [anon_sym_IMP] = ACTIONS(2518),
    [anon_sym_SEL] = ACTIONS(2518),
    [anon_sym_Class] = ACTIONS(2518),
    [anon_sym_id] = ACTIONS(2518),
  },
  [STATE(262)] = {
    [sym_identifier] = ACTIONS(2522),
    [aux_sym_preproc_include_token1] = ACTIONS(2522),
    [aux_sym_preproc_include_token2] = ACTIONS(2522),
    [aux_sym_preproc_def_token1] = ACTIONS(2522),
    [aux_sym_preproc_if_token1] = ACTIONS(2522),
    [aux_sym_preproc_if_token2] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2522),
    [aux_sym_preproc_else_token1] = ACTIONS(2522),
    [aux_sym_preproc_elif_token1] = ACTIONS(2522),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2522),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2522),
    [sym_preproc_directive] = ACTIONS(2522),
    [anon_sym_LPAREN2] = ACTIONS(2524),
    [anon_sym_BANG] = ACTIONS(2524),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_DASH] = ACTIONS(2522),
    [anon_sym_PLUS] = ACTIONS(2522),
    [anon_sym_STAR] = ACTIONS(2524),
    [anon_sym_CARET] = ACTIONS(2524),
    [anon_sym_AMP] = ACTIONS(2524),
    [anon_sym_SEMI] = ACTIONS(2524),
    [anon_sym___extension__] = ACTIONS(2522),
    [anon_sym_typedef] = ACTIONS(2522),
    [anon_sym_extern] = ACTIONS(2522),
    [anon_sym___attribute__] = ACTIONS(2522),
    [anon_sym___attribute] = ACTIONS(2522),
    [anon_sym_noreturn] = ACTIONS(2522),
    [anon_sym_LBRACK] = ACTIONS(2524),
    [anon_sym___declspec] = ACTIONS(2522),
    [anon_sym___cdecl] = ACTIONS(2522),
    [anon_sym___clrcall] = ACTIONS(2522),
    [anon_sym___stdcall] = ACTIONS(2522),
    [anon_sym___fastcall] = ACTIONS(2522),
    [anon_sym___thiscall] = ACTIONS(2522),
    [anon_sym___vectorcall] = ACTIONS(2522),
    [anon_sym_LBRACE] = ACTIONS(2524),
    [anon_sym_signed] = ACTIONS(2522),
    [anon_sym_unsigned] = ACTIONS(2522),
    [anon_sym_long] = ACTIONS(2522),
    [anon_sym_short] = ACTIONS(2522),
    [anon_sym_static] = ACTIONS(2522),
    [anon_sym_ATautoreleasepool] = ACTIONS(2524),
    [anon_sym_auto] = ACTIONS(2522),
    [anon_sym_register] = ACTIONS(2522),
    [anon_sym_inline] = ACTIONS(2522),
    [anon_sym___inline] = ACTIONS(2522),
    [anon_sym___inline__] = ACTIONS(2522),
    [anon_sym___forceinline] = ACTIONS(2522),
    [anon_sym_thread_local] = ACTIONS(2522),
    [anon_sym___thread] = ACTIONS(2522),
    [anon_sym_CG_EXTERN] = ACTIONS(2522),
    [anon_sym_CG_INLINE] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2522),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2522),
    [anon_sym_IBOutlet] = ACTIONS(2522),
    [anon_sym_IBInspectable] = ACTIONS(2522),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2522),
    [anon_sym_NS_INLINE] = ACTIONS(2522),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2522),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2522),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2522),
    [anon_sym_const] = ACTIONS(2522),
    [anon_sym_constexpr] = ACTIONS(2522),
    [anon_sym_volatile] = ACTIONS(2522),
    [anon_sym_restrict] = ACTIONS(2522),
    [anon_sym___restrict__] = ACTIONS(2522),
    [anon_sym__Atomic] = ACTIONS(2522),
    [anon_sym__Noreturn] = ACTIONS(2522),
    [anon_sym__Nonnull] = ACTIONS(2522),
    [anon_sym_nullable] = ACTIONS(2522),
    [anon_sym__Complex] = ACTIONS(2522),
    [anon_sym__Nullable] = ACTIONS(2522),
    [anon_sym__Nullable_result] = ACTIONS(2522),
    [anon_sym__Null_unspecified] = ACTIONS(2522),
    [anon_sym___autoreleasing] = ACTIONS(2522),
    [anon_sym___block] = ACTIONS(2522),
    [anon_sym___bridge] = ACTIONS(2522),
    [anon_sym___bridge_retained] = ACTIONS(2522),
    [anon_sym___bridge_transfer] = ACTIONS(2522),
    [anon_sym___complex] = ACTIONS(2522),
    [anon_sym___const] = ACTIONS(2522),
    [anon_sym___imag] = ACTIONS(2522),
    [anon_sym___kindof] = ACTIONS(2522),
    [anon_sym___nonnull] = ACTIONS(2522),
    [anon_sym___nullable] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2522),
    [anon_sym___real] = ACTIONS(2522),
    [anon_sym___strong] = ACTIONS(2522),
    [anon_sym___unsafe_unretained] = ACTIONS(2522),
    [anon_sym___unused] = ACTIONS(2522),
    [anon_sym___weak] = ACTIONS(2522),
    [anon_sym_alignas] = ACTIONS(2522),
    [anon_sym__Alignas] = ACTIONS(2522),
    [sym_primitive_type] = ACTIONS(2522),
    [anon_sym_enum] = ACTIONS(2522),
    [anon_sym_struct] = ACTIONS(2522),
    [anon_sym_union] = ACTIONS(2522),
    [anon_sym_if] = ACTIONS(2522),
    [anon_sym_else] = ACTIONS(2522),
    [anon_sym_switch] = ACTIONS(2522),
    [anon_sym_case] = ACTIONS(2522),
    [anon_sym_default] = ACTIONS(2522),
    [anon_sym_while] = ACTIONS(2522),
    [anon_sym_do] = ACTIONS(2522),
    [anon_sym_for] = ACTIONS(2522),
    [anon_sym_in] = ACTIONS(2522),
    [anon_sym_return] = ACTIONS(2522),
    [anon_sym_break] = ACTIONS(2522),
    [anon_sym_continue] = ACTIONS(2522),
    [anon_sym_goto] = ACTIONS(2522),
    [anon_sym_DASH_DASH] = ACTIONS(2524),
    [anon_sym_PLUS_PLUS] = ACTIONS(2524),
    [anon_sym_sizeof] = ACTIONS(2522),
    [anon_sym___alignof__] = ACTIONS(2522),
    [anon_sym___alignof] = ACTIONS(2522),
    [anon_sym__alignof] = ACTIONS(2522),
    [anon_sym_alignof] = ACTIONS(2522),
    [anon_sym__Alignof] = ACTIONS(2522),
    [anon_sym_offsetof] = ACTIONS(2522),
    [anon_sym__Generic] = ACTIONS(2522),
    [anon_sym_asm] = ACTIONS(2522),
    [anon_sym___asm__] = ACTIONS(2522),
    [anon_sym___asm] = ACTIONS(2522),
    [sym_number_literal] = ACTIONS(2524),
    [anon_sym_L_SQUOTE] = ACTIONS(2524),
    [anon_sym_u_SQUOTE] = ACTIONS(2524),
    [anon_sym_U_SQUOTE] = ACTIONS(2524),
    [anon_sym_u8_SQUOTE] = ACTIONS(2524),
    [anon_sym_SQUOTE] = ACTIONS(2524),
    [anon_sym_AT] = ACTIONS(2522),
    [anon_sym_DQUOTE] = ACTIONS(2524),
    [anon_sym_L_DQUOTE] = ACTIONS(2524),
    [anon_sym_u_DQUOTE] = ACTIONS(2524),
    [anon_sym_U_DQUOTE] = ACTIONS(2524),
    [anon_sym_u8_DQUOTE] = ACTIONS(2524),
    [sym_true] = ACTIONS(2522),
    [sym_false] = ACTIONS(2522),
    [anon_sym_NULL] = ACTIONS(2522),
    [anon_sym_nullptr] = ACTIONS(2522),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2522),
    [anon_sym___typeof] = ACTIONS(2522),
    [anon_sym_typeof] = ACTIONS(2522),
    [anon_sym_ATimport] = ACTIONS(2524),
    [aux_sym_preproc_undef_token1] = ACTIONS(2522),
    [anon_sym_POUND] = ACTIONS(2522),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2522),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2522),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2522),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_API_AVAILABLE] = ACTIONS(2522),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_API_DEPRECATED] = ACTIONS(2522),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym___deprecated_msg] = ACTIONS(2522),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2522),
    [anon_sym_ATprotocol] = ACTIONS(2524),
    [anon_sym_ATinterface] = ACTIONS(2524),
    [anon_sym_ATimplementation] = ACTIONS(2524),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2524),
    [anon_sym_ATsynthesize] = ACTIONS(2524),
    [anon_sym_ATdynamic] = ACTIONS(2524),
    [anon_sym_ATtry] = ACTIONS(2524),
    [anon_sym___try] = ACTIONS(2522),
    [anon_sym_ATthrow] = ACTIONS(2524),
    [anon_sym_ATselector] = ACTIONS(2524),
    [anon_sym_ATavailable] = ACTIONS(2524),
    [anon_sym___builtin_available] = ACTIONS(2522),
    [anon_sym_va_arg] = ACTIONS(2522),
    [anon_sym_ATencode] = ACTIONS(2524),
    [anon_sym_ATsynchronized] = ACTIONS(2524),
    [anon_sym_BOOL] = ACTIONS(2522),
    [anon_sym_IMP] = ACTIONS(2522),
    [anon_sym_SEL] = ACTIONS(2522),
    [anon_sym_Class] = ACTIONS(2522),
    [anon_sym_id] = ACTIONS(2522),
  },
  [STATE(263)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [aux_sym_preproc_else_token1] = ACTIONS(2510),
    [aux_sym_preproc_elif_token1] = ACTIONS(2510),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(264)] = {
    [sym_identifier] = ACTIONS(2526),
    [aux_sym_preproc_include_token1] = ACTIONS(2526),
    [aux_sym_preproc_include_token2] = ACTIONS(2526),
    [aux_sym_preproc_def_token1] = ACTIONS(2526),
    [aux_sym_preproc_if_token1] = ACTIONS(2526),
    [aux_sym_preproc_if_token2] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2526),
    [aux_sym_preproc_else_token1] = ACTIONS(2526),
    [aux_sym_preproc_elif_token1] = ACTIONS(2526),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2526),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2526),
    [sym_preproc_directive] = ACTIONS(2526),
    [anon_sym_LPAREN2] = ACTIONS(2528),
    [anon_sym_BANG] = ACTIONS(2528),
    [anon_sym_TILDE] = ACTIONS(2528),
    [anon_sym_DASH] = ACTIONS(2526),
    [anon_sym_PLUS] = ACTIONS(2526),
    [anon_sym_STAR] = ACTIONS(2528),
    [anon_sym_CARET] = ACTIONS(2528),
    [anon_sym_AMP] = ACTIONS(2528),
    [anon_sym_SEMI] = ACTIONS(2528),
    [anon_sym___extension__] = ACTIONS(2526),
    [anon_sym_typedef] = ACTIONS(2526),
    [anon_sym_extern] = ACTIONS(2526),
    [anon_sym___attribute__] = ACTIONS(2526),
    [anon_sym___attribute] = ACTIONS(2526),
    [anon_sym_noreturn] = ACTIONS(2526),
    [anon_sym_LBRACK] = ACTIONS(2528),
    [anon_sym___declspec] = ACTIONS(2526),
    [anon_sym___cdecl] = ACTIONS(2526),
    [anon_sym___clrcall] = ACTIONS(2526),
    [anon_sym___stdcall] = ACTIONS(2526),
    [anon_sym___fastcall] = ACTIONS(2526),
    [anon_sym___thiscall] = ACTIONS(2526),
    [anon_sym___vectorcall] = ACTIONS(2526),
    [anon_sym_LBRACE] = ACTIONS(2528),
    [anon_sym_signed] = ACTIONS(2526),
    [anon_sym_unsigned] = ACTIONS(2526),
    [anon_sym_long] = ACTIONS(2526),
    [anon_sym_short] = ACTIONS(2526),
    [anon_sym_static] = ACTIONS(2526),
    [anon_sym_ATautoreleasepool] = ACTIONS(2528),
    [anon_sym_auto] = ACTIONS(2526),
    [anon_sym_register] = ACTIONS(2526),
    [anon_sym_inline] = ACTIONS(2526),
    [anon_sym___inline] = ACTIONS(2526),
    [anon_sym___inline__] = ACTIONS(2526),
    [anon_sym___forceinline] = ACTIONS(2526),
    [anon_sym_thread_local] = ACTIONS(2526),
    [anon_sym___thread] = ACTIONS(2526),
    [anon_sym_CG_EXTERN] = ACTIONS(2526),
    [anon_sym_CG_INLINE] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2526),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2526),
    [anon_sym_IBOutlet] = ACTIONS(2526),
    [anon_sym_IBInspectable] = ACTIONS(2526),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2526),
    [anon_sym_NS_INLINE] = ACTIONS(2526),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2526),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2526),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2526),
    [anon_sym_const] = ACTIONS(2526),
    [anon_sym_constexpr] = ACTIONS(2526),
    [anon_sym_volatile] = ACTIONS(2526),
    [anon_sym_restrict] = ACTIONS(2526),
    [anon_sym___restrict__] = ACTIONS(2526),
    [anon_sym__Atomic] = ACTIONS(2526),
    [anon_sym__Noreturn] = ACTIONS(2526),
    [anon_sym__Nonnull] = ACTIONS(2526),
    [anon_sym_nullable] = ACTIONS(2526),
    [anon_sym__Complex] = ACTIONS(2526),
    [anon_sym__Nullable] = ACTIONS(2526),
    [anon_sym__Nullable_result] = ACTIONS(2526),
    [anon_sym__Null_unspecified] = ACTIONS(2526),
    [anon_sym___autoreleasing] = ACTIONS(2526),
    [anon_sym___block] = ACTIONS(2526),
    [anon_sym___bridge] = ACTIONS(2526),
    [anon_sym___bridge_retained] = ACTIONS(2526),
    [anon_sym___bridge_transfer] = ACTIONS(2526),
    [anon_sym___complex] = ACTIONS(2526),
    [anon_sym___const] = ACTIONS(2526),
    [anon_sym___imag] = ACTIONS(2526),
    [anon_sym___kindof] = ACTIONS(2526),
    [anon_sym___nonnull] = ACTIONS(2526),
    [anon_sym___nullable] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2526),
    [anon_sym___real] = ACTIONS(2526),
    [anon_sym___strong] = ACTIONS(2526),
    [anon_sym___unsafe_unretained] = ACTIONS(2526),
    [anon_sym___unused] = ACTIONS(2526),
    [anon_sym___weak] = ACTIONS(2526),
    [anon_sym_alignas] = ACTIONS(2526),
    [anon_sym__Alignas] = ACTIONS(2526),
    [sym_primitive_type] = ACTIONS(2526),
    [anon_sym_enum] = ACTIONS(2526),
    [anon_sym_struct] = ACTIONS(2526),
    [anon_sym_union] = ACTIONS(2526),
    [anon_sym_if] = ACTIONS(2526),
    [anon_sym_else] = ACTIONS(2526),
    [anon_sym_switch] = ACTIONS(2526),
    [anon_sym_case] = ACTIONS(2526),
    [anon_sym_default] = ACTIONS(2526),
    [anon_sym_while] = ACTIONS(2526),
    [anon_sym_do] = ACTIONS(2526),
    [anon_sym_for] = ACTIONS(2526),
    [anon_sym_in] = ACTIONS(2526),
    [anon_sym_return] = ACTIONS(2526),
    [anon_sym_break] = ACTIONS(2526),
    [anon_sym_continue] = ACTIONS(2526),
    [anon_sym_goto] = ACTIONS(2526),
    [anon_sym_DASH_DASH] = ACTIONS(2528),
    [anon_sym_PLUS_PLUS] = ACTIONS(2528),
    [anon_sym_sizeof] = ACTIONS(2526),
    [anon_sym___alignof__] = ACTIONS(2526),
    [anon_sym___alignof] = ACTIONS(2526),
    [anon_sym__alignof] = ACTIONS(2526),
    [anon_sym_alignof] = ACTIONS(2526),
    [anon_sym__Alignof] = ACTIONS(2526),
    [anon_sym_offsetof] = ACTIONS(2526),
    [anon_sym__Generic] = ACTIONS(2526),
    [anon_sym_asm] = ACTIONS(2526),
    [anon_sym___asm__] = ACTIONS(2526),
    [anon_sym___asm] = ACTIONS(2526),
    [sym_number_literal] = ACTIONS(2528),
    [anon_sym_L_SQUOTE] = ACTIONS(2528),
    [anon_sym_u_SQUOTE] = ACTIONS(2528),
    [anon_sym_U_SQUOTE] = ACTIONS(2528),
    [anon_sym_u8_SQUOTE] = ACTIONS(2528),
    [anon_sym_SQUOTE] = ACTIONS(2528),
    [anon_sym_AT] = ACTIONS(2526),
    [anon_sym_DQUOTE] = ACTIONS(2528),
    [anon_sym_L_DQUOTE] = ACTIONS(2528),
    [anon_sym_u_DQUOTE] = ACTIONS(2528),
    [anon_sym_U_DQUOTE] = ACTIONS(2528),
    [anon_sym_u8_DQUOTE] = ACTIONS(2528),
    [sym_true] = ACTIONS(2526),
    [sym_false] = ACTIONS(2526),
    [anon_sym_NULL] = ACTIONS(2526),
    [anon_sym_nullptr] = ACTIONS(2526),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2526),
    [anon_sym___typeof] = ACTIONS(2526),
    [anon_sym_typeof] = ACTIONS(2526),
    [anon_sym_ATimport] = ACTIONS(2528),
    [aux_sym_preproc_undef_token1] = ACTIONS(2526),
    [anon_sym_POUND] = ACTIONS(2526),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2526),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2526),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2526),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_API_AVAILABLE] = ACTIONS(2526),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_API_DEPRECATED] = ACTIONS(2526),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym___deprecated_msg] = ACTIONS(2526),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2526),
    [anon_sym_ATprotocol] = ACTIONS(2528),
    [anon_sym_ATinterface] = ACTIONS(2528),
    [anon_sym_ATimplementation] = ACTIONS(2528),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2528),
    [anon_sym_ATsynthesize] = ACTIONS(2528),
    [anon_sym_ATdynamic] = ACTIONS(2528),
    [anon_sym_ATtry] = ACTIONS(2528),
    [anon_sym___try] = ACTIONS(2526),
    [anon_sym_ATthrow] = ACTIONS(2528),
    [anon_sym_ATselector] = ACTIONS(2528),
    [anon_sym_ATavailable] = ACTIONS(2528),
    [anon_sym___builtin_available] = ACTIONS(2526),
    [anon_sym_va_arg] = ACTIONS(2526),
    [anon_sym_ATencode] = ACTIONS(2528),
    [anon_sym_ATsynchronized] = ACTIONS(2528),
    [anon_sym_BOOL] = ACTIONS(2526),
    [anon_sym_IMP] = ACTIONS(2526),
    [anon_sym_SEL] = ACTIONS(2526),
    [anon_sym_Class] = ACTIONS(2526),
    [anon_sym_id] = ACTIONS(2526),
  },
  [STATE(265)] = {
    [sym_identifier] = ACTIONS(2530),
    [aux_sym_preproc_include_token1] = ACTIONS(2530),
    [aux_sym_preproc_include_token2] = ACTIONS(2530),
    [aux_sym_preproc_def_token1] = ACTIONS(2530),
    [aux_sym_preproc_if_token1] = ACTIONS(2530),
    [aux_sym_preproc_if_token2] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2530),
    [aux_sym_preproc_else_token1] = ACTIONS(2530),
    [aux_sym_preproc_elif_token1] = ACTIONS(2530),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2530),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2530),
    [sym_preproc_directive] = ACTIONS(2530),
    [anon_sym_LPAREN2] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2530),
    [anon_sym_PLUS] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym___extension__] = ACTIONS(2530),
    [anon_sym_typedef] = ACTIONS(2530),
    [anon_sym_extern] = ACTIONS(2530),
    [anon_sym___attribute__] = ACTIONS(2530),
    [anon_sym___attribute] = ACTIONS(2530),
    [anon_sym_noreturn] = ACTIONS(2530),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym___declspec] = ACTIONS(2530),
    [anon_sym___cdecl] = ACTIONS(2530),
    [anon_sym___clrcall] = ACTIONS(2530),
    [anon_sym___stdcall] = ACTIONS(2530),
    [anon_sym___fastcall] = ACTIONS(2530),
    [anon_sym___thiscall] = ACTIONS(2530),
    [anon_sym___vectorcall] = ACTIONS(2530),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_signed] = ACTIONS(2530),
    [anon_sym_unsigned] = ACTIONS(2530),
    [anon_sym_long] = ACTIONS(2530),
    [anon_sym_short] = ACTIONS(2530),
    [anon_sym_static] = ACTIONS(2530),
    [anon_sym_ATautoreleasepool] = ACTIONS(2532),
    [anon_sym_auto] = ACTIONS(2530),
    [anon_sym_register] = ACTIONS(2530),
    [anon_sym_inline] = ACTIONS(2530),
    [anon_sym___inline] = ACTIONS(2530),
    [anon_sym___inline__] = ACTIONS(2530),
    [anon_sym___forceinline] = ACTIONS(2530),
    [anon_sym_thread_local] = ACTIONS(2530),
    [anon_sym___thread] = ACTIONS(2530),
    [anon_sym_CG_EXTERN] = ACTIONS(2530),
    [anon_sym_CG_INLINE] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2530),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2530),
    [anon_sym_IBOutlet] = ACTIONS(2530),
    [anon_sym_IBInspectable] = ACTIONS(2530),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2530),
    [anon_sym_NS_INLINE] = ACTIONS(2530),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2530),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2530),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2530),
    [anon_sym_const] = ACTIONS(2530),
    [anon_sym_constexpr] = ACTIONS(2530),
    [anon_sym_volatile] = ACTIONS(2530),
    [anon_sym_restrict] = ACTIONS(2530),
    [anon_sym___restrict__] = ACTIONS(2530),
    [anon_sym__Atomic] = ACTIONS(2530),
    [anon_sym__Noreturn] = ACTIONS(2530),
    [anon_sym__Nonnull] = ACTIONS(2530),
    [anon_sym_nullable] = ACTIONS(2530),
    [anon_sym__Complex] = ACTIONS(2530),
    [anon_sym__Nullable] = ACTIONS(2530),
    [anon_sym__Nullable_result] = ACTIONS(2530),
    [anon_sym__Null_unspecified] = ACTIONS(2530),
    [anon_sym___autoreleasing] = ACTIONS(2530),
    [anon_sym___block] = ACTIONS(2530),
    [anon_sym___bridge] = ACTIONS(2530),
    [anon_sym___bridge_retained] = ACTIONS(2530),
    [anon_sym___bridge_transfer] = ACTIONS(2530),
    [anon_sym___complex] = ACTIONS(2530),
    [anon_sym___const] = ACTIONS(2530),
    [anon_sym___imag] = ACTIONS(2530),
    [anon_sym___kindof] = ACTIONS(2530),
    [anon_sym___nonnull] = ACTIONS(2530),
    [anon_sym___nullable] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2530),
    [anon_sym___real] = ACTIONS(2530),
    [anon_sym___strong] = ACTIONS(2530),
    [anon_sym___unsafe_unretained] = ACTIONS(2530),
    [anon_sym___unused] = ACTIONS(2530),
    [anon_sym___weak] = ACTIONS(2530),
    [anon_sym_alignas] = ACTIONS(2530),
    [anon_sym__Alignas] = ACTIONS(2530),
    [sym_primitive_type] = ACTIONS(2530),
    [anon_sym_enum] = ACTIONS(2530),
    [anon_sym_struct] = ACTIONS(2530),
    [anon_sym_union] = ACTIONS(2530),
    [anon_sym_if] = ACTIONS(2530),
    [anon_sym_else] = ACTIONS(2530),
    [anon_sym_switch] = ACTIONS(2530),
    [anon_sym_case] = ACTIONS(2530),
    [anon_sym_default] = ACTIONS(2530),
    [anon_sym_while] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2530),
    [anon_sym_for] = ACTIONS(2530),
    [anon_sym_in] = ACTIONS(2530),
    [anon_sym_return] = ACTIONS(2530),
    [anon_sym_break] = ACTIONS(2530),
    [anon_sym_continue] = ACTIONS(2530),
    [anon_sym_goto] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_sizeof] = ACTIONS(2530),
    [anon_sym___alignof__] = ACTIONS(2530),
    [anon_sym___alignof] = ACTIONS(2530),
    [anon_sym__alignof] = ACTIONS(2530),
    [anon_sym_alignof] = ACTIONS(2530),
    [anon_sym__Alignof] = ACTIONS(2530),
    [anon_sym_offsetof] = ACTIONS(2530),
    [anon_sym__Generic] = ACTIONS(2530),
    [anon_sym_asm] = ACTIONS(2530),
    [anon_sym___asm__] = ACTIONS(2530),
    [anon_sym___asm] = ACTIONS(2530),
    [sym_number_literal] = ACTIONS(2532),
    [anon_sym_L_SQUOTE] = ACTIONS(2532),
    [anon_sym_u_SQUOTE] = ACTIONS(2532),
    [anon_sym_U_SQUOTE] = ACTIONS(2532),
    [anon_sym_u8_SQUOTE] = ACTIONS(2532),
    [anon_sym_SQUOTE] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2530),
    [anon_sym_DQUOTE] = ACTIONS(2532),
    [anon_sym_L_DQUOTE] = ACTIONS(2532),
    [anon_sym_u_DQUOTE] = ACTIONS(2532),
    [anon_sym_U_DQUOTE] = ACTIONS(2532),
    [anon_sym_u8_DQUOTE] = ACTIONS(2532),
    [sym_true] = ACTIONS(2530),
    [sym_false] = ACTIONS(2530),
    [anon_sym_NULL] = ACTIONS(2530),
    [anon_sym_nullptr] = ACTIONS(2530),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2530),
    [anon_sym___typeof] = ACTIONS(2530),
    [anon_sym_typeof] = ACTIONS(2530),
    [anon_sym_ATimport] = ACTIONS(2532),
    [aux_sym_preproc_undef_token1] = ACTIONS(2530),
    [anon_sym_POUND] = ACTIONS(2530),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2530),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2530),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2530),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_API_AVAILABLE] = ACTIONS(2530),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_API_DEPRECATED] = ACTIONS(2530),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym___deprecated_msg] = ACTIONS(2530),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2530),
    [anon_sym_ATprotocol] = ACTIONS(2532),
    [anon_sym_ATinterface] = ACTIONS(2532),
    [anon_sym_ATimplementation] = ACTIONS(2532),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2532),
    [anon_sym_ATsynthesize] = ACTIONS(2532),
    [anon_sym_ATdynamic] = ACTIONS(2532),
    [anon_sym_ATtry] = ACTIONS(2532),
    [anon_sym___try] = ACTIONS(2530),
    [anon_sym_ATthrow] = ACTIONS(2532),
    [anon_sym_ATselector] = ACTIONS(2532),
    [anon_sym_ATavailable] = ACTIONS(2532),
    [anon_sym___builtin_available] = ACTIONS(2530),
    [anon_sym_va_arg] = ACTIONS(2530),
    [anon_sym_ATencode] = ACTIONS(2532),
    [anon_sym_ATsynchronized] = ACTIONS(2532),
    [anon_sym_BOOL] = ACTIONS(2530),
    [anon_sym_IMP] = ACTIONS(2530),
    [anon_sym_SEL] = ACTIONS(2530),
    [anon_sym_Class] = ACTIONS(2530),
    [anon_sym_id] = ACTIONS(2530),
  },
  [STATE(266)] = {
    [sym_identifier] = ACTIONS(2534),
    [aux_sym_preproc_include_token1] = ACTIONS(2534),
    [aux_sym_preproc_include_token2] = ACTIONS(2534),
    [aux_sym_preproc_def_token1] = ACTIONS(2534),
    [aux_sym_preproc_if_token1] = ACTIONS(2534),
    [aux_sym_preproc_if_token2] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2534),
    [aux_sym_preproc_else_token1] = ACTIONS(2534),
    [aux_sym_preproc_elif_token1] = ACTIONS(2534),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2534),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2534),
    [sym_preproc_directive] = ACTIONS(2534),
    [anon_sym_LPAREN2] = ACTIONS(2536),
    [anon_sym_BANG] = ACTIONS(2536),
    [anon_sym_TILDE] = ACTIONS(2536),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2536),
    [anon_sym_CARET] = ACTIONS(2536),
    [anon_sym_AMP] = ACTIONS(2536),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym___extension__] = ACTIONS(2534),
    [anon_sym_typedef] = ACTIONS(2534),
    [anon_sym_extern] = ACTIONS(2534),
    [anon_sym___attribute__] = ACTIONS(2534),
    [anon_sym___attribute] = ACTIONS(2534),
    [anon_sym_noreturn] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2536),
    [anon_sym___declspec] = ACTIONS(2534),
    [anon_sym___cdecl] = ACTIONS(2534),
    [anon_sym___clrcall] = ACTIONS(2534),
    [anon_sym___stdcall] = ACTIONS(2534),
    [anon_sym___fastcall] = ACTIONS(2534),
    [anon_sym___thiscall] = ACTIONS(2534),
    [anon_sym___vectorcall] = ACTIONS(2534),
    [anon_sym_LBRACE] = ACTIONS(2536),
    [anon_sym_signed] = ACTIONS(2534),
    [anon_sym_unsigned] = ACTIONS(2534),
    [anon_sym_long] = ACTIONS(2534),
    [anon_sym_short] = ACTIONS(2534),
    [anon_sym_static] = ACTIONS(2534),
    [anon_sym_ATautoreleasepool] = ACTIONS(2536),
    [anon_sym_auto] = ACTIONS(2534),
    [anon_sym_register] = ACTIONS(2534),
    [anon_sym_inline] = ACTIONS(2534),
    [anon_sym___inline] = ACTIONS(2534),
    [anon_sym___inline__] = ACTIONS(2534),
    [anon_sym___forceinline] = ACTIONS(2534),
    [anon_sym_thread_local] = ACTIONS(2534),
    [anon_sym___thread] = ACTIONS(2534),
    [anon_sym_CG_EXTERN] = ACTIONS(2534),
    [anon_sym_CG_INLINE] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2534),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2534),
    [anon_sym_IBOutlet] = ACTIONS(2534),
    [anon_sym_IBInspectable] = ACTIONS(2534),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2534),
    [anon_sym_NS_INLINE] = ACTIONS(2534),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2534),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2534),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2534),
    [anon_sym_const] = ACTIONS(2534),
    [anon_sym_constexpr] = ACTIONS(2534),
    [anon_sym_volatile] = ACTIONS(2534),
    [anon_sym_restrict] = ACTIONS(2534),
    [anon_sym___restrict__] = ACTIONS(2534),
    [anon_sym__Atomic] = ACTIONS(2534),
    [anon_sym__Noreturn] = ACTIONS(2534),
    [anon_sym__Nonnull] = ACTIONS(2534),
    [anon_sym_nullable] = ACTIONS(2534),
    [anon_sym__Complex] = ACTIONS(2534),
    [anon_sym__Nullable] = ACTIONS(2534),
    [anon_sym__Nullable_result] = ACTIONS(2534),
    [anon_sym__Null_unspecified] = ACTIONS(2534),
    [anon_sym___autoreleasing] = ACTIONS(2534),
    [anon_sym___block] = ACTIONS(2534),
    [anon_sym___bridge] = ACTIONS(2534),
    [anon_sym___bridge_retained] = ACTIONS(2534),
    [anon_sym___bridge_transfer] = ACTIONS(2534),
    [anon_sym___complex] = ACTIONS(2534),
    [anon_sym___const] = ACTIONS(2534),
    [anon_sym___imag] = ACTIONS(2534),
    [anon_sym___kindof] = ACTIONS(2534),
    [anon_sym___nonnull] = ACTIONS(2534),
    [anon_sym___nullable] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2534),
    [anon_sym___real] = ACTIONS(2534),
    [anon_sym___strong] = ACTIONS(2534),
    [anon_sym___unsafe_unretained] = ACTIONS(2534),
    [anon_sym___unused] = ACTIONS(2534),
    [anon_sym___weak] = ACTIONS(2534),
    [anon_sym_alignas] = ACTIONS(2534),
    [anon_sym__Alignas] = ACTIONS(2534),
    [sym_primitive_type] = ACTIONS(2534),
    [anon_sym_enum] = ACTIONS(2534),
    [anon_sym_struct] = ACTIONS(2534),
    [anon_sym_union] = ACTIONS(2534),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_else] = ACTIONS(2534),
    [anon_sym_switch] = ACTIONS(2534),
    [anon_sym_case] = ACTIONS(2534),
    [anon_sym_default] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_do] = ACTIONS(2534),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_in] = ACTIONS(2534),
    [anon_sym_return] = ACTIONS(2534),
    [anon_sym_break] = ACTIONS(2534),
    [anon_sym_continue] = ACTIONS(2534),
    [anon_sym_goto] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_sizeof] = ACTIONS(2534),
    [anon_sym___alignof__] = ACTIONS(2534),
    [anon_sym___alignof] = ACTIONS(2534),
    [anon_sym__alignof] = ACTIONS(2534),
    [anon_sym_alignof] = ACTIONS(2534),
    [anon_sym__Alignof] = ACTIONS(2534),
    [anon_sym_offsetof] = ACTIONS(2534),
    [anon_sym__Generic] = ACTIONS(2534),
    [anon_sym_asm] = ACTIONS(2534),
    [anon_sym___asm__] = ACTIONS(2534),
    [anon_sym___asm] = ACTIONS(2534),
    [sym_number_literal] = ACTIONS(2536),
    [anon_sym_L_SQUOTE] = ACTIONS(2536),
    [anon_sym_u_SQUOTE] = ACTIONS(2536),
    [anon_sym_U_SQUOTE] = ACTIONS(2536),
    [anon_sym_u8_SQUOTE] = ACTIONS(2536),
    [anon_sym_SQUOTE] = ACTIONS(2536),
    [anon_sym_AT] = ACTIONS(2534),
    [anon_sym_DQUOTE] = ACTIONS(2536),
    [anon_sym_L_DQUOTE] = ACTIONS(2536),
    [anon_sym_u_DQUOTE] = ACTIONS(2536),
    [anon_sym_U_DQUOTE] = ACTIONS(2536),
    [anon_sym_u8_DQUOTE] = ACTIONS(2536),
    [sym_true] = ACTIONS(2534),
    [sym_false] = ACTIONS(2534),
    [anon_sym_NULL] = ACTIONS(2534),
    [anon_sym_nullptr] = ACTIONS(2534),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2534),
    [anon_sym___typeof] = ACTIONS(2534),
    [anon_sym_typeof] = ACTIONS(2534),
    [anon_sym_ATimport] = ACTIONS(2536),
    [aux_sym_preproc_undef_token1] = ACTIONS(2534),
    [anon_sym_POUND] = ACTIONS(2534),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2534),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2534),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2534),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_API_AVAILABLE] = ACTIONS(2534),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_API_DEPRECATED] = ACTIONS(2534),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym___deprecated_msg] = ACTIONS(2534),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2534),
    [anon_sym_ATprotocol] = ACTIONS(2536),
    [anon_sym_ATinterface] = ACTIONS(2536),
    [anon_sym_ATimplementation] = ACTIONS(2536),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2536),
    [anon_sym_ATsynthesize] = ACTIONS(2536),
    [anon_sym_ATdynamic] = ACTIONS(2536),
    [anon_sym_ATtry] = ACTIONS(2536),
    [anon_sym___try] = ACTIONS(2534),
    [anon_sym_ATthrow] = ACTIONS(2536),
    [anon_sym_ATselector] = ACTIONS(2536),
    [anon_sym_ATavailable] = ACTIONS(2536),
    [anon_sym___builtin_available] = ACTIONS(2534),
    [anon_sym_va_arg] = ACTIONS(2534),
    [anon_sym_ATencode] = ACTIONS(2536),
    [anon_sym_ATsynchronized] = ACTIONS(2536),
    [anon_sym_BOOL] = ACTIONS(2534),
    [anon_sym_IMP] = ACTIONS(2534),
    [anon_sym_SEL] = ACTIONS(2534),
    [anon_sym_Class] = ACTIONS(2534),
    [anon_sym_id] = ACTIONS(2534),
  },
  [STATE(267)] = {
    [sym_identifier] = ACTIONS(2538),
    [aux_sym_preproc_include_token1] = ACTIONS(2538),
    [aux_sym_preproc_include_token2] = ACTIONS(2538),
    [aux_sym_preproc_def_token1] = ACTIONS(2538),
    [aux_sym_preproc_if_token1] = ACTIONS(2538),
    [aux_sym_preproc_if_token2] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2538),
    [aux_sym_preproc_else_token1] = ACTIONS(2538),
    [aux_sym_preproc_elif_token1] = ACTIONS(2538),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2538),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2538),
    [sym_preproc_directive] = ACTIONS(2538),
    [anon_sym_LPAREN2] = ACTIONS(2540),
    [anon_sym_BANG] = ACTIONS(2540),
    [anon_sym_TILDE] = ACTIONS(2540),
    [anon_sym_DASH] = ACTIONS(2538),
    [anon_sym_PLUS] = ACTIONS(2538),
    [anon_sym_STAR] = ACTIONS(2540),
    [anon_sym_CARET] = ACTIONS(2540),
    [anon_sym_AMP] = ACTIONS(2540),
    [anon_sym_SEMI] = ACTIONS(2540),
    [anon_sym___extension__] = ACTIONS(2538),
    [anon_sym_typedef] = ACTIONS(2538),
    [anon_sym_extern] = ACTIONS(2538),
    [anon_sym___attribute__] = ACTIONS(2538),
    [anon_sym___attribute] = ACTIONS(2538),
    [anon_sym_noreturn] = ACTIONS(2538),
    [anon_sym_LBRACK] = ACTIONS(2540),
    [anon_sym___declspec] = ACTIONS(2538),
    [anon_sym___cdecl] = ACTIONS(2538),
    [anon_sym___clrcall] = ACTIONS(2538),
    [anon_sym___stdcall] = ACTIONS(2538),
    [anon_sym___fastcall] = ACTIONS(2538),
    [anon_sym___thiscall] = ACTIONS(2538),
    [anon_sym___vectorcall] = ACTIONS(2538),
    [anon_sym_LBRACE] = ACTIONS(2540),
    [anon_sym_signed] = ACTIONS(2538),
    [anon_sym_unsigned] = ACTIONS(2538),
    [anon_sym_long] = ACTIONS(2538),
    [anon_sym_short] = ACTIONS(2538),
    [anon_sym_static] = ACTIONS(2538),
    [anon_sym_ATautoreleasepool] = ACTIONS(2540),
    [anon_sym_auto] = ACTIONS(2538),
    [anon_sym_register] = ACTIONS(2538),
    [anon_sym_inline] = ACTIONS(2538),
    [anon_sym___inline] = ACTIONS(2538),
    [anon_sym___inline__] = ACTIONS(2538),
    [anon_sym___forceinline] = ACTIONS(2538),
    [anon_sym_thread_local] = ACTIONS(2538),
    [anon_sym___thread] = ACTIONS(2538),
    [anon_sym_CG_EXTERN] = ACTIONS(2538),
    [anon_sym_CG_INLINE] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2538),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2538),
    [anon_sym_IBOutlet] = ACTIONS(2538),
    [anon_sym_IBInspectable] = ACTIONS(2538),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2538),
    [anon_sym_NS_INLINE] = ACTIONS(2538),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2538),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2538),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2538),
    [anon_sym_const] = ACTIONS(2538),
    [anon_sym_constexpr] = ACTIONS(2538),
    [anon_sym_volatile] = ACTIONS(2538),
    [anon_sym_restrict] = ACTIONS(2538),
    [anon_sym___restrict__] = ACTIONS(2538),
    [anon_sym__Atomic] = ACTIONS(2538),
    [anon_sym__Noreturn] = ACTIONS(2538),
    [anon_sym__Nonnull] = ACTIONS(2538),
    [anon_sym_nullable] = ACTIONS(2538),
    [anon_sym__Complex] = ACTIONS(2538),
    [anon_sym__Nullable] = ACTIONS(2538),
    [anon_sym__Nullable_result] = ACTIONS(2538),
    [anon_sym__Null_unspecified] = ACTIONS(2538),
    [anon_sym___autoreleasing] = ACTIONS(2538),
    [anon_sym___block] = ACTIONS(2538),
    [anon_sym___bridge] = ACTIONS(2538),
    [anon_sym___bridge_retained] = ACTIONS(2538),
    [anon_sym___bridge_transfer] = ACTIONS(2538),
    [anon_sym___complex] = ACTIONS(2538),
    [anon_sym___const] = ACTIONS(2538),
    [anon_sym___imag] = ACTIONS(2538),
    [anon_sym___kindof] = ACTIONS(2538),
    [anon_sym___nonnull] = ACTIONS(2538),
    [anon_sym___nullable] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2538),
    [anon_sym___real] = ACTIONS(2538),
    [anon_sym___strong] = ACTIONS(2538),
    [anon_sym___unsafe_unretained] = ACTIONS(2538),
    [anon_sym___unused] = ACTIONS(2538),
    [anon_sym___weak] = ACTIONS(2538),
    [anon_sym_alignas] = ACTIONS(2538),
    [anon_sym__Alignas] = ACTIONS(2538),
    [sym_primitive_type] = ACTIONS(2538),
    [anon_sym_enum] = ACTIONS(2538),
    [anon_sym_struct] = ACTIONS(2538),
    [anon_sym_union] = ACTIONS(2538),
    [anon_sym_if] = ACTIONS(2538),
    [anon_sym_else] = ACTIONS(2538),
    [anon_sym_switch] = ACTIONS(2538),
    [anon_sym_case] = ACTIONS(2538),
    [anon_sym_default] = ACTIONS(2538),
    [anon_sym_while] = ACTIONS(2538),
    [anon_sym_do] = ACTIONS(2538),
    [anon_sym_for] = ACTIONS(2538),
    [anon_sym_in] = ACTIONS(2538),
    [anon_sym_return] = ACTIONS(2538),
    [anon_sym_break] = ACTIONS(2538),
    [anon_sym_continue] = ACTIONS(2538),
    [anon_sym_goto] = ACTIONS(2538),
    [anon_sym_DASH_DASH] = ACTIONS(2540),
    [anon_sym_PLUS_PLUS] = ACTIONS(2540),
    [anon_sym_sizeof] = ACTIONS(2538),
    [anon_sym___alignof__] = ACTIONS(2538),
    [anon_sym___alignof] = ACTIONS(2538),
    [anon_sym__alignof] = ACTIONS(2538),
    [anon_sym_alignof] = ACTIONS(2538),
    [anon_sym__Alignof] = ACTIONS(2538),
    [anon_sym_offsetof] = ACTIONS(2538),
    [anon_sym__Generic] = ACTIONS(2538),
    [anon_sym_asm] = ACTIONS(2538),
    [anon_sym___asm__] = ACTIONS(2538),
    [anon_sym___asm] = ACTIONS(2538),
    [sym_number_literal] = ACTIONS(2540),
    [anon_sym_L_SQUOTE] = ACTIONS(2540),
    [anon_sym_u_SQUOTE] = ACTIONS(2540),
    [anon_sym_U_SQUOTE] = ACTIONS(2540),
    [anon_sym_u8_SQUOTE] = ACTIONS(2540),
    [anon_sym_SQUOTE] = ACTIONS(2540),
    [anon_sym_AT] = ACTIONS(2538),
    [anon_sym_DQUOTE] = ACTIONS(2540),
    [anon_sym_L_DQUOTE] = ACTIONS(2540),
    [anon_sym_u_DQUOTE] = ACTIONS(2540),
    [anon_sym_U_DQUOTE] = ACTIONS(2540),
    [anon_sym_u8_DQUOTE] = ACTIONS(2540),
    [sym_true] = ACTIONS(2538),
    [sym_false] = ACTIONS(2538),
    [anon_sym_NULL] = ACTIONS(2538),
    [anon_sym_nullptr] = ACTIONS(2538),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2538),
    [anon_sym___typeof] = ACTIONS(2538),
    [anon_sym_typeof] = ACTIONS(2538),
    [anon_sym_ATimport] = ACTIONS(2540),
    [aux_sym_preproc_undef_token1] = ACTIONS(2538),
    [anon_sym_POUND] = ACTIONS(2538),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2538),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2538),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2538),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_API_AVAILABLE] = ACTIONS(2538),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_API_DEPRECATED] = ACTIONS(2538),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym___deprecated_msg] = ACTIONS(2538),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2538),
    [anon_sym_ATprotocol] = ACTIONS(2540),
    [anon_sym_ATinterface] = ACTIONS(2540),
    [anon_sym_ATimplementation] = ACTIONS(2540),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2540),
    [anon_sym_ATsynthesize] = ACTIONS(2540),
    [anon_sym_ATdynamic] = ACTIONS(2540),
    [anon_sym_ATtry] = ACTIONS(2540),
    [anon_sym___try] = ACTIONS(2538),
    [anon_sym_ATthrow] = ACTIONS(2540),
    [anon_sym_ATselector] = ACTIONS(2540),
    [anon_sym_ATavailable] = ACTIONS(2540),
    [anon_sym___builtin_available] = ACTIONS(2538),
    [anon_sym_va_arg] = ACTIONS(2538),
    [anon_sym_ATencode] = ACTIONS(2540),
    [anon_sym_ATsynchronized] = ACTIONS(2540),
    [anon_sym_BOOL] = ACTIONS(2538),
    [anon_sym_IMP] = ACTIONS(2538),
    [anon_sym_SEL] = ACTIONS(2538),
    [anon_sym_Class] = ACTIONS(2538),
    [anon_sym_id] = ACTIONS(2538),
  },
  [STATE(268)] = {
    [sym_identifier] = ACTIONS(2542),
    [aux_sym_preproc_include_token1] = ACTIONS(2542),
    [aux_sym_preproc_include_token2] = ACTIONS(2542),
    [aux_sym_preproc_def_token1] = ACTIONS(2542),
    [aux_sym_preproc_if_token1] = ACTIONS(2542),
    [aux_sym_preproc_if_token2] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2542),
    [aux_sym_preproc_else_token1] = ACTIONS(2542),
    [aux_sym_preproc_elif_token1] = ACTIONS(2542),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2542),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2542),
    [sym_preproc_directive] = ACTIONS(2542),
    [anon_sym_LPAREN2] = ACTIONS(2544),
    [anon_sym_BANG] = ACTIONS(2544),
    [anon_sym_TILDE] = ACTIONS(2544),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_PLUS] = ACTIONS(2542),
    [anon_sym_STAR] = ACTIONS(2544),
    [anon_sym_CARET] = ACTIONS(2544),
    [anon_sym_AMP] = ACTIONS(2544),
    [anon_sym_SEMI] = ACTIONS(2544),
    [anon_sym___extension__] = ACTIONS(2542),
    [anon_sym_typedef] = ACTIONS(2542),
    [anon_sym_extern] = ACTIONS(2542),
    [anon_sym___attribute__] = ACTIONS(2542),
    [anon_sym___attribute] = ACTIONS(2542),
    [anon_sym_noreturn] = ACTIONS(2542),
    [anon_sym_LBRACK] = ACTIONS(2544),
    [anon_sym___declspec] = ACTIONS(2542),
    [anon_sym___cdecl] = ACTIONS(2542),
    [anon_sym___clrcall] = ACTIONS(2542),
    [anon_sym___stdcall] = ACTIONS(2542),
    [anon_sym___fastcall] = ACTIONS(2542),
    [anon_sym___thiscall] = ACTIONS(2542),
    [anon_sym___vectorcall] = ACTIONS(2542),
    [anon_sym_LBRACE] = ACTIONS(2544),
    [anon_sym_signed] = ACTIONS(2542),
    [anon_sym_unsigned] = ACTIONS(2542),
    [anon_sym_long] = ACTIONS(2542),
    [anon_sym_short] = ACTIONS(2542),
    [anon_sym_static] = ACTIONS(2542),
    [anon_sym_ATautoreleasepool] = ACTIONS(2544),
    [anon_sym_auto] = ACTIONS(2542),
    [anon_sym_register] = ACTIONS(2542),
    [anon_sym_inline] = ACTIONS(2542),
    [anon_sym___inline] = ACTIONS(2542),
    [anon_sym___inline__] = ACTIONS(2542),
    [anon_sym___forceinline] = ACTIONS(2542),
    [anon_sym_thread_local] = ACTIONS(2542),
    [anon_sym___thread] = ACTIONS(2542),
    [anon_sym_CG_EXTERN] = ACTIONS(2542),
    [anon_sym_CG_INLINE] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2542),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2542),
    [anon_sym_IBOutlet] = ACTIONS(2542),
    [anon_sym_IBInspectable] = ACTIONS(2542),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2542),
    [anon_sym_NS_INLINE] = ACTIONS(2542),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2542),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2542),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2542),
    [anon_sym_const] = ACTIONS(2542),
    [anon_sym_constexpr] = ACTIONS(2542),
    [anon_sym_volatile] = ACTIONS(2542),
    [anon_sym_restrict] = ACTIONS(2542),
    [anon_sym___restrict__] = ACTIONS(2542),
    [anon_sym__Atomic] = ACTIONS(2542),
    [anon_sym__Noreturn] = ACTIONS(2542),
    [anon_sym__Nonnull] = ACTIONS(2542),
    [anon_sym_nullable] = ACTIONS(2542),
    [anon_sym__Complex] = ACTIONS(2542),
    [anon_sym__Nullable] = ACTIONS(2542),
    [anon_sym__Nullable_result] = ACTIONS(2542),
    [anon_sym__Null_unspecified] = ACTIONS(2542),
    [anon_sym___autoreleasing] = ACTIONS(2542),
    [anon_sym___block] = ACTIONS(2542),
    [anon_sym___bridge] = ACTIONS(2542),
    [anon_sym___bridge_retained] = ACTIONS(2542),
    [anon_sym___bridge_transfer] = ACTIONS(2542),
    [anon_sym___complex] = ACTIONS(2542),
    [anon_sym___const] = ACTIONS(2542),
    [anon_sym___imag] = ACTIONS(2542),
    [anon_sym___kindof] = ACTIONS(2542),
    [anon_sym___nonnull] = ACTIONS(2542),
    [anon_sym___nullable] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2542),
    [anon_sym___real] = ACTIONS(2542),
    [anon_sym___strong] = ACTIONS(2542),
    [anon_sym___unsafe_unretained] = ACTIONS(2542),
    [anon_sym___unused] = ACTIONS(2542),
    [anon_sym___weak] = ACTIONS(2542),
    [anon_sym_alignas] = ACTIONS(2542),
    [anon_sym__Alignas] = ACTIONS(2542),
    [sym_primitive_type] = ACTIONS(2542),
    [anon_sym_enum] = ACTIONS(2542),
    [anon_sym_struct] = ACTIONS(2542),
    [anon_sym_union] = ACTIONS(2542),
    [anon_sym_if] = ACTIONS(2542),
    [anon_sym_else] = ACTIONS(2542),
    [anon_sym_switch] = ACTIONS(2542),
    [anon_sym_case] = ACTIONS(2542),
    [anon_sym_default] = ACTIONS(2542),
    [anon_sym_while] = ACTIONS(2542),
    [anon_sym_do] = ACTIONS(2542),
    [anon_sym_for] = ACTIONS(2542),
    [anon_sym_in] = ACTIONS(2542),
    [anon_sym_return] = ACTIONS(2542),
    [anon_sym_break] = ACTIONS(2542),
    [anon_sym_continue] = ACTIONS(2542),
    [anon_sym_goto] = ACTIONS(2542),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_sizeof] = ACTIONS(2542),
    [anon_sym___alignof__] = ACTIONS(2542),
    [anon_sym___alignof] = ACTIONS(2542),
    [anon_sym__alignof] = ACTIONS(2542),
    [anon_sym_alignof] = ACTIONS(2542),
    [anon_sym__Alignof] = ACTIONS(2542),
    [anon_sym_offsetof] = ACTIONS(2542),
    [anon_sym__Generic] = ACTIONS(2542),
    [anon_sym_asm] = ACTIONS(2542),
    [anon_sym___asm__] = ACTIONS(2542),
    [anon_sym___asm] = ACTIONS(2542),
    [sym_number_literal] = ACTIONS(2544),
    [anon_sym_L_SQUOTE] = ACTIONS(2544),
    [anon_sym_u_SQUOTE] = ACTIONS(2544),
    [anon_sym_U_SQUOTE] = ACTIONS(2544),
    [anon_sym_u8_SQUOTE] = ACTIONS(2544),
    [anon_sym_SQUOTE] = ACTIONS(2544),
    [anon_sym_AT] = ACTIONS(2542),
    [anon_sym_DQUOTE] = ACTIONS(2544),
    [anon_sym_L_DQUOTE] = ACTIONS(2544),
    [anon_sym_u_DQUOTE] = ACTIONS(2544),
    [anon_sym_U_DQUOTE] = ACTIONS(2544),
    [anon_sym_u8_DQUOTE] = ACTIONS(2544),
    [sym_true] = ACTIONS(2542),
    [sym_false] = ACTIONS(2542),
    [anon_sym_NULL] = ACTIONS(2542),
    [anon_sym_nullptr] = ACTIONS(2542),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2542),
    [anon_sym___typeof] = ACTIONS(2542),
    [anon_sym_typeof] = ACTIONS(2542),
    [anon_sym_ATimport] = ACTIONS(2544),
    [aux_sym_preproc_undef_token1] = ACTIONS(2542),
    [anon_sym_POUND] = ACTIONS(2542),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2542),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2542),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2542),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_API_AVAILABLE] = ACTIONS(2542),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_API_DEPRECATED] = ACTIONS(2542),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym___deprecated_msg] = ACTIONS(2542),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2542),
    [anon_sym_ATprotocol] = ACTIONS(2544),
    [anon_sym_ATinterface] = ACTIONS(2544),
    [anon_sym_ATimplementation] = ACTIONS(2544),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2544),
    [anon_sym_ATsynthesize] = ACTIONS(2544),
    [anon_sym_ATdynamic] = ACTIONS(2544),
    [anon_sym_ATtry] = ACTIONS(2544),
    [anon_sym___try] = ACTIONS(2542),
    [anon_sym_ATthrow] = ACTIONS(2544),
    [anon_sym_ATselector] = ACTIONS(2544),
    [anon_sym_ATavailable] = ACTIONS(2544),
    [anon_sym___builtin_available] = ACTIONS(2542),
    [anon_sym_va_arg] = ACTIONS(2542),
    [anon_sym_ATencode] = ACTIONS(2544),
    [anon_sym_ATsynchronized] = ACTIONS(2544),
    [anon_sym_BOOL] = ACTIONS(2542),
    [anon_sym_IMP] = ACTIONS(2542),
    [anon_sym_SEL] = ACTIONS(2542),
    [anon_sym_Class] = ACTIONS(2542),
    [anon_sym_id] = ACTIONS(2542),
  },
  [STATE(269)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token2] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [aux_sym_preproc_else_token1] = ACTIONS(2546),
    [aux_sym_preproc_elif_token1] = ACTIONS(2546),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(270)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token2] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [aux_sym_preproc_else_token1] = ACTIONS(2546),
    [aux_sym_preproc_elif_token1] = ACTIONS(2546),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(271)] = {
    [sym_identifier] = ACTIONS(2550),
    [aux_sym_preproc_include_token1] = ACTIONS(2550),
    [aux_sym_preproc_include_token2] = ACTIONS(2550),
    [aux_sym_preproc_def_token1] = ACTIONS(2550),
    [aux_sym_preproc_if_token1] = ACTIONS(2550),
    [aux_sym_preproc_if_token2] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2550),
    [aux_sym_preproc_else_token1] = ACTIONS(2550),
    [aux_sym_preproc_elif_token1] = ACTIONS(2550),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2550),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2550),
    [sym_preproc_directive] = ACTIONS(2550),
    [anon_sym_LPAREN2] = ACTIONS(2552),
    [anon_sym_BANG] = ACTIONS(2552),
    [anon_sym_TILDE] = ACTIONS(2552),
    [anon_sym_DASH] = ACTIONS(2550),
    [anon_sym_PLUS] = ACTIONS(2550),
    [anon_sym_STAR] = ACTIONS(2552),
    [anon_sym_CARET] = ACTIONS(2552),
    [anon_sym_AMP] = ACTIONS(2552),
    [anon_sym_SEMI] = ACTIONS(2552),
    [anon_sym___extension__] = ACTIONS(2550),
    [anon_sym_typedef] = ACTIONS(2550),
    [anon_sym_extern] = ACTIONS(2550),
    [anon_sym___attribute__] = ACTIONS(2550),
    [anon_sym___attribute] = ACTIONS(2550),
    [anon_sym_noreturn] = ACTIONS(2550),
    [anon_sym_LBRACK] = ACTIONS(2552),
    [anon_sym___declspec] = ACTIONS(2550),
    [anon_sym___cdecl] = ACTIONS(2550),
    [anon_sym___clrcall] = ACTIONS(2550),
    [anon_sym___stdcall] = ACTIONS(2550),
    [anon_sym___fastcall] = ACTIONS(2550),
    [anon_sym___thiscall] = ACTIONS(2550),
    [anon_sym___vectorcall] = ACTIONS(2550),
    [anon_sym_LBRACE] = ACTIONS(2552),
    [anon_sym_signed] = ACTIONS(2550),
    [anon_sym_unsigned] = ACTIONS(2550),
    [anon_sym_long] = ACTIONS(2550),
    [anon_sym_short] = ACTIONS(2550),
    [anon_sym_static] = ACTIONS(2550),
    [anon_sym_ATautoreleasepool] = ACTIONS(2552),
    [anon_sym_auto] = ACTIONS(2550),
    [anon_sym_register] = ACTIONS(2550),
    [anon_sym_inline] = ACTIONS(2550),
    [anon_sym___inline] = ACTIONS(2550),
    [anon_sym___inline__] = ACTIONS(2550),
    [anon_sym___forceinline] = ACTIONS(2550),
    [anon_sym_thread_local] = ACTIONS(2550),
    [anon_sym___thread] = ACTIONS(2550),
    [anon_sym_CG_EXTERN] = ACTIONS(2550),
    [anon_sym_CG_INLINE] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2550),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2550),
    [anon_sym_IBOutlet] = ACTIONS(2550),
    [anon_sym_IBInspectable] = ACTIONS(2550),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2550),
    [anon_sym_NS_INLINE] = ACTIONS(2550),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2550),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2550),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2550),
    [anon_sym_const] = ACTIONS(2550),
    [anon_sym_constexpr] = ACTIONS(2550),
    [anon_sym_volatile] = ACTIONS(2550),
    [anon_sym_restrict] = ACTIONS(2550),
    [anon_sym___restrict__] = ACTIONS(2550),
    [anon_sym__Atomic] = ACTIONS(2550),
    [anon_sym__Noreturn] = ACTIONS(2550),
    [anon_sym__Nonnull] = ACTIONS(2550),
    [anon_sym_nullable] = ACTIONS(2550),
    [anon_sym__Complex] = ACTIONS(2550),
    [anon_sym__Nullable] = ACTIONS(2550),
    [anon_sym__Nullable_result] = ACTIONS(2550),
    [anon_sym__Null_unspecified] = ACTIONS(2550),
    [anon_sym___autoreleasing] = ACTIONS(2550),
    [anon_sym___block] = ACTIONS(2550),
    [anon_sym___bridge] = ACTIONS(2550),
    [anon_sym___bridge_retained] = ACTIONS(2550),
    [anon_sym___bridge_transfer] = ACTIONS(2550),
    [anon_sym___complex] = ACTIONS(2550),
    [anon_sym___const] = ACTIONS(2550),
    [anon_sym___imag] = ACTIONS(2550),
    [anon_sym___kindof] = ACTIONS(2550),
    [anon_sym___nonnull] = ACTIONS(2550),
    [anon_sym___nullable] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2550),
    [anon_sym___real] = ACTIONS(2550),
    [anon_sym___strong] = ACTIONS(2550),
    [anon_sym___unsafe_unretained] = ACTIONS(2550),
    [anon_sym___unused] = ACTIONS(2550),
    [anon_sym___weak] = ACTIONS(2550),
    [anon_sym_alignas] = ACTIONS(2550),
    [anon_sym__Alignas] = ACTIONS(2550),
    [sym_primitive_type] = ACTIONS(2550),
    [anon_sym_enum] = ACTIONS(2550),
    [anon_sym_struct] = ACTIONS(2550),
    [anon_sym_union] = ACTIONS(2550),
    [anon_sym_if] = ACTIONS(2550),
    [anon_sym_else] = ACTIONS(2550),
    [anon_sym_switch] = ACTIONS(2550),
    [anon_sym_case] = ACTIONS(2550),
    [anon_sym_default] = ACTIONS(2550),
    [anon_sym_while] = ACTIONS(2550),
    [anon_sym_do] = ACTIONS(2550),
    [anon_sym_for] = ACTIONS(2550),
    [anon_sym_in] = ACTIONS(2550),
    [anon_sym_return] = ACTIONS(2550),
    [anon_sym_break] = ACTIONS(2550),
    [anon_sym_continue] = ACTIONS(2550),
    [anon_sym_goto] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2552),
    [anon_sym_PLUS_PLUS] = ACTIONS(2552),
    [anon_sym_sizeof] = ACTIONS(2550),
    [anon_sym___alignof__] = ACTIONS(2550),
    [anon_sym___alignof] = ACTIONS(2550),
    [anon_sym__alignof] = ACTIONS(2550),
    [anon_sym_alignof] = ACTIONS(2550),
    [anon_sym__Alignof] = ACTIONS(2550),
    [anon_sym_offsetof] = ACTIONS(2550),
    [anon_sym__Generic] = ACTIONS(2550),
    [anon_sym_asm] = ACTIONS(2550),
    [anon_sym___asm__] = ACTIONS(2550),
    [anon_sym___asm] = ACTIONS(2550),
    [sym_number_literal] = ACTIONS(2552),
    [anon_sym_L_SQUOTE] = ACTIONS(2552),
    [anon_sym_u_SQUOTE] = ACTIONS(2552),
    [anon_sym_U_SQUOTE] = ACTIONS(2552),
    [anon_sym_u8_SQUOTE] = ACTIONS(2552),
    [anon_sym_SQUOTE] = ACTIONS(2552),
    [anon_sym_AT] = ACTIONS(2550),
    [anon_sym_DQUOTE] = ACTIONS(2552),
    [anon_sym_L_DQUOTE] = ACTIONS(2552),
    [anon_sym_u_DQUOTE] = ACTIONS(2552),
    [anon_sym_U_DQUOTE] = ACTIONS(2552),
    [anon_sym_u8_DQUOTE] = ACTIONS(2552),
    [sym_true] = ACTIONS(2550),
    [sym_false] = ACTIONS(2550),
    [anon_sym_NULL] = ACTIONS(2550),
    [anon_sym_nullptr] = ACTIONS(2550),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2550),
    [anon_sym___typeof] = ACTIONS(2550),
    [anon_sym_typeof] = ACTIONS(2550),
    [anon_sym_ATimport] = ACTIONS(2552),
    [aux_sym_preproc_undef_token1] = ACTIONS(2550),
    [anon_sym_POUND] = ACTIONS(2550),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2550),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2550),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2550),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_API_AVAILABLE] = ACTIONS(2550),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_API_DEPRECATED] = ACTIONS(2550),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym___deprecated_msg] = ACTIONS(2550),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2550),
    [anon_sym_ATprotocol] = ACTIONS(2552),
    [anon_sym_ATinterface] = ACTIONS(2552),
    [anon_sym_ATimplementation] = ACTIONS(2552),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2552),
    [anon_sym_ATsynthesize] = ACTIONS(2552),
    [anon_sym_ATdynamic] = ACTIONS(2552),
    [anon_sym_ATtry] = ACTIONS(2552),
    [anon_sym___try] = ACTIONS(2550),
    [anon_sym_ATthrow] = ACTIONS(2552),
    [anon_sym_ATselector] = ACTIONS(2552),
    [anon_sym_ATavailable] = ACTIONS(2552),
    [anon_sym___builtin_available] = ACTIONS(2550),
    [anon_sym_va_arg] = ACTIONS(2550),
    [anon_sym_ATencode] = ACTIONS(2552),
    [anon_sym_ATsynchronized] = ACTIONS(2552),
    [anon_sym_BOOL] = ACTIONS(2550),
    [anon_sym_IMP] = ACTIONS(2550),
    [anon_sym_SEL] = ACTIONS(2550),
    [anon_sym_Class] = ACTIONS(2550),
    [anon_sym_id] = ACTIONS(2550),
  },
  [STATE(272)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token2] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [aux_sym_preproc_else_token1] = ACTIONS(2554),
    [aux_sym_preproc_elif_token1] = ACTIONS(2554),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(273)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token2] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [aux_sym_preproc_else_token1] = ACTIONS(2554),
    [aux_sym_preproc_elif_token1] = ACTIONS(2554),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(274)] = {
    [sym_identifier] = ACTIONS(2558),
    [aux_sym_preproc_include_token1] = ACTIONS(2558),
    [aux_sym_preproc_include_token2] = ACTIONS(2558),
    [aux_sym_preproc_def_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token2] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2558),
    [aux_sym_preproc_else_token1] = ACTIONS(2558),
    [aux_sym_preproc_elif_token1] = ACTIONS(2558),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2558),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2558),
    [sym_preproc_directive] = ACTIONS(2558),
    [anon_sym_LPAREN2] = ACTIONS(2560),
    [anon_sym_BANG] = ACTIONS(2560),
    [anon_sym_TILDE] = ACTIONS(2560),
    [anon_sym_DASH] = ACTIONS(2558),
    [anon_sym_PLUS] = ACTIONS(2558),
    [anon_sym_STAR] = ACTIONS(2560),
    [anon_sym_CARET] = ACTIONS(2560),
    [anon_sym_AMP] = ACTIONS(2560),
    [anon_sym_SEMI] = ACTIONS(2560),
    [anon_sym___extension__] = ACTIONS(2558),
    [anon_sym_typedef] = ACTIONS(2558),
    [anon_sym_extern] = ACTIONS(2558),
    [anon_sym___attribute__] = ACTIONS(2558),
    [anon_sym___attribute] = ACTIONS(2558),
    [anon_sym_noreturn] = ACTIONS(2558),
    [anon_sym_LBRACK] = ACTIONS(2560),
    [anon_sym___declspec] = ACTIONS(2558),
    [anon_sym___cdecl] = ACTIONS(2558),
    [anon_sym___clrcall] = ACTIONS(2558),
    [anon_sym___stdcall] = ACTIONS(2558),
    [anon_sym___fastcall] = ACTIONS(2558),
    [anon_sym___thiscall] = ACTIONS(2558),
    [anon_sym___vectorcall] = ACTIONS(2558),
    [anon_sym_LBRACE] = ACTIONS(2560),
    [anon_sym_signed] = ACTIONS(2558),
    [anon_sym_unsigned] = ACTIONS(2558),
    [anon_sym_long] = ACTIONS(2558),
    [anon_sym_short] = ACTIONS(2558),
    [anon_sym_static] = ACTIONS(2558),
    [anon_sym_ATautoreleasepool] = ACTIONS(2560),
    [anon_sym_auto] = ACTIONS(2558),
    [anon_sym_register] = ACTIONS(2558),
    [anon_sym_inline] = ACTIONS(2558),
    [anon_sym___inline] = ACTIONS(2558),
    [anon_sym___inline__] = ACTIONS(2558),
    [anon_sym___forceinline] = ACTIONS(2558),
    [anon_sym_thread_local] = ACTIONS(2558),
    [anon_sym___thread] = ACTIONS(2558),
    [anon_sym_CG_EXTERN] = ACTIONS(2558),
    [anon_sym_CG_INLINE] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2558),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2558),
    [anon_sym_IBOutlet] = ACTIONS(2558),
    [anon_sym_IBInspectable] = ACTIONS(2558),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2558),
    [anon_sym_NS_INLINE] = ACTIONS(2558),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2558),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2558),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2558),
    [anon_sym_const] = ACTIONS(2558),
    [anon_sym_constexpr] = ACTIONS(2558),
    [anon_sym_volatile] = ACTIONS(2558),
    [anon_sym_restrict] = ACTIONS(2558),
    [anon_sym___restrict__] = ACTIONS(2558),
    [anon_sym__Atomic] = ACTIONS(2558),
    [anon_sym__Noreturn] = ACTIONS(2558),
    [anon_sym__Nonnull] = ACTIONS(2558),
    [anon_sym_nullable] = ACTIONS(2558),
    [anon_sym__Complex] = ACTIONS(2558),
    [anon_sym__Nullable] = ACTIONS(2558),
    [anon_sym__Nullable_result] = ACTIONS(2558),
    [anon_sym__Null_unspecified] = ACTIONS(2558),
    [anon_sym___autoreleasing] = ACTIONS(2558),
    [anon_sym___block] = ACTIONS(2558),
    [anon_sym___bridge] = ACTIONS(2558),
    [anon_sym___bridge_retained] = ACTIONS(2558),
    [anon_sym___bridge_transfer] = ACTIONS(2558),
    [anon_sym___complex] = ACTIONS(2558),
    [anon_sym___const] = ACTIONS(2558),
    [anon_sym___imag] = ACTIONS(2558),
    [anon_sym___kindof] = ACTIONS(2558),
    [anon_sym___nonnull] = ACTIONS(2558),
    [anon_sym___nullable] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2558),
    [anon_sym___real] = ACTIONS(2558),
    [anon_sym___strong] = ACTIONS(2558),
    [anon_sym___unsafe_unretained] = ACTIONS(2558),
    [anon_sym___unused] = ACTIONS(2558),
    [anon_sym___weak] = ACTIONS(2558),
    [anon_sym_alignas] = ACTIONS(2558),
    [anon_sym__Alignas] = ACTIONS(2558),
    [sym_primitive_type] = ACTIONS(2558),
    [anon_sym_enum] = ACTIONS(2558),
    [anon_sym_struct] = ACTIONS(2558),
    [anon_sym_union] = ACTIONS(2558),
    [anon_sym_if] = ACTIONS(2558),
    [anon_sym_else] = ACTIONS(2558),
    [anon_sym_switch] = ACTIONS(2558),
    [anon_sym_case] = ACTIONS(2558),
    [anon_sym_default] = ACTIONS(2558),
    [anon_sym_while] = ACTIONS(2558),
    [anon_sym_do] = ACTIONS(2558),
    [anon_sym_for] = ACTIONS(2558),
    [anon_sym_in] = ACTIONS(2558),
    [anon_sym_return] = ACTIONS(2558),
    [anon_sym_break] = ACTIONS(2558),
    [anon_sym_continue] = ACTIONS(2558),
    [anon_sym_goto] = ACTIONS(2558),
    [anon_sym_DASH_DASH] = ACTIONS(2560),
    [anon_sym_PLUS_PLUS] = ACTIONS(2560),
    [anon_sym_sizeof] = ACTIONS(2558),
    [anon_sym___alignof__] = ACTIONS(2558),
    [anon_sym___alignof] = ACTIONS(2558),
    [anon_sym__alignof] = ACTIONS(2558),
    [anon_sym_alignof] = ACTIONS(2558),
    [anon_sym__Alignof] = ACTIONS(2558),
    [anon_sym_offsetof] = ACTIONS(2558),
    [anon_sym__Generic] = ACTIONS(2558),
    [anon_sym_asm] = ACTIONS(2558),
    [anon_sym___asm__] = ACTIONS(2558),
    [anon_sym___asm] = ACTIONS(2558),
    [sym_number_literal] = ACTIONS(2560),
    [anon_sym_L_SQUOTE] = ACTIONS(2560),
    [anon_sym_u_SQUOTE] = ACTIONS(2560),
    [anon_sym_U_SQUOTE] = ACTIONS(2560),
    [anon_sym_u8_SQUOTE] = ACTIONS(2560),
    [anon_sym_SQUOTE] = ACTIONS(2560),
    [anon_sym_AT] = ACTIONS(2558),
    [anon_sym_DQUOTE] = ACTIONS(2560),
    [anon_sym_L_DQUOTE] = ACTIONS(2560),
    [anon_sym_u_DQUOTE] = ACTIONS(2560),
    [anon_sym_U_DQUOTE] = ACTIONS(2560),
    [anon_sym_u8_DQUOTE] = ACTIONS(2560),
    [sym_true] = ACTIONS(2558),
    [sym_false] = ACTIONS(2558),
    [anon_sym_NULL] = ACTIONS(2558),
    [anon_sym_nullptr] = ACTIONS(2558),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2558),
    [anon_sym___typeof] = ACTIONS(2558),
    [anon_sym_typeof] = ACTIONS(2558),
    [anon_sym_ATimport] = ACTIONS(2560),
    [aux_sym_preproc_undef_token1] = ACTIONS(2558),
    [anon_sym_POUND] = ACTIONS(2558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2558),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2558),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2558),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_API_AVAILABLE] = ACTIONS(2558),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_API_DEPRECATED] = ACTIONS(2558),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym___deprecated_msg] = ACTIONS(2558),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2558),
    [anon_sym_ATprotocol] = ACTIONS(2560),
    [anon_sym_ATinterface] = ACTIONS(2560),
    [anon_sym_ATimplementation] = ACTIONS(2560),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2560),
    [anon_sym_ATsynthesize] = ACTIONS(2560),
    [anon_sym_ATdynamic] = ACTIONS(2560),
    [anon_sym_ATtry] = ACTIONS(2560),
    [anon_sym___try] = ACTIONS(2558),
    [anon_sym_ATthrow] = ACTIONS(2560),
    [anon_sym_ATselector] = ACTIONS(2560),
    [anon_sym_ATavailable] = ACTIONS(2560),
    [anon_sym___builtin_available] = ACTIONS(2558),
    [anon_sym_va_arg] = ACTIONS(2558),
    [anon_sym_ATencode] = ACTIONS(2560),
    [anon_sym_ATsynchronized] = ACTIONS(2560),
    [anon_sym_BOOL] = ACTIONS(2558),
    [anon_sym_IMP] = ACTIONS(2558),
    [anon_sym_SEL] = ACTIONS(2558),
    [anon_sym_Class] = ACTIONS(2558),
    [anon_sym_id] = ACTIONS(2558),
  },
  [STATE(275)] = {
    [sym_catch_clause] = STATE(275),
    [aux_sym_try_statement_repeat1] = STATE(275),
    [sym_identifier] = ACTIONS(2318),
    [aux_sym_preproc_include_token1] = ACTIONS(2318),
    [aux_sym_preproc_include_token2] = ACTIONS(2318),
    [aux_sym_preproc_def_token1] = ACTIONS(2318),
    [aux_sym_preproc_if_token1] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2318),
    [sym_preproc_directive] = ACTIONS(2318),
    [anon_sym_LPAREN2] = ACTIONS(2320),
    [anon_sym_BANG] = ACTIONS(2320),
    [anon_sym_TILDE] = ACTIONS(2320),
    [anon_sym_DASH] = ACTIONS(2318),
    [anon_sym_PLUS] = ACTIONS(2318),
    [anon_sym_STAR] = ACTIONS(2320),
    [anon_sym_CARET] = ACTIONS(2320),
    [anon_sym_AMP] = ACTIONS(2320),
    [anon_sym_SEMI] = ACTIONS(2320),
    [anon_sym___extension__] = ACTIONS(2318),
    [anon_sym_typedef] = ACTIONS(2318),
    [anon_sym_extern] = ACTIONS(2318),
    [anon_sym___attribute__] = ACTIONS(2318),
    [anon_sym___attribute] = ACTIONS(2318),
    [anon_sym_noreturn] = ACTIONS(2318),
    [anon_sym_LBRACK] = ACTIONS(2320),
    [anon_sym___declspec] = ACTIONS(2318),
    [anon_sym___cdecl] = ACTIONS(2318),
    [anon_sym___clrcall] = ACTIONS(2318),
    [anon_sym___stdcall] = ACTIONS(2318),
    [anon_sym___fastcall] = ACTIONS(2318),
    [anon_sym___thiscall] = ACTIONS(2318),
    [anon_sym___vectorcall] = ACTIONS(2318),
    [anon_sym_LBRACE] = ACTIONS(2320),
    [anon_sym_RBRACE] = ACTIONS(2320),
    [anon_sym_signed] = ACTIONS(2318),
    [anon_sym_unsigned] = ACTIONS(2318),
    [anon_sym_long] = ACTIONS(2318),
    [anon_sym_short] = ACTIONS(2318),
    [anon_sym_static] = ACTIONS(2318),
    [anon_sym_ATautoreleasepool] = ACTIONS(2320),
    [anon_sym_auto] = ACTIONS(2318),
    [anon_sym_register] = ACTIONS(2318),
    [anon_sym_inline] = ACTIONS(2318),
    [anon_sym___inline] = ACTIONS(2318),
    [anon_sym___inline__] = ACTIONS(2318),
    [anon_sym___forceinline] = ACTIONS(2318),
    [anon_sym_thread_local] = ACTIONS(2318),
    [anon_sym___thread] = ACTIONS(2318),
    [anon_sym_CG_EXTERN] = ACTIONS(2318),
    [anon_sym_CG_INLINE] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2318),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2318),
    [anon_sym_IBOutlet] = ACTIONS(2318),
    [anon_sym_IBInspectable] = ACTIONS(2318),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2318),
    [anon_sym_NS_INLINE] = ACTIONS(2318),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2318),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2318),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2318),
    [anon_sym_const] = ACTIONS(2318),
    [anon_sym_constexpr] = ACTIONS(2318),
    [anon_sym_volatile] = ACTIONS(2318),
    [anon_sym_restrict] = ACTIONS(2318),
    [anon_sym___restrict__] = ACTIONS(2318),
    [anon_sym__Atomic] = ACTIONS(2318),
    [anon_sym__Noreturn] = ACTIONS(2318),
    [anon_sym__Nonnull] = ACTIONS(2318),
    [anon_sym_nullable] = ACTIONS(2318),
    [anon_sym__Complex] = ACTIONS(2318),
    [anon_sym__Nullable] = ACTIONS(2318),
    [anon_sym__Nullable_result] = ACTIONS(2318),
    [anon_sym__Null_unspecified] = ACTIONS(2318),
    [anon_sym___autoreleasing] = ACTIONS(2318),
    [anon_sym___block] = ACTIONS(2318),
    [anon_sym___bridge] = ACTIONS(2318),
    [anon_sym___bridge_retained] = ACTIONS(2318),
    [anon_sym___bridge_transfer] = ACTIONS(2318),
    [anon_sym___complex] = ACTIONS(2318),
    [anon_sym___const] = ACTIONS(2318),
    [anon_sym___imag] = ACTIONS(2318),
    [anon_sym___kindof] = ACTIONS(2318),
    [anon_sym___nonnull] = ACTIONS(2318),
    [anon_sym___nullable] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2318),
    [anon_sym___real] = ACTIONS(2318),
    [anon_sym___strong] = ACTIONS(2318),
    [anon_sym___unsafe_unretained] = ACTIONS(2318),
    [anon_sym___unused] = ACTIONS(2318),
    [anon_sym___weak] = ACTIONS(2318),
    [anon_sym_alignas] = ACTIONS(2318),
    [anon_sym__Alignas] = ACTIONS(2318),
    [sym_primitive_type] = ACTIONS(2318),
    [anon_sym_enum] = ACTIONS(2318),
    [anon_sym_struct] = ACTIONS(2318),
    [anon_sym_union] = ACTIONS(2318),
    [anon_sym_if] = ACTIONS(2318),
    [anon_sym_else] = ACTIONS(2318),
    [anon_sym_switch] = ACTIONS(2318),
    [anon_sym_case] = ACTIONS(2318),
    [anon_sym_default] = ACTIONS(2318),
    [anon_sym_while] = ACTIONS(2318),
    [anon_sym_do] = ACTIONS(2318),
    [anon_sym_for] = ACTIONS(2318),
    [anon_sym_in] = ACTIONS(2318),
    [anon_sym_return] = ACTIONS(2318),
    [anon_sym_break] = ACTIONS(2318),
    [anon_sym_continue] = ACTIONS(2318),
    [anon_sym_goto] = ACTIONS(2318),
    [anon_sym_DASH_DASH] = ACTIONS(2320),
    [anon_sym_PLUS_PLUS] = ACTIONS(2320),
    [anon_sym_sizeof] = ACTIONS(2318),
    [anon_sym___alignof__] = ACTIONS(2318),
    [anon_sym___alignof] = ACTIONS(2318),
    [anon_sym__alignof] = ACTIONS(2318),
    [anon_sym_alignof] = ACTIONS(2318),
    [anon_sym__Alignof] = ACTIONS(2318),
    [anon_sym_offsetof] = ACTIONS(2318),
    [anon_sym__Generic] = ACTIONS(2318),
    [anon_sym_asm] = ACTIONS(2318),
    [anon_sym___asm__] = ACTIONS(2318),
    [anon_sym___asm] = ACTIONS(2318),
    [sym_number_literal] = ACTIONS(2320),
    [anon_sym_L_SQUOTE] = ACTIONS(2320),
    [anon_sym_u_SQUOTE] = ACTIONS(2320),
    [anon_sym_U_SQUOTE] = ACTIONS(2320),
    [anon_sym_u8_SQUOTE] = ACTIONS(2320),
    [anon_sym_SQUOTE] = ACTIONS(2320),
    [anon_sym_AT] = ACTIONS(2318),
    [anon_sym_DQUOTE] = ACTIONS(2320),
    [anon_sym_L_DQUOTE] = ACTIONS(2320),
    [anon_sym_u_DQUOTE] = ACTIONS(2320),
    [anon_sym_U_DQUOTE] = ACTIONS(2320),
    [anon_sym_u8_DQUOTE] = ACTIONS(2320),
    [sym_true] = ACTIONS(2318),
    [sym_false] = ACTIONS(2318),
    [anon_sym_NULL] = ACTIONS(2318),
    [anon_sym_nullptr] = ACTIONS(2318),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2318),
    [anon_sym___typeof] = ACTIONS(2318),
    [anon_sym_typeof] = ACTIONS(2318),
    [anon_sym_ATimport] = ACTIONS(2320),
    [aux_sym_preproc_undef_token1] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2318),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2318),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2318),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2318),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_API_AVAILABLE] = ACTIONS(2318),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_API_DEPRECATED] = ACTIONS(2318),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym___deprecated_msg] = ACTIONS(2318),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2318),
    [anon_sym_ATprotocol] = ACTIONS(2320),
    [anon_sym_ATinterface] = ACTIONS(2320),
    [anon_sym_ATimplementation] = ACTIONS(2320),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2320),
    [anon_sym_ATtry] = ACTIONS(2320),
    [anon_sym___try] = ACTIONS(2318),
    [anon_sym_ATcatch] = ACTIONS(2562),
    [anon_sym___catch] = ACTIONS(2565),
    [anon_sym_ATfinally] = ACTIONS(2320),
    [anon_sym___finally] = ACTIONS(2318),
    [anon_sym_ATthrow] = ACTIONS(2320),
    [anon_sym_ATselector] = ACTIONS(2320),
    [anon_sym_ATavailable] = ACTIONS(2320),
    [anon_sym___builtin_available] = ACTIONS(2318),
    [anon_sym_va_arg] = ACTIONS(2318),
    [anon_sym_ATencode] = ACTIONS(2320),
    [anon_sym_ATsynchronized] = ACTIONS(2320),
    [anon_sym_BOOL] = ACTIONS(2318),
    [anon_sym_IMP] = ACTIONS(2318),
    [anon_sym_SEL] = ACTIONS(2318),
    [anon_sym_Class] = ACTIONS(2318),
    [anon_sym_id] = ACTIONS(2318),
  },
  [STATE(276)] = {
    [sym_identifier] = ACTIONS(2568),
    [aux_sym_preproc_include_token1] = ACTIONS(2568),
    [aux_sym_preproc_include_token2] = ACTIONS(2568),
    [aux_sym_preproc_def_token1] = ACTIONS(2568),
    [aux_sym_preproc_if_token1] = ACTIONS(2568),
    [aux_sym_preproc_if_token2] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2568),
    [aux_sym_preproc_else_token1] = ACTIONS(2568),
    [aux_sym_preproc_elif_token1] = ACTIONS(2568),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2568),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2568),
    [sym_preproc_directive] = ACTIONS(2568),
    [anon_sym_LPAREN2] = ACTIONS(2570),
    [anon_sym_BANG] = ACTIONS(2570),
    [anon_sym_TILDE] = ACTIONS(2570),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_PLUS] = ACTIONS(2568),
    [anon_sym_STAR] = ACTIONS(2570),
    [anon_sym_CARET] = ACTIONS(2570),
    [anon_sym_AMP] = ACTIONS(2570),
    [anon_sym_SEMI] = ACTIONS(2570),
    [anon_sym___extension__] = ACTIONS(2568),
    [anon_sym_typedef] = ACTIONS(2568),
    [anon_sym_extern] = ACTIONS(2568),
    [anon_sym___attribute__] = ACTIONS(2568),
    [anon_sym___attribute] = ACTIONS(2568),
    [anon_sym_noreturn] = ACTIONS(2568),
    [anon_sym_LBRACK] = ACTIONS(2570),
    [anon_sym___declspec] = ACTIONS(2568),
    [anon_sym___cdecl] = ACTIONS(2568),
    [anon_sym___clrcall] = ACTIONS(2568),
    [anon_sym___stdcall] = ACTIONS(2568),
    [anon_sym___fastcall] = ACTIONS(2568),
    [anon_sym___thiscall] = ACTIONS(2568),
    [anon_sym___vectorcall] = ACTIONS(2568),
    [anon_sym_LBRACE] = ACTIONS(2570),
    [anon_sym_signed] = ACTIONS(2568),
    [anon_sym_unsigned] = ACTIONS(2568),
    [anon_sym_long] = ACTIONS(2568),
    [anon_sym_short] = ACTIONS(2568),
    [anon_sym_static] = ACTIONS(2568),
    [anon_sym_ATautoreleasepool] = ACTIONS(2570),
    [anon_sym_auto] = ACTIONS(2568),
    [anon_sym_register] = ACTIONS(2568),
    [anon_sym_inline] = ACTIONS(2568),
    [anon_sym___inline] = ACTIONS(2568),
    [anon_sym___inline__] = ACTIONS(2568),
    [anon_sym___forceinline] = ACTIONS(2568),
    [anon_sym_thread_local] = ACTIONS(2568),
    [anon_sym___thread] = ACTIONS(2568),
    [anon_sym_CG_EXTERN] = ACTIONS(2568),
    [anon_sym_CG_INLINE] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2568),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2568),
    [anon_sym_IBOutlet] = ACTIONS(2568),
    [anon_sym_IBInspectable] = ACTIONS(2568),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2568),
    [anon_sym_NS_INLINE] = ACTIONS(2568),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2568),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2568),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2568),
    [anon_sym_const] = ACTIONS(2568),
    [anon_sym_constexpr] = ACTIONS(2568),
    [anon_sym_volatile] = ACTIONS(2568),
    [anon_sym_restrict] = ACTIONS(2568),
    [anon_sym___restrict__] = ACTIONS(2568),
    [anon_sym__Atomic] = ACTIONS(2568),
    [anon_sym__Noreturn] = ACTIONS(2568),
    [anon_sym__Nonnull] = ACTIONS(2568),
    [anon_sym_nullable] = ACTIONS(2568),
    [anon_sym__Complex] = ACTIONS(2568),
    [anon_sym__Nullable] = ACTIONS(2568),
    [anon_sym__Nullable_result] = ACTIONS(2568),
    [anon_sym__Null_unspecified] = ACTIONS(2568),
    [anon_sym___autoreleasing] = ACTIONS(2568),
    [anon_sym___block] = ACTIONS(2568),
    [anon_sym___bridge] = ACTIONS(2568),
    [anon_sym___bridge_retained] = ACTIONS(2568),
    [anon_sym___bridge_transfer] = ACTIONS(2568),
    [anon_sym___complex] = ACTIONS(2568),
    [anon_sym___const] = ACTIONS(2568),
    [anon_sym___imag] = ACTIONS(2568),
    [anon_sym___kindof] = ACTIONS(2568),
    [anon_sym___nonnull] = ACTIONS(2568),
    [anon_sym___nullable] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2568),
    [anon_sym___real] = ACTIONS(2568),
    [anon_sym___strong] = ACTIONS(2568),
    [anon_sym___unsafe_unretained] = ACTIONS(2568),
    [anon_sym___unused] = ACTIONS(2568),
    [anon_sym___weak] = ACTIONS(2568),
    [anon_sym_alignas] = ACTIONS(2568),
    [anon_sym__Alignas] = ACTIONS(2568),
    [sym_primitive_type] = ACTIONS(2568),
    [anon_sym_enum] = ACTIONS(2568),
    [anon_sym_struct] = ACTIONS(2568),
    [anon_sym_union] = ACTIONS(2568),
    [anon_sym_if] = ACTIONS(2568),
    [anon_sym_else] = ACTIONS(2568),
    [anon_sym_switch] = ACTIONS(2568),
    [anon_sym_case] = ACTIONS(2568),
    [anon_sym_default] = ACTIONS(2568),
    [anon_sym_while] = ACTIONS(2568),
    [anon_sym_do] = ACTIONS(2568),
    [anon_sym_for] = ACTIONS(2568),
    [anon_sym_in] = ACTIONS(2568),
    [anon_sym_return] = ACTIONS(2568),
    [anon_sym_break] = ACTIONS(2568),
    [anon_sym_continue] = ACTIONS(2568),
    [anon_sym_goto] = ACTIONS(2568),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_sizeof] = ACTIONS(2568),
    [anon_sym___alignof__] = ACTIONS(2568),
    [anon_sym___alignof] = ACTIONS(2568),
    [anon_sym__alignof] = ACTIONS(2568),
    [anon_sym_alignof] = ACTIONS(2568),
    [anon_sym__Alignof] = ACTIONS(2568),
    [anon_sym_offsetof] = ACTIONS(2568),
    [anon_sym__Generic] = ACTIONS(2568),
    [anon_sym_asm] = ACTIONS(2568),
    [anon_sym___asm__] = ACTIONS(2568),
    [anon_sym___asm] = ACTIONS(2568),
    [sym_number_literal] = ACTIONS(2570),
    [anon_sym_L_SQUOTE] = ACTIONS(2570),
    [anon_sym_u_SQUOTE] = ACTIONS(2570),
    [anon_sym_U_SQUOTE] = ACTIONS(2570),
    [anon_sym_u8_SQUOTE] = ACTIONS(2570),
    [anon_sym_SQUOTE] = ACTIONS(2570),
    [anon_sym_AT] = ACTIONS(2568),
    [anon_sym_DQUOTE] = ACTIONS(2570),
    [anon_sym_L_DQUOTE] = ACTIONS(2570),
    [anon_sym_u_DQUOTE] = ACTIONS(2570),
    [anon_sym_U_DQUOTE] = ACTIONS(2570),
    [anon_sym_u8_DQUOTE] = ACTIONS(2570),
    [sym_true] = ACTIONS(2568),
    [sym_false] = ACTIONS(2568),
    [anon_sym_NULL] = ACTIONS(2568),
    [anon_sym_nullptr] = ACTIONS(2568),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2568),
    [anon_sym___typeof] = ACTIONS(2568),
    [anon_sym_typeof] = ACTIONS(2568),
    [anon_sym_ATimport] = ACTIONS(2570),
    [aux_sym_preproc_undef_token1] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2568),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2568),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2568),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2568),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_API_AVAILABLE] = ACTIONS(2568),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_API_DEPRECATED] = ACTIONS(2568),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym___deprecated_msg] = ACTIONS(2568),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2568),
    [anon_sym_ATprotocol] = ACTIONS(2570),
    [anon_sym_ATinterface] = ACTIONS(2570),
    [anon_sym_ATimplementation] = ACTIONS(2570),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2570),
    [anon_sym_ATsynthesize] = ACTIONS(2570),
    [anon_sym_ATdynamic] = ACTIONS(2570),
    [anon_sym_ATtry] = ACTIONS(2570),
    [anon_sym___try] = ACTIONS(2568),
    [anon_sym_ATthrow] = ACTIONS(2570),
    [anon_sym_ATselector] = ACTIONS(2570),
    [anon_sym_ATavailable] = ACTIONS(2570),
    [anon_sym___builtin_available] = ACTIONS(2568),
    [anon_sym_va_arg] = ACTIONS(2568),
    [anon_sym_ATencode] = ACTIONS(2570),
    [anon_sym_ATsynchronized] = ACTIONS(2570),
    [anon_sym_BOOL] = ACTIONS(2568),
    [anon_sym_IMP] = ACTIONS(2568),
    [anon_sym_SEL] = ACTIONS(2568),
    [anon_sym_Class] = ACTIONS(2568),
    [anon_sym_id] = ACTIONS(2568),
  },
  [STATE(277)] = {
    [sym_identifier] = ACTIONS(2572),
    [aux_sym_preproc_include_token1] = ACTIONS(2572),
    [aux_sym_preproc_include_token2] = ACTIONS(2572),
    [aux_sym_preproc_def_token1] = ACTIONS(2572),
    [aux_sym_preproc_if_token1] = ACTIONS(2572),
    [aux_sym_preproc_if_token2] = ACTIONS(2572),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2572),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2572),
    [aux_sym_preproc_else_token1] = ACTIONS(2572),
    [aux_sym_preproc_elif_token1] = ACTIONS(2572),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2572),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2572),
    [sym_preproc_directive] = ACTIONS(2572),
    [anon_sym_LPAREN2] = ACTIONS(2574),
    [anon_sym_BANG] = ACTIONS(2574),
    [anon_sym_TILDE] = ACTIONS(2574),
    [anon_sym_DASH] = ACTIONS(2572),
    [anon_sym_PLUS] = ACTIONS(2572),
    [anon_sym_STAR] = ACTIONS(2574),
    [anon_sym_CARET] = ACTIONS(2574),
    [anon_sym_AMP] = ACTIONS(2574),
    [anon_sym_SEMI] = ACTIONS(2574),
    [anon_sym___extension__] = ACTIONS(2572),
    [anon_sym_typedef] = ACTIONS(2572),
    [anon_sym_extern] = ACTIONS(2572),
    [anon_sym___attribute__] = ACTIONS(2572),
    [anon_sym___attribute] = ACTIONS(2572),
    [anon_sym_noreturn] = ACTIONS(2572),
    [anon_sym_LBRACK] = ACTIONS(2574),
    [anon_sym___declspec] = ACTIONS(2572),
    [anon_sym___cdecl] = ACTIONS(2572),
    [anon_sym___clrcall] = ACTIONS(2572),
    [anon_sym___stdcall] = ACTIONS(2572),
    [anon_sym___fastcall] = ACTIONS(2572),
    [anon_sym___thiscall] = ACTIONS(2572),
    [anon_sym___vectorcall] = ACTIONS(2572),
    [anon_sym_LBRACE] = ACTIONS(2574),
    [anon_sym_signed] = ACTIONS(2572),
    [anon_sym_unsigned] = ACTIONS(2572),
    [anon_sym_long] = ACTIONS(2572),
    [anon_sym_short] = ACTIONS(2572),
    [anon_sym_static] = ACTIONS(2572),
    [anon_sym_ATautoreleasepool] = ACTIONS(2574),
    [anon_sym_auto] = ACTIONS(2572),
    [anon_sym_register] = ACTIONS(2572),
    [anon_sym_inline] = ACTIONS(2572),
    [anon_sym___inline] = ACTIONS(2572),
    [anon_sym___inline__] = ACTIONS(2572),
    [anon_sym___forceinline] = ACTIONS(2572),
    [anon_sym_thread_local] = ACTIONS(2572),
    [anon_sym___thread] = ACTIONS(2572),
    [anon_sym_CG_EXTERN] = ACTIONS(2572),
    [anon_sym_CG_INLINE] = ACTIONS(2572),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2572),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2572),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2572),
    [anon_sym_IBOutlet] = ACTIONS(2572),
    [anon_sym_IBInspectable] = ACTIONS(2572),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2572),
    [anon_sym_NS_INLINE] = ACTIONS(2572),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2572),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2572),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2572),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2572),
    [anon_sym_const] = ACTIONS(2572),
    [anon_sym_constexpr] = ACTIONS(2572),
    [anon_sym_volatile] = ACTIONS(2572),
    [anon_sym_restrict] = ACTIONS(2572),
    [anon_sym___restrict__] = ACTIONS(2572),
    [anon_sym__Atomic] = ACTIONS(2572),
    [anon_sym__Noreturn] = ACTIONS(2572),
    [anon_sym__Nonnull] = ACTIONS(2572),
    [anon_sym_nullable] = ACTIONS(2572),
    [anon_sym__Complex] = ACTIONS(2572),
    [anon_sym__Nullable] = ACTIONS(2572),
    [anon_sym__Nullable_result] = ACTIONS(2572),
    [anon_sym__Null_unspecified] = ACTIONS(2572),
    [anon_sym___autoreleasing] = ACTIONS(2572),
    [anon_sym___block] = ACTIONS(2572),
    [anon_sym___bridge] = ACTIONS(2572),
    [anon_sym___bridge_retained] = ACTIONS(2572),
    [anon_sym___bridge_transfer] = ACTIONS(2572),
    [anon_sym___complex] = ACTIONS(2572),
    [anon_sym___const] = ACTIONS(2572),
    [anon_sym___imag] = ACTIONS(2572),
    [anon_sym___kindof] = ACTIONS(2572),
    [anon_sym___nonnull] = ACTIONS(2572),
    [anon_sym___nullable] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2572),
    [anon_sym___real] = ACTIONS(2572),
    [anon_sym___strong] = ACTIONS(2572),
    [anon_sym___unsafe_unretained] = ACTIONS(2572),
    [anon_sym___unused] = ACTIONS(2572),
    [anon_sym___weak] = ACTIONS(2572),
    [anon_sym_alignas] = ACTIONS(2572),
    [anon_sym__Alignas] = ACTIONS(2572),
    [sym_primitive_type] = ACTIONS(2572),
    [anon_sym_enum] = ACTIONS(2572),
    [anon_sym_struct] = ACTIONS(2572),
    [anon_sym_union] = ACTIONS(2572),
    [anon_sym_if] = ACTIONS(2572),
    [anon_sym_switch] = ACTIONS(2572),
    [anon_sym_case] = ACTIONS(2572),
    [anon_sym_default] = ACTIONS(2572),
    [anon_sym_while] = ACTIONS(2572),
    [anon_sym_do] = ACTIONS(2572),
    [anon_sym_for] = ACTIONS(2572),
    [anon_sym_in] = ACTIONS(2572),
    [anon_sym_return] = ACTIONS(2572),
    [anon_sym_break] = ACTIONS(2572),
    [anon_sym_continue] = ACTIONS(2572),
    [anon_sym_goto] = ACTIONS(2572),
    [anon_sym_DASH_DASH] = ACTIONS(2574),
    [anon_sym_PLUS_PLUS] = ACTIONS(2574),
    [anon_sym_sizeof] = ACTIONS(2572),
    [anon_sym___alignof__] = ACTIONS(2572),
    [anon_sym___alignof] = ACTIONS(2572),
    [anon_sym__alignof] = ACTIONS(2572),
    [anon_sym_alignof] = ACTIONS(2572),
    [anon_sym__Alignof] = ACTIONS(2572),
    [anon_sym_offsetof] = ACTIONS(2572),
    [anon_sym__Generic] = ACTIONS(2572),
    [anon_sym_asm] = ACTIONS(2572),
    [anon_sym___asm__] = ACTIONS(2572),
    [anon_sym___asm] = ACTIONS(2572),
    [sym_number_literal] = ACTIONS(2574),
    [anon_sym_L_SQUOTE] = ACTIONS(2574),
    [anon_sym_u_SQUOTE] = ACTIONS(2574),
    [anon_sym_U_SQUOTE] = ACTIONS(2574),
    [anon_sym_u8_SQUOTE] = ACTIONS(2574),
    [anon_sym_SQUOTE] = ACTIONS(2574),
    [anon_sym_AT] = ACTIONS(2572),
    [anon_sym_DQUOTE] = ACTIONS(2574),
    [anon_sym_L_DQUOTE] = ACTIONS(2574),
    [anon_sym_u_DQUOTE] = ACTIONS(2574),
    [anon_sym_U_DQUOTE] = ACTIONS(2574),
    [anon_sym_u8_DQUOTE] = ACTIONS(2574),
    [sym_true] = ACTIONS(2572),
    [sym_false] = ACTIONS(2572),
    [anon_sym_NULL] = ACTIONS(2572),
    [anon_sym_nullptr] = ACTIONS(2572),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2572),
    [anon_sym___typeof] = ACTIONS(2572),
    [anon_sym_typeof] = ACTIONS(2572),
    [anon_sym_ATimport] = ACTIONS(2574),
    [aux_sym_preproc_undef_token1] = ACTIONS(2572),
    [anon_sym_POUND] = ACTIONS(2572),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2572),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2572),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2572),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2572),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2572),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2572),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2572),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_API_AVAILABLE] = ACTIONS(2572),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_API_DEPRECATED] = ACTIONS(2572),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2572),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym___deprecated_msg] = ACTIONS(2572),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2572),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2572),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2572),
    [anon_sym_ATprotocol] = ACTIONS(2574),
    [anon_sym_ATinterface] = ACTIONS(2574),
    [anon_sym_ATimplementation] = ACTIONS(2574),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2574),
    [anon_sym_ATsynthesize] = ACTIONS(2574),
    [anon_sym_ATdynamic] = ACTIONS(2574),
    [anon_sym_ATtry] = ACTIONS(2574),
    [anon_sym___try] = ACTIONS(2572),
    [anon_sym_ATthrow] = ACTIONS(2574),
    [anon_sym_ATselector] = ACTIONS(2574),
    [anon_sym_ATavailable] = ACTIONS(2574),
    [anon_sym___builtin_available] = ACTIONS(2572),
    [anon_sym_va_arg] = ACTIONS(2572),
    [anon_sym_ATencode] = ACTIONS(2574),
    [anon_sym_ATsynchronized] = ACTIONS(2574),
    [anon_sym_BOOL] = ACTIONS(2572),
    [anon_sym_IMP] = ACTIONS(2572),
    [anon_sym_SEL] = ACTIONS(2572),
    [anon_sym_Class] = ACTIONS(2572),
    [anon_sym_id] = ACTIONS(2572),
  },
  [STATE(278)] = {
    [sym_identifier] = ACTIONS(2576),
    [aux_sym_preproc_include_token1] = ACTIONS(2576),
    [aux_sym_preproc_include_token2] = ACTIONS(2576),
    [aux_sym_preproc_def_token1] = ACTIONS(2576),
    [aux_sym_preproc_if_token1] = ACTIONS(2576),
    [aux_sym_preproc_if_token2] = ACTIONS(2576),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2576),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2576),
    [aux_sym_preproc_else_token1] = ACTIONS(2576),
    [aux_sym_preproc_elif_token1] = ACTIONS(2576),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2576),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2576),
    [sym_preproc_directive] = ACTIONS(2576),
    [anon_sym_LPAREN2] = ACTIONS(2578),
    [anon_sym_BANG] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2578),
    [anon_sym_DASH] = ACTIONS(2576),
    [anon_sym_PLUS] = ACTIONS(2576),
    [anon_sym_STAR] = ACTIONS(2578),
    [anon_sym_CARET] = ACTIONS(2578),
    [anon_sym_AMP] = ACTIONS(2578),
    [anon_sym_SEMI] = ACTIONS(2578),
    [anon_sym___extension__] = ACTIONS(2576),
    [anon_sym_typedef] = ACTIONS(2576),
    [anon_sym_extern] = ACTIONS(2576),
    [anon_sym___attribute__] = ACTIONS(2576),
    [anon_sym___attribute] = ACTIONS(2576),
    [anon_sym_noreturn] = ACTIONS(2576),
    [anon_sym_LBRACK] = ACTIONS(2578),
    [anon_sym___declspec] = ACTIONS(2576),
    [anon_sym___cdecl] = ACTIONS(2576),
    [anon_sym___clrcall] = ACTIONS(2576),
    [anon_sym___stdcall] = ACTIONS(2576),
    [anon_sym___fastcall] = ACTIONS(2576),
    [anon_sym___thiscall] = ACTIONS(2576),
    [anon_sym___vectorcall] = ACTIONS(2576),
    [anon_sym_LBRACE] = ACTIONS(2578),
    [anon_sym_signed] = ACTIONS(2576),
    [anon_sym_unsigned] = ACTIONS(2576),
    [anon_sym_long] = ACTIONS(2576),
    [anon_sym_short] = ACTIONS(2576),
    [anon_sym_static] = ACTIONS(2576),
    [anon_sym_ATautoreleasepool] = ACTIONS(2578),
    [anon_sym_auto] = ACTIONS(2576),
    [anon_sym_register] = ACTIONS(2576),
    [anon_sym_inline] = ACTIONS(2576),
    [anon_sym___inline] = ACTIONS(2576),
    [anon_sym___inline__] = ACTIONS(2576),
    [anon_sym___forceinline] = ACTIONS(2576),
    [anon_sym_thread_local] = ACTIONS(2576),
    [anon_sym___thread] = ACTIONS(2576),
    [anon_sym_CG_EXTERN] = ACTIONS(2576),
    [anon_sym_CG_INLINE] = ACTIONS(2576),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2576),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2576),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2576),
    [anon_sym_IBOutlet] = ACTIONS(2576),
    [anon_sym_IBInspectable] = ACTIONS(2576),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2576),
    [anon_sym_NS_INLINE] = ACTIONS(2576),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2576),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2576),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2576),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2576),
    [anon_sym_const] = ACTIONS(2576),
    [anon_sym_constexpr] = ACTIONS(2576),
    [anon_sym_volatile] = ACTIONS(2576),
    [anon_sym_restrict] = ACTIONS(2576),
    [anon_sym___restrict__] = ACTIONS(2576),
    [anon_sym__Atomic] = ACTIONS(2576),
    [anon_sym__Noreturn] = ACTIONS(2576),
    [anon_sym__Nonnull] = ACTIONS(2576),
    [anon_sym_nullable] = ACTIONS(2576),
    [anon_sym__Complex] = ACTIONS(2576),
    [anon_sym__Nullable] = ACTIONS(2576),
    [anon_sym__Nullable_result] = ACTIONS(2576),
    [anon_sym__Null_unspecified] = ACTIONS(2576),
    [anon_sym___autoreleasing] = ACTIONS(2576),
    [anon_sym___block] = ACTIONS(2576),
    [anon_sym___bridge] = ACTIONS(2576),
    [anon_sym___bridge_retained] = ACTIONS(2576),
    [anon_sym___bridge_transfer] = ACTIONS(2576),
    [anon_sym___complex] = ACTIONS(2576),
    [anon_sym___const] = ACTIONS(2576),
    [anon_sym___imag] = ACTIONS(2576),
    [anon_sym___kindof] = ACTIONS(2576),
    [anon_sym___nonnull] = ACTIONS(2576),
    [anon_sym___nullable] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2576),
    [anon_sym___real] = ACTIONS(2576),
    [anon_sym___strong] = ACTIONS(2576),
    [anon_sym___unsafe_unretained] = ACTIONS(2576),
    [anon_sym___unused] = ACTIONS(2576),
    [anon_sym___weak] = ACTIONS(2576),
    [anon_sym_alignas] = ACTIONS(2576),
    [anon_sym__Alignas] = ACTIONS(2576),
    [sym_primitive_type] = ACTIONS(2576),
    [anon_sym_enum] = ACTIONS(2576),
    [anon_sym_struct] = ACTIONS(2576),
    [anon_sym_union] = ACTIONS(2576),
    [anon_sym_if] = ACTIONS(2576),
    [anon_sym_switch] = ACTIONS(2576),
    [anon_sym_case] = ACTIONS(2576),
    [anon_sym_default] = ACTIONS(2576),
    [anon_sym_while] = ACTIONS(2576),
    [anon_sym_do] = ACTIONS(2576),
    [anon_sym_for] = ACTIONS(2576),
    [anon_sym_in] = ACTIONS(2576),
    [anon_sym_return] = ACTIONS(2576),
    [anon_sym_break] = ACTIONS(2576),
    [anon_sym_continue] = ACTIONS(2576),
    [anon_sym_goto] = ACTIONS(2576),
    [anon_sym_DASH_DASH] = ACTIONS(2578),
    [anon_sym_PLUS_PLUS] = ACTIONS(2578),
    [anon_sym_sizeof] = ACTIONS(2576),
    [anon_sym___alignof__] = ACTIONS(2576),
    [anon_sym___alignof] = ACTIONS(2576),
    [anon_sym__alignof] = ACTIONS(2576),
    [anon_sym_alignof] = ACTIONS(2576),
    [anon_sym__Alignof] = ACTIONS(2576),
    [anon_sym_offsetof] = ACTIONS(2576),
    [anon_sym__Generic] = ACTIONS(2576),
    [anon_sym_asm] = ACTIONS(2576),
    [anon_sym___asm__] = ACTIONS(2576),
    [anon_sym___asm] = ACTIONS(2576),
    [sym_number_literal] = ACTIONS(2578),
    [anon_sym_L_SQUOTE] = ACTIONS(2578),
    [anon_sym_u_SQUOTE] = ACTIONS(2578),
    [anon_sym_U_SQUOTE] = ACTIONS(2578),
    [anon_sym_u8_SQUOTE] = ACTIONS(2578),
    [anon_sym_SQUOTE] = ACTIONS(2578),
    [anon_sym_AT] = ACTIONS(2576),
    [anon_sym_DQUOTE] = ACTIONS(2578),
    [anon_sym_L_DQUOTE] = ACTIONS(2578),
    [anon_sym_u_DQUOTE] = ACTIONS(2578),
    [anon_sym_U_DQUOTE] = ACTIONS(2578),
    [anon_sym_u8_DQUOTE] = ACTIONS(2578),
    [sym_true] = ACTIONS(2576),
    [sym_false] = ACTIONS(2576),
    [anon_sym_NULL] = ACTIONS(2576),
    [anon_sym_nullptr] = ACTIONS(2576),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2576),
    [anon_sym___typeof] = ACTIONS(2576),
    [anon_sym_typeof] = ACTIONS(2576),
    [anon_sym_ATimport] = ACTIONS(2578),
    [aux_sym_preproc_undef_token1] = ACTIONS(2576),
    [anon_sym_POUND] = ACTIONS(2576),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2576),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2576),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2576),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2576),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2576),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2576),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2576),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_API_AVAILABLE] = ACTIONS(2576),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_API_DEPRECATED] = ACTIONS(2576),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2576),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym___deprecated_msg] = ACTIONS(2576),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2576),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2576),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2576),
    [anon_sym_ATprotocol] = ACTIONS(2578),
    [anon_sym_ATinterface] = ACTIONS(2578),
    [anon_sym_ATimplementation] = ACTIONS(2578),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2578),
    [anon_sym_ATsynthesize] = ACTIONS(2578),
    [anon_sym_ATdynamic] = ACTIONS(2578),
    [anon_sym_ATtry] = ACTIONS(2578),
    [anon_sym___try] = ACTIONS(2576),
    [anon_sym_ATthrow] = ACTIONS(2578),
    [anon_sym_ATselector] = ACTIONS(2578),
    [anon_sym_ATavailable] = ACTIONS(2578),
    [anon_sym___builtin_available] = ACTIONS(2576),
    [anon_sym_va_arg] = ACTIONS(2576),
    [anon_sym_ATencode] = ACTIONS(2578),
    [anon_sym_ATsynchronized] = ACTIONS(2578),
    [anon_sym_BOOL] = ACTIONS(2576),
    [anon_sym_IMP] = ACTIONS(2576),
    [anon_sym_SEL] = ACTIONS(2576),
    [anon_sym_Class] = ACTIONS(2576),
    [anon_sym_id] = ACTIONS(2576),
  },
  [STATE(279)] = {
    [sym_identifier] = ACTIONS(2580),
    [aux_sym_preproc_include_token1] = ACTIONS(2580),
    [aux_sym_preproc_include_token2] = ACTIONS(2580),
    [aux_sym_preproc_def_token1] = ACTIONS(2580),
    [aux_sym_preproc_if_token1] = ACTIONS(2580),
    [aux_sym_preproc_if_token2] = ACTIONS(2580),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2580),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2580),
    [aux_sym_preproc_else_token1] = ACTIONS(2580),
    [aux_sym_preproc_elif_token1] = ACTIONS(2580),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2580),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2580),
    [sym_preproc_directive] = ACTIONS(2580),
    [anon_sym_LPAREN2] = ACTIONS(2582),
    [anon_sym_BANG] = ACTIONS(2582),
    [anon_sym_TILDE] = ACTIONS(2582),
    [anon_sym_DASH] = ACTIONS(2580),
    [anon_sym_PLUS] = ACTIONS(2580),
    [anon_sym_STAR] = ACTIONS(2582),
    [anon_sym_CARET] = ACTIONS(2582),
    [anon_sym_AMP] = ACTIONS(2582),
    [anon_sym_SEMI] = ACTIONS(2582),
    [anon_sym___extension__] = ACTIONS(2580),
    [anon_sym_typedef] = ACTIONS(2580),
    [anon_sym_extern] = ACTIONS(2580),
    [anon_sym___attribute__] = ACTIONS(2580),
    [anon_sym___attribute] = ACTIONS(2580),
    [anon_sym_noreturn] = ACTIONS(2580),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym___declspec] = ACTIONS(2580),
    [anon_sym___cdecl] = ACTIONS(2580),
    [anon_sym___clrcall] = ACTIONS(2580),
    [anon_sym___stdcall] = ACTIONS(2580),
    [anon_sym___fastcall] = ACTIONS(2580),
    [anon_sym___thiscall] = ACTIONS(2580),
    [anon_sym___vectorcall] = ACTIONS(2580),
    [anon_sym_LBRACE] = ACTIONS(2582),
    [anon_sym_signed] = ACTIONS(2580),
    [anon_sym_unsigned] = ACTIONS(2580),
    [anon_sym_long] = ACTIONS(2580),
    [anon_sym_short] = ACTIONS(2580),
    [anon_sym_static] = ACTIONS(2580),
    [anon_sym_ATautoreleasepool] = ACTIONS(2582),
    [anon_sym_auto] = ACTIONS(2580),
    [anon_sym_register] = ACTIONS(2580),
    [anon_sym_inline] = ACTIONS(2580),
    [anon_sym___inline] = ACTIONS(2580),
    [anon_sym___inline__] = ACTIONS(2580),
    [anon_sym___forceinline] = ACTIONS(2580),
    [anon_sym_thread_local] = ACTIONS(2580),
    [anon_sym___thread] = ACTIONS(2580),
    [anon_sym_CG_EXTERN] = ACTIONS(2580),
    [anon_sym_CG_INLINE] = ACTIONS(2580),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2580),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2580),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2580),
    [anon_sym_IBOutlet] = ACTIONS(2580),
    [anon_sym_IBInspectable] = ACTIONS(2580),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2580),
    [anon_sym_NS_INLINE] = ACTIONS(2580),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2580),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2580),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2580),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2580),
    [anon_sym_const] = ACTIONS(2580),
    [anon_sym_constexpr] = ACTIONS(2580),
    [anon_sym_volatile] = ACTIONS(2580),
    [anon_sym_restrict] = ACTIONS(2580),
    [anon_sym___restrict__] = ACTIONS(2580),
    [anon_sym__Atomic] = ACTIONS(2580),
    [anon_sym__Noreturn] = ACTIONS(2580),
    [anon_sym__Nonnull] = ACTIONS(2580),
    [anon_sym_nullable] = ACTIONS(2580),
    [anon_sym__Complex] = ACTIONS(2580),
    [anon_sym__Nullable] = ACTIONS(2580),
    [anon_sym__Nullable_result] = ACTIONS(2580),
    [anon_sym__Null_unspecified] = ACTIONS(2580),
    [anon_sym___autoreleasing] = ACTIONS(2580),
    [anon_sym___block] = ACTIONS(2580),
    [anon_sym___bridge] = ACTIONS(2580),
    [anon_sym___bridge_retained] = ACTIONS(2580),
    [anon_sym___bridge_transfer] = ACTIONS(2580),
    [anon_sym___complex] = ACTIONS(2580),
    [anon_sym___const] = ACTIONS(2580),
    [anon_sym___imag] = ACTIONS(2580),
    [anon_sym___kindof] = ACTIONS(2580),
    [anon_sym___nonnull] = ACTIONS(2580),
    [anon_sym___nullable] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2580),
    [anon_sym___real] = ACTIONS(2580),
    [anon_sym___strong] = ACTIONS(2580),
    [anon_sym___unsafe_unretained] = ACTIONS(2580),
    [anon_sym___unused] = ACTIONS(2580),
    [anon_sym___weak] = ACTIONS(2580),
    [anon_sym_alignas] = ACTIONS(2580),
    [anon_sym__Alignas] = ACTIONS(2580),
    [sym_primitive_type] = ACTIONS(2580),
    [anon_sym_enum] = ACTIONS(2580),
    [anon_sym_struct] = ACTIONS(2580),
    [anon_sym_union] = ACTIONS(2580),
    [anon_sym_if] = ACTIONS(2580),
    [anon_sym_switch] = ACTIONS(2580),
    [anon_sym_case] = ACTIONS(2580),
    [anon_sym_default] = ACTIONS(2580),
    [anon_sym_while] = ACTIONS(2580),
    [anon_sym_do] = ACTIONS(2580),
    [anon_sym_for] = ACTIONS(2580),
    [anon_sym_in] = ACTIONS(2580),
    [anon_sym_return] = ACTIONS(2580),
    [anon_sym_break] = ACTIONS(2580),
    [anon_sym_continue] = ACTIONS(2580),
    [anon_sym_goto] = ACTIONS(2580),
    [anon_sym_DASH_DASH] = ACTIONS(2582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2582),
    [anon_sym_sizeof] = ACTIONS(2580),
    [anon_sym___alignof__] = ACTIONS(2580),
    [anon_sym___alignof] = ACTIONS(2580),
    [anon_sym__alignof] = ACTIONS(2580),
    [anon_sym_alignof] = ACTIONS(2580),
    [anon_sym__Alignof] = ACTIONS(2580),
    [anon_sym_offsetof] = ACTIONS(2580),
    [anon_sym__Generic] = ACTIONS(2580),
    [anon_sym_asm] = ACTIONS(2580),
    [anon_sym___asm__] = ACTIONS(2580),
    [anon_sym___asm] = ACTIONS(2580),
    [sym_number_literal] = ACTIONS(2582),
    [anon_sym_L_SQUOTE] = ACTIONS(2582),
    [anon_sym_u_SQUOTE] = ACTIONS(2582),
    [anon_sym_U_SQUOTE] = ACTIONS(2582),
    [anon_sym_u8_SQUOTE] = ACTIONS(2582),
    [anon_sym_SQUOTE] = ACTIONS(2582),
    [anon_sym_AT] = ACTIONS(2580),
    [anon_sym_DQUOTE] = ACTIONS(2582),
    [anon_sym_L_DQUOTE] = ACTIONS(2582),
    [anon_sym_u_DQUOTE] = ACTIONS(2582),
    [anon_sym_U_DQUOTE] = ACTIONS(2582),
    [anon_sym_u8_DQUOTE] = ACTIONS(2582),
    [sym_true] = ACTIONS(2580),
    [sym_false] = ACTIONS(2580),
    [anon_sym_NULL] = ACTIONS(2580),
    [anon_sym_nullptr] = ACTIONS(2580),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2580),
    [anon_sym___typeof] = ACTIONS(2580),
    [anon_sym_typeof] = ACTIONS(2580),
    [anon_sym_ATimport] = ACTIONS(2582),
    [aux_sym_preproc_undef_token1] = ACTIONS(2580),
    [anon_sym_POUND] = ACTIONS(2580),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2580),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2580),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2580),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2580),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2580),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2580),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2580),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_API_AVAILABLE] = ACTIONS(2580),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_API_DEPRECATED] = ACTIONS(2580),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2580),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym___deprecated_msg] = ACTIONS(2580),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2580),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2580),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2580),
    [anon_sym_ATprotocol] = ACTIONS(2582),
    [anon_sym_ATinterface] = ACTIONS(2582),
    [anon_sym_ATimplementation] = ACTIONS(2582),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2582),
    [anon_sym_ATsynthesize] = ACTIONS(2582),
    [anon_sym_ATdynamic] = ACTIONS(2582),
    [anon_sym_ATtry] = ACTIONS(2582),
    [anon_sym___try] = ACTIONS(2580),
    [anon_sym_ATthrow] = ACTIONS(2582),
    [anon_sym_ATselector] = ACTIONS(2582),
    [anon_sym_ATavailable] = ACTIONS(2582),
    [anon_sym___builtin_available] = ACTIONS(2580),
    [anon_sym_va_arg] = ACTIONS(2580),
    [anon_sym_ATencode] = ACTIONS(2582),
    [anon_sym_ATsynchronized] = ACTIONS(2582),
    [anon_sym_BOOL] = ACTIONS(2580),
    [anon_sym_IMP] = ACTIONS(2580),
    [anon_sym_SEL] = ACTIONS(2580),
    [anon_sym_Class] = ACTIONS(2580),
    [anon_sym_id] = ACTIONS(2580),
  },
  [STATE(280)] = {
    [sym_identifier] = ACTIONS(2584),
    [aux_sym_preproc_include_token1] = ACTIONS(2584),
    [aux_sym_preproc_include_token2] = ACTIONS(2584),
    [aux_sym_preproc_def_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token2] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2584),
    [aux_sym_preproc_else_token1] = ACTIONS(2584),
    [aux_sym_preproc_elif_token1] = ACTIONS(2584),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2584),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2584),
    [sym_preproc_directive] = ACTIONS(2584),
    [anon_sym_LPAREN2] = ACTIONS(2586),
    [anon_sym_BANG] = ACTIONS(2586),
    [anon_sym_TILDE] = ACTIONS(2586),
    [anon_sym_DASH] = ACTIONS(2584),
    [anon_sym_PLUS] = ACTIONS(2584),
    [anon_sym_STAR] = ACTIONS(2586),
    [anon_sym_CARET] = ACTIONS(2586),
    [anon_sym_AMP] = ACTIONS(2586),
    [anon_sym_SEMI] = ACTIONS(2586),
    [anon_sym___extension__] = ACTIONS(2584),
    [anon_sym_typedef] = ACTIONS(2584),
    [anon_sym_extern] = ACTIONS(2584),
    [anon_sym___attribute__] = ACTIONS(2584),
    [anon_sym___attribute] = ACTIONS(2584),
    [anon_sym_noreturn] = ACTIONS(2584),
    [anon_sym_LBRACK] = ACTIONS(2586),
    [anon_sym___declspec] = ACTIONS(2584),
    [anon_sym___cdecl] = ACTIONS(2584),
    [anon_sym___clrcall] = ACTIONS(2584),
    [anon_sym___stdcall] = ACTIONS(2584),
    [anon_sym___fastcall] = ACTIONS(2584),
    [anon_sym___thiscall] = ACTIONS(2584),
    [anon_sym___vectorcall] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_signed] = ACTIONS(2584),
    [anon_sym_unsigned] = ACTIONS(2584),
    [anon_sym_long] = ACTIONS(2584),
    [anon_sym_short] = ACTIONS(2584),
    [anon_sym_static] = ACTIONS(2584),
    [anon_sym_ATautoreleasepool] = ACTIONS(2586),
    [anon_sym_auto] = ACTIONS(2584),
    [anon_sym_register] = ACTIONS(2584),
    [anon_sym_inline] = ACTIONS(2584),
    [anon_sym___inline] = ACTIONS(2584),
    [anon_sym___inline__] = ACTIONS(2584),
    [anon_sym___forceinline] = ACTIONS(2584),
    [anon_sym_thread_local] = ACTIONS(2584),
    [anon_sym___thread] = ACTIONS(2584),
    [anon_sym_CG_EXTERN] = ACTIONS(2584),
    [anon_sym_CG_INLINE] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2584),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2584),
    [anon_sym_IBOutlet] = ACTIONS(2584),
    [anon_sym_IBInspectable] = ACTIONS(2584),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2584),
    [anon_sym_NS_INLINE] = ACTIONS(2584),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2584),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2584),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2584),
    [anon_sym_const] = ACTIONS(2584),
    [anon_sym_constexpr] = ACTIONS(2584),
    [anon_sym_volatile] = ACTIONS(2584),
    [anon_sym_restrict] = ACTIONS(2584),
    [anon_sym___restrict__] = ACTIONS(2584),
    [anon_sym__Atomic] = ACTIONS(2584),
    [anon_sym__Noreturn] = ACTIONS(2584),
    [anon_sym__Nonnull] = ACTIONS(2584),
    [anon_sym_nullable] = ACTIONS(2584),
    [anon_sym__Complex] = ACTIONS(2584),
    [anon_sym__Nullable] = ACTIONS(2584),
    [anon_sym__Nullable_result] = ACTIONS(2584),
    [anon_sym__Null_unspecified] = ACTIONS(2584),
    [anon_sym___autoreleasing] = ACTIONS(2584),
    [anon_sym___block] = ACTIONS(2584),
    [anon_sym___bridge] = ACTIONS(2584),
    [anon_sym___bridge_retained] = ACTIONS(2584),
    [anon_sym___bridge_transfer] = ACTIONS(2584),
    [anon_sym___complex] = ACTIONS(2584),
    [anon_sym___const] = ACTIONS(2584),
    [anon_sym___imag] = ACTIONS(2584),
    [anon_sym___kindof] = ACTIONS(2584),
    [anon_sym___nonnull] = ACTIONS(2584),
    [anon_sym___nullable] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2584),
    [anon_sym___real] = ACTIONS(2584),
    [anon_sym___strong] = ACTIONS(2584),
    [anon_sym___unsafe_unretained] = ACTIONS(2584),
    [anon_sym___unused] = ACTIONS(2584),
    [anon_sym___weak] = ACTIONS(2584),
    [anon_sym_alignas] = ACTIONS(2584),
    [anon_sym__Alignas] = ACTIONS(2584),
    [sym_primitive_type] = ACTIONS(2584),
    [anon_sym_enum] = ACTIONS(2584),
    [anon_sym_struct] = ACTIONS(2584),
    [anon_sym_union] = ACTIONS(2584),
    [anon_sym_if] = ACTIONS(2584),
    [anon_sym_switch] = ACTIONS(2584),
    [anon_sym_case] = ACTIONS(2584),
    [anon_sym_default] = ACTIONS(2584),
    [anon_sym_while] = ACTIONS(2584),
    [anon_sym_do] = ACTIONS(2584),
    [anon_sym_for] = ACTIONS(2584),
    [anon_sym_in] = ACTIONS(2584),
    [anon_sym_return] = ACTIONS(2584),
    [anon_sym_break] = ACTIONS(2584),
    [anon_sym_continue] = ACTIONS(2584),
    [anon_sym_goto] = ACTIONS(2584),
    [anon_sym_DASH_DASH] = ACTIONS(2586),
    [anon_sym_PLUS_PLUS] = ACTIONS(2586),
    [anon_sym_sizeof] = ACTIONS(2584),
    [anon_sym___alignof__] = ACTIONS(2584),
    [anon_sym___alignof] = ACTIONS(2584),
    [anon_sym__alignof] = ACTIONS(2584),
    [anon_sym_alignof] = ACTIONS(2584),
    [anon_sym__Alignof] = ACTIONS(2584),
    [anon_sym_offsetof] = ACTIONS(2584),
    [anon_sym__Generic] = ACTIONS(2584),
    [anon_sym_asm] = ACTIONS(2584),
    [anon_sym___asm__] = ACTIONS(2584),
    [anon_sym___asm] = ACTIONS(2584),
    [sym_number_literal] = ACTIONS(2586),
    [anon_sym_L_SQUOTE] = ACTIONS(2586),
    [anon_sym_u_SQUOTE] = ACTIONS(2586),
    [anon_sym_U_SQUOTE] = ACTIONS(2586),
    [anon_sym_u8_SQUOTE] = ACTIONS(2586),
    [anon_sym_SQUOTE] = ACTIONS(2586),
    [anon_sym_AT] = ACTIONS(2584),
    [anon_sym_DQUOTE] = ACTIONS(2586),
    [anon_sym_L_DQUOTE] = ACTIONS(2586),
    [anon_sym_u_DQUOTE] = ACTIONS(2586),
    [anon_sym_U_DQUOTE] = ACTIONS(2586),
    [anon_sym_u8_DQUOTE] = ACTIONS(2586),
    [sym_true] = ACTIONS(2584),
    [sym_false] = ACTIONS(2584),
    [anon_sym_NULL] = ACTIONS(2584),
    [anon_sym_nullptr] = ACTIONS(2584),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2584),
    [anon_sym___typeof] = ACTIONS(2584),
    [anon_sym_typeof] = ACTIONS(2584),
    [anon_sym_ATimport] = ACTIONS(2586),
    [aux_sym_preproc_undef_token1] = ACTIONS(2584),
    [anon_sym_POUND] = ACTIONS(2584),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2584),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2584),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2584),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_API_AVAILABLE] = ACTIONS(2584),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_API_DEPRECATED] = ACTIONS(2584),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym___deprecated_msg] = ACTIONS(2584),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2584),
    [anon_sym_ATprotocol] = ACTIONS(2586),
    [anon_sym_ATinterface] = ACTIONS(2586),
    [anon_sym_ATimplementation] = ACTIONS(2586),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2586),
    [anon_sym_ATsynthesize] = ACTIONS(2586),
    [anon_sym_ATdynamic] = ACTIONS(2586),
    [anon_sym_ATtry] = ACTIONS(2586),
    [anon_sym___try] = ACTIONS(2584),
    [anon_sym_ATthrow] = ACTIONS(2586),
    [anon_sym_ATselector] = ACTIONS(2586),
    [anon_sym_ATavailable] = ACTIONS(2586),
    [anon_sym___builtin_available] = ACTIONS(2584),
    [anon_sym_va_arg] = ACTIONS(2584),
    [anon_sym_ATencode] = ACTIONS(2586),
    [anon_sym_ATsynchronized] = ACTIONS(2586),
    [anon_sym_BOOL] = ACTIONS(2584),
    [anon_sym_IMP] = ACTIONS(2584),
    [anon_sym_SEL] = ACTIONS(2584),
    [anon_sym_Class] = ACTIONS(2584),
    [anon_sym_id] = ACTIONS(2584),
  },
  [STATE(281)] = {
    [sym_identifier] = ACTIONS(2588),
    [aux_sym_preproc_include_token1] = ACTIONS(2588),
    [aux_sym_preproc_include_token2] = ACTIONS(2588),
    [aux_sym_preproc_def_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token2] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2588),
    [aux_sym_preproc_else_token1] = ACTIONS(2588),
    [aux_sym_preproc_elif_token1] = ACTIONS(2588),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2588),
    [sym_preproc_directive] = ACTIONS(2588),
    [anon_sym_LPAREN2] = ACTIONS(2590),
    [anon_sym_BANG] = ACTIONS(2590),
    [anon_sym_TILDE] = ACTIONS(2590),
    [anon_sym_DASH] = ACTIONS(2588),
    [anon_sym_PLUS] = ACTIONS(2588),
    [anon_sym_STAR] = ACTIONS(2590),
    [anon_sym_CARET] = ACTIONS(2590),
    [anon_sym_AMP] = ACTIONS(2590),
    [anon_sym_SEMI] = ACTIONS(2590),
    [anon_sym___extension__] = ACTIONS(2588),
    [anon_sym_typedef] = ACTIONS(2588),
    [anon_sym_extern] = ACTIONS(2588),
    [anon_sym___attribute__] = ACTIONS(2588),
    [anon_sym___attribute] = ACTIONS(2588),
    [anon_sym_noreturn] = ACTIONS(2588),
    [anon_sym_LBRACK] = ACTIONS(2590),
    [anon_sym___declspec] = ACTIONS(2588),
    [anon_sym___cdecl] = ACTIONS(2588),
    [anon_sym___clrcall] = ACTIONS(2588),
    [anon_sym___stdcall] = ACTIONS(2588),
    [anon_sym___fastcall] = ACTIONS(2588),
    [anon_sym___thiscall] = ACTIONS(2588),
    [anon_sym___vectorcall] = ACTIONS(2588),
    [anon_sym_LBRACE] = ACTIONS(2590),
    [anon_sym_signed] = ACTIONS(2588),
    [anon_sym_unsigned] = ACTIONS(2588),
    [anon_sym_long] = ACTIONS(2588),
    [anon_sym_short] = ACTIONS(2588),
    [anon_sym_static] = ACTIONS(2588),
    [anon_sym_ATautoreleasepool] = ACTIONS(2590),
    [anon_sym_auto] = ACTIONS(2588),
    [anon_sym_register] = ACTIONS(2588),
    [anon_sym_inline] = ACTIONS(2588),
    [anon_sym___inline] = ACTIONS(2588),
    [anon_sym___inline__] = ACTIONS(2588),
    [anon_sym___forceinline] = ACTIONS(2588),
    [anon_sym_thread_local] = ACTIONS(2588),
    [anon_sym___thread] = ACTIONS(2588),
    [anon_sym_CG_EXTERN] = ACTIONS(2588),
    [anon_sym_CG_INLINE] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2588),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2588),
    [anon_sym_IBOutlet] = ACTIONS(2588),
    [anon_sym_IBInspectable] = ACTIONS(2588),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2588),
    [anon_sym_NS_INLINE] = ACTIONS(2588),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2588),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2588),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2588),
    [anon_sym_const] = ACTIONS(2588),
    [anon_sym_constexpr] = ACTIONS(2588),
    [anon_sym_volatile] = ACTIONS(2588),
    [anon_sym_restrict] = ACTIONS(2588),
    [anon_sym___restrict__] = ACTIONS(2588),
    [anon_sym__Atomic] = ACTIONS(2588),
    [anon_sym__Noreturn] = ACTIONS(2588),
    [anon_sym__Nonnull] = ACTIONS(2588),
    [anon_sym_nullable] = ACTIONS(2588),
    [anon_sym__Complex] = ACTIONS(2588),
    [anon_sym__Nullable] = ACTIONS(2588),
    [anon_sym__Nullable_result] = ACTIONS(2588),
    [anon_sym__Null_unspecified] = ACTIONS(2588),
    [anon_sym___autoreleasing] = ACTIONS(2588),
    [anon_sym___block] = ACTIONS(2588),
    [anon_sym___bridge] = ACTIONS(2588),
    [anon_sym___bridge_retained] = ACTIONS(2588),
    [anon_sym___bridge_transfer] = ACTIONS(2588),
    [anon_sym___complex] = ACTIONS(2588),
    [anon_sym___const] = ACTIONS(2588),
    [anon_sym___imag] = ACTIONS(2588),
    [anon_sym___kindof] = ACTIONS(2588),
    [anon_sym___nonnull] = ACTIONS(2588),
    [anon_sym___nullable] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2588),
    [anon_sym___real] = ACTIONS(2588),
    [anon_sym___strong] = ACTIONS(2588),
    [anon_sym___unsafe_unretained] = ACTIONS(2588),
    [anon_sym___unused] = ACTIONS(2588),
    [anon_sym___weak] = ACTIONS(2588),
    [anon_sym_alignas] = ACTIONS(2588),
    [anon_sym__Alignas] = ACTIONS(2588),
    [sym_primitive_type] = ACTIONS(2588),
    [anon_sym_enum] = ACTIONS(2588),
    [anon_sym_struct] = ACTIONS(2588),
    [anon_sym_union] = ACTIONS(2588),
    [anon_sym_if] = ACTIONS(2588),
    [anon_sym_switch] = ACTIONS(2588),
    [anon_sym_case] = ACTIONS(2588),
    [anon_sym_default] = ACTIONS(2588),
    [anon_sym_while] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2588),
    [anon_sym_for] = ACTIONS(2588),
    [anon_sym_in] = ACTIONS(2588),
    [anon_sym_return] = ACTIONS(2588),
    [anon_sym_break] = ACTIONS(2588),
    [anon_sym_continue] = ACTIONS(2588),
    [anon_sym_goto] = ACTIONS(2588),
    [anon_sym_DASH_DASH] = ACTIONS(2590),
    [anon_sym_PLUS_PLUS] = ACTIONS(2590),
    [anon_sym_sizeof] = ACTIONS(2588),
    [anon_sym___alignof__] = ACTIONS(2588),
    [anon_sym___alignof] = ACTIONS(2588),
    [anon_sym__alignof] = ACTIONS(2588),
    [anon_sym_alignof] = ACTIONS(2588),
    [anon_sym__Alignof] = ACTIONS(2588),
    [anon_sym_offsetof] = ACTIONS(2588),
    [anon_sym__Generic] = ACTIONS(2588),
    [anon_sym_asm] = ACTIONS(2588),
    [anon_sym___asm__] = ACTIONS(2588),
    [anon_sym___asm] = ACTIONS(2588),
    [sym_number_literal] = ACTIONS(2590),
    [anon_sym_L_SQUOTE] = ACTIONS(2590),
    [anon_sym_u_SQUOTE] = ACTIONS(2590),
    [anon_sym_U_SQUOTE] = ACTIONS(2590),
    [anon_sym_u8_SQUOTE] = ACTIONS(2590),
    [anon_sym_SQUOTE] = ACTIONS(2590),
    [anon_sym_AT] = ACTIONS(2588),
    [anon_sym_DQUOTE] = ACTIONS(2590),
    [anon_sym_L_DQUOTE] = ACTIONS(2590),
    [anon_sym_u_DQUOTE] = ACTIONS(2590),
    [anon_sym_U_DQUOTE] = ACTIONS(2590),
    [anon_sym_u8_DQUOTE] = ACTIONS(2590),
    [sym_true] = ACTIONS(2588),
    [sym_false] = ACTIONS(2588),
    [anon_sym_NULL] = ACTIONS(2588),
    [anon_sym_nullptr] = ACTIONS(2588),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2588),
    [anon_sym___typeof] = ACTIONS(2588),
    [anon_sym_typeof] = ACTIONS(2588),
    [anon_sym_ATimport] = ACTIONS(2590),
    [aux_sym_preproc_undef_token1] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2588),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2588),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2588),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2588),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_API_AVAILABLE] = ACTIONS(2588),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_API_DEPRECATED] = ACTIONS(2588),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym___deprecated_msg] = ACTIONS(2588),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2588),
    [anon_sym_ATprotocol] = ACTIONS(2590),
    [anon_sym_ATinterface] = ACTIONS(2590),
    [anon_sym_ATimplementation] = ACTIONS(2590),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2590),
    [anon_sym_ATsynthesize] = ACTIONS(2590),
    [anon_sym_ATdynamic] = ACTIONS(2590),
    [anon_sym_ATtry] = ACTIONS(2590),
    [anon_sym___try] = ACTIONS(2588),
    [anon_sym_ATthrow] = ACTIONS(2590),
    [anon_sym_ATselector] = ACTIONS(2590),
    [anon_sym_ATavailable] = ACTIONS(2590),
    [anon_sym___builtin_available] = ACTIONS(2588),
    [anon_sym_va_arg] = ACTIONS(2588),
    [anon_sym_ATencode] = ACTIONS(2590),
    [anon_sym_ATsynchronized] = ACTIONS(2590),
    [anon_sym_BOOL] = ACTIONS(2588),
    [anon_sym_IMP] = ACTIONS(2588),
    [anon_sym_SEL] = ACTIONS(2588),
    [anon_sym_Class] = ACTIONS(2588),
    [anon_sym_id] = ACTIONS(2588),
  },
  [STATE(282)] = {
    [sym_identifier] = ACTIONS(2592),
    [aux_sym_preproc_include_token1] = ACTIONS(2592),
    [aux_sym_preproc_include_token2] = ACTIONS(2592),
    [aux_sym_preproc_def_token1] = ACTIONS(2592),
    [aux_sym_preproc_if_token1] = ACTIONS(2592),
    [aux_sym_preproc_if_token2] = ACTIONS(2592),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2592),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2592),
    [aux_sym_preproc_else_token1] = ACTIONS(2592),
    [aux_sym_preproc_elif_token1] = ACTIONS(2592),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2592),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2592),
    [sym_preproc_directive] = ACTIONS(2592),
    [anon_sym_LPAREN2] = ACTIONS(2594),
    [anon_sym_BANG] = ACTIONS(2594),
    [anon_sym_TILDE] = ACTIONS(2594),
    [anon_sym_DASH] = ACTIONS(2592),
    [anon_sym_PLUS] = ACTIONS(2592),
    [anon_sym_STAR] = ACTIONS(2594),
    [anon_sym_CARET] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(2594),
    [anon_sym_SEMI] = ACTIONS(2594),
    [anon_sym___extension__] = ACTIONS(2592),
    [anon_sym_typedef] = ACTIONS(2592),
    [anon_sym_extern] = ACTIONS(2592),
    [anon_sym___attribute__] = ACTIONS(2592),
    [anon_sym___attribute] = ACTIONS(2592),
    [anon_sym_noreturn] = ACTIONS(2592),
    [anon_sym_LBRACK] = ACTIONS(2594),
    [anon_sym___declspec] = ACTIONS(2592),
    [anon_sym___cdecl] = ACTIONS(2592),
    [anon_sym___clrcall] = ACTIONS(2592),
    [anon_sym___stdcall] = ACTIONS(2592),
    [anon_sym___fastcall] = ACTIONS(2592),
    [anon_sym___thiscall] = ACTIONS(2592),
    [anon_sym___vectorcall] = ACTIONS(2592),
    [anon_sym_LBRACE] = ACTIONS(2594),
    [anon_sym_signed] = ACTIONS(2592),
    [anon_sym_unsigned] = ACTIONS(2592),
    [anon_sym_long] = ACTIONS(2592),
    [anon_sym_short] = ACTIONS(2592),
    [anon_sym_static] = ACTIONS(2592),
    [anon_sym_ATautoreleasepool] = ACTIONS(2594),
    [anon_sym_auto] = ACTIONS(2592),
    [anon_sym_register] = ACTIONS(2592),
    [anon_sym_inline] = ACTIONS(2592),
    [anon_sym___inline] = ACTIONS(2592),
    [anon_sym___inline__] = ACTIONS(2592),
    [anon_sym___forceinline] = ACTIONS(2592),
    [anon_sym_thread_local] = ACTIONS(2592),
    [anon_sym___thread] = ACTIONS(2592),
    [anon_sym_CG_EXTERN] = ACTIONS(2592),
    [anon_sym_CG_INLINE] = ACTIONS(2592),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2592),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2592),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2592),
    [anon_sym_IBOutlet] = ACTIONS(2592),
    [anon_sym_IBInspectable] = ACTIONS(2592),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2592),
    [anon_sym_NS_INLINE] = ACTIONS(2592),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2592),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2592),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2592),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2592),
    [anon_sym_const] = ACTIONS(2592),
    [anon_sym_constexpr] = ACTIONS(2592),
    [anon_sym_volatile] = ACTIONS(2592),
    [anon_sym_restrict] = ACTIONS(2592),
    [anon_sym___restrict__] = ACTIONS(2592),
    [anon_sym__Atomic] = ACTIONS(2592),
    [anon_sym__Noreturn] = ACTIONS(2592),
    [anon_sym__Nonnull] = ACTIONS(2592),
    [anon_sym_nullable] = ACTIONS(2592),
    [anon_sym__Complex] = ACTIONS(2592),
    [anon_sym__Nullable] = ACTIONS(2592),
    [anon_sym__Nullable_result] = ACTIONS(2592),
    [anon_sym__Null_unspecified] = ACTIONS(2592),
    [anon_sym___autoreleasing] = ACTIONS(2592),
    [anon_sym___block] = ACTIONS(2592),
    [anon_sym___bridge] = ACTIONS(2592),
    [anon_sym___bridge_retained] = ACTIONS(2592),
    [anon_sym___bridge_transfer] = ACTIONS(2592),
    [anon_sym___complex] = ACTIONS(2592),
    [anon_sym___const] = ACTIONS(2592),
    [anon_sym___imag] = ACTIONS(2592),
    [anon_sym___kindof] = ACTIONS(2592),
    [anon_sym___nonnull] = ACTIONS(2592),
    [anon_sym___nullable] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2592),
    [anon_sym___real] = ACTIONS(2592),
    [anon_sym___strong] = ACTIONS(2592),
    [anon_sym___unsafe_unretained] = ACTIONS(2592),
    [anon_sym___unused] = ACTIONS(2592),
    [anon_sym___weak] = ACTIONS(2592),
    [anon_sym_alignas] = ACTIONS(2592),
    [anon_sym__Alignas] = ACTIONS(2592),
    [sym_primitive_type] = ACTIONS(2592),
    [anon_sym_enum] = ACTIONS(2592),
    [anon_sym_struct] = ACTIONS(2592),
    [anon_sym_union] = ACTIONS(2592),
    [anon_sym_if] = ACTIONS(2592),
    [anon_sym_switch] = ACTIONS(2592),
    [anon_sym_case] = ACTIONS(2592),
    [anon_sym_default] = ACTIONS(2592),
    [anon_sym_while] = ACTIONS(2592),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_for] = ACTIONS(2592),
    [anon_sym_in] = ACTIONS(2592),
    [anon_sym_return] = ACTIONS(2592),
    [anon_sym_break] = ACTIONS(2592),
    [anon_sym_continue] = ACTIONS(2592),
    [anon_sym_goto] = ACTIONS(2592),
    [anon_sym_DASH_DASH] = ACTIONS(2594),
    [anon_sym_PLUS_PLUS] = ACTIONS(2594),
    [anon_sym_sizeof] = ACTIONS(2592),
    [anon_sym___alignof__] = ACTIONS(2592),
    [anon_sym___alignof] = ACTIONS(2592),
    [anon_sym__alignof] = ACTIONS(2592),
    [anon_sym_alignof] = ACTIONS(2592),
    [anon_sym__Alignof] = ACTIONS(2592),
    [anon_sym_offsetof] = ACTIONS(2592),
    [anon_sym__Generic] = ACTIONS(2592),
    [anon_sym_asm] = ACTIONS(2592),
    [anon_sym___asm__] = ACTIONS(2592),
    [anon_sym___asm] = ACTIONS(2592),
    [sym_number_literal] = ACTIONS(2594),
    [anon_sym_L_SQUOTE] = ACTIONS(2594),
    [anon_sym_u_SQUOTE] = ACTIONS(2594),
    [anon_sym_U_SQUOTE] = ACTIONS(2594),
    [anon_sym_u8_SQUOTE] = ACTIONS(2594),
    [anon_sym_SQUOTE] = ACTIONS(2594),
    [anon_sym_AT] = ACTIONS(2592),
    [anon_sym_DQUOTE] = ACTIONS(2594),
    [anon_sym_L_DQUOTE] = ACTIONS(2594),
    [anon_sym_u_DQUOTE] = ACTIONS(2594),
    [anon_sym_U_DQUOTE] = ACTIONS(2594),
    [anon_sym_u8_DQUOTE] = ACTIONS(2594),
    [sym_true] = ACTIONS(2592),
    [sym_false] = ACTIONS(2592),
    [anon_sym_NULL] = ACTIONS(2592),
    [anon_sym_nullptr] = ACTIONS(2592),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2592),
    [anon_sym___typeof] = ACTIONS(2592),
    [anon_sym_typeof] = ACTIONS(2592),
    [anon_sym_ATimport] = ACTIONS(2594),
    [aux_sym_preproc_undef_token1] = ACTIONS(2592),
    [anon_sym_POUND] = ACTIONS(2592),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2592),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2592),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2592),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2592),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2592),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2592),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2592),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_API_AVAILABLE] = ACTIONS(2592),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_API_DEPRECATED] = ACTIONS(2592),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2592),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym___deprecated_msg] = ACTIONS(2592),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2592),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2592),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2592),
    [anon_sym_ATprotocol] = ACTIONS(2594),
    [anon_sym_ATinterface] = ACTIONS(2594),
    [anon_sym_ATimplementation] = ACTIONS(2594),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2594),
    [anon_sym_ATsynthesize] = ACTIONS(2594),
    [anon_sym_ATdynamic] = ACTIONS(2594),
    [anon_sym_ATtry] = ACTIONS(2594),
    [anon_sym___try] = ACTIONS(2592),
    [anon_sym_ATthrow] = ACTIONS(2594),
    [anon_sym_ATselector] = ACTIONS(2594),
    [anon_sym_ATavailable] = ACTIONS(2594),
    [anon_sym___builtin_available] = ACTIONS(2592),
    [anon_sym_va_arg] = ACTIONS(2592),
    [anon_sym_ATencode] = ACTIONS(2594),
    [anon_sym_ATsynchronized] = ACTIONS(2594),
    [anon_sym_BOOL] = ACTIONS(2592),
    [anon_sym_IMP] = ACTIONS(2592),
    [anon_sym_SEL] = ACTIONS(2592),
    [anon_sym_Class] = ACTIONS(2592),
    [anon_sym_id] = ACTIONS(2592),
  },
  [STATE(283)] = {
    [sym_identifier] = ACTIONS(2596),
    [aux_sym_preproc_include_token1] = ACTIONS(2596),
    [aux_sym_preproc_include_token2] = ACTIONS(2596),
    [aux_sym_preproc_def_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token2] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2596),
    [aux_sym_preproc_else_token1] = ACTIONS(2596),
    [aux_sym_preproc_elif_token1] = ACTIONS(2596),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2596),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2596),
    [sym_preproc_directive] = ACTIONS(2596),
    [anon_sym_LPAREN2] = ACTIONS(2598),
    [anon_sym_BANG] = ACTIONS(2598),
    [anon_sym_TILDE] = ACTIONS(2598),
    [anon_sym_DASH] = ACTIONS(2596),
    [anon_sym_PLUS] = ACTIONS(2596),
    [anon_sym_STAR] = ACTIONS(2598),
    [anon_sym_CARET] = ACTIONS(2598),
    [anon_sym_AMP] = ACTIONS(2598),
    [anon_sym_SEMI] = ACTIONS(2598),
    [anon_sym___extension__] = ACTIONS(2596),
    [anon_sym_typedef] = ACTIONS(2596),
    [anon_sym_extern] = ACTIONS(2596),
    [anon_sym___attribute__] = ACTIONS(2596),
    [anon_sym___attribute] = ACTIONS(2596),
    [anon_sym_noreturn] = ACTIONS(2596),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym___declspec] = ACTIONS(2596),
    [anon_sym___cdecl] = ACTIONS(2596),
    [anon_sym___clrcall] = ACTIONS(2596),
    [anon_sym___stdcall] = ACTIONS(2596),
    [anon_sym___fastcall] = ACTIONS(2596),
    [anon_sym___thiscall] = ACTIONS(2596),
    [anon_sym___vectorcall] = ACTIONS(2596),
    [anon_sym_LBRACE] = ACTIONS(2598),
    [anon_sym_signed] = ACTIONS(2596),
    [anon_sym_unsigned] = ACTIONS(2596),
    [anon_sym_long] = ACTIONS(2596),
    [anon_sym_short] = ACTIONS(2596),
    [anon_sym_static] = ACTIONS(2596),
    [anon_sym_ATautoreleasepool] = ACTIONS(2598),
    [anon_sym_auto] = ACTIONS(2596),
    [anon_sym_register] = ACTIONS(2596),
    [anon_sym_inline] = ACTIONS(2596),
    [anon_sym___inline] = ACTIONS(2596),
    [anon_sym___inline__] = ACTIONS(2596),
    [anon_sym___forceinline] = ACTIONS(2596),
    [anon_sym_thread_local] = ACTIONS(2596),
    [anon_sym___thread] = ACTIONS(2596),
    [anon_sym_CG_EXTERN] = ACTIONS(2596),
    [anon_sym_CG_INLINE] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2596),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2596),
    [anon_sym_IBOutlet] = ACTIONS(2596),
    [anon_sym_IBInspectable] = ACTIONS(2596),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2596),
    [anon_sym_NS_INLINE] = ACTIONS(2596),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2596),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2596),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2596),
    [anon_sym_const] = ACTIONS(2596),
    [anon_sym_constexpr] = ACTIONS(2596),
    [anon_sym_volatile] = ACTIONS(2596),
    [anon_sym_restrict] = ACTIONS(2596),
    [anon_sym___restrict__] = ACTIONS(2596),
    [anon_sym__Atomic] = ACTIONS(2596),
    [anon_sym__Noreturn] = ACTIONS(2596),
    [anon_sym__Nonnull] = ACTIONS(2596),
    [anon_sym_nullable] = ACTIONS(2596),
    [anon_sym__Complex] = ACTIONS(2596),
    [anon_sym__Nullable] = ACTIONS(2596),
    [anon_sym__Nullable_result] = ACTIONS(2596),
    [anon_sym__Null_unspecified] = ACTIONS(2596),
    [anon_sym___autoreleasing] = ACTIONS(2596),
    [anon_sym___block] = ACTIONS(2596),
    [anon_sym___bridge] = ACTIONS(2596),
    [anon_sym___bridge_retained] = ACTIONS(2596),
    [anon_sym___bridge_transfer] = ACTIONS(2596),
    [anon_sym___complex] = ACTIONS(2596),
    [anon_sym___const] = ACTIONS(2596),
    [anon_sym___imag] = ACTIONS(2596),
    [anon_sym___kindof] = ACTIONS(2596),
    [anon_sym___nonnull] = ACTIONS(2596),
    [anon_sym___nullable] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2596),
    [anon_sym___real] = ACTIONS(2596),
    [anon_sym___strong] = ACTIONS(2596),
    [anon_sym___unsafe_unretained] = ACTIONS(2596),
    [anon_sym___unused] = ACTIONS(2596),
    [anon_sym___weak] = ACTIONS(2596),
    [anon_sym_alignas] = ACTIONS(2596),
    [anon_sym__Alignas] = ACTIONS(2596),
    [sym_primitive_type] = ACTIONS(2596),
    [anon_sym_enum] = ACTIONS(2596),
    [anon_sym_struct] = ACTIONS(2596),
    [anon_sym_union] = ACTIONS(2596),
    [anon_sym_if] = ACTIONS(2596),
    [anon_sym_switch] = ACTIONS(2596),
    [anon_sym_case] = ACTIONS(2596),
    [anon_sym_default] = ACTIONS(2596),
    [anon_sym_while] = ACTIONS(2596),
    [anon_sym_do] = ACTIONS(2596),
    [anon_sym_for] = ACTIONS(2596),
    [anon_sym_in] = ACTIONS(2596),
    [anon_sym_return] = ACTIONS(2596),
    [anon_sym_break] = ACTIONS(2596),
    [anon_sym_continue] = ACTIONS(2596),
    [anon_sym_goto] = ACTIONS(2596),
    [anon_sym_DASH_DASH] = ACTIONS(2598),
    [anon_sym_PLUS_PLUS] = ACTIONS(2598),
    [anon_sym_sizeof] = ACTIONS(2596),
    [anon_sym___alignof__] = ACTIONS(2596),
    [anon_sym___alignof] = ACTIONS(2596),
    [anon_sym__alignof] = ACTIONS(2596),
    [anon_sym_alignof] = ACTIONS(2596),
    [anon_sym__Alignof] = ACTIONS(2596),
    [anon_sym_offsetof] = ACTIONS(2596),
    [anon_sym__Generic] = ACTIONS(2596),
    [anon_sym_asm] = ACTIONS(2596),
    [anon_sym___asm__] = ACTIONS(2596),
    [anon_sym___asm] = ACTIONS(2596),
    [sym_number_literal] = ACTIONS(2598),
    [anon_sym_L_SQUOTE] = ACTIONS(2598),
    [anon_sym_u_SQUOTE] = ACTIONS(2598),
    [anon_sym_U_SQUOTE] = ACTIONS(2598),
    [anon_sym_u8_SQUOTE] = ACTIONS(2598),
    [anon_sym_SQUOTE] = ACTIONS(2598),
    [anon_sym_AT] = ACTIONS(2596),
    [anon_sym_DQUOTE] = ACTIONS(2598),
    [anon_sym_L_DQUOTE] = ACTIONS(2598),
    [anon_sym_u_DQUOTE] = ACTIONS(2598),
    [anon_sym_U_DQUOTE] = ACTIONS(2598),
    [anon_sym_u8_DQUOTE] = ACTIONS(2598),
    [sym_true] = ACTIONS(2596),
    [sym_false] = ACTIONS(2596),
    [anon_sym_NULL] = ACTIONS(2596),
    [anon_sym_nullptr] = ACTIONS(2596),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2596),
    [anon_sym___typeof] = ACTIONS(2596),
    [anon_sym_typeof] = ACTIONS(2596),
    [anon_sym_ATimport] = ACTIONS(2598),
    [aux_sym_preproc_undef_token1] = ACTIONS(2596),
    [anon_sym_POUND] = ACTIONS(2596),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2596),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2596),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2596),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_API_AVAILABLE] = ACTIONS(2596),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_API_DEPRECATED] = ACTIONS(2596),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym___deprecated_msg] = ACTIONS(2596),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2596),
    [anon_sym_ATprotocol] = ACTIONS(2598),
    [anon_sym_ATinterface] = ACTIONS(2598),
    [anon_sym_ATimplementation] = ACTIONS(2598),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2598),
    [anon_sym_ATsynthesize] = ACTIONS(2598),
    [anon_sym_ATdynamic] = ACTIONS(2598),
    [anon_sym_ATtry] = ACTIONS(2598),
    [anon_sym___try] = ACTIONS(2596),
    [anon_sym_ATthrow] = ACTIONS(2598),
    [anon_sym_ATselector] = ACTIONS(2598),
    [anon_sym_ATavailable] = ACTIONS(2598),
    [anon_sym___builtin_available] = ACTIONS(2596),
    [anon_sym_va_arg] = ACTIONS(2596),
    [anon_sym_ATencode] = ACTIONS(2598),
    [anon_sym_ATsynchronized] = ACTIONS(2598),
    [anon_sym_BOOL] = ACTIONS(2596),
    [anon_sym_IMP] = ACTIONS(2596),
    [anon_sym_SEL] = ACTIONS(2596),
    [anon_sym_Class] = ACTIONS(2596),
    [anon_sym_id] = ACTIONS(2596),
  },
  [STATE(284)] = {
    [sym_identifier] = ACTIONS(2600),
    [aux_sym_preproc_include_token1] = ACTIONS(2600),
    [aux_sym_preproc_include_token2] = ACTIONS(2600),
    [aux_sym_preproc_def_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token2] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2600),
    [aux_sym_preproc_else_token1] = ACTIONS(2600),
    [aux_sym_preproc_elif_token1] = ACTIONS(2600),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2600),
    [sym_preproc_directive] = ACTIONS(2600),
    [anon_sym_LPAREN2] = ACTIONS(2602),
    [anon_sym_BANG] = ACTIONS(2602),
    [anon_sym_TILDE] = ACTIONS(2602),
    [anon_sym_DASH] = ACTIONS(2600),
    [anon_sym_PLUS] = ACTIONS(2600),
    [anon_sym_STAR] = ACTIONS(2602),
    [anon_sym_CARET] = ACTIONS(2602),
    [anon_sym_AMP] = ACTIONS(2602),
    [anon_sym_SEMI] = ACTIONS(2602),
    [anon_sym___extension__] = ACTIONS(2600),
    [anon_sym_typedef] = ACTIONS(2600),
    [anon_sym_extern] = ACTIONS(2600),
    [anon_sym___attribute__] = ACTIONS(2600),
    [anon_sym___attribute] = ACTIONS(2600),
    [anon_sym_noreturn] = ACTIONS(2600),
    [anon_sym_LBRACK] = ACTIONS(2602),
    [anon_sym___declspec] = ACTIONS(2600),
    [anon_sym___cdecl] = ACTIONS(2600),
    [anon_sym___clrcall] = ACTIONS(2600),
    [anon_sym___stdcall] = ACTIONS(2600),
    [anon_sym___fastcall] = ACTIONS(2600),
    [anon_sym___thiscall] = ACTIONS(2600),
    [anon_sym___vectorcall] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_signed] = ACTIONS(2600),
    [anon_sym_unsigned] = ACTIONS(2600),
    [anon_sym_long] = ACTIONS(2600),
    [anon_sym_short] = ACTIONS(2600),
    [anon_sym_static] = ACTIONS(2600),
    [anon_sym_ATautoreleasepool] = ACTIONS(2602),
    [anon_sym_auto] = ACTIONS(2600),
    [anon_sym_register] = ACTIONS(2600),
    [anon_sym_inline] = ACTIONS(2600),
    [anon_sym___inline] = ACTIONS(2600),
    [anon_sym___inline__] = ACTIONS(2600),
    [anon_sym___forceinline] = ACTIONS(2600),
    [anon_sym_thread_local] = ACTIONS(2600),
    [anon_sym___thread] = ACTIONS(2600),
    [anon_sym_CG_EXTERN] = ACTIONS(2600),
    [anon_sym_CG_INLINE] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2600),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2600),
    [anon_sym_IBOutlet] = ACTIONS(2600),
    [anon_sym_IBInspectable] = ACTIONS(2600),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2600),
    [anon_sym_NS_INLINE] = ACTIONS(2600),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2600),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2600),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2600),
    [anon_sym_const] = ACTIONS(2600),
    [anon_sym_constexpr] = ACTIONS(2600),
    [anon_sym_volatile] = ACTIONS(2600),
    [anon_sym_restrict] = ACTIONS(2600),
    [anon_sym___restrict__] = ACTIONS(2600),
    [anon_sym__Atomic] = ACTIONS(2600),
    [anon_sym__Noreturn] = ACTIONS(2600),
    [anon_sym__Nonnull] = ACTIONS(2600),
    [anon_sym_nullable] = ACTIONS(2600),
    [anon_sym__Complex] = ACTIONS(2600),
    [anon_sym__Nullable] = ACTIONS(2600),
    [anon_sym__Nullable_result] = ACTIONS(2600),
    [anon_sym__Null_unspecified] = ACTIONS(2600),
    [anon_sym___autoreleasing] = ACTIONS(2600),
    [anon_sym___block] = ACTIONS(2600),
    [anon_sym___bridge] = ACTIONS(2600),
    [anon_sym___bridge_retained] = ACTIONS(2600),
    [anon_sym___bridge_transfer] = ACTIONS(2600),
    [anon_sym___complex] = ACTIONS(2600),
    [anon_sym___const] = ACTIONS(2600),
    [anon_sym___imag] = ACTIONS(2600),
    [anon_sym___kindof] = ACTIONS(2600),
    [anon_sym___nonnull] = ACTIONS(2600),
    [anon_sym___nullable] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2600),
    [anon_sym___real] = ACTIONS(2600),
    [anon_sym___strong] = ACTIONS(2600),
    [anon_sym___unsafe_unretained] = ACTIONS(2600),
    [anon_sym___unused] = ACTIONS(2600),
    [anon_sym___weak] = ACTIONS(2600),
    [anon_sym_alignas] = ACTIONS(2600),
    [anon_sym__Alignas] = ACTIONS(2600),
    [sym_primitive_type] = ACTIONS(2600),
    [anon_sym_enum] = ACTIONS(2600),
    [anon_sym_struct] = ACTIONS(2600),
    [anon_sym_union] = ACTIONS(2600),
    [anon_sym_if] = ACTIONS(2600),
    [anon_sym_switch] = ACTIONS(2600),
    [anon_sym_case] = ACTIONS(2600),
    [anon_sym_default] = ACTIONS(2600),
    [anon_sym_while] = ACTIONS(2600),
    [anon_sym_do] = ACTIONS(2600),
    [anon_sym_for] = ACTIONS(2600),
    [anon_sym_in] = ACTIONS(2600),
    [anon_sym_return] = ACTIONS(2600),
    [anon_sym_break] = ACTIONS(2600),
    [anon_sym_continue] = ACTIONS(2600),
    [anon_sym_goto] = ACTIONS(2600),
    [anon_sym_DASH_DASH] = ACTIONS(2602),
    [anon_sym_PLUS_PLUS] = ACTIONS(2602),
    [anon_sym_sizeof] = ACTIONS(2600),
    [anon_sym___alignof__] = ACTIONS(2600),
    [anon_sym___alignof] = ACTIONS(2600),
    [anon_sym__alignof] = ACTIONS(2600),
    [anon_sym_alignof] = ACTIONS(2600),
    [anon_sym__Alignof] = ACTIONS(2600),
    [anon_sym_offsetof] = ACTIONS(2600),
    [anon_sym__Generic] = ACTIONS(2600),
    [anon_sym_asm] = ACTIONS(2600),
    [anon_sym___asm__] = ACTIONS(2600),
    [anon_sym___asm] = ACTIONS(2600),
    [sym_number_literal] = ACTIONS(2602),
    [anon_sym_L_SQUOTE] = ACTIONS(2602),
    [anon_sym_u_SQUOTE] = ACTIONS(2602),
    [anon_sym_U_SQUOTE] = ACTIONS(2602),
    [anon_sym_u8_SQUOTE] = ACTIONS(2602),
    [anon_sym_SQUOTE] = ACTIONS(2602),
    [anon_sym_AT] = ACTIONS(2600),
    [anon_sym_DQUOTE] = ACTIONS(2602),
    [anon_sym_L_DQUOTE] = ACTIONS(2602),
    [anon_sym_u_DQUOTE] = ACTIONS(2602),
    [anon_sym_U_DQUOTE] = ACTIONS(2602),
    [anon_sym_u8_DQUOTE] = ACTIONS(2602),
    [sym_true] = ACTIONS(2600),
    [sym_false] = ACTIONS(2600),
    [anon_sym_NULL] = ACTIONS(2600),
    [anon_sym_nullptr] = ACTIONS(2600),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2600),
    [anon_sym___typeof] = ACTIONS(2600),
    [anon_sym_typeof] = ACTIONS(2600),
    [anon_sym_ATimport] = ACTIONS(2602),
    [aux_sym_preproc_undef_token1] = ACTIONS(2600),
    [anon_sym_POUND] = ACTIONS(2600),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2600),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2600),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2600),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_API_AVAILABLE] = ACTIONS(2600),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_API_DEPRECATED] = ACTIONS(2600),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym___deprecated_msg] = ACTIONS(2600),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2600),
    [anon_sym_ATprotocol] = ACTIONS(2602),
    [anon_sym_ATinterface] = ACTIONS(2602),
    [anon_sym_ATimplementation] = ACTIONS(2602),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2602),
    [anon_sym_ATsynthesize] = ACTIONS(2602),
    [anon_sym_ATdynamic] = ACTIONS(2602),
    [anon_sym_ATtry] = ACTIONS(2602),
    [anon_sym___try] = ACTIONS(2600),
    [anon_sym_ATthrow] = ACTIONS(2602),
    [anon_sym_ATselector] = ACTIONS(2602),
    [anon_sym_ATavailable] = ACTIONS(2602),
    [anon_sym___builtin_available] = ACTIONS(2600),
    [anon_sym_va_arg] = ACTIONS(2600),
    [anon_sym_ATencode] = ACTIONS(2602),
    [anon_sym_ATsynchronized] = ACTIONS(2602),
    [anon_sym_BOOL] = ACTIONS(2600),
    [anon_sym_IMP] = ACTIONS(2600),
    [anon_sym_SEL] = ACTIONS(2600),
    [anon_sym_Class] = ACTIONS(2600),
    [anon_sym_id] = ACTIONS(2600),
  },
  [STATE(285)] = {
    [sym_identifier] = ACTIONS(2604),
    [aux_sym_preproc_include_token1] = ACTIONS(2604),
    [aux_sym_preproc_include_token2] = ACTIONS(2604),
    [aux_sym_preproc_def_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token2] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2604),
    [aux_sym_preproc_else_token1] = ACTIONS(2604),
    [aux_sym_preproc_elif_token1] = ACTIONS(2604),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2604),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2604),
    [sym_preproc_directive] = ACTIONS(2604),
    [anon_sym_LPAREN2] = ACTIONS(2606),
    [anon_sym_BANG] = ACTIONS(2606),
    [anon_sym_TILDE] = ACTIONS(2606),
    [anon_sym_DASH] = ACTIONS(2604),
    [anon_sym_PLUS] = ACTIONS(2604),
    [anon_sym_STAR] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(2606),
    [anon_sym_AMP] = ACTIONS(2606),
    [anon_sym_SEMI] = ACTIONS(2606),
    [anon_sym___extension__] = ACTIONS(2604),
    [anon_sym_typedef] = ACTIONS(2604),
    [anon_sym_extern] = ACTIONS(2604),
    [anon_sym___attribute__] = ACTIONS(2604),
    [anon_sym___attribute] = ACTIONS(2604),
    [anon_sym_noreturn] = ACTIONS(2604),
    [anon_sym_LBRACK] = ACTIONS(2606),
    [anon_sym___declspec] = ACTIONS(2604),
    [anon_sym___cdecl] = ACTIONS(2604),
    [anon_sym___clrcall] = ACTIONS(2604),
    [anon_sym___stdcall] = ACTIONS(2604),
    [anon_sym___fastcall] = ACTIONS(2604),
    [anon_sym___thiscall] = ACTIONS(2604),
    [anon_sym___vectorcall] = ACTIONS(2604),
    [anon_sym_LBRACE] = ACTIONS(2606),
    [anon_sym_signed] = ACTIONS(2604),
    [anon_sym_unsigned] = ACTIONS(2604),
    [anon_sym_long] = ACTIONS(2604),
    [anon_sym_short] = ACTIONS(2604),
    [anon_sym_static] = ACTIONS(2604),
    [anon_sym_ATautoreleasepool] = ACTIONS(2606),
    [anon_sym_auto] = ACTIONS(2604),
    [anon_sym_register] = ACTIONS(2604),
    [anon_sym_inline] = ACTIONS(2604),
    [anon_sym___inline] = ACTIONS(2604),
    [anon_sym___inline__] = ACTIONS(2604),
    [anon_sym___forceinline] = ACTIONS(2604),
    [anon_sym_thread_local] = ACTIONS(2604),
    [anon_sym___thread] = ACTIONS(2604),
    [anon_sym_CG_EXTERN] = ACTIONS(2604),
    [anon_sym_CG_INLINE] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2604),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2604),
    [anon_sym_IBOutlet] = ACTIONS(2604),
    [anon_sym_IBInspectable] = ACTIONS(2604),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2604),
    [anon_sym_NS_INLINE] = ACTIONS(2604),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2604),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2604),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2604),
    [anon_sym_const] = ACTIONS(2604),
    [anon_sym_constexpr] = ACTIONS(2604),
    [anon_sym_volatile] = ACTIONS(2604),
    [anon_sym_restrict] = ACTIONS(2604),
    [anon_sym___restrict__] = ACTIONS(2604),
    [anon_sym__Atomic] = ACTIONS(2604),
    [anon_sym__Noreturn] = ACTIONS(2604),
    [anon_sym__Nonnull] = ACTIONS(2604),
    [anon_sym_nullable] = ACTIONS(2604),
    [anon_sym__Complex] = ACTIONS(2604),
    [anon_sym__Nullable] = ACTIONS(2604),
    [anon_sym__Nullable_result] = ACTIONS(2604),
    [anon_sym__Null_unspecified] = ACTIONS(2604),
    [anon_sym___autoreleasing] = ACTIONS(2604),
    [anon_sym___block] = ACTIONS(2604),
    [anon_sym___bridge] = ACTIONS(2604),
    [anon_sym___bridge_retained] = ACTIONS(2604),
    [anon_sym___bridge_transfer] = ACTIONS(2604),
    [anon_sym___complex] = ACTIONS(2604),
    [anon_sym___const] = ACTIONS(2604),
    [anon_sym___imag] = ACTIONS(2604),
    [anon_sym___kindof] = ACTIONS(2604),
    [anon_sym___nonnull] = ACTIONS(2604),
    [anon_sym___nullable] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2604),
    [anon_sym___real] = ACTIONS(2604),
    [anon_sym___strong] = ACTIONS(2604),
    [anon_sym___unsafe_unretained] = ACTIONS(2604),
    [anon_sym___unused] = ACTIONS(2604),
    [anon_sym___weak] = ACTIONS(2604),
    [anon_sym_alignas] = ACTIONS(2604),
    [anon_sym__Alignas] = ACTIONS(2604),
    [sym_primitive_type] = ACTIONS(2604),
    [anon_sym_enum] = ACTIONS(2604),
    [anon_sym_struct] = ACTIONS(2604),
    [anon_sym_union] = ACTIONS(2604),
    [anon_sym_if] = ACTIONS(2604),
    [anon_sym_switch] = ACTIONS(2604),
    [anon_sym_case] = ACTIONS(2604),
    [anon_sym_default] = ACTIONS(2604),
    [anon_sym_while] = ACTIONS(2604),
    [anon_sym_do] = ACTIONS(2604),
    [anon_sym_for] = ACTIONS(2604),
    [anon_sym_in] = ACTIONS(2604),
    [anon_sym_return] = ACTIONS(2604),
    [anon_sym_break] = ACTIONS(2604),
    [anon_sym_continue] = ACTIONS(2604),
    [anon_sym_goto] = ACTIONS(2604),
    [anon_sym_DASH_DASH] = ACTIONS(2606),
    [anon_sym_PLUS_PLUS] = ACTIONS(2606),
    [anon_sym_sizeof] = ACTIONS(2604),
    [anon_sym___alignof__] = ACTIONS(2604),
    [anon_sym___alignof] = ACTIONS(2604),
    [anon_sym__alignof] = ACTIONS(2604),
    [anon_sym_alignof] = ACTIONS(2604),
    [anon_sym__Alignof] = ACTIONS(2604),
    [anon_sym_offsetof] = ACTIONS(2604),
    [anon_sym__Generic] = ACTIONS(2604),
    [anon_sym_asm] = ACTIONS(2604),
    [anon_sym___asm__] = ACTIONS(2604),
    [anon_sym___asm] = ACTIONS(2604),
    [sym_number_literal] = ACTIONS(2606),
    [anon_sym_L_SQUOTE] = ACTIONS(2606),
    [anon_sym_u_SQUOTE] = ACTIONS(2606),
    [anon_sym_U_SQUOTE] = ACTIONS(2606),
    [anon_sym_u8_SQUOTE] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2606),
    [anon_sym_AT] = ACTIONS(2604),
    [anon_sym_DQUOTE] = ACTIONS(2606),
    [anon_sym_L_DQUOTE] = ACTIONS(2606),
    [anon_sym_u_DQUOTE] = ACTIONS(2606),
    [anon_sym_U_DQUOTE] = ACTIONS(2606),
    [anon_sym_u8_DQUOTE] = ACTIONS(2606),
    [sym_true] = ACTIONS(2604),
    [sym_false] = ACTIONS(2604),
    [anon_sym_NULL] = ACTIONS(2604),
    [anon_sym_nullptr] = ACTIONS(2604),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2604),
    [anon_sym___typeof] = ACTIONS(2604),
    [anon_sym_typeof] = ACTIONS(2604),
    [anon_sym_ATimport] = ACTIONS(2606),
    [aux_sym_preproc_undef_token1] = ACTIONS(2604),
    [anon_sym_POUND] = ACTIONS(2604),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2604),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2604),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2604),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_API_AVAILABLE] = ACTIONS(2604),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_API_DEPRECATED] = ACTIONS(2604),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym___deprecated_msg] = ACTIONS(2604),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2604),
    [anon_sym_ATprotocol] = ACTIONS(2606),
    [anon_sym_ATinterface] = ACTIONS(2606),
    [anon_sym_ATimplementation] = ACTIONS(2606),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2606),
    [anon_sym_ATsynthesize] = ACTIONS(2606),
    [anon_sym_ATdynamic] = ACTIONS(2606),
    [anon_sym_ATtry] = ACTIONS(2606),
    [anon_sym___try] = ACTIONS(2604),
    [anon_sym_ATthrow] = ACTIONS(2606),
    [anon_sym_ATselector] = ACTIONS(2606),
    [anon_sym_ATavailable] = ACTIONS(2606),
    [anon_sym___builtin_available] = ACTIONS(2604),
    [anon_sym_va_arg] = ACTIONS(2604),
    [anon_sym_ATencode] = ACTIONS(2606),
    [anon_sym_ATsynchronized] = ACTIONS(2606),
    [anon_sym_BOOL] = ACTIONS(2604),
    [anon_sym_IMP] = ACTIONS(2604),
    [anon_sym_SEL] = ACTIONS(2604),
    [anon_sym_Class] = ACTIONS(2604),
    [anon_sym_id] = ACTIONS(2604),
  },
  [STATE(286)] = {
    [sym_identifier] = ACTIONS(2608),
    [aux_sym_preproc_include_token1] = ACTIONS(2608),
    [aux_sym_preproc_include_token2] = ACTIONS(2608),
    [aux_sym_preproc_def_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token2] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2608),
    [aux_sym_preproc_else_token1] = ACTIONS(2608),
    [aux_sym_preproc_elif_token1] = ACTIONS(2608),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2608),
    [sym_preproc_directive] = ACTIONS(2608),
    [anon_sym_LPAREN2] = ACTIONS(2610),
    [anon_sym_BANG] = ACTIONS(2610),
    [anon_sym_TILDE] = ACTIONS(2610),
    [anon_sym_DASH] = ACTIONS(2608),
    [anon_sym_PLUS] = ACTIONS(2608),
    [anon_sym_STAR] = ACTIONS(2610),
    [anon_sym_CARET] = ACTIONS(2610),
    [anon_sym_AMP] = ACTIONS(2610),
    [anon_sym_SEMI] = ACTIONS(2610),
    [anon_sym___extension__] = ACTIONS(2608),
    [anon_sym_typedef] = ACTIONS(2608),
    [anon_sym_extern] = ACTIONS(2608),
    [anon_sym___attribute__] = ACTIONS(2608),
    [anon_sym___attribute] = ACTIONS(2608),
    [anon_sym_noreturn] = ACTIONS(2608),
    [anon_sym_LBRACK] = ACTIONS(2610),
    [anon_sym___declspec] = ACTIONS(2608),
    [anon_sym___cdecl] = ACTIONS(2608),
    [anon_sym___clrcall] = ACTIONS(2608),
    [anon_sym___stdcall] = ACTIONS(2608),
    [anon_sym___fastcall] = ACTIONS(2608),
    [anon_sym___thiscall] = ACTIONS(2608),
    [anon_sym___vectorcall] = ACTIONS(2608),
    [anon_sym_LBRACE] = ACTIONS(2610),
    [anon_sym_signed] = ACTIONS(2608),
    [anon_sym_unsigned] = ACTIONS(2608),
    [anon_sym_long] = ACTIONS(2608),
    [anon_sym_short] = ACTIONS(2608),
    [anon_sym_static] = ACTIONS(2608),
    [anon_sym_ATautoreleasepool] = ACTIONS(2610),
    [anon_sym_auto] = ACTIONS(2608),
    [anon_sym_register] = ACTIONS(2608),
    [anon_sym_inline] = ACTIONS(2608),
    [anon_sym___inline] = ACTIONS(2608),
    [anon_sym___inline__] = ACTIONS(2608),
    [anon_sym___forceinline] = ACTIONS(2608),
    [anon_sym_thread_local] = ACTIONS(2608),
    [anon_sym___thread] = ACTIONS(2608),
    [anon_sym_CG_EXTERN] = ACTIONS(2608),
    [anon_sym_CG_INLINE] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2608),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2608),
    [anon_sym_IBOutlet] = ACTIONS(2608),
    [anon_sym_IBInspectable] = ACTIONS(2608),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2608),
    [anon_sym_NS_INLINE] = ACTIONS(2608),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2608),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2608),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2608),
    [anon_sym_const] = ACTIONS(2608),
    [anon_sym_constexpr] = ACTIONS(2608),
    [anon_sym_volatile] = ACTIONS(2608),
    [anon_sym_restrict] = ACTIONS(2608),
    [anon_sym___restrict__] = ACTIONS(2608),
    [anon_sym__Atomic] = ACTIONS(2608),
    [anon_sym__Noreturn] = ACTIONS(2608),
    [anon_sym__Nonnull] = ACTIONS(2608),
    [anon_sym_nullable] = ACTIONS(2608),
    [anon_sym__Complex] = ACTIONS(2608),
    [anon_sym__Nullable] = ACTIONS(2608),
    [anon_sym__Nullable_result] = ACTIONS(2608),
    [anon_sym__Null_unspecified] = ACTIONS(2608),
    [anon_sym___autoreleasing] = ACTIONS(2608),
    [anon_sym___block] = ACTIONS(2608),
    [anon_sym___bridge] = ACTIONS(2608),
    [anon_sym___bridge_retained] = ACTIONS(2608),
    [anon_sym___bridge_transfer] = ACTIONS(2608),
    [anon_sym___complex] = ACTIONS(2608),
    [anon_sym___const] = ACTIONS(2608),
    [anon_sym___imag] = ACTIONS(2608),
    [anon_sym___kindof] = ACTIONS(2608),
    [anon_sym___nonnull] = ACTIONS(2608),
    [anon_sym___nullable] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2608),
    [anon_sym___real] = ACTIONS(2608),
    [anon_sym___strong] = ACTIONS(2608),
    [anon_sym___unsafe_unretained] = ACTIONS(2608),
    [anon_sym___unused] = ACTIONS(2608),
    [anon_sym___weak] = ACTIONS(2608),
    [anon_sym_alignas] = ACTIONS(2608),
    [anon_sym__Alignas] = ACTIONS(2608),
    [sym_primitive_type] = ACTIONS(2608),
    [anon_sym_enum] = ACTIONS(2608),
    [anon_sym_struct] = ACTIONS(2608),
    [anon_sym_union] = ACTIONS(2608),
    [anon_sym_if] = ACTIONS(2608),
    [anon_sym_switch] = ACTIONS(2608),
    [anon_sym_case] = ACTIONS(2608),
    [anon_sym_default] = ACTIONS(2608),
    [anon_sym_while] = ACTIONS(2608),
    [anon_sym_do] = ACTIONS(2608),
    [anon_sym_for] = ACTIONS(2608),
    [anon_sym_in] = ACTIONS(2608),
    [anon_sym_return] = ACTIONS(2608),
    [anon_sym_break] = ACTIONS(2608),
    [anon_sym_continue] = ACTIONS(2608),
    [anon_sym_goto] = ACTIONS(2608),
    [anon_sym_DASH_DASH] = ACTIONS(2610),
    [anon_sym_PLUS_PLUS] = ACTIONS(2610),
    [anon_sym_sizeof] = ACTIONS(2608),
    [anon_sym___alignof__] = ACTIONS(2608),
    [anon_sym___alignof] = ACTIONS(2608),
    [anon_sym__alignof] = ACTIONS(2608),
    [anon_sym_alignof] = ACTIONS(2608),
    [anon_sym__Alignof] = ACTIONS(2608),
    [anon_sym_offsetof] = ACTIONS(2608),
    [anon_sym__Generic] = ACTIONS(2608),
    [anon_sym_asm] = ACTIONS(2608),
    [anon_sym___asm__] = ACTIONS(2608),
    [anon_sym___asm] = ACTIONS(2608),
    [sym_number_literal] = ACTIONS(2610),
    [anon_sym_L_SQUOTE] = ACTIONS(2610),
    [anon_sym_u_SQUOTE] = ACTIONS(2610),
    [anon_sym_U_SQUOTE] = ACTIONS(2610),
    [anon_sym_u8_SQUOTE] = ACTIONS(2610),
    [anon_sym_SQUOTE] = ACTIONS(2610),
    [anon_sym_AT] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [anon_sym_L_DQUOTE] = ACTIONS(2610),
    [anon_sym_u_DQUOTE] = ACTIONS(2610),
    [anon_sym_U_DQUOTE] = ACTIONS(2610),
    [anon_sym_u8_DQUOTE] = ACTIONS(2610),
    [sym_true] = ACTIONS(2608),
    [sym_false] = ACTIONS(2608),
    [anon_sym_NULL] = ACTIONS(2608),
    [anon_sym_nullptr] = ACTIONS(2608),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2608),
    [anon_sym___typeof] = ACTIONS(2608),
    [anon_sym_typeof] = ACTIONS(2608),
    [anon_sym_ATimport] = ACTIONS(2610),
    [aux_sym_preproc_undef_token1] = ACTIONS(2608),
    [anon_sym_POUND] = ACTIONS(2608),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2608),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2608),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2608),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_API_AVAILABLE] = ACTIONS(2608),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_API_DEPRECATED] = ACTIONS(2608),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym___deprecated_msg] = ACTIONS(2608),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2608),
    [anon_sym_ATprotocol] = ACTIONS(2610),
    [anon_sym_ATinterface] = ACTIONS(2610),
    [anon_sym_ATimplementation] = ACTIONS(2610),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2610),
    [anon_sym_ATsynthesize] = ACTIONS(2610),
    [anon_sym_ATdynamic] = ACTIONS(2610),
    [anon_sym_ATtry] = ACTIONS(2610),
    [anon_sym___try] = ACTIONS(2608),
    [anon_sym_ATthrow] = ACTIONS(2610),
    [anon_sym_ATselector] = ACTIONS(2610),
    [anon_sym_ATavailable] = ACTIONS(2610),
    [anon_sym___builtin_available] = ACTIONS(2608),
    [anon_sym_va_arg] = ACTIONS(2608),
    [anon_sym_ATencode] = ACTIONS(2610),
    [anon_sym_ATsynchronized] = ACTIONS(2610),
    [anon_sym_BOOL] = ACTIONS(2608),
    [anon_sym_IMP] = ACTIONS(2608),
    [anon_sym_SEL] = ACTIONS(2608),
    [anon_sym_Class] = ACTIONS(2608),
    [anon_sym_id] = ACTIONS(2608),
  },
  [STATE(287)] = {
    [sym_else_clause] = STATE(420),
    [sym_identifier] = ACTIONS(2394),
    [aux_sym_preproc_include_token1] = ACTIONS(2394),
    [aux_sym_preproc_include_token2] = ACTIONS(2394),
    [aux_sym_preproc_def_token1] = ACTIONS(2394),
    [aux_sym_preproc_if_token1] = ACTIONS(2394),
    [aux_sym_preproc_if_token2] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2394),
    [aux_sym_preproc_else_token1] = ACTIONS(2394),
    [aux_sym_preproc_elif_token1] = ACTIONS(2394),
    [sym_preproc_directive] = ACTIONS(2394),
    [anon_sym_LPAREN2] = ACTIONS(2396),
    [anon_sym_BANG] = ACTIONS(2396),
    [anon_sym_TILDE] = ACTIONS(2396),
    [anon_sym_DASH] = ACTIONS(2394),
    [anon_sym_PLUS] = ACTIONS(2394),
    [anon_sym_STAR] = ACTIONS(2396),
    [anon_sym_CARET] = ACTIONS(2396),
    [anon_sym_AMP] = ACTIONS(2396),
    [anon_sym_SEMI] = ACTIONS(2396),
    [anon_sym___extension__] = ACTIONS(2394),
    [anon_sym_typedef] = ACTIONS(2394),
    [anon_sym_extern] = ACTIONS(2394),
    [anon_sym___attribute__] = ACTIONS(2394),
    [anon_sym___attribute] = ACTIONS(2394),
    [anon_sym_noreturn] = ACTIONS(2394),
    [anon_sym_LBRACK] = ACTIONS(2396),
    [anon_sym___declspec] = ACTIONS(2394),
    [anon_sym___cdecl] = ACTIONS(2394),
    [anon_sym___clrcall] = ACTIONS(2394),
    [anon_sym___stdcall] = ACTIONS(2394),
    [anon_sym___fastcall] = ACTIONS(2394),
    [anon_sym___thiscall] = ACTIONS(2394),
    [anon_sym___vectorcall] = ACTIONS(2394),
    [anon_sym_LBRACE] = ACTIONS(2396),
    [anon_sym_signed] = ACTIONS(2394),
    [anon_sym_unsigned] = ACTIONS(2394),
    [anon_sym_long] = ACTIONS(2394),
    [anon_sym_short] = ACTIONS(2394),
    [anon_sym_static] = ACTIONS(2394),
    [anon_sym_ATautoreleasepool] = ACTIONS(2396),
    [anon_sym_auto] = ACTIONS(2394),
    [anon_sym_register] = ACTIONS(2394),
    [anon_sym_inline] = ACTIONS(2394),
    [anon_sym___inline] = ACTIONS(2394),
    [anon_sym___inline__] = ACTIONS(2394),
    [anon_sym___forceinline] = ACTIONS(2394),
    [anon_sym_thread_local] = ACTIONS(2394),
    [anon_sym___thread] = ACTIONS(2394),
    [anon_sym_CG_EXTERN] = ACTIONS(2394),
    [anon_sym_CG_INLINE] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2394),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2394),
    [anon_sym_IBOutlet] = ACTIONS(2394),
    [anon_sym_IBInspectable] = ACTIONS(2394),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2394),
    [anon_sym_NS_INLINE] = ACTIONS(2394),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2394),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2394),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2394),
    [anon_sym_const] = ACTIONS(2394),
    [anon_sym_constexpr] = ACTIONS(2394),
    [anon_sym_volatile] = ACTIONS(2394),
    [anon_sym_restrict] = ACTIONS(2394),
    [anon_sym___restrict__] = ACTIONS(2394),
    [anon_sym__Atomic] = ACTIONS(2394),
    [anon_sym__Noreturn] = ACTIONS(2394),
    [anon_sym__Nonnull] = ACTIONS(2394),
    [anon_sym_nullable] = ACTIONS(2394),
    [anon_sym__Complex] = ACTIONS(2394),
    [anon_sym__Nullable] = ACTIONS(2394),
    [anon_sym__Nullable_result] = ACTIONS(2394),
    [anon_sym__Null_unspecified] = ACTIONS(2394),
    [anon_sym___autoreleasing] = ACTIONS(2394),
    [anon_sym___block] = ACTIONS(2394),
    [anon_sym___bridge] = ACTIONS(2394),
    [anon_sym___bridge_retained] = ACTIONS(2394),
    [anon_sym___bridge_transfer] = ACTIONS(2394),
    [anon_sym___complex] = ACTIONS(2394),
    [anon_sym___const] = ACTIONS(2394),
    [anon_sym___imag] = ACTIONS(2394),
    [anon_sym___kindof] = ACTIONS(2394),
    [anon_sym___nonnull] = ACTIONS(2394),
    [anon_sym___nullable] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2394),
    [anon_sym___real] = ACTIONS(2394),
    [anon_sym___strong] = ACTIONS(2394),
    [anon_sym___unsafe_unretained] = ACTIONS(2394),
    [anon_sym___unused] = ACTIONS(2394),
    [anon_sym___weak] = ACTIONS(2394),
    [anon_sym_alignas] = ACTIONS(2394),
    [anon_sym__Alignas] = ACTIONS(2394),
    [sym_primitive_type] = ACTIONS(2394),
    [anon_sym_enum] = ACTIONS(2394),
    [anon_sym_struct] = ACTIONS(2394),
    [anon_sym_union] = ACTIONS(2394),
    [anon_sym_if] = ACTIONS(2394),
    [anon_sym_else] = ACTIONS(2612),
    [anon_sym_switch] = ACTIONS(2394),
    [anon_sym_case] = ACTIONS(2394),
    [anon_sym_default] = ACTIONS(2394),
    [anon_sym_while] = ACTIONS(2394),
    [anon_sym_do] = ACTIONS(2394),
    [anon_sym_for] = ACTIONS(2394),
    [anon_sym_in] = ACTIONS(2394),
    [anon_sym_return] = ACTIONS(2394),
    [anon_sym_break] = ACTIONS(2394),
    [anon_sym_continue] = ACTIONS(2394),
    [anon_sym_goto] = ACTIONS(2394),
    [anon_sym_DASH_DASH] = ACTIONS(2396),
    [anon_sym_PLUS_PLUS] = ACTIONS(2396),
    [anon_sym_sizeof] = ACTIONS(2394),
    [anon_sym___alignof__] = ACTIONS(2394),
    [anon_sym___alignof] = ACTIONS(2394),
    [anon_sym__alignof] = ACTIONS(2394),
    [anon_sym_alignof] = ACTIONS(2394),
    [anon_sym__Alignof] = ACTIONS(2394),
    [anon_sym_offsetof] = ACTIONS(2394),
    [anon_sym__Generic] = ACTIONS(2394),
    [anon_sym_asm] = ACTIONS(2394),
    [anon_sym___asm__] = ACTIONS(2394),
    [anon_sym___asm] = ACTIONS(2394),
    [sym_number_literal] = ACTIONS(2396),
    [anon_sym_L_SQUOTE] = ACTIONS(2396),
    [anon_sym_u_SQUOTE] = ACTIONS(2396),
    [anon_sym_U_SQUOTE] = ACTIONS(2396),
    [anon_sym_u8_SQUOTE] = ACTIONS(2396),
    [anon_sym_SQUOTE] = ACTIONS(2396),
    [anon_sym_AT] = ACTIONS(2394),
    [anon_sym_DQUOTE] = ACTIONS(2396),
    [anon_sym_L_DQUOTE] = ACTIONS(2396),
    [anon_sym_u_DQUOTE] = ACTIONS(2396),
    [anon_sym_U_DQUOTE] = ACTIONS(2396),
    [anon_sym_u8_DQUOTE] = ACTIONS(2396),
    [sym_true] = ACTIONS(2394),
    [sym_false] = ACTIONS(2394),
    [anon_sym_NULL] = ACTIONS(2394),
    [anon_sym_nullptr] = ACTIONS(2394),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2394),
    [anon_sym___typeof] = ACTIONS(2394),
    [anon_sym_typeof] = ACTIONS(2394),
    [anon_sym_ATimport] = ACTIONS(2396),
    [aux_sym_preproc_undef_token1] = ACTIONS(2394),
    [anon_sym_POUND] = ACTIONS(2394),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2394),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2394),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2394),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_API_AVAILABLE] = ACTIONS(2394),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_API_DEPRECATED] = ACTIONS(2394),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym___deprecated_msg] = ACTIONS(2394),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2394),
    [anon_sym_ATprotocol] = ACTIONS(2396),
    [anon_sym_ATinterface] = ACTIONS(2396),
    [anon_sym_ATimplementation] = ACTIONS(2396),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2396),
    [anon_sym_ATsynthesize] = ACTIONS(2396),
    [anon_sym_ATdynamic] = ACTIONS(2396),
    [anon_sym_ATtry] = ACTIONS(2396),
    [anon_sym___try] = ACTIONS(2394),
    [anon_sym_ATthrow] = ACTIONS(2396),
    [anon_sym_ATselector] = ACTIONS(2396),
    [anon_sym_ATavailable] = ACTIONS(2396),
    [anon_sym___builtin_available] = ACTIONS(2394),
    [anon_sym_va_arg] = ACTIONS(2394),
    [anon_sym_ATencode] = ACTIONS(2396),
    [anon_sym_ATsynchronized] = ACTIONS(2396),
    [anon_sym_BOOL] = ACTIONS(2394),
    [anon_sym_IMP] = ACTIONS(2394),
    [anon_sym_SEL] = ACTIONS(2394),
    [anon_sym_Class] = ACTIONS(2394),
    [anon_sym_id] = ACTIONS(2394),
  },
  [STATE(288)] = {
    [sym_identifier] = ACTIONS(2614),
    [aux_sym_preproc_include_token1] = ACTIONS(2614),
    [aux_sym_preproc_include_token2] = ACTIONS(2614),
    [aux_sym_preproc_def_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token2] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2614),
    [aux_sym_preproc_else_token1] = ACTIONS(2614),
    [aux_sym_preproc_elif_token1] = ACTIONS(2614),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2614),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2614),
    [sym_preproc_directive] = ACTIONS(2614),
    [anon_sym_LPAREN2] = ACTIONS(2616),
    [anon_sym_BANG] = ACTIONS(2616),
    [anon_sym_TILDE] = ACTIONS(2616),
    [anon_sym_DASH] = ACTIONS(2614),
    [anon_sym_PLUS] = ACTIONS(2614),
    [anon_sym_STAR] = ACTIONS(2616),
    [anon_sym_CARET] = ACTIONS(2616),
    [anon_sym_AMP] = ACTIONS(2616),
    [anon_sym_SEMI] = ACTIONS(2616),
    [anon_sym___extension__] = ACTIONS(2614),
    [anon_sym_typedef] = ACTIONS(2614),
    [anon_sym_extern] = ACTIONS(2614),
    [anon_sym___attribute__] = ACTIONS(2614),
    [anon_sym___attribute] = ACTIONS(2614),
    [anon_sym_noreturn] = ACTIONS(2614),
    [anon_sym_LBRACK] = ACTIONS(2616),
    [anon_sym___declspec] = ACTIONS(2614),
    [anon_sym___cdecl] = ACTIONS(2614),
    [anon_sym___clrcall] = ACTIONS(2614),
    [anon_sym___stdcall] = ACTIONS(2614),
    [anon_sym___fastcall] = ACTIONS(2614),
    [anon_sym___thiscall] = ACTIONS(2614),
    [anon_sym___vectorcall] = ACTIONS(2614),
    [anon_sym_LBRACE] = ACTIONS(2616),
    [anon_sym_signed] = ACTIONS(2614),
    [anon_sym_unsigned] = ACTIONS(2614),
    [anon_sym_long] = ACTIONS(2614),
    [anon_sym_short] = ACTIONS(2614),
    [anon_sym_static] = ACTIONS(2614),
    [anon_sym_ATautoreleasepool] = ACTIONS(2616),
    [anon_sym_auto] = ACTIONS(2614),
    [anon_sym_register] = ACTIONS(2614),
    [anon_sym_inline] = ACTIONS(2614),
    [anon_sym___inline] = ACTIONS(2614),
    [anon_sym___inline__] = ACTIONS(2614),
    [anon_sym___forceinline] = ACTIONS(2614),
    [anon_sym_thread_local] = ACTIONS(2614),
    [anon_sym___thread] = ACTIONS(2614),
    [anon_sym_CG_EXTERN] = ACTIONS(2614),
    [anon_sym_CG_INLINE] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2614),
    [anon_sym_IBOutlet] = ACTIONS(2614),
    [anon_sym_IBInspectable] = ACTIONS(2614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2614),
    [anon_sym_NS_INLINE] = ACTIONS(2614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2614),
    [anon_sym_const] = ACTIONS(2614),
    [anon_sym_constexpr] = ACTIONS(2614),
    [anon_sym_volatile] = ACTIONS(2614),
    [anon_sym_restrict] = ACTIONS(2614),
    [anon_sym___restrict__] = ACTIONS(2614),
    [anon_sym__Atomic] = ACTIONS(2614),
    [anon_sym__Noreturn] = ACTIONS(2614),
    [anon_sym__Nonnull] = ACTIONS(2614),
    [anon_sym_nullable] = ACTIONS(2614),
    [anon_sym__Complex] = ACTIONS(2614),
    [anon_sym__Nullable] = ACTIONS(2614),
    [anon_sym__Nullable_result] = ACTIONS(2614),
    [anon_sym__Null_unspecified] = ACTIONS(2614),
    [anon_sym___autoreleasing] = ACTIONS(2614),
    [anon_sym___block] = ACTIONS(2614),
    [anon_sym___bridge] = ACTIONS(2614),
    [anon_sym___bridge_retained] = ACTIONS(2614),
    [anon_sym___bridge_transfer] = ACTIONS(2614),
    [anon_sym___complex] = ACTIONS(2614),
    [anon_sym___const] = ACTIONS(2614),
    [anon_sym___imag] = ACTIONS(2614),
    [anon_sym___kindof] = ACTIONS(2614),
    [anon_sym___nonnull] = ACTIONS(2614),
    [anon_sym___nullable] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2614),
    [anon_sym___real] = ACTIONS(2614),
    [anon_sym___strong] = ACTIONS(2614),
    [anon_sym___unsafe_unretained] = ACTIONS(2614),
    [anon_sym___unused] = ACTIONS(2614),
    [anon_sym___weak] = ACTIONS(2614),
    [anon_sym_alignas] = ACTIONS(2614),
    [anon_sym__Alignas] = ACTIONS(2614),
    [sym_primitive_type] = ACTIONS(2614),
    [anon_sym_enum] = ACTIONS(2614),
    [anon_sym_struct] = ACTIONS(2614),
    [anon_sym_union] = ACTIONS(2614),
    [anon_sym_if] = ACTIONS(2614),
    [anon_sym_switch] = ACTIONS(2614),
    [anon_sym_case] = ACTIONS(2614),
    [anon_sym_default] = ACTIONS(2614),
    [anon_sym_while] = ACTIONS(2614),
    [anon_sym_do] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2614),
    [anon_sym_in] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_break] = ACTIONS(2614),
    [anon_sym_continue] = ACTIONS(2614),
    [anon_sym_goto] = ACTIONS(2614),
    [anon_sym_DASH_DASH] = ACTIONS(2616),
    [anon_sym_PLUS_PLUS] = ACTIONS(2616),
    [anon_sym_sizeof] = ACTIONS(2614),
    [anon_sym___alignof__] = ACTIONS(2614),
    [anon_sym___alignof] = ACTIONS(2614),
    [anon_sym__alignof] = ACTIONS(2614),
    [anon_sym_alignof] = ACTIONS(2614),
    [anon_sym__Alignof] = ACTIONS(2614),
    [anon_sym_offsetof] = ACTIONS(2614),
    [anon_sym__Generic] = ACTIONS(2614),
    [anon_sym_asm] = ACTIONS(2614),
    [anon_sym___asm__] = ACTIONS(2614),
    [anon_sym___asm] = ACTIONS(2614),
    [sym_number_literal] = ACTIONS(2616),
    [anon_sym_L_SQUOTE] = ACTIONS(2616),
    [anon_sym_u_SQUOTE] = ACTIONS(2616),
    [anon_sym_U_SQUOTE] = ACTIONS(2616),
    [anon_sym_u8_SQUOTE] = ACTIONS(2616),
    [anon_sym_SQUOTE] = ACTIONS(2616),
    [anon_sym_AT] = ACTIONS(2614),
    [anon_sym_DQUOTE] = ACTIONS(2616),
    [anon_sym_L_DQUOTE] = ACTIONS(2616),
    [anon_sym_u_DQUOTE] = ACTIONS(2616),
    [anon_sym_U_DQUOTE] = ACTIONS(2616),
    [anon_sym_u8_DQUOTE] = ACTIONS(2616),
    [sym_true] = ACTIONS(2614),
    [sym_false] = ACTIONS(2614),
    [anon_sym_NULL] = ACTIONS(2614),
    [anon_sym_nullptr] = ACTIONS(2614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2614),
    [anon_sym___typeof] = ACTIONS(2614),
    [anon_sym_typeof] = ACTIONS(2614),
    [anon_sym_ATimport] = ACTIONS(2616),
    [aux_sym_preproc_undef_token1] = ACTIONS(2614),
    [anon_sym_POUND] = ACTIONS(2614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_API_AVAILABLE] = ACTIONS(2614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_API_DEPRECATED] = ACTIONS(2614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym___deprecated_msg] = ACTIONS(2614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2614),
    [anon_sym_ATprotocol] = ACTIONS(2616),
    [anon_sym_ATinterface] = ACTIONS(2616),
    [anon_sym_ATimplementation] = ACTIONS(2616),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2616),
    [anon_sym_ATsynthesize] = ACTIONS(2616),
    [anon_sym_ATdynamic] = ACTIONS(2616),
    [anon_sym_ATtry] = ACTIONS(2616),
    [anon_sym___try] = ACTIONS(2614),
    [anon_sym_ATthrow] = ACTIONS(2616),
    [anon_sym_ATselector] = ACTIONS(2616),
    [anon_sym_ATavailable] = ACTIONS(2616),
    [anon_sym___builtin_available] = ACTIONS(2614),
    [anon_sym_va_arg] = ACTIONS(2614),
    [anon_sym_ATencode] = ACTIONS(2616),
    [anon_sym_ATsynchronized] = ACTIONS(2616),
    [anon_sym_BOOL] = ACTIONS(2614),
    [anon_sym_IMP] = ACTIONS(2614),
    [anon_sym_SEL] = ACTIONS(2614),
    [anon_sym_Class] = ACTIONS(2614),
    [anon_sym_id] = ACTIONS(2614),
  },
  [STATE(289)] = {
    [sym_identifier] = ACTIONS(2618),
    [aux_sym_preproc_include_token1] = ACTIONS(2618),
    [aux_sym_preproc_include_token2] = ACTIONS(2618),
    [aux_sym_preproc_def_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token2] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2618),
    [aux_sym_preproc_else_token1] = ACTIONS(2618),
    [aux_sym_preproc_elif_token1] = ACTIONS(2618),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2618),
    [sym_preproc_directive] = ACTIONS(2618),
    [anon_sym_LPAREN2] = ACTIONS(2620),
    [anon_sym_BANG] = ACTIONS(2620),
    [anon_sym_TILDE] = ACTIONS(2620),
    [anon_sym_DASH] = ACTIONS(2618),
    [anon_sym_PLUS] = ACTIONS(2618),
    [anon_sym_STAR] = ACTIONS(2620),
    [anon_sym_CARET] = ACTIONS(2620),
    [anon_sym_AMP] = ACTIONS(2620),
    [anon_sym_SEMI] = ACTIONS(2620),
    [anon_sym___extension__] = ACTIONS(2618),
    [anon_sym_typedef] = ACTIONS(2618),
    [anon_sym_extern] = ACTIONS(2618),
    [anon_sym___attribute__] = ACTIONS(2618),
    [anon_sym___attribute] = ACTIONS(2618),
    [anon_sym_noreturn] = ACTIONS(2618),
    [anon_sym_LBRACK] = ACTIONS(2620),
    [anon_sym___declspec] = ACTIONS(2618),
    [anon_sym___cdecl] = ACTIONS(2618),
    [anon_sym___clrcall] = ACTIONS(2618),
    [anon_sym___stdcall] = ACTIONS(2618),
    [anon_sym___fastcall] = ACTIONS(2618),
    [anon_sym___thiscall] = ACTIONS(2618),
    [anon_sym___vectorcall] = ACTIONS(2618),
    [anon_sym_LBRACE] = ACTIONS(2620),
    [anon_sym_signed] = ACTIONS(2618),
    [anon_sym_unsigned] = ACTIONS(2618),
    [anon_sym_long] = ACTIONS(2618),
    [anon_sym_short] = ACTIONS(2618),
    [anon_sym_static] = ACTIONS(2618),
    [anon_sym_ATautoreleasepool] = ACTIONS(2620),
    [anon_sym_auto] = ACTIONS(2618),
    [anon_sym_register] = ACTIONS(2618),
    [anon_sym_inline] = ACTIONS(2618),
    [anon_sym___inline] = ACTIONS(2618),
    [anon_sym___inline__] = ACTIONS(2618),
    [anon_sym___forceinline] = ACTIONS(2618),
    [anon_sym_thread_local] = ACTIONS(2618),
    [anon_sym___thread] = ACTIONS(2618),
    [anon_sym_CG_EXTERN] = ACTIONS(2618),
    [anon_sym_CG_INLINE] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2618),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2618),
    [anon_sym_IBOutlet] = ACTIONS(2618),
    [anon_sym_IBInspectable] = ACTIONS(2618),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2618),
    [anon_sym_NS_INLINE] = ACTIONS(2618),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2618),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2618),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2618),
    [anon_sym_const] = ACTIONS(2618),
    [anon_sym_constexpr] = ACTIONS(2618),
    [anon_sym_volatile] = ACTIONS(2618),
    [anon_sym_restrict] = ACTIONS(2618),
    [anon_sym___restrict__] = ACTIONS(2618),
    [anon_sym__Atomic] = ACTIONS(2618),
    [anon_sym__Noreturn] = ACTIONS(2618),
    [anon_sym__Nonnull] = ACTIONS(2618),
    [anon_sym_nullable] = ACTIONS(2618),
    [anon_sym__Complex] = ACTIONS(2618),
    [anon_sym__Nullable] = ACTIONS(2618),
    [anon_sym__Nullable_result] = ACTIONS(2618),
    [anon_sym__Null_unspecified] = ACTIONS(2618),
    [anon_sym___autoreleasing] = ACTIONS(2618),
    [anon_sym___block] = ACTIONS(2618),
    [anon_sym___bridge] = ACTIONS(2618),
    [anon_sym___bridge_retained] = ACTIONS(2618),
    [anon_sym___bridge_transfer] = ACTIONS(2618),
    [anon_sym___complex] = ACTIONS(2618),
    [anon_sym___const] = ACTIONS(2618),
    [anon_sym___imag] = ACTIONS(2618),
    [anon_sym___kindof] = ACTIONS(2618),
    [anon_sym___nonnull] = ACTIONS(2618),
    [anon_sym___nullable] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2618),
    [anon_sym___real] = ACTIONS(2618),
    [anon_sym___strong] = ACTIONS(2618),
    [anon_sym___unsafe_unretained] = ACTIONS(2618),
    [anon_sym___unused] = ACTIONS(2618),
    [anon_sym___weak] = ACTIONS(2618),
    [anon_sym_alignas] = ACTIONS(2618),
    [anon_sym__Alignas] = ACTIONS(2618),
    [sym_primitive_type] = ACTIONS(2618),
    [anon_sym_enum] = ACTIONS(2618),
    [anon_sym_struct] = ACTIONS(2618),
    [anon_sym_union] = ACTIONS(2618),
    [anon_sym_if] = ACTIONS(2618),
    [anon_sym_switch] = ACTIONS(2618),
    [anon_sym_case] = ACTIONS(2618),
    [anon_sym_default] = ACTIONS(2618),
    [anon_sym_while] = ACTIONS(2618),
    [anon_sym_do] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2618),
    [anon_sym_in] = ACTIONS(2618),
    [anon_sym_return] = ACTIONS(2618),
    [anon_sym_break] = ACTIONS(2618),
    [anon_sym_continue] = ACTIONS(2618),
    [anon_sym_goto] = ACTIONS(2618),
    [anon_sym_DASH_DASH] = ACTIONS(2620),
    [anon_sym_PLUS_PLUS] = ACTIONS(2620),
    [anon_sym_sizeof] = ACTIONS(2618),
    [anon_sym___alignof__] = ACTIONS(2618),
    [anon_sym___alignof] = ACTIONS(2618),
    [anon_sym__alignof] = ACTIONS(2618),
    [anon_sym_alignof] = ACTIONS(2618),
    [anon_sym__Alignof] = ACTIONS(2618),
    [anon_sym_offsetof] = ACTIONS(2618),
    [anon_sym__Generic] = ACTIONS(2618),
    [anon_sym_asm] = ACTIONS(2618),
    [anon_sym___asm__] = ACTIONS(2618),
    [anon_sym___asm] = ACTIONS(2618),
    [sym_number_literal] = ACTIONS(2620),
    [anon_sym_L_SQUOTE] = ACTIONS(2620),
    [anon_sym_u_SQUOTE] = ACTIONS(2620),
    [anon_sym_U_SQUOTE] = ACTIONS(2620),
    [anon_sym_u8_SQUOTE] = ACTIONS(2620),
    [anon_sym_SQUOTE] = ACTIONS(2620),
    [anon_sym_AT] = ACTIONS(2618),
    [anon_sym_DQUOTE] = ACTIONS(2620),
    [anon_sym_L_DQUOTE] = ACTIONS(2620),
    [anon_sym_u_DQUOTE] = ACTIONS(2620),
    [anon_sym_U_DQUOTE] = ACTIONS(2620),
    [anon_sym_u8_DQUOTE] = ACTIONS(2620),
    [sym_true] = ACTIONS(2618),
    [sym_false] = ACTIONS(2618),
    [anon_sym_NULL] = ACTIONS(2618),
    [anon_sym_nullptr] = ACTIONS(2618),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2618),
    [anon_sym___typeof] = ACTIONS(2618),
    [anon_sym_typeof] = ACTIONS(2618),
    [anon_sym_ATimport] = ACTIONS(2620),
    [aux_sym_preproc_undef_token1] = ACTIONS(2618),
    [anon_sym_POUND] = ACTIONS(2618),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2618),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2618),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2618),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_API_AVAILABLE] = ACTIONS(2618),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_API_DEPRECATED] = ACTIONS(2618),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym___deprecated_msg] = ACTIONS(2618),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2618),
    [anon_sym_ATprotocol] = ACTIONS(2620),
    [anon_sym_ATinterface] = ACTIONS(2620),
    [anon_sym_ATimplementation] = ACTIONS(2620),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2620),
    [anon_sym_ATsynthesize] = ACTIONS(2620),
    [anon_sym_ATdynamic] = ACTIONS(2620),
    [anon_sym_ATtry] = ACTIONS(2620),
    [anon_sym___try] = ACTIONS(2618),
    [anon_sym_ATthrow] = ACTIONS(2620),
    [anon_sym_ATselector] = ACTIONS(2620),
    [anon_sym_ATavailable] = ACTIONS(2620),
    [anon_sym___builtin_available] = ACTIONS(2618),
    [anon_sym_va_arg] = ACTIONS(2618),
    [anon_sym_ATencode] = ACTIONS(2620),
    [anon_sym_ATsynchronized] = ACTIONS(2620),
    [anon_sym_BOOL] = ACTIONS(2618),
    [anon_sym_IMP] = ACTIONS(2618),
    [anon_sym_SEL] = ACTIONS(2618),
    [anon_sym_Class] = ACTIONS(2618),
    [anon_sym_id] = ACTIONS(2618),
  },
  [STATE(290)] = {
    [sym_identifier] = ACTIONS(2622),
    [aux_sym_preproc_include_token1] = ACTIONS(2622),
    [aux_sym_preproc_include_token2] = ACTIONS(2622),
    [aux_sym_preproc_def_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token2] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2622),
    [aux_sym_preproc_else_token1] = ACTIONS(2622),
    [aux_sym_preproc_elif_token1] = ACTIONS(2622),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2622),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2622),
    [sym_preproc_directive] = ACTIONS(2622),
    [anon_sym_LPAREN2] = ACTIONS(2624),
    [anon_sym_BANG] = ACTIONS(2624),
    [anon_sym_TILDE] = ACTIONS(2624),
    [anon_sym_DASH] = ACTIONS(2622),
    [anon_sym_PLUS] = ACTIONS(2622),
    [anon_sym_STAR] = ACTIONS(2624),
    [anon_sym_CARET] = ACTIONS(2624),
    [anon_sym_AMP] = ACTIONS(2624),
    [anon_sym_SEMI] = ACTIONS(2624),
    [anon_sym___extension__] = ACTIONS(2622),
    [anon_sym_typedef] = ACTIONS(2622),
    [anon_sym_extern] = ACTIONS(2622),
    [anon_sym___attribute__] = ACTIONS(2622),
    [anon_sym___attribute] = ACTIONS(2622),
    [anon_sym_noreturn] = ACTIONS(2622),
    [anon_sym_LBRACK] = ACTIONS(2624),
    [anon_sym___declspec] = ACTIONS(2622),
    [anon_sym___cdecl] = ACTIONS(2622),
    [anon_sym___clrcall] = ACTIONS(2622),
    [anon_sym___stdcall] = ACTIONS(2622),
    [anon_sym___fastcall] = ACTIONS(2622),
    [anon_sym___thiscall] = ACTIONS(2622),
    [anon_sym___vectorcall] = ACTIONS(2622),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_signed] = ACTIONS(2622),
    [anon_sym_unsigned] = ACTIONS(2622),
    [anon_sym_long] = ACTIONS(2622),
    [anon_sym_short] = ACTIONS(2622),
    [anon_sym_static] = ACTIONS(2622),
    [anon_sym_ATautoreleasepool] = ACTIONS(2624),
    [anon_sym_auto] = ACTIONS(2622),
    [anon_sym_register] = ACTIONS(2622),
    [anon_sym_inline] = ACTIONS(2622),
    [anon_sym___inline] = ACTIONS(2622),
    [anon_sym___inline__] = ACTIONS(2622),
    [anon_sym___forceinline] = ACTIONS(2622),
    [anon_sym_thread_local] = ACTIONS(2622),
    [anon_sym___thread] = ACTIONS(2622),
    [anon_sym_CG_EXTERN] = ACTIONS(2622),
    [anon_sym_CG_INLINE] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2622),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2622),
    [anon_sym_IBOutlet] = ACTIONS(2622),
    [anon_sym_IBInspectable] = ACTIONS(2622),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2622),
    [anon_sym_NS_INLINE] = ACTIONS(2622),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2622),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2622),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2622),
    [anon_sym_const] = ACTIONS(2622),
    [anon_sym_constexpr] = ACTIONS(2622),
    [anon_sym_volatile] = ACTIONS(2622),
    [anon_sym_restrict] = ACTIONS(2622),
    [anon_sym___restrict__] = ACTIONS(2622),
    [anon_sym__Atomic] = ACTIONS(2622),
    [anon_sym__Noreturn] = ACTIONS(2622),
    [anon_sym__Nonnull] = ACTIONS(2622),
    [anon_sym_nullable] = ACTIONS(2622),
    [anon_sym__Complex] = ACTIONS(2622),
    [anon_sym__Nullable] = ACTIONS(2622),
    [anon_sym__Nullable_result] = ACTIONS(2622),
    [anon_sym__Null_unspecified] = ACTIONS(2622),
    [anon_sym___autoreleasing] = ACTIONS(2622),
    [anon_sym___block] = ACTIONS(2622),
    [anon_sym___bridge] = ACTIONS(2622),
    [anon_sym___bridge_retained] = ACTIONS(2622),
    [anon_sym___bridge_transfer] = ACTIONS(2622),
    [anon_sym___complex] = ACTIONS(2622),
    [anon_sym___const] = ACTIONS(2622),
    [anon_sym___imag] = ACTIONS(2622),
    [anon_sym___kindof] = ACTIONS(2622),
    [anon_sym___nonnull] = ACTIONS(2622),
    [anon_sym___nullable] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2622),
    [anon_sym___real] = ACTIONS(2622),
    [anon_sym___strong] = ACTIONS(2622),
    [anon_sym___unsafe_unretained] = ACTIONS(2622),
    [anon_sym___unused] = ACTIONS(2622),
    [anon_sym___weak] = ACTIONS(2622),
    [anon_sym_alignas] = ACTIONS(2622),
    [anon_sym__Alignas] = ACTIONS(2622),
    [sym_primitive_type] = ACTIONS(2622),
    [anon_sym_enum] = ACTIONS(2622),
    [anon_sym_struct] = ACTIONS(2622),
    [anon_sym_union] = ACTIONS(2622),
    [anon_sym_if] = ACTIONS(2622),
    [anon_sym_switch] = ACTIONS(2622),
    [anon_sym_case] = ACTIONS(2622),
    [anon_sym_default] = ACTIONS(2622),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_do] = ACTIONS(2622),
    [anon_sym_for] = ACTIONS(2622),
    [anon_sym_in] = ACTIONS(2622),
    [anon_sym_return] = ACTIONS(2622),
    [anon_sym_break] = ACTIONS(2622),
    [anon_sym_continue] = ACTIONS(2622),
    [anon_sym_goto] = ACTIONS(2622),
    [anon_sym_DASH_DASH] = ACTIONS(2624),
    [anon_sym_PLUS_PLUS] = ACTIONS(2624),
    [anon_sym_sizeof] = ACTIONS(2622),
    [anon_sym___alignof__] = ACTIONS(2622),
    [anon_sym___alignof] = ACTIONS(2622),
    [anon_sym__alignof] = ACTIONS(2622),
    [anon_sym_alignof] = ACTIONS(2622),
    [anon_sym__Alignof] = ACTIONS(2622),
    [anon_sym_offsetof] = ACTIONS(2622),
    [anon_sym__Generic] = ACTIONS(2622),
    [anon_sym_asm] = ACTIONS(2622),
    [anon_sym___asm__] = ACTIONS(2622),
    [anon_sym___asm] = ACTIONS(2622),
    [sym_number_literal] = ACTIONS(2624),
    [anon_sym_L_SQUOTE] = ACTIONS(2624),
    [anon_sym_u_SQUOTE] = ACTIONS(2624),
    [anon_sym_U_SQUOTE] = ACTIONS(2624),
    [anon_sym_u8_SQUOTE] = ACTIONS(2624),
    [anon_sym_SQUOTE] = ACTIONS(2624),
    [anon_sym_AT] = ACTIONS(2622),
    [anon_sym_DQUOTE] = ACTIONS(2624),
    [anon_sym_L_DQUOTE] = ACTIONS(2624),
    [anon_sym_u_DQUOTE] = ACTIONS(2624),
    [anon_sym_U_DQUOTE] = ACTIONS(2624),
    [anon_sym_u8_DQUOTE] = ACTIONS(2624),
    [sym_true] = ACTIONS(2622),
    [sym_false] = ACTIONS(2622),
    [anon_sym_NULL] = ACTIONS(2622),
    [anon_sym_nullptr] = ACTIONS(2622),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2622),
    [anon_sym___typeof] = ACTIONS(2622),
    [anon_sym_typeof] = ACTIONS(2622),
    [anon_sym_ATimport] = ACTIONS(2624),
    [aux_sym_preproc_undef_token1] = ACTIONS(2622),
    [anon_sym_POUND] = ACTIONS(2622),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2622),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2622),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2622),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_API_AVAILABLE] = ACTIONS(2622),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_API_DEPRECATED] = ACTIONS(2622),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym___deprecated_msg] = ACTIONS(2622),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2622),
    [anon_sym_ATprotocol] = ACTIONS(2624),
    [anon_sym_ATinterface] = ACTIONS(2624),
    [anon_sym_ATimplementation] = ACTIONS(2624),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2624),
    [anon_sym_ATsynthesize] = ACTIONS(2624),
    [anon_sym_ATdynamic] = ACTIONS(2624),
    [anon_sym_ATtry] = ACTIONS(2624),
    [anon_sym___try] = ACTIONS(2622),
    [anon_sym_ATthrow] = ACTIONS(2624),
    [anon_sym_ATselector] = ACTIONS(2624),
    [anon_sym_ATavailable] = ACTIONS(2624),
    [anon_sym___builtin_available] = ACTIONS(2622),
    [anon_sym_va_arg] = ACTIONS(2622),
    [anon_sym_ATencode] = ACTIONS(2624),
    [anon_sym_ATsynchronized] = ACTIONS(2624),
    [anon_sym_BOOL] = ACTIONS(2622),
    [anon_sym_IMP] = ACTIONS(2622),
    [anon_sym_SEL] = ACTIONS(2622),
    [anon_sym_Class] = ACTIONS(2622),
    [anon_sym_id] = ACTIONS(2622),
  },
  [STATE(291)] = {
    [sym_identifier] = ACTIONS(2626),
    [aux_sym_preproc_include_token1] = ACTIONS(2626),
    [aux_sym_preproc_include_token2] = ACTIONS(2626),
    [aux_sym_preproc_def_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token2] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2626),
    [aux_sym_preproc_else_token1] = ACTIONS(2626),
    [aux_sym_preproc_elif_token1] = ACTIONS(2626),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2626),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2626),
    [sym_preproc_directive] = ACTIONS(2626),
    [anon_sym_LPAREN2] = ACTIONS(2628),
    [anon_sym_BANG] = ACTIONS(2628),
    [anon_sym_TILDE] = ACTIONS(2628),
    [anon_sym_DASH] = ACTIONS(2626),
    [anon_sym_PLUS] = ACTIONS(2626),
    [anon_sym_STAR] = ACTIONS(2628),
    [anon_sym_CARET] = ACTIONS(2628),
    [anon_sym_AMP] = ACTIONS(2628),
    [anon_sym_SEMI] = ACTIONS(2628),
    [anon_sym___extension__] = ACTIONS(2626),
    [anon_sym_typedef] = ACTIONS(2626),
    [anon_sym_extern] = ACTIONS(2626),
    [anon_sym___attribute__] = ACTIONS(2626),
    [anon_sym___attribute] = ACTIONS(2626),
    [anon_sym_noreturn] = ACTIONS(2626),
    [anon_sym_LBRACK] = ACTIONS(2628),
    [anon_sym___declspec] = ACTIONS(2626),
    [anon_sym___cdecl] = ACTIONS(2626),
    [anon_sym___clrcall] = ACTIONS(2626),
    [anon_sym___stdcall] = ACTIONS(2626),
    [anon_sym___fastcall] = ACTIONS(2626),
    [anon_sym___thiscall] = ACTIONS(2626),
    [anon_sym___vectorcall] = ACTIONS(2626),
    [anon_sym_LBRACE] = ACTIONS(2628),
    [anon_sym_signed] = ACTIONS(2626),
    [anon_sym_unsigned] = ACTIONS(2626),
    [anon_sym_long] = ACTIONS(2626),
    [anon_sym_short] = ACTIONS(2626),
    [anon_sym_static] = ACTIONS(2626),
    [anon_sym_ATautoreleasepool] = ACTIONS(2628),
    [anon_sym_auto] = ACTIONS(2626),
    [anon_sym_register] = ACTIONS(2626),
    [anon_sym_inline] = ACTIONS(2626),
    [anon_sym___inline] = ACTIONS(2626),
    [anon_sym___inline__] = ACTIONS(2626),
    [anon_sym___forceinline] = ACTIONS(2626),
    [anon_sym_thread_local] = ACTIONS(2626),
    [anon_sym___thread] = ACTIONS(2626),
    [anon_sym_CG_EXTERN] = ACTIONS(2626),
    [anon_sym_CG_INLINE] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2626),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2626),
    [anon_sym_IBOutlet] = ACTIONS(2626),
    [anon_sym_IBInspectable] = ACTIONS(2626),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2626),
    [anon_sym_NS_INLINE] = ACTIONS(2626),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2626),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2626),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2626),
    [anon_sym_const] = ACTIONS(2626),
    [anon_sym_constexpr] = ACTIONS(2626),
    [anon_sym_volatile] = ACTIONS(2626),
    [anon_sym_restrict] = ACTIONS(2626),
    [anon_sym___restrict__] = ACTIONS(2626),
    [anon_sym__Atomic] = ACTIONS(2626),
    [anon_sym__Noreturn] = ACTIONS(2626),
    [anon_sym__Nonnull] = ACTIONS(2626),
    [anon_sym_nullable] = ACTIONS(2626),
    [anon_sym__Complex] = ACTIONS(2626),
    [anon_sym__Nullable] = ACTIONS(2626),
    [anon_sym__Nullable_result] = ACTIONS(2626),
    [anon_sym__Null_unspecified] = ACTIONS(2626),
    [anon_sym___autoreleasing] = ACTIONS(2626),
    [anon_sym___block] = ACTIONS(2626),
    [anon_sym___bridge] = ACTIONS(2626),
    [anon_sym___bridge_retained] = ACTIONS(2626),
    [anon_sym___bridge_transfer] = ACTIONS(2626),
    [anon_sym___complex] = ACTIONS(2626),
    [anon_sym___const] = ACTIONS(2626),
    [anon_sym___imag] = ACTIONS(2626),
    [anon_sym___kindof] = ACTIONS(2626),
    [anon_sym___nonnull] = ACTIONS(2626),
    [anon_sym___nullable] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2626),
    [anon_sym___real] = ACTIONS(2626),
    [anon_sym___strong] = ACTIONS(2626),
    [anon_sym___unsafe_unretained] = ACTIONS(2626),
    [anon_sym___unused] = ACTIONS(2626),
    [anon_sym___weak] = ACTIONS(2626),
    [anon_sym_alignas] = ACTIONS(2626),
    [anon_sym__Alignas] = ACTIONS(2626),
    [sym_primitive_type] = ACTIONS(2626),
    [anon_sym_enum] = ACTIONS(2626),
    [anon_sym_struct] = ACTIONS(2626),
    [anon_sym_union] = ACTIONS(2626),
    [anon_sym_if] = ACTIONS(2626),
    [anon_sym_switch] = ACTIONS(2626),
    [anon_sym_case] = ACTIONS(2626),
    [anon_sym_default] = ACTIONS(2626),
    [anon_sym_while] = ACTIONS(2626),
    [anon_sym_do] = ACTIONS(2626),
    [anon_sym_for] = ACTIONS(2626),
    [anon_sym_in] = ACTIONS(2626),
    [anon_sym_return] = ACTIONS(2626),
    [anon_sym_break] = ACTIONS(2626),
    [anon_sym_continue] = ACTIONS(2626),
    [anon_sym_goto] = ACTIONS(2626),
    [anon_sym_DASH_DASH] = ACTIONS(2628),
    [anon_sym_PLUS_PLUS] = ACTIONS(2628),
    [anon_sym_sizeof] = ACTIONS(2626),
    [anon_sym___alignof__] = ACTIONS(2626),
    [anon_sym___alignof] = ACTIONS(2626),
    [anon_sym__alignof] = ACTIONS(2626),
    [anon_sym_alignof] = ACTIONS(2626),
    [anon_sym__Alignof] = ACTIONS(2626),
    [anon_sym_offsetof] = ACTIONS(2626),
    [anon_sym__Generic] = ACTIONS(2626),
    [anon_sym_asm] = ACTIONS(2626),
    [anon_sym___asm__] = ACTIONS(2626),
    [anon_sym___asm] = ACTIONS(2626),
    [sym_number_literal] = ACTIONS(2628),
    [anon_sym_L_SQUOTE] = ACTIONS(2628),
    [anon_sym_u_SQUOTE] = ACTIONS(2628),
    [anon_sym_U_SQUOTE] = ACTIONS(2628),
    [anon_sym_u8_SQUOTE] = ACTIONS(2628),
    [anon_sym_SQUOTE] = ACTIONS(2628),
    [anon_sym_AT] = ACTIONS(2626),
    [anon_sym_DQUOTE] = ACTIONS(2628),
    [anon_sym_L_DQUOTE] = ACTIONS(2628),
    [anon_sym_u_DQUOTE] = ACTIONS(2628),
    [anon_sym_U_DQUOTE] = ACTIONS(2628),
    [anon_sym_u8_DQUOTE] = ACTIONS(2628),
    [sym_true] = ACTIONS(2626),
    [sym_false] = ACTIONS(2626),
    [anon_sym_NULL] = ACTIONS(2626),
    [anon_sym_nullptr] = ACTIONS(2626),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2626),
    [anon_sym___typeof] = ACTIONS(2626),
    [anon_sym_typeof] = ACTIONS(2626),
    [anon_sym_ATimport] = ACTIONS(2628),
    [aux_sym_preproc_undef_token1] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2626),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2626),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2626),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2626),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_API_AVAILABLE] = ACTIONS(2626),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_API_DEPRECATED] = ACTIONS(2626),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym___deprecated_msg] = ACTIONS(2626),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2626),
    [anon_sym_ATprotocol] = ACTIONS(2628),
    [anon_sym_ATinterface] = ACTIONS(2628),
    [anon_sym_ATimplementation] = ACTIONS(2628),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2628),
    [anon_sym_ATsynthesize] = ACTIONS(2628),
    [anon_sym_ATdynamic] = ACTIONS(2628),
    [anon_sym_ATtry] = ACTIONS(2628),
    [anon_sym___try] = ACTIONS(2626),
    [anon_sym_ATthrow] = ACTIONS(2628),
    [anon_sym_ATselector] = ACTIONS(2628),
    [anon_sym_ATavailable] = ACTIONS(2628),
    [anon_sym___builtin_available] = ACTIONS(2626),
    [anon_sym_va_arg] = ACTIONS(2626),
    [anon_sym_ATencode] = ACTIONS(2628),
    [anon_sym_ATsynchronized] = ACTIONS(2628),
    [anon_sym_BOOL] = ACTIONS(2626),
    [anon_sym_IMP] = ACTIONS(2626),
    [anon_sym_SEL] = ACTIONS(2626),
    [anon_sym_Class] = ACTIONS(2626),
    [anon_sym_id] = ACTIONS(2626),
  },
  [STATE(292)] = {
    [sym_identifier] = ACTIONS(2630),
    [aux_sym_preproc_include_token1] = ACTIONS(2630),
    [aux_sym_preproc_include_token2] = ACTIONS(2630),
    [aux_sym_preproc_def_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token2] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2630),
    [aux_sym_preproc_else_token1] = ACTIONS(2630),
    [aux_sym_preproc_elif_token1] = ACTIONS(2630),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2630),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2630),
    [sym_preproc_directive] = ACTIONS(2630),
    [anon_sym_LPAREN2] = ACTIONS(2632),
    [anon_sym_BANG] = ACTIONS(2632),
    [anon_sym_TILDE] = ACTIONS(2632),
    [anon_sym_DASH] = ACTIONS(2630),
    [anon_sym_PLUS] = ACTIONS(2630),
    [anon_sym_STAR] = ACTIONS(2632),
    [anon_sym_CARET] = ACTIONS(2632),
    [anon_sym_AMP] = ACTIONS(2632),
    [anon_sym_SEMI] = ACTIONS(2632),
    [anon_sym___extension__] = ACTIONS(2630),
    [anon_sym_typedef] = ACTIONS(2630),
    [anon_sym_extern] = ACTIONS(2630),
    [anon_sym___attribute__] = ACTIONS(2630),
    [anon_sym___attribute] = ACTIONS(2630),
    [anon_sym_noreturn] = ACTIONS(2630),
    [anon_sym_LBRACK] = ACTIONS(2632),
    [anon_sym___declspec] = ACTIONS(2630),
    [anon_sym___cdecl] = ACTIONS(2630),
    [anon_sym___clrcall] = ACTIONS(2630),
    [anon_sym___stdcall] = ACTIONS(2630),
    [anon_sym___fastcall] = ACTIONS(2630),
    [anon_sym___thiscall] = ACTIONS(2630),
    [anon_sym___vectorcall] = ACTIONS(2630),
    [anon_sym_LBRACE] = ACTIONS(2632),
    [anon_sym_signed] = ACTIONS(2630),
    [anon_sym_unsigned] = ACTIONS(2630),
    [anon_sym_long] = ACTIONS(2630),
    [anon_sym_short] = ACTIONS(2630),
    [anon_sym_static] = ACTIONS(2630),
    [anon_sym_ATautoreleasepool] = ACTIONS(2632),
    [anon_sym_auto] = ACTIONS(2630),
    [anon_sym_register] = ACTIONS(2630),
    [anon_sym_inline] = ACTIONS(2630),
    [anon_sym___inline] = ACTIONS(2630),
    [anon_sym___inline__] = ACTIONS(2630),
    [anon_sym___forceinline] = ACTIONS(2630),
    [anon_sym_thread_local] = ACTIONS(2630),
    [anon_sym___thread] = ACTIONS(2630),
    [anon_sym_CG_EXTERN] = ACTIONS(2630),
    [anon_sym_CG_INLINE] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2630),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2630),
    [anon_sym_IBOutlet] = ACTIONS(2630),
    [anon_sym_IBInspectable] = ACTIONS(2630),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2630),
    [anon_sym_NS_INLINE] = ACTIONS(2630),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2630),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2630),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2630),
    [anon_sym_const] = ACTIONS(2630),
    [anon_sym_constexpr] = ACTIONS(2630),
    [anon_sym_volatile] = ACTIONS(2630),
    [anon_sym_restrict] = ACTIONS(2630),
    [anon_sym___restrict__] = ACTIONS(2630),
    [anon_sym__Atomic] = ACTIONS(2630),
    [anon_sym__Noreturn] = ACTIONS(2630),
    [anon_sym__Nonnull] = ACTIONS(2630),
    [anon_sym_nullable] = ACTIONS(2630),
    [anon_sym__Complex] = ACTIONS(2630),
    [anon_sym__Nullable] = ACTIONS(2630),
    [anon_sym__Nullable_result] = ACTIONS(2630),
    [anon_sym__Null_unspecified] = ACTIONS(2630),
    [anon_sym___autoreleasing] = ACTIONS(2630),
    [anon_sym___block] = ACTIONS(2630),
    [anon_sym___bridge] = ACTIONS(2630),
    [anon_sym___bridge_retained] = ACTIONS(2630),
    [anon_sym___bridge_transfer] = ACTIONS(2630),
    [anon_sym___complex] = ACTIONS(2630),
    [anon_sym___const] = ACTIONS(2630),
    [anon_sym___imag] = ACTIONS(2630),
    [anon_sym___kindof] = ACTIONS(2630),
    [anon_sym___nonnull] = ACTIONS(2630),
    [anon_sym___nullable] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2630),
    [anon_sym___real] = ACTIONS(2630),
    [anon_sym___strong] = ACTIONS(2630),
    [anon_sym___unsafe_unretained] = ACTIONS(2630),
    [anon_sym___unused] = ACTIONS(2630),
    [anon_sym___weak] = ACTIONS(2630),
    [anon_sym_alignas] = ACTIONS(2630),
    [anon_sym__Alignas] = ACTIONS(2630),
    [sym_primitive_type] = ACTIONS(2630),
    [anon_sym_enum] = ACTIONS(2630),
    [anon_sym_struct] = ACTIONS(2630),
    [anon_sym_union] = ACTIONS(2630),
    [anon_sym_if] = ACTIONS(2630),
    [anon_sym_switch] = ACTIONS(2630),
    [anon_sym_case] = ACTIONS(2630),
    [anon_sym_default] = ACTIONS(2630),
    [anon_sym_while] = ACTIONS(2630),
    [anon_sym_do] = ACTIONS(2630),
    [anon_sym_for] = ACTIONS(2630),
    [anon_sym_in] = ACTIONS(2630),
    [anon_sym_return] = ACTIONS(2630),
    [anon_sym_break] = ACTIONS(2630),
    [anon_sym_continue] = ACTIONS(2630),
    [anon_sym_goto] = ACTIONS(2630),
    [anon_sym_DASH_DASH] = ACTIONS(2632),
    [anon_sym_PLUS_PLUS] = ACTIONS(2632),
    [anon_sym_sizeof] = ACTIONS(2630),
    [anon_sym___alignof__] = ACTIONS(2630),
    [anon_sym___alignof] = ACTIONS(2630),
    [anon_sym__alignof] = ACTIONS(2630),
    [anon_sym_alignof] = ACTIONS(2630),
    [anon_sym__Alignof] = ACTIONS(2630),
    [anon_sym_offsetof] = ACTIONS(2630),
    [anon_sym__Generic] = ACTIONS(2630),
    [anon_sym_asm] = ACTIONS(2630),
    [anon_sym___asm__] = ACTIONS(2630),
    [anon_sym___asm] = ACTIONS(2630),
    [sym_number_literal] = ACTIONS(2632),
    [anon_sym_L_SQUOTE] = ACTIONS(2632),
    [anon_sym_u_SQUOTE] = ACTIONS(2632),
    [anon_sym_U_SQUOTE] = ACTIONS(2632),
    [anon_sym_u8_SQUOTE] = ACTIONS(2632),
    [anon_sym_SQUOTE] = ACTIONS(2632),
    [anon_sym_AT] = ACTIONS(2630),
    [anon_sym_DQUOTE] = ACTIONS(2632),
    [anon_sym_L_DQUOTE] = ACTIONS(2632),
    [anon_sym_u_DQUOTE] = ACTIONS(2632),
    [anon_sym_U_DQUOTE] = ACTIONS(2632),
    [anon_sym_u8_DQUOTE] = ACTIONS(2632),
    [sym_true] = ACTIONS(2630),
    [sym_false] = ACTIONS(2630),
    [anon_sym_NULL] = ACTIONS(2630),
    [anon_sym_nullptr] = ACTIONS(2630),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2630),
    [anon_sym___typeof] = ACTIONS(2630),
    [anon_sym_typeof] = ACTIONS(2630),
    [anon_sym_ATimport] = ACTIONS(2632),
    [aux_sym_preproc_undef_token1] = ACTIONS(2630),
    [anon_sym_POUND] = ACTIONS(2630),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2630),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2630),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2630),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_API_AVAILABLE] = ACTIONS(2630),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_API_DEPRECATED] = ACTIONS(2630),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym___deprecated_msg] = ACTIONS(2630),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2630),
    [anon_sym_ATprotocol] = ACTIONS(2632),
    [anon_sym_ATinterface] = ACTIONS(2632),
    [anon_sym_ATimplementation] = ACTIONS(2632),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2632),
    [anon_sym_ATsynthesize] = ACTIONS(2632),
    [anon_sym_ATdynamic] = ACTIONS(2632),
    [anon_sym_ATtry] = ACTIONS(2632),
    [anon_sym___try] = ACTIONS(2630),
    [anon_sym_ATthrow] = ACTIONS(2632),
    [anon_sym_ATselector] = ACTIONS(2632),
    [anon_sym_ATavailable] = ACTIONS(2632),
    [anon_sym___builtin_available] = ACTIONS(2630),
    [anon_sym_va_arg] = ACTIONS(2630),
    [anon_sym_ATencode] = ACTIONS(2632),
    [anon_sym_ATsynchronized] = ACTIONS(2632),
    [anon_sym_BOOL] = ACTIONS(2630),
    [anon_sym_IMP] = ACTIONS(2630),
    [anon_sym_SEL] = ACTIONS(2630),
    [anon_sym_Class] = ACTIONS(2630),
    [anon_sym_id] = ACTIONS(2630),
  },
  [STATE(293)] = {
    [sym_identifier] = ACTIONS(2634),
    [aux_sym_preproc_include_token1] = ACTIONS(2634),
    [aux_sym_preproc_include_token2] = ACTIONS(2634),
    [aux_sym_preproc_def_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token2] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2634),
    [aux_sym_preproc_else_token1] = ACTIONS(2634),
    [aux_sym_preproc_elif_token1] = ACTIONS(2634),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2634),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2634),
    [sym_preproc_directive] = ACTIONS(2634),
    [anon_sym_LPAREN2] = ACTIONS(2636),
    [anon_sym_BANG] = ACTIONS(2636),
    [anon_sym_TILDE] = ACTIONS(2636),
    [anon_sym_DASH] = ACTIONS(2634),
    [anon_sym_PLUS] = ACTIONS(2634),
    [anon_sym_STAR] = ACTIONS(2636),
    [anon_sym_CARET] = ACTIONS(2636),
    [anon_sym_AMP] = ACTIONS(2636),
    [anon_sym_SEMI] = ACTIONS(2636),
    [anon_sym___extension__] = ACTIONS(2634),
    [anon_sym_typedef] = ACTIONS(2634),
    [anon_sym_extern] = ACTIONS(2634),
    [anon_sym___attribute__] = ACTIONS(2634),
    [anon_sym___attribute] = ACTIONS(2634),
    [anon_sym_noreturn] = ACTIONS(2634),
    [anon_sym_LBRACK] = ACTIONS(2636),
    [anon_sym___declspec] = ACTIONS(2634),
    [anon_sym___cdecl] = ACTIONS(2634),
    [anon_sym___clrcall] = ACTIONS(2634),
    [anon_sym___stdcall] = ACTIONS(2634),
    [anon_sym___fastcall] = ACTIONS(2634),
    [anon_sym___thiscall] = ACTIONS(2634),
    [anon_sym___vectorcall] = ACTIONS(2634),
    [anon_sym_LBRACE] = ACTIONS(2636),
    [anon_sym_signed] = ACTIONS(2634),
    [anon_sym_unsigned] = ACTIONS(2634),
    [anon_sym_long] = ACTIONS(2634),
    [anon_sym_short] = ACTIONS(2634),
    [anon_sym_static] = ACTIONS(2634),
    [anon_sym_ATautoreleasepool] = ACTIONS(2636),
    [anon_sym_auto] = ACTIONS(2634),
    [anon_sym_register] = ACTIONS(2634),
    [anon_sym_inline] = ACTIONS(2634),
    [anon_sym___inline] = ACTIONS(2634),
    [anon_sym___inline__] = ACTIONS(2634),
    [anon_sym___forceinline] = ACTIONS(2634),
    [anon_sym_thread_local] = ACTIONS(2634),
    [anon_sym___thread] = ACTIONS(2634),
    [anon_sym_CG_EXTERN] = ACTIONS(2634),
    [anon_sym_CG_INLINE] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2634),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2634),
    [anon_sym_IBOutlet] = ACTIONS(2634),
    [anon_sym_IBInspectable] = ACTIONS(2634),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2634),
    [anon_sym_NS_INLINE] = ACTIONS(2634),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2634),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2634),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2634),
    [anon_sym_const] = ACTIONS(2634),
    [anon_sym_constexpr] = ACTIONS(2634),
    [anon_sym_volatile] = ACTIONS(2634),
    [anon_sym_restrict] = ACTIONS(2634),
    [anon_sym___restrict__] = ACTIONS(2634),
    [anon_sym__Atomic] = ACTIONS(2634),
    [anon_sym__Noreturn] = ACTIONS(2634),
    [anon_sym__Nonnull] = ACTIONS(2634),
    [anon_sym_nullable] = ACTIONS(2634),
    [anon_sym__Complex] = ACTIONS(2634),
    [anon_sym__Nullable] = ACTIONS(2634),
    [anon_sym__Nullable_result] = ACTIONS(2634),
    [anon_sym__Null_unspecified] = ACTIONS(2634),
    [anon_sym___autoreleasing] = ACTIONS(2634),
    [anon_sym___block] = ACTIONS(2634),
    [anon_sym___bridge] = ACTIONS(2634),
    [anon_sym___bridge_retained] = ACTIONS(2634),
    [anon_sym___bridge_transfer] = ACTIONS(2634),
    [anon_sym___complex] = ACTIONS(2634),
    [anon_sym___const] = ACTIONS(2634),
    [anon_sym___imag] = ACTIONS(2634),
    [anon_sym___kindof] = ACTIONS(2634),
    [anon_sym___nonnull] = ACTIONS(2634),
    [anon_sym___nullable] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2634),
    [anon_sym___real] = ACTIONS(2634),
    [anon_sym___strong] = ACTIONS(2634),
    [anon_sym___unsafe_unretained] = ACTIONS(2634),
    [anon_sym___unused] = ACTIONS(2634),
    [anon_sym___weak] = ACTIONS(2634),
    [anon_sym_alignas] = ACTIONS(2634),
    [anon_sym__Alignas] = ACTIONS(2634),
    [sym_primitive_type] = ACTIONS(2634),
    [anon_sym_enum] = ACTIONS(2634),
    [anon_sym_struct] = ACTIONS(2634),
    [anon_sym_union] = ACTIONS(2634),
    [anon_sym_if] = ACTIONS(2634),
    [anon_sym_switch] = ACTIONS(2634),
    [anon_sym_case] = ACTIONS(2634),
    [anon_sym_default] = ACTIONS(2634),
    [anon_sym_while] = ACTIONS(2634),
    [anon_sym_do] = ACTIONS(2634),
    [anon_sym_for] = ACTIONS(2634),
    [anon_sym_in] = ACTIONS(2634),
    [anon_sym_return] = ACTIONS(2634),
    [anon_sym_break] = ACTIONS(2634),
    [anon_sym_continue] = ACTIONS(2634),
    [anon_sym_goto] = ACTIONS(2634),
    [anon_sym_DASH_DASH] = ACTIONS(2636),
    [anon_sym_PLUS_PLUS] = ACTIONS(2636),
    [anon_sym_sizeof] = ACTIONS(2634),
    [anon_sym___alignof__] = ACTIONS(2634),
    [anon_sym___alignof] = ACTIONS(2634),
    [anon_sym__alignof] = ACTIONS(2634),
    [anon_sym_alignof] = ACTIONS(2634),
    [anon_sym__Alignof] = ACTIONS(2634),
    [anon_sym_offsetof] = ACTIONS(2634),
    [anon_sym__Generic] = ACTIONS(2634),
    [anon_sym_asm] = ACTIONS(2634),
    [anon_sym___asm__] = ACTIONS(2634),
    [anon_sym___asm] = ACTIONS(2634),
    [sym_number_literal] = ACTIONS(2636),
    [anon_sym_L_SQUOTE] = ACTIONS(2636),
    [anon_sym_u_SQUOTE] = ACTIONS(2636),
    [anon_sym_U_SQUOTE] = ACTIONS(2636),
    [anon_sym_u8_SQUOTE] = ACTIONS(2636),
    [anon_sym_SQUOTE] = ACTIONS(2636),
    [anon_sym_AT] = ACTIONS(2634),
    [anon_sym_DQUOTE] = ACTIONS(2636),
    [anon_sym_L_DQUOTE] = ACTIONS(2636),
    [anon_sym_u_DQUOTE] = ACTIONS(2636),
    [anon_sym_U_DQUOTE] = ACTIONS(2636),
    [anon_sym_u8_DQUOTE] = ACTIONS(2636),
    [sym_true] = ACTIONS(2634),
    [sym_false] = ACTIONS(2634),
    [anon_sym_NULL] = ACTIONS(2634),
    [anon_sym_nullptr] = ACTIONS(2634),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2634),
    [anon_sym___typeof] = ACTIONS(2634),
    [anon_sym_typeof] = ACTIONS(2634),
    [anon_sym_ATimport] = ACTIONS(2636),
    [aux_sym_preproc_undef_token1] = ACTIONS(2634),
    [anon_sym_POUND] = ACTIONS(2634),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2634),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2634),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2634),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_API_AVAILABLE] = ACTIONS(2634),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_API_DEPRECATED] = ACTIONS(2634),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym___deprecated_msg] = ACTIONS(2634),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2634),
    [anon_sym_ATprotocol] = ACTIONS(2636),
    [anon_sym_ATinterface] = ACTIONS(2636),
    [anon_sym_ATimplementation] = ACTIONS(2636),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2636),
    [anon_sym_ATsynthesize] = ACTIONS(2636),
    [anon_sym_ATdynamic] = ACTIONS(2636),
    [anon_sym_ATtry] = ACTIONS(2636),
    [anon_sym___try] = ACTIONS(2634),
    [anon_sym_ATthrow] = ACTIONS(2636),
    [anon_sym_ATselector] = ACTIONS(2636),
    [anon_sym_ATavailable] = ACTIONS(2636),
    [anon_sym___builtin_available] = ACTIONS(2634),
    [anon_sym_va_arg] = ACTIONS(2634),
    [anon_sym_ATencode] = ACTIONS(2636),
    [anon_sym_ATsynchronized] = ACTIONS(2636),
    [anon_sym_BOOL] = ACTIONS(2634),
    [anon_sym_IMP] = ACTIONS(2634),
    [anon_sym_SEL] = ACTIONS(2634),
    [anon_sym_Class] = ACTIONS(2634),
    [anon_sym_id] = ACTIONS(2634),
  },
  [STATE(294)] = {
    [sym_identifier] = ACTIONS(2638),
    [aux_sym_preproc_include_token1] = ACTIONS(2638),
    [aux_sym_preproc_include_token2] = ACTIONS(2638),
    [aux_sym_preproc_def_token1] = ACTIONS(2638),
    [aux_sym_preproc_if_token1] = ACTIONS(2638),
    [aux_sym_preproc_if_token2] = ACTIONS(2638),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2638),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2638),
    [aux_sym_preproc_else_token1] = ACTIONS(2638),
    [aux_sym_preproc_elif_token1] = ACTIONS(2638),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2638),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2638),
    [sym_preproc_directive] = ACTIONS(2638),
    [anon_sym_LPAREN2] = ACTIONS(2640),
    [anon_sym_BANG] = ACTIONS(2640),
    [anon_sym_TILDE] = ACTIONS(2640),
    [anon_sym_DASH] = ACTIONS(2638),
    [anon_sym_PLUS] = ACTIONS(2638),
    [anon_sym_STAR] = ACTIONS(2640),
    [anon_sym_CARET] = ACTIONS(2640),
    [anon_sym_AMP] = ACTIONS(2640),
    [anon_sym_SEMI] = ACTIONS(2640),
    [anon_sym___extension__] = ACTIONS(2638),
    [anon_sym_typedef] = ACTIONS(2638),
    [anon_sym_extern] = ACTIONS(2638),
    [anon_sym___attribute__] = ACTIONS(2638),
    [anon_sym___attribute] = ACTIONS(2638),
    [anon_sym_noreturn] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2640),
    [anon_sym___declspec] = ACTIONS(2638),
    [anon_sym___cdecl] = ACTIONS(2638),
    [anon_sym___clrcall] = ACTIONS(2638),
    [anon_sym___stdcall] = ACTIONS(2638),
    [anon_sym___fastcall] = ACTIONS(2638),
    [anon_sym___thiscall] = ACTIONS(2638),
    [anon_sym___vectorcall] = ACTIONS(2638),
    [anon_sym_LBRACE] = ACTIONS(2640),
    [anon_sym_signed] = ACTIONS(2638),
    [anon_sym_unsigned] = ACTIONS(2638),
    [anon_sym_long] = ACTIONS(2638),
    [anon_sym_short] = ACTIONS(2638),
    [anon_sym_static] = ACTIONS(2638),
    [anon_sym_ATautoreleasepool] = ACTIONS(2640),
    [anon_sym_auto] = ACTIONS(2638),
    [anon_sym_register] = ACTIONS(2638),
    [anon_sym_inline] = ACTIONS(2638),
    [anon_sym___inline] = ACTIONS(2638),
    [anon_sym___inline__] = ACTIONS(2638),
    [anon_sym___forceinline] = ACTIONS(2638),
    [anon_sym_thread_local] = ACTIONS(2638),
    [anon_sym___thread] = ACTIONS(2638),
    [anon_sym_CG_EXTERN] = ACTIONS(2638),
    [anon_sym_CG_INLINE] = ACTIONS(2638),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2638),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2638),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2638),
    [anon_sym_IBOutlet] = ACTIONS(2638),
    [anon_sym_IBInspectable] = ACTIONS(2638),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2638),
    [anon_sym_NS_INLINE] = ACTIONS(2638),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2638),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2638),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2638),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2638),
    [anon_sym_const] = ACTIONS(2638),
    [anon_sym_constexpr] = ACTIONS(2638),
    [anon_sym_volatile] = ACTIONS(2638),
    [anon_sym_restrict] = ACTIONS(2638),
    [anon_sym___restrict__] = ACTIONS(2638),
    [anon_sym__Atomic] = ACTIONS(2638),
    [anon_sym__Noreturn] = ACTIONS(2638),
    [anon_sym__Nonnull] = ACTIONS(2638),
    [anon_sym_nullable] = ACTIONS(2638),
    [anon_sym__Complex] = ACTIONS(2638),
    [anon_sym__Nullable] = ACTIONS(2638),
    [anon_sym__Nullable_result] = ACTIONS(2638),
    [anon_sym__Null_unspecified] = ACTIONS(2638),
    [anon_sym___autoreleasing] = ACTIONS(2638),
    [anon_sym___block] = ACTIONS(2638),
    [anon_sym___bridge] = ACTIONS(2638),
    [anon_sym___bridge_retained] = ACTIONS(2638),
    [anon_sym___bridge_transfer] = ACTIONS(2638),
    [anon_sym___complex] = ACTIONS(2638),
    [anon_sym___const] = ACTIONS(2638),
    [anon_sym___imag] = ACTIONS(2638),
    [anon_sym___kindof] = ACTIONS(2638),
    [anon_sym___nonnull] = ACTIONS(2638),
    [anon_sym___nullable] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2638),
    [anon_sym___real] = ACTIONS(2638),
    [anon_sym___strong] = ACTIONS(2638),
    [anon_sym___unsafe_unretained] = ACTIONS(2638),
    [anon_sym___unused] = ACTIONS(2638),
    [anon_sym___weak] = ACTIONS(2638),
    [anon_sym_alignas] = ACTIONS(2638),
    [anon_sym__Alignas] = ACTIONS(2638),
    [sym_primitive_type] = ACTIONS(2638),
    [anon_sym_enum] = ACTIONS(2638),
    [anon_sym_struct] = ACTIONS(2638),
    [anon_sym_union] = ACTIONS(2638),
    [anon_sym_if] = ACTIONS(2638),
    [anon_sym_switch] = ACTIONS(2638),
    [anon_sym_case] = ACTIONS(2638),
    [anon_sym_default] = ACTIONS(2638),
    [anon_sym_while] = ACTIONS(2638),
    [anon_sym_do] = ACTIONS(2638),
    [anon_sym_for] = ACTIONS(2638),
    [anon_sym_in] = ACTIONS(2638),
    [anon_sym_return] = ACTIONS(2638),
    [anon_sym_break] = ACTIONS(2638),
    [anon_sym_continue] = ACTIONS(2638),
    [anon_sym_goto] = ACTIONS(2638),
    [anon_sym_DASH_DASH] = ACTIONS(2640),
    [anon_sym_PLUS_PLUS] = ACTIONS(2640),
    [anon_sym_sizeof] = ACTIONS(2638),
    [anon_sym___alignof__] = ACTIONS(2638),
    [anon_sym___alignof] = ACTIONS(2638),
    [anon_sym__alignof] = ACTIONS(2638),
    [anon_sym_alignof] = ACTIONS(2638),
    [anon_sym__Alignof] = ACTIONS(2638),
    [anon_sym_offsetof] = ACTIONS(2638),
    [anon_sym__Generic] = ACTIONS(2638),
    [anon_sym_asm] = ACTIONS(2638),
    [anon_sym___asm__] = ACTIONS(2638),
    [anon_sym___asm] = ACTIONS(2638),
    [sym_number_literal] = ACTIONS(2640),
    [anon_sym_L_SQUOTE] = ACTIONS(2640),
    [anon_sym_u_SQUOTE] = ACTIONS(2640),
    [anon_sym_U_SQUOTE] = ACTIONS(2640),
    [anon_sym_u8_SQUOTE] = ACTIONS(2640),
    [anon_sym_SQUOTE] = ACTIONS(2640),
    [anon_sym_AT] = ACTIONS(2638),
    [anon_sym_DQUOTE] = ACTIONS(2640),
    [anon_sym_L_DQUOTE] = ACTIONS(2640),
    [anon_sym_u_DQUOTE] = ACTIONS(2640),
    [anon_sym_U_DQUOTE] = ACTIONS(2640),
    [anon_sym_u8_DQUOTE] = ACTIONS(2640),
    [sym_true] = ACTIONS(2638),
    [sym_false] = ACTIONS(2638),
    [anon_sym_NULL] = ACTIONS(2638),
    [anon_sym_nullptr] = ACTIONS(2638),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2638),
    [anon_sym___typeof] = ACTIONS(2638),
    [anon_sym_typeof] = ACTIONS(2638),
    [anon_sym_ATimport] = ACTIONS(2640),
    [aux_sym_preproc_undef_token1] = ACTIONS(2638),
    [anon_sym_POUND] = ACTIONS(2638),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2638),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2638),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2638),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2638),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2638),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2638),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2638),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_API_AVAILABLE] = ACTIONS(2638),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_API_DEPRECATED] = ACTIONS(2638),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2638),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym___deprecated_msg] = ACTIONS(2638),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2638),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2638),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2638),
    [anon_sym_ATprotocol] = ACTIONS(2640),
    [anon_sym_ATinterface] = ACTIONS(2640),
    [anon_sym_ATimplementation] = ACTIONS(2640),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2640),
    [anon_sym_ATsynthesize] = ACTIONS(2640),
    [anon_sym_ATdynamic] = ACTIONS(2640),
    [anon_sym_ATtry] = ACTIONS(2640),
    [anon_sym___try] = ACTIONS(2638),
    [anon_sym_ATthrow] = ACTIONS(2640),
    [anon_sym_ATselector] = ACTIONS(2640),
    [anon_sym_ATavailable] = ACTIONS(2640),
    [anon_sym___builtin_available] = ACTIONS(2638),
    [anon_sym_va_arg] = ACTIONS(2638),
    [anon_sym_ATencode] = ACTIONS(2640),
    [anon_sym_ATsynchronized] = ACTIONS(2640),
    [anon_sym_BOOL] = ACTIONS(2638),
    [anon_sym_IMP] = ACTIONS(2638),
    [anon_sym_SEL] = ACTIONS(2638),
    [anon_sym_Class] = ACTIONS(2638),
    [anon_sym_id] = ACTIONS(2638),
  },
  [STATE(295)] = {
    [sym_identifier] = ACTIONS(2642),
    [aux_sym_preproc_include_token1] = ACTIONS(2642),
    [aux_sym_preproc_include_token2] = ACTIONS(2642),
    [aux_sym_preproc_def_token1] = ACTIONS(2642),
    [aux_sym_preproc_if_token1] = ACTIONS(2642),
    [aux_sym_preproc_if_token2] = ACTIONS(2642),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2642),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2642),
    [aux_sym_preproc_else_token1] = ACTIONS(2642),
    [aux_sym_preproc_elif_token1] = ACTIONS(2642),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2642),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2642),
    [sym_preproc_directive] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_BANG] = ACTIONS(2644),
    [anon_sym_TILDE] = ACTIONS(2644),
    [anon_sym_DASH] = ACTIONS(2642),
    [anon_sym_PLUS] = ACTIONS(2642),
    [anon_sym_STAR] = ACTIONS(2644),
    [anon_sym_CARET] = ACTIONS(2644),
    [anon_sym_AMP] = ACTIONS(2644),
    [anon_sym_SEMI] = ACTIONS(2644),
    [anon_sym___extension__] = ACTIONS(2642),
    [anon_sym_typedef] = ACTIONS(2642),
    [anon_sym_extern] = ACTIONS(2642),
    [anon_sym___attribute__] = ACTIONS(2642),
    [anon_sym___attribute] = ACTIONS(2642),
    [anon_sym_noreturn] = ACTIONS(2642),
    [anon_sym_LBRACK] = ACTIONS(2644),
    [anon_sym___declspec] = ACTIONS(2642),
    [anon_sym___cdecl] = ACTIONS(2642),
    [anon_sym___clrcall] = ACTIONS(2642),
    [anon_sym___stdcall] = ACTIONS(2642),
    [anon_sym___fastcall] = ACTIONS(2642),
    [anon_sym___thiscall] = ACTIONS(2642),
    [anon_sym___vectorcall] = ACTIONS(2642),
    [anon_sym_LBRACE] = ACTIONS(2644),
    [anon_sym_signed] = ACTIONS(2642),
    [anon_sym_unsigned] = ACTIONS(2642),
    [anon_sym_long] = ACTIONS(2642),
    [anon_sym_short] = ACTIONS(2642),
    [anon_sym_static] = ACTIONS(2642),
    [anon_sym_ATautoreleasepool] = ACTIONS(2644),
    [anon_sym_auto] = ACTIONS(2642),
    [anon_sym_register] = ACTIONS(2642),
    [anon_sym_inline] = ACTIONS(2642),
    [anon_sym___inline] = ACTIONS(2642),
    [anon_sym___inline__] = ACTIONS(2642),
    [anon_sym___forceinline] = ACTIONS(2642),
    [anon_sym_thread_local] = ACTIONS(2642),
    [anon_sym___thread] = ACTIONS(2642),
    [anon_sym_CG_EXTERN] = ACTIONS(2642),
    [anon_sym_CG_INLINE] = ACTIONS(2642),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2642),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2642),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2642),
    [anon_sym_IBOutlet] = ACTIONS(2642),
    [anon_sym_IBInspectable] = ACTIONS(2642),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2642),
    [anon_sym_NS_INLINE] = ACTIONS(2642),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2642),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2642),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2642),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2642),
    [anon_sym_const] = ACTIONS(2642),
    [anon_sym_constexpr] = ACTIONS(2642),
    [anon_sym_volatile] = ACTIONS(2642),
    [anon_sym_restrict] = ACTIONS(2642),
    [anon_sym___restrict__] = ACTIONS(2642),
    [anon_sym__Atomic] = ACTIONS(2642),
    [anon_sym__Noreturn] = ACTIONS(2642),
    [anon_sym__Nonnull] = ACTIONS(2642),
    [anon_sym_nullable] = ACTIONS(2642),
    [anon_sym__Complex] = ACTIONS(2642),
    [anon_sym__Nullable] = ACTIONS(2642),
    [anon_sym__Nullable_result] = ACTIONS(2642),
    [anon_sym__Null_unspecified] = ACTIONS(2642),
    [anon_sym___autoreleasing] = ACTIONS(2642),
    [anon_sym___block] = ACTIONS(2642),
    [anon_sym___bridge] = ACTIONS(2642),
    [anon_sym___bridge_retained] = ACTIONS(2642),
    [anon_sym___bridge_transfer] = ACTIONS(2642),
    [anon_sym___complex] = ACTIONS(2642),
    [anon_sym___const] = ACTIONS(2642),
    [anon_sym___imag] = ACTIONS(2642),
    [anon_sym___kindof] = ACTIONS(2642),
    [anon_sym___nonnull] = ACTIONS(2642),
    [anon_sym___nullable] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2642),
    [anon_sym___real] = ACTIONS(2642),
    [anon_sym___strong] = ACTIONS(2642),
    [anon_sym___unsafe_unretained] = ACTIONS(2642),
    [anon_sym___unused] = ACTIONS(2642),
    [anon_sym___weak] = ACTIONS(2642),
    [anon_sym_alignas] = ACTIONS(2642),
    [anon_sym__Alignas] = ACTIONS(2642),
    [sym_primitive_type] = ACTIONS(2642),
    [anon_sym_enum] = ACTIONS(2642),
    [anon_sym_struct] = ACTIONS(2642),
    [anon_sym_union] = ACTIONS(2642),
    [anon_sym_if] = ACTIONS(2642),
    [anon_sym_switch] = ACTIONS(2642),
    [anon_sym_case] = ACTIONS(2642),
    [anon_sym_default] = ACTIONS(2642),
    [anon_sym_while] = ACTIONS(2642),
    [anon_sym_do] = ACTIONS(2642),
    [anon_sym_for] = ACTIONS(2642),
    [anon_sym_in] = ACTIONS(2642),
    [anon_sym_return] = ACTIONS(2642),
    [anon_sym_break] = ACTIONS(2642),
    [anon_sym_continue] = ACTIONS(2642),
    [anon_sym_goto] = ACTIONS(2642),
    [anon_sym_DASH_DASH] = ACTIONS(2644),
    [anon_sym_PLUS_PLUS] = ACTIONS(2644),
    [anon_sym_sizeof] = ACTIONS(2642),
    [anon_sym___alignof__] = ACTIONS(2642),
    [anon_sym___alignof] = ACTIONS(2642),
    [anon_sym__alignof] = ACTIONS(2642),
    [anon_sym_alignof] = ACTIONS(2642),
    [anon_sym__Alignof] = ACTIONS(2642),
    [anon_sym_offsetof] = ACTIONS(2642),
    [anon_sym__Generic] = ACTIONS(2642),
    [anon_sym_asm] = ACTIONS(2642),
    [anon_sym___asm__] = ACTIONS(2642),
    [anon_sym___asm] = ACTIONS(2642),
    [sym_number_literal] = ACTIONS(2644),
    [anon_sym_L_SQUOTE] = ACTIONS(2644),
    [anon_sym_u_SQUOTE] = ACTIONS(2644),
    [anon_sym_U_SQUOTE] = ACTIONS(2644),
    [anon_sym_u8_SQUOTE] = ACTIONS(2644),
    [anon_sym_SQUOTE] = ACTIONS(2644),
    [anon_sym_AT] = ACTIONS(2642),
    [anon_sym_DQUOTE] = ACTIONS(2644),
    [anon_sym_L_DQUOTE] = ACTIONS(2644),
    [anon_sym_u_DQUOTE] = ACTIONS(2644),
    [anon_sym_U_DQUOTE] = ACTIONS(2644),
    [anon_sym_u8_DQUOTE] = ACTIONS(2644),
    [sym_true] = ACTIONS(2642),
    [sym_false] = ACTIONS(2642),
    [anon_sym_NULL] = ACTIONS(2642),
    [anon_sym_nullptr] = ACTIONS(2642),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2642),
    [anon_sym___typeof] = ACTIONS(2642),
    [anon_sym_typeof] = ACTIONS(2642),
    [anon_sym_ATimport] = ACTIONS(2644),
    [aux_sym_preproc_undef_token1] = ACTIONS(2642),
    [anon_sym_POUND] = ACTIONS(2642),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2642),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2642),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2642),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2642),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2642),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2642),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2642),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_API_AVAILABLE] = ACTIONS(2642),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_API_DEPRECATED] = ACTIONS(2642),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2642),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym___deprecated_msg] = ACTIONS(2642),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2642),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2642),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2642),
    [anon_sym_ATprotocol] = ACTIONS(2644),
    [anon_sym_ATinterface] = ACTIONS(2644),
    [anon_sym_ATimplementation] = ACTIONS(2644),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2644),
    [anon_sym_ATsynthesize] = ACTIONS(2644),
    [anon_sym_ATdynamic] = ACTIONS(2644),
    [anon_sym_ATtry] = ACTIONS(2644),
    [anon_sym___try] = ACTIONS(2642),
    [anon_sym_ATthrow] = ACTIONS(2644),
    [anon_sym_ATselector] = ACTIONS(2644),
    [anon_sym_ATavailable] = ACTIONS(2644),
    [anon_sym___builtin_available] = ACTIONS(2642),
    [anon_sym_va_arg] = ACTIONS(2642),
    [anon_sym_ATencode] = ACTIONS(2644),
    [anon_sym_ATsynchronized] = ACTIONS(2644),
    [anon_sym_BOOL] = ACTIONS(2642),
    [anon_sym_IMP] = ACTIONS(2642),
    [anon_sym_SEL] = ACTIONS(2642),
    [anon_sym_Class] = ACTIONS(2642),
    [anon_sym_id] = ACTIONS(2642),
  },
  [STATE(296)] = {
    [sym_identifier] = ACTIONS(2646),
    [aux_sym_preproc_include_token1] = ACTIONS(2646),
    [aux_sym_preproc_include_token2] = ACTIONS(2646),
    [aux_sym_preproc_def_token1] = ACTIONS(2646),
    [aux_sym_preproc_if_token1] = ACTIONS(2646),
    [aux_sym_preproc_if_token2] = ACTIONS(2646),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2646),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2646),
    [aux_sym_preproc_else_token1] = ACTIONS(2646),
    [aux_sym_preproc_elif_token1] = ACTIONS(2646),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2646),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2646),
    [sym_preproc_directive] = ACTIONS(2646),
    [anon_sym_LPAREN2] = ACTIONS(2648),
    [anon_sym_BANG] = ACTIONS(2648),
    [anon_sym_TILDE] = ACTIONS(2648),
    [anon_sym_DASH] = ACTIONS(2646),
    [anon_sym_PLUS] = ACTIONS(2646),
    [anon_sym_STAR] = ACTIONS(2648),
    [anon_sym_CARET] = ACTIONS(2648),
    [anon_sym_AMP] = ACTIONS(2648),
    [anon_sym_SEMI] = ACTIONS(2648),
    [anon_sym___extension__] = ACTIONS(2646),
    [anon_sym_typedef] = ACTIONS(2646),
    [anon_sym_extern] = ACTIONS(2646),
    [anon_sym___attribute__] = ACTIONS(2646),
    [anon_sym___attribute] = ACTIONS(2646),
    [anon_sym_noreturn] = ACTIONS(2646),
    [anon_sym_LBRACK] = ACTIONS(2648),
    [anon_sym___declspec] = ACTIONS(2646),
    [anon_sym___cdecl] = ACTIONS(2646),
    [anon_sym___clrcall] = ACTIONS(2646),
    [anon_sym___stdcall] = ACTIONS(2646),
    [anon_sym___fastcall] = ACTIONS(2646),
    [anon_sym___thiscall] = ACTIONS(2646),
    [anon_sym___vectorcall] = ACTIONS(2646),
    [anon_sym_LBRACE] = ACTIONS(2648),
    [anon_sym_signed] = ACTIONS(2646),
    [anon_sym_unsigned] = ACTIONS(2646),
    [anon_sym_long] = ACTIONS(2646),
    [anon_sym_short] = ACTIONS(2646),
    [anon_sym_static] = ACTIONS(2646),
    [anon_sym_ATautoreleasepool] = ACTIONS(2648),
    [anon_sym_auto] = ACTIONS(2646),
    [anon_sym_register] = ACTIONS(2646),
    [anon_sym_inline] = ACTIONS(2646),
    [anon_sym___inline] = ACTIONS(2646),
    [anon_sym___inline__] = ACTIONS(2646),
    [anon_sym___forceinline] = ACTIONS(2646),
    [anon_sym_thread_local] = ACTIONS(2646),
    [anon_sym___thread] = ACTIONS(2646),
    [anon_sym_CG_EXTERN] = ACTIONS(2646),
    [anon_sym_CG_INLINE] = ACTIONS(2646),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2646),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2646),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2646),
    [anon_sym_IBOutlet] = ACTIONS(2646),
    [anon_sym_IBInspectable] = ACTIONS(2646),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2646),
    [anon_sym_NS_INLINE] = ACTIONS(2646),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2646),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2646),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2646),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2646),
    [anon_sym_const] = ACTIONS(2646),
    [anon_sym_constexpr] = ACTIONS(2646),
    [anon_sym_volatile] = ACTIONS(2646),
    [anon_sym_restrict] = ACTIONS(2646),
    [anon_sym___restrict__] = ACTIONS(2646),
    [anon_sym__Atomic] = ACTIONS(2646),
    [anon_sym__Noreturn] = ACTIONS(2646),
    [anon_sym__Nonnull] = ACTIONS(2646),
    [anon_sym_nullable] = ACTIONS(2646),
    [anon_sym__Complex] = ACTIONS(2646),
    [anon_sym__Nullable] = ACTIONS(2646),
    [anon_sym__Nullable_result] = ACTIONS(2646),
    [anon_sym__Null_unspecified] = ACTIONS(2646),
    [anon_sym___autoreleasing] = ACTIONS(2646),
    [anon_sym___block] = ACTIONS(2646),
    [anon_sym___bridge] = ACTIONS(2646),
    [anon_sym___bridge_retained] = ACTIONS(2646),
    [anon_sym___bridge_transfer] = ACTIONS(2646),
    [anon_sym___complex] = ACTIONS(2646),
    [anon_sym___const] = ACTIONS(2646),
    [anon_sym___imag] = ACTIONS(2646),
    [anon_sym___kindof] = ACTIONS(2646),
    [anon_sym___nonnull] = ACTIONS(2646),
    [anon_sym___nullable] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2646),
    [anon_sym___real] = ACTIONS(2646),
    [anon_sym___strong] = ACTIONS(2646),
    [anon_sym___unsafe_unretained] = ACTIONS(2646),
    [anon_sym___unused] = ACTIONS(2646),
    [anon_sym___weak] = ACTIONS(2646),
    [anon_sym_alignas] = ACTIONS(2646),
    [anon_sym__Alignas] = ACTIONS(2646),
    [sym_primitive_type] = ACTIONS(2646),
    [anon_sym_enum] = ACTIONS(2646),
    [anon_sym_struct] = ACTIONS(2646),
    [anon_sym_union] = ACTIONS(2646),
    [anon_sym_if] = ACTIONS(2646),
    [anon_sym_switch] = ACTIONS(2646),
    [anon_sym_case] = ACTIONS(2646),
    [anon_sym_default] = ACTIONS(2646),
    [anon_sym_while] = ACTIONS(2646),
    [anon_sym_do] = ACTIONS(2646),
    [anon_sym_for] = ACTIONS(2646),
    [anon_sym_in] = ACTIONS(2646),
    [anon_sym_return] = ACTIONS(2646),
    [anon_sym_break] = ACTIONS(2646),
    [anon_sym_continue] = ACTIONS(2646),
    [anon_sym_goto] = ACTIONS(2646),
    [anon_sym_DASH_DASH] = ACTIONS(2648),
    [anon_sym_PLUS_PLUS] = ACTIONS(2648),
    [anon_sym_sizeof] = ACTIONS(2646),
    [anon_sym___alignof__] = ACTIONS(2646),
    [anon_sym___alignof] = ACTIONS(2646),
    [anon_sym__alignof] = ACTIONS(2646),
    [anon_sym_alignof] = ACTIONS(2646),
    [anon_sym__Alignof] = ACTIONS(2646),
    [anon_sym_offsetof] = ACTIONS(2646),
    [anon_sym__Generic] = ACTIONS(2646),
    [anon_sym_asm] = ACTIONS(2646),
    [anon_sym___asm__] = ACTIONS(2646),
    [anon_sym___asm] = ACTIONS(2646),
    [sym_number_literal] = ACTIONS(2648),
    [anon_sym_L_SQUOTE] = ACTIONS(2648),
    [anon_sym_u_SQUOTE] = ACTIONS(2648),
    [anon_sym_U_SQUOTE] = ACTIONS(2648),
    [anon_sym_u8_SQUOTE] = ACTIONS(2648),
    [anon_sym_SQUOTE] = ACTIONS(2648),
    [anon_sym_AT] = ACTIONS(2646),
    [anon_sym_DQUOTE] = ACTIONS(2648),
    [anon_sym_L_DQUOTE] = ACTIONS(2648),
    [anon_sym_u_DQUOTE] = ACTIONS(2648),
    [anon_sym_U_DQUOTE] = ACTIONS(2648),
    [anon_sym_u8_DQUOTE] = ACTIONS(2648),
    [sym_true] = ACTIONS(2646),
    [sym_false] = ACTIONS(2646),
    [anon_sym_NULL] = ACTIONS(2646),
    [anon_sym_nullptr] = ACTIONS(2646),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2646),
    [anon_sym___typeof] = ACTIONS(2646),
    [anon_sym_typeof] = ACTIONS(2646),
    [anon_sym_ATimport] = ACTIONS(2648),
    [aux_sym_preproc_undef_token1] = ACTIONS(2646),
    [anon_sym_POUND] = ACTIONS(2646),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2646),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2646),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2646),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2646),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2646),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2646),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2646),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_API_AVAILABLE] = ACTIONS(2646),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_API_DEPRECATED] = ACTIONS(2646),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2646),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym___deprecated_msg] = ACTIONS(2646),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2646),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2646),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2646),
    [anon_sym_ATprotocol] = ACTIONS(2648),
    [anon_sym_ATinterface] = ACTIONS(2648),
    [anon_sym_ATimplementation] = ACTIONS(2648),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2648),
    [anon_sym_ATsynthesize] = ACTIONS(2648),
    [anon_sym_ATdynamic] = ACTIONS(2648),
    [anon_sym_ATtry] = ACTIONS(2648),
    [anon_sym___try] = ACTIONS(2646),
    [anon_sym_ATthrow] = ACTIONS(2648),
    [anon_sym_ATselector] = ACTIONS(2648),
    [anon_sym_ATavailable] = ACTIONS(2648),
    [anon_sym___builtin_available] = ACTIONS(2646),
    [anon_sym_va_arg] = ACTIONS(2646),
    [anon_sym_ATencode] = ACTIONS(2648),
    [anon_sym_ATsynchronized] = ACTIONS(2648),
    [anon_sym_BOOL] = ACTIONS(2646),
    [anon_sym_IMP] = ACTIONS(2646),
    [anon_sym_SEL] = ACTIONS(2646),
    [anon_sym_Class] = ACTIONS(2646),
    [anon_sym_id] = ACTIONS(2646),
  },
  [STATE(297)] = {
    [sym_identifier] = ACTIONS(2650),
    [aux_sym_preproc_include_token1] = ACTIONS(2650),
    [aux_sym_preproc_include_token2] = ACTIONS(2650),
    [aux_sym_preproc_def_token1] = ACTIONS(2650),
    [aux_sym_preproc_if_token1] = ACTIONS(2650),
    [aux_sym_preproc_if_token2] = ACTIONS(2650),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2650),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2650),
    [aux_sym_preproc_else_token1] = ACTIONS(2650),
    [aux_sym_preproc_elif_token1] = ACTIONS(2650),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2650),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2650),
    [sym_preproc_directive] = ACTIONS(2650),
    [anon_sym_LPAREN2] = ACTIONS(2652),
    [anon_sym_BANG] = ACTIONS(2652),
    [anon_sym_TILDE] = ACTIONS(2652),
    [anon_sym_DASH] = ACTIONS(2650),
    [anon_sym_PLUS] = ACTIONS(2650),
    [anon_sym_STAR] = ACTIONS(2652),
    [anon_sym_CARET] = ACTIONS(2652),
    [anon_sym_AMP] = ACTIONS(2652),
    [anon_sym_SEMI] = ACTIONS(2652),
    [anon_sym___extension__] = ACTIONS(2650),
    [anon_sym_typedef] = ACTIONS(2650),
    [anon_sym_extern] = ACTIONS(2650),
    [anon_sym___attribute__] = ACTIONS(2650),
    [anon_sym___attribute] = ACTIONS(2650),
    [anon_sym_noreturn] = ACTIONS(2650),
    [anon_sym_LBRACK] = ACTIONS(2652),
    [anon_sym___declspec] = ACTIONS(2650),
    [anon_sym___cdecl] = ACTIONS(2650),
    [anon_sym___clrcall] = ACTIONS(2650),
    [anon_sym___stdcall] = ACTIONS(2650),
    [anon_sym___fastcall] = ACTIONS(2650),
    [anon_sym___thiscall] = ACTIONS(2650),
    [anon_sym___vectorcall] = ACTIONS(2650),
    [anon_sym_LBRACE] = ACTIONS(2652),
    [anon_sym_signed] = ACTIONS(2650),
    [anon_sym_unsigned] = ACTIONS(2650),
    [anon_sym_long] = ACTIONS(2650),
    [anon_sym_short] = ACTIONS(2650),
    [anon_sym_static] = ACTIONS(2650),
    [anon_sym_ATautoreleasepool] = ACTIONS(2652),
    [anon_sym_auto] = ACTIONS(2650),
    [anon_sym_register] = ACTIONS(2650),
    [anon_sym_inline] = ACTIONS(2650),
    [anon_sym___inline] = ACTIONS(2650),
    [anon_sym___inline__] = ACTIONS(2650),
    [anon_sym___forceinline] = ACTIONS(2650),
    [anon_sym_thread_local] = ACTIONS(2650),
    [anon_sym___thread] = ACTIONS(2650),
    [anon_sym_CG_EXTERN] = ACTIONS(2650),
    [anon_sym_CG_INLINE] = ACTIONS(2650),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2650),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2650),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2650),
    [anon_sym_IBOutlet] = ACTIONS(2650),
    [anon_sym_IBInspectable] = ACTIONS(2650),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2650),
    [anon_sym_NS_INLINE] = ACTIONS(2650),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2650),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2650),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2650),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2650),
    [anon_sym_const] = ACTIONS(2650),
    [anon_sym_constexpr] = ACTIONS(2650),
    [anon_sym_volatile] = ACTIONS(2650),
    [anon_sym_restrict] = ACTIONS(2650),
    [anon_sym___restrict__] = ACTIONS(2650),
    [anon_sym__Atomic] = ACTIONS(2650),
    [anon_sym__Noreturn] = ACTIONS(2650),
    [anon_sym__Nonnull] = ACTIONS(2650),
    [anon_sym_nullable] = ACTIONS(2650),
    [anon_sym__Complex] = ACTIONS(2650),
    [anon_sym__Nullable] = ACTIONS(2650),
    [anon_sym__Nullable_result] = ACTIONS(2650),
    [anon_sym__Null_unspecified] = ACTIONS(2650),
    [anon_sym___autoreleasing] = ACTIONS(2650),
    [anon_sym___block] = ACTIONS(2650),
    [anon_sym___bridge] = ACTIONS(2650),
    [anon_sym___bridge_retained] = ACTIONS(2650),
    [anon_sym___bridge_transfer] = ACTIONS(2650),
    [anon_sym___complex] = ACTIONS(2650),
    [anon_sym___const] = ACTIONS(2650),
    [anon_sym___imag] = ACTIONS(2650),
    [anon_sym___kindof] = ACTIONS(2650),
    [anon_sym___nonnull] = ACTIONS(2650),
    [anon_sym___nullable] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2650),
    [anon_sym___real] = ACTIONS(2650),
    [anon_sym___strong] = ACTIONS(2650),
    [anon_sym___unsafe_unretained] = ACTIONS(2650),
    [anon_sym___unused] = ACTIONS(2650),
    [anon_sym___weak] = ACTIONS(2650),
    [anon_sym_alignas] = ACTIONS(2650),
    [anon_sym__Alignas] = ACTIONS(2650),
    [sym_primitive_type] = ACTIONS(2650),
    [anon_sym_enum] = ACTIONS(2650),
    [anon_sym_struct] = ACTIONS(2650),
    [anon_sym_union] = ACTIONS(2650),
    [anon_sym_if] = ACTIONS(2650),
    [anon_sym_switch] = ACTIONS(2650),
    [anon_sym_case] = ACTIONS(2650),
    [anon_sym_default] = ACTIONS(2650),
    [anon_sym_while] = ACTIONS(2650),
    [anon_sym_do] = ACTIONS(2650),
    [anon_sym_for] = ACTIONS(2650),
    [anon_sym_in] = ACTIONS(2650),
    [anon_sym_return] = ACTIONS(2650),
    [anon_sym_break] = ACTIONS(2650),
    [anon_sym_continue] = ACTIONS(2650),
    [anon_sym_goto] = ACTIONS(2650),
    [anon_sym_DASH_DASH] = ACTIONS(2652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2652),
    [anon_sym_sizeof] = ACTIONS(2650),
    [anon_sym___alignof__] = ACTIONS(2650),
    [anon_sym___alignof] = ACTIONS(2650),
    [anon_sym__alignof] = ACTIONS(2650),
    [anon_sym_alignof] = ACTIONS(2650),
    [anon_sym__Alignof] = ACTIONS(2650),
    [anon_sym_offsetof] = ACTIONS(2650),
    [anon_sym__Generic] = ACTIONS(2650),
    [anon_sym_asm] = ACTIONS(2650),
    [anon_sym___asm__] = ACTIONS(2650),
    [anon_sym___asm] = ACTIONS(2650),
    [sym_number_literal] = ACTIONS(2652),
    [anon_sym_L_SQUOTE] = ACTIONS(2652),
    [anon_sym_u_SQUOTE] = ACTIONS(2652),
    [anon_sym_U_SQUOTE] = ACTIONS(2652),
    [anon_sym_u8_SQUOTE] = ACTIONS(2652),
    [anon_sym_SQUOTE] = ACTIONS(2652),
    [anon_sym_AT] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_L_DQUOTE] = ACTIONS(2652),
    [anon_sym_u_DQUOTE] = ACTIONS(2652),
    [anon_sym_U_DQUOTE] = ACTIONS(2652),
    [anon_sym_u8_DQUOTE] = ACTIONS(2652),
    [sym_true] = ACTIONS(2650),
    [sym_false] = ACTIONS(2650),
    [anon_sym_NULL] = ACTIONS(2650),
    [anon_sym_nullptr] = ACTIONS(2650),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2650),
    [anon_sym___typeof] = ACTIONS(2650),
    [anon_sym_typeof] = ACTIONS(2650),
    [anon_sym_ATimport] = ACTIONS(2652),
    [aux_sym_preproc_undef_token1] = ACTIONS(2650),
    [anon_sym_POUND] = ACTIONS(2650),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2650),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2650),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2650),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2650),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2650),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2650),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2650),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_API_AVAILABLE] = ACTIONS(2650),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_API_DEPRECATED] = ACTIONS(2650),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2650),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym___deprecated_msg] = ACTIONS(2650),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2650),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2650),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2650),
    [anon_sym_ATprotocol] = ACTIONS(2652),
    [anon_sym_ATinterface] = ACTIONS(2652),
    [anon_sym_ATimplementation] = ACTIONS(2652),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2652),
    [anon_sym_ATsynthesize] = ACTIONS(2652),
    [anon_sym_ATdynamic] = ACTIONS(2652),
    [anon_sym_ATtry] = ACTIONS(2652),
    [anon_sym___try] = ACTIONS(2650),
    [anon_sym_ATthrow] = ACTIONS(2652),
    [anon_sym_ATselector] = ACTIONS(2652),
    [anon_sym_ATavailable] = ACTIONS(2652),
    [anon_sym___builtin_available] = ACTIONS(2650),
    [anon_sym_va_arg] = ACTIONS(2650),
    [anon_sym_ATencode] = ACTIONS(2652),
    [anon_sym_ATsynchronized] = ACTIONS(2652),
    [anon_sym_BOOL] = ACTIONS(2650),
    [anon_sym_IMP] = ACTIONS(2650),
    [anon_sym_SEL] = ACTIONS(2650),
    [anon_sym_Class] = ACTIONS(2650),
    [anon_sym_id] = ACTIONS(2650),
  },
  [STATE(298)] = {
    [sym_identifier] = ACTIONS(2654),
    [aux_sym_preproc_include_token1] = ACTIONS(2654),
    [aux_sym_preproc_include_token2] = ACTIONS(2654),
    [aux_sym_preproc_def_token1] = ACTIONS(2654),
    [aux_sym_preproc_if_token1] = ACTIONS(2654),
    [aux_sym_preproc_if_token2] = ACTIONS(2654),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2654),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2654),
    [aux_sym_preproc_else_token1] = ACTIONS(2654),
    [aux_sym_preproc_elif_token1] = ACTIONS(2654),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2654),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2654),
    [sym_preproc_directive] = ACTIONS(2654),
    [anon_sym_LPAREN2] = ACTIONS(2656),
    [anon_sym_BANG] = ACTIONS(2656),
    [anon_sym_TILDE] = ACTIONS(2656),
    [anon_sym_DASH] = ACTIONS(2654),
    [anon_sym_PLUS] = ACTIONS(2654),
    [anon_sym_STAR] = ACTIONS(2656),
    [anon_sym_CARET] = ACTIONS(2656),
    [anon_sym_AMP] = ACTIONS(2656),
    [anon_sym_SEMI] = ACTIONS(2656),
    [anon_sym___extension__] = ACTIONS(2654),
    [anon_sym_typedef] = ACTIONS(2654),
    [anon_sym_extern] = ACTIONS(2654),
    [anon_sym___attribute__] = ACTIONS(2654),
    [anon_sym___attribute] = ACTIONS(2654),
    [anon_sym_noreturn] = ACTIONS(2654),
    [anon_sym_LBRACK] = ACTIONS(2656),
    [anon_sym___declspec] = ACTIONS(2654),
    [anon_sym___cdecl] = ACTIONS(2654),
    [anon_sym___clrcall] = ACTIONS(2654),
    [anon_sym___stdcall] = ACTIONS(2654),
    [anon_sym___fastcall] = ACTIONS(2654),
    [anon_sym___thiscall] = ACTIONS(2654),
    [anon_sym___vectorcall] = ACTIONS(2654),
    [anon_sym_LBRACE] = ACTIONS(2656),
    [anon_sym_signed] = ACTIONS(2654),
    [anon_sym_unsigned] = ACTIONS(2654),
    [anon_sym_long] = ACTIONS(2654),
    [anon_sym_short] = ACTIONS(2654),
    [anon_sym_static] = ACTIONS(2654),
    [anon_sym_ATautoreleasepool] = ACTIONS(2656),
    [anon_sym_auto] = ACTIONS(2654),
    [anon_sym_register] = ACTIONS(2654),
    [anon_sym_inline] = ACTIONS(2654),
    [anon_sym___inline] = ACTIONS(2654),
    [anon_sym___inline__] = ACTIONS(2654),
    [anon_sym___forceinline] = ACTIONS(2654),
    [anon_sym_thread_local] = ACTIONS(2654),
    [anon_sym___thread] = ACTIONS(2654),
    [anon_sym_CG_EXTERN] = ACTIONS(2654),
    [anon_sym_CG_INLINE] = ACTIONS(2654),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2654),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2654),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2654),
    [anon_sym_IBOutlet] = ACTIONS(2654),
    [anon_sym_IBInspectable] = ACTIONS(2654),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2654),
    [anon_sym_NS_INLINE] = ACTIONS(2654),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2654),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2654),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2654),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2654),
    [anon_sym_const] = ACTIONS(2654),
    [anon_sym_constexpr] = ACTIONS(2654),
    [anon_sym_volatile] = ACTIONS(2654),
    [anon_sym_restrict] = ACTIONS(2654),
    [anon_sym___restrict__] = ACTIONS(2654),
    [anon_sym__Atomic] = ACTIONS(2654),
    [anon_sym__Noreturn] = ACTIONS(2654),
    [anon_sym__Nonnull] = ACTIONS(2654),
    [anon_sym_nullable] = ACTIONS(2654),
    [anon_sym__Complex] = ACTIONS(2654),
    [anon_sym__Nullable] = ACTIONS(2654),
    [anon_sym__Nullable_result] = ACTIONS(2654),
    [anon_sym__Null_unspecified] = ACTIONS(2654),
    [anon_sym___autoreleasing] = ACTIONS(2654),
    [anon_sym___block] = ACTIONS(2654),
    [anon_sym___bridge] = ACTIONS(2654),
    [anon_sym___bridge_retained] = ACTIONS(2654),
    [anon_sym___bridge_transfer] = ACTIONS(2654),
    [anon_sym___complex] = ACTIONS(2654),
    [anon_sym___const] = ACTIONS(2654),
    [anon_sym___imag] = ACTIONS(2654),
    [anon_sym___kindof] = ACTIONS(2654),
    [anon_sym___nonnull] = ACTIONS(2654),
    [anon_sym___nullable] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2654),
    [anon_sym___real] = ACTIONS(2654),
    [anon_sym___strong] = ACTIONS(2654),
    [anon_sym___unsafe_unretained] = ACTIONS(2654),
    [anon_sym___unused] = ACTIONS(2654),
    [anon_sym___weak] = ACTIONS(2654),
    [anon_sym_alignas] = ACTIONS(2654),
    [anon_sym__Alignas] = ACTIONS(2654),
    [sym_primitive_type] = ACTIONS(2654),
    [anon_sym_enum] = ACTIONS(2654),
    [anon_sym_struct] = ACTIONS(2654),
    [anon_sym_union] = ACTIONS(2654),
    [anon_sym_if] = ACTIONS(2654),
    [anon_sym_switch] = ACTIONS(2654),
    [anon_sym_case] = ACTIONS(2654),
    [anon_sym_default] = ACTIONS(2654),
    [anon_sym_while] = ACTIONS(2654),
    [anon_sym_do] = ACTIONS(2654),
    [anon_sym_for] = ACTIONS(2654),
    [anon_sym_in] = ACTIONS(2654),
    [anon_sym_return] = ACTIONS(2654),
    [anon_sym_break] = ACTIONS(2654),
    [anon_sym_continue] = ACTIONS(2654),
    [anon_sym_goto] = ACTIONS(2654),
    [anon_sym_DASH_DASH] = ACTIONS(2656),
    [anon_sym_PLUS_PLUS] = ACTIONS(2656),
    [anon_sym_sizeof] = ACTIONS(2654),
    [anon_sym___alignof__] = ACTIONS(2654),
    [anon_sym___alignof] = ACTIONS(2654),
    [anon_sym__alignof] = ACTIONS(2654),
    [anon_sym_alignof] = ACTIONS(2654),
    [anon_sym__Alignof] = ACTIONS(2654),
    [anon_sym_offsetof] = ACTIONS(2654),
    [anon_sym__Generic] = ACTIONS(2654),
    [anon_sym_asm] = ACTIONS(2654),
    [anon_sym___asm__] = ACTIONS(2654),
    [anon_sym___asm] = ACTIONS(2654),
    [sym_number_literal] = ACTIONS(2656),
    [anon_sym_L_SQUOTE] = ACTIONS(2656),
    [anon_sym_u_SQUOTE] = ACTIONS(2656),
    [anon_sym_U_SQUOTE] = ACTIONS(2656),
    [anon_sym_u8_SQUOTE] = ACTIONS(2656),
    [anon_sym_SQUOTE] = ACTIONS(2656),
    [anon_sym_AT] = ACTIONS(2654),
    [anon_sym_DQUOTE] = ACTIONS(2656),
    [anon_sym_L_DQUOTE] = ACTIONS(2656),
    [anon_sym_u_DQUOTE] = ACTIONS(2656),
    [anon_sym_U_DQUOTE] = ACTIONS(2656),
    [anon_sym_u8_DQUOTE] = ACTIONS(2656),
    [sym_true] = ACTIONS(2654),
    [sym_false] = ACTIONS(2654),
    [anon_sym_NULL] = ACTIONS(2654),
    [anon_sym_nullptr] = ACTIONS(2654),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2654),
    [anon_sym___typeof] = ACTIONS(2654),
    [anon_sym_typeof] = ACTIONS(2654),
    [anon_sym_ATimport] = ACTIONS(2656),
    [aux_sym_preproc_undef_token1] = ACTIONS(2654),
    [anon_sym_POUND] = ACTIONS(2654),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2654),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2654),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2654),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2654),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2654),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2654),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2654),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_API_AVAILABLE] = ACTIONS(2654),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_API_DEPRECATED] = ACTIONS(2654),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2654),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym___deprecated_msg] = ACTIONS(2654),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2654),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2654),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2654),
    [anon_sym_ATprotocol] = ACTIONS(2656),
    [anon_sym_ATinterface] = ACTIONS(2656),
    [anon_sym_ATimplementation] = ACTIONS(2656),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2656),
    [anon_sym_ATsynthesize] = ACTIONS(2656),
    [anon_sym_ATdynamic] = ACTIONS(2656),
    [anon_sym_ATtry] = ACTIONS(2656),
    [anon_sym___try] = ACTIONS(2654),
    [anon_sym_ATthrow] = ACTIONS(2656),
    [anon_sym_ATselector] = ACTIONS(2656),
    [anon_sym_ATavailable] = ACTIONS(2656),
    [anon_sym___builtin_available] = ACTIONS(2654),
    [anon_sym_va_arg] = ACTIONS(2654),
    [anon_sym_ATencode] = ACTIONS(2656),
    [anon_sym_ATsynchronized] = ACTIONS(2656),
    [anon_sym_BOOL] = ACTIONS(2654),
    [anon_sym_IMP] = ACTIONS(2654),
    [anon_sym_SEL] = ACTIONS(2654),
    [anon_sym_Class] = ACTIONS(2654),
    [anon_sym_id] = ACTIONS(2654),
  },
  [STATE(299)] = {
    [sym_identifier] = ACTIONS(2658),
    [aux_sym_preproc_include_token1] = ACTIONS(2658),
    [aux_sym_preproc_include_token2] = ACTIONS(2658),
    [aux_sym_preproc_def_token1] = ACTIONS(2658),
    [aux_sym_preproc_if_token1] = ACTIONS(2658),
    [aux_sym_preproc_if_token2] = ACTIONS(2658),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2658),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2658),
    [aux_sym_preproc_else_token1] = ACTIONS(2658),
    [aux_sym_preproc_elif_token1] = ACTIONS(2658),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2658),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2658),
    [sym_preproc_directive] = ACTIONS(2658),
    [anon_sym_LPAREN2] = ACTIONS(2660),
    [anon_sym_BANG] = ACTIONS(2660),
    [anon_sym_TILDE] = ACTIONS(2660),
    [anon_sym_DASH] = ACTIONS(2658),
    [anon_sym_PLUS] = ACTIONS(2658),
    [anon_sym_STAR] = ACTIONS(2660),
    [anon_sym_CARET] = ACTIONS(2660),
    [anon_sym_AMP] = ACTIONS(2660),
    [anon_sym_SEMI] = ACTIONS(2660),
    [anon_sym___extension__] = ACTIONS(2658),
    [anon_sym_typedef] = ACTIONS(2658),
    [anon_sym_extern] = ACTIONS(2658),
    [anon_sym___attribute__] = ACTIONS(2658),
    [anon_sym___attribute] = ACTIONS(2658),
    [anon_sym_noreturn] = ACTIONS(2658),
    [anon_sym_LBRACK] = ACTIONS(2660),
    [anon_sym___declspec] = ACTIONS(2658),
    [anon_sym___cdecl] = ACTIONS(2658),
    [anon_sym___clrcall] = ACTIONS(2658),
    [anon_sym___stdcall] = ACTIONS(2658),
    [anon_sym___fastcall] = ACTIONS(2658),
    [anon_sym___thiscall] = ACTIONS(2658),
    [anon_sym___vectorcall] = ACTIONS(2658),
    [anon_sym_LBRACE] = ACTIONS(2660),
    [anon_sym_signed] = ACTIONS(2658),
    [anon_sym_unsigned] = ACTIONS(2658),
    [anon_sym_long] = ACTIONS(2658),
    [anon_sym_short] = ACTIONS(2658),
    [anon_sym_static] = ACTIONS(2658),
    [anon_sym_ATautoreleasepool] = ACTIONS(2660),
    [anon_sym_auto] = ACTIONS(2658),
    [anon_sym_register] = ACTIONS(2658),
    [anon_sym_inline] = ACTIONS(2658),
    [anon_sym___inline] = ACTIONS(2658),
    [anon_sym___inline__] = ACTIONS(2658),
    [anon_sym___forceinline] = ACTIONS(2658),
    [anon_sym_thread_local] = ACTIONS(2658),
    [anon_sym___thread] = ACTIONS(2658),
    [anon_sym_CG_EXTERN] = ACTIONS(2658),
    [anon_sym_CG_INLINE] = ACTIONS(2658),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2658),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2658),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2658),
    [anon_sym_IBOutlet] = ACTIONS(2658),
    [anon_sym_IBInspectable] = ACTIONS(2658),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2658),
    [anon_sym_NS_INLINE] = ACTIONS(2658),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2658),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2658),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2658),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2658),
    [anon_sym_const] = ACTIONS(2658),
    [anon_sym_constexpr] = ACTIONS(2658),
    [anon_sym_volatile] = ACTIONS(2658),
    [anon_sym_restrict] = ACTIONS(2658),
    [anon_sym___restrict__] = ACTIONS(2658),
    [anon_sym__Atomic] = ACTIONS(2658),
    [anon_sym__Noreturn] = ACTIONS(2658),
    [anon_sym__Nonnull] = ACTIONS(2658),
    [anon_sym_nullable] = ACTIONS(2658),
    [anon_sym__Complex] = ACTIONS(2658),
    [anon_sym__Nullable] = ACTIONS(2658),
    [anon_sym__Nullable_result] = ACTIONS(2658),
    [anon_sym__Null_unspecified] = ACTIONS(2658),
    [anon_sym___autoreleasing] = ACTIONS(2658),
    [anon_sym___block] = ACTIONS(2658),
    [anon_sym___bridge] = ACTIONS(2658),
    [anon_sym___bridge_retained] = ACTIONS(2658),
    [anon_sym___bridge_transfer] = ACTIONS(2658),
    [anon_sym___complex] = ACTIONS(2658),
    [anon_sym___const] = ACTIONS(2658),
    [anon_sym___imag] = ACTIONS(2658),
    [anon_sym___kindof] = ACTIONS(2658),
    [anon_sym___nonnull] = ACTIONS(2658),
    [anon_sym___nullable] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2658),
    [anon_sym___real] = ACTIONS(2658),
    [anon_sym___strong] = ACTIONS(2658),
    [anon_sym___unsafe_unretained] = ACTIONS(2658),
    [anon_sym___unused] = ACTIONS(2658),
    [anon_sym___weak] = ACTIONS(2658),
    [anon_sym_alignas] = ACTIONS(2658),
    [anon_sym__Alignas] = ACTIONS(2658),
    [sym_primitive_type] = ACTIONS(2658),
    [anon_sym_enum] = ACTIONS(2658),
    [anon_sym_struct] = ACTIONS(2658),
    [anon_sym_union] = ACTIONS(2658),
    [anon_sym_if] = ACTIONS(2658),
    [anon_sym_switch] = ACTIONS(2658),
    [anon_sym_case] = ACTIONS(2658),
    [anon_sym_default] = ACTIONS(2658),
    [anon_sym_while] = ACTIONS(2658),
    [anon_sym_do] = ACTIONS(2658),
    [anon_sym_for] = ACTIONS(2658),
    [anon_sym_in] = ACTIONS(2658),
    [anon_sym_return] = ACTIONS(2658),
    [anon_sym_break] = ACTIONS(2658),
    [anon_sym_continue] = ACTIONS(2658),
    [anon_sym_goto] = ACTIONS(2658),
    [anon_sym_DASH_DASH] = ACTIONS(2660),
    [anon_sym_PLUS_PLUS] = ACTIONS(2660),
    [anon_sym_sizeof] = ACTIONS(2658),
    [anon_sym___alignof__] = ACTIONS(2658),
    [anon_sym___alignof] = ACTIONS(2658),
    [anon_sym__alignof] = ACTIONS(2658),
    [anon_sym_alignof] = ACTIONS(2658),
    [anon_sym__Alignof] = ACTIONS(2658),
    [anon_sym_offsetof] = ACTIONS(2658),
    [anon_sym__Generic] = ACTIONS(2658),
    [anon_sym_asm] = ACTIONS(2658),
    [anon_sym___asm__] = ACTIONS(2658),
    [anon_sym___asm] = ACTIONS(2658),
    [sym_number_literal] = ACTIONS(2660),
    [anon_sym_L_SQUOTE] = ACTIONS(2660),
    [anon_sym_u_SQUOTE] = ACTIONS(2660),
    [anon_sym_U_SQUOTE] = ACTIONS(2660),
    [anon_sym_u8_SQUOTE] = ACTIONS(2660),
    [anon_sym_SQUOTE] = ACTIONS(2660),
    [anon_sym_AT] = ACTIONS(2658),
    [anon_sym_DQUOTE] = ACTIONS(2660),
    [anon_sym_L_DQUOTE] = ACTIONS(2660),
    [anon_sym_u_DQUOTE] = ACTIONS(2660),
    [anon_sym_U_DQUOTE] = ACTIONS(2660),
    [anon_sym_u8_DQUOTE] = ACTIONS(2660),
    [sym_true] = ACTIONS(2658),
    [sym_false] = ACTIONS(2658),
    [anon_sym_NULL] = ACTIONS(2658),
    [anon_sym_nullptr] = ACTIONS(2658),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2658),
    [anon_sym___typeof] = ACTIONS(2658),
    [anon_sym_typeof] = ACTIONS(2658),
    [anon_sym_ATimport] = ACTIONS(2660),
    [aux_sym_preproc_undef_token1] = ACTIONS(2658),
    [anon_sym_POUND] = ACTIONS(2658),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2658),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2658),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2658),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2658),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2658),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2658),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2658),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_API_AVAILABLE] = ACTIONS(2658),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_API_DEPRECATED] = ACTIONS(2658),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2658),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym___deprecated_msg] = ACTIONS(2658),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2658),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2658),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2658),
    [anon_sym_ATprotocol] = ACTIONS(2660),
    [anon_sym_ATinterface] = ACTIONS(2660),
    [anon_sym_ATimplementation] = ACTIONS(2660),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2660),
    [anon_sym_ATsynthesize] = ACTIONS(2660),
    [anon_sym_ATdynamic] = ACTIONS(2660),
    [anon_sym_ATtry] = ACTIONS(2660),
    [anon_sym___try] = ACTIONS(2658),
    [anon_sym_ATthrow] = ACTIONS(2660),
    [anon_sym_ATselector] = ACTIONS(2660),
    [anon_sym_ATavailable] = ACTIONS(2660),
    [anon_sym___builtin_available] = ACTIONS(2658),
    [anon_sym_va_arg] = ACTIONS(2658),
    [anon_sym_ATencode] = ACTIONS(2660),
    [anon_sym_ATsynchronized] = ACTIONS(2660),
    [anon_sym_BOOL] = ACTIONS(2658),
    [anon_sym_IMP] = ACTIONS(2658),
    [anon_sym_SEL] = ACTIONS(2658),
    [anon_sym_Class] = ACTIONS(2658),
    [anon_sym_id] = ACTIONS(2658),
  },
  [STATE(300)] = {
    [sym_identifier] = ACTIONS(2662),
    [aux_sym_preproc_include_token1] = ACTIONS(2662),
    [aux_sym_preproc_include_token2] = ACTIONS(2662),
    [aux_sym_preproc_def_token1] = ACTIONS(2662),
    [aux_sym_preproc_if_token1] = ACTIONS(2662),
    [aux_sym_preproc_if_token2] = ACTIONS(2662),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2662),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2662),
    [aux_sym_preproc_else_token1] = ACTIONS(2662),
    [aux_sym_preproc_elif_token1] = ACTIONS(2662),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2662),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2662),
    [sym_preproc_directive] = ACTIONS(2662),
    [anon_sym_LPAREN2] = ACTIONS(2665),
    [anon_sym_BANG] = ACTIONS(2665),
    [anon_sym_TILDE] = ACTIONS(2665),
    [anon_sym_DASH] = ACTIONS(2662),
    [anon_sym_PLUS] = ACTIONS(2662),
    [anon_sym_STAR] = ACTIONS(2665),
    [anon_sym_CARET] = ACTIONS(2665),
    [anon_sym_AMP] = ACTIONS(2665),
    [anon_sym_SEMI] = ACTIONS(2665),
    [anon_sym___extension__] = ACTIONS(2662),
    [anon_sym_typedef] = ACTIONS(2662),
    [anon_sym_extern] = ACTIONS(2662),
    [anon_sym___attribute__] = ACTIONS(2662),
    [anon_sym___attribute] = ACTIONS(2662),
    [anon_sym_noreturn] = ACTIONS(2662),
    [anon_sym_LBRACK] = ACTIONS(2665),
    [anon_sym___declspec] = ACTIONS(2662),
    [anon_sym___cdecl] = ACTIONS(2662),
    [anon_sym___clrcall] = ACTIONS(2662),
    [anon_sym___stdcall] = ACTIONS(2662),
    [anon_sym___fastcall] = ACTIONS(2662),
    [anon_sym___thiscall] = ACTIONS(2662),
    [anon_sym___vectorcall] = ACTIONS(2662),
    [anon_sym_LBRACE] = ACTIONS(2665),
    [anon_sym_signed] = ACTIONS(2662),
    [anon_sym_unsigned] = ACTIONS(2662),
    [anon_sym_long] = ACTIONS(2662),
    [anon_sym_short] = ACTIONS(2662),
    [anon_sym_static] = ACTIONS(2662),
    [anon_sym_ATautoreleasepool] = ACTIONS(2665),
    [anon_sym_auto] = ACTIONS(2662),
    [anon_sym_register] = ACTIONS(2662),
    [anon_sym_inline] = ACTIONS(2662),
    [anon_sym___inline] = ACTIONS(2662),
    [anon_sym___inline__] = ACTIONS(2662),
    [anon_sym___forceinline] = ACTIONS(2662),
    [anon_sym_thread_local] = ACTIONS(2662),
    [anon_sym___thread] = ACTIONS(2662),
    [anon_sym_CG_EXTERN] = ACTIONS(2662),
    [anon_sym_CG_INLINE] = ACTIONS(2662),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2662),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2662),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2662),
    [anon_sym_IBOutlet] = ACTIONS(2662),
    [anon_sym_IBInspectable] = ACTIONS(2662),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2662),
    [anon_sym_NS_INLINE] = ACTIONS(2662),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2662),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2662),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2662),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2662),
    [anon_sym_const] = ACTIONS(2662),
    [anon_sym_constexpr] = ACTIONS(2662),
    [anon_sym_volatile] = ACTIONS(2662),
    [anon_sym_restrict] = ACTIONS(2662),
    [anon_sym___restrict__] = ACTIONS(2662),
    [anon_sym__Atomic] = ACTIONS(2662),
    [anon_sym__Noreturn] = ACTIONS(2662),
    [anon_sym__Nonnull] = ACTIONS(2662),
    [anon_sym_nullable] = ACTIONS(2662),
    [anon_sym__Complex] = ACTIONS(2662),
    [anon_sym__Nullable] = ACTIONS(2662),
    [anon_sym__Nullable_result] = ACTIONS(2662),
    [anon_sym__Null_unspecified] = ACTIONS(2662),
    [anon_sym___autoreleasing] = ACTIONS(2662),
    [anon_sym___block] = ACTIONS(2662),
    [anon_sym___bridge] = ACTIONS(2662),
    [anon_sym___bridge_retained] = ACTIONS(2662),
    [anon_sym___bridge_transfer] = ACTIONS(2662),
    [anon_sym___complex] = ACTIONS(2662),
    [anon_sym___const] = ACTIONS(2662),
    [anon_sym___imag] = ACTIONS(2662),
    [anon_sym___kindof] = ACTIONS(2662),
    [anon_sym___nonnull] = ACTIONS(2662),
    [anon_sym___nullable] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2662),
    [anon_sym___real] = ACTIONS(2662),
    [anon_sym___strong] = ACTIONS(2662),
    [anon_sym___unsafe_unretained] = ACTIONS(2662),
    [anon_sym___unused] = ACTIONS(2662),
    [anon_sym___weak] = ACTIONS(2662),
    [anon_sym_alignas] = ACTIONS(2662),
    [anon_sym__Alignas] = ACTIONS(2662),
    [sym_primitive_type] = ACTIONS(2662),
    [anon_sym_enum] = ACTIONS(2662),
    [anon_sym_struct] = ACTIONS(2662),
    [anon_sym_union] = ACTIONS(2662),
    [anon_sym_if] = ACTIONS(2662),
    [anon_sym_switch] = ACTIONS(2662),
    [anon_sym_case] = ACTIONS(2662),
    [anon_sym_default] = ACTIONS(2662),
    [anon_sym_while] = ACTIONS(2662),
    [anon_sym_do] = ACTIONS(2662),
    [anon_sym_for] = ACTIONS(2662),
    [anon_sym_in] = ACTIONS(2662),
    [anon_sym_return] = ACTIONS(2662),
    [anon_sym_break] = ACTIONS(2662),
    [anon_sym_continue] = ACTIONS(2662),
    [anon_sym_goto] = ACTIONS(2662),
    [anon_sym_DASH_DASH] = ACTIONS(2665),
    [anon_sym_PLUS_PLUS] = ACTIONS(2665),
    [anon_sym_sizeof] = ACTIONS(2662),
    [anon_sym___alignof__] = ACTIONS(2662),
    [anon_sym___alignof] = ACTIONS(2662),
    [anon_sym__alignof] = ACTIONS(2662),
    [anon_sym_alignof] = ACTIONS(2662),
    [anon_sym__Alignof] = ACTIONS(2662),
    [anon_sym_offsetof] = ACTIONS(2662),
    [anon_sym__Generic] = ACTIONS(2662),
    [anon_sym_asm] = ACTIONS(2662),
    [anon_sym___asm__] = ACTIONS(2662),
    [anon_sym___asm] = ACTIONS(2662),
    [sym_number_literal] = ACTIONS(2665),
    [anon_sym_L_SQUOTE] = ACTIONS(2665),
    [anon_sym_u_SQUOTE] = ACTIONS(2665),
    [anon_sym_U_SQUOTE] = ACTIONS(2665),
    [anon_sym_u8_SQUOTE] = ACTIONS(2665),
    [anon_sym_SQUOTE] = ACTIONS(2665),
    [anon_sym_AT] = ACTIONS(2662),
    [anon_sym_DQUOTE] = ACTIONS(2665),
    [anon_sym_L_DQUOTE] = ACTIONS(2665),
    [anon_sym_u_DQUOTE] = ACTIONS(2665),
    [anon_sym_U_DQUOTE] = ACTIONS(2665),
    [anon_sym_u8_DQUOTE] = ACTIONS(2665),
    [sym_true] = ACTIONS(2662),
    [sym_false] = ACTIONS(2662),
    [anon_sym_NULL] = ACTIONS(2662),
    [anon_sym_nullptr] = ACTIONS(2662),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2662),
    [anon_sym___typeof] = ACTIONS(2662),
    [anon_sym_typeof] = ACTIONS(2662),
    [anon_sym_ATimport] = ACTIONS(2665),
    [aux_sym_preproc_undef_token1] = ACTIONS(2662),
    [anon_sym_POUND] = ACTIONS(2662),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2662),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2662),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2662),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2662),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2662),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2662),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2662),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_API_AVAILABLE] = ACTIONS(2662),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_API_DEPRECATED] = ACTIONS(2662),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2662),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym___deprecated_msg] = ACTIONS(2662),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2662),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2662),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2662),
    [anon_sym_ATprotocol] = ACTIONS(2665),
    [anon_sym_ATinterface] = ACTIONS(2665),
    [anon_sym_ATimplementation] = ACTIONS(2665),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2665),
    [anon_sym_ATsynthesize] = ACTIONS(2665),
    [anon_sym_ATdynamic] = ACTIONS(2665),
    [anon_sym_ATtry] = ACTIONS(2665),
    [anon_sym___try] = ACTIONS(2662),
    [anon_sym_ATthrow] = ACTIONS(2665),
    [anon_sym_ATselector] = ACTIONS(2665),
    [anon_sym_ATavailable] = ACTIONS(2665),
    [anon_sym___builtin_available] = ACTIONS(2662),
    [anon_sym_va_arg] = ACTIONS(2662),
    [anon_sym_ATencode] = ACTIONS(2665),
    [anon_sym_ATsynchronized] = ACTIONS(2665),
    [anon_sym_BOOL] = ACTIONS(2662),
    [anon_sym_IMP] = ACTIONS(2662),
    [anon_sym_SEL] = ACTIONS(2662),
    [anon_sym_Class] = ACTIONS(2662),
    [anon_sym_id] = ACTIONS(2662),
  },
  [STATE(301)] = {
    [sym_identifier] = ACTIONS(2668),
    [aux_sym_preproc_include_token1] = ACTIONS(2668),
    [aux_sym_preproc_include_token2] = ACTIONS(2668),
    [aux_sym_preproc_def_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token2] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2668),
    [aux_sym_preproc_else_token1] = ACTIONS(2668),
    [aux_sym_preproc_elif_token1] = ACTIONS(2668),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2668),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2668),
    [sym_preproc_directive] = ACTIONS(2668),
    [anon_sym_LPAREN2] = ACTIONS(2670),
    [anon_sym_BANG] = ACTIONS(2670),
    [anon_sym_TILDE] = ACTIONS(2670),
    [anon_sym_DASH] = ACTIONS(2668),
    [anon_sym_PLUS] = ACTIONS(2668),
    [anon_sym_STAR] = ACTIONS(2670),
    [anon_sym_CARET] = ACTIONS(2670),
    [anon_sym_AMP] = ACTIONS(2670),
    [anon_sym_SEMI] = ACTIONS(2670),
    [anon_sym___extension__] = ACTIONS(2668),
    [anon_sym_typedef] = ACTIONS(2668),
    [anon_sym_extern] = ACTIONS(2668),
    [anon_sym___attribute__] = ACTIONS(2668),
    [anon_sym___attribute] = ACTIONS(2668),
    [anon_sym_noreturn] = ACTIONS(2668),
    [anon_sym_LBRACK] = ACTIONS(2670),
    [anon_sym___declspec] = ACTIONS(2668),
    [anon_sym___cdecl] = ACTIONS(2668),
    [anon_sym___clrcall] = ACTIONS(2668),
    [anon_sym___stdcall] = ACTIONS(2668),
    [anon_sym___fastcall] = ACTIONS(2668),
    [anon_sym___thiscall] = ACTIONS(2668),
    [anon_sym___vectorcall] = ACTIONS(2668),
    [anon_sym_LBRACE] = ACTIONS(2670),
    [anon_sym_signed] = ACTIONS(2668),
    [anon_sym_unsigned] = ACTIONS(2668),
    [anon_sym_long] = ACTIONS(2668),
    [anon_sym_short] = ACTIONS(2668),
    [anon_sym_static] = ACTIONS(2668),
    [anon_sym_ATautoreleasepool] = ACTIONS(2670),
    [anon_sym_auto] = ACTIONS(2668),
    [anon_sym_register] = ACTIONS(2668),
    [anon_sym_inline] = ACTIONS(2668),
    [anon_sym___inline] = ACTIONS(2668),
    [anon_sym___inline__] = ACTIONS(2668),
    [anon_sym___forceinline] = ACTIONS(2668),
    [anon_sym_thread_local] = ACTIONS(2668),
    [anon_sym___thread] = ACTIONS(2668),
    [anon_sym_CG_EXTERN] = ACTIONS(2668),
    [anon_sym_CG_INLINE] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2668),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2668),
    [anon_sym_IBOutlet] = ACTIONS(2668),
    [anon_sym_IBInspectable] = ACTIONS(2668),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2668),
    [anon_sym_NS_INLINE] = ACTIONS(2668),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2668),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2668),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2668),
    [anon_sym_const] = ACTIONS(2668),
    [anon_sym_constexpr] = ACTIONS(2668),
    [anon_sym_volatile] = ACTIONS(2668),
    [anon_sym_restrict] = ACTIONS(2668),
    [anon_sym___restrict__] = ACTIONS(2668),
    [anon_sym__Atomic] = ACTIONS(2668),
    [anon_sym__Noreturn] = ACTIONS(2668),
    [anon_sym__Nonnull] = ACTIONS(2668),
    [anon_sym_nullable] = ACTIONS(2668),
    [anon_sym__Complex] = ACTIONS(2668),
    [anon_sym__Nullable] = ACTIONS(2668),
    [anon_sym__Nullable_result] = ACTIONS(2668),
    [anon_sym__Null_unspecified] = ACTIONS(2668),
    [anon_sym___autoreleasing] = ACTIONS(2668),
    [anon_sym___block] = ACTIONS(2668),
    [anon_sym___bridge] = ACTIONS(2668),
    [anon_sym___bridge_retained] = ACTIONS(2668),
    [anon_sym___bridge_transfer] = ACTIONS(2668),
    [anon_sym___complex] = ACTIONS(2668),
    [anon_sym___const] = ACTIONS(2668),
    [anon_sym___imag] = ACTIONS(2668),
    [anon_sym___kindof] = ACTIONS(2668),
    [anon_sym___nonnull] = ACTIONS(2668),
    [anon_sym___nullable] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2668),
    [anon_sym___real] = ACTIONS(2668),
    [anon_sym___strong] = ACTIONS(2668),
    [anon_sym___unsafe_unretained] = ACTIONS(2668),
    [anon_sym___unused] = ACTIONS(2668),
    [anon_sym___weak] = ACTIONS(2668),
    [anon_sym_alignas] = ACTIONS(2668),
    [anon_sym__Alignas] = ACTIONS(2668),
    [sym_primitive_type] = ACTIONS(2668),
    [anon_sym_enum] = ACTIONS(2668),
    [anon_sym_struct] = ACTIONS(2668),
    [anon_sym_union] = ACTIONS(2668),
    [anon_sym_if] = ACTIONS(2668),
    [anon_sym_switch] = ACTIONS(2668),
    [anon_sym_case] = ACTIONS(2668),
    [anon_sym_default] = ACTIONS(2668),
    [anon_sym_while] = ACTIONS(2668),
    [anon_sym_do] = ACTIONS(2668),
    [anon_sym_for] = ACTIONS(2668),
    [anon_sym_in] = ACTIONS(2668),
    [anon_sym_return] = ACTIONS(2668),
    [anon_sym_break] = ACTIONS(2668),
    [anon_sym_continue] = ACTIONS(2668),
    [anon_sym_goto] = ACTIONS(2668),
    [anon_sym_DASH_DASH] = ACTIONS(2670),
    [anon_sym_PLUS_PLUS] = ACTIONS(2670),
    [anon_sym_sizeof] = ACTIONS(2668),
    [anon_sym___alignof__] = ACTIONS(2668),
    [anon_sym___alignof] = ACTIONS(2668),
    [anon_sym__alignof] = ACTIONS(2668),
    [anon_sym_alignof] = ACTIONS(2668),
    [anon_sym__Alignof] = ACTIONS(2668),
    [anon_sym_offsetof] = ACTIONS(2668),
    [anon_sym__Generic] = ACTIONS(2668),
    [anon_sym_asm] = ACTIONS(2668),
    [anon_sym___asm__] = ACTIONS(2668),
    [anon_sym___asm] = ACTIONS(2668),
    [sym_number_literal] = ACTIONS(2670),
    [anon_sym_L_SQUOTE] = ACTIONS(2670),
    [anon_sym_u_SQUOTE] = ACTIONS(2670),
    [anon_sym_U_SQUOTE] = ACTIONS(2670),
    [anon_sym_u8_SQUOTE] = ACTIONS(2670),
    [anon_sym_SQUOTE] = ACTIONS(2670),
    [anon_sym_AT] = ACTIONS(2668),
    [anon_sym_DQUOTE] = ACTIONS(2670),
    [anon_sym_L_DQUOTE] = ACTIONS(2670),
    [anon_sym_u_DQUOTE] = ACTIONS(2670),
    [anon_sym_U_DQUOTE] = ACTIONS(2670),
    [anon_sym_u8_DQUOTE] = ACTIONS(2670),
    [sym_true] = ACTIONS(2668),
    [sym_false] = ACTIONS(2668),
    [anon_sym_NULL] = ACTIONS(2668),
    [anon_sym_nullptr] = ACTIONS(2668),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2668),
    [anon_sym___typeof] = ACTIONS(2668),
    [anon_sym_typeof] = ACTIONS(2668),
    [anon_sym_ATimport] = ACTIONS(2670),
    [aux_sym_preproc_undef_token1] = ACTIONS(2668),
    [anon_sym_POUND] = ACTIONS(2668),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2668),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2668),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2668),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_API_AVAILABLE] = ACTIONS(2668),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_API_DEPRECATED] = ACTIONS(2668),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym___deprecated_msg] = ACTIONS(2668),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2668),
    [anon_sym_ATprotocol] = ACTIONS(2670),
    [anon_sym_ATinterface] = ACTIONS(2670),
    [anon_sym_ATimplementation] = ACTIONS(2670),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2670),
    [anon_sym_ATsynthesize] = ACTIONS(2670),
    [anon_sym_ATdynamic] = ACTIONS(2670),
    [anon_sym_ATtry] = ACTIONS(2670),
    [anon_sym___try] = ACTIONS(2668),
    [anon_sym_ATthrow] = ACTIONS(2670),
    [anon_sym_ATselector] = ACTIONS(2670),
    [anon_sym_ATavailable] = ACTIONS(2670),
    [anon_sym___builtin_available] = ACTIONS(2668),
    [anon_sym_va_arg] = ACTIONS(2668),
    [anon_sym_ATencode] = ACTIONS(2670),
    [anon_sym_ATsynchronized] = ACTIONS(2670),
    [anon_sym_BOOL] = ACTIONS(2668),
    [anon_sym_IMP] = ACTIONS(2668),
    [anon_sym_SEL] = ACTIONS(2668),
    [anon_sym_Class] = ACTIONS(2668),
    [anon_sym_id] = ACTIONS(2668),
  },
  [STATE(302)] = {
    [sym_identifier] = ACTIONS(2672),
    [aux_sym_preproc_include_token1] = ACTIONS(2672),
    [aux_sym_preproc_include_token2] = ACTIONS(2672),
    [aux_sym_preproc_def_token1] = ACTIONS(2672),
    [aux_sym_preproc_if_token1] = ACTIONS(2672),
    [aux_sym_preproc_if_token2] = ACTIONS(2672),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2672),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2672),
    [aux_sym_preproc_else_token1] = ACTIONS(2672),
    [aux_sym_preproc_elif_token1] = ACTIONS(2672),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2672),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2672),
    [sym_preproc_directive] = ACTIONS(2672),
    [anon_sym_LPAREN2] = ACTIONS(2674),
    [anon_sym_BANG] = ACTIONS(2674),
    [anon_sym_TILDE] = ACTIONS(2674),
    [anon_sym_DASH] = ACTIONS(2672),
    [anon_sym_PLUS] = ACTIONS(2672),
    [anon_sym_STAR] = ACTIONS(2674),
    [anon_sym_CARET] = ACTIONS(2674),
    [anon_sym_AMP] = ACTIONS(2674),
    [anon_sym_SEMI] = ACTIONS(2674),
    [anon_sym___extension__] = ACTIONS(2672),
    [anon_sym_typedef] = ACTIONS(2672),
    [anon_sym_extern] = ACTIONS(2672),
    [anon_sym___attribute__] = ACTIONS(2672),
    [anon_sym___attribute] = ACTIONS(2672),
    [anon_sym_noreturn] = ACTIONS(2672),
    [anon_sym_LBRACK] = ACTIONS(2674),
    [anon_sym___declspec] = ACTIONS(2672),
    [anon_sym___cdecl] = ACTIONS(2672),
    [anon_sym___clrcall] = ACTIONS(2672),
    [anon_sym___stdcall] = ACTIONS(2672),
    [anon_sym___fastcall] = ACTIONS(2672),
    [anon_sym___thiscall] = ACTIONS(2672),
    [anon_sym___vectorcall] = ACTIONS(2672),
    [anon_sym_LBRACE] = ACTIONS(2674),
    [anon_sym_signed] = ACTIONS(2672),
    [anon_sym_unsigned] = ACTIONS(2672),
    [anon_sym_long] = ACTIONS(2672),
    [anon_sym_short] = ACTIONS(2672),
    [anon_sym_static] = ACTIONS(2672),
    [anon_sym_ATautoreleasepool] = ACTIONS(2674),
    [anon_sym_auto] = ACTIONS(2672),
    [anon_sym_register] = ACTIONS(2672),
    [anon_sym_inline] = ACTIONS(2672),
    [anon_sym___inline] = ACTIONS(2672),
    [anon_sym___inline__] = ACTIONS(2672),
    [anon_sym___forceinline] = ACTIONS(2672),
    [anon_sym_thread_local] = ACTIONS(2672),
    [anon_sym___thread] = ACTIONS(2672),
    [anon_sym_CG_EXTERN] = ACTIONS(2672),
    [anon_sym_CG_INLINE] = ACTIONS(2672),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2672),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2672),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2672),
    [anon_sym_IBOutlet] = ACTIONS(2672),
    [anon_sym_IBInspectable] = ACTIONS(2672),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2672),
    [anon_sym_NS_INLINE] = ACTIONS(2672),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2672),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2672),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2672),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2672),
    [anon_sym_const] = ACTIONS(2672),
    [anon_sym_constexpr] = ACTIONS(2672),
    [anon_sym_volatile] = ACTIONS(2672),
    [anon_sym_restrict] = ACTIONS(2672),
    [anon_sym___restrict__] = ACTIONS(2672),
    [anon_sym__Atomic] = ACTIONS(2672),
    [anon_sym__Noreturn] = ACTIONS(2672),
    [anon_sym__Nonnull] = ACTIONS(2672),
    [anon_sym_nullable] = ACTIONS(2672),
    [anon_sym__Complex] = ACTIONS(2672),
    [anon_sym__Nullable] = ACTIONS(2672),
    [anon_sym__Nullable_result] = ACTIONS(2672),
    [anon_sym__Null_unspecified] = ACTIONS(2672),
    [anon_sym___autoreleasing] = ACTIONS(2672),
    [anon_sym___block] = ACTIONS(2672),
    [anon_sym___bridge] = ACTIONS(2672),
    [anon_sym___bridge_retained] = ACTIONS(2672),
    [anon_sym___bridge_transfer] = ACTIONS(2672),
    [anon_sym___complex] = ACTIONS(2672),
    [anon_sym___const] = ACTIONS(2672),
    [anon_sym___imag] = ACTIONS(2672),
    [anon_sym___kindof] = ACTIONS(2672),
    [anon_sym___nonnull] = ACTIONS(2672),
    [anon_sym___nullable] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2672),
    [anon_sym___real] = ACTIONS(2672),
    [anon_sym___strong] = ACTIONS(2672),
    [anon_sym___unsafe_unretained] = ACTIONS(2672),
    [anon_sym___unused] = ACTIONS(2672),
    [anon_sym___weak] = ACTIONS(2672),
    [anon_sym_alignas] = ACTIONS(2672),
    [anon_sym__Alignas] = ACTIONS(2672),
    [sym_primitive_type] = ACTIONS(2672),
    [anon_sym_enum] = ACTIONS(2672),
    [anon_sym_struct] = ACTIONS(2672),
    [anon_sym_union] = ACTIONS(2672),
    [anon_sym_if] = ACTIONS(2672),
    [anon_sym_switch] = ACTIONS(2672),
    [anon_sym_case] = ACTIONS(2672),
    [anon_sym_default] = ACTIONS(2672),
    [anon_sym_while] = ACTIONS(2672),
    [anon_sym_do] = ACTIONS(2672),
    [anon_sym_for] = ACTIONS(2672),
    [anon_sym_in] = ACTIONS(2672),
    [anon_sym_return] = ACTIONS(2672),
    [anon_sym_break] = ACTIONS(2672),
    [anon_sym_continue] = ACTIONS(2672),
    [anon_sym_goto] = ACTIONS(2672),
    [anon_sym_DASH_DASH] = ACTIONS(2674),
    [anon_sym_PLUS_PLUS] = ACTIONS(2674),
    [anon_sym_sizeof] = ACTIONS(2672),
    [anon_sym___alignof__] = ACTIONS(2672),
    [anon_sym___alignof] = ACTIONS(2672),
    [anon_sym__alignof] = ACTIONS(2672),
    [anon_sym_alignof] = ACTIONS(2672),
    [anon_sym__Alignof] = ACTIONS(2672),
    [anon_sym_offsetof] = ACTIONS(2672),
    [anon_sym__Generic] = ACTIONS(2672),
    [anon_sym_asm] = ACTIONS(2672),
    [anon_sym___asm__] = ACTIONS(2672),
    [anon_sym___asm] = ACTIONS(2672),
    [sym_number_literal] = ACTIONS(2674),
    [anon_sym_L_SQUOTE] = ACTIONS(2674),
    [anon_sym_u_SQUOTE] = ACTIONS(2674),
    [anon_sym_U_SQUOTE] = ACTIONS(2674),
    [anon_sym_u8_SQUOTE] = ACTIONS(2674),
    [anon_sym_SQUOTE] = ACTIONS(2674),
    [anon_sym_AT] = ACTIONS(2672),
    [anon_sym_DQUOTE] = ACTIONS(2674),
    [anon_sym_L_DQUOTE] = ACTIONS(2674),
    [anon_sym_u_DQUOTE] = ACTIONS(2674),
    [anon_sym_U_DQUOTE] = ACTIONS(2674),
    [anon_sym_u8_DQUOTE] = ACTIONS(2674),
    [sym_true] = ACTIONS(2672),
    [sym_false] = ACTIONS(2672),
    [anon_sym_NULL] = ACTIONS(2672),
    [anon_sym_nullptr] = ACTIONS(2672),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2672),
    [anon_sym___typeof] = ACTIONS(2672),
    [anon_sym_typeof] = ACTIONS(2672),
    [anon_sym_ATimport] = ACTIONS(2674),
    [aux_sym_preproc_undef_token1] = ACTIONS(2672),
    [anon_sym_POUND] = ACTIONS(2672),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2672),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2672),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2672),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2672),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2672),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2672),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2672),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_API_AVAILABLE] = ACTIONS(2672),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_API_DEPRECATED] = ACTIONS(2672),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2672),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym___deprecated_msg] = ACTIONS(2672),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2672),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2672),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2672),
    [anon_sym_ATprotocol] = ACTIONS(2674),
    [anon_sym_ATinterface] = ACTIONS(2674),
    [anon_sym_ATimplementation] = ACTIONS(2674),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2674),
    [anon_sym_ATsynthesize] = ACTIONS(2674),
    [anon_sym_ATdynamic] = ACTIONS(2674),
    [anon_sym_ATtry] = ACTIONS(2674),
    [anon_sym___try] = ACTIONS(2672),
    [anon_sym_ATthrow] = ACTIONS(2674),
    [anon_sym_ATselector] = ACTIONS(2674),
    [anon_sym_ATavailable] = ACTIONS(2674),
    [anon_sym___builtin_available] = ACTIONS(2672),
    [anon_sym_va_arg] = ACTIONS(2672),
    [anon_sym_ATencode] = ACTIONS(2674),
    [anon_sym_ATsynchronized] = ACTIONS(2674),
    [anon_sym_BOOL] = ACTIONS(2672),
    [anon_sym_IMP] = ACTIONS(2672),
    [anon_sym_SEL] = ACTIONS(2672),
    [anon_sym_Class] = ACTIONS(2672),
    [anon_sym_id] = ACTIONS(2672),
  },
  [STATE(303)] = {
    [sym_identifier] = ACTIONS(2676),
    [aux_sym_preproc_include_token1] = ACTIONS(2676),
    [aux_sym_preproc_include_token2] = ACTIONS(2676),
    [aux_sym_preproc_def_token1] = ACTIONS(2676),
    [aux_sym_preproc_if_token1] = ACTIONS(2676),
    [aux_sym_preproc_if_token2] = ACTIONS(2676),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2676),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2676),
    [aux_sym_preproc_else_token1] = ACTIONS(2676),
    [aux_sym_preproc_elif_token1] = ACTIONS(2676),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2676),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2676),
    [sym_preproc_directive] = ACTIONS(2676),
    [anon_sym_LPAREN2] = ACTIONS(2678),
    [anon_sym_BANG] = ACTIONS(2678),
    [anon_sym_TILDE] = ACTIONS(2678),
    [anon_sym_DASH] = ACTIONS(2676),
    [anon_sym_PLUS] = ACTIONS(2676),
    [anon_sym_STAR] = ACTIONS(2678),
    [anon_sym_CARET] = ACTIONS(2678),
    [anon_sym_AMP] = ACTIONS(2678),
    [anon_sym_SEMI] = ACTIONS(2678),
    [anon_sym___extension__] = ACTIONS(2676),
    [anon_sym_typedef] = ACTIONS(2676),
    [anon_sym_extern] = ACTIONS(2676),
    [anon_sym___attribute__] = ACTIONS(2676),
    [anon_sym___attribute] = ACTIONS(2676),
    [anon_sym_noreturn] = ACTIONS(2676),
    [anon_sym_LBRACK] = ACTIONS(2678),
    [anon_sym___declspec] = ACTIONS(2676),
    [anon_sym___cdecl] = ACTIONS(2676),
    [anon_sym___clrcall] = ACTIONS(2676),
    [anon_sym___stdcall] = ACTIONS(2676),
    [anon_sym___fastcall] = ACTIONS(2676),
    [anon_sym___thiscall] = ACTIONS(2676),
    [anon_sym___vectorcall] = ACTIONS(2676),
    [anon_sym_LBRACE] = ACTIONS(2678),
    [anon_sym_signed] = ACTIONS(2676),
    [anon_sym_unsigned] = ACTIONS(2676),
    [anon_sym_long] = ACTIONS(2676),
    [anon_sym_short] = ACTIONS(2676),
    [anon_sym_static] = ACTIONS(2676),
    [anon_sym_ATautoreleasepool] = ACTIONS(2678),
    [anon_sym_auto] = ACTIONS(2676),
    [anon_sym_register] = ACTIONS(2676),
    [anon_sym_inline] = ACTIONS(2676),
    [anon_sym___inline] = ACTIONS(2676),
    [anon_sym___inline__] = ACTIONS(2676),
    [anon_sym___forceinline] = ACTIONS(2676),
    [anon_sym_thread_local] = ACTIONS(2676),
    [anon_sym___thread] = ACTIONS(2676),
    [anon_sym_CG_EXTERN] = ACTIONS(2676),
    [anon_sym_CG_INLINE] = ACTIONS(2676),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2676),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2676),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2676),
    [anon_sym_IBOutlet] = ACTIONS(2676),
    [anon_sym_IBInspectable] = ACTIONS(2676),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2676),
    [anon_sym_NS_INLINE] = ACTIONS(2676),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2676),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2676),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2676),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2676),
    [anon_sym_const] = ACTIONS(2676),
    [anon_sym_constexpr] = ACTIONS(2676),
    [anon_sym_volatile] = ACTIONS(2676),
    [anon_sym_restrict] = ACTIONS(2676),
    [anon_sym___restrict__] = ACTIONS(2676),
    [anon_sym__Atomic] = ACTIONS(2676),
    [anon_sym__Noreturn] = ACTIONS(2676),
    [anon_sym__Nonnull] = ACTIONS(2676),
    [anon_sym_nullable] = ACTIONS(2676),
    [anon_sym__Complex] = ACTIONS(2676),
    [anon_sym__Nullable] = ACTIONS(2676),
    [anon_sym__Nullable_result] = ACTIONS(2676),
    [anon_sym__Null_unspecified] = ACTIONS(2676),
    [anon_sym___autoreleasing] = ACTIONS(2676),
    [anon_sym___block] = ACTIONS(2676),
    [anon_sym___bridge] = ACTIONS(2676),
    [anon_sym___bridge_retained] = ACTIONS(2676),
    [anon_sym___bridge_transfer] = ACTIONS(2676),
    [anon_sym___complex] = ACTIONS(2676),
    [anon_sym___const] = ACTIONS(2676),
    [anon_sym___imag] = ACTIONS(2676),
    [anon_sym___kindof] = ACTIONS(2676),
    [anon_sym___nonnull] = ACTIONS(2676),
    [anon_sym___nullable] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2676),
    [anon_sym___real] = ACTIONS(2676),
    [anon_sym___strong] = ACTIONS(2676),
    [anon_sym___unsafe_unretained] = ACTIONS(2676),
    [anon_sym___unused] = ACTIONS(2676),
    [anon_sym___weak] = ACTIONS(2676),
    [anon_sym_alignas] = ACTIONS(2676),
    [anon_sym__Alignas] = ACTIONS(2676),
    [sym_primitive_type] = ACTIONS(2676),
    [anon_sym_enum] = ACTIONS(2676),
    [anon_sym_struct] = ACTIONS(2676),
    [anon_sym_union] = ACTIONS(2676),
    [anon_sym_if] = ACTIONS(2676),
    [anon_sym_switch] = ACTIONS(2676),
    [anon_sym_case] = ACTIONS(2676),
    [anon_sym_default] = ACTIONS(2676),
    [anon_sym_while] = ACTIONS(2676),
    [anon_sym_do] = ACTIONS(2676),
    [anon_sym_for] = ACTIONS(2676),
    [anon_sym_in] = ACTIONS(2676),
    [anon_sym_return] = ACTIONS(2676),
    [anon_sym_break] = ACTIONS(2676),
    [anon_sym_continue] = ACTIONS(2676),
    [anon_sym_goto] = ACTIONS(2676),
    [anon_sym_DASH_DASH] = ACTIONS(2678),
    [anon_sym_PLUS_PLUS] = ACTIONS(2678),
    [anon_sym_sizeof] = ACTIONS(2676),
    [anon_sym___alignof__] = ACTIONS(2676),
    [anon_sym___alignof] = ACTIONS(2676),
    [anon_sym__alignof] = ACTIONS(2676),
    [anon_sym_alignof] = ACTIONS(2676),
    [anon_sym__Alignof] = ACTIONS(2676),
    [anon_sym_offsetof] = ACTIONS(2676),
    [anon_sym__Generic] = ACTIONS(2676),
    [anon_sym_asm] = ACTIONS(2676),
    [anon_sym___asm__] = ACTIONS(2676),
    [anon_sym___asm] = ACTIONS(2676),
    [sym_number_literal] = ACTIONS(2678),
    [anon_sym_L_SQUOTE] = ACTIONS(2678),
    [anon_sym_u_SQUOTE] = ACTIONS(2678),
    [anon_sym_U_SQUOTE] = ACTIONS(2678),
    [anon_sym_u8_SQUOTE] = ACTIONS(2678),
    [anon_sym_SQUOTE] = ACTIONS(2678),
    [anon_sym_AT] = ACTIONS(2676),
    [anon_sym_DQUOTE] = ACTIONS(2678),
    [anon_sym_L_DQUOTE] = ACTIONS(2678),
    [anon_sym_u_DQUOTE] = ACTIONS(2678),
    [anon_sym_U_DQUOTE] = ACTIONS(2678),
    [anon_sym_u8_DQUOTE] = ACTIONS(2678),
    [sym_true] = ACTIONS(2676),
    [sym_false] = ACTIONS(2676),
    [anon_sym_NULL] = ACTIONS(2676),
    [anon_sym_nullptr] = ACTIONS(2676),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2676),
    [anon_sym___typeof] = ACTIONS(2676),
    [anon_sym_typeof] = ACTIONS(2676),
    [anon_sym_ATimport] = ACTIONS(2678),
    [aux_sym_preproc_undef_token1] = ACTIONS(2676),
    [anon_sym_POUND] = ACTIONS(2676),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2676),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2676),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2676),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2676),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2676),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2676),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2676),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_API_AVAILABLE] = ACTIONS(2676),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_API_DEPRECATED] = ACTIONS(2676),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2676),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym___deprecated_msg] = ACTIONS(2676),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2676),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2676),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2676),
    [anon_sym_ATprotocol] = ACTIONS(2678),
    [anon_sym_ATinterface] = ACTIONS(2678),
    [anon_sym_ATimplementation] = ACTIONS(2678),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2678),
    [anon_sym_ATsynthesize] = ACTIONS(2678),
    [anon_sym_ATdynamic] = ACTIONS(2678),
    [anon_sym_ATtry] = ACTIONS(2678),
    [anon_sym___try] = ACTIONS(2676),
    [anon_sym_ATthrow] = ACTIONS(2678),
    [anon_sym_ATselector] = ACTIONS(2678),
    [anon_sym_ATavailable] = ACTIONS(2678),
    [anon_sym___builtin_available] = ACTIONS(2676),
    [anon_sym_va_arg] = ACTIONS(2676),
    [anon_sym_ATencode] = ACTIONS(2678),
    [anon_sym_ATsynchronized] = ACTIONS(2678),
    [anon_sym_BOOL] = ACTIONS(2676),
    [anon_sym_IMP] = ACTIONS(2676),
    [anon_sym_SEL] = ACTIONS(2676),
    [anon_sym_Class] = ACTIONS(2676),
    [anon_sym_id] = ACTIONS(2676),
  },
  [STATE(304)] = {
    [sym_identifier] = ACTIONS(2680),
    [aux_sym_preproc_include_token1] = ACTIONS(2680),
    [aux_sym_preproc_include_token2] = ACTIONS(2680),
    [aux_sym_preproc_def_token1] = ACTIONS(2680),
    [aux_sym_preproc_if_token1] = ACTIONS(2680),
    [aux_sym_preproc_if_token2] = ACTIONS(2680),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2680),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2680),
    [aux_sym_preproc_else_token1] = ACTIONS(2680),
    [aux_sym_preproc_elif_token1] = ACTIONS(2680),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2680),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2680),
    [sym_preproc_directive] = ACTIONS(2680),
    [anon_sym_LPAREN2] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2680),
    [anon_sym_PLUS] = ACTIONS(2680),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym___extension__] = ACTIONS(2680),
    [anon_sym_typedef] = ACTIONS(2680),
    [anon_sym_extern] = ACTIONS(2680),
    [anon_sym___attribute__] = ACTIONS(2680),
    [anon_sym___attribute] = ACTIONS(2680),
    [anon_sym_noreturn] = ACTIONS(2680),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym___declspec] = ACTIONS(2680),
    [anon_sym___cdecl] = ACTIONS(2680),
    [anon_sym___clrcall] = ACTIONS(2680),
    [anon_sym___stdcall] = ACTIONS(2680),
    [anon_sym___fastcall] = ACTIONS(2680),
    [anon_sym___thiscall] = ACTIONS(2680),
    [anon_sym___vectorcall] = ACTIONS(2680),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_signed] = ACTIONS(2680),
    [anon_sym_unsigned] = ACTIONS(2680),
    [anon_sym_long] = ACTIONS(2680),
    [anon_sym_short] = ACTIONS(2680),
    [anon_sym_static] = ACTIONS(2680),
    [anon_sym_ATautoreleasepool] = ACTIONS(2682),
    [anon_sym_auto] = ACTIONS(2680),
    [anon_sym_register] = ACTIONS(2680),
    [anon_sym_inline] = ACTIONS(2680),
    [anon_sym___inline] = ACTIONS(2680),
    [anon_sym___inline__] = ACTIONS(2680),
    [anon_sym___forceinline] = ACTIONS(2680),
    [anon_sym_thread_local] = ACTIONS(2680),
    [anon_sym___thread] = ACTIONS(2680),
    [anon_sym_CG_EXTERN] = ACTIONS(2680),
    [anon_sym_CG_INLINE] = ACTIONS(2680),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2680),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2680),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2680),
    [anon_sym_IBOutlet] = ACTIONS(2680),
    [anon_sym_IBInspectable] = ACTIONS(2680),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2680),
    [anon_sym_NS_INLINE] = ACTIONS(2680),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2680),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2680),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2680),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2680),
    [anon_sym_const] = ACTIONS(2680),
    [anon_sym_constexpr] = ACTIONS(2680),
    [anon_sym_volatile] = ACTIONS(2680),
    [anon_sym_restrict] = ACTIONS(2680),
    [anon_sym___restrict__] = ACTIONS(2680),
    [anon_sym__Atomic] = ACTIONS(2680),
    [anon_sym__Noreturn] = ACTIONS(2680),
    [anon_sym__Nonnull] = ACTIONS(2680),
    [anon_sym_nullable] = ACTIONS(2680),
    [anon_sym__Complex] = ACTIONS(2680),
    [anon_sym__Nullable] = ACTIONS(2680),
    [anon_sym__Nullable_result] = ACTIONS(2680),
    [anon_sym__Null_unspecified] = ACTIONS(2680),
    [anon_sym___autoreleasing] = ACTIONS(2680),
    [anon_sym___block] = ACTIONS(2680),
    [anon_sym___bridge] = ACTIONS(2680),
    [anon_sym___bridge_retained] = ACTIONS(2680),
    [anon_sym___bridge_transfer] = ACTIONS(2680),
    [anon_sym___complex] = ACTIONS(2680),
    [anon_sym___const] = ACTIONS(2680),
    [anon_sym___imag] = ACTIONS(2680),
    [anon_sym___kindof] = ACTIONS(2680),
    [anon_sym___nonnull] = ACTIONS(2680),
    [anon_sym___nullable] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2680),
    [anon_sym___real] = ACTIONS(2680),
    [anon_sym___strong] = ACTIONS(2680),
    [anon_sym___unsafe_unretained] = ACTIONS(2680),
    [anon_sym___unused] = ACTIONS(2680),
    [anon_sym___weak] = ACTIONS(2680),
    [anon_sym_alignas] = ACTIONS(2680),
    [anon_sym__Alignas] = ACTIONS(2680),
    [sym_primitive_type] = ACTIONS(2680),
    [anon_sym_enum] = ACTIONS(2680),
    [anon_sym_struct] = ACTIONS(2680),
    [anon_sym_union] = ACTIONS(2680),
    [anon_sym_if] = ACTIONS(2680),
    [anon_sym_switch] = ACTIONS(2680),
    [anon_sym_case] = ACTIONS(2680),
    [anon_sym_default] = ACTIONS(2680),
    [anon_sym_while] = ACTIONS(2680),
    [anon_sym_do] = ACTIONS(2680),
    [anon_sym_for] = ACTIONS(2680),
    [anon_sym_in] = ACTIONS(2680),
    [anon_sym_return] = ACTIONS(2680),
    [anon_sym_break] = ACTIONS(2680),
    [anon_sym_continue] = ACTIONS(2680),
    [anon_sym_goto] = ACTIONS(2680),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_sizeof] = ACTIONS(2680),
    [anon_sym___alignof__] = ACTIONS(2680),
    [anon_sym___alignof] = ACTIONS(2680),
    [anon_sym__alignof] = ACTIONS(2680),
    [anon_sym_alignof] = ACTIONS(2680),
    [anon_sym__Alignof] = ACTIONS(2680),
    [anon_sym_offsetof] = ACTIONS(2680),
    [anon_sym__Generic] = ACTIONS(2680),
    [anon_sym_asm] = ACTIONS(2680),
    [anon_sym___asm__] = ACTIONS(2680),
    [anon_sym___asm] = ACTIONS(2680),
    [sym_number_literal] = ACTIONS(2682),
    [anon_sym_L_SQUOTE] = ACTIONS(2682),
    [anon_sym_u_SQUOTE] = ACTIONS(2682),
    [anon_sym_U_SQUOTE] = ACTIONS(2682),
    [anon_sym_u8_SQUOTE] = ACTIONS(2682),
    [anon_sym_SQUOTE] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2680),
    [anon_sym_DQUOTE] = ACTIONS(2682),
    [anon_sym_L_DQUOTE] = ACTIONS(2682),
    [anon_sym_u_DQUOTE] = ACTIONS(2682),
    [anon_sym_U_DQUOTE] = ACTIONS(2682),
    [anon_sym_u8_DQUOTE] = ACTIONS(2682),
    [sym_true] = ACTIONS(2680),
    [sym_false] = ACTIONS(2680),
    [anon_sym_NULL] = ACTIONS(2680),
    [anon_sym_nullptr] = ACTIONS(2680),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2680),
    [anon_sym___typeof] = ACTIONS(2680),
    [anon_sym_typeof] = ACTIONS(2680),
    [anon_sym_ATimport] = ACTIONS(2682),
    [aux_sym_preproc_undef_token1] = ACTIONS(2680),
    [anon_sym_POUND] = ACTIONS(2680),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2680),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2680),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2680),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2680),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2680),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2680),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2680),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_API_AVAILABLE] = ACTIONS(2680),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_API_DEPRECATED] = ACTIONS(2680),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2680),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym___deprecated_msg] = ACTIONS(2680),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2680),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2680),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2680),
    [anon_sym_ATprotocol] = ACTIONS(2682),
    [anon_sym_ATinterface] = ACTIONS(2682),
    [anon_sym_ATimplementation] = ACTIONS(2682),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2682),
    [anon_sym_ATsynthesize] = ACTIONS(2682),
    [anon_sym_ATdynamic] = ACTIONS(2682),
    [anon_sym_ATtry] = ACTIONS(2682),
    [anon_sym___try] = ACTIONS(2680),
    [anon_sym_ATthrow] = ACTIONS(2682),
    [anon_sym_ATselector] = ACTIONS(2682),
    [anon_sym_ATavailable] = ACTIONS(2682),
    [anon_sym___builtin_available] = ACTIONS(2680),
    [anon_sym_va_arg] = ACTIONS(2680),
    [anon_sym_ATencode] = ACTIONS(2682),
    [anon_sym_ATsynchronized] = ACTIONS(2682),
    [anon_sym_BOOL] = ACTIONS(2680),
    [anon_sym_IMP] = ACTIONS(2680),
    [anon_sym_SEL] = ACTIONS(2680),
    [anon_sym_Class] = ACTIONS(2680),
    [anon_sym_id] = ACTIONS(2680),
  },
  [STATE(305)] = {
    [sym_identifier] = ACTIONS(2684),
    [aux_sym_preproc_include_token1] = ACTIONS(2684),
    [aux_sym_preproc_include_token2] = ACTIONS(2684),
    [aux_sym_preproc_def_token1] = ACTIONS(2684),
    [aux_sym_preproc_if_token1] = ACTIONS(2684),
    [aux_sym_preproc_if_token2] = ACTIONS(2684),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2684),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2684),
    [aux_sym_preproc_else_token1] = ACTIONS(2684),
    [aux_sym_preproc_elif_token1] = ACTIONS(2684),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2684),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2684),
    [sym_preproc_directive] = ACTIONS(2684),
    [anon_sym_LPAREN2] = ACTIONS(2686),
    [anon_sym_BANG] = ACTIONS(2686),
    [anon_sym_TILDE] = ACTIONS(2686),
    [anon_sym_DASH] = ACTIONS(2684),
    [anon_sym_PLUS] = ACTIONS(2684),
    [anon_sym_STAR] = ACTIONS(2686),
    [anon_sym_CARET] = ACTIONS(2686),
    [anon_sym_AMP] = ACTIONS(2686),
    [anon_sym_SEMI] = ACTIONS(2686),
    [anon_sym___extension__] = ACTIONS(2684),
    [anon_sym_typedef] = ACTIONS(2684),
    [anon_sym_extern] = ACTIONS(2684),
    [anon_sym___attribute__] = ACTIONS(2684),
    [anon_sym___attribute] = ACTIONS(2684),
    [anon_sym_noreturn] = ACTIONS(2684),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym___declspec] = ACTIONS(2684),
    [anon_sym___cdecl] = ACTIONS(2684),
    [anon_sym___clrcall] = ACTIONS(2684),
    [anon_sym___stdcall] = ACTIONS(2684),
    [anon_sym___fastcall] = ACTIONS(2684),
    [anon_sym___thiscall] = ACTIONS(2684),
    [anon_sym___vectorcall] = ACTIONS(2684),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_signed] = ACTIONS(2684),
    [anon_sym_unsigned] = ACTIONS(2684),
    [anon_sym_long] = ACTIONS(2684),
    [anon_sym_short] = ACTIONS(2684),
    [anon_sym_static] = ACTIONS(2684),
    [anon_sym_ATautoreleasepool] = ACTIONS(2686),
    [anon_sym_auto] = ACTIONS(2684),
    [anon_sym_register] = ACTIONS(2684),
    [anon_sym_inline] = ACTIONS(2684),
    [anon_sym___inline] = ACTIONS(2684),
    [anon_sym___inline__] = ACTIONS(2684),
    [anon_sym___forceinline] = ACTIONS(2684),
    [anon_sym_thread_local] = ACTIONS(2684),
    [anon_sym___thread] = ACTIONS(2684),
    [anon_sym_CG_EXTERN] = ACTIONS(2684),
    [anon_sym_CG_INLINE] = ACTIONS(2684),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2684),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2684),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2684),
    [anon_sym_IBOutlet] = ACTIONS(2684),
    [anon_sym_IBInspectable] = ACTIONS(2684),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2684),
    [anon_sym_NS_INLINE] = ACTIONS(2684),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2684),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2684),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2684),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2684),
    [anon_sym_const] = ACTIONS(2684),
    [anon_sym_constexpr] = ACTIONS(2684),
    [anon_sym_volatile] = ACTIONS(2684),
    [anon_sym_restrict] = ACTIONS(2684),
    [anon_sym___restrict__] = ACTIONS(2684),
    [anon_sym__Atomic] = ACTIONS(2684),
    [anon_sym__Noreturn] = ACTIONS(2684),
    [anon_sym__Nonnull] = ACTIONS(2684),
    [anon_sym_nullable] = ACTIONS(2684),
    [anon_sym__Complex] = ACTIONS(2684),
    [anon_sym__Nullable] = ACTIONS(2684),
    [anon_sym__Nullable_result] = ACTIONS(2684),
    [anon_sym__Null_unspecified] = ACTIONS(2684),
    [anon_sym___autoreleasing] = ACTIONS(2684),
    [anon_sym___block] = ACTIONS(2684),
    [anon_sym___bridge] = ACTIONS(2684),
    [anon_sym___bridge_retained] = ACTIONS(2684),
    [anon_sym___bridge_transfer] = ACTIONS(2684),
    [anon_sym___complex] = ACTIONS(2684),
    [anon_sym___const] = ACTIONS(2684),
    [anon_sym___imag] = ACTIONS(2684),
    [anon_sym___kindof] = ACTIONS(2684),
    [anon_sym___nonnull] = ACTIONS(2684),
    [anon_sym___nullable] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2684),
    [anon_sym___real] = ACTIONS(2684),
    [anon_sym___strong] = ACTIONS(2684),
    [anon_sym___unsafe_unretained] = ACTIONS(2684),
    [anon_sym___unused] = ACTIONS(2684),
    [anon_sym___weak] = ACTIONS(2684),
    [anon_sym_alignas] = ACTIONS(2684),
    [anon_sym__Alignas] = ACTIONS(2684),
    [sym_primitive_type] = ACTIONS(2684),
    [anon_sym_enum] = ACTIONS(2684),
    [anon_sym_struct] = ACTIONS(2684),
    [anon_sym_union] = ACTIONS(2684),
    [anon_sym_if] = ACTIONS(2684),
    [anon_sym_switch] = ACTIONS(2684),
    [anon_sym_case] = ACTIONS(2684),
    [anon_sym_default] = ACTIONS(2684),
    [anon_sym_while] = ACTIONS(2684),
    [anon_sym_do] = ACTIONS(2684),
    [anon_sym_for] = ACTIONS(2684),
    [anon_sym_in] = ACTIONS(2684),
    [anon_sym_return] = ACTIONS(2684),
    [anon_sym_break] = ACTIONS(2684),
    [anon_sym_continue] = ACTIONS(2684),
    [anon_sym_goto] = ACTIONS(2684),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_sizeof] = ACTIONS(2684),
    [anon_sym___alignof__] = ACTIONS(2684),
    [anon_sym___alignof] = ACTIONS(2684),
    [anon_sym__alignof] = ACTIONS(2684),
    [anon_sym_alignof] = ACTIONS(2684),
    [anon_sym__Alignof] = ACTIONS(2684),
    [anon_sym_offsetof] = ACTIONS(2684),
    [anon_sym__Generic] = ACTIONS(2684),
    [anon_sym_asm] = ACTIONS(2684),
    [anon_sym___asm__] = ACTIONS(2684),
    [anon_sym___asm] = ACTIONS(2684),
    [sym_number_literal] = ACTIONS(2686),
    [anon_sym_L_SQUOTE] = ACTIONS(2686),
    [anon_sym_u_SQUOTE] = ACTIONS(2686),
    [anon_sym_U_SQUOTE] = ACTIONS(2686),
    [anon_sym_u8_SQUOTE] = ACTIONS(2686),
    [anon_sym_SQUOTE] = ACTIONS(2686),
    [anon_sym_AT] = ACTIONS(2684),
    [anon_sym_DQUOTE] = ACTIONS(2686),
    [anon_sym_L_DQUOTE] = ACTIONS(2686),
    [anon_sym_u_DQUOTE] = ACTIONS(2686),
    [anon_sym_U_DQUOTE] = ACTIONS(2686),
    [anon_sym_u8_DQUOTE] = ACTIONS(2686),
    [sym_true] = ACTIONS(2684),
    [sym_false] = ACTIONS(2684),
    [anon_sym_NULL] = ACTIONS(2684),
    [anon_sym_nullptr] = ACTIONS(2684),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2684),
    [anon_sym___typeof] = ACTIONS(2684),
    [anon_sym_typeof] = ACTIONS(2684),
    [anon_sym_ATimport] = ACTIONS(2686),
    [aux_sym_preproc_undef_token1] = ACTIONS(2684),
    [anon_sym_POUND] = ACTIONS(2684),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2684),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2684),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2684),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2684),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2684),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2684),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2684),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_API_AVAILABLE] = ACTIONS(2684),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_API_DEPRECATED] = ACTIONS(2684),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2684),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym___deprecated_msg] = ACTIONS(2684),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2684),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2684),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2684),
    [anon_sym_ATprotocol] = ACTIONS(2686),
    [anon_sym_ATinterface] = ACTIONS(2686),
    [anon_sym_ATimplementation] = ACTIONS(2686),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2686),
    [anon_sym_ATsynthesize] = ACTIONS(2686),
    [anon_sym_ATdynamic] = ACTIONS(2686),
    [anon_sym_ATtry] = ACTIONS(2686),
    [anon_sym___try] = ACTIONS(2684),
    [anon_sym_ATthrow] = ACTIONS(2686),
    [anon_sym_ATselector] = ACTIONS(2686),
    [anon_sym_ATavailable] = ACTIONS(2686),
    [anon_sym___builtin_available] = ACTIONS(2684),
    [anon_sym_va_arg] = ACTIONS(2684),
    [anon_sym_ATencode] = ACTIONS(2686),
    [anon_sym_ATsynchronized] = ACTIONS(2686),
    [anon_sym_BOOL] = ACTIONS(2684),
    [anon_sym_IMP] = ACTIONS(2684),
    [anon_sym_SEL] = ACTIONS(2684),
    [anon_sym_Class] = ACTIONS(2684),
    [anon_sym_id] = ACTIONS(2684),
  },
  [STATE(306)] = {
    [sym_identifier] = ACTIONS(2688),
    [aux_sym_preproc_include_token1] = ACTIONS(2688),
    [aux_sym_preproc_include_token2] = ACTIONS(2688),
    [aux_sym_preproc_def_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token2] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2688),
    [aux_sym_preproc_else_token1] = ACTIONS(2688),
    [aux_sym_preproc_elif_token1] = ACTIONS(2688),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2688),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2688),
    [sym_preproc_directive] = ACTIONS(2688),
    [anon_sym_LPAREN2] = ACTIONS(2690),
    [anon_sym_BANG] = ACTIONS(2690),
    [anon_sym_TILDE] = ACTIONS(2690),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2690),
    [anon_sym_CARET] = ACTIONS(2690),
    [anon_sym_AMP] = ACTIONS(2690),
    [anon_sym_SEMI] = ACTIONS(2690),
    [anon_sym___extension__] = ACTIONS(2688),
    [anon_sym_typedef] = ACTIONS(2688),
    [anon_sym_extern] = ACTIONS(2688),
    [anon_sym___attribute__] = ACTIONS(2688),
    [anon_sym___attribute] = ACTIONS(2688),
    [anon_sym_noreturn] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym___declspec] = ACTIONS(2688),
    [anon_sym___cdecl] = ACTIONS(2688),
    [anon_sym___clrcall] = ACTIONS(2688),
    [anon_sym___stdcall] = ACTIONS(2688),
    [anon_sym___fastcall] = ACTIONS(2688),
    [anon_sym___thiscall] = ACTIONS(2688),
    [anon_sym___vectorcall] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_signed] = ACTIONS(2688),
    [anon_sym_unsigned] = ACTIONS(2688),
    [anon_sym_long] = ACTIONS(2688),
    [anon_sym_short] = ACTIONS(2688),
    [anon_sym_static] = ACTIONS(2688),
    [anon_sym_ATautoreleasepool] = ACTIONS(2690),
    [anon_sym_auto] = ACTIONS(2688),
    [anon_sym_register] = ACTIONS(2688),
    [anon_sym_inline] = ACTIONS(2688),
    [anon_sym___inline] = ACTIONS(2688),
    [anon_sym___inline__] = ACTIONS(2688),
    [anon_sym___forceinline] = ACTIONS(2688),
    [anon_sym_thread_local] = ACTIONS(2688),
    [anon_sym___thread] = ACTIONS(2688),
    [anon_sym_CG_EXTERN] = ACTIONS(2688),
    [anon_sym_CG_INLINE] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2688),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2688),
    [anon_sym_IBOutlet] = ACTIONS(2688),
    [anon_sym_IBInspectable] = ACTIONS(2688),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2688),
    [anon_sym_NS_INLINE] = ACTIONS(2688),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2688),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2688),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2688),
    [anon_sym_const] = ACTIONS(2688),
    [anon_sym_constexpr] = ACTIONS(2688),
    [anon_sym_volatile] = ACTIONS(2688),
    [anon_sym_restrict] = ACTIONS(2688),
    [anon_sym___restrict__] = ACTIONS(2688),
    [anon_sym__Atomic] = ACTIONS(2688),
    [anon_sym__Noreturn] = ACTIONS(2688),
    [anon_sym__Nonnull] = ACTIONS(2688),
    [anon_sym_nullable] = ACTIONS(2688),
    [anon_sym__Complex] = ACTIONS(2688),
    [anon_sym__Nullable] = ACTIONS(2688),
    [anon_sym__Nullable_result] = ACTIONS(2688),
    [anon_sym__Null_unspecified] = ACTIONS(2688),
    [anon_sym___autoreleasing] = ACTIONS(2688),
    [anon_sym___block] = ACTIONS(2688),
    [anon_sym___bridge] = ACTIONS(2688),
    [anon_sym___bridge_retained] = ACTIONS(2688),
    [anon_sym___bridge_transfer] = ACTIONS(2688),
    [anon_sym___complex] = ACTIONS(2688),
    [anon_sym___const] = ACTIONS(2688),
    [anon_sym___imag] = ACTIONS(2688),
    [anon_sym___kindof] = ACTIONS(2688),
    [anon_sym___nonnull] = ACTIONS(2688),
    [anon_sym___nullable] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2688),
    [anon_sym___real] = ACTIONS(2688),
    [anon_sym___strong] = ACTIONS(2688),
    [anon_sym___unsafe_unretained] = ACTIONS(2688),
    [anon_sym___unused] = ACTIONS(2688),
    [anon_sym___weak] = ACTIONS(2688),
    [anon_sym_alignas] = ACTIONS(2688),
    [anon_sym__Alignas] = ACTIONS(2688),
    [sym_primitive_type] = ACTIONS(2688),
    [anon_sym_enum] = ACTIONS(2688),
    [anon_sym_struct] = ACTIONS(2688),
    [anon_sym_union] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_switch] = ACTIONS(2688),
    [anon_sym_case] = ACTIONS(2688),
    [anon_sym_default] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_in] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_sizeof] = ACTIONS(2688),
    [anon_sym___alignof__] = ACTIONS(2688),
    [anon_sym___alignof] = ACTIONS(2688),
    [anon_sym__alignof] = ACTIONS(2688),
    [anon_sym_alignof] = ACTIONS(2688),
    [anon_sym__Alignof] = ACTIONS(2688),
    [anon_sym_offsetof] = ACTIONS(2688),
    [anon_sym__Generic] = ACTIONS(2688),
    [anon_sym_asm] = ACTIONS(2688),
    [anon_sym___asm__] = ACTIONS(2688),
    [anon_sym___asm] = ACTIONS(2688),
    [sym_number_literal] = ACTIONS(2690),
    [anon_sym_L_SQUOTE] = ACTIONS(2690),
    [anon_sym_u_SQUOTE] = ACTIONS(2690),
    [anon_sym_U_SQUOTE] = ACTIONS(2690),
    [anon_sym_u8_SQUOTE] = ACTIONS(2690),
    [anon_sym_SQUOTE] = ACTIONS(2690),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_DQUOTE] = ACTIONS(2690),
    [anon_sym_L_DQUOTE] = ACTIONS(2690),
    [anon_sym_u_DQUOTE] = ACTIONS(2690),
    [anon_sym_U_DQUOTE] = ACTIONS(2690),
    [anon_sym_u8_DQUOTE] = ACTIONS(2690),
    [sym_true] = ACTIONS(2688),
    [sym_false] = ACTIONS(2688),
    [anon_sym_NULL] = ACTIONS(2688),
    [anon_sym_nullptr] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2688),
    [anon_sym___typeof] = ACTIONS(2688),
    [anon_sym_typeof] = ACTIONS(2688),
    [anon_sym_ATimport] = ACTIONS(2690),
    [aux_sym_preproc_undef_token1] = ACTIONS(2688),
    [anon_sym_POUND] = ACTIONS(2688),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2688),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2688),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2688),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_API_AVAILABLE] = ACTIONS(2688),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_API_DEPRECATED] = ACTIONS(2688),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym___deprecated_msg] = ACTIONS(2688),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2688),
    [anon_sym_ATprotocol] = ACTIONS(2690),
    [anon_sym_ATinterface] = ACTIONS(2690),
    [anon_sym_ATimplementation] = ACTIONS(2690),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2690),
    [anon_sym_ATsynthesize] = ACTIONS(2690),
    [anon_sym_ATdynamic] = ACTIONS(2690),
    [anon_sym_ATtry] = ACTIONS(2690),
    [anon_sym___try] = ACTIONS(2688),
    [anon_sym_ATthrow] = ACTIONS(2690),
    [anon_sym_ATselector] = ACTIONS(2690),
    [anon_sym_ATavailable] = ACTIONS(2690),
    [anon_sym___builtin_available] = ACTIONS(2688),
    [anon_sym_va_arg] = ACTIONS(2688),
    [anon_sym_ATencode] = ACTIONS(2690),
    [anon_sym_ATsynchronized] = ACTIONS(2690),
    [anon_sym_BOOL] = ACTIONS(2688),
    [anon_sym_IMP] = ACTIONS(2688),
    [anon_sym_SEL] = ACTIONS(2688),
    [anon_sym_Class] = ACTIONS(2688),
    [anon_sym_id] = ACTIONS(2688),
  },
  [STATE(307)] = {
    [sym_identifier] = ACTIONS(2692),
    [aux_sym_preproc_include_token1] = ACTIONS(2692),
    [aux_sym_preproc_include_token2] = ACTIONS(2692),
    [aux_sym_preproc_def_token1] = ACTIONS(2692),
    [aux_sym_preproc_if_token1] = ACTIONS(2692),
    [aux_sym_preproc_if_token2] = ACTIONS(2692),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2692),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2692),
    [aux_sym_preproc_else_token1] = ACTIONS(2692),
    [aux_sym_preproc_elif_token1] = ACTIONS(2692),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2692),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2692),
    [sym_preproc_directive] = ACTIONS(2692),
    [anon_sym_LPAREN2] = ACTIONS(2694),
    [anon_sym_BANG] = ACTIONS(2694),
    [anon_sym_TILDE] = ACTIONS(2694),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2694),
    [anon_sym_CARET] = ACTIONS(2694),
    [anon_sym_AMP] = ACTIONS(2694),
    [anon_sym_SEMI] = ACTIONS(2694),
    [anon_sym___extension__] = ACTIONS(2692),
    [anon_sym_typedef] = ACTIONS(2692),
    [anon_sym_extern] = ACTIONS(2692),
    [anon_sym___attribute__] = ACTIONS(2692),
    [anon_sym___attribute] = ACTIONS(2692),
    [anon_sym_noreturn] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym___declspec] = ACTIONS(2692),
    [anon_sym___cdecl] = ACTIONS(2692),
    [anon_sym___clrcall] = ACTIONS(2692),
    [anon_sym___stdcall] = ACTIONS(2692),
    [anon_sym___fastcall] = ACTIONS(2692),
    [anon_sym___thiscall] = ACTIONS(2692),
    [anon_sym___vectorcall] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_signed] = ACTIONS(2692),
    [anon_sym_unsigned] = ACTIONS(2692),
    [anon_sym_long] = ACTIONS(2692),
    [anon_sym_short] = ACTIONS(2692),
    [anon_sym_static] = ACTIONS(2692),
    [anon_sym_ATautoreleasepool] = ACTIONS(2694),
    [anon_sym_auto] = ACTIONS(2692),
    [anon_sym_register] = ACTIONS(2692),
    [anon_sym_inline] = ACTIONS(2692),
    [anon_sym___inline] = ACTIONS(2692),
    [anon_sym___inline__] = ACTIONS(2692),
    [anon_sym___forceinline] = ACTIONS(2692),
    [anon_sym_thread_local] = ACTIONS(2692),
    [anon_sym___thread] = ACTIONS(2692),
    [anon_sym_CG_EXTERN] = ACTIONS(2692),
    [anon_sym_CG_INLINE] = ACTIONS(2692),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2692),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2692),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2692),
    [anon_sym_IBOutlet] = ACTIONS(2692),
    [anon_sym_IBInspectable] = ACTIONS(2692),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2692),
    [anon_sym_NS_INLINE] = ACTIONS(2692),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2692),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2692),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2692),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2692),
    [anon_sym_const] = ACTIONS(2692),
    [anon_sym_constexpr] = ACTIONS(2692),
    [anon_sym_volatile] = ACTIONS(2692),
    [anon_sym_restrict] = ACTIONS(2692),
    [anon_sym___restrict__] = ACTIONS(2692),
    [anon_sym__Atomic] = ACTIONS(2692),
    [anon_sym__Noreturn] = ACTIONS(2692),
    [anon_sym__Nonnull] = ACTIONS(2692),
    [anon_sym_nullable] = ACTIONS(2692),
    [anon_sym__Complex] = ACTIONS(2692),
    [anon_sym__Nullable] = ACTIONS(2692),
    [anon_sym__Nullable_result] = ACTIONS(2692),
    [anon_sym__Null_unspecified] = ACTIONS(2692),
    [anon_sym___autoreleasing] = ACTIONS(2692),
    [anon_sym___block] = ACTIONS(2692),
    [anon_sym___bridge] = ACTIONS(2692),
    [anon_sym___bridge_retained] = ACTIONS(2692),
    [anon_sym___bridge_transfer] = ACTIONS(2692),
    [anon_sym___complex] = ACTIONS(2692),
    [anon_sym___const] = ACTIONS(2692),
    [anon_sym___imag] = ACTIONS(2692),
    [anon_sym___kindof] = ACTIONS(2692),
    [anon_sym___nonnull] = ACTIONS(2692),
    [anon_sym___nullable] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2692),
    [anon_sym___real] = ACTIONS(2692),
    [anon_sym___strong] = ACTIONS(2692),
    [anon_sym___unsafe_unretained] = ACTIONS(2692),
    [anon_sym___unused] = ACTIONS(2692),
    [anon_sym___weak] = ACTIONS(2692),
    [anon_sym_alignas] = ACTIONS(2692),
    [anon_sym__Alignas] = ACTIONS(2692),
    [sym_primitive_type] = ACTIONS(2692),
    [anon_sym_enum] = ACTIONS(2692),
    [anon_sym_struct] = ACTIONS(2692),
    [anon_sym_union] = ACTIONS(2692),
    [anon_sym_if] = ACTIONS(2692),
    [anon_sym_switch] = ACTIONS(2692),
    [anon_sym_case] = ACTIONS(2692),
    [anon_sym_default] = ACTIONS(2692),
    [anon_sym_while] = ACTIONS(2692),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_for] = ACTIONS(2692),
    [anon_sym_in] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_sizeof] = ACTIONS(2692),
    [anon_sym___alignof__] = ACTIONS(2692),
    [anon_sym___alignof] = ACTIONS(2692),
    [anon_sym__alignof] = ACTIONS(2692),
    [anon_sym_alignof] = ACTIONS(2692),
    [anon_sym__Alignof] = ACTIONS(2692),
    [anon_sym_offsetof] = ACTIONS(2692),
    [anon_sym__Generic] = ACTIONS(2692),
    [anon_sym_asm] = ACTIONS(2692),
    [anon_sym___asm__] = ACTIONS(2692),
    [anon_sym___asm] = ACTIONS(2692),
    [sym_number_literal] = ACTIONS(2694),
    [anon_sym_L_SQUOTE] = ACTIONS(2694),
    [anon_sym_u_SQUOTE] = ACTIONS(2694),
    [anon_sym_U_SQUOTE] = ACTIONS(2694),
    [anon_sym_u8_SQUOTE] = ACTIONS(2694),
    [anon_sym_SQUOTE] = ACTIONS(2694),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_DQUOTE] = ACTIONS(2694),
    [anon_sym_L_DQUOTE] = ACTIONS(2694),
    [anon_sym_u_DQUOTE] = ACTIONS(2694),
    [anon_sym_U_DQUOTE] = ACTIONS(2694),
    [anon_sym_u8_DQUOTE] = ACTIONS(2694),
    [sym_true] = ACTIONS(2692),
    [sym_false] = ACTIONS(2692),
    [anon_sym_NULL] = ACTIONS(2692),
    [anon_sym_nullptr] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2692),
    [anon_sym___typeof] = ACTIONS(2692),
    [anon_sym_typeof] = ACTIONS(2692),
    [anon_sym_ATimport] = ACTIONS(2694),
    [aux_sym_preproc_undef_token1] = ACTIONS(2692),
    [anon_sym_POUND] = ACTIONS(2692),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2692),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2692),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2692),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2692),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2692),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2692),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2692),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_API_AVAILABLE] = ACTIONS(2692),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_API_DEPRECATED] = ACTIONS(2692),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2692),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym___deprecated_msg] = ACTIONS(2692),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2692),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2692),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2692),
    [anon_sym_ATprotocol] = ACTIONS(2694),
    [anon_sym_ATinterface] = ACTIONS(2694),
    [anon_sym_ATimplementation] = ACTIONS(2694),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2694),
    [anon_sym_ATsynthesize] = ACTIONS(2694),
    [anon_sym_ATdynamic] = ACTIONS(2694),
    [anon_sym_ATtry] = ACTIONS(2694),
    [anon_sym___try] = ACTIONS(2692),
    [anon_sym_ATthrow] = ACTIONS(2694),
    [anon_sym_ATselector] = ACTIONS(2694),
    [anon_sym_ATavailable] = ACTIONS(2694),
    [anon_sym___builtin_available] = ACTIONS(2692),
    [anon_sym_va_arg] = ACTIONS(2692),
    [anon_sym_ATencode] = ACTIONS(2694),
    [anon_sym_ATsynchronized] = ACTIONS(2694),
    [anon_sym_BOOL] = ACTIONS(2692),
    [anon_sym_IMP] = ACTIONS(2692),
    [anon_sym_SEL] = ACTIONS(2692),
    [anon_sym_Class] = ACTIONS(2692),
    [anon_sym_id] = ACTIONS(2692),
  },
  [STATE(308)] = {
    [sym_identifier] = ACTIONS(2696),
    [aux_sym_preproc_include_token1] = ACTIONS(2696),
    [aux_sym_preproc_include_token2] = ACTIONS(2696),
    [aux_sym_preproc_def_token1] = ACTIONS(2696),
    [aux_sym_preproc_if_token1] = ACTIONS(2696),
    [aux_sym_preproc_if_token2] = ACTIONS(2696),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2696),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2696),
    [aux_sym_preproc_else_token1] = ACTIONS(2696),
    [aux_sym_preproc_elif_token1] = ACTIONS(2696),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2696),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2696),
    [sym_preproc_directive] = ACTIONS(2696),
    [anon_sym_LPAREN2] = ACTIONS(2698),
    [anon_sym_BANG] = ACTIONS(2698),
    [anon_sym_TILDE] = ACTIONS(2698),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2698),
    [anon_sym_CARET] = ACTIONS(2698),
    [anon_sym_AMP] = ACTIONS(2698),
    [anon_sym_SEMI] = ACTIONS(2698),
    [anon_sym___extension__] = ACTIONS(2696),
    [anon_sym_typedef] = ACTIONS(2696),
    [anon_sym_extern] = ACTIONS(2696),
    [anon_sym___attribute__] = ACTIONS(2696),
    [anon_sym___attribute] = ACTIONS(2696),
    [anon_sym_noreturn] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym___declspec] = ACTIONS(2696),
    [anon_sym___cdecl] = ACTIONS(2696),
    [anon_sym___clrcall] = ACTIONS(2696),
    [anon_sym___stdcall] = ACTIONS(2696),
    [anon_sym___fastcall] = ACTIONS(2696),
    [anon_sym___thiscall] = ACTIONS(2696),
    [anon_sym___vectorcall] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_signed] = ACTIONS(2696),
    [anon_sym_unsigned] = ACTIONS(2696),
    [anon_sym_long] = ACTIONS(2696),
    [anon_sym_short] = ACTIONS(2696),
    [anon_sym_static] = ACTIONS(2696),
    [anon_sym_ATautoreleasepool] = ACTIONS(2698),
    [anon_sym_auto] = ACTIONS(2696),
    [anon_sym_register] = ACTIONS(2696),
    [anon_sym_inline] = ACTIONS(2696),
    [anon_sym___inline] = ACTIONS(2696),
    [anon_sym___inline__] = ACTIONS(2696),
    [anon_sym___forceinline] = ACTIONS(2696),
    [anon_sym_thread_local] = ACTIONS(2696),
    [anon_sym___thread] = ACTIONS(2696),
    [anon_sym_CG_EXTERN] = ACTIONS(2696),
    [anon_sym_CG_INLINE] = ACTIONS(2696),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2696),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2696),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2696),
    [anon_sym_IBOutlet] = ACTIONS(2696),
    [anon_sym_IBInspectable] = ACTIONS(2696),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2696),
    [anon_sym_NS_INLINE] = ACTIONS(2696),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2696),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2696),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2696),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2696),
    [anon_sym_const] = ACTIONS(2696),
    [anon_sym_constexpr] = ACTIONS(2696),
    [anon_sym_volatile] = ACTIONS(2696),
    [anon_sym_restrict] = ACTIONS(2696),
    [anon_sym___restrict__] = ACTIONS(2696),
    [anon_sym__Atomic] = ACTIONS(2696),
    [anon_sym__Noreturn] = ACTIONS(2696),
    [anon_sym__Nonnull] = ACTIONS(2696),
    [anon_sym_nullable] = ACTIONS(2696),
    [anon_sym__Complex] = ACTIONS(2696),
    [anon_sym__Nullable] = ACTIONS(2696),
    [anon_sym__Nullable_result] = ACTIONS(2696),
    [anon_sym__Null_unspecified] = ACTIONS(2696),
    [anon_sym___autoreleasing] = ACTIONS(2696),
    [anon_sym___block] = ACTIONS(2696),
    [anon_sym___bridge] = ACTIONS(2696),
    [anon_sym___bridge_retained] = ACTIONS(2696),
    [anon_sym___bridge_transfer] = ACTIONS(2696),
    [anon_sym___complex] = ACTIONS(2696),
    [anon_sym___const] = ACTIONS(2696),
    [anon_sym___imag] = ACTIONS(2696),
    [anon_sym___kindof] = ACTIONS(2696),
    [anon_sym___nonnull] = ACTIONS(2696),
    [anon_sym___nullable] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2696),
    [anon_sym___real] = ACTIONS(2696),
    [anon_sym___strong] = ACTIONS(2696),
    [anon_sym___unsafe_unretained] = ACTIONS(2696),
    [anon_sym___unused] = ACTIONS(2696),
    [anon_sym___weak] = ACTIONS(2696),
    [anon_sym_alignas] = ACTIONS(2696),
    [anon_sym__Alignas] = ACTIONS(2696),
    [sym_primitive_type] = ACTIONS(2696),
    [anon_sym_enum] = ACTIONS(2696),
    [anon_sym_struct] = ACTIONS(2696),
    [anon_sym_union] = ACTIONS(2696),
    [anon_sym_if] = ACTIONS(2696),
    [anon_sym_switch] = ACTIONS(2696),
    [anon_sym_case] = ACTIONS(2696),
    [anon_sym_default] = ACTIONS(2696),
    [anon_sym_while] = ACTIONS(2696),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_for] = ACTIONS(2696),
    [anon_sym_in] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_sizeof] = ACTIONS(2696),
    [anon_sym___alignof__] = ACTIONS(2696),
    [anon_sym___alignof] = ACTIONS(2696),
    [anon_sym__alignof] = ACTIONS(2696),
    [anon_sym_alignof] = ACTIONS(2696),
    [anon_sym__Alignof] = ACTIONS(2696),
    [anon_sym_offsetof] = ACTIONS(2696),
    [anon_sym__Generic] = ACTIONS(2696),
    [anon_sym_asm] = ACTIONS(2696),
    [anon_sym___asm__] = ACTIONS(2696),
    [anon_sym___asm] = ACTIONS(2696),
    [sym_number_literal] = ACTIONS(2698),
    [anon_sym_L_SQUOTE] = ACTIONS(2698),
    [anon_sym_u_SQUOTE] = ACTIONS(2698),
    [anon_sym_U_SQUOTE] = ACTIONS(2698),
    [anon_sym_u8_SQUOTE] = ACTIONS(2698),
    [anon_sym_SQUOTE] = ACTIONS(2698),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_DQUOTE] = ACTIONS(2698),
    [anon_sym_L_DQUOTE] = ACTIONS(2698),
    [anon_sym_u_DQUOTE] = ACTIONS(2698),
    [anon_sym_U_DQUOTE] = ACTIONS(2698),
    [anon_sym_u8_DQUOTE] = ACTIONS(2698),
    [sym_true] = ACTIONS(2696),
    [sym_false] = ACTIONS(2696),
    [anon_sym_NULL] = ACTIONS(2696),
    [anon_sym_nullptr] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2696),
    [anon_sym___typeof] = ACTIONS(2696),
    [anon_sym_typeof] = ACTIONS(2696),
    [anon_sym_ATimport] = ACTIONS(2698),
    [aux_sym_preproc_undef_token1] = ACTIONS(2696),
    [anon_sym_POUND] = ACTIONS(2696),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2696),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2696),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2696),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2696),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2696),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2696),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2696),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_API_AVAILABLE] = ACTIONS(2696),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_API_DEPRECATED] = ACTIONS(2696),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2696),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym___deprecated_msg] = ACTIONS(2696),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2696),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2696),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2696),
    [anon_sym_ATprotocol] = ACTIONS(2698),
    [anon_sym_ATinterface] = ACTIONS(2698),
    [anon_sym_ATimplementation] = ACTIONS(2698),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2698),
    [anon_sym_ATsynthesize] = ACTIONS(2698),
    [anon_sym_ATdynamic] = ACTIONS(2698),
    [anon_sym_ATtry] = ACTIONS(2698),
    [anon_sym___try] = ACTIONS(2696),
    [anon_sym_ATthrow] = ACTIONS(2698),
    [anon_sym_ATselector] = ACTIONS(2698),
    [anon_sym_ATavailable] = ACTIONS(2698),
    [anon_sym___builtin_available] = ACTIONS(2696),
    [anon_sym_va_arg] = ACTIONS(2696),
    [anon_sym_ATencode] = ACTIONS(2698),
    [anon_sym_ATsynchronized] = ACTIONS(2698),
    [anon_sym_BOOL] = ACTIONS(2696),
    [anon_sym_IMP] = ACTIONS(2696),
    [anon_sym_SEL] = ACTIONS(2696),
    [anon_sym_Class] = ACTIONS(2696),
    [anon_sym_id] = ACTIONS(2696),
  },
  [STATE(309)] = {
    [sym_identifier] = ACTIONS(2700),
    [aux_sym_preproc_include_token1] = ACTIONS(2700),
    [aux_sym_preproc_include_token2] = ACTIONS(2700),
    [aux_sym_preproc_def_token1] = ACTIONS(2700),
    [aux_sym_preproc_if_token1] = ACTIONS(2700),
    [aux_sym_preproc_if_token2] = ACTIONS(2700),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2700),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2700),
    [aux_sym_preproc_else_token1] = ACTIONS(2700),
    [aux_sym_preproc_elif_token1] = ACTIONS(2700),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2700),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2700),
    [sym_preproc_directive] = ACTIONS(2700),
    [anon_sym_LPAREN2] = ACTIONS(2702),
    [anon_sym_BANG] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2700),
    [anon_sym_PLUS] = ACTIONS(2700),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym___extension__] = ACTIONS(2700),
    [anon_sym_typedef] = ACTIONS(2700),
    [anon_sym_extern] = ACTIONS(2700),
    [anon_sym___attribute__] = ACTIONS(2700),
    [anon_sym___attribute] = ACTIONS(2700),
    [anon_sym_noreturn] = ACTIONS(2700),
    [anon_sym_LBRACK] = ACTIONS(2702),
    [anon_sym___declspec] = ACTIONS(2700),
    [anon_sym___cdecl] = ACTIONS(2700),
    [anon_sym___clrcall] = ACTIONS(2700),
    [anon_sym___stdcall] = ACTIONS(2700),
    [anon_sym___fastcall] = ACTIONS(2700),
    [anon_sym___thiscall] = ACTIONS(2700),
    [anon_sym___vectorcall] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2702),
    [anon_sym_signed] = ACTIONS(2700),
    [anon_sym_unsigned] = ACTIONS(2700),
    [anon_sym_long] = ACTIONS(2700),
    [anon_sym_short] = ACTIONS(2700),
    [anon_sym_static] = ACTIONS(2700),
    [anon_sym_ATautoreleasepool] = ACTIONS(2702),
    [anon_sym_auto] = ACTIONS(2700),
    [anon_sym_register] = ACTIONS(2700),
    [anon_sym_inline] = ACTIONS(2700),
    [anon_sym___inline] = ACTIONS(2700),
    [anon_sym___inline__] = ACTIONS(2700),
    [anon_sym___forceinline] = ACTIONS(2700),
    [anon_sym_thread_local] = ACTIONS(2700),
    [anon_sym___thread] = ACTIONS(2700),
    [anon_sym_CG_EXTERN] = ACTIONS(2700),
    [anon_sym_CG_INLINE] = ACTIONS(2700),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2700),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2700),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2700),
    [anon_sym_IBOutlet] = ACTIONS(2700),
    [anon_sym_IBInspectable] = ACTIONS(2700),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2700),
    [anon_sym_NS_INLINE] = ACTIONS(2700),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2700),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2700),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2700),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2700),
    [anon_sym_const] = ACTIONS(2700),
    [anon_sym_constexpr] = ACTIONS(2700),
    [anon_sym_volatile] = ACTIONS(2700),
    [anon_sym_restrict] = ACTIONS(2700),
    [anon_sym___restrict__] = ACTIONS(2700),
    [anon_sym__Atomic] = ACTIONS(2700),
    [anon_sym__Noreturn] = ACTIONS(2700),
    [anon_sym__Nonnull] = ACTIONS(2700),
    [anon_sym_nullable] = ACTIONS(2700),
    [anon_sym__Complex] = ACTIONS(2700),
    [anon_sym__Nullable] = ACTIONS(2700),
    [anon_sym__Nullable_result] = ACTIONS(2700),
    [anon_sym__Null_unspecified] = ACTIONS(2700),
    [anon_sym___autoreleasing] = ACTIONS(2700),
    [anon_sym___block] = ACTIONS(2700),
    [anon_sym___bridge] = ACTIONS(2700),
    [anon_sym___bridge_retained] = ACTIONS(2700),
    [anon_sym___bridge_transfer] = ACTIONS(2700),
    [anon_sym___complex] = ACTIONS(2700),
    [anon_sym___const] = ACTIONS(2700),
    [anon_sym___imag] = ACTIONS(2700),
    [anon_sym___kindof] = ACTIONS(2700),
    [anon_sym___nonnull] = ACTIONS(2700),
    [anon_sym___nullable] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2700),
    [anon_sym___real] = ACTIONS(2700),
    [anon_sym___strong] = ACTIONS(2700),
    [anon_sym___unsafe_unretained] = ACTIONS(2700),
    [anon_sym___unused] = ACTIONS(2700),
    [anon_sym___weak] = ACTIONS(2700),
    [anon_sym_alignas] = ACTIONS(2700),
    [anon_sym__Alignas] = ACTIONS(2700),
    [sym_primitive_type] = ACTIONS(2700),
    [anon_sym_enum] = ACTIONS(2700),
    [anon_sym_struct] = ACTIONS(2700),
    [anon_sym_union] = ACTIONS(2700),
    [anon_sym_if] = ACTIONS(2700),
    [anon_sym_switch] = ACTIONS(2700),
    [anon_sym_case] = ACTIONS(2700),
    [anon_sym_default] = ACTIONS(2700),
    [anon_sym_while] = ACTIONS(2700),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_for] = ACTIONS(2700),
    [anon_sym_in] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_DASH_DASH] = ACTIONS(2702),
    [anon_sym_PLUS_PLUS] = ACTIONS(2702),
    [anon_sym_sizeof] = ACTIONS(2700),
    [anon_sym___alignof__] = ACTIONS(2700),
    [anon_sym___alignof] = ACTIONS(2700),
    [anon_sym__alignof] = ACTIONS(2700),
    [anon_sym_alignof] = ACTIONS(2700),
    [anon_sym__Alignof] = ACTIONS(2700),
    [anon_sym_offsetof] = ACTIONS(2700),
    [anon_sym__Generic] = ACTIONS(2700),
    [anon_sym_asm] = ACTIONS(2700),
    [anon_sym___asm__] = ACTIONS(2700),
    [anon_sym___asm] = ACTIONS(2700),
    [sym_number_literal] = ACTIONS(2702),
    [anon_sym_L_SQUOTE] = ACTIONS(2702),
    [anon_sym_u_SQUOTE] = ACTIONS(2702),
    [anon_sym_U_SQUOTE] = ACTIONS(2702),
    [anon_sym_u8_SQUOTE] = ACTIONS(2702),
    [anon_sym_SQUOTE] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_DQUOTE] = ACTIONS(2702),
    [anon_sym_L_DQUOTE] = ACTIONS(2702),
    [anon_sym_u_DQUOTE] = ACTIONS(2702),
    [anon_sym_U_DQUOTE] = ACTIONS(2702),
    [anon_sym_u8_DQUOTE] = ACTIONS(2702),
    [sym_true] = ACTIONS(2700),
    [sym_false] = ACTIONS(2700),
    [anon_sym_NULL] = ACTIONS(2700),
    [anon_sym_nullptr] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2700),
    [anon_sym___typeof] = ACTIONS(2700),
    [anon_sym_typeof] = ACTIONS(2700),
    [anon_sym_ATimport] = ACTIONS(2702),
    [aux_sym_preproc_undef_token1] = ACTIONS(2700),
    [anon_sym_POUND] = ACTIONS(2700),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2700),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2700),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2700),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2700),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2700),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2700),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2700),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_API_AVAILABLE] = ACTIONS(2700),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_API_DEPRECATED] = ACTIONS(2700),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2700),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym___deprecated_msg] = ACTIONS(2700),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2700),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2700),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2700),
    [anon_sym_ATprotocol] = ACTIONS(2702),
    [anon_sym_ATinterface] = ACTIONS(2702),
    [anon_sym_ATimplementation] = ACTIONS(2702),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2702),
    [anon_sym_ATsynthesize] = ACTIONS(2702),
    [anon_sym_ATdynamic] = ACTIONS(2702),
    [anon_sym_ATtry] = ACTIONS(2702),
    [anon_sym___try] = ACTIONS(2700),
    [anon_sym_ATthrow] = ACTIONS(2702),
    [anon_sym_ATselector] = ACTIONS(2702),
    [anon_sym_ATavailable] = ACTIONS(2702),
    [anon_sym___builtin_available] = ACTIONS(2700),
    [anon_sym_va_arg] = ACTIONS(2700),
    [anon_sym_ATencode] = ACTIONS(2702),
    [anon_sym_ATsynchronized] = ACTIONS(2702),
    [anon_sym_BOOL] = ACTIONS(2700),
    [anon_sym_IMP] = ACTIONS(2700),
    [anon_sym_SEL] = ACTIONS(2700),
    [anon_sym_Class] = ACTIONS(2700),
    [anon_sym_id] = ACTIONS(2700),
  },
  [STATE(310)] = {
    [sym_identifier] = ACTIONS(2704),
    [aux_sym_preproc_include_token1] = ACTIONS(2704),
    [aux_sym_preproc_include_token2] = ACTIONS(2704),
    [aux_sym_preproc_def_token1] = ACTIONS(2704),
    [aux_sym_preproc_if_token1] = ACTIONS(2704),
    [aux_sym_preproc_if_token2] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2704),
    [aux_sym_preproc_else_token1] = ACTIONS(2704),
    [aux_sym_preproc_elif_token1] = ACTIONS(2704),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2704),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2704),
    [sym_preproc_directive] = ACTIONS(2704),
    [anon_sym_LPAREN2] = ACTIONS(2706),
    [anon_sym_BANG] = ACTIONS(2706),
    [anon_sym_TILDE] = ACTIONS(2706),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_STAR] = ACTIONS(2706),
    [anon_sym_CARET] = ACTIONS(2706),
    [anon_sym_AMP] = ACTIONS(2706),
    [anon_sym_SEMI] = ACTIONS(2706),
    [anon_sym___extension__] = ACTIONS(2704),
    [anon_sym_typedef] = ACTIONS(2704),
    [anon_sym_extern] = ACTIONS(2704),
    [anon_sym___attribute__] = ACTIONS(2704),
    [anon_sym___attribute] = ACTIONS(2704),
    [anon_sym_noreturn] = ACTIONS(2704),
    [anon_sym_LBRACK] = ACTIONS(2706),
    [anon_sym___declspec] = ACTIONS(2704),
    [anon_sym___cdecl] = ACTIONS(2704),
    [anon_sym___clrcall] = ACTIONS(2704),
    [anon_sym___stdcall] = ACTIONS(2704),
    [anon_sym___fastcall] = ACTIONS(2704),
    [anon_sym___thiscall] = ACTIONS(2704),
    [anon_sym___vectorcall] = ACTIONS(2704),
    [anon_sym_LBRACE] = ACTIONS(2706),
    [anon_sym_signed] = ACTIONS(2704),
    [anon_sym_unsigned] = ACTIONS(2704),
    [anon_sym_long] = ACTIONS(2704),
    [anon_sym_short] = ACTIONS(2704),
    [anon_sym_static] = ACTIONS(2704),
    [anon_sym_ATautoreleasepool] = ACTIONS(2706),
    [anon_sym_auto] = ACTIONS(2704),
    [anon_sym_register] = ACTIONS(2704),
    [anon_sym_inline] = ACTIONS(2704),
    [anon_sym___inline] = ACTIONS(2704),
    [anon_sym___inline__] = ACTIONS(2704),
    [anon_sym___forceinline] = ACTIONS(2704),
    [anon_sym_thread_local] = ACTIONS(2704),
    [anon_sym___thread] = ACTIONS(2704),
    [anon_sym_CG_EXTERN] = ACTIONS(2704),
    [anon_sym_CG_INLINE] = ACTIONS(2704),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2704),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2704),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2704),
    [anon_sym_IBOutlet] = ACTIONS(2704),
    [anon_sym_IBInspectable] = ACTIONS(2704),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2704),
    [anon_sym_NS_INLINE] = ACTIONS(2704),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2704),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2704),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2704),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2704),
    [anon_sym_const] = ACTIONS(2704),
    [anon_sym_constexpr] = ACTIONS(2704),
    [anon_sym_volatile] = ACTIONS(2704),
    [anon_sym_restrict] = ACTIONS(2704),
    [anon_sym___restrict__] = ACTIONS(2704),
    [anon_sym__Atomic] = ACTIONS(2704),
    [anon_sym__Noreturn] = ACTIONS(2704),
    [anon_sym__Nonnull] = ACTIONS(2704),
    [anon_sym_nullable] = ACTIONS(2704),
    [anon_sym__Complex] = ACTIONS(2704),
    [anon_sym__Nullable] = ACTIONS(2704),
    [anon_sym__Nullable_result] = ACTIONS(2704),
    [anon_sym__Null_unspecified] = ACTIONS(2704),
    [anon_sym___autoreleasing] = ACTIONS(2704),
    [anon_sym___block] = ACTIONS(2704),
    [anon_sym___bridge] = ACTIONS(2704),
    [anon_sym___bridge_retained] = ACTIONS(2704),
    [anon_sym___bridge_transfer] = ACTIONS(2704),
    [anon_sym___complex] = ACTIONS(2704),
    [anon_sym___const] = ACTIONS(2704),
    [anon_sym___imag] = ACTIONS(2704),
    [anon_sym___kindof] = ACTIONS(2704),
    [anon_sym___nonnull] = ACTIONS(2704),
    [anon_sym___nullable] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2704),
    [anon_sym___real] = ACTIONS(2704),
    [anon_sym___strong] = ACTIONS(2704),
    [anon_sym___unsafe_unretained] = ACTIONS(2704),
    [anon_sym___unused] = ACTIONS(2704),
    [anon_sym___weak] = ACTIONS(2704),
    [anon_sym_alignas] = ACTIONS(2704),
    [anon_sym__Alignas] = ACTIONS(2704),
    [sym_primitive_type] = ACTIONS(2704),
    [anon_sym_enum] = ACTIONS(2704),
    [anon_sym_struct] = ACTIONS(2704),
    [anon_sym_union] = ACTIONS(2704),
    [anon_sym_if] = ACTIONS(2704),
    [anon_sym_switch] = ACTIONS(2704),
    [anon_sym_case] = ACTIONS(2704),
    [anon_sym_default] = ACTIONS(2704),
    [anon_sym_while] = ACTIONS(2704),
    [anon_sym_do] = ACTIONS(2704),
    [anon_sym_for] = ACTIONS(2704),
    [anon_sym_in] = ACTIONS(2704),
    [anon_sym_return] = ACTIONS(2704),
    [anon_sym_break] = ACTIONS(2704),
    [anon_sym_continue] = ACTIONS(2704),
    [anon_sym_goto] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2706),
    [anon_sym_PLUS_PLUS] = ACTIONS(2706),
    [anon_sym_sizeof] = ACTIONS(2704),
    [anon_sym___alignof__] = ACTIONS(2704),
    [anon_sym___alignof] = ACTIONS(2704),
    [anon_sym__alignof] = ACTIONS(2704),
    [anon_sym_alignof] = ACTIONS(2704),
    [anon_sym__Alignof] = ACTIONS(2704),
    [anon_sym_offsetof] = ACTIONS(2704),
    [anon_sym__Generic] = ACTIONS(2704),
    [anon_sym_asm] = ACTIONS(2704),
    [anon_sym___asm__] = ACTIONS(2704),
    [anon_sym___asm] = ACTIONS(2704),
    [sym_number_literal] = ACTIONS(2706),
    [anon_sym_L_SQUOTE] = ACTIONS(2706),
    [anon_sym_u_SQUOTE] = ACTIONS(2706),
    [anon_sym_U_SQUOTE] = ACTIONS(2706),
    [anon_sym_u8_SQUOTE] = ACTIONS(2706),
    [anon_sym_SQUOTE] = ACTIONS(2706),
    [anon_sym_AT] = ACTIONS(2704),
    [anon_sym_DQUOTE] = ACTIONS(2706),
    [anon_sym_L_DQUOTE] = ACTIONS(2706),
    [anon_sym_u_DQUOTE] = ACTIONS(2706),
    [anon_sym_U_DQUOTE] = ACTIONS(2706),
    [anon_sym_u8_DQUOTE] = ACTIONS(2706),
    [sym_true] = ACTIONS(2704),
    [sym_false] = ACTIONS(2704),
    [anon_sym_NULL] = ACTIONS(2704),
    [anon_sym_nullptr] = ACTIONS(2704),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2704),
    [anon_sym___typeof] = ACTIONS(2704),
    [anon_sym_typeof] = ACTIONS(2704),
    [anon_sym_ATimport] = ACTIONS(2706),
    [aux_sym_preproc_undef_token1] = ACTIONS(2704),
    [anon_sym_POUND] = ACTIONS(2704),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2704),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2704),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2704),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2704),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2704),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2704),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2704),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_API_AVAILABLE] = ACTIONS(2704),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_API_DEPRECATED] = ACTIONS(2704),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2704),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym___deprecated_msg] = ACTIONS(2704),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2704),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2704),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2704),
    [anon_sym_ATprotocol] = ACTIONS(2706),
    [anon_sym_ATinterface] = ACTIONS(2706),
    [anon_sym_ATimplementation] = ACTIONS(2706),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2706),
    [anon_sym_ATsynthesize] = ACTIONS(2706),
    [anon_sym_ATdynamic] = ACTIONS(2706),
    [anon_sym_ATtry] = ACTIONS(2706),
    [anon_sym___try] = ACTIONS(2704),
    [anon_sym_ATthrow] = ACTIONS(2706),
    [anon_sym_ATselector] = ACTIONS(2706),
    [anon_sym_ATavailable] = ACTIONS(2706),
    [anon_sym___builtin_available] = ACTIONS(2704),
    [anon_sym_va_arg] = ACTIONS(2704),
    [anon_sym_ATencode] = ACTIONS(2706),
    [anon_sym_ATsynchronized] = ACTIONS(2706),
    [anon_sym_BOOL] = ACTIONS(2704),
    [anon_sym_IMP] = ACTIONS(2704),
    [anon_sym_SEL] = ACTIONS(2704),
    [anon_sym_Class] = ACTIONS(2704),
    [anon_sym_id] = ACTIONS(2704),
  },
  [STATE(311)] = {
    [sym_identifier] = ACTIONS(2708),
    [aux_sym_preproc_include_token1] = ACTIONS(2708),
    [aux_sym_preproc_include_token2] = ACTIONS(2708),
    [aux_sym_preproc_def_token1] = ACTIONS(2708),
    [aux_sym_preproc_if_token1] = ACTIONS(2708),
    [aux_sym_preproc_if_token2] = ACTIONS(2708),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2708),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2708),
    [aux_sym_preproc_else_token1] = ACTIONS(2708),
    [aux_sym_preproc_elif_token1] = ACTIONS(2708),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2708),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2708),
    [sym_preproc_directive] = ACTIONS(2708),
    [anon_sym_LPAREN2] = ACTIONS(2710),
    [anon_sym_BANG] = ACTIONS(2710),
    [anon_sym_TILDE] = ACTIONS(2710),
    [anon_sym_DASH] = ACTIONS(2708),
    [anon_sym_PLUS] = ACTIONS(2708),
    [anon_sym_STAR] = ACTIONS(2710),
    [anon_sym_CARET] = ACTIONS(2710),
    [anon_sym_AMP] = ACTIONS(2710),
    [anon_sym_SEMI] = ACTIONS(2710),
    [anon_sym___extension__] = ACTIONS(2708),
    [anon_sym_typedef] = ACTIONS(2708),
    [anon_sym_extern] = ACTIONS(2708),
    [anon_sym___attribute__] = ACTIONS(2708),
    [anon_sym___attribute] = ACTIONS(2708),
    [anon_sym_noreturn] = ACTIONS(2708),
    [anon_sym_LBRACK] = ACTIONS(2710),
    [anon_sym___declspec] = ACTIONS(2708),
    [anon_sym___cdecl] = ACTIONS(2708),
    [anon_sym___clrcall] = ACTIONS(2708),
    [anon_sym___stdcall] = ACTIONS(2708),
    [anon_sym___fastcall] = ACTIONS(2708),
    [anon_sym___thiscall] = ACTIONS(2708),
    [anon_sym___vectorcall] = ACTIONS(2708),
    [anon_sym_LBRACE] = ACTIONS(2710),
    [anon_sym_signed] = ACTIONS(2708),
    [anon_sym_unsigned] = ACTIONS(2708),
    [anon_sym_long] = ACTIONS(2708),
    [anon_sym_short] = ACTIONS(2708),
    [anon_sym_static] = ACTIONS(2708),
    [anon_sym_ATautoreleasepool] = ACTIONS(2710),
    [anon_sym_auto] = ACTIONS(2708),
    [anon_sym_register] = ACTIONS(2708),
    [anon_sym_inline] = ACTIONS(2708),
    [anon_sym___inline] = ACTIONS(2708),
    [anon_sym___inline__] = ACTIONS(2708),
    [anon_sym___forceinline] = ACTIONS(2708),
    [anon_sym_thread_local] = ACTIONS(2708),
    [anon_sym___thread] = ACTIONS(2708),
    [anon_sym_CG_EXTERN] = ACTIONS(2708),
    [anon_sym_CG_INLINE] = ACTIONS(2708),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2708),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2708),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2708),
    [anon_sym_IBOutlet] = ACTIONS(2708),
    [anon_sym_IBInspectable] = ACTIONS(2708),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2708),
    [anon_sym_NS_INLINE] = ACTIONS(2708),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2708),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2708),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2708),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2708),
    [anon_sym_const] = ACTIONS(2708),
    [anon_sym_constexpr] = ACTIONS(2708),
    [anon_sym_volatile] = ACTIONS(2708),
    [anon_sym_restrict] = ACTIONS(2708),
    [anon_sym___restrict__] = ACTIONS(2708),
    [anon_sym__Atomic] = ACTIONS(2708),
    [anon_sym__Noreturn] = ACTIONS(2708),
    [anon_sym__Nonnull] = ACTIONS(2708),
    [anon_sym_nullable] = ACTIONS(2708),
    [anon_sym__Complex] = ACTIONS(2708),
    [anon_sym__Nullable] = ACTIONS(2708),
    [anon_sym__Nullable_result] = ACTIONS(2708),
    [anon_sym__Null_unspecified] = ACTIONS(2708),
    [anon_sym___autoreleasing] = ACTIONS(2708),
    [anon_sym___block] = ACTIONS(2708),
    [anon_sym___bridge] = ACTIONS(2708),
    [anon_sym___bridge_retained] = ACTIONS(2708),
    [anon_sym___bridge_transfer] = ACTIONS(2708),
    [anon_sym___complex] = ACTIONS(2708),
    [anon_sym___const] = ACTIONS(2708),
    [anon_sym___imag] = ACTIONS(2708),
    [anon_sym___kindof] = ACTIONS(2708),
    [anon_sym___nonnull] = ACTIONS(2708),
    [anon_sym___nullable] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2708),
    [anon_sym___real] = ACTIONS(2708),
    [anon_sym___strong] = ACTIONS(2708),
    [anon_sym___unsafe_unretained] = ACTIONS(2708),
    [anon_sym___unused] = ACTIONS(2708),
    [anon_sym___weak] = ACTIONS(2708),
    [anon_sym_alignas] = ACTIONS(2708),
    [anon_sym__Alignas] = ACTIONS(2708),
    [sym_primitive_type] = ACTIONS(2708),
    [anon_sym_enum] = ACTIONS(2708),
    [anon_sym_struct] = ACTIONS(2708),
    [anon_sym_union] = ACTIONS(2708),
    [anon_sym_if] = ACTIONS(2708),
    [anon_sym_switch] = ACTIONS(2708),
    [anon_sym_case] = ACTIONS(2708),
    [anon_sym_default] = ACTIONS(2708),
    [anon_sym_while] = ACTIONS(2708),
    [anon_sym_do] = ACTIONS(2708),
    [anon_sym_for] = ACTIONS(2708),
    [anon_sym_in] = ACTIONS(2708),
    [anon_sym_return] = ACTIONS(2708),
    [anon_sym_break] = ACTIONS(2708),
    [anon_sym_continue] = ACTIONS(2708),
    [anon_sym_goto] = ACTIONS(2708),
    [anon_sym_DASH_DASH] = ACTIONS(2710),
    [anon_sym_PLUS_PLUS] = ACTIONS(2710),
    [anon_sym_sizeof] = ACTIONS(2708),
    [anon_sym___alignof__] = ACTIONS(2708),
    [anon_sym___alignof] = ACTIONS(2708),
    [anon_sym__alignof] = ACTIONS(2708),
    [anon_sym_alignof] = ACTIONS(2708),
    [anon_sym__Alignof] = ACTIONS(2708),
    [anon_sym_offsetof] = ACTIONS(2708),
    [anon_sym__Generic] = ACTIONS(2708),
    [anon_sym_asm] = ACTIONS(2708),
    [anon_sym___asm__] = ACTIONS(2708),
    [anon_sym___asm] = ACTIONS(2708),
    [sym_number_literal] = ACTIONS(2710),
    [anon_sym_L_SQUOTE] = ACTIONS(2710),
    [anon_sym_u_SQUOTE] = ACTIONS(2710),
    [anon_sym_U_SQUOTE] = ACTIONS(2710),
    [anon_sym_u8_SQUOTE] = ACTIONS(2710),
    [anon_sym_SQUOTE] = ACTIONS(2710),
    [anon_sym_AT] = ACTIONS(2708),
    [anon_sym_DQUOTE] = ACTIONS(2710),
    [anon_sym_L_DQUOTE] = ACTIONS(2710),
    [anon_sym_u_DQUOTE] = ACTIONS(2710),
    [anon_sym_U_DQUOTE] = ACTIONS(2710),
    [anon_sym_u8_DQUOTE] = ACTIONS(2710),
    [sym_true] = ACTIONS(2708),
    [sym_false] = ACTIONS(2708),
    [anon_sym_NULL] = ACTIONS(2708),
    [anon_sym_nullptr] = ACTIONS(2708),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2708),
    [anon_sym___typeof] = ACTIONS(2708),
    [anon_sym_typeof] = ACTIONS(2708),
    [anon_sym_ATimport] = ACTIONS(2710),
    [aux_sym_preproc_undef_token1] = ACTIONS(2708),
    [anon_sym_POUND] = ACTIONS(2708),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2708),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2708),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2708),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2708),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2708),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2708),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2708),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_API_AVAILABLE] = ACTIONS(2708),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_API_DEPRECATED] = ACTIONS(2708),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2708),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym___deprecated_msg] = ACTIONS(2708),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2708),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2708),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2708),
    [anon_sym_ATprotocol] = ACTIONS(2710),
    [anon_sym_ATinterface] = ACTIONS(2710),
    [anon_sym_ATimplementation] = ACTIONS(2710),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2710),
    [anon_sym_ATsynthesize] = ACTIONS(2710),
    [anon_sym_ATdynamic] = ACTIONS(2710),
    [anon_sym_ATtry] = ACTIONS(2710),
    [anon_sym___try] = ACTIONS(2708),
    [anon_sym_ATthrow] = ACTIONS(2710),
    [anon_sym_ATselector] = ACTIONS(2710),
    [anon_sym_ATavailable] = ACTIONS(2710),
    [anon_sym___builtin_available] = ACTIONS(2708),
    [anon_sym_va_arg] = ACTIONS(2708),
    [anon_sym_ATencode] = ACTIONS(2710),
    [anon_sym_ATsynchronized] = ACTIONS(2710),
    [anon_sym_BOOL] = ACTIONS(2708),
    [anon_sym_IMP] = ACTIONS(2708),
    [anon_sym_SEL] = ACTIONS(2708),
    [anon_sym_Class] = ACTIONS(2708),
    [anon_sym_id] = ACTIONS(2708),
  },
  [STATE(312)] = {
    [sym_identifier] = ACTIONS(2712),
    [aux_sym_preproc_include_token1] = ACTIONS(2712),
    [aux_sym_preproc_include_token2] = ACTIONS(2712),
    [aux_sym_preproc_def_token1] = ACTIONS(2712),
    [aux_sym_preproc_if_token1] = ACTIONS(2712),
    [aux_sym_preproc_if_token2] = ACTIONS(2712),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2712),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2712),
    [aux_sym_preproc_else_token1] = ACTIONS(2712),
    [aux_sym_preproc_elif_token1] = ACTIONS(2712),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2712),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2712),
    [sym_preproc_directive] = ACTIONS(2712),
    [anon_sym_LPAREN2] = ACTIONS(2714),
    [anon_sym_BANG] = ACTIONS(2714),
    [anon_sym_TILDE] = ACTIONS(2714),
    [anon_sym_DASH] = ACTIONS(2712),
    [anon_sym_PLUS] = ACTIONS(2712),
    [anon_sym_STAR] = ACTIONS(2714),
    [anon_sym_CARET] = ACTIONS(2714),
    [anon_sym_AMP] = ACTIONS(2714),
    [anon_sym_SEMI] = ACTIONS(2714),
    [anon_sym___extension__] = ACTIONS(2712),
    [anon_sym_typedef] = ACTIONS(2712),
    [anon_sym_extern] = ACTIONS(2712),
    [anon_sym___attribute__] = ACTIONS(2712),
    [anon_sym___attribute] = ACTIONS(2712),
    [anon_sym_noreturn] = ACTIONS(2712),
    [anon_sym_LBRACK] = ACTIONS(2714),
    [anon_sym___declspec] = ACTIONS(2712),
    [anon_sym___cdecl] = ACTIONS(2712),
    [anon_sym___clrcall] = ACTIONS(2712),
    [anon_sym___stdcall] = ACTIONS(2712),
    [anon_sym___fastcall] = ACTIONS(2712),
    [anon_sym___thiscall] = ACTIONS(2712),
    [anon_sym___vectorcall] = ACTIONS(2712),
    [anon_sym_LBRACE] = ACTIONS(2714),
    [anon_sym_signed] = ACTIONS(2712),
    [anon_sym_unsigned] = ACTIONS(2712),
    [anon_sym_long] = ACTIONS(2712),
    [anon_sym_short] = ACTIONS(2712),
    [anon_sym_static] = ACTIONS(2712),
    [anon_sym_ATautoreleasepool] = ACTIONS(2714),
    [anon_sym_auto] = ACTIONS(2712),
    [anon_sym_register] = ACTIONS(2712),
    [anon_sym_inline] = ACTIONS(2712),
    [anon_sym___inline] = ACTIONS(2712),
    [anon_sym___inline__] = ACTIONS(2712),
    [anon_sym___forceinline] = ACTIONS(2712),
    [anon_sym_thread_local] = ACTIONS(2712),
    [anon_sym___thread] = ACTIONS(2712),
    [anon_sym_CG_EXTERN] = ACTIONS(2712),
    [anon_sym_CG_INLINE] = ACTIONS(2712),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2712),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2712),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2712),
    [anon_sym_IBOutlet] = ACTIONS(2712),
    [anon_sym_IBInspectable] = ACTIONS(2712),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2712),
    [anon_sym_NS_INLINE] = ACTIONS(2712),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2712),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2712),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2712),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2712),
    [anon_sym_const] = ACTIONS(2712),
    [anon_sym_constexpr] = ACTIONS(2712),
    [anon_sym_volatile] = ACTIONS(2712),
    [anon_sym_restrict] = ACTIONS(2712),
    [anon_sym___restrict__] = ACTIONS(2712),
    [anon_sym__Atomic] = ACTIONS(2712),
    [anon_sym__Noreturn] = ACTIONS(2712),
    [anon_sym__Nonnull] = ACTIONS(2712),
    [anon_sym_nullable] = ACTIONS(2712),
    [anon_sym__Complex] = ACTIONS(2712),
    [anon_sym__Nullable] = ACTIONS(2712),
    [anon_sym__Nullable_result] = ACTIONS(2712),
    [anon_sym__Null_unspecified] = ACTIONS(2712),
    [anon_sym___autoreleasing] = ACTIONS(2712),
    [anon_sym___block] = ACTIONS(2712),
    [anon_sym___bridge] = ACTIONS(2712),
    [anon_sym___bridge_retained] = ACTIONS(2712),
    [anon_sym___bridge_transfer] = ACTIONS(2712),
    [anon_sym___complex] = ACTIONS(2712),
    [anon_sym___const] = ACTIONS(2712),
    [anon_sym___imag] = ACTIONS(2712),
    [anon_sym___kindof] = ACTIONS(2712),
    [anon_sym___nonnull] = ACTIONS(2712),
    [anon_sym___nullable] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2712),
    [anon_sym___real] = ACTIONS(2712),
    [anon_sym___strong] = ACTIONS(2712),
    [anon_sym___unsafe_unretained] = ACTIONS(2712),
    [anon_sym___unused] = ACTIONS(2712),
    [anon_sym___weak] = ACTIONS(2712),
    [anon_sym_alignas] = ACTIONS(2712),
    [anon_sym__Alignas] = ACTIONS(2712),
    [sym_primitive_type] = ACTIONS(2712),
    [anon_sym_enum] = ACTIONS(2712),
    [anon_sym_struct] = ACTIONS(2712),
    [anon_sym_union] = ACTIONS(2712),
    [anon_sym_if] = ACTIONS(2712),
    [anon_sym_switch] = ACTIONS(2712),
    [anon_sym_case] = ACTIONS(2712),
    [anon_sym_default] = ACTIONS(2712),
    [anon_sym_while] = ACTIONS(2712),
    [anon_sym_do] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2712),
    [anon_sym_in] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_break] = ACTIONS(2712),
    [anon_sym_continue] = ACTIONS(2712),
    [anon_sym_goto] = ACTIONS(2712),
    [anon_sym_DASH_DASH] = ACTIONS(2714),
    [anon_sym_PLUS_PLUS] = ACTIONS(2714),
    [anon_sym_sizeof] = ACTIONS(2712),
    [anon_sym___alignof__] = ACTIONS(2712),
    [anon_sym___alignof] = ACTIONS(2712),
    [anon_sym__alignof] = ACTIONS(2712),
    [anon_sym_alignof] = ACTIONS(2712),
    [anon_sym__Alignof] = ACTIONS(2712),
    [anon_sym_offsetof] = ACTIONS(2712),
    [anon_sym__Generic] = ACTIONS(2712),
    [anon_sym_asm] = ACTIONS(2712),
    [anon_sym___asm__] = ACTIONS(2712),
    [anon_sym___asm] = ACTIONS(2712),
    [sym_number_literal] = ACTIONS(2714),
    [anon_sym_L_SQUOTE] = ACTIONS(2714),
    [anon_sym_u_SQUOTE] = ACTIONS(2714),
    [anon_sym_U_SQUOTE] = ACTIONS(2714),
    [anon_sym_u8_SQUOTE] = ACTIONS(2714),
    [anon_sym_SQUOTE] = ACTIONS(2714),
    [anon_sym_AT] = ACTIONS(2712),
    [anon_sym_DQUOTE] = ACTIONS(2714),
    [anon_sym_L_DQUOTE] = ACTIONS(2714),
    [anon_sym_u_DQUOTE] = ACTIONS(2714),
    [anon_sym_U_DQUOTE] = ACTIONS(2714),
    [anon_sym_u8_DQUOTE] = ACTIONS(2714),
    [sym_true] = ACTIONS(2712),
    [sym_false] = ACTIONS(2712),
    [anon_sym_NULL] = ACTIONS(2712),
    [anon_sym_nullptr] = ACTIONS(2712),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2712),
    [anon_sym___typeof] = ACTIONS(2712),
    [anon_sym_typeof] = ACTIONS(2712),
    [anon_sym_ATimport] = ACTIONS(2714),
    [aux_sym_preproc_undef_token1] = ACTIONS(2712),
    [anon_sym_POUND] = ACTIONS(2712),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2712),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2712),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2712),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2712),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2712),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2712),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2712),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_API_AVAILABLE] = ACTIONS(2712),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_API_DEPRECATED] = ACTIONS(2712),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2712),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym___deprecated_msg] = ACTIONS(2712),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2712),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2712),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2712),
    [anon_sym_ATprotocol] = ACTIONS(2714),
    [anon_sym_ATinterface] = ACTIONS(2714),
    [anon_sym_ATimplementation] = ACTIONS(2714),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2714),
    [anon_sym_ATsynthesize] = ACTIONS(2714),
    [anon_sym_ATdynamic] = ACTIONS(2714),
    [anon_sym_ATtry] = ACTIONS(2714),
    [anon_sym___try] = ACTIONS(2712),
    [anon_sym_ATthrow] = ACTIONS(2714),
    [anon_sym_ATselector] = ACTIONS(2714),
    [anon_sym_ATavailable] = ACTIONS(2714),
    [anon_sym___builtin_available] = ACTIONS(2712),
    [anon_sym_va_arg] = ACTIONS(2712),
    [anon_sym_ATencode] = ACTIONS(2714),
    [anon_sym_ATsynchronized] = ACTIONS(2714),
    [anon_sym_BOOL] = ACTIONS(2712),
    [anon_sym_IMP] = ACTIONS(2712),
    [anon_sym_SEL] = ACTIONS(2712),
    [anon_sym_Class] = ACTIONS(2712),
    [anon_sym_id] = ACTIONS(2712),
  },
  [STATE(313)] = {
    [sym_identifier] = ACTIONS(2716),
    [aux_sym_preproc_include_token1] = ACTIONS(2716),
    [aux_sym_preproc_include_token2] = ACTIONS(2716),
    [aux_sym_preproc_def_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token2] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2716),
    [aux_sym_preproc_else_token1] = ACTIONS(2716),
    [aux_sym_preproc_elif_token1] = ACTIONS(2716),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2716),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2716),
    [sym_preproc_directive] = ACTIONS(2716),
    [anon_sym_LPAREN2] = ACTIONS(2718),
    [anon_sym_BANG] = ACTIONS(2718),
    [anon_sym_TILDE] = ACTIONS(2718),
    [anon_sym_DASH] = ACTIONS(2716),
    [anon_sym_PLUS] = ACTIONS(2716),
    [anon_sym_STAR] = ACTIONS(2718),
    [anon_sym_CARET] = ACTIONS(2718),
    [anon_sym_AMP] = ACTIONS(2718),
    [anon_sym_SEMI] = ACTIONS(2718),
    [anon_sym___extension__] = ACTIONS(2716),
    [anon_sym_typedef] = ACTIONS(2716),
    [anon_sym_extern] = ACTIONS(2716),
    [anon_sym___attribute__] = ACTIONS(2716),
    [anon_sym___attribute] = ACTIONS(2716),
    [anon_sym_noreturn] = ACTIONS(2716),
    [anon_sym_LBRACK] = ACTIONS(2718),
    [anon_sym___declspec] = ACTIONS(2716),
    [anon_sym___cdecl] = ACTIONS(2716),
    [anon_sym___clrcall] = ACTIONS(2716),
    [anon_sym___stdcall] = ACTIONS(2716),
    [anon_sym___fastcall] = ACTIONS(2716),
    [anon_sym___thiscall] = ACTIONS(2716),
    [anon_sym___vectorcall] = ACTIONS(2716),
    [anon_sym_LBRACE] = ACTIONS(2718),
    [anon_sym_signed] = ACTIONS(2716),
    [anon_sym_unsigned] = ACTIONS(2716),
    [anon_sym_long] = ACTIONS(2716),
    [anon_sym_short] = ACTIONS(2716),
    [anon_sym_static] = ACTIONS(2716),
    [anon_sym_ATautoreleasepool] = ACTIONS(2718),
    [anon_sym_auto] = ACTIONS(2716),
    [anon_sym_register] = ACTIONS(2716),
    [anon_sym_inline] = ACTIONS(2716),
    [anon_sym___inline] = ACTIONS(2716),
    [anon_sym___inline__] = ACTIONS(2716),
    [anon_sym___forceinline] = ACTIONS(2716),
    [anon_sym_thread_local] = ACTIONS(2716),
    [anon_sym___thread] = ACTIONS(2716),
    [anon_sym_CG_EXTERN] = ACTIONS(2716),
    [anon_sym_CG_INLINE] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2716),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2716),
    [anon_sym_IBOutlet] = ACTIONS(2716),
    [anon_sym_IBInspectable] = ACTIONS(2716),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2716),
    [anon_sym_NS_INLINE] = ACTIONS(2716),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2716),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2716),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2716),
    [anon_sym_const] = ACTIONS(2716),
    [anon_sym_constexpr] = ACTIONS(2716),
    [anon_sym_volatile] = ACTIONS(2716),
    [anon_sym_restrict] = ACTIONS(2716),
    [anon_sym___restrict__] = ACTIONS(2716),
    [anon_sym__Atomic] = ACTIONS(2716),
    [anon_sym__Noreturn] = ACTIONS(2716),
    [anon_sym__Nonnull] = ACTIONS(2716),
    [anon_sym_nullable] = ACTIONS(2716),
    [anon_sym__Complex] = ACTIONS(2716),
    [anon_sym__Nullable] = ACTIONS(2716),
    [anon_sym__Nullable_result] = ACTIONS(2716),
    [anon_sym__Null_unspecified] = ACTIONS(2716),
    [anon_sym___autoreleasing] = ACTIONS(2716),
    [anon_sym___block] = ACTIONS(2716),
    [anon_sym___bridge] = ACTIONS(2716),
    [anon_sym___bridge_retained] = ACTIONS(2716),
    [anon_sym___bridge_transfer] = ACTIONS(2716),
    [anon_sym___complex] = ACTIONS(2716),
    [anon_sym___const] = ACTIONS(2716),
    [anon_sym___imag] = ACTIONS(2716),
    [anon_sym___kindof] = ACTIONS(2716),
    [anon_sym___nonnull] = ACTIONS(2716),
    [anon_sym___nullable] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2716),
    [anon_sym___real] = ACTIONS(2716),
    [anon_sym___strong] = ACTIONS(2716),
    [anon_sym___unsafe_unretained] = ACTIONS(2716),
    [anon_sym___unused] = ACTIONS(2716),
    [anon_sym___weak] = ACTIONS(2716),
    [anon_sym_alignas] = ACTIONS(2716),
    [anon_sym__Alignas] = ACTIONS(2716),
    [sym_primitive_type] = ACTIONS(2716),
    [anon_sym_enum] = ACTIONS(2716),
    [anon_sym_struct] = ACTIONS(2716),
    [anon_sym_union] = ACTIONS(2716),
    [anon_sym_if] = ACTIONS(2716),
    [anon_sym_switch] = ACTIONS(2716),
    [anon_sym_case] = ACTIONS(2716),
    [anon_sym_default] = ACTIONS(2716),
    [anon_sym_while] = ACTIONS(2716),
    [anon_sym_do] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2716),
    [anon_sym_in] = ACTIONS(2716),
    [anon_sym_return] = ACTIONS(2716),
    [anon_sym_break] = ACTIONS(2716),
    [anon_sym_continue] = ACTIONS(2716),
    [anon_sym_goto] = ACTIONS(2716),
    [anon_sym_DASH_DASH] = ACTIONS(2718),
    [anon_sym_PLUS_PLUS] = ACTIONS(2718),
    [anon_sym_sizeof] = ACTIONS(2716),
    [anon_sym___alignof__] = ACTIONS(2716),
    [anon_sym___alignof] = ACTIONS(2716),
    [anon_sym__alignof] = ACTIONS(2716),
    [anon_sym_alignof] = ACTIONS(2716),
    [anon_sym__Alignof] = ACTIONS(2716),
    [anon_sym_offsetof] = ACTIONS(2716),
    [anon_sym__Generic] = ACTIONS(2716),
    [anon_sym_asm] = ACTIONS(2716),
    [anon_sym___asm__] = ACTIONS(2716),
    [anon_sym___asm] = ACTIONS(2716),
    [sym_number_literal] = ACTIONS(2718),
    [anon_sym_L_SQUOTE] = ACTIONS(2718),
    [anon_sym_u_SQUOTE] = ACTIONS(2718),
    [anon_sym_U_SQUOTE] = ACTIONS(2718),
    [anon_sym_u8_SQUOTE] = ACTIONS(2718),
    [anon_sym_SQUOTE] = ACTIONS(2718),
    [anon_sym_AT] = ACTIONS(2716),
    [anon_sym_DQUOTE] = ACTIONS(2718),
    [anon_sym_L_DQUOTE] = ACTIONS(2718),
    [anon_sym_u_DQUOTE] = ACTIONS(2718),
    [anon_sym_U_DQUOTE] = ACTIONS(2718),
    [anon_sym_u8_DQUOTE] = ACTIONS(2718),
    [sym_true] = ACTIONS(2716),
    [sym_false] = ACTIONS(2716),
    [anon_sym_NULL] = ACTIONS(2716),
    [anon_sym_nullptr] = ACTIONS(2716),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2716),
    [anon_sym___typeof] = ACTIONS(2716),
    [anon_sym_typeof] = ACTIONS(2716),
    [anon_sym_ATimport] = ACTIONS(2718),
    [aux_sym_preproc_undef_token1] = ACTIONS(2716),
    [anon_sym_POUND] = ACTIONS(2716),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2716),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2716),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2716),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_API_AVAILABLE] = ACTIONS(2716),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_API_DEPRECATED] = ACTIONS(2716),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym___deprecated_msg] = ACTIONS(2716),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2716),
    [anon_sym_ATprotocol] = ACTIONS(2718),
    [anon_sym_ATinterface] = ACTIONS(2718),
    [anon_sym_ATimplementation] = ACTIONS(2718),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2718),
    [anon_sym_ATsynthesize] = ACTIONS(2718),
    [anon_sym_ATdynamic] = ACTIONS(2718),
    [anon_sym_ATtry] = ACTIONS(2718),
    [anon_sym___try] = ACTIONS(2716),
    [anon_sym_ATthrow] = ACTIONS(2718),
    [anon_sym_ATselector] = ACTIONS(2718),
    [anon_sym_ATavailable] = ACTIONS(2718),
    [anon_sym___builtin_available] = ACTIONS(2716),
    [anon_sym_va_arg] = ACTIONS(2716),
    [anon_sym_ATencode] = ACTIONS(2718),
    [anon_sym_ATsynchronized] = ACTIONS(2718),
    [anon_sym_BOOL] = ACTIONS(2716),
    [anon_sym_IMP] = ACTIONS(2716),
    [anon_sym_SEL] = ACTIONS(2716),
    [anon_sym_Class] = ACTIONS(2716),
    [anon_sym_id] = ACTIONS(2716),
  },
  [STATE(314)] = {
    [sym_identifier] = ACTIONS(2720),
    [aux_sym_preproc_include_token1] = ACTIONS(2720),
    [aux_sym_preproc_include_token2] = ACTIONS(2720),
    [aux_sym_preproc_def_token1] = ACTIONS(2720),
    [aux_sym_preproc_if_token1] = ACTIONS(2720),
    [aux_sym_preproc_if_token2] = ACTIONS(2720),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2720),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2720),
    [aux_sym_preproc_else_token1] = ACTIONS(2720),
    [aux_sym_preproc_elif_token1] = ACTIONS(2720),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2720),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2720),
    [sym_preproc_directive] = ACTIONS(2720),
    [anon_sym_LPAREN2] = ACTIONS(2722),
    [anon_sym_BANG] = ACTIONS(2722),
    [anon_sym_TILDE] = ACTIONS(2722),
    [anon_sym_DASH] = ACTIONS(2720),
    [anon_sym_PLUS] = ACTIONS(2720),
    [anon_sym_STAR] = ACTIONS(2722),
    [anon_sym_CARET] = ACTIONS(2722),
    [anon_sym_AMP] = ACTIONS(2722),
    [anon_sym_SEMI] = ACTIONS(2722),
    [anon_sym___extension__] = ACTIONS(2720),
    [anon_sym_typedef] = ACTIONS(2720),
    [anon_sym_extern] = ACTIONS(2720),
    [anon_sym___attribute__] = ACTIONS(2720),
    [anon_sym___attribute] = ACTIONS(2720),
    [anon_sym_noreturn] = ACTIONS(2720),
    [anon_sym_LBRACK] = ACTIONS(2722),
    [anon_sym___declspec] = ACTIONS(2720),
    [anon_sym___cdecl] = ACTIONS(2720),
    [anon_sym___clrcall] = ACTIONS(2720),
    [anon_sym___stdcall] = ACTIONS(2720),
    [anon_sym___fastcall] = ACTIONS(2720),
    [anon_sym___thiscall] = ACTIONS(2720),
    [anon_sym___vectorcall] = ACTIONS(2720),
    [anon_sym_LBRACE] = ACTIONS(2722),
    [anon_sym_signed] = ACTIONS(2720),
    [anon_sym_unsigned] = ACTIONS(2720),
    [anon_sym_long] = ACTIONS(2720),
    [anon_sym_short] = ACTIONS(2720),
    [anon_sym_static] = ACTIONS(2720),
    [anon_sym_ATautoreleasepool] = ACTIONS(2722),
    [anon_sym_auto] = ACTIONS(2720),
    [anon_sym_register] = ACTIONS(2720),
    [anon_sym_inline] = ACTIONS(2720),
    [anon_sym___inline] = ACTIONS(2720),
    [anon_sym___inline__] = ACTIONS(2720),
    [anon_sym___forceinline] = ACTIONS(2720),
    [anon_sym_thread_local] = ACTIONS(2720),
    [anon_sym___thread] = ACTIONS(2720),
    [anon_sym_CG_EXTERN] = ACTIONS(2720),
    [anon_sym_CG_INLINE] = ACTIONS(2720),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2720),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2720),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2720),
    [anon_sym_IBOutlet] = ACTIONS(2720),
    [anon_sym_IBInspectable] = ACTIONS(2720),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2720),
    [anon_sym_NS_INLINE] = ACTIONS(2720),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2720),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2720),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2720),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2720),
    [anon_sym_const] = ACTIONS(2720),
    [anon_sym_constexpr] = ACTIONS(2720),
    [anon_sym_volatile] = ACTIONS(2720),
    [anon_sym_restrict] = ACTIONS(2720),
    [anon_sym___restrict__] = ACTIONS(2720),
    [anon_sym__Atomic] = ACTIONS(2720),
    [anon_sym__Noreturn] = ACTIONS(2720),
    [anon_sym__Nonnull] = ACTIONS(2720),
    [anon_sym_nullable] = ACTIONS(2720),
    [anon_sym__Complex] = ACTIONS(2720),
    [anon_sym__Nullable] = ACTIONS(2720),
    [anon_sym__Nullable_result] = ACTIONS(2720),
    [anon_sym__Null_unspecified] = ACTIONS(2720),
    [anon_sym___autoreleasing] = ACTIONS(2720),
    [anon_sym___block] = ACTIONS(2720),
    [anon_sym___bridge] = ACTIONS(2720),
    [anon_sym___bridge_retained] = ACTIONS(2720),
    [anon_sym___bridge_transfer] = ACTIONS(2720),
    [anon_sym___complex] = ACTIONS(2720),
    [anon_sym___const] = ACTIONS(2720),
    [anon_sym___imag] = ACTIONS(2720),
    [anon_sym___kindof] = ACTIONS(2720),
    [anon_sym___nonnull] = ACTIONS(2720),
    [anon_sym___nullable] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2720),
    [anon_sym___real] = ACTIONS(2720),
    [anon_sym___strong] = ACTIONS(2720),
    [anon_sym___unsafe_unretained] = ACTIONS(2720),
    [anon_sym___unused] = ACTIONS(2720),
    [anon_sym___weak] = ACTIONS(2720),
    [anon_sym_alignas] = ACTIONS(2720),
    [anon_sym__Alignas] = ACTIONS(2720),
    [sym_primitive_type] = ACTIONS(2720),
    [anon_sym_enum] = ACTIONS(2720),
    [anon_sym_struct] = ACTIONS(2720),
    [anon_sym_union] = ACTIONS(2720),
    [anon_sym_if] = ACTIONS(2720),
    [anon_sym_switch] = ACTIONS(2720),
    [anon_sym_case] = ACTIONS(2720),
    [anon_sym_default] = ACTIONS(2720),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_do] = ACTIONS(2720),
    [anon_sym_for] = ACTIONS(2720),
    [anon_sym_in] = ACTIONS(2720),
    [anon_sym_return] = ACTIONS(2720),
    [anon_sym_break] = ACTIONS(2720),
    [anon_sym_continue] = ACTIONS(2720),
    [anon_sym_goto] = ACTIONS(2720),
    [anon_sym_DASH_DASH] = ACTIONS(2722),
    [anon_sym_PLUS_PLUS] = ACTIONS(2722),
    [anon_sym_sizeof] = ACTIONS(2720),
    [anon_sym___alignof__] = ACTIONS(2720),
    [anon_sym___alignof] = ACTIONS(2720),
    [anon_sym__alignof] = ACTIONS(2720),
    [anon_sym_alignof] = ACTIONS(2720),
    [anon_sym__Alignof] = ACTIONS(2720),
    [anon_sym_offsetof] = ACTIONS(2720),
    [anon_sym__Generic] = ACTIONS(2720),
    [anon_sym_asm] = ACTIONS(2720),
    [anon_sym___asm__] = ACTIONS(2720),
    [anon_sym___asm] = ACTIONS(2720),
    [sym_number_literal] = ACTIONS(2722),
    [anon_sym_L_SQUOTE] = ACTIONS(2722),
    [anon_sym_u_SQUOTE] = ACTIONS(2722),
    [anon_sym_U_SQUOTE] = ACTIONS(2722),
    [anon_sym_u8_SQUOTE] = ACTIONS(2722),
    [anon_sym_SQUOTE] = ACTIONS(2722),
    [anon_sym_AT] = ACTIONS(2720),
    [anon_sym_DQUOTE] = ACTIONS(2722),
    [anon_sym_L_DQUOTE] = ACTIONS(2722),
    [anon_sym_u_DQUOTE] = ACTIONS(2722),
    [anon_sym_U_DQUOTE] = ACTIONS(2722),
    [anon_sym_u8_DQUOTE] = ACTIONS(2722),
    [sym_true] = ACTIONS(2720),
    [sym_false] = ACTIONS(2720),
    [anon_sym_NULL] = ACTIONS(2720),
    [anon_sym_nullptr] = ACTIONS(2720),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2720),
    [anon_sym___typeof] = ACTIONS(2720),
    [anon_sym_typeof] = ACTIONS(2720),
    [anon_sym_ATimport] = ACTIONS(2722),
    [aux_sym_preproc_undef_token1] = ACTIONS(2720),
    [anon_sym_POUND] = ACTIONS(2720),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2720),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2720),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2720),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2720),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2720),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2720),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2720),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_API_AVAILABLE] = ACTIONS(2720),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_API_DEPRECATED] = ACTIONS(2720),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2720),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym___deprecated_msg] = ACTIONS(2720),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2720),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2720),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2720),
    [anon_sym_ATprotocol] = ACTIONS(2722),
    [anon_sym_ATinterface] = ACTIONS(2722),
    [anon_sym_ATimplementation] = ACTIONS(2722),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2722),
    [anon_sym_ATsynthesize] = ACTIONS(2722),
    [anon_sym_ATdynamic] = ACTIONS(2722),
    [anon_sym_ATtry] = ACTIONS(2722),
    [anon_sym___try] = ACTIONS(2720),
    [anon_sym_ATthrow] = ACTIONS(2722),
    [anon_sym_ATselector] = ACTIONS(2722),
    [anon_sym_ATavailable] = ACTIONS(2722),
    [anon_sym___builtin_available] = ACTIONS(2720),
    [anon_sym_va_arg] = ACTIONS(2720),
    [anon_sym_ATencode] = ACTIONS(2722),
    [anon_sym_ATsynchronized] = ACTIONS(2722),
    [anon_sym_BOOL] = ACTIONS(2720),
    [anon_sym_IMP] = ACTIONS(2720),
    [anon_sym_SEL] = ACTIONS(2720),
    [anon_sym_Class] = ACTIONS(2720),
    [anon_sym_id] = ACTIONS(2720),
  },
  [STATE(315)] = {
    [sym_identifier] = ACTIONS(2724),
    [aux_sym_preproc_include_token1] = ACTIONS(2724),
    [aux_sym_preproc_include_token2] = ACTIONS(2724),
    [aux_sym_preproc_def_token1] = ACTIONS(2724),
    [aux_sym_preproc_if_token1] = ACTIONS(2724),
    [aux_sym_preproc_if_token2] = ACTIONS(2724),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2724),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2724),
    [aux_sym_preproc_else_token1] = ACTIONS(2724),
    [aux_sym_preproc_elif_token1] = ACTIONS(2724),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2724),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2724),
    [sym_preproc_directive] = ACTIONS(2724),
    [anon_sym_LPAREN2] = ACTIONS(2726),
    [anon_sym_BANG] = ACTIONS(2726),
    [anon_sym_TILDE] = ACTIONS(2726),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2726),
    [anon_sym_CARET] = ACTIONS(2726),
    [anon_sym_AMP] = ACTIONS(2726),
    [anon_sym_SEMI] = ACTIONS(2726),
    [anon_sym___extension__] = ACTIONS(2724),
    [anon_sym_typedef] = ACTIONS(2724),
    [anon_sym_extern] = ACTIONS(2724),
    [anon_sym___attribute__] = ACTIONS(2724),
    [anon_sym___attribute] = ACTIONS(2724),
    [anon_sym_noreturn] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2726),
    [anon_sym___declspec] = ACTIONS(2724),
    [anon_sym___cdecl] = ACTIONS(2724),
    [anon_sym___clrcall] = ACTIONS(2724),
    [anon_sym___stdcall] = ACTIONS(2724),
    [anon_sym___fastcall] = ACTIONS(2724),
    [anon_sym___thiscall] = ACTIONS(2724),
    [anon_sym___vectorcall] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2726),
    [anon_sym_signed] = ACTIONS(2724),
    [anon_sym_unsigned] = ACTIONS(2724),
    [anon_sym_long] = ACTIONS(2724),
    [anon_sym_short] = ACTIONS(2724),
    [anon_sym_static] = ACTIONS(2724),
    [anon_sym_ATautoreleasepool] = ACTIONS(2726),
    [anon_sym_auto] = ACTIONS(2724),
    [anon_sym_register] = ACTIONS(2724),
    [anon_sym_inline] = ACTIONS(2724),
    [anon_sym___inline] = ACTIONS(2724),
    [anon_sym___inline__] = ACTIONS(2724),
    [anon_sym___forceinline] = ACTIONS(2724),
    [anon_sym_thread_local] = ACTIONS(2724),
    [anon_sym___thread] = ACTIONS(2724),
    [anon_sym_CG_EXTERN] = ACTIONS(2724),
    [anon_sym_CG_INLINE] = ACTIONS(2724),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2724),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2724),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2724),
    [anon_sym_IBOutlet] = ACTIONS(2724),
    [anon_sym_IBInspectable] = ACTIONS(2724),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2724),
    [anon_sym_NS_INLINE] = ACTIONS(2724),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2724),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2724),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2724),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2724),
    [anon_sym_const] = ACTIONS(2724),
    [anon_sym_constexpr] = ACTIONS(2724),
    [anon_sym_volatile] = ACTIONS(2724),
    [anon_sym_restrict] = ACTIONS(2724),
    [anon_sym___restrict__] = ACTIONS(2724),
    [anon_sym__Atomic] = ACTIONS(2724),
    [anon_sym__Noreturn] = ACTIONS(2724),
    [anon_sym__Nonnull] = ACTIONS(2724),
    [anon_sym_nullable] = ACTIONS(2724),
    [anon_sym__Complex] = ACTIONS(2724),
    [anon_sym__Nullable] = ACTIONS(2724),
    [anon_sym__Nullable_result] = ACTIONS(2724),
    [anon_sym__Null_unspecified] = ACTIONS(2724),
    [anon_sym___autoreleasing] = ACTIONS(2724),
    [anon_sym___block] = ACTIONS(2724),
    [anon_sym___bridge] = ACTIONS(2724),
    [anon_sym___bridge_retained] = ACTIONS(2724),
    [anon_sym___bridge_transfer] = ACTIONS(2724),
    [anon_sym___complex] = ACTIONS(2724),
    [anon_sym___const] = ACTIONS(2724),
    [anon_sym___imag] = ACTIONS(2724),
    [anon_sym___kindof] = ACTIONS(2724),
    [anon_sym___nonnull] = ACTIONS(2724),
    [anon_sym___nullable] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2724),
    [anon_sym___real] = ACTIONS(2724),
    [anon_sym___strong] = ACTIONS(2724),
    [anon_sym___unsafe_unretained] = ACTIONS(2724),
    [anon_sym___unused] = ACTIONS(2724),
    [anon_sym___weak] = ACTIONS(2724),
    [anon_sym_alignas] = ACTIONS(2724),
    [anon_sym__Alignas] = ACTIONS(2724),
    [sym_primitive_type] = ACTIONS(2724),
    [anon_sym_enum] = ACTIONS(2724),
    [anon_sym_struct] = ACTIONS(2724),
    [anon_sym_union] = ACTIONS(2724),
    [anon_sym_if] = ACTIONS(2724),
    [anon_sym_switch] = ACTIONS(2724),
    [anon_sym_case] = ACTIONS(2724),
    [anon_sym_default] = ACTIONS(2724),
    [anon_sym_while] = ACTIONS(2724),
    [anon_sym_do] = ACTIONS(2724),
    [anon_sym_for] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_return] = ACTIONS(2724),
    [anon_sym_break] = ACTIONS(2724),
    [anon_sym_continue] = ACTIONS(2724),
    [anon_sym_goto] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2726),
    [anon_sym_PLUS_PLUS] = ACTIONS(2726),
    [anon_sym_sizeof] = ACTIONS(2724),
    [anon_sym___alignof__] = ACTIONS(2724),
    [anon_sym___alignof] = ACTIONS(2724),
    [anon_sym__alignof] = ACTIONS(2724),
    [anon_sym_alignof] = ACTIONS(2724),
    [anon_sym__Alignof] = ACTIONS(2724),
    [anon_sym_offsetof] = ACTIONS(2724),
    [anon_sym__Generic] = ACTIONS(2724),
    [anon_sym_asm] = ACTIONS(2724),
    [anon_sym___asm__] = ACTIONS(2724),
    [anon_sym___asm] = ACTIONS(2724),
    [sym_number_literal] = ACTIONS(2726),
    [anon_sym_L_SQUOTE] = ACTIONS(2726),
    [anon_sym_u_SQUOTE] = ACTIONS(2726),
    [anon_sym_U_SQUOTE] = ACTIONS(2726),
    [anon_sym_u8_SQUOTE] = ACTIONS(2726),
    [anon_sym_SQUOTE] = ACTIONS(2726),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2726),
    [anon_sym_L_DQUOTE] = ACTIONS(2726),
    [anon_sym_u_DQUOTE] = ACTIONS(2726),
    [anon_sym_U_DQUOTE] = ACTIONS(2726),
    [anon_sym_u8_DQUOTE] = ACTIONS(2726),
    [sym_true] = ACTIONS(2724),
    [sym_false] = ACTIONS(2724),
    [anon_sym_NULL] = ACTIONS(2724),
    [anon_sym_nullptr] = ACTIONS(2724),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2724),
    [anon_sym___typeof] = ACTIONS(2724),
    [anon_sym_typeof] = ACTIONS(2724),
    [anon_sym_ATimport] = ACTIONS(2726),
    [aux_sym_preproc_undef_token1] = ACTIONS(2724),
    [anon_sym_POUND] = ACTIONS(2724),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2724),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2724),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2724),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2724),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2724),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2724),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2724),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_API_AVAILABLE] = ACTIONS(2724),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_API_DEPRECATED] = ACTIONS(2724),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2724),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym___deprecated_msg] = ACTIONS(2724),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2724),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2724),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2724),
    [anon_sym_ATprotocol] = ACTIONS(2726),
    [anon_sym_ATinterface] = ACTIONS(2726),
    [anon_sym_ATimplementation] = ACTIONS(2726),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2726),
    [anon_sym_ATsynthesize] = ACTIONS(2726),
    [anon_sym_ATdynamic] = ACTIONS(2726),
    [anon_sym_ATtry] = ACTIONS(2726),
    [anon_sym___try] = ACTIONS(2724),
    [anon_sym_ATthrow] = ACTIONS(2726),
    [anon_sym_ATselector] = ACTIONS(2726),
    [anon_sym_ATavailable] = ACTIONS(2726),
    [anon_sym___builtin_available] = ACTIONS(2724),
    [anon_sym_va_arg] = ACTIONS(2724),
    [anon_sym_ATencode] = ACTIONS(2726),
    [anon_sym_ATsynchronized] = ACTIONS(2726),
    [anon_sym_BOOL] = ACTIONS(2724),
    [anon_sym_IMP] = ACTIONS(2724),
    [anon_sym_SEL] = ACTIONS(2724),
    [anon_sym_Class] = ACTIONS(2724),
    [anon_sym_id] = ACTIONS(2724),
  },
  [STATE(316)] = {
    [sym_identifier] = ACTIONS(2728),
    [aux_sym_preproc_include_token1] = ACTIONS(2728),
    [aux_sym_preproc_include_token2] = ACTIONS(2728),
    [aux_sym_preproc_def_token1] = ACTIONS(2728),
    [aux_sym_preproc_if_token1] = ACTIONS(2728),
    [aux_sym_preproc_if_token2] = ACTIONS(2728),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2728),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2728),
    [aux_sym_preproc_else_token1] = ACTIONS(2728),
    [aux_sym_preproc_elif_token1] = ACTIONS(2728),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2728),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2728),
    [sym_preproc_directive] = ACTIONS(2728),
    [anon_sym_LPAREN2] = ACTIONS(2730),
    [anon_sym_BANG] = ACTIONS(2730),
    [anon_sym_TILDE] = ACTIONS(2730),
    [anon_sym_DASH] = ACTIONS(2728),
    [anon_sym_PLUS] = ACTIONS(2728),
    [anon_sym_STAR] = ACTIONS(2730),
    [anon_sym_CARET] = ACTIONS(2730),
    [anon_sym_AMP] = ACTIONS(2730),
    [anon_sym_SEMI] = ACTIONS(2730),
    [anon_sym___extension__] = ACTIONS(2728),
    [anon_sym_typedef] = ACTIONS(2728),
    [anon_sym_extern] = ACTIONS(2728),
    [anon_sym___attribute__] = ACTIONS(2728),
    [anon_sym___attribute] = ACTIONS(2728),
    [anon_sym_noreturn] = ACTIONS(2728),
    [anon_sym_LBRACK] = ACTIONS(2730),
    [anon_sym___declspec] = ACTIONS(2728),
    [anon_sym___cdecl] = ACTIONS(2728),
    [anon_sym___clrcall] = ACTIONS(2728),
    [anon_sym___stdcall] = ACTIONS(2728),
    [anon_sym___fastcall] = ACTIONS(2728),
    [anon_sym___thiscall] = ACTIONS(2728),
    [anon_sym___vectorcall] = ACTIONS(2728),
    [anon_sym_LBRACE] = ACTIONS(2730),
    [anon_sym_signed] = ACTIONS(2728),
    [anon_sym_unsigned] = ACTIONS(2728),
    [anon_sym_long] = ACTIONS(2728),
    [anon_sym_short] = ACTIONS(2728),
    [anon_sym_static] = ACTIONS(2728),
    [anon_sym_ATautoreleasepool] = ACTIONS(2730),
    [anon_sym_auto] = ACTIONS(2728),
    [anon_sym_register] = ACTIONS(2728),
    [anon_sym_inline] = ACTIONS(2728),
    [anon_sym___inline] = ACTIONS(2728),
    [anon_sym___inline__] = ACTIONS(2728),
    [anon_sym___forceinline] = ACTIONS(2728),
    [anon_sym_thread_local] = ACTIONS(2728),
    [anon_sym___thread] = ACTIONS(2728),
    [anon_sym_CG_EXTERN] = ACTIONS(2728),
    [anon_sym_CG_INLINE] = ACTIONS(2728),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2728),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2728),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2728),
    [anon_sym_IBOutlet] = ACTIONS(2728),
    [anon_sym_IBInspectable] = ACTIONS(2728),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2728),
    [anon_sym_NS_INLINE] = ACTIONS(2728),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2728),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2728),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2728),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2728),
    [anon_sym_const] = ACTIONS(2728),
    [anon_sym_constexpr] = ACTIONS(2728),
    [anon_sym_volatile] = ACTIONS(2728),
    [anon_sym_restrict] = ACTIONS(2728),
    [anon_sym___restrict__] = ACTIONS(2728),
    [anon_sym__Atomic] = ACTIONS(2728),
    [anon_sym__Noreturn] = ACTIONS(2728),
    [anon_sym__Nonnull] = ACTIONS(2728),
    [anon_sym_nullable] = ACTIONS(2728),
    [anon_sym__Complex] = ACTIONS(2728),
    [anon_sym__Nullable] = ACTIONS(2728),
    [anon_sym__Nullable_result] = ACTIONS(2728),
    [anon_sym__Null_unspecified] = ACTIONS(2728),
    [anon_sym___autoreleasing] = ACTIONS(2728),
    [anon_sym___block] = ACTIONS(2728),
    [anon_sym___bridge] = ACTIONS(2728),
    [anon_sym___bridge_retained] = ACTIONS(2728),
    [anon_sym___bridge_transfer] = ACTIONS(2728),
    [anon_sym___complex] = ACTIONS(2728),
    [anon_sym___const] = ACTIONS(2728),
    [anon_sym___imag] = ACTIONS(2728),
    [anon_sym___kindof] = ACTIONS(2728),
    [anon_sym___nonnull] = ACTIONS(2728),
    [anon_sym___nullable] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2728),
    [anon_sym___real] = ACTIONS(2728),
    [anon_sym___strong] = ACTIONS(2728),
    [anon_sym___unsafe_unretained] = ACTIONS(2728),
    [anon_sym___unused] = ACTIONS(2728),
    [anon_sym___weak] = ACTIONS(2728),
    [anon_sym_alignas] = ACTIONS(2728),
    [anon_sym__Alignas] = ACTIONS(2728),
    [sym_primitive_type] = ACTIONS(2728),
    [anon_sym_enum] = ACTIONS(2728),
    [anon_sym_struct] = ACTIONS(2728),
    [anon_sym_union] = ACTIONS(2728),
    [anon_sym_if] = ACTIONS(2728),
    [anon_sym_switch] = ACTIONS(2728),
    [anon_sym_case] = ACTIONS(2728),
    [anon_sym_default] = ACTIONS(2728),
    [anon_sym_while] = ACTIONS(2728),
    [anon_sym_do] = ACTIONS(2728),
    [anon_sym_for] = ACTIONS(2728),
    [anon_sym_in] = ACTIONS(2728),
    [anon_sym_return] = ACTIONS(2728),
    [anon_sym_break] = ACTIONS(2728),
    [anon_sym_continue] = ACTIONS(2728),
    [anon_sym_goto] = ACTIONS(2728),
    [anon_sym_DASH_DASH] = ACTIONS(2730),
    [anon_sym_PLUS_PLUS] = ACTIONS(2730),
    [anon_sym_sizeof] = ACTIONS(2728),
    [anon_sym___alignof__] = ACTIONS(2728),
    [anon_sym___alignof] = ACTIONS(2728),
    [anon_sym__alignof] = ACTIONS(2728),
    [anon_sym_alignof] = ACTIONS(2728),
    [anon_sym__Alignof] = ACTIONS(2728),
    [anon_sym_offsetof] = ACTIONS(2728),
    [anon_sym__Generic] = ACTIONS(2728),
    [anon_sym_asm] = ACTIONS(2728),
    [anon_sym___asm__] = ACTIONS(2728),
    [anon_sym___asm] = ACTIONS(2728),
    [sym_number_literal] = ACTIONS(2730),
    [anon_sym_L_SQUOTE] = ACTIONS(2730),
    [anon_sym_u_SQUOTE] = ACTIONS(2730),
    [anon_sym_U_SQUOTE] = ACTIONS(2730),
    [anon_sym_u8_SQUOTE] = ACTIONS(2730),
    [anon_sym_SQUOTE] = ACTIONS(2730),
    [anon_sym_AT] = ACTIONS(2728),
    [anon_sym_DQUOTE] = ACTIONS(2730),
    [anon_sym_L_DQUOTE] = ACTIONS(2730),
    [anon_sym_u_DQUOTE] = ACTIONS(2730),
    [anon_sym_U_DQUOTE] = ACTIONS(2730),
    [anon_sym_u8_DQUOTE] = ACTIONS(2730),
    [sym_true] = ACTIONS(2728),
    [sym_false] = ACTIONS(2728),
    [anon_sym_NULL] = ACTIONS(2728),
    [anon_sym_nullptr] = ACTIONS(2728),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2728),
    [anon_sym___typeof] = ACTIONS(2728),
    [anon_sym_typeof] = ACTIONS(2728),
    [anon_sym_ATimport] = ACTIONS(2730),
    [aux_sym_preproc_undef_token1] = ACTIONS(2728),
    [anon_sym_POUND] = ACTIONS(2728),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2728),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2728),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2728),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2728),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2728),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2728),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2728),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_API_AVAILABLE] = ACTIONS(2728),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_API_DEPRECATED] = ACTIONS(2728),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2728),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym___deprecated_msg] = ACTIONS(2728),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2728),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2728),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2728),
    [anon_sym_ATprotocol] = ACTIONS(2730),
    [anon_sym_ATinterface] = ACTIONS(2730),
    [anon_sym_ATimplementation] = ACTIONS(2730),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2730),
    [anon_sym_ATsynthesize] = ACTIONS(2730),
    [anon_sym_ATdynamic] = ACTIONS(2730),
    [anon_sym_ATtry] = ACTIONS(2730),
    [anon_sym___try] = ACTIONS(2728),
    [anon_sym_ATthrow] = ACTIONS(2730),
    [anon_sym_ATselector] = ACTIONS(2730),
    [anon_sym_ATavailable] = ACTIONS(2730),
    [anon_sym___builtin_available] = ACTIONS(2728),
    [anon_sym_va_arg] = ACTIONS(2728),
    [anon_sym_ATencode] = ACTIONS(2730),
    [anon_sym_ATsynchronized] = ACTIONS(2730),
    [anon_sym_BOOL] = ACTIONS(2728),
    [anon_sym_IMP] = ACTIONS(2728),
    [anon_sym_SEL] = ACTIONS(2728),
    [anon_sym_Class] = ACTIONS(2728),
    [anon_sym_id] = ACTIONS(2728),
  },
  [STATE(317)] = {
    [sym_identifier] = ACTIONS(2732),
    [aux_sym_preproc_include_token1] = ACTIONS(2732),
    [aux_sym_preproc_include_token2] = ACTIONS(2732),
    [aux_sym_preproc_def_token1] = ACTIONS(2732),
    [aux_sym_preproc_if_token1] = ACTIONS(2732),
    [aux_sym_preproc_if_token2] = ACTIONS(2732),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2732),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2732),
    [aux_sym_preproc_else_token1] = ACTIONS(2732),
    [aux_sym_preproc_elif_token1] = ACTIONS(2732),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2732),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2732),
    [sym_preproc_directive] = ACTIONS(2732),
    [anon_sym_LPAREN2] = ACTIONS(2734),
    [anon_sym_BANG] = ACTIONS(2734),
    [anon_sym_TILDE] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2732),
    [anon_sym_PLUS] = ACTIONS(2732),
    [anon_sym_STAR] = ACTIONS(2734),
    [anon_sym_CARET] = ACTIONS(2734),
    [anon_sym_AMP] = ACTIONS(2734),
    [anon_sym_SEMI] = ACTIONS(2734),
    [anon_sym___extension__] = ACTIONS(2732),
    [anon_sym_typedef] = ACTIONS(2732),
    [anon_sym_extern] = ACTIONS(2732),
    [anon_sym___attribute__] = ACTIONS(2732),
    [anon_sym___attribute] = ACTIONS(2732),
    [anon_sym_noreturn] = ACTIONS(2732),
    [anon_sym_LBRACK] = ACTIONS(2734),
    [anon_sym___declspec] = ACTIONS(2732),
    [anon_sym___cdecl] = ACTIONS(2732),
    [anon_sym___clrcall] = ACTIONS(2732),
    [anon_sym___stdcall] = ACTIONS(2732),
    [anon_sym___fastcall] = ACTIONS(2732),
    [anon_sym___thiscall] = ACTIONS(2732),
    [anon_sym___vectorcall] = ACTIONS(2732),
    [anon_sym_LBRACE] = ACTIONS(2734),
    [anon_sym_signed] = ACTIONS(2732),
    [anon_sym_unsigned] = ACTIONS(2732),
    [anon_sym_long] = ACTIONS(2732),
    [anon_sym_short] = ACTIONS(2732),
    [anon_sym_static] = ACTIONS(2732),
    [anon_sym_ATautoreleasepool] = ACTIONS(2734),
    [anon_sym_auto] = ACTIONS(2732),
    [anon_sym_register] = ACTIONS(2732),
    [anon_sym_inline] = ACTIONS(2732),
    [anon_sym___inline] = ACTIONS(2732),
    [anon_sym___inline__] = ACTIONS(2732),
    [anon_sym___forceinline] = ACTIONS(2732),
    [anon_sym_thread_local] = ACTIONS(2732),
    [anon_sym___thread] = ACTIONS(2732),
    [anon_sym_CG_EXTERN] = ACTIONS(2732),
    [anon_sym_CG_INLINE] = ACTIONS(2732),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2732),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2732),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2732),
    [anon_sym_IBOutlet] = ACTIONS(2732),
    [anon_sym_IBInspectable] = ACTIONS(2732),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2732),
    [anon_sym_NS_INLINE] = ACTIONS(2732),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2732),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2732),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2732),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2732),
    [anon_sym_const] = ACTIONS(2732),
    [anon_sym_constexpr] = ACTIONS(2732),
    [anon_sym_volatile] = ACTIONS(2732),
    [anon_sym_restrict] = ACTIONS(2732),
    [anon_sym___restrict__] = ACTIONS(2732),
    [anon_sym__Atomic] = ACTIONS(2732),
    [anon_sym__Noreturn] = ACTIONS(2732),
    [anon_sym__Nonnull] = ACTIONS(2732),
    [anon_sym_nullable] = ACTIONS(2732),
    [anon_sym__Complex] = ACTIONS(2732),
    [anon_sym__Nullable] = ACTIONS(2732),
    [anon_sym__Nullable_result] = ACTIONS(2732),
    [anon_sym__Null_unspecified] = ACTIONS(2732),
    [anon_sym___autoreleasing] = ACTIONS(2732),
    [anon_sym___block] = ACTIONS(2732),
    [anon_sym___bridge] = ACTIONS(2732),
    [anon_sym___bridge_retained] = ACTIONS(2732),
    [anon_sym___bridge_transfer] = ACTIONS(2732),
    [anon_sym___complex] = ACTIONS(2732),
    [anon_sym___const] = ACTIONS(2732),
    [anon_sym___imag] = ACTIONS(2732),
    [anon_sym___kindof] = ACTIONS(2732),
    [anon_sym___nonnull] = ACTIONS(2732),
    [anon_sym___nullable] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2732),
    [anon_sym___real] = ACTIONS(2732),
    [anon_sym___strong] = ACTIONS(2732),
    [anon_sym___unsafe_unretained] = ACTIONS(2732),
    [anon_sym___unused] = ACTIONS(2732),
    [anon_sym___weak] = ACTIONS(2732),
    [anon_sym_alignas] = ACTIONS(2732),
    [anon_sym__Alignas] = ACTIONS(2732),
    [sym_primitive_type] = ACTIONS(2732),
    [anon_sym_enum] = ACTIONS(2732),
    [anon_sym_struct] = ACTIONS(2732),
    [anon_sym_union] = ACTIONS(2732),
    [anon_sym_if] = ACTIONS(2732),
    [anon_sym_switch] = ACTIONS(2732),
    [anon_sym_case] = ACTIONS(2732),
    [anon_sym_default] = ACTIONS(2732),
    [anon_sym_while] = ACTIONS(2732),
    [anon_sym_do] = ACTIONS(2732),
    [anon_sym_for] = ACTIONS(2732),
    [anon_sym_in] = ACTIONS(2732),
    [anon_sym_return] = ACTIONS(2732),
    [anon_sym_break] = ACTIONS(2732),
    [anon_sym_continue] = ACTIONS(2732),
    [anon_sym_goto] = ACTIONS(2732),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_sizeof] = ACTIONS(2732),
    [anon_sym___alignof__] = ACTIONS(2732),
    [anon_sym___alignof] = ACTIONS(2732),
    [anon_sym__alignof] = ACTIONS(2732),
    [anon_sym_alignof] = ACTIONS(2732),
    [anon_sym__Alignof] = ACTIONS(2732),
    [anon_sym_offsetof] = ACTIONS(2732),
    [anon_sym__Generic] = ACTIONS(2732),
    [anon_sym_asm] = ACTIONS(2732),
    [anon_sym___asm__] = ACTIONS(2732),
    [anon_sym___asm] = ACTIONS(2732),
    [sym_number_literal] = ACTIONS(2734),
    [anon_sym_L_SQUOTE] = ACTIONS(2734),
    [anon_sym_u_SQUOTE] = ACTIONS(2734),
    [anon_sym_U_SQUOTE] = ACTIONS(2734),
    [anon_sym_u8_SQUOTE] = ACTIONS(2734),
    [anon_sym_SQUOTE] = ACTIONS(2734),
    [anon_sym_AT] = ACTIONS(2732),
    [anon_sym_DQUOTE] = ACTIONS(2734),
    [anon_sym_L_DQUOTE] = ACTIONS(2734),
    [anon_sym_u_DQUOTE] = ACTIONS(2734),
    [anon_sym_U_DQUOTE] = ACTIONS(2734),
    [anon_sym_u8_DQUOTE] = ACTIONS(2734),
    [sym_true] = ACTIONS(2732),
    [sym_false] = ACTIONS(2732),
    [anon_sym_NULL] = ACTIONS(2732),
    [anon_sym_nullptr] = ACTIONS(2732),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2732),
    [anon_sym___typeof] = ACTIONS(2732),
    [anon_sym_typeof] = ACTIONS(2732),
    [anon_sym_ATimport] = ACTIONS(2734),
    [aux_sym_preproc_undef_token1] = ACTIONS(2732),
    [anon_sym_POUND] = ACTIONS(2732),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2732),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2732),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2732),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2732),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2732),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2732),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2732),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_API_AVAILABLE] = ACTIONS(2732),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_API_DEPRECATED] = ACTIONS(2732),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2732),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym___deprecated_msg] = ACTIONS(2732),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2732),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2732),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2732),
    [anon_sym_ATprotocol] = ACTIONS(2734),
    [anon_sym_ATinterface] = ACTIONS(2734),
    [anon_sym_ATimplementation] = ACTIONS(2734),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2734),
    [anon_sym_ATsynthesize] = ACTIONS(2734),
    [anon_sym_ATdynamic] = ACTIONS(2734),
    [anon_sym_ATtry] = ACTIONS(2734),
    [anon_sym___try] = ACTIONS(2732),
    [anon_sym_ATthrow] = ACTIONS(2734),
    [anon_sym_ATselector] = ACTIONS(2734),
    [anon_sym_ATavailable] = ACTIONS(2734),
    [anon_sym___builtin_available] = ACTIONS(2732),
    [anon_sym_va_arg] = ACTIONS(2732),
    [anon_sym_ATencode] = ACTIONS(2734),
    [anon_sym_ATsynchronized] = ACTIONS(2734),
    [anon_sym_BOOL] = ACTIONS(2732),
    [anon_sym_IMP] = ACTIONS(2732),
    [anon_sym_SEL] = ACTIONS(2732),
    [anon_sym_Class] = ACTIONS(2732),
    [anon_sym_id] = ACTIONS(2732),
  },
  [STATE(318)] = {
    [sym_identifier] = ACTIONS(2736),
    [aux_sym_preproc_include_token1] = ACTIONS(2736),
    [aux_sym_preproc_include_token2] = ACTIONS(2736),
    [aux_sym_preproc_def_token1] = ACTIONS(2736),
    [aux_sym_preproc_if_token1] = ACTIONS(2736),
    [aux_sym_preproc_if_token2] = ACTIONS(2736),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2736),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2736),
    [aux_sym_preproc_else_token1] = ACTIONS(2736),
    [aux_sym_preproc_elif_token1] = ACTIONS(2736),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2736),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2736),
    [sym_preproc_directive] = ACTIONS(2736),
    [anon_sym_LPAREN2] = ACTIONS(2738),
    [anon_sym_BANG] = ACTIONS(2738),
    [anon_sym_TILDE] = ACTIONS(2738),
    [anon_sym_DASH] = ACTIONS(2736),
    [anon_sym_PLUS] = ACTIONS(2736),
    [anon_sym_STAR] = ACTIONS(2738),
    [anon_sym_CARET] = ACTIONS(2738),
    [anon_sym_AMP] = ACTIONS(2738),
    [anon_sym_SEMI] = ACTIONS(2738),
    [anon_sym___extension__] = ACTIONS(2736),
    [anon_sym_typedef] = ACTIONS(2736),
    [anon_sym_extern] = ACTIONS(2736),
    [anon_sym___attribute__] = ACTIONS(2736),
    [anon_sym___attribute] = ACTIONS(2736),
    [anon_sym_noreturn] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(2738),
    [anon_sym___declspec] = ACTIONS(2736),
    [anon_sym___cdecl] = ACTIONS(2736),
    [anon_sym___clrcall] = ACTIONS(2736),
    [anon_sym___stdcall] = ACTIONS(2736),
    [anon_sym___fastcall] = ACTIONS(2736),
    [anon_sym___thiscall] = ACTIONS(2736),
    [anon_sym___vectorcall] = ACTIONS(2736),
    [anon_sym_LBRACE] = ACTIONS(2738),
    [anon_sym_signed] = ACTIONS(2736),
    [anon_sym_unsigned] = ACTIONS(2736),
    [anon_sym_long] = ACTIONS(2736),
    [anon_sym_short] = ACTIONS(2736),
    [anon_sym_static] = ACTIONS(2736),
    [anon_sym_ATautoreleasepool] = ACTIONS(2738),
    [anon_sym_auto] = ACTIONS(2736),
    [anon_sym_register] = ACTIONS(2736),
    [anon_sym_inline] = ACTIONS(2736),
    [anon_sym___inline] = ACTIONS(2736),
    [anon_sym___inline__] = ACTIONS(2736),
    [anon_sym___forceinline] = ACTIONS(2736),
    [anon_sym_thread_local] = ACTIONS(2736),
    [anon_sym___thread] = ACTIONS(2736),
    [anon_sym_CG_EXTERN] = ACTIONS(2736),
    [anon_sym_CG_INLINE] = ACTIONS(2736),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2736),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2736),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2736),
    [anon_sym_IBOutlet] = ACTIONS(2736),
    [anon_sym_IBInspectable] = ACTIONS(2736),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2736),
    [anon_sym_NS_INLINE] = ACTIONS(2736),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2736),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2736),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2736),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2736),
    [anon_sym_const] = ACTIONS(2736),
    [anon_sym_constexpr] = ACTIONS(2736),
    [anon_sym_volatile] = ACTIONS(2736),
    [anon_sym_restrict] = ACTIONS(2736),
    [anon_sym___restrict__] = ACTIONS(2736),
    [anon_sym__Atomic] = ACTIONS(2736),
    [anon_sym__Noreturn] = ACTIONS(2736),
    [anon_sym__Nonnull] = ACTIONS(2736),
    [anon_sym_nullable] = ACTIONS(2736),
    [anon_sym__Complex] = ACTIONS(2736),
    [anon_sym__Nullable] = ACTIONS(2736),
    [anon_sym__Nullable_result] = ACTIONS(2736),
    [anon_sym__Null_unspecified] = ACTIONS(2736),
    [anon_sym___autoreleasing] = ACTIONS(2736),
    [anon_sym___block] = ACTIONS(2736),
    [anon_sym___bridge] = ACTIONS(2736),
    [anon_sym___bridge_retained] = ACTIONS(2736),
    [anon_sym___bridge_transfer] = ACTIONS(2736),
    [anon_sym___complex] = ACTIONS(2736),
    [anon_sym___const] = ACTIONS(2736),
    [anon_sym___imag] = ACTIONS(2736),
    [anon_sym___kindof] = ACTIONS(2736),
    [anon_sym___nonnull] = ACTIONS(2736),
    [anon_sym___nullable] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2736),
    [anon_sym___real] = ACTIONS(2736),
    [anon_sym___strong] = ACTIONS(2736),
    [anon_sym___unsafe_unretained] = ACTIONS(2736),
    [anon_sym___unused] = ACTIONS(2736),
    [anon_sym___weak] = ACTIONS(2736),
    [anon_sym_alignas] = ACTIONS(2736),
    [anon_sym__Alignas] = ACTIONS(2736),
    [sym_primitive_type] = ACTIONS(2736),
    [anon_sym_enum] = ACTIONS(2736),
    [anon_sym_struct] = ACTIONS(2736),
    [anon_sym_union] = ACTIONS(2736),
    [anon_sym_if] = ACTIONS(2736),
    [anon_sym_switch] = ACTIONS(2736),
    [anon_sym_case] = ACTIONS(2736),
    [anon_sym_default] = ACTIONS(2736),
    [anon_sym_while] = ACTIONS(2736),
    [anon_sym_do] = ACTIONS(2736),
    [anon_sym_for] = ACTIONS(2736),
    [anon_sym_in] = ACTIONS(2736),
    [anon_sym_return] = ACTIONS(2736),
    [anon_sym_break] = ACTIONS(2736),
    [anon_sym_continue] = ACTIONS(2736),
    [anon_sym_goto] = ACTIONS(2736),
    [anon_sym_DASH_DASH] = ACTIONS(2738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2738),
    [anon_sym_sizeof] = ACTIONS(2736),
    [anon_sym___alignof__] = ACTIONS(2736),
    [anon_sym___alignof] = ACTIONS(2736),
    [anon_sym__alignof] = ACTIONS(2736),
    [anon_sym_alignof] = ACTIONS(2736),
    [anon_sym__Alignof] = ACTIONS(2736),
    [anon_sym_offsetof] = ACTIONS(2736),
    [anon_sym__Generic] = ACTIONS(2736),
    [anon_sym_asm] = ACTIONS(2736),
    [anon_sym___asm__] = ACTIONS(2736),
    [anon_sym___asm] = ACTIONS(2736),
    [sym_number_literal] = ACTIONS(2738),
    [anon_sym_L_SQUOTE] = ACTIONS(2738),
    [anon_sym_u_SQUOTE] = ACTIONS(2738),
    [anon_sym_U_SQUOTE] = ACTIONS(2738),
    [anon_sym_u8_SQUOTE] = ACTIONS(2738),
    [anon_sym_SQUOTE] = ACTIONS(2738),
    [anon_sym_AT] = ACTIONS(2736),
    [anon_sym_DQUOTE] = ACTIONS(2738),
    [anon_sym_L_DQUOTE] = ACTIONS(2738),
    [anon_sym_u_DQUOTE] = ACTIONS(2738),
    [anon_sym_U_DQUOTE] = ACTIONS(2738),
    [anon_sym_u8_DQUOTE] = ACTIONS(2738),
    [sym_true] = ACTIONS(2736),
    [sym_false] = ACTIONS(2736),
    [anon_sym_NULL] = ACTIONS(2736),
    [anon_sym_nullptr] = ACTIONS(2736),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2736),
    [anon_sym___typeof] = ACTIONS(2736),
    [anon_sym_typeof] = ACTIONS(2736),
    [anon_sym_ATimport] = ACTIONS(2738),
    [aux_sym_preproc_undef_token1] = ACTIONS(2736),
    [anon_sym_POUND] = ACTIONS(2736),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2736),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2736),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2736),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2736),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2736),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2736),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2736),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_API_AVAILABLE] = ACTIONS(2736),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_API_DEPRECATED] = ACTIONS(2736),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2736),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym___deprecated_msg] = ACTIONS(2736),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2736),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2736),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2736),
    [anon_sym_ATprotocol] = ACTIONS(2738),
    [anon_sym_ATinterface] = ACTIONS(2738),
    [anon_sym_ATimplementation] = ACTIONS(2738),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2738),
    [anon_sym_ATsynthesize] = ACTIONS(2738),
    [anon_sym_ATdynamic] = ACTIONS(2738),
    [anon_sym_ATtry] = ACTIONS(2738),
    [anon_sym___try] = ACTIONS(2736),
    [anon_sym_ATthrow] = ACTIONS(2738),
    [anon_sym_ATselector] = ACTIONS(2738),
    [anon_sym_ATavailable] = ACTIONS(2738),
    [anon_sym___builtin_available] = ACTIONS(2736),
    [anon_sym_va_arg] = ACTIONS(2736),
    [anon_sym_ATencode] = ACTIONS(2738),
    [anon_sym_ATsynchronized] = ACTIONS(2738),
    [anon_sym_BOOL] = ACTIONS(2736),
    [anon_sym_IMP] = ACTIONS(2736),
    [anon_sym_SEL] = ACTIONS(2736),
    [anon_sym_Class] = ACTIONS(2736),
    [anon_sym_id] = ACTIONS(2736),
  },
  [STATE(319)] = {
    [sym_identifier] = ACTIONS(2740),
    [aux_sym_preproc_include_token1] = ACTIONS(2740),
    [aux_sym_preproc_include_token2] = ACTIONS(2740),
    [aux_sym_preproc_def_token1] = ACTIONS(2740),
    [aux_sym_preproc_if_token1] = ACTIONS(2740),
    [aux_sym_preproc_if_token2] = ACTIONS(2740),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2740),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2740),
    [aux_sym_preproc_else_token1] = ACTIONS(2740),
    [aux_sym_preproc_elif_token1] = ACTIONS(2740),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2740),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2740),
    [sym_preproc_directive] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_BANG] = ACTIONS(2742),
    [anon_sym_TILDE] = ACTIONS(2742),
    [anon_sym_DASH] = ACTIONS(2740),
    [anon_sym_PLUS] = ACTIONS(2740),
    [anon_sym_STAR] = ACTIONS(2742),
    [anon_sym_CARET] = ACTIONS(2742),
    [anon_sym_AMP] = ACTIONS(2742),
    [anon_sym_SEMI] = ACTIONS(2742),
    [anon_sym___extension__] = ACTIONS(2740),
    [anon_sym_typedef] = ACTIONS(2740),
    [anon_sym_extern] = ACTIONS(2740),
    [anon_sym___attribute__] = ACTIONS(2740),
    [anon_sym___attribute] = ACTIONS(2740),
    [anon_sym_noreturn] = ACTIONS(2740),
    [anon_sym_LBRACK] = ACTIONS(2742),
    [anon_sym___declspec] = ACTIONS(2740),
    [anon_sym___cdecl] = ACTIONS(2740),
    [anon_sym___clrcall] = ACTIONS(2740),
    [anon_sym___stdcall] = ACTIONS(2740),
    [anon_sym___fastcall] = ACTIONS(2740),
    [anon_sym___thiscall] = ACTIONS(2740),
    [anon_sym___vectorcall] = ACTIONS(2740),
    [anon_sym_LBRACE] = ACTIONS(2742),
    [anon_sym_signed] = ACTIONS(2740),
    [anon_sym_unsigned] = ACTIONS(2740),
    [anon_sym_long] = ACTIONS(2740),
    [anon_sym_short] = ACTIONS(2740),
    [anon_sym_static] = ACTIONS(2740),
    [anon_sym_ATautoreleasepool] = ACTIONS(2742),
    [anon_sym_auto] = ACTIONS(2740),
    [anon_sym_register] = ACTIONS(2740),
    [anon_sym_inline] = ACTIONS(2740),
    [anon_sym___inline] = ACTIONS(2740),
    [anon_sym___inline__] = ACTIONS(2740),
    [anon_sym___forceinline] = ACTIONS(2740),
    [anon_sym_thread_local] = ACTIONS(2740),
    [anon_sym___thread] = ACTIONS(2740),
    [anon_sym_CG_EXTERN] = ACTIONS(2740),
    [anon_sym_CG_INLINE] = ACTIONS(2740),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2740),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2740),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2740),
    [anon_sym_IBOutlet] = ACTIONS(2740),
    [anon_sym_IBInspectable] = ACTIONS(2740),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2740),
    [anon_sym_NS_INLINE] = ACTIONS(2740),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2740),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2740),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2740),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2740),
    [anon_sym_const] = ACTIONS(2740),
    [anon_sym_constexpr] = ACTIONS(2740),
    [anon_sym_volatile] = ACTIONS(2740),
    [anon_sym_restrict] = ACTIONS(2740),
    [anon_sym___restrict__] = ACTIONS(2740),
    [anon_sym__Atomic] = ACTIONS(2740),
    [anon_sym__Noreturn] = ACTIONS(2740),
    [anon_sym__Nonnull] = ACTIONS(2740),
    [anon_sym_nullable] = ACTIONS(2740),
    [anon_sym__Complex] = ACTIONS(2740),
    [anon_sym__Nullable] = ACTIONS(2740),
    [anon_sym__Nullable_result] = ACTIONS(2740),
    [anon_sym__Null_unspecified] = ACTIONS(2740),
    [anon_sym___autoreleasing] = ACTIONS(2740),
    [anon_sym___block] = ACTIONS(2740),
    [anon_sym___bridge] = ACTIONS(2740),
    [anon_sym___bridge_retained] = ACTIONS(2740),
    [anon_sym___bridge_transfer] = ACTIONS(2740),
    [anon_sym___complex] = ACTIONS(2740),
    [anon_sym___const] = ACTIONS(2740),
    [anon_sym___imag] = ACTIONS(2740),
    [anon_sym___kindof] = ACTIONS(2740),
    [anon_sym___nonnull] = ACTIONS(2740),
    [anon_sym___nullable] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2740),
    [anon_sym___real] = ACTIONS(2740),
    [anon_sym___strong] = ACTIONS(2740),
    [anon_sym___unsafe_unretained] = ACTIONS(2740),
    [anon_sym___unused] = ACTIONS(2740),
    [anon_sym___weak] = ACTIONS(2740),
    [anon_sym_alignas] = ACTIONS(2740),
    [anon_sym__Alignas] = ACTIONS(2740),
    [sym_primitive_type] = ACTIONS(2740),
    [anon_sym_enum] = ACTIONS(2740),
    [anon_sym_struct] = ACTIONS(2740),
    [anon_sym_union] = ACTIONS(2740),
    [anon_sym_if] = ACTIONS(2740),
    [anon_sym_switch] = ACTIONS(2740),
    [anon_sym_case] = ACTIONS(2740),
    [anon_sym_default] = ACTIONS(2740),
    [anon_sym_while] = ACTIONS(2740),
    [anon_sym_do] = ACTIONS(2740),
    [anon_sym_for] = ACTIONS(2740),
    [anon_sym_in] = ACTIONS(2740),
    [anon_sym_return] = ACTIONS(2740),
    [anon_sym_break] = ACTIONS(2740),
    [anon_sym_continue] = ACTIONS(2740),
    [anon_sym_goto] = ACTIONS(2740),
    [anon_sym_DASH_DASH] = ACTIONS(2742),
    [anon_sym_PLUS_PLUS] = ACTIONS(2742),
    [anon_sym_sizeof] = ACTIONS(2740),
    [anon_sym___alignof__] = ACTIONS(2740),
    [anon_sym___alignof] = ACTIONS(2740),
    [anon_sym__alignof] = ACTIONS(2740),
    [anon_sym_alignof] = ACTIONS(2740),
    [anon_sym__Alignof] = ACTIONS(2740),
    [anon_sym_offsetof] = ACTIONS(2740),
    [anon_sym__Generic] = ACTIONS(2740),
    [anon_sym_asm] = ACTIONS(2740),
    [anon_sym___asm__] = ACTIONS(2740),
    [anon_sym___asm] = ACTIONS(2740),
    [sym_number_literal] = ACTIONS(2742),
    [anon_sym_L_SQUOTE] = ACTIONS(2742),
    [anon_sym_u_SQUOTE] = ACTIONS(2742),
    [anon_sym_U_SQUOTE] = ACTIONS(2742),
    [anon_sym_u8_SQUOTE] = ACTIONS(2742),
    [anon_sym_SQUOTE] = ACTIONS(2742),
    [anon_sym_AT] = ACTIONS(2740),
    [anon_sym_DQUOTE] = ACTIONS(2742),
    [anon_sym_L_DQUOTE] = ACTIONS(2742),
    [anon_sym_u_DQUOTE] = ACTIONS(2742),
    [anon_sym_U_DQUOTE] = ACTIONS(2742),
    [anon_sym_u8_DQUOTE] = ACTIONS(2742),
    [sym_true] = ACTIONS(2740),
    [sym_false] = ACTIONS(2740),
    [anon_sym_NULL] = ACTIONS(2740),
    [anon_sym_nullptr] = ACTIONS(2740),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2740),
    [anon_sym___typeof] = ACTIONS(2740),
    [anon_sym_typeof] = ACTIONS(2740),
    [anon_sym_ATimport] = ACTIONS(2742),
    [aux_sym_preproc_undef_token1] = ACTIONS(2740),
    [anon_sym_POUND] = ACTIONS(2740),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2740),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2740),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2740),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2740),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2740),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2740),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2740),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_API_AVAILABLE] = ACTIONS(2740),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_API_DEPRECATED] = ACTIONS(2740),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2740),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym___deprecated_msg] = ACTIONS(2740),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2740),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2740),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2740),
    [anon_sym_ATprotocol] = ACTIONS(2742),
    [anon_sym_ATinterface] = ACTIONS(2742),
    [anon_sym_ATimplementation] = ACTIONS(2742),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2742),
    [anon_sym_ATsynthesize] = ACTIONS(2742),
    [anon_sym_ATdynamic] = ACTIONS(2742),
    [anon_sym_ATtry] = ACTIONS(2742),
    [anon_sym___try] = ACTIONS(2740),
    [anon_sym_ATthrow] = ACTIONS(2742),
    [anon_sym_ATselector] = ACTIONS(2742),
    [anon_sym_ATavailable] = ACTIONS(2742),
    [anon_sym___builtin_available] = ACTIONS(2740),
    [anon_sym_va_arg] = ACTIONS(2740),
    [anon_sym_ATencode] = ACTIONS(2742),
    [anon_sym_ATsynchronized] = ACTIONS(2742),
    [anon_sym_BOOL] = ACTIONS(2740),
    [anon_sym_IMP] = ACTIONS(2740),
    [anon_sym_SEL] = ACTIONS(2740),
    [anon_sym_Class] = ACTIONS(2740),
    [anon_sym_id] = ACTIONS(2740),
  },
  [STATE(320)] = {
    [sym_identifier] = ACTIONS(2744),
    [aux_sym_preproc_include_token1] = ACTIONS(2744),
    [aux_sym_preproc_include_token2] = ACTIONS(2744),
    [aux_sym_preproc_def_token1] = ACTIONS(2744),
    [aux_sym_preproc_if_token1] = ACTIONS(2744),
    [aux_sym_preproc_if_token2] = ACTIONS(2744),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2744),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2744),
    [aux_sym_preproc_else_token1] = ACTIONS(2744),
    [aux_sym_preproc_elif_token1] = ACTIONS(2744),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2744),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2744),
    [sym_preproc_directive] = ACTIONS(2744),
    [anon_sym_LPAREN2] = ACTIONS(2746),
    [anon_sym_BANG] = ACTIONS(2746),
    [anon_sym_TILDE] = ACTIONS(2746),
    [anon_sym_DASH] = ACTIONS(2744),
    [anon_sym_PLUS] = ACTIONS(2744),
    [anon_sym_STAR] = ACTIONS(2746),
    [anon_sym_CARET] = ACTIONS(2746),
    [anon_sym_AMP] = ACTIONS(2746),
    [anon_sym_SEMI] = ACTIONS(2746),
    [anon_sym___extension__] = ACTIONS(2744),
    [anon_sym_typedef] = ACTIONS(2744),
    [anon_sym_extern] = ACTIONS(2744),
    [anon_sym___attribute__] = ACTIONS(2744),
    [anon_sym___attribute] = ACTIONS(2744),
    [anon_sym_noreturn] = ACTIONS(2744),
    [anon_sym_LBRACK] = ACTIONS(2746),
    [anon_sym___declspec] = ACTIONS(2744),
    [anon_sym___cdecl] = ACTIONS(2744),
    [anon_sym___clrcall] = ACTIONS(2744),
    [anon_sym___stdcall] = ACTIONS(2744),
    [anon_sym___fastcall] = ACTIONS(2744),
    [anon_sym___thiscall] = ACTIONS(2744),
    [anon_sym___vectorcall] = ACTIONS(2744),
    [anon_sym_LBRACE] = ACTIONS(2746),
    [anon_sym_signed] = ACTIONS(2744),
    [anon_sym_unsigned] = ACTIONS(2744),
    [anon_sym_long] = ACTIONS(2744),
    [anon_sym_short] = ACTIONS(2744),
    [anon_sym_static] = ACTIONS(2744),
    [anon_sym_ATautoreleasepool] = ACTIONS(2746),
    [anon_sym_auto] = ACTIONS(2744),
    [anon_sym_register] = ACTIONS(2744),
    [anon_sym_inline] = ACTIONS(2744),
    [anon_sym___inline] = ACTIONS(2744),
    [anon_sym___inline__] = ACTIONS(2744),
    [anon_sym___forceinline] = ACTIONS(2744),
    [anon_sym_thread_local] = ACTIONS(2744),
    [anon_sym___thread] = ACTIONS(2744),
    [anon_sym_CG_EXTERN] = ACTIONS(2744),
    [anon_sym_CG_INLINE] = ACTIONS(2744),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2744),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2744),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2744),
    [anon_sym_IBOutlet] = ACTIONS(2744),
    [anon_sym_IBInspectable] = ACTIONS(2744),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2744),
    [anon_sym_NS_INLINE] = ACTIONS(2744),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2744),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2744),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2744),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2744),
    [anon_sym_const] = ACTIONS(2744),
    [anon_sym_constexpr] = ACTIONS(2744),
    [anon_sym_volatile] = ACTIONS(2744),
    [anon_sym_restrict] = ACTIONS(2744),
    [anon_sym___restrict__] = ACTIONS(2744),
    [anon_sym__Atomic] = ACTIONS(2744),
    [anon_sym__Noreturn] = ACTIONS(2744),
    [anon_sym__Nonnull] = ACTIONS(2744),
    [anon_sym_nullable] = ACTIONS(2744),
    [anon_sym__Complex] = ACTIONS(2744),
    [anon_sym__Nullable] = ACTIONS(2744),
    [anon_sym__Nullable_result] = ACTIONS(2744),
    [anon_sym__Null_unspecified] = ACTIONS(2744),
    [anon_sym___autoreleasing] = ACTIONS(2744),
    [anon_sym___block] = ACTIONS(2744),
    [anon_sym___bridge] = ACTIONS(2744),
    [anon_sym___bridge_retained] = ACTIONS(2744),
    [anon_sym___bridge_transfer] = ACTIONS(2744),
    [anon_sym___complex] = ACTIONS(2744),
    [anon_sym___const] = ACTIONS(2744),
    [anon_sym___imag] = ACTIONS(2744),
    [anon_sym___kindof] = ACTIONS(2744),
    [anon_sym___nonnull] = ACTIONS(2744),
    [anon_sym___nullable] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2744),
    [anon_sym___real] = ACTIONS(2744),
    [anon_sym___strong] = ACTIONS(2744),
    [anon_sym___unsafe_unretained] = ACTIONS(2744),
    [anon_sym___unused] = ACTIONS(2744),
    [anon_sym___weak] = ACTIONS(2744),
    [anon_sym_alignas] = ACTIONS(2744),
    [anon_sym__Alignas] = ACTIONS(2744),
    [sym_primitive_type] = ACTIONS(2744),
    [anon_sym_enum] = ACTIONS(2744),
    [anon_sym_struct] = ACTIONS(2744),
    [anon_sym_union] = ACTIONS(2744),
    [anon_sym_if] = ACTIONS(2744),
    [anon_sym_switch] = ACTIONS(2744),
    [anon_sym_case] = ACTIONS(2744),
    [anon_sym_default] = ACTIONS(2744),
    [anon_sym_while] = ACTIONS(2744),
    [anon_sym_do] = ACTIONS(2744),
    [anon_sym_for] = ACTIONS(2744),
    [anon_sym_in] = ACTIONS(2744),
    [anon_sym_return] = ACTIONS(2744),
    [anon_sym_break] = ACTIONS(2744),
    [anon_sym_continue] = ACTIONS(2744),
    [anon_sym_goto] = ACTIONS(2744),
    [anon_sym_DASH_DASH] = ACTIONS(2746),
    [anon_sym_PLUS_PLUS] = ACTIONS(2746),
    [anon_sym_sizeof] = ACTIONS(2744),
    [anon_sym___alignof__] = ACTIONS(2744),
    [anon_sym___alignof] = ACTIONS(2744),
    [anon_sym__alignof] = ACTIONS(2744),
    [anon_sym_alignof] = ACTIONS(2744),
    [anon_sym__Alignof] = ACTIONS(2744),
    [anon_sym_offsetof] = ACTIONS(2744),
    [anon_sym__Generic] = ACTIONS(2744),
    [anon_sym_asm] = ACTIONS(2744),
    [anon_sym___asm__] = ACTIONS(2744),
    [anon_sym___asm] = ACTIONS(2744),
    [sym_number_literal] = ACTIONS(2746),
    [anon_sym_L_SQUOTE] = ACTIONS(2746),
    [anon_sym_u_SQUOTE] = ACTIONS(2746),
    [anon_sym_U_SQUOTE] = ACTIONS(2746),
    [anon_sym_u8_SQUOTE] = ACTIONS(2746),
    [anon_sym_SQUOTE] = ACTIONS(2746),
    [anon_sym_AT] = ACTIONS(2744),
    [anon_sym_DQUOTE] = ACTIONS(2746),
    [anon_sym_L_DQUOTE] = ACTIONS(2746),
    [anon_sym_u_DQUOTE] = ACTIONS(2746),
    [anon_sym_U_DQUOTE] = ACTIONS(2746),
    [anon_sym_u8_DQUOTE] = ACTIONS(2746),
    [sym_true] = ACTIONS(2744),
    [sym_false] = ACTIONS(2744),
    [anon_sym_NULL] = ACTIONS(2744),
    [anon_sym_nullptr] = ACTIONS(2744),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2744),
    [anon_sym___typeof] = ACTIONS(2744),
    [anon_sym_typeof] = ACTIONS(2744),
    [anon_sym_ATimport] = ACTIONS(2746),
    [aux_sym_preproc_undef_token1] = ACTIONS(2744),
    [anon_sym_POUND] = ACTIONS(2744),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2744),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2744),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2744),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2744),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2744),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2744),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2744),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_API_AVAILABLE] = ACTIONS(2744),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_API_DEPRECATED] = ACTIONS(2744),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2744),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym___deprecated_msg] = ACTIONS(2744),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2744),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2744),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2744),
    [anon_sym_ATprotocol] = ACTIONS(2746),
    [anon_sym_ATinterface] = ACTIONS(2746),
    [anon_sym_ATimplementation] = ACTIONS(2746),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2746),
    [anon_sym_ATsynthesize] = ACTIONS(2746),
    [anon_sym_ATdynamic] = ACTIONS(2746),
    [anon_sym_ATtry] = ACTIONS(2746),
    [anon_sym___try] = ACTIONS(2744),
    [anon_sym_ATthrow] = ACTIONS(2746),
    [anon_sym_ATselector] = ACTIONS(2746),
    [anon_sym_ATavailable] = ACTIONS(2746),
    [anon_sym___builtin_available] = ACTIONS(2744),
    [anon_sym_va_arg] = ACTIONS(2744),
    [anon_sym_ATencode] = ACTIONS(2746),
    [anon_sym_ATsynchronized] = ACTIONS(2746),
    [anon_sym_BOOL] = ACTIONS(2744),
    [anon_sym_IMP] = ACTIONS(2744),
    [anon_sym_SEL] = ACTIONS(2744),
    [anon_sym_Class] = ACTIONS(2744),
    [anon_sym_id] = ACTIONS(2744),
  },
  [STATE(321)] = {
    [sym_identifier] = ACTIONS(2748),
    [aux_sym_preproc_include_token1] = ACTIONS(2748),
    [aux_sym_preproc_include_token2] = ACTIONS(2748),
    [aux_sym_preproc_def_token1] = ACTIONS(2748),
    [aux_sym_preproc_if_token1] = ACTIONS(2748),
    [aux_sym_preproc_if_token2] = ACTIONS(2748),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2748),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2748),
    [aux_sym_preproc_else_token1] = ACTIONS(2748),
    [aux_sym_preproc_elif_token1] = ACTIONS(2748),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2748),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2748),
    [sym_preproc_directive] = ACTIONS(2748),
    [anon_sym_LPAREN2] = ACTIONS(2750),
    [anon_sym_BANG] = ACTIONS(2750),
    [anon_sym_TILDE] = ACTIONS(2750),
    [anon_sym_DASH] = ACTIONS(2748),
    [anon_sym_PLUS] = ACTIONS(2748),
    [anon_sym_STAR] = ACTIONS(2750),
    [anon_sym_CARET] = ACTIONS(2750),
    [anon_sym_AMP] = ACTIONS(2750),
    [anon_sym_SEMI] = ACTIONS(2750),
    [anon_sym___extension__] = ACTIONS(2748),
    [anon_sym_typedef] = ACTIONS(2748),
    [anon_sym_extern] = ACTIONS(2748),
    [anon_sym___attribute__] = ACTIONS(2748),
    [anon_sym___attribute] = ACTIONS(2748),
    [anon_sym_noreturn] = ACTIONS(2748),
    [anon_sym_LBRACK] = ACTIONS(2750),
    [anon_sym___declspec] = ACTIONS(2748),
    [anon_sym___cdecl] = ACTIONS(2748),
    [anon_sym___clrcall] = ACTIONS(2748),
    [anon_sym___stdcall] = ACTIONS(2748),
    [anon_sym___fastcall] = ACTIONS(2748),
    [anon_sym___thiscall] = ACTIONS(2748),
    [anon_sym___vectorcall] = ACTIONS(2748),
    [anon_sym_LBRACE] = ACTIONS(2750),
    [anon_sym_signed] = ACTIONS(2748),
    [anon_sym_unsigned] = ACTIONS(2748),
    [anon_sym_long] = ACTIONS(2748),
    [anon_sym_short] = ACTIONS(2748),
    [anon_sym_static] = ACTIONS(2748),
    [anon_sym_ATautoreleasepool] = ACTIONS(2750),
    [anon_sym_auto] = ACTIONS(2748),
    [anon_sym_register] = ACTIONS(2748),
    [anon_sym_inline] = ACTIONS(2748),
    [anon_sym___inline] = ACTIONS(2748),
    [anon_sym___inline__] = ACTIONS(2748),
    [anon_sym___forceinline] = ACTIONS(2748),
    [anon_sym_thread_local] = ACTIONS(2748),
    [anon_sym___thread] = ACTIONS(2748),
    [anon_sym_CG_EXTERN] = ACTIONS(2748),
    [anon_sym_CG_INLINE] = ACTIONS(2748),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2748),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2748),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2748),
    [anon_sym_IBOutlet] = ACTIONS(2748),
    [anon_sym_IBInspectable] = ACTIONS(2748),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2748),
    [anon_sym_NS_INLINE] = ACTIONS(2748),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2748),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2748),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2748),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2748),
    [anon_sym_const] = ACTIONS(2748),
    [anon_sym_constexpr] = ACTIONS(2748),
    [anon_sym_volatile] = ACTIONS(2748),
    [anon_sym_restrict] = ACTIONS(2748),
    [anon_sym___restrict__] = ACTIONS(2748),
    [anon_sym__Atomic] = ACTIONS(2748),
    [anon_sym__Noreturn] = ACTIONS(2748),
    [anon_sym__Nonnull] = ACTIONS(2748),
    [anon_sym_nullable] = ACTIONS(2748),
    [anon_sym__Complex] = ACTIONS(2748),
    [anon_sym__Nullable] = ACTIONS(2748),
    [anon_sym__Nullable_result] = ACTIONS(2748),
    [anon_sym__Null_unspecified] = ACTIONS(2748),
    [anon_sym___autoreleasing] = ACTIONS(2748),
    [anon_sym___block] = ACTIONS(2748),
    [anon_sym___bridge] = ACTIONS(2748),
    [anon_sym___bridge_retained] = ACTIONS(2748),
    [anon_sym___bridge_transfer] = ACTIONS(2748),
    [anon_sym___complex] = ACTIONS(2748),
    [anon_sym___const] = ACTIONS(2748),
    [anon_sym___imag] = ACTIONS(2748),
    [anon_sym___kindof] = ACTIONS(2748),
    [anon_sym___nonnull] = ACTIONS(2748),
    [anon_sym___nullable] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2748),
    [anon_sym___real] = ACTIONS(2748),
    [anon_sym___strong] = ACTIONS(2748),
    [anon_sym___unsafe_unretained] = ACTIONS(2748),
    [anon_sym___unused] = ACTIONS(2748),
    [anon_sym___weak] = ACTIONS(2748),
    [anon_sym_alignas] = ACTIONS(2748),
    [anon_sym__Alignas] = ACTIONS(2748),
    [sym_primitive_type] = ACTIONS(2748),
    [anon_sym_enum] = ACTIONS(2748),
    [anon_sym_struct] = ACTIONS(2748),
    [anon_sym_union] = ACTIONS(2748),
    [anon_sym_if] = ACTIONS(2748),
    [anon_sym_switch] = ACTIONS(2748),
    [anon_sym_case] = ACTIONS(2748),
    [anon_sym_default] = ACTIONS(2748),
    [anon_sym_while] = ACTIONS(2748),
    [anon_sym_do] = ACTIONS(2748),
    [anon_sym_for] = ACTIONS(2748),
    [anon_sym_in] = ACTIONS(2748),
    [anon_sym_return] = ACTIONS(2748),
    [anon_sym_break] = ACTIONS(2748),
    [anon_sym_continue] = ACTIONS(2748),
    [anon_sym_goto] = ACTIONS(2748),
    [anon_sym_DASH_DASH] = ACTIONS(2750),
    [anon_sym_PLUS_PLUS] = ACTIONS(2750),
    [anon_sym_sizeof] = ACTIONS(2748),
    [anon_sym___alignof__] = ACTIONS(2748),
    [anon_sym___alignof] = ACTIONS(2748),
    [anon_sym__alignof] = ACTIONS(2748),
    [anon_sym_alignof] = ACTIONS(2748),
    [anon_sym__Alignof] = ACTIONS(2748),
    [anon_sym_offsetof] = ACTIONS(2748),
    [anon_sym__Generic] = ACTIONS(2748),
    [anon_sym_asm] = ACTIONS(2748),
    [anon_sym___asm__] = ACTIONS(2748),
    [anon_sym___asm] = ACTIONS(2748),
    [sym_number_literal] = ACTIONS(2750),
    [anon_sym_L_SQUOTE] = ACTIONS(2750),
    [anon_sym_u_SQUOTE] = ACTIONS(2750),
    [anon_sym_U_SQUOTE] = ACTIONS(2750),
    [anon_sym_u8_SQUOTE] = ACTIONS(2750),
    [anon_sym_SQUOTE] = ACTIONS(2750),
    [anon_sym_AT] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_L_DQUOTE] = ACTIONS(2750),
    [anon_sym_u_DQUOTE] = ACTIONS(2750),
    [anon_sym_U_DQUOTE] = ACTIONS(2750),
    [anon_sym_u8_DQUOTE] = ACTIONS(2750),
    [sym_true] = ACTIONS(2748),
    [sym_false] = ACTIONS(2748),
    [anon_sym_NULL] = ACTIONS(2748),
    [anon_sym_nullptr] = ACTIONS(2748),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2748),
    [anon_sym___typeof] = ACTIONS(2748),
    [anon_sym_typeof] = ACTIONS(2748),
    [anon_sym_ATimport] = ACTIONS(2750),
    [aux_sym_preproc_undef_token1] = ACTIONS(2748),
    [anon_sym_POUND] = ACTIONS(2748),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2748),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2748),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2748),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2748),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2748),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2748),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2748),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_API_AVAILABLE] = ACTIONS(2748),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_API_DEPRECATED] = ACTIONS(2748),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2748),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym___deprecated_msg] = ACTIONS(2748),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2748),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2748),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2748),
    [anon_sym_ATprotocol] = ACTIONS(2750),
    [anon_sym_ATinterface] = ACTIONS(2750),
    [anon_sym_ATimplementation] = ACTIONS(2750),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2750),
    [anon_sym_ATsynthesize] = ACTIONS(2750),
    [anon_sym_ATdynamic] = ACTIONS(2750),
    [anon_sym_ATtry] = ACTIONS(2750),
    [anon_sym___try] = ACTIONS(2748),
    [anon_sym_ATthrow] = ACTIONS(2750),
    [anon_sym_ATselector] = ACTIONS(2750),
    [anon_sym_ATavailable] = ACTIONS(2750),
    [anon_sym___builtin_available] = ACTIONS(2748),
    [anon_sym_va_arg] = ACTIONS(2748),
    [anon_sym_ATencode] = ACTIONS(2750),
    [anon_sym_ATsynchronized] = ACTIONS(2750),
    [anon_sym_BOOL] = ACTIONS(2748),
    [anon_sym_IMP] = ACTIONS(2748),
    [anon_sym_SEL] = ACTIONS(2748),
    [anon_sym_Class] = ACTIONS(2748),
    [anon_sym_id] = ACTIONS(2748),
  },
  [STATE(322)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token2] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [aux_sym_preproc_else_token1] = ACTIONS(2752),
    [aux_sym_preproc_elif_token1] = ACTIONS(2752),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATsynthesize] = ACTIONS(2754),
    [anon_sym_ATdynamic] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(323)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token2] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [aux_sym_preproc_else_token1] = ACTIONS(2752),
    [aux_sym_preproc_elif_token1] = ACTIONS(2752),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATsynthesize] = ACTIONS(2754),
    [anon_sym_ATdynamic] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(324)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token2] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [aux_sym_preproc_else_token1] = ACTIONS(2752),
    [aux_sym_preproc_elif_token1] = ACTIONS(2752),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATsynthesize] = ACTIONS(2754),
    [anon_sym_ATdynamic] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(325)] = {
    [sym_identifier] = ACTIONS(2756),
    [aux_sym_preproc_include_token1] = ACTIONS(2756),
    [aux_sym_preproc_include_token2] = ACTIONS(2756),
    [aux_sym_preproc_def_token1] = ACTIONS(2756),
    [aux_sym_preproc_if_token1] = ACTIONS(2756),
    [aux_sym_preproc_if_token2] = ACTIONS(2756),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2756),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2756),
    [aux_sym_preproc_else_token1] = ACTIONS(2756),
    [aux_sym_preproc_elif_token1] = ACTIONS(2756),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2756),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2756),
    [sym_preproc_directive] = ACTIONS(2756),
    [anon_sym_LPAREN2] = ACTIONS(2758),
    [anon_sym_BANG] = ACTIONS(2758),
    [anon_sym_TILDE] = ACTIONS(2758),
    [anon_sym_DASH] = ACTIONS(2756),
    [anon_sym_PLUS] = ACTIONS(2756),
    [anon_sym_STAR] = ACTIONS(2758),
    [anon_sym_CARET] = ACTIONS(2758),
    [anon_sym_AMP] = ACTIONS(2758),
    [anon_sym_SEMI] = ACTIONS(2758),
    [anon_sym___extension__] = ACTIONS(2756),
    [anon_sym_typedef] = ACTIONS(2756),
    [anon_sym_extern] = ACTIONS(2756),
    [anon_sym___attribute__] = ACTIONS(2756),
    [anon_sym___attribute] = ACTIONS(2756),
    [anon_sym_noreturn] = ACTIONS(2756),
    [anon_sym_LBRACK] = ACTIONS(2758),
    [anon_sym___declspec] = ACTIONS(2756),
    [anon_sym___cdecl] = ACTIONS(2756),
    [anon_sym___clrcall] = ACTIONS(2756),
    [anon_sym___stdcall] = ACTIONS(2756),
    [anon_sym___fastcall] = ACTIONS(2756),
    [anon_sym___thiscall] = ACTIONS(2756),
    [anon_sym___vectorcall] = ACTIONS(2756),
    [anon_sym_LBRACE] = ACTIONS(2758),
    [anon_sym_signed] = ACTIONS(2756),
    [anon_sym_unsigned] = ACTIONS(2756),
    [anon_sym_long] = ACTIONS(2756),
    [anon_sym_short] = ACTIONS(2756),
    [anon_sym_static] = ACTIONS(2756),
    [anon_sym_ATautoreleasepool] = ACTIONS(2758),
    [anon_sym_auto] = ACTIONS(2756),
    [anon_sym_register] = ACTIONS(2756),
    [anon_sym_inline] = ACTIONS(2756),
    [anon_sym___inline] = ACTIONS(2756),
    [anon_sym___inline__] = ACTIONS(2756),
    [anon_sym___forceinline] = ACTIONS(2756),
    [anon_sym_thread_local] = ACTIONS(2756),
    [anon_sym___thread] = ACTIONS(2756),
    [anon_sym_CG_EXTERN] = ACTIONS(2756),
    [anon_sym_CG_INLINE] = ACTIONS(2756),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2756),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2756),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2756),
    [anon_sym_IBOutlet] = ACTIONS(2756),
    [anon_sym_IBInspectable] = ACTIONS(2756),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2756),
    [anon_sym_NS_INLINE] = ACTIONS(2756),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2756),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2756),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2756),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2756),
    [anon_sym_const] = ACTIONS(2756),
    [anon_sym_constexpr] = ACTIONS(2756),
    [anon_sym_volatile] = ACTIONS(2756),
    [anon_sym_restrict] = ACTIONS(2756),
    [anon_sym___restrict__] = ACTIONS(2756),
    [anon_sym__Atomic] = ACTIONS(2756),
    [anon_sym__Noreturn] = ACTIONS(2756),
    [anon_sym__Nonnull] = ACTIONS(2756),
    [anon_sym_nullable] = ACTIONS(2756),
    [anon_sym__Complex] = ACTIONS(2756),
    [anon_sym__Nullable] = ACTIONS(2756),
    [anon_sym__Nullable_result] = ACTIONS(2756),
    [anon_sym__Null_unspecified] = ACTIONS(2756),
    [anon_sym___autoreleasing] = ACTIONS(2756),
    [anon_sym___block] = ACTIONS(2756),
    [anon_sym___bridge] = ACTIONS(2756),
    [anon_sym___bridge_retained] = ACTIONS(2756),
    [anon_sym___bridge_transfer] = ACTIONS(2756),
    [anon_sym___complex] = ACTIONS(2756),
    [anon_sym___const] = ACTIONS(2756),
    [anon_sym___imag] = ACTIONS(2756),
    [anon_sym___kindof] = ACTIONS(2756),
    [anon_sym___nonnull] = ACTIONS(2756),
    [anon_sym___nullable] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2756),
    [anon_sym___real] = ACTIONS(2756),
    [anon_sym___strong] = ACTIONS(2756),
    [anon_sym___unsafe_unretained] = ACTIONS(2756),
    [anon_sym___unused] = ACTIONS(2756),
    [anon_sym___weak] = ACTIONS(2756),
    [anon_sym_alignas] = ACTIONS(2756),
    [anon_sym__Alignas] = ACTIONS(2756),
    [sym_primitive_type] = ACTIONS(2756),
    [anon_sym_enum] = ACTIONS(2756),
    [anon_sym_struct] = ACTIONS(2756),
    [anon_sym_union] = ACTIONS(2756),
    [anon_sym_if] = ACTIONS(2756),
    [anon_sym_switch] = ACTIONS(2756),
    [anon_sym_case] = ACTIONS(2756),
    [anon_sym_default] = ACTIONS(2756),
    [anon_sym_while] = ACTIONS(2756),
    [anon_sym_do] = ACTIONS(2756),
    [anon_sym_for] = ACTIONS(2756),
    [anon_sym_in] = ACTIONS(2756),
    [anon_sym_return] = ACTIONS(2756),
    [anon_sym_break] = ACTIONS(2756),
    [anon_sym_continue] = ACTIONS(2756),
    [anon_sym_goto] = ACTIONS(2756),
    [anon_sym_DASH_DASH] = ACTIONS(2758),
    [anon_sym_PLUS_PLUS] = ACTIONS(2758),
    [anon_sym_sizeof] = ACTIONS(2756),
    [anon_sym___alignof__] = ACTIONS(2756),
    [anon_sym___alignof] = ACTIONS(2756),
    [anon_sym__alignof] = ACTIONS(2756),
    [anon_sym_alignof] = ACTIONS(2756),
    [anon_sym__Alignof] = ACTIONS(2756),
    [anon_sym_offsetof] = ACTIONS(2756),
    [anon_sym__Generic] = ACTIONS(2756),
    [anon_sym_asm] = ACTIONS(2756),
    [anon_sym___asm__] = ACTIONS(2756),
    [anon_sym___asm] = ACTIONS(2756),
    [sym_number_literal] = ACTIONS(2758),
    [anon_sym_L_SQUOTE] = ACTIONS(2758),
    [anon_sym_u_SQUOTE] = ACTIONS(2758),
    [anon_sym_U_SQUOTE] = ACTIONS(2758),
    [anon_sym_u8_SQUOTE] = ACTIONS(2758),
    [anon_sym_SQUOTE] = ACTIONS(2758),
    [anon_sym_AT] = ACTIONS(2756),
    [anon_sym_DQUOTE] = ACTIONS(2758),
    [anon_sym_L_DQUOTE] = ACTIONS(2758),
    [anon_sym_u_DQUOTE] = ACTIONS(2758),
    [anon_sym_U_DQUOTE] = ACTIONS(2758),
    [anon_sym_u8_DQUOTE] = ACTIONS(2758),
    [sym_true] = ACTIONS(2756),
    [sym_false] = ACTIONS(2756),
    [anon_sym_NULL] = ACTIONS(2756),
    [anon_sym_nullptr] = ACTIONS(2756),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2756),
    [anon_sym___typeof] = ACTIONS(2756),
    [anon_sym_typeof] = ACTIONS(2756),
    [anon_sym_ATimport] = ACTIONS(2758),
    [aux_sym_preproc_undef_token1] = ACTIONS(2756),
    [anon_sym_POUND] = ACTIONS(2756),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2756),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2756),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2756),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2756),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2756),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2756),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2756),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_API_AVAILABLE] = ACTIONS(2756),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_API_DEPRECATED] = ACTIONS(2756),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2756),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym___deprecated_msg] = ACTIONS(2756),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2756),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2756),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2756),
    [anon_sym_ATprotocol] = ACTIONS(2758),
    [anon_sym_ATinterface] = ACTIONS(2758),
    [anon_sym_ATimplementation] = ACTIONS(2758),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2758),
    [anon_sym_ATsynthesize] = ACTIONS(2758),
    [anon_sym_ATdynamic] = ACTIONS(2758),
    [anon_sym_ATtry] = ACTIONS(2758),
    [anon_sym___try] = ACTIONS(2756),
    [anon_sym_ATthrow] = ACTIONS(2758),
    [anon_sym_ATselector] = ACTIONS(2758),
    [anon_sym_ATavailable] = ACTIONS(2758),
    [anon_sym___builtin_available] = ACTIONS(2756),
    [anon_sym_va_arg] = ACTIONS(2756),
    [anon_sym_ATencode] = ACTIONS(2758),
    [anon_sym_ATsynchronized] = ACTIONS(2758),
    [anon_sym_BOOL] = ACTIONS(2756),
    [anon_sym_IMP] = ACTIONS(2756),
    [anon_sym_SEL] = ACTIONS(2756),
    [anon_sym_Class] = ACTIONS(2756),
    [anon_sym_id] = ACTIONS(2756),
  },
  [STATE(326)] = {
    [sym_identifier] = ACTIONS(2760),
    [aux_sym_preproc_include_token1] = ACTIONS(2760),
    [aux_sym_preproc_include_token2] = ACTIONS(2760),
    [aux_sym_preproc_def_token1] = ACTIONS(2760),
    [aux_sym_preproc_if_token1] = ACTIONS(2760),
    [aux_sym_preproc_if_token2] = ACTIONS(2760),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2760),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2760),
    [aux_sym_preproc_else_token1] = ACTIONS(2760),
    [aux_sym_preproc_elif_token1] = ACTIONS(2760),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2760),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2760),
    [sym_preproc_directive] = ACTIONS(2760),
    [anon_sym_LPAREN2] = ACTIONS(2762),
    [anon_sym_BANG] = ACTIONS(2762),
    [anon_sym_TILDE] = ACTIONS(2762),
    [anon_sym_DASH] = ACTIONS(2760),
    [anon_sym_PLUS] = ACTIONS(2760),
    [anon_sym_STAR] = ACTIONS(2762),
    [anon_sym_CARET] = ACTIONS(2762),
    [anon_sym_AMP] = ACTIONS(2762),
    [anon_sym_SEMI] = ACTIONS(2762),
    [anon_sym___extension__] = ACTIONS(2760),
    [anon_sym_typedef] = ACTIONS(2760),
    [anon_sym_extern] = ACTIONS(2760),
    [anon_sym___attribute__] = ACTIONS(2760),
    [anon_sym___attribute] = ACTIONS(2760),
    [anon_sym_noreturn] = ACTIONS(2760),
    [anon_sym_LBRACK] = ACTIONS(2762),
    [anon_sym___declspec] = ACTIONS(2760),
    [anon_sym___cdecl] = ACTIONS(2760),
    [anon_sym___clrcall] = ACTIONS(2760),
    [anon_sym___stdcall] = ACTIONS(2760),
    [anon_sym___fastcall] = ACTIONS(2760),
    [anon_sym___thiscall] = ACTIONS(2760),
    [anon_sym___vectorcall] = ACTIONS(2760),
    [anon_sym_LBRACE] = ACTIONS(2762),
    [anon_sym_signed] = ACTIONS(2760),
    [anon_sym_unsigned] = ACTIONS(2760),
    [anon_sym_long] = ACTIONS(2760),
    [anon_sym_short] = ACTIONS(2760),
    [anon_sym_static] = ACTIONS(2760),
    [anon_sym_ATautoreleasepool] = ACTIONS(2762),
    [anon_sym_auto] = ACTIONS(2760),
    [anon_sym_register] = ACTIONS(2760),
    [anon_sym_inline] = ACTIONS(2760),
    [anon_sym___inline] = ACTIONS(2760),
    [anon_sym___inline__] = ACTIONS(2760),
    [anon_sym___forceinline] = ACTIONS(2760),
    [anon_sym_thread_local] = ACTIONS(2760),
    [anon_sym___thread] = ACTIONS(2760),
    [anon_sym_CG_EXTERN] = ACTIONS(2760),
    [anon_sym_CG_INLINE] = ACTIONS(2760),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2760),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2760),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2760),
    [anon_sym_IBOutlet] = ACTIONS(2760),
    [anon_sym_IBInspectable] = ACTIONS(2760),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2760),
    [anon_sym_NS_INLINE] = ACTIONS(2760),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2760),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2760),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2760),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2760),
    [anon_sym_const] = ACTIONS(2760),
    [anon_sym_constexpr] = ACTIONS(2760),
    [anon_sym_volatile] = ACTIONS(2760),
    [anon_sym_restrict] = ACTIONS(2760),
    [anon_sym___restrict__] = ACTIONS(2760),
    [anon_sym__Atomic] = ACTIONS(2760),
    [anon_sym__Noreturn] = ACTIONS(2760),
    [anon_sym__Nonnull] = ACTIONS(2760),
    [anon_sym_nullable] = ACTIONS(2760),
    [anon_sym__Complex] = ACTIONS(2760),
    [anon_sym__Nullable] = ACTIONS(2760),
    [anon_sym__Nullable_result] = ACTIONS(2760),
    [anon_sym__Null_unspecified] = ACTIONS(2760),
    [anon_sym___autoreleasing] = ACTIONS(2760),
    [anon_sym___block] = ACTIONS(2760),
    [anon_sym___bridge] = ACTIONS(2760),
    [anon_sym___bridge_retained] = ACTIONS(2760),
    [anon_sym___bridge_transfer] = ACTIONS(2760),
    [anon_sym___complex] = ACTIONS(2760),
    [anon_sym___const] = ACTIONS(2760),
    [anon_sym___imag] = ACTIONS(2760),
    [anon_sym___kindof] = ACTIONS(2760),
    [anon_sym___nonnull] = ACTIONS(2760),
    [anon_sym___nullable] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2760),
    [anon_sym___real] = ACTIONS(2760),
    [anon_sym___strong] = ACTIONS(2760),
    [anon_sym___unsafe_unretained] = ACTIONS(2760),
    [anon_sym___unused] = ACTIONS(2760),
    [anon_sym___weak] = ACTIONS(2760),
    [anon_sym_alignas] = ACTIONS(2760),
    [anon_sym__Alignas] = ACTIONS(2760),
    [sym_primitive_type] = ACTIONS(2760),
    [anon_sym_enum] = ACTIONS(2760),
    [anon_sym_struct] = ACTIONS(2760),
    [anon_sym_union] = ACTIONS(2760),
    [anon_sym_if] = ACTIONS(2760),
    [anon_sym_switch] = ACTIONS(2760),
    [anon_sym_case] = ACTIONS(2760),
    [anon_sym_default] = ACTIONS(2760),
    [anon_sym_while] = ACTIONS(2760),
    [anon_sym_do] = ACTIONS(2760),
    [anon_sym_for] = ACTIONS(2760),
    [anon_sym_in] = ACTIONS(2760),
    [anon_sym_return] = ACTIONS(2760),
    [anon_sym_break] = ACTIONS(2760),
    [anon_sym_continue] = ACTIONS(2760),
    [anon_sym_goto] = ACTIONS(2760),
    [anon_sym_DASH_DASH] = ACTIONS(2762),
    [anon_sym_PLUS_PLUS] = ACTIONS(2762),
    [anon_sym_sizeof] = ACTIONS(2760),
    [anon_sym___alignof__] = ACTIONS(2760),
    [anon_sym___alignof] = ACTIONS(2760),
    [anon_sym__alignof] = ACTIONS(2760),
    [anon_sym_alignof] = ACTIONS(2760),
    [anon_sym__Alignof] = ACTIONS(2760),
    [anon_sym_offsetof] = ACTIONS(2760),
    [anon_sym__Generic] = ACTIONS(2760),
    [anon_sym_asm] = ACTIONS(2760),
    [anon_sym___asm__] = ACTIONS(2760),
    [anon_sym___asm] = ACTIONS(2760),
    [sym_number_literal] = ACTIONS(2762),
    [anon_sym_L_SQUOTE] = ACTIONS(2762),
    [anon_sym_u_SQUOTE] = ACTIONS(2762),
    [anon_sym_U_SQUOTE] = ACTIONS(2762),
    [anon_sym_u8_SQUOTE] = ACTIONS(2762),
    [anon_sym_SQUOTE] = ACTIONS(2762),
    [anon_sym_AT] = ACTIONS(2760),
    [anon_sym_DQUOTE] = ACTIONS(2762),
    [anon_sym_L_DQUOTE] = ACTIONS(2762),
    [anon_sym_u_DQUOTE] = ACTIONS(2762),
    [anon_sym_U_DQUOTE] = ACTIONS(2762),
    [anon_sym_u8_DQUOTE] = ACTIONS(2762),
    [sym_true] = ACTIONS(2760),
    [sym_false] = ACTIONS(2760),
    [anon_sym_NULL] = ACTIONS(2760),
    [anon_sym_nullptr] = ACTIONS(2760),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2760),
    [anon_sym___typeof] = ACTIONS(2760),
    [anon_sym_typeof] = ACTIONS(2760),
    [anon_sym_ATimport] = ACTIONS(2762),
    [aux_sym_preproc_undef_token1] = ACTIONS(2760),
    [anon_sym_POUND] = ACTIONS(2760),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2760),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2760),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2760),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2760),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2760),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2760),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2760),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_API_AVAILABLE] = ACTIONS(2760),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_API_DEPRECATED] = ACTIONS(2760),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2760),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym___deprecated_msg] = ACTIONS(2760),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2760),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2760),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2760),
    [anon_sym_ATprotocol] = ACTIONS(2762),
    [anon_sym_ATinterface] = ACTIONS(2762),
    [anon_sym_ATimplementation] = ACTIONS(2762),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2762),
    [anon_sym_ATsynthesize] = ACTIONS(2762),
    [anon_sym_ATdynamic] = ACTIONS(2762),
    [anon_sym_ATtry] = ACTIONS(2762),
    [anon_sym___try] = ACTIONS(2760),
    [anon_sym_ATthrow] = ACTIONS(2762),
    [anon_sym_ATselector] = ACTIONS(2762),
    [anon_sym_ATavailable] = ACTIONS(2762),
    [anon_sym___builtin_available] = ACTIONS(2760),
    [anon_sym_va_arg] = ACTIONS(2760),
    [anon_sym_ATencode] = ACTIONS(2762),
    [anon_sym_ATsynchronized] = ACTIONS(2762),
    [anon_sym_BOOL] = ACTIONS(2760),
    [anon_sym_IMP] = ACTIONS(2760),
    [anon_sym_SEL] = ACTIONS(2760),
    [anon_sym_Class] = ACTIONS(2760),
    [anon_sym_id] = ACTIONS(2760),
  },
  [STATE(327)] = {
    [sym_identifier] = ACTIONS(2764),
    [aux_sym_preproc_include_token1] = ACTIONS(2764),
    [aux_sym_preproc_include_token2] = ACTIONS(2764),
    [aux_sym_preproc_def_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token2] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2764),
    [aux_sym_preproc_else_token1] = ACTIONS(2764),
    [aux_sym_preproc_elif_token1] = ACTIONS(2764),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2764),
    [sym_preproc_directive] = ACTIONS(2764),
    [anon_sym_LPAREN2] = ACTIONS(2766),
    [anon_sym_BANG] = ACTIONS(2766),
    [anon_sym_TILDE] = ACTIONS(2766),
    [anon_sym_DASH] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(2764),
    [anon_sym_STAR] = ACTIONS(2766),
    [anon_sym_CARET] = ACTIONS(2766),
    [anon_sym_AMP] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2766),
    [anon_sym___extension__] = ACTIONS(2764),
    [anon_sym_typedef] = ACTIONS(2764),
    [anon_sym_extern] = ACTIONS(2764),
    [anon_sym___attribute__] = ACTIONS(2764),
    [anon_sym___attribute] = ACTIONS(2764),
    [anon_sym_noreturn] = ACTIONS(2764),
    [anon_sym_LBRACK] = ACTIONS(2766),
    [anon_sym___declspec] = ACTIONS(2764),
    [anon_sym___cdecl] = ACTIONS(2764),
    [anon_sym___clrcall] = ACTIONS(2764),
    [anon_sym___stdcall] = ACTIONS(2764),
    [anon_sym___fastcall] = ACTIONS(2764),
    [anon_sym___thiscall] = ACTIONS(2764),
    [anon_sym___vectorcall] = ACTIONS(2764),
    [anon_sym_LBRACE] = ACTIONS(2766),
    [anon_sym_signed] = ACTIONS(2764),
    [anon_sym_unsigned] = ACTIONS(2764),
    [anon_sym_long] = ACTIONS(2764),
    [anon_sym_short] = ACTIONS(2764),
    [anon_sym_static] = ACTIONS(2764),
    [anon_sym_ATautoreleasepool] = ACTIONS(2766),
    [anon_sym_auto] = ACTIONS(2764),
    [anon_sym_register] = ACTIONS(2764),
    [anon_sym_inline] = ACTIONS(2764),
    [anon_sym___inline] = ACTIONS(2764),
    [anon_sym___inline__] = ACTIONS(2764),
    [anon_sym___forceinline] = ACTIONS(2764),
    [anon_sym_thread_local] = ACTIONS(2764),
    [anon_sym___thread] = ACTIONS(2764),
    [anon_sym_CG_EXTERN] = ACTIONS(2764),
    [anon_sym_CG_INLINE] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2764),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2764),
    [anon_sym_IBOutlet] = ACTIONS(2764),
    [anon_sym_IBInspectable] = ACTIONS(2764),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2764),
    [anon_sym_NS_INLINE] = ACTIONS(2764),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2764),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2764),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2764),
    [anon_sym_const] = ACTIONS(2764),
    [anon_sym_constexpr] = ACTIONS(2764),
    [anon_sym_volatile] = ACTIONS(2764),
    [anon_sym_restrict] = ACTIONS(2764),
    [anon_sym___restrict__] = ACTIONS(2764),
    [anon_sym__Atomic] = ACTIONS(2764),
    [anon_sym__Noreturn] = ACTIONS(2764),
    [anon_sym__Nonnull] = ACTIONS(2764),
    [anon_sym_nullable] = ACTIONS(2764),
    [anon_sym__Complex] = ACTIONS(2764),
    [anon_sym__Nullable] = ACTIONS(2764),
    [anon_sym__Nullable_result] = ACTIONS(2764),
    [anon_sym__Null_unspecified] = ACTIONS(2764),
    [anon_sym___autoreleasing] = ACTIONS(2764),
    [anon_sym___block] = ACTIONS(2764),
    [anon_sym___bridge] = ACTIONS(2764),
    [anon_sym___bridge_retained] = ACTIONS(2764),
    [anon_sym___bridge_transfer] = ACTIONS(2764),
    [anon_sym___complex] = ACTIONS(2764),
    [anon_sym___const] = ACTIONS(2764),
    [anon_sym___imag] = ACTIONS(2764),
    [anon_sym___kindof] = ACTIONS(2764),
    [anon_sym___nonnull] = ACTIONS(2764),
    [anon_sym___nullable] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2764),
    [anon_sym___real] = ACTIONS(2764),
    [anon_sym___strong] = ACTIONS(2764),
    [anon_sym___unsafe_unretained] = ACTIONS(2764),
    [anon_sym___unused] = ACTIONS(2764),
    [anon_sym___weak] = ACTIONS(2764),
    [anon_sym_alignas] = ACTIONS(2764),
    [anon_sym__Alignas] = ACTIONS(2764),
    [sym_primitive_type] = ACTIONS(2764),
    [anon_sym_enum] = ACTIONS(2764),
    [anon_sym_struct] = ACTIONS(2764),
    [anon_sym_union] = ACTIONS(2764),
    [anon_sym_if] = ACTIONS(2764),
    [anon_sym_switch] = ACTIONS(2764),
    [anon_sym_case] = ACTIONS(2764),
    [anon_sym_default] = ACTIONS(2764),
    [anon_sym_while] = ACTIONS(2764),
    [anon_sym_do] = ACTIONS(2764),
    [anon_sym_for] = ACTIONS(2764),
    [anon_sym_in] = ACTIONS(2764),
    [anon_sym_return] = ACTIONS(2764),
    [anon_sym_break] = ACTIONS(2764),
    [anon_sym_continue] = ACTIONS(2764),
    [anon_sym_goto] = ACTIONS(2764),
    [anon_sym_DASH_DASH] = ACTIONS(2766),
    [anon_sym_PLUS_PLUS] = ACTIONS(2766),
    [anon_sym_sizeof] = ACTIONS(2764),
    [anon_sym___alignof__] = ACTIONS(2764),
    [anon_sym___alignof] = ACTIONS(2764),
    [anon_sym__alignof] = ACTIONS(2764),
    [anon_sym_alignof] = ACTIONS(2764),
    [anon_sym__Alignof] = ACTIONS(2764),
    [anon_sym_offsetof] = ACTIONS(2764),
    [anon_sym__Generic] = ACTIONS(2764),
    [anon_sym_asm] = ACTIONS(2764),
    [anon_sym___asm__] = ACTIONS(2764),
    [anon_sym___asm] = ACTIONS(2764),
    [sym_number_literal] = ACTIONS(2766),
    [anon_sym_L_SQUOTE] = ACTIONS(2766),
    [anon_sym_u_SQUOTE] = ACTIONS(2766),
    [anon_sym_U_SQUOTE] = ACTIONS(2766),
    [anon_sym_u8_SQUOTE] = ACTIONS(2766),
    [anon_sym_SQUOTE] = ACTIONS(2766),
    [anon_sym_AT] = ACTIONS(2764),
    [anon_sym_DQUOTE] = ACTIONS(2766),
    [anon_sym_L_DQUOTE] = ACTIONS(2766),
    [anon_sym_u_DQUOTE] = ACTIONS(2766),
    [anon_sym_U_DQUOTE] = ACTIONS(2766),
    [anon_sym_u8_DQUOTE] = ACTIONS(2766),
    [sym_true] = ACTIONS(2764),
    [sym_false] = ACTIONS(2764),
    [anon_sym_NULL] = ACTIONS(2764),
    [anon_sym_nullptr] = ACTIONS(2764),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2764),
    [anon_sym___typeof] = ACTIONS(2764),
    [anon_sym_typeof] = ACTIONS(2764),
    [anon_sym_ATimport] = ACTIONS(2766),
    [aux_sym_preproc_undef_token1] = ACTIONS(2764),
    [anon_sym_POUND] = ACTIONS(2764),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2764),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2764),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2764),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_API_AVAILABLE] = ACTIONS(2764),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_API_DEPRECATED] = ACTIONS(2764),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym___deprecated_msg] = ACTIONS(2764),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2764),
    [anon_sym_ATprotocol] = ACTIONS(2766),
    [anon_sym_ATinterface] = ACTIONS(2766),
    [anon_sym_ATimplementation] = ACTIONS(2766),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2766),
    [anon_sym_ATsynthesize] = ACTIONS(2766),
    [anon_sym_ATdynamic] = ACTIONS(2766),
    [anon_sym_ATtry] = ACTIONS(2766),
    [anon_sym___try] = ACTIONS(2764),
    [anon_sym_ATthrow] = ACTIONS(2766),
    [anon_sym_ATselector] = ACTIONS(2766),
    [anon_sym_ATavailable] = ACTIONS(2766),
    [anon_sym___builtin_available] = ACTIONS(2764),
    [anon_sym_va_arg] = ACTIONS(2764),
    [anon_sym_ATencode] = ACTIONS(2766),
    [anon_sym_ATsynchronized] = ACTIONS(2766),
    [anon_sym_BOOL] = ACTIONS(2764),
    [anon_sym_IMP] = ACTIONS(2764),
    [anon_sym_SEL] = ACTIONS(2764),
    [anon_sym_Class] = ACTIONS(2764),
    [anon_sym_id] = ACTIONS(2764),
  },
  [STATE(328)] = {
    [sym_identifier] = ACTIONS(2764),
    [aux_sym_preproc_include_token1] = ACTIONS(2764),
    [aux_sym_preproc_include_token2] = ACTIONS(2764),
    [aux_sym_preproc_def_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token2] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2764),
    [aux_sym_preproc_else_token1] = ACTIONS(2764),
    [aux_sym_preproc_elif_token1] = ACTIONS(2764),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2764),
    [sym_preproc_directive] = ACTIONS(2764),
    [anon_sym_LPAREN2] = ACTIONS(2766),
    [anon_sym_BANG] = ACTIONS(2766),
    [anon_sym_TILDE] = ACTIONS(2766),
    [anon_sym_DASH] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(2764),
    [anon_sym_STAR] = ACTIONS(2766),
    [anon_sym_CARET] = ACTIONS(2766),
    [anon_sym_AMP] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2766),
    [anon_sym___extension__] = ACTIONS(2764),
    [anon_sym_typedef] = ACTIONS(2764),
    [anon_sym_extern] = ACTIONS(2764),
    [anon_sym___attribute__] = ACTIONS(2764),
    [anon_sym___attribute] = ACTIONS(2764),
    [anon_sym_noreturn] = ACTIONS(2764),
    [anon_sym_LBRACK] = ACTIONS(2766),
    [anon_sym___declspec] = ACTIONS(2764),
    [anon_sym___cdecl] = ACTIONS(2764),
    [anon_sym___clrcall] = ACTIONS(2764),
    [anon_sym___stdcall] = ACTIONS(2764),
    [anon_sym___fastcall] = ACTIONS(2764),
    [anon_sym___thiscall] = ACTIONS(2764),
    [anon_sym___vectorcall] = ACTIONS(2764),
    [anon_sym_LBRACE] = ACTIONS(2766),
    [anon_sym_signed] = ACTIONS(2764),
    [anon_sym_unsigned] = ACTIONS(2764),
    [anon_sym_long] = ACTIONS(2764),
    [anon_sym_short] = ACTIONS(2764),
    [anon_sym_static] = ACTIONS(2764),
    [anon_sym_ATautoreleasepool] = ACTIONS(2766),
    [anon_sym_auto] = ACTIONS(2764),
    [anon_sym_register] = ACTIONS(2764),
    [anon_sym_inline] = ACTIONS(2764),
    [anon_sym___inline] = ACTIONS(2764),
    [anon_sym___inline__] = ACTIONS(2764),
    [anon_sym___forceinline] = ACTIONS(2764),
    [anon_sym_thread_local] = ACTIONS(2764),
    [anon_sym___thread] = ACTIONS(2764),
    [anon_sym_CG_EXTERN] = ACTIONS(2764),
    [anon_sym_CG_INLINE] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2764),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2764),
    [anon_sym_IBOutlet] = ACTIONS(2764),
    [anon_sym_IBInspectable] = ACTIONS(2764),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2764),
    [anon_sym_NS_INLINE] = ACTIONS(2764),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2764),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2764),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2764),
    [anon_sym_const] = ACTIONS(2764),
    [anon_sym_constexpr] = ACTIONS(2764),
    [anon_sym_volatile] = ACTIONS(2764),
    [anon_sym_restrict] = ACTIONS(2764),
    [anon_sym___restrict__] = ACTIONS(2764),
    [anon_sym__Atomic] = ACTIONS(2764),
    [anon_sym__Noreturn] = ACTIONS(2764),
    [anon_sym__Nonnull] = ACTIONS(2764),
    [anon_sym_nullable] = ACTIONS(2764),
    [anon_sym__Complex] = ACTIONS(2764),
    [anon_sym__Nullable] = ACTIONS(2764),
    [anon_sym__Nullable_result] = ACTIONS(2764),
    [anon_sym__Null_unspecified] = ACTIONS(2764),
    [anon_sym___autoreleasing] = ACTIONS(2764),
    [anon_sym___block] = ACTIONS(2764),
    [anon_sym___bridge] = ACTIONS(2764),
    [anon_sym___bridge_retained] = ACTIONS(2764),
    [anon_sym___bridge_transfer] = ACTIONS(2764),
    [anon_sym___complex] = ACTIONS(2764),
    [anon_sym___const] = ACTIONS(2764),
    [anon_sym___imag] = ACTIONS(2764),
    [anon_sym___kindof] = ACTIONS(2764),
    [anon_sym___nonnull] = ACTIONS(2764),
    [anon_sym___nullable] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2764),
    [anon_sym___real] = ACTIONS(2764),
    [anon_sym___strong] = ACTIONS(2764),
    [anon_sym___unsafe_unretained] = ACTIONS(2764),
    [anon_sym___unused] = ACTIONS(2764),
    [anon_sym___weak] = ACTIONS(2764),
    [anon_sym_alignas] = ACTIONS(2764),
    [anon_sym__Alignas] = ACTIONS(2764),
    [sym_primitive_type] = ACTIONS(2764),
    [anon_sym_enum] = ACTIONS(2764),
    [anon_sym_struct] = ACTIONS(2764),
    [anon_sym_union] = ACTIONS(2764),
    [anon_sym_if] = ACTIONS(2764),
    [anon_sym_switch] = ACTIONS(2764),
    [anon_sym_case] = ACTIONS(2764),
    [anon_sym_default] = ACTIONS(2764),
    [anon_sym_while] = ACTIONS(2764),
    [anon_sym_do] = ACTIONS(2764),
    [anon_sym_for] = ACTIONS(2764),
    [anon_sym_in] = ACTIONS(2764),
    [anon_sym_return] = ACTIONS(2764),
    [anon_sym_break] = ACTIONS(2764),
    [anon_sym_continue] = ACTIONS(2764),
    [anon_sym_goto] = ACTIONS(2764),
    [anon_sym_DASH_DASH] = ACTIONS(2766),
    [anon_sym_PLUS_PLUS] = ACTIONS(2766),
    [anon_sym_sizeof] = ACTIONS(2764),
    [anon_sym___alignof__] = ACTIONS(2764),
    [anon_sym___alignof] = ACTIONS(2764),
    [anon_sym__alignof] = ACTIONS(2764),
    [anon_sym_alignof] = ACTIONS(2764),
    [anon_sym__Alignof] = ACTIONS(2764),
    [anon_sym_offsetof] = ACTIONS(2764),
    [anon_sym__Generic] = ACTIONS(2764),
    [anon_sym_asm] = ACTIONS(2764),
    [anon_sym___asm__] = ACTIONS(2764),
    [anon_sym___asm] = ACTIONS(2764),
    [sym_number_literal] = ACTIONS(2766),
    [anon_sym_L_SQUOTE] = ACTIONS(2766),
    [anon_sym_u_SQUOTE] = ACTIONS(2766),
    [anon_sym_U_SQUOTE] = ACTIONS(2766),
    [anon_sym_u8_SQUOTE] = ACTIONS(2766),
    [anon_sym_SQUOTE] = ACTIONS(2766),
    [anon_sym_AT] = ACTIONS(2764),
    [anon_sym_DQUOTE] = ACTIONS(2766),
    [anon_sym_L_DQUOTE] = ACTIONS(2766),
    [anon_sym_u_DQUOTE] = ACTIONS(2766),
    [anon_sym_U_DQUOTE] = ACTIONS(2766),
    [anon_sym_u8_DQUOTE] = ACTIONS(2766),
    [sym_true] = ACTIONS(2764),
    [sym_false] = ACTIONS(2764),
    [anon_sym_NULL] = ACTIONS(2764),
    [anon_sym_nullptr] = ACTIONS(2764),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2764),
    [anon_sym___typeof] = ACTIONS(2764),
    [anon_sym_typeof] = ACTIONS(2764),
    [anon_sym_ATimport] = ACTIONS(2766),
    [aux_sym_preproc_undef_token1] = ACTIONS(2764),
    [anon_sym_POUND] = ACTIONS(2764),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2764),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2764),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2764),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_API_AVAILABLE] = ACTIONS(2764),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_API_DEPRECATED] = ACTIONS(2764),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym___deprecated_msg] = ACTIONS(2764),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2764),
    [anon_sym_ATprotocol] = ACTIONS(2766),
    [anon_sym_ATinterface] = ACTIONS(2766),
    [anon_sym_ATimplementation] = ACTIONS(2766),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2766),
    [anon_sym_ATsynthesize] = ACTIONS(2766),
    [anon_sym_ATdynamic] = ACTIONS(2766),
    [anon_sym_ATtry] = ACTIONS(2766),
    [anon_sym___try] = ACTIONS(2764),
    [anon_sym_ATthrow] = ACTIONS(2766),
    [anon_sym_ATselector] = ACTIONS(2766),
    [anon_sym_ATavailable] = ACTIONS(2766),
    [anon_sym___builtin_available] = ACTIONS(2764),
    [anon_sym_va_arg] = ACTIONS(2764),
    [anon_sym_ATencode] = ACTIONS(2766),
    [anon_sym_ATsynchronized] = ACTIONS(2766),
    [anon_sym_BOOL] = ACTIONS(2764),
    [anon_sym_IMP] = ACTIONS(2764),
    [anon_sym_SEL] = ACTIONS(2764),
    [anon_sym_Class] = ACTIONS(2764),
    [anon_sym_id] = ACTIONS(2764),
  },
  [STATE(329)] = {
    [sym_identifier] = ACTIONS(2768),
    [aux_sym_preproc_include_token1] = ACTIONS(2768),
    [aux_sym_preproc_include_token2] = ACTIONS(2768),
    [aux_sym_preproc_def_token1] = ACTIONS(2768),
    [aux_sym_preproc_if_token1] = ACTIONS(2768),
    [aux_sym_preproc_if_token2] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2768),
    [aux_sym_preproc_else_token1] = ACTIONS(2768),
    [aux_sym_preproc_elif_token1] = ACTIONS(2768),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2768),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2768),
    [sym_preproc_directive] = ACTIONS(2768),
    [anon_sym_LPAREN2] = ACTIONS(2770),
    [anon_sym_BANG] = ACTIONS(2770),
    [anon_sym_TILDE] = ACTIONS(2770),
    [anon_sym_DASH] = ACTIONS(2768),
    [anon_sym_PLUS] = ACTIONS(2768),
    [anon_sym_STAR] = ACTIONS(2770),
    [anon_sym_CARET] = ACTIONS(2770),
    [anon_sym_AMP] = ACTIONS(2770),
    [anon_sym_SEMI] = ACTIONS(2770),
    [anon_sym___extension__] = ACTIONS(2768),
    [anon_sym_typedef] = ACTIONS(2768),
    [anon_sym_extern] = ACTIONS(2768),
    [anon_sym___attribute__] = ACTIONS(2768),
    [anon_sym___attribute] = ACTIONS(2768),
    [anon_sym_noreturn] = ACTIONS(2768),
    [anon_sym_LBRACK] = ACTIONS(2770),
    [anon_sym___declspec] = ACTIONS(2768),
    [anon_sym___cdecl] = ACTIONS(2768),
    [anon_sym___clrcall] = ACTIONS(2768),
    [anon_sym___stdcall] = ACTIONS(2768),
    [anon_sym___fastcall] = ACTIONS(2768),
    [anon_sym___thiscall] = ACTIONS(2768),
    [anon_sym___vectorcall] = ACTIONS(2768),
    [anon_sym_LBRACE] = ACTIONS(2770),
    [anon_sym_signed] = ACTIONS(2768),
    [anon_sym_unsigned] = ACTIONS(2768),
    [anon_sym_long] = ACTIONS(2768),
    [anon_sym_short] = ACTIONS(2768),
    [anon_sym_static] = ACTIONS(2768),
    [anon_sym_ATautoreleasepool] = ACTIONS(2770),
    [anon_sym_auto] = ACTIONS(2768),
    [anon_sym_register] = ACTIONS(2768),
    [anon_sym_inline] = ACTIONS(2768),
    [anon_sym___inline] = ACTIONS(2768),
    [anon_sym___inline__] = ACTIONS(2768),
    [anon_sym___forceinline] = ACTIONS(2768),
    [anon_sym_thread_local] = ACTIONS(2768),
    [anon_sym___thread] = ACTIONS(2768),
    [anon_sym_CG_EXTERN] = ACTIONS(2768),
    [anon_sym_CG_INLINE] = ACTIONS(2768),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2768),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2768),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2768),
    [anon_sym_IBOutlet] = ACTIONS(2768),
    [anon_sym_IBInspectable] = ACTIONS(2768),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2768),
    [anon_sym_NS_INLINE] = ACTIONS(2768),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2768),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2768),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2768),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2768),
    [anon_sym_const] = ACTIONS(2768),
    [anon_sym_constexpr] = ACTIONS(2768),
    [anon_sym_volatile] = ACTIONS(2768),
    [anon_sym_restrict] = ACTIONS(2768),
    [anon_sym___restrict__] = ACTIONS(2768),
    [anon_sym__Atomic] = ACTIONS(2768),
    [anon_sym__Noreturn] = ACTIONS(2768),
    [anon_sym__Nonnull] = ACTIONS(2768),
    [anon_sym_nullable] = ACTIONS(2768),
    [anon_sym__Complex] = ACTIONS(2768),
    [anon_sym__Nullable] = ACTIONS(2768),
    [anon_sym__Nullable_result] = ACTIONS(2768),
    [anon_sym__Null_unspecified] = ACTIONS(2768),
    [anon_sym___autoreleasing] = ACTIONS(2768),
    [anon_sym___block] = ACTIONS(2768),
    [anon_sym___bridge] = ACTIONS(2768),
    [anon_sym___bridge_retained] = ACTIONS(2768),
    [anon_sym___bridge_transfer] = ACTIONS(2768),
    [anon_sym___complex] = ACTIONS(2768),
    [anon_sym___const] = ACTIONS(2768),
    [anon_sym___imag] = ACTIONS(2768),
    [anon_sym___kindof] = ACTIONS(2768),
    [anon_sym___nonnull] = ACTIONS(2768),
    [anon_sym___nullable] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2768),
    [anon_sym___real] = ACTIONS(2768),
    [anon_sym___strong] = ACTIONS(2768),
    [anon_sym___unsafe_unretained] = ACTIONS(2768),
    [anon_sym___unused] = ACTIONS(2768),
    [anon_sym___weak] = ACTIONS(2768),
    [anon_sym_alignas] = ACTIONS(2768),
    [anon_sym__Alignas] = ACTIONS(2768),
    [sym_primitive_type] = ACTIONS(2768),
    [anon_sym_enum] = ACTIONS(2768),
    [anon_sym_struct] = ACTIONS(2768),
    [anon_sym_union] = ACTIONS(2768),
    [anon_sym_if] = ACTIONS(2768),
    [anon_sym_switch] = ACTIONS(2768),
    [anon_sym_case] = ACTIONS(2768),
    [anon_sym_default] = ACTIONS(2768),
    [anon_sym_while] = ACTIONS(2768),
    [anon_sym_do] = ACTIONS(2768),
    [anon_sym_for] = ACTIONS(2768),
    [anon_sym_in] = ACTIONS(2768),
    [anon_sym_return] = ACTIONS(2768),
    [anon_sym_break] = ACTIONS(2768),
    [anon_sym_continue] = ACTIONS(2768),
    [anon_sym_goto] = ACTIONS(2768),
    [anon_sym_DASH_DASH] = ACTIONS(2770),
    [anon_sym_PLUS_PLUS] = ACTIONS(2770),
    [anon_sym_sizeof] = ACTIONS(2768),
    [anon_sym___alignof__] = ACTIONS(2768),
    [anon_sym___alignof] = ACTIONS(2768),
    [anon_sym__alignof] = ACTIONS(2768),
    [anon_sym_alignof] = ACTIONS(2768),
    [anon_sym__Alignof] = ACTIONS(2768),
    [anon_sym_offsetof] = ACTIONS(2768),
    [anon_sym__Generic] = ACTIONS(2768),
    [anon_sym_asm] = ACTIONS(2768),
    [anon_sym___asm__] = ACTIONS(2768),
    [anon_sym___asm] = ACTIONS(2768),
    [sym_number_literal] = ACTIONS(2770),
    [anon_sym_L_SQUOTE] = ACTIONS(2770),
    [anon_sym_u_SQUOTE] = ACTIONS(2770),
    [anon_sym_U_SQUOTE] = ACTIONS(2770),
    [anon_sym_u8_SQUOTE] = ACTIONS(2770),
    [anon_sym_SQUOTE] = ACTIONS(2770),
    [anon_sym_AT] = ACTIONS(2768),
    [anon_sym_DQUOTE] = ACTIONS(2770),
    [anon_sym_L_DQUOTE] = ACTIONS(2770),
    [anon_sym_u_DQUOTE] = ACTIONS(2770),
    [anon_sym_U_DQUOTE] = ACTIONS(2770),
    [anon_sym_u8_DQUOTE] = ACTIONS(2770),
    [sym_true] = ACTIONS(2768),
    [sym_false] = ACTIONS(2768),
    [anon_sym_NULL] = ACTIONS(2768),
    [anon_sym_nullptr] = ACTIONS(2768),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2768),
    [anon_sym___typeof] = ACTIONS(2768),
    [anon_sym_typeof] = ACTIONS(2768),
    [anon_sym_ATimport] = ACTIONS(2770),
    [aux_sym_preproc_undef_token1] = ACTIONS(2768),
    [anon_sym_POUND] = ACTIONS(2768),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2768),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2768),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2768),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2768),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2768),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2768),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2768),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_API_AVAILABLE] = ACTIONS(2768),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_API_DEPRECATED] = ACTIONS(2768),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2768),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym___deprecated_msg] = ACTIONS(2768),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2768),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2768),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2768),
    [anon_sym_ATprotocol] = ACTIONS(2770),
    [anon_sym_ATinterface] = ACTIONS(2770),
    [anon_sym_ATimplementation] = ACTIONS(2770),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2770),
    [anon_sym_ATsynthesize] = ACTIONS(2770),
    [anon_sym_ATdynamic] = ACTIONS(2770),
    [anon_sym_ATtry] = ACTIONS(2770),
    [anon_sym___try] = ACTIONS(2768),
    [anon_sym_ATthrow] = ACTIONS(2770),
    [anon_sym_ATselector] = ACTIONS(2770),
    [anon_sym_ATavailable] = ACTIONS(2770),
    [anon_sym___builtin_available] = ACTIONS(2768),
    [anon_sym_va_arg] = ACTIONS(2768),
    [anon_sym_ATencode] = ACTIONS(2770),
    [anon_sym_ATsynchronized] = ACTIONS(2770),
    [anon_sym_BOOL] = ACTIONS(2768),
    [anon_sym_IMP] = ACTIONS(2768),
    [anon_sym_SEL] = ACTIONS(2768),
    [anon_sym_Class] = ACTIONS(2768),
    [anon_sym_id] = ACTIONS(2768),
  },
  [STATE(330)] = {
    [sym_identifier] = ACTIONS(2772),
    [aux_sym_preproc_include_token1] = ACTIONS(2772),
    [aux_sym_preproc_include_token2] = ACTIONS(2772),
    [aux_sym_preproc_def_token1] = ACTIONS(2772),
    [aux_sym_preproc_if_token1] = ACTIONS(2772),
    [aux_sym_preproc_if_token2] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2772),
    [aux_sym_preproc_else_token1] = ACTIONS(2772),
    [aux_sym_preproc_elif_token1] = ACTIONS(2772),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2772),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2772),
    [sym_preproc_directive] = ACTIONS(2772),
    [anon_sym_LPAREN2] = ACTIONS(2774),
    [anon_sym_BANG] = ACTIONS(2774),
    [anon_sym_TILDE] = ACTIONS(2774),
    [anon_sym_DASH] = ACTIONS(2772),
    [anon_sym_PLUS] = ACTIONS(2772),
    [anon_sym_STAR] = ACTIONS(2774),
    [anon_sym_CARET] = ACTIONS(2774),
    [anon_sym_AMP] = ACTIONS(2774),
    [anon_sym_SEMI] = ACTIONS(2774),
    [anon_sym___extension__] = ACTIONS(2772),
    [anon_sym_typedef] = ACTIONS(2772),
    [anon_sym_extern] = ACTIONS(2772),
    [anon_sym___attribute__] = ACTIONS(2772),
    [anon_sym___attribute] = ACTIONS(2772),
    [anon_sym_noreturn] = ACTIONS(2772),
    [anon_sym_LBRACK] = ACTIONS(2774),
    [anon_sym___declspec] = ACTIONS(2772),
    [anon_sym___cdecl] = ACTIONS(2772),
    [anon_sym___clrcall] = ACTIONS(2772),
    [anon_sym___stdcall] = ACTIONS(2772),
    [anon_sym___fastcall] = ACTIONS(2772),
    [anon_sym___thiscall] = ACTIONS(2772),
    [anon_sym___vectorcall] = ACTIONS(2772),
    [anon_sym_LBRACE] = ACTIONS(2774),
    [anon_sym_signed] = ACTIONS(2772),
    [anon_sym_unsigned] = ACTIONS(2772),
    [anon_sym_long] = ACTIONS(2772),
    [anon_sym_short] = ACTIONS(2772),
    [anon_sym_static] = ACTIONS(2772),
    [anon_sym_ATautoreleasepool] = ACTIONS(2774),
    [anon_sym_auto] = ACTIONS(2772),
    [anon_sym_register] = ACTIONS(2772),
    [anon_sym_inline] = ACTIONS(2772),
    [anon_sym___inline] = ACTIONS(2772),
    [anon_sym___inline__] = ACTIONS(2772),
    [anon_sym___forceinline] = ACTIONS(2772),
    [anon_sym_thread_local] = ACTIONS(2772),
    [anon_sym___thread] = ACTIONS(2772),
    [anon_sym_CG_EXTERN] = ACTIONS(2772),
    [anon_sym_CG_INLINE] = ACTIONS(2772),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2772),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2772),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2772),
    [anon_sym_IBOutlet] = ACTIONS(2772),
    [anon_sym_IBInspectable] = ACTIONS(2772),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2772),
    [anon_sym_NS_INLINE] = ACTIONS(2772),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2772),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2772),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2772),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2772),
    [anon_sym_const] = ACTIONS(2772),
    [anon_sym_constexpr] = ACTIONS(2772),
    [anon_sym_volatile] = ACTIONS(2772),
    [anon_sym_restrict] = ACTIONS(2772),
    [anon_sym___restrict__] = ACTIONS(2772),
    [anon_sym__Atomic] = ACTIONS(2772),
    [anon_sym__Noreturn] = ACTIONS(2772),
    [anon_sym__Nonnull] = ACTIONS(2772),
    [anon_sym_nullable] = ACTIONS(2772),
    [anon_sym__Complex] = ACTIONS(2772),
    [anon_sym__Nullable] = ACTIONS(2772),
    [anon_sym__Nullable_result] = ACTIONS(2772),
    [anon_sym__Null_unspecified] = ACTIONS(2772),
    [anon_sym___autoreleasing] = ACTIONS(2772),
    [anon_sym___block] = ACTIONS(2772),
    [anon_sym___bridge] = ACTIONS(2772),
    [anon_sym___bridge_retained] = ACTIONS(2772),
    [anon_sym___bridge_transfer] = ACTIONS(2772),
    [anon_sym___complex] = ACTIONS(2772),
    [anon_sym___const] = ACTIONS(2772),
    [anon_sym___imag] = ACTIONS(2772),
    [anon_sym___kindof] = ACTIONS(2772),
    [anon_sym___nonnull] = ACTIONS(2772),
    [anon_sym___nullable] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2772),
    [anon_sym___real] = ACTIONS(2772),
    [anon_sym___strong] = ACTIONS(2772),
    [anon_sym___unsafe_unretained] = ACTIONS(2772),
    [anon_sym___unused] = ACTIONS(2772),
    [anon_sym___weak] = ACTIONS(2772),
    [anon_sym_alignas] = ACTIONS(2772),
    [anon_sym__Alignas] = ACTIONS(2772),
    [sym_primitive_type] = ACTIONS(2772),
    [anon_sym_enum] = ACTIONS(2772),
    [anon_sym_struct] = ACTIONS(2772),
    [anon_sym_union] = ACTIONS(2772),
    [anon_sym_if] = ACTIONS(2772),
    [anon_sym_switch] = ACTIONS(2772),
    [anon_sym_case] = ACTIONS(2772),
    [anon_sym_default] = ACTIONS(2772),
    [anon_sym_while] = ACTIONS(2772),
    [anon_sym_do] = ACTIONS(2772),
    [anon_sym_for] = ACTIONS(2772),
    [anon_sym_in] = ACTIONS(2772),
    [anon_sym_return] = ACTIONS(2772),
    [anon_sym_break] = ACTIONS(2772),
    [anon_sym_continue] = ACTIONS(2772),
    [anon_sym_goto] = ACTIONS(2772),
    [anon_sym_DASH_DASH] = ACTIONS(2774),
    [anon_sym_PLUS_PLUS] = ACTIONS(2774),
    [anon_sym_sizeof] = ACTIONS(2772),
    [anon_sym___alignof__] = ACTIONS(2772),
    [anon_sym___alignof] = ACTIONS(2772),
    [anon_sym__alignof] = ACTIONS(2772),
    [anon_sym_alignof] = ACTIONS(2772),
    [anon_sym__Alignof] = ACTIONS(2772),
    [anon_sym_offsetof] = ACTIONS(2772),
    [anon_sym__Generic] = ACTIONS(2772),
    [anon_sym_asm] = ACTIONS(2772),
    [anon_sym___asm__] = ACTIONS(2772),
    [anon_sym___asm] = ACTIONS(2772),
    [sym_number_literal] = ACTIONS(2774),
    [anon_sym_L_SQUOTE] = ACTIONS(2774),
    [anon_sym_u_SQUOTE] = ACTIONS(2774),
    [anon_sym_U_SQUOTE] = ACTIONS(2774),
    [anon_sym_u8_SQUOTE] = ACTIONS(2774),
    [anon_sym_SQUOTE] = ACTIONS(2774),
    [anon_sym_AT] = ACTIONS(2772),
    [anon_sym_DQUOTE] = ACTIONS(2774),
    [anon_sym_L_DQUOTE] = ACTIONS(2774),
    [anon_sym_u_DQUOTE] = ACTIONS(2774),
    [anon_sym_U_DQUOTE] = ACTIONS(2774),
    [anon_sym_u8_DQUOTE] = ACTIONS(2774),
    [sym_true] = ACTIONS(2772),
    [sym_false] = ACTIONS(2772),
    [anon_sym_NULL] = ACTIONS(2772),
    [anon_sym_nullptr] = ACTIONS(2772),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2772),
    [anon_sym___typeof] = ACTIONS(2772),
    [anon_sym_typeof] = ACTIONS(2772),
    [anon_sym_ATimport] = ACTIONS(2774),
    [aux_sym_preproc_undef_token1] = ACTIONS(2772),
    [anon_sym_POUND] = ACTIONS(2772),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2772),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2772),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2772),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2772),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2772),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2772),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2772),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_API_AVAILABLE] = ACTIONS(2772),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_API_DEPRECATED] = ACTIONS(2772),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2772),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym___deprecated_msg] = ACTIONS(2772),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2772),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2772),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2772),
    [anon_sym_ATprotocol] = ACTIONS(2774),
    [anon_sym_ATinterface] = ACTIONS(2774),
    [anon_sym_ATimplementation] = ACTIONS(2774),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2774),
    [anon_sym_ATsynthesize] = ACTIONS(2774),
    [anon_sym_ATdynamic] = ACTIONS(2774),
    [anon_sym_ATtry] = ACTIONS(2774),
    [anon_sym___try] = ACTIONS(2772),
    [anon_sym_ATthrow] = ACTIONS(2774),
    [anon_sym_ATselector] = ACTIONS(2774),
    [anon_sym_ATavailable] = ACTIONS(2774),
    [anon_sym___builtin_available] = ACTIONS(2772),
    [anon_sym_va_arg] = ACTIONS(2772),
    [anon_sym_ATencode] = ACTIONS(2774),
    [anon_sym_ATsynchronized] = ACTIONS(2774),
    [anon_sym_BOOL] = ACTIONS(2772),
    [anon_sym_IMP] = ACTIONS(2772),
    [anon_sym_SEL] = ACTIONS(2772),
    [anon_sym_Class] = ACTIONS(2772),
    [anon_sym_id] = ACTIONS(2772),
  },
  [STATE(331)] = {
    [sym_identifier] = ACTIONS(2776),
    [aux_sym_preproc_include_token1] = ACTIONS(2776),
    [aux_sym_preproc_include_token2] = ACTIONS(2776),
    [aux_sym_preproc_def_token1] = ACTIONS(2776),
    [aux_sym_preproc_if_token1] = ACTIONS(2776),
    [aux_sym_preproc_if_token2] = ACTIONS(2776),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2776),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2776),
    [aux_sym_preproc_else_token1] = ACTIONS(2776),
    [aux_sym_preproc_elif_token1] = ACTIONS(2776),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2776),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2776),
    [sym_preproc_directive] = ACTIONS(2776),
    [anon_sym_LPAREN2] = ACTIONS(2778),
    [anon_sym_BANG] = ACTIONS(2778),
    [anon_sym_TILDE] = ACTIONS(2778),
    [anon_sym_DASH] = ACTIONS(2776),
    [anon_sym_PLUS] = ACTIONS(2776),
    [anon_sym_STAR] = ACTIONS(2778),
    [anon_sym_CARET] = ACTIONS(2778),
    [anon_sym_AMP] = ACTIONS(2778),
    [anon_sym_SEMI] = ACTIONS(2778),
    [anon_sym___extension__] = ACTIONS(2776),
    [anon_sym_typedef] = ACTIONS(2776),
    [anon_sym_extern] = ACTIONS(2776),
    [anon_sym___attribute__] = ACTIONS(2776),
    [anon_sym___attribute] = ACTIONS(2776),
    [anon_sym_noreturn] = ACTIONS(2776),
    [anon_sym_LBRACK] = ACTIONS(2778),
    [anon_sym___declspec] = ACTIONS(2776),
    [anon_sym___cdecl] = ACTIONS(2776),
    [anon_sym___clrcall] = ACTIONS(2776),
    [anon_sym___stdcall] = ACTIONS(2776),
    [anon_sym___fastcall] = ACTIONS(2776),
    [anon_sym___thiscall] = ACTIONS(2776),
    [anon_sym___vectorcall] = ACTIONS(2776),
    [anon_sym_LBRACE] = ACTIONS(2778),
    [anon_sym_signed] = ACTIONS(2776),
    [anon_sym_unsigned] = ACTIONS(2776),
    [anon_sym_long] = ACTIONS(2776),
    [anon_sym_short] = ACTIONS(2776),
    [anon_sym_static] = ACTIONS(2776),
    [anon_sym_ATautoreleasepool] = ACTIONS(2778),
    [anon_sym_auto] = ACTIONS(2776),
    [anon_sym_register] = ACTIONS(2776),
    [anon_sym_inline] = ACTIONS(2776),
    [anon_sym___inline] = ACTIONS(2776),
    [anon_sym___inline__] = ACTIONS(2776),
    [anon_sym___forceinline] = ACTIONS(2776),
    [anon_sym_thread_local] = ACTIONS(2776),
    [anon_sym___thread] = ACTIONS(2776),
    [anon_sym_CG_EXTERN] = ACTIONS(2776),
    [anon_sym_CG_INLINE] = ACTIONS(2776),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2776),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2776),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2776),
    [anon_sym_IBOutlet] = ACTIONS(2776),
    [anon_sym_IBInspectable] = ACTIONS(2776),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2776),
    [anon_sym_NS_INLINE] = ACTIONS(2776),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2776),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2776),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2776),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2776),
    [anon_sym_const] = ACTIONS(2776),
    [anon_sym_constexpr] = ACTIONS(2776),
    [anon_sym_volatile] = ACTIONS(2776),
    [anon_sym_restrict] = ACTIONS(2776),
    [anon_sym___restrict__] = ACTIONS(2776),
    [anon_sym__Atomic] = ACTIONS(2776),
    [anon_sym__Noreturn] = ACTIONS(2776),
    [anon_sym__Nonnull] = ACTIONS(2776),
    [anon_sym_nullable] = ACTIONS(2776),
    [anon_sym__Complex] = ACTIONS(2776),
    [anon_sym__Nullable] = ACTIONS(2776),
    [anon_sym__Nullable_result] = ACTIONS(2776),
    [anon_sym__Null_unspecified] = ACTIONS(2776),
    [anon_sym___autoreleasing] = ACTIONS(2776),
    [anon_sym___block] = ACTIONS(2776),
    [anon_sym___bridge] = ACTIONS(2776),
    [anon_sym___bridge_retained] = ACTIONS(2776),
    [anon_sym___bridge_transfer] = ACTIONS(2776),
    [anon_sym___complex] = ACTIONS(2776),
    [anon_sym___const] = ACTIONS(2776),
    [anon_sym___imag] = ACTIONS(2776),
    [anon_sym___kindof] = ACTIONS(2776),
    [anon_sym___nonnull] = ACTIONS(2776),
    [anon_sym___nullable] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2776),
    [anon_sym___real] = ACTIONS(2776),
    [anon_sym___strong] = ACTIONS(2776),
    [anon_sym___unsafe_unretained] = ACTIONS(2776),
    [anon_sym___unused] = ACTIONS(2776),
    [anon_sym___weak] = ACTIONS(2776),
    [anon_sym_alignas] = ACTIONS(2776),
    [anon_sym__Alignas] = ACTIONS(2776),
    [sym_primitive_type] = ACTIONS(2776),
    [anon_sym_enum] = ACTIONS(2776),
    [anon_sym_struct] = ACTIONS(2776),
    [anon_sym_union] = ACTIONS(2776),
    [anon_sym_if] = ACTIONS(2776),
    [anon_sym_switch] = ACTIONS(2776),
    [anon_sym_case] = ACTIONS(2776),
    [anon_sym_default] = ACTIONS(2776),
    [anon_sym_while] = ACTIONS(2776),
    [anon_sym_do] = ACTIONS(2776),
    [anon_sym_for] = ACTIONS(2776),
    [anon_sym_in] = ACTIONS(2776),
    [anon_sym_return] = ACTIONS(2776),
    [anon_sym_break] = ACTIONS(2776),
    [anon_sym_continue] = ACTIONS(2776),
    [anon_sym_goto] = ACTIONS(2776),
    [anon_sym_DASH_DASH] = ACTIONS(2778),
    [anon_sym_PLUS_PLUS] = ACTIONS(2778),
    [anon_sym_sizeof] = ACTIONS(2776),
    [anon_sym___alignof__] = ACTIONS(2776),
    [anon_sym___alignof] = ACTIONS(2776),
    [anon_sym__alignof] = ACTIONS(2776),
    [anon_sym_alignof] = ACTIONS(2776),
    [anon_sym__Alignof] = ACTIONS(2776),
    [anon_sym_offsetof] = ACTIONS(2776),
    [anon_sym__Generic] = ACTIONS(2776),
    [anon_sym_asm] = ACTIONS(2776),
    [anon_sym___asm__] = ACTIONS(2776),
    [anon_sym___asm] = ACTIONS(2776),
    [sym_number_literal] = ACTIONS(2778),
    [anon_sym_L_SQUOTE] = ACTIONS(2778),
    [anon_sym_u_SQUOTE] = ACTIONS(2778),
    [anon_sym_U_SQUOTE] = ACTIONS(2778),
    [anon_sym_u8_SQUOTE] = ACTIONS(2778),
    [anon_sym_SQUOTE] = ACTIONS(2778),
    [anon_sym_AT] = ACTIONS(2776),
    [anon_sym_DQUOTE] = ACTIONS(2778),
    [anon_sym_L_DQUOTE] = ACTIONS(2778),
    [anon_sym_u_DQUOTE] = ACTIONS(2778),
    [anon_sym_U_DQUOTE] = ACTIONS(2778),
    [anon_sym_u8_DQUOTE] = ACTIONS(2778),
    [sym_true] = ACTIONS(2776),
    [sym_false] = ACTIONS(2776),
    [anon_sym_NULL] = ACTIONS(2776),
    [anon_sym_nullptr] = ACTIONS(2776),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2776),
    [anon_sym___typeof] = ACTIONS(2776),
    [anon_sym_typeof] = ACTIONS(2776),
    [anon_sym_ATimport] = ACTIONS(2778),
    [aux_sym_preproc_undef_token1] = ACTIONS(2776),
    [anon_sym_POUND] = ACTIONS(2776),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2776),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2776),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_API_AVAILABLE] = ACTIONS(2776),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_API_DEPRECATED] = ACTIONS(2776),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2776),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym___deprecated_msg] = ACTIONS(2776),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2776),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2776),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2776),
    [anon_sym_ATprotocol] = ACTIONS(2778),
    [anon_sym_ATinterface] = ACTIONS(2778),
    [anon_sym_ATimplementation] = ACTIONS(2778),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2778),
    [anon_sym_ATsynthesize] = ACTIONS(2778),
    [anon_sym_ATdynamic] = ACTIONS(2778),
    [anon_sym_ATtry] = ACTIONS(2778),
    [anon_sym___try] = ACTIONS(2776),
    [anon_sym_ATthrow] = ACTIONS(2778),
    [anon_sym_ATselector] = ACTIONS(2778),
    [anon_sym_ATavailable] = ACTIONS(2778),
    [anon_sym___builtin_available] = ACTIONS(2776),
    [anon_sym_va_arg] = ACTIONS(2776),
    [anon_sym_ATencode] = ACTIONS(2778),
    [anon_sym_ATsynchronized] = ACTIONS(2778),
    [anon_sym_BOOL] = ACTIONS(2776),
    [anon_sym_IMP] = ACTIONS(2776),
    [anon_sym_SEL] = ACTIONS(2776),
    [anon_sym_Class] = ACTIONS(2776),
    [anon_sym_id] = ACTIONS(2776),
  },
  [STATE(332)] = {
    [sym_identifier] = ACTIONS(2780),
    [aux_sym_preproc_include_token1] = ACTIONS(2780),
    [aux_sym_preproc_include_token2] = ACTIONS(2780),
    [aux_sym_preproc_def_token1] = ACTIONS(2780),
    [aux_sym_preproc_if_token1] = ACTIONS(2780),
    [aux_sym_preproc_if_token2] = ACTIONS(2780),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2780),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2780),
    [aux_sym_preproc_else_token1] = ACTIONS(2780),
    [aux_sym_preproc_elif_token1] = ACTIONS(2780),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2780),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2780),
    [sym_preproc_directive] = ACTIONS(2780),
    [anon_sym_LPAREN2] = ACTIONS(2782),
    [anon_sym_BANG] = ACTIONS(2782),
    [anon_sym_TILDE] = ACTIONS(2782),
    [anon_sym_DASH] = ACTIONS(2780),
    [anon_sym_PLUS] = ACTIONS(2780),
    [anon_sym_STAR] = ACTIONS(2782),
    [anon_sym_CARET] = ACTIONS(2782),
    [anon_sym_AMP] = ACTIONS(2782),
    [anon_sym_SEMI] = ACTIONS(2782),
    [anon_sym___extension__] = ACTIONS(2780),
    [anon_sym_typedef] = ACTIONS(2780),
    [anon_sym_extern] = ACTIONS(2780),
    [anon_sym___attribute__] = ACTIONS(2780),
    [anon_sym___attribute] = ACTIONS(2780),
    [anon_sym_noreturn] = ACTIONS(2780),
    [anon_sym_LBRACK] = ACTIONS(2782),
    [anon_sym___declspec] = ACTIONS(2780),
    [anon_sym___cdecl] = ACTIONS(2780),
    [anon_sym___clrcall] = ACTIONS(2780),
    [anon_sym___stdcall] = ACTIONS(2780),
    [anon_sym___fastcall] = ACTIONS(2780),
    [anon_sym___thiscall] = ACTIONS(2780),
    [anon_sym___vectorcall] = ACTIONS(2780),
    [anon_sym_LBRACE] = ACTIONS(2782),
    [anon_sym_signed] = ACTIONS(2780),
    [anon_sym_unsigned] = ACTIONS(2780),
    [anon_sym_long] = ACTIONS(2780),
    [anon_sym_short] = ACTIONS(2780),
    [anon_sym_static] = ACTIONS(2780),
    [anon_sym_ATautoreleasepool] = ACTIONS(2782),
    [anon_sym_auto] = ACTIONS(2780),
    [anon_sym_register] = ACTIONS(2780),
    [anon_sym_inline] = ACTIONS(2780),
    [anon_sym___inline] = ACTIONS(2780),
    [anon_sym___inline__] = ACTIONS(2780),
    [anon_sym___forceinline] = ACTIONS(2780),
    [anon_sym_thread_local] = ACTIONS(2780),
    [anon_sym___thread] = ACTIONS(2780),
    [anon_sym_CG_EXTERN] = ACTIONS(2780),
    [anon_sym_CG_INLINE] = ACTIONS(2780),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2780),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2780),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2780),
    [anon_sym_IBOutlet] = ACTIONS(2780),
    [anon_sym_IBInspectable] = ACTIONS(2780),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2780),
    [anon_sym_NS_INLINE] = ACTIONS(2780),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2780),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2780),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2780),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2780),
    [anon_sym_const] = ACTIONS(2780),
    [anon_sym_constexpr] = ACTIONS(2780),
    [anon_sym_volatile] = ACTIONS(2780),
    [anon_sym_restrict] = ACTIONS(2780),
    [anon_sym___restrict__] = ACTIONS(2780),
    [anon_sym__Atomic] = ACTIONS(2780),
    [anon_sym__Noreturn] = ACTIONS(2780),
    [anon_sym__Nonnull] = ACTIONS(2780),
    [anon_sym_nullable] = ACTIONS(2780),
    [anon_sym__Complex] = ACTIONS(2780),
    [anon_sym__Nullable] = ACTIONS(2780),
    [anon_sym__Nullable_result] = ACTIONS(2780),
    [anon_sym__Null_unspecified] = ACTIONS(2780),
    [anon_sym___autoreleasing] = ACTIONS(2780),
    [anon_sym___block] = ACTIONS(2780),
    [anon_sym___bridge] = ACTIONS(2780),
    [anon_sym___bridge_retained] = ACTIONS(2780),
    [anon_sym___bridge_transfer] = ACTIONS(2780),
    [anon_sym___complex] = ACTIONS(2780),
    [anon_sym___const] = ACTIONS(2780),
    [anon_sym___imag] = ACTIONS(2780),
    [anon_sym___kindof] = ACTIONS(2780),
    [anon_sym___nonnull] = ACTIONS(2780),
    [anon_sym___nullable] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2780),
    [anon_sym___real] = ACTIONS(2780),
    [anon_sym___strong] = ACTIONS(2780),
    [anon_sym___unsafe_unretained] = ACTIONS(2780),
    [anon_sym___unused] = ACTIONS(2780),
    [anon_sym___weak] = ACTIONS(2780),
    [anon_sym_alignas] = ACTIONS(2780),
    [anon_sym__Alignas] = ACTIONS(2780),
    [sym_primitive_type] = ACTIONS(2780),
    [anon_sym_enum] = ACTIONS(2780),
    [anon_sym_struct] = ACTIONS(2780),
    [anon_sym_union] = ACTIONS(2780),
    [anon_sym_if] = ACTIONS(2780),
    [anon_sym_switch] = ACTIONS(2780),
    [anon_sym_case] = ACTIONS(2780),
    [anon_sym_default] = ACTIONS(2780),
    [anon_sym_while] = ACTIONS(2780),
    [anon_sym_do] = ACTIONS(2780),
    [anon_sym_for] = ACTIONS(2780),
    [anon_sym_in] = ACTIONS(2780),
    [anon_sym_return] = ACTIONS(2780),
    [anon_sym_break] = ACTIONS(2780),
    [anon_sym_continue] = ACTIONS(2780),
    [anon_sym_goto] = ACTIONS(2780),
    [anon_sym_DASH_DASH] = ACTIONS(2782),
    [anon_sym_PLUS_PLUS] = ACTIONS(2782),
    [anon_sym_sizeof] = ACTIONS(2780),
    [anon_sym___alignof__] = ACTIONS(2780),
    [anon_sym___alignof] = ACTIONS(2780),
    [anon_sym__alignof] = ACTIONS(2780),
    [anon_sym_alignof] = ACTIONS(2780),
    [anon_sym__Alignof] = ACTIONS(2780),
    [anon_sym_offsetof] = ACTIONS(2780),
    [anon_sym__Generic] = ACTIONS(2780),
    [anon_sym_asm] = ACTIONS(2780),
    [anon_sym___asm__] = ACTIONS(2780),
    [anon_sym___asm] = ACTIONS(2780),
    [sym_number_literal] = ACTIONS(2782),
    [anon_sym_L_SQUOTE] = ACTIONS(2782),
    [anon_sym_u_SQUOTE] = ACTIONS(2782),
    [anon_sym_U_SQUOTE] = ACTIONS(2782),
    [anon_sym_u8_SQUOTE] = ACTIONS(2782),
    [anon_sym_SQUOTE] = ACTIONS(2782),
    [anon_sym_AT] = ACTIONS(2780),
    [anon_sym_DQUOTE] = ACTIONS(2782),
    [anon_sym_L_DQUOTE] = ACTIONS(2782),
    [anon_sym_u_DQUOTE] = ACTIONS(2782),
    [anon_sym_U_DQUOTE] = ACTIONS(2782),
    [anon_sym_u8_DQUOTE] = ACTIONS(2782),
    [sym_true] = ACTIONS(2780),
    [sym_false] = ACTIONS(2780),
    [anon_sym_NULL] = ACTIONS(2780),
    [anon_sym_nullptr] = ACTIONS(2780),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2780),
    [anon_sym___typeof] = ACTIONS(2780),
    [anon_sym_typeof] = ACTIONS(2780),
    [anon_sym_ATimport] = ACTIONS(2782),
    [aux_sym_preproc_undef_token1] = ACTIONS(2780),
    [anon_sym_POUND] = ACTIONS(2780),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2780),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2780),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2780),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2780),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2780),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2780),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2780),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_API_AVAILABLE] = ACTIONS(2780),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_API_DEPRECATED] = ACTIONS(2780),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2780),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym___deprecated_msg] = ACTIONS(2780),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2780),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2780),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2780),
    [anon_sym_ATprotocol] = ACTIONS(2782),
    [anon_sym_ATinterface] = ACTIONS(2782),
    [anon_sym_ATimplementation] = ACTIONS(2782),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2782),
    [anon_sym_ATsynthesize] = ACTIONS(2782),
    [anon_sym_ATdynamic] = ACTIONS(2782),
    [anon_sym_ATtry] = ACTIONS(2782),
    [anon_sym___try] = ACTIONS(2780),
    [anon_sym_ATthrow] = ACTIONS(2782),
    [anon_sym_ATselector] = ACTIONS(2782),
    [anon_sym_ATavailable] = ACTIONS(2782),
    [anon_sym___builtin_available] = ACTIONS(2780),
    [anon_sym_va_arg] = ACTIONS(2780),
    [anon_sym_ATencode] = ACTIONS(2782),
    [anon_sym_ATsynchronized] = ACTIONS(2782),
    [anon_sym_BOOL] = ACTIONS(2780),
    [anon_sym_IMP] = ACTIONS(2780),
    [anon_sym_SEL] = ACTIONS(2780),
    [anon_sym_Class] = ACTIONS(2780),
    [anon_sym_id] = ACTIONS(2780),
  },
  [STATE(333)] = {
    [sym_identifier] = ACTIONS(2784),
    [aux_sym_preproc_include_token1] = ACTIONS(2784),
    [aux_sym_preproc_include_token2] = ACTIONS(2784),
    [aux_sym_preproc_def_token1] = ACTIONS(2784),
    [aux_sym_preproc_if_token1] = ACTIONS(2784),
    [aux_sym_preproc_if_token2] = ACTIONS(2784),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2784),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2784),
    [aux_sym_preproc_else_token1] = ACTIONS(2784),
    [aux_sym_preproc_elif_token1] = ACTIONS(2784),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2784),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2784),
    [sym_preproc_directive] = ACTIONS(2784),
    [anon_sym_LPAREN2] = ACTIONS(2786),
    [anon_sym_BANG] = ACTIONS(2786),
    [anon_sym_TILDE] = ACTIONS(2786),
    [anon_sym_DASH] = ACTIONS(2784),
    [anon_sym_PLUS] = ACTIONS(2784),
    [anon_sym_STAR] = ACTIONS(2786),
    [anon_sym_CARET] = ACTIONS(2786),
    [anon_sym_AMP] = ACTIONS(2786),
    [anon_sym_SEMI] = ACTIONS(2786),
    [anon_sym___extension__] = ACTIONS(2784),
    [anon_sym_typedef] = ACTIONS(2784),
    [anon_sym_extern] = ACTIONS(2784),
    [anon_sym___attribute__] = ACTIONS(2784),
    [anon_sym___attribute] = ACTIONS(2784),
    [anon_sym_noreturn] = ACTIONS(2784),
    [anon_sym_LBRACK] = ACTIONS(2786),
    [anon_sym___declspec] = ACTIONS(2784),
    [anon_sym___cdecl] = ACTIONS(2784),
    [anon_sym___clrcall] = ACTIONS(2784),
    [anon_sym___stdcall] = ACTIONS(2784),
    [anon_sym___fastcall] = ACTIONS(2784),
    [anon_sym___thiscall] = ACTIONS(2784),
    [anon_sym___vectorcall] = ACTIONS(2784),
    [anon_sym_LBRACE] = ACTIONS(2786),
    [anon_sym_signed] = ACTIONS(2784),
    [anon_sym_unsigned] = ACTIONS(2784),
    [anon_sym_long] = ACTIONS(2784),
    [anon_sym_short] = ACTIONS(2784),
    [anon_sym_static] = ACTIONS(2784),
    [anon_sym_ATautoreleasepool] = ACTIONS(2786),
    [anon_sym_auto] = ACTIONS(2784),
    [anon_sym_register] = ACTIONS(2784),
    [anon_sym_inline] = ACTIONS(2784),
    [anon_sym___inline] = ACTIONS(2784),
    [anon_sym___inline__] = ACTIONS(2784),
    [anon_sym___forceinline] = ACTIONS(2784),
    [anon_sym_thread_local] = ACTIONS(2784),
    [anon_sym___thread] = ACTIONS(2784),
    [anon_sym_CG_EXTERN] = ACTIONS(2784),
    [anon_sym_CG_INLINE] = ACTIONS(2784),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2784),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2784),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2784),
    [anon_sym_IBOutlet] = ACTIONS(2784),
    [anon_sym_IBInspectable] = ACTIONS(2784),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2784),
    [anon_sym_NS_INLINE] = ACTIONS(2784),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2784),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2784),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2784),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2784),
    [anon_sym_const] = ACTIONS(2784),
    [anon_sym_constexpr] = ACTIONS(2784),
    [anon_sym_volatile] = ACTIONS(2784),
    [anon_sym_restrict] = ACTIONS(2784),
    [anon_sym___restrict__] = ACTIONS(2784),
    [anon_sym__Atomic] = ACTIONS(2784),
    [anon_sym__Noreturn] = ACTIONS(2784),
    [anon_sym__Nonnull] = ACTIONS(2784),
    [anon_sym_nullable] = ACTIONS(2784),
    [anon_sym__Complex] = ACTIONS(2784),
    [anon_sym__Nullable] = ACTIONS(2784),
    [anon_sym__Nullable_result] = ACTIONS(2784),
    [anon_sym__Null_unspecified] = ACTIONS(2784),
    [anon_sym___autoreleasing] = ACTIONS(2784),
    [anon_sym___block] = ACTIONS(2784),
    [anon_sym___bridge] = ACTIONS(2784),
    [anon_sym___bridge_retained] = ACTIONS(2784),
    [anon_sym___bridge_transfer] = ACTIONS(2784),
    [anon_sym___complex] = ACTIONS(2784),
    [anon_sym___const] = ACTIONS(2784),
    [anon_sym___imag] = ACTIONS(2784),
    [anon_sym___kindof] = ACTIONS(2784),
    [anon_sym___nonnull] = ACTIONS(2784),
    [anon_sym___nullable] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2784),
    [anon_sym___real] = ACTIONS(2784),
    [anon_sym___strong] = ACTIONS(2784),
    [anon_sym___unsafe_unretained] = ACTIONS(2784),
    [anon_sym___unused] = ACTIONS(2784),
    [anon_sym___weak] = ACTIONS(2784),
    [anon_sym_alignas] = ACTIONS(2784),
    [anon_sym__Alignas] = ACTIONS(2784),
    [sym_primitive_type] = ACTIONS(2784),
    [anon_sym_enum] = ACTIONS(2784),
    [anon_sym_struct] = ACTIONS(2784),
    [anon_sym_union] = ACTIONS(2784),
    [anon_sym_if] = ACTIONS(2784),
    [anon_sym_switch] = ACTIONS(2784),
    [anon_sym_case] = ACTIONS(2784),
    [anon_sym_default] = ACTIONS(2784),
    [anon_sym_while] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2784),
    [anon_sym_for] = ACTIONS(2784),
    [anon_sym_in] = ACTIONS(2784),
    [anon_sym_return] = ACTIONS(2784),
    [anon_sym_break] = ACTIONS(2784),
    [anon_sym_continue] = ACTIONS(2784),
    [anon_sym_goto] = ACTIONS(2784),
    [anon_sym_DASH_DASH] = ACTIONS(2786),
    [anon_sym_PLUS_PLUS] = ACTIONS(2786),
    [anon_sym_sizeof] = ACTIONS(2784),
    [anon_sym___alignof__] = ACTIONS(2784),
    [anon_sym___alignof] = ACTIONS(2784),
    [anon_sym__alignof] = ACTIONS(2784),
    [anon_sym_alignof] = ACTIONS(2784),
    [anon_sym__Alignof] = ACTIONS(2784),
    [anon_sym_offsetof] = ACTIONS(2784),
    [anon_sym__Generic] = ACTIONS(2784),
    [anon_sym_asm] = ACTIONS(2784),
    [anon_sym___asm__] = ACTIONS(2784),
    [anon_sym___asm] = ACTIONS(2784),
    [sym_number_literal] = ACTIONS(2786),
    [anon_sym_L_SQUOTE] = ACTIONS(2786),
    [anon_sym_u_SQUOTE] = ACTIONS(2786),
    [anon_sym_U_SQUOTE] = ACTIONS(2786),
    [anon_sym_u8_SQUOTE] = ACTIONS(2786),
    [anon_sym_SQUOTE] = ACTIONS(2786),
    [anon_sym_AT] = ACTIONS(2784),
    [anon_sym_DQUOTE] = ACTIONS(2786),
    [anon_sym_L_DQUOTE] = ACTIONS(2786),
    [anon_sym_u_DQUOTE] = ACTIONS(2786),
    [anon_sym_U_DQUOTE] = ACTIONS(2786),
    [anon_sym_u8_DQUOTE] = ACTIONS(2786),
    [sym_true] = ACTIONS(2784),
    [sym_false] = ACTIONS(2784),
    [anon_sym_NULL] = ACTIONS(2784),
    [anon_sym_nullptr] = ACTIONS(2784),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2784),
    [anon_sym___typeof] = ACTIONS(2784),
    [anon_sym_typeof] = ACTIONS(2784),
    [anon_sym_ATimport] = ACTIONS(2786),
    [aux_sym_preproc_undef_token1] = ACTIONS(2784),
    [anon_sym_POUND] = ACTIONS(2784),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2784),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2784),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2784),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2784),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2784),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2784),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2784),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_API_AVAILABLE] = ACTIONS(2784),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_API_DEPRECATED] = ACTIONS(2784),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2784),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym___deprecated_msg] = ACTIONS(2784),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2784),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2784),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2784),
    [anon_sym_ATprotocol] = ACTIONS(2786),
    [anon_sym_ATinterface] = ACTIONS(2786),
    [anon_sym_ATimplementation] = ACTIONS(2786),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2786),
    [anon_sym_ATsynthesize] = ACTIONS(2786),
    [anon_sym_ATdynamic] = ACTIONS(2786),
    [anon_sym_ATtry] = ACTIONS(2786),
    [anon_sym___try] = ACTIONS(2784),
    [anon_sym_ATthrow] = ACTIONS(2786),
    [anon_sym_ATselector] = ACTIONS(2786),
    [anon_sym_ATavailable] = ACTIONS(2786),
    [anon_sym___builtin_available] = ACTIONS(2784),
    [anon_sym_va_arg] = ACTIONS(2784),
    [anon_sym_ATencode] = ACTIONS(2786),
    [anon_sym_ATsynchronized] = ACTIONS(2786),
    [anon_sym_BOOL] = ACTIONS(2784),
    [anon_sym_IMP] = ACTIONS(2784),
    [anon_sym_SEL] = ACTIONS(2784),
    [anon_sym_Class] = ACTIONS(2784),
    [anon_sym_id] = ACTIONS(2784),
  },
  [STATE(334)] = {
    [sym_identifier] = ACTIONS(2788),
    [aux_sym_preproc_include_token1] = ACTIONS(2788),
    [aux_sym_preproc_include_token2] = ACTIONS(2788),
    [aux_sym_preproc_def_token1] = ACTIONS(2788),
    [aux_sym_preproc_if_token1] = ACTIONS(2788),
    [aux_sym_preproc_if_token2] = ACTIONS(2788),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2788),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2788),
    [aux_sym_preproc_else_token1] = ACTIONS(2788),
    [aux_sym_preproc_elif_token1] = ACTIONS(2788),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2788),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2788),
    [sym_preproc_directive] = ACTIONS(2788),
    [anon_sym_LPAREN2] = ACTIONS(2790),
    [anon_sym_BANG] = ACTIONS(2790),
    [anon_sym_TILDE] = ACTIONS(2790),
    [anon_sym_DASH] = ACTIONS(2788),
    [anon_sym_PLUS] = ACTIONS(2788),
    [anon_sym_STAR] = ACTIONS(2790),
    [anon_sym_CARET] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(2790),
    [anon_sym_SEMI] = ACTIONS(2790),
    [anon_sym___extension__] = ACTIONS(2788),
    [anon_sym_typedef] = ACTIONS(2788),
    [anon_sym_extern] = ACTIONS(2788),
    [anon_sym___attribute__] = ACTIONS(2788),
    [anon_sym___attribute] = ACTIONS(2788),
    [anon_sym_noreturn] = ACTIONS(2788),
    [anon_sym_LBRACK] = ACTIONS(2790),
    [anon_sym___declspec] = ACTIONS(2788),
    [anon_sym___cdecl] = ACTIONS(2788),
    [anon_sym___clrcall] = ACTIONS(2788),
    [anon_sym___stdcall] = ACTIONS(2788),
    [anon_sym___fastcall] = ACTIONS(2788),
    [anon_sym___thiscall] = ACTIONS(2788),
    [anon_sym___vectorcall] = ACTIONS(2788),
    [anon_sym_LBRACE] = ACTIONS(2790),
    [anon_sym_signed] = ACTIONS(2788),
    [anon_sym_unsigned] = ACTIONS(2788),
    [anon_sym_long] = ACTIONS(2788),
    [anon_sym_short] = ACTIONS(2788),
    [anon_sym_static] = ACTIONS(2788),
    [anon_sym_ATautoreleasepool] = ACTIONS(2790),
    [anon_sym_auto] = ACTIONS(2788),
    [anon_sym_register] = ACTIONS(2788),
    [anon_sym_inline] = ACTIONS(2788),
    [anon_sym___inline] = ACTIONS(2788),
    [anon_sym___inline__] = ACTIONS(2788),
    [anon_sym___forceinline] = ACTIONS(2788),
    [anon_sym_thread_local] = ACTIONS(2788),
    [anon_sym___thread] = ACTIONS(2788),
    [anon_sym_CG_EXTERN] = ACTIONS(2788),
    [anon_sym_CG_INLINE] = ACTIONS(2788),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2788),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2788),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2788),
    [anon_sym_IBOutlet] = ACTIONS(2788),
    [anon_sym_IBInspectable] = ACTIONS(2788),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2788),
    [anon_sym_NS_INLINE] = ACTIONS(2788),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2788),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2788),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2788),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2788),
    [anon_sym_const] = ACTIONS(2788),
    [anon_sym_constexpr] = ACTIONS(2788),
    [anon_sym_volatile] = ACTIONS(2788),
    [anon_sym_restrict] = ACTIONS(2788),
    [anon_sym___restrict__] = ACTIONS(2788),
    [anon_sym__Atomic] = ACTIONS(2788),
    [anon_sym__Noreturn] = ACTIONS(2788),
    [anon_sym__Nonnull] = ACTIONS(2788),
    [anon_sym_nullable] = ACTIONS(2788),
    [anon_sym__Complex] = ACTIONS(2788),
    [anon_sym__Nullable] = ACTIONS(2788),
    [anon_sym__Nullable_result] = ACTIONS(2788),
    [anon_sym__Null_unspecified] = ACTIONS(2788),
    [anon_sym___autoreleasing] = ACTIONS(2788),
    [anon_sym___block] = ACTIONS(2788),
    [anon_sym___bridge] = ACTIONS(2788),
    [anon_sym___bridge_retained] = ACTIONS(2788),
    [anon_sym___bridge_transfer] = ACTIONS(2788),
    [anon_sym___complex] = ACTIONS(2788),
    [anon_sym___const] = ACTIONS(2788),
    [anon_sym___imag] = ACTIONS(2788),
    [anon_sym___kindof] = ACTIONS(2788),
    [anon_sym___nonnull] = ACTIONS(2788),
    [anon_sym___nullable] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2788),
    [anon_sym___real] = ACTIONS(2788),
    [anon_sym___strong] = ACTIONS(2788),
    [anon_sym___unsafe_unretained] = ACTIONS(2788),
    [anon_sym___unused] = ACTIONS(2788),
    [anon_sym___weak] = ACTIONS(2788),
    [anon_sym_alignas] = ACTIONS(2788),
    [anon_sym__Alignas] = ACTIONS(2788),
    [sym_primitive_type] = ACTIONS(2788),
    [anon_sym_enum] = ACTIONS(2788),
    [anon_sym_struct] = ACTIONS(2788),
    [anon_sym_union] = ACTIONS(2788),
    [anon_sym_if] = ACTIONS(2788),
    [anon_sym_switch] = ACTIONS(2788),
    [anon_sym_case] = ACTIONS(2788),
    [anon_sym_default] = ACTIONS(2788),
    [anon_sym_while] = ACTIONS(2788),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_for] = ACTIONS(2788),
    [anon_sym_in] = ACTIONS(2788),
    [anon_sym_return] = ACTIONS(2788),
    [anon_sym_break] = ACTIONS(2788),
    [anon_sym_continue] = ACTIONS(2788),
    [anon_sym_goto] = ACTIONS(2788),
    [anon_sym_DASH_DASH] = ACTIONS(2790),
    [anon_sym_PLUS_PLUS] = ACTIONS(2790),
    [anon_sym_sizeof] = ACTIONS(2788),
    [anon_sym___alignof__] = ACTIONS(2788),
    [anon_sym___alignof] = ACTIONS(2788),
    [anon_sym__alignof] = ACTIONS(2788),
    [anon_sym_alignof] = ACTIONS(2788),
    [anon_sym__Alignof] = ACTIONS(2788),
    [anon_sym_offsetof] = ACTIONS(2788),
    [anon_sym__Generic] = ACTIONS(2788),
    [anon_sym_asm] = ACTIONS(2788),
    [anon_sym___asm__] = ACTIONS(2788),
    [anon_sym___asm] = ACTIONS(2788),
    [sym_number_literal] = ACTIONS(2790),
    [anon_sym_L_SQUOTE] = ACTIONS(2790),
    [anon_sym_u_SQUOTE] = ACTIONS(2790),
    [anon_sym_U_SQUOTE] = ACTIONS(2790),
    [anon_sym_u8_SQUOTE] = ACTIONS(2790),
    [anon_sym_SQUOTE] = ACTIONS(2790),
    [anon_sym_AT] = ACTIONS(2788),
    [anon_sym_DQUOTE] = ACTIONS(2790),
    [anon_sym_L_DQUOTE] = ACTIONS(2790),
    [anon_sym_u_DQUOTE] = ACTIONS(2790),
    [anon_sym_U_DQUOTE] = ACTIONS(2790),
    [anon_sym_u8_DQUOTE] = ACTIONS(2790),
    [sym_true] = ACTIONS(2788),
    [sym_false] = ACTIONS(2788),
    [anon_sym_NULL] = ACTIONS(2788),
    [anon_sym_nullptr] = ACTIONS(2788),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2788),
    [anon_sym___typeof] = ACTIONS(2788),
    [anon_sym_typeof] = ACTIONS(2788),
    [anon_sym_ATimport] = ACTIONS(2790),
    [aux_sym_preproc_undef_token1] = ACTIONS(2788),
    [anon_sym_POUND] = ACTIONS(2788),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2788),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2788),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2788),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2788),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2788),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2788),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2788),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_API_AVAILABLE] = ACTIONS(2788),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_API_DEPRECATED] = ACTIONS(2788),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2788),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym___deprecated_msg] = ACTIONS(2788),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2788),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2788),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2788),
    [anon_sym_ATprotocol] = ACTIONS(2790),
    [anon_sym_ATinterface] = ACTIONS(2790),
    [anon_sym_ATimplementation] = ACTIONS(2790),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2790),
    [anon_sym_ATsynthesize] = ACTIONS(2790),
    [anon_sym_ATdynamic] = ACTIONS(2790),
    [anon_sym_ATtry] = ACTIONS(2790),
    [anon_sym___try] = ACTIONS(2788),
    [anon_sym_ATthrow] = ACTIONS(2790),
    [anon_sym_ATselector] = ACTIONS(2790),
    [anon_sym_ATavailable] = ACTIONS(2790),
    [anon_sym___builtin_available] = ACTIONS(2788),
    [anon_sym_va_arg] = ACTIONS(2788),
    [anon_sym_ATencode] = ACTIONS(2790),
    [anon_sym_ATsynchronized] = ACTIONS(2790),
    [anon_sym_BOOL] = ACTIONS(2788),
    [anon_sym_IMP] = ACTIONS(2788),
    [anon_sym_SEL] = ACTIONS(2788),
    [anon_sym_Class] = ACTIONS(2788),
    [anon_sym_id] = ACTIONS(2788),
  },
  [STATE(335)] = {
    [sym_identifier] = ACTIONS(2792),
    [aux_sym_preproc_include_token1] = ACTIONS(2792),
    [aux_sym_preproc_include_token2] = ACTIONS(2792),
    [aux_sym_preproc_def_token1] = ACTIONS(2792),
    [aux_sym_preproc_if_token1] = ACTIONS(2792),
    [aux_sym_preproc_if_token2] = ACTIONS(2792),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2792),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2792),
    [aux_sym_preproc_else_token1] = ACTIONS(2792),
    [aux_sym_preproc_elif_token1] = ACTIONS(2792),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2792),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2792),
    [sym_preproc_directive] = ACTIONS(2792),
    [anon_sym_LPAREN2] = ACTIONS(2794),
    [anon_sym_BANG] = ACTIONS(2794),
    [anon_sym_TILDE] = ACTIONS(2794),
    [anon_sym_DASH] = ACTIONS(2792),
    [anon_sym_PLUS] = ACTIONS(2792),
    [anon_sym_STAR] = ACTIONS(2794),
    [anon_sym_CARET] = ACTIONS(2794),
    [anon_sym_AMP] = ACTIONS(2794),
    [anon_sym_SEMI] = ACTIONS(2794),
    [anon_sym___extension__] = ACTIONS(2792),
    [anon_sym_typedef] = ACTIONS(2792),
    [anon_sym_extern] = ACTIONS(2792),
    [anon_sym___attribute__] = ACTIONS(2792),
    [anon_sym___attribute] = ACTIONS(2792),
    [anon_sym_noreturn] = ACTIONS(2792),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym___declspec] = ACTIONS(2792),
    [anon_sym___cdecl] = ACTIONS(2792),
    [anon_sym___clrcall] = ACTIONS(2792),
    [anon_sym___stdcall] = ACTIONS(2792),
    [anon_sym___fastcall] = ACTIONS(2792),
    [anon_sym___thiscall] = ACTIONS(2792),
    [anon_sym___vectorcall] = ACTIONS(2792),
    [anon_sym_LBRACE] = ACTIONS(2794),
    [anon_sym_signed] = ACTIONS(2792),
    [anon_sym_unsigned] = ACTIONS(2792),
    [anon_sym_long] = ACTIONS(2792),
    [anon_sym_short] = ACTIONS(2792),
    [anon_sym_static] = ACTIONS(2792),
    [anon_sym_ATautoreleasepool] = ACTIONS(2794),
    [anon_sym_auto] = ACTIONS(2792),
    [anon_sym_register] = ACTIONS(2792),
    [anon_sym_inline] = ACTIONS(2792),
    [anon_sym___inline] = ACTIONS(2792),
    [anon_sym___inline__] = ACTIONS(2792),
    [anon_sym___forceinline] = ACTIONS(2792),
    [anon_sym_thread_local] = ACTIONS(2792),
    [anon_sym___thread] = ACTIONS(2792),
    [anon_sym_CG_EXTERN] = ACTIONS(2792),
    [anon_sym_CG_INLINE] = ACTIONS(2792),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2792),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2792),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2792),
    [anon_sym_IBOutlet] = ACTIONS(2792),
    [anon_sym_IBInspectable] = ACTIONS(2792),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2792),
    [anon_sym_NS_INLINE] = ACTIONS(2792),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2792),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2792),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2792),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2792),
    [anon_sym_const] = ACTIONS(2792),
    [anon_sym_constexpr] = ACTIONS(2792),
    [anon_sym_volatile] = ACTIONS(2792),
    [anon_sym_restrict] = ACTIONS(2792),
    [anon_sym___restrict__] = ACTIONS(2792),
    [anon_sym__Atomic] = ACTIONS(2792),
    [anon_sym__Noreturn] = ACTIONS(2792),
    [anon_sym__Nonnull] = ACTIONS(2792),
    [anon_sym_nullable] = ACTIONS(2792),
    [anon_sym__Complex] = ACTIONS(2792),
    [anon_sym__Nullable] = ACTIONS(2792),
    [anon_sym__Nullable_result] = ACTIONS(2792),
    [anon_sym__Null_unspecified] = ACTIONS(2792),
    [anon_sym___autoreleasing] = ACTIONS(2792),
    [anon_sym___block] = ACTIONS(2792),
    [anon_sym___bridge] = ACTIONS(2792),
    [anon_sym___bridge_retained] = ACTIONS(2792),
    [anon_sym___bridge_transfer] = ACTIONS(2792),
    [anon_sym___complex] = ACTIONS(2792),
    [anon_sym___const] = ACTIONS(2792),
    [anon_sym___imag] = ACTIONS(2792),
    [anon_sym___kindof] = ACTIONS(2792),
    [anon_sym___nonnull] = ACTIONS(2792),
    [anon_sym___nullable] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2792),
    [anon_sym___real] = ACTIONS(2792),
    [anon_sym___strong] = ACTIONS(2792),
    [anon_sym___unsafe_unretained] = ACTIONS(2792),
    [anon_sym___unused] = ACTIONS(2792),
    [anon_sym___weak] = ACTIONS(2792),
    [anon_sym_alignas] = ACTIONS(2792),
    [anon_sym__Alignas] = ACTIONS(2792),
    [sym_primitive_type] = ACTIONS(2792),
    [anon_sym_enum] = ACTIONS(2792),
    [anon_sym_struct] = ACTIONS(2792),
    [anon_sym_union] = ACTIONS(2792),
    [anon_sym_if] = ACTIONS(2792),
    [anon_sym_switch] = ACTIONS(2792),
    [anon_sym_case] = ACTIONS(2792),
    [anon_sym_default] = ACTIONS(2792),
    [anon_sym_while] = ACTIONS(2792),
    [anon_sym_do] = ACTIONS(2792),
    [anon_sym_for] = ACTIONS(2792),
    [anon_sym_in] = ACTIONS(2792),
    [anon_sym_return] = ACTIONS(2792),
    [anon_sym_break] = ACTIONS(2792),
    [anon_sym_continue] = ACTIONS(2792),
    [anon_sym_goto] = ACTIONS(2792),
    [anon_sym_DASH_DASH] = ACTIONS(2794),
    [anon_sym_PLUS_PLUS] = ACTIONS(2794),
    [anon_sym_sizeof] = ACTIONS(2792),
    [anon_sym___alignof__] = ACTIONS(2792),
    [anon_sym___alignof] = ACTIONS(2792),
    [anon_sym__alignof] = ACTIONS(2792),
    [anon_sym_alignof] = ACTIONS(2792),
    [anon_sym__Alignof] = ACTIONS(2792),
    [anon_sym_offsetof] = ACTIONS(2792),
    [anon_sym__Generic] = ACTIONS(2792),
    [anon_sym_asm] = ACTIONS(2792),
    [anon_sym___asm__] = ACTIONS(2792),
    [anon_sym___asm] = ACTIONS(2792),
    [sym_number_literal] = ACTIONS(2794),
    [anon_sym_L_SQUOTE] = ACTIONS(2794),
    [anon_sym_u_SQUOTE] = ACTIONS(2794),
    [anon_sym_U_SQUOTE] = ACTIONS(2794),
    [anon_sym_u8_SQUOTE] = ACTIONS(2794),
    [anon_sym_SQUOTE] = ACTIONS(2794),
    [anon_sym_AT] = ACTIONS(2792),
    [anon_sym_DQUOTE] = ACTIONS(2794),
    [anon_sym_L_DQUOTE] = ACTIONS(2794),
    [anon_sym_u_DQUOTE] = ACTIONS(2794),
    [anon_sym_U_DQUOTE] = ACTIONS(2794),
    [anon_sym_u8_DQUOTE] = ACTIONS(2794),
    [sym_true] = ACTIONS(2792),
    [sym_false] = ACTIONS(2792),
    [anon_sym_NULL] = ACTIONS(2792),
    [anon_sym_nullptr] = ACTIONS(2792),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2792),
    [anon_sym___typeof] = ACTIONS(2792),
    [anon_sym_typeof] = ACTIONS(2792),
    [anon_sym_ATimport] = ACTIONS(2794),
    [aux_sym_preproc_undef_token1] = ACTIONS(2792),
    [anon_sym_POUND] = ACTIONS(2792),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2792),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2792),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2792),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2792),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2792),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2792),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2792),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_API_AVAILABLE] = ACTIONS(2792),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_API_DEPRECATED] = ACTIONS(2792),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2792),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym___deprecated_msg] = ACTIONS(2792),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2792),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2792),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2792),
    [anon_sym_ATprotocol] = ACTIONS(2794),
    [anon_sym_ATinterface] = ACTIONS(2794),
    [anon_sym_ATimplementation] = ACTIONS(2794),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2794),
    [anon_sym_ATsynthesize] = ACTIONS(2794),
    [anon_sym_ATdynamic] = ACTIONS(2794),
    [anon_sym_ATtry] = ACTIONS(2794),
    [anon_sym___try] = ACTIONS(2792),
    [anon_sym_ATthrow] = ACTIONS(2794),
    [anon_sym_ATselector] = ACTIONS(2794),
    [anon_sym_ATavailable] = ACTIONS(2794),
    [anon_sym___builtin_available] = ACTIONS(2792),
    [anon_sym_va_arg] = ACTIONS(2792),
    [anon_sym_ATencode] = ACTIONS(2794),
    [anon_sym_ATsynchronized] = ACTIONS(2794),
    [anon_sym_BOOL] = ACTIONS(2792),
    [anon_sym_IMP] = ACTIONS(2792),
    [anon_sym_SEL] = ACTIONS(2792),
    [anon_sym_Class] = ACTIONS(2792),
    [anon_sym_id] = ACTIONS(2792),
  },
  [STATE(336)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token2] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [aux_sym_preproc_else_token1] = ACTIONS(2796),
    [aux_sym_preproc_elif_token1] = ACTIONS(2796),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATsynthesize] = ACTIONS(2798),
    [anon_sym_ATdynamic] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(337)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token2] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [aux_sym_preproc_else_token1] = ACTIONS(2796),
    [aux_sym_preproc_elif_token1] = ACTIONS(2796),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATsynthesize] = ACTIONS(2798),
    [anon_sym_ATdynamic] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(338)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token2] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [aux_sym_preproc_else_token1] = ACTIONS(2796),
    [aux_sym_preproc_elif_token1] = ACTIONS(2796),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATsynthesize] = ACTIONS(2798),
    [anon_sym_ATdynamic] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(339)] = {
    [sym_identifier] = ACTIONS(2800),
    [aux_sym_preproc_include_token1] = ACTIONS(2800),
    [aux_sym_preproc_include_token2] = ACTIONS(2800),
    [aux_sym_preproc_def_token1] = ACTIONS(2800),
    [aux_sym_preproc_if_token1] = ACTIONS(2800),
    [aux_sym_preproc_if_token2] = ACTIONS(2800),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2800),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2800),
    [aux_sym_preproc_else_token1] = ACTIONS(2800),
    [aux_sym_preproc_elif_token1] = ACTIONS(2800),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2800),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2800),
    [sym_preproc_directive] = ACTIONS(2800),
    [anon_sym_LPAREN2] = ACTIONS(2802),
    [anon_sym_BANG] = ACTIONS(2802),
    [anon_sym_TILDE] = ACTIONS(2802),
    [anon_sym_DASH] = ACTIONS(2800),
    [anon_sym_PLUS] = ACTIONS(2800),
    [anon_sym_STAR] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(2802),
    [anon_sym_AMP] = ACTIONS(2802),
    [anon_sym_SEMI] = ACTIONS(2802),
    [anon_sym___extension__] = ACTIONS(2800),
    [anon_sym_typedef] = ACTIONS(2800),
    [anon_sym_extern] = ACTIONS(2800),
    [anon_sym___attribute__] = ACTIONS(2800),
    [anon_sym___attribute] = ACTIONS(2800),
    [anon_sym_noreturn] = ACTIONS(2800),
    [anon_sym_LBRACK] = ACTIONS(2802),
    [anon_sym___declspec] = ACTIONS(2800),
    [anon_sym___cdecl] = ACTIONS(2800),
    [anon_sym___clrcall] = ACTIONS(2800),
    [anon_sym___stdcall] = ACTIONS(2800),
    [anon_sym___fastcall] = ACTIONS(2800),
    [anon_sym___thiscall] = ACTIONS(2800),
    [anon_sym___vectorcall] = ACTIONS(2800),
    [anon_sym_LBRACE] = ACTIONS(2802),
    [anon_sym_signed] = ACTIONS(2800),
    [anon_sym_unsigned] = ACTIONS(2800),
    [anon_sym_long] = ACTIONS(2800),
    [anon_sym_short] = ACTIONS(2800),
    [anon_sym_static] = ACTIONS(2800),
    [anon_sym_ATautoreleasepool] = ACTIONS(2802),
    [anon_sym_auto] = ACTIONS(2800),
    [anon_sym_register] = ACTIONS(2800),
    [anon_sym_inline] = ACTIONS(2800),
    [anon_sym___inline] = ACTIONS(2800),
    [anon_sym___inline__] = ACTIONS(2800),
    [anon_sym___forceinline] = ACTIONS(2800),
    [anon_sym_thread_local] = ACTIONS(2800),
    [anon_sym___thread] = ACTIONS(2800),
    [anon_sym_CG_EXTERN] = ACTIONS(2800),
    [anon_sym_CG_INLINE] = ACTIONS(2800),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2800),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2800),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2800),
    [anon_sym_IBOutlet] = ACTIONS(2800),
    [anon_sym_IBInspectable] = ACTIONS(2800),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2800),
    [anon_sym_NS_INLINE] = ACTIONS(2800),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2800),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2800),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2800),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2800),
    [anon_sym_const] = ACTIONS(2800),
    [anon_sym_constexpr] = ACTIONS(2800),
    [anon_sym_volatile] = ACTIONS(2800),
    [anon_sym_restrict] = ACTIONS(2800),
    [anon_sym___restrict__] = ACTIONS(2800),
    [anon_sym__Atomic] = ACTIONS(2800),
    [anon_sym__Noreturn] = ACTIONS(2800),
    [anon_sym__Nonnull] = ACTIONS(2800),
    [anon_sym_nullable] = ACTIONS(2800),
    [anon_sym__Complex] = ACTIONS(2800),
    [anon_sym__Nullable] = ACTIONS(2800),
    [anon_sym__Nullable_result] = ACTIONS(2800),
    [anon_sym__Null_unspecified] = ACTIONS(2800),
    [anon_sym___autoreleasing] = ACTIONS(2800),
    [anon_sym___block] = ACTIONS(2800),
    [anon_sym___bridge] = ACTIONS(2800),
    [anon_sym___bridge_retained] = ACTIONS(2800),
    [anon_sym___bridge_transfer] = ACTIONS(2800),
    [anon_sym___complex] = ACTIONS(2800),
    [anon_sym___const] = ACTIONS(2800),
    [anon_sym___imag] = ACTIONS(2800),
    [anon_sym___kindof] = ACTIONS(2800),
    [anon_sym___nonnull] = ACTIONS(2800),
    [anon_sym___nullable] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2800),
    [anon_sym___real] = ACTIONS(2800),
    [anon_sym___strong] = ACTIONS(2800),
    [anon_sym___unsafe_unretained] = ACTIONS(2800),
    [anon_sym___unused] = ACTIONS(2800),
    [anon_sym___weak] = ACTIONS(2800),
    [anon_sym_alignas] = ACTIONS(2800),
    [anon_sym__Alignas] = ACTIONS(2800),
    [sym_primitive_type] = ACTIONS(2800),
    [anon_sym_enum] = ACTIONS(2800),
    [anon_sym_struct] = ACTIONS(2800),
    [anon_sym_union] = ACTIONS(2800),
    [anon_sym_if] = ACTIONS(2800),
    [anon_sym_switch] = ACTIONS(2800),
    [anon_sym_case] = ACTIONS(2800),
    [anon_sym_default] = ACTIONS(2800),
    [anon_sym_while] = ACTIONS(2800),
    [anon_sym_do] = ACTIONS(2800),
    [anon_sym_for] = ACTIONS(2800),
    [anon_sym_in] = ACTIONS(2800),
    [anon_sym_return] = ACTIONS(2800),
    [anon_sym_break] = ACTIONS(2800),
    [anon_sym_continue] = ACTIONS(2800),
    [anon_sym_goto] = ACTIONS(2800),
    [anon_sym_DASH_DASH] = ACTIONS(2802),
    [anon_sym_PLUS_PLUS] = ACTIONS(2802),
    [anon_sym_sizeof] = ACTIONS(2800),
    [anon_sym___alignof__] = ACTIONS(2800),
    [anon_sym___alignof] = ACTIONS(2800),
    [anon_sym__alignof] = ACTIONS(2800),
    [anon_sym_alignof] = ACTIONS(2800),
    [anon_sym__Alignof] = ACTIONS(2800),
    [anon_sym_offsetof] = ACTIONS(2800),
    [anon_sym__Generic] = ACTIONS(2800),
    [anon_sym_asm] = ACTIONS(2800),
    [anon_sym___asm__] = ACTIONS(2800),
    [anon_sym___asm] = ACTIONS(2800),
    [sym_number_literal] = ACTIONS(2802),
    [anon_sym_L_SQUOTE] = ACTIONS(2802),
    [anon_sym_u_SQUOTE] = ACTIONS(2802),
    [anon_sym_U_SQUOTE] = ACTIONS(2802),
    [anon_sym_u8_SQUOTE] = ACTIONS(2802),
    [anon_sym_SQUOTE] = ACTIONS(2802),
    [anon_sym_AT] = ACTIONS(2800),
    [anon_sym_DQUOTE] = ACTIONS(2802),
    [anon_sym_L_DQUOTE] = ACTIONS(2802),
    [anon_sym_u_DQUOTE] = ACTIONS(2802),
    [anon_sym_U_DQUOTE] = ACTIONS(2802),
    [anon_sym_u8_DQUOTE] = ACTIONS(2802),
    [sym_true] = ACTIONS(2800),
    [sym_false] = ACTIONS(2800),
    [anon_sym_NULL] = ACTIONS(2800),
    [anon_sym_nullptr] = ACTIONS(2800),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2800),
    [anon_sym___typeof] = ACTIONS(2800),
    [anon_sym_typeof] = ACTIONS(2800),
    [anon_sym_ATimport] = ACTIONS(2802),
    [aux_sym_preproc_undef_token1] = ACTIONS(2800),
    [anon_sym_POUND] = ACTIONS(2800),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2800),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2800),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2800),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2800),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2800),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2800),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2800),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_API_AVAILABLE] = ACTIONS(2800),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_API_DEPRECATED] = ACTIONS(2800),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2800),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym___deprecated_msg] = ACTIONS(2800),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2800),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2800),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2800),
    [anon_sym_ATprotocol] = ACTIONS(2802),
    [anon_sym_ATinterface] = ACTIONS(2802),
    [anon_sym_ATimplementation] = ACTIONS(2802),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2802),
    [anon_sym_ATsynthesize] = ACTIONS(2802),
    [anon_sym_ATdynamic] = ACTIONS(2802),
    [anon_sym_ATtry] = ACTIONS(2802),
    [anon_sym___try] = ACTIONS(2800),
    [anon_sym_ATthrow] = ACTIONS(2802),
    [anon_sym_ATselector] = ACTIONS(2802),
    [anon_sym_ATavailable] = ACTIONS(2802),
    [anon_sym___builtin_available] = ACTIONS(2800),
    [anon_sym_va_arg] = ACTIONS(2800),
    [anon_sym_ATencode] = ACTIONS(2802),
    [anon_sym_ATsynchronized] = ACTIONS(2802),
    [anon_sym_BOOL] = ACTIONS(2800),
    [anon_sym_IMP] = ACTIONS(2800),
    [anon_sym_SEL] = ACTIONS(2800),
    [anon_sym_Class] = ACTIONS(2800),
    [anon_sym_id] = ACTIONS(2800),
  },
  [STATE(340)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token2] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [aux_sym_preproc_else_token1] = ACTIONS(2804),
    [aux_sym_preproc_elif_token1] = ACTIONS(2804),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATsynthesize] = ACTIONS(2806),
    [anon_sym_ATdynamic] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(341)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token2] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [aux_sym_preproc_else_token1] = ACTIONS(2804),
    [aux_sym_preproc_elif_token1] = ACTIONS(2804),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATsynthesize] = ACTIONS(2806),
    [anon_sym_ATdynamic] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(342)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token2] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [aux_sym_preproc_else_token1] = ACTIONS(2804),
    [aux_sym_preproc_elif_token1] = ACTIONS(2804),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATsynthesize] = ACTIONS(2806),
    [anon_sym_ATdynamic] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(343)] = {
    [sym_identifier] = ACTIONS(2808),
    [aux_sym_preproc_include_token1] = ACTIONS(2808),
    [aux_sym_preproc_include_token2] = ACTIONS(2808),
    [aux_sym_preproc_def_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token2] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2808),
    [aux_sym_preproc_else_token1] = ACTIONS(2808),
    [aux_sym_preproc_elif_token1] = ACTIONS(2808),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2808),
    [sym_preproc_directive] = ACTIONS(2808),
    [anon_sym_LPAREN2] = ACTIONS(2810),
    [anon_sym_BANG] = ACTIONS(2810),
    [anon_sym_TILDE] = ACTIONS(2810),
    [anon_sym_DASH] = ACTIONS(2808),
    [anon_sym_PLUS] = ACTIONS(2808),
    [anon_sym_STAR] = ACTIONS(2810),
    [anon_sym_CARET] = ACTIONS(2810),
    [anon_sym_AMP] = ACTIONS(2810),
    [anon_sym_SEMI] = ACTIONS(2810),
    [anon_sym___extension__] = ACTIONS(2808),
    [anon_sym_typedef] = ACTIONS(2808),
    [anon_sym_extern] = ACTIONS(2808),
    [anon_sym___attribute__] = ACTIONS(2808),
    [anon_sym___attribute] = ACTIONS(2808),
    [anon_sym_noreturn] = ACTIONS(2808),
    [anon_sym_LBRACK] = ACTIONS(2810),
    [anon_sym___declspec] = ACTIONS(2808),
    [anon_sym___cdecl] = ACTIONS(2808),
    [anon_sym___clrcall] = ACTIONS(2808),
    [anon_sym___stdcall] = ACTIONS(2808),
    [anon_sym___fastcall] = ACTIONS(2808),
    [anon_sym___thiscall] = ACTIONS(2808),
    [anon_sym___vectorcall] = ACTIONS(2808),
    [anon_sym_LBRACE] = ACTIONS(2810),
    [anon_sym_signed] = ACTIONS(2808),
    [anon_sym_unsigned] = ACTIONS(2808),
    [anon_sym_long] = ACTIONS(2808),
    [anon_sym_short] = ACTIONS(2808),
    [anon_sym_static] = ACTIONS(2808),
    [anon_sym_ATautoreleasepool] = ACTIONS(2810),
    [anon_sym_auto] = ACTIONS(2808),
    [anon_sym_register] = ACTIONS(2808),
    [anon_sym_inline] = ACTIONS(2808),
    [anon_sym___inline] = ACTIONS(2808),
    [anon_sym___inline__] = ACTIONS(2808),
    [anon_sym___forceinline] = ACTIONS(2808),
    [anon_sym_thread_local] = ACTIONS(2808),
    [anon_sym___thread] = ACTIONS(2808),
    [anon_sym_CG_EXTERN] = ACTIONS(2808),
    [anon_sym_CG_INLINE] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2808),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2808),
    [anon_sym_IBOutlet] = ACTIONS(2808),
    [anon_sym_IBInspectable] = ACTIONS(2808),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2808),
    [anon_sym_NS_INLINE] = ACTIONS(2808),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2808),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2808),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2808),
    [anon_sym_const] = ACTIONS(2808),
    [anon_sym_constexpr] = ACTIONS(2808),
    [anon_sym_volatile] = ACTIONS(2808),
    [anon_sym_restrict] = ACTIONS(2808),
    [anon_sym___restrict__] = ACTIONS(2808),
    [anon_sym__Atomic] = ACTIONS(2808),
    [anon_sym__Noreturn] = ACTIONS(2808),
    [anon_sym__Nonnull] = ACTIONS(2808),
    [anon_sym_nullable] = ACTIONS(2808),
    [anon_sym__Complex] = ACTIONS(2808),
    [anon_sym__Nullable] = ACTIONS(2808),
    [anon_sym__Nullable_result] = ACTIONS(2808),
    [anon_sym__Null_unspecified] = ACTIONS(2808),
    [anon_sym___autoreleasing] = ACTIONS(2808),
    [anon_sym___block] = ACTIONS(2808),
    [anon_sym___bridge] = ACTIONS(2808),
    [anon_sym___bridge_retained] = ACTIONS(2808),
    [anon_sym___bridge_transfer] = ACTIONS(2808),
    [anon_sym___complex] = ACTIONS(2808),
    [anon_sym___const] = ACTIONS(2808),
    [anon_sym___imag] = ACTIONS(2808),
    [anon_sym___kindof] = ACTIONS(2808),
    [anon_sym___nonnull] = ACTIONS(2808),
    [anon_sym___nullable] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2808),
    [anon_sym___real] = ACTIONS(2808),
    [anon_sym___strong] = ACTIONS(2808),
    [anon_sym___unsafe_unretained] = ACTIONS(2808),
    [anon_sym___unused] = ACTIONS(2808),
    [anon_sym___weak] = ACTIONS(2808),
    [anon_sym_alignas] = ACTIONS(2808),
    [anon_sym__Alignas] = ACTIONS(2808),
    [sym_primitive_type] = ACTIONS(2808),
    [anon_sym_enum] = ACTIONS(2808),
    [anon_sym_struct] = ACTIONS(2808),
    [anon_sym_union] = ACTIONS(2808),
    [anon_sym_if] = ACTIONS(2808),
    [anon_sym_switch] = ACTIONS(2808),
    [anon_sym_case] = ACTIONS(2808),
    [anon_sym_default] = ACTIONS(2808),
    [anon_sym_while] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2808),
    [anon_sym_in] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_break] = ACTIONS(2808),
    [anon_sym_continue] = ACTIONS(2808),
    [anon_sym_goto] = ACTIONS(2808),
    [anon_sym_DASH_DASH] = ACTIONS(2810),
    [anon_sym_PLUS_PLUS] = ACTIONS(2810),
    [anon_sym_sizeof] = ACTIONS(2808),
    [anon_sym___alignof__] = ACTIONS(2808),
    [anon_sym___alignof] = ACTIONS(2808),
    [anon_sym__alignof] = ACTIONS(2808),
    [anon_sym_alignof] = ACTIONS(2808),
    [anon_sym__Alignof] = ACTIONS(2808),
    [anon_sym_offsetof] = ACTIONS(2808),
    [anon_sym__Generic] = ACTIONS(2808),
    [anon_sym_asm] = ACTIONS(2808),
    [anon_sym___asm__] = ACTIONS(2808),
    [anon_sym___asm] = ACTIONS(2808),
    [sym_number_literal] = ACTIONS(2810),
    [anon_sym_L_SQUOTE] = ACTIONS(2810),
    [anon_sym_u_SQUOTE] = ACTIONS(2810),
    [anon_sym_U_SQUOTE] = ACTIONS(2810),
    [anon_sym_u8_SQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2810),
    [anon_sym_AT] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_L_DQUOTE] = ACTIONS(2810),
    [anon_sym_u_DQUOTE] = ACTIONS(2810),
    [anon_sym_U_DQUOTE] = ACTIONS(2810),
    [anon_sym_u8_DQUOTE] = ACTIONS(2810),
    [sym_true] = ACTIONS(2808),
    [sym_false] = ACTIONS(2808),
    [anon_sym_NULL] = ACTIONS(2808),
    [anon_sym_nullptr] = ACTIONS(2808),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2808),
    [anon_sym___typeof] = ACTIONS(2808),
    [anon_sym_typeof] = ACTIONS(2808),
    [anon_sym_ATimport] = ACTIONS(2810),
    [aux_sym_preproc_undef_token1] = ACTIONS(2808),
    [anon_sym_POUND] = ACTIONS(2808),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2808),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2808),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2808),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_API_AVAILABLE] = ACTIONS(2808),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_API_DEPRECATED] = ACTIONS(2808),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym___deprecated_msg] = ACTIONS(2808),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2808),
    [anon_sym_ATprotocol] = ACTIONS(2810),
    [anon_sym_ATinterface] = ACTIONS(2810),
    [anon_sym_ATimplementation] = ACTIONS(2810),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2810),
    [anon_sym_ATsynthesize] = ACTIONS(2810),
    [anon_sym_ATdynamic] = ACTIONS(2810),
    [anon_sym_ATtry] = ACTIONS(2810),
    [anon_sym___try] = ACTIONS(2808),
    [anon_sym_ATthrow] = ACTIONS(2810),
    [anon_sym_ATselector] = ACTIONS(2810),
    [anon_sym_ATavailable] = ACTIONS(2810),
    [anon_sym___builtin_available] = ACTIONS(2808),
    [anon_sym_va_arg] = ACTIONS(2808),
    [anon_sym_ATencode] = ACTIONS(2810),
    [anon_sym_ATsynchronized] = ACTIONS(2810),
    [anon_sym_BOOL] = ACTIONS(2808),
    [anon_sym_IMP] = ACTIONS(2808),
    [anon_sym_SEL] = ACTIONS(2808),
    [anon_sym_Class] = ACTIONS(2808),
    [anon_sym_id] = ACTIONS(2808),
  },
  [STATE(344)] = {
    [sym_identifier] = ACTIONS(2808),
    [aux_sym_preproc_include_token1] = ACTIONS(2808),
    [aux_sym_preproc_include_token2] = ACTIONS(2808),
    [aux_sym_preproc_def_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token2] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2808),
    [aux_sym_preproc_else_token1] = ACTIONS(2808),
    [aux_sym_preproc_elif_token1] = ACTIONS(2808),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2808),
    [sym_preproc_directive] = ACTIONS(2808),
    [anon_sym_LPAREN2] = ACTIONS(2810),
    [anon_sym_BANG] = ACTIONS(2810),
    [anon_sym_TILDE] = ACTIONS(2810),
    [anon_sym_DASH] = ACTIONS(2808),
    [anon_sym_PLUS] = ACTIONS(2808),
    [anon_sym_STAR] = ACTIONS(2810),
    [anon_sym_CARET] = ACTIONS(2810),
    [anon_sym_AMP] = ACTIONS(2810),
    [anon_sym_SEMI] = ACTIONS(2810),
    [anon_sym___extension__] = ACTIONS(2808),
    [anon_sym_typedef] = ACTIONS(2808),
    [anon_sym_extern] = ACTIONS(2808),
    [anon_sym___attribute__] = ACTIONS(2808),
    [anon_sym___attribute] = ACTIONS(2808),
    [anon_sym_noreturn] = ACTIONS(2808),
    [anon_sym_LBRACK] = ACTIONS(2810),
    [anon_sym___declspec] = ACTIONS(2808),
    [anon_sym___cdecl] = ACTIONS(2808),
    [anon_sym___clrcall] = ACTIONS(2808),
    [anon_sym___stdcall] = ACTIONS(2808),
    [anon_sym___fastcall] = ACTIONS(2808),
    [anon_sym___thiscall] = ACTIONS(2808),
    [anon_sym___vectorcall] = ACTIONS(2808),
    [anon_sym_LBRACE] = ACTIONS(2810),
    [anon_sym_signed] = ACTIONS(2808),
    [anon_sym_unsigned] = ACTIONS(2808),
    [anon_sym_long] = ACTIONS(2808),
    [anon_sym_short] = ACTIONS(2808),
    [anon_sym_static] = ACTIONS(2808),
    [anon_sym_ATautoreleasepool] = ACTIONS(2810),
    [anon_sym_auto] = ACTIONS(2808),
    [anon_sym_register] = ACTIONS(2808),
    [anon_sym_inline] = ACTIONS(2808),
    [anon_sym___inline] = ACTIONS(2808),
    [anon_sym___inline__] = ACTIONS(2808),
    [anon_sym___forceinline] = ACTIONS(2808),
    [anon_sym_thread_local] = ACTIONS(2808),
    [anon_sym___thread] = ACTIONS(2808),
    [anon_sym_CG_EXTERN] = ACTIONS(2808),
    [anon_sym_CG_INLINE] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2808),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2808),
    [anon_sym_IBOutlet] = ACTIONS(2808),
    [anon_sym_IBInspectable] = ACTIONS(2808),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2808),
    [anon_sym_NS_INLINE] = ACTIONS(2808),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2808),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2808),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2808),
    [anon_sym_const] = ACTIONS(2808),
    [anon_sym_constexpr] = ACTIONS(2808),
    [anon_sym_volatile] = ACTIONS(2808),
    [anon_sym_restrict] = ACTIONS(2808),
    [anon_sym___restrict__] = ACTIONS(2808),
    [anon_sym__Atomic] = ACTIONS(2808),
    [anon_sym__Noreturn] = ACTIONS(2808),
    [anon_sym__Nonnull] = ACTIONS(2808),
    [anon_sym_nullable] = ACTIONS(2808),
    [anon_sym__Complex] = ACTIONS(2808),
    [anon_sym__Nullable] = ACTIONS(2808),
    [anon_sym__Nullable_result] = ACTIONS(2808),
    [anon_sym__Null_unspecified] = ACTIONS(2808),
    [anon_sym___autoreleasing] = ACTIONS(2808),
    [anon_sym___block] = ACTIONS(2808),
    [anon_sym___bridge] = ACTIONS(2808),
    [anon_sym___bridge_retained] = ACTIONS(2808),
    [anon_sym___bridge_transfer] = ACTIONS(2808),
    [anon_sym___complex] = ACTIONS(2808),
    [anon_sym___const] = ACTIONS(2808),
    [anon_sym___imag] = ACTIONS(2808),
    [anon_sym___kindof] = ACTIONS(2808),
    [anon_sym___nonnull] = ACTIONS(2808),
    [anon_sym___nullable] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2808),
    [anon_sym___real] = ACTIONS(2808),
    [anon_sym___strong] = ACTIONS(2808),
    [anon_sym___unsafe_unretained] = ACTIONS(2808),
    [anon_sym___unused] = ACTIONS(2808),
    [anon_sym___weak] = ACTIONS(2808),
    [anon_sym_alignas] = ACTIONS(2808),
    [anon_sym__Alignas] = ACTIONS(2808),
    [sym_primitive_type] = ACTIONS(2808),
    [anon_sym_enum] = ACTIONS(2808),
    [anon_sym_struct] = ACTIONS(2808),
    [anon_sym_union] = ACTIONS(2808),
    [anon_sym_if] = ACTIONS(2808),
    [anon_sym_switch] = ACTIONS(2808),
    [anon_sym_case] = ACTIONS(2808),
    [anon_sym_default] = ACTIONS(2808),
    [anon_sym_while] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2808),
    [anon_sym_in] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_break] = ACTIONS(2808),
    [anon_sym_continue] = ACTIONS(2808),
    [anon_sym_goto] = ACTIONS(2808),
    [anon_sym_DASH_DASH] = ACTIONS(2810),
    [anon_sym_PLUS_PLUS] = ACTIONS(2810),
    [anon_sym_sizeof] = ACTIONS(2808),
    [anon_sym___alignof__] = ACTIONS(2808),
    [anon_sym___alignof] = ACTIONS(2808),
    [anon_sym__alignof] = ACTIONS(2808),
    [anon_sym_alignof] = ACTIONS(2808),
    [anon_sym__Alignof] = ACTIONS(2808),
    [anon_sym_offsetof] = ACTIONS(2808),
    [anon_sym__Generic] = ACTIONS(2808),
    [anon_sym_asm] = ACTIONS(2808),
    [anon_sym___asm__] = ACTIONS(2808),
    [anon_sym___asm] = ACTIONS(2808),
    [sym_number_literal] = ACTIONS(2810),
    [anon_sym_L_SQUOTE] = ACTIONS(2810),
    [anon_sym_u_SQUOTE] = ACTIONS(2810),
    [anon_sym_U_SQUOTE] = ACTIONS(2810),
    [anon_sym_u8_SQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2810),
    [anon_sym_AT] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_L_DQUOTE] = ACTIONS(2810),
    [anon_sym_u_DQUOTE] = ACTIONS(2810),
    [anon_sym_U_DQUOTE] = ACTIONS(2810),
    [anon_sym_u8_DQUOTE] = ACTIONS(2810),
    [sym_true] = ACTIONS(2808),
    [sym_false] = ACTIONS(2808),
    [anon_sym_NULL] = ACTIONS(2808),
    [anon_sym_nullptr] = ACTIONS(2808),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2808),
    [anon_sym___typeof] = ACTIONS(2808),
    [anon_sym_typeof] = ACTIONS(2808),
    [anon_sym_ATimport] = ACTIONS(2810),
    [aux_sym_preproc_undef_token1] = ACTIONS(2808),
    [anon_sym_POUND] = ACTIONS(2808),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2808),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2808),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2808),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_API_AVAILABLE] = ACTIONS(2808),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_API_DEPRECATED] = ACTIONS(2808),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym___deprecated_msg] = ACTIONS(2808),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2808),
    [anon_sym_ATprotocol] = ACTIONS(2810),
    [anon_sym_ATinterface] = ACTIONS(2810),
    [anon_sym_ATimplementation] = ACTIONS(2810),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2810),
    [anon_sym_ATsynthesize] = ACTIONS(2810),
    [anon_sym_ATdynamic] = ACTIONS(2810),
    [anon_sym_ATtry] = ACTIONS(2810),
    [anon_sym___try] = ACTIONS(2808),
    [anon_sym_ATthrow] = ACTIONS(2810),
    [anon_sym_ATselector] = ACTIONS(2810),
    [anon_sym_ATavailable] = ACTIONS(2810),
    [anon_sym___builtin_available] = ACTIONS(2808),
    [anon_sym_va_arg] = ACTIONS(2808),
    [anon_sym_ATencode] = ACTIONS(2810),
    [anon_sym_ATsynchronized] = ACTIONS(2810),
    [anon_sym_BOOL] = ACTIONS(2808),
    [anon_sym_IMP] = ACTIONS(2808),
    [anon_sym_SEL] = ACTIONS(2808),
    [anon_sym_Class] = ACTIONS(2808),
    [anon_sym_id] = ACTIONS(2808),
  },
  [STATE(345)] = {
    [sym_identifier] = ACTIONS(2812),
    [aux_sym_preproc_include_token1] = ACTIONS(2812),
    [aux_sym_preproc_include_token2] = ACTIONS(2812),
    [aux_sym_preproc_def_token1] = ACTIONS(2812),
    [aux_sym_preproc_if_token1] = ACTIONS(2812),
    [aux_sym_preproc_if_token2] = ACTIONS(2812),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2812),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2812),
    [aux_sym_preproc_else_token1] = ACTIONS(2812),
    [aux_sym_preproc_elif_token1] = ACTIONS(2812),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2812),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2812),
    [sym_preproc_directive] = ACTIONS(2812),
    [anon_sym_LPAREN2] = ACTIONS(2814),
    [anon_sym_BANG] = ACTIONS(2814),
    [anon_sym_TILDE] = ACTIONS(2814),
    [anon_sym_DASH] = ACTIONS(2812),
    [anon_sym_PLUS] = ACTIONS(2812),
    [anon_sym_STAR] = ACTIONS(2814),
    [anon_sym_CARET] = ACTIONS(2814),
    [anon_sym_AMP] = ACTIONS(2814),
    [anon_sym_SEMI] = ACTIONS(2814),
    [anon_sym___extension__] = ACTIONS(2812),
    [anon_sym_typedef] = ACTIONS(2812),
    [anon_sym_extern] = ACTIONS(2812),
    [anon_sym___attribute__] = ACTIONS(2812),
    [anon_sym___attribute] = ACTIONS(2812),
    [anon_sym_noreturn] = ACTIONS(2812),
    [anon_sym_LBRACK] = ACTIONS(2814),
    [anon_sym___declspec] = ACTIONS(2812),
    [anon_sym___cdecl] = ACTIONS(2812),
    [anon_sym___clrcall] = ACTIONS(2812),
    [anon_sym___stdcall] = ACTIONS(2812),
    [anon_sym___fastcall] = ACTIONS(2812),
    [anon_sym___thiscall] = ACTIONS(2812),
    [anon_sym___vectorcall] = ACTIONS(2812),
    [anon_sym_LBRACE] = ACTIONS(2814),
    [anon_sym_signed] = ACTIONS(2812),
    [anon_sym_unsigned] = ACTIONS(2812),
    [anon_sym_long] = ACTIONS(2812),
    [anon_sym_short] = ACTIONS(2812),
    [anon_sym_static] = ACTIONS(2812),
    [anon_sym_ATautoreleasepool] = ACTIONS(2814),
    [anon_sym_auto] = ACTIONS(2812),
    [anon_sym_register] = ACTIONS(2812),
    [anon_sym_inline] = ACTIONS(2812),
    [anon_sym___inline] = ACTIONS(2812),
    [anon_sym___inline__] = ACTIONS(2812),
    [anon_sym___forceinline] = ACTIONS(2812),
    [anon_sym_thread_local] = ACTIONS(2812),
    [anon_sym___thread] = ACTIONS(2812),
    [anon_sym_CG_EXTERN] = ACTIONS(2812),
    [anon_sym_CG_INLINE] = ACTIONS(2812),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2812),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2812),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2812),
    [anon_sym_IBOutlet] = ACTIONS(2812),
    [anon_sym_IBInspectable] = ACTIONS(2812),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2812),
    [anon_sym_NS_INLINE] = ACTIONS(2812),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2812),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2812),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2812),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2812),
    [anon_sym_const] = ACTIONS(2812),
    [anon_sym_constexpr] = ACTIONS(2812),
    [anon_sym_volatile] = ACTIONS(2812),
    [anon_sym_restrict] = ACTIONS(2812),
    [anon_sym___restrict__] = ACTIONS(2812),
    [anon_sym__Atomic] = ACTIONS(2812),
    [anon_sym__Noreturn] = ACTIONS(2812),
    [anon_sym__Nonnull] = ACTIONS(2812),
    [anon_sym_nullable] = ACTIONS(2812),
    [anon_sym__Complex] = ACTIONS(2812),
    [anon_sym__Nullable] = ACTIONS(2812),
    [anon_sym__Nullable_result] = ACTIONS(2812),
    [anon_sym__Null_unspecified] = ACTIONS(2812),
    [anon_sym___autoreleasing] = ACTIONS(2812),
    [anon_sym___block] = ACTIONS(2812),
    [anon_sym___bridge] = ACTIONS(2812),
    [anon_sym___bridge_retained] = ACTIONS(2812),
    [anon_sym___bridge_transfer] = ACTIONS(2812),
    [anon_sym___complex] = ACTIONS(2812),
    [anon_sym___const] = ACTIONS(2812),
    [anon_sym___imag] = ACTIONS(2812),
    [anon_sym___kindof] = ACTIONS(2812),
    [anon_sym___nonnull] = ACTIONS(2812),
    [anon_sym___nullable] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2812),
    [anon_sym___real] = ACTIONS(2812),
    [anon_sym___strong] = ACTIONS(2812),
    [anon_sym___unsafe_unretained] = ACTIONS(2812),
    [anon_sym___unused] = ACTIONS(2812),
    [anon_sym___weak] = ACTIONS(2812),
    [anon_sym_alignas] = ACTIONS(2812),
    [anon_sym__Alignas] = ACTIONS(2812),
    [sym_primitive_type] = ACTIONS(2812),
    [anon_sym_enum] = ACTIONS(2812),
    [anon_sym_struct] = ACTIONS(2812),
    [anon_sym_union] = ACTIONS(2812),
    [anon_sym_if] = ACTIONS(2812),
    [anon_sym_switch] = ACTIONS(2812),
    [anon_sym_case] = ACTIONS(2812),
    [anon_sym_default] = ACTIONS(2812),
    [anon_sym_while] = ACTIONS(2812),
    [anon_sym_do] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2812),
    [anon_sym_in] = ACTIONS(2812),
    [anon_sym_return] = ACTIONS(2812),
    [anon_sym_break] = ACTIONS(2812),
    [anon_sym_continue] = ACTIONS(2812),
    [anon_sym_goto] = ACTIONS(2812),
    [anon_sym_DASH_DASH] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_sizeof] = ACTIONS(2812),
    [anon_sym___alignof__] = ACTIONS(2812),
    [anon_sym___alignof] = ACTIONS(2812),
    [anon_sym__alignof] = ACTIONS(2812),
    [anon_sym_alignof] = ACTIONS(2812),
    [anon_sym__Alignof] = ACTIONS(2812),
    [anon_sym_offsetof] = ACTIONS(2812),
    [anon_sym__Generic] = ACTIONS(2812),
    [anon_sym_asm] = ACTIONS(2812),
    [anon_sym___asm__] = ACTIONS(2812),
    [anon_sym___asm] = ACTIONS(2812),
    [sym_number_literal] = ACTIONS(2814),
    [anon_sym_L_SQUOTE] = ACTIONS(2814),
    [anon_sym_u_SQUOTE] = ACTIONS(2814),
    [anon_sym_U_SQUOTE] = ACTIONS(2814),
    [anon_sym_u8_SQUOTE] = ACTIONS(2814),
    [anon_sym_SQUOTE] = ACTIONS(2814),
    [anon_sym_AT] = ACTIONS(2812),
    [anon_sym_DQUOTE] = ACTIONS(2814),
    [anon_sym_L_DQUOTE] = ACTIONS(2814),
    [anon_sym_u_DQUOTE] = ACTIONS(2814),
    [anon_sym_U_DQUOTE] = ACTIONS(2814),
    [anon_sym_u8_DQUOTE] = ACTIONS(2814),
    [sym_true] = ACTIONS(2812),
    [sym_false] = ACTIONS(2812),
    [anon_sym_NULL] = ACTIONS(2812),
    [anon_sym_nullptr] = ACTIONS(2812),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2812),
    [anon_sym___typeof] = ACTIONS(2812),
    [anon_sym_typeof] = ACTIONS(2812),
    [anon_sym_ATimport] = ACTIONS(2814),
    [aux_sym_preproc_undef_token1] = ACTIONS(2812),
    [anon_sym_POUND] = ACTIONS(2812),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2812),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2812),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2812),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2812),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2812),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2812),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2812),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_API_AVAILABLE] = ACTIONS(2812),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_API_DEPRECATED] = ACTIONS(2812),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2812),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym___deprecated_msg] = ACTIONS(2812),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2812),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2812),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2812),
    [anon_sym_ATprotocol] = ACTIONS(2814),
    [anon_sym_ATinterface] = ACTIONS(2814),
    [anon_sym_ATimplementation] = ACTIONS(2814),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2814),
    [anon_sym_ATsynthesize] = ACTIONS(2814),
    [anon_sym_ATdynamic] = ACTIONS(2814),
    [anon_sym_ATtry] = ACTIONS(2814),
    [anon_sym___try] = ACTIONS(2812),
    [anon_sym_ATthrow] = ACTIONS(2814),
    [anon_sym_ATselector] = ACTIONS(2814),
    [anon_sym_ATavailable] = ACTIONS(2814),
    [anon_sym___builtin_available] = ACTIONS(2812),
    [anon_sym_va_arg] = ACTIONS(2812),
    [anon_sym_ATencode] = ACTIONS(2814),
    [anon_sym_ATsynchronized] = ACTIONS(2814),
    [anon_sym_BOOL] = ACTIONS(2812),
    [anon_sym_IMP] = ACTIONS(2812),
    [anon_sym_SEL] = ACTIONS(2812),
    [anon_sym_Class] = ACTIONS(2812),
    [anon_sym_id] = ACTIONS(2812),
  },
  [STATE(346)] = {
    [sym_identifier] = ACTIONS(2816),
    [aux_sym_preproc_include_token1] = ACTIONS(2816),
    [aux_sym_preproc_include_token2] = ACTIONS(2816),
    [aux_sym_preproc_def_token1] = ACTIONS(2816),
    [aux_sym_preproc_if_token1] = ACTIONS(2816),
    [aux_sym_preproc_if_token2] = ACTIONS(2816),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2816),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2816),
    [aux_sym_preproc_else_token1] = ACTIONS(2816),
    [aux_sym_preproc_elif_token1] = ACTIONS(2816),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2816),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2816),
    [sym_preproc_directive] = ACTIONS(2816),
    [anon_sym_LPAREN2] = ACTIONS(2818),
    [anon_sym_BANG] = ACTIONS(2818),
    [anon_sym_TILDE] = ACTIONS(2818),
    [anon_sym_DASH] = ACTIONS(2816),
    [anon_sym_PLUS] = ACTIONS(2816),
    [anon_sym_STAR] = ACTIONS(2818),
    [anon_sym_CARET] = ACTIONS(2818),
    [anon_sym_AMP] = ACTIONS(2818),
    [anon_sym_SEMI] = ACTIONS(2818),
    [anon_sym___extension__] = ACTIONS(2816),
    [anon_sym_typedef] = ACTIONS(2816),
    [anon_sym_extern] = ACTIONS(2816),
    [anon_sym___attribute__] = ACTIONS(2816),
    [anon_sym___attribute] = ACTIONS(2816),
    [anon_sym_noreturn] = ACTIONS(2816),
    [anon_sym_LBRACK] = ACTIONS(2818),
    [anon_sym___declspec] = ACTIONS(2816),
    [anon_sym___cdecl] = ACTIONS(2816),
    [anon_sym___clrcall] = ACTIONS(2816),
    [anon_sym___stdcall] = ACTIONS(2816),
    [anon_sym___fastcall] = ACTIONS(2816),
    [anon_sym___thiscall] = ACTIONS(2816),
    [anon_sym___vectorcall] = ACTIONS(2816),
    [anon_sym_LBRACE] = ACTIONS(2818),
    [anon_sym_signed] = ACTIONS(2816),
    [anon_sym_unsigned] = ACTIONS(2816),
    [anon_sym_long] = ACTIONS(2816),
    [anon_sym_short] = ACTIONS(2816),
    [anon_sym_static] = ACTIONS(2816),
    [anon_sym_ATautoreleasepool] = ACTIONS(2818),
    [anon_sym_auto] = ACTIONS(2816),
    [anon_sym_register] = ACTIONS(2816),
    [anon_sym_inline] = ACTIONS(2816),
    [anon_sym___inline] = ACTIONS(2816),
    [anon_sym___inline__] = ACTIONS(2816),
    [anon_sym___forceinline] = ACTIONS(2816),
    [anon_sym_thread_local] = ACTIONS(2816),
    [anon_sym___thread] = ACTIONS(2816),
    [anon_sym_CG_EXTERN] = ACTIONS(2816),
    [anon_sym_CG_INLINE] = ACTIONS(2816),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2816),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2816),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2816),
    [anon_sym_IBOutlet] = ACTIONS(2816),
    [anon_sym_IBInspectable] = ACTIONS(2816),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2816),
    [anon_sym_NS_INLINE] = ACTIONS(2816),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2816),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2816),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2816),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2816),
    [anon_sym_const] = ACTIONS(2816),
    [anon_sym_constexpr] = ACTIONS(2816),
    [anon_sym_volatile] = ACTIONS(2816),
    [anon_sym_restrict] = ACTIONS(2816),
    [anon_sym___restrict__] = ACTIONS(2816),
    [anon_sym__Atomic] = ACTIONS(2816),
    [anon_sym__Noreturn] = ACTIONS(2816),
    [anon_sym__Nonnull] = ACTIONS(2816),
    [anon_sym_nullable] = ACTIONS(2816),
    [anon_sym__Complex] = ACTIONS(2816),
    [anon_sym__Nullable] = ACTIONS(2816),
    [anon_sym__Nullable_result] = ACTIONS(2816),
    [anon_sym__Null_unspecified] = ACTIONS(2816),
    [anon_sym___autoreleasing] = ACTIONS(2816),
    [anon_sym___block] = ACTIONS(2816),
    [anon_sym___bridge] = ACTIONS(2816),
    [anon_sym___bridge_retained] = ACTIONS(2816),
    [anon_sym___bridge_transfer] = ACTIONS(2816),
    [anon_sym___complex] = ACTIONS(2816),
    [anon_sym___const] = ACTIONS(2816),
    [anon_sym___imag] = ACTIONS(2816),
    [anon_sym___kindof] = ACTIONS(2816),
    [anon_sym___nonnull] = ACTIONS(2816),
    [anon_sym___nullable] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2816),
    [anon_sym___real] = ACTIONS(2816),
    [anon_sym___strong] = ACTIONS(2816),
    [anon_sym___unsafe_unretained] = ACTIONS(2816),
    [anon_sym___unused] = ACTIONS(2816),
    [anon_sym___weak] = ACTIONS(2816),
    [anon_sym_alignas] = ACTIONS(2816),
    [anon_sym__Alignas] = ACTIONS(2816),
    [sym_primitive_type] = ACTIONS(2816),
    [anon_sym_enum] = ACTIONS(2816),
    [anon_sym_struct] = ACTIONS(2816),
    [anon_sym_union] = ACTIONS(2816),
    [anon_sym_if] = ACTIONS(2816),
    [anon_sym_switch] = ACTIONS(2816),
    [anon_sym_case] = ACTIONS(2816),
    [anon_sym_default] = ACTIONS(2816),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_do] = ACTIONS(2816),
    [anon_sym_for] = ACTIONS(2816),
    [anon_sym_in] = ACTIONS(2816),
    [anon_sym_return] = ACTIONS(2816),
    [anon_sym_break] = ACTIONS(2816),
    [anon_sym_continue] = ACTIONS(2816),
    [anon_sym_goto] = ACTIONS(2816),
    [anon_sym_DASH_DASH] = ACTIONS(2818),
    [anon_sym_PLUS_PLUS] = ACTIONS(2818),
    [anon_sym_sizeof] = ACTIONS(2816),
    [anon_sym___alignof__] = ACTIONS(2816),
    [anon_sym___alignof] = ACTIONS(2816),
    [anon_sym__alignof] = ACTIONS(2816),
    [anon_sym_alignof] = ACTIONS(2816),
    [anon_sym__Alignof] = ACTIONS(2816),
    [anon_sym_offsetof] = ACTIONS(2816),
    [anon_sym__Generic] = ACTIONS(2816),
    [anon_sym_asm] = ACTIONS(2816),
    [anon_sym___asm__] = ACTIONS(2816),
    [anon_sym___asm] = ACTIONS(2816),
    [sym_number_literal] = ACTIONS(2818),
    [anon_sym_L_SQUOTE] = ACTIONS(2818),
    [anon_sym_u_SQUOTE] = ACTIONS(2818),
    [anon_sym_U_SQUOTE] = ACTIONS(2818),
    [anon_sym_u8_SQUOTE] = ACTIONS(2818),
    [anon_sym_SQUOTE] = ACTIONS(2818),
    [anon_sym_AT] = ACTIONS(2816),
    [anon_sym_DQUOTE] = ACTIONS(2818),
    [anon_sym_L_DQUOTE] = ACTIONS(2818),
    [anon_sym_u_DQUOTE] = ACTIONS(2818),
    [anon_sym_U_DQUOTE] = ACTIONS(2818),
    [anon_sym_u8_DQUOTE] = ACTIONS(2818),
    [sym_true] = ACTIONS(2816),
    [sym_false] = ACTIONS(2816),
    [anon_sym_NULL] = ACTIONS(2816),
    [anon_sym_nullptr] = ACTIONS(2816),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2816),
    [anon_sym___typeof] = ACTIONS(2816),
    [anon_sym_typeof] = ACTIONS(2816),
    [anon_sym_ATimport] = ACTIONS(2818),
    [aux_sym_preproc_undef_token1] = ACTIONS(2816),
    [anon_sym_POUND] = ACTIONS(2816),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2816),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2816),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2816),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2816),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2816),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2816),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2816),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_API_AVAILABLE] = ACTIONS(2816),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_API_DEPRECATED] = ACTIONS(2816),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2816),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym___deprecated_msg] = ACTIONS(2816),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2816),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2816),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2816),
    [anon_sym_ATprotocol] = ACTIONS(2818),
    [anon_sym_ATinterface] = ACTIONS(2818),
    [anon_sym_ATimplementation] = ACTIONS(2818),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2818),
    [anon_sym_ATsynthesize] = ACTIONS(2818),
    [anon_sym_ATdynamic] = ACTIONS(2818),
    [anon_sym_ATtry] = ACTIONS(2818),
    [anon_sym___try] = ACTIONS(2816),
    [anon_sym_ATthrow] = ACTIONS(2818),
    [anon_sym_ATselector] = ACTIONS(2818),
    [anon_sym_ATavailable] = ACTIONS(2818),
    [anon_sym___builtin_available] = ACTIONS(2816),
    [anon_sym_va_arg] = ACTIONS(2816),
    [anon_sym_ATencode] = ACTIONS(2818),
    [anon_sym_ATsynchronized] = ACTIONS(2818),
    [anon_sym_BOOL] = ACTIONS(2816),
    [anon_sym_IMP] = ACTIONS(2816),
    [anon_sym_SEL] = ACTIONS(2816),
    [anon_sym_Class] = ACTIONS(2816),
    [anon_sym_id] = ACTIONS(2816),
  },
  [STATE(347)] = {
    [sym_identifier] = ACTIONS(2820),
    [aux_sym_preproc_include_token1] = ACTIONS(2820),
    [aux_sym_preproc_include_token2] = ACTIONS(2820),
    [aux_sym_preproc_def_token1] = ACTIONS(2820),
    [aux_sym_preproc_if_token1] = ACTIONS(2820),
    [aux_sym_preproc_if_token2] = ACTIONS(2820),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2820),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2820),
    [aux_sym_preproc_else_token1] = ACTIONS(2820),
    [aux_sym_preproc_elif_token1] = ACTIONS(2820),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2820),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2820),
    [sym_preproc_directive] = ACTIONS(2820),
    [anon_sym_LPAREN2] = ACTIONS(2822),
    [anon_sym_BANG] = ACTIONS(2822),
    [anon_sym_TILDE] = ACTIONS(2822),
    [anon_sym_DASH] = ACTIONS(2820),
    [anon_sym_PLUS] = ACTIONS(2820),
    [anon_sym_STAR] = ACTIONS(2822),
    [anon_sym_CARET] = ACTIONS(2822),
    [anon_sym_AMP] = ACTIONS(2822),
    [anon_sym_SEMI] = ACTIONS(2822),
    [anon_sym___extension__] = ACTIONS(2820),
    [anon_sym_typedef] = ACTIONS(2820),
    [anon_sym_extern] = ACTIONS(2820),
    [anon_sym___attribute__] = ACTIONS(2820),
    [anon_sym___attribute] = ACTIONS(2820),
    [anon_sym_noreturn] = ACTIONS(2820),
    [anon_sym_LBRACK] = ACTIONS(2822),
    [anon_sym___declspec] = ACTIONS(2820),
    [anon_sym___cdecl] = ACTIONS(2820),
    [anon_sym___clrcall] = ACTIONS(2820),
    [anon_sym___stdcall] = ACTIONS(2820),
    [anon_sym___fastcall] = ACTIONS(2820),
    [anon_sym___thiscall] = ACTIONS(2820),
    [anon_sym___vectorcall] = ACTIONS(2820),
    [anon_sym_LBRACE] = ACTIONS(2822),
    [anon_sym_signed] = ACTIONS(2820),
    [anon_sym_unsigned] = ACTIONS(2820),
    [anon_sym_long] = ACTIONS(2820),
    [anon_sym_short] = ACTIONS(2820),
    [anon_sym_static] = ACTIONS(2820),
    [anon_sym_ATautoreleasepool] = ACTIONS(2822),
    [anon_sym_auto] = ACTIONS(2820),
    [anon_sym_register] = ACTIONS(2820),
    [anon_sym_inline] = ACTIONS(2820),
    [anon_sym___inline] = ACTIONS(2820),
    [anon_sym___inline__] = ACTIONS(2820),
    [anon_sym___forceinline] = ACTIONS(2820),
    [anon_sym_thread_local] = ACTIONS(2820),
    [anon_sym___thread] = ACTIONS(2820),
    [anon_sym_CG_EXTERN] = ACTIONS(2820),
    [anon_sym_CG_INLINE] = ACTIONS(2820),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2820),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2820),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2820),
    [anon_sym_IBOutlet] = ACTIONS(2820),
    [anon_sym_IBInspectable] = ACTIONS(2820),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2820),
    [anon_sym_NS_INLINE] = ACTIONS(2820),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2820),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2820),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2820),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2820),
    [anon_sym_const] = ACTIONS(2820),
    [anon_sym_constexpr] = ACTIONS(2820),
    [anon_sym_volatile] = ACTIONS(2820),
    [anon_sym_restrict] = ACTIONS(2820),
    [anon_sym___restrict__] = ACTIONS(2820),
    [anon_sym__Atomic] = ACTIONS(2820),
    [anon_sym__Noreturn] = ACTIONS(2820),
    [anon_sym__Nonnull] = ACTIONS(2820),
    [anon_sym_nullable] = ACTIONS(2820),
    [anon_sym__Complex] = ACTIONS(2820),
    [anon_sym__Nullable] = ACTIONS(2820),
    [anon_sym__Nullable_result] = ACTIONS(2820),
    [anon_sym__Null_unspecified] = ACTIONS(2820),
    [anon_sym___autoreleasing] = ACTIONS(2820),
    [anon_sym___block] = ACTIONS(2820),
    [anon_sym___bridge] = ACTIONS(2820),
    [anon_sym___bridge_retained] = ACTIONS(2820),
    [anon_sym___bridge_transfer] = ACTIONS(2820),
    [anon_sym___complex] = ACTIONS(2820),
    [anon_sym___const] = ACTIONS(2820),
    [anon_sym___imag] = ACTIONS(2820),
    [anon_sym___kindof] = ACTIONS(2820),
    [anon_sym___nonnull] = ACTIONS(2820),
    [anon_sym___nullable] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2820),
    [anon_sym___real] = ACTIONS(2820),
    [anon_sym___strong] = ACTIONS(2820),
    [anon_sym___unsafe_unretained] = ACTIONS(2820),
    [anon_sym___unused] = ACTIONS(2820),
    [anon_sym___weak] = ACTIONS(2820),
    [anon_sym_alignas] = ACTIONS(2820),
    [anon_sym__Alignas] = ACTIONS(2820),
    [sym_primitive_type] = ACTIONS(2820),
    [anon_sym_enum] = ACTIONS(2820),
    [anon_sym_struct] = ACTIONS(2820),
    [anon_sym_union] = ACTIONS(2820),
    [anon_sym_if] = ACTIONS(2820),
    [anon_sym_switch] = ACTIONS(2820),
    [anon_sym_case] = ACTIONS(2820),
    [anon_sym_default] = ACTIONS(2820),
    [anon_sym_while] = ACTIONS(2820),
    [anon_sym_do] = ACTIONS(2820),
    [anon_sym_for] = ACTIONS(2820),
    [anon_sym_in] = ACTIONS(2820),
    [anon_sym_return] = ACTIONS(2820),
    [anon_sym_break] = ACTIONS(2820),
    [anon_sym_continue] = ACTIONS(2820),
    [anon_sym_goto] = ACTIONS(2820),
    [anon_sym_DASH_DASH] = ACTIONS(2822),
    [anon_sym_PLUS_PLUS] = ACTIONS(2822),
    [anon_sym_sizeof] = ACTIONS(2820),
    [anon_sym___alignof__] = ACTIONS(2820),
    [anon_sym___alignof] = ACTIONS(2820),
    [anon_sym__alignof] = ACTIONS(2820),
    [anon_sym_alignof] = ACTIONS(2820),
    [anon_sym__Alignof] = ACTIONS(2820),
    [anon_sym_offsetof] = ACTIONS(2820),
    [anon_sym__Generic] = ACTIONS(2820),
    [anon_sym_asm] = ACTIONS(2820),
    [anon_sym___asm__] = ACTIONS(2820),
    [anon_sym___asm] = ACTIONS(2820),
    [sym_number_literal] = ACTIONS(2822),
    [anon_sym_L_SQUOTE] = ACTIONS(2822),
    [anon_sym_u_SQUOTE] = ACTIONS(2822),
    [anon_sym_U_SQUOTE] = ACTIONS(2822),
    [anon_sym_u8_SQUOTE] = ACTIONS(2822),
    [anon_sym_SQUOTE] = ACTIONS(2822),
    [anon_sym_AT] = ACTIONS(2820),
    [anon_sym_DQUOTE] = ACTIONS(2822),
    [anon_sym_L_DQUOTE] = ACTIONS(2822),
    [anon_sym_u_DQUOTE] = ACTIONS(2822),
    [anon_sym_U_DQUOTE] = ACTIONS(2822),
    [anon_sym_u8_DQUOTE] = ACTIONS(2822),
    [sym_true] = ACTIONS(2820),
    [sym_false] = ACTIONS(2820),
    [anon_sym_NULL] = ACTIONS(2820),
    [anon_sym_nullptr] = ACTIONS(2820),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2820),
    [anon_sym___typeof] = ACTIONS(2820),
    [anon_sym_typeof] = ACTIONS(2820),
    [anon_sym_ATimport] = ACTIONS(2822),
    [aux_sym_preproc_undef_token1] = ACTIONS(2820),
    [anon_sym_POUND] = ACTIONS(2820),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2820),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2820),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2820),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2820),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2820),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2820),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2820),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_API_AVAILABLE] = ACTIONS(2820),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_API_DEPRECATED] = ACTIONS(2820),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2820),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym___deprecated_msg] = ACTIONS(2820),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2820),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2820),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2820),
    [anon_sym_ATprotocol] = ACTIONS(2822),
    [anon_sym_ATinterface] = ACTIONS(2822),
    [anon_sym_ATimplementation] = ACTIONS(2822),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2822),
    [anon_sym_ATsynthesize] = ACTIONS(2822),
    [anon_sym_ATdynamic] = ACTIONS(2822),
    [anon_sym_ATtry] = ACTIONS(2822),
    [anon_sym___try] = ACTIONS(2820),
    [anon_sym_ATthrow] = ACTIONS(2822),
    [anon_sym_ATselector] = ACTIONS(2822),
    [anon_sym_ATavailable] = ACTIONS(2822),
    [anon_sym___builtin_available] = ACTIONS(2820),
    [anon_sym_va_arg] = ACTIONS(2820),
    [anon_sym_ATencode] = ACTIONS(2822),
    [anon_sym_ATsynchronized] = ACTIONS(2822),
    [anon_sym_BOOL] = ACTIONS(2820),
    [anon_sym_IMP] = ACTIONS(2820),
    [anon_sym_SEL] = ACTIONS(2820),
    [anon_sym_Class] = ACTIONS(2820),
    [anon_sym_id] = ACTIONS(2820),
  },
  [STATE(348)] = {
    [sym_identifier] = ACTIONS(2824),
    [aux_sym_preproc_include_token1] = ACTIONS(2824),
    [aux_sym_preproc_include_token2] = ACTIONS(2824),
    [aux_sym_preproc_def_token1] = ACTIONS(2824),
    [aux_sym_preproc_if_token1] = ACTIONS(2824),
    [aux_sym_preproc_if_token2] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2824),
    [aux_sym_preproc_else_token1] = ACTIONS(2824),
    [aux_sym_preproc_elif_token1] = ACTIONS(2824),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2824),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2824),
    [sym_preproc_directive] = ACTIONS(2824),
    [anon_sym_LPAREN2] = ACTIONS(2826),
    [anon_sym_BANG] = ACTIONS(2826),
    [anon_sym_TILDE] = ACTIONS(2826),
    [anon_sym_DASH] = ACTIONS(2824),
    [anon_sym_PLUS] = ACTIONS(2824),
    [anon_sym_STAR] = ACTIONS(2826),
    [anon_sym_CARET] = ACTIONS(2826),
    [anon_sym_AMP] = ACTIONS(2826),
    [anon_sym_SEMI] = ACTIONS(2826),
    [anon_sym___extension__] = ACTIONS(2824),
    [anon_sym_typedef] = ACTIONS(2824),
    [anon_sym_extern] = ACTIONS(2824),
    [anon_sym___attribute__] = ACTIONS(2824),
    [anon_sym___attribute] = ACTIONS(2824),
    [anon_sym_noreturn] = ACTIONS(2824),
    [anon_sym_LBRACK] = ACTIONS(2826),
    [anon_sym___declspec] = ACTIONS(2824),
    [anon_sym___cdecl] = ACTIONS(2824),
    [anon_sym___clrcall] = ACTIONS(2824),
    [anon_sym___stdcall] = ACTIONS(2824),
    [anon_sym___fastcall] = ACTIONS(2824),
    [anon_sym___thiscall] = ACTIONS(2824),
    [anon_sym___vectorcall] = ACTIONS(2824),
    [anon_sym_LBRACE] = ACTIONS(2826),
    [anon_sym_signed] = ACTIONS(2824),
    [anon_sym_unsigned] = ACTIONS(2824),
    [anon_sym_long] = ACTIONS(2824),
    [anon_sym_short] = ACTIONS(2824),
    [anon_sym_static] = ACTIONS(2824),
    [anon_sym_ATautoreleasepool] = ACTIONS(2826),
    [anon_sym_auto] = ACTIONS(2824),
    [anon_sym_register] = ACTIONS(2824),
    [anon_sym_inline] = ACTIONS(2824),
    [anon_sym___inline] = ACTIONS(2824),
    [anon_sym___inline__] = ACTIONS(2824),
    [anon_sym___forceinline] = ACTIONS(2824),
    [anon_sym_thread_local] = ACTIONS(2824),
    [anon_sym___thread] = ACTIONS(2824),
    [anon_sym_CG_EXTERN] = ACTIONS(2824),
    [anon_sym_CG_INLINE] = ACTIONS(2824),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2824),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2824),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2824),
    [anon_sym_IBOutlet] = ACTIONS(2824),
    [anon_sym_IBInspectable] = ACTIONS(2824),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2824),
    [anon_sym_NS_INLINE] = ACTIONS(2824),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2824),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2824),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2824),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2824),
    [anon_sym_const] = ACTIONS(2824),
    [anon_sym_constexpr] = ACTIONS(2824),
    [anon_sym_volatile] = ACTIONS(2824),
    [anon_sym_restrict] = ACTIONS(2824),
    [anon_sym___restrict__] = ACTIONS(2824),
    [anon_sym__Atomic] = ACTIONS(2824),
    [anon_sym__Noreturn] = ACTIONS(2824),
    [anon_sym__Nonnull] = ACTIONS(2824),
    [anon_sym_nullable] = ACTIONS(2824),
    [anon_sym__Complex] = ACTIONS(2824),
    [anon_sym__Nullable] = ACTIONS(2824),
    [anon_sym__Nullable_result] = ACTIONS(2824),
    [anon_sym__Null_unspecified] = ACTIONS(2824),
    [anon_sym___autoreleasing] = ACTIONS(2824),
    [anon_sym___block] = ACTIONS(2824),
    [anon_sym___bridge] = ACTIONS(2824),
    [anon_sym___bridge_retained] = ACTIONS(2824),
    [anon_sym___bridge_transfer] = ACTIONS(2824),
    [anon_sym___complex] = ACTIONS(2824),
    [anon_sym___const] = ACTIONS(2824),
    [anon_sym___imag] = ACTIONS(2824),
    [anon_sym___kindof] = ACTIONS(2824),
    [anon_sym___nonnull] = ACTIONS(2824),
    [anon_sym___nullable] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2824),
    [anon_sym___real] = ACTIONS(2824),
    [anon_sym___strong] = ACTIONS(2824),
    [anon_sym___unsafe_unretained] = ACTIONS(2824),
    [anon_sym___unused] = ACTIONS(2824),
    [anon_sym___weak] = ACTIONS(2824),
    [anon_sym_alignas] = ACTIONS(2824),
    [anon_sym__Alignas] = ACTIONS(2824),
    [sym_primitive_type] = ACTIONS(2824),
    [anon_sym_enum] = ACTIONS(2824),
    [anon_sym_struct] = ACTIONS(2824),
    [anon_sym_union] = ACTIONS(2824),
    [anon_sym_if] = ACTIONS(2824),
    [anon_sym_switch] = ACTIONS(2824),
    [anon_sym_case] = ACTIONS(2824),
    [anon_sym_default] = ACTIONS(2824),
    [anon_sym_while] = ACTIONS(2824),
    [anon_sym_do] = ACTIONS(2824),
    [anon_sym_for] = ACTIONS(2824),
    [anon_sym_in] = ACTIONS(2824),
    [anon_sym_return] = ACTIONS(2824),
    [anon_sym_break] = ACTIONS(2824),
    [anon_sym_continue] = ACTIONS(2824),
    [anon_sym_goto] = ACTIONS(2824),
    [anon_sym_DASH_DASH] = ACTIONS(2826),
    [anon_sym_PLUS_PLUS] = ACTIONS(2826),
    [anon_sym_sizeof] = ACTIONS(2824),
    [anon_sym___alignof__] = ACTIONS(2824),
    [anon_sym___alignof] = ACTIONS(2824),
    [anon_sym__alignof] = ACTIONS(2824),
    [anon_sym_alignof] = ACTIONS(2824),
    [anon_sym__Alignof] = ACTIONS(2824),
    [anon_sym_offsetof] = ACTIONS(2824),
    [anon_sym__Generic] = ACTIONS(2824),
    [anon_sym_asm] = ACTIONS(2824),
    [anon_sym___asm__] = ACTIONS(2824),
    [anon_sym___asm] = ACTIONS(2824),
    [sym_number_literal] = ACTIONS(2826),
    [anon_sym_L_SQUOTE] = ACTIONS(2826),
    [anon_sym_u_SQUOTE] = ACTIONS(2826),
    [anon_sym_U_SQUOTE] = ACTIONS(2826),
    [anon_sym_u8_SQUOTE] = ACTIONS(2826),
    [anon_sym_SQUOTE] = ACTIONS(2826),
    [anon_sym_AT] = ACTIONS(2824),
    [anon_sym_DQUOTE] = ACTIONS(2826),
    [anon_sym_L_DQUOTE] = ACTIONS(2826),
    [anon_sym_u_DQUOTE] = ACTIONS(2826),
    [anon_sym_U_DQUOTE] = ACTIONS(2826),
    [anon_sym_u8_DQUOTE] = ACTIONS(2826),
    [sym_true] = ACTIONS(2824),
    [sym_false] = ACTIONS(2824),
    [anon_sym_NULL] = ACTIONS(2824),
    [anon_sym_nullptr] = ACTIONS(2824),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2824),
    [anon_sym___typeof] = ACTIONS(2824),
    [anon_sym_typeof] = ACTIONS(2824),
    [anon_sym_ATimport] = ACTIONS(2826),
    [aux_sym_preproc_undef_token1] = ACTIONS(2824),
    [anon_sym_POUND] = ACTIONS(2824),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2824),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2824),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2824),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2824),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2824),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2824),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2824),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_API_AVAILABLE] = ACTIONS(2824),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_API_DEPRECATED] = ACTIONS(2824),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2824),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym___deprecated_msg] = ACTIONS(2824),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2824),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2824),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2824),
    [anon_sym_ATprotocol] = ACTIONS(2826),
    [anon_sym_ATinterface] = ACTIONS(2826),
    [anon_sym_ATimplementation] = ACTIONS(2826),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2826),
    [anon_sym_ATsynthesize] = ACTIONS(2826),
    [anon_sym_ATdynamic] = ACTIONS(2826),
    [anon_sym_ATtry] = ACTIONS(2826),
    [anon_sym___try] = ACTIONS(2824),
    [anon_sym_ATthrow] = ACTIONS(2826),
    [anon_sym_ATselector] = ACTIONS(2826),
    [anon_sym_ATavailable] = ACTIONS(2826),
    [anon_sym___builtin_available] = ACTIONS(2824),
    [anon_sym_va_arg] = ACTIONS(2824),
    [anon_sym_ATencode] = ACTIONS(2826),
    [anon_sym_ATsynchronized] = ACTIONS(2826),
    [anon_sym_BOOL] = ACTIONS(2824),
    [anon_sym_IMP] = ACTIONS(2824),
    [anon_sym_SEL] = ACTIONS(2824),
    [anon_sym_Class] = ACTIONS(2824),
    [anon_sym_id] = ACTIONS(2824),
  },
  [STATE(349)] = {
    [sym_identifier] = ACTIONS(2828),
    [aux_sym_preproc_include_token1] = ACTIONS(2828),
    [aux_sym_preproc_include_token2] = ACTIONS(2828),
    [aux_sym_preproc_def_token1] = ACTIONS(2828),
    [aux_sym_preproc_if_token1] = ACTIONS(2828),
    [aux_sym_preproc_if_token2] = ACTIONS(2828),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2828),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2828),
    [aux_sym_preproc_else_token1] = ACTIONS(2828),
    [aux_sym_preproc_elif_token1] = ACTIONS(2828),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2828),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2828),
    [sym_preproc_directive] = ACTIONS(2828),
    [anon_sym_LPAREN2] = ACTIONS(2830),
    [anon_sym_BANG] = ACTIONS(2830),
    [anon_sym_TILDE] = ACTIONS(2830),
    [anon_sym_DASH] = ACTIONS(2828),
    [anon_sym_PLUS] = ACTIONS(2828),
    [anon_sym_STAR] = ACTIONS(2830),
    [anon_sym_CARET] = ACTIONS(2830),
    [anon_sym_AMP] = ACTIONS(2830),
    [anon_sym_SEMI] = ACTIONS(2830),
    [anon_sym___extension__] = ACTIONS(2828),
    [anon_sym_typedef] = ACTIONS(2828),
    [anon_sym_extern] = ACTIONS(2828),
    [anon_sym___attribute__] = ACTIONS(2828),
    [anon_sym___attribute] = ACTIONS(2828),
    [anon_sym_noreturn] = ACTIONS(2828),
    [anon_sym_LBRACK] = ACTIONS(2830),
    [anon_sym___declspec] = ACTIONS(2828),
    [anon_sym___cdecl] = ACTIONS(2828),
    [anon_sym___clrcall] = ACTIONS(2828),
    [anon_sym___stdcall] = ACTIONS(2828),
    [anon_sym___fastcall] = ACTIONS(2828),
    [anon_sym___thiscall] = ACTIONS(2828),
    [anon_sym___vectorcall] = ACTIONS(2828),
    [anon_sym_LBRACE] = ACTIONS(2830),
    [anon_sym_signed] = ACTIONS(2828),
    [anon_sym_unsigned] = ACTIONS(2828),
    [anon_sym_long] = ACTIONS(2828),
    [anon_sym_short] = ACTIONS(2828),
    [anon_sym_static] = ACTIONS(2828),
    [anon_sym_ATautoreleasepool] = ACTIONS(2830),
    [anon_sym_auto] = ACTIONS(2828),
    [anon_sym_register] = ACTIONS(2828),
    [anon_sym_inline] = ACTIONS(2828),
    [anon_sym___inline] = ACTIONS(2828),
    [anon_sym___inline__] = ACTIONS(2828),
    [anon_sym___forceinline] = ACTIONS(2828),
    [anon_sym_thread_local] = ACTIONS(2828),
    [anon_sym___thread] = ACTIONS(2828),
    [anon_sym_CG_EXTERN] = ACTIONS(2828),
    [anon_sym_CG_INLINE] = ACTIONS(2828),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2828),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2828),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2828),
    [anon_sym_IBOutlet] = ACTIONS(2828),
    [anon_sym_IBInspectable] = ACTIONS(2828),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2828),
    [anon_sym_NS_INLINE] = ACTIONS(2828),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2828),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2828),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2828),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2828),
    [anon_sym_const] = ACTIONS(2828),
    [anon_sym_constexpr] = ACTIONS(2828),
    [anon_sym_volatile] = ACTIONS(2828),
    [anon_sym_restrict] = ACTIONS(2828),
    [anon_sym___restrict__] = ACTIONS(2828),
    [anon_sym__Atomic] = ACTIONS(2828),
    [anon_sym__Noreturn] = ACTIONS(2828),
    [anon_sym__Nonnull] = ACTIONS(2828),
    [anon_sym_nullable] = ACTIONS(2828),
    [anon_sym__Complex] = ACTIONS(2828),
    [anon_sym__Nullable] = ACTIONS(2828),
    [anon_sym__Nullable_result] = ACTIONS(2828),
    [anon_sym__Null_unspecified] = ACTIONS(2828),
    [anon_sym___autoreleasing] = ACTIONS(2828),
    [anon_sym___block] = ACTIONS(2828),
    [anon_sym___bridge] = ACTIONS(2828),
    [anon_sym___bridge_retained] = ACTIONS(2828),
    [anon_sym___bridge_transfer] = ACTIONS(2828),
    [anon_sym___complex] = ACTIONS(2828),
    [anon_sym___const] = ACTIONS(2828),
    [anon_sym___imag] = ACTIONS(2828),
    [anon_sym___kindof] = ACTIONS(2828),
    [anon_sym___nonnull] = ACTIONS(2828),
    [anon_sym___nullable] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2828),
    [anon_sym___real] = ACTIONS(2828),
    [anon_sym___strong] = ACTIONS(2828),
    [anon_sym___unsafe_unretained] = ACTIONS(2828),
    [anon_sym___unused] = ACTIONS(2828),
    [anon_sym___weak] = ACTIONS(2828),
    [anon_sym_alignas] = ACTIONS(2828),
    [anon_sym__Alignas] = ACTIONS(2828),
    [sym_primitive_type] = ACTIONS(2828),
    [anon_sym_enum] = ACTIONS(2828),
    [anon_sym_struct] = ACTIONS(2828),
    [anon_sym_union] = ACTIONS(2828),
    [anon_sym_if] = ACTIONS(2828),
    [anon_sym_switch] = ACTIONS(2828),
    [anon_sym_case] = ACTIONS(2828),
    [anon_sym_default] = ACTIONS(2828),
    [anon_sym_while] = ACTIONS(2828),
    [anon_sym_do] = ACTIONS(2828),
    [anon_sym_for] = ACTIONS(2828),
    [anon_sym_in] = ACTIONS(2828),
    [anon_sym_return] = ACTIONS(2828),
    [anon_sym_break] = ACTIONS(2828),
    [anon_sym_continue] = ACTIONS(2828),
    [anon_sym_goto] = ACTIONS(2828),
    [anon_sym_DASH_DASH] = ACTIONS(2830),
    [anon_sym_PLUS_PLUS] = ACTIONS(2830),
    [anon_sym_sizeof] = ACTIONS(2828),
    [anon_sym___alignof__] = ACTIONS(2828),
    [anon_sym___alignof] = ACTIONS(2828),
    [anon_sym__alignof] = ACTIONS(2828),
    [anon_sym_alignof] = ACTIONS(2828),
    [anon_sym__Alignof] = ACTIONS(2828),
    [anon_sym_offsetof] = ACTIONS(2828),
    [anon_sym__Generic] = ACTIONS(2828),
    [anon_sym_asm] = ACTIONS(2828),
    [anon_sym___asm__] = ACTIONS(2828),
    [anon_sym___asm] = ACTIONS(2828),
    [sym_number_literal] = ACTIONS(2830),
    [anon_sym_L_SQUOTE] = ACTIONS(2830),
    [anon_sym_u_SQUOTE] = ACTIONS(2830),
    [anon_sym_U_SQUOTE] = ACTIONS(2830),
    [anon_sym_u8_SQUOTE] = ACTIONS(2830),
    [anon_sym_SQUOTE] = ACTIONS(2830),
    [anon_sym_AT] = ACTIONS(2828),
    [anon_sym_DQUOTE] = ACTIONS(2830),
    [anon_sym_L_DQUOTE] = ACTIONS(2830),
    [anon_sym_u_DQUOTE] = ACTIONS(2830),
    [anon_sym_U_DQUOTE] = ACTIONS(2830),
    [anon_sym_u8_DQUOTE] = ACTIONS(2830),
    [sym_true] = ACTIONS(2828),
    [sym_false] = ACTIONS(2828),
    [anon_sym_NULL] = ACTIONS(2828),
    [anon_sym_nullptr] = ACTIONS(2828),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2828),
    [anon_sym___typeof] = ACTIONS(2828),
    [anon_sym_typeof] = ACTIONS(2828),
    [anon_sym_ATimport] = ACTIONS(2830),
    [aux_sym_preproc_undef_token1] = ACTIONS(2828),
    [anon_sym_POUND] = ACTIONS(2828),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2828),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2828),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2828),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2828),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2828),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2828),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2828),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_API_AVAILABLE] = ACTIONS(2828),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_API_DEPRECATED] = ACTIONS(2828),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2828),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym___deprecated_msg] = ACTIONS(2828),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2828),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2828),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2828),
    [anon_sym_ATprotocol] = ACTIONS(2830),
    [anon_sym_ATinterface] = ACTIONS(2830),
    [anon_sym_ATimplementation] = ACTIONS(2830),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2830),
    [anon_sym_ATsynthesize] = ACTIONS(2830),
    [anon_sym_ATdynamic] = ACTIONS(2830),
    [anon_sym_ATtry] = ACTIONS(2830),
    [anon_sym___try] = ACTIONS(2828),
    [anon_sym_ATthrow] = ACTIONS(2830),
    [anon_sym_ATselector] = ACTIONS(2830),
    [anon_sym_ATavailable] = ACTIONS(2830),
    [anon_sym___builtin_available] = ACTIONS(2828),
    [anon_sym_va_arg] = ACTIONS(2828),
    [anon_sym_ATencode] = ACTIONS(2830),
    [anon_sym_ATsynchronized] = ACTIONS(2830),
    [anon_sym_BOOL] = ACTIONS(2828),
    [anon_sym_IMP] = ACTIONS(2828),
    [anon_sym_SEL] = ACTIONS(2828),
    [anon_sym_Class] = ACTIONS(2828),
    [anon_sym_id] = ACTIONS(2828),
  },
  [STATE(350)] = {
    [sym_identifier] = ACTIONS(2832),
    [aux_sym_preproc_include_token1] = ACTIONS(2832),
    [aux_sym_preproc_include_token2] = ACTIONS(2832),
    [aux_sym_preproc_def_token1] = ACTIONS(2832),
    [aux_sym_preproc_if_token1] = ACTIONS(2832),
    [aux_sym_preproc_if_token2] = ACTIONS(2832),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2832),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2832),
    [aux_sym_preproc_else_token1] = ACTIONS(2832),
    [aux_sym_preproc_elif_token1] = ACTIONS(2832),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2832),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2832),
    [sym_preproc_directive] = ACTIONS(2832),
    [anon_sym_LPAREN2] = ACTIONS(2834),
    [anon_sym_BANG] = ACTIONS(2834),
    [anon_sym_TILDE] = ACTIONS(2834),
    [anon_sym_DASH] = ACTIONS(2832),
    [anon_sym_PLUS] = ACTIONS(2832),
    [anon_sym_STAR] = ACTIONS(2834),
    [anon_sym_CARET] = ACTIONS(2834),
    [anon_sym_AMP] = ACTIONS(2834),
    [anon_sym_SEMI] = ACTIONS(2834),
    [anon_sym___extension__] = ACTIONS(2832),
    [anon_sym_typedef] = ACTIONS(2832),
    [anon_sym_extern] = ACTIONS(2832),
    [anon_sym___attribute__] = ACTIONS(2832),
    [anon_sym___attribute] = ACTIONS(2832),
    [anon_sym_noreturn] = ACTIONS(2832),
    [anon_sym_LBRACK] = ACTIONS(2834),
    [anon_sym___declspec] = ACTIONS(2832),
    [anon_sym___cdecl] = ACTIONS(2832),
    [anon_sym___clrcall] = ACTIONS(2832),
    [anon_sym___stdcall] = ACTIONS(2832),
    [anon_sym___fastcall] = ACTIONS(2832),
    [anon_sym___thiscall] = ACTIONS(2832),
    [anon_sym___vectorcall] = ACTIONS(2832),
    [anon_sym_LBRACE] = ACTIONS(2834),
    [anon_sym_signed] = ACTIONS(2832),
    [anon_sym_unsigned] = ACTIONS(2832),
    [anon_sym_long] = ACTIONS(2832),
    [anon_sym_short] = ACTIONS(2832),
    [anon_sym_static] = ACTIONS(2832),
    [anon_sym_ATautoreleasepool] = ACTIONS(2834),
    [anon_sym_auto] = ACTIONS(2832),
    [anon_sym_register] = ACTIONS(2832),
    [anon_sym_inline] = ACTIONS(2832),
    [anon_sym___inline] = ACTIONS(2832),
    [anon_sym___inline__] = ACTIONS(2832),
    [anon_sym___forceinline] = ACTIONS(2832),
    [anon_sym_thread_local] = ACTIONS(2832),
    [anon_sym___thread] = ACTIONS(2832),
    [anon_sym_CG_EXTERN] = ACTIONS(2832),
    [anon_sym_CG_INLINE] = ACTIONS(2832),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2832),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2832),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2832),
    [anon_sym_IBOutlet] = ACTIONS(2832),
    [anon_sym_IBInspectable] = ACTIONS(2832),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2832),
    [anon_sym_NS_INLINE] = ACTIONS(2832),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2832),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2832),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2832),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2832),
    [anon_sym_const] = ACTIONS(2832),
    [anon_sym_constexpr] = ACTIONS(2832),
    [anon_sym_volatile] = ACTIONS(2832),
    [anon_sym_restrict] = ACTIONS(2832),
    [anon_sym___restrict__] = ACTIONS(2832),
    [anon_sym__Atomic] = ACTIONS(2832),
    [anon_sym__Noreturn] = ACTIONS(2832),
    [anon_sym__Nonnull] = ACTIONS(2832),
    [anon_sym_nullable] = ACTIONS(2832),
    [anon_sym__Complex] = ACTIONS(2832),
    [anon_sym__Nullable] = ACTIONS(2832),
    [anon_sym__Nullable_result] = ACTIONS(2832),
    [anon_sym__Null_unspecified] = ACTIONS(2832),
    [anon_sym___autoreleasing] = ACTIONS(2832),
    [anon_sym___block] = ACTIONS(2832),
    [anon_sym___bridge] = ACTIONS(2832),
    [anon_sym___bridge_retained] = ACTIONS(2832),
    [anon_sym___bridge_transfer] = ACTIONS(2832),
    [anon_sym___complex] = ACTIONS(2832),
    [anon_sym___const] = ACTIONS(2832),
    [anon_sym___imag] = ACTIONS(2832),
    [anon_sym___kindof] = ACTIONS(2832),
    [anon_sym___nonnull] = ACTIONS(2832),
    [anon_sym___nullable] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2832),
    [anon_sym___real] = ACTIONS(2832),
    [anon_sym___strong] = ACTIONS(2832),
    [anon_sym___unsafe_unretained] = ACTIONS(2832),
    [anon_sym___unused] = ACTIONS(2832),
    [anon_sym___weak] = ACTIONS(2832),
    [anon_sym_alignas] = ACTIONS(2832),
    [anon_sym__Alignas] = ACTIONS(2832),
    [sym_primitive_type] = ACTIONS(2832),
    [anon_sym_enum] = ACTIONS(2832),
    [anon_sym_struct] = ACTIONS(2832),
    [anon_sym_union] = ACTIONS(2832),
    [anon_sym_if] = ACTIONS(2832),
    [anon_sym_switch] = ACTIONS(2832),
    [anon_sym_case] = ACTIONS(2832),
    [anon_sym_default] = ACTIONS(2832),
    [anon_sym_while] = ACTIONS(2832),
    [anon_sym_do] = ACTIONS(2832),
    [anon_sym_for] = ACTIONS(2832),
    [anon_sym_in] = ACTIONS(2832),
    [anon_sym_return] = ACTIONS(2832),
    [anon_sym_break] = ACTIONS(2832),
    [anon_sym_continue] = ACTIONS(2832),
    [anon_sym_goto] = ACTIONS(2832),
    [anon_sym_DASH_DASH] = ACTIONS(2834),
    [anon_sym_PLUS_PLUS] = ACTIONS(2834),
    [anon_sym_sizeof] = ACTIONS(2832),
    [anon_sym___alignof__] = ACTIONS(2832),
    [anon_sym___alignof] = ACTIONS(2832),
    [anon_sym__alignof] = ACTIONS(2832),
    [anon_sym_alignof] = ACTIONS(2832),
    [anon_sym__Alignof] = ACTIONS(2832),
    [anon_sym_offsetof] = ACTIONS(2832),
    [anon_sym__Generic] = ACTIONS(2832),
    [anon_sym_asm] = ACTIONS(2832),
    [anon_sym___asm__] = ACTIONS(2832),
    [anon_sym___asm] = ACTIONS(2832),
    [sym_number_literal] = ACTIONS(2834),
    [anon_sym_L_SQUOTE] = ACTIONS(2834),
    [anon_sym_u_SQUOTE] = ACTIONS(2834),
    [anon_sym_U_SQUOTE] = ACTIONS(2834),
    [anon_sym_u8_SQUOTE] = ACTIONS(2834),
    [anon_sym_SQUOTE] = ACTIONS(2834),
    [anon_sym_AT] = ACTIONS(2832),
    [anon_sym_DQUOTE] = ACTIONS(2834),
    [anon_sym_L_DQUOTE] = ACTIONS(2834),
    [anon_sym_u_DQUOTE] = ACTIONS(2834),
    [anon_sym_U_DQUOTE] = ACTIONS(2834),
    [anon_sym_u8_DQUOTE] = ACTIONS(2834),
    [sym_true] = ACTIONS(2832),
    [sym_false] = ACTIONS(2832),
    [anon_sym_NULL] = ACTIONS(2832),
    [anon_sym_nullptr] = ACTIONS(2832),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2832),
    [anon_sym___typeof] = ACTIONS(2832),
    [anon_sym_typeof] = ACTIONS(2832),
    [anon_sym_ATimport] = ACTIONS(2834),
    [aux_sym_preproc_undef_token1] = ACTIONS(2832),
    [anon_sym_POUND] = ACTIONS(2832),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2832),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2832),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2832),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2832),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2832),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2832),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2832),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_API_AVAILABLE] = ACTIONS(2832),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_API_DEPRECATED] = ACTIONS(2832),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2832),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym___deprecated_msg] = ACTIONS(2832),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2832),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2832),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2832),
    [anon_sym_ATprotocol] = ACTIONS(2834),
    [anon_sym_ATinterface] = ACTIONS(2834),
    [anon_sym_ATimplementation] = ACTIONS(2834),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2834),
    [anon_sym_ATsynthesize] = ACTIONS(2834),
    [anon_sym_ATdynamic] = ACTIONS(2834),
    [anon_sym_ATtry] = ACTIONS(2834),
    [anon_sym___try] = ACTIONS(2832),
    [anon_sym_ATthrow] = ACTIONS(2834),
    [anon_sym_ATselector] = ACTIONS(2834),
    [anon_sym_ATavailable] = ACTIONS(2834),
    [anon_sym___builtin_available] = ACTIONS(2832),
    [anon_sym_va_arg] = ACTIONS(2832),
    [anon_sym_ATencode] = ACTIONS(2834),
    [anon_sym_ATsynchronized] = ACTIONS(2834),
    [anon_sym_BOOL] = ACTIONS(2832),
    [anon_sym_IMP] = ACTIONS(2832),
    [anon_sym_SEL] = ACTIONS(2832),
    [anon_sym_Class] = ACTIONS(2832),
    [anon_sym_id] = ACTIONS(2832),
  },
  [STATE(351)] = {
    [sym_identifier] = ACTIONS(2836),
    [aux_sym_preproc_include_token1] = ACTIONS(2836),
    [aux_sym_preproc_include_token2] = ACTIONS(2836),
    [aux_sym_preproc_def_token1] = ACTIONS(2836),
    [aux_sym_preproc_if_token1] = ACTIONS(2836),
    [aux_sym_preproc_if_token2] = ACTIONS(2836),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2836),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2836),
    [aux_sym_preproc_else_token1] = ACTIONS(2836),
    [aux_sym_preproc_elif_token1] = ACTIONS(2836),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2836),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2836),
    [sym_preproc_directive] = ACTIONS(2836),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_BANG] = ACTIONS(2838),
    [anon_sym_TILDE] = ACTIONS(2838),
    [anon_sym_DASH] = ACTIONS(2836),
    [anon_sym_PLUS] = ACTIONS(2836),
    [anon_sym_STAR] = ACTIONS(2838),
    [anon_sym_CARET] = ACTIONS(2838),
    [anon_sym_AMP] = ACTIONS(2838),
    [anon_sym_SEMI] = ACTIONS(2838),
    [anon_sym___extension__] = ACTIONS(2836),
    [anon_sym_typedef] = ACTIONS(2836),
    [anon_sym_extern] = ACTIONS(2836),
    [anon_sym___attribute__] = ACTIONS(2836),
    [anon_sym___attribute] = ACTIONS(2836),
    [anon_sym_noreturn] = ACTIONS(2836),
    [anon_sym_LBRACK] = ACTIONS(2838),
    [anon_sym___declspec] = ACTIONS(2836),
    [anon_sym___cdecl] = ACTIONS(2836),
    [anon_sym___clrcall] = ACTIONS(2836),
    [anon_sym___stdcall] = ACTIONS(2836),
    [anon_sym___fastcall] = ACTIONS(2836),
    [anon_sym___thiscall] = ACTIONS(2836),
    [anon_sym___vectorcall] = ACTIONS(2836),
    [anon_sym_LBRACE] = ACTIONS(2838),
    [anon_sym_signed] = ACTIONS(2836),
    [anon_sym_unsigned] = ACTIONS(2836),
    [anon_sym_long] = ACTIONS(2836),
    [anon_sym_short] = ACTIONS(2836),
    [anon_sym_static] = ACTIONS(2836),
    [anon_sym_ATautoreleasepool] = ACTIONS(2838),
    [anon_sym_auto] = ACTIONS(2836),
    [anon_sym_register] = ACTIONS(2836),
    [anon_sym_inline] = ACTIONS(2836),
    [anon_sym___inline] = ACTIONS(2836),
    [anon_sym___inline__] = ACTIONS(2836),
    [anon_sym___forceinline] = ACTIONS(2836),
    [anon_sym_thread_local] = ACTIONS(2836),
    [anon_sym___thread] = ACTIONS(2836),
    [anon_sym_CG_EXTERN] = ACTIONS(2836),
    [anon_sym_CG_INLINE] = ACTIONS(2836),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2836),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2836),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2836),
    [anon_sym_IBOutlet] = ACTIONS(2836),
    [anon_sym_IBInspectable] = ACTIONS(2836),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2836),
    [anon_sym_NS_INLINE] = ACTIONS(2836),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2836),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2836),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2836),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2836),
    [anon_sym_const] = ACTIONS(2836),
    [anon_sym_constexpr] = ACTIONS(2836),
    [anon_sym_volatile] = ACTIONS(2836),
    [anon_sym_restrict] = ACTIONS(2836),
    [anon_sym___restrict__] = ACTIONS(2836),
    [anon_sym__Atomic] = ACTIONS(2836),
    [anon_sym__Noreturn] = ACTIONS(2836),
    [anon_sym__Nonnull] = ACTIONS(2836),
    [anon_sym_nullable] = ACTIONS(2836),
    [anon_sym__Complex] = ACTIONS(2836),
    [anon_sym__Nullable] = ACTIONS(2836),
    [anon_sym__Nullable_result] = ACTIONS(2836),
    [anon_sym__Null_unspecified] = ACTIONS(2836),
    [anon_sym___autoreleasing] = ACTIONS(2836),
    [anon_sym___block] = ACTIONS(2836),
    [anon_sym___bridge] = ACTIONS(2836),
    [anon_sym___bridge_retained] = ACTIONS(2836),
    [anon_sym___bridge_transfer] = ACTIONS(2836),
    [anon_sym___complex] = ACTIONS(2836),
    [anon_sym___const] = ACTIONS(2836),
    [anon_sym___imag] = ACTIONS(2836),
    [anon_sym___kindof] = ACTIONS(2836),
    [anon_sym___nonnull] = ACTIONS(2836),
    [anon_sym___nullable] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2836),
    [anon_sym___real] = ACTIONS(2836),
    [anon_sym___strong] = ACTIONS(2836),
    [anon_sym___unsafe_unretained] = ACTIONS(2836),
    [anon_sym___unused] = ACTIONS(2836),
    [anon_sym___weak] = ACTIONS(2836),
    [anon_sym_alignas] = ACTIONS(2836),
    [anon_sym__Alignas] = ACTIONS(2836),
    [sym_primitive_type] = ACTIONS(2836),
    [anon_sym_enum] = ACTIONS(2836),
    [anon_sym_struct] = ACTIONS(2836),
    [anon_sym_union] = ACTIONS(2836),
    [anon_sym_if] = ACTIONS(2836),
    [anon_sym_switch] = ACTIONS(2836),
    [anon_sym_case] = ACTIONS(2836),
    [anon_sym_default] = ACTIONS(2836),
    [anon_sym_while] = ACTIONS(2836),
    [anon_sym_do] = ACTIONS(2836),
    [anon_sym_for] = ACTIONS(2836),
    [anon_sym_in] = ACTIONS(2836),
    [anon_sym_return] = ACTIONS(2836),
    [anon_sym_break] = ACTIONS(2836),
    [anon_sym_continue] = ACTIONS(2836),
    [anon_sym_goto] = ACTIONS(2836),
    [anon_sym_DASH_DASH] = ACTIONS(2838),
    [anon_sym_PLUS_PLUS] = ACTIONS(2838),
    [anon_sym_sizeof] = ACTIONS(2836),
    [anon_sym___alignof__] = ACTIONS(2836),
    [anon_sym___alignof] = ACTIONS(2836),
    [anon_sym__alignof] = ACTIONS(2836),
    [anon_sym_alignof] = ACTIONS(2836),
    [anon_sym__Alignof] = ACTIONS(2836),
    [anon_sym_offsetof] = ACTIONS(2836),
    [anon_sym__Generic] = ACTIONS(2836),
    [anon_sym_asm] = ACTIONS(2836),
    [anon_sym___asm__] = ACTIONS(2836),
    [anon_sym___asm] = ACTIONS(2836),
    [sym_number_literal] = ACTIONS(2838),
    [anon_sym_L_SQUOTE] = ACTIONS(2838),
    [anon_sym_u_SQUOTE] = ACTIONS(2838),
    [anon_sym_U_SQUOTE] = ACTIONS(2838),
    [anon_sym_u8_SQUOTE] = ACTIONS(2838),
    [anon_sym_SQUOTE] = ACTIONS(2838),
    [anon_sym_AT] = ACTIONS(2836),
    [anon_sym_DQUOTE] = ACTIONS(2838),
    [anon_sym_L_DQUOTE] = ACTIONS(2838),
    [anon_sym_u_DQUOTE] = ACTIONS(2838),
    [anon_sym_U_DQUOTE] = ACTIONS(2838),
    [anon_sym_u8_DQUOTE] = ACTIONS(2838),
    [sym_true] = ACTIONS(2836),
    [sym_false] = ACTIONS(2836),
    [anon_sym_NULL] = ACTIONS(2836),
    [anon_sym_nullptr] = ACTIONS(2836),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2836),
    [anon_sym___typeof] = ACTIONS(2836),
    [anon_sym_typeof] = ACTIONS(2836),
    [anon_sym_ATimport] = ACTIONS(2838),
    [aux_sym_preproc_undef_token1] = ACTIONS(2836),
    [anon_sym_POUND] = ACTIONS(2836),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2836),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2836),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2836),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2836),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2836),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2836),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2836),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_API_AVAILABLE] = ACTIONS(2836),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_API_DEPRECATED] = ACTIONS(2836),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2836),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym___deprecated_msg] = ACTIONS(2836),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2836),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2836),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2836),
    [anon_sym_ATprotocol] = ACTIONS(2838),
    [anon_sym_ATinterface] = ACTIONS(2838),
    [anon_sym_ATimplementation] = ACTIONS(2838),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2838),
    [anon_sym_ATsynthesize] = ACTIONS(2838),
    [anon_sym_ATdynamic] = ACTIONS(2838),
    [anon_sym_ATtry] = ACTIONS(2838),
    [anon_sym___try] = ACTIONS(2836),
    [anon_sym_ATthrow] = ACTIONS(2838),
    [anon_sym_ATselector] = ACTIONS(2838),
    [anon_sym_ATavailable] = ACTIONS(2838),
    [anon_sym___builtin_available] = ACTIONS(2836),
    [anon_sym_va_arg] = ACTIONS(2836),
    [anon_sym_ATencode] = ACTIONS(2838),
    [anon_sym_ATsynchronized] = ACTIONS(2838),
    [anon_sym_BOOL] = ACTIONS(2836),
    [anon_sym_IMP] = ACTIONS(2836),
    [anon_sym_SEL] = ACTIONS(2836),
    [anon_sym_Class] = ACTIONS(2836),
    [anon_sym_id] = ACTIONS(2836),
  },
  [STATE(352)] = {
    [sym_identifier] = ACTIONS(2840),
    [aux_sym_preproc_include_token1] = ACTIONS(2840),
    [aux_sym_preproc_include_token2] = ACTIONS(2840),
    [aux_sym_preproc_def_token1] = ACTIONS(2840),
    [aux_sym_preproc_if_token1] = ACTIONS(2840),
    [aux_sym_preproc_if_token2] = ACTIONS(2840),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2840),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2840),
    [aux_sym_preproc_else_token1] = ACTIONS(2840),
    [aux_sym_preproc_elif_token1] = ACTIONS(2840),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2840),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2840),
    [sym_preproc_directive] = ACTIONS(2840),
    [anon_sym_LPAREN2] = ACTIONS(2842),
    [anon_sym_BANG] = ACTIONS(2842),
    [anon_sym_TILDE] = ACTIONS(2842),
    [anon_sym_DASH] = ACTIONS(2840),
    [anon_sym_PLUS] = ACTIONS(2840),
    [anon_sym_STAR] = ACTIONS(2842),
    [anon_sym_CARET] = ACTIONS(2842),
    [anon_sym_AMP] = ACTIONS(2842),
    [anon_sym_SEMI] = ACTIONS(2842),
    [anon_sym___extension__] = ACTIONS(2840),
    [anon_sym_typedef] = ACTIONS(2840),
    [anon_sym_extern] = ACTIONS(2840),
    [anon_sym___attribute__] = ACTIONS(2840),
    [anon_sym___attribute] = ACTIONS(2840),
    [anon_sym_noreturn] = ACTIONS(2840),
    [anon_sym_LBRACK] = ACTIONS(2842),
    [anon_sym___declspec] = ACTIONS(2840),
    [anon_sym___cdecl] = ACTIONS(2840),
    [anon_sym___clrcall] = ACTIONS(2840),
    [anon_sym___stdcall] = ACTIONS(2840),
    [anon_sym___fastcall] = ACTIONS(2840),
    [anon_sym___thiscall] = ACTIONS(2840),
    [anon_sym___vectorcall] = ACTIONS(2840),
    [anon_sym_LBRACE] = ACTIONS(2842),
    [anon_sym_signed] = ACTIONS(2840),
    [anon_sym_unsigned] = ACTIONS(2840),
    [anon_sym_long] = ACTIONS(2840),
    [anon_sym_short] = ACTIONS(2840),
    [anon_sym_static] = ACTIONS(2840),
    [anon_sym_ATautoreleasepool] = ACTIONS(2842),
    [anon_sym_auto] = ACTIONS(2840),
    [anon_sym_register] = ACTIONS(2840),
    [anon_sym_inline] = ACTIONS(2840),
    [anon_sym___inline] = ACTIONS(2840),
    [anon_sym___inline__] = ACTIONS(2840),
    [anon_sym___forceinline] = ACTIONS(2840),
    [anon_sym_thread_local] = ACTIONS(2840),
    [anon_sym___thread] = ACTIONS(2840),
    [anon_sym_CG_EXTERN] = ACTIONS(2840),
    [anon_sym_CG_INLINE] = ACTIONS(2840),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2840),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2840),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2840),
    [anon_sym_IBOutlet] = ACTIONS(2840),
    [anon_sym_IBInspectable] = ACTIONS(2840),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2840),
    [anon_sym_NS_INLINE] = ACTIONS(2840),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2840),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2840),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2840),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2840),
    [anon_sym_const] = ACTIONS(2840),
    [anon_sym_constexpr] = ACTIONS(2840),
    [anon_sym_volatile] = ACTIONS(2840),
    [anon_sym_restrict] = ACTIONS(2840),
    [anon_sym___restrict__] = ACTIONS(2840),
    [anon_sym__Atomic] = ACTIONS(2840),
    [anon_sym__Noreturn] = ACTIONS(2840),
    [anon_sym__Nonnull] = ACTIONS(2840),
    [anon_sym_nullable] = ACTIONS(2840),
    [anon_sym__Complex] = ACTIONS(2840),
    [anon_sym__Nullable] = ACTIONS(2840),
    [anon_sym__Nullable_result] = ACTIONS(2840),
    [anon_sym__Null_unspecified] = ACTIONS(2840),
    [anon_sym___autoreleasing] = ACTIONS(2840),
    [anon_sym___block] = ACTIONS(2840),
    [anon_sym___bridge] = ACTIONS(2840),
    [anon_sym___bridge_retained] = ACTIONS(2840),
    [anon_sym___bridge_transfer] = ACTIONS(2840),
    [anon_sym___complex] = ACTIONS(2840),
    [anon_sym___const] = ACTIONS(2840),
    [anon_sym___imag] = ACTIONS(2840),
    [anon_sym___kindof] = ACTIONS(2840),
    [anon_sym___nonnull] = ACTIONS(2840),
    [anon_sym___nullable] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2840),
    [anon_sym___real] = ACTIONS(2840),
    [anon_sym___strong] = ACTIONS(2840),
    [anon_sym___unsafe_unretained] = ACTIONS(2840),
    [anon_sym___unused] = ACTIONS(2840),
    [anon_sym___weak] = ACTIONS(2840),
    [anon_sym_alignas] = ACTIONS(2840),
    [anon_sym__Alignas] = ACTIONS(2840),
    [sym_primitive_type] = ACTIONS(2840),
    [anon_sym_enum] = ACTIONS(2840),
    [anon_sym_struct] = ACTIONS(2840),
    [anon_sym_union] = ACTIONS(2840),
    [anon_sym_if] = ACTIONS(2840),
    [anon_sym_switch] = ACTIONS(2840),
    [anon_sym_case] = ACTIONS(2840),
    [anon_sym_default] = ACTIONS(2840),
    [anon_sym_while] = ACTIONS(2840),
    [anon_sym_do] = ACTIONS(2840),
    [anon_sym_for] = ACTIONS(2840),
    [anon_sym_in] = ACTIONS(2840),
    [anon_sym_return] = ACTIONS(2840),
    [anon_sym_break] = ACTIONS(2840),
    [anon_sym_continue] = ACTIONS(2840),
    [anon_sym_goto] = ACTIONS(2840),
    [anon_sym_DASH_DASH] = ACTIONS(2842),
    [anon_sym_PLUS_PLUS] = ACTIONS(2842),
    [anon_sym_sizeof] = ACTIONS(2840),
    [anon_sym___alignof__] = ACTIONS(2840),
    [anon_sym___alignof] = ACTIONS(2840),
    [anon_sym__alignof] = ACTIONS(2840),
    [anon_sym_alignof] = ACTIONS(2840),
    [anon_sym__Alignof] = ACTIONS(2840),
    [anon_sym_offsetof] = ACTIONS(2840),
    [anon_sym__Generic] = ACTIONS(2840),
    [anon_sym_asm] = ACTIONS(2840),
    [anon_sym___asm__] = ACTIONS(2840),
    [anon_sym___asm] = ACTIONS(2840),
    [sym_number_literal] = ACTIONS(2842),
    [anon_sym_L_SQUOTE] = ACTIONS(2842),
    [anon_sym_u_SQUOTE] = ACTIONS(2842),
    [anon_sym_U_SQUOTE] = ACTIONS(2842),
    [anon_sym_u8_SQUOTE] = ACTIONS(2842),
    [anon_sym_SQUOTE] = ACTIONS(2842),
    [anon_sym_AT] = ACTIONS(2840),
    [anon_sym_DQUOTE] = ACTIONS(2842),
    [anon_sym_L_DQUOTE] = ACTIONS(2842),
    [anon_sym_u_DQUOTE] = ACTIONS(2842),
    [anon_sym_U_DQUOTE] = ACTIONS(2842),
    [anon_sym_u8_DQUOTE] = ACTIONS(2842),
    [sym_true] = ACTIONS(2840),
    [sym_false] = ACTIONS(2840),
    [anon_sym_NULL] = ACTIONS(2840),
    [anon_sym_nullptr] = ACTIONS(2840),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2840),
    [anon_sym___typeof] = ACTIONS(2840),
    [anon_sym_typeof] = ACTIONS(2840),
    [anon_sym_ATimport] = ACTIONS(2842),
    [aux_sym_preproc_undef_token1] = ACTIONS(2840),
    [anon_sym_POUND] = ACTIONS(2840),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2840),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2840),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2840),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2840),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2840),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2840),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2840),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_API_AVAILABLE] = ACTIONS(2840),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_API_DEPRECATED] = ACTIONS(2840),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2840),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym___deprecated_msg] = ACTIONS(2840),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2840),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2840),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2840),
    [anon_sym_ATprotocol] = ACTIONS(2842),
    [anon_sym_ATinterface] = ACTIONS(2842),
    [anon_sym_ATimplementation] = ACTIONS(2842),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2842),
    [anon_sym_ATsynthesize] = ACTIONS(2842),
    [anon_sym_ATdynamic] = ACTIONS(2842),
    [anon_sym_ATtry] = ACTIONS(2842),
    [anon_sym___try] = ACTIONS(2840),
    [anon_sym_ATthrow] = ACTIONS(2842),
    [anon_sym_ATselector] = ACTIONS(2842),
    [anon_sym_ATavailable] = ACTIONS(2842),
    [anon_sym___builtin_available] = ACTIONS(2840),
    [anon_sym_va_arg] = ACTIONS(2840),
    [anon_sym_ATencode] = ACTIONS(2842),
    [anon_sym_ATsynchronized] = ACTIONS(2842),
    [anon_sym_BOOL] = ACTIONS(2840),
    [anon_sym_IMP] = ACTIONS(2840),
    [anon_sym_SEL] = ACTIONS(2840),
    [anon_sym_Class] = ACTIONS(2840),
    [anon_sym_id] = ACTIONS(2840),
  },
  [STATE(353)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token2] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [aux_sym_preproc_else_token1] = ACTIONS(2844),
    [aux_sym_preproc_elif_token1] = ACTIONS(2844),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATsynthesize] = ACTIONS(2846),
    [anon_sym_ATdynamic] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(354)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token2] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [aux_sym_preproc_else_token1] = ACTIONS(2844),
    [aux_sym_preproc_elif_token1] = ACTIONS(2844),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATsynthesize] = ACTIONS(2846),
    [anon_sym_ATdynamic] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(355)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token2] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [aux_sym_preproc_else_token1] = ACTIONS(2844),
    [aux_sym_preproc_elif_token1] = ACTIONS(2844),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATsynthesize] = ACTIONS(2846),
    [anon_sym_ATdynamic] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(356)] = {
    [sym_identifier] = ACTIONS(2848),
    [aux_sym_preproc_include_token1] = ACTIONS(2848),
    [aux_sym_preproc_include_token2] = ACTIONS(2848),
    [aux_sym_preproc_def_token1] = ACTIONS(2848),
    [aux_sym_preproc_if_token1] = ACTIONS(2848),
    [aux_sym_preproc_if_token2] = ACTIONS(2848),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2848),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2848),
    [aux_sym_preproc_else_token1] = ACTIONS(2848),
    [aux_sym_preproc_elif_token1] = ACTIONS(2848),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2848),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2848),
    [sym_preproc_directive] = ACTIONS(2848),
    [anon_sym_LPAREN2] = ACTIONS(2850),
    [anon_sym_BANG] = ACTIONS(2850),
    [anon_sym_TILDE] = ACTIONS(2850),
    [anon_sym_DASH] = ACTIONS(2848),
    [anon_sym_PLUS] = ACTIONS(2848),
    [anon_sym_STAR] = ACTIONS(2850),
    [anon_sym_CARET] = ACTIONS(2850),
    [anon_sym_AMP] = ACTIONS(2850),
    [anon_sym_SEMI] = ACTIONS(2850),
    [anon_sym___extension__] = ACTIONS(2848),
    [anon_sym_typedef] = ACTIONS(2848),
    [anon_sym_extern] = ACTIONS(2848),
    [anon_sym___attribute__] = ACTIONS(2848),
    [anon_sym___attribute] = ACTIONS(2848),
    [anon_sym_noreturn] = ACTIONS(2848),
    [anon_sym_LBRACK] = ACTIONS(2850),
    [anon_sym___declspec] = ACTIONS(2848),
    [anon_sym___cdecl] = ACTIONS(2848),
    [anon_sym___clrcall] = ACTIONS(2848),
    [anon_sym___stdcall] = ACTIONS(2848),
    [anon_sym___fastcall] = ACTIONS(2848),
    [anon_sym___thiscall] = ACTIONS(2848),
    [anon_sym___vectorcall] = ACTIONS(2848),
    [anon_sym_LBRACE] = ACTIONS(2850),
    [anon_sym_signed] = ACTIONS(2848),
    [anon_sym_unsigned] = ACTIONS(2848),
    [anon_sym_long] = ACTIONS(2848),
    [anon_sym_short] = ACTIONS(2848),
    [anon_sym_static] = ACTIONS(2848),
    [anon_sym_ATautoreleasepool] = ACTIONS(2850),
    [anon_sym_auto] = ACTIONS(2848),
    [anon_sym_register] = ACTIONS(2848),
    [anon_sym_inline] = ACTIONS(2848),
    [anon_sym___inline] = ACTIONS(2848),
    [anon_sym___inline__] = ACTIONS(2848),
    [anon_sym___forceinline] = ACTIONS(2848),
    [anon_sym_thread_local] = ACTIONS(2848),
    [anon_sym___thread] = ACTIONS(2848),
    [anon_sym_CG_EXTERN] = ACTIONS(2848),
    [anon_sym_CG_INLINE] = ACTIONS(2848),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2848),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2848),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2848),
    [anon_sym_IBOutlet] = ACTIONS(2848),
    [anon_sym_IBInspectable] = ACTIONS(2848),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2848),
    [anon_sym_NS_INLINE] = ACTIONS(2848),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2848),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2848),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2848),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2848),
    [anon_sym_const] = ACTIONS(2848),
    [anon_sym_constexpr] = ACTIONS(2848),
    [anon_sym_volatile] = ACTIONS(2848),
    [anon_sym_restrict] = ACTIONS(2848),
    [anon_sym___restrict__] = ACTIONS(2848),
    [anon_sym__Atomic] = ACTIONS(2848),
    [anon_sym__Noreturn] = ACTIONS(2848),
    [anon_sym__Nonnull] = ACTIONS(2848),
    [anon_sym_nullable] = ACTIONS(2848),
    [anon_sym__Complex] = ACTIONS(2848),
    [anon_sym__Nullable] = ACTIONS(2848),
    [anon_sym__Nullable_result] = ACTIONS(2848),
    [anon_sym__Null_unspecified] = ACTIONS(2848),
    [anon_sym___autoreleasing] = ACTIONS(2848),
    [anon_sym___block] = ACTIONS(2848),
    [anon_sym___bridge] = ACTIONS(2848),
    [anon_sym___bridge_retained] = ACTIONS(2848),
    [anon_sym___bridge_transfer] = ACTIONS(2848),
    [anon_sym___complex] = ACTIONS(2848),
    [anon_sym___const] = ACTIONS(2848),
    [anon_sym___imag] = ACTIONS(2848),
    [anon_sym___kindof] = ACTIONS(2848),
    [anon_sym___nonnull] = ACTIONS(2848),
    [anon_sym___nullable] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2848),
    [anon_sym___real] = ACTIONS(2848),
    [anon_sym___strong] = ACTIONS(2848),
    [anon_sym___unsafe_unretained] = ACTIONS(2848),
    [anon_sym___unused] = ACTIONS(2848),
    [anon_sym___weak] = ACTIONS(2848),
    [anon_sym_alignas] = ACTIONS(2848),
    [anon_sym__Alignas] = ACTIONS(2848),
    [sym_primitive_type] = ACTIONS(2848),
    [anon_sym_enum] = ACTIONS(2848),
    [anon_sym_struct] = ACTIONS(2848),
    [anon_sym_union] = ACTIONS(2848),
    [anon_sym_if] = ACTIONS(2848),
    [anon_sym_switch] = ACTIONS(2848),
    [anon_sym_case] = ACTIONS(2848),
    [anon_sym_default] = ACTIONS(2848),
    [anon_sym_while] = ACTIONS(2848),
    [anon_sym_do] = ACTIONS(2848),
    [anon_sym_for] = ACTIONS(2848),
    [anon_sym_in] = ACTIONS(2848),
    [anon_sym_return] = ACTIONS(2848),
    [anon_sym_break] = ACTIONS(2848),
    [anon_sym_continue] = ACTIONS(2848),
    [anon_sym_goto] = ACTIONS(2848),
    [anon_sym_DASH_DASH] = ACTIONS(2850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2850),
    [anon_sym_sizeof] = ACTIONS(2848),
    [anon_sym___alignof__] = ACTIONS(2848),
    [anon_sym___alignof] = ACTIONS(2848),
    [anon_sym__alignof] = ACTIONS(2848),
    [anon_sym_alignof] = ACTIONS(2848),
    [anon_sym__Alignof] = ACTIONS(2848),
    [anon_sym_offsetof] = ACTIONS(2848),
    [anon_sym__Generic] = ACTIONS(2848),
    [anon_sym_asm] = ACTIONS(2848),
    [anon_sym___asm__] = ACTIONS(2848),
    [anon_sym___asm] = ACTIONS(2848),
    [sym_number_literal] = ACTIONS(2850),
    [anon_sym_L_SQUOTE] = ACTIONS(2850),
    [anon_sym_u_SQUOTE] = ACTIONS(2850),
    [anon_sym_U_SQUOTE] = ACTIONS(2850),
    [anon_sym_u8_SQUOTE] = ACTIONS(2850),
    [anon_sym_SQUOTE] = ACTIONS(2850),
    [anon_sym_AT] = ACTIONS(2848),
    [anon_sym_DQUOTE] = ACTIONS(2850),
    [anon_sym_L_DQUOTE] = ACTIONS(2850),
    [anon_sym_u_DQUOTE] = ACTIONS(2850),
    [anon_sym_U_DQUOTE] = ACTIONS(2850),
    [anon_sym_u8_DQUOTE] = ACTIONS(2850),
    [sym_true] = ACTIONS(2848),
    [sym_false] = ACTIONS(2848),
    [anon_sym_NULL] = ACTIONS(2848),
    [anon_sym_nullptr] = ACTIONS(2848),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2848),
    [anon_sym___typeof] = ACTIONS(2848),
    [anon_sym_typeof] = ACTIONS(2848),
    [anon_sym_ATimport] = ACTIONS(2850),
    [aux_sym_preproc_undef_token1] = ACTIONS(2848),
    [anon_sym_POUND] = ACTIONS(2848),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2848),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2848),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2848),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2848),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2848),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2848),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2848),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_API_AVAILABLE] = ACTIONS(2848),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_API_DEPRECATED] = ACTIONS(2848),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2848),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym___deprecated_msg] = ACTIONS(2848),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2848),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2848),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2848),
    [anon_sym_ATprotocol] = ACTIONS(2850),
    [anon_sym_ATinterface] = ACTIONS(2850),
    [anon_sym_ATimplementation] = ACTIONS(2850),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2850),
    [anon_sym_ATsynthesize] = ACTIONS(2850),
    [anon_sym_ATdynamic] = ACTIONS(2850),
    [anon_sym_ATtry] = ACTIONS(2850),
    [anon_sym___try] = ACTIONS(2848),
    [anon_sym_ATthrow] = ACTIONS(2850),
    [anon_sym_ATselector] = ACTIONS(2850),
    [anon_sym_ATavailable] = ACTIONS(2850),
    [anon_sym___builtin_available] = ACTIONS(2848),
    [anon_sym_va_arg] = ACTIONS(2848),
    [anon_sym_ATencode] = ACTIONS(2850),
    [anon_sym_ATsynchronized] = ACTIONS(2850),
    [anon_sym_BOOL] = ACTIONS(2848),
    [anon_sym_IMP] = ACTIONS(2848),
    [anon_sym_SEL] = ACTIONS(2848),
    [anon_sym_Class] = ACTIONS(2848),
    [anon_sym_id] = ACTIONS(2848),
  },
  [STATE(357)] = {
    [sym_identifier] = ACTIONS(2852),
    [aux_sym_preproc_include_token1] = ACTIONS(2852),
    [aux_sym_preproc_include_token2] = ACTIONS(2852),
    [aux_sym_preproc_def_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token2] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2852),
    [aux_sym_preproc_else_token1] = ACTIONS(2852),
    [aux_sym_preproc_elif_token1] = ACTIONS(2852),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2852),
    [sym_preproc_directive] = ACTIONS(2852),
    [anon_sym_LPAREN2] = ACTIONS(2854),
    [anon_sym_BANG] = ACTIONS(2854),
    [anon_sym_TILDE] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2852),
    [anon_sym_PLUS] = ACTIONS(2852),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_SEMI] = ACTIONS(2854),
    [anon_sym___extension__] = ACTIONS(2852),
    [anon_sym_typedef] = ACTIONS(2852),
    [anon_sym_extern] = ACTIONS(2852),
    [anon_sym___attribute__] = ACTIONS(2852),
    [anon_sym___attribute] = ACTIONS(2852),
    [anon_sym_noreturn] = ACTIONS(2852),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym___declspec] = ACTIONS(2852),
    [anon_sym___cdecl] = ACTIONS(2852),
    [anon_sym___clrcall] = ACTIONS(2852),
    [anon_sym___stdcall] = ACTIONS(2852),
    [anon_sym___fastcall] = ACTIONS(2852),
    [anon_sym___thiscall] = ACTIONS(2852),
    [anon_sym___vectorcall] = ACTIONS(2852),
    [anon_sym_LBRACE] = ACTIONS(2854),
    [anon_sym_signed] = ACTIONS(2852),
    [anon_sym_unsigned] = ACTIONS(2852),
    [anon_sym_long] = ACTIONS(2852),
    [anon_sym_short] = ACTIONS(2852),
    [anon_sym_static] = ACTIONS(2852),
    [anon_sym_ATautoreleasepool] = ACTIONS(2854),
    [anon_sym_auto] = ACTIONS(2852),
    [anon_sym_register] = ACTIONS(2852),
    [anon_sym_inline] = ACTIONS(2852),
    [anon_sym___inline] = ACTIONS(2852),
    [anon_sym___inline__] = ACTIONS(2852),
    [anon_sym___forceinline] = ACTIONS(2852),
    [anon_sym_thread_local] = ACTIONS(2852),
    [anon_sym___thread] = ACTIONS(2852),
    [anon_sym_CG_EXTERN] = ACTIONS(2852),
    [anon_sym_CG_INLINE] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2852),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2852),
    [anon_sym_IBOutlet] = ACTIONS(2852),
    [anon_sym_IBInspectable] = ACTIONS(2852),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2852),
    [anon_sym_NS_INLINE] = ACTIONS(2852),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2852),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2852),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2852),
    [anon_sym_const] = ACTIONS(2852),
    [anon_sym_constexpr] = ACTIONS(2852),
    [anon_sym_volatile] = ACTIONS(2852),
    [anon_sym_restrict] = ACTIONS(2852),
    [anon_sym___restrict__] = ACTIONS(2852),
    [anon_sym__Atomic] = ACTIONS(2852),
    [anon_sym__Noreturn] = ACTIONS(2852),
    [anon_sym__Nonnull] = ACTIONS(2852),
    [anon_sym_nullable] = ACTIONS(2852),
    [anon_sym__Complex] = ACTIONS(2852),
    [anon_sym__Nullable] = ACTIONS(2852),
    [anon_sym__Nullable_result] = ACTIONS(2852),
    [anon_sym__Null_unspecified] = ACTIONS(2852),
    [anon_sym___autoreleasing] = ACTIONS(2852),
    [anon_sym___block] = ACTIONS(2852),
    [anon_sym___bridge] = ACTIONS(2852),
    [anon_sym___bridge_retained] = ACTIONS(2852),
    [anon_sym___bridge_transfer] = ACTIONS(2852),
    [anon_sym___complex] = ACTIONS(2852),
    [anon_sym___const] = ACTIONS(2852),
    [anon_sym___imag] = ACTIONS(2852),
    [anon_sym___kindof] = ACTIONS(2852),
    [anon_sym___nonnull] = ACTIONS(2852),
    [anon_sym___nullable] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2852),
    [anon_sym___real] = ACTIONS(2852),
    [anon_sym___strong] = ACTIONS(2852),
    [anon_sym___unsafe_unretained] = ACTIONS(2852),
    [anon_sym___unused] = ACTIONS(2852),
    [anon_sym___weak] = ACTIONS(2852),
    [anon_sym_alignas] = ACTIONS(2852),
    [anon_sym__Alignas] = ACTIONS(2852),
    [sym_primitive_type] = ACTIONS(2852),
    [anon_sym_enum] = ACTIONS(2852),
    [anon_sym_struct] = ACTIONS(2852),
    [anon_sym_union] = ACTIONS(2852),
    [anon_sym_if] = ACTIONS(2852),
    [anon_sym_switch] = ACTIONS(2852),
    [anon_sym_case] = ACTIONS(2852),
    [anon_sym_default] = ACTIONS(2852),
    [anon_sym_while] = ACTIONS(2852),
    [anon_sym_do] = ACTIONS(2852),
    [anon_sym_for] = ACTIONS(2852),
    [anon_sym_in] = ACTIONS(2852),
    [anon_sym_return] = ACTIONS(2852),
    [anon_sym_break] = ACTIONS(2852),
    [anon_sym_continue] = ACTIONS(2852),
    [anon_sym_goto] = ACTIONS(2852),
    [anon_sym_DASH_DASH] = ACTIONS(2854),
    [anon_sym_PLUS_PLUS] = ACTIONS(2854),
    [anon_sym_sizeof] = ACTIONS(2852),
    [anon_sym___alignof__] = ACTIONS(2852),
    [anon_sym___alignof] = ACTIONS(2852),
    [anon_sym__alignof] = ACTIONS(2852),
    [anon_sym_alignof] = ACTIONS(2852),
    [anon_sym__Alignof] = ACTIONS(2852),
    [anon_sym_offsetof] = ACTIONS(2852),
    [anon_sym__Generic] = ACTIONS(2852),
    [anon_sym_asm] = ACTIONS(2852),
    [anon_sym___asm__] = ACTIONS(2852),
    [anon_sym___asm] = ACTIONS(2852),
    [sym_number_literal] = ACTIONS(2854),
    [anon_sym_L_SQUOTE] = ACTIONS(2854),
    [anon_sym_u_SQUOTE] = ACTIONS(2854),
    [anon_sym_U_SQUOTE] = ACTIONS(2854),
    [anon_sym_u8_SQUOTE] = ACTIONS(2854),
    [anon_sym_SQUOTE] = ACTIONS(2854),
    [anon_sym_AT] = ACTIONS(2852),
    [anon_sym_DQUOTE] = ACTIONS(2854),
    [anon_sym_L_DQUOTE] = ACTIONS(2854),
    [anon_sym_u_DQUOTE] = ACTIONS(2854),
    [anon_sym_U_DQUOTE] = ACTIONS(2854),
    [anon_sym_u8_DQUOTE] = ACTIONS(2854),
    [sym_true] = ACTIONS(2852),
    [sym_false] = ACTIONS(2852),
    [anon_sym_NULL] = ACTIONS(2852),
    [anon_sym_nullptr] = ACTIONS(2852),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2852),
    [anon_sym___typeof] = ACTIONS(2852),
    [anon_sym_typeof] = ACTIONS(2852),
    [anon_sym_ATimport] = ACTIONS(2854),
    [aux_sym_preproc_undef_token1] = ACTIONS(2852),
    [anon_sym_POUND] = ACTIONS(2852),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2852),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2852),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2852),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_API_AVAILABLE] = ACTIONS(2852),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_API_DEPRECATED] = ACTIONS(2852),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym___deprecated_msg] = ACTIONS(2852),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2852),
    [anon_sym_ATprotocol] = ACTIONS(2854),
    [anon_sym_ATinterface] = ACTIONS(2854),
    [anon_sym_ATimplementation] = ACTIONS(2854),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2854),
    [anon_sym_ATsynthesize] = ACTIONS(2854),
    [anon_sym_ATdynamic] = ACTIONS(2854),
    [anon_sym_ATtry] = ACTIONS(2854),
    [anon_sym___try] = ACTIONS(2852),
    [anon_sym_ATthrow] = ACTIONS(2854),
    [anon_sym_ATselector] = ACTIONS(2854),
    [anon_sym_ATavailable] = ACTIONS(2854),
    [anon_sym___builtin_available] = ACTIONS(2852),
    [anon_sym_va_arg] = ACTIONS(2852),
    [anon_sym_ATencode] = ACTIONS(2854),
    [anon_sym_ATsynchronized] = ACTIONS(2854),
    [anon_sym_BOOL] = ACTIONS(2852),
    [anon_sym_IMP] = ACTIONS(2852),
    [anon_sym_SEL] = ACTIONS(2852),
    [anon_sym_Class] = ACTIONS(2852),
    [anon_sym_id] = ACTIONS(2852),
  },
  [STATE(358)] = {
    [sym_identifier] = ACTIONS(2856),
    [aux_sym_preproc_include_token1] = ACTIONS(2856),
    [aux_sym_preproc_include_token2] = ACTIONS(2856),
    [aux_sym_preproc_def_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token2] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2856),
    [aux_sym_preproc_else_token1] = ACTIONS(2856),
    [aux_sym_preproc_elif_token1] = ACTIONS(2856),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2856),
    [sym_preproc_directive] = ACTIONS(2856),
    [anon_sym_LPAREN2] = ACTIONS(2858),
    [anon_sym_BANG] = ACTIONS(2858),
    [anon_sym_TILDE] = ACTIONS(2858),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_STAR] = ACTIONS(2858),
    [anon_sym_CARET] = ACTIONS(2858),
    [anon_sym_AMP] = ACTIONS(2858),
    [anon_sym_SEMI] = ACTIONS(2858),
    [anon_sym___extension__] = ACTIONS(2856),
    [anon_sym_typedef] = ACTIONS(2856),
    [anon_sym_extern] = ACTIONS(2856),
    [anon_sym___attribute__] = ACTIONS(2856),
    [anon_sym___attribute] = ACTIONS(2856),
    [anon_sym_noreturn] = ACTIONS(2856),
    [anon_sym_LBRACK] = ACTIONS(2858),
    [anon_sym___declspec] = ACTIONS(2856),
    [anon_sym___cdecl] = ACTIONS(2856),
    [anon_sym___clrcall] = ACTIONS(2856),
    [anon_sym___stdcall] = ACTIONS(2856),
    [anon_sym___fastcall] = ACTIONS(2856),
    [anon_sym___thiscall] = ACTIONS(2856),
    [anon_sym___vectorcall] = ACTIONS(2856),
    [anon_sym_LBRACE] = ACTIONS(2858),
    [anon_sym_signed] = ACTIONS(2856),
    [anon_sym_unsigned] = ACTIONS(2856),
    [anon_sym_long] = ACTIONS(2856),
    [anon_sym_short] = ACTIONS(2856),
    [anon_sym_static] = ACTIONS(2856),
    [anon_sym_ATautoreleasepool] = ACTIONS(2858),
    [anon_sym_auto] = ACTIONS(2856),
    [anon_sym_register] = ACTIONS(2856),
    [anon_sym_inline] = ACTIONS(2856),
    [anon_sym___inline] = ACTIONS(2856),
    [anon_sym___inline__] = ACTIONS(2856),
    [anon_sym___forceinline] = ACTIONS(2856),
    [anon_sym_thread_local] = ACTIONS(2856),
    [anon_sym___thread] = ACTIONS(2856),
    [anon_sym_CG_EXTERN] = ACTIONS(2856),
    [anon_sym_CG_INLINE] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2856),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2856),
    [anon_sym_IBOutlet] = ACTIONS(2856),
    [anon_sym_IBInspectable] = ACTIONS(2856),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2856),
    [anon_sym_NS_INLINE] = ACTIONS(2856),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2856),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2856),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2856),
    [anon_sym_const] = ACTIONS(2856),
    [anon_sym_constexpr] = ACTIONS(2856),
    [anon_sym_volatile] = ACTIONS(2856),
    [anon_sym_restrict] = ACTIONS(2856),
    [anon_sym___restrict__] = ACTIONS(2856),
    [anon_sym__Atomic] = ACTIONS(2856),
    [anon_sym__Noreturn] = ACTIONS(2856),
    [anon_sym__Nonnull] = ACTIONS(2856),
    [anon_sym_nullable] = ACTIONS(2856),
    [anon_sym__Complex] = ACTIONS(2856),
    [anon_sym__Nullable] = ACTIONS(2856),
    [anon_sym__Nullable_result] = ACTIONS(2856),
    [anon_sym__Null_unspecified] = ACTIONS(2856),
    [anon_sym___autoreleasing] = ACTIONS(2856),
    [anon_sym___block] = ACTIONS(2856),
    [anon_sym___bridge] = ACTIONS(2856),
    [anon_sym___bridge_retained] = ACTIONS(2856),
    [anon_sym___bridge_transfer] = ACTIONS(2856),
    [anon_sym___complex] = ACTIONS(2856),
    [anon_sym___const] = ACTIONS(2856),
    [anon_sym___imag] = ACTIONS(2856),
    [anon_sym___kindof] = ACTIONS(2856),
    [anon_sym___nonnull] = ACTIONS(2856),
    [anon_sym___nullable] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2856),
    [anon_sym___real] = ACTIONS(2856),
    [anon_sym___strong] = ACTIONS(2856),
    [anon_sym___unsafe_unretained] = ACTIONS(2856),
    [anon_sym___unused] = ACTIONS(2856),
    [anon_sym___weak] = ACTIONS(2856),
    [anon_sym_alignas] = ACTIONS(2856),
    [anon_sym__Alignas] = ACTIONS(2856),
    [sym_primitive_type] = ACTIONS(2856),
    [anon_sym_enum] = ACTIONS(2856),
    [anon_sym_struct] = ACTIONS(2856),
    [anon_sym_union] = ACTIONS(2856),
    [anon_sym_if] = ACTIONS(2856),
    [anon_sym_switch] = ACTIONS(2856),
    [anon_sym_case] = ACTIONS(2856),
    [anon_sym_default] = ACTIONS(2856),
    [anon_sym_while] = ACTIONS(2856),
    [anon_sym_do] = ACTIONS(2856),
    [anon_sym_for] = ACTIONS(2856),
    [anon_sym_in] = ACTIONS(2856),
    [anon_sym_return] = ACTIONS(2856),
    [anon_sym_break] = ACTIONS(2856),
    [anon_sym_continue] = ACTIONS(2856),
    [anon_sym_goto] = ACTIONS(2856),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_sizeof] = ACTIONS(2856),
    [anon_sym___alignof__] = ACTIONS(2856),
    [anon_sym___alignof] = ACTIONS(2856),
    [anon_sym__alignof] = ACTIONS(2856),
    [anon_sym_alignof] = ACTIONS(2856),
    [anon_sym__Alignof] = ACTIONS(2856),
    [anon_sym_offsetof] = ACTIONS(2856),
    [anon_sym__Generic] = ACTIONS(2856),
    [anon_sym_asm] = ACTIONS(2856),
    [anon_sym___asm__] = ACTIONS(2856),
    [anon_sym___asm] = ACTIONS(2856),
    [sym_number_literal] = ACTIONS(2858),
    [anon_sym_L_SQUOTE] = ACTIONS(2858),
    [anon_sym_u_SQUOTE] = ACTIONS(2858),
    [anon_sym_U_SQUOTE] = ACTIONS(2858),
    [anon_sym_u8_SQUOTE] = ACTIONS(2858),
    [anon_sym_SQUOTE] = ACTIONS(2858),
    [anon_sym_AT] = ACTIONS(2856),
    [anon_sym_DQUOTE] = ACTIONS(2858),
    [anon_sym_L_DQUOTE] = ACTIONS(2858),
    [anon_sym_u_DQUOTE] = ACTIONS(2858),
    [anon_sym_U_DQUOTE] = ACTIONS(2858),
    [anon_sym_u8_DQUOTE] = ACTIONS(2858),
    [sym_true] = ACTIONS(2856),
    [sym_false] = ACTIONS(2856),
    [anon_sym_NULL] = ACTIONS(2856),
    [anon_sym_nullptr] = ACTIONS(2856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2856),
    [anon_sym___typeof] = ACTIONS(2856),
    [anon_sym_typeof] = ACTIONS(2856),
    [anon_sym_ATimport] = ACTIONS(2858),
    [aux_sym_preproc_undef_token1] = ACTIONS(2856),
    [anon_sym_POUND] = ACTIONS(2856),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2856),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2856),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2856),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_API_AVAILABLE] = ACTIONS(2856),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_API_DEPRECATED] = ACTIONS(2856),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym___deprecated_msg] = ACTIONS(2856),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2856),
    [anon_sym_ATprotocol] = ACTIONS(2858),
    [anon_sym_ATinterface] = ACTIONS(2858),
    [anon_sym_ATimplementation] = ACTIONS(2858),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2858),
    [anon_sym_ATsynthesize] = ACTIONS(2858),
    [anon_sym_ATdynamic] = ACTIONS(2858),
    [anon_sym_ATtry] = ACTIONS(2858),
    [anon_sym___try] = ACTIONS(2856),
    [anon_sym_ATthrow] = ACTIONS(2858),
    [anon_sym_ATselector] = ACTIONS(2858),
    [anon_sym_ATavailable] = ACTIONS(2858),
    [anon_sym___builtin_available] = ACTIONS(2856),
    [anon_sym_va_arg] = ACTIONS(2856),
    [anon_sym_ATencode] = ACTIONS(2858),
    [anon_sym_ATsynchronized] = ACTIONS(2858),
    [anon_sym_BOOL] = ACTIONS(2856),
    [anon_sym_IMP] = ACTIONS(2856),
    [anon_sym_SEL] = ACTIONS(2856),
    [anon_sym_Class] = ACTIONS(2856),
    [anon_sym_id] = ACTIONS(2856),
  },
  [STATE(359)] = {
    [sym_identifier] = ACTIONS(2860),
    [aux_sym_preproc_include_token1] = ACTIONS(2860),
    [aux_sym_preproc_include_token2] = ACTIONS(2860),
    [aux_sym_preproc_def_token1] = ACTIONS(2860),
    [aux_sym_preproc_if_token1] = ACTIONS(2860),
    [aux_sym_preproc_if_token2] = ACTIONS(2860),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2860),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2860),
    [aux_sym_preproc_else_token1] = ACTIONS(2860),
    [aux_sym_preproc_elif_token1] = ACTIONS(2860),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2860),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2860),
    [sym_preproc_directive] = ACTIONS(2860),
    [anon_sym_LPAREN2] = ACTIONS(2862),
    [anon_sym_BANG] = ACTIONS(2862),
    [anon_sym_TILDE] = ACTIONS(2862),
    [anon_sym_DASH] = ACTIONS(2860),
    [anon_sym_PLUS] = ACTIONS(2860),
    [anon_sym_STAR] = ACTIONS(2862),
    [anon_sym_CARET] = ACTIONS(2862),
    [anon_sym_AMP] = ACTIONS(2862),
    [anon_sym_SEMI] = ACTIONS(2862),
    [anon_sym___extension__] = ACTIONS(2860),
    [anon_sym_typedef] = ACTIONS(2860),
    [anon_sym_extern] = ACTIONS(2860),
    [anon_sym___attribute__] = ACTIONS(2860),
    [anon_sym___attribute] = ACTIONS(2860),
    [anon_sym_noreturn] = ACTIONS(2860),
    [anon_sym_LBRACK] = ACTIONS(2862),
    [anon_sym___declspec] = ACTIONS(2860),
    [anon_sym___cdecl] = ACTIONS(2860),
    [anon_sym___clrcall] = ACTIONS(2860),
    [anon_sym___stdcall] = ACTIONS(2860),
    [anon_sym___fastcall] = ACTIONS(2860),
    [anon_sym___thiscall] = ACTIONS(2860),
    [anon_sym___vectorcall] = ACTIONS(2860),
    [anon_sym_LBRACE] = ACTIONS(2862),
    [anon_sym_signed] = ACTIONS(2860),
    [anon_sym_unsigned] = ACTIONS(2860),
    [anon_sym_long] = ACTIONS(2860),
    [anon_sym_short] = ACTIONS(2860),
    [anon_sym_static] = ACTIONS(2860),
    [anon_sym_ATautoreleasepool] = ACTIONS(2862),
    [anon_sym_auto] = ACTIONS(2860),
    [anon_sym_register] = ACTIONS(2860),
    [anon_sym_inline] = ACTIONS(2860),
    [anon_sym___inline] = ACTIONS(2860),
    [anon_sym___inline__] = ACTIONS(2860),
    [anon_sym___forceinline] = ACTIONS(2860),
    [anon_sym_thread_local] = ACTIONS(2860),
    [anon_sym___thread] = ACTIONS(2860),
    [anon_sym_CG_EXTERN] = ACTIONS(2860),
    [anon_sym_CG_INLINE] = ACTIONS(2860),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2860),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2860),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2860),
    [anon_sym_IBOutlet] = ACTIONS(2860),
    [anon_sym_IBInspectable] = ACTIONS(2860),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2860),
    [anon_sym_NS_INLINE] = ACTIONS(2860),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2860),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2860),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2860),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2860),
    [anon_sym_const] = ACTIONS(2860),
    [anon_sym_constexpr] = ACTIONS(2860),
    [anon_sym_volatile] = ACTIONS(2860),
    [anon_sym_restrict] = ACTIONS(2860),
    [anon_sym___restrict__] = ACTIONS(2860),
    [anon_sym__Atomic] = ACTIONS(2860),
    [anon_sym__Noreturn] = ACTIONS(2860),
    [anon_sym__Nonnull] = ACTIONS(2860),
    [anon_sym_nullable] = ACTIONS(2860),
    [anon_sym__Complex] = ACTIONS(2860),
    [anon_sym__Nullable] = ACTIONS(2860),
    [anon_sym__Nullable_result] = ACTIONS(2860),
    [anon_sym__Null_unspecified] = ACTIONS(2860),
    [anon_sym___autoreleasing] = ACTIONS(2860),
    [anon_sym___block] = ACTIONS(2860),
    [anon_sym___bridge] = ACTIONS(2860),
    [anon_sym___bridge_retained] = ACTIONS(2860),
    [anon_sym___bridge_transfer] = ACTIONS(2860),
    [anon_sym___complex] = ACTIONS(2860),
    [anon_sym___const] = ACTIONS(2860),
    [anon_sym___imag] = ACTIONS(2860),
    [anon_sym___kindof] = ACTIONS(2860),
    [anon_sym___nonnull] = ACTIONS(2860),
    [anon_sym___nullable] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2860),
    [anon_sym___real] = ACTIONS(2860),
    [anon_sym___strong] = ACTIONS(2860),
    [anon_sym___unsafe_unretained] = ACTIONS(2860),
    [anon_sym___unused] = ACTIONS(2860),
    [anon_sym___weak] = ACTIONS(2860),
    [anon_sym_alignas] = ACTIONS(2860),
    [anon_sym__Alignas] = ACTIONS(2860),
    [sym_primitive_type] = ACTIONS(2860),
    [anon_sym_enum] = ACTIONS(2860),
    [anon_sym_struct] = ACTIONS(2860),
    [anon_sym_union] = ACTIONS(2860),
    [anon_sym_if] = ACTIONS(2860),
    [anon_sym_switch] = ACTIONS(2860),
    [anon_sym_case] = ACTIONS(2860),
    [anon_sym_default] = ACTIONS(2860),
    [anon_sym_while] = ACTIONS(2860),
    [anon_sym_do] = ACTIONS(2860),
    [anon_sym_for] = ACTIONS(2860),
    [anon_sym_in] = ACTIONS(2860),
    [anon_sym_return] = ACTIONS(2860),
    [anon_sym_break] = ACTIONS(2860),
    [anon_sym_continue] = ACTIONS(2860),
    [anon_sym_goto] = ACTIONS(2860),
    [anon_sym_DASH_DASH] = ACTIONS(2862),
    [anon_sym_PLUS_PLUS] = ACTIONS(2862),
    [anon_sym_sizeof] = ACTIONS(2860),
    [anon_sym___alignof__] = ACTIONS(2860),
    [anon_sym___alignof] = ACTIONS(2860),
    [anon_sym__alignof] = ACTIONS(2860),
    [anon_sym_alignof] = ACTIONS(2860),
    [anon_sym__Alignof] = ACTIONS(2860),
    [anon_sym_offsetof] = ACTIONS(2860),
    [anon_sym__Generic] = ACTIONS(2860),
    [anon_sym_asm] = ACTIONS(2860),
    [anon_sym___asm__] = ACTIONS(2860),
    [anon_sym___asm] = ACTIONS(2860),
    [sym_number_literal] = ACTIONS(2862),
    [anon_sym_L_SQUOTE] = ACTIONS(2862),
    [anon_sym_u_SQUOTE] = ACTIONS(2862),
    [anon_sym_U_SQUOTE] = ACTIONS(2862),
    [anon_sym_u8_SQUOTE] = ACTIONS(2862),
    [anon_sym_SQUOTE] = ACTIONS(2862),
    [anon_sym_AT] = ACTIONS(2860),
    [anon_sym_DQUOTE] = ACTIONS(2862),
    [anon_sym_L_DQUOTE] = ACTIONS(2862),
    [anon_sym_u_DQUOTE] = ACTIONS(2862),
    [anon_sym_U_DQUOTE] = ACTIONS(2862),
    [anon_sym_u8_DQUOTE] = ACTIONS(2862),
    [sym_true] = ACTIONS(2860),
    [sym_false] = ACTIONS(2860),
    [anon_sym_NULL] = ACTIONS(2860),
    [anon_sym_nullptr] = ACTIONS(2860),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2860),
    [anon_sym___typeof] = ACTIONS(2860),
    [anon_sym_typeof] = ACTIONS(2860),
    [anon_sym_ATimport] = ACTIONS(2862),
    [aux_sym_preproc_undef_token1] = ACTIONS(2860),
    [anon_sym_POUND] = ACTIONS(2860),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2860),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2860),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2860),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2860),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2860),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2860),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2860),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_API_AVAILABLE] = ACTIONS(2860),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_API_DEPRECATED] = ACTIONS(2860),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2860),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym___deprecated_msg] = ACTIONS(2860),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2860),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2860),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2860),
    [anon_sym_ATprotocol] = ACTIONS(2862),
    [anon_sym_ATinterface] = ACTIONS(2862),
    [anon_sym_ATimplementation] = ACTIONS(2862),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2862),
    [anon_sym_ATsynthesize] = ACTIONS(2862),
    [anon_sym_ATdynamic] = ACTIONS(2862),
    [anon_sym_ATtry] = ACTIONS(2862),
    [anon_sym___try] = ACTIONS(2860),
    [anon_sym_ATthrow] = ACTIONS(2862),
    [anon_sym_ATselector] = ACTIONS(2862),
    [anon_sym_ATavailable] = ACTIONS(2862),
    [anon_sym___builtin_available] = ACTIONS(2860),
    [anon_sym_va_arg] = ACTIONS(2860),
    [anon_sym_ATencode] = ACTIONS(2862),
    [anon_sym_ATsynchronized] = ACTIONS(2862),
    [anon_sym_BOOL] = ACTIONS(2860),
    [anon_sym_IMP] = ACTIONS(2860),
    [anon_sym_SEL] = ACTIONS(2860),
    [anon_sym_Class] = ACTIONS(2860),
    [anon_sym_id] = ACTIONS(2860),
  },
  [STATE(360)] = {
    [sym_identifier] = ACTIONS(2864),
    [aux_sym_preproc_include_token1] = ACTIONS(2864),
    [aux_sym_preproc_include_token2] = ACTIONS(2864),
    [aux_sym_preproc_def_token1] = ACTIONS(2864),
    [aux_sym_preproc_if_token1] = ACTIONS(2864),
    [aux_sym_preproc_if_token2] = ACTIONS(2864),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2864),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2864),
    [aux_sym_preproc_else_token1] = ACTIONS(2864),
    [aux_sym_preproc_elif_token1] = ACTIONS(2864),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2864),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2864),
    [sym_preproc_directive] = ACTIONS(2864),
    [anon_sym_LPAREN2] = ACTIONS(2866),
    [anon_sym_BANG] = ACTIONS(2866),
    [anon_sym_TILDE] = ACTIONS(2866),
    [anon_sym_DASH] = ACTIONS(2864),
    [anon_sym_PLUS] = ACTIONS(2864),
    [anon_sym_STAR] = ACTIONS(2866),
    [anon_sym_CARET] = ACTIONS(2866),
    [anon_sym_AMP] = ACTIONS(2866),
    [anon_sym_SEMI] = ACTIONS(2866),
    [anon_sym___extension__] = ACTIONS(2864),
    [anon_sym_typedef] = ACTIONS(2864),
    [anon_sym_extern] = ACTIONS(2864),
    [anon_sym___attribute__] = ACTIONS(2864),
    [anon_sym___attribute] = ACTIONS(2864),
    [anon_sym_noreturn] = ACTIONS(2864),
    [anon_sym_LBRACK] = ACTIONS(2866),
    [anon_sym___declspec] = ACTIONS(2864),
    [anon_sym___cdecl] = ACTIONS(2864),
    [anon_sym___clrcall] = ACTIONS(2864),
    [anon_sym___stdcall] = ACTIONS(2864),
    [anon_sym___fastcall] = ACTIONS(2864),
    [anon_sym___thiscall] = ACTIONS(2864),
    [anon_sym___vectorcall] = ACTIONS(2864),
    [anon_sym_LBRACE] = ACTIONS(2866),
    [anon_sym_signed] = ACTIONS(2864),
    [anon_sym_unsigned] = ACTIONS(2864),
    [anon_sym_long] = ACTIONS(2864),
    [anon_sym_short] = ACTIONS(2864),
    [anon_sym_static] = ACTIONS(2864),
    [anon_sym_ATautoreleasepool] = ACTIONS(2866),
    [anon_sym_auto] = ACTIONS(2864),
    [anon_sym_register] = ACTIONS(2864),
    [anon_sym_inline] = ACTIONS(2864),
    [anon_sym___inline] = ACTIONS(2864),
    [anon_sym___inline__] = ACTIONS(2864),
    [anon_sym___forceinline] = ACTIONS(2864),
    [anon_sym_thread_local] = ACTIONS(2864),
    [anon_sym___thread] = ACTIONS(2864),
    [anon_sym_CG_EXTERN] = ACTIONS(2864),
    [anon_sym_CG_INLINE] = ACTIONS(2864),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2864),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2864),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2864),
    [anon_sym_IBOutlet] = ACTIONS(2864),
    [anon_sym_IBInspectable] = ACTIONS(2864),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2864),
    [anon_sym_NS_INLINE] = ACTIONS(2864),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2864),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2864),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2864),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2864),
    [anon_sym_const] = ACTIONS(2864),
    [anon_sym_constexpr] = ACTIONS(2864),
    [anon_sym_volatile] = ACTIONS(2864),
    [anon_sym_restrict] = ACTIONS(2864),
    [anon_sym___restrict__] = ACTIONS(2864),
    [anon_sym__Atomic] = ACTIONS(2864),
    [anon_sym__Noreturn] = ACTIONS(2864),
    [anon_sym__Nonnull] = ACTIONS(2864),
    [anon_sym_nullable] = ACTIONS(2864),
    [anon_sym__Complex] = ACTIONS(2864),
    [anon_sym__Nullable] = ACTIONS(2864),
    [anon_sym__Nullable_result] = ACTIONS(2864),
    [anon_sym__Null_unspecified] = ACTIONS(2864),
    [anon_sym___autoreleasing] = ACTIONS(2864),
    [anon_sym___block] = ACTIONS(2864),
    [anon_sym___bridge] = ACTIONS(2864),
    [anon_sym___bridge_retained] = ACTIONS(2864),
    [anon_sym___bridge_transfer] = ACTIONS(2864),
    [anon_sym___complex] = ACTIONS(2864),
    [anon_sym___const] = ACTIONS(2864),
    [anon_sym___imag] = ACTIONS(2864),
    [anon_sym___kindof] = ACTIONS(2864),
    [anon_sym___nonnull] = ACTIONS(2864),
    [anon_sym___nullable] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2864),
    [anon_sym___real] = ACTIONS(2864),
    [anon_sym___strong] = ACTIONS(2864),
    [anon_sym___unsafe_unretained] = ACTIONS(2864),
    [anon_sym___unused] = ACTIONS(2864),
    [anon_sym___weak] = ACTIONS(2864),
    [anon_sym_alignas] = ACTIONS(2864),
    [anon_sym__Alignas] = ACTIONS(2864),
    [sym_primitive_type] = ACTIONS(2864),
    [anon_sym_enum] = ACTIONS(2864),
    [anon_sym_struct] = ACTIONS(2864),
    [anon_sym_union] = ACTIONS(2864),
    [anon_sym_if] = ACTIONS(2864),
    [anon_sym_switch] = ACTIONS(2864),
    [anon_sym_case] = ACTIONS(2864),
    [anon_sym_default] = ACTIONS(2864),
    [anon_sym_while] = ACTIONS(2864),
    [anon_sym_do] = ACTIONS(2864),
    [anon_sym_for] = ACTIONS(2864),
    [anon_sym_in] = ACTIONS(2864),
    [anon_sym_return] = ACTIONS(2864),
    [anon_sym_break] = ACTIONS(2864),
    [anon_sym_continue] = ACTIONS(2864),
    [anon_sym_goto] = ACTIONS(2864),
    [anon_sym_DASH_DASH] = ACTIONS(2866),
    [anon_sym_PLUS_PLUS] = ACTIONS(2866),
    [anon_sym_sizeof] = ACTIONS(2864),
    [anon_sym___alignof__] = ACTIONS(2864),
    [anon_sym___alignof] = ACTIONS(2864),
    [anon_sym__alignof] = ACTIONS(2864),
    [anon_sym_alignof] = ACTIONS(2864),
    [anon_sym__Alignof] = ACTIONS(2864),
    [anon_sym_offsetof] = ACTIONS(2864),
    [anon_sym__Generic] = ACTIONS(2864),
    [anon_sym_asm] = ACTIONS(2864),
    [anon_sym___asm__] = ACTIONS(2864),
    [anon_sym___asm] = ACTIONS(2864),
    [sym_number_literal] = ACTIONS(2866),
    [anon_sym_L_SQUOTE] = ACTIONS(2866),
    [anon_sym_u_SQUOTE] = ACTIONS(2866),
    [anon_sym_U_SQUOTE] = ACTIONS(2866),
    [anon_sym_u8_SQUOTE] = ACTIONS(2866),
    [anon_sym_SQUOTE] = ACTIONS(2866),
    [anon_sym_AT] = ACTIONS(2864),
    [anon_sym_DQUOTE] = ACTIONS(2866),
    [anon_sym_L_DQUOTE] = ACTIONS(2866),
    [anon_sym_u_DQUOTE] = ACTIONS(2866),
    [anon_sym_U_DQUOTE] = ACTIONS(2866),
    [anon_sym_u8_DQUOTE] = ACTIONS(2866),
    [sym_true] = ACTIONS(2864),
    [sym_false] = ACTIONS(2864),
    [anon_sym_NULL] = ACTIONS(2864),
    [anon_sym_nullptr] = ACTIONS(2864),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2864),
    [anon_sym___typeof] = ACTIONS(2864),
    [anon_sym_typeof] = ACTIONS(2864),
    [anon_sym_ATimport] = ACTIONS(2866),
    [aux_sym_preproc_undef_token1] = ACTIONS(2864),
    [anon_sym_POUND] = ACTIONS(2864),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2864),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2864),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2864),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2864),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2864),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2864),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2864),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_API_AVAILABLE] = ACTIONS(2864),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_API_DEPRECATED] = ACTIONS(2864),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2864),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym___deprecated_msg] = ACTIONS(2864),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2864),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2864),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2864),
    [anon_sym_ATprotocol] = ACTIONS(2866),
    [anon_sym_ATinterface] = ACTIONS(2866),
    [anon_sym_ATimplementation] = ACTIONS(2866),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2866),
    [anon_sym_ATsynthesize] = ACTIONS(2866),
    [anon_sym_ATdynamic] = ACTIONS(2866),
    [anon_sym_ATtry] = ACTIONS(2866),
    [anon_sym___try] = ACTIONS(2864),
    [anon_sym_ATthrow] = ACTIONS(2866),
    [anon_sym_ATselector] = ACTIONS(2866),
    [anon_sym_ATavailable] = ACTIONS(2866),
    [anon_sym___builtin_available] = ACTIONS(2864),
    [anon_sym_va_arg] = ACTIONS(2864),
    [anon_sym_ATencode] = ACTIONS(2866),
    [anon_sym_ATsynchronized] = ACTIONS(2866),
    [anon_sym_BOOL] = ACTIONS(2864),
    [anon_sym_IMP] = ACTIONS(2864),
    [anon_sym_SEL] = ACTIONS(2864),
    [anon_sym_Class] = ACTIONS(2864),
    [anon_sym_id] = ACTIONS(2864),
  },
  [STATE(361)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(362)] = {
    [sym_identifier] = ACTIONS(2542),
    [aux_sym_preproc_include_token1] = ACTIONS(2542),
    [aux_sym_preproc_include_token2] = ACTIONS(2542),
    [aux_sym_preproc_def_token1] = ACTIONS(2542),
    [aux_sym_preproc_if_token1] = ACTIONS(2542),
    [aux_sym_preproc_if_token2] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2542),
    [aux_sym_preproc_else_token1] = ACTIONS(2542),
    [aux_sym_preproc_elif_token1] = ACTIONS(2542),
    [sym_preproc_directive] = ACTIONS(2542),
    [anon_sym_LPAREN2] = ACTIONS(2544),
    [anon_sym_BANG] = ACTIONS(2544),
    [anon_sym_TILDE] = ACTIONS(2544),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_PLUS] = ACTIONS(2542),
    [anon_sym_STAR] = ACTIONS(2544),
    [anon_sym_CARET] = ACTIONS(2544),
    [anon_sym_AMP] = ACTIONS(2544),
    [anon_sym_SEMI] = ACTIONS(2544),
    [anon_sym___extension__] = ACTIONS(2542),
    [anon_sym_typedef] = ACTIONS(2542),
    [anon_sym_extern] = ACTIONS(2542),
    [anon_sym___attribute__] = ACTIONS(2542),
    [anon_sym___attribute] = ACTIONS(2542),
    [anon_sym_noreturn] = ACTIONS(2542),
    [anon_sym_LBRACK] = ACTIONS(2544),
    [anon_sym___declspec] = ACTIONS(2542),
    [anon_sym___cdecl] = ACTIONS(2542),
    [anon_sym___clrcall] = ACTIONS(2542),
    [anon_sym___stdcall] = ACTIONS(2542),
    [anon_sym___fastcall] = ACTIONS(2542),
    [anon_sym___thiscall] = ACTIONS(2542),
    [anon_sym___vectorcall] = ACTIONS(2542),
    [anon_sym_LBRACE] = ACTIONS(2544),
    [anon_sym_signed] = ACTIONS(2542),
    [anon_sym_unsigned] = ACTIONS(2542),
    [anon_sym_long] = ACTIONS(2542),
    [anon_sym_short] = ACTIONS(2542),
    [anon_sym_static] = ACTIONS(2542),
    [anon_sym_ATautoreleasepool] = ACTIONS(2544),
    [anon_sym_auto] = ACTIONS(2542),
    [anon_sym_register] = ACTIONS(2542),
    [anon_sym_inline] = ACTIONS(2542),
    [anon_sym___inline] = ACTIONS(2542),
    [anon_sym___inline__] = ACTIONS(2542),
    [anon_sym___forceinline] = ACTIONS(2542),
    [anon_sym_thread_local] = ACTIONS(2542),
    [anon_sym___thread] = ACTIONS(2542),
    [anon_sym_CG_EXTERN] = ACTIONS(2542),
    [anon_sym_CG_INLINE] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2542),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2542),
    [anon_sym_IBOutlet] = ACTIONS(2542),
    [anon_sym_IBInspectable] = ACTIONS(2542),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2542),
    [anon_sym_NS_INLINE] = ACTIONS(2542),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2542),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2542),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2542),
    [anon_sym_const] = ACTIONS(2542),
    [anon_sym_constexpr] = ACTIONS(2542),
    [anon_sym_volatile] = ACTIONS(2542),
    [anon_sym_restrict] = ACTIONS(2542),
    [anon_sym___restrict__] = ACTIONS(2542),
    [anon_sym__Atomic] = ACTIONS(2542),
    [anon_sym__Noreturn] = ACTIONS(2542),
    [anon_sym__Nonnull] = ACTIONS(2542),
    [anon_sym_nullable] = ACTIONS(2542),
    [anon_sym__Complex] = ACTIONS(2542),
    [anon_sym__Nullable] = ACTIONS(2542),
    [anon_sym__Nullable_result] = ACTIONS(2542),
    [anon_sym__Null_unspecified] = ACTIONS(2542),
    [anon_sym___autoreleasing] = ACTIONS(2542),
    [anon_sym___block] = ACTIONS(2542),
    [anon_sym___bridge] = ACTIONS(2542),
    [anon_sym___bridge_retained] = ACTIONS(2542),
    [anon_sym___bridge_transfer] = ACTIONS(2542),
    [anon_sym___complex] = ACTIONS(2542),
    [anon_sym___const] = ACTIONS(2542),
    [anon_sym___imag] = ACTIONS(2542),
    [anon_sym___kindof] = ACTIONS(2542),
    [anon_sym___nonnull] = ACTIONS(2542),
    [anon_sym___nullable] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2542),
    [anon_sym___real] = ACTIONS(2542),
    [anon_sym___strong] = ACTIONS(2542),
    [anon_sym___unsafe_unretained] = ACTIONS(2542),
    [anon_sym___unused] = ACTIONS(2542),
    [anon_sym___weak] = ACTIONS(2542),
    [anon_sym_alignas] = ACTIONS(2542),
    [anon_sym__Alignas] = ACTIONS(2542),
    [sym_primitive_type] = ACTIONS(2542),
    [anon_sym_enum] = ACTIONS(2542),
    [anon_sym_struct] = ACTIONS(2542),
    [anon_sym_union] = ACTIONS(2542),
    [anon_sym_if] = ACTIONS(2542),
    [anon_sym_else] = ACTIONS(2542),
    [anon_sym_switch] = ACTIONS(2542),
    [anon_sym_case] = ACTIONS(2542),
    [anon_sym_default] = ACTIONS(2542),
    [anon_sym_while] = ACTIONS(2542),
    [anon_sym_do] = ACTIONS(2542),
    [anon_sym_for] = ACTIONS(2542),
    [anon_sym_in] = ACTIONS(2542),
    [anon_sym_return] = ACTIONS(2542),
    [anon_sym_break] = ACTIONS(2542),
    [anon_sym_continue] = ACTIONS(2542),
    [anon_sym_goto] = ACTIONS(2542),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_sizeof] = ACTIONS(2542),
    [anon_sym___alignof__] = ACTIONS(2542),
    [anon_sym___alignof] = ACTIONS(2542),
    [anon_sym__alignof] = ACTIONS(2542),
    [anon_sym_alignof] = ACTIONS(2542),
    [anon_sym__Alignof] = ACTIONS(2542),
    [anon_sym_offsetof] = ACTIONS(2542),
    [anon_sym__Generic] = ACTIONS(2542),
    [anon_sym_asm] = ACTIONS(2542),
    [anon_sym___asm__] = ACTIONS(2542),
    [anon_sym___asm] = ACTIONS(2542),
    [sym_number_literal] = ACTIONS(2544),
    [anon_sym_L_SQUOTE] = ACTIONS(2544),
    [anon_sym_u_SQUOTE] = ACTIONS(2544),
    [anon_sym_U_SQUOTE] = ACTIONS(2544),
    [anon_sym_u8_SQUOTE] = ACTIONS(2544),
    [anon_sym_SQUOTE] = ACTIONS(2544),
    [anon_sym_AT] = ACTIONS(2542),
    [anon_sym_DQUOTE] = ACTIONS(2544),
    [anon_sym_L_DQUOTE] = ACTIONS(2544),
    [anon_sym_u_DQUOTE] = ACTIONS(2544),
    [anon_sym_U_DQUOTE] = ACTIONS(2544),
    [anon_sym_u8_DQUOTE] = ACTIONS(2544),
    [sym_true] = ACTIONS(2542),
    [sym_false] = ACTIONS(2542),
    [anon_sym_NULL] = ACTIONS(2542),
    [anon_sym_nullptr] = ACTIONS(2542),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2542),
    [anon_sym___typeof] = ACTIONS(2542),
    [anon_sym_typeof] = ACTIONS(2542),
    [anon_sym_ATimport] = ACTIONS(2544),
    [aux_sym_preproc_undef_token1] = ACTIONS(2542),
    [anon_sym_POUND] = ACTIONS(2542),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2542),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2542),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2542),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_API_AVAILABLE] = ACTIONS(2542),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_API_DEPRECATED] = ACTIONS(2542),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym___deprecated_msg] = ACTIONS(2542),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2542),
    [anon_sym_ATprotocol] = ACTIONS(2544),
    [anon_sym_ATinterface] = ACTIONS(2544),
    [anon_sym_ATimplementation] = ACTIONS(2544),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2544),
    [anon_sym_ATsynthesize] = ACTIONS(2544),
    [anon_sym_ATdynamic] = ACTIONS(2544),
    [anon_sym_ATtry] = ACTIONS(2544),
    [anon_sym___try] = ACTIONS(2542),
    [anon_sym_ATthrow] = ACTIONS(2544),
    [anon_sym_ATselector] = ACTIONS(2544),
    [anon_sym_ATavailable] = ACTIONS(2544),
    [anon_sym___builtin_available] = ACTIONS(2542),
    [anon_sym_va_arg] = ACTIONS(2542),
    [anon_sym_ATencode] = ACTIONS(2544),
    [anon_sym_ATsynchronized] = ACTIONS(2544),
    [anon_sym_BOOL] = ACTIONS(2542),
    [anon_sym_IMP] = ACTIONS(2542),
    [anon_sym_SEL] = ACTIONS(2542),
    [anon_sym_Class] = ACTIONS(2542),
    [anon_sym_id] = ACTIONS(2542),
  },
  [STATE(363)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token2] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [aux_sym_preproc_else_token1] = ACTIONS(2446),
    [aux_sym_preproc_elif_token1] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(364)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token2] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [aux_sym_preproc_else_token1] = ACTIONS(2446),
    [aux_sym_preproc_elif_token1] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(365)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token2] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [aux_sym_preproc_else_token1] = ACTIONS(2438),
    [aux_sym_preproc_elif_token1] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(366)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token2] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [aux_sym_preproc_else_token1] = ACTIONS(2442),
    [aux_sym_preproc_elif_token1] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(367)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token2] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [aux_sym_preproc_else_token1] = ACTIONS(2546),
    [aux_sym_preproc_elif_token1] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(368)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token2] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [aux_sym_preproc_else_token1] = ACTIONS(2546),
    [aux_sym_preproc_elif_token1] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(369)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [aux_sym_preproc_else_token1] = ACTIONS(2510),
    [aux_sym_preproc_elif_token1] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(370)] = {
    [sym_identifier] = ACTIONS(2550),
    [aux_sym_preproc_include_token1] = ACTIONS(2550),
    [aux_sym_preproc_include_token2] = ACTIONS(2550),
    [aux_sym_preproc_def_token1] = ACTIONS(2550),
    [aux_sym_preproc_if_token1] = ACTIONS(2550),
    [aux_sym_preproc_if_token2] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2550),
    [aux_sym_preproc_else_token1] = ACTIONS(2550),
    [aux_sym_preproc_elif_token1] = ACTIONS(2550),
    [sym_preproc_directive] = ACTIONS(2550),
    [anon_sym_LPAREN2] = ACTIONS(2552),
    [anon_sym_BANG] = ACTIONS(2552),
    [anon_sym_TILDE] = ACTIONS(2552),
    [anon_sym_DASH] = ACTIONS(2550),
    [anon_sym_PLUS] = ACTIONS(2550),
    [anon_sym_STAR] = ACTIONS(2552),
    [anon_sym_CARET] = ACTIONS(2552),
    [anon_sym_AMP] = ACTIONS(2552),
    [anon_sym_SEMI] = ACTIONS(2552),
    [anon_sym___extension__] = ACTIONS(2550),
    [anon_sym_typedef] = ACTIONS(2550),
    [anon_sym_extern] = ACTIONS(2550),
    [anon_sym___attribute__] = ACTIONS(2550),
    [anon_sym___attribute] = ACTIONS(2550),
    [anon_sym_noreturn] = ACTIONS(2550),
    [anon_sym_LBRACK] = ACTIONS(2552),
    [anon_sym___declspec] = ACTIONS(2550),
    [anon_sym___cdecl] = ACTIONS(2550),
    [anon_sym___clrcall] = ACTIONS(2550),
    [anon_sym___stdcall] = ACTIONS(2550),
    [anon_sym___fastcall] = ACTIONS(2550),
    [anon_sym___thiscall] = ACTIONS(2550),
    [anon_sym___vectorcall] = ACTIONS(2550),
    [anon_sym_LBRACE] = ACTIONS(2552),
    [anon_sym_signed] = ACTIONS(2550),
    [anon_sym_unsigned] = ACTIONS(2550),
    [anon_sym_long] = ACTIONS(2550),
    [anon_sym_short] = ACTIONS(2550),
    [anon_sym_static] = ACTIONS(2550),
    [anon_sym_ATautoreleasepool] = ACTIONS(2552),
    [anon_sym_auto] = ACTIONS(2550),
    [anon_sym_register] = ACTIONS(2550),
    [anon_sym_inline] = ACTIONS(2550),
    [anon_sym___inline] = ACTIONS(2550),
    [anon_sym___inline__] = ACTIONS(2550),
    [anon_sym___forceinline] = ACTIONS(2550),
    [anon_sym_thread_local] = ACTIONS(2550),
    [anon_sym___thread] = ACTIONS(2550),
    [anon_sym_CG_EXTERN] = ACTIONS(2550),
    [anon_sym_CG_INLINE] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2550),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2550),
    [anon_sym_IBOutlet] = ACTIONS(2550),
    [anon_sym_IBInspectable] = ACTIONS(2550),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2550),
    [anon_sym_NS_INLINE] = ACTIONS(2550),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2550),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2550),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2550),
    [anon_sym_const] = ACTIONS(2550),
    [anon_sym_constexpr] = ACTIONS(2550),
    [anon_sym_volatile] = ACTIONS(2550),
    [anon_sym_restrict] = ACTIONS(2550),
    [anon_sym___restrict__] = ACTIONS(2550),
    [anon_sym__Atomic] = ACTIONS(2550),
    [anon_sym__Noreturn] = ACTIONS(2550),
    [anon_sym__Nonnull] = ACTIONS(2550),
    [anon_sym_nullable] = ACTIONS(2550),
    [anon_sym__Complex] = ACTIONS(2550),
    [anon_sym__Nullable] = ACTIONS(2550),
    [anon_sym__Nullable_result] = ACTIONS(2550),
    [anon_sym__Null_unspecified] = ACTIONS(2550),
    [anon_sym___autoreleasing] = ACTIONS(2550),
    [anon_sym___block] = ACTIONS(2550),
    [anon_sym___bridge] = ACTIONS(2550),
    [anon_sym___bridge_retained] = ACTIONS(2550),
    [anon_sym___bridge_transfer] = ACTIONS(2550),
    [anon_sym___complex] = ACTIONS(2550),
    [anon_sym___const] = ACTIONS(2550),
    [anon_sym___imag] = ACTIONS(2550),
    [anon_sym___kindof] = ACTIONS(2550),
    [anon_sym___nonnull] = ACTIONS(2550),
    [anon_sym___nullable] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2550),
    [anon_sym___real] = ACTIONS(2550),
    [anon_sym___strong] = ACTIONS(2550),
    [anon_sym___unsafe_unretained] = ACTIONS(2550),
    [anon_sym___unused] = ACTIONS(2550),
    [anon_sym___weak] = ACTIONS(2550),
    [anon_sym_alignas] = ACTIONS(2550),
    [anon_sym__Alignas] = ACTIONS(2550),
    [sym_primitive_type] = ACTIONS(2550),
    [anon_sym_enum] = ACTIONS(2550),
    [anon_sym_struct] = ACTIONS(2550),
    [anon_sym_union] = ACTIONS(2550),
    [anon_sym_if] = ACTIONS(2550),
    [anon_sym_else] = ACTIONS(2550),
    [anon_sym_switch] = ACTIONS(2550),
    [anon_sym_case] = ACTIONS(2550),
    [anon_sym_default] = ACTIONS(2550),
    [anon_sym_while] = ACTIONS(2550),
    [anon_sym_do] = ACTIONS(2550),
    [anon_sym_for] = ACTIONS(2550),
    [anon_sym_in] = ACTIONS(2550),
    [anon_sym_return] = ACTIONS(2550),
    [anon_sym_break] = ACTIONS(2550),
    [anon_sym_continue] = ACTIONS(2550),
    [anon_sym_goto] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2552),
    [anon_sym_PLUS_PLUS] = ACTIONS(2552),
    [anon_sym_sizeof] = ACTIONS(2550),
    [anon_sym___alignof__] = ACTIONS(2550),
    [anon_sym___alignof] = ACTIONS(2550),
    [anon_sym__alignof] = ACTIONS(2550),
    [anon_sym_alignof] = ACTIONS(2550),
    [anon_sym__Alignof] = ACTIONS(2550),
    [anon_sym_offsetof] = ACTIONS(2550),
    [anon_sym__Generic] = ACTIONS(2550),
    [anon_sym_asm] = ACTIONS(2550),
    [anon_sym___asm__] = ACTIONS(2550),
    [anon_sym___asm] = ACTIONS(2550),
    [sym_number_literal] = ACTIONS(2552),
    [anon_sym_L_SQUOTE] = ACTIONS(2552),
    [anon_sym_u_SQUOTE] = ACTIONS(2552),
    [anon_sym_U_SQUOTE] = ACTIONS(2552),
    [anon_sym_u8_SQUOTE] = ACTIONS(2552),
    [anon_sym_SQUOTE] = ACTIONS(2552),
    [anon_sym_AT] = ACTIONS(2550),
    [anon_sym_DQUOTE] = ACTIONS(2552),
    [anon_sym_L_DQUOTE] = ACTIONS(2552),
    [anon_sym_u_DQUOTE] = ACTIONS(2552),
    [anon_sym_U_DQUOTE] = ACTIONS(2552),
    [anon_sym_u8_DQUOTE] = ACTIONS(2552),
    [sym_true] = ACTIONS(2550),
    [sym_false] = ACTIONS(2550),
    [anon_sym_NULL] = ACTIONS(2550),
    [anon_sym_nullptr] = ACTIONS(2550),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2550),
    [anon_sym___typeof] = ACTIONS(2550),
    [anon_sym_typeof] = ACTIONS(2550),
    [anon_sym_ATimport] = ACTIONS(2552),
    [aux_sym_preproc_undef_token1] = ACTIONS(2550),
    [anon_sym_POUND] = ACTIONS(2550),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2550),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2550),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2550),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_API_AVAILABLE] = ACTIONS(2550),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_API_DEPRECATED] = ACTIONS(2550),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym___deprecated_msg] = ACTIONS(2550),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2550),
    [anon_sym_ATprotocol] = ACTIONS(2552),
    [anon_sym_ATinterface] = ACTIONS(2552),
    [anon_sym_ATimplementation] = ACTIONS(2552),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2552),
    [anon_sym_ATsynthesize] = ACTIONS(2552),
    [anon_sym_ATdynamic] = ACTIONS(2552),
    [anon_sym_ATtry] = ACTIONS(2552),
    [anon_sym___try] = ACTIONS(2550),
    [anon_sym_ATthrow] = ACTIONS(2552),
    [anon_sym_ATselector] = ACTIONS(2552),
    [anon_sym_ATavailable] = ACTIONS(2552),
    [anon_sym___builtin_available] = ACTIONS(2550),
    [anon_sym_va_arg] = ACTIONS(2550),
    [anon_sym_ATencode] = ACTIONS(2552),
    [anon_sym_ATsynchronized] = ACTIONS(2552),
    [anon_sym_BOOL] = ACTIONS(2550),
    [anon_sym_IMP] = ACTIONS(2550),
    [anon_sym_SEL] = ACTIONS(2550),
    [anon_sym_Class] = ACTIONS(2550),
    [anon_sym_id] = ACTIONS(2550),
  },
  [STATE(371)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [aux_sym_preproc_else_token1] = ACTIONS(2510),
    [aux_sym_preproc_elif_token1] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(372)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [aux_sym_preproc_else_token1] = ACTIONS(2510),
    [aux_sym_preproc_elif_token1] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(373)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_RBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATcatch] = ACTIONS(2348),
    [anon_sym___catch] = ACTIONS(2346),
    [anon_sym_ATfinally] = ACTIONS(2348),
    [anon_sym___finally] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(374)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(375)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(376)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(377)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(378)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(379)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(380)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_RBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATcatch] = ACTIONS(2352),
    [anon_sym___catch] = ACTIONS(2350),
    [anon_sym_ATfinally] = ACTIONS(2352),
    [anon_sym___finally] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(381)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_RBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATcatch] = ACTIONS(2362),
    [anon_sym___catch] = ACTIONS(2360),
    [anon_sym_ATfinally] = ACTIONS(2362),
    [anon_sym___finally] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(382)] = {
    [ts_builtin_sym_end] = ACTIONS(2370),
    [sym_identifier] = ACTIONS(2368),
    [aux_sym_preproc_include_token1] = ACTIONS(2368),
    [aux_sym_preproc_include_token2] = ACTIONS(2368),
    [aux_sym_preproc_def_token1] = ACTIONS(2368),
    [aux_sym_preproc_if_token1] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2368),
    [sym_preproc_directive] = ACTIONS(2368),
    [anon_sym_LPAREN2] = ACTIONS(2370),
    [anon_sym_BANG] = ACTIONS(2370),
    [anon_sym_TILDE] = ACTIONS(2370),
    [anon_sym_DASH] = ACTIONS(2368),
    [anon_sym_PLUS] = ACTIONS(2368),
    [anon_sym_STAR] = ACTIONS(2370),
    [anon_sym_CARET] = ACTIONS(2370),
    [anon_sym_AMP] = ACTIONS(2370),
    [anon_sym_SEMI] = ACTIONS(2370),
    [anon_sym___extension__] = ACTIONS(2368),
    [anon_sym_typedef] = ACTIONS(2368),
    [anon_sym_extern] = ACTIONS(2368),
    [anon_sym___attribute__] = ACTIONS(2368),
    [anon_sym___attribute] = ACTIONS(2368),
    [anon_sym_noreturn] = ACTIONS(2368),
    [anon_sym_LBRACK] = ACTIONS(2370),
    [anon_sym___declspec] = ACTIONS(2368),
    [anon_sym___cdecl] = ACTIONS(2368),
    [anon_sym___clrcall] = ACTIONS(2368),
    [anon_sym___stdcall] = ACTIONS(2368),
    [anon_sym___fastcall] = ACTIONS(2368),
    [anon_sym___thiscall] = ACTIONS(2368),
    [anon_sym___vectorcall] = ACTIONS(2368),
    [anon_sym_LBRACE] = ACTIONS(2370),
    [anon_sym_signed] = ACTIONS(2368),
    [anon_sym_unsigned] = ACTIONS(2368),
    [anon_sym_long] = ACTIONS(2368),
    [anon_sym_short] = ACTIONS(2368),
    [anon_sym_static] = ACTIONS(2368),
    [anon_sym_ATautoreleasepool] = ACTIONS(2370),
    [anon_sym_auto] = ACTIONS(2368),
    [anon_sym_register] = ACTIONS(2368),
    [anon_sym_inline] = ACTIONS(2368),
    [anon_sym___inline] = ACTIONS(2368),
    [anon_sym___inline__] = ACTIONS(2368),
    [anon_sym___forceinline] = ACTIONS(2368),
    [anon_sym_thread_local] = ACTIONS(2368),
    [anon_sym___thread] = ACTIONS(2368),
    [anon_sym_CG_EXTERN] = ACTIONS(2368),
    [anon_sym_CG_INLINE] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2368),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2368),
    [anon_sym_IBOutlet] = ACTIONS(2368),
    [anon_sym_IBInspectable] = ACTIONS(2368),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2368),
    [anon_sym_NS_INLINE] = ACTIONS(2368),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2368),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2368),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2368),
    [anon_sym_const] = ACTIONS(2368),
    [anon_sym_constexpr] = ACTIONS(2368),
    [anon_sym_volatile] = ACTIONS(2368),
    [anon_sym_restrict] = ACTIONS(2368),
    [anon_sym___restrict__] = ACTIONS(2368),
    [anon_sym__Atomic] = ACTIONS(2368),
    [anon_sym__Noreturn] = ACTIONS(2368),
    [anon_sym__Nonnull] = ACTIONS(2368),
    [anon_sym_nullable] = ACTIONS(2368),
    [anon_sym__Complex] = ACTIONS(2368),
    [anon_sym__Nullable] = ACTIONS(2368),
    [anon_sym__Nullable_result] = ACTIONS(2368),
    [anon_sym__Null_unspecified] = ACTIONS(2368),
    [anon_sym___autoreleasing] = ACTIONS(2368),
    [anon_sym___block] = ACTIONS(2368),
    [anon_sym___bridge] = ACTIONS(2368),
    [anon_sym___bridge_retained] = ACTIONS(2368),
    [anon_sym___bridge_transfer] = ACTIONS(2368),
    [anon_sym___complex] = ACTIONS(2368),
    [anon_sym___const] = ACTIONS(2368),
    [anon_sym___imag] = ACTIONS(2368),
    [anon_sym___kindof] = ACTIONS(2368),
    [anon_sym___nonnull] = ACTIONS(2368),
    [anon_sym___nullable] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2368),
    [anon_sym___real] = ACTIONS(2368),
    [anon_sym___strong] = ACTIONS(2368),
    [anon_sym___unsafe_unretained] = ACTIONS(2368),
    [anon_sym___unused] = ACTIONS(2368),
    [anon_sym___weak] = ACTIONS(2368),
    [anon_sym_alignas] = ACTIONS(2368),
    [anon_sym__Alignas] = ACTIONS(2368),
    [sym_primitive_type] = ACTIONS(2368),
    [anon_sym_enum] = ACTIONS(2368),
    [anon_sym_struct] = ACTIONS(2368),
    [anon_sym_union] = ACTIONS(2368),
    [anon_sym_if] = ACTIONS(2368),
    [anon_sym_else] = ACTIONS(2368),
    [anon_sym_switch] = ACTIONS(2368),
    [anon_sym_case] = ACTIONS(2368),
    [anon_sym_default] = ACTIONS(2368),
    [anon_sym_while] = ACTIONS(2368),
    [anon_sym_do] = ACTIONS(2368),
    [anon_sym_for] = ACTIONS(2368),
    [anon_sym_in] = ACTIONS(2368),
    [anon_sym_return] = ACTIONS(2368),
    [anon_sym_break] = ACTIONS(2368),
    [anon_sym_continue] = ACTIONS(2368),
    [anon_sym_goto] = ACTIONS(2368),
    [anon_sym_DASH_DASH] = ACTIONS(2370),
    [anon_sym_PLUS_PLUS] = ACTIONS(2370),
    [anon_sym_sizeof] = ACTIONS(2368),
    [anon_sym___alignof__] = ACTIONS(2368),
    [anon_sym___alignof] = ACTIONS(2368),
    [anon_sym__alignof] = ACTIONS(2368),
    [anon_sym_alignof] = ACTIONS(2368),
    [anon_sym__Alignof] = ACTIONS(2368),
    [anon_sym_offsetof] = ACTIONS(2368),
    [anon_sym__Generic] = ACTIONS(2368),
    [anon_sym_asm] = ACTIONS(2368),
    [anon_sym___asm__] = ACTIONS(2368),
    [anon_sym___asm] = ACTIONS(2368),
    [sym_number_literal] = ACTIONS(2370),
    [anon_sym_L_SQUOTE] = ACTIONS(2370),
    [anon_sym_u_SQUOTE] = ACTIONS(2370),
    [anon_sym_U_SQUOTE] = ACTIONS(2370),
    [anon_sym_u8_SQUOTE] = ACTIONS(2370),
    [anon_sym_SQUOTE] = ACTIONS(2370),
    [anon_sym_AT] = ACTIONS(2368),
    [anon_sym_DQUOTE] = ACTIONS(2370),
    [anon_sym_L_DQUOTE] = ACTIONS(2370),
    [anon_sym_u_DQUOTE] = ACTIONS(2370),
    [anon_sym_U_DQUOTE] = ACTIONS(2370),
    [anon_sym_u8_DQUOTE] = ACTIONS(2370),
    [sym_true] = ACTIONS(2368),
    [sym_false] = ACTIONS(2368),
    [anon_sym_NULL] = ACTIONS(2368),
    [anon_sym_nullptr] = ACTIONS(2368),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2368),
    [anon_sym___typeof] = ACTIONS(2368),
    [anon_sym_typeof] = ACTIONS(2368),
    [anon_sym_ATimport] = ACTIONS(2370),
    [aux_sym_preproc_undef_token1] = ACTIONS(2368),
    [anon_sym_POUND] = ACTIONS(2368),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2368),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2368),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2368),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_API_AVAILABLE] = ACTIONS(2368),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_API_DEPRECATED] = ACTIONS(2368),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym___deprecated_msg] = ACTIONS(2368),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2368),
    [anon_sym_ATprotocol] = ACTIONS(2370),
    [anon_sym_ATinterface] = ACTIONS(2370),
    [anon_sym_ATimplementation] = ACTIONS(2370),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2370),
    [anon_sym_ATtry] = ACTIONS(2370),
    [anon_sym___try] = ACTIONS(2368),
    [anon_sym_ATcatch] = ACTIONS(2370),
    [anon_sym___catch] = ACTIONS(2368),
    [anon_sym_ATfinally] = ACTIONS(2370),
    [anon_sym___finally] = ACTIONS(2368),
    [anon_sym_ATthrow] = ACTIONS(2370),
    [anon_sym_ATselector] = ACTIONS(2370),
    [anon_sym_ATavailable] = ACTIONS(2370),
    [anon_sym___builtin_available] = ACTIONS(2368),
    [anon_sym_va_arg] = ACTIONS(2368),
    [anon_sym_ATencode] = ACTIONS(2370),
    [anon_sym_ATsynchronized] = ACTIONS(2370),
    [anon_sym_BOOL] = ACTIONS(2368),
    [anon_sym_IMP] = ACTIONS(2368),
    [anon_sym_SEL] = ACTIONS(2368),
    [anon_sym_Class] = ACTIONS(2368),
    [anon_sym_id] = ACTIONS(2368),
  },
  [STATE(383)] = {
    [sym_identifier] = ACTIONS(2450),
    [aux_sym_preproc_include_token1] = ACTIONS(2450),
    [aux_sym_preproc_include_token2] = ACTIONS(2450),
    [aux_sym_preproc_def_token1] = ACTIONS(2450),
    [aux_sym_preproc_if_token1] = ACTIONS(2450),
    [aux_sym_preproc_if_token2] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2450),
    [aux_sym_preproc_else_token1] = ACTIONS(2450),
    [aux_sym_preproc_elif_token1] = ACTIONS(2450),
    [sym_preproc_directive] = ACTIONS(2450),
    [anon_sym_LPAREN2] = ACTIONS(2452),
    [anon_sym_BANG] = ACTIONS(2452),
    [anon_sym_TILDE] = ACTIONS(2452),
    [anon_sym_DASH] = ACTIONS(2450),
    [anon_sym_PLUS] = ACTIONS(2450),
    [anon_sym_STAR] = ACTIONS(2452),
    [anon_sym_CARET] = ACTIONS(2452),
    [anon_sym_AMP] = ACTIONS(2452),
    [anon_sym_SEMI] = ACTIONS(2452),
    [anon_sym___extension__] = ACTIONS(2450),
    [anon_sym_typedef] = ACTIONS(2450),
    [anon_sym_extern] = ACTIONS(2450),
    [anon_sym___attribute__] = ACTIONS(2450),
    [anon_sym___attribute] = ACTIONS(2450),
    [anon_sym_noreturn] = ACTIONS(2450),
    [anon_sym_LBRACK] = ACTIONS(2452),
    [anon_sym___declspec] = ACTIONS(2450),
    [anon_sym___cdecl] = ACTIONS(2450),
    [anon_sym___clrcall] = ACTIONS(2450),
    [anon_sym___stdcall] = ACTIONS(2450),
    [anon_sym___fastcall] = ACTIONS(2450),
    [anon_sym___thiscall] = ACTIONS(2450),
    [anon_sym___vectorcall] = ACTIONS(2450),
    [anon_sym_LBRACE] = ACTIONS(2452),
    [anon_sym_signed] = ACTIONS(2450),
    [anon_sym_unsigned] = ACTIONS(2450),
    [anon_sym_long] = ACTIONS(2450),
    [anon_sym_short] = ACTIONS(2450),
    [anon_sym_static] = ACTIONS(2450),
    [anon_sym_ATautoreleasepool] = ACTIONS(2452),
    [anon_sym_auto] = ACTIONS(2450),
    [anon_sym_register] = ACTIONS(2450),
    [anon_sym_inline] = ACTIONS(2450),
    [anon_sym___inline] = ACTIONS(2450),
    [anon_sym___inline__] = ACTIONS(2450),
    [anon_sym___forceinline] = ACTIONS(2450),
    [anon_sym_thread_local] = ACTIONS(2450),
    [anon_sym___thread] = ACTIONS(2450),
    [anon_sym_CG_EXTERN] = ACTIONS(2450),
    [anon_sym_CG_INLINE] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2450),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2450),
    [anon_sym_IBOutlet] = ACTIONS(2450),
    [anon_sym_IBInspectable] = ACTIONS(2450),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2450),
    [anon_sym_NS_INLINE] = ACTIONS(2450),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2450),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2450),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2450),
    [anon_sym_const] = ACTIONS(2450),
    [anon_sym_constexpr] = ACTIONS(2450),
    [anon_sym_volatile] = ACTIONS(2450),
    [anon_sym_restrict] = ACTIONS(2450),
    [anon_sym___restrict__] = ACTIONS(2450),
    [anon_sym__Atomic] = ACTIONS(2450),
    [anon_sym__Noreturn] = ACTIONS(2450),
    [anon_sym__Nonnull] = ACTIONS(2450),
    [anon_sym_nullable] = ACTIONS(2450),
    [anon_sym__Complex] = ACTIONS(2450),
    [anon_sym__Nullable] = ACTIONS(2450),
    [anon_sym__Nullable_result] = ACTIONS(2450),
    [anon_sym__Null_unspecified] = ACTIONS(2450),
    [anon_sym___autoreleasing] = ACTIONS(2450),
    [anon_sym___block] = ACTIONS(2450),
    [anon_sym___bridge] = ACTIONS(2450),
    [anon_sym___bridge_retained] = ACTIONS(2450),
    [anon_sym___bridge_transfer] = ACTIONS(2450),
    [anon_sym___complex] = ACTIONS(2450),
    [anon_sym___const] = ACTIONS(2450),
    [anon_sym___imag] = ACTIONS(2450),
    [anon_sym___kindof] = ACTIONS(2450),
    [anon_sym___nonnull] = ACTIONS(2450),
    [anon_sym___nullable] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2450),
    [anon_sym___real] = ACTIONS(2450),
    [anon_sym___strong] = ACTIONS(2450),
    [anon_sym___unsafe_unretained] = ACTIONS(2450),
    [anon_sym___unused] = ACTIONS(2450),
    [anon_sym___weak] = ACTIONS(2450),
    [anon_sym_alignas] = ACTIONS(2450),
    [anon_sym__Alignas] = ACTIONS(2450),
    [sym_primitive_type] = ACTIONS(2450),
    [anon_sym_enum] = ACTIONS(2450),
    [anon_sym_struct] = ACTIONS(2450),
    [anon_sym_union] = ACTIONS(2450),
    [anon_sym_if] = ACTIONS(2450),
    [anon_sym_else] = ACTIONS(2450),
    [anon_sym_switch] = ACTIONS(2450),
    [anon_sym_case] = ACTIONS(2450),
    [anon_sym_default] = ACTIONS(2450),
    [anon_sym_while] = ACTIONS(2450),
    [anon_sym_do] = ACTIONS(2450),
    [anon_sym_for] = ACTIONS(2450),
    [anon_sym_in] = ACTIONS(2450),
    [anon_sym_return] = ACTIONS(2450),
    [anon_sym_break] = ACTIONS(2450),
    [anon_sym_continue] = ACTIONS(2450),
    [anon_sym_goto] = ACTIONS(2450),
    [anon_sym_DASH_DASH] = ACTIONS(2452),
    [anon_sym_PLUS_PLUS] = ACTIONS(2452),
    [anon_sym_sizeof] = ACTIONS(2450),
    [anon_sym___alignof__] = ACTIONS(2450),
    [anon_sym___alignof] = ACTIONS(2450),
    [anon_sym__alignof] = ACTIONS(2450),
    [anon_sym_alignof] = ACTIONS(2450),
    [anon_sym__Alignof] = ACTIONS(2450),
    [anon_sym_offsetof] = ACTIONS(2450),
    [anon_sym__Generic] = ACTIONS(2450),
    [anon_sym_asm] = ACTIONS(2450),
    [anon_sym___asm__] = ACTIONS(2450),
    [anon_sym___asm] = ACTIONS(2450),
    [sym_number_literal] = ACTIONS(2452),
    [anon_sym_L_SQUOTE] = ACTIONS(2452),
    [anon_sym_u_SQUOTE] = ACTIONS(2452),
    [anon_sym_U_SQUOTE] = ACTIONS(2452),
    [anon_sym_u8_SQUOTE] = ACTIONS(2452),
    [anon_sym_SQUOTE] = ACTIONS(2452),
    [anon_sym_AT] = ACTIONS(2450),
    [anon_sym_DQUOTE] = ACTIONS(2452),
    [anon_sym_L_DQUOTE] = ACTIONS(2452),
    [anon_sym_u_DQUOTE] = ACTIONS(2452),
    [anon_sym_U_DQUOTE] = ACTIONS(2452),
    [anon_sym_u8_DQUOTE] = ACTIONS(2452),
    [sym_true] = ACTIONS(2450),
    [sym_false] = ACTIONS(2450),
    [anon_sym_NULL] = ACTIONS(2450),
    [anon_sym_nullptr] = ACTIONS(2450),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2450),
    [anon_sym___typeof] = ACTIONS(2450),
    [anon_sym_typeof] = ACTIONS(2450),
    [anon_sym_ATimport] = ACTIONS(2452),
    [aux_sym_preproc_undef_token1] = ACTIONS(2450),
    [anon_sym_POUND] = ACTIONS(2450),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2450),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2450),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2450),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_API_AVAILABLE] = ACTIONS(2450),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_API_DEPRECATED] = ACTIONS(2450),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym___deprecated_msg] = ACTIONS(2450),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2450),
    [anon_sym_ATprotocol] = ACTIONS(2452),
    [anon_sym_ATinterface] = ACTIONS(2452),
    [anon_sym_ATimplementation] = ACTIONS(2452),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2452),
    [anon_sym_ATsynthesize] = ACTIONS(2452),
    [anon_sym_ATdynamic] = ACTIONS(2452),
    [anon_sym_ATtry] = ACTIONS(2452),
    [anon_sym___try] = ACTIONS(2450),
    [anon_sym_ATthrow] = ACTIONS(2452),
    [anon_sym_ATselector] = ACTIONS(2452),
    [anon_sym_ATavailable] = ACTIONS(2452),
    [anon_sym___builtin_available] = ACTIONS(2450),
    [anon_sym_va_arg] = ACTIONS(2450),
    [anon_sym_ATencode] = ACTIONS(2452),
    [anon_sym_ATsynchronized] = ACTIONS(2452),
    [anon_sym_BOOL] = ACTIONS(2450),
    [anon_sym_IMP] = ACTIONS(2450),
    [anon_sym_SEL] = ACTIONS(2450),
    [anon_sym_Class] = ACTIONS(2450),
    [anon_sym_id] = ACTIONS(2450),
  },
  [STATE(384)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token2] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [aux_sym_preproc_else_token1] = ACTIONS(2554),
    [aux_sym_preproc_elif_token1] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(385)] = {
    [ts_builtin_sym_end] = ACTIONS(2348),
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATcatch] = ACTIONS(2348),
    [anon_sym___catch] = ACTIONS(2346),
    [anon_sym_ATfinally] = ACTIONS(2348),
    [anon_sym___finally] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(386)] = {
    [ts_builtin_sym_end] = ACTIONS(2352),
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATcatch] = ACTIONS(2352),
    [anon_sym___catch] = ACTIONS(2350),
    [anon_sym_ATfinally] = ACTIONS(2352),
    [anon_sym___finally] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(387)] = {
    [ts_builtin_sym_end] = ACTIONS(2362),
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATcatch] = ACTIONS(2362),
    [anon_sym___catch] = ACTIONS(2360),
    [anon_sym_ATfinally] = ACTIONS(2362),
    [anon_sym___finally] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(388)] = {
    [sym_identifier] = ACTIONS(2458),
    [aux_sym_preproc_include_token1] = ACTIONS(2458),
    [aux_sym_preproc_include_token2] = ACTIONS(2458),
    [aux_sym_preproc_def_token1] = ACTIONS(2458),
    [aux_sym_preproc_if_token1] = ACTIONS(2458),
    [aux_sym_preproc_if_token2] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2458),
    [aux_sym_preproc_else_token1] = ACTIONS(2458),
    [aux_sym_preproc_elif_token1] = ACTIONS(2458),
    [sym_preproc_directive] = ACTIONS(2458),
    [anon_sym_LPAREN2] = ACTIONS(2460),
    [anon_sym_BANG] = ACTIONS(2460),
    [anon_sym_TILDE] = ACTIONS(2460),
    [anon_sym_DASH] = ACTIONS(2458),
    [anon_sym_PLUS] = ACTIONS(2458),
    [anon_sym_STAR] = ACTIONS(2460),
    [anon_sym_CARET] = ACTIONS(2460),
    [anon_sym_AMP] = ACTIONS(2460),
    [anon_sym_SEMI] = ACTIONS(2460),
    [anon_sym___extension__] = ACTIONS(2458),
    [anon_sym_typedef] = ACTIONS(2458),
    [anon_sym_extern] = ACTIONS(2458),
    [anon_sym___attribute__] = ACTIONS(2458),
    [anon_sym___attribute] = ACTIONS(2458),
    [anon_sym_noreturn] = ACTIONS(2458),
    [anon_sym_LBRACK] = ACTIONS(2460),
    [anon_sym___declspec] = ACTIONS(2458),
    [anon_sym___cdecl] = ACTIONS(2458),
    [anon_sym___clrcall] = ACTIONS(2458),
    [anon_sym___stdcall] = ACTIONS(2458),
    [anon_sym___fastcall] = ACTIONS(2458),
    [anon_sym___thiscall] = ACTIONS(2458),
    [anon_sym___vectorcall] = ACTIONS(2458),
    [anon_sym_LBRACE] = ACTIONS(2460),
    [anon_sym_signed] = ACTIONS(2458),
    [anon_sym_unsigned] = ACTIONS(2458),
    [anon_sym_long] = ACTIONS(2458),
    [anon_sym_short] = ACTIONS(2458),
    [anon_sym_static] = ACTIONS(2458),
    [anon_sym_ATautoreleasepool] = ACTIONS(2460),
    [anon_sym_auto] = ACTIONS(2458),
    [anon_sym_register] = ACTIONS(2458),
    [anon_sym_inline] = ACTIONS(2458),
    [anon_sym___inline] = ACTIONS(2458),
    [anon_sym___inline__] = ACTIONS(2458),
    [anon_sym___forceinline] = ACTIONS(2458),
    [anon_sym_thread_local] = ACTIONS(2458),
    [anon_sym___thread] = ACTIONS(2458),
    [anon_sym_CG_EXTERN] = ACTIONS(2458),
    [anon_sym_CG_INLINE] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2458),
    [anon_sym_IBOutlet] = ACTIONS(2458),
    [anon_sym_IBInspectable] = ACTIONS(2458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2458),
    [anon_sym_NS_INLINE] = ACTIONS(2458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2458),
    [anon_sym_const] = ACTIONS(2458),
    [anon_sym_constexpr] = ACTIONS(2458),
    [anon_sym_volatile] = ACTIONS(2458),
    [anon_sym_restrict] = ACTIONS(2458),
    [anon_sym___restrict__] = ACTIONS(2458),
    [anon_sym__Atomic] = ACTIONS(2458),
    [anon_sym__Noreturn] = ACTIONS(2458),
    [anon_sym__Nonnull] = ACTIONS(2458),
    [anon_sym_nullable] = ACTIONS(2458),
    [anon_sym__Complex] = ACTIONS(2458),
    [anon_sym__Nullable] = ACTIONS(2458),
    [anon_sym__Nullable_result] = ACTIONS(2458),
    [anon_sym__Null_unspecified] = ACTIONS(2458),
    [anon_sym___autoreleasing] = ACTIONS(2458),
    [anon_sym___block] = ACTIONS(2458),
    [anon_sym___bridge] = ACTIONS(2458),
    [anon_sym___bridge_retained] = ACTIONS(2458),
    [anon_sym___bridge_transfer] = ACTIONS(2458),
    [anon_sym___complex] = ACTIONS(2458),
    [anon_sym___const] = ACTIONS(2458),
    [anon_sym___imag] = ACTIONS(2458),
    [anon_sym___kindof] = ACTIONS(2458),
    [anon_sym___nonnull] = ACTIONS(2458),
    [anon_sym___nullable] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2458),
    [anon_sym___real] = ACTIONS(2458),
    [anon_sym___strong] = ACTIONS(2458),
    [anon_sym___unsafe_unretained] = ACTIONS(2458),
    [anon_sym___unused] = ACTIONS(2458),
    [anon_sym___weak] = ACTIONS(2458),
    [anon_sym_alignas] = ACTIONS(2458),
    [anon_sym__Alignas] = ACTIONS(2458),
    [sym_primitive_type] = ACTIONS(2458),
    [anon_sym_enum] = ACTIONS(2458),
    [anon_sym_struct] = ACTIONS(2458),
    [anon_sym_union] = ACTIONS(2458),
    [anon_sym_if] = ACTIONS(2458),
    [anon_sym_else] = ACTIONS(2458),
    [anon_sym_switch] = ACTIONS(2458),
    [anon_sym_case] = ACTIONS(2458),
    [anon_sym_default] = ACTIONS(2458),
    [anon_sym_while] = ACTIONS(2458),
    [anon_sym_do] = ACTIONS(2458),
    [anon_sym_for] = ACTIONS(2458),
    [anon_sym_in] = ACTIONS(2458),
    [anon_sym_return] = ACTIONS(2458),
    [anon_sym_break] = ACTIONS(2458),
    [anon_sym_continue] = ACTIONS(2458),
    [anon_sym_goto] = ACTIONS(2458),
    [anon_sym_DASH_DASH] = ACTIONS(2460),
    [anon_sym_PLUS_PLUS] = ACTIONS(2460),
    [anon_sym_sizeof] = ACTIONS(2458),
    [anon_sym___alignof__] = ACTIONS(2458),
    [anon_sym___alignof] = ACTIONS(2458),
    [anon_sym__alignof] = ACTIONS(2458),
    [anon_sym_alignof] = ACTIONS(2458),
    [anon_sym__Alignof] = ACTIONS(2458),
    [anon_sym_offsetof] = ACTIONS(2458),
    [anon_sym__Generic] = ACTIONS(2458),
    [anon_sym_asm] = ACTIONS(2458),
    [anon_sym___asm__] = ACTIONS(2458),
    [anon_sym___asm] = ACTIONS(2458),
    [sym_number_literal] = ACTIONS(2460),
    [anon_sym_L_SQUOTE] = ACTIONS(2460),
    [anon_sym_u_SQUOTE] = ACTIONS(2460),
    [anon_sym_U_SQUOTE] = ACTIONS(2460),
    [anon_sym_u8_SQUOTE] = ACTIONS(2460),
    [anon_sym_SQUOTE] = ACTIONS(2460),
    [anon_sym_AT] = ACTIONS(2458),
    [anon_sym_DQUOTE] = ACTIONS(2460),
    [anon_sym_L_DQUOTE] = ACTIONS(2460),
    [anon_sym_u_DQUOTE] = ACTIONS(2460),
    [anon_sym_U_DQUOTE] = ACTIONS(2460),
    [anon_sym_u8_DQUOTE] = ACTIONS(2460),
    [sym_true] = ACTIONS(2458),
    [sym_false] = ACTIONS(2458),
    [anon_sym_NULL] = ACTIONS(2458),
    [anon_sym_nullptr] = ACTIONS(2458),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2458),
    [anon_sym___typeof] = ACTIONS(2458),
    [anon_sym_typeof] = ACTIONS(2458),
    [anon_sym_ATimport] = ACTIONS(2460),
    [aux_sym_preproc_undef_token1] = ACTIONS(2458),
    [anon_sym_POUND] = ACTIONS(2458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_API_AVAILABLE] = ACTIONS(2458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_API_DEPRECATED] = ACTIONS(2458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym___deprecated_msg] = ACTIONS(2458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2458),
    [anon_sym_ATprotocol] = ACTIONS(2460),
    [anon_sym_ATinterface] = ACTIONS(2460),
    [anon_sym_ATimplementation] = ACTIONS(2460),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2460),
    [anon_sym_ATsynthesize] = ACTIONS(2460),
    [anon_sym_ATdynamic] = ACTIONS(2460),
    [anon_sym_ATtry] = ACTIONS(2460),
    [anon_sym___try] = ACTIONS(2458),
    [anon_sym_ATthrow] = ACTIONS(2460),
    [anon_sym_ATselector] = ACTIONS(2460),
    [anon_sym_ATavailable] = ACTIONS(2460),
    [anon_sym___builtin_available] = ACTIONS(2458),
    [anon_sym_va_arg] = ACTIONS(2458),
    [anon_sym_ATencode] = ACTIONS(2460),
    [anon_sym_ATsynchronized] = ACTIONS(2460),
    [anon_sym_BOOL] = ACTIONS(2458),
    [anon_sym_IMP] = ACTIONS(2458),
    [anon_sym_SEL] = ACTIONS(2458),
    [anon_sym_Class] = ACTIONS(2458),
    [anon_sym_id] = ACTIONS(2458),
  },
  [STATE(389)] = {
    [sym_identifier] = ACTIONS(2462),
    [aux_sym_preproc_include_token1] = ACTIONS(2462),
    [aux_sym_preproc_include_token2] = ACTIONS(2462),
    [aux_sym_preproc_def_token1] = ACTIONS(2462),
    [aux_sym_preproc_if_token1] = ACTIONS(2462),
    [aux_sym_preproc_if_token2] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2462),
    [aux_sym_preproc_else_token1] = ACTIONS(2462),
    [aux_sym_preproc_elif_token1] = ACTIONS(2462),
    [sym_preproc_directive] = ACTIONS(2462),
    [anon_sym_LPAREN2] = ACTIONS(2464),
    [anon_sym_BANG] = ACTIONS(2464),
    [anon_sym_TILDE] = ACTIONS(2464),
    [anon_sym_DASH] = ACTIONS(2462),
    [anon_sym_PLUS] = ACTIONS(2462),
    [anon_sym_STAR] = ACTIONS(2464),
    [anon_sym_CARET] = ACTIONS(2464),
    [anon_sym_AMP] = ACTIONS(2464),
    [anon_sym_SEMI] = ACTIONS(2464),
    [anon_sym___extension__] = ACTIONS(2462),
    [anon_sym_typedef] = ACTIONS(2462),
    [anon_sym_extern] = ACTIONS(2462),
    [anon_sym___attribute__] = ACTIONS(2462),
    [anon_sym___attribute] = ACTIONS(2462),
    [anon_sym_noreturn] = ACTIONS(2462),
    [anon_sym_LBRACK] = ACTIONS(2464),
    [anon_sym___declspec] = ACTIONS(2462),
    [anon_sym___cdecl] = ACTIONS(2462),
    [anon_sym___clrcall] = ACTIONS(2462),
    [anon_sym___stdcall] = ACTIONS(2462),
    [anon_sym___fastcall] = ACTIONS(2462),
    [anon_sym___thiscall] = ACTIONS(2462),
    [anon_sym___vectorcall] = ACTIONS(2462),
    [anon_sym_LBRACE] = ACTIONS(2464),
    [anon_sym_signed] = ACTIONS(2462),
    [anon_sym_unsigned] = ACTIONS(2462),
    [anon_sym_long] = ACTIONS(2462),
    [anon_sym_short] = ACTIONS(2462),
    [anon_sym_static] = ACTIONS(2462),
    [anon_sym_ATautoreleasepool] = ACTIONS(2464),
    [anon_sym_auto] = ACTIONS(2462),
    [anon_sym_register] = ACTIONS(2462),
    [anon_sym_inline] = ACTIONS(2462),
    [anon_sym___inline] = ACTIONS(2462),
    [anon_sym___inline__] = ACTIONS(2462),
    [anon_sym___forceinline] = ACTIONS(2462),
    [anon_sym_thread_local] = ACTIONS(2462),
    [anon_sym___thread] = ACTIONS(2462),
    [anon_sym_CG_EXTERN] = ACTIONS(2462),
    [anon_sym_CG_INLINE] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2462),
    [anon_sym_IBOutlet] = ACTIONS(2462),
    [anon_sym_IBInspectable] = ACTIONS(2462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2462),
    [anon_sym_NS_INLINE] = ACTIONS(2462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2462),
    [anon_sym_const] = ACTIONS(2462),
    [anon_sym_constexpr] = ACTIONS(2462),
    [anon_sym_volatile] = ACTIONS(2462),
    [anon_sym_restrict] = ACTIONS(2462),
    [anon_sym___restrict__] = ACTIONS(2462),
    [anon_sym__Atomic] = ACTIONS(2462),
    [anon_sym__Noreturn] = ACTIONS(2462),
    [anon_sym__Nonnull] = ACTIONS(2462),
    [anon_sym_nullable] = ACTIONS(2462),
    [anon_sym__Complex] = ACTIONS(2462),
    [anon_sym__Nullable] = ACTIONS(2462),
    [anon_sym__Nullable_result] = ACTIONS(2462),
    [anon_sym__Null_unspecified] = ACTIONS(2462),
    [anon_sym___autoreleasing] = ACTIONS(2462),
    [anon_sym___block] = ACTIONS(2462),
    [anon_sym___bridge] = ACTIONS(2462),
    [anon_sym___bridge_retained] = ACTIONS(2462),
    [anon_sym___bridge_transfer] = ACTIONS(2462),
    [anon_sym___complex] = ACTIONS(2462),
    [anon_sym___const] = ACTIONS(2462),
    [anon_sym___imag] = ACTIONS(2462),
    [anon_sym___kindof] = ACTIONS(2462),
    [anon_sym___nonnull] = ACTIONS(2462),
    [anon_sym___nullable] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2462),
    [anon_sym___real] = ACTIONS(2462),
    [anon_sym___strong] = ACTIONS(2462),
    [anon_sym___unsafe_unretained] = ACTIONS(2462),
    [anon_sym___unused] = ACTIONS(2462),
    [anon_sym___weak] = ACTIONS(2462),
    [anon_sym_alignas] = ACTIONS(2462),
    [anon_sym__Alignas] = ACTIONS(2462),
    [sym_primitive_type] = ACTIONS(2462),
    [anon_sym_enum] = ACTIONS(2462),
    [anon_sym_struct] = ACTIONS(2462),
    [anon_sym_union] = ACTIONS(2462),
    [anon_sym_if] = ACTIONS(2462),
    [anon_sym_else] = ACTIONS(2462),
    [anon_sym_switch] = ACTIONS(2462),
    [anon_sym_case] = ACTIONS(2462),
    [anon_sym_default] = ACTIONS(2462),
    [anon_sym_while] = ACTIONS(2462),
    [anon_sym_do] = ACTIONS(2462),
    [anon_sym_for] = ACTIONS(2462),
    [anon_sym_in] = ACTIONS(2462),
    [anon_sym_return] = ACTIONS(2462),
    [anon_sym_break] = ACTIONS(2462),
    [anon_sym_continue] = ACTIONS(2462),
    [anon_sym_goto] = ACTIONS(2462),
    [anon_sym_DASH_DASH] = ACTIONS(2464),
    [anon_sym_PLUS_PLUS] = ACTIONS(2464),
    [anon_sym_sizeof] = ACTIONS(2462),
    [anon_sym___alignof__] = ACTIONS(2462),
    [anon_sym___alignof] = ACTIONS(2462),
    [anon_sym__alignof] = ACTIONS(2462),
    [anon_sym_alignof] = ACTIONS(2462),
    [anon_sym__Alignof] = ACTIONS(2462),
    [anon_sym_offsetof] = ACTIONS(2462),
    [anon_sym__Generic] = ACTIONS(2462),
    [anon_sym_asm] = ACTIONS(2462),
    [anon_sym___asm__] = ACTIONS(2462),
    [anon_sym___asm] = ACTIONS(2462),
    [sym_number_literal] = ACTIONS(2464),
    [anon_sym_L_SQUOTE] = ACTIONS(2464),
    [anon_sym_u_SQUOTE] = ACTIONS(2464),
    [anon_sym_U_SQUOTE] = ACTIONS(2464),
    [anon_sym_u8_SQUOTE] = ACTIONS(2464),
    [anon_sym_SQUOTE] = ACTIONS(2464),
    [anon_sym_AT] = ACTIONS(2462),
    [anon_sym_DQUOTE] = ACTIONS(2464),
    [anon_sym_L_DQUOTE] = ACTIONS(2464),
    [anon_sym_u_DQUOTE] = ACTIONS(2464),
    [anon_sym_U_DQUOTE] = ACTIONS(2464),
    [anon_sym_u8_DQUOTE] = ACTIONS(2464),
    [sym_true] = ACTIONS(2462),
    [sym_false] = ACTIONS(2462),
    [anon_sym_NULL] = ACTIONS(2462),
    [anon_sym_nullptr] = ACTIONS(2462),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2462),
    [anon_sym___typeof] = ACTIONS(2462),
    [anon_sym_typeof] = ACTIONS(2462),
    [anon_sym_ATimport] = ACTIONS(2464),
    [aux_sym_preproc_undef_token1] = ACTIONS(2462),
    [anon_sym_POUND] = ACTIONS(2462),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2462),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_API_AVAILABLE] = ACTIONS(2462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_API_DEPRECATED] = ACTIONS(2462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym___deprecated_msg] = ACTIONS(2462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2462),
    [anon_sym_ATprotocol] = ACTIONS(2464),
    [anon_sym_ATinterface] = ACTIONS(2464),
    [anon_sym_ATimplementation] = ACTIONS(2464),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2464),
    [anon_sym_ATsynthesize] = ACTIONS(2464),
    [anon_sym_ATdynamic] = ACTIONS(2464),
    [anon_sym_ATtry] = ACTIONS(2464),
    [anon_sym___try] = ACTIONS(2462),
    [anon_sym_ATthrow] = ACTIONS(2464),
    [anon_sym_ATselector] = ACTIONS(2464),
    [anon_sym_ATavailable] = ACTIONS(2464),
    [anon_sym___builtin_available] = ACTIONS(2462),
    [anon_sym_va_arg] = ACTIONS(2462),
    [anon_sym_ATencode] = ACTIONS(2464),
    [anon_sym_ATsynchronized] = ACTIONS(2464),
    [anon_sym_BOOL] = ACTIONS(2462),
    [anon_sym_IMP] = ACTIONS(2462),
    [anon_sym_SEL] = ACTIONS(2462),
    [anon_sym_Class] = ACTIONS(2462),
    [anon_sym_id] = ACTIONS(2462),
  },
  [STATE(390)] = {
    [sym_identifier] = ACTIONS(2466),
    [aux_sym_preproc_include_token1] = ACTIONS(2466),
    [aux_sym_preproc_include_token2] = ACTIONS(2466),
    [aux_sym_preproc_def_token1] = ACTIONS(2466),
    [aux_sym_preproc_if_token1] = ACTIONS(2466),
    [aux_sym_preproc_if_token2] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2466),
    [aux_sym_preproc_else_token1] = ACTIONS(2466),
    [aux_sym_preproc_elif_token1] = ACTIONS(2466),
    [sym_preproc_directive] = ACTIONS(2466),
    [anon_sym_LPAREN2] = ACTIONS(2468),
    [anon_sym_BANG] = ACTIONS(2468),
    [anon_sym_TILDE] = ACTIONS(2468),
    [anon_sym_DASH] = ACTIONS(2466),
    [anon_sym_PLUS] = ACTIONS(2466),
    [anon_sym_STAR] = ACTIONS(2468),
    [anon_sym_CARET] = ACTIONS(2468),
    [anon_sym_AMP] = ACTIONS(2468),
    [anon_sym_SEMI] = ACTIONS(2468),
    [anon_sym___extension__] = ACTIONS(2466),
    [anon_sym_typedef] = ACTIONS(2466),
    [anon_sym_extern] = ACTIONS(2466),
    [anon_sym___attribute__] = ACTIONS(2466),
    [anon_sym___attribute] = ACTIONS(2466),
    [anon_sym_noreturn] = ACTIONS(2466),
    [anon_sym_LBRACK] = ACTIONS(2468),
    [anon_sym___declspec] = ACTIONS(2466),
    [anon_sym___cdecl] = ACTIONS(2466),
    [anon_sym___clrcall] = ACTIONS(2466),
    [anon_sym___stdcall] = ACTIONS(2466),
    [anon_sym___fastcall] = ACTIONS(2466),
    [anon_sym___thiscall] = ACTIONS(2466),
    [anon_sym___vectorcall] = ACTIONS(2466),
    [anon_sym_LBRACE] = ACTIONS(2468),
    [anon_sym_signed] = ACTIONS(2466),
    [anon_sym_unsigned] = ACTIONS(2466),
    [anon_sym_long] = ACTIONS(2466),
    [anon_sym_short] = ACTIONS(2466),
    [anon_sym_static] = ACTIONS(2466),
    [anon_sym_ATautoreleasepool] = ACTIONS(2468),
    [anon_sym_auto] = ACTIONS(2466),
    [anon_sym_register] = ACTIONS(2466),
    [anon_sym_inline] = ACTIONS(2466),
    [anon_sym___inline] = ACTIONS(2466),
    [anon_sym___inline__] = ACTIONS(2466),
    [anon_sym___forceinline] = ACTIONS(2466),
    [anon_sym_thread_local] = ACTIONS(2466),
    [anon_sym___thread] = ACTIONS(2466),
    [anon_sym_CG_EXTERN] = ACTIONS(2466),
    [anon_sym_CG_INLINE] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2466),
    [anon_sym_IBOutlet] = ACTIONS(2466),
    [anon_sym_IBInspectable] = ACTIONS(2466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2466),
    [anon_sym_NS_INLINE] = ACTIONS(2466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2466),
    [anon_sym_const] = ACTIONS(2466),
    [anon_sym_constexpr] = ACTIONS(2466),
    [anon_sym_volatile] = ACTIONS(2466),
    [anon_sym_restrict] = ACTIONS(2466),
    [anon_sym___restrict__] = ACTIONS(2466),
    [anon_sym__Atomic] = ACTIONS(2466),
    [anon_sym__Noreturn] = ACTIONS(2466),
    [anon_sym__Nonnull] = ACTIONS(2466),
    [anon_sym_nullable] = ACTIONS(2466),
    [anon_sym__Complex] = ACTIONS(2466),
    [anon_sym__Nullable] = ACTIONS(2466),
    [anon_sym__Nullable_result] = ACTIONS(2466),
    [anon_sym__Null_unspecified] = ACTIONS(2466),
    [anon_sym___autoreleasing] = ACTIONS(2466),
    [anon_sym___block] = ACTIONS(2466),
    [anon_sym___bridge] = ACTIONS(2466),
    [anon_sym___bridge_retained] = ACTIONS(2466),
    [anon_sym___bridge_transfer] = ACTIONS(2466),
    [anon_sym___complex] = ACTIONS(2466),
    [anon_sym___const] = ACTIONS(2466),
    [anon_sym___imag] = ACTIONS(2466),
    [anon_sym___kindof] = ACTIONS(2466),
    [anon_sym___nonnull] = ACTIONS(2466),
    [anon_sym___nullable] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2466),
    [anon_sym___real] = ACTIONS(2466),
    [anon_sym___strong] = ACTIONS(2466),
    [anon_sym___unsafe_unretained] = ACTIONS(2466),
    [anon_sym___unused] = ACTIONS(2466),
    [anon_sym___weak] = ACTIONS(2466),
    [anon_sym_alignas] = ACTIONS(2466),
    [anon_sym__Alignas] = ACTIONS(2466),
    [sym_primitive_type] = ACTIONS(2466),
    [anon_sym_enum] = ACTIONS(2466),
    [anon_sym_struct] = ACTIONS(2466),
    [anon_sym_union] = ACTIONS(2466),
    [anon_sym_if] = ACTIONS(2466),
    [anon_sym_else] = ACTIONS(2466),
    [anon_sym_switch] = ACTIONS(2466),
    [anon_sym_case] = ACTIONS(2466),
    [anon_sym_default] = ACTIONS(2466),
    [anon_sym_while] = ACTIONS(2466),
    [anon_sym_do] = ACTIONS(2466),
    [anon_sym_for] = ACTIONS(2466),
    [anon_sym_in] = ACTIONS(2466),
    [anon_sym_return] = ACTIONS(2466),
    [anon_sym_break] = ACTIONS(2466),
    [anon_sym_continue] = ACTIONS(2466),
    [anon_sym_goto] = ACTIONS(2466),
    [anon_sym_DASH_DASH] = ACTIONS(2468),
    [anon_sym_PLUS_PLUS] = ACTIONS(2468),
    [anon_sym_sizeof] = ACTIONS(2466),
    [anon_sym___alignof__] = ACTIONS(2466),
    [anon_sym___alignof] = ACTIONS(2466),
    [anon_sym__alignof] = ACTIONS(2466),
    [anon_sym_alignof] = ACTIONS(2466),
    [anon_sym__Alignof] = ACTIONS(2466),
    [anon_sym_offsetof] = ACTIONS(2466),
    [anon_sym__Generic] = ACTIONS(2466),
    [anon_sym_asm] = ACTIONS(2466),
    [anon_sym___asm__] = ACTIONS(2466),
    [anon_sym___asm] = ACTIONS(2466),
    [sym_number_literal] = ACTIONS(2468),
    [anon_sym_L_SQUOTE] = ACTIONS(2468),
    [anon_sym_u_SQUOTE] = ACTIONS(2468),
    [anon_sym_U_SQUOTE] = ACTIONS(2468),
    [anon_sym_u8_SQUOTE] = ACTIONS(2468),
    [anon_sym_SQUOTE] = ACTIONS(2468),
    [anon_sym_AT] = ACTIONS(2466),
    [anon_sym_DQUOTE] = ACTIONS(2468),
    [anon_sym_L_DQUOTE] = ACTIONS(2468),
    [anon_sym_u_DQUOTE] = ACTIONS(2468),
    [anon_sym_U_DQUOTE] = ACTIONS(2468),
    [anon_sym_u8_DQUOTE] = ACTIONS(2468),
    [sym_true] = ACTIONS(2466),
    [sym_false] = ACTIONS(2466),
    [anon_sym_NULL] = ACTIONS(2466),
    [anon_sym_nullptr] = ACTIONS(2466),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2466),
    [anon_sym___typeof] = ACTIONS(2466),
    [anon_sym_typeof] = ACTIONS(2466),
    [anon_sym_ATimport] = ACTIONS(2468),
    [aux_sym_preproc_undef_token1] = ACTIONS(2466),
    [anon_sym_POUND] = ACTIONS(2466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_API_AVAILABLE] = ACTIONS(2466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_API_DEPRECATED] = ACTIONS(2466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym___deprecated_msg] = ACTIONS(2466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2466),
    [anon_sym_ATprotocol] = ACTIONS(2468),
    [anon_sym_ATinterface] = ACTIONS(2468),
    [anon_sym_ATimplementation] = ACTIONS(2468),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2468),
    [anon_sym_ATsynthesize] = ACTIONS(2468),
    [anon_sym_ATdynamic] = ACTIONS(2468),
    [anon_sym_ATtry] = ACTIONS(2468),
    [anon_sym___try] = ACTIONS(2466),
    [anon_sym_ATthrow] = ACTIONS(2468),
    [anon_sym_ATselector] = ACTIONS(2468),
    [anon_sym_ATavailable] = ACTIONS(2468),
    [anon_sym___builtin_available] = ACTIONS(2466),
    [anon_sym_va_arg] = ACTIONS(2466),
    [anon_sym_ATencode] = ACTIONS(2468),
    [anon_sym_ATsynchronized] = ACTIONS(2468),
    [anon_sym_BOOL] = ACTIONS(2466),
    [anon_sym_IMP] = ACTIONS(2466),
    [anon_sym_SEL] = ACTIONS(2466),
    [anon_sym_Class] = ACTIONS(2466),
    [anon_sym_id] = ACTIONS(2466),
  },
  [STATE(391)] = {
    [sym_identifier] = ACTIONS(2470),
    [aux_sym_preproc_include_token1] = ACTIONS(2470),
    [aux_sym_preproc_include_token2] = ACTIONS(2470),
    [aux_sym_preproc_def_token1] = ACTIONS(2470),
    [aux_sym_preproc_if_token1] = ACTIONS(2470),
    [aux_sym_preproc_if_token2] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2470),
    [aux_sym_preproc_else_token1] = ACTIONS(2470),
    [aux_sym_preproc_elif_token1] = ACTIONS(2470),
    [sym_preproc_directive] = ACTIONS(2470),
    [anon_sym_LPAREN2] = ACTIONS(2472),
    [anon_sym_BANG] = ACTIONS(2472),
    [anon_sym_TILDE] = ACTIONS(2472),
    [anon_sym_DASH] = ACTIONS(2470),
    [anon_sym_PLUS] = ACTIONS(2470),
    [anon_sym_STAR] = ACTIONS(2472),
    [anon_sym_CARET] = ACTIONS(2472),
    [anon_sym_AMP] = ACTIONS(2472),
    [anon_sym_SEMI] = ACTIONS(2472),
    [anon_sym___extension__] = ACTIONS(2470),
    [anon_sym_typedef] = ACTIONS(2470),
    [anon_sym_extern] = ACTIONS(2470),
    [anon_sym___attribute__] = ACTIONS(2470),
    [anon_sym___attribute] = ACTIONS(2470),
    [anon_sym_noreturn] = ACTIONS(2470),
    [anon_sym_LBRACK] = ACTIONS(2472),
    [anon_sym___declspec] = ACTIONS(2470),
    [anon_sym___cdecl] = ACTIONS(2470),
    [anon_sym___clrcall] = ACTIONS(2470),
    [anon_sym___stdcall] = ACTIONS(2470),
    [anon_sym___fastcall] = ACTIONS(2470),
    [anon_sym___thiscall] = ACTIONS(2470),
    [anon_sym___vectorcall] = ACTIONS(2470),
    [anon_sym_LBRACE] = ACTIONS(2472),
    [anon_sym_signed] = ACTIONS(2470),
    [anon_sym_unsigned] = ACTIONS(2470),
    [anon_sym_long] = ACTIONS(2470),
    [anon_sym_short] = ACTIONS(2470),
    [anon_sym_static] = ACTIONS(2470),
    [anon_sym_ATautoreleasepool] = ACTIONS(2472),
    [anon_sym_auto] = ACTIONS(2470),
    [anon_sym_register] = ACTIONS(2470),
    [anon_sym_inline] = ACTIONS(2470),
    [anon_sym___inline] = ACTIONS(2470),
    [anon_sym___inline__] = ACTIONS(2470),
    [anon_sym___forceinline] = ACTIONS(2470),
    [anon_sym_thread_local] = ACTIONS(2470),
    [anon_sym___thread] = ACTIONS(2470),
    [anon_sym_CG_EXTERN] = ACTIONS(2470),
    [anon_sym_CG_INLINE] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2470),
    [anon_sym_IBOutlet] = ACTIONS(2470),
    [anon_sym_IBInspectable] = ACTIONS(2470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2470),
    [anon_sym_NS_INLINE] = ACTIONS(2470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2470),
    [anon_sym_const] = ACTIONS(2470),
    [anon_sym_constexpr] = ACTIONS(2470),
    [anon_sym_volatile] = ACTIONS(2470),
    [anon_sym_restrict] = ACTIONS(2470),
    [anon_sym___restrict__] = ACTIONS(2470),
    [anon_sym__Atomic] = ACTIONS(2470),
    [anon_sym__Noreturn] = ACTIONS(2470),
    [anon_sym__Nonnull] = ACTIONS(2470),
    [anon_sym_nullable] = ACTIONS(2470),
    [anon_sym__Complex] = ACTIONS(2470),
    [anon_sym__Nullable] = ACTIONS(2470),
    [anon_sym__Nullable_result] = ACTIONS(2470),
    [anon_sym__Null_unspecified] = ACTIONS(2470),
    [anon_sym___autoreleasing] = ACTIONS(2470),
    [anon_sym___block] = ACTIONS(2470),
    [anon_sym___bridge] = ACTIONS(2470),
    [anon_sym___bridge_retained] = ACTIONS(2470),
    [anon_sym___bridge_transfer] = ACTIONS(2470),
    [anon_sym___complex] = ACTIONS(2470),
    [anon_sym___const] = ACTIONS(2470),
    [anon_sym___imag] = ACTIONS(2470),
    [anon_sym___kindof] = ACTIONS(2470),
    [anon_sym___nonnull] = ACTIONS(2470),
    [anon_sym___nullable] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2470),
    [anon_sym___real] = ACTIONS(2470),
    [anon_sym___strong] = ACTIONS(2470),
    [anon_sym___unsafe_unretained] = ACTIONS(2470),
    [anon_sym___unused] = ACTIONS(2470),
    [anon_sym___weak] = ACTIONS(2470),
    [anon_sym_alignas] = ACTIONS(2470),
    [anon_sym__Alignas] = ACTIONS(2470),
    [sym_primitive_type] = ACTIONS(2470),
    [anon_sym_enum] = ACTIONS(2470),
    [anon_sym_struct] = ACTIONS(2470),
    [anon_sym_union] = ACTIONS(2470),
    [anon_sym_if] = ACTIONS(2470),
    [anon_sym_else] = ACTIONS(2470),
    [anon_sym_switch] = ACTIONS(2470),
    [anon_sym_case] = ACTIONS(2470),
    [anon_sym_default] = ACTIONS(2470),
    [anon_sym_while] = ACTIONS(2470),
    [anon_sym_do] = ACTIONS(2470),
    [anon_sym_for] = ACTIONS(2470),
    [anon_sym_in] = ACTIONS(2470),
    [anon_sym_return] = ACTIONS(2470),
    [anon_sym_break] = ACTIONS(2470),
    [anon_sym_continue] = ACTIONS(2470),
    [anon_sym_goto] = ACTIONS(2470),
    [anon_sym_DASH_DASH] = ACTIONS(2472),
    [anon_sym_PLUS_PLUS] = ACTIONS(2472),
    [anon_sym_sizeof] = ACTIONS(2470),
    [anon_sym___alignof__] = ACTIONS(2470),
    [anon_sym___alignof] = ACTIONS(2470),
    [anon_sym__alignof] = ACTIONS(2470),
    [anon_sym_alignof] = ACTIONS(2470),
    [anon_sym__Alignof] = ACTIONS(2470),
    [anon_sym_offsetof] = ACTIONS(2470),
    [anon_sym__Generic] = ACTIONS(2470),
    [anon_sym_asm] = ACTIONS(2470),
    [anon_sym___asm__] = ACTIONS(2470),
    [anon_sym___asm] = ACTIONS(2470),
    [sym_number_literal] = ACTIONS(2472),
    [anon_sym_L_SQUOTE] = ACTIONS(2472),
    [anon_sym_u_SQUOTE] = ACTIONS(2472),
    [anon_sym_U_SQUOTE] = ACTIONS(2472),
    [anon_sym_u8_SQUOTE] = ACTIONS(2472),
    [anon_sym_SQUOTE] = ACTIONS(2472),
    [anon_sym_AT] = ACTIONS(2470),
    [anon_sym_DQUOTE] = ACTIONS(2472),
    [anon_sym_L_DQUOTE] = ACTIONS(2472),
    [anon_sym_u_DQUOTE] = ACTIONS(2472),
    [anon_sym_U_DQUOTE] = ACTIONS(2472),
    [anon_sym_u8_DQUOTE] = ACTIONS(2472),
    [sym_true] = ACTIONS(2470),
    [sym_false] = ACTIONS(2470),
    [anon_sym_NULL] = ACTIONS(2470),
    [anon_sym_nullptr] = ACTIONS(2470),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2470),
    [anon_sym___typeof] = ACTIONS(2470),
    [anon_sym_typeof] = ACTIONS(2470),
    [anon_sym_ATimport] = ACTIONS(2472),
    [aux_sym_preproc_undef_token1] = ACTIONS(2470),
    [anon_sym_POUND] = ACTIONS(2470),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_API_AVAILABLE] = ACTIONS(2470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_API_DEPRECATED] = ACTIONS(2470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym___deprecated_msg] = ACTIONS(2470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2470),
    [anon_sym_ATprotocol] = ACTIONS(2472),
    [anon_sym_ATinterface] = ACTIONS(2472),
    [anon_sym_ATimplementation] = ACTIONS(2472),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2472),
    [anon_sym_ATsynthesize] = ACTIONS(2472),
    [anon_sym_ATdynamic] = ACTIONS(2472),
    [anon_sym_ATtry] = ACTIONS(2472),
    [anon_sym___try] = ACTIONS(2470),
    [anon_sym_ATthrow] = ACTIONS(2472),
    [anon_sym_ATselector] = ACTIONS(2472),
    [anon_sym_ATavailable] = ACTIONS(2472),
    [anon_sym___builtin_available] = ACTIONS(2470),
    [anon_sym_va_arg] = ACTIONS(2470),
    [anon_sym_ATencode] = ACTIONS(2472),
    [anon_sym_ATsynchronized] = ACTIONS(2472),
    [anon_sym_BOOL] = ACTIONS(2470),
    [anon_sym_IMP] = ACTIONS(2470),
    [anon_sym_SEL] = ACTIONS(2470),
    [anon_sym_Class] = ACTIONS(2470),
    [anon_sym_id] = ACTIONS(2470),
  },
  [STATE(392)] = {
    [sym_identifier] = ACTIONS(2568),
    [aux_sym_preproc_include_token1] = ACTIONS(2568),
    [aux_sym_preproc_include_token2] = ACTIONS(2568),
    [aux_sym_preproc_def_token1] = ACTIONS(2568),
    [aux_sym_preproc_if_token1] = ACTIONS(2568),
    [aux_sym_preproc_if_token2] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2568),
    [aux_sym_preproc_else_token1] = ACTIONS(2568),
    [aux_sym_preproc_elif_token1] = ACTIONS(2568),
    [sym_preproc_directive] = ACTIONS(2568),
    [anon_sym_LPAREN2] = ACTIONS(2570),
    [anon_sym_BANG] = ACTIONS(2570),
    [anon_sym_TILDE] = ACTIONS(2570),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_PLUS] = ACTIONS(2568),
    [anon_sym_STAR] = ACTIONS(2570),
    [anon_sym_CARET] = ACTIONS(2570),
    [anon_sym_AMP] = ACTIONS(2570),
    [anon_sym_SEMI] = ACTIONS(2570),
    [anon_sym___extension__] = ACTIONS(2568),
    [anon_sym_typedef] = ACTIONS(2568),
    [anon_sym_extern] = ACTIONS(2568),
    [anon_sym___attribute__] = ACTIONS(2568),
    [anon_sym___attribute] = ACTIONS(2568),
    [anon_sym_noreturn] = ACTIONS(2568),
    [anon_sym_LBRACK] = ACTIONS(2570),
    [anon_sym___declspec] = ACTIONS(2568),
    [anon_sym___cdecl] = ACTIONS(2568),
    [anon_sym___clrcall] = ACTIONS(2568),
    [anon_sym___stdcall] = ACTIONS(2568),
    [anon_sym___fastcall] = ACTIONS(2568),
    [anon_sym___thiscall] = ACTIONS(2568),
    [anon_sym___vectorcall] = ACTIONS(2568),
    [anon_sym_LBRACE] = ACTIONS(2570),
    [anon_sym_signed] = ACTIONS(2568),
    [anon_sym_unsigned] = ACTIONS(2568),
    [anon_sym_long] = ACTIONS(2568),
    [anon_sym_short] = ACTIONS(2568),
    [anon_sym_static] = ACTIONS(2568),
    [anon_sym_ATautoreleasepool] = ACTIONS(2570),
    [anon_sym_auto] = ACTIONS(2568),
    [anon_sym_register] = ACTIONS(2568),
    [anon_sym_inline] = ACTIONS(2568),
    [anon_sym___inline] = ACTIONS(2568),
    [anon_sym___inline__] = ACTIONS(2568),
    [anon_sym___forceinline] = ACTIONS(2568),
    [anon_sym_thread_local] = ACTIONS(2568),
    [anon_sym___thread] = ACTIONS(2568),
    [anon_sym_CG_EXTERN] = ACTIONS(2568),
    [anon_sym_CG_INLINE] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2568),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2568),
    [anon_sym_IBOutlet] = ACTIONS(2568),
    [anon_sym_IBInspectable] = ACTIONS(2568),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2568),
    [anon_sym_NS_INLINE] = ACTIONS(2568),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2568),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2568),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2568),
    [anon_sym_const] = ACTIONS(2568),
    [anon_sym_constexpr] = ACTIONS(2568),
    [anon_sym_volatile] = ACTIONS(2568),
    [anon_sym_restrict] = ACTIONS(2568),
    [anon_sym___restrict__] = ACTIONS(2568),
    [anon_sym__Atomic] = ACTIONS(2568),
    [anon_sym__Noreturn] = ACTIONS(2568),
    [anon_sym__Nonnull] = ACTIONS(2568),
    [anon_sym_nullable] = ACTIONS(2568),
    [anon_sym__Complex] = ACTIONS(2568),
    [anon_sym__Nullable] = ACTIONS(2568),
    [anon_sym__Nullable_result] = ACTIONS(2568),
    [anon_sym__Null_unspecified] = ACTIONS(2568),
    [anon_sym___autoreleasing] = ACTIONS(2568),
    [anon_sym___block] = ACTIONS(2568),
    [anon_sym___bridge] = ACTIONS(2568),
    [anon_sym___bridge_retained] = ACTIONS(2568),
    [anon_sym___bridge_transfer] = ACTIONS(2568),
    [anon_sym___complex] = ACTIONS(2568),
    [anon_sym___const] = ACTIONS(2568),
    [anon_sym___imag] = ACTIONS(2568),
    [anon_sym___kindof] = ACTIONS(2568),
    [anon_sym___nonnull] = ACTIONS(2568),
    [anon_sym___nullable] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2568),
    [anon_sym___real] = ACTIONS(2568),
    [anon_sym___strong] = ACTIONS(2568),
    [anon_sym___unsafe_unretained] = ACTIONS(2568),
    [anon_sym___unused] = ACTIONS(2568),
    [anon_sym___weak] = ACTIONS(2568),
    [anon_sym_alignas] = ACTIONS(2568),
    [anon_sym__Alignas] = ACTIONS(2568),
    [sym_primitive_type] = ACTIONS(2568),
    [anon_sym_enum] = ACTIONS(2568),
    [anon_sym_struct] = ACTIONS(2568),
    [anon_sym_union] = ACTIONS(2568),
    [anon_sym_if] = ACTIONS(2568),
    [anon_sym_else] = ACTIONS(2568),
    [anon_sym_switch] = ACTIONS(2568),
    [anon_sym_case] = ACTIONS(2568),
    [anon_sym_default] = ACTIONS(2568),
    [anon_sym_while] = ACTIONS(2568),
    [anon_sym_do] = ACTIONS(2568),
    [anon_sym_for] = ACTIONS(2568),
    [anon_sym_in] = ACTIONS(2568),
    [anon_sym_return] = ACTIONS(2568),
    [anon_sym_break] = ACTIONS(2568),
    [anon_sym_continue] = ACTIONS(2568),
    [anon_sym_goto] = ACTIONS(2568),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_sizeof] = ACTIONS(2568),
    [anon_sym___alignof__] = ACTIONS(2568),
    [anon_sym___alignof] = ACTIONS(2568),
    [anon_sym__alignof] = ACTIONS(2568),
    [anon_sym_alignof] = ACTIONS(2568),
    [anon_sym__Alignof] = ACTIONS(2568),
    [anon_sym_offsetof] = ACTIONS(2568),
    [anon_sym__Generic] = ACTIONS(2568),
    [anon_sym_asm] = ACTIONS(2568),
    [anon_sym___asm__] = ACTIONS(2568),
    [anon_sym___asm] = ACTIONS(2568),
    [sym_number_literal] = ACTIONS(2570),
    [anon_sym_L_SQUOTE] = ACTIONS(2570),
    [anon_sym_u_SQUOTE] = ACTIONS(2570),
    [anon_sym_U_SQUOTE] = ACTIONS(2570),
    [anon_sym_u8_SQUOTE] = ACTIONS(2570),
    [anon_sym_SQUOTE] = ACTIONS(2570),
    [anon_sym_AT] = ACTIONS(2568),
    [anon_sym_DQUOTE] = ACTIONS(2570),
    [anon_sym_L_DQUOTE] = ACTIONS(2570),
    [anon_sym_u_DQUOTE] = ACTIONS(2570),
    [anon_sym_U_DQUOTE] = ACTIONS(2570),
    [anon_sym_u8_DQUOTE] = ACTIONS(2570),
    [sym_true] = ACTIONS(2568),
    [sym_false] = ACTIONS(2568),
    [anon_sym_NULL] = ACTIONS(2568),
    [anon_sym_nullptr] = ACTIONS(2568),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2568),
    [anon_sym___typeof] = ACTIONS(2568),
    [anon_sym_typeof] = ACTIONS(2568),
    [anon_sym_ATimport] = ACTIONS(2570),
    [aux_sym_preproc_undef_token1] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2568),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2568),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2568),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2568),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_API_AVAILABLE] = ACTIONS(2568),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_API_DEPRECATED] = ACTIONS(2568),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym___deprecated_msg] = ACTIONS(2568),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2568),
    [anon_sym_ATprotocol] = ACTIONS(2570),
    [anon_sym_ATinterface] = ACTIONS(2570),
    [anon_sym_ATimplementation] = ACTIONS(2570),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2570),
    [anon_sym_ATsynthesize] = ACTIONS(2570),
    [anon_sym_ATdynamic] = ACTIONS(2570),
    [anon_sym_ATtry] = ACTIONS(2570),
    [anon_sym___try] = ACTIONS(2568),
    [anon_sym_ATthrow] = ACTIONS(2570),
    [anon_sym_ATselector] = ACTIONS(2570),
    [anon_sym_ATavailable] = ACTIONS(2570),
    [anon_sym___builtin_available] = ACTIONS(2568),
    [anon_sym_va_arg] = ACTIONS(2568),
    [anon_sym_ATencode] = ACTIONS(2570),
    [anon_sym_ATsynchronized] = ACTIONS(2570),
    [anon_sym_BOOL] = ACTIONS(2568),
    [anon_sym_IMP] = ACTIONS(2568),
    [anon_sym_SEL] = ACTIONS(2568),
    [anon_sym_Class] = ACTIONS(2568),
    [anon_sym_id] = ACTIONS(2568),
  },
  [STATE(393)] = {
    [sym_identifier] = ACTIONS(2478),
    [aux_sym_preproc_include_token1] = ACTIONS(2478),
    [aux_sym_preproc_include_token2] = ACTIONS(2478),
    [aux_sym_preproc_def_token1] = ACTIONS(2478),
    [aux_sym_preproc_if_token1] = ACTIONS(2478),
    [aux_sym_preproc_if_token2] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2478),
    [aux_sym_preproc_else_token1] = ACTIONS(2478),
    [aux_sym_preproc_elif_token1] = ACTIONS(2478),
    [sym_preproc_directive] = ACTIONS(2478),
    [anon_sym_LPAREN2] = ACTIONS(2480),
    [anon_sym_BANG] = ACTIONS(2480),
    [anon_sym_TILDE] = ACTIONS(2480),
    [anon_sym_DASH] = ACTIONS(2478),
    [anon_sym_PLUS] = ACTIONS(2478),
    [anon_sym_STAR] = ACTIONS(2480),
    [anon_sym_CARET] = ACTIONS(2480),
    [anon_sym_AMP] = ACTIONS(2480),
    [anon_sym_SEMI] = ACTIONS(2480),
    [anon_sym___extension__] = ACTIONS(2478),
    [anon_sym_typedef] = ACTIONS(2478),
    [anon_sym_extern] = ACTIONS(2478),
    [anon_sym___attribute__] = ACTIONS(2478),
    [anon_sym___attribute] = ACTIONS(2478),
    [anon_sym_noreturn] = ACTIONS(2478),
    [anon_sym_LBRACK] = ACTIONS(2480),
    [anon_sym___declspec] = ACTIONS(2478),
    [anon_sym___cdecl] = ACTIONS(2478),
    [anon_sym___clrcall] = ACTIONS(2478),
    [anon_sym___stdcall] = ACTIONS(2478),
    [anon_sym___fastcall] = ACTIONS(2478),
    [anon_sym___thiscall] = ACTIONS(2478),
    [anon_sym___vectorcall] = ACTIONS(2478),
    [anon_sym_LBRACE] = ACTIONS(2480),
    [anon_sym_signed] = ACTIONS(2478),
    [anon_sym_unsigned] = ACTIONS(2478),
    [anon_sym_long] = ACTIONS(2478),
    [anon_sym_short] = ACTIONS(2478),
    [anon_sym_static] = ACTIONS(2478),
    [anon_sym_ATautoreleasepool] = ACTIONS(2480),
    [anon_sym_auto] = ACTIONS(2478),
    [anon_sym_register] = ACTIONS(2478),
    [anon_sym_inline] = ACTIONS(2478),
    [anon_sym___inline] = ACTIONS(2478),
    [anon_sym___inline__] = ACTIONS(2478),
    [anon_sym___forceinline] = ACTIONS(2478),
    [anon_sym_thread_local] = ACTIONS(2478),
    [anon_sym___thread] = ACTIONS(2478),
    [anon_sym_CG_EXTERN] = ACTIONS(2478),
    [anon_sym_CG_INLINE] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2478),
    [anon_sym_IBOutlet] = ACTIONS(2478),
    [anon_sym_IBInspectable] = ACTIONS(2478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2478),
    [anon_sym_NS_INLINE] = ACTIONS(2478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2478),
    [anon_sym_const] = ACTIONS(2478),
    [anon_sym_constexpr] = ACTIONS(2478),
    [anon_sym_volatile] = ACTIONS(2478),
    [anon_sym_restrict] = ACTIONS(2478),
    [anon_sym___restrict__] = ACTIONS(2478),
    [anon_sym__Atomic] = ACTIONS(2478),
    [anon_sym__Noreturn] = ACTIONS(2478),
    [anon_sym__Nonnull] = ACTIONS(2478),
    [anon_sym_nullable] = ACTIONS(2478),
    [anon_sym__Complex] = ACTIONS(2478),
    [anon_sym__Nullable] = ACTIONS(2478),
    [anon_sym__Nullable_result] = ACTIONS(2478),
    [anon_sym__Null_unspecified] = ACTIONS(2478),
    [anon_sym___autoreleasing] = ACTIONS(2478),
    [anon_sym___block] = ACTIONS(2478),
    [anon_sym___bridge] = ACTIONS(2478),
    [anon_sym___bridge_retained] = ACTIONS(2478),
    [anon_sym___bridge_transfer] = ACTIONS(2478),
    [anon_sym___complex] = ACTIONS(2478),
    [anon_sym___const] = ACTIONS(2478),
    [anon_sym___imag] = ACTIONS(2478),
    [anon_sym___kindof] = ACTIONS(2478),
    [anon_sym___nonnull] = ACTIONS(2478),
    [anon_sym___nullable] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2478),
    [anon_sym___real] = ACTIONS(2478),
    [anon_sym___strong] = ACTIONS(2478),
    [anon_sym___unsafe_unretained] = ACTIONS(2478),
    [anon_sym___unused] = ACTIONS(2478),
    [anon_sym___weak] = ACTIONS(2478),
    [anon_sym_alignas] = ACTIONS(2478),
    [anon_sym__Alignas] = ACTIONS(2478),
    [sym_primitive_type] = ACTIONS(2478),
    [anon_sym_enum] = ACTIONS(2478),
    [anon_sym_struct] = ACTIONS(2478),
    [anon_sym_union] = ACTIONS(2478),
    [anon_sym_if] = ACTIONS(2478),
    [anon_sym_else] = ACTIONS(2478),
    [anon_sym_switch] = ACTIONS(2478),
    [anon_sym_case] = ACTIONS(2478),
    [anon_sym_default] = ACTIONS(2478),
    [anon_sym_while] = ACTIONS(2478),
    [anon_sym_do] = ACTIONS(2478),
    [anon_sym_for] = ACTIONS(2478),
    [anon_sym_in] = ACTIONS(2478),
    [anon_sym_return] = ACTIONS(2478),
    [anon_sym_break] = ACTIONS(2478),
    [anon_sym_continue] = ACTIONS(2478),
    [anon_sym_goto] = ACTIONS(2478),
    [anon_sym_DASH_DASH] = ACTIONS(2480),
    [anon_sym_PLUS_PLUS] = ACTIONS(2480),
    [anon_sym_sizeof] = ACTIONS(2478),
    [anon_sym___alignof__] = ACTIONS(2478),
    [anon_sym___alignof] = ACTIONS(2478),
    [anon_sym__alignof] = ACTIONS(2478),
    [anon_sym_alignof] = ACTIONS(2478),
    [anon_sym__Alignof] = ACTIONS(2478),
    [anon_sym_offsetof] = ACTIONS(2478),
    [anon_sym__Generic] = ACTIONS(2478),
    [anon_sym_asm] = ACTIONS(2478),
    [anon_sym___asm__] = ACTIONS(2478),
    [anon_sym___asm] = ACTIONS(2478),
    [sym_number_literal] = ACTIONS(2480),
    [anon_sym_L_SQUOTE] = ACTIONS(2480),
    [anon_sym_u_SQUOTE] = ACTIONS(2480),
    [anon_sym_U_SQUOTE] = ACTIONS(2480),
    [anon_sym_u8_SQUOTE] = ACTIONS(2480),
    [anon_sym_SQUOTE] = ACTIONS(2480),
    [anon_sym_AT] = ACTIONS(2478),
    [anon_sym_DQUOTE] = ACTIONS(2480),
    [anon_sym_L_DQUOTE] = ACTIONS(2480),
    [anon_sym_u_DQUOTE] = ACTIONS(2480),
    [anon_sym_U_DQUOTE] = ACTIONS(2480),
    [anon_sym_u8_DQUOTE] = ACTIONS(2480),
    [sym_true] = ACTIONS(2478),
    [sym_false] = ACTIONS(2478),
    [anon_sym_NULL] = ACTIONS(2478),
    [anon_sym_nullptr] = ACTIONS(2478),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2478),
    [anon_sym___typeof] = ACTIONS(2478),
    [anon_sym_typeof] = ACTIONS(2478),
    [anon_sym_ATimport] = ACTIONS(2480),
    [aux_sym_preproc_undef_token1] = ACTIONS(2478),
    [anon_sym_POUND] = ACTIONS(2478),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_API_AVAILABLE] = ACTIONS(2478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_API_DEPRECATED] = ACTIONS(2478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym___deprecated_msg] = ACTIONS(2478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2478),
    [anon_sym_ATprotocol] = ACTIONS(2480),
    [anon_sym_ATinterface] = ACTIONS(2480),
    [anon_sym_ATimplementation] = ACTIONS(2480),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2480),
    [anon_sym_ATsynthesize] = ACTIONS(2480),
    [anon_sym_ATdynamic] = ACTIONS(2480),
    [anon_sym_ATtry] = ACTIONS(2480),
    [anon_sym___try] = ACTIONS(2478),
    [anon_sym_ATthrow] = ACTIONS(2480),
    [anon_sym_ATselector] = ACTIONS(2480),
    [anon_sym_ATavailable] = ACTIONS(2480),
    [anon_sym___builtin_available] = ACTIONS(2478),
    [anon_sym_va_arg] = ACTIONS(2478),
    [anon_sym_ATencode] = ACTIONS(2480),
    [anon_sym_ATsynchronized] = ACTIONS(2480),
    [anon_sym_BOOL] = ACTIONS(2478),
    [anon_sym_IMP] = ACTIONS(2478),
    [anon_sym_SEL] = ACTIONS(2478),
    [anon_sym_Class] = ACTIONS(2478),
    [anon_sym_id] = ACTIONS(2478),
  },
  [STATE(394)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token2] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [aux_sym_preproc_else_token1] = ACTIONS(2554),
    [aux_sym_preproc_elif_token1] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(395)] = {
    [sym_identifier] = ACTIONS(2482),
    [aux_sym_preproc_include_token1] = ACTIONS(2482),
    [aux_sym_preproc_include_token2] = ACTIONS(2482),
    [aux_sym_preproc_def_token1] = ACTIONS(2482),
    [aux_sym_preproc_if_token1] = ACTIONS(2482),
    [aux_sym_preproc_if_token2] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2482),
    [aux_sym_preproc_else_token1] = ACTIONS(2482),
    [aux_sym_preproc_elif_token1] = ACTIONS(2482),
    [sym_preproc_directive] = ACTIONS(2482),
    [anon_sym_LPAREN2] = ACTIONS(2484),
    [anon_sym_BANG] = ACTIONS(2484),
    [anon_sym_TILDE] = ACTIONS(2484),
    [anon_sym_DASH] = ACTIONS(2482),
    [anon_sym_PLUS] = ACTIONS(2482),
    [anon_sym_STAR] = ACTIONS(2484),
    [anon_sym_CARET] = ACTIONS(2484),
    [anon_sym_AMP] = ACTIONS(2484),
    [anon_sym_SEMI] = ACTIONS(2484),
    [anon_sym___extension__] = ACTIONS(2482),
    [anon_sym_typedef] = ACTIONS(2482),
    [anon_sym_extern] = ACTIONS(2482),
    [anon_sym___attribute__] = ACTIONS(2482),
    [anon_sym___attribute] = ACTIONS(2482),
    [anon_sym_noreturn] = ACTIONS(2482),
    [anon_sym_LBRACK] = ACTIONS(2484),
    [anon_sym___declspec] = ACTIONS(2482),
    [anon_sym___cdecl] = ACTIONS(2482),
    [anon_sym___clrcall] = ACTIONS(2482),
    [anon_sym___stdcall] = ACTIONS(2482),
    [anon_sym___fastcall] = ACTIONS(2482),
    [anon_sym___thiscall] = ACTIONS(2482),
    [anon_sym___vectorcall] = ACTIONS(2482),
    [anon_sym_LBRACE] = ACTIONS(2484),
    [anon_sym_signed] = ACTIONS(2482),
    [anon_sym_unsigned] = ACTIONS(2482),
    [anon_sym_long] = ACTIONS(2482),
    [anon_sym_short] = ACTIONS(2482),
    [anon_sym_static] = ACTIONS(2482),
    [anon_sym_ATautoreleasepool] = ACTIONS(2484),
    [anon_sym_auto] = ACTIONS(2482),
    [anon_sym_register] = ACTIONS(2482),
    [anon_sym_inline] = ACTIONS(2482),
    [anon_sym___inline] = ACTIONS(2482),
    [anon_sym___inline__] = ACTIONS(2482),
    [anon_sym___forceinline] = ACTIONS(2482),
    [anon_sym_thread_local] = ACTIONS(2482),
    [anon_sym___thread] = ACTIONS(2482),
    [anon_sym_CG_EXTERN] = ACTIONS(2482),
    [anon_sym_CG_INLINE] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2482),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2482),
    [anon_sym_IBOutlet] = ACTIONS(2482),
    [anon_sym_IBInspectable] = ACTIONS(2482),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2482),
    [anon_sym_NS_INLINE] = ACTIONS(2482),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2482),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2482),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2482),
    [anon_sym_const] = ACTIONS(2482),
    [anon_sym_constexpr] = ACTIONS(2482),
    [anon_sym_volatile] = ACTIONS(2482),
    [anon_sym_restrict] = ACTIONS(2482),
    [anon_sym___restrict__] = ACTIONS(2482),
    [anon_sym__Atomic] = ACTIONS(2482),
    [anon_sym__Noreturn] = ACTIONS(2482),
    [anon_sym__Nonnull] = ACTIONS(2482),
    [anon_sym_nullable] = ACTIONS(2482),
    [anon_sym__Complex] = ACTIONS(2482),
    [anon_sym__Nullable] = ACTIONS(2482),
    [anon_sym__Nullable_result] = ACTIONS(2482),
    [anon_sym__Null_unspecified] = ACTIONS(2482),
    [anon_sym___autoreleasing] = ACTIONS(2482),
    [anon_sym___block] = ACTIONS(2482),
    [anon_sym___bridge] = ACTIONS(2482),
    [anon_sym___bridge_retained] = ACTIONS(2482),
    [anon_sym___bridge_transfer] = ACTIONS(2482),
    [anon_sym___complex] = ACTIONS(2482),
    [anon_sym___const] = ACTIONS(2482),
    [anon_sym___imag] = ACTIONS(2482),
    [anon_sym___kindof] = ACTIONS(2482),
    [anon_sym___nonnull] = ACTIONS(2482),
    [anon_sym___nullable] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2482),
    [anon_sym___real] = ACTIONS(2482),
    [anon_sym___strong] = ACTIONS(2482),
    [anon_sym___unsafe_unretained] = ACTIONS(2482),
    [anon_sym___unused] = ACTIONS(2482),
    [anon_sym___weak] = ACTIONS(2482),
    [anon_sym_alignas] = ACTIONS(2482),
    [anon_sym__Alignas] = ACTIONS(2482),
    [sym_primitive_type] = ACTIONS(2482),
    [anon_sym_enum] = ACTIONS(2482),
    [anon_sym_struct] = ACTIONS(2482),
    [anon_sym_union] = ACTIONS(2482),
    [anon_sym_if] = ACTIONS(2482),
    [anon_sym_else] = ACTIONS(2482),
    [anon_sym_switch] = ACTIONS(2482),
    [anon_sym_case] = ACTIONS(2482),
    [anon_sym_default] = ACTIONS(2482),
    [anon_sym_while] = ACTIONS(2482),
    [anon_sym_do] = ACTIONS(2482),
    [anon_sym_for] = ACTIONS(2482),
    [anon_sym_in] = ACTIONS(2482),
    [anon_sym_return] = ACTIONS(2482),
    [anon_sym_break] = ACTIONS(2482),
    [anon_sym_continue] = ACTIONS(2482),
    [anon_sym_goto] = ACTIONS(2482),
    [anon_sym_DASH_DASH] = ACTIONS(2484),
    [anon_sym_PLUS_PLUS] = ACTIONS(2484),
    [anon_sym_sizeof] = ACTIONS(2482),
    [anon_sym___alignof__] = ACTIONS(2482),
    [anon_sym___alignof] = ACTIONS(2482),
    [anon_sym__alignof] = ACTIONS(2482),
    [anon_sym_alignof] = ACTIONS(2482),
    [anon_sym__Alignof] = ACTIONS(2482),
    [anon_sym_offsetof] = ACTIONS(2482),
    [anon_sym__Generic] = ACTIONS(2482),
    [anon_sym_asm] = ACTIONS(2482),
    [anon_sym___asm__] = ACTIONS(2482),
    [anon_sym___asm] = ACTIONS(2482),
    [sym_number_literal] = ACTIONS(2484),
    [anon_sym_L_SQUOTE] = ACTIONS(2484),
    [anon_sym_u_SQUOTE] = ACTIONS(2484),
    [anon_sym_U_SQUOTE] = ACTIONS(2484),
    [anon_sym_u8_SQUOTE] = ACTIONS(2484),
    [anon_sym_SQUOTE] = ACTIONS(2484),
    [anon_sym_AT] = ACTIONS(2482),
    [anon_sym_DQUOTE] = ACTIONS(2484),
    [anon_sym_L_DQUOTE] = ACTIONS(2484),
    [anon_sym_u_DQUOTE] = ACTIONS(2484),
    [anon_sym_U_DQUOTE] = ACTIONS(2484),
    [anon_sym_u8_DQUOTE] = ACTIONS(2484),
    [sym_true] = ACTIONS(2482),
    [sym_false] = ACTIONS(2482),
    [anon_sym_NULL] = ACTIONS(2482),
    [anon_sym_nullptr] = ACTIONS(2482),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2482),
    [anon_sym___typeof] = ACTIONS(2482),
    [anon_sym_typeof] = ACTIONS(2482),
    [anon_sym_ATimport] = ACTIONS(2484),
    [aux_sym_preproc_undef_token1] = ACTIONS(2482),
    [anon_sym_POUND] = ACTIONS(2482),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2482),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2482),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2482),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_API_AVAILABLE] = ACTIONS(2482),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_API_DEPRECATED] = ACTIONS(2482),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym___deprecated_msg] = ACTIONS(2482),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2482),
    [anon_sym_ATprotocol] = ACTIONS(2484),
    [anon_sym_ATinterface] = ACTIONS(2484),
    [anon_sym_ATimplementation] = ACTIONS(2484),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2484),
    [anon_sym_ATsynthesize] = ACTIONS(2484),
    [anon_sym_ATdynamic] = ACTIONS(2484),
    [anon_sym_ATtry] = ACTIONS(2484),
    [anon_sym___try] = ACTIONS(2482),
    [anon_sym_ATthrow] = ACTIONS(2484),
    [anon_sym_ATselector] = ACTIONS(2484),
    [anon_sym_ATavailable] = ACTIONS(2484),
    [anon_sym___builtin_available] = ACTIONS(2482),
    [anon_sym_va_arg] = ACTIONS(2482),
    [anon_sym_ATencode] = ACTIONS(2484),
    [anon_sym_ATsynchronized] = ACTIONS(2484),
    [anon_sym_BOOL] = ACTIONS(2482),
    [anon_sym_IMP] = ACTIONS(2482),
    [anon_sym_SEL] = ACTIONS(2482),
    [anon_sym_Class] = ACTIONS(2482),
    [anon_sym_id] = ACTIONS(2482),
  },
  [STATE(396)] = {
    [sym_identifier] = ACTIONS(2558),
    [aux_sym_preproc_include_token1] = ACTIONS(2558),
    [aux_sym_preproc_include_token2] = ACTIONS(2558),
    [aux_sym_preproc_def_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token2] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2558),
    [aux_sym_preproc_else_token1] = ACTIONS(2558),
    [aux_sym_preproc_elif_token1] = ACTIONS(2558),
    [sym_preproc_directive] = ACTIONS(2558),
    [anon_sym_LPAREN2] = ACTIONS(2560),
    [anon_sym_BANG] = ACTIONS(2560),
    [anon_sym_TILDE] = ACTIONS(2560),
    [anon_sym_DASH] = ACTIONS(2558),
    [anon_sym_PLUS] = ACTIONS(2558),
    [anon_sym_STAR] = ACTIONS(2560),
    [anon_sym_CARET] = ACTIONS(2560),
    [anon_sym_AMP] = ACTIONS(2560),
    [anon_sym_SEMI] = ACTIONS(2560),
    [anon_sym___extension__] = ACTIONS(2558),
    [anon_sym_typedef] = ACTIONS(2558),
    [anon_sym_extern] = ACTIONS(2558),
    [anon_sym___attribute__] = ACTIONS(2558),
    [anon_sym___attribute] = ACTIONS(2558),
    [anon_sym_noreturn] = ACTIONS(2558),
    [anon_sym_LBRACK] = ACTIONS(2560),
    [anon_sym___declspec] = ACTIONS(2558),
    [anon_sym___cdecl] = ACTIONS(2558),
    [anon_sym___clrcall] = ACTIONS(2558),
    [anon_sym___stdcall] = ACTIONS(2558),
    [anon_sym___fastcall] = ACTIONS(2558),
    [anon_sym___thiscall] = ACTIONS(2558),
    [anon_sym___vectorcall] = ACTIONS(2558),
    [anon_sym_LBRACE] = ACTIONS(2560),
    [anon_sym_signed] = ACTIONS(2558),
    [anon_sym_unsigned] = ACTIONS(2558),
    [anon_sym_long] = ACTIONS(2558),
    [anon_sym_short] = ACTIONS(2558),
    [anon_sym_static] = ACTIONS(2558),
    [anon_sym_ATautoreleasepool] = ACTIONS(2560),
    [anon_sym_auto] = ACTIONS(2558),
    [anon_sym_register] = ACTIONS(2558),
    [anon_sym_inline] = ACTIONS(2558),
    [anon_sym___inline] = ACTIONS(2558),
    [anon_sym___inline__] = ACTIONS(2558),
    [anon_sym___forceinline] = ACTIONS(2558),
    [anon_sym_thread_local] = ACTIONS(2558),
    [anon_sym___thread] = ACTIONS(2558),
    [anon_sym_CG_EXTERN] = ACTIONS(2558),
    [anon_sym_CG_INLINE] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2558),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2558),
    [anon_sym_IBOutlet] = ACTIONS(2558),
    [anon_sym_IBInspectable] = ACTIONS(2558),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2558),
    [anon_sym_NS_INLINE] = ACTIONS(2558),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2558),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2558),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2558),
    [anon_sym_const] = ACTIONS(2558),
    [anon_sym_constexpr] = ACTIONS(2558),
    [anon_sym_volatile] = ACTIONS(2558),
    [anon_sym_restrict] = ACTIONS(2558),
    [anon_sym___restrict__] = ACTIONS(2558),
    [anon_sym__Atomic] = ACTIONS(2558),
    [anon_sym__Noreturn] = ACTIONS(2558),
    [anon_sym__Nonnull] = ACTIONS(2558),
    [anon_sym_nullable] = ACTIONS(2558),
    [anon_sym__Complex] = ACTIONS(2558),
    [anon_sym__Nullable] = ACTIONS(2558),
    [anon_sym__Nullable_result] = ACTIONS(2558),
    [anon_sym__Null_unspecified] = ACTIONS(2558),
    [anon_sym___autoreleasing] = ACTIONS(2558),
    [anon_sym___block] = ACTIONS(2558),
    [anon_sym___bridge] = ACTIONS(2558),
    [anon_sym___bridge_retained] = ACTIONS(2558),
    [anon_sym___bridge_transfer] = ACTIONS(2558),
    [anon_sym___complex] = ACTIONS(2558),
    [anon_sym___const] = ACTIONS(2558),
    [anon_sym___imag] = ACTIONS(2558),
    [anon_sym___kindof] = ACTIONS(2558),
    [anon_sym___nonnull] = ACTIONS(2558),
    [anon_sym___nullable] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2558),
    [anon_sym___real] = ACTIONS(2558),
    [anon_sym___strong] = ACTIONS(2558),
    [anon_sym___unsafe_unretained] = ACTIONS(2558),
    [anon_sym___unused] = ACTIONS(2558),
    [anon_sym___weak] = ACTIONS(2558),
    [anon_sym_alignas] = ACTIONS(2558),
    [anon_sym__Alignas] = ACTIONS(2558),
    [sym_primitive_type] = ACTIONS(2558),
    [anon_sym_enum] = ACTIONS(2558),
    [anon_sym_struct] = ACTIONS(2558),
    [anon_sym_union] = ACTIONS(2558),
    [anon_sym_if] = ACTIONS(2558),
    [anon_sym_else] = ACTIONS(2558),
    [anon_sym_switch] = ACTIONS(2558),
    [anon_sym_case] = ACTIONS(2558),
    [anon_sym_default] = ACTIONS(2558),
    [anon_sym_while] = ACTIONS(2558),
    [anon_sym_do] = ACTIONS(2558),
    [anon_sym_for] = ACTIONS(2558),
    [anon_sym_in] = ACTIONS(2558),
    [anon_sym_return] = ACTIONS(2558),
    [anon_sym_break] = ACTIONS(2558),
    [anon_sym_continue] = ACTIONS(2558),
    [anon_sym_goto] = ACTIONS(2558),
    [anon_sym_DASH_DASH] = ACTIONS(2560),
    [anon_sym_PLUS_PLUS] = ACTIONS(2560),
    [anon_sym_sizeof] = ACTIONS(2558),
    [anon_sym___alignof__] = ACTIONS(2558),
    [anon_sym___alignof] = ACTIONS(2558),
    [anon_sym__alignof] = ACTIONS(2558),
    [anon_sym_alignof] = ACTIONS(2558),
    [anon_sym__Alignof] = ACTIONS(2558),
    [anon_sym_offsetof] = ACTIONS(2558),
    [anon_sym__Generic] = ACTIONS(2558),
    [anon_sym_asm] = ACTIONS(2558),
    [anon_sym___asm__] = ACTIONS(2558),
    [anon_sym___asm] = ACTIONS(2558),
    [sym_number_literal] = ACTIONS(2560),
    [anon_sym_L_SQUOTE] = ACTIONS(2560),
    [anon_sym_u_SQUOTE] = ACTIONS(2560),
    [anon_sym_U_SQUOTE] = ACTIONS(2560),
    [anon_sym_u8_SQUOTE] = ACTIONS(2560),
    [anon_sym_SQUOTE] = ACTIONS(2560),
    [anon_sym_AT] = ACTIONS(2558),
    [anon_sym_DQUOTE] = ACTIONS(2560),
    [anon_sym_L_DQUOTE] = ACTIONS(2560),
    [anon_sym_u_DQUOTE] = ACTIONS(2560),
    [anon_sym_U_DQUOTE] = ACTIONS(2560),
    [anon_sym_u8_DQUOTE] = ACTIONS(2560),
    [sym_true] = ACTIONS(2558),
    [sym_false] = ACTIONS(2558),
    [anon_sym_NULL] = ACTIONS(2558),
    [anon_sym_nullptr] = ACTIONS(2558),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2558),
    [anon_sym___typeof] = ACTIONS(2558),
    [anon_sym_typeof] = ACTIONS(2558),
    [anon_sym_ATimport] = ACTIONS(2560),
    [aux_sym_preproc_undef_token1] = ACTIONS(2558),
    [anon_sym_POUND] = ACTIONS(2558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2558),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2558),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2558),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_API_AVAILABLE] = ACTIONS(2558),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_API_DEPRECATED] = ACTIONS(2558),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym___deprecated_msg] = ACTIONS(2558),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2558),
    [anon_sym_ATprotocol] = ACTIONS(2560),
    [anon_sym_ATinterface] = ACTIONS(2560),
    [anon_sym_ATimplementation] = ACTIONS(2560),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2560),
    [anon_sym_ATsynthesize] = ACTIONS(2560),
    [anon_sym_ATdynamic] = ACTIONS(2560),
    [anon_sym_ATtry] = ACTIONS(2560),
    [anon_sym___try] = ACTIONS(2558),
    [anon_sym_ATthrow] = ACTIONS(2560),
    [anon_sym_ATselector] = ACTIONS(2560),
    [anon_sym_ATavailable] = ACTIONS(2560),
    [anon_sym___builtin_available] = ACTIONS(2558),
    [anon_sym_va_arg] = ACTIONS(2558),
    [anon_sym_ATencode] = ACTIONS(2560),
    [anon_sym_ATsynchronized] = ACTIONS(2560),
    [anon_sym_BOOL] = ACTIONS(2558),
    [anon_sym_IMP] = ACTIONS(2558),
    [anon_sym_SEL] = ACTIONS(2558),
    [anon_sym_Class] = ACTIONS(2558),
    [anon_sym_id] = ACTIONS(2558),
  },
  [STATE(397)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [aux_sym_preproc_else_token1] = ACTIONS(2428),
    [aux_sym_preproc_elif_token1] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(398)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(399)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(400)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(401)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(402)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(403)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(404)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [aux_sym_preproc_else_token1] = ACTIONS(2428),
    [aux_sym_preproc_elif_token1] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(405)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token2] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [aux_sym_preproc_else_token1] = ACTIONS(2442),
    [aux_sym_preproc_elif_token1] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(406)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token2] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [aux_sym_preproc_else_token1] = ACTIONS(2346),
    [aux_sym_preproc_elif_token1] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATsynthesize] = ACTIONS(2348),
    [anon_sym_ATdynamic] = ACTIONS(2348),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(407)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token2] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [aux_sym_preproc_else_token1] = ACTIONS(2350),
    [aux_sym_preproc_elif_token1] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATsynthesize] = ACTIONS(2352),
    [anon_sym_ATdynamic] = ACTIONS(2352),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(408)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token2] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [aux_sym_preproc_else_token1] = ACTIONS(2474),
    [aux_sym_preproc_elif_token1] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(409)] = {
    [sym_identifier] = ACTIONS(2368),
    [aux_sym_preproc_include_token1] = ACTIONS(2368),
    [aux_sym_preproc_include_token2] = ACTIONS(2368),
    [aux_sym_preproc_def_token1] = ACTIONS(2368),
    [aux_sym_preproc_if_token1] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2368),
    [sym_preproc_directive] = ACTIONS(2368),
    [anon_sym_LPAREN2] = ACTIONS(2370),
    [anon_sym_BANG] = ACTIONS(2370),
    [anon_sym_TILDE] = ACTIONS(2370),
    [anon_sym_DASH] = ACTIONS(2368),
    [anon_sym_PLUS] = ACTIONS(2368),
    [anon_sym_STAR] = ACTIONS(2370),
    [anon_sym_CARET] = ACTIONS(2370),
    [anon_sym_AMP] = ACTIONS(2370),
    [anon_sym_SEMI] = ACTIONS(2370),
    [anon_sym___extension__] = ACTIONS(2368),
    [anon_sym_typedef] = ACTIONS(2368),
    [anon_sym_extern] = ACTIONS(2368),
    [anon_sym___attribute__] = ACTIONS(2368),
    [anon_sym___attribute] = ACTIONS(2368),
    [anon_sym_noreturn] = ACTIONS(2368),
    [anon_sym_LBRACK] = ACTIONS(2370),
    [anon_sym___declspec] = ACTIONS(2368),
    [anon_sym___cdecl] = ACTIONS(2368),
    [anon_sym___clrcall] = ACTIONS(2368),
    [anon_sym___stdcall] = ACTIONS(2368),
    [anon_sym___fastcall] = ACTIONS(2368),
    [anon_sym___thiscall] = ACTIONS(2368),
    [anon_sym___vectorcall] = ACTIONS(2368),
    [anon_sym_LBRACE] = ACTIONS(2370),
    [anon_sym_RBRACE] = ACTIONS(2370),
    [anon_sym_signed] = ACTIONS(2368),
    [anon_sym_unsigned] = ACTIONS(2368),
    [anon_sym_long] = ACTIONS(2368),
    [anon_sym_short] = ACTIONS(2368),
    [anon_sym_static] = ACTIONS(2368),
    [anon_sym_ATautoreleasepool] = ACTIONS(2370),
    [anon_sym_auto] = ACTIONS(2368),
    [anon_sym_register] = ACTIONS(2368),
    [anon_sym_inline] = ACTIONS(2368),
    [anon_sym___inline] = ACTIONS(2368),
    [anon_sym___inline__] = ACTIONS(2368),
    [anon_sym___forceinline] = ACTIONS(2368),
    [anon_sym_thread_local] = ACTIONS(2368),
    [anon_sym___thread] = ACTIONS(2368),
    [anon_sym_CG_EXTERN] = ACTIONS(2368),
    [anon_sym_CG_INLINE] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2368),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2368),
    [anon_sym_IBOutlet] = ACTIONS(2368),
    [anon_sym_IBInspectable] = ACTIONS(2368),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2368),
    [anon_sym_NS_INLINE] = ACTIONS(2368),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2368),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2368),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2368),
    [anon_sym_const] = ACTIONS(2368),
    [anon_sym_constexpr] = ACTIONS(2368),
    [anon_sym_volatile] = ACTIONS(2368),
    [anon_sym_restrict] = ACTIONS(2368),
    [anon_sym___restrict__] = ACTIONS(2368),
    [anon_sym__Atomic] = ACTIONS(2368),
    [anon_sym__Noreturn] = ACTIONS(2368),
    [anon_sym__Nonnull] = ACTIONS(2368),
    [anon_sym_nullable] = ACTIONS(2368),
    [anon_sym__Complex] = ACTIONS(2368),
    [anon_sym__Nullable] = ACTIONS(2368),
    [anon_sym__Nullable_result] = ACTIONS(2368),
    [anon_sym__Null_unspecified] = ACTIONS(2368),
    [anon_sym___autoreleasing] = ACTIONS(2368),
    [anon_sym___block] = ACTIONS(2368),
    [anon_sym___bridge] = ACTIONS(2368),
    [anon_sym___bridge_retained] = ACTIONS(2368),
    [anon_sym___bridge_transfer] = ACTIONS(2368),
    [anon_sym___complex] = ACTIONS(2368),
    [anon_sym___const] = ACTIONS(2368),
    [anon_sym___imag] = ACTIONS(2368),
    [anon_sym___kindof] = ACTIONS(2368),
    [anon_sym___nonnull] = ACTIONS(2368),
    [anon_sym___nullable] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2368),
    [anon_sym___real] = ACTIONS(2368),
    [anon_sym___strong] = ACTIONS(2368),
    [anon_sym___unsafe_unretained] = ACTIONS(2368),
    [anon_sym___unused] = ACTIONS(2368),
    [anon_sym___weak] = ACTIONS(2368),
    [anon_sym_alignas] = ACTIONS(2368),
    [anon_sym__Alignas] = ACTIONS(2368),
    [sym_primitive_type] = ACTIONS(2368),
    [anon_sym_enum] = ACTIONS(2368),
    [anon_sym_struct] = ACTIONS(2368),
    [anon_sym_union] = ACTIONS(2368),
    [anon_sym_if] = ACTIONS(2368),
    [anon_sym_else] = ACTIONS(2368),
    [anon_sym_switch] = ACTIONS(2368),
    [anon_sym_case] = ACTIONS(2368),
    [anon_sym_default] = ACTIONS(2368),
    [anon_sym_while] = ACTIONS(2368),
    [anon_sym_do] = ACTIONS(2368),
    [anon_sym_for] = ACTIONS(2368),
    [anon_sym_in] = ACTIONS(2368),
    [anon_sym_return] = ACTIONS(2368),
    [anon_sym_break] = ACTIONS(2368),
    [anon_sym_continue] = ACTIONS(2368),
    [anon_sym_goto] = ACTIONS(2368),
    [anon_sym_DASH_DASH] = ACTIONS(2370),
    [anon_sym_PLUS_PLUS] = ACTIONS(2370),
    [anon_sym_sizeof] = ACTIONS(2368),
    [anon_sym___alignof__] = ACTIONS(2368),
    [anon_sym___alignof] = ACTIONS(2368),
    [anon_sym__alignof] = ACTIONS(2368),
    [anon_sym_alignof] = ACTIONS(2368),
    [anon_sym__Alignof] = ACTIONS(2368),
    [anon_sym_offsetof] = ACTIONS(2368),
    [anon_sym__Generic] = ACTIONS(2368),
    [anon_sym_asm] = ACTIONS(2368),
    [anon_sym___asm__] = ACTIONS(2368),
    [anon_sym___asm] = ACTIONS(2368),
    [sym_number_literal] = ACTIONS(2370),
    [anon_sym_L_SQUOTE] = ACTIONS(2370),
    [anon_sym_u_SQUOTE] = ACTIONS(2370),
    [anon_sym_U_SQUOTE] = ACTIONS(2370),
    [anon_sym_u8_SQUOTE] = ACTIONS(2370),
    [anon_sym_SQUOTE] = ACTIONS(2370),
    [anon_sym_AT] = ACTIONS(2368),
    [anon_sym_DQUOTE] = ACTIONS(2370),
    [anon_sym_L_DQUOTE] = ACTIONS(2370),
    [anon_sym_u_DQUOTE] = ACTIONS(2370),
    [anon_sym_U_DQUOTE] = ACTIONS(2370),
    [anon_sym_u8_DQUOTE] = ACTIONS(2370),
    [sym_true] = ACTIONS(2368),
    [sym_false] = ACTIONS(2368),
    [anon_sym_NULL] = ACTIONS(2368),
    [anon_sym_nullptr] = ACTIONS(2368),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2368),
    [anon_sym___typeof] = ACTIONS(2368),
    [anon_sym_typeof] = ACTIONS(2368),
    [anon_sym_ATimport] = ACTIONS(2370),
    [aux_sym_preproc_undef_token1] = ACTIONS(2368),
    [anon_sym_POUND] = ACTIONS(2368),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2368),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2368),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2368),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_API_AVAILABLE] = ACTIONS(2368),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_API_DEPRECATED] = ACTIONS(2368),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym___deprecated_msg] = ACTIONS(2368),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2368),
    [anon_sym_ATprotocol] = ACTIONS(2370),
    [anon_sym_ATinterface] = ACTIONS(2370),
    [anon_sym_ATimplementation] = ACTIONS(2370),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2370),
    [anon_sym_ATtry] = ACTIONS(2370),
    [anon_sym___try] = ACTIONS(2368),
    [anon_sym_ATcatch] = ACTIONS(2370),
    [anon_sym___catch] = ACTIONS(2368),
    [anon_sym_ATfinally] = ACTIONS(2370),
    [anon_sym___finally] = ACTIONS(2368),
    [anon_sym_ATthrow] = ACTIONS(2370),
    [anon_sym_ATselector] = ACTIONS(2370),
    [anon_sym_ATavailable] = ACTIONS(2370),
    [anon_sym___builtin_available] = ACTIONS(2368),
    [anon_sym_va_arg] = ACTIONS(2368),
    [anon_sym_ATencode] = ACTIONS(2370),
    [anon_sym_ATsynchronized] = ACTIONS(2370),
    [anon_sym_BOOL] = ACTIONS(2368),
    [anon_sym_IMP] = ACTIONS(2368),
    [anon_sym_SEL] = ACTIONS(2368),
    [anon_sym_Class] = ACTIONS(2368),
    [anon_sym_id] = ACTIONS(2368),
  },
  [STATE(410)] = {
    [sym_identifier] = ACTIONS(2486),
    [aux_sym_preproc_include_token1] = ACTIONS(2486),
    [aux_sym_preproc_include_token2] = ACTIONS(2486),
    [aux_sym_preproc_def_token1] = ACTIONS(2486),
    [aux_sym_preproc_if_token1] = ACTIONS(2486),
    [aux_sym_preproc_if_token2] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2486),
    [aux_sym_preproc_else_token1] = ACTIONS(2486),
    [aux_sym_preproc_elif_token1] = ACTIONS(2486),
    [sym_preproc_directive] = ACTIONS(2486),
    [anon_sym_LPAREN2] = ACTIONS(2488),
    [anon_sym_BANG] = ACTIONS(2488),
    [anon_sym_TILDE] = ACTIONS(2488),
    [anon_sym_DASH] = ACTIONS(2486),
    [anon_sym_PLUS] = ACTIONS(2486),
    [anon_sym_STAR] = ACTIONS(2488),
    [anon_sym_CARET] = ACTIONS(2488),
    [anon_sym_AMP] = ACTIONS(2488),
    [anon_sym_SEMI] = ACTIONS(2488),
    [anon_sym___extension__] = ACTIONS(2486),
    [anon_sym_typedef] = ACTIONS(2486),
    [anon_sym_extern] = ACTIONS(2486),
    [anon_sym___attribute__] = ACTIONS(2486),
    [anon_sym___attribute] = ACTIONS(2486),
    [anon_sym_noreturn] = ACTIONS(2486),
    [anon_sym_LBRACK] = ACTIONS(2488),
    [anon_sym___declspec] = ACTIONS(2486),
    [anon_sym___cdecl] = ACTIONS(2486),
    [anon_sym___clrcall] = ACTIONS(2486),
    [anon_sym___stdcall] = ACTIONS(2486),
    [anon_sym___fastcall] = ACTIONS(2486),
    [anon_sym___thiscall] = ACTIONS(2486),
    [anon_sym___vectorcall] = ACTIONS(2486),
    [anon_sym_LBRACE] = ACTIONS(2488),
    [anon_sym_signed] = ACTIONS(2486),
    [anon_sym_unsigned] = ACTIONS(2486),
    [anon_sym_long] = ACTIONS(2486),
    [anon_sym_short] = ACTIONS(2486),
    [anon_sym_static] = ACTIONS(2486),
    [anon_sym_ATautoreleasepool] = ACTIONS(2488),
    [anon_sym_auto] = ACTIONS(2486),
    [anon_sym_register] = ACTIONS(2486),
    [anon_sym_inline] = ACTIONS(2486),
    [anon_sym___inline] = ACTIONS(2486),
    [anon_sym___inline__] = ACTIONS(2486),
    [anon_sym___forceinline] = ACTIONS(2486),
    [anon_sym_thread_local] = ACTIONS(2486),
    [anon_sym___thread] = ACTIONS(2486),
    [anon_sym_CG_EXTERN] = ACTIONS(2486),
    [anon_sym_CG_INLINE] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2486),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2486),
    [anon_sym_IBOutlet] = ACTIONS(2486),
    [anon_sym_IBInspectable] = ACTIONS(2486),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2486),
    [anon_sym_NS_INLINE] = ACTIONS(2486),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2486),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2486),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2486),
    [anon_sym_const] = ACTIONS(2486),
    [anon_sym_constexpr] = ACTIONS(2486),
    [anon_sym_volatile] = ACTIONS(2486),
    [anon_sym_restrict] = ACTIONS(2486),
    [anon_sym___restrict__] = ACTIONS(2486),
    [anon_sym__Atomic] = ACTIONS(2486),
    [anon_sym__Noreturn] = ACTIONS(2486),
    [anon_sym__Nonnull] = ACTIONS(2486),
    [anon_sym_nullable] = ACTIONS(2486),
    [anon_sym__Complex] = ACTIONS(2486),
    [anon_sym__Nullable] = ACTIONS(2486),
    [anon_sym__Nullable_result] = ACTIONS(2486),
    [anon_sym__Null_unspecified] = ACTIONS(2486),
    [anon_sym___autoreleasing] = ACTIONS(2486),
    [anon_sym___block] = ACTIONS(2486),
    [anon_sym___bridge] = ACTIONS(2486),
    [anon_sym___bridge_retained] = ACTIONS(2486),
    [anon_sym___bridge_transfer] = ACTIONS(2486),
    [anon_sym___complex] = ACTIONS(2486),
    [anon_sym___const] = ACTIONS(2486),
    [anon_sym___imag] = ACTIONS(2486),
    [anon_sym___kindof] = ACTIONS(2486),
    [anon_sym___nonnull] = ACTIONS(2486),
    [anon_sym___nullable] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2486),
    [anon_sym___real] = ACTIONS(2486),
    [anon_sym___strong] = ACTIONS(2486),
    [anon_sym___unsafe_unretained] = ACTIONS(2486),
    [anon_sym___unused] = ACTIONS(2486),
    [anon_sym___weak] = ACTIONS(2486),
    [anon_sym_alignas] = ACTIONS(2486),
    [anon_sym__Alignas] = ACTIONS(2486),
    [sym_primitive_type] = ACTIONS(2486),
    [anon_sym_enum] = ACTIONS(2486),
    [anon_sym_struct] = ACTIONS(2486),
    [anon_sym_union] = ACTIONS(2486),
    [anon_sym_if] = ACTIONS(2486),
    [anon_sym_else] = ACTIONS(2486),
    [anon_sym_switch] = ACTIONS(2486),
    [anon_sym_case] = ACTIONS(2486),
    [anon_sym_default] = ACTIONS(2486),
    [anon_sym_while] = ACTIONS(2486),
    [anon_sym_do] = ACTIONS(2486),
    [anon_sym_for] = ACTIONS(2486),
    [anon_sym_in] = ACTIONS(2486),
    [anon_sym_return] = ACTIONS(2486),
    [anon_sym_break] = ACTIONS(2486),
    [anon_sym_continue] = ACTIONS(2486),
    [anon_sym_goto] = ACTIONS(2486),
    [anon_sym_DASH_DASH] = ACTIONS(2488),
    [anon_sym_PLUS_PLUS] = ACTIONS(2488),
    [anon_sym_sizeof] = ACTIONS(2486),
    [anon_sym___alignof__] = ACTIONS(2486),
    [anon_sym___alignof] = ACTIONS(2486),
    [anon_sym__alignof] = ACTIONS(2486),
    [anon_sym_alignof] = ACTIONS(2486),
    [anon_sym__Alignof] = ACTIONS(2486),
    [anon_sym_offsetof] = ACTIONS(2486),
    [anon_sym__Generic] = ACTIONS(2486),
    [anon_sym_asm] = ACTIONS(2486),
    [anon_sym___asm__] = ACTIONS(2486),
    [anon_sym___asm] = ACTIONS(2486),
    [sym_number_literal] = ACTIONS(2488),
    [anon_sym_L_SQUOTE] = ACTIONS(2488),
    [anon_sym_u_SQUOTE] = ACTIONS(2488),
    [anon_sym_U_SQUOTE] = ACTIONS(2488),
    [anon_sym_u8_SQUOTE] = ACTIONS(2488),
    [anon_sym_SQUOTE] = ACTIONS(2488),
    [anon_sym_AT] = ACTIONS(2486),
    [anon_sym_DQUOTE] = ACTIONS(2488),
    [anon_sym_L_DQUOTE] = ACTIONS(2488),
    [anon_sym_u_DQUOTE] = ACTIONS(2488),
    [anon_sym_U_DQUOTE] = ACTIONS(2488),
    [anon_sym_u8_DQUOTE] = ACTIONS(2488),
    [sym_true] = ACTIONS(2486),
    [sym_false] = ACTIONS(2486),
    [anon_sym_NULL] = ACTIONS(2486),
    [anon_sym_nullptr] = ACTIONS(2486),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2486),
    [anon_sym___typeof] = ACTIONS(2486),
    [anon_sym_typeof] = ACTIONS(2486),
    [anon_sym_ATimport] = ACTIONS(2488),
    [aux_sym_preproc_undef_token1] = ACTIONS(2486),
    [anon_sym_POUND] = ACTIONS(2486),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2486),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2486),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2486),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_API_AVAILABLE] = ACTIONS(2486),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_API_DEPRECATED] = ACTIONS(2486),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym___deprecated_msg] = ACTIONS(2486),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2486),
    [anon_sym_ATprotocol] = ACTIONS(2488),
    [anon_sym_ATinterface] = ACTIONS(2488),
    [anon_sym_ATimplementation] = ACTIONS(2488),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2488),
    [anon_sym_ATsynthesize] = ACTIONS(2488),
    [anon_sym_ATdynamic] = ACTIONS(2488),
    [anon_sym_ATtry] = ACTIONS(2488),
    [anon_sym___try] = ACTIONS(2486),
    [anon_sym_ATthrow] = ACTIONS(2488),
    [anon_sym_ATselector] = ACTIONS(2488),
    [anon_sym_ATavailable] = ACTIONS(2488),
    [anon_sym___builtin_available] = ACTIONS(2486),
    [anon_sym_va_arg] = ACTIONS(2486),
    [anon_sym_ATencode] = ACTIONS(2488),
    [anon_sym_ATsynchronized] = ACTIONS(2488),
    [anon_sym_BOOL] = ACTIONS(2486),
    [anon_sym_IMP] = ACTIONS(2486),
    [anon_sym_SEL] = ACTIONS(2486),
    [anon_sym_Class] = ACTIONS(2486),
    [anon_sym_id] = ACTIONS(2486),
  },
  [STATE(411)] = {
    [sym_identifier] = ACTIONS(2490),
    [aux_sym_preproc_include_token1] = ACTIONS(2490),
    [aux_sym_preproc_include_token2] = ACTIONS(2490),
    [aux_sym_preproc_def_token1] = ACTIONS(2490),
    [aux_sym_preproc_if_token1] = ACTIONS(2490),
    [aux_sym_preproc_if_token2] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2490),
    [aux_sym_preproc_else_token1] = ACTIONS(2490),
    [aux_sym_preproc_elif_token1] = ACTIONS(2490),
    [sym_preproc_directive] = ACTIONS(2490),
    [anon_sym_LPAREN2] = ACTIONS(2492),
    [anon_sym_BANG] = ACTIONS(2492),
    [anon_sym_TILDE] = ACTIONS(2492),
    [anon_sym_DASH] = ACTIONS(2490),
    [anon_sym_PLUS] = ACTIONS(2490),
    [anon_sym_STAR] = ACTIONS(2492),
    [anon_sym_CARET] = ACTIONS(2492),
    [anon_sym_AMP] = ACTIONS(2492),
    [anon_sym_SEMI] = ACTIONS(2492),
    [anon_sym___extension__] = ACTIONS(2490),
    [anon_sym_typedef] = ACTIONS(2490),
    [anon_sym_extern] = ACTIONS(2490),
    [anon_sym___attribute__] = ACTIONS(2490),
    [anon_sym___attribute] = ACTIONS(2490),
    [anon_sym_noreturn] = ACTIONS(2490),
    [anon_sym_LBRACK] = ACTIONS(2492),
    [anon_sym___declspec] = ACTIONS(2490),
    [anon_sym___cdecl] = ACTIONS(2490),
    [anon_sym___clrcall] = ACTIONS(2490),
    [anon_sym___stdcall] = ACTIONS(2490),
    [anon_sym___fastcall] = ACTIONS(2490),
    [anon_sym___thiscall] = ACTIONS(2490),
    [anon_sym___vectorcall] = ACTIONS(2490),
    [anon_sym_LBRACE] = ACTIONS(2492),
    [anon_sym_signed] = ACTIONS(2490),
    [anon_sym_unsigned] = ACTIONS(2490),
    [anon_sym_long] = ACTIONS(2490),
    [anon_sym_short] = ACTIONS(2490),
    [anon_sym_static] = ACTIONS(2490),
    [anon_sym_ATautoreleasepool] = ACTIONS(2492),
    [anon_sym_auto] = ACTIONS(2490),
    [anon_sym_register] = ACTIONS(2490),
    [anon_sym_inline] = ACTIONS(2490),
    [anon_sym___inline] = ACTIONS(2490),
    [anon_sym___inline__] = ACTIONS(2490),
    [anon_sym___forceinline] = ACTIONS(2490),
    [anon_sym_thread_local] = ACTIONS(2490),
    [anon_sym___thread] = ACTIONS(2490),
    [anon_sym_CG_EXTERN] = ACTIONS(2490),
    [anon_sym_CG_INLINE] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2490),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2490),
    [anon_sym_IBOutlet] = ACTIONS(2490),
    [anon_sym_IBInspectable] = ACTIONS(2490),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2490),
    [anon_sym_NS_INLINE] = ACTIONS(2490),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2490),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2490),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2490),
    [anon_sym_const] = ACTIONS(2490),
    [anon_sym_constexpr] = ACTIONS(2490),
    [anon_sym_volatile] = ACTIONS(2490),
    [anon_sym_restrict] = ACTIONS(2490),
    [anon_sym___restrict__] = ACTIONS(2490),
    [anon_sym__Atomic] = ACTIONS(2490),
    [anon_sym__Noreturn] = ACTIONS(2490),
    [anon_sym__Nonnull] = ACTIONS(2490),
    [anon_sym_nullable] = ACTIONS(2490),
    [anon_sym__Complex] = ACTIONS(2490),
    [anon_sym__Nullable] = ACTIONS(2490),
    [anon_sym__Nullable_result] = ACTIONS(2490),
    [anon_sym__Null_unspecified] = ACTIONS(2490),
    [anon_sym___autoreleasing] = ACTIONS(2490),
    [anon_sym___block] = ACTIONS(2490),
    [anon_sym___bridge] = ACTIONS(2490),
    [anon_sym___bridge_retained] = ACTIONS(2490),
    [anon_sym___bridge_transfer] = ACTIONS(2490),
    [anon_sym___complex] = ACTIONS(2490),
    [anon_sym___const] = ACTIONS(2490),
    [anon_sym___imag] = ACTIONS(2490),
    [anon_sym___kindof] = ACTIONS(2490),
    [anon_sym___nonnull] = ACTIONS(2490),
    [anon_sym___nullable] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2490),
    [anon_sym___real] = ACTIONS(2490),
    [anon_sym___strong] = ACTIONS(2490),
    [anon_sym___unsafe_unretained] = ACTIONS(2490),
    [anon_sym___unused] = ACTIONS(2490),
    [anon_sym___weak] = ACTIONS(2490),
    [anon_sym_alignas] = ACTIONS(2490),
    [anon_sym__Alignas] = ACTIONS(2490),
    [sym_primitive_type] = ACTIONS(2490),
    [anon_sym_enum] = ACTIONS(2490),
    [anon_sym_struct] = ACTIONS(2490),
    [anon_sym_union] = ACTIONS(2490),
    [anon_sym_if] = ACTIONS(2490),
    [anon_sym_else] = ACTIONS(2490),
    [anon_sym_switch] = ACTIONS(2490),
    [anon_sym_case] = ACTIONS(2490),
    [anon_sym_default] = ACTIONS(2490),
    [anon_sym_while] = ACTIONS(2490),
    [anon_sym_do] = ACTIONS(2490),
    [anon_sym_for] = ACTIONS(2490),
    [anon_sym_in] = ACTIONS(2490),
    [anon_sym_return] = ACTIONS(2490),
    [anon_sym_break] = ACTIONS(2490),
    [anon_sym_continue] = ACTIONS(2490),
    [anon_sym_goto] = ACTIONS(2490),
    [anon_sym_DASH_DASH] = ACTIONS(2492),
    [anon_sym_PLUS_PLUS] = ACTIONS(2492),
    [anon_sym_sizeof] = ACTIONS(2490),
    [anon_sym___alignof__] = ACTIONS(2490),
    [anon_sym___alignof] = ACTIONS(2490),
    [anon_sym__alignof] = ACTIONS(2490),
    [anon_sym_alignof] = ACTIONS(2490),
    [anon_sym__Alignof] = ACTIONS(2490),
    [anon_sym_offsetof] = ACTIONS(2490),
    [anon_sym__Generic] = ACTIONS(2490),
    [anon_sym_asm] = ACTIONS(2490),
    [anon_sym___asm__] = ACTIONS(2490),
    [anon_sym___asm] = ACTIONS(2490),
    [sym_number_literal] = ACTIONS(2492),
    [anon_sym_L_SQUOTE] = ACTIONS(2492),
    [anon_sym_u_SQUOTE] = ACTIONS(2492),
    [anon_sym_U_SQUOTE] = ACTIONS(2492),
    [anon_sym_u8_SQUOTE] = ACTIONS(2492),
    [anon_sym_SQUOTE] = ACTIONS(2492),
    [anon_sym_AT] = ACTIONS(2490),
    [anon_sym_DQUOTE] = ACTIONS(2492),
    [anon_sym_L_DQUOTE] = ACTIONS(2492),
    [anon_sym_u_DQUOTE] = ACTIONS(2492),
    [anon_sym_U_DQUOTE] = ACTIONS(2492),
    [anon_sym_u8_DQUOTE] = ACTIONS(2492),
    [sym_true] = ACTIONS(2490),
    [sym_false] = ACTIONS(2490),
    [anon_sym_NULL] = ACTIONS(2490),
    [anon_sym_nullptr] = ACTIONS(2490),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2490),
    [anon_sym___typeof] = ACTIONS(2490),
    [anon_sym_typeof] = ACTIONS(2490),
    [anon_sym_ATimport] = ACTIONS(2492),
    [aux_sym_preproc_undef_token1] = ACTIONS(2490),
    [anon_sym_POUND] = ACTIONS(2490),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2490),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2490),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2490),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_API_AVAILABLE] = ACTIONS(2490),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_API_DEPRECATED] = ACTIONS(2490),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym___deprecated_msg] = ACTIONS(2490),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2490),
    [anon_sym_ATprotocol] = ACTIONS(2492),
    [anon_sym_ATinterface] = ACTIONS(2492),
    [anon_sym_ATimplementation] = ACTIONS(2492),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2492),
    [anon_sym_ATsynthesize] = ACTIONS(2492),
    [anon_sym_ATdynamic] = ACTIONS(2492),
    [anon_sym_ATtry] = ACTIONS(2492),
    [anon_sym___try] = ACTIONS(2490),
    [anon_sym_ATthrow] = ACTIONS(2492),
    [anon_sym_ATselector] = ACTIONS(2492),
    [anon_sym_ATavailable] = ACTIONS(2492),
    [anon_sym___builtin_available] = ACTIONS(2490),
    [anon_sym_va_arg] = ACTIONS(2490),
    [anon_sym_ATencode] = ACTIONS(2492),
    [anon_sym_ATsynchronized] = ACTIONS(2492),
    [anon_sym_BOOL] = ACTIONS(2490),
    [anon_sym_IMP] = ACTIONS(2490),
    [anon_sym_SEL] = ACTIONS(2490),
    [anon_sym_Class] = ACTIONS(2490),
    [anon_sym_id] = ACTIONS(2490),
  },
  [STATE(412)] = {
    [sym_identifier] = ACTIONS(2494),
    [aux_sym_preproc_include_token1] = ACTIONS(2494),
    [aux_sym_preproc_include_token2] = ACTIONS(2494),
    [aux_sym_preproc_def_token1] = ACTIONS(2494),
    [aux_sym_preproc_if_token1] = ACTIONS(2494),
    [aux_sym_preproc_if_token2] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2494),
    [aux_sym_preproc_else_token1] = ACTIONS(2494),
    [aux_sym_preproc_elif_token1] = ACTIONS(2494),
    [sym_preproc_directive] = ACTIONS(2494),
    [anon_sym_LPAREN2] = ACTIONS(2496),
    [anon_sym_BANG] = ACTIONS(2496),
    [anon_sym_TILDE] = ACTIONS(2496),
    [anon_sym_DASH] = ACTIONS(2494),
    [anon_sym_PLUS] = ACTIONS(2494),
    [anon_sym_STAR] = ACTIONS(2496),
    [anon_sym_CARET] = ACTIONS(2496),
    [anon_sym_AMP] = ACTIONS(2496),
    [anon_sym_SEMI] = ACTIONS(2496),
    [anon_sym___extension__] = ACTIONS(2494),
    [anon_sym_typedef] = ACTIONS(2494),
    [anon_sym_extern] = ACTIONS(2494),
    [anon_sym___attribute__] = ACTIONS(2494),
    [anon_sym___attribute] = ACTIONS(2494),
    [anon_sym_noreturn] = ACTIONS(2494),
    [anon_sym_LBRACK] = ACTIONS(2496),
    [anon_sym___declspec] = ACTIONS(2494),
    [anon_sym___cdecl] = ACTIONS(2494),
    [anon_sym___clrcall] = ACTIONS(2494),
    [anon_sym___stdcall] = ACTIONS(2494),
    [anon_sym___fastcall] = ACTIONS(2494),
    [anon_sym___thiscall] = ACTIONS(2494),
    [anon_sym___vectorcall] = ACTIONS(2494),
    [anon_sym_LBRACE] = ACTIONS(2496),
    [anon_sym_signed] = ACTIONS(2494),
    [anon_sym_unsigned] = ACTIONS(2494),
    [anon_sym_long] = ACTIONS(2494),
    [anon_sym_short] = ACTIONS(2494),
    [anon_sym_static] = ACTIONS(2494),
    [anon_sym_ATautoreleasepool] = ACTIONS(2496),
    [anon_sym_auto] = ACTIONS(2494),
    [anon_sym_register] = ACTIONS(2494),
    [anon_sym_inline] = ACTIONS(2494),
    [anon_sym___inline] = ACTIONS(2494),
    [anon_sym___inline__] = ACTIONS(2494),
    [anon_sym___forceinline] = ACTIONS(2494),
    [anon_sym_thread_local] = ACTIONS(2494),
    [anon_sym___thread] = ACTIONS(2494),
    [anon_sym_CG_EXTERN] = ACTIONS(2494),
    [anon_sym_CG_INLINE] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2494),
    [anon_sym_IBOutlet] = ACTIONS(2494),
    [anon_sym_IBInspectable] = ACTIONS(2494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2494),
    [anon_sym_NS_INLINE] = ACTIONS(2494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2494),
    [anon_sym_const] = ACTIONS(2494),
    [anon_sym_constexpr] = ACTIONS(2494),
    [anon_sym_volatile] = ACTIONS(2494),
    [anon_sym_restrict] = ACTIONS(2494),
    [anon_sym___restrict__] = ACTIONS(2494),
    [anon_sym__Atomic] = ACTIONS(2494),
    [anon_sym__Noreturn] = ACTIONS(2494),
    [anon_sym__Nonnull] = ACTIONS(2494),
    [anon_sym_nullable] = ACTIONS(2494),
    [anon_sym__Complex] = ACTIONS(2494),
    [anon_sym__Nullable] = ACTIONS(2494),
    [anon_sym__Nullable_result] = ACTIONS(2494),
    [anon_sym__Null_unspecified] = ACTIONS(2494),
    [anon_sym___autoreleasing] = ACTIONS(2494),
    [anon_sym___block] = ACTIONS(2494),
    [anon_sym___bridge] = ACTIONS(2494),
    [anon_sym___bridge_retained] = ACTIONS(2494),
    [anon_sym___bridge_transfer] = ACTIONS(2494),
    [anon_sym___complex] = ACTIONS(2494),
    [anon_sym___const] = ACTIONS(2494),
    [anon_sym___imag] = ACTIONS(2494),
    [anon_sym___kindof] = ACTIONS(2494),
    [anon_sym___nonnull] = ACTIONS(2494),
    [anon_sym___nullable] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2494),
    [anon_sym___real] = ACTIONS(2494),
    [anon_sym___strong] = ACTIONS(2494),
    [anon_sym___unsafe_unretained] = ACTIONS(2494),
    [anon_sym___unused] = ACTIONS(2494),
    [anon_sym___weak] = ACTIONS(2494),
    [anon_sym_alignas] = ACTIONS(2494),
    [anon_sym__Alignas] = ACTIONS(2494),
    [sym_primitive_type] = ACTIONS(2494),
    [anon_sym_enum] = ACTIONS(2494),
    [anon_sym_struct] = ACTIONS(2494),
    [anon_sym_union] = ACTIONS(2494),
    [anon_sym_if] = ACTIONS(2494),
    [anon_sym_else] = ACTIONS(2494),
    [anon_sym_switch] = ACTIONS(2494),
    [anon_sym_case] = ACTIONS(2494),
    [anon_sym_default] = ACTIONS(2494),
    [anon_sym_while] = ACTIONS(2494),
    [anon_sym_do] = ACTIONS(2494),
    [anon_sym_for] = ACTIONS(2494),
    [anon_sym_in] = ACTIONS(2494),
    [anon_sym_return] = ACTIONS(2494),
    [anon_sym_break] = ACTIONS(2494),
    [anon_sym_continue] = ACTIONS(2494),
    [anon_sym_goto] = ACTIONS(2494),
    [anon_sym_DASH_DASH] = ACTIONS(2496),
    [anon_sym_PLUS_PLUS] = ACTIONS(2496),
    [anon_sym_sizeof] = ACTIONS(2494),
    [anon_sym___alignof__] = ACTIONS(2494),
    [anon_sym___alignof] = ACTIONS(2494),
    [anon_sym__alignof] = ACTIONS(2494),
    [anon_sym_alignof] = ACTIONS(2494),
    [anon_sym__Alignof] = ACTIONS(2494),
    [anon_sym_offsetof] = ACTIONS(2494),
    [anon_sym__Generic] = ACTIONS(2494),
    [anon_sym_asm] = ACTIONS(2494),
    [anon_sym___asm__] = ACTIONS(2494),
    [anon_sym___asm] = ACTIONS(2494),
    [sym_number_literal] = ACTIONS(2496),
    [anon_sym_L_SQUOTE] = ACTIONS(2496),
    [anon_sym_u_SQUOTE] = ACTIONS(2496),
    [anon_sym_U_SQUOTE] = ACTIONS(2496),
    [anon_sym_u8_SQUOTE] = ACTIONS(2496),
    [anon_sym_SQUOTE] = ACTIONS(2496),
    [anon_sym_AT] = ACTIONS(2494),
    [anon_sym_DQUOTE] = ACTIONS(2496),
    [anon_sym_L_DQUOTE] = ACTIONS(2496),
    [anon_sym_u_DQUOTE] = ACTIONS(2496),
    [anon_sym_U_DQUOTE] = ACTIONS(2496),
    [anon_sym_u8_DQUOTE] = ACTIONS(2496),
    [sym_true] = ACTIONS(2494),
    [sym_false] = ACTIONS(2494),
    [anon_sym_NULL] = ACTIONS(2494),
    [anon_sym_nullptr] = ACTIONS(2494),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2494),
    [anon_sym___typeof] = ACTIONS(2494),
    [anon_sym_typeof] = ACTIONS(2494),
    [anon_sym_ATimport] = ACTIONS(2496),
    [aux_sym_preproc_undef_token1] = ACTIONS(2494),
    [anon_sym_POUND] = ACTIONS(2494),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_API_AVAILABLE] = ACTIONS(2494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_API_DEPRECATED] = ACTIONS(2494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym___deprecated_msg] = ACTIONS(2494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2494),
    [anon_sym_ATprotocol] = ACTIONS(2496),
    [anon_sym_ATinterface] = ACTIONS(2496),
    [anon_sym_ATimplementation] = ACTIONS(2496),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2496),
    [anon_sym_ATsynthesize] = ACTIONS(2496),
    [anon_sym_ATdynamic] = ACTIONS(2496),
    [anon_sym_ATtry] = ACTIONS(2496),
    [anon_sym___try] = ACTIONS(2494),
    [anon_sym_ATthrow] = ACTIONS(2496),
    [anon_sym_ATselector] = ACTIONS(2496),
    [anon_sym_ATavailable] = ACTIONS(2496),
    [anon_sym___builtin_available] = ACTIONS(2494),
    [anon_sym_va_arg] = ACTIONS(2494),
    [anon_sym_ATencode] = ACTIONS(2496),
    [anon_sym_ATsynchronized] = ACTIONS(2496),
    [anon_sym_BOOL] = ACTIONS(2494),
    [anon_sym_IMP] = ACTIONS(2494),
    [anon_sym_SEL] = ACTIONS(2494),
    [anon_sym_Class] = ACTIONS(2494),
    [anon_sym_id] = ACTIONS(2494),
  },
  [STATE(413)] = {
    [sym_identifier] = ACTIONS(2498),
    [aux_sym_preproc_include_token1] = ACTIONS(2498),
    [aux_sym_preproc_include_token2] = ACTIONS(2498),
    [aux_sym_preproc_def_token1] = ACTIONS(2498),
    [aux_sym_preproc_if_token1] = ACTIONS(2498),
    [aux_sym_preproc_if_token2] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2498),
    [aux_sym_preproc_else_token1] = ACTIONS(2498),
    [aux_sym_preproc_elif_token1] = ACTIONS(2498),
    [sym_preproc_directive] = ACTIONS(2498),
    [anon_sym_LPAREN2] = ACTIONS(2500),
    [anon_sym_BANG] = ACTIONS(2500),
    [anon_sym_TILDE] = ACTIONS(2500),
    [anon_sym_DASH] = ACTIONS(2498),
    [anon_sym_PLUS] = ACTIONS(2498),
    [anon_sym_STAR] = ACTIONS(2500),
    [anon_sym_CARET] = ACTIONS(2500),
    [anon_sym_AMP] = ACTIONS(2500),
    [anon_sym_SEMI] = ACTIONS(2500),
    [anon_sym___extension__] = ACTIONS(2498),
    [anon_sym_typedef] = ACTIONS(2498),
    [anon_sym_extern] = ACTIONS(2498),
    [anon_sym___attribute__] = ACTIONS(2498),
    [anon_sym___attribute] = ACTIONS(2498),
    [anon_sym_noreturn] = ACTIONS(2498),
    [anon_sym_LBRACK] = ACTIONS(2500),
    [anon_sym___declspec] = ACTIONS(2498),
    [anon_sym___cdecl] = ACTIONS(2498),
    [anon_sym___clrcall] = ACTIONS(2498),
    [anon_sym___stdcall] = ACTIONS(2498),
    [anon_sym___fastcall] = ACTIONS(2498),
    [anon_sym___thiscall] = ACTIONS(2498),
    [anon_sym___vectorcall] = ACTIONS(2498),
    [anon_sym_LBRACE] = ACTIONS(2500),
    [anon_sym_signed] = ACTIONS(2498),
    [anon_sym_unsigned] = ACTIONS(2498),
    [anon_sym_long] = ACTIONS(2498),
    [anon_sym_short] = ACTIONS(2498),
    [anon_sym_static] = ACTIONS(2498),
    [anon_sym_ATautoreleasepool] = ACTIONS(2500),
    [anon_sym_auto] = ACTIONS(2498),
    [anon_sym_register] = ACTIONS(2498),
    [anon_sym_inline] = ACTIONS(2498),
    [anon_sym___inline] = ACTIONS(2498),
    [anon_sym___inline__] = ACTIONS(2498),
    [anon_sym___forceinline] = ACTIONS(2498),
    [anon_sym_thread_local] = ACTIONS(2498),
    [anon_sym___thread] = ACTIONS(2498),
    [anon_sym_CG_EXTERN] = ACTIONS(2498),
    [anon_sym_CG_INLINE] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2498),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2498),
    [anon_sym_IBOutlet] = ACTIONS(2498),
    [anon_sym_IBInspectable] = ACTIONS(2498),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2498),
    [anon_sym_NS_INLINE] = ACTIONS(2498),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2498),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2498),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2498),
    [anon_sym_const] = ACTIONS(2498),
    [anon_sym_constexpr] = ACTIONS(2498),
    [anon_sym_volatile] = ACTIONS(2498),
    [anon_sym_restrict] = ACTIONS(2498),
    [anon_sym___restrict__] = ACTIONS(2498),
    [anon_sym__Atomic] = ACTIONS(2498),
    [anon_sym__Noreturn] = ACTIONS(2498),
    [anon_sym__Nonnull] = ACTIONS(2498),
    [anon_sym_nullable] = ACTIONS(2498),
    [anon_sym__Complex] = ACTIONS(2498),
    [anon_sym__Nullable] = ACTIONS(2498),
    [anon_sym__Nullable_result] = ACTIONS(2498),
    [anon_sym__Null_unspecified] = ACTIONS(2498),
    [anon_sym___autoreleasing] = ACTIONS(2498),
    [anon_sym___block] = ACTIONS(2498),
    [anon_sym___bridge] = ACTIONS(2498),
    [anon_sym___bridge_retained] = ACTIONS(2498),
    [anon_sym___bridge_transfer] = ACTIONS(2498),
    [anon_sym___complex] = ACTIONS(2498),
    [anon_sym___const] = ACTIONS(2498),
    [anon_sym___imag] = ACTIONS(2498),
    [anon_sym___kindof] = ACTIONS(2498),
    [anon_sym___nonnull] = ACTIONS(2498),
    [anon_sym___nullable] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2498),
    [anon_sym___real] = ACTIONS(2498),
    [anon_sym___strong] = ACTIONS(2498),
    [anon_sym___unsafe_unretained] = ACTIONS(2498),
    [anon_sym___unused] = ACTIONS(2498),
    [anon_sym___weak] = ACTIONS(2498),
    [anon_sym_alignas] = ACTIONS(2498),
    [anon_sym__Alignas] = ACTIONS(2498),
    [sym_primitive_type] = ACTIONS(2498),
    [anon_sym_enum] = ACTIONS(2498),
    [anon_sym_struct] = ACTIONS(2498),
    [anon_sym_union] = ACTIONS(2498),
    [anon_sym_if] = ACTIONS(2498),
    [anon_sym_else] = ACTIONS(2498),
    [anon_sym_switch] = ACTIONS(2498),
    [anon_sym_case] = ACTIONS(2498),
    [anon_sym_default] = ACTIONS(2498),
    [anon_sym_while] = ACTIONS(2498),
    [anon_sym_do] = ACTIONS(2498),
    [anon_sym_for] = ACTIONS(2498),
    [anon_sym_in] = ACTIONS(2498),
    [anon_sym_return] = ACTIONS(2498),
    [anon_sym_break] = ACTIONS(2498),
    [anon_sym_continue] = ACTIONS(2498),
    [anon_sym_goto] = ACTIONS(2498),
    [anon_sym_DASH_DASH] = ACTIONS(2500),
    [anon_sym_PLUS_PLUS] = ACTIONS(2500),
    [anon_sym_sizeof] = ACTIONS(2498),
    [anon_sym___alignof__] = ACTIONS(2498),
    [anon_sym___alignof] = ACTIONS(2498),
    [anon_sym__alignof] = ACTIONS(2498),
    [anon_sym_alignof] = ACTIONS(2498),
    [anon_sym__Alignof] = ACTIONS(2498),
    [anon_sym_offsetof] = ACTIONS(2498),
    [anon_sym__Generic] = ACTIONS(2498),
    [anon_sym_asm] = ACTIONS(2498),
    [anon_sym___asm__] = ACTIONS(2498),
    [anon_sym___asm] = ACTIONS(2498),
    [sym_number_literal] = ACTIONS(2500),
    [anon_sym_L_SQUOTE] = ACTIONS(2500),
    [anon_sym_u_SQUOTE] = ACTIONS(2500),
    [anon_sym_U_SQUOTE] = ACTIONS(2500),
    [anon_sym_u8_SQUOTE] = ACTIONS(2500),
    [anon_sym_SQUOTE] = ACTIONS(2500),
    [anon_sym_AT] = ACTIONS(2498),
    [anon_sym_DQUOTE] = ACTIONS(2500),
    [anon_sym_L_DQUOTE] = ACTIONS(2500),
    [anon_sym_u_DQUOTE] = ACTIONS(2500),
    [anon_sym_U_DQUOTE] = ACTIONS(2500),
    [anon_sym_u8_DQUOTE] = ACTIONS(2500),
    [sym_true] = ACTIONS(2498),
    [sym_false] = ACTIONS(2498),
    [anon_sym_NULL] = ACTIONS(2498),
    [anon_sym_nullptr] = ACTIONS(2498),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2498),
    [anon_sym___typeof] = ACTIONS(2498),
    [anon_sym_typeof] = ACTIONS(2498),
    [anon_sym_ATimport] = ACTIONS(2500),
    [aux_sym_preproc_undef_token1] = ACTIONS(2498),
    [anon_sym_POUND] = ACTIONS(2498),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2498),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2498),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2498),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_API_AVAILABLE] = ACTIONS(2498),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_API_DEPRECATED] = ACTIONS(2498),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym___deprecated_msg] = ACTIONS(2498),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2498),
    [anon_sym_ATprotocol] = ACTIONS(2500),
    [anon_sym_ATinterface] = ACTIONS(2500),
    [anon_sym_ATimplementation] = ACTIONS(2500),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2500),
    [anon_sym_ATsynthesize] = ACTIONS(2500),
    [anon_sym_ATdynamic] = ACTIONS(2500),
    [anon_sym_ATtry] = ACTIONS(2500),
    [anon_sym___try] = ACTIONS(2498),
    [anon_sym_ATthrow] = ACTIONS(2500),
    [anon_sym_ATselector] = ACTIONS(2500),
    [anon_sym_ATavailable] = ACTIONS(2500),
    [anon_sym___builtin_available] = ACTIONS(2498),
    [anon_sym_va_arg] = ACTIONS(2498),
    [anon_sym_ATencode] = ACTIONS(2500),
    [anon_sym_ATsynchronized] = ACTIONS(2500),
    [anon_sym_BOOL] = ACTIONS(2498),
    [anon_sym_IMP] = ACTIONS(2498),
    [anon_sym_SEL] = ACTIONS(2498),
    [anon_sym_Class] = ACTIONS(2498),
    [anon_sym_id] = ACTIONS(2498),
  },
  [STATE(414)] = {
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token2] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [aux_sym_preproc_else_token1] = ACTIONS(2306),
    [aux_sym_preproc_elif_token1] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATsynthesize] = ACTIONS(2308),
    [anon_sym_ATdynamic] = ACTIONS(2308),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(415)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token2] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [aux_sym_preproc_else_token1] = ACTIONS(2474),
    [aux_sym_preproc_elif_token1] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(416)] = {
    [sym_identifier] = ACTIONS(2502),
    [aux_sym_preproc_include_token1] = ACTIONS(2502),
    [aux_sym_preproc_include_token2] = ACTIONS(2502),
    [aux_sym_preproc_def_token1] = ACTIONS(2502),
    [aux_sym_preproc_if_token1] = ACTIONS(2502),
    [aux_sym_preproc_if_token2] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2502),
    [aux_sym_preproc_else_token1] = ACTIONS(2502),
    [aux_sym_preproc_elif_token1] = ACTIONS(2502),
    [sym_preproc_directive] = ACTIONS(2502),
    [anon_sym_LPAREN2] = ACTIONS(2504),
    [anon_sym_BANG] = ACTIONS(2504),
    [anon_sym_TILDE] = ACTIONS(2504),
    [anon_sym_DASH] = ACTIONS(2502),
    [anon_sym_PLUS] = ACTIONS(2502),
    [anon_sym_STAR] = ACTIONS(2504),
    [anon_sym_CARET] = ACTIONS(2504),
    [anon_sym_AMP] = ACTIONS(2504),
    [anon_sym_SEMI] = ACTIONS(2504),
    [anon_sym___extension__] = ACTIONS(2502),
    [anon_sym_typedef] = ACTIONS(2502),
    [anon_sym_extern] = ACTIONS(2502),
    [anon_sym___attribute__] = ACTIONS(2502),
    [anon_sym___attribute] = ACTIONS(2502),
    [anon_sym_noreturn] = ACTIONS(2502),
    [anon_sym_LBRACK] = ACTIONS(2504),
    [anon_sym___declspec] = ACTIONS(2502),
    [anon_sym___cdecl] = ACTIONS(2502),
    [anon_sym___clrcall] = ACTIONS(2502),
    [anon_sym___stdcall] = ACTIONS(2502),
    [anon_sym___fastcall] = ACTIONS(2502),
    [anon_sym___thiscall] = ACTIONS(2502),
    [anon_sym___vectorcall] = ACTIONS(2502),
    [anon_sym_LBRACE] = ACTIONS(2504),
    [anon_sym_signed] = ACTIONS(2502),
    [anon_sym_unsigned] = ACTIONS(2502),
    [anon_sym_long] = ACTIONS(2502),
    [anon_sym_short] = ACTIONS(2502),
    [anon_sym_static] = ACTIONS(2502),
    [anon_sym_ATautoreleasepool] = ACTIONS(2504),
    [anon_sym_auto] = ACTIONS(2502),
    [anon_sym_register] = ACTIONS(2502),
    [anon_sym_inline] = ACTIONS(2502),
    [anon_sym___inline] = ACTIONS(2502),
    [anon_sym___inline__] = ACTIONS(2502),
    [anon_sym___forceinline] = ACTIONS(2502),
    [anon_sym_thread_local] = ACTIONS(2502),
    [anon_sym___thread] = ACTIONS(2502),
    [anon_sym_CG_EXTERN] = ACTIONS(2502),
    [anon_sym_CG_INLINE] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2502),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2502),
    [anon_sym_IBOutlet] = ACTIONS(2502),
    [anon_sym_IBInspectable] = ACTIONS(2502),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2502),
    [anon_sym_NS_INLINE] = ACTIONS(2502),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2502),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2502),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2502),
    [anon_sym_const] = ACTIONS(2502),
    [anon_sym_constexpr] = ACTIONS(2502),
    [anon_sym_volatile] = ACTIONS(2502),
    [anon_sym_restrict] = ACTIONS(2502),
    [anon_sym___restrict__] = ACTIONS(2502),
    [anon_sym__Atomic] = ACTIONS(2502),
    [anon_sym__Noreturn] = ACTIONS(2502),
    [anon_sym__Nonnull] = ACTIONS(2502),
    [anon_sym_nullable] = ACTIONS(2502),
    [anon_sym__Complex] = ACTIONS(2502),
    [anon_sym__Nullable] = ACTIONS(2502),
    [anon_sym__Nullable_result] = ACTIONS(2502),
    [anon_sym__Null_unspecified] = ACTIONS(2502),
    [anon_sym___autoreleasing] = ACTIONS(2502),
    [anon_sym___block] = ACTIONS(2502),
    [anon_sym___bridge] = ACTIONS(2502),
    [anon_sym___bridge_retained] = ACTIONS(2502),
    [anon_sym___bridge_transfer] = ACTIONS(2502),
    [anon_sym___complex] = ACTIONS(2502),
    [anon_sym___const] = ACTIONS(2502),
    [anon_sym___imag] = ACTIONS(2502),
    [anon_sym___kindof] = ACTIONS(2502),
    [anon_sym___nonnull] = ACTIONS(2502),
    [anon_sym___nullable] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2502),
    [anon_sym___real] = ACTIONS(2502),
    [anon_sym___strong] = ACTIONS(2502),
    [anon_sym___unsafe_unretained] = ACTIONS(2502),
    [anon_sym___unused] = ACTIONS(2502),
    [anon_sym___weak] = ACTIONS(2502),
    [anon_sym_alignas] = ACTIONS(2502),
    [anon_sym__Alignas] = ACTIONS(2502),
    [sym_primitive_type] = ACTIONS(2502),
    [anon_sym_enum] = ACTIONS(2502),
    [anon_sym_struct] = ACTIONS(2502),
    [anon_sym_union] = ACTIONS(2502),
    [anon_sym_if] = ACTIONS(2502),
    [anon_sym_else] = ACTIONS(2502),
    [anon_sym_switch] = ACTIONS(2502),
    [anon_sym_case] = ACTIONS(2502),
    [anon_sym_default] = ACTIONS(2502),
    [anon_sym_while] = ACTIONS(2502),
    [anon_sym_do] = ACTIONS(2502),
    [anon_sym_for] = ACTIONS(2502),
    [anon_sym_in] = ACTIONS(2502),
    [anon_sym_return] = ACTIONS(2502),
    [anon_sym_break] = ACTIONS(2502),
    [anon_sym_continue] = ACTIONS(2502),
    [anon_sym_goto] = ACTIONS(2502),
    [anon_sym_DASH_DASH] = ACTIONS(2504),
    [anon_sym_PLUS_PLUS] = ACTIONS(2504),
    [anon_sym_sizeof] = ACTIONS(2502),
    [anon_sym___alignof__] = ACTIONS(2502),
    [anon_sym___alignof] = ACTIONS(2502),
    [anon_sym__alignof] = ACTIONS(2502),
    [anon_sym_alignof] = ACTIONS(2502),
    [anon_sym__Alignof] = ACTIONS(2502),
    [anon_sym_offsetof] = ACTIONS(2502),
    [anon_sym__Generic] = ACTIONS(2502),
    [anon_sym_asm] = ACTIONS(2502),
    [anon_sym___asm__] = ACTIONS(2502),
    [anon_sym___asm] = ACTIONS(2502),
    [sym_number_literal] = ACTIONS(2504),
    [anon_sym_L_SQUOTE] = ACTIONS(2504),
    [anon_sym_u_SQUOTE] = ACTIONS(2504),
    [anon_sym_U_SQUOTE] = ACTIONS(2504),
    [anon_sym_u8_SQUOTE] = ACTIONS(2504),
    [anon_sym_SQUOTE] = ACTIONS(2504),
    [anon_sym_AT] = ACTIONS(2502),
    [anon_sym_DQUOTE] = ACTIONS(2504),
    [anon_sym_L_DQUOTE] = ACTIONS(2504),
    [anon_sym_u_DQUOTE] = ACTIONS(2504),
    [anon_sym_U_DQUOTE] = ACTIONS(2504),
    [anon_sym_u8_DQUOTE] = ACTIONS(2504),
    [sym_true] = ACTIONS(2502),
    [sym_false] = ACTIONS(2502),
    [anon_sym_NULL] = ACTIONS(2502),
    [anon_sym_nullptr] = ACTIONS(2502),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2502),
    [anon_sym___typeof] = ACTIONS(2502),
    [anon_sym_typeof] = ACTIONS(2502),
    [anon_sym_ATimport] = ACTIONS(2504),
    [aux_sym_preproc_undef_token1] = ACTIONS(2502),
    [anon_sym_POUND] = ACTIONS(2502),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2502),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2502),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2502),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_API_AVAILABLE] = ACTIONS(2502),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_API_DEPRECATED] = ACTIONS(2502),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym___deprecated_msg] = ACTIONS(2502),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2502),
    [anon_sym_ATprotocol] = ACTIONS(2504),
    [anon_sym_ATinterface] = ACTIONS(2504),
    [anon_sym_ATimplementation] = ACTIONS(2504),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2504),
    [anon_sym_ATsynthesize] = ACTIONS(2504),
    [anon_sym_ATdynamic] = ACTIONS(2504),
    [anon_sym_ATtry] = ACTIONS(2504),
    [anon_sym___try] = ACTIONS(2502),
    [anon_sym_ATthrow] = ACTIONS(2504),
    [anon_sym_ATselector] = ACTIONS(2504),
    [anon_sym_ATavailable] = ACTIONS(2504),
    [anon_sym___builtin_available] = ACTIONS(2502),
    [anon_sym_va_arg] = ACTIONS(2502),
    [anon_sym_ATencode] = ACTIONS(2504),
    [anon_sym_ATsynchronized] = ACTIONS(2504),
    [anon_sym_BOOL] = ACTIONS(2502),
    [anon_sym_IMP] = ACTIONS(2502),
    [anon_sym_SEL] = ACTIONS(2502),
    [anon_sym_Class] = ACTIONS(2502),
    [anon_sym_id] = ACTIONS(2502),
  },
  [STATE(417)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token2] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [aux_sym_preproc_else_token1] = ACTIONS(2360),
    [aux_sym_preproc_elif_token1] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATsynthesize] = ACTIONS(2362),
    [anon_sym_ATdynamic] = ACTIONS(2362),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(418)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [aux_sym_preproc_else_token1] = ACTIONS(2428),
    [aux_sym_preproc_elif_token1] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(419)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token2] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [aux_sym_preproc_else_token1] = ACTIONS(2416),
    [aux_sym_preproc_elif_token1] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(420)] = {
    [sym_identifier] = ACTIONS(2506),
    [aux_sym_preproc_include_token1] = ACTIONS(2506),
    [aux_sym_preproc_include_token2] = ACTIONS(2506),
    [aux_sym_preproc_def_token1] = ACTIONS(2506),
    [aux_sym_preproc_if_token1] = ACTIONS(2506),
    [aux_sym_preproc_if_token2] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2506),
    [aux_sym_preproc_else_token1] = ACTIONS(2506),
    [aux_sym_preproc_elif_token1] = ACTIONS(2506),
    [sym_preproc_directive] = ACTIONS(2506),
    [anon_sym_LPAREN2] = ACTIONS(2508),
    [anon_sym_BANG] = ACTIONS(2508),
    [anon_sym_TILDE] = ACTIONS(2508),
    [anon_sym_DASH] = ACTIONS(2506),
    [anon_sym_PLUS] = ACTIONS(2506),
    [anon_sym_STAR] = ACTIONS(2508),
    [anon_sym_CARET] = ACTIONS(2508),
    [anon_sym_AMP] = ACTIONS(2508),
    [anon_sym_SEMI] = ACTIONS(2508),
    [anon_sym___extension__] = ACTIONS(2506),
    [anon_sym_typedef] = ACTIONS(2506),
    [anon_sym_extern] = ACTIONS(2506),
    [anon_sym___attribute__] = ACTIONS(2506),
    [anon_sym___attribute] = ACTIONS(2506),
    [anon_sym_noreturn] = ACTIONS(2506),
    [anon_sym_LBRACK] = ACTIONS(2508),
    [anon_sym___declspec] = ACTIONS(2506),
    [anon_sym___cdecl] = ACTIONS(2506),
    [anon_sym___clrcall] = ACTIONS(2506),
    [anon_sym___stdcall] = ACTIONS(2506),
    [anon_sym___fastcall] = ACTIONS(2506),
    [anon_sym___thiscall] = ACTIONS(2506),
    [anon_sym___vectorcall] = ACTIONS(2506),
    [anon_sym_LBRACE] = ACTIONS(2508),
    [anon_sym_signed] = ACTIONS(2506),
    [anon_sym_unsigned] = ACTIONS(2506),
    [anon_sym_long] = ACTIONS(2506),
    [anon_sym_short] = ACTIONS(2506),
    [anon_sym_static] = ACTIONS(2506),
    [anon_sym_ATautoreleasepool] = ACTIONS(2508),
    [anon_sym_auto] = ACTIONS(2506),
    [anon_sym_register] = ACTIONS(2506),
    [anon_sym_inline] = ACTIONS(2506),
    [anon_sym___inline] = ACTIONS(2506),
    [anon_sym___inline__] = ACTIONS(2506),
    [anon_sym___forceinline] = ACTIONS(2506),
    [anon_sym_thread_local] = ACTIONS(2506),
    [anon_sym___thread] = ACTIONS(2506),
    [anon_sym_CG_EXTERN] = ACTIONS(2506),
    [anon_sym_CG_INLINE] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2506),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2506),
    [anon_sym_IBOutlet] = ACTIONS(2506),
    [anon_sym_IBInspectable] = ACTIONS(2506),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2506),
    [anon_sym_NS_INLINE] = ACTIONS(2506),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2506),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2506),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2506),
    [anon_sym_const] = ACTIONS(2506),
    [anon_sym_constexpr] = ACTIONS(2506),
    [anon_sym_volatile] = ACTIONS(2506),
    [anon_sym_restrict] = ACTIONS(2506),
    [anon_sym___restrict__] = ACTIONS(2506),
    [anon_sym__Atomic] = ACTIONS(2506),
    [anon_sym__Noreturn] = ACTIONS(2506),
    [anon_sym__Nonnull] = ACTIONS(2506),
    [anon_sym_nullable] = ACTIONS(2506),
    [anon_sym__Complex] = ACTIONS(2506),
    [anon_sym__Nullable] = ACTIONS(2506),
    [anon_sym__Nullable_result] = ACTIONS(2506),
    [anon_sym__Null_unspecified] = ACTIONS(2506),
    [anon_sym___autoreleasing] = ACTIONS(2506),
    [anon_sym___block] = ACTIONS(2506),
    [anon_sym___bridge] = ACTIONS(2506),
    [anon_sym___bridge_retained] = ACTIONS(2506),
    [anon_sym___bridge_transfer] = ACTIONS(2506),
    [anon_sym___complex] = ACTIONS(2506),
    [anon_sym___const] = ACTIONS(2506),
    [anon_sym___imag] = ACTIONS(2506),
    [anon_sym___kindof] = ACTIONS(2506),
    [anon_sym___nonnull] = ACTIONS(2506),
    [anon_sym___nullable] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2506),
    [anon_sym___real] = ACTIONS(2506),
    [anon_sym___strong] = ACTIONS(2506),
    [anon_sym___unsafe_unretained] = ACTIONS(2506),
    [anon_sym___unused] = ACTIONS(2506),
    [anon_sym___weak] = ACTIONS(2506),
    [anon_sym_alignas] = ACTIONS(2506),
    [anon_sym__Alignas] = ACTIONS(2506),
    [sym_primitive_type] = ACTIONS(2506),
    [anon_sym_enum] = ACTIONS(2506),
    [anon_sym_struct] = ACTIONS(2506),
    [anon_sym_union] = ACTIONS(2506),
    [anon_sym_if] = ACTIONS(2506),
    [anon_sym_else] = ACTIONS(2506),
    [anon_sym_switch] = ACTIONS(2506),
    [anon_sym_case] = ACTIONS(2506),
    [anon_sym_default] = ACTIONS(2506),
    [anon_sym_while] = ACTIONS(2506),
    [anon_sym_do] = ACTIONS(2506),
    [anon_sym_for] = ACTIONS(2506),
    [anon_sym_in] = ACTIONS(2506),
    [anon_sym_return] = ACTIONS(2506),
    [anon_sym_break] = ACTIONS(2506),
    [anon_sym_continue] = ACTIONS(2506),
    [anon_sym_goto] = ACTIONS(2506),
    [anon_sym_DASH_DASH] = ACTIONS(2508),
    [anon_sym_PLUS_PLUS] = ACTIONS(2508),
    [anon_sym_sizeof] = ACTIONS(2506),
    [anon_sym___alignof__] = ACTIONS(2506),
    [anon_sym___alignof] = ACTIONS(2506),
    [anon_sym__alignof] = ACTIONS(2506),
    [anon_sym_alignof] = ACTIONS(2506),
    [anon_sym__Alignof] = ACTIONS(2506),
    [anon_sym_offsetof] = ACTIONS(2506),
    [anon_sym__Generic] = ACTIONS(2506),
    [anon_sym_asm] = ACTIONS(2506),
    [anon_sym___asm__] = ACTIONS(2506),
    [anon_sym___asm] = ACTIONS(2506),
    [sym_number_literal] = ACTIONS(2508),
    [anon_sym_L_SQUOTE] = ACTIONS(2508),
    [anon_sym_u_SQUOTE] = ACTIONS(2508),
    [anon_sym_U_SQUOTE] = ACTIONS(2508),
    [anon_sym_u8_SQUOTE] = ACTIONS(2508),
    [anon_sym_SQUOTE] = ACTIONS(2508),
    [anon_sym_AT] = ACTIONS(2506),
    [anon_sym_DQUOTE] = ACTIONS(2508),
    [anon_sym_L_DQUOTE] = ACTIONS(2508),
    [anon_sym_u_DQUOTE] = ACTIONS(2508),
    [anon_sym_U_DQUOTE] = ACTIONS(2508),
    [anon_sym_u8_DQUOTE] = ACTIONS(2508),
    [sym_true] = ACTIONS(2506),
    [sym_false] = ACTIONS(2506),
    [anon_sym_NULL] = ACTIONS(2506),
    [anon_sym_nullptr] = ACTIONS(2506),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2506),
    [anon_sym___typeof] = ACTIONS(2506),
    [anon_sym_typeof] = ACTIONS(2506),
    [anon_sym_ATimport] = ACTIONS(2508),
    [aux_sym_preproc_undef_token1] = ACTIONS(2506),
    [anon_sym_POUND] = ACTIONS(2506),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2506),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2506),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2506),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_API_AVAILABLE] = ACTIONS(2506),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_API_DEPRECATED] = ACTIONS(2506),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym___deprecated_msg] = ACTIONS(2506),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2506),
    [anon_sym_ATprotocol] = ACTIONS(2508),
    [anon_sym_ATinterface] = ACTIONS(2508),
    [anon_sym_ATimplementation] = ACTIONS(2508),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2508),
    [anon_sym_ATsynthesize] = ACTIONS(2508),
    [anon_sym_ATdynamic] = ACTIONS(2508),
    [anon_sym_ATtry] = ACTIONS(2508),
    [anon_sym___try] = ACTIONS(2506),
    [anon_sym_ATthrow] = ACTIONS(2508),
    [anon_sym_ATselector] = ACTIONS(2508),
    [anon_sym_ATavailable] = ACTIONS(2508),
    [anon_sym___builtin_available] = ACTIONS(2506),
    [anon_sym_va_arg] = ACTIONS(2506),
    [anon_sym_ATencode] = ACTIONS(2508),
    [anon_sym_ATsynchronized] = ACTIONS(2508),
    [anon_sym_BOOL] = ACTIONS(2506),
    [anon_sym_IMP] = ACTIONS(2506),
    [anon_sym_SEL] = ACTIONS(2506),
    [anon_sym_Class] = ACTIONS(2506),
    [anon_sym_id] = ACTIONS(2506),
  },
  [STATE(421)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token2] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [aux_sym_preproc_else_token1] = ACTIONS(2420),
    [aux_sym_preproc_elif_token1] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(422)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token2] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [aux_sym_preproc_else_token1] = ACTIONS(2424),
    [aux_sym_preproc_elif_token1] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(423)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token2] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [aux_sym_preproc_else_token1] = ACTIONS(2416),
    [aux_sym_preproc_elif_token1] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(424)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(425)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token2] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [aux_sym_preproc_else_token1] = ACTIONS(2424),
    [aux_sym_preproc_elif_token1] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(426)] = {
    [sym_identifier] = ACTIONS(2514),
    [aux_sym_preproc_include_token1] = ACTIONS(2514),
    [aux_sym_preproc_include_token2] = ACTIONS(2514),
    [aux_sym_preproc_def_token1] = ACTIONS(2514),
    [aux_sym_preproc_if_token1] = ACTIONS(2514),
    [aux_sym_preproc_if_token2] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2514),
    [aux_sym_preproc_else_token1] = ACTIONS(2514),
    [aux_sym_preproc_elif_token1] = ACTIONS(2514),
    [sym_preproc_directive] = ACTIONS(2514),
    [anon_sym_LPAREN2] = ACTIONS(2516),
    [anon_sym_BANG] = ACTIONS(2516),
    [anon_sym_TILDE] = ACTIONS(2516),
    [anon_sym_DASH] = ACTIONS(2514),
    [anon_sym_PLUS] = ACTIONS(2514),
    [anon_sym_STAR] = ACTIONS(2516),
    [anon_sym_CARET] = ACTIONS(2516),
    [anon_sym_AMP] = ACTIONS(2516),
    [anon_sym_SEMI] = ACTIONS(2516),
    [anon_sym___extension__] = ACTIONS(2514),
    [anon_sym_typedef] = ACTIONS(2514),
    [anon_sym_extern] = ACTIONS(2514),
    [anon_sym___attribute__] = ACTIONS(2514),
    [anon_sym___attribute] = ACTIONS(2514),
    [anon_sym_noreturn] = ACTIONS(2514),
    [anon_sym_LBRACK] = ACTIONS(2516),
    [anon_sym___declspec] = ACTIONS(2514),
    [anon_sym___cdecl] = ACTIONS(2514),
    [anon_sym___clrcall] = ACTIONS(2514),
    [anon_sym___stdcall] = ACTIONS(2514),
    [anon_sym___fastcall] = ACTIONS(2514),
    [anon_sym___thiscall] = ACTIONS(2514),
    [anon_sym___vectorcall] = ACTIONS(2514),
    [anon_sym_LBRACE] = ACTIONS(2516),
    [anon_sym_signed] = ACTIONS(2514),
    [anon_sym_unsigned] = ACTIONS(2514),
    [anon_sym_long] = ACTIONS(2514),
    [anon_sym_short] = ACTIONS(2514),
    [anon_sym_static] = ACTIONS(2514),
    [anon_sym_ATautoreleasepool] = ACTIONS(2516),
    [anon_sym_auto] = ACTIONS(2514),
    [anon_sym_register] = ACTIONS(2514),
    [anon_sym_inline] = ACTIONS(2514),
    [anon_sym___inline] = ACTIONS(2514),
    [anon_sym___inline__] = ACTIONS(2514),
    [anon_sym___forceinline] = ACTIONS(2514),
    [anon_sym_thread_local] = ACTIONS(2514),
    [anon_sym___thread] = ACTIONS(2514),
    [anon_sym_CG_EXTERN] = ACTIONS(2514),
    [anon_sym_CG_INLINE] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2514),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2514),
    [anon_sym_IBOutlet] = ACTIONS(2514),
    [anon_sym_IBInspectable] = ACTIONS(2514),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2514),
    [anon_sym_NS_INLINE] = ACTIONS(2514),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2514),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2514),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2514),
    [anon_sym_const] = ACTIONS(2514),
    [anon_sym_constexpr] = ACTIONS(2514),
    [anon_sym_volatile] = ACTIONS(2514),
    [anon_sym_restrict] = ACTIONS(2514),
    [anon_sym___restrict__] = ACTIONS(2514),
    [anon_sym__Atomic] = ACTIONS(2514),
    [anon_sym__Noreturn] = ACTIONS(2514),
    [anon_sym__Nonnull] = ACTIONS(2514),
    [anon_sym_nullable] = ACTIONS(2514),
    [anon_sym__Complex] = ACTIONS(2514),
    [anon_sym__Nullable] = ACTIONS(2514),
    [anon_sym__Nullable_result] = ACTIONS(2514),
    [anon_sym__Null_unspecified] = ACTIONS(2514),
    [anon_sym___autoreleasing] = ACTIONS(2514),
    [anon_sym___block] = ACTIONS(2514),
    [anon_sym___bridge] = ACTIONS(2514),
    [anon_sym___bridge_retained] = ACTIONS(2514),
    [anon_sym___bridge_transfer] = ACTIONS(2514),
    [anon_sym___complex] = ACTIONS(2514),
    [anon_sym___const] = ACTIONS(2514),
    [anon_sym___imag] = ACTIONS(2514),
    [anon_sym___kindof] = ACTIONS(2514),
    [anon_sym___nonnull] = ACTIONS(2514),
    [anon_sym___nullable] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2514),
    [anon_sym___real] = ACTIONS(2514),
    [anon_sym___strong] = ACTIONS(2514),
    [anon_sym___unsafe_unretained] = ACTIONS(2514),
    [anon_sym___unused] = ACTIONS(2514),
    [anon_sym___weak] = ACTIONS(2514),
    [anon_sym_alignas] = ACTIONS(2514),
    [anon_sym__Alignas] = ACTIONS(2514),
    [sym_primitive_type] = ACTIONS(2514),
    [anon_sym_enum] = ACTIONS(2514),
    [anon_sym_struct] = ACTIONS(2514),
    [anon_sym_union] = ACTIONS(2514),
    [anon_sym_if] = ACTIONS(2514),
    [anon_sym_else] = ACTIONS(2514),
    [anon_sym_switch] = ACTIONS(2514),
    [anon_sym_case] = ACTIONS(2514),
    [anon_sym_default] = ACTIONS(2514),
    [anon_sym_while] = ACTIONS(2514),
    [anon_sym_do] = ACTIONS(2514),
    [anon_sym_for] = ACTIONS(2514),
    [anon_sym_in] = ACTIONS(2514),
    [anon_sym_return] = ACTIONS(2514),
    [anon_sym_break] = ACTIONS(2514),
    [anon_sym_continue] = ACTIONS(2514),
    [anon_sym_goto] = ACTIONS(2514),
    [anon_sym_DASH_DASH] = ACTIONS(2516),
    [anon_sym_PLUS_PLUS] = ACTIONS(2516),
    [anon_sym_sizeof] = ACTIONS(2514),
    [anon_sym___alignof__] = ACTIONS(2514),
    [anon_sym___alignof] = ACTIONS(2514),
    [anon_sym__alignof] = ACTIONS(2514),
    [anon_sym_alignof] = ACTIONS(2514),
    [anon_sym__Alignof] = ACTIONS(2514),
    [anon_sym_offsetof] = ACTIONS(2514),
    [anon_sym__Generic] = ACTIONS(2514),
    [anon_sym_asm] = ACTIONS(2514),
    [anon_sym___asm__] = ACTIONS(2514),
    [anon_sym___asm] = ACTIONS(2514),
    [sym_number_literal] = ACTIONS(2516),
    [anon_sym_L_SQUOTE] = ACTIONS(2516),
    [anon_sym_u_SQUOTE] = ACTIONS(2516),
    [anon_sym_U_SQUOTE] = ACTIONS(2516),
    [anon_sym_u8_SQUOTE] = ACTIONS(2516),
    [anon_sym_SQUOTE] = ACTIONS(2516),
    [anon_sym_AT] = ACTIONS(2514),
    [anon_sym_DQUOTE] = ACTIONS(2516),
    [anon_sym_L_DQUOTE] = ACTIONS(2516),
    [anon_sym_u_DQUOTE] = ACTIONS(2516),
    [anon_sym_U_DQUOTE] = ACTIONS(2516),
    [anon_sym_u8_DQUOTE] = ACTIONS(2516),
    [sym_true] = ACTIONS(2514),
    [sym_false] = ACTIONS(2514),
    [anon_sym_NULL] = ACTIONS(2514),
    [anon_sym_nullptr] = ACTIONS(2514),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2514),
    [anon_sym___typeof] = ACTIONS(2514),
    [anon_sym_typeof] = ACTIONS(2514),
    [anon_sym_ATimport] = ACTIONS(2516),
    [aux_sym_preproc_undef_token1] = ACTIONS(2514),
    [anon_sym_POUND] = ACTIONS(2514),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2514),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2514),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2514),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_API_AVAILABLE] = ACTIONS(2514),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_API_DEPRECATED] = ACTIONS(2514),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym___deprecated_msg] = ACTIONS(2514),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2514),
    [anon_sym_ATprotocol] = ACTIONS(2516),
    [anon_sym_ATinterface] = ACTIONS(2516),
    [anon_sym_ATimplementation] = ACTIONS(2516),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2516),
    [anon_sym_ATsynthesize] = ACTIONS(2516),
    [anon_sym_ATdynamic] = ACTIONS(2516),
    [anon_sym_ATtry] = ACTIONS(2516),
    [anon_sym___try] = ACTIONS(2514),
    [anon_sym_ATthrow] = ACTIONS(2516),
    [anon_sym_ATselector] = ACTIONS(2516),
    [anon_sym_ATavailable] = ACTIONS(2516),
    [anon_sym___builtin_available] = ACTIONS(2514),
    [anon_sym_va_arg] = ACTIONS(2514),
    [anon_sym_ATencode] = ACTIONS(2516),
    [anon_sym_ATsynchronized] = ACTIONS(2516),
    [anon_sym_BOOL] = ACTIONS(2514),
    [anon_sym_IMP] = ACTIONS(2514),
    [anon_sym_SEL] = ACTIONS(2514),
    [anon_sym_Class] = ACTIONS(2514),
    [anon_sym_id] = ACTIONS(2514),
  },
  [STATE(427)] = {
    [ts_builtin_sym_end] = ACTIONS(2366),
    [sym_identifier] = ACTIONS(2364),
    [aux_sym_preproc_include_token1] = ACTIONS(2364),
    [aux_sym_preproc_include_token2] = ACTIONS(2364),
    [aux_sym_preproc_def_token1] = ACTIONS(2364),
    [aux_sym_preproc_if_token1] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2364),
    [sym_preproc_directive] = ACTIONS(2364),
    [anon_sym_LPAREN2] = ACTIONS(2366),
    [anon_sym_BANG] = ACTIONS(2366),
    [anon_sym_TILDE] = ACTIONS(2366),
    [anon_sym_DASH] = ACTIONS(2364),
    [anon_sym_PLUS] = ACTIONS(2364),
    [anon_sym_STAR] = ACTIONS(2366),
    [anon_sym_CARET] = ACTIONS(2366),
    [anon_sym_AMP] = ACTIONS(2366),
    [anon_sym_SEMI] = ACTIONS(2366),
    [anon_sym___extension__] = ACTIONS(2364),
    [anon_sym_typedef] = ACTIONS(2364),
    [anon_sym_extern] = ACTIONS(2364),
    [anon_sym___attribute__] = ACTIONS(2364),
    [anon_sym___attribute] = ACTIONS(2364),
    [anon_sym_noreturn] = ACTIONS(2364),
    [anon_sym_LBRACK] = ACTIONS(2366),
    [anon_sym___declspec] = ACTIONS(2364),
    [anon_sym___cdecl] = ACTIONS(2364),
    [anon_sym___clrcall] = ACTIONS(2364),
    [anon_sym___stdcall] = ACTIONS(2364),
    [anon_sym___fastcall] = ACTIONS(2364),
    [anon_sym___thiscall] = ACTIONS(2364),
    [anon_sym___vectorcall] = ACTIONS(2364),
    [anon_sym_LBRACE] = ACTIONS(2366),
    [anon_sym_signed] = ACTIONS(2364),
    [anon_sym_unsigned] = ACTIONS(2364),
    [anon_sym_long] = ACTIONS(2364),
    [anon_sym_short] = ACTIONS(2364),
    [anon_sym_static] = ACTIONS(2364),
    [anon_sym_ATautoreleasepool] = ACTIONS(2366),
    [anon_sym_auto] = ACTIONS(2364),
    [anon_sym_register] = ACTIONS(2364),
    [anon_sym_inline] = ACTIONS(2364),
    [anon_sym___inline] = ACTIONS(2364),
    [anon_sym___inline__] = ACTIONS(2364),
    [anon_sym___forceinline] = ACTIONS(2364),
    [anon_sym_thread_local] = ACTIONS(2364),
    [anon_sym___thread] = ACTIONS(2364),
    [anon_sym_CG_EXTERN] = ACTIONS(2364),
    [anon_sym_CG_INLINE] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2364),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2364),
    [anon_sym_IBOutlet] = ACTIONS(2364),
    [anon_sym_IBInspectable] = ACTIONS(2364),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2364),
    [anon_sym_NS_INLINE] = ACTIONS(2364),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2364),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2364),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2364),
    [anon_sym_const] = ACTIONS(2364),
    [anon_sym_constexpr] = ACTIONS(2364),
    [anon_sym_volatile] = ACTIONS(2364),
    [anon_sym_restrict] = ACTIONS(2364),
    [anon_sym___restrict__] = ACTIONS(2364),
    [anon_sym__Atomic] = ACTIONS(2364),
    [anon_sym__Noreturn] = ACTIONS(2364),
    [anon_sym__Nonnull] = ACTIONS(2364),
    [anon_sym_nullable] = ACTIONS(2364),
    [anon_sym__Complex] = ACTIONS(2364),
    [anon_sym__Nullable] = ACTIONS(2364),
    [anon_sym__Nullable_result] = ACTIONS(2364),
    [anon_sym__Null_unspecified] = ACTIONS(2364),
    [anon_sym___autoreleasing] = ACTIONS(2364),
    [anon_sym___block] = ACTIONS(2364),
    [anon_sym___bridge] = ACTIONS(2364),
    [anon_sym___bridge_retained] = ACTIONS(2364),
    [anon_sym___bridge_transfer] = ACTIONS(2364),
    [anon_sym___complex] = ACTIONS(2364),
    [anon_sym___const] = ACTIONS(2364),
    [anon_sym___imag] = ACTIONS(2364),
    [anon_sym___kindof] = ACTIONS(2364),
    [anon_sym___nonnull] = ACTIONS(2364),
    [anon_sym___nullable] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2364),
    [anon_sym___real] = ACTIONS(2364),
    [anon_sym___strong] = ACTIONS(2364),
    [anon_sym___unsafe_unretained] = ACTIONS(2364),
    [anon_sym___unused] = ACTIONS(2364),
    [anon_sym___weak] = ACTIONS(2364),
    [anon_sym_alignas] = ACTIONS(2364),
    [anon_sym__Alignas] = ACTIONS(2364),
    [sym_primitive_type] = ACTIONS(2364),
    [anon_sym_enum] = ACTIONS(2364),
    [anon_sym_struct] = ACTIONS(2364),
    [anon_sym_union] = ACTIONS(2364),
    [anon_sym_if] = ACTIONS(2364),
    [anon_sym_else] = ACTIONS(2364),
    [anon_sym_switch] = ACTIONS(2364),
    [anon_sym_case] = ACTIONS(2364),
    [anon_sym_default] = ACTIONS(2364),
    [anon_sym_while] = ACTIONS(2364),
    [anon_sym_do] = ACTIONS(2364),
    [anon_sym_for] = ACTIONS(2364),
    [anon_sym_in] = ACTIONS(2364),
    [anon_sym_return] = ACTIONS(2364),
    [anon_sym_break] = ACTIONS(2364),
    [anon_sym_continue] = ACTIONS(2364),
    [anon_sym_goto] = ACTIONS(2364),
    [anon_sym_DASH_DASH] = ACTIONS(2366),
    [anon_sym_PLUS_PLUS] = ACTIONS(2366),
    [anon_sym_sizeof] = ACTIONS(2364),
    [anon_sym___alignof__] = ACTIONS(2364),
    [anon_sym___alignof] = ACTIONS(2364),
    [anon_sym__alignof] = ACTIONS(2364),
    [anon_sym_alignof] = ACTIONS(2364),
    [anon_sym__Alignof] = ACTIONS(2364),
    [anon_sym_offsetof] = ACTIONS(2364),
    [anon_sym__Generic] = ACTIONS(2364),
    [anon_sym_asm] = ACTIONS(2364),
    [anon_sym___asm__] = ACTIONS(2364),
    [anon_sym___asm] = ACTIONS(2364),
    [sym_number_literal] = ACTIONS(2366),
    [anon_sym_L_SQUOTE] = ACTIONS(2366),
    [anon_sym_u_SQUOTE] = ACTIONS(2366),
    [anon_sym_U_SQUOTE] = ACTIONS(2366),
    [anon_sym_u8_SQUOTE] = ACTIONS(2366),
    [anon_sym_SQUOTE] = ACTIONS(2366),
    [anon_sym_AT] = ACTIONS(2364),
    [anon_sym_DQUOTE] = ACTIONS(2366),
    [anon_sym_L_DQUOTE] = ACTIONS(2366),
    [anon_sym_u_DQUOTE] = ACTIONS(2366),
    [anon_sym_U_DQUOTE] = ACTIONS(2366),
    [anon_sym_u8_DQUOTE] = ACTIONS(2366),
    [sym_true] = ACTIONS(2364),
    [sym_false] = ACTIONS(2364),
    [anon_sym_NULL] = ACTIONS(2364),
    [anon_sym_nullptr] = ACTIONS(2364),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2364),
    [anon_sym___typeof] = ACTIONS(2364),
    [anon_sym_typeof] = ACTIONS(2364),
    [anon_sym_ATimport] = ACTIONS(2366),
    [aux_sym_preproc_undef_token1] = ACTIONS(2364),
    [anon_sym_POUND] = ACTIONS(2364),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2364),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2364),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2364),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_API_AVAILABLE] = ACTIONS(2364),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_API_DEPRECATED] = ACTIONS(2364),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym___deprecated_msg] = ACTIONS(2364),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2364),
    [anon_sym_ATprotocol] = ACTIONS(2366),
    [anon_sym_ATinterface] = ACTIONS(2366),
    [anon_sym_ATimplementation] = ACTIONS(2366),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2366),
    [anon_sym_ATtry] = ACTIONS(2366),
    [anon_sym___try] = ACTIONS(2364),
    [anon_sym_ATcatch] = ACTIONS(2366),
    [anon_sym___catch] = ACTIONS(2364),
    [anon_sym_ATfinally] = ACTIONS(2366),
    [anon_sym___finally] = ACTIONS(2364),
    [anon_sym_ATthrow] = ACTIONS(2366),
    [anon_sym_ATselector] = ACTIONS(2366),
    [anon_sym_ATavailable] = ACTIONS(2366),
    [anon_sym___builtin_available] = ACTIONS(2364),
    [anon_sym_va_arg] = ACTIONS(2364),
    [anon_sym_ATencode] = ACTIONS(2366),
    [anon_sym_ATsynchronized] = ACTIONS(2366),
    [anon_sym_BOOL] = ACTIONS(2364),
    [anon_sym_IMP] = ACTIONS(2364),
    [anon_sym_SEL] = ACTIONS(2364),
    [anon_sym_Class] = ACTIONS(2364),
    [anon_sym_id] = ACTIONS(2364),
  },
  [STATE(428)] = {
    [sym_identifier] = ACTIONS(2518),
    [aux_sym_preproc_include_token1] = ACTIONS(2518),
    [aux_sym_preproc_include_token2] = ACTIONS(2518),
    [aux_sym_preproc_def_token1] = ACTIONS(2518),
    [aux_sym_preproc_if_token1] = ACTIONS(2518),
    [aux_sym_preproc_if_token2] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2518),
    [aux_sym_preproc_else_token1] = ACTIONS(2518),
    [aux_sym_preproc_elif_token1] = ACTIONS(2518),
    [sym_preproc_directive] = ACTIONS(2518),
    [anon_sym_LPAREN2] = ACTIONS(2520),
    [anon_sym_BANG] = ACTIONS(2520),
    [anon_sym_TILDE] = ACTIONS(2520),
    [anon_sym_DASH] = ACTIONS(2518),
    [anon_sym_PLUS] = ACTIONS(2518),
    [anon_sym_STAR] = ACTIONS(2520),
    [anon_sym_CARET] = ACTIONS(2520),
    [anon_sym_AMP] = ACTIONS(2520),
    [anon_sym_SEMI] = ACTIONS(2520),
    [anon_sym___extension__] = ACTIONS(2518),
    [anon_sym_typedef] = ACTIONS(2518),
    [anon_sym_extern] = ACTIONS(2518),
    [anon_sym___attribute__] = ACTIONS(2518),
    [anon_sym___attribute] = ACTIONS(2518),
    [anon_sym_noreturn] = ACTIONS(2518),
    [anon_sym_LBRACK] = ACTIONS(2520),
    [anon_sym___declspec] = ACTIONS(2518),
    [anon_sym___cdecl] = ACTIONS(2518),
    [anon_sym___clrcall] = ACTIONS(2518),
    [anon_sym___stdcall] = ACTIONS(2518),
    [anon_sym___fastcall] = ACTIONS(2518),
    [anon_sym___thiscall] = ACTIONS(2518),
    [anon_sym___vectorcall] = ACTIONS(2518),
    [anon_sym_LBRACE] = ACTIONS(2520),
    [anon_sym_signed] = ACTIONS(2518),
    [anon_sym_unsigned] = ACTIONS(2518),
    [anon_sym_long] = ACTIONS(2518),
    [anon_sym_short] = ACTIONS(2518),
    [anon_sym_static] = ACTIONS(2518),
    [anon_sym_ATautoreleasepool] = ACTIONS(2520),
    [anon_sym_auto] = ACTIONS(2518),
    [anon_sym_register] = ACTIONS(2518),
    [anon_sym_inline] = ACTIONS(2518),
    [anon_sym___inline] = ACTIONS(2518),
    [anon_sym___inline__] = ACTIONS(2518),
    [anon_sym___forceinline] = ACTIONS(2518),
    [anon_sym_thread_local] = ACTIONS(2518),
    [anon_sym___thread] = ACTIONS(2518),
    [anon_sym_CG_EXTERN] = ACTIONS(2518),
    [anon_sym_CG_INLINE] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2518),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2518),
    [anon_sym_IBOutlet] = ACTIONS(2518),
    [anon_sym_IBInspectable] = ACTIONS(2518),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2518),
    [anon_sym_NS_INLINE] = ACTIONS(2518),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2518),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2518),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2518),
    [anon_sym_const] = ACTIONS(2518),
    [anon_sym_constexpr] = ACTIONS(2518),
    [anon_sym_volatile] = ACTIONS(2518),
    [anon_sym_restrict] = ACTIONS(2518),
    [anon_sym___restrict__] = ACTIONS(2518),
    [anon_sym__Atomic] = ACTIONS(2518),
    [anon_sym__Noreturn] = ACTIONS(2518),
    [anon_sym__Nonnull] = ACTIONS(2518),
    [anon_sym_nullable] = ACTIONS(2518),
    [anon_sym__Complex] = ACTIONS(2518),
    [anon_sym__Nullable] = ACTIONS(2518),
    [anon_sym__Nullable_result] = ACTIONS(2518),
    [anon_sym__Null_unspecified] = ACTIONS(2518),
    [anon_sym___autoreleasing] = ACTIONS(2518),
    [anon_sym___block] = ACTIONS(2518),
    [anon_sym___bridge] = ACTIONS(2518),
    [anon_sym___bridge_retained] = ACTIONS(2518),
    [anon_sym___bridge_transfer] = ACTIONS(2518),
    [anon_sym___complex] = ACTIONS(2518),
    [anon_sym___const] = ACTIONS(2518),
    [anon_sym___imag] = ACTIONS(2518),
    [anon_sym___kindof] = ACTIONS(2518),
    [anon_sym___nonnull] = ACTIONS(2518),
    [anon_sym___nullable] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2518),
    [anon_sym___real] = ACTIONS(2518),
    [anon_sym___strong] = ACTIONS(2518),
    [anon_sym___unsafe_unretained] = ACTIONS(2518),
    [anon_sym___unused] = ACTIONS(2518),
    [anon_sym___weak] = ACTIONS(2518),
    [anon_sym_alignas] = ACTIONS(2518),
    [anon_sym__Alignas] = ACTIONS(2518),
    [sym_primitive_type] = ACTIONS(2518),
    [anon_sym_enum] = ACTIONS(2518),
    [anon_sym_struct] = ACTIONS(2518),
    [anon_sym_union] = ACTIONS(2518),
    [anon_sym_if] = ACTIONS(2518),
    [anon_sym_else] = ACTIONS(2518),
    [anon_sym_switch] = ACTIONS(2518),
    [anon_sym_case] = ACTIONS(2518),
    [anon_sym_default] = ACTIONS(2518),
    [anon_sym_while] = ACTIONS(2518),
    [anon_sym_do] = ACTIONS(2518),
    [anon_sym_for] = ACTIONS(2518),
    [anon_sym_in] = ACTIONS(2518),
    [anon_sym_return] = ACTIONS(2518),
    [anon_sym_break] = ACTIONS(2518),
    [anon_sym_continue] = ACTIONS(2518),
    [anon_sym_goto] = ACTIONS(2518),
    [anon_sym_DASH_DASH] = ACTIONS(2520),
    [anon_sym_PLUS_PLUS] = ACTIONS(2520),
    [anon_sym_sizeof] = ACTIONS(2518),
    [anon_sym___alignof__] = ACTIONS(2518),
    [anon_sym___alignof] = ACTIONS(2518),
    [anon_sym__alignof] = ACTIONS(2518),
    [anon_sym_alignof] = ACTIONS(2518),
    [anon_sym__Alignof] = ACTIONS(2518),
    [anon_sym_offsetof] = ACTIONS(2518),
    [anon_sym__Generic] = ACTIONS(2518),
    [anon_sym_asm] = ACTIONS(2518),
    [anon_sym___asm__] = ACTIONS(2518),
    [anon_sym___asm] = ACTIONS(2518),
    [sym_number_literal] = ACTIONS(2520),
    [anon_sym_L_SQUOTE] = ACTIONS(2520),
    [anon_sym_u_SQUOTE] = ACTIONS(2520),
    [anon_sym_U_SQUOTE] = ACTIONS(2520),
    [anon_sym_u8_SQUOTE] = ACTIONS(2520),
    [anon_sym_SQUOTE] = ACTIONS(2520),
    [anon_sym_AT] = ACTIONS(2518),
    [anon_sym_DQUOTE] = ACTIONS(2520),
    [anon_sym_L_DQUOTE] = ACTIONS(2520),
    [anon_sym_u_DQUOTE] = ACTIONS(2520),
    [anon_sym_U_DQUOTE] = ACTIONS(2520),
    [anon_sym_u8_DQUOTE] = ACTIONS(2520),
    [sym_true] = ACTIONS(2518),
    [sym_false] = ACTIONS(2518),
    [anon_sym_NULL] = ACTIONS(2518),
    [anon_sym_nullptr] = ACTIONS(2518),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2518),
    [anon_sym___typeof] = ACTIONS(2518),
    [anon_sym_typeof] = ACTIONS(2518),
    [anon_sym_ATimport] = ACTIONS(2520),
    [aux_sym_preproc_undef_token1] = ACTIONS(2518),
    [anon_sym_POUND] = ACTIONS(2518),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2518),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2518),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2518),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_API_AVAILABLE] = ACTIONS(2518),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_API_DEPRECATED] = ACTIONS(2518),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym___deprecated_msg] = ACTIONS(2518),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2518),
    [anon_sym_ATprotocol] = ACTIONS(2520),
    [anon_sym_ATinterface] = ACTIONS(2520),
    [anon_sym_ATimplementation] = ACTIONS(2520),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2520),
    [anon_sym_ATsynthesize] = ACTIONS(2520),
    [anon_sym_ATdynamic] = ACTIONS(2520),
    [anon_sym_ATtry] = ACTIONS(2520),
    [anon_sym___try] = ACTIONS(2518),
    [anon_sym_ATthrow] = ACTIONS(2520),
    [anon_sym_ATselector] = ACTIONS(2520),
    [anon_sym_ATavailable] = ACTIONS(2520),
    [anon_sym___builtin_available] = ACTIONS(2518),
    [anon_sym_va_arg] = ACTIONS(2518),
    [anon_sym_ATencode] = ACTIONS(2520),
    [anon_sym_ATsynchronized] = ACTIONS(2520),
    [anon_sym_BOOL] = ACTIONS(2518),
    [anon_sym_IMP] = ACTIONS(2518),
    [anon_sym_SEL] = ACTIONS(2518),
    [anon_sym_Class] = ACTIONS(2518),
    [anon_sym_id] = ACTIONS(2518),
  },
  [STATE(429)] = {
    [sym_identifier] = ACTIONS(2522),
    [aux_sym_preproc_include_token1] = ACTIONS(2522),
    [aux_sym_preproc_include_token2] = ACTIONS(2522),
    [aux_sym_preproc_def_token1] = ACTIONS(2522),
    [aux_sym_preproc_if_token1] = ACTIONS(2522),
    [aux_sym_preproc_if_token2] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2522),
    [aux_sym_preproc_else_token1] = ACTIONS(2522),
    [aux_sym_preproc_elif_token1] = ACTIONS(2522),
    [sym_preproc_directive] = ACTIONS(2522),
    [anon_sym_LPAREN2] = ACTIONS(2524),
    [anon_sym_BANG] = ACTIONS(2524),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_DASH] = ACTIONS(2522),
    [anon_sym_PLUS] = ACTIONS(2522),
    [anon_sym_STAR] = ACTIONS(2524),
    [anon_sym_CARET] = ACTIONS(2524),
    [anon_sym_AMP] = ACTIONS(2524),
    [anon_sym_SEMI] = ACTIONS(2524),
    [anon_sym___extension__] = ACTIONS(2522),
    [anon_sym_typedef] = ACTIONS(2522),
    [anon_sym_extern] = ACTIONS(2522),
    [anon_sym___attribute__] = ACTIONS(2522),
    [anon_sym___attribute] = ACTIONS(2522),
    [anon_sym_noreturn] = ACTIONS(2522),
    [anon_sym_LBRACK] = ACTIONS(2524),
    [anon_sym___declspec] = ACTIONS(2522),
    [anon_sym___cdecl] = ACTIONS(2522),
    [anon_sym___clrcall] = ACTIONS(2522),
    [anon_sym___stdcall] = ACTIONS(2522),
    [anon_sym___fastcall] = ACTIONS(2522),
    [anon_sym___thiscall] = ACTIONS(2522),
    [anon_sym___vectorcall] = ACTIONS(2522),
    [anon_sym_LBRACE] = ACTIONS(2524),
    [anon_sym_signed] = ACTIONS(2522),
    [anon_sym_unsigned] = ACTIONS(2522),
    [anon_sym_long] = ACTIONS(2522),
    [anon_sym_short] = ACTIONS(2522),
    [anon_sym_static] = ACTIONS(2522),
    [anon_sym_ATautoreleasepool] = ACTIONS(2524),
    [anon_sym_auto] = ACTIONS(2522),
    [anon_sym_register] = ACTIONS(2522),
    [anon_sym_inline] = ACTIONS(2522),
    [anon_sym___inline] = ACTIONS(2522),
    [anon_sym___inline__] = ACTIONS(2522),
    [anon_sym___forceinline] = ACTIONS(2522),
    [anon_sym_thread_local] = ACTIONS(2522),
    [anon_sym___thread] = ACTIONS(2522),
    [anon_sym_CG_EXTERN] = ACTIONS(2522),
    [anon_sym_CG_INLINE] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2522),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2522),
    [anon_sym_IBOutlet] = ACTIONS(2522),
    [anon_sym_IBInspectable] = ACTIONS(2522),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2522),
    [anon_sym_NS_INLINE] = ACTIONS(2522),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2522),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2522),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2522),
    [anon_sym_const] = ACTIONS(2522),
    [anon_sym_constexpr] = ACTIONS(2522),
    [anon_sym_volatile] = ACTIONS(2522),
    [anon_sym_restrict] = ACTIONS(2522),
    [anon_sym___restrict__] = ACTIONS(2522),
    [anon_sym__Atomic] = ACTIONS(2522),
    [anon_sym__Noreturn] = ACTIONS(2522),
    [anon_sym__Nonnull] = ACTIONS(2522),
    [anon_sym_nullable] = ACTIONS(2522),
    [anon_sym__Complex] = ACTIONS(2522),
    [anon_sym__Nullable] = ACTIONS(2522),
    [anon_sym__Nullable_result] = ACTIONS(2522),
    [anon_sym__Null_unspecified] = ACTIONS(2522),
    [anon_sym___autoreleasing] = ACTIONS(2522),
    [anon_sym___block] = ACTIONS(2522),
    [anon_sym___bridge] = ACTIONS(2522),
    [anon_sym___bridge_retained] = ACTIONS(2522),
    [anon_sym___bridge_transfer] = ACTIONS(2522),
    [anon_sym___complex] = ACTIONS(2522),
    [anon_sym___const] = ACTIONS(2522),
    [anon_sym___imag] = ACTIONS(2522),
    [anon_sym___kindof] = ACTIONS(2522),
    [anon_sym___nonnull] = ACTIONS(2522),
    [anon_sym___nullable] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2522),
    [anon_sym___real] = ACTIONS(2522),
    [anon_sym___strong] = ACTIONS(2522),
    [anon_sym___unsafe_unretained] = ACTIONS(2522),
    [anon_sym___unused] = ACTIONS(2522),
    [anon_sym___weak] = ACTIONS(2522),
    [anon_sym_alignas] = ACTIONS(2522),
    [anon_sym__Alignas] = ACTIONS(2522),
    [sym_primitive_type] = ACTIONS(2522),
    [anon_sym_enum] = ACTIONS(2522),
    [anon_sym_struct] = ACTIONS(2522),
    [anon_sym_union] = ACTIONS(2522),
    [anon_sym_if] = ACTIONS(2522),
    [anon_sym_else] = ACTIONS(2522),
    [anon_sym_switch] = ACTIONS(2522),
    [anon_sym_case] = ACTIONS(2522),
    [anon_sym_default] = ACTIONS(2522),
    [anon_sym_while] = ACTIONS(2522),
    [anon_sym_do] = ACTIONS(2522),
    [anon_sym_for] = ACTIONS(2522),
    [anon_sym_in] = ACTIONS(2522),
    [anon_sym_return] = ACTIONS(2522),
    [anon_sym_break] = ACTIONS(2522),
    [anon_sym_continue] = ACTIONS(2522),
    [anon_sym_goto] = ACTIONS(2522),
    [anon_sym_DASH_DASH] = ACTIONS(2524),
    [anon_sym_PLUS_PLUS] = ACTIONS(2524),
    [anon_sym_sizeof] = ACTIONS(2522),
    [anon_sym___alignof__] = ACTIONS(2522),
    [anon_sym___alignof] = ACTIONS(2522),
    [anon_sym__alignof] = ACTIONS(2522),
    [anon_sym_alignof] = ACTIONS(2522),
    [anon_sym__Alignof] = ACTIONS(2522),
    [anon_sym_offsetof] = ACTIONS(2522),
    [anon_sym__Generic] = ACTIONS(2522),
    [anon_sym_asm] = ACTIONS(2522),
    [anon_sym___asm__] = ACTIONS(2522),
    [anon_sym___asm] = ACTIONS(2522),
    [sym_number_literal] = ACTIONS(2524),
    [anon_sym_L_SQUOTE] = ACTIONS(2524),
    [anon_sym_u_SQUOTE] = ACTIONS(2524),
    [anon_sym_U_SQUOTE] = ACTIONS(2524),
    [anon_sym_u8_SQUOTE] = ACTIONS(2524),
    [anon_sym_SQUOTE] = ACTIONS(2524),
    [anon_sym_AT] = ACTIONS(2522),
    [anon_sym_DQUOTE] = ACTIONS(2524),
    [anon_sym_L_DQUOTE] = ACTIONS(2524),
    [anon_sym_u_DQUOTE] = ACTIONS(2524),
    [anon_sym_U_DQUOTE] = ACTIONS(2524),
    [anon_sym_u8_DQUOTE] = ACTIONS(2524),
    [sym_true] = ACTIONS(2522),
    [sym_false] = ACTIONS(2522),
    [anon_sym_NULL] = ACTIONS(2522),
    [anon_sym_nullptr] = ACTIONS(2522),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2522),
    [anon_sym___typeof] = ACTIONS(2522),
    [anon_sym_typeof] = ACTIONS(2522),
    [anon_sym_ATimport] = ACTIONS(2524),
    [aux_sym_preproc_undef_token1] = ACTIONS(2522),
    [anon_sym_POUND] = ACTIONS(2522),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2522),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2522),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2522),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_API_AVAILABLE] = ACTIONS(2522),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_API_DEPRECATED] = ACTIONS(2522),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym___deprecated_msg] = ACTIONS(2522),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2522),
    [anon_sym_ATprotocol] = ACTIONS(2524),
    [anon_sym_ATinterface] = ACTIONS(2524),
    [anon_sym_ATimplementation] = ACTIONS(2524),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2524),
    [anon_sym_ATsynthesize] = ACTIONS(2524),
    [anon_sym_ATdynamic] = ACTIONS(2524),
    [anon_sym_ATtry] = ACTIONS(2524),
    [anon_sym___try] = ACTIONS(2522),
    [anon_sym_ATthrow] = ACTIONS(2524),
    [anon_sym_ATselector] = ACTIONS(2524),
    [anon_sym_ATavailable] = ACTIONS(2524),
    [anon_sym___builtin_available] = ACTIONS(2522),
    [anon_sym_va_arg] = ACTIONS(2522),
    [anon_sym_ATencode] = ACTIONS(2524),
    [anon_sym_ATsynchronized] = ACTIONS(2524),
    [anon_sym_BOOL] = ACTIONS(2522),
    [anon_sym_IMP] = ACTIONS(2522),
    [anon_sym_SEL] = ACTIONS(2522),
    [anon_sym_Class] = ACTIONS(2522),
    [anon_sym_id] = ACTIONS(2522),
  },
  [STATE(430)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token2] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [aux_sym_preproc_else_token1] = ACTIONS(2420),
    [aux_sym_preproc_elif_token1] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(431)] = {
    [sym_identifier] = ACTIONS(2526),
    [aux_sym_preproc_include_token1] = ACTIONS(2526),
    [aux_sym_preproc_include_token2] = ACTIONS(2526),
    [aux_sym_preproc_def_token1] = ACTIONS(2526),
    [aux_sym_preproc_if_token1] = ACTIONS(2526),
    [aux_sym_preproc_if_token2] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2526),
    [aux_sym_preproc_else_token1] = ACTIONS(2526),
    [aux_sym_preproc_elif_token1] = ACTIONS(2526),
    [sym_preproc_directive] = ACTIONS(2526),
    [anon_sym_LPAREN2] = ACTIONS(2528),
    [anon_sym_BANG] = ACTIONS(2528),
    [anon_sym_TILDE] = ACTIONS(2528),
    [anon_sym_DASH] = ACTIONS(2526),
    [anon_sym_PLUS] = ACTIONS(2526),
    [anon_sym_STAR] = ACTIONS(2528),
    [anon_sym_CARET] = ACTIONS(2528),
    [anon_sym_AMP] = ACTIONS(2528),
    [anon_sym_SEMI] = ACTIONS(2528),
    [anon_sym___extension__] = ACTIONS(2526),
    [anon_sym_typedef] = ACTIONS(2526),
    [anon_sym_extern] = ACTIONS(2526),
    [anon_sym___attribute__] = ACTIONS(2526),
    [anon_sym___attribute] = ACTIONS(2526),
    [anon_sym_noreturn] = ACTIONS(2526),
    [anon_sym_LBRACK] = ACTIONS(2528),
    [anon_sym___declspec] = ACTIONS(2526),
    [anon_sym___cdecl] = ACTIONS(2526),
    [anon_sym___clrcall] = ACTIONS(2526),
    [anon_sym___stdcall] = ACTIONS(2526),
    [anon_sym___fastcall] = ACTIONS(2526),
    [anon_sym___thiscall] = ACTIONS(2526),
    [anon_sym___vectorcall] = ACTIONS(2526),
    [anon_sym_LBRACE] = ACTIONS(2528),
    [anon_sym_signed] = ACTIONS(2526),
    [anon_sym_unsigned] = ACTIONS(2526),
    [anon_sym_long] = ACTIONS(2526),
    [anon_sym_short] = ACTIONS(2526),
    [anon_sym_static] = ACTIONS(2526),
    [anon_sym_ATautoreleasepool] = ACTIONS(2528),
    [anon_sym_auto] = ACTIONS(2526),
    [anon_sym_register] = ACTIONS(2526),
    [anon_sym_inline] = ACTIONS(2526),
    [anon_sym___inline] = ACTIONS(2526),
    [anon_sym___inline__] = ACTIONS(2526),
    [anon_sym___forceinline] = ACTIONS(2526),
    [anon_sym_thread_local] = ACTIONS(2526),
    [anon_sym___thread] = ACTIONS(2526),
    [anon_sym_CG_EXTERN] = ACTIONS(2526),
    [anon_sym_CG_INLINE] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2526),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2526),
    [anon_sym_IBOutlet] = ACTIONS(2526),
    [anon_sym_IBInspectable] = ACTIONS(2526),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2526),
    [anon_sym_NS_INLINE] = ACTIONS(2526),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2526),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2526),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2526),
    [anon_sym_const] = ACTIONS(2526),
    [anon_sym_constexpr] = ACTIONS(2526),
    [anon_sym_volatile] = ACTIONS(2526),
    [anon_sym_restrict] = ACTIONS(2526),
    [anon_sym___restrict__] = ACTIONS(2526),
    [anon_sym__Atomic] = ACTIONS(2526),
    [anon_sym__Noreturn] = ACTIONS(2526),
    [anon_sym__Nonnull] = ACTIONS(2526),
    [anon_sym_nullable] = ACTIONS(2526),
    [anon_sym__Complex] = ACTIONS(2526),
    [anon_sym__Nullable] = ACTIONS(2526),
    [anon_sym__Nullable_result] = ACTIONS(2526),
    [anon_sym__Null_unspecified] = ACTIONS(2526),
    [anon_sym___autoreleasing] = ACTIONS(2526),
    [anon_sym___block] = ACTIONS(2526),
    [anon_sym___bridge] = ACTIONS(2526),
    [anon_sym___bridge_retained] = ACTIONS(2526),
    [anon_sym___bridge_transfer] = ACTIONS(2526),
    [anon_sym___complex] = ACTIONS(2526),
    [anon_sym___const] = ACTIONS(2526),
    [anon_sym___imag] = ACTIONS(2526),
    [anon_sym___kindof] = ACTIONS(2526),
    [anon_sym___nonnull] = ACTIONS(2526),
    [anon_sym___nullable] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2526),
    [anon_sym___real] = ACTIONS(2526),
    [anon_sym___strong] = ACTIONS(2526),
    [anon_sym___unsafe_unretained] = ACTIONS(2526),
    [anon_sym___unused] = ACTIONS(2526),
    [anon_sym___weak] = ACTIONS(2526),
    [anon_sym_alignas] = ACTIONS(2526),
    [anon_sym__Alignas] = ACTIONS(2526),
    [sym_primitive_type] = ACTIONS(2526),
    [anon_sym_enum] = ACTIONS(2526),
    [anon_sym_struct] = ACTIONS(2526),
    [anon_sym_union] = ACTIONS(2526),
    [anon_sym_if] = ACTIONS(2526),
    [anon_sym_else] = ACTIONS(2526),
    [anon_sym_switch] = ACTIONS(2526),
    [anon_sym_case] = ACTIONS(2526),
    [anon_sym_default] = ACTIONS(2526),
    [anon_sym_while] = ACTIONS(2526),
    [anon_sym_do] = ACTIONS(2526),
    [anon_sym_for] = ACTIONS(2526),
    [anon_sym_in] = ACTIONS(2526),
    [anon_sym_return] = ACTIONS(2526),
    [anon_sym_break] = ACTIONS(2526),
    [anon_sym_continue] = ACTIONS(2526),
    [anon_sym_goto] = ACTIONS(2526),
    [anon_sym_DASH_DASH] = ACTIONS(2528),
    [anon_sym_PLUS_PLUS] = ACTIONS(2528),
    [anon_sym_sizeof] = ACTIONS(2526),
    [anon_sym___alignof__] = ACTIONS(2526),
    [anon_sym___alignof] = ACTIONS(2526),
    [anon_sym__alignof] = ACTIONS(2526),
    [anon_sym_alignof] = ACTIONS(2526),
    [anon_sym__Alignof] = ACTIONS(2526),
    [anon_sym_offsetof] = ACTIONS(2526),
    [anon_sym__Generic] = ACTIONS(2526),
    [anon_sym_asm] = ACTIONS(2526),
    [anon_sym___asm__] = ACTIONS(2526),
    [anon_sym___asm] = ACTIONS(2526),
    [sym_number_literal] = ACTIONS(2528),
    [anon_sym_L_SQUOTE] = ACTIONS(2528),
    [anon_sym_u_SQUOTE] = ACTIONS(2528),
    [anon_sym_U_SQUOTE] = ACTIONS(2528),
    [anon_sym_u8_SQUOTE] = ACTIONS(2528),
    [anon_sym_SQUOTE] = ACTIONS(2528),
    [anon_sym_AT] = ACTIONS(2526),
    [anon_sym_DQUOTE] = ACTIONS(2528),
    [anon_sym_L_DQUOTE] = ACTIONS(2528),
    [anon_sym_u_DQUOTE] = ACTIONS(2528),
    [anon_sym_U_DQUOTE] = ACTIONS(2528),
    [anon_sym_u8_DQUOTE] = ACTIONS(2528),
    [sym_true] = ACTIONS(2526),
    [sym_false] = ACTIONS(2526),
    [anon_sym_NULL] = ACTIONS(2526),
    [anon_sym_nullptr] = ACTIONS(2526),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2526),
    [anon_sym___typeof] = ACTIONS(2526),
    [anon_sym_typeof] = ACTIONS(2526),
    [anon_sym_ATimport] = ACTIONS(2528),
    [aux_sym_preproc_undef_token1] = ACTIONS(2526),
    [anon_sym_POUND] = ACTIONS(2526),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2526),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2526),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2526),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_API_AVAILABLE] = ACTIONS(2526),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_API_DEPRECATED] = ACTIONS(2526),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym___deprecated_msg] = ACTIONS(2526),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2526),
    [anon_sym_ATprotocol] = ACTIONS(2528),
    [anon_sym_ATinterface] = ACTIONS(2528),
    [anon_sym_ATimplementation] = ACTIONS(2528),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2528),
    [anon_sym_ATsynthesize] = ACTIONS(2528),
    [anon_sym_ATdynamic] = ACTIONS(2528),
    [anon_sym_ATtry] = ACTIONS(2528),
    [anon_sym___try] = ACTIONS(2526),
    [anon_sym_ATthrow] = ACTIONS(2528),
    [anon_sym_ATselector] = ACTIONS(2528),
    [anon_sym_ATavailable] = ACTIONS(2528),
    [anon_sym___builtin_available] = ACTIONS(2526),
    [anon_sym_va_arg] = ACTIONS(2526),
    [anon_sym_ATencode] = ACTIONS(2528),
    [anon_sym_ATsynchronized] = ACTIONS(2528),
    [anon_sym_BOOL] = ACTIONS(2526),
    [anon_sym_IMP] = ACTIONS(2526),
    [anon_sym_SEL] = ACTIONS(2526),
    [anon_sym_Class] = ACTIONS(2526),
    [anon_sym_id] = ACTIONS(2526),
  },
  [STATE(432)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(433)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(434)] = {
    [sym_identifier] = ACTIONS(2530),
    [aux_sym_preproc_include_token1] = ACTIONS(2530),
    [aux_sym_preproc_include_token2] = ACTIONS(2530),
    [aux_sym_preproc_def_token1] = ACTIONS(2530),
    [aux_sym_preproc_if_token1] = ACTIONS(2530),
    [aux_sym_preproc_if_token2] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2530),
    [aux_sym_preproc_else_token1] = ACTIONS(2530),
    [aux_sym_preproc_elif_token1] = ACTIONS(2530),
    [sym_preproc_directive] = ACTIONS(2530),
    [anon_sym_LPAREN2] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2530),
    [anon_sym_PLUS] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym___extension__] = ACTIONS(2530),
    [anon_sym_typedef] = ACTIONS(2530),
    [anon_sym_extern] = ACTIONS(2530),
    [anon_sym___attribute__] = ACTIONS(2530),
    [anon_sym___attribute] = ACTIONS(2530),
    [anon_sym_noreturn] = ACTIONS(2530),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym___declspec] = ACTIONS(2530),
    [anon_sym___cdecl] = ACTIONS(2530),
    [anon_sym___clrcall] = ACTIONS(2530),
    [anon_sym___stdcall] = ACTIONS(2530),
    [anon_sym___fastcall] = ACTIONS(2530),
    [anon_sym___thiscall] = ACTIONS(2530),
    [anon_sym___vectorcall] = ACTIONS(2530),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_signed] = ACTIONS(2530),
    [anon_sym_unsigned] = ACTIONS(2530),
    [anon_sym_long] = ACTIONS(2530),
    [anon_sym_short] = ACTIONS(2530),
    [anon_sym_static] = ACTIONS(2530),
    [anon_sym_ATautoreleasepool] = ACTIONS(2532),
    [anon_sym_auto] = ACTIONS(2530),
    [anon_sym_register] = ACTIONS(2530),
    [anon_sym_inline] = ACTIONS(2530),
    [anon_sym___inline] = ACTIONS(2530),
    [anon_sym___inline__] = ACTIONS(2530),
    [anon_sym___forceinline] = ACTIONS(2530),
    [anon_sym_thread_local] = ACTIONS(2530),
    [anon_sym___thread] = ACTIONS(2530),
    [anon_sym_CG_EXTERN] = ACTIONS(2530),
    [anon_sym_CG_INLINE] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2530),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2530),
    [anon_sym_IBOutlet] = ACTIONS(2530),
    [anon_sym_IBInspectable] = ACTIONS(2530),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2530),
    [anon_sym_NS_INLINE] = ACTIONS(2530),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2530),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2530),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2530),
    [anon_sym_const] = ACTIONS(2530),
    [anon_sym_constexpr] = ACTIONS(2530),
    [anon_sym_volatile] = ACTIONS(2530),
    [anon_sym_restrict] = ACTIONS(2530),
    [anon_sym___restrict__] = ACTIONS(2530),
    [anon_sym__Atomic] = ACTIONS(2530),
    [anon_sym__Noreturn] = ACTIONS(2530),
    [anon_sym__Nonnull] = ACTIONS(2530),
    [anon_sym_nullable] = ACTIONS(2530),
    [anon_sym__Complex] = ACTIONS(2530),
    [anon_sym__Nullable] = ACTIONS(2530),
    [anon_sym__Nullable_result] = ACTIONS(2530),
    [anon_sym__Null_unspecified] = ACTIONS(2530),
    [anon_sym___autoreleasing] = ACTIONS(2530),
    [anon_sym___block] = ACTIONS(2530),
    [anon_sym___bridge] = ACTIONS(2530),
    [anon_sym___bridge_retained] = ACTIONS(2530),
    [anon_sym___bridge_transfer] = ACTIONS(2530),
    [anon_sym___complex] = ACTIONS(2530),
    [anon_sym___const] = ACTIONS(2530),
    [anon_sym___imag] = ACTIONS(2530),
    [anon_sym___kindof] = ACTIONS(2530),
    [anon_sym___nonnull] = ACTIONS(2530),
    [anon_sym___nullable] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2530),
    [anon_sym___real] = ACTIONS(2530),
    [anon_sym___strong] = ACTIONS(2530),
    [anon_sym___unsafe_unretained] = ACTIONS(2530),
    [anon_sym___unused] = ACTIONS(2530),
    [anon_sym___weak] = ACTIONS(2530),
    [anon_sym_alignas] = ACTIONS(2530),
    [anon_sym__Alignas] = ACTIONS(2530),
    [sym_primitive_type] = ACTIONS(2530),
    [anon_sym_enum] = ACTIONS(2530),
    [anon_sym_struct] = ACTIONS(2530),
    [anon_sym_union] = ACTIONS(2530),
    [anon_sym_if] = ACTIONS(2530),
    [anon_sym_else] = ACTIONS(2530),
    [anon_sym_switch] = ACTIONS(2530),
    [anon_sym_case] = ACTIONS(2530),
    [anon_sym_default] = ACTIONS(2530),
    [anon_sym_while] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2530),
    [anon_sym_for] = ACTIONS(2530),
    [anon_sym_in] = ACTIONS(2530),
    [anon_sym_return] = ACTIONS(2530),
    [anon_sym_break] = ACTIONS(2530),
    [anon_sym_continue] = ACTIONS(2530),
    [anon_sym_goto] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_sizeof] = ACTIONS(2530),
    [anon_sym___alignof__] = ACTIONS(2530),
    [anon_sym___alignof] = ACTIONS(2530),
    [anon_sym__alignof] = ACTIONS(2530),
    [anon_sym_alignof] = ACTIONS(2530),
    [anon_sym__Alignof] = ACTIONS(2530),
    [anon_sym_offsetof] = ACTIONS(2530),
    [anon_sym__Generic] = ACTIONS(2530),
    [anon_sym_asm] = ACTIONS(2530),
    [anon_sym___asm__] = ACTIONS(2530),
    [anon_sym___asm] = ACTIONS(2530),
    [sym_number_literal] = ACTIONS(2532),
    [anon_sym_L_SQUOTE] = ACTIONS(2532),
    [anon_sym_u_SQUOTE] = ACTIONS(2532),
    [anon_sym_U_SQUOTE] = ACTIONS(2532),
    [anon_sym_u8_SQUOTE] = ACTIONS(2532),
    [anon_sym_SQUOTE] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2530),
    [anon_sym_DQUOTE] = ACTIONS(2532),
    [anon_sym_L_DQUOTE] = ACTIONS(2532),
    [anon_sym_u_DQUOTE] = ACTIONS(2532),
    [anon_sym_U_DQUOTE] = ACTIONS(2532),
    [anon_sym_u8_DQUOTE] = ACTIONS(2532),
    [sym_true] = ACTIONS(2530),
    [sym_false] = ACTIONS(2530),
    [anon_sym_NULL] = ACTIONS(2530),
    [anon_sym_nullptr] = ACTIONS(2530),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2530),
    [anon_sym___typeof] = ACTIONS(2530),
    [anon_sym_typeof] = ACTIONS(2530),
    [anon_sym_ATimport] = ACTIONS(2532),
    [aux_sym_preproc_undef_token1] = ACTIONS(2530),
    [anon_sym_POUND] = ACTIONS(2530),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2530),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2530),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2530),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_API_AVAILABLE] = ACTIONS(2530),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_API_DEPRECATED] = ACTIONS(2530),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym___deprecated_msg] = ACTIONS(2530),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2530),
    [anon_sym_ATprotocol] = ACTIONS(2532),
    [anon_sym_ATinterface] = ACTIONS(2532),
    [anon_sym_ATimplementation] = ACTIONS(2532),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2532),
    [anon_sym_ATsynthesize] = ACTIONS(2532),
    [anon_sym_ATdynamic] = ACTIONS(2532),
    [anon_sym_ATtry] = ACTIONS(2532),
    [anon_sym___try] = ACTIONS(2530),
    [anon_sym_ATthrow] = ACTIONS(2532),
    [anon_sym_ATselector] = ACTIONS(2532),
    [anon_sym_ATavailable] = ACTIONS(2532),
    [anon_sym___builtin_available] = ACTIONS(2530),
    [anon_sym_va_arg] = ACTIONS(2530),
    [anon_sym_ATencode] = ACTIONS(2532),
    [anon_sym_ATsynchronized] = ACTIONS(2532),
    [anon_sym_BOOL] = ACTIONS(2530),
    [anon_sym_IMP] = ACTIONS(2530),
    [anon_sym_SEL] = ACTIONS(2530),
    [anon_sym_Class] = ACTIONS(2530),
    [anon_sym_id] = ACTIONS(2530),
  },
  [STATE(435)] = {
    [sym_identifier] = ACTIONS(2534),
    [aux_sym_preproc_include_token1] = ACTIONS(2534),
    [aux_sym_preproc_include_token2] = ACTIONS(2534),
    [aux_sym_preproc_def_token1] = ACTIONS(2534),
    [aux_sym_preproc_if_token1] = ACTIONS(2534),
    [aux_sym_preproc_if_token2] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2534),
    [aux_sym_preproc_else_token1] = ACTIONS(2534),
    [aux_sym_preproc_elif_token1] = ACTIONS(2534),
    [sym_preproc_directive] = ACTIONS(2534),
    [anon_sym_LPAREN2] = ACTIONS(2536),
    [anon_sym_BANG] = ACTIONS(2536),
    [anon_sym_TILDE] = ACTIONS(2536),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2536),
    [anon_sym_CARET] = ACTIONS(2536),
    [anon_sym_AMP] = ACTIONS(2536),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym___extension__] = ACTIONS(2534),
    [anon_sym_typedef] = ACTIONS(2534),
    [anon_sym_extern] = ACTIONS(2534),
    [anon_sym___attribute__] = ACTIONS(2534),
    [anon_sym___attribute] = ACTIONS(2534),
    [anon_sym_noreturn] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2536),
    [anon_sym___declspec] = ACTIONS(2534),
    [anon_sym___cdecl] = ACTIONS(2534),
    [anon_sym___clrcall] = ACTIONS(2534),
    [anon_sym___stdcall] = ACTIONS(2534),
    [anon_sym___fastcall] = ACTIONS(2534),
    [anon_sym___thiscall] = ACTIONS(2534),
    [anon_sym___vectorcall] = ACTIONS(2534),
    [anon_sym_LBRACE] = ACTIONS(2536),
    [anon_sym_signed] = ACTIONS(2534),
    [anon_sym_unsigned] = ACTIONS(2534),
    [anon_sym_long] = ACTIONS(2534),
    [anon_sym_short] = ACTIONS(2534),
    [anon_sym_static] = ACTIONS(2534),
    [anon_sym_ATautoreleasepool] = ACTIONS(2536),
    [anon_sym_auto] = ACTIONS(2534),
    [anon_sym_register] = ACTIONS(2534),
    [anon_sym_inline] = ACTIONS(2534),
    [anon_sym___inline] = ACTIONS(2534),
    [anon_sym___inline__] = ACTIONS(2534),
    [anon_sym___forceinline] = ACTIONS(2534),
    [anon_sym_thread_local] = ACTIONS(2534),
    [anon_sym___thread] = ACTIONS(2534),
    [anon_sym_CG_EXTERN] = ACTIONS(2534),
    [anon_sym_CG_INLINE] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2534),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2534),
    [anon_sym_IBOutlet] = ACTIONS(2534),
    [anon_sym_IBInspectable] = ACTIONS(2534),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2534),
    [anon_sym_NS_INLINE] = ACTIONS(2534),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2534),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2534),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2534),
    [anon_sym_const] = ACTIONS(2534),
    [anon_sym_constexpr] = ACTIONS(2534),
    [anon_sym_volatile] = ACTIONS(2534),
    [anon_sym_restrict] = ACTIONS(2534),
    [anon_sym___restrict__] = ACTIONS(2534),
    [anon_sym__Atomic] = ACTIONS(2534),
    [anon_sym__Noreturn] = ACTIONS(2534),
    [anon_sym__Nonnull] = ACTIONS(2534),
    [anon_sym_nullable] = ACTIONS(2534),
    [anon_sym__Complex] = ACTIONS(2534),
    [anon_sym__Nullable] = ACTIONS(2534),
    [anon_sym__Nullable_result] = ACTIONS(2534),
    [anon_sym__Null_unspecified] = ACTIONS(2534),
    [anon_sym___autoreleasing] = ACTIONS(2534),
    [anon_sym___block] = ACTIONS(2534),
    [anon_sym___bridge] = ACTIONS(2534),
    [anon_sym___bridge_retained] = ACTIONS(2534),
    [anon_sym___bridge_transfer] = ACTIONS(2534),
    [anon_sym___complex] = ACTIONS(2534),
    [anon_sym___const] = ACTIONS(2534),
    [anon_sym___imag] = ACTIONS(2534),
    [anon_sym___kindof] = ACTIONS(2534),
    [anon_sym___nonnull] = ACTIONS(2534),
    [anon_sym___nullable] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2534),
    [anon_sym___real] = ACTIONS(2534),
    [anon_sym___strong] = ACTIONS(2534),
    [anon_sym___unsafe_unretained] = ACTIONS(2534),
    [anon_sym___unused] = ACTIONS(2534),
    [anon_sym___weak] = ACTIONS(2534),
    [anon_sym_alignas] = ACTIONS(2534),
    [anon_sym__Alignas] = ACTIONS(2534),
    [sym_primitive_type] = ACTIONS(2534),
    [anon_sym_enum] = ACTIONS(2534),
    [anon_sym_struct] = ACTIONS(2534),
    [anon_sym_union] = ACTIONS(2534),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_else] = ACTIONS(2534),
    [anon_sym_switch] = ACTIONS(2534),
    [anon_sym_case] = ACTIONS(2534),
    [anon_sym_default] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_do] = ACTIONS(2534),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_in] = ACTIONS(2534),
    [anon_sym_return] = ACTIONS(2534),
    [anon_sym_break] = ACTIONS(2534),
    [anon_sym_continue] = ACTIONS(2534),
    [anon_sym_goto] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_sizeof] = ACTIONS(2534),
    [anon_sym___alignof__] = ACTIONS(2534),
    [anon_sym___alignof] = ACTIONS(2534),
    [anon_sym__alignof] = ACTIONS(2534),
    [anon_sym_alignof] = ACTIONS(2534),
    [anon_sym__Alignof] = ACTIONS(2534),
    [anon_sym_offsetof] = ACTIONS(2534),
    [anon_sym__Generic] = ACTIONS(2534),
    [anon_sym_asm] = ACTIONS(2534),
    [anon_sym___asm__] = ACTIONS(2534),
    [anon_sym___asm] = ACTIONS(2534),
    [sym_number_literal] = ACTIONS(2536),
    [anon_sym_L_SQUOTE] = ACTIONS(2536),
    [anon_sym_u_SQUOTE] = ACTIONS(2536),
    [anon_sym_U_SQUOTE] = ACTIONS(2536),
    [anon_sym_u8_SQUOTE] = ACTIONS(2536),
    [anon_sym_SQUOTE] = ACTIONS(2536),
    [anon_sym_AT] = ACTIONS(2534),
    [anon_sym_DQUOTE] = ACTIONS(2536),
    [anon_sym_L_DQUOTE] = ACTIONS(2536),
    [anon_sym_u_DQUOTE] = ACTIONS(2536),
    [anon_sym_U_DQUOTE] = ACTIONS(2536),
    [anon_sym_u8_DQUOTE] = ACTIONS(2536),
    [sym_true] = ACTIONS(2534),
    [sym_false] = ACTIONS(2534),
    [anon_sym_NULL] = ACTIONS(2534),
    [anon_sym_nullptr] = ACTIONS(2534),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2534),
    [anon_sym___typeof] = ACTIONS(2534),
    [anon_sym_typeof] = ACTIONS(2534),
    [anon_sym_ATimport] = ACTIONS(2536),
    [aux_sym_preproc_undef_token1] = ACTIONS(2534),
    [anon_sym_POUND] = ACTIONS(2534),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2534),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2534),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2534),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_API_AVAILABLE] = ACTIONS(2534),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_API_DEPRECATED] = ACTIONS(2534),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym___deprecated_msg] = ACTIONS(2534),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2534),
    [anon_sym_ATprotocol] = ACTIONS(2536),
    [anon_sym_ATinterface] = ACTIONS(2536),
    [anon_sym_ATimplementation] = ACTIONS(2536),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2536),
    [anon_sym_ATsynthesize] = ACTIONS(2536),
    [anon_sym_ATdynamic] = ACTIONS(2536),
    [anon_sym_ATtry] = ACTIONS(2536),
    [anon_sym___try] = ACTIONS(2534),
    [anon_sym_ATthrow] = ACTIONS(2536),
    [anon_sym_ATselector] = ACTIONS(2536),
    [anon_sym_ATavailable] = ACTIONS(2536),
    [anon_sym___builtin_available] = ACTIONS(2534),
    [anon_sym_va_arg] = ACTIONS(2534),
    [anon_sym_ATencode] = ACTIONS(2536),
    [anon_sym_ATsynchronized] = ACTIONS(2536),
    [anon_sym_BOOL] = ACTIONS(2534),
    [anon_sym_IMP] = ACTIONS(2534),
    [anon_sym_SEL] = ACTIONS(2534),
    [anon_sym_Class] = ACTIONS(2534),
    [anon_sym_id] = ACTIONS(2534),
  },
  [STATE(436)] = {
    [sym_identifier] = ACTIONS(2538),
    [aux_sym_preproc_include_token1] = ACTIONS(2538),
    [aux_sym_preproc_include_token2] = ACTIONS(2538),
    [aux_sym_preproc_def_token1] = ACTIONS(2538),
    [aux_sym_preproc_if_token1] = ACTIONS(2538),
    [aux_sym_preproc_if_token2] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2538),
    [aux_sym_preproc_else_token1] = ACTIONS(2538),
    [aux_sym_preproc_elif_token1] = ACTIONS(2538),
    [sym_preproc_directive] = ACTIONS(2538),
    [anon_sym_LPAREN2] = ACTIONS(2540),
    [anon_sym_BANG] = ACTIONS(2540),
    [anon_sym_TILDE] = ACTIONS(2540),
    [anon_sym_DASH] = ACTIONS(2538),
    [anon_sym_PLUS] = ACTIONS(2538),
    [anon_sym_STAR] = ACTIONS(2540),
    [anon_sym_CARET] = ACTIONS(2540),
    [anon_sym_AMP] = ACTIONS(2540),
    [anon_sym_SEMI] = ACTIONS(2540),
    [anon_sym___extension__] = ACTIONS(2538),
    [anon_sym_typedef] = ACTIONS(2538),
    [anon_sym_extern] = ACTIONS(2538),
    [anon_sym___attribute__] = ACTIONS(2538),
    [anon_sym___attribute] = ACTIONS(2538),
    [anon_sym_noreturn] = ACTIONS(2538),
    [anon_sym_LBRACK] = ACTIONS(2540),
    [anon_sym___declspec] = ACTIONS(2538),
    [anon_sym___cdecl] = ACTIONS(2538),
    [anon_sym___clrcall] = ACTIONS(2538),
    [anon_sym___stdcall] = ACTIONS(2538),
    [anon_sym___fastcall] = ACTIONS(2538),
    [anon_sym___thiscall] = ACTIONS(2538),
    [anon_sym___vectorcall] = ACTIONS(2538),
    [anon_sym_LBRACE] = ACTIONS(2540),
    [anon_sym_signed] = ACTIONS(2538),
    [anon_sym_unsigned] = ACTIONS(2538),
    [anon_sym_long] = ACTIONS(2538),
    [anon_sym_short] = ACTIONS(2538),
    [anon_sym_static] = ACTIONS(2538),
    [anon_sym_ATautoreleasepool] = ACTIONS(2540),
    [anon_sym_auto] = ACTIONS(2538),
    [anon_sym_register] = ACTIONS(2538),
    [anon_sym_inline] = ACTIONS(2538),
    [anon_sym___inline] = ACTIONS(2538),
    [anon_sym___inline__] = ACTIONS(2538),
    [anon_sym___forceinline] = ACTIONS(2538),
    [anon_sym_thread_local] = ACTIONS(2538),
    [anon_sym___thread] = ACTIONS(2538),
    [anon_sym_CG_EXTERN] = ACTIONS(2538),
    [anon_sym_CG_INLINE] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2538),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2538),
    [anon_sym_IBOutlet] = ACTIONS(2538),
    [anon_sym_IBInspectable] = ACTIONS(2538),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2538),
    [anon_sym_NS_INLINE] = ACTIONS(2538),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2538),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2538),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2538),
    [anon_sym_const] = ACTIONS(2538),
    [anon_sym_constexpr] = ACTIONS(2538),
    [anon_sym_volatile] = ACTIONS(2538),
    [anon_sym_restrict] = ACTIONS(2538),
    [anon_sym___restrict__] = ACTIONS(2538),
    [anon_sym__Atomic] = ACTIONS(2538),
    [anon_sym__Noreturn] = ACTIONS(2538),
    [anon_sym__Nonnull] = ACTIONS(2538),
    [anon_sym_nullable] = ACTIONS(2538),
    [anon_sym__Complex] = ACTIONS(2538),
    [anon_sym__Nullable] = ACTIONS(2538),
    [anon_sym__Nullable_result] = ACTIONS(2538),
    [anon_sym__Null_unspecified] = ACTIONS(2538),
    [anon_sym___autoreleasing] = ACTIONS(2538),
    [anon_sym___block] = ACTIONS(2538),
    [anon_sym___bridge] = ACTIONS(2538),
    [anon_sym___bridge_retained] = ACTIONS(2538),
    [anon_sym___bridge_transfer] = ACTIONS(2538),
    [anon_sym___complex] = ACTIONS(2538),
    [anon_sym___const] = ACTIONS(2538),
    [anon_sym___imag] = ACTIONS(2538),
    [anon_sym___kindof] = ACTIONS(2538),
    [anon_sym___nonnull] = ACTIONS(2538),
    [anon_sym___nullable] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2538),
    [anon_sym___real] = ACTIONS(2538),
    [anon_sym___strong] = ACTIONS(2538),
    [anon_sym___unsafe_unretained] = ACTIONS(2538),
    [anon_sym___unused] = ACTIONS(2538),
    [anon_sym___weak] = ACTIONS(2538),
    [anon_sym_alignas] = ACTIONS(2538),
    [anon_sym__Alignas] = ACTIONS(2538),
    [sym_primitive_type] = ACTIONS(2538),
    [anon_sym_enum] = ACTIONS(2538),
    [anon_sym_struct] = ACTIONS(2538),
    [anon_sym_union] = ACTIONS(2538),
    [anon_sym_if] = ACTIONS(2538),
    [anon_sym_else] = ACTIONS(2538),
    [anon_sym_switch] = ACTIONS(2538),
    [anon_sym_case] = ACTIONS(2538),
    [anon_sym_default] = ACTIONS(2538),
    [anon_sym_while] = ACTIONS(2538),
    [anon_sym_do] = ACTIONS(2538),
    [anon_sym_for] = ACTIONS(2538),
    [anon_sym_in] = ACTIONS(2538),
    [anon_sym_return] = ACTIONS(2538),
    [anon_sym_break] = ACTIONS(2538),
    [anon_sym_continue] = ACTIONS(2538),
    [anon_sym_goto] = ACTIONS(2538),
    [anon_sym_DASH_DASH] = ACTIONS(2540),
    [anon_sym_PLUS_PLUS] = ACTIONS(2540),
    [anon_sym_sizeof] = ACTIONS(2538),
    [anon_sym___alignof__] = ACTIONS(2538),
    [anon_sym___alignof] = ACTIONS(2538),
    [anon_sym__alignof] = ACTIONS(2538),
    [anon_sym_alignof] = ACTIONS(2538),
    [anon_sym__Alignof] = ACTIONS(2538),
    [anon_sym_offsetof] = ACTIONS(2538),
    [anon_sym__Generic] = ACTIONS(2538),
    [anon_sym_asm] = ACTIONS(2538),
    [anon_sym___asm__] = ACTIONS(2538),
    [anon_sym___asm] = ACTIONS(2538),
    [sym_number_literal] = ACTIONS(2540),
    [anon_sym_L_SQUOTE] = ACTIONS(2540),
    [anon_sym_u_SQUOTE] = ACTIONS(2540),
    [anon_sym_U_SQUOTE] = ACTIONS(2540),
    [anon_sym_u8_SQUOTE] = ACTIONS(2540),
    [anon_sym_SQUOTE] = ACTIONS(2540),
    [anon_sym_AT] = ACTIONS(2538),
    [anon_sym_DQUOTE] = ACTIONS(2540),
    [anon_sym_L_DQUOTE] = ACTIONS(2540),
    [anon_sym_u_DQUOTE] = ACTIONS(2540),
    [anon_sym_U_DQUOTE] = ACTIONS(2540),
    [anon_sym_u8_DQUOTE] = ACTIONS(2540),
    [sym_true] = ACTIONS(2538),
    [sym_false] = ACTIONS(2538),
    [anon_sym_NULL] = ACTIONS(2538),
    [anon_sym_nullptr] = ACTIONS(2538),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2538),
    [anon_sym___typeof] = ACTIONS(2538),
    [anon_sym_typeof] = ACTIONS(2538),
    [anon_sym_ATimport] = ACTIONS(2540),
    [aux_sym_preproc_undef_token1] = ACTIONS(2538),
    [anon_sym_POUND] = ACTIONS(2538),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2538),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2538),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2538),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_API_AVAILABLE] = ACTIONS(2538),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_API_DEPRECATED] = ACTIONS(2538),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym___deprecated_msg] = ACTIONS(2538),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2538),
    [anon_sym_ATprotocol] = ACTIONS(2540),
    [anon_sym_ATinterface] = ACTIONS(2540),
    [anon_sym_ATimplementation] = ACTIONS(2540),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2540),
    [anon_sym_ATsynthesize] = ACTIONS(2540),
    [anon_sym_ATdynamic] = ACTIONS(2540),
    [anon_sym_ATtry] = ACTIONS(2540),
    [anon_sym___try] = ACTIONS(2538),
    [anon_sym_ATthrow] = ACTIONS(2540),
    [anon_sym_ATselector] = ACTIONS(2540),
    [anon_sym_ATavailable] = ACTIONS(2540),
    [anon_sym___builtin_available] = ACTIONS(2538),
    [anon_sym_va_arg] = ACTIONS(2538),
    [anon_sym_ATencode] = ACTIONS(2540),
    [anon_sym_ATsynchronized] = ACTIONS(2540),
    [anon_sym_BOOL] = ACTIONS(2538),
    [anon_sym_IMP] = ACTIONS(2538),
    [anon_sym_SEL] = ACTIONS(2538),
    [anon_sym_Class] = ACTIONS(2538),
    [anon_sym_id] = ACTIONS(2538),
  },
  [STATE(437)] = {
    [sym_identifier] = ACTIONS(2412),
    [aux_sym_preproc_include_token1] = ACTIONS(2412),
    [aux_sym_preproc_include_token2] = ACTIONS(2412),
    [aux_sym_preproc_def_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token2] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2412),
    [aux_sym_preproc_else_token1] = ACTIONS(2412),
    [aux_sym_preproc_elif_token1] = ACTIONS(2412),
    [sym_preproc_directive] = ACTIONS(2412),
    [anon_sym_LPAREN2] = ACTIONS(2414),
    [anon_sym_BANG] = ACTIONS(2414),
    [anon_sym_TILDE] = ACTIONS(2414),
    [anon_sym_DASH] = ACTIONS(2412),
    [anon_sym_PLUS] = ACTIONS(2412),
    [anon_sym_STAR] = ACTIONS(2414),
    [anon_sym_CARET] = ACTIONS(2414),
    [anon_sym_AMP] = ACTIONS(2414),
    [anon_sym_SEMI] = ACTIONS(2414),
    [anon_sym___extension__] = ACTIONS(2412),
    [anon_sym_typedef] = ACTIONS(2412),
    [anon_sym_extern] = ACTIONS(2412),
    [anon_sym___attribute__] = ACTIONS(2412),
    [anon_sym___attribute] = ACTIONS(2412),
    [anon_sym_noreturn] = ACTIONS(2412),
    [anon_sym_LBRACK] = ACTIONS(2414),
    [anon_sym___declspec] = ACTIONS(2412),
    [anon_sym___cdecl] = ACTIONS(2412),
    [anon_sym___clrcall] = ACTIONS(2412),
    [anon_sym___stdcall] = ACTIONS(2412),
    [anon_sym___fastcall] = ACTIONS(2412),
    [anon_sym___thiscall] = ACTIONS(2412),
    [anon_sym___vectorcall] = ACTIONS(2412),
    [anon_sym_LBRACE] = ACTIONS(2414),
    [anon_sym_signed] = ACTIONS(2412),
    [anon_sym_unsigned] = ACTIONS(2412),
    [anon_sym_long] = ACTIONS(2412),
    [anon_sym_short] = ACTIONS(2412),
    [anon_sym_static] = ACTIONS(2412),
    [anon_sym_ATautoreleasepool] = ACTIONS(2414),
    [anon_sym_auto] = ACTIONS(2412),
    [anon_sym_register] = ACTIONS(2412),
    [anon_sym_inline] = ACTIONS(2412),
    [anon_sym___inline] = ACTIONS(2412),
    [anon_sym___inline__] = ACTIONS(2412),
    [anon_sym___forceinline] = ACTIONS(2412),
    [anon_sym_thread_local] = ACTIONS(2412),
    [anon_sym___thread] = ACTIONS(2412),
    [anon_sym_CG_EXTERN] = ACTIONS(2412),
    [anon_sym_CG_INLINE] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2412),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2412),
    [anon_sym_IBOutlet] = ACTIONS(2412),
    [anon_sym_IBInspectable] = ACTIONS(2412),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2412),
    [anon_sym_NS_INLINE] = ACTIONS(2412),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2412),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2412),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2412),
    [anon_sym_const] = ACTIONS(2412),
    [anon_sym_constexpr] = ACTIONS(2412),
    [anon_sym_volatile] = ACTIONS(2412),
    [anon_sym_restrict] = ACTIONS(2412),
    [anon_sym___restrict__] = ACTIONS(2412),
    [anon_sym__Atomic] = ACTIONS(2412),
    [anon_sym__Noreturn] = ACTIONS(2412),
    [anon_sym__Nonnull] = ACTIONS(2412),
    [anon_sym_nullable] = ACTIONS(2412),
    [anon_sym__Complex] = ACTIONS(2412),
    [anon_sym__Nullable] = ACTIONS(2412),
    [anon_sym__Nullable_result] = ACTIONS(2412),
    [anon_sym__Null_unspecified] = ACTIONS(2412),
    [anon_sym___autoreleasing] = ACTIONS(2412),
    [anon_sym___block] = ACTIONS(2412),
    [anon_sym___bridge] = ACTIONS(2412),
    [anon_sym___bridge_retained] = ACTIONS(2412),
    [anon_sym___bridge_transfer] = ACTIONS(2412),
    [anon_sym___complex] = ACTIONS(2412),
    [anon_sym___const] = ACTIONS(2412),
    [anon_sym___imag] = ACTIONS(2412),
    [anon_sym___kindof] = ACTIONS(2412),
    [anon_sym___nonnull] = ACTIONS(2412),
    [anon_sym___nullable] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2412),
    [anon_sym___real] = ACTIONS(2412),
    [anon_sym___strong] = ACTIONS(2412),
    [anon_sym___unsafe_unretained] = ACTIONS(2412),
    [anon_sym___unused] = ACTIONS(2412),
    [anon_sym___weak] = ACTIONS(2412),
    [anon_sym_alignas] = ACTIONS(2412),
    [anon_sym__Alignas] = ACTIONS(2412),
    [sym_primitive_type] = ACTIONS(2412),
    [anon_sym_enum] = ACTIONS(2412),
    [anon_sym_struct] = ACTIONS(2412),
    [anon_sym_union] = ACTIONS(2412),
    [anon_sym_if] = ACTIONS(2412),
    [anon_sym_else] = ACTIONS(2412),
    [anon_sym_switch] = ACTIONS(2412),
    [anon_sym_case] = ACTIONS(2412),
    [anon_sym_default] = ACTIONS(2412),
    [anon_sym_while] = ACTIONS(2412),
    [anon_sym_do] = ACTIONS(2412),
    [anon_sym_for] = ACTIONS(2412),
    [anon_sym_in] = ACTIONS(2412),
    [anon_sym_return] = ACTIONS(2412),
    [anon_sym_break] = ACTIONS(2412),
    [anon_sym_continue] = ACTIONS(2412),
    [anon_sym_goto] = ACTIONS(2412),
    [anon_sym_DASH_DASH] = ACTIONS(2414),
    [anon_sym_PLUS_PLUS] = ACTIONS(2414),
    [anon_sym_sizeof] = ACTIONS(2412),
    [anon_sym___alignof__] = ACTIONS(2412),
    [anon_sym___alignof] = ACTIONS(2412),
    [anon_sym__alignof] = ACTIONS(2412),
    [anon_sym_alignof] = ACTIONS(2412),
    [anon_sym__Alignof] = ACTIONS(2412),
    [anon_sym_offsetof] = ACTIONS(2412),
    [anon_sym__Generic] = ACTIONS(2412),
    [anon_sym_asm] = ACTIONS(2412),
    [anon_sym___asm__] = ACTIONS(2412),
    [anon_sym___asm] = ACTIONS(2412),
    [sym_number_literal] = ACTIONS(2414),
    [anon_sym_L_SQUOTE] = ACTIONS(2414),
    [anon_sym_u_SQUOTE] = ACTIONS(2414),
    [anon_sym_U_SQUOTE] = ACTIONS(2414),
    [anon_sym_u8_SQUOTE] = ACTIONS(2414),
    [anon_sym_SQUOTE] = ACTIONS(2414),
    [anon_sym_AT] = ACTIONS(2412),
    [anon_sym_DQUOTE] = ACTIONS(2414),
    [anon_sym_L_DQUOTE] = ACTIONS(2414),
    [anon_sym_u_DQUOTE] = ACTIONS(2414),
    [anon_sym_U_DQUOTE] = ACTIONS(2414),
    [anon_sym_u8_DQUOTE] = ACTIONS(2414),
    [sym_true] = ACTIONS(2412),
    [sym_false] = ACTIONS(2412),
    [anon_sym_NULL] = ACTIONS(2412),
    [anon_sym_nullptr] = ACTIONS(2412),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2412),
    [anon_sym___typeof] = ACTIONS(2412),
    [anon_sym_typeof] = ACTIONS(2412),
    [anon_sym_ATimport] = ACTIONS(2414),
    [aux_sym_preproc_undef_token1] = ACTIONS(2412),
    [anon_sym_POUND] = ACTIONS(2412),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2412),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2412),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2412),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_API_AVAILABLE] = ACTIONS(2412),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_API_DEPRECATED] = ACTIONS(2412),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym___deprecated_msg] = ACTIONS(2412),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2412),
    [anon_sym_ATprotocol] = ACTIONS(2414),
    [anon_sym_ATinterface] = ACTIONS(2414),
    [anon_sym_ATimplementation] = ACTIONS(2414),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2414),
    [anon_sym_ATsynthesize] = ACTIONS(2414),
    [anon_sym_ATdynamic] = ACTIONS(2414),
    [anon_sym_ATtry] = ACTIONS(2414),
    [anon_sym___try] = ACTIONS(2412),
    [anon_sym_ATthrow] = ACTIONS(2414),
    [anon_sym_ATselector] = ACTIONS(2414),
    [anon_sym_ATavailable] = ACTIONS(2414),
    [anon_sym___builtin_available] = ACTIONS(2412),
    [anon_sym_va_arg] = ACTIONS(2412),
    [anon_sym_ATencode] = ACTIONS(2414),
    [anon_sym_ATsynchronized] = ACTIONS(2414),
    [anon_sym_BOOL] = ACTIONS(2412),
    [anon_sym_IMP] = ACTIONS(2412),
    [anon_sym_SEL] = ACTIONS(2412),
    [anon_sym_Class] = ACTIONS(2412),
    [anon_sym_id] = ACTIONS(2412),
  },
  [STATE(438)] = {
    [sym_identifier] = ACTIONS(2364),
    [aux_sym_preproc_include_token1] = ACTIONS(2364),
    [aux_sym_preproc_include_token2] = ACTIONS(2364),
    [aux_sym_preproc_def_token1] = ACTIONS(2364),
    [aux_sym_preproc_if_token1] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2364),
    [sym_preproc_directive] = ACTIONS(2364),
    [anon_sym_LPAREN2] = ACTIONS(2366),
    [anon_sym_BANG] = ACTIONS(2366),
    [anon_sym_TILDE] = ACTIONS(2366),
    [anon_sym_DASH] = ACTIONS(2364),
    [anon_sym_PLUS] = ACTIONS(2364),
    [anon_sym_STAR] = ACTIONS(2366),
    [anon_sym_CARET] = ACTIONS(2366),
    [anon_sym_AMP] = ACTIONS(2366),
    [anon_sym_SEMI] = ACTIONS(2366),
    [anon_sym___extension__] = ACTIONS(2364),
    [anon_sym_typedef] = ACTIONS(2364),
    [anon_sym_extern] = ACTIONS(2364),
    [anon_sym___attribute__] = ACTIONS(2364),
    [anon_sym___attribute] = ACTIONS(2364),
    [anon_sym_noreturn] = ACTIONS(2364),
    [anon_sym_LBRACK] = ACTIONS(2366),
    [anon_sym___declspec] = ACTIONS(2364),
    [anon_sym___cdecl] = ACTIONS(2364),
    [anon_sym___clrcall] = ACTIONS(2364),
    [anon_sym___stdcall] = ACTIONS(2364),
    [anon_sym___fastcall] = ACTIONS(2364),
    [anon_sym___thiscall] = ACTIONS(2364),
    [anon_sym___vectorcall] = ACTIONS(2364),
    [anon_sym_LBRACE] = ACTIONS(2366),
    [anon_sym_RBRACE] = ACTIONS(2366),
    [anon_sym_signed] = ACTIONS(2364),
    [anon_sym_unsigned] = ACTIONS(2364),
    [anon_sym_long] = ACTIONS(2364),
    [anon_sym_short] = ACTIONS(2364),
    [anon_sym_static] = ACTIONS(2364),
    [anon_sym_ATautoreleasepool] = ACTIONS(2366),
    [anon_sym_auto] = ACTIONS(2364),
    [anon_sym_register] = ACTIONS(2364),
    [anon_sym_inline] = ACTIONS(2364),
    [anon_sym___inline] = ACTIONS(2364),
    [anon_sym___inline__] = ACTIONS(2364),
    [anon_sym___forceinline] = ACTIONS(2364),
    [anon_sym_thread_local] = ACTIONS(2364),
    [anon_sym___thread] = ACTIONS(2364),
    [anon_sym_CG_EXTERN] = ACTIONS(2364),
    [anon_sym_CG_INLINE] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2364),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2364),
    [anon_sym_IBOutlet] = ACTIONS(2364),
    [anon_sym_IBInspectable] = ACTIONS(2364),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2364),
    [anon_sym_NS_INLINE] = ACTIONS(2364),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2364),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2364),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2364),
    [anon_sym_const] = ACTIONS(2364),
    [anon_sym_constexpr] = ACTIONS(2364),
    [anon_sym_volatile] = ACTIONS(2364),
    [anon_sym_restrict] = ACTIONS(2364),
    [anon_sym___restrict__] = ACTIONS(2364),
    [anon_sym__Atomic] = ACTIONS(2364),
    [anon_sym__Noreturn] = ACTIONS(2364),
    [anon_sym__Nonnull] = ACTIONS(2364),
    [anon_sym_nullable] = ACTIONS(2364),
    [anon_sym__Complex] = ACTIONS(2364),
    [anon_sym__Nullable] = ACTIONS(2364),
    [anon_sym__Nullable_result] = ACTIONS(2364),
    [anon_sym__Null_unspecified] = ACTIONS(2364),
    [anon_sym___autoreleasing] = ACTIONS(2364),
    [anon_sym___block] = ACTIONS(2364),
    [anon_sym___bridge] = ACTIONS(2364),
    [anon_sym___bridge_retained] = ACTIONS(2364),
    [anon_sym___bridge_transfer] = ACTIONS(2364),
    [anon_sym___complex] = ACTIONS(2364),
    [anon_sym___const] = ACTIONS(2364),
    [anon_sym___imag] = ACTIONS(2364),
    [anon_sym___kindof] = ACTIONS(2364),
    [anon_sym___nonnull] = ACTIONS(2364),
    [anon_sym___nullable] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2364),
    [anon_sym___real] = ACTIONS(2364),
    [anon_sym___strong] = ACTIONS(2364),
    [anon_sym___unsafe_unretained] = ACTIONS(2364),
    [anon_sym___unused] = ACTIONS(2364),
    [anon_sym___weak] = ACTIONS(2364),
    [anon_sym_alignas] = ACTIONS(2364),
    [anon_sym__Alignas] = ACTIONS(2364),
    [sym_primitive_type] = ACTIONS(2364),
    [anon_sym_enum] = ACTIONS(2364),
    [anon_sym_struct] = ACTIONS(2364),
    [anon_sym_union] = ACTIONS(2364),
    [anon_sym_if] = ACTIONS(2364),
    [anon_sym_else] = ACTIONS(2364),
    [anon_sym_switch] = ACTIONS(2364),
    [anon_sym_case] = ACTIONS(2364),
    [anon_sym_default] = ACTIONS(2364),
    [anon_sym_while] = ACTIONS(2364),
    [anon_sym_do] = ACTIONS(2364),
    [anon_sym_for] = ACTIONS(2364),
    [anon_sym_in] = ACTIONS(2364),
    [anon_sym_return] = ACTIONS(2364),
    [anon_sym_break] = ACTIONS(2364),
    [anon_sym_continue] = ACTIONS(2364),
    [anon_sym_goto] = ACTIONS(2364),
    [anon_sym_DASH_DASH] = ACTIONS(2366),
    [anon_sym_PLUS_PLUS] = ACTIONS(2366),
    [anon_sym_sizeof] = ACTIONS(2364),
    [anon_sym___alignof__] = ACTIONS(2364),
    [anon_sym___alignof] = ACTIONS(2364),
    [anon_sym__alignof] = ACTIONS(2364),
    [anon_sym_alignof] = ACTIONS(2364),
    [anon_sym__Alignof] = ACTIONS(2364),
    [anon_sym_offsetof] = ACTIONS(2364),
    [anon_sym__Generic] = ACTIONS(2364),
    [anon_sym_asm] = ACTIONS(2364),
    [anon_sym___asm__] = ACTIONS(2364),
    [anon_sym___asm] = ACTIONS(2364),
    [sym_number_literal] = ACTIONS(2366),
    [anon_sym_L_SQUOTE] = ACTIONS(2366),
    [anon_sym_u_SQUOTE] = ACTIONS(2366),
    [anon_sym_U_SQUOTE] = ACTIONS(2366),
    [anon_sym_u8_SQUOTE] = ACTIONS(2366),
    [anon_sym_SQUOTE] = ACTIONS(2366),
    [anon_sym_AT] = ACTIONS(2364),
    [anon_sym_DQUOTE] = ACTIONS(2366),
    [anon_sym_L_DQUOTE] = ACTIONS(2366),
    [anon_sym_u_DQUOTE] = ACTIONS(2366),
    [anon_sym_U_DQUOTE] = ACTIONS(2366),
    [anon_sym_u8_DQUOTE] = ACTIONS(2366),
    [sym_true] = ACTIONS(2364),
    [sym_false] = ACTIONS(2364),
    [anon_sym_NULL] = ACTIONS(2364),
    [anon_sym_nullptr] = ACTIONS(2364),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2364),
    [anon_sym___typeof] = ACTIONS(2364),
    [anon_sym_typeof] = ACTIONS(2364),
    [anon_sym_ATimport] = ACTIONS(2366),
    [aux_sym_preproc_undef_token1] = ACTIONS(2364),
    [anon_sym_POUND] = ACTIONS(2364),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2364),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2364),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2364),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_API_AVAILABLE] = ACTIONS(2364),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_API_DEPRECATED] = ACTIONS(2364),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym___deprecated_msg] = ACTIONS(2364),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2364),
    [anon_sym_ATprotocol] = ACTIONS(2366),
    [anon_sym_ATinterface] = ACTIONS(2366),
    [anon_sym_ATimplementation] = ACTIONS(2366),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2366),
    [anon_sym_ATtry] = ACTIONS(2366),
    [anon_sym___try] = ACTIONS(2364),
    [anon_sym_ATcatch] = ACTIONS(2366),
    [anon_sym___catch] = ACTIONS(2364),
    [anon_sym_ATfinally] = ACTIONS(2366),
    [anon_sym___finally] = ACTIONS(2364),
    [anon_sym_ATthrow] = ACTIONS(2366),
    [anon_sym_ATselector] = ACTIONS(2366),
    [anon_sym_ATavailable] = ACTIONS(2366),
    [anon_sym___builtin_available] = ACTIONS(2364),
    [anon_sym_va_arg] = ACTIONS(2364),
    [anon_sym_ATencode] = ACTIONS(2366),
    [anon_sym_ATsynchronized] = ACTIONS(2366),
    [anon_sym_BOOL] = ACTIONS(2364),
    [anon_sym_IMP] = ACTIONS(2364),
    [anon_sym_SEL] = ACTIONS(2364),
    [anon_sym_Class] = ACTIONS(2364),
    [anon_sym_id] = ACTIONS(2364),
  },
  [STATE(439)] = {
    [sym_identifier] = ACTIONS(2454),
    [aux_sym_preproc_include_token1] = ACTIONS(2454),
    [aux_sym_preproc_include_token2] = ACTIONS(2454),
    [aux_sym_preproc_def_token1] = ACTIONS(2454),
    [aux_sym_preproc_if_token1] = ACTIONS(2454),
    [aux_sym_preproc_if_token2] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2454),
    [aux_sym_preproc_else_token1] = ACTIONS(2454),
    [aux_sym_preproc_elif_token1] = ACTIONS(2454),
    [sym_preproc_directive] = ACTIONS(2454),
    [anon_sym_LPAREN2] = ACTIONS(2456),
    [anon_sym_BANG] = ACTIONS(2456),
    [anon_sym_TILDE] = ACTIONS(2456),
    [anon_sym_DASH] = ACTIONS(2454),
    [anon_sym_PLUS] = ACTIONS(2454),
    [anon_sym_STAR] = ACTIONS(2456),
    [anon_sym_CARET] = ACTIONS(2456),
    [anon_sym_AMP] = ACTIONS(2456),
    [anon_sym_SEMI] = ACTIONS(2456),
    [anon_sym___extension__] = ACTIONS(2454),
    [anon_sym_typedef] = ACTIONS(2454),
    [anon_sym_extern] = ACTIONS(2454),
    [anon_sym___attribute__] = ACTIONS(2454),
    [anon_sym___attribute] = ACTIONS(2454),
    [anon_sym_noreturn] = ACTIONS(2454),
    [anon_sym_LBRACK] = ACTIONS(2456),
    [anon_sym___declspec] = ACTIONS(2454),
    [anon_sym___cdecl] = ACTIONS(2454),
    [anon_sym___clrcall] = ACTIONS(2454),
    [anon_sym___stdcall] = ACTIONS(2454),
    [anon_sym___fastcall] = ACTIONS(2454),
    [anon_sym___thiscall] = ACTIONS(2454),
    [anon_sym___vectorcall] = ACTIONS(2454),
    [anon_sym_LBRACE] = ACTIONS(2456),
    [anon_sym_signed] = ACTIONS(2454),
    [anon_sym_unsigned] = ACTIONS(2454),
    [anon_sym_long] = ACTIONS(2454),
    [anon_sym_short] = ACTIONS(2454),
    [anon_sym_static] = ACTIONS(2454),
    [anon_sym_ATautoreleasepool] = ACTIONS(2456),
    [anon_sym_auto] = ACTIONS(2454),
    [anon_sym_register] = ACTIONS(2454),
    [anon_sym_inline] = ACTIONS(2454),
    [anon_sym___inline] = ACTIONS(2454),
    [anon_sym___inline__] = ACTIONS(2454),
    [anon_sym___forceinline] = ACTIONS(2454),
    [anon_sym_thread_local] = ACTIONS(2454),
    [anon_sym___thread] = ACTIONS(2454),
    [anon_sym_CG_EXTERN] = ACTIONS(2454),
    [anon_sym_CG_INLINE] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2454),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2454),
    [anon_sym_IBOutlet] = ACTIONS(2454),
    [anon_sym_IBInspectable] = ACTIONS(2454),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2454),
    [anon_sym_NS_INLINE] = ACTIONS(2454),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2454),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2454),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2454),
    [anon_sym_const] = ACTIONS(2454),
    [anon_sym_constexpr] = ACTIONS(2454),
    [anon_sym_volatile] = ACTIONS(2454),
    [anon_sym_restrict] = ACTIONS(2454),
    [anon_sym___restrict__] = ACTIONS(2454),
    [anon_sym__Atomic] = ACTIONS(2454),
    [anon_sym__Noreturn] = ACTIONS(2454),
    [anon_sym__Nonnull] = ACTIONS(2454),
    [anon_sym_nullable] = ACTIONS(2454),
    [anon_sym__Complex] = ACTIONS(2454),
    [anon_sym__Nullable] = ACTIONS(2454),
    [anon_sym__Nullable_result] = ACTIONS(2454),
    [anon_sym__Null_unspecified] = ACTIONS(2454),
    [anon_sym___autoreleasing] = ACTIONS(2454),
    [anon_sym___block] = ACTIONS(2454),
    [anon_sym___bridge] = ACTIONS(2454),
    [anon_sym___bridge_retained] = ACTIONS(2454),
    [anon_sym___bridge_transfer] = ACTIONS(2454),
    [anon_sym___complex] = ACTIONS(2454),
    [anon_sym___const] = ACTIONS(2454),
    [anon_sym___imag] = ACTIONS(2454),
    [anon_sym___kindof] = ACTIONS(2454),
    [anon_sym___nonnull] = ACTIONS(2454),
    [anon_sym___nullable] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2454),
    [anon_sym___real] = ACTIONS(2454),
    [anon_sym___strong] = ACTIONS(2454),
    [anon_sym___unsafe_unretained] = ACTIONS(2454),
    [anon_sym___unused] = ACTIONS(2454),
    [anon_sym___weak] = ACTIONS(2454),
    [anon_sym_alignas] = ACTIONS(2454),
    [anon_sym__Alignas] = ACTIONS(2454),
    [sym_primitive_type] = ACTIONS(2454),
    [anon_sym_enum] = ACTIONS(2454),
    [anon_sym_struct] = ACTIONS(2454),
    [anon_sym_union] = ACTIONS(2454),
    [anon_sym_if] = ACTIONS(2454),
    [anon_sym_else] = ACTIONS(2454),
    [anon_sym_switch] = ACTIONS(2454),
    [anon_sym_case] = ACTIONS(2454),
    [anon_sym_default] = ACTIONS(2454),
    [anon_sym_while] = ACTIONS(2454),
    [anon_sym_do] = ACTIONS(2454),
    [anon_sym_for] = ACTIONS(2454),
    [anon_sym_in] = ACTIONS(2454),
    [anon_sym_return] = ACTIONS(2454),
    [anon_sym_break] = ACTIONS(2454),
    [anon_sym_continue] = ACTIONS(2454),
    [anon_sym_goto] = ACTIONS(2454),
    [anon_sym_DASH_DASH] = ACTIONS(2456),
    [anon_sym_PLUS_PLUS] = ACTIONS(2456),
    [anon_sym_sizeof] = ACTIONS(2454),
    [anon_sym___alignof__] = ACTIONS(2454),
    [anon_sym___alignof] = ACTIONS(2454),
    [anon_sym__alignof] = ACTIONS(2454),
    [anon_sym_alignof] = ACTIONS(2454),
    [anon_sym__Alignof] = ACTIONS(2454),
    [anon_sym_offsetof] = ACTIONS(2454),
    [anon_sym__Generic] = ACTIONS(2454),
    [anon_sym_asm] = ACTIONS(2454),
    [anon_sym___asm__] = ACTIONS(2454),
    [anon_sym___asm] = ACTIONS(2454),
    [sym_number_literal] = ACTIONS(2456),
    [anon_sym_L_SQUOTE] = ACTIONS(2456),
    [anon_sym_u_SQUOTE] = ACTIONS(2456),
    [anon_sym_U_SQUOTE] = ACTIONS(2456),
    [anon_sym_u8_SQUOTE] = ACTIONS(2456),
    [anon_sym_SQUOTE] = ACTIONS(2456),
    [anon_sym_AT] = ACTIONS(2454),
    [anon_sym_DQUOTE] = ACTIONS(2456),
    [anon_sym_L_DQUOTE] = ACTIONS(2456),
    [anon_sym_u_DQUOTE] = ACTIONS(2456),
    [anon_sym_U_DQUOTE] = ACTIONS(2456),
    [anon_sym_u8_DQUOTE] = ACTIONS(2456),
    [sym_true] = ACTIONS(2454),
    [sym_false] = ACTIONS(2454),
    [anon_sym_NULL] = ACTIONS(2454),
    [anon_sym_nullptr] = ACTIONS(2454),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2454),
    [anon_sym___typeof] = ACTIONS(2454),
    [anon_sym_typeof] = ACTIONS(2454),
    [anon_sym_ATimport] = ACTIONS(2456),
    [aux_sym_preproc_undef_token1] = ACTIONS(2454),
    [anon_sym_POUND] = ACTIONS(2454),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2454),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2454),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2454),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_API_AVAILABLE] = ACTIONS(2454),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_API_DEPRECATED] = ACTIONS(2454),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym___deprecated_msg] = ACTIONS(2454),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2454),
    [anon_sym_ATprotocol] = ACTIONS(2456),
    [anon_sym_ATinterface] = ACTIONS(2456),
    [anon_sym_ATimplementation] = ACTIONS(2456),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2456),
    [anon_sym_ATsynthesize] = ACTIONS(2456),
    [anon_sym_ATdynamic] = ACTIONS(2456),
    [anon_sym_ATtry] = ACTIONS(2456),
    [anon_sym___try] = ACTIONS(2454),
    [anon_sym_ATthrow] = ACTIONS(2456),
    [anon_sym_ATselector] = ACTIONS(2456),
    [anon_sym_ATavailable] = ACTIONS(2456),
    [anon_sym___builtin_available] = ACTIONS(2454),
    [anon_sym_va_arg] = ACTIONS(2454),
    [anon_sym_ATencode] = ACTIONS(2456),
    [anon_sym_ATsynchronized] = ACTIONS(2456),
    [anon_sym_BOOL] = ACTIONS(2454),
    [anon_sym_IMP] = ACTIONS(2454),
    [anon_sym_SEL] = ACTIONS(2454),
    [anon_sym_Class] = ACTIONS(2454),
    [anon_sym_id] = ACTIONS(2454),
  },
  [STATE(440)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token2] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [aux_sym_preproc_else_token1] = ACTIONS(2438),
    [aux_sym_preproc_elif_token1] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(441)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [aux_sym_preproc_else_token1] = ACTIONS(2408),
    [aux_sym_preproc_elif_token1] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(442)] = {
    [sym_identifier] = ACTIONS(2696),
    [aux_sym_preproc_include_token1] = ACTIONS(2696),
    [aux_sym_preproc_include_token2] = ACTIONS(2696),
    [aux_sym_preproc_def_token1] = ACTIONS(2696),
    [aux_sym_preproc_if_token1] = ACTIONS(2696),
    [aux_sym_preproc_if_token2] = ACTIONS(2696),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2696),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2696),
    [aux_sym_preproc_else_token1] = ACTIONS(2696),
    [aux_sym_preproc_elif_token1] = ACTIONS(2696),
    [sym_preproc_directive] = ACTIONS(2696),
    [anon_sym_LPAREN2] = ACTIONS(2698),
    [anon_sym_BANG] = ACTIONS(2698),
    [anon_sym_TILDE] = ACTIONS(2698),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2698),
    [anon_sym_CARET] = ACTIONS(2698),
    [anon_sym_AMP] = ACTIONS(2698),
    [anon_sym_SEMI] = ACTIONS(2698),
    [anon_sym___extension__] = ACTIONS(2696),
    [anon_sym_typedef] = ACTIONS(2696),
    [anon_sym_extern] = ACTIONS(2696),
    [anon_sym___attribute__] = ACTIONS(2696),
    [anon_sym___attribute] = ACTIONS(2696),
    [anon_sym_noreturn] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym___declspec] = ACTIONS(2696),
    [anon_sym___cdecl] = ACTIONS(2696),
    [anon_sym___clrcall] = ACTIONS(2696),
    [anon_sym___stdcall] = ACTIONS(2696),
    [anon_sym___fastcall] = ACTIONS(2696),
    [anon_sym___thiscall] = ACTIONS(2696),
    [anon_sym___vectorcall] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_signed] = ACTIONS(2696),
    [anon_sym_unsigned] = ACTIONS(2696),
    [anon_sym_long] = ACTIONS(2696),
    [anon_sym_short] = ACTIONS(2696),
    [anon_sym_static] = ACTIONS(2696),
    [anon_sym_ATautoreleasepool] = ACTIONS(2698),
    [anon_sym_auto] = ACTIONS(2696),
    [anon_sym_register] = ACTIONS(2696),
    [anon_sym_inline] = ACTIONS(2696),
    [anon_sym___inline] = ACTIONS(2696),
    [anon_sym___inline__] = ACTIONS(2696),
    [anon_sym___forceinline] = ACTIONS(2696),
    [anon_sym_thread_local] = ACTIONS(2696),
    [anon_sym___thread] = ACTIONS(2696),
    [anon_sym_CG_EXTERN] = ACTIONS(2696),
    [anon_sym_CG_INLINE] = ACTIONS(2696),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2696),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2696),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2696),
    [anon_sym_IBOutlet] = ACTIONS(2696),
    [anon_sym_IBInspectable] = ACTIONS(2696),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2696),
    [anon_sym_NS_INLINE] = ACTIONS(2696),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2696),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2696),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2696),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2696),
    [anon_sym_const] = ACTIONS(2696),
    [anon_sym_constexpr] = ACTIONS(2696),
    [anon_sym_volatile] = ACTIONS(2696),
    [anon_sym_restrict] = ACTIONS(2696),
    [anon_sym___restrict__] = ACTIONS(2696),
    [anon_sym__Atomic] = ACTIONS(2696),
    [anon_sym__Noreturn] = ACTIONS(2696),
    [anon_sym__Nonnull] = ACTIONS(2696),
    [anon_sym_nullable] = ACTIONS(2696),
    [anon_sym__Complex] = ACTIONS(2696),
    [anon_sym__Nullable] = ACTIONS(2696),
    [anon_sym__Nullable_result] = ACTIONS(2696),
    [anon_sym__Null_unspecified] = ACTIONS(2696),
    [anon_sym___autoreleasing] = ACTIONS(2696),
    [anon_sym___block] = ACTIONS(2696),
    [anon_sym___bridge] = ACTIONS(2696),
    [anon_sym___bridge_retained] = ACTIONS(2696),
    [anon_sym___bridge_transfer] = ACTIONS(2696),
    [anon_sym___complex] = ACTIONS(2696),
    [anon_sym___const] = ACTIONS(2696),
    [anon_sym___imag] = ACTIONS(2696),
    [anon_sym___kindof] = ACTIONS(2696),
    [anon_sym___nonnull] = ACTIONS(2696),
    [anon_sym___nullable] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2696),
    [anon_sym___real] = ACTIONS(2696),
    [anon_sym___strong] = ACTIONS(2696),
    [anon_sym___unsafe_unretained] = ACTIONS(2696),
    [anon_sym___unused] = ACTIONS(2696),
    [anon_sym___weak] = ACTIONS(2696),
    [anon_sym_alignas] = ACTIONS(2696),
    [anon_sym__Alignas] = ACTIONS(2696),
    [sym_primitive_type] = ACTIONS(2696),
    [anon_sym_enum] = ACTIONS(2696),
    [anon_sym_struct] = ACTIONS(2696),
    [anon_sym_union] = ACTIONS(2696),
    [anon_sym_if] = ACTIONS(2696),
    [anon_sym_switch] = ACTIONS(2696),
    [anon_sym_case] = ACTIONS(2696),
    [anon_sym_default] = ACTIONS(2696),
    [anon_sym_while] = ACTIONS(2696),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_for] = ACTIONS(2696),
    [anon_sym_in] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_sizeof] = ACTIONS(2696),
    [anon_sym___alignof__] = ACTIONS(2696),
    [anon_sym___alignof] = ACTIONS(2696),
    [anon_sym__alignof] = ACTIONS(2696),
    [anon_sym_alignof] = ACTIONS(2696),
    [anon_sym__Alignof] = ACTIONS(2696),
    [anon_sym_offsetof] = ACTIONS(2696),
    [anon_sym__Generic] = ACTIONS(2696),
    [anon_sym_asm] = ACTIONS(2696),
    [anon_sym___asm__] = ACTIONS(2696),
    [anon_sym___asm] = ACTIONS(2696),
    [sym_number_literal] = ACTIONS(2698),
    [anon_sym_L_SQUOTE] = ACTIONS(2698),
    [anon_sym_u_SQUOTE] = ACTIONS(2698),
    [anon_sym_U_SQUOTE] = ACTIONS(2698),
    [anon_sym_u8_SQUOTE] = ACTIONS(2698),
    [anon_sym_SQUOTE] = ACTIONS(2698),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_DQUOTE] = ACTIONS(2698),
    [anon_sym_L_DQUOTE] = ACTIONS(2698),
    [anon_sym_u_DQUOTE] = ACTIONS(2698),
    [anon_sym_U_DQUOTE] = ACTIONS(2698),
    [anon_sym_u8_DQUOTE] = ACTIONS(2698),
    [sym_true] = ACTIONS(2696),
    [sym_false] = ACTIONS(2696),
    [anon_sym_NULL] = ACTIONS(2696),
    [anon_sym_nullptr] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2696),
    [anon_sym___typeof] = ACTIONS(2696),
    [anon_sym_typeof] = ACTIONS(2696),
    [anon_sym_ATimport] = ACTIONS(2698),
    [aux_sym_preproc_undef_token1] = ACTIONS(2696),
    [anon_sym_POUND] = ACTIONS(2696),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2696),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2696),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2696),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2696),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2696),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2696),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2696),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_API_AVAILABLE] = ACTIONS(2696),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_API_DEPRECATED] = ACTIONS(2696),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2696),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym___deprecated_msg] = ACTIONS(2696),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2696),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2696),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2696),
    [anon_sym_ATprotocol] = ACTIONS(2698),
    [anon_sym_ATinterface] = ACTIONS(2698),
    [anon_sym_ATimplementation] = ACTIONS(2698),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2698),
    [anon_sym_ATsynthesize] = ACTIONS(2698),
    [anon_sym_ATdynamic] = ACTIONS(2698),
    [anon_sym_ATtry] = ACTIONS(2698),
    [anon_sym___try] = ACTIONS(2696),
    [anon_sym_ATthrow] = ACTIONS(2698),
    [anon_sym_ATselector] = ACTIONS(2698),
    [anon_sym_ATavailable] = ACTIONS(2698),
    [anon_sym___builtin_available] = ACTIONS(2696),
    [anon_sym_va_arg] = ACTIONS(2696),
    [anon_sym_ATencode] = ACTIONS(2698),
    [anon_sym_ATsynchronized] = ACTIONS(2698),
    [anon_sym_BOOL] = ACTIONS(2696),
    [anon_sym_IMP] = ACTIONS(2696),
    [anon_sym_SEL] = ACTIONS(2696),
    [anon_sym_Class] = ACTIONS(2696),
    [anon_sym_id] = ACTIONS(2696),
  },
  [STATE(443)] = {
    [sym_identifier] = ACTIONS(2720),
    [aux_sym_preproc_include_token1] = ACTIONS(2720),
    [aux_sym_preproc_include_token2] = ACTIONS(2720),
    [aux_sym_preproc_def_token1] = ACTIONS(2720),
    [aux_sym_preproc_if_token1] = ACTIONS(2720),
    [aux_sym_preproc_if_token2] = ACTIONS(2720),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2720),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2720),
    [aux_sym_preproc_else_token1] = ACTIONS(2720),
    [aux_sym_preproc_elif_token1] = ACTIONS(2720),
    [sym_preproc_directive] = ACTIONS(2720),
    [anon_sym_LPAREN2] = ACTIONS(2722),
    [anon_sym_BANG] = ACTIONS(2722),
    [anon_sym_TILDE] = ACTIONS(2722),
    [anon_sym_DASH] = ACTIONS(2720),
    [anon_sym_PLUS] = ACTIONS(2720),
    [anon_sym_STAR] = ACTIONS(2722),
    [anon_sym_CARET] = ACTIONS(2722),
    [anon_sym_AMP] = ACTIONS(2722),
    [anon_sym_SEMI] = ACTIONS(2722),
    [anon_sym___extension__] = ACTIONS(2720),
    [anon_sym_typedef] = ACTIONS(2720),
    [anon_sym_extern] = ACTIONS(2720),
    [anon_sym___attribute__] = ACTIONS(2720),
    [anon_sym___attribute] = ACTIONS(2720),
    [anon_sym_noreturn] = ACTIONS(2720),
    [anon_sym_LBRACK] = ACTIONS(2722),
    [anon_sym___declspec] = ACTIONS(2720),
    [anon_sym___cdecl] = ACTIONS(2720),
    [anon_sym___clrcall] = ACTIONS(2720),
    [anon_sym___stdcall] = ACTIONS(2720),
    [anon_sym___fastcall] = ACTIONS(2720),
    [anon_sym___thiscall] = ACTIONS(2720),
    [anon_sym___vectorcall] = ACTIONS(2720),
    [anon_sym_LBRACE] = ACTIONS(2722),
    [anon_sym_signed] = ACTIONS(2720),
    [anon_sym_unsigned] = ACTIONS(2720),
    [anon_sym_long] = ACTIONS(2720),
    [anon_sym_short] = ACTIONS(2720),
    [anon_sym_static] = ACTIONS(2720),
    [anon_sym_ATautoreleasepool] = ACTIONS(2722),
    [anon_sym_auto] = ACTIONS(2720),
    [anon_sym_register] = ACTIONS(2720),
    [anon_sym_inline] = ACTIONS(2720),
    [anon_sym___inline] = ACTIONS(2720),
    [anon_sym___inline__] = ACTIONS(2720),
    [anon_sym___forceinline] = ACTIONS(2720),
    [anon_sym_thread_local] = ACTIONS(2720),
    [anon_sym___thread] = ACTIONS(2720),
    [anon_sym_CG_EXTERN] = ACTIONS(2720),
    [anon_sym_CG_INLINE] = ACTIONS(2720),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2720),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2720),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2720),
    [anon_sym_IBOutlet] = ACTIONS(2720),
    [anon_sym_IBInspectable] = ACTIONS(2720),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2720),
    [anon_sym_NS_INLINE] = ACTIONS(2720),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2720),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2720),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2720),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2720),
    [anon_sym_const] = ACTIONS(2720),
    [anon_sym_constexpr] = ACTIONS(2720),
    [anon_sym_volatile] = ACTIONS(2720),
    [anon_sym_restrict] = ACTIONS(2720),
    [anon_sym___restrict__] = ACTIONS(2720),
    [anon_sym__Atomic] = ACTIONS(2720),
    [anon_sym__Noreturn] = ACTIONS(2720),
    [anon_sym__Nonnull] = ACTIONS(2720),
    [anon_sym_nullable] = ACTIONS(2720),
    [anon_sym__Complex] = ACTIONS(2720),
    [anon_sym__Nullable] = ACTIONS(2720),
    [anon_sym__Nullable_result] = ACTIONS(2720),
    [anon_sym__Null_unspecified] = ACTIONS(2720),
    [anon_sym___autoreleasing] = ACTIONS(2720),
    [anon_sym___block] = ACTIONS(2720),
    [anon_sym___bridge] = ACTIONS(2720),
    [anon_sym___bridge_retained] = ACTIONS(2720),
    [anon_sym___bridge_transfer] = ACTIONS(2720),
    [anon_sym___complex] = ACTIONS(2720),
    [anon_sym___const] = ACTIONS(2720),
    [anon_sym___imag] = ACTIONS(2720),
    [anon_sym___kindof] = ACTIONS(2720),
    [anon_sym___nonnull] = ACTIONS(2720),
    [anon_sym___nullable] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2720),
    [anon_sym___real] = ACTIONS(2720),
    [anon_sym___strong] = ACTIONS(2720),
    [anon_sym___unsafe_unretained] = ACTIONS(2720),
    [anon_sym___unused] = ACTIONS(2720),
    [anon_sym___weak] = ACTIONS(2720),
    [anon_sym_alignas] = ACTIONS(2720),
    [anon_sym__Alignas] = ACTIONS(2720),
    [sym_primitive_type] = ACTIONS(2720),
    [anon_sym_enum] = ACTIONS(2720),
    [anon_sym_struct] = ACTIONS(2720),
    [anon_sym_union] = ACTIONS(2720),
    [anon_sym_if] = ACTIONS(2720),
    [anon_sym_switch] = ACTIONS(2720),
    [anon_sym_case] = ACTIONS(2720),
    [anon_sym_default] = ACTIONS(2720),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_do] = ACTIONS(2720),
    [anon_sym_for] = ACTIONS(2720),
    [anon_sym_in] = ACTIONS(2720),
    [anon_sym_return] = ACTIONS(2720),
    [anon_sym_break] = ACTIONS(2720),
    [anon_sym_continue] = ACTIONS(2720),
    [anon_sym_goto] = ACTIONS(2720),
    [anon_sym_DASH_DASH] = ACTIONS(2722),
    [anon_sym_PLUS_PLUS] = ACTIONS(2722),
    [anon_sym_sizeof] = ACTIONS(2720),
    [anon_sym___alignof__] = ACTIONS(2720),
    [anon_sym___alignof] = ACTIONS(2720),
    [anon_sym__alignof] = ACTIONS(2720),
    [anon_sym_alignof] = ACTIONS(2720),
    [anon_sym__Alignof] = ACTIONS(2720),
    [anon_sym_offsetof] = ACTIONS(2720),
    [anon_sym__Generic] = ACTIONS(2720),
    [anon_sym_asm] = ACTIONS(2720),
    [anon_sym___asm__] = ACTIONS(2720),
    [anon_sym___asm] = ACTIONS(2720),
    [sym_number_literal] = ACTIONS(2722),
    [anon_sym_L_SQUOTE] = ACTIONS(2722),
    [anon_sym_u_SQUOTE] = ACTIONS(2722),
    [anon_sym_U_SQUOTE] = ACTIONS(2722),
    [anon_sym_u8_SQUOTE] = ACTIONS(2722),
    [anon_sym_SQUOTE] = ACTIONS(2722),
    [anon_sym_AT] = ACTIONS(2720),
    [anon_sym_DQUOTE] = ACTIONS(2722),
    [anon_sym_L_DQUOTE] = ACTIONS(2722),
    [anon_sym_u_DQUOTE] = ACTIONS(2722),
    [anon_sym_U_DQUOTE] = ACTIONS(2722),
    [anon_sym_u8_DQUOTE] = ACTIONS(2722),
    [sym_true] = ACTIONS(2720),
    [sym_false] = ACTIONS(2720),
    [anon_sym_NULL] = ACTIONS(2720),
    [anon_sym_nullptr] = ACTIONS(2720),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2720),
    [anon_sym___typeof] = ACTIONS(2720),
    [anon_sym_typeof] = ACTIONS(2720),
    [anon_sym_ATimport] = ACTIONS(2722),
    [aux_sym_preproc_undef_token1] = ACTIONS(2720),
    [anon_sym_POUND] = ACTIONS(2720),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2720),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2720),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2720),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2720),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2720),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2720),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2720),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_API_AVAILABLE] = ACTIONS(2720),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_API_DEPRECATED] = ACTIONS(2720),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2720),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym___deprecated_msg] = ACTIONS(2720),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2720),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2720),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2720),
    [anon_sym_ATprotocol] = ACTIONS(2722),
    [anon_sym_ATinterface] = ACTIONS(2722),
    [anon_sym_ATimplementation] = ACTIONS(2722),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2722),
    [anon_sym_ATsynthesize] = ACTIONS(2722),
    [anon_sym_ATdynamic] = ACTIONS(2722),
    [anon_sym_ATtry] = ACTIONS(2722),
    [anon_sym___try] = ACTIONS(2720),
    [anon_sym_ATthrow] = ACTIONS(2722),
    [anon_sym_ATselector] = ACTIONS(2722),
    [anon_sym_ATavailable] = ACTIONS(2722),
    [anon_sym___builtin_available] = ACTIONS(2720),
    [anon_sym_va_arg] = ACTIONS(2720),
    [anon_sym_ATencode] = ACTIONS(2722),
    [anon_sym_ATsynchronized] = ACTIONS(2722),
    [anon_sym_BOOL] = ACTIONS(2720),
    [anon_sym_IMP] = ACTIONS(2720),
    [anon_sym_SEL] = ACTIONS(2720),
    [anon_sym_Class] = ACTIONS(2720),
    [anon_sym_id] = ACTIONS(2720),
  },
  [STATE(444)] = {
    [sym_identifier] = ACTIONS(2724),
    [aux_sym_preproc_include_token1] = ACTIONS(2724),
    [aux_sym_preproc_include_token2] = ACTIONS(2724),
    [aux_sym_preproc_def_token1] = ACTIONS(2724),
    [aux_sym_preproc_if_token1] = ACTIONS(2724),
    [aux_sym_preproc_if_token2] = ACTIONS(2724),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2724),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2724),
    [aux_sym_preproc_else_token1] = ACTIONS(2724),
    [aux_sym_preproc_elif_token1] = ACTIONS(2724),
    [sym_preproc_directive] = ACTIONS(2724),
    [anon_sym_LPAREN2] = ACTIONS(2726),
    [anon_sym_BANG] = ACTIONS(2726),
    [anon_sym_TILDE] = ACTIONS(2726),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2726),
    [anon_sym_CARET] = ACTIONS(2726),
    [anon_sym_AMP] = ACTIONS(2726),
    [anon_sym_SEMI] = ACTIONS(2726),
    [anon_sym___extension__] = ACTIONS(2724),
    [anon_sym_typedef] = ACTIONS(2724),
    [anon_sym_extern] = ACTIONS(2724),
    [anon_sym___attribute__] = ACTIONS(2724),
    [anon_sym___attribute] = ACTIONS(2724),
    [anon_sym_noreturn] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2726),
    [anon_sym___declspec] = ACTIONS(2724),
    [anon_sym___cdecl] = ACTIONS(2724),
    [anon_sym___clrcall] = ACTIONS(2724),
    [anon_sym___stdcall] = ACTIONS(2724),
    [anon_sym___fastcall] = ACTIONS(2724),
    [anon_sym___thiscall] = ACTIONS(2724),
    [anon_sym___vectorcall] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2726),
    [anon_sym_signed] = ACTIONS(2724),
    [anon_sym_unsigned] = ACTIONS(2724),
    [anon_sym_long] = ACTIONS(2724),
    [anon_sym_short] = ACTIONS(2724),
    [anon_sym_static] = ACTIONS(2724),
    [anon_sym_ATautoreleasepool] = ACTIONS(2726),
    [anon_sym_auto] = ACTIONS(2724),
    [anon_sym_register] = ACTIONS(2724),
    [anon_sym_inline] = ACTIONS(2724),
    [anon_sym___inline] = ACTIONS(2724),
    [anon_sym___inline__] = ACTIONS(2724),
    [anon_sym___forceinline] = ACTIONS(2724),
    [anon_sym_thread_local] = ACTIONS(2724),
    [anon_sym___thread] = ACTIONS(2724),
    [anon_sym_CG_EXTERN] = ACTIONS(2724),
    [anon_sym_CG_INLINE] = ACTIONS(2724),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2724),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2724),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2724),
    [anon_sym_IBOutlet] = ACTIONS(2724),
    [anon_sym_IBInspectable] = ACTIONS(2724),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2724),
    [anon_sym_NS_INLINE] = ACTIONS(2724),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2724),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2724),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2724),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2724),
    [anon_sym_const] = ACTIONS(2724),
    [anon_sym_constexpr] = ACTIONS(2724),
    [anon_sym_volatile] = ACTIONS(2724),
    [anon_sym_restrict] = ACTIONS(2724),
    [anon_sym___restrict__] = ACTIONS(2724),
    [anon_sym__Atomic] = ACTIONS(2724),
    [anon_sym__Noreturn] = ACTIONS(2724),
    [anon_sym__Nonnull] = ACTIONS(2724),
    [anon_sym_nullable] = ACTIONS(2724),
    [anon_sym__Complex] = ACTIONS(2724),
    [anon_sym__Nullable] = ACTIONS(2724),
    [anon_sym__Nullable_result] = ACTIONS(2724),
    [anon_sym__Null_unspecified] = ACTIONS(2724),
    [anon_sym___autoreleasing] = ACTIONS(2724),
    [anon_sym___block] = ACTIONS(2724),
    [anon_sym___bridge] = ACTIONS(2724),
    [anon_sym___bridge_retained] = ACTIONS(2724),
    [anon_sym___bridge_transfer] = ACTIONS(2724),
    [anon_sym___complex] = ACTIONS(2724),
    [anon_sym___const] = ACTIONS(2724),
    [anon_sym___imag] = ACTIONS(2724),
    [anon_sym___kindof] = ACTIONS(2724),
    [anon_sym___nonnull] = ACTIONS(2724),
    [anon_sym___nullable] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2724),
    [anon_sym___real] = ACTIONS(2724),
    [anon_sym___strong] = ACTIONS(2724),
    [anon_sym___unsafe_unretained] = ACTIONS(2724),
    [anon_sym___unused] = ACTIONS(2724),
    [anon_sym___weak] = ACTIONS(2724),
    [anon_sym_alignas] = ACTIONS(2724),
    [anon_sym__Alignas] = ACTIONS(2724),
    [sym_primitive_type] = ACTIONS(2724),
    [anon_sym_enum] = ACTIONS(2724),
    [anon_sym_struct] = ACTIONS(2724),
    [anon_sym_union] = ACTIONS(2724),
    [anon_sym_if] = ACTIONS(2724),
    [anon_sym_switch] = ACTIONS(2724),
    [anon_sym_case] = ACTIONS(2724),
    [anon_sym_default] = ACTIONS(2724),
    [anon_sym_while] = ACTIONS(2724),
    [anon_sym_do] = ACTIONS(2724),
    [anon_sym_for] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_return] = ACTIONS(2724),
    [anon_sym_break] = ACTIONS(2724),
    [anon_sym_continue] = ACTIONS(2724),
    [anon_sym_goto] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2726),
    [anon_sym_PLUS_PLUS] = ACTIONS(2726),
    [anon_sym_sizeof] = ACTIONS(2724),
    [anon_sym___alignof__] = ACTIONS(2724),
    [anon_sym___alignof] = ACTIONS(2724),
    [anon_sym__alignof] = ACTIONS(2724),
    [anon_sym_alignof] = ACTIONS(2724),
    [anon_sym__Alignof] = ACTIONS(2724),
    [anon_sym_offsetof] = ACTIONS(2724),
    [anon_sym__Generic] = ACTIONS(2724),
    [anon_sym_asm] = ACTIONS(2724),
    [anon_sym___asm__] = ACTIONS(2724),
    [anon_sym___asm] = ACTIONS(2724),
    [sym_number_literal] = ACTIONS(2726),
    [anon_sym_L_SQUOTE] = ACTIONS(2726),
    [anon_sym_u_SQUOTE] = ACTIONS(2726),
    [anon_sym_U_SQUOTE] = ACTIONS(2726),
    [anon_sym_u8_SQUOTE] = ACTIONS(2726),
    [anon_sym_SQUOTE] = ACTIONS(2726),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2726),
    [anon_sym_L_DQUOTE] = ACTIONS(2726),
    [anon_sym_u_DQUOTE] = ACTIONS(2726),
    [anon_sym_U_DQUOTE] = ACTIONS(2726),
    [anon_sym_u8_DQUOTE] = ACTIONS(2726),
    [sym_true] = ACTIONS(2724),
    [sym_false] = ACTIONS(2724),
    [anon_sym_NULL] = ACTIONS(2724),
    [anon_sym_nullptr] = ACTIONS(2724),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2724),
    [anon_sym___typeof] = ACTIONS(2724),
    [anon_sym_typeof] = ACTIONS(2724),
    [anon_sym_ATimport] = ACTIONS(2726),
    [aux_sym_preproc_undef_token1] = ACTIONS(2724),
    [anon_sym_POUND] = ACTIONS(2724),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2724),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2724),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2724),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2724),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2724),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2724),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2724),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_API_AVAILABLE] = ACTIONS(2724),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_API_DEPRECATED] = ACTIONS(2724),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2724),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym___deprecated_msg] = ACTIONS(2724),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2724),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2724),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2724),
    [anon_sym_ATprotocol] = ACTIONS(2726),
    [anon_sym_ATinterface] = ACTIONS(2726),
    [anon_sym_ATimplementation] = ACTIONS(2726),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2726),
    [anon_sym_ATsynthesize] = ACTIONS(2726),
    [anon_sym_ATdynamic] = ACTIONS(2726),
    [anon_sym_ATtry] = ACTIONS(2726),
    [anon_sym___try] = ACTIONS(2724),
    [anon_sym_ATthrow] = ACTIONS(2726),
    [anon_sym_ATselector] = ACTIONS(2726),
    [anon_sym_ATavailable] = ACTIONS(2726),
    [anon_sym___builtin_available] = ACTIONS(2724),
    [anon_sym_va_arg] = ACTIONS(2724),
    [anon_sym_ATencode] = ACTIONS(2726),
    [anon_sym_ATsynchronized] = ACTIONS(2726),
    [anon_sym_BOOL] = ACTIONS(2724),
    [anon_sym_IMP] = ACTIONS(2724),
    [anon_sym_SEL] = ACTIONS(2724),
    [anon_sym_Class] = ACTIONS(2724),
    [anon_sym_id] = ACTIONS(2724),
  },
  [STATE(445)] = {
    [sym_identifier] = ACTIONS(2596),
    [aux_sym_preproc_include_token1] = ACTIONS(2596),
    [aux_sym_preproc_include_token2] = ACTIONS(2596),
    [aux_sym_preproc_def_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token2] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2596),
    [aux_sym_preproc_else_token1] = ACTIONS(2596),
    [aux_sym_preproc_elif_token1] = ACTIONS(2596),
    [sym_preproc_directive] = ACTIONS(2596),
    [anon_sym_LPAREN2] = ACTIONS(2598),
    [anon_sym_BANG] = ACTIONS(2598),
    [anon_sym_TILDE] = ACTIONS(2598),
    [anon_sym_DASH] = ACTIONS(2596),
    [anon_sym_PLUS] = ACTIONS(2596),
    [anon_sym_STAR] = ACTIONS(2598),
    [anon_sym_CARET] = ACTIONS(2598),
    [anon_sym_AMP] = ACTIONS(2598),
    [anon_sym_SEMI] = ACTIONS(2598),
    [anon_sym___extension__] = ACTIONS(2596),
    [anon_sym_typedef] = ACTIONS(2596),
    [anon_sym_extern] = ACTIONS(2596),
    [anon_sym___attribute__] = ACTIONS(2596),
    [anon_sym___attribute] = ACTIONS(2596),
    [anon_sym_noreturn] = ACTIONS(2596),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym___declspec] = ACTIONS(2596),
    [anon_sym___cdecl] = ACTIONS(2596),
    [anon_sym___clrcall] = ACTIONS(2596),
    [anon_sym___stdcall] = ACTIONS(2596),
    [anon_sym___fastcall] = ACTIONS(2596),
    [anon_sym___thiscall] = ACTIONS(2596),
    [anon_sym___vectorcall] = ACTIONS(2596),
    [anon_sym_LBRACE] = ACTIONS(2598),
    [anon_sym_signed] = ACTIONS(2596),
    [anon_sym_unsigned] = ACTIONS(2596),
    [anon_sym_long] = ACTIONS(2596),
    [anon_sym_short] = ACTIONS(2596),
    [anon_sym_static] = ACTIONS(2596),
    [anon_sym_ATautoreleasepool] = ACTIONS(2598),
    [anon_sym_auto] = ACTIONS(2596),
    [anon_sym_register] = ACTIONS(2596),
    [anon_sym_inline] = ACTIONS(2596),
    [anon_sym___inline] = ACTIONS(2596),
    [anon_sym___inline__] = ACTIONS(2596),
    [anon_sym___forceinline] = ACTIONS(2596),
    [anon_sym_thread_local] = ACTIONS(2596),
    [anon_sym___thread] = ACTIONS(2596),
    [anon_sym_CG_EXTERN] = ACTIONS(2596),
    [anon_sym_CG_INLINE] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2596),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2596),
    [anon_sym_IBOutlet] = ACTIONS(2596),
    [anon_sym_IBInspectable] = ACTIONS(2596),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2596),
    [anon_sym_NS_INLINE] = ACTIONS(2596),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2596),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2596),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2596),
    [anon_sym_const] = ACTIONS(2596),
    [anon_sym_constexpr] = ACTIONS(2596),
    [anon_sym_volatile] = ACTIONS(2596),
    [anon_sym_restrict] = ACTIONS(2596),
    [anon_sym___restrict__] = ACTIONS(2596),
    [anon_sym__Atomic] = ACTIONS(2596),
    [anon_sym__Noreturn] = ACTIONS(2596),
    [anon_sym__Nonnull] = ACTIONS(2596),
    [anon_sym_nullable] = ACTIONS(2596),
    [anon_sym__Complex] = ACTIONS(2596),
    [anon_sym__Nullable] = ACTIONS(2596),
    [anon_sym__Nullable_result] = ACTIONS(2596),
    [anon_sym__Null_unspecified] = ACTIONS(2596),
    [anon_sym___autoreleasing] = ACTIONS(2596),
    [anon_sym___block] = ACTIONS(2596),
    [anon_sym___bridge] = ACTIONS(2596),
    [anon_sym___bridge_retained] = ACTIONS(2596),
    [anon_sym___bridge_transfer] = ACTIONS(2596),
    [anon_sym___complex] = ACTIONS(2596),
    [anon_sym___const] = ACTIONS(2596),
    [anon_sym___imag] = ACTIONS(2596),
    [anon_sym___kindof] = ACTIONS(2596),
    [anon_sym___nonnull] = ACTIONS(2596),
    [anon_sym___nullable] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2596),
    [anon_sym___real] = ACTIONS(2596),
    [anon_sym___strong] = ACTIONS(2596),
    [anon_sym___unsafe_unretained] = ACTIONS(2596),
    [anon_sym___unused] = ACTIONS(2596),
    [anon_sym___weak] = ACTIONS(2596),
    [anon_sym_alignas] = ACTIONS(2596),
    [anon_sym__Alignas] = ACTIONS(2596),
    [sym_primitive_type] = ACTIONS(2596),
    [anon_sym_enum] = ACTIONS(2596),
    [anon_sym_struct] = ACTIONS(2596),
    [anon_sym_union] = ACTIONS(2596),
    [anon_sym_if] = ACTIONS(2596),
    [anon_sym_switch] = ACTIONS(2596),
    [anon_sym_case] = ACTIONS(2596),
    [anon_sym_default] = ACTIONS(2596),
    [anon_sym_while] = ACTIONS(2596),
    [anon_sym_do] = ACTIONS(2596),
    [anon_sym_for] = ACTIONS(2596),
    [anon_sym_in] = ACTIONS(2596),
    [anon_sym_return] = ACTIONS(2596),
    [anon_sym_break] = ACTIONS(2596),
    [anon_sym_continue] = ACTIONS(2596),
    [anon_sym_goto] = ACTIONS(2596),
    [anon_sym_DASH_DASH] = ACTIONS(2598),
    [anon_sym_PLUS_PLUS] = ACTIONS(2598),
    [anon_sym_sizeof] = ACTIONS(2596),
    [anon_sym___alignof__] = ACTIONS(2596),
    [anon_sym___alignof] = ACTIONS(2596),
    [anon_sym__alignof] = ACTIONS(2596),
    [anon_sym_alignof] = ACTIONS(2596),
    [anon_sym__Alignof] = ACTIONS(2596),
    [anon_sym_offsetof] = ACTIONS(2596),
    [anon_sym__Generic] = ACTIONS(2596),
    [anon_sym_asm] = ACTIONS(2596),
    [anon_sym___asm__] = ACTIONS(2596),
    [anon_sym___asm] = ACTIONS(2596),
    [sym_number_literal] = ACTIONS(2598),
    [anon_sym_L_SQUOTE] = ACTIONS(2598),
    [anon_sym_u_SQUOTE] = ACTIONS(2598),
    [anon_sym_U_SQUOTE] = ACTIONS(2598),
    [anon_sym_u8_SQUOTE] = ACTIONS(2598),
    [anon_sym_SQUOTE] = ACTIONS(2598),
    [anon_sym_AT] = ACTIONS(2596),
    [anon_sym_DQUOTE] = ACTIONS(2598),
    [anon_sym_L_DQUOTE] = ACTIONS(2598),
    [anon_sym_u_DQUOTE] = ACTIONS(2598),
    [anon_sym_U_DQUOTE] = ACTIONS(2598),
    [anon_sym_u8_DQUOTE] = ACTIONS(2598),
    [sym_true] = ACTIONS(2596),
    [sym_false] = ACTIONS(2596),
    [anon_sym_NULL] = ACTIONS(2596),
    [anon_sym_nullptr] = ACTIONS(2596),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2596),
    [anon_sym___typeof] = ACTIONS(2596),
    [anon_sym_typeof] = ACTIONS(2596),
    [anon_sym_ATimport] = ACTIONS(2598),
    [aux_sym_preproc_undef_token1] = ACTIONS(2596),
    [anon_sym_POUND] = ACTIONS(2596),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2596),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2596),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2596),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_API_AVAILABLE] = ACTIONS(2596),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_API_DEPRECATED] = ACTIONS(2596),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym___deprecated_msg] = ACTIONS(2596),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2596),
    [anon_sym_ATprotocol] = ACTIONS(2598),
    [anon_sym_ATinterface] = ACTIONS(2598),
    [anon_sym_ATimplementation] = ACTIONS(2598),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2598),
    [anon_sym_ATsynthesize] = ACTIONS(2598),
    [anon_sym_ATdynamic] = ACTIONS(2598),
    [anon_sym_ATtry] = ACTIONS(2598),
    [anon_sym___try] = ACTIONS(2596),
    [anon_sym_ATthrow] = ACTIONS(2598),
    [anon_sym_ATselector] = ACTIONS(2598),
    [anon_sym_ATavailable] = ACTIONS(2598),
    [anon_sym___builtin_available] = ACTIONS(2596),
    [anon_sym_va_arg] = ACTIONS(2596),
    [anon_sym_ATencode] = ACTIONS(2598),
    [anon_sym_ATsynchronized] = ACTIONS(2598),
    [anon_sym_BOOL] = ACTIONS(2596),
    [anon_sym_IMP] = ACTIONS(2596),
    [anon_sym_SEL] = ACTIONS(2596),
    [anon_sym_Class] = ACTIONS(2596),
    [anon_sym_id] = ACTIONS(2596),
  },
  [STATE(446)] = {
    [sym_identifier] = ACTIONS(2600),
    [aux_sym_preproc_include_token1] = ACTIONS(2600),
    [aux_sym_preproc_include_token2] = ACTIONS(2600),
    [aux_sym_preproc_def_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token2] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2600),
    [aux_sym_preproc_else_token1] = ACTIONS(2600),
    [aux_sym_preproc_elif_token1] = ACTIONS(2600),
    [sym_preproc_directive] = ACTIONS(2600),
    [anon_sym_LPAREN2] = ACTIONS(2602),
    [anon_sym_BANG] = ACTIONS(2602),
    [anon_sym_TILDE] = ACTIONS(2602),
    [anon_sym_DASH] = ACTIONS(2600),
    [anon_sym_PLUS] = ACTIONS(2600),
    [anon_sym_STAR] = ACTIONS(2602),
    [anon_sym_CARET] = ACTIONS(2602),
    [anon_sym_AMP] = ACTIONS(2602),
    [anon_sym_SEMI] = ACTIONS(2602),
    [anon_sym___extension__] = ACTIONS(2600),
    [anon_sym_typedef] = ACTIONS(2600),
    [anon_sym_extern] = ACTIONS(2600),
    [anon_sym___attribute__] = ACTIONS(2600),
    [anon_sym___attribute] = ACTIONS(2600),
    [anon_sym_noreturn] = ACTIONS(2600),
    [anon_sym_LBRACK] = ACTIONS(2602),
    [anon_sym___declspec] = ACTIONS(2600),
    [anon_sym___cdecl] = ACTIONS(2600),
    [anon_sym___clrcall] = ACTIONS(2600),
    [anon_sym___stdcall] = ACTIONS(2600),
    [anon_sym___fastcall] = ACTIONS(2600),
    [anon_sym___thiscall] = ACTIONS(2600),
    [anon_sym___vectorcall] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_signed] = ACTIONS(2600),
    [anon_sym_unsigned] = ACTIONS(2600),
    [anon_sym_long] = ACTIONS(2600),
    [anon_sym_short] = ACTIONS(2600),
    [anon_sym_static] = ACTIONS(2600),
    [anon_sym_ATautoreleasepool] = ACTIONS(2602),
    [anon_sym_auto] = ACTIONS(2600),
    [anon_sym_register] = ACTIONS(2600),
    [anon_sym_inline] = ACTIONS(2600),
    [anon_sym___inline] = ACTIONS(2600),
    [anon_sym___inline__] = ACTIONS(2600),
    [anon_sym___forceinline] = ACTIONS(2600),
    [anon_sym_thread_local] = ACTIONS(2600),
    [anon_sym___thread] = ACTIONS(2600),
    [anon_sym_CG_EXTERN] = ACTIONS(2600),
    [anon_sym_CG_INLINE] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2600),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2600),
    [anon_sym_IBOutlet] = ACTIONS(2600),
    [anon_sym_IBInspectable] = ACTIONS(2600),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2600),
    [anon_sym_NS_INLINE] = ACTIONS(2600),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2600),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2600),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2600),
    [anon_sym_const] = ACTIONS(2600),
    [anon_sym_constexpr] = ACTIONS(2600),
    [anon_sym_volatile] = ACTIONS(2600),
    [anon_sym_restrict] = ACTIONS(2600),
    [anon_sym___restrict__] = ACTIONS(2600),
    [anon_sym__Atomic] = ACTIONS(2600),
    [anon_sym__Noreturn] = ACTIONS(2600),
    [anon_sym__Nonnull] = ACTIONS(2600),
    [anon_sym_nullable] = ACTIONS(2600),
    [anon_sym__Complex] = ACTIONS(2600),
    [anon_sym__Nullable] = ACTIONS(2600),
    [anon_sym__Nullable_result] = ACTIONS(2600),
    [anon_sym__Null_unspecified] = ACTIONS(2600),
    [anon_sym___autoreleasing] = ACTIONS(2600),
    [anon_sym___block] = ACTIONS(2600),
    [anon_sym___bridge] = ACTIONS(2600),
    [anon_sym___bridge_retained] = ACTIONS(2600),
    [anon_sym___bridge_transfer] = ACTIONS(2600),
    [anon_sym___complex] = ACTIONS(2600),
    [anon_sym___const] = ACTIONS(2600),
    [anon_sym___imag] = ACTIONS(2600),
    [anon_sym___kindof] = ACTIONS(2600),
    [anon_sym___nonnull] = ACTIONS(2600),
    [anon_sym___nullable] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2600),
    [anon_sym___real] = ACTIONS(2600),
    [anon_sym___strong] = ACTIONS(2600),
    [anon_sym___unsafe_unretained] = ACTIONS(2600),
    [anon_sym___unused] = ACTIONS(2600),
    [anon_sym___weak] = ACTIONS(2600),
    [anon_sym_alignas] = ACTIONS(2600),
    [anon_sym__Alignas] = ACTIONS(2600),
    [sym_primitive_type] = ACTIONS(2600),
    [anon_sym_enum] = ACTIONS(2600),
    [anon_sym_struct] = ACTIONS(2600),
    [anon_sym_union] = ACTIONS(2600),
    [anon_sym_if] = ACTIONS(2600),
    [anon_sym_switch] = ACTIONS(2600),
    [anon_sym_case] = ACTIONS(2600),
    [anon_sym_default] = ACTIONS(2600),
    [anon_sym_while] = ACTIONS(2600),
    [anon_sym_do] = ACTIONS(2600),
    [anon_sym_for] = ACTIONS(2600),
    [anon_sym_in] = ACTIONS(2600),
    [anon_sym_return] = ACTIONS(2600),
    [anon_sym_break] = ACTIONS(2600),
    [anon_sym_continue] = ACTIONS(2600),
    [anon_sym_goto] = ACTIONS(2600),
    [anon_sym_DASH_DASH] = ACTIONS(2602),
    [anon_sym_PLUS_PLUS] = ACTIONS(2602),
    [anon_sym_sizeof] = ACTIONS(2600),
    [anon_sym___alignof__] = ACTIONS(2600),
    [anon_sym___alignof] = ACTIONS(2600),
    [anon_sym__alignof] = ACTIONS(2600),
    [anon_sym_alignof] = ACTIONS(2600),
    [anon_sym__Alignof] = ACTIONS(2600),
    [anon_sym_offsetof] = ACTIONS(2600),
    [anon_sym__Generic] = ACTIONS(2600),
    [anon_sym_asm] = ACTIONS(2600),
    [anon_sym___asm__] = ACTIONS(2600),
    [anon_sym___asm] = ACTIONS(2600),
    [sym_number_literal] = ACTIONS(2602),
    [anon_sym_L_SQUOTE] = ACTIONS(2602),
    [anon_sym_u_SQUOTE] = ACTIONS(2602),
    [anon_sym_U_SQUOTE] = ACTIONS(2602),
    [anon_sym_u8_SQUOTE] = ACTIONS(2602),
    [anon_sym_SQUOTE] = ACTIONS(2602),
    [anon_sym_AT] = ACTIONS(2600),
    [anon_sym_DQUOTE] = ACTIONS(2602),
    [anon_sym_L_DQUOTE] = ACTIONS(2602),
    [anon_sym_u_DQUOTE] = ACTIONS(2602),
    [anon_sym_U_DQUOTE] = ACTIONS(2602),
    [anon_sym_u8_DQUOTE] = ACTIONS(2602),
    [sym_true] = ACTIONS(2600),
    [sym_false] = ACTIONS(2600),
    [anon_sym_NULL] = ACTIONS(2600),
    [anon_sym_nullptr] = ACTIONS(2600),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2600),
    [anon_sym___typeof] = ACTIONS(2600),
    [anon_sym_typeof] = ACTIONS(2600),
    [anon_sym_ATimport] = ACTIONS(2602),
    [aux_sym_preproc_undef_token1] = ACTIONS(2600),
    [anon_sym_POUND] = ACTIONS(2600),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2600),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2600),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2600),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_API_AVAILABLE] = ACTIONS(2600),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_API_DEPRECATED] = ACTIONS(2600),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym___deprecated_msg] = ACTIONS(2600),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2600),
    [anon_sym_ATprotocol] = ACTIONS(2602),
    [anon_sym_ATinterface] = ACTIONS(2602),
    [anon_sym_ATimplementation] = ACTIONS(2602),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2602),
    [anon_sym_ATsynthesize] = ACTIONS(2602),
    [anon_sym_ATdynamic] = ACTIONS(2602),
    [anon_sym_ATtry] = ACTIONS(2602),
    [anon_sym___try] = ACTIONS(2600),
    [anon_sym_ATthrow] = ACTIONS(2602),
    [anon_sym_ATselector] = ACTIONS(2602),
    [anon_sym_ATavailable] = ACTIONS(2602),
    [anon_sym___builtin_available] = ACTIONS(2600),
    [anon_sym_va_arg] = ACTIONS(2600),
    [anon_sym_ATencode] = ACTIONS(2602),
    [anon_sym_ATsynchronized] = ACTIONS(2602),
    [anon_sym_BOOL] = ACTIONS(2600),
    [anon_sym_IMP] = ACTIONS(2600),
    [anon_sym_SEL] = ACTIONS(2600),
    [anon_sym_Class] = ACTIONS(2600),
    [anon_sym_id] = ACTIONS(2600),
  },
  [STATE(447)] = {
    [sym_identifier] = ACTIONS(2768),
    [aux_sym_preproc_include_token1] = ACTIONS(2768),
    [aux_sym_preproc_include_token2] = ACTIONS(2768),
    [aux_sym_preproc_def_token1] = ACTIONS(2768),
    [aux_sym_preproc_if_token1] = ACTIONS(2768),
    [aux_sym_preproc_if_token2] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2768),
    [aux_sym_preproc_else_token1] = ACTIONS(2768),
    [aux_sym_preproc_elif_token1] = ACTIONS(2768),
    [sym_preproc_directive] = ACTIONS(2768),
    [anon_sym_LPAREN2] = ACTIONS(2770),
    [anon_sym_BANG] = ACTIONS(2770),
    [anon_sym_TILDE] = ACTIONS(2770),
    [anon_sym_DASH] = ACTIONS(2768),
    [anon_sym_PLUS] = ACTIONS(2768),
    [anon_sym_STAR] = ACTIONS(2770),
    [anon_sym_CARET] = ACTIONS(2770),
    [anon_sym_AMP] = ACTIONS(2770),
    [anon_sym_SEMI] = ACTIONS(2770),
    [anon_sym___extension__] = ACTIONS(2768),
    [anon_sym_typedef] = ACTIONS(2768),
    [anon_sym_extern] = ACTIONS(2768),
    [anon_sym___attribute__] = ACTIONS(2768),
    [anon_sym___attribute] = ACTIONS(2768),
    [anon_sym_noreturn] = ACTIONS(2768),
    [anon_sym_LBRACK] = ACTIONS(2770),
    [anon_sym___declspec] = ACTIONS(2768),
    [anon_sym___cdecl] = ACTIONS(2768),
    [anon_sym___clrcall] = ACTIONS(2768),
    [anon_sym___stdcall] = ACTIONS(2768),
    [anon_sym___fastcall] = ACTIONS(2768),
    [anon_sym___thiscall] = ACTIONS(2768),
    [anon_sym___vectorcall] = ACTIONS(2768),
    [anon_sym_LBRACE] = ACTIONS(2770),
    [anon_sym_signed] = ACTIONS(2768),
    [anon_sym_unsigned] = ACTIONS(2768),
    [anon_sym_long] = ACTIONS(2768),
    [anon_sym_short] = ACTIONS(2768),
    [anon_sym_static] = ACTIONS(2768),
    [anon_sym_ATautoreleasepool] = ACTIONS(2770),
    [anon_sym_auto] = ACTIONS(2768),
    [anon_sym_register] = ACTIONS(2768),
    [anon_sym_inline] = ACTIONS(2768),
    [anon_sym___inline] = ACTIONS(2768),
    [anon_sym___inline__] = ACTIONS(2768),
    [anon_sym___forceinline] = ACTIONS(2768),
    [anon_sym_thread_local] = ACTIONS(2768),
    [anon_sym___thread] = ACTIONS(2768),
    [anon_sym_CG_EXTERN] = ACTIONS(2768),
    [anon_sym_CG_INLINE] = ACTIONS(2768),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2768),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2768),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2768),
    [anon_sym_IBOutlet] = ACTIONS(2768),
    [anon_sym_IBInspectable] = ACTIONS(2768),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2768),
    [anon_sym_NS_INLINE] = ACTIONS(2768),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2768),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2768),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2768),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2768),
    [anon_sym_const] = ACTIONS(2768),
    [anon_sym_constexpr] = ACTIONS(2768),
    [anon_sym_volatile] = ACTIONS(2768),
    [anon_sym_restrict] = ACTIONS(2768),
    [anon_sym___restrict__] = ACTIONS(2768),
    [anon_sym__Atomic] = ACTIONS(2768),
    [anon_sym__Noreturn] = ACTIONS(2768),
    [anon_sym__Nonnull] = ACTIONS(2768),
    [anon_sym_nullable] = ACTIONS(2768),
    [anon_sym__Complex] = ACTIONS(2768),
    [anon_sym__Nullable] = ACTIONS(2768),
    [anon_sym__Nullable_result] = ACTIONS(2768),
    [anon_sym__Null_unspecified] = ACTIONS(2768),
    [anon_sym___autoreleasing] = ACTIONS(2768),
    [anon_sym___block] = ACTIONS(2768),
    [anon_sym___bridge] = ACTIONS(2768),
    [anon_sym___bridge_retained] = ACTIONS(2768),
    [anon_sym___bridge_transfer] = ACTIONS(2768),
    [anon_sym___complex] = ACTIONS(2768),
    [anon_sym___const] = ACTIONS(2768),
    [anon_sym___imag] = ACTIONS(2768),
    [anon_sym___kindof] = ACTIONS(2768),
    [anon_sym___nonnull] = ACTIONS(2768),
    [anon_sym___nullable] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2768),
    [anon_sym___real] = ACTIONS(2768),
    [anon_sym___strong] = ACTIONS(2768),
    [anon_sym___unsafe_unretained] = ACTIONS(2768),
    [anon_sym___unused] = ACTIONS(2768),
    [anon_sym___weak] = ACTIONS(2768),
    [anon_sym_alignas] = ACTIONS(2768),
    [anon_sym__Alignas] = ACTIONS(2768),
    [sym_primitive_type] = ACTIONS(2768),
    [anon_sym_enum] = ACTIONS(2768),
    [anon_sym_struct] = ACTIONS(2768),
    [anon_sym_union] = ACTIONS(2768),
    [anon_sym_if] = ACTIONS(2768),
    [anon_sym_switch] = ACTIONS(2768),
    [anon_sym_case] = ACTIONS(2768),
    [anon_sym_default] = ACTIONS(2768),
    [anon_sym_while] = ACTIONS(2768),
    [anon_sym_do] = ACTIONS(2768),
    [anon_sym_for] = ACTIONS(2768),
    [anon_sym_in] = ACTIONS(2768),
    [anon_sym_return] = ACTIONS(2768),
    [anon_sym_break] = ACTIONS(2768),
    [anon_sym_continue] = ACTIONS(2768),
    [anon_sym_goto] = ACTIONS(2768),
    [anon_sym_DASH_DASH] = ACTIONS(2770),
    [anon_sym_PLUS_PLUS] = ACTIONS(2770),
    [anon_sym_sizeof] = ACTIONS(2768),
    [anon_sym___alignof__] = ACTIONS(2768),
    [anon_sym___alignof] = ACTIONS(2768),
    [anon_sym__alignof] = ACTIONS(2768),
    [anon_sym_alignof] = ACTIONS(2768),
    [anon_sym__Alignof] = ACTIONS(2768),
    [anon_sym_offsetof] = ACTIONS(2768),
    [anon_sym__Generic] = ACTIONS(2768),
    [anon_sym_asm] = ACTIONS(2768),
    [anon_sym___asm__] = ACTIONS(2768),
    [anon_sym___asm] = ACTIONS(2768),
    [sym_number_literal] = ACTIONS(2770),
    [anon_sym_L_SQUOTE] = ACTIONS(2770),
    [anon_sym_u_SQUOTE] = ACTIONS(2770),
    [anon_sym_U_SQUOTE] = ACTIONS(2770),
    [anon_sym_u8_SQUOTE] = ACTIONS(2770),
    [anon_sym_SQUOTE] = ACTIONS(2770),
    [anon_sym_AT] = ACTIONS(2768),
    [anon_sym_DQUOTE] = ACTIONS(2770),
    [anon_sym_L_DQUOTE] = ACTIONS(2770),
    [anon_sym_u_DQUOTE] = ACTIONS(2770),
    [anon_sym_U_DQUOTE] = ACTIONS(2770),
    [anon_sym_u8_DQUOTE] = ACTIONS(2770),
    [sym_true] = ACTIONS(2768),
    [sym_false] = ACTIONS(2768),
    [anon_sym_NULL] = ACTIONS(2768),
    [anon_sym_nullptr] = ACTIONS(2768),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2768),
    [anon_sym___typeof] = ACTIONS(2768),
    [anon_sym_typeof] = ACTIONS(2768),
    [anon_sym_ATimport] = ACTIONS(2770),
    [aux_sym_preproc_undef_token1] = ACTIONS(2768),
    [anon_sym_POUND] = ACTIONS(2768),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2768),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2768),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2768),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2768),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2768),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2768),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2768),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_API_AVAILABLE] = ACTIONS(2768),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_API_DEPRECATED] = ACTIONS(2768),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2768),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym___deprecated_msg] = ACTIONS(2768),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2768),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2768),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2768),
    [anon_sym_ATprotocol] = ACTIONS(2770),
    [anon_sym_ATinterface] = ACTIONS(2770),
    [anon_sym_ATimplementation] = ACTIONS(2770),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2770),
    [anon_sym_ATsynthesize] = ACTIONS(2770),
    [anon_sym_ATdynamic] = ACTIONS(2770),
    [anon_sym_ATtry] = ACTIONS(2770),
    [anon_sym___try] = ACTIONS(2768),
    [anon_sym_ATthrow] = ACTIONS(2770),
    [anon_sym_ATselector] = ACTIONS(2770),
    [anon_sym_ATavailable] = ACTIONS(2770),
    [anon_sym___builtin_available] = ACTIONS(2768),
    [anon_sym_va_arg] = ACTIONS(2768),
    [anon_sym_ATencode] = ACTIONS(2770),
    [anon_sym_ATsynchronized] = ACTIONS(2770),
    [anon_sym_BOOL] = ACTIONS(2768),
    [anon_sym_IMP] = ACTIONS(2768),
    [anon_sym_SEL] = ACTIONS(2768),
    [anon_sym_Class] = ACTIONS(2768),
    [anon_sym_id] = ACTIONS(2768),
  },
  [STATE(448)] = {
    [sym_identifier] = ACTIONS(2772),
    [aux_sym_preproc_include_token1] = ACTIONS(2772),
    [aux_sym_preproc_include_token2] = ACTIONS(2772),
    [aux_sym_preproc_def_token1] = ACTIONS(2772),
    [aux_sym_preproc_if_token1] = ACTIONS(2772),
    [aux_sym_preproc_if_token2] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2772),
    [aux_sym_preproc_else_token1] = ACTIONS(2772),
    [aux_sym_preproc_elif_token1] = ACTIONS(2772),
    [sym_preproc_directive] = ACTIONS(2772),
    [anon_sym_LPAREN2] = ACTIONS(2774),
    [anon_sym_BANG] = ACTIONS(2774),
    [anon_sym_TILDE] = ACTIONS(2774),
    [anon_sym_DASH] = ACTIONS(2772),
    [anon_sym_PLUS] = ACTIONS(2772),
    [anon_sym_STAR] = ACTIONS(2774),
    [anon_sym_CARET] = ACTIONS(2774),
    [anon_sym_AMP] = ACTIONS(2774),
    [anon_sym_SEMI] = ACTIONS(2774),
    [anon_sym___extension__] = ACTIONS(2772),
    [anon_sym_typedef] = ACTIONS(2772),
    [anon_sym_extern] = ACTIONS(2772),
    [anon_sym___attribute__] = ACTIONS(2772),
    [anon_sym___attribute] = ACTIONS(2772),
    [anon_sym_noreturn] = ACTIONS(2772),
    [anon_sym_LBRACK] = ACTIONS(2774),
    [anon_sym___declspec] = ACTIONS(2772),
    [anon_sym___cdecl] = ACTIONS(2772),
    [anon_sym___clrcall] = ACTIONS(2772),
    [anon_sym___stdcall] = ACTIONS(2772),
    [anon_sym___fastcall] = ACTIONS(2772),
    [anon_sym___thiscall] = ACTIONS(2772),
    [anon_sym___vectorcall] = ACTIONS(2772),
    [anon_sym_LBRACE] = ACTIONS(2774),
    [anon_sym_signed] = ACTIONS(2772),
    [anon_sym_unsigned] = ACTIONS(2772),
    [anon_sym_long] = ACTIONS(2772),
    [anon_sym_short] = ACTIONS(2772),
    [anon_sym_static] = ACTIONS(2772),
    [anon_sym_ATautoreleasepool] = ACTIONS(2774),
    [anon_sym_auto] = ACTIONS(2772),
    [anon_sym_register] = ACTIONS(2772),
    [anon_sym_inline] = ACTIONS(2772),
    [anon_sym___inline] = ACTIONS(2772),
    [anon_sym___inline__] = ACTIONS(2772),
    [anon_sym___forceinline] = ACTIONS(2772),
    [anon_sym_thread_local] = ACTIONS(2772),
    [anon_sym___thread] = ACTIONS(2772),
    [anon_sym_CG_EXTERN] = ACTIONS(2772),
    [anon_sym_CG_INLINE] = ACTIONS(2772),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2772),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2772),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2772),
    [anon_sym_IBOutlet] = ACTIONS(2772),
    [anon_sym_IBInspectable] = ACTIONS(2772),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2772),
    [anon_sym_NS_INLINE] = ACTIONS(2772),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2772),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2772),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2772),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2772),
    [anon_sym_const] = ACTIONS(2772),
    [anon_sym_constexpr] = ACTIONS(2772),
    [anon_sym_volatile] = ACTIONS(2772),
    [anon_sym_restrict] = ACTIONS(2772),
    [anon_sym___restrict__] = ACTIONS(2772),
    [anon_sym__Atomic] = ACTIONS(2772),
    [anon_sym__Noreturn] = ACTIONS(2772),
    [anon_sym__Nonnull] = ACTIONS(2772),
    [anon_sym_nullable] = ACTIONS(2772),
    [anon_sym__Complex] = ACTIONS(2772),
    [anon_sym__Nullable] = ACTIONS(2772),
    [anon_sym__Nullable_result] = ACTIONS(2772),
    [anon_sym__Null_unspecified] = ACTIONS(2772),
    [anon_sym___autoreleasing] = ACTIONS(2772),
    [anon_sym___block] = ACTIONS(2772),
    [anon_sym___bridge] = ACTIONS(2772),
    [anon_sym___bridge_retained] = ACTIONS(2772),
    [anon_sym___bridge_transfer] = ACTIONS(2772),
    [anon_sym___complex] = ACTIONS(2772),
    [anon_sym___const] = ACTIONS(2772),
    [anon_sym___imag] = ACTIONS(2772),
    [anon_sym___kindof] = ACTIONS(2772),
    [anon_sym___nonnull] = ACTIONS(2772),
    [anon_sym___nullable] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2772),
    [anon_sym___real] = ACTIONS(2772),
    [anon_sym___strong] = ACTIONS(2772),
    [anon_sym___unsafe_unretained] = ACTIONS(2772),
    [anon_sym___unused] = ACTIONS(2772),
    [anon_sym___weak] = ACTIONS(2772),
    [anon_sym_alignas] = ACTIONS(2772),
    [anon_sym__Alignas] = ACTIONS(2772),
    [sym_primitive_type] = ACTIONS(2772),
    [anon_sym_enum] = ACTIONS(2772),
    [anon_sym_struct] = ACTIONS(2772),
    [anon_sym_union] = ACTIONS(2772),
    [anon_sym_if] = ACTIONS(2772),
    [anon_sym_switch] = ACTIONS(2772),
    [anon_sym_case] = ACTIONS(2772),
    [anon_sym_default] = ACTIONS(2772),
    [anon_sym_while] = ACTIONS(2772),
    [anon_sym_do] = ACTIONS(2772),
    [anon_sym_for] = ACTIONS(2772),
    [anon_sym_in] = ACTIONS(2772),
    [anon_sym_return] = ACTIONS(2772),
    [anon_sym_break] = ACTIONS(2772),
    [anon_sym_continue] = ACTIONS(2772),
    [anon_sym_goto] = ACTIONS(2772),
    [anon_sym_DASH_DASH] = ACTIONS(2774),
    [anon_sym_PLUS_PLUS] = ACTIONS(2774),
    [anon_sym_sizeof] = ACTIONS(2772),
    [anon_sym___alignof__] = ACTIONS(2772),
    [anon_sym___alignof] = ACTIONS(2772),
    [anon_sym__alignof] = ACTIONS(2772),
    [anon_sym_alignof] = ACTIONS(2772),
    [anon_sym__Alignof] = ACTIONS(2772),
    [anon_sym_offsetof] = ACTIONS(2772),
    [anon_sym__Generic] = ACTIONS(2772),
    [anon_sym_asm] = ACTIONS(2772),
    [anon_sym___asm__] = ACTIONS(2772),
    [anon_sym___asm] = ACTIONS(2772),
    [sym_number_literal] = ACTIONS(2774),
    [anon_sym_L_SQUOTE] = ACTIONS(2774),
    [anon_sym_u_SQUOTE] = ACTIONS(2774),
    [anon_sym_U_SQUOTE] = ACTIONS(2774),
    [anon_sym_u8_SQUOTE] = ACTIONS(2774),
    [anon_sym_SQUOTE] = ACTIONS(2774),
    [anon_sym_AT] = ACTIONS(2772),
    [anon_sym_DQUOTE] = ACTIONS(2774),
    [anon_sym_L_DQUOTE] = ACTIONS(2774),
    [anon_sym_u_DQUOTE] = ACTIONS(2774),
    [anon_sym_U_DQUOTE] = ACTIONS(2774),
    [anon_sym_u8_DQUOTE] = ACTIONS(2774),
    [sym_true] = ACTIONS(2772),
    [sym_false] = ACTIONS(2772),
    [anon_sym_NULL] = ACTIONS(2772),
    [anon_sym_nullptr] = ACTIONS(2772),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2772),
    [anon_sym___typeof] = ACTIONS(2772),
    [anon_sym_typeof] = ACTIONS(2772),
    [anon_sym_ATimport] = ACTIONS(2774),
    [aux_sym_preproc_undef_token1] = ACTIONS(2772),
    [anon_sym_POUND] = ACTIONS(2772),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2772),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2772),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2772),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2772),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2772),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2772),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2772),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_API_AVAILABLE] = ACTIONS(2772),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_API_DEPRECATED] = ACTIONS(2772),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2772),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym___deprecated_msg] = ACTIONS(2772),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2772),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2772),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2772),
    [anon_sym_ATprotocol] = ACTIONS(2774),
    [anon_sym_ATinterface] = ACTIONS(2774),
    [anon_sym_ATimplementation] = ACTIONS(2774),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2774),
    [anon_sym_ATsynthesize] = ACTIONS(2774),
    [anon_sym_ATdynamic] = ACTIONS(2774),
    [anon_sym_ATtry] = ACTIONS(2774),
    [anon_sym___try] = ACTIONS(2772),
    [anon_sym_ATthrow] = ACTIONS(2774),
    [anon_sym_ATselector] = ACTIONS(2774),
    [anon_sym_ATavailable] = ACTIONS(2774),
    [anon_sym___builtin_available] = ACTIONS(2772),
    [anon_sym_va_arg] = ACTIONS(2772),
    [anon_sym_ATencode] = ACTIONS(2774),
    [anon_sym_ATsynchronized] = ACTIONS(2774),
    [anon_sym_BOOL] = ACTIONS(2772),
    [anon_sym_IMP] = ACTIONS(2772),
    [anon_sym_SEL] = ACTIONS(2772),
    [anon_sym_Class] = ACTIONS(2772),
    [anon_sym_id] = ACTIONS(2772),
  },
  [STATE(449)] = {
    [sym_identifier] = ACTIONS(2728),
    [aux_sym_preproc_include_token1] = ACTIONS(2728),
    [aux_sym_preproc_include_token2] = ACTIONS(2728),
    [aux_sym_preproc_def_token1] = ACTIONS(2728),
    [aux_sym_preproc_if_token1] = ACTIONS(2728),
    [aux_sym_preproc_if_token2] = ACTIONS(2728),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2728),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2728),
    [aux_sym_preproc_else_token1] = ACTIONS(2728),
    [aux_sym_preproc_elif_token1] = ACTIONS(2728),
    [sym_preproc_directive] = ACTIONS(2728),
    [anon_sym_LPAREN2] = ACTIONS(2730),
    [anon_sym_BANG] = ACTIONS(2730),
    [anon_sym_TILDE] = ACTIONS(2730),
    [anon_sym_DASH] = ACTIONS(2728),
    [anon_sym_PLUS] = ACTIONS(2728),
    [anon_sym_STAR] = ACTIONS(2730),
    [anon_sym_CARET] = ACTIONS(2730),
    [anon_sym_AMP] = ACTIONS(2730),
    [anon_sym_SEMI] = ACTIONS(2730),
    [anon_sym___extension__] = ACTIONS(2728),
    [anon_sym_typedef] = ACTIONS(2728),
    [anon_sym_extern] = ACTIONS(2728),
    [anon_sym___attribute__] = ACTIONS(2728),
    [anon_sym___attribute] = ACTIONS(2728),
    [anon_sym_noreturn] = ACTIONS(2728),
    [anon_sym_LBRACK] = ACTIONS(2730),
    [anon_sym___declspec] = ACTIONS(2728),
    [anon_sym___cdecl] = ACTIONS(2728),
    [anon_sym___clrcall] = ACTIONS(2728),
    [anon_sym___stdcall] = ACTIONS(2728),
    [anon_sym___fastcall] = ACTIONS(2728),
    [anon_sym___thiscall] = ACTIONS(2728),
    [anon_sym___vectorcall] = ACTIONS(2728),
    [anon_sym_LBRACE] = ACTIONS(2730),
    [anon_sym_signed] = ACTIONS(2728),
    [anon_sym_unsigned] = ACTIONS(2728),
    [anon_sym_long] = ACTIONS(2728),
    [anon_sym_short] = ACTIONS(2728),
    [anon_sym_static] = ACTIONS(2728),
    [anon_sym_ATautoreleasepool] = ACTIONS(2730),
    [anon_sym_auto] = ACTIONS(2728),
    [anon_sym_register] = ACTIONS(2728),
    [anon_sym_inline] = ACTIONS(2728),
    [anon_sym___inline] = ACTIONS(2728),
    [anon_sym___inline__] = ACTIONS(2728),
    [anon_sym___forceinline] = ACTIONS(2728),
    [anon_sym_thread_local] = ACTIONS(2728),
    [anon_sym___thread] = ACTIONS(2728),
    [anon_sym_CG_EXTERN] = ACTIONS(2728),
    [anon_sym_CG_INLINE] = ACTIONS(2728),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2728),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2728),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2728),
    [anon_sym_IBOutlet] = ACTIONS(2728),
    [anon_sym_IBInspectable] = ACTIONS(2728),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2728),
    [anon_sym_NS_INLINE] = ACTIONS(2728),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2728),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2728),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2728),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2728),
    [anon_sym_const] = ACTIONS(2728),
    [anon_sym_constexpr] = ACTIONS(2728),
    [anon_sym_volatile] = ACTIONS(2728),
    [anon_sym_restrict] = ACTIONS(2728),
    [anon_sym___restrict__] = ACTIONS(2728),
    [anon_sym__Atomic] = ACTIONS(2728),
    [anon_sym__Noreturn] = ACTIONS(2728),
    [anon_sym__Nonnull] = ACTIONS(2728),
    [anon_sym_nullable] = ACTIONS(2728),
    [anon_sym__Complex] = ACTIONS(2728),
    [anon_sym__Nullable] = ACTIONS(2728),
    [anon_sym__Nullable_result] = ACTIONS(2728),
    [anon_sym__Null_unspecified] = ACTIONS(2728),
    [anon_sym___autoreleasing] = ACTIONS(2728),
    [anon_sym___block] = ACTIONS(2728),
    [anon_sym___bridge] = ACTIONS(2728),
    [anon_sym___bridge_retained] = ACTIONS(2728),
    [anon_sym___bridge_transfer] = ACTIONS(2728),
    [anon_sym___complex] = ACTIONS(2728),
    [anon_sym___const] = ACTIONS(2728),
    [anon_sym___imag] = ACTIONS(2728),
    [anon_sym___kindof] = ACTIONS(2728),
    [anon_sym___nonnull] = ACTIONS(2728),
    [anon_sym___nullable] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2728),
    [anon_sym___real] = ACTIONS(2728),
    [anon_sym___strong] = ACTIONS(2728),
    [anon_sym___unsafe_unretained] = ACTIONS(2728),
    [anon_sym___unused] = ACTIONS(2728),
    [anon_sym___weak] = ACTIONS(2728),
    [anon_sym_alignas] = ACTIONS(2728),
    [anon_sym__Alignas] = ACTIONS(2728),
    [sym_primitive_type] = ACTIONS(2728),
    [anon_sym_enum] = ACTIONS(2728),
    [anon_sym_struct] = ACTIONS(2728),
    [anon_sym_union] = ACTIONS(2728),
    [anon_sym_if] = ACTIONS(2728),
    [anon_sym_switch] = ACTIONS(2728),
    [anon_sym_case] = ACTIONS(2728),
    [anon_sym_default] = ACTIONS(2728),
    [anon_sym_while] = ACTIONS(2728),
    [anon_sym_do] = ACTIONS(2728),
    [anon_sym_for] = ACTIONS(2728),
    [anon_sym_in] = ACTIONS(2728),
    [anon_sym_return] = ACTIONS(2728),
    [anon_sym_break] = ACTIONS(2728),
    [anon_sym_continue] = ACTIONS(2728),
    [anon_sym_goto] = ACTIONS(2728),
    [anon_sym_DASH_DASH] = ACTIONS(2730),
    [anon_sym_PLUS_PLUS] = ACTIONS(2730),
    [anon_sym_sizeof] = ACTIONS(2728),
    [anon_sym___alignof__] = ACTIONS(2728),
    [anon_sym___alignof] = ACTIONS(2728),
    [anon_sym__alignof] = ACTIONS(2728),
    [anon_sym_alignof] = ACTIONS(2728),
    [anon_sym__Alignof] = ACTIONS(2728),
    [anon_sym_offsetof] = ACTIONS(2728),
    [anon_sym__Generic] = ACTIONS(2728),
    [anon_sym_asm] = ACTIONS(2728),
    [anon_sym___asm__] = ACTIONS(2728),
    [anon_sym___asm] = ACTIONS(2728),
    [sym_number_literal] = ACTIONS(2730),
    [anon_sym_L_SQUOTE] = ACTIONS(2730),
    [anon_sym_u_SQUOTE] = ACTIONS(2730),
    [anon_sym_U_SQUOTE] = ACTIONS(2730),
    [anon_sym_u8_SQUOTE] = ACTIONS(2730),
    [anon_sym_SQUOTE] = ACTIONS(2730),
    [anon_sym_AT] = ACTIONS(2728),
    [anon_sym_DQUOTE] = ACTIONS(2730),
    [anon_sym_L_DQUOTE] = ACTIONS(2730),
    [anon_sym_u_DQUOTE] = ACTIONS(2730),
    [anon_sym_U_DQUOTE] = ACTIONS(2730),
    [anon_sym_u8_DQUOTE] = ACTIONS(2730),
    [sym_true] = ACTIONS(2728),
    [sym_false] = ACTIONS(2728),
    [anon_sym_NULL] = ACTIONS(2728),
    [anon_sym_nullptr] = ACTIONS(2728),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2728),
    [anon_sym___typeof] = ACTIONS(2728),
    [anon_sym_typeof] = ACTIONS(2728),
    [anon_sym_ATimport] = ACTIONS(2730),
    [aux_sym_preproc_undef_token1] = ACTIONS(2728),
    [anon_sym_POUND] = ACTIONS(2728),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2728),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2728),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2728),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2728),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2728),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2728),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2728),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_API_AVAILABLE] = ACTIONS(2728),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_API_DEPRECATED] = ACTIONS(2728),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2728),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym___deprecated_msg] = ACTIONS(2728),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2728),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2728),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2728),
    [anon_sym_ATprotocol] = ACTIONS(2730),
    [anon_sym_ATinterface] = ACTIONS(2730),
    [anon_sym_ATimplementation] = ACTIONS(2730),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2730),
    [anon_sym_ATsynthesize] = ACTIONS(2730),
    [anon_sym_ATdynamic] = ACTIONS(2730),
    [anon_sym_ATtry] = ACTIONS(2730),
    [anon_sym___try] = ACTIONS(2728),
    [anon_sym_ATthrow] = ACTIONS(2730),
    [anon_sym_ATselector] = ACTIONS(2730),
    [anon_sym_ATavailable] = ACTIONS(2730),
    [anon_sym___builtin_available] = ACTIONS(2728),
    [anon_sym_va_arg] = ACTIONS(2728),
    [anon_sym_ATencode] = ACTIONS(2730),
    [anon_sym_ATsynchronized] = ACTIONS(2730),
    [anon_sym_BOOL] = ACTIONS(2728),
    [anon_sym_IMP] = ACTIONS(2728),
    [anon_sym_SEL] = ACTIONS(2728),
    [anon_sym_Class] = ACTIONS(2728),
    [anon_sym_id] = ACTIONS(2728),
  },
  [STATE(450)] = {
    [sym_identifier] = ACTIONS(2732),
    [aux_sym_preproc_include_token1] = ACTIONS(2732),
    [aux_sym_preproc_include_token2] = ACTIONS(2732),
    [aux_sym_preproc_def_token1] = ACTIONS(2732),
    [aux_sym_preproc_if_token1] = ACTIONS(2732),
    [aux_sym_preproc_if_token2] = ACTIONS(2732),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2732),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2732),
    [aux_sym_preproc_else_token1] = ACTIONS(2732),
    [aux_sym_preproc_elif_token1] = ACTIONS(2732),
    [sym_preproc_directive] = ACTIONS(2732),
    [anon_sym_LPAREN2] = ACTIONS(2734),
    [anon_sym_BANG] = ACTIONS(2734),
    [anon_sym_TILDE] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2732),
    [anon_sym_PLUS] = ACTIONS(2732),
    [anon_sym_STAR] = ACTIONS(2734),
    [anon_sym_CARET] = ACTIONS(2734),
    [anon_sym_AMP] = ACTIONS(2734),
    [anon_sym_SEMI] = ACTIONS(2734),
    [anon_sym___extension__] = ACTIONS(2732),
    [anon_sym_typedef] = ACTIONS(2732),
    [anon_sym_extern] = ACTIONS(2732),
    [anon_sym___attribute__] = ACTIONS(2732),
    [anon_sym___attribute] = ACTIONS(2732),
    [anon_sym_noreturn] = ACTIONS(2732),
    [anon_sym_LBRACK] = ACTIONS(2734),
    [anon_sym___declspec] = ACTIONS(2732),
    [anon_sym___cdecl] = ACTIONS(2732),
    [anon_sym___clrcall] = ACTIONS(2732),
    [anon_sym___stdcall] = ACTIONS(2732),
    [anon_sym___fastcall] = ACTIONS(2732),
    [anon_sym___thiscall] = ACTIONS(2732),
    [anon_sym___vectorcall] = ACTIONS(2732),
    [anon_sym_LBRACE] = ACTIONS(2734),
    [anon_sym_signed] = ACTIONS(2732),
    [anon_sym_unsigned] = ACTIONS(2732),
    [anon_sym_long] = ACTIONS(2732),
    [anon_sym_short] = ACTIONS(2732),
    [anon_sym_static] = ACTIONS(2732),
    [anon_sym_ATautoreleasepool] = ACTIONS(2734),
    [anon_sym_auto] = ACTIONS(2732),
    [anon_sym_register] = ACTIONS(2732),
    [anon_sym_inline] = ACTIONS(2732),
    [anon_sym___inline] = ACTIONS(2732),
    [anon_sym___inline__] = ACTIONS(2732),
    [anon_sym___forceinline] = ACTIONS(2732),
    [anon_sym_thread_local] = ACTIONS(2732),
    [anon_sym___thread] = ACTIONS(2732),
    [anon_sym_CG_EXTERN] = ACTIONS(2732),
    [anon_sym_CG_INLINE] = ACTIONS(2732),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2732),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2732),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2732),
    [anon_sym_IBOutlet] = ACTIONS(2732),
    [anon_sym_IBInspectable] = ACTIONS(2732),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2732),
    [anon_sym_NS_INLINE] = ACTIONS(2732),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2732),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2732),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2732),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2732),
    [anon_sym_const] = ACTIONS(2732),
    [anon_sym_constexpr] = ACTIONS(2732),
    [anon_sym_volatile] = ACTIONS(2732),
    [anon_sym_restrict] = ACTIONS(2732),
    [anon_sym___restrict__] = ACTIONS(2732),
    [anon_sym__Atomic] = ACTIONS(2732),
    [anon_sym__Noreturn] = ACTIONS(2732),
    [anon_sym__Nonnull] = ACTIONS(2732),
    [anon_sym_nullable] = ACTIONS(2732),
    [anon_sym__Complex] = ACTIONS(2732),
    [anon_sym__Nullable] = ACTIONS(2732),
    [anon_sym__Nullable_result] = ACTIONS(2732),
    [anon_sym__Null_unspecified] = ACTIONS(2732),
    [anon_sym___autoreleasing] = ACTIONS(2732),
    [anon_sym___block] = ACTIONS(2732),
    [anon_sym___bridge] = ACTIONS(2732),
    [anon_sym___bridge_retained] = ACTIONS(2732),
    [anon_sym___bridge_transfer] = ACTIONS(2732),
    [anon_sym___complex] = ACTIONS(2732),
    [anon_sym___const] = ACTIONS(2732),
    [anon_sym___imag] = ACTIONS(2732),
    [anon_sym___kindof] = ACTIONS(2732),
    [anon_sym___nonnull] = ACTIONS(2732),
    [anon_sym___nullable] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2732),
    [anon_sym___real] = ACTIONS(2732),
    [anon_sym___strong] = ACTIONS(2732),
    [anon_sym___unsafe_unretained] = ACTIONS(2732),
    [anon_sym___unused] = ACTIONS(2732),
    [anon_sym___weak] = ACTIONS(2732),
    [anon_sym_alignas] = ACTIONS(2732),
    [anon_sym__Alignas] = ACTIONS(2732),
    [sym_primitive_type] = ACTIONS(2732),
    [anon_sym_enum] = ACTIONS(2732),
    [anon_sym_struct] = ACTIONS(2732),
    [anon_sym_union] = ACTIONS(2732),
    [anon_sym_if] = ACTIONS(2732),
    [anon_sym_switch] = ACTIONS(2732),
    [anon_sym_case] = ACTIONS(2732),
    [anon_sym_default] = ACTIONS(2732),
    [anon_sym_while] = ACTIONS(2732),
    [anon_sym_do] = ACTIONS(2732),
    [anon_sym_for] = ACTIONS(2732),
    [anon_sym_in] = ACTIONS(2732),
    [anon_sym_return] = ACTIONS(2732),
    [anon_sym_break] = ACTIONS(2732),
    [anon_sym_continue] = ACTIONS(2732),
    [anon_sym_goto] = ACTIONS(2732),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_sizeof] = ACTIONS(2732),
    [anon_sym___alignof__] = ACTIONS(2732),
    [anon_sym___alignof] = ACTIONS(2732),
    [anon_sym__alignof] = ACTIONS(2732),
    [anon_sym_alignof] = ACTIONS(2732),
    [anon_sym__Alignof] = ACTIONS(2732),
    [anon_sym_offsetof] = ACTIONS(2732),
    [anon_sym__Generic] = ACTIONS(2732),
    [anon_sym_asm] = ACTIONS(2732),
    [anon_sym___asm__] = ACTIONS(2732),
    [anon_sym___asm] = ACTIONS(2732),
    [sym_number_literal] = ACTIONS(2734),
    [anon_sym_L_SQUOTE] = ACTIONS(2734),
    [anon_sym_u_SQUOTE] = ACTIONS(2734),
    [anon_sym_U_SQUOTE] = ACTIONS(2734),
    [anon_sym_u8_SQUOTE] = ACTIONS(2734),
    [anon_sym_SQUOTE] = ACTIONS(2734),
    [anon_sym_AT] = ACTIONS(2732),
    [anon_sym_DQUOTE] = ACTIONS(2734),
    [anon_sym_L_DQUOTE] = ACTIONS(2734),
    [anon_sym_u_DQUOTE] = ACTIONS(2734),
    [anon_sym_U_DQUOTE] = ACTIONS(2734),
    [anon_sym_u8_DQUOTE] = ACTIONS(2734),
    [sym_true] = ACTIONS(2732),
    [sym_false] = ACTIONS(2732),
    [anon_sym_NULL] = ACTIONS(2732),
    [anon_sym_nullptr] = ACTIONS(2732),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2732),
    [anon_sym___typeof] = ACTIONS(2732),
    [anon_sym_typeof] = ACTIONS(2732),
    [anon_sym_ATimport] = ACTIONS(2734),
    [aux_sym_preproc_undef_token1] = ACTIONS(2732),
    [anon_sym_POUND] = ACTIONS(2732),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2732),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2732),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2732),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2732),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2732),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2732),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2732),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_API_AVAILABLE] = ACTIONS(2732),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_API_DEPRECATED] = ACTIONS(2732),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2732),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym___deprecated_msg] = ACTIONS(2732),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2732),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2732),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2732),
    [anon_sym_ATprotocol] = ACTIONS(2734),
    [anon_sym_ATinterface] = ACTIONS(2734),
    [anon_sym_ATimplementation] = ACTIONS(2734),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2734),
    [anon_sym_ATsynthesize] = ACTIONS(2734),
    [anon_sym_ATdynamic] = ACTIONS(2734),
    [anon_sym_ATtry] = ACTIONS(2734),
    [anon_sym___try] = ACTIONS(2732),
    [anon_sym_ATthrow] = ACTIONS(2734),
    [anon_sym_ATselector] = ACTIONS(2734),
    [anon_sym_ATavailable] = ACTIONS(2734),
    [anon_sym___builtin_available] = ACTIONS(2732),
    [anon_sym_va_arg] = ACTIONS(2732),
    [anon_sym_ATencode] = ACTIONS(2734),
    [anon_sym_ATsynchronized] = ACTIONS(2734),
    [anon_sym_BOOL] = ACTIONS(2732),
    [anon_sym_IMP] = ACTIONS(2732),
    [anon_sym_SEL] = ACTIONS(2732),
    [anon_sym_Class] = ACTIONS(2732),
    [anon_sym_id] = ACTIONS(2732),
  },
  [STATE(451)] = {
    [sym_identifier] = ACTIONS(2736),
    [aux_sym_preproc_include_token1] = ACTIONS(2736),
    [aux_sym_preproc_include_token2] = ACTIONS(2736),
    [aux_sym_preproc_def_token1] = ACTIONS(2736),
    [aux_sym_preproc_if_token1] = ACTIONS(2736),
    [aux_sym_preproc_if_token2] = ACTIONS(2736),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2736),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2736),
    [aux_sym_preproc_else_token1] = ACTIONS(2736),
    [aux_sym_preproc_elif_token1] = ACTIONS(2736),
    [sym_preproc_directive] = ACTIONS(2736),
    [anon_sym_LPAREN2] = ACTIONS(2738),
    [anon_sym_BANG] = ACTIONS(2738),
    [anon_sym_TILDE] = ACTIONS(2738),
    [anon_sym_DASH] = ACTIONS(2736),
    [anon_sym_PLUS] = ACTIONS(2736),
    [anon_sym_STAR] = ACTIONS(2738),
    [anon_sym_CARET] = ACTIONS(2738),
    [anon_sym_AMP] = ACTIONS(2738),
    [anon_sym_SEMI] = ACTIONS(2738),
    [anon_sym___extension__] = ACTIONS(2736),
    [anon_sym_typedef] = ACTIONS(2736),
    [anon_sym_extern] = ACTIONS(2736),
    [anon_sym___attribute__] = ACTIONS(2736),
    [anon_sym___attribute] = ACTIONS(2736),
    [anon_sym_noreturn] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(2738),
    [anon_sym___declspec] = ACTIONS(2736),
    [anon_sym___cdecl] = ACTIONS(2736),
    [anon_sym___clrcall] = ACTIONS(2736),
    [anon_sym___stdcall] = ACTIONS(2736),
    [anon_sym___fastcall] = ACTIONS(2736),
    [anon_sym___thiscall] = ACTIONS(2736),
    [anon_sym___vectorcall] = ACTIONS(2736),
    [anon_sym_LBRACE] = ACTIONS(2738),
    [anon_sym_signed] = ACTIONS(2736),
    [anon_sym_unsigned] = ACTIONS(2736),
    [anon_sym_long] = ACTIONS(2736),
    [anon_sym_short] = ACTIONS(2736),
    [anon_sym_static] = ACTIONS(2736),
    [anon_sym_ATautoreleasepool] = ACTIONS(2738),
    [anon_sym_auto] = ACTIONS(2736),
    [anon_sym_register] = ACTIONS(2736),
    [anon_sym_inline] = ACTIONS(2736),
    [anon_sym___inline] = ACTIONS(2736),
    [anon_sym___inline__] = ACTIONS(2736),
    [anon_sym___forceinline] = ACTIONS(2736),
    [anon_sym_thread_local] = ACTIONS(2736),
    [anon_sym___thread] = ACTIONS(2736),
    [anon_sym_CG_EXTERN] = ACTIONS(2736),
    [anon_sym_CG_INLINE] = ACTIONS(2736),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2736),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2736),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2736),
    [anon_sym_IBOutlet] = ACTIONS(2736),
    [anon_sym_IBInspectable] = ACTIONS(2736),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2736),
    [anon_sym_NS_INLINE] = ACTIONS(2736),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2736),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2736),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2736),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2736),
    [anon_sym_const] = ACTIONS(2736),
    [anon_sym_constexpr] = ACTIONS(2736),
    [anon_sym_volatile] = ACTIONS(2736),
    [anon_sym_restrict] = ACTIONS(2736),
    [anon_sym___restrict__] = ACTIONS(2736),
    [anon_sym__Atomic] = ACTIONS(2736),
    [anon_sym__Noreturn] = ACTIONS(2736),
    [anon_sym__Nonnull] = ACTIONS(2736),
    [anon_sym_nullable] = ACTIONS(2736),
    [anon_sym__Complex] = ACTIONS(2736),
    [anon_sym__Nullable] = ACTIONS(2736),
    [anon_sym__Nullable_result] = ACTIONS(2736),
    [anon_sym__Null_unspecified] = ACTIONS(2736),
    [anon_sym___autoreleasing] = ACTIONS(2736),
    [anon_sym___block] = ACTIONS(2736),
    [anon_sym___bridge] = ACTIONS(2736),
    [anon_sym___bridge_retained] = ACTIONS(2736),
    [anon_sym___bridge_transfer] = ACTIONS(2736),
    [anon_sym___complex] = ACTIONS(2736),
    [anon_sym___const] = ACTIONS(2736),
    [anon_sym___imag] = ACTIONS(2736),
    [anon_sym___kindof] = ACTIONS(2736),
    [anon_sym___nonnull] = ACTIONS(2736),
    [anon_sym___nullable] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2736),
    [anon_sym___real] = ACTIONS(2736),
    [anon_sym___strong] = ACTIONS(2736),
    [anon_sym___unsafe_unretained] = ACTIONS(2736),
    [anon_sym___unused] = ACTIONS(2736),
    [anon_sym___weak] = ACTIONS(2736),
    [anon_sym_alignas] = ACTIONS(2736),
    [anon_sym__Alignas] = ACTIONS(2736),
    [sym_primitive_type] = ACTIONS(2736),
    [anon_sym_enum] = ACTIONS(2736),
    [anon_sym_struct] = ACTIONS(2736),
    [anon_sym_union] = ACTIONS(2736),
    [anon_sym_if] = ACTIONS(2736),
    [anon_sym_switch] = ACTIONS(2736),
    [anon_sym_case] = ACTIONS(2736),
    [anon_sym_default] = ACTIONS(2736),
    [anon_sym_while] = ACTIONS(2736),
    [anon_sym_do] = ACTIONS(2736),
    [anon_sym_for] = ACTIONS(2736),
    [anon_sym_in] = ACTIONS(2736),
    [anon_sym_return] = ACTIONS(2736),
    [anon_sym_break] = ACTIONS(2736),
    [anon_sym_continue] = ACTIONS(2736),
    [anon_sym_goto] = ACTIONS(2736),
    [anon_sym_DASH_DASH] = ACTIONS(2738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2738),
    [anon_sym_sizeof] = ACTIONS(2736),
    [anon_sym___alignof__] = ACTIONS(2736),
    [anon_sym___alignof] = ACTIONS(2736),
    [anon_sym__alignof] = ACTIONS(2736),
    [anon_sym_alignof] = ACTIONS(2736),
    [anon_sym__Alignof] = ACTIONS(2736),
    [anon_sym_offsetof] = ACTIONS(2736),
    [anon_sym__Generic] = ACTIONS(2736),
    [anon_sym_asm] = ACTIONS(2736),
    [anon_sym___asm__] = ACTIONS(2736),
    [anon_sym___asm] = ACTIONS(2736),
    [sym_number_literal] = ACTIONS(2738),
    [anon_sym_L_SQUOTE] = ACTIONS(2738),
    [anon_sym_u_SQUOTE] = ACTIONS(2738),
    [anon_sym_U_SQUOTE] = ACTIONS(2738),
    [anon_sym_u8_SQUOTE] = ACTIONS(2738),
    [anon_sym_SQUOTE] = ACTIONS(2738),
    [anon_sym_AT] = ACTIONS(2736),
    [anon_sym_DQUOTE] = ACTIONS(2738),
    [anon_sym_L_DQUOTE] = ACTIONS(2738),
    [anon_sym_u_DQUOTE] = ACTIONS(2738),
    [anon_sym_U_DQUOTE] = ACTIONS(2738),
    [anon_sym_u8_DQUOTE] = ACTIONS(2738),
    [sym_true] = ACTIONS(2736),
    [sym_false] = ACTIONS(2736),
    [anon_sym_NULL] = ACTIONS(2736),
    [anon_sym_nullptr] = ACTIONS(2736),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2736),
    [anon_sym___typeof] = ACTIONS(2736),
    [anon_sym_typeof] = ACTIONS(2736),
    [anon_sym_ATimport] = ACTIONS(2738),
    [aux_sym_preproc_undef_token1] = ACTIONS(2736),
    [anon_sym_POUND] = ACTIONS(2736),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2736),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2736),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2736),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2736),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2736),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2736),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2736),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_API_AVAILABLE] = ACTIONS(2736),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_API_DEPRECATED] = ACTIONS(2736),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2736),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym___deprecated_msg] = ACTIONS(2736),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2736),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2736),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2736),
    [anon_sym_ATprotocol] = ACTIONS(2738),
    [anon_sym_ATinterface] = ACTIONS(2738),
    [anon_sym_ATimplementation] = ACTIONS(2738),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2738),
    [anon_sym_ATsynthesize] = ACTIONS(2738),
    [anon_sym_ATdynamic] = ACTIONS(2738),
    [anon_sym_ATtry] = ACTIONS(2738),
    [anon_sym___try] = ACTIONS(2736),
    [anon_sym_ATthrow] = ACTIONS(2738),
    [anon_sym_ATselector] = ACTIONS(2738),
    [anon_sym_ATavailable] = ACTIONS(2738),
    [anon_sym___builtin_available] = ACTIONS(2736),
    [anon_sym_va_arg] = ACTIONS(2736),
    [anon_sym_ATencode] = ACTIONS(2738),
    [anon_sym_ATsynchronized] = ACTIONS(2738),
    [anon_sym_BOOL] = ACTIONS(2736),
    [anon_sym_IMP] = ACTIONS(2736),
    [anon_sym_SEL] = ACTIONS(2736),
    [anon_sym_Class] = ACTIONS(2736),
    [anon_sym_id] = ACTIONS(2736),
  },
  [STATE(452)] = {
    [sym_identifier] = ACTIONS(2792),
    [aux_sym_preproc_include_token1] = ACTIONS(2792),
    [aux_sym_preproc_include_token2] = ACTIONS(2792),
    [aux_sym_preproc_def_token1] = ACTIONS(2792),
    [aux_sym_preproc_if_token1] = ACTIONS(2792),
    [aux_sym_preproc_if_token2] = ACTIONS(2792),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2792),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2792),
    [aux_sym_preproc_else_token1] = ACTIONS(2792),
    [aux_sym_preproc_elif_token1] = ACTIONS(2792),
    [sym_preproc_directive] = ACTIONS(2792),
    [anon_sym_LPAREN2] = ACTIONS(2794),
    [anon_sym_BANG] = ACTIONS(2794),
    [anon_sym_TILDE] = ACTIONS(2794),
    [anon_sym_DASH] = ACTIONS(2792),
    [anon_sym_PLUS] = ACTIONS(2792),
    [anon_sym_STAR] = ACTIONS(2794),
    [anon_sym_CARET] = ACTIONS(2794),
    [anon_sym_AMP] = ACTIONS(2794),
    [anon_sym_SEMI] = ACTIONS(2794),
    [anon_sym___extension__] = ACTIONS(2792),
    [anon_sym_typedef] = ACTIONS(2792),
    [anon_sym_extern] = ACTIONS(2792),
    [anon_sym___attribute__] = ACTIONS(2792),
    [anon_sym___attribute] = ACTIONS(2792),
    [anon_sym_noreturn] = ACTIONS(2792),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym___declspec] = ACTIONS(2792),
    [anon_sym___cdecl] = ACTIONS(2792),
    [anon_sym___clrcall] = ACTIONS(2792),
    [anon_sym___stdcall] = ACTIONS(2792),
    [anon_sym___fastcall] = ACTIONS(2792),
    [anon_sym___thiscall] = ACTIONS(2792),
    [anon_sym___vectorcall] = ACTIONS(2792),
    [anon_sym_LBRACE] = ACTIONS(2794),
    [anon_sym_signed] = ACTIONS(2792),
    [anon_sym_unsigned] = ACTIONS(2792),
    [anon_sym_long] = ACTIONS(2792),
    [anon_sym_short] = ACTIONS(2792),
    [anon_sym_static] = ACTIONS(2792),
    [anon_sym_ATautoreleasepool] = ACTIONS(2794),
    [anon_sym_auto] = ACTIONS(2792),
    [anon_sym_register] = ACTIONS(2792),
    [anon_sym_inline] = ACTIONS(2792),
    [anon_sym___inline] = ACTIONS(2792),
    [anon_sym___inline__] = ACTIONS(2792),
    [anon_sym___forceinline] = ACTIONS(2792),
    [anon_sym_thread_local] = ACTIONS(2792),
    [anon_sym___thread] = ACTIONS(2792),
    [anon_sym_CG_EXTERN] = ACTIONS(2792),
    [anon_sym_CG_INLINE] = ACTIONS(2792),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2792),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2792),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2792),
    [anon_sym_IBOutlet] = ACTIONS(2792),
    [anon_sym_IBInspectable] = ACTIONS(2792),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2792),
    [anon_sym_NS_INLINE] = ACTIONS(2792),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2792),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2792),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2792),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2792),
    [anon_sym_const] = ACTIONS(2792),
    [anon_sym_constexpr] = ACTIONS(2792),
    [anon_sym_volatile] = ACTIONS(2792),
    [anon_sym_restrict] = ACTIONS(2792),
    [anon_sym___restrict__] = ACTIONS(2792),
    [anon_sym__Atomic] = ACTIONS(2792),
    [anon_sym__Noreturn] = ACTIONS(2792),
    [anon_sym__Nonnull] = ACTIONS(2792),
    [anon_sym_nullable] = ACTIONS(2792),
    [anon_sym__Complex] = ACTIONS(2792),
    [anon_sym__Nullable] = ACTIONS(2792),
    [anon_sym__Nullable_result] = ACTIONS(2792),
    [anon_sym__Null_unspecified] = ACTIONS(2792),
    [anon_sym___autoreleasing] = ACTIONS(2792),
    [anon_sym___block] = ACTIONS(2792),
    [anon_sym___bridge] = ACTIONS(2792),
    [anon_sym___bridge_retained] = ACTIONS(2792),
    [anon_sym___bridge_transfer] = ACTIONS(2792),
    [anon_sym___complex] = ACTIONS(2792),
    [anon_sym___const] = ACTIONS(2792),
    [anon_sym___imag] = ACTIONS(2792),
    [anon_sym___kindof] = ACTIONS(2792),
    [anon_sym___nonnull] = ACTIONS(2792),
    [anon_sym___nullable] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2792),
    [anon_sym___real] = ACTIONS(2792),
    [anon_sym___strong] = ACTIONS(2792),
    [anon_sym___unsafe_unretained] = ACTIONS(2792),
    [anon_sym___unused] = ACTIONS(2792),
    [anon_sym___weak] = ACTIONS(2792),
    [anon_sym_alignas] = ACTIONS(2792),
    [anon_sym__Alignas] = ACTIONS(2792),
    [sym_primitive_type] = ACTIONS(2792),
    [anon_sym_enum] = ACTIONS(2792),
    [anon_sym_struct] = ACTIONS(2792),
    [anon_sym_union] = ACTIONS(2792),
    [anon_sym_if] = ACTIONS(2792),
    [anon_sym_switch] = ACTIONS(2792),
    [anon_sym_case] = ACTIONS(2792),
    [anon_sym_default] = ACTIONS(2792),
    [anon_sym_while] = ACTIONS(2792),
    [anon_sym_do] = ACTIONS(2792),
    [anon_sym_for] = ACTIONS(2792),
    [anon_sym_in] = ACTIONS(2792),
    [anon_sym_return] = ACTIONS(2792),
    [anon_sym_break] = ACTIONS(2792),
    [anon_sym_continue] = ACTIONS(2792),
    [anon_sym_goto] = ACTIONS(2792),
    [anon_sym_DASH_DASH] = ACTIONS(2794),
    [anon_sym_PLUS_PLUS] = ACTIONS(2794),
    [anon_sym_sizeof] = ACTIONS(2792),
    [anon_sym___alignof__] = ACTIONS(2792),
    [anon_sym___alignof] = ACTIONS(2792),
    [anon_sym__alignof] = ACTIONS(2792),
    [anon_sym_alignof] = ACTIONS(2792),
    [anon_sym__Alignof] = ACTIONS(2792),
    [anon_sym_offsetof] = ACTIONS(2792),
    [anon_sym__Generic] = ACTIONS(2792),
    [anon_sym_asm] = ACTIONS(2792),
    [anon_sym___asm__] = ACTIONS(2792),
    [anon_sym___asm] = ACTIONS(2792),
    [sym_number_literal] = ACTIONS(2794),
    [anon_sym_L_SQUOTE] = ACTIONS(2794),
    [anon_sym_u_SQUOTE] = ACTIONS(2794),
    [anon_sym_U_SQUOTE] = ACTIONS(2794),
    [anon_sym_u8_SQUOTE] = ACTIONS(2794),
    [anon_sym_SQUOTE] = ACTIONS(2794),
    [anon_sym_AT] = ACTIONS(2792),
    [anon_sym_DQUOTE] = ACTIONS(2794),
    [anon_sym_L_DQUOTE] = ACTIONS(2794),
    [anon_sym_u_DQUOTE] = ACTIONS(2794),
    [anon_sym_U_DQUOTE] = ACTIONS(2794),
    [anon_sym_u8_DQUOTE] = ACTIONS(2794),
    [sym_true] = ACTIONS(2792),
    [sym_false] = ACTIONS(2792),
    [anon_sym_NULL] = ACTIONS(2792),
    [anon_sym_nullptr] = ACTIONS(2792),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2792),
    [anon_sym___typeof] = ACTIONS(2792),
    [anon_sym_typeof] = ACTIONS(2792),
    [anon_sym_ATimport] = ACTIONS(2794),
    [aux_sym_preproc_undef_token1] = ACTIONS(2792),
    [anon_sym_POUND] = ACTIONS(2792),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2792),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2792),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2792),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2792),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2792),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2792),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2792),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_API_AVAILABLE] = ACTIONS(2792),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_API_DEPRECATED] = ACTIONS(2792),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2792),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym___deprecated_msg] = ACTIONS(2792),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2792),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2792),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2792),
    [anon_sym_ATprotocol] = ACTIONS(2794),
    [anon_sym_ATinterface] = ACTIONS(2794),
    [anon_sym_ATimplementation] = ACTIONS(2794),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2794),
    [anon_sym_ATsynthesize] = ACTIONS(2794),
    [anon_sym_ATdynamic] = ACTIONS(2794),
    [anon_sym_ATtry] = ACTIONS(2794),
    [anon_sym___try] = ACTIONS(2792),
    [anon_sym_ATthrow] = ACTIONS(2794),
    [anon_sym_ATselector] = ACTIONS(2794),
    [anon_sym_ATavailable] = ACTIONS(2794),
    [anon_sym___builtin_available] = ACTIONS(2792),
    [anon_sym_va_arg] = ACTIONS(2792),
    [anon_sym_ATencode] = ACTIONS(2794),
    [anon_sym_ATsynchronized] = ACTIONS(2794),
    [anon_sym_BOOL] = ACTIONS(2792),
    [anon_sym_IMP] = ACTIONS(2792),
    [anon_sym_SEL] = ACTIONS(2792),
    [anon_sym_Class] = ACTIONS(2792),
    [anon_sym_id] = ACTIONS(2792),
  },
  [STATE(453)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token2] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [aux_sym_preproc_else_token1] = ACTIONS(2796),
    [aux_sym_preproc_elif_token1] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATsynthesize] = ACTIONS(2798),
    [anon_sym_ATdynamic] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(454)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token2] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [aux_sym_preproc_else_token1] = ACTIONS(2796),
    [aux_sym_preproc_elif_token1] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATsynthesize] = ACTIONS(2798),
    [anon_sym_ATdynamic] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(455)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token2] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [aux_sym_preproc_else_token1] = ACTIONS(2796),
    [aux_sym_preproc_elif_token1] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATsynthesize] = ACTIONS(2798),
    [anon_sym_ATdynamic] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(456)] = {
    [sym_identifier] = ACTIONS(2800),
    [aux_sym_preproc_include_token1] = ACTIONS(2800),
    [aux_sym_preproc_include_token2] = ACTIONS(2800),
    [aux_sym_preproc_def_token1] = ACTIONS(2800),
    [aux_sym_preproc_if_token1] = ACTIONS(2800),
    [aux_sym_preproc_if_token2] = ACTIONS(2800),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2800),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2800),
    [aux_sym_preproc_else_token1] = ACTIONS(2800),
    [aux_sym_preproc_elif_token1] = ACTIONS(2800),
    [sym_preproc_directive] = ACTIONS(2800),
    [anon_sym_LPAREN2] = ACTIONS(2802),
    [anon_sym_BANG] = ACTIONS(2802),
    [anon_sym_TILDE] = ACTIONS(2802),
    [anon_sym_DASH] = ACTIONS(2800),
    [anon_sym_PLUS] = ACTIONS(2800),
    [anon_sym_STAR] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(2802),
    [anon_sym_AMP] = ACTIONS(2802),
    [anon_sym_SEMI] = ACTIONS(2802),
    [anon_sym___extension__] = ACTIONS(2800),
    [anon_sym_typedef] = ACTIONS(2800),
    [anon_sym_extern] = ACTIONS(2800),
    [anon_sym___attribute__] = ACTIONS(2800),
    [anon_sym___attribute] = ACTIONS(2800),
    [anon_sym_noreturn] = ACTIONS(2800),
    [anon_sym_LBRACK] = ACTIONS(2802),
    [anon_sym___declspec] = ACTIONS(2800),
    [anon_sym___cdecl] = ACTIONS(2800),
    [anon_sym___clrcall] = ACTIONS(2800),
    [anon_sym___stdcall] = ACTIONS(2800),
    [anon_sym___fastcall] = ACTIONS(2800),
    [anon_sym___thiscall] = ACTIONS(2800),
    [anon_sym___vectorcall] = ACTIONS(2800),
    [anon_sym_LBRACE] = ACTIONS(2802),
    [anon_sym_signed] = ACTIONS(2800),
    [anon_sym_unsigned] = ACTIONS(2800),
    [anon_sym_long] = ACTIONS(2800),
    [anon_sym_short] = ACTIONS(2800),
    [anon_sym_static] = ACTIONS(2800),
    [anon_sym_ATautoreleasepool] = ACTIONS(2802),
    [anon_sym_auto] = ACTIONS(2800),
    [anon_sym_register] = ACTIONS(2800),
    [anon_sym_inline] = ACTIONS(2800),
    [anon_sym___inline] = ACTIONS(2800),
    [anon_sym___inline__] = ACTIONS(2800),
    [anon_sym___forceinline] = ACTIONS(2800),
    [anon_sym_thread_local] = ACTIONS(2800),
    [anon_sym___thread] = ACTIONS(2800),
    [anon_sym_CG_EXTERN] = ACTIONS(2800),
    [anon_sym_CG_INLINE] = ACTIONS(2800),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2800),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2800),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2800),
    [anon_sym_IBOutlet] = ACTIONS(2800),
    [anon_sym_IBInspectable] = ACTIONS(2800),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2800),
    [anon_sym_NS_INLINE] = ACTIONS(2800),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2800),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2800),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2800),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2800),
    [anon_sym_const] = ACTIONS(2800),
    [anon_sym_constexpr] = ACTIONS(2800),
    [anon_sym_volatile] = ACTIONS(2800),
    [anon_sym_restrict] = ACTIONS(2800),
    [anon_sym___restrict__] = ACTIONS(2800),
    [anon_sym__Atomic] = ACTIONS(2800),
    [anon_sym__Noreturn] = ACTIONS(2800),
    [anon_sym__Nonnull] = ACTIONS(2800),
    [anon_sym_nullable] = ACTIONS(2800),
    [anon_sym__Complex] = ACTIONS(2800),
    [anon_sym__Nullable] = ACTIONS(2800),
    [anon_sym__Nullable_result] = ACTIONS(2800),
    [anon_sym__Null_unspecified] = ACTIONS(2800),
    [anon_sym___autoreleasing] = ACTIONS(2800),
    [anon_sym___block] = ACTIONS(2800),
    [anon_sym___bridge] = ACTIONS(2800),
    [anon_sym___bridge_retained] = ACTIONS(2800),
    [anon_sym___bridge_transfer] = ACTIONS(2800),
    [anon_sym___complex] = ACTIONS(2800),
    [anon_sym___const] = ACTIONS(2800),
    [anon_sym___imag] = ACTIONS(2800),
    [anon_sym___kindof] = ACTIONS(2800),
    [anon_sym___nonnull] = ACTIONS(2800),
    [anon_sym___nullable] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2800),
    [anon_sym___real] = ACTIONS(2800),
    [anon_sym___strong] = ACTIONS(2800),
    [anon_sym___unsafe_unretained] = ACTIONS(2800),
    [anon_sym___unused] = ACTIONS(2800),
    [anon_sym___weak] = ACTIONS(2800),
    [anon_sym_alignas] = ACTIONS(2800),
    [anon_sym__Alignas] = ACTIONS(2800),
    [sym_primitive_type] = ACTIONS(2800),
    [anon_sym_enum] = ACTIONS(2800),
    [anon_sym_struct] = ACTIONS(2800),
    [anon_sym_union] = ACTIONS(2800),
    [anon_sym_if] = ACTIONS(2800),
    [anon_sym_switch] = ACTIONS(2800),
    [anon_sym_case] = ACTIONS(2800),
    [anon_sym_default] = ACTIONS(2800),
    [anon_sym_while] = ACTIONS(2800),
    [anon_sym_do] = ACTIONS(2800),
    [anon_sym_for] = ACTIONS(2800),
    [anon_sym_in] = ACTIONS(2800),
    [anon_sym_return] = ACTIONS(2800),
    [anon_sym_break] = ACTIONS(2800),
    [anon_sym_continue] = ACTIONS(2800),
    [anon_sym_goto] = ACTIONS(2800),
    [anon_sym_DASH_DASH] = ACTIONS(2802),
    [anon_sym_PLUS_PLUS] = ACTIONS(2802),
    [anon_sym_sizeof] = ACTIONS(2800),
    [anon_sym___alignof__] = ACTIONS(2800),
    [anon_sym___alignof] = ACTIONS(2800),
    [anon_sym__alignof] = ACTIONS(2800),
    [anon_sym_alignof] = ACTIONS(2800),
    [anon_sym__Alignof] = ACTIONS(2800),
    [anon_sym_offsetof] = ACTIONS(2800),
    [anon_sym__Generic] = ACTIONS(2800),
    [anon_sym_asm] = ACTIONS(2800),
    [anon_sym___asm__] = ACTIONS(2800),
    [anon_sym___asm] = ACTIONS(2800),
    [sym_number_literal] = ACTIONS(2802),
    [anon_sym_L_SQUOTE] = ACTIONS(2802),
    [anon_sym_u_SQUOTE] = ACTIONS(2802),
    [anon_sym_U_SQUOTE] = ACTIONS(2802),
    [anon_sym_u8_SQUOTE] = ACTIONS(2802),
    [anon_sym_SQUOTE] = ACTIONS(2802),
    [anon_sym_AT] = ACTIONS(2800),
    [anon_sym_DQUOTE] = ACTIONS(2802),
    [anon_sym_L_DQUOTE] = ACTIONS(2802),
    [anon_sym_u_DQUOTE] = ACTIONS(2802),
    [anon_sym_U_DQUOTE] = ACTIONS(2802),
    [anon_sym_u8_DQUOTE] = ACTIONS(2802),
    [sym_true] = ACTIONS(2800),
    [sym_false] = ACTIONS(2800),
    [anon_sym_NULL] = ACTIONS(2800),
    [anon_sym_nullptr] = ACTIONS(2800),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2800),
    [anon_sym___typeof] = ACTIONS(2800),
    [anon_sym_typeof] = ACTIONS(2800),
    [anon_sym_ATimport] = ACTIONS(2802),
    [aux_sym_preproc_undef_token1] = ACTIONS(2800),
    [anon_sym_POUND] = ACTIONS(2800),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2800),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2800),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2800),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2800),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2800),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2800),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2800),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_API_AVAILABLE] = ACTIONS(2800),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_API_DEPRECATED] = ACTIONS(2800),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2800),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym___deprecated_msg] = ACTIONS(2800),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2800),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2800),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2800),
    [anon_sym_ATprotocol] = ACTIONS(2802),
    [anon_sym_ATinterface] = ACTIONS(2802),
    [anon_sym_ATimplementation] = ACTIONS(2802),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2802),
    [anon_sym_ATsynthesize] = ACTIONS(2802),
    [anon_sym_ATdynamic] = ACTIONS(2802),
    [anon_sym_ATtry] = ACTIONS(2802),
    [anon_sym___try] = ACTIONS(2800),
    [anon_sym_ATthrow] = ACTIONS(2802),
    [anon_sym_ATselector] = ACTIONS(2802),
    [anon_sym_ATavailable] = ACTIONS(2802),
    [anon_sym___builtin_available] = ACTIONS(2800),
    [anon_sym_va_arg] = ACTIONS(2800),
    [anon_sym_ATencode] = ACTIONS(2802),
    [anon_sym_ATsynchronized] = ACTIONS(2802),
    [anon_sym_BOOL] = ACTIONS(2800),
    [anon_sym_IMP] = ACTIONS(2800),
    [anon_sym_SEL] = ACTIONS(2800),
    [anon_sym_Class] = ACTIONS(2800),
    [anon_sym_id] = ACTIONS(2800),
  },
  [STATE(457)] = {
    [sym_identifier] = ACTIONS(2824),
    [aux_sym_preproc_include_token1] = ACTIONS(2824),
    [aux_sym_preproc_include_token2] = ACTIONS(2824),
    [aux_sym_preproc_def_token1] = ACTIONS(2824),
    [aux_sym_preproc_if_token1] = ACTIONS(2824),
    [aux_sym_preproc_if_token2] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2824),
    [aux_sym_preproc_else_token1] = ACTIONS(2824),
    [aux_sym_preproc_elif_token1] = ACTIONS(2824),
    [sym_preproc_directive] = ACTIONS(2824),
    [anon_sym_LPAREN2] = ACTIONS(2826),
    [anon_sym_BANG] = ACTIONS(2826),
    [anon_sym_TILDE] = ACTIONS(2826),
    [anon_sym_DASH] = ACTIONS(2824),
    [anon_sym_PLUS] = ACTIONS(2824),
    [anon_sym_STAR] = ACTIONS(2826),
    [anon_sym_CARET] = ACTIONS(2826),
    [anon_sym_AMP] = ACTIONS(2826),
    [anon_sym_SEMI] = ACTIONS(2826),
    [anon_sym___extension__] = ACTIONS(2824),
    [anon_sym_typedef] = ACTIONS(2824),
    [anon_sym_extern] = ACTIONS(2824),
    [anon_sym___attribute__] = ACTIONS(2824),
    [anon_sym___attribute] = ACTIONS(2824),
    [anon_sym_noreturn] = ACTIONS(2824),
    [anon_sym_LBRACK] = ACTIONS(2826),
    [anon_sym___declspec] = ACTIONS(2824),
    [anon_sym___cdecl] = ACTIONS(2824),
    [anon_sym___clrcall] = ACTIONS(2824),
    [anon_sym___stdcall] = ACTIONS(2824),
    [anon_sym___fastcall] = ACTIONS(2824),
    [anon_sym___thiscall] = ACTIONS(2824),
    [anon_sym___vectorcall] = ACTIONS(2824),
    [anon_sym_LBRACE] = ACTIONS(2826),
    [anon_sym_signed] = ACTIONS(2824),
    [anon_sym_unsigned] = ACTIONS(2824),
    [anon_sym_long] = ACTIONS(2824),
    [anon_sym_short] = ACTIONS(2824),
    [anon_sym_static] = ACTIONS(2824),
    [anon_sym_ATautoreleasepool] = ACTIONS(2826),
    [anon_sym_auto] = ACTIONS(2824),
    [anon_sym_register] = ACTIONS(2824),
    [anon_sym_inline] = ACTIONS(2824),
    [anon_sym___inline] = ACTIONS(2824),
    [anon_sym___inline__] = ACTIONS(2824),
    [anon_sym___forceinline] = ACTIONS(2824),
    [anon_sym_thread_local] = ACTIONS(2824),
    [anon_sym___thread] = ACTIONS(2824),
    [anon_sym_CG_EXTERN] = ACTIONS(2824),
    [anon_sym_CG_INLINE] = ACTIONS(2824),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2824),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2824),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2824),
    [anon_sym_IBOutlet] = ACTIONS(2824),
    [anon_sym_IBInspectable] = ACTIONS(2824),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2824),
    [anon_sym_NS_INLINE] = ACTIONS(2824),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2824),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2824),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2824),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2824),
    [anon_sym_const] = ACTIONS(2824),
    [anon_sym_constexpr] = ACTIONS(2824),
    [anon_sym_volatile] = ACTIONS(2824),
    [anon_sym_restrict] = ACTIONS(2824),
    [anon_sym___restrict__] = ACTIONS(2824),
    [anon_sym__Atomic] = ACTIONS(2824),
    [anon_sym__Noreturn] = ACTIONS(2824),
    [anon_sym__Nonnull] = ACTIONS(2824),
    [anon_sym_nullable] = ACTIONS(2824),
    [anon_sym__Complex] = ACTIONS(2824),
    [anon_sym__Nullable] = ACTIONS(2824),
    [anon_sym__Nullable_result] = ACTIONS(2824),
    [anon_sym__Null_unspecified] = ACTIONS(2824),
    [anon_sym___autoreleasing] = ACTIONS(2824),
    [anon_sym___block] = ACTIONS(2824),
    [anon_sym___bridge] = ACTIONS(2824),
    [anon_sym___bridge_retained] = ACTIONS(2824),
    [anon_sym___bridge_transfer] = ACTIONS(2824),
    [anon_sym___complex] = ACTIONS(2824),
    [anon_sym___const] = ACTIONS(2824),
    [anon_sym___imag] = ACTIONS(2824),
    [anon_sym___kindof] = ACTIONS(2824),
    [anon_sym___nonnull] = ACTIONS(2824),
    [anon_sym___nullable] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2824),
    [anon_sym___real] = ACTIONS(2824),
    [anon_sym___strong] = ACTIONS(2824),
    [anon_sym___unsafe_unretained] = ACTIONS(2824),
    [anon_sym___unused] = ACTIONS(2824),
    [anon_sym___weak] = ACTIONS(2824),
    [anon_sym_alignas] = ACTIONS(2824),
    [anon_sym__Alignas] = ACTIONS(2824),
    [sym_primitive_type] = ACTIONS(2824),
    [anon_sym_enum] = ACTIONS(2824),
    [anon_sym_struct] = ACTIONS(2824),
    [anon_sym_union] = ACTIONS(2824),
    [anon_sym_if] = ACTIONS(2824),
    [anon_sym_switch] = ACTIONS(2824),
    [anon_sym_case] = ACTIONS(2824),
    [anon_sym_default] = ACTIONS(2824),
    [anon_sym_while] = ACTIONS(2824),
    [anon_sym_do] = ACTIONS(2824),
    [anon_sym_for] = ACTIONS(2824),
    [anon_sym_in] = ACTIONS(2824),
    [anon_sym_return] = ACTIONS(2824),
    [anon_sym_break] = ACTIONS(2824),
    [anon_sym_continue] = ACTIONS(2824),
    [anon_sym_goto] = ACTIONS(2824),
    [anon_sym_DASH_DASH] = ACTIONS(2826),
    [anon_sym_PLUS_PLUS] = ACTIONS(2826),
    [anon_sym_sizeof] = ACTIONS(2824),
    [anon_sym___alignof__] = ACTIONS(2824),
    [anon_sym___alignof] = ACTIONS(2824),
    [anon_sym__alignof] = ACTIONS(2824),
    [anon_sym_alignof] = ACTIONS(2824),
    [anon_sym__Alignof] = ACTIONS(2824),
    [anon_sym_offsetof] = ACTIONS(2824),
    [anon_sym__Generic] = ACTIONS(2824),
    [anon_sym_asm] = ACTIONS(2824),
    [anon_sym___asm__] = ACTIONS(2824),
    [anon_sym___asm] = ACTIONS(2824),
    [sym_number_literal] = ACTIONS(2826),
    [anon_sym_L_SQUOTE] = ACTIONS(2826),
    [anon_sym_u_SQUOTE] = ACTIONS(2826),
    [anon_sym_U_SQUOTE] = ACTIONS(2826),
    [anon_sym_u8_SQUOTE] = ACTIONS(2826),
    [anon_sym_SQUOTE] = ACTIONS(2826),
    [anon_sym_AT] = ACTIONS(2824),
    [anon_sym_DQUOTE] = ACTIONS(2826),
    [anon_sym_L_DQUOTE] = ACTIONS(2826),
    [anon_sym_u_DQUOTE] = ACTIONS(2826),
    [anon_sym_U_DQUOTE] = ACTIONS(2826),
    [anon_sym_u8_DQUOTE] = ACTIONS(2826),
    [sym_true] = ACTIONS(2824),
    [sym_false] = ACTIONS(2824),
    [anon_sym_NULL] = ACTIONS(2824),
    [anon_sym_nullptr] = ACTIONS(2824),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2824),
    [anon_sym___typeof] = ACTIONS(2824),
    [anon_sym_typeof] = ACTIONS(2824),
    [anon_sym_ATimport] = ACTIONS(2826),
    [aux_sym_preproc_undef_token1] = ACTIONS(2824),
    [anon_sym_POUND] = ACTIONS(2824),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2824),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2824),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2824),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2824),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2824),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2824),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2824),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_API_AVAILABLE] = ACTIONS(2824),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_API_DEPRECATED] = ACTIONS(2824),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2824),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym___deprecated_msg] = ACTIONS(2824),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2824),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2824),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2824),
    [anon_sym_ATprotocol] = ACTIONS(2826),
    [anon_sym_ATinterface] = ACTIONS(2826),
    [anon_sym_ATimplementation] = ACTIONS(2826),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2826),
    [anon_sym_ATsynthesize] = ACTIONS(2826),
    [anon_sym_ATdynamic] = ACTIONS(2826),
    [anon_sym_ATtry] = ACTIONS(2826),
    [anon_sym___try] = ACTIONS(2824),
    [anon_sym_ATthrow] = ACTIONS(2826),
    [anon_sym_ATselector] = ACTIONS(2826),
    [anon_sym_ATavailable] = ACTIONS(2826),
    [anon_sym___builtin_available] = ACTIONS(2824),
    [anon_sym_va_arg] = ACTIONS(2824),
    [anon_sym_ATencode] = ACTIONS(2826),
    [anon_sym_ATsynchronized] = ACTIONS(2826),
    [anon_sym_BOOL] = ACTIONS(2824),
    [anon_sym_IMP] = ACTIONS(2824),
    [anon_sym_SEL] = ACTIONS(2824),
    [anon_sym_Class] = ACTIONS(2824),
    [anon_sym_id] = ACTIONS(2824),
  },
  [STATE(458)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token2] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [aux_sym_preproc_else_token1] = ACTIONS(2804),
    [aux_sym_preproc_elif_token1] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATsynthesize] = ACTIONS(2806),
    [anon_sym_ATdynamic] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(459)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token2] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [aux_sym_preproc_else_token1] = ACTIONS(2804),
    [aux_sym_preproc_elif_token1] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATsynthesize] = ACTIONS(2806),
    [anon_sym_ATdynamic] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(460)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token2] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [aux_sym_preproc_else_token1] = ACTIONS(2804),
    [aux_sym_preproc_elif_token1] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATsynthesize] = ACTIONS(2806),
    [anon_sym_ATdynamic] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(461)] = {
    [sym_identifier] = ACTIONS(2808),
    [aux_sym_preproc_include_token1] = ACTIONS(2808),
    [aux_sym_preproc_include_token2] = ACTIONS(2808),
    [aux_sym_preproc_def_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token2] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2808),
    [aux_sym_preproc_else_token1] = ACTIONS(2808),
    [aux_sym_preproc_elif_token1] = ACTIONS(2808),
    [sym_preproc_directive] = ACTIONS(2808),
    [anon_sym_LPAREN2] = ACTIONS(2810),
    [anon_sym_BANG] = ACTIONS(2810),
    [anon_sym_TILDE] = ACTIONS(2810),
    [anon_sym_DASH] = ACTIONS(2808),
    [anon_sym_PLUS] = ACTIONS(2808),
    [anon_sym_STAR] = ACTIONS(2810),
    [anon_sym_CARET] = ACTIONS(2810),
    [anon_sym_AMP] = ACTIONS(2810),
    [anon_sym_SEMI] = ACTIONS(2810),
    [anon_sym___extension__] = ACTIONS(2808),
    [anon_sym_typedef] = ACTIONS(2808),
    [anon_sym_extern] = ACTIONS(2808),
    [anon_sym___attribute__] = ACTIONS(2808),
    [anon_sym___attribute] = ACTIONS(2808),
    [anon_sym_noreturn] = ACTIONS(2808),
    [anon_sym_LBRACK] = ACTIONS(2810),
    [anon_sym___declspec] = ACTIONS(2808),
    [anon_sym___cdecl] = ACTIONS(2808),
    [anon_sym___clrcall] = ACTIONS(2808),
    [anon_sym___stdcall] = ACTIONS(2808),
    [anon_sym___fastcall] = ACTIONS(2808),
    [anon_sym___thiscall] = ACTIONS(2808),
    [anon_sym___vectorcall] = ACTIONS(2808),
    [anon_sym_LBRACE] = ACTIONS(2810),
    [anon_sym_signed] = ACTIONS(2808),
    [anon_sym_unsigned] = ACTIONS(2808),
    [anon_sym_long] = ACTIONS(2808),
    [anon_sym_short] = ACTIONS(2808),
    [anon_sym_static] = ACTIONS(2808),
    [anon_sym_ATautoreleasepool] = ACTIONS(2810),
    [anon_sym_auto] = ACTIONS(2808),
    [anon_sym_register] = ACTIONS(2808),
    [anon_sym_inline] = ACTIONS(2808),
    [anon_sym___inline] = ACTIONS(2808),
    [anon_sym___inline__] = ACTIONS(2808),
    [anon_sym___forceinline] = ACTIONS(2808),
    [anon_sym_thread_local] = ACTIONS(2808),
    [anon_sym___thread] = ACTIONS(2808),
    [anon_sym_CG_EXTERN] = ACTIONS(2808),
    [anon_sym_CG_INLINE] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2808),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2808),
    [anon_sym_IBOutlet] = ACTIONS(2808),
    [anon_sym_IBInspectable] = ACTIONS(2808),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2808),
    [anon_sym_NS_INLINE] = ACTIONS(2808),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2808),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2808),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2808),
    [anon_sym_const] = ACTIONS(2808),
    [anon_sym_constexpr] = ACTIONS(2808),
    [anon_sym_volatile] = ACTIONS(2808),
    [anon_sym_restrict] = ACTIONS(2808),
    [anon_sym___restrict__] = ACTIONS(2808),
    [anon_sym__Atomic] = ACTIONS(2808),
    [anon_sym__Noreturn] = ACTIONS(2808),
    [anon_sym__Nonnull] = ACTIONS(2808),
    [anon_sym_nullable] = ACTIONS(2808),
    [anon_sym__Complex] = ACTIONS(2808),
    [anon_sym__Nullable] = ACTIONS(2808),
    [anon_sym__Nullable_result] = ACTIONS(2808),
    [anon_sym__Null_unspecified] = ACTIONS(2808),
    [anon_sym___autoreleasing] = ACTIONS(2808),
    [anon_sym___block] = ACTIONS(2808),
    [anon_sym___bridge] = ACTIONS(2808),
    [anon_sym___bridge_retained] = ACTIONS(2808),
    [anon_sym___bridge_transfer] = ACTIONS(2808),
    [anon_sym___complex] = ACTIONS(2808),
    [anon_sym___const] = ACTIONS(2808),
    [anon_sym___imag] = ACTIONS(2808),
    [anon_sym___kindof] = ACTIONS(2808),
    [anon_sym___nonnull] = ACTIONS(2808),
    [anon_sym___nullable] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2808),
    [anon_sym___real] = ACTIONS(2808),
    [anon_sym___strong] = ACTIONS(2808),
    [anon_sym___unsafe_unretained] = ACTIONS(2808),
    [anon_sym___unused] = ACTIONS(2808),
    [anon_sym___weak] = ACTIONS(2808),
    [anon_sym_alignas] = ACTIONS(2808),
    [anon_sym__Alignas] = ACTIONS(2808),
    [sym_primitive_type] = ACTIONS(2808),
    [anon_sym_enum] = ACTIONS(2808),
    [anon_sym_struct] = ACTIONS(2808),
    [anon_sym_union] = ACTIONS(2808),
    [anon_sym_if] = ACTIONS(2808),
    [anon_sym_switch] = ACTIONS(2808),
    [anon_sym_case] = ACTIONS(2808),
    [anon_sym_default] = ACTIONS(2808),
    [anon_sym_while] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2808),
    [anon_sym_in] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_break] = ACTIONS(2808),
    [anon_sym_continue] = ACTIONS(2808),
    [anon_sym_goto] = ACTIONS(2808),
    [anon_sym_DASH_DASH] = ACTIONS(2810),
    [anon_sym_PLUS_PLUS] = ACTIONS(2810),
    [anon_sym_sizeof] = ACTIONS(2808),
    [anon_sym___alignof__] = ACTIONS(2808),
    [anon_sym___alignof] = ACTIONS(2808),
    [anon_sym__alignof] = ACTIONS(2808),
    [anon_sym_alignof] = ACTIONS(2808),
    [anon_sym__Alignof] = ACTIONS(2808),
    [anon_sym_offsetof] = ACTIONS(2808),
    [anon_sym__Generic] = ACTIONS(2808),
    [anon_sym_asm] = ACTIONS(2808),
    [anon_sym___asm__] = ACTIONS(2808),
    [anon_sym___asm] = ACTIONS(2808),
    [sym_number_literal] = ACTIONS(2810),
    [anon_sym_L_SQUOTE] = ACTIONS(2810),
    [anon_sym_u_SQUOTE] = ACTIONS(2810),
    [anon_sym_U_SQUOTE] = ACTIONS(2810),
    [anon_sym_u8_SQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2810),
    [anon_sym_AT] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_L_DQUOTE] = ACTIONS(2810),
    [anon_sym_u_DQUOTE] = ACTIONS(2810),
    [anon_sym_U_DQUOTE] = ACTIONS(2810),
    [anon_sym_u8_DQUOTE] = ACTIONS(2810),
    [sym_true] = ACTIONS(2808),
    [sym_false] = ACTIONS(2808),
    [anon_sym_NULL] = ACTIONS(2808),
    [anon_sym_nullptr] = ACTIONS(2808),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2808),
    [anon_sym___typeof] = ACTIONS(2808),
    [anon_sym_typeof] = ACTIONS(2808),
    [anon_sym_ATimport] = ACTIONS(2810),
    [aux_sym_preproc_undef_token1] = ACTIONS(2808),
    [anon_sym_POUND] = ACTIONS(2808),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2808),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2808),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2808),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_API_AVAILABLE] = ACTIONS(2808),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_API_DEPRECATED] = ACTIONS(2808),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym___deprecated_msg] = ACTIONS(2808),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2808),
    [anon_sym_ATprotocol] = ACTIONS(2810),
    [anon_sym_ATinterface] = ACTIONS(2810),
    [anon_sym_ATimplementation] = ACTIONS(2810),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2810),
    [anon_sym_ATsynthesize] = ACTIONS(2810),
    [anon_sym_ATdynamic] = ACTIONS(2810),
    [anon_sym_ATtry] = ACTIONS(2810),
    [anon_sym___try] = ACTIONS(2808),
    [anon_sym_ATthrow] = ACTIONS(2810),
    [anon_sym_ATselector] = ACTIONS(2810),
    [anon_sym_ATavailable] = ACTIONS(2810),
    [anon_sym___builtin_available] = ACTIONS(2808),
    [anon_sym_va_arg] = ACTIONS(2808),
    [anon_sym_ATencode] = ACTIONS(2810),
    [anon_sym_ATsynchronized] = ACTIONS(2810),
    [anon_sym_BOOL] = ACTIONS(2808),
    [anon_sym_IMP] = ACTIONS(2808),
    [anon_sym_SEL] = ACTIONS(2808),
    [anon_sym_Class] = ACTIONS(2808),
    [anon_sym_id] = ACTIONS(2808),
  },
  [STATE(462)] = {
    [sym_catch_clause] = STATE(572),
    [sym_finally_clause] = STATE(971),
    [aux_sym_try_statement_repeat1] = STATE(572),
    [ts_builtin_sym_end] = ACTIONS(2308),
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATcatch] = ACTIONS(2868),
    [anon_sym___catch] = ACTIONS(2870),
    [anon_sym_ATfinally] = ACTIONS(2872),
    [anon_sym___finally] = ACTIONS(2874),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(463)] = {
    [sym_identifier] = ACTIONS(2808),
    [aux_sym_preproc_include_token1] = ACTIONS(2808),
    [aux_sym_preproc_include_token2] = ACTIONS(2808),
    [aux_sym_preproc_def_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token2] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2808),
    [aux_sym_preproc_else_token1] = ACTIONS(2808),
    [aux_sym_preproc_elif_token1] = ACTIONS(2808),
    [sym_preproc_directive] = ACTIONS(2808),
    [anon_sym_LPAREN2] = ACTIONS(2810),
    [anon_sym_BANG] = ACTIONS(2810),
    [anon_sym_TILDE] = ACTIONS(2810),
    [anon_sym_DASH] = ACTIONS(2808),
    [anon_sym_PLUS] = ACTIONS(2808),
    [anon_sym_STAR] = ACTIONS(2810),
    [anon_sym_CARET] = ACTIONS(2810),
    [anon_sym_AMP] = ACTIONS(2810),
    [anon_sym_SEMI] = ACTIONS(2810),
    [anon_sym___extension__] = ACTIONS(2808),
    [anon_sym_typedef] = ACTIONS(2808),
    [anon_sym_extern] = ACTIONS(2808),
    [anon_sym___attribute__] = ACTIONS(2808),
    [anon_sym___attribute] = ACTIONS(2808),
    [anon_sym_noreturn] = ACTIONS(2808),
    [anon_sym_LBRACK] = ACTIONS(2810),
    [anon_sym___declspec] = ACTIONS(2808),
    [anon_sym___cdecl] = ACTIONS(2808),
    [anon_sym___clrcall] = ACTIONS(2808),
    [anon_sym___stdcall] = ACTIONS(2808),
    [anon_sym___fastcall] = ACTIONS(2808),
    [anon_sym___thiscall] = ACTIONS(2808),
    [anon_sym___vectorcall] = ACTIONS(2808),
    [anon_sym_LBRACE] = ACTIONS(2810),
    [anon_sym_signed] = ACTIONS(2808),
    [anon_sym_unsigned] = ACTIONS(2808),
    [anon_sym_long] = ACTIONS(2808),
    [anon_sym_short] = ACTIONS(2808),
    [anon_sym_static] = ACTIONS(2808),
    [anon_sym_ATautoreleasepool] = ACTIONS(2810),
    [anon_sym_auto] = ACTIONS(2808),
    [anon_sym_register] = ACTIONS(2808),
    [anon_sym_inline] = ACTIONS(2808),
    [anon_sym___inline] = ACTIONS(2808),
    [anon_sym___inline__] = ACTIONS(2808),
    [anon_sym___forceinline] = ACTIONS(2808),
    [anon_sym_thread_local] = ACTIONS(2808),
    [anon_sym___thread] = ACTIONS(2808),
    [anon_sym_CG_EXTERN] = ACTIONS(2808),
    [anon_sym_CG_INLINE] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2808),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2808),
    [anon_sym_IBOutlet] = ACTIONS(2808),
    [anon_sym_IBInspectable] = ACTIONS(2808),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2808),
    [anon_sym_NS_INLINE] = ACTIONS(2808),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2808),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2808),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2808),
    [anon_sym_const] = ACTIONS(2808),
    [anon_sym_constexpr] = ACTIONS(2808),
    [anon_sym_volatile] = ACTIONS(2808),
    [anon_sym_restrict] = ACTIONS(2808),
    [anon_sym___restrict__] = ACTIONS(2808),
    [anon_sym__Atomic] = ACTIONS(2808),
    [anon_sym__Noreturn] = ACTIONS(2808),
    [anon_sym__Nonnull] = ACTIONS(2808),
    [anon_sym_nullable] = ACTIONS(2808),
    [anon_sym__Complex] = ACTIONS(2808),
    [anon_sym__Nullable] = ACTIONS(2808),
    [anon_sym__Nullable_result] = ACTIONS(2808),
    [anon_sym__Null_unspecified] = ACTIONS(2808),
    [anon_sym___autoreleasing] = ACTIONS(2808),
    [anon_sym___block] = ACTIONS(2808),
    [anon_sym___bridge] = ACTIONS(2808),
    [anon_sym___bridge_retained] = ACTIONS(2808),
    [anon_sym___bridge_transfer] = ACTIONS(2808),
    [anon_sym___complex] = ACTIONS(2808),
    [anon_sym___const] = ACTIONS(2808),
    [anon_sym___imag] = ACTIONS(2808),
    [anon_sym___kindof] = ACTIONS(2808),
    [anon_sym___nonnull] = ACTIONS(2808),
    [anon_sym___nullable] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2808),
    [anon_sym___real] = ACTIONS(2808),
    [anon_sym___strong] = ACTIONS(2808),
    [anon_sym___unsafe_unretained] = ACTIONS(2808),
    [anon_sym___unused] = ACTIONS(2808),
    [anon_sym___weak] = ACTIONS(2808),
    [anon_sym_alignas] = ACTIONS(2808),
    [anon_sym__Alignas] = ACTIONS(2808),
    [sym_primitive_type] = ACTIONS(2808),
    [anon_sym_enum] = ACTIONS(2808),
    [anon_sym_struct] = ACTIONS(2808),
    [anon_sym_union] = ACTIONS(2808),
    [anon_sym_if] = ACTIONS(2808),
    [anon_sym_switch] = ACTIONS(2808),
    [anon_sym_case] = ACTIONS(2808),
    [anon_sym_default] = ACTIONS(2808),
    [anon_sym_while] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2808),
    [anon_sym_in] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_break] = ACTIONS(2808),
    [anon_sym_continue] = ACTIONS(2808),
    [anon_sym_goto] = ACTIONS(2808),
    [anon_sym_DASH_DASH] = ACTIONS(2810),
    [anon_sym_PLUS_PLUS] = ACTIONS(2810),
    [anon_sym_sizeof] = ACTIONS(2808),
    [anon_sym___alignof__] = ACTIONS(2808),
    [anon_sym___alignof] = ACTIONS(2808),
    [anon_sym__alignof] = ACTIONS(2808),
    [anon_sym_alignof] = ACTIONS(2808),
    [anon_sym__Alignof] = ACTIONS(2808),
    [anon_sym_offsetof] = ACTIONS(2808),
    [anon_sym__Generic] = ACTIONS(2808),
    [anon_sym_asm] = ACTIONS(2808),
    [anon_sym___asm__] = ACTIONS(2808),
    [anon_sym___asm] = ACTIONS(2808),
    [sym_number_literal] = ACTIONS(2810),
    [anon_sym_L_SQUOTE] = ACTIONS(2810),
    [anon_sym_u_SQUOTE] = ACTIONS(2810),
    [anon_sym_U_SQUOTE] = ACTIONS(2810),
    [anon_sym_u8_SQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2810),
    [anon_sym_AT] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_L_DQUOTE] = ACTIONS(2810),
    [anon_sym_u_DQUOTE] = ACTIONS(2810),
    [anon_sym_U_DQUOTE] = ACTIONS(2810),
    [anon_sym_u8_DQUOTE] = ACTIONS(2810),
    [sym_true] = ACTIONS(2808),
    [sym_false] = ACTIONS(2808),
    [anon_sym_NULL] = ACTIONS(2808),
    [anon_sym_nullptr] = ACTIONS(2808),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2808),
    [anon_sym___typeof] = ACTIONS(2808),
    [anon_sym_typeof] = ACTIONS(2808),
    [anon_sym_ATimport] = ACTIONS(2810),
    [aux_sym_preproc_undef_token1] = ACTIONS(2808),
    [anon_sym_POUND] = ACTIONS(2808),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2808),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2808),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2808),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_API_AVAILABLE] = ACTIONS(2808),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_API_DEPRECATED] = ACTIONS(2808),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym___deprecated_msg] = ACTIONS(2808),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2808),
    [anon_sym_ATprotocol] = ACTIONS(2810),
    [anon_sym_ATinterface] = ACTIONS(2810),
    [anon_sym_ATimplementation] = ACTIONS(2810),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2810),
    [anon_sym_ATsynthesize] = ACTIONS(2810),
    [anon_sym_ATdynamic] = ACTIONS(2810),
    [anon_sym_ATtry] = ACTIONS(2810),
    [anon_sym___try] = ACTIONS(2808),
    [anon_sym_ATthrow] = ACTIONS(2810),
    [anon_sym_ATselector] = ACTIONS(2810),
    [anon_sym_ATavailable] = ACTIONS(2810),
    [anon_sym___builtin_available] = ACTIONS(2808),
    [anon_sym_va_arg] = ACTIONS(2808),
    [anon_sym_ATencode] = ACTIONS(2810),
    [anon_sym_ATsynchronized] = ACTIONS(2810),
    [anon_sym_BOOL] = ACTIONS(2808),
    [anon_sym_IMP] = ACTIONS(2808),
    [anon_sym_SEL] = ACTIONS(2808),
    [anon_sym_Class] = ACTIONS(2808),
    [anon_sym_id] = ACTIONS(2808),
  },
  [STATE(464)] = {
    [sym_identifier] = ACTIONS(2812),
    [aux_sym_preproc_include_token1] = ACTIONS(2812),
    [aux_sym_preproc_include_token2] = ACTIONS(2812),
    [aux_sym_preproc_def_token1] = ACTIONS(2812),
    [aux_sym_preproc_if_token1] = ACTIONS(2812),
    [aux_sym_preproc_if_token2] = ACTIONS(2812),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2812),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2812),
    [aux_sym_preproc_else_token1] = ACTIONS(2812),
    [aux_sym_preproc_elif_token1] = ACTIONS(2812),
    [sym_preproc_directive] = ACTIONS(2812),
    [anon_sym_LPAREN2] = ACTIONS(2814),
    [anon_sym_BANG] = ACTIONS(2814),
    [anon_sym_TILDE] = ACTIONS(2814),
    [anon_sym_DASH] = ACTIONS(2812),
    [anon_sym_PLUS] = ACTIONS(2812),
    [anon_sym_STAR] = ACTIONS(2814),
    [anon_sym_CARET] = ACTIONS(2814),
    [anon_sym_AMP] = ACTIONS(2814),
    [anon_sym_SEMI] = ACTIONS(2814),
    [anon_sym___extension__] = ACTIONS(2812),
    [anon_sym_typedef] = ACTIONS(2812),
    [anon_sym_extern] = ACTIONS(2812),
    [anon_sym___attribute__] = ACTIONS(2812),
    [anon_sym___attribute] = ACTIONS(2812),
    [anon_sym_noreturn] = ACTIONS(2812),
    [anon_sym_LBRACK] = ACTIONS(2814),
    [anon_sym___declspec] = ACTIONS(2812),
    [anon_sym___cdecl] = ACTIONS(2812),
    [anon_sym___clrcall] = ACTIONS(2812),
    [anon_sym___stdcall] = ACTIONS(2812),
    [anon_sym___fastcall] = ACTIONS(2812),
    [anon_sym___thiscall] = ACTIONS(2812),
    [anon_sym___vectorcall] = ACTIONS(2812),
    [anon_sym_LBRACE] = ACTIONS(2814),
    [anon_sym_signed] = ACTIONS(2812),
    [anon_sym_unsigned] = ACTIONS(2812),
    [anon_sym_long] = ACTIONS(2812),
    [anon_sym_short] = ACTIONS(2812),
    [anon_sym_static] = ACTIONS(2812),
    [anon_sym_ATautoreleasepool] = ACTIONS(2814),
    [anon_sym_auto] = ACTIONS(2812),
    [anon_sym_register] = ACTIONS(2812),
    [anon_sym_inline] = ACTIONS(2812),
    [anon_sym___inline] = ACTIONS(2812),
    [anon_sym___inline__] = ACTIONS(2812),
    [anon_sym___forceinline] = ACTIONS(2812),
    [anon_sym_thread_local] = ACTIONS(2812),
    [anon_sym___thread] = ACTIONS(2812),
    [anon_sym_CG_EXTERN] = ACTIONS(2812),
    [anon_sym_CG_INLINE] = ACTIONS(2812),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2812),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2812),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2812),
    [anon_sym_IBOutlet] = ACTIONS(2812),
    [anon_sym_IBInspectable] = ACTIONS(2812),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2812),
    [anon_sym_NS_INLINE] = ACTIONS(2812),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2812),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2812),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2812),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2812),
    [anon_sym_const] = ACTIONS(2812),
    [anon_sym_constexpr] = ACTIONS(2812),
    [anon_sym_volatile] = ACTIONS(2812),
    [anon_sym_restrict] = ACTIONS(2812),
    [anon_sym___restrict__] = ACTIONS(2812),
    [anon_sym__Atomic] = ACTIONS(2812),
    [anon_sym__Noreturn] = ACTIONS(2812),
    [anon_sym__Nonnull] = ACTIONS(2812),
    [anon_sym_nullable] = ACTIONS(2812),
    [anon_sym__Complex] = ACTIONS(2812),
    [anon_sym__Nullable] = ACTIONS(2812),
    [anon_sym__Nullable_result] = ACTIONS(2812),
    [anon_sym__Null_unspecified] = ACTIONS(2812),
    [anon_sym___autoreleasing] = ACTIONS(2812),
    [anon_sym___block] = ACTIONS(2812),
    [anon_sym___bridge] = ACTIONS(2812),
    [anon_sym___bridge_retained] = ACTIONS(2812),
    [anon_sym___bridge_transfer] = ACTIONS(2812),
    [anon_sym___complex] = ACTIONS(2812),
    [anon_sym___const] = ACTIONS(2812),
    [anon_sym___imag] = ACTIONS(2812),
    [anon_sym___kindof] = ACTIONS(2812),
    [anon_sym___nonnull] = ACTIONS(2812),
    [anon_sym___nullable] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2812),
    [anon_sym___real] = ACTIONS(2812),
    [anon_sym___strong] = ACTIONS(2812),
    [anon_sym___unsafe_unretained] = ACTIONS(2812),
    [anon_sym___unused] = ACTIONS(2812),
    [anon_sym___weak] = ACTIONS(2812),
    [anon_sym_alignas] = ACTIONS(2812),
    [anon_sym__Alignas] = ACTIONS(2812),
    [sym_primitive_type] = ACTIONS(2812),
    [anon_sym_enum] = ACTIONS(2812),
    [anon_sym_struct] = ACTIONS(2812),
    [anon_sym_union] = ACTIONS(2812),
    [anon_sym_if] = ACTIONS(2812),
    [anon_sym_switch] = ACTIONS(2812),
    [anon_sym_case] = ACTIONS(2812),
    [anon_sym_default] = ACTIONS(2812),
    [anon_sym_while] = ACTIONS(2812),
    [anon_sym_do] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2812),
    [anon_sym_in] = ACTIONS(2812),
    [anon_sym_return] = ACTIONS(2812),
    [anon_sym_break] = ACTIONS(2812),
    [anon_sym_continue] = ACTIONS(2812),
    [anon_sym_goto] = ACTIONS(2812),
    [anon_sym_DASH_DASH] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_sizeof] = ACTIONS(2812),
    [anon_sym___alignof__] = ACTIONS(2812),
    [anon_sym___alignof] = ACTIONS(2812),
    [anon_sym__alignof] = ACTIONS(2812),
    [anon_sym_alignof] = ACTIONS(2812),
    [anon_sym__Alignof] = ACTIONS(2812),
    [anon_sym_offsetof] = ACTIONS(2812),
    [anon_sym__Generic] = ACTIONS(2812),
    [anon_sym_asm] = ACTIONS(2812),
    [anon_sym___asm__] = ACTIONS(2812),
    [anon_sym___asm] = ACTIONS(2812),
    [sym_number_literal] = ACTIONS(2814),
    [anon_sym_L_SQUOTE] = ACTIONS(2814),
    [anon_sym_u_SQUOTE] = ACTIONS(2814),
    [anon_sym_U_SQUOTE] = ACTIONS(2814),
    [anon_sym_u8_SQUOTE] = ACTIONS(2814),
    [anon_sym_SQUOTE] = ACTIONS(2814),
    [anon_sym_AT] = ACTIONS(2812),
    [anon_sym_DQUOTE] = ACTIONS(2814),
    [anon_sym_L_DQUOTE] = ACTIONS(2814),
    [anon_sym_u_DQUOTE] = ACTIONS(2814),
    [anon_sym_U_DQUOTE] = ACTIONS(2814),
    [anon_sym_u8_DQUOTE] = ACTIONS(2814),
    [sym_true] = ACTIONS(2812),
    [sym_false] = ACTIONS(2812),
    [anon_sym_NULL] = ACTIONS(2812),
    [anon_sym_nullptr] = ACTIONS(2812),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2812),
    [anon_sym___typeof] = ACTIONS(2812),
    [anon_sym_typeof] = ACTIONS(2812),
    [anon_sym_ATimport] = ACTIONS(2814),
    [aux_sym_preproc_undef_token1] = ACTIONS(2812),
    [anon_sym_POUND] = ACTIONS(2812),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2812),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2812),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2812),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2812),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2812),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2812),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2812),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_API_AVAILABLE] = ACTIONS(2812),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_API_DEPRECATED] = ACTIONS(2812),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2812),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym___deprecated_msg] = ACTIONS(2812),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2812),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2812),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2812),
    [anon_sym_ATprotocol] = ACTIONS(2814),
    [anon_sym_ATinterface] = ACTIONS(2814),
    [anon_sym_ATimplementation] = ACTIONS(2814),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2814),
    [anon_sym_ATsynthesize] = ACTIONS(2814),
    [anon_sym_ATdynamic] = ACTIONS(2814),
    [anon_sym_ATtry] = ACTIONS(2814),
    [anon_sym___try] = ACTIONS(2812),
    [anon_sym_ATthrow] = ACTIONS(2814),
    [anon_sym_ATselector] = ACTIONS(2814),
    [anon_sym_ATavailable] = ACTIONS(2814),
    [anon_sym___builtin_available] = ACTIONS(2812),
    [anon_sym_va_arg] = ACTIONS(2812),
    [anon_sym_ATencode] = ACTIONS(2814),
    [anon_sym_ATsynchronized] = ACTIONS(2814),
    [anon_sym_BOOL] = ACTIONS(2812),
    [anon_sym_IMP] = ACTIONS(2812),
    [anon_sym_SEL] = ACTIONS(2812),
    [anon_sym_Class] = ACTIONS(2812),
    [anon_sym_id] = ACTIONS(2812),
  },
  [STATE(465)] = {
    [sym_identifier] = ACTIONS(2816),
    [aux_sym_preproc_include_token1] = ACTIONS(2816),
    [aux_sym_preproc_include_token2] = ACTIONS(2816),
    [aux_sym_preproc_def_token1] = ACTIONS(2816),
    [aux_sym_preproc_if_token1] = ACTIONS(2816),
    [aux_sym_preproc_if_token2] = ACTIONS(2816),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2816),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2816),
    [aux_sym_preproc_else_token1] = ACTIONS(2816),
    [aux_sym_preproc_elif_token1] = ACTIONS(2816),
    [sym_preproc_directive] = ACTIONS(2816),
    [anon_sym_LPAREN2] = ACTIONS(2818),
    [anon_sym_BANG] = ACTIONS(2818),
    [anon_sym_TILDE] = ACTIONS(2818),
    [anon_sym_DASH] = ACTIONS(2816),
    [anon_sym_PLUS] = ACTIONS(2816),
    [anon_sym_STAR] = ACTIONS(2818),
    [anon_sym_CARET] = ACTIONS(2818),
    [anon_sym_AMP] = ACTIONS(2818),
    [anon_sym_SEMI] = ACTIONS(2818),
    [anon_sym___extension__] = ACTIONS(2816),
    [anon_sym_typedef] = ACTIONS(2816),
    [anon_sym_extern] = ACTIONS(2816),
    [anon_sym___attribute__] = ACTIONS(2816),
    [anon_sym___attribute] = ACTIONS(2816),
    [anon_sym_noreturn] = ACTIONS(2816),
    [anon_sym_LBRACK] = ACTIONS(2818),
    [anon_sym___declspec] = ACTIONS(2816),
    [anon_sym___cdecl] = ACTIONS(2816),
    [anon_sym___clrcall] = ACTIONS(2816),
    [anon_sym___stdcall] = ACTIONS(2816),
    [anon_sym___fastcall] = ACTIONS(2816),
    [anon_sym___thiscall] = ACTIONS(2816),
    [anon_sym___vectorcall] = ACTIONS(2816),
    [anon_sym_LBRACE] = ACTIONS(2818),
    [anon_sym_signed] = ACTIONS(2816),
    [anon_sym_unsigned] = ACTIONS(2816),
    [anon_sym_long] = ACTIONS(2816),
    [anon_sym_short] = ACTIONS(2816),
    [anon_sym_static] = ACTIONS(2816),
    [anon_sym_ATautoreleasepool] = ACTIONS(2818),
    [anon_sym_auto] = ACTIONS(2816),
    [anon_sym_register] = ACTIONS(2816),
    [anon_sym_inline] = ACTIONS(2816),
    [anon_sym___inline] = ACTIONS(2816),
    [anon_sym___inline__] = ACTIONS(2816),
    [anon_sym___forceinline] = ACTIONS(2816),
    [anon_sym_thread_local] = ACTIONS(2816),
    [anon_sym___thread] = ACTIONS(2816),
    [anon_sym_CG_EXTERN] = ACTIONS(2816),
    [anon_sym_CG_INLINE] = ACTIONS(2816),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2816),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2816),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2816),
    [anon_sym_IBOutlet] = ACTIONS(2816),
    [anon_sym_IBInspectable] = ACTIONS(2816),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2816),
    [anon_sym_NS_INLINE] = ACTIONS(2816),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2816),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2816),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2816),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2816),
    [anon_sym_const] = ACTIONS(2816),
    [anon_sym_constexpr] = ACTIONS(2816),
    [anon_sym_volatile] = ACTIONS(2816),
    [anon_sym_restrict] = ACTIONS(2816),
    [anon_sym___restrict__] = ACTIONS(2816),
    [anon_sym__Atomic] = ACTIONS(2816),
    [anon_sym__Noreturn] = ACTIONS(2816),
    [anon_sym__Nonnull] = ACTIONS(2816),
    [anon_sym_nullable] = ACTIONS(2816),
    [anon_sym__Complex] = ACTIONS(2816),
    [anon_sym__Nullable] = ACTIONS(2816),
    [anon_sym__Nullable_result] = ACTIONS(2816),
    [anon_sym__Null_unspecified] = ACTIONS(2816),
    [anon_sym___autoreleasing] = ACTIONS(2816),
    [anon_sym___block] = ACTIONS(2816),
    [anon_sym___bridge] = ACTIONS(2816),
    [anon_sym___bridge_retained] = ACTIONS(2816),
    [anon_sym___bridge_transfer] = ACTIONS(2816),
    [anon_sym___complex] = ACTIONS(2816),
    [anon_sym___const] = ACTIONS(2816),
    [anon_sym___imag] = ACTIONS(2816),
    [anon_sym___kindof] = ACTIONS(2816),
    [anon_sym___nonnull] = ACTIONS(2816),
    [anon_sym___nullable] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2816),
    [anon_sym___real] = ACTIONS(2816),
    [anon_sym___strong] = ACTIONS(2816),
    [anon_sym___unsafe_unretained] = ACTIONS(2816),
    [anon_sym___unused] = ACTIONS(2816),
    [anon_sym___weak] = ACTIONS(2816),
    [anon_sym_alignas] = ACTIONS(2816),
    [anon_sym__Alignas] = ACTIONS(2816),
    [sym_primitive_type] = ACTIONS(2816),
    [anon_sym_enum] = ACTIONS(2816),
    [anon_sym_struct] = ACTIONS(2816),
    [anon_sym_union] = ACTIONS(2816),
    [anon_sym_if] = ACTIONS(2816),
    [anon_sym_switch] = ACTIONS(2816),
    [anon_sym_case] = ACTIONS(2816),
    [anon_sym_default] = ACTIONS(2816),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_do] = ACTIONS(2816),
    [anon_sym_for] = ACTIONS(2816),
    [anon_sym_in] = ACTIONS(2816),
    [anon_sym_return] = ACTIONS(2816),
    [anon_sym_break] = ACTIONS(2816),
    [anon_sym_continue] = ACTIONS(2816),
    [anon_sym_goto] = ACTIONS(2816),
    [anon_sym_DASH_DASH] = ACTIONS(2818),
    [anon_sym_PLUS_PLUS] = ACTIONS(2818),
    [anon_sym_sizeof] = ACTIONS(2816),
    [anon_sym___alignof__] = ACTIONS(2816),
    [anon_sym___alignof] = ACTIONS(2816),
    [anon_sym__alignof] = ACTIONS(2816),
    [anon_sym_alignof] = ACTIONS(2816),
    [anon_sym__Alignof] = ACTIONS(2816),
    [anon_sym_offsetof] = ACTIONS(2816),
    [anon_sym__Generic] = ACTIONS(2816),
    [anon_sym_asm] = ACTIONS(2816),
    [anon_sym___asm__] = ACTIONS(2816),
    [anon_sym___asm] = ACTIONS(2816),
    [sym_number_literal] = ACTIONS(2818),
    [anon_sym_L_SQUOTE] = ACTIONS(2818),
    [anon_sym_u_SQUOTE] = ACTIONS(2818),
    [anon_sym_U_SQUOTE] = ACTIONS(2818),
    [anon_sym_u8_SQUOTE] = ACTIONS(2818),
    [anon_sym_SQUOTE] = ACTIONS(2818),
    [anon_sym_AT] = ACTIONS(2816),
    [anon_sym_DQUOTE] = ACTIONS(2818),
    [anon_sym_L_DQUOTE] = ACTIONS(2818),
    [anon_sym_u_DQUOTE] = ACTIONS(2818),
    [anon_sym_U_DQUOTE] = ACTIONS(2818),
    [anon_sym_u8_DQUOTE] = ACTIONS(2818),
    [sym_true] = ACTIONS(2816),
    [sym_false] = ACTIONS(2816),
    [anon_sym_NULL] = ACTIONS(2816),
    [anon_sym_nullptr] = ACTIONS(2816),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2816),
    [anon_sym___typeof] = ACTIONS(2816),
    [anon_sym_typeof] = ACTIONS(2816),
    [anon_sym_ATimport] = ACTIONS(2818),
    [aux_sym_preproc_undef_token1] = ACTIONS(2816),
    [anon_sym_POUND] = ACTIONS(2816),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2816),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2816),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2816),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2816),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2816),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2816),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2816),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_API_AVAILABLE] = ACTIONS(2816),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_API_DEPRECATED] = ACTIONS(2816),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2816),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym___deprecated_msg] = ACTIONS(2816),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2816),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2816),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2816),
    [anon_sym_ATprotocol] = ACTIONS(2818),
    [anon_sym_ATinterface] = ACTIONS(2818),
    [anon_sym_ATimplementation] = ACTIONS(2818),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2818),
    [anon_sym_ATsynthesize] = ACTIONS(2818),
    [anon_sym_ATdynamic] = ACTIONS(2818),
    [anon_sym_ATtry] = ACTIONS(2818),
    [anon_sym___try] = ACTIONS(2816),
    [anon_sym_ATthrow] = ACTIONS(2818),
    [anon_sym_ATselector] = ACTIONS(2818),
    [anon_sym_ATavailable] = ACTIONS(2818),
    [anon_sym___builtin_available] = ACTIONS(2816),
    [anon_sym_va_arg] = ACTIONS(2816),
    [anon_sym_ATencode] = ACTIONS(2818),
    [anon_sym_ATsynchronized] = ACTIONS(2818),
    [anon_sym_BOOL] = ACTIONS(2816),
    [anon_sym_IMP] = ACTIONS(2816),
    [anon_sym_SEL] = ACTIONS(2816),
    [anon_sym_Class] = ACTIONS(2816),
    [anon_sym_id] = ACTIONS(2816),
  },
  [STATE(466)] = {
    [sym_identifier] = ACTIONS(2820),
    [aux_sym_preproc_include_token1] = ACTIONS(2820),
    [aux_sym_preproc_include_token2] = ACTIONS(2820),
    [aux_sym_preproc_def_token1] = ACTIONS(2820),
    [aux_sym_preproc_if_token1] = ACTIONS(2820),
    [aux_sym_preproc_if_token2] = ACTIONS(2820),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2820),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2820),
    [aux_sym_preproc_else_token1] = ACTIONS(2820),
    [aux_sym_preproc_elif_token1] = ACTIONS(2820),
    [sym_preproc_directive] = ACTIONS(2820),
    [anon_sym_LPAREN2] = ACTIONS(2822),
    [anon_sym_BANG] = ACTIONS(2822),
    [anon_sym_TILDE] = ACTIONS(2822),
    [anon_sym_DASH] = ACTIONS(2820),
    [anon_sym_PLUS] = ACTIONS(2820),
    [anon_sym_STAR] = ACTIONS(2822),
    [anon_sym_CARET] = ACTIONS(2822),
    [anon_sym_AMP] = ACTIONS(2822),
    [anon_sym_SEMI] = ACTIONS(2822),
    [anon_sym___extension__] = ACTIONS(2820),
    [anon_sym_typedef] = ACTIONS(2820),
    [anon_sym_extern] = ACTIONS(2820),
    [anon_sym___attribute__] = ACTIONS(2820),
    [anon_sym___attribute] = ACTIONS(2820),
    [anon_sym_noreturn] = ACTIONS(2820),
    [anon_sym_LBRACK] = ACTIONS(2822),
    [anon_sym___declspec] = ACTIONS(2820),
    [anon_sym___cdecl] = ACTIONS(2820),
    [anon_sym___clrcall] = ACTIONS(2820),
    [anon_sym___stdcall] = ACTIONS(2820),
    [anon_sym___fastcall] = ACTIONS(2820),
    [anon_sym___thiscall] = ACTIONS(2820),
    [anon_sym___vectorcall] = ACTIONS(2820),
    [anon_sym_LBRACE] = ACTIONS(2822),
    [anon_sym_signed] = ACTIONS(2820),
    [anon_sym_unsigned] = ACTIONS(2820),
    [anon_sym_long] = ACTIONS(2820),
    [anon_sym_short] = ACTIONS(2820),
    [anon_sym_static] = ACTIONS(2820),
    [anon_sym_ATautoreleasepool] = ACTIONS(2822),
    [anon_sym_auto] = ACTIONS(2820),
    [anon_sym_register] = ACTIONS(2820),
    [anon_sym_inline] = ACTIONS(2820),
    [anon_sym___inline] = ACTIONS(2820),
    [anon_sym___inline__] = ACTIONS(2820),
    [anon_sym___forceinline] = ACTIONS(2820),
    [anon_sym_thread_local] = ACTIONS(2820),
    [anon_sym___thread] = ACTIONS(2820),
    [anon_sym_CG_EXTERN] = ACTIONS(2820),
    [anon_sym_CG_INLINE] = ACTIONS(2820),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2820),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2820),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2820),
    [anon_sym_IBOutlet] = ACTIONS(2820),
    [anon_sym_IBInspectable] = ACTIONS(2820),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2820),
    [anon_sym_NS_INLINE] = ACTIONS(2820),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2820),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2820),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2820),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2820),
    [anon_sym_const] = ACTIONS(2820),
    [anon_sym_constexpr] = ACTIONS(2820),
    [anon_sym_volatile] = ACTIONS(2820),
    [anon_sym_restrict] = ACTIONS(2820),
    [anon_sym___restrict__] = ACTIONS(2820),
    [anon_sym__Atomic] = ACTIONS(2820),
    [anon_sym__Noreturn] = ACTIONS(2820),
    [anon_sym__Nonnull] = ACTIONS(2820),
    [anon_sym_nullable] = ACTIONS(2820),
    [anon_sym__Complex] = ACTIONS(2820),
    [anon_sym__Nullable] = ACTIONS(2820),
    [anon_sym__Nullable_result] = ACTIONS(2820),
    [anon_sym__Null_unspecified] = ACTIONS(2820),
    [anon_sym___autoreleasing] = ACTIONS(2820),
    [anon_sym___block] = ACTIONS(2820),
    [anon_sym___bridge] = ACTIONS(2820),
    [anon_sym___bridge_retained] = ACTIONS(2820),
    [anon_sym___bridge_transfer] = ACTIONS(2820),
    [anon_sym___complex] = ACTIONS(2820),
    [anon_sym___const] = ACTIONS(2820),
    [anon_sym___imag] = ACTIONS(2820),
    [anon_sym___kindof] = ACTIONS(2820),
    [anon_sym___nonnull] = ACTIONS(2820),
    [anon_sym___nullable] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2820),
    [anon_sym___real] = ACTIONS(2820),
    [anon_sym___strong] = ACTIONS(2820),
    [anon_sym___unsafe_unretained] = ACTIONS(2820),
    [anon_sym___unused] = ACTIONS(2820),
    [anon_sym___weak] = ACTIONS(2820),
    [anon_sym_alignas] = ACTIONS(2820),
    [anon_sym__Alignas] = ACTIONS(2820),
    [sym_primitive_type] = ACTIONS(2820),
    [anon_sym_enum] = ACTIONS(2820),
    [anon_sym_struct] = ACTIONS(2820),
    [anon_sym_union] = ACTIONS(2820),
    [anon_sym_if] = ACTIONS(2820),
    [anon_sym_switch] = ACTIONS(2820),
    [anon_sym_case] = ACTIONS(2820),
    [anon_sym_default] = ACTIONS(2820),
    [anon_sym_while] = ACTIONS(2820),
    [anon_sym_do] = ACTIONS(2820),
    [anon_sym_for] = ACTIONS(2820),
    [anon_sym_in] = ACTIONS(2820),
    [anon_sym_return] = ACTIONS(2820),
    [anon_sym_break] = ACTIONS(2820),
    [anon_sym_continue] = ACTIONS(2820),
    [anon_sym_goto] = ACTIONS(2820),
    [anon_sym_DASH_DASH] = ACTIONS(2822),
    [anon_sym_PLUS_PLUS] = ACTIONS(2822),
    [anon_sym_sizeof] = ACTIONS(2820),
    [anon_sym___alignof__] = ACTIONS(2820),
    [anon_sym___alignof] = ACTIONS(2820),
    [anon_sym__alignof] = ACTIONS(2820),
    [anon_sym_alignof] = ACTIONS(2820),
    [anon_sym__Alignof] = ACTIONS(2820),
    [anon_sym_offsetof] = ACTIONS(2820),
    [anon_sym__Generic] = ACTIONS(2820),
    [anon_sym_asm] = ACTIONS(2820),
    [anon_sym___asm__] = ACTIONS(2820),
    [anon_sym___asm] = ACTIONS(2820),
    [sym_number_literal] = ACTIONS(2822),
    [anon_sym_L_SQUOTE] = ACTIONS(2822),
    [anon_sym_u_SQUOTE] = ACTIONS(2822),
    [anon_sym_U_SQUOTE] = ACTIONS(2822),
    [anon_sym_u8_SQUOTE] = ACTIONS(2822),
    [anon_sym_SQUOTE] = ACTIONS(2822),
    [anon_sym_AT] = ACTIONS(2820),
    [anon_sym_DQUOTE] = ACTIONS(2822),
    [anon_sym_L_DQUOTE] = ACTIONS(2822),
    [anon_sym_u_DQUOTE] = ACTIONS(2822),
    [anon_sym_U_DQUOTE] = ACTIONS(2822),
    [anon_sym_u8_DQUOTE] = ACTIONS(2822),
    [sym_true] = ACTIONS(2820),
    [sym_false] = ACTIONS(2820),
    [anon_sym_NULL] = ACTIONS(2820),
    [anon_sym_nullptr] = ACTIONS(2820),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2820),
    [anon_sym___typeof] = ACTIONS(2820),
    [anon_sym_typeof] = ACTIONS(2820),
    [anon_sym_ATimport] = ACTIONS(2822),
    [aux_sym_preproc_undef_token1] = ACTIONS(2820),
    [anon_sym_POUND] = ACTIONS(2820),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2820),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2820),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2820),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2820),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2820),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2820),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2820),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_API_AVAILABLE] = ACTIONS(2820),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_API_DEPRECATED] = ACTIONS(2820),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2820),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym___deprecated_msg] = ACTIONS(2820),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2820),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2820),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2820),
    [anon_sym_ATprotocol] = ACTIONS(2822),
    [anon_sym_ATinterface] = ACTIONS(2822),
    [anon_sym_ATimplementation] = ACTIONS(2822),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2822),
    [anon_sym_ATsynthesize] = ACTIONS(2822),
    [anon_sym_ATdynamic] = ACTIONS(2822),
    [anon_sym_ATtry] = ACTIONS(2822),
    [anon_sym___try] = ACTIONS(2820),
    [anon_sym_ATthrow] = ACTIONS(2822),
    [anon_sym_ATselector] = ACTIONS(2822),
    [anon_sym_ATavailable] = ACTIONS(2822),
    [anon_sym___builtin_available] = ACTIONS(2820),
    [anon_sym_va_arg] = ACTIONS(2820),
    [anon_sym_ATencode] = ACTIONS(2822),
    [anon_sym_ATsynchronized] = ACTIONS(2822),
    [anon_sym_BOOL] = ACTIONS(2820),
    [anon_sym_IMP] = ACTIONS(2820),
    [anon_sym_SEL] = ACTIONS(2820),
    [anon_sym_Class] = ACTIONS(2820),
    [anon_sym_id] = ACTIONS(2820),
  },
  [STATE(467)] = {
    [sym_identifier] = ACTIONS(2864),
    [aux_sym_preproc_include_token1] = ACTIONS(2864),
    [aux_sym_preproc_include_token2] = ACTIONS(2864),
    [aux_sym_preproc_def_token1] = ACTIONS(2864),
    [aux_sym_preproc_if_token1] = ACTIONS(2864),
    [aux_sym_preproc_if_token2] = ACTIONS(2864),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2864),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2864),
    [aux_sym_preproc_else_token1] = ACTIONS(2864),
    [aux_sym_preproc_elif_token1] = ACTIONS(2864),
    [sym_preproc_directive] = ACTIONS(2864),
    [anon_sym_LPAREN2] = ACTIONS(2866),
    [anon_sym_BANG] = ACTIONS(2866),
    [anon_sym_TILDE] = ACTIONS(2866),
    [anon_sym_DASH] = ACTIONS(2864),
    [anon_sym_PLUS] = ACTIONS(2864),
    [anon_sym_STAR] = ACTIONS(2866),
    [anon_sym_CARET] = ACTIONS(2866),
    [anon_sym_AMP] = ACTIONS(2866),
    [anon_sym_SEMI] = ACTIONS(2866),
    [anon_sym___extension__] = ACTIONS(2864),
    [anon_sym_typedef] = ACTIONS(2864),
    [anon_sym_extern] = ACTIONS(2864),
    [anon_sym___attribute__] = ACTIONS(2864),
    [anon_sym___attribute] = ACTIONS(2864),
    [anon_sym_noreturn] = ACTIONS(2864),
    [anon_sym_LBRACK] = ACTIONS(2866),
    [anon_sym___declspec] = ACTIONS(2864),
    [anon_sym___cdecl] = ACTIONS(2864),
    [anon_sym___clrcall] = ACTIONS(2864),
    [anon_sym___stdcall] = ACTIONS(2864),
    [anon_sym___fastcall] = ACTIONS(2864),
    [anon_sym___thiscall] = ACTIONS(2864),
    [anon_sym___vectorcall] = ACTIONS(2864),
    [anon_sym_LBRACE] = ACTIONS(2866),
    [anon_sym_signed] = ACTIONS(2864),
    [anon_sym_unsigned] = ACTIONS(2864),
    [anon_sym_long] = ACTIONS(2864),
    [anon_sym_short] = ACTIONS(2864),
    [anon_sym_static] = ACTIONS(2864),
    [anon_sym_ATautoreleasepool] = ACTIONS(2866),
    [anon_sym_auto] = ACTIONS(2864),
    [anon_sym_register] = ACTIONS(2864),
    [anon_sym_inline] = ACTIONS(2864),
    [anon_sym___inline] = ACTIONS(2864),
    [anon_sym___inline__] = ACTIONS(2864),
    [anon_sym___forceinline] = ACTIONS(2864),
    [anon_sym_thread_local] = ACTIONS(2864),
    [anon_sym___thread] = ACTIONS(2864),
    [anon_sym_CG_EXTERN] = ACTIONS(2864),
    [anon_sym_CG_INLINE] = ACTIONS(2864),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2864),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2864),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2864),
    [anon_sym_IBOutlet] = ACTIONS(2864),
    [anon_sym_IBInspectable] = ACTIONS(2864),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2864),
    [anon_sym_NS_INLINE] = ACTIONS(2864),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2864),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2864),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2864),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2864),
    [anon_sym_const] = ACTIONS(2864),
    [anon_sym_constexpr] = ACTIONS(2864),
    [anon_sym_volatile] = ACTIONS(2864),
    [anon_sym_restrict] = ACTIONS(2864),
    [anon_sym___restrict__] = ACTIONS(2864),
    [anon_sym__Atomic] = ACTIONS(2864),
    [anon_sym__Noreturn] = ACTIONS(2864),
    [anon_sym__Nonnull] = ACTIONS(2864),
    [anon_sym_nullable] = ACTIONS(2864),
    [anon_sym__Complex] = ACTIONS(2864),
    [anon_sym__Nullable] = ACTIONS(2864),
    [anon_sym__Nullable_result] = ACTIONS(2864),
    [anon_sym__Null_unspecified] = ACTIONS(2864),
    [anon_sym___autoreleasing] = ACTIONS(2864),
    [anon_sym___block] = ACTIONS(2864),
    [anon_sym___bridge] = ACTIONS(2864),
    [anon_sym___bridge_retained] = ACTIONS(2864),
    [anon_sym___bridge_transfer] = ACTIONS(2864),
    [anon_sym___complex] = ACTIONS(2864),
    [anon_sym___const] = ACTIONS(2864),
    [anon_sym___imag] = ACTIONS(2864),
    [anon_sym___kindof] = ACTIONS(2864),
    [anon_sym___nonnull] = ACTIONS(2864),
    [anon_sym___nullable] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2864),
    [anon_sym___real] = ACTIONS(2864),
    [anon_sym___strong] = ACTIONS(2864),
    [anon_sym___unsafe_unretained] = ACTIONS(2864),
    [anon_sym___unused] = ACTIONS(2864),
    [anon_sym___weak] = ACTIONS(2864),
    [anon_sym_alignas] = ACTIONS(2864),
    [anon_sym__Alignas] = ACTIONS(2864),
    [sym_primitive_type] = ACTIONS(2864),
    [anon_sym_enum] = ACTIONS(2864),
    [anon_sym_struct] = ACTIONS(2864),
    [anon_sym_union] = ACTIONS(2864),
    [anon_sym_if] = ACTIONS(2864),
    [anon_sym_switch] = ACTIONS(2864),
    [anon_sym_case] = ACTIONS(2864),
    [anon_sym_default] = ACTIONS(2864),
    [anon_sym_while] = ACTIONS(2864),
    [anon_sym_do] = ACTIONS(2864),
    [anon_sym_for] = ACTIONS(2864),
    [anon_sym_in] = ACTIONS(2864),
    [anon_sym_return] = ACTIONS(2864),
    [anon_sym_break] = ACTIONS(2864),
    [anon_sym_continue] = ACTIONS(2864),
    [anon_sym_goto] = ACTIONS(2864),
    [anon_sym_DASH_DASH] = ACTIONS(2866),
    [anon_sym_PLUS_PLUS] = ACTIONS(2866),
    [anon_sym_sizeof] = ACTIONS(2864),
    [anon_sym___alignof__] = ACTIONS(2864),
    [anon_sym___alignof] = ACTIONS(2864),
    [anon_sym__alignof] = ACTIONS(2864),
    [anon_sym_alignof] = ACTIONS(2864),
    [anon_sym__Alignof] = ACTIONS(2864),
    [anon_sym_offsetof] = ACTIONS(2864),
    [anon_sym__Generic] = ACTIONS(2864),
    [anon_sym_asm] = ACTIONS(2864),
    [anon_sym___asm__] = ACTIONS(2864),
    [anon_sym___asm] = ACTIONS(2864),
    [sym_number_literal] = ACTIONS(2866),
    [anon_sym_L_SQUOTE] = ACTIONS(2866),
    [anon_sym_u_SQUOTE] = ACTIONS(2866),
    [anon_sym_U_SQUOTE] = ACTIONS(2866),
    [anon_sym_u8_SQUOTE] = ACTIONS(2866),
    [anon_sym_SQUOTE] = ACTIONS(2866),
    [anon_sym_AT] = ACTIONS(2864),
    [anon_sym_DQUOTE] = ACTIONS(2866),
    [anon_sym_L_DQUOTE] = ACTIONS(2866),
    [anon_sym_u_DQUOTE] = ACTIONS(2866),
    [anon_sym_U_DQUOTE] = ACTIONS(2866),
    [anon_sym_u8_DQUOTE] = ACTIONS(2866),
    [sym_true] = ACTIONS(2864),
    [sym_false] = ACTIONS(2864),
    [anon_sym_NULL] = ACTIONS(2864),
    [anon_sym_nullptr] = ACTIONS(2864),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2864),
    [anon_sym___typeof] = ACTIONS(2864),
    [anon_sym_typeof] = ACTIONS(2864),
    [anon_sym_ATimport] = ACTIONS(2866),
    [aux_sym_preproc_undef_token1] = ACTIONS(2864),
    [anon_sym_POUND] = ACTIONS(2864),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2864),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2864),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2864),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2864),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2864),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2864),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2864),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_API_AVAILABLE] = ACTIONS(2864),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_API_DEPRECATED] = ACTIONS(2864),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2864),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym___deprecated_msg] = ACTIONS(2864),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2864),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2864),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2864),
    [anon_sym_ATprotocol] = ACTIONS(2866),
    [anon_sym_ATinterface] = ACTIONS(2866),
    [anon_sym_ATimplementation] = ACTIONS(2866),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2866),
    [anon_sym_ATsynthesize] = ACTIONS(2866),
    [anon_sym_ATdynamic] = ACTIONS(2866),
    [anon_sym_ATtry] = ACTIONS(2866),
    [anon_sym___try] = ACTIONS(2864),
    [anon_sym_ATthrow] = ACTIONS(2866),
    [anon_sym_ATselector] = ACTIONS(2866),
    [anon_sym_ATavailable] = ACTIONS(2866),
    [anon_sym___builtin_available] = ACTIONS(2864),
    [anon_sym_va_arg] = ACTIONS(2864),
    [anon_sym_ATencode] = ACTIONS(2866),
    [anon_sym_ATsynchronized] = ACTIONS(2866),
    [anon_sym_BOOL] = ACTIONS(2864),
    [anon_sym_IMP] = ACTIONS(2864),
    [anon_sym_SEL] = ACTIONS(2864),
    [anon_sym_Class] = ACTIONS(2864),
    [anon_sym_id] = ACTIONS(2864),
  },
  [STATE(468)] = {
    [sym_identifier] = ACTIONS(2692),
    [aux_sym_preproc_include_token1] = ACTIONS(2692),
    [aux_sym_preproc_include_token2] = ACTIONS(2692),
    [aux_sym_preproc_def_token1] = ACTIONS(2692),
    [aux_sym_preproc_if_token1] = ACTIONS(2692),
    [aux_sym_preproc_if_token2] = ACTIONS(2692),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2692),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2692),
    [aux_sym_preproc_else_token1] = ACTIONS(2692),
    [aux_sym_preproc_elif_token1] = ACTIONS(2692),
    [sym_preproc_directive] = ACTIONS(2692),
    [anon_sym_LPAREN2] = ACTIONS(2694),
    [anon_sym_BANG] = ACTIONS(2694),
    [anon_sym_TILDE] = ACTIONS(2694),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2694),
    [anon_sym_CARET] = ACTIONS(2694),
    [anon_sym_AMP] = ACTIONS(2694),
    [anon_sym_SEMI] = ACTIONS(2694),
    [anon_sym___extension__] = ACTIONS(2692),
    [anon_sym_typedef] = ACTIONS(2692),
    [anon_sym_extern] = ACTIONS(2692),
    [anon_sym___attribute__] = ACTIONS(2692),
    [anon_sym___attribute] = ACTIONS(2692),
    [anon_sym_noreturn] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym___declspec] = ACTIONS(2692),
    [anon_sym___cdecl] = ACTIONS(2692),
    [anon_sym___clrcall] = ACTIONS(2692),
    [anon_sym___stdcall] = ACTIONS(2692),
    [anon_sym___fastcall] = ACTIONS(2692),
    [anon_sym___thiscall] = ACTIONS(2692),
    [anon_sym___vectorcall] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_signed] = ACTIONS(2692),
    [anon_sym_unsigned] = ACTIONS(2692),
    [anon_sym_long] = ACTIONS(2692),
    [anon_sym_short] = ACTIONS(2692),
    [anon_sym_static] = ACTIONS(2692),
    [anon_sym_ATautoreleasepool] = ACTIONS(2694),
    [anon_sym_auto] = ACTIONS(2692),
    [anon_sym_register] = ACTIONS(2692),
    [anon_sym_inline] = ACTIONS(2692),
    [anon_sym___inline] = ACTIONS(2692),
    [anon_sym___inline__] = ACTIONS(2692),
    [anon_sym___forceinline] = ACTIONS(2692),
    [anon_sym_thread_local] = ACTIONS(2692),
    [anon_sym___thread] = ACTIONS(2692),
    [anon_sym_CG_EXTERN] = ACTIONS(2692),
    [anon_sym_CG_INLINE] = ACTIONS(2692),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2692),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2692),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2692),
    [anon_sym_IBOutlet] = ACTIONS(2692),
    [anon_sym_IBInspectable] = ACTIONS(2692),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2692),
    [anon_sym_NS_INLINE] = ACTIONS(2692),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2692),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2692),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2692),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2692),
    [anon_sym_const] = ACTIONS(2692),
    [anon_sym_constexpr] = ACTIONS(2692),
    [anon_sym_volatile] = ACTIONS(2692),
    [anon_sym_restrict] = ACTIONS(2692),
    [anon_sym___restrict__] = ACTIONS(2692),
    [anon_sym__Atomic] = ACTIONS(2692),
    [anon_sym__Noreturn] = ACTIONS(2692),
    [anon_sym__Nonnull] = ACTIONS(2692),
    [anon_sym_nullable] = ACTIONS(2692),
    [anon_sym__Complex] = ACTIONS(2692),
    [anon_sym__Nullable] = ACTIONS(2692),
    [anon_sym__Nullable_result] = ACTIONS(2692),
    [anon_sym__Null_unspecified] = ACTIONS(2692),
    [anon_sym___autoreleasing] = ACTIONS(2692),
    [anon_sym___block] = ACTIONS(2692),
    [anon_sym___bridge] = ACTIONS(2692),
    [anon_sym___bridge_retained] = ACTIONS(2692),
    [anon_sym___bridge_transfer] = ACTIONS(2692),
    [anon_sym___complex] = ACTIONS(2692),
    [anon_sym___const] = ACTIONS(2692),
    [anon_sym___imag] = ACTIONS(2692),
    [anon_sym___kindof] = ACTIONS(2692),
    [anon_sym___nonnull] = ACTIONS(2692),
    [anon_sym___nullable] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2692),
    [anon_sym___real] = ACTIONS(2692),
    [anon_sym___strong] = ACTIONS(2692),
    [anon_sym___unsafe_unretained] = ACTIONS(2692),
    [anon_sym___unused] = ACTIONS(2692),
    [anon_sym___weak] = ACTIONS(2692),
    [anon_sym_alignas] = ACTIONS(2692),
    [anon_sym__Alignas] = ACTIONS(2692),
    [sym_primitive_type] = ACTIONS(2692),
    [anon_sym_enum] = ACTIONS(2692),
    [anon_sym_struct] = ACTIONS(2692),
    [anon_sym_union] = ACTIONS(2692),
    [anon_sym_if] = ACTIONS(2692),
    [anon_sym_switch] = ACTIONS(2692),
    [anon_sym_case] = ACTIONS(2692),
    [anon_sym_default] = ACTIONS(2692),
    [anon_sym_while] = ACTIONS(2692),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_for] = ACTIONS(2692),
    [anon_sym_in] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_sizeof] = ACTIONS(2692),
    [anon_sym___alignof__] = ACTIONS(2692),
    [anon_sym___alignof] = ACTIONS(2692),
    [anon_sym__alignof] = ACTIONS(2692),
    [anon_sym_alignof] = ACTIONS(2692),
    [anon_sym__Alignof] = ACTIONS(2692),
    [anon_sym_offsetof] = ACTIONS(2692),
    [anon_sym__Generic] = ACTIONS(2692),
    [anon_sym_asm] = ACTIONS(2692),
    [anon_sym___asm__] = ACTIONS(2692),
    [anon_sym___asm] = ACTIONS(2692),
    [sym_number_literal] = ACTIONS(2694),
    [anon_sym_L_SQUOTE] = ACTIONS(2694),
    [anon_sym_u_SQUOTE] = ACTIONS(2694),
    [anon_sym_U_SQUOTE] = ACTIONS(2694),
    [anon_sym_u8_SQUOTE] = ACTIONS(2694),
    [anon_sym_SQUOTE] = ACTIONS(2694),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_DQUOTE] = ACTIONS(2694),
    [anon_sym_L_DQUOTE] = ACTIONS(2694),
    [anon_sym_u_DQUOTE] = ACTIONS(2694),
    [anon_sym_U_DQUOTE] = ACTIONS(2694),
    [anon_sym_u8_DQUOTE] = ACTIONS(2694),
    [sym_true] = ACTIONS(2692),
    [sym_false] = ACTIONS(2692),
    [anon_sym_NULL] = ACTIONS(2692),
    [anon_sym_nullptr] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2692),
    [anon_sym___typeof] = ACTIONS(2692),
    [anon_sym_typeof] = ACTIONS(2692),
    [anon_sym_ATimport] = ACTIONS(2694),
    [aux_sym_preproc_undef_token1] = ACTIONS(2692),
    [anon_sym_POUND] = ACTIONS(2692),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2692),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2692),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2692),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2692),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2692),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2692),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2692),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_API_AVAILABLE] = ACTIONS(2692),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_API_DEPRECATED] = ACTIONS(2692),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2692),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym___deprecated_msg] = ACTIONS(2692),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2692),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2692),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2692),
    [anon_sym_ATprotocol] = ACTIONS(2694),
    [anon_sym_ATinterface] = ACTIONS(2694),
    [anon_sym_ATimplementation] = ACTIONS(2694),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2694),
    [anon_sym_ATsynthesize] = ACTIONS(2694),
    [anon_sym_ATdynamic] = ACTIONS(2694),
    [anon_sym_ATtry] = ACTIONS(2694),
    [anon_sym___try] = ACTIONS(2692),
    [anon_sym_ATthrow] = ACTIONS(2694),
    [anon_sym_ATselector] = ACTIONS(2694),
    [anon_sym_ATavailable] = ACTIONS(2694),
    [anon_sym___builtin_available] = ACTIONS(2692),
    [anon_sym_va_arg] = ACTIONS(2692),
    [anon_sym_ATencode] = ACTIONS(2694),
    [anon_sym_ATsynchronized] = ACTIONS(2694),
    [anon_sym_BOOL] = ACTIONS(2692),
    [anon_sym_IMP] = ACTIONS(2692),
    [anon_sym_SEL] = ACTIONS(2692),
    [anon_sym_Class] = ACTIONS(2692),
    [anon_sym_id] = ACTIONS(2692),
  },
  [STATE(469)] = {
    [sym_identifier] = ACTIONS(2650),
    [aux_sym_preproc_include_token1] = ACTIONS(2650),
    [aux_sym_preproc_include_token2] = ACTIONS(2650),
    [aux_sym_preproc_def_token1] = ACTIONS(2650),
    [aux_sym_preproc_if_token1] = ACTIONS(2650),
    [aux_sym_preproc_if_token2] = ACTIONS(2650),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2650),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2650),
    [aux_sym_preproc_else_token1] = ACTIONS(2650),
    [aux_sym_preproc_elif_token1] = ACTIONS(2650),
    [sym_preproc_directive] = ACTIONS(2650),
    [anon_sym_LPAREN2] = ACTIONS(2652),
    [anon_sym_BANG] = ACTIONS(2652),
    [anon_sym_TILDE] = ACTIONS(2652),
    [anon_sym_DASH] = ACTIONS(2650),
    [anon_sym_PLUS] = ACTIONS(2650),
    [anon_sym_STAR] = ACTIONS(2652),
    [anon_sym_CARET] = ACTIONS(2652),
    [anon_sym_AMP] = ACTIONS(2652),
    [anon_sym_SEMI] = ACTIONS(2652),
    [anon_sym___extension__] = ACTIONS(2650),
    [anon_sym_typedef] = ACTIONS(2650),
    [anon_sym_extern] = ACTIONS(2650),
    [anon_sym___attribute__] = ACTIONS(2650),
    [anon_sym___attribute] = ACTIONS(2650),
    [anon_sym_noreturn] = ACTIONS(2650),
    [anon_sym_LBRACK] = ACTIONS(2652),
    [anon_sym___declspec] = ACTIONS(2650),
    [anon_sym___cdecl] = ACTIONS(2650),
    [anon_sym___clrcall] = ACTIONS(2650),
    [anon_sym___stdcall] = ACTIONS(2650),
    [anon_sym___fastcall] = ACTIONS(2650),
    [anon_sym___thiscall] = ACTIONS(2650),
    [anon_sym___vectorcall] = ACTIONS(2650),
    [anon_sym_LBRACE] = ACTIONS(2652),
    [anon_sym_signed] = ACTIONS(2650),
    [anon_sym_unsigned] = ACTIONS(2650),
    [anon_sym_long] = ACTIONS(2650),
    [anon_sym_short] = ACTIONS(2650),
    [anon_sym_static] = ACTIONS(2650),
    [anon_sym_ATautoreleasepool] = ACTIONS(2652),
    [anon_sym_auto] = ACTIONS(2650),
    [anon_sym_register] = ACTIONS(2650),
    [anon_sym_inline] = ACTIONS(2650),
    [anon_sym___inline] = ACTIONS(2650),
    [anon_sym___inline__] = ACTIONS(2650),
    [anon_sym___forceinline] = ACTIONS(2650),
    [anon_sym_thread_local] = ACTIONS(2650),
    [anon_sym___thread] = ACTIONS(2650),
    [anon_sym_CG_EXTERN] = ACTIONS(2650),
    [anon_sym_CG_INLINE] = ACTIONS(2650),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2650),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2650),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2650),
    [anon_sym_IBOutlet] = ACTIONS(2650),
    [anon_sym_IBInspectable] = ACTIONS(2650),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2650),
    [anon_sym_NS_INLINE] = ACTIONS(2650),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2650),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2650),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2650),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2650),
    [anon_sym_const] = ACTIONS(2650),
    [anon_sym_constexpr] = ACTIONS(2650),
    [anon_sym_volatile] = ACTIONS(2650),
    [anon_sym_restrict] = ACTIONS(2650),
    [anon_sym___restrict__] = ACTIONS(2650),
    [anon_sym__Atomic] = ACTIONS(2650),
    [anon_sym__Noreturn] = ACTIONS(2650),
    [anon_sym__Nonnull] = ACTIONS(2650),
    [anon_sym_nullable] = ACTIONS(2650),
    [anon_sym__Complex] = ACTIONS(2650),
    [anon_sym__Nullable] = ACTIONS(2650),
    [anon_sym__Nullable_result] = ACTIONS(2650),
    [anon_sym__Null_unspecified] = ACTIONS(2650),
    [anon_sym___autoreleasing] = ACTIONS(2650),
    [anon_sym___block] = ACTIONS(2650),
    [anon_sym___bridge] = ACTIONS(2650),
    [anon_sym___bridge_retained] = ACTIONS(2650),
    [anon_sym___bridge_transfer] = ACTIONS(2650),
    [anon_sym___complex] = ACTIONS(2650),
    [anon_sym___const] = ACTIONS(2650),
    [anon_sym___imag] = ACTIONS(2650),
    [anon_sym___kindof] = ACTIONS(2650),
    [anon_sym___nonnull] = ACTIONS(2650),
    [anon_sym___nullable] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2650),
    [anon_sym___real] = ACTIONS(2650),
    [anon_sym___strong] = ACTIONS(2650),
    [anon_sym___unsafe_unretained] = ACTIONS(2650),
    [anon_sym___unused] = ACTIONS(2650),
    [anon_sym___weak] = ACTIONS(2650),
    [anon_sym_alignas] = ACTIONS(2650),
    [anon_sym__Alignas] = ACTIONS(2650),
    [sym_primitive_type] = ACTIONS(2650),
    [anon_sym_enum] = ACTIONS(2650),
    [anon_sym_struct] = ACTIONS(2650),
    [anon_sym_union] = ACTIONS(2650),
    [anon_sym_if] = ACTIONS(2650),
    [anon_sym_switch] = ACTIONS(2650),
    [anon_sym_case] = ACTIONS(2650),
    [anon_sym_default] = ACTIONS(2650),
    [anon_sym_while] = ACTIONS(2650),
    [anon_sym_do] = ACTIONS(2650),
    [anon_sym_for] = ACTIONS(2650),
    [anon_sym_in] = ACTIONS(2650),
    [anon_sym_return] = ACTIONS(2650),
    [anon_sym_break] = ACTIONS(2650),
    [anon_sym_continue] = ACTIONS(2650),
    [anon_sym_goto] = ACTIONS(2650),
    [anon_sym_DASH_DASH] = ACTIONS(2652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2652),
    [anon_sym_sizeof] = ACTIONS(2650),
    [anon_sym___alignof__] = ACTIONS(2650),
    [anon_sym___alignof] = ACTIONS(2650),
    [anon_sym__alignof] = ACTIONS(2650),
    [anon_sym_alignof] = ACTIONS(2650),
    [anon_sym__Alignof] = ACTIONS(2650),
    [anon_sym_offsetof] = ACTIONS(2650),
    [anon_sym__Generic] = ACTIONS(2650),
    [anon_sym_asm] = ACTIONS(2650),
    [anon_sym___asm__] = ACTIONS(2650),
    [anon_sym___asm] = ACTIONS(2650),
    [sym_number_literal] = ACTIONS(2652),
    [anon_sym_L_SQUOTE] = ACTIONS(2652),
    [anon_sym_u_SQUOTE] = ACTIONS(2652),
    [anon_sym_U_SQUOTE] = ACTIONS(2652),
    [anon_sym_u8_SQUOTE] = ACTIONS(2652),
    [anon_sym_SQUOTE] = ACTIONS(2652),
    [anon_sym_AT] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_L_DQUOTE] = ACTIONS(2652),
    [anon_sym_u_DQUOTE] = ACTIONS(2652),
    [anon_sym_U_DQUOTE] = ACTIONS(2652),
    [anon_sym_u8_DQUOTE] = ACTIONS(2652),
    [sym_true] = ACTIONS(2650),
    [sym_false] = ACTIONS(2650),
    [anon_sym_NULL] = ACTIONS(2650),
    [anon_sym_nullptr] = ACTIONS(2650),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2650),
    [anon_sym___typeof] = ACTIONS(2650),
    [anon_sym_typeof] = ACTIONS(2650),
    [anon_sym_ATimport] = ACTIONS(2652),
    [aux_sym_preproc_undef_token1] = ACTIONS(2650),
    [anon_sym_POUND] = ACTIONS(2650),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2650),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2650),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2650),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2650),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2650),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2650),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2650),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_API_AVAILABLE] = ACTIONS(2650),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_API_DEPRECATED] = ACTIONS(2650),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2650),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym___deprecated_msg] = ACTIONS(2650),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2650),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2650),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2650),
    [anon_sym_ATprotocol] = ACTIONS(2652),
    [anon_sym_ATinterface] = ACTIONS(2652),
    [anon_sym_ATimplementation] = ACTIONS(2652),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2652),
    [anon_sym_ATsynthesize] = ACTIONS(2652),
    [anon_sym_ATdynamic] = ACTIONS(2652),
    [anon_sym_ATtry] = ACTIONS(2652),
    [anon_sym___try] = ACTIONS(2650),
    [anon_sym_ATthrow] = ACTIONS(2652),
    [anon_sym_ATselector] = ACTIONS(2652),
    [anon_sym_ATavailable] = ACTIONS(2652),
    [anon_sym___builtin_available] = ACTIONS(2650),
    [anon_sym_va_arg] = ACTIONS(2650),
    [anon_sym_ATencode] = ACTIONS(2652),
    [anon_sym_ATsynchronized] = ACTIONS(2652),
    [anon_sym_BOOL] = ACTIONS(2650),
    [anon_sym_IMP] = ACTIONS(2650),
    [anon_sym_SEL] = ACTIONS(2650),
    [anon_sym_Class] = ACTIONS(2650),
    [anon_sym_id] = ACTIONS(2650),
  },
  [STATE(470)] = {
    [sym_identifier] = ACTIONS(2828),
    [aux_sym_preproc_include_token1] = ACTIONS(2828),
    [aux_sym_preproc_include_token2] = ACTIONS(2828),
    [aux_sym_preproc_def_token1] = ACTIONS(2828),
    [aux_sym_preproc_if_token1] = ACTIONS(2828),
    [aux_sym_preproc_if_token2] = ACTIONS(2828),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2828),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2828),
    [aux_sym_preproc_else_token1] = ACTIONS(2828),
    [aux_sym_preproc_elif_token1] = ACTIONS(2828),
    [sym_preproc_directive] = ACTIONS(2828),
    [anon_sym_LPAREN2] = ACTIONS(2830),
    [anon_sym_BANG] = ACTIONS(2830),
    [anon_sym_TILDE] = ACTIONS(2830),
    [anon_sym_DASH] = ACTIONS(2828),
    [anon_sym_PLUS] = ACTIONS(2828),
    [anon_sym_STAR] = ACTIONS(2830),
    [anon_sym_CARET] = ACTIONS(2830),
    [anon_sym_AMP] = ACTIONS(2830),
    [anon_sym_SEMI] = ACTIONS(2830),
    [anon_sym___extension__] = ACTIONS(2828),
    [anon_sym_typedef] = ACTIONS(2828),
    [anon_sym_extern] = ACTIONS(2828),
    [anon_sym___attribute__] = ACTIONS(2828),
    [anon_sym___attribute] = ACTIONS(2828),
    [anon_sym_noreturn] = ACTIONS(2828),
    [anon_sym_LBRACK] = ACTIONS(2830),
    [anon_sym___declspec] = ACTIONS(2828),
    [anon_sym___cdecl] = ACTIONS(2828),
    [anon_sym___clrcall] = ACTIONS(2828),
    [anon_sym___stdcall] = ACTIONS(2828),
    [anon_sym___fastcall] = ACTIONS(2828),
    [anon_sym___thiscall] = ACTIONS(2828),
    [anon_sym___vectorcall] = ACTIONS(2828),
    [anon_sym_LBRACE] = ACTIONS(2830),
    [anon_sym_signed] = ACTIONS(2828),
    [anon_sym_unsigned] = ACTIONS(2828),
    [anon_sym_long] = ACTIONS(2828),
    [anon_sym_short] = ACTIONS(2828),
    [anon_sym_static] = ACTIONS(2828),
    [anon_sym_ATautoreleasepool] = ACTIONS(2830),
    [anon_sym_auto] = ACTIONS(2828),
    [anon_sym_register] = ACTIONS(2828),
    [anon_sym_inline] = ACTIONS(2828),
    [anon_sym___inline] = ACTIONS(2828),
    [anon_sym___inline__] = ACTIONS(2828),
    [anon_sym___forceinline] = ACTIONS(2828),
    [anon_sym_thread_local] = ACTIONS(2828),
    [anon_sym___thread] = ACTIONS(2828),
    [anon_sym_CG_EXTERN] = ACTIONS(2828),
    [anon_sym_CG_INLINE] = ACTIONS(2828),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2828),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2828),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2828),
    [anon_sym_IBOutlet] = ACTIONS(2828),
    [anon_sym_IBInspectable] = ACTIONS(2828),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2828),
    [anon_sym_NS_INLINE] = ACTIONS(2828),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2828),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2828),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2828),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2828),
    [anon_sym_const] = ACTIONS(2828),
    [anon_sym_constexpr] = ACTIONS(2828),
    [anon_sym_volatile] = ACTIONS(2828),
    [anon_sym_restrict] = ACTIONS(2828),
    [anon_sym___restrict__] = ACTIONS(2828),
    [anon_sym__Atomic] = ACTIONS(2828),
    [anon_sym__Noreturn] = ACTIONS(2828),
    [anon_sym__Nonnull] = ACTIONS(2828),
    [anon_sym_nullable] = ACTIONS(2828),
    [anon_sym__Complex] = ACTIONS(2828),
    [anon_sym__Nullable] = ACTIONS(2828),
    [anon_sym__Nullable_result] = ACTIONS(2828),
    [anon_sym__Null_unspecified] = ACTIONS(2828),
    [anon_sym___autoreleasing] = ACTIONS(2828),
    [anon_sym___block] = ACTIONS(2828),
    [anon_sym___bridge] = ACTIONS(2828),
    [anon_sym___bridge_retained] = ACTIONS(2828),
    [anon_sym___bridge_transfer] = ACTIONS(2828),
    [anon_sym___complex] = ACTIONS(2828),
    [anon_sym___const] = ACTIONS(2828),
    [anon_sym___imag] = ACTIONS(2828),
    [anon_sym___kindof] = ACTIONS(2828),
    [anon_sym___nonnull] = ACTIONS(2828),
    [anon_sym___nullable] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2828),
    [anon_sym___real] = ACTIONS(2828),
    [anon_sym___strong] = ACTIONS(2828),
    [anon_sym___unsafe_unretained] = ACTIONS(2828),
    [anon_sym___unused] = ACTIONS(2828),
    [anon_sym___weak] = ACTIONS(2828),
    [anon_sym_alignas] = ACTIONS(2828),
    [anon_sym__Alignas] = ACTIONS(2828),
    [sym_primitive_type] = ACTIONS(2828),
    [anon_sym_enum] = ACTIONS(2828),
    [anon_sym_struct] = ACTIONS(2828),
    [anon_sym_union] = ACTIONS(2828),
    [anon_sym_if] = ACTIONS(2828),
    [anon_sym_switch] = ACTIONS(2828),
    [anon_sym_case] = ACTIONS(2828),
    [anon_sym_default] = ACTIONS(2828),
    [anon_sym_while] = ACTIONS(2828),
    [anon_sym_do] = ACTIONS(2828),
    [anon_sym_for] = ACTIONS(2828),
    [anon_sym_in] = ACTIONS(2828),
    [anon_sym_return] = ACTIONS(2828),
    [anon_sym_break] = ACTIONS(2828),
    [anon_sym_continue] = ACTIONS(2828),
    [anon_sym_goto] = ACTIONS(2828),
    [anon_sym_DASH_DASH] = ACTIONS(2830),
    [anon_sym_PLUS_PLUS] = ACTIONS(2830),
    [anon_sym_sizeof] = ACTIONS(2828),
    [anon_sym___alignof__] = ACTIONS(2828),
    [anon_sym___alignof] = ACTIONS(2828),
    [anon_sym__alignof] = ACTIONS(2828),
    [anon_sym_alignof] = ACTIONS(2828),
    [anon_sym__Alignof] = ACTIONS(2828),
    [anon_sym_offsetof] = ACTIONS(2828),
    [anon_sym__Generic] = ACTIONS(2828),
    [anon_sym_asm] = ACTIONS(2828),
    [anon_sym___asm__] = ACTIONS(2828),
    [anon_sym___asm] = ACTIONS(2828),
    [sym_number_literal] = ACTIONS(2830),
    [anon_sym_L_SQUOTE] = ACTIONS(2830),
    [anon_sym_u_SQUOTE] = ACTIONS(2830),
    [anon_sym_U_SQUOTE] = ACTIONS(2830),
    [anon_sym_u8_SQUOTE] = ACTIONS(2830),
    [anon_sym_SQUOTE] = ACTIONS(2830),
    [anon_sym_AT] = ACTIONS(2828),
    [anon_sym_DQUOTE] = ACTIONS(2830),
    [anon_sym_L_DQUOTE] = ACTIONS(2830),
    [anon_sym_u_DQUOTE] = ACTIONS(2830),
    [anon_sym_U_DQUOTE] = ACTIONS(2830),
    [anon_sym_u8_DQUOTE] = ACTIONS(2830),
    [sym_true] = ACTIONS(2828),
    [sym_false] = ACTIONS(2828),
    [anon_sym_NULL] = ACTIONS(2828),
    [anon_sym_nullptr] = ACTIONS(2828),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2828),
    [anon_sym___typeof] = ACTIONS(2828),
    [anon_sym_typeof] = ACTIONS(2828),
    [anon_sym_ATimport] = ACTIONS(2830),
    [aux_sym_preproc_undef_token1] = ACTIONS(2828),
    [anon_sym_POUND] = ACTIONS(2828),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2828),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2828),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2828),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2828),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2828),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2828),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2828),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_API_AVAILABLE] = ACTIONS(2828),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_API_DEPRECATED] = ACTIONS(2828),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2828),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym___deprecated_msg] = ACTIONS(2828),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2828),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2828),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2828),
    [anon_sym_ATprotocol] = ACTIONS(2830),
    [anon_sym_ATinterface] = ACTIONS(2830),
    [anon_sym_ATimplementation] = ACTIONS(2830),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2830),
    [anon_sym_ATsynthesize] = ACTIONS(2830),
    [anon_sym_ATdynamic] = ACTIONS(2830),
    [anon_sym_ATtry] = ACTIONS(2830),
    [anon_sym___try] = ACTIONS(2828),
    [anon_sym_ATthrow] = ACTIONS(2830),
    [anon_sym_ATselector] = ACTIONS(2830),
    [anon_sym_ATavailable] = ACTIONS(2830),
    [anon_sym___builtin_available] = ACTIONS(2828),
    [anon_sym_va_arg] = ACTIONS(2828),
    [anon_sym_ATencode] = ACTIONS(2830),
    [anon_sym_ATsynchronized] = ACTIONS(2830),
    [anon_sym_BOOL] = ACTIONS(2828),
    [anon_sym_IMP] = ACTIONS(2828),
    [anon_sym_SEL] = ACTIONS(2828),
    [anon_sym_Class] = ACTIONS(2828),
    [anon_sym_id] = ACTIONS(2828),
  },
  [STATE(471)] = {
    [sym_identifier] = ACTIONS(2658),
    [aux_sym_preproc_include_token1] = ACTIONS(2658),
    [aux_sym_preproc_include_token2] = ACTIONS(2658),
    [aux_sym_preproc_def_token1] = ACTIONS(2658),
    [aux_sym_preproc_if_token1] = ACTIONS(2658),
    [aux_sym_preproc_if_token2] = ACTIONS(2658),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2658),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2658),
    [aux_sym_preproc_else_token1] = ACTIONS(2658),
    [aux_sym_preproc_elif_token1] = ACTIONS(2658),
    [sym_preproc_directive] = ACTIONS(2658),
    [anon_sym_LPAREN2] = ACTIONS(2660),
    [anon_sym_BANG] = ACTIONS(2660),
    [anon_sym_TILDE] = ACTIONS(2660),
    [anon_sym_DASH] = ACTIONS(2658),
    [anon_sym_PLUS] = ACTIONS(2658),
    [anon_sym_STAR] = ACTIONS(2660),
    [anon_sym_CARET] = ACTIONS(2660),
    [anon_sym_AMP] = ACTIONS(2660),
    [anon_sym_SEMI] = ACTIONS(2660),
    [anon_sym___extension__] = ACTIONS(2658),
    [anon_sym_typedef] = ACTIONS(2658),
    [anon_sym_extern] = ACTIONS(2658),
    [anon_sym___attribute__] = ACTIONS(2658),
    [anon_sym___attribute] = ACTIONS(2658),
    [anon_sym_noreturn] = ACTIONS(2658),
    [anon_sym_LBRACK] = ACTIONS(2660),
    [anon_sym___declspec] = ACTIONS(2658),
    [anon_sym___cdecl] = ACTIONS(2658),
    [anon_sym___clrcall] = ACTIONS(2658),
    [anon_sym___stdcall] = ACTIONS(2658),
    [anon_sym___fastcall] = ACTIONS(2658),
    [anon_sym___thiscall] = ACTIONS(2658),
    [anon_sym___vectorcall] = ACTIONS(2658),
    [anon_sym_LBRACE] = ACTIONS(2660),
    [anon_sym_signed] = ACTIONS(2658),
    [anon_sym_unsigned] = ACTIONS(2658),
    [anon_sym_long] = ACTIONS(2658),
    [anon_sym_short] = ACTIONS(2658),
    [anon_sym_static] = ACTIONS(2658),
    [anon_sym_ATautoreleasepool] = ACTIONS(2660),
    [anon_sym_auto] = ACTIONS(2658),
    [anon_sym_register] = ACTIONS(2658),
    [anon_sym_inline] = ACTIONS(2658),
    [anon_sym___inline] = ACTIONS(2658),
    [anon_sym___inline__] = ACTIONS(2658),
    [anon_sym___forceinline] = ACTIONS(2658),
    [anon_sym_thread_local] = ACTIONS(2658),
    [anon_sym___thread] = ACTIONS(2658),
    [anon_sym_CG_EXTERN] = ACTIONS(2658),
    [anon_sym_CG_INLINE] = ACTIONS(2658),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2658),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2658),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2658),
    [anon_sym_IBOutlet] = ACTIONS(2658),
    [anon_sym_IBInspectable] = ACTIONS(2658),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2658),
    [anon_sym_NS_INLINE] = ACTIONS(2658),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2658),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2658),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2658),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2658),
    [anon_sym_const] = ACTIONS(2658),
    [anon_sym_constexpr] = ACTIONS(2658),
    [anon_sym_volatile] = ACTIONS(2658),
    [anon_sym_restrict] = ACTIONS(2658),
    [anon_sym___restrict__] = ACTIONS(2658),
    [anon_sym__Atomic] = ACTIONS(2658),
    [anon_sym__Noreturn] = ACTIONS(2658),
    [anon_sym__Nonnull] = ACTIONS(2658),
    [anon_sym_nullable] = ACTIONS(2658),
    [anon_sym__Complex] = ACTIONS(2658),
    [anon_sym__Nullable] = ACTIONS(2658),
    [anon_sym__Nullable_result] = ACTIONS(2658),
    [anon_sym__Null_unspecified] = ACTIONS(2658),
    [anon_sym___autoreleasing] = ACTIONS(2658),
    [anon_sym___block] = ACTIONS(2658),
    [anon_sym___bridge] = ACTIONS(2658),
    [anon_sym___bridge_retained] = ACTIONS(2658),
    [anon_sym___bridge_transfer] = ACTIONS(2658),
    [anon_sym___complex] = ACTIONS(2658),
    [anon_sym___const] = ACTIONS(2658),
    [anon_sym___imag] = ACTIONS(2658),
    [anon_sym___kindof] = ACTIONS(2658),
    [anon_sym___nonnull] = ACTIONS(2658),
    [anon_sym___nullable] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2658),
    [anon_sym___real] = ACTIONS(2658),
    [anon_sym___strong] = ACTIONS(2658),
    [anon_sym___unsafe_unretained] = ACTIONS(2658),
    [anon_sym___unused] = ACTIONS(2658),
    [anon_sym___weak] = ACTIONS(2658),
    [anon_sym_alignas] = ACTIONS(2658),
    [anon_sym__Alignas] = ACTIONS(2658),
    [sym_primitive_type] = ACTIONS(2658),
    [anon_sym_enum] = ACTIONS(2658),
    [anon_sym_struct] = ACTIONS(2658),
    [anon_sym_union] = ACTIONS(2658),
    [anon_sym_if] = ACTIONS(2658),
    [anon_sym_switch] = ACTIONS(2658),
    [anon_sym_case] = ACTIONS(2658),
    [anon_sym_default] = ACTIONS(2658),
    [anon_sym_while] = ACTIONS(2658),
    [anon_sym_do] = ACTIONS(2658),
    [anon_sym_for] = ACTIONS(2658),
    [anon_sym_in] = ACTIONS(2658),
    [anon_sym_return] = ACTIONS(2658),
    [anon_sym_break] = ACTIONS(2658),
    [anon_sym_continue] = ACTIONS(2658),
    [anon_sym_goto] = ACTIONS(2658),
    [anon_sym_DASH_DASH] = ACTIONS(2660),
    [anon_sym_PLUS_PLUS] = ACTIONS(2660),
    [anon_sym_sizeof] = ACTIONS(2658),
    [anon_sym___alignof__] = ACTIONS(2658),
    [anon_sym___alignof] = ACTIONS(2658),
    [anon_sym__alignof] = ACTIONS(2658),
    [anon_sym_alignof] = ACTIONS(2658),
    [anon_sym__Alignof] = ACTIONS(2658),
    [anon_sym_offsetof] = ACTIONS(2658),
    [anon_sym__Generic] = ACTIONS(2658),
    [anon_sym_asm] = ACTIONS(2658),
    [anon_sym___asm__] = ACTIONS(2658),
    [anon_sym___asm] = ACTIONS(2658),
    [sym_number_literal] = ACTIONS(2660),
    [anon_sym_L_SQUOTE] = ACTIONS(2660),
    [anon_sym_u_SQUOTE] = ACTIONS(2660),
    [anon_sym_U_SQUOTE] = ACTIONS(2660),
    [anon_sym_u8_SQUOTE] = ACTIONS(2660),
    [anon_sym_SQUOTE] = ACTIONS(2660),
    [anon_sym_AT] = ACTIONS(2658),
    [anon_sym_DQUOTE] = ACTIONS(2660),
    [anon_sym_L_DQUOTE] = ACTIONS(2660),
    [anon_sym_u_DQUOTE] = ACTIONS(2660),
    [anon_sym_U_DQUOTE] = ACTIONS(2660),
    [anon_sym_u8_DQUOTE] = ACTIONS(2660),
    [sym_true] = ACTIONS(2658),
    [sym_false] = ACTIONS(2658),
    [anon_sym_NULL] = ACTIONS(2658),
    [anon_sym_nullptr] = ACTIONS(2658),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2658),
    [anon_sym___typeof] = ACTIONS(2658),
    [anon_sym_typeof] = ACTIONS(2658),
    [anon_sym_ATimport] = ACTIONS(2660),
    [aux_sym_preproc_undef_token1] = ACTIONS(2658),
    [anon_sym_POUND] = ACTIONS(2658),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2658),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2658),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2658),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2658),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2658),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2658),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2658),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_API_AVAILABLE] = ACTIONS(2658),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_API_DEPRECATED] = ACTIONS(2658),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2658),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym___deprecated_msg] = ACTIONS(2658),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2658),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2658),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2658),
    [anon_sym_ATprotocol] = ACTIONS(2660),
    [anon_sym_ATinterface] = ACTIONS(2660),
    [anon_sym_ATimplementation] = ACTIONS(2660),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2660),
    [anon_sym_ATsynthesize] = ACTIONS(2660),
    [anon_sym_ATdynamic] = ACTIONS(2660),
    [anon_sym_ATtry] = ACTIONS(2660),
    [anon_sym___try] = ACTIONS(2658),
    [anon_sym_ATthrow] = ACTIONS(2660),
    [anon_sym_ATselector] = ACTIONS(2660),
    [anon_sym_ATavailable] = ACTIONS(2660),
    [anon_sym___builtin_available] = ACTIONS(2658),
    [anon_sym_va_arg] = ACTIONS(2658),
    [anon_sym_ATencode] = ACTIONS(2660),
    [anon_sym_ATsynchronized] = ACTIONS(2660),
    [anon_sym_BOOL] = ACTIONS(2658),
    [anon_sym_IMP] = ACTIONS(2658),
    [anon_sym_SEL] = ACTIONS(2658),
    [anon_sym_Class] = ACTIONS(2658),
    [anon_sym_id] = ACTIONS(2658),
  },
  [STATE(472)] = {
    [sym_identifier] = ACTIONS(2832),
    [aux_sym_preproc_include_token1] = ACTIONS(2832),
    [aux_sym_preproc_include_token2] = ACTIONS(2832),
    [aux_sym_preproc_def_token1] = ACTIONS(2832),
    [aux_sym_preproc_if_token1] = ACTIONS(2832),
    [aux_sym_preproc_if_token2] = ACTIONS(2832),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2832),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2832),
    [aux_sym_preproc_else_token1] = ACTIONS(2832),
    [aux_sym_preproc_elif_token1] = ACTIONS(2832),
    [sym_preproc_directive] = ACTIONS(2832),
    [anon_sym_LPAREN2] = ACTIONS(2834),
    [anon_sym_BANG] = ACTIONS(2834),
    [anon_sym_TILDE] = ACTIONS(2834),
    [anon_sym_DASH] = ACTIONS(2832),
    [anon_sym_PLUS] = ACTIONS(2832),
    [anon_sym_STAR] = ACTIONS(2834),
    [anon_sym_CARET] = ACTIONS(2834),
    [anon_sym_AMP] = ACTIONS(2834),
    [anon_sym_SEMI] = ACTIONS(2834),
    [anon_sym___extension__] = ACTIONS(2832),
    [anon_sym_typedef] = ACTIONS(2832),
    [anon_sym_extern] = ACTIONS(2832),
    [anon_sym___attribute__] = ACTIONS(2832),
    [anon_sym___attribute] = ACTIONS(2832),
    [anon_sym_noreturn] = ACTIONS(2832),
    [anon_sym_LBRACK] = ACTIONS(2834),
    [anon_sym___declspec] = ACTIONS(2832),
    [anon_sym___cdecl] = ACTIONS(2832),
    [anon_sym___clrcall] = ACTIONS(2832),
    [anon_sym___stdcall] = ACTIONS(2832),
    [anon_sym___fastcall] = ACTIONS(2832),
    [anon_sym___thiscall] = ACTIONS(2832),
    [anon_sym___vectorcall] = ACTIONS(2832),
    [anon_sym_LBRACE] = ACTIONS(2834),
    [anon_sym_signed] = ACTIONS(2832),
    [anon_sym_unsigned] = ACTIONS(2832),
    [anon_sym_long] = ACTIONS(2832),
    [anon_sym_short] = ACTIONS(2832),
    [anon_sym_static] = ACTIONS(2832),
    [anon_sym_ATautoreleasepool] = ACTIONS(2834),
    [anon_sym_auto] = ACTIONS(2832),
    [anon_sym_register] = ACTIONS(2832),
    [anon_sym_inline] = ACTIONS(2832),
    [anon_sym___inline] = ACTIONS(2832),
    [anon_sym___inline__] = ACTIONS(2832),
    [anon_sym___forceinline] = ACTIONS(2832),
    [anon_sym_thread_local] = ACTIONS(2832),
    [anon_sym___thread] = ACTIONS(2832),
    [anon_sym_CG_EXTERN] = ACTIONS(2832),
    [anon_sym_CG_INLINE] = ACTIONS(2832),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2832),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2832),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2832),
    [anon_sym_IBOutlet] = ACTIONS(2832),
    [anon_sym_IBInspectable] = ACTIONS(2832),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2832),
    [anon_sym_NS_INLINE] = ACTIONS(2832),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2832),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2832),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2832),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2832),
    [anon_sym_const] = ACTIONS(2832),
    [anon_sym_constexpr] = ACTIONS(2832),
    [anon_sym_volatile] = ACTIONS(2832),
    [anon_sym_restrict] = ACTIONS(2832),
    [anon_sym___restrict__] = ACTIONS(2832),
    [anon_sym__Atomic] = ACTIONS(2832),
    [anon_sym__Noreturn] = ACTIONS(2832),
    [anon_sym__Nonnull] = ACTIONS(2832),
    [anon_sym_nullable] = ACTIONS(2832),
    [anon_sym__Complex] = ACTIONS(2832),
    [anon_sym__Nullable] = ACTIONS(2832),
    [anon_sym__Nullable_result] = ACTIONS(2832),
    [anon_sym__Null_unspecified] = ACTIONS(2832),
    [anon_sym___autoreleasing] = ACTIONS(2832),
    [anon_sym___block] = ACTIONS(2832),
    [anon_sym___bridge] = ACTIONS(2832),
    [anon_sym___bridge_retained] = ACTIONS(2832),
    [anon_sym___bridge_transfer] = ACTIONS(2832),
    [anon_sym___complex] = ACTIONS(2832),
    [anon_sym___const] = ACTIONS(2832),
    [anon_sym___imag] = ACTIONS(2832),
    [anon_sym___kindof] = ACTIONS(2832),
    [anon_sym___nonnull] = ACTIONS(2832),
    [anon_sym___nullable] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2832),
    [anon_sym___real] = ACTIONS(2832),
    [anon_sym___strong] = ACTIONS(2832),
    [anon_sym___unsafe_unretained] = ACTIONS(2832),
    [anon_sym___unused] = ACTIONS(2832),
    [anon_sym___weak] = ACTIONS(2832),
    [anon_sym_alignas] = ACTIONS(2832),
    [anon_sym__Alignas] = ACTIONS(2832),
    [sym_primitive_type] = ACTIONS(2832),
    [anon_sym_enum] = ACTIONS(2832),
    [anon_sym_struct] = ACTIONS(2832),
    [anon_sym_union] = ACTIONS(2832),
    [anon_sym_if] = ACTIONS(2832),
    [anon_sym_switch] = ACTIONS(2832),
    [anon_sym_case] = ACTIONS(2832),
    [anon_sym_default] = ACTIONS(2832),
    [anon_sym_while] = ACTIONS(2832),
    [anon_sym_do] = ACTIONS(2832),
    [anon_sym_for] = ACTIONS(2832),
    [anon_sym_in] = ACTIONS(2832),
    [anon_sym_return] = ACTIONS(2832),
    [anon_sym_break] = ACTIONS(2832),
    [anon_sym_continue] = ACTIONS(2832),
    [anon_sym_goto] = ACTIONS(2832),
    [anon_sym_DASH_DASH] = ACTIONS(2834),
    [anon_sym_PLUS_PLUS] = ACTIONS(2834),
    [anon_sym_sizeof] = ACTIONS(2832),
    [anon_sym___alignof__] = ACTIONS(2832),
    [anon_sym___alignof] = ACTIONS(2832),
    [anon_sym__alignof] = ACTIONS(2832),
    [anon_sym_alignof] = ACTIONS(2832),
    [anon_sym__Alignof] = ACTIONS(2832),
    [anon_sym_offsetof] = ACTIONS(2832),
    [anon_sym__Generic] = ACTIONS(2832),
    [anon_sym_asm] = ACTIONS(2832),
    [anon_sym___asm__] = ACTIONS(2832),
    [anon_sym___asm] = ACTIONS(2832),
    [sym_number_literal] = ACTIONS(2834),
    [anon_sym_L_SQUOTE] = ACTIONS(2834),
    [anon_sym_u_SQUOTE] = ACTIONS(2834),
    [anon_sym_U_SQUOTE] = ACTIONS(2834),
    [anon_sym_u8_SQUOTE] = ACTIONS(2834),
    [anon_sym_SQUOTE] = ACTIONS(2834),
    [anon_sym_AT] = ACTIONS(2832),
    [anon_sym_DQUOTE] = ACTIONS(2834),
    [anon_sym_L_DQUOTE] = ACTIONS(2834),
    [anon_sym_u_DQUOTE] = ACTIONS(2834),
    [anon_sym_U_DQUOTE] = ACTIONS(2834),
    [anon_sym_u8_DQUOTE] = ACTIONS(2834),
    [sym_true] = ACTIONS(2832),
    [sym_false] = ACTIONS(2832),
    [anon_sym_NULL] = ACTIONS(2832),
    [anon_sym_nullptr] = ACTIONS(2832),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2832),
    [anon_sym___typeof] = ACTIONS(2832),
    [anon_sym_typeof] = ACTIONS(2832),
    [anon_sym_ATimport] = ACTIONS(2834),
    [aux_sym_preproc_undef_token1] = ACTIONS(2832),
    [anon_sym_POUND] = ACTIONS(2832),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2832),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2832),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2832),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2832),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2832),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2832),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2832),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_API_AVAILABLE] = ACTIONS(2832),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_API_DEPRECATED] = ACTIONS(2832),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2832),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym___deprecated_msg] = ACTIONS(2832),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2832),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2832),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2832),
    [anon_sym_ATprotocol] = ACTIONS(2834),
    [anon_sym_ATinterface] = ACTIONS(2834),
    [anon_sym_ATimplementation] = ACTIONS(2834),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2834),
    [anon_sym_ATsynthesize] = ACTIONS(2834),
    [anon_sym_ATdynamic] = ACTIONS(2834),
    [anon_sym_ATtry] = ACTIONS(2834),
    [anon_sym___try] = ACTIONS(2832),
    [anon_sym_ATthrow] = ACTIONS(2834),
    [anon_sym_ATselector] = ACTIONS(2834),
    [anon_sym_ATavailable] = ACTIONS(2834),
    [anon_sym___builtin_available] = ACTIONS(2832),
    [anon_sym_va_arg] = ACTIONS(2832),
    [anon_sym_ATencode] = ACTIONS(2834),
    [anon_sym_ATsynchronized] = ACTIONS(2834),
    [anon_sym_BOOL] = ACTIONS(2832),
    [anon_sym_IMP] = ACTIONS(2832),
    [anon_sym_SEL] = ACTIONS(2832),
    [anon_sym_Class] = ACTIONS(2832),
    [anon_sym_id] = ACTIONS(2832),
  },
  [STATE(473)] = {
    [sym_identifier] = ACTIONS(2836),
    [aux_sym_preproc_include_token1] = ACTIONS(2836),
    [aux_sym_preproc_include_token2] = ACTIONS(2836),
    [aux_sym_preproc_def_token1] = ACTIONS(2836),
    [aux_sym_preproc_if_token1] = ACTIONS(2836),
    [aux_sym_preproc_if_token2] = ACTIONS(2836),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2836),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2836),
    [aux_sym_preproc_else_token1] = ACTIONS(2836),
    [aux_sym_preproc_elif_token1] = ACTIONS(2836),
    [sym_preproc_directive] = ACTIONS(2836),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_BANG] = ACTIONS(2838),
    [anon_sym_TILDE] = ACTIONS(2838),
    [anon_sym_DASH] = ACTIONS(2836),
    [anon_sym_PLUS] = ACTIONS(2836),
    [anon_sym_STAR] = ACTIONS(2838),
    [anon_sym_CARET] = ACTIONS(2838),
    [anon_sym_AMP] = ACTIONS(2838),
    [anon_sym_SEMI] = ACTIONS(2838),
    [anon_sym___extension__] = ACTIONS(2836),
    [anon_sym_typedef] = ACTIONS(2836),
    [anon_sym_extern] = ACTIONS(2836),
    [anon_sym___attribute__] = ACTIONS(2836),
    [anon_sym___attribute] = ACTIONS(2836),
    [anon_sym_noreturn] = ACTIONS(2836),
    [anon_sym_LBRACK] = ACTIONS(2838),
    [anon_sym___declspec] = ACTIONS(2836),
    [anon_sym___cdecl] = ACTIONS(2836),
    [anon_sym___clrcall] = ACTIONS(2836),
    [anon_sym___stdcall] = ACTIONS(2836),
    [anon_sym___fastcall] = ACTIONS(2836),
    [anon_sym___thiscall] = ACTIONS(2836),
    [anon_sym___vectorcall] = ACTIONS(2836),
    [anon_sym_LBRACE] = ACTIONS(2838),
    [anon_sym_signed] = ACTIONS(2836),
    [anon_sym_unsigned] = ACTIONS(2836),
    [anon_sym_long] = ACTIONS(2836),
    [anon_sym_short] = ACTIONS(2836),
    [anon_sym_static] = ACTIONS(2836),
    [anon_sym_ATautoreleasepool] = ACTIONS(2838),
    [anon_sym_auto] = ACTIONS(2836),
    [anon_sym_register] = ACTIONS(2836),
    [anon_sym_inline] = ACTIONS(2836),
    [anon_sym___inline] = ACTIONS(2836),
    [anon_sym___inline__] = ACTIONS(2836),
    [anon_sym___forceinline] = ACTIONS(2836),
    [anon_sym_thread_local] = ACTIONS(2836),
    [anon_sym___thread] = ACTIONS(2836),
    [anon_sym_CG_EXTERN] = ACTIONS(2836),
    [anon_sym_CG_INLINE] = ACTIONS(2836),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2836),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2836),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2836),
    [anon_sym_IBOutlet] = ACTIONS(2836),
    [anon_sym_IBInspectable] = ACTIONS(2836),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2836),
    [anon_sym_NS_INLINE] = ACTIONS(2836),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2836),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2836),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2836),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2836),
    [anon_sym_const] = ACTIONS(2836),
    [anon_sym_constexpr] = ACTIONS(2836),
    [anon_sym_volatile] = ACTIONS(2836),
    [anon_sym_restrict] = ACTIONS(2836),
    [anon_sym___restrict__] = ACTIONS(2836),
    [anon_sym__Atomic] = ACTIONS(2836),
    [anon_sym__Noreturn] = ACTIONS(2836),
    [anon_sym__Nonnull] = ACTIONS(2836),
    [anon_sym_nullable] = ACTIONS(2836),
    [anon_sym__Complex] = ACTIONS(2836),
    [anon_sym__Nullable] = ACTIONS(2836),
    [anon_sym__Nullable_result] = ACTIONS(2836),
    [anon_sym__Null_unspecified] = ACTIONS(2836),
    [anon_sym___autoreleasing] = ACTIONS(2836),
    [anon_sym___block] = ACTIONS(2836),
    [anon_sym___bridge] = ACTIONS(2836),
    [anon_sym___bridge_retained] = ACTIONS(2836),
    [anon_sym___bridge_transfer] = ACTIONS(2836),
    [anon_sym___complex] = ACTIONS(2836),
    [anon_sym___const] = ACTIONS(2836),
    [anon_sym___imag] = ACTIONS(2836),
    [anon_sym___kindof] = ACTIONS(2836),
    [anon_sym___nonnull] = ACTIONS(2836),
    [anon_sym___nullable] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2836),
    [anon_sym___real] = ACTIONS(2836),
    [anon_sym___strong] = ACTIONS(2836),
    [anon_sym___unsafe_unretained] = ACTIONS(2836),
    [anon_sym___unused] = ACTIONS(2836),
    [anon_sym___weak] = ACTIONS(2836),
    [anon_sym_alignas] = ACTIONS(2836),
    [anon_sym__Alignas] = ACTIONS(2836),
    [sym_primitive_type] = ACTIONS(2836),
    [anon_sym_enum] = ACTIONS(2836),
    [anon_sym_struct] = ACTIONS(2836),
    [anon_sym_union] = ACTIONS(2836),
    [anon_sym_if] = ACTIONS(2836),
    [anon_sym_switch] = ACTIONS(2836),
    [anon_sym_case] = ACTIONS(2836),
    [anon_sym_default] = ACTIONS(2836),
    [anon_sym_while] = ACTIONS(2836),
    [anon_sym_do] = ACTIONS(2836),
    [anon_sym_for] = ACTIONS(2836),
    [anon_sym_in] = ACTIONS(2836),
    [anon_sym_return] = ACTIONS(2836),
    [anon_sym_break] = ACTIONS(2836),
    [anon_sym_continue] = ACTIONS(2836),
    [anon_sym_goto] = ACTIONS(2836),
    [anon_sym_DASH_DASH] = ACTIONS(2838),
    [anon_sym_PLUS_PLUS] = ACTIONS(2838),
    [anon_sym_sizeof] = ACTIONS(2836),
    [anon_sym___alignof__] = ACTIONS(2836),
    [anon_sym___alignof] = ACTIONS(2836),
    [anon_sym__alignof] = ACTIONS(2836),
    [anon_sym_alignof] = ACTIONS(2836),
    [anon_sym__Alignof] = ACTIONS(2836),
    [anon_sym_offsetof] = ACTIONS(2836),
    [anon_sym__Generic] = ACTIONS(2836),
    [anon_sym_asm] = ACTIONS(2836),
    [anon_sym___asm__] = ACTIONS(2836),
    [anon_sym___asm] = ACTIONS(2836),
    [sym_number_literal] = ACTIONS(2838),
    [anon_sym_L_SQUOTE] = ACTIONS(2838),
    [anon_sym_u_SQUOTE] = ACTIONS(2838),
    [anon_sym_U_SQUOTE] = ACTIONS(2838),
    [anon_sym_u8_SQUOTE] = ACTIONS(2838),
    [anon_sym_SQUOTE] = ACTIONS(2838),
    [anon_sym_AT] = ACTIONS(2836),
    [anon_sym_DQUOTE] = ACTIONS(2838),
    [anon_sym_L_DQUOTE] = ACTIONS(2838),
    [anon_sym_u_DQUOTE] = ACTIONS(2838),
    [anon_sym_U_DQUOTE] = ACTIONS(2838),
    [anon_sym_u8_DQUOTE] = ACTIONS(2838),
    [sym_true] = ACTIONS(2836),
    [sym_false] = ACTIONS(2836),
    [anon_sym_NULL] = ACTIONS(2836),
    [anon_sym_nullptr] = ACTIONS(2836),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2836),
    [anon_sym___typeof] = ACTIONS(2836),
    [anon_sym_typeof] = ACTIONS(2836),
    [anon_sym_ATimport] = ACTIONS(2838),
    [aux_sym_preproc_undef_token1] = ACTIONS(2836),
    [anon_sym_POUND] = ACTIONS(2836),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2836),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2836),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2836),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2836),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2836),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2836),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2836),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_API_AVAILABLE] = ACTIONS(2836),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_API_DEPRECATED] = ACTIONS(2836),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2836),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym___deprecated_msg] = ACTIONS(2836),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2836),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2836),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2836),
    [anon_sym_ATprotocol] = ACTIONS(2838),
    [anon_sym_ATinterface] = ACTIONS(2838),
    [anon_sym_ATimplementation] = ACTIONS(2838),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2838),
    [anon_sym_ATsynthesize] = ACTIONS(2838),
    [anon_sym_ATdynamic] = ACTIONS(2838),
    [anon_sym_ATtry] = ACTIONS(2838),
    [anon_sym___try] = ACTIONS(2836),
    [anon_sym_ATthrow] = ACTIONS(2838),
    [anon_sym_ATselector] = ACTIONS(2838),
    [anon_sym_ATavailable] = ACTIONS(2838),
    [anon_sym___builtin_available] = ACTIONS(2836),
    [anon_sym_va_arg] = ACTIONS(2836),
    [anon_sym_ATencode] = ACTIONS(2838),
    [anon_sym_ATsynchronized] = ACTIONS(2838),
    [anon_sym_BOOL] = ACTIONS(2836),
    [anon_sym_IMP] = ACTIONS(2836),
    [anon_sym_SEL] = ACTIONS(2836),
    [anon_sym_Class] = ACTIONS(2836),
    [anon_sym_id] = ACTIONS(2836),
  },
  [STATE(474)] = {
    [sym_identifier] = ACTIONS(2662),
    [aux_sym_preproc_include_token1] = ACTIONS(2662),
    [aux_sym_preproc_include_token2] = ACTIONS(2662),
    [aux_sym_preproc_def_token1] = ACTIONS(2662),
    [aux_sym_preproc_if_token1] = ACTIONS(2662),
    [aux_sym_preproc_if_token2] = ACTIONS(2662),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2662),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2662),
    [aux_sym_preproc_else_token1] = ACTIONS(2662),
    [aux_sym_preproc_elif_token1] = ACTIONS(2662),
    [sym_preproc_directive] = ACTIONS(2662),
    [anon_sym_LPAREN2] = ACTIONS(2665),
    [anon_sym_BANG] = ACTIONS(2665),
    [anon_sym_TILDE] = ACTIONS(2665),
    [anon_sym_DASH] = ACTIONS(2662),
    [anon_sym_PLUS] = ACTIONS(2662),
    [anon_sym_STAR] = ACTIONS(2665),
    [anon_sym_CARET] = ACTIONS(2665),
    [anon_sym_AMP] = ACTIONS(2665),
    [anon_sym_SEMI] = ACTIONS(2665),
    [anon_sym___extension__] = ACTIONS(2662),
    [anon_sym_typedef] = ACTIONS(2662),
    [anon_sym_extern] = ACTIONS(2662),
    [anon_sym___attribute__] = ACTIONS(2662),
    [anon_sym___attribute] = ACTIONS(2662),
    [anon_sym_noreturn] = ACTIONS(2662),
    [anon_sym_LBRACK] = ACTIONS(2665),
    [anon_sym___declspec] = ACTIONS(2662),
    [anon_sym___cdecl] = ACTIONS(2662),
    [anon_sym___clrcall] = ACTIONS(2662),
    [anon_sym___stdcall] = ACTIONS(2662),
    [anon_sym___fastcall] = ACTIONS(2662),
    [anon_sym___thiscall] = ACTIONS(2662),
    [anon_sym___vectorcall] = ACTIONS(2662),
    [anon_sym_LBRACE] = ACTIONS(2665),
    [anon_sym_signed] = ACTIONS(2662),
    [anon_sym_unsigned] = ACTIONS(2662),
    [anon_sym_long] = ACTIONS(2662),
    [anon_sym_short] = ACTIONS(2662),
    [anon_sym_static] = ACTIONS(2662),
    [anon_sym_ATautoreleasepool] = ACTIONS(2665),
    [anon_sym_auto] = ACTIONS(2662),
    [anon_sym_register] = ACTIONS(2662),
    [anon_sym_inline] = ACTIONS(2662),
    [anon_sym___inline] = ACTIONS(2662),
    [anon_sym___inline__] = ACTIONS(2662),
    [anon_sym___forceinline] = ACTIONS(2662),
    [anon_sym_thread_local] = ACTIONS(2662),
    [anon_sym___thread] = ACTIONS(2662),
    [anon_sym_CG_EXTERN] = ACTIONS(2662),
    [anon_sym_CG_INLINE] = ACTIONS(2662),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2662),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2662),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2662),
    [anon_sym_IBOutlet] = ACTIONS(2662),
    [anon_sym_IBInspectable] = ACTIONS(2662),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2662),
    [anon_sym_NS_INLINE] = ACTIONS(2662),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2662),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2662),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2662),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2662),
    [anon_sym_const] = ACTIONS(2662),
    [anon_sym_constexpr] = ACTIONS(2662),
    [anon_sym_volatile] = ACTIONS(2662),
    [anon_sym_restrict] = ACTIONS(2662),
    [anon_sym___restrict__] = ACTIONS(2662),
    [anon_sym__Atomic] = ACTIONS(2662),
    [anon_sym__Noreturn] = ACTIONS(2662),
    [anon_sym__Nonnull] = ACTIONS(2662),
    [anon_sym_nullable] = ACTIONS(2662),
    [anon_sym__Complex] = ACTIONS(2662),
    [anon_sym__Nullable] = ACTIONS(2662),
    [anon_sym__Nullable_result] = ACTIONS(2662),
    [anon_sym__Null_unspecified] = ACTIONS(2662),
    [anon_sym___autoreleasing] = ACTIONS(2662),
    [anon_sym___block] = ACTIONS(2662),
    [anon_sym___bridge] = ACTIONS(2662),
    [anon_sym___bridge_retained] = ACTIONS(2662),
    [anon_sym___bridge_transfer] = ACTIONS(2662),
    [anon_sym___complex] = ACTIONS(2662),
    [anon_sym___const] = ACTIONS(2662),
    [anon_sym___imag] = ACTIONS(2662),
    [anon_sym___kindof] = ACTIONS(2662),
    [anon_sym___nonnull] = ACTIONS(2662),
    [anon_sym___nullable] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2662),
    [anon_sym___real] = ACTIONS(2662),
    [anon_sym___strong] = ACTIONS(2662),
    [anon_sym___unsafe_unretained] = ACTIONS(2662),
    [anon_sym___unused] = ACTIONS(2662),
    [anon_sym___weak] = ACTIONS(2662),
    [anon_sym_alignas] = ACTIONS(2662),
    [anon_sym__Alignas] = ACTIONS(2662),
    [sym_primitive_type] = ACTIONS(2662),
    [anon_sym_enum] = ACTIONS(2662),
    [anon_sym_struct] = ACTIONS(2662),
    [anon_sym_union] = ACTIONS(2662),
    [anon_sym_if] = ACTIONS(2662),
    [anon_sym_switch] = ACTIONS(2662),
    [anon_sym_case] = ACTIONS(2662),
    [anon_sym_default] = ACTIONS(2662),
    [anon_sym_while] = ACTIONS(2662),
    [anon_sym_do] = ACTIONS(2662),
    [anon_sym_for] = ACTIONS(2662),
    [anon_sym_in] = ACTIONS(2662),
    [anon_sym_return] = ACTIONS(2662),
    [anon_sym_break] = ACTIONS(2662),
    [anon_sym_continue] = ACTIONS(2662),
    [anon_sym_goto] = ACTIONS(2662),
    [anon_sym_DASH_DASH] = ACTIONS(2665),
    [anon_sym_PLUS_PLUS] = ACTIONS(2665),
    [anon_sym_sizeof] = ACTIONS(2662),
    [anon_sym___alignof__] = ACTIONS(2662),
    [anon_sym___alignof] = ACTIONS(2662),
    [anon_sym__alignof] = ACTIONS(2662),
    [anon_sym_alignof] = ACTIONS(2662),
    [anon_sym__Alignof] = ACTIONS(2662),
    [anon_sym_offsetof] = ACTIONS(2662),
    [anon_sym__Generic] = ACTIONS(2662),
    [anon_sym_asm] = ACTIONS(2662),
    [anon_sym___asm__] = ACTIONS(2662),
    [anon_sym___asm] = ACTIONS(2662),
    [sym_number_literal] = ACTIONS(2665),
    [anon_sym_L_SQUOTE] = ACTIONS(2665),
    [anon_sym_u_SQUOTE] = ACTIONS(2665),
    [anon_sym_U_SQUOTE] = ACTIONS(2665),
    [anon_sym_u8_SQUOTE] = ACTIONS(2665),
    [anon_sym_SQUOTE] = ACTIONS(2665),
    [anon_sym_AT] = ACTIONS(2662),
    [anon_sym_DQUOTE] = ACTIONS(2665),
    [anon_sym_L_DQUOTE] = ACTIONS(2665),
    [anon_sym_u_DQUOTE] = ACTIONS(2665),
    [anon_sym_U_DQUOTE] = ACTIONS(2665),
    [anon_sym_u8_DQUOTE] = ACTIONS(2665),
    [sym_true] = ACTIONS(2662),
    [sym_false] = ACTIONS(2662),
    [anon_sym_NULL] = ACTIONS(2662),
    [anon_sym_nullptr] = ACTIONS(2662),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2662),
    [anon_sym___typeof] = ACTIONS(2662),
    [anon_sym_typeof] = ACTIONS(2662),
    [anon_sym_ATimport] = ACTIONS(2665),
    [aux_sym_preproc_undef_token1] = ACTIONS(2662),
    [anon_sym_POUND] = ACTIONS(2662),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2662),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2662),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2662),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2662),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2662),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2662),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2662),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_API_AVAILABLE] = ACTIONS(2662),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_API_DEPRECATED] = ACTIONS(2662),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2662),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym___deprecated_msg] = ACTIONS(2662),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2662),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2662),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2662),
    [anon_sym_ATprotocol] = ACTIONS(2665),
    [anon_sym_ATinterface] = ACTIONS(2665),
    [anon_sym_ATimplementation] = ACTIONS(2665),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2665),
    [anon_sym_ATsynthesize] = ACTIONS(2665),
    [anon_sym_ATdynamic] = ACTIONS(2665),
    [anon_sym_ATtry] = ACTIONS(2665),
    [anon_sym___try] = ACTIONS(2662),
    [anon_sym_ATthrow] = ACTIONS(2665),
    [anon_sym_ATselector] = ACTIONS(2665),
    [anon_sym_ATavailable] = ACTIONS(2665),
    [anon_sym___builtin_available] = ACTIONS(2662),
    [anon_sym_va_arg] = ACTIONS(2662),
    [anon_sym_ATencode] = ACTIONS(2665),
    [anon_sym_ATsynchronized] = ACTIONS(2665),
    [anon_sym_BOOL] = ACTIONS(2662),
    [anon_sym_IMP] = ACTIONS(2662),
    [anon_sym_SEL] = ACTIONS(2662),
    [anon_sym_Class] = ACTIONS(2662),
    [anon_sym_id] = ACTIONS(2662),
  },
  [STATE(475)] = {
    [sym_identifier] = ACTIONS(2840),
    [aux_sym_preproc_include_token1] = ACTIONS(2840),
    [aux_sym_preproc_include_token2] = ACTIONS(2840),
    [aux_sym_preproc_def_token1] = ACTIONS(2840),
    [aux_sym_preproc_if_token1] = ACTIONS(2840),
    [aux_sym_preproc_if_token2] = ACTIONS(2840),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2840),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2840),
    [aux_sym_preproc_else_token1] = ACTIONS(2840),
    [aux_sym_preproc_elif_token1] = ACTIONS(2840),
    [sym_preproc_directive] = ACTIONS(2840),
    [anon_sym_LPAREN2] = ACTIONS(2842),
    [anon_sym_BANG] = ACTIONS(2842),
    [anon_sym_TILDE] = ACTIONS(2842),
    [anon_sym_DASH] = ACTIONS(2840),
    [anon_sym_PLUS] = ACTIONS(2840),
    [anon_sym_STAR] = ACTIONS(2842),
    [anon_sym_CARET] = ACTIONS(2842),
    [anon_sym_AMP] = ACTIONS(2842),
    [anon_sym_SEMI] = ACTIONS(2842),
    [anon_sym___extension__] = ACTIONS(2840),
    [anon_sym_typedef] = ACTIONS(2840),
    [anon_sym_extern] = ACTIONS(2840),
    [anon_sym___attribute__] = ACTIONS(2840),
    [anon_sym___attribute] = ACTIONS(2840),
    [anon_sym_noreturn] = ACTIONS(2840),
    [anon_sym_LBRACK] = ACTIONS(2842),
    [anon_sym___declspec] = ACTIONS(2840),
    [anon_sym___cdecl] = ACTIONS(2840),
    [anon_sym___clrcall] = ACTIONS(2840),
    [anon_sym___stdcall] = ACTIONS(2840),
    [anon_sym___fastcall] = ACTIONS(2840),
    [anon_sym___thiscall] = ACTIONS(2840),
    [anon_sym___vectorcall] = ACTIONS(2840),
    [anon_sym_LBRACE] = ACTIONS(2842),
    [anon_sym_signed] = ACTIONS(2840),
    [anon_sym_unsigned] = ACTIONS(2840),
    [anon_sym_long] = ACTIONS(2840),
    [anon_sym_short] = ACTIONS(2840),
    [anon_sym_static] = ACTIONS(2840),
    [anon_sym_ATautoreleasepool] = ACTIONS(2842),
    [anon_sym_auto] = ACTIONS(2840),
    [anon_sym_register] = ACTIONS(2840),
    [anon_sym_inline] = ACTIONS(2840),
    [anon_sym___inline] = ACTIONS(2840),
    [anon_sym___inline__] = ACTIONS(2840),
    [anon_sym___forceinline] = ACTIONS(2840),
    [anon_sym_thread_local] = ACTIONS(2840),
    [anon_sym___thread] = ACTIONS(2840),
    [anon_sym_CG_EXTERN] = ACTIONS(2840),
    [anon_sym_CG_INLINE] = ACTIONS(2840),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2840),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2840),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2840),
    [anon_sym_IBOutlet] = ACTIONS(2840),
    [anon_sym_IBInspectable] = ACTIONS(2840),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2840),
    [anon_sym_NS_INLINE] = ACTIONS(2840),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2840),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2840),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2840),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2840),
    [anon_sym_const] = ACTIONS(2840),
    [anon_sym_constexpr] = ACTIONS(2840),
    [anon_sym_volatile] = ACTIONS(2840),
    [anon_sym_restrict] = ACTIONS(2840),
    [anon_sym___restrict__] = ACTIONS(2840),
    [anon_sym__Atomic] = ACTIONS(2840),
    [anon_sym__Noreturn] = ACTIONS(2840),
    [anon_sym__Nonnull] = ACTIONS(2840),
    [anon_sym_nullable] = ACTIONS(2840),
    [anon_sym__Complex] = ACTIONS(2840),
    [anon_sym__Nullable] = ACTIONS(2840),
    [anon_sym__Nullable_result] = ACTIONS(2840),
    [anon_sym__Null_unspecified] = ACTIONS(2840),
    [anon_sym___autoreleasing] = ACTIONS(2840),
    [anon_sym___block] = ACTIONS(2840),
    [anon_sym___bridge] = ACTIONS(2840),
    [anon_sym___bridge_retained] = ACTIONS(2840),
    [anon_sym___bridge_transfer] = ACTIONS(2840),
    [anon_sym___complex] = ACTIONS(2840),
    [anon_sym___const] = ACTIONS(2840),
    [anon_sym___imag] = ACTIONS(2840),
    [anon_sym___kindof] = ACTIONS(2840),
    [anon_sym___nonnull] = ACTIONS(2840),
    [anon_sym___nullable] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2840),
    [anon_sym___real] = ACTIONS(2840),
    [anon_sym___strong] = ACTIONS(2840),
    [anon_sym___unsafe_unretained] = ACTIONS(2840),
    [anon_sym___unused] = ACTIONS(2840),
    [anon_sym___weak] = ACTIONS(2840),
    [anon_sym_alignas] = ACTIONS(2840),
    [anon_sym__Alignas] = ACTIONS(2840),
    [sym_primitive_type] = ACTIONS(2840),
    [anon_sym_enum] = ACTIONS(2840),
    [anon_sym_struct] = ACTIONS(2840),
    [anon_sym_union] = ACTIONS(2840),
    [anon_sym_if] = ACTIONS(2840),
    [anon_sym_switch] = ACTIONS(2840),
    [anon_sym_case] = ACTIONS(2840),
    [anon_sym_default] = ACTIONS(2840),
    [anon_sym_while] = ACTIONS(2840),
    [anon_sym_do] = ACTIONS(2840),
    [anon_sym_for] = ACTIONS(2840),
    [anon_sym_in] = ACTIONS(2840),
    [anon_sym_return] = ACTIONS(2840),
    [anon_sym_break] = ACTIONS(2840),
    [anon_sym_continue] = ACTIONS(2840),
    [anon_sym_goto] = ACTIONS(2840),
    [anon_sym_DASH_DASH] = ACTIONS(2842),
    [anon_sym_PLUS_PLUS] = ACTIONS(2842),
    [anon_sym_sizeof] = ACTIONS(2840),
    [anon_sym___alignof__] = ACTIONS(2840),
    [anon_sym___alignof] = ACTIONS(2840),
    [anon_sym__alignof] = ACTIONS(2840),
    [anon_sym_alignof] = ACTIONS(2840),
    [anon_sym__Alignof] = ACTIONS(2840),
    [anon_sym_offsetof] = ACTIONS(2840),
    [anon_sym__Generic] = ACTIONS(2840),
    [anon_sym_asm] = ACTIONS(2840),
    [anon_sym___asm__] = ACTIONS(2840),
    [anon_sym___asm] = ACTIONS(2840),
    [sym_number_literal] = ACTIONS(2842),
    [anon_sym_L_SQUOTE] = ACTIONS(2842),
    [anon_sym_u_SQUOTE] = ACTIONS(2842),
    [anon_sym_U_SQUOTE] = ACTIONS(2842),
    [anon_sym_u8_SQUOTE] = ACTIONS(2842),
    [anon_sym_SQUOTE] = ACTIONS(2842),
    [anon_sym_AT] = ACTIONS(2840),
    [anon_sym_DQUOTE] = ACTIONS(2842),
    [anon_sym_L_DQUOTE] = ACTIONS(2842),
    [anon_sym_u_DQUOTE] = ACTIONS(2842),
    [anon_sym_U_DQUOTE] = ACTIONS(2842),
    [anon_sym_u8_DQUOTE] = ACTIONS(2842),
    [sym_true] = ACTIONS(2840),
    [sym_false] = ACTIONS(2840),
    [anon_sym_NULL] = ACTIONS(2840),
    [anon_sym_nullptr] = ACTIONS(2840),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2840),
    [anon_sym___typeof] = ACTIONS(2840),
    [anon_sym_typeof] = ACTIONS(2840),
    [anon_sym_ATimport] = ACTIONS(2842),
    [aux_sym_preproc_undef_token1] = ACTIONS(2840),
    [anon_sym_POUND] = ACTIONS(2840),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2840),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2840),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2840),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2840),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2840),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2840),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2840),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_API_AVAILABLE] = ACTIONS(2840),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_API_DEPRECATED] = ACTIONS(2840),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2840),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym___deprecated_msg] = ACTIONS(2840),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2840),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2840),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2840),
    [anon_sym_ATprotocol] = ACTIONS(2842),
    [anon_sym_ATinterface] = ACTIONS(2842),
    [anon_sym_ATimplementation] = ACTIONS(2842),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2842),
    [anon_sym_ATsynthesize] = ACTIONS(2842),
    [anon_sym_ATdynamic] = ACTIONS(2842),
    [anon_sym_ATtry] = ACTIONS(2842),
    [anon_sym___try] = ACTIONS(2840),
    [anon_sym_ATthrow] = ACTIONS(2842),
    [anon_sym_ATselector] = ACTIONS(2842),
    [anon_sym_ATavailable] = ACTIONS(2842),
    [anon_sym___builtin_available] = ACTIONS(2840),
    [anon_sym_va_arg] = ACTIONS(2840),
    [anon_sym_ATencode] = ACTIONS(2842),
    [anon_sym_ATsynchronized] = ACTIONS(2842),
    [anon_sym_BOOL] = ACTIONS(2840),
    [anon_sym_IMP] = ACTIONS(2840),
    [anon_sym_SEL] = ACTIONS(2840),
    [anon_sym_Class] = ACTIONS(2840),
    [anon_sym_id] = ACTIONS(2840),
  },
  [STATE(476)] = {
    [sym_identifier] = ACTIONS(2740),
    [aux_sym_preproc_include_token1] = ACTIONS(2740),
    [aux_sym_preproc_include_token2] = ACTIONS(2740),
    [aux_sym_preproc_def_token1] = ACTIONS(2740),
    [aux_sym_preproc_if_token1] = ACTIONS(2740),
    [aux_sym_preproc_if_token2] = ACTIONS(2740),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2740),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2740),
    [aux_sym_preproc_else_token1] = ACTIONS(2740),
    [aux_sym_preproc_elif_token1] = ACTIONS(2740),
    [sym_preproc_directive] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_BANG] = ACTIONS(2742),
    [anon_sym_TILDE] = ACTIONS(2742),
    [anon_sym_DASH] = ACTIONS(2740),
    [anon_sym_PLUS] = ACTIONS(2740),
    [anon_sym_STAR] = ACTIONS(2742),
    [anon_sym_CARET] = ACTIONS(2742),
    [anon_sym_AMP] = ACTIONS(2742),
    [anon_sym_SEMI] = ACTIONS(2742),
    [anon_sym___extension__] = ACTIONS(2740),
    [anon_sym_typedef] = ACTIONS(2740),
    [anon_sym_extern] = ACTIONS(2740),
    [anon_sym___attribute__] = ACTIONS(2740),
    [anon_sym___attribute] = ACTIONS(2740),
    [anon_sym_noreturn] = ACTIONS(2740),
    [anon_sym_LBRACK] = ACTIONS(2742),
    [anon_sym___declspec] = ACTIONS(2740),
    [anon_sym___cdecl] = ACTIONS(2740),
    [anon_sym___clrcall] = ACTIONS(2740),
    [anon_sym___stdcall] = ACTIONS(2740),
    [anon_sym___fastcall] = ACTIONS(2740),
    [anon_sym___thiscall] = ACTIONS(2740),
    [anon_sym___vectorcall] = ACTIONS(2740),
    [anon_sym_LBRACE] = ACTIONS(2742),
    [anon_sym_signed] = ACTIONS(2740),
    [anon_sym_unsigned] = ACTIONS(2740),
    [anon_sym_long] = ACTIONS(2740),
    [anon_sym_short] = ACTIONS(2740),
    [anon_sym_static] = ACTIONS(2740),
    [anon_sym_ATautoreleasepool] = ACTIONS(2742),
    [anon_sym_auto] = ACTIONS(2740),
    [anon_sym_register] = ACTIONS(2740),
    [anon_sym_inline] = ACTIONS(2740),
    [anon_sym___inline] = ACTIONS(2740),
    [anon_sym___inline__] = ACTIONS(2740),
    [anon_sym___forceinline] = ACTIONS(2740),
    [anon_sym_thread_local] = ACTIONS(2740),
    [anon_sym___thread] = ACTIONS(2740),
    [anon_sym_CG_EXTERN] = ACTIONS(2740),
    [anon_sym_CG_INLINE] = ACTIONS(2740),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2740),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2740),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2740),
    [anon_sym_IBOutlet] = ACTIONS(2740),
    [anon_sym_IBInspectable] = ACTIONS(2740),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2740),
    [anon_sym_NS_INLINE] = ACTIONS(2740),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2740),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2740),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2740),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2740),
    [anon_sym_const] = ACTIONS(2740),
    [anon_sym_constexpr] = ACTIONS(2740),
    [anon_sym_volatile] = ACTIONS(2740),
    [anon_sym_restrict] = ACTIONS(2740),
    [anon_sym___restrict__] = ACTIONS(2740),
    [anon_sym__Atomic] = ACTIONS(2740),
    [anon_sym__Noreturn] = ACTIONS(2740),
    [anon_sym__Nonnull] = ACTIONS(2740),
    [anon_sym_nullable] = ACTIONS(2740),
    [anon_sym__Complex] = ACTIONS(2740),
    [anon_sym__Nullable] = ACTIONS(2740),
    [anon_sym__Nullable_result] = ACTIONS(2740),
    [anon_sym__Null_unspecified] = ACTIONS(2740),
    [anon_sym___autoreleasing] = ACTIONS(2740),
    [anon_sym___block] = ACTIONS(2740),
    [anon_sym___bridge] = ACTIONS(2740),
    [anon_sym___bridge_retained] = ACTIONS(2740),
    [anon_sym___bridge_transfer] = ACTIONS(2740),
    [anon_sym___complex] = ACTIONS(2740),
    [anon_sym___const] = ACTIONS(2740),
    [anon_sym___imag] = ACTIONS(2740),
    [anon_sym___kindof] = ACTIONS(2740),
    [anon_sym___nonnull] = ACTIONS(2740),
    [anon_sym___nullable] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2740),
    [anon_sym___real] = ACTIONS(2740),
    [anon_sym___strong] = ACTIONS(2740),
    [anon_sym___unsafe_unretained] = ACTIONS(2740),
    [anon_sym___unused] = ACTIONS(2740),
    [anon_sym___weak] = ACTIONS(2740),
    [anon_sym_alignas] = ACTIONS(2740),
    [anon_sym__Alignas] = ACTIONS(2740),
    [sym_primitive_type] = ACTIONS(2740),
    [anon_sym_enum] = ACTIONS(2740),
    [anon_sym_struct] = ACTIONS(2740),
    [anon_sym_union] = ACTIONS(2740),
    [anon_sym_if] = ACTIONS(2740),
    [anon_sym_switch] = ACTIONS(2740),
    [anon_sym_case] = ACTIONS(2740),
    [anon_sym_default] = ACTIONS(2740),
    [anon_sym_while] = ACTIONS(2740),
    [anon_sym_do] = ACTIONS(2740),
    [anon_sym_for] = ACTIONS(2740),
    [anon_sym_in] = ACTIONS(2740),
    [anon_sym_return] = ACTIONS(2740),
    [anon_sym_break] = ACTIONS(2740),
    [anon_sym_continue] = ACTIONS(2740),
    [anon_sym_goto] = ACTIONS(2740),
    [anon_sym_DASH_DASH] = ACTIONS(2742),
    [anon_sym_PLUS_PLUS] = ACTIONS(2742),
    [anon_sym_sizeof] = ACTIONS(2740),
    [anon_sym___alignof__] = ACTIONS(2740),
    [anon_sym___alignof] = ACTIONS(2740),
    [anon_sym__alignof] = ACTIONS(2740),
    [anon_sym_alignof] = ACTIONS(2740),
    [anon_sym__Alignof] = ACTIONS(2740),
    [anon_sym_offsetof] = ACTIONS(2740),
    [anon_sym__Generic] = ACTIONS(2740),
    [anon_sym_asm] = ACTIONS(2740),
    [anon_sym___asm__] = ACTIONS(2740),
    [anon_sym___asm] = ACTIONS(2740),
    [sym_number_literal] = ACTIONS(2742),
    [anon_sym_L_SQUOTE] = ACTIONS(2742),
    [anon_sym_u_SQUOTE] = ACTIONS(2742),
    [anon_sym_U_SQUOTE] = ACTIONS(2742),
    [anon_sym_u8_SQUOTE] = ACTIONS(2742),
    [anon_sym_SQUOTE] = ACTIONS(2742),
    [anon_sym_AT] = ACTIONS(2740),
    [anon_sym_DQUOTE] = ACTIONS(2742),
    [anon_sym_L_DQUOTE] = ACTIONS(2742),
    [anon_sym_u_DQUOTE] = ACTIONS(2742),
    [anon_sym_U_DQUOTE] = ACTIONS(2742),
    [anon_sym_u8_DQUOTE] = ACTIONS(2742),
    [sym_true] = ACTIONS(2740),
    [sym_false] = ACTIONS(2740),
    [anon_sym_NULL] = ACTIONS(2740),
    [anon_sym_nullptr] = ACTIONS(2740),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2740),
    [anon_sym___typeof] = ACTIONS(2740),
    [anon_sym_typeof] = ACTIONS(2740),
    [anon_sym_ATimport] = ACTIONS(2742),
    [aux_sym_preproc_undef_token1] = ACTIONS(2740),
    [anon_sym_POUND] = ACTIONS(2740),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2740),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2740),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2740),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2740),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2740),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2740),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2740),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_API_AVAILABLE] = ACTIONS(2740),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_API_DEPRECATED] = ACTIONS(2740),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2740),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym___deprecated_msg] = ACTIONS(2740),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2740),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2740),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2740),
    [anon_sym_ATprotocol] = ACTIONS(2742),
    [anon_sym_ATinterface] = ACTIONS(2742),
    [anon_sym_ATimplementation] = ACTIONS(2742),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2742),
    [anon_sym_ATsynthesize] = ACTIONS(2742),
    [anon_sym_ATdynamic] = ACTIONS(2742),
    [anon_sym_ATtry] = ACTIONS(2742),
    [anon_sym___try] = ACTIONS(2740),
    [anon_sym_ATthrow] = ACTIONS(2742),
    [anon_sym_ATselector] = ACTIONS(2742),
    [anon_sym_ATavailable] = ACTIONS(2742),
    [anon_sym___builtin_available] = ACTIONS(2740),
    [anon_sym_va_arg] = ACTIONS(2740),
    [anon_sym_ATencode] = ACTIONS(2742),
    [anon_sym_ATsynchronized] = ACTIONS(2742),
    [anon_sym_BOOL] = ACTIONS(2740),
    [anon_sym_IMP] = ACTIONS(2740),
    [anon_sym_SEL] = ACTIONS(2740),
    [anon_sym_Class] = ACTIONS(2740),
    [anon_sym_id] = ACTIONS(2740),
  },
  [STATE(477)] = {
    [sym_identifier] = ACTIONS(2700),
    [aux_sym_preproc_include_token1] = ACTIONS(2700),
    [aux_sym_preproc_include_token2] = ACTIONS(2700),
    [aux_sym_preproc_def_token1] = ACTIONS(2700),
    [aux_sym_preproc_if_token1] = ACTIONS(2700),
    [aux_sym_preproc_if_token2] = ACTIONS(2700),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2700),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2700),
    [aux_sym_preproc_else_token1] = ACTIONS(2700),
    [aux_sym_preproc_elif_token1] = ACTIONS(2700),
    [sym_preproc_directive] = ACTIONS(2700),
    [anon_sym_LPAREN2] = ACTIONS(2702),
    [anon_sym_BANG] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2700),
    [anon_sym_PLUS] = ACTIONS(2700),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym___extension__] = ACTIONS(2700),
    [anon_sym_typedef] = ACTIONS(2700),
    [anon_sym_extern] = ACTIONS(2700),
    [anon_sym___attribute__] = ACTIONS(2700),
    [anon_sym___attribute] = ACTIONS(2700),
    [anon_sym_noreturn] = ACTIONS(2700),
    [anon_sym_LBRACK] = ACTIONS(2702),
    [anon_sym___declspec] = ACTIONS(2700),
    [anon_sym___cdecl] = ACTIONS(2700),
    [anon_sym___clrcall] = ACTIONS(2700),
    [anon_sym___stdcall] = ACTIONS(2700),
    [anon_sym___fastcall] = ACTIONS(2700),
    [anon_sym___thiscall] = ACTIONS(2700),
    [anon_sym___vectorcall] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2702),
    [anon_sym_signed] = ACTIONS(2700),
    [anon_sym_unsigned] = ACTIONS(2700),
    [anon_sym_long] = ACTIONS(2700),
    [anon_sym_short] = ACTIONS(2700),
    [anon_sym_static] = ACTIONS(2700),
    [anon_sym_ATautoreleasepool] = ACTIONS(2702),
    [anon_sym_auto] = ACTIONS(2700),
    [anon_sym_register] = ACTIONS(2700),
    [anon_sym_inline] = ACTIONS(2700),
    [anon_sym___inline] = ACTIONS(2700),
    [anon_sym___inline__] = ACTIONS(2700),
    [anon_sym___forceinline] = ACTIONS(2700),
    [anon_sym_thread_local] = ACTIONS(2700),
    [anon_sym___thread] = ACTIONS(2700),
    [anon_sym_CG_EXTERN] = ACTIONS(2700),
    [anon_sym_CG_INLINE] = ACTIONS(2700),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2700),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2700),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2700),
    [anon_sym_IBOutlet] = ACTIONS(2700),
    [anon_sym_IBInspectable] = ACTIONS(2700),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2700),
    [anon_sym_NS_INLINE] = ACTIONS(2700),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2700),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2700),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2700),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2700),
    [anon_sym_const] = ACTIONS(2700),
    [anon_sym_constexpr] = ACTIONS(2700),
    [anon_sym_volatile] = ACTIONS(2700),
    [anon_sym_restrict] = ACTIONS(2700),
    [anon_sym___restrict__] = ACTIONS(2700),
    [anon_sym__Atomic] = ACTIONS(2700),
    [anon_sym__Noreturn] = ACTIONS(2700),
    [anon_sym__Nonnull] = ACTIONS(2700),
    [anon_sym_nullable] = ACTIONS(2700),
    [anon_sym__Complex] = ACTIONS(2700),
    [anon_sym__Nullable] = ACTIONS(2700),
    [anon_sym__Nullable_result] = ACTIONS(2700),
    [anon_sym__Null_unspecified] = ACTIONS(2700),
    [anon_sym___autoreleasing] = ACTIONS(2700),
    [anon_sym___block] = ACTIONS(2700),
    [anon_sym___bridge] = ACTIONS(2700),
    [anon_sym___bridge_retained] = ACTIONS(2700),
    [anon_sym___bridge_transfer] = ACTIONS(2700),
    [anon_sym___complex] = ACTIONS(2700),
    [anon_sym___const] = ACTIONS(2700),
    [anon_sym___imag] = ACTIONS(2700),
    [anon_sym___kindof] = ACTIONS(2700),
    [anon_sym___nonnull] = ACTIONS(2700),
    [anon_sym___nullable] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2700),
    [anon_sym___real] = ACTIONS(2700),
    [anon_sym___strong] = ACTIONS(2700),
    [anon_sym___unsafe_unretained] = ACTIONS(2700),
    [anon_sym___unused] = ACTIONS(2700),
    [anon_sym___weak] = ACTIONS(2700),
    [anon_sym_alignas] = ACTIONS(2700),
    [anon_sym__Alignas] = ACTIONS(2700),
    [sym_primitive_type] = ACTIONS(2700),
    [anon_sym_enum] = ACTIONS(2700),
    [anon_sym_struct] = ACTIONS(2700),
    [anon_sym_union] = ACTIONS(2700),
    [anon_sym_if] = ACTIONS(2700),
    [anon_sym_switch] = ACTIONS(2700),
    [anon_sym_case] = ACTIONS(2700),
    [anon_sym_default] = ACTIONS(2700),
    [anon_sym_while] = ACTIONS(2700),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_for] = ACTIONS(2700),
    [anon_sym_in] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_DASH_DASH] = ACTIONS(2702),
    [anon_sym_PLUS_PLUS] = ACTIONS(2702),
    [anon_sym_sizeof] = ACTIONS(2700),
    [anon_sym___alignof__] = ACTIONS(2700),
    [anon_sym___alignof] = ACTIONS(2700),
    [anon_sym__alignof] = ACTIONS(2700),
    [anon_sym_alignof] = ACTIONS(2700),
    [anon_sym__Alignof] = ACTIONS(2700),
    [anon_sym_offsetof] = ACTIONS(2700),
    [anon_sym__Generic] = ACTIONS(2700),
    [anon_sym_asm] = ACTIONS(2700),
    [anon_sym___asm__] = ACTIONS(2700),
    [anon_sym___asm] = ACTIONS(2700),
    [sym_number_literal] = ACTIONS(2702),
    [anon_sym_L_SQUOTE] = ACTIONS(2702),
    [anon_sym_u_SQUOTE] = ACTIONS(2702),
    [anon_sym_U_SQUOTE] = ACTIONS(2702),
    [anon_sym_u8_SQUOTE] = ACTIONS(2702),
    [anon_sym_SQUOTE] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_DQUOTE] = ACTIONS(2702),
    [anon_sym_L_DQUOTE] = ACTIONS(2702),
    [anon_sym_u_DQUOTE] = ACTIONS(2702),
    [anon_sym_U_DQUOTE] = ACTIONS(2702),
    [anon_sym_u8_DQUOTE] = ACTIONS(2702),
    [sym_true] = ACTIONS(2700),
    [sym_false] = ACTIONS(2700),
    [anon_sym_NULL] = ACTIONS(2700),
    [anon_sym_nullptr] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2700),
    [anon_sym___typeof] = ACTIONS(2700),
    [anon_sym_typeof] = ACTIONS(2700),
    [anon_sym_ATimport] = ACTIONS(2702),
    [aux_sym_preproc_undef_token1] = ACTIONS(2700),
    [anon_sym_POUND] = ACTIONS(2700),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2700),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2700),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2700),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2700),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2700),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2700),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2700),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_API_AVAILABLE] = ACTIONS(2700),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_API_DEPRECATED] = ACTIONS(2700),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2700),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym___deprecated_msg] = ACTIONS(2700),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2700),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2700),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2700),
    [anon_sym_ATprotocol] = ACTIONS(2702),
    [anon_sym_ATinterface] = ACTIONS(2702),
    [anon_sym_ATimplementation] = ACTIONS(2702),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2702),
    [anon_sym_ATsynthesize] = ACTIONS(2702),
    [anon_sym_ATdynamic] = ACTIONS(2702),
    [anon_sym_ATtry] = ACTIONS(2702),
    [anon_sym___try] = ACTIONS(2700),
    [anon_sym_ATthrow] = ACTIONS(2702),
    [anon_sym_ATselector] = ACTIONS(2702),
    [anon_sym_ATavailable] = ACTIONS(2702),
    [anon_sym___builtin_available] = ACTIONS(2700),
    [anon_sym_va_arg] = ACTIONS(2700),
    [anon_sym_ATencode] = ACTIONS(2702),
    [anon_sym_ATsynchronized] = ACTIONS(2702),
    [anon_sym_BOOL] = ACTIONS(2700),
    [anon_sym_IMP] = ACTIONS(2700),
    [anon_sym_SEL] = ACTIONS(2700),
    [anon_sym_Class] = ACTIONS(2700),
    [anon_sym_id] = ACTIONS(2700),
  },
  [STATE(478)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token2] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [aux_sym_preproc_else_token1] = ACTIONS(2844),
    [aux_sym_preproc_elif_token1] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATsynthesize] = ACTIONS(2846),
    [anon_sym_ATdynamic] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(479)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token2] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [aux_sym_preproc_else_token1] = ACTIONS(2844),
    [aux_sym_preproc_elif_token1] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATsynthesize] = ACTIONS(2846),
    [anon_sym_ATdynamic] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(480)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token2] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [aux_sym_preproc_else_token1] = ACTIONS(2844),
    [aux_sym_preproc_elif_token1] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATsynthesize] = ACTIONS(2846),
    [anon_sym_ATdynamic] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(481)] = {
    [sym_identifier] = ACTIONS(2848),
    [aux_sym_preproc_include_token1] = ACTIONS(2848),
    [aux_sym_preproc_include_token2] = ACTIONS(2848),
    [aux_sym_preproc_def_token1] = ACTIONS(2848),
    [aux_sym_preproc_if_token1] = ACTIONS(2848),
    [aux_sym_preproc_if_token2] = ACTIONS(2848),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2848),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2848),
    [aux_sym_preproc_else_token1] = ACTIONS(2848),
    [aux_sym_preproc_elif_token1] = ACTIONS(2848),
    [sym_preproc_directive] = ACTIONS(2848),
    [anon_sym_LPAREN2] = ACTIONS(2850),
    [anon_sym_BANG] = ACTIONS(2850),
    [anon_sym_TILDE] = ACTIONS(2850),
    [anon_sym_DASH] = ACTIONS(2848),
    [anon_sym_PLUS] = ACTIONS(2848),
    [anon_sym_STAR] = ACTIONS(2850),
    [anon_sym_CARET] = ACTIONS(2850),
    [anon_sym_AMP] = ACTIONS(2850),
    [anon_sym_SEMI] = ACTIONS(2850),
    [anon_sym___extension__] = ACTIONS(2848),
    [anon_sym_typedef] = ACTIONS(2848),
    [anon_sym_extern] = ACTIONS(2848),
    [anon_sym___attribute__] = ACTIONS(2848),
    [anon_sym___attribute] = ACTIONS(2848),
    [anon_sym_noreturn] = ACTIONS(2848),
    [anon_sym_LBRACK] = ACTIONS(2850),
    [anon_sym___declspec] = ACTIONS(2848),
    [anon_sym___cdecl] = ACTIONS(2848),
    [anon_sym___clrcall] = ACTIONS(2848),
    [anon_sym___stdcall] = ACTIONS(2848),
    [anon_sym___fastcall] = ACTIONS(2848),
    [anon_sym___thiscall] = ACTIONS(2848),
    [anon_sym___vectorcall] = ACTIONS(2848),
    [anon_sym_LBRACE] = ACTIONS(2850),
    [anon_sym_signed] = ACTIONS(2848),
    [anon_sym_unsigned] = ACTIONS(2848),
    [anon_sym_long] = ACTIONS(2848),
    [anon_sym_short] = ACTIONS(2848),
    [anon_sym_static] = ACTIONS(2848),
    [anon_sym_ATautoreleasepool] = ACTIONS(2850),
    [anon_sym_auto] = ACTIONS(2848),
    [anon_sym_register] = ACTIONS(2848),
    [anon_sym_inline] = ACTIONS(2848),
    [anon_sym___inline] = ACTIONS(2848),
    [anon_sym___inline__] = ACTIONS(2848),
    [anon_sym___forceinline] = ACTIONS(2848),
    [anon_sym_thread_local] = ACTIONS(2848),
    [anon_sym___thread] = ACTIONS(2848),
    [anon_sym_CG_EXTERN] = ACTIONS(2848),
    [anon_sym_CG_INLINE] = ACTIONS(2848),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2848),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2848),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2848),
    [anon_sym_IBOutlet] = ACTIONS(2848),
    [anon_sym_IBInspectable] = ACTIONS(2848),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2848),
    [anon_sym_NS_INLINE] = ACTIONS(2848),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2848),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2848),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2848),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2848),
    [anon_sym_const] = ACTIONS(2848),
    [anon_sym_constexpr] = ACTIONS(2848),
    [anon_sym_volatile] = ACTIONS(2848),
    [anon_sym_restrict] = ACTIONS(2848),
    [anon_sym___restrict__] = ACTIONS(2848),
    [anon_sym__Atomic] = ACTIONS(2848),
    [anon_sym__Noreturn] = ACTIONS(2848),
    [anon_sym__Nonnull] = ACTIONS(2848),
    [anon_sym_nullable] = ACTIONS(2848),
    [anon_sym__Complex] = ACTIONS(2848),
    [anon_sym__Nullable] = ACTIONS(2848),
    [anon_sym__Nullable_result] = ACTIONS(2848),
    [anon_sym__Null_unspecified] = ACTIONS(2848),
    [anon_sym___autoreleasing] = ACTIONS(2848),
    [anon_sym___block] = ACTIONS(2848),
    [anon_sym___bridge] = ACTIONS(2848),
    [anon_sym___bridge_retained] = ACTIONS(2848),
    [anon_sym___bridge_transfer] = ACTIONS(2848),
    [anon_sym___complex] = ACTIONS(2848),
    [anon_sym___const] = ACTIONS(2848),
    [anon_sym___imag] = ACTIONS(2848),
    [anon_sym___kindof] = ACTIONS(2848),
    [anon_sym___nonnull] = ACTIONS(2848),
    [anon_sym___nullable] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2848),
    [anon_sym___real] = ACTIONS(2848),
    [anon_sym___strong] = ACTIONS(2848),
    [anon_sym___unsafe_unretained] = ACTIONS(2848),
    [anon_sym___unused] = ACTIONS(2848),
    [anon_sym___weak] = ACTIONS(2848),
    [anon_sym_alignas] = ACTIONS(2848),
    [anon_sym__Alignas] = ACTIONS(2848),
    [sym_primitive_type] = ACTIONS(2848),
    [anon_sym_enum] = ACTIONS(2848),
    [anon_sym_struct] = ACTIONS(2848),
    [anon_sym_union] = ACTIONS(2848),
    [anon_sym_if] = ACTIONS(2848),
    [anon_sym_switch] = ACTIONS(2848),
    [anon_sym_case] = ACTIONS(2848),
    [anon_sym_default] = ACTIONS(2848),
    [anon_sym_while] = ACTIONS(2848),
    [anon_sym_do] = ACTIONS(2848),
    [anon_sym_for] = ACTIONS(2848),
    [anon_sym_in] = ACTIONS(2848),
    [anon_sym_return] = ACTIONS(2848),
    [anon_sym_break] = ACTIONS(2848),
    [anon_sym_continue] = ACTIONS(2848),
    [anon_sym_goto] = ACTIONS(2848),
    [anon_sym_DASH_DASH] = ACTIONS(2850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2850),
    [anon_sym_sizeof] = ACTIONS(2848),
    [anon_sym___alignof__] = ACTIONS(2848),
    [anon_sym___alignof] = ACTIONS(2848),
    [anon_sym__alignof] = ACTIONS(2848),
    [anon_sym_alignof] = ACTIONS(2848),
    [anon_sym__Alignof] = ACTIONS(2848),
    [anon_sym_offsetof] = ACTIONS(2848),
    [anon_sym__Generic] = ACTIONS(2848),
    [anon_sym_asm] = ACTIONS(2848),
    [anon_sym___asm__] = ACTIONS(2848),
    [anon_sym___asm] = ACTIONS(2848),
    [sym_number_literal] = ACTIONS(2850),
    [anon_sym_L_SQUOTE] = ACTIONS(2850),
    [anon_sym_u_SQUOTE] = ACTIONS(2850),
    [anon_sym_U_SQUOTE] = ACTIONS(2850),
    [anon_sym_u8_SQUOTE] = ACTIONS(2850),
    [anon_sym_SQUOTE] = ACTIONS(2850),
    [anon_sym_AT] = ACTIONS(2848),
    [anon_sym_DQUOTE] = ACTIONS(2850),
    [anon_sym_L_DQUOTE] = ACTIONS(2850),
    [anon_sym_u_DQUOTE] = ACTIONS(2850),
    [anon_sym_U_DQUOTE] = ACTIONS(2850),
    [anon_sym_u8_DQUOTE] = ACTIONS(2850),
    [sym_true] = ACTIONS(2848),
    [sym_false] = ACTIONS(2848),
    [anon_sym_NULL] = ACTIONS(2848),
    [anon_sym_nullptr] = ACTIONS(2848),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2848),
    [anon_sym___typeof] = ACTIONS(2848),
    [anon_sym_typeof] = ACTIONS(2848),
    [anon_sym_ATimport] = ACTIONS(2850),
    [aux_sym_preproc_undef_token1] = ACTIONS(2848),
    [anon_sym_POUND] = ACTIONS(2848),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2848),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2848),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2848),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2848),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2848),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2848),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2848),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_API_AVAILABLE] = ACTIONS(2848),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_API_DEPRECATED] = ACTIONS(2848),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2848),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym___deprecated_msg] = ACTIONS(2848),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2848),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2848),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2848),
    [anon_sym_ATprotocol] = ACTIONS(2850),
    [anon_sym_ATinterface] = ACTIONS(2850),
    [anon_sym_ATimplementation] = ACTIONS(2850),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2850),
    [anon_sym_ATsynthesize] = ACTIONS(2850),
    [anon_sym_ATdynamic] = ACTIONS(2850),
    [anon_sym_ATtry] = ACTIONS(2850),
    [anon_sym___try] = ACTIONS(2848),
    [anon_sym_ATthrow] = ACTIONS(2850),
    [anon_sym_ATselector] = ACTIONS(2850),
    [anon_sym_ATavailable] = ACTIONS(2850),
    [anon_sym___builtin_available] = ACTIONS(2848),
    [anon_sym_va_arg] = ACTIONS(2848),
    [anon_sym_ATencode] = ACTIONS(2850),
    [anon_sym_ATsynchronized] = ACTIONS(2850),
    [anon_sym_BOOL] = ACTIONS(2848),
    [anon_sym_IMP] = ACTIONS(2848),
    [anon_sym_SEL] = ACTIONS(2848),
    [anon_sym_Class] = ACTIONS(2848),
    [anon_sym_id] = ACTIONS(2848),
  },
  [STATE(482)] = {
    [sym_identifier] = ACTIONS(2572),
    [aux_sym_preproc_include_token1] = ACTIONS(2572),
    [aux_sym_preproc_include_token2] = ACTIONS(2572),
    [aux_sym_preproc_def_token1] = ACTIONS(2572),
    [aux_sym_preproc_if_token1] = ACTIONS(2572),
    [aux_sym_preproc_if_token2] = ACTIONS(2572),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2572),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2572),
    [aux_sym_preproc_else_token1] = ACTIONS(2572),
    [aux_sym_preproc_elif_token1] = ACTIONS(2572),
    [sym_preproc_directive] = ACTIONS(2572),
    [anon_sym_LPAREN2] = ACTIONS(2574),
    [anon_sym_BANG] = ACTIONS(2574),
    [anon_sym_TILDE] = ACTIONS(2574),
    [anon_sym_DASH] = ACTIONS(2572),
    [anon_sym_PLUS] = ACTIONS(2572),
    [anon_sym_STAR] = ACTIONS(2574),
    [anon_sym_CARET] = ACTIONS(2574),
    [anon_sym_AMP] = ACTIONS(2574),
    [anon_sym_SEMI] = ACTIONS(2574),
    [anon_sym___extension__] = ACTIONS(2572),
    [anon_sym_typedef] = ACTIONS(2572),
    [anon_sym_extern] = ACTIONS(2572),
    [anon_sym___attribute__] = ACTIONS(2572),
    [anon_sym___attribute] = ACTIONS(2572),
    [anon_sym_noreturn] = ACTIONS(2572),
    [anon_sym_LBRACK] = ACTIONS(2574),
    [anon_sym___declspec] = ACTIONS(2572),
    [anon_sym___cdecl] = ACTIONS(2572),
    [anon_sym___clrcall] = ACTIONS(2572),
    [anon_sym___stdcall] = ACTIONS(2572),
    [anon_sym___fastcall] = ACTIONS(2572),
    [anon_sym___thiscall] = ACTIONS(2572),
    [anon_sym___vectorcall] = ACTIONS(2572),
    [anon_sym_LBRACE] = ACTIONS(2574),
    [anon_sym_signed] = ACTIONS(2572),
    [anon_sym_unsigned] = ACTIONS(2572),
    [anon_sym_long] = ACTIONS(2572),
    [anon_sym_short] = ACTIONS(2572),
    [anon_sym_static] = ACTIONS(2572),
    [anon_sym_ATautoreleasepool] = ACTIONS(2574),
    [anon_sym_auto] = ACTIONS(2572),
    [anon_sym_register] = ACTIONS(2572),
    [anon_sym_inline] = ACTIONS(2572),
    [anon_sym___inline] = ACTIONS(2572),
    [anon_sym___inline__] = ACTIONS(2572),
    [anon_sym___forceinline] = ACTIONS(2572),
    [anon_sym_thread_local] = ACTIONS(2572),
    [anon_sym___thread] = ACTIONS(2572),
    [anon_sym_CG_EXTERN] = ACTIONS(2572),
    [anon_sym_CG_INLINE] = ACTIONS(2572),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2572),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2572),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2572),
    [anon_sym_IBOutlet] = ACTIONS(2572),
    [anon_sym_IBInspectable] = ACTIONS(2572),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2572),
    [anon_sym_NS_INLINE] = ACTIONS(2572),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2572),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2572),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2572),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2572),
    [anon_sym_const] = ACTIONS(2572),
    [anon_sym_constexpr] = ACTIONS(2572),
    [anon_sym_volatile] = ACTIONS(2572),
    [anon_sym_restrict] = ACTIONS(2572),
    [anon_sym___restrict__] = ACTIONS(2572),
    [anon_sym__Atomic] = ACTIONS(2572),
    [anon_sym__Noreturn] = ACTIONS(2572),
    [anon_sym__Nonnull] = ACTIONS(2572),
    [anon_sym_nullable] = ACTIONS(2572),
    [anon_sym__Complex] = ACTIONS(2572),
    [anon_sym__Nullable] = ACTIONS(2572),
    [anon_sym__Nullable_result] = ACTIONS(2572),
    [anon_sym__Null_unspecified] = ACTIONS(2572),
    [anon_sym___autoreleasing] = ACTIONS(2572),
    [anon_sym___block] = ACTIONS(2572),
    [anon_sym___bridge] = ACTIONS(2572),
    [anon_sym___bridge_retained] = ACTIONS(2572),
    [anon_sym___bridge_transfer] = ACTIONS(2572),
    [anon_sym___complex] = ACTIONS(2572),
    [anon_sym___const] = ACTIONS(2572),
    [anon_sym___imag] = ACTIONS(2572),
    [anon_sym___kindof] = ACTIONS(2572),
    [anon_sym___nonnull] = ACTIONS(2572),
    [anon_sym___nullable] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2572),
    [anon_sym___real] = ACTIONS(2572),
    [anon_sym___strong] = ACTIONS(2572),
    [anon_sym___unsafe_unretained] = ACTIONS(2572),
    [anon_sym___unused] = ACTIONS(2572),
    [anon_sym___weak] = ACTIONS(2572),
    [anon_sym_alignas] = ACTIONS(2572),
    [anon_sym__Alignas] = ACTIONS(2572),
    [sym_primitive_type] = ACTIONS(2572),
    [anon_sym_enum] = ACTIONS(2572),
    [anon_sym_struct] = ACTIONS(2572),
    [anon_sym_union] = ACTIONS(2572),
    [anon_sym_if] = ACTIONS(2572),
    [anon_sym_switch] = ACTIONS(2572),
    [anon_sym_case] = ACTIONS(2572),
    [anon_sym_default] = ACTIONS(2572),
    [anon_sym_while] = ACTIONS(2572),
    [anon_sym_do] = ACTIONS(2572),
    [anon_sym_for] = ACTIONS(2572),
    [anon_sym_in] = ACTIONS(2572),
    [anon_sym_return] = ACTIONS(2572),
    [anon_sym_break] = ACTIONS(2572),
    [anon_sym_continue] = ACTIONS(2572),
    [anon_sym_goto] = ACTIONS(2572),
    [anon_sym_DASH_DASH] = ACTIONS(2574),
    [anon_sym_PLUS_PLUS] = ACTIONS(2574),
    [anon_sym_sizeof] = ACTIONS(2572),
    [anon_sym___alignof__] = ACTIONS(2572),
    [anon_sym___alignof] = ACTIONS(2572),
    [anon_sym__alignof] = ACTIONS(2572),
    [anon_sym_alignof] = ACTIONS(2572),
    [anon_sym__Alignof] = ACTIONS(2572),
    [anon_sym_offsetof] = ACTIONS(2572),
    [anon_sym__Generic] = ACTIONS(2572),
    [anon_sym_asm] = ACTIONS(2572),
    [anon_sym___asm__] = ACTIONS(2572),
    [anon_sym___asm] = ACTIONS(2572),
    [sym_number_literal] = ACTIONS(2574),
    [anon_sym_L_SQUOTE] = ACTIONS(2574),
    [anon_sym_u_SQUOTE] = ACTIONS(2574),
    [anon_sym_U_SQUOTE] = ACTIONS(2574),
    [anon_sym_u8_SQUOTE] = ACTIONS(2574),
    [anon_sym_SQUOTE] = ACTIONS(2574),
    [anon_sym_AT] = ACTIONS(2572),
    [anon_sym_DQUOTE] = ACTIONS(2574),
    [anon_sym_L_DQUOTE] = ACTIONS(2574),
    [anon_sym_u_DQUOTE] = ACTIONS(2574),
    [anon_sym_U_DQUOTE] = ACTIONS(2574),
    [anon_sym_u8_DQUOTE] = ACTIONS(2574),
    [sym_true] = ACTIONS(2572),
    [sym_false] = ACTIONS(2572),
    [anon_sym_NULL] = ACTIONS(2572),
    [anon_sym_nullptr] = ACTIONS(2572),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2572),
    [anon_sym___typeof] = ACTIONS(2572),
    [anon_sym_typeof] = ACTIONS(2572),
    [anon_sym_ATimport] = ACTIONS(2574),
    [aux_sym_preproc_undef_token1] = ACTIONS(2572),
    [anon_sym_POUND] = ACTIONS(2572),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2572),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2572),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2572),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2572),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2572),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2572),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2572),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_API_AVAILABLE] = ACTIONS(2572),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_API_DEPRECATED] = ACTIONS(2572),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2572),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym___deprecated_msg] = ACTIONS(2572),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2572),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2572),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2572),
    [anon_sym_ATprotocol] = ACTIONS(2574),
    [anon_sym_ATinterface] = ACTIONS(2574),
    [anon_sym_ATimplementation] = ACTIONS(2574),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2574),
    [anon_sym_ATsynthesize] = ACTIONS(2574),
    [anon_sym_ATdynamic] = ACTIONS(2574),
    [anon_sym_ATtry] = ACTIONS(2574),
    [anon_sym___try] = ACTIONS(2572),
    [anon_sym_ATthrow] = ACTIONS(2574),
    [anon_sym_ATselector] = ACTIONS(2574),
    [anon_sym_ATavailable] = ACTIONS(2574),
    [anon_sym___builtin_available] = ACTIONS(2572),
    [anon_sym_va_arg] = ACTIONS(2572),
    [anon_sym_ATencode] = ACTIONS(2574),
    [anon_sym_ATsynchronized] = ACTIONS(2574),
    [anon_sym_BOOL] = ACTIONS(2572),
    [anon_sym_IMP] = ACTIONS(2572),
    [anon_sym_SEL] = ACTIONS(2572),
    [anon_sym_Class] = ACTIONS(2572),
    [anon_sym_id] = ACTIONS(2572),
  },
  [STATE(483)] = {
    [sym_identifier] = ACTIONS(2716),
    [aux_sym_preproc_include_token1] = ACTIONS(2716),
    [aux_sym_preproc_include_token2] = ACTIONS(2716),
    [aux_sym_preproc_def_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token2] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2716),
    [aux_sym_preproc_else_token1] = ACTIONS(2716),
    [aux_sym_preproc_elif_token1] = ACTIONS(2716),
    [sym_preproc_directive] = ACTIONS(2716),
    [anon_sym_LPAREN2] = ACTIONS(2718),
    [anon_sym_BANG] = ACTIONS(2718),
    [anon_sym_TILDE] = ACTIONS(2718),
    [anon_sym_DASH] = ACTIONS(2716),
    [anon_sym_PLUS] = ACTIONS(2716),
    [anon_sym_STAR] = ACTIONS(2718),
    [anon_sym_CARET] = ACTIONS(2718),
    [anon_sym_AMP] = ACTIONS(2718),
    [anon_sym_SEMI] = ACTIONS(2718),
    [anon_sym___extension__] = ACTIONS(2716),
    [anon_sym_typedef] = ACTIONS(2716),
    [anon_sym_extern] = ACTIONS(2716),
    [anon_sym___attribute__] = ACTIONS(2716),
    [anon_sym___attribute] = ACTIONS(2716),
    [anon_sym_noreturn] = ACTIONS(2716),
    [anon_sym_LBRACK] = ACTIONS(2718),
    [anon_sym___declspec] = ACTIONS(2716),
    [anon_sym___cdecl] = ACTIONS(2716),
    [anon_sym___clrcall] = ACTIONS(2716),
    [anon_sym___stdcall] = ACTIONS(2716),
    [anon_sym___fastcall] = ACTIONS(2716),
    [anon_sym___thiscall] = ACTIONS(2716),
    [anon_sym___vectorcall] = ACTIONS(2716),
    [anon_sym_LBRACE] = ACTIONS(2718),
    [anon_sym_signed] = ACTIONS(2716),
    [anon_sym_unsigned] = ACTIONS(2716),
    [anon_sym_long] = ACTIONS(2716),
    [anon_sym_short] = ACTIONS(2716),
    [anon_sym_static] = ACTIONS(2716),
    [anon_sym_ATautoreleasepool] = ACTIONS(2718),
    [anon_sym_auto] = ACTIONS(2716),
    [anon_sym_register] = ACTIONS(2716),
    [anon_sym_inline] = ACTIONS(2716),
    [anon_sym___inline] = ACTIONS(2716),
    [anon_sym___inline__] = ACTIONS(2716),
    [anon_sym___forceinline] = ACTIONS(2716),
    [anon_sym_thread_local] = ACTIONS(2716),
    [anon_sym___thread] = ACTIONS(2716),
    [anon_sym_CG_EXTERN] = ACTIONS(2716),
    [anon_sym_CG_INLINE] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2716),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2716),
    [anon_sym_IBOutlet] = ACTIONS(2716),
    [anon_sym_IBInspectable] = ACTIONS(2716),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2716),
    [anon_sym_NS_INLINE] = ACTIONS(2716),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2716),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2716),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2716),
    [anon_sym_const] = ACTIONS(2716),
    [anon_sym_constexpr] = ACTIONS(2716),
    [anon_sym_volatile] = ACTIONS(2716),
    [anon_sym_restrict] = ACTIONS(2716),
    [anon_sym___restrict__] = ACTIONS(2716),
    [anon_sym__Atomic] = ACTIONS(2716),
    [anon_sym__Noreturn] = ACTIONS(2716),
    [anon_sym__Nonnull] = ACTIONS(2716),
    [anon_sym_nullable] = ACTIONS(2716),
    [anon_sym__Complex] = ACTIONS(2716),
    [anon_sym__Nullable] = ACTIONS(2716),
    [anon_sym__Nullable_result] = ACTIONS(2716),
    [anon_sym__Null_unspecified] = ACTIONS(2716),
    [anon_sym___autoreleasing] = ACTIONS(2716),
    [anon_sym___block] = ACTIONS(2716),
    [anon_sym___bridge] = ACTIONS(2716),
    [anon_sym___bridge_retained] = ACTIONS(2716),
    [anon_sym___bridge_transfer] = ACTIONS(2716),
    [anon_sym___complex] = ACTIONS(2716),
    [anon_sym___const] = ACTIONS(2716),
    [anon_sym___imag] = ACTIONS(2716),
    [anon_sym___kindof] = ACTIONS(2716),
    [anon_sym___nonnull] = ACTIONS(2716),
    [anon_sym___nullable] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2716),
    [anon_sym___real] = ACTIONS(2716),
    [anon_sym___strong] = ACTIONS(2716),
    [anon_sym___unsafe_unretained] = ACTIONS(2716),
    [anon_sym___unused] = ACTIONS(2716),
    [anon_sym___weak] = ACTIONS(2716),
    [anon_sym_alignas] = ACTIONS(2716),
    [anon_sym__Alignas] = ACTIONS(2716),
    [sym_primitive_type] = ACTIONS(2716),
    [anon_sym_enum] = ACTIONS(2716),
    [anon_sym_struct] = ACTIONS(2716),
    [anon_sym_union] = ACTIONS(2716),
    [anon_sym_if] = ACTIONS(2716),
    [anon_sym_switch] = ACTIONS(2716),
    [anon_sym_case] = ACTIONS(2716),
    [anon_sym_default] = ACTIONS(2716),
    [anon_sym_while] = ACTIONS(2716),
    [anon_sym_do] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2716),
    [anon_sym_in] = ACTIONS(2716),
    [anon_sym_return] = ACTIONS(2716),
    [anon_sym_break] = ACTIONS(2716),
    [anon_sym_continue] = ACTIONS(2716),
    [anon_sym_goto] = ACTIONS(2716),
    [anon_sym_DASH_DASH] = ACTIONS(2718),
    [anon_sym_PLUS_PLUS] = ACTIONS(2718),
    [anon_sym_sizeof] = ACTIONS(2716),
    [anon_sym___alignof__] = ACTIONS(2716),
    [anon_sym___alignof] = ACTIONS(2716),
    [anon_sym__alignof] = ACTIONS(2716),
    [anon_sym_alignof] = ACTIONS(2716),
    [anon_sym__Alignof] = ACTIONS(2716),
    [anon_sym_offsetof] = ACTIONS(2716),
    [anon_sym__Generic] = ACTIONS(2716),
    [anon_sym_asm] = ACTIONS(2716),
    [anon_sym___asm__] = ACTIONS(2716),
    [anon_sym___asm] = ACTIONS(2716),
    [sym_number_literal] = ACTIONS(2718),
    [anon_sym_L_SQUOTE] = ACTIONS(2718),
    [anon_sym_u_SQUOTE] = ACTIONS(2718),
    [anon_sym_U_SQUOTE] = ACTIONS(2718),
    [anon_sym_u8_SQUOTE] = ACTIONS(2718),
    [anon_sym_SQUOTE] = ACTIONS(2718),
    [anon_sym_AT] = ACTIONS(2716),
    [anon_sym_DQUOTE] = ACTIONS(2718),
    [anon_sym_L_DQUOTE] = ACTIONS(2718),
    [anon_sym_u_DQUOTE] = ACTIONS(2718),
    [anon_sym_U_DQUOTE] = ACTIONS(2718),
    [anon_sym_u8_DQUOTE] = ACTIONS(2718),
    [sym_true] = ACTIONS(2716),
    [sym_false] = ACTIONS(2716),
    [anon_sym_NULL] = ACTIONS(2716),
    [anon_sym_nullptr] = ACTIONS(2716),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2716),
    [anon_sym___typeof] = ACTIONS(2716),
    [anon_sym_typeof] = ACTIONS(2716),
    [anon_sym_ATimport] = ACTIONS(2718),
    [aux_sym_preproc_undef_token1] = ACTIONS(2716),
    [anon_sym_POUND] = ACTIONS(2716),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2716),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2716),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2716),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_API_AVAILABLE] = ACTIONS(2716),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_API_DEPRECATED] = ACTIONS(2716),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym___deprecated_msg] = ACTIONS(2716),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2716),
    [anon_sym_ATprotocol] = ACTIONS(2718),
    [anon_sym_ATinterface] = ACTIONS(2718),
    [anon_sym_ATimplementation] = ACTIONS(2718),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2718),
    [anon_sym_ATsynthesize] = ACTIONS(2718),
    [anon_sym_ATdynamic] = ACTIONS(2718),
    [anon_sym_ATtry] = ACTIONS(2718),
    [anon_sym___try] = ACTIONS(2716),
    [anon_sym_ATthrow] = ACTIONS(2718),
    [anon_sym_ATselector] = ACTIONS(2718),
    [anon_sym_ATavailable] = ACTIONS(2718),
    [anon_sym___builtin_available] = ACTIONS(2716),
    [anon_sym_va_arg] = ACTIONS(2716),
    [anon_sym_ATencode] = ACTIONS(2718),
    [anon_sym_ATsynchronized] = ACTIONS(2718),
    [anon_sym_BOOL] = ACTIONS(2716),
    [anon_sym_IMP] = ACTIONS(2716),
    [anon_sym_SEL] = ACTIONS(2716),
    [anon_sym_Class] = ACTIONS(2716),
    [anon_sym_id] = ACTIONS(2716),
  },
  [STATE(484)] = {
    [sym_else_clause] = STATE(589),
    [sym_identifier] = ACTIONS(2394),
    [aux_sym_preproc_include_token1] = ACTIONS(2394),
    [aux_sym_preproc_include_token2] = ACTIONS(2394),
    [aux_sym_preproc_def_token1] = ACTIONS(2394),
    [aux_sym_preproc_if_token1] = ACTIONS(2394),
    [aux_sym_preproc_if_token2] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2394),
    [sym_preproc_directive] = ACTIONS(2394),
    [anon_sym_LPAREN2] = ACTIONS(2396),
    [anon_sym_BANG] = ACTIONS(2396),
    [anon_sym_TILDE] = ACTIONS(2396),
    [anon_sym_DASH] = ACTIONS(2394),
    [anon_sym_PLUS] = ACTIONS(2394),
    [anon_sym_STAR] = ACTIONS(2396),
    [anon_sym_CARET] = ACTIONS(2396),
    [anon_sym_AMP] = ACTIONS(2396),
    [anon_sym_SEMI] = ACTIONS(2396),
    [anon_sym___extension__] = ACTIONS(2394),
    [anon_sym_typedef] = ACTIONS(2394),
    [anon_sym_extern] = ACTIONS(2394),
    [anon_sym___attribute__] = ACTIONS(2394),
    [anon_sym___attribute] = ACTIONS(2394),
    [anon_sym_noreturn] = ACTIONS(2394),
    [anon_sym_LBRACK] = ACTIONS(2396),
    [anon_sym___declspec] = ACTIONS(2394),
    [anon_sym___cdecl] = ACTIONS(2394),
    [anon_sym___clrcall] = ACTIONS(2394),
    [anon_sym___stdcall] = ACTIONS(2394),
    [anon_sym___fastcall] = ACTIONS(2394),
    [anon_sym___thiscall] = ACTIONS(2394),
    [anon_sym___vectorcall] = ACTIONS(2394),
    [anon_sym_LBRACE] = ACTIONS(2396),
    [anon_sym_signed] = ACTIONS(2394),
    [anon_sym_unsigned] = ACTIONS(2394),
    [anon_sym_long] = ACTIONS(2394),
    [anon_sym_short] = ACTIONS(2394),
    [anon_sym_static] = ACTIONS(2394),
    [anon_sym_ATautoreleasepool] = ACTIONS(2396),
    [anon_sym_auto] = ACTIONS(2394),
    [anon_sym_register] = ACTIONS(2394),
    [anon_sym_inline] = ACTIONS(2394),
    [anon_sym___inline] = ACTIONS(2394),
    [anon_sym___inline__] = ACTIONS(2394),
    [anon_sym___forceinline] = ACTIONS(2394),
    [anon_sym_thread_local] = ACTIONS(2394),
    [anon_sym___thread] = ACTIONS(2394),
    [anon_sym_CG_EXTERN] = ACTIONS(2394),
    [anon_sym_CG_INLINE] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2394),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2394),
    [anon_sym_IBOutlet] = ACTIONS(2394),
    [anon_sym_IBInspectable] = ACTIONS(2394),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2394),
    [anon_sym_NS_INLINE] = ACTIONS(2394),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2394),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2394),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2394),
    [anon_sym_const] = ACTIONS(2394),
    [anon_sym_constexpr] = ACTIONS(2394),
    [anon_sym_volatile] = ACTIONS(2394),
    [anon_sym_restrict] = ACTIONS(2394),
    [anon_sym___restrict__] = ACTIONS(2394),
    [anon_sym__Atomic] = ACTIONS(2394),
    [anon_sym__Noreturn] = ACTIONS(2394),
    [anon_sym__Nonnull] = ACTIONS(2394),
    [anon_sym_nullable] = ACTIONS(2394),
    [anon_sym__Complex] = ACTIONS(2394),
    [anon_sym__Nullable] = ACTIONS(2394),
    [anon_sym__Nullable_result] = ACTIONS(2394),
    [anon_sym__Null_unspecified] = ACTIONS(2394),
    [anon_sym___autoreleasing] = ACTIONS(2394),
    [anon_sym___block] = ACTIONS(2394),
    [anon_sym___bridge] = ACTIONS(2394),
    [anon_sym___bridge_retained] = ACTIONS(2394),
    [anon_sym___bridge_transfer] = ACTIONS(2394),
    [anon_sym___complex] = ACTIONS(2394),
    [anon_sym___const] = ACTIONS(2394),
    [anon_sym___imag] = ACTIONS(2394),
    [anon_sym___kindof] = ACTIONS(2394),
    [anon_sym___nonnull] = ACTIONS(2394),
    [anon_sym___nullable] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2394),
    [anon_sym___real] = ACTIONS(2394),
    [anon_sym___strong] = ACTIONS(2394),
    [anon_sym___unsafe_unretained] = ACTIONS(2394),
    [anon_sym___unused] = ACTIONS(2394),
    [anon_sym___weak] = ACTIONS(2394),
    [anon_sym_alignas] = ACTIONS(2394),
    [anon_sym__Alignas] = ACTIONS(2394),
    [sym_primitive_type] = ACTIONS(2394),
    [anon_sym_enum] = ACTIONS(2394),
    [anon_sym_struct] = ACTIONS(2394),
    [anon_sym_union] = ACTIONS(2394),
    [anon_sym_if] = ACTIONS(2394),
    [anon_sym_else] = ACTIONS(2876),
    [anon_sym_switch] = ACTIONS(2394),
    [anon_sym_case] = ACTIONS(2394),
    [anon_sym_default] = ACTIONS(2394),
    [anon_sym_while] = ACTIONS(2394),
    [anon_sym_do] = ACTIONS(2394),
    [anon_sym_for] = ACTIONS(2394),
    [anon_sym_in] = ACTIONS(2394),
    [anon_sym_return] = ACTIONS(2394),
    [anon_sym_break] = ACTIONS(2394),
    [anon_sym_continue] = ACTIONS(2394),
    [anon_sym_goto] = ACTIONS(2394),
    [anon_sym_DASH_DASH] = ACTIONS(2396),
    [anon_sym_PLUS_PLUS] = ACTIONS(2396),
    [anon_sym_sizeof] = ACTIONS(2394),
    [anon_sym___alignof__] = ACTIONS(2394),
    [anon_sym___alignof] = ACTIONS(2394),
    [anon_sym__alignof] = ACTIONS(2394),
    [anon_sym_alignof] = ACTIONS(2394),
    [anon_sym__Alignof] = ACTIONS(2394),
    [anon_sym_offsetof] = ACTIONS(2394),
    [anon_sym__Generic] = ACTIONS(2394),
    [anon_sym_asm] = ACTIONS(2394),
    [anon_sym___asm__] = ACTIONS(2394),
    [anon_sym___asm] = ACTIONS(2394),
    [sym_number_literal] = ACTIONS(2396),
    [anon_sym_L_SQUOTE] = ACTIONS(2396),
    [anon_sym_u_SQUOTE] = ACTIONS(2396),
    [anon_sym_U_SQUOTE] = ACTIONS(2396),
    [anon_sym_u8_SQUOTE] = ACTIONS(2396),
    [anon_sym_SQUOTE] = ACTIONS(2396),
    [anon_sym_AT] = ACTIONS(2394),
    [anon_sym_DQUOTE] = ACTIONS(2396),
    [anon_sym_L_DQUOTE] = ACTIONS(2396),
    [anon_sym_u_DQUOTE] = ACTIONS(2396),
    [anon_sym_U_DQUOTE] = ACTIONS(2396),
    [anon_sym_u8_DQUOTE] = ACTIONS(2396),
    [sym_true] = ACTIONS(2394),
    [sym_false] = ACTIONS(2394),
    [anon_sym_NULL] = ACTIONS(2394),
    [anon_sym_nullptr] = ACTIONS(2394),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2394),
    [anon_sym___typeof] = ACTIONS(2394),
    [anon_sym_typeof] = ACTIONS(2394),
    [anon_sym_ATimport] = ACTIONS(2396),
    [aux_sym_preproc_undef_token1] = ACTIONS(2394),
    [anon_sym_POUND] = ACTIONS(2394),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2394),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2394),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2394),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_API_AVAILABLE] = ACTIONS(2394),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_API_DEPRECATED] = ACTIONS(2394),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym___deprecated_msg] = ACTIONS(2394),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2394),
    [anon_sym_ATprotocol] = ACTIONS(2396),
    [anon_sym_ATinterface] = ACTIONS(2396),
    [anon_sym_ATimplementation] = ACTIONS(2396),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2396),
    [anon_sym_ATsynthesize] = ACTIONS(2396),
    [anon_sym_ATdynamic] = ACTIONS(2396),
    [anon_sym_ATtry] = ACTIONS(2396),
    [anon_sym___try] = ACTIONS(2394),
    [anon_sym_ATthrow] = ACTIONS(2396),
    [anon_sym_ATselector] = ACTIONS(2396),
    [anon_sym_ATavailable] = ACTIONS(2396),
    [anon_sym___builtin_available] = ACTIONS(2394),
    [anon_sym_va_arg] = ACTIONS(2394),
    [anon_sym_ATencode] = ACTIONS(2396),
    [anon_sym_ATsynchronized] = ACTIONS(2396),
    [anon_sym_BOOL] = ACTIONS(2394),
    [anon_sym_IMP] = ACTIONS(2394),
    [anon_sym_SEL] = ACTIONS(2394),
    [anon_sym_Class] = ACTIONS(2394),
    [anon_sym_id] = ACTIONS(2394),
  },
  [STATE(485)] = {
    [sym_identifier] = ACTIONS(2576),
    [aux_sym_preproc_include_token1] = ACTIONS(2576),
    [aux_sym_preproc_include_token2] = ACTIONS(2576),
    [aux_sym_preproc_def_token1] = ACTIONS(2576),
    [aux_sym_preproc_if_token1] = ACTIONS(2576),
    [aux_sym_preproc_if_token2] = ACTIONS(2576),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2576),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2576),
    [aux_sym_preproc_else_token1] = ACTIONS(2576),
    [aux_sym_preproc_elif_token1] = ACTIONS(2576),
    [sym_preproc_directive] = ACTIONS(2576),
    [anon_sym_LPAREN2] = ACTIONS(2578),
    [anon_sym_BANG] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2578),
    [anon_sym_DASH] = ACTIONS(2576),
    [anon_sym_PLUS] = ACTIONS(2576),
    [anon_sym_STAR] = ACTIONS(2578),
    [anon_sym_CARET] = ACTIONS(2578),
    [anon_sym_AMP] = ACTIONS(2578),
    [anon_sym_SEMI] = ACTIONS(2578),
    [anon_sym___extension__] = ACTIONS(2576),
    [anon_sym_typedef] = ACTIONS(2576),
    [anon_sym_extern] = ACTIONS(2576),
    [anon_sym___attribute__] = ACTIONS(2576),
    [anon_sym___attribute] = ACTIONS(2576),
    [anon_sym_noreturn] = ACTIONS(2576),
    [anon_sym_LBRACK] = ACTIONS(2578),
    [anon_sym___declspec] = ACTIONS(2576),
    [anon_sym___cdecl] = ACTIONS(2576),
    [anon_sym___clrcall] = ACTIONS(2576),
    [anon_sym___stdcall] = ACTIONS(2576),
    [anon_sym___fastcall] = ACTIONS(2576),
    [anon_sym___thiscall] = ACTIONS(2576),
    [anon_sym___vectorcall] = ACTIONS(2576),
    [anon_sym_LBRACE] = ACTIONS(2578),
    [anon_sym_signed] = ACTIONS(2576),
    [anon_sym_unsigned] = ACTIONS(2576),
    [anon_sym_long] = ACTIONS(2576),
    [anon_sym_short] = ACTIONS(2576),
    [anon_sym_static] = ACTIONS(2576),
    [anon_sym_ATautoreleasepool] = ACTIONS(2578),
    [anon_sym_auto] = ACTIONS(2576),
    [anon_sym_register] = ACTIONS(2576),
    [anon_sym_inline] = ACTIONS(2576),
    [anon_sym___inline] = ACTIONS(2576),
    [anon_sym___inline__] = ACTIONS(2576),
    [anon_sym___forceinline] = ACTIONS(2576),
    [anon_sym_thread_local] = ACTIONS(2576),
    [anon_sym___thread] = ACTIONS(2576),
    [anon_sym_CG_EXTERN] = ACTIONS(2576),
    [anon_sym_CG_INLINE] = ACTIONS(2576),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2576),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2576),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2576),
    [anon_sym_IBOutlet] = ACTIONS(2576),
    [anon_sym_IBInspectable] = ACTIONS(2576),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2576),
    [anon_sym_NS_INLINE] = ACTIONS(2576),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2576),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2576),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2576),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2576),
    [anon_sym_const] = ACTIONS(2576),
    [anon_sym_constexpr] = ACTIONS(2576),
    [anon_sym_volatile] = ACTIONS(2576),
    [anon_sym_restrict] = ACTIONS(2576),
    [anon_sym___restrict__] = ACTIONS(2576),
    [anon_sym__Atomic] = ACTIONS(2576),
    [anon_sym__Noreturn] = ACTIONS(2576),
    [anon_sym__Nonnull] = ACTIONS(2576),
    [anon_sym_nullable] = ACTIONS(2576),
    [anon_sym__Complex] = ACTIONS(2576),
    [anon_sym__Nullable] = ACTIONS(2576),
    [anon_sym__Nullable_result] = ACTIONS(2576),
    [anon_sym__Null_unspecified] = ACTIONS(2576),
    [anon_sym___autoreleasing] = ACTIONS(2576),
    [anon_sym___block] = ACTIONS(2576),
    [anon_sym___bridge] = ACTIONS(2576),
    [anon_sym___bridge_retained] = ACTIONS(2576),
    [anon_sym___bridge_transfer] = ACTIONS(2576),
    [anon_sym___complex] = ACTIONS(2576),
    [anon_sym___const] = ACTIONS(2576),
    [anon_sym___imag] = ACTIONS(2576),
    [anon_sym___kindof] = ACTIONS(2576),
    [anon_sym___nonnull] = ACTIONS(2576),
    [anon_sym___nullable] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2576),
    [anon_sym___real] = ACTIONS(2576),
    [anon_sym___strong] = ACTIONS(2576),
    [anon_sym___unsafe_unretained] = ACTIONS(2576),
    [anon_sym___unused] = ACTIONS(2576),
    [anon_sym___weak] = ACTIONS(2576),
    [anon_sym_alignas] = ACTIONS(2576),
    [anon_sym__Alignas] = ACTIONS(2576),
    [sym_primitive_type] = ACTIONS(2576),
    [anon_sym_enum] = ACTIONS(2576),
    [anon_sym_struct] = ACTIONS(2576),
    [anon_sym_union] = ACTIONS(2576),
    [anon_sym_if] = ACTIONS(2576),
    [anon_sym_switch] = ACTIONS(2576),
    [anon_sym_case] = ACTIONS(2576),
    [anon_sym_default] = ACTIONS(2576),
    [anon_sym_while] = ACTIONS(2576),
    [anon_sym_do] = ACTIONS(2576),
    [anon_sym_for] = ACTIONS(2576),
    [anon_sym_in] = ACTIONS(2576),
    [anon_sym_return] = ACTIONS(2576),
    [anon_sym_break] = ACTIONS(2576),
    [anon_sym_continue] = ACTIONS(2576),
    [anon_sym_goto] = ACTIONS(2576),
    [anon_sym_DASH_DASH] = ACTIONS(2578),
    [anon_sym_PLUS_PLUS] = ACTIONS(2578),
    [anon_sym_sizeof] = ACTIONS(2576),
    [anon_sym___alignof__] = ACTIONS(2576),
    [anon_sym___alignof] = ACTIONS(2576),
    [anon_sym__alignof] = ACTIONS(2576),
    [anon_sym_alignof] = ACTIONS(2576),
    [anon_sym__Alignof] = ACTIONS(2576),
    [anon_sym_offsetof] = ACTIONS(2576),
    [anon_sym__Generic] = ACTIONS(2576),
    [anon_sym_asm] = ACTIONS(2576),
    [anon_sym___asm__] = ACTIONS(2576),
    [anon_sym___asm] = ACTIONS(2576),
    [sym_number_literal] = ACTIONS(2578),
    [anon_sym_L_SQUOTE] = ACTIONS(2578),
    [anon_sym_u_SQUOTE] = ACTIONS(2578),
    [anon_sym_U_SQUOTE] = ACTIONS(2578),
    [anon_sym_u8_SQUOTE] = ACTIONS(2578),
    [anon_sym_SQUOTE] = ACTIONS(2578),
    [anon_sym_AT] = ACTIONS(2576),
    [anon_sym_DQUOTE] = ACTIONS(2578),
    [anon_sym_L_DQUOTE] = ACTIONS(2578),
    [anon_sym_u_DQUOTE] = ACTIONS(2578),
    [anon_sym_U_DQUOTE] = ACTIONS(2578),
    [anon_sym_u8_DQUOTE] = ACTIONS(2578),
    [sym_true] = ACTIONS(2576),
    [sym_false] = ACTIONS(2576),
    [anon_sym_NULL] = ACTIONS(2576),
    [anon_sym_nullptr] = ACTIONS(2576),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2576),
    [anon_sym___typeof] = ACTIONS(2576),
    [anon_sym_typeof] = ACTIONS(2576),
    [anon_sym_ATimport] = ACTIONS(2578),
    [aux_sym_preproc_undef_token1] = ACTIONS(2576),
    [anon_sym_POUND] = ACTIONS(2576),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2576),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2576),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2576),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2576),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2576),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2576),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2576),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_API_AVAILABLE] = ACTIONS(2576),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_API_DEPRECATED] = ACTIONS(2576),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2576),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym___deprecated_msg] = ACTIONS(2576),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2576),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2576),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2576),
    [anon_sym_ATprotocol] = ACTIONS(2578),
    [anon_sym_ATinterface] = ACTIONS(2578),
    [anon_sym_ATimplementation] = ACTIONS(2578),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2578),
    [anon_sym_ATsynthesize] = ACTIONS(2578),
    [anon_sym_ATdynamic] = ACTIONS(2578),
    [anon_sym_ATtry] = ACTIONS(2578),
    [anon_sym___try] = ACTIONS(2576),
    [anon_sym_ATthrow] = ACTIONS(2578),
    [anon_sym_ATselector] = ACTIONS(2578),
    [anon_sym_ATavailable] = ACTIONS(2578),
    [anon_sym___builtin_available] = ACTIONS(2576),
    [anon_sym_va_arg] = ACTIONS(2576),
    [anon_sym_ATencode] = ACTIONS(2578),
    [anon_sym_ATsynchronized] = ACTIONS(2578),
    [anon_sym_BOOL] = ACTIONS(2576),
    [anon_sym_IMP] = ACTIONS(2576),
    [anon_sym_SEL] = ACTIONS(2576),
    [anon_sym_Class] = ACTIONS(2576),
    [anon_sym_id] = ACTIONS(2576),
  },
  [STATE(486)] = {
    [sym_identifier] = ACTIONS(2630),
    [aux_sym_preproc_include_token1] = ACTIONS(2630),
    [aux_sym_preproc_include_token2] = ACTIONS(2630),
    [aux_sym_preproc_def_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token2] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2630),
    [aux_sym_preproc_else_token1] = ACTIONS(2630),
    [aux_sym_preproc_elif_token1] = ACTIONS(2630),
    [sym_preproc_directive] = ACTIONS(2630),
    [anon_sym_LPAREN2] = ACTIONS(2632),
    [anon_sym_BANG] = ACTIONS(2632),
    [anon_sym_TILDE] = ACTIONS(2632),
    [anon_sym_DASH] = ACTIONS(2630),
    [anon_sym_PLUS] = ACTIONS(2630),
    [anon_sym_STAR] = ACTIONS(2632),
    [anon_sym_CARET] = ACTIONS(2632),
    [anon_sym_AMP] = ACTIONS(2632),
    [anon_sym_SEMI] = ACTIONS(2632),
    [anon_sym___extension__] = ACTIONS(2630),
    [anon_sym_typedef] = ACTIONS(2630),
    [anon_sym_extern] = ACTIONS(2630),
    [anon_sym___attribute__] = ACTIONS(2630),
    [anon_sym___attribute] = ACTIONS(2630),
    [anon_sym_noreturn] = ACTIONS(2630),
    [anon_sym_LBRACK] = ACTIONS(2632),
    [anon_sym___declspec] = ACTIONS(2630),
    [anon_sym___cdecl] = ACTIONS(2630),
    [anon_sym___clrcall] = ACTIONS(2630),
    [anon_sym___stdcall] = ACTIONS(2630),
    [anon_sym___fastcall] = ACTIONS(2630),
    [anon_sym___thiscall] = ACTIONS(2630),
    [anon_sym___vectorcall] = ACTIONS(2630),
    [anon_sym_LBRACE] = ACTIONS(2632),
    [anon_sym_signed] = ACTIONS(2630),
    [anon_sym_unsigned] = ACTIONS(2630),
    [anon_sym_long] = ACTIONS(2630),
    [anon_sym_short] = ACTIONS(2630),
    [anon_sym_static] = ACTIONS(2630),
    [anon_sym_ATautoreleasepool] = ACTIONS(2632),
    [anon_sym_auto] = ACTIONS(2630),
    [anon_sym_register] = ACTIONS(2630),
    [anon_sym_inline] = ACTIONS(2630),
    [anon_sym___inline] = ACTIONS(2630),
    [anon_sym___inline__] = ACTIONS(2630),
    [anon_sym___forceinline] = ACTIONS(2630),
    [anon_sym_thread_local] = ACTIONS(2630),
    [anon_sym___thread] = ACTIONS(2630),
    [anon_sym_CG_EXTERN] = ACTIONS(2630),
    [anon_sym_CG_INLINE] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2630),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2630),
    [anon_sym_IBOutlet] = ACTIONS(2630),
    [anon_sym_IBInspectable] = ACTIONS(2630),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2630),
    [anon_sym_NS_INLINE] = ACTIONS(2630),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2630),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2630),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2630),
    [anon_sym_const] = ACTIONS(2630),
    [anon_sym_constexpr] = ACTIONS(2630),
    [anon_sym_volatile] = ACTIONS(2630),
    [anon_sym_restrict] = ACTIONS(2630),
    [anon_sym___restrict__] = ACTIONS(2630),
    [anon_sym__Atomic] = ACTIONS(2630),
    [anon_sym__Noreturn] = ACTIONS(2630),
    [anon_sym__Nonnull] = ACTIONS(2630),
    [anon_sym_nullable] = ACTIONS(2630),
    [anon_sym__Complex] = ACTIONS(2630),
    [anon_sym__Nullable] = ACTIONS(2630),
    [anon_sym__Nullable_result] = ACTIONS(2630),
    [anon_sym__Null_unspecified] = ACTIONS(2630),
    [anon_sym___autoreleasing] = ACTIONS(2630),
    [anon_sym___block] = ACTIONS(2630),
    [anon_sym___bridge] = ACTIONS(2630),
    [anon_sym___bridge_retained] = ACTIONS(2630),
    [anon_sym___bridge_transfer] = ACTIONS(2630),
    [anon_sym___complex] = ACTIONS(2630),
    [anon_sym___const] = ACTIONS(2630),
    [anon_sym___imag] = ACTIONS(2630),
    [anon_sym___kindof] = ACTIONS(2630),
    [anon_sym___nonnull] = ACTIONS(2630),
    [anon_sym___nullable] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2630),
    [anon_sym___real] = ACTIONS(2630),
    [anon_sym___strong] = ACTIONS(2630),
    [anon_sym___unsafe_unretained] = ACTIONS(2630),
    [anon_sym___unused] = ACTIONS(2630),
    [anon_sym___weak] = ACTIONS(2630),
    [anon_sym_alignas] = ACTIONS(2630),
    [anon_sym__Alignas] = ACTIONS(2630),
    [sym_primitive_type] = ACTIONS(2630),
    [anon_sym_enum] = ACTIONS(2630),
    [anon_sym_struct] = ACTIONS(2630),
    [anon_sym_union] = ACTIONS(2630),
    [anon_sym_if] = ACTIONS(2630),
    [anon_sym_switch] = ACTIONS(2630),
    [anon_sym_case] = ACTIONS(2630),
    [anon_sym_default] = ACTIONS(2630),
    [anon_sym_while] = ACTIONS(2630),
    [anon_sym_do] = ACTIONS(2630),
    [anon_sym_for] = ACTIONS(2630),
    [anon_sym_in] = ACTIONS(2630),
    [anon_sym_return] = ACTIONS(2630),
    [anon_sym_break] = ACTIONS(2630),
    [anon_sym_continue] = ACTIONS(2630),
    [anon_sym_goto] = ACTIONS(2630),
    [anon_sym_DASH_DASH] = ACTIONS(2632),
    [anon_sym_PLUS_PLUS] = ACTIONS(2632),
    [anon_sym_sizeof] = ACTIONS(2630),
    [anon_sym___alignof__] = ACTIONS(2630),
    [anon_sym___alignof] = ACTIONS(2630),
    [anon_sym__alignof] = ACTIONS(2630),
    [anon_sym_alignof] = ACTIONS(2630),
    [anon_sym__Alignof] = ACTIONS(2630),
    [anon_sym_offsetof] = ACTIONS(2630),
    [anon_sym__Generic] = ACTIONS(2630),
    [anon_sym_asm] = ACTIONS(2630),
    [anon_sym___asm__] = ACTIONS(2630),
    [anon_sym___asm] = ACTIONS(2630),
    [sym_number_literal] = ACTIONS(2632),
    [anon_sym_L_SQUOTE] = ACTIONS(2632),
    [anon_sym_u_SQUOTE] = ACTIONS(2632),
    [anon_sym_U_SQUOTE] = ACTIONS(2632),
    [anon_sym_u8_SQUOTE] = ACTIONS(2632),
    [anon_sym_SQUOTE] = ACTIONS(2632),
    [anon_sym_AT] = ACTIONS(2630),
    [anon_sym_DQUOTE] = ACTIONS(2632),
    [anon_sym_L_DQUOTE] = ACTIONS(2632),
    [anon_sym_u_DQUOTE] = ACTIONS(2632),
    [anon_sym_U_DQUOTE] = ACTIONS(2632),
    [anon_sym_u8_DQUOTE] = ACTIONS(2632),
    [sym_true] = ACTIONS(2630),
    [sym_false] = ACTIONS(2630),
    [anon_sym_NULL] = ACTIONS(2630),
    [anon_sym_nullptr] = ACTIONS(2630),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2630),
    [anon_sym___typeof] = ACTIONS(2630),
    [anon_sym_typeof] = ACTIONS(2630),
    [anon_sym_ATimport] = ACTIONS(2632),
    [aux_sym_preproc_undef_token1] = ACTIONS(2630),
    [anon_sym_POUND] = ACTIONS(2630),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2630),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2630),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2630),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_API_AVAILABLE] = ACTIONS(2630),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_API_DEPRECATED] = ACTIONS(2630),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym___deprecated_msg] = ACTIONS(2630),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2630),
    [anon_sym_ATprotocol] = ACTIONS(2632),
    [anon_sym_ATinterface] = ACTIONS(2632),
    [anon_sym_ATimplementation] = ACTIONS(2632),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2632),
    [anon_sym_ATsynthesize] = ACTIONS(2632),
    [anon_sym_ATdynamic] = ACTIONS(2632),
    [anon_sym_ATtry] = ACTIONS(2632),
    [anon_sym___try] = ACTIONS(2630),
    [anon_sym_ATthrow] = ACTIONS(2632),
    [anon_sym_ATselector] = ACTIONS(2632),
    [anon_sym_ATavailable] = ACTIONS(2632),
    [anon_sym___builtin_available] = ACTIONS(2630),
    [anon_sym_va_arg] = ACTIONS(2630),
    [anon_sym_ATencode] = ACTIONS(2632),
    [anon_sym_ATsynchronized] = ACTIONS(2632),
    [anon_sym_BOOL] = ACTIONS(2630),
    [anon_sym_IMP] = ACTIONS(2630),
    [anon_sym_SEL] = ACTIONS(2630),
    [anon_sym_Class] = ACTIONS(2630),
    [anon_sym_id] = ACTIONS(2630),
  },
  [STATE(487)] = {
    [sym_identifier] = ACTIONS(2634),
    [aux_sym_preproc_include_token1] = ACTIONS(2634),
    [aux_sym_preproc_include_token2] = ACTIONS(2634),
    [aux_sym_preproc_def_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token2] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2634),
    [aux_sym_preproc_else_token1] = ACTIONS(2634),
    [aux_sym_preproc_elif_token1] = ACTIONS(2634),
    [sym_preproc_directive] = ACTIONS(2634),
    [anon_sym_LPAREN2] = ACTIONS(2636),
    [anon_sym_BANG] = ACTIONS(2636),
    [anon_sym_TILDE] = ACTIONS(2636),
    [anon_sym_DASH] = ACTIONS(2634),
    [anon_sym_PLUS] = ACTIONS(2634),
    [anon_sym_STAR] = ACTIONS(2636),
    [anon_sym_CARET] = ACTIONS(2636),
    [anon_sym_AMP] = ACTIONS(2636),
    [anon_sym_SEMI] = ACTIONS(2636),
    [anon_sym___extension__] = ACTIONS(2634),
    [anon_sym_typedef] = ACTIONS(2634),
    [anon_sym_extern] = ACTIONS(2634),
    [anon_sym___attribute__] = ACTIONS(2634),
    [anon_sym___attribute] = ACTIONS(2634),
    [anon_sym_noreturn] = ACTIONS(2634),
    [anon_sym_LBRACK] = ACTIONS(2636),
    [anon_sym___declspec] = ACTIONS(2634),
    [anon_sym___cdecl] = ACTIONS(2634),
    [anon_sym___clrcall] = ACTIONS(2634),
    [anon_sym___stdcall] = ACTIONS(2634),
    [anon_sym___fastcall] = ACTIONS(2634),
    [anon_sym___thiscall] = ACTIONS(2634),
    [anon_sym___vectorcall] = ACTIONS(2634),
    [anon_sym_LBRACE] = ACTIONS(2636),
    [anon_sym_signed] = ACTIONS(2634),
    [anon_sym_unsigned] = ACTIONS(2634),
    [anon_sym_long] = ACTIONS(2634),
    [anon_sym_short] = ACTIONS(2634),
    [anon_sym_static] = ACTIONS(2634),
    [anon_sym_ATautoreleasepool] = ACTIONS(2636),
    [anon_sym_auto] = ACTIONS(2634),
    [anon_sym_register] = ACTIONS(2634),
    [anon_sym_inline] = ACTIONS(2634),
    [anon_sym___inline] = ACTIONS(2634),
    [anon_sym___inline__] = ACTIONS(2634),
    [anon_sym___forceinline] = ACTIONS(2634),
    [anon_sym_thread_local] = ACTIONS(2634),
    [anon_sym___thread] = ACTIONS(2634),
    [anon_sym_CG_EXTERN] = ACTIONS(2634),
    [anon_sym_CG_INLINE] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2634),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2634),
    [anon_sym_IBOutlet] = ACTIONS(2634),
    [anon_sym_IBInspectable] = ACTIONS(2634),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2634),
    [anon_sym_NS_INLINE] = ACTIONS(2634),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2634),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2634),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2634),
    [anon_sym_const] = ACTIONS(2634),
    [anon_sym_constexpr] = ACTIONS(2634),
    [anon_sym_volatile] = ACTIONS(2634),
    [anon_sym_restrict] = ACTIONS(2634),
    [anon_sym___restrict__] = ACTIONS(2634),
    [anon_sym__Atomic] = ACTIONS(2634),
    [anon_sym__Noreturn] = ACTIONS(2634),
    [anon_sym__Nonnull] = ACTIONS(2634),
    [anon_sym_nullable] = ACTIONS(2634),
    [anon_sym__Complex] = ACTIONS(2634),
    [anon_sym__Nullable] = ACTIONS(2634),
    [anon_sym__Nullable_result] = ACTIONS(2634),
    [anon_sym__Null_unspecified] = ACTIONS(2634),
    [anon_sym___autoreleasing] = ACTIONS(2634),
    [anon_sym___block] = ACTIONS(2634),
    [anon_sym___bridge] = ACTIONS(2634),
    [anon_sym___bridge_retained] = ACTIONS(2634),
    [anon_sym___bridge_transfer] = ACTIONS(2634),
    [anon_sym___complex] = ACTIONS(2634),
    [anon_sym___const] = ACTIONS(2634),
    [anon_sym___imag] = ACTIONS(2634),
    [anon_sym___kindof] = ACTIONS(2634),
    [anon_sym___nonnull] = ACTIONS(2634),
    [anon_sym___nullable] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2634),
    [anon_sym___real] = ACTIONS(2634),
    [anon_sym___strong] = ACTIONS(2634),
    [anon_sym___unsafe_unretained] = ACTIONS(2634),
    [anon_sym___unused] = ACTIONS(2634),
    [anon_sym___weak] = ACTIONS(2634),
    [anon_sym_alignas] = ACTIONS(2634),
    [anon_sym__Alignas] = ACTIONS(2634),
    [sym_primitive_type] = ACTIONS(2634),
    [anon_sym_enum] = ACTIONS(2634),
    [anon_sym_struct] = ACTIONS(2634),
    [anon_sym_union] = ACTIONS(2634),
    [anon_sym_if] = ACTIONS(2634),
    [anon_sym_switch] = ACTIONS(2634),
    [anon_sym_case] = ACTIONS(2634),
    [anon_sym_default] = ACTIONS(2634),
    [anon_sym_while] = ACTIONS(2634),
    [anon_sym_do] = ACTIONS(2634),
    [anon_sym_for] = ACTIONS(2634),
    [anon_sym_in] = ACTIONS(2634),
    [anon_sym_return] = ACTIONS(2634),
    [anon_sym_break] = ACTIONS(2634),
    [anon_sym_continue] = ACTIONS(2634),
    [anon_sym_goto] = ACTIONS(2634),
    [anon_sym_DASH_DASH] = ACTIONS(2636),
    [anon_sym_PLUS_PLUS] = ACTIONS(2636),
    [anon_sym_sizeof] = ACTIONS(2634),
    [anon_sym___alignof__] = ACTIONS(2634),
    [anon_sym___alignof] = ACTIONS(2634),
    [anon_sym__alignof] = ACTIONS(2634),
    [anon_sym_alignof] = ACTIONS(2634),
    [anon_sym__Alignof] = ACTIONS(2634),
    [anon_sym_offsetof] = ACTIONS(2634),
    [anon_sym__Generic] = ACTIONS(2634),
    [anon_sym_asm] = ACTIONS(2634),
    [anon_sym___asm__] = ACTIONS(2634),
    [anon_sym___asm] = ACTIONS(2634),
    [sym_number_literal] = ACTIONS(2636),
    [anon_sym_L_SQUOTE] = ACTIONS(2636),
    [anon_sym_u_SQUOTE] = ACTIONS(2636),
    [anon_sym_U_SQUOTE] = ACTIONS(2636),
    [anon_sym_u8_SQUOTE] = ACTIONS(2636),
    [anon_sym_SQUOTE] = ACTIONS(2636),
    [anon_sym_AT] = ACTIONS(2634),
    [anon_sym_DQUOTE] = ACTIONS(2636),
    [anon_sym_L_DQUOTE] = ACTIONS(2636),
    [anon_sym_u_DQUOTE] = ACTIONS(2636),
    [anon_sym_U_DQUOTE] = ACTIONS(2636),
    [anon_sym_u8_DQUOTE] = ACTIONS(2636),
    [sym_true] = ACTIONS(2634),
    [sym_false] = ACTIONS(2634),
    [anon_sym_NULL] = ACTIONS(2634),
    [anon_sym_nullptr] = ACTIONS(2634),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2634),
    [anon_sym___typeof] = ACTIONS(2634),
    [anon_sym_typeof] = ACTIONS(2634),
    [anon_sym_ATimport] = ACTIONS(2636),
    [aux_sym_preproc_undef_token1] = ACTIONS(2634),
    [anon_sym_POUND] = ACTIONS(2634),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2634),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2634),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2634),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_API_AVAILABLE] = ACTIONS(2634),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_API_DEPRECATED] = ACTIONS(2634),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym___deprecated_msg] = ACTIONS(2634),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2634),
    [anon_sym_ATprotocol] = ACTIONS(2636),
    [anon_sym_ATinterface] = ACTIONS(2636),
    [anon_sym_ATimplementation] = ACTIONS(2636),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2636),
    [anon_sym_ATsynthesize] = ACTIONS(2636),
    [anon_sym_ATdynamic] = ACTIONS(2636),
    [anon_sym_ATtry] = ACTIONS(2636),
    [anon_sym___try] = ACTIONS(2634),
    [anon_sym_ATthrow] = ACTIONS(2636),
    [anon_sym_ATselector] = ACTIONS(2636),
    [anon_sym_ATavailable] = ACTIONS(2636),
    [anon_sym___builtin_available] = ACTIONS(2634),
    [anon_sym_va_arg] = ACTIONS(2634),
    [anon_sym_ATencode] = ACTIONS(2636),
    [anon_sym_ATsynchronized] = ACTIONS(2636),
    [anon_sym_BOOL] = ACTIONS(2634),
    [anon_sym_IMP] = ACTIONS(2634),
    [anon_sym_SEL] = ACTIONS(2634),
    [anon_sym_Class] = ACTIONS(2634),
    [anon_sym_id] = ACTIONS(2634),
  },
  [STATE(488)] = {
    [sym_identifier] = ACTIONS(2672),
    [aux_sym_preproc_include_token1] = ACTIONS(2672),
    [aux_sym_preproc_include_token2] = ACTIONS(2672),
    [aux_sym_preproc_def_token1] = ACTIONS(2672),
    [aux_sym_preproc_if_token1] = ACTIONS(2672),
    [aux_sym_preproc_if_token2] = ACTIONS(2672),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2672),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2672),
    [aux_sym_preproc_else_token1] = ACTIONS(2672),
    [aux_sym_preproc_elif_token1] = ACTIONS(2672),
    [sym_preproc_directive] = ACTIONS(2672),
    [anon_sym_LPAREN2] = ACTIONS(2674),
    [anon_sym_BANG] = ACTIONS(2674),
    [anon_sym_TILDE] = ACTIONS(2674),
    [anon_sym_DASH] = ACTIONS(2672),
    [anon_sym_PLUS] = ACTIONS(2672),
    [anon_sym_STAR] = ACTIONS(2674),
    [anon_sym_CARET] = ACTIONS(2674),
    [anon_sym_AMP] = ACTIONS(2674),
    [anon_sym_SEMI] = ACTIONS(2674),
    [anon_sym___extension__] = ACTIONS(2672),
    [anon_sym_typedef] = ACTIONS(2672),
    [anon_sym_extern] = ACTIONS(2672),
    [anon_sym___attribute__] = ACTIONS(2672),
    [anon_sym___attribute] = ACTIONS(2672),
    [anon_sym_noreturn] = ACTIONS(2672),
    [anon_sym_LBRACK] = ACTIONS(2674),
    [anon_sym___declspec] = ACTIONS(2672),
    [anon_sym___cdecl] = ACTIONS(2672),
    [anon_sym___clrcall] = ACTIONS(2672),
    [anon_sym___stdcall] = ACTIONS(2672),
    [anon_sym___fastcall] = ACTIONS(2672),
    [anon_sym___thiscall] = ACTIONS(2672),
    [anon_sym___vectorcall] = ACTIONS(2672),
    [anon_sym_LBRACE] = ACTIONS(2674),
    [anon_sym_signed] = ACTIONS(2672),
    [anon_sym_unsigned] = ACTIONS(2672),
    [anon_sym_long] = ACTIONS(2672),
    [anon_sym_short] = ACTIONS(2672),
    [anon_sym_static] = ACTIONS(2672),
    [anon_sym_ATautoreleasepool] = ACTIONS(2674),
    [anon_sym_auto] = ACTIONS(2672),
    [anon_sym_register] = ACTIONS(2672),
    [anon_sym_inline] = ACTIONS(2672),
    [anon_sym___inline] = ACTIONS(2672),
    [anon_sym___inline__] = ACTIONS(2672),
    [anon_sym___forceinline] = ACTIONS(2672),
    [anon_sym_thread_local] = ACTIONS(2672),
    [anon_sym___thread] = ACTIONS(2672),
    [anon_sym_CG_EXTERN] = ACTIONS(2672),
    [anon_sym_CG_INLINE] = ACTIONS(2672),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2672),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2672),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2672),
    [anon_sym_IBOutlet] = ACTIONS(2672),
    [anon_sym_IBInspectable] = ACTIONS(2672),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2672),
    [anon_sym_NS_INLINE] = ACTIONS(2672),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2672),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2672),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2672),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2672),
    [anon_sym_const] = ACTIONS(2672),
    [anon_sym_constexpr] = ACTIONS(2672),
    [anon_sym_volatile] = ACTIONS(2672),
    [anon_sym_restrict] = ACTIONS(2672),
    [anon_sym___restrict__] = ACTIONS(2672),
    [anon_sym__Atomic] = ACTIONS(2672),
    [anon_sym__Noreturn] = ACTIONS(2672),
    [anon_sym__Nonnull] = ACTIONS(2672),
    [anon_sym_nullable] = ACTIONS(2672),
    [anon_sym__Complex] = ACTIONS(2672),
    [anon_sym__Nullable] = ACTIONS(2672),
    [anon_sym__Nullable_result] = ACTIONS(2672),
    [anon_sym__Null_unspecified] = ACTIONS(2672),
    [anon_sym___autoreleasing] = ACTIONS(2672),
    [anon_sym___block] = ACTIONS(2672),
    [anon_sym___bridge] = ACTIONS(2672),
    [anon_sym___bridge_retained] = ACTIONS(2672),
    [anon_sym___bridge_transfer] = ACTIONS(2672),
    [anon_sym___complex] = ACTIONS(2672),
    [anon_sym___const] = ACTIONS(2672),
    [anon_sym___imag] = ACTIONS(2672),
    [anon_sym___kindof] = ACTIONS(2672),
    [anon_sym___nonnull] = ACTIONS(2672),
    [anon_sym___nullable] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2672),
    [anon_sym___real] = ACTIONS(2672),
    [anon_sym___strong] = ACTIONS(2672),
    [anon_sym___unsafe_unretained] = ACTIONS(2672),
    [anon_sym___unused] = ACTIONS(2672),
    [anon_sym___weak] = ACTIONS(2672),
    [anon_sym_alignas] = ACTIONS(2672),
    [anon_sym__Alignas] = ACTIONS(2672),
    [sym_primitive_type] = ACTIONS(2672),
    [anon_sym_enum] = ACTIONS(2672),
    [anon_sym_struct] = ACTIONS(2672),
    [anon_sym_union] = ACTIONS(2672),
    [anon_sym_if] = ACTIONS(2672),
    [anon_sym_switch] = ACTIONS(2672),
    [anon_sym_case] = ACTIONS(2672),
    [anon_sym_default] = ACTIONS(2672),
    [anon_sym_while] = ACTIONS(2672),
    [anon_sym_do] = ACTIONS(2672),
    [anon_sym_for] = ACTIONS(2672),
    [anon_sym_in] = ACTIONS(2672),
    [anon_sym_return] = ACTIONS(2672),
    [anon_sym_break] = ACTIONS(2672),
    [anon_sym_continue] = ACTIONS(2672),
    [anon_sym_goto] = ACTIONS(2672),
    [anon_sym_DASH_DASH] = ACTIONS(2674),
    [anon_sym_PLUS_PLUS] = ACTIONS(2674),
    [anon_sym_sizeof] = ACTIONS(2672),
    [anon_sym___alignof__] = ACTIONS(2672),
    [anon_sym___alignof] = ACTIONS(2672),
    [anon_sym__alignof] = ACTIONS(2672),
    [anon_sym_alignof] = ACTIONS(2672),
    [anon_sym__Alignof] = ACTIONS(2672),
    [anon_sym_offsetof] = ACTIONS(2672),
    [anon_sym__Generic] = ACTIONS(2672),
    [anon_sym_asm] = ACTIONS(2672),
    [anon_sym___asm__] = ACTIONS(2672),
    [anon_sym___asm] = ACTIONS(2672),
    [sym_number_literal] = ACTIONS(2674),
    [anon_sym_L_SQUOTE] = ACTIONS(2674),
    [anon_sym_u_SQUOTE] = ACTIONS(2674),
    [anon_sym_U_SQUOTE] = ACTIONS(2674),
    [anon_sym_u8_SQUOTE] = ACTIONS(2674),
    [anon_sym_SQUOTE] = ACTIONS(2674),
    [anon_sym_AT] = ACTIONS(2672),
    [anon_sym_DQUOTE] = ACTIONS(2674),
    [anon_sym_L_DQUOTE] = ACTIONS(2674),
    [anon_sym_u_DQUOTE] = ACTIONS(2674),
    [anon_sym_U_DQUOTE] = ACTIONS(2674),
    [anon_sym_u8_DQUOTE] = ACTIONS(2674),
    [sym_true] = ACTIONS(2672),
    [sym_false] = ACTIONS(2672),
    [anon_sym_NULL] = ACTIONS(2672),
    [anon_sym_nullptr] = ACTIONS(2672),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2672),
    [anon_sym___typeof] = ACTIONS(2672),
    [anon_sym_typeof] = ACTIONS(2672),
    [anon_sym_ATimport] = ACTIONS(2674),
    [aux_sym_preproc_undef_token1] = ACTIONS(2672),
    [anon_sym_POUND] = ACTIONS(2672),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2672),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2672),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2672),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2672),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2672),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2672),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2672),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_API_AVAILABLE] = ACTIONS(2672),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_API_DEPRECATED] = ACTIONS(2672),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2672),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym___deprecated_msg] = ACTIONS(2672),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2672),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2672),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2672),
    [anon_sym_ATprotocol] = ACTIONS(2674),
    [anon_sym_ATinterface] = ACTIONS(2674),
    [anon_sym_ATimplementation] = ACTIONS(2674),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2674),
    [anon_sym_ATsynthesize] = ACTIONS(2674),
    [anon_sym_ATdynamic] = ACTIONS(2674),
    [anon_sym_ATtry] = ACTIONS(2674),
    [anon_sym___try] = ACTIONS(2672),
    [anon_sym_ATthrow] = ACTIONS(2674),
    [anon_sym_ATselector] = ACTIONS(2674),
    [anon_sym_ATavailable] = ACTIONS(2674),
    [anon_sym___builtin_available] = ACTIONS(2672),
    [anon_sym_va_arg] = ACTIONS(2672),
    [anon_sym_ATencode] = ACTIONS(2674),
    [anon_sym_ATsynchronized] = ACTIONS(2674),
    [anon_sym_BOOL] = ACTIONS(2672),
    [anon_sym_IMP] = ACTIONS(2672),
    [anon_sym_SEL] = ACTIONS(2672),
    [anon_sym_Class] = ACTIONS(2672),
    [anon_sym_id] = ACTIONS(2672),
  },
  [STATE(489)] = {
    [sym_identifier] = ACTIONS(2592),
    [aux_sym_preproc_include_token1] = ACTIONS(2592),
    [aux_sym_preproc_include_token2] = ACTIONS(2592),
    [aux_sym_preproc_def_token1] = ACTIONS(2592),
    [aux_sym_preproc_if_token1] = ACTIONS(2592),
    [aux_sym_preproc_if_token2] = ACTIONS(2592),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2592),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2592),
    [aux_sym_preproc_else_token1] = ACTIONS(2592),
    [aux_sym_preproc_elif_token1] = ACTIONS(2592),
    [sym_preproc_directive] = ACTIONS(2592),
    [anon_sym_LPAREN2] = ACTIONS(2594),
    [anon_sym_BANG] = ACTIONS(2594),
    [anon_sym_TILDE] = ACTIONS(2594),
    [anon_sym_DASH] = ACTIONS(2592),
    [anon_sym_PLUS] = ACTIONS(2592),
    [anon_sym_STAR] = ACTIONS(2594),
    [anon_sym_CARET] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(2594),
    [anon_sym_SEMI] = ACTIONS(2594),
    [anon_sym___extension__] = ACTIONS(2592),
    [anon_sym_typedef] = ACTIONS(2592),
    [anon_sym_extern] = ACTIONS(2592),
    [anon_sym___attribute__] = ACTIONS(2592),
    [anon_sym___attribute] = ACTIONS(2592),
    [anon_sym_noreturn] = ACTIONS(2592),
    [anon_sym_LBRACK] = ACTIONS(2594),
    [anon_sym___declspec] = ACTIONS(2592),
    [anon_sym___cdecl] = ACTIONS(2592),
    [anon_sym___clrcall] = ACTIONS(2592),
    [anon_sym___stdcall] = ACTIONS(2592),
    [anon_sym___fastcall] = ACTIONS(2592),
    [anon_sym___thiscall] = ACTIONS(2592),
    [anon_sym___vectorcall] = ACTIONS(2592),
    [anon_sym_LBRACE] = ACTIONS(2594),
    [anon_sym_signed] = ACTIONS(2592),
    [anon_sym_unsigned] = ACTIONS(2592),
    [anon_sym_long] = ACTIONS(2592),
    [anon_sym_short] = ACTIONS(2592),
    [anon_sym_static] = ACTIONS(2592),
    [anon_sym_ATautoreleasepool] = ACTIONS(2594),
    [anon_sym_auto] = ACTIONS(2592),
    [anon_sym_register] = ACTIONS(2592),
    [anon_sym_inline] = ACTIONS(2592),
    [anon_sym___inline] = ACTIONS(2592),
    [anon_sym___inline__] = ACTIONS(2592),
    [anon_sym___forceinline] = ACTIONS(2592),
    [anon_sym_thread_local] = ACTIONS(2592),
    [anon_sym___thread] = ACTIONS(2592),
    [anon_sym_CG_EXTERN] = ACTIONS(2592),
    [anon_sym_CG_INLINE] = ACTIONS(2592),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2592),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2592),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2592),
    [anon_sym_IBOutlet] = ACTIONS(2592),
    [anon_sym_IBInspectable] = ACTIONS(2592),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2592),
    [anon_sym_NS_INLINE] = ACTIONS(2592),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2592),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2592),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2592),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2592),
    [anon_sym_const] = ACTIONS(2592),
    [anon_sym_constexpr] = ACTIONS(2592),
    [anon_sym_volatile] = ACTIONS(2592),
    [anon_sym_restrict] = ACTIONS(2592),
    [anon_sym___restrict__] = ACTIONS(2592),
    [anon_sym__Atomic] = ACTIONS(2592),
    [anon_sym__Noreturn] = ACTIONS(2592),
    [anon_sym__Nonnull] = ACTIONS(2592),
    [anon_sym_nullable] = ACTIONS(2592),
    [anon_sym__Complex] = ACTIONS(2592),
    [anon_sym__Nullable] = ACTIONS(2592),
    [anon_sym__Nullable_result] = ACTIONS(2592),
    [anon_sym__Null_unspecified] = ACTIONS(2592),
    [anon_sym___autoreleasing] = ACTIONS(2592),
    [anon_sym___block] = ACTIONS(2592),
    [anon_sym___bridge] = ACTIONS(2592),
    [anon_sym___bridge_retained] = ACTIONS(2592),
    [anon_sym___bridge_transfer] = ACTIONS(2592),
    [anon_sym___complex] = ACTIONS(2592),
    [anon_sym___const] = ACTIONS(2592),
    [anon_sym___imag] = ACTIONS(2592),
    [anon_sym___kindof] = ACTIONS(2592),
    [anon_sym___nonnull] = ACTIONS(2592),
    [anon_sym___nullable] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2592),
    [anon_sym___real] = ACTIONS(2592),
    [anon_sym___strong] = ACTIONS(2592),
    [anon_sym___unsafe_unretained] = ACTIONS(2592),
    [anon_sym___unused] = ACTIONS(2592),
    [anon_sym___weak] = ACTIONS(2592),
    [anon_sym_alignas] = ACTIONS(2592),
    [anon_sym__Alignas] = ACTIONS(2592),
    [sym_primitive_type] = ACTIONS(2592),
    [anon_sym_enum] = ACTIONS(2592),
    [anon_sym_struct] = ACTIONS(2592),
    [anon_sym_union] = ACTIONS(2592),
    [anon_sym_if] = ACTIONS(2592),
    [anon_sym_switch] = ACTIONS(2592),
    [anon_sym_case] = ACTIONS(2592),
    [anon_sym_default] = ACTIONS(2592),
    [anon_sym_while] = ACTIONS(2592),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_for] = ACTIONS(2592),
    [anon_sym_in] = ACTIONS(2592),
    [anon_sym_return] = ACTIONS(2592),
    [anon_sym_break] = ACTIONS(2592),
    [anon_sym_continue] = ACTIONS(2592),
    [anon_sym_goto] = ACTIONS(2592),
    [anon_sym_DASH_DASH] = ACTIONS(2594),
    [anon_sym_PLUS_PLUS] = ACTIONS(2594),
    [anon_sym_sizeof] = ACTIONS(2592),
    [anon_sym___alignof__] = ACTIONS(2592),
    [anon_sym___alignof] = ACTIONS(2592),
    [anon_sym__alignof] = ACTIONS(2592),
    [anon_sym_alignof] = ACTIONS(2592),
    [anon_sym__Alignof] = ACTIONS(2592),
    [anon_sym_offsetof] = ACTIONS(2592),
    [anon_sym__Generic] = ACTIONS(2592),
    [anon_sym_asm] = ACTIONS(2592),
    [anon_sym___asm__] = ACTIONS(2592),
    [anon_sym___asm] = ACTIONS(2592),
    [sym_number_literal] = ACTIONS(2594),
    [anon_sym_L_SQUOTE] = ACTIONS(2594),
    [anon_sym_u_SQUOTE] = ACTIONS(2594),
    [anon_sym_U_SQUOTE] = ACTIONS(2594),
    [anon_sym_u8_SQUOTE] = ACTIONS(2594),
    [anon_sym_SQUOTE] = ACTIONS(2594),
    [anon_sym_AT] = ACTIONS(2592),
    [anon_sym_DQUOTE] = ACTIONS(2594),
    [anon_sym_L_DQUOTE] = ACTIONS(2594),
    [anon_sym_u_DQUOTE] = ACTIONS(2594),
    [anon_sym_U_DQUOTE] = ACTIONS(2594),
    [anon_sym_u8_DQUOTE] = ACTIONS(2594),
    [sym_true] = ACTIONS(2592),
    [sym_false] = ACTIONS(2592),
    [anon_sym_NULL] = ACTIONS(2592),
    [anon_sym_nullptr] = ACTIONS(2592),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2592),
    [anon_sym___typeof] = ACTIONS(2592),
    [anon_sym_typeof] = ACTIONS(2592),
    [anon_sym_ATimport] = ACTIONS(2594),
    [aux_sym_preproc_undef_token1] = ACTIONS(2592),
    [anon_sym_POUND] = ACTIONS(2592),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2592),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2592),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2592),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2592),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2592),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2592),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2592),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_API_AVAILABLE] = ACTIONS(2592),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_API_DEPRECATED] = ACTIONS(2592),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2592),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym___deprecated_msg] = ACTIONS(2592),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2592),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2592),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2592),
    [anon_sym_ATprotocol] = ACTIONS(2594),
    [anon_sym_ATinterface] = ACTIONS(2594),
    [anon_sym_ATimplementation] = ACTIONS(2594),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2594),
    [anon_sym_ATsynthesize] = ACTIONS(2594),
    [anon_sym_ATdynamic] = ACTIONS(2594),
    [anon_sym_ATtry] = ACTIONS(2594),
    [anon_sym___try] = ACTIONS(2592),
    [anon_sym_ATthrow] = ACTIONS(2594),
    [anon_sym_ATselector] = ACTIONS(2594),
    [anon_sym_ATavailable] = ACTIONS(2594),
    [anon_sym___builtin_available] = ACTIONS(2592),
    [anon_sym_va_arg] = ACTIONS(2592),
    [anon_sym_ATencode] = ACTIONS(2594),
    [anon_sym_ATsynchronized] = ACTIONS(2594),
    [anon_sym_BOOL] = ACTIONS(2592),
    [anon_sym_IMP] = ACTIONS(2592),
    [anon_sym_SEL] = ACTIONS(2592),
    [anon_sym_Class] = ACTIONS(2592),
    [anon_sym_id] = ACTIONS(2592),
  },
  [STATE(490)] = {
    [sym_identifier] = ACTIONS(2860),
    [aux_sym_preproc_include_token1] = ACTIONS(2860),
    [aux_sym_preproc_include_token2] = ACTIONS(2860),
    [aux_sym_preproc_def_token1] = ACTIONS(2860),
    [aux_sym_preproc_if_token1] = ACTIONS(2860),
    [aux_sym_preproc_if_token2] = ACTIONS(2860),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2860),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2860),
    [aux_sym_preproc_else_token1] = ACTIONS(2860),
    [aux_sym_preproc_elif_token1] = ACTIONS(2860),
    [sym_preproc_directive] = ACTIONS(2860),
    [anon_sym_LPAREN2] = ACTIONS(2862),
    [anon_sym_BANG] = ACTIONS(2862),
    [anon_sym_TILDE] = ACTIONS(2862),
    [anon_sym_DASH] = ACTIONS(2860),
    [anon_sym_PLUS] = ACTIONS(2860),
    [anon_sym_STAR] = ACTIONS(2862),
    [anon_sym_CARET] = ACTIONS(2862),
    [anon_sym_AMP] = ACTIONS(2862),
    [anon_sym_SEMI] = ACTIONS(2862),
    [anon_sym___extension__] = ACTIONS(2860),
    [anon_sym_typedef] = ACTIONS(2860),
    [anon_sym_extern] = ACTIONS(2860),
    [anon_sym___attribute__] = ACTIONS(2860),
    [anon_sym___attribute] = ACTIONS(2860),
    [anon_sym_noreturn] = ACTIONS(2860),
    [anon_sym_LBRACK] = ACTIONS(2862),
    [anon_sym___declspec] = ACTIONS(2860),
    [anon_sym___cdecl] = ACTIONS(2860),
    [anon_sym___clrcall] = ACTIONS(2860),
    [anon_sym___stdcall] = ACTIONS(2860),
    [anon_sym___fastcall] = ACTIONS(2860),
    [anon_sym___thiscall] = ACTIONS(2860),
    [anon_sym___vectorcall] = ACTIONS(2860),
    [anon_sym_LBRACE] = ACTIONS(2862),
    [anon_sym_signed] = ACTIONS(2860),
    [anon_sym_unsigned] = ACTIONS(2860),
    [anon_sym_long] = ACTIONS(2860),
    [anon_sym_short] = ACTIONS(2860),
    [anon_sym_static] = ACTIONS(2860),
    [anon_sym_ATautoreleasepool] = ACTIONS(2862),
    [anon_sym_auto] = ACTIONS(2860),
    [anon_sym_register] = ACTIONS(2860),
    [anon_sym_inline] = ACTIONS(2860),
    [anon_sym___inline] = ACTIONS(2860),
    [anon_sym___inline__] = ACTIONS(2860),
    [anon_sym___forceinline] = ACTIONS(2860),
    [anon_sym_thread_local] = ACTIONS(2860),
    [anon_sym___thread] = ACTIONS(2860),
    [anon_sym_CG_EXTERN] = ACTIONS(2860),
    [anon_sym_CG_INLINE] = ACTIONS(2860),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2860),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2860),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2860),
    [anon_sym_IBOutlet] = ACTIONS(2860),
    [anon_sym_IBInspectable] = ACTIONS(2860),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2860),
    [anon_sym_NS_INLINE] = ACTIONS(2860),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2860),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2860),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2860),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2860),
    [anon_sym_const] = ACTIONS(2860),
    [anon_sym_constexpr] = ACTIONS(2860),
    [anon_sym_volatile] = ACTIONS(2860),
    [anon_sym_restrict] = ACTIONS(2860),
    [anon_sym___restrict__] = ACTIONS(2860),
    [anon_sym__Atomic] = ACTIONS(2860),
    [anon_sym__Noreturn] = ACTIONS(2860),
    [anon_sym__Nonnull] = ACTIONS(2860),
    [anon_sym_nullable] = ACTIONS(2860),
    [anon_sym__Complex] = ACTIONS(2860),
    [anon_sym__Nullable] = ACTIONS(2860),
    [anon_sym__Nullable_result] = ACTIONS(2860),
    [anon_sym__Null_unspecified] = ACTIONS(2860),
    [anon_sym___autoreleasing] = ACTIONS(2860),
    [anon_sym___block] = ACTIONS(2860),
    [anon_sym___bridge] = ACTIONS(2860),
    [anon_sym___bridge_retained] = ACTIONS(2860),
    [anon_sym___bridge_transfer] = ACTIONS(2860),
    [anon_sym___complex] = ACTIONS(2860),
    [anon_sym___const] = ACTIONS(2860),
    [anon_sym___imag] = ACTIONS(2860),
    [anon_sym___kindof] = ACTIONS(2860),
    [anon_sym___nonnull] = ACTIONS(2860),
    [anon_sym___nullable] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2860),
    [anon_sym___real] = ACTIONS(2860),
    [anon_sym___strong] = ACTIONS(2860),
    [anon_sym___unsafe_unretained] = ACTIONS(2860),
    [anon_sym___unused] = ACTIONS(2860),
    [anon_sym___weak] = ACTIONS(2860),
    [anon_sym_alignas] = ACTIONS(2860),
    [anon_sym__Alignas] = ACTIONS(2860),
    [sym_primitive_type] = ACTIONS(2860),
    [anon_sym_enum] = ACTIONS(2860),
    [anon_sym_struct] = ACTIONS(2860),
    [anon_sym_union] = ACTIONS(2860),
    [anon_sym_if] = ACTIONS(2860),
    [anon_sym_switch] = ACTIONS(2860),
    [anon_sym_case] = ACTIONS(2860),
    [anon_sym_default] = ACTIONS(2860),
    [anon_sym_while] = ACTIONS(2860),
    [anon_sym_do] = ACTIONS(2860),
    [anon_sym_for] = ACTIONS(2860),
    [anon_sym_in] = ACTIONS(2860),
    [anon_sym_return] = ACTIONS(2860),
    [anon_sym_break] = ACTIONS(2860),
    [anon_sym_continue] = ACTIONS(2860),
    [anon_sym_goto] = ACTIONS(2860),
    [anon_sym_DASH_DASH] = ACTIONS(2862),
    [anon_sym_PLUS_PLUS] = ACTIONS(2862),
    [anon_sym_sizeof] = ACTIONS(2860),
    [anon_sym___alignof__] = ACTIONS(2860),
    [anon_sym___alignof] = ACTIONS(2860),
    [anon_sym__alignof] = ACTIONS(2860),
    [anon_sym_alignof] = ACTIONS(2860),
    [anon_sym__Alignof] = ACTIONS(2860),
    [anon_sym_offsetof] = ACTIONS(2860),
    [anon_sym__Generic] = ACTIONS(2860),
    [anon_sym_asm] = ACTIONS(2860),
    [anon_sym___asm__] = ACTIONS(2860),
    [anon_sym___asm] = ACTIONS(2860),
    [sym_number_literal] = ACTIONS(2862),
    [anon_sym_L_SQUOTE] = ACTIONS(2862),
    [anon_sym_u_SQUOTE] = ACTIONS(2862),
    [anon_sym_U_SQUOTE] = ACTIONS(2862),
    [anon_sym_u8_SQUOTE] = ACTIONS(2862),
    [anon_sym_SQUOTE] = ACTIONS(2862),
    [anon_sym_AT] = ACTIONS(2860),
    [anon_sym_DQUOTE] = ACTIONS(2862),
    [anon_sym_L_DQUOTE] = ACTIONS(2862),
    [anon_sym_u_DQUOTE] = ACTIONS(2862),
    [anon_sym_U_DQUOTE] = ACTIONS(2862),
    [anon_sym_u8_DQUOTE] = ACTIONS(2862),
    [sym_true] = ACTIONS(2860),
    [sym_false] = ACTIONS(2860),
    [anon_sym_NULL] = ACTIONS(2860),
    [anon_sym_nullptr] = ACTIONS(2860),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2860),
    [anon_sym___typeof] = ACTIONS(2860),
    [anon_sym_typeof] = ACTIONS(2860),
    [anon_sym_ATimport] = ACTIONS(2862),
    [aux_sym_preproc_undef_token1] = ACTIONS(2860),
    [anon_sym_POUND] = ACTIONS(2860),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2860),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2860),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2860),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2860),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2860),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2860),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2860),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_API_AVAILABLE] = ACTIONS(2860),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_API_DEPRECATED] = ACTIONS(2860),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2860),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym___deprecated_msg] = ACTIONS(2860),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2860),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2860),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2860),
    [anon_sym_ATprotocol] = ACTIONS(2862),
    [anon_sym_ATinterface] = ACTIONS(2862),
    [anon_sym_ATimplementation] = ACTIONS(2862),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2862),
    [anon_sym_ATsynthesize] = ACTIONS(2862),
    [anon_sym_ATdynamic] = ACTIONS(2862),
    [anon_sym_ATtry] = ACTIONS(2862),
    [anon_sym___try] = ACTIONS(2860),
    [anon_sym_ATthrow] = ACTIONS(2862),
    [anon_sym_ATselector] = ACTIONS(2862),
    [anon_sym_ATavailable] = ACTIONS(2862),
    [anon_sym___builtin_available] = ACTIONS(2860),
    [anon_sym_va_arg] = ACTIONS(2860),
    [anon_sym_ATencode] = ACTIONS(2862),
    [anon_sym_ATsynchronized] = ACTIONS(2862),
    [anon_sym_BOOL] = ACTIONS(2860),
    [anon_sym_IMP] = ACTIONS(2860),
    [anon_sym_SEL] = ACTIONS(2860),
    [anon_sym_Class] = ACTIONS(2860),
    [anon_sym_id] = ACTIONS(2860),
  },
  [STATE(491)] = {
    [sym_identifier] = ACTIONS(2580),
    [aux_sym_preproc_include_token1] = ACTIONS(2580),
    [aux_sym_preproc_include_token2] = ACTIONS(2580),
    [aux_sym_preproc_def_token1] = ACTIONS(2580),
    [aux_sym_preproc_if_token1] = ACTIONS(2580),
    [aux_sym_preproc_if_token2] = ACTIONS(2580),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2580),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2580),
    [aux_sym_preproc_else_token1] = ACTIONS(2580),
    [aux_sym_preproc_elif_token1] = ACTIONS(2580),
    [sym_preproc_directive] = ACTIONS(2580),
    [anon_sym_LPAREN2] = ACTIONS(2582),
    [anon_sym_BANG] = ACTIONS(2582),
    [anon_sym_TILDE] = ACTIONS(2582),
    [anon_sym_DASH] = ACTIONS(2580),
    [anon_sym_PLUS] = ACTIONS(2580),
    [anon_sym_STAR] = ACTIONS(2582),
    [anon_sym_CARET] = ACTIONS(2582),
    [anon_sym_AMP] = ACTIONS(2582),
    [anon_sym_SEMI] = ACTIONS(2582),
    [anon_sym___extension__] = ACTIONS(2580),
    [anon_sym_typedef] = ACTIONS(2580),
    [anon_sym_extern] = ACTIONS(2580),
    [anon_sym___attribute__] = ACTIONS(2580),
    [anon_sym___attribute] = ACTIONS(2580),
    [anon_sym_noreturn] = ACTIONS(2580),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym___declspec] = ACTIONS(2580),
    [anon_sym___cdecl] = ACTIONS(2580),
    [anon_sym___clrcall] = ACTIONS(2580),
    [anon_sym___stdcall] = ACTIONS(2580),
    [anon_sym___fastcall] = ACTIONS(2580),
    [anon_sym___thiscall] = ACTIONS(2580),
    [anon_sym___vectorcall] = ACTIONS(2580),
    [anon_sym_LBRACE] = ACTIONS(2582),
    [anon_sym_signed] = ACTIONS(2580),
    [anon_sym_unsigned] = ACTIONS(2580),
    [anon_sym_long] = ACTIONS(2580),
    [anon_sym_short] = ACTIONS(2580),
    [anon_sym_static] = ACTIONS(2580),
    [anon_sym_ATautoreleasepool] = ACTIONS(2582),
    [anon_sym_auto] = ACTIONS(2580),
    [anon_sym_register] = ACTIONS(2580),
    [anon_sym_inline] = ACTIONS(2580),
    [anon_sym___inline] = ACTIONS(2580),
    [anon_sym___inline__] = ACTIONS(2580),
    [anon_sym___forceinline] = ACTIONS(2580),
    [anon_sym_thread_local] = ACTIONS(2580),
    [anon_sym___thread] = ACTIONS(2580),
    [anon_sym_CG_EXTERN] = ACTIONS(2580),
    [anon_sym_CG_INLINE] = ACTIONS(2580),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2580),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2580),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2580),
    [anon_sym_IBOutlet] = ACTIONS(2580),
    [anon_sym_IBInspectable] = ACTIONS(2580),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2580),
    [anon_sym_NS_INLINE] = ACTIONS(2580),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2580),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2580),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2580),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2580),
    [anon_sym_const] = ACTIONS(2580),
    [anon_sym_constexpr] = ACTIONS(2580),
    [anon_sym_volatile] = ACTIONS(2580),
    [anon_sym_restrict] = ACTIONS(2580),
    [anon_sym___restrict__] = ACTIONS(2580),
    [anon_sym__Atomic] = ACTIONS(2580),
    [anon_sym__Noreturn] = ACTIONS(2580),
    [anon_sym__Nonnull] = ACTIONS(2580),
    [anon_sym_nullable] = ACTIONS(2580),
    [anon_sym__Complex] = ACTIONS(2580),
    [anon_sym__Nullable] = ACTIONS(2580),
    [anon_sym__Nullable_result] = ACTIONS(2580),
    [anon_sym__Null_unspecified] = ACTIONS(2580),
    [anon_sym___autoreleasing] = ACTIONS(2580),
    [anon_sym___block] = ACTIONS(2580),
    [anon_sym___bridge] = ACTIONS(2580),
    [anon_sym___bridge_retained] = ACTIONS(2580),
    [anon_sym___bridge_transfer] = ACTIONS(2580),
    [anon_sym___complex] = ACTIONS(2580),
    [anon_sym___const] = ACTIONS(2580),
    [anon_sym___imag] = ACTIONS(2580),
    [anon_sym___kindof] = ACTIONS(2580),
    [anon_sym___nonnull] = ACTIONS(2580),
    [anon_sym___nullable] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2580),
    [anon_sym___real] = ACTIONS(2580),
    [anon_sym___strong] = ACTIONS(2580),
    [anon_sym___unsafe_unretained] = ACTIONS(2580),
    [anon_sym___unused] = ACTIONS(2580),
    [anon_sym___weak] = ACTIONS(2580),
    [anon_sym_alignas] = ACTIONS(2580),
    [anon_sym__Alignas] = ACTIONS(2580),
    [sym_primitive_type] = ACTIONS(2580),
    [anon_sym_enum] = ACTIONS(2580),
    [anon_sym_struct] = ACTIONS(2580),
    [anon_sym_union] = ACTIONS(2580),
    [anon_sym_if] = ACTIONS(2580),
    [anon_sym_switch] = ACTIONS(2580),
    [anon_sym_case] = ACTIONS(2580),
    [anon_sym_default] = ACTIONS(2580),
    [anon_sym_while] = ACTIONS(2580),
    [anon_sym_do] = ACTIONS(2580),
    [anon_sym_for] = ACTIONS(2580),
    [anon_sym_in] = ACTIONS(2580),
    [anon_sym_return] = ACTIONS(2580),
    [anon_sym_break] = ACTIONS(2580),
    [anon_sym_continue] = ACTIONS(2580),
    [anon_sym_goto] = ACTIONS(2580),
    [anon_sym_DASH_DASH] = ACTIONS(2582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2582),
    [anon_sym_sizeof] = ACTIONS(2580),
    [anon_sym___alignof__] = ACTIONS(2580),
    [anon_sym___alignof] = ACTIONS(2580),
    [anon_sym__alignof] = ACTIONS(2580),
    [anon_sym_alignof] = ACTIONS(2580),
    [anon_sym__Alignof] = ACTIONS(2580),
    [anon_sym_offsetof] = ACTIONS(2580),
    [anon_sym__Generic] = ACTIONS(2580),
    [anon_sym_asm] = ACTIONS(2580),
    [anon_sym___asm__] = ACTIONS(2580),
    [anon_sym___asm] = ACTIONS(2580),
    [sym_number_literal] = ACTIONS(2582),
    [anon_sym_L_SQUOTE] = ACTIONS(2582),
    [anon_sym_u_SQUOTE] = ACTIONS(2582),
    [anon_sym_U_SQUOTE] = ACTIONS(2582),
    [anon_sym_u8_SQUOTE] = ACTIONS(2582),
    [anon_sym_SQUOTE] = ACTIONS(2582),
    [anon_sym_AT] = ACTIONS(2580),
    [anon_sym_DQUOTE] = ACTIONS(2582),
    [anon_sym_L_DQUOTE] = ACTIONS(2582),
    [anon_sym_u_DQUOTE] = ACTIONS(2582),
    [anon_sym_U_DQUOTE] = ACTIONS(2582),
    [anon_sym_u8_DQUOTE] = ACTIONS(2582),
    [sym_true] = ACTIONS(2580),
    [sym_false] = ACTIONS(2580),
    [anon_sym_NULL] = ACTIONS(2580),
    [anon_sym_nullptr] = ACTIONS(2580),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2580),
    [anon_sym___typeof] = ACTIONS(2580),
    [anon_sym_typeof] = ACTIONS(2580),
    [anon_sym_ATimport] = ACTIONS(2582),
    [aux_sym_preproc_undef_token1] = ACTIONS(2580),
    [anon_sym_POUND] = ACTIONS(2580),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2580),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2580),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2580),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2580),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2580),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2580),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2580),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_API_AVAILABLE] = ACTIONS(2580),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_API_DEPRECATED] = ACTIONS(2580),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2580),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym___deprecated_msg] = ACTIONS(2580),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2580),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2580),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2580),
    [anon_sym_ATprotocol] = ACTIONS(2582),
    [anon_sym_ATinterface] = ACTIONS(2582),
    [anon_sym_ATimplementation] = ACTIONS(2582),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2582),
    [anon_sym_ATsynthesize] = ACTIONS(2582),
    [anon_sym_ATdynamic] = ACTIONS(2582),
    [anon_sym_ATtry] = ACTIONS(2582),
    [anon_sym___try] = ACTIONS(2580),
    [anon_sym_ATthrow] = ACTIONS(2582),
    [anon_sym_ATselector] = ACTIONS(2582),
    [anon_sym_ATavailable] = ACTIONS(2582),
    [anon_sym___builtin_available] = ACTIONS(2580),
    [anon_sym_va_arg] = ACTIONS(2580),
    [anon_sym_ATencode] = ACTIONS(2582),
    [anon_sym_ATsynchronized] = ACTIONS(2582),
    [anon_sym_BOOL] = ACTIONS(2580),
    [anon_sym_IMP] = ACTIONS(2580),
    [anon_sym_SEL] = ACTIONS(2580),
    [anon_sym_Class] = ACTIONS(2580),
    [anon_sym_id] = ACTIONS(2580),
  },
  [STATE(492)] = {
    [sym_identifier] = ACTIONS(2604),
    [aux_sym_preproc_include_token1] = ACTIONS(2604),
    [aux_sym_preproc_include_token2] = ACTIONS(2604),
    [aux_sym_preproc_def_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token2] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2604),
    [aux_sym_preproc_else_token1] = ACTIONS(2604),
    [aux_sym_preproc_elif_token1] = ACTIONS(2604),
    [sym_preproc_directive] = ACTIONS(2604),
    [anon_sym_LPAREN2] = ACTIONS(2606),
    [anon_sym_BANG] = ACTIONS(2606),
    [anon_sym_TILDE] = ACTIONS(2606),
    [anon_sym_DASH] = ACTIONS(2604),
    [anon_sym_PLUS] = ACTIONS(2604),
    [anon_sym_STAR] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(2606),
    [anon_sym_AMP] = ACTIONS(2606),
    [anon_sym_SEMI] = ACTIONS(2606),
    [anon_sym___extension__] = ACTIONS(2604),
    [anon_sym_typedef] = ACTIONS(2604),
    [anon_sym_extern] = ACTIONS(2604),
    [anon_sym___attribute__] = ACTIONS(2604),
    [anon_sym___attribute] = ACTIONS(2604),
    [anon_sym_noreturn] = ACTIONS(2604),
    [anon_sym_LBRACK] = ACTIONS(2606),
    [anon_sym___declspec] = ACTIONS(2604),
    [anon_sym___cdecl] = ACTIONS(2604),
    [anon_sym___clrcall] = ACTIONS(2604),
    [anon_sym___stdcall] = ACTIONS(2604),
    [anon_sym___fastcall] = ACTIONS(2604),
    [anon_sym___thiscall] = ACTIONS(2604),
    [anon_sym___vectorcall] = ACTIONS(2604),
    [anon_sym_LBRACE] = ACTIONS(2606),
    [anon_sym_signed] = ACTIONS(2604),
    [anon_sym_unsigned] = ACTIONS(2604),
    [anon_sym_long] = ACTIONS(2604),
    [anon_sym_short] = ACTIONS(2604),
    [anon_sym_static] = ACTIONS(2604),
    [anon_sym_ATautoreleasepool] = ACTIONS(2606),
    [anon_sym_auto] = ACTIONS(2604),
    [anon_sym_register] = ACTIONS(2604),
    [anon_sym_inline] = ACTIONS(2604),
    [anon_sym___inline] = ACTIONS(2604),
    [anon_sym___inline__] = ACTIONS(2604),
    [anon_sym___forceinline] = ACTIONS(2604),
    [anon_sym_thread_local] = ACTIONS(2604),
    [anon_sym___thread] = ACTIONS(2604),
    [anon_sym_CG_EXTERN] = ACTIONS(2604),
    [anon_sym_CG_INLINE] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2604),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2604),
    [anon_sym_IBOutlet] = ACTIONS(2604),
    [anon_sym_IBInspectable] = ACTIONS(2604),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2604),
    [anon_sym_NS_INLINE] = ACTIONS(2604),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2604),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2604),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2604),
    [anon_sym_const] = ACTIONS(2604),
    [anon_sym_constexpr] = ACTIONS(2604),
    [anon_sym_volatile] = ACTIONS(2604),
    [anon_sym_restrict] = ACTIONS(2604),
    [anon_sym___restrict__] = ACTIONS(2604),
    [anon_sym__Atomic] = ACTIONS(2604),
    [anon_sym__Noreturn] = ACTIONS(2604),
    [anon_sym__Nonnull] = ACTIONS(2604),
    [anon_sym_nullable] = ACTIONS(2604),
    [anon_sym__Complex] = ACTIONS(2604),
    [anon_sym__Nullable] = ACTIONS(2604),
    [anon_sym__Nullable_result] = ACTIONS(2604),
    [anon_sym__Null_unspecified] = ACTIONS(2604),
    [anon_sym___autoreleasing] = ACTIONS(2604),
    [anon_sym___block] = ACTIONS(2604),
    [anon_sym___bridge] = ACTIONS(2604),
    [anon_sym___bridge_retained] = ACTIONS(2604),
    [anon_sym___bridge_transfer] = ACTIONS(2604),
    [anon_sym___complex] = ACTIONS(2604),
    [anon_sym___const] = ACTIONS(2604),
    [anon_sym___imag] = ACTIONS(2604),
    [anon_sym___kindof] = ACTIONS(2604),
    [anon_sym___nonnull] = ACTIONS(2604),
    [anon_sym___nullable] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2604),
    [anon_sym___real] = ACTIONS(2604),
    [anon_sym___strong] = ACTIONS(2604),
    [anon_sym___unsafe_unretained] = ACTIONS(2604),
    [anon_sym___unused] = ACTIONS(2604),
    [anon_sym___weak] = ACTIONS(2604),
    [anon_sym_alignas] = ACTIONS(2604),
    [anon_sym__Alignas] = ACTIONS(2604),
    [sym_primitive_type] = ACTIONS(2604),
    [anon_sym_enum] = ACTIONS(2604),
    [anon_sym_struct] = ACTIONS(2604),
    [anon_sym_union] = ACTIONS(2604),
    [anon_sym_if] = ACTIONS(2604),
    [anon_sym_switch] = ACTIONS(2604),
    [anon_sym_case] = ACTIONS(2604),
    [anon_sym_default] = ACTIONS(2604),
    [anon_sym_while] = ACTIONS(2604),
    [anon_sym_do] = ACTIONS(2604),
    [anon_sym_for] = ACTIONS(2604),
    [anon_sym_in] = ACTIONS(2604),
    [anon_sym_return] = ACTIONS(2604),
    [anon_sym_break] = ACTIONS(2604),
    [anon_sym_continue] = ACTIONS(2604),
    [anon_sym_goto] = ACTIONS(2604),
    [anon_sym_DASH_DASH] = ACTIONS(2606),
    [anon_sym_PLUS_PLUS] = ACTIONS(2606),
    [anon_sym_sizeof] = ACTIONS(2604),
    [anon_sym___alignof__] = ACTIONS(2604),
    [anon_sym___alignof] = ACTIONS(2604),
    [anon_sym__alignof] = ACTIONS(2604),
    [anon_sym_alignof] = ACTIONS(2604),
    [anon_sym__Alignof] = ACTIONS(2604),
    [anon_sym_offsetof] = ACTIONS(2604),
    [anon_sym__Generic] = ACTIONS(2604),
    [anon_sym_asm] = ACTIONS(2604),
    [anon_sym___asm__] = ACTIONS(2604),
    [anon_sym___asm] = ACTIONS(2604),
    [sym_number_literal] = ACTIONS(2606),
    [anon_sym_L_SQUOTE] = ACTIONS(2606),
    [anon_sym_u_SQUOTE] = ACTIONS(2606),
    [anon_sym_U_SQUOTE] = ACTIONS(2606),
    [anon_sym_u8_SQUOTE] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2606),
    [anon_sym_AT] = ACTIONS(2604),
    [anon_sym_DQUOTE] = ACTIONS(2606),
    [anon_sym_L_DQUOTE] = ACTIONS(2606),
    [anon_sym_u_DQUOTE] = ACTIONS(2606),
    [anon_sym_U_DQUOTE] = ACTIONS(2606),
    [anon_sym_u8_DQUOTE] = ACTIONS(2606),
    [sym_true] = ACTIONS(2604),
    [sym_false] = ACTIONS(2604),
    [anon_sym_NULL] = ACTIONS(2604),
    [anon_sym_nullptr] = ACTIONS(2604),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2604),
    [anon_sym___typeof] = ACTIONS(2604),
    [anon_sym_typeof] = ACTIONS(2604),
    [anon_sym_ATimport] = ACTIONS(2606),
    [aux_sym_preproc_undef_token1] = ACTIONS(2604),
    [anon_sym_POUND] = ACTIONS(2604),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2604),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2604),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2604),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_API_AVAILABLE] = ACTIONS(2604),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_API_DEPRECATED] = ACTIONS(2604),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym___deprecated_msg] = ACTIONS(2604),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2604),
    [anon_sym_ATprotocol] = ACTIONS(2606),
    [anon_sym_ATinterface] = ACTIONS(2606),
    [anon_sym_ATimplementation] = ACTIONS(2606),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2606),
    [anon_sym_ATsynthesize] = ACTIONS(2606),
    [anon_sym_ATdynamic] = ACTIONS(2606),
    [anon_sym_ATtry] = ACTIONS(2606),
    [anon_sym___try] = ACTIONS(2604),
    [anon_sym_ATthrow] = ACTIONS(2606),
    [anon_sym_ATselector] = ACTIONS(2606),
    [anon_sym_ATavailable] = ACTIONS(2606),
    [anon_sym___builtin_available] = ACTIONS(2604),
    [anon_sym_va_arg] = ACTIONS(2604),
    [anon_sym_ATencode] = ACTIONS(2606),
    [anon_sym_ATsynchronized] = ACTIONS(2606),
    [anon_sym_BOOL] = ACTIONS(2604),
    [anon_sym_IMP] = ACTIONS(2604),
    [anon_sym_SEL] = ACTIONS(2604),
    [anon_sym_Class] = ACTIONS(2604),
    [anon_sym_id] = ACTIONS(2604),
  },
  [STATE(493)] = {
    [sym_identifier] = ACTIONS(2638),
    [aux_sym_preproc_include_token1] = ACTIONS(2638),
    [aux_sym_preproc_include_token2] = ACTIONS(2638),
    [aux_sym_preproc_def_token1] = ACTIONS(2638),
    [aux_sym_preproc_if_token1] = ACTIONS(2638),
    [aux_sym_preproc_if_token2] = ACTIONS(2638),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2638),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2638),
    [aux_sym_preproc_else_token1] = ACTIONS(2638),
    [aux_sym_preproc_elif_token1] = ACTIONS(2638),
    [sym_preproc_directive] = ACTIONS(2638),
    [anon_sym_LPAREN2] = ACTIONS(2640),
    [anon_sym_BANG] = ACTIONS(2640),
    [anon_sym_TILDE] = ACTIONS(2640),
    [anon_sym_DASH] = ACTIONS(2638),
    [anon_sym_PLUS] = ACTIONS(2638),
    [anon_sym_STAR] = ACTIONS(2640),
    [anon_sym_CARET] = ACTIONS(2640),
    [anon_sym_AMP] = ACTIONS(2640),
    [anon_sym_SEMI] = ACTIONS(2640),
    [anon_sym___extension__] = ACTIONS(2638),
    [anon_sym_typedef] = ACTIONS(2638),
    [anon_sym_extern] = ACTIONS(2638),
    [anon_sym___attribute__] = ACTIONS(2638),
    [anon_sym___attribute] = ACTIONS(2638),
    [anon_sym_noreturn] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2640),
    [anon_sym___declspec] = ACTIONS(2638),
    [anon_sym___cdecl] = ACTIONS(2638),
    [anon_sym___clrcall] = ACTIONS(2638),
    [anon_sym___stdcall] = ACTIONS(2638),
    [anon_sym___fastcall] = ACTIONS(2638),
    [anon_sym___thiscall] = ACTIONS(2638),
    [anon_sym___vectorcall] = ACTIONS(2638),
    [anon_sym_LBRACE] = ACTIONS(2640),
    [anon_sym_signed] = ACTIONS(2638),
    [anon_sym_unsigned] = ACTIONS(2638),
    [anon_sym_long] = ACTIONS(2638),
    [anon_sym_short] = ACTIONS(2638),
    [anon_sym_static] = ACTIONS(2638),
    [anon_sym_ATautoreleasepool] = ACTIONS(2640),
    [anon_sym_auto] = ACTIONS(2638),
    [anon_sym_register] = ACTIONS(2638),
    [anon_sym_inline] = ACTIONS(2638),
    [anon_sym___inline] = ACTIONS(2638),
    [anon_sym___inline__] = ACTIONS(2638),
    [anon_sym___forceinline] = ACTIONS(2638),
    [anon_sym_thread_local] = ACTIONS(2638),
    [anon_sym___thread] = ACTIONS(2638),
    [anon_sym_CG_EXTERN] = ACTIONS(2638),
    [anon_sym_CG_INLINE] = ACTIONS(2638),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2638),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2638),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2638),
    [anon_sym_IBOutlet] = ACTIONS(2638),
    [anon_sym_IBInspectable] = ACTIONS(2638),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2638),
    [anon_sym_NS_INLINE] = ACTIONS(2638),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2638),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2638),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2638),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2638),
    [anon_sym_const] = ACTIONS(2638),
    [anon_sym_constexpr] = ACTIONS(2638),
    [anon_sym_volatile] = ACTIONS(2638),
    [anon_sym_restrict] = ACTIONS(2638),
    [anon_sym___restrict__] = ACTIONS(2638),
    [anon_sym__Atomic] = ACTIONS(2638),
    [anon_sym__Noreturn] = ACTIONS(2638),
    [anon_sym__Nonnull] = ACTIONS(2638),
    [anon_sym_nullable] = ACTIONS(2638),
    [anon_sym__Complex] = ACTIONS(2638),
    [anon_sym__Nullable] = ACTIONS(2638),
    [anon_sym__Nullable_result] = ACTIONS(2638),
    [anon_sym__Null_unspecified] = ACTIONS(2638),
    [anon_sym___autoreleasing] = ACTIONS(2638),
    [anon_sym___block] = ACTIONS(2638),
    [anon_sym___bridge] = ACTIONS(2638),
    [anon_sym___bridge_retained] = ACTIONS(2638),
    [anon_sym___bridge_transfer] = ACTIONS(2638),
    [anon_sym___complex] = ACTIONS(2638),
    [anon_sym___const] = ACTIONS(2638),
    [anon_sym___imag] = ACTIONS(2638),
    [anon_sym___kindof] = ACTIONS(2638),
    [anon_sym___nonnull] = ACTIONS(2638),
    [anon_sym___nullable] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2638),
    [anon_sym___real] = ACTIONS(2638),
    [anon_sym___strong] = ACTIONS(2638),
    [anon_sym___unsafe_unretained] = ACTIONS(2638),
    [anon_sym___unused] = ACTIONS(2638),
    [anon_sym___weak] = ACTIONS(2638),
    [anon_sym_alignas] = ACTIONS(2638),
    [anon_sym__Alignas] = ACTIONS(2638),
    [sym_primitive_type] = ACTIONS(2638),
    [anon_sym_enum] = ACTIONS(2638),
    [anon_sym_struct] = ACTIONS(2638),
    [anon_sym_union] = ACTIONS(2638),
    [anon_sym_if] = ACTIONS(2638),
    [anon_sym_switch] = ACTIONS(2638),
    [anon_sym_case] = ACTIONS(2638),
    [anon_sym_default] = ACTIONS(2638),
    [anon_sym_while] = ACTIONS(2638),
    [anon_sym_do] = ACTIONS(2638),
    [anon_sym_for] = ACTIONS(2638),
    [anon_sym_in] = ACTIONS(2638),
    [anon_sym_return] = ACTIONS(2638),
    [anon_sym_break] = ACTIONS(2638),
    [anon_sym_continue] = ACTIONS(2638),
    [anon_sym_goto] = ACTIONS(2638),
    [anon_sym_DASH_DASH] = ACTIONS(2640),
    [anon_sym_PLUS_PLUS] = ACTIONS(2640),
    [anon_sym_sizeof] = ACTIONS(2638),
    [anon_sym___alignof__] = ACTIONS(2638),
    [anon_sym___alignof] = ACTIONS(2638),
    [anon_sym__alignof] = ACTIONS(2638),
    [anon_sym_alignof] = ACTIONS(2638),
    [anon_sym__Alignof] = ACTIONS(2638),
    [anon_sym_offsetof] = ACTIONS(2638),
    [anon_sym__Generic] = ACTIONS(2638),
    [anon_sym_asm] = ACTIONS(2638),
    [anon_sym___asm__] = ACTIONS(2638),
    [anon_sym___asm] = ACTIONS(2638),
    [sym_number_literal] = ACTIONS(2640),
    [anon_sym_L_SQUOTE] = ACTIONS(2640),
    [anon_sym_u_SQUOTE] = ACTIONS(2640),
    [anon_sym_U_SQUOTE] = ACTIONS(2640),
    [anon_sym_u8_SQUOTE] = ACTIONS(2640),
    [anon_sym_SQUOTE] = ACTIONS(2640),
    [anon_sym_AT] = ACTIONS(2638),
    [anon_sym_DQUOTE] = ACTIONS(2640),
    [anon_sym_L_DQUOTE] = ACTIONS(2640),
    [anon_sym_u_DQUOTE] = ACTIONS(2640),
    [anon_sym_U_DQUOTE] = ACTIONS(2640),
    [anon_sym_u8_DQUOTE] = ACTIONS(2640),
    [sym_true] = ACTIONS(2638),
    [sym_false] = ACTIONS(2638),
    [anon_sym_NULL] = ACTIONS(2638),
    [anon_sym_nullptr] = ACTIONS(2638),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2638),
    [anon_sym___typeof] = ACTIONS(2638),
    [anon_sym_typeof] = ACTIONS(2638),
    [anon_sym_ATimport] = ACTIONS(2640),
    [aux_sym_preproc_undef_token1] = ACTIONS(2638),
    [anon_sym_POUND] = ACTIONS(2638),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2638),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2638),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2638),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2638),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2638),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2638),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2638),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_API_AVAILABLE] = ACTIONS(2638),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_API_DEPRECATED] = ACTIONS(2638),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2638),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym___deprecated_msg] = ACTIONS(2638),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2638),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2638),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2638),
    [anon_sym_ATprotocol] = ACTIONS(2640),
    [anon_sym_ATinterface] = ACTIONS(2640),
    [anon_sym_ATimplementation] = ACTIONS(2640),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2640),
    [anon_sym_ATsynthesize] = ACTIONS(2640),
    [anon_sym_ATdynamic] = ACTIONS(2640),
    [anon_sym_ATtry] = ACTIONS(2640),
    [anon_sym___try] = ACTIONS(2638),
    [anon_sym_ATthrow] = ACTIONS(2640),
    [anon_sym_ATselector] = ACTIONS(2640),
    [anon_sym_ATavailable] = ACTIONS(2640),
    [anon_sym___builtin_available] = ACTIONS(2638),
    [anon_sym_va_arg] = ACTIONS(2638),
    [anon_sym_ATencode] = ACTIONS(2640),
    [anon_sym_ATsynchronized] = ACTIONS(2640),
    [anon_sym_BOOL] = ACTIONS(2638),
    [anon_sym_IMP] = ACTIONS(2638),
    [anon_sym_SEL] = ACTIONS(2638),
    [anon_sym_Class] = ACTIONS(2638),
    [anon_sym_id] = ACTIONS(2638),
  },
  [STATE(494)] = {
    [sym_identifier] = ACTIONS(2744),
    [aux_sym_preproc_include_token1] = ACTIONS(2744),
    [aux_sym_preproc_include_token2] = ACTIONS(2744),
    [aux_sym_preproc_def_token1] = ACTIONS(2744),
    [aux_sym_preproc_if_token1] = ACTIONS(2744),
    [aux_sym_preproc_if_token2] = ACTIONS(2744),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2744),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2744),
    [aux_sym_preproc_else_token1] = ACTIONS(2744),
    [aux_sym_preproc_elif_token1] = ACTIONS(2744),
    [sym_preproc_directive] = ACTIONS(2744),
    [anon_sym_LPAREN2] = ACTIONS(2746),
    [anon_sym_BANG] = ACTIONS(2746),
    [anon_sym_TILDE] = ACTIONS(2746),
    [anon_sym_DASH] = ACTIONS(2744),
    [anon_sym_PLUS] = ACTIONS(2744),
    [anon_sym_STAR] = ACTIONS(2746),
    [anon_sym_CARET] = ACTIONS(2746),
    [anon_sym_AMP] = ACTIONS(2746),
    [anon_sym_SEMI] = ACTIONS(2746),
    [anon_sym___extension__] = ACTIONS(2744),
    [anon_sym_typedef] = ACTIONS(2744),
    [anon_sym_extern] = ACTIONS(2744),
    [anon_sym___attribute__] = ACTIONS(2744),
    [anon_sym___attribute] = ACTIONS(2744),
    [anon_sym_noreturn] = ACTIONS(2744),
    [anon_sym_LBRACK] = ACTIONS(2746),
    [anon_sym___declspec] = ACTIONS(2744),
    [anon_sym___cdecl] = ACTIONS(2744),
    [anon_sym___clrcall] = ACTIONS(2744),
    [anon_sym___stdcall] = ACTIONS(2744),
    [anon_sym___fastcall] = ACTIONS(2744),
    [anon_sym___thiscall] = ACTIONS(2744),
    [anon_sym___vectorcall] = ACTIONS(2744),
    [anon_sym_LBRACE] = ACTIONS(2746),
    [anon_sym_signed] = ACTIONS(2744),
    [anon_sym_unsigned] = ACTIONS(2744),
    [anon_sym_long] = ACTIONS(2744),
    [anon_sym_short] = ACTIONS(2744),
    [anon_sym_static] = ACTIONS(2744),
    [anon_sym_ATautoreleasepool] = ACTIONS(2746),
    [anon_sym_auto] = ACTIONS(2744),
    [anon_sym_register] = ACTIONS(2744),
    [anon_sym_inline] = ACTIONS(2744),
    [anon_sym___inline] = ACTIONS(2744),
    [anon_sym___inline__] = ACTIONS(2744),
    [anon_sym___forceinline] = ACTIONS(2744),
    [anon_sym_thread_local] = ACTIONS(2744),
    [anon_sym___thread] = ACTIONS(2744),
    [anon_sym_CG_EXTERN] = ACTIONS(2744),
    [anon_sym_CG_INLINE] = ACTIONS(2744),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2744),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2744),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2744),
    [anon_sym_IBOutlet] = ACTIONS(2744),
    [anon_sym_IBInspectable] = ACTIONS(2744),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2744),
    [anon_sym_NS_INLINE] = ACTIONS(2744),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2744),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2744),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2744),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2744),
    [anon_sym_const] = ACTIONS(2744),
    [anon_sym_constexpr] = ACTIONS(2744),
    [anon_sym_volatile] = ACTIONS(2744),
    [anon_sym_restrict] = ACTIONS(2744),
    [anon_sym___restrict__] = ACTIONS(2744),
    [anon_sym__Atomic] = ACTIONS(2744),
    [anon_sym__Noreturn] = ACTIONS(2744),
    [anon_sym__Nonnull] = ACTIONS(2744),
    [anon_sym_nullable] = ACTIONS(2744),
    [anon_sym__Complex] = ACTIONS(2744),
    [anon_sym__Nullable] = ACTIONS(2744),
    [anon_sym__Nullable_result] = ACTIONS(2744),
    [anon_sym__Null_unspecified] = ACTIONS(2744),
    [anon_sym___autoreleasing] = ACTIONS(2744),
    [anon_sym___block] = ACTIONS(2744),
    [anon_sym___bridge] = ACTIONS(2744),
    [anon_sym___bridge_retained] = ACTIONS(2744),
    [anon_sym___bridge_transfer] = ACTIONS(2744),
    [anon_sym___complex] = ACTIONS(2744),
    [anon_sym___const] = ACTIONS(2744),
    [anon_sym___imag] = ACTIONS(2744),
    [anon_sym___kindof] = ACTIONS(2744),
    [anon_sym___nonnull] = ACTIONS(2744),
    [anon_sym___nullable] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2744),
    [anon_sym___real] = ACTIONS(2744),
    [anon_sym___strong] = ACTIONS(2744),
    [anon_sym___unsafe_unretained] = ACTIONS(2744),
    [anon_sym___unused] = ACTIONS(2744),
    [anon_sym___weak] = ACTIONS(2744),
    [anon_sym_alignas] = ACTIONS(2744),
    [anon_sym__Alignas] = ACTIONS(2744),
    [sym_primitive_type] = ACTIONS(2744),
    [anon_sym_enum] = ACTIONS(2744),
    [anon_sym_struct] = ACTIONS(2744),
    [anon_sym_union] = ACTIONS(2744),
    [anon_sym_if] = ACTIONS(2744),
    [anon_sym_switch] = ACTIONS(2744),
    [anon_sym_case] = ACTIONS(2744),
    [anon_sym_default] = ACTIONS(2744),
    [anon_sym_while] = ACTIONS(2744),
    [anon_sym_do] = ACTIONS(2744),
    [anon_sym_for] = ACTIONS(2744),
    [anon_sym_in] = ACTIONS(2744),
    [anon_sym_return] = ACTIONS(2744),
    [anon_sym_break] = ACTIONS(2744),
    [anon_sym_continue] = ACTIONS(2744),
    [anon_sym_goto] = ACTIONS(2744),
    [anon_sym_DASH_DASH] = ACTIONS(2746),
    [anon_sym_PLUS_PLUS] = ACTIONS(2746),
    [anon_sym_sizeof] = ACTIONS(2744),
    [anon_sym___alignof__] = ACTIONS(2744),
    [anon_sym___alignof] = ACTIONS(2744),
    [anon_sym__alignof] = ACTIONS(2744),
    [anon_sym_alignof] = ACTIONS(2744),
    [anon_sym__Alignof] = ACTIONS(2744),
    [anon_sym_offsetof] = ACTIONS(2744),
    [anon_sym__Generic] = ACTIONS(2744),
    [anon_sym_asm] = ACTIONS(2744),
    [anon_sym___asm__] = ACTIONS(2744),
    [anon_sym___asm] = ACTIONS(2744),
    [sym_number_literal] = ACTIONS(2746),
    [anon_sym_L_SQUOTE] = ACTIONS(2746),
    [anon_sym_u_SQUOTE] = ACTIONS(2746),
    [anon_sym_U_SQUOTE] = ACTIONS(2746),
    [anon_sym_u8_SQUOTE] = ACTIONS(2746),
    [anon_sym_SQUOTE] = ACTIONS(2746),
    [anon_sym_AT] = ACTIONS(2744),
    [anon_sym_DQUOTE] = ACTIONS(2746),
    [anon_sym_L_DQUOTE] = ACTIONS(2746),
    [anon_sym_u_DQUOTE] = ACTIONS(2746),
    [anon_sym_U_DQUOTE] = ACTIONS(2746),
    [anon_sym_u8_DQUOTE] = ACTIONS(2746),
    [sym_true] = ACTIONS(2744),
    [sym_false] = ACTIONS(2744),
    [anon_sym_NULL] = ACTIONS(2744),
    [anon_sym_nullptr] = ACTIONS(2744),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2744),
    [anon_sym___typeof] = ACTIONS(2744),
    [anon_sym_typeof] = ACTIONS(2744),
    [anon_sym_ATimport] = ACTIONS(2746),
    [aux_sym_preproc_undef_token1] = ACTIONS(2744),
    [anon_sym_POUND] = ACTIONS(2744),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2744),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2744),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2744),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2744),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2744),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2744),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2744),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_API_AVAILABLE] = ACTIONS(2744),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_API_DEPRECATED] = ACTIONS(2744),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2744),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym___deprecated_msg] = ACTIONS(2744),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2744),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2744),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2744),
    [anon_sym_ATprotocol] = ACTIONS(2746),
    [anon_sym_ATinterface] = ACTIONS(2746),
    [anon_sym_ATimplementation] = ACTIONS(2746),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2746),
    [anon_sym_ATsynthesize] = ACTIONS(2746),
    [anon_sym_ATdynamic] = ACTIONS(2746),
    [anon_sym_ATtry] = ACTIONS(2746),
    [anon_sym___try] = ACTIONS(2744),
    [anon_sym_ATthrow] = ACTIONS(2746),
    [anon_sym_ATselector] = ACTIONS(2746),
    [anon_sym_ATavailable] = ACTIONS(2746),
    [anon_sym___builtin_available] = ACTIONS(2744),
    [anon_sym_va_arg] = ACTIONS(2744),
    [anon_sym_ATencode] = ACTIONS(2746),
    [anon_sym_ATsynchronized] = ACTIONS(2746),
    [anon_sym_BOOL] = ACTIONS(2744),
    [anon_sym_IMP] = ACTIONS(2744),
    [anon_sym_SEL] = ACTIONS(2744),
    [anon_sym_Class] = ACTIONS(2744),
    [anon_sym_id] = ACTIONS(2744),
  },
  [STATE(495)] = {
    [sym_identifier] = ACTIONS(2856),
    [aux_sym_preproc_include_token1] = ACTIONS(2856),
    [aux_sym_preproc_include_token2] = ACTIONS(2856),
    [aux_sym_preproc_def_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token2] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2856),
    [aux_sym_preproc_else_token1] = ACTIONS(2856),
    [aux_sym_preproc_elif_token1] = ACTIONS(2856),
    [sym_preproc_directive] = ACTIONS(2856),
    [anon_sym_LPAREN2] = ACTIONS(2858),
    [anon_sym_BANG] = ACTIONS(2858),
    [anon_sym_TILDE] = ACTIONS(2858),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_STAR] = ACTIONS(2858),
    [anon_sym_CARET] = ACTIONS(2858),
    [anon_sym_AMP] = ACTIONS(2858),
    [anon_sym_SEMI] = ACTIONS(2858),
    [anon_sym___extension__] = ACTIONS(2856),
    [anon_sym_typedef] = ACTIONS(2856),
    [anon_sym_extern] = ACTIONS(2856),
    [anon_sym___attribute__] = ACTIONS(2856),
    [anon_sym___attribute] = ACTIONS(2856),
    [anon_sym_noreturn] = ACTIONS(2856),
    [anon_sym_LBRACK] = ACTIONS(2858),
    [anon_sym___declspec] = ACTIONS(2856),
    [anon_sym___cdecl] = ACTIONS(2856),
    [anon_sym___clrcall] = ACTIONS(2856),
    [anon_sym___stdcall] = ACTIONS(2856),
    [anon_sym___fastcall] = ACTIONS(2856),
    [anon_sym___thiscall] = ACTIONS(2856),
    [anon_sym___vectorcall] = ACTIONS(2856),
    [anon_sym_LBRACE] = ACTIONS(2858),
    [anon_sym_signed] = ACTIONS(2856),
    [anon_sym_unsigned] = ACTIONS(2856),
    [anon_sym_long] = ACTIONS(2856),
    [anon_sym_short] = ACTIONS(2856),
    [anon_sym_static] = ACTIONS(2856),
    [anon_sym_ATautoreleasepool] = ACTIONS(2858),
    [anon_sym_auto] = ACTIONS(2856),
    [anon_sym_register] = ACTIONS(2856),
    [anon_sym_inline] = ACTIONS(2856),
    [anon_sym___inline] = ACTIONS(2856),
    [anon_sym___inline__] = ACTIONS(2856),
    [anon_sym___forceinline] = ACTIONS(2856),
    [anon_sym_thread_local] = ACTIONS(2856),
    [anon_sym___thread] = ACTIONS(2856),
    [anon_sym_CG_EXTERN] = ACTIONS(2856),
    [anon_sym_CG_INLINE] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2856),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2856),
    [anon_sym_IBOutlet] = ACTIONS(2856),
    [anon_sym_IBInspectable] = ACTIONS(2856),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2856),
    [anon_sym_NS_INLINE] = ACTIONS(2856),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2856),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2856),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2856),
    [anon_sym_const] = ACTIONS(2856),
    [anon_sym_constexpr] = ACTIONS(2856),
    [anon_sym_volatile] = ACTIONS(2856),
    [anon_sym_restrict] = ACTIONS(2856),
    [anon_sym___restrict__] = ACTIONS(2856),
    [anon_sym__Atomic] = ACTIONS(2856),
    [anon_sym__Noreturn] = ACTIONS(2856),
    [anon_sym__Nonnull] = ACTIONS(2856),
    [anon_sym_nullable] = ACTIONS(2856),
    [anon_sym__Complex] = ACTIONS(2856),
    [anon_sym__Nullable] = ACTIONS(2856),
    [anon_sym__Nullable_result] = ACTIONS(2856),
    [anon_sym__Null_unspecified] = ACTIONS(2856),
    [anon_sym___autoreleasing] = ACTIONS(2856),
    [anon_sym___block] = ACTIONS(2856),
    [anon_sym___bridge] = ACTIONS(2856),
    [anon_sym___bridge_retained] = ACTIONS(2856),
    [anon_sym___bridge_transfer] = ACTIONS(2856),
    [anon_sym___complex] = ACTIONS(2856),
    [anon_sym___const] = ACTIONS(2856),
    [anon_sym___imag] = ACTIONS(2856),
    [anon_sym___kindof] = ACTIONS(2856),
    [anon_sym___nonnull] = ACTIONS(2856),
    [anon_sym___nullable] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2856),
    [anon_sym___real] = ACTIONS(2856),
    [anon_sym___strong] = ACTIONS(2856),
    [anon_sym___unsafe_unretained] = ACTIONS(2856),
    [anon_sym___unused] = ACTIONS(2856),
    [anon_sym___weak] = ACTIONS(2856),
    [anon_sym_alignas] = ACTIONS(2856),
    [anon_sym__Alignas] = ACTIONS(2856),
    [sym_primitive_type] = ACTIONS(2856),
    [anon_sym_enum] = ACTIONS(2856),
    [anon_sym_struct] = ACTIONS(2856),
    [anon_sym_union] = ACTIONS(2856),
    [anon_sym_if] = ACTIONS(2856),
    [anon_sym_switch] = ACTIONS(2856),
    [anon_sym_case] = ACTIONS(2856),
    [anon_sym_default] = ACTIONS(2856),
    [anon_sym_while] = ACTIONS(2856),
    [anon_sym_do] = ACTIONS(2856),
    [anon_sym_for] = ACTIONS(2856),
    [anon_sym_in] = ACTIONS(2856),
    [anon_sym_return] = ACTIONS(2856),
    [anon_sym_break] = ACTIONS(2856),
    [anon_sym_continue] = ACTIONS(2856),
    [anon_sym_goto] = ACTIONS(2856),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_sizeof] = ACTIONS(2856),
    [anon_sym___alignof__] = ACTIONS(2856),
    [anon_sym___alignof] = ACTIONS(2856),
    [anon_sym__alignof] = ACTIONS(2856),
    [anon_sym_alignof] = ACTIONS(2856),
    [anon_sym__Alignof] = ACTIONS(2856),
    [anon_sym_offsetof] = ACTIONS(2856),
    [anon_sym__Generic] = ACTIONS(2856),
    [anon_sym_asm] = ACTIONS(2856),
    [anon_sym___asm__] = ACTIONS(2856),
    [anon_sym___asm] = ACTIONS(2856),
    [sym_number_literal] = ACTIONS(2858),
    [anon_sym_L_SQUOTE] = ACTIONS(2858),
    [anon_sym_u_SQUOTE] = ACTIONS(2858),
    [anon_sym_U_SQUOTE] = ACTIONS(2858),
    [anon_sym_u8_SQUOTE] = ACTIONS(2858),
    [anon_sym_SQUOTE] = ACTIONS(2858),
    [anon_sym_AT] = ACTIONS(2856),
    [anon_sym_DQUOTE] = ACTIONS(2858),
    [anon_sym_L_DQUOTE] = ACTIONS(2858),
    [anon_sym_u_DQUOTE] = ACTIONS(2858),
    [anon_sym_U_DQUOTE] = ACTIONS(2858),
    [anon_sym_u8_DQUOTE] = ACTIONS(2858),
    [sym_true] = ACTIONS(2856),
    [sym_false] = ACTIONS(2856),
    [anon_sym_NULL] = ACTIONS(2856),
    [anon_sym_nullptr] = ACTIONS(2856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2856),
    [anon_sym___typeof] = ACTIONS(2856),
    [anon_sym_typeof] = ACTIONS(2856),
    [anon_sym_ATimport] = ACTIONS(2858),
    [aux_sym_preproc_undef_token1] = ACTIONS(2856),
    [anon_sym_POUND] = ACTIONS(2856),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2856),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2856),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2856),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_API_AVAILABLE] = ACTIONS(2856),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_API_DEPRECATED] = ACTIONS(2856),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym___deprecated_msg] = ACTIONS(2856),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2856),
    [anon_sym_ATprotocol] = ACTIONS(2858),
    [anon_sym_ATinterface] = ACTIONS(2858),
    [anon_sym_ATimplementation] = ACTIONS(2858),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2858),
    [anon_sym_ATsynthesize] = ACTIONS(2858),
    [anon_sym_ATdynamic] = ACTIONS(2858),
    [anon_sym_ATtry] = ACTIONS(2858),
    [anon_sym___try] = ACTIONS(2856),
    [anon_sym_ATthrow] = ACTIONS(2858),
    [anon_sym_ATselector] = ACTIONS(2858),
    [anon_sym_ATavailable] = ACTIONS(2858),
    [anon_sym___builtin_available] = ACTIONS(2856),
    [anon_sym_va_arg] = ACTIONS(2856),
    [anon_sym_ATencode] = ACTIONS(2858),
    [anon_sym_ATsynchronized] = ACTIONS(2858),
    [anon_sym_BOOL] = ACTIONS(2856),
    [anon_sym_IMP] = ACTIONS(2856),
    [anon_sym_SEL] = ACTIONS(2856),
    [anon_sym_Class] = ACTIONS(2856),
    [anon_sym_id] = ACTIONS(2856),
  },
  [STATE(496)] = {
    [sym_identifier] = ACTIONS(2748),
    [aux_sym_preproc_include_token1] = ACTIONS(2748),
    [aux_sym_preproc_include_token2] = ACTIONS(2748),
    [aux_sym_preproc_def_token1] = ACTIONS(2748),
    [aux_sym_preproc_if_token1] = ACTIONS(2748),
    [aux_sym_preproc_if_token2] = ACTIONS(2748),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2748),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2748),
    [aux_sym_preproc_else_token1] = ACTIONS(2748),
    [aux_sym_preproc_elif_token1] = ACTIONS(2748),
    [sym_preproc_directive] = ACTIONS(2748),
    [anon_sym_LPAREN2] = ACTIONS(2750),
    [anon_sym_BANG] = ACTIONS(2750),
    [anon_sym_TILDE] = ACTIONS(2750),
    [anon_sym_DASH] = ACTIONS(2748),
    [anon_sym_PLUS] = ACTIONS(2748),
    [anon_sym_STAR] = ACTIONS(2750),
    [anon_sym_CARET] = ACTIONS(2750),
    [anon_sym_AMP] = ACTIONS(2750),
    [anon_sym_SEMI] = ACTIONS(2750),
    [anon_sym___extension__] = ACTIONS(2748),
    [anon_sym_typedef] = ACTIONS(2748),
    [anon_sym_extern] = ACTIONS(2748),
    [anon_sym___attribute__] = ACTIONS(2748),
    [anon_sym___attribute] = ACTIONS(2748),
    [anon_sym_noreturn] = ACTIONS(2748),
    [anon_sym_LBRACK] = ACTIONS(2750),
    [anon_sym___declspec] = ACTIONS(2748),
    [anon_sym___cdecl] = ACTIONS(2748),
    [anon_sym___clrcall] = ACTIONS(2748),
    [anon_sym___stdcall] = ACTIONS(2748),
    [anon_sym___fastcall] = ACTIONS(2748),
    [anon_sym___thiscall] = ACTIONS(2748),
    [anon_sym___vectorcall] = ACTIONS(2748),
    [anon_sym_LBRACE] = ACTIONS(2750),
    [anon_sym_signed] = ACTIONS(2748),
    [anon_sym_unsigned] = ACTIONS(2748),
    [anon_sym_long] = ACTIONS(2748),
    [anon_sym_short] = ACTIONS(2748),
    [anon_sym_static] = ACTIONS(2748),
    [anon_sym_ATautoreleasepool] = ACTIONS(2750),
    [anon_sym_auto] = ACTIONS(2748),
    [anon_sym_register] = ACTIONS(2748),
    [anon_sym_inline] = ACTIONS(2748),
    [anon_sym___inline] = ACTIONS(2748),
    [anon_sym___inline__] = ACTIONS(2748),
    [anon_sym___forceinline] = ACTIONS(2748),
    [anon_sym_thread_local] = ACTIONS(2748),
    [anon_sym___thread] = ACTIONS(2748),
    [anon_sym_CG_EXTERN] = ACTIONS(2748),
    [anon_sym_CG_INLINE] = ACTIONS(2748),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2748),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2748),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2748),
    [anon_sym_IBOutlet] = ACTIONS(2748),
    [anon_sym_IBInspectable] = ACTIONS(2748),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2748),
    [anon_sym_NS_INLINE] = ACTIONS(2748),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2748),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2748),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2748),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2748),
    [anon_sym_const] = ACTIONS(2748),
    [anon_sym_constexpr] = ACTIONS(2748),
    [anon_sym_volatile] = ACTIONS(2748),
    [anon_sym_restrict] = ACTIONS(2748),
    [anon_sym___restrict__] = ACTIONS(2748),
    [anon_sym__Atomic] = ACTIONS(2748),
    [anon_sym__Noreturn] = ACTIONS(2748),
    [anon_sym__Nonnull] = ACTIONS(2748),
    [anon_sym_nullable] = ACTIONS(2748),
    [anon_sym__Complex] = ACTIONS(2748),
    [anon_sym__Nullable] = ACTIONS(2748),
    [anon_sym__Nullable_result] = ACTIONS(2748),
    [anon_sym__Null_unspecified] = ACTIONS(2748),
    [anon_sym___autoreleasing] = ACTIONS(2748),
    [anon_sym___block] = ACTIONS(2748),
    [anon_sym___bridge] = ACTIONS(2748),
    [anon_sym___bridge_retained] = ACTIONS(2748),
    [anon_sym___bridge_transfer] = ACTIONS(2748),
    [anon_sym___complex] = ACTIONS(2748),
    [anon_sym___const] = ACTIONS(2748),
    [anon_sym___imag] = ACTIONS(2748),
    [anon_sym___kindof] = ACTIONS(2748),
    [anon_sym___nonnull] = ACTIONS(2748),
    [anon_sym___nullable] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2748),
    [anon_sym___real] = ACTIONS(2748),
    [anon_sym___strong] = ACTIONS(2748),
    [anon_sym___unsafe_unretained] = ACTIONS(2748),
    [anon_sym___unused] = ACTIONS(2748),
    [anon_sym___weak] = ACTIONS(2748),
    [anon_sym_alignas] = ACTIONS(2748),
    [anon_sym__Alignas] = ACTIONS(2748),
    [sym_primitive_type] = ACTIONS(2748),
    [anon_sym_enum] = ACTIONS(2748),
    [anon_sym_struct] = ACTIONS(2748),
    [anon_sym_union] = ACTIONS(2748),
    [anon_sym_if] = ACTIONS(2748),
    [anon_sym_switch] = ACTIONS(2748),
    [anon_sym_case] = ACTIONS(2748),
    [anon_sym_default] = ACTIONS(2748),
    [anon_sym_while] = ACTIONS(2748),
    [anon_sym_do] = ACTIONS(2748),
    [anon_sym_for] = ACTIONS(2748),
    [anon_sym_in] = ACTIONS(2748),
    [anon_sym_return] = ACTIONS(2748),
    [anon_sym_break] = ACTIONS(2748),
    [anon_sym_continue] = ACTIONS(2748),
    [anon_sym_goto] = ACTIONS(2748),
    [anon_sym_DASH_DASH] = ACTIONS(2750),
    [anon_sym_PLUS_PLUS] = ACTIONS(2750),
    [anon_sym_sizeof] = ACTIONS(2748),
    [anon_sym___alignof__] = ACTIONS(2748),
    [anon_sym___alignof] = ACTIONS(2748),
    [anon_sym__alignof] = ACTIONS(2748),
    [anon_sym_alignof] = ACTIONS(2748),
    [anon_sym__Alignof] = ACTIONS(2748),
    [anon_sym_offsetof] = ACTIONS(2748),
    [anon_sym__Generic] = ACTIONS(2748),
    [anon_sym_asm] = ACTIONS(2748),
    [anon_sym___asm__] = ACTIONS(2748),
    [anon_sym___asm] = ACTIONS(2748),
    [sym_number_literal] = ACTIONS(2750),
    [anon_sym_L_SQUOTE] = ACTIONS(2750),
    [anon_sym_u_SQUOTE] = ACTIONS(2750),
    [anon_sym_U_SQUOTE] = ACTIONS(2750),
    [anon_sym_u8_SQUOTE] = ACTIONS(2750),
    [anon_sym_SQUOTE] = ACTIONS(2750),
    [anon_sym_AT] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_L_DQUOTE] = ACTIONS(2750),
    [anon_sym_u_DQUOTE] = ACTIONS(2750),
    [anon_sym_U_DQUOTE] = ACTIONS(2750),
    [anon_sym_u8_DQUOTE] = ACTIONS(2750),
    [sym_true] = ACTIONS(2748),
    [sym_false] = ACTIONS(2748),
    [anon_sym_NULL] = ACTIONS(2748),
    [anon_sym_nullptr] = ACTIONS(2748),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2748),
    [anon_sym___typeof] = ACTIONS(2748),
    [anon_sym_typeof] = ACTIONS(2748),
    [anon_sym_ATimport] = ACTIONS(2750),
    [aux_sym_preproc_undef_token1] = ACTIONS(2748),
    [anon_sym_POUND] = ACTIONS(2748),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2748),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2748),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2748),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2748),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2748),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2748),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2748),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_API_AVAILABLE] = ACTIONS(2748),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_API_DEPRECATED] = ACTIONS(2748),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2748),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym___deprecated_msg] = ACTIONS(2748),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2748),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2748),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2748),
    [anon_sym_ATprotocol] = ACTIONS(2750),
    [anon_sym_ATinterface] = ACTIONS(2750),
    [anon_sym_ATimplementation] = ACTIONS(2750),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2750),
    [anon_sym_ATsynthesize] = ACTIONS(2750),
    [anon_sym_ATdynamic] = ACTIONS(2750),
    [anon_sym_ATtry] = ACTIONS(2750),
    [anon_sym___try] = ACTIONS(2748),
    [anon_sym_ATthrow] = ACTIONS(2750),
    [anon_sym_ATselector] = ACTIONS(2750),
    [anon_sym_ATavailable] = ACTIONS(2750),
    [anon_sym___builtin_available] = ACTIONS(2748),
    [anon_sym_va_arg] = ACTIONS(2748),
    [anon_sym_ATencode] = ACTIONS(2750),
    [anon_sym_ATsynchronized] = ACTIONS(2750),
    [anon_sym_BOOL] = ACTIONS(2748),
    [anon_sym_IMP] = ACTIONS(2748),
    [anon_sym_SEL] = ACTIONS(2748),
    [anon_sym_Class] = ACTIONS(2748),
    [anon_sym_id] = ACTIONS(2748),
  },
  [STATE(497)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token2] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [aux_sym_preproc_else_token1] = ACTIONS(2752),
    [aux_sym_preproc_elif_token1] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATsynthesize] = ACTIONS(2754),
    [anon_sym_ATdynamic] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(498)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token2] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [aux_sym_preproc_else_token1] = ACTIONS(2752),
    [aux_sym_preproc_elif_token1] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATsynthesize] = ACTIONS(2754),
    [anon_sym_ATdynamic] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(499)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token2] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [aux_sym_preproc_else_token1] = ACTIONS(2752),
    [aux_sym_preproc_elif_token1] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATsynthesize] = ACTIONS(2754),
    [anon_sym_ATdynamic] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(500)] = {
    [sym_identifier] = ACTIONS(2756),
    [aux_sym_preproc_include_token1] = ACTIONS(2756),
    [aux_sym_preproc_include_token2] = ACTIONS(2756),
    [aux_sym_preproc_def_token1] = ACTIONS(2756),
    [aux_sym_preproc_if_token1] = ACTIONS(2756),
    [aux_sym_preproc_if_token2] = ACTIONS(2756),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2756),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2756),
    [aux_sym_preproc_else_token1] = ACTIONS(2756),
    [aux_sym_preproc_elif_token1] = ACTIONS(2756),
    [sym_preproc_directive] = ACTIONS(2756),
    [anon_sym_LPAREN2] = ACTIONS(2758),
    [anon_sym_BANG] = ACTIONS(2758),
    [anon_sym_TILDE] = ACTIONS(2758),
    [anon_sym_DASH] = ACTIONS(2756),
    [anon_sym_PLUS] = ACTIONS(2756),
    [anon_sym_STAR] = ACTIONS(2758),
    [anon_sym_CARET] = ACTIONS(2758),
    [anon_sym_AMP] = ACTIONS(2758),
    [anon_sym_SEMI] = ACTIONS(2758),
    [anon_sym___extension__] = ACTIONS(2756),
    [anon_sym_typedef] = ACTIONS(2756),
    [anon_sym_extern] = ACTIONS(2756),
    [anon_sym___attribute__] = ACTIONS(2756),
    [anon_sym___attribute] = ACTIONS(2756),
    [anon_sym_noreturn] = ACTIONS(2756),
    [anon_sym_LBRACK] = ACTIONS(2758),
    [anon_sym___declspec] = ACTIONS(2756),
    [anon_sym___cdecl] = ACTIONS(2756),
    [anon_sym___clrcall] = ACTIONS(2756),
    [anon_sym___stdcall] = ACTIONS(2756),
    [anon_sym___fastcall] = ACTIONS(2756),
    [anon_sym___thiscall] = ACTIONS(2756),
    [anon_sym___vectorcall] = ACTIONS(2756),
    [anon_sym_LBRACE] = ACTIONS(2758),
    [anon_sym_signed] = ACTIONS(2756),
    [anon_sym_unsigned] = ACTIONS(2756),
    [anon_sym_long] = ACTIONS(2756),
    [anon_sym_short] = ACTIONS(2756),
    [anon_sym_static] = ACTIONS(2756),
    [anon_sym_ATautoreleasepool] = ACTIONS(2758),
    [anon_sym_auto] = ACTIONS(2756),
    [anon_sym_register] = ACTIONS(2756),
    [anon_sym_inline] = ACTIONS(2756),
    [anon_sym___inline] = ACTIONS(2756),
    [anon_sym___inline__] = ACTIONS(2756),
    [anon_sym___forceinline] = ACTIONS(2756),
    [anon_sym_thread_local] = ACTIONS(2756),
    [anon_sym___thread] = ACTIONS(2756),
    [anon_sym_CG_EXTERN] = ACTIONS(2756),
    [anon_sym_CG_INLINE] = ACTIONS(2756),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2756),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2756),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2756),
    [anon_sym_IBOutlet] = ACTIONS(2756),
    [anon_sym_IBInspectable] = ACTIONS(2756),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2756),
    [anon_sym_NS_INLINE] = ACTIONS(2756),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2756),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2756),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2756),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2756),
    [anon_sym_const] = ACTIONS(2756),
    [anon_sym_constexpr] = ACTIONS(2756),
    [anon_sym_volatile] = ACTIONS(2756),
    [anon_sym_restrict] = ACTIONS(2756),
    [anon_sym___restrict__] = ACTIONS(2756),
    [anon_sym__Atomic] = ACTIONS(2756),
    [anon_sym__Noreturn] = ACTIONS(2756),
    [anon_sym__Nonnull] = ACTIONS(2756),
    [anon_sym_nullable] = ACTIONS(2756),
    [anon_sym__Complex] = ACTIONS(2756),
    [anon_sym__Nullable] = ACTIONS(2756),
    [anon_sym__Nullable_result] = ACTIONS(2756),
    [anon_sym__Null_unspecified] = ACTIONS(2756),
    [anon_sym___autoreleasing] = ACTIONS(2756),
    [anon_sym___block] = ACTIONS(2756),
    [anon_sym___bridge] = ACTIONS(2756),
    [anon_sym___bridge_retained] = ACTIONS(2756),
    [anon_sym___bridge_transfer] = ACTIONS(2756),
    [anon_sym___complex] = ACTIONS(2756),
    [anon_sym___const] = ACTIONS(2756),
    [anon_sym___imag] = ACTIONS(2756),
    [anon_sym___kindof] = ACTIONS(2756),
    [anon_sym___nonnull] = ACTIONS(2756),
    [anon_sym___nullable] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2756),
    [anon_sym___real] = ACTIONS(2756),
    [anon_sym___strong] = ACTIONS(2756),
    [anon_sym___unsafe_unretained] = ACTIONS(2756),
    [anon_sym___unused] = ACTIONS(2756),
    [anon_sym___weak] = ACTIONS(2756),
    [anon_sym_alignas] = ACTIONS(2756),
    [anon_sym__Alignas] = ACTIONS(2756),
    [sym_primitive_type] = ACTIONS(2756),
    [anon_sym_enum] = ACTIONS(2756),
    [anon_sym_struct] = ACTIONS(2756),
    [anon_sym_union] = ACTIONS(2756),
    [anon_sym_if] = ACTIONS(2756),
    [anon_sym_switch] = ACTIONS(2756),
    [anon_sym_case] = ACTIONS(2756),
    [anon_sym_default] = ACTIONS(2756),
    [anon_sym_while] = ACTIONS(2756),
    [anon_sym_do] = ACTIONS(2756),
    [anon_sym_for] = ACTIONS(2756),
    [anon_sym_in] = ACTIONS(2756),
    [anon_sym_return] = ACTIONS(2756),
    [anon_sym_break] = ACTIONS(2756),
    [anon_sym_continue] = ACTIONS(2756),
    [anon_sym_goto] = ACTIONS(2756),
    [anon_sym_DASH_DASH] = ACTIONS(2758),
    [anon_sym_PLUS_PLUS] = ACTIONS(2758),
    [anon_sym_sizeof] = ACTIONS(2756),
    [anon_sym___alignof__] = ACTIONS(2756),
    [anon_sym___alignof] = ACTIONS(2756),
    [anon_sym__alignof] = ACTIONS(2756),
    [anon_sym_alignof] = ACTIONS(2756),
    [anon_sym__Alignof] = ACTIONS(2756),
    [anon_sym_offsetof] = ACTIONS(2756),
    [anon_sym__Generic] = ACTIONS(2756),
    [anon_sym_asm] = ACTIONS(2756),
    [anon_sym___asm__] = ACTIONS(2756),
    [anon_sym___asm] = ACTIONS(2756),
    [sym_number_literal] = ACTIONS(2758),
    [anon_sym_L_SQUOTE] = ACTIONS(2758),
    [anon_sym_u_SQUOTE] = ACTIONS(2758),
    [anon_sym_U_SQUOTE] = ACTIONS(2758),
    [anon_sym_u8_SQUOTE] = ACTIONS(2758),
    [anon_sym_SQUOTE] = ACTIONS(2758),
    [anon_sym_AT] = ACTIONS(2756),
    [anon_sym_DQUOTE] = ACTIONS(2758),
    [anon_sym_L_DQUOTE] = ACTIONS(2758),
    [anon_sym_u_DQUOTE] = ACTIONS(2758),
    [anon_sym_U_DQUOTE] = ACTIONS(2758),
    [anon_sym_u8_DQUOTE] = ACTIONS(2758),
    [sym_true] = ACTIONS(2756),
    [sym_false] = ACTIONS(2756),
    [anon_sym_NULL] = ACTIONS(2756),
    [anon_sym_nullptr] = ACTIONS(2756),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2756),
    [anon_sym___typeof] = ACTIONS(2756),
    [anon_sym_typeof] = ACTIONS(2756),
    [anon_sym_ATimport] = ACTIONS(2758),
    [aux_sym_preproc_undef_token1] = ACTIONS(2756),
    [anon_sym_POUND] = ACTIONS(2756),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2756),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2756),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2756),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2756),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2756),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2756),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2756),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_API_AVAILABLE] = ACTIONS(2756),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_API_DEPRECATED] = ACTIONS(2756),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2756),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym___deprecated_msg] = ACTIONS(2756),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2756),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2756),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2756),
    [anon_sym_ATprotocol] = ACTIONS(2758),
    [anon_sym_ATinterface] = ACTIONS(2758),
    [anon_sym_ATimplementation] = ACTIONS(2758),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2758),
    [anon_sym_ATsynthesize] = ACTIONS(2758),
    [anon_sym_ATdynamic] = ACTIONS(2758),
    [anon_sym_ATtry] = ACTIONS(2758),
    [anon_sym___try] = ACTIONS(2756),
    [anon_sym_ATthrow] = ACTIONS(2758),
    [anon_sym_ATselector] = ACTIONS(2758),
    [anon_sym_ATavailable] = ACTIONS(2758),
    [anon_sym___builtin_available] = ACTIONS(2756),
    [anon_sym_va_arg] = ACTIONS(2756),
    [anon_sym_ATencode] = ACTIONS(2758),
    [anon_sym_ATsynchronized] = ACTIONS(2758),
    [anon_sym_BOOL] = ACTIONS(2756),
    [anon_sym_IMP] = ACTIONS(2756),
    [anon_sym_SEL] = ACTIONS(2756),
    [anon_sym_Class] = ACTIONS(2756),
    [anon_sym_id] = ACTIONS(2756),
  },
  [STATE(501)] = {
    [sym_identifier] = ACTIONS(2760),
    [aux_sym_preproc_include_token1] = ACTIONS(2760),
    [aux_sym_preproc_include_token2] = ACTIONS(2760),
    [aux_sym_preproc_def_token1] = ACTIONS(2760),
    [aux_sym_preproc_if_token1] = ACTIONS(2760),
    [aux_sym_preproc_if_token2] = ACTIONS(2760),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2760),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2760),
    [aux_sym_preproc_else_token1] = ACTIONS(2760),
    [aux_sym_preproc_elif_token1] = ACTIONS(2760),
    [sym_preproc_directive] = ACTIONS(2760),
    [anon_sym_LPAREN2] = ACTIONS(2762),
    [anon_sym_BANG] = ACTIONS(2762),
    [anon_sym_TILDE] = ACTIONS(2762),
    [anon_sym_DASH] = ACTIONS(2760),
    [anon_sym_PLUS] = ACTIONS(2760),
    [anon_sym_STAR] = ACTIONS(2762),
    [anon_sym_CARET] = ACTIONS(2762),
    [anon_sym_AMP] = ACTIONS(2762),
    [anon_sym_SEMI] = ACTIONS(2762),
    [anon_sym___extension__] = ACTIONS(2760),
    [anon_sym_typedef] = ACTIONS(2760),
    [anon_sym_extern] = ACTIONS(2760),
    [anon_sym___attribute__] = ACTIONS(2760),
    [anon_sym___attribute] = ACTIONS(2760),
    [anon_sym_noreturn] = ACTIONS(2760),
    [anon_sym_LBRACK] = ACTIONS(2762),
    [anon_sym___declspec] = ACTIONS(2760),
    [anon_sym___cdecl] = ACTIONS(2760),
    [anon_sym___clrcall] = ACTIONS(2760),
    [anon_sym___stdcall] = ACTIONS(2760),
    [anon_sym___fastcall] = ACTIONS(2760),
    [anon_sym___thiscall] = ACTIONS(2760),
    [anon_sym___vectorcall] = ACTIONS(2760),
    [anon_sym_LBRACE] = ACTIONS(2762),
    [anon_sym_signed] = ACTIONS(2760),
    [anon_sym_unsigned] = ACTIONS(2760),
    [anon_sym_long] = ACTIONS(2760),
    [anon_sym_short] = ACTIONS(2760),
    [anon_sym_static] = ACTIONS(2760),
    [anon_sym_ATautoreleasepool] = ACTIONS(2762),
    [anon_sym_auto] = ACTIONS(2760),
    [anon_sym_register] = ACTIONS(2760),
    [anon_sym_inline] = ACTIONS(2760),
    [anon_sym___inline] = ACTIONS(2760),
    [anon_sym___inline__] = ACTIONS(2760),
    [anon_sym___forceinline] = ACTIONS(2760),
    [anon_sym_thread_local] = ACTIONS(2760),
    [anon_sym___thread] = ACTIONS(2760),
    [anon_sym_CG_EXTERN] = ACTIONS(2760),
    [anon_sym_CG_INLINE] = ACTIONS(2760),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2760),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2760),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2760),
    [anon_sym_IBOutlet] = ACTIONS(2760),
    [anon_sym_IBInspectable] = ACTIONS(2760),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2760),
    [anon_sym_NS_INLINE] = ACTIONS(2760),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2760),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2760),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2760),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2760),
    [anon_sym_const] = ACTIONS(2760),
    [anon_sym_constexpr] = ACTIONS(2760),
    [anon_sym_volatile] = ACTIONS(2760),
    [anon_sym_restrict] = ACTIONS(2760),
    [anon_sym___restrict__] = ACTIONS(2760),
    [anon_sym__Atomic] = ACTIONS(2760),
    [anon_sym__Noreturn] = ACTIONS(2760),
    [anon_sym__Nonnull] = ACTIONS(2760),
    [anon_sym_nullable] = ACTIONS(2760),
    [anon_sym__Complex] = ACTIONS(2760),
    [anon_sym__Nullable] = ACTIONS(2760),
    [anon_sym__Nullable_result] = ACTIONS(2760),
    [anon_sym__Null_unspecified] = ACTIONS(2760),
    [anon_sym___autoreleasing] = ACTIONS(2760),
    [anon_sym___block] = ACTIONS(2760),
    [anon_sym___bridge] = ACTIONS(2760),
    [anon_sym___bridge_retained] = ACTIONS(2760),
    [anon_sym___bridge_transfer] = ACTIONS(2760),
    [anon_sym___complex] = ACTIONS(2760),
    [anon_sym___const] = ACTIONS(2760),
    [anon_sym___imag] = ACTIONS(2760),
    [anon_sym___kindof] = ACTIONS(2760),
    [anon_sym___nonnull] = ACTIONS(2760),
    [anon_sym___nullable] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2760),
    [anon_sym___real] = ACTIONS(2760),
    [anon_sym___strong] = ACTIONS(2760),
    [anon_sym___unsafe_unretained] = ACTIONS(2760),
    [anon_sym___unused] = ACTIONS(2760),
    [anon_sym___weak] = ACTIONS(2760),
    [anon_sym_alignas] = ACTIONS(2760),
    [anon_sym__Alignas] = ACTIONS(2760),
    [sym_primitive_type] = ACTIONS(2760),
    [anon_sym_enum] = ACTIONS(2760),
    [anon_sym_struct] = ACTIONS(2760),
    [anon_sym_union] = ACTIONS(2760),
    [anon_sym_if] = ACTIONS(2760),
    [anon_sym_switch] = ACTIONS(2760),
    [anon_sym_case] = ACTIONS(2760),
    [anon_sym_default] = ACTIONS(2760),
    [anon_sym_while] = ACTIONS(2760),
    [anon_sym_do] = ACTIONS(2760),
    [anon_sym_for] = ACTIONS(2760),
    [anon_sym_in] = ACTIONS(2760),
    [anon_sym_return] = ACTIONS(2760),
    [anon_sym_break] = ACTIONS(2760),
    [anon_sym_continue] = ACTIONS(2760),
    [anon_sym_goto] = ACTIONS(2760),
    [anon_sym_DASH_DASH] = ACTIONS(2762),
    [anon_sym_PLUS_PLUS] = ACTIONS(2762),
    [anon_sym_sizeof] = ACTIONS(2760),
    [anon_sym___alignof__] = ACTIONS(2760),
    [anon_sym___alignof] = ACTIONS(2760),
    [anon_sym__alignof] = ACTIONS(2760),
    [anon_sym_alignof] = ACTIONS(2760),
    [anon_sym__Alignof] = ACTIONS(2760),
    [anon_sym_offsetof] = ACTIONS(2760),
    [anon_sym__Generic] = ACTIONS(2760),
    [anon_sym_asm] = ACTIONS(2760),
    [anon_sym___asm__] = ACTIONS(2760),
    [anon_sym___asm] = ACTIONS(2760),
    [sym_number_literal] = ACTIONS(2762),
    [anon_sym_L_SQUOTE] = ACTIONS(2762),
    [anon_sym_u_SQUOTE] = ACTIONS(2762),
    [anon_sym_U_SQUOTE] = ACTIONS(2762),
    [anon_sym_u8_SQUOTE] = ACTIONS(2762),
    [anon_sym_SQUOTE] = ACTIONS(2762),
    [anon_sym_AT] = ACTIONS(2760),
    [anon_sym_DQUOTE] = ACTIONS(2762),
    [anon_sym_L_DQUOTE] = ACTIONS(2762),
    [anon_sym_u_DQUOTE] = ACTIONS(2762),
    [anon_sym_U_DQUOTE] = ACTIONS(2762),
    [anon_sym_u8_DQUOTE] = ACTIONS(2762),
    [sym_true] = ACTIONS(2760),
    [sym_false] = ACTIONS(2760),
    [anon_sym_NULL] = ACTIONS(2760),
    [anon_sym_nullptr] = ACTIONS(2760),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2760),
    [anon_sym___typeof] = ACTIONS(2760),
    [anon_sym_typeof] = ACTIONS(2760),
    [anon_sym_ATimport] = ACTIONS(2762),
    [aux_sym_preproc_undef_token1] = ACTIONS(2760),
    [anon_sym_POUND] = ACTIONS(2760),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2760),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2760),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2760),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2760),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2760),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2760),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2760),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_API_AVAILABLE] = ACTIONS(2760),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_API_DEPRECATED] = ACTIONS(2760),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2760),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym___deprecated_msg] = ACTIONS(2760),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2760),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2760),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2760),
    [anon_sym_ATprotocol] = ACTIONS(2762),
    [anon_sym_ATinterface] = ACTIONS(2762),
    [anon_sym_ATimplementation] = ACTIONS(2762),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2762),
    [anon_sym_ATsynthesize] = ACTIONS(2762),
    [anon_sym_ATdynamic] = ACTIONS(2762),
    [anon_sym_ATtry] = ACTIONS(2762),
    [anon_sym___try] = ACTIONS(2760),
    [anon_sym_ATthrow] = ACTIONS(2762),
    [anon_sym_ATselector] = ACTIONS(2762),
    [anon_sym_ATavailable] = ACTIONS(2762),
    [anon_sym___builtin_available] = ACTIONS(2760),
    [anon_sym_va_arg] = ACTIONS(2760),
    [anon_sym_ATencode] = ACTIONS(2762),
    [anon_sym_ATsynchronized] = ACTIONS(2762),
    [anon_sym_BOOL] = ACTIONS(2760),
    [anon_sym_IMP] = ACTIONS(2760),
    [anon_sym_SEL] = ACTIONS(2760),
    [anon_sym_Class] = ACTIONS(2760),
    [anon_sym_id] = ACTIONS(2760),
  },
  [STATE(502)] = {
    [sym_identifier] = ACTIONS(2614),
    [aux_sym_preproc_include_token1] = ACTIONS(2614),
    [aux_sym_preproc_include_token2] = ACTIONS(2614),
    [aux_sym_preproc_def_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token2] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2614),
    [aux_sym_preproc_else_token1] = ACTIONS(2614),
    [aux_sym_preproc_elif_token1] = ACTIONS(2614),
    [sym_preproc_directive] = ACTIONS(2614),
    [anon_sym_LPAREN2] = ACTIONS(2616),
    [anon_sym_BANG] = ACTIONS(2616),
    [anon_sym_TILDE] = ACTIONS(2616),
    [anon_sym_DASH] = ACTIONS(2614),
    [anon_sym_PLUS] = ACTIONS(2614),
    [anon_sym_STAR] = ACTIONS(2616),
    [anon_sym_CARET] = ACTIONS(2616),
    [anon_sym_AMP] = ACTIONS(2616),
    [anon_sym_SEMI] = ACTIONS(2616),
    [anon_sym___extension__] = ACTIONS(2614),
    [anon_sym_typedef] = ACTIONS(2614),
    [anon_sym_extern] = ACTIONS(2614),
    [anon_sym___attribute__] = ACTIONS(2614),
    [anon_sym___attribute] = ACTIONS(2614),
    [anon_sym_noreturn] = ACTIONS(2614),
    [anon_sym_LBRACK] = ACTIONS(2616),
    [anon_sym___declspec] = ACTIONS(2614),
    [anon_sym___cdecl] = ACTIONS(2614),
    [anon_sym___clrcall] = ACTIONS(2614),
    [anon_sym___stdcall] = ACTIONS(2614),
    [anon_sym___fastcall] = ACTIONS(2614),
    [anon_sym___thiscall] = ACTIONS(2614),
    [anon_sym___vectorcall] = ACTIONS(2614),
    [anon_sym_LBRACE] = ACTIONS(2616),
    [anon_sym_signed] = ACTIONS(2614),
    [anon_sym_unsigned] = ACTIONS(2614),
    [anon_sym_long] = ACTIONS(2614),
    [anon_sym_short] = ACTIONS(2614),
    [anon_sym_static] = ACTIONS(2614),
    [anon_sym_ATautoreleasepool] = ACTIONS(2616),
    [anon_sym_auto] = ACTIONS(2614),
    [anon_sym_register] = ACTIONS(2614),
    [anon_sym_inline] = ACTIONS(2614),
    [anon_sym___inline] = ACTIONS(2614),
    [anon_sym___inline__] = ACTIONS(2614),
    [anon_sym___forceinline] = ACTIONS(2614),
    [anon_sym_thread_local] = ACTIONS(2614),
    [anon_sym___thread] = ACTIONS(2614),
    [anon_sym_CG_EXTERN] = ACTIONS(2614),
    [anon_sym_CG_INLINE] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2614),
    [anon_sym_IBOutlet] = ACTIONS(2614),
    [anon_sym_IBInspectable] = ACTIONS(2614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2614),
    [anon_sym_NS_INLINE] = ACTIONS(2614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2614),
    [anon_sym_const] = ACTIONS(2614),
    [anon_sym_constexpr] = ACTIONS(2614),
    [anon_sym_volatile] = ACTIONS(2614),
    [anon_sym_restrict] = ACTIONS(2614),
    [anon_sym___restrict__] = ACTIONS(2614),
    [anon_sym__Atomic] = ACTIONS(2614),
    [anon_sym__Noreturn] = ACTIONS(2614),
    [anon_sym__Nonnull] = ACTIONS(2614),
    [anon_sym_nullable] = ACTIONS(2614),
    [anon_sym__Complex] = ACTIONS(2614),
    [anon_sym__Nullable] = ACTIONS(2614),
    [anon_sym__Nullable_result] = ACTIONS(2614),
    [anon_sym__Null_unspecified] = ACTIONS(2614),
    [anon_sym___autoreleasing] = ACTIONS(2614),
    [anon_sym___block] = ACTIONS(2614),
    [anon_sym___bridge] = ACTIONS(2614),
    [anon_sym___bridge_retained] = ACTIONS(2614),
    [anon_sym___bridge_transfer] = ACTIONS(2614),
    [anon_sym___complex] = ACTIONS(2614),
    [anon_sym___const] = ACTIONS(2614),
    [anon_sym___imag] = ACTIONS(2614),
    [anon_sym___kindof] = ACTIONS(2614),
    [anon_sym___nonnull] = ACTIONS(2614),
    [anon_sym___nullable] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2614),
    [anon_sym___real] = ACTIONS(2614),
    [anon_sym___strong] = ACTIONS(2614),
    [anon_sym___unsafe_unretained] = ACTIONS(2614),
    [anon_sym___unused] = ACTIONS(2614),
    [anon_sym___weak] = ACTIONS(2614),
    [anon_sym_alignas] = ACTIONS(2614),
    [anon_sym__Alignas] = ACTIONS(2614),
    [sym_primitive_type] = ACTIONS(2614),
    [anon_sym_enum] = ACTIONS(2614),
    [anon_sym_struct] = ACTIONS(2614),
    [anon_sym_union] = ACTIONS(2614),
    [anon_sym_if] = ACTIONS(2614),
    [anon_sym_switch] = ACTIONS(2614),
    [anon_sym_case] = ACTIONS(2614),
    [anon_sym_default] = ACTIONS(2614),
    [anon_sym_while] = ACTIONS(2614),
    [anon_sym_do] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2614),
    [anon_sym_in] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_break] = ACTIONS(2614),
    [anon_sym_continue] = ACTIONS(2614),
    [anon_sym_goto] = ACTIONS(2614),
    [anon_sym_DASH_DASH] = ACTIONS(2616),
    [anon_sym_PLUS_PLUS] = ACTIONS(2616),
    [anon_sym_sizeof] = ACTIONS(2614),
    [anon_sym___alignof__] = ACTIONS(2614),
    [anon_sym___alignof] = ACTIONS(2614),
    [anon_sym__alignof] = ACTIONS(2614),
    [anon_sym_alignof] = ACTIONS(2614),
    [anon_sym__Alignof] = ACTIONS(2614),
    [anon_sym_offsetof] = ACTIONS(2614),
    [anon_sym__Generic] = ACTIONS(2614),
    [anon_sym_asm] = ACTIONS(2614),
    [anon_sym___asm__] = ACTIONS(2614),
    [anon_sym___asm] = ACTIONS(2614),
    [sym_number_literal] = ACTIONS(2616),
    [anon_sym_L_SQUOTE] = ACTIONS(2616),
    [anon_sym_u_SQUOTE] = ACTIONS(2616),
    [anon_sym_U_SQUOTE] = ACTIONS(2616),
    [anon_sym_u8_SQUOTE] = ACTIONS(2616),
    [anon_sym_SQUOTE] = ACTIONS(2616),
    [anon_sym_AT] = ACTIONS(2614),
    [anon_sym_DQUOTE] = ACTIONS(2616),
    [anon_sym_L_DQUOTE] = ACTIONS(2616),
    [anon_sym_u_DQUOTE] = ACTIONS(2616),
    [anon_sym_U_DQUOTE] = ACTIONS(2616),
    [anon_sym_u8_DQUOTE] = ACTIONS(2616),
    [sym_true] = ACTIONS(2614),
    [sym_false] = ACTIONS(2614),
    [anon_sym_NULL] = ACTIONS(2614),
    [anon_sym_nullptr] = ACTIONS(2614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2614),
    [anon_sym___typeof] = ACTIONS(2614),
    [anon_sym_typeof] = ACTIONS(2614),
    [anon_sym_ATimport] = ACTIONS(2616),
    [aux_sym_preproc_undef_token1] = ACTIONS(2614),
    [anon_sym_POUND] = ACTIONS(2614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_API_AVAILABLE] = ACTIONS(2614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_API_DEPRECATED] = ACTIONS(2614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym___deprecated_msg] = ACTIONS(2614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2614),
    [anon_sym_ATprotocol] = ACTIONS(2616),
    [anon_sym_ATinterface] = ACTIONS(2616),
    [anon_sym_ATimplementation] = ACTIONS(2616),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2616),
    [anon_sym_ATsynthesize] = ACTIONS(2616),
    [anon_sym_ATdynamic] = ACTIONS(2616),
    [anon_sym_ATtry] = ACTIONS(2616),
    [anon_sym___try] = ACTIONS(2614),
    [anon_sym_ATthrow] = ACTIONS(2616),
    [anon_sym_ATselector] = ACTIONS(2616),
    [anon_sym_ATavailable] = ACTIONS(2616),
    [anon_sym___builtin_available] = ACTIONS(2614),
    [anon_sym_va_arg] = ACTIONS(2614),
    [anon_sym_ATencode] = ACTIONS(2616),
    [anon_sym_ATsynchronized] = ACTIONS(2616),
    [anon_sym_BOOL] = ACTIONS(2614),
    [anon_sym_IMP] = ACTIONS(2614),
    [anon_sym_SEL] = ACTIONS(2614),
    [anon_sym_Class] = ACTIONS(2614),
    [anon_sym_id] = ACTIONS(2614),
  },
  [STATE(503)] = {
    [sym_identifier] = ACTIONS(2588),
    [aux_sym_preproc_include_token1] = ACTIONS(2588),
    [aux_sym_preproc_include_token2] = ACTIONS(2588),
    [aux_sym_preproc_def_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token2] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2588),
    [aux_sym_preproc_else_token1] = ACTIONS(2588),
    [aux_sym_preproc_elif_token1] = ACTIONS(2588),
    [sym_preproc_directive] = ACTIONS(2588),
    [anon_sym_LPAREN2] = ACTIONS(2590),
    [anon_sym_BANG] = ACTIONS(2590),
    [anon_sym_TILDE] = ACTIONS(2590),
    [anon_sym_DASH] = ACTIONS(2588),
    [anon_sym_PLUS] = ACTIONS(2588),
    [anon_sym_STAR] = ACTIONS(2590),
    [anon_sym_CARET] = ACTIONS(2590),
    [anon_sym_AMP] = ACTIONS(2590),
    [anon_sym_SEMI] = ACTIONS(2590),
    [anon_sym___extension__] = ACTIONS(2588),
    [anon_sym_typedef] = ACTIONS(2588),
    [anon_sym_extern] = ACTIONS(2588),
    [anon_sym___attribute__] = ACTIONS(2588),
    [anon_sym___attribute] = ACTIONS(2588),
    [anon_sym_noreturn] = ACTIONS(2588),
    [anon_sym_LBRACK] = ACTIONS(2590),
    [anon_sym___declspec] = ACTIONS(2588),
    [anon_sym___cdecl] = ACTIONS(2588),
    [anon_sym___clrcall] = ACTIONS(2588),
    [anon_sym___stdcall] = ACTIONS(2588),
    [anon_sym___fastcall] = ACTIONS(2588),
    [anon_sym___thiscall] = ACTIONS(2588),
    [anon_sym___vectorcall] = ACTIONS(2588),
    [anon_sym_LBRACE] = ACTIONS(2590),
    [anon_sym_signed] = ACTIONS(2588),
    [anon_sym_unsigned] = ACTIONS(2588),
    [anon_sym_long] = ACTIONS(2588),
    [anon_sym_short] = ACTIONS(2588),
    [anon_sym_static] = ACTIONS(2588),
    [anon_sym_ATautoreleasepool] = ACTIONS(2590),
    [anon_sym_auto] = ACTIONS(2588),
    [anon_sym_register] = ACTIONS(2588),
    [anon_sym_inline] = ACTIONS(2588),
    [anon_sym___inline] = ACTIONS(2588),
    [anon_sym___inline__] = ACTIONS(2588),
    [anon_sym___forceinline] = ACTIONS(2588),
    [anon_sym_thread_local] = ACTIONS(2588),
    [anon_sym___thread] = ACTIONS(2588),
    [anon_sym_CG_EXTERN] = ACTIONS(2588),
    [anon_sym_CG_INLINE] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2588),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2588),
    [anon_sym_IBOutlet] = ACTIONS(2588),
    [anon_sym_IBInspectable] = ACTIONS(2588),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2588),
    [anon_sym_NS_INLINE] = ACTIONS(2588),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2588),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2588),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2588),
    [anon_sym_const] = ACTIONS(2588),
    [anon_sym_constexpr] = ACTIONS(2588),
    [anon_sym_volatile] = ACTIONS(2588),
    [anon_sym_restrict] = ACTIONS(2588),
    [anon_sym___restrict__] = ACTIONS(2588),
    [anon_sym__Atomic] = ACTIONS(2588),
    [anon_sym__Noreturn] = ACTIONS(2588),
    [anon_sym__Nonnull] = ACTIONS(2588),
    [anon_sym_nullable] = ACTIONS(2588),
    [anon_sym__Complex] = ACTIONS(2588),
    [anon_sym__Nullable] = ACTIONS(2588),
    [anon_sym__Nullable_result] = ACTIONS(2588),
    [anon_sym__Null_unspecified] = ACTIONS(2588),
    [anon_sym___autoreleasing] = ACTIONS(2588),
    [anon_sym___block] = ACTIONS(2588),
    [anon_sym___bridge] = ACTIONS(2588),
    [anon_sym___bridge_retained] = ACTIONS(2588),
    [anon_sym___bridge_transfer] = ACTIONS(2588),
    [anon_sym___complex] = ACTIONS(2588),
    [anon_sym___const] = ACTIONS(2588),
    [anon_sym___imag] = ACTIONS(2588),
    [anon_sym___kindof] = ACTIONS(2588),
    [anon_sym___nonnull] = ACTIONS(2588),
    [anon_sym___nullable] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2588),
    [anon_sym___real] = ACTIONS(2588),
    [anon_sym___strong] = ACTIONS(2588),
    [anon_sym___unsafe_unretained] = ACTIONS(2588),
    [anon_sym___unused] = ACTIONS(2588),
    [anon_sym___weak] = ACTIONS(2588),
    [anon_sym_alignas] = ACTIONS(2588),
    [anon_sym__Alignas] = ACTIONS(2588),
    [sym_primitive_type] = ACTIONS(2588),
    [anon_sym_enum] = ACTIONS(2588),
    [anon_sym_struct] = ACTIONS(2588),
    [anon_sym_union] = ACTIONS(2588),
    [anon_sym_if] = ACTIONS(2588),
    [anon_sym_switch] = ACTIONS(2588),
    [anon_sym_case] = ACTIONS(2588),
    [anon_sym_default] = ACTIONS(2588),
    [anon_sym_while] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2588),
    [anon_sym_for] = ACTIONS(2588),
    [anon_sym_in] = ACTIONS(2588),
    [anon_sym_return] = ACTIONS(2588),
    [anon_sym_break] = ACTIONS(2588),
    [anon_sym_continue] = ACTIONS(2588),
    [anon_sym_goto] = ACTIONS(2588),
    [anon_sym_DASH_DASH] = ACTIONS(2590),
    [anon_sym_PLUS_PLUS] = ACTIONS(2590),
    [anon_sym_sizeof] = ACTIONS(2588),
    [anon_sym___alignof__] = ACTIONS(2588),
    [anon_sym___alignof] = ACTIONS(2588),
    [anon_sym__alignof] = ACTIONS(2588),
    [anon_sym_alignof] = ACTIONS(2588),
    [anon_sym__Alignof] = ACTIONS(2588),
    [anon_sym_offsetof] = ACTIONS(2588),
    [anon_sym__Generic] = ACTIONS(2588),
    [anon_sym_asm] = ACTIONS(2588),
    [anon_sym___asm__] = ACTIONS(2588),
    [anon_sym___asm] = ACTIONS(2588),
    [sym_number_literal] = ACTIONS(2590),
    [anon_sym_L_SQUOTE] = ACTIONS(2590),
    [anon_sym_u_SQUOTE] = ACTIONS(2590),
    [anon_sym_U_SQUOTE] = ACTIONS(2590),
    [anon_sym_u8_SQUOTE] = ACTIONS(2590),
    [anon_sym_SQUOTE] = ACTIONS(2590),
    [anon_sym_AT] = ACTIONS(2588),
    [anon_sym_DQUOTE] = ACTIONS(2590),
    [anon_sym_L_DQUOTE] = ACTIONS(2590),
    [anon_sym_u_DQUOTE] = ACTIONS(2590),
    [anon_sym_U_DQUOTE] = ACTIONS(2590),
    [anon_sym_u8_DQUOTE] = ACTIONS(2590),
    [sym_true] = ACTIONS(2588),
    [sym_false] = ACTIONS(2588),
    [anon_sym_NULL] = ACTIONS(2588),
    [anon_sym_nullptr] = ACTIONS(2588),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2588),
    [anon_sym___typeof] = ACTIONS(2588),
    [anon_sym_typeof] = ACTIONS(2588),
    [anon_sym_ATimport] = ACTIONS(2590),
    [aux_sym_preproc_undef_token1] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2588),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2588),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2588),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2588),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_API_AVAILABLE] = ACTIONS(2588),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_API_DEPRECATED] = ACTIONS(2588),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym___deprecated_msg] = ACTIONS(2588),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2588),
    [anon_sym_ATprotocol] = ACTIONS(2590),
    [anon_sym_ATinterface] = ACTIONS(2590),
    [anon_sym_ATimplementation] = ACTIONS(2590),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2590),
    [anon_sym_ATsynthesize] = ACTIONS(2590),
    [anon_sym_ATdynamic] = ACTIONS(2590),
    [anon_sym_ATtry] = ACTIONS(2590),
    [anon_sym___try] = ACTIONS(2588),
    [anon_sym_ATthrow] = ACTIONS(2590),
    [anon_sym_ATselector] = ACTIONS(2590),
    [anon_sym_ATavailable] = ACTIONS(2590),
    [anon_sym___builtin_available] = ACTIONS(2588),
    [anon_sym_va_arg] = ACTIONS(2588),
    [anon_sym_ATencode] = ACTIONS(2590),
    [anon_sym_ATsynchronized] = ACTIONS(2590),
    [anon_sym_BOOL] = ACTIONS(2588),
    [anon_sym_IMP] = ACTIONS(2588),
    [anon_sym_SEL] = ACTIONS(2588),
    [anon_sym_Class] = ACTIONS(2588),
    [anon_sym_id] = ACTIONS(2588),
  },
  [STATE(504)] = {
    [sym_identifier] = ACTIONS(2852),
    [aux_sym_preproc_include_token1] = ACTIONS(2852),
    [aux_sym_preproc_include_token2] = ACTIONS(2852),
    [aux_sym_preproc_def_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token2] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2852),
    [aux_sym_preproc_else_token1] = ACTIONS(2852),
    [aux_sym_preproc_elif_token1] = ACTIONS(2852),
    [sym_preproc_directive] = ACTIONS(2852),
    [anon_sym_LPAREN2] = ACTIONS(2854),
    [anon_sym_BANG] = ACTIONS(2854),
    [anon_sym_TILDE] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2852),
    [anon_sym_PLUS] = ACTIONS(2852),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_SEMI] = ACTIONS(2854),
    [anon_sym___extension__] = ACTIONS(2852),
    [anon_sym_typedef] = ACTIONS(2852),
    [anon_sym_extern] = ACTIONS(2852),
    [anon_sym___attribute__] = ACTIONS(2852),
    [anon_sym___attribute] = ACTIONS(2852),
    [anon_sym_noreturn] = ACTIONS(2852),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym___declspec] = ACTIONS(2852),
    [anon_sym___cdecl] = ACTIONS(2852),
    [anon_sym___clrcall] = ACTIONS(2852),
    [anon_sym___stdcall] = ACTIONS(2852),
    [anon_sym___fastcall] = ACTIONS(2852),
    [anon_sym___thiscall] = ACTIONS(2852),
    [anon_sym___vectorcall] = ACTIONS(2852),
    [anon_sym_LBRACE] = ACTIONS(2854),
    [anon_sym_signed] = ACTIONS(2852),
    [anon_sym_unsigned] = ACTIONS(2852),
    [anon_sym_long] = ACTIONS(2852),
    [anon_sym_short] = ACTIONS(2852),
    [anon_sym_static] = ACTIONS(2852),
    [anon_sym_ATautoreleasepool] = ACTIONS(2854),
    [anon_sym_auto] = ACTIONS(2852),
    [anon_sym_register] = ACTIONS(2852),
    [anon_sym_inline] = ACTIONS(2852),
    [anon_sym___inline] = ACTIONS(2852),
    [anon_sym___inline__] = ACTIONS(2852),
    [anon_sym___forceinline] = ACTIONS(2852),
    [anon_sym_thread_local] = ACTIONS(2852),
    [anon_sym___thread] = ACTIONS(2852),
    [anon_sym_CG_EXTERN] = ACTIONS(2852),
    [anon_sym_CG_INLINE] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2852),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2852),
    [anon_sym_IBOutlet] = ACTIONS(2852),
    [anon_sym_IBInspectable] = ACTIONS(2852),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2852),
    [anon_sym_NS_INLINE] = ACTIONS(2852),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2852),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2852),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2852),
    [anon_sym_const] = ACTIONS(2852),
    [anon_sym_constexpr] = ACTIONS(2852),
    [anon_sym_volatile] = ACTIONS(2852),
    [anon_sym_restrict] = ACTIONS(2852),
    [anon_sym___restrict__] = ACTIONS(2852),
    [anon_sym__Atomic] = ACTIONS(2852),
    [anon_sym__Noreturn] = ACTIONS(2852),
    [anon_sym__Nonnull] = ACTIONS(2852),
    [anon_sym_nullable] = ACTIONS(2852),
    [anon_sym__Complex] = ACTIONS(2852),
    [anon_sym__Nullable] = ACTIONS(2852),
    [anon_sym__Nullable_result] = ACTIONS(2852),
    [anon_sym__Null_unspecified] = ACTIONS(2852),
    [anon_sym___autoreleasing] = ACTIONS(2852),
    [anon_sym___block] = ACTIONS(2852),
    [anon_sym___bridge] = ACTIONS(2852),
    [anon_sym___bridge_retained] = ACTIONS(2852),
    [anon_sym___bridge_transfer] = ACTIONS(2852),
    [anon_sym___complex] = ACTIONS(2852),
    [anon_sym___const] = ACTIONS(2852),
    [anon_sym___imag] = ACTIONS(2852),
    [anon_sym___kindof] = ACTIONS(2852),
    [anon_sym___nonnull] = ACTIONS(2852),
    [anon_sym___nullable] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2852),
    [anon_sym___real] = ACTIONS(2852),
    [anon_sym___strong] = ACTIONS(2852),
    [anon_sym___unsafe_unretained] = ACTIONS(2852),
    [anon_sym___unused] = ACTIONS(2852),
    [anon_sym___weak] = ACTIONS(2852),
    [anon_sym_alignas] = ACTIONS(2852),
    [anon_sym__Alignas] = ACTIONS(2852),
    [sym_primitive_type] = ACTIONS(2852),
    [anon_sym_enum] = ACTIONS(2852),
    [anon_sym_struct] = ACTIONS(2852),
    [anon_sym_union] = ACTIONS(2852),
    [anon_sym_if] = ACTIONS(2852),
    [anon_sym_switch] = ACTIONS(2852),
    [anon_sym_case] = ACTIONS(2852),
    [anon_sym_default] = ACTIONS(2852),
    [anon_sym_while] = ACTIONS(2852),
    [anon_sym_do] = ACTIONS(2852),
    [anon_sym_for] = ACTIONS(2852),
    [anon_sym_in] = ACTIONS(2852),
    [anon_sym_return] = ACTIONS(2852),
    [anon_sym_break] = ACTIONS(2852),
    [anon_sym_continue] = ACTIONS(2852),
    [anon_sym_goto] = ACTIONS(2852),
    [anon_sym_DASH_DASH] = ACTIONS(2854),
    [anon_sym_PLUS_PLUS] = ACTIONS(2854),
    [anon_sym_sizeof] = ACTIONS(2852),
    [anon_sym___alignof__] = ACTIONS(2852),
    [anon_sym___alignof] = ACTIONS(2852),
    [anon_sym__alignof] = ACTIONS(2852),
    [anon_sym_alignof] = ACTIONS(2852),
    [anon_sym__Alignof] = ACTIONS(2852),
    [anon_sym_offsetof] = ACTIONS(2852),
    [anon_sym__Generic] = ACTIONS(2852),
    [anon_sym_asm] = ACTIONS(2852),
    [anon_sym___asm__] = ACTIONS(2852),
    [anon_sym___asm] = ACTIONS(2852),
    [sym_number_literal] = ACTIONS(2854),
    [anon_sym_L_SQUOTE] = ACTIONS(2854),
    [anon_sym_u_SQUOTE] = ACTIONS(2854),
    [anon_sym_U_SQUOTE] = ACTIONS(2854),
    [anon_sym_u8_SQUOTE] = ACTIONS(2854),
    [anon_sym_SQUOTE] = ACTIONS(2854),
    [anon_sym_AT] = ACTIONS(2852),
    [anon_sym_DQUOTE] = ACTIONS(2854),
    [anon_sym_L_DQUOTE] = ACTIONS(2854),
    [anon_sym_u_DQUOTE] = ACTIONS(2854),
    [anon_sym_U_DQUOTE] = ACTIONS(2854),
    [anon_sym_u8_DQUOTE] = ACTIONS(2854),
    [sym_true] = ACTIONS(2852),
    [sym_false] = ACTIONS(2852),
    [anon_sym_NULL] = ACTIONS(2852),
    [anon_sym_nullptr] = ACTIONS(2852),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2852),
    [anon_sym___typeof] = ACTIONS(2852),
    [anon_sym_typeof] = ACTIONS(2852),
    [anon_sym_ATimport] = ACTIONS(2854),
    [aux_sym_preproc_undef_token1] = ACTIONS(2852),
    [anon_sym_POUND] = ACTIONS(2852),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2852),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2852),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2852),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_API_AVAILABLE] = ACTIONS(2852),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_API_DEPRECATED] = ACTIONS(2852),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym___deprecated_msg] = ACTIONS(2852),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2852),
    [anon_sym_ATprotocol] = ACTIONS(2854),
    [anon_sym_ATinterface] = ACTIONS(2854),
    [anon_sym_ATimplementation] = ACTIONS(2854),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2854),
    [anon_sym_ATsynthesize] = ACTIONS(2854),
    [anon_sym_ATdynamic] = ACTIONS(2854),
    [anon_sym_ATtry] = ACTIONS(2854),
    [anon_sym___try] = ACTIONS(2852),
    [anon_sym_ATthrow] = ACTIONS(2854),
    [anon_sym_ATselector] = ACTIONS(2854),
    [anon_sym_ATavailable] = ACTIONS(2854),
    [anon_sym___builtin_available] = ACTIONS(2852),
    [anon_sym_va_arg] = ACTIONS(2852),
    [anon_sym_ATencode] = ACTIONS(2854),
    [anon_sym_ATsynchronized] = ACTIONS(2854),
    [anon_sym_BOOL] = ACTIONS(2852),
    [anon_sym_IMP] = ACTIONS(2852),
    [anon_sym_SEL] = ACTIONS(2852),
    [anon_sym_Class] = ACTIONS(2852),
    [anon_sym_id] = ACTIONS(2852),
  },
  [STATE(505)] = {
    [sym_identifier] = ACTIONS(2764),
    [aux_sym_preproc_include_token1] = ACTIONS(2764),
    [aux_sym_preproc_include_token2] = ACTIONS(2764),
    [aux_sym_preproc_def_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token2] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2764),
    [aux_sym_preproc_else_token1] = ACTIONS(2764),
    [aux_sym_preproc_elif_token1] = ACTIONS(2764),
    [sym_preproc_directive] = ACTIONS(2764),
    [anon_sym_LPAREN2] = ACTIONS(2766),
    [anon_sym_BANG] = ACTIONS(2766),
    [anon_sym_TILDE] = ACTIONS(2766),
    [anon_sym_DASH] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(2764),
    [anon_sym_STAR] = ACTIONS(2766),
    [anon_sym_CARET] = ACTIONS(2766),
    [anon_sym_AMP] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2766),
    [anon_sym___extension__] = ACTIONS(2764),
    [anon_sym_typedef] = ACTIONS(2764),
    [anon_sym_extern] = ACTIONS(2764),
    [anon_sym___attribute__] = ACTIONS(2764),
    [anon_sym___attribute] = ACTIONS(2764),
    [anon_sym_noreturn] = ACTIONS(2764),
    [anon_sym_LBRACK] = ACTIONS(2766),
    [anon_sym___declspec] = ACTIONS(2764),
    [anon_sym___cdecl] = ACTIONS(2764),
    [anon_sym___clrcall] = ACTIONS(2764),
    [anon_sym___stdcall] = ACTIONS(2764),
    [anon_sym___fastcall] = ACTIONS(2764),
    [anon_sym___thiscall] = ACTIONS(2764),
    [anon_sym___vectorcall] = ACTIONS(2764),
    [anon_sym_LBRACE] = ACTIONS(2766),
    [anon_sym_signed] = ACTIONS(2764),
    [anon_sym_unsigned] = ACTIONS(2764),
    [anon_sym_long] = ACTIONS(2764),
    [anon_sym_short] = ACTIONS(2764),
    [anon_sym_static] = ACTIONS(2764),
    [anon_sym_ATautoreleasepool] = ACTIONS(2766),
    [anon_sym_auto] = ACTIONS(2764),
    [anon_sym_register] = ACTIONS(2764),
    [anon_sym_inline] = ACTIONS(2764),
    [anon_sym___inline] = ACTIONS(2764),
    [anon_sym___inline__] = ACTIONS(2764),
    [anon_sym___forceinline] = ACTIONS(2764),
    [anon_sym_thread_local] = ACTIONS(2764),
    [anon_sym___thread] = ACTIONS(2764),
    [anon_sym_CG_EXTERN] = ACTIONS(2764),
    [anon_sym_CG_INLINE] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2764),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2764),
    [anon_sym_IBOutlet] = ACTIONS(2764),
    [anon_sym_IBInspectable] = ACTIONS(2764),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2764),
    [anon_sym_NS_INLINE] = ACTIONS(2764),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2764),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2764),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2764),
    [anon_sym_const] = ACTIONS(2764),
    [anon_sym_constexpr] = ACTIONS(2764),
    [anon_sym_volatile] = ACTIONS(2764),
    [anon_sym_restrict] = ACTIONS(2764),
    [anon_sym___restrict__] = ACTIONS(2764),
    [anon_sym__Atomic] = ACTIONS(2764),
    [anon_sym__Noreturn] = ACTIONS(2764),
    [anon_sym__Nonnull] = ACTIONS(2764),
    [anon_sym_nullable] = ACTIONS(2764),
    [anon_sym__Complex] = ACTIONS(2764),
    [anon_sym__Nullable] = ACTIONS(2764),
    [anon_sym__Nullable_result] = ACTIONS(2764),
    [anon_sym__Null_unspecified] = ACTIONS(2764),
    [anon_sym___autoreleasing] = ACTIONS(2764),
    [anon_sym___block] = ACTIONS(2764),
    [anon_sym___bridge] = ACTIONS(2764),
    [anon_sym___bridge_retained] = ACTIONS(2764),
    [anon_sym___bridge_transfer] = ACTIONS(2764),
    [anon_sym___complex] = ACTIONS(2764),
    [anon_sym___const] = ACTIONS(2764),
    [anon_sym___imag] = ACTIONS(2764),
    [anon_sym___kindof] = ACTIONS(2764),
    [anon_sym___nonnull] = ACTIONS(2764),
    [anon_sym___nullable] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2764),
    [anon_sym___real] = ACTIONS(2764),
    [anon_sym___strong] = ACTIONS(2764),
    [anon_sym___unsafe_unretained] = ACTIONS(2764),
    [anon_sym___unused] = ACTIONS(2764),
    [anon_sym___weak] = ACTIONS(2764),
    [anon_sym_alignas] = ACTIONS(2764),
    [anon_sym__Alignas] = ACTIONS(2764),
    [sym_primitive_type] = ACTIONS(2764),
    [anon_sym_enum] = ACTIONS(2764),
    [anon_sym_struct] = ACTIONS(2764),
    [anon_sym_union] = ACTIONS(2764),
    [anon_sym_if] = ACTIONS(2764),
    [anon_sym_switch] = ACTIONS(2764),
    [anon_sym_case] = ACTIONS(2764),
    [anon_sym_default] = ACTIONS(2764),
    [anon_sym_while] = ACTIONS(2764),
    [anon_sym_do] = ACTIONS(2764),
    [anon_sym_for] = ACTIONS(2764),
    [anon_sym_in] = ACTIONS(2764),
    [anon_sym_return] = ACTIONS(2764),
    [anon_sym_break] = ACTIONS(2764),
    [anon_sym_continue] = ACTIONS(2764),
    [anon_sym_goto] = ACTIONS(2764),
    [anon_sym_DASH_DASH] = ACTIONS(2766),
    [anon_sym_PLUS_PLUS] = ACTIONS(2766),
    [anon_sym_sizeof] = ACTIONS(2764),
    [anon_sym___alignof__] = ACTIONS(2764),
    [anon_sym___alignof] = ACTIONS(2764),
    [anon_sym__alignof] = ACTIONS(2764),
    [anon_sym_alignof] = ACTIONS(2764),
    [anon_sym__Alignof] = ACTIONS(2764),
    [anon_sym_offsetof] = ACTIONS(2764),
    [anon_sym__Generic] = ACTIONS(2764),
    [anon_sym_asm] = ACTIONS(2764),
    [anon_sym___asm__] = ACTIONS(2764),
    [anon_sym___asm] = ACTIONS(2764),
    [sym_number_literal] = ACTIONS(2766),
    [anon_sym_L_SQUOTE] = ACTIONS(2766),
    [anon_sym_u_SQUOTE] = ACTIONS(2766),
    [anon_sym_U_SQUOTE] = ACTIONS(2766),
    [anon_sym_u8_SQUOTE] = ACTIONS(2766),
    [anon_sym_SQUOTE] = ACTIONS(2766),
    [anon_sym_AT] = ACTIONS(2764),
    [anon_sym_DQUOTE] = ACTIONS(2766),
    [anon_sym_L_DQUOTE] = ACTIONS(2766),
    [anon_sym_u_DQUOTE] = ACTIONS(2766),
    [anon_sym_U_DQUOTE] = ACTIONS(2766),
    [anon_sym_u8_DQUOTE] = ACTIONS(2766),
    [sym_true] = ACTIONS(2764),
    [sym_false] = ACTIONS(2764),
    [anon_sym_NULL] = ACTIONS(2764),
    [anon_sym_nullptr] = ACTIONS(2764),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2764),
    [anon_sym___typeof] = ACTIONS(2764),
    [anon_sym_typeof] = ACTIONS(2764),
    [anon_sym_ATimport] = ACTIONS(2766),
    [aux_sym_preproc_undef_token1] = ACTIONS(2764),
    [anon_sym_POUND] = ACTIONS(2764),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2764),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2764),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2764),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_API_AVAILABLE] = ACTIONS(2764),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_API_DEPRECATED] = ACTIONS(2764),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym___deprecated_msg] = ACTIONS(2764),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2764),
    [anon_sym_ATprotocol] = ACTIONS(2766),
    [anon_sym_ATinterface] = ACTIONS(2766),
    [anon_sym_ATimplementation] = ACTIONS(2766),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2766),
    [anon_sym_ATsynthesize] = ACTIONS(2766),
    [anon_sym_ATdynamic] = ACTIONS(2766),
    [anon_sym_ATtry] = ACTIONS(2766),
    [anon_sym___try] = ACTIONS(2764),
    [anon_sym_ATthrow] = ACTIONS(2766),
    [anon_sym_ATselector] = ACTIONS(2766),
    [anon_sym_ATavailable] = ACTIONS(2766),
    [anon_sym___builtin_available] = ACTIONS(2764),
    [anon_sym_va_arg] = ACTIONS(2764),
    [anon_sym_ATencode] = ACTIONS(2766),
    [anon_sym_ATsynchronized] = ACTIONS(2766),
    [anon_sym_BOOL] = ACTIONS(2764),
    [anon_sym_IMP] = ACTIONS(2764),
    [anon_sym_SEL] = ACTIONS(2764),
    [anon_sym_Class] = ACTIONS(2764),
    [anon_sym_id] = ACTIONS(2764),
  },
  [STATE(506)] = {
    [sym_identifier] = ACTIONS(2764),
    [aux_sym_preproc_include_token1] = ACTIONS(2764),
    [aux_sym_preproc_include_token2] = ACTIONS(2764),
    [aux_sym_preproc_def_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token2] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2764),
    [aux_sym_preproc_else_token1] = ACTIONS(2764),
    [aux_sym_preproc_elif_token1] = ACTIONS(2764),
    [sym_preproc_directive] = ACTIONS(2764),
    [anon_sym_LPAREN2] = ACTIONS(2766),
    [anon_sym_BANG] = ACTIONS(2766),
    [anon_sym_TILDE] = ACTIONS(2766),
    [anon_sym_DASH] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(2764),
    [anon_sym_STAR] = ACTIONS(2766),
    [anon_sym_CARET] = ACTIONS(2766),
    [anon_sym_AMP] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2766),
    [anon_sym___extension__] = ACTIONS(2764),
    [anon_sym_typedef] = ACTIONS(2764),
    [anon_sym_extern] = ACTIONS(2764),
    [anon_sym___attribute__] = ACTIONS(2764),
    [anon_sym___attribute] = ACTIONS(2764),
    [anon_sym_noreturn] = ACTIONS(2764),
    [anon_sym_LBRACK] = ACTIONS(2766),
    [anon_sym___declspec] = ACTIONS(2764),
    [anon_sym___cdecl] = ACTIONS(2764),
    [anon_sym___clrcall] = ACTIONS(2764),
    [anon_sym___stdcall] = ACTIONS(2764),
    [anon_sym___fastcall] = ACTIONS(2764),
    [anon_sym___thiscall] = ACTIONS(2764),
    [anon_sym___vectorcall] = ACTIONS(2764),
    [anon_sym_LBRACE] = ACTIONS(2766),
    [anon_sym_signed] = ACTIONS(2764),
    [anon_sym_unsigned] = ACTIONS(2764),
    [anon_sym_long] = ACTIONS(2764),
    [anon_sym_short] = ACTIONS(2764),
    [anon_sym_static] = ACTIONS(2764),
    [anon_sym_ATautoreleasepool] = ACTIONS(2766),
    [anon_sym_auto] = ACTIONS(2764),
    [anon_sym_register] = ACTIONS(2764),
    [anon_sym_inline] = ACTIONS(2764),
    [anon_sym___inline] = ACTIONS(2764),
    [anon_sym___inline__] = ACTIONS(2764),
    [anon_sym___forceinline] = ACTIONS(2764),
    [anon_sym_thread_local] = ACTIONS(2764),
    [anon_sym___thread] = ACTIONS(2764),
    [anon_sym_CG_EXTERN] = ACTIONS(2764),
    [anon_sym_CG_INLINE] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2764),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2764),
    [anon_sym_IBOutlet] = ACTIONS(2764),
    [anon_sym_IBInspectable] = ACTIONS(2764),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2764),
    [anon_sym_NS_INLINE] = ACTIONS(2764),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2764),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2764),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2764),
    [anon_sym_const] = ACTIONS(2764),
    [anon_sym_constexpr] = ACTIONS(2764),
    [anon_sym_volatile] = ACTIONS(2764),
    [anon_sym_restrict] = ACTIONS(2764),
    [anon_sym___restrict__] = ACTIONS(2764),
    [anon_sym__Atomic] = ACTIONS(2764),
    [anon_sym__Noreturn] = ACTIONS(2764),
    [anon_sym__Nonnull] = ACTIONS(2764),
    [anon_sym_nullable] = ACTIONS(2764),
    [anon_sym__Complex] = ACTIONS(2764),
    [anon_sym__Nullable] = ACTIONS(2764),
    [anon_sym__Nullable_result] = ACTIONS(2764),
    [anon_sym__Null_unspecified] = ACTIONS(2764),
    [anon_sym___autoreleasing] = ACTIONS(2764),
    [anon_sym___block] = ACTIONS(2764),
    [anon_sym___bridge] = ACTIONS(2764),
    [anon_sym___bridge_retained] = ACTIONS(2764),
    [anon_sym___bridge_transfer] = ACTIONS(2764),
    [anon_sym___complex] = ACTIONS(2764),
    [anon_sym___const] = ACTIONS(2764),
    [anon_sym___imag] = ACTIONS(2764),
    [anon_sym___kindof] = ACTIONS(2764),
    [anon_sym___nonnull] = ACTIONS(2764),
    [anon_sym___nullable] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2764),
    [anon_sym___real] = ACTIONS(2764),
    [anon_sym___strong] = ACTIONS(2764),
    [anon_sym___unsafe_unretained] = ACTIONS(2764),
    [anon_sym___unused] = ACTIONS(2764),
    [anon_sym___weak] = ACTIONS(2764),
    [anon_sym_alignas] = ACTIONS(2764),
    [anon_sym__Alignas] = ACTIONS(2764),
    [sym_primitive_type] = ACTIONS(2764),
    [anon_sym_enum] = ACTIONS(2764),
    [anon_sym_struct] = ACTIONS(2764),
    [anon_sym_union] = ACTIONS(2764),
    [anon_sym_if] = ACTIONS(2764),
    [anon_sym_switch] = ACTIONS(2764),
    [anon_sym_case] = ACTIONS(2764),
    [anon_sym_default] = ACTIONS(2764),
    [anon_sym_while] = ACTIONS(2764),
    [anon_sym_do] = ACTIONS(2764),
    [anon_sym_for] = ACTIONS(2764),
    [anon_sym_in] = ACTIONS(2764),
    [anon_sym_return] = ACTIONS(2764),
    [anon_sym_break] = ACTIONS(2764),
    [anon_sym_continue] = ACTIONS(2764),
    [anon_sym_goto] = ACTIONS(2764),
    [anon_sym_DASH_DASH] = ACTIONS(2766),
    [anon_sym_PLUS_PLUS] = ACTIONS(2766),
    [anon_sym_sizeof] = ACTIONS(2764),
    [anon_sym___alignof__] = ACTIONS(2764),
    [anon_sym___alignof] = ACTIONS(2764),
    [anon_sym__alignof] = ACTIONS(2764),
    [anon_sym_alignof] = ACTIONS(2764),
    [anon_sym__Alignof] = ACTIONS(2764),
    [anon_sym_offsetof] = ACTIONS(2764),
    [anon_sym__Generic] = ACTIONS(2764),
    [anon_sym_asm] = ACTIONS(2764),
    [anon_sym___asm__] = ACTIONS(2764),
    [anon_sym___asm] = ACTIONS(2764),
    [sym_number_literal] = ACTIONS(2766),
    [anon_sym_L_SQUOTE] = ACTIONS(2766),
    [anon_sym_u_SQUOTE] = ACTIONS(2766),
    [anon_sym_U_SQUOTE] = ACTIONS(2766),
    [anon_sym_u8_SQUOTE] = ACTIONS(2766),
    [anon_sym_SQUOTE] = ACTIONS(2766),
    [anon_sym_AT] = ACTIONS(2764),
    [anon_sym_DQUOTE] = ACTIONS(2766),
    [anon_sym_L_DQUOTE] = ACTIONS(2766),
    [anon_sym_u_DQUOTE] = ACTIONS(2766),
    [anon_sym_U_DQUOTE] = ACTIONS(2766),
    [anon_sym_u8_DQUOTE] = ACTIONS(2766),
    [sym_true] = ACTIONS(2764),
    [sym_false] = ACTIONS(2764),
    [anon_sym_NULL] = ACTIONS(2764),
    [anon_sym_nullptr] = ACTIONS(2764),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2764),
    [anon_sym___typeof] = ACTIONS(2764),
    [anon_sym_typeof] = ACTIONS(2764),
    [anon_sym_ATimport] = ACTIONS(2766),
    [aux_sym_preproc_undef_token1] = ACTIONS(2764),
    [anon_sym_POUND] = ACTIONS(2764),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2764),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2764),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2764),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_API_AVAILABLE] = ACTIONS(2764),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_API_DEPRECATED] = ACTIONS(2764),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym___deprecated_msg] = ACTIONS(2764),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2764),
    [anon_sym_ATprotocol] = ACTIONS(2766),
    [anon_sym_ATinterface] = ACTIONS(2766),
    [anon_sym_ATimplementation] = ACTIONS(2766),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2766),
    [anon_sym_ATsynthesize] = ACTIONS(2766),
    [anon_sym_ATdynamic] = ACTIONS(2766),
    [anon_sym_ATtry] = ACTIONS(2766),
    [anon_sym___try] = ACTIONS(2764),
    [anon_sym_ATthrow] = ACTIONS(2766),
    [anon_sym_ATselector] = ACTIONS(2766),
    [anon_sym_ATavailable] = ACTIONS(2766),
    [anon_sym___builtin_available] = ACTIONS(2764),
    [anon_sym_va_arg] = ACTIONS(2764),
    [anon_sym_ATencode] = ACTIONS(2766),
    [anon_sym_ATsynchronized] = ACTIONS(2766),
    [anon_sym_BOOL] = ACTIONS(2764),
    [anon_sym_IMP] = ACTIONS(2764),
    [anon_sym_SEL] = ACTIONS(2764),
    [anon_sym_Class] = ACTIONS(2764),
    [anon_sym_id] = ACTIONS(2764),
  },
  [STATE(507)] = {
    [sym_identifier] = ACTIONS(2676),
    [aux_sym_preproc_include_token1] = ACTIONS(2676),
    [aux_sym_preproc_include_token2] = ACTIONS(2676),
    [aux_sym_preproc_def_token1] = ACTIONS(2676),
    [aux_sym_preproc_if_token1] = ACTIONS(2676),
    [aux_sym_preproc_if_token2] = ACTIONS(2676),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2676),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2676),
    [aux_sym_preproc_else_token1] = ACTIONS(2676),
    [aux_sym_preproc_elif_token1] = ACTIONS(2676),
    [sym_preproc_directive] = ACTIONS(2676),
    [anon_sym_LPAREN2] = ACTIONS(2678),
    [anon_sym_BANG] = ACTIONS(2678),
    [anon_sym_TILDE] = ACTIONS(2678),
    [anon_sym_DASH] = ACTIONS(2676),
    [anon_sym_PLUS] = ACTIONS(2676),
    [anon_sym_STAR] = ACTIONS(2678),
    [anon_sym_CARET] = ACTIONS(2678),
    [anon_sym_AMP] = ACTIONS(2678),
    [anon_sym_SEMI] = ACTIONS(2678),
    [anon_sym___extension__] = ACTIONS(2676),
    [anon_sym_typedef] = ACTIONS(2676),
    [anon_sym_extern] = ACTIONS(2676),
    [anon_sym___attribute__] = ACTIONS(2676),
    [anon_sym___attribute] = ACTIONS(2676),
    [anon_sym_noreturn] = ACTIONS(2676),
    [anon_sym_LBRACK] = ACTIONS(2678),
    [anon_sym___declspec] = ACTIONS(2676),
    [anon_sym___cdecl] = ACTIONS(2676),
    [anon_sym___clrcall] = ACTIONS(2676),
    [anon_sym___stdcall] = ACTIONS(2676),
    [anon_sym___fastcall] = ACTIONS(2676),
    [anon_sym___thiscall] = ACTIONS(2676),
    [anon_sym___vectorcall] = ACTIONS(2676),
    [anon_sym_LBRACE] = ACTIONS(2678),
    [anon_sym_signed] = ACTIONS(2676),
    [anon_sym_unsigned] = ACTIONS(2676),
    [anon_sym_long] = ACTIONS(2676),
    [anon_sym_short] = ACTIONS(2676),
    [anon_sym_static] = ACTIONS(2676),
    [anon_sym_ATautoreleasepool] = ACTIONS(2678),
    [anon_sym_auto] = ACTIONS(2676),
    [anon_sym_register] = ACTIONS(2676),
    [anon_sym_inline] = ACTIONS(2676),
    [anon_sym___inline] = ACTIONS(2676),
    [anon_sym___inline__] = ACTIONS(2676),
    [anon_sym___forceinline] = ACTIONS(2676),
    [anon_sym_thread_local] = ACTIONS(2676),
    [anon_sym___thread] = ACTIONS(2676),
    [anon_sym_CG_EXTERN] = ACTIONS(2676),
    [anon_sym_CG_INLINE] = ACTIONS(2676),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2676),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2676),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2676),
    [anon_sym_IBOutlet] = ACTIONS(2676),
    [anon_sym_IBInspectable] = ACTIONS(2676),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2676),
    [anon_sym_NS_INLINE] = ACTIONS(2676),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2676),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2676),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2676),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2676),
    [anon_sym_const] = ACTIONS(2676),
    [anon_sym_constexpr] = ACTIONS(2676),
    [anon_sym_volatile] = ACTIONS(2676),
    [anon_sym_restrict] = ACTIONS(2676),
    [anon_sym___restrict__] = ACTIONS(2676),
    [anon_sym__Atomic] = ACTIONS(2676),
    [anon_sym__Noreturn] = ACTIONS(2676),
    [anon_sym__Nonnull] = ACTIONS(2676),
    [anon_sym_nullable] = ACTIONS(2676),
    [anon_sym__Complex] = ACTIONS(2676),
    [anon_sym__Nullable] = ACTIONS(2676),
    [anon_sym__Nullable_result] = ACTIONS(2676),
    [anon_sym__Null_unspecified] = ACTIONS(2676),
    [anon_sym___autoreleasing] = ACTIONS(2676),
    [anon_sym___block] = ACTIONS(2676),
    [anon_sym___bridge] = ACTIONS(2676),
    [anon_sym___bridge_retained] = ACTIONS(2676),
    [anon_sym___bridge_transfer] = ACTIONS(2676),
    [anon_sym___complex] = ACTIONS(2676),
    [anon_sym___const] = ACTIONS(2676),
    [anon_sym___imag] = ACTIONS(2676),
    [anon_sym___kindof] = ACTIONS(2676),
    [anon_sym___nonnull] = ACTIONS(2676),
    [anon_sym___nullable] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2676),
    [anon_sym___real] = ACTIONS(2676),
    [anon_sym___strong] = ACTIONS(2676),
    [anon_sym___unsafe_unretained] = ACTIONS(2676),
    [anon_sym___unused] = ACTIONS(2676),
    [anon_sym___weak] = ACTIONS(2676),
    [anon_sym_alignas] = ACTIONS(2676),
    [anon_sym__Alignas] = ACTIONS(2676),
    [sym_primitive_type] = ACTIONS(2676),
    [anon_sym_enum] = ACTIONS(2676),
    [anon_sym_struct] = ACTIONS(2676),
    [anon_sym_union] = ACTIONS(2676),
    [anon_sym_if] = ACTIONS(2676),
    [anon_sym_switch] = ACTIONS(2676),
    [anon_sym_case] = ACTIONS(2676),
    [anon_sym_default] = ACTIONS(2676),
    [anon_sym_while] = ACTIONS(2676),
    [anon_sym_do] = ACTIONS(2676),
    [anon_sym_for] = ACTIONS(2676),
    [anon_sym_in] = ACTIONS(2676),
    [anon_sym_return] = ACTIONS(2676),
    [anon_sym_break] = ACTIONS(2676),
    [anon_sym_continue] = ACTIONS(2676),
    [anon_sym_goto] = ACTIONS(2676),
    [anon_sym_DASH_DASH] = ACTIONS(2678),
    [anon_sym_PLUS_PLUS] = ACTIONS(2678),
    [anon_sym_sizeof] = ACTIONS(2676),
    [anon_sym___alignof__] = ACTIONS(2676),
    [anon_sym___alignof] = ACTIONS(2676),
    [anon_sym__alignof] = ACTIONS(2676),
    [anon_sym_alignof] = ACTIONS(2676),
    [anon_sym__Alignof] = ACTIONS(2676),
    [anon_sym_offsetof] = ACTIONS(2676),
    [anon_sym__Generic] = ACTIONS(2676),
    [anon_sym_asm] = ACTIONS(2676),
    [anon_sym___asm__] = ACTIONS(2676),
    [anon_sym___asm] = ACTIONS(2676),
    [sym_number_literal] = ACTIONS(2678),
    [anon_sym_L_SQUOTE] = ACTIONS(2678),
    [anon_sym_u_SQUOTE] = ACTIONS(2678),
    [anon_sym_U_SQUOTE] = ACTIONS(2678),
    [anon_sym_u8_SQUOTE] = ACTIONS(2678),
    [anon_sym_SQUOTE] = ACTIONS(2678),
    [anon_sym_AT] = ACTIONS(2676),
    [anon_sym_DQUOTE] = ACTIONS(2678),
    [anon_sym_L_DQUOTE] = ACTIONS(2678),
    [anon_sym_u_DQUOTE] = ACTIONS(2678),
    [anon_sym_U_DQUOTE] = ACTIONS(2678),
    [anon_sym_u8_DQUOTE] = ACTIONS(2678),
    [sym_true] = ACTIONS(2676),
    [sym_false] = ACTIONS(2676),
    [anon_sym_NULL] = ACTIONS(2676),
    [anon_sym_nullptr] = ACTIONS(2676),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2676),
    [anon_sym___typeof] = ACTIONS(2676),
    [anon_sym_typeof] = ACTIONS(2676),
    [anon_sym_ATimport] = ACTIONS(2678),
    [aux_sym_preproc_undef_token1] = ACTIONS(2676),
    [anon_sym_POUND] = ACTIONS(2676),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2676),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2676),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2676),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2676),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2676),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2676),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2676),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_API_AVAILABLE] = ACTIONS(2676),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_API_DEPRECATED] = ACTIONS(2676),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2676),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym___deprecated_msg] = ACTIONS(2676),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2676),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2676),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2676),
    [anon_sym_ATprotocol] = ACTIONS(2678),
    [anon_sym_ATinterface] = ACTIONS(2678),
    [anon_sym_ATimplementation] = ACTIONS(2678),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2678),
    [anon_sym_ATsynthesize] = ACTIONS(2678),
    [anon_sym_ATdynamic] = ACTIONS(2678),
    [anon_sym_ATtry] = ACTIONS(2678),
    [anon_sym___try] = ACTIONS(2676),
    [anon_sym_ATthrow] = ACTIONS(2678),
    [anon_sym_ATselector] = ACTIONS(2678),
    [anon_sym_ATavailable] = ACTIONS(2678),
    [anon_sym___builtin_available] = ACTIONS(2676),
    [anon_sym_va_arg] = ACTIONS(2676),
    [anon_sym_ATencode] = ACTIONS(2678),
    [anon_sym_ATsynchronized] = ACTIONS(2678),
    [anon_sym_BOOL] = ACTIONS(2676),
    [anon_sym_IMP] = ACTIONS(2676),
    [anon_sym_SEL] = ACTIONS(2676),
    [anon_sym_Class] = ACTIONS(2676),
    [anon_sym_id] = ACTIONS(2676),
  },
  [STATE(508)] = {
    [sym_identifier] = ACTIONS(2680),
    [aux_sym_preproc_include_token1] = ACTIONS(2680),
    [aux_sym_preproc_include_token2] = ACTIONS(2680),
    [aux_sym_preproc_def_token1] = ACTIONS(2680),
    [aux_sym_preproc_if_token1] = ACTIONS(2680),
    [aux_sym_preproc_if_token2] = ACTIONS(2680),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2680),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2680),
    [aux_sym_preproc_else_token1] = ACTIONS(2680),
    [aux_sym_preproc_elif_token1] = ACTIONS(2680),
    [sym_preproc_directive] = ACTIONS(2680),
    [anon_sym_LPAREN2] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2680),
    [anon_sym_PLUS] = ACTIONS(2680),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym___extension__] = ACTIONS(2680),
    [anon_sym_typedef] = ACTIONS(2680),
    [anon_sym_extern] = ACTIONS(2680),
    [anon_sym___attribute__] = ACTIONS(2680),
    [anon_sym___attribute] = ACTIONS(2680),
    [anon_sym_noreturn] = ACTIONS(2680),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym___declspec] = ACTIONS(2680),
    [anon_sym___cdecl] = ACTIONS(2680),
    [anon_sym___clrcall] = ACTIONS(2680),
    [anon_sym___stdcall] = ACTIONS(2680),
    [anon_sym___fastcall] = ACTIONS(2680),
    [anon_sym___thiscall] = ACTIONS(2680),
    [anon_sym___vectorcall] = ACTIONS(2680),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_signed] = ACTIONS(2680),
    [anon_sym_unsigned] = ACTIONS(2680),
    [anon_sym_long] = ACTIONS(2680),
    [anon_sym_short] = ACTIONS(2680),
    [anon_sym_static] = ACTIONS(2680),
    [anon_sym_ATautoreleasepool] = ACTIONS(2682),
    [anon_sym_auto] = ACTIONS(2680),
    [anon_sym_register] = ACTIONS(2680),
    [anon_sym_inline] = ACTIONS(2680),
    [anon_sym___inline] = ACTIONS(2680),
    [anon_sym___inline__] = ACTIONS(2680),
    [anon_sym___forceinline] = ACTIONS(2680),
    [anon_sym_thread_local] = ACTIONS(2680),
    [anon_sym___thread] = ACTIONS(2680),
    [anon_sym_CG_EXTERN] = ACTIONS(2680),
    [anon_sym_CG_INLINE] = ACTIONS(2680),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2680),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2680),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2680),
    [anon_sym_IBOutlet] = ACTIONS(2680),
    [anon_sym_IBInspectable] = ACTIONS(2680),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2680),
    [anon_sym_NS_INLINE] = ACTIONS(2680),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2680),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2680),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2680),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2680),
    [anon_sym_const] = ACTIONS(2680),
    [anon_sym_constexpr] = ACTIONS(2680),
    [anon_sym_volatile] = ACTIONS(2680),
    [anon_sym_restrict] = ACTIONS(2680),
    [anon_sym___restrict__] = ACTIONS(2680),
    [anon_sym__Atomic] = ACTIONS(2680),
    [anon_sym__Noreturn] = ACTIONS(2680),
    [anon_sym__Nonnull] = ACTIONS(2680),
    [anon_sym_nullable] = ACTIONS(2680),
    [anon_sym__Complex] = ACTIONS(2680),
    [anon_sym__Nullable] = ACTIONS(2680),
    [anon_sym__Nullable_result] = ACTIONS(2680),
    [anon_sym__Null_unspecified] = ACTIONS(2680),
    [anon_sym___autoreleasing] = ACTIONS(2680),
    [anon_sym___block] = ACTIONS(2680),
    [anon_sym___bridge] = ACTIONS(2680),
    [anon_sym___bridge_retained] = ACTIONS(2680),
    [anon_sym___bridge_transfer] = ACTIONS(2680),
    [anon_sym___complex] = ACTIONS(2680),
    [anon_sym___const] = ACTIONS(2680),
    [anon_sym___imag] = ACTIONS(2680),
    [anon_sym___kindof] = ACTIONS(2680),
    [anon_sym___nonnull] = ACTIONS(2680),
    [anon_sym___nullable] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2680),
    [anon_sym___real] = ACTIONS(2680),
    [anon_sym___strong] = ACTIONS(2680),
    [anon_sym___unsafe_unretained] = ACTIONS(2680),
    [anon_sym___unused] = ACTIONS(2680),
    [anon_sym___weak] = ACTIONS(2680),
    [anon_sym_alignas] = ACTIONS(2680),
    [anon_sym__Alignas] = ACTIONS(2680),
    [sym_primitive_type] = ACTIONS(2680),
    [anon_sym_enum] = ACTIONS(2680),
    [anon_sym_struct] = ACTIONS(2680),
    [anon_sym_union] = ACTIONS(2680),
    [anon_sym_if] = ACTIONS(2680),
    [anon_sym_switch] = ACTIONS(2680),
    [anon_sym_case] = ACTIONS(2680),
    [anon_sym_default] = ACTIONS(2680),
    [anon_sym_while] = ACTIONS(2680),
    [anon_sym_do] = ACTIONS(2680),
    [anon_sym_for] = ACTIONS(2680),
    [anon_sym_in] = ACTIONS(2680),
    [anon_sym_return] = ACTIONS(2680),
    [anon_sym_break] = ACTIONS(2680),
    [anon_sym_continue] = ACTIONS(2680),
    [anon_sym_goto] = ACTIONS(2680),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_sizeof] = ACTIONS(2680),
    [anon_sym___alignof__] = ACTIONS(2680),
    [anon_sym___alignof] = ACTIONS(2680),
    [anon_sym__alignof] = ACTIONS(2680),
    [anon_sym_alignof] = ACTIONS(2680),
    [anon_sym__Alignof] = ACTIONS(2680),
    [anon_sym_offsetof] = ACTIONS(2680),
    [anon_sym__Generic] = ACTIONS(2680),
    [anon_sym_asm] = ACTIONS(2680),
    [anon_sym___asm__] = ACTIONS(2680),
    [anon_sym___asm] = ACTIONS(2680),
    [sym_number_literal] = ACTIONS(2682),
    [anon_sym_L_SQUOTE] = ACTIONS(2682),
    [anon_sym_u_SQUOTE] = ACTIONS(2682),
    [anon_sym_U_SQUOTE] = ACTIONS(2682),
    [anon_sym_u8_SQUOTE] = ACTIONS(2682),
    [anon_sym_SQUOTE] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2680),
    [anon_sym_DQUOTE] = ACTIONS(2682),
    [anon_sym_L_DQUOTE] = ACTIONS(2682),
    [anon_sym_u_DQUOTE] = ACTIONS(2682),
    [anon_sym_U_DQUOTE] = ACTIONS(2682),
    [anon_sym_u8_DQUOTE] = ACTIONS(2682),
    [sym_true] = ACTIONS(2680),
    [sym_false] = ACTIONS(2680),
    [anon_sym_NULL] = ACTIONS(2680),
    [anon_sym_nullptr] = ACTIONS(2680),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2680),
    [anon_sym___typeof] = ACTIONS(2680),
    [anon_sym_typeof] = ACTIONS(2680),
    [anon_sym_ATimport] = ACTIONS(2682),
    [aux_sym_preproc_undef_token1] = ACTIONS(2680),
    [anon_sym_POUND] = ACTIONS(2680),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2680),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2680),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2680),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2680),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2680),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2680),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2680),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_API_AVAILABLE] = ACTIONS(2680),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_API_DEPRECATED] = ACTIONS(2680),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2680),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym___deprecated_msg] = ACTIONS(2680),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2680),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2680),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2680),
    [anon_sym_ATprotocol] = ACTIONS(2682),
    [anon_sym_ATinterface] = ACTIONS(2682),
    [anon_sym_ATimplementation] = ACTIONS(2682),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2682),
    [anon_sym_ATsynthesize] = ACTIONS(2682),
    [anon_sym_ATdynamic] = ACTIONS(2682),
    [anon_sym_ATtry] = ACTIONS(2682),
    [anon_sym___try] = ACTIONS(2680),
    [anon_sym_ATthrow] = ACTIONS(2682),
    [anon_sym_ATselector] = ACTIONS(2682),
    [anon_sym_ATavailable] = ACTIONS(2682),
    [anon_sym___builtin_available] = ACTIONS(2680),
    [anon_sym_va_arg] = ACTIONS(2680),
    [anon_sym_ATencode] = ACTIONS(2682),
    [anon_sym_ATsynchronized] = ACTIONS(2682),
    [anon_sym_BOOL] = ACTIONS(2680),
    [anon_sym_IMP] = ACTIONS(2680),
    [anon_sym_SEL] = ACTIONS(2680),
    [anon_sym_Class] = ACTIONS(2680),
    [anon_sym_id] = ACTIONS(2680),
  },
  [STATE(509)] = {
    [sym_identifier] = ACTIONS(2708),
    [aux_sym_preproc_include_token1] = ACTIONS(2708),
    [aux_sym_preproc_include_token2] = ACTIONS(2708),
    [aux_sym_preproc_def_token1] = ACTIONS(2708),
    [aux_sym_preproc_if_token1] = ACTIONS(2708),
    [aux_sym_preproc_if_token2] = ACTIONS(2708),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2708),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2708),
    [aux_sym_preproc_else_token1] = ACTIONS(2708),
    [aux_sym_preproc_elif_token1] = ACTIONS(2708),
    [sym_preproc_directive] = ACTIONS(2708),
    [anon_sym_LPAREN2] = ACTIONS(2710),
    [anon_sym_BANG] = ACTIONS(2710),
    [anon_sym_TILDE] = ACTIONS(2710),
    [anon_sym_DASH] = ACTIONS(2708),
    [anon_sym_PLUS] = ACTIONS(2708),
    [anon_sym_STAR] = ACTIONS(2710),
    [anon_sym_CARET] = ACTIONS(2710),
    [anon_sym_AMP] = ACTIONS(2710),
    [anon_sym_SEMI] = ACTIONS(2710),
    [anon_sym___extension__] = ACTIONS(2708),
    [anon_sym_typedef] = ACTIONS(2708),
    [anon_sym_extern] = ACTIONS(2708),
    [anon_sym___attribute__] = ACTIONS(2708),
    [anon_sym___attribute] = ACTIONS(2708),
    [anon_sym_noreturn] = ACTIONS(2708),
    [anon_sym_LBRACK] = ACTIONS(2710),
    [anon_sym___declspec] = ACTIONS(2708),
    [anon_sym___cdecl] = ACTIONS(2708),
    [anon_sym___clrcall] = ACTIONS(2708),
    [anon_sym___stdcall] = ACTIONS(2708),
    [anon_sym___fastcall] = ACTIONS(2708),
    [anon_sym___thiscall] = ACTIONS(2708),
    [anon_sym___vectorcall] = ACTIONS(2708),
    [anon_sym_LBRACE] = ACTIONS(2710),
    [anon_sym_signed] = ACTIONS(2708),
    [anon_sym_unsigned] = ACTIONS(2708),
    [anon_sym_long] = ACTIONS(2708),
    [anon_sym_short] = ACTIONS(2708),
    [anon_sym_static] = ACTIONS(2708),
    [anon_sym_ATautoreleasepool] = ACTIONS(2710),
    [anon_sym_auto] = ACTIONS(2708),
    [anon_sym_register] = ACTIONS(2708),
    [anon_sym_inline] = ACTIONS(2708),
    [anon_sym___inline] = ACTIONS(2708),
    [anon_sym___inline__] = ACTIONS(2708),
    [anon_sym___forceinline] = ACTIONS(2708),
    [anon_sym_thread_local] = ACTIONS(2708),
    [anon_sym___thread] = ACTIONS(2708),
    [anon_sym_CG_EXTERN] = ACTIONS(2708),
    [anon_sym_CG_INLINE] = ACTIONS(2708),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2708),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2708),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2708),
    [anon_sym_IBOutlet] = ACTIONS(2708),
    [anon_sym_IBInspectable] = ACTIONS(2708),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2708),
    [anon_sym_NS_INLINE] = ACTIONS(2708),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2708),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2708),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2708),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2708),
    [anon_sym_const] = ACTIONS(2708),
    [anon_sym_constexpr] = ACTIONS(2708),
    [anon_sym_volatile] = ACTIONS(2708),
    [anon_sym_restrict] = ACTIONS(2708),
    [anon_sym___restrict__] = ACTIONS(2708),
    [anon_sym__Atomic] = ACTIONS(2708),
    [anon_sym__Noreturn] = ACTIONS(2708),
    [anon_sym__Nonnull] = ACTIONS(2708),
    [anon_sym_nullable] = ACTIONS(2708),
    [anon_sym__Complex] = ACTIONS(2708),
    [anon_sym__Nullable] = ACTIONS(2708),
    [anon_sym__Nullable_result] = ACTIONS(2708),
    [anon_sym__Null_unspecified] = ACTIONS(2708),
    [anon_sym___autoreleasing] = ACTIONS(2708),
    [anon_sym___block] = ACTIONS(2708),
    [anon_sym___bridge] = ACTIONS(2708),
    [anon_sym___bridge_retained] = ACTIONS(2708),
    [anon_sym___bridge_transfer] = ACTIONS(2708),
    [anon_sym___complex] = ACTIONS(2708),
    [anon_sym___const] = ACTIONS(2708),
    [anon_sym___imag] = ACTIONS(2708),
    [anon_sym___kindof] = ACTIONS(2708),
    [anon_sym___nonnull] = ACTIONS(2708),
    [anon_sym___nullable] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2708),
    [anon_sym___real] = ACTIONS(2708),
    [anon_sym___strong] = ACTIONS(2708),
    [anon_sym___unsafe_unretained] = ACTIONS(2708),
    [anon_sym___unused] = ACTIONS(2708),
    [anon_sym___weak] = ACTIONS(2708),
    [anon_sym_alignas] = ACTIONS(2708),
    [anon_sym__Alignas] = ACTIONS(2708),
    [sym_primitive_type] = ACTIONS(2708),
    [anon_sym_enum] = ACTIONS(2708),
    [anon_sym_struct] = ACTIONS(2708),
    [anon_sym_union] = ACTIONS(2708),
    [anon_sym_if] = ACTIONS(2708),
    [anon_sym_switch] = ACTIONS(2708),
    [anon_sym_case] = ACTIONS(2708),
    [anon_sym_default] = ACTIONS(2708),
    [anon_sym_while] = ACTIONS(2708),
    [anon_sym_do] = ACTIONS(2708),
    [anon_sym_for] = ACTIONS(2708),
    [anon_sym_in] = ACTIONS(2708),
    [anon_sym_return] = ACTIONS(2708),
    [anon_sym_break] = ACTIONS(2708),
    [anon_sym_continue] = ACTIONS(2708),
    [anon_sym_goto] = ACTIONS(2708),
    [anon_sym_DASH_DASH] = ACTIONS(2710),
    [anon_sym_PLUS_PLUS] = ACTIONS(2710),
    [anon_sym_sizeof] = ACTIONS(2708),
    [anon_sym___alignof__] = ACTIONS(2708),
    [anon_sym___alignof] = ACTIONS(2708),
    [anon_sym__alignof] = ACTIONS(2708),
    [anon_sym_alignof] = ACTIONS(2708),
    [anon_sym__Alignof] = ACTIONS(2708),
    [anon_sym_offsetof] = ACTIONS(2708),
    [anon_sym__Generic] = ACTIONS(2708),
    [anon_sym_asm] = ACTIONS(2708),
    [anon_sym___asm__] = ACTIONS(2708),
    [anon_sym___asm] = ACTIONS(2708),
    [sym_number_literal] = ACTIONS(2710),
    [anon_sym_L_SQUOTE] = ACTIONS(2710),
    [anon_sym_u_SQUOTE] = ACTIONS(2710),
    [anon_sym_U_SQUOTE] = ACTIONS(2710),
    [anon_sym_u8_SQUOTE] = ACTIONS(2710),
    [anon_sym_SQUOTE] = ACTIONS(2710),
    [anon_sym_AT] = ACTIONS(2708),
    [anon_sym_DQUOTE] = ACTIONS(2710),
    [anon_sym_L_DQUOTE] = ACTIONS(2710),
    [anon_sym_u_DQUOTE] = ACTIONS(2710),
    [anon_sym_U_DQUOTE] = ACTIONS(2710),
    [anon_sym_u8_DQUOTE] = ACTIONS(2710),
    [sym_true] = ACTIONS(2708),
    [sym_false] = ACTIONS(2708),
    [anon_sym_NULL] = ACTIONS(2708),
    [anon_sym_nullptr] = ACTIONS(2708),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2708),
    [anon_sym___typeof] = ACTIONS(2708),
    [anon_sym_typeof] = ACTIONS(2708),
    [anon_sym_ATimport] = ACTIONS(2710),
    [aux_sym_preproc_undef_token1] = ACTIONS(2708),
    [anon_sym_POUND] = ACTIONS(2708),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2708),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2708),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2708),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2708),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2708),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2708),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2708),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_API_AVAILABLE] = ACTIONS(2708),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_API_DEPRECATED] = ACTIONS(2708),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2708),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym___deprecated_msg] = ACTIONS(2708),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2708),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2708),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2708),
    [anon_sym_ATprotocol] = ACTIONS(2710),
    [anon_sym_ATinterface] = ACTIONS(2710),
    [anon_sym_ATimplementation] = ACTIONS(2710),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2710),
    [anon_sym_ATsynthesize] = ACTIONS(2710),
    [anon_sym_ATdynamic] = ACTIONS(2710),
    [anon_sym_ATtry] = ACTIONS(2710),
    [anon_sym___try] = ACTIONS(2708),
    [anon_sym_ATthrow] = ACTIONS(2710),
    [anon_sym_ATselector] = ACTIONS(2710),
    [anon_sym_ATavailable] = ACTIONS(2710),
    [anon_sym___builtin_available] = ACTIONS(2708),
    [anon_sym_va_arg] = ACTIONS(2708),
    [anon_sym_ATencode] = ACTIONS(2710),
    [anon_sym_ATsynchronized] = ACTIONS(2710),
    [anon_sym_BOOL] = ACTIONS(2708),
    [anon_sym_IMP] = ACTIONS(2708),
    [anon_sym_SEL] = ACTIONS(2708),
    [anon_sym_Class] = ACTIONS(2708),
    [anon_sym_id] = ACTIONS(2708),
  },
  [STATE(510)] = {
    [sym_identifier] = ACTIONS(2622),
    [aux_sym_preproc_include_token1] = ACTIONS(2622),
    [aux_sym_preproc_include_token2] = ACTIONS(2622),
    [aux_sym_preproc_def_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token2] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2622),
    [aux_sym_preproc_else_token1] = ACTIONS(2622),
    [aux_sym_preproc_elif_token1] = ACTIONS(2622),
    [sym_preproc_directive] = ACTIONS(2622),
    [anon_sym_LPAREN2] = ACTIONS(2624),
    [anon_sym_BANG] = ACTIONS(2624),
    [anon_sym_TILDE] = ACTIONS(2624),
    [anon_sym_DASH] = ACTIONS(2622),
    [anon_sym_PLUS] = ACTIONS(2622),
    [anon_sym_STAR] = ACTIONS(2624),
    [anon_sym_CARET] = ACTIONS(2624),
    [anon_sym_AMP] = ACTIONS(2624),
    [anon_sym_SEMI] = ACTIONS(2624),
    [anon_sym___extension__] = ACTIONS(2622),
    [anon_sym_typedef] = ACTIONS(2622),
    [anon_sym_extern] = ACTIONS(2622),
    [anon_sym___attribute__] = ACTIONS(2622),
    [anon_sym___attribute] = ACTIONS(2622),
    [anon_sym_noreturn] = ACTIONS(2622),
    [anon_sym_LBRACK] = ACTIONS(2624),
    [anon_sym___declspec] = ACTIONS(2622),
    [anon_sym___cdecl] = ACTIONS(2622),
    [anon_sym___clrcall] = ACTIONS(2622),
    [anon_sym___stdcall] = ACTIONS(2622),
    [anon_sym___fastcall] = ACTIONS(2622),
    [anon_sym___thiscall] = ACTIONS(2622),
    [anon_sym___vectorcall] = ACTIONS(2622),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_signed] = ACTIONS(2622),
    [anon_sym_unsigned] = ACTIONS(2622),
    [anon_sym_long] = ACTIONS(2622),
    [anon_sym_short] = ACTIONS(2622),
    [anon_sym_static] = ACTIONS(2622),
    [anon_sym_ATautoreleasepool] = ACTIONS(2624),
    [anon_sym_auto] = ACTIONS(2622),
    [anon_sym_register] = ACTIONS(2622),
    [anon_sym_inline] = ACTIONS(2622),
    [anon_sym___inline] = ACTIONS(2622),
    [anon_sym___inline__] = ACTIONS(2622),
    [anon_sym___forceinline] = ACTIONS(2622),
    [anon_sym_thread_local] = ACTIONS(2622),
    [anon_sym___thread] = ACTIONS(2622),
    [anon_sym_CG_EXTERN] = ACTIONS(2622),
    [anon_sym_CG_INLINE] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2622),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2622),
    [anon_sym_IBOutlet] = ACTIONS(2622),
    [anon_sym_IBInspectable] = ACTIONS(2622),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2622),
    [anon_sym_NS_INLINE] = ACTIONS(2622),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2622),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2622),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2622),
    [anon_sym_const] = ACTIONS(2622),
    [anon_sym_constexpr] = ACTIONS(2622),
    [anon_sym_volatile] = ACTIONS(2622),
    [anon_sym_restrict] = ACTIONS(2622),
    [anon_sym___restrict__] = ACTIONS(2622),
    [anon_sym__Atomic] = ACTIONS(2622),
    [anon_sym__Noreturn] = ACTIONS(2622),
    [anon_sym__Nonnull] = ACTIONS(2622),
    [anon_sym_nullable] = ACTIONS(2622),
    [anon_sym__Complex] = ACTIONS(2622),
    [anon_sym__Nullable] = ACTIONS(2622),
    [anon_sym__Nullable_result] = ACTIONS(2622),
    [anon_sym__Null_unspecified] = ACTIONS(2622),
    [anon_sym___autoreleasing] = ACTIONS(2622),
    [anon_sym___block] = ACTIONS(2622),
    [anon_sym___bridge] = ACTIONS(2622),
    [anon_sym___bridge_retained] = ACTIONS(2622),
    [anon_sym___bridge_transfer] = ACTIONS(2622),
    [anon_sym___complex] = ACTIONS(2622),
    [anon_sym___const] = ACTIONS(2622),
    [anon_sym___imag] = ACTIONS(2622),
    [anon_sym___kindof] = ACTIONS(2622),
    [anon_sym___nonnull] = ACTIONS(2622),
    [anon_sym___nullable] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2622),
    [anon_sym___real] = ACTIONS(2622),
    [anon_sym___strong] = ACTIONS(2622),
    [anon_sym___unsafe_unretained] = ACTIONS(2622),
    [anon_sym___unused] = ACTIONS(2622),
    [anon_sym___weak] = ACTIONS(2622),
    [anon_sym_alignas] = ACTIONS(2622),
    [anon_sym__Alignas] = ACTIONS(2622),
    [sym_primitive_type] = ACTIONS(2622),
    [anon_sym_enum] = ACTIONS(2622),
    [anon_sym_struct] = ACTIONS(2622),
    [anon_sym_union] = ACTIONS(2622),
    [anon_sym_if] = ACTIONS(2622),
    [anon_sym_switch] = ACTIONS(2622),
    [anon_sym_case] = ACTIONS(2622),
    [anon_sym_default] = ACTIONS(2622),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_do] = ACTIONS(2622),
    [anon_sym_for] = ACTIONS(2622),
    [anon_sym_in] = ACTIONS(2622),
    [anon_sym_return] = ACTIONS(2622),
    [anon_sym_break] = ACTIONS(2622),
    [anon_sym_continue] = ACTIONS(2622),
    [anon_sym_goto] = ACTIONS(2622),
    [anon_sym_DASH_DASH] = ACTIONS(2624),
    [anon_sym_PLUS_PLUS] = ACTIONS(2624),
    [anon_sym_sizeof] = ACTIONS(2622),
    [anon_sym___alignof__] = ACTIONS(2622),
    [anon_sym___alignof] = ACTIONS(2622),
    [anon_sym__alignof] = ACTIONS(2622),
    [anon_sym_alignof] = ACTIONS(2622),
    [anon_sym__Alignof] = ACTIONS(2622),
    [anon_sym_offsetof] = ACTIONS(2622),
    [anon_sym__Generic] = ACTIONS(2622),
    [anon_sym_asm] = ACTIONS(2622),
    [anon_sym___asm__] = ACTIONS(2622),
    [anon_sym___asm] = ACTIONS(2622),
    [sym_number_literal] = ACTIONS(2624),
    [anon_sym_L_SQUOTE] = ACTIONS(2624),
    [anon_sym_u_SQUOTE] = ACTIONS(2624),
    [anon_sym_U_SQUOTE] = ACTIONS(2624),
    [anon_sym_u8_SQUOTE] = ACTIONS(2624),
    [anon_sym_SQUOTE] = ACTIONS(2624),
    [anon_sym_AT] = ACTIONS(2622),
    [anon_sym_DQUOTE] = ACTIONS(2624),
    [anon_sym_L_DQUOTE] = ACTIONS(2624),
    [anon_sym_u_DQUOTE] = ACTIONS(2624),
    [anon_sym_U_DQUOTE] = ACTIONS(2624),
    [anon_sym_u8_DQUOTE] = ACTIONS(2624),
    [sym_true] = ACTIONS(2622),
    [sym_false] = ACTIONS(2622),
    [anon_sym_NULL] = ACTIONS(2622),
    [anon_sym_nullptr] = ACTIONS(2622),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2622),
    [anon_sym___typeof] = ACTIONS(2622),
    [anon_sym_typeof] = ACTIONS(2622),
    [anon_sym_ATimport] = ACTIONS(2624),
    [aux_sym_preproc_undef_token1] = ACTIONS(2622),
    [anon_sym_POUND] = ACTIONS(2622),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2622),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2622),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2622),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_API_AVAILABLE] = ACTIONS(2622),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_API_DEPRECATED] = ACTIONS(2622),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym___deprecated_msg] = ACTIONS(2622),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2622),
    [anon_sym_ATprotocol] = ACTIONS(2624),
    [anon_sym_ATinterface] = ACTIONS(2624),
    [anon_sym_ATimplementation] = ACTIONS(2624),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2624),
    [anon_sym_ATsynthesize] = ACTIONS(2624),
    [anon_sym_ATdynamic] = ACTIONS(2624),
    [anon_sym_ATtry] = ACTIONS(2624),
    [anon_sym___try] = ACTIONS(2622),
    [anon_sym_ATthrow] = ACTIONS(2624),
    [anon_sym_ATselector] = ACTIONS(2624),
    [anon_sym_ATavailable] = ACTIONS(2624),
    [anon_sym___builtin_available] = ACTIONS(2622),
    [anon_sym_va_arg] = ACTIONS(2622),
    [anon_sym_ATencode] = ACTIONS(2624),
    [anon_sym_ATsynchronized] = ACTIONS(2624),
    [anon_sym_BOOL] = ACTIONS(2622),
    [anon_sym_IMP] = ACTIONS(2622),
    [anon_sym_SEL] = ACTIONS(2622),
    [anon_sym_Class] = ACTIONS(2622),
    [anon_sym_id] = ACTIONS(2622),
  },
  [STATE(511)] = {
    [sym_identifier] = ACTIONS(2712),
    [aux_sym_preproc_include_token1] = ACTIONS(2712),
    [aux_sym_preproc_include_token2] = ACTIONS(2712),
    [aux_sym_preproc_def_token1] = ACTIONS(2712),
    [aux_sym_preproc_if_token1] = ACTIONS(2712),
    [aux_sym_preproc_if_token2] = ACTIONS(2712),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2712),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2712),
    [aux_sym_preproc_else_token1] = ACTIONS(2712),
    [aux_sym_preproc_elif_token1] = ACTIONS(2712),
    [sym_preproc_directive] = ACTIONS(2712),
    [anon_sym_LPAREN2] = ACTIONS(2714),
    [anon_sym_BANG] = ACTIONS(2714),
    [anon_sym_TILDE] = ACTIONS(2714),
    [anon_sym_DASH] = ACTIONS(2712),
    [anon_sym_PLUS] = ACTIONS(2712),
    [anon_sym_STAR] = ACTIONS(2714),
    [anon_sym_CARET] = ACTIONS(2714),
    [anon_sym_AMP] = ACTIONS(2714),
    [anon_sym_SEMI] = ACTIONS(2714),
    [anon_sym___extension__] = ACTIONS(2712),
    [anon_sym_typedef] = ACTIONS(2712),
    [anon_sym_extern] = ACTIONS(2712),
    [anon_sym___attribute__] = ACTIONS(2712),
    [anon_sym___attribute] = ACTIONS(2712),
    [anon_sym_noreturn] = ACTIONS(2712),
    [anon_sym_LBRACK] = ACTIONS(2714),
    [anon_sym___declspec] = ACTIONS(2712),
    [anon_sym___cdecl] = ACTIONS(2712),
    [anon_sym___clrcall] = ACTIONS(2712),
    [anon_sym___stdcall] = ACTIONS(2712),
    [anon_sym___fastcall] = ACTIONS(2712),
    [anon_sym___thiscall] = ACTIONS(2712),
    [anon_sym___vectorcall] = ACTIONS(2712),
    [anon_sym_LBRACE] = ACTIONS(2714),
    [anon_sym_signed] = ACTIONS(2712),
    [anon_sym_unsigned] = ACTIONS(2712),
    [anon_sym_long] = ACTIONS(2712),
    [anon_sym_short] = ACTIONS(2712),
    [anon_sym_static] = ACTIONS(2712),
    [anon_sym_ATautoreleasepool] = ACTIONS(2714),
    [anon_sym_auto] = ACTIONS(2712),
    [anon_sym_register] = ACTIONS(2712),
    [anon_sym_inline] = ACTIONS(2712),
    [anon_sym___inline] = ACTIONS(2712),
    [anon_sym___inline__] = ACTIONS(2712),
    [anon_sym___forceinline] = ACTIONS(2712),
    [anon_sym_thread_local] = ACTIONS(2712),
    [anon_sym___thread] = ACTIONS(2712),
    [anon_sym_CG_EXTERN] = ACTIONS(2712),
    [anon_sym_CG_INLINE] = ACTIONS(2712),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2712),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2712),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2712),
    [anon_sym_IBOutlet] = ACTIONS(2712),
    [anon_sym_IBInspectable] = ACTIONS(2712),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2712),
    [anon_sym_NS_INLINE] = ACTIONS(2712),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2712),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2712),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2712),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2712),
    [anon_sym_const] = ACTIONS(2712),
    [anon_sym_constexpr] = ACTIONS(2712),
    [anon_sym_volatile] = ACTIONS(2712),
    [anon_sym_restrict] = ACTIONS(2712),
    [anon_sym___restrict__] = ACTIONS(2712),
    [anon_sym__Atomic] = ACTIONS(2712),
    [anon_sym__Noreturn] = ACTIONS(2712),
    [anon_sym__Nonnull] = ACTIONS(2712),
    [anon_sym_nullable] = ACTIONS(2712),
    [anon_sym__Complex] = ACTIONS(2712),
    [anon_sym__Nullable] = ACTIONS(2712),
    [anon_sym__Nullable_result] = ACTIONS(2712),
    [anon_sym__Null_unspecified] = ACTIONS(2712),
    [anon_sym___autoreleasing] = ACTIONS(2712),
    [anon_sym___block] = ACTIONS(2712),
    [anon_sym___bridge] = ACTIONS(2712),
    [anon_sym___bridge_retained] = ACTIONS(2712),
    [anon_sym___bridge_transfer] = ACTIONS(2712),
    [anon_sym___complex] = ACTIONS(2712),
    [anon_sym___const] = ACTIONS(2712),
    [anon_sym___imag] = ACTIONS(2712),
    [anon_sym___kindof] = ACTIONS(2712),
    [anon_sym___nonnull] = ACTIONS(2712),
    [anon_sym___nullable] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2712),
    [anon_sym___real] = ACTIONS(2712),
    [anon_sym___strong] = ACTIONS(2712),
    [anon_sym___unsafe_unretained] = ACTIONS(2712),
    [anon_sym___unused] = ACTIONS(2712),
    [anon_sym___weak] = ACTIONS(2712),
    [anon_sym_alignas] = ACTIONS(2712),
    [anon_sym__Alignas] = ACTIONS(2712),
    [sym_primitive_type] = ACTIONS(2712),
    [anon_sym_enum] = ACTIONS(2712),
    [anon_sym_struct] = ACTIONS(2712),
    [anon_sym_union] = ACTIONS(2712),
    [anon_sym_if] = ACTIONS(2712),
    [anon_sym_switch] = ACTIONS(2712),
    [anon_sym_case] = ACTIONS(2712),
    [anon_sym_default] = ACTIONS(2712),
    [anon_sym_while] = ACTIONS(2712),
    [anon_sym_do] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2712),
    [anon_sym_in] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_break] = ACTIONS(2712),
    [anon_sym_continue] = ACTIONS(2712),
    [anon_sym_goto] = ACTIONS(2712),
    [anon_sym_DASH_DASH] = ACTIONS(2714),
    [anon_sym_PLUS_PLUS] = ACTIONS(2714),
    [anon_sym_sizeof] = ACTIONS(2712),
    [anon_sym___alignof__] = ACTIONS(2712),
    [anon_sym___alignof] = ACTIONS(2712),
    [anon_sym__alignof] = ACTIONS(2712),
    [anon_sym_alignof] = ACTIONS(2712),
    [anon_sym__Alignof] = ACTIONS(2712),
    [anon_sym_offsetof] = ACTIONS(2712),
    [anon_sym__Generic] = ACTIONS(2712),
    [anon_sym_asm] = ACTIONS(2712),
    [anon_sym___asm__] = ACTIONS(2712),
    [anon_sym___asm] = ACTIONS(2712),
    [sym_number_literal] = ACTIONS(2714),
    [anon_sym_L_SQUOTE] = ACTIONS(2714),
    [anon_sym_u_SQUOTE] = ACTIONS(2714),
    [anon_sym_U_SQUOTE] = ACTIONS(2714),
    [anon_sym_u8_SQUOTE] = ACTIONS(2714),
    [anon_sym_SQUOTE] = ACTIONS(2714),
    [anon_sym_AT] = ACTIONS(2712),
    [anon_sym_DQUOTE] = ACTIONS(2714),
    [anon_sym_L_DQUOTE] = ACTIONS(2714),
    [anon_sym_u_DQUOTE] = ACTIONS(2714),
    [anon_sym_U_DQUOTE] = ACTIONS(2714),
    [anon_sym_u8_DQUOTE] = ACTIONS(2714),
    [sym_true] = ACTIONS(2712),
    [sym_false] = ACTIONS(2712),
    [anon_sym_NULL] = ACTIONS(2712),
    [anon_sym_nullptr] = ACTIONS(2712),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2712),
    [anon_sym___typeof] = ACTIONS(2712),
    [anon_sym_typeof] = ACTIONS(2712),
    [anon_sym_ATimport] = ACTIONS(2714),
    [aux_sym_preproc_undef_token1] = ACTIONS(2712),
    [anon_sym_POUND] = ACTIONS(2712),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2712),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2712),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2712),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2712),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2712),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2712),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2712),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_API_AVAILABLE] = ACTIONS(2712),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_API_DEPRECATED] = ACTIONS(2712),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2712),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym___deprecated_msg] = ACTIONS(2712),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2712),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2712),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2712),
    [anon_sym_ATprotocol] = ACTIONS(2714),
    [anon_sym_ATinterface] = ACTIONS(2714),
    [anon_sym_ATimplementation] = ACTIONS(2714),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2714),
    [anon_sym_ATsynthesize] = ACTIONS(2714),
    [anon_sym_ATdynamic] = ACTIONS(2714),
    [anon_sym_ATtry] = ACTIONS(2714),
    [anon_sym___try] = ACTIONS(2712),
    [anon_sym_ATthrow] = ACTIONS(2714),
    [anon_sym_ATselector] = ACTIONS(2714),
    [anon_sym_ATavailable] = ACTIONS(2714),
    [anon_sym___builtin_available] = ACTIONS(2712),
    [anon_sym_va_arg] = ACTIONS(2712),
    [anon_sym_ATencode] = ACTIONS(2714),
    [anon_sym_ATsynchronized] = ACTIONS(2714),
    [anon_sym_BOOL] = ACTIONS(2712),
    [anon_sym_IMP] = ACTIONS(2712),
    [anon_sym_SEL] = ACTIONS(2712),
    [anon_sym_Class] = ACTIONS(2712),
    [anon_sym_id] = ACTIONS(2712),
  },
  [STATE(512)] = {
    [sym_identifier] = ACTIONS(2684),
    [aux_sym_preproc_include_token1] = ACTIONS(2684),
    [aux_sym_preproc_include_token2] = ACTIONS(2684),
    [aux_sym_preproc_def_token1] = ACTIONS(2684),
    [aux_sym_preproc_if_token1] = ACTIONS(2684),
    [aux_sym_preproc_if_token2] = ACTIONS(2684),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2684),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2684),
    [aux_sym_preproc_else_token1] = ACTIONS(2684),
    [aux_sym_preproc_elif_token1] = ACTIONS(2684),
    [sym_preproc_directive] = ACTIONS(2684),
    [anon_sym_LPAREN2] = ACTIONS(2686),
    [anon_sym_BANG] = ACTIONS(2686),
    [anon_sym_TILDE] = ACTIONS(2686),
    [anon_sym_DASH] = ACTIONS(2684),
    [anon_sym_PLUS] = ACTIONS(2684),
    [anon_sym_STAR] = ACTIONS(2686),
    [anon_sym_CARET] = ACTIONS(2686),
    [anon_sym_AMP] = ACTIONS(2686),
    [anon_sym_SEMI] = ACTIONS(2686),
    [anon_sym___extension__] = ACTIONS(2684),
    [anon_sym_typedef] = ACTIONS(2684),
    [anon_sym_extern] = ACTIONS(2684),
    [anon_sym___attribute__] = ACTIONS(2684),
    [anon_sym___attribute] = ACTIONS(2684),
    [anon_sym_noreturn] = ACTIONS(2684),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym___declspec] = ACTIONS(2684),
    [anon_sym___cdecl] = ACTIONS(2684),
    [anon_sym___clrcall] = ACTIONS(2684),
    [anon_sym___stdcall] = ACTIONS(2684),
    [anon_sym___fastcall] = ACTIONS(2684),
    [anon_sym___thiscall] = ACTIONS(2684),
    [anon_sym___vectorcall] = ACTIONS(2684),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_signed] = ACTIONS(2684),
    [anon_sym_unsigned] = ACTIONS(2684),
    [anon_sym_long] = ACTIONS(2684),
    [anon_sym_short] = ACTIONS(2684),
    [anon_sym_static] = ACTIONS(2684),
    [anon_sym_ATautoreleasepool] = ACTIONS(2686),
    [anon_sym_auto] = ACTIONS(2684),
    [anon_sym_register] = ACTIONS(2684),
    [anon_sym_inline] = ACTIONS(2684),
    [anon_sym___inline] = ACTIONS(2684),
    [anon_sym___inline__] = ACTIONS(2684),
    [anon_sym___forceinline] = ACTIONS(2684),
    [anon_sym_thread_local] = ACTIONS(2684),
    [anon_sym___thread] = ACTIONS(2684),
    [anon_sym_CG_EXTERN] = ACTIONS(2684),
    [anon_sym_CG_INLINE] = ACTIONS(2684),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2684),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2684),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2684),
    [anon_sym_IBOutlet] = ACTIONS(2684),
    [anon_sym_IBInspectable] = ACTIONS(2684),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2684),
    [anon_sym_NS_INLINE] = ACTIONS(2684),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2684),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2684),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2684),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2684),
    [anon_sym_const] = ACTIONS(2684),
    [anon_sym_constexpr] = ACTIONS(2684),
    [anon_sym_volatile] = ACTIONS(2684),
    [anon_sym_restrict] = ACTIONS(2684),
    [anon_sym___restrict__] = ACTIONS(2684),
    [anon_sym__Atomic] = ACTIONS(2684),
    [anon_sym__Noreturn] = ACTIONS(2684),
    [anon_sym__Nonnull] = ACTIONS(2684),
    [anon_sym_nullable] = ACTIONS(2684),
    [anon_sym__Complex] = ACTIONS(2684),
    [anon_sym__Nullable] = ACTIONS(2684),
    [anon_sym__Nullable_result] = ACTIONS(2684),
    [anon_sym__Null_unspecified] = ACTIONS(2684),
    [anon_sym___autoreleasing] = ACTIONS(2684),
    [anon_sym___block] = ACTIONS(2684),
    [anon_sym___bridge] = ACTIONS(2684),
    [anon_sym___bridge_retained] = ACTIONS(2684),
    [anon_sym___bridge_transfer] = ACTIONS(2684),
    [anon_sym___complex] = ACTIONS(2684),
    [anon_sym___const] = ACTIONS(2684),
    [anon_sym___imag] = ACTIONS(2684),
    [anon_sym___kindof] = ACTIONS(2684),
    [anon_sym___nonnull] = ACTIONS(2684),
    [anon_sym___nullable] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2684),
    [anon_sym___real] = ACTIONS(2684),
    [anon_sym___strong] = ACTIONS(2684),
    [anon_sym___unsafe_unretained] = ACTIONS(2684),
    [anon_sym___unused] = ACTIONS(2684),
    [anon_sym___weak] = ACTIONS(2684),
    [anon_sym_alignas] = ACTIONS(2684),
    [anon_sym__Alignas] = ACTIONS(2684),
    [sym_primitive_type] = ACTIONS(2684),
    [anon_sym_enum] = ACTIONS(2684),
    [anon_sym_struct] = ACTIONS(2684),
    [anon_sym_union] = ACTIONS(2684),
    [anon_sym_if] = ACTIONS(2684),
    [anon_sym_switch] = ACTIONS(2684),
    [anon_sym_case] = ACTIONS(2684),
    [anon_sym_default] = ACTIONS(2684),
    [anon_sym_while] = ACTIONS(2684),
    [anon_sym_do] = ACTIONS(2684),
    [anon_sym_for] = ACTIONS(2684),
    [anon_sym_in] = ACTIONS(2684),
    [anon_sym_return] = ACTIONS(2684),
    [anon_sym_break] = ACTIONS(2684),
    [anon_sym_continue] = ACTIONS(2684),
    [anon_sym_goto] = ACTIONS(2684),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_sizeof] = ACTIONS(2684),
    [anon_sym___alignof__] = ACTIONS(2684),
    [anon_sym___alignof] = ACTIONS(2684),
    [anon_sym__alignof] = ACTIONS(2684),
    [anon_sym_alignof] = ACTIONS(2684),
    [anon_sym__Alignof] = ACTIONS(2684),
    [anon_sym_offsetof] = ACTIONS(2684),
    [anon_sym__Generic] = ACTIONS(2684),
    [anon_sym_asm] = ACTIONS(2684),
    [anon_sym___asm__] = ACTIONS(2684),
    [anon_sym___asm] = ACTIONS(2684),
    [sym_number_literal] = ACTIONS(2686),
    [anon_sym_L_SQUOTE] = ACTIONS(2686),
    [anon_sym_u_SQUOTE] = ACTIONS(2686),
    [anon_sym_U_SQUOTE] = ACTIONS(2686),
    [anon_sym_u8_SQUOTE] = ACTIONS(2686),
    [anon_sym_SQUOTE] = ACTIONS(2686),
    [anon_sym_AT] = ACTIONS(2684),
    [anon_sym_DQUOTE] = ACTIONS(2686),
    [anon_sym_L_DQUOTE] = ACTIONS(2686),
    [anon_sym_u_DQUOTE] = ACTIONS(2686),
    [anon_sym_U_DQUOTE] = ACTIONS(2686),
    [anon_sym_u8_DQUOTE] = ACTIONS(2686),
    [sym_true] = ACTIONS(2684),
    [sym_false] = ACTIONS(2684),
    [anon_sym_NULL] = ACTIONS(2684),
    [anon_sym_nullptr] = ACTIONS(2684),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2684),
    [anon_sym___typeof] = ACTIONS(2684),
    [anon_sym_typeof] = ACTIONS(2684),
    [anon_sym_ATimport] = ACTIONS(2686),
    [aux_sym_preproc_undef_token1] = ACTIONS(2684),
    [anon_sym_POUND] = ACTIONS(2684),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2684),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2684),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2684),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2684),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2684),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2684),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2684),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_API_AVAILABLE] = ACTIONS(2684),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_API_DEPRECATED] = ACTIONS(2684),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2684),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym___deprecated_msg] = ACTIONS(2684),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2684),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2684),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2684),
    [anon_sym_ATprotocol] = ACTIONS(2686),
    [anon_sym_ATinterface] = ACTIONS(2686),
    [anon_sym_ATimplementation] = ACTIONS(2686),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2686),
    [anon_sym_ATsynthesize] = ACTIONS(2686),
    [anon_sym_ATdynamic] = ACTIONS(2686),
    [anon_sym_ATtry] = ACTIONS(2686),
    [anon_sym___try] = ACTIONS(2684),
    [anon_sym_ATthrow] = ACTIONS(2686),
    [anon_sym_ATselector] = ACTIONS(2686),
    [anon_sym_ATavailable] = ACTIONS(2686),
    [anon_sym___builtin_available] = ACTIONS(2684),
    [anon_sym_va_arg] = ACTIONS(2684),
    [anon_sym_ATencode] = ACTIONS(2686),
    [anon_sym_ATsynchronized] = ACTIONS(2686),
    [anon_sym_BOOL] = ACTIONS(2684),
    [anon_sym_IMP] = ACTIONS(2684),
    [anon_sym_SEL] = ACTIONS(2684),
    [anon_sym_Class] = ACTIONS(2684),
    [anon_sym_id] = ACTIONS(2684),
  },
  [STATE(513)] = {
    [sym_identifier] = ACTIONS(2776),
    [aux_sym_preproc_include_token1] = ACTIONS(2776),
    [aux_sym_preproc_include_token2] = ACTIONS(2776),
    [aux_sym_preproc_def_token1] = ACTIONS(2776),
    [aux_sym_preproc_if_token1] = ACTIONS(2776),
    [aux_sym_preproc_if_token2] = ACTIONS(2776),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2776),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2776),
    [aux_sym_preproc_else_token1] = ACTIONS(2776),
    [aux_sym_preproc_elif_token1] = ACTIONS(2776),
    [sym_preproc_directive] = ACTIONS(2776),
    [anon_sym_LPAREN2] = ACTIONS(2778),
    [anon_sym_BANG] = ACTIONS(2778),
    [anon_sym_TILDE] = ACTIONS(2778),
    [anon_sym_DASH] = ACTIONS(2776),
    [anon_sym_PLUS] = ACTIONS(2776),
    [anon_sym_STAR] = ACTIONS(2778),
    [anon_sym_CARET] = ACTIONS(2778),
    [anon_sym_AMP] = ACTIONS(2778),
    [anon_sym_SEMI] = ACTIONS(2778),
    [anon_sym___extension__] = ACTIONS(2776),
    [anon_sym_typedef] = ACTIONS(2776),
    [anon_sym_extern] = ACTIONS(2776),
    [anon_sym___attribute__] = ACTIONS(2776),
    [anon_sym___attribute] = ACTIONS(2776),
    [anon_sym_noreturn] = ACTIONS(2776),
    [anon_sym_LBRACK] = ACTIONS(2778),
    [anon_sym___declspec] = ACTIONS(2776),
    [anon_sym___cdecl] = ACTIONS(2776),
    [anon_sym___clrcall] = ACTIONS(2776),
    [anon_sym___stdcall] = ACTIONS(2776),
    [anon_sym___fastcall] = ACTIONS(2776),
    [anon_sym___thiscall] = ACTIONS(2776),
    [anon_sym___vectorcall] = ACTIONS(2776),
    [anon_sym_LBRACE] = ACTIONS(2778),
    [anon_sym_signed] = ACTIONS(2776),
    [anon_sym_unsigned] = ACTIONS(2776),
    [anon_sym_long] = ACTIONS(2776),
    [anon_sym_short] = ACTIONS(2776),
    [anon_sym_static] = ACTIONS(2776),
    [anon_sym_ATautoreleasepool] = ACTIONS(2778),
    [anon_sym_auto] = ACTIONS(2776),
    [anon_sym_register] = ACTIONS(2776),
    [anon_sym_inline] = ACTIONS(2776),
    [anon_sym___inline] = ACTIONS(2776),
    [anon_sym___inline__] = ACTIONS(2776),
    [anon_sym___forceinline] = ACTIONS(2776),
    [anon_sym_thread_local] = ACTIONS(2776),
    [anon_sym___thread] = ACTIONS(2776),
    [anon_sym_CG_EXTERN] = ACTIONS(2776),
    [anon_sym_CG_INLINE] = ACTIONS(2776),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2776),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2776),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2776),
    [anon_sym_IBOutlet] = ACTIONS(2776),
    [anon_sym_IBInspectable] = ACTIONS(2776),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2776),
    [anon_sym_NS_INLINE] = ACTIONS(2776),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2776),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2776),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2776),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2776),
    [anon_sym_const] = ACTIONS(2776),
    [anon_sym_constexpr] = ACTIONS(2776),
    [anon_sym_volatile] = ACTIONS(2776),
    [anon_sym_restrict] = ACTIONS(2776),
    [anon_sym___restrict__] = ACTIONS(2776),
    [anon_sym__Atomic] = ACTIONS(2776),
    [anon_sym__Noreturn] = ACTIONS(2776),
    [anon_sym__Nonnull] = ACTIONS(2776),
    [anon_sym_nullable] = ACTIONS(2776),
    [anon_sym__Complex] = ACTIONS(2776),
    [anon_sym__Nullable] = ACTIONS(2776),
    [anon_sym__Nullable_result] = ACTIONS(2776),
    [anon_sym__Null_unspecified] = ACTIONS(2776),
    [anon_sym___autoreleasing] = ACTIONS(2776),
    [anon_sym___block] = ACTIONS(2776),
    [anon_sym___bridge] = ACTIONS(2776),
    [anon_sym___bridge_retained] = ACTIONS(2776),
    [anon_sym___bridge_transfer] = ACTIONS(2776),
    [anon_sym___complex] = ACTIONS(2776),
    [anon_sym___const] = ACTIONS(2776),
    [anon_sym___imag] = ACTIONS(2776),
    [anon_sym___kindof] = ACTIONS(2776),
    [anon_sym___nonnull] = ACTIONS(2776),
    [anon_sym___nullable] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2776),
    [anon_sym___real] = ACTIONS(2776),
    [anon_sym___strong] = ACTIONS(2776),
    [anon_sym___unsafe_unretained] = ACTIONS(2776),
    [anon_sym___unused] = ACTIONS(2776),
    [anon_sym___weak] = ACTIONS(2776),
    [anon_sym_alignas] = ACTIONS(2776),
    [anon_sym__Alignas] = ACTIONS(2776),
    [sym_primitive_type] = ACTIONS(2776),
    [anon_sym_enum] = ACTIONS(2776),
    [anon_sym_struct] = ACTIONS(2776),
    [anon_sym_union] = ACTIONS(2776),
    [anon_sym_if] = ACTIONS(2776),
    [anon_sym_switch] = ACTIONS(2776),
    [anon_sym_case] = ACTIONS(2776),
    [anon_sym_default] = ACTIONS(2776),
    [anon_sym_while] = ACTIONS(2776),
    [anon_sym_do] = ACTIONS(2776),
    [anon_sym_for] = ACTIONS(2776),
    [anon_sym_in] = ACTIONS(2776),
    [anon_sym_return] = ACTIONS(2776),
    [anon_sym_break] = ACTIONS(2776),
    [anon_sym_continue] = ACTIONS(2776),
    [anon_sym_goto] = ACTIONS(2776),
    [anon_sym_DASH_DASH] = ACTIONS(2778),
    [anon_sym_PLUS_PLUS] = ACTIONS(2778),
    [anon_sym_sizeof] = ACTIONS(2776),
    [anon_sym___alignof__] = ACTIONS(2776),
    [anon_sym___alignof] = ACTIONS(2776),
    [anon_sym__alignof] = ACTIONS(2776),
    [anon_sym_alignof] = ACTIONS(2776),
    [anon_sym__Alignof] = ACTIONS(2776),
    [anon_sym_offsetof] = ACTIONS(2776),
    [anon_sym__Generic] = ACTIONS(2776),
    [anon_sym_asm] = ACTIONS(2776),
    [anon_sym___asm__] = ACTIONS(2776),
    [anon_sym___asm] = ACTIONS(2776),
    [sym_number_literal] = ACTIONS(2778),
    [anon_sym_L_SQUOTE] = ACTIONS(2778),
    [anon_sym_u_SQUOTE] = ACTIONS(2778),
    [anon_sym_U_SQUOTE] = ACTIONS(2778),
    [anon_sym_u8_SQUOTE] = ACTIONS(2778),
    [anon_sym_SQUOTE] = ACTIONS(2778),
    [anon_sym_AT] = ACTIONS(2776),
    [anon_sym_DQUOTE] = ACTIONS(2778),
    [anon_sym_L_DQUOTE] = ACTIONS(2778),
    [anon_sym_u_DQUOTE] = ACTIONS(2778),
    [anon_sym_U_DQUOTE] = ACTIONS(2778),
    [anon_sym_u8_DQUOTE] = ACTIONS(2778),
    [sym_true] = ACTIONS(2776),
    [sym_false] = ACTIONS(2776),
    [anon_sym_NULL] = ACTIONS(2776),
    [anon_sym_nullptr] = ACTIONS(2776),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2776),
    [anon_sym___typeof] = ACTIONS(2776),
    [anon_sym_typeof] = ACTIONS(2776),
    [anon_sym_ATimport] = ACTIONS(2778),
    [aux_sym_preproc_undef_token1] = ACTIONS(2776),
    [anon_sym_POUND] = ACTIONS(2776),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2776),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2776),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_API_AVAILABLE] = ACTIONS(2776),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_API_DEPRECATED] = ACTIONS(2776),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2776),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym___deprecated_msg] = ACTIONS(2776),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2776),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2776),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2776),
    [anon_sym_ATprotocol] = ACTIONS(2778),
    [anon_sym_ATinterface] = ACTIONS(2778),
    [anon_sym_ATimplementation] = ACTIONS(2778),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2778),
    [anon_sym_ATsynthesize] = ACTIONS(2778),
    [anon_sym_ATdynamic] = ACTIONS(2778),
    [anon_sym_ATtry] = ACTIONS(2778),
    [anon_sym___try] = ACTIONS(2776),
    [anon_sym_ATthrow] = ACTIONS(2778),
    [anon_sym_ATselector] = ACTIONS(2778),
    [anon_sym_ATavailable] = ACTIONS(2778),
    [anon_sym___builtin_available] = ACTIONS(2776),
    [anon_sym_va_arg] = ACTIONS(2776),
    [anon_sym_ATencode] = ACTIONS(2778),
    [anon_sym_ATsynchronized] = ACTIONS(2778),
    [anon_sym_BOOL] = ACTIONS(2776),
    [anon_sym_IMP] = ACTIONS(2776),
    [anon_sym_SEL] = ACTIONS(2776),
    [anon_sym_Class] = ACTIONS(2776),
    [anon_sym_id] = ACTIONS(2776),
  },
  [STATE(514)] = {
    [sym_identifier] = ACTIONS(2642),
    [aux_sym_preproc_include_token1] = ACTIONS(2642),
    [aux_sym_preproc_include_token2] = ACTIONS(2642),
    [aux_sym_preproc_def_token1] = ACTIONS(2642),
    [aux_sym_preproc_if_token1] = ACTIONS(2642),
    [aux_sym_preproc_if_token2] = ACTIONS(2642),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2642),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2642),
    [aux_sym_preproc_else_token1] = ACTIONS(2642),
    [aux_sym_preproc_elif_token1] = ACTIONS(2642),
    [sym_preproc_directive] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_BANG] = ACTIONS(2644),
    [anon_sym_TILDE] = ACTIONS(2644),
    [anon_sym_DASH] = ACTIONS(2642),
    [anon_sym_PLUS] = ACTIONS(2642),
    [anon_sym_STAR] = ACTIONS(2644),
    [anon_sym_CARET] = ACTIONS(2644),
    [anon_sym_AMP] = ACTIONS(2644),
    [anon_sym_SEMI] = ACTIONS(2644),
    [anon_sym___extension__] = ACTIONS(2642),
    [anon_sym_typedef] = ACTIONS(2642),
    [anon_sym_extern] = ACTIONS(2642),
    [anon_sym___attribute__] = ACTIONS(2642),
    [anon_sym___attribute] = ACTIONS(2642),
    [anon_sym_noreturn] = ACTIONS(2642),
    [anon_sym_LBRACK] = ACTIONS(2644),
    [anon_sym___declspec] = ACTIONS(2642),
    [anon_sym___cdecl] = ACTIONS(2642),
    [anon_sym___clrcall] = ACTIONS(2642),
    [anon_sym___stdcall] = ACTIONS(2642),
    [anon_sym___fastcall] = ACTIONS(2642),
    [anon_sym___thiscall] = ACTIONS(2642),
    [anon_sym___vectorcall] = ACTIONS(2642),
    [anon_sym_LBRACE] = ACTIONS(2644),
    [anon_sym_signed] = ACTIONS(2642),
    [anon_sym_unsigned] = ACTIONS(2642),
    [anon_sym_long] = ACTIONS(2642),
    [anon_sym_short] = ACTIONS(2642),
    [anon_sym_static] = ACTIONS(2642),
    [anon_sym_ATautoreleasepool] = ACTIONS(2644),
    [anon_sym_auto] = ACTIONS(2642),
    [anon_sym_register] = ACTIONS(2642),
    [anon_sym_inline] = ACTIONS(2642),
    [anon_sym___inline] = ACTIONS(2642),
    [anon_sym___inline__] = ACTIONS(2642),
    [anon_sym___forceinline] = ACTIONS(2642),
    [anon_sym_thread_local] = ACTIONS(2642),
    [anon_sym___thread] = ACTIONS(2642),
    [anon_sym_CG_EXTERN] = ACTIONS(2642),
    [anon_sym_CG_INLINE] = ACTIONS(2642),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2642),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2642),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2642),
    [anon_sym_IBOutlet] = ACTIONS(2642),
    [anon_sym_IBInspectable] = ACTIONS(2642),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2642),
    [anon_sym_NS_INLINE] = ACTIONS(2642),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2642),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2642),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2642),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2642),
    [anon_sym_const] = ACTIONS(2642),
    [anon_sym_constexpr] = ACTIONS(2642),
    [anon_sym_volatile] = ACTIONS(2642),
    [anon_sym_restrict] = ACTIONS(2642),
    [anon_sym___restrict__] = ACTIONS(2642),
    [anon_sym__Atomic] = ACTIONS(2642),
    [anon_sym__Noreturn] = ACTIONS(2642),
    [anon_sym__Nonnull] = ACTIONS(2642),
    [anon_sym_nullable] = ACTIONS(2642),
    [anon_sym__Complex] = ACTIONS(2642),
    [anon_sym__Nullable] = ACTIONS(2642),
    [anon_sym__Nullable_result] = ACTIONS(2642),
    [anon_sym__Null_unspecified] = ACTIONS(2642),
    [anon_sym___autoreleasing] = ACTIONS(2642),
    [anon_sym___block] = ACTIONS(2642),
    [anon_sym___bridge] = ACTIONS(2642),
    [anon_sym___bridge_retained] = ACTIONS(2642),
    [anon_sym___bridge_transfer] = ACTIONS(2642),
    [anon_sym___complex] = ACTIONS(2642),
    [anon_sym___const] = ACTIONS(2642),
    [anon_sym___imag] = ACTIONS(2642),
    [anon_sym___kindof] = ACTIONS(2642),
    [anon_sym___nonnull] = ACTIONS(2642),
    [anon_sym___nullable] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2642),
    [anon_sym___real] = ACTIONS(2642),
    [anon_sym___strong] = ACTIONS(2642),
    [anon_sym___unsafe_unretained] = ACTIONS(2642),
    [anon_sym___unused] = ACTIONS(2642),
    [anon_sym___weak] = ACTIONS(2642),
    [anon_sym_alignas] = ACTIONS(2642),
    [anon_sym__Alignas] = ACTIONS(2642),
    [sym_primitive_type] = ACTIONS(2642),
    [anon_sym_enum] = ACTIONS(2642),
    [anon_sym_struct] = ACTIONS(2642),
    [anon_sym_union] = ACTIONS(2642),
    [anon_sym_if] = ACTIONS(2642),
    [anon_sym_switch] = ACTIONS(2642),
    [anon_sym_case] = ACTIONS(2642),
    [anon_sym_default] = ACTIONS(2642),
    [anon_sym_while] = ACTIONS(2642),
    [anon_sym_do] = ACTIONS(2642),
    [anon_sym_for] = ACTIONS(2642),
    [anon_sym_in] = ACTIONS(2642),
    [anon_sym_return] = ACTIONS(2642),
    [anon_sym_break] = ACTIONS(2642),
    [anon_sym_continue] = ACTIONS(2642),
    [anon_sym_goto] = ACTIONS(2642),
    [anon_sym_DASH_DASH] = ACTIONS(2644),
    [anon_sym_PLUS_PLUS] = ACTIONS(2644),
    [anon_sym_sizeof] = ACTIONS(2642),
    [anon_sym___alignof__] = ACTIONS(2642),
    [anon_sym___alignof] = ACTIONS(2642),
    [anon_sym__alignof] = ACTIONS(2642),
    [anon_sym_alignof] = ACTIONS(2642),
    [anon_sym__Alignof] = ACTIONS(2642),
    [anon_sym_offsetof] = ACTIONS(2642),
    [anon_sym__Generic] = ACTIONS(2642),
    [anon_sym_asm] = ACTIONS(2642),
    [anon_sym___asm__] = ACTIONS(2642),
    [anon_sym___asm] = ACTIONS(2642),
    [sym_number_literal] = ACTIONS(2644),
    [anon_sym_L_SQUOTE] = ACTIONS(2644),
    [anon_sym_u_SQUOTE] = ACTIONS(2644),
    [anon_sym_U_SQUOTE] = ACTIONS(2644),
    [anon_sym_u8_SQUOTE] = ACTIONS(2644),
    [anon_sym_SQUOTE] = ACTIONS(2644),
    [anon_sym_AT] = ACTIONS(2642),
    [anon_sym_DQUOTE] = ACTIONS(2644),
    [anon_sym_L_DQUOTE] = ACTIONS(2644),
    [anon_sym_u_DQUOTE] = ACTIONS(2644),
    [anon_sym_U_DQUOTE] = ACTIONS(2644),
    [anon_sym_u8_DQUOTE] = ACTIONS(2644),
    [sym_true] = ACTIONS(2642),
    [sym_false] = ACTIONS(2642),
    [anon_sym_NULL] = ACTIONS(2642),
    [anon_sym_nullptr] = ACTIONS(2642),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2642),
    [anon_sym___typeof] = ACTIONS(2642),
    [anon_sym_typeof] = ACTIONS(2642),
    [anon_sym_ATimport] = ACTIONS(2644),
    [aux_sym_preproc_undef_token1] = ACTIONS(2642),
    [anon_sym_POUND] = ACTIONS(2642),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2642),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2642),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2642),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2642),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2642),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2642),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2642),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_API_AVAILABLE] = ACTIONS(2642),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_API_DEPRECATED] = ACTIONS(2642),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2642),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym___deprecated_msg] = ACTIONS(2642),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2642),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2642),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2642),
    [anon_sym_ATprotocol] = ACTIONS(2644),
    [anon_sym_ATinterface] = ACTIONS(2644),
    [anon_sym_ATimplementation] = ACTIONS(2644),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2644),
    [anon_sym_ATsynthesize] = ACTIONS(2644),
    [anon_sym_ATdynamic] = ACTIONS(2644),
    [anon_sym_ATtry] = ACTIONS(2644),
    [anon_sym___try] = ACTIONS(2642),
    [anon_sym_ATthrow] = ACTIONS(2644),
    [anon_sym_ATselector] = ACTIONS(2644),
    [anon_sym_ATavailable] = ACTIONS(2644),
    [anon_sym___builtin_available] = ACTIONS(2642),
    [anon_sym_va_arg] = ACTIONS(2642),
    [anon_sym_ATencode] = ACTIONS(2644),
    [anon_sym_ATsynchronized] = ACTIONS(2644),
    [anon_sym_BOOL] = ACTIONS(2642),
    [anon_sym_IMP] = ACTIONS(2642),
    [anon_sym_SEL] = ACTIONS(2642),
    [anon_sym_Class] = ACTIONS(2642),
    [anon_sym_id] = ACTIONS(2642),
  },
  [STATE(515)] = {
    [sym_identifier] = ACTIONS(2654),
    [aux_sym_preproc_include_token1] = ACTIONS(2654),
    [aux_sym_preproc_include_token2] = ACTIONS(2654),
    [aux_sym_preproc_def_token1] = ACTIONS(2654),
    [aux_sym_preproc_if_token1] = ACTIONS(2654),
    [aux_sym_preproc_if_token2] = ACTIONS(2654),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2654),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2654),
    [aux_sym_preproc_else_token1] = ACTIONS(2654),
    [aux_sym_preproc_elif_token1] = ACTIONS(2654),
    [sym_preproc_directive] = ACTIONS(2654),
    [anon_sym_LPAREN2] = ACTIONS(2656),
    [anon_sym_BANG] = ACTIONS(2656),
    [anon_sym_TILDE] = ACTIONS(2656),
    [anon_sym_DASH] = ACTIONS(2654),
    [anon_sym_PLUS] = ACTIONS(2654),
    [anon_sym_STAR] = ACTIONS(2656),
    [anon_sym_CARET] = ACTIONS(2656),
    [anon_sym_AMP] = ACTIONS(2656),
    [anon_sym_SEMI] = ACTIONS(2656),
    [anon_sym___extension__] = ACTIONS(2654),
    [anon_sym_typedef] = ACTIONS(2654),
    [anon_sym_extern] = ACTIONS(2654),
    [anon_sym___attribute__] = ACTIONS(2654),
    [anon_sym___attribute] = ACTIONS(2654),
    [anon_sym_noreturn] = ACTIONS(2654),
    [anon_sym_LBRACK] = ACTIONS(2656),
    [anon_sym___declspec] = ACTIONS(2654),
    [anon_sym___cdecl] = ACTIONS(2654),
    [anon_sym___clrcall] = ACTIONS(2654),
    [anon_sym___stdcall] = ACTIONS(2654),
    [anon_sym___fastcall] = ACTIONS(2654),
    [anon_sym___thiscall] = ACTIONS(2654),
    [anon_sym___vectorcall] = ACTIONS(2654),
    [anon_sym_LBRACE] = ACTIONS(2656),
    [anon_sym_signed] = ACTIONS(2654),
    [anon_sym_unsigned] = ACTIONS(2654),
    [anon_sym_long] = ACTIONS(2654),
    [anon_sym_short] = ACTIONS(2654),
    [anon_sym_static] = ACTIONS(2654),
    [anon_sym_ATautoreleasepool] = ACTIONS(2656),
    [anon_sym_auto] = ACTIONS(2654),
    [anon_sym_register] = ACTIONS(2654),
    [anon_sym_inline] = ACTIONS(2654),
    [anon_sym___inline] = ACTIONS(2654),
    [anon_sym___inline__] = ACTIONS(2654),
    [anon_sym___forceinline] = ACTIONS(2654),
    [anon_sym_thread_local] = ACTIONS(2654),
    [anon_sym___thread] = ACTIONS(2654),
    [anon_sym_CG_EXTERN] = ACTIONS(2654),
    [anon_sym_CG_INLINE] = ACTIONS(2654),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2654),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2654),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2654),
    [anon_sym_IBOutlet] = ACTIONS(2654),
    [anon_sym_IBInspectable] = ACTIONS(2654),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2654),
    [anon_sym_NS_INLINE] = ACTIONS(2654),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2654),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2654),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2654),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2654),
    [anon_sym_const] = ACTIONS(2654),
    [anon_sym_constexpr] = ACTIONS(2654),
    [anon_sym_volatile] = ACTIONS(2654),
    [anon_sym_restrict] = ACTIONS(2654),
    [anon_sym___restrict__] = ACTIONS(2654),
    [anon_sym__Atomic] = ACTIONS(2654),
    [anon_sym__Noreturn] = ACTIONS(2654),
    [anon_sym__Nonnull] = ACTIONS(2654),
    [anon_sym_nullable] = ACTIONS(2654),
    [anon_sym__Complex] = ACTIONS(2654),
    [anon_sym__Nullable] = ACTIONS(2654),
    [anon_sym__Nullable_result] = ACTIONS(2654),
    [anon_sym__Null_unspecified] = ACTIONS(2654),
    [anon_sym___autoreleasing] = ACTIONS(2654),
    [anon_sym___block] = ACTIONS(2654),
    [anon_sym___bridge] = ACTIONS(2654),
    [anon_sym___bridge_retained] = ACTIONS(2654),
    [anon_sym___bridge_transfer] = ACTIONS(2654),
    [anon_sym___complex] = ACTIONS(2654),
    [anon_sym___const] = ACTIONS(2654),
    [anon_sym___imag] = ACTIONS(2654),
    [anon_sym___kindof] = ACTIONS(2654),
    [anon_sym___nonnull] = ACTIONS(2654),
    [anon_sym___nullable] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2654),
    [anon_sym___real] = ACTIONS(2654),
    [anon_sym___strong] = ACTIONS(2654),
    [anon_sym___unsafe_unretained] = ACTIONS(2654),
    [anon_sym___unused] = ACTIONS(2654),
    [anon_sym___weak] = ACTIONS(2654),
    [anon_sym_alignas] = ACTIONS(2654),
    [anon_sym__Alignas] = ACTIONS(2654),
    [sym_primitive_type] = ACTIONS(2654),
    [anon_sym_enum] = ACTIONS(2654),
    [anon_sym_struct] = ACTIONS(2654),
    [anon_sym_union] = ACTIONS(2654),
    [anon_sym_if] = ACTIONS(2654),
    [anon_sym_switch] = ACTIONS(2654),
    [anon_sym_case] = ACTIONS(2654),
    [anon_sym_default] = ACTIONS(2654),
    [anon_sym_while] = ACTIONS(2654),
    [anon_sym_do] = ACTIONS(2654),
    [anon_sym_for] = ACTIONS(2654),
    [anon_sym_in] = ACTIONS(2654),
    [anon_sym_return] = ACTIONS(2654),
    [anon_sym_break] = ACTIONS(2654),
    [anon_sym_continue] = ACTIONS(2654),
    [anon_sym_goto] = ACTIONS(2654),
    [anon_sym_DASH_DASH] = ACTIONS(2656),
    [anon_sym_PLUS_PLUS] = ACTIONS(2656),
    [anon_sym_sizeof] = ACTIONS(2654),
    [anon_sym___alignof__] = ACTIONS(2654),
    [anon_sym___alignof] = ACTIONS(2654),
    [anon_sym__alignof] = ACTIONS(2654),
    [anon_sym_alignof] = ACTIONS(2654),
    [anon_sym__Alignof] = ACTIONS(2654),
    [anon_sym_offsetof] = ACTIONS(2654),
    [anon_sym__Generic] = ACTIONS(2654),
    [anon_sym_asm] = ACTIONS(2654),
    [anon_sym___asm__] = ACTIONS(2654),
    [anon_sym___asm] = ACTIONS(2654),
    [sym_number_literal] = ACTIONS(2656),
    [anon_sym_L_SQUOTE] = ACTIONS(2656),
    [anon_sym_u_SQUOTE] = ACTIONS(2656),
    [anon_sym_U_SQUOTE] = ACTIONS(2656),
    [anon_sym_u8_SQUOTE] = ACTIONS(2656),
    [anon_sym_SQUOTE] = ACTIONS(2656),
    [anon_sym_AT] = ACTIONS(2654),
    [anon_sym_DQUOTE] = ACTIONS(2656),
    [anon_sym_L_DQUOTE] = ACTIONS(2656),
    [anon_sym_u_DQUOTE] = ACTIONS(2656),
    [anon_sym_U_DQUOTE] = ACTIONS(2656),
    [anon_sym_u8_DQUOTE] = ACTIONS(2656),
    [sym_true] = ACTIONS(2654),
    [sym_false] = ACTIONS(2654),
    [anon_sym_NULL] = ACTIONS(2654),
    [anon_sym_nullptr] = ACTIONS(2654),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2654),
    [anon_sym___typeof] = ACTIONS(2654),
    [anon_sym_typeof] = ACTIONS(2654),
    [anon_sym_ATimport] = ACTIONS(2656),
    [aux_sym_preproc_undef_token1] = ACTIONS(2654),
    [anon_sym_POUND] = ACTIONS(2654),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2654),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2654),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2654),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2654),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2654),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2654),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2654),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_API_AVAILABLE] = ACTIONS(2654),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_API_DEPRECATED] = ACTIONS(2654),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2654),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym___deprecated_msg] = ACTIONS(2654),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2654),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2654),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2654),
    [anon_sym_ATprotocol] = ACTIONS(2656),
    [anon_sym_ATinterface] = ACTIONS(2656),
    [anon_sym_ATimplementation] = ACTIONS(2656),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2656),
    [anon_sym_ATsynthesize] = ACTIONS(2656),
    [anon_sym_ATdynamic] = ACTIONS(2656),
    [anon_sym_ATtry] = ACTIONS(2656),
    [anon_sym___try] = ACTIONS(2654),
    [anon_sym_ATthrow] = ACTIONS(2656),
    [anon_sym_ATselector] = ACTIONS(2656),
    [anon_sym_ATavailable] = ACTIONS(2656),
    [anon_sym___builtin_available] = ACTIONS(2654),
    [anon_sym_va_arg] = ACTIONS(2654),
    [anon_sym_ATencode] = ACTIONS(2656),
    [anon_sym_ATsynchronized] = ACTIONS(2656),
    [anon_sym_BOOL] = ACTIONS(2654),
    [anon_sym_IMP] = ACTIONS(2654),
    [anon_sym_SEL] = ACTIONS(2654),
    [anon_sym_Class] = ACTIONS(2654),
    [anon_sym_id] = ACTIONS(2654),
  },
  [STATE(516)] = {
    [sym_identifier] = ACTIONS(2688),
    [aux_sym_preproc_include_token1] = ACTIONS(2688),
    [aux_sym_preproc_include_token2] = ACTIONS(2688),
    [aux_sym_preproc_def_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token2] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2688),
    [aux_sym_preproc_else_token1] = ACTIONS(2688),
    [aux_sym_preproc_elif_token1] = ACTIONS(2688),
    [sym_preproc_directive] = ACTIONS(2688),
    [anon_sym_LPAREN2] = ACTIONS(2690),
    [anon_sym_BANG] = ACTIONS(2690),
    [anon_sym_TILDE] = ACTIONS(2690),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2690),
    [anon_sym_CARET] = ACTIONS(2690),
    [anon_sym_AMP] = ACTIONS(2690),
    [anon_sym_SEMI] = ACTIONS(2690),
    [anon_sym___extension__] = ACTIONS(2688),
    [anon_sym_typedef] = ACTIONS(2688),
    [anon_sym_extern] = ACTIONS(2688),
    [anon_sym___attribute__] = ACTIONS(2688),
    [anon_sym___attribute] = ACTIONS(2688),
    [anon_sym_noreturn] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym___declspec] = ACTIONS(2688),
    [anon_sym___cdecl] = ACTIONS(2688),
    [anon_sym___clrcall] = ACTIONS(2688),
    [anon_sym___stdcall] = ACTIONS(2688),
    [anon_sym___fastcall] = ACTIONS(2688),
    [anon_sym___thiscall] = ACTIONS(2688),
    [anon_sym___vectorcall] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_signed] = ACTIONS(2688),
    [anon_sym_unsigned] = ACTIONS(2688),
    [anon_sym_long] = ACTIONS(2688),
    [anon_sym_short] = ACTIONS(2688),
    [anon_sym_static] = ACTIONS(2688),
    [anon_sym_ATautoreleasepool] = ACTIONS(2690),
    [anon_sym_auto] = ACTIONS(2688),
    [anon_sym_register] = ACTIONS(2688),
    [anon_sym_inline] = ACTIONS(2688),
    [anon_sym___inline] = ACTIONS(2688),
    [anon_sym___inline__] = ACTIONS(2688),
    [anon_sym___forceinline] = ACTIONS(2688),
    [anon_sym_thread_local] = ACTIONS(2688),
    [anon_sym___thread] = ACTIONS(2688),
    [anon_sym_CG_EXTERN] = ACTIONS(2688),
    [anon_sym_CG_INLINE] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2688),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2688),
    [anon_sym_IBOutlet] = ACTIONS(2688),
    [anon_sym_IBInspectable] = ACTIONS(2688),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2688),
    [anon_sym_NS_INLINE] = ACTIONS(2688),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2688),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2688),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2688),
    [anon_sym_const] = ACTIONS(2688),
    [anon_sym_constexpr] = ACTIONS(2688),
    [anon_sym_volatile] = ACTIONS(2688),
    [anon_sym_restrict] = ACTIONS(2688),
    [anon_sym___restrict__] = ACTIONS(2688),
    [anon_sym__Atomic] = ACTIONS(2688),
    [anon_sym__Noreturn] = ACTIONS(2688),
    [anon_sym__Nonnull] = ACTIONS(2688),
    [anon_sym_nullable] = ACTIONS(2688),
    [anon_sym__Complex] = ACTIONS(2688),
    [anon_sym__Nullable] = ACTIONS(2688),
    [anon_sym__Nullable_result] = ACTIONS(2688),
    [anon_sym__Null_unspecified] = ACTIONS(2688),
    [anon_sym___autoreleasing] = ACTIONS(2688),
    [anon_sym___block] = ACTIONS(2688),
    [anon_sym___bridge] = ACTIONS(2688),
    [anon_sym___bridge_retained] = ACTIONS(2688),
    [anon_sym___bridge_transfer] = ACTIONS(2688),
    [anon_sym___complex] = ACTIONS(2688),
    [anon_sym___const] = ACTIONS(2688),
    [anon_sym___imag] = ACTIONS(2688),
    [anon_sym___kindof] = ACTIONS(2688),
    [anon_sym___nonnull] = ACTIONS(2688),
    [anon_sym___nullable] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2688),
    [anon_sym___real] = ACTIONS(2688),
    [anon_sym___strong] = ACTIONS(2688),
    [anon_sym___unsafe_unretained] = ACTIONS(2688),
    [anon_sym___unused] = ACTIONS(2688),
    [anon_sym___weak] = ACTIONS(2688),
    [anon_sym_alignas] = ACTIONS(2688),
    [anon_sym__Alignas] = ACTIONS(2688),
    [sym_primitive_type] = ACTIONS(2688),
    [anon_sym_enum] = ACTIONS(2688),
    [anon_sym_struct] = ACTIONS(2688),
    [anon_sym_union] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_switch] = ACTIONS(2688),
    [anon_sym_case] = ACTIONS(2688),
    [anon_sym_default] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_in] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_sizeof] = ACTIONS(2688),
    [anon_sym___alignof__] = ACTIONS(2688),
    [anon_sym___alignof] = ACTIONS(2688),
    [anon_sym__alignof] = ACTIONS(2688),
    [anon_sym_alignof] = ACTIONS(2688),
    [anon_sym__Alignof] = ACTIONS(2688),
    [anon_sym_offsetof] = ACTIONS(2688),
    [anon_sym__Generic] = ACTIONS(2688),
    [anon_sym_asm] = ACTIONS(2688),
    [anon_sym___asm__] = ACTIONS(2688),
    [anon_sym___asm] = ACTIONS(2688),
    [sym_number_literal] = ACTIONS(2690),
    [anon_sym_L_SQUOTE] = ACTIONS(2690),
    [anon_sym_u_SQUOTE] = ACTIONS(2690),
    [anon_sym_U_SQUOTE] = ACTIONS(2690),
    [anon_sym_u8_SQUOTE] = ACTIONS(2690),
    [anon_sym_SQUOTE] = ACTIONS(2690),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_DQUOTE] = ACTIONS(2690),
    [anon_sym_L_DQUOTE] = ACTIONS(2690),
    [anon_sym_u_DQUOTE] = ACTIONS(2690),
    [anon_sym_U_DQUOTE] = ACTIONS(2690),
    [anon_sym_u8_DQUOTE] = ACTIONS(2690),
    [sym_true] = ACTIONS(2688),
    [sym_false] = ACTIONS(2688),
    [anon_sym_NULL] = ACTIONS(2688),
    [anon_sym_nullptr] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2688),
    [anon_sym___typeof] = ACTIONS(2688),
    [anon_sym_typeof] = ACTIONS(2688),
    [anon_sym_ATimport] = ACTIONS(2690),
    [aux_sym_preproc_undef_token1] = ACTIONS(2688),
    [anon_sym_POUND] = ACTIONS(2688),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2688),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2688),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2688),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_API_AVAILABLE] = ACTIONS(2688),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_API_DEPRECATED] = ACTIONS(2688),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym___deprecated_msg] = ACTIONS(2688),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2688),
    [anon_sym_ATprotocol] = ACTIONS(2690),
    [anon_sym_ATinterface] = ACTIONS(2690),
    [anon_sym_ATimplementation] = ACTIONS(2690),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2690),
    [anon_sym_ATsynthesize] = ACTIONS(2690),
    [anon_sym_ATdynamic] = ACTIONS(2690),
    [anon_sym_ATtry] = ACTIONS(2690),
    [anon_sym___try] = ACTIONS(2688),
    [anon_sym_ATthrow] = ACTIONS(2690),
    [anon_sym_ATselector] = ACTIONS(2690),
    [anon_sym_ATavailable] = ACTIONS(2690),
    [anon_sym___builtin_available] = ACTIONS(2688),
    [anon_sym_va_arg] = ACTIONS(2688),
    [anon_sym_ATencode] = ACTIONS(2690),
    [anon_sym_ATsynchronized] = ACTIONS(2690),
    [anon_sym_BOOL] = ACTIONS(2688),
    [anon_sym_IMP] = ACTIONS(2688),
    [anon_sym_SEL] = ACTIONS(2688),
    [anon_sym_Class] = ACTIONS(2688),
    [anon_sym_id] = ACTIONS(2688),
  },
  [STATE(517)] = {
    [sym_identifier] = ACTIONS(2584),
    [aux_sym_preproc_include_token1] = ACTIONS(2584),
    [aux_sym_preproc_include_token2] = ACTIONS(2584),
    [aux_sym_preproc_def_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token2] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2584),
    [aux_sym_preproc_else_token1] = ACTIONS(2584),
    [aux_sym_preproc_elif_token1] = ACTIONS(2584),
    [sym_preproc_directive] = ACTIONS(2584),
    [anon_sym_LPAREN2] = ACTIONS(2586),
    [anon_sym_BANG] = ACTIONS(2586),
    [anon_sym_TILDE] = ACTIONS(2586),
    [anon_sym_DASH] = ACTIONS(2584),
    [anon_sym_PLUS] = ACTIONS(2584),
    [anon_sym_STAR] = ACTIONS(2586),
    [anon_sym_CARET] = ACTIONS(2586),
    [anon_sym_AMP] = ACTIONS(2586),
    [anon_sym_SEMI] = ACTIONS(2586),
    [anon_sym___extension__] = ACTIONS(2584),
    [anon_sym_typedef] = ACTIONS(2584),
    [anon_sym_extern] = ACTIONS(2584),
    [anon_sym___attribute__] = ACTIONS(2584),
    [anon_sym___attribute] = ACTIONS(2584),
    [anon_sym_noreturn] = ACTIONS(2584),
    [anon_sym_LBRACK] = ACTIONS(2586),
    [anon_sym___declspec] = ACTIONS(2584),
    [anon_sym___cdecl] = ACTIONS(2584),
    [anon_sym___clrcall] = ACTIONS(2584),
    [anon_sym___stdcall] = ACTIONS(2584),
    [anon_sym___fastcall] = ACTIONS(2584),
    [anon_sym___thiscall] = ACTIONS(2584),
    [anon_sym___vectorcall] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_signed] = ACTIONS(2584),
    [anon_sym_unsigned] = ACTIONS(2584),
    [anon_sym_long] = ACTIONS(2584),
    [anon_sym_short] = ACTIONS(2584),
    [anon_sym_static] = ACTIONS(2584),
    [anon_sym_ATautoreleasepool] = ACTIONS(2586),
    [anon_sym_auto] = ACTIONS(2584),
    [anon_sym_register] = ACTIONS(2584),
    [anon_sym_inline] = ACTIONS(2584),
    [anon_sym___inline] = ACTIONS(2584),
    [anon_sym___inline__] = ACTIONS(2584),
    [anon_sym___forceinline] = ACTIONS(2584),
    [anon_sym_thread_local] = ACTIONS(2584),
    [anon_sym___thread] = ACTIONS(2584),
    [anon_sym_CG_EXTERN] = ACTIONS(2584),
    [anon_sym_CG_INLINE] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2584),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2584),
    [anon_sym_IBOutlet] = ACTIONS(2584),
    [anon_sym_IBInspectable] = ACTIONS(2584),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2584),
    [anon_sym_NS_INLINE] = ACTIONS(2584),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2584),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2584),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2584),
    [anon_sym_const] = ACTIONS(2584),
    [anon_sym_constexpr] = ACTIONS(2584),
    [anon_sym_volatile] = ACTIONS(2584),
    [anon_sym_restrict] = ACTIONS(2584),
    [anon_sym___restrict__] = ACTIONS(2584),
    [anon_sym__Atomic] = ACTIONS(2584),
    [anon_sym__Noreturn] = ACTIONS(2584),
    [anon_sym__Nonnull] = ACTIONS(2584),
    [anon_sym_nullable] = ACTIONS(2584),
    [anon_sym__Complex] = ACTIONS(2584),
    [anon_sym__Nullable] = ACTIONS(2584),
    [anon_sym__Nullable_result] = ACTIONS(2584),
    [anon_sym__Null_unspecified] = ACTIONS(2584),
    [anon_sym___autoreleasing] = ACTIONS(2584),
    [anon_sym___block] = ACTIONS(2584),
    [anon_sym___bridge] = ACTIONS(2584),
    [anon_sym___bridge_retained] = ACTIONS(2584),
    [anon_sym___bridge_transfer] = ACTIONS(2584),
    [anon_sym___complex] = ACTIONS(2584),
    [anon_sym___const] = ACTIONS(2584),
    [anon_sym___imag] = ACTIONS(2584),
    [anon_sym___kindof] = ACTIONS(2584),
    [anon_sym___nonnull] = ACTIONS(2584),
    [anon_sym___nullable] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2584),
    [anon_sym___real] = ACTIONS(2584),
    [anon_sym___strong] = ACTIONS(2584),
    [anon_sym___unsafe_unretained] = ACTIONS(2584),
    [anon_sym___unused] = ACTIONS(2584),
    [anon_sym___weak] = ACTIONS(2584),
    [anon_sym_alignas] = ACTIONS(2584),
    [anon_sym__Alignas] = ACTIONS(2584),
    [sym_primitive_type] = ACTIONS(2584),
    [anon_sym_enum] = ACTIONS(2584),
    [anon_sym_struct] = ACTIONS(2584),
    [anon_sym_union] = ACTIONS(2584),
    [anon_sym_if] = ACTIONS(2584),
    [anon_sym_switch] = ACTIONS(2584),
    [anon_sym_case] = ACTIONS(2584),
    [anon_sym_default] = ACTIONS(2584),
    [anon_sym_while] = ACTIONS(2584),
    [anon_sym_do] = ACTIONS(2584),
    [anon_sym_for] = ACTIONS(2584),
    [anon_sym_in] = ACTIONS(2584),
    [anon_sym_return] = ACTIONS(2584),
    [anon_sym_break] = ACTIONS(2584),
    [anon_sym_continue] = ACTIONS(2584),
    [anon_sym_goto] = ACTIONS(2584),
    [anon_sym_DASH_DASH] = ACTIONS(2586),
    [anon_sym_PLUS_PLUS] = ACTIONS(2586),
    [anon_sym_sizeof] = ACTIONS(2584),
    [anon_sym___alignof__] = ACTIONS(2584),
    [anon_sym___alignof] = ACTIONS(2584),
    [anon_sym__alignof] = ACTIONS(2584),
    [anon_sym_alignof] = ACTIONS(2584),
    [anon_sym__Alignof] = ACTIONS(2584),
    [anon_sym_offsetof] = ACTIONS(2584),
    [anon_sym__Generic] = ACTIONS(2584),
    [anon_sym_asm] = ACTIONS(2584),
    [anon_sym___asm__] = ACTIONS(2584),
    [anon_sym___asm] = ACTIONS(2584),
    [sym_number_literal] = ACTIONS(2586),
    [anon_sym_L_SQUOTE] = ACTIONS(2586),
    [anon_sym_u_SQUOTE] = ACTIONS(2586),
    [anon_sym_U_SQUOTE] = ACTIONS(2586),
    [anon_sym_u8_SQUOTE] = ACTIONS(2586),
    [anon_sym_SQUOTE] = ACTIONS(2586),
    [anon_sym_AT] = ACTIONS(2584),
    [anon_sym_DQUOTE] = ACTIONS(2586),
    [anon_sym_L_DQUOTE] = ACTIONS(2586),
    [anon_sym_u_DQUOTE] = ACTIONS(2586),
    [anon_sym_U_DQUOTE] = ACTIONS(2586),
    [anon_sym_u8_DQUOTE] = ACTIONS(2586),
    [sym_true] = ACTIONS(2584),
    [sym_false] = ACTIONS(2584),
    [anon_sym_NULL] = ACTIONS(2584),
    [anon_sym_nullptr] = ACTIONS(2584),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2584),
    [anon_sym___typeof] = ACTIONS(2584),
    [anon_sym_typeof] = ACTIONS(2584),
    [anon_sym_ATimport] = ACTIONS(2586),
    [aux_sym_preproc_undef_token1] = ACTIONS(2584),
    [anon_sym_POUND] = ACTIONS(2584),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2584),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2584),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2584),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_API_AVAILABLE] = ACTIONS(2584),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_API_DEPRECATED] = ACTIONS(2584),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym___deprecated_msg] = ACTIONS(2584),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2584),
    [anon_sym_ATprotocol] = ACTIONS(2586),
    [anon_sym_ATinterface] = ACTIONS(2586),
    [anon_sym_ATimplementation] = ACTIONS(2586),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2586),
    [anon_sym_ATsynthesize] = ACTIONS(2586),
    [anon_sym_ATdynamic] = ACTIONS(2586),
    [anon_sym_ATtry] = ACTIONS(2586),
    [anon_sym___try] = ACTIONS(2584),
    [anon_sym_ATthrow] = ACTIONS(2586),
    [anon_sym_ATselector] = ACTIONS(2586),
    [anon_sym_ATavailable] = ACTIONS(2586),
    [anon_sym___builtin_available] = ACTIONS(2584),
    [anon_sym_va_arg] = ACTIONS(2584),
    [anon_sym_ATencode] = ACTIONS(2586),
    [anon_sym_ATsynchronized] = ACTIONS(2586),
    [anon_sym_BOOL] = ACTIONS(2584),
    [anon_sym_IMP] = ACTIONS(2584),
    [anon_sym_SEL] = ACTIONS(2584),
    [anon_sym_Class] = ACTIONS(2584),
    [anon_sym_id] = ACTIONS(2584),
  },
  [STATE(518)] = {
    [sym_identifier] = ACTIONS(2608),
    [aux_sym_preproc_include_token1] = ACTIONS(2608),
    [aux_sym_preproc_include_token2] = ACTIONS(2608),
    [aux_sym_preproc_def_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token2] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2608),
    [aux_sym_preproc_else_token1] = ACTIONS(2608),
    [aux_sym_preproc_elif_token1] = ACTIONS(2608),
    [sym_preproc_directive] = ACTIONS(2608),
    [anon_sym_LPAREN2] = ACTIONS(2610),
    [anon_sym_BANG] = ACTIONS(2610),
    [anon_sym_TILDE] = ACTIONS(2610),
    [anon_sym_DASH] = ACTIONS(2608),
    [anon_sym_PLUS] = ACTIONS(2608),
    [anon_sym_STAR] = ACTIONS(2610),
    [anon_sym_CARET] = ACTIONS(2610),
    [anon_sym_AMP] = ACTIONS(2610),
    [anon_sym_SEMI] = ACTIONS(2610),
    [anon_sym___extension__] = ACTIONS(2608),
    [anon_sym_typedef] = ACTIONS(2608),
    [anon_sym_extern] = ACTIONS(2608),
    [anon_sym___attribute__] = ACTIONS(2608),
    [anon_sym___attribute] = ACTIONS(2608),
    [anon_sym_noreturn] = ACTIONS(2608),
    [anon_sym_LBRACK] = ACTIONS(2610),
    [anon_sym___declspec] = ACTIONS(2608),
    [anon_sym___cdecl] = ACTIONS(2608),
    [anon_sym___clrcall] = ACTIONS(2608),
    [anon_sym___stdcall] = ACTIONS(2608),
    [anon_sym___fastcall] = ACTIONS(2608),
    [anon_sym___thiscall] = ACTIONS(2608),
    [anon_sym___vectorcall] = ACTIONS(2608),
    [anon_sym_LBRACE] = ACTIONS(2610),
    [anon_sym_signed] = ACTIONS(2608),
    [anon_sym_unsigned] = ACTIONS(2608),
    [anon_sym_long] = ACTIONS(2608),
    [anon_sym_short] = ACTIONS(2608),
    [anon_sym_static] = ACTIONS(2608),
    [anon_sym_ATautoreleasepool] = ACTIONS(2610),
    [anon_sym_auto] = ACTIONS(2608),
    [anon_sym_register] = ACTIONS(2608),
    [anon_sym_inline] = ACTIONS(2608),
    [anon_sym___inline] = ACTIONS(2608),
    [anon_sym___inline__] = ACTIONS(2608),
    [anon_sym___forceinline] = ACTIONS(2608),
    [anon_sym_thread_local] = ACTIONS(2608),
    [anon_sym___thread] = ACTIONS(2608),
    [anon_sym_CG_EXTERN] = ACTIONS(2608),
    [anon_sym_CG_INLINE] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2608),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2608),
    [anon_sym_IBOutlet] = ACTIONS(2608),
    [anon_sym_IBInspectable] = ACTIONS(2608),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2608),
    [anon_sym_NS_INLINE] = ACTIONS(2608),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2608),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2608),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2608),
    [anon_sym_const] = ACTIONS(2608),
    [anon_sym_constexpr] = ACTIONS(2608),
    [anon_sym_volatile] = ACTIONS(2608),
    [anon_sym_restrict] = ACTIONS(2608),
    [anon_sym___restrict__] = ACTIONS(2608),
    [anon_sym__Atomic] = ACTIONS(2608),
    [anon_sym__Noreturn] = ACTIONS(2608),
    [anon_sym__Nonnull] = ACTIONS(2608),
    [anon_sym_nullable] = ACTIONS(2608),
    [anon_sym__Complex] = ACTIONS(2608),
    [anon_sym__Nullable] = ACTIONS(2608),
    [anon_sym__Nullable_result] = ACTIONS(2608),
    [anon_sym__Null_unspecified] = ACTIONS(2608),
    [anon_sym___autoreleasing] = ACTIONS(2608),
    [anon_sym___block] = ACTIONS(2608),
    [anon_sym___bridge] = ACTIONS(2608),
    [anon_sym___bridge_retained] = ACTIONS(2608),
    [anon_sym___bridge_transfer] = ACTIONS(2608),
    [anon_sym___complex] = ACTIONS(2608),
    [anon_sym___const] = ACTIONS(2608),
    [anon_sym___imag] = ACTIONS(2608),
    [anon_sym___kindof] = ACTIONS(2608),
    [anon_sym___nonnull] = ACTIONS(2608),
    [anon_sym___nullable] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2608),
    [anon_sym___real] = ACTIONS(2608),
    [anon_sym___strong] = ACTIONS(2608),
    [anon_sym___unsafe_unretained] = ACTIONS(2608),
    [anon_sym___unused] = ACTIONS(2608),
    [anon_sym___weak] = ACTIONS(2608),
    [anon_sym_alignas] = ACTIONS(2608),
    [anon_sym__Alignas] = ACTIONS(2608),
    [sym_primitive_type] = ACTIONS(2608),
    [anon_sym_enum] = ACTIONS(2608),
    [anon_sym_struct] = ACTIONS(2608),
    [anon_sym_union] = ACTIONS(2608),
    [anon_sym_if] = ACTIONS(2608),
    [anon_sym_switch] = ACTIONS(2608),
    [anon_sym_case] = ACTIONS(2608),
    [anon_sym_default] = ACTIONS(2608),
    [anon_sym_while] = ACTIONS(2608),
    [anon_sym_do] = ACTIONS(2608),
    [anon_sym_for] = ACTIONS(2608),
    [anon_sym_in] = ACTIONS(2608),
    [anon_sym_return] = ACTIONS(2608),
    [anon_sym_break] = ACTIONS(2608),
    [anon_sym_continue] = ACTIONS(2608),
    [anon_sym_goto] = ACTIONS(2608),
    [anon_sym_DASH_DASH] = ACTIONS(2610),
    [anon_sym_PLUS_PLUS] = ACTIONS(2610),
    [anon_sym_sizeof] = ACTIONS(2608),
    [anon_sym___alignof__] = ACTIONS(2608),
    [anon_sym___alignof] = ACTIONS(2608),
    [anon_sym__alignof] = ACTIONS(2608),
    [anon_sym_alignof] = ACTIONS(2608),
    [anon_sym__Alignof] = ACTIONS(2608),
    [anon_sym_offsetof] = ACTIONS(2608),
    [anon_sym__Generic] = ACTIONS(2608),
    [anon_sym_asm] = ACTIONS(2608),
    [anon_sym___asm__] = ACTIONS(2608),
    [anon_sym___asm] = ACTIONS(2608),
    [sym_number_literal] = ACTIONS(2610),
    [anon_sym_L_SQUOTE] = ACTIONS(2610),
    [anon_sym_u_SQUOTE] = ACTIONS(2610),
    [anon_sym_U_SQUOTE] = ACTIONS(2610),
    [anon_sym_u8_SQUOTE] = ACTIONS(2610),
    [anon_sym_SQUOTE] = ACTIONS(2610),
    [anon_sym_AT] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [anon_sym_L_DQUOTE] = ACTIONS(2610),
    [anon_sym_u_DQUOTE] = ACTIONS(2610),
    [anon_sym_U_DQUOTE] = ACTIONS(2610),
    [anon_sym_u8_DQUOTE] = ACTIONS(2610),
    [sym_true] = ACTIONS(2608),
    [sym_false] = ACTIONS(2608),
    [anon_sym_NULL] = ACTIONS(2608),
    [anon_sym_nullptr] = ACTIONS(2608),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2608),
    [anon_sym___typeof] = ACTIONS(2608),
    [anon_sym_typeof] = ACTIONS(2608),
    [anon_sym_ATimport] = ACTIONS(2610),
    [aux_sym_preproc_undef_token1] = ACTIONS(2608),
    [anon_sym_POUND] = ACTIONS(2608),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2608),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2608),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2608),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_API_AVAILABLE] = ACTIONS(2608),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_API_DEPRECATED] = ACTIONS(2608),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym___deprecated_msg] = ACTIONS(2608),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2608),
    [anon_sym_ATprotocol] = ACTIONS(2610),
    [anon_sym_ATinterface] = ACTIONS(2610),
    [anon_sym_ATimplementation] = ACTIONS(2610),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2610),
    [anon_sym_ATsynthesize] = ACTIONS(2610),
    [anon_sym_ATdynamic] = ACTIONS(2610),
    [anon_sym_ATtry] = ACTIONS(2610),
    [anon_sym___try] = ACTIONS(2608),
    [anon_sym_ATthrow] = ACTIONS(2610),
    [anon_sym_ATselector] = ACTIONS(2610),
    [anon_sym_ATavailable] = ACTIONS(2610),
    [anon_sym___builtin_available] = ACTIONS(2608),
    [anon_sym_va_arg] = ACTIONS(2608),
    [anon_sym_ATencode] = ACTIONS(2610),
    [anon_sym_ATsynchronized] = ACTIONS(2610),
    [anon_sym_BOOL] = ACTIONS(2608),
    [anon_sym_IMP] = ACTIONS(2608),
    [anon_sym_SEL] = ACTIONS(2608),
    [anon_sym_Class] = ACTIONS(2608),
    [anon_sym_id] = ACTIONS(2608),
  },
  [STATE(519)] = {
    [sym_identifier] = ACTIONS(2618),
    [aux_sym_preproc_include_token1] = ACTIONS(2618),
    [aux_sym_preproc_include_token2] = ACTIONS(2618),
    [aux_sym_preproc_def_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token2] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2618),
    [aux_sym_preproc_else_token1] = ACTIONS(2618),
    [aux_sym_preproc_elif_token1] = ACTIONS(2618),
    [sym_preproc_directive] = ACTIONS(2618),
    [anon_sym_LPAREN2] = ACTIONS(2620),
    [anon_sym_BANG] = ACTIONS(2620),
    [anon_sym_TILDE] = ACTIONS(2620),
    [anon_sym_DASH] = ACTIONS(2618),
    [anon_sym_PLUS] = ACTIONS(2618),
    [anon_sym_STAR] = ACTIONS(2620),
    [anon_sym_CARET] = ACTIONS(2620),
    [anon_sym_AMP] = ACTIONS(2620),
    [anon_sym_SEMI] = ACTIONS(2620),
    [anon_sym___extension__] = ACTIONS(2618),
    [anon_sym_typedef] = ACTIONS(2618),
    [anon_sym_extern] = ACTIONS(2618),
    [anon_sym___attribute__] = ACTIONS(2618),
    [anon_sym___attribute] = ACTIONS(2618),
    [anon_sym_noreturn] = ACTIONS(2618),
    [anon_sym_LBRACK] = ACTIONS(2620),
    [anon_sym___declspec] = ACTIONS(2618),
    [anon_sym___cdecl] = ACTIONS(2618),
    [anon_sym___clrcall] = ACTIONS(2618),
    [anon_sym___stdcall] = ACTIONS(2618),
    [anon_sym___fastcall] = ACTIONS(2618),
    [anon_sym___thiscall] = ACTIONS(2618),
    [anon_sym___vectorcall] = ACTIONS(2618),
    [anon_sym_LBRACE] = ACTIONS(2620),
    [anon_sym_signed] = ACTIONS(2618),
    [anon_sym_unsigned] = ACTIONS(2618),
    [anon_sym_long] = ACTIONS(2618),
    [anon_sym_short] = ACTIONS(2618),
    [anon_sym_static] = ACTIONS(2618),
    [anon_sym_ATautoreleasepool] = ACTIONS(2620),
    [anon_sym_auto] = ACTIONS(2618),
    [anon_sym_register] = ACTIONS(2618),
    [anon_sym_inline] = ACTIONS(2618),
    [anon_sym___inline] = ACTIONS(2618),
    [anon_sym___inline__] = ACTIONS(2618),
    [anon_sym___forceinline] = ACTIONS(2618),
    [anon_sym_thread_local] = ACTIONS(2618),
    [anon_sym___thread] = ACTIONS(2618),
    [anon_sym_CG_EXTERN] = ACTIONS(2618),
    [anon_sym_CG_INLINE] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2618),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2618),
    [anon_sym_IBOutlet] = ACTIONS(2618),
    [anon_sym_IBInspectable] = ACTIONS(2618),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2618),
    [anon_sym_NS_INLINE] = ACTIONS(2618),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2618),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2618),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2618),
    [anon_sym_const] = ACTIONS(2618),
    [anon_sym_constexpr] = ACTIONS(2618),
    [anon_sym_volatile] = ACTIONS(2618),
    [anon_sym_restrict] = ACTIONS(2618),
    [anon_sym___restrict__] = ACTIONS(2618),
    [anon_sym__Atomic] = ACTIONS(2618),
    [anon_sym__Noreturn] = ACTIONS(2618),
    [anon_sym__Nonnull] = ACTIONS(2618),
    [anon_sym_nullable] = ACTIONS(2618),
    [anon_sym__Complex] = ACTIONS(2618),
    [anon_sym__Nullable] = ACTIONS(2618),
    [anon_sym__Nullable_result] = ACTIONS(2618),
    [anon_sym__Null_unspecified] = ACTIONS(2618),
    [anon_sym___autoreleasing] = ACTIONS(2618),
    [anon_sym___block] = ACTIONS(2618),
    [anon_sym___bridge] = ACTIONS(2618),
    [anon_sym___bridge_retained] = ACTIONS(2618),
    [anon_sym___bridge_transfer] = ACTIONS(2618),
    [anon_sym___complex] = ACTIONS(2618),
    [anon_sym___const] = ACTIONS(2618),
    [anon_sym___imag] = ACTIONS(2618),
    [anon_sym___kindof] = ACTIONS(2618),
    [anon_sym___nonnull] = ACTIONS(2618),
    [anon_sym___nullable] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2618),
    [anon_sym___real] = ACTIONS(2618),
    [anon_sym___strong] = ACTIONS(2618),
    [anon_sym___unsafe_unretained] = ACTIONS(2618),
    [anon_sym___unused] = ACTIONS(2618),
    [anon_sym___weak] = ACTIONS(2618),
    [anon_sym_alignas] = ACTIONS(2618),
    [anon_sym__Alignas] = ACTIONS(2618),
    [sym_primitive_type] = ACTIONS(2618),
    [anon_sym_enum] = ACTIONS(2618),
    [anon_sym_struct] = ACTIONS(2618),
    [anon_sym_union] = ACTIONS(2618),
    [anon_sym_if] = ACTIONS(2618),
    [anon_sym_switch] = ACTIONS(2618),
    [anon_sym_case] = ACTIONS(2618),
    [anon_sym_default] = ACTIONS(2618),
    [anon_sym_while] = ACTIONS(2618),
    [anon_sym_do] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2618),
    [anon_sym_in] = ACTIONS(2618),
    [anon_sym_return] = ACTIONS(2618),
    [anon_sym_break] = ACTIONS(2618),
    [anon_sym_continue] = ACTIONS(2618),
    [anon_sym_goto] = ACTIONS(2618),
    [anon_sym_DASH_DASH] = ACTIONS(2620),
    [anon_sym_PLUS_PLUS] = ACTIONS(2620),
    [anon_sym_sizeof] = ACTIONS(2618),
    [anon_sym___alignof__] = ACTIONS(2618),
    [anon_sym___alignof] = ACTIONS(2618),
    [anon_sym__alignof] = ACTIONS(2618),
    [anon_sym_alignof] = ACTIONS(2618),
    [anon_sym__Alignof] = ACTIONS(2618),
    [anon_sym_offsetof] = ACTIONS(2618),
    [anon_sym__Generic] = ACTIONS(2618),
    [anon_sym_asm] = ACTIONS(2618),
    [anon_sym___asm__] = ACTIONS(2618),
    [anon_sym___asm] = ACTIONS(2618),
    [sym_number_literal] = ACTIONS(2620),
    [anon_sym_L_SQUOTE] = ACTIONS(2620),
    [anon_sym_u_SQUOTE] = ACTIONS(2620),
    [anon_sym_U_SQUOTE] = ACTIONS(2620),
    [anon_sym_u8_SQUOTE] = ACTIONS(2620),
    [anon_sym_SQUOTE] = ACTIONS(2620),
    [anon_sym_AT] = ACTIONS(2618),
    [anon_sym_DQUOTE] = ACTIONS(2620),
    [anon_sym_L_DQUOTE] = ACTIONS(2620),
    [anon_sym_u_DQUOTE] = ACTIONS(2620),
    [anon_sym_U_DQUOTE] = ACTIONS(2620),
    [anon_sym_u8_DQUOTE] = ACTIONS(2620),
    [sym_true] = ACTIONS(2618),
    [sym_false] = ACTIONS(2618),
    [anon_sym_NULL] = ACTIONS(2618),
    [anon_sym_nullptr] = ACTIONS(2618),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2618),
    [anon_sym___typeof] = ACTIONS(2618),
    [anon_sym_typeof] = ACTIONS(2618),
    [anon_sym_ATimport] = ACTIONS(2620),
    [aux_sym_preproc_undef_token1] = ACTIONS(2618),
    [anon_sym_POUND] = ACTIONS(2618),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2618),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2618),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2618),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_API_AVAILABLE] = ACTIONS(2618),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_API_DEPRECATED] = ACTIONS(2618),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym___deprecated_msg] = ACTIONS(2618),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2618),
    [anon_sym_ATprotocol] = ACTIONS(2620),
    [anon_sym_ATinterface] = ACTIONS(2620),
    [anon_sym_ATimplementation] = ACTIONS(2620),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2620),
    [anon_sym_ATsynthesize] = ACTIONS(2620),
    [anon_sym_ATdynamic] = ACTIONS(2620),
    [anon_sym_ATtry] = ACTIONS(2620),
    [anon_sym___try] = ACTIONS(2618),
    [anon_sym_ATthrow] = ACTIONS(2620),
    [anon_sym_ATselector] = ACTIONS(2620),
    [anon_sym_ATavailable] = ACTIONS(2620),
    [anon_sym___builtin_available] = ACTIONS(2618),
    [anon_sym_va_arg] = ACTIONS(2618),
    [anon_sym_ATencode] = ACTIONS(2620),
    [anon_sym_ATsynchronized] = ACTIONS(2620),
    [anon_sym_BOOL] = ACTIONS(2618),
    [anon_sym_IMP] = ACTIONS(2618),
    [anon_sym_SEL] = ACTIONS(2618),
    [anon_sym_Class] = ACTIONS(2618),
    [anon_sym_id] = ACTIONS(2618),
  },
  [STATE(520)] = {
    [sym_identifier] = ACTIONS(2704),
    [aux_sym_preproc_include_token1] = ACTIONS(2704),
    [aux_sym_preproc_include_token2] = ACTIONS(2704),
    [aux_sym_preproc_def_token1] = ACTIONS(2704),
    [aux_sym_preproc_if_token1] = ACTIONS(2704),
    [aux_sym_preproc_if_token2] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2704),
    [aux_sym_preproc_else_token1] = ACTIONS(2704),
    [aux_sym_preproc_elif_token1] = ACTIONS(2704),
    [sym_preproc_directive] = ACTIONS(2704),
    [anon_sym_LPAREN2] = ACTIONS(2706),
    [anon_sym_BANG] = ACTIONS(2706),
    [anon_sym_TILDE] = ACTIONS(2706),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_STAR] = ACTIONS(2706),
    [anon_sym_CARET] = ACTIONS(2706),
    [anon_sym_AMP] = ACTIONS(2706),
    [anon_sym_SEMI] = ACTIONS(2706),
    [anon_sym___extension__] = ACTIONS(2704),
    [anon_sym_typedef] = ACTIONS(2704),
    [anon_sym_extern] = ACTIONS(2704),
    [anon_sym___attribute__] = ACTIONS(2704),
    [anon_sym___attribute] = ACTIONS(2704),
    [anon_sym_noreturn] = ACTIONS(2704),
    [anon_sym_LBRACK] = ACTIONS(2706),
    [anon_sym___declspec] = ACTIONS(2704),
    [anon_sym___cdecl] = ACTIONS(2704),
    [anon_sym___clrcall] = ACTIONS(2704),
    [anon_sym___stdcall] = ACTIONS(2704),
    [anon_sym___fastcall] = ACTIONS(2704),
    [anon_sym___thiscall] = ACTIONS(2704),
    [anon_sym___vectorcall] = ACTIONS(2704),
    [anon_sym_LBRACE] = ACTIONS(2706),
    [anon_sym_signed] = ACTIONS(2704),
    [anon_sym_unsigned] = ACTIONS(2704),
    [anon_sym_long] = ACTIONS(2704),
    [anon_sym_short] = ACTIONS(2704),
    [anon_sym_static] = ACTIONS(2704),
    [anon_sym_ATautoreleasepool] = ACTIONS(2706),
    [anon_sym_auto] = ACTIONS(2704),
    [anon_sym_register] = ACTIONS(2704),
    [anon_sym_inline] = ACTIONS(2704),
    [anon_sym___inline] = ACTIONS(2704),
    [anon_sym___inline__] = ACTIONS(2704),
    [anon_sym___forceinline] = ACTIONS(2704),
    [anon_sym_thread_local] = ACTIONS(2704),
    [anon_sym___thread] = ACTIONS(2704),
    [anon_sym_CG_EXTERN] = ACTIONS(2704),
    [anon_sym_CG_INLINE] = ACTIONS(2704),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2704),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2704),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2704),
    [anon_sym_IBOutlet] = ACTIONS(2704),
    [anon_sym_IBInspectable] = ACTIONS(2704),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2704),
    [anon_sym_NS_INLINE] = ACTIONS(2704),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2704),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2704),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2704),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2704),
    [anon_sym_const] = ACTIONS(2704),
    [anon_sym_constexpr] = ACTIONS(2704),
    [anon_sym_volatile] = ACTIONS(2704),
    [anon_sym_restrict] = ACTIONS(2704),
    [anon_sym___restrict__] = ACTIONS(2704),
    [anon_sym__Atomic] = ACTIONS(2704),
    [anon_sym__Noreturn] = ACTIONS(2704),
    [anon_sym__Nonnull] = ACTIONS(2704),
    [anon_sym_nullable] = ACTIONS(2704),
    [anon_sym__Complex] = ACTIONS(2704),
    [anon_sym__Nullable] = ACTIONS(2704),
    [anon_sym__Nullable_result] = ACTIONS(2704),
    [anon_sym__Null_unspecified] = ACTIONS(2704),
    [anon_sym___autoreleasing] = ACTIONS(2704),
    [anon_sym___block] = ACTIONS(2704),
    [anon_sym___bridge] = ACTIONS(2704),
    [anon_sym___bridge_retained] = ACTIONS(2704),
    [anon_sym___bridge_transfer] = ACTIONS(2704),
    [anon_sym___complex] = ACTIONS(2704),
    [anon_sym___const] = ACTIONS(2704),
    [anon_sym___imag] = ACTIONS(2704),
    [anon_sym___kindof] = ACTIONS(2704),
    [anon_sym___nonnull] = ACTIONS(2704),
    [anon_sym___nullable] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2704),
    [anon_sym___real] = ACTIONS(2704),
    [anon_sym___strong] = ACTIONS(2704),
    [anon_sym___unsafe_unretained] = ACTIONS(2704),
    [anon_sym___unused] = ACTIONS(2704),
    [anon_sym___weak] = ACTIONS(2704),
    [anon_sym_alignas] = ACTIONS(2704),
    [anon_sym__Alignas] = ACTIONS(2704),
    [sym_primitive_type] = ACTIONS(2704),
    [anon_sym_enum] = ACTIONS(2704),
    [anon_sym_struct] = ACTIONS(2704),
    [anon_sym_union] = ACTIONS(2704),
    [anon_sym_if] = ACTIONS(2704),
    [anon_sym_switch] = ACTIONS(2704),
    [anon_sym_case] = ACTIONS(2704),
    [anon_sym_default] = ACTIONS(2704),
    [anon_sym_while] = ACTIONS(2704),
    [anon_sym_do] = ACTIONS(2704),
    [anon_sym_for] = ACTIONS(2704),
    [anon_sym_in] = ACTIONS(2704),
    [anon_sym_return] = ACTIONS(2704),
    [anon_sym_break] = ACTIONS(2704),
    [anon_sym_continue] = ACTIONS(2704),
    [anon_sym_goto] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2706),
    [anon_sym_PLUS_PLUS] = ACTIONS(2706),
    [anon_sym_sizeof] = ACTIONS(2704),
    [anon_sym___alignof__] = ACTIONS(2704),
    [anon_sym___alignof] = ACTIONS(2704),
    [anon_sym__alignof] = ACTIONS(2704),
    [anon_sym_alignof] = ACTIONS(2704),
    [anon_sym__Alignof] = ACTIONS(2704),
    [anon_sym_offsetof] = ACTIONS(2704),
    [anon_sym__Generic] = ACTIONS(2704),
    [anon_sym_asm] = ACTIONS(2704),
    [anon_sym___asm__] = ACTIONS(2704),
    [anon_sym___asm] = ACTIONS(2704),
    [sym_number_literal] = ACTIONS(2706),
    [anon_sym_L_SQUOTE] = ACTIONS(2706),
    [anon_sym_u_SQUOTE] = ACTIONS(2706),
    [anon_sym_U_SQUOTE] = ACTIONS(2706),
    [anon_sym_u8_SQUOTE] = ACTIONS(2706),
    [anon_sym_SQUOTE] = ACTIONS(2706),
    [anon_sym_AT] = ACTIONS(2704),
    [anon_sym_DQUOTE] = ACTIONS(2706),
    [anon_sym_L_DQUOTE] = ACTIONS(2706),
    [anon_sym_u_DQUOTE] = ACTIONS(2706),
    [anon_sym_U_DQUOTE] = ACTIONS(2706),
    [anon_sym_u8_DQUOTE] = ACTIONS(2706),
    [sym_true] = ACTIONS(2704),
    [sym_false] = ACTIONS(2704),
    [anon_sym_NULL] = ACTIONS(2704),
    [anon_sym_nullptr] = ACTIONS(2704),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2704),
    [anon_sym___typeof] = ACTIONS(2704),
    [anon_sym_typeof] = ACTIONS(2704),
    [anon_sym_ATimport] = ACTIONS(2706),
    [aux_sym_preproc_undef_token1] = ACTIONS(2704),
    [anon_sym_POUND] = ACTIONS(2704),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2704),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2704),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2704),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2704),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2704),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2704),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2704),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_API_AVAILABLE] = ACTIONS(2704),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_API_DEPRECATED] = ACTIONS(2704),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2704),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym___deprecated_msg] = ACTIONS(2704),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2704),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2704),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2704),
    [anon_sym_ATprotocol] = ACTIONS(2706),
    [anon_sym_ATinterface] = ACTIONS(2706),
    [anon_sym_ATimplementation] = ACTIONS(2706),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2706),
    [anon_sym_ATsynthesize] = ACTIONS(2706),
    [anon_sym_ATdynamic] = ACTIONS(2706),
    [anon_sym_ATtry] = ACTIONS(2706),
    [anon_sym___try] = ACTIONS(2704),
    [anon_sym_ATthrow] = ACTIONS(2706),
    [anon_sym_ATselector] = ACTIONS(2706),
    [anon_sym_ATavailable] = ACTIONS(2706),
    [anon_sym___builtin_available] = ACTIONS(2704),
    [anon_sym_va_arg] = ACTIONS(2704),
    [anon_sym_ATencode] = ACTIONS(2706),
    [anon_sym_ATsynchronized] = ACTIONS(2706),
    [anon_sym_BOOL] = ACTIONS(2704),
    [anon_sym_IMP] = ACTIONS(2704),
    [anon_sym_SEL] = ACTIONS(2704),
    [anon_sym_Class] = ACTIONS(2704),
    [anon_sym_id] = ACTIONS(2704),
  },
  [STATE(521)] = {
    [sym_identifier] = ACTIONS(2780),
    [aux_sym_preproc_include_token1] = ACTIONS(2780),
    [aux_sym_preproc_include_token2] = ACTIONS(2780),
    [aux_sym_preproc_def_token1] = ACTIONS(2780),
    [aux_sym_preproc_if_token1] = ACTIONS(2780),
    [aux_sym_preproc_if_token2] = ACTIONS(2780),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2780),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2780),
    [aux_sym_preproc_else_token1] = ACTIONS(2780),
    [aux_sym_preproc_elif_token1] = ACTIONS(2780),
    [sym_preproc_directive] = ACTIONS(2780),
    [anon_sym_LPAREN2] = ACTIONS(2782),
    [anon_sym_BANG] = ACTIONS(2782),
    [anon_sym_TILDE] = ACTIONS(2782),
    [anon_sym_DASH] = ACTIONS(2780),
    [anon_sym_PLUS] = ACTIONS(2780),
    [anon_sym_STAR] = ACTIONS(2782),
    [anon_sym_CARET] = ACTIONS(2782),
    [anon_sym_AMP] = ACTIONS(2782),
    [anon_sym_SEMI] = ACTIONS(2782),
    [anon_sym___extension__] = ACTIONS(2780),
    [anon_sym_typedef] = ACTIONS(2780),
    [anon_sym_extern] = ACTIONS(2780),
    [anon_sym___attribute__] = ACTIONS(2780),
    [anon_sym___attribute] = ACTIONS(2780),
    [anon_sym_noreturn] = ACTIONS(2780),
    [anon_sym_LBRACK] = ACTIONS(2782),
    [anon_sym___declspec] = ACTIONS(2780),
    [anon_sym___cdecl] = ACTIONS(2780),
    [anon_sym___clrcall] = ACTIONS(2780),
    [anon_sym___stdcall] = ACTIONS(2780),
    [anon_sym___fastcall] = ACTIONS(2780),
    [anon_sym___thiscall] = ACTIONS(2780),
    [anon_sym___vectorcall] = ACTIONS(2780),
    [anon_sym_LBRACE] = ACTIONS(2782),
    [anon_sym_signed] = ACTIONS(2780),
    [anon_sym_unsigned] = ACTIONS(2780),
    [anon_sym_long] = ACTIONS(2780),
    [anon_sym_short] = ACTIONS(2780),
    [anon_sym_static] = ACTIONS(2780),
    [anon_sym_ATautoreleasepool] = ACTIONS(2782),
    [anon_sym_auto] = ACTIONS(2780),
    [anon_sym_register] = ACTIONS(2780),
    [anon_sym_inline] = ACTIONS(2780),
    [anon_sym___inline] = ACTIONS(2780),
    [anon_sym___inline__] = ACTIONS(2780),
    [anon_sym___forceinline] = ACTIONS(2780),
    [anon_sym_thread_local] = ACTIONS(2780),
    [anon_sym___thread] = ACTIONS(2780),
    [anon_sym_CG_EXTERN] = ACTIONS(2780),
    [anon_sym_CG_INLINE] = ACTIONS(2780),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2780),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2780),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2780),
    [anon_sym_IBOutlet] = ACTIONS(2780),
    [anon_sym_IBInspectable] = ACTIONS(2780),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2780),
    [anon_sym_NS_INLINE] = ACTIONS(2780),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2780),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2780),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2780),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2780),
    [anon_sym_const] = ACTIONS(2780),
    [anon_sym_constexpr] = ACTIONS(2780),
    [anon_sym_volatile] = ACTIONS(2780),
    [anon_sym_restrict] = ACTIONS(2780),
    [anon_sym___restrict__] = ACTIONS(2780),
    [anon_sym__Atomic] = ACTIONS(2780),
    [anon_sym__Noreturn] = ACTIONS(2780),
    [anon_sym__Nonnull] = ACTIONS(2780),
    [anon_sym_nullable] = ACTIONS(2780),
    [anon_sym__Complex] = ACTIONS(2780),
    [anon_sym__Nullable] = ACTIONS(2780),
    [anon_sym__Nullable_result] = ACTIONS(2780),
    [anon_sym__Null_unspecified] = ACTIONS(2780),
    [anon_sym___autoreleasing] = ACTIONS(2780),
    [anon_sym___block] = ACTIONS(2780),
    [anon_sym___bridge] = ACTIONS(2780),
    [anon_sym___bridge_retained] = ACTIONS(2780),
    [anon_sym___bridge_transfer] = ACTIONS(2780),
    [anon_sym___complex] = ACTIONS(2780),
    [anon_sym___const] = ACTIONS(2780),
    [anon_sym___imag] = ACTIONS(2780),
    [anon_sym___kindof] = ACTIONS(2780),
    [anon_sym___nonnull] = ACTIONS(2780),
    [anon_sym___nullable] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2780),
    [anon_sym___real] = ACTIONS(2780),
    [anon_sym___strong] = ACTIONS(2780),
    [anon_sym___unsafe_unretained] = ACTIONS(2780),
    [anon_sym___unused] = ACTIONS(2780),
    [anon_sym___weak] = ACTIONS(2780),
    [anon_sym_alignas] = ACTIONS(2780),
    [anon_sym__Alignas] = ACTIONS(2780),
    [sym_primitive_type] = ACTIONS(2780),
    [anon_sym_enum] = ACTIONS(2780),
    [anon_sym_struct] = ACTIONS(2780),
    [anon_sym_union] = ACTIONS(2780),
    [anon_sym_if] = ACTIONS(2780),
    [anon_sym_switch] = ACTIONS(2780),
    [anon_sym_case] = ACTIONS(2780),
    [anon_sym_default] = ACTIONS(2780),
    [anon_sym_while] = ACTIONS(2780),
    [anon_sym_do] = ACTIONS(2780),
    [anon_sym_for] = ACTIONS(2780),
    [anon_sym_in] = ACTIONS(2780),
    [anon_sym_return] = ACTIONS(2780),
    [anon_sym_break] = ACTIONS(2780),
    [anon_sym_continue] = ACTIONS(2780),
    [anon_sym_goto] = ACTIONS(2780),
    [anon_sym_DASH_DASH] = ACTIONS(2782),
    [anon_sym_PLUS_PLUS] = ACTIONS(2782),
    [anon_sym_sizeof] = ACTIONS(2780),
    [anon_sym___alignof__] = ACTIONS(2780),
    [anon_sym___alignof] = ACTIONS(2780),
    [anon_sym__alignof] = ACTIONS(2780),
    [anon_sym_alignof] = ACTIONS(2780),
    [anon_sym__Alignof] = ACTIONS(2780),
    [anon_sym_offsetof] = ACTIONS(2780),
    [anon_sym__Generic] = ACTIONS(2780),
    [anon_sym_asm] = ACTIONS(2780),
    [anon_sym___asm__] = ACTIONS(2780),
    [anon_sym___asm] = ACTIONS(2780),
    [sym_number_literal] = ACTIONS(2782),
    [anon_sym_L_SQUOTE] = ACTIONS(2782),
    [anon_sym_u_SQUOTE] = ACTIONS(2782),
    [anon_sym_U_SQUOTE] = ACTIONS(2782),
    [anon_sym_u8_SQUOTE] = ACTIONS(2782),
    [anon_sym_SQUOTE] = ACTIONS(2782),
    [anon_sym_AT] = ACTIONS(2780),
    [anon_sym_DQUOTE] = ACTIONS(2782),
    [anon_sym_L_DQUOTE] = ACTIONS(2782),
    [anon_sym_u_DQUOTE] = ACTIONS(2782),
    [anon_sym_U_DQUOTE] = ACTIONS(2782),
    [anon_sym_u8_DQUOTE] = ACTIONS(2782),
    [sym_true] = ACTIONS(2780),
    [sym_false] = ACTIONS(2780),
    [anon_sym_NULL] = ACTIONS(2780),
    [anon_sym_nullptr] = ACTIONS(2780),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2780),
    [anon_sym___typeof] = ACTIONS(2780),
    [anon_sym_typeof] = ACTIONS(2780),
    [anon_sym_ATimport] = ACTIONS(2782),
    [aux_sym_preproc_undef_token1] = ACTIONS(2780),
    [anon_sym_POUND] = ACTIONS(2780),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2780),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2780),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2780),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2780),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2780),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2780),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2780),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_API_AVAILABLE] = ACTIONS(2780),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_API_DEPRECATED] = ACTIONS(2780),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2780),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym___deprecated_msg] = ACTIONS(2780),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2780),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2780),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2780),
    [anon_sym_ATprotocol] = ACTIONS(2782),
    [anon_sym_ATinterface] = ACTIONS(2782),
    [anon_sym_ATimplementation] = ACTIONS(2782),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2782),
    [anon_sym_ATsynthesize] = ACTIONS(2782),
    [anon_sym_ATdynamic] = ACTIONS(2782),
    [anon_sym_ATtry] = ACTIONS(2782),
    [anon_sym___try] = ACTIONS(2780),
    [anon_sym_ATthrow] = ACTIONS(2782),
    [anon_sym_ATselector] = ACTIONS(2782),
    [anon_sym_ATavailable] = ACTIONS(2782),
    [anon_sym___builtin_available] = ACTIONS(2780),
    [anon_sym_va_arg] = ACTIONS(2780),
    [anon_sym_ATencode] = ACTIONS(2782),
    [anon_sym_ATsynchronized] = ACTIONS(2782),
    [anon_sym_BOOL] = ACTIONS(2780),
    [anon_sym_IMP] = ACTIONS(2780),
    [anon_sym_SEL] = ACTIONS(2780),
    [anon_sym_Class] = ACTIONS(2780),
    [anon_sym_id] = ACTIONS(2780),
  },
  [STATE(522)] = {
    [sym_identifier] = ACTIONS(2784),
    [aux_sym_preproc_include_token1] = ACTIONS(2784),
    [aux_sym_preproc_include_token2] = ACTIONS(2784),
    [aux_sym_preproc_def_token1] = ACTIONS(2784),
    [aux_sym_preproc_if_token1] = ACTIONS(2784),
    [aux_sym_preproc_if_token2] = ACTIONS(2784),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2784),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2784),
    [aux_sym_preproc_else_token1] = ACTIONS(2784),
    [aux_sym_preproc_elif_token1] = ACTIONS(2784),
    [sym_preproc_directive] = ACTIONS(2784),
    [anon_sym_LPAREN2] = ACTIONS(2786),
    [anon_sym_BANG] = ACTIONS(2786),
    [anon_sym_TILDE] = ACTIONS(2786),
    [anon_sym_DASH] = ACTIONS(2784),
    [anon_sym_PLUS] = ACTIONS(2784),
    [anon_sym_STAR] = ACTIONS(2786),
    [anon_sym_CARET] = ACTIONS(2786),
    [anon_sym_AMP] = ACTIONS(2786),
    [anon_sym_SEMI] = ACTIONS(2786),
    [anon_sym___extension__] = ACTIONS(2784),
    [anon_sym_typedef] = ACTIONS(2784),
    [anon_sym_extern] = ACTIONS(2784),
    [anon_sym___attribute__] = ACTIONS(2784),
    [anon_sym___attribute] = ACTIONS(2784),
    [anon_sym_noreturn] = ACTIONS(2784),
    [anon_sym_LBRACK] = ACTIONS(2786),
    [anon_sym___declspec] = ACTIONS(2784),
    [anon_sym___cdecl] = ACTIONS(2784),
    [anon_sym___clrcall] = ACTIONS(2784),
    [anon_sym___stdcall] = ACTIONS(2784),
    [anon_sym___fastcall] = ACTIONS(2784),
    [anon_sym___thiscall] = ACTIONS(2784),
    [anon_sym___vectorcall] = ACTIONS(2784),
    [anon_sym_LBRACE] = ACTIONS(2786),
    [anon_sym_signed] = ACTIONS(2784),
    [anon_sym_unsigned] = ACTIONS(2784),
    [anon_sym_long] = ACTIONS(2784),
    [anon_sym_short] = ACTIONS(2784),
    [anon_sym_static] = ACTIONS(2784),
    [anon_sym_ATautoreleasepool] = ACTIONS(2786),
    [anon_sym_auto] = ACTIONS(2784),
    [anon_sym_register] = ACTIONS(2784),
    [anon_sym_inline] = ACTIONS(2784),
    [anon_sym___inline] = ACTIONS(2784),
    [anon_sym___inline__] = ACTIONS(2784),
    [anon_sym___forceinline] = ACTIONS(2784),
    [anon_sym_thread_local] = ACTIONS(2784),
    [anon_sym___thread] = ACTIONS(2784),
    [anon_sym_CG_EXTERN] = ACTIONS(2784),
    [anon_sym_CG_INLINE] = ACTIONS(2784),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2784),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2784),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2784),
    [anon_sym_IBOutlet] = ACTIONS(2784),
    [anon_sym_IBInspectable] = ACTIONS(2784),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2784),
    [anon_sym_NS_INLINE] = ACTIONS(2784),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2784),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2784),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2784),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2784),
    [anon_sym_const] = ACTIONS(2784),
    [anon_sym_constexpr] = ACTIONS(2784),
    [anon_sym_volatile] = ACTIONS(2784),
    [anon_sym_restrict] = ACTIONS(2784),
    [anon_sym___restrict__] = ACTIONS(2784),
    [anon_sym__Atomic] = ACTIONS(2784),
    [anon_sym__Noreturn] = ACTIONS(2784),
    [anon_sym__Nonnull] = ACTIONS(2784),
    [anon_sym_nullable] = ACTIONS(2784),
    [anon_sym__Complex] = ACTIONS(2784),
    [anon_sym__Nullable] = ACTIONS(2784),
    [anon_sym__Nullable_result] = ACTIONS(2784),
    [anon_sym__Null_unspecified] = ACTIONS(2784),
    [anon_sym___autoreleasing] = ACTIONS(2784),
    [anon_sym___block] = ACTIONS(2784),
    [anon_sym___bridge] = ACTIONS(2784),
    [anon_sym___bridge_retained] = ACTIONS(2784),
    [anon_sym___bridge_transfer] = ACTIONS(2784),
    [anon_sym___complex] = ACTIONS(2784),
    [anon_sym___const] = ACTIONS(2784),
    [anon_sym___imag] = ACTIONS(2784),
    [anon_sym___kindof] = ACTIONS(2784),
    [anon_sym___nonnull] = ACTIONS(2784),
    [anon_sym___nullable] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2784),
    [anon_sym___real] = ACTIONS(2784),
    [anon_sym___strong] = ACTIONS(2784),
    [anon_sym___unsafe_unretained] = ACTIONS(2784),
    [anon_sym___unused] = ACTIONS(2784),
    [anon_sym___weak] = ACTIONS(2784),
    [anon_sym_alignas] = ACTIONS(2784),
    [anon_sym__Alignas] = ACTIONS(2784),
    [sym_primitive_type] = ACTIONS(2784),
    [anon_sym_enum] = ACTIONS(2784),
    [anon_sym_struct] = ACTIONS(2784),
    [anon_sym_union] = ACTIONS(2784),
    [anon_sym_if] = ACTIONS(2784),
    [anon_sym_switch] = ACTIONS(2784),
    [anon_sym_case] = ACTIONS(2784),
    [anon_sym_default] = ACTIONS(2784),
    [anon_sym_while] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2784),
    [anon_sym_for] = ACTIONS(2784),
    [anon_sym_in] = ACTIONS(2784),
    [anon_sym_return] = ACTIONS(2784),
    [anon_sym_break] = ACTIONS(2784),
    [anon_sym_continue] = ACTIONS(2784),
    [anon_sym_goto] = ACTIONS(2784),
    [anon_sym_DASH_DASH] = ACTIONS(2786),
    [anon_sym_PLUS_PLUS] = ACTIONS(2786),
    [anon_sym_sizeof] = ACTIONS(2784),
    [anon_sym___alignof__] = ACTIONS(2784),
    [anon_sym___alignof] = ACTIONS(2784),
    [anon_sym__alignof] = ACTIONS(2784),
    [anon_sym_alignof] = ACTIONS(2784),
    [anon_sym__Alignof] = ACTIONS(2784),
    [anon_sym_offsetof] = ACTIONS(2784),
    [anon_sym__Generic] = ACTIONS(2784),
    [anon_sym_asm] = ACTIONS(2784),
    [anon_sym___asm__] = ACTIONS(2784),
    [anon_sym___asm] = ACTIONS(2784),
    [sym_number_literal] = ACTIONS(2786),
    [anon_sym_L_SQUOTE] = ACTIONS(2786),
    [anon_sym_u_SQUOTE] = ACTIONS(2786),
    [anon_sym_U_SQUOTE] = ACTIONS(2786),
    [anon_sym_u8_SQUOTE] = ACTIONS(2786),
    [anon_sym_SQUOTE] = ACTIONS(2786),
    [anon_sym_AT] = ACTIONS(2784),
    [anon_sym_DQUOTE] = ACTIONS(2786),
    [anon_sym_L_DQUOTE] = ACTIONS(2786),
    [anon_sym_u_DQUOTE] = ACTIONS(2786),
    [anon_sym_U_DQUOTE] = ACTIONS(2786),
    [anon_sym_u8_DQUOTE] = ACTIONS(2786),
    [sym_true] = ACTIONS(2784),
    [sym_false] = ACTIONS(2784),
    [anon_sym_NULL] = ACTIONS(2784),
    [anon_sym_nullptr] = ACTIONS(2784),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2784),
    [anon_sym___typeof] = ACTIONS(2784),
    [anon_sym_typeof] = ACTIONS(2784),
    [anon_sym_ATimport] = ACTIONS(2786),
    [aux_sym_preproc_undef_token1] = ACTIONS(2784),
    [anon_sym_POUND] = ACTIONS(2784),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2784),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2784),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2784),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2784),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2784),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2784),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2784),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_API_AVAILABLE] = ACTIONS(2784),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_API_DEPRECATED] = ACTIONS(2784),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2784),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym___deprecated_msg] = ACTIONS(2784),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2784),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2784),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2784),
    [anon_sym_ATprotocol] = ACTIONS(2786),
    [anon_sym_ATinterface] = ACTIONS(2786),
    [anon_sym_ATimplementation] = ACTIONS(2786),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2786),
    [anon_sym_ATsynthesize] = ACTIONS(2786),
    [anon_sym_ATdynamic] = ACTIONS(2786),
    [anon_sym_ATtry] = ACTIONS(2786),
    [anon_sym___try] = ACTIONS(2784),
    [anon_sym_ATthrow] = ACTIONS(2786),
    [anon_sym_ATselector] = ACTIONS(2786),
    [anon_sym_ATavailable] = ACTIONS(2786),
    [anon_sym___builtin_available] = ACTIONS(2784),
    [anon_sym_va_arg] = ACTIONS(2784),
    [anon_sym_ATencode] = ACTIONS(2786),
    [anon_sym_ATsynchronized] = ACTIONS(2786),
    [anon_sym_BOOL] = ACTIONS(2784),
    [anon_sym_IMP] = ACTIONS(2784),
    [anon_sym_SEL] = ACTIONS(2784),
    [anon_sym_Class] = ACTIONS(2784),
    [anon_sym_id] = ACTIONS(2784),
  },
  [STATE(523)] = {
    [sym_identifier] = ACTIONS(2788),
    [aux_sym_preproc_include_token1] = ACTIONS(2788),
    [aux_sym_preproc_include_token2] = ACTIONS(2788),
    [aux_sym_preproc_def_token1] = ACTIONS(2788),
    [aux_sym_preproc_if_token1] = ACTIONS(2788),
    [aux_sym_preproc_if_token2] = ACTIONS(2788),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2788),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2788),
    [aux_sym_preproc_else_token1] = ACTIONS(2788),
    [aux_sym_preproc_elif_token1] = ACTIONS(2788),
    [sym_preproc_directive] = ACTIONS(2788),
    [anon_sym_LPAREN2] = ACTIONS(2790),
    [anon_sym_BANG] = ACTIONS(2790),
    [anon_sym_TILDE] = ACTIONS(2790),
    [anon_sym_DASH] = ACTIONS(2788),
    [anon_sym_PLUS] = ACTIONS(2788),
    [anon_sym_STAR] = ACTIONS(2790),
    [anon_sym_CARET] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(2790),
    [anon_sym_SEMI] = ACTIONS(2790),
    [anon_sym___extension__] = ACTIONS(2788),
    [anon_sym_typedef] = ACTIONS(2788),
    [anon_sym_extern] = ACTIONS(2788),
    [anon_sym___attribute__] = ACTIONS(2788),
    [anon_sym___attribute] = ACTIONS(2788),
    [anon_sym_noreturn] = ACTIONS(2788),
    [anon_sym_LBRACK] = ACTIONS(2790),
    [anon_sym___declspec] = ACTIONS(2788),
    [anon_sym___cdecl] = ACTIONS(2788),
    [anon_sym___clrcall] = ACTIONS(2788),
    [anon_sym___stdcall] = ACTIONS(2788),
    [anon_sym___fastcall] = ACTIONS(2788),
    [anon_sym___thiscall] = ACTIONS(2788),
    [anon_sym___vectorcall] = ACTIONS(2788),
    [anon_sym_LBRACE] = ACTIONS(2790),
    [anon_sym_signed] = ACTIONS(2788),
    [anon_sym_unsigned] = ACTIONS(2788),
    [anon_sym_long] = ACTIONS(2788),
    [anon_sym_short] = ACTIONS(2788),
    [anon_sym_static] = ACTIONS(2788),
    [anon_sym_ATautoreleasepool] = ACTIONS(2790),
    [anon_sym_auto] = ACTIONS(2788),
    [anon_sym_register] = ACTIONS(2788),
    [anon_sym_inline] = ACTIONS(2788),
    [anon_sym___inline] = ACTIONS(2788),
    [anon_sym___inline__] = ACTIONS(2788),
    [anon_sym___forceinline] = ACTIONS(2788),
    [anon_sym_thread_local] = ACTIONS(2788),
    [anon_sym___thread] = ACTIONS(2788),
    [anon_sym_CG_EXTERN] = ACTIONS(2788),
    [anon_sym_CG_INLINE] = ACTIONS(2788),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2788),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2788),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2788),
    [anon_sym_IBOutlet] = ACTIONS(2788),
    [anon_sym_IBInspectable] = ACTIONS(2788),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2788),
    [anon_sym_NS_INLINE] = ACTIONS(2788),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2788),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2788),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2788),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2788),
    [anon_sym_const] = ACTIONS(2788),
    [anon_sym_constexpr] = ACTIONS(2788),
    [anon_sym_volatile] = ACTIONS(2788),
    [anon_sym_restrict] = ACTIONS(2788),
    [anon_sym___restrict__] = ACTIONS(2788),
    [anon_sym__Atomic] = ACTIONS(2788),
    [anon_sym__Noreturn] = ACTIONS(2788),
    [anon_sym__Nonnull] = ACTIONS(2788),
    [anon_sym_nullable] = ACTIONS(2788),
    [anon_sym__Complex] = ACTIONS(2788),
    [anon_sym__Nullable] = ACTIONS(2788),
    [anon_sym__Nullable_result] = ACTIONS(2788),
    [anon_sym__Null_unspecified] = ACTIONS(2788),
    [anon_sym___autoreleasing] = ACTIONS(2788),
    [anon_sym___block] = ACTIONS(2788),
    [anon_sym___bridge] = ACTIONS(2788),
    [anon_sym___bridge_retained] = ACTIONS(2788),
    [anon_sym___bridge_transfer] = ACTIONS(2788),
    [anon_sym___complex] = ACTIONS(2788),
    [anon_sym___const] = ACTIONS(2788),
    [anon_sym___imag] = ACTIONS(2788),
    [anon_sym___kindof] = ACTIONS(2788),
    [anon_sym___nonnull] = ACTIONS(2788),
    [anon_sym___nullable] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2788),
    [anon_sym___real] = ACTIONS(2788),
    [anon_sym___strong] = ACTIONS(2788),
    [anon_sym___unsafe_unretained] = ACTIONS(2788),
    [anon_sym___unused] = ACTIONS(2788),
    [anon_sym___weak] = ACTIONS(2788),
    [anon_sym_alignas] = ACTIONS(2788),
    [anon_sym__Alignas] = ACTIONS(2788),
    [sym_primitive_type] = ACTIONS(2788),
    [anon_sym_enum] = ACTIONS(2788),
    [anon_sym_struct] = ACTIONS(2788),
    [anon_sym_union] = ACTIONS(2788),
    [anon_sym_if] = ACTIONS(2788),
    [anon_sym_switch] = ACTIONS(2788),
    [anon_sym_case] = ACTIONS(2788),
    [anon_sym_default] = ACTIONS(2788),
    [anon_sym_while] = ACTIONS(2788),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_for] = ACTIONS(2788),
    [anon_sym_in] = ACTIONS(2788),
    [anon_sym_return] = ACTIONS(2788),
    [anon_sym_break] = ACTIONS(2788),
    [anon_sym_continue] = ACTIONS(2788),
    [anon_sym_goto] = ACTIONS(2788),
    [anon_sym_DASH_DASH] = ACTIONS(2790),
    [anon_sym_PLUS_PLUS] = ACTIONS(2790),
    [anon_sym_sizeof] = ACTIONS(2788),
    [anon_sym___alignof__] = ACTIONS(2788),
    [anon_sym___alignof] = ACTIONS(2788),
    [anon_sym__alignof] = ACTIONS(2788),
    [anon_sym_alignof] = ACTIONS(2788),
    [anon_sym__Alignof] = ACTIONS(2788),
    [anon_sym_offsetof] = ACTIONS(2788),
    [anon_sym__Generic] = ACTIONS(2788),
    [anon_sym_asm] = ACTIONS(2788),
    [anon_sym___asm__] = ACTIONS(2788),
    [anon_sym___asm] = ACTIONS(2788),
    [sym_number_literal] = ACTIONS(2790),
    [anon_sym_L_SQUOTE] = ACTIONS(2790),
    [anon_sym_u_SQUOTE] = ACTIONS(2790),
    [anon_sym_U_SQUOTE] = ACTIONS(2790),
    [anon_sym_u8_SQUOTE] = ACTIONS(2790),
    [anon_sym_SQUOTE] = ACTIONS(2790),
    [anon_sym_AT] = ACTIONS(2788),
    [anon_sym_DQUOTE] = ACTIONS(2790),
    [anon_sym_L_DQUOTE] = ACTIONS(2790),
    [anon_sym_u_DQUOTE] = ACTIONS(2790),
    [anon_sym_U_DQUOTE] = ACTIONS(2790),
    [anon_sym_u8_DQUOTE] = ACTIONS(2790),
    [sym_true] = ACTIONS(2788),
    [sym_false] = ACTIONS(2788),
    [anon_sym_NULL] = ACTIONS(2788),
    [anon_sym_nullptr] = ACTIONS(2788),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2788),
    [anon_sym___typeof] = ACTIONS(2788),
    [anon_sym_typeof] = ACTIONS(2788),
    [anon_sym_ATimport] = ACTIONS(2790),
    [aux_sym_preproc_undef_token1] = ACTIONS(2788),
    [anon_sym_POUND] = ACTIONS(2788),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2788),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2788),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2788),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2788),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2788),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2788),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2788),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_API_AVAILABLE] = ACTIONS(2788),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_API_DEPRECATED] = ACTIONS(2788),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2788),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym___deprecated_msg] = ACTIONS(2788),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2788),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2788),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2788),
    [anon_sym_ATprotocol] = ACTIONS(2790),
    [anon_sym_ATinterface] = ACTIONS(2790),
    [anon_sym_ATimplementation] = ACTIONS(2790),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2790),
    [anon_sym_ATsynthesize] = ACTIONS(2790),
    [anon_sym_ATdynamic] = ACTIONS(2790),
    [anon_sym_ATtry] = ACTIONS(2790),
    [anon_sym___try] = ACTIONS(2788),
    [anon_sym_ATthrow] = ACTIONS(2790),
    [anon_sym_ATselector] = ACTIONS(2790),
    [anon_sym_ATavailable] = ACTIONS(2790),
    [anon_sym___builtin_available] = ACTIONS(2788),
    [anon_sym_va_arg] = ACTIONS(2788),
    [anon_sym_ATencode] = ACTIONS(2790),
    [anon_sym_ATsynchronized] = ACTIONS(2790),
    [anon_sym_BOOL] = ACTIONS(2788),
    [anon_sym_IMP] = ACTIONS(2788),
    [anon_sym_SEL] = ACTIONS(2788),
    [anon_sym_Class] = ACTIONS(2788),
    [anon_sym_id] = ACTIONS(2788),
  },
  [STATE(524)] = {
    [sym_identifier] = ACTIONS(2626),
    [aux_sym_preproc_include_token1] = ACTIONS(2626),
    [aux_sym_preproc_include_token2] = ACTIONS(2626),
    [aux_sym_preproc_def_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token2] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2626),
    [aux_sym_preproc_else_token1] = ACTIONS(2626),
    [aux_sym_preproc_elif_token1] = ACTIONS(2626),
    [sym_preproc_directive] = ACTIONS(2626),
    [anon_sym_LPAREN2] = ACTIONS(2628),
    [anon_sym_BANG] = ACTIONS(2628),
    [anon_sym_TILDE] = ACTIONS(2628),
    [anon_sym_DASH] = ACTIONS(2626),
    [anon_sym_PLUS] = ACTIONS(2626),
    [anon_sym_STAR] = ACTIONS(2628),
    [anon_sym_CARET] = ACTIONS(2628),
    [anon_sym_AMP] = ACTIONS(2628),
    [anon_sym_SEMI] = ACTIONS(2628),
    [anon_sym___extension__] = ACTIONS(2626),
    [anon_sym_typedef] = ACTIONS(2626),
    [anon_sym_extern] = ACTIONS(2626),
    [anon_sym___attribute__] = ACTIONS(2626),
    [anon_sym___attribute] = ACTIONS(2626),
    [anon_sym_noreturn] = ACTIONS(2626),
    [anon_sym_LBRACK] = ACTIONS(2628),
    [anon_sym___declspec] = ACTIONS(2626),
    [anon_sym___cdecl] = ACTIONS(2626),
    [anon_sym___clrcall] = ACTIONS(2626),
    [anon_sym___stdcall] = ACTIONS(2626),
    [anon_sym___fastcall] = ACTIONS(2626),
    [anon_sym___thiscall] = ACTIONS(2626),
    [anon_sym___vectorcall] = ACTIONS(2626),
    [anon_sym_LBRACE] = ACTIONS(2628),
    [anon_sym_signed] = ACTIONS(2626),
    [anon_sym_unsigned] = ACTIONS(2626),
    [anon_sym_long] = ACTIONS(2626),
    [anon_sym_short] = ACTIONS(2626),
    [anon_sym_static] = ACTIONS(2626),
    [anon_sym_ATautoreleasepool] = ACTIONS(2628),
    [anon_sym_auto] = ACTIONS(2626),
    [anon_sym_register] = ACTIONS(2626),
    [anon_sym_inline] = ACTIONS(2626),
    [anon_sym___inline] = ACTIONS(2626),
    [anon_sym___inline__] = ACTIONS(2626),
    [anon_sym___forceinline] = ACTIONS(2626),
    [anon_sym_thread_local] = ACTIONS(2626),
    [anon_sym___thread] = ACTIONS(2626),
    [anon_sym_CG_EXTERN] = ACTIONS(2626),
    [anon_sym_CG_INLINE] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2626),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2626),
    [anon_sym_IBOutlet] = ACTIONS(2626),
    [anon_sym_IBInspectable] = ACTIONS(2626),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2626),
    [anon_sym_NS_INLINE] = ACTIONS(2626),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2626),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2626),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2626),
    [anon_sym_const] = ACTIONS(2626),
    [anon_sym_constexpr] = ACTIONS(2626),
    [anon_sym_volatile] = ACTIONS(2626),
    [anon_sym_restrict] = ACTIONS(2626),
    [anon_sym___restrict__] = ACTIONS(2626),
    [anon_sym__Atomic] = ACTIONS(2626),
    [anon_sym__Noreturn] = ACTIONS(2626),
    [anon_sym__Nonnull] = ACTIONS(2626),
    [anon_sym_nullable] = ACTIONS(2626),
    [anon_sym__Complex] = ACTIONS(2626),
    [anon_sym__Nullable] = ACTIONS(2626),
    [anon_sym__Nullable_result] = ACTIONS(2626),
    [anon_sym__Null_unspecified] = ACTIONS(2626),
    [anon_sym___autoreleasing] = ACTIONS(2626),
    [anon_sym___block] = ACTIONS(2626),
    [anon_sym___bridge] = ACTIONS(2626),
    [anon_sym___bridge_retained] = ACTIONS(2626),
    [anon_sym___bridge_transfer] = ACTIONS(2626),
    [anon_sym___complex] = ACTIONS(2626),
    [anon_sym___const] = ACTIONS(2626),
    [anon_sym___imag] = ACTIONS(2626),
    [anon_sym___kindof] = ACTIONS(2626),
    [anon_sym___nonnull] = ACTIONS(2626),
    [anon_sym___nullable] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2626),
    [anon_sym___real] = ACTIONS(2626),
    [anon_sym___strong] = ACTIONS(2626),
    [anon_sym___unsafe_unretained] = ACTIONS(2626),
    [anon_sym___unused] = ACTIONS(2626),
    [anon_sym___weak] = ACTIONS(2626),
    [anon_sym_alignas] = ACTIONS(2626),
    [anon_sym__Alignas] = ACTIONS(2626),
    [sym_primitive_type] = ACTIONS(2626),
    [anon_sym_enum] = ACTIONS(2626),
    [anon_sym_struct] = ACTIONS(2626),
    [anon_sym_union] = ACTIONS(2626),
    [anon_sym_if] = ACTIONS(2626),
    [anon_sym_switch] = ACTIONS(2626),
    [anon_sym_case] = ACTIONS(2626),
    [anon_sym_default] = ACTIONS(2626),
    [anon_sym_while] = ACTIONS(2626),
    [anon_sym_do] = ACTIONS(2626),
    [anon_sym_for] = ACTIONS(2626),
    [anon_sym_in] = ACTIONS(2626),
    [anon_sym_return] = ACTIONS(2626),
    [anon_sym_break] = ACTIONS(2626),
    [anon_sym_continue] = ACTIONS(2626),
    [anon_sym_goto] = ACTIONS(2626),
    [anon_sym_DASH_DASH] = ACTIONS(2628),
    [anon_sym_PLUS_PLUS] = ACTIONS(2628),
    [anon_sym_sizeof] = ACTIONS(2626),
    [anon_sym___alignof__] = ACTIONS(2626),
    [anon_sym___alignof] = ACTIONS(2626),
    [anon_sym__alignof] = ACTIONS(2626),
    [anon_sym_alignof] = ACTIONS(2626),
    [anon_sym__Alignof] = ACTIONS(2626),
    [anon_sym_offsetof] = ACTIONS(2626),
    [anon_sym__Generic] = ACTIONS(2626),
    [anon_sym_asm] = ACTIONS(2626),
    [anon_sym___asm__] = ACTIONS(2626),
    [anon_sym___asm] = ACTIONS(2626),
    [sym_number_literal] = ACTIONS(2628),
    [anon_sym_L_SQUOTE] = ACTIONS(2628),
    [anon_sym_u_SQUOTE] = ACTIONS(2628),
    [anon_sym_U_SQUOTE] = ACTIONS(2628),
    [anon_sym_u8_SQUOTE] = ACTIONS(2628),
    [anon_sym_SQUOTE] = ACTIONS(2628),
    [anon_sym_AT] = ACTIONS(2626),
    [anon_sym_DQUOTE] = ACTIONS(2628),
    [anon_sym_L_DQUOTE] = ACTIONS(2628),
    [anon_sym_u_DQUOTE] = ACTIONS(2628),
    [anon_sym_U_DQUOTE] = ACTIONS(2628),
    [anon_sym_u8_DQUOTE] = ACTIONS(2628),
    [sym_true] = ACTIONS(2626),
    [sym_false] = ACTIONS(2626),
    [anon_sym_NULL] = ACTIONS(2626),
    [anon_sym_nullptr] = ACTIONS(2626),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2626),
    [anon_sym___typeof] = ACTIONS(2626),
    [anon_sym_typeof] = ACTIONS(2626),
    [anon_sym_ATimport] = ACTIONS(2628),
    [aux_sym_preproc_undef_token1] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2626),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2626),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2626),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2626),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_API_AVAILABLE] = ACTIONS(2626),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_API_DEPRECATED] = ACTIONS(2626),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym___deprecated_msg] = ACTIONS(2626),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2626),
    [anon_sym_ATprotocol] = ACTIONS(2628),
    [anon_sym_ATinterface] = ACTIONS(2628),
    [anon_sym_ATimplementation] = ACTIONS(2628),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2628),
    [anon_sym_ATsynthesize] = ACTIONS(2628),
    [anon_sym_ATdynamic] = ACTIONS(2628),
    [anon_sym_ATtry] = ACTIONS(2628),
    [anon_sym___try] = ACTIONS(2626),
    [anon_sym_ATthrow] = ACTIONS(2628),
    [anon_sym_ATselector] = ACTIONS(2628),
    [anon_sym_ATavailable] = ACTIONS(2628),
    [anon_sym___builtin_available] = ACTIONS(2626),
    [anon_sym_va_arg] = ACTIONS(2626),
    [anon_sym_ATencode] = ACTIONS(2628),
    [anon_sym_ATsynchronized] = ACTIONS(2628),
    [anon_sym_BOOL] = ACTIONS(2626),
    [anon_sym_IMP] = ACTIONS(2626),
    [anon_sym_SEL] = ACTIONS(2626),
    [anon_sym_Class] = ACTIONS(2626),
    [anon_sym_id] = ACTIONS(2626),
  },
  [STATE(525)] = {
    [sym_identifier] = ACTIONS(2646),
    [aux_sym_preproc_include_token1] = ACTIONS(2646),
    [aux_sym_preproc_include_token2] = ACTIONS(2646),
    [aux_sym_preproc_def_token1] = ACTIONS(2646),
    [aux_sym_preproc_if_token1] = ACTIONS(2646),
    [aux_sym_preproc_if_token2] = ACTIONS(2646),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2646),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2646),
    [aux_sym_preproc_else_token1] = ACTIONS(2646),
    [aux_sym_preproc_elif_token1] = ACTIONS(2646),
    [sym_preproc_directive] = ACTIONS(2646),
    [anon_sym_LPAREN2] = ACTIONS(2648),
    [anon_sym_BANG] = ACTIONS(2648),
    [anon_sym_TILDE] = ACTIONS(2648),
    [anon_sym_DASH] = ACTIONS(2646),
    [anon_sym_PLUS] = ACTIONS(2646),
    [anon_sym_STAR] = ACTIONS(2648),
    [anon_sym_CARET] = ACTIONS(2648),
    [anon_sym_AMP] = ACTIONS(2648),
    [anon_sym_SEMI] = ACTIONS(2648),
    [anon_sym___extension__] = ACTIONS(2646),
    [anon_sym_typedef] = ACTIONS(2646),
    [anon_sym_extern] = ACTIONS(2646),
    [anon_sym___attribute__] = ACTIONS(2646),
    [anon_sym___attribute] = ACTIONS(2646),
    [anon_sym_noreturn] = ACTIONS(2646),
    [anon_sym_LBRACK] = ACTIONS(2648),
    [anon_sym___declspec] = ACTIONS(2646),
    [anon_sym___cdecl] = ACTIONS(2646),
    [anon_sym___clrcall] = ACTIONS(2646),
    [anon_sym___stdcall] = ACTIONS(2646),
    [anon_sym___fastcall] = ACTIONS(2646),
    [anon_sym___thiscall] = ACTIONS(2646),
    [anon_sym___vectorcall] = ACTIONS(2646),
    [anon_sym_LBRACE] = ACTIONS(2648),
    [anon_sym_signed] = ACTIONS(2646),
    [anon_sym_unsigned] = ACTIONS(2646),
    [anon_sym_long] = ACTIONS(2646),
    [anon_sym_short] = ACTIONS(2646),
    [anon_sym_static] = ACTIONS(2646),
    [anon_sym_ATautoreleasepool] = ACTIONS(2648),
    [anon_sym_auto] = ACTIONS(2646),
    [anon_sym_register] = ACTIONS(2646),
    [anon_sym_inline] = ACTIONS(2646),
    [anon_sym___inline] = ACTIONS(2646),
    [anon_sym___inline__] = ACTIONS(2646),
    [anon_sym___forceinline] = ACTIONS(2646),
    [anon_sym_thread_local] = ACTIONS(2646),
    [anon_sym___thread] = ACTIONS(2646),
    [anon_sym_CG_EXTERN] = ACTIONS(2646),
    [anon_sym_CG_INLINE] = ACTIONS(2646),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2646),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2646),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2646),
    [anon_sym_IBOutlet] = ACTIONS(2646),
    [anon_sym_IBInspectable] = ACTIONS(2646),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2646),
    [anon_sym_NS_INLINE] = ACTIONS(2646),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2646),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2646),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2646),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2646),
    [anon_sym_const] = ACTIONS(2646),
    [anon_sym_constexpr] = ACTIONS(2646),
    [anon_sym_volatile] = ACTIONS(2646),
    [anon_sym_restrict] = ACTIONS(2646),
    [anon_sym___restrict__] = ACTIONS(2646),
    [anon_sym__Atomic] = ACTIONS(2646),
    [anon_sym__Noreturn] = ACTIONS(2646),
    [anon_sym__Nonnull] = ACTIONS(2646),
    [anon_sym_nullable] = ACTIONS(2646),
    [anon_sym__Complex] = ACTIONS(2646),
    [anon_sym__Nullable] = ACTIONS(2646),
    [anon_sym__Nullable_result] = ACTIONS(2646),
    [anon_sym__Null_unspecified] = ACTIONS(2646),
    [anon_sym___autoreleasing] = ACTIONS(2646),
    [anon_sym___block] = ACTIONS(2646),
    [anon_sym___bridge] = ACTIONS(2646),
    [anon_sym___bridge_retained] = ACTIONS(2646),
    [anon_sym___bridge_transfer] = ACTIONS(2646),
    [anon_sym___complex] = ACTIONS(2646),
    [anon_sym___const] = ACTIONS(2646),
    [anon_sym___imag] = ACTIONS(2646),
    [anon_sym___kindof] = ACTIONS(2646),
    [anon_sym___nonnull] = ACTIONS(2646),
    [anon_sym___nullable] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2646),
    [anon_sym___real] = ACTIONS(2646),
    [anon_sym___strong] = ACTIONS(2646),
    [anon_sym___unsafe_unretained] = ACTIONS(2646),
    [anon_sym___unused] = ACTIONS(2646),
    [anon_sym___weak] = ACTIONS(2646),
    [anon_sym_alignas] = ACTIONS(2646),
    [anon_sym__Alignas] = ACTIONS(2646),
    [sym_primitive_type] = ACTIONS(2646),
    [anon_sym_enum] = ACTIONS(2646),
    [anon_sym_struct] = ACTIONS(2646),
    [anon_sym_union] = ACTIONS(2646),
    [anon_sym_if] = ACTIONS(2646),
    [anon_sym_switch] = ACTIONS(2646),
    [anon_sym_case] = ACTIONS(2646),
    [anon_sym_default] = ACTIONS(2646),
    [anon_sym_while] = ACTIONS(2646),
    [anon_sym_do] = ACTIONS(2646),
    [anon_sym_for] = ACTIONS(2646),
    [anon_sym_in] = ACTIONS(2646),
    [anon_sym_return] = ACTIONS(2646),
    [anon_sym_break] = ACTIONS(2646),
    [anon_sym_continue] = ACTIONS(2646),
    [anon_sym_goto] = ACTIONS(2646),
    [anon_sym_DASH_DASH] = ACTIONS(2648),
    [anon_sym_PLUS_PLUS] = ACTIONS(2648),
    [anon_sym_sizeof] = ACTIONS(2646),
    [anon_sym___alignof__] = ACTIONS(2646),
    [anon_sym___alignof] = ACTIONS(2646),
    [anon_sym__alignof] = ACTIONS(2646),
    [anon_sym_alignof] = ACTIONS(2646),
    [anon_sym__Alignof] = ACTIONS(2646),
    [anon_sym_offsetof] = ACTIONS(2646),
    [anon_sym__Generic] = ACTIONS(2646),
    [anon_sym_asm] = ACTIONS(2646),
    [anon_sym___asm__] = ACTIONS(2646),
    [anon_sym___asm] = ACTIONS(2646),
    [sym_number_literal] = ACTIONS(2648),
    [anon_sym_L_SQUOTE] = ACTIONS(2648),
    [anon_sym_u_SQUOTE] = ACTIONS(2648),
    [anon_sym_U_SQUOTE] = ACTIONS(2648),
    [anon_sym_u8_SQUOTE] = ACTIONS(2648),
    [anon_sym_SQUOTE] = ACTIONS(2648),
    [anon_sym_AT] = ACTIONS(2646),
    [anon_sym_DQUOTE] = ACTIONS(2648),
    [anon_sym_L_DQUOTE] = ACTIONS(2648),
    [anon_sym_u_DQUOTE] = ACTIONS(2648),
    [anon_sym_U_DQUOTE] = ACTIONS(2648),
    [anon_sym_u8_DQUOTE] = ACTIONS(2648),
    [sym_true] = ACTIONS(2646),
    [sym_false] = ACTIONS(2646),
    [anon_sym_NULL] = ACTIONS(2646),
    [anon_sym_nullptr] = ACTIONS(2646),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2646),
    [anon_sym___typeof] = ACTIONS(2646),
    [anon_sym_typeof] = ACTIONS(2646),
    [anon_sym_ATimport] = ACTIONS(2648),
    [aux_sym_preproc_undef_token1] = ACTIONS(2646),
    [anon_sym_POUND] = ACTIONS(2646),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2646),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2646),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2646),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2646),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2646),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2646),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2646),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_API_AVAILABLE] = ACTIONS(2646),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_API_DEPRECATED] = ACTIONS(2646),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2646),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym___deprecated_msg] = ACTIONS(2646),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2646),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2646),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2646),
    [anon_sym_ATprotocol] = ACTIONS(2648),
    [anon_sym_ATinterface] = ACTIONS(2648),
    [anon_sym_ATimplementation] = ACTIONS(2648),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2648),
    [anon_sym_ATsynthesize] = ACTIONS(2648),
    [anon_sym_ATdynamic] = ACTIONS(2648),
    [anon_sym_ATtry] = ACTIONS(2648),
    [anon_sym___try] = ACTIONS(2646),
    [anon_sym_ATthrow] = ACTIONS(2648),
    [anon_sym_ATselector] = ACTIONS(2648),
    [anon_sym_ATavailable] = ACTIONS(2648),
    [anon_sym___builtin_available] = ACTIONS(2646),
    [anon_sym_va_arg] = ACTIONS(2646),
    [anon_sym_ATencode] = ACTIONS(2648),
    [anon_sym_ATsynchronized] = ACTIONS(2648),
    [anon_sym_BOOL] = ACTIONS(2646),
    [anon_sym_IMP] = ACTIONS(2646),
    [anon_sym_SEL] = ACTIONS(2646),
    [anon_sym_Class] = ACTIONS(2646),
    [anon_sym_id] = ACTIONS(2646),
  },
  [STATE(526)] = {
    [sym_identifier] = ACTIONS(2668),
    [aux_sym_preproc_include_token1] = ACTIONS(2668),
    [aux_sym_preproc_include_token2] = ACTIONS(2668),
    [aux_sym_preproc_def_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token2] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2668),
    [aux_sym_preproc_else_token1] = ACTIONS(2668),
    [aux_sym_preproc_elif_token1] = ACTIONS(2668),
    [sym_preproc_directive] = ACTIONS(2668),
    [anon_sym_LPAREN2] = ACTIONS(2670),
    [anon_sym_BANG] = ACTIONS(2670),
    [anon_sym_TILDE] = ACTIONS(2670),
    [anon_sym_DASH] = ACTIONS(2668),
    [anon_sym_PLUS] = ACTIONS(2668),
    [anon_sym_STAR] = ACTIONS(2670),
    [anon_sym_CARET] = ACTIONS(2670),
    [anon_sym_AMP] = ACTIONS(2670),
    [anon_sym_SEMI] = ACTIONS(2670),
    [anon_sym___extension__] = ACTIONS(2668),
    [anon_sym_typedef] = ACTIONS(2668),
    [anon_sym_extern] = ACTIONS(2668),
    [anon_sym___attribute__] = ACTIONS(2668),
    [anon_sym___attribute] = ACTIONS(2668),
    [anon_sym_noreturn] = ACTIONS(2668),
    [anon_sym_LBRACK] = ACTIONS(2670),
    [anon_sym___declspec] = ACTIONS(2668),
    [anon_sym___cdecl] = ACTIONS(2668),
    [anon_sym___clrcall] = ACTIONS(2668),
    [anon_sym___stdcall] = ACTIONS(2668),
    [anon_sym___fastcall] = ACTIONS(2668),
    [anon_sym___thiscall] = ACTIONS(2668),
    [anon_sym___vectorcall] = ACTIONS(2668),
    [anon_sym_LBRACE] = ACTIONS(2670),
    [anon_sym_signed] = ACTIONS(2668),
    [anon_sym_unsigned] = ACTIONS(2668),
    [anon_sym_long] = ACTIONS(2668),
    [anon_sym_short] = ACTIONS(2668),
    [anon_sym_static] = ACTIONS(2668),
    [anon_sym_ATautoreleasepool] = ACTIONS(2670),
    [anon_sym_auto] = ACTIONS(2668),
    [anon_sym_register] = ACTIONS(2668),
    [anon_sym_inline] = ACTIONS(2668),
    [anon_sym___inline] = ACTIONS(2668),
    [anon_sym___inline__] = ACTIONS(2668),
    [anon_sym___forceinline] = ACTIONS(2668),
    [anon_sym_thread_local] = ACTIONS(2668),
    [anon_sym___thread] = ACTIONS(2668),
    [anon_sym_CG_EXTERN] = ACTIONS(2668),
    [anon_sym_CG_INLINE] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2668),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2668),
    [anon_sym_IBOutlet] = ACTIONS(2668),
    [anon_sym_IBInspectable] = ACTIONS(2668),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2668),
    [anon_sym_NS_INLINE] = ACTIONS(2668),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2668),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2668),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2668),
    [anon_sym_const] = ACTIONS(2668),
    [anon_sym_constexpr] = ACTIONS(2668),
    [anon_sym_volatile] = ACTIONS(2668),
    [anon_sym_restrict] = ACTIONS(2668),
    [anon_sym___restrict__] = ACTIONS(2668),
    [anon_sym__Atomic] = ACTIONS(2668),
    [anon_sym__Noreturn] = ACTIONS(2668),
    [anon_sym__Nonnull] = ACTIONS(2668),
    [anon_sym_nullable] = ACTIONS(2668),
    [anon_sym__Complex] = ACTIONS(2668),
    [anon_sym__Nullable] = ACTIONS(2668),
    [anon_sym__Nullable_result] = ACTIONS(2668),
    [anon_sym__Null_unspecified] = ACTIONS(2668),
    [anon_sym___autoreleasing] = ACTIONS(2668),
    [anon_sym___block] = ACTIONS(2668),
    [anon_sym___bridge] = ACTIONS(2668),
    [anon_sym___bridge_retained] = ACTIONS(2668),
    [anon_sym___bridge_transfer] = ACTIONS(2668),
    [anon_sym___complex] = ACTIONS(2668),
    [anon_sym___const] = ACTIONS(2668),
    [anon_sym___imag] = ACTIONS(2668),
    [anon_sym___kindof] = ACTIONS(2668),
    [anon_sym___nonnull] = ACTIONS(2668),
    [anon_sym___nullable] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2668),
    [anon_sym___real] = ACTIONS(2668),
    [anon_sym___strong] = ACTIONS(2668),
    [anon_sym___unsafe_unretained] = ACTIONS(2668),
    [anon_sym___unused] = ACTIONS(2668),
    [anon_sym___weak] = ACTIONS(2668),
    [anon_sym_alignas] = ACTIONS(2668),
    [anon_sym__Alignas] = ACTIONS(2668),
    [sym_primitive_type] = ACTIONS(2668),
    [anon_sym_enum] = ACTIONS(2668),
    [anon_sym_struct] = ACTIONS(2668),
    [anon_sym_union] = ACTIONS(2668),
    [anon_sym_if] = ACTIONS(2668),
    [anon_sym_switch] = ACTIONS(2668),
    [anon_sym_case] = ACTIONS(2668),
    [anon_sym_default] = ACTIONS(2668),
    [anon_sym_while] = ACTIONS(2668),
    [anon_sym_do] = ACTIONS(2668),
    [anon_sym_for] = ACTIONS(2668),
    [anon_sym_in] = ACTIONS(2668),
    [anon_sym_return] = ACTIONS(2668),
    [anon_sym_break] = ACTIONS(2668),
    [anon_sym_continue] = ACTIONS(2668),
    [anon_sym_goto] = ACTIONS(2668),
    [anon_sym_DASH_DASH] = ACTIONS(2670),
    [anon_sym_PLUS_PLUS] = ACTIONS(2670),
    [anon_sym_sizeof] = ACTIONS(2668),
    [anon_sym___alignof__] = ACTIONS(2668),
    [anon_sym___alignof] = ACTIONS(2668),
    [anon_sym__alignof] = ACTIONS(2668),
    [anon_sym_alignof] = ACTIONS(2668),
    [anon_sym__Alignof] = ACTIONS(2668),
    [anon_sym_offsetof] = ACTIONS(2668),
    [anon_sym__Generic] = ACTIONS(2668),
    [anon_sym_asm] = ACTIONS(2668),
    [anon_sym___asm__] = ACTIONS(2668),
    [anon_sym___asm] = ACTIONS(2668),
    [sym_number_literal] = ACTIONS(2670),
    [anon_sym_L_SQUOTE] = ACTIONS(2670),
    [anon_sym_u_SQUOTE] = ACTIONS(2670),
    [anon_sym_U_SQUOTE] = ACTIONS(2670),
    [anon_sym_u8_SQUOTE] = ACTIONS(2670),
    [anon_sym_SQUOTE] = ACTIONS(2670),
    [anon_sym_AT] = ACTIONS(2668),
    [anon_sym_DQUOTE] = ACTIONS(2670),
    [anon_sym_L_DQUOTE] = ACTIONS(2670),
    [anon_sym_u_DQUOTE] = ACTIONS(2670),
    [anon_sym_U_DQUOTE] = ACTIONS(2670),
    [anon_sym_u8_DQUOTE] = ACTIONS(2670),
    [sym_true] = ACTIONS(2668),
    [sym_false] = ACTIONS(2668),
    [anon_sym_NULL] = ACTIONS(2668),
    [anon_sym_nullptr] = ACTIONS(2668),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2668),
    [anon_sym___typeof] = ACTIONS(2668),
    [anon_sym_typeof] = ACTIONS(2668),
    [anon_sym_ATimport] = ACTIONS(2670),
    [aux_sym_preproc_undef_token1] = ACTIONS(2668),
    [anon_sym_POUND] = ACTIONS(2668),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2668),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2668),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2668),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_API_AVAILABLE] = ACTIONS(2668),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_API_DEPRECATED] = ACTIONS(2668),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym___deprecated_msg] = ACTIONS(2668),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2668),
    [anon_sym_ATprotocol] = ACTIONS(2670),
    [anon_sym_ATinterface] = ACTIONS(2670),
    [anon_sym_ATimplementation] = ACTIONS(2670),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2670),
    [anon_sym_ATsynthesize] = ACTIONS(2670),
    [anon_sym_ATdynamic] = ACTIONS(2670),
    [anon_sym_ATtry] = ACTIONS(2670),
    [anon_sym___try] = ACTIONS(2668),
    [anon_sym_ATthrow] = ACTIONS(2670),
    [anon_sym_ATselector] = ACTIONS(2670),
    [anon_sym_ATavailable] = ACTIONS(2670),
    [anon_sym___builtin_available] = ACTIONS(2668),
    [anon_sym_va_arg] = ACTIONS(2668),
    [anon_sym_ATencode] = ACTIONS(2670),
    [anon_sym_ATsynchronized] = ACTIONS(2670),
    [anon_sym_BOOL] = ACTIONS(2668),
    [anon_sym_IMP] = ACTIONS(2668),
    [anon_sym_SEL] = ACTIONS(2668),
    [anon_sym_Class] = ACTIONS(2668),
    [anon_sym_id] = ACTIONS(2668),
  },
  [STATE(527)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(528)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token2] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(529)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token2] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(530)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token2] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(531)] = {
    [sym_identifier] = ACTIONS(2514),
    [aux_sym_preproc_include_token1] = ACTIONS(2514),
    [aux_sym_preproc_include_token2] = ACTIONS(2514),
    [aux_sym_preproc_def_token1] = ACTIONS(2514),
    [aux_sym_preproc_if_token1] = ACTIONS(2514),
    [aux_sym_preproc_if_token2] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2514),
    [sym_preproc_directive] = ACTIONS(2514),
    [anon_sym_LPAREN2] = ACTIONS(2516),
    [anon_sym_BANG] = ACTIONS(2516),
    [anon_sym_TILDE] = ACTIONS(2516),
    [anon_sym_DASH] = ACTIONS(2514),
    [anon_sym_PLUS] = ACTIONS(2514),
    [anon_sym_STAR] = ACTIONS(2516),
    [anon_sym_CARET] = ACTIONS(2516),
    [anon_sym_AMP] = ACTIONS(2516),
    [anon_sym_SEMI] = ACTIONS(2516),
    [anon_sym___extension__] = ACTIONS(2514),
    [anon_sym_typedef] = ACTIONS(2514),
    [anon_sym_extern] = ACTIONS(2514),
    [anon_sym___attribute__] = ACTIONS(2514),
    [anon_sym___attribute] = ACTIONS(2514),
    [anon_sym_noreturn] = ACTIONS(2514),
    [anon_sym_LBRACK] = ACTIONS(2516),
    [anon_sym___declspec] = ACTIONS(2514),
    [anon_sym___cdecl] = ACTIONS(2514),
    [anon_sym___clrcall] = ACTIONS(2514),
    [anon_sym___stdcall] = ACTIONS(2514),
    [anon_sym___fastcall] = ACTIONS(2514),
    [anon_sym___thiscall] = ACTIONS(2514),
    [anon_sym___vectorcall] = ACTIONS(2514),
    [anon_sym_LBRACE] = ACTIONS(2516),
    [anon_sym_signed] = ACTIONS(2514),
    [anon_sym_unsigned] = ACTIONS(2514),
    [anon_sym_long] = ACTIONS(2514),
    [anon_sym_short] = ACTIONS(2514),
    [anon_sym_static] = ACTIONS(2514),
    [anon_sym_ATautoreleasepool] = ACTIONS(2516),
    [anon_sym_auto] = ACTIONS(2514),
    [anon_sym_register] = ACTIONS(2514),
    [anon_sym_inline] = ACTIONS(2514),
    [anon_sym___inline] = ACTIONS(2514),
    [anon_sym___inline__] = ACTIONS(2514),
    [anon_sym___forceinline] = ACTIONS(2514),
    [anon_sym_thread_local] = ACTIONS(2514),
    [anon_sym___thread] = ACTIONS(2514),
    [anon_sym_CG_EXTERN] = ACTIONS(2514),
    [anon_sym_CG_INLINE] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2514),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2514),
    [anon_sym_IBOutlet] = ACTIONS(2514),
    [anon_sym_IBInspectable] = ACTIONS(2514),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2514),
    [anon_sym_NS_INLINE] = ACTIONS(2514),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2514),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2514),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2514),
    [anon_sym_const] = ACTIONS(2514),
    [anon_sym_constexpr] = ACTIONS(2514),
    [anon_sym_volatile] = ACTIONS(2514),
    [anon_sym_restrict] = ACTIONS(2514),
    [anon_sym___restrict__] = ACTIONS(2514),
    [anon_sym__Atomic] = ACTIONS(2514),
    [anon_sym__Noreturn] = ACTIONS(2514),
    [anon_sym__Nonnull] = ACTIONS(2514),
    [anon_sym_nullable] = ACTIONS(2514),
    [anon_sym__Complex] = ACTIONS(2514),
    [anon_sym__Nullable] = ACTIONS(2514),
    [anon_sym__Nullable_result] = ACTIONS(2514),
    [anon_sym__Null_unspecified] = ACTIONS(2514),
    [anon_sym___autoreleasing] = ACTIONS(2514),
    [anon_sym___block] = ACTIONS(2514),
    [anon_sym___bridge] = ACTIONS(2514),
    [anon_sym___bridge_retained] = ACTIONS(2514),
    [anon_sym___bridge_transfer] = ACTIONS(2514),
    [anon_sym___complex] = ACTIONS(2514),
    [anon_sym___const] = ACTIONS(2514),
    [anon_sym___imag] = ACTIONS(2514),
    [anon_sym___kindof] = ACTIONS(2514),
    [anon_sym___nonnull] = ACTIONS(2514),
    [anon_sym___nullable] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2514),
    [anon_sym___real] = ACTIONS(2514),
    [anon_sym___strong] = ACTIONS(2514),
    [anon_sym___unsafe_unretained] = ACTIONS(2514),
    [anon_sym___unused] = ACTIONS(2514),
    [anon_sym___weak] = ACTIONS(2514),
    [anon_sym_alignas] = ACTIONS(2514),
    [anon_sym__Alignas] = ACTIONS(2514),
    [sym_primitive_type] = ACTIONS(2514),
    [anon_sym_enum] = ACTIONS(2514),
    [anon_sym_struct] = ACTIONS(2514),
    [anon_sym_union] = ACTIONS(2514),
    [anon_sym_if] = ACTIONS(2514),
    [anon_sym_else] = ACTIONS(2514),
    [anon_sym_switch] = ACTIONS(2514),
    [anon_sym_case] = ACTIONS(2514),
    [anon_sym_default] = ACTIONS(2514),
    [anon_sym_while] = ACTIONS(2514),
    [anon_sym_do] = ACTIONS(2514),
    [anon_sym_for] = ACTIONS(2514),
    [anon_sym_in] = ACTIONS(2514),
    [anon_sym_return] = ACTIONS(2514),
    [anon_sym_break] = ACTIONS(2514),
    [anon_sym_continue] = ACTIONS(2514),
    [anon_sym_goto] = ACTIONS(2514),
    [anon_sym_DASH_DASH] = ACTIONS(2516),
    [anon_sym_PLUS_PLUS] = ACTIONS(2516),
    [anon_sym_sizeof] = ACTIONS(2514),
    [anon_sym___alignof__] = ACTIONS(2514),
    [anon_sym___alignof] = ACTIONS(2514),
    [anon_sym__alignof] = ACTIONS(2514),
    [anon_sym_alignof] = ACTIONS(2514),
    [anon_sym__Alignof] = ACTIONS(2514),
    [anon_sym_offsetof] = ACTIONS(2514),
    [anon_sym__Generic] = ACTIONS(2514),
    [anon_sym_asm] = ACTIONS(2514),
    [anon_sym___asm__] = ACTIONS(2514),
    [anon_sym___asm] = ACTIONS(2514),
    [sym_number_literal] = ACTIONS(2516),
    [anon_sym_L_SQUOTE] = ACTIONS(2516),
    [anon_sym_u_SQUOTE] = ACTIONS(2516),
    [anon_sym_U_SQUOTE] = ACTIONS(2516),
    [anon_sym_u8_SQUOTE] = ACTIONS(2516),
    [anon_sym_SQUOTE] = ACTIONS(2516),
    [anon_sym_AT] = ACTIONS(2514),
    [anon_sym_DQUOTE] = ACTIONS(2516),
    [anon_sym_L_DQUOTE] = ACTIONS(2516),
    [anon_sym_u_DQUOTE] = ACTIONS(2516),
    [anon_sym_U_DQUOTE] = ACTIONS(2516),
    [anon_sym_u8_DQUOTE] = ACTIONS(2516),
    [sym_true] = ACTIONS(2514),
    [sym_false] = ACTIONS(2514),
    [anon_sym_NULL] = ACTIONS(2514),
    [anon_sym_nullptr] = ACTIONS(2514),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2514),
    [anon_sym___typeof] = ACTIONS(2514),
    [anon_sym_typeof] = ACTIONS(2514),
    [anon_sym_ATimport] = ACTIONS(2516),
    [aux_sym_preproc_undef_token1] = ACTIONS(2514),
    [anon_sym_POUND] = ACTIONS(2514),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2514),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2514),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2514),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_API_AVAILABLE] = ACTIONS(2514),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_API_DEPRECATED] = ACTIONS(2514),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym___deprecated_msg] = ACTIONS(2514),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2514),
    [anon_sym_ATprotocol] = ACTIONS(2516),
    [anon_sym_ATinterface] = ACTIONS(2516),
    [anon_sym_ATimplementation] = ACTIONS(2516),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2516),
    [anon_sym_ATsynthesize] = ACTIONS(2516),
    [anon_sym_ATdynamic] = ACTIONS(2516),
    [anon_sym_ATtry] = ACTIONS(2516),
    [anon_sym___try] = ACTIONS(2514),
    [anon_sym_ATthrow] = ACTIONS(2516),
    [anon_sym_ATselector] = ACTIONS(2516),
    [anon_sym_ATavailable] = ACTIONS(2516),
    [anon_sym___builtin_available] = ACTIONS(2514),
    [anon_sym_va_arg] = ACTIONS(2514),
    [anon_sym_ATencode] = ACTIONS(2516),
    [anon_sym_ATsynchronized] = ACTIONS(2516),
    [anon_sym_BOOL] = ACTIONS(2514),
    [anon_sym_IMP] = ACTIONS(2514),
    [anon_sym_SEL] = ACTIONS(2514),
    [anon_sym_Class] = ACTIONS(2514),
    [anon_sym_id] = ACTIONS(2514),
  },
  [STATE(532)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token2] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(533)] = {
    [sym_identifier] = ACTIONS(2518),
    [aux_sym_preproc_include_token1] = ACTIONS(2518),
    [aux_sym_preproc_include_token2] = ACTIONS(2518),
    [aux_sym_preproc_def_token1] = ACTIONS(2518),
    [aux_sym_preproc_if_token1] = ACTIONS(2518),
    [aux_sym_preproc_if_token2] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2518),
    [sym_preproc_directive] = ACTIONS(2518),
    [anon_sym_LPAREN2] = ACTIONS(2520),
    [anon_sym_BANG] = ACTIONS(2520),
    [anon_sym_TILDE] = ACTIONS(2520),
    [anon_sym_DASH] = ACTIONS(2518),
    [anon_sym_PLUS] = ACTIONS(2518),
    [anon_sym_STAR] = ACTIONS(2520),
    [anon_sym_CARET] = ACTIONS(2520),
    [anon_sym_AMP] = ACTIONS(2520),
    [anon_sym_SEMI] = ACTIONS(2520),
    [anon_sym___extension__] = ACTIONS(2518),
    [anon_sym_typedef] = ACTIONS(2518),
    [anon_sym_extern] = ACTIONS(2518),
    [anon_sym___attribute__] = ACTIONS(2518),
    [anon_sym___attribute] = ACTIONS(2518),
    [anon_sym_noreturn] = ACTIONS(2518),
    [anon_sym_LBRACK] = ACTIONS(2520),
    [anon_sym___declspec] = ACTIONS(2518),
    [anon_sym___cdecl] = ACTIONS(2518),
    [anon_sym___clrcall] = ACTIONS(2518),
    [anon_sym___stdcall] = ACTIONS(2518),
    [anon_sym___fastcall] = ACTIONS(2518),
    [anon_sym___thiscall] = ACTIONS(2518),
    [anon_sym___vectorcall] = ACTIONS(2518),
    [anon_sym_LBRACE] = ACTIONS(2520),
    [anon_sym_signed] = ACTIONS(2518),
    [anon_sym_unsigned] = ACTIONS(2518),
    [anon_sym_long] = ACTIONS(2518),
    [anon_sym_short] = ACTIONS(2518),
    [anon_sym_static] = ACTIONS(2518),
    [anon_sym_ATautoreleasepool] = ACTIONS(2520),
    [anon_sym_auto] = ACTIONS(2518),
    [anon_sym_register] = ACTIONS(2518),
    [anon_sym_inline] = ACTIONS(2518),
    [anon_sym___inline] = ACTIONS(2518),
    [anon_sym___inline__] = ACTIONS(2518),
    [anon_sym___forceinline] = ACTIONS(2518),
    [anon_sym_thread_local] = ACTIONS(2518),
    [anon_sym___thread] = ACTIONS(2518),
    [anon_sym_CG_EXTERN] = ACTIONS(2518),
    [anon_sym_CG_INLINE] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2518),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2518),
    [anon_sym_IBOutlet] = ACTIONS(2518),
    [anon_sym_IBInspectable] = ACTIONS(2518),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2518),
    [anon_sym_NS_INLINE] = ACTIONS(2518),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2518),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2518),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2518),
    [anon_sym_const] = ACTIONS(2518),
    [anon_sym_constexpr] = ACTIONS(2518),
    [anon_sym_volatile] = ACTIONS(2518),
    [anon_sym_restrict] = ACTIONS(2518),
    [anon_sym___restrict__] = ACTIONS(2518),
    [anon_sym__Atomic] = ACTIONS(2518),
    [anon_sym__Noreturn] = ACTIONS(2518),
    [anon_sym__Nonnull] = ACTIONS(2518),
    [anon_sym_nullable] = ACTIONS(2518),
    [anon_sym__Complex] = ACTIONS(2518),
    [anon_sym__Nullable] = ACTIONS(2518),
    [anon_sym__Nullable_result] = ACTIONS(2518),
    [anon_sym__Null_unspecified] = ACTIONS(2518),
    [anon_sym___autoreleasing] = ACTIONS(2518),
    [anon_sym___block] = ACTIONS(2518),
    [anon_sym___bridge] = ACTIONS(2518),
    [anon_sym___bridge_retained] = ACTIONS(2518),
    [anon_sym___bridge_transfer] = ACTIONS(2518),
    [anon_sym___complex] = ACTIONS(2518),
    [anon_sym___const] = ACTIONS(2518),
    [anon_sym___imag] = ACTIONS(2518),
    [anon_sym___kindof] = ACTIONS(2518),
    [anon_sym___nonnull] = ACTIONS(2518),
    [anon_sym___nullable] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2518),
    [anon_sym___real] = ACTIONS(2518),
    [anon_sym___strong] = ACTIONS(2518),
    [anon_sym___unsafe_unretained] = ACTIONS(2518),
    [anon_sym___unused] = ACTIONS(2518),
    [anon_sym___weak] = ACTIONS(2518),
    [anon_sym_alignas] = ACTIONS(2518),
    [anon_sym__Alignas] = ACTIONS(2518),
    [sym_primitive_type] = ACTIONS(2518),
    [anon_sym_enum] = ACTIONS(2518),
    [anon_sym_struct] = ACTIONS(2518),
    [anon_sym_union] = ACTIONS(2518),
    [anon_sym_if] = ACTIONS(2518),
    [anon_sym_else] = ACTIONS(2518),
    [anon_sym_switch] = ACTIONS(2518),
    [anon_sym_case] = ACTIONS(2518),
    [anon_sym_default] = ACTIONS(2518),
    [anon_sym_while] = ACTIONS(2518),
    [anon_sym_do] = ACTIONS(2518),
    [anon_sym_for] = ACTIONS(2518),
    [anon_sym_in] = ACTIONS(2518),
    [anon_sym_return] = ACTIONS(2518),
    [anon_sym_break] = ACTIONS(2518),
    [anon_sym_continue] = ACTIONS(2518),
    [anon_sym_goto] = ACTIONS(2518),
    [anon_sym_DASH_DASH] = ACTIONS(2520),
    [anon_sym_PLUS_PLUS] = ACTIONS(2520),
    [anon_sym_sizeof] = ACTIONS(2518),
    [anon_sym___alignof__] = ACTIONS(2518),
    [anon_sym___alignof] = ACTIONS(2518),
    [anon_sym__alignof] = ACTIONS(2518),
    [anon_sym_alignof] = ACTIONS(2518),
    [anon_sym__Alignof] = ACTIONS(2518),
    [anon_sym_offsetof] = ACTIONS(2518),
    [anon_sym__Generic] = ACTIONS(2518),
    [anon_sym_asm] = ACTIONS(2518),
    [anon_sym___asm__] = ACTIONS(2518),
    [anon_sym___asm] = ACTIONS(2518),
    [sym_number_literal] = ACTIONS(2520),
    [anon_sym_L_SQUOTE] = ACTIONS(2520),
    [anon_sym_u_SQUOTE] = ACTIONS(2520),
    [anon_sym_U_SQUOTE] = ACTIONS(2520),
    [anon_sym_u8_SQUOTE] = ACTIONS(2520),
    [anon_sym_SQUOTE] = ACTIONS(2520),
    [anon_sym_AT] = ACTIONS(2518),
    [anon_sym_DQUOTE] = ACTIONS(2520),
    [anon_sym_L_DQUOTE] = ACTIONS(2520),
    [anon_sym_u_DQUOTE] = ACTIONS(2520),
    [anon_sym_U_DQUOTE] = ACTIONS(2520),
    [anon_sym_u8_DQUOTE] = ACTIONS(2520),
    [sym_true] = ACTIONS(2518),
    [sym_false] = ACTIONS(2518),
    [anon_sym_NULL] = ACTIONS(2518),
    [anon_sym_nullptr] = ACTIONS(2518),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2518),
    [anon_sym___typeof] = ACTIONS(2518),
    [anon_sym_typeof] = ACTIONS(2518),
    [anon_sym_ATimport] = ACTIONS(2520),
    [aux_sym_preproc_undef_token1] = ACTIONS(2518),
    [anon_sym_POUND] = ACTIONS(2518),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2518),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2518),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2518),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_API_AVAILABLE] = ACTIONS(2518),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_API_DEPRECATED] = ACTIONS(2518),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym___deprecated_msg] = ACTIONS(2518),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2518),
    [anon_sym_ATprotocol] = ACTIONS(2520),
    [anon_sym_ATinterface] = ACTIONS(2520),
    [anon_sym_ATimplementation] = ACTIONS(2520),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2520),
    [anon_sym_ATsynthesize] = ACTIONS(2520),
    [anon_sym_ATdynamic] = ACTIONS(2520),
    [anon_sym_ATtry] = ACTIONS(2520),
    [anon_sym___try] = ACTIONS(2518),
    [anon_sym_ATthrow] = ACTIONS(2520),
    [anon_sym_ATselector] = ACTIONS(2520),
    [anon_sym_ATavailable] = ACTIONS(2520),
    [anon_sym___builtin_available] = ACTIONS(2518),
    [anon_sym_va_arg] = ACTIONS(2518),
    [anon_sym_ATencode] = ACTIONS(2520),
    [anon_sym_ATsynchronized] = ACTIONS(2520),
    [anon_sym_BOOL] = ACTIONS(2518),
    [anon_sym_IMP] = ACTIONS(2518),
    [anon_sym_SEL] = ACTIONS(2518),
    [anon_sym_Class] = ACTIONS(2518),
    [anon_sym_id] = ACTIONS(2518),
  },
  [STATE(534)] = {
    [sym_identifier] = ACTIONS(2522),
    [aux_sym_preproc_include_token1] = ACTIONS(2522),
    [aux_sym_preproc_include_token2] = ACTIONS(2522),
    [aux_sym_preproc_def_token1] = ACTIONS(2522),
    [aux_sym_preproc_if_token1] = ACTIONS(2522),
    [aux_sym_preproc_if_token2] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2522),
    [sym_preproc_directive] = ACTIONS(2522),
    [anon_sym_LPAREN2] = ACTIONS(2524),
    [anon_sym_BANG] = ACTIONS(2524),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_DASH] = ACTIONS(2522),
    [anon_sym_PLUS] = ACTIONS(2522),
    [anon_sym_STAR] = ACTIONS(2524),
    [anon_sym_CARET] = ACTIONS(2524),
    [anon_sym_AMP] = ACTIONS(2524),
    [anon_sym_SEMI] = ACTIONS(2524),
    [anon_sym___extension__] = ACTIONS(2522),
    [anon_sym_typedef] = ACTIONS(2522),
    [anon_sym_extern] = ACTIONS(2522),
    [anon_sym___attribute__] = ACTIONS(2522),
    [anon_sym___attribute] = ACTIONS(2522),
    [anon_sym_noreturn] = ACTIONS(2522),
    [anon_sym_LBRACK] = ACTIONS(2524),
    [anon_sym___declspec] = ACTIONS(2522),
    [anon_sym___cdecl] = ACTIONS(2522),
    [anon_sym___clrcall] = ACTIONS(2522),
    [anon_sym___stdcall] = ACTIONS(2522),
    [anon_sym___fastcall] = ACTIONS(2522),
    [anon_sym___thiscall] = ACTIONS(2522),
    [anon_sym___vectorcall] = ACTIONS(2522),
    [anon_sym_LBRACE] = ACTIONS(2524),
    [anon_sym_signed] = ACTIONS(2522),
    [anon_sym_unsigned] = ACTIONS(2522),
    [anon_sym_long] = ACTIONS(2522),
    [anon_sym_short] = ACTIONS(2522),
    [anon_sym_static] = ACTIONS(2522),
    [anon_sym_ATautoreleasepool] = ACTIONS(2524),
    [anon_sym_auto] = ACTIONS(2522),
    [anon_sym_register] = ACTIONS(2522),
    [anon_sym_inline] = ACTIONS(2522),
    [anon_sym___inline] = ACTIONS(2522),
    [anon_sym___inline__] = ACTIONS(2522),
    [anon_sym___forceinline] = ACTIONS(2522),
    [anon_sym_thread_local] = ACTIONS(2522),
    [anon_sym___thread] = ACTIONS(2522),
    [anon_sym_CG_EXTERN] = ACTIONS(2522),
    [anon_sym_CG_INLINE] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2522),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2522),
    [anon_sym_IBOutlet] = ACTIONS(2522),
    [anon_sym_IBInspectable] = ACTIONS(2522),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2522),
    [anon_sym_NS_INLINE] = ACTIONS(2522),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2522),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2522),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2522),
    [anon_sym_const] = ACTIONS(2522),
    [anon_sym_constexpr] = ACTIONS(2522),
    [anon_sym_volatile] = ACTIONS(2522),
    [anon_sym_restrict] = ACTIONS(2522),
    [anon_sym___restrict__] = ACTIONS(2522),
    [anon_sym__Atomic] = ACTIONS(2522),
    [anon_sym__Noreturn] = ACTIONS(2522),
    [anon_sym__Nonnull] = ACTIONS(2522),
    [anon_sym_nullable] = ACTIONS(2522),
    [anon_sym__Complex] = ACTIONS(2522),
    [anon_sym__Nullable] = ACTIONS(2522),
    [anon_sym__Nullable_result] = ACTIONS(2522),
    [anon_sym__Null_unspecified] = ACTIONS(2522),
    [anon_sym___autoreleasing] = ACTIONS(2522),
    [anon_sym___block] = ACTIONS(2522),
    [anon_sym___bridge] = ACTIONS(2522),
    [anon_sym___bridge_retained] = ACTIONS(2522),
    [anon_sym___bridge_transfer] = ACTIONS(2522),
    [anon_sym___complex] = ACTIONS(2522),
    [anon_sym___const] = ACTIONS(2522),
    [anon_sym___imag] = ACTIONS(2522),
    [anon_sym___kindof] = ACTIONS(2522),
    [anon_sym___nonnull] = ACTIONS(2522),
    [anon_sym___nullable] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2522),
    [anon_sym___real] = ACTIONS(2522),
    [anon_sym___strong] = ACTIONS(2522),
    [anon_sym___unsafe_unretained] = ACTIONS(2522),
    [anon_sym___unused] = ACTIONS(2522),
    [anon_sym___weak] = ACTIONS(2522),
    [anon_sym_alignas] = ACTIONS(2522),
    [anon_sym__Alignas] = ACTIONS(2522),
    [sym_primitive_type] = ACTIONS(2522),
    [anon_sym_enum] = ACTIONS(2522),
    [anon_sym_struct] = ACTIONS(2522),
    [anon_sym_union] = ACTIONS(2522),
    [anon_sym_if] = ACTIONS(2522),
    [anon_sym_else] = ACTIONS(2522),
    [anon_sym_switch] = ACTIONS(2522),
    [anon_sym_case] = ACTIONS(2522),
    [anon_sym_default] = ACTIONS(2522),
    [anon_sym_while] = ACTIONS(2522),
    [anon_sym_do] = ACTIONS(2522),
    [anon_sym_for] = ACTIONS(2522),
    [anon_sym_in] = ACTIONS(2522),
    [anon_sym_return] = ACTIONS(2522),
    [anon_sym_break] = ACTIONS(2522),
    [anon_sym_continue] = ACTIONS(2522),
    [anon_sym_goto] = ACTIONS(2522),
    [anon_sym_DASH_DASH] = ACTIONS(2524),
    [anon_sym_PLUS_PLUS] = ACTIONS(2524),
    [anon_sym_sizeof] = ACTIONS(2522),
    [anon_sym___alignof__] = ACTIONS(2522),
    [anon_sym___alignof] = ACTIONS(2522),
    [anon_sym__alignof] = ACTIONS(2522),
    [anon_sym_alignof] = ACTIONS(2522),
    [anon_sym__Alignof] = ACTIONS(2522),
    [anon_sym_offsetof] = ACTIONS(2522),
    [anon_sym__Generic] = ACTIONS(2522),
    [anon_sym_asm] = ACTIONS(2522),
    [anon_sym___asm__] = ACTIONS(2522),
    [anon_sym___asm] = ACTIONS(2522),
    [sym_number_literal] = ACTIONS(2524),
    [anon_sym_L_SQUOTE] = ACTIONS(2524),
    [anon_sym_u_SQUOTE] = ACTIONS(2524),
    [anon_sym_U_SQUOTE] = ACTIONS(2524),
    [anon_sym_u8_SQUOTE] = ACTIONS(2524),
    [anon_sym_SQUOTE] = ACTIONS(2524),
    [anon_sym_AT] = ACTIONS(2522),
    [anon_sym_DQUOTE] = ACTIONS(2524),
    [anon_sym_L_DQUOTE] = ACTIONS(2524),
    [anon_sym_u_DQUOTE] = ACTIONS(2524),
    [anon_sym_U_DQUOTE] = ACTIONS(2524),
    [anon_sym_u8_DQUOTE] = ACTIONS(2524),
    [sym_true] = ACTIONS(2522),
    [sym_false] = ACTIONS(2522),
    [anon_sym_NULL] = ACTIONS(2522),
    [anon_sym_nullptr] = ACTIONS(2522),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2522),
    [anon_sym___typeof] = ACTIONS(2522),
    [anon_sym_typeof] = ACTIONS(2522),
    [anon_sym_ATimport] = ACTIONS(2524),
    [aux_sym_preproc_undef_token1] = ACTIONS(2522),
    [anon_sym_POUND] = ACTIONS(2522),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2522),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2522),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2522),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_API_AVAILABLE] = ACTIONS(2522),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_API_DEPRECATED] = ACTIONS(2522),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym___deprecated_msg] = ACTIONS(2522),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2522),
    [anon_sym_ATprotocol] = ACTIONS(2524),
    [anon_sym_ATinterface] = ACTIONS(2524),
    [anon_sym_ATimplementation] = ACTIONS(2524),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2524),
    [anon_sym_ATsynthesize] = ACTIONS(2524),
    [anon_sym_ATdynamic] = ACTIONS(2524),
    [anon_sym_ATtry] = ACTIONS(2524),
    [anon_sym___try] = ACTIONS(2522),
    [anon_sym_ATthrow] = ACTIONS(2524),
    [anon_sym_ATselector] = ACTIONS(2524),
    [anon_sym_ATavailable] = ACTIONS(2524),
    [anon_sym___builtin_available] = ACTIONS(2522),
    [anon_sym_va_arg] = ACTIONS(2522),
    [anon_sym_ATencode] = ACTIONS(2524),
    [anon_sym_ATsynchronized] = ACTIONS(2524),
    [anon_sym_BOOL] = ACTIONS(2522),
    [anon_sym_IMP] = ACTIONS(2522),
    [anon_sym_SEL] = ACTIONS(2522),
    [anon_sym_Class] = ACTIONS(2522),
    [anon_sym_id] = ACTIONS(2522),
  },
  [STATE(535)] = {
    [sym_identifier] = ACTIONS(2454),
    [aux_sym_preproc_include_token1] = ACTIONS(2454),
    [aux_sym_preproc_include_token2] = ACTIONS(2454),
    [aux_sym_preproc_def_token1] = ACTIONS(2454),
    [aux_sym_preproc_if_token1] = ACTIONS(2454),
    [aux_sym_preproc_if_token2] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2454),
    [sym_preproc_directive] = ACTIONS(2454),
    [anon_sym_LPAREN2] = ACTIONS(2456),
    [anon_sym_BANG] = ACTIONS(2456),
    [anon_sym_TILDE] = ACTIONS(2456),
    [anon_sym_DASH] = ACTIONS(2454),
    [anon_sym_PLUS] = ACTIONS(2454),
    [anon_sym_STAR] = ACTIONS(2456),
    [anon_sym_CARET] = ACTIONS(2456),
    [anon_sym_AMP] = ACTIONS(2456),
    [anon_sym_SEMI] = ACTIONS(2456),
    [anon_sym___extension__] = ACTIONS(2454),
    [anon_sym_typedef] = ACTIONS(2454),
    [anon_sym_extern] = ACTIONS(2454),
    [anon_sym___attribute__] = ACTIONS(2454),
    [anon_sym___attribute] = ACTIONS(2454),
    [anon_sym_noreturn] = ACTIONS(2454),
    [anon_sym_LBRACK] = ACTIONS(2456),
    [anon_sym___declspec] = ACTIONS(2454),
    [anon_sym___cdecl] = ACTIONS(2454),
    [anon_sym___clrcall] = ACTIONS(2454),
    [anon_sym___stdcall] = ACTIONS(2454),
    [anon_sym___fastcall] = ACTIONS(2454),
    [anon_sym___thiscall] = ACTIONS(2454),
    [anon_sym___vectorcall] = ACTIONS(2454),
    [anon_sym_LBRACE] = ACTIONS(2456),
    [anon_sym_signed] = ACTIONS(2454),
    [anon_sym_unsigned] = ACTIONS(2454),
    [anon_sym_long] = ACTIONS(2454),
    [anon_sym_short] = ACTIONS(2454),
    [anon_sym_static] = ACTIONS(2454),
    [anon_sym_ATautoreleasepool] = ACTIONS(2456),
    [anon_sym_auto] = ACTIONS(2454),
    [anon_sym_register] = ACTIONS(2454),
    [anon_sym_inline] = ACTIONS(2454),
    [anon_sym___inline] = ACTIONS(2454),
    [anon_sym___inline__] = ACTIONS(2454),
    [anon_sym___forceinline] = ACTIONS(2454),
    [anon_sym_thread_local] = ACTIONS(2454),
    [anon_sym___thread] = ACTIONS(2454),
    [anon_sym_CG_EXTERN] = ACTIONS(2454),
    [anon_sym_CG_INLINE] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2454),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2454),
    [anon_sym_IBOutlet] = ACTIONS(2454),
    [anon_sym_IBInspectable] = ACTIONS(2454),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2454),
    [anon_sym_NS_INLINE] = ACTIONS(2454),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2454),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2454),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2454),
    [anon_sym_const] = ACTIONS(2454),
    [anon_sym_constexpr] = ACTIONS(2454),
    [anon_sym_volatile] = ACTIONS(2454),
    [anon_sym_restrict] = ACTIONS(2454),
    [anon_sym___restrict__] = ACTIONS(2454),
    [anon_sym__Atomic] = ACTIONS(2454),
    [anon_sym__Noreturn] = ACTIONS(2454),
    [anon_sym__Nonnull] = ACTIONS(2454),
    [anon_sym_nullable] = ACTIONS(2454),
    [anon_sym__Complex] = ACTIONS(2454),
    [anon_sym__Nullable] = ACTIONS(2454),
    [anon_sym__Nullable_result] = ACTIONS(2454),
    [anon_sym__Null_unspecified] = ACTIONS(2454),
    [anon_sym___autoreleasing] = ACTIONS(2454),
    [anon_sym___block] = ACTIONS(2454),
    [anon_sym___bridge] = ACTIONS(2454),
    [anon_sym___bridge_retained] = ACTIONS(2454),
    [anon_sym___bridge_transfer] = ACTIONS(2454),
    [anon_sym___complex] = ACTIONS(2454),
    [anon_sym___const] = ACTIONS(2454),
    [anon_sym___imag] = ACTIONS(2454),
    [anon_sym___kindof] = ACTIONS(2454),
    [anon_sym___nonnull] = ACTIONS(2454),
    [anon_sym___nullable] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2454),
    [anon_sym___real] = ACTIONS(2454),
    [anon_sym___strong] = ACTIONS(2454),
    [anon_sym___unsafe_unretained] = ACTIONS(2454),
    [anon_sym___unused] = ACTIONS(2454),
    [anon_sym___weak] = ACTIONS(2454),
    [anon_sym_alignas] = ACTIONS(2454),
    [anon_sym__Alignas] = ACTIONS(2454),
    [sym_primitive_type] = ACTIONS(2454),
    [anon_sym_enum] = ACTIONS(2454),
    [anon_sym_struct] = ACTIONS(2454),
    [anon_sym_union] = ACTIONS(2454),
    [anon_sym_if] = ACTIONS(2454),
    [anon_sym_else] = ACTIONS(2454),
    [anon_sym_switch] = ACTIONS(2454),
    [anon_sym_case] = ACTIONS(2454),
    [anon_sym_default] = ACTIONS(2454),
    [anon_sym_while] = ACTIONS(2454),
    [anon_sym_do] = ACTIONS(2454),
    [anon_sym_for] = ACTIONS(2454),
    [anon_sym_in] = ACTIONS(2454),
    [anon_sym_return] = ACTIONS(2454),
    [anon_sym_break] = ACTIONS(2454),
    [anon_sym_continue] = ACTIONS(2454),
    [anon_sym_goto] = ACTIONS(2454),
    [anon_sym_DASH_DASH] = ACTIONS(2456),
    [anon_sym_PLUS_PLUS] = ACTIONS(2456),
    [anon_sym_sizeof] = ACTIONS(2454),
    [anon_sym___alignof__] = ACTIONS(2454),
    [anon_sym___alignof] = ACTIONS(2454),
    [anon_sym__alignof] = ACTIONS(2454),
    [anon_sym_alignof] = ACTIONS(2454),
    [anon_sym__Alignof] = ACTIONS(2454),
    [anon_sym_offsetof] = ACTIONS(2454),
    [anon_sym__Generic] = ACTIONS(2454),
    [anon_sym_asm] = ACTIONS(2454),
    [anon_sym___asm__] = ACTIONS(2454),
    [anon_sym___asm] = ACTIONS(2454),
    [sym_number_literal] = ACTIONS(2456),
    [anon_sym_L_SQUOTE] = ACTIONS(2456),
    [anon_sym_u_SQUOTE] = ACTIONS(2456),
    [anon_sym_U_SQUOTE] = ACTIONS(2456),
    [anon_sym_u8_SQUOTE] = ACTIONS(2456),
    [anon_sym_SQUOTE] = ACTIONS(2456),
    [anon_sym_AT] = ACTIONS(2454),
    [anon_sym_DQUOTE] = ACTIONS(2456),
    [anon_sym_L_DQUOTE] = ACTIONS(2456),
    [anon_sym_u_DQUOTE] = ACTIONS(2456),
    [anon_sym_U_DQUOTE] = ACTIONS(2456),
    [anon_sym_u8_DQUOTE] = ACTIONS(2456),
    [sym_true] = ACTIONS(2454),
    [sym_false] = ACTIONS(2454),
    [anon_sym_NULL] = ACTIONS(2454),
    [anon_sym_nullptr] = ACTIONS(2454),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2454),
    [anon_sym___typeof] = ACTIONS(2454),
    [anon_sym_typeof] = ACTIONS(2454),
    [anon_sym_ATimport] = ACTIONS(2456),
    [aux_sym_preproc_undef_token1] = ACTIONS(2454),
    [anon_sym_POUND] = ACTIONS(2454),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2454),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2454),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2454),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_API_AVAILABLE] = ACTIONS(2454),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_API_DEPRECATED] = ACTIONS(2454),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym___deprecated_msg] = ACTIONS(2454),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2454),
    [anon_sym_ATprotocol] = ACTIONS(2456),
    [anon_sym_ATinterface] = ACTIONS(2456),
    [anon_sym_ATimplementation] = ACTIONS(2456),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2456),
    [anon_sym_ATsynthesize] = ACTIONS(2456),
    [anon_sym_ATdynamic] = ACTIONS(2456),
    [anon_sym_ATtry] = ACTIONS(2456),
    [anon_sym___try] = ACTIONS(2454),
    [anon_sym_ATthrow] = ACTIONS(2456),
    [anon_sym_ATselector] = ACTIONS(2456),
    [anon_sym_ATavailable] = ACTIONS(2456),
    [anon_sym___builtin_available] = ACTIONS(2454),
    [anon_sym_va_arg] = ACTIONS(2454),
    [anon_sym_ATencode] = ACTIONS(2456),
    [anon_sym_ATsynchronized] = ACTIONS(2456),
    [anon_sym_BOOL] = ACTIONS(2454),
    [anon_sym_IMP] = ACTIONS(2454),
    [anon_sym_SEL] = ACTIONS(2454),
    [anon_sym_Class] = ACTIONS(2454),
    [anon_sym_id] = ACTIONS(2454),
  },
  [STATE(536)] = {
    [sym_identifier] = ACTIONS(2526),
    [aux_sym_preproc_include_token1] = ACTIONS(2526),
    [aux_sym_preproc_include_token2] = ACTIONS(2526),
    [aux_sym_preproc_def_token1] = ACTIONS(2526),
    [aux_sym_preproc_if_token1] = ACTIONS(2526),
    [aux_sym_preproc_if_token2] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2526),
    [sym_preproc_directive] = ACTIONS(2526),
    [anon_sym_LPAREN2] = ACTIONS(2528),
    [anon_sym_BANG] = ACTIONS(2528),
    [anon_sym_TILDE] = ACTIONS(2528),
    [anon_sym_DASH] = ACTIONS(2526),
    [anon_sym_PLUS] = ACTIONS(2526),
    [anon_sym_STAR] = ACTIONS(2528),
    [anon_sym_CARET] = ACTIONS(2528),
    [anon_sym_AMP] = ACTIONS(2528),
    [anon_sym_SEMI] = ACTIONS(2528),
    [anon_sym___extension__] = ACTIONS(2526),
    [anon_sym_typedef] = ACTIONS(2526),
    [anon_sym_extern] = ACTIONS(2526),
    [anon_sym___attribute__] = ACTIONS(2526),
    [anon_sym___attribute] = ACTIONS(2526),
    [anon_sym_noreturn] = ACTIONS(2526),
    [anon_sym_LBRACK] = ACTIONS(2528),
    [anon_sym___declspec] = ACTIONS(2526),
    [anon_sym___cdecl] = ACTIONS(2526),
    [anon_sym___clrcall] = ACTIONS(2526),
    [anon_sym___stdcall] = ACTIONS(2526),
    [anon_sym___fastcall] = ACTIONS(2526),
    [anon_sym___thiscall] = ACTIONS(2526),
    [anon_sym___vectorcall] = ACTIONS(2526),
    [anon_sym_LBRACE] = ACTIONS(2528),
    [anon_sym_signed] = ACTIONS(2526),
    [anon_sym_unsigned] = ACTIONS(2526),
    [anon_sym_long] = ACTIONS(2526),
    [anon_sym_short] = ACTIONS(2526),
    [anon_sym_static] = ACTIONS(2526),
    [anon_sym_ATautoreleasepool] = ACTIONS(2528),
    [anon_sym_auto] = ACTIONS(2526),
    [anon_sym_register] = ACTIONS(2526),
    [anon_sym_inline] = ACTIONS(2526),
    [anon_sym___inline] = ACTIONS(2526),
    [anon_sym___inline__] = ACTIONS(2526),
    [anon_sym___forceinline] = ACTIONS(2526),
    [anon_sym_thread_local] = ACTIONS(2526),
    [anon_sym___thread] = ACTIONS(2526),
    [anon_sym_CG_EXTERN] = ACTIONS(2526),
    [anon_sym_CG_INLINE] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2526),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2526),
    [anon_sym_IBOutlet] = ACTIONS(2526),
    [anon_sym_IBInspectable] = ACTIONS(2526),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2526),
    [anon_sym_NS_INLINE] = ACTIONS(2526),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2526),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2526),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2526),
    [anon_sym_const] = ACTIONS(2526),
    [anon_sym_constexpr] = ACTIONS(2526),
    [anon_sym_volatile] = ACTIONS(2526),
    [anon_sym_restrict] = ACTIONS(2526),
    [anon_sym___restrict__] = ACTIONS(2526),
    [anon_sym__Atomic] = ACTIONS(2526),
    [anon_sym__Noreturn] = ACTIONS(2526),
    [anon_sym__Nonnull] = ACTIONS(2526),
    [anon_sym_nullable] = ACTIONS(2526),
    [anon_sym__Complex] = ACTIONS(2526),
    [anon_sym__Nullable] = ACTIONS(2526),
    [anon_sym__Nullable_result] = ACTIONS(2526),
    [anon_sym__Null_unspecified] = ACTIONS(2526),
    [anon_sym___autoreleasing] = ACTIONS(2526),
    [anon_sym___block] = ACTIONS(2526),
    [anon_sym___bridge] = ACTIONS(2526),
    [anon_sym___bridge_retained] = ACTIONS(2526),
    [anon_sym___bridge_transfer] = ACTIONS(2526),
    [anon_sym___complex] = ACTIONS(2526),
    [anon_sym___const] = ACTIONS(2526),
    [anon_sym___imag] = ACTIONS(2526),
    [anon_sym___kindof] = ACTIONS(2526),
    [anon_sym___nonnull] = ACTIONS(2526),
    [anon_sym___nullable] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2526),
    [anon_sym___real] = ACTIONS(2526),
    [anon_sym___strong] = ACTIONS(2526),
    [anon_sym___unsafe_unretained] = ACTIONS(2526),
    [anon_sym___unused] = ACTIONS(2526),
    [anon_sym___weak] = ACTIONS(2526),
    [anon_sym_alignas] = ACTIONS(2526),
    [anon_sym__Alignas] = ACTIONS(2526),
    [sym_primitive_type] = ACTIONS(2526),
    [anon_sym_enum] = ACTIONS(2526),
    [anon_sym_struct] = ACTIONS(2526),
    [anon_sym_union] = ACTIONS(2526),
    [anon_sym_if] = ACTIONS(2526),
    [anon_sym_else] = ACTIONS(2526),
    [anon_sym_switch] = ACTIONS(2526),
    [anon_sym_case] = ACTIONS(2526),
    [anon_sym_default] = ACTIONS(2526),
    [anon_sym_while] = ACTIONS(2526),
    [anon_sym_do] = ACTIONS(2526),
    [anon_sym_for] = ACTIONS(2526),
    [anon_sym_in] = ACTIONS(2526),
    [anon_sym_return] = ACTIONS(2526),
    [anon_sym_break] = ACTIONS(2526),
    [anon_sym_continue] = ACTIONS(2526),
    [anon_sym_goto] = ACTIONS(2526),
    [anon_sym_DASH_DASH] = ACTIONS(2528),
    [anon_sym_PLUS_PLUS] = ACTIONS(2528),
    [anon_sym_sizeof] = ACTIONS(2526),
    [anon_sym___alignof__] = ACTIONS(2526),
    [anon_sym___alignof] = ACTIONS(2526),
    [anon_sym__alignof] = ACTIONS(2526),
    [anon_sym_alignof] = ACTIONS(2526),
    [anon_sym__Alignof] = ACTIONS(2526),
    [anon_sym_offsetof] = ACTIONS(2526),
    [anon_sym__Generic] = ACTIONS(2526),
    [anon_sym_asm] = ACTIONS(2526),
    [anon_sym___asm__] = ACTIONS(2526),
    [anon_sym___asm] = ACTIONS(2526),
    [sym_number_literal] = ACTIONS(2528),
    [anon_sym_L_SQUOTE] = ACTIONS(2528),
    [anon_sym_u_SQUOTE] = ACTIONS(2528),
    [anon_sym_U_SQUOTE] = ACTIONS(2528),
    [anon_sym_u8_SQUOTE] = ACTIONS(2528),
    [anon_sym_SQUOTE] = ACTIONS(2528),
    [anon_sym_AT] = ACTIONS(2526),
    [anon_sym_DQUOTE] = ACTIONS(2528),
    [anon_sym_L_DQUOTE] = ACTIONS(2528),
    [anon_sym_u_DQUOTE] = ACTIONS(2528),
    [anon_sym_U_DQUOTE] = ACTIONS(2528),
    [anon_sym_u8_DQUOTE] = ACTIONS(2528),
    [sym_true] = ACTIONS(2526),
    [sym_false] = ACTIONS(2526),
    [anon_sym_NULL] = ACTIONS(2526),
    [anon_sym_nullptr] = ACTIONS(2526),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2526),
    [anon_sym___typeof] = ACTIONS(2526),
    [anon_sym_typeof] = ACTIONS(2526),
    [anon_sym_ATimport] = ACTIONS(2528),
    [aux_sym_preproc_undef_token1] = ACTIONS(2526),
    [anon_sym_POUND] = ACTIONS(2526),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2526),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2526),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2526),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_API_AVAILABLE] = ACTIONS(2526),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_API_DEPRECATED] = ACTIONS(2526),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym___deprecated_msg] = ACTIONS(2526),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2526),
    [anon_sym_ATprotocol] = ACTIONS(2528),
    [anon_sym_ATinterface] = ACTIONS(2528),
    [anon_sym_ATimplementation] = ACTIONS(2528),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2528),
    [anon_sym_ATsynthesize] = ACTIONS(2528),
    [anon_sym_ATdynamic] = ACTIONS(2528),
    [anon_sym_ATtry] = ACTIONS(2528),
    [anon_sym___try] = ACTIONS(2526),
    [anon_sym_ATthrow] = ACTIONS(2528),
    [anon_sym_ATselector] = ACTIONS(2528),
    [anon_sym_ATavailable] = ACTIONS(2528),
    [anon_sym___builtin_available] = ACTIONS(2526),
    [anon_sym_va_arg] = ACTIONS(2526),
    [anon_sym_ATencode] = ACTIONS(2528),
    [anon_sym_ATsynchronized] = ACTIONS(2528),
    [anon_sym_BOOL] = ACTIONS(2526),
    [anon_sym_IMP] = ACTIONS(2526),
    [anon_sym_SEL] = ACTIONS(2526),
    [anon_sym_Class] = ACTIONS(2526),
    [anon_sym_id] = ACTIONS(2526),
  },
  [STATE(537)] = {
    [sym_identifier] = ACTIONS(2530),
    [aux_sym_preproc_include_token1] = ACTIONS(2530),
    [aux_sym_preproc_include_token2] = ACTIONS(2530),
    [aux_sym_preproc_def_token1] = ACTIONS(2530),
    [aux_sym_preproc_if_token1] = ACTIONS(2530),
    [aux_sym_preproc_if_token2] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2530),
    [sym_preproc_directive] = ACTIONS(2530),
    [anon_sym_LPAREN2] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2530),
    [anon_sym_PLUS] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym___extension__] = ACTIONS(2530),
    [anon_sym_typedef] = ACTIONS(2530),
    [anon_sym_extern] = ACTIONS(2530),
    [anon_sym___attribute__] = ACTIONS(2530),
    [anon_sym___attribute] = ACTIONS(2530),
    [anon_sym_noreturn] = ACTIONS(2530),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym___declspec] = ACTIONS(2530),
    [anon_sym___cdecl] = ACTIONS(2530),
    [anon_sym___clrcall] = ACTIONS(2530),
    [anon_sym___stdcall] = ACTIONS(2530),
    [anon_sym___fastcall] = ACTIONS(2530),
    [anon_sym___thiscall] = ACTIONS(2530),
    [anon_sym___vectorcall] = ACTIONS(2530),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_signed] = ACTIONS(2530),
    [anon_sym_unsigned] = ACTIONS(2530),
    [anon_sym_long] = ACTIONS(2530),
    [anon_sym_short] = ACTIONS(2530),
    [anon_sym_static] = ACTIONS(2530),
    [anon_sym_ATautoreleasepool] = ACTIONS(2532),
    [anon_sym_auto] = ACTIONS(2530),
    [anon_sym_register] = ACTIONS(2530),
    [anon_sym_inline] = ACTIONS(2530),
    [anon_sym___inline] = ACTIONS(2530),
    [anon_sym___inline__] = ACTIONS(2530),
    [anon_sym___forceinline] = ACTIONS(2530),
    [anon_sym_thread_local] = ACTIONS(2530),
    [anon_sym___thread] = ACTIONS(2530),
    [anon_sym_CG_EXTERN] = ACTIONS(2530),
    [anon_sym_CG_INLINE] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2530),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2530),
    [anon_sym_IBOutlet] = ACTIONS(2530),
    [anon_sym_IBInspectable] = ACTIONS(2530),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2530),
    [anon_sym_NS_INLINE] = ACTIONS(2530),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2530),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2530),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2530),
    [anon_sym_const] = ACTIONS(2530),
    [anon_sym_constexpr] = ACTIONS(2530),
    [anon_sym_volatile] = ACTIONS(2530),
    [anon_sym_restrict] = ACTIONS(2530),
    [anon_sym___restrict__] = ACTIONS(2530),
    [anon_sym__Atomic] = ACTIONS(2530),
    [anon_sym__Noreturn] = ACTIONS(2530),
    [anon_sym__Nonnull] = ACTIONS(2530),
    [anon_sym_nullable] = ACTIONS(2530),
    [anon_sym__Complex] = ACTIONS(2530),
    [anon_sym__Nullable] = ACTIONS(2530),
    [anon_sym__Nullable_result] = ACTIONS(2530),
    [anon_sym__Null_unspecified] = ACTIONS(2530),
    [anon_sym___autoreleasing] = ACTIONS(2530),
    [anon_sym___block] = ACTIONS(2530),
    [anon_sym___bridge] = ACTIONS(2530),
    [anon_sym___bridge_retained] = ACTIONS(2530),
    [anon_sym___bridge_transfer] = ACTIONS(2530),
    [anon_sym___complex] = ACTIONS(2530),
    [anon_sym___const] = ACTIONS(2530),
    [anon_sym___imag] = ACTIONS(2530),
    [anon_sym___kindof] = ACTIONS(2530),
    [anon_sym___nonnull] = ACTIONS(2530),
    [anon_sym___nullable] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2530),
    [anon_sym___real] = ACTIONS(2530),
    [anon_sym___strong] = ACTIONS(2530),
    [anon_sym___unsafe_unretained] = ACTIONS(2530),
    [anon_sym___unused] = ACTIONS(2530),
    [anon_sym___weak] = ACTIONS(2530),
    [anon_sym_alignas] = ACTIONS(2530),
    [anon_sym__Alignas] = ACTIONS(2530),
    [sym_primitive_type] = ACTIONS(2530),
    [anon_sym_enum] = ACTIONS(2530),
    [anon_sym_struct] = ACTIONS(2530),
    [anon_sym_union] = ACTIONS(2530),
    [anon_sym_if] = ACTIONS(2530),
    [anon_sym_else] = ACTIONS(2530),
    [anon_sym_switch] = ACTIONS(2530),
    [anon_sym_case] = ACTIONS(2530),
    [anon_sym_default] = ACTIONS(2530),
    [anon_sym_while] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2530),
    [anon_sym_for] = ACTIONS(2530),
    [anon_sym_in] = ACTIONS(2530),
    [anon_sym_return] = ACTIONS(2530),
    [anon_sym_break] = ACTIONS(2530),
    [anon_sym_continue] = ACTIONS(2530),
    [anon_sym_goto] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_sizeof] = ACTIONS(2530),
    [anon_sym___alignof__] = ACTIONS(2530),
    [anon_sym___alignof] = ACTIONS(2530),
    [anon_sym__alignof] = ACTIONS(2530),
    [anon_sym_alignof] = ACTIONS(2530),
    [anon_sym__Alignof] = ACTIONS(2530),
    [anon_sym_offsetof] = ACTIONS(2530),
    [anon_sym__Generic] = ACTIONS(2530),
    [anon_sym_asm] = ACTIONS(2530),
    [anon_sym___asm__] = ACTIONS(2530),
    [anon_sym___asm] = ACTIONS(2530),
    [sym_number_literal] = ACTIONS(2532),
    [anon_sym_L_SQUOTE] = ACTIONS(2532),
    [anon_sym_u_SQUOTE] = ACTIONS(2532),
    [anon_sym_U_SQUOTE] = ACTIONS(2532),
    [anon_sym_u8_SQUOTE] = ACTIONS(2532),
    [anon_sym_SQUOTE] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2530),
    [anon_sym_DQUOTE] = ACTIONS(2532),
    [anon_sym_L_DQUOTE] = ACTIONS(2532),
    [anon_sym_u_DQUOTE] = ACTIONS(2532),
    [anon_sym_U_DQUOTE] = ACTIONS(2532),
    [anon_sym_u8_DQUOTE] = ACTIONS(2532),
    [sym_true] = ACTIONS(2530),
    [sym_false] = ACTIONS(2530),
    [anon_sym_NULL] = ACTIONS(2530),
    [anon_sym_nullptr] = ACTIONS(2530),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2530),
    [anon_sym___typeof] = ACTIONS(2530),
    [anon_sym_typeof] = ACTIONS(2530),
    [anon_sym_ATimport] = ACTIONS(2532),
    [aux_sym_preproc_undef_token1] = ACTIONS(2530),
    [anon_sym_POUND] = ACTIONS(2530),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2530),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2530),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2530),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_API_AVAILABLE] = ACTIONS(2530),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_API_DEPRECATED] = ACTIONS(2530),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym___deprecated_msg] = ACTIONS(2530),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2530),
    [anon_sym_ATprotocol] = ACTIONS(2532),
    [anon_sym_ATinterface] = ACTIONS(2532),
    [anon_sym_ATimplementation] = ACTIONS(2532),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2532),
    [anon_sym_ATsynthesize] = ACTIONS(2532),
    [anon_sym_ATdynamic] = ACTIONS(2532),
    [anon_sym_ATtry] = ACTIONS(2532),
    [anon_sym___try] = ACTIONS(2530),
    [anon_sym_ATthrow] = ACTIONS(2532),
    [anon_sym_ATselector] = ACTIONS(2532),
    [anon_sym_ATavailable] = ACTIONS(2532),
    [anon_sym___builtin_available] = ACTIONS(2530),
    [anon_sym_va_arg] = ACTIONS(2530),
    [anon_sym_ATencode] = ACTIONS(2532),
    [anon_sym_ATsynchronized] = ACTIONS(2532),
    [anon_sym_BOOL] = ACTIONS(2530),
    [anon_sym_IMP] = ACTIONS(2530),
    [anon_sym_SEL] = ACTIONS(2530),
    [anon_sym_Class] = ACTIONS(2530),
    [anon_sym_id] = ACTIONS(2530),
  },
  [STATE(538)] = {
    [sym_identifier] = ACTIONS(2534),
    [aux_sym_preproc_include_token1] = ACTIONS(2534),
    [aux_sym_preproc_include_token2] = ACTIONS(2534),
    [aux_sym_preproc_def_token1] = ACTIONS(2534),
    [aux_sym_preproc_if_token1] = ACTIONS(2534),
    [aux_sym_preproc_if_token2] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2534),
    [sym_preproc_directive] = ACTIONS(2534),
    [anon_sym_LPAREN2] = ACTIONS(2536),
    [anon_sym_BANG] = ACTIONS(2536),
    [anon_sym_TILDE] = ACTIONS(2536),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2536),
    [anon_sym_CARET] = ACTIONS(2536),
    [anon_sym_AMP] = ACTIONS(2536),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym___extension__] = ACTIONS(2534),
    [anon_sym_typedef] = ACTIONS(2534),
    [anon_sym_extern] = ACTIONS(2534),
    [anon_sym___attribute__] = ACTIONS(2534),
    [anon_sym___attribute] = ACTIONS(2534),
    [anon_sym_noreturn] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2536),
    [anon_sym___declspec] = ACTIONS(2534),
    [anon_sym___cdecl] = ACTIONS(2534),
    [anon_sym___clrcall] = ACTIONS(2534),
    [anon_sym___stdcall] = ACTIONS(2534),
    [anon_sym___fastcall] = ACTIONS(2534),
    [anon_sym___thiscall] = ACTIONS(2534),
    [anon_sym___vectorcall] = ACTIONS(2534),
    [anon_sym_LBRACE] = ACTIONS(2536),
    [anon_sym_signed] = ACTIONS(2534),
    [anon_sym_unsigned] = ACTIONS(2534),
    [anon_sym_long] = ACTIONS(2534),
    [anon_sym_short] = ACTIONS(2534),
    [anon_sym_static] = ACTIONS(2534),
    [anon_sym_ATautoreleasepool] = ACTIONS(2536),
    [anon_sym_auto] = ACTIONS(2534),
    [anon_sym_register] = ACTIONS(2534),
    [anon_sym_inline] = ACTIONS(2534),
    [anon_sym___inline] = ACTIONS(2534),
    [anon_sym___inline__] = ACTIONS(2534),
    [anon_sym___forceinline] = ACTIONS(2534),
    [anon_sym_thread_local] = ACTIONS(2534),
    [anon_sym___thread] = ACTIONS(2534),
    [anon_sym_CG_EXTERN] = ACTIONS(2534),
    [anon_sym_CG_INLINE] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2534),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2534),
    [anon_sym_IBOutlet] = ACTIONS(2534),
    [anon_sym_IBInspectable] = ACTIONS(2534),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2534),
    [anon_sym_NS_INLINE] = ACTIONS(2534),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2534),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2534),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2534),
    [anon_sym_const] = ACTIONS(2534),
    [anon_sym_constexpr] = ACTIONS(2534),
    [anon_sym_volatile] = ACTIONS(2534),
    [anon_sym_restrict] = ACTIONS(2534),
    [anon_sym___restrict__] = ACTIONS(2534),
    [anon_sym__Atomic] = ACTIONS(2534),
    [anon_sym__Noreturn] = ACTIONS(2534),
    [anon_sym__Nonnull] = ACTIONS(2534),
    [anon_sym_nullable] = ACTIONS(2534),
    [anon_sym__Complex] = ACTIONS(2534),
    [anon_sym__Nullable] = ACTIONS(2534),
    [anon_sym__Nullable_result] = ACTIONS(2534),
    [anon_sym__Null_unspecified] = ACTIONS(2534),
    [anon_sym___autoreleasing] = ACTIONS(2534),
    [anon_sym___block] = ACTIONS(2534),
    [anon_sym___bridge] = ACTIONS(2534),
    [anon_sym___bridge_retained] = ACTIONS(2534),
    [anon_sym___bridge_transfer] = ACTIONS(2534),
    [anon_sym___complex] = ACTIONS(2534),
    [anon_sym___const] = ACTIONS(2534),
    [anon_sym___imag] = ACTIONS(2534),
    [anon_sym___kindof] = ACTIONS(2534),
    [anon_sym___nonnull] = ACTIONS(2534),
    [anon_sym___nullable] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2534),
    [anon_sym___real] = ACTIONS(2534),
    [anon_sym___strong] = ACTIONS(2534),
    [anon_sym___unsafe_unretained] = ACTIONS(2534),
    [anon_sym___unused] = ACTIONS(2534),
    [anon_sym___weak] = ACTIONS(2534),
    [anon_sym_alignas] = ACTIONS(2534),
    [anon_sym__Alignas] = ACTIONS(2534),
    [sym_primitive_type] = ACTIONS(2534),
    [anon_sym_enum] = ACTIONS(2534),
    [anon_sym_struct] = ACTIONS(2534),
    [anon_sym_union] = ACTIONS(2534),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_else] = ACTIONS(2534),
    [anon_sym_switch] = ACTIONS(2534),
    [anon_sym_case] = ACTIONS(2534),
    [anon_sym_default] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_do] = ACTIONS(2534),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_in] = ACTIONS(2534),
    [anon_sym_return] = ACTIONS(2534),
    [anon_sym_break] = ACTIONS(2534),
    [anon_sym_continue] = ACTIONS(2534),
    [anon_sym_goto] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_sizeof] = ACTIONS(2534),
    [anon_sym___alignof__] = ACTIONS(2534),
    [anon_sym___alignof] = ACTIONS(2534),
    [anon_sym__alignof] = ACTIONS(2534),
    [anon_sym_alignof] = ACTIONS(2534),
    [anon_sym__Alignof] = ACTIONS(2534),
    [anon_sym_offsetof] = ACTIONS(2534),
    [anon_sym__Generic] = ACTIONS(2534),
    [anon_sym_asm] = ACTIONS(2534),
    [anon_sym___asm__] = ACTIONS(2534),
    [anon_sym___asm] = ACTIONS(2534),
    [sym_number_literal] = ACTIONS(2536),
    [anon_sym_L_SQUOTE] = ACTIONS(2536),
    [anon_sym_u_SQUOTE] = ACTIONS(2536),
    [anon_sym_U_SQUOTE] = ACTIONS(2536),
    [anon_sym_u8_SQUOTE] = ACTIONS(2536),
    [anon_sym_SQUOTE] = ACTIONS(2536),
    [anon_sym_AT] = ACTIONS(2534),
    [anon_sym_DQUOTE] = ACTIONS(2536),
    [anon_sym_L_DQUOTE] = ACTIONS(2536),
    [anon_sym_u_DQUOTE] = ACTIONS(2536),
    [anon_sym_U_DQUOTE] = ACTIONS(2536),
    [anon_sym_u8_DQUOTE] = ACTIONS(2536),
    [sym_true] = ACTIONS(2534),
    [sym_false] = ACTIONS(2534),
    [anon_sym_NULL] = ACTIONS(2534),
    [anon_sym_nullptr] = ACTIONS(2534),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2534),
    [anon_sym___typeof] = ACTIONS(2534),
    [anon_sym_typeof] = ACTIONS(2534),
    [anon_sym_ATimport] = ACTIONS(2536),
    [aux_sym_preproc_undef_token1] = ACTIONS(2534),
    [anon_sym_POUND] = ACTIONS(2534),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2534),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2534),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2534),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_API_AVAILABLE] = ACTIONS(2534),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_API_DEPRECATED] = ACTIONS(2534),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym___deprecated_msg] = ACTIONS(2534),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2534),
    [anon_sym_ATprotocol] = ACTIONS(2536),
    [anon_sym_ATinterface] = ACTIONS(2536),
    [anon_sym_ATimplementation] = ACTIONS(2536),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2536),
    [anon_sym_ATsynthesize] = ACTIONS(2536),
    [anon_sym_ATdynamic] = ACTIONS(2536),
    [anon_sym_ATtry] = ACTIONS(2536),
    [anon_sym___try] = ACTIONS(2534),
    [anon_sym_ATthrow] = ACTIONS(2536),
    [anon_sym_ATselector] = ACTIONS(2536),
    [anon_sym_ATavailable] = ACTIONS(2536),
    [anon_sym___builtin_available] = ACTIONS(2534),
    [anon_sym_va_arg] = ACTIONS(2534),
    [anon_sym_ATencode] = ACTIONS(2536),
    [anon_sym_ATsynchronized] = ACTIONS(2536),
    [anon_sym_BOOL] = ACTIONS(2534),
    [anon_sym_IMP] = ACTIONS(2534),
    [anon_sym_SEL] = ACTIONS(2534),
    [anon_sym_Class] = ACTIONS(2534),
    [anon_sym_id] = ACTIONS(2534),
  },
  [STATE(539)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token2] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(540)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token2] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATsynthesize] = ACTIONS(2362),
    [anon_sym_ATdynamic] = ACTIONS(2362),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(541)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token2] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(542)] = {
    [sym_identifier] = ACTIONS(2538),
    [aux_sym_preproc_include_token1] = ACTIONS(2538),
    [aux_sym_preproc_include_token2] = ACTIONS(2538),
    [aux_sym_preproc_def_token1] = ACTIONS(2538),
    [aux_sym_preproc_if_token1] = ACTIONS(2538),
    [aux_sym_preproc_if_token2] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2538),
    [sym_preproc_directive] = ACTIONS(2538),
    [anon_sym_LPAREN2] = ACTIONS(2540),
    [anon_sym_BANG] = ACTIONS(2540),
    [anon_sym_TILDE] = ACTIONS(2540),
    [anon_sym_DASH] = ACTIONS(2538),
    [anon_sym_PLUS] = ACTIONS(2538),
    [anon_sym_STAR] = ACTIONS(2540),
    [anon_sym_CARET] = ACTIONS(2540),
    [anon_sym_AMP] = ACTIONS(2540),
    [anon_sym_SEMI] = ACTIONS(2540),
    [anon_sym___extension__] = ACTIONS(2538),
    [anon_sym_typedef] = ACTIONS(2538),
    [anon_sym_extern] = ACTIONS(2538),
    [anon_sym___attribute__] = ACTIONS(2538),
    [anon_sym___attribute] = ACTIONS(2538),
    [anon_sym_noreturn] = ACTIONS(2538),
    [anon_sym_LBRACK] = ACTIONS(2540),
    [anon_sym___declspec] = ACTIONS(2538),
    [anon_sym___cdecl] = ACTIONS(2538),
    [anon_sym___clrcall] = ACTIONS(2538),
    [anon_sym___stdcall] = ACTIONS(2538),
    [anon_sym___fastcall] = ACTIONS(2538),
    [anon_sym___thiscall] = ACTIONS(2538),
    [anon_sym___vectorcall] = ACTIONS(2538),
    [anon_sym_LBRACE] = ACTIONS(2540),
    [anon_sym_signed] = ACTIONS(2538),
    [anon_sym_unsigned] = ACTIONS(2538),
    [anon_sym_long] = ACTIONS(2538),
    [anon_sym_short] = ACTIONS(2538),
    [anon_sym_static] = ACTIONS(2538),
    [anon_sym_ATautoreleasepool] = ACTIONS(2540),
    [anon_sym_auto] = ACTIONS(2538),
    [anon_sym_register] = ACTIONS(2538),
    [anon_sym_inline] = ACTIONS(2538),
    [anon_sym___inline] = ACTIONS(2538),
    [anon_sym___inline__] = ACTIONS(2538),
    [anon_sym___forceinline] = ACTIONS(2538),
    [anon_sym_thread_local] = ACTIONS(2538),
    [anon_sym___thread] = ACTIONS(2538),
    [anon_sym_CG_EXTERN] = ACTIONS(2538),
    [anon_sym_CG_INLINE] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2538),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2538),
    [anon_sym_IBOutlet] = ACTIONS(2538),
    [anon_sym_IBInspectable] = ACTIONS(2538),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2538),
    [anon_sym_NS_INLINE] = ACTIONS(2538),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2538),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2538),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2538),
    [anon_sym_const] = ACTIONS(2538),
    [anon_sym_constexpr] = ACTIONS(2538),
    [anon_sym_volatile] = ACTIONS(2538),
    [anon_sym_restrict] = ACTIONS(2538),
    [anon_sym___restrict__] = ACTIONS(2538),
    [anon_sym__Atomic] = ACTIONS(2538),
    [anon_sym__Noreturn] = ACTIONS(2538),
    [anon_sym__Nonnull] = ACTIONS(2538),
    [anon_sym_nullable] = ACTIONS(2538),
    [anon_sym__Complex] = ACTIONS(2538),
    [anon_sym__Nullable] = ACTIONS(2538),
    [anon_sym__Nullable_result] = ACTIONS(2538),
    [anon_sym__Null_unspecified] = ACTIONS(2538),
    [anon_sym___autoreleasing] = ACTIONS(2538),
    [anon_sym___block] = ACTIONS(2538),
    [anon_sym___bridge] = ACTIONS(2538),
    [anon_sym___bridge_retained] = ACTIONS(2538),
    [anon_sym___bridge_transfer] = ACTIONS(2538),
    [anon_sym___complex] = ACTIONS(2538),
    [anon_sym___const] = ACTIONS(2538),
    [anon_sym___imag] = ACTIONS(2538),
    [anon_sym___kindof] = ACTIONS(2538),
    [anon_sym___nonnull] = ACTIONS(2538),
    [anon_sym___nullable] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2538),
    [anon_sym___real] = ACTIONS(2538),
    [anon_sym___strong] = ACTIONS(2538),
    [anon_sym___unsafe_unretained] = ACTIONS(2538),
    [anon_sym___unused] = ACTIONS(2538),
    [anon_sym___weak] = ACTIONS(2538),
    [anon_sym_alignas] = ACTIONS(2538),
    [anon_sym__Alignas] = ACTIONS(2538),
    [sym_primitive_type] = ACTIONS(2538),
    [anon_sym_enum] = ACTIONS(2538),
    [anon_sym_struct] = ACTIONS(2538),
    [anon_sym_union] = ACTIONS(2538),
    [anon_sym_if] = ACTIONS(2538),
    [anon_sym_else] = ACTIONS(2538),
    [anon_sym_switch] = ACTIONS(2538),
    [anon_sym_case] = ACTIONS(2538),
    [anon_sym_default] = ACTIONS(2538),
    [anon_sym_while] = ACTIONS(2538),
    [anon_sym_do] = ACTIONS(2538),
    [anon_sym_for] = ACTIONS(2538),
    [anon_sym_in] = ACTIONS(2538),
    [anon_sym_return] = ACTIONS(2538),
    [anon_sym_break] = ACTIONS(2538),
    [anon_sym_continue] = ACTIONS(2538),
    [anon_sym_goto] = ACTIONS(2538),
    [anon_sym_DASH_DASH] = ACTIONS(2540),
    [anon_sym_PLUS_PLUS] = ACTIONS(2540),
    [anon_sym_sizeof] = ACTIONS(2538),
    [anon_sym___alignof__] = ACTIONS(2538),
    [anon_sym___alignof] = ACTIONS(2538),
    [anon_sym__alignof] = ACTIONS(2538),
    [anon_sym_alignof] = ACTIONS(2538),
    [anon_sym__Alignof] = ACTIONS(2538),
    [anon_sym_offsetof] = ACTIONS(2538),
    [anon_sym__Generic] = ACTIONS(2538),
    [anon_sym_asm] = ACTIONS(2538),
    [anon_sym___asm__] = ACTIONS(2538),
    [anon_sym___asm] = ACTIONS(2538),
    [sym_number_literal] = ACTIONS(2540),
    [anon_sym_L_SQUOTE] = ACTIONS(2540),
    [anon_sym_u_SQUOTE] = ACTIONS(2540),
    [anon_sym_U_SQUOTE] = ACTIONS(2540),
    [anon_sym_u8_SQUOTE] = ACTIONS(2540),
    [anon_sym_SQUOTE] = ACTIONS(2540),
    [anon_sym_AT] = ACTIONS(2538),
    [anon_sym_DQUOTE] = ACTIONS(2540),
    [anon_sym_L_DQUOTE] = ACTIONS(2540),
    [anon_sym_u_DQUOTE] = ACTIONS(2540),
    [anon_sym_U_DQUOTE] = ACTIONS(2540),
    [anon_sym_u8_DQUOTE] = ACTIONS(2540),
    [sym_true] = ACTIONS(2538),
    [sym_false] = ACTIONS(2538),
    [anon_sym_NULL] = ACTIONS(2538),
    [anon_sym_nullptr] = ACTIONS(2538),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2538),
    [anon_sym___typeof] = ACTIONS(2538),
    [anon_sym_typeof] = ACTIONS(2538),
    [anon_sym_ATimport] = ACTIONS(2540),
    [aux_sym_preproc_undef_token1] = ACTIONS(2538),
    [anon_sym_POUND] = ACTIONS(2538),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2538),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2538),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2538),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_API_AVAILABLE] = ACTIONS(2538),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_API_DEPRECATED] = ACTIONS(2538),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym___deprecated_msg] = ACTIONS(2538),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2538),
    [anon_sym_ATprotocol] = ACTIONS(2540),
    [anon_sym_ATinterface] = ACTIONS(2540),
    [anon_sym_ATimplementation] = ACTIONS(2540),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2540),
    [anon_sym_ATsynthesize] = ACTIONS(2540),
    [anon_sym_ATdynamic] = ACTIONS(2540),
    [anon_sym_ATtry] = ACTIONS(2540),
    [anon_sym___try] = ACTIONS(2538),
    [anon_sym_ATthrow] = ACTIONS(2540),
    [anon_sym_ATselector] = ACTIONS(2540),
    [anon_sym_ATavailable] = ACTIONS(2540),
    [anon_sym___builtin_available] = ACTIONS(2538),
    [anon_sym_va_arg] = ACTIONS(2538),
    [anon_sym_ATencode] = ACTIONS(2540),
    [anon_sym_ATsynchronized] = ACTIONS(2540),
    [anon_sym_BOOL] = ACTIONS(2538),
    [anon_sym_IMP] = ACTIONS(2538),
    [anon_sym_SEL] = ACTIONS(2538),
    [anon_sym_Class] = ACTIONS(2538),
    [anon_sym_id] = ACTIONS(2538),
  },
  [STATE(543)] = {
    [sym_identifier] = ACTIONS(2450),
    [aux_sym_preproc_include_token1] = ACTIONS(2450),
    [aux_sym_preproc_include_token2] = ACTIONS(2450),
    [aux_sym_preproc_def_token1] = ACTIONS(2450),
    [aux_sym_preproc_if_token1] = ACTIONS(2450),
    [aux_sym_preproc_if_token2] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2450),
    [sym_preproc_directive] = ACTIONS(2450),
    [anon_sym_LPAREN2] = ACTIONS(2452),
    [anon_sym_BANG] = ACTIONS(2452),
    [anon_sym_TILDE] = ACTIONS(2452),
    [anon_sym_DASH] = ACTIONS(2450),
    [anon_sym_PLUS] = ACTIONS(2450),
    [anon_sym_STAR] = ACTIONS(2452),
    [anon_sym_CARET] = ACTIONS(2452),
    [anon_sym_AMP] = ACTIONS(2452),
    [anon_sym_SEMI] = ACTIONS(2452),
    [anon_sym___extension__] = ACTIONS(2450),
    [anon_sym_typedef] = ACTIONS(2450),
    [anon_sym_extern] = ACTIONS(2450),
    [anon_sym___attribute__] = ACTIONS(2450),
    [anon_sym___attribute] = ACTIONS(2450),
    [anon_sym_noreturn] = ACTIONS(2450),
    [anon_sym_LBRACK] = ACTIONS(2452),
    [anon_sym___declspec] = ACTIONS(2450),
    [anon_sym___cdecl] = ACTIONS(2450),
    [anon_sym___clrcall] = ACTIONS(2450),
    [anon_sym___stdcall] = ACTIONS(2450),
    [anon_sym___fastcall] = ACTIONS(2450),
    [anon_sym___thiscall] = ACTIONS(2450),
    [anon_sym___vectorcall] = ACTIONS(2450),
    [anon_sym_LBRACE] = ACTIONS(2452),
    [anon_sym_signed] = ACTIONS(2450),
    [anon_sym_unsigned] = ACTIONS(2450),
    [anon_sym_long] = ACTIONS(2450),
    [anon_sym_short] = ACTIONS(2450),
    [anon_sym_static] = ACTIONS(2450),
    [anon_sym_ATautoreleasepool] = ACTIONS(2452),
    [anon_sym_auto] = ACTIONS(2450),
    [anon_sym_register] = ACTIONS(2450),
    [anon_sym_inline] = ACTIONS(2450),
    [anon_sym___inline] = ACTIONS(2450),
    [anon_sym___inline__] = ACTIONS(2450),
    [anon_sym___forceinline] = ACTIONS(2450),
    [anon_sym_thread_local] = ACTIONS(2450),
    [anon_sym___thread] = ACTIONS(2450),
    [anon_sym_CG_EXTERN] = ACTIONS(2450),
    [anon_sym_CG_INLINE] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2450),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2450),
    [anon_sym_IBOutlet] = ACTIONS(2450),
    [anon_sym_IBInspectable] = ACTIONS(2450),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2450),
    [anon_sym_NS_INLINE] = ACTIONS(2450),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2450),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2450),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2450),
    [anon_sym_const] = ACTIONS(2450),
    [anon_sym_constexpr] = ACTIONS(2450),
    [anon_sym_volatile] = ACTIONS(2450),
    [anon_sym_restrict] = ACTIONS(2450),
    [anon_sym___restrict__] = ACTIONS(2450),
    [anon_sym__Atomic] = ACTIONS(2450),
    [anon_sym__Noreturn] = ACTIONS(2450),
    [anon_sym__Nonnull] = ACTIONS(2450),
    [anon_sym_nullable] = ACTIONS(2450),
    [anon_sym__Complex] = ACTIONS(2450),
    [anon_sym__Nullable] = ACTIONS(2450),
    [anon_sym__Nullable_result] = ACTIONS(2450),
    [anon_sym__Null_unspecified] = ACTIONS(2450),
    [anon_sym___autoreleasing] = ACTIONS(2450),
    [anon_sym___block] = ACTIONS(2450),
    [anon_sym___bridge] = ACTIONS(2450),
    [anon_sym___bridge_retained] = ACTIONS(2450),
    [anon_sym___bridge_transfer] = ACTIONS(2450),
    [anon_sym___complex] = ACTIONS(2450),
    [anon_sym___const] = ACTIONS(2450),
    [anon_sym___imag] = ACTIONS(2450),
    [anon_sym___kindof] = ACTIONS(2450),
    [anon_sym___nonnull] = ACTIONS(2450),
    [anon_sym___nullable] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2450),
    [anon_sym___real] = ACTIONS(2450),
    [anon_sym___strong] = ACTIONS(2450),
    [anon_sym___unsafe_unretained] = ACTIONS(2450),
    [anon_sym___unused] = ACTIONS(2450),
    [anon_sym___weak] = ACTIONS(2450),
    [anon_sym_alignas] = ACTIONS(2450),
    [anon_sym__Alignas] = ACTIONS(2450),
    [sym_primitive_type] = ACTIONS(2450),
    [anon_sym_enum] = ACTIONS(2450),
    [anon_sym_struct] = ACTIONS(2450),
    [anon_sym_union] = ACTIONS(2450),
    [anon_sym_if] = ACTIONS(2450),
    [anon_sym_else] = ACTIONS(2450),
    [anon_sym_switch] = ACTIONS(2450),
    [anon_sym_case] = ACTIONS(2450),
    [anon_sym_default] = ACTIONS(2450),
    [anon_sym_while] = ACTIONS(2450),
    [anon_sym_do] = ACTIONS(2450),
    [anon_sym_for] = ACTIONS(2450),
    [anon_sym_in] = ACTIONS(2450),
    [anon_sym_return] = ACTIONS(2450),
    [anon_sym_break] = ACTIONS(2450),
    [anon_sym_continue] = ACTIONS(2450),
    [anon_sym_goto] = ACTIONS(2450),
    [anon_sym_DASH_DASH] = ACTIONS(2452),
    [anon_sym_PLUS_PLUS] = ACTIONS(2452),
    [anon_sym_sizeof] = ACTIONS(2450),
    [anon_sym___alignof__] = ACTIONS(2450),
    [anon_sym___alignof] = ACTIONS(2450),
    [anon_sym__alignof] = ACTIONS(2450),
    [anon_sym_alignof] = ACTIONS(2450),
    [anon_sym__Alignof] = ACTIONS(2450),
    [anon_sym_offsetof] = ACTIONS(2450),
    [anon_sym__Generic] = ACTIONS(2450),
    [anon_sym_asm] = ACTIONS(2450),
    [anon_sym___asm__] = ACTIONS(2450),
    [anon_sym___asm] = ACTIONS(2450),
    [sym_number_literal] = ACTIONS(2452),
    [anon_sym_L_SQUOTE] = ACTIONS(2452),
    [anon_sym_u_SQUOTE] = ACTIONS(2452),
    [anon_sym_U_SQUOTE] = ACTIONS(2452),
    [anon_sym_u8_SQUOTE] = ACTIONS(2452),
    [anon_sym_SQUOTE] = ACTIONS(2452),
    [anon_sym_AT] = ACTIONS(2450),
    [anon_sym_DQUOTE] = ACTIONS(2452),
    [anon_sym_L_DQUOTE] = ACTIONS(2452),
    [anon_sym_u_DQUOTE] = ACTIONS(2452),
    [anon_sym_U_DQUOTE] = ACTIONS(2452),
    [anon_sym_u8_DQUOTE] = ACTIONS(2452),
    [sym_true] = ACTIONS(2450),
    [sym_false] = ACTIONS(2450),
    [anon_sym_NULL] = ACTIONS(2450),
    [anon_sym_nullptr] = ACTIONS(2450),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2450),
    [anon_sym___typeof] = ACTIONS(2450),
    [anon_sym_typeof] = ACTIONS(2450),
    [anon_sym_ATimport] = ACTIONS(2452),
    [aux_sym_preproc_undef_token1] = ACTIONS(2450),
    [anon_sym_POUND] = ACTIONS(2450),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2450),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2450),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2450),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_API_AVAILABLE] = ACTIONS(2450),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_API_DEPRECATED] = ACTIONS(2450),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym___deprecated_msg] = ACTIONS(2450),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2450),
    [anon_sym_ATprotocol] = ACTIONS(2452),
    [anon_sym_ATinterface] = ACTIONS(2452),
    [anon_sym_ATimplementation] = ACTIONS(2452),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2452),
    [anon_sym_ATsynthesize] = ACTIONS(2452),
    [anon_sym_ATdynamic] = ACTIONS(2452),
    [anon_sym_ATtry] = ACTIONS(2452),
    [anon_sym___try] = ACTIONS(2450),
    [anon_sym_ATthrow] = ACTIONS(2452),
    [anon_sym_ATselector] = ACTIONS(2452),
    [anon_sym_ATavailable] = ACTIONS(2452),
    [anon_sym___builtin_available] = ACTIONS(2450),
    [anon_sym_va_arg] = ACTIONS(2450),
    [anon_sym_ATencode] = ACTIONS(2452),
    [anon_sym_ATsynchronized] = ACTIONS(2452),
    [anon_sym_BOOL] = ACTIONS(2450),
    [anon_sym_IMP] = ACTIONS(2450),
    [anon_sym_SEL] = ACTIONS(2450),
    [anon_sym_Class] = ACTIONS(2450),
    [anon_sym_id] = ACTIONS(2450),
  },
  [STATE(544)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token2] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(545)] = {
    [sym_identifier] = ACTIONS(2542),
    [aux_sym_preproc_include_token1] = ACTIONS(2542),
    [aux_sym_preproc_include_token2] = ACTIONS(2542),
    [aux_sym_preproc_def_token1] = ACTIONS(2542),
    [aux_sym_preproc_if_token1] = ACTIONS(2542),
    [aux_sym_preproc_if_token2] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2542),
    [sym_preproc_directive] = ACTIONS(2542),
    [anon_sym_LPAREN2] = ACTIONS(2544),
    [anon_sym_BANG] = ACTIONS(2544),
    [anon_sym_TILDE] = ACTIONS(2544),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_PLUS] = ACTIONS(2542),
    [anon_sym_STAR] = ACTIONS(2544),
    [anon_sym_CARET] = ACTIONS(2544),
    [anon_sym_AMP] = ACTIONS(2544),
    [anon_sym_SEMI] = ACTIONS(2544),
    [anon_sym___extension__] = ACTIONS(2542),
    [anon_sym_typedef] = ACTIONS(2542),
    [anon_sym_extern] = ACTIONS(2542),
    [anon_sym___attribute__] = ACTIONS(2542),
    [anon_sym___attribute] = ACTIONS(2542),
    [anon_sym_noreturn] = ACTIONS(2542),
    [anon_sym_LBRACK] = ACTIONS(2544),
    [anon_sym___declspec] = ACTIONS(2542),
    [anon_sym___cdecl] = ACTIONS(2542),
    [anon_sym___clrcall] = ACTIONS(2542),
    [anon_sym___stdcall] = ACTIONS(2542),
    [anon_sym___fastcall] = ACTIONS(2542),
    [anon_sym___thiscall] = ACTIONS(2542),
    [anon_sym___vectorcall] = ACTIONS(2542),
    [anon_sym_LBRACE] = ACTIONS(2544),
    [anon_sym_signed] = ACTIONS(2542),
    [anon_sym_unsigned] = ACTIONS(2542),
    [anon_sym_long] = ACTIONS(2542),
    [anon_sym_short] = ACTIONS(2542),
    [anon_sym_static] = ACTIONS(2542),
    [anon_sym_ATautoreleasepool] = ACTIONS(2544),
    [anon_sym_auto] = ACTIONS(2542),
    [anon_sym_register] = ACTIONS(2542),
    [anon_sym_inline] = ACTIONS(2542),
    [anon_sym___inline] = ACTIONS(2542),
    [anon_sym___inline__] = ACTIONS(2542),
    [anon_sym___forceinline] = ACTIONS(2542),
    [anon_sym_thread_local] = ACTIONS(2542),
    [anon_sym___thread] = ACTIONS(2542),
    [anon_sym_CG_EXTERN] = ACTIONS(2542),
    [anon_sym_CG_INLINE] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2542),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2542),
    [anon_sym_IBOutlet] = ACTIONS(2542),
    [anon_sym_IBInspectable] = ACTIONS(2542),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2542),
    [anon_sym_NS_INLINE] = ACTIONS(2542),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2542),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2542),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2542),
    [anon_sym_const] = ACTIONS(2542),
    [anon_sym_constexpr] = ACTIONS(2542),
    [anon_sym_volatile] = ACTIONS(2542),
    [anon_sym_restrict] = ACTIONS(2542),
    [anon_sym___restrict__] = ACTIONS(2542),
    [anon_sym__Atomic] = ACTIONS(2542),
    [anon_sym__Noreturn] = ACTIONS(2542),
    [anon_sym__Nonnull] = ACTIONS(2542),
    [anon_sym_nullable] = ACTIONS(2542),
    [anon_sym__Complex] = ACTIONS(2542),
    [anon_sym__Nullable] = ACTIONS(2542),
    [anon_sym__Nullable_result] = ACTIONS(2542),
    [anon_sym__Null_unspecified] = ACTIONS(2542),
    [anon_sym___autoreleasing] = ACTIONS(2542),
    [anon_sym___block] = ACTIONS(2542),
    [anon_sym___bridge] = ACTIONS(2542),
    [anon_sym___bridge_retained] = ACTIONS(2542),
    [anon_sym___bridge_transfer] = ACTIONS(2542),
    [anon_sym___complex] = ACTIONS(2542),
    [anon_sym___const] = ACTIONS(2542),
    [anon_sym___imag] = ACTIONS(2542),
    [anon_sym___kindof] = ACTIONS(2542),
    [anon_sym___nonnull] = ACTIONS(2542),
    [anon_sym___nullable] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2542),
    [anon_sym___real] = ACTIONS(2542),
    [anon_sym___strong] = ACTIONS(2542),
    [anon_sym___unsafe_unretained] = ACTIONS(2542),
    [anon_sym___unused] = ACTIONS(2542),
    [anon_sym___weak] = ACTIONS(2542),
    [anon_sym_alignas] = ACTIONS(2542),
    [anon_sym__Alignas] = ACTIONS(2542),
    [sym_primitive_type] = ACTIONS(2542),
    [anon_sym_enum] = ACTIONS(2542),
    [anon_sym_struct] = ACTIONS(2542),
    [anon_sym_union] = ACTIONS(2542),
    [anon_sym_if] = ACTIONS(2542),
    [anon_sym_else] = ACTIONS(2542),
    [anon_sym_switch] = ACTIONS(2542),
    [anon_sym_case] = ACTIONS(2542),
    [anon_sym_default] = ACTIONS(2542),
    [anon_sym_while] = ACTIONS(2542),
    [anon_sym_do] = ACTIONS(2542),
    [anon_sym_for] = ACTIONS(2542),
    [anon_sym_in] = ACTIONS(2542),
    [anon_sym_return] = ACTIONS(2542),
    [anon_sym_break] = ACTIONS(2542),
    [anon_sym_continue] = ACTIONS(2542),
    [anon_sym_goto] = ACTIONS(2542),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_sizeof] = ACTIONS(2542),
    [anon_sym___alignof__] = ACTIONS(2542),
    [anon_sym___alignof] = ACTIONS(2542),
    [anon_sym__alignof] = ACTIONS(2542),
    [anon_sym_alignof] = ACTIONS(2542),
    [anon_sym__Alignof] = ACTIONS(2542),
    [anon_sym_offsetof] = ACTIONS(2542),
    [anon_sym__Generic] = ACTIONS(2542),
    [anon_sym_asm] = ACTIONS(2542),
    [anon_sym___asm__] = ACTIONS(2542),
    [anon_sym___asm] = ACTIONS(2542),
    [sym_number_literal] = ACTIONS(2544),
    [anon_sym_L_SQUOTE] = ACTIONS(2544),
    [anon_sym_u_SQUOTE] = ACTIONS(2544),
    [anon_sym_U_SQUOTE] = ACTIONS(2544),
    [anon_sym_u8_SQUOTE] = ACTIONS(2544),
    [anon_sym_SQUOTE] = ACTIONS(2544),
    [anon_sym_AT] = ACTIONS(2542),
    [anon_sym_DQUOTE] = ACTIONS(2544),
    [anon_sym_L_DQUOTE] = ACTIONS(2544),
    [anon_sym_u_DQUOTE] = ACTIONS(2544),
    [anon_sym_U_DQUOTE] = ACTIONS(2544),
    [anon_sym_u8_DQUOTE] = ACTIONS(2544),
    [sym_true] = ACTIONS(2542),
    [sym_false] = ACTIONS(2542),
    [anon_sym_NULL] = ACTIONS(2542),
    [anon_sym_nullptr] = ACTIONS(2542),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2542),
    [anon_sym___typeof] = ACTIONS(2542),
    [anon_sym_typeof] = ACTIONS(2542),
    [anon_sym_ATimport] = ACTIONS(2544),
    [aux_sym_preproc_undef_token1] = ACTIONS(2542),
    [anon_sym_POUND] = ACTIONS(2542),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2542),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2542),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2542),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_API_AVAILABLE] = ACTIONS(2542),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_API_DEPRECATED] = ACTIONS(2542),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym___deprecated_msg] = ACTIONS(2542),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2542),
    [anon_sym_ATprotocol] = ACTIONS(2544),
    [anon_sym_ATinterface] = ACTIONS(2544),
    [anon_sym_ATimplementation] = ACTIONS(2544),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2544),
    [anon_sym_ATsynthesize] = ACTIONS(2544),
    [anon_sym_ATdynamic] = ACTIONS(2544),
    [anon_sym_ATtry] = ACTIONS(2544),
    [anon_sym___try] = ACTIONS(2542),
    [anon_sym_ATthrow] = ACTIONS(2544),
    [anon_sym_ATselector] = ACTIONS(2544),
    [anon_sym_ATavailable] = ACTIONS(2544),
    [anon_sym___builtin_available] = ACTIONS(2542),
    [anon_sym_va_arg] = ACTIONS(2542),
    [anon_sym_ATencode] = ACTIONS(2544),
    [anon_sym_ATsynchronized] = ACTIONS(2544),
    [anon_sym_BOOL] = ACTIONS(2542),
    [anon_sym_IMP] = ACTIONS(2542),
    [anon_sym_SEL] = ACTIONS(2542),
    [anon_sym_Class] = ACTIONS(2542),
    [anon_sym_id] = ACTIONS(2542),
  },
  [STATE(546)] = {
    [sym_identifier] = ACTIONS(2458),
    [aux_sym_preproc_include_token1] = ACTIONS(2458),
    [aux_sym_preproc_include_token2] = ACTIONS(2458),
    [aux_sym_preproc_def_token1] = ACTIONS(2458),
    [aux_sym_preproc_if_token1] = ACTIONS(2458),
    [aux_sym_preproc_if_token2] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2458),
    [sym_preproc_directive] = ACTIONS(2458),
    [anon_sym_LPAREN2] = ACTIONS(2460),
    [anon_sym_BANG] = ACTIONS(2460),
    [anon_sym_TILDE] = ACTIONS(2460),
    [anon_sym_DASH] = ACTIONS(2458),
    [anon_sym_PLUS] = ACTIONS(2458),
    [anon_sym_STAR] = ACTIONS(2460),
    [anon_sym_CARET] = ACTIONS(2460),
    [anon_sym_AMP] = ACTIONS(2460),
    [anon_sym_SEMI] = ACTIONS(2460),
    [anon_sym___extension__] = ACTIONS(2458),
    [anon_sym_typedef] = ACTIONS(2458),
    [anon_sym_extern] = ACTIONS(2458),
    [anon_sym___attribute__] = ACTIONS(2458),
    [anon_sym___attribute] = ACTIONS(2458),
    [anon_sym_noreturn] = ACTIONS(2458),
    [anon_sym_LBRACK] = ACTIONS(2460),
    [anon_sym___declspec] = ACTIONS(2458),
    [anon_sym___cdecl] = ACTIONS(2458),
    [anon_sym___clrcall] = ACTIONS(2458),
    [anon_sym___stdcall] = ACTIONS(2458),
    [anon_sym___fastcall] = ACTIONS(2458),
    [anon_sym___thiscall] = ACTIONS(2458),
    [anon_sym___vectorcall] = ACTIONS(2458),
    [anon_sym_LBRACE] = ACTIONS(2460),
    [anon_sym_signed] = ACTIONS(2458),
    [anon_sym_unsigned] = ACTIONS(2458),
    [anon_sym_long] = ACTIONS(2458),
    [anon_sym_short] = ACTIONS(2458),
    [anon_sym_static] = ACTIONS(2458),
    [anon_sym_ATautoreleasepool] = ACTIONS(2460),
    [anon_sym_auto] = ACTIONS(2458),
    [anon_sym_register] = ACTIONS(2458),
    [anon_sym_inline] = ACTIONS(2458),
    [anon_sym___inline] = ACTIONS(2458),
    [anon_sym___inline__] = ACTIONS(2458),
    [anon_sym___forceinline] = ACTIONS(2458),
    [anon_sym_thread_local] = ACTIONS(2458),
    [anon_sym___thread] = ACTIONS(2458),
    [anon_sym_CG_EXTERN] = ACTIONS(2458),
    [anon_sym_CG_INLINE] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2458),
    [anon_sym_IBOutlet] = ACTIONS(2458),
    [anon_sym_IBInspectable] = ACTIONS(2458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2458),
    [anon_sym_NS_INLINE] = ACTIONS(2458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2458),
    [anon_sym_const] = ACTIONS(2458),
    [anon_sym_constexpr] = ACTIONS(2458),
    [anon_sym_volatile] = ACTIONS(2458),
    [anon_sym_restrict] = ACTIONS(2458),
    [anon_sym___restrict__] = ACTIONS(2458),
    [anon_sym__Atomic] = ACTIONS(2458),
    [anon_sym__Noreturn] = ACTIONS(2458),
    [anon_sym__Nonnull] = ACTIONS(2458),
    [anon_sym_nullable] = ACTIONS(2458),
    [anon_sym__Complex] = ACTIONS(2458),
    [anon_sym__Nullable] = ACTIONS(2458),
    [anon_sym__Nullable_result] = ACTIONS(2458),
    [anon_sym__Null_unspecified] = ACTIONS(2458),
    [anon_sym___autoreleasing] = ACTIONS(2458),
    [anon_sym___block] = ACTIONS(2458),
    [anon_sym___bridge] = ACTIONS(2458),
    [anon_sym___bridge_retained] = ACTIONS(2458),
    [anon_sym___bridge_transfer] = ACTIONS(2458),
    [anon_sym___complex] = ACTIONS(2458),
    [anon_sym___const] = ACTIONS(2458),
    [anon_sym___imag] = ACTIONS(2458),
    [anon_sym___kindof] = ACTIONS(2458),
    [anon_sym___nonnull] = ACTIONS(2458),
    [anon_sym___nullable] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2458),
    [anon_sym___real] = ACTIONS(2458),
    [anon_sym___strong] = ACTIONS(2458),
    [anon_sym___unsafe_unretained] = ACTIONS(2458),
    [anon_sym___unused] = ACTIONS(2458),
    [anon_sym___weak] = ACTIONS(2458),
    [anon_sym_alignas] = ACTIONS(2458),
    [anon_sym__Alignas] = ACTIONS(2458),
    [sym_primitive_type] = ACTIONS(2458),
    [anon_sym_enum] = ACTIONS(2458),
    [anon_sym_struct] = ACTIONS(2458),
    [anon_sym_union] = ACTIONS(2458),
    [anon_sym_if] = ACTIONS(2458),
    [anon_sym_else] = ACTIONS(2458),
    [anon_sym_switch] = ACTIONS(2458),
    [anon_sym_case] = ACTIONS(2458),
    [anon_sym_default] = ACTIONS(2458),
    [anon_sym_while] = ACTIONS(2458),
    [anon_sym_do] = ACTIONS(2458),
    [anon_sym_for] = ACTIONS(2458),
    [anon_sym_in] = ACTIONS(2458),
    [anon_sym_return] = ACTIONS(2458),
    [anon_sym_break] = ACTIONS(2458),
    [anon_sym_continue] = ACTIONS(2458),
    [anon_sym_goto] = ACTIONS(2458),
    [anon_sym_DASH_DASH] = ACTIONS(2460),
    [anon_sym_PLUS_PLUS] = ACTIONS(2460),
    [anon_sym_sizeof] = ACTIONS(2458),
    [anon_sym___alignof__] = ACTIONS(2458),
    [anon_sym___alignof] = ACTIONS(2458),
    [anon_sym__alignof] = ACTIONS(2458),
    [anon_sym_alignof] = ACTIONS(2458),
    [anon_sym__Alignof] = ACTIONS(2458),
    [anon_sym_offsetof] = ACTIONS(2458),
    [anon_sym__Generic] = ACTIONS(2458),
    [anon_sym_asm] = ACTIONS(2458),
    [anon_sym___asm__] = ACTIONS(2458),
    [anon_sym___asm] = ACTIONS(2458),
    [sym_number_literal] = ACTIONS(2460),
    [anon_sym_L_SQUOTE] = ACTIONS(2460),
    [anon_sym_u_SQUOTE] = ACTIONS(2460),
    [anon_sym_U_SQUOTE] = ACTIONS(2460),
    [anon_sym_u8_SQUOTE] = ACTIONS(2460),
    [anon_sym_SQUOTE] = ACTIONS(2460),
    [anon_sym_AT] = ACTIONS(2458),
    [anon_sym_DQUOTE] = ACTIONS(2460),
    [anon_sym_L_DQUOTE] = ACTIONS(2460),
    [anon_sym_u_DQUOTE] = ACTIONS(2460),
    [anon_sym_U_DQUOTE] = ACTIONS(2460),
    [anon_sym_u8_DQUOTE] = ACTIONS(2460),
    [sym_true] = ACTIONS(2458),
    [sym_false] = ACTIONS(2458),
    [anon_sym_NULL] = ACTIONS(2458),
    [anon_sym_nullptr] = ACTIONS(2458),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2458),
    [anon_sym___typeof] = ACTIONS(2458),
    [anon_sym_typeof] = ACTIONS(2458),
    [anon_sym_ATimport] = ACTIONS(2460),
    [aux_sym_preproc_undef_token1] = ACTIONS(2458),
    [anon_sym_POUND] = ACTIONS(2458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_API_AVAILABLE] = ACTIONS(2458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_API_DEPRECATED] = ACTIONS(2458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym___deprecated_msg] = ACTIONS(2458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2458),
    [anon_sym_ATprotocol] = ACTIONS(2460),
    [anon_sym_ATinterface] = ACTIONS(2460),
    [anon_sym_ATimplementation] = ACTIONS(2460),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2460),
    [anon_sym_ATsynthesize] = ACTIONS(2460),
    [anon_sym_ATdynamic] = ACTIONS(2460),
    [anon_sym_ATtry] = ACTIONS(2460),
    [anon_sym___try] = ACTIONS(2458),
    [anon_sym_ATthrow] = ACTIONS(2460),
    [anon_sym_ATselector] = ACTIONS(2460),
    [anon_sym_ATavailable] = ACTIONS(2460),
    [anon_sym___builtin_available] = ACTIONS(2458),
    [anon_sym_va_arg] = ACTIONS(2458),
    [anon_sym_ATencode] = ACTIONS(2460),
    [anon_sym_ATsynchronized] = ACTIONS(2460),
    [anon_sym_BOOL] = ACTIONS(2458),
    [anon_sym_IMP] = ACTIONS(2458),
    [anon_sym_SEL] = ACTIONS(2458),
    [anon_sym_Class] = ACTIONS(2458),
    [anon_sym_id] = ACTIONS(2458),
  },
  [STATE(547)] = {
    [sym_identifier] = ACTIONS(2550),
    [aux_sym_preproc_include_token1] = ACTIONS(2550),
    [aux_sym_preproc_include_token2] = ACTIONS(2550),
    [aux_sym_preproc_def_token1] = ACTIONS(2550),
    [aux_sym_preproc_if_token1] = ACTIONS(2550),
    [aux_sym_preproc_if_token2] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2550),
    [sym_preproc_directive] = ACTIONS(2550),
    [anon_sym_LPAREN2] = ACTIONS(2552),
    [anon_sym_BANG] = ACTIONS(2552),
    [anon_sym_TILDE] = ACTIONS(2552),
    [anon_sym_DASH] = ACTIONS(2550),
    [anon_sym_PLUS] = ACTIONS(2550),
    [anon_sym_STAR] = ACTIONS(2552),
    [anon_sym_CARET] = ACTIONS(2552),
    [anon_sym_AMP] = ACTIONS(2552),
    [anon_sym_SEMI] = ACTIONS(2552),
    [anon_sym___extension__] = ACTIONS(2550),
    [anon_sym_typedef] = ACTIONS(2550),
    [anon_sym_extern] = ACTIONS(2550),
    [anon_sym___attribute__] = ACTIONS(2550),
    [anon_sym___attribute] = ACTIONS(2550),
    [anon_sym_noreturn] = ACTIONS(2550),
    [anon_sym_LBRACK] = ACTIONS(2552),
    [anon_sym___declspec] = ACTIONS(2550),
    [anon_sym___cdecl] = ACTIONS(2550),
    [anon_sym___clrcall] = ACTIONS(2550),
    [anon_sym___stdcall] = ACTIONS(2550),
    [anon_sym___fastcall] = ACTIONS(2550),
    [anon_sym___thiscall] = ACTIONS(2550),
    [anon_sym___vectorcall] = ACTIONS(2550),
    [anon_sym_LBRACE] = ACTIONS(2552),
    [anon_sym_signed] = ACTIONS(2550),
    [anon_sym_unsigned] = ACTIONS(2550),
    [anon_sym_long] = ACTIONS(2550),
    [anon_sym_short] = ACTIONS(2550),
    [anon_sym_static] = ACTIONS(2550),
    [anon_sym_ATautoreleasepool] = ACTIONS(2552),
    [anon_sym_auto] = ACTIONS(2550),
    [anon_sym_register] = ACTIONS(2550),
    [anon_sym_inline] = ACTIONS(2550),
    [anon_sym___inline] = ACTIONS(2550),
    [anon_sym___inline__] = ACTIONS(2550),
    [anon_sym___forceinline] = ACTIONS(2550),
    [anon_sym_thread_local] = ACTIONS(2550),
    [anon_sym___thread] = ACTIONS(2550),
    [anon_sym_CG_EXTERN] = ACTIONS(2550),
    [anon_sym_CG_INLINE] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2550),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2550),
    [anon_sym_IBOutlet] = ACTIONS(2550),
    [anon_sym_IBInspectable] = ACTIONS(2550),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2550),
    [anon_sym_NS_INLINE] = ACTIONS(2550),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2550),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2550),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2550),
    [anon_sym_const] = ACTIONS(2550),
    [anon_sym_constexpr] = ACTIONS(2550),
    [anon_sym_volatile] = ACTIONS(2550),
    [anon_sym_restrict] = ACTIONS(2550),
    [anon_sym___restrict__] = ACTIONS(2550),
    [anon_sym__Atomic] = ACTIONS(2550),
    [anon_sym__Noreturn] = ACTIONS(2550),
    [anon_sym__Nonnull] = ACTIONS(2550),
    [anon_sym_nullable] = ACTIONS(2550),
    [anon_sym__Complex] = ACTIONS(2550),
    [anon_sym__Nullable] = ACTIONS(2550),
    [anon_sym__Nullable_result] = ACTIONS(2550),
    [anon_sym__Null_unspecified] = ACTIONS(2550),
    [anon_sym___autoreleasing] = ACTIONS(2550),
    [anon_sym___block] = ACTIONS(2550),
    [anon_sym___bridge] = ACTIONS(2550),
    [anon_sym___bridge_retained] = ACTIONS(2550),
    [anon_sym___bridge_transfer] = ACTIONS(2550),
    [anon_sym___complex] = ACTIONS(2550),
    [anon_sym___const] = ACTIONS(2550),
    [anon_sym___imag] = ACTIONS(2550),
    [anon_sym___kindof] = ACTIONS(2550),
    [anon_sym___nonnull] = ACTIONS(2550),
    [anon_sym___nullable] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2550),
    [anon_sym___real] = ACTIONS(2550),
    [anon_sym___strong] = ACTIONS(2550),
    [anon_sym___unsafe_unretained] = ACTIONS(2550),
    [anon_sym___unused] = ACTIONS(2550),
    [anon_sym___weak] = ACTIONS(2550),
    [anon_sym_alignas] = ACTIONS(2550),
    [anon_sym__Alignas] = ACTIONS(2550),
    [sym_primitive_type] = ACTIONS(2550),
    [anon_sym_enum] = ACTIONS(2550),
    [anon_sym_struct] = ACTIONS(2550),
    [anon_sym_union] = ACTIONS(2550),
    [anon_sym_if] = ACTIONS(2550),
    [anon_sym_else] = ACTIONS(2550),
    [anon_sym_switch] = ACTIONS(2550),
    [anon_sym_case] = ACTIONS(2550),
    [anon_sym_default] = ACTIONS(2550),
    [anon_sym_while] = ACTIONS(2550),
    [anon_sym_do] = ACTIONS(2550),
    [anon_sym_for] = ACTIONS(2550),
    [anon_sym_in] = ACTIONS(2550),
    [anon_sym_return] = ACTIONS(2550),
    [anon_sym_break] = ACTIONS(2550),
    [anon_sym_continue] = ACTIONS(2550),
    [anon_sym_goto] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2552),
    [anon_sym_PLUS_PLUS] = ACTIONS(2552),
    [anon_sym_sizeof] = ACTIONS(2550),
    [anon_sym___alignof__] = ACTIONS(2550),
    [anon_sym___alignof] = ACTIONS(2550),
    [anon_sym__alignof] = ACTIONS(2550),
    [anon_sym_alignof] = ACTIONS(2550),
    [anon_sym__Alignof] = ACTIONS(2550),
    [anon_sym_offsetof] = ACTIONS(2550),
    [anon_sym__Generic] = ACTIONS(2550),
    [anon_sym_asm] = ACTIONS(2550),
    [anon_sym___asm__] = ACTIONS(2550),
    [anon_sym___asm] = ACTIONS(2550),
    [sym_number_literal] = ACTIONS(2552),
    [anon_sym_L_SQUOTE] = ACTIONS(2552),
    [anon_sym_u_SQUOTE] = ACTIONS(2552),
    [anon_sym_U_SQUOTE] = ACTIONS(2552),
    [anon_sym_u8_SQUOTE] = ACTIONS(2552),
    [anon_sym_SQUOTE] = ACTIONS(2552),
    [anon_sym_AT] = ACTIONS(2550),
    [anon_sym_DQUOTE] = ACTIONS(2552),
    [anon_sym_L_DQUOTE] = ACTIONS(2552),
    [anon_sym_u_DQUOTE] = ACTIONS(2552),
    [anon_sym_U_DQUOTE] = ACTIONS(2552),
    [anon_sym_u8_DQUOTE] = ACTIONS(2552),
    [sym_true] = ACTIONS(2550),
    [sym_false] = ACTIONS(2550),
    [anon_sym_NULL] = ACTIONS(2550),
    [anon_sym_nullptr] = ACTIONS(2550),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2550),
    [anon_sym___typeof] = ACTIONS(2550),
    [anon_sym_typeof] = ACTIONS(2550),
    [anon_sym_ATimport] = ACTIONS(2552),
    [aux_sym_preproc_undef_token1] = ACTIONS(2550),
    [anon_sym_POUND] = ACTIONS(2550),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2550),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2550),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2550),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_API_AVAILABLE] = ACTIONS(2550),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_API_DEPRECATED] = ACTIONS(2550),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym___deprecated_msg] = ACTIONS(2550),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2550),
    [anon_sym_ATprotocol] = ACTIONS(2552),
    [anon_sym_ATinterface] = ACTIONS(2552),
    [anon_sym_ATimplementation] = ACTIONS(2552),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2552),
    [anon_sym_ATsynthesize] = ACTIONS(2552),
    [anon_sym_ATdynamic] = ACTIONS(2552),
    [anon_sym_ATtry] = ACTIONS(2552),
    [anon_sym___try] = ACTIONS(2550),
    [anon_sym_ATthrow] = ACTIONS(2552),
    [anon_sym_ATselector] = ACTIONS(2552),
    [anon_sym_ATavailable] = ACTIONS(2552),
    [anon_sym___builtin_available] = ACTIONS(2550),
    [anon_sym_va_arg] = ACTIONS(2550),
    [anon_sym_ATencode] = ACTIONS(2552),
    [anon_sym_ATsynchronized] = ACTIONS(2552),
    [anon_sym_BOOL] = ACTIONS(2550),
    [anon_sym_IMP] = ACTIONS(2550),
    [anon_sym_SEL] = ACTIONS(2550),
    [anon_sym_Class] = ACTIONS(2550),
    [anon_sym_id] = ACTIONS(2550),
  },
  [STATE(548)] = {
    [sym_identifier] = ACTIONS(2462),
    [aux_sym_preproc_include_token1] = ACTIONS(2462),
    [aux_sym_preproc_include_token2] = ACTIONS(2462),
    [aux_sym_preproc_def_token1] = ACTIONS(2462),
    [aux_sym_preproc_if_token1] = ACTIONS(2462),
    [aux_sym_preproc_if_token2] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2462),
    [sym_preproc_directive] = ACTIONS(2462),
    [anon_sym_LPAREN2] = ACTIONS(2464),
    [anon_sym_BANG] = ACTIONS(2464),
    [anon_sym_TILDE] = ACTIONS(2464),
    [anon_sym_DASH] = ACTIONS(2462),
    [anon_sym_PLUS] = ACTIONS(2462),
    [anon_sym_STAR] = ACTIONS(2464),
    [anon_sym_CARET] = ACTIONS(2464),
    [anon_sym_AMP] = ACTIONS(2464),
    [anon_sym_SEMI] = ACTIONS(2464),
    [anon_sym___extension__] = ACTIONS(2462),
    [anon_sym_typedef] = ACTIONS(2462),
    [anon_sym_extern] = ACTIONS(2462),
    [anon_sym___attribute__] = ACTIONS(2462),
    [anon_sym___attribute] = ACTIONS(2462),
    [anon_sym_noreturn] = ACTIONS(2462),
    [anon_sym_LBRACK] = ACTIONS(2464),
    [anon_sym___declspec] = ACTIONS(2462),
    [anon_sym___cdecl] = ACTIONS(2462),
    [anon_sym___clrcall] = ACTIONS(2462),
    [anon_sym___stdcall] = ACTIONS(2462),
    [anon_sym___fastcall] = ACTIONS(2462),
    [anon_sym___thiscall] = ACTIONS(2462),
    [anon_sym___vectorcall] = ACTIONS(2462),
    [anon_sym_LBRACE] = ACTIONS(2464),
    [anon_sym_signed] = ACTIONS(2462),
    [anon_sym_unsigned] = ACTIONS(2462),
    [anon_sym_long] = ACTIONS(2462),
    [anon_sym_short] = ACTIONS(2462),
    [anon_sym_static] = ACTIONS(2462),
    [anon_sym_ATautoreleasepool] = ACTIONS(2464),
    [anon_sym_auto] = ACTIONS(2462),
    [anon_sym_register] = ACTIONS(2462),
    [anon_sym_inline] = ACTIONS(2462),
    [anon_sym___inline] = ACTIONS(2462),
    [anon_sym___inline__] = ACTIONS(2462),
    [anon_sym___forceinline] = ACTIONS(2462),
    [anon_sym_thread_local] = ACTIONS(2462),
    [anon_sym___thread] = ACTIONS(2462),
    [anon_sym_CG_EXTERN] = ACTIONS(2462),
    [anon_sym_CG_INLINE] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2462),
    [anon_sym_IBOutlet] = ACTIONS(2462),
    [anon_sym_IBInspectable] = ACTIONS(2462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2462),
    [anon_sym_NS_INLINE] = ACTIONS(2462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2462),
    [anon_sym_const] = ACTIONS(2462),
    [anon_sym_constexpr] = ACTIONS(2462),
    [anon_sym_volatile] = ACTIONS(2462),
    [anon_sym_restrict] = ACTIONS(2462),
    [anon_sym___restrict__] = ACTIONS(2462),
    [anon_sym__Atomic] = ACTIONS(2462),
    [anon_sym__Noreturn] = ACTIONS(2462),
    [anon_sym__Nonnull] = ACTIONS(2462),
    [anon_sym_nullable] = ACTIONS(2462),
    [anon_sym__Complex] = ACTIONS(2462),
    [anon_sym__Nullable] = ACTIONS(2462),
    [anon_sym__Nullable_result] = ACTIONS(2462),
    [anon_sym__Null_unspecified] = ACTIONS(2462),
    [anon_sym___autoreleasing] = ACTIONS(2462),
    [anon_sym___block] = ACTIONS(2462),
    [anon_sym___bridge] = ACTIONS(2462),
    [anon_sym___bridge_retained] = ACTIONS(2462),
    [anon_sym___bridge_transfer] = ACTIONS(2462),
    [anon_sym___complex] = ACTIONS(2462),
    [anon_sym___const] = ACTIONS(2462),
    [anon_sym___imag] = ACTIONS(2462),
    [anon_sym___kindof] = ACTIONS(2462),
    [anon_sym___nonnull] = ACTIONS(2462),
    [anon_sym___nullable] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2462),
    [anon_sym___real] = ACTIONS(2462),
    [anon_sym___strong] = ACTIONS(2462),
    [anon_sym___unsafe_unretained] = ACTIONS(2462),
    [anon_sym___unused] = ACTIONS(2462),
    [anon_sym___weak] = ACTIONS(2462),
    [anon_sym_alignas] = ACTIONS(2462),
    [anon_sym__Alignas] = ACTIONS(2462),
    [sym_primitive_type] = ACTIONS(2462),
    [anon_sym_enum] = ACTIONS(2462),
    [anon_sym_struct] = ACTIONS(2462),
    [anon_sym_union] = ACTIONS(2462),
    [anon_sym_if] = ACTIONS(2462),
    [anon_sym_else] = ACTIONS(2462),
    [anon_sym_switch] = ACTIONS(2462),
    [anon_sym_case] = ACTIONS(2462),
    [anon_sym_default] = ACTIONS(2462),
    [anon_sym_while] = ACTIONS(2462),
    [anon_sym_do] = ACTIONS(2462),
    [anon_sym_for] = ACTIONS(2462),
    [anon_sym_in] = ACTIONS(2462),
    [anon_sym_return] = ACTIONS(2462),
    [anon_sym_break] = ACTIONS(2462),
    [anon_sym_continue] = ACTIONS(2462),
    [anon_sym_goto] = ACTIONS(2462),
    [anon_sym_DASH_DASH] = ACTIONS(2464),
    [anon_sym_PLUS_PLUS] = ACTIONS(2464),
    [anon_sym_sizeof] = ACTIONS(2462),
    [anon_sym___alignof__] = ACTIONS(2462),
    [anon_sym___alignof] = ACTIONS(2462),
    [anon_sym__alignof] = ACTIONS(2462),
    [anon_sym_alignof] = ACTIONS(2462),
    [anon_sym__Alignof] = ACTIONS(2462),
    [anon_sym_offsetof] = ACTIONS(2462),
    [anon_sym__Generic] = ACTIONS(2462),
    [anon_sym_asm] = ACTIONS(2462),
    [anon_sym___asm__] = ACTIONS(2462),
    [anon_sym___asm] = ACTIONS(2462),
    [sym_number_literal] = ACTIONS(2464),
    [anon_sym_L_SQUOTE] = ACTIONS(2464),
    [anon_sym_u_SQUOTE] = ACTIONS(2464),
    [anon_sym_U_SQUOTE] = ACTIONS(2464),
    [anon_sym_u8_SQUOTE] = ACTIONS(2464),
    [anon_sym_SQUOTE] = ACTIONS(2464),
    [anon_sym_AT] = ACTIONS(2462),
    [anon_sym_DQUOTE] = ACTIONS(2464),
    [anon_sym_L_DQUOTE] = ACTIONS(2464),
    [anon_sym_u_DQUOTE] = ACTIONS(2464),
    [anon_sym_U_DQUOTE] = ACTIONS(2464),
    [anon_sym_u8_DQUOTE] = ACTIONS(2464),
    [sym_true] = ACTIONS(2462),
    [sym_false] = ACTIONS(2462),
    [anon_sym_NULL] = ACTIONS(2462),
    [anon_sym_nullptr] = ACTIONS(2462),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2462),
    [anon_sym___typeof] = ACTIONS(2462),
    [anon_sym_typeof] = ACTIONS(2462),
    [anon_sym_ATimport] = ACTIONS(2464),
    [aux_sym_preproc_undef_token1] = ACTIONS(2462),
    [anon_sym_POUND] = ACTIONS(2462),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2462),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_API_AVAILABLE] = ACTIONS(2462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_API_DEPRECATED] = ACTIONS(2462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym___deprecated_msg] = ACTIONS(2462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2462),
    [anon_sym_ATprotocol] = ACTIONS(2464),
    [anon_sym_ATinterface] = ACTIONS(2464),
    [anon_sym_ATimplementation] = ACTIONS(2464),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2464),
    [anon_sym_ATsynthesize] = ACTIONS(2464),
    [anon_sym_ATdynamic] = ACTIONS(2464),
    [anon_sym_ATtry] = ACTIONS(2464),
    [anon_sym___try] = ACTIONS(2462),
    [anon_sym_ATthrow] = ACTIONS(2464),
    [anon_sym_ATselector] = ACTIONS(2464),
    [anon_sym_ATavailable] = ACTIONS(2464),
    [anon_sym___builtin_available] = ACTIONS(2462),
    [anon_sym_va_arg] = ACTIONS(2462),
    [anon_sym_ATencode] = ACTIONS(2464),
    [anon_sym_ATsynchronized] = ACTIONS(2464),
    [anon_sym_BOOL] = ACTIONS(2462),
    [anon_sym_IMP] = ACTIONS(2462),
    [anon_sym_SEL] = ACTIONS(2462),
    [anon_sym_Class] = ACTIONS(2462),
    [anon_sym_id] = ACTIONS(2462),
  },
  [STATE(549)] = {
    [sym_identifier] = ACTIONS(2466),
    [aux_sym_preproc_include_token1] = ACTIONS(2466),
    [aux_sym_preproc_include_token2] = ACTIONS(2466),
    [aux_sym_preproc_def_token1] = ACTIONS(2466),
    [aux_sym_preproc_if_token1] = ACTIONS(2466),
    [aux_sym_preproc_if_token2] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2466),
    [sym_preproc_directive] = ACTIONS(2466),
    [anon_sym_LPAREN2] = ACTIONS(2468),
    [anon_sym_BANG] = ACTIONS(2468),
    [anon_sym_TILDE] = ACTIONS(2468),
    [anon_sym_DASH] = ACTIONS(2466),
    [anon_sym_PLUS] = ACTIONS(2466),
    [anon_sym_STAR] = ACTIONS(2468),
    [anon_sym_CARET] = ACTIONS(2468),
    [anon_sym_AMP] = ACTIONS(2468),
    [anon_sym_SEMI] = ACTIONS(2468),
    [anon_sym___extension__] = ACTIONS(2466),
    [anon_sym_typedef] = ACTIONS(2466),
    [anon_sym_extern] = ACTIONS(2466),
    [anon_sym___attribute__] = ACTIONS(2466),
    [anon_sym___attribute] = ACTIONS(2466),
    [anon_sym_noreturn] = ACTIONS(2466),
    [anon_sym_LBRACK] = ACTIONS(2468),
    [anon_sym___declspec] = ACTIONS(2466),
    [anon_sym___cdecl] = ACTIONS(2466),
    [anon_sym___clrcall] = ACTIONS(2466),
    [anon_sym___stdcall] = ACTIONS(2466),
    [anon_sym___fastcall] = ACTIONS(2466),
    [anon_sym___thiscall] = ACTIONS(2466),
    [anon_sym___vectorcall] = ACTIONS(2466),
    [anon_sym_LBRACE] = ACTIONS(2468),
    [anon_sym_signed] = ACTIONS(2466),
    [anon_sym_unsigned] = ACTIONS(2466),
    [anon_sym_long] = ACTIONS(2466),
    [anon_sym_short] = ACTIONS(2466),
    [anon_sym_static] = ACTIONS(2466),
    [anon_sym_ATautoreleasepool] = ACTIONS(2468),
    [anon_sym_auto] = ACTIONS(2466),
    [anon_sym_register] = ACTIONS(2466),
    [anon_sym_inline] = ACTIONS(2466),
    [anon_sym___inline] = ACTIONS(2466),
    [anon_sym___inline__] = ACTIONS(2466),
    [anon_sym___forceinline] = ACTIONS(2466),
    [anon_sym_thread_local] = ACTIONS(2466),
    [anon_sym___thread] = ACTIONS(2466),
    [anon_sym_CG_EXTERN] = ACTIONS(2466),
    [anon_sym_CG_INLINE] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2466),
    [anon_sym_IBOutlet] = ACTIONS(2466),
    [anon_sym_IBInspectable] = ACTIONS(2466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2466),
    [anon_sym_NS_INLINE] = ACTIONS(2466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2466),
    [anon_sym_const] = ACTIONS(2466),
    [anon_sym_constexpr] = ACTIONS(2466),
    [anon_sym_volatile] = ACTIONS(2466),
    [anon_sym_restrict] = ACTIONS(2466),
    [anon_sym___restrict__] = ACTIONS(2466),
    [anon_sym__Atomic] = ACTIONS(2466),
    [anon_sym__Noreturn] = ACTIONS(2466),
    [anon_sym__Nonnull] = ACTIONS(2466),
    [anon_sym_nullable] = ACTIONS(2466),
    [anon_sym__Complex] = ACTIONS(2466),
    [anon_sym__Nullable] = ACTIONS(2466),
    [anon_sym__Nullable_result] = ACTIONS(2466),
    [anon_sym__Null_unspecified] = ACTIONS(2466),
    [anon_sym___autoreleasing] = ACTIONS(2466),
    [anon_sym___block] = ACTIONS(2466),
    [anon_sym___bridge] = ACTIONS(2466),
    [anon_sym___bridge_retained] = ACTIONS(2466),
    [anon_sym___bridge_transfer] = ACTIONS(2466),
    [anon_sym___complex] = ACTIONS(2466),
    [anon_sym___const] = ACTIONS(2466),
    [anon_sym___imag] = ACTIONS(2466),
    [anon_sym___kindof] = ACTIONS(2466),
    [anon_sym___nonnull] = ACTIONS(2466),
    [anon_sym___nullable] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2466),
    [anon_sym___real] = ACTIONS(2466),
    [anon_sym___strong] = ACTIONS(2466),
    [anon_sym___unsafe_unretained] = ACTIONS(2466),
    [anon_sym___unused] = ACTIONS(2466),
    [anon_sym___weak] = ACTIONS(2466),
    [anon_sym_alignas] = ACTIONS(2466),
    [anon_sym__Alignas] = ACTIONS(2466),
    [sym_primitive_type] = ACTIONS(2466),
    [anon_sym_enum] = ACTIONS(2466),
    [anon_sym_struct] = ACTIONS(2466),
    [anon_sym_union] = ACTIONS(2466),
    [anon_sym_if] = ACTIONS(2466),
    [anon_sym_else] = ACTIONS(2466),
    [anon_sym_switch] = ACTIONS(2466),
    [anon_sym_case] = ACTIONS(2466),
    [anon_sym_default] = ACTIONS(2466),
    [anon_sym_while] = ACTIONS(2466),
    [anon_sym_do] = ACTIONS(2466),
    [anon_sym_for] = ACTIONS(2466),
    [anon_sym_in] = ACTIONS(2466),
    [anon_sym_return] = ACTIONS(2466),
    [anon_sym_break] = ACTIONS(2466),
    [anon_sym_continue] = ACTIONS(2466),
    [anon_sym_goto] = ACTIONS(2466),
    [anon_sym_DASH_DASH] = ACTIONS(2468),
    [anon_sym_PLUS_PLUS] = ACTIONS(2468),
    [anon_sym_sizeof] = ACTIONS(2466),
    [anon_sym___alignof__] = ACTIONS(2466),
    [anon_sym___alignof] = ACTIONS(2466),
    [anon_sym__alignof] = ACTIONS(2466),
    [anon_sym_alignof] = ACTIONS(2466),
    [anon_sym__Alignof] = ACTIONS(2466),
    [anon_sym_offsetof] = ACTIONS(2466),
    [anon_sym__Generic] = ACTIONS(2466),
    [anon_sym_asm] = ACTIONS(2466),
    [anon_sym___asm__] = ACTIONS(2466),
    [anon_sym___asm] = ACTIONS(2466),
    [sym_number_literal] = ACTIONS(2468),
    [anon_sym_L_SQUOTE] = ACTIONS(2468),
    [anon_sym_u_SQUOTE] = ACTIONS(2468),
    [anon_sym_U_SQUOTE] = ACTIONS(2468),
    [anon_sym_u8_SQUOTE] = ACTIONS(2468),
    [anon_sym_SQUOTE] = ACTIONS(2468),
    [anon_sym_AT] = ACTIONS(2466),
    [anon_sym_DQUOTE] = ACTIONS(2468),
    [anon_sym_L_DQUOTE] = ACTIONS(2468),
    [anon_sym_u_DQUOTE] = ACTIONS(2468),
    [anon_sym_U_DQUOTE] = ACTIONS(2468),
    [anon_sym_u8_DQUOTE] = ACTIONS(2468),
    [sym_true] = ACTIONS(2466),
    [sym_false] = ACTIONS(2466),
    [anon_sym_NULL] = ACTIONS(2466),
    [anon_sym_nullptr] = ACTIONS(2466),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2466),
    [anon_sym___typeof] = ACTIONS(2466),
    [anon_sym_typeof] = ACTIONS(2466),
    [anon_sym_ATimport] = ACTIONS(2468),
    [aux_sym_preproc_undef_token1] = ACTIONS(2466),
    [anon_sym_POUND] = ACTIONS(2466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_API_AVAILABLE] = ACTIONS(2466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_API_DEPRECATED] = ACTIONS(2466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym___deprecated_msg] = ACTIONS(2466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2466),
    [anon_sym_ATprotocol] = ACTIONS(2468),
    [anon_sym_ATinterface] = ACTIONS(2468),
    [anon_sym_ATimplementation] = ACTIONS(2468),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2468),
    [anon_sym_ATsynthesize] = ACTIONS(2468),
    [anon_sym_ATdynamic] = ACTIONS(2468),
    [anon_sym_ATtry] = ACTIONS(2468),
    [anon_sym___try] = ACTIONS(2466),
    [anon_sym_ATthrow] = ACTIONS(2468),
    [anon_sym_ATselector] = ACTIONS(2468),
    [anon_sym_ATavailable] = ACTIONS(2468),
    [anon_sym___builtin_available] = ACTIONS(2466),
    [anon_sym_va_arg] = ACTIONS(2466),
    [anon_sym_ATencode] = ACTIONS(2468),
    [anon_sym_ATsynchronized] = ACTIONS(2468),
    [anon_sym_BOOL] = ACTIONS(2466),
    [anon_sym_IMP] = ACTIONS(2466),
    [anon_sym_SEL] = ACTIONS(2466),
    [anon_sym_Class] = ACTIONS(2466),
    [anon_sym_id] = ACTIONS(2466),
  },
  [STATE(550)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(551)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(552)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(553)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(554)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(555)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(556)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(557)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(558)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(559)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(560)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(561)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(562)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(563)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(564)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(565)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token2] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATsynthesize] = ACTIONS(2410),
    [anon_sym_ATdynamic] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(566)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token2] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATsynthesize] = ACTIONS(2352),
    [anon_sym_ATdynamic] = ACTIONS(2352),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(567)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token2] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(568)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token2] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(569)] = {
    [sym_identifier] = ACTIONS(2568),
    [aux_sym_preproc_include_token1] = ACTIONS(2568),
    [aux_sym_preproc_include_token2] = ACTIONS(2568),
    [aux_sym_preproc_def_token1] = ACTIONS(2568),
    [aux_sym_preproc_if_token1] = ACTIONS(2568),
    [aux_sym_preproc_if_token2] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2568),
    [sym_preproc_directive] = ACTIONS(2568),
    [anon_sym_LPAREN2] = ACTIONS(2570),
    [anon_sym_BANG] = ACTIONS(2570),
    [anon_sym_TILDE] = ACTIONS(2570),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_PLUS] = ACTIONS(2568),
    [anon_sym_STAR] = ACTIONS(2570),
    [anon_sym_CARET] = ACTIONS(2570),
    [anon_sym_AMP] = ACTIONS(2570),
    [anon_sym_SEMI] = ACTIONS(2570),
    [anon_sym___extension__] = ACTIONS(2568),
    [anon_sym_typedef] = ACTIONS(2568),
    [anon_sym_extern] = ACTIONS(2568),
    [anon_sym___attribute__] = ACTIONS(2568),
    [anon_sym___attribute] = ACTIONS(2568),
    [anon_sym_noreturn] = ACTIONS(2568),
    [anon_sym_LBRACK] = ACTIONS(2570),
    [anon_sym___declspec] = ACTIONS(2568),
    [anon_sym___cdecl] = ACTIONS(2568),
    [anon_sym___clrcall] = ACTIONS(2568),
    [anon_sym___stdcall] = ACTIONS(2568),
    [anon_sym___fastcall] = ACTIONS(2568),
    [anon_sym___thiscall] = ACTIONS(2568),
    [anon_sym___vectorcall] = ACTIONS(2568),
    [anon_sym_LBRACE] = ACTIONS(2570),
    [anon_sym_signed] = ACTIONS(2568),
    [anon_sym_unsigned] = ACTIONS(2568),
    [anon_sym_long] = ACTIONS(2568),
    [anon_sym_short] = ACTIONS(2568),
    [anon_sym_static] = ACTIONS(2568),
    [anon_sym_ATautoreleasepool] = ACTIONS(2570),
    [anon_sym_auto] = ACTIONS(2568),
    [anon_sym_register] = ACTIONS(2568),
    [anon_sym_inline] = ACTIONS(2568),
    [anon_sym___inline] = ACTIONS(2568),
    [anon_sym___inline__] = ACTIONS(2568),
    [anon_sym___forceinline] = ACTIONS(2568),
    [anon_sym_thread_local] = ACTIONS(2568),
    [anon_sym___thread] = ACTIONS(2568),
    [anon_sym_CG_EXTERN] = ACTIONS(2568),
    [anon_sym_CG_INLINE] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2568),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2568),
    [anon_sym_IBOutlet] = ACTIONS(2568),
    [anon_sym_IBInspectable] = ACTIONS(2568),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2568),
    [anon_sym_NS_INLINE] = ACTIONS(2568),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2568),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2568),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2568),
    [anon_sym_const] = ACTIONS(2568),
    [anon_sym_constexpr] = ACTIONS(2568),
    [anon_sym_volatile] = ACTIONS(2568),
    [anon_sym_restrict] = ACTIONS(2568),
    [anon_sym___restrict__] = ACTIONS(2568),
    [anon_sym__Atomic] = ACTIONS(2568),
    [anon_sym__Noreturn] = ACTIONS(2568),
    [anon_sym__Nonnull] = ACTIONS(2568),
    [anon_sym_nullable] = ACTIONS(2568),
    [anon_sym__Complex] = ACTIONS(2568),
    [anon_sym__Nullable] = ACTIONS(2568),
    [anon_sym__Nullable_result] = ACTIONS(2568),
    [anon_sym__Null_unspecified] = ACTIONS(2568),
    [anon_sym___autoreleasing] = ACTIONS(2568),
    [anon_sym___block] = ACTIONS(2568),
    [anon_sym___bridge] = ACTIONS(2568),
    [anon_sym___bridge_retained] = ACTIONS(2568),
    [anon_sym___bridge_transfer] = ACTIONS(2568),
    [anon_sym___complex] = ACTIONS(2568),
    [anon_sym___const] = ACTIONS(2568),
    [anon_sym___imag] = ACTIONS(2568),
    [anon_sym___kindof] = ACTIONS(2568),
    [anon_sym___nonnull] = ACTIONS(2568),
    [anon_sym___nullable] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2568),
    [anon_sym___real] = ACTIONS(2568),
    [anon_sym___strong] = ACTIONS(2568),
    [anon_sym___unsafe_unretained] = ACTIONS(2568),
    [anon_sym___unused] = ACTIONS(2568),
    [anon_sym___weak] = ACTIONS(2568),
    [anon_sym_alignas] = ACTIONS(2568),
    [anon_sym__Alignas] = ACTIONS(2568),
    [sym_primitive_type] = ACTIONS(2568),
    [anon_sym_enum] = ACTIONS(2568),
    [anon_sym_struct] = ACTIONS(2568),
    [anon_sym_union] = ACTIONS(2568),
    [anon_sym_if] = ACTIONS(2568),
    [anon_sym_else] = ACTIONS(2568),
    [anon_sym_switch] = ACTIONS(2568),
    [anon_sym_case] = ACTIONS(2568),
    [anon_sym_default] = ACTIONS(2568),
    [anon_sym_while] = ACTIONS(2568),
    [anon_sym_do] = ACTIONS(2568),
    [anon_sym_for] = ACTIONS(2568),
    [anon_sym_in] = ACTIONS(2568),
    [anon_sym_return] = ACTIONS(2568),
    [anon_sym_break] = ACTIONS(2568),
    [anon_sym_continue] = ACTIONS(2568),
    [anon_sym_goto] = ACTIONS(2568),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_sizeof] = ACTIONS(2568),
    [anon_sym___alignof__] = ACTIONS(2568),
    [anon_sym___alignof] = ACTIONS(2568),
    [anon_sym__alignof] = ACTIONS(2568),
    [anon_sym_alignof] = ACTIONS(2568),
    [anon_sym__Alignof] = ACTIONS(2568),
    [anon_sym_offsetof] = ACTIONS(2568),
    [anon_sym__Generic] = ACTIONS(2568),
    [anon_sym_asm] = ACTIONS(2568),
    [anon_sym___asm__] = ACTIONS(2568),
    [anon_sym___asm] = ACTIONS(2568),
    [sym_number_literal] = ACTIONS(2570),
    [anon_sym_L_SQUOTE] = ACTIONS(2570),
    [anon_sym_u_SQUOTE] = ACTIONS(2570),
    [anon_sym_U_SQUOTE] = ACTIONS(2570),
    [anon_sym_u8_SQUOTE] = ACTIONS(2570),
    [anon_sym_SQUOTE] = ACTIONS(2570),
    [anon_sym_AT] = ACTIONS(2568),
    [anon_sym_DQUOTE] = ACTIONS(2570),
    [anon_sym_L_DQUOTE] = ACTIONS(2570),
    [anon_sym_u_DQUOTE] = ACTIONS(2570),
    [anon_sym_U_DQUOTE] = ACTIONS(2570),
    [anon_sym_u8_DQUOTE] = ACTIONS(2570),
    [sym_true] = ACTIONS(2568),
    [sym_false] = ACTIONS(2568),
    [anon_sym_NULL] = ACTIONS(2568),
    [anon_sym_nullptr] = ACTIONS(2568),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2568),
    [anon_sym___typeof] = ACTIONS(2568),
    [anon_sym_typeof] = ACTIONS(2568),
    [anon_sym_ATimport] = ACTIONS(2570),
    [aux_sym_preproc_undef_token1] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2568),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2568),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2568),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2568),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_API_AVAILABLE] = ACTIONS(2568),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_API_DEPRECATED] = ACTIONS(2568),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym___deprecated_msg] = ACTIONS(2568),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2568),
    [anon_sym_ATprotocol] = ACTIONS(2570),
    [anon_sym_ATinterface] = ACTIONS(2570),
    [anon_sym_ATimplementation] = ACTIONS(2570),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2570),
    [anon_sym_ATsynthesize] = ACTIONS(2570),
    [anon_sym_ATdynamic] = ACTIONS(2570),
    [anon_sym_ATtry] = ACTIONS(2570),
    [anon_sym___try] = ACTIONS(2568),
    [anon_sym_ATthrow] = ACTIONS(2570),
    [anon_sym_ATselector] = ACTIONS(2570),
    [anon_sym_ATavailable] = ACTIONS(2570),
    [anon_sym___builtin_available] = ACTIONS(2568),
    [anon_sym_va_arg] = ACTIONS(2568),
    [anon_sym_ATencode] = ACTIONS(2570),
    [anon_sym_ATsynchronized] = ACTIONS(2570),
    [anon_sym_BOOL] = ACTIONS(2568),
    [anon_sym_IMP] = ACTIONS(2568),
    [anon_sym_SEL] = ACTIONS(2568),
    [anon_sym_Class] = ACTIONS(2568),
    [anon_sym_id] = ACTIONS(2568),
  },
  [STATE(570)] = {
    [sym_identifier] = ACTIONS(2478),
    [aux_sym_preproc_include_token1] = ACTIONS(2478),
    [aux_sym_preproc_include_token2] = ACTIONS(2478),
    [aux_sym_preproc_def_token1] = ACTIONS(2478),
    [aux_sym_preproc_if_token1] = ACTIONS(2478),
    [aux_sym_preproc_if_token2] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2478),
    [sym_preproc_directive] = ACTIONS(2478),
    [anon_sym_LPAREN2] = ACTIONS(2480),
    [anon_sym_BANG] = ACTIONS(2480),
    [anon_sym_TILDE] = ACTIONS(2480),
    [anon_sym_DASH] = ACTIONS(2478),
    [anon_sym_PLUS] = ACTIONS(2478),
    [anon_sym_STAR] = ACTIONS(2480),
    [anon_sym_CARET] = ACTIONS(2480),
    [anon_sym_AMP] = ACTIONS(2480),
    [anon_sym_SEMI] = ACTIONS(2480),
    [anon_sym___extension__] = ACTIONS(2478),
    [anon_sym_typedef] = ACTIONS(2478),
    [anon_sym_extern] = ACTIONS(2478),
    [anon_sym___attribute__] = ACTIONS(2478),
    [anon_sym___attribute] = ACTIONS(2478),
    [anon_sym_noreturn] = ACTIONS(2478),
    [anon_sym_LBRACK] = ACTIONS(2480),
    [anon_sym___declspec] = ACTIONS(2478),
    [anon_sym___cdecl] = ACTIONS(2478),
    [anon_sym___clrcall] = ACTIONS(2478),
    [anon_sym___stdcall] = ACTIONS(2478),
    [anon_sym___fastcall] = ACTIONS(2478),
    [anon_sym___thiscall] = ACTIONS(2478),
    [anon_sym___vectorcall] = ACTIONS(2478),
    [anon_sym_LBRACE] = ACTIONS(2480),
    [anon_sym_signed] = ACTIONS(2478),
    [anon_sym_unsigned] = ACTIONS(2478),
    [anon_sym_long] = ACTIONS(2478),
    [anon_sym_short] = ACTIONS(2478),
    [anon_sym_static] = ACTIONS(2478),
    [anon_sym_ATautoreleasepool] = ACTIONS(2480),
    [anon_sym_auto] = ACTIONS(2478),
    [anon_sym_register] = ACTIONS(2478),
    [anon_sym_inline] = ACTIONS(2478),
    [anon_sym___inline] = ACTIONS(2478),
    [anon_sym___inline__] = ACTIONS(2478),
    [anon_sym___forceinline] = ACTIONS(2478),
    [anon_sym_thread_local] = ACTIONS(2478),
    [anon_sym___thread] = ACTIONS(2478),
    [anon_sym_CG_EXTERN] = ACTIONS(2478),
    [anon_sym_CG_INLINE] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2478),
    [anon_sym_IBOutlet] = ACTIONS(2478),
    [anon_sym_IBInspectable] = ACTIONS(2478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2478),
    [anon_sym_NS_INLINE] = ACTIONS(2478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2478),
    [anon_sym_const] = ACTIONS(2478),
    [anon_sym_constexpr] = ACTIONS(2478),
    [anon_sym_volatile] = ACTIONS(2478),
    [anon_sym_restrict] = ACTIONS(2478),
    [anon_sym___restrict__] = ACTIONS(2478),
    [anon_sym__Atomic] = ACTIONS(2478),
    [anon_sym__Noreturn] = ACTIONS(2478),
    [anon_sym__Nonnull] = ACTIONS(2478),
    [anon_sym_nullable] = ACTIONS(2478),
    [anon_sym__Complex] = ACTIONS(2478),
    [anon_sym__Nullable] = ACTIONS(2478),
    [anon_sym__Nullable_result] = ACTIONS(2478),
    [anon_sym__Null_unspecified] = ACTIONS(2478),
    [anon_sym___autoreleasing] = ACTIONS(2478),
    [anon_sym___block] = ACTIONS(2478),
    [anon_sym___bridge] = ACTIONS(2478),
    [anon_sym___bridge_retained] = ACTIONS(2478),
    [anon_sym___bridge_transfer] = ACTIONS(2478),
    [anon_sym___complex] = ACTIONS(2478),
    [anon_sym___const] = ACTIONS(2478),
    [anon_sym___imag] = ACTIONS(2478),
    [anon_sym___kindof] = ACTIONS(2478),
    [anon_sym___nonnull] = ACTIONS(2478),
    [anon_sym___nullable] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2478),
    [anon_sym___real] = ACTIONS(2478),
    [anon_sym___strong] = ACTIONS(2478),
    [anon_sym___unsafe_unretained] = ACTIONS(2478),
    [anon_sym___unused] = ACTIONS(2478),
    [anon_sym___weak] = ACTIONS(2478),
    [anon_sym_alignas] = ACTIONS(2478),
    [anon_sym__Alignas] = ACTIONS(2478),
    [sym_primitive_type] = ACTIONS(2478),
    [anon_sym_enum] = ACTIONS(2478),
    [anon_sym_struct] = ACTIONS(2478),
    [anon_sym_union] = ACTIONS(2478),
    [anon_sym_if] = ACTIONS(2478),
    [anon_sym_else] = ACTIONS(2478),
    [anon_sym_switch] = ACTIONS(2478),
    [anon_sym_case] = ACTIONS(2478),
    [anon_sym_default] = ACTIONS(2478),
    [anon_sym_while] = ACTIONS(2478),
    [anon_sym_do] = ACTIONS(2478),
    [anon_sym_for] = ACTIONS(2478),
    [anon_sym_in] = ACTIONS(2478),
    [anon_sym_return] = ACTIONS(2478),
    [anon_sym_break] = ACTIONS(2478),
    [anon_sym_continue] = ACTIONS(2478),
    [anon_sym_goto] = ACTIONS(2478),
    [anon_sym_DASH_DASH] = ACTIONS(2480),
    [anon_sym_PLUS_PLUS] = ACTIONS(2480),
    [anon_sym_sizeof] = ACTIONS(2478),
    [anon_sym___alignof__] = ACTIONS(2478),
    [anon_sym___alignof] = ACTIONS(2478),
    [anon_sym__alignof] = ACTIONS(2478),
    [anon_sym_alignof] = ACTIONS(2478),
    [anon_sym__Alignof] = ACTIONS(2478),
    [anon_sym_offsetof] = ACTIONS(2478),
    [anon_sym__Generic] = ACTIONS(2478),
    [anon_sym_asm] = ACTIONS(2478),
    [anon_sym___asm__] = ACTIONS(2478),
    [anon_sym___asm] = ACTIONS(2478),
    [sym_number_literal] = ACTIONS(2480),
    [anon_sym_L_SQUOTE] = ACTIONS(2480),
    [anon_sym_u_SQUOTE] = ACTIONS(2480),
    [anon_sym_U_SQUOTE] = ACTIONS(2480),
    [anon_sym_u8_SQUOTE] = ACTIONS(2480),
    [anon_sym_SQUOTE] = ACTIONS(2480),
    [anon_sym_AT] = ACTIONS(2478),
    [anon_sym_DQUOTE] = ACTIONS(2480),
    [anon_sym_L_DQUOTE] = ACTIONS(2480),
    [anon_sym_u_DQUOTE] = ACTIONS(2480),
    [anon_sym_U_DQUOTE] = ACTIONS(2480),
    [anon_sym_u8_DQUOTE] = ACTIONS(2480),
    [sym_true] = ACTIONS(2478),
    [sym_false] = ACTIONS(2478),
    [anon_sym_NULL] = ACTIONS(2478),
    [anon_sym_nullptr] = ACTIONS(2478),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2478),
    [anon_sym___typeof] = ACTIONS(2478),
    [anon_sym_typeof] = ACTIONS(2478),
    [anon_sym_ATimport] = ACTIONS(2480),
    [aux_sym_preproc_undef_token1] = ACTIONS(2478),
    [anon_sym_POUND] = ACTIONS(2478),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_API_AVAILABLE] = ACTIONS(2478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_API_DEPRECATED] = ACTIONS(2478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym___deprecated_msg] = ACTIONS(2478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2478),
    [anon_sym_ATprotocol] = ACTIONS(2480),
    [anon_sym_ATinterface] = ACTIONS(2480),
    [anon_sym_ATimplementation] = ACTIONS(2480),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2480),
    [anon_sym_ATsynthesize] = ACTIONS(2480),
    [anon_sym_ATdynamic] = ACTIONS(2480),
    [anon_sym_ATtry] = ACTIONS(2480),
    [anon_sym___try] = ACTIONS(2478),
    [anon_sym_ATthrow] = ACTIONS(2480),
    [anon_sym_ATselector] = ACTIONS(2480),
    [anon_sym_ATavailable] = ACTIONS(2480),
    [anon_sym___builtin_available] = ACTIONS(2478),
    [anon_sym_va_arg] = ACTIONS(2478),
    [anon_sym_ATencode] = ACTIONS(2480),
    [anon_sym_ATsynchronized] = ACTIONS(2480),
    [anon_sym_BOOL] = ACTIONS(2478),
    [anon_sym_IMP] = ACTIONS(2478),
    [anon_sym_SEL] = ACTIONS(2478),
    [anon_sym_Class] = ACTIONS(2478),
    [anon_sym_id] = ACTIONS(2478),
  },
  [STATE(571)] = {
    [sym_identifier] = ACTIONS(2412),
    [aux_sym_preproc_include_token1] = ACTIONS(2412),
    [aux_sym_preproc_include_token2] = ACTIONS(2412),
    [aux_sym_preproc_def_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token2] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2412),
    [sym_preproc_directive] = ACTIONS(2412),
    [anon_sym_LPAREN2] = ACTIONS(2414),
    [anon_sym_BANG] = ACTIONS(2414),
    [anon_sym_TILDE] = ACTIONS(2414),
    [anon_sym_DASH] = ACTIONS(2412),
    [anon_sym_PLUS] = ACTIONS(2412),
    [anon_sym_STAR] = ACTIONS(2414),
    [anon_sym_CARET] = ACTIONS(2414),
    [anon_sym_AMP] = ACTIONS(2414),
    [anon_sym_SEMI] = ACTIONS(2414),
    [anon_sym___extension__] = ACTIONS(2412),
    [anon_sym_typedef] = ACTIONS(2412),
    [anon_sym_extern] = ACTIONS(2412),
    [anon_sym___attribute__] = ACTIONS(2412),
    [anon_sym___attribute] = ACTIONS(2412),
    [anon_sym_noreturn] = ACTIONS(2412),
    [anon_sym_LBRACK] = ACTIONS(2414),
    [anon_sym___declspec] = ACTIONS(2412),
    [anon_sym___cdecl] = ACTIONS(2412),
    [anon_sym___clrcall] = ACTIONS(2412),
    [anon_sym___stdcall] = ACTIONS(2412),
    [anon_sym___fastcall] = ACTIONS(2412),
    [anon_sym___thiscall] = ACTIONS(2412),
    [anon_sym___vectorcall] = ACTIONS(2412),
    [anon_sym_LBRACE] = ACTIONS(2414),
    [anon_sym_signed] = ACTIONS(2412),
    [anon_sym_unsigned] = ACTIONS(2412),
    [anon_sym_long] = ACTIONS(2412),
    [anon_sym_short] = ACTIONS(2412),
    [anon_sym_static] = ACTIONS(2412),
    [anon_sym_ATautoreleasepool] = ACTIONS(2414),
    [anon_sym_auto] = ACTIONS(2412),
    [anon_sym_register] = ACTIONS(2412),
    [anon_sym_inline] = ACTIONS(2412),
    [anon_sym___inline] = ACTIONS(2412),
    [anon_sym___inline__] = ACTIONS(2412),
    [anon_sym___forceinline] = ACTIONS(2412),
    [anon_sym_thread_local] = ACTIONS(2412),
    [anon_sym___thread] = ACTIONS(2412),
    [anon_sym_CG_EXTERN] = ACTIONS(2412),
    [anon_sym_CG_INLINE] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2412),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2412),
    [anon_sym_IBOutlet] = ACTIONS(2412),
    [anon_sym_IBInspectable] = ACTIONS(2412),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2412),
    [anon_sym_NS_INLINE] = ACTIONS(2412),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2412),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2412),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2412),
    [anon_sym_const] = ACTIONS(2412),
    [anon_sym_constexpr] = ACTIONS(2412),
    [anon_sym_volatile] = ACTIONS(2412),
    [anon_sym_restrict] = ACTIONS(2412),
    [anon_sym___restrict__] = ACTIONS(2412),
    [anon_sym__Atomic] = ACTIONS(2412),
    [anon_sym__Noreturn] = ACTIONS(2412),
    [anon_sym__Nonnull] = ACTIONS(2412),
    [anon_sym_nullable] = ACTIONS(2412),
    [anon_sym__Complex] = ACTIONS(2412),
    [anon_sym__Nullable] = ACTIONS(2412),
    [anon_sym__Nullable_result] = ACTIONS(2412),
    [anon_sym__Null_unspecified] = ACTIONS(2412),
    [anon_sym___autoreleasing] = ACTIONS(2412),
    [anon_sym___block] = ACTIONS(2412),
    [anon_sym___bridge] = ACTIONS(2412),
    [anon_sym___bridge_retained] = ACTIONS(2412),
    [anon_sym___bridge_transfer] = ACTIONS(2412),
    [anon_sym___complex] = ACTIONS(2412),
    [anon_sym___const] = ACTIONS(2412),
    [anon_sym___imag] = ACTIONS(2412),
    [anon_sym___kindof] = ACTIONS(2412),
    [anon_sym___nonnull] = ACTIONS(2412),
    [anon_sym___nullable] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2412),
    [anon_sym___real] = ACTIONS(2412),
    [anon_sym___strong] = ACTIONS(2412),
    [anon_sym___unsafe_unretained] = ACTIONS(2412),
    [anon_sym___unused] = ACTIONS(2412),
    [anon_sym___weak] = ACTIONS(2412),
    [anon_sym_alignas] = ACTIONS(2412),
    [anon_sym__Alignas] = ACTIONS(2412),
    [sym_primitive_type] = ACTIONS(2412),
    [anon_sym_enum] = ACTIONS(2412),
    [anon_sym_struct] = ACTIONS(2412),
    [anon_sym_union] = ACTIONS(2412),
    [anon_sym_if] = ACTIONS(2412),
    [anon_sym_else] = ACTIONS(2412),
    [anon_sym_switch] = ACTIONS(2412),
    [anon_sym_case] = ACTIONS(2412),
    [anon_sym_default] = ACTIONS(2412),
    [anon_sym_while] = ACTIONS(2412),
    [anon_sym_do] = ACTIONS(2412),
    [anon_sym_for] = ACTIONS(2412),
    [anon_sym_in] = ACTIONS(2412),
    [anon_sym_return] = ACTIONS(2412),
    [anon_sym_break] = ACTIONS(2412),
    [anon_sym_continue] = ACTIONS(2412),
    [anon_sym_goto] = ACTIONS(2412),
    [anon_sym_DASH_DASH] = ACTIONS(2414),
    [anon_sym_PLUS_PLUS] = ACTIONS(2414),
    [anon_sym_sizeof] = ACTIONS(2412),
    [anon_sym___alignof__] = ACTIONS(2412),
    [anon_sym___alignof] = ACTIONS(2412),
    [anon_sym__alignof] = ACTIONS(2412),
    [anon_sym_alignof] = ACTIONS(2412),
    [anon_sym__Alignof] = ACTIONS(2412),
    [anon_sym_offsetof] = ACTIONS(2412),
    [anon_sym__Generic] = ACTIONS(2412),
    [anon_sym_asm] = ACTIONS(2412),
    [anon_sym___asm__] = ACTIONS(2412),
    [anon_sym___asm] = ACTIONS(2412),
    [sym_number_literal] = ACTIONS(2414),
    [anon_sym_L_SQUOTE] = ACTIONS(2414),
    [anon_sym_u_SQUOTE] = ACTIONS(2414),
    [anon_sym_U_SQUOTE] = ACTIONS(2414),
    [anon_sym_u8_SQUOTE] = ACTIONS(2414),
    [anon_sym_SQUOTE] = ACTIONS(2414),
    [anon_sym_AT] = ACTIONS(2412),
    [anon_sym_DQUOTE] = ACTIONS(2414),
    [anon_sym_L_DQUOTE] = ACTIONS(2414),
    [anon_sym_u_DQUOTE] = ACTIONS(2414),
    [anon_sym_U_DQUOTE] = ACTIONS(2414),
    [anon_sym_u8_DQUOTE] = ACTIONS(2414),
    [sym_true] = ACTIONS(2412),
    [sym_false] = ACTIONS(2412),
    [anon_sym_NULL] = ACTIONS(2412),
    [anon_sym_nullptr] = ACTIONS(2412),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2412),
    [anon_sym___typeof] = ACTIONS(2412),
    [anon_sym_typeof] = ACTIONS(2412),
    [anon_sym_ATimport] = ACTIONS(2414),
    [aux_sym_preproc_undef_token1] = ACTIONS(2412),
    [anon_sym_POUND] = ACTIONS(2412),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2412),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2412),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2412),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_API_AVAILABLE] = ACTIONS(2412),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_API_DEPRECATED] = ACTIONS(2412),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym___deprecated_msg] = ACTIONS(2412),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2412),
    [anon_sym_ATprotocol] = ACTIONS(2414),
    [anon_sym_ATinterface] = ACTIONS(2414),
    [anon_sym_ATimplementation] = ACTIONS(2414),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2414),
    [anon_sym_ATsynthesize] = ACTIONS(2414),
    [anon_sym_ATdynamic] = ACTIONS(2414),
    [anon_sym_ATtry] = ACTIONS(2414),
    [anon_sym___try] = ACTIONS(2412),
    [anon_sym_ATthrow] = ACTIONS(2414),
    [anon_sym_ATselector] = ACTIONS(2414),
    [anon_sym_ATavailable] = ACTIONS(2414),
    [anon_sym___builtin_available] = ACTIONS(2412),
    [anon_sym_va_arg] = ACTIONS(2412),
    [anon_sym_ATencode] = ACTIONS(2414),
    [anon_sym_ATsynchronized] = ACTIONS(2414),
    [anon_sym_BOOL] = ACTIONS(2412),
    [anon_sym_IMP] = ACTIONS(2412),
    [anon_sym_SEL] = ACTIONS(2412),
    [anon_sym_Class] = ACTIONS(2412),
    [anon_sym_id] = ACTIONS(2412),
  },
  [STATE(572)] = {
    [sym_catch_clause] = STATE(572),
    [aux_sym_try_statement_repeat1] = STATE(572),
    [ts_builtin_sym_end] = ACTIONS(2320),
    [sym_identifier] = ACTIONS(2318),
    [aux_sym_preproc_include_token1] = ACTIONS(2318),
    [aux_sym_preproc_include_token2] = ACTIONS(2318),
    [aux_sym_preproc_def_token1] = ACTIONS(2318),
    [aux_sym_preproc_if_token1] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2318),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2318),
    [sym_preproc_directive] = ACTIONS(2318),
    [anon_sym_LPAREN2] = ACTIONS(2320),
    [anon_sym_BANG] = ACTIONS(2320),
    [anon_sym_TILDE] = ACTIONS(2320),
    [anon_sym_DASH] = ACTIONS(2318),
    [anon_sym_PLUS] = ACTIONS(2318),
    [anon_sym_STAR] = ACTIONS(2320),
    [anon_sym_CARET] = ACTIONS(2320),
    [anon_sym_AMP] = ACTIONS(2320),
    [anon_sym_SEMI] = ACTIONS(2320),
    [anon_sym___extension__] = ACTIONS(2318),
    [anon_sym_typedef] = ACTIONS(2318),
    [anon_sym_extern] = ACTIONS(2318),
    [anon_sym___attribute__] = ACTIONS(2318),
    [anon_sym___attribute] = ACTIONS(2318),
    [anon_sym_noreturn] = ACTIONS(2318),
    [anon_sym_LBRACK] = ACTIONS(2320),
    [anon_sym___declspec] = ACTIONS(2318),
    [anon_sym___cdecl] = ACTIONS(2318),
    [anon_sym___clrcall] = ACTIONS(2318),
    [anon_sym___stdcall] = ACTIONS(2318),
    [anon_sym___fastcall] = ACTIONS(2318),
    [anon_sym___thiscall] = ACTIONS(2318),
    [anon_sym___vectorcall] = ACTIONS(2318),
    [anon_sym_LBRACE] = ACTIONS(2320),
    [anon_sym_signed] = ACTIONS(2318),
    [anon_sym_unsigned] = ACTIONS(2318),
    [anon_sym_long] = ACTIONS(2318),
    [anon_sym_short] = ACTIONS(2318),
    [anon_sym_static] = ACTIONS(2318),
    [anon_sym_ATautoreleasepool] = ACTIONS(2320),
    [anon_sym_auto] = ACTIONS(2318),
    [anon_sym_register] = ACTIONS(2318),
    [anon_sym_inline] = ACTIONS(2318),
    [anon_sym___inline] = ACTIONS(2318),
    [anon_sym___inline__] = ACTIONS(2318),
    [anon_sym___forceinline] = ACTIONS(2318),
    [anon_sym_thread_local] = ACTIONS(2318),
    [anon_sym___thread] = ACTIONS(2318),
    [anon_sym_CG_EXTERN] = ACTIONS(2318),
    [anon_sym_CG_INLINE] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2318),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2318),
    [anon_sym_IBOutlet] = ACTIONS(2318),
    [anon_sym_IBInspectable] = ACTIONS(2318),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2318),
    [anon_sym_NS_INLINE] = ACTIONS(2318),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2318),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2318),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2318),
    [anon_sym_const] = ACTIONS(2318),
    [anon_sym_constexpr] = ACTIONS(2318),
    [anon_sym_volatile] = ACTIONS(2318),
    [anon_sym_restrict] = ACTIONS(2318),
    [anon_sym___restrict__] = ACTIONS(2318),
    [anon_sym__Atomic] = ACTIONS(2318),
    [anon_sym__Noreturn] = ACTIONS(2318),
    [anon_sym__Nonnull] = ACTIONS(2318),
    [anon_sym_nullable] = ACTIONS(2318),
    [anon_sym__Complex] = ACTIONS(2318),
    [anon_sym__Nullable] = ACTIONS(2318),
    [anon_sym__Nullable_result] = ACTIONS(2318),
    [anon_sym__Null_unspecified] = ACTIONS(2318),
    [anon_sym___autoreleasing] = ACTIONS(2318),
    [anon_sym___block] = ACTIONS(2318),
    [anon_sym___bridge] = ACTIONS(2318),
    [anon_sym___bridge_retained] = ACTIONS(2318),
    [anon_sym___bridge_transfer] = ACTIONS(2318),
    [anon_sym___complex] = ACTIONS(2318),
    [anon_sym___const] = ACTIONS(2318),
    [anon_sym___imag] = ACTIONS(2318),
    [anon_sym___kindof] = ACTIONS(2318),
    [anon_sym___nonnull] = ACTIONS(2318),
    [anon_sym___nullable] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2318),
    [anon_sym___real] = ACTIONS(2318),
    [anon_sym___strong] = ACTIONS(2318),
    [anon_sym___unsafe_unretained] = ACTIONS(2318),
    [anon_sym___unused] = ACTIONS(2318),
    [anon_sym___weak] = ACTIONS(2318),
    [anon_sym_alignas] = ACTIONS(2318),
    [anon_sym__Alignas] = ACTIONS(2318),
    [sym_primitive_type] = ACTIONS(2318),
    [anon_sym_enum] = ACTIONS(2318),
    [anon_sym_struct] = ACTIONS(2318),
    [anon_sym_union] = ACTIONS(2318),
    [anon_sym_if] = ACTIONS(2318),
    [anon_sym_else] = ACTIONS(2318),
    [anon_sym_switch] = ACTIONS(2318),
    [anon_sym_case] = ACTIONS(2318),
    [anon_sym_default] = ACTIONS(2318),
    [anon_sym_while] = ACTIONS(2318),
    [anon_sym_do] = ACTIONS(2318),
    [anon_sym_for] = ACTIONS(2318),
    [anon_sym_in] = ACTIONS(2318),
    [anon_sym_return] = ACTIONS(2318),
    [anon_sym_break] = ACTIONS(2318),
    [anon_sym_continue] = ACTIONS(2318),
    [anon_sym_goto] = ACTIONS(2318),
    [anon_sym_DASH_DASH] = ACTIONS(2320),
    [anon_sym_PLUS_PLUS] = ACTIONS(2320),
    [anon_sym_sizeof] = ACTIONS(2318),
    [anon_sym___alignof__] = ACTIONS(2318),
    [anon_sym___alignof] = ACTIONS(2318),
    [anon_sym__alignof] = ACTIONS(2318),
    [anon_sym_alignof] = ACTIONS(2318),
    [anon_sym__Alignof] = ACTIONS(2318),
    [anon_sym_offsetof] = ACTIONS(2318),
    [anon_sym__Generic] = ACTIONS(2318),
    [anon_sym_asm] = ACTIONS(2318),
    [anon_sym___asm__] = ACTIONS(2318),
    [anon_sym___asm] = ACTIONS(2318),
    [sym_number_literal] = ACTIONS(2320),
    [anon_sym_L_SQUOTE] = ACTIONS(2320),
    [anon_sym_u_SQUOTE] = ACTIONS(2320),
    [anon_sym_U_SQUOTE] = ACTIONS(2320),
    [anon_sym_u8_SQUOTE] = ACTIONS(2320),
    [anon_sym_SQUOTE] = ACTIONS(2320),
    [anon_sym_AT] = ACTIONS(2318),
    [anon_sym_DQUOTE] = ACTIONS(2320),
    [anon_sym_L_DQUOTE] = ACTIONS(2320),
    [anon_sym_u_DQUOTE] = ACTIONS(2320),
    [anon_sym_U_DQUOTE] = ACTIONS(2320),
    [anon_sym_u8_DQUOTE] = ACTIONS(2320),
    [sym_true] = ACTIONS(2318),
    [sym_false] = ACTIONS(2318),
    [anon_sym_NULL] = ACTIONS(2318),
    [anon_sym_nullptr] = ACTIONS(2318),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2318),
    [anon_sym___typeof] = ACTIONS(2318),
    [anon_sym_typeof] = ACTIONS(2318),
    [anon_sym_ATimport] = ACTIONS(2320),
    [aux_sym_preproc_undef_token1] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2318),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2318),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2318),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2318),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_API_AVAILABLE] = ACTIONS(2318),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_API_DEPRECATED] = ACTIONS(2318),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym___deprecated_msg] = ACTIONS(2318),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2318),
    [anon_sym_ATprotocol] = ACTIONS(2320),
    [anon_sym_ATinterface] = ACTIONS(2320),
    [anon_sym_ATimplementation] = ACTIONS(2320),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2320),
    [anon_sym_ATcatch] = ACTIONS(2878),
    [anon_sym___catch] = ACTIONS(2881),
    [anon_sym_ATfinally] = ACTIONS(2320),
    [anon_sym___finally] = ACTIONS(2318),
    [anon_sym_ATselector] = ACTIONS(2320),
    [anon_sym_ATavailable] = ACTIONS(2320),
    [anon_sym___builtin_available] = ACTIONS(2318),
    [anon_sym_va_arg] = ACTIONS(2318),
    [anon_sym_ATencode] = ACTIONS(2320),
    [anon_sym_BOOL] = ACTIONS(2318),
    [anon_sym_IMP] = ACTIONS(2318),
    [anon_sym_SEL] = ACTIONS(2318),
    [anon_sym_Class] = ACTIONS(2318),
    [anon_sym_id] = ACTIONS(2318),
  },
  [STATE(573)] = {
    [sym_identifier] = ACTIONS(2482),
    [aux_sym_preproc_include_token1] = ACTIONS(2482),
    [aux_sym_preproc_include_token2] = ACTIONS(2482),
    [aux_sym_preproc_def_token1] = ACTIONS(2482),
    [aux_sym_preproc_if_token1] = ACTIONS(2482),
    [aux_sym_preproc_if_token2] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2482),
    [sym_preproc_directive] = ACTIONS(2482),
    [anon_sym_LPAREN2] = ACTIONS(2484),
    [anon_sym_BANG] = ACTIONS(2484),
    [anon_sym_TILDE] = ACTIONS(2484),
    [anon_sym_DASH] = ACTIONS(2482),
    [anon_sym_PLUS] = ACTIONS(2482),
    [anon_sym_STAR] = ACTIONS(2484),
    [anon_sym_CARET] = ACTIONS(2484),
    [anon_sym_AMP] = ACTIONS(2484),
    [anon_sym_SEMI] = ACTIONS(2484),
    [anon_sym___extension__] = ACTIONS(2482),
    [anon_sym_typedef] = ACTIONS(2482),
    [anon_sym_extern] = ACTIONS(2482),
    [anon_sym___attribute__] = ACTIONS(2482),
    [anon_sym___attribute] = ACTIONS(2482),
    [anon_sym_noreturn] = ACTIONS(2482),
    [anon_sym_LBRACK] = ACTIONS(2484),
    [anon_sym___declspec] = ACTIONS(2482),
    [anon_sym___cdecl] = ACTIONS(2482),
    [anon_sym___clrcall] = ACTIONS(2482),
    [anon_sym___stdcall] = ACTIONS(2482),
    [anon_sym___fastcall] = ACTIONS(2482),
    [anon_sym___thiscall] = ACTIONS(2482),
    [anon_sym___vectorcall] = ACTIONS(2482),
    [anon_sym_LBRACE] = ACTIONS(2484),
    [anon_sym_signed] = ACTIONS(2482),
    [anon_sym_unsigned] = ACTIONS(2482),
    [anon_sym_long] = ACTIONS(2482),
    [anon_sym_short] = ACTIONS(2482),
    [anon_sym_static] = ACTIONS(2482),
    [anon_sym_ATautoreleasepool] = ACTIONS(2484),
    [anon_sym_auto] = ACTIONS(2482),
    [anon_sym_register] = ACTIONS(2482),
    [anon_sym_inline] = ACTIONS(2482),
    [anon_sym___inline] = ACTIONS(2482),
    [anon_sym___inline__] = ACTIONS(2482),
    [anon_sym___forceinline] = ACTIONS(2482),
    [anon_sym_thread_local] = ACTIONS(2482),
    [anon_sym___thread] = ACTIONS(2482),
    [anon_sym_CG_EXTERN] = ACTIONS(2482),
    [anon_sym_CG_INLINE] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2482),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2482),
    [anon_sym_IBOutlet] = ACTIONS(2482),
    [anon_sym_IBInspectable] = ACTIONS(2482),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2482),
    [anon_sym_NS_INLINE] = ACTIONS(2482),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2482),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2482),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2482),
    [anon_sym_const] = ACTIONS(2482),
    [anon_sym_constexpr] = ACTIONS(2482),
    [anon_sym_volatile] = ACTIONS(2482),
    [anon_sym_restrict] = ACTIONS(2482),
    [anon_sym___restrict__] = ACTIONS(2482),
    [anon_sym__Atomic] = ACTIONS(2482),
    [anon_sym__Noreturn] = ACTIONS(2482),
    [anon_sym__Nonnull] = ACTIONS(2482),
    [anon_sym_nullable] = ACTIONS(2482),
    [anon_sym__Complex] = ACTIONS(2482),
    [anon_sym__Nullable] = ACTIONS(2482),
    [anon_sym__Nullable_result] = ACTIONS(2482),
    [anon_sym__Null_unspecified] = ACTIONS(2482),
    [anon_sym___autoreleasing] = ACTIONS(2482),
    [anon_sym___block] = ACTIONS(2482),
    [anon_sym___bridge] = ACTIONS(2482),
    [anon_sym___bridge_retained] = ACTIONS(2482),
    [anon_sym___bridge_transfer] = ACTIONS(2482),
    [anon_sym___complex] = ACTIONS(2482),
    [anon_sym___const] = ACTIONS(2482),
    [anon_sym___imag] = ACTIONS(2482),
    [anon_sym___kindof] = ACTIONS(2482),
    [anon_sym___nonnull] = ACTIONS(2482),
    [anon_sym___nullable] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2482),
    [anon_sym___real] = ACTIONS(2482),
    [anon_sym___strong] = ACTIONS(2482),
    [anon_sym___unsafe_unretained] = ACTIONS(2482),
    [anon_sym___unused] = ACTIONS(2482),
    [anon_sym___weak] = ACTIONS(2482),
    [anon_sym_alignas] = ACTIONS(2482),
    [anon_sym__Alignas] = ACTIONS(2482),
    [sym_primitive_type] = ACTIONS(2482),
    [anon_sym_enum] = ACTIONS(2482),
    [anon_sym_struct] = ACTIONS(2482),
    [anon_sym_union] = ACTIONS(2482),
    [anon_sym_if] = ACTIONS(2482),
    [anon_sym_else] = ACTIONS(2482),
    [anon_sym_switch] = ACTIONS(2482),
    [anon_sym_case] = ACTIONS(2482),
    [anon_sym_default] = ACTIONS(2482),
    [anon_sym_while] = ACTIONS(2482),
    [anon_sym_do] = ACTIONS(2482),
    [anon_sym_for] = ACTIONS(2482),
    [anon_sym_in] = ACTIONS(2482),
    [anon_sym_return] = ACTIONS(2482),
    [anon_sym_break] = ACTIONS(2482),
    [anon_sym_continue] = ACTIONS(2482),
    [anon_sym_goto] = ACTIONS(2482),
    [anon_sym_DASH_DASH] = ACTIONS(2484),
    [anon_sym_PLUS_PLUS] = ACTIONS(2484),
    [anon_sym_sizeof] = ACTIONS(2482),
    [anon_sym___alignof__] = ACTIONS(2482),
    [anon_sym___alignof] = ACTIONS(2482),
    [anon_sym__alignof] = ACTIONS(2482),
    [anon_sym_alignof] = ACTIONS(2482),
    [anon_sym__Alignof] = ACTIONS(2482),
    [anon_sym_offsetof] = ACTIONS(2482),
    [anon_sym__Generic] = ACTIONS(2482),
    [anon_sym_asm] = ACTIONS(2482),
    [anon_sym___asm__] = ACTIONS(2482),
    [anon_sym___asm] = ACTIONS(2482),
    [sym_number_literal] = ACTIONS(2484),
    [anon_sym_L_SQUOTE] = ACTIONS(2484),
    [anon_sym_u_SQUOTE] = ACTIONS(2484),
    [anon_sym_U_SQUOTE] = ACTIONS(2484),
    [anon_sym_u8_SQUOTE] = ACTIONS(2484),
    [anon_sym_SQUOTE] = ACTIONS(2484),
    [anon_sym_AT] = ACTIONS(2482),
    [anon_sym_DQUOTE] = ACTIONS(2484),
    [anon_sym_L_DQUOTE] = ACTIONS(2484),
    [anon_sym_u_DQUOTE] = ACTIONS(2484),
    [anon_sym_U_DQUOTE] = ACTIONS(2484),
    [anon_sym_u8_DQUOTE] = ACTIONS(2484),
    [sym_true] = ACTIONS(2482),
    [sym_false] = ACTIONS(2482),
    [anon_sym_NULL] = ACTIONS(2482),
    [anon_sym_nullptr] = ACTIONS(2482),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2482),
    [anon_sym___typeof] = ACTIONS(2482),
    [anon_sym_typeof] = ACTIONS(2482),
    [anon_sym_ATimport] = ACTIONS(2484),
    [aux_sym_preproc_undef_token1] = ACTIONS(2482),
    [anon_sym_POUND] = ACTIONS(2482),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2482),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2482),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2482),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_API_AVAILABLE] = ACTIONS(2482),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_API_DEPRECATED] = ACTIONS(2482),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym___deprecated_msg] = ACTIONS(2482),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2482),
    [anon_sym_ATprotocol] = ACTIONS(2484),
    [anon_sym_ATinterface] = ACTIONS(2484),
    [anon_sym_ATimplementation] = ACTIONS(2484),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2484),
    [anon_sym_ATsynthesize] = ACTIONS(2484),
    [anon_sym_ATdynamic] = ACTIONS(2484),
    [anon_sym_ATtry] = ACTIONS(2484),
    [anon_sym___try] = ACTIONS(2482),
    [anon_sym_ATthrow] = ACTIONS(2484),
    [anon_sym_ATselector] = ACTIONS(2484),
    [anon_sym_ATavailable] = ACTIONS(2484),
    [anon_sym___builtin_available] = ACTIONS(2482),
    [anon_sym_va_arg] = ACTIONS(2482),
    [anon_sym_ATencode] = ACTIONS(2484),
    [anon_sym_ATsynchronized] = ACTIONS(2484),
    [anon_sym_BOOL] = ACTIONS(2482),
    [anon_sym_IMP] = ACTIONS(2482),
    [anon_sym_SEL] = ACTIONS(2482),
    [anon_sym_Class] = ACTIONS(2482),
    [anon_sym_id] = ACTIONS(2482),
  },
  [STATE(574)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token2] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(575)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token2] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(576)] = {
    [sym_identifier] = ACTIONS(2486),
    [aux_sym_preproc_include_token1] = ACTIONS(2486),
    [aux_sym_preproc_include_token2] = ACTIONS(2486),
    [aux_sym_preproc_def_token1] = ACTIONS(2486),
    [aux_sym_preproc_if_token1] = ACTIONS(2486),
    [aux_sym_preproc_if_token2] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2486),
    [sym_preproc_directive] = ACTIONS(2486),
    [anon_sym_LPAREN2] = ACTIONS(2488),
    [anon_sym_BANG] = ACTIONS(2488),
    [anon_sym_TILDE] = ACTIONS(2488),
    [anon_sym_DASH] = ACTIONS(2486),
    [anon_sym_PLUS] = ACTIONS(2486),
    [anon_sym_STAR] = ACTIONS(2488),
    [anon_sym_CARET] = ACTIONS(2488),
    [anon_sym_AMP] = ACTIONS(2488),
    [anon_sym_SEMI] = ACTIONS(2488),
    [anon_sym___extension__] = ACTIONS(2486),
    [anon_sym_typedef] = ACTIONS(2486),
    [anon_sym_extern] = ACTIONS(2486),
    [anon_sym___attribute__] = ACTIONS(2486),
    [anon_sym___attribute] = ACTIONS(2486),
    [anon_sym_noreturn] = ACTIONS(2486),
    [anon_sym_LBRACK] = ACTIONS(2488),
    [anon_sym___declspec] = ACTIONS(2486),
    [anon_sym___cdecl] = ACTIONS(2486),
    [anon_sym___clrcall] = ACTIONS(2486),
    [anon_sym___stdcall] = ACTIONS(2486),
    [anon_sym___fastcall] = ACTIONS(2486),
    [anon_sym___thiscall] = ACTIONS(2486),
    [anon_sym___vectorcall] = ACTIONS(2486),
    [anon_sym_LBRACE] = ACTIONS(2488),
    [anon_sym_signed] = ACTIONS(2486),
    [anon_sym_unsigned] = ACTIONS(2486),
    [anon_sym_long] = ACTIONS(2486),
    [anon_sym_short] = ACTIONS(2486),
    [anon_sym_static] = ACTIONS(2486),
    [anon_sym_ATautoreleasepool] = ACTIONS(2488),
    [anon_sym_auto] = ACTIONS(2486),
    [anon_sym_register] = ACTIONS(2486),
    [anon_sym_inline] = ACTIONS(2486),
    [anon_sym___inline] = ACTIONS(2486),
    [anon_sym___inline__] = ACTIONS(2486),
    [anon_sym___forceinline] = ACTIONS(2486),
    [anon_sym_thread_local] = ACTIONS(2486),
    [anon_sym___thread] = ACTIONS(2486),
    [anon_sym_CG_EXTERN] = ACTIONS(2486),
    [anon_sym_CG_INLINE] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2486),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2486),
    [anon_sym_IBOutlet] = ACTIONS(2486),
    [anon_sym_IBInspectable] = ACTIONS(2486),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2486),
    [anon_sym_NS_INLINE] = ACTIONS(2486),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2486),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2486),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2486),
    [anon_sym_const] = ACTIONS(2486),
    [anon_sym_constexpr] = ACTIONS(2486),
    [anon_sym_volatile] = ACTIONS(2486),
    [anon_sym_restrict] = ACTIONS(2486),
    [anon_sym___restrict__] = ACTIONS(2486),
    [anon_sym__Atomic] = ACTIONS(2486),
    [anon_sym__Noreturn] = ACTIONS(2486),
    [anon_sym__Nonnull] = ACTIONS(2486),
    [anon_sym_nullable] = ACTIONS(2486),
    [anon_sym__Complex] = ACTIONS(2486),
    [anon_sym__Nullable] = ACTIONS(2486),
    [anon_sym__Nullable_result] = ACTIONS(2486),
    [anon_sym__Null_unspecified] = ACTIONS(2486),
    [anon_sym___autoreleasing] = ACTIONS(2486),
    [anon_sym___block] = ACTIONS(2486),
    [anon_sym___bridge] = ACTIONS(2486),
    [anon_sym___bridge_retained] = ACTIONS(2486),
    [anon_sym___bridge_transfer] = ACTIONS(2486),
    [anon_sym___complex] = ACTIONS(2486),
    [anon_sym___const] = ACTIONS(2486),
    [anon_sym___imag] = ACTIONS(2486),
    [anon_sym___kindof] = ACTIONS(2486),
    [anon_sym___nonnull] = ACTIONS(2486),
    [anon_sym___nullable] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2486),
    [anon_sym___real] = ACTIONS(2486),
    [anon_sym___strong] = ACTIONS(2486),
    [anon_sym___unsafe_unretained] = ACTIONS(2486),
    [anon_sym___unused] = ACTIONS(2486),
    [anon_sym___weak] = ACTIONS(2486),
    [anon_sym_alignas] = ACTIONS(2486),
    [anon_sym__Alignas] = ACTIONS(2486),
    [sym_primitive_type] = ACTIONS(2486),
    [anon_sym_enum] = ACTIONS(2486),
    [anon_sym_struct] = ACTIONS(2486),
    [anon_sym_union] = ACTIONS(2486),
    [anon_sym_if] = ACTIONS(2486),
    [anon_sym_else] = ACTIONS(2486),
    [anon_sym_switch] = ACTIONS(2486),
    [anon_sym_case] = ACTIONS(2486),
    [anon_sym_default] = ACTIONS(2486),
    [anon_sym_while] = ACTIONS(2486),
    [anon_sym_do] = ACTIONS(2486),
    [anon_sym_for] = ACTIONS(2486),
    [anon_sym_in] = ACTIONS(2486),
    [anon_sym_return] = ACTIONS(2486),
    [anon_sym_break] = ACTIONS(2486),
    [anon_sym_continue] = ACTIONS(2486),
    [anon_sym_goto] = ACTIONS(2486),
    [anon_sym_DASH_DASH] = ACTIONS(2488),
    [anon_sym_PLUS_PLUS] = ACTIONS(2488),
    [anon_sym_sizeof] = ACTIONS(2486),
    [anon_sym___alignof__] = ACTIONS(2486),
    [anon_sym___alignof] = ACTIONS(2486),
    [anon_sym__alignof] = ACTIONS(2486),
    [anon_sym_alignof] = ACTIONS(2486),
    [anon_sym__Alignof] = ACTIONS(2486),
    [anon_sym_offsetof] = ACTIONS(2486),
    [anon_sym__Generic] = ACTIONS(2486),
    [anon_sym_asm] = ACTIONS(2486),
    [anon_sym___asm__] = ACTIONS(2486),
    [anon_sym___asm] = ACTIONS(2486),
    [sym_number_literal] = ACTIONS(2488),
    [anon_sym_L_SQUOTE] = ACTIONS(2488),
    [anon_sym_u_SQUOTE] = ACTIONS(2488),
    [anon_sym_U_SQUOTE] = ACTIONS(2488),
    [anon_sym_u8_SQUOTE] = ACTIONS(2488),
    [anon_sym_SQUOTE] = ACTIONS(2488),
    [anon_sym_AT] = ACTIONS(2486),
    [anon_sym_DQUOTE] = ACTIONS(2488),
    [anon_sym_L_DQUOTE] = ACTIONS(2488),
    [anon_sym_u_DQUOTE] = ACTIONS(2488),
    [anon_sym_U_DQUOTE] = ACTIONS(2488),
    [anon_sym_u8_DQUOTE] = ACTIONS(2488),
    [sym_true] = ACTIONS(2486),
    [sym_false] = ACTIONS(2486),
    [anon_sym_NULL] = ACTIONS(2486),
    [anon_sym_nullptr] = ACTIONS(2486),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2486),
    [anon_sym___typeof] = ACTIONS(2486),
    [anon_sym_typeof] = ACTIONS(2486),
    [anon_sym_ATimport] = ACTIONS(2488),
    [aux_sym_preproc_undef_token1] = ACTIONS(2486),
    [anon_sym_POUND] = ACTIONS(2486),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2486),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2486),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2486),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_API_AVAILABLE] = ACTIONS(2486),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_API_DEPRECATED] = ACTIONS(2486),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym___deprecated_msg] = ACTIONS(2486),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2486),
    [anon_sym_ATprotocol] = ACTIONS(2488),
    [anon_sym_ATinterface] = ACTIONS(2488),
    [anon_sym_ATimplementation] = ACTIONS(2488),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2488),
    [anon_sym_ATsynthesize] = ACTIONS(2488),
    [anon_sym_ATdynamic] = ACTIONS(2488),
    [anon_sym_ATtry] = ACTIONS(2488),
    [anon_sym___try] = ACTIONS(2486),
    [anon_sym_ATthrow] = ACTIONS(2488),
    [anon_sym_ATselector] = ACTIONS(2488),
    [anon_sym_ATavailable] = ACTIONS(2488),
    [anon_sym___builtin_available] = ACTIONS(2486),
    [anon_sym_va_arg] = ACTIONS(2486),
    [anon_sym_ATencode] = ACTIONS(2488),
    [anon_sym_ATsynchronized] = ACTIONS(2488),
    [anon_sym_BOOL] = ACTIONS(2486),
    [anon_sym_IMP] = ACTIONS(2486),
    [anon_sym_SEL] = ACTIONS(2486),
    [anon_sym_Class] = ACTIONS(2486),
    [anon_sym_id] = ACTIONS(2486),
  },
  [STATE(577)] = {
    [sym_identifier] = ACTIONS(2490),
    [aux_sym_preproc_include_token1] = ACTIONS(2490),
    [aux_sym_preproc_include_token2] = ACTIONS(2490),
    [aux_sym_preproc_def_token1] = ACTIONS(2490),
    [aux_sym_preproc_if_token1] = ACTIONS(2490),
    [aux_sym_preproc_if_token2] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2490),
    [sym_preproc_directive] = ACTIONS(2490),
    [anon_sym_LPAREN2] = ACTIONS(2492),
    [anon_sym_BANG] = ACTIONS(2492),
    [anon_sym_TILDE] = ACTIONS(2492),
    [anon_sym_DASH] = ACTIONS(2490),
    [anon_sym_PLUS] = ACTIONS(2490),
    [anon_sym_STAR] = ACTIONS(2492),
    [anon_sym_CARET] = ACTIONS(2492),
    [anon_sym_AMP] = ACTIONS(2492),
    [anon_sym_SEMI] = ACTIONS(2492),
    [anon_sym___extension__] = ACTIONS(2490),
    [anon_sym_typedef] = ACTIONS(2490),
    [anon_sym_extern] = ACTIONS(2490),
    [anon_sym___attribute__] = ACTIONS(2490),
    [anon_sym___attribute] = ACTIONS(2490),
    [anon_sym_noreturn] = ACTIONS(2490),
    [anon_sym_LBRACK] = ACTIONS(2492),
    [anon_sym___declspec] = ACTIONS(2490),
    [anon_sym___cdecl] = ACTIONS(2490),
    [anon_sym___clrcall] = ACTIONS(2490),
    [anon_sym___stdcall] = ACTIONS(2490),
    [anon_sym___fastcall] = ACTIONS(2490),
    [anon_sym___thiscall] = ACTIONS(2490),
    [anon_sym___vectorcall] = ACTIONS(2490),
    [anon_sym_LBRACE] = ACTIONS(2492),
    [anon_sym_signed] = ACTIONS(2490),
    [anon_sym_unsigned] = ACTIONS(2490),
    [anon_sym_long] = ACTIONS(2490),
    [anon_sym_short] = ACTIONS(2490),
    [anon_sym_static] = ACTIONS(2490),
    [anon_sym_ATautoreleasepool] = ACTIONS(2492),
    [anon_sym_auto] = ACTIONS(2490),
    [anon_sym_register] = ACTIONS(2490),
    [anon_sym_inline] = ACTIONS(2490),
    [anon_sym___inline] = ACTIONS(2490),
    [anon_sym___inline__] = ACTIONS(2490),
    [anon_sym___forceinline] = ACTIONS(2490),
    [anon_sym_thread_local] = ACTIONS(2490),
    [anon_sym___thread] = ACTIONS(2490),
    [anon_sym_CG_EXTERN] = ACTIONS(2490),
    [anon_sym_CG_INLINE] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2490),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2490),
    [anon_sym_IBOutlet] = ACTIONS(2490),
    [anon_sym_IBInspectable] = ACTIONS(2490),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2490),
    [anon_sym_NS_INLINE] = ACTIONS(2490),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2490),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2490),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2490),
    [anon_sym_const] = ACTIONS(2490),
    [anon_sym_constexpr] = ACTIONS(2490),
    [anon_sym_volatile] = ACTIONS(2490),
    [anon_sym_restrict] = ACTIONS(2490),
    [anon_sym___restrict__] = ACTIONS(2490),
    [anon_sym__Atomic] = ACTIONS(2490),
    [anon_sym__Noreturn] = ACTIONS(2490),
    [anon_sym__Nonnull] = ACTIONS(2490),
    [anon_sym_nullable] = ACTIONS(2490),
    [anon_sym__Complex] = ACTIONS(2490),
    [anon_sym__Nullable] = ACTIONS(2490),
    [anon_sym__Nullable_result] = ACTIONS(2490),
    [anon_sym__Null_unspecified] = ACTIONS(2490),
    [anon_sym___autoreleasing] = ACTIONS(2490),
    [anon_sym___block] = ACTIONS(2490),
    [anon_sym___bridge] = ACTIONS(2490),
    [anon_sym___bridge_retained] = ACTIONS(2490),
    [anon_sym___bridge_transfer] = ACTIONS(2490),
    [anon_sym___complex] = ACTIONS(2490),
    [anon_sym___const] = ACTIONS(2490),
    [anon_sym___imag] = ACTIONS(2490),
    [anon_sym___kindof] = ACTIONS(2490),
    [anon_sym___nonnull] = ACTIONS(2490),
    [anon_sym___nullable] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2490),
    [anon_sym___real] = ACTIONS(2490),
    [anon_sym___strong] = ACTIONS(2490),
    [anon_sym___unsafe_unretained] = ACTIONS(2490),
    [anon_sym___unused] = ACTIONS(2490),
    [anon_sym___weak] = ACTIONS(2490),
    [anon_sym_alignas] = ACTIONS(2490),
    [anon_sym__Alignas] = ACTIONS(2490),
    [sym_primitive_type] = ACTIONS(2490),
    [anon_sym_enum] = ACTIONS(2490),
    [anon_sym_struct] = ACTIONS(2490),
    [anon_sym_union] = ACTIONS(2490),
    [anon_sym_if] = ACTIONS(2490),
    [anon_sym_else] = ACTIONS(2490),
    [anon_sym_switch] = ACTIONS(2490),
    [anon_sym_case] = ACTIONS(2490),
    [anon_sym_default] = ACTIONS(2490),
    [anon_sym_while] = ACTIONS(2490),
    [anon_sym_do] = ACTIONS(2490),
    [anon_sym_for] = ACTIONS(2490),
    [anon_sym_in] = ACTIONS(2490),
    [anon_sym_return] = ACTIONS(2490),
    [anon_sym_break] = ACTIONS(2490),
    [anon_sym_continue] = ACTIONS(2490),
    [anon_sym_goto] = ACTIONS(2490),
    [anon_sym_DASH_DASH] = ACTIONS(2492),
    [anon_sym_PLUS_PLUS] = ACTIONS(2492),
    [anon_sym_sizeof] = ACTIONS(2490),
    [anon_sym___alignof__] = ACTIONS(2490),
    [anon_sym___alignof] = ACTIONS(2490),
    [anon_sym__alignof] = ACTIONS(2490),
    [anon_sym_alignof] = ACTIONS(2490),
    [anon_sym__Alignof] = ACTIONS(2490),
    [anon_sym_offsetof] = ACTIONS(2490),
    [anon_sym__Generic] = ACTIONS(2490),
    [anon_sym_asm] = ACTIONS(2490),
    [anon_sym___asm__] = ACTIONS(2490),
    [anon_sym___asm] = ACTIONS(2490),
    [sym_number_literal] = ACTIONS(2492),
    [anon_sym_L_SQUOTE] = ACTIONS(2492),
    [anon_sym_u_SQUOTE] = ACTIONS(2492),
    [anon_sym_U_SQUOTE] = ACTIONS(2492),
    [anon_sym_u8_SQUOTE] = ACTIONS(2492),
    [anon_sym_SQUOTE] = ACTIONS(2492),
    [anon_sym_AT] = ACTIONS(2490),
    [anon_sym_DQUOTE] = ACTIONS(2492),
    [anon_sym_L_DQUOTE] = ACTIONS(2492),
    [anon_sym_u_DQUOTE] = ACTIONS(2492),
    [anon_sym_U_DQUOTE] = ACTIONS(2492),
    [anon_sym_u8_DQUOTE] = ACTIONS(2492),
    [sym_true] = ACTIONS(2490),
    [sym_false] = ACTIONS(2490),
    [anon_sym_NULL] = ACTIONS(2490),
    [anon_sym_nullptr] = ACTIONS(2490),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2490),
    [anon_sym___typeof] = ACTIONS(2490),
    [anon_sym_typeof] = ACTIONS(2490),
    [anon_sym_ATimport] = ACTIONS(2492),
    [aux_sym_preproc_undef_token1] = ACTIONS(2490),
    [anon_sym_POUND] = ACTIONS(2490),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2490),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2490),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2490),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_API_AVAILABLE] = ACTIONS(2490),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_API_DEPRECATED] = ACTIONS(2490),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym___deprecated_msg] = ACTIONS(2490),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2490),
    [anon_sym_ATprotocol] = ACTIONS(2492),
    [anon_sym_ATinterface] = ACTIONS(2492),
    [anon_sym_ATimplementation] = ACTIONS(2492),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2492),
    [anon_sym_ATsynthesize] = ACTIONS(2492),
    [anon_sym_ATdynamic] = ACTIONS(2492),
    [anon_sym_ATtry] = ACTIONS(2492),
    [anon_sym___try] = ACTIONS(2490),
    [anon_sym_ATthrow] = ACTIONS(2492),
    [anon_sym_ATselector] = ACTIONS(2492),
    [anon_sym_ATavailable] = ACTIONS(2492),
    [anon_sym___builtin_available] = ACTIONS(2490),
    [anon_sym_va_arg] = ACTIONS(2490),
    [anon_sym_ATencode] = ACTIONS(2492),
    [anon_sym_ATsynchronized] = ACTIONS(2492),
    [anon_sym_BOOL] = ACTIONS(2490),
    [anon_sym_IMP] = ACTIONS(2490),
    [anon_sym_SEL] = ACTIONS(2490),
    [anon_sym_Class] = ACTIONS(2490),
    [anon_sym_id] = ACTIONS(2490),
  },
  [STATE(578)] = {
    [sym_identifier] = ACTIONS(2494),
    [aux_sym_preproc_include_token1] = ACTIONS(2494),
    [aux_sym_preproc_include_token2] = ACTIONS(2494),
    [aux_sym_preproc_def_token1] = ACTIONS(2494),
    [aux_sym_preproc_if_token1] = ACTIONS(2494),
    [aux_sym_preproc_if_token2] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2494),
    [sym_preproc_directive] = ACTIONS(2494),
    [anon_sym_LPAREN2] = ACTIONS(2496),
    [anon_sym_BANG] = ACTIONS(2496),
    [anon_sym_TILDE] = ACTIONS(2496),
    [anon_sym_DASH] = ACTIONS(2494),
    [anon_sym_PLUS] = ACTIONS(2494),
    [anon_sym_STAR] = ACTIONS(2496),
    [anon_sym_CARET] = ACTIONS(2496),
    [anon_sym_AMP] = ACTIONS(2496),
    [anon_sym_SEMI] = ACTIONS(2496),
    [anon_sym___extension__] = ACTIONS(2494),
    [anon_sym_typedef] = ACTIONS(2494),
    [anon_sym_extern] = ACTIONS(2494),
    [anon_sym___attribute__] = ACTIONS(2494),
    [anon_sym___attribute] = ACTIONS(2494),
    [anon_sym_noreturn] = ACTIONS(2494),
    [anon_sym_LBRACK] = ACTIONS(2496),
    [anon_sym___declspec] = ACTIONS(2494),
    [anon_sym___cdecl] = ACTIONS(2494),
    [anon_sym___clrcall] = ACTIONS(2494),
    [anon_sym___stdcall] = ACTIONS(2494),
    [anon_sym___fastcall] = ACTIONS(2494),
    [anon_sym___thiscall] = ACTIONS(2494),
    [anon_sym___vectorcall] = ACTIONS(2494),
    [anon_sym_LBRACE] = ACTIONS(2496),
    [anon_sym_signed] = ACTIONS(2494),
    [anon_sym_unsigned] = ACTIONS(2494),
    [anon_sym_long] = ACTIONS(2494),
    [anon_sym_short] = ACTIONS(2494),
    [anon_sym_static] = ACTIONS(2494),
    [anon_sym_ATautoreleasepool] = ACTIONS(2496),
    [anon_sym_auto] = ACTIONS(2494),
    [anon_sym_register] = ACTIONS(2494),
    [anon_sym_inline] = ACTIONS(2494),
    [anon_sym___inline] = ACTIONS(2494),
    [anon_sym___inline__] = ACTIONS(2494),
    [anon_sym___forceinline] = ACTIONS(2494),
    [anon_sym_thread_local] = ACTIONS(2494),
    [anon_sym___thread] = ACTIONS(2494),
    [anon_sym_CG_EXTERN] = ACTIONS(2494),
    [anon_sym_CG_INLINE] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2494),
    [anon_sym_IBOutlet] = ACTIONS(2494),
    [anon_sym_IBInspectable] = ACTIONS(2494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2494),
    [anon_sym_NS_INLINE] = ACTIONS(2494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2494),
    [anon_sym_const] = ACTIONS(2494),
    [anon_sym_constexpr] = ACTIONS(2494),
    [anon_sym_volatile] = ACTIONS(2494),
    [anon_sym_restrict] = ACTIONS(2494),
    [anon_sym___restrict__] = ACTIONS(2494),
    [anon_sym__Atomic] = ACTIONS(2494),
    [anon_sym__Noreturn] = ACTIONS(2494),
    [anon_sym__Nonnull] = ACTIONS(2494),
    [anon_sym_nullable] = ACTIONS(2494),
    [anon_sym__Complex] = ACTIONS(2494),
    [anon_sym__Nullable] = ACTIONS(2494),
    [anon_sym__Nullable_result] = ACTIONS(2494),
    [anon_sym__Null_unspecified] = ACTIONS(2494),
    [anon_sym___autoreleasing] = ACTIONS(2494),
    [anon_sym___block] = ACTIONS(2494),
    [anon_sym___bridge] = ACTIONS(2494),
    [anon_sym___bridge_retained] = ACTIONS(2494),
    [anon_sym___bridge_transfer] = ACTIONS(2494),
    [anon_sym___complex] = ACTIONS(2494),
    [anon_sym___const] = ACTIONS(2494),
    [anon_sym___imag] = ACTIONS(2494),
    [anon_sym___kindof] = ACTIONS(2494),
    [anon_sym___nonnull] = ACTIONS(2494),
    [anon_sym___nullable] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2494),
    [anon_sym___real] = ACTIONS(2494),
    [anon_sym___strong] = ACTIONS(2494),
    [anon_sym___unsafe_unretained] = ACTIONS(2494),
    [anon_sym___unused] = ACTIONS(2494),
    [anon_sym___weak] = ACTIONS(2494),
    [anon_sym_alignas] = ACTIONS(2494),
    [anon_sym__Alignas] = ACTIONS(2494),
    [sym_primitive_type] = ACTIONS(2494),
    [anon_sym_enum] = ACTIONS(2494),
    [anon_sym_struct] = ACTIONS(2494),
    [anon_sym_union] = ACTIONS(2494),
    [anon_sym_if] = ACTIONS(2494),
    [anon_sym_else] = ACTIONS(2494),
    [anon_sym_switch] = ACTIONS(2494),
    [anon_sym_case] = ACTIONS(2494),
    [anon_sym_default] = ACTIONS(2494),
    [anon_sym_while] = ACTIONS(2494),
    [anon_sym_do] = ACTIONS(2494),
    [anon_sym_for] = ACTIONS(2494),
    [anon_sym_in] = ACTIONS(2494),
    [anon_sym_return] = ACTIONS(2494),
    [anon_sym_break] = ACTIONS(2494),
    [anon_sym_continue] = ACTIONS(2494),
    [anon_sym_goto] = ACTIONS(2494),
    [anon_sym_DASH_DASH] = ACTIONS(2496),
    [anon_sym_PLUS_PLUS] = ACTIONS(2496),
    [anon_sym_sizeof] = ACTIONS(2494),
    [anon_sym___alignof__] = ACTIONS(2494),
    [anon_sym___alignof] = ACTIONS(2494),
    [anon_sym__alignof] = ACTIONS(2494),
    [anon_sym_alignof] = ACTIONS(2494),
    [anon_sym__Alignof] = ACTIONS(2494),
    [anon_sym_offsetof] = ACTIONS(2494),
    [anon_sym__Generic] = ACTIONS(2494),
    [anon_sym_asm] = ACTIONS(2494),
    [anon_sym___asm__] = ACTIONS(2494),
    [anon_sym___asm] = ACTIONS(2494),
    [sym_number_literal] = ACTIONS(2496),
    [anon_sym_L_SQUOTE] = ACTIONS(2496),
    [anon_sym_u_SQUOTE] = ACTIONS(2496),
    [anon_sym_U_SQUOTE] = ACTIONS(2496),
    [anon_sym_u8_SQUOTE] = ACTIONS(2496),
    [anon_sym_SQUOTE] = ACTIONS(2496),
    [anon_sym_AT] = ACTIONS(2494),
    [anon_sym_DQUOTE] = ACTIONS(2496),
    [anon_sym_L_DQUOTE] = ACTIONS(2496),
    [anon_sym_u_DQUOTE] = ACTIONS(2496),
    [anon_sym_U_DQUOTE] = ACTIONS(2496),
    [anon_sym_u8_DQUOTE] = ACTIONS(2496),
    [sym_true] = ACTIONS(2494),
    [sym_false] = ACTIONS(2494),
    [anon_sym_NULL] = ACTIONS(2494),
    [anon_sym_nullptr] = ACTIONS(2494),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2494),
    [anon_sym___typeof] = ACTIONS(2494),
    [anon_sym_typeof] = ACTIONS(2494),
    [anon_sym_ATimport] = ACTIONS(2496),
    [aux_sym_preproc_undef_token1] = ACTIONS(2494),
    [anon_sym_POUND] = ACTIONS(2494),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_API_AVAILABLE] = ACTIONS(2494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_API_DEPRECATED] = ACTIONS(2494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym___deprecated_msg] = ACTIONS(2494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2494),
    [anon_sym_ATprotocol] = ACTIONS(2496),
    [anon_sym_ATinterface] = ACTIONS(2496),
    [anon_sym_ATimplementation] = ACTIONS(2496),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2496),
    [anon_sym_ATsynthesize] = ACTIONS(2496),
    [anon_sym_ATdynamic] = ACTIONS(2496),
    [anon_sym_ATtry] = ACTIONS(2496),
    [anon_sym___try] = ACTIONS(2494),
    [anon_sym_ATthrow] = ACTIONS(2496),
    [anon_sym_ATselector] = ACTIONS(2496),
    [anon_sym_ATavailable] = ACTIONS(2496),
    [anon_sym___builtin_available] = ACTIONS(2494),
    [anon_sym_va_arg] = ACTIONS(2494),
    [anon_sym_ATencode] = ACTIONS(2496),
    [anon_sym_ATsynchronized] = ACTIONS(2496),
    [anon_sym_BOOL] = ACTIONS(2494),
    [anon_sym_IMP] = ACTIONS(2494),
    [anon_sym_SEL] = ACTIONS(2494),
    [anon_sym_Class] = ACTIONS(2494),
    [anon_sym_id] = ACTIONS(2494),
  },
  [STATE(579)] = {
    [sym_identifier] = ACTIONS(2498),
    [aux_sym_preproc_include_token1] = ACTIONS(2498),
    [aux_sym_preproc_include_token2] = ACTIONS(2498),
    [aux_sym_preproc_def_token1] = ACTIONS(2498),
    [aux_sym_preproc_if_token1] = ACTIONS(2498),
    [aux_sym_preproc_if_token2] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2498),
    [sym_preproc_directive] = ACTIONS(2498),
    [anon_sym_LPAREN2] = ACTIONS(2500),
    [anon_sym_BANG] = ACTIONS(2500),
    [anon_sym_TILDE] = ACTIONS(2500),
    [anon_sym_DASH] = ACTIONS(2498),
    [anon_sym_PLUS] = ACTIONS(2498),
    [anon_sym_STAR] = ACTIONS(2500),
    [anon_sym_CARET] = ACTIONS(2500),
    [anon_sym_AMP] = ACTIONS(2500),
    [anon_sym_SEMI] = ACTIONS(2500),
    [anon_sym___extension__] = ACTIONS(2498),
    [anon_sym_typedef] = ACTIONS(2498),
    [anon_sym_extern] = ACTIONS(2498),
    [anon_sym___attribute__] = ACTIONS(2498),
    [anon_sym___attribute] = ACTIONS(2498),
    [anon_sym_noreturn] = ACTIONS(2498),
    [anon_sym_LBRACK] = ACTIONS(2500),
    [anon_sym___declspec] = ACTIONS(2498),
    [anon_sym___cdecl] = ACTIONS(2498),
    [anon_sym___clrcall] = ACTIONS(2498),
    [anon_sym___stdcall] = ACTIONS(2498),
    [anon_sym___fastcall] = ACTIONS(2498),
    [anon_sym___thiscall] = ACTIONS(2498),
    [anon_sym___vectorcall] = ACTIONS(2498),
    [anon_sym_LBRACE] = ACTIONS(2500),
    [anon_sym_signed] = ACTIONS(2498),
    [anon_sym_unsigned] = ACTIONS(2498),
    [anon_sym_long] = ACTIONS(2498),
    [anon_sym_short] = ACTIONS(2498),
    [anon_sym_static] = ACTIONS(2498),
    [anon_sym_ATautoreleasepool] = ACTIONS(2500),
    [anon_sym_auto] = ACTIONS(2498),
    [anon_sym_register] = ACTIONS(2498),
    [anon_sym_inline] = ACTIONS(2498),
    [anon_sym___inline] = ACTIONS(2498),
    [anon_sym___inline__] = ACTIONS(2498),
    [anon_sym___forceinline] = ACTIONS(2498),
    [anon_sym_thread_local] = ACTIONS(2498),
    [anon_sym___thread] = ACTIONS(2498),
    [anon_sym_CG_EXTERN] = ACTIONS(2498),
    [anon_sym_CG_INLINE] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2498),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2498),
    [anon_sym_IBOutlet] = ACTIONS(2498),
    [anon_sym_IBInspectable] = ACTIONS(2498),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2498),
    [anon_sym_NS_INLINE] = ACTIONS(2498),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2498),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2498),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2498),
    [anon_sym_const] = ACTIONS(2498),
    [anon_sym_constexpr] = ACTIONS(2498),
    [anon_sym_volatile] = ACTIONS(2498),
    [anon_sym_restrict] = ACTIONS(2498),
    [anon_sym___restrict__] = ACTIONS(2498),
    [anon_sym__Atomic] = ACTIONS(2498),
    [anon_sym__Noreturn] = ACTIONS(2498),
    [anon_sym__Nonnull] = ACTIONS(2498),
    [anon_sym_nullable] = ACTIONS(2498),
    [anon_sym__Complex] = ACTIONS(2498),
    [anon_sym__Nullable] = ACTIONS(2498),
    [anon_sym__Nullable_result] = ACTIONS(2498),
    [anon_sym__Null_unspecified] = ACTIONS(2498),
    [anon_sym___autoreleasing] = ACTIONS(2498),
    [anon_sym___block] = ACTIONS(2498),
    [anon_sym___bridge] = ACTIONS(2498),
    [anon_sym___bridge_retained] = ACTIONS(2498),
    [anon_sym___bridge_transfer] = ACTIONS(2498),
    [anon_sym___complex] = ACTIONS(2498),
    [anon_sym___const] = ACTIONS(2498),
    [anon_sym___imag] = ACTIONS(2498),
    [anon_sym___kindof] = ACTIONS(2498),
    [anon_sym___nonnull] = ACTIONS(2498),
    [anon_sym___nullable] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2498),
    [anon_sym___real] = ACTIONS(2498),
    [anon_sym___strong] = ACTIONS(2498),
    [anon_sym___unsafe_unretained] = ACTIONS(2498),
    [anon_sym___unused] = ACTIONS(2498),
    [anon_sym___weak] = ACTIONS(2498),
    [anon_sym_alignas] = ACTIONS(2498),
    [anon_sym__Alignas] = ACTIONS(2498),
    [sym_primitive_type] = ACTIONS(2498),
    [anon_sym_enum] = ACTIONS(2498),
    [anon_sym_struct] = ACTIONS(2498),
    [anon_sym_union] = ACTIONS(2498),
    [anon_sym_if] = ACTIONS(2498),
    [anon_sym_else] = ACTIONS(2498),
    [anon_sym_switch] = ACTIONS(2498),
    [anon_sym_case] = ACTIONS(2498),
    [anon_sym_default] = ACTIONS(2498),
    [anon_sym_while] = ACTIONS(2498),
    [anon_sym_do] = ACTIONS(2498),
    [anon_sym_for] = ACTIONS(2498),
    [anon_sym_in] = ACTIONS(2498),
    [anon_sym_return] = ACTIONS(2498),
    [anon_sym_break] = ACTIONS(2498),
    [anon_sym_continue] = ACTIONS(2498),
    [anon_sym_goto] = ACTIONS(2498),
    [anon_sym_DASH_DASH] = ACTIONS(2500),
    [anon_sym_PLUS_PLUS] = ACTIONS(2500),
    [anon_sym_sizeof] = ACTIONS(2498),
    [anon_sym___alignof__] = ACTIONS(2498),
    [anon_sym___alignof] = ACTIONS(2498),
    [anon_sym__alignof] = ACTIONS(2498),
    [anon_sym_alignof] = ACTIONS(2498),
    [anon_sym__Alignof] = ACTIONS(2498),
    [anon_sym_offsetof] = ACTIONS(2498),
    [anon_sym__Generic] = ACTIONS(2498),
    [anon_sym_asm] = ACTIONS(2498),
    [anon_sym___asm__] = ACTIONS(2498),
    [anon_sym___asm] = ACTIONS(2498),
    [sym_number_literal] = ACTIONS(2500),
    [anon_sym_L_SQUOTE] = ACTIONS(2500),
    [anon_sym_u_SQUOTE] = ACTIONS(2500),
    [anon_sym_U_SQUOTE] = ACTIONS(2500),
    [anon_sym_u8_SQUOTE] = ACTIONS(2500),
    [anon_sym_SQUOTE] = ACTIONS(2500),
    [anon_sym_AT] = ACTIONS(2498),
    [anon_sym_DQUOTE] = ACTIONS(2500),
    [anon_sym_L_DQUOTE] = ACTIONS(2500),
    [anon_sym_u_DQUOTE] = ACTIONS(2500),
    [anon_sym_U_DQUOTE] = ACTIONS(2500),
    [anon_sym_u8_DQUOTE] = ACTIONS(2500),
    [sym_true] = ACTIONS(2498),
    [sym_false] = ACTIONS(2498),
    [anon_sym_NULL] = ACTIONS(2498),
    [anon_sym_nullptr] = ACTIONS(2498),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2498),
    [anon_sym___typeof] = ACTIONS(2498),
    [anon_sym_typeof] = ACTIONS(2498),
    [anon_sym_ATimport] = ACTIONS(2500),
    [aux_sym_preproc_undef_token1] = ACTIONS(2498),
    [anon_sym_POUND] = ACTIONS(2498),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2498),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2498),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2498),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_API_AVAILABLE] = ACTIONS(2498),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_API_DEPRECATED] = ACTIONS(2498),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym___deprecated_msg] = ACTIONS(2498),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2498),
    [anon_sym_ATprotocol] = ACTIONS(2500),
    [anon_sym_ATinterface] = ACTIONS(2500),
    [anon_sym_ATimplementation] = ACTIONS(2500),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2500),
    [anon_sym_ATsynthesize] = ACTIONS(2500),
    [anon_sym_ATdynamic] = ACTIONS(2500),
    [anon_sym_ATtry] = ACTIONS(2500),
    [anon_sym___try] = ACTIONS(2498),
    [anon_sym_ATthrow] = ACTIONS(2500),
    [anon_sym_ATselector] = ACTIONS(2500),
    [anon_sym_ATavailable] = ACTIONS(2500),
    [anon_sym___builtin_available] = ACTIONS(2498),
    [anon_sym_va_arg] = ACTIONS(2498),
    [anon_sym_ATencode] = ACTIONS(2500),
    [anon_sym_ATsynchronized] = ACTIONS(2500),
    [anon_sym_BOOL] = ACTIONS(2498),
    [anon_sym_IMP] = ACTIONS(2498),
    [anon_sym_SEL] = ACTIONS(2498),
    [anon_sym_Class] = ACTIONS(2498),
    [anon_sym_id] = ACTIONS(2498),
  },
  [STATE(580)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token2] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(581)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(582)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(583)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(584)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token2] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATsynthesize] = ACTIONS(2348),
    [anon_sym_ATdynamic] = ACTIONS(2348),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(585)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token2] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(586)] = {
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token2] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATsynthesize] = ACTIONS(2308),
    [anon_sym_ATdynamic] = ACTIONS(2308),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(587)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token2] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(588)] = {
    [sym_identifier] = ACTIONS(2502),
    [aux_sym_preproc_include_token1] = ACTIONS(2502),
    [aux_sym_preproc_include_token2] = ACTIONS(2502),
    [aux_sym_preproc_def_token1] = ACTIONS(2502),
    [aux_sym_preproc_if_token1] = ACTIONS(2502),
    [aux_sym_preproc_if_token2] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2502),
    [sym_preproc_directive] = ACTIONS(2502),
    [anon_sym_LPAREN2] = ACTIONS(2504),
    [anon_sym_BANG] = ACTIONS(2504),
    [anon_sym_TILDE] = ACTIONS(2504),
    [anon_sym_DASH] = ACTIONS(2502),
    [anon_sym_PLUS] = ACTIONS(2502),
    [anon_sym_STAR] = ACTIONS(2504),
    [anon_sym_CARET] = ACTIONS(2504),
    [anon_sym_AMP] = ACTIONS(2504),
    [anon_sym_SEMI] = ACTIONS(2504),
    [anon_sym___extension__] = ACTIONS(2502),
    [anon_sym_typedef] = ACTIONS(2502),
    [anon_sym_extern] = ACTIONS(2502),
    [anon_sym___attribute__] = ACTIONS(2502),
    [anon_sym___attribute] = ACTIONS(2502),
    [anon_sym_noreturn] = ACTIONS(2502),
    [anon_sym_LBRACK] = ACTIONS(2504),
    [anon_sym___declspec] = ACTIONS(2502),
    [anon_sym___cdecl] = ACTIONS(2502),
    [anon_sym___clrcall] = ACTIONS(2502),
    [anon_sym___stdcall] = ACTIONS(2502),
    [anon_sym___fastcall] = ACTIONS(2502),
    [anon_sym___thiscall] = ACTIONS(2502),
    [anon_sym___vectorcall] = ACTIONS(2502),
    [anon_sym_LBRACE] = ACTIONS(2504),
    [anon_sym_signed] = ACTIONS(2502),
    [anon_sym_unsigned] = ACTIONS(2502),
    [anon_sym_long] = ACTIONS(2502),
    [anon_sym_short] = ACTIONS(2502),
    [anon_sym_static] = ACTIONS(2502),
    [anon_sym_ATautoreleasepool] = ACTIONS(2504),
    [anon_sym_auto] = ACTIONS(2502),
    [anon_sym_register] = ACTIONS(2502),
    [anon_sym_inline] = ACTIONS(2502),
    [anon_sym___inline] = ACTIONS(2502),
    [anon_sym___inline__] = ACTIONS(2502),
    [anon_sym___forceinline] = ACTIONS(2502),
    [anon_sym_thread_local] = ACTIONS(2502),
    [anon_sym___thread] = ACTIONS(2502),
    [anon_sym_CG_EXTERN] = ACTIONS(2502),
    [anon_sym_CG_INLINE] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2502),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2502),
    [anon_sym_IBOutlet] = ACTIONS(2502),
    [anon_sym_IBInspectable] = ACTIONS(2502),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2502),
    [anon_sym_NS_INLINE] = ACTIONS(2502),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2502),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2502),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2502),
    [anon_sym_const] = ACTIONS(2502),
    [anon_sym_constexpr] = ACTIONS(2502),
    [anon_sym_volatile] = ACTIONS(2502),
    [anon_sym_restrict] = ACTIONS(2502),
    [anon_sym___restrict__] = ACTIONS(2502),
    [anon_sym__Atomic] = ACTIONS(2502),
    [anon_sym__Noreturn] = ACTIONS(2502),
    [anon_sym__Nonnull] = ACTIONS(2502),
    [anon_sym_nullable] = ACTIONS(2502),
    [anon_sym__Complex] = ACTIONS(2502),
    [anon_sym__Nullable] = ACTIONS(2502),
    [anon_sym__Nullable_result] = ACTIONS(2502),
    [anon_sym__Null_unspecified] = ACTIONS(2502),
    [anon_sym___autoreleasing] = ACTIONS(2502),
    [anon_sym___block] = ACTIONS(2502),
    [anon_sym___bridge] = ACTIONS(2502),
    [anon_sym___bridge_retained] = ACTIONS(2502),
    [anon_sym___bridge_transfer] = ACTIONS(2502),
    [anon_sym___complex] = ACTIONS(2502),
    [anon_sym___const] = ACTIONS(2502),
    [anon_sym___imag] = ACTIONS(2502),
    [anon_sym___kindof] = ACTIONS(2502),
    [anon_sym___nonnull] = ACTIONS(2502),
    [anon_sym___nullable] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2502),
    [anon_sym___real] = ACTIONS(2502),
    [anon_sym___strong] = ACTIONS(2502),
    [anon_sym___unsafe_unretained] = ACTIONS(2502),
    [anon_sym___unused] = ACTIONS(2502),
    [anon_sym___weak] = ACTIONS(2502),
    [anon_sym_alignas] = ACTIONS(2502),
    [anon_sym__Alignas] = ACTIONS(2502),
    [sym_primitive_type] = ACTIONS(2502),
    [anon_sym_enum] = ACTIONS(2502),
    [anon_sym_struct] = ACTIONS(2502),
    [anon_sym_union] = ACTIONS(2502),
    [anon_sym_if] = ACTIONS(2502),
    [anon_sym_else] = ACTIONS(2502),
    [anon_sym_switch] = ACTIONS(2502),
    [anon_sym_case] = ACTIONS(2502),
    [anon_sym_default] = ACTIONS(2502),
    [anon_sym_while] = ACTIONS(2502),
    [anon_sym_do] = ACTIONS(2502),
    [anon_sym_for] = ACTIONS(2502),
    [anon_sym_in] = ACTIONS(2502),
    [anon_sym_return] = ACTIONS(2502),
    [anon_sym_break] = ACTIONS(2502),
    [anon_sym_continue] = ACTIONS(2502),
    [anon_sym_goto] = ACTIONS(2502),
    [anon_sym_DASH_DASH] = ACTIONS(2504),
    [anon_sym_PLUS_PLUS] = ACTIONS(2504),
    [anon_sym_sizeof] = ACTIONS(2502),
    [anon_sym___alignof__] = ACTIONS(2502),
    [anon_sym___alignof] = ACTIONS(2502),
    [anon_sym__alignof] = ACTIONS(2502),
    [anon_sym_alignof] = ACTIONS(2502),
    [anon_sym__Alignof] = ACTIONS(2502),
    [anon_sym_offsetof] = ACTIONS(2502),
    [anon_sym__Generic] = ACTIONS(2502),
    [anon_sym_asm] = ACTIONS(2502),
    [anon_sym___asm__] = ACTIONS(2502),
    [anon_sym___asm] = ACTIONS(2502),
    [sym_number_literal] = ACTIONS(2504),
    [anon_sym_L_SQUOTE] = ACTIONS(2504),
    [anon_sym_u_SQUOTE] = ACTIONS(2504),
    [anon_sym_U_SQUOTE] = ACTIONS(2504),
    [anon_sym_u8_SQUOTE] = ACTIONS(2504),
    [anon_sym_SQUOTE] = ACTIONS(2504),
    [anon_sym_AT] = ACTIONS(2502),
    [anon_sym_DQUOTE] = ACTIONS(2504),
    [anon_sym_L_DQUOTE] = ACTIONS(2504),
    [anon_sym_u_DQUOTE] = ACTIONS(2504),
    [anon_sym_U_DQUOTE] = ACTIONS(2504),
    [anon_sym_u8_DQUOTE] = ACTIONS(2504),
    [sym_true] = ACTIONS(2502),
    [sym_false] = ACTIONS(2502),
    [anon_sym_NULL] = ACTIONS(2502),
    [anon_sym_nullptr] = ACTIONS(2502),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2502),
    [anon_sym___typeof] = ACTIONS(2502),
    [anon_sym_typeof] = ACTIONS(2502),
    [anon_sym_ATimport] = ACTIONS(2504),
    [aux_sym_preproc_undef_token1] = ACTIONS(2502),
    [anon_sym_POUND] = ACTIONS(2502),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2502),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2502),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2502),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_API_AVAILABLE] = ACTIONS(2502),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_API_DEPRECATED] = ACTIONS(2502),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym___deprecated_msg] = ACTIONS(2502),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2502),
    [anon_sym_ATprotocol] = ACTIONS(2504),
    [anon_sym_ATinterface] = ACTIONS(2504),
    [anon_sym_ATimplementation] = ACTIONS(2504),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2504),
    [anon_sym_ATsynthesize] = ACTIONS(2504),
    [anon_sym_ATdynamic] = ACTIONS(2504),
    [anon_sym_ATtry] = ACTIONS(2504),
    [anon_sym___try] = ACTIONS(2502),
    [anon_sym_ATthrow] = ACTIONS(2504),
    [anon_sym_ATselector] = ACTIONS(2504),
    [anon_sym_ATavailable] = ACTIONS(2504),
    [anon_sym___builtin_available] = ACTIONS(2502),
    [anon_sym_va_arg] = ACTIONS(2502),
    [anon_sym_ATencode] = ACTIONS(2504),
    [anon_sym_ATsynchronized] = ACTIONS(2504),
    [anon_sym_BOOL] = ACTIONS(2502),
    [anon_sym_IMP] = ACTIONS(2502),
    [anon_sym_SEL] = ACTIONS(2502),
    [anon_sym_Class] = ACTIONS(2502),
    [anon_sym_id] = ACTIONS(2502),
  },
  [STATE(589)] = {
    [sym_identifier] = ACTIONS(2506),
    [aux_sym_preproc_include_token1] = ACTIONS(2506),
    [aux_sym_preproc_include_token2] = ACTIONS(2506),
    [aux_sym_preproc_def_token1] = ACTIONS(2506),
    [aux_sym_preproc_if_token1] = ACTIONS(2506),
    [aux_sym_preproc_if_token2] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2506),
    [sym_preproc_directive] = ACTIONS(2506),
    [anon_sym_LPAREN2] = ACTIONS(2508),
    [anon_sym_BANG] = ACTIONS(2508),
    [anon_sym_TILDE] = ACTIONS(2508),
    [anon_sym_DASH] = ACTIONS(2506),
    [anon_sym_PLUS] = ACTIONS(2506),
    [anon_sym_STAR] = ACTIONS(2508),
    [anon_sym_CARET] = ACTIONS(2508),
    [anon_sym_AMP] = ACTIONS(2508),
    [anon_sym_SEMI] = ACTIONS(2508),
    [anon_sym___extension__] = ACTIONS(2506),
    [anon_sym_typedef] = ACTIONS(2506),
    [anon_sym_extern] = ACTIONS(2506),
    [anon_sym___attribute__] = ACTIONS(2506),
    [anon_sym___attribute] = ACTIONS(2506),
    [anon_sym_noreturn] = ACTIONS(2506),
    [anon_sym_LBRACK] = ACTIONS(2508),
    [anon_sym___declspec] = ACTIONS(2506),
    [anon_sym___cdecl] = ACTIONS(2506),
    [anon_sym___clrcall] = ACTIONS(2506),
    [anon_sym___stdcall] = ACTIONS(2506),
    [anon_sym___fastcall] = ACTIONS(2506),
    [anon_sym___thiscall] = ACTIONS(2506),
    [anon_sym___vectorcall] = ACTIONS(2506),
    [anon_sym_LBRACE] = ACTIONS(2508),
    [anon_sym_signed] = ACTIONS(2506),
    [anon_sym_unsigned] = ACTIONS(2506),
    [anon_sym_long] = ACTIONS(2506),
    [anon_sym_short] = ACTIONS(2506),
    [anon_sym_static] = ACTIONS(2506),
    [anon_sym_ATautoreleasepool] = ACTIONS(2508),
    [anon_sym_auto] = ACTIONS(2506),
    [anon_sym_register] = ACTIONS(2506),
    [anon_sym_inline] = ACTIONS(2506),
    [anon_sym___inline] = ACTIONS(2506),
    [anon_sym___inline__] = ACTIONS(2506),
    [anon_sym___forceinline] = ACTIONS(2506),
    [anon_sym_thread_local] = ACTIONS(2506),
    [anon_sym___thread] = ACTIONS(2506),
    [anon_sym_CG_EXTERN] = ACTIONS(2506),
    [anon_sym_CG_INLINE] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2506),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2506),
    [anon_sym_IBOutlet] = ACTIONS(2506),
    [anon_sym_IBInspectable] = ACTIONS(2506),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2506),
    [anon_sym_NS_INLINE] = ACTIONS(2506),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2506),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2506),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2506),
    [anon_sym_const] = ACTIONS(2506),
    [anon_sym_constexpr] = ACTIONS(2506),
    [anon_sym_volatile] = ACTIONS(2506),
    [anon_sym_restrict] = ACTIONS(2506),
    [anon_sym___restrict__] = ACTIONS(2506),
    [anon_sym__Atomic] = ACTIONS(2506),
    [anon_sym__Noreturn] = ACTIONS(2506),
    [anon_sym__Nonnull] = ACTIONS(2506),
    [anon_sym_nullable] = ACTIONS(2506),
    [anon_sym__Complex] = ACTIONS(2506),
    [anon_sym__Nullable] = ACTIONS(2506),
    [anon_sym__Nullable_result] = ACTIONS(2506),
    [anon_sym__Null_unspecified] = ACTIONS(2506),
    [anon_sym___autoreleasing] = ACTIONS(2506),
    [anon_sym___block] = ACTIONS(2506),
    [anon_sym___bridge] = ACTIONS(2506),
    [anon_sym___bridge_retained] = ACTIONS(2506),
    [anon_sym___bridge_transfer] = ACTIONS(2506),
    [anon_sym___complex] = ACTIONS(2506),
    [anon_sym___const] = ACTIONS(2506),
    [anon_sym___imag] = ACTIONS(2506),
    [anon_sym___kindof] = ACTIONS(2506),
    [anon_sym___nonnull] = ACTIONS(2506),
    [anon_sym___nullable] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2506),
    [anon_sym___real] = ACTIONS(2506),
    [anon_sym___strong] = ACTIONS(2506),
    [anon_sym___unsafe_unretained] = ACTIONS(2506),
    [anon_sym___unused] = ACTIONS(2506),
    [anon_sym___weak] = ACTIONS(2506),
    [anon_sym_alignas] = ACTIONS(2506),
    [anon_sym__Alignas] = ACTIONS(2506),
    [sym_primitive_type] = ACTIONS(2506),
    [anon_sym_enum] = ACTIONS(2506),
    [anon_sym_struct] = ACTIONS(2506),
    [anon_sym_union] = ACTIONS(2506),
    [anon_sym_if] = ACTIONS(2506),
    [anon_sym_else] = ACTIONS(2506),
    [anon_sym_switch] = ACTIONS(2506),
    [anon_sym_case] = ACTIONS(2506),
    [anon_sym_default] = ACTIONS(2506),
    [anon_sym_while] = ACTIONS(2506),
    [anon_sym_do] = ACTIONS(2506),
    [anon_sym_for] = ACTIONS(2506),
    [anon_sym_in] = ACTIONS(2506),
    [anon_sym_return] = ACTIONS(2506),
    [anon_sym_break] = ACTIONS(2506),
    [anon_sym_continue] = ACTIONS(2506),
    [anon_sym_goto] = ACTIONS(2506),
    [anon_sym_DASH_DASH] = ACTIONS(2508),
    [anon_sym_PLUS_PLUS] = ACTIONS(2508),
    [anon_sym_sizeof] = ACTIONS(2506),
    [anon_sym___alignof__] = ACTIONS(2506),
    [anon_sym___alignof] = ACTIONS(2506),
    [anon_sym__alignof] = ACTIONS(2506),
    [anon_sym_alignof] = ACTIONS(2506),
    [anon_sym__Alignof] = ACTIONS(2506),
    [anon_sym_offsetof] = ACTIONS(2506),
    [anon_sym__Generic] = ACTIONS(2506),
    [anon_sym_asm] = ACTIONS(2506),
    [anon_sym___asm__] = ACTIONS(2506),
    [anon_sym___asm] = ACTIONS(2506),
    [sym_number_literal] = ACTIONS(2508),
    [anon_sym_L_SQUOTE] = ACTIONS(2508),
    [anon_sym_u_SQUOTE] = ACTIONS(2508),
    [anon_sym_U_SQUOTE] = ACTIONS(2508),
    [anon_sym_u8_SQUOTE] = ACTIONS(2508),
    [anon_sym_SQUOTE] = ACTIONS(2508),
    [anon_sym_AT] = ACTIONS(2506),
    [anon_sym_DQUOTE] = ACTIONS(2508),
    [anon_sym_L_DQUOTE] = ACTIONS(2508),
    [anon_sym_u_DQUOTE] = ACTIONS(2508),
    [anon_sym_U_DQUOTE] = ACTIONS(2508),
    [anon_sym_u8_DQUOTE] = ACTIONS(2508),
    [sym_true] = ACTIONS(2506),
    [sym_false] = ACTIONS(2506),
    [anon_sym_NULL] = ACTIONS(2506),
    [anon_sym_nullptr] = ACTIONS(2506),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2506),
    [anon_sym___typeof] = ACTIONS(2506),
    [anon_sym_typeof] = ACTIONS(2506),
    [anon_sym_ATimport] = ACTIONS(2508),
    [aux_sym_preproc_undef_token1] = ACTIONS(2506),
    [anon_sym_POUND] = ACTIONS(2506),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2506),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2506),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2506),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_API_AVAILABLE] = ACTIONS(2506),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_API_DEPRECATED] = ACTIONS(2506),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym___deprecated_msg] = ACTIONS(2506),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2506),
    [anon_sym_ATprotocol] = ACTIONS(2508),
    [anon_sym_ATinterface] = ACTIONS(2508),
    [anon_sym_ATimplementation] = ACTIONS(2508),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2508),
    [anon_sym_ATsynthesize] = ACTIONS(2508),
    [anon_sym_ATdynamic] = ACTIONS(2508),
    [anon_sym_ATtry] = ACTIONS(2508),
    [anon_sym___try] = ACTIONS(2506),
    [anon_sym_ATthrow] = ACTIONS(2508),
    [anon_sym_ATselector] = ACTIONS(2508),
    [anon_sym_ATavailable] = ACTIONS(2508),
    [anon_sym___builtin_available] = ACTIONS(2506),
    [anon_sym_va_arg] = ACTIONS(2506),
    [anon_sym_ATencode] = ACTIONS(2508),
    [anon_sym_ATsynchronized] = ACTIONS(2508),
    [anon_sym_BOOL] = ACTIONS(2506),
    [anon_sym_IMP] = ACTIONS(2506),
    [anon_sym_SEL] = ACTIONS(2506),
    [anon_sym_Class] = ACTIONS(2506),
    [anon_sym_id] = ACTIONS(2506),
  },
  [STATE(590)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token2] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(591)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(592)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(593)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(594)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token2] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(595)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2220),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2222),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(2884),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2222),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2222),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2222),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2222),
    [anon_sym_GT_GT] = ACTIONS(2222),
    [anon_sym___extension__] = ACTIONS(2220),
    [anon_sym_extern] = ACTIONS(2220),
    [anon_sym___attribute__] = ACTIONS(2220),
    [anon_sym___attribute] = ACTIONS(2220),
    [anon_sym_noreturn] = ACTIONS(2220),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym___declspec] = ACTIONS(2220),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_RBRACE] = ACTIONS(2222),
    [anon_sym_static] = ACTIONS(2220),
    [anon_sym_auto] = ACTIONS(2220),
    [anon_sym_register] = ACTIONS(2220),
    [anon_sym_inline] = ACTIONS(2220),
    [anon_sym___inline] = ACTIONS(2220),
    [anon_sym___inline__] = ACTIONS(2220),
    [anon_sym___forceinline] = ACTIONS(2220),
    [anon_sym_thread_local] = ACTIONS(2220),
    [anon_sym___thread] = ACTIONS(2220),
    [anon_sym_CG_EXTERN] = ACTIONS(2220),
    [anon_sym_CG_INLINE] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2220),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2220),
    [anon_sym_IBOutlet] = ACTIONS(2220),
    [anon_sym_IBInspectable] = ACTIONS(2220),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2220),
    [anon_sym_NS_INLINE] = ACTIONS(2220),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2220),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2220),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2220),
    [anon_sym_const] = ACTIONS(2220),
    [anon_sym_constexpr] = ACTIONS(2220),
    [anon_sym_volatile] = ACTIONS(2220),
    [anon_sym_restrict] = ACTIONS(2220),
    [anon_sym___restrict__] = ACTIONS(2220),
    [anon_sym__Atomic] = ACTIONS(2220),
    [anon_sym__Noreturn] = ACTIONS(2220),
    [anon_sym__Nonnull] = ACTIONS(2220),
    [anon_sym_nullable] = ACTIONS(2220),
    [anon_sym__Complex] = ACTIONS(2220),
    [anon_sym__Nullable] = ACTIONS(2220),
    [anon_sym__Nullable_result] = ACTIONS(2220),
    [anon_sym__Null_unspecified] = ACTIONS(2220),
    [anon_sym___autoreleasing] = ACTIONS(2220),
    [anon_sym___block] = ACTIONS(2220),
    [anon_sym___bridge] = ACTIONS(2220),
    [anon_sym___bridge_retained] = ACTIONS(2220),
    [anon_sym___bridge_transfer] = ACTIONS(2220),
    [anon_sym___complex] = ACTIONS(2220),
    [anon_sym___const] = ACTIONS(2220),
    [anon_sym___imag] = ACTIONS(2220),
    [anon_sym___kindof] = ACTIONS(2220),
    [anon_sym___nonnull] = ACTIONS(2220),
    [anon_sym___nullable] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2220),
    [anon_sym___real] = ACTIONS(2220),
    [anon_sym___strong] = ACTIONS(2220),
    [anon_sym___unsafe_unretained] = ACTIONS(2220),
    [anon_sym___unused] = ACTIONS(2220),
    [anon_sym___weak] = ACTIONS(2220),
    [anon_sym_alignas] = ACTIONS(2220),
    [anon_sym__Alignas] = ACTIONS(2220),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2220),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2220),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2220),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_API_AVAILABLE] = ACTIONS(2220),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_API_DEPRECATED] = ACTIONS(2220),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym___deprecated_msg] = ACTIONS(2220),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2220),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(596)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token2] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(597)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token2] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(598)] = {
    [sym_identifier] = ACTIONS(2558),
    [aux_sym_preproc_include_token1] = ACTIONS(2558),
    [aux_sym_preproc_include_token2] = ACTIONS(2558),
    [aux_sym_preproc_def_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token2] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2558),
    [sym_preproc_directive] = ACTIONS(2558),
    [anon_sym_LPAREN2] = ACTIONS(2560),
    [anon_sym_BANG] = ACTIONS(2560),
    [anon_sym_TILDE] = ACTIONS(2560),
    [anon_sym_DASH] = ACTIONS(2558),
    [anon_sym_PLUS] = ACTIONS(2558),
    [anon_sym_STAR] = ACTIONS(2560),
    [anon_sym_CARET] = ACTIONS(2560),
    [anon_sym_AMP] = ACTIONS(2560),
    [anon_sym_SEMI] = ACTIONS(2560),
    [anon_sym___extension__] = ACTIONS(2558),
    [anon_sym_typedef] = ACTIONS(2558),
    [anon_sym_extern] = ACTIONS(2558),
    [anon_sym___attribute__] = ACTIONS(2558),
    [anon_sym___attribute] = ACTIONS(2558),
    [anon_sym_noreturn] = ACTIONS(2558),
    [anon_sym_LBRACK] = ACTIONS(2560),
    [anon_sym___declspec] = ACTIONS(2558),
    [anon_sym___cdecl] = ACTIONS(2558),
    [anon_sym___clrcall] = ACTIONS(2558),
    [anon_sym___stdcall] = ACTIONS(2558),
    [anon_sym___fastcall] = ACTIONS(2558),
    [anon_sym___thiscall] = ACTIONS(2558),
    [anon_sym___vectorcall] = ACTIONS(2558),
    [anon_sym_LBRACE] = ACTIONS(2560),
    [anon_sym_signed] = ACTIONS(2558),
    [anon_sym_unsigned] = ACTIONS(2558),
    [anon_sym_long] = ACTIONS(2558),
    [anon_sym_short] = ACTIONS(2558),
    [anon_sym_static] = ACTIONS(2558),
    [anon_sym_ATautoreleasepool] = ACTIONS(2560),
    [anon_sym_auto] = ACTIONS(2558),
    [anon_sym_register] = ACTIONS(2558),
    [anon_sym_inline] = ACTIONS(2558),
    [anon_sym___inline] = ACTIONS(2558),
    [anon_sym___inline__] = ACTIONS(2558),
    [anon_sym___forceinline] = ACTIONS(2558),
    [anon_sym_thread_local] = ACTIONS(2558),
    [anon_sym___thread] = ACTIONS(2558),
    [anon_sym_CG_EXTERN] = ACTIONS(2558),
    [anon_sym_CG_INLINE] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2558),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2558),
    [anon_sym_IBOutlet] = ACTIONS(2558),
    [anon_sym_IBInspectable] = ACTIONS(2558),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2558),
    [anon_sym_NS_INLINE] = ACTIONS(2558),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2558),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2558),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2558),
    [anon_sym_const] = ACTIONS(2558),
    [anon_sym_constexpr] = ACTIONS(2558),
    [anon_sym_volatile] = ACTIONS(2558),
    [anon_sym_restrict] = ACTIONS(2558),
    [anon_sym___restrict__] = ACTIONS(2558),
    [anon_sym__Atomic] = ACTIONS(2558),
    [anon_sym__Noreturn] = ACTIONS(2558),
    [anon_sym__Nonnull] = ACTIONS(2558),
    [anon_sym_nullable] = ACTIONS(2558),
    [anon_sym__Complex] = ACTIONS(2558),
    [anon_sym__Nullable] = ACTIONS(2558),
    [anon_sym__Nullable_result] = ACTIONS(2558),
    [anon_sym__Null_unspecified] = ACTIONS(2558),
    [anon_sym___autoreleasing] = ACTIONS(2558),
    [anon_sym___block] = ACTIONS(2558),
    [anon_sym___bridge] = ACTIONS(2558),
    [anon_sym___bridge_retained] = ACTIONS(2558),
    [anon_sym___bridge_transfer] = ACTIONS(2558),
    [anon_sym___complex] = ACTIONS(2558),
    [anon_sym___const] = ACTIONS(2558),
    [anon_sym___imag] = ACTIONS(2558),
    [anon_sym___kindof] = ACTIONS(2558),
    [anon_sym___nonnull] = ACTIONS(2558),
    [anon_sym___nullable] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2558),
    [anon_sym___real] = ACTIONS(2558),
    [anon_sym___strong] = ACTIONS(2558),
    [anon_sym___unsafe_unretained] = ACTIONS(2558),
    [anon_sym___unused] = ACTIONS(2558),
    [anon_sym___weak] = ACTIONS(2558),
    [anon_sym_alignas] = ACTIONS(2558),
    [anon_sym__Alignas] = ACTIONS(2558),
    [sym_primitive_type] = ACTIONS(2558),
    [anon_sym_enum] = ACTIONS(2558),
    [anon_sym_struct] = ACTIONS(2558),
    [anon_sym_union] = ACTIONS(2558),
    [anon_sym_if] = ACTIONS(2558),
    [anon_sym_else] = ACTIONS(2558),
    [anon_sym_switch] = ACTIONS(2558),
    [anon_sym_case] = ACTIONS(2558),
    [anon_sym_default] = ACTIONS(2558),
    [anon_sym_while] = ACTIONS(2558),
    [anon_sym_do] = ACTIONS(2558),
    [anon_sym_for] = ACTIONS(2558),
    [anon_sym_in] = ACTIONS(2558),
    [anon_sym_return] = ACTIONS(2558),
    [anon_sym_break] = ACTIONS(2558),
    [anon_sym_continue] = ACTIONS(2558),
    [anon_sym_goto] = ACTIONS(2558),
    [anon_sym_DASH_DASH] = ACTIONS(2560),
    [anon_sym_PLUS_PLUS] = ACTIONS(2560),
    [anon_sym_sizeof] = ACTIONS(2558),
    [anon_sym___alignof__] = ACTIONS(2558),
    [anon_sym___alignof] = ACTIONS(2558),
    [anon_sym__alignof] = ACTIONS(2558),
    [anon_sym_alignof] = ACTIONS(2558),
    [anon_sym__Alignof] = ACTIONS(2558),
    [anon_sym_offsetof] = ACTIONS(2558),
    [anon_sym__Generic] = ACTIONS(2558),
    [anon_sym_asm] = ACTIONS(2558),
    [anon_sym___asm__] = ACTIONS(2558),
    [anon_sym___asm] = ACTIONS(2558),
    [sym_number_literal] = ACTIONS(2560),
    [anon_sym_L_SQUOTE] = ACTIONS(2560),
    [anon_sym_u_SQUOTE] = ACTIONS(2560),
    [anon_sym_U_SQUOTE] = ACTIONS(2560),
    [anon_sym_u8_SQUOTE] = ACTIONS(2560),
    [anon_sym_SQUOTE] = ACTIONS(2560),
    [anon_sym_AT] = ACTIONS(2558),
    [anon_sym_DQUOTE] = ACTIONS(2560),
    [anon_sym_L_DQUOTE] = ACTIONS(2560),
    [anon_sym_u_DQUOTE] = ACTIONS(2560),
    [anon_sym_U_DQUOTE] = ACTIONS(2560),
    [anon_sym_u8_DQUOTE] = ACTIONS(2560),
    [sym_true] = ACTIONS(2558),
    [sym_false] = ACTIONS(2558),
    [anon_sym_NULL] = ACTIONS(2558),
    [anon_sym_nullptr] = ACTIONS(2558),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2558),
    [anon_sym___typeof] = ACTIONS(2558),
    [anon_sym_typeof] = ACTIONS(2558),
    [anon_sym_ATimport] = ACTIONS(2560),
    [aux_sym_preproc_undef_token1] = ACTIONS(2558),
    [anon_sym_POUND] = ACTIONS(2558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2558),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2558),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2558),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_API_AVAILABLE] = ACTIONS(2558),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_API_DEPRECATED] = ACTIONS(2558),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym___deprecated_msg] = ACTIONS(2558),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2558),
    [anon_sym_ATprotocol] = ACTIONS(2560),
    [anon_sym_ATinterface] = ACTIONS(2560),
    [anon_sym_ATimplementation] = ACTIONS(2560),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2560),
    [anon_sym_ATsynthesize] = ACTIONS(2560),
    [anon_sym_ATdynamic] = ACTIONS(2560),
    [anon_sym_ATtry] = ACTIONS(2560),
    [anon_sym___try] = ACTIONS(2558),
    [anon_sym_ATthrow] = ACTIONS(2560),
    [anon_sym_ATselector] = ACTIONS(2560),
    [anon_sym_ATavailable] = ACTIONS(2560),
    [anon_sym___builtin_available] = ACTIONS(2558),
    [anon_sym_va_arg] = ACTIONS(2558),
    [anon_sym_ATencode] = ACTIONS(2560),
    [anon_sym_ATsynchronized] = ACTIONS(2560),
    [anon_sym_BOOL] = ACTIONS(2558),
    [anon_sym_IMP] = ACTIONS(2558),
    [anon_sym_SEL] = ACTIONS(2558),
    [anon_sym_Class] = ACTIONS(2558),
    [anon_sym_id] = ACTIONS(2558),
  },
  [STATE(599)] = {
    [sym_identifier] = ACTIONS(2470),
    [aux_sym_preproc_include_token1] = ACTIONS(2470),
    [aux_sym_preproc_include_token2] = ACTIONS(2470),
    [aux_sym_preproc_def_token1] = ACTIONS(2470),
    [aux_sym_preproc_if_token1] = ACTIONS(2470),
    [aux_sym_preproc_if_token2] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2470),
    [sym_preproc_directive] = ACTIONS(2470),
    [anon_sym_LPAREN2] = ACTIONS(2472),
    [anon_sym_BANG] = ACTIONS(2472),
    [anon_sym_TILDE] = ACTIONS(2472),
    [anon_sym_DASH] = ACTIONS(2470),
    [anon_sym_PLUS] = ACTIONS(2470),
    [anon_sym_STAR] = ACTIONS(2472),
    [anon_sym_CARET] = ACTIONS(2472),
    [anon_sym_AMP] = ACTIONS(2472),
    [anon_sym_SEMI] = ACTIONS(2472),
    [anon_sym___extension__] = ACTIONS(2470),
    [anon_sym_typedef] = ACTIONS(2470),
    [anon_sym_extern] = ACTIONS(2470),
    [anon_sym___attribute__] = ACTIONS(2470),
    [anon_sym___attribute] = ACTIONS(2470),
    [anon_sym_noreturn] = ACTIONS(2470),
    [anon_sym_LBRACK] = ACTIONS(2472),
    [anon_sym___declspec] = ACTIONS(2470),
    [anon_sym___cdecl] = ACTIONS(2470),
    [anon_sym___clrcall] = ACTIONS(2470),
    [anon_sym___stdcall] = ACTIONS(2470),
    [anon_sym___fastcall] = ACTIONS(2470),
    [anon_sym___thiscall] = ACTIONS(2470),
    [anon_sym___vectorcall] = ACTIONS(2470),
    [anon_sym_LBRACE] = ACTIONS(2472),
    [anon_sym_signed] = ACTIONS(2470),
    [anon_sym_unsigned] = ACTIONS(2470),
    [anon_sym_long] = ACTIONS(2470),
    [anon_sym_short] = ACTIONS(2470),
    [anon_sym_static] = ACTIONS(2470),
    [anon_sym_ATautoreleasepool] = ACTIONS(2472),
    [anon_sym_auto] = ACTIONS(2470),
    [anon_sym_register] = ACTIONS(2470),
    [anon_sym_inline] = ACTIONS(2470),
    [anon_sym___inline] = ACTIONS(2470),
    [anon_sym___inline__] = ACTIONS(2470),
    [anon_sym___forceinline] = ACTIONS(2470),
    [anon_sym_thread_local] = ACTIONS(2470),
    [anon_sym___thread] = ACTIONS(2470),
    [anon_sym_CG_EXTERN] = ACTIONS(2470),
    [anon_sym_CG_INLINE] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2470),
    [anon_sym_IBOutlet] = ACTIONS(2470),
    [anon_sym_IBInspectable] = ACTIONS(2470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2470),
    [anon_sym_NS_INLINE] = ACTIONS(2470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2470),
    [anon_sym_const] = ACTIONS(2470),
    [anon_sym_constexpr] = ACTIONS(2470),
    [anon_sym_volatile] = ACTIONS(2470),
    [anon_sym_restrict] = ACTIONS(2470),
    [anon_sym___restrict__] = ACTIONS(2470),
    [anon_sym__Atomic] = ACTIONS(2470),
    [anon_sym__Noreturn] = ACTIONS(2470),
    [anon_sym__Nonnull] = ACTIONS(2470),
    [anon_sym_nullable] = ACTIONS(2470),
    [anon_sym__Complex] = ACTIONS(2470),
    [anon_sym__Nullable] = ACTIONS(2470),
    [anon_sym__Nullable_result] = ACTIONS(2470),
    [anon_sym__Null_unspecified] = ACTIONS(2470),
    [anon_sym___autoreleasing] = ACTIONS(2470),
    [anon_sym___block] = ACTIONS(2470),
    [anon_sym___bridge] = ACTIONS(2470),
    [anon_sym___bridge_retained] = ACTIONS(2470),
    [anon_sym___bridge_transfer] = ACTIONS(2470),
    [anon_sym___complex] = ACTIONS(2470),
    [anon_sym___const] = ACTIONS(2470),
    [anon_sym___imag] = ACTIONS(2470),
    [anon_sym___kindof] = ACTIONS(2470),
    [anon_sym___nonnull] = ACTIONS(2470),
    [anon_sym___nullable] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2470),
    [anon_sym___real] = ACTIONS(2470),
    [anon_sym___strong] = ACTIONS(2470),
    [anon_sym___unsafe_unretained] = ACTIONS(2470),
    [anon_sym___unused] = ACTIONS(2470),
    [anon_sym___weak] = ACTIONS(2470),
    [anon_sym_alignas] = ACTIONS(2470),
    [anon_sym__Alignas] = ACTIONS(2470),
    [sym_primitive_type] = ACTIONS(2470),
    [anon_sym_enum] = ACTIONS(2470),
    [anon_sym_struct] = ACTIONS(2470),
    [anon_sym_union] = ACTIONS(2470),
    [anon_sym_if] = ACTIONS(2470),
    [anon_sym_else] = ACTIONS(2470),
    [anon_sym_switch] = ACTIONS(2470),
    [anon_sym_case] = ACTIONS(2470),
    [anon_sym_default] = ACTIONS(2470),
    [anon_sym_while] = ACTIONS(2470),
    [anon_sym_do] = ACTIONS(2470),
    [anon_sym_for] = ACTIONS(2470),
    [anon_sym_in] = ACTIONS(2470),
    [anon_sym_return] = ACTIONS(2470),
    [anon_sym_break] = ACTIONS(2470),
    [anon_sym_continue] = ACTIONS(2470),
    [anon_sym_goto] = ACTIONS(2470),
    [anon_sym_DASH_DASH] = ACTIONS(2472),
    [anon_sym_PLUS_PLUS] = ACTIONS(2472),
    [anon_sym_sizeof] = ACTIONS(2470),
    [anon_sym___alignof__] = ACTIONS(2470),
    [anon_sym___alignof] = ACTIONS(2470),
    [anon_sym__alignof] = ACTIONS(2470),
    [anon_sym_alignof] = ACTIONS(2470),
    [anon_sym__Alignof] = ACTIONS(2470),
    [anon_sym_offsetof] = ACTIONS(2470),
    [anon_sym__Generic] = ACTIONS(2470),
    [anon_sym_asm] = ACTIONS(2470),
    [anon_sym___asm__] = ACTIONS(2470),
    [anon_sym___asm] = ACTIONS(2470),
    [sym_number_literal] = ACTIONS(2472),
    [anon_sym_L_SQUOTE] = ACTIONS(2472),
    [anon_sym_u_SQUOTE] = ACTIONS(2472),
    [anon_sym_U_SQUOTE] = ACTIONS(2472),
    [anon_sym_u8_SQUOTE] = ACTIONS(2472),
    [anon_sym_SQUOTE] = ACTIONS(2472),
    [anon_sym_AT] = ACTIONS(2470),
    [anon_sym_DQUOTE] = ACTIONS(2472),
    [anon_sym_L_DQUOTE] = ACTIONS(2472),
    [anon_sym_u_DQUOTE] = ACTIONS(2472),
    [anon_sym_U_DQUOTE] = ACTIONS(2472),
    [anon_sym_u8_DQUOTE] = ACTIONS(2472),
    [sym_true] = ACTIONS(2470),
    [sym_false] = ACTIONS(2470),
    [anon_sym_NULL] = ACTIONS(2470),
    [anon_sym_nullptr] = ACTIONS(2470),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2470),
    [anon_sym___typeof] = ACTIONS(2470),
    [anon_sym_typeof] = ACTIONS(2470),
    [anon_sym_ATimport] = ACTIONS(2472),
    [aux_sym_preproc_undef_token1] = ACTIONS(2470),
    [anon_sym_POUND] = ACTIONS(2470),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_API_AVAILABLE] = ACTIONS(2470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_API_DEPRECATED] = ACTIONS(2470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym___deprecated_msg] = ACTIONS(2470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2470),
    [anon_sym_ATprotocol] = ACTIONS(2472),
    [anon_sym_ATinterface] = ACTIONS(2472),
    [anon_sym_ATimplementation] = ACTIONS(2472),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2472),
    [anon_sym_ATsynthesize] = ACTIONS(2472),
    [anon_sym_ATdynamic] = ACTIONS(2472),
    [anon_sym_ATtry] = ACTIONS(2472),
    [anon_sym___try] = ACTIONS(2470),
    [anon_sym_ATthrow] = ACTIONS(2472),
    [anon_sym_ATselector] = ACTIONS(2472),
    [anon_sym_ATavailable] = ACTIONS(2472),
    [anon_sym___builtin_available] = ACTIONS(2470),
    [anon_sym_va_arg] = ACTIONS(2470),
    [anon_sym_ATencode] = ACTIONS(2472),
    [anon_sym_ATsynchronized] = ACTIONS(2472),
    [anon_sym_BOOL] = ACTIONS(2470),
    [anon_sym_IMP] = ACTIONS(2470),
    [anon_sym_SEL] = ACTIONS(2470),
    [anon_sym_Class] = ACTIONS(2470),
    [anon_sym_id] = ACTIONS(2470),
  },
  [STATE(600)] = {
    [sym_identifier] = ACTIONS(2650),
    [aux_sym_preproc_include_token1] = ACTIONS(2650),
    [aux_sym_preproc_include_token2] = ACTIONS(2650),
    [aux_sym_preproc_def_token1] = ACTIONS(2650),
    [aux_sym_preproc_if_token1] = ACTIONS(2650),
    [aux_sym_preproc_if_token2] = ACTIONS(2650),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2650),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2650),
    [sym_preproc_directive] = ACTIONS(2650),
    [anon_sym_LPAREN2] = ACTIONS(2652),
    [anon_sym_BANG] = ACTIONS(2652),
    [anon_sym_TILDE] = ACTIONS(2652),
    [anon_sym_DASH] = ACTIONS(2650),
    [anon_sym_PLUS] = ACTIONS(2650),
    [anon_sym_STAR] = ACTIONS(2652),
    [anon_sym_CARET] = ACTIONS(2652),
    [anon_sym_AMP] = ACTIONS(2652),
    [anon_sym_SEMI] = ACTIONS(2652),
    [anon_sym___extension__] = ACTIONS(2650),
    [anon_sym_typedef] = ACTIONS(2650),
    [anon_sym_extern] = ACTIONS(2650),
    [anon_sym___attribute__] = ACTIONS(2650),
    [anon_sym___attribute] = ACTIONS(2650),
    [anon_sym_noreturn] = ACTIONS(2650),
    [anon_sym_LBRACK] = ACTIONS(2652),
    [anon_sym___declspec] = ACTIONS(2650),
    [anon_sym___cdecl] = ACTIONS(2650),
    [anon_sym___clrcall] = ACTIONS(2650),
    [anon_sym___stdcall] = ACTIONS(2650),
    [anon_sym___fastcall] = ACTIONS(2650),
    [anon_sym___thiscall] = ACTIONS(2650),
    [anon_sym___vectorcall] = ACTIONS(2650),
    [anon_sym_LBRACE] = ACTIONS(2652),
    [anon_sym_signed] = ACTIONS(2650),
    [anon_sym_unsigned] = ACTIONS(2650),
    [anon_sym_long] = ACTIONS(2650),
    [anon_sym_short] = ACTIONS(2650),
    [anon_sym_static] = ACTIONS(2650),
    [anon_sym_ATautoreleasepool] = ACTIONS(2652),
    [anon_sym_auto] = ACTIONS(2650),
    [anon_sym_register] = ACTIONS(2650),
    [anon_sym_inline] = ACTIONS(2650),
    [anon_sym___inline] = ACTIONS(2650),
    [anon_sym___inline__] = ACTIONS(2650),
    [anon_sym___forceinline] = ACTIONS(2650),
    [anon_sym_thread_local] = ACTIONS(2650),
    [anon_sym___thread] = ACTIONS(2650),
    [anon_sym_CG_EXTERN] = ACTIONS(2650),
    [anon_sym_CG_INLINE] = ACTIONS(2650),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2650),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2650),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2650),
    [anon_sym_IBOutlet] = ACTIONS(2650),
    [anon_sym_IBInspectable] = ACTIONS(2650),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2650),
    [anon_sym_NS_INLINE] = ACTIONS(2650),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2650),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2650),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2650),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2650),
    [anon_sym_const] = ACTIONS(2650),
    [anon_sym_constexpr] = ACTIONS(2650),
    [anon_sym_volatile] = ACTIONS(2650),
    [anon_sym_restrict] = ACTIONS(2650),
    [anon_sym___restrict__] = ACTIONS(2650),
    [anon_sym__Atomic] = ACTIONS(2650),
    [anon_sym__Noreturn] = ACTIONS(2650),
    [anon_sym__Nonnull] = ACTIONS(2650),
    [anon_sym_nullable] = ACTIONS(2650),
    [anon_sym__Complex] = ACTIONS(2650),
    [anon_sym__Nullable] = ACTIONS(2650),
    [anon_sym__Nullable_result] = ACTIONS(2650),
    [anon_sym__Null_unspecified] = ACTIONS(2650),
    [anon_sym___autoreleasing] = ACTIONS(2650),
    [anon_sym___block] = ACTIONS(2650),
    [anon_sym___bridge] = ACTIONS(2650),
    [anon_sym___bridge_retained] = ACTIONS(2650),
    [anon_sym___bridge_transfer] = ACTIONS(2650),
    [anon_sym___complex] = ACTIONS(2650),
    [anon_sym___const] = ACTIONS(2650),
    [anon_sym___imag] = ACTIONS(2650),
    [anon_sym___kindof] = ACTIONS(2650),
    [anon_sym___nonnull] = ACTIONS(2650),
    [anon_sym___nullable] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2650),
    [anon_sym___real] = ACTIONS(2650),
    [anon_sym___strong] = ACTIONS(2650),
    [anon_sym___unsafe_unretained] = ACTIONS(2650),
    [anon_sym___unused] = ACTIONS(2650),
    [anon_sym___weak] = ACTIONS(2650),
    [anon_sym_alignas] = ACTIONS(2650),
    [anon_sym__Alignas] = ACTIONS(2650),
    [sym_primitive_type] = ACTIONS(2650),
    [anon_sym_enum] = ACTIONS(2650),
    [anon_sym_struct] = ACTIONS(2650),
    [anon_sym_union] = ACTIONS(2650),
    [anon_sym_if] = ACTIONS(2650),
    [anon_sym_switch] = ACTIONS(2650),
    [anon_sym_case] = ACTIONS(2650),
    [anon_sym_default] = ACTIONS(2650),
    [anon_sym_while] = ACTIONS(2650),
    [anon_sym_do] = ACTIONS(2650),
    [anon_sym_for] = ACTIONS(2650),
    [anon_sym_in] = ACTIONS(2650),
    [anon_sym_return] = ACTIONS(2650),
    [anon_sym_break] = ACTIONS(2650),
    [anon_sym_continue] = ACTIONS(2650),
    [anon_sym_goto] = ACTIONS(2650),
    [anon_sym_DASH_DASH] = ACTIONS(2652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2652),
    [anon_sym_sizeof] = ACTIONS(2650),
    [anon_sym___alignof__] = ACTIONS(2650),
    [anon_sym___alignof] = ACTIONS(2650),
    [anon_sym__alignof] = ACTIONS(2650),
    [anon_sym_alignof] = ACTIONS(2650),
    [anon_sym__Alignof] = ACTIONS(2650),
    [anon_sym_offsetof] = ACTIONS(2650),
    [anon_sym__Generic] = ACTIONS(2650),
    [anon_sym_asm] = ACTIONS(2650),
    [anon_sym___asm__] = ACTIONS(2650),
    [anon_sym___asm] = ACTIONS(2650),
    [sym_number_literal] = ACTIONS(2652),
    [anon_sym_L_SQUOTE] = ACTIONS(2652),
    [anon_sym_u_SQUOTE] = ACTIONS(2652),
    [anon_sym_U_SQUOTE] = ACTIONS(2652),
    [anon_sym_u8_SQUOTE] = ACTIONS(2652),
    [anon_sym_SQUOTE] = ACTIONS(2652),
    [anon_sym_AT] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_L_DQUOTE] = ACTIONS(2652),
    [anon_sym_u_DQUOTE] = ACTIONS(2652),
    [anon_sym_U_DQUOTE] = ACTIONS(2652),
    [anon_sym_u8_DQUOTE] = ACTIONS(2652),
    [sym_true] = ACTIONS(2650),
    [sym_false] = ACTIONS(2650),
    [anon_sym_NULL] = ACTIONS(2650),
    [anon_sym_nullptr] = ACTIONS(2650),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2650),
    [anon_sym___typeof] = ACTIONS(2650),
    [anon_sym_typeof] = ACTIONS(2650),
    [anon_sym_ATimport] = ACTIONS(2652),
    [aux_sym_preproc_undef_token1] = ACTIONS(2650),
    [anon_sym_POUND] = ACTIONS(2650),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2650),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2650),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2650),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2650),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2650),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2650),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2650),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_API_AVAILABLE] = ACTIONS(2650),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_API_DEPRECATED] = ACTIONS(2650),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2650),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym___deprecated_msg] = ACTIONS(2650),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2650),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2650),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2650),
    [anon_sym_ATprotocol] = ACTIONS(2652),
    [anon_sym_ATinterface] = ACTIONS(2652),
    [anon_sym_ATimplementation] = ACTIONS(2652),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2652),
    [anon_sym_ATsynthesize] = ACTIONS(2652),
    [anon_sym_ATdynamic] = ACTIONS(2652),
    [anon_sym_ATtry] = ACTIONS(2652),
    [anon_sym___try] = ACTIONS(2650),
    [anon_sym_ATthrow] = ACTIONS(2652),
    [anon_sym_ATselector] = ACTIONS(2652),
    [anon_sym_ATavailable] = ACTIONS(2652),
    [anon_sym___builtin_available] = ACTIONS(2650),
    [anon_sym_va_arg] = ACTIONS(2650),
    [anon_sym_ATencode] = ACTIONS(2652),
    [anon_sym_ATsynchronized] = ACTIONS(2652),
    [anon_sym_BOOL] = ACTIONS(2650),
    [anon_sym_IMP] = ACTIONS(2650),
    [anon_sym_SEL] = ACTIONS(2650),
    [anon_sym_Class] = ACTIONS(2650),
    [anon_sym_id] = ACTIONS(2650),
  },
  [STATE(601)] = {
    [sym_identifier] = ACTIONS(2744),
    [aux_sym_preproc_include_token1] = ACTIONS(2744),
    [aux_sym_preproc_include_token2] = ACTIONS(2744),
    [aux_sym_preproc_def_token1] = ACTIONS(2744),
    [aux_sym_preproc_if_token1] = ACTIONS(2744),
    [aux_sym_preproc_if_token2] = ACTIONS(2744),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2744),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2744),
    [sym_preproc_directive] = ACTIONS(2744),
    [anon_sym_LPAREN2] = ACTIONS(2746),
    [anon_sym_BANG] = ACTIONS(2746),
    [anon_sym_TILDE] = ACTIONS(2746),
    [anon_sym_DASH] = ACTIONS(2744),
    [anon_sym_PLUS] = ACTIONS(2744),
    [anon_sym_STAR] = ACTIONS(2746),
    [anon_sym_CARET] = ACTIONS(2746),
    [anon_sym_AMP] = ACTIONS(2746),
    [anon_sym_SEMI] = ACTIONS(2746),
    [anon_sym___extension__] = ACTIONS(2744),
    [anon_sym_typedef] = ACTIONS(2744),
    [anon_sym_extern] = ACTIONS(2744),
    [anon_sym___attribute__] = ACTIONS(2744),
    [anon_sym___attribute] = ACTIONS(2744),
    [anon_sym_noreturn] = ACTIONS(2744),
    [anon_sym_LBRACK] = ACTIONS(2746),
    [anon_sym___declspec] = ACTIONS(2744),
    [anon_sym___cdecl] = ACTIONS(2744),
    [anon_sym___clrcall] = ACTIONS(2744),
    [anon_sym___stdcall] = ACTIONS(2744),
    [anon_sym___fastcall] = ACTIONS(2744),
    [anon_sym___thiscall] = ACTIONS(2744),
    [anon_sym___vectorcall] = ACTIONS(2744),
    [anon_sym_LBRACE] = ACTIONS(2746),
    [anon_sym_signed] = ACTIONS(2744),
    [anon_sym_unsigned] = ACTIONS(2744),
    [anon_sym_long] = ACTIONS(2744),
    [anon_sym_short] = ACTIONS(2744),
    [anon_sym_static] = ACTIONS(2744),
    [anon_sym_ATautoreleasepool] = ACTIONS(2746),
    [anon_sym_auto] = ACTIONS(2744),
    [anon_sym_register] = ACTIONS(2744),
    [anon_sym_inline] = ACTIONS(2744),
    [anon_sym___inline] = ACTIONS(2744),
    [anon_sym___inline__] = ACTIONS(2744),
    [anon_sym___forceinline] = ACTIONS(2744),
    [anon_sym_thread_local] = ACTIONS(2744),
    [anon_sym___thread] = ACTIONS(2744),
    [anon_sym_CG_EXTERN] = ACTIONS(2744),
    [anon_sym_CG_INLINE] = ACTIONS(2744),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2744),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2744),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2744),
    [anon_sym_IBOutlet] = ACTIONS(2744),
    [anon_sym_IBInspectable] = ACTIONS(2744),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2744),
    [anon_sym_NS_INLINE] = ACTIONS(2744),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2744),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2744),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2744),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2744),
    [anon_sym_const] = ACTIONS(2744),
    [anon_sym_constexpr] = ACTIONS(2744),
    [anon_sym_volatile] = ACTIONS(2744),
    [anon_sym_restrict] = ACTIONS(2744),
    [anon_sym___restrict__] = ACTIONS(2744),
    [anon_sym__Atomic] = ACTIONS(2744),
    [anon_sym__Noreturn] = ACTIONS(2744),
    [anon_sym__Nonnull] = ACTIONS(2744),
    [anon_sym_nullable] = ACTIONS(2744),
    [anon_sym__Complex] = ACTIONS(2744),
    [anon_sym__Nullable] = ACTIONS(2744),
    [anon_sym__Nullable_result] = ACTIONS(2744),
    [anon_sym__Null_unspecified] = ACTIONS(2744),
    [anon_sym___autoreleasing] = ACTIONS(2744),
    [anon_sym___block] = ACTIONS(2744),
    [anon_sym___bridge] = ACTIONS(2744),
    [anon_sym___bridge_retained] = ACTIONS(2744),
    [anon_sym___bridge_transfer] = ACTIONS(2744),
    [anon_sym___complex] = ACTIONS(2744),
    [anon_sym___const] = ACTIONS(2744),
    [anon_sym___imag] = ACTIONS(2744),
    [anon_sym___kindof] = ACTIONS(2744),
    [anon_sym___nonnull] = ACTIONS(2744),
    [anon_sym___nullable] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2744),
    [anon_sym___real] = ACTIONS(2744),
    [anon_sym___strong] = ACTIONS(2744),
    [anon_sym___unsafe_unretained] = ACTIONS(2744),
    [anon_sym___unused] = ACTIONS(2744),
    [anon_sym___weak] = ACTIONS(2744),
    [anon_sym_alignas] = ACTIONS(2744),
    [anon_sym__Alignas] = ACTIONS(2744),
    [sym_primitive_type] = ACTIONS(2744),
    [anon_sym_enum] = ACTIONS(2744),
    [anon_sym_struct] = ACTIONS(2744),
    [anon_sym_union] = ACTIONS(2744),
    [anon_sym_if] = ACTIONS(2744),
    [anon_sym_switch] = ACTIONS(2744),
    [anon_sym_case] = ACTIONS(2744),
    [anon_sym_default] = ACTIONS(2744),
    [anon_sym_while] = ACTIONS(2744),
    [anon_sym_do] = ACTIONS(2744),
    [anon_sym_for] = ACTIONS(2744),
    [anon_sym_in] = ACTIONS(2744),
    [anon_sym_return] = ACTIONS(2744),
    [anon_sym_break] = ACTIONS(2744),
    [anon_sym_continue] = ACTIONS(2744),
    [anon_sym_goto] = ACTIONS(2744),
    [anon_sym_DASH_DASH] = ACTIONS(2746),
    [anon_sym_PLUS_PLUS] = ACTIONS(2746),
    [anon_sym_sizeof] = ACTIONS(2744),
    [anon_sym___alignof__] = ACTIONS(2744),
    [anon_sym___alignof] = ACTIONS(2744),
    [anon_sym__alignof] = ACTIONS(2744),
    [anon_sym_alignof] = ACTIONS(2744),
    [anon_sym__Alignof] = ACTIONS(2744),
    [anon_sym_offsetof] = ACTIONS(2744),
    [anon_sym__Generic] = ACTIONS(2744),
    [anon_sym_asm] = ACTIONS(2744),
    [anon_sym___asm__] = ACTIONS(2744),
    [anon_sym___asm] = ACTIONS(2744),
    [sym_number_literal] = ACTIONS(2746),
    [anon_sym_L_SQUOTE] = ACTIONS(2746),
    [anon_sym_u_SQUOTE] = ACTIONS(2746),
    [anon_sym_U_SQUOTE] = ACTIONS(2746),
    [anon_sym_u8_SQUOTE] = ACTIONS(2746),
    [anon_sym_SQUOTE] = ACTIONS(2746),
    [anon_sym_AT] = ACTIONS(2744),
    [anon_sym_DQUOTE] = ACTIONS(2746),
    [anon_sym_L_DQUOTE] = ACTIONS(2746),
    [anon_sym_u_DQUOTE] = ACTIONS(2746),
    [anon_sym_U_DQUOTE] = ACTIONS(2746),
    [anon_sym_u8_DQUOTE] = ACTIONS(2746),
    [sym_true] = ACTIONS(2744),
    [sym_false] = ACTIONS(2744),
    [anon_sym_NULL] = ACTIONS(2744),
    [anon_sym_nullptr] = ACTIONS(2744),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2744),
    [anon_sym___typeof] = ACTIONS(2744),
    [anon_sym_typeof] = ACTIONS(2744),
    [anon_sym_ATimport] = ACTIONS(2746),
    [aux_sym_preproc_undef_token1] = ACTIONS(2744),
    [anon_sym_POUND] = ACTIONS(2744),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2744),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2744),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2744),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2744),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2744),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2744),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2744),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_API_AVAILABLE] = ACTIONS(2744),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_API_DEPRECATED] = ACTIONS(2744),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2744),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym___deprecated_msg] = ACTIONS(2744),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2744),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2744),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2744),
    [anon_sym_ATprotocol] = ACTIONS(2746),
    [anon_sym_ATinterface] = ACTIONS(2746),
    [anon_sym_ATimplementation] = ACTIONS(2746),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2746),
    [anon_sym_ATsynthesize] = ACTIONS(2746),
    [anon_sym_ATdynamic] = ACTIONS(2746),
    [anon_sym_ATtry] = ACTIONS(2746),
    [anon_sym___try] = ACTIONS(2744),
    [anon_sym_ATthrow] = ACTIONS(2746),
    [anon_sym_ATselector] = ACTIONS(2746),
    [anon_sym_ATavailable] = ACTIONS(2746),
    [anon_sym___builtin_available] = ACTIONS(2744),
    [anon_sym_va_arg] = ACTIONS(2744),
    [anon_sym_ATencode] = ACTIONS(2746),
    [anon_sym_ATsynchronized] = ACTIONS(2746),
    [anon_sym_BOOL] = ACTIONS(2744),
    [anon_sym_IMP] = ACTIONS(2744),
    [anon_sym_SEL] = ACTIONS(2744),
    [anon_sym_Class] = ACTIONS(2744),
    [anon_sym_id] = ACTIONS(2744),
  },
  [STATE(602)] = {
    [sym_identifier] = ACTIONS(2748),
    [aux_sym_preproc_include_token1] = ACTIONS(2748),
    [aux_sym_preproc_include_token2] = ACTIONS(2748),
    [aux_sym_preproc_def_token1] = ACTIONS(2748),
    [aux_sym_preproc_if_token1] = ACTIONS(2748),
    [aux_sym_preproc_if_token2] = ACTIONS(2748),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2748),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2748),
    [sym_preproc_directive] = ACTIONS(2748),
    [anon_sym_LPAREN2] = ACTIONS(2750),
    [anon_sym_BANG] = ACTIONS(2750),
    [anon_sym_TILDE] = ACTIONS(2750),
    [anon_sym_DASH] = ACTIONS(2748),
    [anon_sym_PLUS] = ACTIONS(2748),
    [anon_sym_STAR] = ACTIONS(2750),
    [anon_sym_CARET] = ACTIONS(2750),
    [anon_sym_AMP] = ACTIONS(2750),
    [anon_sym_SEMI] = ACTIONS(2750),
    [anon_sym___extension__] = ACTIONS(2748),
    [anon_sym_typedef] = ACTIONS(2748),
    [anon_sym_extern] = ACTIONS(2748),
    [anon_sym___attribute__] = ACTIONS(2748),
    [anon_sym___attribute] = ACTIONS(2748),
    [anon_sym_noreturn] = ACTIONS(2748),
    [anon_sym_LBRACK] = ACTIONS(2750),
    [anon_sym___declspec] = ACTIONS(2748),
    [anon_sym___cdecl] = ACTIONS(2748),
    [anon_sym___clrcall] = ACTIONS(2748),
    [anon_sym___stdcall] = ACTIONS(2748),
    [anon_sym___fastcall] = ACTIONS(2748),
    [anon_sym___thiscall] = ACTIONS(2748),
    [anon_sym___vectorcall] = ACTIONS(2748),
    [anon_sym_LBRACE] = ACTIONS(2750),
    [anon_sym_signed] = ACTIONS(2748),
    [anon_sym_unsigned] = ACTIONS(2748),
    [anon_sym_long] = ACTIONS(2748),
    [anon_sym_short] = ACTIONS(2748),
    [anon_sym_static] = ACTIONS(2748),
    [anon_sym_ATautoreleasepool] = ACTIONS(2750),
    [anon_sym_auto] = ACTIONS(2748),
    [anon_sym_register] = ACTIONS(2748),
    [anon_sym_inline] = ACTIONS(2748),
    [anon_sym___inline] = ACTIONS(2748),
    [anon_sym___inline__] = ACTIONS(2748),
    [anon_sym___forceinline] = ACTIONS(2748),
    [anon_sym_thread_local] = ACTIONS(2748),
    [anon_sym___thread] = ACTIONS(2748),
    [anon_sym_CG_EXTERN] = ACTIONS(2748),
    [anon_sym_CG_INLINE] = ACTIONS(2748),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2748),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2748),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2748),
    [anon_sym_IBOutlet] = ACTIONS(2748),
    [anon_sym_IBInspectable] = ACTIONS(2748),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2748),
    [anon_sym_NS_INLINE] = ACTIONS(2748),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2748),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2748),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2748),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2748),
    [anon_sym_const] = ACTIONS(2748),
    [anon_sym_constexpr] = ACTIONS(2748),
    [anon_sym_volatile] = ACTIONS(2748),
    [anon_sym_restrict] = ACTIONS(2748),
    [anon_sym___restrict__] = ACTIONS(2748),
    [anon_sym__Atomic] = ACTIONS(2748),
    [anon_sym__Noreturn] = ACTIONS(2748),
    [anon_sym__Nonnull] = ACTIONS(2748),
    [anon_sym_nullable] = ACTIONS(2748),
    [anon_sym__Complex] = ACTIONS(2748),
    [anon_sym__Nullable] = ACTIONS(2748),
    [anon_sym__Nullable_result] = ACTIONS(2748),
    [anon_sym__Null_unspecified] = ACTIONS(2748),
    [anon_sym___autoreleasing] = ACTIONS(2748),
    [anon_sym___block] = ACTIONS(2748),
    [anon_sym___bridge] = ACTIONS(2748),
    [anon_sym___bridge_retained] = ACTIONS(2748),
    [anon_sym___bridge_transfer] = ACTIONS(2748),
    [anon_sym___complex] = ACTIONS(2748),
    [anon_sym___const] = ACTIONS(2748),
    [anon_sym___imag] = ACTIONS(2748),
    [anon_sym___kindof] = ACTIONS(2748),
    [anon_sym___nonnull] = ACTIONS(2748),
    [anon_sym___nullable] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2748),
    [anon_sym___real] = ACTIONS(2748),
    [anon_sym___strong] = ACTIONS(2748),
    [anon_sym___unsafe_unretained] = ACTIONS(2748),
    [anon_sym___unused] = ACTIONS(2748),
    [anon_sym___weak] = ACTIONS(2748),
    [anon_sym_alignas] = ACTIONS(2748),
    [anon_sym__Alignas] = ACTIONS(2748),
    [sym_primitive_type] = ACTIONS(2748),
    [anon_sym_enum] = ACTIONS(2748),
    [anon_sym_struct] = ACTIONS(2748),
    [anon_sym_union] = ACTIONS(2748),
    [anon_sym_if] = ACTIONS(2748),
    [anon_sym_switch] = ACTIONS(2748),
    [anon_sym_case] = ACTIONS(2748),
    [anon_sym_default] = ACTIONS(2748),
    [anon_sym_while] = ACTIONS(2748),
    [anon_sym_do] = ACTIONS(2748),
    [anon_sym_for] = ACTIONS(2748),
    [anon_sym_in] = ACTIONS(2748),
    [anon_sym_return] = ACTIONS(2748),
    [anon_sym_break] = ACTIONS(2748),
    [anon_sym_continue] = ACTIONS(2748),
    [anon_sym_goto] = ACTIONS(2748),
    [anon_sym_DASH_DASH] = ACTIONS(2750),
    [anon_sym_PLUS_PLUS] = ACTIONS(2750),
    [anon_sym_sizeof] = ACTIONS(2748),
    [anon_sym___alignof__] = ACTIONS(2748),
    [anon_sym___alignof] = ACTIONS(2748),
    [anon_sym__alignof] = ACTIONS(2748),
    [anon_sym_alignof] = ACTIONS(2748),
    [anon_sym__Alignof] = ACTIONS(2748),
    [anon_sym_offsetof] = ACTIONS(2748),
    [anon_sym__Generic] = ACTIONS(2748),
    [anon_sym_asm] = ACTIONS(2748),
    [anon_sym___asm__] = ACTIONS(2748),
    [anon_sym___asm] = ACTIONS(2748),
    [sym_number_literal] = ACTIONS(2750),
    [anon_sym_L_SQUOTE] = ACTIONS(2750),
    [anon_sym_u_SQUOTE] = ACTIONS(2750),
    [anon_sym_U_SQUOTE] = ACTIONS(2750),
    [anon_sym_u8_SQUOTE] = ACTIONS(2750),
    [anon_sym_SQUOTE] = ACTIONS(2750),
    [anon_sym_AT] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_L_DQUOTE] = ACTIONS(2750),
    [anon_sym_u_DQUOTE] = ACTIONS(2750),
    [anon_sym_U_DQUOTE] = ACTIONS(2750),
    [anon_sym_u8_DQUOTE] = ACTIONS(2750),
    [sym_true] = ACTIONS(2748),
    [sym_false] = ACTIONS(2748),
    [anon_sym_NULL] = ACTIONS(2748),
    [anon_sym_nullptr] = ACTIONS(2748),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2748),
    [anon_sym___typeof] = ACTIONS(2748),
    [anon_sym_typeof] = ACTIONS(2748),
    [anon_sym_ATimport] = ACTIONS(2750),
    [aux_sym_preproc_undef_token1] = ACTIONS(2748),
    [anon_sym_POUND] = ACTIONS(2748),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2748),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2748),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2748),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2748),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2748),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2748),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2748),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_API_AVAILABLE] = ACTIONS(2748),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_API_DEPRECATED] = ACTIONS(2748),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2748),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym___deprecated_msg] = ACTIONS(2748),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2748),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2748),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2748),
    [anon_sym_ATprotocol] = ACTIONS(2750),
    [anon_sym_ATinterface] = ACTIONS(2750),
    [anon_sym_ATimplementation] = ACTIONS(2750),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2750),
    [anon_sym_ATsynthesize] = ACTIONS(2750),
    [anon_sym_ATdynamic] = ACTIONS(2750),
    [anon_sym_ATtry] = ACTIONS(2750),
    [anon_sym___try] = ACTIONS(2748),
    [anon_sym_ATthrow] = ACTIONS(2750),
    [anon_sym_ATselector] = ACTIONS(2750),
    [anon_sym_ATavailable] = ACTIONS(2750),
    [anon_sym___builtin_available] = ACTIONS(2748),
    [anon_sym_va_arg] = ACTIONS(2748),
    [anon_sym_ATencode] = ACTIONS(2750),
    [anon_sym_ATsynchronized] = ACTIONS(2750),
    [anon_sym_BOOL] = ACTIONS(2748),
    [anon_sym_IMP] = ACTIONS(2748),
    [anon_sym_SEL] = ACTIONS(2748),
    [anon_sym_Class] = ACTIONS(2748),
    [anon_sym_id] = ACTIONS(2748),
  },
  [STATE(603)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token2] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATsynthesize] = ACTIONS(2754),
    [anon_sym_ATdynamic] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(604)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token2] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATsynthesize] = ACTIONS(2754),
    [anon_sym_ATdynamic] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(605)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token2] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATsynthesize] = ACTIONS(2754),
    [anon_sym_ATdynamic] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(606)] = {
    [sym_identifier] = ACTIONS(2756),
    [aux_sym_preproc_include_token1] = ACTIONS(2756),
    [aux_sym_preproc_include_token2] = ACTIONS(2756),
    [aux_sym_preproc_def_token1] = ACTIONS(2756),
    [aux_sym_preproc_if_token1] = ACTIONS(2756),
    [aux_sym_preproc_if_token2] = ACTIONS(2756),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2756),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2756),
    [sym_preproc_directive] = ACTIONS(2756),
    [anon_sym_LPAREN2] = ACTIONS(2758),
    [anon_sym_BANG] = ACTIONS(2758),
    [anon_sym_TILDE] = ACTIONS(2758),
    [anon_sym_DASH] = ACTIONS(2756),
    [anon_sym_PLUS] = ACTIONS(2756),
    [anon_sym_STAR] = ACTIONS(2758),
    [anon_sym_CARET] = ACTIONS(2758),
    [anon_sym_AMP] = ACTIONS(2758),
    [anon_sym_SEMI] = ACTIONS(2758),
    [anon_sym___extension__] = ACTIONS(2756),
    [anon_sym_typedef] = ACTIONS(2756),
    [anon_sym_extern] = ACTIONS(2756),
    [anon_sym___attribute__] = ACTIONS(2756),
    [anon_sym___attribute] = ACTIONS(2756),
    [anon_sym_noreturn] = ACTIONS(2756),
    [anon_sym_LBRACK] = ACTIONS(2758),
    [anon_sym___declspec] = ACTIONS(2756),
    [anon_sym___cdecl] = ACTIONS(2756),
    [anon_sym___clrcall] = ACTIONS(2756),
    [anon_sym___stdcall] = ACTIONS(2756),
    [anon_sym___fastcall] = ACTIONS(2756),
    [anon_sym___thiscall] = ACTIONS(2756),
    [anon_sym___vectorcall] = ACTIONS(2756),
    [anon_sym_LBRACE] = ACTIONS(2758),
    [anon_sym_signed] = ACTIONS(2756),
    [anon_sym_unsigned] = ACTIONS(2756),
    [anon_sym_long] = ACTIONS(2756),
    [anon_sym_short] = ACTIONS(2756),
    [anon_sym_static] = ACTIONS(2756),
    [anon_sym_ATautoreleasepool] = ACTIONS(2758),
    [anon_sym_auto] = ACTIONS(2756),
    [anon_sym_register] = ACTIONS(2756),
    [anon_sym_inline] = ACTIONS(2756),
    [anon_sym___inline] = ACTIONS(2756),
    [anon_sym___inline__] = ACTIONS(2756),
    [anon_sym___forceinline] = ACTIONS(2756),
    [anon_sym_thread_local] = ACTIONS(2756),
    [anon_sym___thread] = ACTIONS(2756),
    [anon_sym_CG_EXTERN] = ACTIONS(2756),
    [anon_sym_CG_INLINE] = ACTIONS(2756),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2756),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2756),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2756),
    [anon_sym_IBOutlet] = ACTIONS(2756),
    [anon_sym_IBInspectable] = ACTIONS(2756),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2756),
    [anon_sym_NS_INLINE] = ACTIONS(2756),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2756),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2756),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2756),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2756),
    [anon_sym_const] = ACTIONS(2756),
    [anon_sym_constexpr] = ACTIONS(2756),
    [anon_sym_volatile] = ACTIONS(2756),
    [anon_sym_restrict] = ACTIONS(2756),
    [anon_sym___restrict__] = ACTIONS(2756),
    [anon_sym__Atomic] = ACTIONS(2756),
    [anon_sym__Noreturn] = ACTIONS(2756),
    [anon_sym__Nonnull] = ACTIONS(2756),
    [anon_sym_nullable] = ACTIONS(2756),
    [anon_sym__Complex] = ACTIONS(2756),
    [anon_sym__Nullable] = ACTIONS(2756),
    [anon_sym__Nullable_result] = ACTIONS(2756),
    [anon_sym__Null_unspecified] = ACTIONS(2756),
    [anon_sym___autoreleasing] = ACTIONS(2756),
    [anon_sym___block] = ACTIONS(2756),
    [anon_sym___bridge] = ACTIONS(2756),
    [anon_sym___bridge_retained] = ACTIONS(2756),
    [anon_sym___bridge_transfer] = ACTIONS(2756),
    [anon_sym___complex] = ACTIONS(2756),
    [anon_sym___const] = ACTIONS(2756),
    [anon_sym___imag] = ACTIONS(2756),
    [anon_sym___kindof] = ACTIONS(2756),
    [anon_sym___nonnull] = ACTIONS(2756),
    [anon_sym___nullable] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2756),
    [anon_sym___real] = ACTIONS(2756),
    [anon_sym___strong] = ACTIONS(2756),
    [anon_sym___unsafe_unretained] = ACTIONS(2756),
    [anon_sym___unused] = ACTIONS(2756),
    [anon_sym___weak] = ACTIONS(2756),
    [anon_sym_alignas] = ACTIONS(2756),
    [anon_sym__Alignas] = ACTIONS(2756),
    [sym_primitive_type] = ACTIONS(2756),
    [anon_sym_enum] = ACTIONS(2756),
    [anon_sym_struct] = ACTIONS(2756),
    [anon_sym_union] = ACTIONS(2756),
    [anon_sym_if] = ACTIONS(2756),
    [anon_sym_switch] = ACTIONS(2756),
    [anon_sym_case] = ACTIONS(2756),
    [anon_sym_default] = ACTIONS(2756),
    [anon_sym_while] = ACTIONS(2756),
    [anon_sym_do] = ACTIONS(2756),
    [anon_sym_for] = ACTIONS(2756),
    [anon_sym_in] = ACTIONS(2756),
    [anon_sym_return] = ACTIONS(2756),
    [anon_sym_break] = ACTIONS(2756),
    [anon_sym_continue] = ACTIONS(2756),
    [anon_sym_goto] = ACTIONS(2756),
    [anon_sym_DASH_DASH] = ACTIONS(2758),
    [anon_sym_PLUS_PLUS] = ACTIONS(2758),
    [anon_sym_sizeof] = ACTIONS(2756),
    [anon_sym___alignof__] = ACTIONS(2756),
    [anon_sym___alignof] = ACTIONS(2756),
    [anon_sym__alignof] = ACTIONS(2756),
    [anon_sym_alignof] = ACTIONS(2756),
    [anon_sym__Alignof] = ACTIONS(2756),
    [anon_sym_offsetof] = ACTIONS(2756),
    [anon_sym__Generic] = ACTIONS(2756),
    [anon_sym_asm] = ACTIONS(2756),
    [anon_sym___asm__] = ACTIONS(2756),
    [anon_sym___asm] = ACTIONS(2756),
    [sym_number_literal] = ACTIONS(2758),
    [anon_sym_L_SQUOTE] = ACTIONS(2758),
    [anon_sym_u_SQUOTE] = ACTIONS(2758),
    [anon_sym_U_SQUOTE] = ACTIONS(2758),
    [anon_sym_u8_SQUOTE] = ACTIONS(2758),
    [anon_sym_SQUOTE] = ACTIONS(2758),
    [anon_sym_AT] = ACTIONS(2756),
    [anon_sym_DQUOTE] = ACTIONS(2758),
    [anon_sym_L_DQUOTE] = ACTIONS(2758),
    [anon_sym_u_DQUOTE] = ACTIONS(2758),
    [anon_sym_U_DQUOTE] = ACTIONS(2758),
    [anon_sym_u8_DQUOTE] = ACTIONS(2758),
    [sym_true] = ACTIONS(2756),
    [sym_false] = ACTIONS(2756),
    [anon_sym_NULL] = ACTIONS(2756),
    [anon_sym_nullptr] = ACTIONS(2756),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2756),
    [anon_sym___typeof] = ACTIONS(2756),
    [anon_sym_typeof] = ACTIONS(2756),
    [anon_sym_ATimport] = ACTIONS(2758),
    [aux_sym_preproc_undef_token1] = ACTIONS(2756),
    [anon_sym_POUND] = ACTIONS(2756),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2756),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2756),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2756),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2756),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2756),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2756),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2756),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_API_AVAILABLE] = ACTIONS(2756),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_API_DEPRECATED] = ACTIONS(2756),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2756),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym___deprecated_msg] = ACTIONS(2756),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2756),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2756),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2756),
    [anon_sym_ATprotocol] = ACTIONS(2758),
    [anon_sym_ATinterface] = ACTIONS(2758),
    [anon_sym_ATimplementation] = ACTIONS(2758),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2758),
    [anon_sym_ATsynthesize] = ACTIONS(2758),
    [anon_sym_ATdynamic] = ACTIONS(2758),
    [anon_sym_ATtry] = ACTIONS(2758),
    [anon_sym___try] = ACTIONS(2756),
    [anon_sym_ATthrow] = ACTIONS(2758),
    [anon_sym_ATselector] = ACTIONS(2758),
    [anon_sym_ATavailable] = ACTIONS(2758),
    [anon_sym___builtin_available] = ACTIONS(2756),
    [anon_sym_va_arg] = ACTIONS(2756),
    [anon_sym_ATencode] = ACTIONS(2758),
    [anon_sym_ATsynchronized] = ACTIONS(2758),
    [anon_sym_BOOL] = ACTIONS(2756),
    [anon_sym_IMP] = ACTIONS(2756),
    [anon_sym_SEL] = ACTIONS(2756),
    [anon_sym_Class] = ACTIONS(2756),
    [anon_sym_id] = ACTIONS(2756),
  },
  [STATE(607)] = {
    [sym_identifier] = ACTIONS(2760),
    [aux_sym_preproc_include_token1] = ACTIONS(2760),
    [aux_sym_preproc_include_token2] = ACTIONS(2760),
    [aux_sym_preproc_def_token1] = ACTIONS(2760),
    [aux_sym_preproc_if_token1] = ACTIONS(2760),
    [aux_sym_preproc_if_token2] = ACTIONS(2760),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2760),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2760),
    [sym_preproc_directive] = ACTIONS(2760),
    [anon_sym_LPAREN2] = ACTIONS(2762),
    [anon_sym_BANG] = ACTIONS(2762),
    [anon_sym_TILDE] = ACTIONS(2762),
    [anon_sym_DASH] = ACTIONS(2760),
    [anon_sym_PLUS] = ACTIONS(2760),
    [anon_sym_STAR] = ACTIONS(2762),
    [anon_sym_CARET] = ACTIONS(2762),
    [anon_sym_AMP] = ACTIONS(2762),
    [anon_sym_SEMI] = ACTIONS(2762),
    [anon_sym___extension__] = ACTIONS(2760),
    [anon_sym_typedef] = ACTIONS(2760),
    [anon_sym_extern] = ACTIONS(2760),
    [anon_sym___attribute__] = ACTIONS(2760),
    [anon_sym___attribute] = ACTIONS(2760),
    [anon_sym_noreturn] = ACTIONS(2760),
    [anon_sym_LBRACK] = ACTIONS(2762),
    [anon_sym___declspec] = ACTIONS(2760),
    [anon_sym___cdecl] = ACTIONS(2760),
    [anon_sym___clrcall] = ACTIONS(2760),
    [anon_sym___stdcall] = ACTIONS(2760),
    [anon_sym___fastcall] = ACTIONS(2760),
    [anon_sym___thiscall] = ACTIONS(2760),
    [anon_sym___vectorcall] = ACTIONS(2760),
    [anon_sym_LBRACE] = ACTIONS(2762),
    [anon_sym_signed] = ACTIONS(2760),
    [anon_sym_unsigned] = ACTIONS(2760),
    [anon_sym_long] = ACTIONS(2760),
    [anon_sym_short] = ACTIONS(2760),
    [anon_sym_static] = ACTIONS(2760),
    [anon_sym_ATautoreleasepool] = ACTIONS(2762),
    [anon_sym_auto] = ACTIONS(2760),
    [anon_sym_register] = ACTIONS(2760),
    [anon_sym_inline] = ACTIONS(2760),
    [anon_sym___inline] = ACTIONS(2760),
    [anon_sym___inline__] = ACTIONS(2760),
    [anon_sym___forceinline] = ACTIONS(2760),
    [anon_sym_thread_local] = ACTIONS(2760),
    [anon_sym___thread] = ACTIONS(2760),
    [anon_sym_CG_EXTERN] = ACTIONS(2760),
    [anon_sym_CG_INLINE] = ACTIONS(2760),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2760),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2760),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2760),
    [anon_sym_IBOutlet] = ACTIONS(2760),
    [anon_sym_IBInspectable] = ACTIONS(2760),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2760),
    [anon_sym_NS_INLINE] = ACTIONS(2760),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2760),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2760),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2760),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2760),
    [anon_sym_const] = ACTIONS(2760),
    [anon_sym_constexpr] = ACTIONS(2760),
    [anon_sym_volatile] = ACTIONS(2760),
    [anon_sym_restrict] = ACTIONS(2760),
    [anon_sym___restrict__] = ACTIONS(2760),
    [anon_sym__Atomic] = ACTIONS(2760),
    [anon_sym__Noreturn] = ACTIONS(2760),
    [anon_sym__Nonnull] = ACTIONS(2760),
    [anon_sym_nullable] = ACTIONS(2760),
    [anon_sym__Complex] = ACTIONS(2760),
    [anon_sym__Nullable] = ACTIONS(2760),
    [anon_sym__Nullable_result] = ACTIONS(2760),
    [anon_sym__Null_unspecified] = ACTIONS(2760),
    [anon_sym___autoreleasing] = ACTIONS(2760),
    [anon_sym___block] = ACTIONS(2760),
    [anon_sym___bridge] = ACTIONS(2760),
    [anon_sym___bridge_retained] = ACTIONS(2760),
    [anon_sym___bridge_transfer] = ACTIONS(2760),
    [anon_sym___complex] = ACTIONS(2760),
    [anon_sym___const] = ACTIONS(2760),
    [anon_sym___imag] = ACTIONS(2760),
    [anon_sym___kindof] = ACTIONS(2760),
    [anon_sym___nonnull] = ACTIONS(2760),
    [anon_sym___nullable] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2760),
    [anon_sym___real] = ACTIONS(2760),
    [anon_sym___strong] = ACTIONS(2760),
    [anon_sym___unsafe_unretained] = ACTIONS(2760),
    [anon_sym___unused] = ACTIONS(2760),
    [anon_sym___weak] = ACTIONS(2760),
    [anon_sym_alignas] = ACTIONS(2760),
    [anon_sym__Alignas] = ACTIONS(2760),
    [sym_primitive_type] = ACTIONS(2760),
    [anon_sym_enum] = ACTIONS(2760),
    [anon_sym_struct] = ACTIONS(2760),
    [anon_sym_union] = ACTIONS(2760),
    [anon_sym_if] = ACTIONS(2760),
    [anon_sym_switch] = ACTIONS(2760),
    [anon_sym_case] = ACTIONS(2760),
    [anon_sym_default] = ACTIONS(2760),
    [anon_sym_while] = ACTIONS(2760),
    [anon_sym_do] = ACTIONS(2760),
    [anon_sym_for] = ACTIONS(2760),
    [anon_sym_in] = ACTIONS(2760),
    [anon_sym_return] = ACTIONS(2760),
    [anon_sym_break] = ACTIONS(2760),
    [anon_sym_continue] = ACTIONS(2760),
    [anon_sym_goto] = ACTIONS(2760),
    [anon_sym_DASH_DASH] = ACTIONS(2762),
    [anon_sym_PLUS_PLUS] = ACTIONS(2762),
    [anon_sym_sizeof] = ACTIONS(2760),
    [anon_sym___alignof__] = ACTIONS(2760),
    [anon_sym___alignof] = ACTIONS(2760),
    [anon_sym__alignof] = ACTIONS(2760),
    [anon_sym_alignof] = ACTIONS(2760),
    [anon_sym__Alignof] = ACTIONS(2760),
    [anon_sym_offsetof] = ACTIONS(2760),
    [anon_sym__Generic] = ACTIONS(2760),
    [anon_sym_asm] = ACTIONS(2760),
    [anon_sym___asm__] = ACTIONS(2760),
    [anon_sym___asm] = ACTIONS(2760),
    [sym_number_literal] = ACTIONS(2762),
    [anon_sym_L_SQUOTE] = ACTIONS(2762),
    [anon_sym_u_SQUOTE] = ACTIONS(2762),
    [anon_sym_U_SQUOTE] = ACTIONS(2762),
    [anon_sym_u8_SQUOTE] = ACTIONS(2762),
    [anon_sym_SQUOTE] = ACTIONS(2762),
    [anon_sym_AT] = ACTIONS(2760),
    [anon_sym_DQUOTE] = ACTIONS(2762),
    [anon_sym_L_DQUOTE] = ACTIONS(2762),
    [anon_sym_u_DQUOTE] = ACTIONS(2762),
    [anon_sym_U_DQUOTE] = ACTIONS(2762),
    [anon_sym_u8_DQUOTE] = ACTIONS(2762),
    [sym_true] = ACTIONS(2760),
    [sym_false] = ACTIONS(2760),
    [anon_sym_NULL] = ACTIONS(2760),
    [anon_sym_nullptr] = ACTIONS(2760),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2760),
    [anon_sym___typeof] = ACTIONS(2760),
    [anon_sym_typeof] = ACTIONS(2760),
    [anon_sym_ATimport] = ACTIONS(2762),
    [aux_sym_preproc_undef_token1] = ACTIONS(2760),
    [anon_sym_POUND] = ACTIONS(2760),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2760),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2760),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2760),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2760),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2760),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2760),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2760),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_API_AVAILABLE] = ACTIONS(2760),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_API_DEPRECATED] = ACTIONS(2760),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2760),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym___deprecated_msg] = ACTIONS(2760),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2760),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2760),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2760),
    [anon_sym_ATprotocol] = ACTIONS(2762),
    [anon_sym_ATinterface] = ACTIONS(2762),
    [anon_sym_ATimplementation] = ACTIONS(2762),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2762),
    [anon_sym_ATsynthesize] = ACTIONS(2762),
    [anon_sym_ATdynamic] = ACTIONS(2762),
    [anon_sym_ATtry] = ACTIONS(2762),
    [anon_sym___try] = ACTIONS(2760),
    [anon_sym_ATthrow] = ACTIONS(2762),
    [anon_sym_ATselector] = ACTIONS(2762),
    [anon_sym_ATavailable] = ACTIONS(2762),
    [anon_sym___builtin_available] = ACTIONS(2760),
    [anon_sym_va_arg] = ACTIONS(2760),
    [anon_sym_ATencode] = ACTIONS(2762),
    [anon_sym_ATsynchronized] = ACTIONS(2762),
    [anon_sym_BOOL] = ACTIONS(2760),
    [anon_sym_IMP] = ACTIONS(2760),
    [anon_sym_SEL] = ACTIONS(2760),
    [anon_sym_Class] = ACTIONS(2760),
    [anon_sym_id] = ACTIONS(2760),
  },
  [STATE(608)] = {
    [sym_identifier] = ACTIONS(2764),
    [aux_sym_preproc_include_token1] = ACTIONS(2764),
    [aux_sym_preproc_include_token2] = ACTIONS(2764),
    [aux_sym_preproc_def_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token2] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2764),
    [sym_preproc_directive] = ACTIONS(2764),
    [anon_sym_LPAREN2] = ACTIONS(2766),
    [anon_sym_BANG] = ACTIONS(2766),
    [anon_sym_TILDE] = ACTIONS(2766),
    [anon_sym_DASH] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(2764),
    [anon_sym_STAR] = ACTIONS(2766),
    [anon_sym_CARET] = ACTIONS(2766),
    [anon_sym_AMP] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2766),
    [anon_sym___extension__] = ACTIONS(2764),
    [anon_sym_typedef] = ACTIONS(2764),
    [anon_sym_extern] = ACTIONS(2764),
    [anon_sym___attribute__] = ACTIONS(2764),
    [anon_sym___attribute] = ACTIONS(2764),
    [anon_sym_noreturn] = ACTIONS(2764),
    [anon_sym_LBRACK] = ACTIONS(2766),
    [anon_sym___declspec] = ACTIONS(2764),
    [anon_sym___cdecl] = ACTIONS(2764),
    [anon_sym___clrcall] = ACTIONS(2764),
    [anon_sym___stdcall] = ACTIONS(2764),
    [anon_sym___fastcall] = ACTIONS(2764),
    [anon_sym___thiscall] = ACTIONS(2764),
    [anon_sym___vectorcall] = ACTIONS(2764),
    [anon_sym_LBRACE] = ACTIONS(2766),
    [anon_sym_signed] = ACTIONS(2764),
    [anon_sym_unsigned] = ACTIONS(2764),
    [anon_sym_long] = ACTIONS(2764),
    [anon_sym_short] = ACTIONS(2764),
    [anon_sym_static] = ACTIONS(2764),
    [anon_sym_ATautoreleasepool] = ACTIONS(2766),
    [anon_sym_auto] = ACTIONS(2764),
    [anon_sym_register] = ACTIONS(2764),
    [anon_sym_inline] = ACTIONS(2764),
    [anon_sym___inline] = ACTIONS(2764),
    [anon_sym___inline__] = ACTIONS(2764),
    [anon_sym___forceinline] = ACTIONS(2764),
    [anon_sym_thread_local] = ACTIONS(2764),
    [anon_sym___thread] = ACTIONS(2764),
    [anon_sym_CG_EXTERN] = ACTIONS(2764),
    [anon_sym_CG_INLINE] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2764),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2764),
    [anon_sym_IBOutlet] = ACTIONS(2764),
    [anon_sym_IBInspectable] = ACTIONS(2764),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2764),
    [anon_sym_NS_INLINE] = ACTIONS(2764),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2764),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2764),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2764),
    [anon_sym_const] = ACTIONS(2764),
    [anon_sym_constexpr] = ACTIONS(2764),
    [anon_sym_volatile] = ACTIONS(2764),
    [anon_sym_restrict] = ACTIONS(2764),
    [anon_sym___restrict__] = ACTIONS(2764),
    [anon_sym__Atomic] = ACTIONS(2764),
    [anon_sym__Noreturn] = ACTIONS(2764),
    [anon_sym__Nonnull] = ACTIONS(2764),
    [anon_sym_nullable] = ACTIONS(2764),
    [anon_sym__Complex] = ACTIONS(2764),
    [anon_sym__Nullable] = ACTIONS(2764),
    [anon_sym__Nullable_result] = ACTIONS(2764),
    [anon_sym__Null_unspecified] = ACTIONS(2764),
    [anon_sym___autoreleasing] = ACTIONS(2764),
    [anon_sym___block] = ACTIONS(2764),
    [anon_sym___bridge] = ACTIONS(2764),
    [anon_sym___bridge_retained] = ACTIONS(2764),
    [anon_sym___bridge_transfer] = ACTIONS(2764),
    [anon_sym___complex] = ACTIONS(2764),
    [anon_sym___const] = ACTIONS(2764),
    [anon_sym___imag] = ACTIONS(2764),
    [anon_sym___kindof] = ACTIONS(2764),
    [anon_sym___nonnull] = ACTIONS(2764),
    [anon_sym___nullable] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2764),
    [anon_sym___real] = ACTIONS(2764),
    [anon_sym___strong] = ACTIONS(2764),
    [anon_sym___unsafe_unretained] = ACTIONS(2764),
    [anon_sym___unused] = ACTIONS(2764),
    [anon_sym___weak] = ACTIONS(2764),
    [anon_sym_alignas] = ACTIONS(2764),
    [anon_sym__Alignas] = ACTIONS(2764),
    [sym_primitive_type] = ACTIONS(2764),
    [anon_sym_enum] = ACTIONS(2764),
    [anon_sym_struct] = ACTIONS(2764),
    [anon_sym_union] = ACTIONS(2764),
    [anon_sym_if] = ACTIONS(2764),
    [anon_sym_switch] = ACTIONS(2764),
    [anon_sym_case] = ACTIONS(2764),
    [anon_sym_default] = ACTIONS(2764),
    [anon_sym_while] = ACTIONS(2764),
    [anon_sym_do] = ACTIONS(2764),
    [anon_sym_for] = ACTIONS(2764),
    [anon_sym_in] = ACTIONS(2764),
    [anon_sym_return] = ACTIONS(2764),
    [anon_sym_break] = ACTIONS(2764),
    [anon_sym_continue] = ACTIONS(2764),
    [anon_sym_goto] = ACTIONS(2764),
    [anon_sym_DASH_DASH] = ACTIONS(2766),
    [anon_sym_PLUS_PLUS] = ACTIONS(2766),
    [anon_sym_sizeof] = ACTIONS(2764),
    [anon_sym___alignof__] = ACTIONS(2764),
    [anon_sym___alignof] = ACTIONS(2764),
    [anon_sym__alignof] = ACTIONS(2764),
    [anon_sym_alignof] = ACTIONS(2764),
    [anon_sym__Alignof] = ACTIONS(2764),
    [anon_sym_offsetof] = ACTIONS(2764),
    [anon_sym__Generic] = ACTIONS(2764),
    [anon_sym_asm] = ACTIONS(2764),
    [anon_sym___asm__] = ACTIONS(2764),
    [anon_sym___asm] = ACTIONS(2764),
    [sym_number_literal] = ACTIONS(2766),
    [anon_sym_L_SQUOTE] = ACTIONS(2766),
    [anon_sym_u_SQUOTE] = ACTIONS(2766),
    [anon_sym_U_SQUOTE] = ACTIONS(2766),
    [anon_sym_u8_SQUOTE] = ACTIONS(2766),
    [anon_sym_SQUOTE] = ACTIONS(2766),
    [anon_sym_AT] = ACTIONS(2764),
    [anon_sym_DQUOTE] = ACTIONS(2766),
    [anon_sym_L_DQUOTE] = ACTIONS(2766),
    [anon_sym_u_DQUOTE] = ACTIONS(2766),
    [anon_sym_U_DQUOTE] = ACTIONS(2766),
    [anon_sym_u8_DQUOTE] = ACTIONS(2766),
    [sym_true] = ACTIONS(2764),
    [sym_false] = ACTIONS(2764),
    [anon_sym_NULL] = ACTIONS(2764),
    [anon_sym_nullptr] = ACTIONS(2764),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2764),
    [anon_sym___typeof] = ACTIONS(2764),
    [anon_sym_typeof] = ACTIONS(2764),
    [anon_sym_ATimport] = ACTIONS(2766),
    [aux_sym_preproc_undef_token1] = ACTIONS(2764),
    [anon_sym_POUND] = ACTIONS(2764),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2764),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2764),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2764),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_API_AVAILABLE] = ACTIONS(2764),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_API_DEPRECATED] = ACTIONS(2764),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym___deprecated_msg] = ACTIONS(2764),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2764),
    [anon_sym_ATprotocol] = ACTIONS(2766),
    [anon_sym_ATinterface] = ACTIONS(2766),
    [anon_sym_ATimplementation] = ACTIONS(2766),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2766),
    [anon_sym_ATsynthesize] = ACTIONS(2766),
    [anon_sym_ATdynamic] = ACTIONS(2766),
    [anon_sym_ATtry] = ACTIONS(2766),
    [anon_sym___try] = ACTIONS(2764),
    [anon_sym_ATthrow] = ACTIONS(2766),
    [anon_sym_ATselector] = ACTIONS(2766),
    [anon_sym_ATavailable] = ACTIONS(2766),
    [anon_sym___builtin_available] = ACTIONS(2764),
    [anon_sym_va_arg] = ACTIONS(2764),
    [anon_sym_ATencode] = ACTIONS(2766),
    [anon_sym_ATsynchronized] = ACTIONS(2766),
    [anon_sym_BOOL] = ACTIONS(2764),
    [anon_sym_IMP] = ACTIONS(2764),
    [anon_sym_SEL] = ACTIONS(2764),
    [anon_sym_Class] = ACTIONS(2764),
    [anon_sym_id] = ACTIONS(2764),
  },
  [STATE(609)] = {
    [sym_identifier] = ACTIONS(2764),
    [aux_sym_preproc_include_token1] = ACTIONS(2764),
    [aux_sym_preproc_include_token2] = ACTIONS(2764),
    [aux_sym_preproc_def_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token2] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2764),
    [sym_preproc_directive] = ACTIONS(2764),
    [anon_sym_LPAREN2] = ACTIONS(2766),
    [anon_sym_BANG] = ACTIONS(2766),
    [anon_sym_TILDE] = ACTIONS(2766),
    [anon_sym_DASH] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(2764),
    [anon_sym_STAR] = ACTIONS(2766),
    [anon_sym_CARET] = ACTIONS(2766),
    [anon_sym_AMP] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2766),
    [anon_sym___extension__] = ACTIONS(2764),
    [anon_sym_typedef] = ACTIONS(2764),
    [anon_sym_extern] = ACTIONS(2764),
    [anon_sym___attribute__] = ACTIONS(2764),
    [anon_sym___attribute] = ACTIONS(2764),
    [anon_sym_noreturn] = ACTIONS(2764),
    [anon_sym_LBRACK] = ACTIONS(2766),
    [anon_sym___declspec] = ACTIONS(2764),
    [anon_sym___cdecl] = ACTIONS(2764),
    [anon_sym___clrcall] = ACTIONS(2764),
    [anon_sym___stdcall] = ACTIONS(2764),
    [anon_sym___fastcall] = ACTIONS(2764),
    [anon_sym___thiscall] = ACTIONS(2764),
    [anon_sym___vectorcall] = ACTIONS(2764),
    [anon_sym_LBRACE] = ACTIONS(2766),
    [anon_sym_signed] = ACTIONS(2764),
    [anon_sym_unsigned] = ACTIONS(2764),
    [anon_sym_long] = ACTIONS(2764),
    [anon_sym_short] = ACTIONS(2764),
    [anon_sym_static] = ACTIONS(2764),
    [anon_sym_ATautoreleasepool] = ACTIONS(2766),
    [anon_sym_auto] = ACTIONS(2764),
    [anon_sym_register] = ACTIONS(2764),
    [anon_sym_inline] = ACTIONS(2764),
    [anon_sym___inline] = ACTIONS(2764),
    [anon_sym___inline__] = ACTIONS(2764),
    [anon_sym___forceinline] = ACTIONS(2764),
    [anon_sym_thread_local] = ACTIONS(2764),
    [anon_sym___thread] = ACTIONS(2764),
    [anon_sym_CG_EXTERN] = ACTIONS(2764),
    [anon_sym_CG_INLINE] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2764),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2764),
    [anon_sym_IBOutlet] = ACTIONS(2764),
    [anon_sym_IBInspectable] = ACTIONS(2764),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2764),
    [anon_sym_NS_INLINE] = ACTIONS(2764),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2764),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2764),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2764),
    [anon_sym_const] = ACTIONS(2764),
    [anon_sym_constexpr] = ACTIONS(2764),
    [anon_sym_volatile] = ACTIONS(2764),
    [anon_sym_restrict] = ACTIONS(2764),
    [anon_sym___restrict__] = ACTIONS(2764),
    [anon_sym__Atomic] = ACTIONS(2764),
    [anon_sym__Noreturn] = ACTIONS(2764),
    [anon_sym__Nonnull] = ACTIONS(2764),
    [anon_sym_nullable] = ACTIONS(2764),
    [anon_sym__Complex] = ACTIONS(2764),
    [anon_sym__Nullable] = ACTIONS(2764),
    [anon_sym__Nullable_result] = ACTIONS(2764),
    [anon_sym__Null_unspecified] = ACTIONS(2764),
    [anon_sym___autoreleasing] = ACTIONS(2764),
    [anon_sym___block] = ACTIONS(2764),
    [anon_sym___bridge] = ACTIONS(2764),
    [anon_sym___bridge_retained] = ACTIONS(2764),
    [anon_sym___bridge_transfer] = ACTIONS(2764),
    [anon_sym___complex] = ACTIONS(2764),
    [anon_sym___const] = ACTIONS(2764),
    [anon_sym___imag] = ACTIONS(2764),
    [anon_sym___kindof] = ACTIONS(2764),
    [anon_sym___nonnull] = ACTIONS(2764),
    [anon_sym___nullable] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2764),
    [anon_sym___real] = ACTIONS(2764),
    [anon_sym___strong] = ACTIONS(2764),
    [anon_sym___unsafe_unretained] = ACTIONS(2764),
    [anon_sym___unused] = ACTIONS(2764),
    [anon_sym___weak] = ACTIONS(2764),
    [anon_sym_alignas] = ACTIONS(2764),
    [anon_sym__Alignas] = ACTIONS(2764),
    [sym_primitive_type] = ACTIONS(2764),
    [anon_sym_enum] = ACTIONS(2764),
    [anon_sym_struct] = ACTIONS(2764),
    [anon_sym_union] = ACTIONS(2764),
    [anon_sym_if] = ACTIONS(2764),
    [anon_sym_switch] = ACTIONS(2764),
    [anon_sym_case] = ACTIONS(2764),
    [anon_sym_default] = ACTIONS(2764),
    [anon_sym_while] = ACTIONS(2764),
    [anon_sym_do] = ACTIONS(2764),
    [anon_sym_for] = ACTIONS(2764),
    [anon_sym_in] = ACTIONS(2764),
    [anon_sym_return] = ACTIONS(2764),
    [anon_sym_break] = ACTIONS(2764),
    [anon_sym_continue] = ACTIONS(2764),
    [anon_sym_goto] = ACTIONS(2764),
    [anon_sym_DASH_DASH] = ACTIONS(2766),
    [anon_sym_PLUS_PLUS] = ACTIONS(2766),
    [anon_sym_sizeof] = ACTIONS(2764),
    [anon_sym___alignof__] = ACTIONS(2764),
    [anon_sym___alignof] = ACTIONS(2764),
    [anon_sym__alignof] = ACTIONS(2764),
    [anon_sym_alignof] = ACTIONS(2764),
    [anon_sym__Alignof] = ACTIONS(2764),
    [anon_sym_offsetof] = ACTIONS(2764),
    [anon_sym__Generic] = ACTIONS(2764),
    [anon_sym_asm] = ACTIONS(2764),
    [anon_sym___asm__] = ACTIONS(2764),
    [anon_sym___asm] = ACTIONS(2764),
    [sym_number_literal] = ACTIONS(2766),
    [anon_sym_L_SQUOTE] = ACTIONS(2766),
    [anon_sym_u_SQUOTE] = ACTIONS(2766),
    [anon_sym_U_SQUOTE] = ACTIONS(2766),
    [anon_sym_u8_SQUOTE] = ACTIONS(2766),
    [anon_sym_SQUOTE] = ACTIONS(2766),
    [anon_sym_AT] = ACTIONS(2764),
    [anon_sym_DQUOTE] = ACTIONS(2766),
    [anon_sym_L_DQUOTE] = ACTIONS(2766),
    [anon_sym_u_DQUOTE] = ACTIONS(2766),
    [anon_sym_U_DQUOTE] = ACTIONS(2766),
    [anon_sym_u8_DQUOTE] = ACTIONS(2766),
    [sym_true] = ACTIONS(2764),
    [sym_false] = ACTIONS(2764),
    [anon_sym_NULL] = ACTIONS(2764),
    [anon_sym_nullptr] = ACTIONS(2764),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2764),
    [anon_sym___typeof] = ACTIONS(2764),
    [anon_sym_typeof] = ACTIONS(2764),
    [anon_sym_ATimport] = ACTIONS(2766),
    [aux_sym_preproc_undef_token1] = ACTIONS(2764),
    [anon_sym_POUND] = ACTIONS(2764),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2764),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2764),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2764),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_API_AVAILABLE] = ACTIONS(2764),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_API_DEPRECATED] = ACTIONS(2764),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym___deprecated_msg] = ACTIONS(2764),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2764),
    [anon_sym_ATprotocol] = ACTIONS(2766),
    [anon_sym_ATinterface] = ACTIONS(2766),
    [anon_sym_ATimplementation] = ACTIONS(2766),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2766),
    [anon_sym_ATsynthesize] = ACTIONS(2766),
    [anon_sym_ATdynamic] = ACTIONS(2766),
    [anon_sym_ATtry] = ACTIONS(2766),
    [anon_sym___try] = ACTIONS(2764),
    [anon_sym_ATthrow] = ACTIONS(2766),
    [anon_sym_ATselector] = ACTIONS(2766),
    [anon_sym_ATavailable] = ACTIONS(2766),
    [anon_sym___builtin_available] = ACTIONS(2764),
    [anon_sym_va_arg] = ACTIONS(2764),
    [anon_sym_ATencode] = ACTIONS(2766),
    [anon_sym_ATsynchronized] = ACTIONS(2766),
    [anon_sym_BOOL] = ACTIONS(2764),
    [anon_sym_IMP] = ACTIONS(2764),
    [anon_sym_SEL] = ACTIONS(2764),
    [anon_sym_Class] = ACTIONS(2764),
    [anon_sym_id] = ACTIONS(2764),
  },
  [STATE(610)] = {
    [sym_identifier] = ACTIONS(2776),
    [aux_sym_preproc_include_token1] = ACTIONS(2776),
    [aux_sym_preproc_include_token2] = ACTIONS(2776),
    [aux_sym_preproc_def_token1] = ACTIONS(2776),
    [aux_sym_preproc_if_token1] = ACTIONS(2776),
    [aux_sym_preproc_if_token2] = ACTIONS(2776),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2776),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2776),
    [sym_preproc_directive] = ACTIONS(2776),
    [anon_sym_LPAREN2] = ACTIONS(2778),
    [anon_sym_BANG] = ACTIONS(2778),
    [anon_sym_TILDE] = ACTIONS(2778),
    [anon_sym_DASH] = ACTIONS(2776),
    [anon_sym_PLUS] = ACTIONS(2776),
    [anon_sym_STAR] = ACTIONS(2778),
    [anon_sym_CARET] = ACTIONS(2778),
    [anon_sym_AMP] = ACTIONS(2778),
    [anon_sym_SEMI] = ACTIONS(2778),
    [anon_sym___extension__] = ACTIONS(2776),
    [anon_sym_typedef] = ACTIONS(2776),
    [anon_sym_extern] = ACTIONS(2776),
    [anon_sym___attribute__] = ACTIONS(2776),
    [anon_sym___attribute] = ACTIONS(2776),
    [anon_sym_noreturn] = ACTIONS(2776),
    [anon_sym_LBRACK] = ACTIONS(2778),
    [anon_sym___declspec] = ACTIONS(2776),
    [anon_sym___cdecl] = ACTIONS(2776),
    [anon_sym___clrcall] = ACTIONS(2776),
    [anon_sym___stdcall] = ACTIONS(2776),
    [anon_sym___fastcall] = ACTIONS(2776),
    [anon_sym___thiscall] = ACTIONS(2776),
    [anon_sym___vectorcall] = ACTIONS(2776),
    [anon_sym_LBRACE] = ACTIONS(2778),
    [anon_sym_signed] = ACTIONS(2776),
    [anon_sym_unsigned] = ACTIONS(2776),
    [anon_sym_long] = ACTIONS(2776),
    [anon_sym_short] = ACTIONS(2776),
    [anon_sym_static] = ACTIONS(2776),
    [anon_sym_ATautoreleasepool] = ACTIONS(2778),
    [anon_sym_auto] = ACTIONS(2776),
    [anon_sym_register] = ACTIONS(2776),
    [anon_sym_inline] = ACTIONS(2776),
    [anon_sym___inline] = ACTIONS(2776),
    [anon_sym___inline__] = ACTIONS(2776),
    [anon_sym___forceinline] = ACTIONS(2776),
    [anon_sym_thread_local] = ACTIONS(2776),
    [anon_sym___thread] = ACTIONS(2776),
    [anon_sym_CG_EXTERN] = ACTIONS(2776),
    [anon_sym_CG_INLINE] = ACTIONS(2776),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2776),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2776),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2776),
    [anon_sym_IBOutlet] = ACTIONS(2776),
    [anon_sym_IBInspectable] = ACTIONS(2776),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2776),
    [anon_sym_NS_INLINE] = ACTIONS(2776),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2776),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2776),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2776),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2776),
    [anon_sym_const] = ACTIONS(2776),
    [anon_sym_constexpr] = ACTIONS(2776),
    [anon_sym_volatile] = ACTIONS(2776),
    [anon_sym_restrict] = ACTIONS(2776),
    [anon_sym___restrict__] = ACTIONS(2776),
    [anon_sym__Atomic] = ACTIONS(2776),
    [anon_sym__Noreturn] = ACTIONS(2776),
    [anon_sym__Nonnull] = ACTIONS(2776),
    [anon_sym_nullable] = ACTIONS(2776),
    [anon_sym__Complex] = ACTIONS(2776),
    [anon_sym__Nullable] = ACTIONS(2776),
    [anon_sym__Nullable_result] = ACTIONS(2776),
    [anon_sym__Null_unspecified] = ACTIONS(2776),
    [anon_sym___autoreleasing] = ACTIONS(2776),
    [anon_sym___block] = ACTIONS(2776),
    [anon_sym___bridge] = ACTIONS(2776),
    [anon_sym___bridge_retained] = ACTIONS(2776),
    [anon_sym___bridge_transfer] = ACTIONS(2776),
    [anon_sym___complex] = ACTIONS(2776),
    [anon_sym___const] = ACTIONS(2776),
    [anon_sym___imag] = ACTIONS(2776),
    [anon_sym___kindof] = ACTIONS(2776),
    [anon_sym___nonnull] = ACTIONS(2776),
    [anon_sym___nullable] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2776),
    [anon_sym___real] = ACTIONS(2776),
    [anon_sym___strong] = ACTIONS(2776),
    [anon_sym___unsafe_unretained] = ACTIONS(2776),
    [anon_sym___unused] = ACTIONS(2776),
    [anon_sym___weak] = ACTIONS(2776),
    [anon_sym_alignas] = ACTIONS(2776),
    [anon_sym__Alignas] = ACTIONS(2776),
    [sym_primitive_type] = ACTIONS(2776),
    [anon_sym_enum] = ACTIONS(2776),
    [anon_sym_struct] = ACTIONS(2776),
    [anon_sym_union] = ACTIONS(2776),
    [anon_sym_if] = ACTIONS(2776),
    [anon_sym_switch] = ACTIONS(2776),
    [anon_sym_case] = ACTIONS(2776),
    [anon_sym_default] = ACTIONS(2776),
    [anon_sym_while] = ACTIONS(2776),
    [anon_sym_do] = ACTIONS(2776),
    [anon_sym_for] = ACTIONS(2776),
    [anon_sym_in] = ACTIONS(2776),
    [anon_sym_return] = ACTIONS(2776),
    [anon_sym_break] = ACTIONS(2776),
    [anon_sym_continue] = ACTIONS(2776),
    [anon_sym_goto] = ACTIONS(2776),
    [anon_sym_DASH_DASH] = ACTIONS(2778),
    [anon_sym_PLUS_PLUS] = ACTIONS(2778),
    [anon_sym_sizeof] = ACTIONS(2776),
    [anon_sym___alignof__] = ACTIONS(2776),
    [anon_sym___alignof] = ACTIONS(2776),
    [anon_sym__alignof] = ACTIONS(2776),
    [anon_sym_alignof] = ACTIONS(2776),
    [anon_sym__Alignof] = ACTIONS(2776),
    [anon_sym_offsetof] = ACTIONS(2776),
    [anon_sym__Generic] = ACTIONS(2776),
    [anon_sym_asm] = ACTIONS(2776),
    [anon_sym___asm__] = ACTIONS(2776),
    [anon_sym___asm] = ACTIONS(2776),
    [sym_number_literal] = ACTIONS(2778),
    [anon_sym_L_SQUOTE] = ACTIONS(2778),
    [anon_sym_u_SQUOTE] = ACTIONS(2778),
    [anon_sym_U_SQUOTE] = ACTIONS(2778),
    [anon_sym_u8_SQUOTE] = ACTIONS(2778),
    [anon_sym_SQUOTE] = ACTIONS(2778),
    [anon_sym_AT] = ACTIONS(2776),
    [anon_sym_DQUOTE] = ACTIONS(2778),
    [anon_sym_L_DQUOTE] = ACTIONS(2778),
    [anon_sym_u_DQUOTE] = ACTIONS(2778),
    [anon_sym_U_DQUOTE] = ACTIONS(2778),
    [anon_sym_u8_DQUOTE] = ACTIONS(2778),
    [sym_true] = ACTIONS(2776),
    [sym_false] = ACTIONS(2776),
    [anon_sym_NULL] = ACTIONS(2776),
    [anon_sym_nullptr] = ACTIONS(2776),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2776),
    [anon_sym___typeof] = ACTIONS(2776),
    [anon_sym_typeof] = ACTIONS(2776),
    [anon_sym_ATimport] = ACTIONS(2778),
    [aux_sym_preproc_undef_token1] = ACTIONS(2776),
    [anon_sym_POUND] = ACTIONS(2776),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2776),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2776),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_API_AVAILABLE] = ACTIONS(2776),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_API_DEPRECATED] = ACTIONS(2776),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2776),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym___deprecated_msg] = ACTIONS(2776),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2776),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2776),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2776),
    [anon_sym_ATprotocol] = ACTIONS(2778),
    [anon_sym_ATinterface] = ACTIONS(2778),
    [anon_sym_ATimplementation] = ACTIONS(2778),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2778),
    [anon_sym_ATsynthesize] = ACTIONS(2778),
    [anon_sym_ATdynamic] = ACTIONS(2778),
    [anon_sym_ATtry] = ACTIONS(2778),
    [anon_sym___try] = ACTIONS(2776),
    [anon_sym_ATthrow] = ACTIONS(2778),
    [anon_sym_ATselector] = ACTIONS(2778),
    [anon_sym_ATavailable] = ACTIONS(2778),
    [anon_sym___builtin_available] = ACTIONS(2776),
    [anon_sym_va_arg] = ACTIONS(2776),
    [anon_sym_ATencode] = ACTIONS(2778),
    [anon_sym_ATsynchronized] = ACTIONS(2778),
    [anon_sym_BOOL] = ACTIONS(2776),
    [anon_sym_IMP] = ACTIONS(2776),
    [anon_sym_SEL] = ACTIONS(2776),
    [anon_sym_Class] = ACTIONS(2776),
    [anon_sym_id] = ACTIONS(2776),
  },
  [STATE(611)] = {
    [sym_identifier] = ACTIONS(2780),
    [aux_sym_preproc_include_token1] = ACTIONS(2780),
    [aux_sym_preproc_include_token2] = ACTIONS(2780),
    [aux_sym_preproc_def_token1] = ACTIONS(2780),
    [aux_sym_preproc_if_token1] = ACTIONS(2780),
    [aux_sym_preproc_if_token2] = ACTIONS(2780),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2780),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2780),
    [sym_preproc_directive] = ACTIONS(2780),
    [anon_sym_LPAREN2] = ACTIONS(2782),
    [anon_sym_BANG] = ACTIONS(2782),
    [anon_sym_TILDE] = ACTIONS(2782),
    [anon_sym_DASH] = ACTIONS(2780),
    [anon_sym_PLUS] = ACTIONS(2780),
    [anon_sym_STAR] = ACTIONS(2782),
    [anon_sym_CARET] = ACTIONS(2782),
    [anon_sym_AMP] = ACTIONS(2782),
    [anon_sym_SEMI] = ACTIONS(2782),
    [anon_sym___extension__] = ACTIONS(2780),
    [anon_sym_typedef] = ACTIONS(2780),
    [anon_sym_extern] = ACTIONS(2780),
    [anon_sym___attribute__] = ACTIONS(2780),
    [anon_sym___attribute] = ACTIONS(2780),
    [anon_sym_noreturn] = ACTIONS(2780),
    [anon_sym_LBRACK] = ACTIONS(2782),
    [anon_sym___declspec] = ACTIONS(2780),
    [anon_sym___cdecl] = ACTIONS(2780),
    [anon_sym___clrcall] = ACTIONS(2780),
    [anon_sym___stdcall] = ACTIONS(2780),
    [anon_sym___fastcall] = ACTIONS(2780),
    [anon_sym___thiscall] = ACTIONS(2780),
    [anon_sym___vectorcall] = ACTIONS(2780),
    [anon_sym_LBRACE] = ACTIONS(2782),
    [anon_sym_signed] = ACTIONS(2780),
    [anon_sym_unsigned] = ACTIONS(2780),
    [anon_sym_long] = ACTIONS(2780),
    [anon_sym_short] = ACTIONS(2780),
    [anon_sym_static] = ACTIONS(2780),
    [anon_sym_ATautoreleasepool] = ACTIONS(2782),
    [anon_sym_auto] = ACTIONS(2780),
    [anon_sym_register] = ACTIONS(2780),
    [anon_sym_inline] = ACTIONS(2780),
    [anon_sym___inline] = ACTIONS(2780),
    [anon_sym___inline__] = ACTIONS(2780),
    [anon_sym___forceinline] = ACTIONS(2780),
    [anon_sym_thread_local] = ACTIONS(2780),
    [anon_sym___thread] = ACTIONS(2780),
    [anon_sym_CG_EXTERN] = ACTIONS(2780),
    [anon_sym_CG_INLINE] = ACTIONS(2780),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2780),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2780),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2780),
    [anon_sym_IBOutlet] = ACTIONS(2780),
    [anon_sym_IBInspectable] = ACTIONS(2780),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2780),
    [anon_sym_NS_INLINE] = ACTIONS(2780),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2780),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2780),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2780),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2780),
    [anon_sym_const] = ACTIONS(2780),
    [anon_sym_constexpr] = ACTIONS(2780),
    [anon_sym_volatile] = ACTIONS(2780),
    [anon_sym_restrict] = ACTIONS(2780),
    [anon_sym___restrict__] = ACTIONS(2780),
    [anon_sym__Atomic] = ACTIONS(2780),
    [anon_sym__Noreturn] = ACTIONS(2780),
    [anon_sym__Nonnull] = ACTIONS(2780),
    [anon_sym_nullable] = ACTIONS(2780),
    [anon_sym__Complex] = ACTIONS(2780),
    [anon_sym__Nullable] = ACTIONS(2780),
    [anon_sym__Nullable_result] = ACTIONS(2780),
    [anon_sym__Null_unspecified] = ACTIONS(2780),
    [anon_sym___autoreleasing] = ACTIONS(2780),
    [anon_sym___block] = ACTIONS(2780),
    [anon_sym___bridge] = ACTIONS(2780),
    [anon_sym___bridge_retained] = ACTIONS(2780),
    [anon_sym___bridge_transfer] = ACTIONS(2780),
    [anon_sym___complex] = ACTIONS(2780),
    [anon_sym___const] = ACTIONS(2780),
    [anon_sym___imag] = ACTIONS(2780),
    [anon_sym___kindof] = ACTIONS(2780),
    [anon_sym___nonnull] = ACTIONS(2780),
    [anon_sym___nullable] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2780),
    [anon_sym___real] = ACTIONS(2780),
    [anon_sym___strong] = ACTIONS(2780),
    [anon_sym___unsafe_unretained] = ACTIONS(2780),
    [anon_sym___unused] = ACTIONS(2780),
    [anon_sym___weak] = ACTIONS(2780),
    [anon_sym_alignas] = ACTIONS(2780),
    [anon_sym__Alignas] = ACTIONS(2780),
    [sym_primitive_type] = ACTIONS(2780),
    [anon_sym_enum] = ACTIONS(2780),
    [anon_sym_struct] = ACTIONS(2780),
    [anon_sym_union] = ACTIONS(2780),
    [anon_sym_if] = ACTIONS(2780),
    [anon_sym_switch] = ACTIONS(2780),
    [anon_sym_case] = ACTIONS(2780),
    [anon_sym_default] = ACTIONS(2780),
    [anon_sym_while] = ACTIONS(2780),
    [anon_sym_do] = ACTIONS(2780),
    [anon_sym_for] = ACTIONS(2780),
    [anon_sym_in] = ACTIONS(2780),
    [anon_sym_return] = ACTIONS(2780),
    [anon_sym_break] = ACTIONS(2780),
    [anon_sym_continue] = ACTIONS(2780),
    [anon_sym_goto] = ACTIONS(2780),
    [anon_sym_DASH_DASH] = ACTIONS(2782),
    [anon_sym_PLUS_PLUS] = ACTIONS(2782),
    [anon_sym_sizeof] = ACTIONS(2780),
    [anon_sym___alignof__] = ACTIONS(2780),
    [anon_sym___alignof] = ACTIONS(2780),
    [anon_sym__alignof] = ACTIONS(2780),
    [anon_sym_alignof] = ACTIONS(2780),
    [anon_sym__Alignof] = ACTIONS(2780),
    [anon_sym_offsetof] = ACTIONS(2780),
    [anon_sym__Generic] = ACTIONS(2780),
    [anon_sym_asm] = ACTIONS(2780),
    [anon_sym___asm__] = ACTIONS(2780),
    [anon_sym___asm] = ACTIONS(2780),
    [sym_number_literal] = ACTIONS(2782),
    [anon_sym_L_SQUOTE] = ACTIONS(2782),
    [anon_sym_u_SQUOTE] = ACTIONS(2782),
    [anon_sym_U_SQUOTE] = ACTIONS(2782),
    [anon_sym_u8_SQUOTE] = ACTIONS(2782),
    [anon_sym_SQUOTE] = ACTIONS(2782),
    [anon_sym_AT] = ACTIONS(2780),
    [anon_sym_DQUOTE] = ACTIONS(2782),
    [anon_sym_L_DQUOTE] = ACTIONS(2782),
    [anon_sym_u_DQUOTE] = ACTIONS(2782),
    [anon_sym_U_DQUOTE] = ACTIONS(2782),
    [anon_sym_u8_DQUOTE] = ACTIONS(2782),
    [sym_true] = ACTIONS(2780),
    [sym_false] = ACTIONS(2780),
    [anon_sym_NULL] = ACTIONS(2780),
    [anon_sym_nullptr] = ACTIONS(2780),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2780),
    [anon_sym___typeof] = ACTIONS(2780),
    [anon_sym_typeof] = ACTIONS(2780),
    [anon_sym_ATimport] = ACTIONS(2782),
    [aux_sym_preproc_undef_token1] = ACTIONS(2780),
    [anon_sym_POUND] = ACTIONS(2780),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2780),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2780),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2780),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2780),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2780),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2780),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2780),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_API_AVAILABLE] = ACTIONS(2780),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_API_DEPRECATED] = ACTIONS(2780),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2780),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym___deprecated_msg] = ACTIONS(2780),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2780),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2780),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2780),
    [anon_sym_ATprotocol] = ACTIONS(2782),
    [anon_sym_ATinterface] = ACTIONS(2782),
    [anon_sym_ATimplementation] = ACTIONS(2782),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2782),
    [anon_sym_ATsynthesize] = ACTIONS(2782),
    [anon_sym_ATdynamic] = ACTIONS(2782),
    [anon_sym_ATtry] = ACTIONS(2782),
    [anon_sym___try] = ACTIONS(2780),
    [anon_sym_ATthrow] = ACTIONS(2782),
    [anon_sym_ATselector] = ACTIONS(2782),
    [anon_sym_ATavailable] = ACTIONS(2782),
    [anon_sym___builtin_available] = ACTIONS(2780),
    [anon_sym_va_arg] = ACTIONS(2780),
    [anon_sym_ATencode] = ACTIONS(2782),
    [anon_sym_ATsynchronized] = ACTIONS(2782),
    [anon_sym_BOOL] = ACTIONS(2780),
    [anon_sym_IMP] = ACTIONS(2780),
    [anon_sym_SEL] = ACTIONS(2780),
    [anon_sym_Class] = ACTIONS(2780),
    [anon_sym_id] = ACTIONS(2780),
  },
  [STATE(612)] = {
    [sym_identifier] = ACTIONS(2784),
    [aux_sym_preproc_include_token1] = ACTIONS(2784),
    [aux_sym_preproc_include_token2] = ACTIONS(2784),
    [aux_sym_preproc_def_token1] = ACTIONS(2784),
    [aux_sym_preproc_if_token1] = ACTIONS(2784),
    [aux_sym_preproc_if_token2] = ACTIONS(2784),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2784),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2784),
    [sym_preproc_directive] = ACTIONS(2784),
    [anon_sym_LPAREN2] = ACTIONS(2786),
    [anon_sym_BANG] = ACTIONS(2786),
    [anon_sym_TILDE] = ACTIONS(2786),
    [anon_sym_DASH] = ACTIONS(2784),
    [anon_sym_PLUS] = ACTIONS(2784),
    [anon_sym_STAR] = ACTIONS(2786),
    [anon_sym_CARET] = ACTIONS(2786),
    [anon_sym_AMP] = ACTIONS(2786),
    [anon_sym_SEMI] = ACTIONS(2786),
    [anon_sym___extension__] = ACTIONS(2784),
    [anon_sym_typedef] = ACTIONS(2784),
    [anon_sym_extern] = ACTIONS(2784),
    [anon_sym___attribute__] = ACTIONS(2784),
    [anon_sym___attribute] = ACTIONS(2784),
    [anon_sym_noreturn] = ACTIONS(2784),
    [anon_sym_LBRACK] = ACTIONS(2786),
    [anon_sym___declspec] = ACTIONS(2784),
    [anon_sym___cdecl] = ACTIONS(2784),
    [anon_sym___clrcall] = ACTIONS(2784),
    [anon_sym___stdcall] = ACTIONS(2784),
    [anon_sym___fastcall] = ACTIONS(2784),
    [anon_sym___thiscall] = ACTIONS(2784),
    [anon_sym___vectorcall] = ACTIONS(2784),
    [anon_sym_LBRACE] = ACTIONS(2786),
    [anon_sym_signed] = ACTIONS(2784),
    [anon_sym_unsigned] = ACTIONS(2784),
    [anon_sym_long] = ACTIONS(2784),
    [anon_sym_short] = ACTIONS(2784),
    [anon_sym_static] = ACTIONS(2784),
    [anon_sym_ATautoreleasepool] = ACTIONS(2786),
    [anon_sym_auto] = ACTIONS(2784),
    [anon_sym_register] = ACTIONS(2784),
    [anon_sym_inline] = ACTIONS(2784),
    [anon_sym___inline] = ACTIONS(2784),
    [anon_sym___inline__] = ACTIONS(2784),
    [anon_sym___forceinline] = ACTIONS(2784),
    [anon_sym_thread_local] = ACTIONS(2784),
    [anon_sym___thread] = ACTIONS(2784),
    [anon_sym_CG_EXTERN] = ACTIONS(2784),
    [anon_sym_CG_INLINE] = ACTIONS(2784),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2784),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2784),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2784),
    [anon_sym_IBOutlet] = ACTIONS(2784),
    [anon_sym_IBInspectable] = ACTIONS(2784),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2784),
    [anon_sym_NS_INLINE] = ACTIONS(2784),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2784),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2784),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2784),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2784),
    [anon_sym_const] = ACTIONS(2784),
    [anon_sym_constexpr] = ACTIONS(2784),
    [anon_sym_volatile] = ACTIONS(2784),
    [anon_sym_restrict] = ACTIONS(2784),
    [anon_sym___restrict__] = ACTIONS(2784),
    [anon_sym__Atomic] = ACTIONS(2784),
    [anon_sym__Noreturn] = ACTIONS(2784),
    [anon_sym__Nonnull] = ACTIONS(2784),
    [anon_sym_nullable] = ACTIONS(2784),
    [anon_sym__Complex] = ACTIONS(2784),
    [anon_sym__Nullable] = ACTIONS(2784),
    [anon_sym__Nullable_result] = ACTIONS(2784),
    [anon_sym__Null_unspecified] = ACTIONS(2784),
    [anon_sym___autoreleasing] = ACTIONS(2784),
    [anon_sym___block] = ACTIONS(2784),
    [anon_sym___bridge] = ACTIONS(2784),
    [anon_sym___bridge_retained] = ACTIONS(2784),
    [anon_sym___bridge_transfer] = ACTIONS(2784),
    [anon_sym___complex] = ACTIONS(2784),
    [anon_sym___const] = ACTIONS(2784),
    [anon_sym___imag] = ACTIONS(2784),
    [anon_sym___kindof] = ACTIONS(2784),
    [anon_sym___nonnull] = ACTIONS(2784),
    [anon_sym___nullable] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2784),
    [anon_sym___real] = ACTIONS(2784),
    [anon_sym___strong] = ACTIONS(2784),
    [anon_sym___unsafe_unretained] = ACTIONS(2784),
    [anon_sym___unused] = ACTIONS(2784),
    [anon_sym___weak] = ACTIONS(2784),
    [anon_sym_alignas] = ACTIONS(2784),
    [anon_sym__Alignas] = ACTIONS(2784),
    [sym_primitive_type] = ACTIONS(2784),
    [anon_sym_enum] = ACTIONS(2784),
    [anon_sym_struct] = ACTIONS(2784),
    [anon_sym_union] = ACTIONS(2784),
    [anon_sym_if] = ACTIONS(2784),
    [anon_sym_switch] = ACTIONS(2784),
    [anon_sym_case] = ACTIONS(2784),
    [anon_sym_default] = ACTIONS(2784),
    [anon_sym_while] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2784),
    [anon_sym_for] = ACTIONS(2784),
    [anon_sym_in] = ACTIONS(2784),
    [anon_sym_return] = ACTIONS(2784),
    [anon_sym_break] = ACTIONS(2784),
    [anon_sym_continue] = ACTIONS(2784),
    [anon_sym_goto] = ACTIONS(2784),
    [anon_sym_DASH_DASH] = ACTIONS(2786),
    [anon_sym_PLUS_PLUS] = ACTIONS(2786),
    [anon_sym_sizeof] = ACTIONS(2784),
    [anon_sym___alignof__] = ACTIONS(2784),
    [anon_sym___alignof] = ACTIONS(2784),
    [anon_sym__alignof] = ACTIONS(2784),
    [anon_sym_alignof] = ACTIONS(2784),
    [anon_sym__Alignof] = ACTIONS(2784),
    [anon_sym_offsetof] = ACTIONS(2784),
    [anon_sym__Generic] = ACTIONS(2784),
    [anon_sym_asm] = ACTIONS(2784),
    [anon_sym___asm__] = ACTIONS(2784),
    [anon_sym___asm] = ACTIONS(2784),
    [sym_number_literal] = ACTIONS(2786),
    [anon_sym_L_SQUOTE] = ACTIONS(2786),
    [anon_sym_u_SQUOTE] = ACTIONS(2786),
    [anon_sym_U_SQUOTE] = ACTIONS(2786),
    [anon_sym_u8_SQUOTE] = ACTIONS(2786),
    [anon_sym_SQUOTE] = ACTIONS(2786),
    [anon_sym_AT] = ACTIONS(2784),
    [anon_sym_DQUOTE] = ACTIONS(2786),
    [anon_sym_L_DQUOTE] = ACTIONS(2786),
    [anon_sym_u_DQUOTE] = ACTIONS(2786),
    [anon_sym_U_DQUOTE] = ACTIONS(2786),
    [anon_sym_u8_DQUOTE] = ACTIONS(2786),
    [sym_true] = ACTIONS(2784),
    [sym_false] = ACTIONS(2784),
    [anon_sym_NULL] = ACTIONS(2784),
    [anon_sym_nullptr] = ACTIONS(2784),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2784),
    [anon_sym___typeof] = ACTIONS(2784),
    [anon_sym_typeof] = ACTIONS(2784),
    [anon_sym_ATimport] = ACTIONS(2786),
    [aux_sym_preproc_undef_token1] = ACTIONS(2784),
    [anon_sym_POUND] = ACTIONS(2784),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2784),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2784),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2784),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2784),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2784),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2784),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2784),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_API_AVAILABLE] = ACTIONS(2784),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_API_DEPRECATED] = ACTIONS(2784),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2784),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym___deprecated_msg] = ACTIONS(2784),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2784),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2784),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2784),
    [anon_sym_ATprotocol] = ACTIONS(2786),
    [anon_sym_ATinterface] = ACTIONS(2786),
    [anon_sym_ATimplementation] = ACTIONS(2786),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2786),
    [anon_sym_ATsynthesize] = ACTIONS(2786),
    [anon_sym_ATdynamic] = ACTIONS(2786),
    [anon_sym_ATtry] = ACTIONS(2786),
    [anon_sym___try] = ACTIONS(2784),
    [anon_sym_ATthrow] = ACTIONS(2786),
    [anon_sym_ATselector] = ACTIONS(2786),
    [anon_sym_ATavailable] = ACTIONS(2786),
    [anon_sym___builtin_available] = ACTIONS(2784),
    [anon_sym_va_arg] = ACTIONS(2784),
    [anon_sym_ATencode] = ACTIONS(2786),
    [anon_sym_ATsynchronized] = ACTIONS(2786),
    [anon_sym_BOOL] = ACTIONS(2784),
    [anon_sym_IMP] = ACTIONS(2784),
    [anon_sym_SEL] = ACTIONS(2784),
    [anon_sym_Class] = ACTIONS(2784),
    [anon_sym_id] = ACTIONS(2784),
  },
  [STATE(613)] = {
    [sym_identifier] = ACTIONS(2788),
    [aux_sym_preproc_include_token1] = ACTIONS(2788),
    [aux_sym_preproc_include_token2] = ACTIONS(2788),
    [aux_sym_preproc_def_token1] = ACTIONS(2788),
    [aux_sym_preproc_if_token1] = ACTIONS(2788),
    [aux_sym_preproc_if_token2] = ACTIONS(2788),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2788),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2788),
    [sym_preproc_directive] = ACTIONS(2788),
    [anon_sym_LPAREN2] = ACTIONS(2790),
    [anon_sym_BANG] = ACTIONS(2790),
    [anon_sym_TILDE] = ACTIONS(2790),
    [anon_sym_DASH] = ACTIONS(2788),
    [anon_sym_PLUS] = ACTIONS(2788),
    [anon_sym_STAR] = ACTIONS(2790),
    [anon_sym_CARET] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(2790),
    [anon_sym_SEMI] = ACTIONS(2790),
    [anon_sym___extension__] = ACTIONS(2788),
    [anon_sym_typedef] = ACTIONS(2788),
    [anon_sym_extern] = ACTIONS(2788),
    [anon_sym___attribute__] = ACTIONS(2788),
    [anon_sym___attribute] = ACTIONS(2788),
    [anon_sym_noreturn] = ACTIONS(2788),
    [anon_sym_LBRACK] = ACTIONS(2790),
    [anon_sym___declspec] = ACTIONS(2788),
    [anon_sym___cdecl] = ACTIONS(2788),
    [anon_sym___clrcall] = ACTIONS(2788),
    [anon_sym___stdcall] = ACTIONS(2788),
    [anon_sym___fastcall] = ACTIONS(2788),
    [anon_sym___thiscall] = ACTIONS(2788),
    [anon_sym___vectorcall] = ACTIONS(2788),
    [anon_sym_LBRACE] = ACTIONS(2790),
    [anon_sym_signed] = ACTIONS(2788),
    [anon_sym_unsigned] = ACTIONS(2788),
    [anon_sym_long] = ACTIONS(2788),
    [anon_sym_short] = ACTIONS(2788),
    [anon_sym_static] = ACTIONS(2788),
    [anon_sym_ATautoreleasepool] = ACTIONS(2790),
    [anon_sym_auto] = ACTIONS(2788),
    [anon_sym_register] = ACTIONS(2788),
    [anon_sym_inline] = ACTIONS(2788),
    [anon_sym___inline] = ACTIONS(2788),
    [anon_sym___inline__] = ACTIONS(2788),
    [anon_sym___forceinline] = ACTIONS(2788),
    [anon_sym_thread_local] = ACTIONS(2788),
    [anon_sym___thread] = ACTIONS(2788),
    [anon_sym_CG_EXTERN] = ACTIONS(2788),
    [anon_sym_CG_INLINE] = ACTIONS(2788),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2788),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2788),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2788),
    [anon_sym_IBOutlet] = ACTIONS(2788),
    [anon_sym_IBInspectable] = ACTIONS(2788),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2788),
    [anon_sym_NS_INLINE] = ACTIONS(2788),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2788),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2788),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2788),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2788),
    [anon_sym_const] = ACTIONS(2788),
    [anon_sym_constexpr] = ACTIONS(2788),
    [anon_sym_volatile] = ACTIONS(2788),
    [anon_sym_restrict] = ACTIONS(2788),
    [anon_sym___restrict__] = ACTIONS(2788),
    [anon_sym__Atomic] = ACTIONS(2788),
    [anon_sym__Noreturn] = ACTIONS(2788),
    [anon_sym__Nonnull] = ACTIONS(2788),
    [anon_sym_nullable] = ACTIONS(2788),
    [anon_sym__Complex] = ACTIONS(2788),
    [anon_sym__Nullable] = ACTIONS(2788),
    [anon_sym__Nullable_result] = ACTIONS(2788),
    [anon_sym__Null_unspecified] = ACTIONS(2788),
    [anon_sym___autoreleasing] = ACTIONS(2788),
    [anon_sym___block] = ACTIONS(2788),
    [anon_sym___bridge] = ACTIONS(2788),
    [anon_sym___bridge_retained] = ACTIONS(2788),
    [anon_sym___bridge_transfer] = ACTIONS(2788),
    [anon_sym___complex] = ACTIONS(2788),
    [anon_sym___const] = ACTIONS(2788),
    [anon_sym___imag] = ACTIONS(2788),
    [anon_sym___kindof] = ACTIONS(2788),
    [anon_sym___nonnull] = ACTIONS(2788),
    [anon_sym___nullable] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2788),
    [anon_sym___real] = ACTIONS(2788),
    [anon_sym___strong] = ACTIONS(2788),
    [anon_sym___unsafe_unretained] = ACTIONS(2788),
    [anon_sym___unused] = ACTIONS(2788),
    [anon_sym___weak] = ACTIONS(2788),
    [anon_sym_alignas] = ACTIONS(2788),
    [anon_sym__Alignas] = ACTIONS(2788),
    [sym_primitive_type] = ACTIONS(2788),
    [anon_sym_enum] = ACTIONS(2788),
    [anon_sym_struct] = ACTIONS(2788),
    [anon_sym_union] = ACTIONS(2788),
    [anon_sym_if] = ACTIONS(2788),
    [anon_sym_switch] = ACTIONS(2788),
    [anon_sym_case] = ACTIONS(2788),
    [anon_sym_default] = ACTIONS(2788),
    [anon_sym_while] = ACTIONS(2788),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_for] = ACTIONS(2788),
    [anon_sym_in] = ACTIONS(2788),
    [anon_sym_return] = ACTIONS(2788),
    [anon_sym_break] = ACTIONS(2788),
    [anon_sym_continue] = ACTIONS(2788),
    [anon_sym_goto] = ACTIONS(2788),
    [anon_sym_DASH_DASH] = ACTIONS(2790),
    [anon_sym_PLUS_PLUS] = ACTIONS(2790),
    [anon_sym_sizeof] = ACTIONS(2788),
    [anon_sym___alignof__] = ACTIONS(2788),
    [anon_sym___alignof] = ACTIONS(2788),
    [anon_sym__alignof] = ACTIONS(2788),
    [anon_sym_alignof] = ACTIONS(2788),
    [anon_sym__Alignof] = ACTIONS(2788),
    [anon_sym_offsetof] = ACTIONS(2788),
    [anon_sym__Generic] = ACTIONS(2788),
    [anon_sym_asm] = ACTIONS(2788),
    [anon_sym___asm__] = ACTIONS(2788),
    [anon_sym___asm] = ACTIONS(2788),
    [sym_number_literal] = ACTIONS(2790),
    [anon_sym_L_SQUOTE] = ACTIONS(2790),
    [anon_sym_u_SQUOTE] = ACTIONS(2790),
    [anon_sym_U_SQUOTE] = ACTIONS(2790),
    [anon_sym_u8_SQUOTE] = ACTIONS(2790),
    [anon_sym_SQUOTE] = ACTIONS(2790),
    [anon_sym_AT] = ACTIONS(2788),
    [anon_sym_DQUOTE] = ACTIONS(2790),
    [anon_sym_L_DQUOTE] = ACTIONS(2790),
    [anon_sym_u_DQUOTE] = ACTIONS(2790),
    [anon_sym_U_DQUOTE] = ACTIONS(2790),
    [anon_sym_u8_DQUOTE] = ACTIONS(2790),
    [sym_true] = ACTIONS(2788),
    [sym_false] = ACTIONS(2788),
    [anon_sym_NULL] = ACTIONS(2788),
    [anon_sym_nullptr] = ACTIONS(2788),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2788),
    [anon_sym___typeof] = ACTIONS(2788),
    [anon_sym_typeof] = ACTIONS(2788),
    [anon_sym_ATimport] = ACTIONS(2790),
    [aux_sym_preproc_undef_token1] = ACTIONS(2788),
    [anon_sym_POUND] = ACTIONS(2788),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2788),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2788),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2788),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2788),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2788),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2788),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2788),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_API_AVAILABLE] = ACTIONS(2788),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_API_DEPRECATED] = ACTIONS(2788),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2788),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym___deprecated_msg] = ACTIONS(2788),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2788),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2788),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2788),
    [anon_sym_ATprotocol] = ACTIONS(2790),
    [anon_sym_ATinterface] = ACTIONS(2790),
    [anon_sym_ATimplementation] = ACTIONS(2790),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2790),
    [anon_sym_ATsynthesize] = ACTIONS(2790),
    [anon_sym_ATdynamic] = ACTIONS(2790),
    [anon_sym_ATtry] = ACTIONS(2790),
    [anon_sym___try] = ACTIONS(2788),
    [anon_sym_ATthrow] = ACTIONS(2790),
    [anon_sym_ATselector] = ACTIONS(2790),
    [anon_sym_ATavailable] = ACTIONS(2790),
    [anon_sym___builtin_available] = ACTIONS(2788),
    [anon_sym_va_arg] = ACTIONS(2788),
    [anon_sym_ATencode] = ACTIONS(2790),
    [anon_sym_ATsynchronized] = ACTIONS(2790),
    [anon_sym_BOOL] = ACTIONS(2788),
    [anon_sym_IMP] = ACTIONS(2788),
    [anon_sym_SEL] = ACTIONS(2788),
    [anon_sym_Class] = ACTIONS(2788),
    [anon_sym_id] = ACTIONS(2788),
  },
  [STATE(614)] = {
    [sym_identifier] = ACTIONS(2792),
    [aux_sym_preproc_include_token1] = ACTIONS(2792),
    [aux_sym_preproc_include_token2] = ACTIONS(2792),
    [aux_sym_preproc_def_token1] = ACTIONS(2792),
    [aux_sym_preproc_if_token1] = ACTIONS(2792),
    [aux_sym_preproc_if_token2] = ACTIONS(2792),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2792),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2792),
    [sym_preproc_directive] = ACTIONS(2792),
    [anon_sym_LPAREN2] = ACTIONS(2794),
    [anon_sym_BANG] = ACTIONS(2794),
    [anon_sym_TILDE] = ACTIONS(2794),
    [anon_sym_DASH] = ACTIONS(2792),
    [anon_sym_PLUS] = ACTIONS(2792),
    [anon_sym_STAR] = ACTIONS(2794),
    [anon_sym_CARET] = ACTIONS(2794),
    [anon_sym_AMP] = ACTIONS(2794),
    [anon_sym_SEMI] = ACTIONS(2794),
    [anon_sym___extension__] = ACTIONS(2792),
    [anon_sym_typedef] = ACTIONS(2792),
    [anon_sym_extern] = ACTIONS(2792),
    [anon_sym___attribute__] = ACTIONS(2792),
    [anon_sym___attribute] = ACTIONS(2792),
    [anon_sym_noreturn] = ACTIONS(2792),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym___declspec] = ACTIONS(2792),
    [anon_sym___cdecl] = ACTIONS(2792),
    [anon_sym___clrcall] = ACTIONS(2792),
    [anon_sym___stdcall] = ACTIONS(2792),
    [anon_sym___fastcall] = ACTIONS(2792),
    [anon_sym___thiscall] = ACTIONS(2792),
    [anon_sym___vectorcall] = ACTIONS(2792),
    [anon_sym_LBRACE] = ACTIONS(2794),
    [anon_sym_signed] = ACTIONS(2792),
    [anon_sym_unsigned] = ACTIONS(2792),
    [anon_sym_long] = ACTIONS(2792),
    [anon_sym_short] = ACTIONS(2792),
    [anon_sym_static] = ACTIONS(2792),
    [anon_sym_ATautoreleasepool] = ACTIONS(2794),
    [anon_sym_auto] = ACTIONS(2792),
    [anon_sym_register] = ACTIONS(2792),
    [anon_sym_inline] = ACTIONS(2792),
    [anon_sym___inline] = ACTIONS(2792),
    [anon_sym___inline__] = ACTIONS(2792),
    [anon_sym___forceinline] = ACTIONS(2792),
    [anon_sym_thread_local] = ACTIONS(2792),
    [anon_sym___thread] = ACTIONS(2792),
    [anon_sym_CG_EXTERN] = ACTIONS(2792),
    [anon_sym_CG_INLINE] = ACTIONS(2792),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2792),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2792),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2792),
    [anon_sym_IBOutlet] = ACTIONS(2792),
    [anon_sym_IBInspectable] = ACTIONS(2792),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2792),
    [anon_sym_NS_INLINE] = ACTIONS(2792),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2792),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2792),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2792),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2792),
    [anon_sym_const] = ACTIONS(2792),
    [anon_sym_constexpr] = ACTIONS(2792),
    [anon_sym_volatile] = ACTIONS(2792),
    [anon_sym_restrict] = ACTIONS(2792),
    [anon_sym___restrict__] = ACTIONS(2792),
    [anon_sym__Atomic] = ACTIONS(2792),
    [anon_sym__Noreturn] = ACTIONS(2792),
    [anon_sym__Nonnull] = ACTIONS(2792),
    [anon_sym_nullable] = ACTIONS(2792),
    [anon_sym__Complex] = ACTIONS(2792),
    [anon_sym__Nullable] = ACTIONS(2792),
    [anon_sym__Nullable_result] = ACTIONS(2792),
    [anon_sym__Null_unspecified] = ACTIONS(2792),
    [anon_sym___autoreleasing] = ACTIONS(2792),
    [anon_sym___block] = ACTIONS(2792),
    [anon_sym___bridge] = ACTIONS(2792),
    [anon_sym___bridge_retained] = ACTIONS(2792),
    [anon_sym___bridge_transfer] = ACTIONS(2792),
    [anon_sym___complex] = ACTIONS(2792),
    [anon_sym___const] = ACTIONS(2792),
    [anon_sym___imag] = ACTIONS(2792),
    [anon_sym___kindof] = ACTIONS(2792),
    [anon_sym___nonnull] = ACTIONS(2792),
    [anon_sym___nullable] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2792),
    [anon_sym___real] = ACTIONS(2792),
    [anon_sym___strong] = ACTIONS(2792),
    [anon_sym___unsafe_unretained] = ACTIONS(2792),
    [anon_sym___unused] = ACTIONS(2792),
    [anon_sym___weak] = ACTIONS(2792),
    [anon_sym_alignas] = ACTIONS(2792),
    [anon_sym__Alignas] = ACTIONS(2792),
    [sym_primitive_type] = ACTIONS(2792),
    [anon_sym_enum] = ACTIONS(2792),
    [anon_sym_struct] = ACTIONS(2792),
    [anon_sym_union] = ACTIONS(2792),
    [anon_sym_if] = ACTIONS(2792),
    [anon_sym_switch] = ACTIONS(2792),
    [anon_sym_case] = ACTIONS(2792),
    [anon_sym_default] = ACTIONS(2792),
    [anon_sym_while] = ACTIONS(2792),
    [anon_sym_do] = ACTIONS(2792),
    [anon_sym_for] = ACTIONS(2792),
    [anon_sym_in] = ACTIONS(2792),
    [anon_sym_return] = ACTIONS(2792),
    [anon_sym_break] = ACTIONS(2792),
    [anon_sym_continue] = ACTIONS(2792),
    [anon_sym_goto] = ACTIONS(2792),
    [anon_sym_DASH_DASH] = ACTIONS(2794),
    [anon_sym_PLUS_PLUS] = ACTIONS(2794),
    [anon_sym_sizeof] = ACTIONS(2792),
    [anon_sym___alignof__] = ACTIONS(2792),
    [anon_sym___alignof] = ACTIONS(2792),
    [anon_sym__alignof] = ACTIONS(2792),
    [anon_sym_alignof] = ACTIONS(2792),
    [anon_sym__Alignof] = ACTIONS(2792),
    [anon_sym_offsetof] = ACTIONS(2792),
    [anon_sym__Generic] = ACTIONS(2792),
    [anon_sym_asm] = ACTIONS(2792),
    [anon_sym___asm__] = ACTIONS(2792),
    [anon_sym___asm] = ACTIONS(2792),
    [sym_number_literal] = ACTIONS(2794),
    [anon_sym_L_SQUOTE] = ACTIONS(2794),
    [anon_sym_u_SQUOTE] = ACTIONS(2794),
    [anon_sym_U_SQUOTE] = ACTIONS(2794),
    [anon_sym_u8_SQUOTE] = ACTIONS(2794),
    [anon_sym_SQUOTE] = ACTIONS(2794),
    [anon_sym_AT] = ACTIONS(2792),
    [anon_sym_DQUOTE] = ACTIONS(2794),
    [anon_sym_L_DQUOTE] = ACTIONS(2794),
    [anon_sym_u_DQUOTE] = ACTIONS(2794),
    [anon_sym_U_DQUOTE] = ACTIONS(2794),
    [anon_sym_u8_DQUOTE] = ACTIONS(2794),
    [sym_true] = ACTIONS(2792),
    [sym_false] = ACTIONS(2792),
    [anon_sym_NULL] = ACTIONS(2792),
    [anon_sym_nullptr] = ACTIONS(2792),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2792),
    [anon_sym___typeof] = ACTIONS(2792),
    [anon_sym_typeof] = ACTIONS(2792),
    [anon_sym_ATimport] = ACTIONS(2794),
    [aux_sym_preproc_undef_token1] = ACTIONS(2792),
    [anon_sym_POUND] = ACTIONS(2792),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2792),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2792),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2792),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2792),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2792),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2792),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2792),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_API_AVAILABLE] = ACTIONS(2792),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_API_DEPRECATED] = ACTIONS(2792),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2792),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym___deprecated_msg] = ACTIONS(2792),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2792),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2792),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2792),
    [anon_sym_ATprotocol] = ACTIONS(2794),
    [anon_sym_ATinterface] = ACTIONS(2794),
    [anon_sym_ATimplementation] = ACTIONS(2794),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2794),
    [anon_sym_ATsynthesize] = ACTIONS(2794),
    [anon_sym_ATdynamic] = ACTIONS(2794),
    [anon_sym_ATtry] = ACTIONS(2794),
    [anon_sym___try] = ACTIONS(2792),
    [anon_sym_ATthrow] = ACTIONS(2794),
    [anon_sym_ATselector] = ACTIONS(2794),
    [anon_sym_ATavailable] = ACTIONS(2794),
    [anon_sym___builtin_available] = ACTIONS(2792),
    [anon_sym_va_arg] = ACTIONS(2792),
    [anon_sym_ATencode] = ACTIONS(2794),
    [anon_sym_ATsynchronized] = ACTIONS(2794),
    [anon_sym_BOOL] = ACTIONS(2792),
    [anon_sym_IMP] = ACTIONS(2792),
    [anon_sym_SEL] = ACTIONS(2792),
    [anon_sym_Class] = ACTIONS(2792),
    [anon_sym_id] = ACTIONS(2792),
  },
  [STATE(615)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token2] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATsynthesize] = ACTIONS(2798),
    [anon_sym_ATdynamic] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(616)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token2] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATsynthesize] = ACTIONS(2798),
    [anon_sym_ATdynamic] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(617)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token2] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATsynthesize] = ACTIONS(2798),
    [anon_sym_ATdynamic] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(618)] = {
    [sym_identifier] = ACTIONS(2800),
    [aux_sym_preproc_include_token1] = ACTIONS(2800),
    [aux_sym_preproc_include_token2] = ACTIONS(2800),
    [aux_sym_preproc_def_token1] = ACTIONS(2800),
    [aux_sym_preproc_if_token1] = ACTIONS(2800),
    [aux_sym_preproc_if_token2] = ACTIONS(2800),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2800),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2800),
    [sym_preproc_directive] = ACTIONS(2800),
    [anon_sym_LPAREN2] = ACTIONS(2802),
    [anon_sym_BANG] = ACTIONS(2802),
    [anon_sym_TILDE] = ACTIONS(2802),
    [anon_sym_DASH] = ACTIONS(2800),
    [anon_sym_PLUS] = ACTIONS(2800),
    [anon_sym_STAR] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(2802),
    [anon_sym_AMP] = ACTIONS(2802),
    [anon_sym_SEMI] = ACTIONS(2802),
    [anon_sym___extension__] = ACTIONS(2800),
    [anon_sym_typedef] = ACTIONS(2800),
    [anon_sym_extern] = ACTIONS(2800),
    [anon_sym___attribute__] = ACTIONS(2800),
    [anon_sym___attribute] = ACTIONS(2800),
    [anon_sym_noreturn] = ACTIONS(2800),
    [anon_sym_LBRACK] = ACTIONS(2802),
    [anon_sym___declspec] = ACTIONS(2800),
    [anon_sym___cdecl] = ACTIONS(2800),
    [anon_sym___clrcall] = ACTIONS(2800),
    [anon_sym___stdcall] = ACTIONS(2800),
    [anon_sym___fastcall] = ACTIONS(2800),
    [anon_sym___thiscall] = ACTIONS(2800),
    [anon_sym___vectorcall] = ACTIONS(2800),
    [anon_sym_LBRACE] = ACTIONS(2802),
    [anon_sym_signed] = ACTIONS(2800),
    [anon_sym_unsigned] = ACTIONS(2800),
    [anon_sym_long] = ACTIONS(2800),
    [anon_sym_short] = ACTIONS(2800),
    [anon_sym_static] = ACTIONS(2800),
    [anon_sym_ATautoreleasepool] = ACTIONS(2802),
    [anon_sym_auto] = ACTIONS(2800),
    [anon_sym_register] = ACTIONS(2800),
    [anon_sym_inline] = ACTIONS(2800),
    [anon_sym___inline] = ACTIONS(2800),
    [anon_sym___inline__] = ACTIONS(2800),
    [anon_sym___forceinline] = ACTIONS(2800),
    [anon_sym_thread_local] = ACTIONS(2800),
    [anon_sym___thread] = ACTIONS(2800),
    [anon_sym_CG_EXTERN] = ACTIONS(2800),
    [anon_sym_CG_INLINE] = ACTIONS(2800),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2800),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2800),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2800),
    [anon_sym_IBOutlet] = ACTIONS(2800),
    [anon_sym_IBInspectable] = ACTIONS(2800),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2800),
    [anon_sym_NS_INLINE] = ACTIONS(2800),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2800),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2800),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2800),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2800),
    [anon_sym_const] = ACTIONS(2800),
    [anon_sym_constexpr] = ACTIONS(2800),
    [anon_sym_volatile] = ACTIONS(2800),
    [anon_sym_restrict] = ACTIONS(2800),
    [anon_sym___restrict__] = ACTIONS(2800),
    [anon_sym__Atomic] = ACTIONS(2800),
    [anon_sym__Noreturn] = ACTIONS(2800),
    [anon_sym__Nonnull] = ACTIONS(2800),
    [anon_sym_nullable] = ACTIONS(2800),
    [anon_sym__Complex] = ACTIONS(2800),
    [anon_sym__Nullable] = ACTIONS(2800),
    [anon_sym__Nullable_result] = ACTIONS(2800),
    [anon_sym__Null_unspecified] = ACTIONS(2800),
    [anon_sym___autoreleasing] = ACTIONS(2800),
    [anon_sym___block] = ACTIONS(2800),
    [anon_sym___bridge] = ACTIONS(2800),
    [anon_sym___bridge_retained] = ACTIONS(2800),
    [anon_sym___bridge_transfer] = ACTIONS(2800),
    [anon_sym___complex] = ACTIONS(2800),
    [anon_sym___const] = ACTIONS(2800),
    [anon_sym___imag] = ACTIONS(2800),
    [anon_sym___kindof] = ACTIONS(2800),
    [anon_sym___nonnull] = ACTIONS(2800),
    [anon_sym___nullable] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2800),
    [anon_sym___real] = ACTIONS(2800),
    [anon_sym___strong] = ACTIONS(2800),
    [anon_sym___unsafe_unretained] = ACTIONS(2800),
    [anon_sym___unused] = ACTIONS(2800),
    [anon_sym___weak] = ACTIONS(2800),
    [anon_sym_alignas] = ACTIONS(2800),
    [anon_sym__Alignas] = ACTIONS(2800),
    [sym_primitive_type] = ACTIONS(2800),
    [anon_sym_enum] = ACTIONS(2800),
    [anon_sym_struct] = ACTIONS(2800),
    [anon_sym_union] = ACTIONS(2800),
    [anon_sym_if] = ACTIONS(2800),
    [anon_sym_switch] = ACTIONS(2800),
    [anon_sym_case] = ACTIONS(2800),
    [anon_sym_default] = ACTIONS(2800),
    [anon_sym_while] = ACTIONS(2800),
    [anon_sym_do] = ACTIONS(2800),
    [anon_sym_for] = ACTIONS(2800),
    [anon_sym_in] = ACTIONS(2800),
    [anon_sym_return] = ACTIONS(2800),
    [anon_sym_break] = ACTIONS(2800),
    [anon_sym_continue] = ACTIONS(2800),
    [anon_sym_goto] = ACTIONS(2800),
    [anon_sym_DASH_DASH] = ACTIONS(2802),
    [anon_sym_PLUS_PLUS] = ACTIONS(2802),
    [anon_sym_sizeof] = ACTIONS(2800),
    [anon_sym___alignof__] = ACTIONS(2800),
    [anon_sym___alignof] = ACTIONS(2800),
    [anon_sym__alignof] = ACTIONS(2800),
    [anon_sym_alignof] = ACTIONS(2800),
    [anon_sym__Alignof] = ACTIONS(2800),
    [anon_sym_offsetof] = ACTIONS(2800),
    [anon_sym__Generic] = ACTIONS(2800),
    [anon_sym_asm] = ACTIONS(2800),
    [anon_sym___asm__] = ACTIONS(2800),
    [anon_sym___asm] = ACTIONS(2800),
    [sym_number_literal] = ACTIONS(2802),
    [anon_sym_L_SQUOTE] = ACTIONS(2802),
    [anon_sym_u_SQUOTE] = ACTIONS(2802),
    [anon_sym_U_SQUOTE] = ACTIONS(2802),
    [anon_sym_u8_SQUOTE] = ACTIONS(2802),
    [anon_sym_SQUOTE] = ACTIONS(2802),
    [anon_sym_AT] = ACTIONS(2800),
    [anon_sym_DQUOTE] = ACTIONS(2802),
    [anon_sym_L_DQUOTE] = ACTIONS(2802),
    [anon_sym_u_DQUOTE] = ACTIONS(2802),
    [anon_sym_U_DQUOTE] = ACTIONS(2802),
    [anon_sym_u8_DQUOTE] = ACTIONS(2802),
    [sym_true] = ACTIONS(2800),
    [sym_false] = ACTIONS(2800),
    [anon_sym_NULL] = ACTIONS(2800),
    [anon_sym_nullptr] = ACTIONS(2800),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2800),
    [anon_sym___typeof] = ACTIONS(2800),
    [anon_sym_typeof] = ACTIONS(2800),
    [anon_sym_ATimport] = ACTIONS(2802),
    [aux_sym_preproc_undef_token1] = ACTIONS(2800),
    [anon_sym_POUND] = ACTIONS(2800),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2800),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2800),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2800),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2800),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2800),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2800),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2800),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_API_AVAILABLE] = ACTIONS(2800),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_API_DEPRECATED] = ACTIONS(2800),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2800),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym___deprecated_msg] = ACTIONS(2800),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2800),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2800),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2800),
    [anon_sym_ATprotocol] = ACTIONS(2802),
    [anon_sym_ATinterface] = ACTIONS(2802),
    [anon_sym_ATimplementation] = ACTIONS(2802),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2802),
    [anon_sym_ATsynthesize] = ACTIONS(2802),
    [anon_sym_ATdynamic] = ACTIONS(2802),
    [anon_sym_ATtry] = ACTIONS(2802),
    [anon_sym___try] = ACTIONS(2800),
    [anon_sym_ATthrow] = ACTIONS(2802),
    [anon_sym_ATselector] = ACTIONS(2802),
    [anon_sym_ATavailable] = ACTIONS(2802),
    [anon_sym___builtin_available] = ACTIONS(2800),
    [anon_sym_va_arg] = ACTIONS(2800),
    [anon_sym_ATencode] = ACTIONS(2802),
    [anon_sym_ATsynchronized] = ACTIONS(2802),
    [anon_sym_BOOL] = ACTIONS(2800),
    [anon_sym_IMP] = ACTIONS(2800),
    [anon_sym_SEL] = ACTIONS(2800),
    [anon_sym_Class] = ACTIONS(2800),
    [anon_sym_id] = ACTIONS(2800),
  },
  [STATE(619)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token2] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATsynthesize] = ACTIONS(2806),
    [anon_sym_ATdynamic] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(620)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token2] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATsynthesize] = ACTIONS(2806),
    [anon_sym_ATdynamic] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(621)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token2] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATsynthesize] = ACTIONS(2806),
    [anon_sym_ATdynamic] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(622)] = {
    [sym_identifier] = ACTIONS(2808),
    [aux_sym_preproc_include_token1] = ACTIONS(2808),
    [aux_sym_preproc_include_token2] = ACTIONS(2808),
    [aux_sym_preproc_def_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token2] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2808),
    [sym_preproc_directive] = ACTIONS(2808),
    [anon_sym_LPAREN2] = ACTIONS(2810),
    [anon_sym_BANG] = ACTIONS(2810),
    [anon_sym_TILDE] = ACTIONS(2810),
    [anon_sym_DASH] = ACTIONS(2808),
    [anon_sym_PLUS] = ACTIONS(2808),
    [anon_sym_STAR] = ACTIONS(2810),
    [anon_sym_CARET] = ACTIONS(2810),
    [anon_sym_AMP] = ACTIONS(2810),
    [anon_sym_SEMI] = ACTIONS(2810),
    [anon_sym___extension__] = ACTIONS(2808),
    [anon_sym_typedef] = ACTIONS(2808),
    [anon_sym_extern] = ACTIONS(2808),
    [anon_sym___attribute__] = ACTIONS(2808),
    [anon_sym___attribute] = ACTIONS(2808),
    [anon_sym_noreturn] = ACTIONS(2808),
    [anon_sym_LBRACK] = ACTIONS(2810),
    [anon_sym___declspec] = ACTIONS(2808),
    [anon_sym___cdecl] = ACTIONS(2808),
    [anon_sym___clrcall] = ACTIONS(2808),
    [anon_sym___stdcall] = ACTIONS(2808),
    [anon_sym___fastcall] = ACTIONS(2808),
    [anon_sym___thiscall] = ACTIONS(2808),
    [anon_sym___vectorcall] = ACTIONS(2808),
    [anon_sym_LBRACE] = ACTIONS(2810),
    [anon_sym_signed] = ACTIONS(2808),
    [anon_sym_unsigned] = ACTIONS(2808),
    [anon_sym_long] = ACTIONS(2808),
    [anon_sym_short] = ACTIONS(2808),
    [anon_sym_static] = ACTIONS(2808),
    [anon_sym_ATautoreleasepool] = ACTIONS(2810),
    [anon_sym_auto] = ACTIONS(2808),
    [anon_sym_register] = ACTIONS(2808),
    [anon_sym_inline] = ACTIONS(2808),
    [anon_sym___inline] = ACTIONS(2808),
    [anon_sym___inline__] = ACTIONS(2808),
    [anon_sym___forceinline] = ACTIONS(2808),
    [anon_sym_thread_local] = ACTIONS(2808),
    [anon_sym___thread] = ACTIONS(2808),
    [anon_sym_CG_EXTERN] = ACTIONS(2808),
    [anon_sym_CG_INLINE] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2808),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2808),
    [anon_sym_IBOutlet] = ACTIONS(2808),
    [anon_sym_IBInspectable] = ACTIONS(2808),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2808),
    [anon_sym_NS_INLINE] = ACTIONS(2808),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2808),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2808),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2808),
    [anon_sym_const] = ACTIONS(2808),
    [anon_sym_constexpr] = ACTIONS(2808),
    [anon_sym_volatile] = ACTIONS(2808),
    [anon_sym_restrict] = ACTIONS(2808),
    [anon_sym___restrict__] = ACTIONS(2808),
    [anon_sym__Atomic] = ACTIONS(2808),
    [anon_sym__Noreturn] = ACTIONS(2808),
    [anon_sym__Nonnull] = ACTIONS(2808),
    [anon_sym_nullable] = ACTIONS(2808),
    [anon_sym__Complex] = ACTIONS(2808),
    [anon_sym__Nullable] = ACTIONS(2808),
    [anon_sym__Nullable_result] = ACTIONS(2808),
    [anon_sym__Null_unspecified] = ACTIONS(2808),
    [anon_sym___autoreleasing] = ACTIONS(2808),
    [anon_sym___block] = ACTIONS(2808),
    [anon_sym___bridge] = ACTIONS(2808),
    [anon_sym___bridge_retained] = ACTIONS(2808),
    [anon_sym___bridge_transfer] = ACTIONS(2808),
    [anon_sym___complex] = ACTIONS(2808),
    [anon_sym___const] = ACTIONS(2808),
    [anon_sym___imag] = ACTIONS(2808),
    [anon_sym___kindof] = ACTIONS(2808),
    [anon_sym___nonnull] = ACTIONS(2808),
    [anon_sym___nullable] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2808),
    [anon_sym___real] = ACTIONS(2808),
    [anon_sym___strong] = ACTIONS(2808),
    [anon_sym___unsafe_unretained] = ACTIONS(2808),
    [anon_sym___unused] = ACTIONS(2808),
    [anon_sym___weak] = ACTIONS(2808),
    [anon_sym_alignas] = ACTIONS(2808),
    [anon_sym__Alignas] = ACTIONS(2808),
    [sym_primitive_type] = ACTIONS(2808),
    [anon_sym_enum] = ACTIONS(2808),
    [anon_sym_struct] = ACTIONS(2808),
    [anon_sym_union] = ACTIONS(2808),
    [anon_sym_if] = ACTIONS(2808),
    [anon_sym_switch] = ACTIONS(2808),
    [anon_sym_case] = ACTIONS(2808),
    [anon_sym_default] = ACTIONS(2808),
    [anon_sym_while] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2808),
    [anon_sym_in] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_break] = ACTIONS(2808),
    [anon_sym_continue] = ACTIONS(2808),
    [anon_sym_goto] = ACTIONS(2808),
    [anon_sym_DASH_DASH] = ACTIONS(2810),
    [anon_sym_PLUS_PLUS] = ACTIONS(2810),
    [anon_sym_sizeof] = ACTIONS(2808),
    [anon_sym___alignof__] = ACTIONS(2808),
    [anon_sym___alignof] = ACTIONS(2808),
    [anon_sym__alignof] = ACTIONS(2808),
    [anon_sym_alignof] = ACTIONS(2808),
    [anon_sym__Alignof] = ACTIONS(2808),
    [anon_sym_offsetof] = ACTIONS(2808),
    [anon_sym__Generic] = ACTIONS(2808),
    [anon_sym_asm] = ACTIONS(2808),
    [anon_sym___asm__] = ACTIONS(2808),
    [anon_sym___asm] = ACTIONS(2808),
    [sym_number_literal] = ACTIONS(2810),
    [anon_sym_L_SQUOTE] = ACTIONS(2810),
    [anon_sym_u_SQUOTE] = ACTIONS(2810),
    [anon_sym_U_SQUOTE] = ACTIONS(2810),
    [anon_sym_u8_SQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2810),
    [anon_sym_AT] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_L_DQUOTE] = ACTIONS(2810),
    [anon_sym_u_DQUOTE] = ACTIONS(2810),
    [anon_sym_U_DQUOTE] = ACTIONS(2810),
    [anon_sym_u8_DQUOTE] = ACTIONS(2810),
    [sym_true] = ACTIONS(2808),
    [sym_false] = ACTIONS(2808),
    [anon_sym_NULL] = ACTIONS(2808),
    [anon_sym_nullptr] = ACTIONS(2808),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2808),
    [anon_sym___typeof] = ACTIONS(2808),
    [anon_sym_typeof] = ACTIONS(2808),
    [anon_sym_ATimport] = ACTIONS(2810),
    [aux_sym_preproc_undef_token1] = ACTIONS(2808),
    [anon_sym_POUND] = ACTIONS(2808),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2808),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2808),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2808),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_API_AVAILABLE] = ACTIONS(2808),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_API_DEPRECATED] = ACTIONS(2808),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym___deprecated_msg] = ACTIONS(2808),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2808),
    [anon_sym_ATprotocol] = ACTIONS(2810),
    [anon_sym_ATinterface] = ACTIONS(2810),
    [anon_sym_ATimplementation] = ACTIONS(2810),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2810),
    [anon_sym_ATsynthesize] = ACTIONS(2810),
    [anon_sym_ATdynamic] = ACTIONS(2810),
    [anon_sym_ATtry] = ACTIONS(2810),
    [anon_sym___try] = ACTIONS(2808),
    [anon_sym_ATthrow] = ACTIONS(2810),
    [anon_sym_ATselector] = ACTIONS(2810),
    [anon_sym_ATavailable] = ACTIONS(2810),
    [anon_sym___builtin_available] = ACTIONS(2808),
    [anon_sym_va_arg] = ACTIONS(2808),
    [anon_sym_ATencode] = ACTIONS(2810),
    [anon_sym_ATsynchronized] = ACTIONS(2810),
    [anon_sym_BOOL] = ACTIONS(2808),
    [anon_sym_IMP] = ACTIONS(2808),
    [anon_sym_SEL] = ACTIONS(2808),
    [anon_sym_Class] = ACTIONS(2808),
    [anon_sym_id] = ACTIONS(2808),
  },
  [STATE(623)] = {
    [sym_identifier] = ACTIONS(2808),
    [aux_sym_preproc_include_token1] = ACTIONS(2808),
    [aux_sym_preproc_include_token2] = ACTIONS(2808),
    [aux_sym_preproc_def_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token2] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2808),
    [sym_preproc_directive] = ACTIONS(2808),
    [anon_sym_LPAREN2] = ACTIONS(2810),
    [anon_sym_BANG] = ACTIONS(2810),
    [anon_sym_TILDE] = ACTIONS(2810),
    [anon_sym_DASH] = ACTIONS(2808),
    [anon_sym_PLUS] = ACTIONS(2808),
    [anon_sym_STAR] = ACTIONS(2810),
    [anon_sym_CARET] = ACTIONS(2810),
    [anon_sym_AMP] = ACTIONS(2810),
    [anon_sym_SEMI] = ACTIONS(2810),
    [anon_sym___extension__] = ACTIONS(2808),
    [anon_sym_typedef] = ACTIONS(2808),
    [anon_sym_extern] = ACTIONS(2808),
    [anon_sym___attribute__] = ACTIONS(2808),
    [anon_sym___attribute] = ACTIONS(2808),
    [anon_sym_noreturn] = ACTIONS(2808),
    [anon_sym_LBRACK] = ACTIONS(2810),
    [anon_sym___declspec] = ACTIONS(2808),
    [anon_sym___cdecl] = ACTIONS(2808),
    [anon_sym___clrcall] = ACTIONS(2808),
    [anon_sym___stdcall] = ACTIONS(2808),
    [anon_sym___fastcall] = ACTIONS(2808),
    [anon_sym___thiscall] = ACTIONS(2808),
    [anon_sym___vectorcall] = ACTIONS(2808),
    [anon_sym_LBRACE] = ACTIONS(2810),
    [anon_sym_signed] = ACTIONS(2808),
    [anon_sym_unsigned] = ACTIONS(2808),
    [anon_sym_long] = ACTIONS(2808),
    [anon_sym_short] = ACTIONS(2808),
    [anon_sym_static] = ACTIONS(2808),
    [anon_sym_ATautoreleasepool] = ACTIONS(2810),
    [anon_sym_auto] = ACTIONS(2808),
    [anon_sym_register] = ACTIONS(2808),
    [anon_sym_inline] = ACTIONS(2808),
    [anon_sym___inline] = ACTIONS(2808),
    [anon_sym___inline__] = ACTIONS(2808),
    [anon_sym___forceinline] = ACTIONS(2808),
    [anon_sym_thread_local] = ACTIONS(2808),
    [anon_sym___thread] = ACTIONS(2808),
    [anon_sym_CG_EXTERN] = ACTIONS(2808),
    [anon_sym_CG_INLINE] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2808),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2808),
    [anon_sym_IBOutlet] = ACTIONS(2808),
    [anon_sym_IBInspectable] = ACTIONS(2808),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2808),
    [anon_sym_NS_INLINE] = ACTIONS(2808),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2808),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2808),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2808),
    [anon_sym_const] = ACTIONS(2808),
    [anon_sym_constexpr] = ACTIONS(2808),
    [anon_sym_volatile] = ACTIONS(2808),
    [anon_sym_restrict] = ACTIONS(2808),
    [anon_sym___restrict__] = ACTIONS(2808),
    [anon_sym__Atomic] = ACTIONS(2808),
    [anon_sym__Noreturn] = ACTIONS(2808),
    [anon_sym__Nonnull] = ACTIONS(2808),
    [anon_sym_nullable] = ACTIONS(2808),
    [anon_sym__Complex] = ACTIONS(2808),
    [anon_sym__Nullable] = ACTIONS(2808),
    [anon_sym__Nullable_result] = ACTIONS(2808),
    [anon_sym__Null_unspecified] = ACTIONS(2808),
    [anon_sym___autoreleasing] = ACTIONS(2808),
    [anon_sym___block] = ACTIONS(2808),
    [anon_sym___bridge] = ACTIONS(2808),
    [anon_sym___bridge_retained] = ACTIONS(2808),
    [anon_sym___bridge_transfer] = ACTIONS(2808),
    [anon_sym___complex] = ACTIONS(2808),
    [anon_sym___const] = ACTIONS(2808),
    [anon_sym___imag] = ACTIONS(2808),
    [anon_sym___kindof] = ACTIONS(2808),
    [anon_sym___nonnull] = ACTIONS(2808),
    [anon_sym___nullable] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2808),
    [anon_sym___real] = ACTIONS(2808),
    [anon_sym___strong] = ACTIONS(2808),
    [anon_sym___unsafe_unretained] = ACTIONS(2808),
    [anon_sym___unused] = ACTIONS(2808),
    [anon_sym___weak] = ACTIONS(2808),
    [anon_sym_alignas] = ACTIONS(2808),
    [anon_sym__Alignas] = ACTIONS(2808),
    [sym_primitive_type] = ACTIONS(2808),
    [anon_sym_enum] = ACTIONS(2808),
    [anon_sym_struct] = ACTIONS(2808),
    [anon_sym_union] = ACTIONS(2808),
    [anon_sym_if] = ACTIONS(2808),
    [anon_sym_switch] = ACTIONS(2808),
    [anon_sym_case] = ACTIONS(2808),
    [anon_sym_default] = ACTIONS(2808),
    [anon_sym_while] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2808),
    [anon_sym_in] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_break] = ACTIONS(2808),
    [anon_sym_continue] = ACTIONS(2808),
    [anon_sym_goto] = ACTIONS(2808),
    [anon_sym_DASH_DASH] = ACTIONS(2810),
    [anon_sym_PLUS_PLUS] = ACTIONS(2810),
    [anon_sym_sizeof] = ACTIONS(2808),
    [anon_sym___alignof__] = ACTIONS(2808),
    [anon_sym___alignof] = ACTIONS(2808),
    [anon_sym__alignof] = ACTIONS(2808),
    [anon_sym_alignof] = ACTIONS(2808),
    [anon_sym__Alignof] = ACTIONS(2808),
    [anon_sym_offsetof] = ACTIONS(2808),
    [anon_sym__Generic] = ACTIONS(2808),
    [anon_sym_asm] = ACTIONS(2808),
    [anon_sym___asm__] = ACTIONS(2808),
    [anon_sym___asm] = ACTIONS(2808),
    [sym_number_literal] = ACTIONS(2810),
    [anon_sym_L_SQUOTE] = ACTIONS(2810),
    [anon_sym_u_SQUOTE] = ACTIONS(2810),
    [anon_sym_U_SQUOTE] = ACTIONS(2810),
    [anon_sym_u8_SQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2810),
    [anon_sym_AT] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_L_DQUOTE] = ACTIONS(2810),
    [anon_sym_u_DQUOTE] = ACTIONS(2810),
    [anon_sym_U_DQUOTE] = ACTIONS(2810),
    [anon_sym_u8_DQUOTE] = ACTIONS(2810),
    [sym_true] = ACTIONS(2808),
    [sym_false] = ACTIONS(2808),
    [anon_sym_NULL] = ACTIONS(2808),
    [anon_sym_nullptr] = ACTIONS(2808),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2808),
    [anon_sym___typeof] = ACTIONS(2808),
    [anon_sym_typeof] = ACTIONS(2808),
    [anon_sym_ATimport] = ACTIONS(2810),
    [aux_sym_preproc_undef_token1] = ACTIONS(2808),
    [anon_sym_POUND] = ACTIONS(2808),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2808),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2808),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2808),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_API_AVAILABLE] = ACTIONS(2808),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_API_DEPRECATED] = ACTIONS(2808),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym___deprecated_msg] = ACTIONS(2808),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2808),
    [anon_sym_ATprotocol] = ACTIONS(2810),
    [anon_sym_ATinterface] = ACTIONS(2810),
    [anon_sym_ATimplementation] = ACTIONS(2810),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2810),
    [anon_sym_ATsynthesize] = ACTIONS(2810),
    [anon_sym_ATdynamic] = ACTIONS(2810),
    [anon_sym_ATtry] = ACTIONS(2810),
    [anon_sym___try] = ACTIONS(2808),
    [anon_sym_ATthrow] = ACTIONS(2810),
    [anon_sym_ATselector] = ACTIONS(2810),
    [anon_sym_ATavailable] = ACTIONS(2810),
    [anon_sym___builtin_available] = ACTIONS(2808),
    [anon_sym_va_arg] = ACTIONS(2808),
    [anon_sym_ATencode] = ACTIONS(2810),
    [anon_sym_ATsynchronized] = ACTIONS(2810),
    [anon_sym_BOOL] = ACTIONS(2808),
    [anon_sym_IMP] = ACTIONS(2808),
    [anon_sym_SEL] = ACTIONS(2808),
    [anon_sym_Class] = ACTIONS(2808),
    [anon_sym_id] = ACTIONS(2808),
  },
  [STATE(624)] = {
    [sym_identifier] = ACTIONS(2812),
    [aux_sym_preproc_include_token1] = ACTIONS(2812),
    [aux_sym_preproc_include_token2] = ACTIONS(2812),
    [aux_sym_preproc_def_token1] = ACTIONS(2812),
    [aux_sym_preproc_if_token1] = ACTIONS(2812),
    [aux_sym_preproc_if_token2] = ACTIONS(2812),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2812),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2812),
    [sym_preproc_directive] = ACTIONS(2812),
    [anon_sym_LPAREN2] = ACTIONS(2814),
    [anon_sym_BANG] = ACTIONS(2814),
    [anon_sym_TILDE] = ACTIONS(2814),
    [anon_sym_DASH] = ACTIONS(2812),
    [anon_sym_PLUS] = ACTIONS(2812),
    [anon_sym_STAR] = ACTIONS(2814),
    [anon_sym_CARET] = ACTIONS(2814),
    [anon_sym_AMP] = ACTIONS(2814),
    [anon_sym_SEMI] = ACTIONS(2814),
    [anon_sym___extension__] = ACTIONS(2812),
    [anon_sym_typedef] = ACTIONS(2812),
    [anon_sym_extern] = ACTIONS(2812),
    [anon_sym___attribute__] = ACTIONS(2812),
    [anon_sym___attribute] = ACTIONS(2812),
    [anon_sym_noreturn] = ACTIONS(2812),
    [anon_sym_LBRACK] = ACTIONS(2814),
    [anon_sym___declspec] = ACTIONS(2812),
    [anon_sym___cdecl] = ACTIONS(2812),
    [anon_sym___clrcall] = ACTIONS(2812),
    [anon_sym___stdcall] = ACTIONS(2812),
    [anon_sym___fastcall] = ACTIONS(2812),
    [anon_sym___thiscall] = ACTIONS(2812),
    [anon_sym___vectorcall] = ACTIONS(2812),
    [anon_sym_LBRACE] = ACTIONS(2814),
    [anon_sym_signed] = ACTIONS(2812),
    [anon_sym_unsigned] = ACTIONS(2812),
    [anon_sym_long] = ACTIONS(2812),
    [anon_sym_short] = ACTIONS(2812),
    [anon_sym_static] = ACTIONS(2812),
    [anon_sym_ATautoreleasepool] = ACTIONS(2814),
    [anon_sym_auto] = ACTIONS(2812),
    [anon_sym_register] = ACTIONS(2812),
    [anon_sym_inline] = ACTIONS(2812),
    [anon_sym___inline] = ACTIONS(2812),
    [anon_sym___inline__] = ACTIONS(2812),
    [anon_sym___forceinline] = ACTIONS(2812),
    [anon_sym_thread_local] = ACTIONS(2812),
    [anon_sym___thread] = ACTIONS(2812),
    [anon_sym_CG_EXTERN] = ACTIONS(2812),
    [anon_sym_CG_INLINE] = ACTIONS(2812),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2812),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2812),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2812),
    [anon_sym_IBOutlet] = ACTIONS(2812),
    [anon_sym_IBInspectable] = ACTIONS(2812),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2812),
    [anon_sym_NS_INLINE] = ACTIONS(2812),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2812),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2812),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2812),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2812),
    [anon_sym_const] = ACTIONS(2812),
    [anon_sym_constexpr] = ACTIONS(2812),
    [anon_sym_volatile] = ACTIONS(2812),
    [anon_sym_restrict] = ACTIONS(2812),
    [anon_sym___restrict__] = ACTIONS(2812),
    [anon_sym__Atomic] = ACTIONS(2812),
    [anon_sym__Noreturn] = ACTIONS(2812),
    [anon_sym__Nonnull] = ACTIONS(2812),
    [anon_sym_nullable] = ACTIONS(2812),
    [anon_sym__Complex] = ACTIONS(2812),
    [anon_sym__Nullable] = ACTIONS(2812),
    [anon_sym__Nullable_result] = ACTIONS(2812),
    [anon_sym__Null_unspecified] = ACTIONS(2812),
    [anon_sym___autoreleasing] = ACTIONS(2812),
    [anon_sym___block] = ACTIONS(2812),
    [anon_sym___bridge] = ACTIONS(2812),
    [anon_sym___bridge_retained] = ACTIONS(2812),
    [anon_sym___bridge_transfer] = ACTIONS(2812),
    [anon_sym___complex] = ACTIONS(2812),
    [anon_sym___const] = ACTIONS(2812),
    [anon_sym___imag] = ACTIONS(2812),
    [anon_sym___kindof] = ACTIONS(2812),
    [anon_sym___nonnull] = ACTIONS(2812),
    [anon_sym___nullable] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2812),
    [anon_sym___real] = ACTIONS(2812),
    [anon_sym___strong] = ACTIONS(2812),
    [anon_sym___unsafe_unretained] = ACTIONS(2812),
    [anon_sym___unused] = ACTIONS(2812),
    [anon_sym___weak] = ACTIONS(2812),
    [anon_sym_alignas] = ACTIONS(2812),
    [anon_sym__Alignas] = ACTIONS(2812),
    [sym_primitive_type] = ACTIONS(2812),
    [anon_sym_enum] = ACTIONS(2812),
    [anon_sym_struct] = ACTIONS(2812),
    [anon_sym_union] = ACTIONS(2812),
    [anon_sym_if] = ACTIONS(2812),
    [anon_sym_switch] = ACTIONS(2812),
    [anon_sym_case] = ACTIONS(2812),
    [anon_sym_default] = ACTIONS(2812),
    [anon_sym_while] = ACTIONS(2812),
    [anon_sym_do] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2812),
    [anon_sym_in] = ACTIONS(2812),
    [anon_sym_return] = ACTIONS(2812),
    [anon_sym_break] = ACTIONS(2812),
    [anon_sym_continue] = ACTIONS(2812),
    [anon_sym_goto] = ACTIONS(2812),
    [anon_sym_DASH_DASH] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_sizeof] = ACTIONS(2812),
    [anon_sym___alignof__] = ACTIONS(2812),
    [anon_sym___alignof] = ACTIONS(2812),
    [anon_sym__alignof] = ACTIONS(2812),
    [anon_sym_alignof] = ACTIONS(2812),
    [anon_sym__Alignof] = ACTIONS(2812),
    [anon_sym_offsetof] = ACTIONS(2812),
    [anon_sym__Generic] = ACTIONS(2812),
    [anon_sym_asm] = ACTIONS(2812),
    [anon_sym___asm__] = ACTIONS(2812),
    [anon_sym___asm] = ACTIONS(2812),
    [sym_number_literal] = ACTIONS(2814),
    [anon_sym_L_SQUOTE] = ACTIONS(2814),
    [anon_sym_u_SQUOTE] = ACTIONS(2814),
    [anon_sym_U_SQUOTE] = ACTIONS(2814),
    [anon_sym_u8_SQUOTE] = ACTIONS(2814),
    [anon_sym_SQUOTE] = ACTIONS(2814),
    [anon_sym_AT] = ACTIONS(2812),
    [anon_sym_DQUOTE] = ACTIONS(2814),
    [anon_sym_L_DQUOTE] = ACTIONS(2814),
    [anon_sym_u_DQUOTE] = ACTIONS(2814),
    [anon_sym_U_DQUOTE] = ACTIONS(2814),
    [anon_sym_u8_DQUOTE] = ACTIONS(2814),
    [sym_true] = ACTIONS(2812),
    [sym_false] = ACTIONS(2812),
    [anon_sym_NULL] = ACTIONS(2812),
    [anon_sym_nullptr] = ACTIONS(2812),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2812),
    [anon_sym___typeof] = ACTIONS(2812),
    [anon_sym_typeof] = ACTIONS(2812),
    [anon_sym_ATimport] = ACTIONS(2814),
    [aux_sym_preproc_undef_token1] = ACTIONS(2812),
    [anon_sym_POUND] = ACTIONS(2812),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2812),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2812),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2812),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2812),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2812),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2812),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2812),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_API_AVAILABLE] = ACTIONS(2812),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_API_DEPRECATED] = ACTIONS(2812),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2812),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym___deprecated_msg] = ACTIONS(2812),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2812),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2812),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2812),
    [anon_sym_ATprotocol] = ACTIONS(2814),
    [anon_sym_ATinterface] = ACTIONS(2814),
    [anon_sym_ATimplementation] = ACTIONS(2814),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2814),
    [anon_sym_ATsynthesize] = ACTIONS(2814),
    [anon_sym_ATdynamic] = ACTIONS(2814),
    [anon_sym_ATtry] = ACTIONS(2814),
    [anon_sym___try] = ACTIONS(2812),
    [anon_sym_ATthrow] = ACTIONS(2814),
    [anon_sym_ATselector] = ACTIONS(2814),
    [anon_sym_ATavailable] = ACTIONS(2814),
    [anon_sym___builtin_available] = ACTIONS(2812),
    [anon_sym_va_arg] = ACTIONS(2812),
    [anon_sym_ATencode] = ACTIONS(2814),
    [anon_sym_ATsynchronized] = ACTIONS(2814),
    [anon_sym_BOOL] = ACTIONS(2812),
    [anon_sym_IMP] = ACTIONS(2812),
    [anon_sym_SEL] = ACTIONS(2812),
    [anon_sym_Class] = ACTIONS(2812),
    [anon_sym_id] = ACTIONS(2812),
  },
  [STATE(625)] = {
    [sym_identifier] = ACTIONS(2816),
    [aux_sym_preproc_include_token1] = ACTIONS(2816),
    [aux_sym_preproc_include_token2] = ACTIONS(2816),
    [aux_sym_preproc_def_token1] = ACTIONS(2816),
    [aux_sym_preproc_if_token1] = ACTIONS(2816),
    [aux_sym_preproc_if_token2] = ACTIONS(2816),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2816),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2816),
    [sym_preproc_directive] = ACTIONS(2816),
    [anon_sym_LPAREN2] = ACTIONS(2818),
    [anon_sym_BANG] = ACTIONS(2818),
    [anon_sym_TILDE] = ACTIONS(2818),
    [anon_sym_DASH] = ACTIONS(2816),
    [anon_sym_PLUS] = ACTIONS(2816),
    [anon_sym_STAR] = ACTIONS(2818),
    [anon_sym_CARET] = ACTIONS(2818),
    [anon_sym_AMP] = ACTIONS(2818),
    [anon_sym_SEMI] = ACTIONS(2818),
    [anon_sym___extension__] = ACTIONS(2816),
    [anon_sym_typedef] = ACTIONS(2816),
    [anon_sym_extern] = ACTIONS(2816),
    [anon_sym___attribute__] = ACTIONS(2816),
    [anon_sym___attribute] = ACTIONS(2816),
    [anon_sym_noreturn] = ACTIONS(2816),
    [anon_sym_LBRACK] = ACTIONS(2818),
    [anon_sym___declspec] = ACTIONS(2816),
    [anon_sym___cdecl] = ACTIONS(2816),
    [anon_sym___clrcall] = ACTIONS(2816),
    [anon_sym___stdcall] = ACTIONS(2816),
    [anon_sym___fastcall] = ACTIONS(2816),
    [anon_sym___thiscall] = ACTIONS(2816),
    [anon_sym___vectorcall] = ACTIONS(2816),
    [anon_sym_LBRACE] = ACTIONS(2818),
    [anon_sym_signed] = ACTIONS(2816),
    [anon_sym_unsigned] = ACTIONS(2816),
    [anon_sym_long] = ACTIONS(2816),
    [anon_sym_short] = ACTIONS(2816),
    [anon_sym_static] = ACTIONS(2816),
    [anon_sym_ATautoreleasepool] = ACTIONS(2818),
    [anon_sym_auto] = ACTIONS(2816),
    [anon_sym_register] = ACTIONS(2816),
    [anon_sym_inline] = ACTIONS(2816),
    [anon_sym___inline] = ACTIONS(2816),
    [anon_sym___inline__] = ACTIONS(2816),
    [anon_sym___forceinline] = ACTIONS(2816),
    [anon_sym_thread_local] = ACTIONS(2816),
    [anon_sym___thread] = ACTIONS(2816),
    [anon_sym_CG_EXTERN] = ACTIONS(2816),
    [anon_sym_CG_INLINE] = ACTIONS(2816),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2816),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2816),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2816),
    [anon_sym_IBOutlet] = ACTIONS(2816),
    [anon_sym_IBInspectable] = ACTIONS(2816),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2816),
    [anon_sym_NS_INLINE] = ACTIONS(2816),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2816),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2816),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2816),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2816),
    [anon_sym_const] = ACTIONS(2816),
    [anon_sym_constexpr] = ACTIONS(2816),
    [anon_sym_volatile] = ACTIONS(2816),
    [anon_sym_restrict] = ACTIONS(2816),
    [anon_sym___restrict__] = ACTIONS(2816),
    [anon_sym__Atomic] = ACTIONS(2816),
    [anon_sym__Noreturn] = ACTIONS(2816),
    [anon_sym__Nonnull] = ACTIONS(2816),
    [anon_sym_nullable] = ACTIONS(2816),
    [anon_sym__Complex] = ACTIONS(2816),
    [anon_sym__Nullable] = ACTIONS(2816),
    [anon_sym__Nullable_result] = ACTIONS(2816),
    [anon_sym__Null_unspecified] = ACTIONS(2816),
    [anon_sym___autoreleasing] = ACTIONS(2816),
    [anon_sym___block] = ACTIONS(2816),
    [anon_sym___bridge] = ACTIONS(2816),
    [anon_sym___bridge_retained] = ACTIONS(2816),
    [anon_sym___bridge_transfer] = ACTIONS(2816),
    [anon_sym___complex] = ACTIONS(2816),
    [anon_sym___const] = ACTIONS(2816),
    [anon_sym___imag] = ACTIONS(2816),
    [anon_sym___kindof] = ACTIONS(2816),
    [anon_sym___nonnull] = ACTIONS(2816),
    [anon_sym___nullable] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2816),
    [anon_sym___real] = ACTIONS(2816),
    [anon_sym___strong] = ACTIONS(2816),
    [anon_sym___unsafe_unretained] = ACTIONS(2816),
    [anon_sym___unused] = ACTIONS(2816),
    [anon_sym___weak] = ACTIONS(2816),
    [anon_sym_alignas] = ACTIONS(2816),
    [anon_sym__Alignas] = ACTIONS(2816),
    [sym_primitive_type] = ACTIONS(2816),
    [anon_sym_enum] = ACTIONS(2816),
    [anon_sym_struct] = ACTIONS(2816),
    [anon_sym_union] = ACTIONS(2816),
    [anon_sym_if] = ACTIONS(2816),
    [anon_sym_switch] = ACTIONS(2816),
    [anon_sym_case] = ACTIONS(2816),
    [anon_sym_default] = ACTIONS(2816),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_do] = ACTIONS(2816),
    [anon_sym_for] = ACTIONS(2816),
    [anon_sym_in] = ACTIONS(2816),
    [anon_sym_return] = ACTIONS(2816),
    [anon_sym_break] = ACTIONS(2816),
    [anon_sym_continue] = ACTIONS(2816),
    [anon_sym_goto] = ACTIONS(2816),
    [anon_sym_DASH_DASH] = ACTIONS(2818),
    [anon_sym_PLUS_PLUS] = ACTIONS(2818),
    [anon_sym_sizeof] = ACTIONS(2816),
    [anon_sym___alignof__] = ACTIONS(2816),
    [anon_sym___alignof] = ACTIONS(2816),
    [anon_sym__alignof] = ACTIONS(2816),
    [anon_sym_alignof] = ACTIONS(2816),
    [anon_sym__Alignof] = ACTIONS(2816),
    [anon_sym_offsetof] = ACTIONS(2816),
    [anon_sym__Generic] = ACTIONS(2816),
    [anon_sym_asm] = ACTIONS(2816),
    [anon_sym___asm__] = ACTIONS(2816),
    [anon_sym___asm] = ACTIONS(2816),
    [sym_number_literal] = ACTIONS(2818),
    [anon_sym_L_SQUOTE] = ACTIONS(2818),
    [anon_sym_u_SQUOTE] = ACTIONS(2818),
    [anon_sym_U_SQUOTE] = ACTIONS(2818),
    [anon_sym_u8_SQUOTE] = ACTIONS(2818),
    [anon_sym_SQUOTE] = ACTIONS(2818),
    [anon_sym_AT] = ACTIONS(2816),
    [anon_sym_DQUOTE] = ACTIONS(2818),
    [anon_sym_L_DQUOTE] = ACTIONS(2818),
    [anon_sym_u_DQUOTE] = ACTIONS(2818),
    [anon_sym_U_DQUOTE] = ACTIONS(2818),
    [anon_sym_u8_DQUOTE] = ACTIONS(2818),
    [sym_true] = ACTIONS(2816),
    [sym_false] = ACTIONS(2816),
    [anon_sym_NULL] = ACTIONS(2816),
    [anon_sym_nullptr] = ACTIONS(2816),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2816),
    [anon_sym___typeof] = ACTIONS(2816),
    [anon_sym_typeof] = ACTIONS(2816),
    [anon_sym_ATimport] = ACTIONS(2818),
    [aux_sym_preproc_undef_token1] = ACTIONS(2816),
    [anon_sym_POUND] = ACTIONS(2816),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2816),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2816),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2816),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2816),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2816),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2816),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2816),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_API_AVAILABLE] = ACTIONS(2816),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_API_DEPRECATED] = ACTIONS(2816),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2816),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym___deprecated_msg] = ACTIONS(2816),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2816),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2816),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2816),
    [anon_sym_ATprotocol] = ACTIONS(2818),
    [anon_sym_ATinterface] = ACTIONS(2818),
    [anon_sym_ATimplementation] = ACTIONS(2818),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2818),
    [anon_sym_ATsynthesize] = ACTIONS(2818),
    [anon_sym_ATdynamic] = ACTIONS(2818),
    [anon_sym_ATtry] = ACTIONS(2818),
    [anon_sym___try] = ACTIONS(2816),
    [anon_sym_ATthrow] = ACTIONS(2818),
    [anon_sym_ATselector] = ACTIONS(2818),
    [anon_sym_ATavailable] = ACTIONS(2818),
    [anon_sym___builtin_available] = ACTIONS(2816),
    [anon_sym_va_arg] = ACTIONS(2816),
    [anon_sym_ATencode] = ACTIONS(2818),
    [anon_sym_ATsynchronized] = ACTIONS(2818),
    [anon_sym_BOOL] = ACTIONS(2816),
    [anon_sym_IMP] = ACTIONS(2816),
    [anon_sym_SEL] = ACTIONS(2816),
    [anon_sym_Class] = ACTIONS(2816),
    [anon_sym_id] = ACTIONS(2816),
  },
  [STATE(626)] = {
    [sym_identifier] = ACTIONS(2820),
    [aux_sym_preproc_include_token1] = ACTIONS(2820),
    [aux_sym_preproc_include_token2] = ACTIONS(2820),
    [aux_sym_preproc_def_token1] = ACTIONS(2820),
    [aux_sym_preproc_if_token1] = ACTIONS(2820),
    [aux_sym_preproc_if_token2] = ACTIONS(2820),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2820),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2820),
    [sym_preproc_directive] = ACTIONS(2820),
    [anon_sym_LPAREN2] = ACTIONS(2822),
    [anon_sym_BANG] = ACTIONS(2822),
    [anon_sym_TILDE] = ACTIONS(2822),
    [anon_sym_DASH] = ACTIONS(2820),
    [anon_sym_PLUS] = ACTIONS(2820),
    [anon_sym_STAR] = ACTIONS(2822),
    [anon_sym_CARET] = ACTIONS(2822),
    [anon_sym_AMP] = ACTIONS(2822),
    [anon_sym_SEMI] = ACTIONS(2822),
    [anon_sym___extension__] = ACTIONS(2820),
    [anon_sym_typedef] = ACTIONS(2820),
    [anon_sym_extern] = ACTIONS(2820),
    [anon_sym___attribute__] = ACTIONS(2820),
    [anon_sym___attribute] = ACTIONS(2820),
    [anon_sym_noreturn] = ACTIONS(2820),
    [anon_sym_LBRACK] = ACTIONS(2822),
    [anon_sym___declspec] = ACTIONS(2820),
    [anon_sym___cdecl] = ACTIONS(2820),
    [anon_sym___clrcall] = ACTIONS(2820),
    [anon_sym___stdcall] = ACTIONS(2820),
    [anon_sym___fastcall] = ACTIONS(2820),
    [anon_sym___thiscall] = ACTIONS(2820),
    [anon_sym___vectorcall] = ACTIONS(2820),
    [anon_sym_LBRACE] = ACTIONS(2822),
    [anon_sym_signed] = ACTIONS(2820),
    [anon_sym_unsigned] = ACTIONS(2820),
    [anon_sym_long] = ACTIONS(2820),
    [anon_sym_short] = ACTIONS(2820),
    [anon_sym_static] = ACTIONS(2820),
    [anon_sym_ATautoreleasepool] = ACTIONS(2822),
    [anon_sym_auto] = ACTIONS(2820),
    [anon_sym_register] = ACTIONS(2820),
    [anon_sym_inline] = ACTIONS(2820),
    [anon_sym___inline] = ACTIONS(2820),
    [anon_sym___inline__] = ACTIONS(2820),
    [anon_sym___forceinline] = ACTIONS(2820),
    [anon_sym_thread_local] = ACTIONS(2820),
    [anon_sym___thread] = ACTIONS(2820),
    [anon_sym_CG_EXTERN] = ACTIONS(2820),
    [anon_sym_CG_INLINE] = ACTIONS(2820),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2820),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2820),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2820),
    [anon_sym_IBOutlet] = ACTIONS(2820),
    [anon_sym_IBInspectable] = ACTIONS(2820),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2820),
    [anon_sym_NS_INLINE] = ACTIONS(2820),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2820),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2820),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2820),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2820),
    [anon_sym_const] = ACTIONS(2820),
    [anon_sym_constexpr] = ACTIONS(2820),
    [anon_sym_volatile] = ACTIONS(2820),
    [anon_sym_restrict] = ACTIONS(2820),
    [anon_sym___restrict__] = ACTIONS(2820),
    [anon_sym__Atomic] = ACTIONS(2820),
    [anon_sym__Noreturn] = ACTIONS(2820),
    [anon_sym__Nonnull] = ACTIONS(2820),
    [anon_sym_nullable] = ACTIONS(2820),
    [anon_sym__Complex] = ACTIONS(2820),
    [anon_sym__Nullable] = ACTIONS(2820),
    [anon_sym__Nullable_result] = ACTIONS(2820),
    [anon_sym__Null_unspecified] = ACTIONS(2820),
    [anon_sym___autoreleasing] = ACTIONS(2820),
    [anon_sym___block] = ACTIONS(2820),
    [anon_sym___bridge] = ACTIONS(2820),
    [anon_sym___bridge_retained] = ACTIONS(2820),
    [anon_sym___bridge_transfer] = ACTIONS(2820),
    [anon_sym___complex] = ACTIONS(2820),
    [anon_sym___const] = ACTIONS(2820),
    [anon_sym___imag] = ACTIONS(2820),
    [anon_sym___kindof] = ACTIONS(2820),
    [anon_sym___nonnull] = ACTIONS(2820),
    [anon_sym___nullable] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2820),
    [anon_sym___real] = ACTIONS(2820),
    [anon_sym___strong] = ACTIONS(2820),
    [anon_sym___unsafe_unretained] = ACTIONS(2820),
    [anon_sym___unused] = ACTIONS(2820),
    [anon_sym___weak] = ACTIONS(2820),
    [anon_sym_alignas] = ACTIONS(2820),
    [anon_sym__Alignas] = ACTIONS(2820),
    [sym_primitive_type] = ACTIONS(2820),
    [anon_sym_enum] = ACTIONS(2820),
    [anon_sym_struct] = ACTIONS(2820),
    [anon_sym_union] = ACTIONS(2820),
    [anon_sym_if] = ACTIONS(2820),
    [anon_sym_switch] = ACTIONS(2820),
    [anon_sym_case] = ACTIONS(2820),
    [anon_sym_default] = ACTIONS(2820),
    [anon_sym_while] = ACTIONS(2820),
    [anon_sym_do] = ACTIONS(2820),
    [anon_sym_for] = ACTIONS(2820),
    [anon_sym_in] = ACTIONS(2820),
    [anon_sym_return] = ACTIONS(2820),
    [anon_sym_break] = ACTIONS(2820),
    [anon_sym_continue] = ACTIONS(2820),
    [anon_sym_goto] = ACTIONS(2820),
    [anon_sym_DASH_DASH] = ACTIONS(2822),
    [anon_sym_PLUS_PLUS] = ACTIONS(2822),
    [anon_sym_sizeof] = ACTIONS(2820),
    [anon_sym___alignof__] = ACTIONS(2820),
    [anon_sym___alignof] = ACTIONS(2820),
    [anon_sym__alignof] = ACTIONS(2820),
    [anon_sym_alignof] = ACTIONS(2820),
    [anon_sym__Alignof] = ACTIONS(2820),
    [anon_sym_offsetof] = ACTIONS(2820),
    [anon_sym__Generic] = ACTIONS(2820),
    [anon_sym_asm] = ACTIONS(2820),
    [anon_sym___asm__] = ACTIONS(2820),
    [anon_sym___asm] = ACTIONS(2820),
    [sym_number_literal] = ACTIONS(2822),
    [anon_sym_L_SQUOTE] = ACTIONS(2822),
    [anon_sym_u_SQUOTE] = ACTIONS(2822),
    [anon_sym_U_SQUOTE] = ACTIONS(2822),
    [anon_sym_u8_SQUOTE] = ACTIONS(2822),
    [anon_sym_SQUOTE] = ACTIONS(2822),
    [anon_sym_AT] = ACTIONS(2820),
    [anon_sym_DQUOTE] = ACTIONS(2822),
    [anon_sym_L_DQUOTE] = ACTIONS(2822),
    [anon_sym_u_DQUOTE] = ACTIONS(2822),
    [anon_sym_U_DQUOTE] = ACTIONS(2822),
    [anon_sym_u8_DQUOTE] = ACTIONS(2822),
    [sym_true] = ACTIONS(2820),
    [sym_false] = ACTIONS(2820),
    [anon_sym_NULL] = ACTIONS(2820),
    [anon_sym_nullptr] = ACTIONS(2820),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2820),
    [anon_sym___typeof] = ACTIONS(2820),
    [anon_sym_typeof] = ACTIONS(2820),
    [anon_sym_ATimport] = ACTIONS(2822),
    [aux_sym_preproc_undef_token1] = ACTIONS(2820),
    [anon_sym_POUND] = ACTIONS(2820),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2820),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2820),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2820),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2820),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2820),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2820),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2820),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_API_AVAILABLE] = ACTIONS(2820),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_API_DEPRECATED] = ACTIONS(2820),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2820),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym___deprecated_msg] = ACTIONS(2820),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2820),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2820),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2820),
    [anon_sym_ATprotocol] = ACTIONS(2822),
    [anon_sym_ATinterface] = ACTIONS(2822),
    [anon_sym_ATimplementation] = ACTIONS(2822),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2822),
    [anon_sym_ATsynthesize] = ACTIONS(2822),
    [anon_sym_ATdynamic] = ACTIONS(2822),
    [anon_sym_ATtry] = ACTIONS(2822),
    [anon_sym___try] = ACTIONS(2820),
    [anon_sym_ATthrow] = ACTIONS(2822),
    [anon_sym_ATselector] = ACTIONS(2822),
    [anon_sym_ATavailable] = ACTIONS(2822),
    [anon_sym___builtin_available] = ACTIONS(2820),
    [anon_sym_va_arg] = ACTIONS(2820),
    [anon_sym_ATencode] = ACTIONS(2822),
    [anon_sym_ATsynchronized] = ACTIONS(2822),
    [anon_sym_BOOL] = ACTIONS(2820),
    [anon_sym_IMP] = ACTIONS(2820),
    [anon_sym_SEL] = ACTIONS(2820),
    [anon_sym_Class] = ACTIONS(2820),
    [anon_sym_id] = ACTIONS(2820),
  },
  [STATE(627)] = {
    [sym_identifier] = ACTIONS(2828),
    [aux_sym_preproc_include_token1] = ACTIONS(2828),
    [aux_sym_preproc_include_token2] = ACTIONS(2828),
    [aux_sym_preproc_def_token1] = ACTIONS(2828),
    [aux_sym_preproc_if_token1] = ACTIONS(2828),
    [aux_sym_preproc_if_token2] = ACTIONS(2828),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2828),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2828),
    [sym_preproc_directive] = ACTIONS(2828),
    [anon_sym_LPAREN2] = ACTIONS(2830),
    [anon_sym_BANG] = ACTIONS(2830),
    [anon_sym_TILDE] = ACTIONS(2830),
    [anon_sym_DASH] = ACTIONS(2828),
    [anon_sym_PLUS] = ACTIONS(2828),
    [anon_sym_STAR] = ACTIONS(2830),
    [anon_sym_CARET] = ACTIONS(2830),
    [anon_sym_AMP] = ACTIONS(2830),
    [anon_sym_SEMI] = ACTIONS(2830),
    [anon_sym___extension__] = ACTIONS(2828),
    [anon_sym_typedef] = ACTIONS(2828),
    [anon_sym_extern] = ACTIONS(2828),
    [anon_sym___attribute__] = ACTIONS(2828),
    [anon_sym___attribute] = ACTIONS(2828),
    [anon_sym_noreturn] = ACTIONS(2828),
    [anon_sym_LBRACK] = ACTIONS(2830),
    [anon_sym___declspec] = ACTIONS(2828),
    [anon_sym___cdecl] = ACTIONS(2828),
    [anon_sym___clrcall] = ACTIONS(2828),
    [anon_sym___stdcall] = ACTIONS(2828),
    [anon_sym___fastcall] = ACTIONS(2828),
    [anon_sym___thiscall] = ACTIONS(2828),
    [anon_sym___vectorcall] = ACTIONS(2828),
    [anon_sym_LBRACE] = ACTIONS(2830),
    [anon_sym_signed] = ACTIONS(2828),
    [anon_sym_unsigned] = ACTIONS(2828),
    [anon_sym_long] = ACTIONS(2828),
    [anon_sym_short] = ACTIONS(2828),
    [anon_sym_static] = ACTIONS(2828),
    [anon_sym_ATautoreleasepool] = ACTIONS(2830),
    [anon_sym_auto] = ACTIONS(2828),
    [anon_sym_register] = ACTIONS(2828),
    [anon_sym_inline] = ACTIONS(2828),
    [anon_sym___inline] = ACTIONS(2828),
    [anon_sym___inline__] = ACTIONS(2828),
    [anon_sym___forceinline] = ACTIONS(2828),
    [anon_sym_thread_local] = ACTIONS(2828),
    [anon_sym___thread] = ACTIONS(2828),
    [anon_sym_CG_EXTERN] = ACTIONS(2828),
    [anon_sym_CG_INLINE] = ACTIONS(2828),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2828),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2828),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2828),
    [anon_sym_IBOutlet] = ACTIONS(2828),
    [anon_sym_IBInspectable] = ACTIONS(2828),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2828),
    [anon_sym_NS_INLINE] = ACTIONS(2828),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2828),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2828),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2828),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2828),
    [anon_sym_const] = ACTIONS(2828),
    [anon_sym_constexpr] = ACTIONS(2828),
    [anon_sym_volatile] = ACTIONS(2828),
    [anon_sym_restrict] = ACTIONS(2828),
    [anon_sym___restrict__] = ACTIONS(2828),
    [anon_sym__Atomic] = ACTIONS(2828),
    [anon_sym__Noreturn] = ACTIONS(2828),
    [anon_sym__Nonnull] = ACTIONS(2828),
    [anon_sym_nullable] = ACTIONS(2828),
    [anon_sym__Complex] = ACTIONS(2828),
    [anon_sym__Nullable] = ACTIONS(2828),
    [anon_sym__Nullable_result] = ACTIONS(2828),
    [anon_sym__Null_unspecified] = ACTIONS(2828),
    [anon_sym___autoreleasing] = ACTIONS(2828),
    [anon_sym___block] = ACTIONS(2828),
    [anon_sym___bridge] = ACTIONS(2828),
    [anon_sym___bridge_retained] = ACTIONS(2828),
    [anon_sym___bridge_transfer] = ACTIONS(2828),
    [anon_sym___complex] = ACTIONS(2828),
    [anon_sym___const] = ACTIONS(2828),
    [anon_sym___imag] = ACTIONS(2828),
    [anon_sym___kindof] = ACTIONS(2828),
    [anon_sym___nonnull] = ACTIONS(2828),
    [anon_sym___nullable] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2828),
    [anon_sym___real] = ACTIONS(2828),
    [anon_sym___strong] = ACTIONS(2828),
    [anon_sym___unsafe_unretained] = ACTIONS(2828),
    [anon_sym___unused] = ACTIONS(2828),
    [anon_sym___weak] = ACTIONS(2828),
    [anon_sym_alignas] = ACTIONS(2828),
    [anon_sym__Alignas] = ACTIONS(2828),
    [sym_primitive_type] = ACTIONS(2828),
    [anon_sym_enum] = ACTIONS(2828),
    [anon_sym_struct] = ACTIONS(2828),
    [anon_sym_union] = ACTIONS(2828),
    [anon_sym_if] = ACTIONS(2828),
    [anon_sym_switch] = ACTIONS(2828),
    [anon_sym_case] = ACTIONS(2828),
    [anon_sym_default] = ACTIONS(2828),
    [anon_sym_while] = ACTIONS(2828),
    [anon_sym_do] = ACTIONS(2828),
    [anon_sym_for] = ACTIONS(2828),
    [anon_sym_in] = ACTIONS(2828),
    [anon_sym_return] = ACTIONS(2828),
    [anon_sym_break] = ACTIONS(2828),
    [anon_sym_continue] = ACTIONS(2828),
    [anon_sym_goto] = ACTIONS(2828),
    [anon_sym_DASH_DASH] = ACTIONS(2830),
    [anon_sym_PLUS_PLUS] = ACTIONS(2830),
    [anon_sym_sizeof] = ACTIONS(2828),
    [anon_sym___alignof__] = ACTIONS(2828),
    [anon_sym___alignof] = ACTIONS(2828),
    [anon_sym__alignof] = ACTIONS(2828),
    [anon_sym_alignof] = ACTIONS(2828),
    [anon_sym__Alignof] = ACTIONS(2828),
    [anon_sym_offsetof] = ACTIONS(2828),
    [anon_sym__Generic] = ACTIONS(2828),
    [anon_sym_asm] = ACTIONS(2828),
    [anon_sym___asm__] = ACTIONS(2828),
    [anon_sym___asm] = ACTIONS(2828),
    [sym_number_literal] = ACTIONS(2830),
    [anon_sym_L_SQUOTE] = ACTIONS(2830),
    [anon_sym_u_SQUOTE] = ACTIONS(2830),
    [anon_sym_U_SQUOTE] = ACTIONS(2830),
    [anon_sym_u8_SQUOTE] = ACTIONS(2830),
    [anon_sym_SQUOTE] = ACTIONS(2830),
    [anon_sym_AT] = ACTIONS(2828),
    [anon_sym_DQUOTE] = ACTIONS(2830),
    [anon_sym_L_DQUOTE] = ACTIONS(2830),
    [anon_sym_u_DQUOTE] = ACTIONS(2830),
    [anon_sym_U_DQUOTE] = ACTIONS(2830),
    [anon_sym_u8_DQUOTE] = ACTIONS(2830),
    [sym_true] = ACTIONS(2828),
    [sym_false] = ACTIONS(2828),
    [anon_sym_NULL] = ACTIONS(2828),
    [anon_sym_nullptr] = ACTIONS(2828),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2828),
    [anon_sym___typeof] = ACTIONS(2828),
    [anon_sym_typeof] = ACTIONS(2828),
    [anon_sym_ATimport] = ACTIONS(2830),
    [aux_sym_preproc_undef_token1] = ACTIONS(2828),
    [anon_sym_POUND] = ACTIONS(2828),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2828),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2828),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2828),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2828),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2828),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2828),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2828),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_API_AVAILABLE] = ACTIONS(2828),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_API_DEPRECATED] = ACTIONS(2828),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2828),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym___deprecated_msg] = ACTIONS(2828),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2828),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2828),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2828),
    [anon_sym_ATprotocol] = ACTIONS(2830),
    [anon_sym_ATinterface] = ACTIONS(2830),
    [anon_sym_ATimplementation] = ACTIONS(2830),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2830),
    [anon_sym_ATsynthesize] = ACTIONS(2830),
    [anon_sym_ATdynamic] = ACTIONS(2830),
    [anon_sym_ATtry] = ACTIONS(2830),
    [anon_sym___try] = ACTIONS(2828),
    [anon_sym_ATthrow] = ACTIONS(2830),
    [anon_sym_ATselector] = ACTIONS(2830),
    [anon_sym_ATavailable] = ACTIONS(2830),
    [anon_sym___builtin_available] = ACTIONS(2828),
    [anon_sym_va_arg] = ACTIONS(2828),
    [anon_sym_ATencode] = ACTIONS(2830),
    [anon_sym_ATsynchronized] = ACTIONS(2830),
    [anon_sym_BOOL] = ACTIONS(2828),
    [anon_sym_IMP] = ACTIONS(2828),
    [anon_sym_SEL] = ACTIONS(2828),
    [anon_sym_Class] = ACTIONS(2828),
    [anon_sym_id] = ACTIONS(2828),
  },
  [STATE(628)] = {
    [sym_identifier] = ACTIONS(2832),
    [aux_sym_preproc_include_token1] = ACTIONS(2832),
    [aux_sym_preproc_include_token2] = ACTIONS(2832),
    [aux_sym_preproc_def_token1] = ACTIONS(2832),
    [aux_sym_preproc_if_token1] = ACTIONS(2832),
    [aux_sym_preproc_if_token2] = ACTIONS(2832),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2832),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2832),
    [sym_preproc_directive] = ACTIONS(2832),
    [anon_sym_LPAREN2] = ACTIONS(2834),
    [anon_sym_BANG] = ACTIONS(2834),
    [anon_sym_TILDE] = ACTIONS(2834),
    [anon_sym_DASH] = ACTIONS(2832),
    [anon_sym_PLUS] = ACTIONS(2832),
    [anon_sym_STAR] = ACTIONS(2834),
    [anon_sym_CARET] = ACTIONS(2834),
    [anon_sym_AMP] = ACTIONS(2834),
    [anon_sym_SEMI] = ACTIONS(2834),
    [anon_sym___extension__] = ACTIONS(2832),
    [anon_sym_typedef] = ACTIONS(2832),
    [anon_sym_extern] = ACTIONS(2832),
    [anon_sym___attribute__] = ACTIONS(2832),
    [anon_sym___attribute] = ACTIONS(2832),
    [anon_sym_noreturn] = ACTIONS(2832),
    [anon_sym_LBRACK] = ACTIONS(2834),
    [anon_sym___declspec] = ACTIONS(2832),
    [anon_sym___cdecl] = ACTIONS(2832),
    [anon_sym___clrcall] = ACTIONS(2832),
    [anon_sym___stdcall] = ACTIONS(2832),
    [anon_sym___fastcall] = ACTIONS(2832),
    [anon_sym___thiscall] = ACTIONS(2832),
    [anon_sym___vectorcall] = ACTIONS(2832),
    [anon_sym_LBRACE] = ACTIONS(2834),
    [anon_sym_signed] = ACTIONS(2832),
    [anon_sym_unsigned] = ACTIONS(2832),
    [anon_sym_long] = ACTIONS(2832),
    [anon_sym_short] = ACTIONS(2832),
    [anon_sym_static] = ACTIONS(2832),
    [anon_sym_ATautoreleasepool] = ACTIONS(2834),
    [anon_sym_auto] = ACTIONS(2832),
    [anon_sym_register] = ACTIONS(2832),
    [anon_sym_inline] = ACTIONS(2832),
    [anon_sym___inline] = ACTIONS(2832),
    [anon_sym___inline__] = ACTIONS(2832),
    [anon_sym___forceinline] = ACTIONS(2832),
    [anon_sym_thread_local] = ACTIONS(2832),
    [anon_sym___thread] = ACTIONS(2832),
    [anon_sym_CG_EXTERN] = ACTIONS(2832),
    [anon_sym_CG_INLINE] = ACTIONS(2832),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2832),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2832),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2832),
    [anon_sym_IBOutlet] = ACTIONS(2832),
    [anon_sym_IBInspectable] = ACTIONS(2832),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2832),
    [anon_sym_NS_INLINE] = ACTIONS(2832),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2832),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2832),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2832),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2832),
    [anon_sym_const] = ACTIONS(2832),
    [anon_sym_constexpr] = ACTIONS(2832),
    [anon_sym_volatile] = ACTIONS(2832),
    [anon_sym_restrict] = ACTIONS(2832),
    [anon_sym___restrict__] = ACTIONS(2832),
    [anon_sym__Atomic] = ACTIONS(2832),
    [anon_sym__Noreturn] = ACTIONS(2832),
    [anon_sym__Nonnull] = ACTIONS(2832),
    [anon_sym_nullable] = ACTIONS(2832),
    [anon_sym__Complex] = ACTIONS(2832),
    [anon_sym__Nullable] = ACTIONS(2832),
    [anon_sym__Nullable_result] = ACTIONS(2832),
    [anon_sym__Null_unspecified] = ACTIONS(2832),
    [anon_sym___autoreleasing] = ACTIONS(2832),
    [anon_sym___block] = ACTIONS(2832),
    [anon_sym___bridge] = ACTIONS(2832),
    [anon_sym___bridge_retained] = ACTIONS(2832),
    [anon_sym___bridge_transfer] = ACTIONS(2832),
    [anon_sym___complex] = ACTIONS(2832),
    [anon_sym___const] = ACTIONS(2832),
    [anon_sym___imag] = ACTIONS(2832),
    [anon_sym___kindof] = ACTIONS(2832),
    [anon_sym___nonnull] = ACTIONS(2832),
    [anon_sym___nullable] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2832),
    [anon_sym___real] = ACTIONS(2832),
    [anon_sym___strong] = ACTIONS(2832),
    [anon_sym___unsafe_unretained] = ACTIONS(2832),
    [anon_sym___unused] = ACTIONS(2832),
    [anon_sym___weak] = ACTIONS(2832),
    [anon_sym_alignas] = ACTIONS(2832),
    [anon_sym__Alignas] = ACTIONS(2832),
    [sym_primitive_type] = ACTIONS(2832),
    [anon_sym_enum] = ACTIONS(2832),
    [anon_sym_struct] = ACTIONS(2832),
    [anon_sym_union] = ACTIONS(2832),
    [anon_sym_if] = ACTIONS(2832),
    [anon_sym_switch] = ACTIONS(2832),
    [anon_sym_case] = ACTIONS(2832),
    [anon_sym_default] = ACTIONS(2832),
    [anon_sym_while] = ACTIONS(2832),
    [anon_sym_do] = ACTIONS(2832),
    [anon_sym_for] = ACTIONS(2832),
    [anon_sym_in] = ACTIONS(2832),
    [anon_sym_return] = ACTIONS(2832),
    [anon_sym_break] = ACTIONS(2832),
    [anon_sym_continue] = ACTIONS(2832),
    [anon_sym_goto] = ACTIONS(2832),
    [anon_sym_DASH_DASH] = ACTIONS(2834),
    [anon_sym_PLUS_PLUS] = ACTIONS(2834),
    [anon_sym_sizeof] = ACTIONS(2832),
    [anon_sym___alignof__] = ACTIONS(2832),
    [anon_sym___alignof] = ACTIONS(2832),
    [anon_sym__alignof] = ACTIONS(2832),
    [anon_sym_alignof] = ACTIONS(2832),
    [anon_sym__Alignof] = ACTIONS(2832),
    [anon_sym_offsetof] = ACTIONS(2832),
    [anon_sym__Generic] = ACTIONS(2832),
    [anon_sym_asm] = ACTIONS(2832),
    [anon_sym___asm__] = ACTIONS(2832),
    [anon_sym___asm] = ACTIONS(2832),
    [sym_number_literal] = ACTIONS(2834),
    [anon_sym_L_SQUOTE] = ACTIONS(2834),
    [anon_sym_u_SQUOTE] = ACTIONS(2834),
    [anon_sym_U_SQUOTE] = ACTIONS(2834),
    [anon_sym_u8_SQUOTE] = ACTIONS(2834),
    [anon_sym_SQUOTE] = ACTIONS(2834),
    [anon_sym_AT] = ACTIONS(2832),
    [anon_sym_DQUOTE] = ACTIONS(2834),
    [anon_sym_L_DQUOTE] = ACTIONS(2834),
    [anon_sym_u_DQUOTE] = ACTIONS(2834),
    [anon_sym_U_DQUOTE] = ACTIONS(2834),
    [anon_sym_u8_DQUOTE] = ACTIONS(2834),
    [sym_true] = ACTIONS(2832),
    [sym_false] = ACTIONS(2832),
    [anon_sym_NULL] = ACTIONS(2832),
    [anon_sym_nullptr] = ACTIONS(2832),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2832),
    [anon_sym___typeof] = ACTIONS(2832),
    [anon_sym_typeof] = ACTIONS(2832),
    [anon_sym_ATimport] = ACTIONS(2834),
    [aux_sym_preproc_undef_token1] = ACTIONS(2832),
    [anon_sym_POUND] = ACTIONS(2832),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2832),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2832),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2832),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2832),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2832),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2832),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2832),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_API_AVAILABLE] = ACTIONS(2832),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_API_DEPRECATED] = ACTIONS(2832),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2832),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym___deprecated_msg] = ACTIONS(2832),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2832),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2832),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2832),
    [anon_sym_ATprotocol] = ACTIONS(2834),
    [anon_sym_ATinterface] = ACTIONS(2834),
    [anon_sym_ATimplementation] = ACTIONS(2834),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2834),
    [anon_sym_ATsynthesize] = ACTIONS(2834),
    [anon_sym_ATdynamic] = ACTIONS(2834),
    [anon_sym_ATtry] = ACTIONS(2834),
    [anon_sym___try] = ACTIONS(2832),
    [anon_sym_ATthrow] = ACTIONS(2834),
    [anon_sym_ATselector] = ACTIONS(2834),
    [anon_sym_ATavailable] = ACTIONS(2834),
    [anon_sym___builtin_available] = ACTIONS(2832),
    [anon_sym_va_arg] = ACTIONS(2832),
    [anon_sym_ATencode] = ACTIONS(2834),
    [anon_sym_ATsynchronized] = ACTIONS(2834),
    [anon_sym_BOOL] = ACTIONS(2832),
    [anon_sym_IMP] = ACTIONS(2832),
    [anon_sym_SEL] = ACTIONS(2832),
    [anon_sym_Class] = ACTIONS(2832),
    [anon_sym_id] = ACTIONS(2832),
  },
  [STATE(629)] = {
    [sym_identifier] = ACTIONS(2836),
    [aux_sym_preproc_include_token1] = ACTIONS(2836),
    [aux_sym_preproc_include_token2] = ACTIONS(2836),
    [aux_sym_preproc_def_token1] = ACTIONS(2836),
    [aux_sym_preproc_if_token1] = ACTIONS(2836),
    [aux_sym_preproc_if_token2] = ACTIONS(2836),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2836),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2836),
    [sym_preproc_directive] = ACTIONS(2836),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_BANG] = ACTIONS(2838),
    [anon_sym_TILDE] = ACTIONS(2838),
    [anon_sym_DASH] = ACTIONS(2836),
    [anon_sym_PLUS] = ACTIONS(2836),
    [anon_sym_STAR] = ACTIONS(2838),
    [anon_sym_CARET] = ACTIONS(2838),
    [anon_sym_AMP] = ACTIONS(2838),
    [anon_sym_SEMI] = ACTIONS(2838),
    [anon_sym___extension__] = ACTIONS(2836),
    [anon_sym_typedef] = ACTIONS(2836),
    [anon_sym_extern] = ACTIONS(2836),
    [anon_sym___attribute__] = ACTIONS(2836),
    [anon_sym___attribute] = ACTIONS(2836),
    [anon_sym_noreturn] = ACTIONS(2836),
    [anon_sym_LBRACK] = ACTIONS(2838),
    [anon_sym___declspec] = ACTIONS(2836),
    [anon_sym___cdecl] = ACTIONS(2836),
    [anon_sym___clrcall] = ACTIONS(2836),
    [anon_sym___stdcall] = ACTIONS(2836),
    [anon_sym___fastcall] = ACTIONS(2836),
    [anon_sym___thiscall] = ACTIONS(2836),
    [anon_sym___vectorcall] = ACTIONS(2836),
    [anon_sym_LBRACE] = ACTIONS(2838),
    [anon_sym_signed] = ACTIONS(2836),
    [anon_sym_unsigned] = ACTIONS(2836),
    [anon_sym_long] = ACTIONS(2836),
    [anon_sym_short] = ACTIONS(2836),
    [anon_sym_static] = ACTIONS(2836),
    [anon_sym_ATautoreleasepool] = ACTIONS(2838),
    [anon_sym_auto] = ACTIONS(2836),
    [anon_sym_register] = ACTIONS(2836),
    [anon_sym_inline] = ACTIONS(2836),
    [anon_sym___inline] = ACTIONS(2836),
    [anon_sym___inline__] = ACTIONS(2836),
    [anon_sym___forceinline] = ACTIONS(2836),
    [anon_sym_thread_local] = ACTIONS(2836),
    [anon_sym___thread] = ACTIONS(2836),
    [anon_sym_CG_EXTERN] = ACTIONS(2836),
    [anon_sym_CG_INLINE] = ACTIONS(2836),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2836),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2836),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2836),
    [anon_sym_IBOutlet] = ACTIONS(2836),
    [anon_sym_IBInspectable] = ACTIONS(2836),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2836),
    [anon_sym_NS_INLINE] = ACTIONS(2836),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2836),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2836),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2836),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2836),
    [anon_sym_const] = ACTIONS(2836),
    [anon_sym_constexpr] = ACTIONS(2836),
    [anon_sym_volatile] = ACTIONS(2836),
    [anon_sym_restrict] = ACTIONS(2836),
    [anon_sym___restrict__] = ACTIONS(2836),
    [anon_sym__Atomic] = ACTIONS(2836),
    [anon_sym__Noreturn] = ACTIONS(2836),
    [anon_sym__Nonnull] = ACTIONS(2836),
    [anon_sym_nullable] = ACTIONS(2836),
    [anon_sym__Complex] = ACTIONS(2836),
    [anon_sym__Nullable] = ACTIONS(2836),
    [anon_sym__Nullable_result] = ACTIONS(2836),
    [anon_sym__Null_unspecified] = ACTIONS(2836),
    [anon_sym___autoreleasing] = ACTIONS(2836),
    [anon_sym___block] = ACTIONS(2836),
    [anon_sym___bridge] = ACTIONS(2836),
    [anon_sym___bridge_retained] = ACTIONS(2836),
    [anon_sym___bridge_transfer] = ACTIONS(2836),
    [anon_sym___complex] = ACTIONS(2836),
    [anon_sym___const] = ACTIONS(2836),
    [anon_sym___imag] = ACTIONS(2836),
    [anon_sym___kindof] = ACTIONS(2836),
    [anon_sym___nonnull] = ACTIONS(2836),
    [anon_sym___nullable] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2836),
    [anon_sym___real] = ACTIONS(2836),
    [anon_sym___strong] = ACTIONS(2836),
    [anon_sym___unsafe_unretained] = ACTIONS(2836),
    [anon_sym___unused] = ACTIONS(2836),
    [anon_sym___weak] = ACTIONS(2836),
    [anon_sym_alignas] = ACTIONS(2836),
    [anon_sym__Alignas] = ACTIONS(2836),
    [sym_primitive_type] = ACTIONS(2836),
    [anon_sym_enum] = ACTIONS(2836),
    [anon_sym_struct] = ACTIONS(2836),
    [anon_sym_union] = ACTIONS(2836),
    [anon_sym_if] = ACTIONS(2836),
    [anon_sym_switch] = ACTIONS(2836),
    [anon_sym_case] = ACTIONS(2836),
    [anon_sym_default] = ACTIONS(2836),
    [anon_sym_while] = ACTIONS(2836),
    [anon_sym_do] = ACTIONS(2836),
    [anon_sym_for] = ACTIONS(2836),
    [anon_sym_in] = ACTIONS(2836),
    [anon_sym_return] = ACTIONS(2836),
    [anon_sym_break] = ACTIONS(2836),
    [anon_sym_continue] = ACTIONS(2836),
    [anon_sym_goto] = ACTIONS(2836),
    [anon_sym_DASH_DASH] = ACTIONS(2838),
    [anon_sym_PLUS_PLUS] = ACTIONS(2838),
    [anon_sym_sizeof] = ACTIONS(2836),
    [anon_sym___alignof__] = ACTIONS(2836),
    [anon_sym___alignof] = ACTIONS(2836),
    [anon_sym__alignof] = ACTIONS(2836),
    [anon_sym_alignof] = ACTIONS(2836),
    [anon_sym__Alignof] = ACTIONS(2836),
    [anon_sym_offsetof] = ACTIONS(2836),
    [anon_sym__Generic] = ACTIONS(2836),
    [anon_sym_asm] = ACTIONS(2836),
    [anon_sym___asm__] = ACTIONS(2836),
    [anon_sym___asm] = ACTIONS(2836),
    [sym_number_literal] = ACTIONS(2838),
    [anon_sym_L_SQUOTE] = ACTIONS(2838),
    [anon_sym_u_SQUOTE] = ACTIONS(2838),
    [anon_sym_U_SQUOTE] = ACTIONS(2838),
    [anon_sym_u8_SQUOTE] = ACTIONS(2838),
    [anon_sym_SQUOTE] = ACTIONS(2838),
    [anon_sym_AT] = ACTIONS(2836),
    [anon_sym_DQUOTE] = ACTIONS(2838),
    [anon_sym_L_DQUOTE] = ACTIONS(2838),
    [anon_sym_u_DQUOTE] = ACTIONS(2838),
    [anon_sym_U_DQUOTE] = ACTIONS(2838),
    [anon_sym_u8_DQUOTE] = ACTIONS(2838),
    [sym_true] = ACTIONS(2836),
    [sym_false] = ACTIONS(2836),
    [anon_sym_NULL] = ACTIONS(2836),
    [anon_sym_nullptr] = ACTIONS(2836),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2836),
    [anon_sym___typeof] = ACTIONS(2836),
    [anon_sym_typeof] = ACTIONS(2836),
    [anon_sym_ATimport] = ACTIONS(2838),
    [aux_sym_preproc_undef_token1] = ACTIONS(2836),
    [anon_sym_POUND] = ACTIONS(2836),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2836),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2836),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2836),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2836),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2836),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2836),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2836),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_API_AVAILABLE] = ACTIONS(2836),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_API_DEPRECATED] = ACTIONS(2836),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2836),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym___deprecated_msg] = ACTIONS(2836),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2836),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2836),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2836),
    [anon_sym_ATprotocol] = ACTIONS(2838),
    [anon_sym_ATinterface] = ACTIONS(2838),
    [anon_sym_ATimplementation] = ACTIONS(2838),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2838),
    [anon_sym_ATsynthesize] = ACTIONS(2838),
    [anon_sym_ATdynamic] = ACTIONS(2838),
    [anon_sym_ATtry] = ACTIONS(2838),
    [anon_sym___try] = ACTIONS(2836),
    [anon_sym_ATthrow] = ACTIONS(2838),
    [anon_sym_ATselector] = ACTIONS(2838),
    [anon_sym_ATavailable] = ACTIONS(2838),
    [anon_sym___builtin_available] = ACTIONS(2836),
    [anon_sym_va_arg] = ACTIONS(2836),
    [anon_sym_ATencode] = ACTIONS(2838),
    [anon_sym_ATsynchronized] = ACTIONS(2838),
    [anon_sym_BOOL] = ACTIONS(2836),
    [anon_sym_IMP] = ACTIONS(2836),
    [anon_sym_SEL] = ACTIONS(2836),
    [anon_sym_Class] = ACTIONS(2836),
    [anon_sym_id] = ACTIONS(2836),
  },
  [STATE(630)] = {
    [sym_identifier] = ACTIONS(2840),
    [aux_sym_preproc_include_token1] = ACTIONS(2840),
    [aux_sym_preproc_include_token2] = ACTIONS(2840),
    [aux_sym_preproc_def_token1] = ACTIONS(2840),
    [aux_sym_preproc_if_token1] = ACTIONS(2840),
    [aux_sym_preproc_if_token2] = ACTIONS(2840),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2840),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2840),
    [sym_preproc_directive] = ACTIONS(2840),
    [anon_sym_LPAREN2] = ACTIONS(2842),
    [anon_sym_BANG] = ACTIONS(2842),
    [anon_sym_TILDE] = ACTIONS(2842),
    [anon_sym_DASH] = ACTIONS(2840),
    [anon_sym_PLUS] = ACTIONS(2840),
    [anon_sym_STAR] = ACTIONS(2842),
    [anon_sym_CARET] = ACTIONS(2842),
    [anon_sym_AMP] = ACTIONS(2842),
    [anon_sym_SEMI] = ACTIONS(2842),
    [anon_sym___extension__] = ACTIONS(2840),
    [anon_sym_typedef] = ACTIONS(2840),
    [anon_sym_extern] = ACTIONS(2840),
    [anon_sym___attribute__] = ACTIONS(2840),
    [anon_sym___attribute] = ACTIONS(2840),
    [anon_sym_noreturn] = ACTIONS(2840),
    [anon_sym_LBRACK] = ACTIONS(2842),
    [anon_sym___declspec] = ACTIONS(2840),
    [anon_sym___cdecl] = ACTIONS(2840),
    [anon_sym___clrcall] = ACTIONS(2840),
    [anon_sym___stdcall] = ACTIONS(2840),
    [anon_sym___fastcall] = ACTIONS(2840),
    [anon_sym___thiscall] = ACTIONS(2840),
    [anon_sym___vectorcall] = ACTIONS(2840),
    [anon_sym_LBRACE] = ACTIONS(2842),
    [anon_sym_signed] = ACTIONS(2840),
    [anon_sym_unsigned] = ACTIONS(2840),
    [anon_sym_long] = ACTIONS(2840),
    [anon_sym_short] = ACTIONS(2840),
    [anon_sym_static] = ACTIONS(2840),
    [anon_sym_ATautoreleasepool] = ACTIONS(2842),
    [anon_sym_auto] = ACTIONS(2840),
    [anon_sym_register] = ACTIONS(2840),
    [anon_sym_inline] = ACTIONS(2840),
    [anon_sym___inline] = ACTIONS(2840),
    [anon_sym___inline__] = ACTIONS(2840),
    [anon_sym___forceinline] = ACTIONS(2840),
    [anon_sym_thread_local] = ACTIONS(2840),
    [anon_sym___thread] = ACTIONS(2840),
    [anon_sym_CG_EXTERN] = ACTIONS(2840),
    [anon_sym_CG_INLINE] = ACTIONS(2840),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2840),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2840),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2840),
    [anon_sym_IBOutlet] = ACTIONS(2840),
    [anon_sym_IBInspectable] = ACTIONS(2840),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2840),
    [anon_sym_NS_INLINE] = ACTIONS(2840),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2840),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2840),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2840),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2840),
    [anon_sym_const] = ACTIONS(2840),
    [anon_sym_constexpr] = ACTIONS(2840),
    [anon_sym_volatile] = ACTIONS(2840),
    [anon_sym_restrict] = ACTIONS(2840),
    [anon_sym___restrict__] = ACTIONS(2840),
    [anon_sym__Atomic] = ACTIONS(2840),
    [anon_sym__Noreturn] = ACTIONS(2840),
    [anon_sym__Nonnull] = ACTIONS(2840),
    [anon_sym_nullable] = ACTIONS(2840),
    [anon_sym__Complex] = ACTIONS(2840),
    [anon_sym__Nullable] = ACTIONS(2840),
    [anon_sym__Nullable_result] = ACTIONS(2840),
    [anon_sym__Null_unspecified] = ACTIONS(2840),
    [anon_sym___autoreleasing] = ACTIONS(2840),
    [anon_sym___block] = ACTIONS(2840),
    [anon_sym___bridge] = ACTIONS(2840),
    [anon_sym___bridge_retained] = ACTIONS(2840),
    [anon_sym___bridge_transfer] = ACTIONS(2840),
    [anon_sym___complex] = ACTIONS(2840),
    [anon_sym___const] = ACTIONS(2840),
    [anon_sym___imag] = ACTIONS(2840),
    [anon_sym___kindof] = ACTIONS(2840),
    [anon_sym___nonnull] = ACTIONS(2840),
    [anon_sym___nullable] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2840),
    [anon_sym___real] = ACTIONS(2840),
    [anon_sym___strong] = ACTIONS(2840),
    [anon_sym___unsafe_unretained] = ACTIONS(2840),
    [anon_sym___unused] = ACTIONS(2840),
    [anon_sym___weak] = ACTIONS(2840),
    [anon_sym_alignas] = ACTIONS(2840),
    [anon_sym__Alignas] = ACTIONS(2840),
    [sym_primitive_type] = ACTIONS(2840),
    [anon_sym_enum] = ACTIONS(2840),
    [anon_sym_struct] = ACTIONS(2840),
    [anon_sym_union] = ACTIONS(2840),
    [anon_sym_if] = ACTIONS(2840),
    [anon_sym_switch] = ACTIONS(2840),
    [anon_sym_case] = ACTIONS(2840),
    [anon_sym_default] = ACTIONS(2840),
    [anon_sym_while] = ACTIONS(2840),
    [anon_sym_do] = ACTIONS(2840),
    [anon_sym_for] = ACTIONS(2840),
    [anon_sym_in] = ACTIONS(2840),
    [anon_sym_return] = ACTIONS(2840),
    [anon_sym_break] = ACTIONS(2840),
    [anon_sym_continue] = ACTIONS(2840),
    [anon_sym_goto] = ACTIONS(2840),
    [anon_sym_DASH_DASH] = ACTIONS(2842),
    [anon_sym_PLUS_PLUS] = ACTIONS(2842),
    [anon_sym_sizeof] = ACTIONS(2840),
    [anon_sym___alignof__] = ACTIONS(2840),
    [anon_sym___alignof] = ACTIONS(2840),
    [anon_sym__alignof] = ACTIONS(2840),
    [anon_sym_alignof] = ACTIONS(2840),
    [anon_sym__Alignof] = ACTIONS(2840),
    [anon_sym_offsetof] = ACTIONS(2840),
    [anon_sym__Generic] = ACTIONS(2840),
    [anon_sym_asm] = ACTIONS(2840),
    [anon_sym___asm__] = ACTIONS(2840),
    [anon_sym___asm] = ACTIONS(2840),
    [sym_number_literal] = ACTIONS(2842),
    [anon_sym_L_SQUOTE] = ACTIONS(2842),
    [anon_sym_u_SQUOTE] = ACTIONS(2842),
    [anon_sym_U_SQUOTE] = ACTIONS(2842),
    [anon_sym_u8_SQUOTE] = ACTIONS(2842),
    [anon_sym_SQUOTE] = ACTIONS(2842),
    [anon_sym_AT] = ACTIONS(2840),
    [anon_sym_DQUOTE] = ACTIONS(2842),
    [anon_sym_L_DQUOTE] = ACTIONS(2842),
    [anon_sym_u_DQUOTE] = ACTIONS(2842),
    [anon_sym_U_DQUOTE] = ACTIONS(2842),
    [anon_sym_u8_DQUOTE] = ACTIONS(2842),
    [sym_true] = ACTIONS(2840),
    [sym_false] = ACTIONS(2840),
    [anon_sym_NULL] = ACTIONS(2840),
    [anon_sym_nullptr] = ACTIONS(2840),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2840),
    [anon_sym___typeof] = ACTIONS(2840),
    [anon_sym_typeof] = ACTIONS(2840),
    [anon_sym_ATimport] = ACTIONS(2842),
    [aux_sym_preproc_undef_token1] = ACTIONS(2840),
    [anon_sym_POUND] = ACTIONS(2840),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2840),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2840),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2840),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2840),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2840),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2840),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2840),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_API_AVAILABLE] = ACTIONS(2840),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_API_DEPRECATED] = ACTIONS(2840),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2840),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym___deprecated_msg] = ACTIONS(2840),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2840),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2840),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2840),
    [anon_sym_ATprotocol] = ACTIONS(2842),
    [anon_sym_ATinterface] = ACTIONS(2842),
    [anon_sym_ATimplementation] = ACTIONS(2842),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2842),
    [anon_sym_ATsynthesize] = ACTIONS(2842),
    [anon_sym_ATdynamic] = ACTIONS(2842),
    [anon_sym_ATtry] = ACTIONS(2842),
    [anon_sym___try] = ACTIONS(2840),
    [anon_sym_ATthrow] = ACTIONS(2842),
    [anon_sym_ATselector] = ACTIONS(2842),
    [anon_sym_ATavailable] = ACTIONS(2842),
    [anon_sym___builtin_available] = ACTIONS(2840),
    [anon_sym_va_arg] = ACTIONS(2840),
    [anon_sym_ATencode] = ACTIONS(2842),
    [anon_sym_ATsynchronized] = ACTIONS(2842),
    [anon_sym_BOOL] = ACTIONS(2840),
    [anon_sym_IMP] = ACTIONS(2840),
    [anon_sym_SEL] = ACTIONS(2840),
    [anon_sym_Class] = ACTIONS(2840),
    [anon_sym_id] = ACTIONS(2840),
  },
  [STATE(631)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token2] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATsynthesize] = ACTIONS(2846),
    [anon_sym_ATdynamic] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(632)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token2] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATsynthesize] = ACTIONS(2846),
    [anon_sym_ATdynamic] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(633)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token2] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATsynthesize] = ACTIONS(2846),
    [anon_sym_ATdynamic] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(634)] = {
    [sym_identifier] = ACTIONS(2848),
    [aux_sym_preproc_include_token1] = ACTIONS(2848),
    [aux_sym_preproc_include_token2] = ACTIONS(2848),
    [aux_sym_preproc_def_token1] = ACTIONS(2848),
    [aux_sym_preproc_if_token1] = ACTIONS(2848),
    [aux_sym_preproc_if_token2] = ACTIONS(2848),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2848),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2848),
    [sym_preproc_directive] = ACTIONS(2848),
    [anon_sym_LPAREN2] = ACTIONS(2850),
    [anon_sym_BANG] = ACTIONS(2850),
    [anon_sym_TILDE] = ACTIONS(2850),
    [anon_sym_DASH] = ACTIONS(2848),
    [anon_sym_PLUS] = ACTIONS(2848),
    [anon_sym_STAR] = ACTIONS(2850),
    [anon_sym_CARET] = ACTIONS(2850),
    [anon_sym_AMP] = ACTIONS(2850),
    [anon_sym_SEMI] = ACTIONS(2850),
    [anon_sym___extension__] = ACTIONS(2848),
    [anon_sym_typedef] = ACTIONS(2848),
    [anon_sym_extern] = ACTIONS(2848),
    [anon_sym___attribute__] = ACTIONS(2848),
    [anon_sym___attribute] = ACTIONS(2848),
    [anon_sym_noreturn] = ACTIONS(2848),
    [anon_sym_LBRACK] = ACTIONS(2850),
    [anon_sym___declspec] = ACTIONS(2848),
    [anon_sym___cdecl] = ACTIONS(2848),
    [anon_sym___clrcall] = ACTIONS(2848),
    [anon_sym___stdcall] = ACTIONS(2848),
    [anon_sym___fastcall] = ACTIONS(2848),
    [anon_sym___thiscall] = ACTIONS(2848),
    [anon_sym___vectorcall] = ACTIONS(2848),
    [anon_sym_LBRACE] = ACTIONS(2850),
    [anon_sym_signed] = ACTIONS(2848),
    [anon_sym_unsigned] = ACTIONS(2848),
    [anon_sym_long] = ACTIONS(2848),
    [anon_sym_short] = ACTIONS(2848),
    [anon_sym_static] = ACTIONS(2848),
    [anon_sym_ATautoreleasepool] = ACTIONS(2850),
    [anon_sym_auto] = ACTIONS(2848),
    [anon_sym_register] = ACTIONS(2848),
    [anon_sym_inline] = ACTIONS(2848),
    [anon_sym___inline] = ACTIONS(2848),
    [anon_sym___inline__] = ACTIONS(2848),
    [anon_sym___forceinline] = ACTIONS(2848),
    [anon_sym_thread_local] = ACTIONS(2848),
    [anon_sym___thread] = ACTIONS(2848),
    [anon_sym_CG_EXTERN] = ACTIONS(2848),
    [anon_sym_CG_INLINE] = ACTIONS(2848),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2848),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2848),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2848),
    [anon_sym_IBOutlet] = ACTIONS(2848),
    [anon_sym_IBInspectable] = ACTIONS(2848),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2848),
    [anon_sym_NS_INLINE] = ACTIONS(2848),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2848),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2848),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2848),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2848),
    [anon_sym_const] = ACTIONS(2848),
    [anon_sym_constexpr] = ACTIONS(2848),
    [anon_sym_volatile] = ACTIONS(2848),
    [anon_sym_restrict] = ACTIONS(2848),
    [anon_sym___restrict__] = ACTIONS(2848),
    [anon_sym__Atomic] = ACTIONS(2848),
    [anon_sym__Noreturn] = ACTIONS(2848),
    [anon_sym__Nonnull] = ACTIONS(2848),
    [anon_sym_nullable] = ACTIONS(2848),
    [anon_sym__Complex] = ACTIONS(2848),
    [anon_sym__Nullable] = ACTIONS(2848),
    [anon_sym__Nullable_result] = ACTIONS(2848),
    [anon_sym__Null_unspecified] = ACTIONS(2848),
    [anon_sym___autoreleasing] = ACTIONS(2848),
    [anon_sym___block] = ACTIONS(2848),
    [anon_sym___bridge] = ACTIONS(2848),
    [anon_sym___bridge_retained] = ACTIONS(2848),
    [anon_sym___bridge_transfer] = ACTIONS(2848),
    [anon_sym___complex] = ACTIONS(2848),
    [anon_sym___const] = ACTIONS(2848),
    [anon_sym___imag] = ACTIONS(2848),
    [anon_sym___kindof] = ACTIONS(2848),
    [anon_sym___nonnull] = ACTIONS(2848),
    [anon_sym___nullable] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2848),
    [anon_sym___real] = ACTIONS(2848),
    [anon_sym___strong] = ACTIONS(2848),
    [anon_sym___unsafe_unretained] = ACTIONS(2848),
    [anon_sym___unused] = ACTIONS(2848),
    [anon_sym___weak] = ACTIONS(2848),
    [anon_sym_alignas] = ACTIONS(2848),
    [anon_sym__Alignas] = ACTIONS(2848),
    [sym_primitive_type] = ACTIONS(2848),
    [anon_sym_enum] = ACTIONS(2848),
    [anon_sym_struct] = ACTIONS(2848),
    [anon_sym_union] = ACTIONS(2848),
    [anon_sym_if] = ACTIONS(2848),
    [anon_sym_switch] = ACTIONS(2848),
    [anon_sym_case] = ACTIONS(2848),
    [anon_sym_default] = ACTIONS(2848),
    [anon_sym_while] = ACTIONS(2848),
    [anon_sym_do] = ACTIONS(2848),
    [anon_sym_for] = ACTIONS(2848),
    [anon_sym_in] = ACTIONS(2848),
    [anon_sym_return] = ACTIONS(2848),
    [anon_sym_break] = ACTIONS(2848),
    [anon_sym_continue] = ACTIONS(2848),
    [anon_sym_goto] = ACTIONS(2848),
    [anon_sym_DASH_DASH] = ACTIONS(2850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2850),
    [anon_sym_sizeof] = ACTIONS(2848),
    [anon_sym___alignof__] = ACTIONS(2848),
    [anon_sym___alignof] = ACTIONS(2848),
    [anon_sym__alignof] = ACTIONS(2848),
    [anon_sym_alignof] = ACTIONS(2848),
    [anon_sym__Alignof] = ACTIONS(2848),
    [anon_sym_offsetof] = ACTIONS(2848),
    [anon_sym__Generic] = ACTIONS(2848),
    [anon_sym_asm] = ACTIONS(2848),
    [anon_sym___asm__] = ACTIONS(2848),
    [anon_sym___asm] = ACTIONS(2848),
    [sym_number_literal] = ACTIONS(2850),
    [anon_sym_L_SQUOTE] = ACTIONS(2850),
    [anon_sym_u_SQUOTE] = ACTIONS(2850),
    [anon_sym_U_SQUOTE] = ACTIONS(2850),
    [anon_sym_u8_SQUOTE] = ACTIONS(2850),
    [anon_sym_SQUOTE] = ACTIONS(2850),
    [anon_sym_AT] = ACTIONS(2848),
    [anon_sym_DQUOTE] = ACTIONS(2850),
    [anon_sym_L_DQUOTE] = ACTIONS(2850),
    [anon_sym_u_DQUOTE] = ACTIONS(2850),
    [anon_sym_U_DQUOTE] = ACTIONS(2850),
    [anon_sym_u8_DQUOTE] = ACTIONS(2850),
    [sym_true] = ACTIONS(2848),
    [sym_false] = ACTIONS(2848),
    [anon_sym_NULL] = ACTIONS(2848),
    [anon_sym_nullptr] = ACTIONS(2848),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2848),
    [anon_sym___typeof] = ACTIONS(2848),
    [anon_sym_typeof] = ACTIONS(2848),
    [anon_sym_ATimport] = ACTIONS(2850),
    [aux_sym_preproc_undef_token1] = ACTIONS(2848),
    [anon_sym_POUND] = ACTIONS(2848),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2848),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2848),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2848),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2848),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2848),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2848),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2848),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_API_AVAILABLE] = ACTIONS(2848),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_API_DEPRECATED] = ACTIONS(2848),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2848),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym___deprecated_msg] = ACTIONS(2848),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2848),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2848),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2848),
    [anon_sym_ATprotocol] = ACTIONS(2850),
    [anon_sym_ATinterface] = ACTIONS(2850),
    [anon_sym_ATimplementation] = ACTIONS(2850),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2850),
    [anon_sym_ATsynthesize] = ACTIONS(2850),
    [anon_sym_ATdynamic] = ACTIONS(2850),
    [anon_sym_ATtry] = ACTIONS(2850),
    [anon_sym___try] = ACTIONS(2848),
    [anon_sym_ATthrow] = ACTIONS(2850),
    [anon_sym_ATselector] = ACTIONS(2850),
    [anon_sym_ATavailable] = ACTIONS(2850),
    [anon_sym___builtin_available] = ACTIONS(2848),
    [anon_sym_va_arg] = ACTIONS(2848),
    [anon_sym_ATencode] = ACTIONS(2850),
    [anon_sym_ATsynchronized] = ACTIONS(2850),
    [anon_sym_BOOL] = ACTIONS(2848),
    [anon_sym_IMP] = ACTIONS(2848),
    [anon_sym_SEL] = ACTIONS(2848),
    [anon_sym_Class] = ACTIONS(2848),
    [anon_sym_id] = ACTIONS(2848),
  },
  [STATE(635)] = {
    [sym_identifier] = ACTIONS(2572),
    [aux_sym_preproc_include_token1] = ACTIONS(2572),
    [aux_sym_preproc_include_token2] = ACTIONS(2572),
    [aux_sym_preproc_def_token1] = ACTIONS(2572),
    [aux_sym_preproc_if_token1] = ACTIONS(2572),
    [aux_sym_preproc_if_token2] = ACTIONS(2572),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2572),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2572),
    [sym_preproc_directive] = ACTIONS(2572),
    [anon_sym_LPAREN2] = ACTIONS(2574),
    [anon_sym_BANG] = ACTIONS(2574),
    [anon_sym_TILDE] = ACTIONS(2574),
    [anon_sym_DASH] = ACTIONS(2572),
    [anon_sym_PLUS] = ACTIONS(2572),
    [anon_sym_STAR] = ACTIONS(2574),
    [anon_sym_CARET] = ACTIONS(2574),
    [anon_sym_AMP] = ACTIONS(2574),
    [anon_sym_SEMI] = ACTIONS(2574),
    [anon_sym___extension__] = ACTIONS(2572),
    [anon_sym_typedef] = ACTIONS(2572),
    [anon_sym_extern] = ACTIONS(2572),
    [anon_sym___attribute__] = ACTIONS(2572),
    [anon_sym___attribute] = ACTIONS(2572),
    [anon_sym_noreturn] = ACTIONS(2572),
    [anon_sym_LBRACK] = ACTIONS(2574),
    [anon_sym___declspec] = ACTIONS(2572),
    [anon_sym___cdecl] = ACTIONS(2572),
    [anon_sym___clrcall] = ACTIONS(2572),
    [anon_sym___stdcall] = ACTIONS(2572),
    [anon_sym___fastcall] = ACTIONS(2572),
    [anon_sym___thiscall] = ACTIONS(2572),
    [anon_sym___vectorcall] = ACTIONS(2572),
    [anon_sym_LBRACE] = ACTIONS(2574),
    [anon_sym_signed] = ACTIONS(2572),
    [anon_sym_unsigned] = ACTIONS(2572),
    [anon_sym_long] = ACTIONS(2572),
    [anon_sym_short] = ACTIONS(2572),
    [anon_sym_static] = ACTIONS(2572),
    [anon_sym_ATautoreleasepool] = ACTIONS(2574),
    [anon_sym_auto] = ACTIONS(2572),
    [anon_sym_register] = ACTIONS(2572),
    [anon_sym_inline] = ACTIONS(2572),
    [anon_sym___inline] = ACTIONS(2572),
    [anon_sym___inline__] = ACTIONS(2572),
    [anon_sym___forceinline] = ACTIONS(2572),
    [anon_sym_thread_local] = ACTIONS(2572),
    [anon_sym___thread] = ACTIONS(2572),
    [anon_sym_CG_EXTERN] = ACTIONS(2572),
    [anon_sym_CG_INLINE] = ACTIONS(2572),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2572),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2572),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2572),
    [anon_sym_IBOutlet] = ACTIONS(2572),
    [anon_sym_IBInspectable] = ACTIONS(2572),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2572),
    [anon_sym_NS_INLINE] = ACTIONS(2572),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2572),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2572),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2572),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2572),
    [anon_sym_const] = ACTIONS(2572),
    [anon_sym_constexpr] = ACTIONS(2572),
    [anon_sym_volatile] = ACTIONS(2572),
    [anon_sym_restrict] = ACTIONS(2572),
    [anon_sym___restrict__] = ACTIONS(2572),
    [anon_sym__Atomic] = ACTIONS(2572),
    [anon_sym__Noreturn] = ACTIONS(2572),
    [anon_sym__Nonnull] = ACTIONS(2572),
    [anon_sym_nullable] = ACTIONS(2572),
    [anon_sym__Complex] = ACTIONS(2572),
    [anon_sym__Nullable] = ACTIONS(2572),
    [anon_sym__Nullable_result] = ACTIONS(2572),
    [anon_sym__Null_unspecified] = ACTIONS(2572),
    [anon_sym___autoreleasing] = ACTIONS(2572),
    [anon_sym___block] = ACTIONS(2572),
    [anon_sym___bridge] = ACTIONS(2572),
    [anon_sym___bridge_retained] = ACTIONS(2572),
    [anon_sym___bridge_transfer] = ACTIONS(2572),
    [anon_sym___complex] = ACTIONS(2572),
    [anon_sym___const] = ACTIONS(2572),
    [anon_sym___imag] = ACTIONS(2572),
    [anon_sym___kindof] = ACTIONS(2572),
    [anon_sym___nonnull] = ACTIONS(2572),
    [anon_sym___nullable] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2572),
    [anon_sym___real] = ACTIONS(2572),
    [anon_sym___strong] = ACTIONS(2572),
    [anon_sym___unsafe_unretained] = ACTIONS(2572),
    [anon_sym___unused] = ACTIONS(2572),
    [anon_sym___weak] = ACTIONS(2572),
    [anon_sym_alignas] = ACTIONS(2572),
    [anon_sym__Alignas] = ACTIONS(2572),
    [sym_primitive_type] = ACTIONS(2572),
    [anon_sym_enum] = ACTIONS(2572),
    [anon_sym_struct] = ACTIONS(2572),
    [anon_sym_union] = ACTIONS(2572),
    [anon_sym_if] = ACTIONS(2572),
    [anon_sym_switch] = ACTIONS(2572),
    [anon_sym_case] = ACTIONS(2572),
    [anon_sym_default] = ACTIONS(2572),
    [anon_sym_while] = ACTIONS(2572),
    [anon_sym_do] = ACTIONS(2572),
    [anon_sym_for] = ACTIONS(2572),
    [anon_sym_in] = ACTIONS(2572),
    [anon_sym_return] = ACTIONS(2572),
    [anon_sym_break] = ACTIONS(2572),
    [anon_sym_continue] = ACTIONS(2572),
    [anon_sym_goto] = ACTIONS(2572),
    [anon_sym_DASH_DASH] = ACTIONS(2574),
    [anon_sym_PLUS_PLUS] = ACTIONS(2574),
    [anon_sym_sizeof] = ACTIONS(2572),
    [anon_sym___alignof__] = ACTIONS(2572),
    [anon_sym___alignof] = ACTIONS(2572),
    [anon_sym__alignof] = ACTIONS(2572),
    [anon_sym_alignof] = ACTIONS(2572),
    [anon_sym__Alignof] = ACTIONS(2572),
    [anon_sym_offsetof] = ACTIONS(2572),
    [anon_sym__Generic] = ACTIONS(2572),
    [anon_sym_asm] = ACTIONS(2572),
    [anon_sym___asm__] = ACTIONS(2572),
    [anon_sym___asm] = ACTIONS(2572),
    [sym_number_literal] = ACTIONS(2574),
    [anon_sym_L_SQUOTE] = ACTIONS(2574),
    [anon_sym_u_SQUOTE] = ACTIONS(2574),
    [anon_sym_U_SQUOTE] = ACTIONS(2574),
    [anon_sym_u8_SQUOTE] = ACTIONS(2574),
    [anon_sym_SQUOTE] = ACTIONS(2574),
    [anon_sym_AT] = ACTIONS(2572),
    [anon_sym_DQUOTE] = ACTIONS(2574),
    [anon_sym_L_DQUOTE] = ACTIONS(2574),
    [anon_sym_u_DQUOTE] = ACTIONS(2574),
    [anon_sym_U_DQUOTE] = ACTIONS(2574),
    [anon_sym_u8_DQUOTE] = ACTIONS(2574),
    [sym_true] = ACTIONS(2572),
    [sym_false] = ACTIONS(2572),
    [anon_sym_NULL] = ACTIONS(2572),
    [anon_sym_nullptr] = ACTIONS(2572),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2572),
    [anon_sym___typeof] = ACTIONS(2572),
    [anon_sym_typeof] = ACTIONS(2572),
    [anon_sym_ATimport] = ACTIONS(2574),
    [aux_sym_preproc_undef_token1] = ACTIONS(2572),
    [anon_sym_POUND] = ACTIONS(2572),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2572),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2572),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2572),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2572),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2572),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2572),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2572),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_API_AVAILABLE] = ACTIONS(2572),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_API_DEPRECATED] = ACTIONS(2572),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2572),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym___deprecated_msg] = ACTIONS(2572),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2572),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2572),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2572),
    [anon_sym_ATprotocol] = ACTIONS(2574),
    [anon_sym_ATinterface] = ACTIONS(2574),
    [anon_sym_ATimplementation] = ACTIONS(2574),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2574),
    [anon_sym_ATsynthesize] = ACTIONS(2574),
    [anon_sym_ATdynamic] = ACTIONS(2574),
    [anon_sym_ATtry] = ACTIONS(2574),
    [anon_sym___try] = ACTIONS(2572),
    [anon_sym_ATthrow] = ACTIONS(2574),
    [anon_sym_ATselector] = ACTIONS(2574),
    [anon_sym_ATavailable] = ACTIONS(2574),
    [anon_sym___builtin_available] = ACTIONS(2572),
    [anon_sym_va_arg] = ACTIONS(2572),
    [anon_sym_ATencode] = ACTIONS(2574),
    [anon_sym_ATsynchronized] = ACTIONS(2574),
    [anon_sym_BOOL] = ACTIONS(2572),
    [anon_sym_IMP] = ACTIONS(2572),
    [anon_sym_SEL] = ACTIONS(2572),
    [anon_sym_Class] = ACTIONS(2572),
    [anon_sym_id] = ACTIONS(2572),
  },
  [STATE(636)] = {
    [sym_identifier] = ACTIONS(2576),
    [aux_sym_preproc_include_token1] = ACTIONS(2576),
    [aux_sym_preproc_include_token2] = ACTIONS(2576),
    [aux_sym_preproc_def_token1] = ACTIONS(2576),
    [aux_sym_preproc_if_token1] = ACTIONS(2576),
    [aux_sym_preproc_if_token2] = ACTIONS(2576),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2576),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2576),
    [sym_preproc_directive] = ACTIONS(2576),
    [anon_sym_LPAREN2] = ACTIONS(2578),
    [anon_sym_BANG] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2578),
    [anon_sym_DASH] = ACTIONS(2576),
    [anon_sym_PLUS] = ACTIONS(2576),
    [anon_sym_STAR] = ACTIONS(2578),
    [anon_sym_CARET] = ACTIONS(2578),
    [anon_sym_AMP] = ACTIONS(2578),
    [anon_sym_SEMI] = ACTIONS(2578),
    [anon_sym___extension__] = ACTIONS(2576),
    [anon_sym_typedef] = ACTIONS(2576),
    [anon_sym_extern] = ACTIONS(2576),
    [anon_sym___attribute__] = ACTIONS(2576),
    [anon_sym___attribute] = ACTIONS(2576),
    [anon_sym_noreturn] = ACTIONS(2576),
    [anon_sym_LBRACK] = ACTIONS(2578),
    [anon_sym___declspec] = ACTIONS(2576),
    [anon_sym___cdecl] = ACTIONS(2576),
    [anon_sym___clrcall] = ACTIONS(2576),
    [anon_sym___stdcall] = ACTIONS(2576),
    [anon_sym___fastcall] = ACTIONS(2576),
    [anon_sym___thiscall] = ACTIONS(2576),
    [anon_sym___vectorcall] = ACTIONS(2576),
    [anon_sym_LBRACE] = ACTIONS(2578),
    [anon_sym_signed] = ACTIONS(2576),
    [anon_sym_unsigned] = ACTIONS(2576),
    [anon_sym_long] = ACTIONS(2576),
    [anon_sym_short] = ACTIONS(2576),
    [anon_sym_static] = ACTIONS(2576),
    [anon_sym_ATautoreleasepool] = ACTIONS(2578),
    [anon_sym_auto] = ACTIONS(2576),
    [anon_sym_register] = ACTIONS(2576),
    [anon_sym_inline] = ACTIONS(2576),
    [anon_sym___inline] = ACTIONS(2576),
    [anon_sym___inline__] = ACTIONS(2576),
    [anon_sym___forceinline] = ACTIONS(2576),
    [anon_sym_thread_local] = ACTIONS(2576),
    [anon_sym___thread] = ACTIONS(2576),
    [anon_sym_CG_EXTERN] = ACTIONS(2576),
    [anon_sym_CG_INLINE] = ACTIONS(2576),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2576),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2576),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2576),
    [anon_sym_IBOutlet] = ACTIONS(2576),
    [anon_sym_IBInspectable] = ACTIONS(2576),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2576),
    [anon_sym_NS_INLINE] = ACTIONS(2576),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2576),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2576),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2576),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2576),
    [anon_sym_const] = ACTIONS(2576),
    [anon_sym_constexpr] = ACTIONS(2576),
    [anon_sym_volatile] = ACTIONS(2576),
    [anon_sym_restrict] = ACTIONS(2576),
    [anon_sym___restrict__] = ACTIONS(2576),
    [anon_sym__Atomic] = ACTIONS(2576),
    [anon_sym__Noreturn] = ACTIONS(2576),
    [anon_sym__Nonnull] = ACTIONS(2576),
    [anon_sym_nullable] = ACTIONS(2576),
    [anon_sym__Complex] = ACTIONS(2576),
    [anon_sym__Nullable] = ACTIONS(2576),
    [anon_sym__Nullable_result] = ACTIONS(2576),
    [anon_sym__Null_unspecified] = ACTIONS(2576),
    [anon_sym___autoreleasing] = ACTIONS(2576),
    [anon_sym___block] = ACTIONS(2576),
    [anon_sym___bridge] = ACTIONS(2576),
    [anon_sym___bridge_retained] = ACTIONS(2576),
    [anon_sym___bridge_transfer] = ACTIONS(2576),
    [anon_sym___complex] = ACTIONS(2576),
    [anon_sym___const] = ACTIONS(2576),
    [anon_sym___imag] = ACTIONS(2576),
    [anon_sym___kindof] = ACTIONS(2576),
    [anon_sym___nonnull] = ACTIONS(2576),
    [anon_sym___nullable] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2576),
    [anon_sym___real] = ACTIONS(2576),
    [anon_sym___strong] = ACTIONS(2576),
    [anon_sym___unsafe_unretained] = ACTIONS(2576),
    [anon_sym___unused] = ACTIONS(2576),
    [anon_sym___weak] = ACTIONS(2576),
    [anon_sym_alignas] = ACTIONS(2576),
    [anon_sym__Alignas] = ACTIONS(2576),
    [sym_primitive_type] = ACTIONS(2576),
    [anon_sym_enum] = ACTIONS(2576),
    [anon_sym_struct] = ACTIONS(2576),
    [anon_sym_union] = ACTIONS(2576),
    [anon_sym_if] = ACTIONS(2576),
    [anon_sym_switch] = ACTIONS(2576),
    [anon_sym_case] = ACTIONS(2576),
    [anon_sym_default] = ACTIONS(2576),
    [anon_sym_while] = ACTIONS(2576),
    [anon_sym_do] = ACTIONS(2576),
    [anon_sym_for] = ACTIONS(2576),
    [anon_sym_in] = ACTIONS(2576),
    [anon_sym_return] = ACTIONS(2576),
    [anon_sym_break] = ACTIONS(2576),
    [anon_sym_continue] = ACTIONS(2576),
    [anon_sym_goto] = ACTIONS(2576),
    [anon_sym_DASH_DASH] = ACTIONS(2578),
    [anon_sym_PLUS_PLUS] = ACTIONS(2578),
    [anon_sym_sizeof] = ACTIONS(2576),
    [anon_sym___alignof__] = ACTIONS(2576),
    [anon_sym___alignof] = ACTIONS(2576),
    [anon_sym__alignof] = ACTIONS(2576),
    [anon_sym_alignof] = ACTIONS(2576),
    [anon_sym__Alignof] = ACTIONS(2576),
    [anon_sym_offsetof] = ACTIONS(2576),
    [anon_sym__Generic] = ACTIONS(2576),
    [anon_sym_asm] = ACTIONS(2576),
    [anon_sym___asm__] = ACTIONS(2576),
    [anon_sym___asm] = ACTIONS(2576),
    [sym_number_literal] = ACTIONS(2578),
    [anon_sym_L_SQUOTE] = ACTIONS(2578),
    [anon_sym_u_SQUOTE] = ACTIONS(2578),
    [anon_sym_U_SQUOTE] = ACTIONS(2578),
    [anon_sym_u8_SQUOTE] = ACTIONS(2578),
    [anon_sym_SQUOTE] = ACTIONS(2578),
    [anon_sym_AT] = ACTIONS(2576),
    [anon_sym_DQUOTE] = ACTIONS(2578),
    [anon_sym_L_DQUOTE] = ACTIONS(2578),
    [anon_sym_u_DQUOTE] = ACTIONS(2578),
    [anon_sym_U_DQUOTE] = ACTIONS(2578),
    [anon_sym_u8_DQUOTE] = ACTIONS(2578),
    [sym_true] = ACTIONS(2576),
    [sym_false] = ACTIONS(2576),
    [anon_sym_NULL] = ACTIONS(2576),
    [anon_sym_nullptr] = ACTIONS(2576),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2576),
    [anon_sym___typeof] = ACTIONS(2576),
    [anon_sym_typeof] = ACTIONS(2576),
    [anon_sym_ATimport] = ACTIONS(2578),
    [aux_sym_preproc_undef_token1] = ACTIONS(2576),
    [anon_sym_POUND] = ACTIONS(2576),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2576),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2576),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2576),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2576),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2576),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2576),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2576),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_API_AVAILABLE] = ACTIONS(2576),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_API_DEPRECATED] = ACTIONS(2576),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2576),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym___deprecated_msg] = ACTIONS(2576),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2576),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2576),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2576),
    [anon_sym_ATprotocol] = ACTIONS(2578),
    [anon_sym_ATinterface] = ACTIONS(2578),
    [anon_sym_ATimplementation] = ACTIONS(2578),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2578),
    [anon_sym_ATsynthesize] = ACTIONS(2578),
    [anon_sym_ATdynamic] = ACTIONS(2578),
    [anon_sym_ATtry] = ACTIONS(2578),
    [anon_sym___try] = ACTIONS(2576),
    [anon_sym_ATthrow] = ACTIONS(2578),
    [anon_sym_ATselector] = ACTIONS(2578),
    [anon_sym_ATavailable] = ACTIONS(2578),
    [anon_sym___builtin_available] = ACTIONS(2576),
    [anon_sym_va_arg] = ACTIONS(2576),
    [anon_sym_ATencode] = ACTIONS(2578),
    [anon_sym_ATsynchronized] = ACTIONS(2578),
    [anon_sym_BOOL] = ACTIONS(2576),
    [anon_sym_IMP] = ACTIONS(2576),
    [anon_sym_SEL] = ACTIONS(2576),
    [anon_sym_Class] = ACTIONS(2576),
    [anon_sym_id] = ACTIONS(2576),
  },
  [STATE(637)] = {
    [sym_identifier] = ACTIONS(2592),
    [aux_sym_preproc_include_token1] = ACTIONS(2592),
    [aux_sym_preproc_include_token2] = ACTIONS(2592),
    [aux_sym_preproc_def_token1] = ACTIONS(2592),
    [aux_sym_preproc_if_token1] = ACTIONS(2592),
    [aux_sym_preproc_if_token2] = ACTIONS(2592),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2592),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2592),
    [sym_preproc_directive] = ACTIONS(2592),
    [anon_sym_LPAREN2] = ACTIONS(2594),
    [anon_sym_BANG] = ACTIONS(2594),
    [anon_sym_TILDE] = ACTIONS(2594),
    [anon_sym_DASH] = ACTIONS(2592),
    [anon_sym_PLUS] = ACTIONS(2592),
    [anon_sym_STAR] = ACTIONS(2594),
    [anon_sym_CARET] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(2594),
    [anon_sym_SEMI] = ACTIONS(2594),
    [anon_sym___extension__] = ACTIONS(2592),
    [anon_sym_typedef] = ACTIONS(2592),
    [anon_sym_extern] = ACTIONS(2592),
    [anon_sym___attribute__] = ACTIONS(2592),
    [anon_sym___attribute] = ACTIONS(2592),
    [anon_sym_noreturn] = ACTIONS(2592),
    [anon_sym_LBRACK] = ACTIONS(2594),
    [anon_sym___declspec] = ACTIONS(2592),
    [anon_sym___cdecl] = ACTIONS(2592),
    [anon_sym___clrcall] = ACTIONS(2592),
    [anon_sym___stdcall] = ACTIONS(2592),
    [anon_sym___fastcall] = ACTIONS(2592),
    [anon_sym___thiscall] = ACTIONS(2592),
    [anon_sym___vectorcall] = ACTIONS(2592),
    [anon_sym_LBRACE] = ACTIONS(2594),
    [anon_sym_signed] = ACTIONS(2592),
    [anon_sym_unsigned] = ACTIONS(2592),
    [anon_sym_long] = ACTIONS(2592),
    [anon_sym_short] = ACTIONS(2592),
    [anon_sym_static] = ACTIONS(2592),
    [anon_sym_ATautoreleasepool] = ACTIONS(2594),
    [anon_sym_auto] = ACTIONS(2592),
    [anon_sym_register] = ACTIONS(2592),
    [anon_sym_inline] = ACTIONS(2592),
    [anon_sym___inline] = ACTIONS(2592),
    [anon_sym___inline__] = ACTIONS(2592),
    [anon_sym___forceinline] = ACTIONS(2592),
    [anon_sym_thread_local] = ACTIONS(2592),
    [anon_sym___thread] = ACTIONS(2592),
    [anon_sym_CG_EXTERN] = ACTIONS(2592),
    [anon_sym_CG_INLINE] = ACTIONS(2592),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2592),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2592),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2592),
    [anon_sym_IBOutlet] = ACTIONS(2592),
    [anon_sym_IBInspectable] = ACTIONS(2592),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2592),
    [anon_sym_NS_INLINE] = ACTIONS(2592),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2592),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2592),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2592),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2592),
    [anon_sym_const] = ACTIONS(2592),
    [anon_sym_constexpr] = ACTIONS(2592),
    [anon_sym_volatile] = ACTIONS(2592),
    [anon_sym_restrict] = ACTIONS(2592),
    [anon_sym___restrict__] = ACTIONS(2592),
    [anon_sym__Atomic] = ACTIONS(2592),
    [anon_sym__Noreturn] = ACTIONS(2592),
    [anon_sym__Nonnull] = ACTIONS(2592),
    [anon_sym_nullable] = ACTIONS(2592),
    [anon_sym__Complex] = ACTIONS(2592),
    [anon_sym__Nullable] = ACTIONS(2592),
    [anon_sym__Nullable_result] = ACTIONS(2592),
    [anon_sym__Null_unspecified] = ACTIONS(2592),
    [anon_sym___autoreleasing] = ACTIONS(2592),
    [anon_sym___block] = ACTIONS(2592),
    [anon_sym___bridge] = ACTIONS(2592),
    [anon_sym___bridge_retained] = ACTIONS(2592),
    [anon_sym___bridge_transfer] = ACTIONS(2592),
    [anon_sym___complex] = ACTIONS(2592),
    [anon_sym___const] = ACTIONS(2592),
    [anon_sym___imag] = ACTIONS(2592),
    [anon_sym___kindof] = ACTIONS(2592),
    [anon_sym___nonnull] = ACTIONS(2592),
    [anon_sym___nullable] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2592),
    [anon_sym___real] = ACTIONS(2592),
    [anon_sym___strong] = ACTIONS(2592),
    [anon_sym___unsafe_unretained] = ACTIONS(2592),
    [anon_sym___unused] = ACTIONS(2592),
    [anon_sym___weak] = ACTIONS(2592),
    [anon_sym_alignas] = ACTIONS(2592),
    [anon_sym__Alignas] = ACTIONS(2592),
    [sym_primitive_type] = ACTIONS(2592),
    [anon_sym_enum] = ACTIONS(2592),
    [anon_sym_struct] = ACTIONS(2592),
    [anon_sym_union] = ACTIONS(2592),
    [anon_sym_if] = ACTIONS(2592),
    [anon_sym_switch] = ACTIONS(2592),
    [anon_sym_case] = ACTIONS(2592),
    [anon_sym_default] = ACTIONS(2592),
    [anon_sym_while] = ACTIONS(2592),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_for] = ACTIONS(2592),
    [anon_sym_in] = ACTIONS(2592),
    [anon_sym_return] = ACTIONS(2592),
    [anon_sym_break] = ACTIONS(2592),
    [anon_sym_continue] = ACTIONS(2592),
    [anon_sym_goto] = ACTIONS(2592),
    [anon_sym_DASH_DASH] = ACTIONS(2594),
    [anon_sym_PLUS_PLUS] = ACTIONS(2594),
    [anon_sym_sizeof] = ACTIONS(2592),
    [anon_sym___alignof__] = ACTIONS(2592),
    [anon_sym___alignof] = ACTIONS(2592),
    [anon_sym__alignof] = ACTIONS(2592),
    [anon_sym_alignof] = ACTIONS(2592),
    [anon_sym__Alignof] = ACTIONS(2592),
    [anon_sym_offsetof] = ACTIONS(2592),
    [anon_sym__Generic] = ACTIONS(2592),
    [anon_sym_asm] = ACTIONS(2592),
    [anon_sym___asm__] = ACTIONS(2592),
    [anon_sym___asm] = ACTIONS(2592),
    [sym_number_literal] = ACTIONS(2594),
    [anon_sym_L_SQUOTE] = ACTIONS(2594),
    [anon_sym_u_SQUOTE] = ACTIONS(2594),
    [anon_sym_U_SQUOTE] = ACTIONS(2594),
    [anon_sym_u8_SQUOTE] = ACTIONS(2594),
    [anon_sym_SQUOTE] = ACTIONS(2594),
    [anon_sym_AT] = ACTIONS(2592),
    [anon_sym_DQUOTE] = ACTIONS(2594),
    [anon_sym_L_DQUOTE] = ACTIONS(2594),
    [anon_sym_u_DQUOTE] = ACTIONS(2594),
    [anon_sym_U_DQUOTE] = ACTIONS(2594),
    [anon_sym_u8_DQUOTE] = ACTIONS(2594),
    [sym_true] = ACTIONS(2592),
    [sym_false] = ACTIONS(2592),
    [anon_sym_NULL] = ACTIONS(2592),
    [anon_sym_nullptr] = ACTIONS(2592),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2592),
    [anon_sym___typeof] = ACTIONS(2592),
    [anon_sym_typeof] = ACTIONS(2592),
    [anon_sym_ATimport] = ACTIONS(2594),
    [aux_sym_preproc_undef_token1] = ACTIONS(2592),
    [anon_sym_POUND] = ACTIONS(2592),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2592),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2592),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2592),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2592),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2592),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2592),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2592),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_API_AVAILABLE] = ACTIONS(2592),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_API_DEPRECATED] = ACTIONS(2592),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2592),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym___deprecated_msg] = ACTIONS(2592),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2592),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2592),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2592),
    [anon_sym_ATprotocol] = ACTIONS(2594),
    [anon_sym_ATinterface] = ACTIONS(2594),
    [anon_sym_ATimplementation] = ACTIONS(2594),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2594),
    [anon_sym_ATsynthesize] = ACTIONS(2594),
    [anon_sym_ATdynamic] = ACTIONS(2594),
    [anon_sym_ATtry] = ACTIONS(2594),
    [anon_sym___try] = ACTIONS(2592),
    [anon_sym_ATthrow] = ACTIONS(2594),
    [anon_sym_ATselector] = ACTIONS(2594),
    [anon_sym_ATavailable] = ACTIONS(2594),
    [anon_sym___builtin_available] = ACTIONS(2592),
    [anon_sym_va_arg] = ACTIONS(2592),
    [anon_sym_ATencode] = ACTIONS(2594),
    [anon_sym_ATsynchronized] = ACTIONS(2594),
    [anon_sym_BOOL] = ACTIONS(2592),
    [anon_sym_IMP] = ACTIONS(2592),
    [anon_sym_SEL] = ACTIONS(2592),
    [anon_sym_Class] = ACTIONS(2592),
    [anon_sym_id] = ACTIONS(2592),
  },
  [STATE(638)] = {
    [sym_identifier] = ACTIONS(2860),
    [aux_sym_preproc_include_token1] = ACTIONS(2860),
    [aux_sym_preproc_include_token2] = ACTIONS(2860),
    [aux_sym_preproc_def_token1] = ACTIONS(2860),
    [aux_sym_preproc_if_token1] = ACTIONS(2860),
    [aux_sym_preproc_if_token2] = ACTIONS(2860),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2860),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2860),
    [sym_preproc_directive] = ACTIONS(2860),
    [anon_sym_LPAREN2] = ACTIONS(2862),
    [anon_sym_BANG] = ACTIONS(2862),
    [anon_sym_TILDE] = ACTIONS(2862),
    [anon_sym_DASH] = ACTIONS(2860),
    [anon_sym_PLUS] = ACTIONS(2860),
    [anon_sym_STAR] = ACTIONS(2862),
    [anon_sym_CARET] = ACTIONS(2862),
    [anon_sym_AMP] = ACTIONS(2862),
    [anon_sym_SEMI] = ACTIONS(2862),
    [anon_sym___extension__] = ACTIONS(2860),
    [anon_sym_typedef] = ACTIONS(2860),
    [anon_sym_extern] = ACTIONS(2860),
    [anon_sym___attribute__] = ACTIONS(2860),
    [anon_sym___attribute] = ACTIONS(2860),
    [anon_sym_noreturn] = ACTIONS(2860),
    [anon_sym_LBRACK] = ACTIONS(2862),
    [anon_sym___declspec] = ACTIONS(2860),
    [anon_sym___cdecl] = ACTIONS(2860),
    [anon_sym___clrcall] = ACTIONS(2860),
    [anon_sym___stdcall] = ACTIONS(2860),
    [anon_sym___fastcall] = ACTIONS(2860),
    [anon_sym___thiscall] = ACTIONS(2860),
    [anon_sym___vectorcall] = ACTIONS(2860),
    [anon_sym_LBRACE] = ACTIONS(2862),
    [anon_sym_signed] = ACTIONS(2860),
    [anon_sym_unsigned] = ACTIONS(2860),
    [anon_sym_long] = ACTIONS(2860),
    [anon_sym_short] = ACTIONS(2860),
    [anon_sym_static] = ACTIONS(2860),
    [anon_sym_ATautoreleasepool] = ACTIONS(2862),
    [anon_sym_auto] = ACTIONS(2860),
    [anon_sym_register] = ACTIONS(2860),
    [anon_sym_inline] = ACTIONS(2860),
    [anon_sym___inline] = ACTIONS(2860),
    [anon_sym___inline__] = ACTIONS(2860),
    [anon_sym___forceinline] = ACTIONS(2860),
    [anon_sym_thread_local] = ACTIONS(2860),
    [anon_sym___thread] = ACTIONS(2860),
    [anon_sym_CG_EXTERN] = ACTIONS(2860),
    [anon_sym_CG_INLINE] = ACTIONS(2860),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2860),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2860),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2860),
    [anon_sym_IBOutlet] = ACTIONS(2860),
    [anon_sym_IBInspectable] = ACTIONS(2860),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2860),
    [anon_sym_NS_INLINE] = ACTIONS(2860),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2860),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2860),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2860),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2860),
    [anon_sym_const] = ACTIONS(2860),
    [anon_sym_constexpr] = ACTIONS(2860),
    [anon_sym_volatile] = ACTIONS(2860),
    [anon_sym_restrict] = ACTIONS(2860),
    [anon_sym___restrict__] = ACTIONS(2860),
    [anon_sym__Atomic] = ACTIONS(2860),
    [anon_sym__Noreturn] = ACTIONS(2860),
    [anon_sym__Nonnull] = ACTIONS(2860),
    [anon_sym_nullable] = ACTIONS(2860),
    [anon_sym__Complex] = ACTIONS(2860),
    [anon_sym__Nullable] = ACTIONS(2860),
    [anon_sym__Nullable_result] = ACTIONS(2860),
    [anon_sym__Null_unspecified] = ACTIONS(2860),
    [anon_sym___autoreleasing] = ACTIONS(2860),
    [anon_sym___block] = ACTIONS(2860),
    [anon_sym___bridge] = ACTIONS(2860),
    [anon_sym___bridge_retained] = ACTIONS(2860),
    [anon_sym___bridge_transfer] = ACTIONS(2860),
    [anon_sym___complex] = ACTIONS(2860),
    [anon_sym___const] = ACTIONS(2860),
    [anon_sym___imag] = ACTIONS(2860),
    [anon_sym___kindof] = ACTIONS(2860),
    [anon_sym___nonnull] = ACTIONS(2860),
    [anon_sym___nullable] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2860),
    [anon_sym___real] = ACTIONS(2860),
    [anon_sym___strong] = ACTIONS(2860),
    [anon_sym___unsafe_unretained] = ACTIONS(2860),
    [anon_sym___unused] = ACTIONS(2860),
    [anon_sym___weak] = ACTIONS(2860),
    [anon_sym_alignas] = ACTIONS(2860),
    [anon_sym__Alignas] = ACTIONS(2860),
    [sym_primitive_type] = ACTIONS(2860),
    [anon_sym_enum] = ACTIONS(2860),
    [anon_sym_struct] = ACTIONS(2860),
    [anon_sym_union] = ACTIONS(2860),
    [anon_sym_if] = ACTIONS(2860),
    [anon_sym_switch] = ACTIONS(2860),
    [anon_sym_case] = ACTIONS(2860),
    [anon_sym_default] = ACTIONS(2860),
    [anon_sym_while] = ACTIONS(2860),
    [anon_sym_do] = ACTIONS(2860),
    [anon_sym_for] = ACTIONS(2860),
    [anon_sym_in] = ACTIONS(2860),
    [anon_sym_return] = ACTIONS(2860),
    [anon_sym_break] = ACTIONS(2860),
    [anon_sym_continue] = ACTIONS(2860),
    [anon_sym_goto] = ACTIONS(2860),
    [anon_sym_DASH_DASH] = ACTIONS(2862),
    [anon_sym_PLUS_PLUS] = ACTIONS(2862),
    [anon_sym_sizeof] = ACTIONS(2860),
    [anon_sym___alignof__] = ACTIONS(2860),
    [anon_sym___alignof] = ACTIONS(2860),
    [anon_sym__alignof] = ACTIONS(2860),
    [anon_sym_alignof] = ACTIONS(2860),
    [anon_sym__Alignof] = ACTIONS(2860),
    [anon_sym_offsetof] = ACTIONS(2860),
    [anon_sym__Generic] = ACTIONS(2860),
    [anon_sym_asm] = ACTIONS(2860),
    [anon_sym___asm__] = ACTIONS(2860),
    [anon_sym___asm] = ACTIONS(2860),
    [sym_number_literal] = ACTIONS(2862),
    [anon_sym_L_SQUOTE] = ACTIONS(2862),
    [anon_sym_u_SQUOTE] = ACTIONS(2862),
    [anon_sym_U_SQUOTE] = ACTIONS(2862),
    [anon_sym_u8_SQUOTE] = ACTIONS(2862),
    [anon_sym_SQUOTE] = ACTIONS(2862),
    [anon_sym_AT] = ACTIONS(2860),
    [anon_sym_DQUOTE] = ACTIONS(2862),
    [anon_sym_L_DQUOTE] = ACTIONS(2862),
    [anon_sym_u_DQUOTE] = ACTIONS(2862),
    [anon_sym_U_DQUOTE] = ACTIONS(2862),
    [anon_sym_u8_DQUOTE] = ACTIONS(2862),
    [sym_true] = ACTIONS(2860),
    [sym_false] = ACTIONS(2860),
    [anon_sym_NULL] = ACTIONS(2860),
    [anon_sym_nullptr] = ACTIONS(2860),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2860),
    [anon_sym___typeof] = ACTIONS(2860),
    [anon_sym_typeof] = ACTIONS(2860),
    [anon_sym_ATimport] = ACTIONS(2862),
    [aux_sym_preproc_undef_token1] = ACTIONS(2860),
    [anon_sym_POUND] = ACTIONS(2860),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2860),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2860),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2860),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2860),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2860),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2860),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2860),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_API_AVAILABLE] = ACTIONS(2860),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_API_DEPRECATED] = ACTIONS(2860),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2860),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym___deprecated_msg] = ACTIONS(2860),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2860),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2860),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2860),
    [anon_sym_ATprotocol] = ACTIONS(2862),
    [anon_sym_ATinterface] = ACTIONS(2862),
    [anon_sym_ATimplementation] = ACTIONS(2862),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2862),
    [anon_sym_ATsynthesize] = ACTIONS(2862),
    [anon_sym_ATdynamic] = ACTIONS(2862),
    [anon_sym_ATtry] = ACTIONS(2862),
    [anon_sym___try] = ACTIONS(2860),
    [anon_sym_ATthrow] = ACTIONS(2862),
    [anon_sym_ATselector] = ACTIONS(2862),
    [anon_sym_ATavailable] = ACTIONS(2862),
    [anon_sym___builtin_available] = ACTIONS(2860),
    [anon_sym_va_arg] = ACTIONS(2860),
    [anon_sym_ATencode] = ACTIONS(2862),
    [anon_sym_ATsynchronized] = ACTIONS(2862),
    [anon_sym_BOOL] = ACTIONS(2860),
    [anon_sym_IMP] = ACTIONS(2860),
    [anon_sym_SEL] = ACTIONS(2860),
    [anon_sym_Class] = ACTIONS(2860),
    [anon_sym_id] = ACTIONS(2860),
  },
  [STATE(639)] = {
    [sym_identifier] = ACTIONS(2580),
    [aux_sym_preproc_include_token1] = ACTIONS(2580),
    [aux_sym_preproc_include_token2] = ACTIONS(2580),
    [aux_sym_preproc_def_token1] = ACTIONS(2580),
    [aux_sym_preproc_if_token1] = ACTIONS(2580),
    [aux_sym_preproc_if_token2] = ACTIONS(2580),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2580),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2580),
    [sym_preproc_directive] = ACTIONS(2580),
    [anon_sym_LPAREN2] = ACTIONS(2582),
    [anon_sym_BANG] = ACTIONS(2582),
    [anon_sym_TILDE] = ACTIONS(2582),
    [anon_sym_DASH] = ACTIONS(2580),
    [anon_sym_PLUS] = ACTIONS(2580),
    [anon_sym_STAR] = ACTIONS(2582),
    [anon_sym_CARET] = ACTIONS(2582),
    [anon_sym_AMP] = ACTIONS(2582),
    [anon_sym_SEMI] = ACTIONS(2582),
    [anon_sym___extension__] = ACTIONS(2580),
    [anon_sym_typedef] = ACTIONS(2580),
    [anon_sym_extern] = ACTIONS(2580),
    [anon_sym___attribute__] = ACTIONS(2580),
    [anon_sym___attribute] = ACTIONS(2580),
    [anon_sym_noreturn] = ACTIONS(2580),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym___declspec] = ACTIONS(2580),
    [anon_sym___cdecl] = ACTIONS(2580),
    [anon_sym___clrcall] = ACTIONS(2580),
    [anon_sym___stdcall] = ACTIONS(2580),
    [anon_sym___fastcall] = ACTIONS(2580),
    [anon_sym___thiscall] = ACTIONS(2580),
    [anon_sym___vectorcall] = ACTIONS(2580),
    [anon_sym_LBRACE] = ACTIONS(2582),
    [anon_sym_signed] = ACTIONS(2580),
    [anon_sym_unsigned] = ACTIONS(2580),
    [anon_sym_long] = ACTIONS(2580),
    [anon_sym_short] = ACTIONS(2580),
    [anon_sym_static] = ACTIONS(2580),
    [anon_sym_ATautoreleasepool] = ACTIONS(2582),
    [anon_sym_auto] = ACTIONS(2580),
    [anon_sym_register] = ACTIONS(2580),
    [anon_sym_inline] = ACTIONS(2580),
    [anon_sym___inline] = ACTIONS(2580),
    [anon_sym___inline__] = ACTIONS(2580),
    [anon_sym___forceinline] = ACTIONS(2580),
    [anon_sym_thread_local] = ACTIONS(2580),
    [anon_sym___thread] = ACTIONS(2580),
    [anon_sym_CG_EXTERN] = ACTIONS(2580),
    [anon_sym_CG_INLINE] = ACTIONS(2580),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2580),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2580),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2580),
    [anon_sym_IBOutlet] = ACTIONS(2580),
    [anon_sym_IBInspectable] = ACTIONS(2580),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2580),
    [anon_sym_NS_INLINE] = ACTIONS(2580),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2580),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2580),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2580),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2580),
    [anon_sym_const] = ACTIONS(2580),
    [anon_sym_constexpr] = ACTIONS(2580),
    [anon_sym_volatile] = ACTIONS(2580),
    [anon_sym_restrict] = ACTIONS(2580),
    [anon_sym___restrict__] = ACTIONS(2580),
    [anon_sym__Atomic] = ACTIONS(2580),
    [anon_sym__Noreturn] = ACTIONS(2580),
    [anon_sym__Nonnull] = ACTIONS(2580),
    [anon_sym_nullable] = ACTIONS(2580),
    [anon_sym__Complex] = ACTIONS(2580),
    [anon_sym__Nullable] = ACTIONS(2580),
    [anon_sym__Nullable_result] = ACTIONS(2580),
    [anon_sym__Null_unspecified] = ACTIONS(2580),
    [anon_sym___autoreleasing] = ACTIONS(2580),
    [anon_sym___block] = ACTIONS(2580),
    [anon_sym___bridge] = ACTIONS(2580),
    [anon_sym___bridge_retained] = ACTIONS(2580),
    [anon_sym___bridge_transfer] = ACTIONS(2580),
    [anon_sym___complex] = ACTIONS(2580),
    [anon_sym___const] = ACTIONS(2580),
    [anon_sym___imag] = ACTIONS(2580),
    [anon_sym___kindof] = ACTIONS(2580),
    [anon_sym___nonnull] = ACTIONS(2580),
    [anon_sym___nullable] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2580),
    [anon_sym___real] = ACTIONS(2580),
    [anon_sym___strong] = ACTIONS(2580),
    [anon_sym___unsafe_unretained] = ACTIONS(2580),
    [anon_sym___unused] = ACTIONS(2580),
    [anon_sym___weak] = ACTIONS(2580),
    [anon_sym_alignas] = ACTIONS(2580),
    [anon_sym__Alignas] = ACTIONS(2580),
    [sym_primitive_type] = ACTIONS(2580),
    [anon_sym_enum] = ACTIONS(2580),
    [anon_sym_struct] = ACTIONS(2580),
    [anon_sym_union] = ACTIONS(2580),
    [anon_sym_if] = ACTIONS(2580),
    [anon_sym_switch] = ACTIONS(2580),
    [anon_sym_case] = ACTIONS(2580),
    [anon_sym_default] = ACTIONS(2580),
    [anon_sym_while] = ACTIONS(2580),
    [anon_sym_do] = ACTIONS(2580),
    [anon_sym_for] = ACTIONS(2580),
    [anon_sym_in] = ACTIONS(2580),
    [anon_sym_return] = ACTIONS(2580),
    [anon_sym_break] = ACTIONS(2580),
    [anon_sym_continue] = ACTIONS(2580),
    [anon_sym_goto] = ACTIONS(2580),
    [anon_sym_DASH_DASH] = ACTIONS(2582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2582),
    [anon_sym_sizeof] = ACTIONS(2580),
    [anon_sym___alignof__] = ACTIONS(2580),
    [anon_sym___alignof] = ACTIONS(2580),
    [anon_sym__alignof] = ACTIONS(2580),
    [anon_sym_alignof] = ACTIONS(2580),
    [anon_sym__Alignof] = ACTIONS(2580),
    [anon_sym_offsetof] = ACTIONS(2580),
    [anon_sym__Generic] = ACTIONS(2580),
    [anon_sym_asm] = ACTIONS(2580),
    [anon_sym___asm__] = ACTIONS(2580),
    [anon_sym___asm] = ACTIONS(2580),
    [sym_number_literal] = ACTIONS(2582),
    [anon_sym_L_SQUOTE] = ACTIONS(2582),
    [anon_sym_u_SQUOTE] = ACTIONS(2582),
    [anon_sym_U_SQUOTE] = ACTIONS(2582),
    [anon_sym_u8_SQUOTE] = ACTIONS(2582),
    [anon_sym_SQUOTE] = ACTIONS(2582),
    [anon_sym_AT] = ACTIONS(2580),
    [anon_sym_DQUOTE] = ACTIONS(2582),
    [anon_sym_L_DQUOTE] = ACTIONS(2582),
    [anon_sym_u_DQUOTE] = ACTIONS(2582),
    [anon_sym_U_DQUOTE] = ACTIONS(2582),
    [anon_sym_u8_DQUOTE] = ACTIONS(2582),
    [sym_true] = ACTIONS(2580),
    [sym_false] = ACTIONS(2580),
    [anon_sym_NULL] = ACTIONS(2580),
    [anon_sym_nullptr] = ACTIONS(2580),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2580),
    [anon_sym___typeof] = ACTIONS(2580),
    [anon_sym_typeof] = ACTIONS(2580),
    [anon_sym_ATimport] = ACTIONS(2582),
    [aux_sym_preproc_undef_token1] = ACTIONS(2580),
    [anon_sym_POUND] = ACTIONS(2580),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2580),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2580),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2580),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2580),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2580),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2580),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2580),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_API_AVAILABLE] = ACTIONS(2580),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_API_DEPRECATED] = ACTIONS(2580),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2580),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym___deprecated_msg] = ACTIONS(2580),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2580),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2580),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2580),
    [anon_sym_ATprotocol] = ACTIONS(2582),
    [anon_sym_ATinterface] = ACTIONS(2582),
    [anon_sym_ATimplementation] = ACTIONS(2582),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2582),
    [anon_sym_ATsynthesize] = ACTIONS(2582),
    [anon_sym_ATdynamic] = ACTIONS(2582),
    [anon_sym_ATtry] = ACTIONS(2582),
    [anon_sym___try] = ACTIONS(2580),
    [anon_sym_ATthrow] = ACTIONS(2582),
    [anon_sym_ATselector] = ACTIONS(2582),
    [anon_sym_ATavailable] = ACTIONS(2582),
    [anon_sym___builtin_available] = ACTIONS(2580),
    [anon_sym_va_arg] = ACTIONS(2580),
    [anon_sym_ATencode] = ACTIONS(2582),
    [anon_sym_ATsynchronized] = ACTIONS(2582),
    [anon_sym_BOOL] = ACTIONS(2580),
    [anon_sym_IMP] = ACTIONS(2580),
    [anon_sym_SEL] = ACTIONS(2580),
    [anon_sym_Class] = ACTIONS(2580),
    [anon_sym_id] = ACTIONS(2580),
  },
  [STATE(640)] = {
    [sym_else_clause] = STATE(696),
    [sym_identifier] = ACTIONS(2394),
    [aux_sym_preproc_include_token1] = ACTIONS(2394),
    [aux_sym_preproc_include_token2] = ACTIONS(2394),
    [aux_sym_preproc_def_token1] = ACTIONS(2394),
    [aux_sym_preproc_if_token1] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2394),
    [sym_preproc_directive] = ACTIONS(2394),
    [anon_sym_LPAREN2] = ACTIONS(2396),
    [anon_sym_BANG] = ACTIONS(2396),
    [anon_sym_TILDE] = ACTIONS(2396),
    [anon_sym_DASH] = ACTIONS(2394),
    [anon_sym_PLUS] = ACTIONS(2394),
    [anon_sym_STAR] = ACTIONS(2396),
    [anon_sym_CARET] = ACTIONS(2396),
    [anon_sym_AMP] = ACTIONS(2396),
    [anon_sym_SEMI] = ACTIONS(2396),
    [anon_sym___extension__] = ACTIONS(2394),
    [anon_sym_typedef] = ACTIONS(2394),
    [anon_sym_extern] = ACTIONS(2394),
    [anon_sym___attribute__] = ACTIONS(2394),
    [anon_sym___attribute] = ACTIONS(2394),
    [anon_sym_noreturn] = ACTIONS(2394),
    [anon_sym_LBRACK] = ACTIONS(2396),
    [anon_sym___declspec] = ACTIONS(2394),
    [anon_sym___cdecl] = ACTIONS(2394),
    [anon_sym___clrcall] = ACTIONS(2394),
    [anon_sym___stdcall] = ACTIONS(2394),
    [anon_sym___fastcall] = ACTIONS(2394),
    [anon_sym___thiscall] = ACTIONS(2394),
    [anon_sym___vectorcall] = ACTIONS(2394),
    [anon_sym_LBRACE] = ACTIONS(2396),
    [anon_sym_RBRACE] = ACTIONS(2396),
    [anon_sym_signed] = ACTIONS(2394),
    [anon_sym_unsigned] = ACTIONS(2394),
    [anon_sym_long] = ACTIONS(2394),
    [anon_sym_short] = ACTIONS(2394),
    [anon_sym_static] = ACTIONS(2394),
    [anon_sym_ATautoreleasepool] = ACTIONS(2396),
    [anon_sym_auto] = ACTIONS(2394),
    [anon_sym_register] = ACTIONS(2394),
    [anon_sym_inline] = ACTIONS(2394),
    [anon_sym___inline] = ACTIONS(2394),
    [anon_sym___inline__] = ACTIONS(2394),
    [anon_sym___forceinline] = ACTIONS(2394),
    [anon_sym_thread_local] = ACTIONS(2394),
    [anon_sym___thread] = ACTIONS(2394),
    [anon_sym_CG_EXTERN] = ACTIONS(2394),
    [anon_sym_CG_INLINE] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2394),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2394),
    [anon_sym_IBOutlet] = ACTIONS(2394),
    [anon_sym_IBInspectable] = ACTIONS(2394),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2394),
    [anon_sym_NS_INLINE] = ACTIONS(2394),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2394),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2394),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2394),
    [anon_sym_const] = ACTIONS(2394),
    [anon_sym_constexpr] = ACTIONS(2394),
    [anon_sym_volatile] = ACTIONS(2394),
    [anon_sym_restrict] = ACTIONS(2394),
    [anon_sym___restrict__] = ACTIONS(2394),
    [anon_sym__Atomic] = ACTIONS(2394),
    [anon_sym__Noreturn] = ACTIONS(2394),
    [anon_sym__Nonnull] = ACTIONS(2394),
    [anon_sym_nullable] = ACTIONS(2394),
    [anon_sym__Complex] = ACTIONS(2394),
    [anon_sym__Nullable] = ACTIONS(2394),
    [anon_sym__Nullable_result] = ACTIONS(2394),
    [anon_sym__Null_unspecified] = ACTIONS(2394),
    [anon_sym___autoreleasing] = ACTIONS(2394),
    [anon_sym___block] = ACTIONS(2394),
    [anon_sym___bridge] = ACTIONS(2394),
    [anon_sym___bridge_retained] = ACTIONS(2394),
    [anon_sym___bridge_transfer] = ACTIONS(2394),
    [anon_sym___complex] = ACTIONS(2394),
    [anon_sym___const] = ACTIONS(2394),
    [anon_sym___imag] = ACTIONS(2394),
    [anon_sym___kindof] = ACTIONS(2394),
    [anon_sym___nonnull] = ACTIONS(2394),
    [anon_sym___nullable] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2394),
    [anon_sym___real] = ACTIONS(2394),
    [anon_sym___strong] = ACTIONS(2394),
    [anon_sym___unsafe_unretained] = ACTIONS(2394),
    [anon_sym___unused] = ACTIONS(2394),
    [anon_sym___weak] = ACTIONS(2394),
    [anon_sym_alignas] = ACTIONS(2394),
    [anon_sym__Alignas] = ACTIONS(2394),
    [sym_primitive_type] = ACTIONS(2394),
    [anon_sym_enum] = ACTIONS(2394),
    [anon_sym_struct] = ACTIONS(2394),
    [anon_sym_union] = ACTIONS(2394),
    [anon_sym_if] = ACTIONS(2394),
    [anon_sym_else] = ACTIONS(2894),
    [anon_sym_switch] = ACTIONS(2394),
    [anon_sym_case] = ACTIONS(2394),
    [anon_sym_default] = ACTIONS(2394),
    [anon_sym_while] = ACTIONS(2394),
    [anon_sym_do] = ACTIONS(2394),
    [anon_sym_for] = ACTIONS(2394),
    [anon_sym_in] = ACTIONS(2394),
    [anon_sym_return] = ACTIONS(2394),
    [anon_sym_break] = ACTIONS(2394),
    [anon_sym_continue] = ACTIONS(2394),
    [anon_sym_goto] = ACTIONS(2394),
    [anon_sym_DASH_DASH] = ACTIONS(2396),
    [anon_sym_PLUS_PLUS] = ACTIONS(2396),
    [anon_sym_sizeof] = ACTIONS(2394),
    [anon_sym___alignof__] = ACTIONS(2394),
    [anon_sym___alignof] = ACTIONS(2394),
    [anon_sym__alignof] = ACTIONS(2394),
    [anon_sym_alignof] = ACTIONS(2394),
    [anon_sym__Alignof] = ACTIONS(2394),
    [anon_sym_offsetof] = ACTIONS(2394),
    [anon_sym__Generic] = ACTIONS(2394),
    [anon_sym_asm] = ACTIONS(2394),
    [anon_sym___asm__] = ACTIONS(2394),
    [anon_sym___asm] = ACTIONS(2394),
    [sym_number_literal] = ACTIONS(2396),
    [anon_sym_L_SQUOTE] = ACTIONS(2396),
    [anon_sym_u_SQUOTE] = ACTIONS(2396),
    [anon_sym_U_SQUOTE] = ACTIONS(2396),
    [anon_sym_u8_SQUOTE] = ACTIONS(2396),
    [anon_sym_SQUOTE] = ACTIONS(2396),
    [anon_sym_AT] = ACTIONS(2394),
    [anon_sym_DQUOTE] = ACTIONS(2396),
    [anon_sym_L_DQUOTE] = ACTIONS(2396),
    [anon_sym_u_DQUOTE] = ACTIONS(2396),
    [anon_sym_U_DQUOTE] = ACTIONS(2396),
    [anon_sym_u8_DQUOTE] = ACTIONS(2396),
    [sym_true] = ACTIONS(2394),
    [sym_false] = ACTIONS(2394),
    [anon_sym_NULL] = ACTIONS(2394),
    [anon_sym_nullptr] = ACTIONS(2394),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2394),
    [anon_sym___typeof] = ACTIONS(2394),
    [anon_sym_typeof] = ACTIONS(2394),
    [anon_sym_ATimport] = ACTIONS(2396),
    [aux_sym_preproc_undef_token1] = ACTIONS(2394),
    [anon_sym_POUND] = ACTIONS(2394),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2394),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2394),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2394),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_API_AVAILABLE] = ACTIONS(2394),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_API_DEPRECATED] = ACTIONS(2394),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym___deprecated_msg] = ACTIONS(2394),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2394),
    [anon_sym_ATprotocol] = ACTIONS(2396),
    [anon_sym_ATinterface] = ACTIONS(2396),
    [anon_sym_ATimplementation] = ACTIONS(2396),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2396),
    [anon_sym_ATtry] = ACTIONS(2396),
    [anon_sym___try] = ACTIONS(2394),
    [anon_sym_ATthrow] = ACTIONS(2396),
    [anon_sym_ATselector] = ACTIONS(2396),
    [anon_sym_ATavailable] = ACTIONS(2396),
    [anon_sym___builtin_available] = ACTIONS(2394),
    [anon_sym_va_arg] = ACTIONS(2394),
    [anon_sym_ATencode] = ACTIONS(2396),
    [anon_sym_ATsynchronized] = ACTIONS(2396),
    [anon_sym_BOOL] = ACTIONS(2394),
    [anon_sym_IMP] = ACTIONS(2394),
    [anon_sym_SEL] = ACTIONS(2394),
    [anon_sym_Class] = ACTIONS(2394),
    [anon_sym_id] = ACTIONS(2394),
  },
  [STATE(641)] = {
    [sym_identifier] = ACTIONS(2856),
    [aux_sym_preproc_include_token1] = ACTIONS(2856),
    [aux_sym_preproc_include_token2] = ACTIONS(2856),
    [aux_sym_preproc_def_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token2] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2856),
    [sym_preproc_directive] = ACTIONS(2856),
    [anon_sym_LPAREN2] = ACTIONS(2858),
    [anon_sym_BANG] = ACTIONS(2858),
    [anon_sym_TILDE] = ACTIONS(2858),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_STAR] = ACTIONS(2858),
    [anon_sym_CARET] = ACTIONS(2858),
    [anon_sym_AMP] = ACTIONS(2858),
    [anon_sym_SEMI] = ACTIONS(2858),
    [anon_sym___extension__] = ACTIONS(2856),
    [anon_sym_typedef] = ACTIONS(2856),
    [anon_sym_extern] = ACTIONS(2856),
    [anon_sym___attribute__] = ACTIONS(2856),
    [anon_sym___attribute] = ACTIONS(2856),
    [anon_sym_noreturn] = ACTIONS(2856),
    [anon_sym_LBRACK] = ACTIONS(2858),
    [anon_sym___declspec] = ACTIONS(2856),
    [anon_sym___cdecl] = ACTIONS(2856),
    [anon_sym___clrcall] = ACTIONS(2856),
    [anon_sym___stdcall] = ACTIONS(2856),
    [anon_sym___fastcall] = ACTIONS(2856),
    [anon_sym___thiscall] = ACTIONS(2856),
    [anon_sym___vectorcall] = ACTIONS(2856),
    [anon_sym_LBRACE] = ACTIONS(2858),
    [anon_sym_signed] = ACTIONS(2856),
    [anon_sym_unsigned] = ACTIONS(2856),
    [anon_sym_long] = ACTIONS(2856),
    [anon_sym_short] = ACTIONS(2856),
    [anon_sym_static] = ACTIONS(2856),
    [anon_sym_ATautoreleasepool] = ACTIONS(2858),
    [anon_sym_auto] = ACTIONS(2856),
    [anon_sym_register] = ACTIONS(2856),
    [anon_sym_inline] = ACTIONS(2856),
    [anon_sym___inline] = ACTIONS(2856),
    [anon_sym___inline__] = ACTIONS(2856),
    [anon_sym___forceinline] = ACTIONS(2856),
    [anon_sym_thread_local] = ACTIONS(2856),
    [anon_sym___thread] = ACTIONS(2856),
    [anon_sym_CG_EXTERN] = ACTIONS(2856),
    [anon_sym_CG_INLINE] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2856),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2856),
    [anon_sym_IBOutlet] = ACTIONS(2856),
    [anon_sym_IBInspectable] = ACTIONS(2856),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2856),
    [anon_sym_NS_INLINE] = ACTIONS(2856),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2856),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2856),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2856),
    [anon_sym_const] = ACTIONS(2856),
    [anon_sym_constexpr] = ACTIONS(2856),
    [anon_sym_volatile] = ACTIONS(2856),
    [anon_sym_restrict] = ACTIONS(2856),
    [anon_sym___restrict__] = ACTIONS(2856),
    [anon_sym__Atomic] = ACTIONS(2856),
    [anon_sym__Noreturn] = ACTIONS(2856),
    [anon_sym__Nonnull] = ACTIONS(2856),
    [anon_sym_nullable] = ACTIONS(2856),
    [anon_sym__Complex] = ACTIONS(2856),
    [anon_sym__Nullable] = ACTIONS(2856),
    [anon_sym__Nullable_result] = ACTIONS(2856),
    [anon_sym__Null_unspecified] = ACTIONS(2856),
    [anon_sym___autoreleasing] = ACTIONS(2856),
    [anon_sym___block] = ACTIONS(2856),
    [anon_sym___bridge] = ACTIONS(2856),
    [anon_sym___bridge_retained] = ACTIONS(2856),
    [anon_sym___bridge_transfer] = ACTIONS(2856),
    [anon_sym___complex] = ACTIONS(2856),
    [anon_sym___const] = ACTIONS(2856),
    [anon_sym___imag] = ACTIONS(2856),
    [anon_sym___kindof] = ACTIONS(2856),
    [anon_sym___nonnull] = ACTIONS(2856),
    [anon_sym___nullable] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2856),
    [anon_sym___real] = ACTIONS(2856),
    [anon_sym___strong] = ACTIONS(2856),
    [anon_sym___unsafe_unretained] = ACTIONS(2856),
    [anon_sym___unused] = ACTIONS(2856),
    [anon_sym___weak] = ACTIONS(2856),
    [anon_sym_alignas] = ACTIONS(2856),
    [anon_sym__Alignas] = ACTIONS(2856),
    [sym_primitive_type] = ACTIONS(2856),
    [anon_sym_enum] = ACTIONS(2856),
    [anon_sym_struct] = ACTIONS(2856),
    [anon_sym_union] = ACTIONS(2856),
    [anon_sym_if] = ACTIONS(2856),
    [anon_sym_switch] = ACTIONS(2856),
    [anon_sym_case] = ACTIONS(2856),
    [anon_sym_default] = ACTIONS(2856),
    [anon_sym_while] = ACTIONS(2856),
    [anon_sym_do] = ACTIONS(2856),
    [anon_sym_for] = ACTIONS(2856),
    [anon_sym_in] = ACTIONS(2856),
    [anon_sym_return] = ACTIONS(2856),
    [anon_sym_break] = ACTIONS(2856),
    [anon_sym_continue] = ACTIONS(2856),
    [anon_sym_goto] = ACTIONS(2856),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_sizeof] = ACTIONS(2856),
    [anon_sym___alignof__] = ACTIONS(2856),
    [anon_sym___alignof] = ACTIONS(2856),
    [anon_sym__alignof] = ACTIONS(2856),
    [anon_sym_alignof] = ACTIONS(2856),
    [anon_sym__Alignof] = ACTIONS(2856),
    [anon_sym_offsetof] = ACTIONS(2856),
    [anon_sym__Generic] = ACTIONS(2856),
    [anon_sym_asm] = ACTIONS(2856),
    [anon_sym___asm__] = ACTIONS(2856),
    [anon_sym___asm] = ACTIONS(2856),
    [sym_number_literal] = ACTIONS(2858),
    [anon_sym_L_SQUOTE] = ACTIONS(2858),
    [anon_sym_u_SQUOTE] = ACTIONS(2858),
    [anon_sym_U_SQUOTE] = ACTIONS(2858),
    [anon_sym_u8_SQUOTE] = ACTIONS(2858),
    [anon_sym_SQUOTE] = ACTIONS(2858),
    [anon_sym_AT] = ACTIONS(2856),
    [anon_sym_DQUOTE] = ACTIONS(2858),
    [anon_sym_L_DQUOTE] = ACTIONS(2858),
    [anon_sym_u_DQUOTE] = ACTIONS(2858),
    [anon_sym_U_DQUOTE] = ACTIONS(2858),
    [anon_sym_u8_DQUOTE] = ACTIONS(2858),
    [sym_true] = ACTIONS(2856),
    [sym_false] = ACTIONS(2856),
    [anon_sym_NULL] = ACTIONS(2856),
    [anon_sym_nullptr] = ACTIONS(2856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2856),
    [anon_sym___typeof] = ACTIONS(2856),
    [anon_sym_typeof] = ACTIONS(2856),
    [anon_sym_ATimport] = ACTIONS(2858),
    [aux_sym_preproc_undef_token1] = ACTIONS(2856),
    [anon_sym_POUND] = ACTIONS(2856),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2856),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2856),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2856),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_API_AVAILABLE] = ACTIONS(2856),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_API_DEPRECATED] = ACTIONS(2856),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym___deprecated_msg] = ACTIONS(2856),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2856),
    [anon_sym_ATprotocol] = ACTIONS(2858),
    [anon_sym_ATinterface] = ACTIONS(2858),
    [anon_sym_ATimplementation] = ACTIONS(2858),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2858),
    [anon_sym_ATsynthesize] = ACTIONS(2858),
    [anon_sym_ATdynamic] = ACTIONS(2858),
    [anon_sym_ATtry] = ACTIONS(2858),
    [anon_sym___try] = ACTIONS(2856),
    [anon_sym_ATthrow] = ACTIONS(2858),
    [anon_sym_ATselector] = ACTIONS(2858),
    [anon_sym_ATavailable] = ACTIONS(2858),
    [anon_sym___builtin_available] = ACTIONS(2856),
    [anon_sym_va_arg] = ACTIONS(2856),
    [anon_sym_ATencode] = ACTIONS(2858),
    [anon_sym_ATsynchronized] = ACTIONS(2858),
    [anon_sym_BOOL] = ACTIONS(2856),
    [anon_sym_IMP] = ACTIONS(2856),
    [anon_sym_SEL] = ACTIONS(2856),
    [anon_sym_Class] = ACTIONS(2856),
    [anon_sym_id] = ACTIONS(2856),
  },
  [STATE(642)] = {
    [sym_identifier] = ACTIONS(2588),
    [aux_sym_preproc_include_token1] = ACTIONS(2588),
    [aux_sym_preproc_include_token2] = ACTIONS(2588),
    [aux_sym_preproc_def_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token2] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2588),
    [sym_preproc_directive] = ACTIONS(2588),
    [anon_sym_LPAREN2] = ACTIONS(2590),
    [anon_sym_BANG] = ACTIONS(2590),
    [anon_sym_TILDE] = ACTIONS(2590),
    [anon_sym_DASH] = ACTIONS(2588),
    [anon_sym_PLUS] = ACTIONS(2588),
    [anon_sym_STAR] = ACTIONS(2590),
    [anon_sym_CARET] = ACTIONS(2590),
    [anon_sym_AMP] = ACTIONS(2590),
    [anon_sym_SEMI] = ACTIONS(2590),
    [anon_sym___extension__] = ACTIONS(2588),
    [anon_sym_typedef] = ACTIONS(2588),
    [anon_sym_extern] = ACTIONS(2588),
    [anon_sym___attribute__] = ACTIONS(2588),
    [anon_sym___attribute] = ACTIONS(2588),
    [anon_sym_noreturn] = ACTIONS(2588),
    [anon_sym_LBRACK] = ACTIONS(2590),
    [anon_sym___declspec] = ACTIONS(2588),
    [anon_sym___cdecl] = ACTIONS(2588),
    [anon_sym___clrcall] = ACTIONS(2588),
    [anon_sym___stdcall] = ACTIONS(2588),
    [anon_sym___fastcall] = ACTIONS(2588),
    [anon_sym___thiscall] = ACTIONS(2588),
    [anon_sym___vectorcall] = ACTIONS(2588),
    [anon_sym_LBRACE] = ACTIONS(2590),
    [anon_sym_signed] = ACTIONS(2588),
    [anon_sym_unsigned] = ACTIONS(2588),
    [anon_sym_long] = ACTIONS(2588),
    [anon_sym_short] = ACTIONS(2588),
    [anon_sym_static] = ACTIONS(2588),
    [anon_sym_ATautoreleasepool] = ACTIONS(2590),
    [anon_sym_auto] = ACTIONS(2588),
    [anon_sym_register] = ACTIONS(2588),
    [anon_sym_inline] = ACTIONS(2588),
    [anon_sym___inline] = ACTIONS(2588),
    [anon_sym___inline__] = ACTIONS(2588),
    [anon_sym___forceinline] = ACTIONS(2588),
    [anon_sym_thread_local] = ACTIONS(2588),
    [anon_sym___thread] = ACTIONS(2588),
    [anon_sym_CG_EXTERN] = ACTIONS(2588),
    [anon_sym_CG_INLINE] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2588),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2588),
    [anon_sym_IBOutlet] = ACTIONS(2588),
    [anon_sym_IBInspectable] = ACTIONS(2588),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2588),
    [anon_sym_NS_INLINE] = ACTIONS(2588),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2588),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2588),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2588),
    [anon_sym_const] = ACTIONS(2588),
    [anon_sym_constexpr] = ACTIONS(2588),
    [anon_sym_volatile] = ACTIONS(2588),
    [anon_sym_restrict] = ACTIONS(2588),
    [anon_sym___restrict__] = ACTIONS(2588),
    [anon_sym__Atomic] = ACTIONS(2588),
    [anon_sym__Noreturn] = ACTIONS(2588),
    [anon_sym__Nonnull] = ACTIONS(2588),
    [anon_sym_nullable] = ACTIONS(2588),
    [anon_sym__Complex] = ACTIONS(2588),
    [anon_sym__Nullable] = ACTIONS(2588),
    [anon_sym__Nullable_result] = ACTIONS(2588),
    [anon_sym__Null_unspecified] = ACTIONS(2588),
    [anon_sym___autoreleasing] = ACTIONS(2588),
    [anon_sym___block] = ACTIONS(2588),
    [anon_sym___bridge] = ACTIONS(2588),
    [anon_sym___bridge_retained] = ACTIONS(2588),
    [anon_sym___bridge_transfer] = ACTIONS(2588),
    [anon_sym___complex] = ACTIONS(2588),
    [anon_sym___const] = ACTIONS(2588),
    [anon_sym___imag] = ACTIONS(2588),
    [anon_sym___kindof] = ACTIONS(2588),
    [anon_sym___nonnull] = ACTIONS(2588),
    [anon_sym___nullable] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2588),
    [anon_sym___real] = ACTIONS(2588),
    [anon_sym___strong] = ACTIONS(2588),
    [anon_sym___unsafe_unretained] = ACTIONS(2588),
    [anon_sym___unused] = ACTIONS(2588),
    [anon_sym___weak] = ACTIONS(2588),
    [anon_sym_alignas] = ACTIONS(2588),
    [anon_sym__Alignas] = ACTIONS(2588),
    [sym_primitive_type] = ACTIONS(2588),
    [anon_sym_enum] = ACTIONS(2588),
    [anon_sym_struct] = ACTIONS(2588),
    [anon_sym_union] = ACTIONS(2588),
    [anon_sym_if] = ACTIONS(2588),
    [anon_sym_switch] = ACTIONS(2588),
    [anon_sym_case] = ACTIONS(2588),
    [anon_sym_default] = ACTIONS(2588),
    [anon_sym_while] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2588),
    [anon_sym_for] = ACTIONS(2588),
    [anon_sym_in] = ACTIONS(2588),
    [anon_sym_return] = ACTIONS(2588),
    [anon_sym_break] = ACTIONS(2588),
    [anon_sym_continue] = ACTIONS(2588),
    [anon_sym_goto] = ACTIONS(2588),
    [anon_sym_DASH_DASH] = ACTIONS(2590),
    [anon_sym_PLUS_PLUS] = ACTIONS(2590),
    [anon_sym_sizeof] = ACTIONS(2588),
    [anon_sym___alignof__] = ACTIONS(2588),
    [anon_sym___alignof] = ACTIONS(2588),
    [anon_sym__alignof] = ACTIONS(2588),
    [anon_sym_alignof] = ACTIONS(2588),
    [anon_sym__Alignof] = ACTIONS(2588),
    [anon_sym_offsetof] = ACTIONS(2588),
    [anon_sym__Generic] = ACTIONS(2588),
    [anon_sym_asm] = ACTIONS(2588),
    [anon_sym___asm__] = ACTIONS(2588),
    [anon_sym___asm] = ACTIONS(2588),
    [sym_number_literal] = ACTIONS(2590),
    [anon_sym_L_SQUOTE] = ACTIONS(2590),
    [anon_sym_u_SQUOTE] = ACTIONS(2590),
    [anon_sym_U_SQUOTE] = ACTIONS(2590),
    [anon_sym_u8_SQUOTE] = ACTIONS(2590),
    [anon_sym_SQUOTE] = ACTIONS(2590),
    [anon_sym_AT] = ACTIONS(2588),
    [anon_sym_DQUOTE] = ACTIONS(2590),
    [anon_sym_L_DQUOTE] = ACTIONS(2590),
    [anon_sym_u_DQUOTE] = ACTIONS(2590),
    [anon_sym_U_DQUOTE] = ACTIONS(2590),
    [anon_sym_u8_DQUOTE] = ACTIONS(2590),
    [sym_true] = ACTIONS(2588),
    [sym_false] = ACTIONS(2588),
    [anon_sym_NULL] = ACTIONS(2588),
    [anon_sym_nullptr] = ACTIONS(2588),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2588),
    [anon_sym___typeof] = ACTIONS(2588),
    [anon_sym_typeof] = ACTIONS(2588),
    [anon_sym_ATimport] = ACTIONS(2590),
    [aux_sym_preproc_undef_token1] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2588),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2588),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2588),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2588),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_API_AVAILABLE] = ACTIONS(2588),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_API_DEPRECATED] = ACTIONS(2588),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym___deprecated_msg] = ACTIONS(2588),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2588),
    [anon_sym_ATprotocol] = ACTIONS(2590),
    [anon_sym_ATinterface] = ACTIONS(2590),
    [anon_sym_ATimplementation] = ACTIONS(2590),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2590),
    [anon_sym_ATsynthesize] = ACTIONS(2590),
    [anon_sym_ATdynamic] = ACTIONS(2590),
    [anon_sym_ATtry] = ACTIONS(2590),
    [anon_sym___try] = ACTIONS(2588),
    [anon_sym_ATthrow] = ACTIONS(2590),
    [anon_sym_ATselector] = ACTIONS(2590),
    [anon_sym_ATavailable] = ACTIONS(2590),
    [anon_sym___builtin_available] = ACTIONS(2588),
    [anon_sym_va_arg] = ACTIONS(2588),
    [anon_sym_ATencode] = ACTIONS(2590),
    [anon_sym_ATsynchronized] = ACTIONS(2590),
    [anon_sym_BOOL] = ACTIONS(2588),
    [anon_sym_IMP] = ACTIONS(2588),
    [anon_sym_SEL] = ACTIONS(2588),
    [anon_sym_Class] = ACTIONS(2588),
    [anon_sym_id] = ACTIONS(2588),
  },
  [STATE(643)] = {
    [sym_identifier] = ACTIONS(2852),
    [aux_sym_preproc_include_token1] = ACTIONS(2852),
    [aux_sym_preproc_include_token2] = ACTIONS(2852),
    [aux_sym_preproc_def_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token2] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2852),
    [sym_preproc_directive] = ACTIONS(2852),
    [anon_sym_LPAREN2] = ACTIONS(2854),
    [anon_sym_BANG] = ACTIONS(2854),
    [anon_sym_TILDE] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2852),
    [anon_sym_PLUS] = ACTIONS(2852),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_SEMI] = ACTIONS(2854),
    [anon_sym___extension__] = ACTIONS(2852),
    [anon_sym_typedef] = ACTIONS(2852),
    [anon_sym_extern] = ACTIONS(2852),
    [anon_sym___attribute__] = ACTIONS(2852),
    [anon_sym___attribute] = ACTIONS(2852),
    [anon_sym_noreturn] = ACTIONS(2852),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym___declspec] = ACTIONS(2852),
    [anon_sym___cdecl] = ACTIONS(2852),
    [anon_sym___clrcall] = ACTIONS(2852),
    [anon_sym___stdcall] = ACTIONS(2852),
    [anon_sym___fastcall] = ACTIONS(2852),
    [anon_sym___thiscall] = ACTIONS(2852),
    [anon_sym___vectorcall] = ACTIONS(2852),
    [anon_sym_LBRACE] = ACTIONS(2854),
    [anon_sym_signed] = ACTIONS(2852),
    [anon_sym_unsigned] = ACTIONS(2852),
    [anon_sym_long] = ACTIONS(2852),
    [anon_sym_short] = ACTIONS(2852),
    [anon_sym_static] = ACTIONS(2852),
    [anon_sym_ATautoreleasepool] = ACTIONS(2854),
    [anon_sym_auto] = ACTIONS(2852),
    [anon_sym_register] = ACTIONS(2852),
    [anon_sym_inline] = ACTIONS(2852),
    [anon_sym___inline] = ACTIONS(2852),
    [anon_sym___inline__] = ACTIONS(2852),
    [anon_sym___forceinline] = ACTIONS(2852),
    [anon_sym_thread_local] = ACTIONS(2852),
    [anon_sym___thread] = ACTIONS(2852),
    [anon_sym_CG_EXTERN] = ACTIONS(2852),
    [anon_sym_CG_INLINE] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2852),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2852),
    [anon_sym_IBOutlet] = ACTIONS(2852),
    [anon_sym_IBInspectable] = ACTIONS(2852),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2852),
    [anon_sym_NS_INLINE] = ACTIONS(2852),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2852),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2852),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2852),
    [anon_sym_const] = ACTIONS(2852),
    [anon_sym_constexpr] = ACTIONS(2852),
    [anon_sym_volatile] = ACTIONS(2852),
    [anon_sym_restrict] = ACTIONS(2852),
    [anon_sym___restrict__] = ACTIONS(2852),
    [anon_sym__Atomic] = ACTIONS(2852),
    [anon_sym__Noreturn] = ACTIONS(2852),
    [anon_sym__Nonnull] = ACTIONS(2852),
    [anon_sym_nullable] = ACTIONS(2852),
    [anon_sym__Complex] = ACTIONS(2852),
    [anon_sym__Nullable] = ACTIONS(2852),
    [anon_sym__Nullable_result] = ACTIONS(2852),
    [anon_sym__Null_unspecified] = ACTIONS(2852),
    [anon_sym___autoreleasing] = ACTIONS(2852),
    [anon_sym___block] = ACTIONS(2852),
    [anon_sym___bridge] = ACTIONS(2852),
    [anon_sym___bridge_retained] = ACTIONS(2852),
    [anon_sym___bridge_transfer] = ACTIONS(2852),
    [anon_sym___complex] = ACTIONS(2852),
    [anon_sym___const] = ACTIONS(2852),
    [anon_sym___imag] = ACTIONS(2852),
    [anon_sym___kindof] = ACTIONS(2852),
    [anon_sym___nonnull] = ACTIONS(2852),
    [anon_sym___nullable] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2852),
    [anon_sym___real] = ACTIONS(2852),
    [anon_sym___strong] = ACTIONS(2852),
    [anon_sym___unsafe_unretained] = ACTIONS(2852),
    [anon_sym___unused] = ACTIONS(2852),
    [anon_sym___weak] = ACTIONS(2852),
    [anon_sym_alignas] = ACTIONS(2852),
    [anon_sym__Alignas] = ACTIONS(2852),
    [sym_primitive_type] = ACTIONS(2852),
    [anon_sym_enum] = ACTIONS(2852),
    [anon_sym_struct] = ACTIONS(2852),
    [anon_sym_union] = ACTIONS(2852),
    [anon_sym_if] = ACTIONS(2852),
    [anon_sym_switch] = ACTIONS(2852),
    [anon_sym_case] = ACTIONS(2852),
    [anon_sym_default] = ACTIONS(2852),
    [anon_sym_while] = ACTIONS(2852),
    [anon_sym_do] = ACTIONS(2852),
    [anon_sym_for] = ACTIONS(2852),
    [anon_sym_in] = ACTIONS(2852),
    [anon_sym_return] = ACTIONS(2852),
    [anon_sym_break] = ACTIONS(2852),
    [anon_sym_continue] = ACTIONS(2852),
    [anon_sym_goto] = ACTIONS(2852),
    [anon_sym_DASH_DASH] = ACTIONS(2854),
    [anon_sym_PLUS_PLUS] = ACTIONS(2854),
    [anon_sym_sizeof] = ACTIONS(2852),
    [anon_sym___alignof__] = ACTIONS(2852),
    [anon_sym___alignof] = ACTIONS(2852),
    [anon_sym__alignof] = ACTIONS(2852),
    [anon_sym_alignof] = ACTIONS(2852),
    [anon_sym__Alignof] = ACTIONS(2852),
    [anon_sym_offsetof] = ACTIONS(2852),
    [anon_sym__Generic] = ACTIONS(2852),
    [anon_sym_asm] = ACTIONS(2852),
    [anon_sym___asm__] = ACTIONS(2852),
    [anon_sym___asm] = ACTIONS(2852),
    [sym_number_literal] = ACTIONS(2854),
    [anon_sym_L_SQUOTE] = ACTIONS(2854),
    [anon_sym_u_SQUOTE] = ACTIONS(2854),
    [anon_sym_U_SQUOTE] = ACTIONS(2854),
    [anon_sym_u8_SQUOTE] = ACTIONS(2854),
    [anon_sym_SQUOTE] = ACTIONS(2854),
    [anon_sym_AT] = ACTIONS(2852),
    [anon_sym_DQUOTE] = ACTIONS(2854),
    [anon_sym_L_DQUOTE] = ACTIONS(2854),
    [anon_sym_u_DQUOTE] = ACTIONS(2854),
    [anon_sym_U_DQUOTE] = ACTIONS(2854),
    [anon_sym_u8_DQUOTE] = ACTIONS(2854),
    [sym_true] = ACTIONS(2852),
    [sym_false] = ACTIONS(2852),
    [anon_sym_NULL] = ACTIONS(2852),
    [anon_sym_nullptr] = ACTIONS(2852),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2852),
    [anon_sym___typeof] = ACTIONS(2852),
    [anon_sym_typeof] = ACTIONS(2852),
    [anon_sym_ATimport] = ACTIONS(2854),
    [aux_sym_preproc_undef_token1] = ACTIONS(2852),
    [anon_sym_POUND] = ACTIONS(2852),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2852),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2852),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2852),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_API_AVAILABLE] = ACTIONS(2852),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_API_DEPRECATED] = ACTIONS(2852),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym___deprecated_msg] = ACTIONS(2852),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2852),
    [anon_sym_ATprotocol] = ACTIONS(2854),
    [anon_sym_ATinterface] = ACTIONS(2854),
    [anon_sym_ATimplementation] = ACTIONS(2854),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2854),
    [anon_sym_ATsynthesize] = ACTIONS(2854),
    [anon_sym_ATdynamic] = ACTIONS(2854),
    [anon_sym_ATtry] = ACTIONS(2854),
    [anon_sym___try] = ACTIONS(2852),
    [anon_sym_ATthrow] = ACTIONS(2854),
    [anon_sym_ATselector] = ACTIONS(2854),
    [anon_sym_ATavailable] = ACTIONS(2854),
    [anon_sym___builtin_available] = ACTIONS(2852),
    [anon_sym_va_arg] = ACTIONS(2852),
    [anon_sym_ATencode] = ACTIONS(2854),
    [anon_sym_ATsynchronized] = ACTIONS(2854),
    [anon_sym_BOOL] = ACTIONS(2852),
    [anon_sym_IMP] = ACTIONS(2852),
    [anon_sym_SEL] = ACTIONS(2852),
    [anon_sym_Class] = ACTIONS(2852),
    [anon_sym_id] = ACTIONS(2852),
  },
  [STATE(644)] = {
    [sym_identifier] = ACTIONS(2740),
    [aux_sym_preproc_include_token1] = ACTIONS(2740),
    [aux_sym_preproc_include_token2] = ACTIONS(2740),
    [aux_sym_preproc_def_token1] = ACTIONS(2740),
    [aux_sym_preproc_if_token1] = ACTIONS(2740),
    [aux_sym_preproc_if_token2] = ACTIONS(2740),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2740),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2740),
    [sym_preproc_directive] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_BANG] = ACTIONS(2742),
    [anon_sym_TILDE] = ACTIONS(2742),
    [anon_sym_DASH] = ACTIONS(2740),
    [anon_sym_PLUS] = ACTIONS(2740),
    [anon_sym_STAR] = ACTIONS(2742),
    [anon_sym_CARET] = ACTIONS(2742),
    [anon_sym_AMP] = ACTIONS(2742),
    [anon_sym_SEMI] = ACTIONS(2742),
    [anon_sym___extension__] = ACTIONS(2740),
    [anon_sym_typedef] = ACTIONS(2740),
    [anon_sym_extern] = ACTIONS(2740),
    [anon_sym___attribute__] = ACTIONS(2740),
    [anon_sym___attribute] = ACTIONS(2740),
    [anon_sym_noreturn] = ACTIONS(2740),
    [anon_sym_LBRACK] = ACTIONS(2742),
    [anon_sym___declspec] = ACTIONS(2740),
    [anon_sym___cdecl] = ACTIONS(2740),
    [anon_sym___clrcall] = ACTIONS(2740),
    [anon_sym___stdcall] = ACTIONS(2740),
    [anon_sym___fastcall] = ACTIONS(2740),
    [anon_sym___thiscall] = ACTIONS(2740),
    [anon_sym___vectorcall] = ACTIONS(2740),
    [anon_sym_LBRACE] = ACTIONS(2742),
    [anon_sym_signed] = ACTIONS(2740),
    [anon_sym_unsigned] = ACTIONS(2740),
    [anon_sym_long] = ACTIONS(2740),
    [anon_sym_short] = ACTIONS(2740),
    [anon_sym_static] = ACTIONS(2740),
    [anon_sym_ATautoreleasepool] = ACTIONS(2742),
    [anon_sym_auto] = ACTIONS(2740),
    [anon_sym_register] = ACTIONS(2740),
    [anon_sym_inline] = ACTIONS(2740),
    [anon_sym___inline] = ACTIONS(2740),
    [anon_sym___inline__] = ACTIONS(2740),
    [anon_sym___forceinline] = ACTIONS(2740),
    [anon_sym_thread_local] = ACTIONS(2740),
    [anon_sym___thread] = ACTIONS(2740),
    [anon_sym_CG_EXTERN] = ACTIONS(2740),
    [anon_sym_CG_INLINE] = ACTIONS(2740),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2740),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2740),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2740),
    [anon_sym_IBOutlet] = ACTIONS(2740),
    [anon_sym_IBInspectable] = ACTIONS(2740),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2740),
    [anon_sym_NS_INLINE] = ACTIONS(2740),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2740),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2740),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2740),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2740),
    [anon_sym_const] = ACTIONS(2740),
    [anon_sym_constexpr] = ACTIONS(2740),
    [anon_sym_volatile] = ACTIONS(2740),
    [anon_sym_restrict] = ACTIONS(2740),
    [anon_sym___restrict__] = ACTIONS(2740),
    [anon_sym__Atomic] = ACTIONS(2740),
    [anon_sym__Noreturn] = ACTIONS(2740),
    [anon_sym__Nonnull] = ACTIONS(2740),
    [anon_sym_nullable] = ACTIONS(2740),
    [anon_sym__Complex] = ACTIONS(2740),
    [anon_sym__Nullable] = ACTIONS(2740),
    [anon_sym__Nullable_result] = ACTIONS(2740),
    [anon_sym__Null_unspecified] = ACTIONS(2740),
    [anon_sym___autoreleasing] = ACTIONS(2740),
    [anon_sym___block] = ACTIONS(2740),
    [anon_sym___bridge] = ACTIONS(2740),
    [anon_sym___bridge_retained] = ACTIONS(2740),
    [anon_sym___bridge_transfer] = ACTIONS(2740),
    [anon_sym___complex] = ACTIONS(2740),
    [anon_sym___const] = ACTIONS(2740),
    [anon_sym___imag] = ACTIONS(2740),
    [anon_sym___kindof] = ACTIONS(2740),
    [anon_sym___nonnull] = ACTIONS(2740),
    [anon_sym___nullable] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2740),
    [anon_sym___real] = ACTIONS(2740),
    [anon_sym___strong] = ACTIONS(2740),
    [anon_sym___unsafe_unretained] = ACTIONS(2740),
    [anon_sym___unused] = ACTIONS(2740),
    [anon_sym___weak] = ACTIONS(2740),
    [anon_sym_alignas] = ACTIONS(2740),
    [anon_sym__Alignas] = ACTIONS(2740),
    [sym_primitive_type] = ACTIONS(2740),
    [anon_sym_enum] = ACTIONS(2740),
    [anon_sym_struct] = ACTIONS(2740),
    [anon_sym_union] = ACTIONS(2740),
    [anon_sym_if] = ACTIONS(2740),
    [anon_sym_switch] = ACTIONS(2740),
    [anon_sym_case] = ACTIONS(2740),
    [anon_sym_default] = ACTIONS(2740),
    [anon_sym_while] = ACTIONS(2740),
    [anon_sym_do] = ACTIONS(2740),
    [anon_sym_for] = ACTIONS(2740),
    [anon_sym_in] = ACTIONS(2740),
    [anon_sym_return] = ACTIONS(2740),
    [anon_sym_break] = ACTIONS(2740),
    [anon_sym_continue] = ACTIONS(2740),
    [anon_sym_goto] = ACTIONS(2740),
    [anon_sym_DASH_DASH] = ACTIONS(2742),
    [anon_sym_PLUS_PLUS] = ACTIONS(2742),
    [anon_sym_sizeof] = ACTIONS(2740),
    [anon_sym___alignof__] = ACTIONS(2740),
    [anon_sym___alignof] = ACTIONS(2740),
    [anon_sym__alignof] = ACTIONS(2740),
    [anon_sym_alignof] = ACTIONS(2740),
    [anon_sym__Alignof] = ACTIONS(2740),
    [anon_sym_offsetof] = ACTIONS(2740),
    [anon_sym__Generic] = ACTIONS(2740),
    [anon_sym_asm] = ACTIONS(2740),
    [anon_sym___asm__] = ACTIONS(2740),
    [anon_sym___asm] = ACTIONS(2740),
    [sym_number_literal] = ACTIONS(2742),
    [anon_sym_L_SQUOTE] = ACTIONS(2742),
    [anon_sym_u_SQUOTE] = ACTIONS(2742),
    [anon_sym_U_SQUOTE] = ACTIONS(2742),
    [anon_sym_u8_SQUOTE] = ACTIONS(2742),
    [anon_sym_SQUOTE] = ACTIONS(2742),
    [anon_sym_AT] = ACTIONS(2740),
    [anon_sym_DQUOTE] = ACTIONS(2742),
    [anon_sym_L_DQUOTE] = ACTIONS(2742),
    [anon_sym_u_DQUOTE] = ACTIONS(2742),
    [anon_sym_U_DQUOTE] = ACTIONS(2742),
    [anon_sym_u8_DQUOTE] = ACTIONS(2742),
    [sym_true] = ACTIONS(2740),
    [sym_false] = ACTIONS(2740),
    [anon_sym_NULL] = ACTIONS(2740),
    [anon_sym_nullptr] = ACTIONS(2740),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2740),
    [anon_sym___typeof] = ACTIONS(2740),
    [anon_sym_typeof] = ACTIONS(2740),
    [anon_sym_ATimport] = ACTIONS(2742),
    [aux_sym_preproc_undef_token1] = ACTIONS(2740),
    [anon_sym_POUND] = ACTIONS(2740),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2740),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2740),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2740),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2740),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2740),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2740),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2740),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_API_AVAILABLE] = ACTIONS(2740),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_API_DEPRECATED] = ACTIONS(2740),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2740),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym___deprecated_msg] = ACTIONS(2740),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2740),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2740),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2740),
    [anon_sym_ATprotocol] = ACTIONS(2742),
    [anon_sym_ATinterface] = ACTIONS(2742),
    [anon_sym_ATimplementation] = ACTIONS(2742),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2742),
    [anon_sym_ATsynthesize] = ACTIONS(2742),
    [anon_sym_ATdynamic] = ACTIONS(2742),
    [anon_sym_ATtry] = ACTIONS(2742),
    [anon_sym___try] = ACTIONS(2740),
    [anon_sym_ATthrow] = ACTIONS(2742),
    [anon_sym_ATselector] = ACTIONS(2742),
    [anon_sym_ATavailable] = ACTIONS(2742),
    [anon_sym___builtin_available] = ACTIONS(2740),
    [anon_sym_va_arg] = ACTIONS(2740),
    [anon_sym_ATencode] = ACTIONS(2742),
    [anon_sym_ATsynchronized] = ACTIONS(2742),
    [anon_sym_BOOL] = ACTIONS(2740),
    [anon_sym_IMP] = ACTIONS(2740),
    [anon_sym_SEL] = ACTIONS(2740),
    [anon_sym_Class] = ACTIONS(2740),
    [anon_sym_id] = ACTIONS(2740),
  },
  [STATE(645)] = {
    [sym_identifier] = ACTIONS(2584),
    [aux_sym_preproc_include_token1] = ACTIONS(2584),
    [aux_sym_preproc_include_token2] = ACTIONS(2584),
    [aux_sym_preproc_def_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token2] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2584),
    [sym_preproc_directive] = ACTIONS(2584),
    [anon_sym_LPAREN2] = ACTIONS(2586),
    [anon_sym_BANG] = ACTIONS(2586),
    [anon_sym_TILDE] = ACTIONS(2586),
    [anon_sym_DASH] = ACTIONS(2584),
    [anon_sym_PLUS] = ACTIONS(2584),
    [anon_sym_STAR] = ACTIONS(2586),
    [anon_sym_CARET] = ACTIONS(2586),
    [anon_sym_AMP] = ACTIONS(2586),
    [anon_sym_SEMI] = ACTIONS(2586),
    [anon_sym___extension__] = ACTIONS(2584),
    [anon_sym_typedef] = ACTIONS(2584),
    [anon_sym_extern] = ACTIONS(2584),
    [anon_sym___attribute__] = ACTIONS(2584),
    [anon_sym___attribute] = ACTIONS(2584),
    [anon_sym_noreturn] = ACTIONS(2584),
    [anon_sym_LBRACK] = ACTIONS(2586),
    [anon_sym___declspec] = ACTIONS(2584),
    [anon_sym___cdecl] = ACTIONS(2584),
    [anon_sym___clrcall] = ACTIONS(2584),
    [anon_sym___stdcall] = ACTIONS(2584),
    [anon_sym___fastcall] = ACTIONS(2584),
    [anon_sym___thiscall] = ACTIONS(2584),
    [anon_sym___vectorcall] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_signed] = ACTIONS(2584),
    [anon_sym_unsigned] = ACTIONS(2584),
    [anon_sym_long] = ACTIONS(2584),
    [anon_sym_short] = ACTIONS(2584),
    [anon_sym_static] = ACTIONS(2584),
    [anon_sym_ATautoreleasepool] = ACTIONS(2586),
    [anon_sym_auto] = ACTIONS(2584),
    [anon_sym_register] = ACTIONS(2584),
    [anon_sym_inline] = ACTIONS(2584),
    [anon_sym___inline] = ACTIONS(2584),
    [anon_sym___inline__] = ACTIONS(2584),
    [anon_sym___forceinline] = ACTIONS(2584),
    [anon_sym_thread_local] = ACTIONS(2584),
    [anon_sym___thread] = ACTIONS(2584),
    [anon_sym_CG_EXTERN] = ACTIONS(2584),
    [anon_sym_CG_INLINE] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2584),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2584),
    [anon_sym_IBOutlet] = ACTIONS(2584),
    [anon_sym_IBInspectable] = ACTIONS(2584),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2584),
    [anon_sym_NS_INLINE] = ACTIONS(2584),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2584),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2584),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2584),
    [anon_sym_const] = ACTIONS(2584),
    [anon_sym_constexpr] = ACTIONS(2584),
    [anon_sym_volatile] = ACTIONS(2584),
    [anon_sym_restrict] = ACTIONS(2584),
    [anon_sym___restrict__] = ACTIONS(2584),
    [anon_sym__Atomic] = ACTIONS(2584),
    [anon_sym__Noreturn] = ACTIONS(2584),
    [anon_sym__Nonnull] = ACTIONS(2584),
    [anon_sym_nullable] = ACTIONS(2584),
    [anon_sym__Complex] = ACTIONS(2584),
    [anon_sym__Nullable] = ACTIONS(2584),
    [anon_sym__Nullable_result] = ACTIONS(2584),
    [anon_sym__Null_unspecified] = ACTIONS(2584),
    [anon_sym___autoreleasing] = ACTIONS(2584),
    [anon_sym___block] = ACTIONS(2584),
    [anon_sym___bridge] = ACTIONS(2584),
    [anon_sym___bridge_retained] = ACTIONS(2584),
    [anon_sym___bridge_transfer] = ACTIONS(2584),
    [anon_sym___complex] = ACTIONS(2584),
    [anon_sym___const] = ACTIONS(2584),
    [anon_sym___imag] = ACTIONS(2584),
    [anon_sym___kindof] = ACTIONS(2584),
    [anon_sym___nonnull] = ACTIONS(2584),
    [anon_sym___nullable] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2584),
    [anon_sym___real] = ACTIONS(2584),
    [anon_sym___strong] = ACTIONS(2584),
    [anon_sym___unsafe_unretained] = ACTIONS(2584),
    [anon_sym___unused] = ACTIONS(2584),
    [anon_sym___weak] = ACTIONS(2584),
    [anon_sym_alignas] = ACTIONS(2584),
    [anon_sym__Alignas] = ACTIONS(2584),
    [sym_primitive_type] = ACTIONS(2584),
    [anon_sym_enum] = ACTIONS(2584),
    [anon_sym_struct] = ACTIONS(2584),
    [anon_sym_union] = ACTIONS(2584),
    [anon_sym_if] = ACTIONS(2584),
    [anon_sym_switch] = ACTIONS(2584),
    [anon_sym_case] = ACTIONS(2584),
    [anon_sym_default] = ACTIONS(2584),
    [anon_sym_while] = ACTIONS(2584),
    [anon_sym_do] = ACTIONS(2584),
    [anon_sym_for] = ACTIONS(2584),
    [anon_sym_in] = ACTIONS(2584),
    [anon_sym_return] = ACTIONS(2584),
    [anon_sym_break] = ACTIONS(2584),
    [anon_sym_continue] = ACTIONS(2584),
    [anon_sym_goto] = ACTIONS(2584),
    [anon_sym_DASH_DASH] = ACTIONS(2586),
    [anon_sym_PLUS_PLUS] = ACTIONS(2586),
    [anon_sym_sizeof] = ACTIONS(2584),
    [anon_sym___alignof__] = ACTIONS(2584),
    [anon_sym___alignof] = ACTIONS(2584),
    [anon_sym__alignof] = ACTIONS(2584),
    [anon_sym_alignof] = ACTIONS(2584),
    [anon_sym__Alignof] = ACTIONS(2584),
    [anon_sym_offsetof] = ACTIONS(2584),
    [anon_sym__Generic] = ACTIONS(2584),
    [anon_sym_asm] = ACTIONS(2584),
    [anon_sym___asm__] = ACTIONS(2584),
    [anon_sym___asm] = ACTIONS(2584),
    [sym_number_literal] = ACTIONS(2586),
    [anon_sym_L_SQUOTE] = ACTIONS(2586),
    [anon_sym_u_SQUOTE] = ACTIONS(2586),
    [anon_sym_U_SQUOTE] = ACTIONS(2586),
    [anon_sym_u8_SQUOTE] = ACTIONS(2586),
    [anon_sym_SQUOTE] = ACTIONS(2586),
    [anon_sym_AT] = ACTIONS(2584),
    [anon_sym_DQUOTE] = ACTIONS(2586),
    [anon_sym_L_DQUOTE] = ACTIONS(2586),
    [anon_sym_u_DQUOTE] = ACTIONS(2586),
    [anon_sym_U_DQUOTE] = ACTIONS(2586),
    [anon_sym_u8_DQUOTE] = ACTIONS(2586),
    [sym_true] = ACTIONS(2584),
    [sym_false] = ACTIONS(2584),
    [anon_sym_NULL] = ACTIONS(2584),
    [anon_sym_nullptr] = ACTIONS(2584),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2584),
    [anon_sym___typeof] = ACTIONS(2584),
    [anon_sym_typeof] = ACTIONS(2584),
    [anon_sym_ATimport] = ACTIONS(2586),
    [aux_sym_preproc_undef_token1] = ACTIONS(2584),
    [anon_sym_POUND] = ACTIONS(2584),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2584),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2584),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2584),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_API_AVAILABLE] = ACTIONS(2584),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_API_DEPRECATED] = ACTIONS(2584),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym___deprecated_msg] = ACTIONS(2584),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2584),
    [anon_sym_ATprotocol] = ACTIONS(2586),
    [anon_sym_ATinterface] = ACTIONS(2586),
    [anon_sym_ATimplementation] = ACTIONS(2586),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2586),
    [anon_sym_ATsynthesize] = ACTIONS(2586),
    [anon_sym_ATdynamic] = ACTIONS(2586),
    [anon_sym_ATtry] = ACTIONS(2586),
    [anon_sym___try] = ACTIONS(2584),
    [anon_sym_ATthrow] = ACTIONS(2586),
    [anon_sym_ATselector] = ACTIONS(2586),
    [anon_sym_ATavailable] = ACTIONS(2586),
    [anon_sym___builtin_available] = ACTIONS(2584),
    [anon_sym_va_arg] = ACTIONS(2584),
    [anon_sym_ATencode] = ACTIONS(2586),
    [anon_sym_ATsynchronized] = ACTIONS(2586),
    [anon_sym_BOOL] = ACTIONS(2584),
    [anon_sym_IMP] = ACTIONS(2584),
    [anon_sym_SEL] = ACTIONS(2584),
    [anon_sym_Class] = ACTIONS(2584),
    [anon_sym_id] = ACTIONS(2584),
  },
  [STATE(646)] = {
    [sym_identifier] = ACTIONS(2608),
    [aux_sym_preproc_include_token1] = ACTIONS(2608),
    [aux_sym_preproc_include_token2] = ACTIONS(2608),
    [aux_sym_preproc_def_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token2] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2608),
    [sym_preproc_directive] = ACTIONS(2608),
    [anon_sym_LPAREN2] = ACTIONS(2610),
    [anon_sym_BANG] = ACTIONS(2610),
    [anon_sym_TILDE] = ACTIONS(2610),
    [anon_sym_DASH] = ACTIONS(2608),
    [anon_sym_PLUS] = ACTIONS(2608),
    [anon_sym_STAR] = ACTIONS(2610),
    [anon_sym_CARET] = ACTIONS(2610),
    [anon_sym_AMP] = ACTIONS(2610),
    [anon_sym_SEMI] = ACTIONS(2610),
    [anon_sym___extension__] = ACTIONS(2608),
    [anon_sym_typedef] = ACTIONS(2608),
    [anon_sym_extern] = ACTIONS(2608),
    [anon_sym___attribute__] = ACTIONS(2608),
    [anon_sym___attribute] = ACTIONS(2608),
    [anon_sym_noreturn] = ACTIONS(2608),
    [anon_sym_LBRACK] = ACTIONS(2610),
    [anon_sym___declspec] = ACTIONS(2608),
    [anon_sym___cdecl] = ACTIONS(2608),
    [anon_sym___clrcall] = ACTIONS(2608),
    [anon_sym___stdcall] = ACTIONS(2608),
    [anon_sym___fastcall] = ACTIONS(2608),
    [anon_sym___thiscall] = ACTIONS(2608),
    [anon_sym___vectorcall] = ACTIONS(2608),
    [anon_sym_LBRACE] = ACTIONS(2610),
    [anon_sym_signed] = ACTIONS(2608),
    [anon_sym_unsigned] = ACTIONS(2608),
    [anon_sym_long] = ACTIONS(2608),
    [anon_sym_short] = ACTIONS(2608),
    [anon_sym_static] = ACTIONS(2608),
    [anon_sym_ATautoreleasepool] = ACTIONS(2610),
    [anon_sym_auto] = ACTIONS(2608),
    [anon_sym_register] = ACTIONS(2608),
    [anon_sym_inline] = ACTIONS(2608),
    [anon_sym___inline] = ACTIONS(2608),
    [anon_sym___inline__] = ACTIONS(2608),
    [anon_sym___forceinline] = ACTIONS(2608),
    [anon_sym_thread_local] = ACTIONS(2608),
    [anon_sym___thread] = ACTIONS(2608),
    [anon_sym_CG_EXTERN] = ACTIONS(2608),
    [anon_sym_CG_INLINE] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2608),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2608),
    [anon_sym_IBOutlet] = ACTIONS(2608),
    [anon_sym_IBInspectable] = ACTIONS(2608),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2608),
    [anon_sym_NS_INLINE] = ACTIONS(2608),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2608),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2608),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2608),
    [anon_sym_const] = ACTIONS(2608),
    [anon_sym_constexpr] = ACTIONS(2608),
    [anon_sym_volatile] = ACTIONS(2608),
    [anon_sym_restrict] = ACTIONS(2608),
    [anon_sym___restrict__] = ACTIONS(2608),
    [anon_sym__Atomic] = ACTIONS(2608),
    [anon_sym__Noreturn] = ACTIONS(2608),
    [anon_sym__Nonnull] = ACTIONS(2608),
    [anon_sym_nullable] = ACTIONS(2608),
    [anon_sym__Complex] = ACTIONS(2608),
    [anon_sym__Nullable] = ACTIONS(2608),
    [anon_sym__Nullable_result] = ACTIONS(2608),
    [anon_sym__Null_unspecified] = ACTIONS(2608),
    [anon_sym___autoreleasing] = ACTIONS(2608),
    [anon_sym___block] = ACTIONS(2608),
    [anon_sym___bridge] = ACTIONS(2608),
    [anon_sym___bridge_retained] = ACTIONS(2608),
    [anon_sym___bridge_transfer] = ACTIONS(2608),
    [anon_sym___complex] = ACTIONS(2608),
    [anon_sym___const] = ACTIONS(2608),
    [anon_sym___imag] = ACTIONS(2608),
    [anon_sym___kindof] = ACTIONS(2608),
    [anon_sym___nonnull] = ACTIONS(2608),
    [anon_sym___nullable] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2608),
    [anon_sym___real] = ACTIONS(2608),
    [anon_sym___strong] = ACTIONS(2608),
    [anon_sym___unsafe_unretained] = ACTIONS(2608),
    [anon_sym___unused] = ACTIONS(2608),
    [anon_sym___weak] = ACTIONS(2608),
    [anon_sym_alignas] = ACTIONS(2608),
    [anon_sym__Alignas] = ACTIONS(2608),
    [sym_primitive_type] = ACTIONS(2608),
    [anon_sym_enum] = ACTIONS(2608),
    [anon_sym_struct] = ACTIONS(2608),
    [anon_sym_union] = ACTIONS(2608),
    [anon_sym_if] = ACTIONS(2608),
    [anon_sym_switch] = ACTIONS(2608),
    [anon_sym_case] = ACTIONS(2608),
    [anon_sym_default] = ACTIONS(2608),
    [anon_sym_while] = ACTIONS(2608),
    [anon_sym_do] = ACTIONS(2608),
    [anon_sym_for] = ACTIONS(2608),
    [anon_sym_in] = ACTIONS(2608),
    [anon_sym_return] = ACTIONS(2608),
    [anon_sym_break] = ACTIONS(2608),
    [anon_sym_continue] = ACTIONS(2608),
    [anon_sym_goto] = ACTIONS(2608),
    [anon_sym_DASH_DASH] = ACTIONS(2610),
    [anon_sym_PLUS_PLUS] = ACTIONS(2610),
    [anon_sym_sizeof] = ACTIONS(2608),
    [anon_sym___alignof__] = ACTIONS(2608),
    [anon_sym___alignof] = ACTIONS(2608),
    [anon_sym__alignof] = ACTIONS(2608),
    [anon_sym_alignof] = ACTIONS(2608),
    [anon_sym__Alignof] = ACTIONS(2608),
    [anon_sym_offsetof] = ACTIONS(2608),
    [anon_sym__Generic] = ACTIONS(2608),
    [anon_sym_asm] = ACTIONS(2608),
    [anon_sym___asm__] = ACTIONS(2608),
    [anon_sym___asm] = ACTIONS(2608),
    [sym_number_literal] = ACTIONS(2610),
    [anon_sym_L_SQUOTE] = ACTIONS(2610),
    [anon_sym_u_SQUOTE] = ACTIONS(2610),
    [anon_sym_U_SQUOTE] = ACTIONS(2610),
    [anon_sym_u8_SQUOTE] = ACTIONS(2610),
    [anon_sym_SQUOTE] = ACTIONS(2610),
    [anon_sym_AT] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [anon_sym_L_DQUOTE] = ACTIONS(2610),
    [anon_sym_u_DQUOTE] = ACTIONS(2610),
    [anon_sym_U_DQUOTE] = ACTIONS(2610),
    [anon_sym_u8_DQUOTE] = ACTIONS(2610),
    [sym_true] = ACTIONS(2608),
    [sym_false] = ACTIONS(2608),
    [anon_sym_NULL] = ACTIONS(2608),
    [anon_sym_nullptr] = ACTIONS(2608),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2608),
    [anon_sym___typeof] = ACTIONS(2608),
    [anon_sym_typeof] = ACTIONS(2608),
    [anon_sym_ATimport] = ACTIONS(2610),
    [aux_sym_preproc_undef_token1] = ACTIONS(2608),
    [anon_sym_POUND] = ACTIONS(2608),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2608),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2608),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2608),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_API_AVAILABLE] = ACTIONS(2608),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_API_DEPRECATED] = ACTIONS(2608),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym___deprecated_msg] = ACTIONS(2608),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2608),
    [anon_sym_ATprotocol] = ACTIONS(2610),
    [anon_sym_ATinterface] = ACTIONS(2610),
    [anon_sym_ATimplementation] = ACTIONS(2610),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2610),
    [anon_sym_ATsynthesize] = ACTIONS(2610),
    [anon_sym_ATdynamic] = ACTIONS(2610),
    [anon_sym_ATtry] = ACTIONS(2610),
    [anon_sym___try] = ACTIONS(2608),
    [anon_sym_ATthrow] = ACTIONS(2610),
    [anon_sym_ATselector] = ACTIONS(2610),
    [anon_sym_ATavailable] = ACTIONS(2610),
    [anon_sym___builtin_available] = ACTIONS(2608),
    [anon_sym_va_arg] = ACTIONS(2608),
    [anon_sym_ATencode] = ACTIONS(2610),
    [anon_sym_ATsynchronized] = ACTIONS(2610),
    [anon_sym_BOOL] = ACTIONS(2608),
    [anon_sym_IMP] = ACTIONS(2608),
    [anon_sym_SEL] = ACTIONS(2608),
    [anon_sym_Class] = ACTIONS(2608),
    [anon_sym_id] = ACTIONS(2608),
  },
  [STATE(647)] = {
    [sym_identifier] = ACTIONS(2618),
    [aux_sym_preproc_include_token1] = ACTIONS(2618),
    [aux_sym_preproc_include_token2] = ACTIONS(2618),
    [aux_sym_preproc_def_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token2] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2618),
    [sym_preproc_directive] = ACTIONS(2618),
    [anon_sym_LPAREN2] = ACTIONS(2620),
    [anon_sym_BANG] = ACTIONS(2620),
    [anon_sym_TILDE] = ACTIONS(2620),
    [anon_sym_DASH] = ACTIONS(2618),
    [anon_sym_PLUS] = ACTIONS(2618),
    [anon_sym_STAR] = ACTIONS(2620),
    [anon_sym_CARET] = ACTIONS(2620),
    [anon_sym_AMP] = ACTIONS(2620),
    [anon_sym_SEMI] = ACTIONS(2620),
    [anon_sym___extension__] = ACTIONS(2618),
    [anon_sym_typedef] = ACTIONS(2618),
    [anon_sym_extern] = ACTIONS(2618),
    [anon_sym___attribute__] = ACTIONS(2618),
    [anon_sym___attribute] = ACTIONS(2618),
    [anon_sym_noreturn] = ACTIONS(2618),
    [anon_sym_LBRACK] = ACTIONS(2620),
    [anon_sym___declspec] = ACTIONS(2618),
    [anon_sym___cdecl] = ACTIONS(2618),
    [anon_sym___clrcall] = ACTIONS(2618),
    [anon_sym___stdcall] = ACTIONS(2618),
    [anon_sym___fastcall] = ACTIONS(2618),
    [anon_sym___thiscall] = ACTIONS(2618),
    [anon_sym___vectorcall] = ACTIONS(2618),
    [anon_sym_LBRACE] = ACTIONS(2620),
    [anon_sym_signed] = ACTIONS(2618),
    [anon_sym_unsigned] = ACTIONS(2618),
    [anon_sym_long] = ACTIONS(2618),
    [anon_sym_short] = ACTIONS(2618),
    [anon_sym_static] = ACTIONS(2618),
    [anon_sym_ATautoreleasepool] = ACTIONS(2620),
    [anon_sym_auto] = ACTIONS(2618),
    [anon_sym_register] = ACTIONS(2618),
    [anon_sym_inline] = ACTIONS(2618),
    [anon_sym___inline] = ACTIONS(2618),
    [anon_sym___inline__] = ACTIONS(2618),
    [anon_sym___forceinline] = ACTIONS(2618),
    [anon_sym_thread_local] = ACTIONS(2618),
    [anon_sym___thread] = ACTIONS(2618),
    [anon_sym_CG_EXTERN] = ACTIONS(2618),
    [anon_sym_CG_INLINE] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2618),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2618),
    [anon_sym_IBOutlet] = ACTIONS(2618),
    [anon_sym_IBInspectable] = ACTIONS(2618),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2618),
    [anon_sym_NS_INLINE] = ACTIONS(2618),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2618),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2618),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2618),
    [anon_sym_const] = ACTIONS(2618),
    [anon_sym_constexpr] = ACTIONS(2618),
    [anon_sym_volatile] = ACTIONS(2618),
    [anon_sym_restrict] = ACTIONS(2618),
    [anon_sym___restrict__] = ACTIONS(2618),
    [anon_sym__Atomic] = ACTIONS(2618),
    [anon_sym__Noreturn] = ACTIONS(2618),
    [anon_sym__Nonnull] = ACTIONS(2618),
    [anon_sym_nullable] = ACTIONS(2618),
    [anon_sym__Complex] = ACTIONS(2618),
    [anon_sym__Nullable] = ACTIONS(2618),
    [anon_sym__Nullable_result] = ACTIONS(2618),
    [anon_sym__Null_unspecified] = ACTIONS(2618),
    [anon_sym___autoreleasing] = ACTIONS(2618),
    [anon_sym___block] = ACTIONS(2618),
    [anon_sym___bridge] = ACTIONS(2618),
    [anon_sym___bridge_retained] = ACTIONS(2618),
    [anon_sym___bridge_transfer] = ACTIONS(2618),
    [anon_sym___complex] = ACTIONS(2618),
    [anon_sym___const] = ACTIONS(2618),
    [anon_sym___imag] = ACTIONS(2618),
    [anon_sym___kindof] = ACTIONS(2618),
    [anon_sym___nonnull] = ACTIONS(2618),
    [anon_sym___nullable] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2618),
    [anon_sym___real] = ACTIONS(2618),
    [anon_sym___strong] = ACTIONS(2618),
    [anon_sym___unsafe_unretained] = ACTIONS(2618),
    [anon_sym___unused] = ACTIONS(2618),
    [anon_sym___weak] = ACTIONS(2618),
    [anon_sym_alignas] = ACTIONS(2618),
    [anon_sym__Alignas] = ACTIONS(2618),
    [sym_primitive_type] = ACTIONS(2618),
    [anon_sym_enum] = ACTIONS(2618),
    [anon_sym_struct] = ACTIONS(2618),
    [anon_sym_union] = ACTIONS(2618),
    [anon_sym_if] = ACTIONS(2618),
    [anon_sym_switch] = ACTIONS(2618),
    [anon_sym_case] = ACTIONS(2618),
    [anon_sym_default] = ACTIONS(2618),
    [anon_sym_while] = ACTIONS(2618),
    [anon_sym_do] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2618),
    [anon_sym_in] = ACTIONS(2618),
    [anon_sym_return] = ACTIONS(2618),
    [anon_sym_break] = ACTIONS(2618),
    [anon_sym_continue] = ACTIONS(2618),
    [anon_sym_goto] = ACTIONS(2618),
    [anon_sym_DASH_DASH] = ACTIONS(2620),
    [anon_sym_PLUS_PLUS] = ACTIONS(2620),
    [anon_sym_sizeof] = ACTIONS(2618),
    [anon_sym___alignof__] = ACTIONS(2618),
    [anon_sym___alignof] = ACTIONS(2618),
    [anon_sym__alignof] = ACTIONS(2618),
    [anon_sym_alignof] = ACTIONS(2618),
    [anon_sym__Alignof] = ACTIONS(2618),
    [anon_sym_offsetof] = ACTIONS(2618),
    [anon_sym__Generic] = ACTIONS(2618),
    [anon_sym_asm] = ACTIONS(2618),
    [anon_sym___asm__] = ACTIONS(2618),
    [anon_sym___asm] = ACTIONS(2618),
    [sym_number_literal] = ACTIONS(2620),
    [anon_sym_L_SQUOTE] = ACTIONS(2620),
    [anon_sym_u_SQUOTE] = ACTIONS(2620),
    [anon_sym_U_SQUOTE] = ACTIONS(2620),
    [anon_sym_u8_SQUOTE] = ACTIONS(2620),
    [anon_sym_SQUOTE] = ACTIONS(2620),
    [anon_sym_AT] = ACTIONS(2618),
    [anon_sym_DQUOTE] = ACTIONS(2620),
    [anon_sym_L_DQUOTE] = ACTIONS(2620),
    [anon_sym_u_DQUOTE] = ACTIONS(2620),
    [anon_sym_U_DQUOTE] = ACTIONS(2620),
    [anon_sym_u8_DQUOTE] = ACTIONS(2620),
    [sym_true] = ACTIONS(2618),
    [sym_false] = ACTIONS(2618),
    [anon_sym_NULL] = ACTIONS(2618),
    [anon_sym_nullptr] = ACTIONS(2618),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2618),
    [anon_sym___typeof] = ACTIONS(2618),
    [anon_sym_typeof] = ACTIONS(2618),
    [anon_sym_ATimport] = ACTIONS(2620),
    [aux_sym_preproc_undef_token1] = ACTIONS(2618),
    [anon_sym_POUND] = ACTIONS(2618),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2618),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2618),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2618),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_API_AVAILABLE] = ACTIONS(2618),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_API_DEPRECATED] = ACTIONS(2618),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym___deprecated_msg] = ACTIONS(2618),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2618),
    [anon_sym_ATprotocol] = ACTIONS(2620),
    [anon_sym_ATinterface] = ACTIONS(2620),
    [anon_sym_ATimplementation] = ACTIONS(2620),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2620),
    [anon_sym_ATsynthesize] = ACTIONS(2620),
    [anon_sym_ATdynamic] = ACTIONS(2620),
    [anon_sym_ATtry] = ACTIONS(2620),
    [anon_sym___try] = ACTIONS(2618),
    [anon_sym_ATthrow] = ACTIONS(2620),
    [anon_sym_ATselector] = ACTIONS(2620),
    [anon_sym_ATavailable] = ACTIONS(2620),
    [anon_sym___builtin_available] = ACTIONS(2618),
    [anon_sym_va_arg] = ACTIONS(2618),
    [anon_sym_ATencode] = ACTIONS(2620),
    [anon_sym_ATsynchronized] = ACTIONS(2620),
    [anon_sym_BOOL] = ACTIONS(2618),
    [anon_sym_IMP] = ACTIONS(2618),
    [anon_sym_SEL] = ACTIONS(2618),
    [anon_sym_Class] = ACTIONS(2618),
    [anon_sym_id] = ACTIONS(2618),
  },
  [STATE(648)] = {
    [sym_identifier] = ACTIONS(2704),
    [aux_sym_preproc_include_token1] = ACTIONS(2704),
    [aux_sym_preproc_include_token2] = ACTIONS(2704),
    [aux_sym_preproc_def_token1] = ACTIONS(2704),
    [aux_sym_preproc_if_token1] = ACTIONS(2704),
    [aux_sym_preproc_if_token2] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2704),
    [sym_preproc_directive] = ACTIONS(2704),
    [anon_sym_LPAREN2] = ACTIONS(2706),
    [anon_sym_BANG] = ACTIONS(2706),
    [anon_sym_TILDE] = ACTIONS(2706),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_STAR] = ACTIONS(2706),
    [anon_sym_CARET] = ACTIONS(2706),
    [anon_sym_AMP] = ACTIONS(2706),
    [anon_sym_SEMI] = ACTIONS(2706),
    [anon_sym___extension__] = ACTIONS(2704),
    [anon_sym_typedef] = ACTIONS(2704),
    [anon_sym_extern] = ACTIONS(2704),
    [anon_sym___attribute__] = ACTIONS(2704),
    [anon_sym___attribute] = ACTIONS(2704),
    [anon_sym_noreturn] = ACTIONS(2704),
    [anon_sym_LBRACK] = ACTIONS(2706),
    [anon_sym___declspec] = ACTIONS(2704),
    [anon_sym___cdecl] = ACTIONS(2704),
    [anon_sym___clrcall] = ACTIONS(2704),
    [anon_sym___stdcall] = ACTIONS(2704),
    [anon_sym___fastcall] = ACTIONS(2704),
    [anon_sym___thiscall] = ACTIONS(2704),
    [anon_sym___vectorcall] = ACTIONS(2704),
    [anon_sym_LBRACE] = ACTIONS(2706),
    [anon_sym_signed] = ACTIONS(2704),
    [anon_sym_unsigned] = ACTIONS(2704),
    [anon_sym_long] = ACTIONS(2704),
    [anon_sym_short] = ACTIONS(2704),
    [anon_sym_static] = ACTIONS(2704),
    [anon_sym_ATautoreleasepool] = ACTIONS(2706),
    [anon_sym_auto] = ACTIONS(2704),
    [anon_sym_register] = ACTIONS(2704),
    [anon_sym_inline] = ACTIONS(2704),
    [anon_sym___inline] = ACTIONS(2704),
    [anon_sym___inline__] = ACTIONS(2704),
    [anon_sym___forceinline] = ACTIONS(2704),
    [anon_sym_thread_local] = ACTIONS(2704),
    [anon_sym___thread] = ACTIONS(2704),
    [anon_sym_CG_EXTERN] = ACTIONS(2704),
    [anon_sym_CG_INLINE] = ACTIONS(2704),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2704),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2704),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2704),
    [anon_sym_IBOutlet] = ACTIONS(2704),
    [anon_sym_IBInspectable] = ACTIONS(2704),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2704),
    [anon_sym_NS_INLINE] = ACTIONS(2704),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2704),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2704),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2704),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2704),
    [anon_sym_const] = ACTIONS(2704),
    [anon_sym_constexpr] = ACTIONS(2704),
    [anon_sym_volatile] = ACTIONS(2704),
    [anon_sym_restrict] = ACTIONS(2704),
    [anon_sym___restrict__] = ACTIONS(2704),
    [anon_sym__Atomic] = ACTIONS(2704),
    [anon_sym__Noreturn] = ACTIONS(2704),
    [anon_sym__Nonnull] = ACTIONS(2704),
    [anon_sym_nullable] = ACTIONS(2704),
    [anon_sym__Complex] = ACTIONS(2704),
    [anon_sym__Nullable] = ACTIONS(2704),
    [anon_sym__Nullable_result] = ACTIONS(2704),
    [anon_sym__Null_unspecified] = ACTIONS(2704),
    [anon_sym___autoreleasing] = ACTIONS(2704),
    [anon_sym___block] = ACTIONS(2704),
    [anon_sym___bridge] = ACTIONS(2704),
    [anon_sym___bridge_retained] = ACTIONS(2704),
    [anon_sym___bridge_transfer] = ACTIONS(2704),
    [anon_sym___complex] = ACTIONS(2704),
    [anon_sym___const] = ACTIONS(2704),
    [anon_sym___imag] = ACTIONS(2704),
    [anon_sym___kindof] = ACTIONS(2704),
    [anon_sym___nonnull] = ACTIONS(2704),
    [anon_sym___nullable] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2704),
    [anon_sym___real] = ACTIONS(2704),
    [anon_sym___strong] = ACTIONS(2704),
    [anon_sym___unsafe_unretained] = ACTIONS(2704),
    [anon_sym___unused] = ACTIONS(2704),
    [anon_sym___weak] = ACTIONS(2704),
    [anon_sym_alignas] = ACTIONS(2704),
    [anon_sym__Alignas] = ACTIONS(2704),
    [sym_primitive_type] = ACTIONS(2704),
    [anon_sym_enum] = ACTIONS(2704),
    [anon_sym_struct] = ACTIONS(2704),
    [anon_sym_union] = ACTIONS(2704),
    [anon_sym_if] = ACTIONS(2704),
    [anon_sym_switch] = ACTIONS(2704),
    [anon_sym_case] = ACTIONS(2704),
    [anon_sym_default] = ACTIONS(2704),
    [anon_sym_while] = ACTIONS(2704),
    [anon_sym_do] = ACTIONS(2704),
    [anon_sym_for] = ACTIONS(2704),
    [anon_sym_in] = ACTIONS(2704),
    [anon_sym_return] = ACTIONS(2704),
    [anon_sym_break] = ACTIONS(2704),
    [anon_sym_continue] = ACTIONS(2704),
    [anon_sym_goto] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2706),
    [anon_sym_PLUS_PLUS] = ACTIONS(2706),
    [anon_sym_sizeof] = ACTIONS(2704),
    [anon_sym___alignof__] = ACTIONS(2704),
    [anon_sym___alignof] = ACTIONS(2704),
    [anon_sym__alignof] = ACTIONS(2704),
    [anon_sym_alignof] = ACTIONS(2704),
    [anon_sym__Alignof] = ACTIONS(2704),
    [anon_sym_offsetof] = ACTIONS(2704),
    [anon_sym__Generic] = ACTIONS(2704),
    [anon_sym_asm] = ACTIONS(2704),
    [anon_sym___asm__] = ACTIONS(2704),
    [anon_sym___asm] = ACTIONS(2704),
    [sym_number_literal] = ACTIONS(2706),
    [anon_sym_L_SQUOTE] = ACTIONS(2706),
    [anon_sym_u_SQUOTE] = ACTIONS(2706),
    [anon_sym_U_SQUOTE] = ACTIONS(2706),
    [anon_sym_u8_SQUOTE] = ACTIONS(2706),
    [anon_sym_SQUOTE] = ACTIONS(2706),
    [anon_sym_AT] = ACTIONS(2704),
    [anon_sym_DQUOTE] = ACTIONS(2706),
    [anon_sym_L_DQUOTE] = ACTIONS(2706),
    [anon_sym_u_DQUOTE] = ACTIONS(2706),
    [anon_sym_U_DQUOTE] = ACTIONS(2706),
    [anon_sym_u8_DQUOTE] = ACTIONS(2706),
    [sym_true] = ACTIONS(2704),
    [sym_false] = ACTIONS(2704),
    [anon_sym_NULL] = ACTIONS(2704),
    [anon_sym_nullptr] = ACTIONS(2704),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2704),
    [anon_sym___typeof] = ACTIONS(2704),
    [anon_sym_typeof] = ACTIONS(2704),
    [anon_sym_ATimport] = ACTIONS(2706),
    [aux_sym_preproc_undef_token1] = ACTIONS(2704),
    [anon_sym_POUND] = ACTIONS(2704),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2704),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2704),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2704),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2704),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2704),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2704),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2704),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_API_AVAILABLE] = ACTIONS(2704),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_API_DEPRECATED] = ACTIONS(2704),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2704),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym___deprecated_msg] = ACTIONS(2704),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2704),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2704),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2704),
    [anon_sym_ATprotocol] = ACTIONS(2706),
    [anon_sym_ATinterface] = ACTIONS(2706),
    [anon_sym_ATimplementation] = ACTIONS(2706),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2706),
    [anon_sym_ATsynthesize] = ACTIONS(2706),
    [anon_sym_ATdynamic] = ACTIONS(2706),
    [anon_sym_ATtry] = ACTIONS(2706),
    [anon_sym___try] = ACTIONS(2704),
    [anon_sym_ATthrow] = ACTIONS(2706),
    [anon_sym_ATselector] = ACTIONS(2706),
    [anon_sym_ATavailable] = ACTIONS(2706),
    [anon_sym___builtin_available] = ACTIONS(2704),
    [anon_sym_va_arg] = ACTIONS(2704),
    [anon_sym_ATencode] = ACTIONS(2706),
    [anon_sym_ATsynchronized] = ACTIONS(2706),
    [anon_sym_BOOL] = ACTIONS(2704),
    [anon_sym_IMP] = ACTIONS(2704),
    [anon_sym_SEL] = ACTIONS(2704),
    [anon_sym_Class] = ACTIONS(2704),
    [anon_sym_id] = ACTIONS(2704),
  },
  [STATE(649)] = {
    [sym_identifier] = ACTIONS(2596),
    [aux_sym_preproc_include_token1] = ACTIONS(2596),
    [aux_sym_preproc_include_token2] = ACTIONS(2596),
    [aux_sym_preproc_def_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token2] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2596),
    [sym_preproc_directive] = ACTIONS(2596),
    [anon_sym_LPAREN2] = ACTIONS(2598),
    [anon_sym_BANG] = ACTIONS(2598),
    [anon_sym_TILDE] = ACTIONS(2598),
    [anon_sym_DASH] = ACTIONS(2596),
    [anon_sym_PLUS] = ACTIONS(2596),
    [anon_sym_STAR] = ACTIONS(2598),
    [anon_sym_CARET] = ACTIONS(2598),
    [anon_sym_AMP] = ACTIONS(2598),
    [anon_sym_SEMI] = ACTIONS(2598),
    [anon_sym___extension__] = ACTIONS(2596),
    [anon_sym_typedef] = ACTIONS(2596),
    [anon_sym_extern] = ACTIONS(2596),
    [anon_sym___attribute__] = ACTIONS(2596),
    [anon_sym___attribute] = ACTIONS(2596),
    [anon_sym_noreturn] = ACTIONS(2596),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym___declspec] = ACTIONS(2596),
    [anon_sym___cdecl] = ACTIONS(2596),
    [anon_sym___clrcall] = ACTIONS(2596),
    [anon_sym___stdcall] = ACTIONS(2596),
    [anon_sym___fastcall] = ACTIONS(2596),
    [anon_sym___thiscall] = ACTIONS(2596),
    [anon_sym___vectorcall] = ACTIONS(2596),
    [anon_sym_LBRACE] = ACTIONS(2598),
    [anon_sym_signed] = ACTIONS(2596),
    [anon_sym_unsigned] = ACTIONS(2596),
    [anon_sym_long] = ACTIONS(2596),
    [anon_sym_short] = ACTIONS(2596),
    [anon_sym_static] = ACTIONS(2596),
    [anon_sym_ATautoreleasepool] = ACTIONS(2598),
    [anon_sym_auto] = ACTIONS(2596),
    [anon_sym_register] = ACTIONS(2596),
    [anon_sym_inline] = ACTIONS(2596),
    [anon_sym___inline] = ACTIONS(2596),
    [anon_sym___inline__] = ACTIONS(2596),
    [anon_sym___forceinline] = ACTIONS(2596),
    [anon_sym_thread_local] = ACTIONS(2596),
    [anon_sym___thread] = ACTIONS(2596),
    [anon_sym_CG_EXTERN] = ACTIONS(2596),
    [anon_sym_CG_INLINE] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2596),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2596),
    [anon_sym_IBOutlet] = ACTIONS(2596),
    [anon_sym_IBInspectable] = ACTIONS(2596),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2596),
    [anon_sym_NS_INLINE] = ACTIONS(2596),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2596),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2596),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2596),
    [anon_sym_const] = ACTIONS(2596),
    [anon_sym_constexpr] = ACTIONS(2596),
    [anon_sym_volatile] = ACTIONS(2596),
    [anon_sym_restrict] = ACTIONS(2596),
    [anon_sym___restrict__] = ACTIONS(2596),
    [anon_sym__Atomic] = ACTIONS(2596),
    [anon_sym__Noreturn] = ACTIONS(2596),
    [anon_sym__Nonnull] = ACTIONS(2596),
    [anon_sym_nullable] = ACTIONS(2596),
    [anon_sym__Complex] = ACTIONS(2596),
    [anon_sym__Nullable] = ACTIONS(2596),
    [anon_sym__Nullable_result] = ACTIONS(2596),
    [anon_sym__Null_unspecified] = ACTIONS(2596),
    [anon_sym___autoreleasing] = ACTIONS(2596),
    [anon_sym___block] = ACTIONS(2596),
    [anon_sym___bridge] = ACTIONS(2596),
    [anon_sym___bridge_retained] = ACTIONS(2596),
    [anon_sym___bridge_transfer] = ACTIONS(2596),
    [anon_sym___complex] = ACTIONS(2596),
    [anon_sym___const] = ACTIONS(2596),
    [anon_sym___imag] = ACTIONS(2596),
    [anon_sym___kindof] = ACTIONS(2596),
    [anon_sym___nonnull] = ACTIONS(2596),
    [anon_sym___nullable] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2596),
    [anon_sym___real] = ACTIONS(2596),
    [anon_sym___strong] = ACTIONS(2596),
    [anon_sym___unsafe_unretained] = ACTIONS(2596),
    [anon_sym___unused] = ACTIONS(2596),
    [anon_sym___weak] = ACTIONS(2596),
    [anon_sym_alignas] = ACTIONS(2596),
    [anon_sym__Alignas] = ACTIONS(2596),
    [sym_primitive_type] = ACTIONS(2596),
    [anon_sym_enum] = ACTIONS(2596),
    [anon_sym_struct] = ACTIONS(2596),
    [anon_sym_union] = ACTIONS(2596),
    [anon_sym_if] = ACTIONS(2596),
    [anon_sym_switch] = ACTIONS(2596),
    [anon_sym_case] = ACTIONS(2596),
    [anon_sym_default] = ACTIONS(2596),
    [anon_sym_while] = ACTIONS(2596),
    [anon_sym_do] = ACTIONS(2596),
    [anon_sym_for] = ACTIONS(2596),
    [anon_sym_in] = ACTIONS(2596),
    [anon_sym_return] = ACTIONS(2596),
    [anon_sym_break] = ACTIONS(2596),
    [anon_sym_continue] = ACTIONS(2596),
    [anon_sym_goto] = ACTIONS(2596),
    [anon_sym_DASH_DASH] = ACTIONS(2598),
    [anon_sym_PLUS_PLUS] = ACTIONS(2598),
    [anon_sym_sizeof] = ACTIONS(2596),
    [anon_sym___alignof__] = ACTIONS(2596),
    [anon_sym___alignof] = ACTIONS(2596),
    [anon_sym__alignof] = ACTIONS(2596),
    [anon_sym_alignof] = ACTIONS(2596),
    [anon_sym__Alignof] = ACTIONS(2596),
    [anon_sym_offsetof] = ACTIONS(2596),
    [anon_sym__Generic] = ACTIONS(2596),
    [anon_sym_asm] = ACTIONS(2596),
    [anon_sym___asm__] = ACTIONS(2596),
    [anon_sym___asm] = ACTIONS(2596),
    [sym_number_literal] = ACTIONS(2598),
    [anon_sym_L_SQUOTE] = ACTIONS(2598),
    [anon_sym_u_SQUOTE] = ACTIONS(2598),
    [anon_sym_U_SQUOTE] = ACTIONS(2598),
    [anon_sym_u8_SQUOTE] = ACTIONS(2598),
    [anon_sym_SQUOTE] = ACTIONS(2598),
    [anon_sym_AT] = ACTIONS(2596),
    [anon_sym_DQUOTE] = ACTIONS(2598),
    [anon_sym_L_DQUOTE] = ACTIONS(2598),
    [anon_sym_u_DQUOTE] = ACTIONS(2598),
    [anon_sym_U_DQUOTE] = ACTIONS(2598),
    [anon_sym_u8_DQUOTE] = ACTIONS(2598),
    [sym_true] = ACTIONS(2596),
    [sym_false] = ACTIONS(2596),
    [anon_sym_NULL] = ACTIONS(2596),
    [anon_sym_nullptr] = ACTIONS(2596),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2596),
    [anon_sym___typeof] = ACTIONS(2596),
    [anon_sym_typeof] = ACTIONS(2596),
    [anon_sym_ATimport] = ACTIONS(2598),
    [aux_sym_preproc_undef_token1] = ACTIONS(2596),
    [anon_sym_POUND] = ACTIONS(2596),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2596),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2596),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2596),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_API_AVAILABLE] = ACTIONS(2596),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_API_DEPRECATED] = ACTIONS(2596),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym___deprecated_msg] = ACTIONS(2596),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2596),
    [anon_sym_ATprotocol] = ACTIONS(2598),
    [anon_sym_ATinterface] = ACTIONS(2598),
    [anon_sym_ATimplementation] = ACTIONS(2598),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2598),
    [anon_sym_ATsynthesize] = ACTIONS(2598),
    [anon_sym_ATdynamic] = ACTIONS(2598),
    [anon_sym_ATtry] = ACTIONS(2598),
    [anon_sym___try] = ACTIONS(2596),
    [anon_sym_ATthrow] = ACTIONS(2598),
    [anon_sym_ATselector] = ACTIONS(2598),
    [anon_sym_ATavailable] = ACTIONS(2598),
    [anon_sym___builtin_available] = ACTIONS(2596),
    [anon_sym_va_arg] = ACTIONS(2596),
    [anon_sym_ATencode] = ACTIONS(2598),
    [anon_sym_ATsynchronized] = ACTIONS(2598),
    [anon_sym_BOOL] = ACTIONS(2596),
    [anon_sym_IMP] = ACTIONS(2596),
    [anon_sym_SEL] = ACTIONS(2596),
    [anon_sym_Class] = ACTIONS(2596),
    [anon_sym_id] = ACTIONS(2596),
  },
  [STATE(650)] = {
    [sym_identifier] = ACTIONS(2600),
    [aux_sym_preproc_include_token1] = ACTIONS(2600),
    [aux_sym_preproc_include_token2] = ACTIONS(2600),
    [aux_sym_preproc_def_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token2] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2600),
    [sym_preproc_directive] = ACTIONS(2600),
    [anon_sym_LPAREN2] = ACTIONS(2602),
    [anon_sym_BANG] = ACTIONS(2602),
    [anon_sym_TILDE] = ACTIONS(2602),
    [anon_sym_DASH] = ACTIONS(2600),
    [anon_sym_PLUS] = ACTIONS(2600),
    [anon_sym_STAR] = ACTIONS(2602),
    [anon_sym_CARET] = ACTIONS(2602),
    [anon_sym_AMP] = ACTIONS(2602),
    [anon_sym_SEMI] = ACTIONS(2602),
    [anon_sym___extension__] = ACTIONS(2600),
    [anon_sym_typedef] = ACTIONS(2600),
    [anon_sym_extern] = ACTIONS(2600),
    [anon_sym___attribute__] = ACTIONS(2600),
    [anon_sym___attribute] = ACTIONS(2600),
    [anon_sym_noreturn] = ACTIONS(2600),
    [anon_sym_LBRACK] = ACTIONS(2602),
    [anon_sym___declspec] = ACTIONS(2600),
    [anon_sym___cdecl] = ACTIONS(2600),
    [anon_sym___clrcall] = ACTIONS(2600),
    [anon_sym___stdcall] = ACTIONS(2600),
    [anon_sym___fastcall] = ACTIONS(2600),
    [anon_sym___thiscall] = ACTIONS(2600),
    [anon_sym___vectorcall] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_signed] = ACTIONS(2600),
    [anon_sym_unsigned] = ACTIONS(2600),
    [anon_sym_long] = ACTIONS(2600),
    [anon_sym_short] = ACTIONS(2600),
    [anon_sym_static] = ACTIONS(2600),
    [anon_sym_ATautoreleasepool] = ACTIONS(2602),
    [anon_sym_auto] = ACTIONS(2600),
    [anon_sym_register] = ACTIONS(2600),
    [anon_sym_inline] = ACTIONS(2600),
    [anon_sym___inline] = ACTIONS(2600),
    [anon_sym___inline__] = ACTIONS(2600),
    [anon_sym___forceinline] = ACTIONS(2600),
    [anon_sym_thread_local] = ACTIONS(2600),
    [anon_sym___thread] = ACTIONS(2600),
    [anon_sym_CG_EXTERN] = ACTIONS(2600),
    [anon_sym_CG_INLINE] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2600),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2600),
    [anon_sym_IBOutlet] = ACTIONS(2600),
    [anon_sym_IBInspectable] = ACTIONS(2600),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2600),
    [anon_sym_NS_INLINE] = ACTIONS(2600),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2600),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2600),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2600),
    [anon_sym_const] = ACTIONS(2600),
    [anon_sym_constexpr] = ACTIONS(2600),
    [anon_sym_volatile] = ACTIONS(2600),
    [anon_sym_restrict] = ACTIONS(2600),
    [anon_sym___restrict__] = ACTIONS(2600),
    [anon_sym__Atomic] = ACTIONS(2600),
    [anon_sym__Noreturn] = ACTIONS(2600),
    [anon_sym__Nonnull] = ACTIONS(2600),
    [anon_sym_nullable] = ACTIONS(2600),
    [anon_sym__Complex] = ACTIONS(2600),
    [anon_sym__Nullable] = ACTIONS(2600),
    [anon_sym__Nullable_result] = ACTIONS(2600),
    [anon_sym__Null_unspecified] = ACTIONS(2600),
    [anon_sym___autoreleasing] = ACTIONS(2600),
    [anon_sym___block] = ACTIONS(2600),
    [anon_sym___bridge] = ACTIONS(2600),
    [anon_sym___bridge_retained] = ACTIONS(2600),
    [anon_sym___bridge_transfer] = ACTIONS(2600),
    [anon_sym___complex] = ACTIONS(2600),
    [anon_sym___const] = ACTIONS(2600),
    [anon_sym___imag] = ACTIONS(2600),
    [anon_sym___kindof] = ACTIONS(2600),
    [anon_sym___nonnull] = ACTIONS(2600),
    [anon_sym___nullable] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2600),
    [anon_sym___real] = ACTIONS(2600),
    [anon_sym___strong] = ACTIONS(2600),
    [anon_sym___unsafe_unretained] = ACTIONS(2600),
    [anon_sym___unused] = ACTIONS(2600),
    [anon_sym___weak] = ACTIONS(2600),
    [anon_sym_alignas] = ACTIONS(2600),
    [anon_sym__Alignas] = ACTIONS(2600),
    [sym_primitive_type] = ACTIONS(2600),
    [anon_sym_enum] = ACTIONS(2600),
    [anon_sym_struct] = ACTIONS(2600),
    [anon_sym_union] = ACTIONS(2600),
    [anon_sym_if] = ACTIONS(2600),
    [anon_sym_switch] = ACTIONS(2600),
    [anon_sym_case] = ACTIONS(2600),
    [anon_sym_default] = ACTIONS(2600),
    [anon_sym_while] = ACTIONS(2600),
    [anon_sym_do] = ACTIONS(2600),
    [anon_sym_for] = ACTIONS(2600),
    [anon_sym_in] = ACTIONS(2600),
    [anon_sym_return] = ACTIONS(2600),
    [anon_sym_break] = ACTIONS(2600),
    [anon_sym_continue] = ACTIONS(2600),
    [anon_sym_goto] = ACTIONS(2600),
    [anon_sym_DASH_DASH] = ACTIONS(2602),
    [anon_sym_PLUS_PLUS] = ACTIONS(2602),
    [anon_sym_sizeof] = ACTIONS(2600),
    [anon_sym___alignof__] = ACTIONS(2600),
    [anon_sym___alignof] = ACTIONS(2600),
    [anon_sym__alignof] = ACTIONS(2600),
    [anon_sym_alignof] = ACTIONS(2600),
    [anon_sym__Alignof] = ACTIONS(2600),
    [anon_sym_offsetof] = ACTIONS(2600),
    [anon_sym__Generic] = ACTIONS(2600),
    [anon_sym_asm] = ACTIONS(2600),
    [anon_sym___asm__] = ACTIONS(2600),
    [anon_sym___asm] = ACTIONS(2600),
    [sym_number_literal] = ACTIONS(2602),
    [anon_sym_L_SQUOTE] = ACTIONS(2602),
    [anon_sym_u_SQUOTE] = ACTIONS(2602),
    [anon_sym_U_SQUOTE] = ACTIONS(2602),
    [anon_sym_u8_SQUOTE] = ACTIONS(2602),
    [anon_sym_SQUOTE] = ACTIONS(2602),
    [anon_sym_AT] = ACTIONS(2600),
    [anon_sym_DQUOTE] = ACTIONS(2602),
    [anon_sym_L_DQUOTE] = ACTIONS(2602),
    [anon_sym_u_DQUOTE] = ACTIONS(2602),
    [anon_sym_U_DQUOTE] = ACTIONS(2602),
    [anon_sym_u8_DQUOTE] = ACTIONS(2602),
    [sym_true] = ACTIONS(2600),
    [sym_false] = ACTIONS(2600),
    [anon_sym_NULL] = ACTIONS(2600),
    [anon_sym_nullptr] = ACTIONS(2600),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2600),
    [anon_sym___typeof] = ACTIONS(2600),
    [anon_sym_typeof] = ACTIONS(2600),
    [anon_sym_ATimport] = ACTIONS(2602),
    [aux_sym_preproc_undef_token1] = ACTIONS(2600),
    [anon_sym_POUND] = ACTIONS(2600),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2600),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2600),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2600),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_API_AVAILABLE] = ACTIONS(2600),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_API_DEPRECATED] = ACTIONS(2600),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym___deprecated_msg] = ACTIONS(2600),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2600),
    [anon_sym_ATprotocol] = ACTIONS(2602),
    [anon_sym_ATinterface] = ACTIONS(2602),
    [anon_sym_ATimplementation] = ACTIONS(2602),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2602),
    [anon_sym_ATsynthesize] = ACTIONS(2602),
    [anon_sym_ATdynamic] = ACTIONS(2602),
    [anon_sym_ATtry] = ACTIONS(2602),
    [anon_sym___try] = ACTIONS(2600),
    [anon_sym_ATthrow] = ACTIONS(2602),
    [anon_sym_ATselector] = ACTIONS(2602),
    [anon_sym_ATavailable] = ACTIONS(2602),
    [anon_sym___builtin_available] = ACTIONS(2600),
    [anon_sym_va_arg] = ACTIONS(2600),
    [anon_sym_ATencode] = ACTIONS(2602),
    [anon_sym_ATsynchronized] = ACTIONS(2602),
    [anon_sym_BOOL] = ACTIONS(2600),
    [anon_sym_IMP] = ACTIONS(2600),
    [anon_sym_SEL] = ACTIONS(2600),
    [anon_sym_Class] = ACTIONS(2600),
    [anon_sym_id] = ACTIONS(2600),
  },
  [STATE(651)] = {
    [sym_identifier] = ACTIONS(2768),
    [aux_sym_preproc_include_token1] = ACTIONS(2768),
    [aux_sym_preproc_include_token2] = ACTIONS(2768),
    [aux_sym_preproc_def_token1] = ACTIONS(2768),
    [aux_sym_preproc_if_token1] = ACTIONS(2768),
    [aux_sym_preproc_if_token2] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2768),
    [sym_preproc_directive] = ACTIONS(2768),
    [anon_sym_LPAREN2] = ACTIONS(2770),
    [anon_sym_BANG] = ACTIONS(2770),
    [anon_sym_TILDE] = ACTIONS(2770),
    [anon_sym_DASH] = ACTIONS(2768),
    [anon_sym_PLUS] = ACTIONS(2768),
    [anon_sym_STAR] = ACTIONS(2770),
    [anon_sym_CARET] = ACTIONS(2770),
    [anon_sym_AMP] = ACTIONS(2770),
    [anon_sym_SEMI] = ACTIONS(2770),
    [anon_sym___extension__] = ACTIONS(2768),
    [anon_sym_typedef] = ACTIONS(2768),
    [anon_sym_extern] = ACTIONS(2768),
    [anon_sym___attribute__] = ACTIONS(2768),
    [anon_sym___attribute] = ACTIONS(2768),
    [anon_sym_noreturn] = ACTIONS(2768),
    [anon_sym_LBRACK] = ACTIONS(2770),
    [anon_sym___declspec] = ACTIONS(2768),
    [anon_sym___cdecl] = ACTIONS(2768),
    [anon_sym___clrcall] = ACTIONS(2768),
    [anon_sym___stdcall] = ACTIONS(2768),
    [anon_sym___fastcall] = ACTIONS(2768),
    [anon_sym___thiscall] = ACTIONS(2768),
    [anon_sym___vectorcall] = ACTIONS(2768),
    [anon_sym_LBRACE] = ACTIONS(2770),
    [anon_sym_signed] = ACTIONS(2768),
    [anon_sym_unsigned] = ACTIONS(2768),
    [anon_sym_long] = ACTIONS(2768),
    [anon_sym_short] = ACTIONS(2768),
    [anon_sym_static] = ACTIONS(2768),
    [anon_sym_ATautoreleasepool] = ACTIONS(2770),
    [anon_sym_auto] = ACTIONS(2768),
    [anon_sym_register] = ACTIONS(2768),
    [anon_sym_inline] = ACTIONS(2768),
    [anon_sym___inline] = ACTIONS(2768),
    [anon_sym___inline__] = ACTIONS(2768),
    [anon_sym___forceinline] = ACTIONS(2768),
    [anon_sym_thread_local] = ACTIONS(2768),
    [anon_sym___thread] = ACTIONS(2768),
    [anon_sym_CG_EXTERN] = ACTIONS(2768),
    [anon_sym_CG_INLINE] = ACTIONS(2768),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2768),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2768),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2768),
    [anon_sym_IBOutlet] = ACTIONS(2768),
    [anon_sym_IBInspectable] = ACTIONS(2768),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2768),
    [anon_sym_NS_INLINE] = ACTIONS(2768),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2768),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2768),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2768),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2768),
    [anon_sym_const] = ACTIONS(2768),
    [anon_sym_constexpr] = ACTIONS(2768),
    [anon_sym_volatile] = ACTIONS(2768),
    [anon_sym_restrict] = ACTIONS(2768),
    [anon_sym___restrict__] = ACTIONS(2768),
    [anon_sym__Atomic] = ACTIONS(2768),
    [anon_sym__Noreturn] = ACTIONS(2768),
    [anon_sym__Nonnull] = ACTIONS(2768),
    [anon_sym_nullable] = ACTIONS(2768),
    [anon_sym__Complex] = ACTIONS(2768),
    [anon_sym__Nullable] = ACTIONS(2768),
    [anon_sym__Nullable_result] = ACTIONS(2768),
    [anon_sym__Null_unspecified] = ACTIONS(2768),
    [anon_sym___autoreleasing] = ACTIONS(2768),
    [anon_sym___block] = ACTIONS(2768),
    [anon_sym___bridge] = ACTIONS(2768),
    [anon_sym___bridge_retained] = ACTIONS(2768),
    [anon_sym___bridge_transfer] = ACTIONS(2768),
    [anon_sym___complex] = ACTIONS(2768),
    [anon_sym___const] = ACTIONS(2768),
    [anon_sym___imag] = ACTIONS(2768),
    [anon_sym___kindof] = ACTIONS(2768),
    [anon_sym___nonnull] = ACTIONS(2768),
    [anon_sym___nullable] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2768),
    [anon_sym___real] = ACTIONS(2768),
    [anon_sym___strong] = ACTIONS(2768),
    [anon_sym___unsafe_unretained] = ACTIONS(2768),
    [anon_sym___unused] = ACTIONS(2768),
    [anon_sym___weak] = ACTIONS(2768),
    [anon_sym_alignas] = ACTIONS(2768),
    [anon_sym__Alignas] = ACTIONS(2768),
    [sym_primitive_type] = ACTIONS(2768),
    [anon_sym_enum] = ACTIONS(2768),
    [anon_sym_struct] = ACTIONS(2768),
    [anon_sym_union] = ACTIONS(2768),
    [anon_sym_if] = ACTIONS(2768),
    [anon_sym_switch] = ACTIONS(2768),
    [anon_sym_case] = ACTIONS(2768),
    [anon_sym_default] = ACTIONS(2768),
    [anon_sym_while] = ACTIONS(2768),
    [anon_sym_do] = ACTIONS(2768),
    [anon_sym_for] = ACTIONS(2768),
    [anon_sym_in] = ACTIONS(2768),
    [anon_sym_return] = ACTIONS(2768),
    [anon_sym_break] = ACTIONS(2768),
    [anon_sym_continue] = ACTIONS(2768),
    [anon_sym_goto] = ACTIONS(2768),
    [anon_sym_DASH_DASH] = ACTIONS(2770),
    [anon_sym_PLUS_PLUS] = ACTIONS(2770),
    [anon_sym_sizeof] = ACTIONS(2768),
    [anon_sym___alignof__] = ACTIONS(2768),
    [anon_sym___alignof] = ACTIONS(2768),
    [anon_sym__alignof] = ACTIONS(2768),
    [anon_sym_alignof] = ACTIONS(2768),
    [anon_sym__Alignof] = ACTIONS(2768),
    [anon_sym_offsetof] = ACTIONS(2768),
    [anon_sym__Generic] = ACTIONS(2768),
    [anon_sym_asm] = ACTIONS(2768),
    [anon_sym___asm__] = ACTIONS(2768),
    [anon_sym___asm] = ACTIONS(2768),
    [sym_number_literal] = ACTIONS(2770),
    [anon_sym_L_SQUOTE] = ACTIONS(2770),
    [anon_sym_u_SQUOTE] = ACTIONS(2770),
    [anon_sym_U_SQUOTE] = ACTIONS(2770),
    [anon_sym_u8_SQUOTE] = ACTIONS(2770),
    [anon_sym_SQUOTE] = ACTIONS(2770),
    [anon_sym_AT] = ACTIONS(2768),
    [anon_sym_DQUOTE] = ACTIONS(2770),
    [anon_sym_L_DQUOTE] = ACTIONS(2770),
    [anon_sym_u_DQUOTE] = ACTIONS(2770),
    [anon_sym_U_DQUOTE] = ACTIONS(2770),
    [anon_sym_u8_DQUOTE] = ACTIONS(2770),
    [sym_true] = ACTIONS(2768),
    [sym_false] = ACTIONS(2768),
    [anon_sym_NULL] = ACTIONS(2768),
    [anon_sym_nullptr] = ACTIONS(2768),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2768),
    [anon_sym___typeof] = ACTIONS(2768),
    [anon_sym_typeof] = ACTIONS(2768),
    [anon_sym_ATimport] = ACTIONS(2770),
    [aux_sym_preproc_undef_token1] = ACTIONS(2768),
    [anon_sym_POUND] = ACTIONS(2768),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2768),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2768),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2768),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2768),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2768),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2768),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2768),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_API_AVAILABLE] = ACTIONS(2768),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_API_DEPRECATED] = ACTIONS(2768),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2768),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym___deprecated_msg] = ACTIONS(2768),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2768),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2768),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2768),
    [anon_sym_ATprotocol] = ACTIONS(2770),
    [anon_sym_ATinterface] = ACTIONS(2770),
    [anon_sym_ATimplementation] = ACTIONS(2770),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2770),
    [anon_sym_ATsynthesize] = ACTIONS(2770),
    [anon_sym_ATdynamic] = ACTIONS(2770),
    [anon_sym_ATtry] = ACTIONS(2770),
    [anon_sym___try] = ACTIONS(2768),
    [anon_sym_ATthrow] = ACTIONS(2770),
    [anon_sym_ATselector] = ACTIONS(2770),
    [anon_sym_ATavailable] = ACTIONS(2770),
    [anon_sym___builtin_available] = ACTIONS(2768),
    [anon_sym_va_arg] = ACTIONS(2768),
    [anon_sym_ATencode] = ACTIONS(2770),
    [anon_sym_ATsynchronized] = ACTIONS(2770),
    [anon_sym_BOOL] = ACTIONS(2768),
    [anon_sym_IMP] = ACTIONS(2768),
    [anon_sym_SEL] = ACTIONS(2768),
    [anon_sym_Class] = ACTIONS(2768),
    [anon_sym_id] = ACTIONS(2768),
  },
  [STATE(652)] = {
    [sym_identifier] = ACTIONS(2772),
    [aux_sym_preproc_include_token1] = ACTIONS(2772),
    [aux_sym_preproc_include_token2] = ACTIONS(2772),
    [aux_sym_preproc_def_token1] = ACTIONS(2772),
    [aux_sym_preproc_if_token1] = ACTIONS(2772),
    [aux_sym_preproc_if_token2] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2772),
    [sym_preproc_directive] = ACTIONS(2772),
    [anon_sym_LPAREN2] = ACTIONS(2774),
    [anon_sym_BANG] = ACTIONS(2774),
    [anon_sym_TILDE] = ACTIONS(2774),
    [anon_sym_DASH] = ACTIONS(2772),
    [anon_sym_PLUS] = ACTIONS(2772),
    [anon_sym_STAR] = ACTIONS(2774),
    [anon_sym_CARET] = ACTIONS(2774),
    [anon_sym_AMP] = ACTIONS(2774),
    [anon_sym_SEMI] = ACTIONS(2774),
    [anon_sym___extension__] = ACTIONS(2772),
    [anon_sym_typedef] = ACTIONS(2772),
    [anon_sym_extern] = ACTIONS(2772),
    [anon_sym___attribute__] = ACTIONS(2772),
    [anon_sym___attribute] = ACTIONS(2772),
    [anon_sym_noreturn] = ACTIONS(2772),
    [anon_sym_LBRACK] = ACTIONS(2774),
    [anon_sym___declspec] = ACTIONS(2772),
    [anon_sym___cdecl] = ACTIONS(2772),
    [anon_sym___clrcall] = ACTIONS(2772),
    [anon_sym___stdcall] = ACTIONS(2772),
    [anon_sym___fastcall] = ACTIONS(2772),
    [anon_sym___thiscall] = ACTIONS(2772),
    [anon_sym___vectorcall] = ACTIONS(2772),
    [anon_sym_LBRACE] = ACTIONS(2774),
    [anon_sym_signed] = ACTIONS(2772),
    [anon_sym_unsigned] = ACTIONS(2772),
    [anon_sym_long] = ACTIONS(2772),
    [anon_sym_short] = ACTIONS(2772),
    [anon_sym_static] = ACTIONS(2772),
    [anon_sym_ATautoreleasepool] = ACTIONS(2774),
    [anon_sym_auto] = ACTIONS(2772),
    [anon_sym_register] = ACTIONS(2772),
    [anon_sym_inline] = ACTIONS(2772),
    [anon_sym___inline] = ACTIONS(2772),
    [anon_sym___inline__] = ACTIONS(2772),
    [anon_sym___forceinline] = ACTIONS(2772),
    [anon_sym_thread_local] = ACTIONS(2772),
    [anon_sym___thread] = ACTIONS(2772),
    [anon_sym_CG_EXTERN] = ACTIONS(2772),
    [anon_sym_CG_INLINE] = ACTIONS(2772),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2772),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2772),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2772),
    [anon_sym_IBOutlet] = ACTIONS(2772),
    [anon_sym_IBInspectable] = ACTIONS(2772),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2772),
    [anon_sym_NS_INLINE] = ACTIONS(2772),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2772),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2772),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2772),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2772),
    [anon_sym_const] = ACTIONS(2772),
    [anon_sym_constexpr] = ACTIONS(2772),
    [anon_sym_volatile] = ACTIONS(2772),
    [anon_sym_restrict] = ACTIONS(2772),
    [anon_sym___restrict__] = ACTIONS(2772),
    [anon_sym__Atomic] = ACTIONS(2772),
    [anon_sym__Noreturn] = ACTIONS(2772),
    [anon_sym__Nonnull] = ACTIONS(2772),
    [anon_sym_nullable] = ACTIONS(2772),
    [anon_sym__Complex] = ACTIONS(2772),
    [anon_sym__Nullable] = ACTIONS(2772),
    [anon_sym__Nullable_result] = ACTIONS(2772),
    [anon_sym__Null_unspecified] = ACTIONS(2772),
    [anon_sym___autoreleasing] = ACTIONS(2772),
    [anon_sym___block] = ACTIONS(2772),
    [anon_sym___bridge] = ACTIONS(2772),
    [anon_sym___bridge_retained] = ACTIONS(2772),
    [anon_sym___bridge_transfer] = ACTIONS(2772),
    [anon_sym___complex] = ACTIONS(2772),
    [anon_sym___const] = ACTIONS(2772),
    [anon_sym___imag] = ACTIONS(2772),
    [anon_sym___kindof] = ACTIONS(2772),
    [anon_sym___nonnull] = ACTIONS(2772),
    [anon_sym___nullable] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2772),
    [anon_sym___real] = ACTIONS(2772),
    [anon_sym___strong] = ACTIONS(2772),
    [anon_sym___unsafe_unretained] = ACTIONS(2772),
    [anon_sym___unused] = ACTIONS(2772),
    [anon_sym___weak] = ACTIONS(2772),
    [anon_sym_alignas] = ACTIONS(2772),
    [anon_sym__Alignas] = ACTIONS(2772),
    [sym_primitive_type] = ACTIONS(2772),
    [anon_sym_enum] = ACTIONS(2772),
    [anon_sym_struct] = ACTIONS(2772),
    [anon_sym_union] = ACTIONS(2772),
    [anon_sym_if] = ACTIONS(2772),
    [anon_sym_switch] = ACTIONS(2772),
    [anon_sym_case] = ACTIONS(2772),
    [anon_sym_default] = ACTIONS(2772),
    [anon_sym_while] = ACTIONS(2772),
    [anon_sym_do] = ACTIONS(2772),
    [anon_sym_for] = ACTIONS(2772),
    [anon_sym_in] = ACTIONS(2772),
    [anon_sym_return] = ACTIONS(2772),
    [anon_sym_break] = ACTIONS(2772),
    [anon_sym_continue] = ACTIONS(2772),
    [anon_sym_goto] = ACTIONS(2772),
    [anon_sym_DASH_DASH] = ACTIONS(2774),
    [anon_sym_PLUS_PLUS] = ACTIONS(2774),
    [anon_sym_sizeof] = ACTIONS(2772),
    [anon_sym___alignof__] = ACTIONS(2772),
    [anon_sym___alignof] = ACTIONS(2772),
    [anon_sym__alignof] = ACTIONS(2772),
    [anon_sym_alignof] = ACTIONS(2772),
    [anon_sym__Alignof] = ACTIONS(2772),
    [anon_sym_offsetof] = ACTIONS(2772),
    [anon_sym__Generic] = ACTIONS(2772),
    [anon_sym_asm] = ACTIONS(2772),
    [anon_sym___asm__] = ACTIONS(2772),
    [anon_sym___asm] = ACTIONS(2772),
    [sym_number_literal] = ACTIONS(2774),
    [anon_sym_L_SQUOTE] = ACTIONS(2774),
    [anon_sym_u_SQUOTE] = ACTIONS(2774),
    [anon_sym_U_SQUOTE] = ACTIONS(2774),
    [anon_sym_u8_SQUOTE] = ACTIONS(2774),
    [anon_sym_SQUOTE] = ACTIONS(2774),
    [anon_sym_AT] = ACTIONS(2772),
    [anon_sym_DQUOTE] = ACTIONS(2774),
    [anon_sym_L_DQUOTE] = ACTIONS(2774),
    [anon_sym_u_DQUOTE] = ACTIONS(2774),
    [anon_sym_U_DQUOTE] = ACTIONS(2774),
    [anon_sym_u8_DQUOTE] = ACTIONS(2774),
    [sym_true] = ACTIONS(2772),
    [sym_false] = ACTIONS(2772),
    [anon_sym_NULL] = ACTIONS(2772),
    [anon_sym_nullptr] = ACTIONS(2772),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2772),
    [anon_sym___typeof] = ACTIONS(2772),
    [anon_sym_typeof] = ACTIONS(2772),
    [anon_sym_ATimport] = ACTIONS(2774),
    [aux_sym_preproc_undef_token1] = ACTIONS(2772),
    [anon_sym_POUND] = ACTIONS(2772),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2772),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2772),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2772),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2772),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2772),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2772),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2772),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_API_AVAILABLE] = ACTIONS(2772),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_API_DEPRECATED] = ACTIONS(2772),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2772),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym___deprecated_msg] = ACTIONS(2772),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2772),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2772),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2772),
    [anon_sym_ATprotocol] = ACTIONS(2774),
    [anon_sym_ATinterface] = ACTIONS(2774),
    [anon_sym_ATimplementation] = ACTIONS(2774),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2774),
    [anon_sym_ATsynthesize] = ACTIONS(2774),
    [anon_sym_ATdynamic] = ACTIONS(2774),
    [anon_sym_ATtry] = ACTIONS(2774),
    [anon_sym___try] = ACTIONS(2772),
    [anon_sym_ATthrow] = ACTIONS(2774),
    [anon_sym_ATselector] = ACTIONS(2774),
    [anon_sym_ATavailable] = ACTIONS(2774),
    [anon_sym___builtin_available] = ACTIONS(2772),
    [anon_sym_va_arg] = ACTIONS(2772),
    [anon_sym_ATencode] = ACTIONS(2774),
    [anon_sym_ATsynchronized] = ACTIONS(2774),
    [anon_sym_BOOL] = ACTIONS(2772),
    [anon_sym_IMP] = ACTIONS(2772),
    [anon_sym_SEL] = ACTIONS(2772),
    [anon_sym_Class] = ACTIONS(2772),
    [anon_sym_id] = ACTIONS(2772),
  },
  [STATE(653)] = {
    [sym_identifier] = ACTIONS(2824),
    [aux_sym_preproc_include_token1] = ACTIONS(2824),
    [aux_sym_preproc_include_token2] = ACTIONS(2824),
    [aux_sym_preproc_def_token1] = ACTIONS(2824),
    [aux_sym_preproc_if_token1] = ACTIONS(2824),
    [aux_sym_preproc_if_token2] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2824),
    [sym_preproc_directive] = ACTIONS(2824),
    [anon_sym_LPAREN2] = ACTIONS(2826),
    [anon_sym_BANG] = ACTIONS(2826),
    [anon_sym_TILDE] = ACTIONS(2826),
    [anon_sym_DASH] = ACTIONS(2824),
    [anon_sym_PLUS] = ACTIONS(2824),
    [anon_sym_STAR] = ACTIONS(2826),
    [anon_sym_CARET] = ACTIONS(2826),
    [anon_sym_AMP] = ACTIONS(2826),
    [anon_sym_SEMI] = ACTIONS(2826),
    [anon_sym___extension__] = ACTIONS(2824),
    [anon_sym_typedef] = ACTIONS(2824),
    [anon_sym_extern] = ACTIONS(2824),
    [anon_sym___attribute__] = ACTIONS(2824),
    [anon_sym___attribute] = ACTIONS(2824),
    [anon_sym_noreturn] = ACTIONS(2824),
    [anon_sym_LBRACK] = ACTIONS(2826),
    [anon_sym___declspec] = ACTIONS(2824),
    [anon_sym___cdecl] = ACTIONS(2824),
    [anon_sym___clrcall] = ACTIONS(2824),
    [anon_sym___stdcall] = ACTIONS(2824),
    [anon_sym___fastcall] = ACTIONS(2824),
    [anon_sym___thiscall] = ACTIONS(2824),
    [anon_sym___vectorcall] = ACTIONS(2824),
    [anon_sym_LBRACE] = ACTIONS(2826),
    [anon_sym_signed] = ACTIONS(2824),
    [anon_sym_unsigned] = ACTIONS(2824),
    [anon_sym_long] = ACTIONS(2824),
    [anon_sym_short] = ACTIONS(2824),
    [anon_sym_static] = ACTIONS(2824),
    [anon_sym_ATautoreleasepool] = ACTIONS(2826),
    [anon_sym_auto] = ACTIONS(2824),
    [anon_sym_register] = ACTIONS(2824),
    [anon_sym_inline] = ACTIONS(2824),
    [anon_sym___inline] = ACTIONS(2824),
    [anon_sym___inline__] = ACTIONS(2824),
    [anon_sym___forceinline] = ACTIONS(2824),
    [anon_sym_thread_local] = ACTIONS(2824),
    [anon_sym___thread] = ACTIONS(2824),
    [anon_sym_CG_EXTERN] = ACTIONS(2824),
    [anon_sym_CG_INLINE] = ACTIONS(2824),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2824),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2824),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2824),
    [anon_sym_IBOutlet] = ACTIONS(2824),
    [anon_sym_IBInspectable] = ACTIONS(2824),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2824),
    [anon_sym_NS_INLINE] = ACTIONS(2824),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2824),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2824),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2824),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2824),
    [anon_sym_const] = ACTIONS(2824),
    [anon_sym_constexpr] = ACTIONS(2824),
    [anon_sym_volatile] = ACTIONS(2824),
    [anon_sym_restrict] = ACTIONS(2824),
    [anon_sym___restrict__] = ACTIONS(2824),
    [anon_sym__Atomic] = ACTIONS(2824),
    [anon_sym__Noreturn] = ACTIONS(2824),
    [anon_sym__Nonnull] = ACTIONS(2824),
    [anon_sym_nullable] = ACTIONS(2824),
    [anon_sym__Complex] = ACTIONS(2824),
    [anon_sym__Nullable] = ACTIONS(2824),
    [anon_sym__Nullable_result] = ACTIONS(2824),
    [anon_sym__Null_unspecified] = ACTIONS(2824),
    [anon_sym___autoreleasing] = ACTIONS(2824),
    [anon_sym___block] = ACTIONS(2824),
    [anon_sym___bridge] = ACTIONS(2824),
    [anon_sym___bridge_retained] = ACTIONS(2824),
    [anon_sym___bridge_transfer] = ACTIONS(2824),
    [anon_sym___complex] = ACTIONS(2824),
    [anon_sym___const] = ACTIONS(2824),
    [anon_sym___imag] = ACTIONS(2824),
    [anon_sym___kindof] = ACTIONS(2824),
    [anon_sym___nonnull] = ACTIONS(2824),
    [anon_sym___nullable] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2824),
    [anon_sym___real] = ACTIONS(2824),
    [anon_sym___strong] = ACTIONS(2824),
    [anon_sym___unsafe_unretained] = ACTIONS(2824),
    [anon_sym___unused] = ACTIONS(2824),
    [anon_sym___weak] = ACTIONS(2824),
    [anon_sym_alignas] = ACTIONS(2824),
    [anon_sym__Alignas] = ACTIONS(2824),
    [sym_primitive_type] = ACTIONS(2824),
    [anon_sym_enum] = ACTIONS(2824),
    [anon_sym_struct] = ACTIONS(2824),
    [anon_sym_union] = ACTIONS(2824),
    [anon_sym_if] = ACTIONS(2824),
    [anon_sym_switch] = ACTIONS(2824),
    [anon_sym_case] = ACTIONS(2824),
    [anon_sym_default] = ACTIONS(2824),
    [anon_sym_while] = ACTIONS(2824),
    [anon_sym_do] = ACTIONS(2824),
    [anon_sym_for] = ACTIONS(2824),
    [anon_sym_in] = ACTIONS(2824),
    [anon_sym_return] = ACTIONS(2824),
    [anon_sym_break] = ACTIONS(2824),
    [anon_sym_continue] = ACTIONS(2824),
    [anon_sym_goto] = ACTIONS(2824),
    [anon_sym_DASH_DASH] = ACTIONS(2826),
    [anon_sym_PLUS_PLUS] = ACTIONS(2826),
    [anon_sym_sizeof] = ACTIONS(2824),
    [anon_sym___alignof__] = ACTIONS(2824),
    [anon_sym___alignof] = ACTIONS(2824),
    [anon_sym__alignof] = ACTIONS(2824),
    [anon_sym_alignof] = ACTIONS(2824),
    [anon_sym__Alignof] = ACTIONS(2824),
    [anon_sym_offsetof] = ACTIONS(2824),
    [anon_sym__Generic] = ACTIONS(2824),
    [anon_sym_asm] = ACTIONS(2824),
    [anon_sym___asm__] = ACTIONS(2824),
    [anon_sym___asm] = ACTIONS(2824),
    [sym_number_literal] = ACTIONS(2826),
    [anon_sym_L_SQUOTE] = ACTIONS(2826),
    [anon_sym_u_SQUOTE] = ACTIONS(2826),
    [anon_sym_U_SQUOTE] = ACTIONS(2826),
    [anon_sym_u8_SQUOTE] = ACTIONS(2826),
    [anon_sym_SQUOTE] = ACTIONS(2826),
    [anon_sym_AT] = ACTIONS(2824),
    [anon_sym_DQUOTE] = ACTIONS(2826),
    [anon_sym_L_DQUOTE] = ACTIONS(2826),
    [anon_sym_u_DQUOTE] = ACTIONS(2826),
    [anon_sym_U_DQUOTE] = ACTIONS(2826),
    [anon_sym_u8_DQUOTE] = ACTIONS(2826),
    [sym_true] = ACTIONS(2824),
    [sym_false] = ACTIONS(2824),
    [anon_sym_NULL] = ACTIONS(2824),
    [anon_sym_nullptr] = ACTIONS(2824),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2824),
    [anon_sym___typeof] = ACTIONS(2824),
    [anon_sym_typeof] = ACTIONS(2824),
    [anon_sym_ATimport] = ACTIONS(2826),
    [aux_sym_preproc_undef_token1] = ACTIONS(2824),
    [anon_sym_POUND] = ACTIONS(2824),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2824),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2824),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2824),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2824),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2824),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2824),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2824),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_API_AVAILABLE] = ACTIONS(2824),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_API_DEPRECATED] = ACTIONS(2824),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2824),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym___deprecated_msg] = ACTIONS(2824),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2824),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2824),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2824),
    [anon_sym_ATprotocol] = ACTIONS(2826),
    [anon_sym_ATinterface] = ACTIONS(2826),
    [anon_sym_ATimplementation] = ACTIONS(2826),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2826),
    [anon_sym_ATsynthesize] = ACTIONS(2826),
    [anon_sym_ATdynamic] = ACTIONS(2826),
    [anon_sym_ATtry] = ACTIONS(2826),
    [anon_sym___try] = ACTIONS(2824),
    [anon_sym_ATthrow] = ACTIONS(2826),
    [anon_sym_ATselector] = ACTIONS(2826),
    [anon_sym_ATavailable] = ACTIONS(2826),
    [anon_sym___builtin_available] = ACTIONS(2824),
    [anon_sym_va_arg] = ACTIONS(2824),
    [anon_sym_ATencode] = ACTIONS(2826),
    [anon_sym_ATsynchronized] = ACTIONS(2826),
    [anon_sym_BOOL] = ACTIONS(2824),
    [anon_sym_IMP] = ACTIONS(2824),
    [anon_sym_SEL] = ACTIONS(2824),
    [anon_sym_Class] = ACTIONS(2824),
    [anon_sym_id] = ACTIONS(2824),
  },
  [STATE(654)] = {
    [sym_identifier] = ACTIONS(2658),
    [aux_sym_preproc_include_token1] = ACTIONS(2658),
    [aux_sym_preproc_include_token2] = ACTIONS(2658),
    [aux_sym_preproc_def_token1] = ACTIONS(2658),
    [aux_sym_preproc_if_token1] = ACTIONS(2658),
    [aux_sym_preproc_if_token2] = ACTIONS(2658),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2658),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2658),
    [sym_preproc_directive] = ACTIONS(2658),
    [anon_sym_LPAREN2] = ACTIONS(2660),
    [anon_sym_BANG] = ACTIONS(2660),
    [anon_sym_TILDE] = ACTIONS(2660),
    [anon_sym_DASH] = ACTIONS(2658),
    [anon_sym_PLUS] = ACTIONS(2658),
    [anon_sym_STAR] = ACTIONS(2660),
    [anon_sym_CARET] = ACTIONS(2660),
    [anon_sym_AMP] = ACTIONS(2660),
    [anon_sym_SEMI] = ACTIONS(2660),
    [anon_sym___extension__] = ACTIONS(2658),
    [anon_sym_typedef] = ACTIONS(2658),
    [anon_sym_extern] = ACTIONS(2658),
    [anon_sym___attribute__] = ACTIONS(2658),
    [anon_sym___attribute] = ACTIONS(2658),
    [anon_sym_noreturn] = ACTIONS(2658),
    [anon_sym_LBRACK] = ACTIONS(2660),
    [anon_sym___declspec] = ACTIONS(2658),
    [anon_sym___cdecl] = ACTIONS(2658),
    [anon_sym___clrcall] = ACTIONS(2658),
    [anon_sym___stdcall] = ACTIONS(2658),
    [anon_sym___fastcall] = ACTIONS(2658),
    [anon_sym___thiscall] = ACTIONS(2658),
    [anon_sym___vectorcall] = ACTIONS(2658),
    [anon_sym_LBRACE] = ACTIONS(2660),
    [anon_sym_signed] = ACTIONS(2658),
    [anon_sym_unsigned] = ACTIONS(2658),
    [anon_sym_long] = ACTIONS(2658),
    [anon_sym_short] = ACTIONS(2658),
    [anon_sym_static] = ACTIONS(2658),
    [anon_sym_ATautoreleasepool] = ACTIONS(2660),
    [anon_sym_auto] = ACTIONS(2658),
    [anon_sym_register] = ACTIONS(2658),
    [anon_sym_inline] = ACTIONS(2658),
    [anon_sym___inline] = ACTIONS(2658),
    [anon_sym___inline__] = ACTIONS(2658),
    [anon_sym___forceinline] = ACTIONS(2658),
    [anon_sym_thread_local] = ACTIONS(2658),
    [anon_sym___thread] = ACTIONS(2658),
    [anon_sym_CG_EXTERN] = ACTIONS(2658),
    [anon_sym_CG_INLINE] = ACTIONS(2658),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2658),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2658),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2658),
    [anon_sym_IBOutlet] = ACTIONS(2658),
    [anon_sym_IBInspectable] = ACTIONS(2658),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2658),
    [anon_sym_NS_INLINE] = ACTIONS(2658),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2658),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2658),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2658),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2658),
    [anon_sym_const] = ACTIONS(2658),
    [anon_sym_constexpr] = ACTIONS(2658),
    [anon_sym_volatile] = ACTIONS(2658),
    [anon_sym_restrict] = ACTIONS(2658),
    [anon_sym___restrict__] = ACTIONS(2658),
    [anon_sym__Atomic] = ACTIONS(2658),
    [anon_sym__Noreturn] = ACTIONS(2658),
    [anon_sym__Nonnull] = ACTIONS(2658),
    [anon_sym_nullable] = ACTIONS(2658),
    [anon_sym__Complex] = ACTIONS(2658),
    [anon_sym__Nullable] = ACTIONS(2658),
    [anon_sym__Nullable_result] = ACTIONS(2658),
    [anon_sym__Null_unspecified] = ACTIONS(2658),
    [anon_sym___autoreleasing] = ACTIONS(2658),
    [anon_sym___block] = ACTIONS(2658),
    [anon_sym___bridge] = ACTIONS(2658),
    [anon_sym___bridge_retained] = ACTIONS(2658),
    [anon_sym___bridge_transfer] = ACTIONS(2658),
    [anon_sym___complex] = ACTIONS(2658),
    [anon_sym___const] = ACTIONS(2658),
    [anon_sym___imag] = ACTIONS(2658),
    [anon_sym___kindof] = ACTIONS(2658),
    [anon_sym___nonnull] = ACTIONS(2658),
    [anon_sym___nullable] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2658),
    [anon_sym___real] = ACTIONS(2658),
    [anon_sym___strong] = ACTIONS(2658),
    [anon_sym___unsafe_unretained] = ACTIONS(2658),
    [anon_sym___unused] = ACTIONS(2658),
    [anon_sym___weak] = ACTIONS(2658),
    [anon_sym_alignas] = ACTIONS(2658),
    [anon_sym__Alignas] = ACTIONS(2658),
    [sym_primitive_type] = ACTIONS(2658),
    [anon_sym_enum] = ACTIONS(2658),
    [anon_sym_struct] = ACTIONS(2658),
    [anon_sym_union] = ACTIONS(2658),
    [anon_sym_if] = ACTIONS(2658),
    [anon_sym_switch] = ACTIONS(2658),
    [anon_sym_case] = ACTIONS(2658),
    [anon_sym_default] = ACTIONS(2658),
    [anon_sym_while] = ACTIONS(2658),
    [anon_sym_do] = ACTIONS(2658),
    [anon_sym_for] = ACTIONS(2658),
    [anon_sym_in] = ACTIONS(2658),
    [anon_sym_return] = ACTIONS(2658),
    [anon_sym_break] = ACTIONS(2658),
    [anon_sym_continue] = ACTIONS(2658),
    [anon_sym_goto] = ACTIONS(2658),
    [anon_sym_DASH_DASH] = ACTIONS(2660),
    [anon_sym_PLUS_PLUS] = ACTIONS(2660),
    [anon_sym_sizeof] = ACTIONS(2658),
    [anon_sym___alignof__] = ACTIONS(2658),
    [anon_sym___alignof] = ACTIONS(2658),
    [anon_sym__alignof] = ACTIONS(2658),
    [anon_sym_alignof] = ACTIONS(2658),
    [anon_sym__Alignof] = ACTIONS(2658),
    [anon_sym_offsetof] = ACTIONS(2658),
    [anon_sym__Generic] = ACTIONS(2658),
    [anon_sym_asm] = ACTIONS(2658),
    [anon_sym___asm__] = ACTIONS(2658),
    [anon_sym___asm] = ACTIONS(2658),
    [sym_number_literal] = ACTIONS(2660),
    [anon_sym_L_SQUOTE] = ACTIONS(2660),
    [anon_sym_u_SQUOTE] = ACTIONS(2660),
    [anon_sym_U_SQUOTE] = ACTIONS(2660),
    [anon_sym_u8_SQUOTE] = ACTIONS(2660),
    [anon_sym_SQUOTE] = ACTIONS(2660),
    [anon_sym_AT] = ACTIONS(2658),
    [anon_sym_DQUOTE] = ACTIONS(2660),
    [anon_sym_L_DQUOTE] = ACTIONS(2660),
    [anon_sym_u_DQUOTE] = ACTIONS(2660),
    [anon_sym_U_DQUOTE] = ACTIONS(2660),
    [anon_sym_u8_DQUOTE] = ACTIONS(2660),
    [sym_true] = ACTIONS(2658),
    [sym_false] = ACTIONS(2658),
    [anon_sym_NULL] = ACTIONS(2658),
    [anon_sym_nullptr] = ACTIONS(2658),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2658),
    [anon_sym___typeof] = ACTIONS(2658),
    [anon_sym_typeof] = ACTIONS(2658),
    [anon_sym_ATimport] = ACTIONS(2660),
    [aux_sym_preproc_undef_token1] = ACTIONS(2658),
    [anon_sym_POUND] = ACTIONS(2658),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2658),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2658),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2658),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2658),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2658),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2658),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2658),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_API_AVAILABLE] = ACTIONS(2658),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_API_DEPRECATED] = ACTIONS(2658),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2658),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym___deprecated_msg] = ACTIONS(2658),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2658),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2658),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2658),
    [anon_sym_ATprotocol] = ACTIONS(2660),
    [anon_sym_ATinterface] = ACTIONS(2660),
    [anon_sym_ATimplementation] = ACTIONS(2660),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2660),
    [anon_sym_ATsynthesize] = ACTIONS(2660),
    [anon_sym_ATdynamic] = ACTIONS(2660),
    [anon_sym_ATtry] = ACTIONS(2660),
    [anon_sym___try] = ACTIONS(2658),
    [anon_sym_ATthrow] = ACTIONS(2660),
    [anon_sym_ATselector] = ACTIONS(2660),
    [anon_sym_ATavailable] = ACTIONS(2660),
    [anon_sym___builtin_available] = ACTIONS(2658),
    [anon_sym_va_arg] = ACTIONS(2658),
    [anon_sym_ATencode] = ACTIONS(2660),
    [anon_sym_ATsynchronized] = ACTIONS(2660),
    [anon_sym_BOOL] = ACTIONS(2658),
    [anon_sym_IMP] = ACTIONS(2658),
    [anon_sym_SEL] = ACTIONS(2658),
    [anon_sym_Class] = ACTIONS(2658),
    [anon_sym_id] = ACTIONS(2658),
  },
  [STATE(655)] = {
    [sym_identifier] = ACTIONS(2662),
    [aux_sym_preproc_include_token1] = ACTIONS(2662),
    [aux_sym_preproc_include_token2] = ACTIONS(2662),
    [aux_sym_preproc_def_token1] = ACTIONS(2662),
    [aux_sym_preproc_if_token1] = ACTIONS(2662),
    [aux_sym_preproc_if_token2] = ACTIONS(2662),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2662),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2662),
    [sym_preproc_directive] = ACTIONS(2662),
    [anon_sym_LPAREN2] = ACTIONS(2665),
    [anon_sym_BANG] = ACTIONS(2665),
    [anon_sym_TILDE] = ACTIONS(2665),
    [anon_sym_DASH] = ACTIONS(2662),
    [anon_sym_PLUS] = ACTIONS(2662),
    [anon_sym_STAR] = ACTIONS(2665),
    [anon_sym_CARET] = ACTIONS(2665),
    [anon_sym_AMP] = ACTIONS(2665),
    [anon_sym_SEMI] = ACTIONS(2665),
    [anon_sym___extension__] = ACTIONS(2662),
    [anon_sym_typedef] = ACTIONS(2662),
    [anon_sym_extern] = ACTIONS(2662),
    [anon_sym___attribute__] = ACTIONS(2662),
    [anon_sym___attribute] = ACTIONS(2662),
    [anon_sym_noreturn] = ACTIONS(2662),
    [anon_sym_LBRACK] = ACTIONS(2665),
    [anon_sym___declspec] = ACTIONS(2662),
    [anon_sym___cdecl] = ACTIONS(2662),
    [anon_sym___clrcall] = ACTIONS(2662),
    [anon_sym___stdcall] = ACTIONS(2662),
    [anon_sym___fastcall] = ACTIONS(2662),
    [anon_sym___thiscall] = ACTIONS(2662),
    [anon_sym___vectorcall] = ACTIONS(2662),
    [anon_sym_LBRACE] = ACTIONS(2665),
    [anon_sym_signed] = ACTIONS(2662),
    [anon_sym_unsigned] = ACTIONS(2662),
    [anon_sym_long] = ACTIONS(2662),
    [anon_sym_short] = ACTIONS(2662),
    [anon_sym_static] = ACTIONS(2662),
    [anon_sym_ATautoreleasepool] = ACTIONS(2665),
    [anon_sym_auto] = ACTIONS(2662),
    [anon_sym_register] = ACTIONS(2662),
    [anon_sym_inline] = ACTIONS(2662),
    [anon_sym___inline] = ACTIONS(2662),
    [anon_sym___inline__] = ACTIONS(2662),
    [anon_sym___forceinline] = ACTIONS(2662),
    [anon_sym_thread_local] = ACTIONS(2662),
    [anon_sym___thread] = ACTIONS(2662),
    [anon_sym_CG_EXTERN] = ACTIONS(2662),
    [anon_sym_CG_INLINE] = ACTIONS(2662),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2662),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2662),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2662),
    [anon_sym_IBOutlet] = ACTIONS(2662),
    [anon_sym_IBInspectable] = ACTIONS(2662),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2662),
    [anon_sym_NS_INLINE] = ACTIONS(2662),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2662),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2662),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2662),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2662),
    [anon_sym_const] = ACTIONS(2662),
    [anon_sym_constexpr] = ACTIONS(2662),
    [anon_sym_volatile] = ACTIONS(2662),
    [anon_sym_restrict] = ACTIONS(2662),
    [anon_sym___restrict__] = ACTIONS(2662),
    [anon_sym__Atomic] = ACTIONS(2662),
    [anon_sym__Noreturn] = ACTIONS(2662),
    [anon_sym__Nonnull] = ACTIONS(2662),
    [anon_sym_nullable] = ACTIONS(2662),
    [anon_sym__Complex] = ACTIONS(2662),
    [anon_sym__Nullable] = ACTIONS(2662),
    [anon_sym__Nullable_result] = ACTIONS(2662),
    [anon_sym__Null_unspecified] = ACTIONS(2662),
    [anon_sym___autoreleasing] = ACTIONS(2662),
    [anon_sym___block] = ACTIONS(2662),
    [anon_sym___bridge] = ACTIONS(2662),
    [anon_sym___bridge_retained] = ACTIONS(2662),
    [anon_sym___bridge_transfer] = ACTIONS(2662),
    [anon_sym___complex] = ACTIONS(2662),
    [anon_sym___const] = ACTIONS(2662),
    [anon_sym___imag] = ACTIONS(2662),
    [anon_sym___kindof] = ACTIONS(2662),
    [anon_sym___nonnull] = ACTIONS(2662),
    [anon_sym___nullable] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2662),
    [anon_sym___real] = ACTIONS(2662),
    [anon_sym___strong] = ACTIONS(2662),
    [anon_sym___unsafe_unretained] = ACTIONS(2662),
    [anon_sym___unused] = ACTIONS(2662),
    [anon_sym___weak] = ACTIONS(2662),
    [anon_sym_alignas] = ACTIONS(2662),
    [anon_sym__Alignas] = ACTIONS(2662),
    [sym_primitive_type] = ACTIONS(2662),
    [anon_sym_enum] = ACTIONS(2662),
    [anon_sym_struct] = ACTIONS(2662),
    [anon_sym_union] = ACTIONS(2662),
    [anon_sym_if] = ACTIONS(2662),
    [anon_sym_switch] = ACTIONS(2662),
    [anon_sym_case] = ACTIONS(2662),
    [anon_sym_default] = ACTIONS(2662),
    [anon_sym_while] = ACTIONS(2662),
    [anon_sym_do] = ACTIONS(2662),
    [anon_sym_for] = ACTIONS(2662),
    [anon_sym_in] = ACTIONS(2662),
    [anon_sym_return] = ACTIONS(2662),
    [anon_sym_break] = ACTIONS(2662),
    [anon_sym_continue] = ACTIONS(2662),
    [anon_sym_goto] = ACTIONS(2662),
    [anon_sym_DASH_DASH] = ACTIONS(2665),
    [anon_sym_PLUS_PLUS] = ACTIONS(2665),
    [anon_sym_sizeof] = ACTIONS(2662),
    [anon_sym___alignof__] = ACTIONS(2662),
    [anon_sym___alignof] = ACTIONS(2662),
    [anon_sym__alignof] = ACTIONS(2662),
    [anon_sym_alignof] = ACTIONS(2662),
    [anon_sym__Alignof] = ACTIONS(2662),
    [anon_sym_offsetof] = ACTIONS(2662),
    [anon_sym__Generic] = ACTIONS(2662),
    [anon_sym_asm] = ACTIONS(2662),
    [anon_sym___asm__] = ACTIONS(2662),
    [anon_sym___asm] = ACTIONS(2662),
    [sym_number_literal] = ACTIONS(2665),
    [anon_sym_L_SQUOTE] = ACTIONS(2665),
    [anon_sym_u_SQUOTE] = ACTIONS(2665),
    [anon_sym_U_SQUOTE] = ACTIONS(2665),
    [anon_sym_u8_SQUOTE] = ACTIONS(2665),
    [anon_sym_SQUOTE] = ACTIONS(2665),
    [anon_sym_AT] = ACTIONS(2662),
    [anon_sym_DQUOTE] = ACTIONS(2665),
    [anon_sym_L_DQUOTE] = ACTIONS(2665),
    [anon_sym_u_DQUOTE] = ACTIONS(2665),
    [anon_sym_U_DQUOTE] = ACTIONS(2665),
    [anon_sym_u8_DQUOTE] = ACTIONS(2665),
    [sym_true] = ACTIONS(2662),
    [sym_false] = ACTIONS(2662),
    [anon_sym_NULL] = ACTIONS(2662),
    [anon_sym_nullptr] = ACTIONS(2662),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2662),
    [anon_sym___typeof] = ACTIONS(2662),
    [anon_sym_typeof] = ACTIONS(2662),
    [anon_sym_ATimport] = ACTIONS(2665),
    [aux_sym_preproc_undef_token1] = ACTIONS(2662),
    [anon_sym_POUND] = ACTIONS(2662),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2662),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2662),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2662),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2662),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2662),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2662),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2662),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_API_AVAILABLE] = ACTIONS(2662),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_API_DEPRECATED] = ACTIONS(2662),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2662),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym___deprecated_msg] = ACTIONS(2662),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2662),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2662),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2662),
    [anon_sym_ATprotocol] = ACTIONS(2665),
    [anon_sym_ATinterface] = ACTIONS(2665),
    [anon_sym_ATimplementation] = ACTIONS(2665),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2665),
    [anon_sym_ATsynthesize] = ACTIONS(2665),
    [anon_sym_ATdynamic] = ACTIONS(2665),
    [anon_sym_ATtry] = ACTIONS(2665),
    [anon_sym___try] = ACTIONS(2662),
    [anon_sym_ATthrow] = ACTIONS(2665),
    [anon_sym_ATselector] = ACTIONS(2665),
    [anon_sym_ATavailable] = ACTIONS(2665),
    [anon_sym___builtin_available] = ACTIONS(2662),
    [anon_sym_va_arg] = ACTIONS(2662),
    [anon_sym_ATencode] = ACTIONS(2665),
    [anon_sym_ATsynchronized] = ACTIONS(2665),
    [anon_sym_BOOL] = ACTIONS(2662),
    [anon_sym_IMP] = ACTIONS(2662),
    [anon_sym_SEL] = ACTIONS(2662),
    [anon_sym_Class] = ACTIONS(2662),
    [anon_sym_id] = ACTIONS(2662),
  },
  [STATE(656)] = {
    [sym_else_clause] = STATE(741),
    [ts_builtin_sym_end] = ACTIONS(2396),
    [sym_identifier] = ACTIONS(2394),
    [aux_sym_preproc_include_token1] = ACTIONS(2394),
    [aux_sym_preproc_include_token2] = ACTIONS(2394),
    [aux_sym_preproc_def_token1] = ACTIONS(2394),
    [aux_sym_preproc_if_token1] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2394),
    [sym_preproc_directive] = ACTIONS(2394),
    [anon_sym_LPAREN2] = ACTIONS(2396),
    [anon_sym_BANG] = ACTIONS(2396),
    [anon_sym_TILDE] = ACTIONS(2396),
    [anon_sym_DASH] = ACTIONS(2394),
    [anon_sym_PLUS] = ACTIONS(2394),
    [anon_sym_STAR] = ACTIONS(2396),
    [anon_sym_CARET] = ACTIONS(2396),
    [anon_sym_AMP] = ACTIONS(2396),
    [anon_sym_SEMI] = ACTIONS(2396),
    [anon_sym___extension__] = ACTIONS(2394),
    [anon_sym_typedef] = ACTIONS(2394),
    [anon_sym_extern] = ACTIONS(2394),
    [anon_sym___attribute__] = ACTIONS(2394),
    [anon_sym___attribute] = ACTIONS(2394),
    [anon_sym_noreturn] = ACTIONS(2394),
    [anon_sym_LBRACK] = ACTIONS(2396),
    [anon_sym___declspec] = ACTIONS(2394),
    [anon_sym___cdecl] = ACTIONS(2394),
    [anon_sym___clrcall] = ACTIONS(2394),
    [anon_sym___stdcall] = ACTIONS(2394),
    [anon_sym___fastcall] = ACTIONS(2394),
    [anon_sym___thiscall] = ACTIONS(2394),
    [anon_sym___vectorcall] = ACTIONS(2394),
    [anon_sym_LBRACE] = ACTIONS(2396),
    [anon_sym_signed] = ACTIONS(2394),
    [anon_sym_unsigned] = ACTIONS(2394),
    [anon_sym_long] = ACTIONS(2394),
    [anon_sym_short] = ACTIONS(2394),
    [anon_sym_static] = ACTIONS(2394),
    [anon_sym_ATautoreleasepool] = ACTIONS(2396),
    [anon_sym_auto] = ACTIONS(2394),
    [anon_sym_register] = ACTIONS(2394),
    [anon_sym_inline] = ACTIONS(2394),
    [anon_sym___inline] = ACTIONS(2394),
    [anon_sym___inline__] = ACTIONS(2394),
    [anon_sym___forceinline] = ACTIONS(2394),
    [anon_sym_thread_local] = ACTIONS(2394),
    [anon_sym___thread] = ACTIONS(2394),
    [anon_sym_CG_EXTERN] = ACTIONS(2394),
    [anon_sym_CG_INLINE] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2394),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2394),
    [anon_sym_IBOutlet] = ACTIONS(2394),
    [anon_sym_IBInspectable] = ACTIONS(2394),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2394),
    [anon_sym_NS_INLINE] = ACTIONS(2394),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2394),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2394),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2394),
    [anon_sym_const] = ACTIONS(2394),
    [anon_sym_constexpr] = ACTIONS(2394),
    [anon_sym_volatile] = ACTIONS(2394),
    [anon_sym_restrict] = ACTIONS(2394),
    [anon_sym___restrict__] = ACTIONS(2394),
    [anon_sym__Atomic] = ACTIONS(2394),
    [anon_sym__Noreturn] = ACTIONS(2394),
    [anon_sym__Nonnull] = ACTIONS(2394),
    [anon_sym_nullable] = ACTIONS(2394),
    [anon_sym__Complex] = ACTIONS(2394),
    [anon_sym__Nullable] = ACTIONS(2394),
    [anon_sym__Nullable_result] = ACTIONS(2394),
    [anon_sym__Null_unspecified] = ACTIONS(2394),
    [anon_sym___autoreleasing] = ACTIONS(2394),
    [anon_sym___block] = ACTIONS(2394),
    [anon_sym___bridge] = ACTIONS(2394),
    [anon_sym___bridge_retained] = ACTIONS(2394),
    [anon_sym___bridge_transfer] = ACTIONS(2394),
    [anon_sym___complex] = ACTIONS(2394),
    [anon_sym___const] = ACTIONS(2394),
    [anon_sym___imag] = ACTIONS(2394),
    [anon_sym___kindof] = ACTIONS(2394),
    [anon_sym___nonnull] = ACTIONS(2394),
    [anon_sym___nullable] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2394),
    [anon_sym___real] = ACTIONS(2394),
    [anon_sym___strong] = ACTIONS(2394),
    [anon_sym___unsafe_unretained] = ACTIONS(2394),
    [anon_sym___unused] = ACTIONS(2394),
    [anon_sym___weak] = ACTIONS(2394),
    [anon_sym_alignas] = ACTIONS(2394),
    [anon_sym__Alignas] = ACTIONS(2394),
    [sym_primitive_type] = ACTIONS(2394),
    [anon_sym_enum] = ACTIONS(2394),
    [anon_sym_struct] = ACTIONS(2394),
    [anon_sym_union] = ACTIONS(2394),
    [anon_sym_if] = ACTIONS(2394),
    [anon_sym_else] = ACTIONS(2896),
    [anon_sym_switch] = ACTIONS(2394),
    [anon_sym_case] = ACTIONS(2394),
    [anon_sym_default] = ACTIONS(2394),
    [anon_sym_while] = ACTIONS(2394),
    [anon_sym_do] = ACTIONS(2394),
    [anon_sym_for] = ACTIONS(2394),
    [anon_sym_in] = ACTIONS(2394),
    [anon_sym_return] = ACTIONS(2394),
    [anon_sym_break] = ACTIONS(2394),
    [anon_sym_continue] = ACTIONS(2394),
    [anon_sym_goto] = ACTIONS(2394),
    [anon_sym_DASH_DASH] = ACTIONS(2396),
    [anon_sym_PLUS_PLUS] = ACTIONS(2396),
    [anon_sym_sizeof] = ACTIONS(2394),
    [anon_sym___alignof__] = ACTIONS(2394),
    [anon_sym___alignof] = ACTIONS(2394),
    [anon_sym__alignof] = ACTIONS(2394),
    [anon_sym_alignof] = ACTIONS(2394),
    [anon_sym__Alignof] = ACTIONS(2394),
    [anon_sym_offsetof] = ACTIONS(2394),
    [anon_sym__Generic] = ACTIONS(2394),
    [anon_sym_asm] = ACTIONS(2394),
    [anon_sym___asm__] = ACTIONS(2394),
    [anon_sym___asm] = ACTIONS(2394),
    [sym_number_literal] = ACTIONS(2396),
    [anon_sym_L_SQUOTE] = ACTIONS(2396),
    [anon_sym_u_SQUOTE] = ACTIONS(2396),
    [anon_sym_U_SQUOTE] = ACTIONS(2396),
    [anon_sym_u8_SQUOTE] = ACTIONS(2396),
    [anon_sym_SQUOTE] = ACTIONS(2396),
    [anon_sym_AT] = ACTIONS(2394),
    [anon_sym_DQUOTE] = ACTIONS(2396),
    [anon_sym_L_DQUOTE] = ACTIONS(2396),
    [anon_sym_u_DQUOTE] = ACTIONS(2396),
    [anon_sym_U_DQUOTE] = ACTIONS(2396),
    [anon_sym_u8_DQUOTE] = ACTIONS(2396),
    [sym_true] = ACTIONS(2394),
    [sym_false] = ACTIONS(2394),
    [anon_sym_NULL] = ACTIONS(2394),
    [anon_sym_nullptr] = ACTIONS(2394),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2394),
    [anon_sym___typeof] = ACTIONS(2394),
    [anon_sym_typeof] = ACTIONS(2394),
    [anon_sym_ATimport] = ACTIONS(2396),
    [aux_sym_preproc_undef_token1] = ACTIONS(2394),
    [anon_sym_POUND] = ACTIONS(2394),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2394),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2394),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2394),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_API_AVAILABLE] = ACTIONS(2394),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_API_DEPRECATED] = ACTIONS(2394),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym___deprecated_msg] = ACTIONS(2394),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2394),
    [anon_sym_ATprotocol] = ACTIONS(2396),
    [anon_sym_ATinterface] = ACTIONS(2396),
    [anon_sym_ATimplementation] = ACTIONS(2396),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2396),
    [anon_sym_ATtry] = ACTIONS(2396),
    [anon_sym___try] = ACTIONS(2394),
    [anon_sym_ATthrow] = ACTIONS(2396),
    [anon_sym_ATselector] = ACTIONS(2396),
    [anon_sym_ATavailable] = ACTIONS(2396),
    [anon_sym___builtin_available] = ACTIONS(2394),
    [anon_sym_va_arg] = ACTIONS(2394),
    [anon_sym_ATencode] = ACTIONS(2396),
    [anon_sym_ATsynchronized] = ACTIONS(2396),
    [anon_sym_BOOL] = ACTIONS(2394),
    [anon_sym_IMP] = ACTIONS(2394),
    [anon_sym_SEL] = ACTIONS(2394),
    [anon_sym_Class] = ACTIONS(2394),
    [anon_sym_id] = ACTIONS(2394),
  },
  [STATE(657)] = {
    [sym_identifier] = ACTIONS(2716),
    [aux_sym_preproc_include_token1] = ACTIONS(2716),
    [aux_sym_preproc_include_token2] = ACTIONS(2716),
    [aux_sym_preproc_def_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token2] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2716),
    [sym_preproc_directive] = ACTIONS(2716),
    [anon_sym_LPAREN2] = ACTIONS(2718),
    [anon_sym_BANG] = ACTIONS(2718),
    [anon_sym_TILDE] = ACTIONS(2718),
    [anon_sym_DASH] = ACTIONS(2716),
    [anon_sym_PLUS] = ACTIONS(2716),
    [anon_sym_STAR] = ACTIONS(2718),
    [anon_sym_CARET] = ACTIONS(2718),
    [anon_sym_AMP] = ACTIONS(2718),
    [anon_sym_SEMI] = ACTIONS(2718),
    [anon_sym___extension__] = ACTIONS(2716),
    [anon_sym_typedef] = ACTIONS(2716),
    [anon_sym_extern] = ACTIONS(2716),
    [anon_sym___attribute__] = ACTIONS(2716),
    [anon_sym___attribute] = ACTIONS(2716),
    [anon_sym_noreturn] = ACTIONS(2716),
    [anon_sym_LBRACK] = ACTIONS(2718),
    [anon_sym___declspec] = ACTIONS(2716),
    [anon_sym___cdecl] = ACTIONS(2716),
    [anon_sym___clrcall] = ACTIONS(2716),
    [anon_sym___stdcall] = ACTIONS(2716),
    [anon_sym___fastcall] = ACTIONS(2716),
    [anon_sym___thiscall] = ACTIONS(2716),
    [anon_sym___vectorcall] = ACTIONS(2716),
    [anon_sym_LBRACE] = ACTIONS(2718),
    [anon_sym_signed] = ACTIONS(2716),
    [anon_sym_unsigned] = ACTIONS(2716),
    [anon_sym_long] = ACTIONS(2716),
    [anon_sym_short] = ACTIONS(2716),
    [anon_sym_static] = ACTIONS(2716),
    [anon_sym_ATautoreleasepool] = ACTIONS(2718),
    [anon_sym_auto] = ACTIONS(2716),
    [anon_sym_register] = ACTIONS(2716),
    [anon_sym_inline] = ACTIONS(2716),
    [anon_sym___inline] = ACTIONS(2716),
    [anon_sym___inline__] = ACTIONS(2716),
    [anon_sym___forceinline] = ACTIONS(2716),
    [anon_sym_thread_local] = ACTIONS(2716),
    [anon_sym___thread] = ACTIONS(2716),
    [anon_sym_CG_EXTERN] = ACTIONS(2716),
    [anon_sym_CG_INLINE] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2716),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2716),
    [anon_sym_IBOutlet] = ACTIONS(2716),
    [anon_sym_IBInspectable] = ACTIONS(2716),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2716),
    [anon_sym_NS_INLINE] = ACTIONS(2716),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2716),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2716),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2716),
    [anon_sym_const] = ACTIONS(2716),
    [anon_sym_constexpr] = ACTIONS(2716),
    [anon_sym_volatile] = ACTIONS(2716),
    [anon_sym_restrict] = ACTIONS(2716),
    [anon_sym___restrict__] = ACTIONS(2716),
    [anon_sym__Atomic] = ACTIONS(2716),
    [anon_sym__Noreturn] = ACTIONS(2716),
    [anon_sym__Nonnull] = ACTIONS(2716),
    [anon_sym_nullable] = ACTIONS(2716),
    [anon_sym__Complex] = ACTIONS(2716),
    [anon_sym__Nullable] = ACTIONS(2716),
    [anon_sym__Nullable_result] = ACTIONS(2716),
    [anon_sym__Null_unspecified] = ACTIONS(2716),
    [anon_sym___autoreleasing] = ACTIONS(2716),
    [anon_sym___block] = ACTIONS(2716),
    [anon_sym___bridge] = ACTIONS(2716),
    [anon_sym___bridge_retained] = ACTIONS(2716),
    [anon_sym___bridge_transfer] = ACTIONS(2716),
    [anon_sym___complex] = ACTIONS(2716),
    [anon_sym___const] = ACTIONS(2716),
    [anon_sym___imag] = ACTIONS(2716),
    [anon_sym___kindof] = ACTIONS(2716),
    [anon_sym___nonnull] = ACTIONS(2716),
    [anon_sym___nullable] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2716),
    [anon_sym___real] = ACTIONS(2716),
    [anon_sym___strong] = ACTIONS(2716),
    [anon_sym___unsafe_unretained] = ACTIONS(2716),
    [anon_sym___unused] = ACTIONS(2716),
    [anon_sym___weak] = ACTIONS(2716),
    [anon_sym_alignas] = ACTIONS(2716),
    [anon_sym__Alignas] = ACTIONS(2716),
    [sym_primitive_type] = ACTIONS(2716),
    [anon_sym_enum] = ACTIONS(2716),
    [anon_sym_struct] = ACTIONS(2716),
    [anon_sym_union] = ACTIONS(2716),
    [anon_sym_if] = ACTIONS(2716),
    [anon_sym_switch] = ACTIONS(2716),
    [anon_sym_case] = ACTIONS(2716),
    [anon_sym_default] = ACTIONS(2716),
    [anon_sym_while] = ACTIONS(2716),
    [anon_sym_do] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2716),
    [anon_sym_in] = ACTIONS(2716),
    [anon_sym_return] = ACTIONS(2716),
    [anon_sym_break] = ACTIONS(2716),
    [anon_sym_continue] = ACTIONS(2716),
    [anon_sym_goto] = ACTIONS(2716),
    [anon_sym_DASH_DASH] = ACTIONS(2718),
    [anon_sym_PLUS_PLUS] = ACTIONS(2718),
    [anon_sym_sizeof] = ACTIONS(2716),
    [anon_sym___alignof__] = ACTIONS(2716),
    [anon_sym___alignof] = ACTIONS(2716),
    [anon_sym__alignof] = ACTIONS(2716),
    [anon_sym_alignof] = ACTIONS(2716),
    [anon_sym__Alignof] = ACTIONS(2716),
    [anon_sym_offsetof] = ACTIONS(2716),
    [anon_sym__Generic] = ACTIONS(2716),
    [anon_sym_asm] = ACTIONS(2716),
    [anon_sym___asm__] = ACTIONS(2716),
    [anon_sym___asm] = ACTIONS(2716),
    [sym_number_literal] = ACTIONS(2718),
    [anon_sym_L_SQUOTE] = ACTIONS(2718),
    [anon_sym_u_SQUOTE] = ACTIONS(2718),
    [anon_sym_U_SQUOTE] = ACTIONS(2718),
    [anon_sym_u8_SQUOTE] = ACTIONS(2718),
    [anon_sym_SQUOTE] = ACTIONS(2718),
    [anon_sym_AT] = ACTIONS(2716),
    [anon_sym_DQUOTE] = ACTIONS(2718),
    [anon_sym_L_DQUOTE] = ACTIONS(2718),
    [anon_sym_u_DQUOTE] = ACTIONS(2718),
    [anon_sym_U_DQUOTE] = ACTIONS(2718),
    [anon_sym_u8_DQUOTE] = ACTIONS(2718),
    [sym_true] = ACTIONS(2716),
    [sym_false] = ACTIONS(2716),
    [anon_sym_NULL] = ACTIONS(2716),
    [anon_sym_nullptr] = ACTIONS(2716),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2716),
    [anon_sym___typeof] = ACTIONS(2716),
    [anon_sym_typeof] = ACTIONS(2716),
    [anon_sym_ATimport] = ACTIONS(2718),
    [aux_sym_preproc_undef_token1] = ACTIONS(2716),
    [anon_sym_POUND] = ACTIONS(2716),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2716),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2716),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2716),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_API_AVAILABLE] = ACTIONS(2716),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_API_DEPRECATED] = ACTIONS(2716),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym___deprecated_msg] = ACTIONS(2716),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2716),
    [anon_sym_ATprotocol] = ACTIONS(2718),
    [anon_sym_ATinterface] = ACTIONS(2718),
    [anon_sym_ATimplementation] = ACTIONS(2718),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2718),
    [anon_sym_ATsynthesize] = ACTIONS(2718),
    [anon_sym_ATdynamic] = ACTIONS(2718),
    [anon_sym_ATtry] = ACTIONS(2718),
    [anon_sym___try] = ACTIONS(2716),
    [anon_sym_ATthrow] = ACTIONS(2718),
    [anon_sym_ATselector] = ACTIONS(2718),
    [anon_sym_ATavailable] = ACTIONS(2718),
    [anon_sym___builtin_available] = ACTIONS(2716),
    [anon_sym_va_arg] = ACTIONS(2716),
    [anon_sym_ATencode] = ACTIONS(2718),
    [anon_sym_ATsynchronized] = ACTIONS(2718),
    [anon_sym_BOOL] = ACTIONS(2716),
    [anon_sym_IMP] = ACTIONS(2716),
    [anon_sym_SEL] = ACTIONS(2716),
    [anon_sym_Class] = ACTIONS(2716),
    [anon_sym_id] = ACTIONS(2716),
  },
  [STATE(658)] = {
    [sym_identifier] = ACTIONS(2634),
    [aux_sym_preproc_include_token1] = ACTIONS(2634),
    [aux_sym_preproc_include_token2] = ACTIONS(2634),
    [aux_sym_preproc_def_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token2] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2634),
    [sym_preproc_directive] = ACTIONS(2634),
    [anon_sym_LPAREN2] = ACTIONS(2636),
    [anon_sym_BANG] = ACTIONS(2636),
    [anon_sym_TILDE] = ACTIONS(2636),
    [anon_sym_DASH] = ACTIONS(2634),
    [anon_sym_PLUS] = ACTIONS(2634),
    [anon_sym_STAR] = ACTIONS(2636),
    [anon_sym_CARET] = ACTIONS(2636),
    [anon_sym_AMP] = ACTIONS(2636),
    [anon_sym_SEMI] = ACTIONS(2636),
    [anon_sym___extension__] = ACTIONS(2634),
    [anon_sym_typedef] = ACTIONS(2634),
    [anon_sym_extern] = ACTIONS(2634),
    [anon_sym___attribute__] = ACTIONS(2634),
    [anon_sym___attribute] = ACTIONS(2634),
    [anon_sym_noreturn] = ACTIONS(2634),
    [anon_sym_LBRACK] = ACTIONS(2636),
    [anon_sym___declspec] = ACTIONS(2634),
    [anon_sym___cdecl] = ACTIONS(2634),
    [anon_sym___clrcall] = ACTIONS(2634),
    [anon_sym___stdcall] = ACTIONS(2634),
    [anon_sym___fastcall] = ACTIONS(2634),
    [anon_sym___thiscall] = ACTIONS(2634),
    [anon_sym___vectorcall] = ACTIONS(2634),
    [anon_sym_LBRACE] = ACTIONS(2636),
    [anon_sym_signed] = ACTIONS(2634),
    [anon_sym_unsigned] = ACTIONS(2634),
    [anon_sym_long] = ACTIONS(2634),
    [anon_sym_short] = ACTIONS(2634),
    [anon_sym_static] = ACTIONS(2634),
    [anon_sym_ATautoreleasepool] = ACTIONS(2636),
    [anon_sym_auto] = ACTIONS(2634),
    [anon_sym_register] = ACTIONS(2634),
    [anon_sym_inline] = ACTIONS(2634),
    [anon_sym___inline] = ACTIONS(2634),
    [anon_sym___inline__] = ACTIONS(2634),
    [anon_sym___forceinline] = ACTIONS(2634),
    [anon_sym_thread_local] = ACTIONS(2634),
    [anon_sym___thread] = ACTIONS(2634),
    [anon_sym_CG_EXTERN] = ACTIONS(2634),
    [anon_sym_CG_INLINE] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2634),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2634),
    [anon_sym_IBOutlet] = ACTIONS(2634),
    [anon_sym_IBInspectable] = ACTIONS(2634),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2634),
    [anon_sym_NS_INLINE] = ACTIONS(2634),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2634),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2634),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2634),
    [anon_sym_const] = ACTIONS(2634),
    [anon_sym_constexpr] = ACTIONS(2634),
    [anon_sym_volatile] = ACTIONS(2634),
    [anon_sym_restrict] = ACTIONS(2634),
    [anon_sym___restrict__] = ACTIONS(2634),
    [anon_sym__Atomic] = ACTIONS(2634),
    [anon_sym__Noreturn] = ACTIONS(2634),
    [anon_sym__Nonnull] = ACTIONS(2634),
    [anon_sym_nullable] = ACTIONS(2634),
    [anon_sym__Complex] = ACTIONS(2634),
    [anon_sym__Nullable] = ACTIONS(2634),
    [anon_sym__Nullable_result] = ACTIONS(2634),
    [anon_sym__Null_unspecified] = ACTIONS(2634),
    [anon_sym___autoreleasing] = ACTIONS(2634),
    [anon_sym___block] = ACTIONS(2634),
    [anon_sym___bridge] = ACTIONS(2634),
    [anon_sym___bridge_retained] = ACTIONS(2634),
    [anon_sym___bridge_transfer] = ACTIONS(2634),
    [anon_sym___complex] = ACTIONS(2634),
    [anon_sym___const] = ACTIONS(2634),
    [anon_sym___imag] = ACTIONS(2634),
    [anon_sym___kindof] = ACTIONS(2634),
    [anon_sym___nonnull] = ACTIONS(2634),
    [anon_sym___nullable] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2634),
    [anon_sym___real] = ACTIONS(2634),
    [anon_sym___strong] = ACTIONS(2634),
    [anon_sym___unsafe_unretained] = ACTIONS(2634),
    [anon_sym___unused] = ACTIONS(2634),
    [anon_sym___weak] = ACTIONS(2634),
    [anon_sym_alignas] = ACTIONS(2634),
    [anon_sym__Alignas] = ACTIONS(2634),
    [sym_primitive_type] = ACTIONS(2634),
    [anon_sym_enum] = ACTIONS(2634),
    [anon_sym_struct] = ACTIONS(2634),
    [anon_sym_union] = ACTIONS(2634),
    [anon_sym_if] = ACTIONS(2634),
    [anon_sym_switch] = ACTIONS(2634),
    [anon_sym_case] = ACTIONS(2634),
    [anon_sym_default] = ACTIONS(2634),
    [anon_sym_while] = ACTIONS(2634),
    [anon_sym_do] = ACTIONS(2634),
    [anon_sym_for] = ACTIONS(2634),
    [anon_sym_in] = ACTIONS(2634),
    [anon_sym_return] = ACTIONS(2634),
    [anon_sym_break] = ACTIONS(2634),
    [anon_sym_continue] = ACTIONS(2634),
    [anon_sym_goto] = ACTIONS(2634),
    [anon_sym_DASH_DASH] = ACTIONS(2636),
    [anon_sym_PLUS_PLUS] = ACTIONS(2636),
    [anon_sym_sizeof] = ACTIONS(2634),
    [anon_sym___alignof__] = ACTIONS(2634),
    [anon_sym___alignof] = ACTIONS(2634),
    [anon_sym__alignof] = ACTIONS(2634),
    [anon_sym_alignof] = ACTIONS(2634),
    [anon_sym__Alignof] = ACTIONS(2634),
    [anon_sym_offsetof] = ACTIONS(2634),
    [anon_sym__Generic] = ACTIONS(2634),
    [anon_sym_asm] = ACTIONS(2634),
    [anon_sym___asm__] = ACTIONS(2634),
    [anon_sym___asm] = ACTIONS(2634),
    [sym_number_literal] = ACTIONS(2636),
    [anon_sym_L_SQUOTE] = ACTIONS(2636),
    [anon_sym_u_SQUOTE] = ACTIONS(2636),
    [anon_sym_U_SQUOTE] = ACTIONS(2636),
    [anon_sym_u8_SQUOTE] = ACTIONS(2636),
    [anon_sym_SQUOTE] = ACTIONS(2636),
    [anon_sym_AT] = ACTIONS(2634),
    [anon_sym_DQUOTE] = ACTIONS(2636),
    [anon_sym_L_DQUOTE] = ACTIONS(2636),
    [anon_sym_u_DQUOTE] = ACTIONS(2636),
    [anon_sym_U_DQUOTE] = ACTIONS(2636),
    [anon_sym_u8_DQUOTE] = ACTIONS(2636),
    [sym_true] = ACTIONS(2634),
    [sym_false] = ACTIONS(2634),
    [anon_sym_NULL] = ACTIONS(2634),
    [anon_sym_nullptr] = ACTIONS(2634),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2634),
    [anon_sym___typeof] = ACTIONS(2634),
    [anon_sym_typeof] = ACTIONS(2634),
    [anon_sym_ATimport] = ACTIONS(2636),
    [aux_sym_preproc_undef_token1] = ACTIONS(2634),
    [anon_sym_POUND] = ACTIONS(2634),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2634),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2634),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2634),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_API_AVAILABLE] = ACTIONS(2634),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_API_DEPRECATED] = ACTIONS(2634),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym___deprecated_msg] = ACTIONS(2634),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2634),
    [anon_sym_ATprotocol] = ACTIONS(2636),
    [anon_sym_ATinterface] = ACTIONS(2636),
    [anon_sym_ATimplementation] = ACTIONS(2636),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2636),
    [anon_sym_ATsynthesize] = ACTIONS(2636),
    [anon_sym_ATdynamic] = ACTIONS(2636),
    [anon_sym_ATtry] = ACTIONS(2636),
    [anon_sym___try] = ACTIONS(2634),
    [anon_sym_ATthrow] = ACTIONS(2636),
    [anon_sym_ATselector] = ACTIONS(2636),
    [anon_sym_ATavailable] = ACTIONS(2636),
    [anon_sym___builtin_available] = ACTIONS(2634),
    [anon_sym_va_arg] = ACTIONS(2634),
    [anon_sym_ATencode] = ACTIONS(2636),
    [anon_sym_ATsynchronized] = ACTIONS(2636),
    [anon_sym_BOOL] = ACTIONS(2634),
    [anon_sym_IMP] = ACTIONS(2634),
    [anon_sym_SEL] = ACTIONS(2634),
    [anon_sym_Class] = ACTIONS(2634),
    [anon_sym_id] = ACTIONS(2634),
  },
  [STATE(659)] = {
    [sym_identifier] = ACTIONS(2668),
    [aux_sym_preproc_include_token1] = ACTIONS(2668),
    [aux_sym_preproc_include_token2] = ACTIONS(2668),
    [aux_sym_preproc_def_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token2] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2668),
    [sym_preproc_directive] = ACTIONS(2668),
    [anon_sym_LPAREN2] = ACTIONS(2670),
    [anon_sym_BANG] = ACTIONS(2670),
    [anon_sym_TILDE] = ACTIONS(2670),
    [anon_sym_DASH] = ACTIONS(2668),
    [anon_sym_PLUS] = ACTIONS(2668),
    [anon_sym_STAR] = ACTIONS(2670),
    [anon_sym_CARET] = ACTIONS(2670),
    [anon_sym_AMP] = ACTIONS(2670),
    [anon_sym_SEMI] = ACTIONS(2670),
    [anon_sym___extension__] = ACTIONS(2668),
    [anon_sym_typedef] = ACTIONS(2668),
    [anon_sym_extern] = ACTIONS(2668),
    [anon_sym___attribute__] = ACTIONS(2668),
    [anon_sym___attribute] = ACTIONS(2668),
    [anon_sym_noreturn] = ACTIONS(2668),
    [anon_sym_LBRACK] = ACTIONS(2670),
    [anon_sym___declspec] = ACTIONS(2668),
    [anon_sym___cdecl] = ACTIONS(2668),
    [anon_sym___clrcall] = ACTIONS(2668),
    [anon_sym___stdcall] = ACTIONS(2668),
    [anon_sym___fastcall] = ACTIONS(2668),
    [anon_sym___thiscall] = ACTIONS(2668),
    [anon_sym___vectorcall] = ACTIONS(2668),
    [anon_sym_LBRACE] = ACTIONS(2670),
    [anon_sym_signed] = ACTIONS(2668),
    [anon_sym_unsigned] = ACTIONS(2668),
    [anon_sym_long] = ACTIONS(2668),
    [anon_sym_short] = ACTIONS(2668),
    [anon_sym_static] = ACTIONS(2668),
    [anon_sym_ATautoreleasepool] = ACTIONS(2670),
    [anon_sym_auto] = ACTIONS(2668),
    [anon_sym_register] = ACTIONS(2668),
    [anon_sym_inline] = ACTIONS(2668),
    [anon_sym___inline] = ACTIONS(2668),
    [anon_sym___inline__] = ACTIONS(2668),
    [anon_sym___forceinline] = ACTIONS(2668),
    [anon_sym_thread_local] = ACTIONS(2668),
    [anon_sym___thread] = ACTIONS(2668),
    [anon_sym_CG_EXTERN] = ACTIONS(2668),
    [anon_sym_CG_INLINE] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2668),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2668),
    [anon_sym_IBOutlet] = ACTIONS(2668),
    [anon_sym_IBInspectable] = ACTIONS(2668),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2668),
    [anon_sym_NS_INLINE] = ACTIONS(2668),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2668),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2668),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2668),
    [anon_sym_const] = ACTIONS(2668),
    [anon_sym_constexpr] = ACTIONS(2668),
    [anon_sym_volatile] = ACTIONS(2668),
    [anon_sym_restrict] = ACTIONS(2668),
    [anon_sym___restrict__] = ACTIONS(2668),
    [anon_sym__Atomic] = ACTIONS(2668),
    [anon_sym__Noreturn] = ACTIONS(2668),
    [anon_sym__Nonnull] = ACTIONS(2668),
    [anon_sym_nullable] = ACTIONS(2668),
    [anon_sym__Complex] = ACTIONS(2668),
    [anon_sym__Nullable] = ACTIONS(2668),
    [anon_sym__Nullable_result] = ACTIONS(2668),
    [anon_sym__Null_unspecified] = ACTIONS(2668),
    [anon_sym___autoreleasing] = ACTIONS(2668),
    [anon_sym___block] = ACTIONS(2668),
    [anon_sym___bridge] = ACTIONS(2668),
    [anon_sym___bridge_retained] = ACTIONS(2668),
    [anon_sym___bridge_transfer] = ACTIONS(2668),
    [anon_sym___complex] = ACTIONS(2668),
    [anon_sym___const] = ACTIONS(2668),
    [anon_sym___imag] = ACTIONS(2668),
    [anon_sym___kindof] = ACTIONS(2668),
    [anon_sym___nonnull] = ACTIONS(2668),
    [anon_sym___nullable] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2668),
    [anon_sym___real] = ACTIONS(2668),
    [anon_sym___strong] = ACTIONS(2668),
    [anon_sym___unsafe_unretained] = ACTIONS(2668),
    [anon_sym___unused] = ACTIONS(2668),
    [anon_sym___weak] = ACTIONS(2668),
    [anon_sym_alignas] = ACTIONS(2668),
    [anon_sym__Alignas] = ACTIONS(2668),
    [sym_primitive_type] = ACTIONS(2668),
    [anon_sym_enum] = ACTIONS(2668),
    [anon_sym_struct] = ACTIONS(2668),
    [anon_sym_union] = ACTIONS(2668),
    [anon_sym_if] = ACTIONS(2668),
    [anon_sym_switch] = ACTIONS(2668),
    [anon_sym_case] = ACTIONS(2668),
    [anon_sym_default] = ACTIONS(2668),
    [anon_sym_while] = ACTIONS(2668),
    [anon_sym_do] = ACTIONS(2668),
    [anon_sym_for] = ACTIONS(2668),
    [anon_sym_in] = ACTIONS(2668),
    [anon_sym_return] = ACTIONS(2668),
    [anon_sym_break] = ACTIONS(2668),
    [anon_sym_continue] = ACTIONS(2668),
    [anon_sym_goto] = ACTIONS(2668),
    [anon_sym_DASH_DASH] = ACTIONS(2670),
    [anon_sym_PLUS_PLUS] = ACTIONS(2670),
    [anon_sym_sizeof] = ACTIONS(2668),
    [anon_sym___alignof__] = ACTIONS(2668),
    [anon_sym___alignof] = ACTIONS(2668),
    [anon_sym__alignof] = ACTIONS(2668),
    [anon_sym_alignof] = ACTIONS(2668),
    [anon_sym__Alignof] = ACTIONS(2668),
    [anon_sym_offsetof] = ACTIONS(2668),
    [anon_sym__Generic] = ACTIONS(2668),
    [anon_sym_asm] = ACTIONS(2668),
    [anon_sym___asm__] = ACTIONS(2668),
    [anon_sym___asm] = ACTIONS(2668),
    [sym_number_literal] = ACTIONS(2670),
    [anon_sym_L_SQUOTE] = ACTIONS(2670),
    [anon_sym_u_SQUOTE] = ACTIONS(2670),
    [anon_sym_U_SQUOTE] = ACTIONS(2670),
    [anon_sym_u8_SQUOTE] = ACTIONS(2670),
    [anon_sym_SQUOTE] = ACTIONS(2670),
    [anon_sym_AT] = ACTIONS(2668),
    [anon_sym_DQUOTE] = ACTIONS(2670),
    [anon_sym_L_DQUOTE] = ACTIONS(2670),
    [anon_sym_u_DQUOTE] = ACTIONS(2670),
    [anon_sym_U_DQUOTE] = ACTIONS(2670),
    [anon_sym_u8_DQUOTE] = ACTIONS(2670),
    [sym_true] = ACTIONS(2668),
    [sym_false] = ACTIONS(2668),
    [anon_sym_NULL] = ACTIONS(2668),
    [anon_sym_nullptr] = ACTIONS(2668),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2668),
    [anon_sym___typeof] = ACTIONS(2668),
    [anon_sym_typeof] = ACTIONS(2668),
    [anon_sym_ATimport] = ACTIONS(2670),
    [aux_sym_preproc_undef_token1] = ACTIONS(2668),
    [anon_sym_POUND] = ACTIONS(2668),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2668),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2668),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2668),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_API_AVAILABLE] = ACTIONS(2668),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_API_DEPRECATED] = ACTIONS(2668),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym___deprecated_msg] = ACTIONS(2668),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2668),
    [anon_sym_ATprotocol] = ACTIONS(2670),
    [anon_sym_ATinterface] = ACTIONS(2670),
    [anon_sym_ATimplementation] = ACTIONS(2670),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2670),
    [anon_sym_ATsynthesize] = ACTIONS(2670),
    [anon_sym_ATdynamic] = ACTIONS(2670),
    [anon_sym_ATtry] = ACTIONS(2670),
    [anon_sym___try] = ACTIONS(2668),
    [anon_sym_ATthrow] = ACTIONS(2670),
    [anon_sym_ATselector] = ACTIONS(2670),
    [anon_sym_ATavailable] = ACTIONS(2670),
    [anon_sym___builtin_available] = ACTIONS(2668),
    [anon_sym_va_arg] = ACTIONS(2668),
    [anon_sym_ATencode] = ACTIONS(2670),
    [anon_sym_ATsynchronized] = ACTIONS(2670),
    [anon_sym_BOOL] = ACTIONS(2668),
    [anon_sym_IMP] = ACTIONS(2668),
    [anon_sym_SEL] = ACTIONS(2668),
    [anon_sym_Class] = ACTIONS(2668),
    [anon_sym_id] = ACTIONS(2668),
  },
  [STATE(660)] = {
    [sym_identifier] = ACTIONS(2604),
    [aux_sym_preproc_include_token1] = ACTIONS(2604),
    [aux_sym_preproc_include_token2] = ACTIONS(2604),
    [aux_sym_preproc_def_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token2] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2604),
    [sym_preproc_directive] = ACTIONS(2604),
    [anon_sym_LPAREN2] = ACTIONS(2606),
    [anon_sym_BANG] = ACTIONS(2606),
    [anon_sym_TILDE] = ACTIONS(2606),
    [anon_sym_DASH] = ACTIONS(2604),
    [anon_sym_PLUS] = ACTIONS(2604),
    [anon_sym_STAR] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(2606),
    [anon_sym_AMP] = ACTIONS(2606),
    [anon_sym_SEMI] = ACTIONS(2606),
    [anon_sym___extension__] = ACTIONS(2604),
    [anon_sym_typedef] = ACTIONS(2604),
    [anon_sym_extern] = ACTIONS(2604),
    [anon_sym___attribute__] = ACTIONS(2604),
    [anon_sym___attribute] = ACTIONS(2604),
    [anon_sym_noreturn] = ACTIONS(2604),
    [anon_sym_LBRACK] = ACTIONS(2606),
    [anon_sym___declspec] = ACTIONS(2604),
    [anon_sym___cdecl] = ACTIONS(2604),
    [anon_sym___clrcall] = ACTIONS(2604),
    [anon_sym___stdcall] = ACTIONS(2604),
    [anon_sym___fastcall] = ACTIONS(2604),
    [anon_sym___thiscall] = ACTIONS(2604),
    [anon_sym___vectorcall] = ACTIONS(2604),
    [anon_sym_LBRACE] = ACTIONS(2606),
    [anon_sym_signed] = ACTIONS(2604),
    [anon_sym_unsigned] = ACTIONS(2604),
    [anon_sym_long] = ACTIONS(2604),
    [anon_sym_short] = ACTIONS(2604),
    [anon_sym_static] = ACTIONS(2604),
    [anon_sym_ATautoreleasepool] = ACTIONS(2606),
    [anon_sym_auto] = ACTIONS(2604),
    [anon_sym_register] = ACTIONS(2604),
    [anon_sym_inline] = ACTIONS(2604),
    [anon_sym___inline] = ACTIONS(2604),
    [anon_sym___inline__] = ACTIONS(2604),
    [anon_sym___forceinline] = ACTIONS(2604),
    [anon_sym_thread_local] = ACTIONS(2604),
    [anon_sym___thread] = ACTIONS(2604),
    [anon_sym_CG_EXTERN] = ACTIONS(2604),
    [anon_sym_CG_INLINE] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2604),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2604),
    [anon_sym_IBOutlet] = ACTIONS(2604),
    [anon_sym_IBInspectable] = ACTIONS(2604),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2604),
    [anon_sym_NS_INLINE] = ACTIONS(2604),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2604),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2604),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2604),
    [anon_sym_const] = ACTIONS(2604),
    [anon_sym_constexpr] = ACTIONS(2604),
    [anon_sym_volatile] = ACTIONS(2604),
    [anon_sym_restrict] = ACTIONS(2604),
    [anon_sym___restrict__] = ACTIONS(2604),
    [anon_sym__Atomic] = ACTIONS(2604),
    [anon_sym__Noreturn] = ACTIONS(2604),
    [anon_sym__Nonnull] = ACTIONS(2604),
    [anon_sym_nullable] = ACTIONS(2604),
    [anon_sym__Complex] = ACTIONS(2604),
    [anon_sym__Nullable] = ACTIONS(2604),
    [anon_sym__Nullable_result] = ACTIONS(2604),
    [anon_sym__Null_unspecified] = ACTIONS(2604),
    [anon_sym___autoreleasing] = ACTIONS(2604),
    [anon_sym___block] = ACTIONS(2604),
    [anon_sym___bridge] = ACTIONS(2604),
    [anon_sym___bridge_retained] = ACTIONS(2604),
    [anon_sym___bridge_transfer] = ACTIONS(2604),
    [anon_sym___complex] = ACTIONS(2604),
    [anon_sym___const] = ACTIONS(2604),
    [anon_sym___imag] = ACTIONS(2604),
    [anon_sym___kindof] = ACTIONS(2604),
    [anon_sym___nonnull] = ACTIONS(2604),
    [anon_sym___nullable] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2604),
    [anon_sym___real] = ACTIONS(2604),
    [anon_sym___strong] = ACTIONS(2604),
    [anon_sym___unsafe_unretained] = ACTIONS(2604),
    [anon_sym___unused] = ACTIONS(2604),
    [anon_sym___weak] = ACTIONS(2604),
    [anon_sym_alignas] = ACTIONS(2604),
    [anon_sym__Alignas] = ACTIONS(2604),
    [sym_primitive_type] = ACTIONS(2604),
    [anon_sym_enum] = ACTIONS(2604),
    [anon_sym_struct] = ACTIONS(2604),
    [anon_sym_union] = ACTIONS(2604),
    [anon_sym_if] = ACTIONS(2604),
    [anon_sym_switch] = ACTIONS(2604),
    [anon_sym_case] = ACTIONS(2604),
    [anon_sym_default] = ACTIONS(2604),
    [anon_sym_while] = ACTIONS(2604),
    [anon_sym_do] = ACTIONS(2604),
    [anon_sym_for] = ACTIONS(2604),
    [anon_sym_in] = ACTIONS(2604),
    [anon_sym_return] = ACTIONS(2604),
    [anon_sym_break] = ACTIONS(2604),
    [anon_sym_continue] = ACTIONS(2604),
    [anon_sym_goto] = ACTIONS(2604),
    [anon_sym_DASH_DASH] = ACTIONS(2606),
    [anon_sym_PLUS_PLUS] = ACTIONS(2606),
    [anon_sym_sizeof] = ACTIONS(2604),
    [anon_sym___alignof__] = ACTIONS(2604),
    [anon_sym___alignof] = ACTIONS(2604),
    [anon_sym__alignof] = ACTIONS(2604),
    [anon_sym_alignof] = ACTIONS(2604),
    [anon_sym__Alignof] = ACTIONS(2604),
    [anon_sym_offsetof] = ACTIONS(2604),
    [anon_sym__Generic] = ACTIONS(2604),
    [anon_sym_asm] = ACTIONS(2604),
    [anon_sym___asm__] = ACTIONS(2604),
    [anon_sym___asm] = ACTIONS(2604),
    [sym_number_literal] = ACTIONS(2606),
    [anon_sym_L_SQUOTE] = ACTIONS(2606),
    [anon_sym_u_SQUOTE] = ACTIONS(2606),
    [anon_sym_U_SQUOTE] = ACTIONS(2606),
    [anon_sym_u8_SQUOTE] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2606),
    [anon_sym_AT] = ACTIONS(2604),
    [anon_sym_DQUOTE] = ACTIONS(2606),
    [anon_sym_L_DQUOTE] = ACTIONS(2606),
    [anon_sym_u_DQUOTE] = ACTIONS(2606),
    [anon_sym_U_DQUOTE] = ACTIONS(2606),
    [anon_sym_u8_DQUOTE] = ACTIONS(2606),
    [sym_true] = ACTIONS(2604),
    [sym_false] = ACTIONS(2604),
    [anon_sym_NULL] = ACTIONS(2604),
    [anon_sym_nullptr] = ACTIONS(2604),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2604),
    [anon_sym___typeof] = ACTIONS(2604),
    [anon_sym_typeof] = ACTIONS(2604),
    [anon_sym_ATimport] = ACTIONS(2606),
    [aux_sym_preproc_undef_token1] = ACTIONS(2604),
    [anon_sym_POUND] = ACTIONS(2604),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2604),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2604),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2604),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_API_AVAILABLE] = ACTIONS(2604),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_API_DEPRECATED] = ACTIONS(2604),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym___deprecated_msg] = ACTIONS(2604),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2604),
    [anon_sym_ATprotocol] = ACTIONS(2606),
    [anon_sym_ATinterface] = ACTIONS(2606),
    [anon_sym_ATimplementation] = ACTIONS(2606),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2606),
    [anon_sym_ATsynthesize] = ACTIONS(2606),
    [anon_sym_ATdynamic] = ACTIONS(2606),
    [anon_sym_ATtry] = ACTIONS(2606),
    [anon_sym___try] = ACTIONS(2604),
    [anon_sym_ATthrow] = ACTIONS(2606),
    [anon_sym_ATselector] = ACTIONS(2606),
    [anon_sym_ATavailable] = ACTIONS(2606),
    [anon_sym___builtin_available] = ACTIONS(2604),
    [anon_sym_va_arg] = ACTIONS(2604),
    [anon_sym_ATencode] = ACTIONS(2606),
    [anon_sym_ATsynchronized] = ACTIONS(2606),
    [anon_sym_BOOL] = ACTIONS(2604),
    [anon_sym_IMP] = ACTIONS(2604),
    [anon_sym_SEL] = ACTIONS(2604),
    [anon_sym_Class] = ACTIONS(2604),
    [anon_sym_id] = ACTIONS(2604),
  },
  [STATE(661)] = {
    [sym_identifier] = ACTIONS(2614),
    [aux_sym_preproc_include_token1] = ACTIONS(2614),
    [aux_sym_preproc_include_token2] = ACTIONS(2614),
    [aux_sym_preproc_def_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token2] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2614),
    [sym_preproc_directive] = ACTIONS(2614),
    [anon_sym_LPAREN2] = ACTIONS(2616),
    [anon_sym_BANG] = ACTIONS(2616),
    [anon_sym_TILDE] = ACTIONS(2616),
    [anon_sym_DASH] = ACTIONS(2614),
    [anon_sym_PLUS] = ACTIONS(2614),
    [anon_sym_STAR] = ACTIONS(2616),
    [anon_sym_CARET] = ACTIONS(2616),
    [anon_sym_AMP] = ACTIONS(2616),
    [anon_sym_SEMI] = ACTIONS(2616),
    [anon_sym___extension__] = ACTIONS(2614),
    [anon_sym_typedef] = ACTIONS(2614),
    [anon_sym_extern] = ACTIONS(2614),
    [anon_sym___attribute__] = ACTIONS(2614),
    [anon_sym___attribute] = ACTIONS(2614),
    [anon_sym_noreturn] = ACTIONS(2614),
    [anon_sym_LBRACK] = ACTIONS(2616),
    [anon_sym___declspec] = ACTIONS(2614),
    [anon_sym___cdecl] = ACTIONS(2614),
    [anon_sym___clrcall] = ACTIONS(2614),
    [anon_sym___stdcall] = ACTIONS(2614),
    [anon_sym___fastcall] = ACTIONS(2614),
    [anon_sym___thiscall] = ACTIONS(2614),
    [anon_sym___vectorcall] = ACTIONS(2614),
    [anon_sym_LBRACE] = ACTIONS(2616),
    [anon_sym_signed] = ACTIONS(2614),
    [anon_sym_unsigned] = ACTIONS(2614),
    [anon_sym_long] = ACTIONS(2614),
    [anon_sym_short] = ACTIONS(2614),
    [anon_sym_static] = ACTIONS(2614),
    [anon_sym_ATautoreleasepool] = ACTIONS(2616),
    [anon_sym_auto] = ACTIONS(2614),
    [anon_sym_register] = ACTIONS(2614),
    [anon_sym_inline] = ACTIONS(2614),
    [anon_sym___inline] = ACTIONS(2614),
    [anon_sym___inline__] = ACTIONS(2614),
    [anon_sym___forceinline] = ACTIONS(2614),
    [anon_sym_thread_local] = ACTIONS(2614),
    [anon_sym___thread] = ACTIONS(2614),
    [anon_sym_CG_EXTERN] = ACTIONS(2614),
    [anon_sym_CG_INLINE] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2614),
    [anon_sym_IBOutlet] = ACTIONS(2614),
    [anon_sym_IBInspectable] = ACTIONS(2614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2614),
    [anon_sym_NS_INLINE] = ACTIONS(2614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2614),
    [anon_sym_const] = ACTIONS(2614),
    [anon_sym_constexpr] = ACTIONS(2614),
    [anon_sym_volatile] = ACTIONS(2614),
    [anon_sym_restrict] = ACTIONS(2614),
    [anon_sym___restrict__] = ACTIONS(2614),
    [anon_sym__Atomic] = ACTIONS(2614),
    [anon_sym__Noreturn] = ACTIONS(2614),
    [anon_sym__Nonnull] = ACTIONS(2614),
    [anon_sym_nullable] = ACTIONS(2614),
    [anon_sym__Complex] = ACTIONS(2614),
    [anon_sym__Nullable] = ACTIONS(2614),
    [anon_sym__Nullable_result] = ACTIONS(2614),
    [anon_sym__Null_unspecified] = ACTIONS(2614),
    [anon_sym___autoreleasing] = ACTIONS(2614),
    [anon_sym___block] = ACTIONS(2614),
    [anon_sym___bridge] = ACTIONS(2614),
    [anon_sym___bridge_retained] = ACTIONS(2614),
    [anon_sym___bridge_transfer] = ACTIONS(2614),
    [anon_sym___complex] = ACTIONS(2614),
    [anon_sym___const] = ACTIONS(2614),
    [anon_sym___imag] = ACTIONS(2614),
    [anon_sym___kindof] = ACTIONS(2614),
    [anon_sym___nonnull] = ACTIONS(2614),
    [anon_sym___nullable] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2614),
    [anon_sym___real] = ACTIONS(2614),
    [anon_sym___strong] = ACTIONS(2614),
    [anon_sym___unsafe_unretained] = ACTIONS(2614),
    [anon_sym___unused] = ACTIONS(2614),
    [anon_sym___weak] = ACTIONS(2614),
    [anon_sym_alignas] = ACTIONS(2614),
    [anon_sym__Alignas] = ACTIONS(2614),
    [sym_primitive_type] = ACTIONS(2614),
    [anon_sym_enum] = ACTIONS(2614),
    [anon_sym_struct] = ACTIONS(2614),
    [anon_sym_union] = ACTIONS(2614),
    [anon_sym_if] = ACTIONS(2614),
    [anon_sym_switch] = ACTIONS(2614),
    [anon_sym_case] = ACTIONS(2614),
    [anon_sym_default] = ACTIONS(2614),
    [anon_sym_while] = ACTIONS(2614),
    [anon_sym_do] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2614),
    [anon_sym_in] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_break] = ACTIONS(2614),
    [anon_sym_continue] = ACTIONS(2614),
    [anon_sym_goto] = ACTIONS(2614),
    [anon_sym_DASH_DASH] = ACTIONS(2616),
    [anon_sym_PLUS_PLUS] = ACTIONS(2616),
    [anon_sym_sizeof] = ACTIONS(2614),
    [anon_sym___alignof__] = ACTIONS(2614),
    [anon_sym___alignof] = ACTIONS(2614),
    [anon_sym__alignof] = ACTIONS(2614),
    [anon_sym_alignof] = ACTIONS(2614),
    [anon_sym__Alignof] = ACTIONS(2614),
    [anon_sym_offsetof] = ACTIONS(2614),
    [anon_sym__Generic] = ACTIONS(2614),
    [anon_sym_asm] = ACTIONS(2614),
    [anon_sym___asm__] = ACTIONS(2614),
    [anon_sym___asm] = ACTIONS(2614),
    [sym_number_literal] = ACTIONS(2616),
    [anon_sym_L_SQUOTE] = ACTIONS(2616),
    [anon_sym_u_SQUOTE] = ACTIONS(2616),
    [anon_sym_U_SQUOTE] = ACTIONS(2616),
    [anon_sym_u8_SQUOTE] = ACTIONS(2616),
    [anon_sym_SQUOTE] = ACTIONS(2616),
    [anon_sym_AT] = ACTIONS(2614),
    [anon_sym_DQUOTE] = ACTIONS(2616),
    [anon_sym_L_DQUOTE] = ACTIONS(2616),
    [anon_sym_u_DQUOTE] = ACTIONS(2616),
    [anon_sym_U_DQUOTE] = ACTIONS(2616),
    [anon_sym_u8_DQUOTE] = ACTIONS(2616),
    [sym_true] = ACTIONS(2614),
    [sym_false] = ACTIONS(2614),
    [anon_sym_NULL] = ACTIONS(2614),
    [anon_sym_nullptr] = ACTIONS(2614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2614),
    [anon_sym___typeof] = ACTIONS(2614),
    [anon_sym_typeof] = ACTIONS(2614),
    [anon_sym_ATimport] = ACTIONS(2616),
    [aux_sym_preproc_undef_token1] = ACTIONS(2614),
    [anon_sym_POUND] = ACTIONS(2614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_API_AVAILABLE] = ACTIONS(2614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_API_DEPRECATED] = ACTIONS(2614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym___deprecated_msg] = ACTIONS(2614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2614),
    [anon_sym_ATprotocol] = ACTIONS(2616),
    [anon_sym_ATinterface] = ACTIONS(2616),
    [anon_sym_ATimplementation] = ACTIONS(2616),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2616),
    [anon_sym_ATsynthesize] = ACTIONS(2616),
    [anon_sym_ATdynamic] = ACTIONS(2616),
    [anon_sym_ATtry] = ACTIONS(2616),
    [anon_sym___try] = ACTIONS(2614),
    [anon_sym_ATthrow] = ACTIONS(2616),
    [anon_sym_ATselector] = ACTIONS(2616),
    [anon_sym_ATavailable] = ACTIONS(2616),
    [anon_sym___builtin_available] = ACTIONS(2614),
    [anon_sym_va_arg] = ACTIONS(2614),
    [anon_sym_ATencode] = ACTIONS(2616),
    [anon_sym_ATsynchronized] = ACTIONS(2616),
    [anon_sym_BOOL] = ACTIONS(2614),
    [anon_sym_IMP] = ACTIONS(2614),
    [anon_sym_SEL] = ACTIONS(2614),
    [anon_sym_Class] = ACTIONS(2614),
    [anon_sym_id] = ACTIONS(2614),
  },
  [STATE(662)] = {
    [sym_identifier] = ACTIONS(2622),
    [aux_sym_preproc_include_token1] = ACTIONS(2622),
    [aux_sym_preproc_include_token2] = ACTIONS(2622),
    [aux_sym_preproc_def_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token2] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2622),
    [sym_preproc_directive] = ACTIONS(2622),
    [anon_sym_LPAREN2] = ACTIONS(2624),
    [anon_sym_BANG] = ACTIONS(2624),
    [anon_sym_TILDE] = ACTIONS(2624),
    [anon_sym_DASH] = ACTIONS(2622),
    [anon_sym_PLUS] = ACTIONS(2622),
    [anon_sym_STAR] = ACTIONS(2624),
    [anon_sym_CARET] = ACTIONS(2624),
    [anon_sym_AMP] = ACTIONS(2624),
    [anon_sym_SEMI] = ACTIONS(2624),
    [anon_sym___extension__] = ACTIONS(2622),
    [anon_sym_typedef] = ACTIONS(2622),
    [anon_sym_extern] = ACTIONS(2622),
    [anon_sym___attribute__] = ACTIONS(2622),
    [anon_sym___attribute] = ACTIONS(2622),
    [anon_sym_noreturn] = ACTIONS(2622),
    [anon_sym_LBRACK] = ACTIONS(2624),
    [anon_sym___declspec] = ACTIONS(2622),
    [anon_sym___cdecl] = ACTIONS(2622),
    [anon_sym___clrcall] = ACTIONS(2622),
    [anon_sym___stdcall] = ACTIONS(2622),
    [anon_sym___fastcall] = ACTIONS(2622),
    [anon_sym___thiscall] = ACTIONS(2622),
    [anon_sym___vectorcall] = ACTIONS(2622),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_signed] = ACTIONS(2622),
    [anon_sym_unsigned] = ACTIONS(2622),
    [anon_sym_long] = ACTIONS(2622),
    [anon_sym_short] = ACTIONS(2622),
    [anon_sym_static] = ACTIONS(2622),
    [anon_sym_ATautoreleasepool] = ACTIONS(2624),
    [anon_sym_auto] = ACTIONS(2622),
    [anon_sym_register] = ACTIONS(2622),
    [anon_sym_inline] = ACTIONS(2622),
    [anon_sym___inline] = ACTIONS(2622),
    [anon_sym___inline__] = ACTIONS(2622),
    [anon_sym___forceinline] = ACTIONS(2622),
    [anon_sym_thread_local] = ACTIONS(2622),
    [anon_sym___thread] = ACTIONS(2622),
    [anon_sym_CG_EXTERN] = ACTIONS(2622),
    [anon_sym_CG_INLINE] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2622),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2622),
    [anon_sym_IBOutlet] = ACTIONS(2622),
    [anon_sym_IBInspectable] = ACTIONS(2622),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2622),
    [anon_sym_NS_INLINE] = ACTIONS(2622),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2622),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2622),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2622),
    [anon_sym_const] = ACTIONS(2622),
    [anon_sym_constexpr] = ACTIONS(2622),
    [anon_sym_volatile] = ACTIONS(2622),
    [anon_sym_restrict] = ACTIONS(2622),
    [anon_sym___restrict__] = ACTIONS(2622),
    [anon_sym__Atomic] = ACTIONS(2622),
    [anon_sym__Noreturn] = ACTIONS(2622),
    [anon_sym__Nonnull] = ACTIONS(2622),
    [anon_sym_nullable] = ACTIONS(2622),
    [anon_sym__Complex] = ACTIONS(2622),
    [anon_sym__Nullable] = ACTIONS(2622),
    [anon_sym__Nullable_result] = ACTIONS(2622),
    [anon_sym__Null_unspecified] = ACTIONS(2622),
    [anon_sym___autoreleasing] = ACTIONS(2622),
    [anon_sym___block] = ACTIONS(2622),
    [anon_sym___bridge] = ACTIONS(2622),
    [anon_sym___bridge_retained] = ACTIONS(2622),
    [anon_sym___bridge_transfer] = ACTIONS(2622),
    [anon_sym___complex] = ACTIONS(2622),
    [anon_sym___const] = ACTIONS(2622),
    [anon_sym___imag] = ACTIONS(2622),
    [anon_sym___kindof] = ACTIONS(2622),
    [anon_sym___nonnull] = ACTIONS(2622),
    [anon_sym___nullable] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2622),
    [anon_sym___real] = ACTIONS(2622),
    [anon_sym___strong] = ACTIONS(2622),
    [anon_sym___unsafe_unretained] = ACTIONS(2622),
    [anon_sym___unused] = ACTIONS(2622),
    [anon_sym___weak] = ACTIONS(2622),
    [anon_sym_alignas] = ACTIONS(2622),
    [anon_sym__Alignas] = ACTIONS(2622),
    [sym_primitive_type] = ACTIONS(2622),
    [anon_sym_enum] = ACTIONS(2622),
    [anon_sym_struct] = ACTIONS(2622),
    [anon_sym_union] = ACTIONS(2622),
    [anon_sym_if] = ACTIONS(2622),
    [anon_sym_switch] = ACTIONS(2622),
    [anon_sym_case] = ACTIONS(2622),
    [anon_sym_default] = ACTIONS(2622),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_do] = ACTIONS(2622),
    [anon_sym_for] = ACTIONS(2622),
    [anon_sym_in] = ACTIONS(2622),
    [anon_sym_return] = ACTIONS(2622),
    [anon_sym_break] = ACTIONS(2622),
    [anon_sym_continue] = ACTIONS(2622),
    [anon_sym_goto] = ACTIONS(2622),
    [anon_sym_DASH_DASH] = ACTIONS(2624),
    [anon_sym_PLUS_PLUS] = ACTIONS(2624),
    [anon_sym_sizeof] = ACTIONS(2622),
    [anon_sym___alignof__] = ACTIONS(2622),
    [anon_sym___alignof] = ACTIONS(2622),
    [anon_sym__alignof] = ACTIONS(2622),
    [anon_sym_alignof] = ACTIONS(2622),
    [anon_sym__Alignof] = ACTIONS(2622),
    [anon_sym_offsetof] = ACTIONS(2622),
    [anon_sym__Generic] = ACTIONS(2622),
    [anon_sym_asm] = ACTIONS(2622),
    [anon_sym___asm__] = ACTIONS(2622),
    [anon_sym___asm] = ACTIONS(2622),
    [sym_number_literal] = ACTIONS(2624),
    [anon_sym_L_SQUOTE] = ACTIONS(2624),
    [anon_sym_u_SQUOTE] = ACTIONS(2624),
    [anon_sym_U_SQUOTE] = ACTIONS(2624),
    [anon_sym_u8_SQUOTE] = ACTIONS(2624),
    [anon_sym_SQUOTE] = ACTIONS(2624),
    [anon_sym_AT] = ACTIONS(2622),
    [anon_sym_DQUOTE] = ACTIONS(2624),
    [anon_sym_L_DQUOTE] = ACTIONS(2624),
    [anon_sym_u_DQUOTE] = ACTIONS(2624),
    [anon_sym_U_DQUOTE] = ACTIONS(2624),
    [anon_sym_u8_DQUOTE] = ACTIONS(2624),
    [sym_true] = ACTIONS(2622),
    [sym_false] = ACTIONS(2622),
    [anon_sym_NULL] = ACTIONS(2622),
    [anon_sym_nullptr] = ACTIONS(2622),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2622),
    [anon_sym___typeof] = ACTIONS(2622),
    [anon_sym_typeof] = ACTIONS(2622),
    [anon_sym_ATimport] = ACTIONS(2624),
    [aux_sym_preproc_undef_token1] = ACTIONS(2622),
    [anon_sym_POUND] = ACTIONS(2622),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2622),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2622),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2622),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_API_AVAILABLE] = ACTIONS(2622),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_API_DEPRECATED] = ACTIONS(2622),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym___deprecated_msg] = ACTIONS(2622),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2622),
    [anon_sym_ATprotocol] = ACTIONS(2624),
    [anon_sym_ATinterface] = ACTIONS(2624),
    [anon_sym_ATimplementation] = ACTIONS(2624),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2624),
    [anon_sym_ATsynthesize] = ACTIONS(2624),
    [anon_sym_ATdynamic] = ACTIONS(2624),
    [anon_sym_ATtry] = ACTIONS(2624),
    [anon_sym___try] = ACTIONS(2622),
    [anon_sym_ATthrow] = ACTIONS(2624),
    [anon_sym_ATselector] = ACTIONS(2624),
    [anon_sym_ATavailable] = ACTIONS(2624),
    [anon_sym___builtin_available] = ACTIONS(2622),
    [anon_sym_va_arg] = ACTIONS(2622),
    [anon_sym_ATencode] = ACTIONS(2624),
    [anon_sym_ATsynchronized] = ACTIONS(2624),
    [anon_sym_BOOL] = ACTIONS(2622),
    [anon_sym_IMP] = ACTIONS(2622),
    [anon_sym_SEL] = ACTIONS(2622),
    [anon_sym_Class] = ACTIONS(2622),
    [anon_sym_id] = ACTIONS(2622),
  },
  [STATE(663)] = {
    [sym_identifier] = ACTIONS(2626),
    [aux_sym_preproc_include_token1] = ACTIONS(2626),
    [aux_sym_preproc_include_token2] = ACTIONS(2626),
    [aux_sym_preproc_def_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token2] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2626),
    [sym_preproc_directive] = ACTIONS(2626),
    [anon_sym_LPAREN2] = ACTIONS(2628),
    [anon_sym_BANG] = ACTIONS(2628),
    [anon_sym_TILDE] = ACTIONS(2628),
    [anon_sym_DASH] = ACTIONS(2626),
    [anon_sym_PLUS] = ACTIONS(2626),
    [anon_sym_STAR] = ACTIONS(2628),
    [anon_sym_CARET] = ACTIONS(2628),
    [anon_sym_AMP] = ACTIONS(2628),
    [anon_sym_SEMI] = ACTIONS(2628),
    [anon_sym___extension__] = ACTIONS(2626),
    [anon_sym_typedef] = ACTIONS(2626),
    [anon_sym_extern] = ACTIONS(2626),
    [anon_sym___attribute__] = ACTIONS(2626),
    [anon_sym___attribute] = ACTIONS(2626),
    [anon_sym_noreturn] = ACTIONS(2626),
    [anon_sym_LBRACK] = ACTIONS(2628),
    [anon_sym___declspec] = ACTIONS(2626),
    [anon_sym___cdecl] = ACTIONS(2626),
    [anon_sym___clrcall] = ACTIONS(2626),
    [anon_sym___stdcall] = ACTIONS(2626),
    [anon_sym___fastcall] = ACTIONS(2626),
    [anon_sym___thiscall] = ACTIONS(2626),
    [anon_sym___vectorcall] = ACTIONS(2626),
    [anon_sym_LBRACE] = ACTIONS(2628),
    [anon_sym_signed] = ACTIONS(2626),
    [anon_sym_unsigned] = ACTIONS(2626),
    [anon_sym_long] = ACTIONS(2626),
    [anon_sym_short] = ACTIONS(2626),
    [anon_sym_static] = ACTIONS(2626),
    [anon_sym_ATautoreleasepool] = ACTIONS(2628),
    [anon_sym_auto] = ACTIONS(2626),
    [anon_sym_register] = ACTIONS(2626),
    [anon_sym_inline] = ACTIONS(2626),
    [anon_sym___inline] = ACTIONS(2626),
    [anon_sym___inline__] = ACTIONS(2626),
    [anon_sym___forceinline] = ACTIONS(2626),
    [anon_sym_thread_local] = ACTIONS(2626),
    [anon_sym___thread] = ACTIONS(2626),
    [anon_sym_CG_EXTERN] = ACTIONS(2626),
    [anon_sym_CG_INLINE] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2626),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2626),
    [anon_sym_IBOutlet] = ACTIONS(2626),
    [anon_sym_IBInspectable] = ACTIONS(2626),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2626),
    [anon_sym_NS_INLINE] = ACTIONS(2626),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2626),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2626),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2626),
    [anon_sym_const] = ACTIONS(2626),
    [anon_sym_constexpr] = ACTIONS(2626),
    [anon_sym_volatile] = ACTIONS(2626),
    [anon_sym_restrict] = ACTIONS(2626),
    [anon_sym___restrict__] = ACTIONS(2626),
    [anon_sym__Atomic] = ACTIONS(2626),
    [anon_sym__Noreturn] = ACTIONS(2626),
    [anon_sym__Nonnull] = ACTIONS(2626),
    [anon_sym_nullable] = ACTIONS(2626),
    [anon_sym__Complex] = ACTIONS(2626),
    [anon_sym__Nullable] = ACTIONS(2626),
    [anon_sym__Nullable_result] = ACTIONS(2626),
    [anon_sym__Null_unspecified] = ACTIONS(2626),
    [anon_sym___autoreleasing] = ACTIONS(2626),
    [anon_sym___block] = ACTIONS(2626),
    [anon_sym___bridge] = ACTIONS(2626),
    [anon_sym___bridge_retained] = ACTIONS(2626),
    [anon_sym___bridge_transfer] = ACTIONS(2626),
    [anon_sym___complex] = ACTIONS(2626),
    [anon_sym___const] = ACTIONS(2626),
    [anon_sym___imag] = ACTIONS(2626),
    [anon_sym___kindof] = ACTIONS(2626),
    [anon_sym___nonnull] = ACTIONS(2626),
    [anon_sym___nullable] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2626),
    [anon_sym___real] = ACTIONS(2626),
    [anon_sym___strong] = ACTIONS(2626),
    [anon_sym___unsafe_unretained] = ACTIONS(2626),
    [anon_sym___unused] = ACTIONS(2626),
    [anon_sym___weak] = ACTIONS(2626),
    [anon_sym_alignas] = ACTIONS(2626),
    [anon_sym__Alignas] = ACTIONS(2626),
    [sym_primitive_type] = ACTIONS(2626),
    [anon_sym_enum] = ACTIONS(2626),
    [anon_sym_struct] = ACTIONS(2626),
    [anon_sym_union] = ACTIONS(2626),
    [anon_sym_if] = ACTIONS(2626),
    [anon_sym_switch] = ACTIONS(2626),
    [anon_sym_case] = ACTIONS(2626),
    [anon_sym_default] = ACTIONS(2626),
    [anon_sym_while] = ACTIONS(2626),
    [anon_sym_do] = ACTIONS(2626),
    [anon_sym_for] = ACTIONS(2626),
    [anon_sym_in] = ACTIONS(2626),
    [anon_sym_return] = ACTIONS(2626),
    [anon_sym_break] = ACTIONS(2626),
    [anon_sym_continue] = ACTIONS(2626),
    [anon_sym_goto] = ACTIONS(2626),
    [anon_sym_DASH_DASH] = ACTIONS(2628),
    [anon_sym_PLUS_PLUS] = ACTIONS(2628),
    [anon_sym_sizeof] = ACTIONS(2626),
    [anon_sym___alignof__] = ACTIONS(2626),
    [anon_sym___alignof] = ACTIONS(2626),
    [anon_sym__alignof] = ACTIONS(2626),
    [anon_sym_alignof] = ACTIONS(2626),
    [anon_sym__Alignof] = ACTIONS(2626),
    [anon_sym_offsetof] = ACTIONS(2626),
    [anon_sym__Generic] = ACTIONS(2626),
    [anon_sym_asm] = ACTIONS(2626),
    [anon_sym___asm__] = ACTIONS(2626),
    [anon_sym___asm] = ACTIONS(2626),
    [sym_number_literal] = ACTIONS(2628),
    [anon_sym_L_SQUOTE] = ACTIONS(2628),
    [anon_sym_u_SQUOTE] = ACTIONS(2628),
    [anon_sym_U_SQUOTE] = ACTIONS(2628),
    [anon_sym_u8_SQUOTE] = ACTIONS(2628),
    [anon_sym_SQUOTE] = ACTIONS(2628),
    [anon_sym_AT] = ACTIONS(2626),
    [anon_sym_DQUOTE] = ACTIONS(2628),
    [anon_sym_L_DQUOTE] = ACTIONS(2628),
    [anon_sym_u_DQUOTE] = ACTIONS(2628),
    [anon_sym_U_DQUOTE] = ACTIONS(2628),
    [anon_sym_u8_DQUOTE] = ACTIONS(2628),
    [sym_true] = ACTIONS(2626),
    [sym_false] = ACTIONS(2626),
    [anon_sym_NULL] = ACTIONS(2626),
    [anon_sym_nullptr] = ACTIONS(2626),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2626),
    [anon_sym___typeof] = ACTIONS(2626),
    [anon_sym_typeof] = ACTIONS(2626),
    [anon_sym_ATimport] = ACTIONS(2628),
    [aux_sym_preproc_undef_token1] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2626),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2626),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2626),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2626),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_API_AVAILABLE] = ACTIONS(2626),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_API_DEPRECATED] = ACTIONS(2626),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym___deprecated_msg] = ACTIONS(2626),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2626),
    [anon_sym_ATprotocol] = ACTIONS(2628),
    [anon_sym_ATinterface] = ACTIONS(2628),
    [anon_sym_ATimplementation] = ACTIONS(2628),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2628),
    [anon_sym_ATsynthesize] = ACTIONS(2628),
    [anon_sym_ATdynamic] = ACTIONS(2628),
    [anon_sym_ATtry] = ACTIONS(2628),
    [anon_sym___try] = ACTIONS(2626),
    [anon_sym_ATthrow] = ACTIONS(2628),
    [anon_sym_ATselector] = ACTIONS(2628),
    [anon_sym_ATavailable] = ACTIONS(2628),
    [anon_sym___builtin_available] = ACTIONS(2626),
    [anon_sym_va_arg] = ACTIONS(2626),
    [anon_sym_ATencode] = ACTIONS(2628),
    [anon_sym_ATsynchronized] = ACTIONS(2628),
    [anon_sym_BOOL] = ACTIONS(2626),
    [anon_sym_IMP] = ACTIONS(2626),
    [anon_sym_SEL] = ACTIONS(2626),
    [anon_sym_Class] = ACTIONS(2626),
    [anon_sym_id] = ACTIONS(2626),
  },
  [STATE(664)] = {
    [sym_identifier] = ACTIONS(2630),
    [aux_sym_preproc_include_token1] = ACTIONS(2630),
    [aux_sym_preproc_include_token2] = ACTIONS(2630),
    [aux_sym_preproc_def_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token2] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2630),
    [sym_preproc_directive] = ACTIONS(2630),
    [anon_sym_LPAREN2] = ACTIONS(2632),
    [anon_sym_BANG] = ACTIONS(2632),
    [anon_sym_TILDE] = ACTIONS(2632),
    [anon_sym_DASH] = ACTIONS(2630),
    [anon_sym_PLUS] = ACTIONS(2630),
    [anon_sym_STAR] = ACTIONS(2632),
    [anon_sym_CARET] = ACTIONS(2632),
    [anon_sym_AMP] = ACTIONS(2632),
    [anon_sym_SEMI] = ACTIONS(2632),
    [anon_sym___extension__] = ACTIONS(2630),
    [anon_sym_typedef] = ACTIONS(2630),
    [anon_sym_extern] = ACTIONS(2630),
    [anon_sym___attribute__] = ACTIONS(2630),
    [anon_sym___attribute] = ACTIONS(2630),
    [anon_sym_noreturn] = ACTIONS(2630),
    [anon_sym_LBRACK] = ACTIONS(2632),
    [anon_sym___declspec] = ACTIONS(2630),
    [anon_sym___cdecl] = ACTIONS(2630),
    [anon_sym___clrcall] = ACTIONS(2630),
    [anon_sym___stdcall] = ACTIONS(2630),
    [anon_sym___fastcall] = ACTIONS(2630),
    [anon_sym___thiscall] = ACTIONS(2630),
    [anon_sym___vectorcall] = ACTIONS(2630),
    [anon_sym_LBRACE] = ACTIONS(2632),
    [anon_sym_signed] = ACTIONS(2630),
    [anon_sym_unsigned] = ACTIONS(2630),
    [anon_sym_long] = ACTIONS(2630),
    [anon_sym_short] = ACTIONS(2630),
    [anon_sym_static] = ACTIONS(2630),
    [anon_sym_ATautoreleasepool] = ACTIONS(2632),
    [anon_sym_auto] = ACTIONS(2630),
    [anon_sym_register] = ACTIONS(2630),
    [anon_sym_inline] = ACTIONS(2630),
    [anon_sym___inline] = ACTIONS(2630),
    [anon_sym___inline__] = ACTIONS(2630),
    [anon_sym___forceinline] = ACTIONS(2630),
    [anon_sym_thread_local] = ACTIONS(2630),
    [anon_sym___thread] = ACTIONS(2630),
    [anon_sym_CG_EXTERN] = ACTIONS(2630),
    [anon_sym_CG_INLINE] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2630),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2630),
    [anon_sym_IBOutlet] = ACTIONS(2630),
    [anon_sym_IBInspectable] = ACTIONS(2630),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2630),
    [anon_sym_NS_INLINE] = ACTIONS(2630),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2630),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2630),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2630),
    [anon_sym_const] = ACTIONS(2630),
    [anon_sym_constexpr] = ACTIONS(2630),
    [anon_sym_volatile] = ACTIONS(2630),
    [anon_sym_restrict] = ACTIONS(2630),
    [anon_sym___restrict__] = ACTIONS(2630),
    [anon_sym__Atomic] = ACTIONS(2630),
    [anon_sym__Noreturn] = ACTIONS(2630),
    [anon_sym__Nonnull] = ACTIONS(2630),
    [anon_sym_nullable] = ACTIONS(2630),
    [anon_sym__Complex] = ACTIONS(2630),
    [anon_sym__Nullable] = ACTIONS(2630),
    [anon_sym__Nullable_result] = ACTIONS(2630),
    [anon_sym__Null_unspecified] = ACTIONS(2630),
    [anon_sym___autoreleasing] = ACTIONS(2630),
    [anon_sym___block] = ACTIONS(2630),
    [anon_sym___bridge] = ACTIONS(2630),
    [anon_sym___bridge_retained] = ACTIONS(2630),
    [anon_sym___bridge_transfer] = ACTIONS(2630),
    [anon_sym___complex] = ACTIONS(2630),
    [anon_sym___const] = ACTIONS(2630),
    [anon_sym___imag] = ACTIONS(2630),
    [anon_sym___kindof] = ACTIONS(2630),
    [anon_sym___nonnull] = ACTIONS(2630),
    [anon_sym___nullable] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2630),
    [anon_sym___real] = ACTIONS(2630),
    [anon_sym___strong] = ACTIONS(2630),
    [anon_sym___unsafe_unretained] = ACTIONS(2630),
    [anon_sym___unused] = ACTIONS(2630),
    [anon_sym___weak] = ACTIONS(2630),
    [anon_sym_alignas] = ACTIONS(2630),
    [anon_sym__Alignas] = ACTIONS(2630),
    [sym_primitive_type] = ACTIONS(2630),
    [anon_sym_enum] = ACTIONS(2630),
    [anon_sym_struct] = ACTIONS(2630),
    [anon_sym_union] = ACTIONS(2630),
    [anon_sym_if] = ACTIONS(2630),
    [anon_sym_switch] = ACTIONS(2630),
    [anon_sym_case] = ACTIONS(2630),
    [anon_sym_default] = ACTIONS(2630),
    [anon_sym_while] = ACTIONS(2630),
    [anon_sym_do] = ACTIONS(2630),
    [anon_sym_for] = ACTIONS(2630),
    [anon_sym_in] = ACTIONS(2630),
    [anon_sym_return] = ACTIONS(2630),
    [anon_sym_break] = ACTIONS(2630),
    [anon_sym_continue] = ACTIONS(2630),
    [anon_sym_goto] = ACTIONS(2630),
    [anon_sym_DASH_DASH] = ACTIONS(2632),
    [anon_sym_PLUS_PLUS] = ACTIONS(2632),
    [anon_sym_sizeof] = ACTIONS(2630),
    [anon_sym___alignof__] = ACTIONS(2630),
    [anon_sym___alignof] = ACTIONS(2630),
    [anon_sym__alignof] = ACTIONS(2630),
    [anon_sym_alignof] = ACTIONS(2630),
    [anon_sym__Alignof] = ACTIONS(2630),
    [anon_sym_offsetof] = ACTIONS(2630),
    [anon_sym__Generic] = ACTIONS(2630),
    [anon_sym_asm] = ACTIONS(2630),
    [anon_sym___asm__] = ACTIONS(2630),
    [anon_sym___asm] = ACTIONS(2630),
    [sym_number_literal] = ACTIONS(2632),
    [anon_sym_L_SQUOTE] = ACTIONS(2632),
    [anon_sym_u_SQUOTE] = ACTIONS(2632),
    [anon_sym_U_SQUOTE] = ACTIONS(2632),
    [anon_sym_u8_SQUOTE] = ACTIONS(2632),
    [anon_sym_SQUOTE] = ACTIONS(2632),
    [anon_sym_AT] = ACTIONS(2630),
    [anon_sym_DQUOTE] = ACTIONS(2632),
    [anon_sym_L_DQUOTE] = ACTIONS(2632),
    [anon_sym_u_DQUOTE] = ACTIONS(2632),
    [anon_sym_U_DQUOTE] = ACTIONS(2632),
    [anon_sym_u8_DQUOTE] = ACTIONS(2632),
    [sym_true] = ACTIONS(2630),
    [sym_false] = ACTIONS(2630),
    [anon_sym_NULL] = ACTIONS(2630),
    [anon_sym_nullptr] = ACTIONS(2630),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2630),
    [anon_sym___typeof] = ACTIONS(2630),
    [anon_sym_typeof] = ACTIONS(2630),
    [anon_sym_ATimport] = ACTIONS(2632),
    [aux_sym_preproc_undef_token1] = ACTIONS(2630),
    [anon_sym_POUND] = ACTIONS(2630),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2630),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2630),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2630),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_API_AVAILABLE] = ACTIONS(2630),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_API_DEPRECATED] = ACTIONS(2630),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym___deprecated_msg] = ACTIONS(2630),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2630),
    [anon_sym_ATprotocol] = ACTIONS(2632),
    [anon_sym_ATinterface] = ACTIONS(2632),
    [anon_sym_ATimplementation] = ACTIONS(2632),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2632),
    [anon_sym_ATsynthesize] = ACTIONS(2632),
    [anon_sym_ATdynamic] = ACTIONS(2632),
    [anon_sym_ATtry] = ACTIONS(2632),
    [anon_sym___try] = ACTIONS(2630),
    [anon_sym_ATthrow] = ACTIONS(2632),
    [anon_sym_ATselector] = ACTIONS(2632),
    [anon_sym_ATavailable] = ACTIONS(2632),
    [anon_sym___builtin_available] = ACTIONS(2630),
    [anon_sym_va_arg] = ACTIONS(2630),
    [anon_sym_ATencode] = ACTIONS(2632),
    [anon_sym_ATsynchronized] = ACTIONS(2632),
    [anon_sym_BOOL] = ACTIONS(2630),
    [anon_sym_IMP] = ACTIONS(2630),
    [anon_sym_SEL] = ACTIONS(2630),
    [anon_sym_Class] = ACTIONS(2630),
    [anon_sym_id] = ACTIONS(2630),
  },
  [STATE(665)] = {
    [sym_identifier] = ACTIONS(2638),
    [aux_sym_preproc_include_token1] = ACTIONS(2638),
    [aux_sym_preproc_include_token2] = ACTIONS(2638),
    [aux_sym_preproc_def_token1] = ACTIONS(2638),
    [aux_sym_preproc_if_token1] = ACTIONS(2638),
    [aux_sym_preproc_if_token2] = ACTIONS(2638),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2638),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2638),
    [sym_preproc_directive] = ACTIONS(2638),
    [anon_sym_LPAREN2] = ACTIONS(2640),
    [anon_sym_BANG] = ACTIONS(2640),
    [anon_sym_TILDE] = ACTIONS(2640),
    [anon_sym_DASH] = ACTIONS(2638),
    [anon_sym_PLUS] = ACTIONS(2638),
    [anon_sym_STAR] = ACTIONS(2640),
    [anon_sym_CARET] = ACTIONS(2640),
    [anon_sym_AMP] = ACTIONS(2640),
    [anon_sym_SEMI] = ACTIONS(2640),
    [anon_sym___extension__] = ACTIONS(2638),
    [anon_sym_typedef] = ACTIONS(2638),
    [anon_sym_extern] = ACTIONS(2638),
    [anon_sym___attribute__] = ACTIONS(2638),
    [anon_sym___attribute] = ACTIONS(2638),
    [anon_sym_noreturn] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2640),
    [anon_sym___declspec] = ACTIONS(2638),
    [anon_sym___cdecl] = ACTIONS(2638),
    [anon_sym___clrcall] = ACTIONS(2638),
    [anon_sym___stdcall] = ACTIONS(2638),
    [anon_sym___fastcall] = ACTIONS(2638),
    [anon_sym___thiscall] = ACTIONS(2638),
    [anon_sym___vectorcall] = ACTIONS(2638),
    [anon_sym_LBRACE] = ACTIONS(2640),
    [anon_sym_signed] = ACTIONS(2638),
    [anon_sym_unsigned] = ACTIONS(2638),
    [anon_sym_long] = ACTIONS(2638),
    [anon_sym_short] = ACTIONS(2638),
    [anon_sym_static] = ACTIONS(2638),
    [anon_sym_ATautoreleasepool] = ACTIONS(2640),
    [anon_sym_auto] = ACTIONS(2638),
    [anon_sym_register] = ACTIONS(2638),
    [anon_sym_inline] = ACTIONS(2638),
    [anon_sym___inline] = ACTIONS(2638),
    [anon_sym___inline__] = ACTIONS(2638),
    [anon_sym___forceinline] = ACTIONS(2638),
    [anon_sym_thread_local] = ACTIONS(2638),
    [anon_sym___thread] = ACTIONS(2638),
    [anon_sym_CG_EXTERN] = ACTIONS(2638),
    [anon_sym_CG_INLINE] = ACTIONS(2638),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2638),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2638),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2638),
    [anon_sym_IBOutlet] = ACTIONS(2638),
    [anon_sym_IBInspectable] = ACTIONS(2638),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2638),
    [anon_sym_NS_INLINE] = ACTIONS(2638),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2638),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2638),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2638),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2638),
    [anon_sym_const] = ACTIONS(2638),
    [anon_sym_constexpr] = ACTIONS(2638),
    [anon_sym_volatile] = ACTIONS(2638),
    [anon_sym_restrict] = ACTIONS(2638),
    [anon_sym___restrict__] = ACTIONS(2638),
    [anon_sym__Atomic] = ACTIONS(2638),
    [anon_sym__Noreturn] = ACTIONS(2638),
    [anon_sym__Nonnull] = ACTIONS(2638),
    [anon_sym_nullable] = ACTIONS(2638),
    [anon_sym__Complex] = ACTIONS(2638),
    [anon_sym__Nullable] = ACTIONS(2638),
    [anon_sym__Nullable_result] = ACTIONS(2638),
    [anon_sym__Null_unspecified] = ACTIONS(2638),
    [anon_sym___autoreleasing] = ACTIONS(2638),
    [anon_sym___block] = ACTIONS(2638),
    [anon_sym___bridge] = ACTIONS(2638),
    [anon_sym___bridge_retained] = ACTIONS(2638),
    [anon_sym___bridge_transfer] = ACTIONS(2638),
    [anon_sym___complex] = ACTIONS(2638),
    [anon_sym___const] = ACTIONS(2638),
    [anon_sym___imag] = ACTIONS(2638),
    [anon_sym___kindof] = ACTIONS(2638),
    [anon_sym___nonnull] = ACTIONS(2638),
    [anon_sym___nullable] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2638),
    [anon_sym___real] = ACTIONS(2638),
    [anon_sym___strong] = ACTIONS(2638),
    [anon_sym___unsafe_unretained] = ACTIONS(2638),
    [anon_sym___unused] = ACTIONS(2638),
    [anon_sym___weak] = ACTIONS(2638),
    [anon_sym_alignas] = ACTIONS(2638),
    [anon_sym__Alignas] = ACTIONS(2638),
    [sym_primitive_type] = ACTIONS(2638),
    [anon_sym_enum] = ACTIONS(2638),
    [anon_sym_struct] = ACTIONS(2638),
    [anon_sym_union] = ACTIONS(2638),
    [anon_sym_if] = ACTIONS(2638),
    [anon_sym_switch] = ACTIONS(2638),
    [anon_sym_case] = ACTIONS(2638),
    [anon_sym_default] = ACTIONS(2638),
    [anon_sym_while] = ACTIONS(2638),
    [anon_sym_do] = ACTIONS(2638),
    [anon_sym_for] = ACTIONS(2638),
    [anon_sym_in] = ACTIONS(2638),
    [anon_sym_return] = ACTIONS(2638),
    [anon_sym_break] = ACTIONS(2638),
    [anon_sym_continue] = ACTIONS(2638),
    [anon_sym_goto] = ACTIONS(2638),
    [anon_sym_DASH_DASH] = ACTIONS(2640),
    [anon_sym_PLUS_PLUS] = ACTIONS(2640),
    [anon_sym_sizeof] = ACTIONS(2638),
    [anon_sym___alignof__] = ACTIONS(2638),
    [anon_sym___alignof] = ACTIONS(2638),
    [anon_sym__alignof] = ACTIONS(2638),
    [anon_sym_alignof] = ACTIONS(2638),
    [anon_sym__Alignof] = ACTIONS(2638),
    [anon_sym_offsetof] = ACTIONS(2638),
    [anon_sym__Generic] = ACTIONS(2638),
    [anon_sym_asm] = ACTIONS(2638),
    [anon_sym___asm__] = ACTIONS(2638),
    [anon_sym___asm] = ACTIONS(2638),
    [sym_number_literal] = ACTIONS(2640),
    [anon_sym_L_SQUOTE] = ACTIONS(2640),
    [anon_sym_u_SQUOTE] = ACTIONS(2640),
    [anon_sym_U_SQUOTE] = ACTIONS(2640),
    [anon_sym_u8_SQUOTE] = ACTIONS(2640),
    [anon_sym_SQUOTE] = ACTIONS(2640),
    [anon_sym_AT] = ACTIONS(2638),
    [anon_sym_DQUOTE] = ACTIONS(2640),
    [anon_sym_L_DQUOTE] = ACTIONS(2640),
    [anon_sym_u_DQUOTE] = ACTIONS(2640),
    [anon_sym_U_DQUOTE] = ACTIONS(2640),
    [anon_sym_u8_DQUOTE] = ACTIONS(2640),
    [sym_true] = ACTIONS(2638),
    [sym_false] = ACTIONS(2638),
    [anon_sym_NULL] = ACTIONS(2638),
    [anon_sym_nullptr] = ACTIONS(2638),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2638),
    [anon_sym___typeof] = ACTIONS(2638),
    [anon_sym_typeof] = ACTIONS(2638),
    [anon_sym_ATimport] = ACTIONS(2640),
    [aux_sym_preproc_undef_token1] = ACTIONS(2638),
    [anon_sym_POUND] = ACTIONS(2638),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2638),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2638),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2638),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2638),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2638),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2638),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2638),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_API_AVAILABLE] = ACTIONS(2638),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_API_DEPRECATED] = ACTIONS(2638),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2638),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym___deprecated_msg] = ACTIONS(2638),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2638),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2638),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2638),
    [anon_sym_ATprotocol] = ACTIONS(2640),
    [anon_sym_ATinterface] = ACTIONS(2640),
    [anon_sym_ATimplementation] = ACTIONS(2640),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2640),
    [anon_sym_ATsynthesize] = ACTIONS(2640),
    [anon_sym_ATdynamic] = ACTIONS(2640),
    [anon_sym_ATtry] = ACTIONS(2640),
    [anon_sym___try] = ACTIONS(2638),
    [anon_sym_ATthrow] = ACTIONS(2640),
    [anon_sym_ATselector] = ACTIONS(2640),
    [anon_sym_ATavailable] = ACTIONS(2640),
    [anon_sym___builtin_available] = ACTIONS(2638),
    [anon_sym_va_arg] = ACTIONS(2638),
    [anon_sym_ATencode] = ACTIONS(2640),
    [anon_sym_ATsynchronized] = ACTIONS(2640),
    [anon_sym_BOOL] = ACTIONS(2638),
    [anon_sym_IMP] = ACTIONS(2638),
    [anon_sym_SEL] = ACTIONS(2638),
    [anon_sym_Class] = ACTIONS(2638),
    [anon_sym_id] = ACTIONS(2638),
  },
  [STATE(666)] = {
    [sym_identifier] = ACTIONS(2642),
    [aux_sym_preproc_include_token1] = ACTIONS(2642),
    [aux_sym_preproc_include_token2] = ACTIONS(2642),
    [aux_sym_preproc_def_token1] = ACTIONS(2642),
    [aux_sym_preproc_if_token1] = ACTIONS(2642),
    [aux_sym_preproc_if_token2] = ACTIONS(2642),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2642),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2642),
    [sym_preproc_directive] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_BANG] = ACTIONS(2644),
    [anon_sym_TILDE] = ACTIONS(2644),
    [anon_sym_DASH] = ACTIONS(2642),
    [anon_sym_PLUS] = ACTIONS(2642),
    [anon_sym_STAR] = ACTIONS(2644),
    [anon_sym_CARET] = ACTIONS(2644),
    [anon_sym_AMP] = ACTIONS(2644),
    [anon_sym_SEMI] = ACTIONS(2644),
    [anon_sym___extension__] = ACTIONS(2642),
    [anon_sym_typedef] = ACTIONS(2642),
    [anon_sym_extern] = ACTIONS(2642),
    [anon_sym___attribute__] = ACTIONS(2642),
    [anon_sym___attribute] = ACTIONS(2642),
    [anon_sym_noreturn] = ACTIONS(2642),
    [anon_sym_LBRACK] = ACTIONS(2644),
    [anon_sym___declspec] = ACTIONS(2642),
    [anon_sym___cdecl] = ACTIONS(2642),
    [anon_sym___clrcall] = ACTIONS(2642),
    [anon_sym___stdcall] = ACTIONS(2642),
    [anon_sym___fastcall] = ACTIONS(2642),
    [anon_sym___thiscall] = ACTIONS(2642),
    [anon_sym___vectorcall] = ACTIONS(2642),
    [anon_sym_LBRACE] = ACTIONS(2644),
    [anon_sym_signed] = ACTIONS(2642),
    [anon_sym_unsigned] = ACTIONS(2642),
    [anon_sym_long] = ACTIONS(2642),
    [anon_sym_short] = ACTIONS(2642),
    [anon_sym_static] = ACTIONS(2642),
    [anon_sym_ATautoreleasepool] = ACTIONS(2644),
    [anon_sym_auto] = ACTIONS(2642),
    [anon_sym_register] = ACTIONS(2642),
    [anon_sym_inline] = ACTIONS(2642),
    [anon_sym___inline] = ACTIONS(2642),
    [anon_sym___inline__] = ACTIONS(2642),
    [anon_sym___forceinline] = ACTIONS(2642),
    [anon_sym_thread_local] = ACTIONS(2642),
    [anon_sym___thread] = ACTIONS(2642),
    [anon_sym_CG_EXTERN] = ACTIONS(2642),
    [anon_sym_CG_INLINE] = ACTIONS(2642),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2642),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2642),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2642),
    [anon_sym_IBOutlet] = ACTIONS(2642),
    [anon_sym_IBInspectable] = ACTIONS(2642),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2642),
    [anon_sym_NS_INLINE] = ACTIONS(2642),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2642),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2642),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2642),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2642),
    [anon_sym_const] = ACTIONS(2642),
    [anon_sym_constexpr] = ACTIONS(2642),
    [anon_sym_volatile] = ACTIONS(2642),
    [anon_sym_restrict] = ACTIONS(2642),
    [anon_sym___restrict__] = ACTIONS(2642),
    [anon_sym__Atomic] = ACTIONS(2642),
    [anon_sym__Noreturn] = ACTIONS(2642),
    [anon_sym__Nonnull] = ACTIONS(2642),
    [anon_sym_nullable] = ACTIONS(2642),
    [anon_sym__Complex] = ACTIONS(2642),
    [anon_sym__Nullable] = ACTIONS(2642),
    [anon_sym__Nullable_result] = ACTIONS(2642),
    [anon_sym__Null_unspecified] = ACTIONS(2642),
    [anon_sym___autoreleasing] = ACTIONS(2642),
    [anon_sym___block] = ACTIONS(2642),
    [anon_sym___bridge] = ACTIONS(2642),
    [anon_sym___bridge_retained] = ACTIONS(2642),
    [anon_sym___bridge_transfer] = ACTIONS(2642),
    [anon_sym___complex] = ACTIONS(2642),
    [anon_sym___const] = ACTIONS(2642),
    [anon_sym___imag] = ACTIONS(2642),
    [anon_sym___kindof] = ACTIONS(2642),
    [anon_sym___nonnull] = ACTIONS(2642),
    [anon_sym___nullable] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2642),
    [anon_sym___real] = ACTIONS(2642),
    [anon_sym___strong] = ACTIONS(2642),
    [anon_sym___unsafe_unretained] = ACTIONS(2642),
    [anon_sym___unused] = ACTIONS(2642),
    [anon_sym___weak] = ACTIONS(2642),
    [anon_sym_alignas] = ACTIONS(2642),
    [anon_sym__Alignas] = ACTIONS(2642),
    [sym_primitive_type] = ACTIONS(2642),
    [anon_sym_enum] = ACTIONS(2642),
    [anon_sym_struct] = ACTIONS(2642),
    [anon_sym_union] = ACTIONS(2642),
    [anon_sym_if] = ACTIONS(2642),
    [anon_sym_switch] = ACTIONS(2642),
    [anon_sym_case] = ACTIONS(2642),
    [anon_sym_default] = ACTIONS(2642),
    [anon_sym_while] = ACTIONS(2642),
    [anon_sym_do] = ACTIONS(2642),
    [anon_sym_for] = ACTIONS(2642),
    [anon_sym_in] = ACTIONS(2642),
    [anon_sym_return] = ACTIONS(2642),
    [anon_sym_break] = ACTIONS(2642),
    [anon_sym_continue] = ACTIONS(2642),
    [anon_sym_goto] = ACTIONS(2642),
    [anon_sym_DASH_DASH] = ACTIONS(2644),
    [anon_sym_PLUS_PLUS] = ACTIONS(2644),
    [anon_sym_sizeof] = ACTIONS(2642),
    [anon_sym___alignof__] = ACTIONS(2642),
    [anon_sym___alignof] = ACTIONS(2642),
    [anon_sym__alignof] = ACTIONS(2642),
    [anon_sym_alignof] = ACTIONS(2642),
    [anon_sym__Alignof] = ACTIONS(2642),
    [anon_sym_offsetof] = ACTIONS(2642),
    [anon_sym__Generic] = ACTIONS(2642),
    [anon_sym_asm] = ACTIONS(2642),
    [anon_sym___asm__] = ACTIONS(2642),
    [anon_sym___asm] = ACTIONS(2642),
    [sym_number_literal] = ACTIONS(2644),
    [anon_sym_L_SQUOTE] = ACTIONS(2644),
    [anon_sym_u_SQUOTE] = ACTIONS(2644),
    [anon_sym_U_SQUOTE] = ACTIONS(2644),
    [anon_sym_u8_SQUOTE] = ACTIONS(2644),
    [anon_sym_SQUOTE] = ACTIONS(2644),
    [anon_sym_AT] = ACTIONS(2642),
    [anon_sym_DQUOTE] = ACTIONS(2644),
    [anon_sym_L_DQUOTE] = ACTIONS(2644),
    [anon_sym_u_DQUOTE] = ACTIONS(2644),
    [anon_sym_U_DQUOTE] = ACTIONS(2644),
    [anon_sym_u8_DQUOTE] = ACTIONS(2644),
    [sym_true] = ACTIONS(2642),
    [sym_false] = ACTIONS(2642),
    [anon_sym_NULL] = ACTIONS(2642),
    [anon_sym_nullptr] = ACTIONS(2642),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2642),
    [anon_sym___typeof] = ACTIONS(2642),
    [anon_sym_typeof] = ACTIONS(2642),
    [anon_sym_ATimport] = ACTIONS(2644),
    [aux_sym_preproc_undef_token1] = ACTIONS(2642),
    [anon_sym_POUND] = ACTIONS(2642),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2642),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2642),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2642),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2642),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2642),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2642),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2642),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_API_AVAILABLE] = ACTIONS(2642),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_API_DEPRECATED] = ACTIONS(2642),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2642),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym___deprecated_msg] = ACTIONS(2642),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2642),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2642),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2642),
    [anon_sym_ATprotocol] = ACTIONS(2644),
    [anon_sym_ATinterface] = ACTIONS(2644),
    [anon_sym_ATimplementation] = ACTIONS(2644),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2644),
    [anon_sym_ATsynthesize] = ACTIONS(2644),
    [anon_sym_ATdynamic] = ACTIONS(2644),
    [anon_sym_ATtry] = ACTIONS(2644),
    [anon_sym___try] = ACTIONS(2642),
    [anon_sym_ATthrow] = ACTIONS(2644),
    [anon_sym_ATselector] = ACTIONS(2644),
    [anon_sym_ATavailable] = ACTIONS(2644),
    [anon_sym___builtin_available] = ACTIONS(2642),
    [anon_sym_va_arg] = ACTIONS(2642),
    [anon_sym_ATencode] = ACTIONS(2644),
    [anon_sym_ATsynchronized] = ACTIONS(2644),
    [anon_sym_BOOL] = ACTIONS(2642),
    [anon_sym_IMP] = ACTIONS(2642),
    [anon_sym_SEL] = ACTIONS(2642),
    [anon_sym_Class] = ACTIONS(2642),
    [anon_sym_id] = ACTIONS(2642),
  },
  [STATE(667)] = {
    [sym_identifier] = ACTIONS(2646),
    [aux_sym_preproc_include_token1] = ACTIONS(2646),
    [aux_sym_preproc_include_token2] = ACTIONS(2646),
    [aux_sym_preproc_def_token1] = ACTIONS(2646),
    [aux_sym_preproc_if_token1] = ACTIONS(2646),
    [aux_sym_preproc_if_token2] = ACTIONS(2646),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2646),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2646),
    [sym_preproc_directive] = ACTIONS(2646),
    [anon_sym_LPAREN2] = ACTIONS(2648),
    [anon_sym_BANG] = ACTIONS(2648),
    [anon_sym_TILDE] = ACTIONS(2648),
    [anon_sym_DASH] = ACTIONS(2646),
    [anon_sym_PLUS] = ACTIONS(2646),
    [anon_sym_STAR] = ACTIONS(2648),
    [anon_sym_CARET] = ACTIONS(2648),
    [anon_sym_AMP] = ACTIONS(2648),
    [anon_sym_SEMI] = ACTIONS(2648),
    [anon_sym___extension__] = ACTIONS(2646),
    [anon_sym_typedef] = ACTIONS(2646),
    [anon_sym_extern] = ACTIONS(2646),
    [anon_sym___attribute__] = ACTIONS(2646),
    [anon_sym___attribute] = ACTIONS(2646),
    [anon_sym_noreturn] = ACTIONS(2646),
    [anon_sym_LBRACK] = ACTIONS(2648),
    [anon_sym___declspec] = ACTIONS(2646),
    [anon_sym___cdecl] = ACTIONS(2646),
    [anon_sym___clrcall] = ACTIONS(2646),
    [anon_sym___stdcall] = ACTIONS(2646),
    [anon_sym___fastcall] = ACTIONS(2646),
    [anon_sym___thiscall] = ACTIONS(2646),
    [anon_sym___vectorcall] = ACTIONS(2646),
    [anon_sym_LBRACE] = ACTIONS(2648),
    [anon_sym_signed] = ACTIONS(2646),
    [anon_sym_unsigned] = ACTIONS(2646),
    [anon_sym_long] = ACTIONS(2646),
    [anon_sym_short] = ACTIONS(2646),
    [anon_sym_static] = ACTIONS(2646),
    [anon_sym_ATautoreleasepool] = ACTIONS(2648),
    [anon_sym_auto] = ACTIONS(2646),
    [anon_sym_register] = ACTIONS(2646),
    [anon_sym_inline] = ACTIONS(2646),
    [anon_sym___inline] = ACTIONS(2646),
    [anon_sym___inline__] = ACTIONS(2646),
    [anon_sym___forceinline] = ACTIONS(2646),
    [anon_sym_thread_local] = ACTIONS(2646),
    [anon_sym___thread] = ACTIONS(2646),
    [anon_sym_CG_EXTERN] = ACTIONS(2646),
    [anon_sym_CG_INLINE] = ACTIONS(2646),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2646),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2646),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2646),
    [anon_sym_IBOutlet] = ACTIONS(2646),
    [anon_sym_IBInspectable] = ACTIONS(2646),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2646),
    [anon_sym_NS_INLINE] = ACTIONS(2646),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2646),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2646),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2646),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2646),
    [anon_sym_const] = ACTIONS(2646),
    [anon_sym_constexpr] = ACTIONS(2646),
    [anon_sym_volatile] = ACTIONS(2646),
    [anon_sym_restrict] = ACTIONS(2646),
    [anon_sym___restrict__] = ACTIONS(2646),
    [anon_sym__Atomic] = ACTIONS(2646),
    [anon_sym__Noreturn] = ACTIONS(2646),
    [anon_sym__Nonnull] = ACTIONS(2646),
    [anon_sym_nullable] = ACTIONS(2646),
    [anon_sym__Complex] = ACTIONS(2646),
    [anon_sym__Nullable] = ACTIONS(2646),
    [anon_sym__Nullable_result] = ACTIONS(2646),
    [anon_sym__Null_unspecified] = ACTIONS(2646),
    [anon_sym___autoreleasing] = ACTIONS(2646),
    [anon_sym___block] = ACTIONS(2646),
    [anon_sym___bridge] = ACTIONS(2646),
    [anon_sym___bridge_retained] = ACTIONS(2646),
    [anon_sym___bridge_transfer] = ACTIONS(2646),
    [anon_sym___complex] = ACTIONS(2646),
    [anon_sym___const] = ACTIONS(2646),
    [anon_sym___imag] = ACTIONS(2646),
    [anon_sym___kindof] = ACTIONS(2646),
    [anon_sym___nonnull] = ACTIONS(2646),
    [anon_sym___nullable] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2646),
    [anon_sym___real] = ACTIONS(2646),
    [anon_sym___strong] = ACTIONS(2646),
    [anon_sym___unsafe_unretained] = ACTIONS(2646),
    [anon_sym___unused] = ACTIONS(2646),
    [anon_sym___weak] = ACTIONS(2646),
    [anon_sym_alignas] = ACTIONS(2646),
    [anon_sym__Alignas] = ACTIONS(2646),
    [sym_primitive_type] = ACTIONS(2646),
    [anon_sym_enum] = ACTIONS(2646),
    [anon_sym_struct] = ACTIONS(2646),
    [anon_sym_union] = ACTIONS(2646),
    [anon_sym_if] = ACTIONS(2646),
    [anon_sym_switch] = ACTIONS(2646),
    [anon_sym_case] = ACTIONS(2646),
    [anon_sym_default] = ACTIONS(2646),
    [anon_sym_while] = ACTIONS(2646),
    [anon_sym_do] = ACTIONS(2646),
    [anon_sym_for] = ACTIONS(2646),
    [anon_sym_in] = ACTIONS(2646),
    [anon_sym_return] = ACTIONS(2646),
    [anon_sym_break] = ACTIONS(2646),
    [anon_sym_continue] = ACTIONS(2646),
    [anon_sym_goto] = ACTIONS(2646),
    [anon_sym_DASH_DASH] = ACTIONS(2648),
    [anon_sym_PLUS_PLUS] = ACTIONS(2648),
    [anon_sym_sizeof] = ACTIONS(2646),
    [anon_sym___alignof__] = ACTIONS(2646),
    [anon_sym___alignof] = ACTIONS(2646),
    [anon_sym__alignof] = ACTIONS(2646),
    [anon_sym_alignof] = ACTIONS(2646),
    [anon_sym__Alignof] = ACTIONS(2646),
    [anon_sym_offsetof] = ACTIONS(2646),
    [anon_sym__Generic] = ACTIONS(2646),
    [anon_sym_asm] = ACTIONS(2646),
    [anon_sym___asm__] = ACTIONS(2646),
    [anon_sym___asm] = ACTIONS(2646),
    [sym_number_literal] = ACTIONS(2648),
    [anon_sym_L_SQUOTE] = ACTIONS(2648),
    [anon_sym_u_SQUOTE] = ACTIONS(2648),
    [anon_sym_U_SQUOTE] = ACTIONS(2648),
    [anon_sym_u8_SQUOTE] = ACTIONS(2648),
    [anon_sym_SQUOTE] = ACTIONS(2648),
    [anon_sym_AT] = ACTIONS(2646),
    [anon_sym_DQUOTE] = ACTIONS(2648),
    [anon_sym_L_DQUOTE] = ACTIONS(2648),
    [anon_sym_u_DQUOTE] = ACTIONS(2648),
    [anon_sym_U_DQUOTE] = ACTIONS(2648),
    [anon_sym_u8_DQUOTE] = ACTIONS(2648),
    [sym_true] = ACTIONS(2646),
    [sym_false] = ACTIONS(2646),
    [anon_sym_NULL] = ACTIONS(2646),
    [anon_sym_nullptr] = ACTIONS(2646),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2646),
    [anon_sym___typeof] = ACTIONS(2646),
    [anon_sym_typeof] = ACTIONS(2646),
    [anon_sym_ATimport] = ACTIONS(2648),
    [aux_sym_preproc_undef_token1] = ACTIONS(2646),
    [anon_sym_POUND] = ACTIONS(2646),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2646),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2646),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2646),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2646),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2646),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2646),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2646),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_API_AVAILABLE] = ACTIONS(2646),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_API_DEPRECATED] = ACTIONS(2646),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2646),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym___deprecated_msg] = ACTIONS(2646),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2646),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2646),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2646),
    [anon_sym_ATprotocol] = ACTIONS(2648),
    [anon_sym_ATinterface] = ACTIONS(2648),
    [anon_sym_ATimplementation] = ACTIONS(2648),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2648),
    [anon_sym_ATsynthesize] = ACTIONS(2648),
    [anon_sym_ATdynamic] = ACTIONS(2648),
    [anon_sym_ATtry] = ACTIONS(2648),
    [anon_sym___try] = ACTIONS(2646),
    [anon_sym_ATthrow] = ACTIONS(2648),
    [anon_sym_ATselector] = ACTIONS(2648),
    [anon_sym_ATavailable] = ACTIONS(2648),
    [anon_sym___builtin_available] = ACTIONS(2646),
    [anon_sym_va_arg] = ACTIONS(2646),
    [anon_sym_ATencode] = ACTIONS(2648),
    [anon_sym_ATsynchronized] = ACTIONS(2648),
    [anon_sym_BOOL] = ACTIONS(2646),
    [anon_sym_IMP] = ACTIONS(2646),
    [anon_sym_SEL] = ACTIONS(2646),
    [anon_sym_Class] = ACTIONS(2646),
    [anon_sym_id] = ACTIONS(2646),
  },
  [STATE(668)] = {
    [sym_identifier] = ACTIONS(2654),
    [aux_sym_preproc_include_token1] = ACTIONS(2654),
    [aux_sym_preproc_include_token2] = ACTIONS(2654),
    [aux_sym_preproc_def_token1] = ACTIONS(2654),
    [aux_sym_preproc_if_token1] = ACTIONS(2654),
    [aux_sym_preproc_if_token2] = ACTIONS(2654),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2654),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2654),
    [sym_preproc_directive] = ACTIONS(2654),
    [anon_sym_LPAREN2] = ACTIONS(2656),
    [anon_sym_BANG] = ACTIONS(2656),
    [anon_sym_TILDE] = ACTIONS(2656),
    [anon_sym_DASH] = ACTIONS(2654),
    [anon_sym_PLUS] = ACTIONS(2654),
    [anon_sym_STAR] = ACTIONS(2656),
    [anon_sym_CARET] = ACTIONS(2656),
    [anon_sym_AMP] = ACTIONS(2656),
    [anon_sym_SEMI] = ACTIONS(2656),
    [anon_sym___extension__] = ACTIONS(2654),
    [anon_sym_typedef] = ACTIONS(2654),
    [anon_sym_extern] = ACTIONS(2654),
    [anon_sym___attribute__] = ACTIONS(2654),
    [anon_sym___attribute] = ACTIONS(2654),
    [anon_sym_noreturn] = ACTIONS(2654),
    [anon_sym_LBRACK] = ACTIONS(2656),
    [anon_sym___declspec] = ACTIONS(2654),
    [anon_sym___cdecl] = ACTIONS(2654),
    [anon_sym___clrcall] = ACTIONS(2654),
    [anon_sym___stdcall] = ACTIONS(2654),
    [anon_sym___fastcall] = ACTIONS(2654),
    [anon_sym___thiscall] = ACTIONS(2654),
    [anon_sym___vectorcall] = ACTIONS(2654),
    [anon_sym_LBRACE] = ACTIONS(2656),
    [anon_sym_signed] = ACTIONS(2654),
    [anon_sym_unsigned] = ACTIONS(2654),
    [anon_sym_long] = ACTIONS(2654),
    [anon_sym_short] = ACTIONS(2654),
    [anon_sym_static] = ACTIONS(2654),
    [anon_sym_ATautoreleasepool] = ACTIONS(2656),
    [anon_sym_auto] = ACTIONS(2654),
    [anon_sym_register] = ACTIONS(2654),
    [anon_sym_inline] = ACTIONS(2654),
    [anon_sym___inline] = ACTIONS(2654),
    [anon_sym___inline__] = ACTIONS(2654),
    [anon_sym___forceinline] = ACTIONS(2654),
    [anon_sym_thread_local] = ACTIONS(2654),
    [anon_sym___thread] = ACTIONS(2654),
    [anon_sym_CG_EXTERN] = ACTIONS(2654),
    [anon_sym_CG_INLINE] = ACTIONS(2654),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2654),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2654),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2654),
    [anon_sym_IBOutlet] = ACTIONS(2654),
    [anon_sym_IBInspectable] = ACTIONS(2654),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2654),
    [anon_sym_NS_INLINE] = ACTIONS(2654),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2654),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2654),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2654),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2654),
    [anon_sym_const] = ACTIONS(2654),
    [anon_sym_constexpr] = ACTIONS(2654),
    [anon_sym_volatile] = ACTIONS(2654),
    [anon_sym_restrict] = ACTIONS(2654),
    [anon_sym___restrict__] = ACTIONS(2654),
    [anon_sym__Atomic] = ACTIONS(2654),
    [anon_sym__Noreturn] = ACTIONS(2654),
    [anon_sym__Nonnull] = ACTIONS(2654),
    [anon_sym_nullable] = ACTIONS(2654),
    [anon_sym__Complex] = ACTIONS(2654),
    [anon_sym__Nullable] = ACTIONS(2654),
    [anon_sym__Nullable_result] = ACTIONS(2654),
    [anon_sym__Null_unspecified] = ACTIONS(2654),
    [anon_sym___autoreleasing] = ACTIONS(2654),
    [anon_sym___block] = ACTIONS(2654),
    [anon_sym___bridge] = ACTIONS(2654),
    [anon_sym___bridge_retained] = ACTIONS(2654),
    [anon_sym___bridge_transfer] = ACTIONS(2654),
    [anon_sym___complex] = ACTIONS(2654),
    [anon_sym___const] = ACTIONS(2654),
    [anon_sym___imag] = ACTIONS(2654),
    [anon_sym___kindof] = ACTIONS(2654),
    [anon_sym___nonnull] = ACTIONS(2654),
    [anon_sym___nullable] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2654),
    [anon_sym___real] = ACTIONS(2654),
    [anon_sym___strong] = ACTIONS(2654),
    [anon_sym___unsafe_unretained] = ACTIONS(2654),
    [anon_sym___unused] = ACTIONS(2654),
    [anon_sym___weak] = ACTIONS(2654),
    [anon_sym_alignas] = ACTIONS(2654),
    [anon_sym__Alignas] = ACTIONS(2654),
    [sym_primitive_type] = ACTIONS(2654),
    [anon_sym_enum] = ACTIONS(2654),
    [anon_sym_struct] = ACTIONS(2654),
    [anon_sym_union] = ACTIONS(2654),
    [anon_sym_if] = ACTIONS(2654),
    [anon_sym_switch] = ACTIONS(2654),
    [anon_sym_case] = ACTIONS(2654),
    [anon_sym_default] = ACTIONS(2654),
    [anon_sym_while] = ACTIONS(2654),
    [anon_sym_do] = ACTIONS(2654),
    [anon_sym_for] = ACTIONS(2654),
    [anon_sym_in] = ACTIONS(2654),
    [anon_sym_return] = ACTIONS(2654),
    [anon_sym_break] = ACTIONS(2654),
    [anon_sym_continue] = ACTIONS(2654),
    [anon_sym_goto] = ACTIONS(2654),
    [anon_sym_DASH_DASH] = ACTIONS(2656),
    [anon_sym_PLUS_PLUS] = ACTIONS(2656),
    [anon_sym_sizeof] = ACTIONS(2654),
    [anon_sym___alignof__] = ACTIONS(2654),
    [anon_sym___alignof] = ACTIONS(2654),
    [anon_sym__alignof] = ACTIONS(2654),
    [anon_sym_alignof] = ACTIONS(2654),
    [anon_sym__Alignof] = ACTIONS(2654),
    [anon_sym_offsetof] = ACTIONS(2654),
    [anon_sym__Generic] = ACTIONS(2654),
    [anon_sym_asm] = ACTIONS(2654),
    [anon_sym___asm__] = ACTIONS(2654),
    [anon_sym___asm] = ACTIONS(2654),
    [sym_number_literal] = ACTIONS(2656),
    [anon_sym_L_SQUOTE] = ACTIONS(2656),
    [anon_sym_u_SQUOTE] = ACTIONS(2656),
    [anon_sym_U_SQUOTE] = ACTIONS(2656),
    [anon_sym_u8_SQUOTE] = ACTIONS(2656),
    [anon_sym_SQUOTE] = ACTIONS(2656),
    [anon_sym_AT] = ACTIONS(2654),
    [anon_sym_DQUOTE] = ACTIONS(2656),
    [anon_sym_L_DQUOTE] = ACTIONS(2656),
    [anon_sym_u_DQUOTE] = ACTIONS(2656),
    [anon_sym_U_DQUOTE] = ACTIONS(2656),
    [anon_sym_u8_DQUOTE] = ACTIONS(2656),
    [sym_true] = ACTIONS(2654),
    [sym_false] = ACTIONS(2654),
    [anon_sym_NULL] = ACTIONS(2654),
    [anon_sym_nullptr] = ACTIONS(2654),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2654),
    [anon_sym___typeof] = ACTIONS(2654),
    [anon_sym_typeof] = ACTIONS(2654),
    [anon_sym_ATimport] = ACTIONS(2656),
    [aux_sym_preproc_undef_token1] = ACTIONS(2654),
    [anon_sym_POUND] = ACTIONS(2654),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2654),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2654),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2654),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2654),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2654),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2654),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2654),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_API_AVAILABLE] = ACTIONS(2654),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_API_DEPRECATED] = ACTIONS(2654),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2654),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym___deprecated_msg] = ACTIONS(2654),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2654),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2654),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2654),
    [anon_sym_ATprotocol] = ACTIONS(2656),
    [anon_sym_ATinterface] = ACTIONS(2656),
    [anon_sym_ATimplementation] = ACTIONS(2656),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2656),
    [anon_sym_ATsynthesize] = ACTIONS(2656),
    [anon_sym_ATdynamic] = ACTIONS(2656),
    [anon_sym_ATtry] = ACTIONS(2656),
    [anon_sym___try] = ACTIONS(2654),
    [anon_sym_ATthrow] = ACTIONS(2656),
    [anon_sym_ATselector] = ACTIONS(2656),
    [anon_sym_ATavailable] = ACTIONS(2656),
    [anon_sym___builtin_available] = ACTIONS(2654),
    [anon_sym_va_arg] = ACTIONS(2654),
    [anon_sym_ATencode] = ACTIONS(2656),
    [anon_sym_ATsynchronized] = ACTIONS(2656),
    [anon_sym_BOOL] = ACTIONS(2654),
    [anon_sym_IMP] = ACTIONS(2654),
    [anon_sym_SEL] = ACTIONS(2654),
    [anon_sym_Class] = ACTIONS(2654),
    [anon_sym_id] = ACTIONS(2654),
  },
  [STATE(669)] = {
    [sym_identifier] = ACTIONS(2672),
    [aux_sym_preproc_include_token1] = ACTIONS(2672),
    [aux_sym_preproc_include_token2] = ACTIONS(2672),
    [aux_sym_preproc_def_token1] = ACTIONS(2672),
    [aux_sym_preproc_if_token1] = ACTIONS(2672),
    [aux_sym_preproc_if_token2] = ACTIONS(2672),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2672),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2672),
    [sym_preproc_directive] = ACTIONS(2672),
    [anon_sym_LPAREN2] = ACTIONS(2674),
    [anon_sym_BANG] = ACTIONS(2674),
    [anon_sym_TILDE] = ACTIONS(2674),
    [anon_sym_DASH] = ACTIONS(2672),
    [anon_sym_PLUS] = ACTIONS(2672),
    [anon_sym_STAR] = ACTIONS(2674),
    [anon_sym_CARET] = ACTIONS(2674),
    [anon_sym_AMP] = ACTIONS(2674),
    [anon_sym_SEMI] = ACTIONS(2674),
    [anon_sym___extension__] = ACTIONS(2672),
    [anon_sym_typedef] = ACTIONS(2672),
    [anon_sym_extern] = ACTIONS(2672),
    [anon_sym___attribute__] = ACTIONS(2672),
    [anon_sym___attribute] = ACTIONS(2672),
    [anon_sym_noreturn] = ACTIONS(2672),
    [anon_sym_LBRACK] = ACTIONS(2674),
    [anon_sym___declspec] = ACTIONS(2672),
    [anon_sym___cdecl] = ACTIONS(2672),
    [anon_sym___clrcall] = ACTIONS(2672),
    [anon_sym___stdcall] = ACTIONS(2672),
    [anon_sym___fastcall] = ACTIONS(2672),
    [anon_sym___thiscall] = ACTIONS(2672),
    [anon_sym___vectorcall] = ACTIONS(2672),
    [anon_sym_LBRACE] = ACTIONS(2674),
    [anon_sym_signed] = ACTIONS(2672),
    [anon_sym_unsigned] = ACTIONS(2672),
    [anon_sym_long] = ACTIONS(2672),
    [anon_sym_short] = ACTIONS(2672),
    [anon_sym_static] = ACTIONS(2672),
    [anon_sym_ATautoreleasepool] = ACTIONS(2674),
    [anon_sym_auto] = ACTIONS(2672),
    [anon_sym_register] = ACTIONS(2672),
    [anon_sym_inline] = ACTIONS(2672),
    [anon_sym___inline] = ACTIONS(2672),
    [anon_sym___inline__] = ACTIONS(2672),
    [anon_sym___forceinline] = ACTIONS(2672),
    [anon_sym_thread_local] = ACTIONS(2672),
    [anon_sym___thread] = ACTIONS(2672),
    [anon_sym_CG_EXTERN] = ACTIONS(2672),
    [anon_sym_CG_INLINE] = ACTIONS(2672),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2672),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2672),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2672),
    [anon_sym_IBOutlet] = ACTIONS(2672),
    [anon_sym_IBInspectable] = ACTIONS(2672),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2672),
    [anon_sym_NS_INLINE] = ACTIONS(2672),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2672),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2672),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2672),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2672),
    [anon_sym_const] = ACTIONS(2672),
    [anon_sym_constexpr] = ACTIONS(2672),
    [anon_sym_volatile] = ACTIONS(2672),
    [anon_sym_restrict] = ACTIONS(2672),
    [anon_sym___restrict__] = ACTIONS(2672),
    [anon_sym__Atomic] = ACTIONS(2672),
    [anon_sym__Noreturn] = ACTIONS(2672),
    [anon_sym__Nonnull] = ACTIONS(2672),
    [anon_sym_nullable] = ACTIONS(2672),
    [anon_sym__Complex] = ACTIONS(2672),
    [anon_sym__Nullable] = ACTIONS(2672),
    [anon_sym__Nullable_result] = ACTIONS(2672),
    [anon_sym__Null_unspecified] = ACTIONS(2672),
    [anon_sym___autoreleasing] = ACTIONS(2672),
    [anon_sym___block] = ACTIONS(2672),
    [anon_sym___bridge] = ACTIONS(2672),
    [anon_sym___bridge_retained] = ACTIONS(2672),
    [anon_sym___bridge_transfer] = ACTIONS(2672),
    [anon_sym___complex] = ACTIONS(2672),
    [anon_sym___const] = ACTIONS(2672),
    [anon_sym___imag] = ACTIONS(2672),
    [anon_sym___kindof] = ACTIONS(2672),
    [anon_sym___nonnull] = ACTIONS(2672),
    [anon_sym___nullable] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2672),
    [anon_sym___real] = ACTIONS(2672),
    [anon_sym___strong] = ACTIONS(2672),
    [anon_sym___unsafe_unretained] = ACTIONS(2672),
    [anon_sym___unused] = ACTIONS(2672),
    [anon_sym___weak] = ACTIONS(2672),
    [anon_sym_alignas] = ACTIONS(2672),
    [anon_sym__Alignas] = ACTIONS(2672),
    [sym_primitive_type] = ACTIONS(2672),
    [anon_sym_enum] = ACTIONS(2672),
    [anon_sym_struct] = ACTIONS(2672),
    [anon_sym_union] = ACTIONS(2672),
    [anon_sym_if] = ACTIONS(2672),
    [anon_sym_switch] = ACTIONS(2672),
    [anon_sym_case] = ACTIONS(2672),
    [anon_sym_default] = ACTIONS(2672),
    [anon_sym_while] = ACTIONS(2672),
    [anon_sym_do] = ACTIONS(2672),
    [anon_sym_for] = ACTIONS(2672),
    [anon_sym_in] = ACTIONS(2672),
    [anon_sym_return] = ACTIONS(2672),
    [anon_sym_break] = ACTIONS(2672),
    [anon_sym_continue] = ACTIONS(2672),
    [anon_sym_goto] = ACTIONS(2672),
    [anon_sym_DASH_DASH] = ACTIONS(2674),
    [anon_sym_PLUS_PLUS] = ACTIONS(2674),
    [anon_sym_sizeof] = ACTIONS(2672),
    [anon_sym___alignof__] = ACTIONS(2672),
    [anon_sym___alignof] = ACTIONS(2672),
    [anon_sym__alignof] = ACTIONS(2672),
    [anon_sym_alignof] = ACTIONS(2672),
    [anon_sym__Alignof] = ACTIONS(2672),
    [anon_sym_offsetof] = ACTIONS(2672),
    [anon_sym__Generic] = ACTIONS(2672),
    [anon_sym_asm] = ACTIONS(2672),
    [anon_sym___asm__] = ACTIONS(2672),
    [anon_sym___asm] = ACTIONS(2672),
    [sym_number_literal] = ACTIONS(2674),
    [anon_sym_L_SQUOTE] = ACTIONS(2674),
    [anon_sym_u_SQUOTE] = ACTIONS(2674),
    [anon_sym_U_SQUOTE] = ACTIONS(2674),
    [anon_sym_u8_SQUOTE] = ACTIONS(2674),
    [anon_sym_SQUOTE] = ACTIONS(2674),
    [anon_sym_AT] = ACTIONS(2672),
    [anon_sym_DQUOTE] = ACTIONS(2674),
    [anon_sym_L_DQUOTE] = ACTIONS(2674),
    [anon_sym_u_DQUOTE] = ACTIONS(2674),
    [anon_sym_U_DQUOTE] = ACTIONS(2674),
    [anon_sym_u8_DQUOTE] = ACTIONS(2674),
    [sym_true] = ACTIONS(2672),
    [sym_false] = ACTIONS(2672),
    [anon_sym_NULL] = ACTIONS(2672),
    [anon_sym_nullptr] = ACTIONS(2672),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2672),
    [anon_sym___typeof] = ACTIONS(2672),
    [anon_sym_typeof] = ACTIONS(2672),
    [anon_sym_ATimport] = ACTIONS(2674),
    [aux_sym_preproc_undef_token1] = ACTIONS(2672),
    [anon_sym_POUND] = ACTIONS(2672),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2672),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2672),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2672),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2672),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2672),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2672),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2672),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_API_AVAILABLE] = ACTIONS(2672),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_API_DEPRECATED] = ACTIONS(2672),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2672),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym___deprecated_msg] = ACTIONS(2672),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2672),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2672),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2672),
    [anon_sym_ATprotocol] = ACTIONS(2674),
    [anon_sym_ATinterface] = ACTIONS(2674),
    [anon_sym_ATimplementation] = ACTIONS(2674),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2674),
    [anon_sym_ATsynthesize] = ACTIONS(2674),
    [anon_sym_ATdynamic] = ACTIONS(2674),
    [anon_sym_ATtry] = ACTIONS(2674),
    [anon_sym___try] = ACTIONS(2672),
    [anon_sym_ATthrow] = ACTIONS(2674),
    [anon_sym_ATselector] = ACTIONS(2674),
    [anon_sym_ATavailable] = ACTIONS(2674),
    [anon_sym___builtin_available] = ACTIONS(2672),
    [anon_sym_va_arg] = ACTIONS(2672),
    [anon_sym_ATencode] = ACTIONS(2674),
    [anon_sym_ATsynchronized] = ACTIONS(2674),
    [anon_sym_BOOL] = ACTIONS(2672),
    [anon_sym_IMP] = ACTIONS(2672),
    [anon_sym_SEL] = ACTIONS(2672),
    [anon_sym_Class] = ACTIONS(2672),
    [anon_sym_id] = ACTIONS(2672),
  },
  [STATE(670)] = {
    [sym_identifier] = ACTIONS(2676),
    [aux_sym_preproc_include_token1] = ACTIONS(2676),
    [aux_sym_preproc_include_token2] = ACTIONS(2676),
    [aux_sym_preproc_def_token1] = ACTIONS(2676),
    [aux_sym_preproc_if_token1] = ACTIONS(2676),
    [aux_sym_preproc_if_token2] = ACTIONS(2676),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2676),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2676),
    [sym_preproc_directive] = ACTIONS(2676),
    [anon_sym_LPAREN2] = ACTIONS(2678),
    [anon_sym_BANG] = ACTIONS(2678),
    [anon_sym_TILDE] = ACTIONS(2678),
    [anon_sym_DASH] = ACTIONS(2676),
    [anon_sym_PLUS] = ACTIONS(2676),
    [anon_sym_STAR] = ACTIONS(2678),
    [anon_sym_CARET] = ACTIONS(2678),
    [anon_sym_AMP] = ACTIONS(2678),
    [anon_sym_SEMI] = ACTIONS(2678),
    [anon_sym___extension__] = ACTIONS(2676),
    [anon_sym_typedef] = ACTIONS(2676),
    [anon_sym_extern] = ACTIONS(2676),
    [anon_sym___attribute__] = ACTIONS(2676),
    [anon_sym___attribute] = ACTIONS(2676),
    [anon_sym_noreturn] = ACTIONS(2676),
    [anon_sym_LBRACK] = ACTIONS(2678),
    [anon_sym___declspec] = ACTIONS(2676),
    [anon_sym___cdecl] = ACTIONS(2676),
    [anon_sym___clrcall] = ACTIONS(2676),
    [anon_sym___stdcall] = ACTIONS(2676),
    [anon_sym___fastcall] = ACTIONS(2676),
    [anon_sym___thiscall] = ACTIONS(2676),
    [anon_sym___vectorcall] = ACTIONS(2676),
    [anon_sym_LBRACE] = ACTIONS(2678),
    [anon_sym_signed] = ACTIONS(2676),
    [anon_sym_unsigned] = ACTIONS(2676),
    [anon_sym_long] = ACTIONS(2676),
    [anon_sym_short] = ACTIONS(2676),
    [anon_sym_static] = ACTIONS(2676),
    [anon_sym_ATautoreleasepool] = ACTIONS(2678),
    [anon_sym_auto] = ACTIONS(2676),
    [anon_sym_register] = ACTIONS(2676),
    [anon_sym_inline] = ACTIONS(2676),
    [anon_sym___inline] = ACTIONS(2676),
    [anon_sym___inline__] = ACTIONS(2676),
    [anon_sym___forceinline] = ACTIONS(2676),
    [anon_sym_thread_local] = ACTIONS(2676),
    [anon_sym___thread] = ACTIONS(2676),
    [anon_sym_CG_EXTERN] = ACTIONS(2676),
    [anon_sym_CG_INLINE] = ACTIONS(2676),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2676),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2676),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2676),
    [anon_sym_IBOutlet] = ACTIONS(2676),
    [anon_sym_IBInspectable] = ACTIONS(2676),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2676),
    [anon_sym_NS_INLINE] = ACTIONS(2676),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2676),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2676),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2676),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2676),
    [anon_sym_const] = ACTIONS(2676),
    [anon_sym_constexpr] = ACTIONS(2676),
    [anon_sym_volatile] = ACTIONS(2676),
    [anon_sym_restrict] = ACTIONS(2676),
    [anon_sym___restrict__] = ACTIONS(2676),
    [anon_sym__Atomic] = ACTIONS(2676),
    [anon_sym__Noreturn] = ACTIONS(2676),
    [anon_sym__Nonnull] = ACTIONS(2676),
    [anon_sym_nullable] = ACTIONS(2676),
    [anon_sym__Complex] = ACTIONS(2676),
    [anon_sym__Nullable] = ACTIONS(2676),
    [anon_sym__Nullable_result] = ACTIONS(2676),
    [anon_sym__Null_unspecified] = ACTIONS(2676),
    [anon_sym___autoreleasing] = ACTIONS(2676),
    [anon_sym___block] = ACTIONS(2676),
    [anon_sym___bridge] = ACTIONS(2676),
    [anon_sym___bridge_retained] = ACTIONS(2676),
    [anon_sym___bridge_transfer] = ACTIONS(2676),
    [anon_sym___complex] = ACTIONS(2676),
    [anon_sym___const] = ACTIONS(2676),
    [anon_sym___imag] = ACTIONS(2676),
    [anon_sym___kindof] = ACTIONS(2676),
    [anon_sym___nonnull] = ACTIONS(2676),
    [anon_sym___nullable] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2676),
    [anon_sym___real] = ACTIONS(2676),
    [anon_sym___strong] = ACTIONS(2676),
    [anon_sym___unsafe_unretained] = ACTIONS(2676),
    [anon_sym___unused] = ACTIONS(2676),
    [anon_sym___weak] = ACTIONS(2676),
    [anon_sym_alignas] = ACTIONS(2676),
    [anon_sym__Alignas] = ACTIONS(2676),
    [sym_primitive_type] = ACTIONS(2676),
    [anon_sym_enum] = ACTIONS(2676),
    [anon_sym_struct] = ACTIONS(2676),
    [anon_sym_union] = ACTIONS(2676),
    [anon_sym_if] = ACTIONS(2676),
    [anon_sym_switch] = ACTIONS(2676),
    [anon_sym_case] = ACTIONS(2676),
    [anon_sym_default] = ACTIONS(2676),
    [anon_sym_while] = ACTIONS(2676),
    [anon_sym_do] = ACTIONS(2676),
    [anon_sym_for] = ACTIONS(2676),
    [anon_sym_in] = ACTIONS(2676),
    [anon_sym_return] = ACTIONS(2676),
    [anon_sym_break] = ACTIONS(2676),
    [anon_sym_continue] = ACTIONS(2676),
    [anon_sym_goto] = ACTIONS(2676),
    [anon_sym_DASH_DASH] = ACTIONS(2678),
    [anon_sym_PLUS_PLUS] = ACTIONS(2678),
    [anon_sym_sizeof] = ACTIONS(2676),
    [anon_sym___alignof__] = ACTIONS(2676),
    [anon_sym___alignof] = ACTIONS(2676),
    [anon_sym__alignof] = ACTIONS(2676),
    [anon_sym_alignof] = ACTIONS(2676),
    [anon_sym__Alignof] = ACTIONS(2676),
    [anon_sym_offsetof] = ACTIONS(2676),
    [anon_sym__Generic] = ACTIONS(2676),
    [anon_sym_asm] = ACTIONS(2676),
    [anon_sym___asm__] = ACTIONS(2676),
    [anon_sym___asm] = ACTIONS(2676),
    [sym_number_literal] = ACTIONS(2678),
    [anon_sym_L_SQUOTE] = ACTIONS(2678),
    [anon_sym_u_SQUOTE] = ACTIONS(2678),
    [anon_sym_U_SQUOTE] = ACTIONS(2678),
    [anon_sym_u8_SQUOTE] = ACTIONS(2678),
    [anon_sym_SQUOTE] = ACTIONS(2678),
    [anon_sym_AT] = ACTIONS(2676),
    [anon_sym_DQUOTE] = ACTIONS(2678),
    [anon_sym_L_DQUOTE] = ACTIONS(2678),
    [anon_sym_u_DQUOTE] = ACTIONS(2678),
    [anon_sym_U_DQUOTE] = ACTIONS(2678),
    [anon_sym_u8_DQUOTE] = ACTIONS(2678),
    [sym_true] = ACTIONS(2676),
    [sym_false] = ACTIONS(2676),
    [anon_sym_NULL] = ACTIONS(2676),
    [anon_sym_nullptr] = ACTIONS(2676),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2676),
    [anon_sym___typeof] = ACTIONS(2676),
    [anon_sym_typeof] = ACTIONS(2676),
    [anon_sym_ATimport] = ACTIONS(2678),
    [aux_sym_preproc_undef_token1] = ACTIONS(2676),
    [anon_sym_POUND] = ACTIONS(2676),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2676),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2676),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2676),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2676),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2676),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2676),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2676),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_API_AVAILABLE] = ACTIONS(2676),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_API_DEPRECATED] = ACTIONS(2676),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2676),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym___deprecated_msg] = ACTIONS(2676),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2676),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2676),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2676),
    [anon_sym_ATprotocol] = ACTIONS(2678),
    [anon_sym_ATinterface] = ACTIONS(2678),
    [anon_sym_ATimplementation] = ACTIONS(2678),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2678),
    [anon_sym_ATsynthesize] = ACTIONS(2678),
    [anon_sym_ATdynamic] = ACTIONS(2678),
    [anon_sym_ATtry] = ACTIONS(2678),
    [anon_sym___try] = ACTIONS(2676),
    [anon_sym_ATthrow] = ACTIONS(2678),
    [anon_sym_ATselector] = ACTIONS(2678),
    [anon_sym_ATavailable] = ACTIONS(2678),
    [anon_sym___builtin_available] = ACTIONS(2676),
    [anon_sym_va_arg] = ACTIONS(2676),
    [anon_sym_ATencode] = ACTIONS(2678),
    [anon_sym_ATsynchronized] = ACTIONS(2678),
    [anon_sym_BOOL] = ACTIONS(2676),
    [anon_sym_IMP] = ACTIONS(2676),
    [anon_sym_SEL] = ACTIONS(2676),
    [anon_sym_Class] = ACTIONS(2676),
    [anon_sym_id] = ACTIONS(2676),
  },
  [STATE(671)] = {
    [sym_identifier] = ACTIONS(2680),
    [aux_sym_preproc_include_token1] = ACTIONS(2680),
    [aux_sym_preproc_include_token2] = ACTIONS(2680),
    [aux_sym_preproc_def_token1] = ACTIONS(2680),
    [aux_sym_preproc_if_token1] = ACTIONS(2680),
    [aux_sym_preproc_if_token2] = ACTIONS(2680),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2680),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2680),
    [sym_preproc_directive] = ACTIONS(2680),
    [anon_sym_LPAREN2] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2680),
    [anon_sym_PLUS] = ACTIONS(2680),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym___extension__] = ACTIONS(2680),
    [anon_sym_typedef] = ACTIONS(2680),
    [anon_sym_extern] = ACTIONS(2680),
    [anon_sym___attribute__] = ACTIONS(2680),
    [anon_sym___attribute] = ACTIONS(2680),
    [anon_sym_noreturn] = ACTIONS(2680),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym___declspec] = ACTIONS(2680),
    [anon_sym___cdecl] = ACTIONS(2680),
    [anon_sym___clrcall] = ACTIONS(2680),
    [anon_sym___stdcall] = ACTIONS(2680),
    [anon_sym___fastcall] = ACTIONS(2680),
    [anon_sym___thiscall] = ACTIONS(2680),
    [anon_sym___vectorcall] = ACTIONS(2680),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_signed] = ACTIONS(2680),
    [anon_sym_unsigned] = ACTIONS(2680),
    [anon_sym_long] = ACTIONS(2680),
    [anon_sym_short] = ACTIONS(2680),
    [anon_sym_static] = ACTIONS(2680),
    [anon_sym_ATautoreleasepool] = ACTIONS(2682),
    [anon_sym_auto] = ACTIONS(2680),
    [anon_sym_register] = ACTIONS(2680),
    [anon_sym_inline] = ACTIONS(2680),
    [anon_sym___inline] = ACTIONS(2680),
    [anon_sym___inline__] = ACTIONS(2680),
    [anon_sym___forceinline] = ACTIONS(2680),
    [anon_sym_thread_local] = ACTIONS(2680),
    [anon_sym___thread] = ACTIONS(2680),
    [anon_sym_CG_EXTERN] = ACTIONS(2680),
    [anon_sym_CG_INLINE] = ACTIONS(2680),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2680),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2680),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2680),
    [anon_sym_IBOutlet] = ACTIONS(2680),
    [anon_sym_IBInspectable] = ACTIONS(2680),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2680),
    [anon_sym_NS_INLINE] = ACTIONS(2680),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2680),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2680),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2680),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2680),
    [anon_sym_const] = ACTIONS(2680),
    [anon_sym_constexpr] = ACTIONS(2680),
    [anon_sym_volatile] = ACTIONS(2680),
    [anon_sym_restrict] = ACTIONS(2680),
    [anon_sym___restrict__] = ACTIONS(2680),
    [anon_sym__Atomic] = ACTIONS(2680),
    [anon_sym__Noreturn] = ACTIONS(2680),
    [anon_sym__Nonnull] = ACTIONS(2680),
    [anon_sym_nullable] = ACTIONS(2680),
    [anon_sym__Complex] = ACTIONS(2680),
    [anon_sym__Nullable] = ACTIONS(2680),
    [anon_sym__Nullable_result] = ACTIONS(2680),
    [anon_sym__Null_unspecified] = ACTIONS(2680),
    [anon_sym___autoreleasing] = ACTIONS(2680),
    [anon_sym___block] = ACTIONS(2680),
    [anon_sym___bridge] = ACTIONS(2680),
    [anon_sym___bridge_retained] = ACTIONS(2680),
    [anon_sym___bridge_transfer] = ACTIONS(2680),
    [anon_sym___complex] = ACTIONS(2680),
    [anon_sym___const] = ACTIONS(2680),
    [anon_sym___imag] = ACTIONS(2680),
    [anon_sym___kindof] = ACTIONS(2680),
    [anon_sym___nonnull] = ACTIONS(2680),
    [anon_sym___nullable] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2680),
    [anon_sym___real] = ACTIONS(2680),
    [anon_sym___strong] = ACTIONS(2680),
    [anon_sym___unsafe_unretained] = ACTIONS(2680),
    [anon_sym___unused] = ACTIONS(2680),
    [anon_sym___weak] = ACTIONS(2680),
    [anon_sym_alignas] = ACTIONS(2680),
    [anon_sym__Alignas] = ACTIONS(2680),
    [sym_primitive_type] = ACTIONS(2680),
    [anon_sym_enum] = ACTIONS(2680),
    [anon_sym_struct] = ACTIONS(2680),
    [anon_sym_union] = ACTIONS(2680),
    [anon_sym_if] = ACTIONS(2680),
    [anon_sym_switch] = ACTIONS(2680),
    [anon_sym_case] = ACTIONS(2680),
    [anon_sym_default] = ACTIONS(2680),
    [anon_sym_while] = ACTIONS(2680),
    [anon_sym_do] = ACTIONS(2680),
    [anon_sym_for] = ACTIONS(2680),
    [anon_sym_in] = ACTIONS(2680),
    [anon_sym_return] = ACTIONS(2680),
    [anon_sym_break] = ACTIONS(2680),
    [anon_sym_continue] = ACTIONS(2680),
    [anon_sym_goto] = ACTIONS(2680),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_sizeof] = ACTIONS(2680),
    [anon_sym___alignof__] = ACTIONS(2680),
    [anon_sym___alignof] = ACTIONS(2680),
    [anon_sym__alignof] = ACTIONS(2680),
    [anon_sym_alignof] = ACTIONS(2680),
    [anon_sym__Alignof] = ACTIONS(2680),
    [anon_sym_offsetof] = ACTIONS(2680),
    [anon_sym__Generic] = ACTIONS(2680),
    [anon_sym_asm] = ACTIONS(2680),
    [anon_sym___asm__] = ACTIONS(2680),
    [anon_sym___asm] = ACTIONS(2680),
    [sym_number_literal] = ACTIONS(2682),
    [anon_sym_L_SQUOTE] = ACTIONS(2682),
    [anon_sym_u_SQUOTE] = ACTIONS(2682),
    [anon_sym_U_SQUOTE] = ACTIONS(2682),
    [anon_sym_u8_SQUOTE] = ACTIONS(2682),
    [anon_sym_SQUOTE] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2680),
    [anon_sym_DQUOTE] = ACTIONS(2682),
    [anon_sym_L_DQUOTE] = ACTIONS(2682),
    [anon_sym_u_DQUOTE] = ACTIONS(2682),
    [anon_sym_U_DQUOTE] = ACTIONS(2682),
    [anon_sym_u8_DQUOTE] = ACTIONS(2682),
    [sym_true] = ACTIONS(2680),
    [sym_false] = ACTIONS(2680),
    [anon_sym_NULL] = ACTIONS(2680),
    [anon_sym_nullptr] = ACTIONS(2680),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2680),
    [anon_sym___typeof] = ACTIONS(2680),
    [anon_sym_typeof] = ACTIONS(2680),
    [anon_sym_ATimport] = ACTIONS(2682),
    [aux_sym_preproc_undef_token1] = ACTIONS(2680),
    [anon_sym_POUND] = ACTIONS(2680),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2680),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2680),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2680),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2680),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2680),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2680),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2680),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_API_AVAILABLE] = ACTIONS(2680),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_API_DEPRECATED] = ACTIONS(2680),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2680),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym___deprecated_msg] = ACTIONS(2680),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2680),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2680),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2680),
    [anon_sym_ATprotocol] = ACTIONS(2682),
    [anon_sym_ATinterface] = ACTIONS(2682),
    [anon_sym_ATimplementation] = ACTIONS(2682),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2682),
    [anon_sym_ATsynthesize] = ACTIONS(2682),
    [anon_sym_ATdynamic] = ACTIONS(2682),
    [anon_sym_ATtry] = ACTIONS(2682),
    [anon_sym___try] = ACTIONS(2680),
    [anon_sym_ATthrow] = ACTIONS(2682),
    [anon_sym_ATselector] = ACTIONS(2682),
    [anon_sym_ATavailable] = ACTIONS(2682),
    [anon_sym___builtin_available] = ACTIONS(2680),
    [anon_sym_va_arg] = ACTIONS(2680),
    [anon_sym_ATencode] = ACTIONS(2682),
    [anon_sym_ATsynchronized] = ACTIONS(2682),
    [anon_sym_BOOL] = ACTIONS(2680),
    [anon_sym_IMP] = ACTIONS(2680),
    [anon_sym_SEL] = ACTIONS(2680),
    [anon_sym_Class] = ACTIONS(2680),
    [anon_sym_id] = ACTIONS(2680),
  },
  [STATE(672)] = {
    [sym_identifier] = ACTIONS(2684),
    [aux_sym_preproc_include_token1] = ACTIONS(2684),
    [aux_sym_preproc_include_token2] = ACTIONS(2684),
    [aux_sym_preproc_def_token1] = ACTIONS(2684),
    [aux_sym_preproc_if_token1] = ACTIONS(2684),
    [aux_sym_preproc_if_token2] = ACTIONS(2684),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2684),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2684),
    [sym_preproc_directive] = ACTIONS(2684),
    [anon_sym_LPAREN2] = ACTIONS(2686),
    [anon_sym_BANG] = ACTIONS(2686),
    [anon_sym_TILDE] = ACTIONS(2686),
    [anon_sym_DASH] = ACTIONS(2684),
    [anon_sym_PLUS] = ACTIONS(2684),
    [anon_sym_STAR] = ACTIONS(2686),
    [anon_sym_CARET] = ACTIONS(2686),
    [anon_sym_AMP] = ACTIONS(2686),
    [anon_sym_SEMI] = ACTIONS(2686),
    [anon_sym___extension__] = ACTIONS(2684),
    [anon_sym_typedef] = ACTIONS(2684),
    [anon_sym_extern] = ACTIONS(2684),
    [anon_sym___attribute__] = ACTIONS(2684),
    [anon_sym___attribute] = ACTIONS(2684),
    [anon_sym_noreturn] = ACTIONS(2684),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym___declspec] = ACTIONS(2684),
    [anon_sym___cdecl] = ACTIONS(2684),
    [anon_sym___clrcall] = ACTIONS(2684),
    [anon_sym___stdcall] = ACTIONS(2684),
    [anon_sym___fastcall] = ACTIONS(2684),
    [anon_sym___thiscall] = ACTIONS(2684),
    [anon_sym___vectorcall] = ACTIONS(2684),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_signed] = ACTIONS(2684),
    [anon_sym_unsigned] = ACTIONS(2684),
    [anon_sym_long] = ACTIONS(2684),
    [anon_sym_short] = ACTIONS(2684),
    [anon_sym_static] = ACTIONS(2684),
    [anon_sym_ATautoreleasepool] = ACTIONS(2686),
    [anon_sym_auto] = ACTIONS(2684),
    [anon_sym_register] = ACTIONS(2684),
    [anon_sym_inline] = ACTIONS(2684),
    [anon_sym___inline] = ACTIONS(2684),
    [anon_sym___inline__] = ACTIONS(2684),
    [anon_sym___forceinline] = ACTIONS(2684),
    [anon_sym_thread_local] = ACTIONS(2684),
    [anon_sym___thread] = ACTIONS(2684),
    [anon_sym_CG_EXTERN] = ACTIONS(2684),
    [anon_sym_CG_INLINE] = ACTIONS(2684),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2684),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2684),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2684),
    [anon_sym_IBOutlet] = ACTIONS(2684),
    [anon_sym_IBInspectable] = ACTIONS(2684),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2684),
    [anon_sym_NS_INLINE] = ACTIONS(2684),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2684),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2684),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2684),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2684),
    [anon_sym_const] = ACTIONS(2684),
    [anon_sym_constexpr] = ACTIONS(2684),
    [anon_sym_volatile] = ACTIONS(2684),
    [anon_sym_restrict] = ACTIONS(2684),
    [anon_sym___restrict__] = ACTIONS(2684),
    [anon_sym__Atomic] = ACTIONS(2684),
    [anon_sym__Noreturn] = ACTIONS(2684),
    [anon_sym__Nonnull] = ACTIONS(2684),
    [anon_sym_nullable] = ACTIONS(2684),
    [anon_sym__Complex] = ACTIONS(2684),
    [anon_sym__Nullable] = ACTIONS(2684),
    [anon_sym__Nullable_result] = ACTIONS(2684),
    [anon_sym__Null_unspecified] = ACTIONS(2684),
    [anon_sym___autoreleasing] = ACTIONS(2684),
    [anon_sym___block] = ACTIONS(2684),
    [anon_sym___bridge] = ACTIONS(2684),
    [anon_sym___bridge_retained] = ACTIONS(2684),
    [anon_sym___bridge_transfer] = ACTIONS(2684),
    [anon_sym___complex] = ACTIONS(2684),
    [anon_sym___const] = ACTIONS(2684),
    [anon_sym___imag] = ACTIONS(2684),
    [anon_sym___kindof] = ACTIONS(2684),
    [anon_sym___nonnull] = ACTIONS(2684),
    [anon_sym___nullable] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2684),
    [anon_sym___real] = ACTIONS(2684),
    [anon_sym___strong] = ACTIONS(2684),
    [anon_sym___unsafe_unretained] = ACTIONS(2684),
    [anon_sym___unused] = ACTIONS(2684),
    [anon_sym___weak] = ACTIONS(2684),
    [anon_sym_alignas] = ACTIONS(2684),
    [anon_sym__Alignas] = ACTIONS(2684),
    [sym_primitive_type] = ACTIONS(2684),
    [anon_sym_enum] = ACTIONS(2684),
    [anon_sym_struct] = ACTIONS(2684),
    [anon_sym_union] = ACTIONS(2684),
    [anon_sym_if] = ACTIONS(2684),
    [anon_sym_switch] = ACTIONS(2684),
    [anon_sym_case] = ACTIONS(2684),
    [anon_sym_default] = ACTIONS(2684),
    [anon_sym_while] = ACTIONS(2684),
    [anon_sym_do] = ACTIONS(2684),
    [anon_sym_for] = ACTIONS(2684),
    [anon_sym_in] = ACTIONS(2684),
    [anon_sym_return] = ACTIONS(2684),
    [anon_sym_break] = ACTIONS(2684),
    [anon_sym_continue] = ACTIONS(2684),
    [anon_sym_goto] = ACTIONS(2684),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_sizeof] = ACTIONS(2684),
    [anon_sym___alignof__] = ACTIONS(2684),
    [anon_sym___alignof] = ACTIONS(2684),
    [anon_sym__alignof] = ACTIONS(2684),
    [anon_sym_alignof] = ACTIONS(2684),
    [anon_sym__Alignof] = ACTIONS(2684),
    [anon_sym_offsetof] = ACTIONS(2684),
    [anon_sym__Generic] = ACTIONS(2684),
    [anon_sym_asm] = ACTIONS(2684),
    [anon_sym___asm__] = ACTIONS(2684),
    [anon_sym___asm] = ACTIONS(2684),
    [sym_number_literal] = ACTIONS(2686),
    [anon_sym_L_SQUOTE] = ACTIONS(2686),
    [anon_sym_u_SQUOTE] = ACTIONS(2686),
    [anon_sym_U_SQUOTE] = ACTIONS(2686),
    [anon_sym_u8_SQUOTE] = ACTIONS(2686),
    [anon_sym_SQUOTE] = ACTIONS(2686),
    [anon_sym_AT] = ACTIONS(2684),
    [anon_sym_DQUOTE] = ACTIONS(2686),
    [anon_sym_L_DQUOTE] = ACTIONS(2686),
    [anon_sym_u_DQUOTE] = ACTIONS(2686),
    [anon_sym_U_DQUOTE] = ACTIONS(2686),
    [anon_sym_u8_DQUOTE] = ACTIONS(2686),
    [sym_true] = ACTIONS(2684),
    [sym_false] = ACTIONS(2684),
    [anon_sym_NULL] = ACTIONS(2684),
    [anon_sym_nullptr] = ACTIONS(2684),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2684),
    [anon_sym___typeof] = ACTIONS(2684),
    [anon_sym_typeof] = ACTIONS(2684),
    [anon_sym_ATimport] = ACTIONS(2686),
    [aux_sym_preproc_undef_token1] = ACTIONS(2684),
    [anon_sym_POUND] = ACTIONS(2684),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2684),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2684),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2684),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2684),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2684),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2684),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2684),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_API_AVAILABLE] = ACTIONS(2684),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_API_DEPRECATED] = ACTIONS(2684),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2684),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym___deprecated_msg] = ACTIONS(2684),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2684),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2684),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2684),
    [anon_sym_ATprotocol] = ACTIONS(2686),
    [anon_sym_ATinterface] = ACTIONS(2686),
    [anon_sym_ATimplementation] = ACTIONS(2686),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2686),
    [anon_sym_ATsynthesize] = ACTIONS(2686),
    [anon_sym_ATdynamic] = ACTIONS(2686),
    [anon_sym_ATtry] = ACTIONS(2686),
    [anon_sym___try] = ACTIONS(2684),
    [anon_sym_ATthrow] = ACTIONS(2686),
    [anon_sym_ATselector] = ACTIONS(2686),
    [anon_sym_ATavailable] = ACTIONS(2686),
    [anon_sym___builtin_available] = ACTIONS(2684),
    [anon_sym_va_arg] = ACTIONS(2684),
    [anon_sym_ATencode] = ACTIONS(2686),
    [anon_sym_ATsynchronized] = ACTIONS(2686),
    [anon_sym_BOOL] = ACTIONS(2684),
    [anon_sym_IMP] = ACTIONS(2684),
    [anon_sym_SEL] = ACTIONS(2684),
    [anon_sym_Class] = ACTIONS(2684),
    [anon_sym_id] = ACTIONS(2684),
  },
  [STATE(673)] = {
    [sym_identifier] = ACTIONS(2864),
    [aux_sym_preproc_include_token1] = ACTIONS(2864),
    [aux_sym_preproc_include_token2] = ACTIONS(2864),
    [aux_sym_preproc_def_token1] = ACTIONS(2864),
    [aux_sym_preproc_if_token1] = ACTIONS(2864),
    [aux_sym_preproc_if_token2] = ACTIONS(2864),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2864),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2864),
    [sym_preproc_directive] = ACTIONS(2864),
    [anon_sym_LPAREN2] = ACTIONS(2866),
    [anon_sym_BANG] = ACTIONS(2866),
    [anon_sym_TILDE] = ACTIONS(2866),
    [anon_sym_DASH] = ACTIONS(2864),
    [anon_sym_PLUS] = ACTIONS(2864),
    [anon_sym_STAR] = ACTIONS(2866),
    [anon_sym_CARET] = ACTIONS(2866),
    [anon_sym_AMP] = ACTIONS(2866),
    [anon_sym_SEMI] = ACTIONS(2866),
    [anon_sym___extension__] = ACTIONS(2864),
    [anon_sym_typedef] = ACTIONS(2864),
    [anon_sym_extern] = ACTIONS(2864),
    [anon_sym___attribute__] = ACTIONS(2864),
    [anon_sym___attribute] = ACTIONS(2864),
    [anon_sym_noreturn] = ACTIONS(2864),
    [anon_sym_LBRACK] = ACTIONS(2866),
    [anon_sym___declspec] = ACTIONS(2864),
    [anon_sym___cdecl] = ACTIONS(2864),
    [anon_sym___clrcall] = ACTIONS(2864),
    [anon_sym___stdcall] = ACTIONS(2864),
    [anon_sym___fastcall] = ACTIONS(2864),
    [anon_sym___thiscall] = ACTIONS(2864),
    [anon_sym___vectorcall] = ACTIONS(2864),
    [anon_sym_LBRACE] = ACTIONS(2866),
    [anon_sym_signed] = ACTIONS(2864),
    [anon_sym_unsigned] = ACTIONS(2864),
    [anon_sym_long] = ACTIONS(2864),
    [anon_sym_short] = ACTIONS(2864),
    [anon_sym_static] = ACTIONS(2864),
    [anon_sym_ATautoreleasepool] = ACTIONS(2866),
    [anon_sym_auto] = ACTIONS(2864),
    [anon_sym_register] = ACTIONS(2864),
    [anon_sym_inline] = ACTIONS(2864),
    [anon_sym___inline] = ACTIONS(2864),
    [anon_sym___inline__] = ACTIONS(2864),
    [anon_sym___forceinline] = ACTIONS(2864),
    [anon_sym_thread_local] = ACTIONS(2864),
    [anon_sym___thread] = ACTIONS(2864),
    [anon_sym_CG_EXTERN] = ACTIONS(2864),
    [anon_sym_CG_INLINE] = ACTIONS(2864),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2864),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2864),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2864),
    [anon_sym_IBOutlet] = ACTIONS(2864),
    [anon_sym_IBInspectable] = ACTIONS(2864),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2864),
    [anon_sym_NS_INLINE] = ACTIONS(2864),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2864),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2864),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2864),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2864),
    [anon_sym_const] = ACTIONS(2864),
    [anon_sym_constexpr] = ACTIONS(2864),
    [anon_sym_volatile] = ACTIONS(2864),
    [anon_sym_restrict] = ACTIONS(2864),
    [anon_sym___restrict__] = ACTIONS(2864),
    [anon_sym__Atomic] = ACTIONS(2864),
    [anon_sym__Noreturn] = ACTIONS(2864),
    [anon_sym__Nonnull] = ACTIONS(2864),
    [anon_sym_nullable] = ACTIONS(2864),
    [anon_sym__Complex] = ACTIONS(2864),
    [anon_sym__Nullable] = ACTIONS(2864),
    [anon_sym__Nullable_result] = ACTIONS(2864),
    [anon_sym__Null_unspecified] = ACTIONS(2864),
    [anon_sym___autoreleasing] = ACTIONS(2864),
    [anon_sym___block] = ACTIONS(2864),
    [anon_sym___bridge] = ACTIONS(2864),
    [anon_sym___bridge_retained] = ACTIONS(2864),
    [anon_sym___bridge_transfer] = ACTIONS(2864),
    [anon_sym___complex] = ACTIONS(2864),
    [anon_sym___const] = ACTIONS(2864),
    [anon_sym___imag] = ACTIONS(2864),
    [anon_sym___kindof] = ACTIONS(2864),
    [anon_sym___nonnull] = ACTIONS(2864),
    [anon_sym___nullable] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2864),
    [anon_sym___real] = ACTIONS(2864),
    [anon_sym___strong] = ACTIONS(2864),
    [anon_sym___unsafe_unretained] = ACTIONS(2864),
    [anon_sym___unused] = ACTIONS(2864),
    [anon_sym___weak] = ACTIONS(2864),
    [anon_sym_alignas] = ACTIONS(2864),
    [anon_sym__Alignas] = ACTIONS(2864),
    [sym_primitive_type] = ACTIONS(2864),
    [anon_sym_enum] = ACTIONS(2864),
    [anon_sym_struct] = ACTIONS(2864),
    [anon_sym_union] = ACTIONS(2864),
    [anon_sym_if] = ACTIONS(2864),
    [anon_sym_switch] = ACTIONS(2864),
    [anon_sym_case] = ACTIONS(2864),
    [anon_sym_default] = ACTIONS(2864),
    [anon_sym_while] = ACTIONS(2864),
    [anon_sym_do] = ACTIONS(2864),
    [anon_sym_for] = ACTIONS(2864),
    [anon_sym_in] = ACTIONS(2864),
    [anon_sym_return] = ACTIONS(2864),
    [anon_sym_break] = ACTIONS(2864),
    [anon_sym_continue] = ACTIONS(2864),
    [anon_sym_goto] = ACTIONS(2864),
    [anon_sym_DASH_DASH] = ACTIONS(2866),
    [anon_sym_PLUS_PLUS] = ACTIONS(2866),
    [anon_sym_sizeof] = ACTIONS(2864),
    [anon_sym___alignof__] = ACTIONS(2864),
    [anon_sym___alignof] = ACTIONS(2864),
    [anon_sym__alignof] = ACTIONS(2864),
    [anon_sym_alignof] = ACTIONS(2864),
    [anon_sym__Alignof] = ACTIONS(2864),
    [anon_sym_offsetof] = ACTIONS(2864),
    [anon_sym__Generic] = ACTIONS(2864),
    [anon_sym_asm] = ACTIONS(2864),
    [anon_sym___asm__] = ACTIONS(2864),
    [anon_sym___asm] = ACTIONS(2864),
    [sym_number_literal] = ACTIONS(2866),
    [anon_sym_L_SQUOTE] = ACTIONS(2866),
    [anon_sym_u_SQUOTE] = ACTIONS(2866),
    [anon_sym_U_SQUOTE] = ACTIONS(2866),
    [anon_sym_u8_SQUOTE] = ACTIONS(2866),
    [anon_sym_SQUOTE] = ACTIONS(2866),
    [anon_sym_AT] = ACTIONS(2864),
    [anon_sym_DQUOTE] = ACTIONS(2866),
    [anon_sym_L_DQUOTE] = ACTIONS(2866),
    [anon_sym_u_DQUOTE] = ACTIONS(2866),
    [anon_sym_U_DQUOTE] = ACTIONS(2866),
    [anon_sym_u8_DQUOTE] = ACTIONS(2866),
    [sym_true] = ACTIONS(2864),
    [sym_false] = ACTIONS(2864),
    [anon_sym_NULL] = ACTIONS(2864),
    [anon_sym_nullptr] = ACTIONS(2864),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2864),
    [anon_sym___typeof] = ACTIONS(2864),
    [anon_sym_typeof] = ACTIONS(2864),
    [anon_sym_ATimport] = ACTIONS(2866),
    [aux_sym_preproc_undef_token1] = ACTIONS(2864),
    [anon_sym_POUND] = ACTIONS(2864),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2864),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2864),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2864),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2864),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2864),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2864),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2864),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_API_AVAILABLE] = ACTIONS(2864),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_API_DEPRECATED] = ACTIONS(2864),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2864),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym___deprecated_msg] = ACTIONS(2864),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2864),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2864),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2864),
    [anon_sym_ATprotocol] = ACTIONS(2866),
    [anon_sym_ATinterface] = ACTIONS(2866),
    [anon_sym_ATimplementation] = ACTIONS(2866),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2866),
    [anon_sym_ATsynthesize] = ACTIONS(2866),
    [anon_sym_ATdynamic] = ACTIONS(2866),
    [anon_sym_ATtry] = ACTIONS(2866),
    [anon_sym___try] = ACTIONS(2864),
    [anon_sym_ATthrow] = ACTIONS(2866),
    [anon_sym_ATselector] = ACTIONS(2866),
    [anon_sym_ATavailable] = ACTIONS(2866),
    [anon_sym___builtin_available] = ACTIONS(2864),
    [anon_sym_va_arg] = ACTIONS(2864),
    [anon_sym_ATencode] = ACTIONS(2866),
    [anon_sym_ATsynchronized] = ACTIONS(2866),
    [anon_sym_BOOL] = ACTIONS(2864),
    [anon_sym_IMP] = ACTIONS(2864),
    [anon_sym_SEL] = ACTIONS(2864),
    [anon_sym_Class] = ACTIONS(2864),
    [anon_sym_id] = ACTIONS(2864),
  },
  [STATE(674)] = {
    [sym_identifier] = ACTIONS(2692),
    [aux_sym_preproc_include_token1] = ACTIONS(2692),
    [aux_sym_preproc_include_token2] = ACTIONS(2692),
    [aux_sym_preproc_def_token1] = ACTIONS(2692),
    [aux_sym_preproc_if_token1] = ACTIONS(2692),
    [aux_sym_preproc_if_token2] = ACTIONS(2692),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2692),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2692),
    [sym_preproc_directive] = ACTIONS(2692),
    [anon_sym_LPAREN2] = ACTIONS(2694),
    [anon_sym_BANG] = ACTIONS(2694),
    [anon_sym_TILDE] = ACTIONS(2694),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2694),
    [anon_sym_CARET] = ACTIONS(2694),
    [anon_sym_AMP] = ACTIONS(2694),
    [anon_sym_SEMI] = ACTIONS(2694),
    [anon_sym___extension__] = ACTIONS(2692),
    [anon_sym_typedef] = ACTIONS(2692),
    [anon_sym_extern] = ACTIONS(2692),
    [anon_sym___attribute__] = ACTIONS(2692),
    [anon_sym___attribute] = ACTIONS(2692),
    [anon_sym_noreturn] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym___declspec] = ACTIONS(2692),
    [anon_sym___cdecl] = ACTIONS(2692),
    [anon_sym___clrcall] = ACTIONS(2692),
    [anon_sym___stdcall] = ACTIONS(2692),
    [anon_sym___fastcall] = ACTIONS(2692),
    [anon_sym___thiscall] = ACTIONS(2692),
    [anon_sym___vectorcall] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_signed] = ACTIONS(2692),
    [anon_sym_unsigned] = ACTIONS(2692),
    [anon_sym_long] = ACTIONS(2692),
    [anon_sym_short] = ACTIONS(2692),
    [anon_sym_static] = ACTIONS(2692),
    [anon_sym_ATautoreleasepool] = ACTIONS(2694),
    [anon_sym_auto] = ACTIONS(2692),
    [anon_sym_register] = ACTIONS(2692),
    [anon_sym_inline] = ACTIONS(2692),
    [anon_sym___inline] = ACTIONS(2692),
    [anon_sym___inline__] = ACTIONS(2692),
    [anon_sym___forceinline] = ACTIONS(2692),
    [anon_sym_thread_local] = ACTIONS(2692),
    [anon_sym___thread] = ACTIONS(2692),
    [anon_sym_CG_EXTERN] = ACTIONS(2692),
    [anon_sym_CG_INLINE] = ACTIONS(2692),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2692),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2692),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2692),
    [anon_sym_IBOutlet] = ACTIONS(2692),
    [anon_sym_IBInspectable] = ACTIONS(2692),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2692),
    [anon_sym_NS_INLINE] = ACTIONS(2692),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2692),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2692),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2692),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2692),
    [anon_sym_const] = ACTIONS(2692),
    [anon_sym_constexpr] = ACTIONS(2692),
    [anon_sym_volatile] = ACTIONS(2692),
    [anon_sym_restrict] = ACTIONS(2692),
    [anon_sym___restrict__] = ACTIONS(2692),
    [anon_sym__Atomic] = ACTIONS(2692),
    [anon_sym__Noreturn] = ACTIONS(2692),
    [anon_sym__Nonnull] = ACTIONS(2692),
    [anon_sym_nullable] = ACTIONS(2692),
    [anon_sym__Complex] = ACTIONS(2692),
    [anon_sym__Nullable] = ACTIONS(2692),
    [anon_sym__Nullable_result] = ACTIONS(2692),
    [anon_sym__Null_unspecified] = ACTIONS(2692),
    [anon_sym___autoreleasing] = ACTIONS(2692),
    [anon_sym___block] = ACTIONS(2692),
    [anon_sym___bridge] = ACTIONS(2692),
    [anon_sym___bridge_retained] = ACTIONS(2692),
    [anon_sym___bridge_transfer] = ACTIONS(2692),
    [anon_sym___complex] = ACTIONS(2692),
    [anon_sym___const] = ACTIONS(2692),
    [anon_sym___imag] = ACTIONS(2692),
    [anon_sym___kindof] = ACTIONS(2692),
    [anon_sym___nonnull] = ACTIONS(2692),
    [anon_sym___nullable] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2692),
    [anon_sym___real] = ACTIONS(2692),
    [anon_sym___strong] = ACTIONS(2692),
    [anon_sym___unsafe_unretained] = ACTIONS(2692),
    [anon_sym___unused] = ACTIONS(2692),
    [anon_sym___weak] = ACTIONS(2692),
    [anon_sym_alignas] = ACTIONS(2692),
    [anon_sym__Alignas] = ACTIONS(2692),
    [sym_primitive_type] = ACTIONS(2692),
    [anon_sym_enum] = ACTIONS(2692),
    [anon_sym_struct] = ACTIONS(2692),
    [anon_sym_union] = ACTIONS(2692),
    [anon_sym_if] = ACTIONS(2692),
    [anon_sym_switch] = ACTIONS(2692),
    [anon_sym_case] = ACTIONS(2692),
    [anon_sym_default] = ACTIONS(2692),
    [anon_sym_while] = ACTIONS(2692),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_for] = ACTIONS(2692),
    [anon_sym_in] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_sizeof] = ACTIONS(2692),
    [anon_sym___alignof__] = ACTIONS(2692),
    [anon_sym___alignof] = ACTIONS(2692),
    [anon_sym__alignof] = ACTIONS(2692),
    [anon_sym_alignof] = ACTIONS(2692),
    [anon_sym__Alignof] = ACTIONS(2692),
    [anon_sym_offsetof] = ACTIONS(2692),
    [anon_sym__Generic] = ACTIONS(2692),
    [anon_sym_asm] = ACTIONS(2692),
    [anon_sym___asm__] = ACTIONS(2692),
    [anon_sym___asm] = ACTIONS(2692),
    [sym_number_literal] = ACTIONS(2694),
    [anon_sym_L_SQUOTE] = ACTIONS(2694),
    [anon_sym_u_SQUOTE] = ACTIONS(2694),
    [anon_sym_U_SQUOTE] = ACTIONS(2694),
    [anon_sym_u8_SQUOTE] = ACTIONS(2694),
    [anon_sym_SQUOTE] = ACTIONS(2694),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_DQUOTE] = ACTIONS(2694),
    [anon_sym_L_DQUOTE] = ACTIONS(2694),
    [anon_sym_u_DQUOTE] = ACTIONS(2694),
    [anon_sym_U_DQUOTE] = ACTIONS(2694),
    [anon_sym_u8_DQUOTE] = ACTIONS(2694),
    [sym_true] = ACTIONS(2692),
    [sym_false] = ACTIONS(2692),
    [anon_sym_NULL] = ACTIONS(2692),
    [anon_sym_nullptr] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2692),
    [anon_sym___typeof] = ACTIONS(2692),
    [anon_sym_typeof] = ACTIONS(2692),
    [anon_sym_ATimport] = ACTIONS(2694),
    [aux_sym_preproc_undef_token1] = ACTIONS(2692),
    [anon_sym_POUND] = ACTIONS(2692),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2692),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2692),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2692),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2692),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2692),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2692),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2692),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_API_AVAILABLE] = ACTIONS(2692),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_API_DEPRECATED] = ACTIONS(2692),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2692),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym___deprecated_msg] = ACTIONS(2692),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2692),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2692),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2692),
    [anon_sym_ATprotocol] = ACTIONS(2694),
    [anon_sym_ATinterface] = ACTIONS(2694),
    [anon_sym_ATimplementation] = ACTIONS(2694),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2694),
    [anon_sym_ATsynthesize] = ACTIONS(2694),
    [anon_sym_ATdynamic] = ACTIONS(2694),
    [anon_sym_ATtry] = ACTIONS(2694),
    [anon_sym___try] = ACTIONS(2692),
    [anon_sym_ATthrow] = ACTIONS(2694),
    [anon_sym_ATselector] = ACTIONS(2694),
    [anon_sym_ATavailable] = ACTIONS(2694),
    [anon_sym___builtin_available] = ACTIONS(2692),
    [anon_sym_va_arg] = ACTIONS(2692),
    [anon_sym_ATencode] = ACTIONS(2694),
    [anon_sym_ATsynchronized] = ACTIONS(2694),
    [anon_sym_BOOL] = ACTIONS(2692),
    [anon_sym_IMP] = ACTIONS(2692),
    [anon_sym_SEL] = ACTIONS(2692),
    [anon_sym_Class] = ACTIONS(2692),
    [anon_sym_id] = ACTIONS(2692),
  },
  [STATE(675)] = {
    [sym_identifier] = ACTIONS(2696),
    [aux_sym_preproc_include_token1] = ACTIONS(2696),
    [aux_sym_preproc_include_token2] = ACTIONS(2696),
    [aux_sym_preproc_def_token1] = ACTIONS(2696),
    [aux_sym_preproc_if_token1] = ACTIONS(2696),
    [aux_sym_preproc_if_token2] = ACTIONS(2696),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2696),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2696),
    [sym_preproc_directive] = ACTIONS(2696),
    [anon_sym_LPAREN2] = ACTIONS(2698),
    [anon_sym_BANG] = ACTIONS(2698),
    [anon_sym_TILDE] = ACTIONS(2698),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2698),
    [anon_sym_CARET] = ACTIONS(2698),
    [anon_sym_AMP] = ACTIONS(2698),
    [anon_sym_SEMI] = ACTIONS(2698),
    [anon_sym___extension__] = ACTIONS(2696),
    [anon_sym_typedef] = ACTIONS(2696),
    [anon_sym_extern] = ACTIONS(2696),
    [anon_sym___attribute__] = ACTIONS(2696),
    [anon_sym___attribute] = ACTIONS(2696),
    [anon_sym_noreturn] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym___declspec] = ACTIONS(2696),
    [anon_sym___cdecl] = ACTIONS(2696),
    [anon_sym___clrcall] = ACTIONS(2696),
    [anon_sym___stdcall] = ACTIONS(2696),
    [anon_sym___fastcall] = ACTIONS(2696),
    [anon_sym___thiscall] = ACTIONS(2696),
    [anon_sym___vectorcall] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_signed] = ACTIONS(2696),
    [anon_sym_unsigned] = ACTIONS(2696),
    [anon_sym_long] = ACTIONS(2696),
    [anon_sym_short] = ACTIONS(2696),
    [anon_sym_static] = ACTIONS(2696),
    [anon_sym_ATautoreleasepool] = ACTIONS(2698),
    [anon_sym_auto] = ACTIONS(2696),
    [anon_sym_register] = ACTIONS(2696),
    [anon_sym_inline] = ACTIONS(2696),
    [anon_sym___inline] = ACTIONS(2696),
    [anon_sym___inline__] = ACTIONS(2696),
    [anon_sym___forceinline] = ACTIONS(2696),
    [anon_sym_thread_local] = ACTIONS(2696),
    [anon_sym___thread] = ACTIONS(2696),
    [anon_sym_CG_EXTERN] = ACTIONS(2696),
    [anon_sym_CG_INLINE] = ACTIONS(2696),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2696),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2696),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2696),
    [anon_sym_IBOutlet] = ACTIONS(2696),
    [anon_sym_IBInspectable] = ACTIONS(2696),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2696),
    [anon_sym_NS_INLINE] = ACTIONS(2696),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2696),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2696),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2696),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2696),
    [anon_sym_const] = ACTIONS(2696),
    [anon_sym_constexpr] = ACTIONS(2696),
    [anon_sym_volatile] = ACTIONS(2696),
    [anon_sym_restrict] = ACTIONS(2696),
    [anon_sym___restrict__] = ACTIONS(2696),
    [anon_sym__Atomic] = ACTIONS(2696),
    [anon_sym__Noreturn] = ACTIONS(2696),
    [anon_sym__Nonnull] = ACTIONS(2696),
    [anon_sym_nullable] = ACTIONS(2696),
    [anon_sym__Complex] = ACTIONS(2696),
    [anon_sym__Nullable] = ACTIONS(2696),
    [anon_sym__Nullable_result] = ACTIONS(2696),
    [anon_sym__Null_unspecified] = ACTIONS(2696),
    [anon_sym___autoreleasing] = ACTIONS(2696),
    [anon_sym___block] = ACTIONS(2696),
    [anon_sym___bridge] = ACTIONS(2696),
    [anon_sym___bridge_retained] = ACTIONS(2696),
    [anon_sym___bridge_transfer] = ACTIONS(2696),
    [anon_sym___complex] = ACTIONS(2696),
    [anon_sym___const] = ACTIONS(2696),
    [anon_sym___imag] = ACTIONS(2696),
    [anon_sym___kindof] = ACTIONS(2696),
    [anon_sym___nonnull] = ACTIONS(2696),
    [anon_sym___nullable] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2696),
    [anon_sym___real] = ACTIONS(2696),
    [anon_sym___strong] = ACTIONS(2696),
    [anon_sym___unsafe_unretained] = ACTIONS(2696),
    [anon_sym___unused] = ACTIONS(2696),
    [anon_sym___weak] = ACTIONS(2696),
    [anon_sym_alignas] = ACTIONS(2696),
    [anon_sym__Alignas] = ACTIONS(2696),
    [sym_primitive_type] = ACTIONS(2696),
    [anon_sym_enum] = ACTIONS(2696),
    [anon_sym_struct] = ACTIONS(2696),
    [anon_sym_union] = ACTIONS(2696),
    [anon_sym_if] = ACTIONS(2696),
    [anon_sym_switch] = ACTIONS(2696),
    [anon_sym_case] = ACTIONS(2696),
    [anon_sym_default] = ACTIONS(2696),
    [anon_sym_while] = ACTIONS(2696),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_for] = ACTIONS(2696),
    [anon_sym_in] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_sizeof] = ACTIONS(2696),
    [anon_sym___alignof__] = ACTIONS(2696),
    [anon_sym___alignof] = ACTIONS(2696),
    [anon_sym__alignof] = ACTIONS(2696),
    [anon_sym_alignof] = ACTIONS(2696),
    [anon_sym__Alignof] = ACTIONS(2696),
    [anon_sym_offsetof] = ACTIONS(2696),
    [anon_sym__Generic] = ACTIONS(2696),
    [anon_sym_asm] = ACTIONS(2696),
    [anon_sym___asm__] = ACTIONS(2696),
    [anon_sym___asm] = ACTIONS(2696),
    [sym_number_literal] = ACTIONS(2698),
    [anon_sym_L_SQUOTE] = ACTIONS(2698),
    [anon_sym_u_SQUOTE] = ACTIONS(2698),
    [anon_sym_U_SQUOTE] = ACTIONS(2698),
    [anon_sym_u8_SQUOTE] = ACTIONS(2698),
    [anon_sym_SQUOTE] = ACTIONS(2698),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_DQUOTE] = ACTIONS(2698),
    [anon_sym_L_DQUOTE] = ACTIONS(2698),
    [anon_sym_u_DQUOTE] = ACTIONS(2698),
    [anon_sym_U_DQUOTE] = ACTIONS(2698),
    [anon_sym_u8_DQUOTE] = ACTIONS(2698),
    [sym_true] = ACTIONS(2696),
    [sym_false] = ACTIONS(2696),
    [anon_sym_NULL] = ACTIONS(2696),
    [anon_sym_nullptr] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2696),
    [anon_sym___typeof] = ACTIONS(2696),
    [anon_sym_typeof] = ACTIONS(2696),
    [anon_sym_ATimport] = ACTIONS(2698),
    [aux_sym_preproc_undef_token1] = ACTIONS(2696),
    [anon_sym_POUND] = ACTIONS(2696),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2696),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2696),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2696),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2696),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2696),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2696),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2696),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_API_AVAILABLE] = ACTIONS(2696),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_API_DEPRECATED] = ACTIONS(2696),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2696),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym___deprecated_msg] = ACTIONS(2696),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2696),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2696),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2696),
    [anon_sym_ATprotocol] = ACTIONS(2698),
    [anon_sym_ATinterface] = ACTIONS(2698),
    [anon_sym_ATimplementation] = ACTIONS(2698),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2698),
    [anon_sym_ATsynthesize] = ACTIONS(2698),
    [anon_sym_ATdynamic] = ACTIONS(2698),
    [anon_sym_ATtry] = ACTIONS(2698),
    [anon_sym___try] = ACTIONS(2696),
    [anon_sym_ATthrow] = ACTIONS(2698),
    [anon_sym_ATselector] = ACTIONS(2698),
    [anon_sym_ATavailable] = ACTIONS(2698),
    [anon_sym___builtin_available] = ACTIONS(2696),
    [anon_sym_va_arg] = ACTIONS(2696),
    [anon_sym_ATencode] = ACTIONS(2698),
    [anon_sym_ATsynchronized] = ACTIONS(2698),
    [anon_sym_BOOL] = ACTIONS(2696),
    [anon_sym_IMP] = ACTIONS(2696),
    [anon_sym_SEL] = ACTIONS(2696),
    [anon_sym_Class] = ACTIONS(2696),
    [anon_sym_id] = ACTIONS(2696),
  },
  [STATE(676)] = {
    [sym_identifier] = ACTIONS(2700),
    [aux_sym_preproc_include_token1] = ACTIONS(2700),
    [aux_sym_preproc_include_token2] = ACTIONS(2700),
    [aux_sym_preproc_def_token1] = ACTIONS(2700),
    [aux_sym_preproc_if_token1] = ACTIONS(2700),
    [aux_sym_preproc_if_token2] = ACTIONS(2700),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2700),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2700),
    [sym_preproc_directive] = ACTIONS(2700),
    [anon_sym_LPAREN2] = ACTIONS(2702),
    [anon_sym_BANG] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2700),
    [anon_sym_PLUS] = ACTIONS(2700),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym___extension__] = ACTIONS(2700),
    [anon_sym_typedef] = ACTIONS(2700),
    [anon_sym_extern] = ACTIONS(2700),
    [anon_sym___attribute__] = ACTIONS(2700),
    [anon_sym___attribute] = ACTIONS(2700),
    [anon_sym_noreturn] = ACTIONS(2700),
    [anon_sym_LBRACK] = ACTIONS(2702),
    [anon_sym___declspec] = ACTIONS(2700),
    [anon_sym___cdecl] = ACTIONS(2700),
    [anon_sym___clrcall] = ACTIONS(2700),
    [anon_sym___stdcall] = ACTIONS(2700),
    [anon_sym___fastcall] = ACTIONS(2700),
    [anon_sym___thiscall] = ACTIONS(2700),
    [anon_sym___vectorcall] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2702),
    [anon_sym_signed] = ACTIONS(2700),
    [anon_sym_unsigned] = ACTIONS(2700),
    [anon_sym_long] = ACTIONS(2700),
    [anon_sym_short] = ACTIONS(2700),
    [anon_sym_static] = ACTIONS(2700),
    [anon_sym_ATautoreleasepool] = ACTIONS(2702),
    [anon_sym_auto] = ACTIONS(2700),
    [anon_sym_register] = ACTIONS(2700),
    [anon_sym_inline] = ACTIONS(2700),
    [anon_sym___inline] = ACTIONS(2700),
    [anon_sym___inline__] = ACTIONS(2700),
    [anon_sym___forceinline] = ACTIONS(2700),
    [anon_sym_thread_local] = ACTIONS(2700),
    [anon_sym___thread] = ACTIONS(2700),
    [anon_sym_CG_EXTERN] = ACTIONS(2700),
    [anon_sym_CG_INLINE] = ACTIONS(2700),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2700),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2700),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2700),
    [anon_sym_IBOutlet] = ACTIONS(2700),
    [anon_sym_IBInspectable] = ACTIONS(2700),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2700),
    [anon_sym_NS_INLINE] = ACTIONS(2700),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2700),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2700),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2700),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2700),
    [anon_sym_const] = ACTIONS(2700),
    [anon_sym_constexpr] = ACTIONS(2700),
    [anon_sym_volatile] = ACTIONS(2700),
    [anon_sym_restrict] = ACTIONS(2700),
    [anon_sym___restrict__] = ACTIONS(2700),
    [anon_sym__Atomic] = ACTIONS(2700),
    [anon_sym__Noreturn] = ACTIONS(2700),
    [anon_sym__Nonnull] = ACTIONS(2700),
    [anon_sym_nullable] = ACTIONS(2700),
    [anon_sym__Complex] = ACTIONS(2700),
    [anon_sym__Nullable] = ACTIONS(2700),
    [anon_sym__Nullable_result] = ACTIONS(2700),
    [anon_sym__Null_unspecified] = ACTIONS(2700),
    [anon_sym___autoreleasing] = ACTIONS(2700),
    [anon_sym___block] = ACTIONS(2700),
    [anon_sym___bridge] = ACTIONS(2700),
    [anon_sym___bridge_retained] = ACTIONS(2700),
    [anon_sym___bridge_transfer] = ACTIONS(2700),
    [anon_sym___complex] = ACTIONS(2700),
    [anon_sym___const] = ACTIONS(2700),
    [anon_sym___imag] = ACTIONS(2700),
    [anon_sym___kindof] = ACTIONS(2700),
    [anon_sym___nonnull] = ACTIONS(2700),
    [anon_sym___nullable] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2700),
    [anon_sym___real] = ACTIONS(2700),
    [anon_sym___strong] = ACTIONS(2700),
    [anon_sym___unsafe_unretained] = ACTIONS(2700),
    [anon_sym___unused] = ACTIONS(2700),
    [anon_sym___weak] = ACTIONS(2700),
    [anon_sym_alignas] = ACTIONS(2700),
    [anon_sym__Alignas] = ACTIONS(2700),
    [sym_primitive_type] = ACTIONS(2700),
    [anon_sym_enum] = ACTIONS(2700),
    [anon_sym_struct] = ACTIONS(2700),
    [anon_sym_union] = ACTIONS(2700),
    [anon_sym_if] = ACTIONS(2700),
    [anon_sym_switch] = ACTIONS(2700),
    [anon_sym_case] = ACTIONS(2700),
    [anon_sym_default] = ACTIONS(2700),
    [anon_sym_while] = ACTIONS(2700),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_for] = ACTIONS(2700),
    [anon_sym_in] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_DASH_DASH] = ACTIONS(2702),
    [anon_sym_PLUS_PLUS] = ACTIONS(2702),
    [anon_sym_sizeof] = ACTIONS(2700),
    [anon_sym___alignof__] = ACTIONS(2700),
    [anon_sym___alignof] = ACTIONS(2700),
    [anon_sym__alignof] = ACTIONS(2700),
    [anon_sym_alignof] = ACTIONS(2700),
    [anon_sym__Alignof] = ACTIONS(2700),
    [anon_sym_offsetof] = ACTIONS(2700),
    [anon_sym__Generic] = ACTIONS(2700),
    [anon_sym_asm] = ACTIONS(2700),
    [anon_sym___asm__] = ACTIONS(2700),
    [anon_sym___asm] = ACTIONS(2700),
    [sym_number_literal] = ACTIONS(2702),
    [anon_sym_L_SQUOTE] = ACTIONS(2702),
    [anon_sym_u_SQUOTE] = ACTIONS(2702),
    [anon_sym_U_SQUOTE] = ACTIONS(2702),
    [anon_sym_u8_SQUOTE] = ACTIONS(2702),
    [anon_sym_SQUOTE] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_DQUOTE] = ACTIONS(2702),
    [anon_sym_L_DQUOTE] = ACTIONS(2702),
    [anon_sym_u_DQUOTE] = ACTIONS(2702),
    [anon_sym_U_DQUOTE] = ACTIONS(2702),
    [anon_sym_u8_DQUOTE] = ACTIONS(2702),
    [sym_true] = ACTIONS(2700),
    [sym_false] = ACTIONS(2700),
    [anon_sym_NULL] = ACTIONS(2700),
    [anon_sym_nullptr] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2700),
    [anon_sym___typeof] = ACTIONS(2700),
    [anon_sym_typeof] = ACTIONS(2700),
    [anon_sym_ATimport] = ACTIONS(2702),
    [aux_sym_preproc_undef_token1] = ACTIONS(2700),
    [anon_sym_POUND] = ACTIONS(2700),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2700),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2700),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2700),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2700),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2700),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2700),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2700),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_API_AVAILABLE] = ACTIONS(2700),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_API_DEPRECATED] = ACTIONS(2700),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2700),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym___deprecated_msg] = ACTIONS(2700),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2700),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2700),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2700),
    [anon_sym_ATprotocol] = ACTIONS(2702),
    [anon_sym_ATinterface] = ACTIONS(2702),
    [anon_sym_ATimplementation] = ACTIONS(2702),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2702),
    [anon_sym_ATsynthesize] = ACTIONS(2702),
    [anon_sym_ATdynamic] = ACTIONS(2702),
    [anon_sym_ATtry] = ACTIONS(2702),
    [anon_sym___try] = ACTIONS(2700),
    [anon_sym_ATthrow] = ACTIONS(2702),
    [anon_sym_ATselector] = ACTIONS(2702),
    [anon_sym_ATavailable] = ACTIONS(2702),
    [anon_sym___builtin_available] = ACTIONS(2700),
    [anon_sym_va_arg] = ACTIONS(2700),
    [anon_sym_ATencode] = ACTIONS(2702),
    [anon_sym_ATsynchronized] = ACTIONS(2702),
    [anon_sym_BOOL] = ACTIONS(2700),
    [anon_sym_IMP] = ACTIONS(2700),
    [anon_sym_SEL] = ACTIONS(2700),
    [anon_sym_Class] = ACTIONS(2700),
    [anon_sym_id] = ACTIONS(2700),
  },
  [STATE(677)] = {
    [sym_identifier] = ACTIONS(2708),
    [aux_sym_preproc_include_token1] = ACTIONS(2708),
    [aux_sym_preproc_include_token2] = ACTIONS(2708),
    [aux_sym_preproc_def_token1] = ACTIONS(2708),
    [aux_sym_preproc_if_token1] = ACTIONS(2708),
    [aux_sym_preproc_if_token2] = ACTIONS(2708),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2708),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2708),
    [sym_preproc_directive] = ACTIONS(2708),
    [anon_sym_LPAREN2] = ACTIONS(2710),
    [anon_sym_BANG] = ACTIONS(2710),
    [anon_sym_TILDE] = ACTIONS(2710),
    [anon_sym_DASH] = ACTIONS(2708),
    [anon_sym_PLUS] = ACTIONS(2708),
    [anon_sym_STAR] = ACTIONS(2710),
    [anon_sym_CARET] = ACTIONS(2710),
    [anon_sym_AMP] = ACTIONS(2710),
    [anon_sym_SEMI] = ACTIONS(2710),
    [anon_sym___extension__] = ACTIONS(2708),
    [anon_sym_typedef] = ACTIONS(2708),
    [anon_sym_extern] = ACTIONS(2708),
    [anon_sym___attribute__] = ACTIONS(2708),
    [anon_sym___attribute] = ACTIONS(2708),
    [anon_sym_noreturn] = ACTIONS(2708),
    [anon_sym_LBRACK] = ACTIONS(2710),
    [anon_sym___declspec] = ACTIONS(2708),
    [anon_sym___cdecl] = ACTIONS(2708),
    [anon_sym___clrcall] = ACTIONS(2708),
    [anon_sym___stdcall] = ACTIONS(2708),
    [anon_sym___fastcall] = ACTIONS(2708),
    [anon_sym___thiscall] = ACTIONS(2708),
    [anon_sym___vectorcall] = ACTIONS(2708),
    [anon_sym_LBRACE] = ACTIONS(2710),
    [anon_sym_signed] = ACTIONS(2708),
    [anon_sym_unsigned] = ACTIONS(2708),
    [anon_sym_long] = ACTIONS(2708),
    [anon_sym_short] = ACTIONS(2708),
    [anon_sym_static] = ACTIONS(2708),
    [anon_sym_ATautoreleasepool] = ACTIONS(2710),
    [anon_sym_auto] = ACTIONS(2708),
    [anon_sym_register] = ACTIONS(2708),
    [anon_sym_inline] = ACTIONS(2708),
    [anon_sym___inline] = ACTIONS(2708),
    [anon_sym___inline__] = ACTIONS(2708),
    [anon_sym___forceinline] = ACTIONS(2708),
    [anon_sym_thread_local] = ACTIONS(2708),
    [anon_sym___thread] = ACTIONS(2708),
    [anon_sym_CG_EXTERN] = ACTIONS(2708),
    [anon_sym_CG_INLINE] = ACTIONS(2708),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2708),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2708),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2708),
    [anon_sym_IBOutlet] = ACTIONS(2708),
    [anon_sym_IBInspectable] = ACTIONS(2708),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2708),
    [anon_sym_NS_INLINE] = ACTIONS(2708),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2708),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2708),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2708),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2708),
    [anon_sym_const] = ACTIONS(2708),
    [anon_sym_constexpr] = ACTIONS(2708),
    [anon_sym_volatile] = ACTIONS(2708),
    [anon_sym_restrict] = ACTIONS(2708),
    [anon_sym___restrict__] = ACTIONS(2708),
    [anon_sym__Atomic] = ACTIONS(2708),
    [anon_sym__Noreturn] = ACTIONS(2708),
    [anon_sym__Nonnull] = ACTIONS(2708),
    [anon_sym_nullable] = ACTIONS(2708),
    [anon_sym__Complex] = ACTIONS(2708),
    [anon_sym__Nullable] = ACTIONS(2708),
    [anon_sym__Nullable_result] = ACTIONS(2708),
    [anon_sym__Null_unspecified] = ACTIONS(2708),
    [anon_sym___autoreleasing] = ACTIONS(2708),
    [anon_sym___block] = ACTIONS(2708),
    [anon_sym___bridge] = ACTIONS(2708),
    [anon_sym___bridge_retained] = ACTIONS(2708),
    [anon_sym___bridge_transfer] = ACTIONS(2708),
    [anon_sym___complex] = ACTIONS(2708),
    [anon_sym___const] = ACTIONS(2708),
    [anon_sym___imag] = ACTIONS(2708),
    [anon_sym___kindof] = ACTIONS(2708),
    [anon_sym___nonnull] = ACTIONS(2708),
    [anon_sym___nullable] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2708),
    [anon_sym___real] = ACTIONS(2708),
    [anon_sym___strong] = ACTIONS(2708),
    [anon_sym___unsafe_unretained] = ACTIONS(2708),
    [anon_sym___unused] = ACTIONS(2708),
    [anon_sym___weak] = ACTIONS(2708),
    [anon_sym_alignas] = ACTIONS(2708),
    [anon_sym__Alignas] = ACTIONS(2708),
    [sym_primitive_type] = ACTIONS(2708),
    [anon_sym_enum] = ACTIONS(2708),
    [anon_sym_struct] = ACTIONS(2708),
    [anon_sym_union] = ACTIONS(2708),
    [anon_sym_if] = ACTIONS(2708),
    [anon_sym_switch] = ACTIONS(2708),
    [anon_sym_case] = ACTIONS(2708),
    [anon_sym_default] = ACTIONS(2708),
    [anon_sym_while] = ACTIONS(2708),
    [anon_sym_do] = ACTIONS(2708),
    [anon_sym_for] = ACTIONS(2708),
    [anon_sym_in] = ACTIONS(2708),
    [anon_sym_return] = ACTIONS(2708),
    [anon_sym_break] = ACTIONS(2708),
    [anon_sym_continue] = ACTIONS(2708),
    [anon_sym_goto] = ACTIONS(2708),
    [anon_sym_DASH_DASH] = ACTIONS(2710),
    [anon_sym_PLUS_PLUS] = ACTIONS(2710),
    [anon_sym_sizeof] = ACTIONS(2708),
    [anon_sym___alignof__] = ACTIONS(2708),
    [anon_sym___alignof] = ACTIONS(2708),
    [anon_sym__alignof] = ACTIONS(2708),
    [anon_sym_alignof] = ACTIONS(2708),
    [anon_sym__Alignof] = ACTIONS(2708),
    [anon_sym_offsetof] = ACTIONS(2708),
    [anon_sym__Generic] = ACTIONS(2708),
    [anon_sym_asm] = ACTIONS(2708),
    [anon_sym___asm__] = ACTIONS(2708),
    [anon_sym___asm] = ACTIONS(2708),
    [sym_number_literal] = ACTIONS(2710),
    [anon_sym_L_SQUOTE] = ACTIONS(2710),
    [anon_sym_u_SQUOTE] = ACTIONS(2710),
    [anon_sym_U_SQUOTE] = ACTIONS(2710),
    [anon_sym_u8_SQUOTE] = ACTIONS(2710),
    [anon_sym_SQUOTE] = ACTIONS(2710),
    [anon_sym_AT] = ACTIONS(2708),
    [anon_sym_DQUOTE] = ACTIONS(2710),
    [anon_sym_L_DQUOTE] = ACTIONS(2710),
    [anon_sym_u_DQUOTE] = ACTIONS(2710),
    [anon_sym_U_DQUOTE] = ACTIONS(2710),
    [anon_sym_u8_DQUOTE] = ACTIONS(2710),
    [sym_true] = ACTIONS(2708),
    [sym_false] = ACTIONS(2708),
    [anon_sym_NULL] = ACTIONS(2708),
    [anon_sym_nullptr] = ACTIONS(2708),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2708),
    [anon_sym___typeof] = ACTIONS(2708),
    [anon_sym_typeof] = ACTIONS(2708),
    [anon_sym_ATimport] = ACTIONS(2710),
    [aux_sym_preproc_undef_token1] = ACTIONS(2708),
    [anon_sym_POUND] = ACTIONS(2708),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2708),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2708),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2708),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2708),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2708),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2708),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2708),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_API_AVAILABLE] = ACTIONS(2708),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_API_DEPRECATED] = ACTIONS(2708),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2708),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym___deprecated_msg] = ACTIONS(2708),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2708),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2708),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2708),
    [anon_sym_ATprotocol] = ACTIONS(2710),
    [anon_sym_ATinterface] = ACTIONS(2710),
    [anon_sym_ATimplementation] = ACTIONS(2710),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2710),
    [anon_sym_ATsynthesize] = ACTIONS(2710),
    [anon_sym_ATdynamic] = ACTIONS(2710),
    [anon_sym_ATtry] = ACTIONS(2710),
    [anon_sym___try] = ACTIONS(2708),
    [anon_sym_ATthrow] = ACTIONS(2710),
    [anon_sym_ATselector] = ACTIONS(2710),
    [anon_sym_ATavailable] = ACTIONS(2710),
    [anon_sym___builtin_available] = ACTIONS(2708),
    [anon_sym_va_arg] = ACTIONS(2708),
    [anon_sym_ATencode] = ACTIONS(2710),
    [anon_sym_ATsynchronized] = ACTIONS(2710),
    [anon_sym_BOOL] = ACTIONS(2708),
    [anon_sym_IMP] = ACTIONS(2708),
    [anon_sym_SEL] = ACTIONS(2708),
    [anon_sym_Class] = ACTIONS(2708),
    [anon_sym_id] = ACTIONS(2708),
  },
  [STATE(678)] = {
    [sym_identifier] = ACTIONS(2712),
    [aux_sym_preproc_include_token1] = ACTIONS(2712),
    [aux_sym_preproc_include_token2] = ACTIONS(2712),
    [aux_sym_preproc_def_token1] = ACTIONS(2712),
    [aux_sym_preproc_if_token1] = ACTIONS(2712),
    [aux_sym_preproc_if_token2] = ACTIONS(2712),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2712),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2712),
    [sym_preproc_directive] = ACTIONS(2712),
    [anon_sym_LPAREN2] = ACTIONS(2714),
    [anon_sym_BANG] = ACTIONS(2714),
    [anon_sym_TILDE] = ACTIONS(2714),
    [anon_sym_DASH] = ACTIONS(2712),
    [anon_sym_PLUS] = ACTIONS(2712),
    [anon_sym_STAR] = ACTIONS(2714),
    [anon_sym_CARET] = ACTIONS(2714),
    [anon_sym_AMP] = ACTIONS(2714),
    [anon_sym_SEMI] = ACTIONS(2714),
    [anon_sym___extension__] = ACTIONS(2712),
    [anon_sym_typedef] = ACTIONS(2712),
    [anon_sym_extern] = ACTIONS(2712),
    [anon_sym___attribute__] = ACTIONS(2712),
    [anon_sym___attribute] = ACTIONS(2712),
    [anon_sym_noreturn] = ACTIONS(2712),
    [anon_sym_LBRACK] = ACTIONS(2714),
    [anon_sym___declspec] = ACTIONS(2712),
    [anon_sym___cdecl] = ACTIONS(2712),
    [anon_sym___clrcall] = ACTIONS(2712),
    [anon_sym___stdcall] = ACTIONS(2712),
    [anon_sym___fastcall] = ACTIONS(2712),
    [anon_sym___thiscall] = ACTIONS(2712),
    [anon_sym___vectorcall] = ACTIONS(2712),
    [anon_sym_LBRACE] = ACTIONS(2714),
    [anon_sym_signed] = ACTIONS(2712),
    [anon_sym_unsigned] = ACTIONS(2712),
    [anon_sym_long] = ACTIONS(2712),
    [anon_sym_short] = ACTIONS(2712),
    [anon_sym_static] = ACTIONS(2712),
    [anon_sym_ATautoreleasepool] = ACTIONS(2714),
    [anon_sym_auto] = ACTIONS(2712),
    [anon_sym_register] = ACTIONS(2712),
    [anon_sym_inline] = ACTIONS(2712),
    [anon_sym___inline] = ACTIONS(2712),
    [anon_sym___inline__] = ACTIONS(2712),
    [anon_sym___forceinline] = ACTIONS(2712),
    [anon_sym_thread_local] = ACTIONS(2712),
    [anon_sym___thread] = ACTIONS(2712),
    [anon_sym_CG_EXTERN] = ACTIONS(2712),
    [anon_sym_CG_INLINE] = ACTIONS(2712),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2712),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2712),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2712),
    [anon_sym_IBOutlet] = ACTIONS(2712),
    [anon_sym_IBInspectable] = ACTIONS(2712),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2712),
    [anon_sym_NS_INLINE] = ACTIONS(2712),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2712),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2712),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2712),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2712),
    [anon_sym_const] = ACTIONS(2712),
    [anon_sym_constexpr] = ACTIONS(2712),
    [anon_sym_volatile] = ACTIONS(2712),
    [anon_sym_restrict] = ACTIONS(2712),
    [anon_sym___restrict__] = ACTIONS(2712),
    [anon_sym__Atomic] = ACTIONS(2712),
    [anon_sym__Noreturn] = ACTIONS(2712),
    [anon_sym__Nonnull] = ACTIONS(2712),
    [anon_sym_nullable] = ACTIONS(2712),
    [anon_sym__Complex] = ACTIONS(2712),
    [anon_sym__Nullable] = ACTIONS(2712),
    [anon_sym__Nullable_result] = ACTIONS(2712),
    [anon_sym__Null_unspecified] = ACTIONS(2712),
    [anon_sym___autoreleasing] = ACTIONS(2712),
    [anon_sym___block] = ACTIONS(2712),
    [anon_sym___bridge] = ACTIONS(2712),
    [anon_sym___bridge_retained] = ACTIONS(2712),
    [anon_sym___bridge_transfer] = ACTIONS(2712),
    [anon_sym___complex] = ACTIONS(2712),
    [anon_sym___const] = ACTIONS(2712),
    [anon_sym___imag] = ACTIONS(2712),
    [anon_sym___kindof] = ACTIONS(2712),
    [anon_sym___nonnull] = ACTIONS(2712),
    [anon_sym___nullable] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2712),
    [anon_sym___real] = ACTIONS(2712),
    [anon_sym___strong] = ACTIONS(2712),
    [anon_sym___unsafe_unretained] = ACTIONS(2712),
    [anon_sym___unused] = ACTIONS(2712),
    [anon_sym___weak] = ACTIONS(2712),
    [anon_sym_alignas] = ACTIONS(2712),
    [anon_sym__Alignas] = ACTIONS(2712),
    [sym_primitive_type] = ACTIONS(2712),
    [anon_sym_enum] = ACTIONS(2712),
    [anon_sym_struct] = ACTIONS(2712),
    [anon_sym_union] = ACTIONS(2712),
    [anon_sym_if] = ACTIONS(2712),
    [anon_sym_switch] = ACTIONS(2712),
    [anon_sym_case] = ACTIONS(2712),
    [anon_sym_default] = ACTIONS(2712),
    [anon_sym_while] = ACTIONS(2712),
    [anon_sym_do] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2712),
    [anon_sym_in] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_break] = ACTIONS(2712),
    [anon_sym_continue] = ACTIONS(2712),
    [anon_sym_goto] = ACTIONS(2712),
    [anon_sym_DASH_DASH] = ACTIONS(2714),
    [anon_sym_PLUS_PLUS] = ACTIONS(2714),
    [anon_sym_sizeof] = ACTIONS(2712),
    [anon_sym___alignof__] = ACTIONS(2712),
    [anon_sym___alignof] = ACTIONS(2712),
    [anon_sym__alignof] = ACTIONS(2712),
    [anon_sym_alignof] = ACTIONS(2712),
    [anon_sym__Alignof] = ACTIONS(2712),
    [anon_sym_offsetof] = ACTIONS(2712),
    [anon_sym__Generic] = ACTIONS(2712),
    [anon_sym_asm] = ACTIONS(2712),
    [anon_sym___asm__] = ACTIONS(2712),
    [anon_sym___asm] = ACTIONS(2712),
    [sym_number_literal] = ACTIONS(2714),
    [anon_sym_L_SQUOTE] = ACTIONS(2714),
    [anon_sym_u_SQUOTE] = ACTIONS(2714),
    [anon_sym_U_SQUOTE] = ACTIONS(2714),
    [anon_sym_u8_SQUOTE] = ACTIONS(2714),
    [anon_sym_SQUOTE] = ACTIONS(2714),
    [anon_sym_AT] = ACTIONS(2712),
    [anon_sym_DQUOTE] = ACTIONS(2714),
    [anon_sym_L_DQUOTE] = ACTIONS(2714),
    [anon_sym_u_DQUOTE] = ACTIONS(2714),
    [anon_sym_U_DQUOTE] = ACTIONS(2714),
    [anon_sym_u8_DQUOTE] = ACTIONS(2714),
    [sym_true] = ACTIONS(2712),
    [sym_false] = ACTIONS(2712),
    [anon_sym_NULL] = ACTIONS(2712),
    [anon_sym_nullptr] = ACTIONS(2712),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2712),
    [anon_sym___typeof] = ACTIONS(2712),
    [anon_sym_typeof] = ACTIONS(2712),
    [anon_sym_ATimport] = ACTIONS(2714),
    [aux_sym_preproc_undef_token1] = ACTIONS(2712),
    [anon_sym_POUND] = ACTIONS(2712),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2712),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2712),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2712),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2712),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2712),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2712),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2712),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_API_AVAILABLE] = ACTIONS(2712),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_API_DEPRECATED] = ACTIONS(2712),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2712),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym___deprecated_msg] = ACTIONS(2712),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2712),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2712),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2712),
    [anon_sym_ATprotocol] = ACTIONS(2714),
    [anon_sym_ATinterface] = ACTIONS(2714),
    [anon_sym_ATimplementation] = ACTIONS(2714),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2714),
    [anon_sym_ATsynthesize] = ACTIONS(2714),
    [anon_sym_ATdynamic] = ACTIONS(2714),
    [anon_sym_ATtry] = ACTIONS(2714),
    [anon_sym___try] = ACTIONS(2712),
    [anon_sym_ATthrow] = ACTIONS(2714),
    [anon_sym_ATselector] = ACTIONS(2714),
    [anon_sym_ATavailable] = ACTIONS(2714),
    [anon_sym___builtin_available] = ACTIONS(2712),
    [anon_sym_va_arg] = ACTIONS(2712),
    [anon_sym_ATencode] = ACTIONS(2714),
    [anon_sym_ATsynchronized] = ACTIONS(2714),
    [anon_sym_BOOL] = ACTIONS(2712),
    [anon_sym_IMP] = ACTIONS(2712),
    [anon_sym_SEL] = ACTIONS(2712),
    [anon_sym_Class] = ACTIONS(2712),
    [anon_sym_id] = ACTIONS(2712),
  },
  [STATE(679)] = {
    [sym_identifier] = ACTIONS(2720),
    [aux_sym_preproc_include_token1] = ACTIONS(2720),
    [aux_sym_preproc_include_token2] = ACTIONS(2720),
    [aux_sym_preproc_def_token1] = ACTIONS(2720),
    [aux_sym_preproc_if_token1] = ACTIONS(2720),
    [aux_sym_preproc_if_token2] = ACTIONS(2720),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2720),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2720),
    [sym_preproc_directive] = ACTIONS(2720),
    [anon_sym_LPAREN2] = ACTIONS(2722),
    [anon_sym_BANG] = ACTIONS(2722),
    [anon_sym_TILDE] = ACTIONS(2722),
    [anon_sym_DASH] = ACTIONS(2720),
    [anon_sym_PLUS] = ACTIONS(2720),
    [anon_sym_STAR] = ACTIONS(2722),
    [anon_sym_CARET] = ACTIONS(2722),
    [anon_sym_AMP] = ACTIONS(2722),
    [anon_sym_SEMI] = ACTIONS(2722),
    [anon_sym___extension__] = ACTIONS(2720),
    [anon_sym_typedef] = ACTIONS(2720),
    [anon_sym_extern] = ACTIONS(2720),
    [anon_sym___attribute__] = ACTIONS(2720),
    [anon_sym___attribute] = ACTIONS(2720),
    [anon_sym_noreturn] = ACTIONS(2720),
    [anon_sym_LBRACK] = ACTIONS(2722),
    [anon_sym___declspec] = ACTIONS(2720),
    [anon_sym___cdecl] = ACTIONS(2720),
    [anon_sym___clrcall] = ACTIONS(2720),
    [anon_sym___stdcall] = ACTIONS(2720),
    [anon_sym___fastcall] = ACTIONS(2720),
    [anon_sym___thiscall] = ACTIONS(2720),
    [anon_sym___vectorcall] = ACTIONS(2720),
    [anon_sym_LBRACE] = ACTIONS(2722),
    [anon_sym_signed] = ACTIONS(2720),
    [anon_sym_unsigned] = ACTIONS(2720),
    [anon_sym_long] = ACTIONS(2720),
    [anon_sym_short] = ACTIONS(2720),
    [anon_sym_static] = ACTIONS(2720),
    [anon_sym_ATautoreleasepool] = ACTIONS(2722),
    [anon_sym_auto] = ACTIONS(2720),
    [anon_sym_register] = ACTIONS(2720),
    [anon_sym_inline] = ACTIONS(2720),
    [anon_sym___inline] = ACTIONS(2720),
    [anon_sym___inline__] = ACTIONS(2720),
    [anon_sym___forceinline] = ACTIONS(2720),
    [anon_sym_thread_local] = ACTIONS(2720),
    [anon_sym___thread] = ACTIONS(2720),
    [anon_sym_CG_EXTERN] = ACTIONS(2720),
    [anon_sym_CG_INLINE] = ACTIONS(2720),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2720),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2720),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2720),
    [anon_sym_IBOutlet] = ACTIONS(2720),
    [anon_sym_IBInspectable] = ACTIONS(2720),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2720),
    [anon_sym_NS_INLINE] = ACTIONS(2720),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2720),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2720),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2720),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2720),
    [anon_sym_const] = ACTIONS(2720),
    [anon_sym_constexpr] = ACTIONS(2720),
    [anon_sym_volatile] = ACTIONS(2720),
    [anon_sym_restrict] = ACTIONS(2720),
    [anon_sym___restrict__] = ACTIONS(2720),
    [anon_sym__Atomic] = ACTIONS(2720),
    [anon_sym__Noreturn] = ACTIONS(2720),
    [anon_sym__Nonnull] = ACTIONS(2720),
    [anon_sym_nullable] = ACTIONS(2720),
    [anon_sym__Complex] = ACTIONS(2720),
    [anon_sym__Nullable] = ACTIONS(2720),
    [anon_sym__Nullable_result] = ACTIONS(2720),
    [anon_sym__Null_unspecified] = ACTIONS(2720),
    [anon_sym___autoreleasing] = ACTIONS(2720),
    [anon_sym___block] = ACTIONS(2720),
    [anon_sym___bridge] = ACTIONS(2720),
    [anon_sym___bridge_retained] = ACTIONS(2720),
    [anon_sym___bridge_transfer] = ACTIONS(2720),
    [anon_sym___complex] = ACTIONS(2720),
    [anon_sym___const] = ACTIONS(2720),
    [anon_sym___imag] = ACTIONS(2720),
    [anon_sym___kindof] = ACTIONS(2720),
    [anon_sym___nonnull] = ACTIONS(2720),
    [anon_sym___nullable] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2720),
    [anon_sym___real] = ACTIONS(2720),
    [anon_sym___strong] = ACTIONS(2720),
    [anon_sym___unsafe_unretained] = ACTIONS(2720),
    [anon_sym___unused] = ACTIONS(2720),
    [anon_sym___weak] = ACTIONS(2720),
    [anon_sym_alignas] = ACTIONS(2720),
    [anon_sym__Alignas] = ACTIONS(2720),
    [sym_primitive_type] = ACTIONS(2720),
    [anon_sym_enum] = ACTIONS(2720),
    [anon_sym_struct] = ACTIONS(2720),
    [anon_sym_union] = ACTIONS(2720),
    [anon_sym_if] = ACTIONS(2720),
    [anon_sym_switch] = ACTIONS(2720),
    [anon_sym_case] = ACTIONS(2720),
    [anon_sym_default] = ACTIONS(2720),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_do] = ACTIONS(2720),
    [anon_sym_for] = ACTIONS(2720),
    [anon_sym_in] = ACTIONS(2720),
    [anon_sym_return] = ACTIONS(2720),
    [anon_sym_break] = ACTIONS(2720),
    [anon_sym_continue] = ACTIONS(2720),
    [anon_sym_goto] = ACTIONS(2720),
    [anon_sym_DASH_DASH] = ACTIONS(2722),
    [anon_sym_PLUS_PLUS] = ACTIONS(2722),
    [anon_sym_sizeof] = ACTIONS(2720),
    [anon_sym___alignof__] = ACTIONS(2720),
    [anon_sym___alignof] = ACTIONS(2720),
    [anon_sym__alignof] = ACTIONS(2720),
    [anon_sym_alignof] = ACTIONS(2720),
    [anon_sym__Alignof] = ACTIONS(2720),
    [anon_sym_offsetof] = ACTIONS(2720),
    [anon_sym__Generic] = ACTIONS(2720),
    [anon_sym_asm] = ACTIONS(2720),
    [anon_sym___asm__] = ACTIONS(2720),
    [anon_sym___asm] = ACTIONS(2720),
    [sym_number_literal] = ACTIONS(2722),
    [anon_sym_L_SQUOTE] = ACTIONS(2722),
    [anon_sym_u_SQUOTE] = ACTIONS(2722),
    [anon_sym_U_SQUOTE] = ACTIONS(2722),
    [anon_sym_u8_SQUOTE] = ACTIONS(2722),
    [anon_sym_SQUOTE] = ACTIONS(2722),
    [anon_sym_AT] = ACTIONS(2720),
    [anon_sym_DQUOTE] = ACTIONS(2722),
    [anon_sym_L_DQUOTE] = ACTIONS(2722),
    [anon_sym_u_DQUOTE] = ACTIONS(2722),
    [anon_sym_U_DQUOTE] = ACTIONS(2722),
    [anon_sym_u8_DQUOTE] = ACTIONS(2722),
    [sym_true] = ACTIONS(2720),
    [sym_false] = ACTIONS(2720),
    [anon_sym_NULL] = ACTIONS(2720),
    [anon_sym_nullptr] = ACTIONS(2720),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2720),
    [anon_sym___typeof] = ACTIONS(2720),
    [anon_sym_typeof] = ACTIONS(2720),
    [anon_sym_ATimport] = ACTIONS(2722),
    [aux_sym_preproc_undef_token1] = ACTIONS(2720),
    [anon_sym_POUND] = ACTIONS(2720),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2720),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2720),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2720),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2720),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2720),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2720),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2720),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_API_AVAILABLE] = ACTIONS(2720),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_API_DEPRECATED] = ACTIONS(2720),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2720),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym___deprecated_msg] = ACTIONS(2720),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2720),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2720),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2720),
    [anon_sym_ATprotocol] = ACTIONS(2722),
    [anon_sym_ATinterface] = ACTIONS(2722),
    [anon_sym_ATimplementation] = ACTIONS(2722),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2722),
    [anon_sym_ATsynthesize] = ACTIONS(2722),
    [anon_sym_ATdynamic] = ACTIONS(2722),
    [anon_sym_ATtry] = ACTIONS(2722),
    [anon_sym___try] = ACTIONS(2720),
    [anon_sym_ATthrow] = ACTIONS(2722),
    [anon_sym_ATselector] = ACTIONS(2722),
    [anon_sym_ATavailable] = ACTIONS(2722),
    [anon_sym___builtin_available] = ACTIONS(2720),
    [anon_sym_va_arg] = ACTIONS(2720),
    [anon_sym_ATencode] = ACTIONS(2722),
    [anon_sym_ATsynchronized] = ACTIONS(2722),
    [anon_sym_BOOL] = ACTIONS(2720),
    [anon_sym_IMP] = ACTIONS(2720),
    [anon_sym_SEL] = ACTIONS(2720),
    [anon_sym_Class] = ACTIONS(2720),
    [anon_sym_id] = ACTIONS(2720),
  },
  [STATE(680)] = {
    [sym_identifier] = ACTIONS(2724),
    [aux_sym_preproc_include_token1] = ACTIONS(2724),
    [aux_sym_preproc_include_token2] = ACTIONS(2724),
    [aux_sym_preproc_def_token1] = ACTIONS(2724),
    [aux_sym_preproc_if_token1] = ACTIONS(2724),
    [aux_sym_preproc_if_token2] = ACTIONS(2724),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2724),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2724),
    [sym_preproc_directive] = ACTIONS(2724),
    [anon_sym_LPAREN2] = ACTIONS(2726),
    [anon_sym_BANG] = ACTIONS(2726),
    [anon_sym_TILDE] = ACTIONS(2726),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2726),
    [anon_sym_CARET] = ACTIONS(2726),
    [anon_sym_AMP] = ACTIONS(2726),
    [anon_sym_SEMI] = ACTIONS(2726),
    [anon_sym___extension__] = ACTIONS(2724),
    [anon_sym_typedef] = ACTIONS(2724),
    [anon_sym_extern] = ACTIONS(2724),
    [anon_sym___attribute__] = ACTIONS(2724),
    [anon_sym___attribute] = ACTIONS(2724),
    [anon_sym_noreturn] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2726),
    [anon_sym___declspec] = ACTIONS(2724),
    [anon_sym___cdecl] = ACTIONS(2724),
    [anon_sym___clrcall] = ACTIONS(2724),
    [anon_sym___stdcall] = ACTIONS(2724),
    [anon_sym___fastcall] = ACTIONS(2724),
    [anon_sym___thiscall] = ACTIONS(2724),
    [anon_sym___vectorcall] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2726),
    [anon_sym_signed] = ACTIONS(2724),
    [anon_sym_unsigned] = ACTIONS(2724),
    [anon_sym_long] = ACTIONS(2724),
    [anon_sym_short] = ACTIONS(2724),
    [anon_sym_static] = ACTIONS(2724),
    [anon_sym_ATautoreleasepool] = ACTIONS(2726),
    [anon_sym_auto] = ACTIONS(2724),
    [anon_sym_register] = ACTIONS(2724),
    [anon_sym_inline] = ACTIONS(2724),
    [anon_sym___inline] = ACTIONS(2724),
    [anon_sym___inline__] = ACTIONS(2724),
    [anon_sym___forceinline] = ACTIONS(2724),
    [anon_sym_thread_local] = ACTIONS(2724),
    [anon_sym___thread] = ACTIONS(2724),
    [anon_sym_CG_EXTERN] = ACTIONS(2724),
    [anon_sym_CG_INLINE] = ACTIONS(2724),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2724),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2724),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2724),
    [anon_sym_IBOutlet] = ACTIONS(2724),
    [anon_sym_IBInspectable] = ACTIONS(2724),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2724),
    [anon_sym_NS_INLINE] = ACTIONS(2724),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2724),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2724),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2724),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2724),
    [anon_sym_const] = ACTIONS(2724),
    [anon_sym_constexpr] = ACTIONS(2724),
    [anon_sym_volatile] = ACTIONS(2724),
    [anon_sym_restrict] = ACTIONS(2724),
    [anon_sym___restrict__] = ACTIONS(2724),
    [anon_sym__Atomic] = ACTIONS(2724),
    [anon_sym__Noreturn] = ACTIONS(2724),
    [anon_sym__Nonnull] = ACTIONS(2724),
    [anon_sym_nullable] = ACTIONS(2724),
    [anon_sym__Complex] = ACTIONS(2724),
    [anon_sym__Nullable] = ACTIONS(2724),
    [anon_sym__Nullable_result] = ACTIONS(2724),
    [anon_sym__Null_unspecified] = ACTIONS(2724),
    [anon_sym___autoreleasing] = ACTIONS(2724),
    [anon_sym___block] = ACTIONS(2724),
    [anon_sym___bridge] = ACTIONS(2724),
    [anon_sym___bridge_retained] = ACTIONS(2724),
    [anon_sym___bridge_transfer] = ACTIONS(2724),
    [anon_sym___complex] = ACTIONS(2724),
    [anon_sym___const] = ACTIONS(2724),
    [anon_sym___imag] = ACTIONS(2724),
    [anon_sym___kindof] = ACTIONS(2724),
    [anon_sym___nonnull] = ACTIONS(2724),
    [anon_sym___nullable] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2724),
    [anon_sym___real] = ACTIONS(2724),
    [anon_sym___strong] = ACTIONS(2724),
    [anon_sym___unsafe_unretained] = ACTIONS(2724),
    [anon_sym___unused] = ACTIONS(2724),
    [anon_sym___weak] = ACTIONS(2724),
    [anon_sym_alignas] = ACTIONS(2724),
    [anon_sym__Alignas] = ACTIONS(2724),
    [sym_primitive_type] = ACTIONS(2724),
    [anon_sym_enum] = ACTIONS(2724),
    [anon_sym_struct] = ACTIONS(2724),
    [anon_sym_union] = ACTIONS(2724),
    [anon_sym_if] = ACTIONS(2724),
    [anon_sym_switch] = ACTIONS(2724),
    [anon_sym_case] = ACTIONS(2724),
    [anon_sym_default] = ACTIONS(2724),
    [anon_sym_while] = ACTIONS(2724),
    [anon_sym_do] = ACTIONS(2724),
    [anon_sym_for] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_return] = ACTIONS(2724),
    [anon_sym_break] = ACTIONS(2724),
    [anon_sym_continue] = ACTIONS(2724),
    [anon_sym_goto] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2726),
    [anon_sym_PLUS_PLUS] = ACTIONS(2726),
    [anon_sym_sizeof] = ACTIONS(2724),
    [anon_sym___alignof__] = ACTIONS(2724),
    [anon_sym___alignof] = ACTIONS(2724),
    [anon_sym__alignof] = ACTIONS(2724),
    [anon_sym_alignof] = ACTIONS(2724),
    [anon_sym__Alignof] = ACTIONS(2724),
    [anon_sym_offsetof] = ACTIONS(2724),
    [anon_sym__Generic] = ACTIONS(2724),
    [anon_sym_asm] = ACTIONS(2724),
    [anon_sym___asm__] = ACTIONS(2724),
    [anon_sym___asm] = ACTIONS(2724),
    [sym_number_literal] = ACTIONS(2726),
    [anon_sym_L_SQUOTE] = ACTIONS(2726),
    [anon_sym_u_SQUOTE] = ACTIONS(2726),
    [anon_sym_U_SQUOTE] = ACTIONS(2726),
    [anon_sym_u8_SQUOTE] = ACTIONS(2726),
    [anon_sym_SQUOTE] = ACTIONS(2726),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2726),
    [anon_sym_L_DQUOTE] = ACTIONS(2726),
    [anon_sym_u_DQUOTE] = ACTIONS(2726),
    [anon_sym_U_DQUOTE] = ACTIONS(2726),
    [anon_sym_u8_DQUOTE] = ACTIONS(2726),
    [sym_true] = ACTIONS(2724),
    [sym_false] = ACTIONS(2724),
    [anon_sym_NULL] = ACTIONS(2724),
    [anon_sym_nullptr] = ACTIONS(2724),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2724),
    [anon_sym___typeof] = ACTIONS(2724),
    [anon_sym_typeof] = ACTIONS(2724),
    [anon_sym_ATimport] = ACTIONS(2726),
    [aux_sym_preproc_undef_token1] = ACTIONS(2724),
    [anon_sym_POUND] = ACTIONS(2724),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2724),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2724),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2724),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2724),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2724),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2724),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2724),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_API_AVAILABLE] = ACTIONS(2724),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_API_DEPRECATED] = ACTIONS(2724),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2724),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym___deprecated_msg] = ACTIONS(2724),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2724),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2724),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2724),
    [anon_sym_ATprotocol] = ACTIONS(2726),
    [anon_sym_ATinterface] = ACTIONS(2726),
    [anon_sym_ATimplementation] = ACTIONS(2726),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2726),
    [anon_sym_ATsynthesize] = ACTIONS(2726),
    [anon_sym_ATdynamic] = ACTIONS(2726),
    [anon_sym_ATtry] = ACTIONS(2726),
    [anon_sym___try] = ACTIONS(2724),
    [anon_sym_ATthrow] = ACTIONS(2726),
    [anon_sym_ATselector] = ACTIONS(2726),
    [anon_sym_ATavailable] = ACTIONS(2726),
    [anon_sym___builtin_available] = ACTIONS(2724),
    [anon_sym_va_arg] = ACTIONS(2724),
    [anon_sym_ATencode] = ACTIONS(2726),
    [anon_sym_ATsynchronized] = ACTIONS(2726),
    [anon_sym_BOOL] = ACTIONS(2724),
    [anon_sym_IMP] = ACTIONS(2724),
    [anon_sym_SEL] = ACTIONS(2724),
    [anon_sym_Class] = ACTIONS(2724),
    [anon_sym_id] = ACTIONS(2724),
  },
  [STATE(681)] = {
    [sym_identifier] = ACTIONS(2728),
    [aux_sym_preproc_include_token1] = ACTIONS(2728),
    [aux_sym_preproc_include_token2] = ACTIONS(2728),
    [aux_sym_preproc_def_token1] = ACTIONS(2728),
    [aux_sym_preproc_if_token1] = ACTIONS(2728),
    [aux_sym_preproc_if_token2] = ACTIONS(2728),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2728),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2728),
    [sym_preproc_directive] = ACTIONS(2728),
    [anon_sym_LPAREN2] = ACTIONS(2730),
    [anon_sym_BANG] = ACTIONS(2730),
    [anon_sym_TILDE] = ACTIONS(2730),
    [anon_sym_DASH] = ACTIONS(2728),
    [anon_sym_PLUS] = ACTIONS(2728),
    [anon_sym_STAR] = ACTIONS(2730),
    [anon_sym_CARET] = ACTIONS(2730),
    [anon_sym_AMP] = ACTIONS(2730),
    [anon_sym_SEMI] = ACTIONS(2730),
    [anon_sym___extension__] = ACTIONS(2728),
    [anon_sym_typedef] = ACTIONS(2728),
    [anon_sym_extern] = ACTIONS(2728),
    [anon_sym___attribute__] = ACTIONS(2728),
    [anon_sym___attribute] = ACTIONS(2728),
    [anon_sym_noreturn] = ACTIONS(2728),
    [anon_sym_LBRACK] = ACTIONS(2730),
    [anon_sym___declspec] = ACTIONS(2728),
    [anon_sym___cdecl] = ACTIONS(2728),
    [anon_sym___clrcall] = ACTIONS(2728),
    [anon_sym___stdcall] = ACTIONS(2728),
    [anon_sym___fastcall] = ACTIONS(2728),
    [anon_sym___thiscall] = ACTIONS(2728),
    [anon_sym___vectorcall] = ACTIONS(2728),
    [anon_sym_LBRACE] = ACTIONS(2730),
    [anon_sym_signed] = ACTIONS(2728),
    [anon_sym_unsigned] = ACTIONS(2728),
    [anon_sym_long] = ACTIONS(2728),
    [anon_sym_short] = ACTIONS(2728),
    [anon_sym_static] = ACTIONS(2728),
    [anon_sym_ATautoreleasepool] = ACTIONS(2730),
    [anon_sym_auto] = ACTIONS(2728),
    [anon_sym_register] = ACTIONS(2728),
    [anon_sym_inline] = ACTIONS(2728),
    [anon_sym___inline] = ACTIONS(2728),
    [anon_sym___inline__] = ACTIONS(2728),
    [anon_sym___forceinline] = ACTIONS(2728),
    [anon_sym_thread_local] = ACTIONS(2728),
    [anon_sym___thread] = ACTIONS(2728),
    [anon_sym_CG_EXTERN] = ACTIONS(2728),
    [anon_sym_CG_INLINE] = ACTIONS(2728),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2728),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2728),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2728),
    [anon_sym_IBOutlet] = ACTIONS(2728),
    [anon_sym_IBInspectable] = ACTIONS(2728),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2728),
    [anon_sym_NS_INLINE] = ACTIONS(2728),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2728),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2728),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2728),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2728),
    [anon_sym_const] = ACTIONS(2728),
    [anon_sym_constexpr] = ACTIONS(2728),
    [anon_sym_volatile] = ACTIONS(2728),
    [anon_sym_restrict] = ACTIONS(2728),
    [anon_sym___restrict__] = ACTIONS(2728),
    [anon_sym__Atomic] = ACTIONS(2728),
    [anon_sym__Noreturn] = ACTIONS(2728),
    [anon_sym__Nonnull] = ACTIONS(2728),
    [anon_sym_nullable] = ACTIONS(2728),
    [anon_sym__Complex] = ACTIONS(2728),
    [anon_sym__Nullable] = ACTIONS(2728),
    [anon_sym__Nullable_result] = ACTIONS(2728),
    [anon_sym__Null_unspecified] = ACTIONS(2728),
    [anon_sym___autoreleasing] = ACTIONS(2728),
    [anon_sym___block] = ACTIONS(2728),
    [anon_sym___bridge] = ACTIONS(2728),
    [anon_sym___bridge_retained] = ACTIONS(2728),
    [anon_sym___bridge_transfer] = ACTIONS(2728),
    [anon_sym___complex] = ACTIONS(2728),
    [anon_sym___const] = ACTIONS(2728),
    [anon_sym___imag] = ACTIONS(2728),
    [anon_sym___kindof] = ACTIONS(2728),
    [anon_sym___nonnull] = ACTIONS(2728),
    [anon_sym___nullable] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2728),
    [anon_sym___real] = ACTIONS(2728),
    [anon_sym___strong] = ACTIONS(2728),
    [anon_sym___unsafe_unretained] = ACTIONS(2728),
    [anon_sym___unused] = ACTIONS(2728),
    [anon_sym___weak] = ACTIONS(2728),
    [anon_sym_alignas] = ACTIONS(2728),
    [anon_sym__Alignas] = ACTIONS(2728),
    [sym_primitive_type] = ACTIONS(2728),
    [anon_sym_enum] = ACTIONS(2728),
    [anon_sym_struct] = ACTIONS(2728),
    [anon_sym_union] = ACTIONS(2728),
    [anon_sym_if] = ACTIONS(2728),
    [anon_sym_switch] = ACTIONS(2728),
    [anon_sym_case] = ACTIONS(2728),
    [anon_sym_default] = ACTIONS(2728),
    [anon_sym_while] = ACTIONS(2728),
    [anon_sym_do] = ACTIONS(2728),
    [anon_sym_for] = ACTIONS(2728),
    [anon_sym_in] = ACTIONS(2728),
    [anon_sym_return] = ACTIONS(2728),
    [anon_sym_break] = ACTIONS(2728),
    [anon_sym_continue] = ACTIONS(2728),
    [anon_sym_goto] = ACTIONS(2728),
    [anon_sym_DASH_DASH] = ACTIONS(2730),
    [anon_sym_PLUS_PLUS] = ACTIONS(2730),
    [anon_sym_sizeof] = ACTIONS(2728),
    [anon_sym___alignof__] = ACTIONS(2728),
    [anon_sym___alignof] = ACTIONS(2728),
    [anon_sym__alignof] = ACTIONS(2728),
    [anon_sym_alignof] = ACTIONS(2728),
    [anon_sym__Alignof] = ACTIONS(2728),
    [anon_sym_offsetof] = ACTIONS(2728),
    [anon_sym__Generic] = ACTIONS(2728),
    [anon_sym_asm] = ACTIONS(2728),
    [anon_sym___asm__] = ACTIONS(2728),
    [anon_sym___asm] = ACTIONS(2728),
    [sym_number_literal] = ACTIONS(2730),
    [anon_sym_L_SQUOTE] = ACTIONS(2730),
    [anon_sym_u_SQUOTE] = ACTIONS(2730),
    [anon_sym_U_SQUOTE] = ACTIONS(2730),
    [anon_sym_u8_SQUOTE] = ACTIONS(2730),
    [anon_sym_SQUOTE] = ACTIONS(2730),
    [anon_sym_AT] = ACTIONS(2728),
    [anon_sym_DQUOTE] = ACTIONS(2730),
    [anon_sym_L_DQUOTE] = ACTIONS(2730),
    [anon_sym_u_DQUOTE] = ACTIONS(2730),
    [anon_sym_U_DQUOTE] = ACTIONS(2730),
    [anon_sym_u8_DQUOTE] = ACTIONS(2730),
    [sym_true] = ACTIONS(2728),
    [sym_false] = ACTIONS(2728),
    [anon_sym_NULL] = ACTIONS(2728),
    [anon_sym_nullptr] = ACTIONS(2728),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2728),
    [anon_sym___typeof] = ACTIONS(2728),
    [anon_sym_typeof] = ACTIONS(2728),
    [anon_sym_ATimport] = ACTIONS(2730),
    [aux_sym_preproc_undef_token1] = ACTIONS(2728),
    [anon_sym_POUND] = ACTIONS(2728),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2728),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2728),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2728),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2728),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2728),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2728),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2728),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_API_AVAILABLE] = ACTIONS(2728),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_API_DEPRECATED] = ACTIONS(2728),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2728),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym___deprecated_msg] = ACTIONS(2728),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2728),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2728),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2728),
    [anon_sym_ATprotocol] = ACTIONS(2730),
    [anon_sym_ATinterface] = ACTIONS(2730),
    [anon_sym_ATimplementation] = ACTIONS(2730),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2730),
    [anon_sym_ATsynthesize] = ACTIONS(2730),
    [anon_sym_ATdynamic] = ACTIONS(2730),
    [anon_sym_ATtry] = ACTIONS(2730),
    [anon_sym___try] = ACTIONS(2728),
    [anon_sym_ATthrow] = ACTIONS(2730),
    [anon_sym_ATselector] = ACTIONS(2730),
    [anon_sym_ATavailable] = ACTIONS(2730),
    [anon_sym___builtin_available] = ACTIONS(2728),
    [anon_sym_va_arg] = ACTIONS(2728),
    [anon_sym_ATencode] = ACTIONS(2730),
    [anon_sym_ATsynchronized] = ACTIONS(2730),
    [anon_sym_BOOL] = ACTIONS(2728),
    [anon_sym_IMP] = ACTIONS(2728),
    [anon_sym_SEL] = ACTIONS(2728),
    [anon_sym_Class] = ACTIONS(2728),
    [anon_sym_id] = ACTIONS(2728),
  },
  [STATE(682)] = {
    [sym_identifier] = ACTIONS(2732),
    [aux_sym_preproc_include_token1] = ACTIONS(2732),
    [aux_sym_preproc_include_token2] = ACTIONS(2732),
    [aux_sym_preproc_def_token1] = ACTIONS(2732),
    [aux_sym_preproc_if_token1] = ACTIONS(2732),
    [aux_sym_preproc_if_token2] = ACTIONS(2732),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2732),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2732),
    [sym_preproc_directive] = ACTIONS(2732),
    [anon_sym_LPAREN2] = ACTIONS(2734),
    [anon_sym_BANG] = ACTIONS(2734),
    [anon_sym_TILDE] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2732),
    [anon_sym_PLUS] = ACTIONS(2732),
    [anon_sym_STAR] = ACTIONS(2734),
    [anon_sym_CARET] = ACTIONS(2734),
    [anon_sym_AMP] = ACTIONS(2734),
    [anon_sym_SEMI] = ACTIONS(2734),
    [anon_sym___extension__] = ACTIONS(2732),
    [anon_sym_typedef] = ACTIONS(2732),
    [anon_sym_extern] = ACTIONS(2732),
    [anon_sym___attribute__] = ACTIONS(2732),
    [anon_sym___attribute] = ACTIONS(2732),
    [anon_sym_noreturn] = ACTIONS(2732),
    [anon_sym_LBRACK] = ACTIONS(2734),
    [anon_sym___declspec] = ACTIONS(2732),
    [anon_sym___cdecl] = ACTIONS(2732),
    [anon_sym___clrcall] = ACTIONS(2732),
    [anon_sym___stdcall] = ACTIONS(2732),
    [anon_sym___fastcall] = ACTIONS(2732),
    [anon_sym___thiscall] = ACTIONS(2732),
    [anon_sym___vectorcall] = ACTIONS(2732),
    [anon_sym_LBRACE] = ACTIONS(2734),
    [anon_sym_signed] = ACTIONS(2732),
    [anon_sym_unsigned] = ACTIONS(2732),
    [anon_sym_long] = ACTIONS(2732),
    [anon_sym_short] = ACTIONS(2732),
    [anon_sym_static] = ACTIONS(2732),
    [anon_sym_ATautoreleasepool] = ACTIONS(2734),
    [anon_sym_auto] = ACTIONS(2732),
    [anon_sym_register] = ACTIONS(2732),
    [anon_sym_inline] = ACTIONS(2732),
    [anon_sym___inline] = ACTIONS(2732),
    [anon_sym___inline__] = ACTIONS(2732),
    [anon_sym___forceinline] = ACTIONS(2732),
    [anon_sym_thread_local] = ACTIONS(2732),
    [anon_sym___thread] = ACTIONS(2732),
    [anon_sym_CG_EXTERN] = ACTIONS(2732),
    [anon_sym_CG_INLINE] = ACTIONS(2732),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2732),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2732),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2732),
    [anon_sym_IBOutlet] = ACTIONS(2732),
    [anon_sym_IBInspectable] = ACTIONS(2732),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2732),
    [anon_sym_NS_INLINE] = ACTIONS(2732),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2732),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2732),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2732),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2732),
    [anon_sym_const] = ACTIONS(2732),
    [anon_sym_constexpr] = ACTIONS(2732),
    [anon_sym_volatile] = ACTIONS(2732),
    [anon_sym_restrict] = ACTIONS(2732),
    [anon_sym___restrict__] = ACTIONS(2732),
    [anon_sym__Atomic] = ACTIONS(2732),
    [anon_sym__Noreturn] = ACTIONS(2732),
    [anon_sym__Nonnull] = ACTIONS(2732),
    [anon_sym_nullable] = ACTIONS(2732),
    [anon_sym__Complex] = ACTIONS(2732),
    [anon_sym__Nullable] = ACTIONS(2732),
    [anon_sym__Nullable_result] = ACTIONS(2732),
    [anon_sym__Null_unspecified] = ACTIONS(2732),
    [anon_sym___autoreleasing] = ACTIONS(2732),
    [anon_sym___block] = ACTIONS(2732),
    [anon_sym___bridge] = ACTIONS(2732),
    [anon_sym___bridge_retained] = ACTIONS(2732),
    [anon_sym___bridge_transfer] = ACTIONS(2732),
    [anon_sym___complex] = ACTIONS(2732),
    [anon_sym___const] = ACTIONS(2732),
    [anon_sym___imag] = ACTIONS(2732),
    [anon_sym___kindof] = ACTIONS(2732),
    [anon_sym___nonnull] = ACTIONS(2732),
    [anon_sym___nullable] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2732),
    [anon_sym___real] = ACTIONS(2732),
    [anon_sym___strong] = ACTIONS(2732),
    [anon_sym___unsafe_unretained] = ACTIONS(2732),
    [anon_sym___unused] = ACTIONS(2732),
    [anon_sym___weak] = ACTIONS(2732),
    [anon_sym_alignas] = ACTIONS(2732),
    [anon_sym__Alignas] = ACTIONS(2732),
    [sym_primitive_type] = ACTIONS(2732),
    [anon_sym_enum] = ACTIONS(2732),
    [anon_sym_struct] = ACTIONS(2732),
    [anon_sym_union] = ACTIONS(2732),
    [anon_sym_if] = ACTIONS(2732),
    [anon_sym_switch] = ACTIONS(2732),
    [anon_sym_case] = ACTIONS(2732),
    [anon_sym_default] = ACTIONS(2732),
    [anon_sym_while] = ACTIONS(2732),
    [anon_sym_do] = ACTIONS(2732),
    [anon_sym_for] = ACTIONS(2732),
    [anon_sym_in] = ACTIONS(2732),
    [anon_sym_return] = ACTIONS(2732),
    [anon_sym_break] = ACTIONS(2732),
    [anon_sym_continue] = ACTIONS(2732),
    [anon_sym_goto] = ACTIONS(2732),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_sizeof] = ACTIONS(2732),
    [anon_sym___alignof__] = ACTIONS(2732),
    [anon_sym___alignof] = ACTIONS(2732),
    [anon_sym__alignof] = ACTIONS(2732),
    [anon_sym_alignof] = ACTIONS(2732),
    [anon_sym__Alignof] = ACTIONS(2732),
    [anon_sym_offsetof] = ACTIONS(2732),
    [anon_sym__Generic] = ACTIONS(2732),
    [anon_sym_asm] = ACTIONS(2732),
    [anon_sym___asm__] = ACTIONS(2732),
    [anon_sym___asm] = ACTIONS(2732),
    [sym_number_literal] = ACTIONS(2734),
    [anon_sym_L_SQUOTE] = ACTIONS(2734),
    [anon_sym_u_SQUOTE] = ACTIONS(2734),
    [anon_sym_U_SQUOTE] = ACTIONS(2734),
    [anon_sym_u8_SQUOTE] = ACTIONS(2734),
    [anon_sym_SQUOTE] = ACTIONS(2734),
    [anon_sym_AT] = ACTIONS(2732),
    [anon_sym_DQUOTE] = ACTIONS(2734),
    [anon_sym_L_DQUOTE] = ACTIONS(2734),
    [anon_sym_u_DQUOTE] = ACTIONS(2734),
    [anon_sym_U_DQUOTE] = ACTIONS(2734),
    [anon_sym_u8_DQUOTE] = ACTIONS(2734),
    [sym_true] = ACTIONS(2732),
    [sym_false] = ACTIONS(2732),
    [anon_sym_NULL] = ACTIONS(2732),
    [anon_sym_nullptr] = ACTIONS(2732),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2732),
    [anon_sym___typeof] = ACTIONS(2732),
    [anon_sym_typeof] = ACTIONS(2732),
    [anon_sym_ATimport] = ACTIONS(2734),
    [aux_sym_preproc_undef_token1] = ACTIONS(2732),
    [anon_sym_POUND] = ACTIONS(2732),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2732),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2732),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2732),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2732),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2732),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2732),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2732),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_API_AVAILABLE] = ACTIONS(2732),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_API_DEPRECATED] = ACTIONS(2732),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2732),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym___deprecated_msg] = ACTIONS(2732),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2732),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2732),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2732),
    [anon_sym_ATprotocol] = ACTIONS(2734),
    [anon_sym_ATinterface] = ACTIONS(2734),
    [anon_sym_ATimplementation] = ACTIONS(2734),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2734),
    [anon_sym_ATsynthesize] = ACTIONS(2734),
    [anon_sym_ATdynamic] = ACTIONS(2734),
    [anon_sym_ATtry] = ACTIONS(2734),
    [anon_sym___try] = ACTIONS(2732),
    [anon_sym_ATthrow] = ACTIONS(2734),
    [anon_sym_ATselector] = ACTIONS(2734),
    [anon_sym_ATavailable] = ACTIONS(2734),
    [anon_sym___builtin_available] = ACTIONS(2732),
    [anon_sym_va_arg] = ACTIONS(2732),
    [anon_sym_ATencode] = ACTIONS(2734),
    [anon_sym_ATsynchronized] = ACTIONS(2734),
    [anon_sym_BOOL] = ACTIONS(2732),
    [anon_sym_IMP] = ACTIONS(2732),
    [anon_sym_SEL] = ACTIONS(2732),
    [anon_sym_Class] = ACTIONS(2732),
    [anon_sym_id] = ACTIONS(2732),
  },
  [STATE(683)] = {
    [sym_identifier] = ACTIONS(2736),
    [aux_sym_preproc_include_token1] = ACTIONS(2736),
    [aux_sym_preproc_include_token2] = ACTIONS(2736),
    [aux_sym_preproc_def_token1] = ACTIONS(2736),
    [aux_sym_preproc_if_token1] = ACTIONS(2736),
    [aux_sym_preproc_if_token2] = ACTIONS(2736),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2736),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2736),
    [sym_preproc_directive] = ACTIONS(2736),
    [anon_sym_LPAREN2] = ACTIONS(2738),
    [anon_sym_BANG] = ACTIONS(2738),
    [anon_sym_TILDE] = ACTIONS(2738),
    [anon_sym_DASH] = ACTIONS(2736),
    [anon_sym_PLUS] = ACTIONS(2736),
    [anon_sym_STAR] = ACTIONS(2738),
    [anon_sym_CARET] = ACTIONS(2738),
    [anon_sym_AMP] = ACTIONS(2738),
    [anon_sym_SEMI] = ACTIONS(2738),
    [anon_sym___extension__] = ACTIONS(2736),
    [anon_sym_typedef] = ACTIONS(2736),
    [anon_sym_extern] = ACTIONS(2736),
    [anon_sym___attribute__] = ACTIONS(2736),
    [anon_sym___attribute] = ACTIONS(2736),
    [anon_sym_noreturn] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(2738),
    [anon_sym___declspec] = ACTIONS(2736),
    [anon_sym___cdecl] = ACTIONS(2736),
    [anon_sym___clrcall] = ACTIONS(2736),
    [anon_sym___stdcall] = ACTIONS(2736),
    [anon_sym___fastcall] = ACTIONS(2736),
    [anon_sym___thiscall] = ACTIONS(2736),
    [anon_sym___vectorcall] = ACTIONS(2736),
    [anon_sym_LBRACE] = ACTIONS(2738),
    [anon_sym_signed] = ACTIONS(2736),
    [anon_sym_unsigned] = ACTIONS(2736),
    [anon_sym_long] = ACTIONS(2736),
    [anon_sym_short] = ACTIONS(2736),
    [anon_sym_static] = ACTIONS(2736),
    [anon_sym_ATautoreleasepool] = ACTIONS(2738),
    [anon_sym_auto] = ACTIONS(2736),
    [anon_sym_register] = ACTIONS(2736),
    [anon_sym_inline] = ACTIONS(2736),
    [anon_sym___inline] = ACTIONS(2736),
    [anon_sym___inline__] = ACTIONS(2736),
    [anon_sym___forceinline] = ACTIONS(2736),
    [anon_sym_thread_local] = ACTIONS(2736),
    [anon_sym___thread] = ACTIONS(2736),
    [anon_sym_CG_EXTERN] = ACTIONS(2736),
    [anon_sym_CG_INLINE] = ACTIONS(2736),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2736),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2736),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2736),
    [anon_sym_IBOutlet] = ACTIONS(2736),
    [anon_sym_IBInspectable] = ACTIONS(2736),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2736),
    [anon_sym_NS_INLINE] = ACTIONS(2736),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2736),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2736),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2736),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2736),
    [anon_sym_const] = ACTIONS(2736),
    [anon_sym_constexpr] = ACTIONS(2736),
    [anon_sym_volatile] = ACTIONS(2736),
    [anon_sym_restrict] = ACTIONS(2736),
    [anon_sym___restrict__] = ACTIONS(2736),
    [anon_sym__Atomic] = ACTIONS(2736),
    [anon_sym__Noreturn] = ACTIONS(2736),
    [anon_sym__Nonnull] = ACTIONS(2736),
    [anon_sym_nullable] = ACTIONS(2736),
    [anon_sym__Complex] = ACTIONS(2736),
    [anon_sym__Nullable] = ACTIONS(2736),
    [anon_sym__Nullable_result] = ACTIONS(2736),
    [anon_sym__Null_unspecified] = ACTIONS(2736),
    [anon_sym___autoreleasing] = ACTIONS(2736),
    [anon_sym___block] = ACTIONS(2736),
    [anon_sym___bridge] = ACTIONS(2736),
    [anon_sym___bridge_retained] = ACTIONS(2736),
    [anon_sym___bridge_transfer] = ACTIONS(2736),
    [anon_sym___complex] = ACTIONS(2736),
    [anon_sym___const] = ACTIONS(2736),
    [anon_sym___imag] = ACTIONS(2736),
    [anon_sym___kindof] = ACTIONS(2736),
    [anon_sym___nonnull] = ACTIONS(2736),
    [anon_sym___nullable] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2736),
    [anon_sym___real] = ACTIONS(2736),
    [anon_sym___strong] = ACTIONS(2736),
    [anon_sym___unsafe_unretained] = ACTIONS(2736),
    [anon_sym___unused] = ACTIONS(2736),
    [anon_sym___weak] = ACTIONS(2736),
    [anon_sym_alignas] = ACTIONS(2736),
    [anon_sym__Alignas] = ACTIONS(2736),
    [sym_primitive_type] = ACTIONS(2736),
    [anon_sym_enum] = ACTIONS(2736),
    [anon_sym_struct] = ACTIONS(2736),
    [anon_sym_union] = ACTIONS(2736),
    [anon_sym_if] = ACTIONS(2736),
    [anon_sym_switch] = ACTIONS(2736),
    [anon_sym_case] = ACTIONS(2736),
    [anon_sym_default] = ACTIONS(2736),
    [anon_sym_while] = ACTIONS(2736),
    [anon_sym_do] = ACTIONS(2736),
    [anon_sym_for] = ACTIONS(2736),
    [anon_sym_in] = ACTIONS(2736),
    [anon_sym_return] = ACTIONS(2736),
    [anon_sym_break] = ACTIONS(2736),
    [anon_sym_continue] = ACTIONS(2736),
    [anon_sym_goto] = ACTIONS(2736),
    [anon_sym_DASH_DASH] = ACTIONS(2738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2738),
    [anon_sym_sizeof] = ACTIONS(2736),
    [anon_sym___alignof__] = ACTIONS(2736),
    [anon_sym___alignof] = ACTIONS(2736),
    [anon_sym__alignof] = ACTIONS(2736),
    [anon_sym_alignof] = ACTIONS(2736),
    [anon_sym__Alignof] = ACTIONS(2736),
    [anon_sym_offsetof] = ACTIONS(2736),
    [anon_sym__Generic] = ACTIONS(2736),
    [anon_sym_asm] = ACTIONS(2736),
    [anon_sym___asm__] = ACTIONS(2736),
    [anon_sym___asm] = ACTIONS(2736),
    [sym_number_literal] = ACTIONS(2738),
    [anon_sym_L_SQUOTE] = ACTIONS(2738),
    [anon_sym_u_SQUOTE] = ACTIONS(2738),
    [anon_sym_U_SQUOTE] = ACTIONS(2738),
    [anon_sym_u8_SQUOTE] = ACTIONS(2738),
    [anon_sym_SQUOTE] = ACTIONS(2738),
    [anon_sym_AT] = ACTIONS(2736),
    [anon_sym_DQUOTE] = ACTIONS(2738),
    [anon_sym_L_DQUOTE] = ACTIONS(2738),
    [anon_sym_u_DQUOTE] = ACTIONS(2738),
    [anon_sym_U_DQUOTE] = ACTIONS(2738),
    [anon_sym_u8_DQUOTE] = ACTIONS(2738),
    [sym_true] = ACTIONS(2736),
    [sym_false] = ACTIONS(2736),
    [anon_sym_NULL] = ACTIONS(2736),
    [anon_sym_nullptr] = ACTIONS(2736),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2736),
    [anon_sym___typeof] = ACTIONS(2736),
    [anon_sym_typeof] = ACTIONS(2736),
    [anon_sym_ATimport] = ACTIONS(2738),
    [aux_sym_preproc_undef_token1] = ACTIONS(2736),
    [anon_sym_POUND] = ACTIONS(2736),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2736),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2736),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2736),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2736),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2736),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2736),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2736),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_API_AVAILABLE] = ACTIONS(2736),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_API_DEPRECATED] = ACTIONS(2736),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2736),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym___deprecated_msg] = ACTIONS(2736),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2736),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2736),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2736),
    [anon_sym_ATprotocol] = ACTIONS(2738),
    [anon_sym_ATinterface] = ACTIONS(2738),
    [anon_sym_ATimplementation] = ACTIONS(2738),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2738),
    [anon_sym_ATsynthesize] = ACTIONS(2738),
    [anon_sym_ATdynamic] = ACTIONS(2738),
    [anon_sym_ATtry] = ACTIONS(2738),
    [anon_sym___try] = ACTIONS(2736),
    [anon_sym_ATthrow] = ACTIONS(2738),
    [anon_sym_ATselector] = ACTIONS(2738),
    [anon_sym_ATavailable] = ACTIONS(2738),
    [anon_sym___builtin_available] = ACTIONS(2736),
    [anon_sym_va_arg] = ACTIONS(2736),
    [anon_sym_ATencode] = ACTIONS(2738),
    [anon_sym_ATsynchronized] = ACTIONS(2738),
    [anon_sym_BOOL] = ACTIONS(2736),
    [anon_sym_IMP] = ACTIONS(2736),
    [anon_sym_SEL] = ACTIONS(2736),
    [anon_sym_Class] = ACTIONS(2736),
    [anon_sym_id] = ACTIONS(2736),
  },
  [STATE(684)] = {
    [sym_identifier] = ACTIONS(2688),
    [aux_sym_preproc_include_token1] = ACTIONS(2688),
    [aux_sym_preproc_include_token2] = ACTIONS(2688),
    [aux_sym_preproc_def_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token2] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2688),
    [sym_preproc_directive] = ACTIONS(2688),
    [anon_sym_LPAREN2] = ACTIONS(2690),
    [anon_sym_BANG] = ACTIONS(2690),
    [anon_sym_TILDE] = ACTIONS(2690),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2690),
    [anon_sym_CARET] = ACTIONS(2690),
    [anon_sym_AMP] = ACTIONS(2690),
    [anon_sym_SEMI] = ACTIONS(2690),
    [anon_sym___extension__] = ACTIONS(2688),
    [anon_sym_typedef] = ACTIONS(2688),
    [anon_sym_extern] = ACTIONS(2688),
    [anon_sym___attribute__] = ACTIONS(2688),
    [anon_sym___attribute] = ACTIONS(2688),
    [anon_sym_noreturn] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym___declspec] = ACTIONS(2688),
    [anon_sym___cdecl] = ACTIONS(2688),
    [anon_sym___clrcall] = ACTIONS(2688),
    [anon_sym___stdcall] = ACTIONS(2688),
    [anon_sym___fastcall] = ACTIONS(2688),
    [anon_sym___thiscall] = ACTIONS(2688),
    [anon_sym___vectorcall] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_signed] = ACTIONS(2688),
    [anon_sym_unsigned] = ACTIONS(2688),
    [anon_sym_long] = ACTIONS(2688),
    [anon_sym_short] = ACTIONS(2688),
    [anon_sym_static] = ACTIONS(2688),
    [anon_sym_ATautoreleasepool] = ACTIONS(2690),
    [anon_sym_auto] = ACTIONS(2688),
    [anon_sym_register] = ACTIONS(2688),
    [anon_sym_inline] = ACTIONS(2688),
    [anon_sym___inline] = ACTIONS(2688),
    [anon_sym___inline__] = ACTIONS(2688),
    [anon_sym___forceinline] = ACTIONS(2688),
    [anon_sym_thread_local] = ACTIONS(2688),
    [anon_sym___thread] = ACTIONS(2688),
    [anon_sym_CG_EXTERN] = ACTIONS(2688),
    [anon_sym_CG_INLINE] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2688),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2688),
    [anon_sym_IBOutlet] = ACTIONS(2688),
    [anon_sym_IBInspectable] = ACTIONS(2688),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2688),
    [anon_sym_NS_INLINE] = ACTIONS(2688),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2688),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2688),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2688),
    [anon_sym_const] = ACTIONS(2688),
    [anon_sym_constexpr] = ACTIONS(2688),
    [anon_sym_volatile] = ACTIONS(2688),
    [anon_sym_restrict] = ACTIONS(2688),
    [anon_sym___restrict__] = ACTIONS(2688),
    [anon_sym__Atomic] = ACTIONS(2688),
    [anon_sym__Noreturn] = ACTIONS(2688),
    [anon_sym__Nonnull] = ACTIONS(2688),
    [anon_sym_nullable] = ACTIONS(2688),
    [anon_sym__Complex] = ACTIONS(2688),
    [anon_sym__Nullable] = ACTIONS(2688),
    [anon_sym__Nullable_result] = ACTIONS(2688),
    [anon_sym__Null_unspecified] = ACTIONS(2688),
    [anon_sym___autoreleasing] = ACTIONS(2688),
    [anon_sym___block] = ACTIONS(2688),
    [anon_sym___bridge] = ACTIONS(2688),
    [anon_sym___bridge_retained] = ACTIONS(2688),
    [anon_sym___bridge_transfer] = ACTIONS(2688),
    [anon_sym___complex] = ACTIONS(2688),
    [anon_sym___const] = ACTIONS(2688),
    [anon_sym___imag] = ACTIONS(2688),
    [anon_sym___kindof] = ACTIONS(2688),
    [anon_sym___nonnull] = ACTIONS(2688),
    [anon_sym___nullable] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2688),
    [anon_sym___real] = ACTIONS(2688),
    [anon_sym___strong] = ACTIONS(2688),
    [anon_sym___unsafe_unretained] = ACTIONS(2688),
    [anon_sym___unused] = ACTIONS(2688),
    [anon_sym___weak] = ACTIONS(2688),
    [anon_sym_alignas] = ACTIONS(2688),
    [anon_sym__Alignas] = ACTIONS(2688),
    [sym_primitive_type] = ACTIONS(2688),
    [anon_sym_enum] = ACTIONS(2688),
    [anon_sym_struct] = ACTIONS(2688),
    [anon_sym_union] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_switch] = ACTIONS(2688),
    [anon_sym_case] = ACTIONS(2688),
    [anon_sym_default] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_in] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_sizeof] = ACTIONS(2688),
    [anon_sym___alignof__] = ACTIONS(2688),
    [anon_sym___alignof] = ACTIONS(2688),
    [anon_sym__alignof] = ACTIONS(2688),
    [anon_sym_alignof] = ACTIONS(2688),
    [anon_sym__Alignof] = ACTIONS(2688),
    [anon_sym_offsetof] = ACTIONS(2688),
    [anon_sym__Generic] = ACTIONS(2688),
    [anon_sym_asm] = ACTIONS(2688),
    [anon_sym___asm__] = ACTIONS(2688),
    [anon_sym___asm] = ACTIONS(2688),
    [sym_number_literal] = ACTIONS(2690),
    [anon_sym_L_SQUOTE] = ACTIONS(2690),
    [anon_sym_u_SQUOTE] = ACTIONS(2690),
    [anon_sym_U_SQUOTE] = ACTIONS(2690),
    [anon_sym_u8_SQUOTE] = ACTIONS(2690),
    [anon_sym_SQUOTE] = ACTIONS(2690),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_DQUOTE] = ACTIONS(2690),
    [anon_sym_L_DQUOTE] = ACTIONS(2690),
    [anon_sym_u_DQUOTE] = ACTIONS(2690),
    [anon_sym_U_DQUOTE] = ACTIONS(2690),
    [anon_sym_u8_DQUOTE] = ACTIONS(2690),
    [sym_true] = ACTIONS(2688),
    [sym_false] = ACTIONS(2688),
    [anon_sym_NULL] = ACTIONS(2688),
    [anon_sym_nullptr] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2688),
    [anon_sym___typeof] = ACTIONS(2688),
    [anon_sym_typeof] = ACTIONS(2688),
    [anon_sym_ATimport] = ACTIONS(2690),
    [aux_sym_preproc_undef_token1] = ACTIONS(2688),
    [anon_sym_POUND] = ACTIONS(2688),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2688),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2688),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2688),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_API_AVAILABLE] = ACTIONS(2688),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_API_DEPRECATED] = ACTIONS(2688),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym___deprecated_msg] = ACTIONS(2688),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2688),
    [anon_sym_ATprotocol] = ACTIONS(2690),
    [anon_sym_ATinterface] = ACTIONS(2690),
    [anon_sym_ATimplementation] = ACTIONS(2690),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2690),
    [anon_sym_ATsynthesize] = ACTIONS(2690),
    [anon_sym_ATdynamic] = ACTIONS(2690),
    [anon_sym_ATtry] = ACTIONS(2690),
    [anon_sym___try] = ACTIONS(2688),
    [anon_sym_ATthrow] = ACTIONS(2690),
    [anon_sym_ATselector] = ACTIONS(2690),
    [anon_sym_ATavailable] = ACTIONS(2690),
    [anon_sym___builtin_available] = ACTIONS(2688),
    [anon_sym_va_arg] = ACTIONS(2688),
    [anon_sym_ATencode] = ACTIONS(2690),
    [anon_sym_ATsynchronized] = ACTIONS(2690),
    [anon_sym_BOOL] = ACTIONS(2688),
    [anon_sym_IMP] = ACTIONS(2688),
    [anon_sym_SEL] = ACTIONS(2688),
    [anon_sym_Class] = ACTIONS(2688),
    [anon_sym_id] = ACTIONS(2688),
  },
  [STATE(685)] = {
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_RBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(686)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_RBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(687)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_RBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(688)] = {
    [ts_builtin_sym_end] = ACTIONS(2430),
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(689)] = {
    [ts_builtin_sym_end] = ACTIONS(2348),
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(690)] = {
    [ts_builtin_sym_end] = ACTIONS(2456),
    [sym_identifier] = ACTIONS(2454),
    [aux_sym_preproc_include_token1] = ACTIONS(2454),
    [aux_sym_preproc_include_token2] = ACTIONS(2454),
    [aux_sym_preproc_def_token1] = ACTIONS(2454),
    [aux_sym_preproc_if_token1] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2454),
    [sym_preproc_directive] = ACTIONS(2454),
    [anon_sym_LPAREN2] = ACTIONS(2456),
    [anon_sym_BANG] = ACTIONS(2456),
    [anon_sym_TILDE] = ACTIONS(2456),
    [anon_sym_DASH] = ACTIONS(2454),
    [anon_sym_PLUS] = ACTIONS(2454),
    [anon_sym_STAR] = ACTIONS(2456),
    [anon_sym_CARET] = ACTIONS(2456),
    [anon_sym_AMP] = ACTIONS(2456),
    [anon_sym_SEMI] = ACTIONS(2456),
    [anon_sym___extension__] = ACTIONS(2454),
    [anon_sym_typedef] = ACTIONS(2454),
    [anon_sym_extern] = ACTIONS(2454),
    [anon_sym___attribute__] = ACTIONS(2454),
    [anon_sym___attribute] = ACTIONS(2454),
    [anon_sym_noreturn] = ACTIONS(2454),
    [anon_sym_LBRACK] = ACTIONS(2456),
    [anon_sym___declspec] = ACTIONS(2454),
    [anon_sym___cdecl] = ACTIONS(2454),
    [anon_sym___clrcall] = ACTIONS(2454),
    [anon_sym___stdcall] = ACTIONS(2454),
    [anon_sym___fastcall] = ACTIONS(2454),
    [anon_sym___thiscall] = ACTIONS(2454),
    [anon_sym___vectorcall] = ACTIONS(2454),
    [anon_sym_LBRACE] = ACTIONS(2456),
    [anon_sym_signed] = ACTIONS(2454),
    [anon_sym_unsigned] = ACTIONS(2454),
    [anon_sym_long] = ACTIONS(2454),
    [anon_sym_short] = ACTIONS(2454),
    [anon_sym_static] = ACTIONS(2454),
    [anon_sym_ATautoreleasepool] = ACTIONS(2456),
    [anon_sym_auto] = ACTIONS(2454),
    [anon_sym_register] = ACTIONS(2454),
    [anon_sym_inline] = ACTIONS(2454),
    [anon_sym___inline] = ACTIONS(2454),
    [anon_sym___inline__] = ACTIONS(2454),
    [anon_sym___forceinline] = ACTIONS(2454),
    [anon_sym_thread_local] = ACTIONS(2454),
    [anon_sym___thread] = ACTIONS(2454),
    [anon_sym_CG_EXTERN] = ACTIONS(2454),
    [anon_sym_CG_INLINE] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2454),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2454),
    [anon_sym_IBOutlet] = ACTIONS(2454),
    [anon_sym_IBInspectable] = ACTIONS(2454),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2454),
    [anon_sym_NS_INLINE] = ACTIONS(2454),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2454),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2454),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2454),
    [anon_sym_const] = ACTIONS(2454),
    [anon_sym_constexpr] = ACTIONS(2454),
    [anon_sym_volatile] = ACTIONS(2454),
    [anon_sym_restrict] = ACTIONS(2454),
    [anon_sym___restrict__] = ACTIONS(2454),
    [anon_sym__Atomic] = ACTIONS(2454),
    [anon_sym__Noreturn] = ACTIONS(2454),
    [anon_sym__Nonnull] = ACTIONS(2454),
    [anon_sym_nullable] = ACTIONS(2454),
    [anon_sym__Complex] = ACTIONS(2454),
    [anon_sym__Nullable] = ACTIONS(2454),
    [anon_sym__Nullable_result] = ACTIONS(2454),
    [anon_sym__Null_unspecified] = ACTIONS(2454),
    [anon_sym___autoreleasing] = ACTIONS(2454),
    [anon_sym___block] = ACTIONS(2454),
    [anon_sym___bridge] = ACTIONS(2454),
    [anon_sym___bridge_retained] = ACTIONS(2454),
    [anon_sym___bridge_transfer] = ACTIONS(2454),
    [anon_sym___complex] = ACTIONS(2454),
    [anon_sym___const] = ACTIONS(2454),
    [anon_sym___imag] = ACTIONS(2454),
    [anon_sym___kindof] = ACTIONS(2454),
    [anon_sym___nonnull] = ACTIONS(2454),
    [anon_sym___nullable] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2454),
    [anon_sym___real] = ACTIONS(2454),
    [anon_sym___strong] = ACTIONS(2454),
    [anon_sym___unsafe_unretained] = ACTIONS(2454),
    [anon_sym___unused] = ACTIONS(2454),
    [anon_sym___weak] = ACTIONS(2454),
    [anon_sym_alignas] = ACTIONS(2454),
    [anon_sym__Alignas] = ACTIONS(2454),
    [sym_primitive_type] = ACTIONS(2454),
    [anon_sym_enum] = ACTIONS(2454),
    [anon_sym_struct] = ACTIONS(2454),
    [anon_sym_union] = ACTIONS(2454),
    [anon_sym_if] = ACTIONS(2454),
    [anon_sym_else] = ACTIONS(2454),
    [anon_sym_switch] = ACTIONS(2454),
    [anon_sym_case] = ACTIONS(2454),
    [anon_sym_default] = ACTIONS(2454),
    [anon_sym_while] = ACTIONS(2454),
    [anon_sym_do] = ACTIONS(2454),
    [anon_sym_for] = ACTIONS(2454),
    [anon_sym_in] = ACTIONS(2454),
    [anon_sym_return] = ACTIONS(2454),
    [anon_sym_break] = ACTIONS(2454),
    [anon_sym_continue] = ACTIONS(2454),
    [anon_sym_goto] = ACTIONS(2454),
    [anon_sym_DASH_DASH] = ACTIONS(2456),
    [anon_sym_PLUS_PLUS] = ACTIONS(2456),
    [anon_sym_sizeof] = ACTIONS(2454),
    [anon_sym___alignof__] = ACTIONS(2454),
    [anon_sym___alignof] = ACTIONS(2454),
    [anon_sym__alignof] = ACTIONS(2454),
    [anon_sym_alignof] = ACTIONS(2454),
    [anon_sym__Alignof] = ACTIONS(2454),
    [anon_sym_offsetof] = ACTIONS(2454),
    [anon_sym__Generic] = ACTIONS(2454),
    [anon_sym_asm] = ACTIONS(2454),
    [anon_sym___asm__] = ACTIONS(2454),
    [anon_sym___asm] = ACTIONS(2454),
    [sym_number_literal] = ACTIONS(2456),
    [anon_sym_L_SQUOTE] = ACTIONS(2456),
    [anon_sym_u_SQUOTE] = ACTIONS(2456),
    [anon_sym_U_SQUOTE] = ACTIONS(2456),
    [anon_sym_u8_SQUOTE] = ACTIONS(2456),
    [anon_sym_SQUOTE] = ACTIONS(2456),
    [anon_sym_AT] = ACTIONS(2454),
    [anon_sym_DQUOTE] = ACTIONS(2456),
    [anon_sym_L_DQUOTE] = ACTIONS(2456),
    [anon_sym_u_DQUOTE] = ACTIONS(2456),
    [anon_sym_U_DQUOTE] = ACTIONS(2456),
    [anon_sym_u8_DQUOTE] = ACTIONS(2456),
    [sym_true] = ACTIONS(2454),
    [sym_false] = ACTIONS(2454),
    [anon_sym_NULL] = ACTIONS(2454),
    [anon_sym_nullptr] = ACTIONS(2454),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2454),
    [anon_sym___typeof] = ACTIONS(2454),
    [anon_sym_typeof] = ACTIONS(2454),
    [anon_sym_ATimport] = ACTIONS(2456),
    [aux_sym_preproc_undef_token1] = ACTIONS(2454),
    [anon_sym_POUND] = ACTIONS(2454),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2454),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2454),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2454),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_API_AVAILABLE] = ACTIONS(2454),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_API_DEPRECATED] = ACTIONS(2454),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym___deprecated_msg] = ACTIONS(2454),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2454),
    [anon_sym_ATprotocol] = ACTIONS(2456),
    [anon_sym_ATinterface] = ACTIONS(2456),
    [anon_sym_ATimplementation] = ACTIONS(2456),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2456),
    [anon_sym_ATtry] = ACTIONS(2456),
    [anon_sym___try] = ACTIONS(2454),
    [anon_sym_ATthrow] = ACTIONS(2456),
    [anon_sym_ATselector] = ACTIONS(2456),
    [anon_sym_ATavailable] = ACTIONS(2456),
    [anon_sym___builtin_available] = ACTIONS(2454),
    [anon_sym_va_arg] = ACTIONS(2454),
    [anon_sym_ATencode] = ACTIONS(2456),
    [anon_sym_ATsynchronized] = ACTIONS(2456),
    [anon_sym_BOOL] = ACTIONS(2454),
    [anon_sym_IMP] = ACTIONS(2454),
    [anon_sym_SEL] = ACTIONS(2454),
    [anon_sym_Class] = ACTIONS(2454),
    [anon_sym_id] = ACTIONS(2454),
  },
  [STATE(691)] = {
    [sym_identifier] = ACTIONS(2494),
    [aux_sym_preproc_include_token1] = ACTIONS(2494),
    [aux_sym_preproc_include_token2] = ACTIONS(2494),
    [aux_sym_preproc_def_token1] = ACTIONS(2494),
    [aux_sym_preproc_if_token1] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2494),
    [sym_preproc_directive] = ACTIONS(2494),
    [anon_sym_LPAREN2] = ACTIONS(2496),
    [anon_sym_BANG] = ACTIONS(2496),
    [anon_sym_TILDE] = ACTIONS(2496),
    [anon_sym_DASH] = ACTIONS(2494),
    [anon_sym_PLUS] = ACTIONS(2494),
    [anon_sym_STAR] = ACTIONS(2496),
    [anon_sym_CARET] = ACTIONS(2496),
    [anon_sym_AMP] = ACTIONS(2496),
    [anon_sym_SEMI] = ACTIONS(2496),
    [anon_sym___extension__] = ACTIONS(2494),
    [anon_sym_typedef] = ACTIONS(2494),
    [anon_sym_extern] = ACTIONS(2494),
    [anon_sym___attribute__] = ACTIONS(2494),
    [anon_sym___attribute] = ACTIONS(2494),
    [anon_sym_noreturn] = ACTIONS(2494),
    [anon_sym_LBRACK] = ACTIONS(2496),
    [anon_sym___declspec] = ACTIONS(2494),
    [anon_sym___cdecl] = ACTIONS(2494),
    [anon_sym___clrcall] = ACTIONS(2494),
    [anon_sym___stdcall] = ACTIONS(2494),
    [anon_sym___fastcall] = ACTIONS(2494),
    [anon_sym___thiscall] = ACTIONS(2494),
    [anon_sym___vectorcall] = ACTIONS(2494),
    [anon_sym_LBRACE] = ACTIONS(2496),
    [anon_sym_RBRACE] = ACTIONS(2496),
    [anon_sym_signed] = ACTIONS(2494),
    [anon_sym_unsigned] = ACTIONS(2494),
    [anon_sym_long] = ACTIONS(2494),
    [anon_sym_short] = ACTIONS(2494),
    [anon_sym_static] = ACTIONS(2494),
    [anon_sym_ATautoreleasepool] = ACTIONS(2496),
    [anon_sym_auto] = ACTIONS(2494),
    [anon_sym_register] = ACTIONS(2494),
    [anon_sym_inline] = ACTIONS(2494),
    [anon_sym___inline] = ACTIONS(2494),
    [anon_sym___inline__] = ACTIONS(2494),
    [anon_sym___forceinline] = ACTIONS(2494),
    [anon_sym_thread_local] = ACTIONS(2494),
    [anon_sym___thread] = ACTIONS(2494),
    [anon_sym_CG_EXTERN] = ACTIONS(2494),
    [anon_sym_CG_INLINE] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2494),
    [anon_sym_IBOutlet] = ACTIONS(2494),
    [anon_sym_IBInspectable] = ACTIONS(2494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2494),
    [anon_sym_NS_INLINE] = ACTIONS(2494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2494),
    [anon_sym_const] = ACTIONS(2494),
    [anon_sym_constexpr] = ACTIONS(2494),
    [anon_sym_volatile] = ACTIONS(2494),
    [anon_sym_restrict] = ACTIONS(2494),
    [anon_sym___restrict__] = ACTIONS(2494),
    [anon_sym__Atomic] = ACTIONS(2494),
    [anon_sym__Noreturn] = ACTIONS(2494),
    [anon_sym__Nonnull] = ACTIONS(2494),
    [anon_sym_nullable] = ACTIONS(2494),
    [anon_sym__Complex] = ACTIONS(2494),
    [anon_sym__Nullable] = ACTIONS(2494),
    [anon_sym__Nullable_result] = ACTIONS(2494),
    [anon_sym__Null_unspecified] = ACTIONS(2494),
    [anon_sym___autoreleasing] = ACTIONS(2494),
    [anon_sym___block] = ACTIONS(2494),
    [anon_sym___bridge] = ACTIONS(2494),
    [anon_sym___bridge_retained] = ACTIONS(2494),
    [anon_sym___bridge_transfer] = ACTIONS(2494),
    [anon_sym___complex] = ACTIONS(2494),
    [anon_sym___const] = ACTIONS(2494),
    [anon_sym___imag] = ACTIONS(2494),
    [anon_sym___kindof] = ACTIONS(2494),
    [anon_sym___nonnull] = ACTIONS(2494),
    [anon_sym___nullable] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2494),
    [anon_sym___real] = ACTIONS(2494),
    [anon_sym___strong] = ACTIONS(2494),
    [anon_sym___unsafe_unretained] = ACTIONS(2494),
    [anon_sym___unused] = ACTIONS(2494),
    [anon_sym___weak] = ACTIONS(2494),
    [anon_sym_alignas] = ACTIONS(2494),
    [anon_sym__Alignas] = ACTIONS(2494),
    [sym_primitive_type] = ACTIONS(2494),
    [anon_sym_enum] = ACTIONS(2494),
    [anon_sym_struct] = ACTIONS(2494),
    [anon_sym_union] = ACTIONS(2494),
    [anon_sym_if] = ACTIONS(2494),
    [anon_sym_else] = ACTIONS(2494),
    [anon_sym_switch] = ACTIONS(2494),
    [anon_sym_case] = ACTIONS(2494),
    [anon_sym_default] = ACTIONS(2494),
    [anon_sym_while] = ACTIONS(2494),
    [anon_sym_do] = ACTIONS(2494),
    [anon_sym_for] = ACTIONS(2494),
    [anon_sym_in] = ACTIONS(2494),
    [anon_sym_return] = ACTIONS(2494),
    [anon_sym_break] = ACTIONS(2494),
    [anon_sym_continue] = ACTIONS(2494),
    [anon_sym_goto] = ACTIONS(2494),
    [anon_sym_DASH_DASH] = ACTIONS(2496),
    [anon_sym_PLUS_PLUS] = ACTIONS(2496),
    [anon_sym_sizeof] = ACTIONS(2494),
    [anon_sym___alignof__] = ACTIONS(2494),
    [anon_sym___alignof] = ACTIONS(2494),
    [anon_sym__alignof] = ACTIONS(2494),
    [anon_sym_alignof] = ACTIONS(2494),
    [anon_sym__Alignof] = ACTIONS(2494),
    [anon_sym_offsetof] = ACTIONS(2494),
    [anon_sym__Generic] = ACTIONS(2494),
    [anon_sym_asm] = ACTIONS(2494),
    [anon_sym___asm__] = ACTIONS(2494),
    [anon_sym___asm] = ACTIONS(2494),
    [sym_number_literal] = ACTIONS(2496),
    [anon_sym_L_SQUOTE] = ACTIONS(2496),
    [anon_sym_u_SQUOTE] = ACTIONS(2496),
    [anon_sym_U_SQUOTE] = ACTIONS(2496),
    [anon_sym_u8_SQUOTE] = ACTIONS(2496),
    [anon_sym_SQUOTE] = ACTIONS(2496),
    [anon_sym_AT] = ACTIONS(2494),
    [anon_sym_DQUOTE] = ACTIONS(2496),
    [anon_sym_L_DQUOTE] = ACTIONS(2496),
    [anon_sym_u_DQUOTE] = ACTIONS(2496),
    [anon_sym_U_DQUOTE] = ACTIONS(2496),
    [anon_sym_u8_DQUOTE] = ACTIONS(2496),
    [sym_true] = ACTIONS(2494),
    [sym_false] = ACTIONS(2494),
    [anon_sym_NULL] = ACTIONS(2494),
    [anon_sym_nullptr] = ACTIONS(2494),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2494),
    [anon_sym___typeof] = ACTIONS(2494),
    [anon_sym_typeof] = ACTIONS(2494),
    [anon_sym_ATimport] = ACTIONS(2496),
    [aux_sym_preproc_undef_token1] = ACTIONS(2494),
    [anon_sym_POUND] = ACTIONS(2494),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_API_AVAILABLE] = ACTIONS(2494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_API_DEPRECATED] = ACTIONS(2494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym___deprecated_msg] = ACTIONS(2494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2494),
    [anon_sym_ATprotocol] = ACTIONS(2496),
    [anon_sym_ATinterface] = ACTIONS(2496),
    [anon_sym_ATimplementation] = ACTIONS(2496),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2496),
    [anon_sym_ATtry] = ACTIONS(2496),
    [anon_sym___try] = ACTIONS(2494),
    [anon_sym_ATthrow] = ACTIONS(2496),
    [anon_sym_ATselector] = ACTIONS(2496),
    [anon_sym_ATavailable] = ACTIONS(2496),
    [anon_sym___builtin_available] = ACTIONS(2494),
    [anon_sym_va_arg] = ACTIONS(2494),
    [anon_sym_ATencode] = ACTIONS(2496),
    [anon_sym_ATsynchronized] = ACTIONS(2496),
    [anon_sym_BOOL] = ACTIONS(2494),
    [anon_sym_IMP] = ACTIONS(2494),
    [anon_sym_SEL] = ACTIONS(2494),
    [anon_sym_Class] = ACTIONS(2494),
    [anon_sym_id] = ACTIONS(2494),
  },
  [STATE(692)] = {
    [ts_builtin_sym_end] = ACTIONS(2430),
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(693)] = {
    [ts_builtin_sym_end] = ACTIONS(2430),
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(694)] = {
    [ts_builtin_sym_end] = ACTIONS(2352),
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(695)] = {
    [sym_identifier] = ACTIONS(2412),
    [aux_sym_preproc_include_token1] = ACTIONS(2412),
    [aux_sym_preproc_include_token2] = ACTIONS(2412),
    [aux_sym_preproc_def_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2412),
    [sym_preproc_directive] = ACTIONS(2412),
    [anon_sym_LPAREN2] = ACTIONS(2414),
    [anon_sym_BANG] = ACTIONS(2414),
    [anon_sym_TILDE] = ACTIONS(2414),
    [anon_sym_DASH] = ACTIONS(2412),
    [anon_sym_PLUS] = ACTIONS(2412),
    [anon_sym_STAR] = ACTIONS(2414),
    [anon_sym_CARET] = ACTIONS(2414),
    [anon_sym_AMP] = ACTIONS(2414),
    [anon_sym_SEMI] = ACTIONS(2414),
    [anon_sym___extension__] = ACTIONS(2412),
    [anon_sym_typedef] = ACTIONS(2412),
    [anon_sym_extern] = ACTIONS(2412),
    [anon_sym___attribute__] = ACTIONS(2412),
    [anon_sym___attribute] = ACTIONS(2412),
    [anon_sym_noreturn] = ACTIONS(2412),
    [anon_sym_LBRACK] = ACTIONS(2414),
    [anon_sym___declspec] = ACTIONS(2412),
    [anon_sym___cdecl] = ACTIONS(2412),
    [anon_sym___clrcall] = ACTIONS(2412),
    [anon_sym___stdcall] = ACTIONS(2412),
    [anon_sym___fastcall] = ACTIONS(2412),
    [anon_sym___thiscall] = ACTIONS(2412),
    [anon_sym___vectorcall] = ACTIONS(2412),
    [anon_sym_LBRACE] = ACTIONS(2414),
    [anon_sym_RBRACE] = ACTIONS(2414),
    [anon_sym_signed] = ACTIONS(2412),
    [anon_sym_unsigned] = ACTIONS(2412),
    [anon_sym_long] = ACTIONS(2412),
    [anon_sym_short] = ACTIONS(2412),
    [anon_sym_static] = ACTIONS(2412),
    [anon_sym_ATautoreleasepool] = ACTIONS(2414),
    [anon_sym_auto] = ACTIONS(2412),
    [anon_sym_register] = ACTIONS(2412),
    [anon_sym_inline] = ACTIONS(2412),
    [anon_sym___inline] = ACTIONS(2412),
    [anon_sym___inline__] = ACTIONS(2412),
    [anon_sym___forceinline] = ACTIONS(2412),
    [anon_sym_thread_local] = ACTIONS(2412),
    [anon_sym___thread] = ACTIONS(2412),
    [anon_sym_CG_EXTERN] = ACTIONS(2412),
    [anon_sym_CG_INLINE] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2412),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2412),
    [anon_sym_IBOutlet] = ACTIONS(2412),
    [anon_sym_IBInspectable] = ACTIONS(2412),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2412),
    [anon_sym_NS_INLINE] = ACTIONS(2412),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2412),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2412),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2412),
    [anon_sym_const] = ACTIONS(2412),
    [anon_sym_constexpr] = ACTIONS(2412),
    [anon_sym_volatile] = ACTIONS(2412),
    [anon_sym_restrict] = ACTIONS(2412),
    [anon_sym___restrict__] = ACTIONS(2412),
    [anon_sym__Atomic] = ACTIONS(2412),
    [anon_sym__Noreturn] = ACTIONS(2412),
    [anon_sym__Nonnull] = ACTIONS(2412),
    [anon_sym_nullable] = ACTIONS(2412),
    [anon_sym__Complex] = ACTIONS(2412),
    [anon_sym__Nullable] = ACTIONS(2412),
    [anon_sym__Nullable_result] = ACTIONS(2412),
    [anon_sym__Null_unspecified] = ACTIONS(2412),
    [anon_sym___autoreleasing] = ACTIONS(2412),
    [anon_sym___block] = ACTIONS(2412),
    [anon_sym___bridge] = ACTIONS(2412),
    [anon_sym___bridge_retained] = ACTIONS(2412),
    [anon_sym___bridge_transfer] = ACTIONS(2412),
    [anon_sym___complex] = ACTIONS(2412),
    [anon_sym___const] = ACTIONS(2412),
    [anon_sym___imag] = ACTIONS(2412),
    [anon_sym___kindof] = ACTIONS(2412),
    [anon_sym___nonnull] = ACTIONS(2412),
    [anon_sym___nullable] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2412),
    [anon_sym___real] = ACTIONS(2412),
    [anon_sym___strong] = ACTIONS(2412),
    [anon_sym___unsafe_unretained] = ACTIONS(2412),
    [anon_sym___unused] = ACTIONS(2412),
    [anon_sym___weak] = ACTIONS(2412),
    [anon_sym_alignas] = ACTIONS(2412),
    [anon_sym__Alignas] = ACTIONS(2412),
    [sym_primitive_type] = ACTIONS(2412),
    [anon_sym_enum] = ACTIONS(2412),
    [anon_sym_struct] = ACTIONS(2412),
    [anon_sym_union] = ACTIONS(2412),
    [anon_sym_if] = ACTIONS(2412),
    [anon_sym_else] = ACTIONS(2412),
    [anon_sym_switch] = ACTIONS(2412),
    [anon_sym_case] = ACTIONS(2412),
    [anon_sym_default] = ACTIONS(2412),
    [anon_sym_while] = ACTIONS(2412),
    [anon_sym_do] = ACTIONS(2412),
    [anon_sym_for] = ACTIONS(2412),
    [anon_sym_in] = ACTIONS(2412),
    [anon_sym_return] = ACTIONS(2412),
    [anon_sym_break] = ACTIONS(2412),
    [anon_sym_continue] = ACTIONS(2412),
    [anon_sym_goto] = ACTIONS(2412),
    [anon_sym_DASH_DASH] = ACTIONS(2414),
    [anon_sym_PLUS_PLUS] = ACTIONS(2414),
    [anon_sym_sizeof] = ACTIONS(2412),
    [anon_sym___alignof__] = ACTIONS(2412),
    [anon_sym___alignof] = ACTIONS(2412),
    [anon_sym__alignof] = ACTIONS(2412),
    [anon_sym_alignof] = ACTIONS(2412),
    [anon_sym__Alignof] = ACTIONS(2412),
    [anon_sym_offsetof] = ACTIONS(2412),
    [anon_sym__Generic] = ACTIONS(2412),
    [anon_sym_asm] = ACTIONS(2412),
    [anon_sym___asm__] = ACTIONS(2412),
    [anon_sym___asm] = ACTIONS(2412),
    [sym_number_literal] = ACTIONS(2414),
    [anon_sym_L_SQUOTE] = ACTIONS(2414),
    [anon_sym_u_SQUOTE] = ACTIONS(2414),
    [anon_sym_U_SQUOTE] = ACTIONS(2414),
    [anon_sym_u8_SQUOTE] = ACTIONS(2414),
    [anon_sym_SQUOTE] = ACTIONS(2414),
    [anon_sym_AT] = ACTIONS(2412),
    [anon_sym_DQUOTE] = ACTIONS(2414),
    [anon_sym_L_DQUOTE] = ACTIONS(2414),
    [anon_sym_u_DQUOTE] = ACTIONS(2414),
    [anon_sym_U_DQUOTE] = ACTIONS(2414),
    [anon_sym_u8_DQUOTE] = ACTIONS(2414),
    [sym_true] = ACTIONS(2412),
    [sym_false] = ACTIONS(2412),
    [anon_sym_NULL] = ACTIONS(2412),
    [anon_sym_nullptr] = ACTIONS(2412),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2412),
    [anon_sym___typeof] = ACTIONS(2412),
    [anon_sym_typeof] = ACTIONS(2412),
    [anon_sym_ATimport] = ACTIONS(2414),
    [aux_sym_preproc_undef_token1] = ACTIONS(2412),
    [anon_sym_POUND] = ACTIONS(2412),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2412),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2412),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2412),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_API_AVAILABLE] = ACTIONS(2412),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_API_DEPRECATED] = ACTIONS(2412),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym___deprecated_msg] = ACTIONS(2412),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2412),
    [anon_sym_ATprotocol] = ACTIONS(2414),
    [anon_sym_ATinterface] = ACTIONS(2414),
    [anon_sym_ATimplementation] = ACTIONS(2414),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2414),
    [anon_sym_ATtry] = ACTIONS(2414),
    [anon_sym___try] = ACTIONS(2412),
    [anon_sym_ATthrow] = ACTIONS(2414),
    [anon_sym_ATselector] = ACTIONS(2414),
    [anon_sym_ATavailable] = ACTIONS(2414),
    [anon_sym___builtin_available] = ACTIONS(2412),
    [anon_sym_va_arg] = ACTIONS(2412),
    [anon_sym_ATencode] = ACTIONS(2414),
    [anon_sym_ATsynchronized] = ACTIONS(2414),
    [anon_sym_BOOL] = ACTIONS(2412),
    [anon_sym_IMP] = ACTIONS(2412),
    [anon_sym_SEL] = ACTIONS(2412),
    [anon_sym_Class] = ACTIONS(2412),
    [anon_sym_id] = ACTIONS(2412),
  },
  [STATE(696)] = {
    [sym_identifier] = ACTIONS(2506),
    [aux_sym_preproc_include_token1] = ACTIONS(2506),
    [aux_sym_preproc_include_token2] = ACTIONS(2506),
    [aux_sym_preproc_def_token1] = ACTIONS(2506),
    [aux_sym_preproc_if_token1] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2506),
    [sym_preproc_directive] = ACTIONS(2506),
    [anon_sym_LPAREN2] = ACTIONS(2508),
    [anon_sym_BANG] = ACTIONS(2508),
    [anon_sym_TILDE] = ACTIONS(2508),
    [anon_sym_DASH] = ACTIONS(2506),
    [anon_sym_PLUS] = ACTIONS(2506),
    [anon_sym_STAR] = ACTIONS(2508),
    [anon_sym_CARET] = ACTIONS(2508),
    [anon_sym_AMP] = ACTIONS(2508),
    [anon_sym_SEMI] = ACTIONS(2508),
    [anon_sym___extension__] = ACTIONS(2506),
    [anon_sym_typedef] = ACTIONS(2506),
    [anon_sym_extern] = ACTIONS(2506),
    [anon_sym___attribute__] = ACTIONS(2506),
    [anon_sym___attribute] = ACTIONS(2506),
    [anon_sym_noreturn] = ACTIONS(2506),
    [anon_sym_LBRACK] = ACTIONS(2508),
    [anon_sym___declspec] = ACTIONS(2506),
    [anon_sym___cdecl] = ACTIONS(2506),
    [anon_sym___clrcall] = ACTIONS(2506),
    [anon_sym___stdcall] = ACTIONS(2506),
    [anon_sym___fastcall] = ACTIONS(2506),
    [anon_sym___thiscall] = ACTIONS(2506),
    [anon_sym___vectorcall] = ACTIONS(2506),
    [anon_sym_LBRACE] = ACTIONS(2508),
    [anon_sym_RBRACE] = ACTIONS(2508),
    [anon_sym_signed] = ACTIONS(2506),
    [anon_sym_unsigned] = ACTIONS(2506),
    [anon_sym_long] = ACTIONS(2506),
    [anon_sym_short] = ACTIONS(2506),
    [anon_sym_static] = ACTIONS(2506),
    [anon_sym_ATautoreleasepool] = ACTIONS(2508),
    [anon_sym_auto] = ACTIONS(2506),
    [anon_sym_register] = ACTIONS(2506),
    [anon_sym_inline] = ACTIONS(2506),
    [anon_sym___inline] = ACTIONS(2506),
    [anon_sym___inline__] = ACTIONS(2506),
    [anon_sym___forceinline] = ACTIONS(2506),
    [anon_sym_thread_local] = ACTIONS(2506),
    [anon_sym___thread] = ACTIONS(2506),
    [anon_sym_CG_EXTERN] = ACTIONS(2506),
    [anon_sym_CG_INLINE] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2506),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2506),
    [anon_sym_IBOutlet] = ACTIONS(2506),
    [anon_sym_IBInspectable] = ACTIONS(2506),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2506),
    [anon_sym_NS_INLINE] = ACTIONS(2506),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2506),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2506),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2506),
    [anon_sym_const] = ACTIONS(2506),
    [anon_sym_constexpr] = ACTIONS(2506),
    [anon_sym_volatile] = ACTIONS(2506),
    [anon_sym_restrict] = ACTIONS(2506),
    [anon_sym___restrict__] = ACTIONS(2506),
    [anon_sym__Atomic] = ACTIONS(2506),
    [anon_sym__Noreturn] = ACTIONS(2506),
    [anon_sym__Nonnull] = ACTIONS(2506),
    [anon_sym_nullable] = ACTIONS(2506),
    [anon_sym__Complex] = ACTIONS(2506),
    [anon_sym__Nullable] = ACTIONS(2506),
    [anon_sym__Nullable_result] = ACTIONS(2506),
    [anon_sym__Null_unspecified] = ACTIONS(2506),
    [anon_sym___autoreleasing] = ACTIONS(2506),
    [anon_sym___block] = ACTIONS(2506),
    [anon_sym___bridge] = ACTIONS(2506),
    [anon_sym___bridge_retained] = ACTIONS(2506),
    [anon_sym___bridge_transfer] = ACTIONS(2506),
    [anon_sym___complex] = ACTIONS(2506),
    [anon_sym___const] = ACTIONS(2506),
    [anon_sym___imag] = ACTIONS(2506),
    [anon_sym___kindof] = ACTIONS(2506),
    [anon_sym___nonnull] = ACTIONS(2506),
    [anon_sym___nullable] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2506),
    [anon_sym___real] = ACTIONS(2506),
    [anon_sym___strong] = ACTIONS(2506),
    [anon_sym___unsafe_unretained] = ACTIONS(2506),
    [anon_sym___unused] = ACTIONS(2506),
    [anon_sym___weak] = ACTIONS(2506),
    [anon_sym_alignas] = ACTIONS(2506),
    [anon_sym__Alignas] = ACTIONS(2506),
    [sym_primitive_type] = ACTIONS(2506),
    [anon_sym_enum] = ACTIONS(2506),
    [anon_sym_struct] = ACTIONS(2506),
    [anon_sym_union] = ACTIONS(2506),
    [anon_sym_if] = ACTIONS(2506),
    [anon_sym_else] = ACTIONS(2506),
    [anon_sym_switch] = ACTIONS(2506),
    [anon_sym_case] = ACTIONS(2506),
    [anon_sym_default] = ACTIONS(2506),
    [anon_sym_while] = ACTIONS(2506),
    [anon_sym_do] = ACTIONS(2506),
    [anon_sym_for] = ACTIONS(2506),
    [anon_sym_in] = ACTIONS(2506),
    [anon_sym_return] = ACTIONS(2506),
    [anon_sym_break] = ACTIONS(2506),
    [anon_sym_continue] = ACTIONS(2506),
    [anon_sym_goto] = ACTIONS(2506),
    [anon_sym_DASH_DASH] = ACTIONS(2508),
    [anon_sym_PLUS_PLUS] = ACTIONS(2508),
    [anon_sym_sizeof] = ACTIONS(2506),
    [anon_sym___alignof__] = ACTIONS(2506),
    [anon_sym___alignof] = ACTIONS(2506),
    [anon_sym__alignof] = ACTIONS(2506),
    [anon_sym_alignof] = ACTIONS(2506),
    [anon_sym__Alignof] = ACTIONS(2506),
    [anon_sym_offsetof] = ACTIONS(2506),
    [anon_sym__Generic] = ACTIONS(2506),
    [anon_sym_asm] = ACTIONS(2506),
    [anon_sym___asm__] = ACTIONS(2506),
    [anon_sym___asm] = ACTIONS(2506),
    [sym_number_literal] = ACTIONS(2508),
    [anon_sym_L_SQUOTE] = ACTIONS(2508),
    [anon_sym_u_SQUOTE] = ACTIONS(2508),
    [anon_sym_U_SQUOTE] = ACTIONS(2508),
    [anon_sym_u8_SQUOTE] = ACTIONS(2508),
    [anon_sym_SQUOTE] = ACTIONS(2508),
    [anon_sym_AT] = ACTIONS(2506),
    [anon_sym_DQUOTE] = ACTIONS(2508),
    [anon_sym_L_DQUOTE] = ACTIONS(2508),
    [anon_sym_u_DQUOTE] = ACTIONS(2508),
    [anon_sym_U_DQUOTE] = ACTIONS(2508),
    [anon_sym_u8_DQUOTE] = ACTIONS(2508),
    [sym_true] = ACTIONS(2506),
    [sym_false] = ACTIONS(2506),
    [anon_sym_NULL] = ACTIONS(2506),
    [anon_sym_nullptr] = ACTIONS(2506),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2506),
    [anon_sym___typeof] = ACTIONS(2506),
    [anon_sym_typeof] = ACTIONS(2506),
    [anon_sym_ATimport] = ACTIONS(2508),
    [aux_sym_preproc_undef_token1] = ACTIONS(2506),
    [anon_sym_POUND] = ACTIONS(2506),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2506),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2506),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2506),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_API_AVAILABLE] = ACTIONS(2506),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_API_DEPRECATED] = ACTIONS(2506),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym___deprecated_msg] = ACTIONS(2506),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2506),
    [anon_sym_ATprotocol] = ACTIONS(2508),
    [anon_sym_ATinterface] = ACTIONS(2508),
    [anon_sym_ATimplementation] = ACTIONS(2508),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2508),
    [anon_sym_ATtry] = ACTIONS(2508),
    [anon_sym___try] = ACTIONS(2506),
    [anon_sym_ATthrow] = ACTIONS(2508),
    [anon_sym_ATselector] = ACTIONS(2508),
    [anon_sym_ATavailable] = ACTIONS(2508),
    [anon_sym___builtin_available] = ACTIONS(2506),
    [anon_sym_va_arg] = ACTIONS(2506),
    [anon_sym_ATencode] = ACTIONS(2508),
    [anon_sym_ATsynchronized] = ACTIONS(2508),
    [anon_sym_BOOL] = ACTIONS(2506),
    [anon_sym_IMP] = ACTIONS(2506),
    [anon_sym_SEL] = ACTIONS(2506),
    [anon_sym_Class] = ACTIONS(2506),
    [anon_sym_id] = ACTIONS(2506),
  },
  [STATE(697)] = {
    [ts_builtin_sym_end] = ACTIONS(2426),
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(698)] = {
    [sym_identifier] = ACTIONS(2462),
    [aux_sym_preproc_include_token1] = ACTIONS(2462),
    [aux_sym_preproc_include_token2] = ACTIONS(2462),
    [aux_sym_preproc_def_token1] = ACTIONS(2462),
    [aux_sym_preproc_if_token1] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2462),
    [sym_preproc_directive] = ACTIONS(2462),
    [anon_sym_LPAREN2] = ACTIONS(2464),
    [anon_sym_BANG] = ACTIONS(2464),
    [anon_sym_TILDE] = ACTIONS(2464),
    [anon_sym_DASH] = ACTIONS(2462),
    [anon_sym_PLUS] = ACTIONS(2462),
    [anon_sym_STAR] = ACTIONS(2464),
    [anon_sym_CARET] = ACTIONS(2464),
    [anon_sym_AMP] = ACTIONS(2464),
    [anon_sym_SEMI] = ACTIONS(2464),
    [anon_sym___extension__] = ACTIONS(2462),
    [anon_sym_typedef] = ACTIONS(2462),
    [anon_sym_extern] = ACTIONS(2462),
    [anon_sym___attribute__] = ACTIONS(2462),
    [anon_sym___attribute] = ACTIONS(2462),
    [anon_sym_noreturn] = ACTIONS(2462),
    [anon_sym_LBRACK] = ACTIONS(2464),
    [anon_sym___declspec] = ACTIONS(2462),
    [anon_sym___cdecl] = ACTIONS(2462),
    [anon_sym___clrcall] = ACTIONS(2462),
    [anon_sym___stdcall] = ACTIONS(2462),
    [anon_sym___fastcall] = ACTIONS(2462),
    [anon_sym___thiscall] = ACTIONS(2462),
    [anon_sym___vectorcall] = ACTIONS(2462),
    [anon_sym_LBRACE] = ACTIONS(2464),
    [anon_sym_RBRACE] = ACTIONS(2464),
    [anon_sym_signed] = ACTIONS(2462),
    [anon_sym_unsigned] = ACTIONS(2462),
    [anon_sym_long] = ACTIONS(2462),
    [anon_sym_short] = ACTIONS(2462),
    [anon_sym_static] = ACTIONS(2462),
    [anon_sym_ATautoreleasepool] = ACTIONS(2464),
    [anon_sym_auto] = ACTIONS(2462),
    [anon_sym_register] = ACTIONS(2462),
    [anon_sym_inline] = ACTIONS(2462),
    [anon_sym___inline] = ACTIONS(2462),
    [anon_sym___inline__] = ACTIONS(2462),
    [anon_sym___forceinline] = ACTIONS(2462),
    [anon_sym_thread_local] = ACTIONS(2462),
    [anon_sym___thread] = ACTIONS(2462),
    [anon_sym_CG_EXTERN] = ACTIONS(2462),
    [anon_sym_CG_INLINE] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2462),
    [anon_sym_IBOutlet] = ACTIONS(2462),
    [anon_sym_IBInspectable] = ACTIONS(2462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2462),
    [anon_sym_NS_INLINE] = ACTIONS(2462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2462),
    [anon_sym_const] = ACTIONS(2462),
    [anon_sym_constexpr] = ACTIONS(2462),
    [anon_sym_volatile] = ACTIONS(2462),
    [anon_sym_restrict] = ACTIONS(2462),
    [anon_sym___restrict__] = ACTIONS(2462),
    [anon_sym__Atomic] = ACTIONS(2462),
    [anon_sym__Noreturn] = ACTIONS(2462),
    [anon_sym__Nonnull] = ACTIONS(2462),
    [anon_sym_nullable] = ACTIONS(2462),
    [anon_sym__Complex] = ACTIONS(2462),
    [anon_sym__Nullable] = ACTIONS(2462),
    [anon_sym__Nullable_result] = ACTIONS(2462),
    [anon_sym__Null_unspecified] = ACTIONS(2462),
    [anon_sym___autoreleasing] = ACTIONS(2462),
    [anon_sym___block] = ACTIONS(2462),
    [anon_sym___bridge] = ACTIONS(2462),
    [anon_sym___bridge_retained] = ACTIONS(2462),
    [anon_sym___bridge_transfer] = ACTIONS(2462),
    [anon_sym___complex] = ACTIONS(2462),
    [anon_sym___const] = ACTIONS(2462),
    [anon_sym___imag] = ACTIONS(2462),
    [anon_sym___kindof] = ACTIONS(2462),
    [anon_sym___nonnull] = ACTIONS(2462),
    [anon_sym___nullable] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2462),
    [anon_sym___real] = ACTIONS(2462),
    [anon_sym___strong] = ACTIONS(2462),
    [anon_sym___unsafe_unretained] = ACTIONS(2462),
    [anon_sym___unused] = ACTIONS(2462),
    [anon_sym___weak] = ACTIONS(2462),
    [anon_sym_alignas] = ACTIONS(2462),
    [anon_sym__Alignas] = ACTIONS(2462),
    [sym_primitive_type] = ACTIONS(2462),
    [anon_sym_enum] = ACTIONS(2462),
    [anon_sym_struct] = ACTIONS(2462),
    [anon_sym_union] = ACTIONS(2462),
    [anon_sym_if] = ACTIONS(2462),
    [anon_sym_else] = ACTIONS(2462),
    [anon_sym_switch] = ACTIONS(2462),
    [anon_sym_case] = ACTIONS(2462),
    [anon_sym_default] = ACTIONS(2462),
    [anon_sym_while] = ACTIONS(2462),
    [anon_sym_do] = ACTIONS(2462),
    [anon_sym_for] = ACTIONS(2462),
    [anon_sym_in] = ACTIONS(2462),
    [anon_sym_return] = ACTIONS(2462),
    [anon_sym_break] = ACTIONS(2462),
    [anon_sym_continue] = ACTIONS(2462),
    [anon_sym_goto] = ACTIONS(2462),
    [anon_sym_DASH_DASH] = ACTIONS(2464),
    [anon_sym_PLUS_PLUS] = ACTIONS(2464),
    [anon_sym_sizeof] = ACTIONS(2462),
    [anon_sym___alignof__] = ACTIONS(2462),
    [anon_sym___alignof] = ACTIONS(2462),
    [anon_sym__alignof] = ACTIONS(2462),
    [anon_sym_alignof] = ACTIONS(2462),
    [anon_sym__Alignof] = ACTIONS(2462),
    [anon_sym_offsetof] = ACTIONS(2462),
    [anon_sym__Generic] = ACTIONS(2462),
    [anon_sym_asm] = ACTIONS(2462),
    [anon_sym___asm__] = ACTIONS(2462),
    [anon_sym___asm] = ACTIONS(2462),
    [sym_number_literal] = ACTIONS(2464),
    [anon_sym_L_SQUOTE] = ACTIONS(2464),
    [anon_sym_u_SQUOTE] = ACTIONS(2464),
    [anon_sym_U_SQUOTE] = ACTIONS(2464),
    [anon_sym_u8_SQUOTE] = ACTIONS(2464),
    [anon_sym_SQUOTE] = ACTIONS(2464),
    [anon_sym_AT] = ACTIONS(2462),
    [anon_sym_DQUOTE] = ACTIONS(2464),
    [anon_sym_L_DQUOTE] = ACTIONS(2464),
    [anon_sym_u_DQUOTE] = ACTIONS(2464),
    [anon_sym_U_DQUOTE] = ACTIONS(2464),
    [anon_sym_u8_DQUOTE] = ACTIONS(2464),
    [sym_true] = ACTIONS(2462),
    [sym_false] = ACTIONS(2462),
    [anon_sym_NULL] = ACTIONS(2462),
    [anon_sym_nullptr] = ACTIONS(2462),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2462),
    [anon_sym___typeof] = ACTIONS(2462),
    [anon_sym_typeof] = ACTIONS(2462),
    [anon_sym_ATimport] = ACTIONS(2464),
    [aux_sym_preproc_undef_token1] = ACTIONS(2462),
    [anon_sym_POUND] = ACTIONS(2462),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2462),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_API_AVAILABLE] = ACTIONS(2462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_API_DEPRECATED] = ACTIONS(2462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym___deprecated_msg] = ACTIONS(2462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2462),
    [anon_sym_ATprotocol] = ACTIONS(2464),
    [anon_sym_ATinterface] = ACTIONS(2464),
    [anon_sym_ATimplementation] = ACTIONS(2464),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2464),
    [anon_sym_ATtry] = ACTIONS(2464),
    [anon_sym___try] = ACTIONS(2462),
    [anon_sym_ATthrow] = ACTIONS(2464),
    [anon_sym_ATselector] = ACTIONS(2464),
    [anon_sym_ATavailable] = ACTIONS(2464),
    [anon_sym___builtin_available] = ACTIONS(2462),
    [anon_sym_va_arg] = ACTIONS(2462),
    [anon_sym_ATencode] = ACTIONS(2464),
    [anon_sym_ATsynchronized] = ACTIONS(2464),
    [anon_sym_BOOL] = ACTIONS(2462),
    [anon_sym_IMP] = ACTIONS(2462),
    [anon_sym_SEL] = ACTIONS(2462),
    [anon_sym_Class] = ACTIONS(2462),
    [anon_sym_id] = ACTIONS(2462),
  },
  [STATE(699)] = {
    [sym_identifier] = ACTIONS(2466),
    [aux_sym_preproc_include_token1] = ACTIONS(2466),
    [aux_sym_preproc_include_token2] = ACTIONS(2466),
    [aux_sym_preproc_def_token1] = ACTIONS(2466),
    [aux_sym_preproc_if_token1] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2466),
    [sym_preproc_directive] = ACTIONS(2466),
    [anon_sym_LPAREN2] = ACTIONS(2468),
    [anon_sym_BANG] = ACTIONS(2468),
    [anon_sym_TILDE] = ACTIONS(2468),
    [anon_sym_DASH] = ACTIONS(2466),
    [anon_sym_PLUS] = ACTIONS(2466),
    [anon_sym_STAR] = ACTIONS(2468),
    [anon_sym_CARET] = ACTIONS(2468),
    [anon_sym_AMP] = ACTIONS(2468),
    [anon_sym_SEMI] = ACTIONS(2468),
    [anon_sym___extension__] = ACTIONS(2466),
    [anon_sym_typedef] = ACTIONS(2466),
    [anon_sym_extern] = ACTIONS(2466),
    [anon_sym___attribute__] = ACTIONS(2466),
    [anon_sym___attribute] = ACTIONS(2466),
    [anon_sym_noreturn] = ACTIONS(2466),
    [anon_sym_LBRACK] = ACTIONS(2468),
    [anon_sym___declspec] = ACTIONS(2466),
    [anon_sym___cdecl] = ACTIONS(2466),
    [anon_sym___clrcall] = ACTIONS(2466),
    [anon_sym___stdcall] = ACTIONS(2466),
    [anon_sym___fastcall] = ACTIONS(2466),
    [anon_sym___thiscall] = ACTIONS(2466),
    [anon_sym___vectorcall] = ACTIONS(2466),
    [anon_sym_LBRACE] = ACTIONS(2468),
    [anon_sym_RBRACE] = ACTIONS(2468),
    [anon_sym_signed] = ACTIONS(2466),
    [anon_sym_unsigned] = ACTIONS(2466),
    [anon_sym_long] = ACTIONS(2466),
    [anon_sym_short] = ACTIONS(2466),
    [anon_sym_static] = ACTIONS(2466),
    [anon_sym_ATautoreleasepool] = ACTIONS(2468),
    [anon_sym_auto] = ACTIONS(2466),
    [anon_sym_register] = ACTIONS(2466),
    [anon_sym_inline] = ACTIONS(2466),
    [anon_sym___inline] = ACTIONS(2466),
    [anon_sym___inline__] = ACTIONS(2466),
    [anon_sym___forceinline] = ACTIONS(2466),
    [anon_sym_thread_local] = ACTIONS(2466),
    [anon_sym___thread] = ACTIONS(2466),
    [anon_sym_CG_EXTERN] = ACTIONS(2466),
    [anon_sym_CG_INLINE] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2466),
    [anon_sym_IBOutlet] = ACTIONS(2466),
    [anon_sym_IBInspectable] = ACTIONS(2466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2466),
    [anon_sym_NS_INLINE] = ACTIONS(2466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2466),
    [anon_sym_const] = ACTIONS(2466),
    [anon_sym_constexpr] = ACTIONS(2466),
    [anon_sym_volatile] = ACTIONS(2466),
    [anon_sym_restrict] = ACTIONS(2466),
    [anon_sym___restrict__] = ACTIONS(2466),
    [anon_sym__Atomic] = ACTIONS(2466),
    [anon_sym__Noreturn] = ACTIONS(2466),
    [anon_sym__Nonnull] = ACTIONS(2466),
    [anon_sym_nullable] = ACTIONS(2466),
    [anon_sym__Complex] = ACTIONS(2466),
    [anon_sym__Nullable] = ACTIONS(2466),
    [anon_sym__Nullable_result] = ACTIONS(2466),
    [anon_sym__Null_unspecified] = ACTIONS(2466),
    [anon_sym___autoreleasing] = ACTIONS(2466),
    [anon_sym___block] = ACTIONS(2466),
    [anon_sym___bridge] = ACTIONS(2466),
    [anon_sym___bridge_retained] = ACTIONS(2466),
    [anon_sym___bridge_transfer] = ACTIONS(2466),
    [anon_sym___complex] = ACTIONS(2466),
    [anon_sym___const] = ACTIONS(2466),
    [anon_sym___imag] = ACTIONS(2466),
    [anon_sym___kindof] = ACTIONS(2466),
    [anon_sym___nonnull] = ACTIONS(2466),
    [anon_sym___nullable] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2466),
    [anon_sym___real] = ACTIONS(2466),
    [anon_sym___strong] = ACTIONS(2466),
    [anon_sym___unsafe_unretained] = ACTIONS(2466),
    [anon_sym___unused] = ACTIONS(2466),
    [anon_sym___weak] = ACTIONS(2466),
    [anon_sym_alignas] = ACTIONS(2466),
    [anon_sym__Alignas] = ACTIONS(2466),
    [sym_primitive_type] = ACTIONS(2466),
    [anon_sym_enum] = ACTIONS(2466),
    [anon_sym_struct] = ACTIONS(2466),
    [anon_sym_union] = ACTIONS(2466),
    [anon_sym_if] = ACTIONS(2466),
    [anon_sym_else] = ACTIONS(2466),
    [anon_sym_switch] = ACTIONS(2466),
    [anon_sym_case] = ACTIONS(2466),
    [anon_sym_default] = ACTIONS(2466),
    [anon_sym_while] = ACTIONS(2466),
    [anon_sym_do] = ACTIONS(2466),
    [anon_sym_for] = ACTIONS(2466),
    [anon_sym_in] = ACTIONS(2466),
    [anon_sym_return] = ACTIONS(2466),
    [anon_sym_break] = ACTIONS(2466),
    [anon_sym_continue] = ACTIONS(2466),
    [anon_sym_goto] = ACTIONS(2466),
    [anon_sym_DASH_DASH] = ACTIONS(2468),
    [anon_sym_PLUS_PLUS] = ACTIONS(2468),
    [anon_sym_sizeof] = ACTIONS(2466),
    [anon_sym___alignof__] = ACTIONS(2466),
    [anon_sym___alignof] = ACTIONS(2466),
    [anon_sym__alignof] = ACTIONS(2466),
    [anon_sym_alignof] = ACTIONS(2466),
    [anon_sym__Alignof] = ACTIONS(2466),
    [anon_sym_offsetof] = ACTIONS(2466),
    [anon_sym__Generic] = ACTIONS(2466),
    [anon_sym_asm] = ACTIONS(2466),
    [anon_sym___asm__] = ACTIONS(2466),
    [anon_sym___asm] = ACTIONS(2466),
    [sym_number_literal] = ACTIONS(2468),
    [anon_sym_L_SQUOTE] = ACTIONS(2468),
    [anon_sym_u_SQUOTE] = ACTIONS(2468),
    [anon_sym_U_SQUOTE] = ACTIONS(2468),
    [anon_sym_u8_SQUOTE] = ACTIONS(2468),
    [anon_sym_SQUOTE] = ACTIONS(2468),
    [anon_sym_AT] = ACTIONS(2466),
    [anon_sym_DQUOTE] = ACTIONS(2468),
    [anon_sym_L_DQUOTE] = ACTIONS(2468),
    [anon_sym_u_DQUOTE] = ACTIONS(2468),
    [anon_sym_U_DQUOTE] = ACTIONS(2468),
    [anon_sym_u8_DQUOTE] = ACTIONS(2468),
    [sym_true] = ACTIONS(2466),
    [sym_false] = ACTIONS(2466),
    [anon_sym_NULL] = ACTIONS(2466),
    [anon_sym_nullptr] = ACTIONS(2466),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2466),
    [anon_sym___typeof] = ACTIONS(2466),
    [anon_sym_typeof] = ACTIONS(2466),
    [anon_sym_ATimport] = ACTIONS(2468),
    [aux_sym_preproc_undef_token1] = ACTIONS(2466),
    [anon_sym_POUND] = ACTIONS(2466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_API_AVAILABLE] = ACTIONS(2466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_API_DEPRECATED] = ACTIONS(2466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym___deprecated_msg] = ACTIONS(2466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2466),
    [anon_sym_ATprotocol] = ACTIONS(2468),
    [anon_sym_ATinterface] = ACTIONS(2468),
    [anon_sym_ATimplementation] = ACTIONS(2468),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2468),
    [anon_sym_ATtry] = ACTIONS(2468),
    [anon_sym___try] = ACTIONS(2466),
    [anon_sym_ATthrow] = ACTIONS(2468),
    [anon_sym_ATselector] = ACTIONS(2468),
    [anon_sym_ATavailable] = ACTIONS(2468),
    [anon_sym___builtin_available] = ACTIONS(2466),
    [anon_sym_va_arg] = ACTIONS(2466),
    [anon_sym_ATencode] = ACTIONS(2468),
    [anon_sym_ATsynchronized] = ACTIONS(2468),
    [anon_sym_BOOL] = ACTIONS(2466),
    [anon_sym_IMP] = ACTIONS(2466),
    [anon_sym_SEL] = ACTIONS(2466),
    [anon_sym_Class] = ACTIONS(2466),
    [anon_sym_id] = ACTIONS(2466),
  },
  [STATE(700)] = {
    [ts_builtin_sym_end] = ACTIONS(2362),
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(701)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_RBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(702)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_RBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(703)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_RBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(704)] = {
    [ts_builtin_sym_end] = ACTIONS(2512),
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(705)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_RBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(706)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_RBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(707)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_RBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(708)] = {
    [sym_identifier] = ACTIONS(2450),
    [aux_sym_preproc_include_token1] = ACTIONS(2450),
    [aux_sym_preproc_include_token2] = ACTIONS(2450),
    [aux_sym_preproc_def_token1] = ACTIONS(2450),
    [aux_sym_preproc_if_token1] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2450),
    [sym_preproc_directive] = ACTIONS(2450),
    [anon_sym_LPAREN2] = ACTIONS(2452),
    [anon_sym_BANG] = ACTIONS(2452),
    [anon_sym_TILDE] = ACTIONS(2452),
    [anon_sym_DASH] = ACTIONS(2450),
    [anon_sym_PLUS] = ACTIONS(2450),
    [anon_sym_STAR] = ACTIONS(2452),
    [anon_sym_CARET] = ACTIONS(2452),
    [anon_sym_AMP] = ACTIONS(2452),
    [anon_sym_SEMI] = ACTIONS(2452),
    [anon_sym___extension__] = ACTIONS(2450),
    [anon_sym_typedef] = ACTIONS(2450),
    [anon_sym_extern] = ACTIONS(2450),
    [anon_sym___attribute__] = ACTIONS(2450),
    [anon_sym___attribute] = ACTIONS(2450),
    [anon_sym_noreturn] = ACTIONS(2450),
    [anon_sym_LBRACK] = ACTIONS(2452),
    [anon_sym___declspec] = ACTIONS(2450),
    [anon_sym___cdecl] = ACTIONS(2450),
    [anon_sym___clrcall] = ACTIONS(2450),
    [anon_sym___stdcall] = ACTIONS(2450),
    [anon_sym___fastcall] = ACTIONS(2450),
    [anon_sym___thiscall] = ACTIONS(2450),
    [anon_sym___vectorcall] = ACTIONS(2450),
    [anon_sym_LBRACE] = ACTIONS(2452),
    [anon_sym_RBRACE] = ACTIONS(2452),
    [anon_sym_signed] = ACTIONS(2450),
    [anon_sym_unsigned] = ACTIONS(2450),
    [anon_sym_long] = ACTIONS(2450),
    [anon_sym_short] = ACTIONS(2450),
    [anon_sym_static] = ACTIONS(2450),
    [anon_sym_ATautoreleasepool] = ACTIONS(2452),
    [anon_sym_auto] = ACTIONS(2450),
    [anon_sym_register] = ACTIONS(2450),
    [anon_sym_inline] = ACTIONS(2450),
    [anon_sym___inline] = ACTIONS(2450),
    [anon_sym___inline__] = ACTIONS(2450),
    [anon_sym___forceinline] = ACTIONS(2450),
    [anon_sym_thread_local] = ACTIONS(2450),
    [anon_sym___thread] = ACTIONS(2450),
    [anon_sym_CG_EXTERN] = ACTIONS(2450),
    [anon_sym_CG_INLINE] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2450),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2450),
    [anon_sym_IBOutlet] = ACTIONS(2450),
    [anon_sym_IBInspectable] = ACTIONS(2450),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2450),
    [anon_sym_NS_INLINE] = ACTIONS(2450),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2450),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2450),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2450),
    [anon_sym_const] = ACTIONS(2450),
    [anon_sym_constexpr] = ACTIONS(2450),
    [anon_sym_volatile] = ACTIONS(2450),
    [anon_sym_restrict] = ACTIONS(2450),
    [anon_sym___restrict__] = ACTIONS(2450),
    [anon_sym__Atomic] = ACTIONS(2450),
    [anon_sym__Noreturn] = ACTIONS(2450),
    [anon_sym__Nonnull] = ACTIONS(2450),
    [anon_sym_nullable] = ACTIONS(2450),
    [anon_sym__Complex] = ACTIONS(2450),
    [anon_sym__Nullable] = ACTIONS(2450),
    [anon_sym__Nullable_result] = ACTIONS(2450),
    [anon_sym__Null_unspecified] = ACTIONS(2450),
    [anon_sym___autoreleasing] = ACTIONS(2450),
    [anon_sym___block] = ACTIONS(2450),
    [anon_sym___bridge] = ACTIONS(2450),
    [anon_sym___bridge_retained] = ACTIONS(2450),
    [anon_sym___bridge_transfer] = ACTIONS(2450),
    [anon_sym___complex] = ACTIONS(2450),
    [anon_sym___const] = ACTIONS(2450),
    [anon_sym___imag] = ACTIONS(2450),
    [anon_sym___kindof] = ACTIONS(2450),
    [anon_sym___nonnull] = ACTIONS(2450),
    [anon_sym___nullable] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2450),
    [anon_sym___real] = ACTIONS(2450),
    [anon_sym___strong] = ACTIONS(2450),
    [anon_sym___unsafe_unretained] = ACTIONS(2450),
    [anon_sym___unused] = ACTIONS(2450),
    [anon_sym___weak] = ACTIONS(2450),
    [anon_sym_alignas] = ACTIONS(2450),
    [anon_sym__Alignas] = ACTIONS(2450),
    [sym_primitive_type] = ACTIONS(2450),
    [anon_sym_enum] = ACTIONS(2450),
    [anon_sym_struct] = ACTIONS(2450),
    [anon_sym_union] = ACTIONS(2450),
    [anon_sym_if] = ACTIONS(2450),
    [anon_sym_else] = ACTIONS(2450),
    [anon_sym_switch] = ACTIONS(2450),
    [anon_sym_case] = ACTIONS(2450),
    [anon_sym_default] = ACTIONS(2450),
    [anon_sym_while] = ACTIONS(2450),
    [anon_sym_do] = ACTIONS(2450),
    [anon_sym_for] = ACTIONS(2450),
    [anon_sym_in] = ACTIONS(2450),
    [anon_sym_return] = ACTIONS(2450),
    [anon_sym_break] = ACTIONS(2450),
    [anon_sym_continue] = ACTIONS(2450),
    [anon_sym_goto] = ACTIONS(2450),
    [anon_sym_DASH_DASH] = ACTIONS(2452),
    [anon_sym_PLUS_PLUS] = ACTIONS(2452),
    [anon_sym_sizeof] = ACTIONS(2450),
    [anon_sym___alignof__] = ACTIONS(2450),
    [anon_sym___alignof] = ACTIONS(2450),
    [anon_sym__alignof] = ACTIONS(2450),
    [anon_sym_alignof] = ACTIONS(2450),
    [anon_sym__Alignof] = ACTIONS(2450),
    [anon_sym_offsetof] = ACTIONS(2450),
    [anon_sym__Generic] = ACTIONS(2450),
    [anon_sym_asm] = ACTIONS(2450),
    [anon_sym___asm__] = ACTIONS(2450),
    [anon_sym___asm] = ACTIONS(2450),
    [sym_number_literal] = ACTIONS(2452),
    [anon_sym_L_SQUOTE] = ACTIONS(2452),
    [anon_sym_u_SQUOTE] = ACTIONS(2452),
    [anon_sym_U_SQUOTE] = ACTIONS(2452),
    [anon_sym_u8_SQUOTE] = ACTIONS(2452),
    [anon_sym_SQUOTE] = ACTIONS(2452),
    [anon_sym_AT] = ACTIONS(2450),
    [anon_sym_DQUOTE] = ACTIONS(2452),
    [anon_sym_L_DQUOTE] = ACTIONS(2452),
    [anon_sym_u_DQUOTE] = ACTIONS(2452),
    [anon_sym_U_DQUOTE] = ACTIONS(2452),
    [anon_sym_u8_DQUOTE] = ACTIONS(2452),
    [sym_true] = ACTIONS(2450),
    [sym_false] = ACTIONS(2450),
    [anon_sym_NULL] = ACTIONS(2450),
    [anon_sym_nullptr] = ACTIONS(2450),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2450),
    [anon_sym___typeof] = ACTIONS(2450),
    [anon_sym_typeof] = ACTIONS(2450),
    [anon_sym_ATimport] = ACTIONS(2452),
    [aux_sym_preproc_undef_token1] = ACTIONS(2450),
    [anon_sym_POUND] = ACTIONS(2450),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2450),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2450),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2450),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_API_AVAILABLE] = ACTIONS(2450),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_API_DEPRECATED] = ACTIONS(2450),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym___deprecated_msg] = ACTIONS(2450),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2450),
    [anon_sym_ATprotocol] = ACTIONS(2452),
    [anon_sym_ATinterface] = ACTIONS(2452),
    [anon_sym_ATimplementation] = ACTIONS(2452),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2452),
    [anon_sym_ATtry] = ACTIONS(2452),
    [anon_sym___try] = ACTIONS(2450),
    [anon_sym_ATthrow] = ACTIONS(2452),
    [anon_sym_ATselector] = ACTIONS(2452),
    [anon_sym_ATavailable] = ACTIONS(2452),
    [anon_sym___builtin_available] = ACTIONS(2450),
    [anon_sym_va_arg] = ACTIONS(2450),
    [anon_sym_ATencode] = ACTIONS(2452),
    [anon_sym_ATsynchronized] = ACTIONS(2452),
    [anon_sym_BOOL] = ACTIONS(2450),
    [anon_sym_IMP] = ACTIONS(2450),
    [anon_sym_SEL] = ACTIONS(2450),
    [anon_sym_Class] = ACTIONS(2450),
    [anon_sym_id] = ACTIONS(2450),
  },
  [STATE(709)] = {
    [ts_builtin_sym_end] = ACTIONS(2452),
    [sym_identifier] = ACTIONS(2450),
    [aux_sym_preproc_include_token1] = ACTIONS(2450),
    [aux_sym_preproc_include_token2] = ACTIONS(2450),
    [aux_sym_preproc_def_token1] = ACTIONS(2450),
    [aux_sym_preproc_if_token1] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2450),
    [sym_preproc_directive] = ACTIONS(2450),
    [anon_sym_LPAREN2] = ACTIONS(2452),
    [anon_sym_BANG] = ACTIONS(2452),
    [anon_sym_TILDE] = ACTIONS(2452),
    [anon_sym_DASH] = ACTIONS(2450),
    [anon_sym_PLUS] = ACTIONS(2450),
    [anon_sym_STAR] = ACTIONS(2452),
    [anon_sym_CARET] = ACTIONS(2452),
    [anon_sym_AMP] = ACTIONS(2452),
    [anon_sym_SEMI] = ACTIONS(2452),
    [anon_sym___extension__] = ACTIONS(2450),
    [anon_sym_typedef] = ACTIONS(2450),
    [anon_sym_extern] = ACTIONS(2450),
    [anon_sym___attribute__] = ACTIONS(2450),
    [anon_sym___attribute] = ACTIONS(2450),
    [anon_sym_noreturn] = ACTIONS(2450),
    [anon_sym_LBRACK] = ACTIONS(2452),
    [anon_sym___declspec] = ACTIONS(2450),
    [anon_sym___cdecl] = ACTIONS(2450),
    [anon_sym___clrcall] = ACTIONS(2450),
    [anon_sym___stdcall] = ACTIONS(2450),
    [anon_sym___fastcall] = ACTIONS(2450),
    [anon_sym___thiscall] = ACTIONS(2450),
    [anon_sym___vectorcall] = ACTIONS(2450),
    [anon_sym_LBRACE] = ACTIONS(2452),
    [anon_sym_signed] = ACTIONS(2450),
    [anon_sym_unsigned] = ACTIONS(2450),
    [anon_sym_long] = ACTIONS(2450),
    [anon_sym_short] = ACTIONS(2450),
    [anon_sym_static] = ACTIONS(2450),
    [anon_sym_ATautoreleasepool] = ACTIONS(2452),
    [anon_sym_auto] = ACTIONS(2450),
    [anon_sym_register] = ACTIONS(2450),
    [anon_sym_inline] = ACTIONS(2450),
    [anon_sym___inline] = ACTIONS(2450),
    [anon_sym___inline__] = ACTIONS(2450),
    [anon_sym___forceinline] = ACTIONS(2450),
    [anon_sym_thread_local] = ACTIONS(2450),
    [anon_sym___thread] = ACTIONS(2450),
    [anon_sym_CG_EXTERN] = ACTIONS(2450),
    [anon_sym_CG_INLINE] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2450),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2450),
    [anon_sym_IBOutlet] = ACTIONS(2450),
    [anon_sym_IBInspectable] = ACTIONS(2450),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2450),
    [anon_sym_NS_INLINE] = ACTIONS(2450),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2450),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2450),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2450),
    [anon_sym_const] = ACTIONS(2450),
    [anon_sym_constexpr] = ACTIONS(2450),
    [anon_sym_volatile] = ACTIONS(2450),
    [anon_sym_restrict] = ACTIONS(2450),
    [anon_sym___restrict__] = ACTIONS(2450),
    [anon_sym__Atomic] = ACTIONS(2450),
    [anon_sym__Noreturn] = ACTIONS(2450),
    [anon_sym__Nonnull] = ACTIONS(2450),
    [anon_sym_nullable] = ACTIONS(2450),
    [anon_sym__Complex] = ACTIONS(2450),
    [anon_sym__Nullable] = ACTIONS(2450),
    [anon_sym__Nullable_result] = ACTIONS(2450),
    [anon_sym__Null_unspecified] = ACTIONS(2450),
    [anon_sym___autoreleasing] = ACTIONS(2450),
    [anon_sym___block] = ACTIONS(2450),
    [anon_sym___bridge] = ACTIONS(2450),
    [anon_sym___bridge_retained] = ACTIONS(2450),
    [anon_sym___bridge_transfer] = ACTIONS(2450),
    [anon_sym___complex] = ACTIONS(2450),
    [anon_sym___const] = ACTIONS(2450),
    [anon_sym___imag] = ACTIONS(2450),
    [anon_sym___kindof] = ACTIONS(2450),
    [anon_sym___nonnull] = ACTIONS(2450),
    [anon_sym___nullable] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2450),
    [anon_sym___real] = ACTIONS(2450),
    [anon_sym___strong] = ACTIONS(2450),
    [anon_sym___unsafe_unretained] = ACTIONS(2450),
    [anon_sym___unused] = ACTIONS(2450),
    [anon_sym___weak] = ACTIONS(2450),
    [anon_sym_alignas] = ACTIONS(2450),
    [anon_sym__Alignas] = ACTIONS(2450),
    [sym_primitive_type] = ACTIONS(2450),
    [anon_sym_enum] = ACTIONS(2450),
    [anon_sym_struct] = ACTIONS(2450),
    [anon_sym_union] = ACTIONS(2450),
    [anon_sym_if] = ACTIONS(2450),
    [anon_sym_else] = ACTIONS(2450),
    [anon_sym_switch] = ACTIONS(2450),
    [anon_sym_case] = ACTIONS(2450),
    [anon_sym_default] = ACTIONS(2450),
    [anon_sym_while] = ACTIONS(2450),
    [anon_sym_do] = ACTIONS(2450),
    [anon_sym_for] = ACTIONS(2450),
    [anon_sym_in] = ACTIONS(2450),
    [anon_sym_return] = ACTIONS(2450),
    [anon_sym_break] = ACTIONS(2450),
    [anon_sym_continue] = ACTIONS(2450),
    [anon_sym_goto] = ACTIONS(2450),
    [anon_sym_DASH_DASH] = ACTIONS(2452),
    [anon_sym_PLUS_PLUS] = ACTIONS(2452),
    [anon_sym_sizeof] = ACTIONS(2450),
    [anon_sym___alignof__] = ACTIONS(2450),
    [anon_sym___alignof] = ACTIONS(2450),
    [anon_sym__alignof] = ACTIONS(2450),
    [anon_sym_alignof] = ACTIONS(2450),
    [anon_sym__Alignof] = ACTIONS(2450),
    [anon_sym_offsetof] = ACTIONS(2450),
    [anon_sym__Generic] = ACTIONS(2450),
    [anon_sym_asm] = ACTIONS(2450),
    [anon_sym___asm__] = ACTIONS(2450),
    [anon_sym___asm] = ACTIONS(2450),
    [sym_number_literal] = ACTIONS(2452),
    [anon_sym_L_SQUOTE] = ACTIONS(2452),
    [anon_sym_u_SQUOTE] = ACTIONS(2452),
    [anon_sym_U_SQUOTE] = ACTIONS(2452),
    [anon_sym_u8_SQUOTE] = ACTIONS(2452),
    [anon_sym_SQUOTE] = ACTIONS(2452),
    [anon_sym_AT] = ACTIONS(2450),
    [anon_sym_DQUOTE] = ACTIONS(2452),
    [anon_sym_L_DQUOTE] = ACTIONS(2452),
    [anon_sym_u_DQUOTE] = ACTIONS(2452),
    [anon_sym_U_DQUOTE] = ACTIONS(2452),
    [anon_sym_u8_DQUOTE] = ACTIONS(2452),
    [sym_true] = ACTIONS(2450),
    [sym_false] = ACTIONS(2450),
    [anon_sym_NULL] = ACTIONS(2450),
    [anon_sym_nullptr] = ACTIONS(2450),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2450),
    [anon_sym___typeof] = ACTIONS(2450),
    [anon_sym_typeof] = ACTIONS(2450),
    [anon_sym_ATimport] = ACTIONS(2452),
    [aux_sym_preproc_undef_token1] = ACTIONS(2450),
    [anon_sym_POUND] = ACTIONS(2450),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2450),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2450),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2450),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_API_AVAILABLE] = ACTIONS(2450),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_API_DEPRECATED] = ACTIONS(2450),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym___deprecated_msg] = ACTIONS(2450),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2450),
    [anon_sym_ATprotocol] = ACTIONS(2452),
    [anon_sym_ATinterface] = ACTIONS(2452),
    [anon_sym_ATimplementation] = ACTIONS(2452),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2452),
    [anon_sym_ATtry] = ACTIONS(2452),
    [anon_sym___try] = ACTIONS(2450),
    [anon_sym_ATthrow] = ACTIONS(2452),
    [anon_sym_ATselector] = ACTIONS(2452),
    [anon_sym_ATavailable] = ACTIONS(2452),
    [anon_sym___builtin_available] = ACTIONS(2450),
    [anon_sym_va_arg] = ACTIONS(2450),
    [anon_sym_ATencode] = ACTIONS(2452),
    [anon_sym_ATsynchronized] = ACTIONS(2452),
    [anon_sym_BOOL] = ACTIONS(2450),
    [anon_sym_IMP] = ACTIONS(2450),
    [anon_sym_SEL] = ACTIONS(2450),
    [anon_sym_Class] = ACTIONS(2450),
    [anon_sym_id] = ACTIONS(2450),
  },
  [STATE(710)] = {
    [ts_builtin_sym_end] = ACTIONS(2512),
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(711)] = {
    [ts_builtin_sym_end] = ACTIONS(2512),
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(712)] = {
    [sym_identifier] = ACTIONS(2470),
    [aux_sym_preproc_include_token1] = ACTIONS(2470),
    [aux_sym_preproc_include_token2] = ACTIONS(2470),
    [aux_sym_preproc_def_token1] = ACTIONS(2470),
    [aux_sym_preproc_if_token1] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2470),
    [sym_preproc_directive] = ACTIONS(2470),
    [anon_sym_LPAREN2] = ACTIONS(2472),
    [anon_sym_BANG] = ACTIONS(2472),
    [anon_sym_TILDE] = ACTIONS(2472),
    [anon_sym_DASH] = ACTIONS(2470),
    [anon_sym_PLUS] = ACTIONS(2470),
    [anon_sym_STAR] = ACTIONS(2472),
    [anon_sym_CARET] = ACTIONS(2472),
    [anon_sym_AMP] = ACTIONS(2472),
    [anon_sym_SEMI] = ACTIONS(2472),
    [anon_sym___extension__] = ACTIONS(2470),
    [anon_sym_typedef] = ACTIONS(2470),
    [anon_sym_extern] = ACTIONS(2470),
    [anon_sym___attribute__] = ACTIONS(2470),
    [anon_sym___attribute] = ACTIONS(2470),
    [anon_sym_noreturn] = ACTIONS(2470),
    [anon_sym_LBRACK] = ACTIONS(2472),
    [anon_sym___declspec] = ACTIONS(2470),
    [anon_sym___cdecl] = ACTIONS(2470),
    [anon_sym___clrcall] = ACTIONS(2470),
    [anon_sym___stdcall] = ACTIONS(2470),
    [anon_sym___fastcall] = ACTIONS(2470),
    [anon_sym___thiscall] = ACTIONS(2470),
    [anon_sym___vectorcall] = ACTIONS(2470),
    [anon_sym_LBRACE] = ACTIONS(2472),
    [anon_sym_RBRACE] = ACTIONS(2472),
    [anon_sym_signed] = ACTIONS(2470),
    [anon_sym_unsigned] = ACTIONS(2470),
    [anon_sym_long] = ACTIONS(2470),
    [anon_sym_short] = ACTIONS(2470),
    [anon_sym_static] = ACTIONS(2470),
    [anon_sym_ATautoreleasepool] = ACTIONS(2472),
    [anon_sym_auto] = ACTIONS(2470),
    [anon_sym_register] = ACTIONS(2470),
    [anon_sym_inline] = ACTIONS(2470),
    [anon_sym___inline] = ACTIONS(2470),
    [anon_sym___inline__] = ACTIONS(2470),
    [anon_sym___forceinline] = ACTIONS(2470),
    [anon_sym_thread_local] = ACTIONS(2470),
    [anon_sym___thread] = ACTIONS(2470),
    [anon_sym_CG_EXTERN] = ACTIONS(2470),
    [anon_sym_CG_INLINE] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2470),
    [anon_sym_IBOutlet] = ACTIONS(2470),
    [anon_sym_IBInspectable] = ACTIONS(2470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2470),
    [anon_sym_NS_INLINE] = ACTIONS(2470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2470),
    [anon_sym_const] = ACTIONS(2470),
    [anon_sym_constexpr] = ACTIONS(2470),
    [anon_sym_volatile] = ACTIONS(2470),
    [anon_sym_restrict] = ACTIONS(2470),
    [anon_sym___restrict__] = ACTIONS(2470),
    [anon_sym__Atomic] = ACTIONS(2470),
    [anon_sym__Noreturn] = ACTIONS(2470),
    [anon_sym__Nonnull] = ACTIONS(2470),
    [anon_sym_nullable] = ACTIONS(2470),
    [anon_sym__Complex] = ACTIONS(2470),
    [anon_sym__Nullable] = ACTIONS(2470),
    [anon_sym__Nullable_result] = ACTIONS(2470),
    [anon_sym__Null_unspecified] = ACTIONS(2470),
    [anon_sym___autoreleasing] = ACTIONS(2470),
    [anon_sym___block] = ACTIONS(2470),
    [anon_sym___bridge] = ACTIONS(2470),
    [anon_sym___bridge_retained] = ACTIONS(2470),
    [anon_sym___bridge_transfer] = ACTIONS(2470),
    [anon_sym___complex] = ACTIONS(2470),
    [anon_sym___const] = ACTIONS(2470),
    [anon_sym___imag] = ACTIONS(2470),
    [anon_sym___kindof] = ACTIONS(2470),
    [anon_sym___nonnull] = ACTIONS(2470),
    [anon_sym___nullable] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2470),
    [anon_sym___real] = ACTIONS(2470),
    [anon_sym___strong] = ACTIONS(2470),
    [anon_sym___unsafe_unretained] = ACTIONS(2470),
    [anon_sym___unused] = ACTIONS(2470),
    [anon_sym___weak] = ACTIONS(2470),
    [anon_sym_alignas] = ACTIONS(2470),
    [anon_sym__Alignas] = ACTIONS(2470),
    [sym_primitive_type] = ACTIONS(2470),
    [anon_sym_enum] = ACTIONS(2470),
    [anon_sym_struct] = ACTIONS(2470),
    [anon_sym_union] = ACTIONS(2470),
    [anon_sym_if] = ACTIONS(2470),
    [anon_sym_else] = ACTIONS(2470),
    [anon_sym_switch] = ACTIONS(2470),
    [anon_sym_case] = ACTIONS(2470),
    [anon_sym_default] = ACTIONS(2470),
    [anon_sym_while] = ACTIONS(2470),
    [anon_sym_do] = ACTIONS(2470),
    [anon_sym_for] = ACTIONS(2470),
    [anon_sym_in] = ACTIONS(2470),
    [anon_sym_return] = ACTIONS(2470),
    [anon_sym_break] = ACTIONS(2470),
    [anon_sym_continue] = ACTIONS(2470),
    [anon_sym_goto] = ACTIONS(2470),
    [anon_sym_DASH_DASH] = ACTIONS(2472),
    [anon_sym_PLUS_PLUS] = ACTIONS(2472),
    [anon_sym_sizeof] = ACTIONS(2470),
    [anon_sym___alignof__] = ACTIONS(2470),
    [anon_sym___alignof] = ACTIONS(2470),
    [anon_sym__alignof] = ACTIONS(2470),
    [anon_sym_alignof] = ACTIONS(2470),
    [anon_sym__Alignof] = ACTIONS(2470),
    [anon_sym_offsetof] = ACTIONS(2470),
    [anon_sym__Generic] = ACTIONS(2470),
    [anon_sym_asm] = ACTIONS(2470),
    [anon_sym___asm__] = ACTIONS(2470),
    [anon_sym___asm] = ACTIONS(2470),
    [sym_number_literal] = ACTIONS(2472),
    [anon_sym_L_SQUOTE] = ACTIONS(2472),
    [anon_sym_u_SQUOTE] = ACTIONS(2472),
    [anon_sym_U_SQUOTE] = ACTIONS(2472),
    [anon_sym_u8_SQUOTE] = ACTIONS(2472),
    [anon_sym_SQUOTE] = ACTIONS(2472),
    [anon_sym_AT] = ACTIONS(2470),
    [anon_sym_DQUOTE] = ACTIONS(2472),
    [anon_sym_L_DQUOTE] = ACTIONS(2472),
    [anon_sym_u_DQUOTE] = ACTIONS(2472),
    [anon_sym_U_DQUOTE] = ACTIONS(2472),
    [anon_sym_u8_DQUOTE] = ACTIONS(2472),
    [sym_true] = ACTIONS(2470),
    [sym_false] = ACTIONS(2470),
    [anon_sym_NULL] = ACTIONS(2470),
    [anon_sym_nullptr] = ACTIONS(2470),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2470),
    [anon_sym___typeof] = ACTIONS(2470),
    [anon_sym_typeof] = ACTIONS(2470),
    [anon_sym_ATimport] = ACTIONS(2472),
    [aux_sym_preproc_undef_token1] = ACTIONS(2470),
    [anon_sym_POUND] = ACTIONS(2470),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_API_AVAILABLE] = ACTIONS(2470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_API_DEPRECATED] = ACTIONS(2470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym___deprecated_msg] = ACTIONS(2470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2470),
    [anon_sym_ATprotocol] = ACTIONS(2472),
    [anon_sym_ATinterface] = ACTIONS(2472),
    [anon_sym_ATimplementation] = ACTIONS(2472),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2472),
    [anon_sym_ATtry] = ACTIONS(2472),
    [anon_sym___try] = ACTIONS(2470),
    [anon_sym_ATthrow] = ACTIONS(2472),
    [anon_sym_ATselector] = ACTIONS(2472),
    [anon_sym_ATavailable] = ACTIONS(2472),
    [anon_sym___builtin_available] = ACTIONS(2470),
    [anon_sym_va_arg] = ACTIONS(2470),
    [anon_sym_ATencode] = ACTIONS(2472),
    [anon_sym_ATsynchronized] = ACTIONS(2472),
    [anon_sym_BOOL] = ACTIONS(2470),
    [anon_sym_IMP] = ACTIONS(2470),
    [anon_sym_SEL] = ACTIONS(2470),
    [anon_sym_Class] = ACTIONS(2470),
    [anon_sym_id] = ACTIONS(2470),
  },
  [STATE(713)] = {
    [ts_builtin_sym_end] = ACTIONS(2460),
    [sym_identifier] = ACTIONS(2458),
    [aux_sym_preproc_include_token1] = ACTIONS(2458),
    [aux_sym_preproc_include_token2] = ACTIONS(2458),
    [aux_sym_preproc_def_token1] = ACTIONS(2458),
    [aux_sym_preproc_if_token1] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2458),
    [sym_preproc_directive] = ACTIONS(2458),
    [anon_sym_LPAREN2] = ACTIONS(2460),
    [anon_sym_BANG] = ACTIONS(2460),
    [anon_sym_TILDE] = ACTIONS(2460),
    [anon_sym_DASH] = ACTIONS(2458),
    [anon_sym_PLUS] = ACTIONS(2458),
    [anon_sym_STAR] = ACTIONS(2460),
    [anon_sym_CARET] = ACTIONS(2460),
    [anon_sym_AMP] = ACTIONS(2460),
    [anon_sym_SEMI] = ACTIONS(2460),
    [anon_sym___extension__] = ACTIONS(2458),
    [anon_sym_typedef] = ACTIONS(2458),
    [anon_sym_extern] = ACTIONS(2458),
    [anon_sym___attribute__] = ACTIONS(2458),
    [anon_sym___attribute] = ACTIONS(2458),
    [anon_sym_noreturn] = ACTIONS(2458),
    [anon_sym_LBRACK] = ACTIONS(2460),
    [anon_sym___declspec] = ACTIONS(2458),
    [anon_sym___cdecl] = ACTIONS(2458),
    [anon_sym___clrcall] = ACTIONS(2458),
    [anon_sym___stdcall] = ACTIONS(2458),
    [anon_sym___fastcall] = ACTIONS(2458),
    [anon_sym___thiscall] = ACTIONS(2458),
    [anon_sym___vectorcall] = ACTIONS(2458),
    [anon_sym_LBRACE] = ACTIONS(2460),
    [anon_sym_signed] = ACTIONS(2458),
    [anon_sym_unsigned] = ACTIONS(2458),
    [anon_sym_long] = ACTIONS(2458),
    [anon_sym_short] = ACTIONS(2458),
    [anon_sym_static] = ACTIONS(2458),
    [anon_sym_ATautoreleasepool] = ACTIONS(2460),
    [anon_sym_auto] = ACTIONS(2458),
    [anon_sym_register] = ACTIONS(2458),
    [anon_sym_inline] = ACTIONS(2458),
    [anon_sym___inline] = ACTIONS(2458),
    [anon_sym___inline__] = ACTIONS(2458),
    [anon_sym___forceinline] = ACTIONS(2458),
    [anon_sym_thread_local] = ACTIONS(2458),
    [anon_sym___thread] = ACTIONS(2458),
    [anon_sym_CG_EXTERN] = ACTIONS(2458),
    [anon_sym_CG_INLINE] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2458),
    [anon_sym_IBOutlet] = ACTIONS(2458),
    [anon_sym_IBInspectable] = ACTIONS(2458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2458),
    [anon_sym_NS_INLINE] = ACTIONS(2458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2458),
    [anon_sym_const] = ACTIONS(2458),
    [anon_sym_constexpr] = ACTIONS(2458),
    [anon_sym_volatile] = ACTIONS(2458),
    [anon_sym_restrict] = ACTIONS(2458),
    [anon_sym___restrict__] = ACTIONS(2458),
    [anon_sym__Atomic] = ACTIONS(2458),
    [anon_sym__Noreturn] = ACTIONS(2458),
    [anon_sym__Nonnull] = ACTIONS(2458),
    [anon_sym_nullable] = ACTIONS(2458),
    [anon_sym__Complex] = ACTIONS(2458),
    [anon_sym__Nullable] = ACTIONS(2458),
    [anon_sym__Nullable_result] = ACTIONS(2458),
    [anon_sym__Null_unspecified] = ACTIONS(2458),
    [anon_sym___autoreleasing] = ACTIONS(2458),
    [anon_sym___block] = ACTIONS(2458),
    [anon_sym___bridge] = ACTIONS(2458),
    [anon_sym___bridge_retained] = ACTIONS(2458),
    [anon_sym___bridge_transfer] = ACTIONS(2458),
    [anon_sym___complex] = ACTIONS(2458),
    [anon_sym___const] = ACTIONS(2458),
    [anon_sym___imag] = ACTIONS(2458),
    [anon_sym___kindof] = ACTIONS(2458),
    [anon_sym___nonnull] = ACTIONS(2458),
    [anon_sym___nullable] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2458),
    [anon_sym___real] = ACTIONS(2458),
    [anon_sym___strong] = ACTIONS(2458),
    [anon_sym___unsafe_unretained] = ACTIONS(2458),
    [anon_sym___unused] = ACTIONS(2458),
    [anon_sym___weak] = ACTIONS(2458),
    [anon_sym_alignas] = ACTIONS(2458),
    [anon_sym__Alignas] = ACTIONS(2458),
    [sym_primitive_type] = ACTIONS(2458),
    [anon_sym_enum] = ACTIONS(2458),
    [anon_sym_struct] = ACTIONS(2458),
    [anon_sym_union] = ACTIONS(2458),
    [anon_sym_if] = ACTIONS(2458),
    [anon_sym_else] = ACTIONS(2458),
    [anon_sym_switch] = ACTIONS(2458),
    [anon_sym_case] = ACTIONS(2458),
    [anon_sym_default] = ACTIONS(2458),
    [anon_sym_while] = ACTIONS(2458),
    [anon_sym_do] = ACTIONS(2458),
    [anon_sym_for] = ACTIONS(2458),
    [anon_sym_in] = ACTIONS(2458),
    [anon_sym_return] = ACTIONS(2458),
    [anon_sym_break] = ACTIONS(2458),
    [anon_sym_continue] = ACTIONS(2458),
    [anon_sym_goto] = ACTIONS(2458),
    [anon_sym_DASH_DASH] = ACTIONS(2460),
    [anon_sym_PLUS_PLUS] = ACTIONS(2460),
    [anon_sym_sizeof] = ACTIONS(2458),
    [anon_sym___alignof__] = ACTIONS(2458),
    [anon_sym___alignof] = ACTIONS(2458),
    [anon_sym__alignof] = ACTIONS(2458),
    [anon_sym_alignof] = ACTIONS(2458),
    [anon_sym__Alignof] = ACTIONS(2458),
    [anon_sym_offsetof] = ACTIONS(2458),
    [anon_sym__Generic] = ACTIONS(2458),
    [anon_sym_asm] = ACTIONS(2458),
    [anon_sym___asm__] = ACTIONS(2458),
    [anon_sym___asm] = ACTIONS(2458),
    [sym_number_literal] = ACTIONS(2460),
    [anon_sym_L_SQUOTE] = ACTIONS(2460),
    [anon_sym_u_SQUOTE] = ACTIONS(2460),
    [anon_sym_U_SQUOTE] = ACTIONS(2460),
    [anon_sym_u8_SQUOTE] = ACTIONS(2460),
    [anon_sym_SQUOTE] = ACTIONS(2460),
    [anon_sym_AT] = ACTIONS(2458),
    [anon_sym_DQUOTE] = ACTIONS(2460),
    [anon_sym_L_DQUOTE] = ACTIONS(2460),
    [anon_sym_u_DQUOTE] = ACTIONS(2460),
    [anon_sym_U_DQUOTE] = ACTIONS(2460),
    [anon_sym_u8_DQUOTE] = ACTIONS(2460),
    [sym_true] = ACTIONS(2458),
    [sym_false] = ACTIONS(2458),
    [anon_sym_NULL] = ACTIONS(2458),
    [anon_sym_nullptr] = ACTIONS(2458),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2458),
    [anon_sym___typeof] = ACTIONS(2458),
    [anon_sym_typeof] = ACTIONS(2458),
    [anon_sym_ATimport] = ACTIONS(2460),
    [aux_sym_preproc_undef_token1] = ACTIONS(2458),
    [anon_sym_POUND] = ACTIONS(2458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_API_AVAILABLE] = ACTIONS(2458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_API_DEPRECATED] = ACTIONS(2458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym___deprecated_msg] = ACTIONS(2458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2458),
    [anon_sym_ATprotocol] = ACTIONS(2460),
    [anon_sym_ATinterface] = ACTIONS(2460),
    [anon_sym_ATimplementation] = ACTIONS(2460),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2460),
    [anon_sym_ATtry] = ACTIONS(2460),
    [anon_sym___try] = ACTIONS(2458),
    [anon_sym_ATthrow] = ACTIONS(2460),
    [anon_sym_ATselector] = ACTIONS(2460),
    [anon_sym_ATavailable] = ACTIONS(2460),
    [anon_sym___builtin_available] = ACTIONS(2458),
    [anon_sym_va_arg] = ACTIONS(2458),
    [anon_sym_ATencode] = ACTIONS(2460),
    [anon_sym_ATsynchronized] = ACTIONS(2460),
    [anon_sym_BOOL] = ACTIONS(2458),
    [anon_sym_IMP] = ACTIONS(2458),
    [anon_sym_SEL] = ACTIONS(2458),
    [anon_sym_Class] = ACTIONS(2458),
    [anon_sym_id] = ACTIONS(2458),
  },
  [STATE(714)] = {
    [ts_builtin_sym_end] = ACTIONS(2464),
    [sym_identifier] = ACTIONS(2462),
    [aux_sym_preproc_include_token1] = ACTIONS(2462),
    [aux_sym_preproc_include_token2] = ACTIONS(2462),
    [aux_sym_preproc_def_token1] = ACTIONS(2462),
    [aux_sym_preproc_if_token1] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2462),
    [sym_preproc_directive] = ACTIONS(2462),
    [anon_sym_LPAREN2] = ACTIONS(2464),
    [anon_sym_BANG] = ACTIONS(2464),
    [anon_sym_TILDE] = ACTIONS(2464),
    [anon_sym_DASH] = ACTIONS(2462),
    [anon_sym_PLUS] = ACTIONS(2462),
    [anon_sym_STAR] = ACTIONS(2464),
    [anon_sym_CARET] = ACTIONS(2464),
    [anon_sym_AMP] = ACTIONS(2464),
    [anon_sym_SEMI] = ACTIONS(2464),
    [anon_sym___extension__] = ACTIONS(2462),
    [anon_sym_typedef] = ACTIONS(2462),
    [anon_sym_extern] = ACTIONS(2462),
    [anon_sym___attribute__] = ACTIONS(2462),
    [anon_sym___attribute] = ACTIONS(2462),
    [anon_sym_noreturn] = ACTIONS(2462),
    [anon_sym_LBRACK] = ACTIONS(2464),
    [anon_sym___declspec] = ACTIONS(2462),
    [anon_sym___cdecl] = ACTIONS(2462),
    [anon_sym___clrcall] = ACTIONS(2462),
    [anon_sym___stdcall] = ACTIONS(2462),
    [anon_sym___fastcall] = ACTIONS(2462),
    [anon_sym___thiscall] = ACTIONS(2462),
    [anon_sym___vectorcall] = ACTIONS(2462),
    [anon_sym_LBRACE] = ACTIONS(2464),
    [anon_sym_signed] = ACTIONS(2462),
    [anon_sym_unsigned] = ACTIONS(2462),
    [anon_sym_long] = ACTIONS(2462),
    [anon_sym_short] = ACTIONS(2462),
    [anon_sym_static] = ACTIONS(2462),
    [anon_sym_ATautoreleasepool] = ACTIONS(2464),
    [anon_sym_auto] = ACTIONS(2462),
    [anon_sym_register] = ACTIONS(2462),
    [anon_sym_inline] = ACTIONS(2462),
    [anon_sym___inline] = ACTIONS(2462),
    [anon_sym___inline__] = ACTIONS(2462),
    [anon_sym___forceinline] = ACTIONS(2462),
    [anon_sym_thread_local] = ACTIONS(2462),
    [anon_sym___thread] = ACTIONS(2462),
    [anon_sym_CG_EXTERN] = ACTIONS(2462),
    [anon_sym_CG_INLINE] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2462),
    [anon_sym_IBOutlet] = ACTIONS(2462),
    [anon_sym_IBInspectable] = ACTIONS(2462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2462),
    [anon_sym_NS_INLINE] = ACTIONS(2462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2462),
    [anon_sym_const] = ACTIONS(2462),
    [anon_sym_constexpr] = ACTIONS(2462),
    [anon_sym_volatile] = ACTIONS(2462),
    [anon_sym_restrict] = ACTIONS(2462),
    [anon_sym___restrict__] = ACTIONS(2462),
    [anon_sym__Atomic] = ACTIONS(2462),
    [anon_sym__Noreturn] = ACTIONS(2462),
    [anon_sym__Nonnull] = ACTIONS(2462),
    [anon_sym_nullable] = ACTIONS(2462),
    [anon_sym__Complex] = ACTIONS(2462),
    [anon_sym__Nullable] = ACTIONS(2462),
    [anon_sym__Nullable_result] = ACTIONS(2462),
    [anon_sym__Null_unspecified] = ACTIONS(2462),
    [anon_sym___autoreleasing] = ACTIONS(2462),
    [anon_sym___block] = ACTIONS(2462),
    [anon_sym___bridge] = ACTIONS(2462),
    [anon_sym___bridge_retained] = ACTIONS(2462),
    [anon_sym___bridge_transfer] = ACTIONS(2462),
    [anon_sym___complex] = ACTIONS(2462),
    [anon_sym___const] = ACTIONS(2462),
    [anon_sym___imag] = ACTIONS(2462),
    [anon_sym___kindof] = ACTIONS(2462),
    [anon_sym___nonnull] = ACTIONS(2462),
    [anon_sym___nullable] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2462),
    [anon_sym___real] = ACTIONS(2462),
    [anon_sym___strong] = ACTIONS(2462),
    [anon_sym___unsafe_unretained] = ACTIONS(2462),
    [anon_sym___unused] = ACTIONS(2462),
    [anon_sym___weak] = ACTIONS(2462),
    [anon_sym_alignas] = ACTIONS(2462),
    [anon_sym__Alignas] = ACTIONS(2462),
    [sym_primitive_type] = ACTIONS(2462),
    [anon_sym_enum] = ACTIONS(2462),
    [anon_sym_struct] = ACTIONS(2462),
    [anon_sym_union] = ACTIONS(2462),
    [anon_sym_if] = ACTIONS(2462),
    [anon_sym_else] = ACTIONS(2462),
    [anon_sym_switch] = ACTIONS(2462),
    [anon_sym_case] = ACTIONS(2462),
    [anon_sym_default] = ACTIONS(2462),
    [anon_sym_while] = ACTIONS(2462),
    [anon_sym_do] = ACTIONS(2462),
    [anon_sym_for] = ACTIONS(2462),
    [anon_sym_in] = ACTIONS(2462),
    [anon_sym_return] = ACTIONS(2462),
    [anon_sym_break] = ACTIONS(2462),
    [anon_sym_continue] = ACTIONS(2462),
    [anon_sym_goto] = ACTIONS(2462),
    [anon_sym_DASH_DASH] = ACTIONS(2464),
    [anon_sym_PLUS_PLUS] = ACTIONS(2464),
    [anon_sym_sizeof] = ACTIONS(2462),
    [anon_sym___alignof__] = ACTIONS(2462),
    [anon_sym___alignof] = ACTIONS(2462),
    [anon_sym__alignof] = ACTIONS(2462),
    [anon_sym_alignof] = ACTIONS(2462),
    [anon_sym__Alignof] = ACTIONS(2462),
    [anon_sym_offsetof] = ACTIONS(2462),
    [anon_sym__Generic] = ACTIONS(2462),
    [anon_sym_asm] = ACTIONS(2462),
    [anon_sym___asm__] = ACTIONS(2462),
    [anon_sym___asm] = ACTIONS(2462),
    [sym_number_literal] = ACTIONS(2464),
    [anon_sym_L_SQUOTE] = ACTIONS(2464),
    [anon_sym_u_SQUOTE] = ACTIONS(2464),
    [anon_sym_U_SQUOTE] = ACTIONS(2464),
    [anon_sym_u8_SQUOTE] = ACTIONS(2464),
    [anon_sym_SQUOTE] = ACTIONS(2464),
    [anon_sym_AT] = ACTIONS(2462),
    [anon_sym_DQUOTE] = ACTIONS(2464),
    [anon_sym_L_DQUOTE] = ACTIONS(2464),
    [anon_sym_u_DQUOTE] = ACTIONS(2464),
    [anon_sym_U_DQUOTE] = ACTIONS(2464),
    [anon_sym_u8_DQUOTE] = ACTIONS(2464),
    [sym_true] = ACTIONS(2462),
    [sym_false] = ACTIONS(2462),
    [anon_sym_NULL] = ACTIONS(2462),
    [anon_sym_nullptr] = ACTIONS(2462),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2462),
    [anon_sym___typeof] = ACTIONS(2462),
    [anon_sym_typeof] = ACTIONS(2462),
    [anon_sym_ATimport] = ACTIONS(2464),
    [aux_sym_preproc_undef_token1] = ACTIONS(2462),
    [anon_sym_POUND] = ACTIONS(2462),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2462),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_API_AVAILABLE] = ACTIONS(2462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_API_DEPRECATED] = ACTIONS(2462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym___deprecated_msg] = ACTIONS(2462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2462),
    [anon_sym_ATprotocol] = ACTIONS(2464),
    [anon_sym_ATinterface] = ACTIONS(2464),
    [anon_sym_ATimplementation] = ACTIONS(2464),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2464),
    [anon_sym_ATtry] = ACTIONS(2464),
    [anon_sym___try] = ACTIONS(2462),
    [anon_sym_ATthrow] = ACTIONS(2464),
    [anon_sym_ATselector] = ACTIONS(2464),
    [anon_sym_ATavailable] = ACTIONS(2464),
    [anon_sym___builtin_available] = ACTIONS(2462),
    [anon_sym_va_arg] = ACTIONS(2462),
    [anon_sym_ATencode] = ACTIONS(2464),
    [anon_sym_ATsynchronized] = ACTIONS(2464),
    [anon_sym_BOOL] = ACTIONS(2462),
    [anon_sym_IMP] = ACTIONS(2462),
    [anon_sym_SEL] = ACTIONS(2462),
    [anon_sym_Class] = ACTIONS(2462),
    [anon_sym_id] = ACTIONS(2462),
  },
  [STATE(715)] = {
    [ts_builtin_sym_end] = ACTIONS(2468),
    [sym_identifier] = ACTIONS(2466),
    [aux_sym_preproc_include_token1] = ACTIONS(2466),
    [aux_sym_preproc_include_token2] = ACTIONS(2466),
    [aux_sym_preproc_def_token1] = ACTIONS(2466),
    [aux_sym_preproc_if_token1] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2466),
    [sym_preproc_directive] = ACTIONS(2466),
    [anon_sym_LPAREN2] = ACTIONS(2468),
    [anon_sym_BANG] = ACTIONS(2468),
    [anon_sym_TILDE] = ACTIONS(2468),
    [anon_sym_DASH] = ACTIONS(2466),
    [anon_sym_PLUS] = ACTIONS(2466),
    [anon_sym_STAR] = ACTIONS(2468),
    [anon_sym_CARET] = ACTIONS(2468),
    [anon_sym_AMP] = ACTIONS(2468),
    [anon_sym_SEMI] = ACTIONS(2468),
    [anon_sym___extension__] = ACTIONS(2466),
    [anon_sym_typedef] = ACTIONS(2466),
    [anon_sym_extern] = ACTIONS(2466),
    [anon_sym___attribute__] = ACTIONS(2466),
    [anon_sym___attribute] = ACTIONS(2466),
    [anon_sym_noreturn] = ACTIONS(2466),
    [anon_sym_LBRACK] = ACTIONS(2468),
    [anon_sym___declspec] = ACTIONS(2466),
    [anon_sym___cdecl] = ACTIONS(2466),
    [anon_sym___clrcall] = ACTIONS(2466),
    [anon_sym___stdcall] = ACTIONS(2466),
    [anon_sym___fastcall] = ACTIONS(2466),
    [anon_sym___thiscall] = ACTIONS(2466),
    [anon_sym___vectorcall] = ACTIONS(2466),
    [anon_sym_LBRACE] = ACTIONS(2468),
    [anon_sym_signed] = ACTIONS(2466),
    [anon_sym_unsigned] = ACTIONS(2466),
    [anon_sym_long] = ACTIONS(2466),
    [anon_sym_short] = ACTIONS(2466),
    [anon_sym_static] = ACTIONS(2466),
    [anon_sym_ATautoreleasepool] = ACTIONS(2468),
    [anon_sym_auto] = ACTIONS(2466),
    [anon_sym_register] = ACTIONS(2466),
    [anon_sym_inline] = ACTIONS(2466),
    [anon_sym___inline] = ACTIONS(2466),
    [anon_sym___inline__] = ACTIONS(2466),
    [anon_sym___forceinline] = ACTIONS(2466),
    [anon_sym_thread_local] = ACTIONS(2466),
    [anon_sym___thread] = ACTIONS(2466),
    [anon_sym_CG_EXTERN] = ACTIONS(2466),
    [anon_sym_CG_INLINE] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2466),
    [anon_sym_IBOutlet] = ACTIONS(2466),
    [anon_sym_IBInspectable] = ACTIONS(2466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2466),
    [anon_sym_NS_INLINE] = ACTIONS(2466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2466),
    [anon_sym_const] = ACTIONS(2466),
    [anon_sym_constexpr] = ACTIONS(2466),
    [anon_sym_volatile] = ACTIONS(2466),
    [anon_sym_restrict] = ACTIONS(2466),
    [anon_sym___restrict__] = ACTIONS(2466),
    [anon_sym__Atomic] = ACTIONS(2466),
    [anon_sym__Noreturn] = ACTIONS(2466),
    [anon_sym__Nonnull] = ACTIONS(2466),
    [anon_sym_nullable] = ACTIONS(2466),
    [anon_sym__Complex] = ACTIONS(2466),
    [anon_sym__Nullable] = ACTIONS(2466),
    [anon_sym__Nullable_result] = ACTIONS(2466),
    [anon_sym__Null_unspecified] = ACTIONS(2466),
    [anon_sym___autoreleasing] = ACTIONS(2466),
    [anon_sym___block] = ACTIONS(2466),
    [anon_sym___bridge] = ACTIONS(2466),
    [anon_sym___bridge_retained] = ACTIONS(2466),
    [anon_sym___bridge_transfer] = ACTIONS(2466),
    [anon_sym___complex] = ACTIONS(2466),
    [anon_sym___const] = ACTIONS(2466),
    [anon_sym___imag] = ACTIONS(2466),
    [anon_sym___kindof] = ACTIONS(2466),
    [anon_sym___nonnull] = ACTIONS(2466),
    [anon_sym___nullable] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2466),
    [anon_sym___real] = ACTIONS(2466),
    [anon_sym___strong] = ACTIONS(2466),
    [anon_sym___unsafe_unretained] = ACTIONS(2466),
    [anon_sym___unused] = ACTIONS(2466),
    [anon_sym___weak] = ACTIONS(2466),
    [anon_sym_alignas] = ACTIONS(2466),
    [anon_sym__Alignas] = ACTIONS(2466),
    [sym_primitive_type] = ACTIONS(2466),
    [anon_sym_enum] = ACTIONS(2466),
    [anon_sym_struct] = ACTIONS(2466),
    [anon_sym_union] = ACTIONS(2466),
    [anon_sym_if] = ACTIONS(2466),
    [anon_sym_else] = ACTIONS(2466),
    [anon_sym_switch] = ACTIONS(2466),
    [anon_sym_case] = ACTIONS(2466),
    [anon_sym_default] = ACTIONS(2466),
    [anon_sym_while] = ACTIONS(2466),
    [anon_sym_do] = ACTIONS(2466),
    [anon_sym_for] = ACTIONS(2466),
    [anon_sym_in] = ACTIONS(2466),
    [anon_sym_return] = ACTIONS(2466),
    [anon_sym_break] = ACTIONS(2466),
    [anon_sym_continue] = ACTIONS(2466),
    [anon_sym_goto] = ACTIONS(2466),
    [anon_sym_DASH_DASH] = ACTIONS(2468),
    [anon_sym_PLUS_PLUS] = ACTIONS(2468),
    [anon_sym_sizeof] = ACTIONS(2466),
    [anon_sym___alignof__] = ACTIONS(2466),
    [anon_sym___alignof] = ACTIONS(2466),
    [anon_sym__alignof] = ACTIONS(2466),
    [anon_sym_alignof] = ACTIONS(2466),
    [anon_sym__Alignof] = ACTIONS(2466),
    [anon_sym_offsetof] = ACTIONS(2466),
    [anon_sym__Generic] = ACTIONS(2466),
    [anon_sym_asm] = ACTIONS(2466),
    [anon_sym___asm__] = ACTIONS(2466),
    [anon_sym___asm] = ACTIONS(2466),
    [sym_number_literal] = ACTIONS(2468),
    [anon_sym_L_SQUOTE] = ACTIONS(2468),
    [anon_sym_u_SQUOTE] = ACTIONS(2468),
    [anon_sym_U_SQUOTE] = ACTIONS(2468),
    [anon_sym_u8_SQUOTE] = ACTIONS(2468),
    [anon_sym_SQUOTE] = ACTIONS(2468),
    [anon_sym_AT] = ACTIONS(2466),
    [anon_sym_DQUOTE] = ACTIONS(2468),
    [anon_sym_L_DQUOTE] = ACTIONS(2468),
    [anon_sym_u_DQUOTE] = ACTIONS(2468),
    [anon_sym_U_DQUOTE] = ACTIONS(2468),
    [anon_sym_u8_DQUOTE] = ACTIONS(2468),
    [sym_true] = ACTIONS(2466),
    [sym_false] = ACTIONS(2466),
    [anon_sym_NULL] = ACTIONS(2466),
    [anon_sym_nullptr] = ACTIONS(2466),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2466),
    [anon_sym___typeof] = ACTIONS(2466),
    [anon_sym_typeof] = ACTIONS(2466),
    [anon_sym_ATimport] = ACTIONS(2468),
    [aux_sym_preproc_undef_token1] = ACTIONS(2466),
    [anon_sym_POUND] = ACTIONS(2466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_API_AVAILABLE] = ACTIONS(2466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_API_DEPRECATED] = ACTIONS(2466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym___deprecated_msg] = ACTIONS(2466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2466),
    [anon_sym_ATprotocol] = ACTIONS(2468),
    [anon_sym_ATinterface] = ACTIONS(2468),
    [anon_sym_ATimplementation] = ACTIONS(2468),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2468),
    [anon_sym_ATtry] = ACTIONS(2468),
    [anon_sym___try] = ACTIONS(2466),
    [anon_sym_ATthrow] = ACTIONS(2468),
    [anon_sym_ATselector] = ACTIONS(2468),
    [anon_sym_ATavailable] = ACTIONS(2468),
    [anon_sym___builtin_available] = ACTIONS(2466),
    [anon_sym_va_arg] = ACTIONS(2466),
    [anon_sym_ATencode] = ACTIONS(2468),
    [anon_sym_ATsynchronized] = ACTIONS(2468),
    [anon_sym_BOOL] = ACTIONS(2466),
    [anon_sym_IMP] = ACTIONS(2466),
    [anon_sym_SEL] = ACTIONS(2466),
    [anon_sym_Class] = ACTIONS(2466),
    [anon_sym_id] = ACTIONS(2466),
  },
  [STATE(716)] = {
    [ts_builtin_sym_end] = ACTIONS(2472),
    [sym_identifier] = ACTIONS(2470),
    [aux_sym_preproc_include_token1] = ACTIONS(2470),
    [aux_sym_preproc_include_token2] = ACTIONS(2470),
    [aux_sym_preproc_def_token1] = ACTIONS(2470),
    [aux_sym_preproc_if_token1] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2470),
    [sym_preproc_directive] = ACTIONS(2470),
    [anon_sym_LPAREN2] = ACTIONS(2472),
    [anon_sym_BANG] = ACTIONS(2472),
    [anon_sym_TILDE] = ACTIONS(2472),
    [anon_sym_DASH] = ACTIONS(2470),
    [anon_sym_PLUS] = ACTIONS(2470),
    [anon_sym_STAR] = ACTIONS(2472),
    [anon_sym_CARET] = ACTIONS(2472),
    [anon_sym_AMP] = ACTIONS(2472),
    [anon_sym_SEMI] = ACTIONS(2472),
    [anon_sym___extension__] = ACTIONS(2470),
    [anon_sym_typedef] = ACTIONS(2470),
    [anon_sym_extern] = ACTIONS(2470),
    [anon_sym___attribute__] = ACTIONS(2470),
    [anon_sym___attribute] = ACTIONS(2470),
    [anon_sym_noreturn] = ACTIONS(2470),
    [anon_sym_LBRACK] = ACTIONS(2472),
    [anon_sym___declspec] = ACTIONS(2470),
    [anon_sym___cdecl] = ACTIONS(2470),
    [anon_sym___clrcall] = ACTIONS(2470),
    [anon_sym___stdcall] = ACTIONS(2470),
    [anon_sym___fastcall] = ACTIONS(2470),
    [anon_sym___thiscall] = ACTIONS(2470),
    [anon_sym___vectorcall] = ACTIONS(2470),
    [anon_sym_LBRACE] = ACTIONS(2472),
    [anon_sym_signed] = ACTIONS(2470),
    [anon_sym_unsigned] = ACTIONS(2470),
    [anon_sym_long] = ACTIONS(2470),
    [anon_sym_short] = ACTIONS(2470),
    [anon_sym_static] = ACTIONS(2470),
    [anon_sym_ATautoreleasepool] = ACTIONS(2472),
    [anon_sym_auto] = ACTIONS(2470),
    [anon_sym_register] = ACTIONS(2470),
    [anon_sym_inline] = ACTIONS(2470),
    [anon_sym___inline] = ACTIONS(2470),
    [anon_sym___inline__] = ACTIONS(2470),
    [anon_sym___forceinline] = ACTIONS(2470),
    [anon_sym_thread_local] = ACTIONS(2470),
    [anon_sym___thread] = ACTIONS(2470),
    [anon_sym_CG_EXTERN] = ACTIONS(2470),
    [anon_sym_CG_INLINE] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2470),
    [anon_sym_IBOutlet] = ACTIONS(2470),
    [anon_sym_IBInspectable] = ACTIONS(2470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2470),
    [anon_sym_NS_INLINE] = ACTIONS(2470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2470),
    [anon_sym_const] = ACTIONS(2470),
    [anon_sym_constexpr] = ACTIONS(2470),
    [anon_sym_volatile] = ACTIONS(2470),
    [anon_sym_restrict] = ACTIONS(2470),
    [anon_sym___restrict__] = ACTIONS(2470),
    [anon_sym__Atomic] = ACTIONS(2470),
    [anon_sym__Noreturn] = ACTIONS(2470),
    [anon_sym__Nonnull] = ACTIONS(2470),
    [anon_sym_nullable] = ACTIONS(2470),
    [anon_sym__Complex] = ACTIONS(2470),
    [anon_sym__Nullable] = ACTIONS(2470),
    [anon_sym__Nullable_result] = ACTIONS(2470),
    [anon_sym__Null_unspecified] = ACTIONS(2470),
    [anon_sym___autoreleasing] = ACTIONS(2470),
    [anon_sym___block] = ACTIONS(2470),
    [anon_sym___bridge] = ACTIONS(2470),
    [anon_sym___bridge_retained] = ACTIONS(2470),
    [anon_sym___bridge_transfer] = ACTIONS(2470),
    [anon_sym___complex] = ACTIONS(2470),
    [anon_sym___const] = ACTIONS(2470),
    [anon_sym___imag] = ACTIONS(2470),
    [anon_sym___kindof] = ACTIONS(2470),
    [anon_sym___nonnull] = ACTIONS(2470),
    [anon_sym___nullable] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2470),
    [anon_sym___real] = ACTIONS(2470),
    [anon_sym___strong] = ACTIONS(2470),
    [anon_sym___unsafe_unretained] = ACTIONS(2470),
    [anon_sym___unused] = ACTIONS(2470),
    [anon_sym___weak] = ACTIONS(2470),
    [anon_sym_alignas] = ACTIONS(2470),
    [anon_sym__Alignas] = ACTIONS(2470),
    [sym_primitive_type] = ACTIONS(2470),
    [anon_sym_enum] = ACTIONS(2470),
    [anon_sym_struct] = ACTIONS(2470),
    [anon_sym_union] = ACTIONS(2470),
    [anon_sym_if] = ACTIONS(2470),
    [anon_sym_else] = ACTIONS(2470),
    [anon_sym_switch] = ACTIONS(2470),
    [anon_sym_case] = ACTIONS(2470),
    [anon_sym_default] = ACTIONS(2470),
    [anon_sym_while] = ACTIONS(2470),
    [anon_sym_do] = ACTIONS(2470),
    [anon_sym_for] = ACTIONS(2470),
    [anon_sym_in] = ACTIONS(2470),
    [anon_sym_return] = ACTIONS(2470),
    [anon_sym_break] = ACTIONS(2470),
    [anon_sym_continue] = ACTIONS(2470),
    [anon_sym_goto] = ACTIONS(2470),
    [anon_sym_DASH_DASH] = ACTIONS(2472),
    [anon_sym_PLUS_PLUS] = ACTIONS(2472),
    [anon_sym_sizeof] = ACTIONS(2470),
    [anon_sym___alignof__] = ACTIONS(2470),
    [anon_sym___alignof] = ACTIONS(2470),
    [anon_sym__alignof] = ACTIONS(2470),
    [anon_sym_alignof] = ACTIONS(2470),
    [anon_sym__Alignof] = ACTIONS(2470),
    [anon_sym_offsetof] = ACTIONS(2470),
    [anon_sym__Generic] = ACTIONS(2470),
    [anon_sym_asm] = ACTIONS(2470),
    [anon_sym___asm__] = ACTIONS(2470),
    [anon_sym___asm] = ACTIONS(2470),
    [sym_number_literal] = ACTIONS(2472),
    [anon_sym_L_SQUOTE] = ACTIONS(2472),
    [anon_sym_u_SQUOTE] = ACTIONS(2472),
    [anon_sym_U_SQUOTE] = ACTIONS(2472),
    [anon_sym_u8_SQUOTE] = ACTIONS(2472),
    [anon_sym_SQUOTE] = ACTIONS(2472),
    [anon_sym_AT] = ACTIONS(2470),
    [anon_sym_DQUOTE] = ACTIONS(2472),
    [anon_sym_L_DQUOTE] = ACTIONS(2472),
    [anon_sym_u_DQUOTE] = ACTIONS(2472),
    [anon_sym_U_DQUOTE] = ACTIONS(2472),
    [anon_sym_u8_DQUOTE] = ACTIONS(2472),
    [sym_true] = ACTIONS(2470),
    [sym_false] = ACTIONS(2470),
    [anon_sym_NULL] = ACTIONS(2470),
    [anon_sym_nullptr] = ACTIONS(2470),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2470),
    [anon_sym___typeof] = ACTIONS(2470),
    [anon_sym_typeof] = ACTIONS(2470),
    [anon_sym_ATimport] = ACTIONS(2472),
    [aux_sym_preproc_undef_token1] = ACTIONS(2470),
    [anon_sym_POUND] = ACTIONS(2470),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_API_AVAILABLE] = ACTIONS(2470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_API_DEPRECATED] = ACTIONS(2470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym___deprecated_msg] = ACTIONS(2470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2470),
    [anon_sym_ATprotocol] = ACTIONS(2472),
    [anon_sym_ATinterface] = ACTIONS(2472),
    [anon_sym_ATimplementation] = ACTIONS(2472),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2472),
    [anon_sym_ATtry] = ACTIONS(2472),
    [anon_sym___try] = ACTIONS(2470),
    [anon_sym_ATthrow] = ACTIONS(2472),
    [anon_sym_ATselector] = ACTIONS(2472),
    [anon_sym_ATavailable] = ACTIONS(2472),
    [anon_sym___builtin_available] = ACTIONS(2470),
    [anon_sym_va_arg] = ACTIONS(2470),
    [anon_sym_ATencode] = ACTIONS(2472),
    [anon_sym_ATsynchronized] = ACTIONS(2472),
    [anon_sym_BOOL] = ACTIONS(2470),
    [anon_sym_IMP] = ACTIONS(2470),
    [anon_sym_SEL] = ACTIONS(2470),
    [anon_sym_Class] = ACTIONS(2470),
    [anon_sym_id] = ACTIONS(2470),
  },
  [STATE(717)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_RBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(718)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_RBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(719)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_RBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(720)] = {
    [ts_builtin_sym_end] = ACTIONS(2570),
    [sym_identifier] = ACTIONS(2568),
    [aux_sym_preproc_include_token1] = ACTIONS(2568),
    [aux_sym_preproc_include_token2] = ACTIONS(2568),
    [aux_sym_preproc_def_token1] = ACTIONS(2568),
    [aux_sym_preproc_if_token1] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2568),
    [sym_preproc_directive] = ACTIONS(2568),
    [anon_sym_LPAREN2] = ACTIONS(2570),
    [anon_sym_BANG] = ACTIONS(2570),
    [anon_sym_TILDE] = ACTIONS(2570),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_PLUS] = ACTIONS(2568),
    [anon_sym_STAR] = ACTIONS(2570),
    [anon_sym_CARET] = ACTIONS(2570),
    [anon_sym_AMP] = ACTIONS(2570),
    [anon_sym_SEMI] = ACTIONS(2570),
    [anon_sym___extension__] = ACTIONS(2568),
    [anon_sym_typedef] = ACTIONS(2568),
    [anon_sym_extern] = ACTIONS(2568),
    [anon_sym___attribute__] = ACTIONS(2568),
    [anon_sym___attribute] = ACTIONS(2568),
    [anon_sym_noreturn] = ACTIONS(2568),
    [anon_sym_LBRACK] = ACTIONS(2570),
    [anon_sym___declspec] = ACTIONS(2568),
    [anon_sym___cdecl] = ACTIONS(2568),
    [anon_sym___clrcall] = ACTIONS(2568),
    [anon_sym___stdcall] = ACTIONS(2568),
    [anon_sym___fastcall] = ACTIONS(2568),
    [anon_sym___thiscall] = ACTIONS(2568),
    [anon_sym___vectorcall] = ACTIONS(2568),
    [anon_sym_LBRACE] = ACTIONS(2570),
    [anon_sym_signed] = ACTIONS(2568),
    [anon_sym_unsigned] = ACTIONS(2568),
    [anon_sym_long] = ACTIONS(2568),
    [anon_sym_short] = ACTIONS(2568),
    [anon_sym_static] = ACTIONS(2568),
    [anon_sym_ATautoreleasepool] = ACTIONS(2570),
    [anon_sym_auto] = ACTIONS(2568),
    [anon_sym_register] = ACTIONS(2568),
    [anon_sym_inline] = ACTIONS(2568),
    [anon_sym___inline] = ACTIONS(2568),
    [anon_sym___inline__] = ACTIONS(2568),
    [anon_sym___forceinline] = ACTIONS(2568),
    [anon_sym_thread_local] = ACTIONS(2568),
    [anon_sym___thread] = ACTIONS(2568),
    [anon_sym_CG_EXTERN] = ACTIONS(2568),
    [anon_sym_CG_INLINE] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2568),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2568),
    [anon_sym_IBOutlet] = ACTIONS(2568),
    [anon_sym_IBInspectable] = ACTIONS(2568),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2568),
    [anon_sym_NS_INLINE] = ACTIONS(2568),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2568),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2568),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2568),
    [anon_sym_const] = ACTIONS(2568),
    [anon_sym_constexpr] = ACTIONS(2568),
    [anon_sym_volatile] = ACTIONS(2568),
    [anon_sym_restrict] = ACTIONS(2568),
    [anon_sym___restrict__] = ACTIONS(2568),
    [anon_sym__Atomic] = ACTIONS(2568),
    [anon_sym__Noreturn] = ACTIONS(2568),
    [anon_sym__Nonnull] = ACTIONS(2568),
    [anon_sym_nullable] = ACTIONS(2568),
    [anon_sym__Complex] = ACTIONS(2568),
    [anon_sym__Nullable] = ACTIONS(2568),
    [anon_sym__Nullable_result] = ACTIONS(2568),
    [anon_sym__Null_unspecified] = ACTIONS(2568),
    [anon_sym___autoreleasing] = ACTIONS(2568),
    [anon_sym___block] = ACTIONS(2568),
    [anon_sym___bridge] = ACTIONS(2568),
    [anon_sym___bridge_retained] = ACTIONS(2568),
    [anon_sym___bridge_transfer] = ACTIONS(2568),
    [anon_sym___complex] = ACTIONS(2568),
    [anon_sym___const] = ACTIONS(2568),
    [anon_sym___imag] = ACTIONS(2568),
    [anon_sym___kindof] = ACTIONS(2568),
    [anon_sym___nonnull] = ACTIONS(2568),
    [anon_sym___nullable] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2568),
    [anon_sym___real] = ACTIONS(2568),
    [anon_sym___strong] = ACTIONS(2568),
    [anon_sym___unsafe_unretained] = ACTIONS(2568),
    [anon_sym___unused] = ACTIONS(2568),
    [anon_sym___weak] = ACTIONS(2568),
    [anon_sym_alignas] = ACTIONS(2568),
    [anon_sym__Alignas] = ACTIONS(2568),
    [sym_primitive_type] = ACTIONS(2568),
    [anon_sym_enum] = ACTIONS(2568),
    [anon_sym_struct] = ACTIONS(2568),
    [anon_sym_union] = ACTIONS(2568),
    [anon_sym_if] = ACTIONS(2568),
    [anon_sym_else] = ACTIONS(2568),
    [anon_sym_switch] = ACTIONS(2568),
    [anon_sym_case] = ACTIONS(2568),
    [anon_sym_default] = ACTIONS(2568),
    [anon_sym_while] = ACTIONS(2568),
    [anon_sym_do] = ACTIONS(2568),
    [anon_sym_for] = ACTIONS(2568),
    [anon_sym_in] = ACTIONS(2568),
    [anon_sym_return] = ACTIONS(2568),
    [anon_sym_break] = ACTIONS(2568),
    [anon_sym_continue] = ACTIONS(2568),
    [anon_sym_goto] = ACTIONS(2568),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_sizeof] = ACTIONS(2568),
    [anon_sym___alignof__] = ACTIONS(2568),
    [anon_sym___alignof] = ACTIONS(2568),
    [anon_sym__alignof] = ACTIONS(2568),
    [anon_sym_alignof] = ACTIONS(2568),
    [anon_sym__Alignof] = ACTIONS(2568),
    [anon_sym_offsetof] = ACTIONS(2568),
    [anon_sym__Generic] = ACTIONS(2568),
    [anon_sym_asm] = ACTIONS(2568),
    [anon_sym___asm__] = ACTIONS(2568),
    [anon_sym___asm] = ACTIONS(2568),
    [sym_number_literal] = ACTIONS(2570),
    [anon_sym_L_SQUOTE] = ACTIONS(2570),
    [anon_sym_u_SQUOTE] = ACTIONS(2570),
    [anon_sym_U_SQUOTE] = ACTIONS(2570),
    [anon_sym_u8_SQUOTE] = ACTIONS(2570),
    [anon_sym_SQUOTE] = ACTIONS(2570),
    [anon_sym_AT] = ACTIONS(2568),
    [anon_sym_DQUOTE] = ACTIONS(2570),
    [anon_sym_L_DQUOTE] = ACTIONS(2570),
    [anon_sym_u_DQUOTE] = ACTIONS(2570),
    [anon_sym_U_DQUOTE] = ACTIONS(2570),
    [anon_sym_u8_DQUOTE] = ACTIONS(2570),
    [sym_true] = ACTIONS(2568),
    [sym_false] = ACTIONS(2568),
    [anon_sym_NULL] = ACTIONS(2568),
    [anon_sym_nullptr] = ACTIONS(2568),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2568),
    [anon_sym___typeof] = ACTIONS(2568),
    [anon_sym_typeof] = ACTIONS(2568),
    [anon_sym_ATimport] = ACTIONS(2570),
    [aux_sym_preproc_undef_token1] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2568),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2568),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2568),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2568),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_API_AVAILABLE] = ACTIONS(2568),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_API_DEPRECATED] = ACTIONS(2568),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym___deprecated_msg] = ACTIONS(2568),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2568),
    [anon_sym_ATprotocol] = ACTIONS(2570),
    [anon_sym_ATinterface] = ACTIONS(2570),
    [anon_sym_ATimplementation] = ACTIONS(2570),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2570),
    [anon_sym_ATtry] = ACTIONS(2570),
    [anon_sym___try] = ACTIONS(2568),
    [anon_sym_ATthrow] = ACTIONS(2570),
    [anon_sym_ATselector] = ACTIONS(2570),
    [anon_sym_ATavailable] = ACTIONS(2570),
    [anon_sym___builtin_available] = ACTIONS(2568),
    [anon_sym_va_arg] = ACTIONS(2568),
    [anon_sym_ATencode] = ACTIONS(2570),
    [anon_sym_ATsynchronized] = ACTIONS(2570),
    [anon_sym_BOOL] = ACTIONS(2568),
    [anon_sym_IMP] = ACTIONS(2568),
    [anon_sym_SEL] = ACTIONS(2568),
    [anon_sym_Class] = ACTIONS(2568),
    [anon_sym_id] = ACTIONS(2568),
  },
  [STATE(721)] = {
    [ts_builtin_sym_end] = ACTIONS(2480),
    [sym_identifier] = ACTIONS(2478),
    [aux_sym_preproc_include_token1] = ACTIONS(2478),
    [aux_sym_preproc_include_token2] = ACTIONS(2478),
    [aux_sym_preproc_def_token1] = ACTIONS(2478),
    [aux_sym_preproc_if_token1] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2478),
    [sym_preproc_directive] = ACTIONS(2478),
    [anon_sym_LPAREN2] = ACTIONS(2480),
    [anon_sym_BANG] = ACTIONS(2480),
    [anon_sym_TILDE] = ACTIONS(2480),
    [anon_sym_DASH] = ACTIONS(2478),
    [anon_sym_PLUS] = ACTIONS(2478),
    [anon_sym_STAR] = ACTIONS(2480),
    [anon_sym_CARET] = ACTIONS(2480),
    [anon_sym_AMP] = ACTIONS(2480),
    [anon_sym_SEMI] = ACTIONS(2480),
    [anon_sym___extension__] = ACTIONS(2478),
    [anon_sym_typedef] = ACTIONS(2478),
    [anon_sym_extern] = ACTIONS(2478),
    [anon_sym___attribute__] = ACTIONS(2478),
    [anon_sym___attribute] = ACTIONS(2478),
    [anon_sym_noreturn] = ACTIONS(2478),
    [anon_sym_LBRACK] = ACTIONS(2480),
    [anon_sym___declspec] = ACTIONS(2478),
    [anon_sym___cdecl] = ACTIONS(2478),
    [anon_sym___clrcall] = ACTIONS(2478),
    [anon_sym___stdcall] = ACTIONS(2478),
    [anon_sym___fastcall] = ACTIONS(2478),
    [anon_sym___thiscall] = ACTIONS(2478),
    [anon_sym___vectorcall] = ACTIONS(2478),
    [anon_sym_LBRACE] = ACTIONS(2480),
    [anon_sym_signed] = ACTIONS(2478),
    [anon_sym_unsigned] = ACTIONS(2478),
    [anon_sym_long] = ACTIONS(2478),
    [anon_sym_short] = ACTIONS(2478),
    [anon_sym_static] = ACTIONS(2478),
    [anon_sym_ATautoreleasepool] = ACTIONS(2480),
    [anon_sym_auto] = ACTIONS(2478),
    [anon_sym_register] = ACTIONS(2478),
    [anon_sym_inline] = ACTIONS(2478),
    [anon_sym___inline] = ACTIONS(2478),
    [anon_sym___inline__] = ACTIONS(2478),
    [anon_sym___forceinline] = ACTIONS(2478),
    [anon_sym_thread_local] = ACTIONS(2478),
    [anon_sym___thread] = ACTIONS(2478),
    [anon_sym_CG_EXTERN] = ACTIONS(2478),
    [anon_sym_CG_INLINE] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2478),
    [anon_sym_IBOutlet] = ACTIONS(2478),
    [anon_sym_IBInspectable] = ACTIONS(2478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2478),
    [anon_sym_NS_INLINE] = ACTIONS(2478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2478),
    [anon_sym_const] = ACTIONS(2478),
    [anon_sym_constexpr] = ACTIONS(2478),
    [anon_sym_volatile] = ACTIONS(2478),
    [anon_sym_restrict] = ACTIONS(2478),
    [anon_sym___restrict__] = ACTIONS(2478),
    [anon_sym__Atomic] = ACTIONS(2478),
    [anon_sym__Noreturn] = ACTIONS(2478),
    [anon_sym__Nonnull] = ACTIONS(2478),
    [anon_sym_nullable] = ACTIONS(2478),
    [anon_sym__Complex] = ACTIONS(2478),
    [anon_sym__Nullable] = ACTIONS(2478),
    [anon_sym__Nullable_result] = ACTIONS(2478),
    [anon_sym__Null_unspecified] = ACTIONS(2478),
    [anon_sym___autoreleasing] = ACTIONS(2478),
    [anon_sym___block] = ACTIONS(2478),
    [anon_sym___bridge] = ACTIONS(2478),
    [anon_sym___bridge_retained] = ACTIONS(2478),
    [anon_sym___bridge_transfer] = ACTIONS(2478),
    [anon_sym___complex] = ACTIONS(2478),
    [anon_sym___const] = ACTIONS(2478),
    [anon_sym___imag] = ACTIONS(2478),
    [anon_sym___kindof] = ACTIONS(2478),
    [anon_sym___nonnull] = ACTIONS(2478),
    [anon_sym___nullable] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2478),
    [anon_sym___real] = ACTIONS(2478),
    [anon_sym___strong] = ACTIONS(2478),
    [anon_sym___unsafe_unretained] = ACTIONS(2478),
    [anon_sym___unused] = ACTIONS(2478),
    [anon_sym___weak] = ACTIONS(2478),
    [anon_sym_alignas] = ACTIONS(2478),
    [anon_sym__Alignas] = ACTIONS(2478),
    [sym_primitive_type] = ACTIONS(2478),
    [anon_sym_enum] = ACTIONS(2478),
    [anon_sym_struct] = ACTIONS(2478),
    [anon_sym_union] = ACTIONS(2478),
    [anon_sym_if] = ACTIONS(2478),
    [anon_sym_else] = ACTIONS(2478),
    [anon_sym_switch] = ACTIONS(2478),
    [anon_sym_case] = ACTIONS(2478),
    [anon_sym_default] = ACTIONS(2478),
    [anon_sym_while] = ACTIONS(2478),
    [anon_sym_do] = ACTIONS(2478),
    [anon_sym_for] = ACTIONS(2478),
    [anon_sym_in] = ACTIONS(2478),
    [anon_sym_return] = ACTIONS(2478),
    [anon_sym_break] = ACTIONS(2478),
    [anon_sym_continue] = ACTIONS(2478),
    [anon_sym_goto] = ACTIONS(2478),
    [anon_sym_DASH_DASH] = ACTIONS(2480),
    [anon_sym_PLUS_PLUS] = ACTIONS(2480),
    [anon_sym_sizeof] = ACTIONS(2478),
    [anon_sym___alignof__] = ACTIONS(2478),
    [anon_sym___alignof] = ACTIONS(2478),
    [anon_sym__alignof] = ACTIONS(2478),
    [anon_sym_alignof] = ACTIONS(2478),
    [anon_sym__Alignof] = ACTIONS(2478),
    [anon_sym_offsetof] = ACTIONS(2478),
    [anon_sym__Generic] = ACTIONS(2478),
    [anon_sym_asm] = ACTIONS(2478),
    [anon_sym___asm__] = ACTIONS(2478),
    [anon_sym___asm] = ACTIONS(2478),
    [sym_number_literal] = ACTIONS(2480),
    [anon_sym_L_SQUOTE] = ACTIONS(2480),
    [anon_sym_u_SQUOTE] = ACTIONS(2480),
    [anon_sym_U_SQUOTE] = ACTIONS(2480),
    [anon_sym_u8_SQUOTE] = ACTIONS(2480),
    [anon_sym_SQUOTE] = ACTIONS(2480),
    [anon_sym_AT] = ACTIONS(2478),
    [anon_sym_DQUOTE] = ACTIONS(2480),
    [anon_sym_L_DQUOTE] = ACTIONS(2480),
    [anon_sym_u_DQUOTE] = ACTIONS(2480),
    [anon_sym_U_DQUOTE] = ACTIONS(2480),
    [anon_sym_u8_DQUOTE] = ACTIONS(2480),
    [sym_true] = ACTIONS(2478),
    [sym_false] = ACTIONS(2478),
    [anon_sym_NULL] = ACTIONS(2478),
    [anon_sym_nullptr] = ACTIONS(2478),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2478),
    [anon_sym___typeof] = ACTIONS(2478),
    [anon_sym_typeof] = ACTIONS(2478),
    [anon_sym_ATimport] = ACTIONS(2480),
    [aux_sym_preproc_undef_token1] = ACTIONS(2478),
    [anon_sym_POUND] = ACTIONS(2478),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_API_AVAILABLE] = ACTIONS(2478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_API_DEPRECATED] = ACTIONS(2478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym___deprecated_msg] = ACTIONS(2478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2478),
    [anon_sym_ATprotocol] = ACTIONS(2480),
    [anon_sym_ATinterface] = ACTIONS(2480),
    [anon_sym_ATimplementation] = ACTIONS(2480),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2480),
    [anon_sym_ATtry] = ACTIONS(2480),
    [anon_sym___try] = ACTIONS(2478),
    [anon_sym_ATthrow] = ACTIONS(2480),
    [anon_sym_ATselector] = ACTIONS(2480),
    [anon_sym_ATavailable] = ACTIONS(2480),
    [anon_sym___builtin_available] = ACTIONS(2478),
    [anon_sym_va_arg] = ACTIONS(2478),
    [anon_sym_ATencode] = ACTIONS(2480),
    [anon_sym_ATsynchronized] = ACTIONS(2480),
    [anon_sym_BOOL] = ACTIONS(2478),
    [anon_sym_IMP] = ACTIONS(2478),
    [anon_sym_SEL] = ACTIONS(2478),
    [anon_sym_Class] = ACTIONS(2478),
    [anon_sym_id] = ACTIONS(2478),
  },
  [STATE(722)] = {
    [ts_builtin_sym_end] = ACTIONS(2484),
    [sym_identifier] = ACTIONS(2482),
    [aux_sym_preproc_include_token1] = ACTIONS(2482),
    [aux_sym_preproc_include_token2] = ACTIONS(2482),
    [aux_sym_preproc_def_token1] = ACTIONS(2482),
    [aux_sym_preproc_if_token1] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2482),
    [sym_preproc_directive] = ACTIONS(2482),
    [anon_sym_LPAREN2] = ACTIONS(2484),
    [anon_sym_BANG] = ACTIONS(2484),
    [anon_sym_TILDE] = ACTIONS(2484),
    [anon_sym_DASH] = ACTIONS(2482),
    [anon_sym_PLUS] = ACTIONS(2482),
    [anon_sym_STAR] = ACTIONS(2484),
    [anon_sym_CARET] = ACTIONS(2484),
    [anon_sym_AMP] = ACTIONS(2484),
    [anon_sym_SEMI] = ACTIONS(2484),
    [anon_sym___extension__] = ACTIONS(2482),
    [anon_sym_typedef] = ACTIONS(2482),
    [anon_sym_extern] = ACTIONS(2482),
    [anon_sym___attribute__] = ACTIONS(2482),
    [anon_sym___attribute] = ACTIONS(2482),
    [anon_sym_noreturn] = ACTIONS(2482),
    [anon_sym_LBRACK] = ACTIONS(2484),
    [anon_sym___declspec] = ACTIONS(2482),
    [anon_sym___cdecl] = ACTIONS(2482),
    [anon_sym___clrcall] = ACTIONS(2482),
    [anon_sym___stdcall] = ACTIONS(2482),
    [anon_sym___fastcall] = ACTIONS(2482),
    [anon_sym___thiscall] = ACTIONS(2482),
    [anon_sym___vectorcall] = ACTIONS(2482),
    [anon_sym_LBRACE] = ACTIONS(2484),
    [anon_sym_signed] = ACTIONS(2482),
    [anon_sym_unsigned] = ACTIONS(2482),
    [anon_sym_long] = ACTIONS(2482),
    [anon_sym_short] = ACTIONS(2482),
    [anon_sym_static] = ACTIONS(2482),
    [anon_sym_ATautoreleasepool] = ACTIONS(2484),
    [anon_sym_auto] = ACTIONS(2482),
    [anon_sym_register] = ACTIONS(2482),
    [anon_sym_inline] = ACTIONS(2482),
    [anon_sym___inline] = ACTIONS(2482),
    [anon_sym___inline__] = ACTIONS(2482),
    [anon_sym___forceinline] = ACTIONS(2482),
    [anon_sym_thread_local] = ACTIONS(2482),
    [anon_sym___thread] = ACTIONS(2482),
    [anon_sym_CG_EXTERN] = ACTIONS(2482),
    [anon_sym_CG_INLINE] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2482),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2482),
    [anon_sym_IBOutlet] = ACTIONS(2482),
    [anon_sym_IBInspectable] = ACTIONS(2482),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2482),
    [anon_sym_NS_INLINE] = ACTIONS(2482),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2482),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2482),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2482),
    [anon_sym_const] = ACTIONS(2482),
    [anon_sym_constexpr] = ACTIONS(2482),
    [anon_sym_volatile] = ACTIONS(2482),
    [anon_sym_restrict] = ACTIONS(2482),
    [anon_sym___restrict__] = ACTIONS(2482),
    [anon_sym__Atomic] = ACTIONS(2482),
    [anon_sym__Noreturn] = ACTIONS(2482),
    [anon_sym__Nonnull] = ACTIONS(2482),
    [anon_sym_nullable] = ACTIONS(2482),
    [anon_sym__Complex] = ACTIONS(2482),
    [anon_sym__Nullable] = ACTIONS(2482),
    [anon_sym__Nullable_result] = ACTIONS(2482),
    [anon_sym__Null_unspecified] = ACTIONS(2482),
    [anon_sym___autoreleasing] = ACTIONS(2482),
    [anon_sym___block] = ACTIONS(2482),
    [anon_sym___bridge] = ACTIONS(2482),
    [anon_sym___bridge_retained] = ACTIONS(2482),
    [anon_sym___bridge_transfer] = ACTIONS(2482),
    [anon_sym___complex] = ACTIONS(2482),
    [anon_sym___const] = ACTIONS(2482),
    [anon_sym___imag] = ACTIONS(2482),
    [anon_sym___kindof] = ACTIONS(2482),
    [anon_sym___nonnull] = ACTIONS(2482),
    [anon_sym___nullable] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2482),
    [anon_sym___real] = ACTIONS(2482),
    [anon_sym___strong] = ACTIONS(2482),
    [anon_sym___unsafe_unretained] = ACTIONS(2482),
    [anon_sym___unused] = ACTIONS(2482),
    [anon_sym___weak] = ACTIONS(2482),
    [anon_sym_alignas] = ACTIONS(2482),
    [anon_sym__Alignas] = ACTIONS(2482),
    [sym_primitive_type] = ACTIONS(2482),
    [anon_sym_enum] = ACTIONS(2482),
    [anon_sym_struct] = ACTIONS(2482),
    [anon_sym_union] = ACTIONS(2482),
    [anon_sym_if] = ACTIONS(2482),
    [anon_sym_else] = ACTIONS(2482),
    [anon_sym_switch] = ACTIONS(2482),
    [anon_sym_case] = ACTIONS(2482),
    [anon_sym_default] = ACTIONS(2482),
    [anon_sym_while] = ACTIONS(2482),
    [anon_sym_do] = ACTIONS(2482),
    [anon_sym_for] = ACTIONS(2482),
    [anon_sym_in] = ACTIONS(2482),
    [anon_sym_return] = ACTIONS(2482),
    [anon_sym_break] = ACTIONS(2482),
    [anon_sym_continue] = ACTIONS(2482),
    [anon_sym_goto] = ACTIONS(2482),
    [anon_sym_DASH_DASH] = ACTIONS(2484),
    [anon_sym_PLUS_PLUS] = ACTIONS(2484),
    [anon_sym_sizeof] = ACTIONS(2482),
    [anon_sym___alignof__] = ACTIONS(2482),
    [anon_sym___alignof] = ACTIONS(2482),
    [anon_sym__alignof] = ACTIONS(2482),
    [anon_sym_alignof] = ACTIONS(2482),
    [anon_sym__Alignof] = ACTIONS(2482),
    [anon_sym_offsetof] = ACTIONS(2482),
    [anon_sym__Generic] = ACTIONS(2482),
    [anon_sym_asm] = ACTIONS(2482),
    [anon_sym___asm__] = ACTIONS(2482),
    [anon_sym___asm] = ACTIONS(2482),
    [sym_number_literal] = ACTIONS(2484),
    [anon_sym_L_SQUOTE] = ACTIONS(2484),
    [anon_sym_u_SQUOTE] = ACTIONS(2484),
    [anon_sym_U_SQUOTE] = ACTIONS(2484),
    [anon_sym_u8_SQUOTE] = ACTIONS(2484),
    [anon_sym_SQUOTE] = ACTIONS(2484),
    [anon_sym_AT] = ACTIONS(2482),
    [anon_sym_DQUOTE] = ACTIONS(2484),
    [anon_sym_L_DQUOTE] = ACTIONS(2484),
    [anon_sym_u_DQUOTE] = ACTIONS(2484),
    [anon_sym_U_DQUOTE] = ACTIONS(2484),
    [anon_sym_u8_DQUOTE] = ACTIONS(2484),
    [sym_true] = ACTIONS(2482),
    [sym_false] = ACTIONS(2482),
    [anon_sym_NULL] = ACTIONS(2482),
    [anon_sym_nullptr] = ACTIONS(2482),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2482),
    [anon_sym___typeof] = ACTIONS(2482),
    [anon_sym_typeof] = ACTIONS(2482),
    [anon_sym_ATimport] = ACTIONS(2484),
    [aux_sym_preproc_undef_token1] = ACTIONS(2482),
    [anon_sym_POUND] = ACTIONS(2482),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2482),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2482),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2482),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_API_AVAILABLE] = ACTIONS(2482),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_API_DEPRECATED] = ACTIONS(2482),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym___deprecated_msg] = ACTIONS(2482),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2482),
    [anon_sym_ATprotocol] = ACTIONS(2484),
    [anon_sym_ATinterface] = ACTIONS(2484),
    [anon_sym_ATimplementation] = ACTIONS(2484),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2484),
    [anon_sym_ATtry] = ACTIONS(2484),
    [anon_sym___try] = ACTIONS(2482),
    [anon_sym_ATthrow] = ACTIONS(2484),
    [anon_sym_ATselector] = ACTIONS(2484),
    [anon_sym_ATavailable] = ACTIONS(2484),
    [anon_sym___builtin_available] = ACTIONS(2482),
    [anon_sym_va_arg] = ACTIONS(2482),
    [anon_sym_ATencode] = ACTIONS(2484),
    [anon_sym_ATsynchronized] = ACTIONS(2484),
    [anon_sym_BOOL] = ACTIONS(2482),
    [anon_sym_IMP] = ACTIONS(2482),
    [anon_sym_SEL] = ACTIONS(2482),
    [anon_sym_Class] = ACTIONS(2482),
    [anon_sym_id] = ACTIONS(2482),
  },
  [STATE(723)] = {
    [ts_builtin_sym_end] = ACTIONS(2366),
    [sym_identifier] = ACTIONS(2364),
    [aux_sym_preproc_include_token1] = ACTIONS(2364),
    [aux_sym_preproc_include_token2] = ACTIONS(2364),
    [aux_sym_preproc_def_token1] = ACTIONS(2364),
    [aux_sym_preproc_if_token1] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2364),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2364),
    [sym_preproc_directive] = ACTIONS(2364),
    [anon_sym_LPAREN2] = ACTIONS(2366),
    [anon_sym_BANG] = ACTIONS(2366),
    [anon_sym_TILDE] = ACTIONS(2366),
    [anon_sym_DASH] = ACTIONS(2364),
    [anon_sym_PLUS] = ACTIONS(2364),
    [anon_sym_STAR] = ACTIONS(2366),
    [anon_sym_CARET] = ACTIONS(2366),
    [anon_sym_AMP] = ACTIONS(2366),
    [anon_sym_SEMI] = ACTIONS(2366),
    [anon_sym___extension__] = ACTIONS(2364),
    [anon_sym_typedef] = ACTIONS(2364),
    [anon_sym_extern] = ACTIONS(2364),
    [anon_sym___attribute__] = ACTIONS(2364),
    [anon_sym___attribute] = ACTIONS(2364),
    [anon_sym_noreturn] = ACTIONS(2364),
    [anon_sym_LBRACK] = ACTIONS(2366),
    [anon_sym___declspec] = ACTIONS(2364),
    [anon_sym___cdecl] = ACTIONS(2364),
    [anon_sym___clrcall] = ACTIONS(2364),
    [anon_sym___stdcall] = ACTIONS(2364),
    [anon_sym___fastcall] = ACTIONS(2364),
    [anon_sym___thiscall] = ACTIONS(2364),
    [anon_sym___vectorcall] = ACTIONS(2364),
    [anon_sym_LBRACE] = ACTIONS(2366),
    [anon_sym_signed] = ACTIONS(2364),
    [anon_sym_unsigned] = ACTIONS(2364),
    [anon_sym_long] = ACTIONS(2364),
    [anon_sym_short] = ACTIONS(2364),
    [anon_sym_static] = ACTIONS(2364),
    [anon_sym_ATautoreleasepool] = ACTIONS(2366),
    [anon_sym_auto] = ACTIONS(2364),
    [anon_sym_register] = ACTIONS(2364),
    [anon_sym_inline] = ACTIONS(2364),
    [anon_sym___inline] = ACTIONS(2364),
    [anon_sym___inline__] = ACTIONS(2364),
    [anon_sym___forceinline] = ACTIONS(2364),
    [anon_sym_thread_local] = ACTIONS(2364),
    [anon_sym___thread] = ACTIONS(2364),
    [anon_sym_CG_EXTERN] = ACTIONS(2364),
    [anon_sym_CG_INLINE] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2364),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2364),
    [anon_sym_IBOutlet] = ACTIONS(2364),
    [anon_sym_IBInspectable] = ACTIONS(2364),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2364),
    [anon_sym_NS_INLINE] = ACTIONS(2364),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2364),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2364),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2364),
    [anon_sym_const] = ACTIONS(2364),
    [anon_sym_constexpr] = ACTIONS(2364),
    [anon_sym_volatile] = ACTIONS(2364),
    [anon_sym_restrict] = ACTIONS(2364),
    [anon_sym___restrict__] = ACTIONS(2364),
    [anon_sym__Atomic] = ACTIONS(2364),
    [anon_sym__Noreturn] = ACTIONS(2364),
    [anon_sym__Nonnull] = ACTIONS(2364),
    [anon_sym_nullable] = ACTIONS(2364),
    [anon_sym__Complex] = ACTIONS(2364),
    [anon_sym__Nullable] = ACTIONS(2364),
    [anon_sym__Nullable_result] = ACTIONS(2364),
    [anon_sym__Null_unspecified] = ACTIONS(2364),
    [anon_sym___autoreleasing] = ACTIONS(2364),
    [anon_sym___block] = ACTIONS(2364),
    [anon_sym___bridge] = ACTIONS(2364),
    [anon_sym___bridge_retained] = ACTIONS(2364),
    [anon_sym___bridge_transfer] = ACTIONS(2364),
    [anon_sym___complex] = ACTIONS(2364),
    [anon_sym___const] = ACTIONS(2364),
    [anon_sym___imag] = ACTIONS(2364),
    [anon_sym___kindof] = ACTIONS(2364),
    [anon_sym___nonnull] = ACTIONS(2364),
    [anon_sym___nullable] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2364),
    [anon_sym___real] = ACTIONS(2364),
    [anon_sym___strong] = ACTIONS(2364),
    [anon_sym___unsafe_unretained] = ACTIONS(2364),
    [anon_sym___unused] = ACTIONS(2364),
    [anon_sym___weak] = ACTIONS(2364),
    [anon_sym_alignas] = ACTIONS(2364),
    [anon_sym__Alignas] = ACTIONS(2364),
    [sym_primitive_type] = ACTIONS(2364),
    [anon_sym_enum] = ACTIONS(2364),
    [anon_sym_struct] = ACTIONS(2364),
    [anon_sym_union] = ACTIONS(2364),
    [anon_sym_if] = ACTIONS(2364),
    [anon_sym_else] = ACTIONS(2364),
    [anon_sym_switch] = ACTIONS(2364),
    [anon_sym_case] = ACTIONS(2364),
    [anon_sym_default] = ACTIONS(2364),
    [anon_sym_while] = ACTIONS(2364),
    [anon_sym_do] = ACTIONS(2364),
    [anon_sym_for] = ACTIONS(2364),
    [anon_sym_in] = ACTIONS(2364),
    [anon_sym_return] = ACTIONS(2364),
    [anon_sym_break] = ACTIONS(2364),
    [anon_sym_continue] = ACTIONS(2364),
    [anon_sym_goto] = ACTIONS(2364),
    [anon_sym_DASH_DASH] = ACTIONS(2366),
    [anon_sym_PLUS_PLUS] = ACTIONS(2366),
    [anon_sym_sizeof] = ACTIONS(2364),
    [anon_sym___alignof__] = ACTIONS(2364),
    [anon_sym___alignof] = ACTIONS(2364),
    [anon_sym__alignof] = ACTIONS(2364),
    [anon_sym_alignof] = ACTIONS(2364),
    [anon_sym__Alignof] = ACTIONS(2364),
    [anon_sym_offsetof] = ACTIONS(2364),
    [anon_sym__Generic] = ACTIONS(2364),
    [anon_sym_asm] = ACTIONS(2364),
    [anon_sym___asm__] = ACTIONS(2364),
    [anon_sym___asm] = ACTIONS(2364),
    [sym_number_literal] = ACTIONS(2366),
    [anon_sym_L_SQUOTE] = ACTIONS(2366),
    [anon_sym_u_SQUOTE] = ACTIONS(2366),
    [anon_sym_U_SQUOTE] = ACTIONS(2366),
    [anon_sym_u8_SQUOTE] = ACTIONS(2366),
    [anon_sym_SQUOTE] = ACTIONS(2366),
    [anon_sym_AT] = ACTIONS(2364),
    [anon_sym_DQUOTE] = ACTIONS(2366),
    [anon_sym_L_DQUOTE] = ACTIONS(2366),
    [anon_sym_u_DQUOTE] = ACTIONS(2366),
    [anon_sym_U_DQUOTE] = ACTIONS(2366),
    [anon_sym_u8_DQUOTE] = ACTIONS(2366),
    [sym_true] = ACTIONS(2364),
    [sym_false] = ACTIONS(2364),
    [anon_sym_NULL] = ACTIONS(2364),
    [anon_sym_nullptr] = ACTIONS(2364),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2364),
    [anon_sym___typeof] = ACTIONS(2364),
    [anon_sym_typeof] = ACTIONS(2364),
    [anon_sym_ATimport] = ACTIONS(2366),
    [aux_sym_preproc_undef_token1] = ACTIONS(2364),
    [anon_sym_POUND] = ACTIONS(2364),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2364),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2364),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2364),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_API_AVAILABLE] = ACTIONS(2364),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_API_DEPRECATED] = ACTIONS(2364),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym___deprecated_msg] = ACTIONS(2364),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2364),
    [anon_sym_ATprotocol] = ACTIONS(2366),
    [anon_sym_ATinterface] = ACTIONS(2366),
    [anon_sym_ATimplementation] = ACTIONS(2366),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2366),
    [anon_sym_ATcatch] = ACTIONS(2366),
    [anon_sym___catch] = ACTIONS(2364),
    [anon_sym_ATfinally] = ACTIONS(2366),
    [anon_sym___finally] = ACTIONS(2364),
    [anon_sym_ATselector] = ACTIONS(2366),
    [anon_sym_ATavailable] = ACTIONS(2366),
    [anon_sym___builtin_available] = ACTIONS(2364),
    [anon_sym_va_arg] = ACTIONS(2364),
    [anon_sym_ATencode] = ACTIONS(2366),
    [anon_sym_BOOL] = ACTIONS(2364),
    [anon_sym_IMP] = ACTIONS(2364),
    [anon_sym_SEL] = ACTIONS(2364),
    [anon_sym_Class] = ACTIONS(2364),
    [anon_sym_id] = ACTIONS(2364),
  },
  [STATE(724)] = {
    [ts_builtin_sym_end] = ACTIONS(2488),
    [sym_identifier] = ACTIONS(2486),
    [aux_sym_preproc_include_token1] = ACTIONS(2486),
    [aux_sym_preproc_include_token2] = ACTIONS(2486),
    [aux_sym_preproc_def_token1] = ACTIONS(2486),
    [aux_sym_preproc_if_token1] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2486),
    [sym_preproc_directive] = ACTIONS(2486),
    [anon_sym_LPAREN2] = ACTIONS(2488),
    [anon_sym_BANG] = ACTIONS(2488),
    [anon_sym_TILDE] = ACTIONS(2488),
    [anon_sym_DASH] = ACTIONS(2486),
    [anon_sym_PLUS] = ACTIONS(2486),
    [anon_sym_STAR] = ACTIONS(2488),
    [anon_sym_CARET] = ACTIONS(2488),
    [anon_sym_AMP] = ACTIONS(2488),
    [anon_sym_SEMI] = ACTIONS(2488),
    [anon_sym___extension__] = ACTIONS(2486),
    [anon_sym_typedef] = ACTIONS(2486),
    [anon_sym_extern] = ACTIONS(2486),
    [anon_sym___attribute__] = ACTIONS(2486),
    [anon_sym___attribute] = ACTIONS(2486),
    [anon_sym_noreturn] = ACTIONS(2486),
    [anon_sym_LBRACK] = ACTIONS(2488),
    [anon_sym___declspec] = ACTIONS(2486),
    [anon_sym___cdecl] = ACTIONS(2486),
    [anon_sym___clrcall] = ACTIONS(2486),
    [anon_sym___stdcall] = ACTIONS(2486),
    [anon_sym___fastcall] = ACTIONS(2486),
    [anon_sym___thiscall] = ACTIONS(2486),
    [anon_sym___vectorcall] = ACTIONS(2486),
    [anon_sym_LBRACE] = ACTIONS(2488),
    [anon_sym_signed] = ACTIONS(2486),
    [anon_sym_unsigned] = ACTIONS(2486),
    [anon_sym_long] = ACTIONS(2486),
    [anon_sym_short] = ACTIONS(2486),
    [anon_sym_static] = ACTIONS(2486),
    [anon_sym_ATautoreleasepool] = ACTIONS(2488),
    [anon_sym_auto] = ACTIONS(2486),
    [anon_sym_register] = ACTIONS(2486),
    [anon_sym_inline] = ACTIONS(2486),
    [anon_sym___inline] = ACTIONS(2486),
    [anon_sym___inline__] = ACTIONS(2486),
    [anon_sym___forceinline] = ACTIONS(2486),
    [anon_sym_thread_local] = ACTIONS(2486),
    [anon_sym___thread] = ACTIONS(2486),
    [anon_sym_CG_EXTERN] = ACTIONS(2486),
    [anon_sym_CG_INLINE] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2486),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2486),
    [anon_sym_IBOutlet] = ACTIONS(2486),
    [anon_sym_IBInspectable] = ACTIONS(2486),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2486),
    [anon_sym_NS_INLINE] = ACTIONS(2486),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2486),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2486),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2486),
    [anon_sym_const] = ACTIONS(2486),
    [anon_sym_constexpr] = ACTIONS(2486),
    [anon_sym_volatile] = ACTIONS(2486),
    [anon_sym_restrict] = ACTIONS(2486),
    [anon_sym___restrict__] = ACTIONS(2486),
    [anon_sym__Atomic] = ACTIONS(2486),
    [anon_sym__Noreturn] = ACTIONS(2486),
    [anon_sym__Nonnull] = ACTIONS(2486),
    [anon_sym_nullable] = ACTIONS(2486),
    [anon_sym__Complex] = ACTIONS(2486),
    [anon_sym__Nullable] = ACTIONS(2486),
    [anon_sym__Nullable_result] = ACTIONS(2486),
    [anon_sym__Null_unspecified] = ACTIONS(2486),
    [anon_sym___autoreleasing] = ACTIONS(2486),
    [anon_sym___block] = ACTIONS(2486),
    [anon_sym___bridge] = ACTIONS(2486),
    [anon_sym___bridge_retained] = ACTIONS(2486),
    [anon_sym___bridge_transfer] = ACTIONS(2486),
    [anon_sym___complex] = ACTIONS(2486),
    [anon_sym___const] = ACTIONS(2486),
    [anon_sym___imag] = ACTIONS(2486),
    [anon_sym___kindof] = ACTIONS(2486),
    [anon_sym___nonnull] = ACTIONS(2486),
    [anon_sym___nullable] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2486),
    [anon_sym___real] = ACTIONS(2486),
    [anon_sym___strong] = ACTIONS(2486),
    [anon_sym___unsafe_unretained] = ACTIONS(2486),
    [anon_sym___unused] = ACTIONS(2486),
    [anon_sym___weak] = ACTIONS(2486),
    [anon_sym_alignas] = ACTIONS(2486),
    [anon_sym__Alignas] = ACTIONS(2486),
    [sym_primitive_type] = ACTIONS(2486),
    [anon_sym_enum] = ACTIONS(2486),
    [anon_sym_struct] = ACTIONS(2486),
    [anon_sym_union] = ACTIONS(2486),
    [anon_sym_if] = ACTIONS(2486),
    [anon_sym_else] = ACTIONS(2486),
    [anon_sym_switch] = ACTIONS(2486),
    [anon_sym_case] = ACTIONS(2486),
    [anon_sym_default] = ACTIONS(2486),
    [anon_sym_while] = ACTIONS(2486),
    [anon_sym_do] = ACTIONS(2486),
    [anon_sym_for] = ACTIONS(2486),
    [anon_sym_in] = ACTIONS(2486),
    [anon_sym_return] = ACTIONS(2486),
    [anon_sym_break] = ACTIONS(2486),
    [anon_sym_continue] = ACTIONS(2486),
    [anon_sym_goto] = ACTIONS(2486),
    [anon_sym_DASH_DASH] = ACTIONS(2488),
    [anon_sym_PLUS_PLUS] = ACTIONS(2488),
    [anon_sym_sizeof] = ACTIONS(2486),
    [anon_sym___alignof__] = ACTIONS(2486),
    [anon_sym___alignof] = ACTIONS(2486),
    [anon_sym__alignof] = ACTIONS(2486),
    [anon_sym_alignof] = ACTIONS(2486),
    [anon_sym__Alignof] = ACTIONS(2486),
    [anon_sym_offsetof] = ACTIONS(2486),
    [anon_sym__Generic] = ACTIONS(2486),
    [anon_sym_asm] = ACTIONS(2486),
    [anon_sym___asm__] = ACTIONS(2486),
    [anon_sym___asm] = ACTIONS(2486),
    [sym_number_literal] = ACTIONS(2488),
    [anon_sym_L_SQUOTE] = ACTIONS(2488),
    [anon_sym_u_SQUOTE] = ACTIONS(2488),
    [anon_sym_U_SQUOTE] = ACTIONS(2488),
    [anon_sym_u8_SQUOTE] = ACTIONS(2488),
    [anon_sym_SQUOTE] = ACTIONS(2488),
    [anon_sym_AT] = ACTIONS(2486),
    [anon_sym_DQUOTE] = ACTIONS(2488),
    [anon_sym_L_DQUOTE] = ACTIONS(2488),
    [anon_sym_u_DQUOTE] = ACTIONS(2488),
    [anon_sym_U_DQUOTE] = ACTIONS(2488),
    [anon_sym_u8_DQUOTE] = ACTIONS(2488),
    [sym_true] = ACTIONS(2486),
    [sym_false] = ACTIONS(2486),
    [anon_sym_NULL] = ACTIONS(2486),
    [anon_sym_nullptr] = ACTIONS(2486),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2486),
    [anon_sym___typeof] = ACTIONS(2486),
    [anon_sym_typeof] = ACTIONS(2486),
    [anon_sym_ATimport] = ACTIONS(2488),
    [aux_sym_preproc_undef_token1] = ACTIONS(2486),
    [anon_sym_POUND] = ACTIONS(2486),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2486),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2486),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2486),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_API_AVAILABLE] = ACTIONS(2486),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_API_DEPRECATED] = ACTIONS(2486),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym___deprecated_msg] = ACTIONS(2486),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2486),
    [anon_sym_ATprotocol] = ACTIONS(2488),
    [anon_sym_ATinterface] = ACTIONS(2488),
    [anon_sym_ATimplementation] = ACTIONS(2488),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2488),
    [anon_sym_ATtry] = ACTIONS(2488),
    [anon_sym___try] = ACTIONS(2486),
    [anon_sym_ATthrow] = ACTIONS(2488),
    [anon_sym_ATselector] = ACTIONS(2488),
    [anon_sym_ATavailable] = ACTIONS(2488),
    [anon_sym___builtin_available] = ACTIONS(2486),
    [anon_sym_va_arg] = ACTIONS(2486),
    [anon_sym_ATencode] = ACTIONS(2488),
    [anon_sym_ATsynchronized] = ACTIONS(2488),
    [anon_sym_BOOL] = ACTIONS(2486),
    [anon_sym_IMP] = ACTIONS(2486),
    [anon_sym_SEL] = ACTIONS(2486),
    [anon_sym_Class] = ACTIONS(2486),
    [anon_sym_id] = ACTIONS(2486),
  },
  [STATE(725)] = {
    [ts_builtin_sym_end] = ACTIONS(2492),
    [sym_identifier] = ACTIONS(2490),
    [aux_sym_preproc_include_token1] = ACTIONS(2490),
    [aux_sym_preproc_include_token2] = ACTIONS(2490),
    [aux_sym_preproc_def_token1] = ACTIONS(2490),
    [aux_sym_preproc_if_token1] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2490),
    [sym_preproc_directive] = ACTIONS(2490),
    [anon_sym_LPAREN2] = ACTIONS(2492),
    [anon_sym_BANG] = ACTIONS(2492),
    [anon_sym_TILDE] = ACTIONS(2492),
    [anon_sym_DASH] = ACTIONS(2490),
    [anon_sym_PLUS] = ACTIONS(2490),
    [anon_sym_STAR] = ACTIONS(2492),
    [anon_sym_CARET] = ACTIONS(2492),
    [anon_sym_AMP] = ACTIONS(2492),
    [anon_sym_SEMI] = ACTIONS(2492),
    [anon_sym___extension__] = ACTIONS(2490),
    [anon_sym_typedef] = ACTIONS(2490),
    [anon_sym_extern] = ACTIONS(2490),
    [anon_sym___attribute__] = ACTIONS(2490),
    [anon_sym___attribute] = ACTIONS(2490),
    [anon_sym_noreturn] = ACTIONS(2490),
    [anon_sym_LBRACK] = ACTIONS(2492),
    [anon_sym___declspec] = ACTIONS(2490),
    [anon_sym___cdecl] = ACTIONS(2490),
    [anon_sym___clrcall] = ACTIONS(2490),
    [anon_sym___stdcall] = ACTIONS(2490),
    [anon_sym___fastcall] = ACTIONS(2490),
    [anon_sym___thiscall] = ACTIONS(2490),
    [anon_sym___vectorcall] = ACTIONS(2490),
    [anon_sym_LBRACE] = ACTIONS(2492),
    [anon_sym_signed] = ACTIONS(2490),
    [anon_sym_unsigned] = ACTIONS(2490),
    [anon_sym_long] = ACTIONS(2490),
    [anon_sym_short] = ACTIONS(2490),
    [anon_sym_static] = ACTIONS(2490),
    [anon_sym_ATautoreleasepool] = ACTIONS(2492),
    [anon_sym_auto] = ACTIONS(2490),
    [anon_sym_register] = ACTIONS(2490),
    [anon_sym_inline] = ACTIONS(2490),
    [anon_sym___inline] = ACTIONS(2490),
    [anon_sym___inline__] = ACTIONS(2490),
    [anon_sym___forceinline] = ACTIONS(2490),
    [anon_sym_thread_local] = ACTIONS(2490),
    [anon_sym___thread] = ACTIONS(2490),
    [anon_sym_CG_EXTERN] = ACTIONS(2490),
    [anon_sym_CG_INLINE] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2490),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2490),
    [anon_sym_IBOutlet] = ACTIONS(2490),
    [anon_sym_IBInspectable] = ACTIONS(2490),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2490),
    [anon_sym_NS_INLINE] = ACTIONS(2490),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2490),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2490),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2490),
    [anon_sym_const] = ACTIONS(2490),
    [anon_sym_constexpr] = ACTIONS(2490),
    [anon_sym_volatile] = ACTIONS(2490),
    [anon_sym_restrict] = ACTIONS(2490),
    [anon_sym___restrict__] = ACTIONS(2490),
    [anon_sym__Atomic] = ACTIONS(2490),
    [anon_sym__Noreturn] = ACTIONS(2490),
    [anon_sym__Nonnull] = ACTIONS(2490),
    [anon_sym_nullable] = ACTIONS(2490),
    [anon_sym__Complex] = ACTIONS(2490),
    [anon_sym__Nullable] = ACTIONS(2490),
    [anon_sym__Nullable_result] = ACTIONS(2490),
    [anon_sym__Null_unspecified] = ACTIONS(2490),
    [anon_sym___autoreleasing] = ACTIONS(2490),
    [anon_sym___block] = ACTIONS(2490),
    [anon_sym___bridge] = ACTIONS(2490),
    [anon_sym___bridge_retained] = ACTIONS(2490),
    [anon_sym___bridge_transfer] = ACTIONS(2490),
    [anon_sym___complex] = ACTIONS(2490),
    [anon_sym___const] = ACTIONS(2490),
    [anon_sym___imag] = ACTIONS(2490),
    [anon_sym___kindof] = ACTIONS(2490),
    [anon_sym___nonnull] = ACTIONS(2490),
    [anon_sym___nullable] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2490),
    [anon_sym___real] = ACTIONS(2490),
    [anon_sym___strong] = ACTIONS(2490),
    [anon_sym___unsafe_unretained] = ACTIONS(2490),
    [anon_sym___unused] = ACTIONS(2490),
    [anon_sym___weak] = ACTIONS(2490),
    [anon_sym_alignas] = ACTIONS(2490),
    [anon_sym__Alignas] = ACTIONS(2490),
    [sym_primitive_type] = ACTIONS(2490),
    [anon_sym_enum] = ACTIONS(2490),
    [anon_sym_struct] = ACTIONS(2490),
    [anon_sym_union] = ACTIONS(2490),
    [anon_sym_if] = ACTIONS(2490),
    [anon_sym_else] = ACTIONS(2490),
    [anon_sym_switch] = ACTIONS(2490),
    [anon_sym_case] = ACTIONS(2490),
    [anon_sym_default] = ACTIONS(2490),
    [anon_sym_while] = ACTIONS(2490),
    [anon_sym_do] = ACTIONS(2490),
    [anon_sym_for] = ACTIONS(2490),
    [anon_sym_in] = ACTIONS(2490),
    [anon_sym_return] = ACTIONS(2490),
    [anon_sym_break] = ACTIONS(2490),
    [anon_sym_continue] = ACTIONS(2490),
    [anon_sym_goto] = ACTIONS(2490),
    [anon_sym_DASH_DASH] = ACTIONS(2492),
    [anon_sym_PLUS_PLUS] = ACTIONS(2492),
    [anon_sym_sizeof] = ACTIONS(2490),
    [anon_sym___alignof__] = ACTIONS(2490),
    [anon_sym___alignof] = ACTIONS(2490),
    [anon_sym__alignof] = ACTIONS(2490),
    [anon_sym_alignof] = ACTIONS(2490),
    [anon_sym__Alignof] = ACTIONS(2490),
    [anon_sym_offsetof] = ACTIONS(2490),
    [anon_sym__Generic] = ACTIONS(2490),
    [anon_sym_asm] = ACTIONS(2490),
    [anon_sym___asm__] = ACTIONS(2490),
    [anon_sym___asm] = ACTIONS(2490),
    [sym_number_literal] = ACTIONS(2492),
    [anon_sym_L_SQUOTE] = ACTIONS(2492),
    [anon_sym_u_SQUOTE] = ACTIONS(2492),
    [anon_sym_U_SQUOTE] = ACTIONS(2492),
    [anon_sym_u8_SQUOTE] = ACTIONS(2492),
    [anon_sym_SQUOTE] = ACTIONS(2492),
    [anon_sym_AT] = ACTIONS(2490),
    [anon_sym_DQUOTE] = ACTIONS(2492),
    [anon_sym_L_DQUOTE] = ACTIONS(2492),
    [anon_sym_u_DQUOTE] = ACTIONS(2492),
    [anon_sym_U_DQUOTE] = ACTIONS(2492),
    [anon_sym_u8_DQUOTE] = ACTIONS(2492),
    [sym_true] = ACTIONS(2490),
    [sym_false] = ACTIONS(2490),
    [anon_sym_NULL] = ACTIONS(2490),
    [anon_sym_nullptr] = ACTIONS(2490),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2490),
    [anon_sym___typeof] = ACTIONS(2490),
    [anon_sym_typeof] = ACTIONS(2490),
    [anon_sym_ATimport] = ACTIONS(2492),
    [aux_sym_preproc_undef_token1] = ACTIONS(2490),
    [anon_sym_POUND] = ACTIONS(2490),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2490),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2490),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2490),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_API_AVAILABLE] = ACTIONS(2490),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_API_DEPRECATED] = ACTIONS(2490),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym___deprecated_msg] = ACTIONS(2490),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2490),
    [anon_sym_ATprotocol] = ACTIONS(2492),
    [anon_sym_ATinterface] = ACTIONS(2492),
    [anon_sym_ATimplementation] = ACTIONS(2492),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2492),
    [anon_sym_ATtry] = ACTIONS(2492),
    [anon_sym___try] = ACTIONS(2490),
    [anon_sym_ATthrow] = ACTIONS(2492),
    [anon_sym_ATselector] = ACTIONS(2492),
    [anon_sym_ATavailable] = ACTIONS(2492),
    [anon_sym___builtin_available] = ACTIONS(2490),
    [anon_sym_va_arg] = ACTIONS(2490),
    [anon_sym_ATencode] = ACTIONS(2492),
    [anon_sym_ATsynchronized] = ACTIONS(2492),
    [anon_sym_BOOL] = ACTIONS(2490),
    [anon_sym_IMP] = ACTIONS(2490),
    [anon_sym_SEL] = ACTIONS(2490),
    [anon_sym_Class] = ACTIONS(2490),
    [anon_sym_id] = ACTIONS(2490),
  },
  [STATE(726)] = {
    [ts_builtin_sym_end] = ACTIONS(2496),
    [sym_identifier] = ACTIONS(2494),
    [aux_sym_preproc_include_token1] = ACTIONS(2494),
    [aux_sym_preproc_include_token2] = ACTIONS(2494),
    [aux_sym_preproc_def_token1] = ACTIONS(2494),
    [aux_sym_preproc_if_token1] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2494),
    [sym_preproc_directive] = ACTIONS(2494),
    [anon_sym_LPAREN2] = ACTIONS(2496),
    [anon_sym_BANG] = ACTIONS(2496),
    [anon_sym_TILDE] = ACTIONS(2496),
    [anon_sym_DASH] = ACTIONS(2494),
    [anon_sym_PLUS] = ACTIONS(2494),
    [anon_sym_STAR] = ACTIONS(2496),
    [anon_sym_CARET] = ACTIONS(2496),
    [anon_sym_AMP] = ACTIONS(2496),
    [anon_sym_SEMI] = ACTIONS(2496),
    [anon_sym___extension__] = ACTIONS(2494),
    [anon_sym_typedef] = ACTIONS(2494),
    [anon_sym_extern] = ACTIONS(2494),
    [anon_sym___attribute__] = ACTIONS(2494),
    [anon_sym___attribute] = ACTIONS(2494),
    [anon_sym_noreturn] = ACTIONS(2494),
    [anon_sym_LBRACK] = ACTIONS(2496),
    [anon_sym___declspec] = ACTIONS(2494),
    [anon_sym___cdecl] = ACTIONS(2494),
    [anon_sym___clrcall] = ACTIONS(2494),
    [anon_sym___stdcall] = ACTIONS(2494),
    [anon_sym___fastcall] = ACTIONS(2494),
    [anon_sym___thiscall] = ACTIONS(2494),
    [anon_sym___vectorcall] = ACTIONS(2494),
    [anon_sym_LBRACE] = ACTIONS(2496),
    [anon_sym_signed] = ACTIONS(2494),
    [anon_sym_unsigned] = ACTIONS(2494),
    [anon_sym_long] = ACTIONS(2494),
    [anon_sym_short] = ACTIONS(2494),
    [anon_sym_static] = ACTIONS(2494),
    [anon_sym_ATautoreleasepool] = ACTIONS(2496),
    [anon_sym_auto] = ACTIONS(2494),
    [anon_sym_register] = ACTIONS(2494),
    [anon_sym_inline] = ACTIONS(2494),
    [anon_sym___inline] = ACTIONS(2494),
    [anon_sym___inline__] = ACTIONS(2494),
    [anon_sym___forceinline] = ACTIONS(2494),
    [anon_sym_thread_local] = ACTIONS(2494),
    [anon_sym___thread] = ACTIONS(2494),
    [anon_sym_CG_EXTERN] = ACTIONS(2494),
    [anon_sym_CG_INLINE] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2494),
    [anon_sym_IBOutlet] = ACTIONS(2494),
    [anon_sym_IBInspectable] = ACTIONS(2494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2494),
    [anon_sym_NS_INLINE] = ACTIONS(2494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2494),
    [anon_sym_const] = ACTIONS(2494),
    [anon_sym_constexpr] = ACTIONS(2494),
    [anon_sym_volatile] = ACTIONS(2494),
    [anon_sym_restrict] = ACTIONS(2494),
    [anon_sym___restrict__] = ACTIONS(2494),
    [anon_sym__Atomic] = ACTIONS(2494),
    [anon_sym__Noreturn] = ACTIONS(2494),
    [anon_sym__Nonnull] = ACTIONS(2494),
    [anon_sym_nullable] = ACTIONS(2494),
    [anon_sym__Complex] = ACTIONS(2494),
    [anon_sym__Nullable] = ACTIONS(2494),
    [anon_sym__Nullable_result] = ACTIONS(2494),
    [anon_sym__Null_unspecified] = ACTIONS(2494),
    [anon_sym___autoreleasing] = ACTIONS(2494),
    [anon_sym___block] = ACTIONS(2494),
    [anon_sym___bridge] = ACTIONS(2494),
    [anon_sym___bridge_retained] = ACTIONS(2494),
    [anon_sym___bridge_transfer] = ACTIONS(2494),
    [anon_sym___complex] = ACTIONS(2494),
    [anon_sym___const] = ACTIONS(2494),
    [anon_sym___imag] = ACTIONS(2494),
    [anon_sym___kindof] = ACTIONS(2494),
    [anon_sym___nonnull] = ACTIONS(2494),
    [anon_sym___nullable] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2494),
    [anon_sym___real] = ACTIONS(2494),
    [anon_sym___strong] = ACTIONS(2494),
    [anon_sym___unsafe_unretained] = ACTIONS(2494),
    [anon_sym___unused] = ACTIONS(2494),
    [anon_sym___weak] = ACTIONS(2494),
    [anon_sym_alignas] = ACTIONS(2494),
    [anon_sym__Alignas] = ACTIONS(2494),
    [sym_primitive_type] = ACTIONS(2494),
    [anon_sym_enum] = ACTIONS(2494),
    [anon_sym_struct] = ACTIONS(2494),
    [anon_sym_union] = ACTIONS(2494),
    [anon_sym_if] = ACTIONS(2494),
    [anon_sym_else] = ACTIONS(2494),
    [anon_sym_switch] = ACTIONS(2494),
    [anon_sym_case] = ACTIONS(2494),
    [anon_sym_default] = ACTIONS(2494),
    [anon_sym_while] = ACTIONS(2494),
    [anon_sym_do] = ACTIONS(2494),
    [anon_sym_for] = ACTIONS(2494),
    [anon_sym_in] = ACTIONS(2494),
    [anon_sym_return] = ACTIONS(2494),
    [anon_sym_break] = ACTIONS(2494),
    [anon_sym_continue] = ACTIONS(2494),
    [anon_sym_goto] = ACTIONS(2494),
    [anon_sym_DASH_DASH] = ACTIONS(2496),
    [anon_sym_PLUS_PLUS] = ACTIONS(2496),
    [anon_sym_sizeof] = ACTIONS(2494),
    [anon_sym___alignof__] = ACTIONS(2494),
    [anon_sym___alignof] = ACTIONS(2494),
    [anon_sym__alignof] = ACTIONS(2494),
    [anon_sym_alignof] = ACTIONS(2494),
    [anon_sym__Alignof] = ACTIONS(2494),
    [anon_sym_offsetof] = ACTIONS(2494),
    [anon_sym__Generic] = ACTIONS(2494),
    [anon_sym_asm] = ACTIONS(2494),
    [anon_sym___asm__] = ACTIONS(2494),
    [anon_sym___asm] = ACTIONS(2494),
    [sym_number_literal] = ACTIONS(2496),
    [anon_sym_L_SQUOTE] = ACTIONS(2496),
    [anon_sym_u_SQUOTE] = ACTIONS(2496),
    [anon_sym_U_SQUOTE] = ACTIONS(2496),
    [anon_sym_u8_SQUOTE] = ACTIONS(2496),
    [anon_sym_SQUOTE] = ACTIONS(2496),
    [anon_sym_AT] = ACTIONS(2494),
    [anon_sym_DQUOTE] = ACTIONS(2496),
    [anon_sym_L_DQUOTE] = ACTIONS(2496),
    [anon_sym_u_DQUOTE] = ACTIONS(2496),
    [anon_sym_U_DQUOTE] = ACTIONS(2496),
    [anon_sym_u8_DQUOTE] = ACTIONS(2496),
    [sym_true] = ACTIONS(2494),
    [sym_false] = ACTIONS(2494),
    [anon_sym_NULL] = ACTIONS(2494),
    [anon_sym_nullptr] = ACTIONS(2494),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2494),
    [anon_sym___typeof] = ACTIONS(2494),
    [anon_sym_typeof] = ACTIONS(2494),
    [anon_sym_ATimport] = ACTIONS(2496),
    [aux_sym_preproc_undef_token1] = ACTIONS(2494),
    [anon_sym_POUND] = ACTIONS(2494),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_API_AVAILABLE] = ACTIONS(2494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_API_DEPRECATED] = ACTIONS(2494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym___deprecated_msg] = ACTIONS(2494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2494),
    [anon_sym_ATprotocol] = ACTIONS(2496),
    [anon_sym_ATinterface] = ACTIONS(2496),
    [anon_sym_ATimplementation] = ACTIONS(2496),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2496),
    [anon_sym_ATtry] = ACTIONS(2496),
    [anon_sym___try] = ACTIONS(2494),
    [anon_sym_ATthrow] = ACTIONS(2496),
    [anon_sym_ATselector] = ACTIONS(2496),
    [anon_sym_ATavailable] = ACTIONS(2496),
    [anon_sym___builtin_available] = ACTIONS(2494),
    [anon_sym_va_arg] = ACTIONS(2494),
    [anon_sym_ATencode] = ACTIONS(2496),
    [anon_sym_ATsynchronized] = ACTIONS(2496),
    [anon_sym_BOOL] = ACTIONS(2494),
    [anon_sym_IMP] = ACTIONS(2494),
    [anon_sym_SEL] = ACTIONS(2494),
    [anon_sym_Class] = ACTIONS(2494),
    [anon_sym_id] = ACTIONS(2494),
  },
  [STATE(727)] = {
    [ts_builtin_sym_end] = ACTIONS(2500),
    [sym_identifier] = ACTIONS(2498),
    [aux_sym_preproc_include_token1] = ACTIONS(2498),
    [aux_sym_preproc_include_token2] = ACTIONS(2498),
    [aux_sym_preproc_def_token1] = ACTIONS(2498),
    [aux_sym_preproc_if_token1] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2498),
    [sym_preproc_directive] = ACTIONS(2498),
    [anon_sym_LPAREN2] = ACTIONS(2500),
    [anon_sym_BANG] = ACTIONS(2500),
    [anon_sym_TILDE] = ACTIONS(2500),
    [anon_sym_DASH] = ACTIONS(2498),
    [anon_sym_PLUS] = ACTIONS(2498),
    [anon_sym_STAR] = ACTIONS(2500),
    [anon_sym_CARET] = ACTIONS(2500),
    [anon_sym_AMP] = ACTIONS(2500),
    [anon_sym_SEMI] = ACTIONS(2500),
    [anon_sym___extension__] = ACTIONS(2498),
    [anon_sym_typedef] = ACTIONS(2498),
    [anon_sym_extern] = ACTIONS(2498),
    [anon_sym___attribute__] = ACTIONS(2498),
    [anon_sym___attribute] = ACTIONS(2498),
    [anon_sym_noreturn] = ACTIONS(2498),
    [anon_sym_LBRACK] = ACTIONS(2500),
    [anon_sym___declspec] = ACTIONS(2498),
    [anon_sym___cdecl] = ACTIONS(2498),
    [anon_sym___clrcall] = ACTIONS(2498),
    [anon_sym___stdcall] = ACTIONS(2498),
    [anon_sym___fastcall] = ACTIONS(2498),
    [anon_sym___thiscall] = ACTIONS(2498),
    [anon_sym___vectorcall] = ACTIONS(2498),
    [anon_sym_LBRACE] = ACTIONS(2500),
    [anon_sym_signed] = ACTIONS(2498),
    [anon_sym_unsigned] = ACTIONS(2498),
    [anon_sym_long] = ACTIONS(2498),
    [anon_sym_short] = ACTIONS(2498),
    [anon_sym_static] = ACTIONS(2498),
    [anon_sym_ATautoreleasepool] = ACTIONS(2500),
    [anon_sym_auto] = ACTIONS(2498),
    [anon_sym_register] = ACTIONS(2498),
    [anon_sym_inline] = ACTIONS(2498),
    [anon_sym___inline] = ACTIONS(2498),
    [anon_sym___inline__] = ACTIONS(2498),
    [anon_sym___forceinline] = ACTIONS(2498),
    [anon_sym_thread_local] = ACTIONS(2498),
    [anon_sym___thread] = ACTIONS(2498),
    [anon_sym_CG_EXTERN] = ACTIONS(2498),
    [anon_sym_CG_INLINE] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2498),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2498),
    [anon_sym_IBOutlet] = ACTIONS(2498),
    [anon_sym_IBInspectable] = ACTIONS(2498),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2498),
    [anon_sym_NS_INLINE] = ACTIONS(2498),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2498),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2498),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2498),
    [anon_sym_const] = ACTIONS(2498),
    [anon_sym_constexpr] = ACTIONS(2498),
    [anon_sym_volatile] = ACTIONS(2498),
    [anon_sym_restrict] = ACTIONS(2498),
    [anon_sym___restrict__] = ACTIONS(2498),
    [anon_sym__Atomic] = ACTIONS(2498),
    [anon_sym__Noreturn] = ACTIONS(2498),
    [anon_sym__Nonnull] = ACTIONS(2498),
    [anon_sym_nullable] = ACTIONS(2498),
    [anon_sym__Complex] = ACTIONS(2498),
    [anon_sym__Nullable] = ACTIONS(2498),
    [anon_sym__Nullable_result] = ACTIONS(2498),
    [anon_sym__Null_unspecified] = ACTIONS(2498),
    [anon_sym___autoreleasing] = ACTIONS(2498),
    [anon_sym___block] = ACTIONS(2498),
    [anon_sym___bridge] = ACTIONS(2498),
    [anon_sym___bridge_retained] = ACTIONS(2498),
    [anon_sym___bridge_transfer] = ACTIONS(2498),
    [anon_sym___complex] = ACTIONS(2498),
    [anon_sym___const] = ACTIONS(2498),
    [anon_sym___imag] = ACTIONS(2498),
    [anon_sym___kindof] = ACTIONS(2498),
    [anon_sym___nonnull] = ACTIONS(2498),
    [anon_sym___nullable] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2498),
    [anon_sym___real] = ACTIONS(2498),
    [anon_sym___strong] = ACTIONS(2498),
    [anon_sym___unsafe_unretained] = ACTIONS(2498),
    [anon_sym___unused] = ACTIONS(2498),
    [anon_sym___weak] = ACTIONS(2498),
    [anon_sym_alignas] = ACTIONS(2498),
    [anon_sym__Alignas] = ACTIONS(2498),
    [sym_primitive_type] = ACTIONS(2498),
    [anon_sym_enum] = ACTIONS(2498),
    [anon_sym_struct] = ACTIONS(2498),
    [anon_sym_union] = ACTIONS(2498),
    [anon_sym_if] = ACTIONS(2498),
    [anon_sym_else] = ACTIONS(2498),
    [anon_sym_switch] = ACTIONS(2498),
    [anon_sym_case] = ACTIONS(2498),
    [anon_sym_default] = ACTIONS(2498),
    [anon_sym_while] = ACTIONS(2498),
    [anon_sym_do] = ACTIONS(2498),
    [anon_sym_for] = ACTIONS(2498),
    [anon_sym_in] = ACTIONS(2498),
    [anon_sym_return] = ACTIONS(2498),
    [anon_sym_break] = ACTIONS(2498),
    [anon_sym_continue] = ACTIONS(2498),
    [anon_sym_goto] = ACTIONS(2498),
    [anon_sym_DASH_DASH] = ACTIONS(2500),
    [anon_sym_PLUS_PLUS] = ACTIONS(2500),
    [anon_sym_sizeof] = ACTIONS(2498),
    [anon_sym___alignof__] = ACTIONS(2498),
    [anon_sym___alignof] = ACTIONS(2498),
    [anon_sym__alignof] = ACTIONS(2498),
    [anon_sym_alignof] = ACTIONS(2498),
    [anon_sym__Alignof] = ACTIONS(2498),
    [anon_sym_offsetof] = ACTIONS(2498),
    [anon_sym__Generic] = ACTIONS(2498),
    [anon_sym_asm] = ACTIONS(2498),
    [anon_sym___asm__] = ACTIONS(2498),
    [anon_sym___asm] = ACTIONS(2498),
    [sym_number_literal] = ACTIONS(2500),
    [anon_sym_L_SQUOTE] = ACTIONS(2500),
    [anon_sym_u_SQUOTE] = ACTIONS(2500),
    [anon_sym_U_SQUOTE] = ACTIONS(2500),
    [anon_sym_u8_SQUOTE] = ACTIONS(2500),
    [anon_sym_SQUOTE] = ACTIONS(2500),
    [anon_sym_AT] = ACTIONS(2498),
    [anon_sym_DQUOTE] = ACTIONS(2500),
    [anon_sym_L_DQUOTE] = ACTIONS(2500),
    [anon_sym_u_DQUOTE] = ACTIONS(2500),
    [anon_sym_U_DQUOTE] = ACTIONS(2500),
    [anon_sym_u8_DQUOTE] = ACTIONS(2500),
    [sym_true] = ACTIONS(2498),
    [sym_false] = ACTIONS(2498),
    [anon_sym_NULL] = ACTIONS(2498),
    [anon_sym_nullptr] = ACTIONS(2498),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2498),
    [anon_sym___typeof] = ACTIONS(2498),
    [anon_sym_typeof] = ACTIONS(2498),
    [anon_sym_ATimport] = ACTIONS(2500),
    [aux_sym_preproc_undef_token1] = ACTIONS(2498),
    [anon_sym_POUND] = ACTIONS(2498),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2498),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2498),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2498),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_API_AVAILABLE] = ACTIONS(2498),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_API_DEPRECATED] = ACTIONS(2498),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym___deprecated_msg] = ACTIONS(2498),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2498),
    [anon_sym_ATprotocol] = ACTIONS(2500),
    [anon_sym_ATinterface] = ACTIONS(2500),
    [anon_sym_ATimplementation] = ACTIONS(2500),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2500),
    [anon_sym_ATtry] = ACTIONS(2500),
    [anon_sym___try] = ACTIONS(2498),
    [anon_sym_ATthrow] = ACTIONS(2500),
    [anon_sym_ATselector] = ACTIONS(2500),
    [anon_sym_ATavailable] = ACTIONS(2500),
    [anon_sym___builtin_available] = ACTIONS(2498),
    [anon_sym_va_arg] = ACTIONS(2498),
    [anon_sym_ATencode] = ACTIONS(2500),
    [anon_sym_ATsynchronized] = ACTIONS(2500),
    [anon_sym_BOOL] = ACTIONS(2498),
    [anon_sym_IMP] = ACTIONS(2498),
    [anon_sym_SEL] = ACTIONS(2498),
    [anon_sym_Class] = ACTIONS(2498),
    [anon_sym_id] = ACTIONS(2498),
  },
  [STATE(728)] = {
    [sym_identifier] = ACTIONS(2526),
    [aux_sym_preproc_include_token1] = ACTIONS(2526),
    [aux_sym_preproc_include_token2] = ACTIONS(2526),
    [aux_sym_preproc_def_token1] = ACTIONS(2526),
    [aux_sym_preproc_if_token1] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2526),
    [sym_preproc_directive] = ACTIONS(2526),
    [anon_sym_LPAREN2] = ACTIONS(2528),
    [anon_sym_BANG] = ACTIONS(2528),
    [anon_sym_TILDE] = ACTIONS(2528),
    [anon_sym_DASH] = ACTIONS(2526),
    [anon_sym_PLUS] = ACTIONS(2526),
    [anon_sym_STAR] = ACTIONS(2528),
    [anon_sym_CARET] = ACTIONS(2528),
    [anon_sym_AMP] = ACTIONS(2528),
    [anon_sym_SEMI] = ACTIONS(2528),
    [anon_sym___extension__] = ACTIONS(2526),
    [anon_sym_typedef] = ACTIONS(2526),
    [anon_sym_extern] = ACTIONS(2526),
    [anon_sym___attribute__] = ACTIONS(2526),
    [anon_sym___attribute] = ACTIONS(2526),
    [anon_sym_noreturn] = ACTIONS(2526),
    [anon_sym_LBRACK] = ACTIONS(2528),
    [anon_sym___declspec] = ACTIONS(2526),
    [anon_sym___cdecl] = ACTIONS(2526),
    [anon_sym___clrcall] = ACTIONS(2526),
    [anon_sym___stdcall] = ACTIONS(2526),
    [anon_sym___fastcall] = ACTIONS(2526),
    [anon_sym___thiscall] = ACTIONS(2526),
    [anon_sym___vectorcall] = ACTIONS(2526),
    [anon_sym_LBRACE] = ACTIONS(2528),
    [anon_sym_RBRACE] = ACTIONS(2528),
    [anon_sym_signed] = ACTIONS(2526),
    [anon_sym_unsigned] = ACTIONS(2526),
    [anon_sym_long] = ACTIONS(2526),
    [anon_sym_short] = ACTIONS(2526),
    [anon_sym_static] = ACTIONS(2526),
    [anon_sym_ATautoreleasepool] = ACTIONS(2528),
    [anon_sym_auto] = ACTIONS(2526),
    [anon_sym_register] = ACTIONS(2526),
    [anon_sym_inline] = ACTIONS(2526),
    [anon_sym___inline] = ACTIONS(2526),
    [anon_sym___inline__] = ACTIONS(2526),
    [anon_sym___forceinline] = ACTIONS(2526),
    [anon_sym_thread_local] = ACTIONS(2526),
    [anon_sym___thread] = ACTIONS(2526),
    [anon_sym_CG_EXTERN] = ACTIONS(2526),
    [anon_sym_CG_INLINE] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2526),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2526),
    [anon_sym_IBOutlet] = ACTIONS(2526),
    [anon_sym_IBInspectable] = ACTIONS(2526),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2526),
    [anon_sym_NS_INLINE] = ACTIONS(2526),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2526),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2526),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2526),
    [anon_sym_const] = ACTIONS(2526),
    [anon_sym_constexpr] = ACTIONS(2526),
    [anon_sym_volatile] = ACTIONS(2526),
    [anon_sym_restrict] = ACTIONS(2526),
    [anon_sym___restrict__] = ACTIONS(2526),
    [anon_sym__Atomic] = ACTIONS(2526),
    [anon_sym__Noreturn] = ACTIONS(2526),
    [anon_sym__Nonnull] = ACTIONS(2526),
    [anon_sym_nullable] = ACTIONS(2526),
    [anon_sym__Complex] = ACTIONS(2526),
    [anon_sym__Nullable] = ACTIONS(2526),
    [anon_sym__Nullable_result] = ACTIONS(2526),
    [anon_sym__Null_unspecified] = ACTIONS(2526),
    [anon_sym___autoreleasing] = ACTIONS(2526),
    [anon_sym___block] = ACTIONS(2526),
    [anon_sym___bridge] = ACTIONS(2526),
    [anon_sym___bridge_retained] = ACTIONS(2526),
    [anon_sym___bridge_transfer] = ACTIONS(2526),
    [anon_sym___complex] = ACTIONS(2526),
    [anon_sym___const] = ACTIONS(2526),
    [anon_sym___imag] = ACTIONS(2526),
    [anon_sym___kindof] = ACTIONS(2526),
    [anon_sym___nonnull] = ACTIONS(2526),
    [anon_sym___nullable] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2526),
    [anon_sym___real] = ACTIONS(2526),
    [anon_sym___strong] = ACTIONS(2526),
    [anon_sym___unsafe_unretained] = ACTIONS(2526),
    [anon_sym___unused] = ACTIONS(2526),
    [anon_sym___weak] = ACTIONS(2526),
    [anon_sym_alignas] = ACTIONS(2526),
    [anon_sym__Alignas] = ACTIONS(2526),
    [sym_primitive_type] = ACTIONS(2526),
    [anon_sym_enum] = ACTIONS(2526),
    [anon_sym_struct] = ACTIONS(2526),
    [anon_sym_union] = ACTIONS(2526),
    [anon_sym_if] = ACTIONS(2526),
    [anon_sym_else] = ACTIONS(2526),
    [anon_sym_switch] = ACTIONS(2526),
    [anon_sym_case] = ACTIONS(2526),
    [anon_sym_default] = ACTIONS(2526),
    [anon_sym_while] = ACTIONS(2526),
    [anon_sym_do] = ACTIONS(2526),
    [anon_sym_for] = ACTIONS(2526),
    [anon_sym_in] = ACTIONS(2526),
    [anon_sym_return] = ACTIONS(2526),
    [anon_sym_break] = ACTIONS(2526),
    [anon_sym_continue] = ACTIONS(2526),
    [anon_sym_goto] = ACTIONS(2526),
    [anon_sym_DASH_DASH] = ACTIONS(2528),
    [anon_sym_PLUS_PLUS] = ACTIONS(2528),
    [anon_sym_sizeof] = ACTIONS(2526),
    [anon_sym___alignof__] = ACTIONS(2526),
    [anon_sym___alignof] = ACTIONS(2526),
    [anon_sym__alignof] = ACTIONS(2526),
    [anon_sym_alignof] = ACTIONS(2526),
    [anon_sym__Alignof] = ACTIONS(2526),
    [anon_sym_offsetof] = ACTIONS(2526),
    [anon_sym__Generic] = ACTIONS(2526),
    [anon_sym_asm] = ACTIONS(2526),
    [anon_sym___asm__] = ACTIONS(2526),
    [anon_sym___asm] = ACTIONS(2526),
    [sym_number_literal] = ACTIONS(2528),
    [anon_sym_L_SQUOTE] = ACTIONS(2528),
    [anon_sym_u_SQUOTE] = ACTIONS(2528),
    [anon_sym_U_SQUOTE] = ACTIONS(2528),
    [anon_sym_u8_SQUOTE] = ACTIONS(2528),
    [anon_sym_SQUOTE] = ACTIONS(2528),
    [anon_sym_AT] = ACTIONS(2526),
    [anon_sym_DQUOTE] = ACTIONS(2528),
    [anon_sym_L_DQUOTE] = ACTIONS(2528),
    [anon_sym_u_DQUOTE] = ACTIONS(2528),
    [anon_sym_U_DQUOTE] = ACTIONS(2528),
    [anon_sym_u8_DQUOTE] = ACTIONS(2528),
    [sym_true] = ACTIONS(2526),
    [sym_false] = ACTIONS(2526),
    [anon_sym_NULL] = ACTIONS(2526),
    [anon_sym_nullptr] = ACTIONS(2526),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2526),
    [anon_sym___typeof] = ACTIONS(2526),
    [anon_sym_typeof] = ACTIONS(2526),
    [anon_sym_ATimport] = ACTIONS(2528),
    [aux_sym_preproc_undef_token1] = ACTIONS(2526),
    [anon_sym_POUND] = ACTIONS(2526),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2526),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2526),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2526),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_API_AVAILABLE] = ACTIONS(2526),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_API_DEPRECATED] = ACTIONS(2526),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym___deprecated_msg] = ACTIONS(2526),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2526),
    [anon_sym_ATprotocol] = ACTIONS(2528),
    [anon_sym_ATinterface] = ACTIONS(2528),
    [anon_sym_ATimplementation] = ACTIONS(2528),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2528),
    [anon_sym_ATtry] = ACTIONS(2528),
    [anon_sym___try] = ACTIONS(2526),
    [anon_sym_ATthrow] = ACTIONS(2528),
    [anon_sym_ATselector] = ACTIONS(2528),
    [anon_sym_ATavailable] = ACTIONS(2528),
    [anon_sym___builtin_available] = ACTIONS(2526),
    [anon_sym_va_arg] = ACTIONS(2526),
    [anon_sym_ATencode] = ACTIONS(2528),
    [anon_sym_ATsynchronized] = ACTIONS(2528),
    [anon_sym_BOOL] = ACTIONS(2526),
    [anon_sym_IMP] = ACTIONS(2526),
    [anon_sym_SEL] = ACTIONS(2526),
    [anon_sym_Class] = ACTIONS(2526),
    [anon_sym_id] = ACTIONS(2526),
  },
  [STATE(729)] = {
    [ts_builtin_sym_end] = ACTIONS(2422),
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(730)] = {
    [sym_identifier] = ACTIONS(2530),
    [aux_sym_preproc_include_token1] = ACTIONS(2530),
    [aux_sym_preproc_include_token2] = ACTIONS(2530),
    [aux_sym_preproc_def_token1] = ACTIONS(2530),
    [aux_sym_preproc_if_token1] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2530),
    [sym_preproc_directive] = ACTIONS(2530),
    [anon_sym_LPAREN2] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2530),
    [anon_sym_PLUS] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym___extension__] = ACTIONS(2530),
    [anon_sym_typedef] = ACTIONS(2530),
    [anon_sym_extern] = ACTIONS(2530),
    [anon_sym___attribute__] = ACTIONS(2530),
    [anon_sym___attribute] = ACTIONS(2530),
    [anon_sym_noreturn] = ACTIONS(2530),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym___declspec] = ACTIONS(2530),
    [anon_sym___cdecl] = ACTIONS(2530),
    [anon_sym___clrcall] = ACTIONS(2530),
    [anon_sym___stdcall] = ACTIONS(2530),
    [anon_sym___fastcall] = ACTIONS(2530),
    [anon_sym___thiscall] = ACTIONS(2530),
    [anon_sym___vectorcall] = ACTIONS(2530),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_signed] = ACTIONS(2530),
    [anon_sym_unsigned] = ACTIONS(2530),
    [anon_sym_long] = ACTIONS(2530),
    [anon_sym_short] = ACTIONS(2530),
    [anon_sym_static] = ACTIONS(2530),
    [anon_sym_ATautoreleasepool] = ACTIONS(2532),
    [anon_sym_auto] = ACTIONS(2530),
    [anon_sym_register] = ACTIONS(2530),
    [anon_sym_inline] = ACTIONS(2530),
    [anon_sym___inline] = ACTIONS(2530),
    [anon_sym___inline__] = ACTIONS(2530),
    [anon_sym___forceinline] = ACTIONS(2530),
    [anon_sym_thread_local] = ACTIONS(2530),
    [anon_sym___thread] = ACTIONS(2530),
    [anon_sym_CG_EXTERN] = ACTIONS(2530),
    [anon_sym_CG_INLINE] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2530),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2530),
    [anon_sym_IBOutlet] = ACTIONS(2530),
    [anon_sym_IBInspectable] = ACTIONS(2530),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2530),
    [anon_sym_NS_INLINE] = ACTIONS(2530),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2530),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2530),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2530),
    [anon_sym_const] = ACTIONS(2530),
    [anon_sym_constexpr] = ACTIONS(2530),
    [anon_sym_volatile] = ACTIONS(2530),
    [anon_sym_restrict] = ACTIONS(2530),
    [anon_sym___restrict__] = ACTIONS(2530),
    [anon_sym__Atomic] = ACTIONS(2530),
    [anon_sym__Noreturn] = ACTIONS(2530),
    [anon_sym__Nonnull] = ACTIONS(2530),
    [anon_sym_nullable] = ACTIONS(2530),
    [anon_sym__Complex] = ACTIONS(2530),
    [anon_sym__Nullable] = ACTIONS(2530),
    [anon_sym__Nullable_result] = ACTIONS(2530),
    [anon_sym__Null_unspecified] = ACTIONS(2530),
    [anon_sym___autoreleasing] = ACTIONS(2530),
    [anon_sym___block] = ACTIONS(2530),
    [anon_sym___bridge] = ACTIONS(2530),
    [anon_sym___bridge_retained] = ACTIONS(2530),
    [anon_sym___bridge_transfer] = ACTIONS(2530),
    [anon_sym___complex] = ACTIONS(2530),
    [anon_sym___const] = ACTIONS(2530),
    [anon_sym___imag] = ACTIONS(2530),
    [anon_sym___kindof] = ACTIONS(2530),
    [anon_sym___nonnull] = ACTIONS(2530),
    [anon_sym___nullable] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2530),
    [anon_sym___real] = ACTIONS(2530),
    [anon_sym___strong] = ACTIONS(2530),
    [anon_sym___unsafe_unretained] = ACTIONS(2530),
    [anon_sym___unused] = ACTIONS(2530),
    [anon_sym___weak] = ACTIONS(2530),
    [anon_sym_alignas] = ACTIONS(2530),
    [anon_sym__Alignas] = ACTIONS(2530),
    [sym_primitive_type] = ACTIONS(2530),
    [anon_sym_enum] = ACTIONS(2530),
    [anon_sym_struct] = ACTIONS(2530),
    [anon_sym_union] = ACTIONS(2530),
    [anon_sym_if] = ACTIONS(2530),
    [anon_sym_else] = ACTIONS(2530),
    [anon_sym_switch] = ACTIONS(2530),
    [anon_sym_case] = ACTIONS(2530),
    [anon_sym_default] = ACTIONS(2530),
    [anon_sym_while] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2530),
    [anon_sym_for] = ACTIONS(2530),
    [anon_sym_in] = ACTIONS(2530),
    [anon_sym_return] = ACTIONS(2530),
    [anon_sym_break] = ACTIONS(2530),
    [anon_sym_continue] = ACTIONS(2530),
    [anon_sym_goto] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_sizeof] = ACTIONS(2530),
    [anon_sym___alignof__] = ACTIONS(2530),
    [anon_sym___alignof] = ACTIONS(2530),
    [anon_sym__alignof] = ACTIONS(2530),
    [anon_sym_alignof] = ACTIONS(2530),
    [anon_sym__Alignof] = ACTIONS(2530),
    [anon_sym_offsetof] = ACTIONS(2530),
    [anon_sym__Generic] = ACTIONS(2530),
    [anon_sym_asm] = ACTIONS(2530),
    [anon_sym___asm__] = ACTIONS(2530),
    [anon_sym___asm] = ACTIONS(2530),
    [sym_number_literal] = ACTIONS(2532),
    [anon_sym_L_SQUOTE] = ACTIONS(2532),
    [anon_sym_u_SQUOTE] = ACTIONS(2532),
    [anon_sym_U_SQUOTE] = ACTIONS(2532),
    [anon_sym_u8_SQUOTE] = ACTIONS(2532),
    [anon_sym_SQUOTE] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2530),
    [anon_sym_DQUOTE] = ACTIONS(2532),
    [anon_sym_L_DQUOTE] = ACTIONS(2532),
    [anon_sym_u_DQUOTE] = ACTIONS(2532),
    [anon_sym_U_DQUOTE] = ACTIONS(2532),
    [anon_sym_u8_DQUOTE] = ACTIONS(2532),
    [sym_true] = ACTIONS(2530),
    [sym_false] = ACTIONS(2530),
    [anon_sym_NULL] = ACTIONS(2530),
    [anon_sym_nullptr] = ACTIONS(2530),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2530),
    [anon_sym___typeof] = ACTIONS(2530),
    [anon_sym_typeof] = ACTIONS(2530),
    [anon_sym_ATimport] = ACTIONS(2532),
    [aux_sym_preproc_undef_token1] = ACTIONS(2530),
    [anon_sym_POUND] = ACTIONS(2530),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2530),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2530),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2530),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_API_AVAILABLE] = ACTIONS(2530),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_API_DEPRECATED] = ACTIONS(2530),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym___deprecated_msg] = ACTIONS(2530),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2530),
    [anon_sym_ATprotocol] = ACTIONS(2532),
    [anon_sym_ATinterface] = ACTIONS(2532),
    [anon_sym_ATimplementation] = ACTIONS(2532),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2532),
    [anon_sym_ATtry] = ACTIONS(2532),
    [anon_sym___try] = ACTIONS(2530),
    [anon_sym_ATthrow] = ACTIONS(2532),
    [anon_sym_ATselector] = ACTIONS(2532),
    [anon_sym_ATavailable] = ACTIONS(2532),
    [anon_sym___builtin_available] = ACTIONS(2530),
    [anon_sym_va_arg] = ACTIONS(2530),
    [anon_sym_ATencode] = ACTIONS(2532),
    [anon_sym_ATsynchronized] = ACTIONS(2532),
    [anon_sym_BOOL] = ACTIONS(2530),
    [anon_sym_IMP] = ACTIONS(2530),
    [anon_sym_SEL] = ACTIONS(2530),
    [anon_sym_Class] = ACTIONS(2530),
    [anon_sym_id] = ACTIONS(2530),
  },
  [STATE(731)] = {
    [sym_identifier] = ACTIONS(2534),
    [aux_sym_preproc_include_token1] = ACTIONS(2534),
    [aux_sym_preproc_include_token2] = ACTIONS(2534),
    [aux_sym_preproc_def_token1] = ACTIONS(2534),
    [aux_sym_preproc_if_token1] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2534),
    [sym_preproc_directive] = ACTIONS(2534),
    [anon_sym_LPAREN2] = ACTIONS(2536),
    [anon_sym_BANG] = ACTIONS(2536),
    [anon_sym_TILDE] = ACTIONS(2536),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2536),
    [anon_sym_CARET] = ACTIONS(2536),
    [anon_sym_AMP] = ACTIONS(2536),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym___extension__] = ACTIONS(2534),
    [anon_sym_typedef] = ACTIONS(2534),
    [anon_sym_extern] = ACTIONS(2534),
    [anon_sym___attribute__] = ACTIONS(2534),
    [anon_sym___attribute] = ACTIONS(2534),
    [anon_sym_noreturn] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2536),
    [anon_sym___declspec] = ACTIONS(2534),
    [anon_sym___cdecl] = ACTIONS(2534),
    [anon_sym___clrcall] = ACTIONS(2534),
    [anon_sym___stdcall] = ACTIONS(2534),
    [anon_sym___fastcall] = ACTIONS(2534),
    [anon_sym___thiscall] = ACTIONS(2534),
    [anon_sym___vectorcall] = ACTIONS(2534),
    [anon_sym_LBRACE] = ACTIONS(2536),
    [anon_sym_RBRACE] = ACTIONS(2536),
    [anon_sym_signed] = ACTIONS(2534),
    [anon_sym_unsigned] = ACTIONS(2534),
    [anon_sym_long] = ACTIONS(2534),
    [anon_sym_short] = ACTIONS(2534),
    [anon_sym_static] = ACTIONS(2534),
    [anon_sym_ATautoreleasepool] = ACTIONS(2536),
    [anon_sym_auto] = ACTIONS(2534),
    [anon_sym_register] = ACTIONS(2534),
    [anon_sym_inline] = ACTIONS(2534),
    [anon_sym___inline] = ACTIONS(2534),
    [anon_sym___inline__] = ACTIONS(2534),
    [anon_sym___forceinline] = ACTIONS(2534),
    [anon_sym_thread_local] = ACTIONS(2534),
    [anon_sym___thread] = ACTIONS(2534),
    [anon_sym_CG_EXTERN] = ACTIONS(2534),
    [anon_sym_CG_INLINE] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2534),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2534),
    [anon_sym_IBOutlet] = ACTIONS(2534),
    [anon_sym_IBInspectable] = ACTIONS(2534),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2534),
    [anon_sym_NS_INLINE] = ACTIONS(2534),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2534),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2534),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2534),
    [anon_sym_const] = ACTIONS(2534),
    [anon_sym_constexpr] = ACTIONS(2534),
    [anon_sym_volatile] = ACTIONS(2534),
    [anon_sym_restrict] = ACTIONS(2534),
    [anon_sym___restrict__] = ACTIONS(2534),
    [anon_sym__Atomic] = ACTIONS(2534),
    [anon_sym__Noreturn] = ACTIONS(2534),
    [anon_sym__Nonnull] = ACTIONS(2534),
    [anon_sym_nullable] = ACTIONS(2534),
    [anon_sym__Complex] = ACTIONS(2534),
    [anon_sym__Nullable] = ACTIONS(2534),
    [anon_sym__Nullable_result] = ACTIONS(2534),
    [anon_sym__Null_unspecified] = ACTIONS(2534),
    [anon_sym___autoreleasing] = ACTIONS(2534),
    [anon_sym___block] = ACTIONS(2534),
    [anon_sym___bridge] = ACTIONS(2534),
    [anon_sym___bridge_retained] = ACTIONS(2534),
    [anon_sym___bridge_transfer] = ACTIONS(2534),
    [anon_sym___complex] = ACTIONS(2534),
    [anon_sym___const] = ACTIONS(2534),
    [anon_sym___imag] = ACTIONS(2534),
    [anon_sym___kindof] = ACTIONS(2534),
    [anon_sym___nonnull] = ACTIONS(2534),
    [anon_sym___nullable] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2534),
    [anon_sym___real] = ACTIONS(2534),
    [anon_sym___strong] = ACTIONS(2534),
    [anon_sym___unsafe_unretained] = ACTIONS(2534),
    [anon_sym___unused] = ACTIONS(2534),
    [anon_sym___weak] = ACTIONS(2534),
    [anon_sym_alignas] = ACTIONS(2534),
    [anon_sym__Alignas] = ACTIONS(2534),
    [sym_primitive_type] = ACTIONS(2534),
    [anon_sym_enum] = ACTIONS(2534),
    [anon_sym_struct] = ACTIONS(2534),
    [anon_sym_union] = ACTIONS(2534),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_else] = ACTIONS(2534),
    [anon_sym_switch] = ACTIONS(2534),
    [anon_sym_case] = ACTIONS(2534),
    [anon_sym_default] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_do] = ACTIONS(2534),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_in] = ACTIONS(2534),
    [anon_sym_return] = ACTIONS(2534),
    [anon_sym_break] = ACTIONS(2534),
    [anon_sym_continue] = ACTIONS(2534),
    [anon_sym_goto] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_sizeof] = ACTIONS(2534),
    [anon_sym___alignof__] = ACTIONS(2534),
    [anon_sym___alignof] = ACTIONS(2534),
    [anon_sym__alignof] = ACTIONS(2534),
    [anon_sym_alignof] = ACTIONS(2534),
    [anon_sym__Alignof] = ACTIONS(2534),
    [anon_sym_offsetof] = ACTIONS(2534),
    [anon_sym__Generic] = ACTIONS(2534),
    [anon_sym_asm] = ACTIONS(2534),
    [anon_sym___asm__] = ACTIONS(2534),
    [anon_sym___asm] = ACTIONS(2534),
    [sym_number_literal] = ACTIONS(2536),
    [anon_sym_L_SQUOTE] = ACTIONS(2536),
    [anon_sym_u_SQUOTE] = ACTIONS(2536),
    [anon_sym_U_SQUOTE] = ACTIONS(2536),
    [anon_sym_u8_SQUOTE] = ACTIONS(2536),
    [anon_sym_SQUOTE] = ACTIONS(2536),
    [anon_sym_AT] = ACTIONS(2534),
    [anon_sym_DQUOTE] = ACTIONS(2536),
    [anon_sym_L_DQUOTE] = ACTIONS(2536),
    [anon_sym_u_DQUOTE] = ACTIONS(2536),
    [anon_sym_U_DQUOTE] = ACTIONS(2536),
    [anon_sym_u8_DQUOTE] = ACTIONS(2536),
    [sym_true] = ACTIONS(2534),
    [sym_false] = ACTIONS(2534),
    [anon_sym_NULL] = ACTIONS(2534),
    [anon_sym_nullptr] = ACTIONS(2534),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2534),
    [anon_sym___typeof] = ACTIONS(2534),
    [anon_sym_typeof] = ACTIONS(2534),
    [anon_sym_ATimport] = ACTIONS(2536),
    [aux_sym_preproc_undef_token1] = ACTIONS(2534),
    [anon_sym_POUND] = ACTIONS(2534),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2534),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2534),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2534),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_API_AVAILABLE] = ACTIONS(2534),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_API_DEPRECATED] = ACTIONS(2534),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym___deprecated_msg] = ACTIONS(2534),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2534),
    [anon_sym_ATprotocol] = ACTIONS(2536),
    [anon_sym_ATinterface] = ACTIONS(2536),
    [anon_sym_ATimplementation] = ACTIONS(2536),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2536),
    [anon_sym_ATtry] = ACTIONS(2536),
    [anon_sym___try] = ACTIONS(2534),
    [anon_sym_ATthrow] = ACTIONS(2536),
    [anon_sym_ATselector] = ACTIONS(2536),
    [anon_sym_ATavailable] = ACTIONS(2536),
    [anon_sym___builtin_available] = ACTIONS(2534),
    [anon_sym_va_arg] = ACTIONS(2534),
    [anon_sym_ATencode] = ACTIONS(2536),
    [anon_sym_ATsynchronized] = ACTIONS(2536),
    [anon_sym_BOOL] = ACTIONS(2534),
    [anon_sym_IMP] = ACTIONS(2534),
    [anon_sym_SEL] = ACTIONS(2534),
    [anon_sym_Class] = ACTIONS(2534),
    [anon_sym_id] = ACTIONS(2534),
  },
  [STATE(732)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_RBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(733)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_RBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(734)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_RBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(735)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_RBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(736)] = {
    [ts_builtin_sym_end] = ACTIONS(2548),
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(737)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_RBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(738)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_RBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(739)] = {
    [ts_builtin_sym_end] = ACTIONS(2308),
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(740)] = {
    [ts_builtin_sym_end] = ACTIONS(2504),
    [sym_identifier] = ACTIONS(2502),
    [aux_sym_preproc_include_token1] = ACTIONS(2502),
    [aux_sym_preproc_include_token2] = ACTIONS(2502),
    [aux_sym_preproc_def_token1] = ACTIONS(2502),
    [aux_sym_preproc_if_token1] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2502),
    [sym_preproc_directive] = ACTIONS(2502),
    [anon_sym_LPAREN2] = ACTIONS(2504),
    [anon_sym_BANG] = ACTIONS(2504),
    [anon_sym_TILDE] = ACTIONS(2504),
    [anon_sym_DASH] = ACTIONS(2502),
    [anon_sym_PLUS] = ACTIONS(2502),
    [anon_sym_STAR] = ACTIONS(2504),
    [anon_sym_CARET] = ACTIONS(2504),
    [anon_sym_AMP] = ACTIONS(2504),
    [anon_sym_SEMI] = ACTIONS(2504),
    [anon_sym___extension__] = ACTIONS(2502),
    [anon_sym_typedef] = ACTIONS(2502),
    [anon_sym_extern] = ACTIONS(2502),
    [anon_sym___attribute__] = ACTIONS(2502),
    [anon_sym___attribute] = ACTIONS(2502),
    [anon_sym_noreturn] = ACTIONS(2502),
    [anon_sym_LBRACK] = ACTIONS(2504),
    [anon_sym___declspec] = ACTIONS(2502),
    [anon_sym___cdecl] = ACTIONS(2502),
    [anon_sym___clrcall] = ACTIONS(2502),
    [anon_sym___stdcall] = ACTIONS(2502),
    [anon_sym___fastcall] = ACTIONS(2502),
    [anon_sym___thiscall] = ACTIONS(2502),
    [anon_sym___vectorcall] = ACTIONS(2502),
    [anon_sym_LBRACE] = ACTIONS(2504),
    [anon_sym_signed] = ACTIONS(2502),
    [anon_sym_unsigned] = ACTIONS(2502),
    [anon_sym_long] = ACTIONS(2502),
    [anon_sym_short] = ACTIONS(2502),
    [anon_sym_static] = ACTIONS(2502),
    [anon_sym_ATautoreleasepool] = ACTIONS(2504),
    [anon_sym_auto] = ACTIONS(2502),
    [anon_sym_register] = ACTIONS(2502),
    [anon_sym_inline] = ACTIONS(2502),
    [anon_sym___inline] = ACTIONS(2502),
    [anon_sym___inline__] = ACTIONS(2502),
    [anon_sym___forceinline] = ACTIONS(2502),
    [anon_sym_thread_local] = ACTIONS(2502),
    [anon_sym___thread] = ACTIONS(2502),
    [anon_sym_CG_EXTERN] = ACTIONS(2502),
    [anon_sym_CG_INLINE] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2502),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2502),
    [anon_sym_IBOutlet] = ACTIONS(2502),
    [anon_sym_IBInspectable] = ACTIONS(2502),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2502),
    [anon_sym_NS_INLINE] = ACTIONS(2502),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2502),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2502),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2502),
    [anon_sym_const] = ACTIONS(2502),
    [anon_sym_constexpr] = ACTIONS(2502),
    [anon_sym_volatile] = ACTIONS(2502),
    [anon_sym_restrict] = ACTIONS(2502),
    [anon_sym___restrict__] = ACTIONS(2502),
    [anon_sym__Atomic] = ACTIONS(2502),
    [anon_sym__Noreturn] = ACTIONS(2502),
    [anon_sym__Nonnull] = ACTIONS(2502),
    [anon_sym_nullable] = ACTIONS(2502),
    [anon_sym__Complex] = ACTIONS(2502),
    [anon_sym__Nullable] = ACTIONS(2502),
    [anon_sym__Nullable_result] = ACTIONS(2502),
    [anon_sym__Null_unspecified] = ACTIONS(2502),
    [anon_sym___autoreleasing] = ACTIONS(2502),
    [anon_sym___block] = ACTIONS(2502),
    [anon_sym___bridge] = ACTIONS(2502),
    [anon_sym___bridge_retained] = ACTIONS(2502),
    [anon_sym___bridge_transfer] = ACTIONS(2502),
    [anon_sym___complex] = ACTIONS(2502),
    [anon_sym___const] = ACTIONS(2502),
    [anon_sym___imag] = ACTIONS(2502),
    [anon_sym___kindof] = ACTIONS(2502),
    [anon_sym___nonnull] = ACTIONS(2502),
    [anon_sym___nullable] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2502),
    [anon_sym___real] = ACTIONS(2502),
    [anon_sym___strong] = ACTIONS(2502),
    [anon_sym___unsafe_unretained] = ACTIONS(2502),
    [anon_sym___unused] = ACTIONS(2502),
    [anon_sym___weak] = ACTIONS(2502),
    [anon_sym_alignas] = ACTIONS(2502),
    [anon_sym__Alignas] = ACTIONS(2502),
    [sym_primitive_type] = ACTIONS(2502),
    [anon_sym_enum] = ACTIONS(2502),
    [anon_sym_struct] = ACTIONS(2502),
    [anon_sym_union] = ACTIONS(2502),
    [anon_sym_if] = ACTIONS(2502),
    [anon_sym_else] = ACTIONS(2502),
    [anon_sym_switch] = ACTIONS(2502),
    [anon_sym_case] = ACTIONS(2502),
    [anon_sym_default] = ACTIONS(2502),
    [anon_sym_while] = ACTIONS(2502),
    [anon_sym_do] = ACTIONS(2502),
    [anon_sym_for] = ACTIONS(2502),
    [anon_sym_in] = ACTIONS(2502),
    [anon_sym_return] = ACTIONS(2502),
    [anon_sym_break] = ACTIONS(2502),
    [anon_sym_continue] = ACTIONS(2502),
    [anon_sym_goto] = ACTIONS(2502),
    [anon_sym_DASH_DASH] = ACTIONS(2504),
    [anon_sym_PLUS_PLUS] = ACTIONS(2504),
    [anon_sym_sizeof] = ACTIONS(2502),
    [anon_sym___alignof__] = ACTIONS(2502),
    [anon_sym___alignof] = ACTIONS(2502),
    [anon_sym__alignof] = ACTIONS(2502),
    [anon_sym_alignof] = ACTIONS(2502),
    [anon_sym__Alignof] = ACTIONS(2502),
    [anon_sym_offsetof] = ACTIONS(2502),
    [anon_sym__Generic] = ACTIONS(2502),
    [anon_sym_asm] = ACTIONS(2502),
    [anon_sym___asm__] = ACTIONS(2502),
    [anon_sym___asm] = ACTIONS(2502),
    [sym_number_literal] = ACTIONS(2504),
    [anon_sym_L_SQUOTE] = ACTIONS(2504),
    [anon_sym_u_SQUOTE] = ACTIONS(2504),
    [anon_sym_U_SQUOTE] = ACTIONS(2504),
    [anon_sym_u8_SQUOTE] = ACTIONS(2504),
    [anon_sym_SQUOTE] = ACTIONS(2504),
    [anon_sym_AT] = ACTIONS(2502),
    [anon_sym_DQUOTE] = ACTIONS(2504),
    [anon_sym_L_DQUOTE] = ACTIONS(2504),
    [anon_sym_u_DQUOTE] = ACTIONS(2504),
    [anon_sym_U_DQUOTE] = ACTIONS(2504),
    [anon_sym_u8_DQUOTE] = ACTIONS(2504),
    [sym_true] = ACTIONS(2502),
    [sym_false] = ACTIONS(2502),
    [anon_sym_NULL] = ACTIONS(2502),
    [anon_sym_nullptr] = ACTIONS(2502),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2502),
    [anon_sym___typeof] = ACTIONS(2502),
    [anon_sym_typeof] = ACTIONS(2502),
    [anon_sym_ATimport] = ACTIONS(2504),
    [aux_sym_preproc_undef_token1] = ACTIONS(2502),
    [anon_sym_POUND] = ACTIONS(2502),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2502),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2502),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2502),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_API_AVAILABLE] = ACTIONS(2502),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_API_DEPRECATED] = ACTIONS(2502),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym___deprecated_msg] = ACTIONS(2502),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2502),
    [anon_sym_ATprotocol] = ACTIONS(2504),
    [anon_sym_ATinterface] = ACTIONS(2504),
    [anon_sym_ATimplementation] = ACTIONS(2504),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2504),
    [anon_sym_ATtry] = ACTIONS(2504),
    [anon_sym___try] = ACTIONS(2502),
    [anon_sym_ATthrow] = ACTIONS(2504),
    [anon_sym_ATselector] = ACTIONS(2504),
    [anon_sym_ATavailable] = ACTIONS(2504),
    [anon_sym___builtin_available] = ACTIONS(2502),
    [anon_sym_va_arg] = ACTIONS(2502),
    [anon_sym_ATencode] = ACTIONS(2504),
    [anon_sym_ATsynchronized] = ACTIONS(2504),
    [anon_sym_BOOL] = ACTIONS(2502),
    [anon_sym_IMP] = ACTIONS(2502),
    [anon_sym_SEL] = ACTIONS(2502),
    [anon_sym_Class] = ACTIONS(2502),
    [anon_sym_id] = ACTIONS(2502),
  },
  [STATE(741)] = {
    [ts_builtin_sym_end] = ACTIONS(2508),
    [sym_identifier] = ACTIONS(2506),
    [aux_sym_preproc_include_token1] = ACTIONS(2506),
    [aux_sym_preproc_include_token2] = ACTIONS(2506),
    [aux_sym_preproc_def_token1] = ACTIONS(2506),
    [aux_sym_preproc_if_token1] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2506),
    [sym_preproc_directive] = ACTIONS(2506),
    [anon_sym_LPAREN2] = ACTIONS(2508),
    [anon_sym_BANG] = ACTIONS(2508),
    [anon_sym_TILDE] = ACTIONS(2508),
    [anon_sym_DASH] = ACTIONS(2506),
    [anon_sym_PLUS] = ACTIONS(2506),
    [anon_sym_STAR] = ACTIONS(2508),
    [anon_sym_CARET] = ACTIONS(2508),
    [anon_sym_AMP] = ACTIONS(2508),
    [anon_sym_SEMI] = ACTIONS(2508),
    [anon_sym___extension__] = ACTIONS(2506),
    [anon_sym_typedef] = ACTIONS(2506),
    [anon_sym_extern] = ACTIONS(2506),
    [anon_sym___attribute__] = ACTIONS(2506),
    [anon_sym___attribute] = ACTIONS(2506),
    [anon_sym_noreturn] = ACTIONS(2506),
    [anon_sym_LBRACK] = ACTIONS(2508),
    [anon_sym___declspec] = ACTIONS(2506),
    [anon_sym___cdecl] = ACTIONS(2506),
    [anon_sym___clrcall] = ACTIONS(2506),
    [anon_sym___stdcall] = ACTIONS(2506),
    [anon_sym___fastcall] = ACTIONS(2506),
    [anon_sym___thiscall] = ACTIONS(2506),
    [anon_sym___vectorcall] = ACTIONS(2506),
    [anon_sym_LBRACE] = ACTIONS(2508),
    [anon_sym_signed] = ACTIONS(2506),
    [anon_sym_unsigned] = ACTIONS(2506),
    [anon_sym_long] = ACTIONS(2506),
    [anon_sym_short] = ACTIONS(2506),
    [anon_sym_static] = ACTIONS(2506),
    [anon_sym_ATautoreleasepool] = ACTIONS(2508),
    [anon_sym_auto] = ACTIONS(2506),
    [anon_sym_register] = ACTIONS(2506),
    [anon_sym_inline] = ACTIONS(2506),
    [anon_sym___inline] = ACTIONS(2506),
    [anon_sym___inline__] = ACTIONS(2506),
    [anon_sym___forceinline] = ACTIONS(2506),
    [anon_sym_thread_local] = ACTIONS(2506),
    [anon_sym___thread] = ACTIONS(2506),
    [anon_sym_CG_EXTERN] = ACTIONS(2506),
    [anon_sym_CG_INLINE] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2506),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2506),
    [anon_sym_IBOutlet] = ACTIONS(2506),
    [anon_sym_IBInspectable] = ACTIONS(2506),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2506),
    [anon_sym_NS_INLINE] = ACTIONS(2506),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2506),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2506),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2506),
    [anon_sym_const] = ACTIONS(2506),
    [anon_sym_constexpr] = ACTIONS(2506),
    [anon_sym_volatile] = ACTIONS(2506),
    [anon_sym_restrict] = ACTIONS(2506),
    [anon_sym___restrict__] = ACTIONS(2506),
    [anon_sym__Atomic] = ACTIONS(2506),
    [anon_sym__Noreturn] = ACTIONS(2506),
    [anon_sym__Nonnull] = ACTIONS(2506),
    [anon_sym_nullable] = ACTIONS(2506),
    [anon_sym__Complex] = ACTIONS(2506),
    [anon_sym__Nullable] = ACTIONS(2506),
    [anon_sym__Nullable_result] = ACTIONS(2506),
    [anon_sym__Null_unspecified] = ACTIONS(2506),
    [anon_sym___autoreleasing] = ACTIONS(2506),
    [anon_sym___block] = ACTIONS(2506),
    [anon_sym___bridge] = ACTIONS(2506),
    [anon_sym___bridge_retained] = ACTIONS(2506),
    [anon_sym___bridge_transfer] = ACTIONS(2506),
    [anon_sym___complex] = ACTIONS(2506),
    [anon_sym___const] = ACTIONS(2506),
    [anon_sym___imag] = ACTIONS(2506),
    [anon_sym___kindof] = ACTIONS(2506),
    [anon_sym___nonnull] = ACTIONS(2506),
    [anon_sym___nullable] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2506),
    [anon_sym___real] = ACTIONS(2506),
    [anon_sym___strong] = ACTIONS(2506),
    [anon_sym___unsafe_unretained] = ACTIONS(2506),
    [anon_sym___unused] = ACTIONS(2506),
    [anon_sym___weak] = ACTIONS(2506),
    [anon_sym_alignas] = ACTIONS(2506),
    [anon_sym__Alignas] = ACTIONS(2506),
    [sym_primitive_type] = ACTIONS(2506),
    [anon_sym_enum] = ACTIONS(2506),
    [anon_sym_struct] = ACTIONS(2506),
    [anon_sym_union] = ACTIONS(2506),
    [anon_sym_if] = ACTIONS(2506),
    [anon_sym_else] = ACTIONS(2506),
    [anon_sym_switch] = ACTIONS(2506),
    [anon_sym_case] = ACTIONS(2506),
    [anon_sym_default] = ACTIONS(2506),
    [anon_sym_while] = ACTIONS(2506),
    [anon_sym_do] = ACTIONS(2506),
    [anon_sym_for] = ACTIONS(2506),
    [anon_sym_in] = ACTIONS(2506),
    [anon_sym_return] = ACTIONS(2506),
    [anon_sym_break] = ACTIONS(2506),
    [anon_sym_continue] = ACTIONS(2506),
    [anon_sym_goto] = ACTIONS(2506),
    [anon_sym_DASH_DASH] = ACTIONS(2508),
    [anon_sym_PLUS_PLUS] = ACTIONS(2508),
    [anon_sym_sizeof] = ACTIONS(2506),
    [anon_sym___alignof__] = ACTIONS(2506),
    [anon_sym___alignof] = ACTIONS(2506),
    [anon_sym__alignof] = ACTIONS(2506),
    [anon_sym_alignof] = ACTIONS(2506),
    [anon_sym__Alignof] = ACTIONS(2506),
    [anon_sym_offsetof] = ACTIONS(2506),
    [anon_sym__Generic] = ACTIONS(2506),
    [anon_sym_asm] = ACTIONS(2506),
    [anon_sym___asm__] = ACTIONS(2506),
    [anon_sym___asm] = ACTIONS(2506),
    [sym_number_literal] = ACTIONS(2508),
    [anon_sym_L_SQUOTE] = ACTIONS(2508),
    [anon_sym_u_SQUOTE] = ACTIONS(2508),
    [anon_sym_U_SQUOTE] = ACTIONS(2508),
    [anon_sym_u8_SQUOTE] = ACTIONS(2508),
    [anon_sym_SQUOTE] = ACTIONS(2508),
    [anon_sym_AT] = ACTIONS(2506),
    [anon_sym_DQUOTE] = ACTIONS(2508),
    [anon_sym_L_DQUOTE] = ACTIONS(2508),
    [anon_sym_u_DQUOTE] = ACTIONS(2508),
    [anon_sym_U_DQUOTE] = ACTIONS(2508),
    [anon_sym_u8_DQUOTE] = ACTIONS(2508),
    [sym_true] = ACTIONS(2506),
    [sym_false] = ACTIONS(2506),
    [anon_sym_NULL] = ACTIONS(2506),
    [anon_sym_nullptr] = ACTIONS(2506),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2506),
    [anon_sym___typeof] = ACTIONS(2506),
    [anon_sym_typeof] = ACTIONS(2506),
    [anon_sym_ATimport] = ACTIONS(2508),
    [aux_sym_preproc_undef_token1] = ACTIONS(2506),
    [anon_sym_POUND] = ACTIONS(2506),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2506),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2506),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2506),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_API_AVAILABLE] = ACTIONS(2506),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_API_DEPRECATED] = ACTIONS(2506),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym___deprecated_msg] = ACTIONS(2506),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2506),
    [anon_sym_ATprotocol] = ACTIONS(2508),
    [anon_sym_ATinterface] = ACTIONS(2508),
    [anon_sym_ATimplementation] = ACTIONS(2508),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2508),
    [anon_sym_ATtry] = ACTIONS(2508),
    [anon_sym___try] = ACTIONS(2506),
    [anon_sym_ATthrow] = ACTIONS(2508),
    [anon_sym_ATselector] = ACTIONS(2508),
    [anon_sym_ATavailable] = ACTIONS(2508),
    [anon_sym___builtin_available] = ACTIONS(2506),
    [anon_sym_va_arg] = ACTIONS(2506),
    [anon_sym_ATencode] = ACTIONS(2508),
    [anon_sym_ATsynchronized] = ACTIONS(2508),
    [anon_sym_BOOL] = ACTIONS(2506),
    [anon_sym_IMP] = ACTIONS(2506),
    [anon_sym_SEL] = ACTIONS(2506),
    [anon_sym_Class] = ACTIONS(2506),
    [anon_sym_id] = ACTIONS(2506),
  },
  [STATE(742)] = {
    [ts_builtin_sym_end] = ACTIONS(2426),
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(743)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_RBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(744)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_RBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(745)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_RBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(746)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_RBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(747)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_RBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(748)] = {
    [ts_builtin_sym_end] = ACTIONS(2548),
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(749)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_RBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(750)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_RBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(751)] = {
    [sym_identifier] = ACTIONS(2550),
    [aux_sym_preproc_include_token1] = ACTIONS(2550),
    [aux_sym_preproc_include_token2] = ACTIONS(2550),
    [aux_sym_preproc_def_token1] = ACTIONS(2550),
    [aux_sym_preproc_if_token1] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2550),
    [sym_preproc_directive] = ACTIONS(2550),
    [anon_sym_LPAREN2] = ACTIONS(2552),
    [anon_sym_BANG] = ACTIONS(2552),
    [anon_sym_TILDE] = ACTIONS(2552),
    [anon_sym_DASH] = ACTIONS(2550),
    [anon_sym_PLUS] = ACTIONS(2550),
    [anon_sym_STAR] = ACTIONS(2552),
    [anon_sym_CARET] = ACTIONS(2552),
    [anon_sym_AMP] = ACTIONS(2552),
    [anon_sym_SEMI] = ACTIONS(2552),
    [anon_sym___extension__] = ACTIONS(2550),
    [anon_sym_typedef] = ACTIONS(2550),
    [anon_sym_extern] = ACTIONS(2550),
    [anon_sym___attribute__] = ACTIONS(2550),
    [anon_sym___attribute] = ACTIONS(2550),
    [anon_sym_noreturn] = ACTIONS(2550),
    [anon_sym_LBRACK] = ACTIONS(2552),
    [anon_sym___declspec] = ACTIONS(2550),
    [anon_sym___cdecl] = ACTIONS(2550),
    [anon_sym___clrcall] = ACTIONS(2550),
    [anon_sym___stdcall] = ACTIONS(2550),
    [anon_sym___fastcall] = ACTIONS(2550),
    [anon_sym___thiscall] = ACTIONS(2550),
    [anon_sym___vectorcall] = ACTIONS(2550),
    [anon_sym_LBRACE] = ACTIONS(2552),
    [anon_sym_RBRACE] = ACTIONS(2552),
    [anon_sym_signed] = ACTIONS(2550),
    [anon_sym_unsigned] = ACTIONS(2550),
    [anon_sym_long] = ACTIONS(2550),
    [anon_sym_short] = ACTIONS(2550),
    [anon_sym_static] = ACTIONS(2550),
    [anon_sym_ATautoreleasepool] = ACTIONS(2552),
    [anon_sym_auto] = ACTIONS(2550),
    [anon_sym_register] = ACTIONS(2550),
    [anon_sym_inline] = ACTIONS(2550),
    [anon_sym___inline] = ACTIONS(2550),
    [anon_sym___inline__] = ACTIONS(2550),
    [anon_sym___forceinline] = ACTIONS(2550),
    [anon_sym_thread_local] = ACTIONS(2550),
    [anon_sym___thread] = ACTIONS(2550),
    [anon_sym_CG_EXTERN] = ACTIONS(2550),
    [anon_sym_CG_INLINE] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2550),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2550),
    [anon_sym_IBOutlet] = ACTIONS(2550),
    [anon_sym_IBInspectable] = ACTIONS(2550),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2550),
    [anon_sym_NS_INLINE] = ACTIONS(2550),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2550),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2550),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2550),
    [anon_sym_const] = ACTIONS(2550),
    [anon_sym_constexpr] = ACTIONS(2550),
    [anon_sym_volatile] = ACTIONS(2550),
    [anon_sym_restrict] = ACTIONS(2550),
    [anon_sym___restrict__] = ACTIONS(2550),
    [anon_sym__Atomic] = ACTIONS(2550),
    [anon_sym__Noreturn] = ACTIONS(2550),
    [anon_sym__Nonnull] = ACTIONS(2550),
    [anon_sym_nullable] = ACTIONS(2550),
    [anon_sym__Complex] = ACTIONS(2550),
    [anon_sym__Nullable] = ACTIONS(2550),
    [anon_sym__Nullable_result] = ACTIONS(2550),
    [anon_sym__Null_unspecified] = ACTIONS(2550),
    [anon_sym___autoreleasing] = ACTIONS(2550),
    [anon_sym___block] = ACTIONS(2550),
    [anon_sym___bridge] = ACTIONS(2550),
    [anon_sym___bridge_retained] = ACTIONS(2550),
    [anon_sym___bridge_transfer] = ACTIONS(2550),
    [anon_sym___complex] = ACTIONS(2550),
    [anon_sym___const] = ACTIONS(2550),
    [anon_sym___imag] = ACTIONS(2550),
    [anon_sym___kindof] = ACTIONS(2550),
    [anon_sym___nonnull] = ACTIONS(2550),
    [anon_sym___nullable] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2550),
    [anon_sym___real] = ACTIONS(2550),
    [anon_sym___strong] = ACTIONS(2550),
    [anon_sym___unsafe_unretained] = ACTIONS(2550),
    [anon_sym___unused] = ACTIONS(2550),
    [anon_sym___weak] = ACTIONS(2550),
    [anon_sym_alignas] = ACTIONS(2550),
    [anon_sym__Alignas] = ACTIONS(2550),
    [sym_primitive_type] = ACTIONS(2550),
    [anon_sym_enum] = ACTIONS(2550),
    [anon_sym_struct] = ACTIONS(2550),
    [anon_sym_union] = ACTIONS(2550),
    [anon_sym_if] = ACTIONS(2550),
    [anon_sym_else] = ACTIONS(2550),
    [anon_sym_switch] = ACTIONS(2550),
    [anon_sym_case] = ACTIONS(2550),
    [anon_sym_default] = ACTIONS(2550),
    [anon_sym_while] = ACTIONS(2550),
    [anon_sym_do] = ACTIONS(2550),
    [anon_sym_for] = ACTIONS(2550),
    [anon_sym_in] = ACTIONS(2550),
    [anon_sym_return] = ACTIONS(2550),
    [anon_sym_break] = ACTIONS(2550),
    [anon_sym_continue] = ACTIONS(2550),
    [anon_sym_goto] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2552),
    [anon_sym_PLUS_PLUS] = ACTIONS(2552),
    [anon_sym_sizeof] = ACTIONS(2550),
    [anon_sym___alignof__] = ACTIONS(2550),
    [anon_sym___alignof] = ACTIONS(2550),
    [anon_sym__alignof] = ACTIONS(2550),
    [anon_sym_alignof] = ACTIONS(2550),
    [anon_sym__Alignof] = ACTIONS(2550),
    [anon_sym_offsetof] = ACTIONS(2550),
    [anon_sym__Generic] = ACTIONS(2550),
    [anon_sym_asm] = ACTIONS(2550),
    [anon_sym___asm__] = ACTIONS(2550),
    [anon_sym___asm] = ACTIONS(2550),
    [sym_number_literal] = ACTIONS(2552),
    [anon_sym_L_SQUOTE] = ACTIONS(2552),
    [anon_sym_u_SQUOTE] = ACTIONS(2552),
    [anon_sym_U_SQUOTE] = ACTIONS(2552),
    [anon_sym_u8_SQUOTE] = ACTIONS(2552),
    [anon_sym_SQUOTE] = ACTIONS(2552),
    [anon_sym_AT] = ACTIONS(2550),
    [anon_sym_DQUOTE] = ACTIONS(2552),
    [anon_sym_L_DQUOTE] = ACTIONS(2552),
    [anon_sym_u_DQUOTE] = ACTIONS(2552),
    [anon_sym_U_DQUOTE] = ACTIONS(2552),
    [anon_sym_u8_DQUOTE] = ACTIONS(2552),
    [sym_true] = ACTIONS(2550),
    [sym_false] = ACTIONS(2550),
    [anon_sym_NULL] = ACTIONS(2550),
    [anon_sym_nullptr] = ACTIONS(2550),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2550),
    [anon_sym___typeof] = ACTIONS(2550),
    [anon_sym_typeof] = ACTIONS(2550),
    [anon_sym_ATimport] = ACTIONS(2552),
    [aux_sym_preproc_undef_token1] = ACTIONS(2550),
    [anon_sym_POUND] = ACTIONS(2550),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2550),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2550),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2550),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_API_AVAILABLE] = ACTIONS(2550),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_API_DEPRECATED] = ACTIONS(2550),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym___deprecated_msg] = ACTIONS(2550),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2550),
    [anon_sym_ATprotocol] = ACTIONS(2552),
    [anon_sym_ATinterface] = ACTIONS(2552),
    [anon_sym_ATimplementation] = ACTIONS(2552),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2552),
    [anon_sym_ATtry] = ACTIONS(2552),
    [anon_sym___try] = ACTIONS(2550),
    [anon_sym_ATthrow] = ACTIONS(2552),
    [anon_sym_ATselector] = ACTIONS(2552),
    [anon_sym_ATavailable] = ACTIONS(2552),
    [anon_sym___builtin_available] = ACTIONS(2550),
    [anon_sym_va_arg] = ACTIONS(2550),
    [anon_sym_ATencode] = ACTIONS(2552),
    [anon_sym_ATsynchronized] = ACTIONS(2552),
    [anon_sym_BOOL] = ACTIONS(2550),
    [anon_sym_IMP] = ACTIONS(2550),
    [anon_sym_SEL] = ACTIONS(2550),
    [anon_sym_Class] = ACTIONS(2550),
    [anon_sym_id] = ACTIONS(2550),
  },
  [STATE(752)] = {
    [ts_builtin_sym_end] = ACTIONS(2348),
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATcatch] = ACTIONS(2348),
    [anon_sym___catch] = ACTIONS(2346),
    [anon_sym_ATfinally] = ACTIONS(2348),
    [anon_sym___finally] = ACTIONS(2346),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(753)] = {
    [ts_builtin_sym_end] = ACTIONS(2352),
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATcatch] = ACTIONS(2352),
    [anon_sym___catch] = ACTIONS(2350),
    [anon_sym_ATfinally] = ACTIONS(2352),
    [anon_sym___finally] = ACTIONS(2350),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(754)] = {
    [ts_builtin_sym_end] = ACTIONS(2362),
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATcatch] = ACTIONS(2362),
    [anon_sym___catch] = ACTIONS(2360),
    [anon_sym_ATfinally] = ACTIONS(2362),
    [anon_sym___finally] = ACTIONS(2360),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(755)] = {
    [ts_builtin_sym_end] = ACTIONS(2516),
    [sym_identifier] = ACTIONS(2514),
    [aux_sym_preproc_include_token1] = ACTIONS(2514),
    [aux_sym_preproc_include_token2] = ACTIONS(2514),
    [aux_sym_preproc_def_token1] = ACTIONS(2514),
    [aux_sym_preproc_if_token1] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2514),
    [sym_preproc_directive] = ACTIONS(2514),
    [anon_sym_LPAREN2] = ACTIONS(2516),
    [anon_sym_BANG] = ACTIONS(2516),
    [anon_sym_TILDE] = ACTIONS(2516),
    [anon_sym_DASH] = ACTIONS(2514),
    [anon_sym_PLUS] = ACTIONS(2514),
    [anon_sym_STAR] = ACTIONS(2516),
    [anon_sym_CARET] = ACTIONS(2516),
    [anon_sym_AMP] = ACTIONS(2516),
    [anon_sym_SEMI] = ACTIONS(2516),
    [anon_sym___extension__] = ACTIONS(2514),
    [anon_sym_typedef] = ACTIONS(2514),
    [anon_sym_extern] = ACTIONS(2514),
    [anon_sym___attribute__] = ACTIONS(2514),
    [anon_sym___attribute] = ACTIONS(2514),
    [anon_sym_noreturn] = ACTIONS(2514),
    [anon_sym_LBRACK] = ACTIONS(2516),
    [anon_sym___declspec] = ACTIONS(2514),
    [anon_sym___cdecl] = ACTIONS(2514),
    [anon_sym___clrcall] = ACTIONS(2514),
    [anon_sym___stdcall] = ACTIONS(2514),
    [anon_sym___fastcall] = ACTIONS(2514),
    [anon_sym___thiscall] = ACTIONS(2514),
    [anon_sym___vectorcall] = ACTIONS(2514),
    [anon_sym_LBRACE] = ACTIONS(2516),
    [anon_sym_signed] = ACTIONS(2514),
    [anon_sym_unsigned] = ACTIONS(2514),
    [anon_sym_long] = ACTIONS(2514),
    [anon_sym_short] = ACTIONS(2514),
    [anon_sym_static] = ACTIONS(2514),
    [anon_sym_ATautoreleasepool] = ACTIONS(2516),
    [anon_sym_auto] = ACTIONS(2514),
    [anon_sym_register] = ACTIONS(2514),
    [anon_sym_inline] = ACTIONS(2514),
    [anon_sym___inline] = ACTIONS(2514),
    [anon_sym___inline__] = ACTIONS(2514),
    [anon_sym___forceinline] = ACTIONS(2514),
    [anon_sym_thread_local] = ACTIONS(2514),
    [anon_sym___thread] = ACTIONS(2514),
    [anon_sym_CG_EXTERN] = ACTIONS(2514),
    [anon_sym_CG_INLINE] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2514),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2514),
    [anon_sym_IBOutlet] = ACTIONS(2514),
    [anon_sym_IBInspectable] = ACTIONS(2514),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2514),
    [anon_sym_NS_INLINE] = ACTIONS(2514),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2514),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2514),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2514),
    [anon_sym_const] = ACTIONS(2514),
    [anon_sym_constexpr] = ACTIONS(2514),
    [anon_sym_volatile] = ACTIONS(2514),
    [anon_sym_restrict] = ACTIONS(2514),
    [anon_sym___restrict__] = ACTIONS(2514),
    [anon_sym__Atomic] = ACTIONS(2514),
    [anon_sym__Noreturn] = ACTIONS(2514),
    [anon_sym__Nonnull] = ACTIONS(2514),
    [anon_sym_nullable] = ACTIONS(2514),
    [anon_sym__Complex] = ACTIONS(2514),
    [anon_sym__Nullable] = ACTIONS(2514),
    [anon_sym__Nullable_result] = ACTIONS(2514),
    [anon_sym__Null_unspecified] = ACTIONS(2514),
    [anon_sym___autoreleasing] = ACTIONS(2514),
    [anon_sym___block] = ACTIONS(2514),
    [anon_sym___bridge] = ACTIONS(2514),
    [anon_sym___bridge_retained] = ACTIONS(2514),
    [anon_sym___bridge_transfer] = ACTIONS(2514),
    [anon_sym___complex] = ACTIONS(2514),
    [anon_sym___const] = ACTIONS(2514),
    [anon_sym___imag] = ACTIONS(2514),
    [anon_sym___kindof] = ACTIONS(2514),
    [anon_sym___nonnull] = ACTIONS(2514),
    [anon_sym___nullable] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2514),
    [anon_sym___real] = ACTIONS(2514),
    [anon_sym___strong] = ACTIONS(2514),
    [anon_sym___unsafe_unretained] = ACTIONS(2514),
    [anon_sym___unused] = ACTIONS(2514),
    [anon_sym___weak] = ACTIONS(2514),
    [anon_sym_alignas] = ACTIONS(2514),
    [anon_sym__Alignas] = ACTIONS(2514),
    [sym_primitive_type] = ACTIONS(2514),
    [anon_sym_enum] = ACTIONS(2514),
    [anon_sym_struct] = ACTIONS(2514),
    [anon_sym_union] = ACTIONS(2514),
    [anon_sym_if] = ACTIONS(2514),
    [anon_sym_else] = ACTIONS(2514),
    [anon_sym_switch] = ACTIONS(2514),
    [anon_sym_case] = ACTIONS(2514),
    [anon_sym_default] = ACTIONS(2514),
    [anon_sym_while] = ACTIONS(2514),
    [anon_sym_do] = ACTIONS(2514),
    [anon_sym_for] = ACTIONS(2514),
    [anon_sym_in] = ACTIONS(2514),
    [anon_sym_return] = ACTIONS(2514),
    [anon_sym_break] = ACTIONS(2514),
    [anon_sym_continue] = ACTIONS(2514),
    [anon_sym_goto] = ACTIONS(2514),
    [anon_sym_DASH_DASH] = ACTIONS(2516),
    [anon_sym_PLUS_PLUS] = ACTIONS(2516),
    [anon_sym_sizeof] = ACTIONS(2514),
    [anon_sym___alignof__] = ACTIONS(2514),
    [anon_sym___alignof] = ACTIONS(2514),
    [anon_sym__alignof] = ACTIONS(2514),
    [anon_sym_alignof] = ACTIONS(2514),
    [anon_sym__Alignof] = ACTIONS(2514),
    [anon_sym_offsetof] = ACTIONS(2514),
    [anon_sym__Generic] = ACTIONS(2514),
    [anon_sym_asm] = ACTIONS(2514),
    [anon_sym___asm__] = ACTIONS(2514),
    [anon_sym___asm] = ACTIONS(2514),
    [sym_number_literal] = ACTIONS(2516),
    [anon_sym_L_SQUOTE] = ACTIONS(2516),
    [anon_sym_u_SQUOTE] = ACTIONS(2516),
    [anon_sym_U_SQUOTE] = ACTIONS(2516),
    [anon_sym_u8_SQUOTE] = ACTIONS(2516),
    [anon_sym_SQUOTE] = ACTIONS(2516),
    [anon_sym_AT] = ACTIONS(2514),
    [anon_sym_DQUOTE] = ACTIONS(2516),
    [anon_sym_L_DQUOTE] = ACTIONS(2516),
    [anon_sym_u_DQUOTE] = ACTIONS(2516),
    [anon_sym_U_DQUOTE] = ACTIONS(2516),
    [anon_sym_u8_DQUOTE] = ACTIONS(2516),
    [sym_true] = ACTIONS(2514),
    [sym_false] = ACTIONS(2514),
    [anon_sym_NULL] = ACTIONS(2514),
    [anon_sym_nullptr] = ACTIONS(2514),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2514),
    [anon_sym___typeof] = ACTIONS(2514),
    [anon_sym_typeof] = ACTIONS(2514),
    [anon_sym_ATimport] = ACTIONS(2516),
    [aux_sym_preproc_undef_token1] = ACTIONS(2514),
    [anon_sym_POUND] = ACTIONS(2514),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2514),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2514),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2514),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_API_AVAILABLE] = ACTIONS(2514),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_API_DEPRECATED] = ACTIONS(2514),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym___deprecated_msg] = ACTIONS(2514),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2514),
    [anon_sym_ATprotocol] = ACTIONS(2516),
    [anon_sym_ATinterface] = ACTIONS(2516),
    [anon_sym_ATimplementation] = ACTIONS(2516),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2516),
    [anon_sym_ATtry] = ACTIONS(2516),
    [anon_sym___try] = ACTIONS(2514),
    [anon_sym_ATthrow] = ACTIONS(2516),
    [anon_sym_ATselector] = ACTIONS(2516),
    [anon_sym_ATavailable] = ACTIONS(2516),
    [anon_sym___builtin_available] = ACTIONS(2514),
    [anon_sym_va_arg] = ACTIONS(2514),
    [anon_sym_ATencode] = ACTIONS(2516),
    [anon_sym_ATsynchronized] = ACTIONS(2516),
    [anon_sym_BOOL] = ACTIONS(2514),
    [anon_sym_IMP] = ACTIONS(2514),
    [anon_sym_SEL] = ACTIONS(2514),
    [anon_sym_Class] = ACTIONS(2514),
    [anon_sym_id] = ACTIONS(2514),
  },
  [STATE(756)] = {
    [sym_identifier] = ACTIONS(2502),
    [aux_sym_preproc_include_token1] = ACTIONS(2502),
    [aux_sym_preproc_include_token2] = ACTIONS(2502),
    [aux_sym_preproc_def_token1] = ACTIONS(2502),
    [aux_sym_preproc_if_token1] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2502),
    [sym_preproc_directive] = ACTIONS(2502),
    [anon_sym_LPAREN2] = ACTIONS(2504),
    [anon_sym_BANG] = ACTIONS(2504),
    [anon_sym_TILDE] = ACTIONS(2504),
    [anon_sym_DASH] = ACTIONS(2502),
    [anon_sym_PLUS] = ACTIONS(2502),
    [anon_sym_STAR] = ACTIONS(2504),
    [anon_sym_CARET] = ACTIONS(2504),
    [anon_sym_AMP] = ACTIONS(2504),
    [anon_sym_SEMI] = ACTIONS(2504),
    [anon_sym___extension__] = ACTIONS(2502),
    [anon_sym_typedef] = ACTIONS(2502),
    [anon_sym_extern] = ACTIONS(2502),
    [anon_sym___attribute__] = ACTIONS(2502),
    [anon_sym___attribute] = ACTIONS(2502),
    [anon_sym_noreturn] = ACTIONS(2502),
    [anon_sym_LBRACK] = ACTIONS(2504),
    [anon_sym___declspec] = ACTIONS(2502),
    [anon_sym___cdecl] = ACTIONS(2502),
    [anon_sym___clrcall] = ACTIONS(2502),
    [anon_sym___stdcall] = ACTIONS(2502),
    [anon_sym___fastcall] = ACTIONS(2502),
    [anon_sym___thiscall] = ACTIONS(2502),
    [anon_sym___vectorcall] = ACTIONS(2502),
    [anon_sym_LBRACE] = ACTIONS(2504),
    [anon_sym_RBRACE] = ACTIONS(2504),
    [anon_sym_signed] = ACTIONS(2502),
    [anon_sym_unsigned] = ACTIONS(2502),
    [anon_sym_long] = ACTIONS(2502),
    [anon_sym_short] = ACTIONS(2502),
    [anon_sym_static] = ACTIONS(2502),
    [anon_sym_ATautoreleasepool] = ACTIONS(2504),
    [anon_sym_auto] = ACTIONS(2502),
    [anon_sym_register] = ACTIONS(2502),
    [anon_sym_inline] = ACTIONS(2502),
    [anon_sym___inline] = ACTIONS(2502),
    [anon_sym___inline__] = ACTIONS(2502),
    [anon_sym___forceinline] = ACTIONS(2502),
    [anon_sym_thread_local] = ACTIONS(2502),
    [anon_sym___thread] = ACTIONS(2502),
    [anon_sym_CG_EXTERN] = ACTIONS(2502),
    [anon_sym_CG_INLINE] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2502),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2502),
    [anon_sym_IBOutlet] = ACTIONS(2502),
    [anon_sym_IBInspectable] = ACTIONS(2502),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2502),
    [anon_sym_NS_INLINE] = ACTIONS(2502),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2502),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2502),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2502),
    [anon_sym_const] = ACTIONS(2502),
    [anon_sym_constexpr] = ACTIONS(2502),
    [anon_sym_volatile] = ACTIONS(2502),
    [anon_sym_restrict] = ACTIONS(2502),
    [anon_sym___restrict__] = ACTIONS(2502),
    [anon_sym__Atomic] = ACTIONS(2502),
    [anon_sym__Noreturn] = ACTIONS(2502),
    [anon_sym__Nonnull] = ACTIONS(2502),
    [anon_sym_nullable] = ACTIONS(2502),
    [anon_sym__Complex] = ACTIONS(2502),
    [anon_sym__Nullable] = ACTIONS(2502),
    [anon_sym__Nullable_result] = ACTIONS(2502),
    [anon_sym__Null_unspecified] = ACTIONS(2502),
    [anon_sym___autoreleasing] = ACTIONS(2502),
    [anon_sym___block] = ACTIONS(2502),
    [anon_sym___bridge] = ACTIONS(2502),
    [anon_sym___bridge_retained] = ACTIONS(2502),
    [anon_sym___bridge_transfer] = ACTIONS(2502),
    [anon_sym___complex] = ACTIONS(2502),
    [anon_sym___const] = ACTIONS(2502),
    [anon_sym___imag] = ACTIONS(2502),
    [anon_sym___kindof] = ACTIONS(2502),
    [anon_sym___nonnull] = ACTIONS(2502),
    [anon_sym___nullable] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2502),
    [anon_sym___real] = ACTIONS(2502),
    [anon_sym___strong] = ACTIONS(2502),
    [anon_sym___unsafe_unretained] = ACTIONS(2502),
    [anon_sym___unused] = ACTIONS(2502),
    [anon_sym___weak] = ACTIONS(2502),
    [anon_sym_alignas] = ACTIONS(2502),
    [anon_sym__Alignas] = ACTIONS(2502),
    [sym_primitive_type] = ACTIONS(2502),
    [anon_sym_enum] = ACTIONS(2502),
    [anon_sym_struct] = ACTIONS(2502),
    [anon_sym_union] = ACTIONS(2502),
    [anon_sym_if] = ACTIONS(2502),
    [anon_sym_else] = ACTIONS(2502),
    [anon_sym_switch] = ACTIONS(2502),
    [anon_sym_case] = ACTIONS(2502),
    [anon_sym_default] = ACTIONS(2502),
    [anon_sym_while] = ACTIONS(2502),
    [anon_sym_do] = ACTIONS(2502),
    [anon_sym_for] = ACTIONS(2502),
    [anon_sym_in] = ACTIONS(2502),
    [anon_sym_return] = ACTIONS(2502),
    [anon_sym_break] = ACTIONS(2502),
    [anon_sym_continue] = ACTIONS(2502),
    [anon_sym_goto] = ACTIONS(2502),
    [anon_sym_DASH_DASH] = ACTIONS(2504),
    [anon_sym_PLUS_PLUS] = ACTIONS(2504),
    [anon_sym_sizeof] = ACTIONS(2502),
    [anon_sym___alignof__] = ACTIONS(2502),
    [anon_sym___alignof] = ACTIONS(2502),
    [anon_sym__alignof] = ACTIONS(2502),
    [anon_sym_alignof] = ACTIONS(2502),
    [anon_sym__Alignof] = ACTIONS(2502),
    [anon_sym_offsetof] = ACTIONS(2502),
    [anon_sym__Generic] = ACTIONS(2502),
    [anon_sym_asm] = ACTIONS(2502),
    [anon_sym___asm__] = ACTIONS(2502),
    [anon_sym___asm] = ACTIONS(2502),
    [sym_number_literal] = ACTIONS(2504),
    [anon_sym_L_SQUOTE] = ACTIONS(2504),
    [anon_sym_u_SQUOTE] = ACTIONS(2504),
    [anon_sym_U_SQUOTE] = ACTIONS(2504),
    [anon_sym_u8_SQUOTE] = ACTIONS(2504),
    [anon_sym_SQUOTE] = ACTIONS(2504),
    [anon_sym_AT] = ACTIONS(2502),
    [anon_sym_DQUOTE] = ACTIONS(2504),
    [anon_sym_L_DQUOTE] = ACTIONS(2504),
    [anon_sym_u_DQUOTE] = ACTIONS(2504),
    [anon_sym_U_DQUOTE] = ACTIONS(2504),
    [anon_sym_u8_DQUOTE] = ACTIONS(2504),
    [sym_true] = ACTIONS(2502),
    [sym_false] = ACTIONS(2502),
    [anon_sym_NULL] = ACTIONS(2502),
    [anon_sym_nullptr] = ACTIONS(2502),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2502),
    [anon_sym___typeof] = ACTIONS(2502),
    [anon_sym_typeof] = ACTIONS(2502),
    [anon_sym_ATimport] = ACTIONS(2504),
    [aux_sym_preproc_undef_token1] = ACTIONS(2502),
    [anon_sym_POUND] = ACTIONS(2502),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2502),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2502),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2502),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_API_AVAILABLE] = ACTIONS(2502),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_API_DEPRECATED] = ACTIONS(2502),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym___deprecated_msg] = ACTIONS(2502),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2502),
    [anon_sym_ATprotocol] = ACTIONS(2504),
    [anon_sym_ATinterface] = ACTIONS(2504),
    [anon_sym_ATimplementation] = ACTIONS(2504),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2504),
    [anon_sym_ATtry] = ACTIONS(2504),
    [anon_sym___try] = ACTIONS(2502),
    [anon_sym_ATthrow] = ACTIONS(2504),
    [anon_sym_ATselector] = ACTIONS(2504),
    [anon_sym_ATavailable] = ACTIONS(2504),
    [anon_sym___builtin_available] = ACTIONS(2502),
    [anon_sym_va_arg] = ACTIONS(2502),
    [anon_sym_ATencode] = ACTIONS(2504),
    [anon_sym_ATsynchronized] = ACTIONS(2504),
    [anon_sym_BOOL] = ACTIONS(2502),
    [anon_sym_IMP] = ACTIONS(2502),
    [anon_sym_SEL] = ACTIONS(2502),
    [anon_sym_Class] = ACTIONS(2502),
    [anon_sym_id] = ACTIONS(2502),
  },
  [STATE(757)] = {
    [ts_builtin_sym_end] = ACTIONS(2520),
    [sym_identifier] = ACTIONS(2518),
    [aux_sym_preproc_include_token1] = ACTIONS(2518),
    [aux_sym_preproc_include_token2] = ACTIONS(2518),
    [aux_sym_preproc_def_token1] = ACTIONS(2518),
    [aux_sym_preproc_if_token1] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2518),
    [sym_preproc_directive] = ACTIONS(2518),
    [anon_sym_LPAREN2] = ACTIONS(2520),
    [anon_sym_BANG] = ACTIONS(2520),
    [anon_sym_TILDE] = ACTIONS(2520),
    [anon_sym_DASH] = ACTIONS(2518),
    [anon_sym_PLUS] = ACTIONS(2518),
    [anon_sym_STAR] = ACTIONS(2520),
    [anon_sym_CARET] = ACTIONS(2520),
    [anon_sym_AMP] = ACTIONS(2520),
    [anon_sym_SEMI] = ACTIONS(2520),
    [anon_sym___extension__] = ACTIONS(2518),
    [anon_sym_typedef] = ACTIONS(2518),
    [anon_sym_extern] = ACTIONS(2518),
    [anon_sym___attribute__] = ACTIONS(2518),
    [anon_sym___attribute] = ACTIONS(2518),
    [anon_sym_noreturn] = ACTIONS(2518),
    [anon_sym_LBRACK] = ACTIONS(2520),
    [anon_sym___declspec] = ACTIONS(2518),
    [anon_sym___cdecl] = ACTIONS(2518),
    [anon_sym___clrcall] = ACTIONS(2518),
    [anon_sym___stdcall] = ACTIONS(2518),
    [anon_sym___fastcall] = ACTIONS(2518),
    [anon_sym___thiscall] = ACTIONS(2518),
    [anon_sym___vectorcall] = ACTIONS(2518),
    [anon_sym_LBRACE] = ACTIONS(2520),
    [anon_sym_signed] = ACTIONS(2518),
    [anon_sym_unsigned] = ACTIONS(2518),
    [anon_sym_long] = ACTIONS(2518),
    [anon_sym_short] = ACTIONS(2518),
    [anon_sym_static] = ACTIONS(2518),
    [anon_sym_ATautoreleasepool] = ACTIONS(2520),
    [anon_sym_auto] = ACTIONS(2518),
    [anon_sym_register] = ACTIONS(2518),
    [anon_sym_inline] = ACTIONS(2518),
    [anon_sym___inline] = ACTIONS(2518),
    [anon_sym___inline__] = ACTIONS(2518),
    [anon_sym___forceinline] = ACTIONS(2518),
    [anon_sym_thread_local] = ACTIONS(2518),
    [anon_sym___thread] = ACTIONS(2518),
    [anon_sym_CG_EXTERN] = ACTIONS(2518),
    [anon_sym_CG_INLINE] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2518),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2518),
    [anon_sym_IBOutlet] = ACTIONS(2518),
    [anon_sym_IBInspectable] = ACTIONS(2518),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2518),
    [anon_sym_NS_INLINE] = ACTIONS(2518),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2518),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2518),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2518),
    [anon_sym_const] = ACTIONS(2518),
    [anon_sym_constexpr] = ACTIONS(2518),
    [anon_sym_volatile] = ACTIONS(2518),
    [anon_sym_restrict] = ACTIONS(2518),
    [anon_sym___restrict__] = ACTIONS(2518),
    [anon_sym__Atomic] = ACTIONS(2518),
    [anon_sym__Noreturn] = ACTIONS(2518),
    [anon_sym__Nonnull] = ACTIONS(2518),
    [anon_sym_nullable] = ACTIONS(2518),
    [anon_sym__Complex] = ACTIONS(2518),
    [anon_sym__Nullable] = ACTIONS(2518),
    [anon_sym__Nullable_result] = ACTIONS(2518),
    [anon_sym__Null_unspecified] = ACTIONS(2518),
    [anon_sym___autoreleasing] = ACTIONS(2518),
    [anon_sym___block] = ACTIONS(2518),
    [anon_sym___bridge] = ACTIONS(2518),
    [anon_sym___bridge_retained] = ACTIONS(2518),
    [anon_sym___bridge_transfer] = ACTIONS(2518),
    [anon_sym___complex] = ACTIONS(2518),
    [anon_sym___const] = ACTIONS(2518),
    [anon_sym___imag] = ACTIONS(2518),
    [anon_sym___kindof] = ACTIONS(2518),
    [anon_sym___nonnull] = ACTIONS(2518),
    [anon_sym___nullable] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2518),
    [anon_sym___real] = ACTIONS(2518),
    [anon_sym___strong] = ACTIONS(2518),
    [anon_sym___unsafe_unretained] = ACTIONS(2518),
    [anon_sym___unused] = ACTIONS(2518),
    [anon_sym___weak] = ACTIONS(2518),
    [anon_sym_alignas] = ACTIONS(2518),
    [anon_sym__Alignas] = ACTIONS(2518),
    [sym_primitive_type] = ACTIONS(2518),
    [anon_sym_enum] = ACTIONS(2518),
    [anon_sym_struct] = ACTIONS(2518),
    [anon_sym_union] = ACTIONS(2518),
    [anon_sym_if] = ACTIONS(2518),
    [anon_sym_else] = ACTIONS(2518),
    [anon_sym_switch] = ACTIONS(2518),
    [anon_sym_case] = ACTIONS(2518),
    [anon_sym_default] = ACTIONS(2518),
    [anon_sym_while] = ACTIONS(2518),
    [anon_sym_do] = ACTIONS(2518),
    [anon_sym_for] = ACTIONS(2518),
    [anon_sym_in] = ACTIONS(2518),
    [anon_sym_return] = ACTIONS(2518),
    [anon_sym_break] = ACTIONS(2518),
    [anon_sym_continue] = ACTIONS(2518),
    [anon_sym_goto] = ACTIONS(2518),
    [anon_sym_DASH_DASH] = ACTIONS(2520),
    [anon_sym_PLUS_PLUS] = ACTIONS(2520),
    [anon_sym_sizeof] = ACTIONS(2518),
    [anon_sym___alignof__] = ACTIONS(2518),
    [anon_sym___alignof] = ACTIONS(2518),
    [anon_sym__alignof] = ACTIONS(2518),
    [anon_sym_alignof] = ACTIONS(2518),
    [anon_sym__Alignof] = ACTIONS(2518),
    [anon_sym_offsetof] = ACTIONS(2518),
    [anon_sym__Generic] = ACTIONS(2518),
    [anon_sym_asm] = ACTIONS(2518),
    [anon_sym___asm__] = ACTIONS(2518),
    [anon_sym___asm] = ACTIONS(2518),
    [sym_number_literal] = ACTIONS(2520),
    [anon_sym_L_SQUOTE] = ACTIONS(2520),
    [anon_sym_u_SQUOTE] = ACTIONS(2520),
    [anon_sym_U_SQUOTE] = ACTIONS(2520),
    [anon_sym_u8_SQUOTE] = ACTIONS(2520),
    [anon_sym_SQUOTE] = ACTIONS(2520),
    [anon_sym_AT] = ACTIONS(2518),
    [anon_sym_DQUOTE] = ACTIONS(2520),
    [anon_sym_L_DQUOTE] = ACTIONS(2520),
    [anon_sym_u_DQUOTE] = ACTIONS(2520),
    [anon_sym_U_DQUOTE] = ACTIONS(2520),
    [anon_sym_u8_DQUOTE] = ACTIONS(2520),
    [sym_true] = ACTIONS(2518),
    [sym_false] = ACTIONS(2518),
    [anon_sym_NULL] = ACTIONS(2518),
    [anon_sym_nullptr] = ACTIONS(2518),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2518),
    [anon_sym___typeof] = ACTIONS(2518),
    [anon_sym_typeof] = ACTIONS(2518),
    [anon_sym_ATimport] = ACTIONS(2520),
    [aux_sym_preproc_undef_token1] = ACTIONS(2518),
    [anon_sym_POUND] = ACTIONS(2518),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2518),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2518),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2518),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_API_AVAILABLE] = ACTIONS(2518),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_API_DEPRECATED] = ACTIONS(2518),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym___deprecated_msg] = ACTIONS(2518),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2518),
    [anon_sym_ATprotocol] = ACTIONS(2520),
    [anon_sym_ATinterface] = ACTIONS(2520),
    [anon_sym_ATimplementation] = ACTIONS(2520),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2520),
    [anon_sym_ATtry] = ACTIONS(2520),
    [anon_sym___try] = ACTIONS(2518),
    [anon_sym_ATthrow] = ACTIONS(2520),
    [anon_sym_ATselector] = ACTIONS(2520),
    [anon_sym_ATavailable] = ACTIONS(2520),
    [anon_sym___builtin_available] = ACTIONS(2518),
    [anon_sym_va_arg] = ACTIONS(2518),
    [anon_sym_ATencode] = ACTIONS(2520),
    [anon_sym_ATsynchronized] = ACTIONS(2520),
    [anon_sym_BOOL] = ACTIONS(2518),
    [anon_sym_IMP] = ACTIONS(2518),
    [anon_sym_SEL] = ACTIONS(2518),
    [anon_sym_Class] = ACTIONS(2518),
    [anon_sym_id] = ACTIONS(2518),
  },
  [STATE(758)] = {
    [ts_builtin_sym_end] = ACTIONS(2524),
    [sym_identifier] = ACTIONS(2522),
    [aux_sym_preproc_include_token1] = ACTIONS(2522),
    [aux_sym_preproc_include_token2] = ACTIONS(2522),
    [aux_sym_preproc_def_token1] = ACTIONS(2522),
    [aux_sym_preproc_if_token1] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2522),
    [sym_preproc_directive] = ACTIONS(2522),
    [anon_sym_LPAREN2] = ACTIONS(2524),
    [anon_sym_BANG] = ACTIONS(2524),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_DASH] = ACTIONS(2522),
    [anon_sym_PLUS] = ACTIONS(2522),
    [anon_sym_STAR] = ACTIONS(2524),
    [anon_sym_CARET] = ACTIONS(2524),
    [anon_sym_AMP] = ACTIONS(2524),
    [anon_sym_SEMI] = ACTIONS(2524),
    [anon_sym___extension__] = ACTIONS(2522),
    [anon_sym_typedef] = ACTIONS(2522),
    [anon_sym_extern] = ACTIONS(2522),
    [anon_sym___attribute__] = ACTIONS(2522),
    [anon_sym___attribute] = ACTIONS(2522),
    [anon_sym_noreturn] = ACTIONS(2522),
    [anon_sym_LBRACK] = ACTIONS(2524),
    [anon_sym___declspec] = ACTIONS(2522),
    [anon_sym___cdecl] = ACTIONS(2522),
    [anon_sym___clrcall] = ACTIONS(2522),
    [anon_sym___stdcall] = ACTIONS(2522),
    [anon_sym___fastcall] = ACTIONS(2522),
    [anon_sym___thiscall] = ACTIONS(2522),
    [anon_sym___vectorcall] = ACTIONS(2522),
    [anon_sym_LBRACE] = ACTIONS(2524),
    [anon_sym_signed] = ACTIONS(2522),
    [anon_sym_unsigned] = ACTIONS(2522),
    [anon_sym_long] = ACTIONS(2522),
    [anon_sym_short] = ACTIONS(2522),
    [anon_sym_static] = ACTIONS(2522),
    [anon_sym_ATautoreleasepool] = ACTIONS(2524),
    [anon_sym_auto] = ACTIONS(2522),
    [anon_sym_register] = ACTIONS(2522),
    [anon_sym_inline] = ACTIONS(2522),
    [anon_sym___inline] = ACTIONS(2522),
    [anon_sym___inline__] = ACTIONS(2522),
    [anon_sym___forceinline] = ACTIONS(2522),
    [anon_sym_thread_local] = ACTIONS(2522),
    [anon_sym___thread] = ACTIONS(2522),
    [anon_sym_CG_EXTERN] = ACTIONS(2522),
    [anon_sym_CG_INLINE] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2522),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2522),
    [anon_sym_IBOutlet] = ACTIONS(2522),
    [anon_sym_IBInspectable] = ACTIONS(2522),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2522),
    [anon_sym_NS_INLINE] = ACTIONS(2522),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2522),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2522),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2522),
    [anon_sym_const] = ACTIONS(2522),
    [anon_sym_constexpr] = ACTIONS(2522),
    [anon_sym_volatile] = ACTIONS(2522),
    [anon_sym_restrict] = ACTIONS(2522),
    [anon_sym___restrict__] = ACTIONS(2522),
    [anon_sym__Atomic] = ACTIONS(2522),
    [anon_sym__Noreturn] = ACTIONS(2522),
    [anon_sym__Nonnull] = ACTIONS(2522),
    [anon_sym_nullable] = ACTIONS(2522),
    [anon_sym__Complex] = ACTIONS(2522),
    [anon_sym__Nullable] = ACTIONS(2522),
    [anon_sym__Nullable_result] = ACTIONS(2522),
    [anon_sym__Null_unspecified] = ACTIONS(2522),
    [anon_sym___autoreleasing] = ACTIONS(2522),
    [anon_sym___block] = ACTIONS(2522),
    [anon_sym___bridge] = ACTIONS(2522),
    [anon_sym___bridge_retained] = ACTIONS(2522),
    [anon_sym___bridge_transfer] = ACTIONS(2522),
    [anon_sym___complex] = ACTIONS(2522),
    [anon_sym___const] = ACTIONS(2522),
    [anon_sym___imag] = ACTIONS(2522),
    [anon_sym___kindof] = ACTIONS(2522),
    [anon_sym___nonnull] = ACTIONS(2522),
    [anon_sym___nullable] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2522),
    [anon_sym___real] = ACTIONS(2522),
    [anon_sym___strong] = ACTIONS(2522),
    [anon_sym___unsafe_unretained] = ACTIONS(2522),
    [anon_sym___unused] = ACTIONS(2522),
    [anon_sym___weak] = ACTIONS(2522),
    [anon_sym_alignas] = ACTIONS(2522),
    [anon_sym__Alignas] = ACTIONS(2522),
    [sym_primitive_type] = ACTIONS(2522),
    [anon_sym_enum] = ACTIONS(2522),
    [anon_sym_struct] = ACTIONS(2522),
    [anon_sym_union] = ACTIONS(2522),
    [anon_sym_if] = ACTIONS(2522),
    [anon_sym_else] = ACTIONS(2522),
    [anon_sym_switch] = ACTIONS(2522),
    [anon_sym_case] = ACTIONS(2522),
    [anon_sym_default] = ACTIONS(2522),
    [anon_sym_while] = ACTIONS(2522),
    [anon_sym_do] = ACTIONS(2522),
    [anon_sym_for] = ACTIONS(2522),
    [anon_sym_in] = ACTIONS(2522),
    [anon_sym_return] = ACTIONS(2522),
    [anon_sym_break] = ACTIONS(2522),
    [anon_sym_continue] = ACTIONS(2522),
    [anon_sym_goto] = ACTIONS(2522),
    [anon_sym_DASH_DASH] = ACTIONS(2524),
    [anon_sym_PLUS_PLUS] = ACTIONS(2524),
    [anon_sym_sizeof] = ACTIONS(2522),
    [anon_sym___alignof__] = ACTIONS(2522),
    [anon_sym___alignof] = ACTIONS(2522),
    [anon_sym__alignof] = ACTIONS(2522),
    [anon_sym_alignof] = ACTIONS(2522),
    [anon_sym__Alignof] = ACTIONS(2522),
    [anon_sym_offsetof] = ACTIONS(2522),
    [anon_sym__Generic] = ACTIONS(2522),
    [anon_sym_asm] = ACTIONS(2522),
    [anon_sym___asm__] = ACTIONS(2522),
    [anon_sym___asm] = ACTIONS(2522),
    [sym_number_literal] = ACTIONS(2524),
    [anon_sym_L_SQUOTE] = ACTIONS(2524),
    [anon_sym_u_SQUOTE] = ACTIONS(2524),
    [anon_sym_U_SQUOTE] = ACTIONS(2524),
    [anon_sym_u8_SQUOTE] = ACTIONS(2524),
    [anon_sym_SQUOTE] = ACTIONS(2524),
    [anon_sym_AT] = ACTIONS(2522),
    [anon_sym_DQUOTE] = ACTIONS(2524),
    [anon_sym_L_DQUOTE] = ACTIONS(2524),
    [anon_sym_u_DQUOTE] = ACTIONS(2524),
    [anon_sym_U_DQUOTE] = ACTIONS(2524),
    [anon_sym_u8_DQUOTE] = ACTIONS(2524),
    [sym_true] = ACTIONS(2522),
    [sym_false] = ACTIONS(2522),
    [anon_sym_NULL] = ACTIONS(2522),
    [anon_sym_nullptr] = ACTIONS(2522),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2522),
    [anon_sym___typeof] = ACTIONS(2522),
    [anon_sym_typeof] = ACTIONS(2522),
    [anon_sym_ATimport] = ACTIONS(2524),
    [aux_sym_preproc_undef_token1] = ACTIONS(2522),
    [anon_sym_POUND] = ACTIONS(2522),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2522),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2522),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2522),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_API_AVAILABLE] = ACTIONS(2522),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_API_DEPRECATED] = ACTIONS(2522),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym___deprecated_msg] = ACTIONS(2522),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2522),
    [anon_sym_ATprotocol] = ACTIONS(2524),
    [anon_sym_ATinterface] = ACTIONS(2524),
    [anon_sym_ATimplementation] = ACTIONS(2524),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2524),
    [anon_sym_ATtry] = ACTIONS(2524),
    [anon_sym___try] = ACTIONS(2522),
    [anon_sym_ATthrow] = ACTIONS(2524),
    [anon_sym_ATselector] = ACTIONS(2524),
    [anon_sym_ATavailable] = ACTIONS(2524),
    [anon_sym___builtin_available] = ACTIONS(2522),
    [anon_sym_va_arg] = ACTIONS(2522),
    [anon_sym_ATencode] = ACTIONS(2524),
    [anon_sym_ATsynchronized] = ACTIONS(2524),
    [anon_sym_BOOL] = ACTIONS(2522),
    [anon_sym_IMP] = ACTIONS(2522),
    [anon_sym_SEL] = ACTIONS(2522),
    [anon_sym_Class] = ACTIONS(2522),
    [anon_sym_id] = ACTIONS(2522),
  },
  [STATE(759)] = {
    [ts_builtin_sym_end] = ACTIONS(2528),
    [sym_identifier] = ACTIONS(2526),
    [aux_sym_preproc_include_token1] = ACTIONS(2526),
    [aux_sym_preproc_include_token2] = ACTIONS(2526),
    [aux_sym_preproc_def_token1] = ACTIONS(2526),
    [aux_sym_preproc_if_token1] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2526),
    [sym_preproc_directive] = ACTIONS(2526),
    [anon_sym_LPAREN2] = ACTIONS(2528),
    [anon_sym_BANG] = ACTIONS(2528),
    [anon_sym_TILDE] = ACTIONS(2528),
    [anon_sym_DASH] = ACTIONS(2526),
    [anon_sym_PLUS] = ACTIONS(2526),
    [anon_sym_STAR] = ACTIONS(2528),
    [anon_sym_CARET] = ACTIONS(2528),
    [anon_sym_AMP] = ACTIONS(2528),
    [anon_sym_SEMI] = ACTIONS(2528),
    [anon_sym___extension__] = ACTIONS(2526),
    [anon_sym_typedef] = ACTIONS(2526),
    [anon_sym_extern] = ACTIONS(2526),
    [anon_sym___attribute__] = ACTIONS(2526),
    [anon_sym___attribute] = ACTIONS(2526),
    [anon_sym_noreturn] = ACTIONS(2526),
    [anon_sym_LBRACK] = ACTIONS(2528),
    [anon_sym___declspec] = ACTIONS(2526),
    [anon_sym___cdecl] = ACTIONS(2526),
    [anon_sym___clrcall] = ACTIONS(2526),
    [anon_sym___stdcall] = ACTIONS(2526),
    [anon_sym___fastcall] = ACTIONS(2526),
    [anon_sym___thiscall] = ACTIONS(2526),
    [anon_sym___vectorcall] = ACTIONS(2526),
    [anon_sym_LBRACE] = ACTIONS(2528),
    [anon_sym_signed] = ACTIONS(2526),
    [anon_sym_unsigned] = ACTIONS(2526),
    [anon_sym_long] = ACTIONS(2526),
    [anon_sym_short] = ACTIONS(2526),
    [anon_sym_static] = ACTIONS(2526),
    [anon_sym_ATautoreleasepool] = ACTIONS(2528),
    [anon_sym_auto] = ACTIONS(2526),
    [anon_sym_register] = ACTIONS(2526),
    [anon_sym_inline] = ACTIONS(2526),
    [anon_sym___inline] = ACTIONS(2526),
    [anon_sym___inline__] = ACTIONS(2526),
    [anon_sym___forceinline] = ACTIONS(2526),
    [anon_sym_thread_local] = ACTIONS(2526),
    [anon_sym___thread] = ACTIONS(2526),
    [anon_sym_CG_EXTERN] = ACTIONS(2526),
    [anon_sym_CG_INLINE] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2526),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2526),
    [anon_sym_IBOutlet] = ACTIONS(2526),
    [anon_sym_IBInspectable] = ACTIONS(2526),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2526),
    [anon_sym_NS_INLINE] = ACTIONS(2526),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2526),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2526),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2526),
    [anon_sym_const] = ACTIONS(2526),
    [anon_sym_constexpr] = ACTIONS(2526),
    [anon_sym_volatile] = ACTIONS(2526),
    [anon_sym_restrict] = ACTIONS(2526),
    [anon_sym___restrict__] = ACTIONS(2526),
    [anon_sym__Atomic] = ACTIONS(2526),
    [anon_sym__Noreturn] = ACTIONS(2526),
    [anon_sym__Nonnull] = ACTIONS(2526),
    [anon_sym_nullable] = ACTIONS(2526),
    [anon_sym__Complex] = ACTIONS(2526),
    [anon_sym__Nullable] = ACTIONS(2526),
    [anon_sym__Nullable_result] = ACTIONS(2526),
    [anon_sym__Null_unspecified] = ACTIONS(2526),
    [anon_sym___autoreleasing] = ACTIONS(2526),
    [anon_sym___block] = ACTIONS(2526),
    [anon_sym___bridge] = ACTIONS(2526),
    [anon_sym___bridge_retained] = ACTIONS(2526),
    [anon_sym___bridge_transfer] = ACTIONS(2526),
    [anon_sym___complex] = ACTIONS(2526),
    [anon_sym___const] = ACTIONS(2526),
    [anon_sym___imag] = ACTIONS(2526),
    [anon_sym___kindof] = ACTIONS(2526),
    [anon_sym___nonnull] = ACTIONS(2526),
    [anon_sym___nullable] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2526),
    [anon_sym___real] = ACTIONS(2526),
    [anon_sym___strong] = ACTIONS(2526),
    [anon_sym___unsafe_unretained] = ACTIONS(2526),
    [anon_sym___unused] = ACTIONS(2526),
    [anon_sym___weak] = ACTIONS(2526),
    [anon_sym_alignas] = ACTIONS(2526),
    [anon_sym__Alignas] = ACTIONS(2526),
    [sym_primitive_type] = ACTIONS(2526),
    [anon_sym_enum] = ACTIONS(2526),
    [anon_sym_struct] = ACTIONS(2526),
    [anon_sym_union] = ACTIONS(2526),
    [anon_sym_if] = ACTIONS(2526),
    [anon_sym_else] = ACTIONS(2526),
    [anon_sym_switch] = ACTIONS(2526),
    [anon_sym_case] = ACTIONS(2526),
    [anon_sym_default] = ACTIONS(2526),
    [anon_sym_while] = ACTIONS(2526),
    [anon_sym_do] = ACTIONS(2526),
    [anon_sym_for] = ACTIONS(2526),
    [anon_sym_in] = ACTIONS(2526),
    [anon_sym_return] = ACTIONS(2526),
    [anon_sym_break] = ACTIONS(2526),
    [anon_sym_continue] = ACTIONS(2526),
    [anon_sym_goto] = ACTIONS(2526),
    [anon_sym_DASH_DASH] = ACTIONS(2528),
    [anon_sym_PLUS_PLUS] = ACTIONS(2528),
    [anon_sym_sizeof] = ACTIONS(2526),
    [anon_sym___alignof__] = ACTIONS(2526),
    [anon_sym___alignof] = ACTIONS(2526),
    [anon_sym__alignof] = ACTIONS(2526),
    [anon_sym_alignof] = ACTIONS(2526),
    [anon_sym__Alignof] = ACTIONS(2526),
    [anon_sym_offsetof] = ACTIONS(2526),
    [anon_sym__Generic] = ACTIONS(2526),
    [anon_sym_asm] = ACTIONS(2526),
    [anon_sym___asm__] = ACTIONS(2526),
    [anon_sym___asm] = ACTIONS(2526),
    [sym_number_literal] = ACTIONS(2528),
    [anon_sym_L_SQUOTE] = ACTIONS(2528),
    [anon_sym_u_SQUOTE] = ACTIONS(2528),
    [anon_sym_U_SQUOTE] = ACTIONS(2528),
    [anon_sym_u8_SQUOTE] = ACTIONS(2528),
    [anon_sym_SQUOTE] = ACTIONS(2528),
    [anon_sym_AT] = ACTIONS(2526),
    [anon_sym_DQUOTE] = ACTIONS(2528),
    [anon_sym_L_DQUOTE] = ACTIONS(2528),
    [anon_sym_u_DQUOTE] = ACTIONS(2528),
    [anon_sym_U_DQUOTE] = ACTIONS(2528),
    [anon_sym_u8_DQUOTE] = ACTIONS(2528),
    [sym_true] = ACTIONS(2526),
    [sym_false] = ACTIONS(2526),
    [anon_sym_NULL] = ACTIONS(2526),
    [anon_sym_nullptr] = ACTIONS(2526),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2526),
    [anon_sym___typeof] = ACTIONS(2526),
    [anon_sym_typeof] = ACTIONS(2526),
    [anon_sym_ATimport] = ACTIONS(2528),
    [aux_sym_preproc_undef_token1] = ACTIONS(2526),
    [anon_sym_POUND] = ACTIONS(2526),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2526),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2526),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2526),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_API_AVAILABLE] = ACTIONS(2526),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_API_DEPRECATED] = ACTIONS(2526),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym___deprecated_msg] = ACTIONS(2526),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2526),
    [anon_sym_ATprotocol] = ACTIONS(2528),
    [anon_sym_ATinterface] = ACTIONS(2528),
    [anon_sym_ATimplementation] = ACTIONS(2528),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2528),
    [anon_sym_ATtry] = ACTIONS(2528),
    [anon_sym___try] = ACTIONS(2526),
    [anon_sym_ATthrow] = ACTIONS(2528),
    [anon_sym_ATselector] = ACTIONS(2528),
    [anon_sym_ATavailable] = ACTIONS(2528),
    [anon_sym___builtin_available] = ACTIONS(2526),
    [anon_sym_va_arg] = ACTIONS(2526),
    [anon_sym_ATencode] = ACTIONS(2528),
    [anon_sym_ATsynchronized] = ACTIONS(2528),
    [anon_sym_BOOL] = ACTIONS(2526),
    [anon_sym_IMP] = ACTIONS(2526),
    [anon_sym_SEL] = ACTIONS(2526),
    [anon_sym_Class] = ACTIONS(2526),
    [anon_sym_id] = ACTIONS(2526),
  },
  [STATE(760)] = {
    [ts_builtin_sym_end] = ACTIONS(2532),
    [sym_identifier] = ACTIONS(2530),
    [aux_sym_preproc_include_token1] = ACTIONS(2530),
    [aux_sym_preproc_include_token2] = ACTIONS(2530),
    [aux_sym_preproc_def_token1] = ACTIONS(2530),
    [aux_sym_preproc_if_token1] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2530),
    [sym_preproc_directive] = ACTIONS(2530),
    [anon_sym_LPAREN2] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2530),
    [anon_sym_PLUS] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym___extension__] = ACTIONS(2530),
    [anon_sym_typedef] = ACTIONS(2530),
    [anon_sym_extern] = ACTIONS(2530),
    [anon_sym___attribute__] = ACTIONS(2530),
    [anon_sym___attribute] = ACTIONS(2530),
    [anon_sym_noreturn] = ACTIONS(2530),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym___declspec] = ACTIONS(2530),
    [anon_sym___cdecl] = ACTIONS(2530),
    [anon_sym___clrcall] = ACTIONS(2530),
    [anon_sym___stdcall] = ACTIONS(2530),
    [anon_sym___fastcall] = ACTIONS(2530),
    [anon_sym___thiscall] = ACTIONS(2530),
    [anon_sym___vectorcall] = ACTIONS(2530),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_signed] = ACTIONS(2530),
    [anon_sym_unsigned] = ACTIONS(2530),
    [anon_sym_long] = ACTIONS(2530),
    [anon_sym_short] = ACTIONS(2530),
    [anon_sym_static] = ACTIONS(2530),
    [anon_sym_ATautoreleasepool] = ACTIONS(2532),
    [anon_sym_auto] = ACTIONS(2530),
    [anon_sym_register] = ACTIONS(2530),
    [anon_sym_inline] = ACTIONS(2530),
    [anon_sym___inline] = ACTIONS(2530),
    [anon_sym___inline__] = ACTIONS(2530),
    [anon_sym___forceinline] = ACTIONS(2530),
    [anon_sym_thread_local] = ACTIONS(2530),
    [anon_sym___thread] = ACTIONS(2530),
    [anon_sym_CG_EXTERN] = ACTIONS(2530),
    [anon_sym_CG_INLINE] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2530),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2530),
    [anon_sym_IBOutlet] = ACTIONS(2530),
    [anon_sym_IBInspectable] = ACTIONS(2530),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2530),
    [anon_sym_NS_INLINE] = ACTIONS(2530),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2530),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2530),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2530),
    [anon_sym_const] = ACTIONS(2530),
    [anon_sym_constexpr] = ACTIONS(2530),
    [anon_sym_volatile] = ACTIONS(2530),
    [anon_sym_restrict] = ACTIONS(2530),
    [anon_sym___restrict__] = ACTIONS(2530),
    [anon_sym__Atomic] = ACTIONS(2530),
    [anon_sym__Noreturn] = ACTIONS(2530),
    [anon_sym__Nonnull] = ACTIONS(2530),
    [anon_sym_nullable] = ACTIONS(2530),
    [anon_sym__Complex] = ACTIONS(2530),
    [anon_sym__Nullable] = ACTIONS(2530),
    [anon_sym__Nullable_result] = ACTIONS(2530),
    [anon_sym__Null_unspecified] = ACTIONS(2530),
    [anon_sym___autoreleasing] = ACTIONS(2530),
    [anon_sym___block] = ACTIONS(2530),
    [anon_sym___bridge] = ACTIONS(2530),
    [anon_sym___bridge_retained] = ACTIONS(2530),
    [anon_sym___bridge_transfer] = ACTIONS(2530),
    [anon_sym___complex] = ACTIONS(2530),
    [anon_sym___const] = ACTIONS(2530),
    [anon_sym___imag] = ACTIONS(2530),
    [anon_sym___kindof] = ACTIONS(2530),
    [anon_sym___nonnull] = ACTIONS(2530),
    [anon_sym___nullable] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2530),
    [anon_sym___real] = ACTIONS(2530),
    [anon_sym___strong] = ACTIONS(2530),
    [anon_sym___unsafe_unretained] = ACTIONS(2530),
    [anon_sym___unused] = ACTIONS(2530),
    [anon_sym___weak] = ACTIONS(2530),
    [anon_sym_alignas] = ACTIONS(2530),
    [anon_sym__Alignas] = ACTIONS(2530),
    [sym_primitive_type] = ACTIONS(2530),
    [anon_sym_enum] = ACTIONS(2530),
    [anon_sym_struct] = ACTIONS(2530),
    [anon_sym_union] = ACTIONS(2530),
    [anon_sym_if] = ACTIONS(2530),
    [anon_sym_else] = ACTIONS(2530),
    [anon_sym_switch] = ACTIONS(2530),
    [anon_sym_case] = ACTIONS(2530),
    [anon_sym_default] = ACTIONS(2530),
    [anon_sym_while] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2530),
    [anon_sym_for] = ACTIONS(2530),
    [anon_sym_in] = ACTIONS(2530),
    [anon_sym_return] = ACTIONS(2530),
    [anon_sym_break] = ACTIONS(2530),
    [anon_sym_continue] = ACTIONS(2530),
    [anon_sym_goto] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_sizeof] = ACTIONS(2530),
    [anon_sym___alignof__] = ACTIONS(2530),
    [anon_sym___alignof] = ACTIONS(2530),
    [anon_sym__alignof] = ACTIONS(2530),
    [anon_sym_alignof] = ACTIONS(2530),
    [anon_sym__Alignof] = ACTIONS(2530),
    [anon_sym_offsetof] = ACTIONS(2530),
    [anon_sym__Generic] = ACTIONS(2530),
    [anon_sym_asm] = ACTIONS(2530),
    [anon_sym___asm__] = ACTIONS(2530),
    [anon_sym___asm] = ACTIONS(2530),
    [sym_number_literal] = ACTIONS(2532),
    [anon_sym_L_SQUOTE] = ACTIONS(2532),
    [anon_sym_u_SQUOTE] = ACTIONS(2532),
    [anon_sym_U_SQUOTE] = ACTIONS(2532),
    [anon_sym_u8_SQUOTE] = ACTIONS(2532),
    [anon_sym_SQUOTE] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2530),
    [anon_sym_DQUOTE] = ACTIONS(2532),
    [anon_sym_L_DQUOTE] = ACTIONS(2532),
    [anon_sym_u_DQUOTE] = ACTIONS(2532),
    [anon_sym_U_DQUOTE] = ACTIONS(2532),
    [anon_sym_u8_DQUOTE] = ACTIONS(2532),
    [sym_true] = ACTIONS(2530),
    [sym_false] = ACTIONS(2530),
    [anon_sym_NULL] = ACTIONS(2530),
    [anon_sym_nullptr] = ACTIONS(2530),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2530),
    [anon_sym___typeof] = ACTIONS(2530),
    [anon_sym_typeof] = ACTIONS(2530),
    [anon_sym_ATimport] = ACTIONS(2532),
    [aux_sym_preproc_undef_token1] = ACTIONS(2530),
    [anon_sym_POUND] = ACTIONS(2530),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2530),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2530),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2530),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_API_AVAILABLE] = ACTIONS(2530),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_API_DEPRECATED] = ACTIONS(2530),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym___deprecated_msg] = ACTIONS(2530),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2530),
    [anon_sym_ATprotocol] = ACTIONS(2532),
    [anon_sym_ATinterface] = ACTIONS(2532),
    [anon_sym_ATimplementation] = ACTIONS(2532),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2532),
    [anon_sym_ATtry] = ACTIONS(2532),
    [anon_sym___try] = ACTIONS(2530),
    [anon_sym_ATthrow] = ACTIONS(2532),
    [anon_sym_ATselector] = ACTIONS(2532),
    [anon_sym_ATavailable] = ACTIONS(2532),
    [anon_sym___builtin_available] = ACTIONS(2530),
    [anon_sym_va_arg] = ACTIONS(2530),
    [anon_sym_ATencode] = ACTIONS(2532),
    [anon_sym_ATsynchronized] = ACTIONS(2532),
    [anon_sym_BOOL] = ACTIONS(2530),
    [anon_sym_IMP] = ACTIONS(2530),
    [anon_sym_SEL] = ACTIONS(2530),
    [anon_sym_Class] = ACTIONS(2530),
    [anon_sym_id] = ACTIONS(2530),
  },
  [STATE(761)] = {
    [ts_builtin_sym_end] = ACTIONS(2536),
    [sym_identifier] = ACTIONS(2534),
    [aux_sym_preproc_include_token1] = ACTIONS(2534),
    [aux_sym_preproc_include_token2] = ACTIONS(2534),
    [aux_sym_preproc_def_token1] = ACTIONS(2534),
    [aux_sym_preproc_if_token1] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2534),
    [sym_preproc_directive] = ACTIONS(2534),
    [anon_sym_LPAREN2] = ACTIONS(2536),
    [anon_sym_BANG] = ACTIONS(2536),
    [anon_sym_TILDE] = ACTIONS(2536),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2536),
    [anon_sym_CARET] = ACTIONS(2536),
    [anon_sym_AMP] = ACTIONS(2536),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym___extension__] = ACTIONS(2534),
    [anon_sym_typedef] = ACTIONS(2534),
    [anon_sym_extern] = ACTIONS(2534),
    [anon_sym___attribute__] = ACTIONS(2534),
    [anon_sym___attribute] = ACTIONS(2534),
    [anon_sym_noreturn] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2536),
    [anon_sym___declspec] = ACTIONS(2534),
    [anon_sym___cdecl] = ACTIONS(2534),
    [anon_sym___clrcall] = ACTIONS(2534),
    [anon_sym___stdcall] = ACTIONS(2534),
    [anon_sym___fastcall] = ACTIONS(2534),
    [anon_sym___thiscall] = ACTIONS(2534),
    [anon_sym___vectorcall] = ACTIONS(2534),
    [anon_sym_LBRACE] = ACTIONS(2536),
    [anon_sym_signed] = ACTIONS(2534),
    [anon_sym_unsigned] = ACTIONS(2534),
    [anon_sym_long] = ACTIONS(2534),
    [anon_sym_short] = ACTIONS(2534),
    [anon_sym_static] = ACTIONS(2534),
    [anon_sym_ATautoreleasepool] = ACTIONS(2536),
    [anon_sym_auto] = ACTIONS(2534),
    [anon_sym_register] = ACTIONS(2534),
    [anon_sym_inline] = ACTIONS(2534),
    [anon_sym___inline] = ACTIONS(2534),
    [anon_sym___inline__] = ACTIONS(2534),
    [anon_sym___forceinline] = ACTIONS(2534),
    [anon_sym_thread_local] = ACTIONS(2534),
    [anon_sym___thread] = ACTIONS(2534),
    [anon_sym_CG_EXTERN] = ACTIONS(2534),
    [anon_sym_CG_INLINE] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2534),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2534),
    [anon_sym_IBOutlet] = ACTIONS(2534),
    [anon_sym_IBInspectable] = ACTIONS(2534),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2534),
    [anon_sym_NS_INLINE] = ACTIONS(2534),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2534),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2534),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2534),
    [anon_sym_const] = ACTIONS(2534),
    [anon_sym_constexpr] = ACTIONS(2534),
    [anon_sym_volatile] = ACTIONS(2534),
    [anon_sym_restrict] = ACTIONS(2534),
    [anon_sym___restrict__] = ACTIONS(2534),
    [anon_sym__Atomic] = ACTIONS(2534),
    [anon_sym__Noreturn] = ACTIONS(2534),
    [anon_sym__Nonnull] = ACTIONS(2534),
    [anon_sym_nullable] = ACTIONS(2534),
    [anon_sym__Complex] = ACTIONS(2534),
    [anon_sym__Nullable] = ACTIONS(2534),
    [anon_sym__Nullable_result] = ACTIONS(2534),
    [anon_sym__Null_unspecified] = ACTIONS(2534),
    [anon_sym___autoreleasing] = ACTIONS(2534),
    [anon_sym___block] = ACTIONS(2534),
    [anon_sym___bridge] = ACTIONS(2534),
    [anon_sym___bridge_retained] = ACTIONS(2534),
    [anon_sym___bridge_transfer] = ACTIONS(2534),
    [anon_sym___complex] = ACTIONS(2534),
    [anon_sym___const] = ACTIONS(2534),
    [anon_sym___imag] = ACTIONS(2534),
    [anon_sym___kindof] = ACTIONS(2534),
    [anon_sym___nonnull] = ACTIONS(2534),
    [anon_sym___nullable] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2534),
    [anon_sym___real] = ACTIONS(2534),
    [anon_sym___strong] = ACTIONS(2534),
    [anon_sym___unsafe_unretained] = ACTIONS(2534),
    [anon_sym___unused] = ACTIONS(2534),
    [anon_sym___weak] = ACTIONS(2534),
    [anon_sym_alignas] = ACTIONS(2534),
    [anon_sym__Alignas] = ACTIONS(2534),
    [sym_primitive_type] = ACTIONS(2534),
    [anon_sym_enum] = ACTIONS(2534),
    [anon_sym_struct] = ACTIONS(2534),
    [anon_sym_union] = ACTIONS(2534),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_else] = ACTIONS(2534),
    [anon_sym_switch] = ACTIONS(2534),
    [anon_sym_case] = ACTIONS(2534),
    [anon_sym_default] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_do] = ACTIONS(2534),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_in] = ACTIONS(2534),
    [anon_sym_return] = ACTIONS(2534),
    [anon_sym_break] = ACTIONS(2534),
    [anon_sym_continue] = ACTIONS(2534),
    [anon_sym_goto] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_sizeof] = ACTIONS(2534),
    [anon_sym___alignof__] = ACTIONS(2534),
    [anon_sym___alignof] = ACTIONS(2534),
    [anon_sym__alignof] = ACTIONS(2534),
    [anon_sym_alignof] = ACTIONS(2534),
    [anon_sym__Alignof] = ACTIONS(2534),
    [anon_sym_offsetof] = ACTIONS(2534),
    [anon_sym__Generic] = ACTIONS(2534),
    [anon_sym_asm] = ACTIONS(2534),
    [anon_sym___asm__] = ACTIONS(2534),
    [anon_sym___asm] = ACTIONS(2534),
    [sym_number_literal] = ACTIONS(2536),
    [anon_sym_L_SQUOTE] = ACTIONS(2536),
    [anon_sym_u_SQUOTE] = ACTIONS(2536),
    [anon_sym_U_SQUOTE] = ACTIONS(2536),
    [anon_sym_u8_SQUOTE] = ACTIONS(2536),
    [anon_sym_SQUOTE] = ACTIONS(2536),
    [anon_sym_AT] = ACTIONS(2534),
    [anon_sym_DQUOTE] = ACTIONS(2536),
    [anon_sym_L_DQUOTE] = ACTIONS(2536),
    [anon_sym_u_DQUOTE] = ACTIONS(2536),
    [anon_sym_U_DQUOTE] = ACTIONS(2536),
    [anon_sym_u8_DQUOTE] = ACTIONS(2536),
    [sym_true] = ACTIONS(2534),
    [sym_false] = ACTIONS(2534),
    [anon_sym_NULL] = ACTIONS(2534),
    [anon_sym_nullptr] = ACTIONS(2534),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2534),
    [anon_sym___typeof] = ACTIONS(2534),
    [anon_sym_typeof] = ACTIONS(2534),
    [anon_sym_ATimport] = ACTIONS(2536),
    [aux_sym_preproc_undef_token1] = ACTIONS(2534),
    [anon_sym_POUND] = ACTIONS(2534),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2534),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2534),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2534),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_API_AVAILABLE] = ACTIONS(2534),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_API_DEPRECATED] = ACTIONS(2534),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym___deprecated_msg] = ACTIONS(2534),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2534),
    [anon_sym_ATprotocol] = ACTIONS(2536),
    [anon_sym_ATinterface] = ACTIONS(2536),
    [anon_sym_ATimplementation] = ACTIONS(2536),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2536),
    [anon_sym_ATtry] = ACTIONS(2536),
    [anon_sym___try] = ACTIONS(2534),
    [anon_sym_ATthrow] = ACTIONS(2536),
    [anon_sym_ATselector] = ACTIONS(2536),
    [anon_sym_ATavailable] = ACTIONS(2536),
    [anon_sym___builtin_available] = ACTIONS(2534),
    [anon_sym_va_arg] = ACTIONS(2534),
    [anon_sym_ATencode] = ACTIONS(2536),
    [anon_sym_ATsynchronized] = ACTIONS(2536),
    [anon_sym_BOOL] = ACTIONS(2534),
    [anon_sym_IMP] = ACTIONS(2534),
    [anon_sym_SEL] = ACTIONS(2534),
    [anon_sym_Class] = ACTIONS(2534),
    [anon_sym_id] = ACTIONS(2534),
  },
  [STATE(762)] = {
    [sym_identifier] = ACTIONS(2558),
    [aux_sym_preproc_include_token1] = ACTIONS(2558),
    [aux_sym_preproc_include_token2] = ACTIONS(2558),
    [aux_sym_preproc_def_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2558),
    [sym_preproc_directive] = ACTIONS(2558),
    [anon_sym_LPAREN2] = ACTIONS(2560),
    [anon_sym_BANG] = ACTIONS(2560),
    [anon_sym_TILDE] = ACTIONS(2560),
    [anon_sym_DASH] = ACTIONS(2558),
    [anon_sym_PLUS] = ACTIONS(2558),
    [anon_sym_STAR] = ACTIONS(2560),
    [anon_sym_CARET] = ACTIONS(2560),
    [anon_sym_AMP] = ACTIONS(2560),
    [anon_sym_SEMI] = ACTIONS(2560),
    [anon_sym___extension__] = ACTIONS(2558),
    [anon_sym_typedef] = ACTIONS(2558),
    [anon_sym_extern] = ACTIONS(2558),
    [anon_sym___attribute__] = ACTIONS(2558),
    [anon_sym___attribute] = ACTIONS(2558),
    [anon_sym_noreturn] = ACTIONS(2558),
    [anon_sym_LBRACK] = ACTIONS(2560),
    [anon_sym___declspec] = ACTIONS(2558),
    [anon_sym___cdecl] = ACTIONS(2558),
    [anon_sym___clrcall] = ACTIONS(2558),
    [anon_sym___stdcall] = ACTIONS(2558),
    [anon_sym___fastcall] = ACTIONS(2558),
    [anon_sym___thiscall] = ACTIONS(2558),
    [anon_sym___vectorcall] = ACTIONS(2558),
    [anon_sym_LBRACE] = ACTIONS(2560),
    [anon_sym_RBRACE] = ACTIONS(2560),
    [anon_sym_signed] = ACTIONS(2558),
    [anon_sym_unsigned] = ACTIONS(2558),
    [anon_sym_long] = ACTIONS(2558),
    [anon_sym_short] = ACTIONS(2558),
    [anon_sym_static] = ACTIONS(2558),
    [anon_sym_ATautoreleasepool] = ACTIONS(2560),
    [anon_sym_auto] = ACTIONS(2558),
    [anon_sym_register] = ACTIONS(2558),
    [anon_sym_inline] = ACTIONS(2558),
    [anon_sym___inline] = ACTIONS(2558),
    [anon_sym___inline__] = ACTIONS(2558),
    [anon_sym___forceinline] = ACTIONS(2558),
    [anon_sym_thread_local] = ACTIONS(2558),
    [anon_sym___thread] = ACTIONS(2558),
    [anon_sym_CG_EXTERN] = ACTIONS(2558),
    [anon_sym_CG_INLINE] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2558),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2558),
    [anon_sym_IBOutlet] = ACTIONS(2558),
    [anon_sym_IBInspectable] = ACTIONS(2558),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2558),
    [anon_sym_NS_INLINE] = ACTIONS(2558),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2558),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2558),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2558),
    [anon_sym_const] = ACTIONS(2558),
    [anon_sym_constexpr] = ACTIONS(2558),
    [anon_sym_volatile] = ACTIONS(2558),
    [anon_sym_restrict] = ACTIONS(2558),
    [anon_sym___restrict__] = ACTIONS(2558),
    [anon_sym__Atomic] = ACTIONS(2558),
    [anon_sym__Noreturn] = ACTIONS(2558),
    [anon_sym__Nonnull] = ACTIONS(2558),
    [anon_sym_nullable] = ACTIONS(2558),
    [anon_sym__Complex] = ACTIONS(2558),
    [anon_sym__Nullable] = ACTIONS(2558),
    [anon_sym__Nullable_result] = ACTIONS(2558),
    [anon_sym__Null_unspecified] = ACTIONS(2558),
    [anon_sym___autoreleasing] = ACTIONS(2558),
    [anon_sym___block] = ACTIONS(2558),
    [anon_sym___bridge] = ACTIONS(2558),
    [anon_sym___bridge_retained] = ACTIONS(2558),
    [anon_sym___bridge_transfer] = ACTIONS(2558),
    [anon_sym___complex] = ACTIONS(2558),
    [anon_sym___const] = ACTIONS(2558),
    [anon_sym___imag] = ACTIONS(2558),
    [anon_sym___kindof] = ACTIONS(2558),
    [anon_sym___nonnull] = ACTIONS(2558),
    [anon_sym___nullable] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2558),
    [anon_sym___real] = ACTIONS(2558),
    [anon_sym___strong] = ACTIONS(2558),
    [anon_sym___unsafe_unretained] = ACTIONS(2558),
    [anon_sym___unused] = ACTIONS(2558),
    [anon_sym___weak] = ACTIONS(2558),
    [anon_sym_alignas] = ACTIONS(2558),
    [anon_sym__Alignas] = ACTIONS(2558),
    [sym_primitive_type] = ACTIONS(2558),
    [anon_sym_enum] = ACTIONS(2558),
    [anon_sym_struct] = ACTIONS(2558),
    [anon_sym_union] = ACTIONS(2558),
    [anon_sym_if] = ACTIONS(2558),
    [anon_sym_else] = ACTIONS(2558),
    [anon_sym_switch] = ACTIONS(2558),
    [anon_sym_case] = ACTIONS(2558),
    [anon_sym_default] = ACTIONS(2558),
    [anon_sym_while] = ACTIONS(2558),
    [anon_sym_do] = ACTIONS(2558),
    [anon_sym_for] = ACTIONS(2558),
    [anon_sym_in] = ACTIONS(2558),
    [anon_sym_return] = ACTIONS(2558),
    [anon_sym_break] = ACTIONS(2558),
    [anon_sym_continue] = ACTIONS(2558),
    [anon_sym_goto] = ACTIONS(2558),
    [anon_sym_DASH_DASH] = ACTIONS(2560),
    [anon_sym_PLUS_PLUS] = ACTIONS(2560),
    [anon_sym_sizeof] = ACTIONS(2558),
    [anon_sym___alignof__] = ACTIONS(2558),
    [anon_sym___alignof] = ACTIONS(2558),
    [anon_sym__alignof] = ACTIONS(2558),
    [anon_sym_alignof] = ACTIONS(2558),
    [anon_sym__Alignof] = ACTIONS(2558),
    [anon_sym_offsetof] = ACTIONS(2558),
    [anon_sym__Generic] = ACTIONS(2558),
    [anon_sym_asm] = ACTIONS(2558),
    [anon_sym___asm__] = ACTIONS(2558),
    [anon_sym___asm] = ACTIONS(2558),
    [sym_number_literal] = ACTIONS(2560),
    [anon_sym_L_SQUOTE] = ACTIONS(2560),
    [anon_sym_u_SQUOTE] = ACTIONS(2560),
    [anon_sym_U_SQUOTE] = ACTIONS(2560),
    [anon_sym_u8_SQUOTE] = ACTIONS(2560),
    [anon_sym_SQUOTE] = ACTIONS(2560),
    [anon_sym_AT] = ACTIONS(2558),
    [anon_sym_DQUOTE] = ACTIONS(2560),
    [anon_sym_L_DQUOTE] = ACTIONS(2560),
    [anon_sym_u_DQUOTE] = ACTIONS(2560),
    [anon_sym_U_DQUOTE] = ACTIONS(2560),
    [anon_sym_u8_DQUOTE] = ACTIONS(2560),
    [sym_true] = ACTIONS(2558),
    [sym_false] = ACTIONS(2558),
    [anon_sym_NULL] = ACTIONS(2558),
    [anon_sym_nullptr] = ACTIONS(2558),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2558),
    [anon_sym___typeof] = ACTIONS(2558),
    [anon_sym_typeof] = ACTIONS(2558),
    [anon_sym_ATimport] = ACTIONS(2560),
    [aux_sym_preproc_undef_token1] = ACTIONS(2558),
    [anon_sym_POUND] = ACTIONS(2558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2558),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2558),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2558),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_API_AVAILABLE] = ACTIONS(2558),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_API_DEPRECATED] = ACTIONS(2558),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym___deprecated_msg] = ACTIONS(2558),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2558),
    [anon_sym_ATprotocol] = ACTIONS(2560),
    [anon_sym_ATinterface] = ACTIONS(2560),
    [anon_sym_ATimplementation] = ACTIONS(2560),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2560),
    [anon_sym_ATtry] = ACTIONS(2560),
    [anon_sym___try] = ACTIONS(2558),
    [anon_sym_ATthrow] = ACTIONS(2560),
    [anon_sym_ATselector] = ACTIONS(2560),
    [anon_sym_ATavailable] = ACTIONS(2560),
    [anon_sym___builtin_available] = ACTIONS(2558),
    [anon_sym_va_arg] = ACTIONS(2558),
    [anon_sym_ATencode] = ACTIONS(2560),
    [anon_sym_ATsynchronized] = ACTIONS(2560),
    [anon_sym_BOOL] = ACTIONS(2558),
    [anon_sym_IMP] = ACTIONS(2558),
    [anon_sym_SEL] = ACTIONS(2558),
    [anon_sym_Class] = ACTIONS(2558),
    [anon_sym_id] = ACTIONS(2558),
  },
  [STATE(763)] = {
    [ts_builtin_sym_end] = ACTIONS(2370),
    [sym_identifier] = ACTIONS(2368),
    [aux_sym_preproc_include_token1] = ACTIONS(2368),
    [aux_sym_preproc_include_token2] = ACTIONS(2368),
    [aux_sym_preproc_def_token1] = ACTIONS(2368),
    [aux_sym_preproc_if_token1] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2368),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2368),
    [sym_preproc_directive] = ACTIONS(2368),
    [anon_sym_LPAREN2] = ACTIONS(2370),
    [anon_sym_BANG] = ACTIONS(2370),
    [anon_sym_TILDE] = ACTIONS(2370),
    [anon_sym_DASH] = ACTIONS(2368),
    [anon_sym_PLUS] = ACTIONS(2368),
    [anon_sym_STAR] = ACTIONS(2370),
    [anon_sym_CARET] = ACTIONS(2370),
    [anon_sym_AMP] = ACTIONS(2370),
    [anon_sym_SEMI] = ACTIONS(2370),
    [anon_sym___extension__] = ACTIONS(2368),
    [anon_sym_typedef] = ACTIONS(2368),
    [anon_sym_extern] = ACTIONS(2368),
    [anon_sym___attribute__] = ACTIONS(2368),
    [anon_sym___attribute] = ACTIONS(2368),
    [anon_sym_noreturn] = ACTIONS(2368),
    [anon_sym_LBRACK] = ACTIONS(2370),
    [anon_sym___declspec] = ACTIONS(2368),
    [anon_sym___cdecl] = ACTIONS(2368),
    [anon_sym___clrcall] = ACTIONS(2368),
    [anon_sym___stdcall] = ACTIONS(2368),
    [anon_sym___fastcall] = ACTIONS(2368),
    [anon_sym___thiscall] = ACTIONS(2368),
    [anon_sym___vectorcall] = ACTIONS(2368),
    [anon_sym_LBRACE] = ACTIONS(2370),
    [anon_sym_signed] = ACTIONS(2368),
    [anon_sym_unsigned] = ACTIONS(2368),
    [anon_sym_long] = ACTIONS(2368),
    [anon_sym_short] = ACTIONS(2368),
    [anon_sym_static] = ACTIONS(2368),
    [anon_sym_ATautoreleasepool] = ACTIONS(2370),
    [anon_sym_auto] = ACTIONS(2368),
    [anon_sym_register] = ACTIONS(2368),
    [anon_sym_inline] = ACTIONS(2368),
    [anon_sym___inline] = ACTIONS(2368),
    [anon_sym___inline__] = ACTIONS(2368),
    [anon_sym___forceinline] = ACTIONS(2368),
    [anon_sym_thread_local] = ACTIONS(2368),
    [anon_sym___thread] = ACTIONS(2368),
    [anon_sym_CG_EXTERN] = ACTIONS(2368),
    [anon_sym_CG_INLINE] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2368),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2368),
    [anon_sym_IBOutlet] = ACTIONS(2368),
    [anon_sym_IBInspectable] = ACTIONS(2368),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2368),
    [anon_sym_NS_INLINE] = ACTIONS(2368),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2368),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2368),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2368),
    [anon_sym_const] = ACTIONS(2368),
    [anon_sym_constexpr] = ACTIONS(2368),
    [anon_sym_volatile] = ACTIONS(2368),
    [anon_sym_restrict] = ACTIONS(2368),
    [anon_sym___restrict__] = ACTIONS(2368),
    [anon_sym__Atomic] = ACTIONS(2368),
    [anon_sym__Noreturn] = ACTIONS(2368),
    [anon_sym__Nonnull] = ACTIONS(2368),
    [anon_sym_nullable] = ACTIONS(2368),
    [anon_sym__Complex] = ACTIONS(2368),
    [anon_sym__Nullable] = ACTIONS(2368),
    [anon_sym__Nullable_result] = ACTIONS(2368),
    [anon_sym__Null_unspecified] = ACTIONS(2368),
    [anon_sym___autoreleasing] = ACTIONS(2368),
    [anon_sym___block] = ACTIONS(2368),
    [anon_sym___bridge] = ACTIONS(2368),
    [anon_sym___bridge_retained] = ACTIONS(2368),
    [anon_sym___bridge_transfer] = ACTIONS(2368),
    [anon_sym___complex] = ACTIONS(2368),
    [anon_sym___const] = ACTIONS(2368),
    [anon_sym___imag] = ACTIONS(2368),
    [anon_sym___kindof] = ACTIONS(2368),
    [anon_sym___nonnull] = ACTIONS(2368),
    [anon_sym___nullable] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2368),
    [anon_sym___real] = ACTIONS(2368),
    [anon_sym___strong] = ACTIONS(2368),
    [anon_sym___unsafe_unretained] = ACTIONS(2368),
    [anon_sym___unused] = ACTIONS(2368),
    [anon_sym___weak] = ACTIONS(2368),
    [anon_sym_alignas] = ACTIONS(2368),
    [anon_sym__Alignas] = ACTIONS(2368),
    [sym_primitive_type] = ACTIONS(2368),
    [anon_sym_enum] = ACTIONS(2368),
    [anon_sym_struct] = ACTIONS(2368),
    [anon_sym_union] = ACTIONS(2368),
    [anon_sym_if] = ACTIONS(2368),
    [anon_sym_else] = ACTIONS(2368),
    [anon_sym_switch] = ACTIONS(2368),
    [anon_sym_case] = ACTIONS(2368),
    [anon_sym_default] = ACTIONS(2368),
    [anon_sym_while] = ACTIONS(2368),
    [anon_sym_do] = ACTIONS(2368),
    [anon_sym_for] = ACTIONS(2368),
    [anon_sym_in] = ACTIONS(2368),
    [anon_sym_return] = ACTIONS(2368),
    [anon_sym_break] = ACTIONS(2368),
    [anon_sym_continue] = ACTIONS(2368),
    [anon_sym_goto] = ACTIONS(2368),
    [anon_sym_DASH_DASH] = ACTIONS(2370),
    [anon_sym_PLUS_PLUS] = ACTIONS(2370),
    [anon_sym_sizeof] = ACTIONS(2368),
    [anon_sym___alignof__] = ACTIONS(2368),
    [anon_sym___alignof] = ACTIONS(2368),
    [anon_sym__alignof] = ACTIONS(2368),
    [anon_sym_alignof] = ACTIONS(2368),
    [anon_sym__Alignof] = ACTIONS(2368),
    [anon_sym_offsetof] = ACTIONS(2368),
    [anon_sym__Generic] = ACTIONS(2368),
    [anon_sym_asm] = ACTIONS(2368),
    [anon_sym___asm__] = ACTIONS(2368),
    [anon_sym___asm] = ACTIONS(2368),
    [sym_number_literal] = ACTIONS(2370),
    [anon_sym_L_SQUOTE] = ACTIONS(2370),
    [anon_sym_u_SQUOTE] = ACTIONS(2370),
    [anon_sym_U_SQUOTE] = ACTIONS(2370),
    [anon_sym_u8_SQUOTE] = ACTIONS(2370),
    [anon_sym_SQUOTE] = ACTIONS(2370),
    [anon_sym_AT] = ACTIONS(2368),
    [anon_sym_DQUOTE] = ACTIONS(2370),
    [anon_sym_L_DQUOTE] = ACTIONS(2370),
    [anon_sym_u_DQUOTE] = ACTIONS(2370),
    [anon_sym_U_DQUOTE] = ACTIONS(2370),
    [anon_sym_u8_DQUOTE] = ACTIONS(2370),
    [sym_true] = ACTIONS(2368),
    [sym_false] = ACTIONS(2368),
    [anon_sym_NULL] = ACTIONS(2368),
    [anon_sym_nullptr] = ACTIONS(2368),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2368),
    [anon_sym___typeof] = ACTIONS(2368),
    [anon_sym_typeof] = ACTIONS(2368),
    [anon_sym_ATimport] = ACTIONS(2370),
    [aux_sym_preproc_undef_token1] = ACTIONS(2368),
    [anon_sym_POUND] = ACTIONS(2368),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2368),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2368),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2368),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_API_AVAILABLE] = ACTIONS(2368),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_API_DEPRECATED] = ACTIONS(2368),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym___deprecated_msg] = ACTIONS(2368),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2368),
    [anon_sym_ATprotocol] = ACTIONS(2370),
    [anon_sym_ATinterface] = ACTIONS(2370),
    [anon_sym_ATimplementation] = ACTIONS(2370),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2370),
    [anon_sym_ATcatch] = ACTIONS(2370),
    [anon_sym___catch] = ACTIONS(2368),
    [anon_sym_ATfinally] = ACTIONS(2370),
    [anon_sym___finally] = ACTIONS(2368),
    [anon_sym_ATselector] = ACTIONS(2370),
    [anon_sym_ATavailable] = ACTIONS(2370),
    [anon_sym___builtin_available] = ACTIONS(2368),
    [anon_sym_va_arg] = ACTIONS(2368),
    [anon_sym_ATencode] = ACTIONS(2370),
    [anon_sym_BOOL] = ACTIONS(2368),
    [anon_sym_IMP] = ACTIONS(2368),
    [anon_sym_SEL] = ACTIONS(2368),
    [anon_sym_Class] = ACTIONS(2368),
    [anon_sym_id] = ACTIONS(2368),
  },
  [STATE(764)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(765)] = {
    [sym_identifier] = ACTIONS(2498),
    [aux_sym_preproc_include_token1] = ACTIONS(2498),
    [aux_sym_preproc_include_token2] = ACTIONS(2498),
    [aux_sym_preproc_def_token1] = ACTIONS(2498),
    [aux_sym_preproc_if_token1] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2498),
    [sym_preproc_directive] = ACTIONS(2498),
    [anon_sym_LPAREN2] = ACTIONS(2500),
    [anon_sym_BANG] = ACTIONS(2500),
    [anon_sym_TILDE] = ACTIONS(2500),
    [anon_sym_DASH] = ACTIONS(2498),
    [anon_sym_PLUS] = ACTIONS(2498),
    [anon_sym_STAR] = ACTIONS(2500),
    [anon_sym_CARET] = ACTIONS(2500),
    [anon_sym_AMP] = ACTIONS(2500),
    [anon_sym_SEMI] = ACTIONS(2500),
    [anon_sym___extension__] = ACTIONS(2498),
    [anon_sym_typedef] = ACTIONS(2498),
    [anon_sym_extern] = ACTIONS(2498),
    [anon_sym___attribute__] = ACTIONS(2498),
    [anon_sym___attribute] = ACTIONS(2498),
    [anon_sym_noreturn] = ACTIONS(2498),
    [anon_sym_LBRACK] = ACTIONS(2500),
    [anon_sym___declspec] = ACTIONS(2498),
    [anon_sym___cdecl] = ACTIONS(2498),
    [anon_sym___clrcall] = ACTIONS(2498),
    [anon_sym___stdcall] = ACTIONS(2498),
    [anon_sym___fastcall] = ACTIONS(2498),
    [anon_sym___thiscall] = ACTIONS(2498),
    [anon_sym___vectorcall] = ACTIONS(2498),
    [anon_sym_LBRACE] = ACTIONS(2500),
    [anon_sym_RBRACE] = ACTIONS(2500),
    [anon_sym_signed] = ACTIONS(2498),
    [anon_sym_unsigned] = ACTIONS(2498),
    [anon_sym_long] = ACTIONS(2498),
    [anon_sym_short] = ACTIONS(2498),
    [anon_sym_static] = ACTIONS(2498),
    [anon_sym_ATautoreleasepool] = ACTIONS(2500),
    [anon_sym_auto] = ACTIONS(2498),
    [anon_sym_register] = ACTIONS(2498),
    [anon_sym_inline] = ACTIONS(2498),
    [anon_sym___inline] = ACTIONS(2498),
    [anon_sym___inline__] = ACTIONS(2498),
    [anon_sym___forceinline] = ACTIONS(2498),
    [anon_sym_thread_local] = ACTIONS(2498),
    [anon_sym___thread] = ACTIONS(2498),
    [anon_sym_CG_EXTERN] = ACTIONS(2498),
    [anon_sym_CG_INLINE] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2498),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2498),
    [anon_sym_IBOutlet] = ACTIONS(2498),
    [anon_sym_IBInspectable] = ACTIONS(2498),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2498),
    [anon_sym_NS_INLINE] = ACTIONS(2498),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2498),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2498),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2498),
    [anon_sym_const] = ACTIONS(2498),
    [anon_sym_constexpr] = ACTIONS(2498),
    [anon_sym_volatile] = ACTIONS(2498),
    [anon_sym_restrict] = ACTIONS(2498),
    [anon_sym___restrict__] = ACTIONS(2498),
    [anon_sym__Atomic] = ACTIONS(2498),
    [anon_sym__Noreturn] = ACTIONS(2498),
    [anon_sym__Nonnull] = ACTIONS(2498),
    [anon_sym_nullable] = ACTIONS(2498),
    [anon_sym__Complex] = ACTIONS(2498),
    [anon_sym__Nullable] = ACTIONS(2498),
    [anon_sym__Nullable_result] = ACTIONS(2498),
    [anon_sym__Null_unspecified] = ACTIONS(2498),
    [anon_sym___autoreleasing] = ACTIONS(2498),
    [anon_sym___block] = ACTIONS(2498),
    [anon_sym___bridge] = ACTIONS(2498),
    [anon_sym___bridge_retained] = ACTIONS(2498),
    [anon_sym___bridge_transfer] = ACTIONS(2498),
    [anon_sym___complex] = ACTIONS(2498),
    [anon_sym___const] = ACTIONS(2498),
    [anon_sym___imag] = ACTIONS(2498),
    [anon_sym___kindof] = ACTIONS(2498),
    [anon_sym___nonnull] = ACTIONS(2498),
    [anon_sym___nullable] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2498),
    [anon_sym___real] = ACTIONS(2498),
    [anon_sym___strong] = ACTIONS(2498),
    [anon_sym___unsafe_unretained] = ACTIONS(2498),
    [anon_sym___unused] = ACTIONS(2498),
    [anon_sym___weak] = ACTIONS(2498),
    [anon_sym_alignas] = ACTIONS(2498),
    [anon_sym__Alignas] = ACTIONS(2498),
    [sym_primitive_type] = ACTIONS(2498),
    [anon_sym_enum] = ACTIONS(2498),
    [anon_sym_struct] = ACTIONS(2498),
    [anon_sym_union] = ACTIONS(2498),
    [anon_sym_if] = ACTIONS(2498),
    [anon_sym_else] = ACTIONS(2498),
    [anon_sym_switch] = ACTIONS(2498),
    [anon_sym_case] = ACTIONS(2498),
    [anon_sym_default] = ACTIONS(2498),
    [anon_sym_while] = ACTIONS(2498),
    [anon_sym_do] = ACTIONS(2498),
    [anon_sym_for] = ACTIONS(2498),
    [anon_sym_in] = ACTIONS(2498),
    [anon_sym_return] = ACTIONS(2498),
    [anon_sym_break] = ACTIONS(2498),
    [anon_sym_continue] = ACTIONS(2498),
    [anon_sym_goto] = ACTIONS(2498),
    [anon_sym_DASH_DASH] = ACTIONS(2500),
    [anon_sym_PLUS_PLUS] = ACTIONS(2500),
    [anon_sym_sizeof] = ACTIONS(2498),
    [anon_sym___alignof__] = ACTIONS(2498),
    [anon_sym___alignof] = ACTIONS(2498),
    [anon_sym__alignof] = ACTIONS(2498),
    [anon_sym_alignof] = ACTIONS(2498),
    [anon_sym__Alignof] = ACTIONS(2498),
    [anon_sym_offsetof] = ACTIONS(2498),
    [anon_sym__Generic] = ACTIONS(2498),
    [anon_sym_asm] = ACTIONS(2498),
    [anon_sym___asm__] = ACTIONS(2498),
    [anon_sym___asm] = ACTIONS(2498),
    [sym_number_literal] = ACTIONS(2500),
    [anon_sym_L_SQUOTE] = ACTIONS(2500),
    [anon_sym_u_SQUOTE] = ACTIONS(2500),
    [anon_sym_U_SQUOTE] = ACTIONS(2500),
    [anon_sym_u8_SQUOTE] = ACTIONS(2500),
    [anon_sym_SQUOTE] = ACTIONS(2500),
    [anon_sym_AT] = ACTIONS(2498),
    [anon_sym_DQUOTE] = ACTIONS(2500),
    [anon_sym_L_DQUOTE] = ACTIONS(2500),
    [anon_sym_u_DQUOTE] = ACTIONS(2500),
    [anon_sym_U_DQUOTE] = ACTIONS(2500),
    [anon_sym_u8_DQUOTE] = ACTIONS(2500),
    [sym_true] = ACTIONS(2498),
    [sym_false] = ACTIONS(2498),
    [anon_sym_NULL] = ACTIONS(2498),
    [anon_sym_nullptr] = ACTIONS(2498),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2498),
    [anon_sym___typeof] = ACTIONS(2498),
    [anon_sym_typeof] = ACTIONS(2498),
    [anon_sym_ATimport] = ACTIONS(2500),
    [aux_sym_preproc_undef_token1] = ACTIONS(2498),
    [anon_sym_POUND] = ACTIONS(2498),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2498),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2498),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2498),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_API_AVAILABLE] = ACTIONS(2498),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_API_DEPRECATED] = ACTIONS(2498),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym___deprecated_msg] = ACTIONS(2498),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2498),
    [anon_sym_ATprotocol] = ACTIONS(2500),
    [anon_sym_ATinterface] = ACTIONS(2500),
    [anon_sym_ATimplementation] = ACTIONS(2500),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2500),
    [anon_sym_ATtry] = ACTIONS(2500),
    [anon_sym___try] = ACTIONS(2498),
    [anon_sym_ATthrow] = ACTIONS(2500),
    [anon_sym_ATselector] = ACTIONS(2500),
    [anon_sym_ATavailable] = ACTIONS(2500),
    [anon_sym___builtin_available] = ACTIONS(2498),
    [anon_sym_va_arg] = ACTIONS(2498),
    [anon_sym_ATencode] = ACTIONS(2500),
    [anon_sym_ATsynchronized] = ACTIONS(2500),
    [anon_sym_BOOL] = ACTIONS(2498),
    [anon_sym_IMP] = ACTIONS(2498),
    [anon_sym_SEL] = ACTIONS(2498),
    [anon_sym_Class] = ACTIONS(2498),
    [anon_sym_id] = ACTIONS(2498),
  },
  [STATE(766)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(767)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(768)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(769)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(770)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(771)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(772)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(773)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(774)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(775)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(776)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(777)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(778)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(779)] = {
    [ts_builtin_sym_end] = ACTIONS(2418),
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(780)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(781)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(782)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_RBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(783)] = {
    [sym_identifier] = ACTIONS(2568),
    [aux_sym_preproc_include_token1] = ACTIONS(2568),
    [aux_sym_preproc_include_token2] = ACTIONS(2568),
    [aux_sym_preproc_def_token1] = ACTIONS(2568),
    [aux_sym_preproc_if_token1] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2568),
    [sym_preproc_directive] = ACTIONS(2568),
    [anon_sym_LPAREN2] = ACTIONS(2570),
    [anon_sym_BANG] = ACTIONS(2570),
    [anon_sym_TILDE] = ACTIONS(2570),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_PLUS] = ACTIONS(2568),
    [anon_sym_STAR] = ACTIONS(2570),
    [anon_sym_CARET] = ACTIONS(2570),
    [anon_sym_AMP] = ACTIONS(2570),
    [anon_sym_SEMI] = ACTIONS(2570),
    [anon_sym___extension__] = ACTIONS(2568),
    [anon_sym_typedef] = ACTIONS(2568),
    [anon_sym_extern] = ACTIONS(2568),
    [anon_sym___attribute__] = ACTIONS(2568),
    [anon_sym___attribute] = ACTIONS(2568),
    [anon_sym_noreturn] = ACTIONS(2568),
    [anon_sym_LBRACK] = ACTIONS(2570),
    [anon_sym___declspec] = ACTIONS(2568),
    [anon_sym___cdecl] = ACTIONS(2568),
    [anon_sym___clrcall] = ACTIONS(2568),
    [anon_sym___stdcall] = ACTIONS(2568),
    [anon_sym___fastcall] = ACTIONS(2568),
    [anon_sym___thiscall] = ACTIONS(2568),
    [anon_sym___vectorcall] = ACTIONS(2568),
    [anon_sym_LBRACE] = ACTIONS(2570),
    [anon_sym_RBRACE] = ACTIONS(2570),
    [anon_sym_signed] = ACTIONS(2568),
    [anon_sym_unsigned] = ACTIONS(2568),
    [anon_sym_long] = ACTIONS(2568),
    [anon_sym_short] = ACTIONS(2568),
    [anon_sym_static] = ACTIONS(2568),
    [anon_sym_ATautoreleasepool] = ACTIONS(2570),
    [anon_sym_auto] = ACTIONS(2568),
    [anon_sym_register] = ACTIONS(2568),
    [anon_sym_inline] = ACTIONS(2568),
    [anon_sym___inline] = ACTIONS(2568),
    [anon_sym___inline__] = ACTIONS(2568),
    [anon_sym___forceinline] = ACTIONS(2568),
    [anon_sym_thread_local] = ACTIONS(2568),
    [anon_sym___thread] = ACTIONS(2568),
    [anon_sym_CG_EXTERN] = ACTIONS(2568),
    [anon_sym_CG_INLINE] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2568),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2568),
    [anon_sym_IBOutlet] = ACTIONS(2568),
    [anon_sym_IBInspectable] = ACTIONS(2568),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2568),
    [anon_sym_NS_INLINE] = ACTIONS(2568),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2568),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2568),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2568),
    [anon_sym_const] = ACTIONS(2568),
    [anon_sym_constexpr] = ACTIONS(2568),
    [anon_sym_volatile] = ACTIONS(2568),
    [anon_sym_restrict] = ACTIONS(2568),
    [anon_sym___restrict__] = ACTIONS(2568),
    [anon_sym__Atomic] = ACTIONS(2568),
    [anon_sym__Noreturn] = ACTIONS(2568),
    [anon_sym__Nonnull] = ACTIONS(2568),
    [anon_sym_nullable] = ACTIONS(2568),
    [anon_sym__Complex] = ACTIONS(2568),
    [anon_sym__Nullable] = ACTIONS(2568),
    [anon_sym__Nullable_result] = ACTIONS(2568),
    [anon_sym__Null_unspecified] = ACTIONS(2568),
    [anon_sym___autoreleasing] = ACTIONS(2568),
    [anon_sym___block] = ACTIONS(2568),
    [anon_sym___bridge] = ACTIONS(2568),
    [anon_sym___bridge_retained] = ACTIONS(2568),
    [anon_sym___bridge_transfer] = ACTIONS(2568),
    [anon_sym___complex] = ACTIONS(2568),
    [anon_sym___const] = ACTIONS(2568),
    [anon_sym___imag] = ACTIONS(2568),
    [anon_sym___kindof] = ACTIONS(2568),
    [anon_sym___nonnull] = ACTIONS(2568),
    [anon_sym___nullable] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2568),
    [anon_sym___real] = ACTIONS(2568),
    [anon_sym___strong] = ACTIONS(2568),
    [anon_sym___unsafe_unretained] = ACTIONS(2568),
    [anon_sym___unused] = ACTIONS(2568),
    [anon_sym___weak] = ACTIONS(2568),
    [anon_sym_alignas] = ACTIONS(2568),
    [anon_sym__Alignas] = ACTIONS(2568),
    [sym_primitive_type] = ACTIONS(2568),
    [anon_sym_enum] = ACTIONS(2568),
    [anon_sym_struct] = ACTIONS(2568),
    [anon_sym_union] = ACTIONS(2568),
    [anon_sym_if] = ACTIONS(2568),
    [anon_sym_else] = ACTIONS(2568),
    [anon_sym_switch] = ACTIONS(2568),
    [anon_sym_case] = ACTIONS(2568),
    [anon_sym_default] = ACTIONS(2568),
    [anon_sym_while] = ACTIONS(2568),
    [anon_sym_do] = ACTIONS(2568),
    [anon_sym_for] = ACTIONS(2568),
    [anon_sym_in] = ACTIONS(2568),
    [anon_sym_return] = ACTIONS(2568),
    [anon_sym_break] = ACTIONS(2568),
    [anon_sym_continue] = ACTIONS(2568),
    [anon_sym_goto] = ACTIONS(2568),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_sizeof] = ACTIONS(2568),
    [anon_sym___alignof__] = ACTIONS(2568),
    [anon_sym___alignof] = ACTIONS(2568),
    [anon_sym__alignof] = ACTIONS(2568),
    [anon_sym_alignof] = ACTIONS(2568),
    [anon_sym__Alignof] = ACTIONS(2568),
    [anon_sym_offsetof] = ACTIONS(2568),
    [anon_sym__Generic] = ACTIONS(2568),
    [anon_sym_asm] = ACTIONS(2568),
    [anon_sym___asm__] = ACTIONS(2568),
    [anon_sym___asm] = ACTIONS(2568),
    [sym_number_literal] = ACTIONS(2570),
    [anon_sym_L_SQUOTE] = ACTIONS(2570),
    [anon_sym_u_SQUOTE] = ACTIONS(2570),
    [anon_sym_U_SQUOTE] = ACTIONS(2570),
    [anon_sym_u8_SQUOTE] = ACTIONS(2570),
    [anon_sym_SQUOTE] = ACTIONS(2570),
    [anon_sym_AT] = ACTIONS(2568),
    [anon_sym_DQUOTE] = ACTIONS(2570),
    [anon_sym_L_DQUOTE] = ACTIONS(2570),
    [anon_sym_u_DQUOTE] = ACTIONS(2570),
    [anon_sym_U_DQUOTE] = ACTIONS(2570),
    [anon_sym_u8_DQUOTE] = ACTIONS(2570),
    [sym_true] = ACTIONS(2568),
    [sym_false] = ACTIONS(2568),
    [anon_sym_NULL] = ACTIONS(2568),
    [anon_sym_nullptr] = ACTIONS(2568),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2568),
    [anon_sym___typeof] = ACTIONS(2568),
    [anon_sym_typeof] = ACTIONS(2568),
    [anon_sym_ATimport] = ACTIONS(2570),
    [aux_sym_preproc_undef_token1] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2568),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2568),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2568),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2568),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_API_AVAILABLE] = ACTIONS(2568),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_API_DEPRECATED] = ACTIONS(2568),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym___deprecated_msg] = ACTIONS(2568),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2568),
    [anon_sym_ATprotocol] = ACTIONS(2570),
    [anon_sym_ATinterface] = ACTIONS(2570),
    [anon_sym_ATimplementation] = ACTIONS(2570),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2570),
    [anon_sym_ATtry] = ACTIONS(2570),
    [anon_sym___try] = ACTIONS(2568),
    [anon_sym_ATthrow] = ACTIONS(2570),
    [anon_sym_ATselector] = ACTIONS(2570),
    [anon_sym_ATavailable] = ACTIONS(2570),
    [anon_sym___builtin_available] = ACTIONS(2568),
    [anon_sym_va_arg] = ACTIONS(2568),
    [anon_sym_ATencode] = ACTIONS(2570),
    [anon_sym_ATsynchronized] = ACTIONS(2570),
    [anon_sym_BOOL] = ACTIONS(2568),
    [anon_sym_IMP] = ACTIONS(2568),
    [anon_sym_SEL] = ACTIONS(2568),
    [anon_sym_Class] = ACTIONS(2568),
    [anon_sym_id] = ACTIONS(2568),
  },
  [STATE(784)] = {
    [sym_identifier] = ACTIONS(2478),
    [aux_sym_preproc_include_token1] = ACTIONS(2478),
    [aux_sym_preproc_include_token2] = ACTIONS(2478),
    [aux_sym_preproc_def_token1] = ACTIONS(2478),
    [aux_sym_preproc_if_token1] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2478),
    [sym_preproc_directive] = ACTIONS(2478),
    [anon_sym_LPAREN2] = ACTIONS(2480),
    [anon_sym_BANG] = ACTIONS(2480),
    [anon_sym_TILDE] = ACTIONS(2480),
    [anon_sym_DASH] = ACTIONS(2478),
    [anon_sym_PLUS] = ACTIONS(2478),
    [anon_sym_STAR] = ACTIONS(2480),
    [anon_sym_CARET] = ACTIONS(2480),
    [anon_sym_AMP] = ACTIONS(2480),
    [anon_sym_SEMI] = ACTIONS(2480),
    [anon_sym___extension__] = ACTIONS(2478),
    [anon_sym_typedef] = ACTIONS(2478),
    [anon_sym_extern] = ACTIONS(2478),
    [anon_sym___attribute__] = ACTIONS(2478),
    [anon_sym___attribute] = ACTIONS(2478),
    [anon_sym_noreturn] = ACTIONS(2478),
    [anon_sym_LBRACK] = ACTIONS(2480),
    [anon_sym___declspec] = ACTIONS(2478),
    [anon_sym___cdecl] = ACTIONS(2478),
    [anon_sym___clrcall] = ACTIONS(2478),
    [anon_sym___stdcall] = ACTIONS(2478),
    [anon_sym___fastcall] = ACTIONS(2478),
    [anon_sym___thiscall] = ACTIONS(2478),
    [anon_sym___vectorcall] = ACTIONS(2478),
    [anon_sym_LBRACE] = ACTIONS(2480),
    [anon_sym_RBRACE] = ACTIONS(2480),
    [anon_sym_signed] = ACTIONS(2478),
    [anon_sym_unsigned] = ACTIONS(2478),
    [anon_sym_long] = ACTIONS(2478),
    [anon_sym_short] = ACTIONS(2478),
    [anon_sym_static] = ACTIONS(2478),
    [anon_sym_ATautoreleasepool] = ACTIONS(2480),
    [anon_sym_auto] = ACTIONS(2478),
    [anon_sym_register] = ACTIONS(2478),
    [anon_sym_inline] = ACTIONS(2478),
    [anon_sym___inline] = ACTIONS(2478),
    [anon_sym___inline__] = ACTIONS(2478),
    [anon_sym___forceinline] = ACTIONS(2478),
    [anon_sym_thread_local] = ACTIONS(2478),
    [anon_sym___thread] = ACTIONS(2478),
    [anon_sym_CG_EXTERN] = ACTIONS(2478),
    [anon_sym_CG_INLINE] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2478),
    [anon_sym_IBOutlet] = ACTIONS(2478),
    [anon_sym_IBInspectable] = ACTIONS(2478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2478),
    [anon_sym_NS_INLINE] = ACTIONS(2478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2478),
    [anon_sym_const] = ACTIONS(2478),
    [anon_sym_constexpr] = ACTIONS(2478),
    [anon_sym_volatile] = ACTIONS(2478),
    [anon_sym_restrict] = ACTIONS(2478),
    [anon_sym___restrict__] = ACTIONS(2478),
    [anon_sym__Atomic] = ACTIONS(2478),
    [anon_sym__Noreturn] = ACTIONS(2478),
    [anon_sym__Nonnull] = ACTIONS(2478),
    [anon_sym_nullable] = ACTIONS(2478),
    [anon_sym__Complex] = ACTIONS(2478),
    [anon_sym__Nullable] = ACTIONS(2478),
    [anon_sym__Nullable_result] = ACTIONS(2478),
    [anon_sym__Null_unspecified] = ACTIONS(2478),
    [anon_sym___autoreleasing] = ACTIONS(2478),
    [anon_sym___block] = ACTIONS(2478),
    [anon_sym___bridge] = ACTIONS(2478),
    [anon_sym___bridge_retained] = ACTIONS(2478),
    [anon_sym___bridge_transfer] = ACTIONS(2478),
    [anon_sym___complex] = ACTIONS(2478),
    [anon_sym___const] = ACTIONS(2478),
    [anon_sym___imag] = ACTIONS(2478),
    [anon_sym___kindof] = ACTIONS(2478),
    [anon_sym___nonnull] = ACTIONS(2478),
    [anon_sym___nullable] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2478),
    [anon_sym___real] = ACTIONS(2478),
    [anon_sym___strong] = ACTIONS(2478),
    [anon_sym___unsafe_unretained] = ACTIONS(2478),
    [anon_sym___unused] = ACTIONS(2478),
    [anon_sym___weak] = ACTIONS(2478),
    [anon_sym_alignas] = ACTIONS(2478),
    [anon_sym__Alignas] = ACTIONS(2478),
    [sym_primitive_type] = ACTIONS(2478),
    [anon_sym_enum] = ACTIONS(2478),
    [anon_sym_struct] = ACTIONS(2478),
    [anon_sym_union] = ACTIONS(2478),
    [anon_sym_if] = ACTIONS(2478),
    [anon_sym_else] = ACTIONS(2478),
    [anon_sym_switch] = ACTIONS(2478),
    [anon_sym_case] = ACTIONS(2478),
    [anon_sym_default] = ACTIONS(2478),
    [anon_sym_while] = ACTIONS(2478),
    [anon_sym_do] = ACTIONS(2478),
    [anon_sym_for] = ACTIONS(2478),
    [anon_sym_in] = ACTIONS(2478),
    [anon_sym_return] = ACTIONS(2478),
    [anon_sym_break] = ACTIONS(2478),
    [anon_sym_continue] = ACTIONS(2478),
    [anon_sym_goto] = ACTIONS(2478),
    [anon_sym_DASH_DASH] = ACTIONS(2480),
    [anon_sym_PLUS_PLUS] = ACTIONS(2480),
    [anon_sym_sizeof] = ACTIONS(2478),
    [anon_sym___alignof__] = ACTIONS(2478),
    [anon_sym___alignof] = ACTIONS(2478),
    [anon_sym__alignof] = ACTIONS(2478),
    [anon_sym_alignof] = ACTIONS(2478),
    [anon_sym__Alignof] = ACTIONS(2478),
    [anon_sym_offsetof] = ACTIONS(2478),
    [anon_sym__Generic] = ACTIONS(2478),
    [anon_sym_asm] = ACTIONS(2478),
    [anon_sym___asm__] = ACTIONS(2478),
    [anon_sym___asm] = ACTIONS(2478),
    [sym_number_literal] = ACTIONS(2480),
    [anon_sym_L_SQUOTE] = ACTIONS(2480),
    [anon_sym_u_SQUOTE] = ACTIONS(2480),
    [anon_sym_U_SQUOTE] = ACTIONS(2480),
    [anon_sym_u8_SQUOTE] = ACTIONS(2480),
    [anon_sym_SQUOTE] = ACTIONS(2480),
    [anon_sym_AT] = ACTIONS(2478),
    [anon_sym_DQUOTE] = ACTIONS(2480),
    [anon_sym_L_DQUOTE] = ACTIONS(2480),
    [anon_sym_u_DQUOTE] = ACTIONS(2480),
    [anon_sym_U_DQUOTE] = ACTIONS(2480),
    [anon_sym_u8_DQUOTE] = ACTIONS(2480),
    [sym_true] = ACTIONS(2478),
    [sym_false] = ACTIONS(2478),
    [anon_sym_NULL] = ACTIONS(2478),
    [anon_sym_nullptr] = ACTIONS(2478),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2478),
    [anon_sym___typeof] = ACTIONS(2478),
    [anon_sym_typeof] = ACTIONS(2478),
    [anon_sym_ATimport] = ACTIONS(2480),
    [aux_sym_preproc_undef_token1] = ACTIONS(2478),
    [anon_sym_POUND] = ACTIONS(2478),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_API_AVAILABLE] = ACTIONS(2478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_API_DEPRECATED] = ACTIONS(2478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym___deprecated_msg] = ACTIONS(2478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2478),
    [anon_sym_ATprotocol] = ACTIONS(2480),
    [anon_sym_ATinterface] = ACTIONS(2480),
    [anon_sym_ATimplementation] = ACTIONS(2480),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2480),
    [anon_sym_ATtry] = ACTIONS(2480),
    [anon_sym___try] = ACTIONS(2478),
    [anon_sym_ATthrow] = ACTIONS(2480),
    [anon_sym_ATselector] = ACTIONS(2480),
    [anon_sym_ATavailable] = ACTIONS(2480),
    [anon_sym___builtin_available] = ACTIONS(2478),
    [anon_sym_va_arg] = ACTIONS(2478),
    [anon_sym_ATencode] = ACTIONS(2480),
    [anon_sym_ATsynchronized] = ACTIONS(2480),
    [anon_sym_BOOL] = ACTIONS(2478),
    [anon_sym_IMP] = ACTIONS(2478),
    [anon_sym_SEL] = ACTIONS(2478),
    [anon_sym_Class] = ACTIONS(2478),
    [anon_sym_id] = ACTIONS(2478),
  },
  [STATE(785)] = {
    [ts_builtin_sym_end] = ACTIONS(2540),
    [sym_identifier] = ACTIONS(2538),
    [aux_sym_preproc_include_token1] = ACTIONS(2538),
    [aux_sym_preproc_include_token2] = ACTIONS(2538),
    [aux_sym_preproc_def_token1] = ACTIONS(2538),
    [aux_sym_preproc_if_token1] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2538),
    [sym_preproc_directive] = ACTIONS(2538),
    [anon_sym_LPAREN2] = ACTIONS(2540),
    [anon_sym_BANG] = ACTIONS(2540),
    [anon_sym_TILDE] = ACTIONS(2540),
    [anon_sym_DASH] = ACTIONS(2538),
    [anon_sym_PLUS] = ACTIONS(2538),
    [anon_sym_STAR] = ACTIONS(2540),
    [anon_sym_CARET] = ACTIONS(2540),
    [anon_sym_AMP] = ACTIONS(2540),
    [anon_sym_SEMI] = ACTIONS(2540),
    [anon_sym___extension__] = ACTIONS(2538),
    [anon_sym_typedef] = ACTIONS(2538),
    [anon_sym_extern] = ACTIONS(2538),
    [anon_sym___attribute__] = ACTIONS(2538),
    [anon_sym___attribute] = ACTIONS(2538),
    [anon_sym_noreturn] = ACTIONS(2538),
    [anon_sym_LBRACK] = ACTIONS(2540),
    [anon_sym___declspec] = ACTIONS(2538),
    [anon_sym___cdecl] = ACTIONS(2538),
    [anon_sym___clrcall] = ACTIONS(2538),
    [anon_sym___stdcall] = ACTIONS(2538),
    [anon_sym___fastcall] = ACTIONS(2538),
    [anon_sym___thiscall] = ACTIONS(2538),
    [anon_sym___vectorcall] = ACTIONS(2538),
    [anon_sym_LBRACE] = ACTIONS(2540),
    [anon_sym_signed] = ACTIONS(2538),
    [anon_sym_unsigned] = ACTIONS(2538),
    [anon_sym_long] = ACTIONS(2538),
    [anon_sym_short] = ACTIONS(2538),
    [anon_sym_static] = ACTIONS(2538),
    [anon_sym_ATautoreleasepool] = ACTIONS(2540),
    [anon_sym_auto] = ACTIONS(2538),
    [anon_sym_register] = ACTIONS(2538),
    [anon_sym_inline] = ACTIONS(2538),
    [anon_sym___inline] = ACTIONS(2538),
    [anon_sym___inline__] = ACTIONS(2538),
    [anon_sym___forceinline] = ACTIONS(2538),
    [anon_sym_thread_local] = ACTIONS(2538),
    [anon_sym___thread] = ACTIONS(2538),
    [anon_sym_CG_EXTERN] = ACTIONS(2538),
    [anon_sym_CG_INLINE] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2538),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2538),
    [anon_sym_IBOutlet] = ACTIONS(2538),
    [anon_sym_IBInspectable] = ACTIONS(2538),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2538),
    [anon_sym_NS_INLINE] = ACTIONS(2538),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2538),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2538),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2538),
    [anon_sym_const] = ACTIONS(2538),
    [anon_sym_constexpr] = ACTIONS(2538),
    [anon_sym_volatile] = ACTIONS(2538),
    [anon_sym_restrict] = ACTIONS(2538),
    [anon_sym___restrict__] = ACTIONS(2538),
    [anon_sym__Atomic] = ACTIONS(2538),
    [anon_sym__Noreturn] = ACTIONS(2538),
    [anon_sym__Nonnull] = ACTIONS(2538),
    [anon_sym_nullable] = ACTIONS(2538),
    [anon_sym__Complex] = ACTIONS(2538),
    [anon_sym__Nullable] = ACTIONS(2538),
    [anon_sym__Nullable_result] = ACTIONS(2538),
    [anon_sym__Null_unspecified] = ACTIONS(2538),
    [anon_sym___autoreleasing] = ACTIONS(2538),
    [anon_sym___block] = ACTIONS(2538),
    [anon_sym___bridge] = ACTIONS(2538),
    [anon_sym___bridge_retained] = ACTIONS(2538),
    [anon_sym___bridge_transfer] = ACTIONS(2538),
    [anon_sym___complex] = ACTIONS(2538),
    [anon_sym___const] = ACTIONS(2538),
    [anon_sym___imag] = ACTIONS(2538),
    [anon_sym___kindof] = ACTIONS(2538),
    [anon_sym___nonnull] = ACTIONS(2538),
    [anon_sym___nullable] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2538),
    [anon_sym___real] = ACTIONS(2538),
    [anon_sym___strong] = ACTIONS(2538),
    [anon_sym___unsafe_unretained] = ACTIONS(2538),
    [anon_sym___unused] = ACTIONS(2538),
    [anon_sym___weak] = ACTIONS(2538),
    [anon_sym_alignas] = ACTIONS(2538),
    [anon_sym__Alignas] = ACTIONS(2538),
    [sym_primitive_type] = ACTIONS(2538),
    [anon_sym_enum] = ACTIONS(2538),
    [anon_sym_struct] = ACTIONS(2538),
    [anon_sym_union] = ACTIONS(2538),
    [anon_sym_if] = ACTIONS(2538),
    [anon_sym_else] = ACTIONS(2538),
    [anon_sym_switch] = ACTIONS(2538),
    [anon_sym_case] = ACTIONS(2538),
    [anon_sym_default] = ACTIONS(2538),
    [anon_sym_while] = ACTIONS(2538),
    [anon_sym_do] = ACTIONS(2538),
    [anon_sym_for] = ACTIONS(2538),
    [anon_sym_in] = ACTIONS(2538),
    [anon_sym_return] = ACTIONS(2538),
    [anon_sym_break] = ACTIONS(2538),
    [anon_sym_continue] = ACTIONS(2538),
    [anon_sym_goto] = ACTIONS(2538),
    [anon_sym_DASH_DASH] = ACTIONS(2540),
    [anon_sym_PLUS_PLUS] = ACTIONS(2540),
    [anon_sym_sizeof] = ACTIONS(2538),
    [anon_sym___alignof__] = ACTIONS(2538),
    [anon_sym___alignof] = ACTIONS(2538),
    [anon_sym__alignof] = ACTIONS(2538),
    [anon_sym_alignof] = ACTIONS(2538),
    [anon_sym__Alignof] = ACTIONS(2538),
    [anon_sym_offsetof] = ACTIONS(2538),
    [anon_sym__Generic] = ACTIONS(2538),
    [anon_sym_asm] = ACTIONS(2538),
    [anon_sym___asm__] = ACTIONS(2538),
    [anon_sym___asm] = ACTIONS(2538),
    [sym_number_literal] = ACTIONS(2540),
    [anon_sym_L_SQUOTE] = ACTIONS(2540),
    [anon_sym_u_SQUOTE] = ACTIONS(2540),
    [anon_sym_U_SQUOTE] = ACTIONS(2540),
    [anon_sym_u8_SQUOTE] = ACTIONS(2540),
    [anon_sym_SQUOTE] = ACTIONS(2540),
    [anon_sym_AT] = ACTIONS(2538),
    [anon_sym_DQUOTE] = ACTIONS(2540),
    [anon_sym_L_DQUOTE] = ACTIONS(2540),
    [anon_sym_u_DQUOTE] = ACTIONS(2540),
    [anon_sym_U_DQUOTE] = ACTIONS(2540),
    [anon_sym_u8_DQUOTE] = ACTIONS(2540),
    [sym_true] = ACTIONS(2538),
    [sym_false] = ACTIONS(2538),
    [anon_sym_NULL] = ACTIONS(2538),
    [anon_sym_nullptr] = ACTIONS(2538),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2538),
    [anon_sym___typeof] = ACTIONS(2538),
    [anon_sym_typeof] = ACTIONS(2538),
    [anon_sym_ATimport] = ACTIONS(2540),
    [aux_sym_preproc_undef_token1] = ACTIONS(2538),
    [anon_sym_POUND] = ACTIONS(2538),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2538),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2538),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2538),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_API_AVAILABLE] = ACTIONS(2538),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_API_DEPRECATED] = ACTIONS(2538),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym___deprecated_msg] = ACTIONS(2538),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2538),
    [anon_sym_ATprotocol] = ACTIONS(2540),
    [anon_sym_ATinterface] = ACTIONS(2540),
    [anon_sym_ATimplementation] = ACTIONS(2540),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2540),
    [anon_sym_ATtry] = ACTIONS(2540),
    [anon_sym___try] = ACTIONS(2538),
    [anon_sym_ATthrow] = ACTIONS(2540),
    [anon_sym_ATselector] = ACTIONS(2540),
    [anon_sym_ATavailable] = ACTIONS(2540),
    [anon_sym___builtin_available] = ACTIONS(2538),
    [anon_sym_va_arg] = ACTIONS(2538),
    [anon_sym_ATencode] = ACTIONS(2540),
    [anon_sym_ATsynchronized] = ACTIONS(2540),
    [anon_sym_BOOL] = ACTIONS(2538),
    [anon_sym_IMP] = ACTIONS(2538),
    [anon_sym_SEL] = ACTIONS(2538),
    [anon_sym_Class] = ACTIONS(2538),
    [anon_sym_id] = ACTIONS(2538),
  },
  [STATE(786)] = {
    [sym_identifier] = ACTIONS(2514),
    [aux_sym_preproc_include_token1] = ACTIONS(2514),
    [aux_sym_preproc_include_token2] = ACTIONS(2514),
    [aux_sym_preproc_def_token1] = ACTIONS(2514),
    [aux_sym_preproc_if_token1] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2514),
    [sym_preproc_directive] = ACTIONS(2514),
    [anon_sym_LPAREN2] = ACTIONS(2516),
    [anon_sym_BANG] = ACTIONS(2516),
    [anon_sym_TILDE] = ACTIONS(2516),
    [anon_sym_DASH] = ACTIONS(2514),
    [anon_sym_PLUS] = ACTIONS(2514),
    [anon_sym_STAR] = ACTIONS(2516),
    [anon_sym_CARET] = ACTIONS(2516),
    [anon_sym_AMP] = ACTIONS(2516),
    [anon_sym_SEMI] = ACTIONS(2516),
    [anon_sym___extension__] = ACTIONS(2514),
    [anon_sym_typedef] = ACTIONS(2514),
    [anon_sym_extern] = ACTIONS(2514),
    [anon_sym___attribute__] = ACTIONS(2514),
    [anon_sym___attribute] = ACTIONS(2514),
    [anon_sym_noreturn] = ACTIONS(2514),
    [anon_sym_LBRACK] = ACTIONS(2516),
    [anon_sym___declspec] = ACTIONS(2514),
    [anon_sym___cdecl] = ACTIONS(2514),
    [anon_sym___clrcall] = ACTIONS(2514),
    [anon_sym___stdcall] = ACTIONS(2514),
    [anon_sym___fastcall] = ACTIONS(2514),
    [anon_sym___thiscall] = ACTIONS(2514),
    [anon_sym___vectorcall] = ACTIONS(2514),
    [anon_sym_LBRACE] = ACTIONS(2516),
    [anon_sym_RBRACE] = ACTIONS(2516),
    [anon_sym_signed] = ACTIONS(2514),
    [anon_sym_unsigned] = ACTIONS(2514),
    [anon_sym_long] = ACTIONS(2514),
    [anon_sym_short] = ACTIONS(2514),
    [anon_sym_static] = ACTIONS(2514),
    [anon_sym_ATautoreleasepool] = ACTIONS(2516),
    [anon_sym_auto] = ACTIONS(2514),
    [anon_sym_register] = ACTIONS(2514),
    [anon_sym_inline] = ACTIONS(2514),
    [anon_sym___inline] = ACTIONS(2514),
    [anon_sym___inline__] = ACTIONS(2514),
    [anon_sym___forceinline] = ACTIONS(2514),
    [anon_sym_thread_local] = ACTIONS(2514),
    [anon_sym___thread] = ACTIONS(2514),
    [anon_sym_CG_EXTERN] = ACTIONS(2514),
    [anon_sym_CG_INLINE] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2514),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2514),
    [anon_sym_IBOutlet] = ACTIONS(2514),
    [anon_sym_IBInspectable] = ACTIONS(2514),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2514),
    [anon_sym_NS_INLINE] = ACTIONS(2514),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2514),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2514),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2514),
    [anon_sym_const] = ACTIONS(2514),
    [anon_sym_constexpr] = ACTIONS(2514),
    [anon_sym_volatile] = ACTIONS(2514),
    [anon_sym_restrict] = ACTIONS(2514),
    [anon_sym___restrict__] = ACTIONS(2514),
    [anon_sym__Atomic] = ACTIONS(2514),
    [anon_sym__Noreturn] = ACTIONS(2514),
    [anon_sym__Nonnull] = ACTIONS(2514),
    [anon_sym_nullable] = ACTIONS(2514),
    [anon_sym__Complex] = ACTIONS(2514),
    [anon_sym__Nullable] = ACTIONS(2514),
    [anon_sym__Nullable_result] = ACTIONS(2514),
    [anon_sym__Null_unspecified] = ACTIONS(2514),
    [anon_sym___autoreleasing] = ACTIONS(2514),
    [anon_sym___block] = ACTIONS(2514),
    [anon_sym___bridge] = ACTIONS(2514),
    [anon_sym___bridge_retained] = ACTIONS(2514),
    [anon_sym___bridge_transfer] = ACTIONS(2514),
    [anon_sym___complex] = ACTIONS(2514),
    [anon_sym___const] = ACTIONS(2514),
    [anon_sym___imag] = ACTIONS(2514),
    [anon_sym___kindof] = ACTIONS(2514),
    [anon_sym___nonnull] = ACTIONS(2514),
    [anon_sym___nullable] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2514),
    [anon_sym___real] = ACTIONS(2514),
    [anon_sym___strong] = ACTIONS(2514),
    [anon_sym___unsafe_unretained] = ACTIONS(2514),
    [anon_sym___unused] = ACTIONS(2514),
    [anon_sym___weak] = ACTIONS(2514),
    [anon_sym_alignas] = ACTIONS(2514),
    [anon_sym__Alignas] = ACTIONS(2514),
    [sym_primitive_type] = ACTIONS(2514),
    [anon_sym_enum] = ACTIONS(2514),
    [anon_sym_struct] = ACTIONS(2514),
    [anon_sym_union] = ACTIONS(2514),
    [anon_sym_if] = ACTIONS(2514),
    [anon_sym_else] = ACTIONS(2514),
    [anon_sym_switch] = ACTIONS(2514),
    [anon_sym_case] = ACTIONS(2514),
    [anon_sym_default] = ACTIONS(2514),
    [anon_sym_while] = ACTIONS(2514),
    [anon_sym_do] = ACTIONS(2514),
    [anon_sym_for] = ACTIONS(2514),
    [anon_sym_in] = ACTIONS(2514),
    [anon_sym_return] = ACTIONS(2514),
    [anon_sym_break] = ACTIONS(2514),
    [anon_sym_continue] = ACTIONS(2514),
    [anon_sym_goto] = ACTIONS(2514),
    [anon_sym_DASH_DASH] = ACTIONS(2516),
    [anon_sym_PLUS_PLUS] = ACTIONS(2516),
    [anon_sym_sizeof] = ACTIONS(2514),
    [anon_sym___alignof__] = ACTIONS(2514),
    [anon_sym___alignof] = ACTIONS(2514),
    [anon_sym__alignof] = ACTIONS(2514),
    [anon_sym_alignof] = ACTIONS(2514),
    [anon_sym__Alignof] = ACTIONS(2514),
    [anon_sym_offsetof] = ACTIONS(2514),
    [anon_sym__Generic] = ACTIONS(2514),
    [anon_sym_asm] = ACTIONS(2514),
    [anon_sym___asm__] = ACTIONS(2514),
    [anon_sym___asm] = ACTIONS(2514),
    [sym_number_literal] = ACTIONS(2516),
    [anon_sym_L_SQUOTE] = ACTIONS(2516),
    [anon_sym_u_SQUOTE] = ACTIONS(2516),
    [anon_sym_U_SQUOTE] = ACTIONS(2516),
    [anon_sym_u8_SQUOTE] = ACTIONS(2516),
    [anon_sym_SQUOTE] = ACTIONS(2516),
    [anon_sym_AT] = ACTIONS(2514),
    [anon_sym_DQUOTE] = ACTIONS(2516),
    [anon_sym_L_DQUOTE] = ACTIONS(2516),
    [anon_sym_u_DQUOTE] = ACTIONS(2516),
    [anon_sym_U_DQUOTE] = ACTIONS(2516),
    [anon_sym_u8_DQUOTE] = ACTIONS(2516),
    [sym_true] = ACTIONS(2514),
    [sym_false] = ACTIONS(2514),
    [anon_sym_NULL] = ACTIONS(2514),
    [anon_sym_nullptr] = ACTIONS(2514),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2514),
    [anon_sym___typeof] = ACTIONS(2514),
    [anon_sym_typeof] = ACTIONS(2514),
    [anon_sym_ATimport] = ACTIONS(2516),
    [aux_sym_preproc_undef_token1] = ACTIONS(2514),
    [anon_sym_POUND] = ACTIONS(2514),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2514),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2514),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2514),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_API_AVAILABLE] = ACTIONS(2514),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_API_DEPRECATED] = ACTIONS(2514),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym___deprecated_msg] = ACTIONS(2514),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2514),
    [anon_sym_ATprotocol] = ACTIONS(2516),
    [anon_sym_ATinterface] = ACTIONS(2516),
    [anon_sym_ATimplementation] = ACTIONS(2516),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2516),
    [anon_sym_ATtry] = ACTIONS(2516),
    [anon_sym___try] = ACTIONS(2514),
    [anon_sym_ATthrow] = ACTIONS(2516),
    [anon_sym_ATselector] = ACTIONS(2516),
    [anon_sym_ATavailable] = ACTIONS(2516),
    [anon_sym___builtin_available] = ACTIONS(2514),
    [anon_sym_va_arg] = ACTIONS(2514),
    [anon_sym_ATencode] = ACTIONS(2516),
    [anon_sym_ATsynchronized] = ACTIONS(2516),
    [anon_sym_BOOL] = ACTIONS(2514),
    [anon_sym_IMP] = ACTIONS(2514),
    [anon_sym_SEL] = ACTIONS(2514),
    [anon_sym_Class] = ACTIONS(2514),
    [anon_sym_id] = ACTIONS(2514),
  },
  [STATE(787)] = {
    [sym_identifier] = ACTIONS(2518),
    [aux_sym_preproc_include_token1] = ACTIONS(2518),
    [aux_sym_preproc_include_token2] = ACTIONS(2518),
    [aux_sym_preproc_def_token1] = ACTIONS(2518),
    [aux_sym_preproc_if_token1] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2518),
    [sym_preproc_directive] = ACTIONS(2518),
    [anon_sym_LPAREN2] = ACTIONS(2520),
    [anon_sym_BANG] = ACTIONS(2520),
    [anon_sym_TILDE] = ACTIONS(2520),
    [anon_sym_DASH] = ACTIONS(2518),
    [anon_sym_PLUS] = ACTIONS(2518),
    [anon_sym_STAR] = ACTIONS(2520),
    [anon_sym_CARET] = ACTIONS(2520),
    [anon_sym_AMP] = ACTIONS(2520),
    [anon_sym_SEMI] = ACTIONS(2520),
    [anon_sym___extension__] = ACTIONS(2518),
    [anon_sym_typedef] = ACTIONS(2518),
    [anon_sym_extern] = ACTIONS(2518),
    [anon_sym___attribute__] = ACTIONS(2518),
    [anon_sym___attribute] = ACTIONS(2518),
    [anon_sym_noreturn] = ACTIONS(2518),
    [anon_sym_LBRACK] = ACTIONS(2520),
    [anon_sym___declspec] = ACTIONS(2518),
    [anon_sym___cdecl] = ACTIONS(2518),
    [anon_sym___clrcall] = ACTIONS(2518),
    [anon_sym___stdcall] = ACTIONS(2518),
    [anon_sym___fastcall] = ACTIONS(2518),
    [anon_sym___thiscall] = ACTIONS(2518),
    [anon_sym___vectorcall] = ACTIONS(2518),
    [anon_sym_LBRACE] = ACTIONS(2520),
    [anon_sym_RBRACE] = ACTIONS(2520),
    [anon_sym_signed] = ACTIONS(2518),
    [anon_sym_unsigned] = ACTIONS(2518),
    [anon_sym_long] = ACTIONS(2518),
    [anon_sym_short] = ACTIONS(2518),
    [anon_sym_static] = ACTIONS(2518),
    [anon_sym_ATautoreleasepool] = ACTIONS(2520),
    [anon_sym_auto] = ACTIONS(2518),
    [anon_sym_register] = ACTIONS(2518),
    [anon_sym_inline] = ACTIONS(2518),
    [anon_sym___inline] = ACTIONS(2518),
    [anon_sym___inline__] = ACTIONS(2518),
    [anon_sym___forceinline] = ACTIONS(2518),
    [anon_sym_thread_local] = ACTIONS(2518),
    [anon_sym___thread] = ACTIONS(2518),
    [anon_sym_CG_EXTERN] = ACTIONS(2518),
    [anon_sym_CG_INLINE] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2518),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2518),
    [anon_sym_IBOutlet] = ACTIONS(2518),
    [anon_sym_IBInspectable] = ACTIONS(2518),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2518),
    [anon_sym_NS_INLINE] = ACTIONS(2518),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2518),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2518),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2518),
    [anon_sym_const] = ACTIONS(2518),
    [anon_sym_constexpr] = ACTIONS(2518),
    [anon_sym_volatile] = ACTIONS(2518),
    [anon_sym_restrict] = ACTIONS(2518),
    [anon_sym___restrict__] = ACTIONS(2518),
    [anon_sym__Atomic] = ACTIONS(2518),
    [anon_sym__Noreturn] = ACTIONS(2518),
    [anon_sym__Nonnull] = ACTIONS(2518),
    [anon_sym_nullable] = ACTIONS(2518),
    [anon_sym__Complex] = ACTIONS(2518),
    [anon_sym__Nullable] = ACTIONS(2518),
    [anon_sym__Nullable_result] = ACTIONS(2518),
    [anon_sym__Null_unspecified] = ACTIONS(2518),
    [anon_sym___autoreleasing] = ACTIONS(2518),
    [anon_sym___block] = ACTIONS(2518),
    [anon_sym___bridge] = ACTIONS(2518),
    [anon_sym___bridge_retained] = ACTIONS(2518),
    [anon_sym___bridge_transfer] = ACTIONS(2518),
    [anon_sym___complex] = ACTIONS(2518),
    [anon_sym___const] = ACTIONS(2518),
    [anon_sym___imag] = ACTIONS(2518),
    [anon_sym___kindof] = ACTIONS(2518),
    [anon_sym___nonnull] = ACTIONS(2518),
    [anon_sym___nullable] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2518),
    [anon_sym___real] = ACTIONS(2518),
    [anon_sym___strong] = ACTIONS(2518),
    [anon_sym___unsafe_unretained] = ACTIONS(2518),
    [anon_sym___unused] = ACTIONS(2518),
    [anon_sym___weak] = ACTIONS(2518),
    [anon_sym_alignas] = ACTIONS(2518),
    [anon_sym__Alignas] = ACTIONS(2518),
    [sym_primitive_type] = ACTIONS(2518),
    [anon_sym_enum] = ACTIONS(2518),
    [anon_sym_struct] = ACTIONS(2518),
    [anon_sym_union] = ACTIONS(2518),
    [anon_sym_if] = ACTIONS(2518),
    [anon_sym_else] = ACTIONS(2518),
    [anon_sym_switch] = ACTIONS(2518),
    [anon_sym_case] = ACTIONS(2518),
    [anon_sym_default] = ACTIONS(2518),
    [anon_sym_while] = ACTIONS(2518),
    [anon_sym_do] = ACTIONS(2518),
    [anon_sym_for] = ACTIONS(2518),
    [anon_sym_in] = ACTIONS(2518),
    [anon_sym_return] = ACTIONS(2518),
    [anon_sym_break] = ACTIONS(2518),
    [anon_sym_continue] = ACTIONS(2518),
    [anon_sym_goto] = ACTIONS(2518),
    [anon_sym_DASH_DASH] = ACTIONS(2520),
    [anon_sym_PLUS_PLUS] = ACTIONS(2520),
    [anon_sym_sizeof] = ACTIONS(2518),
    [anon_sym___alignof__] = ACTIONS(2518),
    [anon_sym___alignof] = ACTIONS(2518),
    [anon_sym__alignof] = ACTIONS(2518),
    [anon_sym_alignof] = ACTIONS(2518),
    [anon_sym__Alignof] = ACTIONS(2518),
    [anon_sym_offsetof] = ACTIONS(2518),
    [anon_sym__Generic] = ACTIONS(2518),
    [anon_sym_asm] = ACTIONS(2518),
    [anon_sym___asm__] = ACTIONS(2518),
    [anon_sym___asm] = ACTIONS(2518),
    [sym_number_literal] = ACTIONS(2520),
    [anon_sym_L_SQUOTE] = ACTIONS(2520),
    [anon_sym_u_SQUOTE] = ACTIONS(2520),
    [anon_sym_U_SQUOTE] = ACTIONS(2520),
    [anon_sym_u8_SQUOTE] = ACTIONS(2520),
    [anon_sym_SQUOTE] = ACTIONS(2520),
    [anon_sym_AT] = ACTIONS(2518),
    [anon_sym_DQUOTE] = ACTIONS(2520),
    [anon_sym_L_DQUOTE] = ACTIONS(2520),
    [anon_sym_u_DQUOTE] = ACTIONS(2520),
    [anon_sym_U_DQUOTE] = ACTIONS(2520),
    [anon_sym_u8_DQUOTE] = ACTIONS(2520),
    [sym_true] = ACTIONS(2518),
    [sym_false] = ACTIONS(2518),
    [anon_sym_NULL] = ACTIONS(2518),
    [anon_sym_nullptr] = ACTIONS(2518),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2518),
    [anon_sym___typeof] = ACTIONS(2518),
    [anon_sym_typeof] = ACTIONS(2518),
    [anon_sym_ATimport] = ACTIONS(2520),
    [aux_sym_preproc_undef_token1] = ACTIONS(2518),
    [anon_sym_POUND] = ACTIONS(2518),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2518),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2518),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2518),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_API_AVAILABLE] = ACTIONS(2518),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_API_DEPRECATED] = ACTIONS(2518),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym___deprecated_msg] = ACTIONS(2518),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2518),
    [anon_sym_ATprotocol] = ACTIONS(2520),
    [anon_sym_ATinterface] = ACTIONS(2520),
    [anon_sym_ATimplementation] = ACTIONS(2520),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2520),
    [anon_sym_ATtry] = ACTIONS(2520),
    [anon_sym___try] = ACTIONS(2518),
    [anon_sym_ATthrow] = ACTIONS(2520),
    [anon_sym_ATselector] = ACTIONS(2520),
    [anon_sym_ATavailable] = ACTIONS(2520),
    [anon_sym___builtin_available] = ACTIONS(2518),
    [anon_sym_va_arg] = ACTIONS(2518),
    [anon_sym_ATencode] = ACTIONS(2520),
    [anon_sym_ATsynchronized] = ACTIONS(2520),
    [anon_sym_BOOL] = ACTIONS(2518),
    [anon_sym_IMP] = ACTIONS(2518),
    [anon_sym_SEL] = ACTIONS(2518),
    [anon_sym_Class] = ACTIONS(2518),
    [anon_sym_id] = ACTIONS(2518),
  },
  [STATE(788)] = {
    [sym_identifier] = ACTIONS(2522),
    [aux_sym_preproc_include_token1] = ACTIONS(2522),
    [aux_sym_preproc_include_token2] = ACTIONS(2522),
    [aux_sym_preproc_def_token1] = ACTIONS(2522),
    [aux_sym_preproc_if_token1] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2522),
    [sym_preproc_directive] = ACTIONS(2522),
    [anon_sym_LPAREN2] = ACTIONS(2524),
    [anon_sym_BANG] = ACTIONS(2524),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_DASH] = ACTIONS(2522),
    [anon_sym_PLUS] = ACTIONS(2522),
    [anon_sym_STAR] = ACTIONS(2524),
    [anon_sym_CARET] = ACTIONS(2524),
    [anon_sym_AMP] = ACTIONS(2524),
    [anon_sym_SEMI] = ACTIONS(2524),
    [anon_sym___extension__] = ACTIONS(2522),
    [anon_sym_typedef] = ACTIONS(2522),
    [anon_sym_extern] = ACTIONS(2522),
    [anon_sym___attribute__] = ACTIONS(2522),
    [anon_sym___attribute] = ACTIONS(2522),
    [anon_sym_noreturn] = ACTIONS(2522),
    [anon_sym_LBRACK] = ACTIONS(2524),
    [anon_sym___declspec] = ACTIONS(2522),
    [anon_sym___cdecl] = ACTIONS(2522),
    [anon_sym___clrcall] = ACTIONS(2522),
    [anon_sym___stdcall] = ACTIONS(2522),
    [anon_sym___fastcall] = ACTIONS(2522),
    [anon_sym___thiscall] = ACTIONS(2522),
    [anon_sym___vectorcall] = ACTIONS(2522),
    [anon_sym_LBRACE] = ACTIONS(2524),
    [anon_sym_RBRACE] = ACTIONS(2524),
    [anon_sym_signed] = ACTIONS(2522),
    [anon_sym_unsigned] = ACTIONS(2522),
    [anon_sym_long] = ACTIONS(2522),
    [anon_sym_short] = ACTIONS(2522),
    [anon_sym_static] = ACTIONS(2522),
    [anon_sym_ATautoreleasepool] = ACTIONS(2524),
    [anon_sym_auto] = ACTIONS(2522),
    [anon_sym_register] = ACTIONS(2522),
    [anon_sym_inline] = ACTIONS(2522),
    [anon_sym___inline] = ACTIONS(2522),
    [anon_sym___inline__] = ACTIONS(2522),
    [anon_sym___forceinline] = ACTIONS(2522),
    [anon_sym_thread_local] = ACTIONS(2522),
    [anon_sym___thread] = ACTIONS(2522),
    [anon_sym_CG_EXTERN] = ACTIONS(2522),
    [anon_sym_CG_INLINE] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2522),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2522),
    [anon_sym_IBOutlet] = ACTIONS(2522),
    [anon_sym_IBInspectable] = ACTIONS(2522),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2522),
    [anon_sym_NS_INLINE] = ACTIONS(2522),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2522),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2522),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2522),
    [anon_sym_const] = ACTIONS(2522),
    [anon_sym_constexpr] = ACTIONS(2522),
    [anon_sym_volatile] = ACTIONS(2522),
    [anon_sym_restrict] = ACTIONS(2522),
    [anon_sym___restrict__] = ACTIONS(2522),
    [anon_sym__Atomic] = ACTIONS(2522),
    [anon_sym__Noreturn] = ACTIONS(2522),
    [anon_sym__Nonnull] = ACTIONS(2522),
    [anon_sym_nullable] = ACTIONS(2522),
    [anon_sym__Complex] = ACTIONS(2522),
    [anon_sym__Nullable] = ACTIONS(2522),
    [anon_sym__Nullable_result] = ACTIONS(2522),
    [anon_sym__Null_unspecified] = ACTIONS(2522),
    [anon_sym___autoreleasing] = ACTIONS(2522),
    [anon_sym___block] = ACTIONS(2522),
    [anon_sym___bridge] = ACTIONS(2522),
    [anon_sym___bridge_retained] = ACTIONS(2522),
    [anon_sym___bridge_transfer] = ACTIONS(2522),
    [anon_sym___complex] = ACTIONS(2522),
    [anon_sym___const] = ACTIONS(2522),
    [anon_sym___imag] = ACTIONS(2522),
    [anon_sym___kindof] = ACTIONS(2522),
    [anon_sym___nonnull] = ACTIONS(2522),
    [anon_sym___nullable] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2522),
    [anon_sym___real] = ACTIONS(2522),
    [anon_sym___strong] = ACTIONS(2522),
    [anon_sym___unsafe_unretained] = ACTIONS(2522),
    [anon_sym___unused] = ACTIONS(2522),
    [anon_sym___weak] = ACTIONS(2522),
    [anon_sym_alignas] = ACTIONS(2522),
    [anon_sym__Alignas] = ACTIONS(2522),
    [sym_primitive_type] = ACTIONS(2522),
    [anon_sym_enum] = ACTIONS(2522),
    [anon_sym_struct] = ACTIONS(2522),
    [anon_sym_union] = ACTIONS(2522),
    [anon_sym_if] = ACTIONS(2522),
    [anon_sym_else] = ACTIONS(2522),
    [anon_sym_switch] = ACTIONS(2522),
    [anon_sym_case] = ACTIONS(2522),
    [anon_sym_default] = ACTIONS(2522),
    [anon_sym_while] = ACTIONS(2522),
    [anon_sym_do] = ACTIONS(2522),
    [anon_sym_for] = ACTIONS(2522),
    [anon_sym_in] = ACTIONS(2522),
    [anon_sym_return] = ACTIONS(2522),
    [anon_sym_break] = ACTIONS(2522),
    [anon_sym_continue] = ACTIONS(2522),
    [anon_sym_goto] = ACTIONS(2522),
    [anon_sym_DASH_DASH] = ACTIONS(2524),
    [anon_sym_PLUS_PLUS] = ACTIONS(2524),
    [anon_sym_sizeof] = ACTIONS(2522),
    [anon_sym___alignof__] = ACTIONS(2522),
    [anon_sym___alignof] = ACTIONS(2522),
    [anon_sym__alignof] = ACTIONS(2522),
    [anon_sym_alignof] = ACTIONS(2522),
    [anon_sym__Alignof] = ACTIONS(2522),
    [anon_sym_offsetof] = ACTIONS(2522),
    [anon_sym__Generic] = ACTIONS(2522),
    [anon_sym_asm] = ACTIONS(2522),
    [anon_sym___asm__] = ACTIONS(2522),
    [anon_sym___asm] = ACTIONS(2522),
    [sym_number_literal] = ACTIONS(2524),
    [anon_sym_L_SQUOTE] = ACTIONS(2524),
    [anon_sym_u_SQUOTE] = ACTIONS(2524),
    [anon_sym_U_SQUOTE] = ACTIONS(2524),
    [anon_sym_u8_SQUOTE] = ACTIONS(2524),
    [anon_sym_SQUOTE] = ACTIONS(2524),
    [anon_sym_AT] = ACTIONS(2522),
    [anon_sym_DQUOTE] = ACTIONS(2524),
    [anon_sym_L_DQUOTE] = ACTIONS(2524),
    [anon_sym_u_DQUOTE] = ACTIONS(2524),
    [anon_sym_U_DQUOTE] = ACTIONS(2524),
    [anon_sym_u8_DQUOTE] = ACTIONS(2524),
    [sym_true] = ACTIONS(2522),
    [sym_false] = ACTIONS(2522),
    [anon_sym_NULL] = ACTIONS(2522),
    [anon_sym_nullptr] = ACTIONS(2522),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2522),
    [anon_sym___typeof] = ACTIONS(2522),
    [anon_sym_typeof] = ACTIONS(2522),
    [anon_sym_ATimport] = ACTIONS(2524),
    [aux_sym_preproc_undef_token1] = ACTIONS(2522),
    [anon_sym_POUND] = ACTIONS(2522),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2522),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2522),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2522),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_API_AVAILABLE] = ACTIONS(2522),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_API_DEPRECATED] = ACTIONS(2522),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym___deprecated_msg] = ACTIONS(2522),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2522),
    [anon_sym_ATprotocol] = ACTIONS(2524),
    [anon_sym_ATinterface] = ACTIONS(2524),
    [anon_sym_ATimplementation] = ACTIONS(2524),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2524),
    [anon_sym_ATtry] = ACTIONS(2524),
    [anon_sym___try] = ACTIONS(2522),
    [anon_sym_ATthrow] = ACTIONS(2524),
    [anon_sym_ATselector] = ACTIONS(2524),
    [anon_sym_ATavailable] = ACTIONS(2524),
    [anon_sym___builtin_available] = ACTIONS(2522),
    [anon_sym_va_arg] = ACTIONS(2522),
    [anon_sym_ATencode] = ACTIONS(2524),
    [anon_sym_ATsynchronized] = ACTIONS(2524),
    [anon_sym_BOOL] = ACTIONS(2522),
    [anon_sym_IMP] = ACTIONS(2522),
    [anon_sym_SEL] = ACTIONS(2522),
    [anon_sym_Class] = ACTIONS(2522),
    [anon_sym_id] = ACTIONS(2522),
  },
  [STATE(789)] = {
    [ts_builtin_sym_end] = ACTIONS(2440),
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(790)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_RBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(791)] = {
    [ts_builtin_sym_end] = ACTIONS(2544),
    [sym_identifier] = ACTIONS(2542),
    [aux_sym_preproc_include_token1] = ACTIONS(2542),
    [aux_sym_preproc_include_token2] = ACTIONS(2542),
    [aux_sym_preproc_def_token1] = ACTIONS(2542),
    [aux_sym_preproc_if_token1] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2542),
    [sym_preproc_directive] = ACTIONS(2542),
    [anon_sym_LPAREN2] = ACTIONS(2544),
    [anon_sym_BANG] = ACTIONS(2544),
    [anon_sym_TILDE] = ACTIONS(2544),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_PLUS] = ACTIONS(2542),
    [anon_sym_STAR] = ACTIONS(2544),
    [anon_sym_CARET] = ACTIONS(2544),
    [anon_sym_AMP] = ACTIONS(2544),
    [anon_sym_SEMI] = ACTIONS(2544),
    [anon_sym___extension__] = ACTIONS(2542),
    [anon_sym_typedef] = ACTIONS(2542),
    [anon_sym_extern] = ACTIONS(2542),
    [anon_sym___attribute__] = ACTIONS(2542),
    [anon_sym___attribute] = ACTIONS(2542),
    [anon_sym_noreturn] = ACTIONS(2542),
    [anon_sym_LBRACK] = ACTIONS(2544),
    [anon_sym___declspec] = ACTIONS(2542),
    [anon_sym___cdecl] = ACTIONS(2542),
    [anon_sym___clrcall] = ACTIONS(2542),
    [anon_sym___stdcall] = ACTIONS(2542),
    [anon_sym___fastcall] = ACTIONS(2542),
    [anon_sym___thiscall] = ACTIONS(2542),
    [anon_sym___vectorcall] = ACTIONS(2542),
    [anon_sym_LBRACE] = ACTIONS(2544),
    [anon_sym_signed] = ACTIONS(2542),
    [anon_sym_unsigned] = ACTIONS(2542),
    [anon_sym_long] = ACTIONS(2542),
    [anon_sym_short] = ACTIONS(2542),
    [anon_sym_static] = ACTIONS(2542),
    [anon_sym_ATautoreleasepool] = ACTIONS(2544),
    [anon_sym_auto] = ACTIONS(2542),
    [anon_sym_register] = ACTIONS(2542),
    [anon_sym_inline] = ACTIONS(2542),
    [anon_sym___inline] = ACTIONS(2542),
    [anon_sym___inline__] = ACTIONS(2542),
    [anon_sym___forceinline] = ACTIONS(2542),
    [anon_sym_thread_local] = ACTIONS(2542),
    [anon_sym___thread] = ACTIONS(2542),
    [anon_sym_CG_EXTERN] = ACTIONS(2542),
    [anon_sym_CG_INLINE] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2542),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2542),
    [anon_sym_IBOutlet] = ACTIONS(2542),
    [anon_sym_IBInspectable] = ACTIONS(2542),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2542),
    [anon_sym_NS_INLINE] = ACTIONS(2542),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2542),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2542),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2542),
    [anon_sym_const] = ACTIONS(2542),
    [anon_sym_constexpr] = ACTIONS(2542),
    [anon_sym_volatile] = ACTIONS(2542),
    [anon_sym_restrict] = ACTIONS(2542),
    [anon_sym___restrict__] = ACTIONS(2542),
    [anon_sym__Atomic] = ACTIONS(2542),
    [anon_sym__Noreturn] = ACTIONS(2542),
    [anon_sym__Nonnull] = ACTIONS(2542),
    [anon_sym_nullable] = ACTIONS(2542),
    [anon_sym__Complex] = ACTIONS(2542),
    [anon_sym__Nullable] = ACTIONS(2542),
    [anon_sym__Nullable_result] = ACTIONS(2542),
    [anon_sym__Null_unspecified] = ACTIONS(2542),
    [anon_sym___autoreleasing] = ACTIONS(2542),
    [anon_sym___block] = ACTIONS(2542),
    [anon_sym___bridge] = ACTIONS(2542),
    [anon_sym___bridge_retained] = ACTIONS(2542),
    [anon_sym___bridge_transfer] = ACTIONS(2542),
    [anon_sym___complex] = ACTIONS(2542),
    [anon_sym___const] = ACTIONS(2542),
    [anon_sym___imag] = ACTIONS(2542),
    [anon_sym___kindof] = ACTIONS(2542),
    [anon_sym___nonnull] = ACTIONS(2542),
    [anon_sym___nullable] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2542),
    [anon_sym___real] = ACTIONS(2542),
    [anon_sym___strong] = ACTIONS(2542),
    [anon_sym___unsafe_unretained] = ACTIONS(2542),
    [anon_sym___unused] = ACTIONS(2542),
    [anon_sym___weak] = ACTIONS(2542),
    [anon_sym_alignas] = ACTIONS(2542),
    [anon_sym__Alignas] = ACTIONS(2542),
    [sym_primitive_type] = ACTIONS(2542),
    [anon_sym_enum] = ACTIONS(2542),
    [anon_sym_struct] = ACTIONS(2542),
    [anon_sym_union] = ACTIONS(2542),
    [anon_sym_if] = ACTIONS(2542),
    [anon_sym_else] = ACTIONS(2542),
    [anon_sym_switch] = ACTIONS(2542),
    [anon_sym_case] = ACTIONS(2542),
    [anon_sym_default] = ACTIONS(2542),
    [anon_sym_while] = ACTIONS(2542),
    [anon_sym_do] = ACTIONS(2542),
    [anon_sym_for] = ACTIONS(2542),
    [anon_sym_in] = ACTIONS(2542),
    [anon_sym_return] = ACTIONS(2542),
    [anon_sym_break] = ACTIONS(2542),
    [anon_sym_continue] = ACTIONS(2542),
    [anon_sym_goto] = ACTIONS(2542),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_sizeof] = ACTIONS(2542),
    [anon_sym___alignof__] = ACTIONS(2542),
    [anon_sym___alignof] = ACTIONS(2542),
    [anon_sym__alignof] = ACTIONS(2542),
    [anon_sym_alignof] = ACTIONS(2542),
    [anon_sym__Alignof] = ACTIONS(2542),
    [anon_sym_offsetof] = ACTIONS(2542),
    [anon_sym__Generic] = ACTIONS(2542),
    [anon_sym_asm] = ACTIONS(2542),
    [anon_sym___asm__] = ACTIONS(2542),
    [anon_sym___asm] = ACTIONS(2542),
    [sym_number_literal] = ACTIONS(2544),
    [anon_sym_L_SQUOTE] = ACTIONS(2544),
    [anon_sym_u_SQUOTE] = ACTIONS(2544),
    [anon_sym_U_SQUOTE] = ACTIONS(2544),
    [anon_sym_u8_SQUOTE] = ACTIONS(2544),
    [anon_sym_SQUOTE] = ACTIONS(2544),
    [anon_sym_AT] = ACTIONS(2542),
    [anon_sym_DQUOTE] = ACTIONS(2544),
    [anon_sym_L_DQUOTE] = ACTIONS(2544),
    [anon_sym_u_DQUOTE] = ACTIONS(2544),
    [anon_sym_U_DQUOTE] = ACTIONS(2544),
    [anon_sym_u8_DQUOTE] = ACTIONS(2544),
    [sym_true] = ACTIONS(2542),
    [sym_false] = ACTIONS(2542),
    [anon_sym_NULL] = ACTIONS(2542),
    [anon_sym_nullptr] = ACTIONS(2542),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2542),
    [anon_sym___typeof] = ACTIONS(2542),
    [anon_sym_typeof] = ACTIONS(2542),
    [anon_sym_ATimport] = ACTIONS(2544),
    [aux_sym_preproc_undef_token1] = ACTIONS(2542),
    [anon_sym_POUND] = ACTIONS(2542),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2542),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2542),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2542),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_API_AVAILABLE] = ACTIONS(2542),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_API_DEPRECATED] = ACTIONS(2542),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym___deprecated_msg] = ACTIONS(2542),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2542),
    [anon_sym_ATprotocol] = ACTIONS(2544),
    [anon_sym_ATinterface] = ACTIONS(2544),
    [anon_sym_ATimplementation] = ACTIONS(2544),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2544),
    [anon_sym_ATtry] = ACTIONS(2544),
    [anon_sym___try] = ACTIONS(2542),
    [anon_sym_ATthrow] = ACTIONS(2544),
    [anon_sym_ATselector] = ACTIONS(2544),
    [anon_sym_ATavailable] = ACTIONS(2544),
    [anon_sym___builtin_available] = ACTIONS(2542),
    [anon_sym_va_arg] = ACTIONS(2542),
    [anon_sym_ATencode] = ACTIONS(2544),
    [anon_sym_ATsynchronized] = ACTIONS(2544),
    [anon_sym_BOOL] = ACTIONS(2542),
    [anon_sym_IMP] = ACTIONS(2542),
    [anon_sym_SEL] = ACTIONS(2542),
    [anon_sym_Class] = ACTIONS(2542),
    [anon_sym_id] = ACTIONS(2542),
  },
  [STATE(792)] = {
    [ts_builtin_sym_end] = ACTIONS(2552),
    [sym_identifier] = ACTIONS(2550),
    [aux_sym_preproc_include_token1] = ACTIONS(2550),
    [aux_sym_preproc_include_token2] = ACTIONS(2550),
    [aux_sym_preproc_def_token1] = ACTIONS(2550),
    [aux_sym_preproc_if_token1] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2550),
    [sym_preproc_directive] = ACTIONS(2550),
    [anon_sym_LPAREN2] = ACTIONS(2552),
    [anon_sym_BANG] = ACTIONS(2552),
    [anon_sym_TILDE] = ACTIONS(2552),
    [anon_sym_DASH] = ACTIONS(2550),
    [anon_sym_PLUS] = ACTIONS(2550),
    [anon_sym_STAR] = ACTIONS(2552),
    [anon_sym_CARET] = ACTIONS(2552),
    [anon_sym_AMP] = ACTIONS(2552),
    [anon_sym_SEMI] = ACTIONS(2552),
    [anon_sym___extension__] = ACTIONS(2550),
    [anon_sym_typedef] = ACTIONS(2550),
    [anon_sym_extern] = ACTIONS(2550),
    [anon_sym___attribute__] = ACTIONS(2550),
    [anon_sym___attribute] = ACTIONS(2550),
    [anon_sym_noreturn] = ACTIONS(2550),
    [anon_sym_LBRACK] = ACTIONS(2552),
    [anon_sym___declspec] = ACTIONS(2550),
    [anon_sym___cdecl] = ACTIONS(2550),
    [anon_sym___clrcall] = ACTIONS(2550),
    [anon_sym___stdcall] = ACTIONS(2550),
    [anon_sym___fastcall] = ACTIONS(2550),
    [anon_sym___thiscall] = ACTIONS(2550),
    [anon_sym___vectorcall] = ACTIONS(2550),
    [anon_sym_LBRACE] = ACTIONS(2552),
    [anon_sym_signed] = ACTIONS(2550),
    [anon_sym_unsigned] = ACTIONS(2550),
    [anon_sym_long] = ACTIONS(2550),
    [anon_sym_short] = ACTIONS(2550),
    [anon_sym_static] = ACTIONS(2550),
    [anon_sym_ATautoreleasepool] = ACTIONS(2552),
    [anon_sym_auto] = ACTIONS(2550),
    [anon_sym_register] = ACTIONS(2550),
    [anon_sym_inline] = ACTIONS(2550),
    [anon_sym___inline] = ACTIONS(2550),
    [anon_sym___inline__] = ACTIONS(2550),
    [anon_sym___forceinline] = ACTIONS(2550),
    [anon_sym_thread_local] = ACTIONS(2550),
    [anon_sym___thread] = ACTIONS(2550),
    [anon_sym_CG_EXTERN] = ACTIONS(2550),
    [anon_sym_CG_INLINE] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2550),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2550),
    [anon_sym_IBOutlet] = ACTIONS(2550),
    [anon_sym_IBInspectable] = ACTIONS(2550),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2550),
    [anon_sym_NS_INLINE] = ACTIONS(2550),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2550),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2550),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2550),
    [anon_sym_const] = ACTIONS(2550),
    [anon_sym_constexpr] = ACTIONS(2550),
    [anon_sym_volatile] = ACTIONS(2550),
    [anon_sym_restrict] = ACTIONS(2550),
    [anon_sym___restrict__] = ACTIONS(2550),
    [anon_sym__Atomic] = ACTIONS(2550),
    [anon_sym__Noreturn] = ACTIONS(2550),
    [anon_sym__Nonnull] = ACTIONS(2550),
    [anon_sym_nullable] = ACTIONS(2550),
    [anon_sym__Complex] = ACTIONS(2550),
    [anon_sym__Nullable] = ACTIONS(2550),
    [anon_sym__Nullable_result] = ACTIONS(2550),
    [anon_sym__Null_unspecified] = ACTIONS(2550),
    [anon_sym___autoreleasing] = ACTIONS(2550),
    [anon_sym___block] = ACTIONS(2550),
    [anon_sym___bridge] = ACTIONS(2550),
    [anon_sym___bridge_retained] = ACTIONS(2550),
    [anon_sym___bridge_transfer] = ACTIONS(2550),
    [anon_sym___complex] = ACTIONS(2550),
    [anon_sym___const] = ACTIONS(2550),
    [anon_sym___imag] = ACTIONS(2550),
    [anon_sym___kindof] = ACTIONS(2550),
    [anon_sym___nonnull] = ACTIONS(2550),
    [anon_sym___nullable] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2550),
    [anon_sym___real] = ACTIONS(2550),
    [anon_sym___strong] = ACTIONS(2550),
    [anon_sym___unsafe_unretained] = ACTIONS(2550),
    [anon_sym___unused] = ACTIONS(2550),
    [anon_sym___weak] = ACTIONS(2550),
    [anon_sym_alignas] = ACTIONS(2550),
    [anon_sym__Alignas] = ACTIONS(2550),
    [sym_primitive_type] = ACTIONS(2550),
    [anon_sym_enum] = ACTIONS(2550),
    [anon_sym_struct] = ACTIONS(2550),
    [anon_sym_union] = ACTIONS(2550),
    [anon_sym_if] = ACTIONS(2550),
    [anon_sym_else] = ACTIONS(2550),
    [anon_sym_switch] = ACTIONS(2550),
    [anon_sym_case] = ACTIONS(2550),
    [anon_sym_default] = ACTIONS(2550),
    [anon_sym_while] = ACTIONS(2550),
    [anon_sym_do] = ACTIONS(2550),
    [anon_sym_for] = ACTIONS(2550),
    [anon_sym_in] = ACTIONS(2550),
    [anon_sym_return] = ACTIONS(2550),
    [anon_sym_break] = ACTIONS(2550),
    [anon_sym_continue] = ACTIONS(2550),
    [anon_sym_goto] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2552),
    [anon_sym_PLUS_PLUS] = ACTIONS(2552),
    [anon_sym_sizeof] = ACTIONS(2550),
    [anon_sym___alignof__] = ACTIONS(2550),
    [anon_sym___alignof] = ACTIONS(2550),
    [anon_sym__alignof] = ACTIONS(2550),
    [anon_sym_alignof] = ACTIONS(2550),
    [anon_sym__Alignof] = ACTIONS(2550),
    [anon_sym_offsetof] = ACTIONS(2550),
    [anon_sym__Generic] = ACTIONS(2550),
    [anon_sym_asm] = ACTIONS(2550),
    [anon_sym___asm__] = ACTIONS(2550),
    [anon_sym___asm] = ACTIONS(2550),
    [sym_number_literal] = ACTIONS(2552),
    [anon_sym_L_SQUOTE] = ACTIONS(2552),
    [anon_sym_u_SQUOTE] = ACTIONS(2552),
    [anon_sym_U_SQUOTE] = ACTIONS(2552),
    [anon_sym_u8_SQUOTE] = ACTIONS(2552),
    [anon_sym_SQUOTE] = ACTIONS(2552),
    [anon_sym_AT] = ACTIONS(2550),
    [anon_sym_DQUOTE] = ACTIONS(2552),
    [anon_sym_L_DQUOTE] = ACTIONS(2552),
    [anon_sym_u_DQUOTE] = ACTIONS(2552),
    [anon_sym_U_DQUOTE] = ACTIONS(2552),
    [anon_sym_u8_DQUOTE] = ACTIONS(2552),
    [sym_true] = ACTIONS(2550),
    [sym_false] = ACTIONS(2550),
    [anon_sym_NULL] = ACTIONS(2550),
    [anon_sym_nullptr] = ACTIONS(2550),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2550),
    [anon_sym___typeof] = ACTIONS(2550),
    [anon_sym_typeof] = ACTIONS(2550),
    [anon_sym_ATimport] = ACTIONS(2552),
    [aux_sym_preproc_undef_token1] = ACTIONS(2550),
    [anon_sym_POUND] = ACTIONS(2550),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2550),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2550),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2550),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_API_AVAILABLE] = ACTIONS(2550),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_API_DEPRECATED] = ACTIONS(2550),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym___deprecated_msg] = ACTIONS(2550),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2550),
    [anon_sym_ATprotocol] = ACTIONS(2552),
    [anon_sym_ATinterface] = ACTIONS(2552),
    [anon_sym_ATimplementation] = ACTIONS(2552),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2552),
    [anon_sym_ATtry] = ACTIONS(2552),
    [anon_sym___try] = ACTIONS(2550),
    [anon_sym_ATthrow] = ACTIONS(2552),
    [anon_sym_ATselector] = ACTIONS(2552),
    [anon_sym_ATavailable] = ACTIONS(2552),
    [anon_sym___builtin_available] = ACTIONS(2550),
    [anon_sym_va_arg] = ACTIONS(2550),
    [anon_sym_ATencode] = ACTIONS(2552),
    [anon_sym_ATsynchronized] = ACTIONS(2552),
    [anon_sym_BOOL] = ACTIONS(2550),
    [anon_sym_IMP] = ACTIONS(2550),
    [anon_sym_SEL] = ACTIONS(2550),
    [anon_sym_Class] = ACTIONS(2550),
    [anon_sym_id] = ACTIONS(2550),
  },
  [STATE(793)] = {
    [ts_builtin_sym_end] = ACTIONS(2444),
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(794)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(795)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(796)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(797)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(798)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(799)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(800)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(801)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(802)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(803)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(804)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(805)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(806)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(807)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(808)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(809)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(810)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(811)] = {
    [sym_identifier] = ACTIONS(2538),
    [aux_sym_preproc_include_token1] = ACTIONS(2538),
    [aux_sym_preproc_include_token2] = ACTIONS(2538),
    [aux_sym_preproc_def_token1] = ACTIONS(2538),
    [aux_sym_preproc_if_token1] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2538),
    [sym_preproc_directive] = ACTIONS(2538),
    [anon_sym_LPAREN2] = ACTIONS(2540),
    [anon_sym_BANG] = ACTIONS(2540),
    [anon_sym_TILDE] = ACTIONS(2540),
    [anon_sym_DASH] = ACTIONS(2538),
    [anon_sym_PLUS] = ACTIONS(2538),
    [anon_sym_STAR] = ACTIONS(2540),
    [anon_sym_CARET] = ACTIONS(2540),
    [anon_sym_AMP] = ACTIONS(2540),
    [anon_sym_SEMI] = ACTIONS(2540),
    [anon_sym___extension__] = ACTIONS(2538),
    [anon_sym_typedef] = ACTIONS(2538),
    [anon_sym_extern] = ACTIONS(2538),
    [anon_sym___attribute__] = ACTIONS(2538),
    [anon_sym___attribute] = ACTIONS(2538),
    [anon_sym_noreturn] = ACTIONS(2538),
    [anon_sym_LBRACK] = ACTIONS(2540),
    [anon_sym___declspec] = ACTIONS(2538),
    [anon_sym___cdecl] = ACTIONS(2538),
    [anon_sym___clrcall] = ACTIONS(2538),
    [anon_sym___stdcall] = ACTIONS(2538),
    [anon_sym___fastcall] = ACTIONS(2538),
    [anon_sym___thiscall] = ACTIONS(2538),
    [anon_sym___vectorcall] = ACTIONS(2538),
    [anon_sym_LBRACE] = ACTIONS(2540),
    [anon_sym_RBRACE] = ACTIONS(2540),
    [anon_sym_signed] = ACTIONS(2538),
    [anon_sym_unsigned] = ACTIONS(2538),
    [anon_sym_long] = ACTIONS(2538),
    [anon_sym_short] = ACTIONS(2538),
    [anon_sym_static] = ACTIONS(2538),
    [anon_sym_ATautoreleasepool] = ACTIONS(2540),
    [anon_sym_auto] = ACTIONS(2538),
    [anon_sym_register] = ACTIONS(2538),
    [anon_sym_inline] = ACTIONS(2538),
    [anon_sym___inline] = ACTIONS(2538),
    [anon_sym___inline__] = ACTIONS(2538),
    [anon_sym___forceinline] = ACTIONS(2538),
    [anon_sym_thread_local] = ACTIONS(2538),
    [anon_sym___thread] = ACTIONS(2538),
    [anon_sym_CG_EXTERN] = ACTIONS(2538),
    [anon_sym_CG_INLINE] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2538),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2538),
    [anon_sym_IBOutlet] = ACTIONS(2538),
    [anon_sym_IBInspectable] = ACTIONS(2538),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2538),
    [anon_sym_NS_INLINE] = ACTIONS(2538),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2538),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2538),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2538),
    [anon_sym_const] = ACTIONS(2538),
    [anon_sym_constexpr] = ACTIONS(2538),
    [anon_sym_volatile] = ACTIONS(2538),
    [anon_sym_restrict] = ACTIONS(2538),
    [anon_sym___restrict__] = ACTIONS(2538),
    [anon_sym__Atomic] = ACTIONS(2538),
    [anon_sym__Noreturn] = ACTIONS(2538),
    [anon_sym__Nonnull] = ACTIONS(2538),
    [anon_sym_nullable] = ACTIONS(2538),
    [anon_sym__Complex] = ACTIONS(2538),
    [anon_sym__Nullable] = ACTIONS(2538),
    [anon_sym__Nullable_result] = ACTIONS(2538),
    [anon_sym__Null_unspecified] = ACTIONS(2538),
    [anon_sym___autoreleasing] = ACTIONS(2538),
    [anon_sym___block] = ACTIONS(2538),
    [anon_sym___bridge] = ACTIONS(2538),
    [anon_sym___bridge_retained] = ACTIONS(2538),
    [anon_sym___bridge_transfer] = ACTIONS(2538),
    [anon_sym___complex] = ACTIONS(2538),
    [anon_sym___const] = ACTIONS(2538),
    [anon_sym___imag] = ACTIONS(2538),
    [anon_sym___kindof] = ACTIONS(2538),
    [anon_sym___nonnull] = ACTIONS(2538),
    [anon_sym___nullable] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2538),
    [anon_sym___real] = ACTIONS(2538),
    [anon_sym___strong] = ACTIONS(2538),
    [anon_sym___unsafe_unretained] = ACTIONS(2538),
    [anon_sym___unused] = ACTIONS(2538),
    [anon_sym___weak] = ACTIONS(2538),
    [anon_sym_alignas] = ACTIONS(2538),
    [anon_sym__Alignas] = ACTIONS(2538),
    [sym_primitive_type] = ACTIONS(2538),
    [anon_sym_enum] = ACTIONS(2538),
    [anon_sym_struct] = ACTIONS(2538),
    [anon_sym_union] = ACTIONS(2538),
    [anon_sym_if] = ACTIONS(2538),
    [anon_sym_else] = ACTIONS(2538),
    [anon_sym_switch] = ACTIONS(2538),
    [anon_sym_case] = ACTIONS(2538),
    [anon_sym_default] = ACTIONS(2538),
    [anon_sym_while] = ACTIONS(2538),
    [anon_sym_do] = ACTIONS(2538),
    [anon_sym_for] = ACTIONS(2538),
    [anon_sym_in] = ACTIONS(2538),
    [anon_sym_return] = ACTIONS(2538),
    [anon_sym_break] = ACTIONS(2538),
    [anon_sym_continue] = ACTIONS(2538),
    [anon_sym_goto] = ACTIONS(2538),
    [anon_sym_DASH_DASH] = ACTIONS(2540),
    [anon_sym_PLUS_PLUS] = ACTIONS(2540),
    [anon_sym_sizeof] = ACTIONS(2538),
    [anon_sym___alignof__] = ACTIONS(2538),
    [anon_sym___alignof] = ACTIONS(2538),
    [anon_sym__alignof] = ACTIONS(2538),
    [anon_sym_alignof] = ACTIONS(2538),
    [anon_sym__Alignof] = ACTIONS(2538),
    [anon_sym_offsetof] = ACTIONS(2538),
    [anon_sym__Generic] = ACTIONS(2538),
    [anon_sym_asm] = ACTIONS(2538),
    [anon_sym___asm__] = ACTIONS(2538),
    [anon_sym___asm] = ACTIONS(2538),
    [sym_number_literal] = ACTIONS(2540),
    [anon_sym_L_SQUOTE] = ACTIONS(2540),
    [anon_sym_u_SQUOTE] = ACTIONS(2540),
    [anon_sym_U_SQUOTE] = ACTIONS(2540),
    [anon_sym_u8_SQUOTE] = ACTIONS(2540),
    [anon_sym_SQUOTE] = ACTIONS(2540),
    [anon_sym_AT] = ACTIONS(2538),
    [anon_sym_DQUOTE] = ACTIONS(2540),
    [anon_sym_L_DQUOTE] = ACTIONS(2540),
    [anon_sym_u_DQUOTE] = ACTIONS(2540),
    [anon_sym_U_DQUOTE] = ACTIONS(2540),
    [anon_sym_u8_DQUOTE] = ACTIONS(2540),
    [sym_true] = ACTIONS(2538),
    [sym_false] = ACTIONS(2538),
    [anon_sym_NULL] = ACTIONS(2538),
    [anon_sym_nullptr] = ACTIONS(2538),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2538),
    [anon_sym___typeof] = ACTIONS(2538),
    [anon_sym_typeof] = ACTIONS(2538),
    [anon_sym_ATimport] = ACTIONS(2540),
    [aux_sym_preproc_undef_token1] = ACTIONS(2538),
    [anon_sym_POUND] = ACTIONS(2538),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2538),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2538),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2538),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_API_AVAILABLE] = ACTIONS(2538),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_API_DEPRECATED] = ACTIONS(2538),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym___deprecated_msg] = ACTIONS(2538),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2538),
    [anon_sym_ATprotocol] = ACTIONS(2540),
    [anon_sym_ATinterface] = ACTIONS(2540),
    [anon_sym_ATimplementation] = ACTIONS(2540),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2540),
    [anon_sym_ATtry] = ACTIONS(2540),
    [anon_sym___try] = ACTIONS(2538),
    [anon_sym_ATthrow] = ACTIONS(2540),
    [anon_sym_ATselector] = ACTIONS(2540),
    [anon_sym_ATavailable] = ACTIONS(2540),
    [anon_sym___builtin_available] = ACTIONS(2538),
    [anon_sym_va_arg] = ACTIONS(2538),
    [anon_sym_ATencode] = ACTIONS(2540),
    [anon_sym_ATsynchronized] = ACTIONS(2540),
    [anon_sym_BOOL] = ACTIONS(2538),
    [anon_sym_IMP] = ACTIONS(2538),
    [anon_sym_SEL] = ACTIONS(2538),
    [anon_sym_Class] = ACTIONS(2538),
    [anon_sym_id] = ACTIONS(2538),
  },
  [STATE(812)] = {
    [ts_builtin_sym_end] = ACTIONS(2448),
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(813)] = {
    [ts_builtin_sym_end] = ACTIONS(2448),
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(814)] = {
    [ts_builtin_sym_end] = ACTIONS(2440),
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(815)] = {
    [sym_identifier] = ACTIONS(2542),
    [aux_sym_preproc_include_token1] = ACTIONS(2542),
    [aux_sym_preproc_include_token2] = ACTIONS(2542),
    [aux_sym_preproc_def_token1] = ACTIONS(2542),
    [aux_sym_preproc_if_token1] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2542),
    [sym_preproc_directive] = ACTIONS(2542),
    [anon_sym_LPAREN2] = ACTIONS(2544),
    [anon_sym_BANG] = ACTIONS(2544),
    [anon_sym_TILDE] = ACTIONS(2544),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_PLUS] = ACTIONS(2542),
    [anon_sym_STAR] = ACTIONS(2544),
    [anon_sym_CARET] = ACTIONS(2544),
    [anon_sym_AMP] = ACTIONS(2544),
    [anon_sym_SEMI] = ACTIONS(2544),
    [anon_sym___extension__] = ACTIONS(2542),
    [anon_sym_typedef] = ACTIONS(2542),
    [anon_sym_extern] = ACTIONS(2542),
    [anon_sym___attribute__] = ACTIONS(2542),
    [anon_sym___attribute] = ACTIONS(2542),
    [anon_sym_noreturn] = ACTIONS(2542),
    [anon_sym_LBRACK] = ACTIONS(2544),
    [anon_sym___declspec] = ACTIONS(2542),
    [anon_sym___cdecl] = ACTIONS(2542),
    [anon_sym___clrcall] = ACTIONS(2542),
    [anon_sym___stdcall] = ACTIONS(2542),
    [anon_sym___fastcall] = ACTIONS(2542),
    [anon_sym___thiscall] = ACTIONS(2542),
    [anon_sym___vectorcall] = ACTIONS(2542),
    [anon_sym_LBRACE] = ACTIONS(2544),
    [anon_sym_RBRACE] = ACTIONS(2544),
    [anon_sym_signed] = ACTIONS(2542),
    [anon_sym_unsigned] = ACTIONS(2542),
    [anon_sym_long] = ACTIONS(2542),
    [anon_sym_short] = ACTIONS(2542),
    [anon_sym_static] = ACTIONS(2542),
    [anon_sym_ATautoreleasepool] = ACTIONS(2544),
    [anon_sym_auto] = ACTIONS(2542),
    [anon_sym_register] = ACTIONS(2542),
    [anon_sym_inline] = ACTIONS(2542),
    [anon_sym___inline] = ACTIONS(2542),
    [anon_sym___inline__] = ACTIONS(2542),
    [anon_sym___forceinline] = ACTIONS(2542),
    [anon_sym_thread_local] = ACTIONS(2542),
    [anon_sym___thread] = ACTIONS(2542),
    [anon_sym_CG_EXTERN] = ACTIONS(2542),
    [anon_sym_CG_INLINE] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2542),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2542),
    [anon_sym_IBOutlet] = ACTIONS(2542),
    [anon_sym_IBInspectable] = ACTIONS(2542),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2542),
    [anon_sym_NS_INLINE] = ACTIONS(2542),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2542),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2542),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2542),
    [anon_sym_const] = ACTIONS(2542),
    [anon_sym_constexpr] = ACTIONS(2542),
    [anon_sym_volatile] = ACTIONS(2542),
    [anon_sym_restrict] = ACTIONS(2542),
    [anon_sym___restrict__] = ACTIONS(2542),
    [anon_sym__Atomic] = ACTIONS(2542),
    [anon_sym__Noreturn] = ACTIONS(2542),
    [anon_sym__Nonnull] = ACTIONS(2542),
    [anon_sym_nullable] = ACTIONS(2542),
    [anon_sym__Complex] = ACTIONS(2542),
    [anon_sym__Nullable] = ACTIONS(2542),
    [anon_sym__Nullable_result] = ACTIONS(2542),
    [anon_sym__Null_unspecified] = ACTIONS(2542),
    [anon_sym___autoreleasing] = ACTIONS(2542),
    [anon_sym___block] = ACTIONS(2542),
    [anon_sym___bridge] = ACTIONS(2542),
    [anon_sym___bridge_retained] = ACTIONS(2542),
    [anon_sym___bridge_transfer] = ACTIONS(2542),
    [anon_sym___complex] = ACTIONS(2542),
    [anon_sym___const] = ACTIONS(2542),
    [anon_sym___imag] = ACTIONS(2542),
    [anon_sym___kindof] = ACTIONS(2542),
    [anon_sym___nonnull] = ACTIONS(2542),
    [anon_sym___nullable] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2542),
    [anon_sym___real] = ACTIONS(2542),
    [anon_sym___strong] = ACTIONS(2542),
    [anon_sym___unsafe_unretained] = ACTIONS(2542),
    [anon_sym___unused] = ACTIONS(2542),
    [anon_sym___weak] = ACTIONS(2542),
    [anon_sym_alignas] = ACTIONS(2542),
    [anon_sym__Alignas] = ACTIONS(2542),
    [sym_primitive_type] = ACTIONS(2542),
    [anon_sym_enum] = ACTIONS(2542),
    [anon_sym_struct] = ACTIONS(2542),
    [anon_sym_union] = ACTIONS(2542),
    [anon_sym_if] = ACTIONS(2542),
    [anon_sym_else] = ACTIONS(2542),
    [anon_sym_switch] = ACTIONS(2542),
    [anon_sym_case] = ACTIONS(2542),
    [anon_sym_default] = ACTIONS(2542),
    [anon_sym_while] = ACTIONS(2542),
    [anon_sym_do] = ACTIONS(2542),
    [anon_sym_for] = ACTIONS(2542),
    [anon_sym_in] = ACTIONS(2542),
    [anon_sym_return] = ACTIONS(2542),
    [anon_sym_break] = ACTIONS(2542),
    [anon_sym_continue] = ACTIONS(2542),
    [anon_sym_goto] = ACTIONS(2542),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_sizeof] = ACTIONS(2542),
    [anon_sym___alignof__] = ACTIONS(2542),
    [anon_sym___alignof] = ACTIONS(2542),
    [anon_sym__alignof] = ACTIONS(2542),
    [anon_sym_alignof] = ACTIONS(2542),
    [anon_sym__Alignof] = ACTIONS(2542),
    [anon_sym_offsetof] = ACTIONS(2542),
    [anon_sym__Generic] = ACTIONS(2542),
    [anon_sym_asm] = ACTIONS(2542),
    [anon_sym___asm__] = ACTIONS(2542),
    [anon_sym___asm] = ACTIONS(2542),
    [sym_number_literal] = ACTIONS(2544),
    [anon_sym_L_SQUOTE] = ACTIONS(2544),
    [anon_sym_u_SQUOTE] = ACTIONS(2544),
    [anon_sym_U_SQUOTE] = ACTIONS(2544),
    [anon_sym_u8_SQUOTE] = ACTIONS(2544),
    [anon_sym_SQUOTE] = ACTIONS(2544),
    [anon_sym_AT] = ACTIONS(2542),
    [anon_sym_DQUOTE] = ACTIONS(2544),
    [anon_sym_L_DQUOTE] = ACTIONS(2544),
    [anon_sym_u_DQUOTE] = ACTIONS(2544),
    [anon_sym_U_DQUOTE] = ACTIONS(2544),
    [anon_sym_u8_DQUOTE] = ACTIONS(2544),
    [sym_true] = ACTIONS(2542),
    [sym_false] = ACTIONS(2542),
    [anon_sym_NULL] = ACTIONS(2542),
    [anon_sym_nullptr] = ACTIONS(2542),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2542),
    [anon_sym___typeof] = ACTIONS(2542),
    [anon_sym_typeof] = ACTIONS(2542),
    [anon_sym_ATimport] = ACTIONS(2544),
    [aux_sym_preproc_undef_token1] = ACTIONS(2542),
    [anon_sym_POUND] = ACTIONS(2542),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2542),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2542),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2542),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_API_AVAILABLE] = ACTIONS(2542),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_API_DEPRECATED] = ACTIONS(2542),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym___deprecated_msg] = ACTIONS(2542),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2542),
    [anon_sym_ATprotocol] = ACTIONS(2544),
    [anon_sym_ATinterface] = ACTIONS(2544),
    [anon_sym_ATimplementation] = ACTIONS(2544),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2544),
    [anon_sym_ATtry] = ACTIONS(2544),
    [anon_sym___try] = ACTIONS(2542),
    [anon_sym_ATthrow] = ACTIONS(2544),
    [anon_sym_ATselector] = ACTIONS(2544),
    [anon_sym_ATavailable] = ACTIONS(2544),
    [anon_sym___builtin_available] = ACTIONS(2542),
    [anon_sym_va_arg] = ACTIONS(2542),
    [anon_sym_ATencode] = ACTIONS(2544),
    [anon_sym_ATsynchronized] = ACTIONS(2544),
    [anon_sym_BOOL] = ACTIONS(2542),
    [anon_sym_IMP] = ACTIONS(2542),
    [anon_sym_SEL] = ACTIONS(2542),
    [anon_sym_Class] = ACTIONS(2542),
    [anon_sym_id] = ACTIONS(2542),
  },
  [STATE(816)] = {
    [ts_builtin_sym_end] = ACTIONS(2556),
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(817)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_RBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(818)] = {
    [ts_builtin_sym_end] = ACTIONS(2556),
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(819)] = {
    [ts_builtin_sym_end] = ACTIONS(2444),
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(820)] = {
    [ts_builtin_sym_end] = ACTIONS(2476),
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(821)] = {
    [ts_builtin_sym_end] = ACTIONS(2476),
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(822)] = {
    [ts_builtin_sym_end] = ACTIONS(2560),
    [sym_identifier] = ACTIONS(2558),
    [aux_sym_preproc_include_token1] = ACTIONS(2558),
    [aux_sym_preproc_include_token2] = ACTIONS(2558),
    [aux_sym_preproc_def_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2558),
    [sym_preproc_directive] = ACTIONS(2558),
    [anon_sym_LPAREN2] = ACTIONS(2560),
    [anon_sym_BANG] = ACTIONS(2560),
    [anon_sym_TILDE] = ACTIONS(2560),
    [anon_sym_DASH] = ACTIONS(2558),
    [anon_sym_PLUS] = ACTIONS(2558),
    [anon_sym_STAR] = ACTIONS(2560),
    [anon_sym_CARET] = ACTIONS(2560),
    [anon_sym_AMP] = ACTIONS(2560),
    [anon_sym_SEMI] = ACTIONS(2560),
    [anon_sym___extension__] = ACTIONS(2558),
    [anon_sym_typedef] = ACTIONS(2558),
    [anon_sym_extern] = ACTIONS(2558),
    [anon_sym___attribute__] = ACTIONS(2558),
    [anon_sym___attribute] = ACTIONS(2558),
    [anon_sym_noreturn] = ACTIONS(2558),
    [anon_sym_LBRACK] = ACTIONS(2560),
    [anon_sym___declspec] = ACTIONS(2558),
    [anon_sym___cdecl] = ACTIONS(2558),
    [anon_sym___clrcall] = ACTIONS(2558),
    [anon_sym___stdcall] = ACTIONS(2558),
    [anon_sym___fastcall] = ACTIONS(2558),
    [anon_sym___thiscall] = ACTIONS(2558),
    [anon_sym___vectorcall] = ACTIONS(2558),
    [anon_sym_LBRACE] = ACTIONS(2560),
    [anon_sym_signed] = ACTIONS(2558),
    [anon_sym_unsigned] = ACTIONS(2558),
    [anon_sym_long] = ACTIONS(2558),
    [anon_sym_short] = ACTIONS(2558),
    [anon_sym_static] = ACTIONS(2558),
    [anon_sym_ATautoreleasepool] = ACTIONS(2560),
    [anon_sym_auto] = ACTIONS(2558),
    [anon_sym_register] = ACTIONS(2558),
    [anon_sym_inline] = ACTIONS(2558),
    [anon_sym___inline] = ACTIONS(2558),
    [anon_sym___inline__] = ACTIONS(2558),
    [anon_sym___forceinline] = ACTIONS(2558),
    [anon_sym_thread_local] = ACTIONS(2558),
    [anon_sym___thread] = ACTIONS(2558),
    [anon_sym_CG_EXTERN] = ACTIONS(2558),
    [anon_sym_CG_INLINE] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2558),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2558),
    [anon_sym_IBOutlet] = ACTIONS(2558),
    [anon_sym_IBInspectable] = ACTIONS(2558),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2558),
    [anon_sym_NS_INLINE] = ACTIONS(2558),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2558),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2558),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2558),
    [anon_sym_const] = ACTIONS(2558),
    [anon_sym_constexpr] = ACTIONS(2558),
    [anon_sym_volatile] = ACTIONS(2558),
    [anon_sym_restrict] = ACTIONS(2558),
    [anon_sym___restrict__] = ACTIONS(2558),
    [anon_sym__Atomic] = ACTIONS(2558),
    [anon_sym__Noreturn] = ACTIONS(2558),
    [anon_sym__Nonnull] = ACTIONS(2558),
    [anon_sym_nullable] = ACTIONS(2558),
    [anon_sym__Complex] = ACTIONS(2558),
    [anon_sym__Nullable] = ACTIONS(2558),
    [anon_sym__Nullable_result] = ACTIONS(2558),
    [anon_sym__Null_unspecified] = ACTIONS(2558),
    [anon_sym___autoreleasing] = ACTIONS(2558),
    [anon_sym___block] = ACTIONS(2558),
    [anon_sym___bridge] = ACTIONS(2558),
    [anon_sym___bridge_retained] = ACTIONS(2558),
    [anon_sym___bridge_transfer] = ACTIONS(2558),
    [anon_sym___complex] = ACTIONS(2558),
    [anon_sym___const] = ACTIONS(2558),
    [anon_sym___imag] = ACTIONS(2558),
    [anon_sym___kindof] = ACTIONS(2558),
    [anon_sym___nonnull] = ACTIONS(2558),
    [anon_sym___nullable] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2558),
    [anon_sym___real] = ACTIONS(2558),
    [anon_sym___strong] = ACTIONS(2558),
    [anon_sym___unsafe_unretained] = ACTIONS(2558),
    [anon_sym___unused] = ACTIONS(2558),
    [anon_sym___weak] = ACTIONS(2558),
    [anon_sym_alignas] = ACTIONS(2558),
    [anon_sym__Alignas] = ACTIONS(2558),
    [sym_primitive_type] = ACTIONS(2558),
    [anon_sym_enum] = ACTIONS(2558),
    [anon_sym_struct] = ACTIONS(2558),
    [anon_sym_union] = ACTIONS(2558),
    [anon_sym_if] = ACTIONS(2558),
    [anon_sym_else] = ACTIONS(2558),
    [anon_sym_switch] = ACTIONS(2558),
    [anon_sym_case] = ACTIONS(2558),
    [anon_sym_default] = ACTIONS(2558),
    [anon_sym_while] = ACTIONS(2558),
    [anon_sym_do] = ACTIONS(2558),
    [anon_sym_for] = ACTIONS(2558),
    [anon_sym_in] = ACTIONS(2558),
    [anon_sym_return] = ACTIONS(2558),
    [anon_sym_break] = ACTIONS(2558),
    [anon_sym_continue] = ACTIONS(2558),
    [anon_sym_goto] = ACTIONS(2558),
    [anon_sym_DASH_DASH] = ACTIONS(2560),
    [anon_sym_PLUS_PLUS] = ACTIONS(2560),
    [anon_sym_sizeof] = ACTIONS(2558),
    [anon_sym___alignof__] = ACTIONS(2558),
    [anon_sym___alignof] = ACTIONS(2558),
    [anon_sym__alignof] = ACTIONS(2558),
    [anon_sym_alignof] = ACTIONS(2558),
    [anon_sym__Alignof] = ACTIONS(2558),
    [anon_sym_offsetof] = ACTIONS(2558),
    [anon_sym__Generic] = ACTIONS(2558),
    [anon_sym_asm] = ACTIONS(2558),
    [anon_sym___asm__] = ACTIONS(2558),
    [anon_sym___asm] = ACTIONS(2558),
    [sym_number_literal] = ACTIONS(2560),
    [anon_sym_L_SQUOTE] = ACTIONS(2560),
    [anon_sym_u_SQUOTE] = ACTIONS(2560),
    [anon_sym_U_SQUOTE] = ACTIONS(2560),
    [anon_sym_u8_SQUOTE] = ACTIONS(2560),
    [anon_sym_SQUOTE] = ACTIONS(2560),
    [anon_sym_AT] = ACTIONS(2558),
    [anon_sym_DQUOTE] = ACTIONS(2560),
    [anon_sym_L_DQUOTE] = ACTIONS(2560),
    [anon_sym_u_DQUOTE] = ACTIONS(2560),
    [anon_sym_U_DQUOTE] = ACTIONS(2560),
    [anon_sym_u8_DQUOTE] = ACTIONS(2560),
    [sym_true] = ACTIONS(2558),
    [sym_false] = ACTIONS(2558),
    [anon_sym_NULL] = ACTIONS(2558),
    [anon_sym_nullptr] = ACTIONS(2558),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2558),
    [anon_sym___typeof] = ACTIONS(2558),
    [anon_sym_typeof] = ACTIONS(2558),
    [anon_sym_ATimport] = ACTIONS(2560),
    [aux_sym_preproc_undef_token1] = ACTIONS(2558),
    [anon_sym_POUND] = ACTIONS(2558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2558),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2558),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2558),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_API_AVAILABLE] = ACTIONS(2558),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_API_DEPRECATED] = ACTIONS(2558),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym___deprecated_msg] = ACTIONS(2558),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2558),
    [anon_sym_ATprotocol] = ACTIONS(2560),
    [anon_sym_ATinterface] = ACTIONS(2560),
    [anon_sym_ATimplementation] = ACTIONS(2560),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2560),
    [anon_sym_ATtry] = ACTIONS(2560),
    [anon_sym___try] = ACTIONS(2558),
    [anon_sym_ATthrow] = ACTIONS(2560),
    [anon_sym_ATselector] = ACTIONS(2560),
    [anon_sym_ATavailable] = ACTIONS(2560),
    [anon_sym___builtin_available] = ACTIONS(2558),
    [anon_sym_va_arg] = ACTIONS(2558),
    [anon_sym_ATencode] = ACTIONS(2560),
    [anon_sym_ATsynchronized] = ACTIONS(2560),
    [anon_sym_BOOL] = ACTIONS(2558),
    [anon_sym_IMP] = ACTIONS(2558),
    [anon_sym_SEL] = ACTIONS(2558),
    [anon_sym_Class] = ACTIONS(2558),
    [anon_sym_id] = ACTIONS(2558),
  },
  [STATE(823)] = {
    [ts_builtin_sym_end] = ACTIONS(2414),
    [sym_identifier] = ACTIONS(2412),
    [aux_sym_preproc_include_token1] = ACTIONS(2412),
    [aux_sym_preproc_include_token2] = ACTIONS(2412),
    [aux_sym_preproc_def_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2412),
    [sym_preproc_directive] = ACTIONS(2412),
    [anon_sym_LPAREN2] = ACTIONS(2414),
    [anon_sym_BANG] = ACTIONS(2414),
    [anon_sym_TILDE] = ACTIONS(2414),
    [anon_sym_DASH] = ACTIONS(2412),
    [anon_sym_PLUS] = ACTIONS(2412),
    [anon_sym_STAR] = ACTIONS(2414),
    [anon_sym_CARET] = ACTIONS(2414),
    [anon_sym_AMP] = ACTIONS(2414),
    [anon_sym_SEMI] = ACTIONS(2414),
    [anon_sym___extension__] = ACTIONS(2412),
    [anon_sym_typedef] = ACTIONS(2412),
    [anon_sym_extern] = ACTIONS(2412),
    [anon_sym___attribute__] = ACTIONS(2412),
    [anon_sym___attribute] = ACTIONS(2412),
    [anon_sym_noreturn] = ACTIONS(2412),
    [anon_sym_LBRACK] = ACTIONS(2414),
    [anon_sym___declspec] = ACTIONS(2412),
    [anon_sym___cdecl] = ACTIONS(2412),
    [anon_sym___clrcall] = ACTIONS(2412),
    [anon_sym___stdcall] = ACTIONS(2412),
    [anon_sym___fastcall] = ACTIONS(2412),
    [anon_sym___thiscall] = ACTIONS(2412),
    [anon_sym___vectorcall] = ACTIONS(2412),
    [anon_sym_LBRACE] = ACTIONS(2414),
    [anon_sym_signed] = ACTIONS(2412),
    [anon_sym_unsigned] = ACTIONS(2412),
    [anon_sym_long] = ACTIONS(2412),
    [anon_sym_short] = ACTIONS(2412),
    [anon_sym_static] = ACTIONS(2412),
    [anon_sym_ATautoreleasepool] = ACTIONS(2414),
    [anon_sym_auto] = ACTIONS(2412),
    [anon_sym_register] = ACTIONS(2412),
    [anon_sym_inline] = ACTIONS(2412),
    [anon_sym___inline] = ACTIONS(2412),
    [anon_sym___inline__] = ACTIONS(2412),
    [anon_sym___forceinline] = ACTIONS(2412),
    [anon_sym_thread_local] = ACTIONS(2412),
    [anon_sym___thread] = ACTIONS(2412),
    [anon_sym_CG_EXTERN] = ACTIONS(2412),
    [anon_sym_CG_INLINE] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2412),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2412),
    [anon_sym_IBOutlet] = ACTIONS(2412),
    [anon_sym_IBInspectable] = ACTIONS(2412),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2412),
    [anon_sym_NS_INLINE] = ACTIONS(2412),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2412),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2412),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2412),
    [anon_sym_const] = ACTIONS(2412),
    [anon_sym_constexpr] = ACTIONS(2412),
    [anon_sym_volatile] = ACTIONS(2412),
    [anon_sym_restrict] = ACTIONS(2412),
    [anon_sym___restrict__] = ACTIONS(2412),
    [anon_sym__Atomic] = ACTIONS(2412),
    [anon_sym__Noreturn] = ACTIONS(2412),
    [anon_sym__Nonnull] = ACTIONS(2412),
    [anon_sym_nullable] = ACTIONS(2412),
    [anon_sym__Complex] = ACTIONS(2412),
    [anon_sym__Nullable] = ACTIONS(2412),
    [anon_sym__Nullable_result] = ACTIONS(2412),
    [anon_sym__Null_unspecified] = ACTIONS(2412),
    [anon_sym___autoreleasing] = ACTIONS(2412),
    [anon_sym___block] = ACTIONS(2412),
    [anon_sym___bridge] = ACTIONS(2412),
    [anon_sym___bridge_retained] = ACTIONS(2412),
    [anon_sym___bridge_transfer] = ACTIONS(2412),
    [anon_sym___complex] = ACTIONS(2412),
    [anon_sym___const] = ACTIONS(2412),
    [anon_sym___imag] = ACTIONS(2412),
    [anon_sym___kindof] = ACTIONS(2412),
    [anon_sym___nonnull] = ACTIONS(2412),
    [anon_sym___nullable] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2412),
    [anon_sym___real] = ACTIONS(2412),
    [anon_sym___strong] = ACTIONS(2412),
    [anon_sym___unsafe_unretained] = ACTIONS(2412),
    [anon_sym___unused] = ACTIONS(2412),
    [anon_sym___weak] = ACTIONS(2412),
    [anon_sym_alignas] = ACTIONS(2412),
    [anon_sym__Alignas] = ACTIONS(2412),
    [sym_primitive_type] = ACTIONS(2412),
    [anon_sym_enum] = ACTIONS(2412),
    [anon_sym_struct] = ACTIONS(2412),
    [anon_sym_union] = ACTIONS(2412),
    [anon_sym_if] = ACTIONS(2412),
    [anon_sym_else] = ACTIONS(2412),
    [anon_sym_switch] = ACTIONS(2412),
    [anon_sym_case] = ACTIONS(2412),
    [anon_sym_default] = ACTIONS(2412),
    [anon_sym_while] = ACTIONS(2412),
    [anon_sym_do] = ACTIONS(2412),
    [anon_sym_for] = ACTIONS(2412),
    [anon_sym_in] = ACTIONS(2412),
    [anon_sym_return] = ACTIONS(2412),
    [anon_sym_break] = ACTIONS(2412),
    [anon_sym_continue] = ACTIONS(2412),
    [anon_sym_goto] = ACTIONS(2412),
    [anon_sym_DASH_DASH] = ACTIONS(2414),
    [anon_sym_PLUS_PLUS] = ACTIONS(2414),
    [anon_sym_sizeof] = ACTIONS(2412),
    [anon_sym___alignof__] = ACTIONS(2412),
    [anon_sym___alignof] = ACTIONS(2412),
    [anon_sym__alignof] = ACTIONS(2412),
    [anon_sym_alignof] = ACTIONS(2412),
    [anon_sym__Alignof] = ACTIONS(2412),
    [anon_sym_offsetof] = ACTIONS(2412),
    [anon_sym__Generic] = ACTIONS(2412),
    [anon_sym_asm] = ACTIONS(2412),
    [anon_sym___asm__] = ACTIONS(2412),
    [anon_sym___asm] = ACTIONS(2412),
    [sym_number_literal] = ACTIONS(2414),
    [anon_sym_L_SQUOTE] = ACTIONS(2414),
    [anon_sym_u_SQUOTE] = ACTIONS(2414),
    [anon_sym_U_SQUOTE] = ACTIONS(2414),
    [anon_sym_u8_SQUOTE] = ACTIONS(2414),
    [anon_sym_SQUOTE] = ACTIONS(2414),
    [anon_sym_AT] = ACTIONS(2412),
    [anon_sym_DQUOTE] = ACTIONS(2414),
    [anon_sym_L_DQUOTE] = ACTIONS(2414),
    [anon_sym_u_DQUOTE] = ACTIONS(2414),
    [anon_sym_U_DQUOTE] = ACTIONS(2414),
    [anon_sym_u8_DQUOTE] = ACTIONS(2414),
    [sym_true] = ACTIONS(2412),
    [sym_false] = ACTIONS(2412),
    [anon_sym_NULL] = ACTIONS(2412),
    [anon_sym_nullptr] = ACTIONS(2412),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2412),
    [anon_sym___typeof] = ACTIONS(2412),
    [anon_sym_typeof] = ACTIONS(2412),
    [anon_sym_ATimport] = ACTIONS(2414),
    [aux_sym_preproc_undef_token1] = ACTIONS(2412),
    [anon_sym_POUND] = ACTIONS(2412),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2412),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2412),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2412),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_API_AVAILABLE] = ACTIONS(2412),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_API_DEPRECATED] = ACTIONS(2412),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym___deprecated_msg] = ACTIONS(2412),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2412),
    [anon_sym_ATprotocol] = ACTIONS(2414),
    [anon_sym_ATinterface] = ACTIONS(2414),
    [anon_sym_ATimplementation] = ACTIONS(2414),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2414),
    [anon_sym_ATtry] = ACTIONS(2414),
    [anon_sym___try] = ACTIONS(2412),
    [anon_sym_ATthrow] = ACTIONS(2414),
    [anon_sym_ATselector] = ACTIONS(2414),
    [anon_sym_ATavailable] = ACTIONS(2414),
    [anon_sym___builtin_available] = ACTIONS(2412),
    [anon_sym_va_arg] = ACTIONS(2412),
    [anon_sym_ATencode] = ACTIONS(2414),
    [anon_sym_ATsynchronized] = ACTIONS(2414),
    [anon_sym_BOOL] = ACTIONS(2412),
    [anon_sym_IMP] = ACTIONS(2412),
    [anon_sym_SEL] = ACTIONS(2412),
    [anon_sym_Class] = ACTIONS(2412),
    [anon_sym_id] = ACTIONS(2412),
  },
  [STATE(824)] = {
    [sym_identifier] = ACTIONS(2458),
    [aux_sym_preproc_include_token1] = ACTIONS(2458),
    [aux_sym_preproc_include_token2] = ACTIONS(2458),
    [aux_sym_preproc_def_token1] = ACTIONS(2458),
    [aux_sym_preproc_if_token1] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2458),
    [sym_preproc_directive] = ACTIONS(2458),
    [anon_sym_LPAREN2] = ACTIONS(2460),
    [anon_sym_BANG] = ACTIONS(2460),
    [anon_sym_TILDE] = ACTIONS(2460),
    [anon_sym_DASH] = ACTIONS(2458),
    [anon_sym_PLUS] = ACTIONS(2458),
    [anon_sym_STAR] = ACTIONS(2460),
    [anon_sym_CARET] = ACTIONS(2460),
    [anon_sym_AMP] = ACTIONS(2460),
    [anon_sym_SEMI] = ACTIONS(2460),
    [anon_sym___extension__] = ACTIONS(2458),
    [anon_sym_typedef] = ACTIONS(2458),
    [anon_sym_extern] = ACTIONS(2458),
    [anon_sym___attribute__] = ACTIONS(2458),
    [anon_sym___attribute] = ACTIONS(2458),
    [anon_sym_noreturn] = ACTIONS(2458),
    [anon_sym_LBRACK] = ACTIONS(2460),
    [anon_sym___declspec] = ACTIONS(2458),
    [anon_sym___cdecl] = ACTIONS(2458),
    [anon_sym___clrcall] = ACTIONS(2458),
    [anon_sym___stdcall] = ACTIONS(2458),
    [anon_sym___fastcall] = ACTIONS(2458),
    [anon_sym___thiscall] = ACTIONS(2458),
    [anon_sym___vectorcall] = ACTIONS(2458),
    [anon_sym_LBRACE] = ACTIONS(2460),
    [anon_sym_RBRACE] = ACTIONS(2460),
    [anon_sym_signed] = ACTIONS(2458),
    [anon_sym_unsigned] = ACTIONS(2458),
    [anon_sym_long] = ACTIONS(2458),
    [anon_sym_short] = ACTIONS(2458),
    [anon_sym_static] = ACTIONS(2458),
    [anon_sym_ATautoreleasepool] = ACTIONS(2460),
    [anon_sym_auto] = ACTIONS(2458),
    [anon_sym_register] = ACTIONS(2458),
    [anon_sym_inline] = ACTIONS(2458),
    [anon_sym___inline] = ACTIONS(2458),
    [anon_sym___inline__] = ACTIONS(2458),
    [anon_sym___forceinline] = ACTIONS(2458),
    [anon_sym_thread_local] = ACTIONS(2458),
    [anon_sym___thread] = ACTIONS(2458),
    [anon_sym_CG_EXTERN] = ACTIONS(2458),
    [anon_sym_CG_INLINE] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2458),
    [anon_sym_IBOutlet] = ACTIONS(2458),
    [anon_sym_IBInspectable] = ACTIONS(2458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2458),
    [anon_sym_NS_INLINE] = ACTIONS(2458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2458),
    [anon_sym_const] = ACTIONS(2458),
    [anon_sym_constexpr] = ACTIONS(2458),
    [anon_sym_volatile] = ACTIONS(2458),
    [anon_sym_restrict] = ACTIONS(2458),
    [anon_sym___restrict__] = ACTIONS(2458),
    [anon_sym__Atomic] = ACTIONS(2458),
    [anon_sym__Noreturn] = ACTIONS(2458),
    [anon_sym__Nonnull] = ACTIONS(2458),
    [anon_sym_nullable] = ACTIONS(2458),
    [anon_sym__Complex] = ACTIONS(2458),
    [anon_sym__Nullable] = ACTIONS(2458),
    [anon_sym__Nullable_result] = ACTIONS(2458),
    [anon_sym__Null_unspecified] = ACTIONS(2458),
    [anon_sym___autoreleasing] = ACTIONS(2458),
    [anon_sym___block] = ACTIONS(2458),
    [anon_sym___bridge] = ACTIONS(2458),
    [anon_sym___bridge_retained] = ACTIONS(2458),
    [anon_sym___bridge_transfer] = ACTIONS(2458),
    [anon_sym___complex] = ACTIONS(2458),
    [anon_sym___const] = ACTIONS(2458),
    [anon_sym___imag] = ACTIONS(2458),
    [anon_sym___kindof] = ACTIONS(2458),
    [anon_sym___nonnull] = ACTIONS(2458),
    [anon_sym___nullable] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2458),
    [anon_sym___real] = ACTIONS(2458),
    [anon_sym___strong] = ACTIONS(2458),
    [anon_sym___unsafe_unretained] = ACTIONS(2458),
    [anon_sym___unused] = ACTIONS(2458),
    [anon_sym___weak] = ACTIONS(2458),
    [anon_sym_alignas] = ACTIONS(2458),
    [anon_sym__Alignas] = ACTIONS(2458),
    [sym_primitive_type] = ACTIONS(2458),
    [anon_sym_enum] = ACTIONS(2458),
    [anon_sym_struct] = ACTIONS(2458),
    [anon_sym_union] = ACTIONS(2458),
    [anon_sym_if] = ACTIONS(2458),
    [anon_sym_else] = ACTIONS(2458),
    [anon_sym_switch] = ACTIONS(2458),
    [anon_sym_case] = ACTIONS(2458),
    [anon_sym_default] = ACTIONS(2458),
    [anon_sym_while] = ACTIONS(2458),
    [anon_sym_do] = ACTIONS(2458),
    [anon_sym_for] = ACTIONS(2458),
    [anon_sym_in] = ACTIONS(2458),
    [anon_sym_return] = ACTIONS(2458),
    [anon_sym_break] = ACTIONS(2458),
    [anon_sym_continue] = ACTIONS(2458),
    [anon_sym_goto] = ACTIONS(2458),
    [anon_sym_DASH_DASH] = ACTIONS(2460),
    [anon_sym_PLUS_PLUS] = ACTIONS(2460),
    [anon_sym_sizeof] = ACTIONS(2458),
    [anon_sym___alignof__] = ACTIONS(2458),
    [anon_sym___alignof] = ACTIONS(2458),
    [anon_sym__alignof] = ACTIONS(2458),
    [anon_sym_alignof] = ACTIONS(2458),
    [anon_sym__Alignof] = ACTIONS(2458),
    [anon_sym_offsetof] = ACTIONS(2458),
    [anon_sym__Generic] = ACTIONS(2458),
    [anon_sym_asm] = ACTIONS(2458),
    [anon_sym___asm__] = ACTIONS(2458),
    [anon_sym___asm] = ACTIONS(2458),
    [sym_number_literal] = ACTIONS(2460),
    [anon_sym_L_SQUOTE] = ACTIONS(2460),
    [anon_sym_u_SQUOTE] = ACTIONS(2460),
    [anon_sym_U_SQUOTE] = ACTIONS(2460),
    [anon_sym_u8_SQUOTE] = ACTIONS(2460),
    [anon_sym_SQUOTE] = ACTIONS(2460),
    [anon_sym_AT] = ACTIONS(2458),
    [anon_sym_DQUOTE] = ACTIONS(2460),
    [anon_sym_L_DQUOTE] = ACTIONS(2460),
    [anon_sym_u_DQUOTE] = ACTIONS(2460),
    [anon_sym_U_DQUOTE] = ACTIONS(2460),
    [anon_sym_u8_DQUOTE] = ACTIONS(2460),
    [sym_true] = ACTIONS(2458),
    [sym_false] = ACTIONS(2458),
    [anon_sym_NULL] = ACTIONS(2458),
    [anon_sym_nullptr] = ACTIONS(2458),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2458),
    [anon_sym___typeof] = ACTIONS(2458),
    [anon_sym_typeof] = ACTIONS(2458),
    [anon_sym_ATimport] = ACTIONS(2460),
    [aux_sym_preproc_undef_token1] = ACTIONS(2458),
    [anon_sym_POUND] = ACTIONS(2458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_API_AVAILABLE] = ACTIONS(2458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_API_DEPRECATED] = ACTIONS(2458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym___deprecated_msg] = ACTIONS(2458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2458),
    [anon_sym_ATprotocol] = ACTIONS(2460),
    [anon_sym_ATinterface] = ACTIONS(2460),
    [anon_sym_ATimplementation] = ACTIONS(2460),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2460),
    [anon_sym_ATtry] = ACTIONS(2460),
    [anon_sym___try] = ACTIONS(2458),
    [anon_sym_ATthrow] = ACTIONS(2460),
    [anon_sym_ATselector] = ACTIONS(2460),
    [anon_sym_ATavailable] = ACTIONS(2460),
    [anon_sym___builtin_available] = ACTIONS(2458),
    [anon_sym_va_arg] = ACTIONS(2458),
    [anon_sym_ATencode] = ACTIONS(2460),
    [anon_sym_ATsynchronized] = ACTIONS(2460),
    [anon_sym_BOOL] = ACTIONS(2458),
    [anon_sym_IMP] = ACTIONS(2458),
    [anon_sym_SEL] = ACTIONS(2458),
    [anon_sym_Class] = ACTIONS(2458),
    [anon_sym_id] = ACTIONS(2458),
  },
  [STATE(825)] = {
    [sym_identifier] = ACTIONS(2454),
    [aux_sym_preproc_include_token1] = ACTIONS(2454),
    [aux_sym_preproc_include_token2] = ACTIONS(2454),
    [aux_sym_preproc_def_token1] = ACTIONS(2454),
    [aux_sym_preproc_if_token1] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2454),
    [sym_preproc_directive] = ACTIONS(2454),
    [anon_sym_LPAREN2] = ACTIONS(2456),
    [anon_sym_BANG] = ACTIONS(2456),
    [anon_sym_TILDE] = ACTIONS(2456),
    [anon_sym_DASH] = ACTIONS(2454),
    [anon_sym_PLUS] = ACTIONS(2454),
    [anon_sym_STAR] = ACTIONS(2456),
    [anon_sym_CARET] = ACTIONS(2456),
    [anon_sym_AMP] = ACTIONS(2456),
    [anon_sym_SEMI] = ACTIONS(2456),
    [anon_sym___extension__] = ACTIONS(2454),
    [anon_sym_typedef] = ACTIONS(2454),
    [anon_sym_extern] = ACTIONS(2454),
    [anon_sym___attribute__] = ACTIONS(2454),
    [anon_sym___attribute] = ACTIONS(2454),
    [anon_sym_noreturn] = ACTIONS(2454),
    [anon_sym_LBRACK] = ACTIONS(2456),
    [anon_sym___declspec] = ACTIONS(2454),
    [anon_sym___cdecl] = ACTIONS(2454),
    [anon_sym___clrcall] = ACTIONS(2454),
    [anon_sym___stdcall] = ACTIONS(2454),
    [anon_sym___fastcall] = ACTIONS(2454),
    [anon_sym___thiscall] = ACTIONS(2454),
    [anon_sym___vectorcall] = ACTIONS(2454),
    [anon_sym_LBRACE] = ACTIONS(2456),
    [anon_sym_RBRACE] = ACTIONS(2456),
    [anon_sym_signed] = ACTIONS(2454),
    [anon_sym_unsigned] = ACTIONS(2454),
    [anon_sym_long] = ACTIONS(2454),
    [anon_sym_short] = ACTIONS(2454),
    [anon_sym_static] = ACTIONS(2454),
    [anon_sym_ATautoreleasepool] = ACTIONS(2456),
    [anon_sym_auto] = ACTIONS(2454),
    [anon_sym_register] = ACTIONS(2454),
    [anon_sym_inline] = ACTIONS(2454),
    [anon_sym___inline] = ACTIONS(2454),
    [anon_sym___inline__] = ACTIONS(2454),
    [anon_sym___forceinline] = ACTIONS(2454),
    [anon_sym_thread_local] = ACTIONS(2454),
    [anon_sym___thread] = ACTIONS(2454),
    [anon_sym_CG_EXTERN] = ACTIONS(2454),
    [anon_sym_CG_INLINE] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2454),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2454),
    [anon_sym_IBOutlet] = ACTIONS(2454),
    [anon_sym_IBInspectable] = ACTIONS(2454),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2454),
    [anon_sym_NS_INLINE] = ACTIONS(2454),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2454),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2454),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2454),
    [anon_sym_const] = ACTIONS(2454),
    [anon_sym_constexpr] = ACTIONS(2454),
    [anon_sym_volatile] = ACTIONS(2454),
    [anon_sym_restrict] = ACTIONS(2454),
    [anon_sym___restrict__] = ACTIONS(2454),
    [anon_sym__Atomic] = ACTIONS(2454),
    [anon_sym__Noreturn] = ACTIONS(2454),
    [anon_sym__Nonnull] = ACTIONS(2454),
    [anon_sym_nullable] = ACTIONS(2454),
    [anon_sym__Complex] = ACTIONS(2454),
    [anon_sym__Nullable] = ACTIONS(2454),
    [anon_sym__Nullable_result] = ACTIONS(2454),
    [anon_sym__Null_unspecified] = ACTIONS(2454),
    [anon_sym___autoreleasing] = ACTIONS(2454),
    [anon_sym___block] = ACTIONS(2454),
    [anon_sym___bridge] = ACTIONS(2454),
    [anon_sym___bridge_retained] = ACTIONS(2454),
    [anon_sym___bridge_transfer] = ACTIONS(2454),
    [anon_sym___complex] = ACTIONS(2454),
    [anon_sym___const] = ACTIONS(2454),
    [anon_sym___imag] = ACTIONS(2454),
    [anon_sym___kindof] = ACTIONS(2454),
    [anon_sym___nonnull] = ACTIONS(2454),
    [anon_sym___nullable] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2454),
    [anon_sym___real] = ACTIONS(2454),
    [anon_sym___strong] = ACTIONS(2454),
    [anon_sym___unsafe_unretained] = ACTIONS(2454),
    [anon_sym___unused] = ACTIONS(2454),
    [anon_sym___weak] = ACTIONS(2454),
    [anon_sym_alignas] = ACTIONS(2454),
    [anon_sym__Alignas] = ACTIONS(2454),
    [sym_primitive_type] = ACTIONS(2454),
    [anon_sym_enum] = ACTIONS(2454),
    [anon_sym_struct] = ACTIONS(2454),
    [anon_sym_union] = ACTIONS(2454),
    [anon_sym_if] = ACTIONS(2454),
    [anon_sym_else] = ACTIONS(2454),
    [anon_sym_switch] = ACTIONS(2454),
    [anon_sym_case] = ACTIONS(2454),
    [anon_sym_default] = ACTIONS(2454),
    [anon_sym_while] = ACTIONS(2454),
    [anon_sym_do] = ACTIONS(2454),
    [anon_sym_for] = ACTIONS(2454),
    [anon_sym_in] = ACTIONS(2454),
    [anon_sym_return] = ACTIONS(2454),
    [anon_sym_break] = ACTIONS(2454),
    [anon_sym_continue] = ACTIONS(2454),
    [anon_sym_goto] = ACTIONS(2454),
    [anon_sym_DASH_DASH] = ACTIONS(2456),
    [anon_sym_PLUS_PLUS] = ACTIONS(2456),
    [anon_sym_sizeof] = ACTIONS(2454),
    [anon_sym___alignof__] = ACTIONS(2454),
    [anon_sym___alignof] = ACTIONS(2454),
    [anon_sym__alignof] = ACTIONS(2454),
    [anon_sym_alignof] = ACTIONS(2454),
    [anon_sym__Alignof] = ACTIONS(2454),
    [anon_sym_offsetof] = ACTIONS(2454),
    [anon_sym__Generic] = ACTIONS(2454),
    [anon_sym_asm] = ACTIONS(2454),
    [anon_sym___asm__] = ACTIONS(2454),
    [anon_sym___asm] = ACTIONS(2454),
    [sym_number_literal] = ACTIONS(2456),
    [anon_sym_L_SQUOTE] = ACTIONS(2456),
    [anon_sym_u_SQUOTE] = ACTIONS(2456),
    [anon_sym_U_SQUOTE] = ACTIONS(2456),
    [anon_sym_u8_SQUOTE] = ACTIONS(2456),
    [anon_sym_SQUOTE] = ACTIONS(2456),
    [anon_sym_AT] = ACTIONS(2454),
    [anon_sym_DQUOTE] = ACTIONS(2456),
    [anon_sym_L_DQUOTE] = ACTIONS(2456),
    [anon_sym_u_DQUOTE] = ACTIONS(2456),
    [anon_sym_U_DQUOTE] = ACTIONS(2456),
    [anon_sym_u8_DQUOTE] = ACTIONS(2456),
    [sym_true] = ACTIONS(2454),
    [sym_false] = ACTIONS(2454),
    [anon_sym_NULL] = ACTIONS(2454),
    [anon_sym_nullptr] = ACTIONS(2454),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2454),
    [anon_sym___typeof] = ACTIONS(2454),
    [anon_sym_typeof] = ACTIONS(2454),
    [anon_sym_ATimport] = ACTIONS(2456),
    [aux_sym_preproc_undef_token1] = ACTIONS(2454),
    [anon_sym_POUND] = ACTIONS(2454),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2454),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2454),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2454),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_API_AVAILABLE] = ACTIONS(2454),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_API_DEPRECATED] = ACTIONS(2454),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym___deprecated_msg] = ACTIONS(2454),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2454),
    [anon_sym_ATprotocol] = ACTIONS(2456),
    [anon_sym_ATinterface] = ACTIONS(2456),
    [anon_sym_ATimplementation] = ACTIONS(2456),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2456),
    [anon_sym_ATtry] = ACTIONS(2456),
    [anon_sym___try] = ACTIONS(2454),
    [anon_sym_ATthrow] = ACTIONS(2456),
    [anon_sym_ATselector] = ACTIONS(2456),
    [anon_sym_ATavailable] = ACTIONS(2456),
    [anon_sym___builtin_available] = ACTIONS(2454),
    [anon_sym_va_arg] = ACTIONS(2454),
    [anon_sym_ATencode] = ACTIONS(2456),
    [anon_sym_ATsynchronized] = ACTIONS(2456),
    [anon_sym_BOOL] = ACTIONS(2454),
    [anon_sym_IMP] = ACTIONS(2454),
    [anon_sym_SEL] = ACTIONS(2454),
    [anon_sym_Class] = ACTIONS(2454),
    [anon_sym_id] = ACTIONS(2454),
  },
  [STATE(826)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(2900),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2222),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2222),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2222),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2222),
    [anon_sym_GT_GT] = ACTIONS(2222),
    [anon_sym_SEMI] = ACTIONS(2222),
    [anon_sym___extension__] = ACTIONS(2220),
    [anon_sym_extern] = ACTIONS(2220),
    [anon_sym___attribute__] = ACTIONS(2220),
    [anon_sym___attribute] = ACTIONS(2220),
    [anon_sym_noreturn] = ACTIONS(2220),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym___declspec] = ACTIONS(2220),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_static] = ACTIONS(2220),
    [anon_sym_auto] = ACTIONS(2220),
    [anon_sym_register] = ACTIONS(2220),
    [anon_sym_inline] = ACTIONS(2220),
    [anon_sym___inline] = ACTIONS(2220),
    [anon_sym___inline__] = ACTIONS(2220),
    [anon_sym___forceinline] = ACTIONS(2220),
    [anon_sym_thread_local] = ACTIONS(2220),
    [anon_sym___thread] = ACTIONS(2220),
    [anon_sym_CG_EXTERN] = ACTIONS(2220),
    [anon_sym_CG_INLINE] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2220),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2220),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2220),
    [anon_sym_IBOutlet] = ACTIONS(2220),
    [anon_sym_IBInspectable] = ACTIONS(2220),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2220),
    [anon_sym_NS_INLINE] = ACTIONS(2220),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2220),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2220),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2220),
    [anon_sym_const] = ACTIONS(2220),
    [anon_sym_constexpr] = ACTIONS(2220),
    [anon_sym_volatile] = ACTIONS(2220),
    [anon_sym_restrict] = ACTIONS(2220),
    [anon_sym___restrict__] = ACTIONS(2220),
    [anon_sym__Atomic] = ACTIONS(2220),
    [anon_sym__Noreturn] = ACTIONS(2220),
    [anon_sym__Nonnull] = ACTIONS(2220),
    [anon_sym_nullable] = ACTIONS(2220),
    [anon_sym__Complex] = ACTIONS(2220),
    [anon_sym__Nullable] = ACTIONS(2220),
    [anon_sym__Nullable_result] = ACTIONS(2220),
    [anon_sym__Null_unspecified] = ACTIONS(2220),
    [anon_sym___autoreleasing] = ACTIONS(2220),
    [anon_sym___block] = ACTIONS(2220),
    [anon_sym___bridge] = ACTIONS(2220),
    [anon_sym___bridge_retained] = ACTIONS(2220),
    [anon_sym___bridge_transfer] = ACTIONS(2220),
    [anon_sym___complex] = ACTIONS(2220),
    [anon_sym___const] = ACTIONS(2220),
    [anon_sym___imag] = ACTIONS(2220),
    [anon_sym___kindof] = ACTIONS(2220),
    [anon_sym___nonnull] = ACTIONS(2220),
    [anon_sym___nullable] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2220),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2220),
    [anon_sym___real] = ACTIONS(2220),
    [anon_sym___strong] = ACTIONS(2220),
    [anon_sym___unsafe_unretained] = ACTIONS(2220),
    [anon_sym___unused] = ACTIONS(2220),
    [anon_sym___weak] = ACTIONS(2220),
    [anon_sym_alignas] = ACTIONS(2220),
    [anon_sym__Alignas] = ACTIONS(2220),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2220),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2220),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2220),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2220),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2220),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_API_AVAILABLE] = ACTIONS(2220),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_API_DEPRECATED] = ACTIONS(2220),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2220),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2220),
    [anon_sym___deprecated_msg] = ACTIONS(2220),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2220),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2220),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2220),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2220),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2220),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(827)] = {
    [sym_identifier] = ACTIONS(2482),
    [aux_sym_preproc_include_token1] = ACTIONS(2482),
    [aux_sym_preproc_include_token2] = ACTIONS(2482),
    [aux_sym_preproc_def_token1] = ACTIONS(2482),
    [aux_sym_preproc_if_token1] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2482),
    [sym_preproc_directive] = ACTIONS(2482),
    [anon_sym_LPAREN2] = ACTIONS(2484),
    [anon_sym_BANG] = ACTIONS(2484),
    [anon_sym_TILDE] = ACTIONS(2484),
    [anon_sym_DASH] = ACTIONS(2482),
    [anon_sym_PLUS] = ACTIONS(2482),
    [anon_sym_STAR] = ACTIONS(2484),
    [anon_sym_CARET] = ACTIONS(2484),
    [anon_sym_AMP] = ACTIONS(2484),
    [anon_sym_SEMI] = ACTIONS(2484),
    [anon_sym___extension__] = ACTIONS(2482),
    [anon_sym_typedef] = ACTIONS(2482),
    [anon_sym_extern] = ACTIONS(2482),
    [anon_sym___attribute__] = ACTIONS(2482),
    [anon_sym___attribute] = ACTIONS(2482),
    [anon_sym_noreturn] = ACTIONS(2482),
    [anon_sym_LBRACK] = ACTIONS(2484),
    [anon_sym___declspec] = ACTIONS(2482),
    [anon_sym___cdecl] = ACTIONS(2482),
    [anon_sym___clrcall] = ACTIONS(2482),
    [anon_sym___stdcall] = ACTIONS(2482),
    [anon_sym___fastcall] = ACTIONS(2482),
    [anon_sym___thiscall] = ACTIONS(2482),
    [anon_sym___vectorcall] = ACTIONS(2482),
    [anon_sym_LBRACE] = ACTIONS(2484),
    [anon_sym_RBRACE] = ACTIONS(2484),
    [anon_sym_signed] = ACTIONS(2482),
    [anon_sym_unsigned] = ACTIONS(2482),
    [anon_sym_long] = ACTIONS(2482),
    [anon_sym_short] = ACTIONS(2482),
    [anon_sym_static] = ACTIONS(2482),
    [anon_sym_ATautoreleasepool] = ACTIONS(2484),
    [anon_sym_auto] = ACTIONS(2482),
    [anon_sym_register] = ACTIONS(2482),
    [anon_sym_inline] = ACTIONS(2482),
    [anon_sym___inline] = ACTIONS(2482),
    [anon_sym___inline__] = ACTIONS(2482),
    [anon_sym___forceinline] = ACTIONS(2482),
    [anon_sym_thread_local] = ACTIONS(2482),
    [anon_sym___thread] = ACTIONS(2482),
    [anon_sym_CG_EXTERN] = ACTIONS(2482),
    [anon_sym_CG_INLINE] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2482),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2482),
    [anon_sym_IBOutlet] = ACTIONS(2482),
    [anon_sym_IBInspectable] = ACTIONS(2482),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2482),
    [anon_sym_NS_INLINE] = ACTIONS(2482),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2482),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2482),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2482),
    [anon_sym_const] = ACTIONS(2482),
    [anon_sym_constexpr] = ACTIONS(2482),
    [anon_sym_volatile] = ACTIONS(2482),
    [anon_sym_restrict] = ACTIONS(2482),
    [anon_sym___restrict__] = ACTIONS(2482),
    [anon_sym__Atomic] = ACTIONS(2482),
    [anon_sym__Noreturn] = ACTIONS(2482),
    [anon_sym__Nonnull] = ACTIONS(2482),
    [anon_sym_nullable] = ACTIONS(2482),
    [anon_sym__Complex] = ACTIONS(2482),
    [anon_sym__Nullable] = ACTIONS(2482),
    [anon_sym__Nullable_result] = ACTIONS(2482),
    [anon_sym__Null_unspecified] = ACTIONS(2482),
    [anon_sym___autoreleasing] = ACTIONS(2482),
    [anon_sym___block] = ACTIONS(2482),
    [anon_sym___bridge] = ACTIONS(2482),
    [anon_sym___bridge_retained] = ACTIONS(2482),
    [anon_sym___bridge_transfer] = ACTIONS(2482),
    [anon_sym___complex] = ACTIONS(2482),
    [anon_sym___const] = ACTIONS(2482),
    [anon_sym___imag] = ACTIONS(2482),
    [anon_sym___kindof] = ACTIONS(2482),
    [anon_sym___nonnull] = ACTIONS(2482),
    [anon_sym___nullable] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2482),
    [anon_sym___real] = ACTIONS(2482),
    [anon_sym___strong] = ACTIONS(2482),
    [anon_sym___unsafe_unretained] = ACTIONS(2482),
    [anon_sym___unused] = ACTIONS(2482),
    [anon_sym___weak] = ACTIONS(2482),
    [anon_sym_alignas] = ACTIONS(2482),
    [anon_sym__Alignas] = ACTIONS(2482),
    [sym_primitive_type] = ACTIONS(2482),
    [anon_sym_enum] = ACTIONS(2482),
    [anon_sym_struct] = ACTIONS(2482),
    [anon_sym_union] = ACTIONS(2482),
    [anon_sym_if] = ACTIONS(2482),
    [anon_sym_else] = ACTIONS(2482),
    [anon_sym_switch] = ACTIONS(2482),
    [anon_sym_case] = ACTIONS(2482),
    [anon_sym_default] = ACTIONS(2482),
    [anon_sym_while] = ACTIONS(2482),
    [anon_sym_do] = ACTIONS(2482),
    [anon_sym_for] = ACTIONS(2482),
    [anon_sym_in] = ACTIONS(2482),
    [anon_sym_return] = ACTIONS(2482),
    [anon_sym_break] = ACTIONS(2482),
    [anon_sym_continue] = ACTIONS(2482),
    [anon_sym_goto] = ACTIONS(2482),
    [anon_sym_DASH_DASH] = ACTIONS(2484),
    [anon_sym_PLUS_PLUS] = ACTIONS(2484),
    [anon_sym_sizeof] = ACTIONS(2482),
    [anon_sym___alignof__] = ACTIONS(2482),
    [anon_sym___alignof] = ACTIONS(2482),
    [anon_sym__alignof] = ACTIONS(2482),
    [anon_sym_alignof] = ACTIONS(2482),
    [anon_sym__Alignof] = ACTIONS(2482),
    [anon_sym_offsetof] = ACTIONS(2482),
    [anon_sym__Generic] = ACTIONS(2482),
    [anon_sym_asm] = ACTIONS(2482),
    [anon_sym___asm__] = ACTIONS(2482),
    [anon_sym___asm] = ACTIONS(2482),
    [sym_number_literal] = ACTIONS(2484),
    [anon_sym_L_SQUOTE] = ACTIONS(2484),
    [anon_sym_u_SQUOTE] = ACTIONS(2484),
    [anon_sym_U_SQUOTE] = ACTIONS(2484),
    [anon_sym_u8_SQUOTE] = ACTIONS(2484),
    [anon_sym_SQUOTE] = ACTIONS(2484),
    [anon_sym_AT] = ACTIONS(2482),
    [anon_sym_DQUOTE] = ACTIONS(2484),
    [anon_sym_L_DQUOTE] = ACTIONS(2484),
    [anon_sym_u_DQUOTE] = ACTIONS(2484),
    [anon_sym_U_DQUOTE] = ACTIONS(2484),
    [anon_sym_u8_DQUOTE] = ACTIONS(2484),
    [sym_true] = ACTIONS(2482),
    [sym_false] = ACTIONS(2482),
    [anon_sym_NULL] = ACTIONS(2482),
    [anon_sym_nullptr] = ACTIONS(2482),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2482),
    [anon_sym___typeof] = ACTIONS(2482),
    [anon_sym_typeof] = ACTIONS(2482),
    [anon_sym_ATimport] = ACTIONS(2484),
    [aux_sym_preproc_undef_token1] = ACTIONS(2482),
    [anon_sym_POUND] = ACTIONS(2482),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2482),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2482),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2482),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_API_AVAILABLE] = ACTIONS(2482),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_API_DEPRECATED] = ACTIONS(2482),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym___deprecated_msg] = ACTIONS(2482),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2482),
    [anon_sym_ATprotocol] = ACTIONS(2484),
    [anon_sym_ATinterface] = ACTIONS(2484),
    [anon_sym_ATimplementation] = ACTIONS(2484),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2484),
    [anon_sym_ATtry] = ACTIONS(2484),
    [anon_sym___try] = ACTIONS(2482),
    [anon_sym_ATthrow] = ACTIONS(2484),
    [anon_sym_ATselector] = ACTIONS(2484),
    [anon_sym_ATavailable] = ACTIONS(2484),
    [anon_sym___builtin_available] = ACTIONS(2482),
    [anon_sym_va_arg] = ACTIONS(2482),
    [anon_sym_ATencode] = ACTIONS(2484),
    [anon_sym_ATsynchronized] = ACTIONS(2484),
    [anon_sym_BOOL] = ACTIONS(2482),
    [anon_sym_IMP] = ACTIONS(2482),
    [anon_sym_SEL] = ACTIONS(2482),
    [anon_sym_Class] = ACTIONS(2482),
    [anon_sym_id] = ACTIONS(2482),
  },
  [STATE(828)] = {
    [sym_identifier] = ACTIONS(2486),
    [aux_sym_preproc_include_token1] = ACTIONS(2486),
    [aux_sym_preproc_include_token2] = ACTIONS(2486),
    [aux_sym_preproc_def_token1] = ACTIONS(2486),
    [aux_sym_preproc_if_token1] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2486),
    [sym_preproc_directive] = ACTIONS(2486),
    [anon_sym_LPAREN2] = ACTIONS(2488),
    [anon_sym_BANG] = ACTIONS(2488),
    [anon_sym_TILDE] = ACTIONS(2488),
    [anon_sym_DASH] = ACTIONS(2486),
    [anon_sym_PLUS] = ACTIONS(2486),
    [anon_sym_STAR] = ACTIONS(2488),
    [anon_sym_CARET] = ACTIONS(2488),
    [anon_sym_AMP] = ACTIONS(2488),
    [anon_sym_SEMI] = ACTIONS(2488),
    [anon_sym___extension__] = ACTIONS(2486),
    [anon_sym_typedef] = ACTIONS(2486),
    [anon_sym_extern] = ACTIONS(2486),
    [anon_sym___attribute__] = ACTIONS(2486),
    [anon_sym___attribute] = ACTIONS(2486),
    [anon_sym_noreturn] = ACTIONS(2486),
    [anon_sym_LBRACK] = ACTIONS(2488),
    [anon_sym___declspec] = ACTIONS(2486),
    [anon_sym___cdecl] = ACTIONS(2486),
    [anon_sym___clrcall] = ACTIONS(2486),
    [anon_sym___stdcall] = ACTIONS(2486),
    [anon_sym___fastcall] = ACTIONS(2486),
    [anon_sym___thiscall] = ACTIONS(2486),
    [anon_sym___vectorcall] = ACTIONS(2486),
    [anon_sym_LBRACE] = ACTIONS(2488),
    [anon_sym_RBRACE] = ACTIONS(2488),
    [anon_sym_signed] = ACTIONS(2486),
    [anon_sym_unsigned] = ACTIONS(2486),
    [anon_sym_long] = ACTIONS(2486),
    [anon_sym_short] = ACTIONS(2486),
    [anon_sym_static] = ACTIONS(2486),
    [anon_sym_ATautoreleasepool] = ACTIONS(2488),
    [anon_sym_auto] = ACTIONS(2486),
    [anon_sym_register] = ACTIONS(2486),
    [anon_sym_inline] = ACTIONS(2486),
    [anon_sym___inline] = ACTIONS(2486),
    [anon_sym___inline__] = ACTIONS(2486),
    [anon_sym___forceinline] = ACTIONS(2486),
    [anon_sym_thread_local] = ACTIONS(2486),
    [anon_sym___thread] = ACTIONS(2486),
    [anon_sym_CG_EXTERN] = ACTIONS(2486),
    [anon_sym_CG_INLINE] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2486),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2486),
    [anon_sym_IBOutlet] = ACTIONS(2486),
    [anon_sym_IBInspectable] = ACTIONS(2486),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2486),
    [anon_sym_NS_INLINE] = ACTIONS(2486),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2486),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2486),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2486),
    [anon_sym_const] = ACTIONS(2486),
    [anon_sym_constexpr] = ACTIONS(2486),
    [anon_sym_volatile] = ACTIONS(2486),
    [anon_sym_restrict] = ACTIONS(2486),
    [anon_sym___restrict__] = ACTIONS(2486),
    [anon_sym__Atomic] = ACTIONS(2486),
    [anon_sym__Noreturn] = ACTIONS(2486),
    [anon_sym__Nonnull] = ACTIONS(2486),
    [anon_sym_nullable] = ACTIONS(2486),
    [anon_sym__Complex] = ACTIONS(2486),
    [anon_sym__Nullable] = ACTIONS(2486),
    [anon_sym__Nullable_result] = ACTIONS(2486),
    [anon_sym__Null_unspecified] = ACTIONS(2486),
    [anon_sym___autoreleasing] = ACTIONS(2486),
    [anon_sym___block] = ACTIONS(2486),
    [anon_sym___bridge] = ACTIONS(2486),
    [anon_sym___bridge_retained] = ACTIONS(2486),
    [anon_sym___bridge_transfer] = ACTIONS(2486),
    [anon_sym___complex] = ACTIONS(2486),
    [anon_sym___const] = ACTIONS(2486),
    [anon_sym___imag] = ACTIONS(2486),
    [anon_sym___kindof] = ACTIONS(2486),
    [anon_sym___nonnull] = ACTIONS(2486),
    [anon_sym___nullable] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2486),
    [anon_sym___real] = ACTIONS(2486),
    [anon_sym___strong] = ACTIONS(2486),
    [anon_sym___unsafe_unretained] = ACTIONS(2486),
    [anon_sym___unused] = ACTIONS(2486),
    [anon_sym___weak] = ACTIONS(2486),
    [anon_sym_alignas] = ACTIONS(2486),
    [anon_sym__Alignas] = ACTIONS(2486),
    [sym_primitive_type] = ACTIONS(2486),
    [anon_sym_enum] = ACTIONS(2486),
    [anon_sym_struct] = ACTIONS(2486),
    [anon_sym_union] = ACTIONS(2486),
    [anon_sym_if] = ACTIONS(2486),
    [anon_sym_else] = ACTIONS(2486),
    [anon_sym_switch] = ACTIONS(2486),
    [anon_sym_case] = ACTIONS(2486),
    [anon_sym_default] = ACTIONS(2486),
    [anon_sym_while] = ACTIONS(2486),
    [anon_sym_do] = ACTIONS(2486),
    [anon_sym_for] = ACTIONS(2486),
    [anon_sym_in] = ACTIONS(2486),
    [anon_sym_return] = ACTIONS(2486),
    [anon_sym_break] = ACTIONS(2486),
    [anon_sym_continue] = ACTIONS(2486),
    [anon_sym_goto] = ACTIONS(2486),
    [anon_sym_DASH_DASH] = ACTIONS(2488),
    [anon_sym_PLUS_PLUS] = ACTIONS(2488),
    [anon_sym_sizeof] = ACTIONS(2486),
    [anon_sym___alignof__] = ACTIONS(2486),
    [anon_sym___alignof] = ACTIONS(2486),
    [anon_sym__alignof] = ACTIONS(2486),
    [anon_sym_alignof] = ACTIONS(2486),
    [anon_sym__Alignof] = ACTIONS(2486),
    [anon_sym_offsetof] = ACTIONS(2486),
    [anon_sym__Generic] = ACTIONS(2486),
    [anon_sym_asm] = ACTIONS(2486),
    [anon_sym___asm__] = ACTIONS(2486),
    [anon_sym___asm] = ACTIONS(2486),
    [sym_number_literal] = ACTIONS(2488),
    [anon_sym_L_SQUOTE] = ACTIONS(2488),
    [anon_sym_u_SQUOTE] = ACTIONS(2488),
    [anon_sym_U_SQUOTE] = ACTIONS(2488),
    [anon_sym_u8_SQUOTE] = ACTIONS(2488),
    [anon_sym_SQUOTE] = ACTIONS(2488),
    [anon_sym_AT] = ACTIONS(2486),
    [anon_sym_DQUOTE] = ACTIONS(2488),
    [anon_sym_L_DQUOTE] = ACTIONS(2488),
    [anon_sym_u_DQUOTE] = ACTIONS(2488),
    [anon_sym_U_DQUOTE] = ACTIONS(2488),
    [anon_sym_u8_DQUOTE] = ACTIONS(2488),
    [sym_true] = ACTIONS(2486),
    [sym_false] = ACTIONS(2486),
    [anon_sym_NULL] = ACTIONS(2486),
    [anon_sym_nullptr] = ACTIONS(2486),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2486),
    [anon_sym___typeof] = ACTIONS(2486),
    [anon_sym_typeof] = ACTIONS(2486),
    [anon_sym_ATimport] = ACTIONS(2488),
    [aux_sym_preproc_undef_token1] = ACTIONS(2486),
    [anon_sym_POUND] = ACTIONS(2486),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2486),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2486),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2486),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_API_AVAILABLE] = ACTIONS(2486),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_API_DEPRECATED] = ACTIONS(2486),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym___deprecated_msg] = ACTIONS(2486),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2486),
    [anon_sym_ATprotocol] = ACTIONS(2488),
    [anon_sym_ATinterface] = ACTIONS(2488),
    [anon_sym_ATimplementation] = ACTIONS(2488),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2488),
    [anon_sym_ATtry] = ACTIONS(2488),
    [anon_sym___try] = ACTIONS(2486),
    [anon_sym_ATthrow] = ACTIONS(2488),
    [anon_sym_ATselector] = ACTIONS(2488),
    [anon_sym_ATavailable] = ACTIONS(2488),
    [anon_sym___builtin_available] = ACTIONS(2486),
    [anon_sym_va_arg] = ACTIONS(2486),
    [anon_sym_ATencode] = ACTIONS(2488),
    [anon_sym_ATsynchronized] = ACTIONS(2488),
    [anon_sym_BOOL] = ACTIONS(2486),
    [anon_sym_IMP] = ACTIONS(2486),
    [anon_sym_SEL] = ACTIONS(2486),
    [anon_sym_Class] = ACTIONS(2486),
    [anon_sym_id] = ACTIONS(2486),
  },
  [STATE(829)] = {
    [sym_identifier] = ACTIONS(2490),
    [aux_sym_preproc_include_token1] = ACTIONS(2490),
    [aux_sym_preproc_include_token2] = ACTIONS(2490),
    [aux_sym_preproc_def_token1] = ACTIONS(2490),
    [aux_sym_preproc_if_token1] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2490),
    [sym_preproc_directive] = ACTIONS(2490),
    [anon_sym_LPAREN2] = ACTIONS(2492),
    [anon_sym_BANG] = ACTIONS(2492),
    [anon_sym_TILDE] = ACTIONS(2492),
    [anon_sym_DASH] = ACTIONS(2490),
    [anon_sym_PLUS] = ACTIONS(2490),
    [anon_sym_STAR] = ACTIONS(2492),
    [anon_sym_CARET] = ACTIONS(2492),
    [anon_sym_AMP] = ACTIONS(2492),
    [anon_sym_SEMI] = ACTIONS(2492),
    [anon_sym___extension__] = ACTIONS(2490),
    [anon_sym_typedef] = ACTIONS(2490),
    [anon_sym_extern] = ACTIONS(2490),
    [anon_sym___attribute__] = ACTIONS(2490),
    [anon_sym___attribute] = ACTIONS(2490),
    [anon_sym_noreturn] = ACTIONS(2490),
    [anon_sym_LBRACK] = ACTIONS(2492),
    [anon_sym___declspec] = ACTIONS(2490),
    [anon_sym___cdecl] = ACTIONS(2490),
    [anon_sym___clrcall] = ACTIONS(2490),
    [anon_sym___stdcall] = ACTIONS(2490),
    [anon_sym___fastcall] = ACTIONS(2490),
    [anon_sym___thiscall] = ACTIONS(2490),
    [anon_sym___vectorcall] = ACTIONS(2490),
    [anon_sym_LBRACE] = ACTIONS(2492),
    [anon_sym_RBRACE] = ACTIONS(2492),
    [anon_sym_signed] = ACTIONS(2490),
    [anon_sym_unsigned] = ACTIONS(2490),
    [anon_sym_long] = ACTIONS(2490),
    [anon_sym_short] = ACTIONS(2490),
    [anon_sym_static] = ACTIONS(2490),
    [anon_sym_ATautoreleasepool] = ACTIONS(2492),
    [anon_sym_auto] = ACTIONS(2490),
    [anon_sym_register] = ACTIONS(2490),
    [anon_sym_inline] = ACTIONS(2490),
    [anon_sym___inline] = ACTIONS(2490),
    [anon_sym___inline__] = ACTIONS(2490),
    [anon_sym___forceinline] = ACTIONS(2490),
    [anon_sym_thread_local] = ACTIONS(2490),
    [anon_sym___thread] = ACTIONS(2490),
    [anon_sym_CG_EXTERN] = ACTIONS(2490),
    [anon_sym_CG_INLINE] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2490),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2490),
    [anon_sym_IBOutlet] = ACTIONS(2490),
    [anon_sym_IBInspectable] = ACTIONS(2490),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2490),
    [anon_sym_NS_INLINE] = ACTIONS(2490),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2490),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2490),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2490),
    [anon_sym_const] = ACTIONS(2490),
    [anon_sym_constexpr] = ACTIONS(2490),
    [anon_sym_volatile] = ACTIONS(2490),
    [anon_sym_restrict] = ACTIONS(2490),
    [anon_sym___restrict__] = ACTIONS(2490),
    [anon_sym__Atomic] = ACTIONS(2490),
    [anon_sym__Noreturn] = ACTIONS(2490),
    [anon_sym__Nonnull] = ACTIONS(2490),
    [anon_sym_nullable] = ACTIONS(2490),
    [anon_sym__Complex] = ACTIONS(2490),
    [anon_sym__Nullable] = ACTIONS(2490),
    [anon_sym__Nullable_result] = ACTIONS(2490),
    [anon_sym__Null_unspecified] = ACTIONS(2490),
    [anon_sym___autoreleasing] = ACTIONS(2490),
    [anon_sym___block] = ACTIONS(2490),
    [anon_sym___bridge] = ACTIONS(2490),
    [anon_sym___bridge_retained] = ACTIONS(2490),
    [anon_sym___bridge_transfer] = ACTIONS(2490),
    [anon_sym___complex] = ACTIONS(2490),
    [anon_sym___const] = ACTIONS(2490),
    [anon_sym___imag] = ACTIONS(2490),
    [anon_sym___kindof] = ACTIONS(2490),
    [anon_sym___nonnull] = ACTIONS(2490),
    [anon_sym___nullable] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2490),
    [anon_sym___real] = ACTIONS(2490),
    [anon_sym___strong] = ACTIONS(2490),
    [anon_sym___unsafe_unretained] = ACTIONS(2490),
    [anon_sym___unused] = ACTIONS(2490),
    [anon_sym___weak] = ACTIONS(2490),
    [anon_sym_alignas] = ACTIONS(2490),
    [anon_sym__Alignas] = ACTIONS(2490),
    [sym_primitive_type] = ACTIONS(2490),
    [anon_sym_enum] = ACTIONS(2490),
    [anon_sym_struct] = ACTIONS(2490),
    [anon_sym_union] = ACTIONS(2490),
    [anon_sym_if] = ACTIONS(2490),
    [anon_sym_else] = ACTIONS(2490),
    [anon_sym_switch] = ACTIONS(2490),
    [anon_sym_case] = ACTIONS(2490),
    [anon_sym_default] = ACTIONS(2490),
    [anon_sym_while] = ACTIONS(2490),
    [anon_sym_do] = ACTIONS(2490),
    [anon_sym_for] = ACTIONS(2490),
    [anon_sym_in] = ACTIONS(2490),
    [anon_sym_return] = ACTIONS(2490),
    [anon_sym_break] = ACTIONS(2490),
    [anon_sym_continue] = ACTIONS(2490),
    [anon_sym_goto] = ACTIONS(2490),
    [anon_sym_DASH_DASH] = ACTIONS(2492),
    [anon_sym_PLUS_PLUS] = ACTIONS(2492),
    [anon_sym_sizeof] = ACTIONS(2490),
    [anon_sym___alignof__] = ACTIONS(2490),
    [anon_sym___alignof] = ACTIONS(2490),
    [anon_sym__alignof] = ACTIONS(2490),
    [anon_sym_alignof] = ACTIONS(2490),
    [anon_sym__Alignof] = ACTIONS(2490),
    [anon_sym_offsetof] = ACTIONS(2490),
    [anon_sym__Generic] = ACTIONS(2490),
    [anon_sym_asm] = ACTIONS(2490),
    [anon_sym___asm__] = ACTIONS(2490),
    [anon_sym___asm] = ACTIONS(2490),
    [sym_number_literal] = ACTIONS(2492),
    [anon_sym_L_SQUOTE] = ACTIONS(2492),
    [anon_sym_u_SQUOTE] = ACTIONS(2492),
    [anon_sym_U_SQUOTE] = ACTIONS(2492),
    [anon_sym_u8_SQUOTE] = ACTIONS(2492),
    [anon_sym_SQUOTE] = ACTIONS(2492),
    [anon_sym_AT] = ACTIONS(2490),
    [anon_sym_DQUOTE] = ACTIONS(2492),
    [anon_sym_L_DQUOTE] = ACTIONS(2492),
    [anon_sym_u_DQUOTE] = ACTIONS(2492),
    [anon_sym_U_DQUOTE] = ACTIONS(2492),
    [anon_sym_u8_DQUOTE] = ACTIONS(2492),
    [sym_true] = ACTIONS(2490),
    [sym_false] = ACTIONS(2490),
    [anon_sym_NULL] = ACTIONS(2490),
    [anon_sym_nullptr] = ACTIONS(2490),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2490),
    [anon_sym___typeof] = ACTIONS(2490),
    [anon_sym_typeof] = ACTIONS(2490),
    [anon_sym_ATimport] = ACTIONS(2492),
    [aux_sym_preproc_undef_token1] = ACTIONS(2490),
    [anon_sym_POUND] = ACTIONS(2490),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2490),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2490),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2490),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_API_AVAILABLE] = ACTIONS(2490),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_API_DEPRECATED] = ACTIONS(2490),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym___deprecated_msg] = ACTIONS(2490),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2490),
    [anon_sym_ATprotocol] = ACTIONS(2492),
    [anon_sym_ATinterface] = ACTIONS(2492),
    [anon_sym_ATimplementation] = ACTIONS(2492),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2492),
    [anon_sym_ATtry] = ACTIONS(2492),
    [anon_sym___try] = ACTIONS(2490),
    [anon_sym_ATthrow] = ACTIONS(2492),
    [anon_sym_ATselector] = ACTIONS(2492),
    [anon_sym_ATavailable] = ACTIONS(2492),
    [anon_sym___builtin_available] = ACTIONS(2490),
    [anon_sym_va_arg] = ACTIONS(2490),
    [anon_sym_ATencode] = ACTIONS(2492),
    [anon_sym_ATsynchronized] = ACTIONS(2492),
    [anon_sym_BOOL] = ACTIONS(2490),
    [anon_sym_IMP] = ACTIONS(2490),
    [anon_sym_SEL] = ACTIONS(2490),
    [anon_sym_Class] = ACTIONS(2490),
    [anon_sym_id] = ACTIONS(2490),
  },
  [STATE(830)] = {
    [ts_builtin_sym_end] = ACTIONS(2418),
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(831)] = {
    [ts_builtin_sym_end] = ACTIONS(2422),
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(832)] = {
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_RBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(833)] = {
    [sym_identifier] = ACTIONS(2700),
    [aux_sym_preproc_include_token1] = ACTIONS(2700),
    [aux_sym_preproc_include_token2] = ACTIONS(2700),
    [aux_sym_preproc_def_token1] = ACTIONS(2700),
    [aux_sym_preproc_if_token1] = ACTIONS(2700),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2700),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2700),
    [sym_preproc_directive] = ACTIONS(2700),
    [anon_sym_LPAREN2] = ACTIONS(2702),
    [anon_sym_BANG] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2700),
    [anon_sym_PLUS] = ACTIONS(2700),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym___extension__] = ACTIONS(2700),
    [anon_sym_typedef] = ACTIONS(2700),
    [anon_sym_extern] = ACTIONS(2700),
    [anon_sym___attribute__] = ACTIONS(2700),
    [anon_sym___attribute] = ACTIONS(2700),
    [anon_sym_noreturn] = ACTIONS(2700),
    [anon_sym_LBRACK] = ACTIONS(2702),
    [anon_sym___declspec] = ACTIONS(2700),
    [anon_sym___cdecl] = ACTIONS(2700),
    [anon_sym___clrcall] = ACTIONS(2700),
    [anon_sym___stdcall] = ACTIONS(2700),
    [anon_sym___fastcall] = ACTIONS(2700),
    [anon_sym___thiscall] = ACTIONS(2700),
    [anon_sym___vectorcall] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2702),
    [anon_sym_RBRACE] = ACTIONS(2702),
    [anon_sym_signed] = ACTIONS(2700),
    [anon_sym_unsigned] = ACTIONS(2700),
    [anon_sym_long] = ACTIONS(2700),
    [anon_sym_short] = ACTIONS(2700),
    [anon_sym_static] = ACTIONS(2700),
    [anon_sym_ATautoreleasepool] = ACTIONS(2702),
    [anon_sym_auto] = ACTIONS(2700),
    [anon_sym_register] = ACTIONS(2700),
    [anon_sym_inline] = ACTIONS(2700),
    [anon_sym___inline] = ACTIONS(2700),
    [anon_sym___inline__] = ACTIONS(2700),
    [anon_sym___forceinline] = ACTIONS(2700),
    [anon_sym_thread_local] = ACTIONS(2700),
    [anon_sym___thread] = ACTIONS(2700),
    [anon_sym_CG_EXTERN] = ACTIONS(2700),
    [anon_sym_CG_INLINE] = ACTIONS(2700),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2700),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2700),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2700),
    [anon_sym_IBOutlet] = ACTIONS(2700),
    [anon_sym_IBInspectable] = ACTIONS(2700),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2700),
    [anon_sym_NS_INLINE] = ACTIONS(2700),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2700),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2700),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2700),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2700),
    [anon_sym_const] = ACTIONS(2700),
    [anon_sym_constexpr] = ACTIONS(2700),
    [anon_sym_volatile] = ACTIONS(2700),
    [anon_sym_restrict] = ACTIONS(2700),
    [anon_sym___restrict__] = ACTIONS(2700),
    [anon_sym__Atomic] = ACTIONS(2700),
    [anon_sym__Noreturn] = ACTIONS(2700),
    [anon_sym__Nonnull] = ACTIONS(2700),
    [anon_sym_nullable] = ACTIONS(2700),
    [anon_sym__Complex] = ACTIONS(2700),
    [anon_sym__Nullable] = ACTIONS(2700),
    [anon_sym__Nullable_result] = ACTIONS(2700),
    [anon_sym__Null_unspecified] = ACTIONS(2700),
    [anon_sym___autoreleasing] = ACTIONS(2700),
    [anon_sym___block] = ACTIONS(2700),
    [anon_sym___bridge] = ACTIONS(2700),
    [anon_sym___bridge_retained] = ACTIONS(2700),
    [anon_sym___bridge_transfer] = ACTIONS(2700),
    [anon_sym___complex] = ACTIONS(2700),
    [anon_sym___const] = ACTIONS(2700),
    [anon_sym___imag] = ACTIONS(2700),
    [anon_sym___kindof] = ACTIONS(2700),
    [anon_sym___nonnull] = ACTIONS(2700),
    [anon_sym___nullable] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2700),
    [anon_sym___real] = ACTIONS(2700),
    [anon_sym___strong] = ACTIONS(2700),
    [anon_sym___unsafe_unretained] = ACTIONS(2700),
    [anon_sym___unused] = ACTIONS(2700),
    [anon_sym___weak] = ACTIONS(2700),
    [anon_sym_alignas] = ACTIONS(2700),
    [anon_sym__Alignas] = ACTIONS(2700),
    [sym_primitive_type] = ACTIONS(2700),
    [anon_sym_enum] = ACTIONS(2700),
    [anon_sym_struct] = ACTIONS(2700),
    [anon_sym_union] = ACTIONS(2700),
    [anon_sym_if] = ACTIONS(2700),
    [anon_sym_switch] = ACTIONS(2700),
    [anon_sym_case] = ACTIONS(2700),
    [anon_sym_default] = ACTIONS(2700),
    [anon_sym_while] = ACTIONS(2700),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_for] = ACTIONS(2700),
    [anon_sym_in] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_DASH_DASH] = ACTIONS(2702),
    [anon_sym_PLUS_PLUS] = ACTIONS(2702),
    [anon_sym_sizeof] = ACTIONS(2700),
    [anon_sym___alignof__] = ACTIONS(2700),
    [anon_sym___alignof] = ACTIONS(2700),
    [anon_sym__alignof] = ACTIONS(2700),
    [anon_sym_alignof] = ACTIONS(2700),
    [anon_sym__Alignof] = ACTIONS(2700),
    [anon_sym_offsetof] = ACTIONS(2700),
    [anon_sym__Generic] = ACTIONS(2700),
    [anon_sym_asm] = ACTIONS(2700),
    [anon_sym___asm__] = ACTIONS(2700),
    [anon_sym___asm] = ACTIONS(2700),
    [sym_number_literal] = ACTIONS(2702),
    [anon_sym_L_SQUOTE] = ACTIONS(2702),
    [anon_sym_u_SQUOTE] = ACTIONS(2702),
    [anon_sym_U_SQUOTE] = ACTIONS(2702),
    [anon_sym_u8_SQUOTE] = ACTIONS(2702),
    [anon_sym_SQUOTE] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_DQUOTE] = ACTIONS(2702),
    [anon_sym_L_DQUOTE] = ACTIONS(2702),
    [anon_sym_u_DQUOTE] = ACTIONS(2702),
    [anon_sym_U_DQUOTE] = ACTIONS(2702),
    [anon_sym_u8_DQUOTE] = ACTIONS(2702),
    [sym_true] = ACTIONS(2700),
    [sym_false] = ACTIONS(2700),
    [anon_sym_NULL] = ACTIONS(2700),
    [anon_sym_nullptr] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2700),
    [anon_sym___typeof] = ACTIONS(2700),
    [anon_sym_typeof] = ACTIONS(2700),
    [anon_sym_ATimport] = ACTIONS(2702),
    [aux_sym_preproc_undef_token1] = ACTIONS(2700),
    [anon_sym_POUND] = ACTIONS(2700),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2700),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2700),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2700),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2700),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2700),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2700),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2700),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_API_AVAILABLE] = ACTIONS(2700),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_API_DEPRECATED] = ACTIONS(2700),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2700),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym___deprecated_msg] = ACTIONS(2700),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2700),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2700),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2700),
    [anon_sym_ATprotocol] = ACTIONS(2702),
    [anon_sym_ATinterface] = ACTIONS(2702),
    [anon_sym_ATimplementation] = ACTIONS(2702),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2702),
    [anon_sym_ATtry] = ACTIONS(2702),
    [anon_sym___try] = ACTIONS(2700),
    [anon_sym_ATthrow] = ACTIONS(2702),
    [anon_sym_ATselector] = ACTIONS(2702),
    [anon_sym_ATavailable] = ACTIONS(2702),
    [anon_sym___builtin_available] = ACTIONS(2700),
    [anon_sym_va_arg] = ACTIONS(2700),
    [anon_sym_ATencode] = ACTIONS(2702),
    [anon_sym_ATsynchronized] = ACTIONS(2702),
    [anon_sym_BOOL] = ACTIONS(2700),
    [anon_sym_IMP] = ACTIONS(2700),
    [anon_sym_SEL] = ACTIONS(2700),
    [anon_sym_Class] = ACTIONS(2700),
    [anon_sym_id] = ACTIONS(2700),
  },
  [STATE(834)] = {
    [sym_identifier] = ACTIONS(2864),
    [aux_sym_preproc_include_token1] = ACTIONS(2864),
    [aux_sym_preproc_include_token2] = ACTIONS(2864),
    [aux_sym_preproc_def_token1] = ACTIONS(2864),
    [aux_sym_preproc_if_token1] = ACTIONS(2864),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2864),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2864),
    [sym_preproc_directive] = ACTIONS(2864),
    [anon_sym_LPAREN2] = ACTIONS(2866),
    [anon_sym_BANG] = ACTIONS(2866),
    [anon_sym_TILDE] = ACTIONS(2866),
    [anon_sym_DASH] = ACTIONS(2864),
    [anon_sym_PLUS] = ACTIONS(2864),
    [anon_sym_STAR] = ACTIONS(2866),
    [anon_sym_CARET] = ACTIONS(2866),
    [anon_sym_AMP] = ACTIONS(2866),
    [anon_sym_SEMI] = ACTIONS(2866),
    [anon_sym___extension__] = ACTIONS(2864),
    [anon_sym_typedef] = ACTIONS(2864),
    [anon_sym_extern] = ACTIONS(2864),
    [anon_sym___attribute__] = ACTIONS(2864),
    [anon_sym___attribute] = ACTIONS(2864),
    [anon_sym_noreturn] = ACTIONS(2864),
    [anon_sym_LBRACK] = ACTIONS(2866),
    [anon_sym___declspec] = ACTIONS(2864),
    [anon_sym___cdecl] = ACTIONS(2864),
    [anon_sym___clrcall] = ACTIONS(2864),
    [anon_sym___stdcall] = ACTIONS(2864),
    [anon_sym___fastcall] = ACTIONS(2864),
    [anon_sym___thiscall] = ACTIONS(2864),
    [anon_sym___vectorcall] = ACTIONS(2864),
    [anon_sym_LBRACE] = ACTIONS(2866),
    [anon_sym_RBRACE] = ACTIONS(2866),
    [anon_sym_signed] = ACTIONS(2864),
    [anon_sym_unsigned] = ACTIONS(2864),
    [anon_sym_long] = ACTIONS(2864),
    [anon_sym_short] = ACTIONS(2864),
    [anon_sym_static] = ACTIONS(2864),
    [anon_sym_ATautoreleasepool] = ACTIONS(2866),
    [anon_sym_auto] = ACTIONS(2864),
    [anon_sym_register] = ACTIONS(2864),
    [anon_sym_inline] = ACTIONS(2864),
    [anon_sym___inline] = ACTIONS(2864),
    [anon_sym___inline__] = ACTIONS(2864),
    [anon_sym___forceinline] = ACTIONS(2864),
    [anon_sym_thread_local] = ACTIONS(2864),
    [anon_sym___thread] = ACTIONS(2864),
    [anon_sym_CG_EXTERN] = ACTIONS(2864),
    [anon_sym_CG_INLINE] = ACTIONS(2864),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2864),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2864),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2864),
    [anon_sym_IBOutlet] = ACTIONS(2864),
    [anon_sym_IBInspectable] = ACTIONS(2864),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2864),
    [anon_sym_NS_INLINE] = ACTIONS(2864),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2864),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2864),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2864),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2864),
    [anon_sym_const] = ACTIONS(2864),
    [anon_sym_constexpr] = ACTIONS(2864),
    [anon_sym_volatile] = ACTIONS(2864),
    [anon_sym_restrict] = ACTIONS(2864),
    [anon_sym___restrict__] = ACTIONS(2864),
    [anon_sym__Atomic] = ACTIONS(2864),
    [anon_sym__Noreturn] = ACTIONS(2864),
    [anon_sym__Nonnull] = ACTIONS(2864),
    [anon_sym_nullable] = ACTIONS(2864),
    [anon_sym__Complex] = ACTIONS(2864),
    [anon_sym__Nullable] = ACTIONS(2864),
    [anon_sym__Nullable_result] = ACTIONS(2864),
    [anon_sym__Null_unspecified] = ACTIONS(2864),
    [anon_sym___autoreleasing] = ACTIONS(2864),
    [anon_sym___block] = ACTIONS(2864),
    [anon_sym___bridge] = ACTIONS(2864),
    [anon_sym___bridge_retained] = ACTIONS(2864),
    [anon_sym___bridge_transfer] = ACTIONS(2864),
    [anon_sym___complex] = ACTIONS(2864),
    [anon_sym___const] = ACTIONS(2864),
    [anon_sym___imag] = ACTIONS(2864),
    [anon_sym___kindof] = ACTIONS(2864),
    [anon_sym___nonnull] = ACTIONS(2864),
    [anon_sym___nullable] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2864),
    [anon_sym___real] = ACTIONS(2864),
    [anon_sym___strong] = ACTIONS(2864),
    [anon_sym___unsafe_unretained] = ACTIONS(2864),
    [anon_sym___unused] = ACTIONS(2864),
    [anon_sym___weak] = ACTIONS(2864),
    [anon_sym_alignas] = ACTIONS(2864),
    [anon_sym__Alignas] = ACTIONS(2864),
    [sym_primitive_type] = ACTIONS(2864),
    [anon_sym_enum] = ACTIONS(2864),
    [anon_sym_struct] = ACTIONS(2864),
    [anon_sym_union] = ACTIONS(2864),
    [anon_sym_if] = ACTIONS(2864),
    [anon_sym_switch] = ACTIONS(2864),
    [anon_sym_case] = ACTIONS(2864),
    [anon_sym_default] = ACTIONS(2864),
    [anon_sym_while] = ACTIONS(2864),
    [anon_sym_do] = ACTIONS(2864),
    [anon_sym_for] = ACTIONS(2864),
    [anon_sym_in] = ACTIONS(2864),
    [anon_sym_return] = ACTIONS(2864),
    [anon_sym_break] = ACTIONS(2864),
    [anon_sym_continue] = ACTIONS(2864),
    [anon_sym_goto] = ACTIONS(2864),
    [anon_sym_DASH_DASH] = ACTIONS(2866),
    [anon_sym_PLUS_PLUS] = ACTIONS(2866),
    [anon_sym_sizeof] = ACTIONS(2864),
    [anon_sym___alignof__] = ACTIONS(2864),
    [anon_sym___alignof] = ACTIONS(2864),
    [anon_sym__alignof] = ACTIONS(2864),
    [anon_sym_alignof] = ACTIONS(2864),
    [anon_sym__Alignof] = ACTIONS(2864),
    [anon_sym_offsetof] = ACTIONS(2864),
    [anon_sym__Generic] = ACTIONS(2864),
    [anon_sym_asm] = ACTIONS(2864),
    [anon_sym___asm__] = ACTIONS(2864),
    [anon_sym___asm] = ACTIONS(2864),
    [sym_number_literal] = ACTIONS(2866),
    [anon_sym_L_SQUOTE] = ACTIONS(2866),
    [anon_sym_u_SQUOTE] = ACTIONS(2866),
    [anon_sym_U_SQUOTE] = ACTIONS(2866),
    [anon_sym_u8_SQUOTE] = ACTIONS(2866),
    [anon_sym_SQUOTE] = ACTIONS(2866),
    [anon_sym_AT] = ACTIONS(2864),
    [anon_sym_DQUOTE] = ACTIONS(2866),
    [anon_sym_L_DQUOTE] = ACTIONS(2866),
    [anon_sym_u_DQUOTE] = ACTIONS(2866),
    [anon_sym_U_DQUOTE] = ACTIONS(2866),
    [anon_sym_u8_DQUOTE] = ACTIONS(2866),
    [sym_true] = ACTIONS(2864),
    [sym_false] = ACTIONS(2864),
    [anon_sym_NULL] = ACTIONS(2864),
    [anon_sym_nullptr] = ACTIONS(2864),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2864),
    [anon_sym___typeof] = ACTIONS(2864),
    [anon_sym_typeof] = ACTIONS(2864),
    [anon_sym_ATimport] = ACTIONS(2866),
    [aux_sym_preproc_undef_token1] = ACTIONS(2864),
    [anon_sym_POUND] = ACTIONS(2864),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2864),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2864),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2864),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2864),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2864),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2864),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2864),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_API_AVAILABLE] = ACTIONS(2864),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_API_DEPRECATED] = ACTIONS(2864),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2864),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym___deprecated_msg] = ACTIONS(2864),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2864),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2864),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2864),
    [anon_sym_ATprotocol] = ACTIONS(2866),
    [anon_sym_ATinterface] = ACTIONS(2866),
    [anon_sym_ATimplementation] = ACTIONS(2866),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2866),
    [anon_sym_ATtry] = ACTIONS(2866),
    [anon_sym___try] = ACTIONS(2864),
    [anon_sym_ATthrow] = ACTIONS(2866),
    [anon_sym_ATselector] = ACTIONS(2866),
    [anon_sym_ATavailable] = ACTIONS(2866),
    [anon_sym___builtin_available] = ACTIONS(2864),
    [anon_sym_va_arg] = ACTIONS(2864),
    [anon_sym_ATencode] = ACTIONS(2866),
    [anon_sym_ATsynchronized] = ACTIONS(2866),
    [anon_sym_BOOL] = ACTIONS(2864),
    [anon_sym_IMP] = ACTIONS(2864),
    [anon_sym_SEL] = ACTIONS(2864),
    [anon_sym_Class] = ACTIONS(2864),
    [anon_sym_id] = ACTIONS(2864),
  },
  [STATE(835)] = {
    [sym_identifier] = ACTIONS(2760),
    [aux_sym_preproc_include_token1] = ACTIONS(2760),
    [aux_sym_preproc_include_token2] = ACTIONS(2760),
    [aux_sym_preproc_def_token1] = ACTIONS(2760),
    [aux_sym_preproc_if_token1] = ACTIONS(2760),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2760),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2760),
    [sym_preproc_directive] = ACTIONS(2760),
    [anon_sym_LPAREN2] = ACTIONS(2762),
    [anon_sym_BANG] = ACTIONS(2762),
    [anon_sym_TILDE] = ACTIONS(2762),
    [anon_sym_DASH] = ACTIONS(2760),
    [anon_sym_PLUS] = ACTIONS(2760),
    [anon_sym_STAR] = ACTIONS(2762),
    [anon_sym_CARET] = ACTIONS(2762),
    [anon_sym_AMP] = ACTIONS(2762),
    [anon_sym_SEMI] = ACTIONS(2762),
    [anon_sym___extension__] = ACTIONS(2760),
    [anon_sym_typedef] = ACTIONS(2760),
    [anon_sym_extern] = ACTIONS(2760),
    [anon_sym___attribute__] = ACTIONS(2760),
    [anon_sym___attribute] = ACTIONS(2760),
    [anon_sym_noreturn] = ACTIONS(2760),
    [anon_sym_LBRACK] = ACTIONS(2762),
    [anon_sym___declspec] = ACTIONS(2760),
    [anon_sym___cdecl] = ACTIONS(2760),
    [anon_sym___clrcall] = ACTIONS(2760),
    [anon_sym___stdcall] = ACTIONS(2760),
    [anon_sym___fastcall] = ACTIONS(2760),
    [anon_sym___thiscall] = ACTIONS(2760),
    [anon_sym___vectorcall] = ACTIONS(2760),
    [anon_sym_LBRACE] = ACTIONS(2762),
    [anon_sym_RBRACE] = ACTIONS(2762),
    [anon_sym_signed] = ACTIONS(2760),
    [anon_sym_unsigned] = ACTIONS(2760),
    [anon_sym_long] = ACTIONS(2760),
    [anon_sym_short] = ACTIONS(2760),
    [anon_sym_static] = ACTIONS(2760),
    [anon_sym_ATautoreleasepool] = ACTIONS(2762),
    [anon_sym_auto] = ACTIONS(2760),
    [anon_sym_register] = ACTIONS(2760),
    [anon_sym_inline] = ACTIONS(2760),
    [anon_sym___inline] = ACTIONS(2760),
    [anon_sym___inline__] = ACTIONS(2760),
    [anon_sym___forceinline] = ACTIONS(2760),
    [anon_sym_thread_local] = ACTIONS(2760),
    [anon_sym___thread] = ACTIONS(2760),
    [anon_sym_CG_EXTERN] = ACTIONS(2760),
    [anon_sym_CG_INLINE] = ACTIONS(2760),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2760),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2760),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2760),
    [anon_sym_IBOutlet] = ACTIONS(2760),
    [anon_sym_IBInspectable] = ACTIONS(2760),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2760),
    [anon_sym_NS_INLINE] = ACTIONS(2760),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2760),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2760),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2760),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2760),
    [anon_sym_const] = ACTIONS(2760),
    [anon_sym_constexpr] = ACTIONS(2760),
    [anon_sym_volatile] = ACTIONS(2760),
    [anon_sym_restrict] = ACTIONS(2760),
    [anon_sym___restrict__] = ACTIONS(2760),
    [anon_sym__Atomic] = ACTIONS(2760),
    [anon_sym__Noreturn] = ACTIONS(2760),
    [anon_sym__Nonnull] = ACTIONS(2760),
    [anon_sym_nullable] = ACTIONS(2760),
    [anon_sym__Complex] = ACTIONS(2760),
    [anon_sym__Nullable] = ACTIONS(2760),
    [anon_sym__Nullable_result] = ACTIONS(2760),
    [anon_sym__Null_unspecified] = ACTIONS(2760),
    [anon_sym___autoreleasing] = ACTIONS(2760),
    [anon_sym___block] = ACTIONS(2760),
    [anon_sym___bridge] = ACTIONS(2760),
    [anon_sym___bridge_retained] = ACTIONS(2760),
    [anon_sym___bridge_transfer] = ACTIONS(2760),
    [anon_sym___complex] = ACTIONS(2760),
    [anon_sym___const] = ACTIONS(2760),
    [anon_sym___imag] = ACTIONS(2760),
    [anon_sym___kindof] = ACTIONS(2760),
    [anon_sym___nonnull] = ACTIONS(2760),
    [anon_sym___nullable] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2760),
    [anon_sym___real] = ACTIONS(2760),
    [anon_sym___strong] = ACTIONS(2760),
    [anon_sym___unsafe_unretained] = ACTIONS(2760),
    [anon_sym___unused] = ACTIONS(2760),
    [anon_sym___weak] = ACTIONS(2760),
    [anon_sym_alignas] = ACTIONS(2760),
    [anon_sym__Alignas] = ACTIONS(2760),
    [sym_primitive_type] = ACTIONS(2760),
    [anon_sym_enum] = ACTIONS(2760),
    [anon_sym_struct] = ACTIONS(2760),
    [anon_sym_union] = ACTIONS(2760),
    [anon_sym_if] = ACTIONS(2760),
    [anon_sym_switch] = ACTIONS(2760),
    [anon_sym_case] = ACTIONS(2760),
    [anon_sym_default] = ACTIONS(2760),
    [anon_sym_while] = ACTIONS(2760),
    [anon_sym_do] = ACTIONS(2760),
    [anon_sym_for] = ACTIONS(2760),
    [anon_sym_in] = ACTIONS(2760),
    [anon_sym_return] = ACTIONS(2760),
    [anon_sym_break] = ACTIONS(2760),
    [anon_sym_continue] = ACTIONS(2760),
    [anon_sym_goto] = ACTIONS(2760),
    [anon_sym_DASH_DASH] = ACTIONS(2762),
    [anon_sym_PLUS_PLUS] = ACTIONS(2762),
    [anon_sym_sizeof] = ACTIONS(2760),
    [anon_sym___alignof__] = ACTIONS(2760),
    [anon_sym___alignof] = ACTIONS(2760),
    [anon_sym__alignof] = ACTIONS(2760),
    [anon_sym_alignof] = ACTIONS(2760),
    [anon_sym__Alignof] = ACTIONS(2760),
    [anon_sym_offsetof] = ACTIONS(2760),
    [anon_sym__Generic] = ACTIONS(2760),
    [anon_sym_asm] = ACTIONS(2760),
    [anon_sym___asm__] = ACTIONS(2760),
    [anon_sym___asm] = ACTIONS(2760),
    [sym_number_literal] = ACTIONS(2762),
    [anon_sym_L_SQUOTE] = ACTIONS(2762),
    [anon_sym_u_SQUOTE] = ACTIONS(2762),
    [anon_sym_U_SQUOTE] = ACTIONS(2762),
    [anon_sym_u8_SQUOTE] = ACTIONS(2762),
    [anon_sym_SQUOTE] = ACTIONS(2762),
    [anon_sym_AT] = ACTIONS(2760),
    [anon_sym_DQUOTE] = ACTIONS(2762),
    [anon_sym_L_DQUOTE] = ACTIONS(2762),
    [anon_sym_u_DQUOTE] = ACTIONS(2762),
    [anon_sym_U_DQUOTE] = ACTIONS(2762),
    [anon_sym_u8_DQUOTE] = ACTIONS(2762),
    [sym_true] = ACTIONS(2760),
    [sym_false] = ACTIONS(2760),
    [anon_sym_NULL] = ACTIONS(2760),
    [anon_sym_nullptr] = ACTIONS(2760),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2760),
    [anon_sym___typeof] = ACTIONS(2760),
    [anon_sym_typeof] = ACTIONS(2760),
    [anon_sym_ATimport] = ACTIONS(2762),
    [aux_sym_preproc_undef_token1] = ACTIONS(2760),
    [anon_sym_POUND] = ACTIONS(2760),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2760),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2760),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2760),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2760),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2760),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2760),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2760),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_API_AVAILABLE] = ACTIONS(2760),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_API_DEPRECATED] = ACTIONS(2760),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2760),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym___deprecated_msg] = ACTIONS(2760),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2760),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2760),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2760),
    [anon_sym_ATprotocol] = ACTIONS(2762),
    [anon_sym_ATinterface] = ACTIONS(2762),
    [anon_sym_ATimplementation] = ACTIONS(2762),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2762),
    [anon_sym_ATtry] = ACTIONS(2762),
    [anon_sym___try] = ACTIONS(2760),
    [anon_sym_ATthrow] = ACTIONS(2762),
    [anon_sym_ATselector] = ACTIONS(2762),
    [anon_sym_ATavailable] = ACTIONS(2762),
    [anon_sym___builtin_available] = ACTIONS(2760),
    [anon_sym_va_arg] = ACTIONS(2760),
    [anon_sym_ATencode] = ACTIONS(2762),
    [anon_sym_ATsynchronized] = ACTIONS(2762),
    [anon_sym_BOOL] = ACTIONS(2760),
    [anon_sym_IMP] = ACTIONS(2760),
    [anon_sym_SEL] = ACTIONS(2760),
    [anon_sym_Class] = ACTIONS(2760),
    [anon_sym_id] = ACTIONS(2760),
  },
  [STATE(836)] = {
    [sym_identifier] = ACTIONS(2860),
    [aux_sym_preproc_include_token1] = ACTIONS(2860),
    [aux_sym_preproc_include_token2] = ACTIONS(2860),
    [aux_sym_preproc_def_token1] = ACTIONS(2860),
    [aux_sym_preproc_if_token1] = ACTIONS(2860),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2860),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2860),
    [sym_preproc_directive] = ACTIONS(2860),
    [anon_sym_LPAREN2] = ACTIONS(2862),
    [anon_sym_BANG] = ACTIONS(2862),
    [anon_sym_TILDE] = ACTIONS(2862),
    [anon_sym_DASH] = ACTIONS(2860),
    [anon_sym_PLUS] = ACTIONS(2860),
    [anon_sym_STAR] = ACTIONS(2862),
    [anon_sym_CARET] = ACTIONS(2862),
    [anon_sym_AMP] = ACTIONS(2862),
    [anon_sym_SEMI] = ACTIONS(2862),
    [anon_sym___extension__] = ACTIONS(2860),
    [anon_sym_typedef] = ACTIONS(2860),
    [anon_sym_extern] = ACTIONS(2860),
    [anon_sym___attribute__] = ACTIONS(2860),
    [anon_sym___attribute] = ACTIONS(2860),
    [anon_sym_noreturn] = ACTIONS(2860),
    [anon_sym_LBRACK] = ACTIONS(2862),
    [anon_sym___declspec] = ACTIONS(2860),
    [anon_sym___cdecl] = ACTIONS(2860),
    [anon_sym___clrcall] = ACTIONS(2860),
    [anon_sym___stdcall] = ACTIONS(2860),
    [anon_sym___fastcall] = ACTIONS(2860),
    [anon_sym___thiscall] = ACTIONS(2860),
    [anon_sym___vectorcall] = ACTIONS(2860),
    [anon_sym_LBRACE] = ACTIONS(2862),
    [anon_sym_RBRACE] = ACTIONS(2862),
    [anon_sym_signed] = ACTIONS(2860),
    [anon_sym_unsigned] = ACTIONS(2860),
    [anon_sym_long] = ACTIONS(2860),
    [anon_sym_short] = ACTIONS(2860),
    [anon_sym_static] = ACTIONS(2860),
    [anon_sym_ATautoreleasepool] = ACTIONS(2862),
    [anon_sym_auto] = ACTIONS(2860),
    [anon_sym_register] = ACTIONS(2860),
    [anon_sym_inline] = ACTIONS(2860),
    [anon_sym___inline] = ACTIONS(2860),
    [anon_sym___inline__] = ACTIONS(2860),
    [anon_sym___forceinline] = ACTIONS(2860),
    [anon_sym_thread_local] = ACTIONS(2860),
    [anon_sym___thread] = ACTIONS(2860),
    [anon_sym_CG_EXTERN] = ACTIONS(2860),
    [anon_sym_CG_INLINE] = ACTIONS(2860),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2860),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2860),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2860),
    [anon_sym_IBOutlet] = ACTIONS(2860),
    [anon_sym_IBInspectable] = ACTIONS(2860),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2860),
    [anon_sym_NS_INLINE] = ACTIONS(2860),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2860),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2860),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2860),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2860),
    [anon_sym_const] = ACTIONS(2860),
    [anon_sym_constexpr] = ACTIONS(2860),
    [anon_sym_volatile] = ACTIONS(2860),
    [anon_sym_restrict] = ACTIONS(2860),
    [anon_sym___restrict__] = ACTIONS(2860),
    [anon_sym__Atomic] = ACTIONS(2860),
    [anon_sym__Noreturn] = ACTIONS(2860),
    [anon_sym__Nonnull] = ACTIONS(2860),
    [anon_sym_nullable] = ACTIONS(2860),
    [anon_sym__Complex] = ACTIONS(2860),
    [anon_sym__Nullable] = ACTIONS(2860),
    [anon_sym__Nullable_result] = ACTIONS(2860),
    [anon_sym__Null_unspecified] = ACTIONS(2860),
    [anon_sym___autoreleasing] = ACTIONS(2860),
    [anon_sym___block] = ACTIONS(2860),
    [anon_sym___bridge] = ACTIONS(2860),
    [anon_sym___bridge_retained] = ACTIONS(2860),
    [anon_sym___bridge_transfer] = ACTIONS(2860),
    [anon_sym___complex] = ACTIONS(2860),
    [anon_sym___const] = ACTIONS(2860),
    [anon_sym___imag] = ACTIONS(2860),
    [anon_sym___kindof] = ACTIONS(2860),
    [anon_sym___nonnull] = ACTIONS(2860),
    [anon_sym___nullable] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2860),
    [anon_sym___real] = ACTIONS(2860),
    [anon_sym___strong] = ACTIONS(2860),
    [anon_sym___unsafe_unretained] = ACTIONS(2860),
    [anon_sym___unused] = ACTIONS(2860),
    [anon_sym___weak] = ACTIONS(2860),
    [anon_sym_alignas] = ACTIONS(2860),
    [anon_sym__Alignas] = ACTIONS(2860),
    [sym_primitive_type] = ACTIONS(2860),
    [anon_sym_enum] = ACTIONS(2860),
    [anon_sym_struct] = ACTIONS(2860),
    [anon_sym_union] = ACTIONS(2860),
    [anon_sym_if] = ACTIONS(2860),
    [anon_sym_switch] = ACTIONS(2860),
    [anon_sym_case] = ACTIONS(2860),
    [anon_sym_default] = ACTIONS(2860),
    [anon_sym_while] = ACTIONS(2860),
    [anon_sym_do] = ACTIONS(2860),
    [anon_sym_for] = ACTIONS(2860),
    [anon_sym_in] = ACTIONS(2860),
    [anon_sym_return] = ACTIONS(2860),
    [anon_sym_break] = ACTIONS(2860),
    [anon_sym_continue] = ACTIONS(2860),
    [anon_sym_goto] = ACTIONS(2860),
    [anon_sym_DASH_DASH] = ACTIONS(2862),
    [anon_sym_PLUS_PLUS] = ACTIONS(2862),
    [anon_sym_sizeof] = ACTIONS(2860),
    [anon_sym___alignof__] = ACTIONS(2860),
    [anon_sym___alignof] = ACTIONS(2860),
    [anon_sym__alignof] = ACTIONS(2860),
    [anon_sym_alignof] = ACTIONS(2860),
    [anon_sym__Alignof] = ACTIONS(2860),
    [anon_sym_offsetof] = ACTIONS(2860),
    [anon_sym__Generic] = ACTIONS(2860),
    [anon_sym_asm] = ACTIONS(2860),
    [anon_sym___asm__] = ACTIONS(2860),
    [anon_sym___asm] = ACTIONS(2860),
    [sym_number_literal] = ACTIONS(2862),
    [anon_sym_L_SQUOTE] = ACTIONS(2862),
    [anon_sym_u_SQUOTE] = ACTIONS(2862),
    [anon_sym_U_SQUOTE] = ACTIONS(2862),
    [anon_sym_u8_SQUOTE] = ACTIONS(2862),
    [anon_sym_SQUOTE] = ACTIONS(2862),
    [anon_sym_AT] = ACTIONS(2860),
    [anon_sym_DQUOTE] = ACTIONS(2862),
    [anon_sym_L_DQUOTE] = ACTIONS(2862),
    [anon_sym_u_DQUOTE] = ACTIONS(2862),
    [anon_sym_U_DQUOTE] = ACTIONS(2862),
    [anon_sym_u8_DQUOTE] = ACTIONS(2862),
    [sym_true] = ACTIONS(2860),
    [sym_false] = ACTIONS(2860),
    [anon_sym_NULL] = ACTIONS(2860),
    [anon_sym_nullptr] = ACTIONS(2860),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2860),
    [anon_sym___typeof] = ACTIONS(2860),
    [anon_sym_typeof] = ACTIONS(2860),
    [anon_sym_ATimport] = ACTIONS(2862),
    [aux_sym_preproc_undef_token1] = ACTIONS(2860),
    [anon_sym_POUND] = ACTIONS(2860),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2860),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2860),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2860),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2860),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2860),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2860),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2860),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_API_AVAILABLE] = ACTIONS(2860),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_API_DEPRECATED] = ACTIONS(2860),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2860),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym___deprecated_msg] = ACTIONS(2860),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2860),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2860),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2860),
    [anon_sym_ATprotocol] = ACTIONS(2862),
    [anon_sym_ATinterface] = ACTIONS(2862),
    [anon_sym_ATimplementation] = ACTIONS(2862),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2862),
    [anon_sym_ATtry] = ACTIONS(2862),
    [anon_sym___try] = ACTIONS(2860),
    [anon_sym_ATthrow] = ACTIONS(2862),
    [anon_sym_ATselector] = ACTIONS(2862),
    [anon_sym_ATavailable] = ACTIONS(2862),
    [anon_sym___builtin_available] = ACTIONS(2860),
    [anon_sym_va_arg] = ACTIONS(2860),
    [anon_sym_ATencode] = ACTIONS(2862),
    [anon_sym_ATsynchronized] = ACTIONS(2862),
    [anon_sym_BOOL] = ACTIONS(2860),
    [anon_sym_IMP] = ACTIONS(2860),
    [anon_sym_SEL] = ACTIONS(2860),
    [anon_sym_Class] = ACTIONS(2860),
    [anon_sym_id] = ACTIONS(2860),
  },
  [STATE(837)] = {
    [sym_identifier] = ACTIONS(2764),
    [aux_sym_preproc_include_token1] = ACTIONS(2764),
    [aux_sym_preproc_include_token2] = ACTIONS(2764),
    [aux_sym_preproc_def_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2764),
    [sym_preproc_directive] = ACTIONS(2764),
    [anon_sym_LPAREN2] = ACTIONS(2766),
    [anon_sym_BANG] = ACTIONS(2766),
    [anon_sym_TILDE] = ACTIONS(2766),
    [anon_sym_DASH] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(2764),
    [anon_sym_STAR] = ACTIONS(2766),
    [anon_sym_CARET] = ACTIONS(2766),
    [anon_sym_AMP] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2766),
    [anon_sym___extension__] = ACTIONS(2764),
    [anon_sym_typedef] = ACTIONS(2764),
    [anon_sym_extern] = ACTIONS(2764),
    [anon_sym___attribute__] = ACTIONS(2764),
    [anon_sym___attribute] = ACTIONS(2764),
    [anon_sym_noreturn] = ACTIONS(2764),
    [anon_sym_LBRACK] = ACTIONS(2766),
    [anon_sym___declspec] = ACTIONS(2764),
    [anon_sym___cdecl] = ACTIONS(2764),
    [anon_sym___clrcall] = ACTIONS(2764),
    [anon_sym___stdcall] = ACTIONS(2764),
    [anon_sym___fastcall] = ACTIONS(2764),
    [anon_sym___thiscall] = ACTIONS(2764),
    [anon_sym___vectorcall] = ACTIONS(2764),
    [anon_sym_LBRACE] = ACTIONS(2766),
    [anon_sym_RBRACE] = ACTIONS(2766),
    [anon_sym_signed] = ACTIONS(2764),
    [anon_sym_unsigned] = ACTIONS(2764),
    [anon_sym_long] = ACTIONS(2764),
    [anon_sym_short] = ACTIONS(2764),
    [anon_sym_static] = ACTIONS(2764),
    [anon_sym_ATautoreleasepool] = ACTIONS(2766),
    [anon_sym_auto] = ACTIONS(2764),
    [anon_sym_register] = ACTIONS(2764),
    [anon_sym_inline] = ACTIONS(2764),
    [anon_sym___inline] = ACTIONS(2764),
    [anon_sym___inline__] = ACTIONS(2764),
    [anon_sym___forceinline] = ACTIONS(2764),
    [anon_sym_thread_local] = ACTIONS(2764),
    [anon_sym___thread] = ACTIONS(2764),
    [anon_sym_CG_EXTERN] = ACTIONS(2764),
    [anon_sym_CG_INLINE] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2764),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2764),
    [anon_sym_IBOutlet] = ACTIONS(2764),
    [anon_sym_IBInspectable] = ACTIONS(2764),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2764),
    [anon_sym_NS_INLINE] = ACTIONS(2764),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2764),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2764),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2764),
    [anon_sym_const] = ACTIONS(2764),
    [anon_sym_constexpr] = ACTIONS(2764),
    [anon_sym_volatile] = ACTIONS(2764),
    [anon_sym_restrict] = ACTIONS(2764),
    [anon_sym___restrict__] = ACTIONS(2764),
    [anon_sym__Atomic] = ACTIONS(2764),
    [anon_sym__Noreturn] = ACTIONS(2764),
    [anon_sym__Nonnull] = ACTIONS(2764),
    [anon_sym_nullable] = ACTIONS(2764),
    [anon_sym__Complex] = ACTIONS(2764),
    [anon_sym__Nullable] = ACTIONS(2764),
    [anon_sym__Nullable_result] = ACTIONS(2764),
    [anon_sym__Null_unspecified] = ACTIONS(2764),
    [anon_sym___autoreleasing] = ACTIONS(2764),
    [anon_sym___block] = ACTIONS(2764),
    [anon_sym___bridge] = ACTIONS(2764),
    [anon_sym___bridge_retained] = ACTIONS(2764),
    [anon_sym___bridge_transfer] = ACTIONS(2764),
    [anon_sym___complex] = ACTIONS(2764),
    [anon_sym___const] = ACTIONS(2764),
    [anon_sym___imag] = ACTIONS(2764),
    [anon_sym___kindof] = ACTIONS(2764),
    [anon_sym___nonnull] = ACTIONS(2764),
    [anon_sym___nullable] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2764),
    [anon_sym___real] = ACTIONS(2764),
    [anon_sym___strong] = ACTIONS(2764),
    [anon_sym___unsafe_unretained] = ACTIONS(2764),
    [anon_sym___unused] = ACTIONS(2764),
    [anon_sym___weak] = ACTIONS(2764),
    [anon_sym_alignas] = ACTIONS(2764),
    [anon_sym__Alignas] = ACTIONS(2764),
    [sym_primitive_type] = ACTIONS(2764),
    [anon_sym_enum] = ACTIONS(2764),
    [anon_sym_struct] = ACTIONS(2764),
    [anon_sym_union] = ACTIONS(2764),
    [anon_sym_if] = ACTIONS(2764),
    [anon_sym_switch] = ACTIONS(2764),
    [anon_sym_case] = ACTIONS(2764),
    [anon_sym_default] = ACTIONS(2764),
    [anon_sym_while] = ACTIONS(2764),
    [anon_sym_do] = ACTIONS(2764),
    [anon_sym_for] = ACTIONS(2764),
    [anon_sym_in] = ACTIONS(2764),
    [anon_sym_return] = ACTIONS(2764),
    [anon_sym_break] = ACTIONS(2764),
    [anon_sym_continue] = ACTIONS(2764),
    [anon_sym_goto] = ACTIONS(2764),
    [anon_sym_DASH_DASH] = ACTIONS(2766),
    [anon_sym_PLUS_PLUS] = ACTIONS(2766),
    [anon_sym_sizeof] = ACTIONS(2764),
    [anon_sym___alignof__] = ACTIONS(2764),
    [anon_sym___alignof] = ACTIONS(2764),
    [anon_sym__alignof] = ACTIONS(2764),
    [anon_sym_alignof] = ACTIONS(2764),
    [anon_sym__Alignof] = ACTIONS(2764),
    [anon_sym_offsetof] = ACTIONS(2764),
    [anon_sym__Generic] = ACTIONS(2764),
    [anon_sym_asm] = ACTIONS(2764),
    [anon_sym___asm__] = ACTIONS(2764),
    [anon_sym___asm] = ACTIONS(2764),
    [sym_number_literal] = ACTIONS(2766),
    [anon_sym_L_SQUOTE] = ACTIONS(2766),
    [anon_sym_u_SQUOTE] = ACTIONS(2766),
    [anon_sym_U_SQUOTE] = ACTIONS(2766),
    [anon_sym_u8_SQUOTE] = ACTIONS(2766),
    [anon_sym_SQUOTE] = ACTIONS(2766),
    [anon_sym_AT] = ACTIONS(2764),
    [anon_sym_DQUOTE] = ACTIONS(2766),
    [anon_sym_L_DQUOTE] = ACTIONS(2766),
    [anon_sym_u_DQUOTE] = ACTIONS(2766),
    [anon_sym_U_DQUOTE] = ACTIONS(2766),
    [anon_sym_u8_DQUOTE] = ACTIONS(2766),
    [sym_true] = ACTIONS(2764),
    [sym_false] = ACTIONS(2764),
    [anon_sym_NULL] = ACTIONS(2764),
    [anon_sym_nullptr] = ACTIONS(2764),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2764),
    [anon_sym___typeof] = ACTIONS(2764),
    [anon_sym_typeof] = ACTIONS(2764),
    [anon_sym_ATimport] = ACTIONS(2766),
    [aux_sym_preproc_undef_token1] = ACTIONS(2764),
    [anon_sym_POUND] = ACTIONS(2764),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2764),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2764),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2764),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_API_AVAILABLE] = ACTIONS(2764),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_API_DEPRECATED] = ACTIONS(2764),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym___deprecated_msg] = ACTIONS(2764),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2764),
    [anon_sym_ATprotocol] = ACTIONS(2766),
    [anon_sym_ATinterface] = ACTIONS(2766),
    [anon_sym_ATimplementation] = ACTIONS(2766),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2766),
    [anon_sym_ATtry] = ACTIONS(2766),
    [anon_sym___try] = ACTIONS(2764),
    [anon_sym_ATthrow] = ACTIONS(2766),
    [anon_sym_ATselector] = ACTIONS(2766),
    [anon_sym_ATavailable] = ACTIONS(2766),
    [anon_sym___builtin_available] = ACTIONS(2764),
    [anon_sym_va_arg] = ACTIONS(2764),
    [anon_sym_ATencode] = ACTIONS(2766),
    [anon_sym_ATsynchronized] = ACTIONS(2766),
    [anon_sym_BOOL] = ACTIONS(2764),
    [anon_sym_IMP] = ACTIONS(2764),
    [anon_sym_SEL] = ACTIONS(2764),
    [anon_sym_Class] = ACTIONS(2764),
    [anon_sym_id] = ACTIONS(2764),
  },
  [STATE(838)] = {
    [sym_identifier] = ACTIONS(2662),
    [aux_sym_preproc_include_token1] = ACTIONS(2662),
    [aux_sym_preproc_include_token2] = ACTIONS(2662),
    [aux_sym_preproc_def_token1] = ACTIONS(2662),
    [aux_sym_preproc_if_token1] = ACTIONS(2662),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2662),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2662),
    [sym_preproc_directive] = ACTIONS(2662),
    [anon_sym_LPAREN2] = ACTIONS(2665),
    [anon_sym_BANG] = ACTIONS(2665),
    [anon_sym_TILDE] = ACTIONS(2665),
    [anon_sym_DASH] = ACTIONS(2662),
    [anon_sym_PLUS] = ACTIONS(2662),
    [anon_sym_STAR] = ACTIONS(2665),
    [anon_sym_CARET] = ACTIONS(2665),
    [anon_sym_AMP] = ACTIONS(2665),
    [anon_sym_SEMI] = ACTIONS(2665),
    [anon_sym___extension__] = ACTIONS(2662),
    [anon_sym_typedef] = ACTIONS(2662),
    [anon_sym_extern] = ACTIONS(2662),
    [anon_sym___attribute__] = ACTIONS(2662),
    [anon_sym___attribute] = ACTIONS(2662),
    [anon_sym_noreturn] = ACTIONS(2662),
    [anon_sym_LBRACK] = ACTIONS(2665),
    [anon_sym___declspec] = ACTIONS(2662),
    [anon_sym___cdecl] = ACTIONS(2662),
    [anon_sym___clrcall] = ACTIONS(2662),
    [anon_sym___stdcall] = ACTIONS(2662),
    [anon_sym___fastcall] = ACTIONS(2662),
    [anon_sym___thiscall] = ACTIONS(2662),
    [anon_sym___vectorcall] = ACTIONS(2662),
    [anon_sym_LBRACE] = ACTIONS(2665),
    [anon_sym_RBRACE] = ACTIONS(2665),
    [anon_sym_signed] = ACTIONS(2662),
    [anon_sym_unsigned] = ACTIONS(2662),
    [anon_sym_long] = ACTIONS(2662),
    [anon_sym_short] = ACTIONS(2662),
    [anon_sym_static] = ACTIONS(2662),
    [anon_sym_ATautoreleasepool] = ACTIONS(2665),
    [anon_sym_auto] = ACTIONS(2662),
    [anon_sym_register] = ACTIONS(2662),
    [anon_sym_inline] = ACTIONS(2662),
    [anon_sym___inline] = ACTIONS(2662),
    [anon_sym___inline__] = ACTIONS(2662),
    [anon_sym___forceinline] = ACTIONS(2662),
    [anon_sym_thread_local] = ACTIONS(2662),
    [anon_sym___thread] = ACTIONS(2662),
    [anon_sym_CG_EXTERN] = ACTIONS(2662),
    [anon_sym_CG_INLINE] = ACTIONS(2662),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2662),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2662),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2662),
    [anon_sym_IBOutlet] = ACTIONS(2662),
    [anon_sym_IBInspectable] = ACTIONS(2662),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2662),
    [anon_sym_NS_INLINE] = ACTIONS(2662),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2662),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2662),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2662),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2662),
    [anon_sym_const] = ACTIONS(2662),
    [anon_sym_constexpr] = ACTIONS(2662),
    [anon_sym_volatile] = ACTIONS(2662),
    [anon_sym_restrict] = ACTIONS(2662),
    [anon_sym___restrict__] = ACTIONS(2662),
    [anon_sym__Atomic] = ACTIONS(2662),
    [anon_sym__Noreturn] = ACTIONS(2662),
    [anon_sym__Nonnull] = ACTIONS(2662),
    [anon_sym_nullable] = ACTIONS(2662),
    [anon_sym__Complex] = ACTIONS(2662),
    [anon_sym__Nullable] = ACTIONS(2662),
    [anon_sym__Nullable_result] = ACTIONS(2662),
    [anon_sym__Null_unspecified] = ACTIONS(2662),
    [anon_sym___autoreleasing] = ACTIONS(2662),
    [anon_sym___block] = ACTIONS(2662),
    [anon_sym___bridge] = ACTIONS(2662),
    [anon_sym___bridge_retained] = ACTIONS(2662),
    [anon_sym___bridge_transfer] = ACTIONS(2662),
    [anon_sym___complex] = ACTIONS(2662),
    [anon_sym___const] = ACTIONS(2662),
    [anon_sym___imag] = ACTIONS(2662),
    [anon_sym___kindof] = ACTIONS(2662),
    [anon_sym___nonnull] = ACTIONS(2662),
    [anon_sym___nullable] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2662),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2662),
    [anon_sym___real] = ACTIONS(2662),
    [anon_sym___strong] = ACTIONS(2662),
    [anon_sym___unsafe_unretained] = ACTIONS(2662),
    [anon_sym___unused] = ACTIONS(2662),
    [anon_sym___weak] = ACTIONS(2662),
    [anon_sym_alignas] = ACTIONS(2662),
    [anon_sym__Alignas] = ACTIONS(2662),
    [sym_primitive_type] = ACTIONS(2662),
    [anon_sym_enum] = ACTIONS(2662),
    [anon_sym_struct] = ACTIONS(2662),
    [anon_sym_union] = ACTIONS(2662),
    [anon_sym_if] = ACTIONS(2662),
    [anon_sym_switch] = ACTIONS(2662),
    [anon_sym_case] = ACTIONS(2662),
    [anon_sym_default] = ACTIONS(2662),
    [anon_sym_while] = ACTIONS(2662),
    [anon_sym_do] = ACTIONS(2662),
    [anon_sym_for] = ACTIONS(2662),
    [anon_sym_in] = ACTIONS(2662),
    [anon_sym_return] = ACTIONS(2662),
    [anon_sym_break] = ACTIONS(2662),
    [anon_sym_continue] = ACTIONS(2662),
    [anon_sym_goto] = ACTIONS(2662),
    [anon_sym_DASH_DASH] = ACTIONS(2665),
    [anon_sym_PLUS_PLUS] = ACTIONS(2665),
    [anon_sym_sizeof] = ACTIONS(2662),
    [anon_sym___alignof__] = ACTIONS(2662),
    [anon_sym___alignof] = ACTIONS(2662),
    [anon_sym__alignof] = ACTIONS(2662),
    [anon_sym_alignof] = ACTIONS(2662),
    [anon_sym__Alignof] = ACTIONS(2662),
    [anon_sym_offsetof] = ACTIONS(2662),
    [anon_sym__Generic] = ACTIONS(2662),
    [anon_sym_asm] = ACTIONS(2662),
    [anon_sym___asm__] = ACTIONS(2662),
    [anon_sym___asm] = ACTIONS(2662),
    [sym_number_literal] = ACTIONS(2665),
    [anon_sym_L_SQUOTE] = ACTIONS(2665),
    [anon_sym_u_SQUOTE] = ACTIONS(2665),
    [anon_sym_U_SQUOTE] = ACTIONS(2665),
    [anon_sym_u8_SQUOTE] = ACTIONS(2665),
    [anon_sym_SQUOTE] = ACTIONS(2665),
    [anon_sym_AT] = ACTIONS(2662),
    [anon_sym_DQUOTE] = ACTIONS(2665),
    [anon_sym_L_DQUOTE] = ACTIONS(2665),
    [anon_sym_u_DQUOTE] = ACTIONS(2665),
    [anon_sym_U_DQUOTE] = ACTIONS(2665),
    [anon_sym_u8_DQUOTE] = ACTIONS(2665),
    [sym_true] = ACTIONS(2662),
    [sym_false] = ACTIONS(2662),
    [anon_sym_NULL] = ACTIONS(2662),
    [anon_sym_nullptr] = ACTIONS(2662),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2662),
    [anon_sym___typeof] = ACTIONS(2662),
    [anon_sym_typeof] = ACTIONS(2662),
    [anon_sym_ATimport] = ACTIONS(2665),
    [aux_sym_preproc_undef_token1] = ACTIONS(2662),
    [anon_sym_POUND] = ACTIONS(2662),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2662),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2662),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2662),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2662),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2662),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2662),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2662),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_API_AVAILABLE] = ACTIONS(2662),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_API_DEPRECATED] = ACTIONS(2662),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2662),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2662),
    [anon_sym___deprecated_msg] = ACTIONS(2662),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2662),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2662),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2662),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2662),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2662),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2662),
    [anon_sym_ATprotocol] = ACTIONS(2665),
    [anon_sym_ATinterface] = ACTIONS(2665),
    [anon_sym_ATimplementation] = ACTIONS(2665),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2665),
    [anon_sym_ATtry] = ACTIONS(2665),
    [anon_sym___try] = ACTIONS(2662),
    [anon_sym_ATthrow] = ACTIONS(2665),
    [anon_sym_ATselector] = ACTIONS(2665),
    [anon_sym_ATavailable] = ACTIONS(2665),
    [anon_sym___builtin_available] = ACTIONS(2662),
    [anon_sym_va_arg] = ACTIONS(2662),
    [anon_sym_ATencode] = ACTIONS(2665),
    [anon_sym_ATsynchronized] = ACTIONS(2665),
    [anon_sym_BOOL] = ACTIONS(2662),
    [anon_sym_IMP] = ACTIONS(2662),
    [anon_sym_SEL] = ACTIONS(2662),
    [anon_sym_Class] = ACTIONS(2662),
    [anon_sym_id] = ACTIONS(2662),
  },
  [STATE(839)] = {
    [sym_identifier] = ACTIONS(2708),
    [aux_sym_preproc_include_token1] = ACTIONS(2708),
    [aux_sym_preproc_include_token2] = ACTIONS(2708),
    [aux_sym_preproc_def_token1] = ACTIONS(2708),
    [aux_sym_preproc_if_token1] = ACTIONS(2708),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2708),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2708),
    [sym_preproc_directive] = ACTIONS(2708),
    [anon_sym_LPAREN2] = ACTIONS(2710),
    [anon_sym_BANG] = ACTIONS(2710),
    [anon_sym_TILDE] = ACTIONS(2710),
    [anon_sym_DASH] = ACTIONS(2708),
    [anon_sym_PLUS] = ACTIONS(2708),
    [anon_sym_STAR] = ACTIONS(2710),
    [anon_sym_CARET] = ACTIONS(2710),
    [anon_sym_AMP] = ACTIONS(2710),
    [anon_sym_SEMI] = ACTIONS(2710),
    [anon_sym___extension__] = ACTIONS(2708),
    [anon_sym_typedef] = ACTIONS(2708),
    [anon_sym_extern] = ACTIONS(2708),
    [anon_sym___attribute__] = ACTIONS(2708),
    [anon_sym___attribute] = ACTIONS(2708),
    [anon_sym_noreturn] = ACTIONS(2708),
    [anon_sym_LBRACK] = ACTIONS(2710),
    [anon_sym___declspec] = ACTIONS(2708),
    [anon_sym___cdecl] = ACTIONS(2708),
    [anon_sym___clrcall] = ACTIONS(2708),
    [anon_sym___stdcall] = ACTIONS(2708),
    [anon_sym___fastcall] = ACTIONS(2708),
    [anon_sym___thiscall] = ACTIONS(2708),
    [anon_sym___vectorcall] = ACTIONS(2708),
    [anon_sym_LBRACE] = ACTIONS(2710),
    [anon_sym_RBRACE] = ACTIONS(2710),
    [anon_sym_signed] = ACTIONS(2708),
    [anon_sym_unsigned] = ACTIONS(2708),
    [anon_sym_long] = ACTIONS(2708),
    [anon_sym_short] = ACTIONS(2708),
    [anon_sym_static] = ACTIONS(2708),
    [anon_sym_ATautoreleasepool] = ACTIONS(2710),
    [anon_sym_auto] = ACTIONS(2708),
    [anon_sym_register] = ACTIONS(2708),
    [anon_sym_inline] = ACTIONS(2708),
    [anon_sym___inline] = ACTIONS(2708),
    [anon_sym___inline__] = ACTIONS(2708),
    [anon_sym___forceinline] = ACTIONS(2708),
    [anon_sym_thread_local] = ACTIONS(2708),
    [anon_sym___thread] = ACTIONS(2708),
    [anon_sym_CG_EXTERN] = ACTIONS(2708),
    [anon_sym_CG_INLINE] = ACTIONS(2708),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2708),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2708),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2708),
    [anon_sym_IBOutlet] = ACTIONS(2708),
    [anon_sym_IBInspectable] = ACTIONS(2708),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2708),
    [anon_sym_NS_INLINE] = ACTIONS(2708),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2708),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2708),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2708),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2708),
    [anon_sym_const] = ACTIONS(2708),
    [anon_sym_constexpr] = ACTIONS(2708),
    [anon_sym_volatile] = ACTIONS(2708),
    [anon_sym_restrict] = ACTIONS(2708),
    [anon_sym___restrict__] = ACTIONS(2708),
    [anon_sym__Atomic] = ACTIONS(2708),
    [anon_sym__Noreturn] = ACTIONS(2708),
    [anon_sym__Nonnull] = ACTIONS(2708),
    [anon_sym_nullable] = ACTIONS(2708),
    [anon_sym__Complex] = ACTIONS(2708),
    [anon_sym__Nullable] = ACTIONS(2708),
    [anon_sym__Nullable_result] = ACTIONS(2708),
    [anon_sym__Null_unspecified] = ACTIONS(2708),
    [anon_sym___autoreleasing] = ACTIONS(2708),
    [anon_sym___block] = ACTIONS(2708),
    [anon_sym___bridge] = ACTIONS(2708),
    [anon_sym___bridge_retained] = ACTIONS(2708),
    [anon_sym___bridge_transfer] = ACTIONS(2708),
    [anon_sym___complex] = ACTIONS(2708),
    [anon_sym___const] = ACTIONS(2708),
    [anon_sym___imag] = ACTIONS(2708),
    [anon_sym___kindof] = ACTIONS(2708),
    [anon_sym___nonnull] = ACTIONS(2708),
    [anon_sym___nullable] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2708),
    [anon_sym___real] = ACTIONS(2708),
    [anon_sym___strong] = ACTIONS(2708),
    [anon_sym___unsafe_unretained] = ACTIONS(2708),
    [anon_sym___unused] = ACTIONS(2708),
    [anon_sym___weak] = ACTIONS(2708),
    [anon_sym_alignas] = ACTIONS(2708),
    [anon_sym__Alignas] = ACTIONS(2708),
    [sym_primitive_type] = ACTIONS(2708),
    [anon_sym_enum] = ACTIONS(2708),
    [anon_sym_struct] = ACTIONS(2708),
    [anon_sym_union] = ACTIONS(2708),
    [anon_sym_if] = ACTIONS(2708),
    [anon_sym_switch] = ACTIONS(2708),
    [anon_sym_case] = ACTIONS(2708),
    [anon_sym_default] = ACTIONS(2708),
    [anon_sym_while] = ACTIONS(2708),
    [anon_sym_do] = ACTIONS(2708),
    [anon_sym_for] = ACTIONS(2708),
    [anon_sym_in] = ACTIONS(2708),
    [anon_sym_return] = ACTIONS(2708),
    [anon_sym_break] = ACTIONS(2708),
    [anon_sym_continue] = ACTIONS(2708),
    [anon_sym_goto] = ACTIONS(2708),
    [anon_sym_DASH_DASH] = ACTIONS(2710),
    [anon_sym_PLUS_PLUS] = ACTIONS(2710),
    [anon_sym_sizeof] = ACTIONS(2708),
    [anon_sym___alignof__] = ACTIONS(2708),
    [anon_sym___alignof] = ACTIONS(2708),
    [anon_sym__alignof] = ACTIONS(2708),
    [anon_sym_alignof] = ACTIONS(2708),
    [anon_sym__Alignof] = ACTIONS(2708),
    [anon_sym_offsetof] = ACTIONS(2708),
    [anon_sym__Generic] = ACTIONS(2708),
    [anon_sym_asm] = ACTIONS(2708),
    [anon_sym___asm__] = ACTIONS(2708),
    [anon_sym___asm] = ACTIONS(2708),
    [sym_number_literal] = ACTIONS(2710),
    [anon_sym_L_SQUOTE] = ACTIONS(2710),
    [anon_sym_u_SQUOTE] = ACTIONS(2710),
    [anon_sym_U_SQUOTE] = ACTIONS(2710),
    [anon_sym_u8_SQUOTE] = ACTIONS(2710),
    [anon_sym_SQUOTE] = ACTIONS(2710),
    [anon_sym_AT] = ACTIONS(2708),
    [anon_sym_DQUOTE] = ACTIONS(2710),
    [anon_sym_L_DQUOTE] = ACTIONS(2710),
    [anon_sym_u_DQUOTE] = ACTIONS(2710),
    [anon_sym_U_DQUOTE] = ACTIONS(2710),
    [anon_sym_u8_DQUOTE] = ACTIONS(2710),
    [sym_true] = ACTIONS(2708),
    [sym_false] = ACTIONS(2708),
    [anon_sym_NULL] = ACTIONS(2708),
    [anon_sym_nullptr] = ACTIONS(2708),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2708),
    [anon_sym___typeof] = ACTIONS(2708),
    [anon_sym_typeof] = ACTIONS(2708),
    [anon_sym_ATimport] = ACTIONS(2710),
    [aux_sym_preproc_undef_token1] = ACTIONS(2708),
    [anon_sym_POUND] = ACTIONS(2708),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2708),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2708),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2708),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2708),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2708),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2708),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2708),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_API_AVAILABLE] = ACTIONS(2708),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_API_DEPRECATED] = ACTIONS(2708),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2708),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym___deprecated_msg] = ACTIONS(2708),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2708),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2708),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2708),
    [anon_sym_ATprotocol] = ACTIONS(2710),
    [anon_sym_ATinterface] = ACTIONS(2710),
    [anon_sym_ATimplementation] = ACTIONS(2710),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2710),
    [anon_sym_ATtry] = ACTIONS(2710),
    [anon_sym___try] = ACTIONS(2708),
    [anon_sym_ATthrow] = ACTIONS(2710),
    [anon_sym_ATselector] = ACTIONS(2710),
    [anon_sym_ATavailable] = ACTIONS(2710),
    [anon_sym___builtin_available] = ACTIONS(2708),
    [anon_sym_va_arg] = ACTIONS(2708),
    [anon_sym_ATencode] = ACTIONS(2710),
    [anon_sym_ATsynchronized] = ACTIONS(2710),
    [anon_sym_BOOL] = ACTIONS(2708),
    [anon_sym_IMP] = ACTIONS(2708),
    [anon_sym_SEL] = ACTIONS(2708),
    [anon_sym_Class] = ACTIONS(2708),
    [anon_sym_id] = ACTIONS(2708),
  },
  [STATE(840)] = {
    [sym_identifier] = ACTIONS(2764),
    [aux_sym_preproc_include_token1] = ACTIONS(2764),
    [aux_sym_preproc_include_token2] = ACTIONS(2764),
    [aux_sym_preproc_def_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2764),
    [sym_preproc_directive] = ACTIONS(2764),
    [anon_sym_LPAREN2] = ACTIONS(2766),
    [anon_sym_BANG] = ACTIONS(2766),
    [anon_sym_TILDE] = ACTIONS(2766),
    [anon_sym_DASH] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(2764),
    [anon_sym_STAR] = ACTIONS(2766),
    [anon_sym_CARET] = ACTIONS(2766),
    [anon_sym_AMP] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2766),
    [anon_sym___extension__] = ACTIONS(2764),
    [anon_sym_typedef] = ACTIONS(2764),
    [anon_sym_extern] = ACTIONS(2764),
    [anon_sym___attribute__] = ACTIONS(2764),
    [anon_sym___attribute] = ACTIONS(2764),
    [anon_sym_noreturn] = ACTIONS(2764),
    [anon_sym_LBRACK] = ACTIONS(2766),
    [anon_sym___declspec] = ACTIONS(2764),
    [anon_sym___cdecl] = ACTIONS(2764),
    [anon_sym___clrcall] = ACTIONS(2764),
    [anon_sym___stdcall] = ACTIONS(2764),
    [anon_sym___fastcall] = ACTIONS(2764),
    [anon_sym___thiscall] = ACTIONS(2764),
    [anon_sym___vectorcall] = ACTIONS(2764),
    [anon_sym_LBRACE] = ACTIONS(2766),
    [anon_sym_RBRACE] = ACTIONS(2766),
    [anon_sym_signed] = ACTIONS(2764),
    [anon_sym_unsigned] = ACTIONS(2764),
    [anon_sym_long] = ACTIONS(2764),
    [anon_sym_short] = ACTIONS(2764),
    [anon_sym_static] = ACTIONS(2764),
    [anon_sym_ATautoreleasepool] = ACTIONS(2766),
    [anon_sym_auto] = ACTIONS(2764),
    [anon_sym_register] = ACTIONS(2764),
    [anon_sym_inline] = ACTIONS(2764),
    [anon_sym___inline] = ACTIONS(2764),
    [anon_sym___inline__] = ACTIONS(2764),
    [anon_sym___forceinline] = ACTIONS(2764),
    [anon_sym_thread_local] = ACTIONS(2764),
    [anon_sym___thread] = ACTIONS(2764),
    [anon_sym_CG_EXTERN] = ACTIONS(2764),
    [anon_sym_CG_INLINE] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2764),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2764),
    [anon_sym_IBOutlet] = ACTIONS(2764),
    [anon_sym_IBInspectable] = ACTIONS(2764),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2764),
    [anon_sym_NS_INLINE] = ACTIONS(2764),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2764),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2764),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2764),
    [anon_sym_const] = ACTIONS(2764),
    [anon_sym_constexpr] = ACTIONS(2764),
    [anon_sym_volatile] = ACTIONS(2764),
    [anon_sym_restrict] = ACTIONS(2764),
    [anon_sym___restrict__] = ACTIONS(2764),
    [anon_sym__Atomic] = ACTIONS(2764),
    [anon_sym__Noreturn] = ACTIONS(2764),
    [anon_sym__Nonnull] = ACTIONS(2764),
    [anon_sym_nullable] = ACTIONS(2764),
    [anon_sym__Complex] = ACTIONS(2764),
    [anon_sym__Nullable] = ACTIONS(2764),
    [anon_sym__Nullable_result] = ACTIONS(2764),
    [anon_sym__Null_unspecified] = ACTIONS(2764),
    [anon_sym___autoreleasing] = ACTIONS(2764),
    [anon_sym___block] = ACTIONS(2764),
    [anon_sym___bridge] = ACTIONS(2764),
    [anon_sym___bridge_retained] = ACTIONS(2764),
    [anon_sym___bridge_transfer] = ACTIONS(2764),
    [anon_sym___complex] = ACTIONS(2764),
    [anon_sym___const] = ACTIONS(2764),
    [anon_sym___imag] = ACTIONS(2764),
    [anon_sym___kindof] = ACTIONS(2764),
    [anon_sym___nonnull] = ACTIONS(2764),
    [anon_sym___nullable] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2764),
    [anon_sym___real] = ACTIONS(2764),
    [anon_sym___strong] = ACTIONS(2764),
    [anon_sym___unsafe_unretained] = ACTIONS(2764),
    [anon_sym___unused] = ACTIONS(2764),
    [anon_sym___weak] = ACTIONS(2764),
    [anon_sym_alignas] = ACTIONS(2764),
    [anon_sym__Alignas] = ACTIONS(2764),
    [sym_primitive_type] = ACTIONS(2764),
    [anon_sym_enum] = ACTIONS(2764),
    [anon_sym_struct] = ACTIONS(2764),
    [anon_sym_union] = ACTIONS(2764),
    [anon_sym_if] = ACTIONS(2764),
    [anon_sym_switch] = ACTIONS(2764),
    [anon_sym_case] = ACTIONS(2764),
    [anon_sym_default] = ACTIONS(2764),
    [anon_sym_while] = ACTIONS(2764),
    [anon_sym_do] = ACTIONS(2764),
    [anon_sym_for] = ACTIONS(2764),
    [anon_sym_in] = ACTIONS(2764),
    [anon_sym_return] = ACTIONS(2764),
    [anon_sym_break] = ACTIONS(2764),
    [anon_sym_continue] = ACTIONS(2764),
    [anon_sym_goto] = ACTIONS(2764),
    [anon_sym_DASH_DASH] = ACTIONS(2766),
    [anon_sym_PLUS_PLUS] = ACTIONS(2766),
    [anon_sym_sizeof] = ACTIONS(2764),
    [anon_sym___alignof__] = ACTIONS(2764),
    [anon_sym___alignof] = ACTIONS(2764),
    [anon_sym__alignof] = ACTIONS(2764),
    [anon_sym_alignof] = ACTIONS(2764),
    [anon_sym__Alignof] = ACTIONS(2764),
    [anon_sym_offsetof] = ACTIONS(2764),
    [anon_sym__Generic] = ACTIONS(2764),
    [anon_sym_asm] = ACTIONS(2764),
    [anon_sym___asm__] = ACTIONS(2764),
    [anon_sym___asm] = ACTIONS(2764),
    [sym_number_literal] = ACTIONS(2766),
    [anon_sym_L_SQUOTE] = ACTIONS(2766),
    [anon_sym_u_SQUOTE] = ACTIONS(2766),
    [anon_sym_U_SQUOTE] = ACTIONS(2766),
    [anon_sym_u8_SQUOTE] = ACTIONS(2766),
    [anon_sym_SQUOTE] = ACTIONS(2766),
    [anon_sym_AT] = ACTIONS(2764),
    [anon_sym_DQUOTE] = ACTIONS(2766),
    [anon_sym_L_DQUOTE] = ACTIONS(2766),
    [anon_sym_u_DQUOTE] = ACTIONS(2766),
    [anon_sym_U_DQUOTE] = ACTIONS(2766),
    [anon_sym_u8_DQUOTE] = ACTIONS(2766),
    [sym_true] = ACTIONS(2764),
    [sym_false] = ACTIONS(2764),
    [anon_sym_NULL] = ACTIONS(2764),
    [anon_sym_nullptr] = ACTIONS(2764),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2764),
    [anon_sym___typeof] = ACTIONS(2764),
    [anon_sym_typeof] = ACTIONS(2764),
    [anon_sym_ATimport] = ACTIONS(2766),
    [aux_sym_preproc_undef_token1] = ACTIONS(2764),
    [anon_sym_POUND] = ACTIONS(2764),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2764),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2764),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2764),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_API_AVAILABLE] = ACTIONS(2764),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_API_DEPRECATED] = ACTIONS(2764),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym___deprecated_msg] = ACTIONS(2764),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2764),
    [anon_sym_ATprotocol] = ACTIONS(2766),
    [anon_sym_ATinterface] = ACTIONS(2766),
    [anon_sym_ATimplementation] = ACTIONS(2766),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2766),
    [anon_sym_ATtry] = ACTIONS(2766),
    [anon_sym___try] = ACTIONS(2764),
    [anon_sym_ATthrow] = ACTIONS(2766),
    [anon_sym_ATselector] = ACTIONS(2766),
    [anon_sym_ATavailable] = ACTIONS(2766),
    [anon_sym___builtin_available] = ACTIONS(2764),
    [anon_sym_va_arg] = ACTIONS(2764),
    [anon_sym_ATencode] = ACTIONS(2766),
    [anon_sym_ATsynchronized] = ACTIONS(2766),
    [anon_sym_BOOL] = ACTIONS(2764),
    [anon_sym_IMP] = ACTIONS(2764),
    [anon_sym_SEL] = ACTIONS(2764),
    [anon_sym_Class] = ACTIONS(2764),
    [anon_sym_id] = ACTIONS(2764),
  },
  [STATE(841)] = {
    [sym_identifier] = ACTIONS(2600),
    [aux_sym_preproc_include_token1] = ACTIONS(2600),
    [aux_sym_preproc_include_token2] = ACTIONS(2600),
    [aux_sym_preproc_def_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2600),
    [sym_preproc_directive] = ACTIONS(2600),
    [anon_sym_LPAREN2] = ACTIONS(2602),
    [anon_sym_BANG] = ACTIONS(2602),
    [anon_sym_TILDE] = ACTIONS(2602),
    [anon_sym_DASH] = ACTIONS(2600),
    [anon_sym_PLUS] = ACTIONS(2600),
    [anon_sym_STAR] = ACTIONS(2602),
    [anon_sym_CARET] = ACTIONS(2602),
    [anon_sym_AMP] = ACTIONS(2602),
    [anon_sym_SEMI] = ACTIONS(2602),
    [anon_sym___extension__] = ACTIONS(2600),
    [anon_sym_typedef] = ACTIONS(2600),
    [anon_sym_extern] = ACTIONS(2600),
    [anon_sym___attribute__] = ACTIONS(2600),
    [anon_sym___attribute] = ACTIONS(2600),
    [anon_sym_noreturn] = ACTIONS(2600),
    [anon_sym_LBRACK] = ACTIONS(2602),
    [anon_sym___declspec] = ACTIONS(2600),
    [anon_sym___cdecl] = ACTIONS(2600),
    [anon_sym___clrcall] = ACTIONS(2600),
    [anon_sym___stdcall] = ACTIONS(2600),
    [anon_sym___fastcall] = ACTIONS(2600),
    [anon_sym___thiscall] = ACTIONS(2600),
    [anon_sym___vectorcall] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_RBRACE] = ACTIONS(2602),
    [anon_sym_signed] = ACTIONS(2600),
    [anon_sym_unsigned] = ACTIONS(2600),
    [anon_sym_long] = ACTIONS(2600),
    [anon_sym_short] = ACTIONS(2600),
    [anon_sym_static] = ACTIONS(2600),
    [anon_sym_ATautoreleasepool] = ACTIONS(2602),
    [anon_sym_auto] = ACTIONS(2600),
    [anon_sym_register] = ACTIONS(2600),
    [anon_sym_inline] = ACTIONS(2600),
    [anon_sym___inline] = ACTIONS(2600),
    [anon_sym___inline__] = ACTIONS(2600),
    [anon_sym___forceinline] = ACTIONS(2600),
    [anon_sym_thread_local] = ACTIONS(2600),
    [anon_sym___thread] = ACTIONS(2600),
    [anon_sym_CG_EXTERN] = ACTIONS(2600),
    [anon_sym_CG_INLINE] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2600),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2600),
    [anon_sym_IBOutlet] = ACTIONS(2600),
    [anon_sym_IBInspectable] = ACTIONS(2600),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2600),
    [anon_sym_NS_INLINE] = ACTIONS(2600),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2600),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2600),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2600),
    [anon_sym_const] = ACTIONS(2600),
    [anon_sym_constexpr] = ACTIONS(2600),
    [anon_sym_volatile] = ACTIONS(2600),
    [anon_sym_restrict] = ACTIONS(2600),
    [anon_sym___restrict__] = ACTIONS(2600),
    [anon_sym__Atomic] = ACTIONS(2600),
    [anon_sym__Noreturn] = ACTIONS(2600),
    [anon_sym__Nonnull] = ACTIONS(2600),
    [anon_sym_nullable] = ACTIONS(2600),
    [anon_sym__Complex] = ACTIONS(2600),
    [anon_sym__Nullable] = ACTIONS(2600),
    [anon_sym__Nullable_result] = ACTIONS(2600),
    [anon_sym__Null_unspecified] = ACTIONS(2600),
    [anon_sym___autoreleasing] = ACTIONS(2600),
    [anon_sym___block] = ACTIONS(2600),
    [anon_sym___bridge] = ACTIONS(2600),
    [anon_sym___bridge_retained] = ACTIONS(2600),
    [anon_sym___bridge_transfer] = ACTIONS(2600),
    [anon_sym___complex] = ACTIONS(2600),
    [anon_sym___const] = ACTIONS(2600),
    [anon_sym___imag] = ACTIONS(2600),
    [anon_sym___kindof] = ACTIONS(2600),
    [anon_sym___nonnull] = ACTIONS(2600),
    [anon_sym___nullable] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2600),
    [anon_sym___real] = ACTIONS(2600),
    [anon_sym___strong] = ACTIONS(2600),
    [anon_sym___unsafe_unretained] = ACTIONS(2600),
    [anon_sym___unused] = ACTIONS(2600),
    [anon_sym___weak] = ACTIONS(2600),
    [anon_sym_alignas] = ACTIONS(2600),
    [anon_sym__Alignas] = ACTIONS(2600),
    [sym_primitive_type] = ACTIONS(2600),
    [anon_sym_enum] = ACTIONS(2600),
    [anon_sym_struct] = ACTIONS(2600),
    [anon_sym_union] = ACTIONS(2600),
    [anon_sym_if] = ACTIONS(2600),
    [anon_sym_switch] = ACTIONS(2600),
    [anon_sym_case] = ACTIONS(2600),
    [anon_sym_default] = ACTIONS(2600),
    [anon_sym_while] = ACTIONS(2600),
    [anon_sym_do] = ACTIONS(2600),
    [anon_sym_for] = ACTIONS(2600),
    [anon_sym_in] = ACTIONS(2600),
    [anon_sym_return] = ACTIONS(2600),
    [anon_sym_break] = ACTIONS(2600),
    [anon_sym_continue] = ACTIONS(2600),
    [anon_sym_goto] = ACTIONS(2600),
    [anon_sym_DASH_DASH] = ACTIONS(2602),
    [anon_sym_PLUS_PLUS] = ACTIONS(2602),
    [anon_sym_sizeof] = ACTIONS(2600),
    [anon_sym___alignof__] = ACTIONS(2600),
    [anon_sym___alignof] = ACTIONS(2600),
    [anon_sym__alignof] = ACTIONS(2600),
    [anon_sym_alignof] = ACTIONS(2600),
    [anon_sym__Alignof] = ACTIONS(2600),
    [anon_sym_offsetof] = ACTIONS(2600),
    [anon_sym__Generic] = ACTIONS(2600),
    [anon_sym_asm] = ACTIONS(2600),
    [anon_sym___asm__] = ACTIONS(2600),
    [anon_sym___asm] = ACTIONS(2600),
    [sym_number_literal] = ACTIONS(2602),
    [anon_sym_L_SQUOTE] = ACTIONS(2602),
    [anon_sym_u_SQUOTE] = ACTIONS(2602),
    [anon_sym_U_SQUOTE] = ACTIONS(2602),
    [anon_sym_u8_SQUOTE] = ACTIONS(2602),
    [anon_sym_SQUOTE] = ACTIONS(2602),
    [anon_sym_AT] = ACTIONS(2600),
    [anon_sym_DQUOTE] = ACTIONS(2602),
    [anon_sym_L_DQUOTE] = ACTIONS(2602),
    [anon_sym_u_DQUOTE] = ACTIONS(2602),
    [anon_sym_U_DQUOTE] = ACTIONS(2602),
    [anon_sym_u8_DQUOTE] = ACTIONS(2602),
    [sym_true] = ACTIONS(2600),
    [sym_false] = ACTIONS(2600),
    [anon_sym_NULL] = ACTIONS(2600),
    [anon_sym_nullptr] = ACTIONS(2600),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2600),
    [anon_sym___typeof] = ACTIONS(2600),
    [anon_sym_typeof] = ACTIONS(2600),
    [anon_sym_ATimport] = ACTIONS(2602),
    [aux_sym_preproc_undef_token1] = ACTIONS(2600),
    [anon_sym_POUND] = ACTIONS(2600),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2600),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2600),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2600),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_API_AVAILABLE] = ACTIONS(2600),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_API_DEPRECATED] = ACTIONS(2600),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym___deprecated_msg] = ACTIONS(2600),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2600),
    [anon_sym_ATprotocol] = ACTIONS(2602),
    [anon_sym_ATinterface] = ACTIONS(2602),
    [anon_sym_ATimplementation] = ACTIONS(2602),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2602),
    [anon_sym_ATtry] = ACTIONS(2602),
    [anon_sym___try] = ACTIONS(2600),
    [anon_sym_ATthrow] = ACTIONS(2602),
    [anon_sym_ATselector] = ACTIONS(2602),
    [anon_sym_ATavailable] = ACTIONS(2602),
    [anon_sym___builtin_available] = ACTIONS(2600),
    [anon_sym_va_arg] = ACTIONS(2600),
    [anon_sym_ATencode] = ACTIONS(2602),
    [anon_sym_ATsynchronized] = ACTIONS(2602),
    [anon_sym_BOOL] = ACTIONS(2600),
    [anon_sym_IMP] = ACTIONS(2600),
    [anon_sym_SEL] = ACTIONS(2600),
    [anon_sym_Class] = ACTIONS(2600),
    [anon_sym_id] = ACTIONS(2600),
  },
  [STATE(842)] = {
    [sym_identifier] = ACTIONS(2580),
    [aux_sym_preproc_include_token1] = ACTIONS(2580),
    [aux_sym_preproc_include_token2] = ACTIONS(2580),
    [aux_sym_preproc_def_token1] = ACTIONS(2580),
    [aux_sym_preproc_if_token1] = ACTIONS(2580),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2580),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2580),
    [sym_preproc_directive] = ACTIONS(2580),
    [anon_sym_LPAREN2] = ACTIONS(2582),
    [anon_sym_BANG] = ACTIONS(2582),
    [anon_sym_TILDE] = ACTIONS(2582),
    [anon_sym_DASH] = ACTIONS(2580),
    [anon_sym_PLUS] = ACTIONS(2580),
    [anon_sym_STAR] = ACTIONS(2582),
    [anon_sym_CARET] = ACTIONS(2582),
    [anon_sym_AMP] = ACTIONS(2582),
    [anon_sym_SEMI] = ACTIONS(2582),
    [anon_sym___extension__] = ACTIONS(2580),
    [anon_sym_typedef] = ACTIONS(2580),
    [anon_sym_extern] = ACTIONS(2580),
    [anon_sym___attribute__] = ACTIONS(2580),
    [anon_sym___attribute] = ACTIONS(2580),
    [anon_sym_noreturn] = ACTIONS(2580),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym___declspec] = ACTIONS(2580),
    [anon_sym___cdecl] = ACTIONS(2580),
    [anon_sym___clrcall] = ACTIONS(2580),
    [anon_sym___stdcall] = ACTIONS(2580),
    [anon_sym___fastcall] = ACTIONS(2580),
    [anon_sym___thiscall] = ACTIONS(2580),
    [anon_sym___vectorcall] = ACTIONS(2580),
    [anon_sym_LBRACE] = ACTIONS(2582),
    [anon_sym_RBRACE] = ACTIONS(2582),
    [anon_sym_signed] = ACTIONS(2580),
    [anon_sym_unsigned] = ACTIONS(2580),
    [anon_sym_long] = ACTIONS(2580),
    [anon_sym_short] = ACTIONS(2580),
    [anon_sym_static] = ACTIONS(2580),
    [anon_sym_ATautoreleasepool] = ACTIONS(2582),
    [anon_sym_auto] = ACTIONS(2580),
    [anon_sym_register] = ACTIONS(2580),
    [anon_sym_inline] = ACTIONS(2580),
    [anon_sym___inline] = ACTIONS(2580),
    [anon_sym___inline__] = ACTIONS(2580),
    [anon_sym___forceinline] = ACTIONS(2580),
    [anon_sym_thread_local] = ACTIONS(2580),
    [anon_sym___thread] = ACTIONS(2580),
    [anon_sym_CG_EXTERN] = ACTIONS(2580),
    [anon_sym_CG_INLINE] = ACTIONS(2580),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2580),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2580),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2580),
    [anon_sym_IBOutlet] = ACTIONS(2580),
    [anon_sym_IBInspectable] = ACTIONS(2580),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2580),
    [anon_sym_NS_INLINE] = ACTIONS(2580),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2580),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2580),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2580),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2580),
    [anon_sym_const] = ACTIONS(2580),
    [anon_sym_constexpr] = ACTIONS(2580),
    [anon_sym_volatile] = ACTIONS(2580),
    [anon_sym_restrict] = ACTIONS(2580),
    [anon_sym___restrict__] = ACTIONS(2580),
    [anon_sym__Atomic] = ACTIONS(2580),
    [anon_sym__Noreturn] = ACTIONS(2580),
    [anon_sym__Nonnull] = ACTIONS(2580),
    [anon_sym_nullable] = ACTIONS(2580),
    [anon_sym__Complex] = ACTIONS(2580),
    [anon_sym__Nullable] = ACTIONS(2580),
    [anon_sym__Nullable_result] = ACTIONS(2580),
    [anon_sym__Null_unspecified] = ACTIONS(2580),
    [anon_sym___autoreleasing] = ACTIONS(2580),
    [anon_sym___block] = ACTIONS(2580),
    [anon_sym___bridge] = ACTIONS(2580),
    [anon_sym___bridge_retained] = ACTIONS(2580),
    [anon_sym___bridge_transfer] = ACTIONS(2580),
    [anon_sym___complex] = ACTIONS(2580),
    [anon_sym___const] = ACTIONS(2580),
    [anon_sym___imag] = ACTIONS(2580),
    [anon_sym___kindof] = ACTIONS(2580),
    [anon_sym___nonnull] = ACTIONS(2580),
    [anon_sym___nullable] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2580),
    [anon_sym___real] = ACTIONS(2580),
    [anon_sym___strong] = ACTIONS(2580),
    [anon_sym___unsafe_unretained] = ACTIONS(2580),
    [anon_sym___unused] = ACTIONS(2580),
    [anon_sym___weak] = ACTIONS(2580),
    [anon_sym_alignas] = ACTIONS(2580),
    [anon_sym__Alignas] = ACTIONS(2580),
    [sym_primitive_type] = ACTIONS(2580),
    [anon_sym_enum] = ACTIONS(2580),
    [anon_sym_struct] = ACTIONS(2580),
    [anon_sym_union] = ACTIONS(2580),
    [anon_sym_if] = ACTIONS(2580),
    [anon_sym_switch] = ACTIONS(2580),
    [anon_sym_case] = ACTIONS(2580),
    [anon_sym_default] = ACTIONS(2580),
    [anon_sym_while] = ACTIONS(2580),
    [anon_sym_do] = ACTIONS(2580),
    [anon_sym_for] = ACTIONS(2580),
    [anon_sym_in] = ACTIONS(2580),
    [anon_sym_return] = ACTIONS(2580),
    [anon_sym_break] = ACTIONS(2580),
    [anon_sym_continue] = ACTIONS(2580),
    [anon_sym_goto] = ACTIONS(2580),
    [anon_sym_DASH_DASH] = ACTIONS(2582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2582),
    [anon_sym_sizeof] = ACTIONS(2580),
    [anon_sym___alignof__] = ACTIONS(2580),
    [anon_sym___alignof] = ACTIONS(2580),
    [anon_sym__alignof] = ACTIONS(2580),
    [anon_sym_alignof] = ACTIONS(2580),
    [anon_sym__Alignof] = ACTIONS(2580),
    [anon_sym_offsetof] = ACTIONS(2580),
    [anon_sym__Generic] = ACTIONS(2580),
    [anon_sym_asm] = ACTIONS(2580),
    [anon_sym___asm__] = ACTIONS(2580),
    [anon_sym___asm] = ACTIONS(2580),
    [sym_number_literal] = ACTIONS(2582),
    [anon_sym_L_SQUOTE] = ACTIONS(2582),
    [anon_sym_u_SQUOTE] = ACTIONS(2582),
    [anon_sym_U_SQUOTE] = ACTIONS(2582),
    [anon_sym_u8_SQUOTE] = ACTIONS(2582),
    [anon_sym_SQUOTE] = ACTIONS(2582),
    [anon_sym_AT] = ACTIONS(2580),
    [anon_sym_DQUOTE] = ACTIONS(2582),
    [anon_sym_L_DQUOTE] = ACTIONS(2582),
    [anon_sym_u_DQUOTE] = ACTIONS(2582),
    [anon_sym_U_DQUOTE] = ACTIONS(2582),
    [anon_sym_u8_DQUOTE] = ACTIONS(2582),
    [sym_true] = ACTIONS(2580),
    [sym_false] = ACTIONS(2580),
    [anon_sym_NULL] = ACTIONS(2580),
    [anon_sym_nullptr] = ACTIONS(2580),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2580),
    [anon_sym___typeof] = ACTIONS(2580),
    [anon_sym_typeof] = ACTIONS(2580),
    [anon_sym_ATimport] = ACTIONS(2582),
    [aux_sym_preproc_undef_token1] = ACTIONS(2580),
    [anon_sym_POUND] = ACTIONS(2580),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2580),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2580),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2580),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2580),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2580),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2580),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2580),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_API_AVAILABLE] = ACTIONS(2580),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_API_DEPRECATED] = ACTIONS(2580),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2580),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym___deprecated_msg] = ACTIONS(2580),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2580),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2580),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2580),
    [anon_sym_ATprotocol] = ACTIONS(2582),
    [anon_sym_ATinterface] = ACTIONS(2582),
    [anon_sym_ATimplementation] = ACTIONS(2582),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2582),
    [anon_sym_ATtry] = ACTIONS(2582),
    [anon_sym___try] = ACTIONS(2580),
    [anon_sym_ATthrow] = ACTIONS(2582),
    [anon_sym_ATselector] = ACTIONS(2582),
    [anon_sym_ATavailable] = ACTIONS(2582),
    [anon_sym___builtin_available] = ACTIONS(2580),
    [anon_sym_va_arg] = ACTIONS(2580),
    [anon_sym_ATencode] = ACTIONS(2582),
    [anon_sym_ATsynchronized] = ACTIONS(2582),
    [anon_sym_BOOL] = ACTIONS(2580),
    [anon_sym_IMP] = ACTIONS(2580),
    [anon_sym_SEL] = ACTIONS(2580),
    [anon_sym_Class] = ACTIONS(2580),
    [anon_sym_id] = ACTIONS(2580),
  },
  [STATE(843)] = {
    [sym_identifier] = ACTIONS(2768),
    [aux_sym_preproc_include_token1] = ACTIONS(2768),
    [aux_sym_preproc_include_token2] = ACTIONS(2768),
    [aux_sym_preproc_def_token1] = ACTIONS(2768),
    [aux_sym_preproc_if_token1] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2768),
    [sym_preproc_directive] = ACTIONS(2768),
    [anon_sym_LPAREN2] = ACTIONS(2770),
    [anon_sym_BANG] = ACTIONS(2770),
    [anon_sym_TILDE] = ACTIONS(2770),
    [anon_sym_DASH] = ACTIONS(2768),
    [anon_sym_PLUS] = ACTIONS(2768),
    [anon_sym_STAR] = ACTIONS(2770),
    [anon_sym_CARET] = ACTIONS(2770),
    [anon_sym_AMP] = ACTIONS(2770),
    [anon_sym_SEMI] = ACTIONS(2770),
    [anon_sym___extension__] = ACTIONS(2768),
    [anon_sym_typedef] = ACTIONS(2768),
    [anon_sym_extern] = ACTIONS(2768),
    [anon_sym___attribute__] = ACTIONS(2768),
    [anon_sym___attribute] = ACTIONS(2768),
    [anon_sym_noreturn] = ACTIONS(2768),
    [anon_sym_LBRACK] = ACTIONS(2770),
    [anon_sym___declspec] = ACTIONS(2768),
    [anon_sym___cdecl] = ACTIONS(2768),
    [anon_sym___clrcall] = ACTIONS(2768),
    [anon_sym___stdcall] = ACTIONS(2768),
    [anon_sym___fastcall] = ACTIONS(2768),
    [anon_sym___thiscall] = ACTIONS(2768),
    [anon_sym___vectorcall] = ACTIONS(2768),
    [anon_sym_LBRACE] = ACTIONS(2770),
    [anon_sym_RBRACE] = ACTIONS(2770),
    [anon_sym_signed] = ACTIONS(2768),
    [anon_sym_unsigned] = ACTIONS(2768),
    [anon_sym_long] = ACTIONS(2768),
    [anon_sym_short] = ACTIONS(2768),
    [anon_sym_static] = ACTIONS(2768),
    [anon_sym_ATautoreleasepool] = ACTIONS(2770),
    [anon_sym_auto] = ACTIONS(2768),
    [anon_sym_register] = ACTIONS(2768),
    [anon_sym_inline] = ACTIONS(2768),
    [anon_sym___inline] = ACTIONS(2768),
    [anon_sym___inline__] = ACTIONS(2768),
    [anon_sym___forceinline] = ACTIONS(2768),
    [anon_sym_thread_local] = ACTIONS(2768),
    [anon_sym___thread] = ACTIONS(2768),
    [anon_sym_CG_EXTERN] = ACTIONS(2768),
    [anon_sym_CG_INLINE] = ACTIONS(2768),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2768),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2768),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2768),
    [anon_sym_IBOutlet] = ACTIONS(2768),
    [anon_sym_IBInspectable] = ACTIONS(2768),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2768),
    [anon_sym_NS_INLINE] = ACTIONS(2768),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2768),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2768),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2768),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2768),
    [anon_sym_const] = ACTIONS(2768),
    [anon_sym_constexpr] = ACTIONS(2768),
    [anon_sym_volatile] = ACTIONS(2768),
    [anon_sym_restrict] = ACTIONS(2768),
    [anon_sym___restrict__] = ACTIONS(2768),
    [anon_sym__Atomic] = ACTIONS(2768),
    [anon_sym__Noreturn] = ACTIONS(2768),
    [anon_sym__Nonnull] = ACTIONS(2768),
    [anon_sym_nullable] = ACTIONS(2768),
    [anon_sym__Complex] = ACTIONS(2768),
    [anon_sym__Nullable] = ACTIONS(2768),
    [anon_sym__Nullable_result] = ACTIONS(2768),
    [anon_sym__Null_unspecified] = ACTIONS(2768),
    [anon_sym___autoreleasing] = ACTIONS(2768),
    [anon_sym___block] = ACTIONS(2768),
    [anon_sym___bridge] = ACTIONS(2768),
    [anon_sym___bridge_retained] = ACTIONS(2768),
    [anon_sym___bridge_transfer] = ACTIONS(2768),
    [anon_sym___complex] = ACTIONS(2768),
    [anon_sym___const] = ACTIONS(2768),
    [anon_sym___imag] = ACTIONS(2768),
    [anon_sym___kindof] = ACTIONS(2768),
    [anon_sym___nonnull] = ACTIONS(2768),
    [anon_sym___nullable] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2768),
    [anon_sym___real] = ACTIONS(2768),
    [anon_sym___strong] = ACTIONS(2768),
    [anon_sym___unsafe_unretained] = ACTIONS(2768),
    [anon_sym___unused] = ACTIONS(2768),
    [anon_sym___weak] = ACTIONS(2768),
    [anon_sym_alignas] = ACTIONS(2768),
    [anon_sym__Alignas] = ACTIONS(2768),
    [sym_primitive_type] = ACTIONS(2768),
    [anon_sym_enum] = ACTIONS(2768),
    [anon_sym_struct] = ACTIONS(2768),
    [anon_sym_union] = ACTIONS(2768),
    [anon_sym_if] = ACTIONS(2768),
    [anon_sym_switch] = ACTIONS(2768),
    [anon_sym_case] = ACTIONS(2768),
    [anon_sym_default] = ACTIONS(2768),
    [anon_sym_while] = ACTIONS(2768),
    [anon_sym_do] = ACTIONS(2768),
    [anon_sym_for] = ACTIONS(2768),
    [anon_sym_in] = ACTIONS(2768),
    [anon_sym_return] = ACTIONS(2768),
    [anon_sym_break] = ACTIONS(2768),
    [anon_sym_continue] = ACTIONS(2768),
    [anon_sym_goto] = ACTIONS(2768),
    [anon_sym_DASH_DASH] = ACTIONS(2770),
    [anon_sym_PLUS_PLUS] = ACTIONS(2770),
    [anon_sym_sizeof] = ACTIONS(2768),
    [anon_sym___alignof__] = ACTIONS(2768),
    [anon_sym___alignof] = ACTIONS(2768),
    [anon_sym__alignof] = ACTIONS(2768),
    [anon_sym_alignof] = ACTIONS(2768),
    [anon_sym__Alignof] = ACTIONS(2768),
    [anon_sym_offsetof] = ACTIONS(2768),
    [anon_sym__Generic] = ACTIONS(2768),
    [anon_sym_asm] = ACTIONS(2768),
    [anon_sym___asm__] = ACTIONS(2768),
    [anon_sym___asm] = ACTIONS(2768),
    [sym_number_literal] = ACTIONS(2770),
    [anon_sym_L_SQUOTE] = ACTIONS(2770),
    [anon_sym_u_SQUOTE] = ACTIONS(2770),
    [anon_sym_U_SQUOTE] = ACTIONS(2770),
    [anon_sym_u8_SQUOTE] = ACTIONS(2770),
    [anon_sym_SQUOTE] = ACTIONS(2770),
    [anon_sym_AT] = ACTIONS(2768),
    [anon_sym_DQUOTE] = ACTIONS(2770),
    [anon_sym_L_DQUOTE] = ACTIONS(2770),
    [anon_sym_u_DQUOTE] = ACTIONS(2770),
    [anon_sym_U_DQUOTE] = ACTIONS(2770),
    [anon_sym_u8_DQUOTE] = ACTIONS(2770),
    [sym_true] = ACTIONS(2768),
    [sym_false] = ACTIONS(2768),
    [anon_sym_NULL] = ACTIONS(2768),
    [anon_sym_nullptr] = ACTIONS(2768),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2768),
    [anon_sym___typeof] = ACTIONS(2768),
    [anon_sym_typeof] = ACTIONS(2768),
    [anon_sym_ATimport] = ACTIONS(2770),
    [aux_sym_preproc_undef_token1] = ACTIONS(2768),
    [anon_sym_POUND] = ACTIONS(2768),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2768),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2768),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2768),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2768),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2768),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2768),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2768),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_API_AVAILABLE] = ACTIONS(2768),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_API_DEPRECATED] = ACTIONS(2768),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2768),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym___deprecated_msg] = ACTIONS(2768),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2768),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2768),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2768),
    [anon_sym_ATprotocol] = ACTIONS(2770),
    [anon_sym_ATinterface] = ACTIONS(2770),
    [anon_sym_ATimplementation] = ACTIONS(2770),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2770),
    [anon_sym_ATtry] = ACTIONS(2770),
    [anon_sym___try] = ACTIONS(2768),
    [anon_sym_ATthrow] = ACTIONS(2770),
    [anon_sym_ATselector] = ACTIONS(2770),
    [anon_sym_ATavailable] = ACTIONS(2770),
    [anon_sym___builtin_available] = ACTIONS(2768),
    [anon_sym_va_arg] = ACTIONS(2768),
    [anon_sym_ATencode] = ACTIONS(2770),
    [anon_sym_ATsynchronized] = ACTIONS(2770),
    [anon_sym_BOOL] = ACTIONS(2768),
    [anon_sym_IMP] = ACTIONS(2768),
    [anon_sym_SEL] = ACTIONS(2768),
    [anon_sym_Class] = ACTIONS(2768),
    [anon_sym_id] = ACTIONS(2768),
  },
  [STATE(844)] = {
    [sym_identifier] = ACTIONS(2772),
    [aux_sym_preproc_include_token1] = ACTIONS(2772),
    [aux_sym_preproc_include_token2] = ACTIONS(2772),
    [aux_sym_preproc_def_token1] = ACTIONS(2772),
    [aux_sym_preproc_if_token1] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2772),
    [sym_preproc_directive] = ACTIONS(2772),
    [anon_sym_LPAREN2] = ACTIONS(2774),
    [anon_sym_BANG] = ACTIONS(2774),
    [anon_sym_TILDE] = ACTIONS(2774),
    [anon_sym_DASH] = ACTIONS(2772),
    [anon_sym_PLUS] = ACTIONS(2772),
    [anon_sym_STAR] = ACTIONS(2774),
    [anon_sym_CARET] = ACTIONS(2774),
    [anon_sym_AMP] = ACTIONS(2774),
    [anon_sym_SEMI] = ACTIONS(2774),
    [anon_sym___extension__] = ACTIONS(2772),
    [anon_sym_typedef] = ACTIONS(2772),
    [anon_sym_extern] = ACTIONS(2772),
    [anon_sym___attribute__] = ACTIONS(2772),
    [anon_sym___attribute] = ACTIONS(2772),
    [anon_sym_noreturn] = ACTIONS(2772),
    [anon_sym_LBRACK] = ACTIONS(2774),
    [anon_sym___declspec] = ACTIONS(2772),
    [anon_sym___cdecl] = ACTIONS(2772),
    [anon_sym___clrcall] = ACTIONS(2772),
    [anon_sym___stdcall] = ACTIONS(2772),
    [anon_sym___fastcall] = ACTIONS(2772),
    [anon_sym___thiscall] = ACTIONS(2772),
    [anon_sym___vectorcall] = ACTIONS(2772),
    [anon_sym_LBRACE] = ACTIONS(2774),
    [anon_sym_RBRACE] = ACTIONS(2774),
    [anon_sym_signed] = ACTIONS(2772),
    [anon_sym_unsigned] = ACTIONS(2772),
    [anon_sym_long] = ACTIONS(2772),
    [anon_sym_short] = ACTIONS(2772),
    [anon_sym_static] = ACTIONS(2772),
    [anon_sym_ATautoreleasepool] = ACTIONS(2774),
    [anon_sym_auto] = ACTIONS(2772),
    [anon_sym_register] = ACTIONS(2772),
    [anon_sym_inline] = ACTIONS(2772),
    [anon_sym___inline] = ACTIONS(2772),
    [anon_sym___inline__] = ACTIONS(2772),
    [anon_sym___forceinline] = ACTIONS(2772),
    [anon_sym_thread_local] = ACTIONS(2772),
    [anon_sym___thread] = ACTIONS(2772),
    [anon_sym_CG_EXTERN] = ACTIONS(2772),
    [anon_sym_CG_INLINE] = ACTIONS(2772),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2772),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2772),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2772),
    [anon_sym_IBOutlet] = ACTIONS(2772),
    [anon_sym_IBInspectable] = ACTIONS(2772),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2772),
    [anon_sym_NS_INLINE] = ACTIONS(2772),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2772),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2772),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2772),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2772),
    [anon_sym_const] = ACTIONS(2772),
    [anon_sym_constexpr] = ACTIONS(2772),
    [anon_sym_volatile] = ACTIONS(2772),
    [anon_sym_restrict] = ACTIONS(2772),
    [anon_sym___restrict__] = ACTIONS(2772),
    [anon_sym__Atomic] = ACTIONS(2772),
    [anon_sym__Noreturn] = ACTIONS(2772),
    [anon_sym__Nonnull] = ACTIONS(2772),
    [anon_sym_nullable] = ACTIONS(2772),
    [anon_sym__Complex] = ACTIONS(2772),
    [anon_sym__Nullable] = ACTIONS(2772),
    [anon_sym__Nullable_result] = ACTIONS(2772),
    [anon_sym__Null_unspecified] = ACTIONS(2772),
    [anon_sym___autoreleasing] = ACTIONS(2772),
    [anon_sym___block] = ACTIONS(2772),
    [anon_sym___bridge] = ACTIONS(2772),
    [anon_sym___bridge_retained] = ACTIONS(2772),
    [anon_sym___bridge_transfer] = ACTIONS(2772),
    [anon_sym___complex] = ACTIONS(2772),
    [anon_sym___const] = ACTIONS(2772),
    [anon_sym___imag] = ACTIONS(2772),
    [anon_sym___kindof] = ACTIONS(2772),
    [anon_sym___nonnull] = ACTIONS(2772),
    [anon_sym___nullable] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2772),
    [anon_sym___real] = ACTIONS(2772),
    [anon_sym___strong] = ACTIONS(2772),
    [anon_sym___unsafe_unretained] = ACTIONS(2772),
    [anon_sym___unused] = ACTIONS(2772),
    [anon_sym___weak] = ACTIONS(2772),
    [anon_sym_alignas] = ACTIONS(2772),
    [anon_sym__Alignas] = ACTIONS(2772),
    [sym_primitive_type] = ACTIONS(2772),
    [anon_sym_enum] = ACTIONS(2772),
    [anon_sym_struct] = ACTIONS(2772),
    [anon_sym_union] = ACTIONS(2772),
    [anon_sym_if] = ACTIONS(2772),
    [anon_sym_switch] = ACTIONS(2772),
    [anon_sym_case] = ACTIONS(2772),
    [anon_sym_default] = ACTIONS(2772),
    [anon_sym_while] = ACTIONS(2772),
    [anon_sym_do] = ACTIONS(2772),
    [anon_sym_for] = ACTIONS(2772),
    [anon_sym_in] = ACTIONS(2772),
    [anon_sym_return] = ACTIONS(2772),
    [anon_sym_break] = ACTIONS(2772),
    [anon_sym_continue] = ACTIONS(2772),
    [anon_sym_goto] = ACTIONS(2772),
    [anon_sym_DASH_DASH] = ACTIONS(2774),
    [anon_sym_PLUS_PLUS] = ACTIONS(2774),
    [anon_sym_sizeof] = ACTIONS(2772),
    [anon_sym___alignof__] = ACTIONS(2772),
    [anon_sym___alignof] = ACTIONS(2772),
    [anon_sym__alignof] = ACTIONS(2772),
    [anon_sym_alignof] = ACTIONS(2772),
    [anon_sym__Alignof] = ACTIONS(2772),
    [anon_sym_offsetof] = ACTIONS(2772),
    [anon_sym__Generic] = ACTIONS(2772),
    [anon_sym_asm] = ACTIONS(2772),
    [anon_sym___asm__] = ACTIONS(2772),
    [anon_sym___asm] = ACTIONS(2772),
    [sym_number_literal] = ACTIONS(2774),
    [anon_sym_L_SQUOTE] = ACTIONS(2774),
    [anon_sym_u_SQUOTE] = ACTIONS(2774),
    [anon_sym_U_SQUOTE] = ACTIONS(2774),
    [anon_sym_u8_SQUOTE] = ACTIONS(2774),
    [anon_sym_SQUOTE] = ACTIONS(2774),
    [anon_sym_AT] = ACTIONS(2772),
    [anon_sym_DQUOTE] = ACTIONS(2774),
    [anon_sym_L_DQUOTE] = ACTIONS(2774),
    [anon_sym_u_DQUOTE] = ACTIONS(2774),
    [anon_sym_U_DQUOTE] = ACTIONS(2774),
    [anon_sym_u8_DQUOTE] = ACTIONS(2774),
    [sym_true] = ACTIONS(2772),
    [sym_false] = ACTIONS(2772),
    [anon_sym_NULL] = ACTIONS(2772),
    [anon_sym_nullptr] = ACTIONS(2772),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2772),
    [anon_sym___typeof] = ACTIONS(2772),
    [anon_sym_typeof] = ACTIONS(2772),
    [anon_sym_ATimport] = ACTIONS(2774),
    [aux_sym_preproc_undef_token1] = ACTIONS(2772),
    [anon_sym_POUND] = ACTIONS(2772),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2772),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2772),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2772),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2772),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2772),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2772),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2772),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_API_AVAILABLE] = ACTIONS(2772),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_API_DEPRECATED] = ACTIONS(2772),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2772),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym___deprecated_msg] = ACTIONS(2772),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2772),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2772),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2772),
    [anon_sym_ATprotocol] = ACTIONS(2774),
    [anon_sym_ATinterface] = ACTIONS(2774),
    [anon_sym_ATimplementation] = ACTIONS(2774),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2774),
    [anon_sym_ATtry] = ACTIONS(2774),
    [anon_sym___try] = ACTIONS(2772),
    [anon_sym_ATthrow] = ACTIONS(2774),
    [anon_sym_ATselector] = ACTIONS(2774),
    [anon_sym_ATavailable] = ACTIONS(2774),
    [anon_sym___builtin_available] = ACTIONS(2772),
    [anon_sym_va_arg] = ACTIONS(2772),
    [anon_sym_ATencode] = ACTIONS(2774),
    [anon_sym_ATsynchronized] = ACTIONS(2774),
    [anon_sym_BOOL] = ACTIONS(2772),
    [anon_sym_IMP] = ACTIONS(2772),
    [anon_sym_SEL] = ACTIONS(2772),
    [anon_sym_Class] = ACTIONS(2772),
    [anon_sym_id] = ACTIONS(2772),
  },
  [STATE(845)] = {
    [sym_identifier] = ACTIONS(2604),
    [aux_sym_preproc_include_token1] = ACTIONS(2604),
    [aux_sym_preproc_include_token2] = ACTIONS(2604),
    [aux_sym_preproc_def_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2604),
    [sym_preproc_directive] = ACTIONS(2604),
    [anon_sym_LPAREN2] = ACTIONS(2606),
    [anon_sym_BANG] = ACTIONS(2606),
    [anon_sym_TILDE] = ACTIONS(2606),
    [anon_sym_DASH] = ACTIONS(2604),
    [anon_sym_PLUS] = ACTIONS(2604),
    [anon_sym_STAR] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(2606),
    [anon_sym_AMP] = ACTIONS(2606),
    [anon_sym_SEMI] = ACTIONS(2606),
    [anon_sym___extension__] = ACTIONS(2604),
    [anon_sym_typedef] = ACTIONS(2604),
    [anon_sym_extern] = ACTIONS(2604),
    [anon_sym___attribute__] = ACTIONS(2604),
    [anon_sym___attribute] = ACTIONS(2604),
    [anon_sym_noreturn] = ACTIONS(2604),
    [anon_sym_LBRACK] = ACTIONS(2606),
    [anon_sym___declspec] = ACTIONS(2604),
    [anon_sym___cdecl] = ACTIONS(2604),
    [anon_sym___clrcall] = ACTIONS(2604),
    [anon_sym___stdcall] = ACTIONS(2604),
    [anon_sym___fastcall] = ACTIONS(2604),
    [anon_sym___thiscall] = ACTIONS(2604),
    [anon_sym___vectorcall] = ACTIONS(2604),
    [anon_sym_LBRACE] = ACTIONS(2606),
    [anon_sym_RBRACE] = ACTIONS(2606),
    [anon_sym_signed] = ACTIONS(2604),
    [anon_sym_unsigned] = ACTIONS(2604),
    [anon_sym_long] = ACTIONS(2604),
    [anon_sym_short] = ACTIONS(2604),
    [anon_sym_static] = ACTIONS(2604),
    [anon_sym_ATautoreleasepool] = ACTIONS(2606),
    [anon_sym_auto] = ACTIONS(2604),
    [anon_sym_register] = ACTIONS(2604),
    [anon_sym_inline] = ACTIONS(2604),
    [anon_sym___inline] = ACTIONS(2604),
    [anon_sym___inline__] = ACTIONS(2604),
    [anon_sym___forceinline] = ACTIONS(2604),
    [anon_sym_thread_local] = ACTIONS(2604),
    [anon_sym___thread] = ACTIONS(2604),
    [anon_sym_CG_EXTERN] = ACTIONS(2604),
    [anon_sym_CG_INLINE] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2604),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2604),
    [anon_sym_IBOutlet] = ACTIONS(2604),
    [anon_sym_IBInspectable] = ACTIONS(2604),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2604),
    [anon_sym_NS_INLINE] = ACTIONS(2604),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2604),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2604),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2604),
    [anon_sym_const] = ACTIONS(2604),
    [anon_sym_constexpr] = ACTIONS(2604),
    [anon_sym_volatile] = ACTIONS(2604),
    [anon_sym_restrict] = ACTIONS(2604),
    [anon_sym___restrict__] = ACTIONS(2604),
    [anon_sym__Atomic] = ACTIONS(2604),
    [anon_sym__Noreturn] = ACTIONS(2604),
    [anon_sym__Nonnull] = ACTIONS(2604),
    [anon_sym_nullable] = ACTIONS(2604),
    [anon_sym__Complex] = ACTIONS(2604),
    [anon_sym__Nullable] = ACTIONS(2604),
    [anon_sym__Nullable_result] = ACTIONS(2604),
    [anon_sym__Null_unspecified] = ACTIONS(2604),
    [anon_sym___autoreleasing] = ACTIONS(2604),
    [anon_sym___block] = ACTIONS(2604),
    [anon_sym___bridge] = ACTIONS(2604),
    [anon_sym___bridge_retained] = ACTIONS(2604),
    [anon_sym___bridge_transfer] = ACTIONS(2604),
    [anon_sym___complex] = ACTIONS(2604),
    [anon_sym___const] = ACTIONS(2604),
    [anon_sym___imag] = ACTIONS(2604),
    [anon_sym___kindof] = ACTIONS(2604),
    [anon_sym___nonnull] = ACTIONS(2604),
    [anon_sym___nullable] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2604),
    [anon_sym___real] = ACTIONS(2604),
    [anon_sym___strong] = ACTIONS(2604),
    [anon_sym___unsafe_unretained] = ACTIONS(2604),
    [anon_sym___unused] = ACTIONS(2604),
    [anon_sym___weak] = ACTIONS(2604),
    [anon_sym_alignas] = ACTIONS(2604),
    [anon_sym__Alignas] = ACTIONS(2604),
    [sym_primitive_type] = ACTIONS(2604),
    [anon_sym_enum] = ACTIONS(2604),
    [anon_sym_struct] = ACTIONS(2604),
    [anon_sym_union] = ACTIONS(2604),
    [anon_sym_if] = ACTIONS(2604),
    [anon_sym_switch] = ACTIONS(2604),
    [anon_sym_case] = ACTIONS(2604),
    [anon_sym_default] = ACTIONS(2604),
    [anon_sym_while] = ACTIONS(2604),
    [anon_sym_do] = ACTIONS(2604),
    [anon_sym_for] = ACTIONS(2604),
    [anon_sym_in] = ACTIONS(2604),
    [anon_sym_return] = ACTIONS(2604),
    [anon_sym_break] = ACTIONS(2604),
    [anon_sym_continue] = ACTIONS(2604),
    [anon_sym_goto] = ACTIONS(2604),
    [anon_sym_DASH_DASH] = ACTIONS(2606),
    [anon_sym_PLUS_PLUS] = ACTIONS(2606),
    [anon_sym_sizeof] = ACTIONS(2604),
    [anon_sym___alignof__] = ACTIONS(2604),
    [anon_sym___alignof] = ACTIONS(2604),
    [anon_sym__alignof] = ACTIONS(2604),
    [anon_sym_alignof] = ACTIONS(2604),
    [anon_sym__Alignof] = ACTIONS(2604),
    [anon_sym_offsetof] = ACTIONS(2604),
    [anon_sym__Generic] = ACTIONS(2604),
    [anon_sym_asm] = ACTIONS(2604),
    [anon_sym___asm__] = ACTIONS(2604),
    [anon_sym___asm] = ACTIONS(2604),
    [sym_number_literal] = ACTIONS(2606),
    [anon_sym_L_SQUOTE] = ACTIONS(2606),
    [anon_sym_u_SQUOTE] = ACTIONS(2606),
    [anon_sym_U_SQUOTE] = ACTIONS(2606),
    [anon_sym_u8_SQUOTE] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2606),
    [anon_sym_AT] = ACTIONS(2604),
    [anon_sym_DQUOTE] = ACTIONS(2606),
    [anon_sym_L_DQUOTE] = ACTIONS(2606),
    [anon_sym_u_DQUOTE] = ACTIONS(2606),
    [anon_sym_U_DQUOTE] = ACTIONS(2606),
    [anon_sym_u8_DQUOTE] = ACTIONS(2606),
    [sym_true] = ACTIONS(2604),
    [sym_false] = ACTIONS(2604),
    [anon_sym_NULL] = ACTIONS(2604),
    [anon_sym_nullptr] = ACTIONS(2604),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2604),
    [anon_sym___typeof] = ACTIONS(2604),
    [anon_sym_typeof] = ACTIONS(2604),
    [anon_sym_ATimport] = ACTIONS(2606),
    [aux_sym_preproc_undef_token1] = ACTIONS(2604),
    [anon_sym_POUND] = ACTIONS(2604),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2604),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2604),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2604),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_API_AVAILABLE] = ACTIONS(2604),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_API_DEPRECATED] = ACTIONS(2604),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym___deprecated_msg] = ACTIONS(2604),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2604),
    [anon_sym_ATprotocol] = ACTIONS(2606),
    [anon_sym_ATinterface] = ACTIONS(2606),
    [anon_sym_ATimplementation] = ACTIONS(2606),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2606),
    [anon_sym_ATtry] = ACTIONS(2606),
    [anon_sym___try] = ACTIONS(2604),
    [anon_sym_ATthrow] = ACTIONS(2606),
    [anon_sym_ATselector] = ACTIONS(2606),
    [anon_sym_ATavailable] = ACTIONS(2606),
    [anon_sym___builtin_available] = ACTIONS(2604),
    [anon_sym_va_arg] = ACTIONS(2604),
    [anon_sym_ATencode] = ACTIONS(2606),
    [anon_sym_ATsynchronized] = ACTIONS(2606),
    [anon_sym_BOOL] = ACTIONS(2604),
    [anon_sym_IMP] = ACTIONS(2604),
    [anon_sym_SEL] = ACTIONS(2604),
    [anon_sym_Class] = ACTIONS(2604),
    [anon_sym_id] = ACTIONS(2604),
  },
  [STATE(846)] = {
    [sym_identifier] = ACTIONS(2684),
    [aux_sym_preproc_include_token1] = ACTIONS(2684),
    [aux_sym_preproc_include_token2] = ACTIONS(2684),
    [aux_sym_preproc_def_token1] = ACTIONS(2684),
    [aux_sym_preproc_if_token1] = ACTIONS(2684),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2684),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2684),
    [sym_preproc_directive] = ACTIONS(2684),
    [anon_sym_LPAREN2] = ACTIONS(2686),
    [anon_sym_BANG] = ACTIONS(2686),
    [anon_sym_TILDE] = ACTIONS(2686),
    [anon_sym_DASH] = ACTIONS(2684),
    [anon_sym_PLUS] = ACTIONS(2684),
    [anon_sym_STAR] = ACTIONS(2686),
    [anon_sym_CARET] = ACTIONS(2686),
    [anon_sym_AMP] = ACTIONS(2686),
    [anon_sym_SEMI] = ACTIONS(2686),
    [anon_sym___extension__] = ACTIONS(2684),
    [anon_sym_typedef] = ACTIONS(2684),
    [anon_sym_extern] = ACTIONS(2684),
    [anon_sym___attribute__] = ACTIONS(2684),
    [anon_sym___attribute] = ACTIONS(2684),
    [anon_sym_noreturn] = ACTIONS(2684),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym___declspec] = ACTIONS(2684),
    [anon_sym___cdecl] = ACTIONS(2684),
    [anon_sym___clrcall] = ACTIONS(2684),
    [anon_sym___stdcall] = ACTIONS(2684),
    [anon_sym___fastcall] = ACTIONS(2684),
    [anon_sym___thiscall] = ACTIONS(2684),
    [anon_sym___vectorcall] = ACTIONS(2684),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_RBRACE] = ACTIONS(2686),
    [anon_sym_signed] = ACTIONS(2684),
    [anon_sym_unsigned] = ACTIONS(2684),
    [anon_sym_long] = ACTIONS(2684),
    [anon_sym_short] = ACTIONS(2684),
    [anon_sym_static] = ACTIONS(2684),
    [anon_sym_ATautoreleasepool] = ACTIONS(2686),
    [anon_sym_auto] = ACTIONS(2684),
    [anon_sym_register] = ACTIONS(2684),
    [anon_sym_inline] = ACTIONS(2684),
    [anon_sym___inline] = ACTIONS(2684),
    [anon_sym___inline__] = ACTIONS(2684),
    [anon_sym___forceinline] = ACTIONS(2684),
    [anon_sym_thread_local] = ACTIONS(2684),
    [anon_sym___thread] = ACTIONS(2684),
    [anon_sym_CG_EXTERN] = ACTIONS(2684),
    [anon_sym_CG_INLINE] = ACTIONS(2684),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2684),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2684),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2684),
    [anon_sym_IBOutlet] = ACTIONS(2684),
    [anon_sym_IBInspectable] = ACTIONS(2684),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2684),
    [anon_sym_NS_INLINE] = ACTIONS(2684),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2684),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2684),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2684),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2684),
    [anon_sym_const] = ACTIONS(2684),
    [anon_sym_constexpr] = ACTIONS(2684),
    [anon_sym_volatile] = ACTIONS(2684),
    [anon_sym_restrict] = ACTIONS(2684),
    [anon_sym___restrict__] = ACTIONS(2684),
    [anon_sym__Atomic] = ACTIONS(2684),
    [anon_sym__Noreturn] = ACTIONS(2684),
    [anon_sym__Nonnull] = ACTIONS(2684),
    [anon_sym_nullable] = ACTIONS(2684),
    [anon_sym__Complex] = ACTIONS(2684),
    [anon_sym__Nullable] = ACTIONS(2684),
    [anon_sym__Nullable_result] = ACTIONS(2684),
    [anon_sym__Null_unspecified] = ACTIONS(2684),
    [anon_sym___autoreleasing] = ACTIONS(2684),
    [anon_sym___block] = ACTIONS(2684),
    [anon_sym___bridge] = ACTIONS(2684),
    [anon_sym___bridge_retained] = ACTIONS(2684),
    [anon_sym___bridge_transfer] = ACTIONS(2684),
    [anon_sym___complex] = ACTIONS(2684),
    [anon_sym___const] = ACTIONS(2684),
    [anon_sym___imag] = ACTIONS(2684),
    [anon_sym___kindof] = ACTIONS(2684),
    [anon_sym___nonnull] = ACTIONS(2684),
    [anon_sym___nullable] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2684),
    [anon_sym___real] = ACTIONS(2684),
    [anon_sym___strong] = ACTIONS(2684),
    [anon_sym___unsafe_unretained] = ACTIONS(2684),
    [anon_sym___unused] = ACTIONS(2684),
    [anon_sym___weak] = ACTIONS(2684),
    [anon_sym_alignas] = ACTIONS(2684),
    [anon_sym__Alignas] = ACTIONS(2684),
    [sym_primitive_type] = ACTIONS(2684),
    [anon_sym_enum] = ACTIONS(2684),
    [anon_sym_struct] = ACTIONS(2684),
    [anon_sym_union] = ACTIONS(2684),
    [anon_sym_if] = ACTIONS(2684),
    [anon_sym_switch] = ACTIONS(2684),
    [anon_sym_case] = ACTIONS(2684),
    [anon_sym_default] = ACTIONS(2684),
    [anon_sym_while] = ACTIONS(2684),
    [anon_sym_do] = ACTIONS(2684),
    [anon_sym_for] = ACTIONS(2684),
    [anon_sym_in] = ACTIONS(2684),
    [anon_sym_return] = ACTIONS(2684),
    [anon_sym_break] = ACTIONS(2684),
    [anon_sym_continue] = ACTIONS(2684),
    [anon_sym_goto] = ACTIONS(2684),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_sizeof] = ACTIONS(2684),
    [anon_sym___alignof__] = ACTIONS(2684),
    [anon_sym___alignof] = ACTIONS(2684),
    [anon_sym__alignof] = ACTIONS(2684),
    [anon_sym_alignof] = ACTIONS(2684),
    [anon_sym__Alignof] = ACTIONS(2684),
    [anon_sym_offsetof] = ACTIONS(2684),
    [anon_sym__Generic] = ACTIONS(2684),
    [anon_sym_asm] = ACTIONS(2684),
    [anon_sym___asm__] = ACTIONS(2684),
    [anon_sym___asm] = ACTIONS(2684),
    [sym_number_literal] = ACTIONS(2686),
    [anon_sym_L_SQUOTE] = ACTIONS(2686),
    [anon_sym_u_SQUOTE] = ACTIONS(2686),
    [anon_sym_U_SQUOTE] = ACTIONS(2686),
    [anon_sym_u8_SQUOTE] = ACTIONS(2686),
    [anon_sym_SQUOTE] = ACTIONS(2686),
    [anon_sym_AT] = ACTIONS(2684),
    [anon_sym_DQUOTE] = ACTIONS(2686),
    [anon_sym_L_DQUOTE] = ACTIONS(2686),
    [anon_sym_u_DQUOTE] = ACTIONS(2686),
    [anon_sym_U_DQUOTE] = ACTIONS(2686),
    [anon_sym_u8_DQUOTE] = ACTIONS(2686),
    [sym_true] = ACTIONS(2684),
    [sym_false] = ACTIONS(2684),
    [anon_sym_NULL] = ACTIONS(2684),
    [anon_sym_nullptr] = ACTIONS(2684),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2684),
    [anon_sym___typeof] = ACTIONS(2684),
    [anon_sym_typeof] = ACTIONS(2684),
    [anon_sym_ATimport] = ACTIONS(2686),
    [aux_sym_preproc_undef_token1] = ACTIONS(2684),
    [anon_sym_POUND] = ACTIONS(2684),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2684),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2684),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2684),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2684),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2684),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2684),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2684),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_API_AVAILABLE] = ACTIONS(2684),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_API_DEPRECATED] = ACTIONS(2684),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2684),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym___deprecated_msg] = ACTIONS(2684),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2684),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2684),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2684),
    [anon_sym_ATprotocol] = ACTIONS(2686),
    [anon_sym_ATinterface] = ACTIONS(2686),
    [anon_sym_ATimplementation] = ACTIONS(2686),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2686),
    [anon_sym_ATtry] = ACTIONS(2686),
    [anon_sym___try] = ACTIONS(2684),
    [anon_sym_ATthrow] = ACTIONS(2686),
    [anon_sym_ATselector] = ACTIONS(2686),
    [anon_sym_ATavailable] = ACTIONS(2686),
    [anon_sym___builtin_available] = ACTIONS(2684),
    [anon_sym_va_arg] = ACTIONS(2684),
    [anon_sym_ATencode] = ACTIONS(2686),
    [anon_sym_ATsynchronized] = ACTIONS(2686),
    [anon_sym_BOOL] = ACTIONS(2684),
    [anon_sym_IMP] = ACTIONS(2684),
    [anon_sym_SEL] = ACTIONS(2684),
    [anon_sym_Class] = ACTIONS(2684),
    [anon_sym_id] = ACTIONS(2684),
  },
  [STATE(847)] = {
    [sym_identifier] = ACTIONS(2692),
    [aux_sym_preproc_include_token1] = ACTIONS(2692),
    [aux_sym_preproc_include_token2] = ACTIONS(2692),
    [aux_sym_preproc_def_token1] = ACTIONS(2692),
    [aux_sym_preproc_if_token1] = ACTIONS(2692),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2692),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2692),
    [sym_preproc_directive] = ACTIONS(2692),
    [anon_sym_LPAREN2] = ACTIONS(2694),
    [anon_sym_BANG] = ACTIONS(2694),
    [anon_sym_TILDE] = ACTIONS(2694),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2694),
    [anon_sym_CARET] = ACTIONS(2694),
    [anon_sym_AMP] = ACTIONS(2694),
    [anon_sym_SEMI] = ACTIONS(2694),
    [anon_sym___extension__] = ACTIONS(2692),
    [anon_sym_typedef] = ACTIONS(2692),
    [anon_sym_extern] = ACTIONS(2692),
    [anon_sym___attribute__] = ACTIONS(2692),
    [anon_sym___attribute] = ACTIONS(2692),
    [anon_sym_noreturn] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym___declspec] = ACTIONS(2692),
    [anon_sym___cdecl] = ACTIONS(2692),
    [anon_sym___clrcall] = ACTIONS(2692),
    [anon_sym___stdcall] = ACTIONS(2692),
    [anon_sym___fastcall] = ACTIONS(2692),
    [anon_sym___thiscall] = ACTIONS(2692),
    [anon_sym___vectorcall] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_RBRACE] = ACTIONS(2694),
    [anon_sym_signed] = ACTIONS(2692),
    [anon_sym_unsigned] = ACTIONS(2692),
    [anon_sym_long] = ACTIONS(2692),
    [anon_sym_short] = ACTIONS(2692),
    [anon_sym_static] = ACTIONS(2692),
    [anon_sym_ATautoreleasepool] = ACTIONS(2694),
    [anon_sym_auto] = ACTIONS(2692),
    [anon_sym_register] = ACTIONS(2692),
    [anon_sym_inline] = ACTIONS(2692),
    [anon_sym___inline] = ACTIONS(2692),
    [anon_sym___inline__] = ACTIONS(2692),
    [anon_sym___forceinline] = ACTIONS(2692),
    [anon_sym_thread_local] = ACTIONS(2692),
    [anon_sym___thread] = ACTIONS(2692),
    [anon_sym_CG_EXTERN] = ACTIONS(2692),
    [anon_sym_CG_INLINE] = ACTIONS(2692),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2692),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2692),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2692),
    [anon_sym_IBOutlet] = ACTIONS(2692),
    [anon_sym_IBInspectable] = ACTIONS(2692),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2692),
    [anon_sym_NS_INLINE] = ACTIONS(2692),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2692),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2692),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2692),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2692),
    [anon_sym_const] = ACTIONS(2692),
    [anon_sym_constexpr] = ACTIONS(2692),
    [anon_sym_volatile] = ACTIONS(2692),
    [anon_sym_restrict] = ACTIONS(2692),
    [anon_sym___restrict__] = ACTIONS(2692),
    [anon_sym__Atomic] = ACTIONS(2692),
    [anon_sym__Noreturn] = ACTIONS(2692),
    [anon_sym__Nonnull] = ACTIONS(2692),
    [anon_sym_nullable] = ACTIONS(2692),
    [anon_sym__Complex] = ACTIONS(2692),
    [anon_sym__Nullable] = ACTIONS(2692),
    [anon_sym__Nullable_result] = ACTIONS(2692),
    [anon_sym__Null_unspecified] = ACTIONS(2692),
    [anon_sym___autoreleasing] = ACTIONS(2692),
    [anon_sym___block] = ACTIONS(2692),
    [anon_sym___bridge] = ACTIONS(2692),
    [anon_sym___bridge_retained] = ACTIONS(2692),
    [anon_sym___bridge_transfer] = ACTIONS(2692),
    [anon_sym___complex] = ACTIONS(2692),
    [anon_sym___const] = ACTIONS(2692),
    [anon_sym___imag] = ACTIONS(2692),
    [anon_sym___kindof] = ACTIONS(2692),
    [anon_sym___nonnull] = ACTIONS(2692),
    [anon_sym___nullable] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2692),
    [anon_sym___real] = ACTIONS(2692),
    [anon_sym___strong] = ACTIONS(2692),
    [anon_sym___unsafe_unretained] = ACTIONS(2692),
    [anon_sym___unused] = ACTIONS(2692),
    [anon_sym___weak] = ACTIONS(2692),
    [anon_sym_alignas] = ACTIONS(2692),
    [anon_sym__Alignas] = ACTIONS(2692),
    [sym_primitive_type] = ACTIONS(2692),
    [anon_sym_enum] = ACTIONS(2692),
    [anon_sym_struct] = ACTIONS(2692),
    [anon_sym_union] = ACTIONS(2692),
    [anon_sym_if] = ACTIONS(2692),
    [anon_sym_switch] = ACTIONS(2692),
    [anon_sym_case] = ACTIONS(2692),
    [anon_sym_default] = ACTIONS(2692),
    [anon_sym_while] = ACTIONS(2692),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_for] = ACTIONS(2692),
    [anon_sym_in] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_sizeof] = ACTIONS(2692),
    [anon_sym___alignof__] = ACTIONS(2692),
    [anon_sym___alignof] = ACTIONS(2692),
    [anon_sym__alignof] = ACTIONS(2692),
    [anon_sym_alignof] = ACTIONS(2692),
    [anon_sym__Alignof] = ACTIONS(2692),
    [anon_sym_offsetof] = ACTIONS(2692),
    [anon_sym__Generic] = ACTIONS(2692),
    [anon_sym_asm] = ACTIONS(2692),
    [anon_sym___asm__] = ACTIONS(2692),
    [anon_sym___asm] = ACTIONS(2692),
    [sym_number_literal] = ACTIONS(2694),
    [anon_sym_L_SQUOTE] = ACTIONS(2694),
    [anon_sym_u_SQUOTE] = ACTIONS(2694),
    [anon_sym_U_SQUOTE] = ACTIONS(2694),
    [anon_sym_u8_SQUOTE] = ACTIONS(2694),
    [anon_sym_SQUOTE] = ACTIONS(2694),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_DQUOTE] = ACTIONS(2694),
    [anon_sym_L_DQUOTE] = ACTIONS(2694),
    [anon_sym_u_DQUOTE] = ACTIONS(2694),
    [anon_sym_U_DQUOTE] = ACTIONS(2694),
    [anon_sym_u8_DQUOTE] = ACTIONS(2694),
    [sym_true] = ACTIONS(2692),
    [sym_false] = ACTIONS(2692),
    [anon_sym_NULL] = ACTIONS(2692),
    [anon_sym_nullptr] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2692),
    [anon_sym___typeof] = ACTIONS(2692),
    [anon_sym_typeof] = ACTIONS(2692),
    [anon_sym_ATimport] = ACTIONS(2694),
    [aux_sym_preproc_undef_token1] = ACTIONS(2692),
    [anon_sym_POUND] = ACTIONS(2692),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2692),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2692),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2692),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2692),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2692),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2692),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2692),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_API_AVAILABLE] = ACTIONS(2692),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_API_DEPRECATED] = ACTIONS(2692),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2692),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym___deprecated_msg] = ACTIONS(2692),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2692),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2692),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2692),
    [anon_sym_ATprotocol] = ACTIONS(2694),
    [anon_sym_ATinterface] = ACTIONS(2694),
    [anon_sym_ATimplementation] = ACTIONS(2694),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2694),
    [anon_sym_ATtry] = ACTIONS(2694),
    [anon_sym___try] = ACTIONS(2692),
    [anon_sym_ATthrow] = ACTIONS(2694),
    [anon_sym_ATselector] = ACTIONS(2694),
    [anon_sym_ATavailable] = ACTIONS(2694),
    [anon_sym___builtin_available] = ACTIONS(2692),
    [anon_sym_va_arg] = ACTIONS(2692),
    [anon_sym_ATencode] = ACTIONS(2694),
    [anon_sym_ATsynchronized] = ACTIONS(2694),
    [anon_sym_BOOL] = ACTIONS(2692),
    [anon_sym_IMP] = ACTIONS(2692),
    [anon_sym_SEL] = ACTIONS(2692),
    [anon_sym_Class] = ACTIONS(2692),
    [anon_sym_id] = ACTIONS(2692),
  },
  [STATE(848)] = {
    [sym_identifier] = ACTIONS(2696),
    [aux_sym_preproc_include_token1] = ACTIONS(2696),
    [aux_sym_preproc_include_token2] = ACTIONS(2696),
    [aux_sym_preproc_def_token1] = ACTIONS(2696),
    [aux_sym_preproc_if_token1] = ACTIONS(2696),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2696),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2696),
    [sym_preproc_directive] = ACTIONS(2696),
    [anon_sym_LPAREN2] = ACTIONS(2698),
    [anon_sym_BANG] = ACTIONS(2698),
    [anon_sym_TILDE] = ACTIONS(2698),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2698),
    [anon_sym_CARET] = ACTIONS(2698),
    [anon_sym_AMP] = ACTIONS(2698),
    [anon_sym_SEMI] = ACTIONS(2698),
    [anon_sym___extension__] = ACTIONS(2696),
    [anon_sym_typedef] = ACTIONS(2696),
    [anon_sym_extern] = ACTIONS(2696),
    [anon_sym___attribute__] = ACTIONS(2696),
    [anon_sym___attribute] = ACTIONS(2696),
    [anon_sym_noreturn] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym___declspec] = ACTIONS(2696),
    [anon_sym___cdecl] = ACTIONS(2696),
    [anon_sym___clrcall] = ACTIONS(2696),
    [anon_sym___stdcall] = ACTIONS(2696),
    [anon_sym___fastcall] = ACTIONS(2696),
    [anon_sym___thiscall] = ACTIONS(2696),
    [anon_sym___vectorcall] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_RBRACE] = ACTIONS(2698),
    [anon_sym_signed] = ACTIONS(2696),
    [anon_sym_unsigned] = ACTIONS(2696),
    [anon_sym_long] = ACTIONS(2696),
    [anon_sym_short] = ACTIONS(2696),
    [anon_sym_static] = ACTIONS(2696),
    [anon_sym_ATautoreleasepool] = ACTIONS(2698),
    [anon_sym_auto] = ACTIONS(2696),
    [anon_sym_register] = ACTIONS(2696),
    [anon_sym_inline] = ACTIONS(2696),
    [anon_sym___inline] = ACTIONS(2696),
    [anon_sym___inline__] = ACTIONS(2696),
    [anon_sym___forceinline] = ACTIONS(2696),
    [anon_sym_thread_local] = ACTIONS(2696),
    [anon_sym___thread] = ACTIONS(2696),
    [anon_sym_CG_EXTERN] = ACTIONS(2696),
    [anon_sym_CG_INLINE] = ACTIONS(2696),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2696),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2696),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2696),
    [anon_sym_IBOutlet] = ACTIONS(2696),
    [anon_sym_IBInspectable] = ACTIONS(2696),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2696),
    [anon_sym_NS_INLINE] = ACTIONS(2696),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2696),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2696),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2696),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2696),
    [anon_sym_const] = ACTIONS(2696),
    [anon_sym_constexpr] = ACTIONS(2696),
    [anon_sym_volatile] = ACTIONS(2696),
    [anon_sym_restrict] = ACTIONS(2696),
    [anon_sym___restrict__] = ACTIONS(2696),
    [anon_sym__Atomic] = ACTIONS(2696),
    [anon_sym__Noreturn] = ACTIONS(2696),
    [anon_sym__Nonnull] = ACTIONS(2696),
    [anon_sym_nullable] = ACTIONS(2696),
    [anon_sym__Complex] = ACTIONS(2696),
    [anon_sym__Nullable] = ACTIONS(2696),
    [anon_sym__Nullable_result] = ACTIONS(2696),
    [anon_sym__Null_unspecified] = ACTIONS(2696),
    [anon_sym___autoreleasing] = ACTIONS(2696),
    [anon_sym___block] = ACTIONS(2696),
    [anon_sym___bridge] = ACTIONS(2696),
    [anon_sym___bridge_retained] = ACTIONS(2696),
    [anon_sym___bridge_transfer] = ACTIONS(2696),
    [anon_sym___complex] = ACTIONS(2696),
    [anon_sym___const] = ACTIONS(2696),
    [anon_sym___imag] = ACTIONS(2696),
    [anon_sym___kindof] = ACTIONS(2696),
    [anon_sym___nonnull] = ACTIONS(2696),
    [anon_sym___nullable] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2696),
    [anon_sym___real] = ACTIONS(2696),
    [anon_sym___strong] = ACTIONS(2696),
    [anon_sym___unsafe_unretained] = ACTIONS(2696),
    [anon_sym___unused] = ACTIONS(2696),
    [anon_sym___weak] = ACTIONS(2696),
    [anon_sym_alignas] = ACTIONS(2696),
    [anon_sym__Alignas] = ACTIONS(2696),
    [sym_primitive_type] = ACTIONS(2696),
    [anon_sym_enum] = ACTIONS(2696),
    [anon_sym_struct] = ACTIONS(2696),
    [anon_sym_union] = ACTIONS(2696),
    [anon_sym_if] = ACTIONS(2696),
    [anon_sym_switch] = ACTIONS(2696),
    [anon_sym_case] = ACTIONS(2696),
    [anon_sym_default] = ACTIONS(2696),
    [anon_sym_while] = ACTIONS(2696),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_for] = ACTIONS(2696),
    [anon_sym_in] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_sizeof] = ACTIONS(2696),
    [anon_sym___alignof__] = ACTIONS(2696),
    [anon_sym___alignof] = ACTIONS(2696),
    [anon_sym__alignof] = ACTIONS(2696),
    [anon_sym_alignof] = ACTIONS(2696),
    [anon_sym__Alignof] = ACTIONS(2696),
    [anon_sym_offsetof] = ACTIONS(2696),
    [anon_sym__Generic] = ACTIONS(2696),
    [anon_sym_asm] = ACTIONS(2696),
    [anon_sym___asm__] = ACTIONS(2696),
    [anon_sym___asm] = ACTIONS(2696),
    [sym_number_literal] = ACTIONS(2698),
    [anon_sym_L_SQUOTE] = ACTIONS(2698),
    [anon_sym_u_SQUOTE] = ACTIONS(2698),
    [anon_sym_U_SQUOTE] = ACTIONS(2698),
    [anon_sym_u8_SQUOTE] = ACTIONS(2698),
    [anon_sym_SQUOTE] = ACTIONS(2698),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_DQUOTE] = ACTIONS(2698),
    [anon_sym_L_DQUOTE] = ACTIONS(2698),
    [anon_sym_u_DQUOTE] = ACTIONS(2698),
    [anon_sym_U_DQUOTE] = ACTIONS(2698),
    [anon_sym_u8_DQUOTE] = ACTIONS(2698),
    [sym_true] = ACTIONS(2696),
    [sym_false] = ACTIONS(2696),
    [anon_sym_NULL] = ACTIONS(2696),
    [anon_sym_nullptr] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2696),
    [anon_sym___typeof] = ACTIONS(2696),
    [anon_sym_typeof] = ACTIONS(2696),
    [anon_sym_ATimport] = ACTIONS(2698),
    [aux_sym_preproc_undef_token1] = ACTIONS(2696),
    [anon_sym_POUND] = ACTIONS(2696),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2696),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2696),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2696),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2696),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2696),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2696),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2696),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_API_AVAILABLE] = ACTIONS(2696),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_API_DEPRECATED] = ACTIONS(2696),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2696),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym___deprecated_msg] = ACTIONS(2696),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2696),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2696),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2696),
    [anon_sym_ATprotocol] = ACTIONS(2698),
    [anon_sym_ATinterface] = ACTIONS(2698),
    [anon_sym_ATimplementation] = ACTIONS(2698),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2698),
    [anon_sym_ATtry] = ACTIONS(2698),
    [anon_sym___try] = ACTIONS(2696),
    [anon_sym_ATthrow] = ACTIONS(2698),
    [anon_sym_ATselector] = ACTIONS(2698),
    [anon_sym_ATavailable] = ACTIONS(2698),
    [anon_sym___builtin_available] = ACTIONS(2696),
    [anon_sym_va_arg] = ACTIONS(2696),
    [anon_sym_ATencode] = ACTIONS(2698),
    [anon_sym_ATsynchronized] = ACTIONS(2698),
    [anon_sym_BOOL] = ACTIONS(2696),
    [anon_sym_IMP] = ACTIONS(2696),
    [anon_sym_SEL] = ACTIONS(2696),
    [anon_sym_Class] = ACTIONS(2696),
    [anon_sym_id] = ACTIONS(2696),
  },
  [STATE(849)] = {
    [sym_identifier] = ACTIONS(2638),
    [aux_sym_preproc_include_token1] = ACTIONS(2638),
    [aux_sym_preproc_include_token2] = ACTIONS(2638),
    [aux_sym_preproc_def_token1] = ACTIONS(2638),
    [aux_sym_preproc_if_token1] = ACTIONS(2638),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2638),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2638),
    [sym_preproc_directive] = ACTIONS(2638),
    [anon_sym_LPAREN2] = ACTIONS(2640),
    [anon_sym_BANG] = ACTIONS(2640),
    [anon_sym_TILDE] = ACTIONS(2640),
    [anon_sym_DASH] = ACTIONS(2638),
    [anon_sym_PLUS] = ACTIONS(2638),
    [anon_sym_STAR] = ACTIONS(2640),
    [anon_sym_CARET] = ACTIONS(2640),
    [anon_sym_AMP] = ACTIONS(2640),
    [anon_sym_SEMI] = ACTIONS(2640),
    [anon_sym___extension__] = ACTIONS(2638),
    [anon_sym_typedef] = ACTIONS(2638),
    [anon_sym_extern] = ACTIONS(2638),
    [anon_sym___attribute__] = ACTIONS(2638),
    [anon_sym___attribute] = ACTIONS(2638),
    [anon_sym_noreturn] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2640),
    [anon_sym___declspec] = ACTIONS(2638),
    [anon_sym___cdecl] = ACTIONS(2638),
    [anon_sym___clrcall] = ACTIONS(2638),
    [anon_sym___stdcall] = ACTIONS(2638),
    [anon_sym___fastcall] = ACTIONS(2638),
    [anon_sym___thiscall] = ACTIONS(2638),
    [anon_sym___vectorcall] = ACTIONS(2638),
    [anon_sym_LBRACE] = ACTIONS(2640),
    [anon_sym_RBRACE] = ACTIONS(2640),
    [anon_sym_signed] = ACTIONS(2638),
    [anon_sym_unsigned] = ACTIONS(2638),
    [anon_sym_long] = ACTIONS(2638),
    [anon_sym_short] = ACTIONS(2638),
    [anon_sym_static] = ACTIONS(2638),
    [anon_sym_ATautoreleasepool] = ACTIONS(2640),
    [anon_sym_auto] = ACTIONS(2638),
    [anon_sym_register] = ACTIONS(2638),
    [anon_sym_inline] = ACTIONS(2638),
    [anon_sym___inline] = ACTIONS(2638),
    [anon_sym___inline__] = ACTIONS(2638),
    [anon_sym___forceinline] = ACTIONS(2638),
    [anon_sym_thread_local] = ACTIONS(2638),
    [anon_sym___thread] = ACTIONS(2638),
    [anon_sym_CG_EXTERN] = ACTIONS(2638),
    [anon_sym_CG_INLINE] = ACTIONS(2638),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2638),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2638),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2638),
    [anon_sym_IBOutlet] = ACTIONS(2638),
    [anon_sym_IBInspectable] = ACTIONS(2638),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2638),
    [anon_sym_NS_INLINE] = ACTIONS(2638),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2638),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2638),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2638),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2638),
    [anon_sym_const] = ACTIONS(2638),
    [anon_sym_constexpr] = ACTIONS(2638),
    [anon_sym_volatile] = ACTIONS(2638),
    [anon_sym_restrict] = ACTIONS(2638),
    [anon_sym___restrict__] = ACTIONS(2638),
    [anon_sym__Atomic] = ACTIONS(2638),
    [anon_sym__Noreturn] = ACTIONS(2638),
    [anon_sym__Nonnull] = ACTIONS(2638),
    [anon_sym_nullable] = ACTIONS(2638),
    [anon_sym__Complex] = ACTIONS(2638),
    [anon_sym__Nullable] = ACTIONS(2638),
    [anon_sym__Nullable_result] = ACTIONS(2638),
    [anon_sym__Null_unspecified] = ACTIONS(2638),
    [anon_sym___autoreleasing] = ACTIONS(2638),
    [anon_sym___block] = ACTIONS(2638),
    [anon_sym___bridge] = ACTIONS(2638),
    [anon_sym___bridge_retained] = ACTIONS(2638),
    [anon_sym___bridge_transfer] = ACTIONS(2638),
    [anon_sym___complex] = ACTIONS(2638),
    [anon_sym___const] = ACTIONS(2638),
    [anon_sym___imag] = ACTIONS(2638),
    [anon_sym___kindof] = ACTIONS(2638),
    [anon_sym___nonnull] = ACTIONS(2638),
    [anon_sym___nullable] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2638),
    [anon_sym___real] = ACTIONS(2638),
    [anon_sym___strong] = ACTIONS(2638),
    [anon_sym___unsafe_unretained] = ACTIONS(2638),
    [anon_sym___unused] = ACTIONS(2638),
    [anon_sym___weak] = ACTIONS(2638),
    [anon_sym_alignas] = ACTIONS(2638),
    [anon_sym__Alignas] = ACTIONS(2638),
    [sym_primitive_type] = ACTIONS(2638),
    [anon_sym_enum] = ACTIONS(2638),
    [anon_sym_struct] = ACTIONS(2638),
    [anon_sym_union] = ACTIONS(2638),
    [anon_sym_if] = ACTIONS(2638),
    [anon_sym_switch] = ACTIONS(2638),
    [anon_sym_case] = ACTIONS(2638),
    [anon_sym_default] = ACTIONS(2638),
    [anon_sym_while] = ACTIONS(2638),
    [anon_sym_do] = ACTIONS(2638),
    [anon_sym_for] = ACTIONS(2638),
    [anon_sym_in] = ACTIONS(2638),
    [anon_sym_return] = ACTIONS(2638),
    [anon_sym_break] = ACTIONS(2638),
    [anon_sym_continue] = ACTIONS(2638),
    [anon_sym_goto] = ACTIONS(2638),
    [anon_sym_DASH_DASH] = ACTIONS(2640),
    [anon_sym_PLUS_PLUS] = ACTIONS(2640),
    [anon_sym_sizeof] = ACTIONS(2638),
    [anon_sym___alignof__] = ACTIONS(2638),
    [anon_sym___alignof] = ACTIONS(2638),
    [anon_sym__alignof] = ACTIONS(2638),
    [anon_sym_alignof] = ACTIONS(2638),
    [anon_sym__Alignof] = ACTIONS(2638),
    [anon_sym_offsetof] = ACTIONS(2638),
    [anon_sym__Generic] = ACTIONS(2638),
    [anon_sym_asm] = ACTIONS(2638),
    [anon_sym___asm__] = ACTIONS(2638),
    [anon_sym___asm] = ACTIONS(2638),
    [sym_number_literal] = ACTIONS(2640),
    [anon_sym_L_SQUOTE] = ACTIONS(2640),
    [anon_sym_u_SQUOTE] = ACTIONS(2640),
    [anon_sym_U_SQUOTE] = ACTIONS(2640),
    [anon_sym_u8_SQUOTE] = ACTIONS(2640),
    [anon_sym_SQUOTE] = ACTIONS(2640),
    [anon_sym_AT] = ACTIONS(2638),
    [anon_sym_DQUOTE] = ACTIONS(2640),
    [anon_sym_L_DQUOTE] = ACTIONS(2640),
    [anon_sym_u_DQUOTE] = ACTIONS(2640),
    [anon_sym_U_DQUOTE] = ACTIONS(2640),
    [anon_sym_u8_DQUOTE] = ACTIONS(2640),
    [sym_true] = ACTIONS(2638),
    [sym_false] = ACTIONS(2638),
    [anon_sym_NULL] = ACTIONS(2638),
    [anon_sym_nullptr] = ACTIONS(2638),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2638),
    [anon_sym___typeof] = ACTIONS(2638),
    [anon_sym_typeof] = ACTIONS(2638),
    [anon_sym_ATimport] = ACTIONS(2640),
    [aux_sym_preproc_undef_token1] = ACTIONS(2638),
    [anon_sym_POUND] = ACTIONS(2638),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2638),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2638),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2638),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2638),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2638),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2638),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2638),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_API_AVAILABLE] = ACTIONS(2638),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_API_DEPRECATED] = ACTIONS(2638),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2638),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym___deprecated_msg] = ACTIONS(2638),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2638),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2638),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2638),
    [anon_sym_ATprotocol] = ACTIONS(2640),
    [anon_sym_ATinterface] = ACTIONS(2640),
    [anon_sym_ATimplementation] = ACTIONS(2640),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2640),
    [anon_sym_ATtry] = ACTIONS(2640),
    [anon_sym___try] = ACTIONS(2638),
    [anon_sym_ATthrow] = ACTIONS(2640),
    [anon_sym_ATselector] = ACTIONS(2640),
    [anon_sym_ATavailable] = ACTIONS(2640),
    [anon_sym___builtin_available] = ACTIONS(2638),
    [anon_sym_va_arg] = ACTIONS(2638),
    [anon_sym_ATencode] = ACTIONS(2640),
    [anon_sym_ATsynchronized] = ACTIONS(2640),
    [anon_sym_BOOL] = ACTIONS(2638),
    [anon_sym_IMP] = ACTIONS(2638),
    [anon_sym_SEL] = ACTIONS(2638),
    [anon_sym_Class] = ACTIONS(2638),
    [anon_sym_id] = ACTIONS(2638),
  },
  [STATE(850)] = {
    [sym_identifier] = ACTIONS(2776),
    [aux_sym_preproc_include_token1] = ACTIONS(2776),
    [aux_sym_preproc_include_token2] = ACTIONS(2776),
    [aux_sym_preproc_def_token1] = ACTIONS(2776),
    [aux_sym_preproc_if_token1] = ACTIONS(2776),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2776),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2776),
    [sym_preproc_directive] = ACTIONS(2776),
    [anon_sym_LPAREN2] = ACTIONS(2778),
    [anon_sym_BANG] = ACTIONS(2778),
    [anon_sym_TILDE] = ACTIONS(2778),
    [anon_sym_DASH] = ACTIONS(2776),
    [anon_sym_PLUS] = ACTIONS(2776),
    [anon_sym_STAR] = ACTIONS(2778),
    [anon_sym_CARET] = ACTIONS(2778),
    [anon_sym_AMP] = ACTIONS(2778),
    [anon_sym_SEMI] = ACTIONS(2778),
    [anon_sym___extension__] = ACTIONS(2776),
    [anon_sym_typedef] = ACTIONS(2776),
    [anon_sym_extern] = ACTIONS(2776),
    [anon_sym___attribute__] = ACTIONS(2776),
    [anon_sym___attribute] = ACTIONS(2776),
    [anon_sym_noreturn] = ACTIONS(2776),
    [anon_sym_LBRACK] = ACTIONS(2778),
    [anon_sym___declspec] = ACTIONS(2776),
    [anon_sym___cdecl] = ACTIONS(2776),
    [anon_sym___clrcall] = ACTIONS(2776),
    [anon_sym___stdcall] = ACTIONS(2776),
    [anon_sym___fastcall] = ACTIONS(2776),
    [anon_sym___thiscall] = ACTIONS(2776),
    [anon_sym___vectorcall] = ACTIONS(2776),
    [anon_sym_LBRACE] = ACTIONS(2778),
    [anon_sym_RBRACE] = ACTIONS(2778),
    [anon_sym_signed] = ACTIONS(2776),
    [anon_sym_unsigned] = ACTIONS(2776),
    [anon_sym_long] = ACTIONS(2776),
    [anon_sym_short] = ACTIONS(2776),
    [anon_sym_static] = ACTIONS(2776),
    [anon_sym_ATautoreleasepool] = ACTIONS(2778),
    [anon_sym_auto] = ACTIONS(2776),
    [anon_sym_register] = ACTIONS(2776),
    [anon_sym_inline] = ACTIONS(2776),
    [anon_sym___inline] = ACTIONS(2776),
    [anon_sym___inline__] = ACTIONS(2776),
    [anon_sym___forceinline] = ACTIONS(2776),
    [anon_sym_thread_local] = ACTIONS(2776),
    [anon_sym___thread] = ACTIONS(2776),
    [anon_sym_CG_EXTERN] = ACTIONS(2776),
    [anon_sym_CG_INLINE] = ACTIONS(2776),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2776),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2776),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2776),
    [anon_sym_IBOutlet] = ACTIONS(2776),
    [anon_sym_IBInspectable] = ACTIONS(2776),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2776),
    [anon_sym_NS_INLINE] = ACTIONS(2776),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2776),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2776),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2776),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2776),
    [anon_sym_const] = ACTIONS(2776),
    [anon_sym_constexpr] = ACTIONS(2776),
    [anon_sym_volatile] = ACTIONS(2776),
    [anon_sym_restrict] = ACTIONS(2776),
    [anon_sym___restrict__] = ACTIONS(2776),
    [anon_sym__Atomic] = ACTIONS(2776),
    [anon_sym__Noreturn] = ACTIONS(2776),
    [anon_sym__Nonnull] = ACTIONS(2776),
    [anon_sym_nullable] = ACTIONS(2776),
    [anon_sym__Complex] = ACTIONS(2776),
    [anon_sym__Nullable] = ACTIONS(2776),
    [anon_sym__Nullable_result] = ACTIONS(2776),
    [anon_sym__Null_unspecified] = ACTIONS(2776),
    [anon_sym___autoreleasing] = ACTIONS(2776),
    [anon_sym___block] = ACTIONS(2776),
    [anon_sym___bridge] = ACTIONS(2776),
    [anon_sym___bridge_retained] = ACTIONS(2776),
    [anon_sym___bridge_transfer] = ACTIONS(2776),
    [anon_sym___complex] = ACTIONS(2776),
    [anon_sym___const] = ACTIONS(2776),
    [anon_sym___imag] = ACTIONS(2776),
    [anon_sym___kindof] = ACTIONS(2776),
    [anon_sym___nonnull] = ACTIONS(2776),
    [anon_sym___nullable] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2776),
    [anon_sym___real] = ACTIONS(2776),
    [anon_sym___strong] = ACTIONS(2776),
    [anon_sym___unsafe_unretained] = ACTIONS(2776),
    [anon_sym___unused] = ACTIONS(2776),
    [anon_sym___weak] = ACTIONS(2776),
    [anon_sym_alignas] = ACTIONS(2776),
    [anon_sym__Alignas] = ACTIONS(2776),
    [sym_primitive_type] = ACTIONS(2776),
    [anon_sym_enum] = ACTIONS(2776),
    [anon_sym_struct] = ACTIONS(2776),
    [anon_sym_union] = ACTIONS(2776),
    [anon_sym_if] = ACTIONS(2776),
    [anon_sym_switch] = ACTIONS(2776),
    [anon_sym_case] = ACTIONS(2776),
    [anon_sym_default] = ACTIONS(2776),
    [anon_sym_while] = ACTIONS(2776),
    [anon_sym_do] = ACTIONS(2776),
    [anon_sym_for] = ACTIONS(2776),
    [anon_sym_in] = ACTIONS(2776),
    [anon_sym_return] = ACTIONS(2776),
    [anon_sym_break] = ACTIONS(2776),
    [anon_sym_continue] = ACTIONS(2776),
    [anon_sym_goto] = ACTIONS(2776),
    [anon_sym_DASH_DASH] = ACTIONS(2778),
    [anon_sym_PLUS_PLUS] = ACTIONS(2778),
    [anon_sym_sizeof] = ACTIONS(2776),
    [anon_sym___alignof__] = ACTIONS(2776),
    [anon_sym___alignof] = ACTIONS(2776),
    [anon_sym__alignof] = ACTIONS(2776),
    [anon_sym_alignof] = ACTIONS(2776),
    [anon_sym__Alignof] = ACTIONS(2776),
    [anon_sym_offsetof] = ACTIONS(2776),
    [anon_sym__Generic] = ACTIONS(2776),
    [anon_sym_asm] = ACTIONS(2776),
    [anon_sym___asm__] = ACTIONS(2776),
    [anon_sym___asm] = ACTIONS(2776),
    [sym_number_literal] = ACTIONS(2778),
    [anon_sym_L_SQUOTE] = ACTIONS(2778),
    [anon_sym_u_SQUOTE] = ACTIONS(2778),
    [anon_sym_U_SQUOTE] = ACTIONS(2778),
    [anon_sym_u8_SQUOTE] = ACTIONS(2778),
    [anon_sym_SQUOTE] = ACTIONS(2778),
    [anon_sym_AT] = ACTIONS(2776),
    [anon_sym_DQUOTE] = ACTIONS(2778),
    [anon_sym_L_DQUOTE] = ACTIONS(2778),
    [anon_sym_u_DQUOTE] = ACTIONS(2778),
    [anon_sym_U_DQUOTE] = ACTIONS(2778),
    [anon_sym_u8_DQUOTE] = ACTIONS(2778),
    [sym_true] = ACTIONS(2776),
    [sym_false] = ACTIONS(2776),
    [anon_sym_NULL] = ACTIONS(2776),
    [anon_sym_nullptr] = ACTIONS(2776),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2776),
    [anon_sym___typeof] = ACTIONS(2776),
    [anon_sym_typeof] = ACTIONS(2776),
    [anon_sym_ATimport] = ACTIONS(2778),
    [aux_sym_preproc_undef_token1] = ACTIONS(2776),
    [anon_sym_POUND] = ACTIONS(2776),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2776),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2776),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_API_AVAILABLE] = ACTIONS(2776),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_API_DEPRECATED] = ACTIONS(2776),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2776),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym___deprecated_msg] = ACTIONS(2776),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2776),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2776),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2776),
    [anon_sym_ATprotocol] = ACTIONS(2778),
    [anon_sym_ATinterface] = ACTIONS(2778),
    [anon_sym_ATimplementation] = ACTIONS(2778),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2778),
    [anon_sym_ATtry] = ACTIONS(2778),
    [anon_sym___try] = ACTIONS(2776),
    [anon_sym_ATthrow] = ACTIONS(2778),
    [anon_sym_ATselector] = ACTIONS(2778),
    [anon_sym_ATavailable] = ACTIONS(2778),
    [anon_sym___builtin_available] = ACTIONS(2776),
    [anon_sym_va_arg] = ACTIONS(2776),
    [anon_sym_ATencode] = ACTIONS(2778),
    [anon_sym_ATsynchronized] = ACTIONS(2778),
    [anon_sym_BOOL] = ACTIONS(2776),
    [anon_sym_IMP] = ACTIONS(2776),
    [anon_sym_SEL] = ACTIONS(2776),
    [anon_sym_Class] = ACTIONS(2776),
    [anon_sym_id] = ACTIONS(2776),
  },
  [STATE(851)] = {
    [sym_identifier] = ACTIONS(2716),
    [aux_sym_preproc_include_token1] = ACTIONS(2716),
    [aux_sym_preproc_include_token2] = ACTIONS(2716),
    [aux_sym_preproc_def_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2716),
    [sym_preproc_directive] = ACTIONS(2716),
    [anon_sym_LPAREN2] = ACTIONS(2718),
    [anon_sym_BANG] = ACTIONS(2718),
    [anon_sym_TILDE] = ACTIONS(2718),
    [anon_sym_DASH] = ACTIONS(2716),
    [anon_sym_PLUS] = ACTIONS(2716),
    [anon_sym_STAR] = ACTIONS(2718),
    [anon_sym_CARET] = ACTIONS(2718),
    [anon_sym_AMP] = ACTIONS(2718),
    [anon_sym_SEMI] = ACTIONS(2718),
    [anon_sym___extension__] = ACTIONS(2716),
    [anon_sym_typedef] = ACTIONS(2716),
    [anon_sym_extern] = ACTIONS(2716),
    [anon_sym___attribute__] = ACTIONS(2716),
    [anon_sym___attribute] = ACTIONS(2716),
    [anon_sym_noreturn] = ACTIONS(2716),
    [anon_sym_LBRACK] = ACTIONS(2718),
    [anon_sym___declspec] = ACTIONS(2716),
    [anon_sym___cdecl] = ACTIONS(2716),
    [anon_sym___clrcall] = ACTIONS(2716),
    [anon_sym___stdcall] = ACTIONS(2716),
    [anon_sym___fastcall] = ACTIONS(2716),
    [anon_sym___thiscall] = ACTIONS(2716),
    [anon_sym___vectorcall] = ACTIONS(2716),
    [anon_sym_LBRACE] = ACTIONS(2718),
    [anon_sym_RBRACE] = ACTIONS(2718),
    [anon_sym_signed] = ACTIONS(2716),
    [anon_sym_unsigned] = ACTIONS(2716),
    [anon_sym_long] = ACTIONS(2716),
    [anon_sym_short] = ACTIONS(2716),
    [anon_sym_static] = ACTIONS(2716),
    [anon_sym_ATautoreleasepool] = ACTIONS(2718),
    [anon_sym_auto] = ACTIONS(2716),
    [anon_sym_register] = ACTIONS(2716),
    [anon_sym_inline] = ACTIONS(2716),
    [anon_sym___inline] = ACTIONS(2716),
    [anon_sym___inline__] = ACTIONS(2716),
    [anon_sym___forceinline] = ACTIONS(2716),
    [anon_sym_thread_local] = ACTIONS(2716),
    [anon_sym___thread] = ACTIONS(2716),
    [anon_sym_CG_EXTERN] = ACTIONS(2716),
    [anon_sym_CG_INLINE] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2716),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2716),
    [anon_sym_IBOutlet] = ACTIONS(2716),
    [anon_sym_IBInspectable] = ACTIONS(2716),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2716),
    [anon_sym_NS_INLINE] = ACTIONS(2716),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2716),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2716),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2716),
    [anon_sym_const] = ACTIONS(2716),
    [anon_sym_constexpr] = ACTIONS(2716),
    [anon_sym_volatile] = ACTIONS(2716),
    [anon_sym_restrict] = ACTIONS(2716),
    [anon_sym___restrict__] = ACTIONS(2716),
    [anon_sym__Atomic] = ACTIONS(2716),
    [anon_sym__Noreturn] = ACTIONS(2716),
    [anon_sym__Nonnull] = ACTIONS(2716),
    [anon_sym_nullable] = ACTIONS(2716),
    [anon_sym__Complex] = ACTIONS(2716),
    [anon_sym__Nullable] = ACTIONS(2716),
    [anon_sym__Nullable_result] = ACTIONS(2716),
    [anon_sym__Null_unspecified] = ACTIONS(2716),
    [anon_sym___autoreleasing] = ACTIONS(2716),
    [anon_sym___block] = ACTIONS(2716),
    [anon_sym___bridge] = ACTIONS(2716),
    [anon_sym___bridge_retained] = ACTIONS(2716),
    [anon_sym___bridge_transfer] = ACTIONS(2716),
    [anon_sym___complex] = ACTIONS(2716),
    [anon_sym___const] = ACTIONS(2716),
    [anon_sym___imag] = ACTIONS(2716),
    [anon_sym___kindof] = ACTIONS(2716),
    [anon_sym___nonnull] = ACTIONS(2716),
    [anon_sym___nullable] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2716),
    [anon_sym___real] = ACTIONS(2716),
    [anon_sym___strong] = ACTIONS(2716),
    [anon_sym___unsafe_unretained] = ACTIONS(2716),
    [anon_sym___unused] = ACTIONS(2716),
    [anon_sym___weak] = ACTIONS(2716),
    [anon_sym_alignas] = ACTIONS(2716),
    [anon_sym__Alignas] = ACTIONS(2716),
    [sym_primitive_type] = ACTIONS(2716),
    [anon_sym_enum] = ACTIONS(2716),
    [anon_sym_struct] = ACTIONS(2716),
    [anon_sym_union] = ACTIONS(2716),
    [anon_sym_if] = ACTIONS(2716),
    [anon_sym_switch] = ACTIONS(2716),
    [anon_sym_case] = ACTIONS(2716),
    [anon_sym_default] = ACTIONS(2716),
    [anon_sym_while] = ACTIONS(2716),
    [anon_sym_do] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2716),
    [anon_sym_in] = ACTIONS(2716),
    [anon_sym_return] = ACTIONS(2716),
    [anon_sym_break] = ACTIONS(2716),
    [anon_sym_continue] = ACTIONS(2716),
    [anon_sym_goto] = ACTIONS(2716),
    [anon_sym_DASH_DASH] = ACTIONS(2718),
    [anon_sym_PLUS_PLUS] = ACTIONS(2718),
    [anon_sym_sizeof] = ACTIONS(2716),
    [anon_sym___alignof__] = ACTIONS(2716),
    [anon_sym___alignof] = ACTIONS(2716),
    [anon_sym__alignof] = ACTIONS(2716),
    [anon_sym_alignof] = ACTIONS(2716),
    [anon_sym__Alignof] = ACTIONS(2716),
    [anon_sym_offsetof] = ACTIONS(2716),
    [anon_sym__Generic] = ACTIONS(2716),
    [anon_sym_asm] = ACTIONS(2716),
    [anon_sym___asm__] = ACTIONS(2716),
    [anon_sym___asm] = ACTIONS(2716),
    [sym_number_literal] = ACTIONS(2718),
    [anon_sym_L_SQUOTE] = ACTIONS(2718),
    [anon_sym_u_SQUOTE] = ACTIONS(2718),
    [anon_sym_U_SQUOTE] = ACTIONS(2718),
    [anon_sym_u8_SQUOTE] = ACTIONS(2718),
    [anon_sym_SQUOTE] = ACTIONS(2718),
    [anon_sym_AT] = ACTIONS(2716),
    [anon_sym_DQUOTE] = ACTIONS(2718),
    [anon_sym_L_DQUOTE] = ACTIONS(2718),
    [anon_sym_u_DQUOTE] = ACTIONS(2718),
    [anon_sym_U_DQUOTE] = ACTIONS(2718),
    [anon_sym_u8_DQUOTE] = ACTIONS(2718),
    [sym_true] = ACTIONS(2716),
    [sym_false] = ACTIONS(2716),
    [anon_sym_NULL] = ACTIONS(2716),
    [anon_sym_nullptr] = ACTIONS(2716),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2716),
    [anon_sym___typeof] = ACTIONS(2716),
    [anon_sym_typeof] = ACTIONS(2716),
    [anon_sym_ATimport] = ACTIONS(2718),
    [aux_sym_preproc_undef_token1] = ACTIONS(2716),
    [anon_sym_POUND] = ACTIONS(2716),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2716),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2716),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2716),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_API_AVAILABLE] = ACTIONS(2716),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_API_DEPRECATED] = ACTIONS(2716),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym___deprecated_msg] = ACTIONS(2716),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2716),
    [anon_sym_ATprotocol] = ACTIONS(2718),
    [anon_sym_ATinterface] = ACTIONS(2718),
    [anon_sym_ATimplementation] = ACTIONS(2718),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2718),
    [anon_sym_ATtry] = ACTIONS(2718),
    [anon_sym___try] = ACTIONS(2716),
    [anon_sym_ATthrow] = ACTIONS(2718),
    [anon_sym_ATselector] = ACTIONS(2718),
    [anon_sym_ATavailable] = ACTIONS(2718),
    [anon_sym___builtin_available] = ACTIONS(2716),
    [anon_sym_va_arg] = ACTIONS(2716),
    [anon_sym_ATencode] = ACTIONS(2718),
    [anon_sym_ATsynchronized] = ACTIONS(2718),
    [anon_sym_BOOL] = ACTIONS(2716),
    [anon_sym_IMP] = ACTIONS(2716),
    [anon_sym_SEL] = ACTIONS(2716),
    [anon_sym_Class] = ACTIONS(2716),
    [anon_sym_id] = ACTIONS(2716),
  },
  [STATE(852)] = {
    [sym_identifier] = ACTIONS(2720),
    [aux_sym_preproc_include_token1] = ACTIONS(2720),
    [aux_sym_preproc_include_token2] = ACTIONS(2720),
    [aux_sym_preproc_def_token1] = ACTIONS(2720),
    [aux_sym_preproc_if_token1] = ACTIONS(2720),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2720),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2720),
    [sym_preproc_directive] = ACTIONS(2720),
    [anon_sym_LPAREN2] = ACTIONS(2722),
    [anon_sym_BANG] = ACTIONS(2722),
    [anon_sym_TILDE] = ACTIONS(2722),
    [anon_sym_DASH] = ACTIONS(2720),
    [anon_sym_PLUS] = ACTIONS(2720),
    [anon_sym_STAR] = ACTIONS(2722),
    [anon_sym_CARET] = ACTIONS(2722),
    [anon_sym_AMP] = ACTIONS(2722),
    [anon_sym_SEMI] = ACTIONS(2722),
    [anon_sym___extension__] = ACTIONS(2720),
    [anon_sym_typedef] = ACTIONS(2720),
    [anon_sym_extern] = ACTIONS(2720),
    [anon_sym___attribute__] = ACTIONS(2720),
    [anon_sym___attribute] = ACTIONS(2720),
    [anon_sym_noreturn] = ACTIONS(2720),
    [anon_sym_LBRACK] = ACTIONS(2722),
    [anon_sym___declspec] = ACTIONS(2720),
    [anon_sym___cdecl] = ACTIONS(2720),
    [anon_sym___clrcall] = ACTIONS(2720),
    [anon_sym___stdcall] = ACTIONS(2720),
    [anon_sym___fastcall] = ACTIONS(2720),
    [anon_sym___thiscall] = ACTIONS(2720),
    [anon_sym___vectorcall] = ACTIONS(2720),
    [anon_sym_LBRACE] = ACTIONS(2722),
    [anon_sym_RBRACE] = ACTIONS(2722),
    [anon_sym_signed] = ACTIONS(2720),
    [anon_sym_unsigned] = ACTIONS(2720),
    [anon_sym_long] = ACTIONS(2720),
    [anon_sym_short] = ACTIONS(2720),
    [anon_sym_static] = ACTIONS(2720),
    [anon_sym_ATautoreleasepool] = ACTIONS(2722),
    [anon_sym_auto] = ACTIONS(2720),
    [anon_sym_register] = ACTIONS(2720),
    [anon_sym_inline] = ACTIONS(2720),
    [anon_sym___inline] = ACTIONS(2720),
    [anon_sym___inline__] = ACTIONS(2720),
    [anon_sym___forceinline] = ACTIONS(2720),
    [anon_sym_thread_local] = ACTIONS(2720),
    [anon_sym___thread] = ACTIONS(2720),
    [anon_sym_CG_EXTERN] = ACTIONS(2720),
    [anon_sym_CG_INLINE] = ACTIONS(2720),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2720),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2720),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2720),
    [anon_sym_IBOutlet] = ACTIONS(2720),
    [anon_sym_IBInspectable] = ACTIONS(2720),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2720),
    [anon_sym_NS_INLINE] = ACTIONS(2720),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2720),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2720),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2720),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2720),
    [anon_sym_const] = ACTIONS(2720),
    [anon_sym_constexpr] = ACTIONS(2720),
    [anon_sym_volatile] = ACTIONS(2720),
    [anon_sym_restrict] = ACTIONS(2720),
    [anon_sym___restrict__] = ACTIONS(2720),
    [anon_sym__Atomic] = ACTIONS(2720),
    [anon_sym__Noreturn] = ACTIONS(2720),
    [anon_sym__Nonnull] = ACTIONS(2720),
    [anon_sym_nullable] = ACTIONS(2720),
    [anon_sym__Complex] = ACTIONS(2720),
    [anon_sym__Nullable] = ACTIONS(2720),
    [anon_sym__Nullable_result] = ACTIONS(2720),
    [anon_sym__Null_unspecified] = ACTIONS(2720),
    [anon_sym___autoreleasing] = ACTIONS(2720),
    [anon_sym___block] = ACTIONS(2720),
    [anon_sym___bridge] = ACTIONS(2720),
    [anon_sym___bridge_retained] = ACTIONS(2720),
    [anon_sym___bridge_transfer] = ACTIONS(2720),
    [anon_sym___complex] = ACTIONS(2720),
    [anon_sym___const] = ACTIONS(2720),
    [anon_sym___imag] = ACTIONS(2720),
    [anon_sym___kindof] = ACTIONS(2720),
    [anon_sym___nonnull] = ACTIONS(2720),
    [anon_sym___nullable] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2720),
    [anon_sym___real] = ACTIONS(2720),
    [anon_sym___strong] = ACTIONS(2720),
    [anon_sym___unsafe_unretained] = ACTIONS(2720),
    [anon_sym___unused] = ACTIONS(2720),
    [anon_sym___weak] = ACTIONS(2720),
    [anon_sym_alignas] = ACTIONS(2720),
    [anon_sym__Alignas] = ACTIONS(2720),
    [sym_primitive_type] = ACTIONS(2720),
    [anon_sym_enum] = ACTIONS(2720),
    [anon_sym_struct] = ACTIONS(2720),
    [anon_sym_union] = ACTIONS(2720),
    [anon_sym_if] = ACTIONS(2720),
    [anon_sym_switch] = ACTIONS(2720),
    [anon_sym_case] = ACTIONS(2720),
    [anon_sym_default] = ACTIONS(2720),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_do] = ACTIONS(2720),
    [anon_sym_for] = ACTIONS(2720),
    [anon_sym_in] = ACTIONS(2720),
    [anon_sym_return] = ACTIONS(2720),
    [anon_sym_break] = ACTIONS(2720),
    [anon_sym_continue] = ACTIONS(2720),
    [anon_sym_goto] = ACTIONS(2720),
    [anon_sym_DASH_DASH] = ACTIONS(2722),
    [anon_sym_PLUS_PLUS] = ACTIONS(2722),
    [anon_sym_sizeof] = ACTIONS(2720),
    [anon_sym___alignof__] = ACTIONS(2720),
    [anon_sym___alignof] = ACTIONS(2720),
    [anon_sym__alignof] = ACTIONS(2720),
    [anon_sym_alignof] = ACTIONS(2720),
    [anon_sym__Alignof] = ACTIONS(2720),
    [anon_sym_offsetof] = ACTIONS(2720),
    [anon_sym__Generic] = ACTIONS(2720),
    [anon_sym_asm] = ACTIONS(2720),
    [anon_sym___asm__] = ACTIONS(2720),
    [anon_sym___asm] = ACTIONS(2720),
    [sym_number_literal] = ACTIONS(2722),
    [anon_sym_L_SQUOTE] = ACTIONS(2722),
    [anon_sym_u_SQUOTE] = ACTIONS(2722),
    [anon_sym_U_SQUOTE] = ACTIONS(2722),
    [anon_sym_u8_SQUOTE] = ACTIONS(2722),
    [anon_sym_SQUOTE] = ACTIONS(2722),
    [anon_sym_AT] = ACTIONS(2720),
    [anon_sym_DQUOTE] = ACTIONS(2722),
    [anon_sym_L_DQUOTE] = ACTIONS(2722),
    [anon_sym_u_DQUOTE] = ACTIONS(2722),
    [anon_sym_U_DQUOTE] = ACTIONS(2722),
    [anon_sym_u8_DQUOTE] = ACTIONS(2722),
    [sym_true] = ACTIONS(2720),
    [sym_false] = ACTIONS(2720),
    [anon_sym_NULL] = ACTIONS(2720),
    [anon_sym_nullptr] = ACTIONS(2720),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2720),
    [anon_sym___typeof] = ACTIONS(2720),
    [anon_sym_typeof] = ACTIONS(2720),
    [anon_sym_ATimport] = ACTIONS(2722),
    [aux_sym_preproc_undef_token1] = ACTIONS(2720),
    [anon_sym_POUND] = ACTIONS(2720),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2720),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2720),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2720),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2720),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2720),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2720),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2720),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_API_AVAILABLE] = ACTIONS(2720),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_API_DEPRECATED] = ACTIONS(2720),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2720),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym___deprecated_msg] = ACTIONS(2720),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2720),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2720),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2720),
    [anon_sym_ATprotocol] = ACTIONS(2722),
    [anon_sym_ATinterface] = ACTIONS(2722),
    [anon_sym_ATimplementation] = ACTIONS(2722),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2722),
    [anon_sym_ATtry] = ACTIONS(2722),
    [anon_sym___try] = ACTIONS(2720),
    [anon_sym_ATthrow] = ACTIONS(2722),
    [anon_sym_ATselector] = ACTIONS(2722),
    [anon_sym_ATavailable] = ACTIONS(2722),
    [anon_sym___builtin_available] = ACTIONS(2720),
    [anon_sym_va_arg] = ACTIONS(2720),
    [anon_sym_ATencode] = ACTIONS(2722),
    [anon_sym_ATsynchronized] = ACTIONS(2722),
    [anon_sym_BOOL] = ACTIONS(2720),
    [anon_sym_IMP] = ACTIONS(2720),
    [anon_sym_SEL] = ACTIONS(2720),
    [anon_sym_Class] = ACTIONS(2720),
    [anon_sym_id] = ACTIONS(2720),
  },
  [STATE(853)] = {
    [sym_identifier] = ACTIONS(2724),
    [aux_sym_preproc_include_token1] = ACTIONS(2724),
    [aux_sym_preproc_include_token2] = ACTIONS(2724),
    [aux_sym_preproc_def_token1] = ACTIONS(2724),
    [aux_sym_preproc_if_token1] = ACTIONS(2724),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2724),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2724),
    [sym_preproc_directive] = ACTIONS(2724),
    [anon_sym_LPAREN2] = ACTIONS(2726),
    [anon_sym_BANG] = ACTIONS(2726),
    [anon_sym_TILDE] = ACTIONS(2726),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2726),
    [anon_sym_CARET] = ACTIONS(2726),
    [anon_sym_AMP] = ACTIONS(2726),
    [anon_sym_SEMI] = ACTIONS(2726),
    [anon_sym___extension__] = ACTIONS(2724),
    [anon_sym_typedef] = ACTIONS(2724),
    [anon_sym_extern] = ACTIONS(2724),
    [anon_sym___attribute__] = ACTIONS(2724),
    [anon_sym___attribute] = ACTIONS(2724),
    [anon_sym_noreturn] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2726),
    [anon_sym___declspec] = ACTIONS(2724),
    [anon_sym___cdecl] = ACTIONS(2724),
    [anon_sym___clrcall] = ACTIONS(2724),
    [anon_sym___stdcall] = ACTIONS(2724),
    [anon_sym___fastcall] = ACTIONS(2724),
    [anon_sym___thiscall] = ACTIONS(2724),
    [anon_sym___vectorcall] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2726),
    [anon_sym_RBRACE] = ACTIONS(2726),
    [anon_sym_signed] = ACTIONS(2724),
    [anon_sym_unsigned] = ACTIONS(2724),
    [anon_sym_long] = ACTIONS(2724),
    [anon_sym_short] = ACTIONS(2724),
    [anon_sym_static] = ACTIONS(2724),
    [anon_sym_ATautoreleasepool] = ACTIONS(2726),
    [anon_sym_auto] = ACTIONS(2724),
    [anon_sym_register] = ACTIONS(2724),
    [anon_sym_inline] = ACTIONS(2724),
    [anon_sym___inline] = ACTIONS(2724),
    [anon_sym___inline__] = ACTIONS(2724),
    [anon_sym___forceinline] = ACTIONS(2724),
    [anon_sym_thread_local] = ACTIONS(2724),
    [anon_sym___thread] = ACTIONS(2724),
    [anon_sym_CG_EXTERN] = ACTIONS(2724),
    [anon_sym_CG_INLINE] = ACTIONS(2724),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2724),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2724),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2724),
    [anon_sym_IBOutlet] = ACTIONS(2724),
    [anon_sym_IBInspectable] = ACTIONS(2724),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2724),
    [anon_sym_NS_INLINE] = ACTIONS(2724),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2724),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2724),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2724),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2724),
    [anon_sym_const] = ACTIONS(2724),
    [anon_sym_constexpr] = ACTIONS(2724),
    [anon_sym_volatile] = ACTIONS(2724),
    [anon_sym_restrict] = ACTIONS(2724),
    [anon_sym___restrict__] = ACTIONS(2724),
    [anon_sym__Atomic] = ACTIONS(2724),
    [anon_sym__Noreturn] = ACTIONS(2724),
    [anon_sym__Nonnull] = ACTIONS(2724),
    [anon_sym_nullable] = ACTIONS(2724),
    [anon_sym__Complex] = ACTIONS(2724),
    [anon_sym__Nullable] = ACTIONS(2724),
    [anon_sym__Nullable_result] = ACTIONS(2724),
    [anon_sym__Null_unspecified] = ACTIONS(2724),
    [anon_sym___autoreleasing] = ACTIONS(2724),
    [anon_sym___block] = ACTIONS(2724),
    [anon_sym___bridge] = ACTIONS(2724),
    [anon_sym___bridge_retained] = ACTIONS(2724),
    [anon_sym___bridge_transfer] = ACTIONS(2724),
    [anon_sym___complex] = ACTIONS(2724),
    [anon_sym___const] = ACTIONS(2724),
    [anon_sym___imag] = ACTIONS(2724),
    [anon_sym___kindof] = ACTIONS(2724),
    [anon_sym___nonnull] = ACTIONS(2724),
    [anon_sym___nullable] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2724),
    [anon_sym___real] = ACTIONS(2724),
    [anon_sym___strong] = ACTIONS(2724),
    [anon_sym___unsafe_unretained] = ACTIONS(2724),
    [anon_sym___unused] = ACTIONS(2724),
    [anon_sym___weak] = ACTIONS(2724),
    [anon_sym_alignas] = ACTIONS(2724),
    [anon_sym__Alignas] = ACTIONS(2724),
    [sym_primitive_type] = ACTIONS(2724),
    [anon_sym_enum] = ACTIONS(2724),
    [anon_sym_struct] = ACTIONS(2724),
    [anon_sym_union] = ACTIONS(2724),
    [anon_sym_if] = ACTIONS(2724),
    [anon_sym_switch] = ACTIONS(2724),
    [anon_sym_case] = ACTIONS(2724),
    [anon_sym_default] = ACTIONS(2724),
    [anon_sym_while] = ACTIONS(2724),
    [anon_sym_do] = ACTIONS(2724),
    [anon_sym_for] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_return] = ACTIONS(2724),
    [anon_sym_break] = ACTIONS(2724),
    [anon_sym_continue] = ACTIONS(2724),
    [anon_sym_goto] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2726),
    [anon_sym_PLUS_PLUS] = ACTIONS(2726),
    [anon_sym_sizeof] = ACTIONS(2724),
    [anon_sym___alignof__] = ACTIONS(2724),
    [anon_sym___alignof] = ACTIONS(2724),
    [anon_sym__alignof] = ACTIONS(2724),
    [anon_sym_alignof] = ACTIONS(2724),
    [anon_sym__Alignof] = ACTIONS(2724),
    [anon_sym_offsetof] = ACTIONS(2724),
    [anon_sym__Generic] = ACTIONS(2724),
    [anon_sym_asm] = ACTIONS(2724),
    [anon_sym___asm__] = ACTIONS(2724),
    [anon_sym___asm] = ACTIONS(2724),
    [sym_number_literal] = ACTIONS(2726),
    [anon_sym_L_SQUOTE] = ACTIONS(2726),
    [anon_sym_u_SQUOTE] = ACTIONS(2726),
    [anon_sym_U_SQUOTE] = ACTIONS(2726),
    [anon_sym_u8_SQUOTE] = ACTIONS(2726),
    [anon_sym_SQUOTE] = ACTIONS(2726),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2726),
    [anon_sym_L_DQUOTE] = ACTIONS(2726),
    [anon_sym_u_DQUOTE] = ACTIONS(2726),
    [anon_sym_U_DQUOTE] = ACTIONS(2726),
    [anon_sym_u8_DQUOTE] = ACTIONS(2726),
    [sym_true] = ACTIONS(2724),
    [sym_false] = ACTIONS(2724),
    [anon_sym_NULL] = ACTIONS(2724),
    [anon_sym_nullptr] = ACTIONS(2724),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2724),
    [anon_sym___typeof] = ACTIONS(2724),
    [anon_sym_typeof] = ACTIONS(2724),
    [anon_sym_ATimport] = ACTIONS(2726),
    [aux_sym_preproc_undef_token1] = ACTIONS(2724),
    [anon_sym_POUND] = ACTIONS(2724),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2724),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2724),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2724),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2724),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2724),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2724),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2724),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_API_AVAILABLE] = ACTIONS(2724),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_API_DEPRECATED] = ACTIONS(2724),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2724),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym___deprecated_msg] = ACTIONS(2724),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2724),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2724),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2724),
    [anon_sym_ATprotocol] = ACTIONS(2726),
    [anon_sym_ATinterface] = ACTIONS(2726),
    [anon_sym_ATimplementation] = ACTIONS(2726),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2726),
    [anon_sym_ATtry] = ACTIONS(2726),
    [anon_sym___try] = ACTIONS(2724),
    [anon_sym_ATthrow] = ACTIONS(2726),
    [anon_sym_ATselector] = ACTIONS(2726),
    [anon_sym_ATavailable] = ACTIONS(2726),
    [anon_sym___builtin_available] = ACTIONS(2724),
    [anon_sym_va_arg] = ACTIONS(2724),
    [anon_sym_ATencode] = ACTIONS(2726),
    [anon_sym_ATsynchronized] = ACTIONS(2726),
    [anon_sym_BOOL] = ACTIONS(2724),
    [anon_sym_IMP] = ACTIONS(2724),
    [anon_sym_SEL] = ACTIONS(2724),
    [anon_sym_Class] = ACTIONS(2724),
    [anon_sym_id] = ACTIONS(2724),
  },
  [STATE(854)] = {
    [sym_identifier] = ACTIONS(2728),
    [aux_sym_preproc_include_token1] = ACTIONS(2728),
    [aux_sym_preproc_include_token2] = ACTIONS(2728),
    [aux_sym_preproc_def_token1] = ACTIONS(2728),
    [aux_sym_preproc_if_token1] = ACTIONS(2728),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2728),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2728),
    [sym_preproc_directive] = ACTIONS(2728),
    [anon_sym_LPAREN2] = ACTIONS(2730),
    [anon_sym_BANG] = ACTIONS(2730),
    [anon_sym_TILDE] = ACTIONS(2730),
    [anon_sym_DASH] = ACTIONS(2728),
    [anon_sym_PLUS] = ACTIONS(2728),
    [anon_sym_STAR] = ACTIONS(2730),
    [anon_sym_CARET] = ACTIONS(2730),
    [anon_sym_AMP] = ACTIONS(2730),
    [anon_sym_SEMI] = ACTIONS(2730),
    [anon_sym___extension__] = ACTIONS(2728),
    [anon_sym_typedef] = ACTIONS(2728),
    [anon_sym_extern] = ACTIONS(2728),
    [anon_sym___attribute__] = ACTIONS(2728),
    [anon_sym___attribute] = ACTIONS(2728),
    [anon_sym_noreturn] = ACTIONS(2728),
    [anon_sym_LBRACK] = ACTIONS(2730),
    [anon_sym___declspec] = ACTIONS(2728),
    [anon_sym___cdecl] = ACTIONS(2728),
    [anon_sym___clrcall] = ACTIONS(2728),
    [anon_sym___stdcall] = ACTIONS(2728),
    [anon_sym___fastcall] = ACTIONS(2728),
    [anon_sym___thiscall] = ACTIONS(2728),
    [anon_sym___vectorcall] = ACTIONS(2728),
    [anon_sym_LBRACE] = ACTIONS(2730),
    [anon_sym_RBRACE] = ACTIONS(2730),
    [anon_sym_signed] = ACTIONS(2728),
    [anon_sym_unsigned] = ACTIONS(2728),
    [anon_sym_long] = ACTIONS(2728),
    [anon_sym_short] = ACTIONS(2728),
    [anon_sym_static] = ACTIONS(2728),
    [anon_sym_ATautoreleasepool] = ACTIONS(2730),
    [anon_sym_auto] = ACTIONS(2728),
    [anon_sym_register] = ACTIONS(2728),
    [anon_sym_inline] = ACTIONS(2728),
    [anon_sym___inline] = ACTIONS(2728),
    [anon_sym___inline__] = ACTIONS(2728),
    [anon_sym___forceinline] = ACTIONS(2728),
    [anon_sym_thread_local] = ACTIONS(2728),
    [anon_sym___thread] = ACTIONS(2728),
    [anon_sym_CG_EXTERN] = ACTIONS(2728),
    [anon_sym_CG_INLINE] = ACTIONS(2728),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2728),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2728),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2728),
    [anon_sym_IBOutlet] = ACTIONS(2728),
    [anon_sym_IBInspectable] = ACTIONS(2728),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2728),
    [anon_sym_NS_INLINE] = ACTIONS(2728),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2728),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2728),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2728),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2728),
    [anon_sym_const] = ACTIONS(2728),
    [anon_sym_constexpr] = ACTIONS(2728),
    [anon_sym_volatile] = ACTIONS(2728),
    [anon_sym_restrict] = ACTIONS(2728),
    [anon_sym___restrict__] = ACTIONS(2728),
    [anon_sym__Atomic] = ACTIONS(2728),
    [anon_sym__Noreturn] = ACTIONS(2728),
    [anon_sym__Nonnull] = ACTIONS(2728),
    [anon_sym_nullable] = ACTIONS(2728),
    [anon_sym__Complex] = ACTIONS(2728),
    [anon_sym__Nullable] = ACTIONS(2728),
    [anon_sym__Nullable_result] = ACTIONS(2728),
    [anon_sym__Null_unspecified] = ACTIONS(2728),
    [anon_sym___autoreleasing] = ACTIONS(2728),
    [anon_sym___block] = ACTIONS(2728),
    [anon_sym___bridge] = ACTIONS(2728),
    [anon_sym___bridge_retained] = ACTIONS(2728),
    [anon_sym___bridge_transfer] = ACTIONS(2728),
    [anon_sym___complex] = ACTIONS(2728),
    [anon_sym___const] = ACTIONS(2728),
    [anon_sym___imag] = ACTIONS(2728),
    [anon_sym___kindof] = ACTIONS(2728),
    [anon_sym___nonnull] = ACTIONS(2728),
    [anon_sym___nullable] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2728),
    [anon_sym___real] = ACTIONS(2728),
    [anon_sym___strong] = ACTIONS(2728),
    [anon_sym___unsafe_unretained] = ACTIONS(2728),
    [anon_sym___unused] = ACTIONS(2728),
    [anon_sym___weak] = ACTIONS(2728),
    [anon_sym_alignas] = ACTIONS(2728),
    [anon_sym__Alignas] = ACTIONS(2728),
    [sym_primitive_type] = ACTIONS(2728),
    [anon_sym_enum] = ACTIONS(2728),
    [anon_sym_struct] = ACTIONS(2728),
    [anon_sym_union] = ACTIONS(2728),
    [anon_sym_if] = ACTIONS(2728),
    [anon_sym_switch] = ACTIONS(2728),
    [anon_sym_case] = ACTIONS(2728),
    [anon_sym_default] = ACTIONS(2728),
    [anon_sym_while] = ACTIONS(2728),
    [anon_sym_do] = ACTIONS(2728),
    [anon_sym_for] = ACTIONS(2728),
    [anon_sym_in] = ACTIONS(2728),
    [anon_sym_return] = ACTIONS(2728),
    [anon_sym_break] = ACTIONS(2728),
    [anon_sym_continue] = ACTIONS(2728),
    [anon_sym_goto] = ACTIONS(2728),
    [anon_sym_DASH_DASH] = ACTIONS(2730),
    [anon_sym_PLUS_PLUS] = ACTIONS(2730),
    [anon_sym_sizeof] = ACTIONS(2728),
    [anon_sym___alignof__] = ACTIONS(2728),
    [anon_sym___alignof] = ACTIONS(2728),
    [anon_sym__alignof] = ACTIONS(2728),
    [anon_sym_alignof] = ACTIONS(2728),
    [anon_sym__Alignof] = ACTIONS(2728),
    [anon_sym_offsetof] = ACTIONS(2728),
    [anon_sym__Generic] = ACTIONS(2728),
    [anon_sym_asm] = ACTIONS(2728),
    [anon_sym___asm__] = ACTIONS(2728),
    [anon_sym___asm] = ACTIONS(2728),
    [sym_number_literal] = ACTIONS(2730),
    [anon_sym_L_SQUOTE] = ACTIONS(2730),
    [anon_sym_u_SQUOTE] = ACTIONS(2730),
    [anon_sym_U_SQUOTE] = ACTIONS(2730),
    [anon_sym_u8_SQUOTE] = ACTIONS(2730),
    [anon_sym_SQUOTE] = ACTIONS(2730),
    [anon_sym_AT] = ACTIONS(2728),
    [anon_sym_DQUOTE] = ACTIONS(2730),
    [anon_sym_L_DQUOTE] = ACTIONS(2730),
    [anon_sym_u_DQUOTE] = ACTIONS(2730),
    [anon_sym_U_DQUOTE] = ACTIONS(2730),
    [anon_sym_u8_DQUOTE] = ACTIONS(2730),
    [sym_true] = ACTIONS(2728),
    [sym_false] = ACTIONS(2728),
    [anon_sym_NULL] = ACTIONS(2728),
    [anon_sym_nullptr] = ACTIONS(2728),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2728),
    [anon_sym___typeof] = ACTIONS(2728),
    [anon_sym_typeof] = ACTIONS(2728),
    [anon_sym_ATimport] = ACTIONS(2730),
    [aux_sym_preproc_undef_token1] = ACTIONS(2728),
    [anon_sym_POUND] = ACTIONS(2728),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2728),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2728),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2728),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2728),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2728),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2728),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2728),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_API_AVAILABLE] = ACTIONS(2728),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_API_DEPRECATED] = ACTIONS(2728),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2728),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym___deprecated_msg] = ACTIONS(2728),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2728),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2728),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2728),
    [anon_sym_ATprotocol] = ACTIONS(2730),
    [anon_sym_ATinterface] = ACTIONS(2730),
    [anon_sym_ATimplementation] = ACTIONS(2730),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2730),
    [anon_sym_ATtry] = ACTIONS(2730),
    [anon_sym___try] = ACTIONS(2728),
    [anon_sym_ATthrow] = ACTIONS(2730),
    [anon_sym_ATselector] = ACTIONS(2730),
    [anon_sym_ATavailable] = ACTIONS(2730),
    [anon_sym___builtin_available] = ACTIONS(2728),
    [anon_sym_va_arg] = ACTIONS(2728),
    [anon_sym_ATencode] = ACTIONS(2730),
    [anon_sym_ATsynchronized] = ACTIONS(2730),
    [anon_sym_BOOL] = ACTIONS(2728),
    [anon_sym_IMP] = ACTIONS(2728),
    [anon_sym_SEL] = ACTIONS(2728),
    [anon_sym_Class] = ACTIONS(2728),
    [anon_sym_id] = ACTIONS(2728),
  },
  [STATE(855)] = {
    [sym_identifier] = ACTIONS(2732),
    [aux_sym_preproc_include_token1] = ACTIONS(2732),
    [aux_sym_preproc_include_token2] = ACTIONS(2732),
    [aux_sym_preproc_def_token1] = ACTIONS(2732),
    [aux_sym_preproc_if_token1] = ACTIONS(2732),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2732),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2732),
    [sym_preproc_directive] = ACTIONS(2732),
    [anon_sym_LPAREN2] = ACTIONS(2734),
    [anon_sym_BANG] = ACTIONS(2734),
    [anon_sym_TILDE] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2732),
    [anon_sym_PLUS] = ACTIONS(2732),
    [anon_sym_STAR] = ACTIONS(2734),
    [anon_sym_CARET] = ACTIONS(2734),
    [anon_sym_AMP] = ACTIONS(2734),
    [anon_sym_SEMI] = ACTIONS(2734),
    [anon_sym___extension__] = ACTIONS(2732),
    [anon_sym_typedef] = ACTIONS(2732),
    [anon_sym_extern] = ACTIONS(2732),
    [anon_sym___attribute__] = ACTIONS(2732),
    [anon_sym___attribute] = ACTIONS(2732),
    [anon_sym_noreturn] = ACTIONS(2732),
    [anon_sym_LBRACK] = ACTIONS(2734),
    [anon_sym___declspec] = ACTIONS(2732),
    [anon_sym___cdecl] = ACTIONS(2732),
    [anon_sym___clrcall] = ACTIONS(2732),
    [anon_sym___stdcall] = ACTIONS(2732),
    [anon_sym___fastcall] = ACTIONS(2732),
    [anon_sym___thiscall] = ACTIONS(2732),
    [anon_sym___vectorcall] = ACTIONS(2732),
    [anon_sym_LBRACE] = ACTIONS(2734),
    [anon_sym_RBRACE] = ACTIONS(2734),
    [anon_sym_signed] = ACTIONS(2732),
    [anon_sym_unsigned] = ACTIONS(2732),
    [anon_sym_long] = ACTIONS(2732),
    [anon_sym_short] = ACTIONS(2732),
    [anon_sym_static] = ACTIONS(2732),
    [anon_sym_ATautoreleasepool] = ACTIONS(2734),
    [anon_sym_auto] = ACTIONS(2732),
    [anon_sym_register] = ACTIONS(2732),
    [anon_sym_inline] = ACTIONS(2732),
    [anon_sym___inline] = ACTIONS(2732),
    [anon_sym___inline__] = ACTIONS(2732),
    [anon_sym___forceinline] = ACTIONS(2732),
    [anon_sym_thread_local] = ACTIONS(2732),
    [anon_sym___thread] = ACTIONS(2732),
    [anon_sym_CG_EXTERN] = ACTIONS(2732),
    [anon_sym_CG_INLINE] = ACTIONS(2732),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2732),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2732),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2732),
    [anon_sym_IBOutlet] = ACTIONS(2732),
    [anon_sym_IBInspectable] = ACTIONS(2732),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2732),
    [anon_sym_NS_INLINE] = ACTIONS(2732),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2732),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2732),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2732),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2732),
    [anon_sym_const] = ACTIONS(2732),
    [anon_sym_constexpr] = ACTIONS(2732),
    [anon_sym_volatile] = ACTIONS(2732),
    [anon_sym_restrict] = ACTIONS(2732),
    [anon_sym___restrict__] = ACTIONS(2732),
    [anon_sym__Atomic] = ACTIONS(2732),
    [anon_sym__Noreturn] = ACTIONS(2732),
    [anon_sym__Nonnull] = ACTIONS(2732),
    [anon_sym_nullable] = ACTIONS(2732),
    [anon_sym__Complex] = ACTIONS(2732),
    [anon_sym__Nullable] = ACTIONS(2732),
    [anon_sym__Nullable_result] = ACTIONS(2732),
    [anon_sym__Null_unspecified] = ACTIONS(2732),
    [anon_sym___autoreleasing] = ACTIONS(2732),
    [anon_sym___block] = ACTIONS(2732),
    [anon_sym___bridge] = ACTIONS(2732),
    [anon_sym___bridge_retained] = ACTIONS(2732),
    [anon_sym___bridge_transfer] = ACTIONS(2732),
    [anon_sym___complex] = ACTIONS(2732),
    [anon_sym___const] = ACTIONS(2732),
    [anon_sym___imag] = ACTIONS(2732),
    [anon_sym___kindof] = ACTIONS(2732),
    [anon_sym___nonnull] = ACTIONS(2732),
    [anon_sym___nullable] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2732),
    [anon_sym___real] = ACTIONS(2732),
    [anon_sym___strong] = ACTIONS(2732),
    [anon_sym___unsafe_unretained] = ACTIONS(2732),
    [anon_sym___unused] = ACTIONS(2732),
    [anon_sym___weak] = ACTIONS(2732),
    [anon_sym_alignas] = ACTIONS(2732),
    [anon_sym__Alignas] = ACTIONS(2732),
    [sym_primitive_type] = ACTIONS(2732),
    [anon_sym_enum] = ACTIONS(2732),
    [anon_sym_struct] = ACTIONS(2732),
    [anon_sym_union] = ACTIONS(2732),
    [anon_sym_if] = ACTIONS(2732),
    [anon_sym_switch] = ACTIONS(2732),
    [anon_sym_case] = ACTIONS(2732),
    [anon_sym_default] = ACTIONS(2732),
    [anon_sym_while] = ACTIONS(2732),
    [anon_sym_do] = ACTIONS(2732),
    [anon_sym_for] = ACTIONS(2732),
    [anon_sym_in] = ACTIONS(2732),
    [anon_sym_return] = ACTIONS(2732),
    [anon_sym_break] = ACTIONS(2732),
    [anon_sym_continue] = ACTIONS(2732),
    [anon_sym_goto] = ACTIONS(2732),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_sizeof] = ACTIONS(2732),
    [anon_sym___alignof__] = ACTIONS(2732),
    [anon_sym___alignof] = ACTIONS(2732),
    [anon_sym__alignof] = ACTIONS(2732),
    [anon_sym_alignof] = ACTIONS(2732),
    [anon_sym__Alignof] = ACTIONS(2732),
    [anon_sym_offsetof] = ACTIONS(2732),
    [anon_sym__Generic] = ACTIONS(2732),
    [anon_sym_asm] = ACTIONS(2732),
    [anon_sym___asm__] = ACTIONS(2732),
    [anon_sym___asm] = ACTIONS(2732),
    [sym_number_literal] = ACTIONS(2734),
    [anon_sym_L_SQUOTE] = ACTIONS(2734),
    [anon_sym_u_SQUOTE] = ACTIONS(2734),
    [anon_sym_U_SQUOTE] = ACTIONS(2734),
    [anon_sym_u8_SQUOTE] = ACTIONS(2734),
    [anon_sym_SQUOTE] = ACTIONS(2734),
    [anon_sym_AT] = ACTIONS(2732),
    [anon_sym_DQUOTE] = ACTIONS(2734),
    [anon_sym_L_DQUOTE] = ACTIONS(2734),
    [anon_sym_u_DQUOTE] = ACTIONS(2734),
    [anon_sym_U_DQUOTE] = ACTIONS(2734),
    [anon_sym_u8_DQUOTE] = ACTIONS(2734),
    [sym_true] = ACTIONS(2732),
    [sym_false] = ACTIONS(2732),
    [anon_sym_NULL] = ACTIONS(2732),
    [anon_sym_nullptr] = ACTIONS(2732),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2732),
    [anon_sym___typeof] = ACTIONS(2732),
    [anon_sym_typeof] = ACTIONS(2732),
    [anon_sym_ATimport] = ACTIONS(2734),
    [aux_sym_preproc_undef_token1] = ACTIONS(2732),
    [anon_sym_POUND] = ACTIONS(2732),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2732),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2732),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2732),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2732),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2732),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2732),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2732),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_API_AVAILABLE] = ACTIONS(2732),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_API_DEPRECATED] = ACTIONS(2732),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2732),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym___deprecated_msg] = ACTIONS(2732),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2732),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2732),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2732),
    [anon_sym_ATprotocol] = ACTIONS(2734),
    [anon_sym_ATinterface] = ACTIONS(2734),
    [anon_sym_ATimplementation] = ACTIONS(2734),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2734),
    [anon_sym_ATtry] = ACTIONS(2734),
    [anon_sym___try] = ACTIONS(2732),
    [anon_sym_ATthrow] = ACTIONS(2734),
    [anon_sym_ATselector] = ACTIONS(2734),
    [anon_sym_ATavailable] = ACTIONS(2734),
    [anon_sym___builtin_available] = ACTIONS(2732),
    [anon_sym_va_arg] = ACTIONS(2732),
    [anon_sym_ATencode] = ACTIONS(2734),
    [anon_sym_ATsynchronized] = ACTIONS(2734),
    [anon_sym_BOOL] = ACTIONS(2732),
    [anon_sym_IMP] = ACTIONS(2732),
    [anon_sym_SEL] = ACTIONS(2732),
    [anon_sym_Class] = ACTIONS(2732),
    [anon_sym_id] = ACTIONS(2732),
  },
  [STATE(856)] = {
    [sym_identifier] = ACTIONS(2736),
    [aux_sym_preproc_include_token1] = ACTIONS(2736),
    [aux_sym_preproc_include_token2] = ACTIONS(2736),
    [aux_sym_preproc_def_token1] = ACTIONS(2736),
    [aux_sym_preproc_if_token1] = ACTIONS(2736),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2736),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2736),
    [sym_preproc_directive] = ACTIONS(2736),
    [anon_sym_LPAREN2] = ACTIONS(2738),
    [anon_sym_BANG] = ACTIONS(2738),
    [anon_sym_TILDE] = ACTIONS(2738),
    [anon_sym_DASH] = ACTIONS(2736),
    [anon_sym_PLUS] = ACTIONS(2736),
    [anon_sym_STAR] = ACTIONS(2738),
    [anon_sym_CARET] = ACTIONS(2738),
    [anon_sym_AMP] = ACTIONS(2738),
    [anon_sym_SEMI] = ACTIONS(2738),
    [anon_sym___extension__] = ACTIONS(2736),
    [anon_sym_typedef] = ACTIONS(2736),
    [anon_sym_extern] = ACTIONS(2736),
    [anon_sym___attribute__] = ACTIONS(2736),
    [anon_sym___attribute] = ACTIONS(2736),
    [anon_sym_noreturn] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(2738),
    [anon_sym___declspec] = ACTIONS(2736),
    [anon_sym___cdecl] = ACTIONS(2736),
    [anon_sym___clrcall] = ACTIONS(2736),
    [anon_sym___stdcall] = ACTIONS(2736),
    [anon_sym___fastcall] = ACTIONS(2736),
    [anon_sym___thiscall] = ACTIONS(2736),
    [anon_sym___vectorcall] = ACTIONS(2736),
    [anon_sym_LBRACE] = ACTIONS(2738),
    [anon_sym_RBRACE] = ACTIONS(2738),
    [anon_sym_signed] = ACTIONS(2736),
    [anon_sym_unsigned] = ACTIONS(2736),
    [anon_sym_long] = ACTIONS(2736),
    [anon_sym_short] = ACTIONS(2736),
    [anon_sym_static] = ACTIONS(2736),
    [anon_sym_ATautoreleasepool] = ACTIONS(2738),
    [anon_sym_auto] = ACTIONS(2736),
    [anon_sym_register] = ACTIONS(2736),
    [anon_sym_inline] = ACTIONS(2736),
    [anon_sym___inline] = ACTIONS(2736),
    [anon_sym___inline__] = ACTIONS(2736),
    [anon_sym___forceinline] = ACTIONS(2736),
    [anon_sym_thread_local] = ACTIONS(2736),
    [anon_sym___thread] = ACTIONS(2736),
    [anon_sym_CG_EXTERN] = ACTIONS(2736),
    [anon_sym_CG_INLINE] = ACTIONS(2736),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2736),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2736),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2736),
    [anon_sym_IBOutlet] = ACTIONS(2736),
    [anon_sym_IBInspectable] = ACTIONS(2736),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2736),
    [anon_sym_NS_INLINE] = ACTIONS(2736),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2736),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2736),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2736),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2736),
    [anon_sym_const] = ACTIONS(2736),
    [anon_sym_constexpr] = ACTIONS(2736),
    [anon_sym_volatile] = ACTIONS(2736),
    [anon_sym_restrict] = ACTIONS(2736),
    [anon_sym___restrict__] = ACTIONS(2736),
    [anon_sym__Atomic] = ACTIONS(2736),
    [anon_sym__Noreturn] = ACTIONS(2736),
    [anon_sym__Nonnull] = ACTIONS(2736),
    [anon_sym_nullable] = ACTIONS(2736),
    [anon_sym__Complex] = ACTIONS(2736),
    [anon_sym__Nullable] = ACTIONS(2736),
    [anon_sym__Nullable_result] = ACTIONS(2736),
    [anon_sym__Null_unspecified] = ACTIONS(2736),
    [anon_sym___autoreleasing] = ACTIONS(2736),
    [anon_sym___block] = ACTIONS(2736),
    [anon_sym___bridge] = ACTIONS(2736),
    [anon_sym___bridge_retained] = ACTIONS(2736),
    [anon_sym___bridge_transfer] = ACTIONS(2736),
    [anon_sym___complex] = ACTIONS(2736),
    [anon_sym___const] = ACTIONS(2736),
    [anon_sym___imag] = ACTIONS(2736),
    [anon_sym___kindof] = ACTIONS(2736),
    [anon_sym___nonnull] = ACTIONS(2736),
    [anon_sym___nullable] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2736),
    [anon_sym___real] = ACTIONS(2736),
    [anon_sym___strong] = ACTIONS(2736),
    [anon_sym___unsafe_unretained] = ACTIONS(2736),
    [anon_sym___unused] = ACTIONS(2736),
    [anon_sym___weak] = ACTIONS(2736),
    [anon_sym_alignas] = ACTIONS(2736),
    [anon_sym__Alignas] = ACTIONS(2736),
    [sym_primitive_type] = ACTIONS(2736),
    [anon_sym_enum] = ACTIONS(2736),
    [anon_sym_struct] = ACTIONS(2736),
    [anon_sym_union] = ACTIONS(2736),
    [anon_sym_if] = ACTIONS(2736),
    [anon_sym_switch] = ACTIONS(2736),
    [anon_sym_case] = ACTIONS(2736),
    [anon_sym_default] = ACTIONS(2736),
    [anon_sym_while] = ACTIONS(2736),
    [anon_sym_do] = ACTIONS(2736),
    [anon_sym_for] = ACTIONS(2736),
    [anon_sym_in] = ACTIONS(2736),
    [anon_sym_return] = ACTIONS(2736),
    [anon_sym_break] = ACTIONS(2736),
    [anon_sym_continue] = ACTIONS(2736),
    [anon_sym_goto] = ACTIONS(2736),
    [anon_sym_DASH_DASH] = ACTIONS(2738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2738),
    [anon_sym_sizeof] = ACTIONS(2736),
    [anon_sym___alignof__] = ACTIONS(2736),
    [anon_sym___alignof] = ACTIONS(2736),
    [anon_sym__alignof] = ACTIONS(2736),
    [anon_sym_alignof] = ACTIONS(2736),
    [anon_sym__Alignof] = ACTIONS(2736),
    [anon_sym_offsetof] = ACTIONS(2736),
    [anon_sym__Generic] = ACTIONS(2736),
    [anon_sym_asm] = ACTIONS(2736),
    [anon_sym___asm__] = ACTIONS(2736),
    [anon_sym___asm] = ACTIONS(2736),
    [sym_number_literal] = ACTIONS(2738),
    [anon_sym_L_SQUOTE] = ACTIONS(2738),
    [anon_sym_u_SQUOTE] = ACTIONS(2738),
    [anon_sym_U_SQUOTE] = ACTIONS(2738),
    [anon_sym_u8_SQUOTE] = ACTIONS(2738),
    [anon_sym_SQUOTE] = ACTIONS(2738),
    [anon_sym_AT] = ACTIONS(2736),
    [anon_sym_DQUOTE] = ACTIONS(2738),
    [anon_sym_L_DQUOTE] = ACTIONS(2738),
    [anon_sym_u_DQUOTE] = ACTIONS(2738),
    [anon_sym_U_DQUOTE] = ACTIONS(2738),
    [anon_sym_u8_DQUOTE] = ACTIONS(2738),
    [sym_true] = ACTIONS(2736),
    [sym_false] = ACTIONS(2736),
    [anon_sym_NULL] = ACTIONS(2736),
    [anon_sym_nullptr] = ACTIONS(2736),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2736),
    [anon_sym___typeof] = ACTIONS(2736),
    [anon_sym_typeof] = ACTIONS(2736),
    [anon_sym_ATimport] = ACTIONS(2738),
    [aux_sym_preproc_undef_token1] = ACTIONS(2736),
    [anon_sym_POUND] = ACTIONS(2736),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2736),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2736),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2736),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2736),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2736),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2736),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2736),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_API_AVAILABLE] = ACTIONS(2736),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_API_DEPRECATED] = ACTIONS(2736),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2736),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym___deprecated_msg] = ACTIONS(2736),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2736),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2736),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2736),
    [anon_sym_ATprotocol] = ACTIONS(2738),
    [anon_sym_ATinterface] = ACTIONS(2738),
    [anon_sym_ATimplementation] = ACTIONS(2738),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2738),
    [anon_sym_ATtry] = ACTIONS(2738),
    [anon_sym___try] = ACTIONS(2736),
    [anon_sym_ATthrow] = ACTIONS(2738),
    [anon_sym_ATselector] = ACTIONS(2738),
    [anon_sym_ATavailable] = ACTIONS(2738),
    [anon_sym___builtin_available] = ACTIONS(2736),
    [anon_sym_va_arg] = ACTIONS(2736),
    [anon_sym_ATencode] = ACTIONS(2738),
    [anon_sym_ATsynchronized] = ACTIONS(2738),
    [anon_sym_BOOL] = ACTIONS(2736),
    [anon_sym_IMP] = ACTIONS(2736),
    [anon_sym_SEL] = ACTIONS(2736),
    [anon_sym_Class] = ACTIONS(2736),
    [anon_sym_id] = ACTIONS(2736),
  },
  [STATE(857)] = {
    [sym_identifier] = ACTIONS(2650),
    [aux_sym_preproc_include_token1] = ACTIONS(2650),
    [aux_sym_preproc_include_token2] = ACTIONS(2650),
    [aux_sym_preproc_def_token1] = ACTIONS(2650),
    [aux_sym_preproc_if_token1] = ACTIONS(2650),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2650),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2650),
    [sym_preproc_directive] = ACTIONS(2650),
    [anon_sym_LPAREN2] = ACTIONS(2652),
    [anon_sym_BANG] = ACTIONS(2652),
    [anon_sym_TILDE] = ACTIONS(2652),
    [anon_sym_DASH] = ACTIONS(2650),
    [anon_sym_PLUS] = ACTIONS(2650),
    [anon_sym_STAR] = ACTIONS(2652),
    [anon_sym_CARET] = ACTIONS(2652),
    [anon_sym_AMP] = ACTIONS(2652),
    [anon_sym_SEMI] = ACTIONS(2652),
    [anon_sym___extension__] = ACTIONS(2650),
    [anon_sym_typedef] = ACTIONS(2650),
    [anon_sym_extern] = ACTIONS(2650),
    [anon_sym___attribute__] = ACTIONS(2650),
    [anon_sym___attribute] = ACTIONS(2650),
    [anon_sym_noreturn] = ACTIONS(2650),
    [anon_sym_LBRACK] = ACTIONS(2652),
    [anon_sym___declspec] = ACTIONS(2650),
    [anon_sym___cdecl] = ACTIONS(2650),
    [anon_sym___clrcall] = ACTIONS(2650),
    [anon_sym___stdcall] = ACTIONS(2650),
    [anon_sym___fastcall] = ACTIONS(2650),
    [anon_sym___thiscall] = ACTIONS(2650),
    [anon_sym___vectorcall] = ACTIONS(2650),
    [anon_sym_LBRACE] = ACTIONS(2652),
    [anon_sym_RBRACE] = ACTIONS(2652),
    [anon_sym_signed] = ACTIONS(2650),
    [anon_sym_unsigned] = ACTIONS(2650),
    [anon_sym_long] = ACTIONS(2650),
    [anon_sym_short] = ACTIONS(2650),
    [anon_sym_static] = ACTIONS(2650),
    [anon_sym_ATautoreleasepool] = ACTIONS(2652),
    [anon_sym_auto] = ACTIONS(2650),
    [anon_sym_register] = ACTIONS(2650),
    [anon_sym_inline] = ACTIONS(2650),
    [anon_sym___inline] = ACTIONS(2650),
    [anon_sym___inline__] = ACTIONS(2650),
    [anon_sym___forceinline] = ACTIONS(2650),
    [anon_sym_thread_local] = ACTIONS(2650),
    [anon_sym___thread] = ACTIONS(2650),
    [anon_sym_CG_EXTERN] = ACTIONS(2650),
    [anon_sym_CG_INLINE] = ACTIONS(2650),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2650),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2650),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2650),
    [anon_sym_IBOutlet] = ACTIONS(2650),
    [anon_sym_IBInspectable] = ACTIONS(2650),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2650),
    [anon_sym_NS_INLINE] = ACTIONS(2650),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2650),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2650),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2650),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2650),
    [anon_sym_const] = ACTIONS(2650),
    [anon_sym_constexpr] = ACTIONS(2650),
    [anon_sym_volatile] = ACTIONS(2650),
    [anon_sym_restrict] = ACTIONS(2650),
    [anon_sym___restrict__] = ACTIONS(2650),
    [anon_sym__Atomic] = ACTIONS(2650),
    [anon_sym__Noreturn] = ACTIONS(2650),
    [anon_sym__Nonnull] = ACTIONS(2650),
    [anon_sym_nullable] = ACTIONS(2650),
    [anon_sym__Complex] = ACTIONS(2650),
    [anon_sym__Nullable] = ACTIONS(2650),
    [anon_sym__Nullable_result] = ACTIONS(2650),
    [anon_sym__Null_unspecified] = ACTIONS(2650),
    [anon_sym___autoreleasing] = ACTIONS(2650),
    [anon_sym___block] = ACTIONS(2650),
    [anon_sym___bridge] = ACTIONS(2650),
    [anon_sym___bridge_retained] = ACTIONS(2650),
    [anon_sym___bridge_transfer] = ACTIONS(2650),
    [anon_sym___complex] = ACTIONS(2650),
    [anon_sym___const] = ACTIONS(2650),
    [anon_sym___imag] = ACTIONS(2650),
    [anon_sym___kindof] = ACTIONS(2650),
    [anon_sym___nonnull] = ACTIONS(2650),
    [anon_sym___nullable] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2650),
    [anon_sym___real] = ACTIONS(2650),
    [anon_sym___strong] = ACTIONS(2650),
    [anon_sym___unsafe_unretained] = ACTIONS(2650),
    [anon_sym___unused] = ACTIONS(2650),
    [anon_sym___weak] = ACTIONS(2650),
    [anon_sym_alignas] = ACTIONS(2650),
    [anon_sym__Alignas] = ACTIONS(2650),
    [sym_primitive_type] = ACTIONS(2650),
    [anon_sym_enum] = ACTIONS(2650),
    [anon_sym_struct] = ACTIONS(2650),
    [anon_sym_union] = ACTIONS(2650),
    [anon_sym_if] = ACTIONS(2650),
    [anon_sym_switch] = ACTIONS(2650),
    [anon_sym_case] = ACTIONS(2650),
    [anon_sym_default] = ACTIONS(2650),
    [anon_sym_while] = ACTIONS(2650),
    [anon_sym_do] = ACTIONS(2650),
    [anon_sym_for] = ACTIONS(2650),
    [anon_sym_in] = ACTIONS(2650),
    [anon_sym_return] = ACTIONS(2650),
    [anon_sym_break] = ACTIONS(2650),
    [anon_sym_continue] = ACTIONS(2650),
    [anon_sym_goto] = ACTIONS(2650),
    [anon_sym_DASH_DASH] = ACTIONS(2652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2652),
    [anon_sym_sizeof] = ACTIONS(2650),
    [anon_sym___alignof__] = ACTIONS(2650),
    [anon_sym___alignof] = ACTIONS(2650),
    [anon_sym__alignof] = ACTIONS(2650),
    [anon_sym_alignof] = ACTIONS(2650),
    [anon_sym__Alignof] = ACTIONS(2650),
    [anon_sym_offsetof] = ACTIONS(2650),
    [anon_sym__Generic] = ACTIONS(2650),
    [anon_sym_asm] = ACTIONS(2650),
    [anon_sym___asm__] = ACTIONS(2650),
    [anon_sym___asm] = ACTIONS(2650),
    [sym_number_literal] = ACTIONS(2652),
    [anon_sym_L_SQUOTE] = ACTIONS(2652),
    [anon_sym_u_SQUOTE] = ACTIONS(2652),
    [anon_sym_U_SQUOTE] = ACTIONS(2652),
    [anon_sym_u8_SQUOTE] = ACTIONS(2652),
    [anon_sym_SQUOTE] = ACTIONS(2652),
    [anon_sym_AT] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_L_DQUOTE] = ACTIONS(2652),
    [anon_sym_u_DQUOTE] = ACTIONS(2652),
    [anon_sym_U_DQUOTE] = ACTIONS(2652),
    [anon_sym_u8_DQUOTE] = ACTIONS(2652),
    [sym_true] = ACTIONS(2650),
    [sym_false] = ACTIONS(2650),
    [anon_sym_NULL] = ACTIONS(2650),
    [anon_sym_nullptr] = ACTIONS(2650),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2650),
    [anon_sym___typeof] = ACTIONS(2650),
    [anon_sym_typeof] = ACTIONS(2650),
    [anon_sym_ATimport] = ACTIONS(2652),
    [aux_sym_preproc_undef_token1] = ACTIONS(2650),
    [anon_sym_POUND] = ACTIONS(2650),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2650),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2650),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2650),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2650),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2650),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2650),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2650),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_API_AVAILABLE] = ACTIONS(2650),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_API_DEPRECATED] = ACTIONS(2650),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2650),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym___deprecated_msg] = ACTIONS(2650),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2650),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2650),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2650),
    [anon_sym_ATprotocol] = ACTIONS(2652),
    [anon_sym_ATinterface] = ACTIONS(2652),
    [anon_sym_ATimplementation] = ACTIONS(2652),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2652),
    [anon_sym_ATtry] = ACTIONS(2652),
    [anon_sym___try] = ACTIONS(2650),
    [anon_sym_ATthrow] = ACTIONS(2652),
    [anon_sym_ATselector] = ACTIONS(2652),
    [anon_sym_ATavailable] = ACTIONS(2652),
    [anon_sym___builtin_available] = ACTIONS(2650),
    [anon_sym_va_arg] = ACTIONS(2650),
    [anon_sym_ATencode] = ACTIONS(2652),
    [anon_sym_ATsynchronized] = ACTIONS(2652),
    [anon_sym_BOOL] = ACTIONS(2650),
    [anon_sym_IMP] = ACTIONS(2650),
    [anon_sym_SEL] = ACTIONS(2650),
    [anon_sym_Class] = ACTIONS(2650),
    [anon_sym_id] = ACTIONS(2650),
  },
  [STATE(858)] = {
    [sym_identifier] = ACTIONS(2672),
    [aux_sym_preproc_include_token1] = ACTIONS(2672),
    [aux_sym_preproc_include_token2] = ACTIONS(2672),
    [aux_sym_preproc_def_token1] = ACTIONS(2672),
    [aux_sym_preproc_if_token1] = ACTIONS(2672),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2672),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2672),
    [sym_preproc_directive] = ACTIONS(2672),
    [anon_sym_LPAREN2] = ACTIONS(2674),
    [anon_sym_BANG] = ACTIONS(2674),
    [anon_sym_TILDE] = ACTIONS(2674),
    [anon_sym_DASH] = ACTIONS(2672),
    [anon_sym_PLUS] = ACTIONS(2672),
    [anon_sym_STAR] = ACTIONS(2674),
    [anon_sym_CARET] = ACTIONS(2674),
    [anon_sym_AMP] = ACTIONS(2674),
    [anon_sym_SEMI] = ACTIONS(2674),
    [anon_sym___extension__] = ACTIONS(2672),
    [anon_sym_typedef] = ACTIONS(2672),
    [anon_sym_extern] = ACTIONS(2672),
    [anon_sym___attribute__] = ACTIONS(2672),
    [anon_sym___attribute] = ACTIONS(2672),
    [anon_sym_noreturn] = ACTIONS(2672),
    [anon_sym_LBRACK] = ACTIONS(2674),
    [anon_sym___declspec] = ACTIONS(2672),
    [anon_sym___cdecl] = ACTIONS(2672),
    [anon_sym___clrcall] = ACTIONS(2672),
    [anon_sym___stdcall] = ACTIONS(2672),
    [anon_sym___fastcall] = ACTIONS(2672),
    [anon_sym___thiscall] = ACTIONS(2672),
    [anon_sym___vectorcall] = ACTIONS(2672),
    [anon_sym_LBRACE] = ACTIONS(2674),
    [anon_sym_RBRACE] = ACTIONS(2674),
    [anon_sym_signed] = ACTIONS(2672),
    [anon_sym_unsigned] = ACTIONS(2672),
    [anon_sym_long] = ACTIONS(2672),
    [anon_sym_short] = ACTIONS(2672),
    [anon_sym_static] = ACTIONS(2672),
    [anon_sym_ATautoreleasepool] = ACTIONS(2674),
    [anon_sym_auto] = ACTIONS(2672),
    [anon_sym_register] = ACTIONS(2672),
    [anon_sym_inline] = ACTIONS(2672),
    [anon_sym___inline] = ACTIONS(2672),
    [anon_sym___inline__] = ACTIONS(2672),
    [anon_sym___forceinline] = ACTIONS(2672),
    [anon_sym_thread_local] = ACTIONS(2672),
    [anon_sym___thread] = ACTIONS(2672),
    [anon_sym_CG_EXTERN] = ACTIONS(2672),
    [anon_sym_CG_INLINE] = ACTIONS(2672),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2672),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2672),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2672),
    [anon_sym_IBOutlet] = ACTIONS(2672),
    [anon_sym_IBInspectable] = ACTIONS(2672),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2672),
    [anon_sym_NS_INLINE] = ACTIONS(2672),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2672),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2672),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2672),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2672),
    [anon_sym_const] = ACTIONS(2672),
    [anon_sym_constexpr] = ACTIONS(2672),
    [anon_sym_volatile] = ACTIONS(2672),
    [anon_sym_restrict] = ACTIONS(2672),
    [anon_sym___restrict__] = ACTIONS(2672),
    [anon_sym__Atomic] = ACTIONS(2672),
    [anon_sym__Noreturn] = ACTIONS(2672),
    [anon_sym__Nonnull] = ACTIONS(2672),
    [anon_sym_nullable] = ACTIONS(2672),
    [anon_sym__Complex] = ACTIONS(2672),
    [anon_sym__Nullable] = ACTIONS(2672),
    [anon_sym__Nullable_result] = ACTIONS(2672),
    [anon_sym__Null_unspecified] = ACTIONS(2672),
    [anon_sym___autoreleasing] = ACTIONS(2672),
    [anon_sym___block] = ACTIONS(2672),
    [anon_sym___bridge] = ACTIONS(2672),
    [anon_sym___bridge_retained] = ACTIONS(2672),
    [anon_sym___bridge_transfer] = ACTIONS(2672),
    [anon_sym___complex] = ACTIONS(2672),
    [anon_sym___const] = ACTIONS(2672),
    [anon_sym___imag] = ACTIONS(2672),
    [anon_sym___kindof] = ACTIONS(2672),
    [anon_sym___nonnull] = ACTIONS(2672),
    [anon_sym___nullable] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2672),
    [anon_sym___real] = ACTIONS(2672),
    [anon_sym___strong] = ACTIONS(2672),
    [anon_sym___unsafe_unretained] = ACTIONS(2672),
    [anon_sym___unused] = ACTIONS(2672),
    [anon_sym___weak] = ACTIONS(2672),
    [anon_sym_alignas] = ACTIONS(2672),
    [anon_sym__Alignas] = ACTIONS(2672),
    [sym_primitive_type] = ACTIONS(2672),
    [anon_sym_enum] = ACTIONS(2672),
    [anon_sym_struct] = ACTIONS(2672),
    [anon_sym_union] = ACTIONS(2672),
    [anon_sym_if] = ACTIONS(2672),
    [anon_sym_switch] = ACTIONS(2672),
    [anon_sym_case] = ACTIONS(2672),
    [anon_sym_default] = ACTIONS(2672),
    [anon_sym_while] = ACTIONS(2672),
    [anon_sym_do] = ACTIONS(2672),
    [anon_sym_for] = ACTIONS(2672),
    [anon_sym_in] = ACTIONS(2672),
    [anon_sym_return] = ACTIONS(2672),
    [anon_sym_break] = ACTIONS(2672),
    [anon_sym_continue] = ACTIONS(2672),
    [anon_sym_goto] = ACTIONS(2672),
    [anon_sym_DASH_DASH] = ACTIONS(2674),
    [anon_sym_PLUS_PLUS] = ACTIONS(2674),
    [anon_sym_sizeof] = ACTIONS(2672),
    [anon_sym___alignof__] = ACTIONS(2672),
    [anon_sym___alignof] = ACTIONS(2672),
    [anon_sym__alignof] = ACTIONS(2672),
    [anon_sym_alignof] = ACTIONS(2672),
    [anon_sym__Alignof] = ACTIONS(2672),
    [anon_sym_offsetof] = ACTIONS(2672),
    [anon_sym__Generic] = ACTIONS(2672),
    [anon_sym_asm] = ACTIONS(2672),
    [anon_sym___asm__] = ACTIONS(2672),
    [anon_sym___asm] = ACTIONS(2672),
    [sym_number_literal] = ACTIONS(2674),
    [anon_sym_L_SQUOTE] = ACTIONS(2674),
    [anon_sym_u_SQUOTE] = ACTIONS(2674),
    [anon_sym_U_SQUOTE] = ACTIONS(2674),
    [anon_sym_u8_SQUOTE] = ACTIONS(2674),
    [anon_sym_SQUOTE] = ACTIONS(2674),
    [anon_sym_AT] = ACTIONS(2672),
    [anon_sym_DQUOTE] = ACTIONS(2674),
    [anon_sym_L_DQUOTE] = ACTIONS(2674),
    [anon_sym_u_DQUOTE] = ACTIONS(2674),
    [anon_sym_U_DQUOTE] = ACTIONS(2674),
    [anon_sym_u8_DQUOTE] = ACTIONS(2674),
    [sym_true] = ACTIONS(2672),
    [sym_false] = ACTIONS(2672),
    [anon_sym_NULL] = ACTIONS(2672),
    [anon_sym_nullptr] = ACTIONS(2672),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2672),
    [anon_sym___typeof] = ACTIONS(2672),
    [anon_sym_typeof] = ACTIONS(2672),
    [anon_sym_ATimport] = ACTIONS(2674),
    [aux_sym_preproc_undef_token1] = ACTIONS(2672),
    [anon_sym_POUND] = ACTIONS(2672),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2672),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2672),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2672),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2672),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2672),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2672),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2672),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_API_AVAILABLE] = ACTIONS(2672),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_API_DEPRECATED] = ACTIONS(2672),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2672),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym___deprecated_msg] = ACTIONS(2672),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2672),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2672),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2672),
    [anon_sym_ATprotocol] = ACTIONS(2674),
    [anon_sym_ATinterface] = ACTIONS(2674),
    [anon_sym_ATimplementation] = ACTIONS(2674),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2674),
    [anon_sym_ATtry] = ACTIONS(2674),
    [anon_sym___try] = ACTIONS(2672),
    [anon_sym_ATthrow] = ACTIONS(2674),
    [anon_sym_ATselector] = ACTIONS(2674),
    [anon_sym_ATavailable] = ACTIONS(2674),
    [anon_sym___builtin_available] = ACTIONS(2672),
    [anon_sym_va_arg] = ACTIONS(2672),
    [anon_sym_ATencode] = ACTIONS(2674),
    [anon_sym_ATsynchronized] = ACTIONS(2674),
    [anon_sym_BOOL] = ACTIONS(2672),
    [anon_sym_IMP] = ACTIONS(2672),
    [anon_sym_SEL] = ACTIONS(2672),
    [anon_sym_Class] = ACTIONS(2672),
    [anon_sym_id] = ACTIONS(2672),
  },
  [STATE(859)] = {
    [sym_identifier] = ACTIONS(2608),
    [aux_sym_preproc_include_token1] = ACTIONS(2608),
    [aux_sym_preproc_include_token2] = ACTIONS(2608),
    [aux_sym_preproc_def_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2608),
    [sym_preproc_directive] = ACTIONS(2608),
    [anon_sym_LPAREN2] = ACTIONS(2610),
    [anon_sym_BANG] = ACTIONS(2610),
    [anon_sym_TILDE] = ACTIONS(2610),
    [anon_sym_DASH] = ACTIONS(2608),
    [anon_sym_PLUS] = ACTIONS(2608),
    [anon_sym_STAR] = ACTIONS(2610),
    [anon_sym_CARET] = ACTIONS(2610),
    [anon_sym_AMP] = ACTIONS(2610),
    [anon_sym_SEMI] = ACTIONS(2610),
    [anon_sym___extension__] = ACTIONS(2608),
    [anon_sym_typedef] = ACTIONS(2608),
    [anon_sym_extern] = ACTIONS(2608),
    [anon_sym___attribute__] = ACTIONS(2608),
    [anon_sym___attribute] = ACTIONS(2608),
    [anon_sym_noreturn] = ACTIONS(2608),
    [anon_sym_LBRACK] = ACTIONS(2610),
    [anon_sym___declspec] = ACTIONS(2608),
    [anon_sym___cdecl] = ACTIONS(2608),
    [anon_sym___clrcall] = ACTIONS(2608),
    [anon_sym___stdcall] = ACTIONS(2608),
    [anon_sym___fastcall] = ACTIONS(2608),
    [anon_sym___thiscall] = ACTIONS(2608),
    [anon_sym___vectorcall] = ACTIONS(2608),
    [anon_sym_LBRACE] = ACTIONS(2610),
    [anon_sym_RBRACE] = ACTIONS(2610),
    [anon_sym_signed] = ACTIONS(2608),
    [anon_sym_unsigned] = ACTIONS(2608),
    [anon_sym_long] = ACTIONS(2608),
    [anon_sym_short] = ACTIONS(2608),
    [anon_sym_static] = ACTIONS(2608),
    [anon_sym_ATautoreleasepool] = ACTIONS(2610),
    [anon_sym_auto] = ACTIONS(2608),
    [anon_sym_register] = ACTIONS(2608),
    [anon_sym_inline] = ACTIONS(2608),
    [anon_sym___inline] = ACTIONS(2608),
    [anon_sym___inline__] = ACTIONS(2608),
    [anon_sym___forceinline] = ACTIONS(2608),
    [anon_sym_thread_local] = ACTIONS(2608),
    [anon_sym___thread] = ACTIONS(2608),
    [anon_sym_CG_EXTERN] = ACTIONS(2608),
    [anon_sym_CG_INLINE] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2608),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2608),
    [anon_sym_IBOutlet] = ACTIONS(2608),
    [anon_sym_IBInspectable] = ACTIONS(2608),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2608),
    [anon_sym_NS_INLINE] = ACTIONS(2608),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2608),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2608),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2608),
    [anon_sym_const] = ACTIONS(2608),
    [anon_sym_constexpr] = ACTIONS(2608),
    [anon_sym_volatile] = ACTIONS(2608),
    [anon_sym_restrict] = ACTIONS(2608),
    [anon_sym___restrict__] = ACTIONS(2608),
    [anon_sym__Atomic] = ACTIONS(2608),
    [anon_sym__Noreturn] = ACTIONS(2608),
    [anon_sym__Nonnull] = ACTIONS(2608),
    [anon_sym_nullable] = ACTIONS(2608),
    [anon_sym__Complex] = ACTIONS(2608),
    [anon_sym__Nullable] = ACTIONS(2608),
    [anon_sym__Nullable_result] = ACTIONS(2608),
    [anon_sym__Null_unspecified] = ACTIONS(2608),
    [anon_sym___autoreleasing] = ACTIONS(2608),
    [anon_sym___block] = ACTIONS(2608),
    [anon_sym___bridge] = ACTIONS(2608),
    [anon_sym___bridge_retained] = ACTIONS(2608),
    [anon_sym___bridge_transfer] = ACTIONS(2608),
    [anon_sym___complex] = ACTIONS(2608),
    [anon_sym___const] = ACTIONS(2608),
    [anon_sym___imag] = ACTIONS(2608),
    [anon_sym___kindof] = ACTIONS(2608),
    [anon_sym___nonnull] = ACTIONS(2608),
    [anon_sym___nullable] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2608),
    [anon_sym___real] = ACTIONS(2608),
    [anon_sym___strong] = ACTIONS(2608),
    [anon_sym___unsafe_unretained] = ACTIONS(2608),
    [anon_sym___unused] = ACTIONS(2608),
    [anon_sym___weak] = ACTIONS(2608),
    [anon_sym_alignas] = ACTIONS(2608),
    [anon_sym__Alignas] = ACTIONS(2608),
    [sym_primitive_type] = ACTIONS(2608),
    [anon_sym_enum] = ACTIONS(2608),
    [anon_sym_struct] = ACTIONS(2608),
    [anon_sym_union] = ACTIONS(2608),
    [anon_sym_if] = ACTIONS(2608),
    [anon_sym_switch] = ACTIONS(2608),
    [anon_sym_case] = ACTIONS(2608),
    [anon_sym_default] = ACTIONS(2608),
    [anon_sym_while] = ACTIONS(2608),
    [anon_sym_do] = ACTIONS(2608),
    [anon_sym_for] = ACTIONS(2608),
    [anon_sym_in] = ACTIONS(2608),
    [anon_sym_return] = ACTIONS(2608),
    [anon_sym_break] = ACTIONS(2608),
    [anon_sym_continue] = ACTIONS(2608),
    [anon_sym_goto] = ACTIONS(2608),
    [anon_sym_DASH_DASH] = ACTIONS(2610),
    [anon_sym_PLUS_PLUS] = ACTIONS(2610),
    [anon_sym_sizeof] = ACTIONS(2608),
    [anon_sym___alignof__] = ACTIONS(2608),
    [anon_sym___alignof] = ACTIONS(2608),
    [anon_sym__alignof] = ACTIONS(2608),
    [anon_sym_alignof] = ACTIONS(2608),
    [anon_sym__Alignof] = ACTIONS(2608),
    [anon_sym_offsetof] = ACTIONS(2608),
    [anon_sym__Generic] = ACTIONS(2608),
    [anon_sym_asm] = ACTIONS(2608),
    [anon_sym___asm__] = ACTIONS(2608),
    [anon_sym___asm] = ACTIONS(2608),
    [sym_number_literal] = ACTIONS(2610),
    [anon_sym_L_SQUOTE] = ACTIONS(2610),
    [anon_sym_u_SQUOTE] = ACTIONS(2610),
    [anon_sym_U_SQUOTE] = ACTIONS(2610),
    [anon_sym_u8_SQUOTE] = ACTIONS(2610),
    [anon_sym_SQUOTE] = ACTIONS(2610),
    [anon_sym_AT] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [anon_sym_L_DQUOTE] = ACTIONS(2610),
    [anon_sym_u_DQUOTE] = ACTIONS(2610),
    [anon_sym_U_DQUOTE] = ACTIONS(2610),
    [anon_sym_u8_DQUOTE] = ACTIONS(2610),
    [sym_true] = ACTIONS(2608),
    [sym_false] = ACTIONS(2608),
    [anon_sym_NULL] = ACTIONS(2608),
    [anon_sym_nullptr] = ACTIONS(2608),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2608),
    [anon_sym___typeof] = ACTIONS(2608),
    [anon_sym_typeof] = ACTIONS(2608),
    [anon_sym_ATimport] = ACTIONS(2610),
    [aux_sym_preproc_undef_token1] = ACTIONS(2608),
    [anon_sym_POUND] = ACTIONS(2608),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2608),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2608),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2608),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_API_AVAILABLE] = ACTIONS(2608),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_API_DEPRECATED] = ACTIONS(2608),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym___deprecated_msg] = ACTIONS(2608),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2608),
    [anon_sym_ATprotocol] = ACTIONS(2610),
    [anon_sym_ATinterface] = ACTIONS(2610),
    [anon_sym_ATimplementation] = ACTIONS(2610),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2610),
    [anon_sym_ATtry] = ACTIONS(2610),
    [anon_sym___try] = ACTIONS(2608),
    [anon_sym_ATthrow] = ACTIONS(2610),
    [anon_sym_ATselector] = ACTIONS(2610),
    [anon_sym_ATavailable] = ACTIONS(2610),
    [anon_sym___builtin_available] = ACTIONS(2608),
    [anon_sym_va_arg] = ACTIONS(2608),
    [anon_sym_ATencode] = ACTIONS(2610),
    [anon_sym_ATsynchronized] = ACTIONS(2610),
    [anon_sym_BOOL] = ACTIONS(2608),
    [anon_sym_IMP] = ACTIONS(2608),
    [anon_sym_SEL] = ACTIONS(2608),
    [anon_sym_Class] = ACTIONS(2608),
    [anon_sym_id] = ACTIONS(2608),
  },
  [STATE(860)] = {
    [sym_identifier] = ACTIONS(2618),
    [aux_sym_preproc_include_token1] = ACTIONS(2618),
    [aux_sym_preproc_include_token2] = ACTIONS(2618),
    [aux_sym_preproc_def_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2618),
    [sym_preproc_directive] = ACTIONS(2618),
    [anon_sym_LPAREN2] = ACTIONS(2620),
    [anon_sym_BANG] = ACTIONS(2620),
    [anon_sym_TILDE] = ACTIONS(2620),
    [anon_sym_DASH] = ACTIONS(2618),
    [anon_sym_PLUS] = ACTIONS(2618),
    [anon_sym_STAR] = ACTIONS(2620),
    [anon_sym_CARET] = ACTIONS(2620),
    [anon_sym_AMP] = ACTIONS(2620),
    [anon_sym_SEMI] = ACTIONS(2620),
    [anon_sym___extension__] = ACTIONS(2618),
    [anon_sym_typedef] = ACTIONS(2618),
    [anon_sym_extern] = ACTIONS(2618),
    [anon_sym___attribute__] = ACTIONS(2618),
    [anon_sym___attribute] = ACTIONS(2618),
    [anon_sym_noreturn] = ACTIONS(2618),
    [anon_sym_LBRACK] = ACTIONS(2620),
    [anon_sym___declspec] = ACTIONS(2618),
    [anon_sym___cdecl] = ACTIONS(2618),
    [anon_sym___clrcall] = ACTIONS(2618),
    [anon_sym___stdcall] = ACTIONS(2618),
    [anon_sym___fastcall] = ACTIONS(2618),
    [anon_sym___thiscall] = ACTIONS(2618),
    [anon_sym___vectorcall] = ACTIONS(2618),
    [anon_sym_LBRACE] = ACTIONS(2620),
    [anon_sym_RBRACE] = ACTIONS(2620),
    [anon_sym_signed] = ACTIONS(2618),
    [anon_sym_unsigned] = ACTIONS(2618),
    [anon_sym_long] = ACTIONS(2618),
    [anon_sym_short] = ACTIONS(2618),
    [anon_sym_static] = ACTIONS(2618),
    [anon_sym_ATautoreleasepool] = ACTIONS(2620),
    [anon_sym_auto] = ACTIONS(2618),
    [anon_sym_register] = ACTIONS(2618),
    [anon_sym_inline] = ACTIONS(2618),
    [anon_sym___inline] = ACTIONS(2618),
    [anon_sym___inline__] = ACTIONS(2618),
    [anon_sym___forceinline] = ACTIONS(2618),
    [anon_sym_thread_local] = ACTIONS(2618),
    [anon_sym___thread] = ACTIONS(2618),
    [anon_sym_CG_EXTERN] = ACTIONS(2618),
    [anon_sym_CG_INLINE] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2618),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2618),
    [anon_sym_IBOutlet] = ACTIONS(2618),
    [anon_sym_IBInspectable] = ACTIONS(2618),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2618),
    [anon_sym_NS_INLINE] = ACTIONS(2618),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2618),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2618),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2618),
    [anon_sym_const] = ACTIONS(2618),
    [anon_sym_constexpr] = ACTIONS(2618),
    [anon_sym_volatile] = ACTIONS(2618),
    [anon_sym_restrict] = ACTIONS(2618),
    [anon_sym___restrict__] = ACTIONS(2618),
    [anon_sym__Atomic] = ACTIONS(2618),
    [anon_sym__Noreturn] = ACTIONS(2618),
    [anon_sym__Nonnull] = ACTIONS(2618),
    [anon_sym_nullable] = ACTIONS(2618),
    [anon_sym__Complex] = ACTIONS(2618),
    [anon_sym__Nullable] = ACTIONS(2618),
    [anon_sym__Nullable_result] = ACTIONS(2618),
    [anon_sym__Null_unspecified] = ACTIONS(2618),
    [anon_sym___autoreleasing] = ACTIONS(2618),
    [anon_sym___block] = ACTIONS(2618),
    [anon_sym___bridge] = ACTIONS(2618),
    [anon_sym___bridge_retained] = ACTIONS(2618),
    [anon_sym___bridge_transfer] = ACTIONS(2618),
    [anon_sym___complex] = ACTIONS(2618),
    [anon_sym___const] = ACTIONS(2618),
    [anon_sym___imag] = ACTIONS(2618),
    [anon_sym___kindof] = ACTIONS(2618),
    [anon_sym___nonnull] = ACTIONS(2618),
    [anon_sym___nullable] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2618),
    [anon_sym___real] = ACTIONS(2618),
    [anon_sym___strong] = ACTIONS(2618),
    [anon_sym___unsafe_unretained] = ACTIONS(2618),
    [anon_sym___unused] = ACTIONS(2618),
    [anon_sym___weak] = ACTIONS(2618),
    [anon_sym_alignas] = ACTIONS(2618),
    [anon_sym__Alignas] = ACTIONS(2618),
    [sym_primitive_type] = ACTIONS(2618),
    [anon_sym_enum] = ACTIONS(2618),
    [anon_sym_struct] = ACTIONS(2618),
    [anon_sym_union] = ACTIONS(2618),
    [anon_sym_if] = ACTIONS(2618),
    [anon_sym_switch] = ACTIONS(2618),
    [anon_sym_case] = ACTIONS(2618),
    [anon_sym_default] = ACTIONS(2618),
    [anon_sym_while] = ACTIONS(2618),
    [anon_sym_do] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2618),
    [anon_sym_in] = ACTIONS(2618),
    [anon_sym_return] = ACTIONS(2618),
    [anon_sym_break] = ACTIONS(2618),
    [anon_sym_continue] = ACTIONS(2618),
    [anon_sym_goto] = ACTIONS(2618),
    [anon_sym_DASH_DASH] = ACTIONS(2620),
    [anon_sym_PLUS_PLUS] = ACTIONS(2620),
    [anon_sym_sizeof] = ACTIONS(2618),
    [anon_sym___alignof__] = ACTIONS(2618),
    [anon_sym___alignof] = ACTIONS(2618),
    [anon_sym__alignof] = ACTIONS(2618),
    [anon_sym_alignof] = ACTIONS(2618),
    [anon_sym__Alignof] = ACTIONS(2618),
    [anon_sym_offsetof] = ACTIONS(2618),
    [anon_sym__Generic] = ACTIONS(2618),
    [anon_sym_asm] = ACTIONS(2618),
    [anon_sym___asm__] = ACTIONS(2618),
    [anon_sym___asm] = ACTIONS(2618),
    [sym_number_literal] = ACTIONS(2620),
    [anon_sym_L_SQUOTE] = ACTIONS(2620),
    [anon_sym_u_SQUOTE] = ACTIONS(2620),
    [anon_sym_U_SQUOTE] = ACTIONS(2620),
    [anon_sym_u8_SQUOTE] = ACTIONS(2620),
    [anon_sym_SQUOTE] = ACTIONS(2620),
    [anon_sym_AT] = ACTIONS(2618),
    [anon_sym_DQUOTE] = ACTIONS(2620),
    [anon_sym_L_DQUOTE] = ACTIONS(2620),
    [anon_sym_u_DQUOTE] = ACTIONS(2620),
    [anon_sym_U_DQUOTE] = ACTIONS(2620),
    [anon_sym_u8_DQUOTE] = ACTIONS(2620),
    [sym_true] = ACTIONS(2618),
    [sym_false] = ACTIONS(2618),
    [anon_sym_NULL] = ACTIONS(2618),
    [anon_sym_nullptr] = ACTIONS(2618),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2618),
    [anon_sym___typeof] = ACTIONS(2618),
    [anon_sym_typeof] = ACTIONS(2618),
    [anon_sym_ATimport] = ACTIONS(2620),
    [aux_sym_preproc_undef_token1] = ACTIONS(2618),
    [anon_sym_POUND] = ACTIONS(2618),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2618),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2618),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2618),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_API_AVAILABLE] = ACTIONS(2618),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_API_DEPRECATED] = ACTIONS(2618),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym___deprecated_msg] = ACTIONS(2618),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2618),
    [anon_sym_ATprotocol] = ACTIONS(2620),
    [anon_sym_ATinterface] = ACTIONS(2620),
    [anon_sym_ATimplementation] = ACTIONS(2620),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2620),
    [anon_sym_ATtry] = ACTIONS(2620),
    [anon_sym___try] = ACTIONS(2618),
    [anon_sym_ATthrow] = ACTIONS(2620),
    [anon_sym_ATselector] = ACTIONS(2620),
    [anon_sym_ATavailable] = ACTIONS(2620),
    [anon_sym___builtin_available] = ACTIONS(2618),
    [anon_sym_va_arg] = ACTIONS(2618),
    [anon_sym_ATencode] = ACTIONS(2620),
    [anon_sym_ATsynchronized] = ACTIONS(2620),
    [anon_sym_BOOL] = ACTIONS(2618),
    [anon_sym_IMP] = ACTIONS(2618),
    [anon_sym_SEL] = ACTIONS(2618),
    [anon_sym_Class] = ACTIONS(2618),
    [anon_sym_id] = ACTIONS(2618),
  },
  [STATE(861)] = {
    [sym_identifier] = ACTIONS(2584),
    [aux_sym_preproc_include_token1] = ACTIONS(2584),
    [aux_sym_preproc_include_token2] = ACTIONS(2584),
    [aux_sym_preproc_def_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2584),
    [sym_preproc_directive] = ACTIONS(2584),
    [anon_sym_LPAREN2] = ACTIONS(2586),
    [anon_sym_BANG] = ACTIONS(2586),
    [anon_sym_TILDE] = ACTIONS(2586),
    [anon_sym_DASH] = ACTIONS(2584),
    [anon_sym_PLUS] = ACTIONS(2584),
    [anon_sym_STAR] = ACTIONS(2586),
    [anon_sym_CARET] = ACTIONS(2586),
    [anon_sym_AMP] = ACTIONS(2586),
    [anon_sym_SEMI] = ACTIONS(2586),
    [anon_sym___extension__] = ACTIONS(2584),
    [anon_sym_typedef] = ACTIONS(2584),
    [anon_sym_extern] = ACTIONS(2584),
    [anon_sym___attribute__] = ACTIONS(2584),
    [anon_sym___attribute] = ACTIONS(2584),
    [anon_sym_noreturn] = ACTIONS(2584),
    [anon_sym_LBRACK] = ACTIONS(2586),
    [anon_sym___declspec] = ACTIONS(2584),
    [anon_sym___cdecl] = ACTIONS(2584),
    [anon_sym___clrcall] = ACTIONS(2584),
    [anon_sym___stdcall] = ACTIONS(2584),
    [anon_sym___fastcall] = ACTIONS(2584),
    [anon_sym___thiscall] = ACTIONS(2584),
    [anon_sym___vectorcall] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_RBRACE] = ACTIONS(2586),
    [anon_sym_signed] = ACTIONS(2584),
    [anon_sym_unsigned] = ACTIONS(2584),
    [anon_sym_long] = ACTIONS(2584),
    [anon_sym_short] = ACTIONS(2584),
    [anon_sym_static] = ACTIONS(2584),
    [anon_sym_ATautoreleasepool] = ACTIONS(2586),
    [anon_sym_auto] = ACTIONS(2584),
    [anon_sym_register] = ACTIONS(2584),
    [anon_sym_inline] = ACTIONS(2584),
    [anon_sym___inline] = ACTIONS(2584),
    [anon_sym___inline__] = ACTIONS(2584),
    [anon_sym___forceinline] = ACTIONS(2584),
    [anon_sym_thread_local] = ACTIONS(2584),
    [anon_sym___thread] = ACTIONS(2584),
    [anon_sym_CG_EXTERN] = ACTIONS(2584),
    [anon_sym_CG_INLINE] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2584),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2584),
    [anon_sym_IBOutlet] = ACTIONS(2584),
    [anon_sym_IBInspectable] = ACTIONS(2584),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2584),
    [anon_sym_NS_INLINE] = ACTIONS(2584),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2584),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2584),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2584),
    [anon_sym_const] = ACTIONS(2584),
    [anon_sym_constexpr] = ACTIONS(2584),
    [anon_sym_volatile] = ACTIONS(2584),
    [anon_sym_restrict] = ACTIONS(2584),
    [anon_sym___restrict__] = ACTIONS(2584),
    [anon_sym__Atomic] = ACTIONS(2584),
    [anon_sym__Noreturn] = ACTIONS(2584),
    [anon_sym__Nonnull] = ACTIONS(2584),
    [anon_sym_nullable] = ACTIONS(2584),
    [anon_sym__Complex] = ACTIONS(2584),
    [anon_sym__Nullable] = ACTIONS(2584),
    [anon_sym__Nullable_result] = ACTIONS(2584),
    [anon_sym__Null_unspecified] = ACTIONS(2584),
    [anon_sym___autoreleasing] = ACTIONS(2584),
    [anon_sym___block] = ACTIONS(2584),
    [anon_sym___bridge] = ACTIONS(2584),
    [anon_sym___bridge_retained] = ACTIONS(2584),
    [anon_sym___bridge_transfer] = ACTIONS(2584),
    [anon_sym___complex] = ACTIONS(2584),
    [anon_sym___const] = ACTIONS(2584),
    [anon_sym___imag] = ACTIONS(2584),
    [anon_sym___kindof] = ACTIONS(2584),
    [anon_sym___nonnull] = ACTIONS(2584),
    [anon_sym___nullable] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2584),
    [anon_sym___real] = ACTIONS(2584),
    [anon_sym___strong] = ACTIONS(2584),
    [anon_sym___unsafe_unretained] = ACTIONS(2584),
    [anon_sym___unused] = ACTIONS(2584),
    [anon_sym___weak] = ACTIONS(2584),
    [anon_sym_alignas] = ACTIONS(2584),
    [anon_sym__Alignas] = ACTIONS(2584),
    [sym_primitive_type] = ACTIONS(2584),
    [anon_sym_enum] = ACTIONS(2584),
    [anon_sym_struct] = ACTIONS(2584),
    [anon_sym_union] = ACTIONS(2584),
    [anon_sym_if] = ACTIONS(2584),
    [anon_sym_switch] = ACTIONS(2584),
    [anon_sym_case] = ACTIONS(2584),
    [anon_sym_default] = ACTIONS(2584),
    [anon_sym_while] = ACTIONS(2584),
    [anon_sym_do] = ACTIONS(2584),
    [anon_sym_for] = ACTIONS(2584),
    [anon_sym_in] = ACTIONS(2584),
    [anon_sym_return] = ACTIONS(2584),
    [anon_sym_break] = ACTIONS(2584),
    [anon_sym_continue] = ACTIONS(2584),
    [anon_sym_goto] = ACTIONS(2584),
    [anon_sym_DASH_DASH] = ACTIONS(2586),
    [anon_sym_PLUS_PLUS] = ACTIONS(2586),
    [anon_sym_sizeof] = ACTIONS(2584),
    [anon_sym___alignof__] = ACTIONS(2584),
    [anon_sym___alignof] = ACTIONS(2584),
    [anon_sym__alignof] = ACTIONS(2584),
    [anon_sym_alignof] = ACTIONS(2584),
    [anon_sym__Alignof] = ACTIONS(2584),
    [anon_sym_offsetof] = ACTIONS(2584),
    [anon_sym__Generic] = ACTIONS(2584),
    [anon_sym_asm] = ACTIONS(2584),
    [anon_sym___asm__] = ACTIONS(2584),
    [anon_sym___asm] = ACTIONS(2584),
    [sym_number_literal] = ACTIONS(2586),
    [anon_sym_L_SQUOTE] = ACTIONS(2586),
    [anon_sym_u_SQUOTE] = ACTIONS(2586),
    [anon_sym_U_SQUOTE] = ACTIONS(2586),
    [anon_sym_u8_SQUOTE] = ACTIONS(2586),
    [anon_sym_SQUOTE] = ACTIONS(2586),
    [anon_sym_AT] = ACTIONS(2584),
    [anon_sym_DQUOTE] = ACTIONS(2586),
    [anon_sym_L_DQUOTE] = ACTIONS(2586),
    [anon_sym_u_DQUOTE] = ACTIONS(2586),
    [anon_sym_U_DQUOTE] = ACTIONS(2586),
    [anon_sym_u8_DQUOTE] = ACTIONS(2586),
    [sym_true] = ACTIONS(2584),
    [sym_false] = ACTIONS(2584),
    [anon_sym_NULL] = ACTIONS(2584),
    [anon_sym_nullptr] = ACTIONS(2584),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2584),
    [anon_sym___typeof] = ACTIONS(2584),
    [anon_sym_typeof] = ACTIONS(2584),
    [anon_sym_ATimport] = ACTIONS(2586),
    [aux_sym_preproc_undef_token1] = ACTIONS(2584),
    [anon_sym_POUND] = ACTIONS(2584),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2584),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2584),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2584),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_API_AVAILABLE] = ACTIONS(2584),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_API_DEPRECATED] = ACTIONS(2584),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym___deprecated_msg] = ACTIONS(2584),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2584),
    [anon_sym_ATprotocol] = ACTIONS(2586),
    [anon_sym_ATinterface] = ACTIONS(2586),
    [anon_sym_ATimplementation] = ACTIONS(2586),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2586),
    [anon_sym_ATtry] = ACTIONS(2586),
    [anon_sym___try] = ACTIONS(2584),
    [anon_sym_ATthrow] = ACTIONS(2586),
    [anon_sym_ATselector] = ACTIONS(2586),
    [anon_sym_ATavailable] = ACTIONS(2586),
    [anon_sym___builtin_available] = ACTIONS(2584),
    [anon_sym_va_arg] = ACTIONS(2584),
    [anon_sym_ATencode] = ACTIONS(2586),
    [anon_sym_ATsynchronized] = ACTIONS(2586),
    [anon_sym_BOOL] = ACTIONS(2584),
    [anon_sym_IMP] = ACTIONS(2584),
    [anon_sym_SEL] = ACTIONS(2584),
    [anon_sym_Class] = ACTIONS(2584),
    [anon_sym_id] = ACTIONS(2584),
  },
  [STATE(862)] = {
    [sym_identifier] = ACTIONS(2780),
    [aux_sym_preproc_include_token1] = ACTIONS(2780),
    [aux_sym_preproc_include_token2] = ACTIONS(2780),
    [aux_sym_preproc_def_token1] = ACTIONS(2780),
    [aux_sym_preproc_if_token1] = ACTIONS(2780),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2780),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2780),
    [sym_preproc_directive] = ACTIONS(2780),
    [anon_sym_LPAREN2] = ACTIONS(2782),
    [anon_sym_BANG] = ACTIONS(2782),
    [anon_sym_TILDE] = ACTIONS(2782),
    [anon_sym_DASH] = ACTIONS(2780),
    [anon_sym_PLUS] = ACTIONS(2780),
    [anon_sym_STAR] = ACTIONS(2782),
    [anon_sym_CARET] = ACTIONS(2782),
    [anon_sym_AMP] = ACTIONS(2782),
    [anon_sym_SEMI] = ACTIONS(2782),
    [anon_sym___extension__] = ACTIONS(2780),
    [anon_sym_typedef] = ACTIONS(2780),
    [anon_sym_extern] = ACTIONS(2780),
    [anon_sym___attribute__] = ACTIONS(2780),
    [anon_sym___attribute] = ACTIONS(2780),
    [anon_sym_noreturn] = ACTIONS(2780),
    [anon_sym_LBRACK] = ACTIONS(2782),
    [anon_sym___declspec] = ACTIONS(2780),
    [anon_sym___cdecl] = ACTIONS(2780),
    [anon_sym___clrcall] = ACTIONS(2780),
    [anon_sym___stdcall] = ACTIONS(2780),
    [anon_sym___fastcall] = ACTIONS(2780),
    [anon_sym___thiscall] = ACTIONS(2780),
    [anon_sym___vectorcall] = ACTIONS(2780),
    [anon_sym_LBRACE] = ACTIONS(2782),
    [anon_sym_RBRACE] = ACTIONS(2782),
    [anon_sym_signed] = ACTIONS(2780),
    [anon_sym_unsigned] = ACTIONS(2780),
    [anon_sym_long] = ACTIONS(2780),
    [anon_sym_short] = ACTIONS(2780),
    [anon_sym_static] = ACTIONS(2780),
    [anon_sym_ATautoreleasepool] = ACTIONS(2782),
    [anon_sym_auto] = ACTIONS(2780),
    [anon_sym_register] = ACTIONS(2780),
    [anon_sym_inline] = ACTIONS(2780),
    [anon_sym___inline] = ACTIONS(2780),
    [anon_sym___inline__] = ACTIONS(2780),
    [anon_sym___forceinline] = ACTIONS(2780),
    [anon_sym_thread_local] = ACTIONS(2780),
    [anon_sym___thread] = ACTIONS(2780),
    [anon_sym_CG_EXTERN] = ACTIONS(2780),
    [anon_sym_CG_INLINE] = ACTIONS(2780),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2780),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2780),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2780),
    [anon_sym_IBOutlet] = ACTIONS(2780),
    [anon_sym_IBInspectable] = ACTIONS(2780),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2780),
    [anon_sym_NS_INLINE] = ACTIONS(2780),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2780),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2780),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2780),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2780),
    [anon_sym_const] = ACTIONS(2780),
    [anon_sym_constexpr] = ACTIONS(2780),
    [anon_sym_volatile] = ACTIONS(2780),
    [anon_sym_restrict] = ACTIONS(2780),
    [anon_sym___restrict__] = ACTIONS(2780),
    [anon_sym__Atomic] = ACTIONS(2780),
    [anon_sym__Noreturn] = ACTIONS(2780),
    [anon_sym__Nonnull] = ACTIONS(2780),
    [anon_sym_nullable] = ACTIONS(2780),
    [anon_sym__Complex] = ACTIONS(2780),
    [anon_sym__Nullable] = ACTIONS(2780),
    [anon_sym__Nullable_result] = ACTIONS(2780),
    [anon_sym__Null_unspecified] = ACTIONS(2780),
    [anon_sym___autoreleasing] = ACTIONS(2780),
    [anon_sym___block] = ACTIONS(2780),
    [anon_sym___bridge] = ACTIONS(2780),
    [anon_sym___bridge_retained] = ACTIONS(2780),
    [anon_sym___bridge_transfer] = ACTIONS(2780),
    [anon_sym___complex] = ACTIONS(2780),
    [anon_sym___const] = ACTIONS(2780),
    [anon_sym___imag] = ACTIONS(2780),
    [anon_sym___kindof] = ACTIONS(2780),
    [anon_sym___nonnull] = ACTIONS(2780),
    [anon_sym___nullable] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2780),
    [anon_sym___real] = ACTIONS(2780),
    [anon_sym___strong] = ACTIONS(2780),
    [anon_sym___unsafe_unretained] = ACTIONS(2780),
    [anon_sym___unused] = ACTIONS(2780),
    [anon_sym___weak] = ACTIONS(2780),
    [anon_sym_alignas] = ACTIONS(2780),
    [anon_sym__Alignas] = ACTIONS(2780),
    [sym_primitive_type] = ACTIONS(2780),
    [anon_sym_enum] = ACTIONS(2780),
    [anon_sym_struct] = ACTIONS(2780),
    [anon_sym_union] = ACTIONS(2780),
    [anon_sym_if] = ACTIONS(2780),
    [anon_sym_switch] = ACTIONS(2780),
    [anon_sym_case] = ACTIONS(2780),
    [anon_sym_default] = ACTIONS(2780),
    [anon_sym_while] = ACTIONS(2780),
    [anon_sym_do] = ACTIONS(2780),
    [anon_sym_for] = ACTIONS(2780),
    [anon_sym_in] = ACTIONS(2780),
    [anon_sym_return] = ACTIONS(2780),
    [anon_sym_break] = ACTIONS(2780),
    [anon_sym_continue] = ACTIONS(2780),
    [anon_sym_goto] = ACTIONS(2780),
    [anon_sym_DASH_DASH] = ACTIONS(2782),
    [anon_sym_PLUS_PLUS] = ACTIONS(2782),
    [anon_sym_sizeof] = ACTIONS(2780),
    [anon_sym___alignof__] = ACTIONS(2780),
    [anon_sym___alignof] = ACTIONS(2780),
    [anon_sym__alignof] = ACTIONS(2780),
    [anon_sym_alignof] = ACTIONS(2780),
    [anon_sym__Alignof] = ACTIONS(2780),
    [anon_sym_offsetof] = ACTIONS(2780),
    [anon_sym__Generic] = ACTIONS(2780),
    [anon_sym_asm] = ACTIONS(2780),
    [anon_sym___asm__] = ACTIONS(2780),
    [anon_sym___asm] = ACTIONS(2780),
    [sym_number_literal] = ACTIONS(2782),
    [anon_sym_L_SQUOTE] = ACTIONS(2782),
    [anon_sym_u_SQUOTE] = ACTIONS(2782),
    [anon_sym_U_SQUOTE] = ACTIONS(2782),
    [anon_sym_u8_SQUOTE] = ACTIONS(2782),
    [anon_sym_SQUOTE] = ACTIONS(2782),
    [anon_sym_AT] = ACTIONS(2780),
    [anon_sym_DQUOTE] = ACTIONS(2782),
    [anon_sym_L_DQUOTE] = ACTIONS(2782),
    [anon_sym_u_DQUOTE] = ACTIONS(2782),
    [anon_sym_U_DQUOTE] = ACTIONS(2782),
    [anon_sym_u8_DQUOTE] = ACTIONS(2782),
    [sym_true] = ACTIONS(2780),
    [sym_false] = ACTIONS(2780),
    [anon_sym_NULL] = ACTIONS(2780),
    [anon_sym_nullptr] = ACTIONS(2780),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2780),
    [anon_sym___typeof] = ACTIONS(2780),
    [anon_sym_typeof] = ACTIONS(2780),
    [anon_sym_ATimport] = ACTIONS(2782),
    [aux_sym_preproc_undef_token1] = ACTIONS(2780),
    [anon_sym_POUND] = ACTIONS(2780),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2780),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2780),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2780),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2780),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2780),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2780),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2780),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_API_AVAILABLE] = ACTIONS(2780),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_API_DEPRECATED] = ACTIONS(2780),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2780),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym___deprecated_msg] = ACTIONS(2780),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2780),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2780),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2780),
    [anon_sym_ATprotocol] = ACTIONS(2782),
    [anon_sym_ATinterface] = ACTIONS(2782),
    [anon_sym_ATimplementation] = ACTIONS(2782),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2782),
    [anon_sym_ATtry] = ACTIONS(2782),
    [anon_sym___try] = ACTIONS(2780),
    [anon_sym_ATthrow] = ACTIONS(2782),
    [anon_sym_ATselector] = ACTIONS(2782),
    [anon_sym_ATavailable] = ACTIONS(2782),
    [anon_sym___builtin_available] = ACTIONS(2780),
    [anon_sym_va_arg] = ACTIONS(2780),
    [anon_sym_ATencode] = ACTIONS(2782),
    [anon_sym_ATsynchronized] = ACTIONS(2782),
    [anon_sym_BOOL] = ACTIONS(2780),
    [anon_sym_IMP] = ACTIONS(2780),
    [anon_sym_SEL] = ACTIONS(2780),
    [anon_sym_Class] = ACTIONS(2780),
    [anon_sym_id] = ACTIONS(2780),
  },
  [STATE(863)] = {
    [sym_identifier] = ACTIONS(2784),
    [aux_sym_preproc_include_token1] = ACTIONS(2784),
    [aux_sym_preproc_include_token2] = ACTIONS(2784),
    [aux_sym_preproc_def_token1] = ACTIONS(2784),
    [aux_sym_preproc_if_token1] = ACTIONS(2784),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2784),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2784),
    [sym_preproc_directive] = ACTIONS(2784),
    [anon_sym_LPAREN2] = ACTIONS(2786),
    [anon_sym_BANG] = ACTIONS(2786),
    [anon_sym_TILDE] = ACTIONS(2786),
    [anon_sym_DASH] = ACTIONS(2784),
    [anon_sym_PLUS] = ACTIONS(2784),
    [anon_sym_STAR] = ACTIONS(2786),
    [anon_sym_CARET] = ACTIONS(2786),
    [anon_sym_AMP] = ACTIONS(2786),
    [anon_sym_SEMI] = ACTIONS(2786),
    [anon_sym___extension__] = ACTIONS(2784),
    [anon_sym_typedef] = ACTIONS(2784),
    [anon_sym_extern] = ACTIONS(2784),
    [anon_sym___attribute__] = ACTIONS(2784),
    [anon_sym___attribute] = ACTIONS(2784),
    [anon_sym_noreturn] = ACTIONS(2784),
    [anon_sym_LBRACK] = ACTIONS(2786),
    [anon_sym___declspec] = ACTIONS(2784),
    [anon_sym___cdecl] = ACTIONS(2784),
    [anon_sym___clrcall] = ACTIONS(2784),
    [anon_sym___stdcall] = ACTIONS(2784),
    [anon_sym___fastcall] = ACTIONS(2784),
    [anon_sym___thiscall] = ACTIONS(2784),
    [anon_sym___vectorcall] = ACTIONS(2784),
    [anon_sym_LBRACE] = ACTIONS(2786),
    [anon_sym_RBRACE] = ACTIONS(2786),
    [anon_sym_signed] = ACTIONS(2784),
    [anon_sym_unsigned] = ACTIONS(2784),
    [anon_sym_long] = ACTIONS(2784),
    [anon_sym_short] = ACTIONS(2784),
    [anon_sym_static] = ACTIONS(2784),
    [anon_sym_ATautoreleasepool] = ACTIONS(2786),
    [anon_sym_auto] = ACTIONS(2784),
    [anon_sym_register] = ACTIONS(2784),
    [anon_sym_inline] = ACTIONS(2784),
    [anon_sym___inline] = ACTIONS(2784),
    [anon_sym___inline__] = ACTIONS(2784),
    [anon_sym___forceinline] = ACTIONS(2784),
    [anon_sym_thread_local] = ACTIONS(2784),
    [anon_sym___thread] = ACTIONS(2784),
    [anon_sym_CG_EXTERN] = ACTIONS(2784),
    [anon_sym_CG_INLINE] = ACTIONS(2784),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2784),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2784),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2784),
    [anon_sym_IBOutlet] = ACTIONS(2784),
    [anon_sym_IBInspectable] = ACTIONS(2784),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2784),
    [anon_sym_NS_INLINE] = ACTIONS(2784),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2784),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2784),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2784),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2784),
    [anon_sym_const] = ACTIONS(2784),
    [anon_sym_constexpr] = ACTIONS(2784),
    [anon_sym_volatile] = ACTIONS(2784),
    [anon_sym_restrict] = ACTIONS(2784),
    [anon_sym___restrict__] = ACTIONS(2784),
    [anon_sym__Atomic] = ACTIONS(2784),
    [anon_sym__Noreturn] = ACTIONS(2784),
    [anon_sym__Nonnull] = ACTIONS(2784),
    [anon_sym_nullable] = ACTIONS(2784),
    [anon_sym__Complex] = ACTIONS(2784),
    [anon_sym__Nullable] = ACTIONS(2784),
    [anon_sym__Nullable_result] = ACTIONS(2784),
    [anon_sym__Null_unspecified] = ACTIONS(2784),
    [anon_sym___autoreleasing] = ACTIONS(2784),
    [anon_sym___block] = ACTIONS(2784),
    [anon_sym___bridge] = ACTIONS(2784),
    [anon_sym___bridge_retained] = ACTIONS(2784),
    [anon_sym___bridge_transfer] = ACTIONS(2784),
    [anon_sym___complex] = ACTIONS(2784),
    [anon_sym___const] = ACTIONS(2784),
    [anon_sym___imag] = ACTIONS(2784),
    [anon_sym___kindof] = ACTIONS(2784),
    [anon_sym___nonnull] = ACTIONS(2784),
    [anon_sym___nullable] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2784),
    [anon_sym___real] = ACTIONS(2784),
    [anon_sym___strong] = ACTIONS(2784),
    [anon_sym___unsafe_unretained] = ACTIONS(2784),
    [anon_sym___unused] = ACTIONS(2784),
    [anon_sym___weak] = ACTIONS(2784),
    [anon_sym_alignas] = ACTIONS(2784),
    [anon_sym__Alignas] = ACTIONS(2784),
    [sym_primitive_type] = ACTIONS(2784),
    [anon_sym_enum] = ACTIONS(2784),
    [anon_sym_struct] = ACTIONS(2784),
    [anon_sym_union] = ACTIONS(2784),
    [anon_sym_if] = ACTIONS(2784),
    [anon_sym_switch] = ACTIONS(2784),
    [anon_sym_case] = ACTIONS(2784),
    [anon_sym_default] = ACTIONS(2784),
    [anon_sym_while] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2784),
    [anon_sym_for] = ACTIONS(2784),
    [anon_sym_in] = ACTIONS(2784),
    [anon_sym_return] = ACTIONS(2784),
    [anon_sym_break] = ACTIONS(2784),
    [anon_sym_continue] = ACTIONS(2784),
    [anon_sym_goto] = ACTIONS(2784),
    [anon_sym_DASH_DASH] = ACTIONS(2786),
    [anon_sym_PLUS_PLUS] = ACTIONS(2786),
    [anon_sym_sizeof] = ACTIONS(2784),
    [anon_sym___alignof__] = ACTIONS(2784),
    [anon_sym___alignof] = ACTIONS(2784),
    [anon_sym__alignof] = ACTIONS(2784),
    [anon_sym_alignof] = ACTIONS(2784),
    [anon_sym__Alignof] = ACTIONS(2784),
    [anon_sym_offsetof] = ACTIONS(2784),
    [anon_sym__Generic] = ACTIONS(2784),
    [anon_sym_asm] = ACTIONS(2784),
    [anon_sym___asm__] = ACTIONS(2784),
    [anon_sym___asm] = ACTIONS(2784),
    [sym_number_literal] = ACTIONS(2786),
    [anon_sym_L_SQUOTE] = ACTIONS(2786),
    [anon_sym_u_SQUOTE] = ACTIONS(2786),
    [anon_sym_U_SQUOTE] = ACTIONS(2786),
    [anon_sym_u8_SQUOTE] = ACTIONS(2786),
    [anon_sym_SQUOTE] = ACTIONS(2786),
    [anon_sym_AT] = ACTIONS(2784),
    [anon_sym_DQUOTE] = ACTIONS(2786),
    [anon_sym_L_DQUOTE] = ACTIONS(2786),
    [anon_sym_u_DQUOTE] = ACTIONS(2786),
    [anon_sym_U_DQUOTE] = ACTIONS(2786),
    [anon_sym_u8_DQUOTE] = ACTIONS(2786),
    [sym_true] = ACTIONS(2784),
    [sym_false] = ACTIONS(2784),
    [anon_sym_NULL] = ACTIONS(2784),
    [anon_sym_nullptr] = ACTIONS(2784),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2784),
    [anon_sym___typeof] = ACTIONS(2784),
    [anon_sym_typeof] = ACTIONS(2784),
    [anon_sym_ATimport] = ACTIONS(2786),
    [aux_sym_preproc_undef_token1] = ACTIONS(2784),
    [anon_sym_POUND] = ACTIONS(2784),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2784),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2784),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2784),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2784),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2784),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2784),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2784),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_API_AVAILABLE] = ACTIONS(2784),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_API_DEPRECATED] = ACTIONS(2784),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2784),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym___deprecated_msg] = ACTIONS(2784),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2784),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2784),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2784),
    [anon_sym_ATprotocol] = ACTIONS(2786),
    [anon_sym_ATinterface] = ACTIONS(2786),
    [anon_sym_ATimplementation] = ACTIONS(2786),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2786),
    [anon_sym_ATtry] = ACTIONS(2786),
    [anon_sym___try] = ACTIONS(2784),
    [anon_sym_ATthrow] = ACTIONS(2786),
    [anon_sym_ATselector] = ACTIONS(2786),
    [anon_sym_ATavailable] = ACTIONS(2786),
    [anon_sym___builtin_available] = ACTIONS(2784),
    [anon_sym_va_arg] = ACTIONS(2784),
    [anon_sym_ATencode] = ACTIONS(2786),
    [anon_sym_ATsynchronized] = ACTIONS(2786),
    [anon_sym_BOOL] = ACTIONS(2784),
    [anon_sym_IMP] = ACTIONS(2784),
    [anon_sym_SEL] = ACTIONS(2784),
    [anon_sym_Class] = ACTIONS(2784),
    [anon_sym_id] = ACTIONS(2784),
  },
  [STATE(864)] = {
    [sym_identifier] = ACTIONS(2788),
    [aux_sym_preproc_include_token1] = ACTIONS(2788),
    [aux_sym_preproc_include_token2] = ACTIONS(2788),
    [aux_sym_preproc_def_token1] = ACTIONS(2788),
    [aux_sym_preproc_if_token1] = ACTIONS(2788),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2788),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2788),
    [sym_preproc_directive] = ACTIONS(2788),
    [anon_sym_LPAREN2] = ACTIONS(2790),
    [anon_sym_BANG] = ACTIONS(2790),
    [anon_sym_TILDE] = ACTIONS(2790),
    [anon_sym_DASH] = ACTIONS(2788),
    [anon_sym_PLUS] = ACTIONS(2788),
    [anon_sym_STAR] = ACTIONS(2790),
    [anon_sym_CARET] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(2790),
    [anon_sym_SEMI] = ACTIONS(2790),
    [anon_sym___extension__] = ACTIONS(2788),
    [anon_sym_typedef] = ACTIONS(2788),
    [anon_sym_extern] = ACTIONS(2788),
    [anon_sym___attribute__] = ACTIONS(2788),
    [anon_sym___attribute] = ACTIONS(2788),
    [anon_sym_noreturn] = ACTIONS(2788),
    [anon_sym_LBRACK] = ACTIONS(2790),
    [anon_sym___declspec] = ACTIONS(2788),
    [anon_sym___cdecl] = ACTIONS(2788),
    [anon_sym___clrcall] = ACTIONS(2788),
    [anon_sym___stdcall] = ACTIONS(2788),
    [anon_sym___fastcall] = ACTIONS(2788),
    [anon_sym___thiscall] = ACTIONS(2788),
    [anon_sym___vectorcall] = ACTIONS(2788),
    [anon_sym_LBRACE] = ACTIONS(2790),
    [anon_sym_RBRACE] = ACTIONS(2790),
    [anon_sym_signed] = ACTIONS(2788),
    [anon_sym_unsigned] = ACTIONS(2788),
    [anon_sym_long] = ACTIONS(2788),
    [anon_sym_short] = ACTIONS(2788),
    [anon_sym_static] = ACTIONS(2788),
    [anon_sym_ATautoreleasepool] = ACTIONS(2790),
    [anon_sym_auto] = ACTIONS(2788),
    [anon_sym_register] = ACTIONS(2788),
    [anon_sym_inline] = ACTIONS(2788),
    [anon_sym___inline] = ACTIONS(2788),
    [anon_sym___inline__] = ACTIONS(2788),
    [anon_sym___forceinline] = ACTIONS(2788),
    [anon_sym_thread_local] = ACTIONS(2788),
    [anon_sym___thread] = ACTIONS(2788),
    [anon_sym_CG_EXTERN] = ACTIONS(2788),
    [anon_sym_CG_INLINE] = ACTIONS(2788),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2788),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2788),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2788),
    [anon_sym_IBOutlet] = ACTIONS(2788),
    [anon_sym_IBInspectable] = ACTIONS(2788),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2788),
    [anon_sym_NS_INLINE] = ACTIONS(2788),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2788),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2788),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2788),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2788),
    [anon_sym_const] = ACTIONS(2788),
    [anon_sym_constexpr] = ACTIONS(2788),
    [anon_sym_volatile] = ACTIONS(2788),
    [anon_sym_restrict] = ACTIONS(2788),
    [anon_sym___restrict__] = ACTIONS(2788),
    [anon_sym__Atomic] = ACTIONS(2788),
    [anon_sym__Noreturn] = ACTIONS(2788),
    [anon_sym__Nonnull] = ACTIONS(2788),
    [anon_sym_nullable] = ACTIONS(2788),
    [anon_sym__Complex] = ACTIONS(2788),
    [anon_sym__Nullable] = ACTIONS(2788),
    [anon_sym__Nullable_result] = ACTIONS(2788),
    [anon_sym__Null_unspecified] = ACTIONS(2788),
    [anon_sym___autoreleasing] = ACTIONS(2788),
    [anon_sym___block] = ACTIONS(2788),
    [anon_sym___bridge] = ACTIONS(2788),
    [anon_sym___bridge_retained] = ACTIONS(2788),
    [anon_sym___bridge_transfer] = ACTIONS(2788),
    [anon_sym___complex] = ACTIONS(2788),
    [anon_sym___const] = ACTIONS(2788),
    [anon_sym___imag] = ACTIONS(2788),
    [anon_sym___kindof] = ACTIONS(2788),
    [anon_sym___nonnull] = ACTIONS(2788),
    [anon_sym___nullable] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2788),
    [anon_sym___real] = ACTIONS(2788),
    [anon_sym___strong] = ACTIONS(2788),
    [anon_sym___unsafe_unretained] = ACTIONS(2788),
    [anon_sym___unused] = ACTIONS(2788),
    [anon_sym___weak] = ACTIONS(2788),
    [anon_sym_alignas] = ACTIONS(2788),
    [anon_sym__Alignas] = ACTIONS(2788),
    [sym_primitive_type] = ACTIONS(2788),
    [anon_sym_enum] = ACTIONS(2788),
    [anon_sym_struct] = ACTIONS(2788),
    [anon_sym_union] = ACTIONS(2788),
    [anon_sym_if] = ACTIONS(2788),
    [anon_sym_switch] = ACTIONS(2788),
    [anon_sym_case] = ACTIONS(2788),
    [anon_sym_default] = ACTIONS(2788),
    [anon_sym_while] = ACTIONS(2788),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_for] = ACTIONS(2788),
    [anon_sym_in] = ACTIONS(2788),
    [anon_sym_return] = ACTIONS(2788),
    [anon_sym_break] = ACTIONS(2788),
    [anon_sym_continue] = ACTIONS(2788),
    [anon_sym_goto] = ACTIONS(2788),
    [anon_sym_DASH_DASH] = ACTIONS(2790),
    [anon_sym_PLUS_PLUS] = ACTIONS(2790),
    [anon_sym_sizeof] = ACTIONS(2788),
    [anon_sym___alignof__] = ACTIONS(2788),
    [anon_sym___alignof] = ACTIONS(2788),
    [anon_sym__alignof] = ACTIONS(2788),
    [anon_sym_alignof] = ACTIONS(2788),
    [anon_sym__Alignof] = ACTIONS(2788),
    [anon_sym_offsetof] = ACTIONS(2788),
    [anon_sym__Generic] = ACTIONS(2788),
    [anon_sym_asm] = ACTIONS(2788),
    [anon_sym___asm__] = ACTIONS(2788),
    [anon_sym___asm] = ACTIONS(2788),
    [sym_number_literal] = ACTIONS(2790),
    [anon_sym_L_SQUOTE] = ACTIONS(2790),
    [anon_sym_u_SQUOTE] = ACTIONS(2790),
    [anon_sym_U_SQUOTE] = ACTIONS(2790),
    [anon_sym_u8_SQUOTE] = ACTIONS(2790),
    [anon_sym_SQUOTE] = ACTIONS(2790),
    [anon_sym_AT] = ACTIONS(2788),
    [anon_sym_DQUOTE] = ACTIONS(2790),
    [anon_sym_L_DQUOTE] = ACTIONS(2790),
    [anon_sym_u_DQUOTE] = ACTIONS(2790),
    [anon_sym_U_DQUOTE] = ACTIONS(2790),
    [anon_sym_u8_DQUOTE] = ACTIONS(2790),
    [sym_true] = ACTIONS(2788),
    [sym_false] = ACTIONS(2788),
    [anon_sym_NULL] = ACTIONS(2788),
    [anon_sym_nullptr] = ACTIONS(2788),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2788),
    [anon_sym___typeof] = ACTIONS(2788),
    [anon_sym_typeof] = ACTIONS(2788),
    [anon_sym_ATimport] = ACTIONS(2790),
    [aux_sym_preproc_undef_token1] = ACTIONS(2788),
    [anon_sym_POUND] = ACTIONS(2788),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2788),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2788),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2788),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2788),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2788),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2788),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2788),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_API_AVAILABLE] = ACTIONS(2788),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_API_DEPRECATED] = ACTIONS(2788),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2788),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym___deprecated_msg] = ACTIONS(2788),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2788),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2788),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2788),
    [anon_sym_ATprotocol] = ACTIONS(2790),
    [anon_sym_ATinterface] = ACTIONS(2790),
    [anon_sym_ATimplementation] = ACTIONS(2790),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2790),
    [anon_sym_ATtry] = ACTIONS(2790),
    [anon_sym___try] = ACTIONS(2788),
    [anon_sym_ATthrow] = ACTIONS(2790),
    [anon_sym_ATselector] = ACTIONS(2790),
    [anon_sym_ATavailable] = ACTIONS(2790),
    [anon_sym___builtin_available] = ACTIONS(2788),
    [anon_sym_va_arg] = ACTIONS(2788),
    [anon_sym_ATencode] = ACTIONS(2790),
    [anon_sym_ATsynchronized] = ACTIONS(2790),
    [anon_sym_BOOL] = ACTIONS(2788),
    [anon_sym_IMP] = ACTIONS(2788),
    [anon_sym_SEL] = ACTIONS(2788),
    [anon_sym_Class] = ACTIONS(2788),
    [anon_sym_id] = ACTIONS(2788),
  },
  [STATE(865)] = {
    [sym_identifier] = ACTIONS(2740),
    [aux_sym_preproc_include_token1] = ACTIONS(2740),
    [aux_sym_preproc_include_token2] = ACTIONS(2740),
    [aux_sym_preproc_def_token1] = ACTIONS(2740),
    [aux_sym_preproc_if_token1] = ACTIONS(2740),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2740),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2740),
    [sym_preproc_directive] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_BANG] = ACTIONS(2742),
    [anon_sym_TILDE] = ACTIONS(2742),
    [anon_sym_DASH] = ACTIONS(2740),
    [anon_sym_PLUS] = ACTIONS(2740),
    [anon_sym_STAR] = ACTIONS(2742),
    [anon_sym_CARET] = ACTIONS(2742),
    [anon_sym_AMP] = ACTIONS(2742),
    [anon_sym_SEMI] = ACTIONS(2742),
    [anon_sym___extension__] = ACTIONS(2740),
    [anon_sym_typedef] = ACTIONS(2740),
    [anon_sym_extern] = ACTIONS(2740),
    [anon_sym___attribute__] = ACTIONS(2740),
    [anon_sym___attribute] = ACTIONS(2740),
    [anon_sym_noreturn] = ACTIONS(2740),
    [anon_sym_LBRACK] = ACTIONS(2742),
    [anon_sym___declspec] = ACTIONS(2740),
    [anon_sym___cdecl] = ACTIONS(2740),
    [anon_sym___clrcall] = ACTIONS(2740),
    [anon_sym___stdcall] = ACTIONS(2740),
    [anon_sym___fastcall] = ACTIONS(2740),
    [anon_sym___thiscall] = ACTIONS(2740),
    [anon_sym___vectorcall] = ACTIONS(2740),
    [anon_sym_LBRACE] = ACTIONS(2742),
    [anon_sym_RBRACE] = ACTIONS(2742),
    [anon_sym_signed] = ACTIONS(2740),
    [anon_sym_unsigned] = ACTIONS(2740),
    [anon_sym_long] = ACTIONS(2740),
    [anon_sym_short] = ACTIONS(2740),
    [anon_sym_static] = ACTIONS(2740),
    [anon_sym_ATautoreleasepool] = ACTIONS(2742),
    [anon_sym_auto] = ACTIONS(2740),
    [anon_sym_register] = ACTIONS(2740),
    [anon_sym_inline] = ACTIONS(2740),
    [anon_sym___inline] = ACTIONS(2740),
    [anon_sym___inline__] = ACTIONS(2740),
    [anon_sym___forceinline] = ACTIONS(2740),
    [anon_sym_thread_local] = ACTIONS(2740),
    [anon_sym___thread] = ACTIONS(2740),
    [anon_sym_CG_EXTERN] = ACTIONS(2740),
    [anon_sym_CG_INLINE] = ACTIONS(2740),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2740),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2740),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2740),
    [anon_sym_IBOutlet] = ACTIONS(2740),
    [anon_sym_IBInspectable] = ACTIONS(2740),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2740),
    [anon_sym_NS_INLINE] = ACTIONS(2740),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2740),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2740),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2740),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2740),
    [anon_sym_const] = ACTIONS(2740),
    [anon_sym_constexpr] = ACTIONS(2740),
    [anon_sym_volatile] = ACTIONS(2740),
    [anon_sym_restrict] = ACTIONS(2740),
    [anon_sym___restrict__] = ACTIONS(2740),
    [anon_sym__Atomic] = ACTIONS(2740),
    [anon_sym__Noreturn] = ACTIONS(2740),
    [anon_sym__Nonnull] = ACTIONS(2740),
    [anon_sym_nullable] = ACTIONS(2740),
    [anon_sym__Complex] = ACTIONS(2740),
    [anon_sym__Nullable] = ACTIONS(2740),
    [anon_sym__Nullable_result] = ACTIONS(2740),
    [anon_sym__Null_unspecified] = ACTIONS(2740),
    [anon_sym___autoreleasing] = ACTIONS(2740),
    [anon_sym___block] = ACTIONS(2740),
    [anon_sym___bridge] = ACTIONS(2740),
    [anon_sym___bridge_retained] = ACTIONS(2740),
    [anon_sym___bridge_transfer] = ACTIONS(2740),
    [anon_sym___complex] = ACTIONS(2740),
    [anon_sym___const] = ACTIONS(2740),
    [anon_sym___imag] = ACTIONS(2740),
    [anon_sym___kindof] = ACTIONS(2740),
    [anon_sym___nonnull] = ACTIONS(2740),
    [anon_sym___nullable] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2740),
    [anon_sym___real] = ACTIONS(2740),
    [anon_sym___strong] = ACTIONS(2740),
    [anon_sym___unsafe_unretained] = ACTIONS(2740),
    [anon_sym___unused] = ACTIONS(2740),
    [anon_sym___weak] = ACTIONS(2740),
    [anon_sym_alignas] = ACTIONS(2740),
    [anon_sym__Alignas] = ACTIONS(2740),
    [sym_primitive_type] = ACTIONS(2740),
    [anon_sym_enum] = ACTIONS(2740),
    [anon_sym_struct] = ACTIONS(2740),
    [anon_sym_union] = ACTIONS(2740),
    [anon_sym_if] = ACTIONS(2740),
    [anon_sym_switch] = ACTIONS(2740),
    [anon_sym_case] = ACTIONS(2740),
    [anon_sym_default] = ACTIONS(2740),
    [anon_sym_while] = ACTIONS(2740),
    [anon_sym_do] = ACTIONS(2740),
    [anon_sym_for] = ACTIONS(2740),
    [anon_sym_in] = ACTIONS(2740),
    [anon_sym_return] = ACTIONS(2740),
    [anon_sym_break] = ACTIONS(2740),
    [anon_sym_continue] = ACTIONS(2740),
    [anon_sym_goto] = ACTIONS(2740),
    [anon_sym_DASH_DASH] = ACTIONS(2742),
    [anon_sym_PLUS_PLUS] = ACTIONS(2742),
    [anon_sym_sizeof] = ACTIONS(2740),
    [anon_sym___alignof__] = ACTIONS(2740),
    [anon_sym___alignof] = ACTIONS(2740),
    [anon_sym__alignof] = ACTIONS(2740),
    [anon_sym_alignof] = ACTIONS(2740),
    [anon_sym__Alignof] = ACTIONS(2740),
    [anon_sym_offsetof] = ACTIONS(2740),
    [anon_sym__Generic] = ACTIONS(2740),
    [anon_sym_asm] = ACTIONS(2740),
    [anon_sym___asm__] = ACTIONS(2740),
    [anon_sym___asm] = ACTIONS(2740),
    [sym_number_literal] = ACTIONS(2742),
    [anon_sym_L_SQUOTE] = ACTIONS(2742),
    [anon_sym_u_SQUOTE] = ACTIONS(2742),
    [anon_sym_U_SQUOTE] = ACTIONS(2742),
    [anon_sym_u8_SQUOTE] = ACTIONS(2742),
    [anon_sym_SQUOTE] = ACTIONS(2742),
    [anon_sym_AT] = ACTIONS(2740),
    [anon_sym_DQUOTE] = ACTIONS(2742),
    [anon_sym_L_DQUOTE] = ACTIONS(2742),
    [anon_sym_u_DQUOTE] = ACTIONS(2742),
    [anon_sym_U_DQUOTE] = ACTIONS(2742),
    [anon_sym_u8_DQUOTE] = ACTIONS(2742),
    [sym_true] = ACTIONS(2740),
    [sym_false] = ACTIONS(2740),
    [anon_sym_NULL] = ACTIONS(2740),
    [anon_sym_nullptr] = ACTIONS(2740),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2740),
    [anon_sym___typeof] = ACTIONS(2740),
    [anon_sym_typeof] = ACTIONS(2740),
    [anon_sym_ATimport] = ACTIONS(2742),
    [aux_sym_preproc_undef_token1] = ACTIONS(2740),
    [anon_sym_POUND] = ACTIONS(2740),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2740),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2740),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2740),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2740),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2740),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2740),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2740),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_API_AVAILABLE] = ACTIONS(2740),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_API_DEPRECATED] = ACTIONS(2740),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2740),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym___deprecated_msg] = ACTIONS(2740),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2740),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2740),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2740),
    [anon_sym_ATprotocol] = ACTIONS(2742),
    [anon_sym_ATinterface] = ACTIONS(2742),
    [anon_sym_ATimplementation] = ACTIONS(2742),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2742),
    [anon_sym_ATtry] = ACTIONS(2742),
    [anon_sym___try] = ACTIONS(2740),
    [anon_sym_ATthrow] = ACTIONS(2742),
    [anon_sym_ATselector] = ACTIONS(2742),
    [anon_sym_ATavailable] = ACTIONS(2742),
    [anon_sym___builtin_available] = ACTIONS(2740),
    [anon_sym_va_arg] = ACTIONS(2740),
    [anon_sym_ATencode] = ACTIONS(2742),
    [anon_sym_ATsynchronized] = ACTIONS(2742),
    [anon_sym_BOOL] = ACTIONS(2740),
    [anon_sym_IMP] = ACTIONS(2740),
    [anon_sym_SEL] = ACTIONS(2740),
    [anon_sym_Class] = ACTIONS(2740),
    [anon_sym_id] = ACTIONS(2740),
  },
  [STATE(866)] = {
    [sym_identifier] = ACTIONS(2642),
    [aux_sym_preproc_include_token1] = ACTIONS(2642),
    [aux_sym_preproc_include_token2] = ACTIONS(2642),
    [aux_sym_preproc_def_token1] = ACTIONS(2642),
    [aux_sym_preproc_if_token1] = ACTIONS(2642),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2642),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2642),
    [sym_preproc_directive] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_BANG] = ACTIONS(2644),
    [anon_sym_TILDE] = ACTIONS(2644),
    [anon_sym_DASH] = ACTIONS(2642),
    [anon_sym_PLUS] = ACTIONS(2642),
    [anon_sym_STAR] = ACTIONS(2644),
    [anon_sym_CARET] = ACTIONS(2644),
    [anon_sym_AMP] = ACTIONS(2644),
    [anon_sym_SEMI] = ACTIONS(2644),
    [anon_sym___extension__] = ACTIONS(2642),
    [anon_sym_typedef] = ACTIONS(2642),
    [anon_sym_extern] = ACTIONS(2642),
    [anon_sym___attribute__] = ACTIONS(2642),
    [anon_sym___attribute] = ACTIONS(2642),
    [anon_sym_noreturn] = ACTIONS(2642),
    [anon_sym_LBRACK] = ACTIONS(2644),
    [anon_sym___declspec] = ACTIONS(2642),
    [anon_sym___cdecl] = ACTIONS(2642),
    [anon_sym___clrcall] = ACTIONS(2642),
    [anon_sym___stdcall] = ACTIONS(2642),
    [anon_sym___fastcall] = ACTIONS(2642),
    [anon_sym___thiscall] = ACTIONS(2642),
    [anon_sym___vectorcall] = ACTIONS(2642),
    [anon_sym_LBRACE] = ACTIONS(2644),
    [anon_sym_RBRACE] = ACTIONS(2644),
    [anon_sym_signed] = ACTIONS(2642),
    [anon_sym_unsigned] = ACTIONS(2642),
    [anon_sym_long] = ACTIONS(2642),
    [anon_sym_short] = ACTIONS(2642),
    [anon_sym_static] = ACTIONS(2642),
    [anon_sym_ATautoreleasepool] = ACTIONS(2644),
    [anon_sym_auto] = ACTIONS(2642),
    [anon_sym_register] = ACTIONS(2642),
    [anon_sym_inline] = ACTIONS(2642),
    [anon_sym___inline] = ACTIONS(2642),
    [anon_sym___inline__] = ACTIONS(2642),
    [anon_sym___forceinline] = ACTIONS(2642),
    [anon_sym_thread_local] = ACTIONS(2642),
    [anon_sym___thread] = ACTIONS(2642),
    [anon_sym_CG_EXTERN] = ACTIONS(2642),
    [anon_sym_CG_INLINE] = ACTIONS(2642),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2642),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2642),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2642),
    [anon_sym_IBOutlet] = ACTIONS(2642),
    [anon_sym_IBInspectable] = ACTIONS(2642),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2642),
    [anon_sym_NS_INLINE] = ACTIONS(2642),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2642),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2642),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2642),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2642),
    [anon_sym_const] = ACTIONS(2642),
    [anon_sym_constexpr] = ACTIONS(2642),
    [anon_sym_volatile] = ACTIONS(2642),
    [anon_sym_restrict] = ACTIONS(2642),
    [anon_sym___restrict__] = ACTIONS(2642),
    [anon_sym__Atomic] = ACTIONS(2642),
    [anon_sym__Noreturn] = ACTIONS(2642),
    [anon_sym__Nonnull] = ACTIONS(2642),
    [anon_sym_nullable] = ACTIONS(2642),
    [anon_sym__Complex] = ACTIONS(2642),
    [anon_sym__Nullable] = ACTIONS(2642),
    [anon_sym__Nullable_result] = ACTIONS(2642),
    [anon_sym__Null_unspecified] = ACTIONS(2642),
    [anon_sym___autoreleasing] = ACTIONS(2642),
    [anon_sym___block] = ACTIONS(2642),
    [anon_sym___bridge] = ACTIONS(2642),
    [anon_sym___bridge_retained] = ACTIONS(2642),
    [anon_sym___bridge_transfer] = ACTIONS(2642),
    [anon_sym___complex] = ACTIONS(2642),
    [anon_sym___const] = ACTIONS(2642),
    [anon_sym___imag] = ACTIONS(2642),
    [anon_sym___kindof] = ACTIONS(2642),
    [anon_sym___nonnull] = ACTIONS(2642),
    [anon_sym___nullable] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2642),
    [anon_sym___real] = ACTIONS(2642),
    [anon_sym___strong] = ACTIONS(2642),
    [anon_sym___unsafe_unretained] = ACTIONS(2642),
    [anon_sym___unused] = ACTIONS(2642),
    [anon_sym___weak] = ACTIONS(2642),
    [anon_sym_alignas] = ACTIONS(2642),
    [anon_sym__Alignas] = ACTIONS(2642),
    [sym_primitive_type] = ACTIONS(2642),
    [anon_sym_enum] = ACTIONS(2642),
    [anon_sym_struct] = ACTIONS(2642),
    [anon_sym_union] = ACTIONS(2642),
    [anon_sym_if] = ACTIONS(2642),
    [anon_sym_switch] = ACTIONS(2642),
    [anon_sym_case] = ACTIONS(2642),
    [anon_sym_default] = ACTIONS(2642),
    [anon_sym_while] = ACTIONS(2642),
    [anon_sym_do] = ACTIONS(2642),
    [anon_sym_for] = ACTIONS(2642),
    [anon_sym_in] = ACTIONS(2642),
    [anon_sym_return] = ACTIONS(2642),
    [anon_sym_break] = ACTIONS(2642),
    [anon_sym_continue] = ACTIONS(2642),
    [anon_sym_goto] = ACTIONS(2642),
    [anon_sym_DASH_DASH] = ACTIONS(2644),
    [anon_sym_PLUS_PLUS] = ACTIONS(2644),
    [anon_sym_sizeof] = ACTIONS(2642),
    [anon_sym___alignof__] = ACTIONS(2642),
    [anon_sym___alignof] = ACTIONS(2642),
    [anon_sym__alignof] = ACTIONS(2642),
    [anon_sym_alignof] = ACTIONS(2642),
    [anon_sym__Alignof] = ACTIONS(2642),
    [anon_sym_offsetof] = ACTIONS(2642),
    [anon_sym__Generic] = ACTIONS(2642),
    [anon_sym_asm] = ACTIONS(2642),
    [anon_sym___asm__] = ACTIONS(2642),
    [anon_sym___asm] = ACTIONS(2642),
    [sym_number_literal] = ACTIONS(2644),
    [anon_sym_L_SQUOTE] = ACTIONS(2644),
    [anon_sym_u_SQUOTE] = ACTIONS(2644),
    [anon_sym_U_SQUOTE] = ACTIONS(2644),
    [anon_sym_u8_SQUOTE] = ACTIONS(2644),
    [anon_sym_SQUOTE] = ACTIONS(2644),
    [anon_sym_AT] = ACTIONS(2642),
    [anon_sym_DQUOTE] = ACTIONS(2644),
    [anon_sym_L_DQUOTE] = ACTIONS(2644),
    [anon_sym_u_DQUOTE] = ACTIONS(2644),
    [anon_sym_U_DQUOTE] = ACTIONS(2644),
    [anon_sym_u8_DQUOTE] = ACTIONS(2644),
    [sym_true] = ACTIONS(2642),
    [sym_false] = ACTIONS(2642),
    [anon_sym_NULL] = ACTIONS(2642),
    [anon_sym_nullptr] = ACTIONS(2642),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2642),
    [anon_sym___typeof] = ACTIONS(2642),
    [anon_sym_typeof] = ACTIONS(2642),
    [anon_sym_ATimport] = ACTIONS(2644),
    [aux_sym_preproc_undef_token1] = ACTIONS(2642),
    [anon_sym_POUND] = ACTIONS(2642),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2642),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2642),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2642),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2642),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2642),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2642),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2642),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_API_AVAILABLE] = ACTIONS(2642),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_API_DEPRECATED] = ACTIONS(2642),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2642),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym___deprecated_msg] = ACTIONS(2642),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2642),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2642),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2642),
    [anon_sym_ATprotocol] = ACTIONS(2644),
    [anon_sym_ATinterface] = ACTIONS(2644),
    [anon_sym_ATimplementation] = ACTIONS(2644),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2644),
    [anon_sym_ATtry] = ACTIONS(2644),
    [anon_sym___try] = ACTIONS(2642),
    [anon_sym_ATthrow] = ACTIONS(2644),
    [anon_sym_ATselector] = ACTIONS(2644),
    [anon_sym_ATavailable] = ACTIONS(2644),
    [anon_sym___builtin_available] = ACTIONS(2642),
    [anon_sym_va_arg] = ACTIONS(2642),
    [anon_sym_ATencode] = ACTIONS(2644),
    [anon_sym_ATsynchronized] = ACTIONS(2644),
    [anon_sym_BOOL] = ACTIONS(2642),
    [anon_sym_IMP] = ACTIONS(2642),
    [anon_sym_SEL] = ACTIONS(2642),
    [anon_sym_Class] = ACTIONS(2642),
    [anon_sym_id] = ACTIONS(2642),
  },
  [STATE(867)] = {
    [sym_identifier] = ACTIONS(2646),
    [aux_sym_preproc_include_token1] = ACTIONS(2646),
    [aux_sym_preproc_include_token2] = ACTIONS(2646),
    [aux_sym_preproc_def_token1] = ACTIONS(2646),
    [aux_sym_preproc_if_token1] = ACTIONS(2646),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2646),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2646),
    [sym_preproc_directive] = ACTIONS(2646),
    [anon_sym_LPAREN2] = ACTIONS(2648),
    [anon_sym_BANG] = ACTIONS(2648),
    [anon_sym_TILDE] = ACTIONS(2648),
    [anon_sym_DASH] = ACTIONS(2646),
    [anon_sym_PLUS] = ACTIONS(2646),
    [anon_sym_STAR] = ACTIONS(2648),
    [anon_sym_CARET] = ACTIONS(2648),
    [anon_sym_AMP] = ACTIONS(2648),
    [anon_sym_SEMI] = ACTIONS(2648),
    [anon_sym___extension__] = ACTIONS(2646),
    [anon_sym_typedef] = ACTIONS(2646),
    [anon_sym_extern] = ACTIONS(2646),
    [anon_sym___attribute__] = ACTIONS(2646),
    [anon_sym___attribute] = ACTIONS(2646),
    [anon_sym_noreturn] = ACTIONS(2646),
    [anon_sym_LBRACK] = ACTIONS(2648),
    [anon_sym___declspec] = ACTIONS(2646),
    [anon_sym___cdecl] = ACTIONS(2646),
    [anon_sym___clrcall] = ACTIONS(2646),
    [anon_sym___stdcall] = ACTIONS(2646),
    [anon_sym___fastcall] = ACTIONS(2646),
    [anon_sym___thiscall] = ACTIONS(2646),
    [anon_sym___vectorcall] = ACTIONS(2646),
    [anon_sym_LBRACE] = ACTIONS(2648),
    [anon_sym_RBRACE] = ACTIONS(2648),
    [anon_sym_signed] = ACTIONS(2646),
    [anon_sym_unsigned] = ACTIONS(2646),
    [anon_sym_long] = ACTIONS(2646),
    [anon_sym_short] = ACTIONS(2646),
    [anon_sym_static] = ACTIONS(2646),
    [anon_sym_ATautoreleasepool] = ACTIONS(2648),
    [anon_sym_auto] = ACTIONS(2646),
    [anon_sym_register] = ACTIONS(2646),
    [anon_sym_inline] = ACTIONS(2646),
    [anon_sym___inline] = ACTIONS(2646),
    [anon_sym___inline__] = ACTIONS(2646),
    [anon_sym___forceinline] = ACTIONS(2646),
    [anon_sym_thread_local] = ACTIONS(2646),
    [anon_sym___thread] = ACTIONS(2646),
    [anon_sym_CG_EXTERN] = ACTIONS(2646),
    [anon_sym_CG_INLINE] = ACTIONS(2646),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2646),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2646),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2646),
    [anon_sym_IBOutlet] = ACTIONS(2646),
    [anon_sym_IBInspectable] = ACTIONS(2646),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2646),
    [anon_sym_NS_INLINE] = ACTIONS(2646),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2646),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2646),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2646),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2646),
    [anon_sym_const] = ACTIONS(2646),
    [anon_sym_constexpr] = ACTIONS(2646),
    [anon_sym_volatile] = ACTIONS(2646),
    [anon_sym_restrict] = ACTIONS(2646),
    [anon_sym___restrict__] = ACTIONS(2646),
    [anon_sym__Atomic] = ACTIONS(2646),
    [anon_sym__Noreturn] = ACTIONS(2646),
    [anon_sym__Nonnull] = ACTIONS(2646),
    [anon_sym_nullable] = ACTIONS(2646),
    [anon_sym__Complex] = ACTIONS(2646),
    [anon_sym__Nullable] = ACTIONS(2646),
    [anon_sym__Nullable_result] = ACTIONS(2646),
    [anon_sym__Null_unspecified] = ACTIONS(2646),
    [anon_sym___autoreleasing] = ACTIONS(2646),
    [anon_sym___block] = ACTIONS(2646),
    [anon_sym___bridge] = ACTIONS(2646),
    [anon_sym___bridge_retained] = ACTIONS(2646),
    [anon_sym___bridge_transfer] = ACTIONS(2646),
    [anon_sym___complex] = ACTIONS(2646),
    [anon_sym___const] = ACTIONS(2646),
    [anon_sym___imag] = ACTIONS(2646),
    [anon_sym___kindof] = ACTIONS(2646),
    [anon_sym___nonnull] = ACTIONS(2646),
    [anon_sym___nullable] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2646),
    [anon_sym___real] = ACTIONS(2646),
    [anon_sym___strong] = ACTIONS(2646),
    [anon_sym___unsafe_unretained] = ACTIONS(2646),
    [anon_sym___unused] = ACTIONS(2646),
    [anon_sym___weak] = ACTIONS(2646),
    [anon_sym_alignas] = ACTIONS(2646),
    [anon_sym__Alignas] = ACTIONS(2646),
    [sym_primitive_type] = ACTIONS(2646),
    [anon_sym_enum] = ACTIONS(2646),
    [anon_sym_struct] = ACTIONS(2646),
    [anon_sym_union] = ACTIONS(2646),
    [anon_sym_if] = ACTIONS(2646),
    [anon_sym_switch] = ACTIONS(2646),
    [anon_sym_case] = ACTIONS(2646),
    [anon_sym_default] = ACTIONS(2646),
    [anon_sym_while] = ACTIONS(2646),
    [anon_sym_do] = ACTIONS(2646),
    [anon_sym_for] = ACTIONS(2646),
    [anon_sym_in] = ACTIONS(2646),
    [anon_sym_return] = ACTIONS(2646),
    [anon_sym_break] = ACTIONS(2646),
    [anon_sym_continue] = ACTIONS(2646),
    [anon_sym_goto] = ACTIONS(2646),
    [anon_sym_DASH_DASH] = ACTIONS(2648),
    [anon_sym_PLUS_PLUS] = ACTIONS(2648),
    [anon_sym_sizeof] = ACTIONS(2646),
    [anon_sym___alignof__] = ACTIONS(2646),
    [anon_sym___alignof] = ACTIONS(2646),
    [anon_sym__alignof] = ACTIONS(2646),
    [anon_sym_alignof] = ACTIONS(2646),
    [anon_sym__Alignof] = ACTIONS(2646),
    [anon_sym_offsetof] = ACTIONS(2646),
    [anon_sym__Generic] = ACTIONS(2646),
    [anon_sym_asm] = ACTIONS(2646),
    [anon_sym___asm__] = ACTIONS(2646),
    [anon_sym___asm] = ACTIONS(2646),
    [sym_number_literal] = ACTIONS(2648),
    [anon_sym_L_SQUOTE] = ACTIONS(2648),
    [anon_sym_u_SQUOTE] = ACTIONS(2648),
    [anon_sym_U_SQUOTE] = ACTIONS(2648),
    [anon_sym_u8_SQUOTE] = ACTIONS(2648),
    [anon_sym_SQUOTE] = ACTIONS(2648),
    [anon_sym_AT] = ACTIONS(2646),
    [anon_sym_DQUOTE] = ACTIONS(2648),
    [anon_sym_L_DQUOTE] = ACTIONS(2648),
    [anon_sym_u_DQUOTE] = ACTIONS(2648),
    [anon_sym_U_DQUOTE] = ACTIONS(2648),
    [anon_sym_u8_DQUOTE] = ACTIONS(2648),
    [sym_true] = ACTIONS(2646),
    [sym_false] = ACTIONS(2646),
    [anon_sym_NULL] = ACTIONS(2646),
    [anon_sym_nullptr] = ACTIONS(2646),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2646),
    [anon_sym___typeof] = ACTIONS(2646),
    [anon_sym_typeof] = ACTIONS(2646),
    [anon_sym_ATimport] = ACTIONS(2648),
    [aux_sym_preproc_undef_token1] = ACTIONS(2646),
    [anon_sym_POUND] = ACTIONS(2646),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2646),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2646),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2646),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2646),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2646),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2646),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2646),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_API_AVAILABLE] = ACTIONS(2646),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_API_DEPRECATED] = ACTIONS(2646),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2646),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym___deprecated_msg] = ACTIONS(2646),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2646),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2646),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2646),
    [anon_sym_ATprotocol] = ACTIONS(2648),
    [anon_sym_ATinterface] = ACTIONS(2648),
    [anon_sym_ATimplementation] = ACTIONS(2648),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2648),
    [anon_sym_ATtry] = ACTIONS(2648),
    [anon_sym___try] = ACTIONS(2646),
    [anon_sym_ATthrow] = ACTIONS(2648),
    [anon_sym_ATselector] = ACTIONS(2648),
    [anon_sym_ATavailable] = ACTIONS(2648),
    [anon_sym___builtin_available] = ACTIONS(2646),
    [anon_sym_va_arg] = ACTIONS(2646),
    [anon_sym_ATencode] = ACTIONS(2648),
    [anon_sym_ATsynchronized] = ACTIONS(2648),
    [anon_sym_BOOL] = ACTIONS(2646),
    [anon_sym_IMP] = ACTIONS(2646),
    [anon_sym_SEL] = ACTIONS(2646),
    [anon_sym_Class] = ACTIONS(2646),
    [anon_sym_id] = ACTIONS(2646),
  },
  [STATE(868)] = {
    [sym_identifier] = ACTIONS(2588),
    [aux_sym_preproc_include_token1] = ACTIONS(2588),
    [aux_sym_preproc_include_token2] = ACTIONS(2588),
    [aux_sym_preproc_def_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2588),
    [sym_preproc_directive] = ACTIONS(2588),
    [anon_sym_LPAREN2] = ACTIONS(2590),
    [anon_sym_BANG] = ACTIONS(2590),
    [anon_sym_TILDE] = ACTIONS(2590),
    [anon_sym_DASH] = ACTIONS(2588),
    [anon_sym_PLUS] = ACTIONS(2588),
    [anon_sym_STAR] = ACTIONS(2590),
    [anon_sym_CARET] = ACTIONS(2590),
    [anon_sym_AMP] = ACTIONS(2590),
    [anon_sym_SEMI] = ACTIONS(2590),
    [anon_sym___extension__] = ACTIONS(2588),
    [anon_sym_typedef] = ACTIONS(2588),
    [anon_sym_extern] = ACTIONS(2588),
    [anon_sym___attribute__] = ACTIONS(2588),
    [anon_sym___attribute] = ACTIONS(2588),
    [anon_sym_noreturn] = ACTIONS(2588),
    [anon_sym_LBRACK] = ACTIONS(2590),
    [anon_sym___declspec] = ACTIONS(2588),
    [anon_sym___cdecl] = ACTIONS(2588),
    [anon_sym___clrcall] = ACTIONS(2588),
    [anon_sym___stdcall] = ACTIONS(2588),
    [anon_sym___fastcall] = ACTIONS(2588),
    [anon_sym___thiscall] = ACTIONS(2588),
    [anon_sym___vectorcall] = ACTIONS(2588),
    [anon_sym_LBRACE] = ACTIONS(2590),
    [anon_sym_RBRACE] = ACTIONS(2590),
    [anon_sym_signed] = ACTIONS(2588),
    [anon_sym_unsigned] = ACTIONS(2588),
    [anon_sym_long] = ACTIONS(2588),
    [anon_sym_short] = ACTIONS(2588),
    [anon_sym_static] = ACTIONS(2588),
    [anon_sym_ATautoreleasepool] = ACTIONS(2590),
    [anon_sym_auto] = ACTIONS(2588),
    [anon_sym_register] = ACTIONS(2588),
    [anon_sym_inline] = ACTIONS(2588),
    [anon_sym___inline] = ACTIONS(2588),
    [anon_sym___inline__] = ACTIONS(2588),
    [anon_sym___forceinline] = ACTIONS(2588),
    [anon_sym_thread_local] = ACTIONS(2588),
    [anon_sym___thread] = ACTIONS(2588),
    [anon_sym_CG_EXTERN] = ACTIONS(2588),
    [anon_sym_CG_INLINE] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2588),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2588),
    [anon_sym_IBOutlet] = ACTIONS(2588),
    [anon_sym_IBInspectable] = ACTIONS(2588),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2588),
    [anon_sym_NS_INLINE] = ACTIONS(2588),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2588),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2588),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2588),
    [anon_sym_const] = ACTIONS(2588),
    [anon_sym_constexpr] = ACTIONS(2588),
    [anon_sym_volatile] = ACTIONS(2588),
    [anon_sym_restrict] = ACTIONS(2588),
    [anon_sym___restrict__] = ACTIONS(2588),
    [anon_sym__Atomic] = ACTIONS(2588),
    [anon_sym__Noreturn] = ACTIONS(2588),
    [anon_sym__Nonnull] = ACTIONS(2588),
    [anon_sym_nullable] = ACTIONS(2588),
    [anon_sym__Complex] = ACTIONS(2588),
    [anon_sym__Nullable] = ACTIONS(2588),
    [anon_sym__Nullable_result] = ACTIONS(2588),
    [anon_sym__Null_unspecified] = ACTIONS(2588),
    [anon_sym___autoreleasing] = ACTIONS(2588),
    [anon_sym___block] = ACTIONS(2588),
    [anon_sym___bridge] = ACTIONS(2588),
    [anon_sym___bridge_retained] = ACTIONS(2588),
    [anon_sym___bridge_transfer] = ACTIONS(2588),
    [anon_sym___complex] = ACTIONS(2588),
    [anon_sym___const] = ACTIONS(2588),
    [anon_sym___imag] = ACTIONS(2588),
    [anon_sym___kindof] = ACTIONS(2588),
    [anon_sym___nonnull] = ACTIONS(2588),
    [anon_sym___nullable] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2588),
    [anon_sym___real] = ACTIONS(2588),
    [anon_sym___strong] = ACTIONS(2588),
    [anon_sym___unsafe_unretained] = ACTIONS(2588),
    [anon_sym___unused] = ACTIONS(2588),
    [anon_sym___weak] = ACTIONS(2588),
    [anon_sym_alignas] = ACTIONS(2588),
    [anon_sym__Alignas] = ACTIONS(2588),
    [sym_primitive_type] = ACTIONS(2588),
    [anon_sym_enum] = ACTIONS(2588),
    [anon_sym_struct] = ACTIONS(2588),
    [anon_sym_union] = ACTIONS(2588),
    [anon_sym_if] = ACTIONS(2588),
    [anon_sym_switch] = ACTIONS(2588),
    [anon_sym_case] = ACTIONS(2588),
    [anon_sym_default] = ACTIONS(2588),
    [anon_sym_while] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2588),
    [anon_sym_for] = ACTIONS(2588),
    [anon_sym_in] = ACTIONS(2588),
    [anon_sym_return] = ACTIONS(2588),
    [anon_sym_break] = ACTIONS(2588),
    [anon_sym_continue] = ACTIONS(2588),
    [anon_sym_goto] = ACTIONS(2588),
    [anon_sym_DASH_DASH] = ACTIONS(2590),
    [anon_sym_PLUS_PLUS] = ACTIONS(2590),
    [anon_sym_sizeof] = ACTIONS(2588),
    [anon_sym___alignof__] = ACTIONS(2588),
    [anon_sym___alignof] = ACTIONS(2588),
    [anon_sym__alignof] = ACTIONS(2588),
    [anon_sym_alignof] = ACTIONS(2588),
    [anon_sym__Alignof] = ACTIONS(2588),
    [anon_sym_offsetof] = ACTIONS(2588),
    [anon_sym__Generic] = ACTIONS(2588),
    [anon_sym_asm] = ACTIONS(2588),
    [anon_sym___asm__] = ACTIONS(2588),
    [anon_sym___asm] = ACTIONS(2588),
    [sym_number_literal] = ACTIONS(2590),
    [anon_sym_L_SQUOTE] = ACTIONS(2590),
    [anon_sym_u_SQUOTE] = ACTIONS(2590),
    [anon_sym_U_SQUOTE] = ACTIONS(2590),
    [anon_sym_u8_SQUOTE] = ACTIONS(2590),
    [anon_sym_SQUOTE] = ACTIONS(2590),
    [anon_sym_AT] = ACTIONS(2588),
    [anon_sym_DQUOTE] = ACTIONS(2590),
    [anon_sym_L_DQUOTE] = ACTIONS(2590),
    [anon_sym_u_DQUOTE] = ACTIONS(2590),
    [anon_sym_U_DQUOTE] = ACTIONS(2590),
    [anon_sym_u8_DQUOTE] = ACTIONS(2590),
    [sym_true] = ACTIONS(2588),
    [sym_false] = ACTIONS(2588),
    [anon_sym_NULL] = ACTIONS(2588),
    [anon_sym_nullptr] = ACTIONS(2588),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2588),
    [anon_sym___typeof] = ACTIONS(2588),
    [anon_sym_typeof] = ACTIONS(2588),
    [anon_sym_ATimport] = ACTIONS(2590),
    [aux_sym_preproc_undef_token1] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2588),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2588),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2588),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2588),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_API_AVAILABLE] = ACTIONS(2588),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_API_DEPRECATED] = ACTIONS(2588),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym___deprecated_msg] = ACTIONS(2588),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2588),
    [anon_sym_ATprotocol] = ACTIONS(2590),
    [anon_sym_ATinterface] = ACTIONS(2590),
    [anon_sym_ATimplementation] = ACTIONS(2590),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2590),
    [anon_sym_ATtry] = ACTIONS(2590),
    [anon_sym___try] = ACTIONS(2588),
    [anon_sym_ATthrow] = ACTIONS(2590),
    [anon_sym_ATselector] = ACTIONS(2590),
    [anon_sym_ATavailable] = ACTIONS(2590),
    [anon_sym___builtin_available] = ACTIONS(2588),
    [anon_sym_va_arg] = ACTIONS(2588),
    [anon_sym_ATencode] = ACTIONS(2590),
    [anon_sym_ATsynchronized] = ACTIONS(2590),
    [anon_sym_BOOL] = ACTIONS(2588),
    [anon_sym_IMP] = ACTIONS(2588),
    [anon_sym_SEL] = ACTIONS(2588),
    [anon_sym_Class] = ACTIONS(2588),
    [anon_sym_id] = ACTIONS(2588),
  },
  [STATE(869)] = {
    [sym_identifier] = ACTIONS(2704),
    [aux_sym_preproc_include_token1] = ACTIONS(2704),
    [aux_sym_preproc_include_token2] = ACTIONS(2704),
    [aux_sym_preproc_def_token1] = ACTIONS(2704),
    [aux_sym_preproc_if_token1] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2704),
    [sym_preproc_directive] = ACTIONS(2704),
    [anon_sym_LPAREN2] = ACTIONS(2706),
    [anon_sym_BANG] = ACTIONS(2706),
    [anon_sym_TILDE] = ACTIONS(2706),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_STAR] = ACTIONS(2706),
    [anon_sym_CARET] = ACTIONS(2706),
    [anon_sym_AMP] = ACTIONS(2706),
    [anon_sym_SEMI] = ACTIONS(2706),
    [anon_sym___extension__] = ACTIONS(2704),
    [anon_sym_typedef] = ACTIONS(2704),
    [anon_sym_extern] = ACTIONS(2704),
    [anon_sym___attribute__] = ACTIONS(2704),
    [anon_sym___attribute] = ACTIONS(2704),
    [anon_sym_noreturn] = ACTIONS(2704),
    [anon_sym_LBRACK] = ACTIONS(2706),
    [anon_sym___declspec] = ACTIONS(2704),
    [anon_sym___cdecl] = ACTIONS(2704),
    [anon_sym___clrcall] = ACTIONS(2704),
    [anon_sym___stdcall] = ACTIONS(2704),
    [anon_sym___fastcall] = ACTIONS(2704),
    [anon_sym___thiscall] = ACTIONS(2704),
    [anon_sym___vectorcall] = ACTIONS(2704),
    [anon_sym_LBRACE] = ACTIONS(2706),
    [anon_sym_RBRACE] = ACTIONS(2706),
    [anon_sym_signed] = ACTIONS(2704),
    [anon_sym_unsigned] = ACTIONS(2704),
    [anon_sym_long] = ACTIONS(2704),
    [anon_sym_short] = ACTIONS(2704),
    [anon_sym_static] = ACTIONS(2704),
    [anon_sym_ATautoreleasepool] = ACTIONS(2706),
    [anon_sym_auto] = ACTIONS(2704),
    [anon_sym_register] = ACTIONS(2704),
    [anon_sym_inline] = ACTIONS(2704),
    [anon_sym___inline] = ACTIONS(2704),
    [anon_sym___inline__] = ACTIONS(2704),
    [anon_sym___forceinline] = ACTIONS(2704),
    [anon_sym_thread_local] = ACTIONS(2704),
    [anon_sym___thread] = ACTIONS(2704),
    [anon_sym_CG_EXTERN] = ACTIONS(2704),
    [anon_sym_CG_INLINE] = ACTIONS(2704),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2704),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2704),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2704),
    [anon_sym_IBOutlet] = ACTIONS(2704),
    [anon_sym_IBInspectable] = ACTIONS(2704),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2704),
    [anon_sym_NS_INLINE] = ACTIONS(2704),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2704),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2704),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2704),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2704),
    [anon_sym_const] = ACTIONS(2704),
    [anon_sym_constexpr] = ACTIONS(2704),
    [anon_sym_volatile] = ACTIONS(2704),
    [anon_sym_restrict] = ACTIONS(2704),
    [anon_sym___restrict__] = ACTIONS(2704),
    [anon_sym__Atomic] = ACTIONS(2704),
    [anon_sym__Noreturn] = ACTIONS(2704),
    [anon_sym__Nonnull] = ACTIONS(2704),
    [anon_sym_nullable] = ACTIONS(2704),
    [anon_sym__Complex] = ACTIONS(2704),
    [anon_sym__Nullable] = ACTIONS(2704),
    [anon_sym__Nullable_result] = ACTIONS(2704),
    [anon_sym__Null_unspecified] = ACTIONS(2704),
    [anon_sym___autoreleasing] = ACTIONS(2704),
    [anon_sym___block] = ACTIONS(2704),
    [anon_sym___bridge] = ACTIONS(2704),
    [anon_sym___bridge_retained] = ACTIONS(2704),
    [anon_sym___bridge_transfer] = ACTIONS(2704),
    [anon_sym___complex] = ACTIONS(2704),
    [anon_sym___const] = ACTIONS(2704),
    [anon_sym___imag] = ACTIONS(2704),
    [anon_sym___kindof] = ACTIONS(2704),
    [anon_sym___nonnull] = ACTIONS(2704),
    [anon_sym___nullable] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2704),
    [anon_sym___real] = ACTIONS(2704),
    [anon_sym___strong] = ACTIONS(2704),
    [anon_sym___unsafe_unretained] = ACTIONS(2704),
    [anon_sym___unused] = ACTIONS(2704),
    [anon_sym___weak] = ACTIONS(2704),
    [anon_sym_alignas] = ACTIONS(2704),
    [anon_sym__Alignas] = ACTIONS(2704),
    [sym_primitive_type] = ACTIONS(2704),
    [anon_sym_enum] = ACTIONS(2704),
    [anon_sym_struct] = ACTIONS(2704),
    [anon_sym_union] = ACTIONS(2704),
    [anon_sym_if] = ACTIONS(2704),
    [anon_sym_switch] = ACTIONS(2704),
    [anon_sym_case] = ACTIONS(2704),
    [anon_sym_default] = ACTIONS(2704),
    [anon_sym_while] = ACTIONS(2704),
    [anon_sym_do] = ACTIONS(2704),
    [anon_sym_for] = ACTIONS(2704),
    [anon_sym_in] = ACTIONS(2704),
    [anon_sym_return] = ACTIONS(2704),
    [anon_sym_break] = ACTIONS(2704),
    [anon_sym_continue] = ACTIONS(2704),
    [anon_sym_goto] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2706),
    [anon_sym_PLUS_PLUS] = ACTIONS(2706),
    [anon_sym_sizeof] = ACTIONS(2704),
    [anon_sym___alignof__] = ACTIONS(2704),
    [anon_sym___alignof] = ACTIONS(2704),
    [anon_sym__alignof] = ACTIONS(2704),
    [anon_sym_alignof] = ACTIONS(2704),
    [anon_sym__Alignof] = ACTIONS(2704),
    [anon_sym_offsetof] = ACTIONS(2704),
    [anon_sym__Generic] = ACTIONS(2704),
    [anon_sym_asm] = ACTIONS(2704),
    [anon_sym___asm__] = ACTIONS(2704),
    [anon_sym___asm] = ACTIONS(2704),
    [sym_number_literal] = ACTIONS(2706),
    [anon_sym_L_SQUOTE] = ACTIONS(2706),
    [anon_sym_u_SQUOTE] = ACTIONS(2706),
    [anon_sym_U_SQUOTE] = ACTIONS(2706),
    [anon_sym_u8_SQUOTE] = ACTIONS(2706),
    [anon_sym_SQUOTE] = ACTIONS(2706),
    [anon_sym_AT] = ACTIONS(2704),
    [anon_sym_DQUOTE] = ACTIONS(2706),
    [anon_sym_L_DQUOTE] = ACTIONS(2706),
    [anon_sym_u_DQUOTE] = ACTIONS(2706),
    [anon_sym_U_DQUOTE] = ACTIONS(2706),
    [anon_sym_u8_DQUOTE] = ACTIONS(2706),
    [sym_true] = ACTIONS(2704),
    [sym_false] = ACTIONS(2704),
    [anon_sym_NULL] = ACTIONS(2704),
    [anon_sym_nullptr] = ACTIONS(2704),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2704),
    [anon_sym___typeof] = ACTIONS(2704),
    [anon_sym_typeof] = ACTIONS(2704),
    [anon_sym_ATimport] = ACTIONS(2706),
    [aux_sym_preproc_undef_token1] = ACTIONS(2704),
    [anon_sym_POUND] = ACTIONS(2704),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2704),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2704),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2704),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2704),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2704),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2704),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2704),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_API_AVAILABLE] = ACTIONS(2704),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_API_DEPRECATED] = ACTIONS(2704),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2704),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym___deprecated_msg] = ACTIONS(2704),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2704),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2704),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2704),
    [anon_sym_ATprotocol] = ACTIONS(2706),
    [anon_sym_ATinterface] = ACTIONS(2706),
    [anon_sym_ATimplementation] = ACTIONS(2706),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2706),
    [anon_sym_ATtry] = ACTIONS(2706),
    [anon_sym___try] = ACTIONS(2704),
    [anon_sym_ATthrow] = ACTIONS(2706),
    [anon_sym_ATselector] = ACTIONS(2706),
    [anon_sym_ATavailable] = ACTIONS(2706),
    [anon_sym___builtin_available] = ACTIONS(2704),
    [anon_sym_va_arg] = ACTIONS(2704),
    [anon_sym_ATencode] = ACTIONS(2706),
    [anon_sym_ATsynchronized] = ACTIONS(2706),
    [anon_sym_BOOL] = ACTIONS(2704),
    [anon_sym_IMP] = ACTIONS(2704),
    [anon_sym_SEL] = ACTIONS(2704),
    [anon_sym_Class] = ACTIONS(2704),
    [anon_sym_id] = ACTIONS(2704),
  },
  [STATE(870)] = {
    [sym_identifier] = ACTIONS(2688),
    [aux_sym_preproc_include_token1] = ACTIONS(2688),
    [aux_sym_preproc_include_token2] = ACTIONS(2688),
    [aux_sym_preproc_def_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2688),
    [sym_preproc_directive] = ACTIONS(2688),
    [anon_sym_LPAREN2] = ACTIONS(2690),
    [anon_sym_BANG] = ACTIONS(2690),
    [anon_sym_TILDE] = ACTIONS(2690),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2690),
    [anon_sym_CARET] = ACTIONS(2690),
    [anon_sym_AMP] = ACTIONS(2690),
    [anon_sym_SEMI] = ACTIONS(2690),
    [anon_sym___extension__] = ACTIONS(2688),
    [anon_sym_typedef] = ACTIONS(2688),
    [anon_sym_extern] = ACTIONS(2688),
    [anon_sym___attribute__] = ACTIONS(2688),
    [anon_sym___attribute] = ACTIONS(2688),
    [anon_sym_noreturn] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym___declspec] = ACTIONS(2688),
    [anon_sym___cdecl] = ACTIONS(2688),
    [anon_sym___clrcall] = ACTIONS(2688),
    [anon_sym___stdcall] = ACTIONS(2688),
    [anon_sym___fastcall] = ACTIONS(2688),
    [anon_sym___thiscall] = ACTIONS(2688),
    [anon_sym___vectorcall] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_RBRACE] = ACTIONS(2690),
    [anon_sym_signed] = ACTIONS(2688),
    [anon_sym_unsigned] = ACTIONS(2688),
    [anon_sym_long] = ACTIONS(2688),
    [anon_sym_short] = ACTIONS(2688),
    [anon_sym_static] = ACTIONS(2688),
    [anon_sym_ATautoreleasepool] = ACTIONS(2690),
    [anon_sym_auto] = ACTIONS(2688),
    [anon_sym_register] = ACTIONS(2688),
    [anon_sym_inline] = ACTIONS(2688),
    [anon_sym___inline] = ACTIONS(2688),
    [anon_sym___inline__] = ACTIONS(2688),
    [anon_sym___forceinline] = ACTIONS(2688),
    [anon_sym_thread_local] = ACTIONS(2688),
    [anon_sym___thread] = ACTIONS(2688),
    [anon_sym_CG_EXTERN] = ACTIONS(2688),
    [anon_sym_CG_INLINE] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2688),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2688),
    [anon_sym_IBOutlet] = ACTIONS(2688),
    [anon_sym_IBInspectable] = ACTIONS(2688),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2688),
    [anon_sym_NS_INLINE] = ACTIONS(2688),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2688),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2688),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2688),
    [anon_sym_const] = ACTIONS(2688),
    [anon_sym_constexpr] = ACTIONS(2688),
    [anon_sym_volatile] = ACTIONS(2688),
    [anon_sym_restrict] = ACTIONS(2688),
    [anon_sym___restrict__] = ACTIONS(2688),
    [anon_sym__Atomic] = ACTIONS(2688),
    [anon_sym__Noreturn] = ACTIONS(2688),
    [anon_sym__Nonnull] = ACTIONS(2688),
    [anon_sym_nullable] = ACTIONS(2688),
    [anon_sym__Complex] = ACTIONS(2688),
    [anon_sym__Nullable] = ACTIONS(2688),
    [anon_sym__Nullable_result] = ACTIONS(2688),
    [anon_sym__Null_unspecified] = ACTIONS(2688),
    [anon_sym___autoreleasing] = ACTIONS(2688),
    [anon_sym___block] = ACTIONS(2688),
    [anon_sym___bridge] = ACTIONS(2688),
    [anon_sym___bridge_retained] = ACTIONS(2688),
    [anon_sym___bridge_transfer] = ACTIONS(2688),
    [anon_sym___complex] = ACTIONS(2688),
    [anon_sym___const] = ACTIONS(2688),
    [anon_sym___imag] = ACTIONS(2688),
    [anon_sym___kindof] = ACTIONS(2688),
    [anon_sym___nonnull] = ACTIONS(2688),
    [anon_sym___nullable] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2688),
    [anon_sym___real] = ACTIONS(2688),
    [anon_sym___strong] = ACTIONS(2688),
    [anon_sym___unsafe_unretained] = ACTIONS(2688),
    [anon_sym___unused] = ACTIONS(2688),
    [anon_sym___weak] = ACTIONS(2688),
    [anon_sym_alignas] = ACTIONS(2688),
    [anon_sym__Alignas] = ACTIONS(2688),
    [sym_primitive_type] = ACTIONS(2688),
    [anon_sym_enum] = ACTIONS(2688),
    [anon_sym_struct] = ACTIONS(2688),
    [anon_sym_union] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_switch] = ACTIONS(2688),
    [anon_sym_case] = ACTIONS(2688),
    [anon_sym_default] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_in] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_sizeof] = ACTIONS(2688),
    [anon_sym___alignof__] = ACTIONS(2688),
    [anon_sym___alignof] = ACTIONS(2688),
    [anon_sym__alignof] = ACTIONS(2688),
    [anon_sym_alignof] = ACTIONS(2688),
    [anon_sym__Alignof] = ACTIONS(2688),
    [anon_sym_offsetof] = ACTIONS(2688),
    [anon_sym__Generic] = ACTIONS(2688),
    [anon_sym_asm] = ACTIONS(2688),
    [anon_sym___asm__] = ACTIONS(2688),
    [anon_sym___asm] = ACTIONS(2688),
    [sym_number_literal] = ACTIONS(2690),
    [anon_sym_L_SQUOTE] = ACTIONS(2690),
    [anon_sym_u_SQUOTE] = ACTIONS(2690),
    [anon_sym_U_SQUOTE] = ACTIONS(2690),
    [anon_sym_u8_SQUOTE] = ACTIONS(2690),
    [anon_sym_SQUOTE] = ACTIONS(2690),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_DQUOTE] = ACTIONS(2690),
    [anon_sym_L_DQUOTE] = ACTIONS(2690),
    [anon_sym_u_DQUOTE] = ACTIONS(2690),
    [anon_sym_U_DQUOTE] = ACTIONS(2690),
    [anon_sym_u8_DQUOTE] = ACTIONS(2690),
    [sym_true] = ACTIONS(2688),
    [sym_false] = ACTIONS(2688),
    [anon_sym_NULL] = ACTIONS(2688),
    [anon_sym_nullptr] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2688),
    [anon_sym___typeof] = ACTIONS(2688),
    [anon_sym_typeof] = ACTIONS(2688),
    [anon_sym_ATimport] = ACTIONS(2690),
    [aux_sym_preproc_undef_token1] = ACTIONS(2688),
    [anon_sym_POUND] = ACTIONS(2688),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2688),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2688),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2688),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_API_AVAILABLE] = ACTIONS(2688),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_API_DEPRECATED] = ACTIONS(2688),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym___deprecated_msg] = ACTIONS(2688),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2688),
    [anon_sym_ATprotocol] = ACTIONS(2690),
    [anon_sym_ATinterface] = ACTIONS(2690),
    [anon_sym_ATimplementation] = ACTIONS(2690),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2690),
    [anon_sym_ATtry] = ACTIONS(2690),
    [anon_sym___try] = ACTIONS(2688),
    [anon_sym_ATthrow] = ACTIONS(2690),
    [anon_sym_ATselector] = ACTIONS(2690),
    [anon_sym_ATavailable] = ACTIONS(2690),
    [anon_sym___builtin_available] = ACTIONS(2688),
    [anon_sym_va_arg] = ACTIONS(2688),
    [anon_sym_ATencode] = ACTIONS(2690),
    [anon_sym_ATsynchronized] = ACTIONS(2690),
    [anon_sym_BOOL] = ACTIONS(2688),
    [anon_sym_IMP] = ACTIONS(2688),
    [anon_sym_SEL] = ACTIONS(2688),
    [anon_sym_Class] = ACTIONS(2688),
    [anon_sym_id] = ACTIONS(2688),
  },
  [STATE(871)] = {
    [sym_identifier] = ACTIONS(2634),
    [aux_sym_preproc_include_token1] = ACTIONS(2634),
    [aux_sym_preproc_include_token2] = ACTIONS(2634),
    [aux_sym_preproc_def_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2634),
    [sym_preproc_directive] = ACTIONS(2634),
    [anon_sym_LPAREN2] = ACTIONS(2636),
    [anon_sym_BANG] = ACTIONS(2636),
    [anon_sym_TILDE] = ACTIONS(2636),
    [anon_sym_DASH] = ACTIONS(2634),
    [anon_sym_PLUS] = ACTIONS(2634),
    [anon_sym_STAR] = ACTIONS(2636),
    [anon_sym_CARET] = ACTIONS(2636),
    [anon_sym_AMP] = ACTIONS(2636),
    [anon_sym_SEMI] = ACTIONS(2636),
    [anon_sym___extension__] = ACTIONS(2634),
    [anon_sym_typedef] = ACTIONS(2634),
    [anon_sym_extern] = ACTIONS(2634),
    [anon_sym___attribute__] = ACTIONS(2634),
    [anon_sym___attribute] = ACTIONS(2634),
    [anon_sym_noreturn] = ACTIONS(2634),
    [anon_sym_LBRACK] = ACTIONS(2636),
    [anon_sym___declspec] = ACTIONS(2634),
    [anon_sym___cdecl] = ACTIONS(2634),
    [anon_sym___clrcall] = ACTIONS(2634),
    [anon_sym___stdcall] = ACTIONS(2634),
    [anon_sym___fastcall] = ACTIONS(2634),
    [anon_sym___thiscall] = ACTIONS(2634),
    [anon_sym___vectorcall] = ACTIONS(2634),
    [anon_sym_LBRACE] = ACTIONS(2636),
    [anon_sym_RBRACE] = ACTIONS(2636),
    [anon_sym_signed] = ACTIONS(2634),
    [anon_sym_unsigned] = ACTIONS(2634),
    [anon_sym_long] = ACTIONS(2634),
    [anon_sym_short] = ACTIONS(2634),
    [anon_sym_static] = ACTIONS(2634),
    [anon_sym_ATautoreleasepool] = ACTIONS(2636),
    [anon_sym_auto] = ACTIONS(2634),
    [anon_sym_register] = ACTIONS(2634),
    [anon_sym_inline] = ACTIONS(2634),
    [anon_sym___inline] = ACTIONS(2634),
    [anon_sym___inline__] = ACTIONS(2634),
    [anon_sym___forceinline] = ACTIONS(2634),
    [anon_sym_thread_local] = ACTIONS(2634),
    [anon_sym___thread] = ACTIONS(2634),
    [anon_sym_CG_EXTERN] = ACTIONS(2634),
    [anon_sym_CG_INLINE] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2634),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2634),
    [anon_sym_IBOutlet] = ACTIONS(2634),
    [anon_sym_IBInspectable] = ACTIONS(2634),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2634),
    [anon_sym_NS_INLINE] = ACTIONS(2634),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2634),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2634),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2634),
    [anon_sym_const] = ACTIONS(2634),
    [anon_sym_constexpr] = ACTIONS(2634),
    [anon_sym_volatile] = ACTIONS(2634),
    [anon_sym_restrict] = ACTIONS(2634),
    [anon_sym___restrict__] = ACTIONS(2634),
    [anon_sym__Atomic] = ACTIONS(2634),
    [anon_sym__Noreturn] = ACTIONS(2634),
    [anon_sym__Nonnull] = ACTIONS(2634),
    [anon_sym_nullable] = ACTIONS(2634),
    [anon_sym__Complex] = ACTIONS(2634),
    [anon_sym__Nullable] = ACTIONS(2634),
    [anon_sym__Nullable_result] = ACTIONS(2634),
    [anon_sym__Null_unspecified] = ACTIONS(2634),
    [anon_sym___autoreleasing] = ACTIONS(2634),
    [anon_sym___block] = ACTIONS(2634),
    [anon_sym___bridge] = ACTIONS(2634),
    [anon_sym___bridge_retained] = ACTIONS(2634),
    [anon_sym___bridge_transfer] = ACTIONS(2634),
    [anon_sym___complex] = ACTIONS(2634),
    [anon_sym___const] = ACTIONS(2634),
    [anon_sym___imag] = ACTIONS(2634),
    [anon_sym___kindof] = ACTIONS(2634),
    [anon_sym___nonnull] = ACTIONS(2634),
    [anon_sym___nullable] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2634),
    [anon_sym___real] = ACTIONS(2634),
    [anon_sym___strong] = ACTIONS(2634),
    [anon_sym___unsafe_unretained] = ACTIONS(2634),
    [anon_sym___unused] = ACTIONS(2634),
    [anon_sym___weak] = ACTIONS(2634),
    [anon_sym_alignas] = ACTIONS(2634),
    [anon_sym__Alignas] = ACTIONS(2634),
    [sym_primitive_type] = ACTIONS(2634),
    [anon_sym_enum] = ACTIONS(2634),
    [anon_sym_struct] = ACTIONS(2634),
    [anon_sym_union] = ACTIONS(2634),
    [anon_sym_if] = ACTIONS(2634),
    [anon_sym_switch] = ACTIONS(2634),
    [anon_sym_case] = ACTIONS(2634),
    [anon_sym_default] = ACTIONS(2634),
    [anon_sym_while] = ACTIONS(2634),
    [anon_sym_do] = ACTIONS(2634),
    [anon_sym_for] = ACTIONS(2634),
    [anon_sym_in] = ACTIONS(2634),
    [anon_sym_return] = ACTIONS(2634),
    [anon_sym_break] = ACTIONS(2634),
    [anon_sym_continue] = ACTIONS(2634),
    [anon_sym_goto] = ACTIONS(2634),
    [anon_sym_DASH_DASH] = ACTIONS(2636),
    [anon_sym_PLUS_PLUS] = ACTIONS(2636),
    [anon_sym_sizeof] = ACTIONS(2634),
    [anon_sym___alignof__] = ACTIONS(2634),
    [anon_sym___alignof] = ACTIONS(2634),
    [anon_sym__alignof] = ACTIONS(2634),
    [anon_sym_alignof] = ACTIONS(2634),
    [anon_sym__Alignof] = ACTIONS(2634),
    [anon_sym_offsetof] = ACTIONS(2634),
    [anon_sym__Generic] = ACTIONS(2634),
    [anon_sym_asm] = ACTIONS(2634),
    [anon_sym___asm__] = ACTIONS(2634),
    [anon_sym___asm] = ACTIONS(2634),
    [sym_number_literal] = ACTIONS(2636),
    [anon_sym_L_SQUOTE] = ACTIONS(2636),
    [anon_sym_u_SQUOTE] = ACTIONS(2636),
    [anon_sym_U_SQUOTE] = ACTIONS(2636),
    [anon_sym_u8_SQUOTE] = ACTIONS(2636),
    [anon_sym_SQUOTE] = ACTIONS(2636),
    [anon_sym_AT] = ACTIONS(2634),
    [anon_sym_DQUOTE] = ACTIONS(2636),
    [anon_sym_L_DQUOTE] = ACTIONS(2636),
    [anon_sym_u_DQUOTE] = ACTIONS(2636),
    [anon_sym_U_DQUOTE] = ACTIONS(2636),
    [anon_sym_u8_DQUOTE] = ACTIONS(2636),
    [sym_true] = ACTIONS(2634),
    [sym_false] = ACTIONS(2634),
    [anon_sym_NULL] = ACTIONS(2634),
    [anon_sym_nullptr] = ACTIONS(2634),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2634),
    [anon_sym___typeof] = ACTIONS(2634),
    [anon_sym_typeof] = ACTIONS(2634),
    [anon_sym_ATimport] = ACTIONS(2636),
    [aux_sym_preproc_undef_token1] = ACTIONS(2634),
    [anon_sym_POUND] = ACTIONS(2634),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2634),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2634),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2634),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_API_AVAILABLE] = ACTIONS(2634),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_API_DEPRECATED] = ACTIONS(2634),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym___deprecated_msg] = ACTIONS(2634),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2634),
    [anon_sym_ATprotocol] = ACTIONS(2636),
    [anon_sym_ATinterface] = ACTIONS(2636),
    [anon_sym_ATimplementation] = ACTIONS(2636),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2636),
    [anon_sym_ATtry] = ACTIONS(2636),
    [anon_sym___try] = ACTIONS(2634),
    [anon_sym_ATthrow] = ACTIONS(2636),
    [anon_sym_ATselector] = ACTIONS(2636),
    [anon_sym_ATavailable] = ACTIONS(2636),
    [anon_sym___builtin_available] = ACTIONS(2634),
    [anon_sym_va_arg] = ACTIONS(2634),
    [anon_sym_ATencode] = ACTIONS(2636),
    [anon_sym_ATsynchronized] = ACTIONS(2636),
    [anon_sym_BOOL] = ACTIONS(2634),
    [anon_sym_IMP] = ACTIONS(2634),
    [anon_sym_SEL] = ACTIONS(2634),
    [anon_sym_Class] = ACTIONS(2634),
    [anon_sym_id] = ACTIONS(2634),
  },
  [STATE(872)] = {
    [sym_identifier] = ACTIONS(2792),
    [aux_sym_preproc_include_token1] = ACTIONS(2792),
    [aux_sym_preproc_include_token2] = ACTIONS(2792),
    [aux_sym_preproc_def_token1] = ACTIONS(2792),
    [aux_sym_preproc_if_token1] = ACTIONS(2792),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2792),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2792),
    [sym_preproc_directive] = ACTIONS(2792),
    [anon_sym_LPAREN2] = ACTIONS(2794),
    [anon_sym_BANG] = ACTIONS(2794),
    [anon_sym_TILDE] = ACTIONS(2794),
    [anon_sym_DASH] = ACTIONS(2792),
    [anon_sym_PLUS] = ACTIONS(2792),
    [anon_sym_STAR] = ACTIONS(2794),
    [anon_sym_CARET] = ACTIONS(2794),
    [anon_sym_AMP] = ACTIONS(2794),
    [anon_sym_SEMI] = ACTIONS(2794),
    [anon_sym___extension__] = ACTIONS(2792),
    [anon_sym_typedef] = ACTIONS(2792),
    [anon_sym_extern] = ACTIONS(2792),
    [anon_sym___attribute__] = ACTIONS(2792),
    [anon_sym___attribute] = ACTIONS(2792),
    [anon_sym_noreturn] = ACTIONS(2792),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym___declspec] = ACTIONS(2792),
    [anon_sym___cdecl] = ACTIONS(2792),
    [anon_sym___clrcall] = ACTIONS(2792),
    [anon_sym___stdcall] = ACTIONS(2792),
    [anon_sym___fastcall] = ACTIONS(2792),
    [anon_sym___thiscall] = ACTIONS(2792),
    [anon_sym___vectorcall] = ACTIONS(2792),
    [anon_sym_LBRACE] = ACTIONS(2794),
    [anon_sym_RBRACE] = ACTIONS(2794),
    [anon_sym_signed] = ACTIONS(2792),
    [anon_sym_unsigned] = ACTIONS(2792),
    [anon_sym_long] = ACTIONS(2792),
    [anon_sym_short] = ACTIONS(2792),
    [anon_sym_static] = ACTIONS(2792),
    [anon_sym_ATautoreleasepool] = ACTIONS(2794),
    [anon_sym_auto] = ACTIONS(2792),
    [anon_sym_register] = ACTIONS(2792),
    [anon_sym_inline] = ACTIONS(2792),
    [anon_sym___inline] = ACTIONS(2792),
    [anon_sym___inline__] = ACTIONS(2792),
    [anon_sym___forceinline] = ACTIONS(2792),
    [anon_sym_thread_local] = ACTIONS(2792),
    [anon_sym___thread] = ACTIONS(2792),
    [anon_sym_CG_EXTERN] = ACTIONS(2792),
    [anon_sym_CG_INLINE] = ACTIONS(2792),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2792),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2792),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2792),
    [anon_sym_IBOutlet] = ACTIONS(2792),
    [anon_sym_IBInspectable] = ACTIONS(2792),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2792),
    [anon_sym_NS_INLINE] = ACTIONS(2792),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2792),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2792),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2792),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2792),
    [anon_sym_const] = ACTIONS(2792),
    [anon_sym_constexpr] = ACTIONS(2792),
    [anon_sym_volatile] = ACTIONS(2792),
    [anon_sym_restrict] = ACTIONS(2792),
    [anon_sym___restrict__] = ACTIONS(2792),
    [anon_sym__Atomic] = ACTIONS(2792),
    [anon_sym__Noreturn] = ACTIONS(2792),
    [anon_sym__Nonnull] = ACTIONS(2792),
    [anon_sym_nullable] = ACTIONS(2792),
    [anon_sym__Complex] = ACTIONS(2792),
    [anon_sym__Nullable] = ACTIONS(2792),
    [anon_sym__Nullable_result] = ACTIONS(2792),
    [anon_sym__Null_unspecified] = ACTIONS(2792),
    [anon_sym___autoreleasing] = ACTIONS(2792),
    [anon_sym___block] = ACTIONS(2792),
    [anon_sym___bridge] = ACTIONS(2792),
    [anon_sym___bridge_retained] = ACTIONS(2792),
    [anon_sym___bridge_transfer] = ACTIONS(2792),
    [anon_sym___complex] = ACTIONS(2792),
    [anon_sym___const] = ACTIONS(2792),
    [anon_sym___imag] = ACTIONS(2792),
    [anon_sym___kindof] = ACTIONS(2792),
    [anon_sym___nonnull] = ACTIONS(2792),
    [anon_sym___nullable] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2792),
    [anon_sym___real] = ACTIONS(2792),
    [anon_sym___strong] = ACTIONS(2792),
    [anon_sym___unsafe_unretained] = ACTIONS(2792),
    [anon_sym___unused] = ACTIONS(2792),
    [anon_sym___weak] = ACTIONS(2792),
    [anon_sym_alignas] = ACTIONS(2792),
    [anon_sym__Alignas] = ACTIONS(2792),
    [sym_primitive_type] = ACTIONS(2792),
    [anon_sym_enum] = ACTIONS(2792),
    [anon_sym_struct] = ACTIONS(2792),
    [anon_sym_union] = ACTIONS(2792),
    [anon_sym_if] = ACTIONS(2792),
    [anon_sym_switch] = ACTIONS(2792),
    [anon_sym_case] = ACTIONS(2792),
    [anon_sym_default] = ACTIONS(2792),
    [anon_sym_while] = ACTIONS(2792),
    [anon_sym_do] = ACTIONS(2792),
    [anon_sym_for] = ACTIONS(2792),
    [anon_sym_in] = ACTIONS(2792),
    [anon_sym_return] = ACTIONS(2792),
    [anon_sym_break] = ACTIONS(2792),
    [anon_sym_continue] = ACTIONS(2792),
    [anon_sym_goto] = ACTIONS(2792),
    [anon_sym_DASH_DASH] = ACTIONS(2794),
    [anon_sym_PLUS_PLUS] = ACTIONS(2794),
    [anon_sym_sizeof] = ACTIONS(2792),
    [anon_sym___alignof__] = ACTIONS(2792),
    [anon_sym___alignof] = ACTIONS(2792),
    [anon_sym__alignof] = ACTIONS(2792),
    [anon_sym_alignof] = ACTIONS(2792),
    [anon_sym__Alignof] = ACTIONS(2792),
    [anon_sym_offsetof] = ACTIONS(2792),
    [anon_sym__Generic] = ACTIONS(2792),
    [anon_sym_asm] = ACTIONS(2792),
    [anon_sym___asm__] = ACTIONS(2792),
    [anon_sym___asm] = ACTIONS(2792),
    [sym_number_literal] = ACTIONS(2794),
    [anon_sym_L_SQUOTE] = ACTIONS(2794),
    [anon_sym_u_SQUOTE] = ACTIONS(2794),
    [anon_sym_U_SQUOTE] = ACTIONS(2794),
    [anon_sym_u8_SQUOTE] = ACTIONS(2794),
    [anon_sym_SQUOTE] = ACTIONS(2794),
    [anon_sym_AT] = ACTIONS(2792),
    [anon_sym_DQUOTE] = ACTIONS(2794),
    [anon_sym_L_DQUOTE] = ACTIONS(2794),
    [anon_sym_u_DQUOTE] = ACTIONS(2794),
    [anon_sym_U_DQUOTE] = ACTIONS(2794),
    [anon_sym_u8_DQUOTE] = ACTIONS(2794),
    [sym_true] = ACTIONS(2792),
    [sym_false] = ACTIONS(2792),
    [anon_sym_NULL] = ACTIONS(2792),
    [anon_sym_nullptr] = ACTIONS(2792),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2792),
    [anon_sym___typeof] = ACTIONS(2792),
    [anon_sym_typeof] = ACTIONS(2792),
    [anon_sym_ATimport] = ACTIONS(2794),
    [aux_sym_preproc_undef_token1] = ACTIONS(2792),
    [anon_sym_POUND] = ACTIONS(2792),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2792),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2792),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2792),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2792),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2792),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2792),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2792),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_API_AVAILABLE] = ACTIONS(2792),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_API_DEPRECATED] = ACTIONS(2792),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2792),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym___deprecated_msg] = ACTIONS(2792),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2792),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2792),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2792),
    [anon_sym_ATprotocol] = ACTIONS(2794),
    [anon_sym_ATinterface] = ACTIONS(2794),
    [anon_sym_ATimplementation] = ACTIONS(2794),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2794),
    [anon_sym_ATtry] = ACTIONS(2794),
    [anon_sym___try] = ACTIONS(2792),
    [anon_sym_ATthrow] = ACTIONS(2794),
    [anon_sym_ATselector] = ACTIONS(2794),
    [anon_sym_ATavailable] = ACTIONS(2794),
    [anon_sym___builtin_available] = ACTIONS(2792),
    [anon_sym_va_arg] = ACTIONS(2792),
    [anon_sym_ATencode] = ACTIONS(2794),
    [anon_sym_ATsynchronized] = ACTIONS(2794),
    [anon_sym_BOOL] = ACTIONS(2792),
    [anon_sym_IMP] = ACTIONS(2792),
    [anon_sym_SEL] = ACTIONS(2792),
    [anon_sym_Class] = ACTIONS(2792),
    [anon_sym_id] = ACTIONS(2792),
  },
  [STATE(873)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_RBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(874)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_RBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(875)] = {
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_RBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATtry] = ACTIONS(2798),
    [anon_sym___try] = ACTIONS(2796),
    [anon_sym_ATthrow] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_ATsynchronized] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(876)] = {
    [sym_identifier] = ACTIONS(2668),
    [aux_sym_preproc_include_token1] = ACTIONS(2668),
    [aux_sym_preproc_include_token2] = ACTIONS(2668),
    [aux_sym_preproc_def_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2668),
    [sym_preproc_directive] = ACTIONS(2668),
    [anon_sym_LPAREN2] = ACTIONS(2670),
    [anon_sym_BANG] = ACTIONS(2670),
    [anon_sym_TILDE] = ACTIONS(2670),
    [anon_sym_DASH] = ACTIONS(2668),
    [anon_sym_PLUS] = ACTIONS(2668),
    [anon_sym_STAR] = ACTIONS(2670),
    [anon_sym_CARET] = ACTIONS(2670),
    [anon_sym_AMP] = ACTIONS(2670),
    [anon_sym_SEMI] = ACTIONS(2670),
    [anon_sym___extension__] = ACTIONS(2668),
    [anon_sym_typedef] = ACTIONS(2668),
    [anon_sym_extern] = ACTIONS(2668),
    [anon_sym___attribute__] = ACTIONS(2668),
    [anon_sym___attribute] = ACTIONS(2668),
    [anon_sym_noreturn] = ACTIONS(2668),
    [anon_sym_LBRACK] = ACTIONS(2670),
    [anon_sym___declspec] = ACTIONS(2668),
    [anon_sym___cdecl] = ACTIONS(2668),
    [anon_sym___clrcall] = ACTIONS(2668),
    [anon_sym___stdcall] = ACTIONS(2668),
    [anon_sym___fastcall] = ACTIONS(2668),
    [anon_sym___thiscall] = ACTIONS(2668),
    [anon_sym___vectorcall] = ACTIONS(2668),
    [anon_sym_LBRACE] = ACTIONS(2670),
    [anon_sym_RBRACE] = ACTIONS(2670),
    [anon_sym_signed] = ACTIONS(2668),
    [anon_sym_unsigned] = ACTIONS(2668),
    [anon_sym_long] = ACTIONS(2668),
    [anon_sym_short] = ACTIONS(2668),
    [anon_sym_static] = ACTIONS(2668),
    [anon_sym_ATautoreleasepool] = ACTIONS(2670),
    [anon_sym_auto] = ACTIONS(2668),
    [anon_sym_register] = ACTIONS(2668),
    [anon_sym_inline] = ACTIONS(2668),
    [anon_sym___inline] = ACTIONS(2668),
    [anon_sym___inline__] = ACTIONS(2668),
    [anon_sym___forceinline] = ACTIONS(2668),
    [anon_sym_thread_local] = ACTIONS(2668),
    [anon_sym___thread] = ACTIONS(2668),
    [anon_sym_CG_EXTERN] = ACTIONS(2668),
    [anon_sym_CG_INLINE] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2668),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2668),
    [anon_sym_IBOutlet] = ACTIONS(2668),
    [anon_sym_IBInspectable] = ACTIONS(2668),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2668),
    [anon_sym_NS_INLINE] = ACTIONS(2668),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2668),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2668),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2668),
    [anon_sym_const] = ACTIONS(2668),
    [anon_sym_constexpr] = ACTIONS(2668),
    [anon_sym_volatile] = ACTIONS(2668),
    [anon_sym_restrict] = ACTIONS(2668),
    [anon_sym___restrict__] = ACTIONS(2668),
    [anon_sym__Atomic] = ACTIONS(2668),
    [anon_sym__Noreturn] = ACTIONS(2668),
    [anon_sym__Nonnull] = ACTIONS(2668),
    [anon_sym_nullable] = ACTIONS(2668),
    [anon_sym__Complex] = ACTIONS(2668),
    [anon_sym__Nullable] = ACTIONS(2668),
    [anon_sym__Nullable_result] = ACTIONS(2668),
    [anon_sym__Null_unspecified] = ACTIONS(2668),
    [anon_sym___autoreleasing] = ACTIONS(2668),
    [anon_sym___block] = ACTIONS(2668),
    [anon_sym___bridge] = ACTIONS(2668),
    [anon_sym___bridge_retained] = ACTIONS(2668),
    [anon_sym___bridge_transfer] = ACTIONS(2668),
    [anon_sym___complex] = ACTIONS(2668),
    [anon_sym___const] = ACTIONS(2668),
    [anon_sym___imag] = ACTIONS(2668),
    [anon_sym___kindof] = ACTIONS(2668),
    [anon_sym___nonnull] = ACTIONS(2668),
    [anon_sym___nullable] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2668),
    [anon_sym___real] = ACTIONS(2668),
    [anon_sym___strong] = ACTIONS(2668),
    [anon_sym___unsafe_unretained] = ACTIONS(2668),
    [anon_sym___unused] = ACTIONS(2668),
    [anon_sym___weak] = ACTIONS(2668),
    [anon_sym_alignas] = ACTIONS(2668),
    [anon_sym__Alignas] = ACTIONS(2668),
    [sym_primitive_type] = ACTIONS(2668),
    [anon_sym_enum] = ACTIONS(2668),
    [anon_sym_struct] = ACTIONS(2668),
    [anon_sym_union] = ACTIONS(2668),
    [anon_sym_if] = ACTIONS(2668),
    [anon_sym_switch] = ACTIONS(2668),
    [anon_sym_case] = ACTIONS(2668),
    [anon_sym_default] = ACTIONS(2668),
    [anon_sym_while] = ACTIONS(2668),
    [anon_sym_do] = ACTIONS(2668),
    [anon_sym_for] = ACTIONS(2668),
    [anon_sym_in] = ACTIONS(2668),
    [anon_sym_return] = ACTIONS(2668),
    [anon_sym_break] = ACTIONS(2668),
    [anon_sym_continue] = ACTIONS(2668),
    [anon_sym_goto] = ACTIONS(2668),
    [anon_sym_DASH_DASH] = ACTIONS(2670),
    [anon_sym_PLUS_PLUS] = ACTIONS(2670),
    [anon_sym_sizeof] = ACTIONS(2668),
    [anon_sym___alignof__] = ACTIONS(2668),
    [anon_sym___alignof] = ACTIONS(2668),
    [anon_sym__alignof] = ACTIONS(2668),
    [anon_sym_alignof] = ACTIONS(2668),
    [anon_sym__Alignof] = ACTIONS(2668),
    [anon_sym_offsetof] = ACTIONS(2668),
    [anon_sym__Generic] = ACTIONS(2668),
    [anon_sym_asm] = ACTIONS(2668),
    [anon_sym___asm__] = ACTIONS(2668),
    [anon_sym___asm] = ACTIONS(2668),
    [sym_number_literal] = ACTIONS(2670),
    [anon_sym_L_SQUOTE] = ACTIONS(2670),
    [anon_sym_u_SQUOTE] = ACTIONS(2670),
    [anon_sym_U_SQUOTE] = ACTIONS(2670),
    [anon_sym_u8_SQUOTE] = ACTIONS(2670),
    [anon_sym_SQUOTE] = ACTIONS(2670),
    [anon_sym_AT] = ACTIONS(2668),
    [anon_sym_DQUOTE] = ACTIONS(2670),
    [anon_sym_L_DQUOTE] = ACTIONS(2670),
    [anon_sym_u_DQUOTE] = ACTIONS(2670),
    [anon_sym_U_DQUOTE] = ACTIONS(2670),
    [anon_sym_u8_DQUOTE] = ACTIONS(2670),
    [sym_true] = ACTIONS(2668),
    [sym_false] = ACTIONS(2668),
    [anon_sym_NULL] = ACTIONS(2668),
    [anon_sym_nullptr] = ACTIONS(2668),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2668),
    [anon_sym___typeof] = ACTIONS(2668),
    [anon_sym_typeof] = ACTIONS(2668),
    [anon_sym_ATimport] = ACTIONS(2670),
    [aux_sym_preproc_undef_token1] = ACTIONS(2668),
    [anon_sym_POUND] = ACTIONS(2668),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2668),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2668),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2668),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_API_AVAILABLE] = ACTIONS(2668),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_API_DEPRECATED] = ACTIONS(2668),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym___deprecated_msg] = ACTIONS(2668),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2668),
    [anon_sym_ATprotocol] = ACTIONS(2670),
    [anon_sym_ATinterface] = ACTIONS(2670),
    [anon_sym_ATimplementation] = ACTIONS(2670),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2670),
    [anon_sym_ATtry] = ACTIONS(2670),
    [anon_sym___try] = ACTIONS(2668),
    [anon_sym_ATthrow] = ACTIONS(2670),
    [anon_sym_ATselector] = ACTIONS(2670),
    [anon_sym_ATavailable] = ACTIONS(2670),
    [anon_sym___builtin_available] = ACTIONS(2668),
    [anon_sym_va_arg] = ACTIONS(2668),
    [anon_sym_ATencode] = ACTIONS(2670),
    [anon_sym_ATsynchronized] = ACTIONS(2670),
    [anon_sym_BOOL] = ACTIONS(2668),
    [anon_sym_IMP] = ACTIONS(2668),
    [anon_sym_SEL] = ACTIONS(2668),
    [anon_sym_Class] = ACTIONS(2668),
    [anon_sym_id] = ACTIONS(2668),
  },
  [STATE(877)] = {
    [sym_identifier] = ACTIONS(2800),
    [aux_sym_preproc_include_token1] = ACTIONS(2800),
    [aux_sym_preproc_include_token2] = ACTIONS(2800),
    [aux_sym_preproc_def_token1] = ACTIONS(2800),
    [aux_sym_preproc_if_token1] = ACTIONS(2800),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2800),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2800),
    [sym_preproc_directive] = ACTIONS(2800),
    [anon_sym_LPAREN2] = ACTIONS(2802),
    [anon_sym_BANG] = ACTIONS(2802),
    [anon_sym_TILDE] = ACTIONS(2802),
    [anon_sym_DASH] = ACTIONS(2800),
    [anon_sym_PLUS] = ACTIONS(2800),
    [anon_sym_STAR] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(2802),
    [anon_sym_AMP] = ACTIONS(2802),
    [anon_sym_SEMI] = ACTIONS(2802),
    [anon_sym___extension__] = ACTIONS(2800),
    [anon_sym_typedef] = ACTIONS(2800),
    [anon_sym_extern] = ACTIONS(2800),
    [anon_sym___attribute__] = ACTIONS(2800),
    [anon_sym___attribute] = ACTIONS(2800),
    [anon_sym_noreturn] = ACTIONS(2800),
    [anon_sym_LBRACK] = ACTIONS(2802),
    [anon_sym___declspec] = ACTIONS(2800),
    [anon_sym___cdecl] = ACTIONS(2800),
    [anon_sym___clrcall] = ACTIONS(2800),
    [anon_sym___stdcall] = ACTIONS(2800),
    [anon_sym___fastcall] = ACTIONS(2800),
    [anon_sym___thiscall] = ACTIONS(2800),
    [anon_sym___vectorcall] = ACTIONS(2800),
    [anon_sym_LBRACE] = ACTIONS(2802),
    [anon_sym_RBRACE] = ACTIONS(2802),
    [anon_sym_signed] = ACTIONS(2800),
    [anon_sym_unsigned] = ACTIONS(2800),
    [anon_sym_long] = ACTIONS(2800),
    [anon_sym_short] = ACTIONS(2800),
    [anon_sym_static] = ACTIONS(2800),
    [anon_sym_ATautoreleasepool] = ACTIONS(2802),
    [anon_sym_auto] = ACTIONS(2800),
    [anon_sym_register] = ACTIONS(2800),
    [anon_sym_inline] = ACTIONS(2800),
    [anon_sym___inline] = ACTIONS(2800),
    [anon_sym___inline__] = ACTIONS(2800),
    [anon_sym___forceinline] = ACTIONS(2800),
    [anon_sym_thread_local] = ACTIONS(2800),
    [anon_sym___thread] = ACTIONS(2800),
    [anon_sym_CG_EXTERN] = ACTIONS(2800),
    [anon_sym_CG_INLINE] = ACTIONS(2800),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2800),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2800),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2800),
    [anon_sym_IBOutlet] = ACTIONS(2800),
    [anon_sym_IBInspectable] = ACTIONS(2800),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2800),
    [anon_sym_NS_INLINE] = ACTIONS(2800),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2800),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2800),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2800),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2800),
    [anon_sym_const] = ACTIONS(2800),
    [anon_sym_constexpr] = ACTIONS(2800),
    [anon_sym_volatile] = ACTIONS(2800),
    [anon_sym_restrict] = ACTIONS(2800),
    [anon_sym___restrict__] = ACTIONS(2800),
    [anon_sym__Atomic] = ACTIONS(2800),
    [anon_sym__Noreturn] = ACTIONS(2800),
    [anon_sym__Nonnull] = ACTIONS(2800),
    [anon_sym_nullable] = ACTIONS(2800),
    [anon_sym__Complex] = ACTIONS(2800),
    [anon_sym__Nullable] = ACTIONS(2800),
    [anon_sym__Nullable_result] = ACTIONS(2800),
    [anon_sym__Null_unspecified] = ACTIONS(2800),
    [anon_sym___autoreleasing] = ACTIONS(2800),
    [anon_sym___block] = ACTIONS(2800),
    [anon_sym___bridge] = ACTIONS(2800),
    [anon_sym___bridge_retained] = ACTIONS(2800),
    [anon_sym___bridge_transfer] = ACTIONS(2800),
    [anon_sym___complex] = ACTIONS(2800),
    [anon_sym___const] = ACTIONS(2800),
    [anon_sym___imag] = ACTIONS(2800),
    [anon_sym___kindof] = ACTIONS(2800),
    [anon_sym___nonnull] = ACTIONS(2800),
    [anon_sym___nullable] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2800),
    [anon_sym___real] = ACTIONS(2800),
    [anon_sym___strong] = ACTIONS(2800),
    [anon_sym___unsafe_unretained] = ACTIONS(2800),
    [anon_sym___unused] = ACTIONS(2800),
    [anon_sym___weak] = ACTIONS(2800),
    [anon_sym_alignas] = ACTIONS(2800),
    [anon_sym__Alignas] = ACTIONS(2800),
    [sym_primitive_type] = ACTIONS(2800),
    [anon_sym_enum] = ACTIONS(2800),
    [anon_sym_struct] = ACTIONS(2800),
    [anon_sym_union] = ACTIONS(2800),
    [anon_sym_if] = ACTIONS(2800),
    [anon_sym_switch] = ACTIONS(2800),
    [anon_sym_case] = ACTIONS(2800),
    [anon_sym_default] = ACTIONS(2800),
    [anon_sym_while] = ACTIONS(2800),
    [anon_sym_do] = ACTIONS(2800),
    [anon_sym_for] = ACTIONS(2800),
    [anon_sym_in] = ACTIONS(2800),
    [anon_sym_return] = ACTIONS(2800),
    [anon_sym_break] = ACTIONS(2800),
    [anon_sym_continue] = ACTIONS(2800),
    [anon_sym_goto] = ACTIONS(2800),
    [anon_sym_DASH_DASH] = ACTIONS(2802),
    [anon_sym_PLUS_PLUS] = ACTIONS(2802),
    [anon_sym_sizeof] = ACTIONS(2800),
    [anon_sym___alignof__] = ACTIONS(2800),
    [anon_sym___alignof] = ACTIONS(2800),
    [anon_sym__alignof] = ACTIONS(2800),
    [anon_sym_alignof] = ACTIONS(2800),
    [anon_sym__Alignof] = ACTIONS(2800),
    [anon_sym_offsetof] = ACTIONS(2800),
    [anon_sym__Generic] = ACTIONS(2800),
    [anon_sym_asm] = ACTIONS(2800),
    [anon_sym___asm__] = ACTIONS(2800),
    [anon_sym___asm] = ACTIONS(2800),
    [sym_number_literal] = ACTIONS(2802),
    [anon_sym_L_SQUOTE] = ACTIONS(2802),
    [anon_sym_u_SQUOTE] = ACTIONS(2802),
    [anon_sym_U_SQUOTE] = ACTIONS(2802),
    [anon_sym_u8_SQUOTE] = ACTIONS(2802),
    [anon_sym_SQUOTE] = ACTIONS(2802),
    [anon_sym_AT] = ACTIONS(2800),
    [anon_sym_DQUOTE] = ACTIONS(2802),
    [anon_sym_L_DQUOTE] = ACTIONS(2802),
    [anon_sym_u_DQUOTE] = ACTIONS(2802),
    [anon_sym_U_DQUOTE] = ACTIONS(2802),
    [anon_sym_u8_DQUOTE] = ACTIONS(2802),
    [sym_true] = ACTIONS(2800),
    [sym_false] = ACTIONS(2800),
    [anon_sym_NULL] = ACTIONS(2800),
    [anon_sym_nullptr] = ACTIONS(2800),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2800),
    [anon_sym___typeof] = ACTIONS(2800),
    [anon_sym_typeof] = ACTIONS(2800),
    [anon_sym_ATimport] = ACTIONS(2802),
    [aux_sym_preproc_undef_token1] = ACTIONS(2800),
    [anon_sym_POUND] = ACTIONS(2800),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2800),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2800),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2800),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2800),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2800),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2800),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2800),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_API_AVAILABLE] = ACTIONS(2800),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_API_DEPRECATED] = ACTIONS(2800),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2800),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym___deprecated_msg] = ACTIONS(2800),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2800),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2800),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2800),
    [anon_sym_ATprotocol] = ACTIONS(2802),
    [anon_sym_ATinterface] = ACTIONS(2802),
    [anon_sym_ATimplementation] = ACTIONS(2802),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2802),
    [anon_sym_ATtry] = ACTIONS(2802),
    [anon_sym___try] = ACTIONS(2800),
    [anon_sym_ATthrow] = ACTIONS(2802),
    [anon_sym_ATselector] = ACTIONS(2802),
    [anon_sym_ATavailable] = ACTIONS(2802),
    [anon_sym___builtin_available] = ACTIONS(2800),
    [anon_sym_va_arg] = ACTIONS(2800),
    [anon_sym_ATencode] = ACTIONS(2802),
    [anon_sym_ATsynchronized] = ACTIONS(2802),
    [anon_sym_BOOL] = ACTIONS(2800),
    [anon_sym_IMP] = ACTIONS(2800),
    [anon_sym_SEL] = ACTIONS(2800),
    [anon_sym_Class] = ACTIONS(2800),
    [anon_sym_id] = ACTIONS(2800),
  },
  [STATE(878)] = {
    [sym_identifier] = ACTIONS(2596),
    [aux_sym_preproc_include_token1] = ACTIONS(2596),
    [aux_sym_preproc_include_token2] = ACTIONS(2596),
    [aux_sym_preproc_def_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2596),
    [sym_preproc_directive] = ACTIONS(2596),
    [anon_sym_LPAREN2] = ACTIONS(2598),
    [anon_sym_BANG] = ACTIONS(2598),
    [anon_sym_TILDE] = ACTIONS(2598),
    [anon_sym_DASH] = ACTIONS(2596),
    [anon_sym_PLUS] = ACTIONS(2596),
    [anon_sym_STAR] = ACTIONS(2598),
    [anon_sym_CARET] = ACTIONS(2598),
    [anon_sym_AMP] = ACTIONS(2598),
    [anon_sym_SEMI] = ACTIONS(2598),
    [anon_sym___extension__] = ACTIONS(2596),
    [anon_sym_typedef] = ACTIONS(2596),
    [anon_sym_extern] = ACTIONS(2596),
    [anon_sym___attribute__] = ACTIONS(2596),
    [anon_sym___attribute] = ACTIONS(2596),
    [anon_sym_noreturn] = ACTIONS(2596),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym___declspec] = ACTIONS(2596),
    [anon_sym___cdecl] = ACTIONS(2596),
    [anon_sym___clrcall] = ACTIONS(2596),
    [anon_sym___stdcall] = ACTIONS(2596),
    [anon_sym___fastcall] = ACTIONS(2596),
    [anon_sym___thiscall] = ACTIONS(2596),
    [anon_sym___vectorcall] = ACTIONS(2596),
    [anon_sym_LBRACE] = ACTIONS(2598),
    [anon_sym_RBRACE] = ACTIONS(2598),
    [anon_sym_signed] = ACTIONS(2596),
    [anon_sym_unsigned] = ACTIONS(2596),
    [anon_sym_long] = ACTIONS(2596),
    [anon_sym_short] = ACTIONS(2596),
    [anon_sym_static] = ACTIONS(2596),
    [anon_sym_ATautoreleasepool] = ACTIONS(2598),
    [anon_sym_auto] = ACTIONS(2596),
    [anon_sym_register] = ACTIONS(2596),
    [anon_sym_inline] = ACTIONS(2596),
    [anon_sym___inline] = ACTIONS(2596),
    [anon_sym___inline__] = ACTIONS(2596),
    [anon_sym___forceinline] = ACTIONS(2596),
    [anon_sym_thread_local] = ACTIONS(2596),
    [anon_sym___thread] = ACTIONS(2596),
    [anon_sym_CG_EXTERN] = ACTIONS(2596),
    [anon_sym_CG_INLINE] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2596),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2596),
    [anon_sym_IBOutlet] = ACTIONS(2596),
    [anon_sym_IBInspectable] = ACTIONS(2596),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2596),
    [anon_sym_NS_INLINE] = ACTIONS(2596),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2596),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2596),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2596),
    [anon_sym_const] = ACTIONS(2596),
    [anon_sym_constexpr] = ACTIONS(2596),
    [anon_sym_volatile] = ACTIONS(2596),
    [anon_sym_restrict] = ACTIONS(2596),
    [anon_sym___restrict__] = ACTIONS(2596),
    [anon_sym__Atomic] = ACTIONS(2596),
    [anon_sym__Noreturn] = ACTIONS(2596),
    [anon_sym__Nonnull] = ACTIONS(2596),
    [anon_sym_nullable] = ACTIONS(2596),
    [anon_sym__Complex] = ACTIONS(2596),
    [anon_sym__Nullable] = ACTIONS(2596),
    [anon_sym__Nullable_result] = ACTIONS(2596),
    [anon_sym__Null_unspecified] = ACTIONS(2596),
    [anon_sym___autoreleasing] = ACTIONS(2596),
    [anon_sym___block] = ACTIONS(2596),
    [anon_sym___bridge] = ACTIONS(2596),
    [anon_sym___bridge_retained] = ACTIONS(2596),
    [anon_sym___bridge_transfer] = ACTIONS(2596),
    [anon_sym___complex] = ACTIONS(2596),
    [anon_sym___const] = ACTIONS(2596),
    [anon_sym___imag] = ACTIONS(2596),
    [anon_sym___kindof] = ACTIONS(2596),
    [anon_sym___nonnull] = ACTIONS(2596),
    [anon_sym___nullable] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2596),
    [anon_sym___real] = ACTIONS(2596),
    [anon_sym___strong] = ACTIONS(2596),
    [anon_sym___unsafe_unretained] = ACTIONS(2596),
    [anon_sym___unused] = ACTIONS(2596),
    [anon_sym___weak] = ACTIONS(2596),
    [anon_sym_alignas] = ACTIONS(2596),
    [anon_sym__Alignas] = ACTIONS(2596),
    [sym_primitive_type] = ACTIONS(2596),
    [anon_sym_enum] = ACTIONS(2596),
    [anon_sym_struct] = ACTIONS(2596),
    [anon_sym_union] = ACTIONS(2596),
    [anon_sym_if] = ACTIONS(2596),
    [anon_sym_switch] = ACTIONS(2596),
    [anon_sym_case] = ACTIONS(2596),
    [anon_sym_default] = ACTIONS(2596),
    [anon_sym_while] = ACTIONS(2596),
    [anon_sym_do] = ACTIONS(2596),
    [anon_sym_for] = ACTIONS(2596),
    [anon_sym_in] = ACTIONS(2596),
    [anon_sym_return] = ACTIONS(2596),
    [anon_sym_break] = ACTIONS(2596),
    [anon_sym_continue] = ACTIONS(2596),
    [anon_sym_goto] = ACTIONS(2596),
    [anon_sym_DASH_DASH] = ACTIONS(2598),
    [anon_sym_PLUS_PLUS] = ACTIONS(2598),
    [anon_sym_sizeof] = ACTIONS(2596),
    [anon_sym___alignof__] = ACTIONS(2596),
    [anon_sym___alignof] = ACTIONS(2596),
    [anon_sym__alignof] = ACTIONS(2596),
    [anon_sym_alignof] = ACTIONS(2596),
    [anon_sym__Alignof] = ACTIONS(2596),
    [anon_sym_offsetof] = ACTIONS(2596),
    [anon_sym__Generic] = ACTIONS(2596),
    [anon_sym_asm] = ACTIONS(2596),
    [anon_sym___asm__] = ACTIONS(2596),
    [anon_sym___asm] = ACTIONS(2596),
    [sym_number_literal] = ACTIONS(2598),
    [anon_sym_L_SQUOTE] = ACTIONS(2598),
    [anon_sym_u_SQUOTE] = ACTIONS(2598),
    [anon_sym_U_SQUOTE] = ACTIONS(2598),
    [anon_sym_u8_SQUOTE] = ACTIONS(2598),
    [anon_sym_SQUOTE] = ACTIONS(2598),
    [anon_sym_AT] = ACTIONS(2596),
    [anon_sym_DQUOTE] = ACTIONS(2598),
    [anon_sym_L_DQUOTE] = ACTIONS(2598),
    [anon_sym_u_DQUOTE] = ACTIONS(2598),
    [anon_sym_U_DQUOTE] = ACTIONS(2598),
    [anon_sym_u8_DQUOTE] = ACTIONS(2598),
    [sym_true] = ACTIONS(2596),
    [sym_false] = ACTIONS(2596),
    [anon_sym_NULL] = ACTIONS(2596),
    [anon_sym_nullptr] = ACTIONS(2596),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2596),
    [anon_sym___typeof] = ACTIONS(2596),
    [anon_sym_typeof] = ACTIONS(2596),
    [anon_sym_ATimport] = ACTIONS(2598),
    [aux_sym_preproc_undef_token1] = ACTIONS(2596),
    [anon_sym_POUND] = ACTIONS(2596),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2596),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2596),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2596),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_API_AVAILABLE] = ACTIONS(2596),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_API_DEPRECATED] = ACTIONS(2596),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym___deprecated_msg] = ACTIONS(2596),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2596),
    [anon_sym_ATprotocol] = ACTIONS(2598),
    [anon_sym_ATinterface] = ACTIONS(2598),
    [anon_sym_ATimplementation] = ACTIONS(2598),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2598),
    [anon_sym_ATtry] = ACTIONS(2598),
    [anon_sym___try] = ACTIONS(2596),
    [anon_sym_ATthrow] = ACTIONS(2598),
    [anon_sym_ATselector] = ACTIONS(2598),
    [anon_sym_ATavailable] = ACTIONS(2598),
    [anon_sym___builtin_available] = ACTIONS(2596),
    [anon_sym_va_arg] = ACTIONS(2596),
    [anon_sym_ATencode] = ACTIONS(2598),
    [anon_sym_ATsynchronized] = ACTIONS(2598),
    [anon_sym_BOOL] = ACTIONS(2596),
    [anon_sym_IMP] = ACTIONS(2596),
    [anon_sym_SEL] = ACTIONS(2596),
    [anon_sym_Class] = ACTIONS(2596),
    [anon_sym_id] = ACTIONS(2596),
  },
  [STATE(879)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_RBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(880)] = {
    [sym_identifier] = ACTIONS(2744),
    [aux_sym_preproc_include_token1] = ACTIONS(2744),
    [aux_sym_preproc_include_token2] = ACTIONS(2744),
    [aux_sym_preproc_def_token1] = ACTIONS(2744),
    [aux_sym_preproc_if_token1] = ACTIONS(2744),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2744),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2744),
    [sym_preproc_directive] = ACTIONS(2744),
    [anon_sym_LPAREN2] = ACTIONS(2746),
    [anon_sym_BANG] = ACTIONS(2746),
    [anon_sym_TILDE] = ACTIONS(2746),
    [anon_sym_DASH] = ACTIONS(2744),
    [anon_sym_PLUS] = ACTIONS(2744),
    [anon_sym_STAR] = ACTIONS(2746),
    [anon_sym_CARET] = ACTIONS(2746),
    [anon_sym_AMP] = ACTIONS(2746),
    [anon_sym_SEMI] = ACTIONS(2746),
    [anon_sym___extension__] = ACTIONS(2744),
    [anon_sym_typedef] = ACTIONS(2744),
    [anon_sym_extern] = ACTIONS(2744),
    [anon_sym___attribute__] = ACTIONS(2744),
    [anon_sym___attribute] = ACTIONS(2744),
    [anon_sym_noreturn] = ACTIONS(2744),
    [anon_sym_LBRACK] = ACTIONS(2746),
    [anon_sym___declspec] = ACTIONS(2744),
    [anon_sym___cdecl] = ACTIONS(2744),
    [anon_sym___clrcall] = ACTIONS(2744),
    [anon_sym___stdcall] = ACTIONS(2744),
    [anon_sym___fastcall] = ACTIONS(2744),
    [anon_sym___thiscall] = ACTIONS(2744),
    [anon_sym___vectorcall] = ACTIONS(2744),
    [anon_sym_LBRACE] = ACTIONS(2746),
    [anon_sym_RBRACE] = ACTIONS(2746),
    [anon_sym_signed] = ACTIONS(2744),
    [anon_sym_unsigned] = ACTIONS(2744),
    [anon_sym_long] = ACTIONS(2744),
    [anon_sym_short] = ACTIONS(2744),
    [anon_sym_static] = ACTIONS(2744),
    [anon_sym_ATautoreleasepool] = ACTIONS(2746),
    [anon_sym_auto] = ACTIONS(2744),
    [anon_sym_register] = ACTIONS(2744),
    [anon_sym_inline] = ACTIONS(2744),
    [anon_sym___inline] = ACTIONS(2744),
    [anon_sym___inline__] = ACTIONS(2744),
    [anon_sym___forceinline] = ACTIONS(2744),
    [anon_sym_thread_local] = ACTIONS(2744),
    [anon_sym___thread] = ACTIONS(2744),
    [anon_sym_CG_EXTERN] = ACTIONS(2744),
    [anon_sym_CG_INLINE] = ACTIONS(2744),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2744),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2744),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2744),
    [anon_sym_IBOutlet] = ACTIONS(2744),
    [anon_sym_IBInspectable] = ACTIONS(2744),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2744),
    [anon_sym_NS_INLINE] = ACTIONS(2744),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2744),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2744),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2744),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2744),
    [anon_sym_const] = ACTIONS(2744),
    [anon_sym_constexpr] = ACTIONS(2744),
    [anon_sym_volatile] = ACTIONS(2744),
    [anon_sym_restrict] = ACTIONS(2744),
    [anon_sym___restrict__] = ACTIONS(2744),
    [anon_sym__Atomic] = ACTIONS(2744),
    [anon_sym__Noreturn] = ACTIONS(2744),
    [anon_sym__Nonnull] = ACTIONS(2744),
    [anon_sym_nullable] = ACTIONS(2744),
    [anon_sym__Complex] = ACTIONS(2744),
    [anon_sym__Nullable] = ACTIONS(2744),
    [anon_sym__Nullable_result] = ACTIONS(2744),
    [anon_sym__Null_unspecified] = ACTIONS(2744),
    [anon_sym___autoreleasing] = ACTIONS(2744),
    [anon_sym___block] = ACTIONS(2744),
    [anon_sym___bridge] = ACTIONS(2744),
    [anon_sym___bridge_retained] = ACTIONS(2744),
    [anon_sym___bridge_transfer] = ACTIONS(2744),
    [anon_sym___complex] = ACTIONS(2744),
    [anon_sym___const] = ACTIONS(2744),
    [anon_sym___imag] = ACTIONS(2744),
    [anon_sym___kindof] = ACTIONS(2744),
    [anon_sym___nonnull] = ACTIONS(2744),
    [anon_sym___nullable] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2744),
    [anon_sym___real] = ACTIONS(2744),
    [anon_sym___strong] = ACTIONS(2744),
    [anon_sym___unsafe_unretained] = ACTIONS(2744),
    [anon_sym___unused] = ACTIONS(2744),
    [anon_sym___weak] = ACTIONS(2744),
    [anon_sym_alignas] = ACTIONS(2744),
    [anon_sym__Alignas] = ACTIONS(2744),
    [sym_primitive_type] = ACTIONS(2744),
    [anon_sym_enum] = ACTIONS(2744),
    [anon_sym_struct] = ACTIONS(2744),
    [anon_sym_union] = ACTIONS(2744),
    [anon_sym_if] = ACTIONS(2744),
    [anon_sym_switch] = ACTIONS(2744),
    [anon_sym_case] = ACTIONS(2744),
    [anon_sym_default] = ACTIONS(2744),
    [anon_sym_while] = ACTIONS(2744),
    [anon_sym_do] = ACTIONS(2744),
    [anon_sym_for] = ACTIONS(2744),
    [anon_sym_in] = ACTIONS(2744),
    [anon_sym_return] = ACTIONS(2744),
    [anon_sym_break] = ACTIONS(2744),
    [anon_sym_continue] = ACTIONS(2744),
    [anon_sym_goto] = ACTIONS(2744),
    [anon_sym_DASH_DASH] = ACTIONS(2746),
    [anon_sym_PLUS_PLUS] = ACTIONS(2746),
    [anon_sym_sizeof] = ACTIONS(2744),
    [anon_sym___alignof__] = ACTIONS(2744),
    [anon_sym___alignof] = ACTIONS(2744),
    [anon_sym__alignof] = ACTIONS(2744),
    [anon_sym_alignof] = ACTIONS(2744),
    [anon_sym__Alignof] = ACTIONS(2744),
    [anon_sym_offsetof] = ACTIONS(2744),
    [anon_sym__Generic] = ACTIONS(2744),
    [anon_sym_asm] = ACTIONS(2744),
    [anon_sym___asm__] = ACTIONS(2744),
    [anon_sym___asm] = ACTIONS(2744),
    [sym_number_literal] = ACTIONS(2746),
    [anon_sym_L_SQUOTE] = ACTIONS(2746),
    [anon_sym_u_SQUOTE] = ACTIONS(2746),
    [anon_sym_U_SQUOTE] = ACTIONS(2746),
    [anon_sym_u8_SQUOTE] = ACTIONS(2746),
    [anon_sym_SQUOTE] = ACTIONS(2746),
    [anon_sym_AT] = ACTIONS(2744),
    [anon_sym_DQUOTE] = ACTIONS(2746),
    [anon_sym_L_DQUOTE] = ACTIONS(2746),
    [anon_sym_u_DQUOTE] = ACTIONS(2746),
    [anon_sym_U_DQUOTE] = ACTIONS(2746),
    [anon_sym_u8_DQUOTE] = ACTIONS(2746),
    [sym_true] = ACTIONS(2744),
    [sym_false] = ACTIONS(2744),
    [anon_sym_NULL] = ACTIONS(2744),
    [anon_sym_nullptr] = ACTIONS(2744),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2744),
    [anon_sym___typeof] = ACTIONS(2744),
    [anon_sym_typeof] = ACTIONS(2744),
    [anon_sym_ATimport] = ACTIONS(2746),
    [aux_sym_preproc_undef_token1] = ACTIONS(2744),
    [anon_sym_POUND] = ACTIONS(2744),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2744),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2744),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2744),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2744),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2744),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2744),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2744),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_API_AVAILABLE] = ACTIONS(2744),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_API_DEPRECATED] = ACTIONS(2744),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2744),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym___deprecated_msg] = ACTIONS(2744),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2744),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2744),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2744),
    [anon_sym_ATprotocol] = ACTIONS(2746),
    [anon_sym_ATinterface] = ACTIONS(2746),
    [anon_sym_ATimplementation] = ACTIONS(2746),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2746),
    [anon_sym_ATtry] = ACTIONS(2746),
    [anon_sym___try] = ACTIONS(2744),
    [anon_sym_ATthrow] = ACTIONS(2746),
    [anon_sym_ATselector] = ACTIONS(2746),
    [anon_sym_ATavailable] = ACTIONS(2746),
    [anon_sym___builtin_available] = ACTIONS(2744),
    [anon_sym_va_arg] = ACTIONS(2744),
    [anon_sym_ATencode] = ACTIONS(2746),
    [anon_sym_ATsynchronized] = ACTIONS(2746),
    [anon_sym_BOOL] = ACTIONS(2744),
    [anon_sym_IMP] = ACTIONS(2744),
    [anon_sym_SEL] = ACTIONS(2744),
    [anon_sym_Class] = ACTIONS(2744),
    [anon_sym_id] = ACTIONS(2744),
  },
  [STATE(881)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_RBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(882)] = {
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_RBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATtry] = ACTIONS(2806),
    [anon_sym___try] = ACTIONS(2804),
    [anon_sym_ATthrow] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_ATsynchronized] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(883)] = {
    [sym_identifier] = ACTIONS(2748),
    [aux_sym_preproc_include_token1] = ACTIONS(2748),
    [aux_sym_preproc_include_token2] = ACTIONS(2748),
    [aux_sym_preproc_def_token1] = ACTIONS(2748),
    [aux_sym_preproc_if_token1] = ACTIONS(2748),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2748),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2748),
    [sym_preproc_directive] = ACTIONS(2748),
    [anon_sym_LPAREN2] = ACTIONS(2750),
    [anon_sym_BANG] = ACTIONS(2750),
    [anon_sym_TILDE] = ACTIONS(2750),
    [anon_sym_DASH] = ACTIONS(2748),
    [anon_sym_PLUS] = ACTIONS(2748),
    [anon_sym_STAR] = ACTIONS(2750),
    [anon_sym_CARET] = ACTIONS(2750),
    [anon_sym_AMP] = ACTIONS(2750),
    [anon_sym_SEMI] = ACTIONS(2750),
    [anon_sym___extension__] = ACTIONS(2748),
    [anon_sym_typedef] = ACTIONS(2748),
    [anon_sym_extern] = ACTIONS(2748),
    [anon_sym___attribute__] = ACTIONS(2748),
    [anon_sym___attribute] = ACTIONS(2748),
    [anon_sym_noreturn] = ACTIONS(2748),
    [anon_sym_LBRACK] = ACTIONS(2750),
    [anon_sym___declspec] = ACTIONS(2748),
    [anon_sym___cdecl] = ACTIONS(2748),
    [anon_sym___clrcall] = ACTIONS(2748),
    [anon_sym___stdcall] = ACTIONS(2748),
    [anon_sym___fastcall] = ACTIONS(2748),
    [anon_sym___thiscall] = ACTIONS(2748),
    [anon_sym___vectorcall] = ACTIONS(2748),
    [anon_sym_LBRACE] = ACTIONS(2750),
    [anon_sym_RBRACE] = ACTIONS(2750),
    [anon_sym_signed] = ACTIONS(2748),
    [anon_sym_unsigned] = ACTIONS(2748),
    [anon_sym_long] = ACTIONS(2748),
    [anon_sym_short] = ACTIONS(2748),
    [anon_sym_static] = ACTIONS(2748),
    [anon_sym_ATautoreleasepool] = ACTIONS(2750),
    [anon_sym_auto] = ACTIONS(2748),
    [anon_sym_register] = ACTIONS(2748),
    [anon_sym_inline] = ACTIONS(2748),
    [anon_sym___inline] = ACTIONS(2748),
    [anon_sym___inline__] = ACTIONS(2748),
    [anon_sym___forceinline] = ACTIONS(2748),
    [anon_sym_thread_local] = ACTIONS(2748),
    [anon_sym___thread] = ACTIONS(2748),
    [anon_sym_CG_EXTERN] = ACTIONS(2748),
    [anon_sym_CG_INLINE] = ACTIONS(2748),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2748),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2748),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2748),
    [anon_sym_IBOutlet] = ACTIONS(2748),
    [anon_sym_IBInspectable] = ACTIONS(2748),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2748),
    [anon_sym_NS_INLINE] = ACTIONS(2748),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2748),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2748),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2748),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2748),
    [anon_sym_const] = ACTIONS(2748),
    [anon_sym_constexpr] = ACTIONS(2748),
    [anon_sym_volatile] = ACTIONS(2748),
    [anon_sym_restrict] = ACTIONS(2748),
    [anon_sym___restrict__] = ACTIONS(2748),
    [anon_sym__Atomic] = ACTIONS(2748),
    [anon_sym__Noreturn] = ACTIONS(2748),
    [anon_sym__Nonnull] = ACTIONS(2748),
    [anon_sym_nullable] = ACTIONS(2748),
    [anon_sym__Complex] = ACTIONS(2748),
    [anon_sym__Nullable] = ACTIONS(2748),
    [anon_sym__Nullable_result] = ACTIONS(2748),
    [anon_sym__Null_unspecified] = ACTIONS(2748),
    [anon_sym___autoreleasing] = ACTIONS(2748),
    [anon_sym___block] = ACTIONS(2748),
    [anon_sym___bridge] = ACTIONS(2748),
    [anon_sym___bridge_retained] = ACTIONS(2748),
    [anon_sym___bridge_transfer] = ACTIONS(2748),
    [anon_sym___complex] = ACTIONS(2748),
    [anon_sym___const] = ACTIONS(2748),
    [anon_sym___imag] = ACTIONS(2748),
    [anon_sym___kindof] = ACTIONS(2748),
    [anon_sym___nonnull] = ACTIONS(2748),
    [anon_sym___nullable] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2748),
    [anon_sym___real] = ACTIONS(2748),
    [anon_sym___strong] = ACTIONS(2748),
    [anon_sym___unsafe_unretained] = ACTIONS(2748),
    [anon_sym___unused] = ACTIONS(2748),
    [anon_sym___weak] = ACTIONS(2748),
    [anon_sym_alignas] = ACTIONS(2748),
    [anon_sym__Alignas] = ACTIONS(2748),
    [sym_primitive_type] = ACTIONS(2748),
    [anon_sym_enum] = ACTIONS(2748),
    [anon_sym_struct] = ACTIONS(2748),
    [anon_sym_union] = ACTIONS(2748),
    [anon_sym_if] = ACTIONS(2748),
    [anon_sym_switch] = ACTIONS(2748),
    [anon_sym_case] = ACTIONS(2748),
    [anon_sym_default] = ACTIONS(2748),
    [anon_sym_while] = ACTIONS(2748),
    [anon_sym_do] = ACTIONS(2748),
    [anon_sym_for] = ACTIONS(2748),
    [anon_sym_in] = ACTIONS(2748),
    [anon_sym_return] = ACTIONS(2748),
    [anon_sym_break] = ACTIONS(2748),
    [anon_sym_continue] = ACTIONS(2748),
    [anon_sym_goto] = ACTIONS(2748),
    [anon_sym_DASH_DASH] = ACTIONS(2750),
    [anon_sym_PLUS_PLUS] = ACTIONS(2750),
    [anon_sym_sizeof] = ACTIONS(2748),
    [anon_sym___alignof__] = ACTIONS(2748),
    [anon_sym___alignof] = ACTIONS(2748),
    [anon_sym__alignof] = ACTIONS(2748),
    [anon_sym_alignof] = ACTIONS(2748),
    [anon_sym__Alignof] = ACTIONS(2748),
    [anon_sym_offsetof] = ACTIONS(2748),
    [anon_sym__Generic] = ACTIONS(2748),
    [anon_sym_asm] = ACTIONS(2748),
    [anon_sym___asm__] = ACTIONS(2748),
    [anon_sym___asm] = ACTIONS(2748),
    [sym_number_literal] = ACTIONS(2750),
    [anon_sym_L_SQUOTE] = ACTIONS(2750),
    [anon_sym_u_SQUOTE] = ACTIONS(2750),
    [anon_sym_U_SQUOTE] = ACTIONS(2750),
    [anon_sym_u8_SQUOTE] = ACTIONS(2750),
    [anon_sym_SQUOTE] = ACTIONS(2750),
    [anon_sym_AT] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_L_DQUOTE] = ACTIONS(2750),
    [anon_sym_u_DQUOTE] = ACTIONS(2750),
    [anon_sym_U_DQUOTE] = ACTIONS(2750),
    [anon_sym_u8_DQUOTE] = ACTIONS(2750),
    [sym_true] = ACTIONS(2748),
    [sym_false] = ACTIONS(2748),
    [anon_sym_NULL] = ACTIONS(2748),
    [anon_sym_nullptr] = ACTIONS(2748),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2748),
    [anon_sym___typeof] = ACTIONS(2748),
    [anon_sym_typeof] = ACTIONS(2748),
    [anon_sym_ATimport] = ACTIONS(2750),
    [aux_sym_preproc_undef_token1] = ACTIONS(2748),
    [anon_sym_POUND] = ACTIONS(2748),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2748),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2748),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2748),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2748),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2748),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2748),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2748),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_API_AVAILABLE] = ACTIONS(2748),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_API_DEPRECATED] = ACTIONS(2748),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2748),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym___deprecated_msg] = ACTIONS(2748),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2748),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2748),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2748),
    [anon_sym_ATprotocol] = ACTIONS(2750),
    [anon_sym_ATinterface] = ACTIONS(2750),
    [anon_sym_ATimplementation] = ACTIONS(2750),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2750),
    [anon_sym_ATtry] = ACTIONS(2750),
    [anon_sym___try] = ACTIONS(2748),
    [anon_sym_ATthrow] = ACTIONS(2750),
    [anon_sym_ATselector] = ACTIONS(2750),
    [anon_sym_ATavailable] = ACTIONS(2750),
    [anon_sym___builtin_available] = ACTIONS(2748),
    [anon_sym_va_arg] = ACTIONS(2748),
    [anon_sym_ATencode] = ACTIONS(2750),
    [anon_sym_ATsynchronized] = ACTIONS(2750),
    [anon_sym_BOOL] = ACTIONS(2748),
    [anon_sym_IMP] = ACTIONS(2748),
    [anon_sym_SEL] = ACTIONS(2748),
    [anon_sym_Class] = ACTIONS(2748),
    [anon_sym_id] = ACTIONS(2748),
  },
  [STATE(884)] = {
    [sym_identifier] = ACTIONS(2808),
    [aux_sym_preproc_include_token1] = ACTIONS(2808),
    [aux_sym_preproc_include_token2] = ACTIONS(2808),
    [aux_sym_preproc_def_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2808),
    [sym_preproc_directive] = ACTIONS(2808),
    [anon_sym_LPAREN2] = ACTIONS(2810),
    [anon_sym_BANG] = ACTIONS(2810),
    [anon_sym_TILDE] = ACTIONS(2810),
    [anon_sym_DASH] = ACTIONS(2808),
    [anon_sym_PLUS] = ACTIONS(2808),
    [anon_sym_STAR] = ACTIONS(2810),
    [anon_sym_CARET] = ACTIONS(2810),
    [anon_sym_AMP] = ACTIONS(2810),
    [anon_sym_SEMI] = ACTIONS(2810),
    [anon_sym___extension__] = ACTIONS(2808),
    [anon_sym_typedef] = ACTIONS(2808),
    [anon_sym_extern] = ACTIONS(2808),
    [anon_sym___attribute__] = ACTIONS(2808),
    [anon_sym___attribute] = ACTIONS(2808),
    [anon_sym_noreturn] = ACTIONS(2808),
    [anon_sym_LBRACK] = ACTIONS(2810),
    [anon_sym___declspec] = ACTIONS(2808),
    [anon_sym___cdecl] = ACTIONS(2808),
    [anon_sym___clrcall] = ACTIONS(2808),
    [anon_sym___stdcall] = ACTIONS(2808),
    [anon_sym___fastcall] = ACTIONS(2808),
    [anon_sym___thiscall] = ACTIONS(2808),
    [anon_sym___vectorcall] = ACTIONS(2808),
    [anon_sym_LBRACE] = ACTIONS(2810),
    [anon_sym_RBRACE] = ACTIONS(2810),
    [anon_sym_signed] = ACTIONS(2808),
    [anon_sym_unsigned] = ACTIONS(2808),
    [anon_sym_long] = ACTIONS(2808),
    [anon_sym_short] = ACTIONS(2808),
    [anon_sym_static] = ACTIONS(2808),
    [anon_sym_ATautoreleasepool] = ACTIONS(2810),
    [anon_sym_auto] = ACTIONS(2808),
    [anon_sym_register] = ACTIONS(2808),
    [anon_sym_inline] = ACTIONS(2808),
    [anon_sym___inline] = ACTIONS(2808),
    [anon_sym___inline__] = ACTIONS(2808),
    [anon_sym___forceinline] = ACTIONS(2808),
    [anon_sym_thread_local] = ACTIONS(2808),
    [anon_sym___thread] = ACTIONS(2808),
    [anon_sym_CG_EXTERN] = ACTIONS(2808),
    [anon_sym_CG_INLINE] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2808),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2808),
    [anon_sym_IBOutlet] = ACTIONS(2808),
    [anon_sym_IBInspectable] = ACTIONS(2808),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2808),
    [anon_sym_NS_INLINE] = ACTIONS(2808),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2808),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2808),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2808),
    [anon_sym_const] = ACTIONS(2808),
    [anon_sym_constexpr] = ACTIONS(2808),
    [anon_sym_volatile] = ACTIONS(2808),
    [anon_sym_restrict] = ACTIONS(2808),
    [anon_sym___restrict__] = ACTIONS(2808),
    [anon_sym__Atomic] = ACTIONS(2808),
    [anon_sym__Noreturn] = ACTIONS(2808),
    [anon_sym__Nonnull] = ACTIONS(2808),
    [anon_sym_nullable] = ACTIONS(2808),
    [anon_sym__Complex] = ACTIONS(2808),
    [anon_sym__Nullable] = ACTIONS(2808),
    [anon_sym__Nullable_result] = ACTIONS(2808),
    [anon_sym__Null_unspecified] = ACTIONS(2808),
    [anon_sym___autoreleasing] = ACTIONS(2808),
    [anon_sym___block] = ACTIONS(2808),
    [anon_sym___bridge] = ACTIONS(2808),
    [anon_sym___bridge_retained] = ACTIONS(2808),
    [anon_sym___bridge_transfer] = ACTIONS(2808),
    [anon_sym___complex] = ACTIONS(2808),
    [anon_sym___const] = ACTIONS(2808),
    [anon_sym___imag] = ACTIONS(2808),
    [anon_sym___kindof] = ACTIONS(2808),
    [anon_sym___nonnull] = ACTIONS(2808),
    [anon_sym___nullable] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2808),
    [anon_sym___real] = ACTIONS(2808),
    [anon_sym___strong] = ACTIONS(2808),
    [anon_sym___unsafe_unretained] = ACTIONS(2808),
    [anon_sym___unused] = ACTIONS(2808),
    [anon_sym___weak] = ACTIONS(2808),
    [anon_sym_alignas] = ACTIONS(2808),
    [anon_sym__Alignas] = ACTIONS(2808),
    [sym_primitive_type] = ACTIONS(2808),
    [anon_sym_enum] = ACTIONS(2808),
    [anon_sym_struct] = ACTIONS(2808),
    [anon_sym_union] = ACTIONS(2808),
    [anon_sym_if] = ACTIONS(2808),
    [anon_sym_switch] = ACTIONS(2808),
    [anon_sym_case] = ACTIONS(2808),
    [anon_sym_default] = ACTIONS(2808),
    [anon_sym_while] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2808),
    [anon_sym_in] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_break] = ACTIONS(2808),
    [anon_sym_continue] = ACTIONS(2808),
    [anon_sym_goto] = ACTIONS(2808),
    [anon_sym_DASH_DASH] = ACTIONS(2810),
    [anon_sym_PLUS_PLUS] = ACTIONS(2810),
    [anon_sym_sizeof] = ACTIONS(2808),
    [anon_sym___alignof__] = ACTIONS(2808),
    [anon_sym___alignof] = ACTIONS(2808),
    [anon_sym__alignof] = ACTIONS(2808),
    [anon_sym_alignof] = ACTIONS(2808),
    [anon_sym__Alignof] = ACTIONS(2808),
    [anon_sym_offsetof] = ACTIONS(2808),
    [anon_sym__Generic] = ACTIONS(2808),
    [anon_sym_asm] = ACTIONS(2808),
    [anon_sym___asm__] = ACTIONS(2808),
    [anon_sym___asm] = ACTIONS(2808),
    [sym_number_literal] = ACTIONS(2810),
    [anon_sym_L_SQUOTE] = ACTIONS(2810),
    [anon_sym_u_SQUOTE] = ACTIONS(2810),
    [anon_sym_U_SQUOTE] = ACTIONS(2810),
    [anon_sym_u8_SQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2810),
    [anon_sym_AT] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_L_DQUOTE] = ACTIONS(2810),
    [anon_sym_u_DQUOTE] = ACTIONS(2810),
    [anon_sym_U_DQUOTE] = ACTIONS(2810),
    [anon_sym_u8_DQUOTE] = ACTIONS(2810),
    [sym_true] = ACTIONS(2808),
    [sym_false] = ACTIONS(2808),
    [anon_sym_NULL] = ACTIONS(2808),
    [anon_sym_nullptr] = ACTIONS(2808),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2808),
    [anon_sym___typeof] = ACTIONS(2808),
    [anon_sym_typeof] = ACTIONS(2808),
    [anon_sym_ATimport] = ACTIONS(2810),
    [aux_sym_preproc_undef_token1] = ACTIONS(2808),
    [anon_sym_POUND] = ACTIONS(2808),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2808),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2808),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2808),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_API_AVAILABLE] = ACTIONS(2808),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_API_DEPRECATED] = ACTIONS(2808),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym___deprecated_msg] = ACTIONS(2808),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2808),
    [anon_sym_ATprotocol] = ACTIONS(2810),
    [anon_sym_ATinterface] = ACTIONS(2810),
    [anon_sym_ATimplementation] = ACTIONS(2810),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2810),
    [anon_sym_ATtry] = ACTIONS(2810),
    [anon_sym___try] = ACTIONS(2808),
    [anon_sym_ATthrow] = ACTIONS(2810),
    [anon_sym_ATselector] = ACTIONS(2810),
    [anon_sym_ATavailable] = ACTIONS(2810),
    [anon_sym___builtin_available] = ACTIONS(2808),
    [anon_sym_va_arg] = ACTIONS(2808),
    [anon_sym_ATencode] = ACTIONS(2810),
    [anon_sym_ATsynchronized] = ACTIONS(2810),
    [anon_sym_BOOL] = ACTIONS(2808),
    [anon_sym_IMP] = ACTIONS(2808),
    [anon_sym_SEL] = ACTIONS(2808),
    [anon_sym_Class] = ACTIONS(2808),
    [anon_sym_id] = ACTIONS(2808),
  },
  [STATE(885)] = {
    [sym_identifier] = ACTIONS(2808),
    [aux_sym_preproc_include_token1] = ACTIONS(2808),
    [aux_sym_preproc_include_token2] = ACTIONS(2808),
    [aux_sym_preproc_def_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2808),
    [sym_preproc_directive] = ACTIONS(2808),
    [anon_sym_LPAREN2] = ACTIONS(2810),
    [anon_sym_BANG] = ACTIONS(2810),
    [anon_sym_TILDE] = ACTIONS(2810),
    [anon_sym_DASH] = ACTIONS(2808),
    [anon_sym_PLUS] = ACTIONS(2808),
    [anon_sym_STAR] = ACTIONS(2810),
    [anon_sym_CARET] = ACTIONS(2810),
    [anon_sym_AMP] = ACTIONS(2810),
    [anon_sym_SEMI] = ACTIONS(2810),
    [anon_sym___extension__] = ACTIONS(2808),
    [anon_sym_typedef] = ACTIONS(2808),
    [anon_sym_extern] = ACTIONS(2808),
    [anon_sym___attribute__] = ACTIONS(2808),
    [anon_sym___attribute] = ACTIONS(2808),
    [anon_sym_noreturn] = ACTIONS(2808),
    [anon_sym_LBRACK] = ACTIONS(2810),
    [anon_sym___declspec] = ACTIONS(2808),
    [anon_sym___cdecl] = ACTIONS(2808),
    [anon_sym___clrcall] = ACTIONS(2808),
    [anon_sym___stdcall] = ACTIONS(2808),
    [anon_sym___fastcall] = ACTIONS(2808),
    [anon_sym___thiscall] = ACTIONS(2808),
    [anon_sym___vectorcall] = ACTIONS(2808),
    [anon_sym_LBRACE] = ACTIONS(2810),
    [anon_sym_RBRACE] = ACTIONS(2810),
    [anon_sym_signed] = ACTIONS(2808),
    [anon_sym_unsigned] = ACTIONS(2808),
    [anon_sym_long] = ACTIONS(2808),
    [anon_sym_short] = ACTIONS(2808),
    [anon_sym_static] = ACTIONS(2808),
    [anon_sym_ATautoreleasepool] = ACTIONS(2810),
    [anon_sym_auto] = ACTIONS(2808),
    [anon_sym_register] = ACTIONS(2808),
    [anon_sym_inline] = ACTIONS(2808),
    [anon_sym___inline] = ACTIONS(2808),
    [anon_sym___inline__] = ACTIONS(2808),
    [anon_sym___forceinline] = ACTIONS(2808),
    [anon_sym_thread_local] = ACTIONS(2808),
    [anon_sym___thread] = ACTIONS(2808),
    [anon_sym_CG_EXTERN] = ACTIONS(2808),
    [anon_sym_CG_INLINE] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2808),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2808),
    [anon_sym_IBOutlet] = ACTIONS(2808),
    [anon_sym_IBInspectable] = ACTIONS(2808),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2808),
    [anon_sym_NS_INLINE] = ACTIONS(2808),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2808),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2808),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2808),
    [anon_sym_const] = ACTIONS(2808),
    [anon_sym_constexpr] = ACTIONS(2808),
    [anon_sym_volatile] = ACTIONS(2808),
    [anon_sym_restrict] = ACTIONS(2808),
    [anon_sym___restrict__] = ACTIONS(2808),
    [anon_sym__Atomic] = ACTIONS(2808),
    [anon_sym__Noreturn] = ACTIONS(2808),
    [anon_sym__Nonnull] = ACTIONS(2808),
    [anon_sym_nullable] = ACTIONS(2808),
    [anon_sym__Complex] = ACTIONS(2808),
    [anon_sym__Nullable] = ACTIONS(2808),
    [anon_sym__Nullable_result] = ACTIONS(2808),
    [anon_sym__Null_unspecified] = ACTIONS(2808),
    [anon_sym___autoreleasing] = ACTIONS(2808),
    [anon_sym___block] = ACTIONS(2808),
    [anon_sym___bridge] = ACTIONS(2808),
    [anon_sym___bridge_retained] = ACTIONS(2808),
    [anon_sym___bridge_transfer] = ACTIONS(2808),
    [anon_sym___complex] = ACTIONS(2808),
    [anon_sym___const] = ACTIONS(2808),
    [anon_sym___imag] = ACTIONS(2808),
    [anon_sym___kindof] = ACTIONS(2808),
    [anon_sym___nonnull] = ACTIONS(2808),
    [anon_sym___nullable] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2808),
    [anon_sym___real] = ACTIONS(2808),
    [anon_sym___strong] = ACTIONS(2808),
    [anon_sym___unsafe_unretained] = ACTIONS(2808),
    [anon_sym___unused] = ACTIONS(2808),
    [anon_sym___weak] = ACTIONS(2808),
    [anon_sym_alignas] = ACTIONS(2808),
    [anon_sym__Alignas] = ACTIONS(2808),
    [sym_primitive_type] = ACTIONS(2808),
    [anon_sym_enum] = ACTIONS(2808),
    [anon_sym_struct] = ACTIONS(2808),
    [anon_sym_union] = ACTIONS(2808),
    [anon_sym_if] = ACTIONS(2808),
    [anon_sym_switch] = ACTIONS(2808),
    [anon_sym_case] = ACTIONS(2808),
    [anon_sym_default] = ACTIONS(2808),
    [anon_sym_while] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2808),
    [anon_sym_in] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_break] = ACTIONS(2808),
    [anon_sym_continue] = ACTIONS(2808),
    [anon_sym_goto] = ACTIONS(2808),
    [anon_sym_DASH_DASH] = ACTIONS(2810),
    [anon_sym_PLUS_PLUS] = ACTIONS(2810),
    [anon_sym_sizeof] = ACTIONS(2808),
    [anon_sym___alignof__] = ACTIONS(2808),
    [anon_sym___alignof] = ACTIONS(2808),
    [anon_sym__alignof] = ACTIONS(2808),
    [anon_sym_alignof] = ACTIONS(2808),
    [anon_sym__Alignof] = ACTIONS(2808),
    [anon_sym_offsetof] = ACTIONS(2808),
    [anon_sym__Generic] = ACTIONS(2808),
    [anon_sym_asm] = ACTIONS(2808),
    [anon_sym___asm__] = ACTIONS(2808),
    [anon_sym___asm] = ACTIONS(2808),
    [sym_number_literal] = ACTIONS(2810),
    [anon_sym_L_SQUOTE] = ACTIONS(2810),
    [anon_sym_u_SQUOTE] = ACTIONS(2810),
    [anon_sym_U_SQUOTE] = ACTIONS(2810),
    [anon_sym_u8_SQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2810),
    [anon_sym_AT] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_L_DQUOTE] = ACTIONS(2810),
    [anon_sym_u_DQUOTE] = ACTIONS(2810),
    [anon_sym_U_DQUOTE] = ACTIONS(2810),
    [anon_sym_u8_DQUOTE] = ACTIONS(2810),
    [sym_true] = ACTIONS(2808),
    [sym_false] = ACTIONS(2808),
    [anon_sym_NULL] = ACTIONS(2808),
    [anon_sym_nullptr] = ACTIONS(2808),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2808),
    [anon_sym___typeof] = ACTIONS(2808),
    [anon_sym_typeof] = ACTIONS(2808),
    [anon_sym_ATimport] = ACTIONS(2810),
    [aux_sym_preproc_undef_token1] = ACTIONS(2808),
    [anon_sym_POUND] = ACTIONS(2808),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2808),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2808),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2808),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_API_AVAILABLE] = ACTIONS(2808),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_API_DEPRECATED] = ACTIONS(2808),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym___deprecated_msg] = ACTIONS(2808),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2808),
    [anon_sym_ATprotocol] = ACTIONS(2810),
    [anon_sym_ATinterface] = ACTIONS(2810),
    [anon_sym_ATimplementation] = ACTIONS(2810),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2810),
    [anon_sym_ATtry] = ACTIONS(2810),
    [anon_sym___try] = ACTIONS(2808),
    [anon_sym_ATthrow] = ACTIONS(2810),
    [anon_sym_ATselector] = ACTIONS(2810),
    [anon_sym_ATavailable] = ACTIONS(2810),
    [anon_sym___builtin_available] = ACTIONS(2808),
    [anon_sym_va_arg] = ACTIONS(2808),
    [anon_sym_ATencode] = ACTIONS(2810),
    [anon_sym_ATsynchronized] = ACTIONS(2810),
    [anon_sym_BOOL] = ACTIONS(2808),
    [anon_sym_IMP] = ACTIONS(2808),
    [anon_sym_SEL] = ACTIONS(2808),
    [anon_sym_Class] = ACTIONS(2808),
    [anon_sym_id] = ACTIONS(2808),
  },
  [STATE(886)] = {
    [sym_identifier] = ACTIONS(2812),
    [aux_sym_preproc_include_token1] = ACTIONS(2812),
    [aux_sym_preproc_include_token2] = ACTIONS(2812),
    [aux_sym_preproc_def_token1] = ACTIONS(2812),
    [aux_sym_preproc_if_token1] = ACTIONS(2812),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2812),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2812),
    [sym_preproc_directive] = ACTIONS(2812),
    [anon_sym_LPAREN2] = ACTIONS(2814),
    [anon_sym_BANG] = ACTIONS(2814),
    [anon_sym_TILDE] = ACTIONS(2814),
    [anon_sym_DASH] = ACTIONS(2812),
    [anon_sym_PLUS] = ACTIONS(2812),
    [anon_sym_STAR] = ACTIONS(2814),
    [anon_sym_CARET] = ACTIONS(2814),
    [anon_sym_AMP] = ACTIONS(2814),
    [anon_sym_SEMI] = ACTIONS(2814),
    [anon_sym___extension__] = ACTIONS(2812),
    [anon_sym_typedef] = ACTIONS(2812),
    [anon_sym_extern] = ACTIONS(2812),
    [anon_sym___attribute__] = ACTIONS(2812),
    [anon_sym___attribute] = ACTIONS(2812),
    [anon_sym_noreturn] = ACTIONS(2812),
    [anon_sym_LBRACK] = ACTIONS(2814),
    [anon_sym___declspec] = ACTIONS(2812),
    [anon_sym___cdecl] = ACTIONS(2812),
    [anon_sym___clrcall] = ACTIONS(2812),
    [anon_sym___stdcall] = ACTIONS(2812),
    [anon_sym___fastcall] = ACTIONS(2812),
    [anon_sym___thiscall] = ACTIONS(2812),
    [anon_sym___vectorcall] = ACTIONS(2812),
    [anon_sym_LBRACE] = ACTIONS(2814),
    [anon_sym_RBRACE] = ACTIONS(2814),
    [anon_sym_signed] = ACTIONS(2812),
    [anon_sym_unsigned] = ACTIONS(2812),
    [anon_sym_long] = ACTIONS(2812),
    [anon_sym_short] = ACTIONS(2812),
    [anon_sym_static] = ACTIONS(2812),
    [anon_sym_ATautoreleasepool] = ACTIONS(2814),
    [anon_sym_auto] = ACTIONS(2812),
    [anon_sym_register] = ACTIONS(2812),
    [anon_sym_inline] = ACTIONS(2812),
    [anon_sym___inline] = ACTIONS(2812),
    [anon_sym___inline__] = ACTIONS(2812),
    [anon_sym___forceinline] = ACTIONS(2812),
    [anon_sym_thread_local] = ACTIONS(2812),
    [anon_sym___thread] = ACTIONS(2812),
    [anon_sym_CG_EXTERN] = ACTIONS(2812),
    [anon_sym_CG_INLINE] = ACTIONS(2812),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2812),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2812),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2812),
    [anon_sym_IBOutlet] = ACTIONS(2812),
    [anon_sym_IBInspectable] = ACTIONS(2812),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2812),
    [anon_sym_NS_INLINE] = ACTIONS(2812),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2812),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2812),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2812),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2812),
    [anon_sym_const] = ACTIONS(2812),
    [anon_sym_constexpr] = ACTIONS(2812),
    [anon_sym_volatile] = ACTIONS(2812),
    [anon_sym_restrict] = ACTIONS(2812),
    [anon_sym___restrict__] = ACTIONS(2812),
    [anon_sym__Atomic] = ACTIONS(2812),
    [anon_sym__Noreturn] = ACTIONS(2812),
    [anon_sym__Nonnull] = ACTIONS(2812),
    [anon_sym_nullable] = ACTIONS(2812),
    [anon_sym__Complex] = ACTIONS(2812),
    [anon_sym__Nullable] = ACTIONS(2812),
    [anon_sym__Nullable_result] = ACTIONS(2812),
    [anon_sym__Null_unspecified] = ACTIONS(2812),
    [anon_sym___autoreleasing] = ACTIONS(2812),
    [anon_sym___block] = ACTIONS(2812),
    [anon_sym___bridge] = ACTIONS(2812),
    [anon_sym___bridge_retained] = ACTIONS(2812),
    [anon_sym___bridge_transfer] = ACTIONS(2812),
    [anon_sym___complex] = ACTIONS(2812),
    [anon_sym___const] = ACTIONS(2812),
    [anon_sym___imag] = ACTIONS(2812),
    [anon_sym___kindof] = ACTIONS(2812),
    [anon_sym___nonnull] = ACTIONS(2812),
    [anon_sym___nullable] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2812),
    [anon_sym___real] = ACTIONS(2812),
    [anon_sym___strong] = ACTIONS(2812),
    [anon_sym___unsafe_unretained] = ACTIONS(2812),
    [anon_sym___unused] = ACTIONS(2812),
    [anon_sym___weak] = ACTIONS(2812),
    [anon_sym_alignas] = ACTIONS(2812),
    [anon_sym__Alignas] = ACTIONS(2812),
    [sym_primitive_type] = ACTIONS(2812),
    [anon_sym_enum] = ACTIONS(2812),
    [anon_sym_struct] = ACTIONS(2812),
    [anon_sym_union] = ACTIONS(2812),
    [anon_sym_if] = ACTIONS(2812),
    [anon_sym_switch] = ACTIONS(2812),
    [anon_sym_case] = ACTIONS(2812),
    [anon_sym_default] = ACTIONS(2812),
    [anon_sym_while] = ACTIONS(2812),
    [anon_sym_do] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2812),
    [anon_sym_in] = ACTIONS(2812),
    [anon_sym_return] = ACTIONS(2812),
    [anon_sym_break] = ACTIONS(2812),
    [anon_sym_continue] = ACTIONS(2812),
    [anon_sym_goto] = ACTIONS(2812),
    [anon_sym_DASH_DASH] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_sizeof] = ACTIONS(2812),
    [anon_sym___alignof__] = ACTIONS(2812),
    [anon_sym___alignof] = ACTIONS(2812),
    [anon_sym__alignof] = ACTIONS(2812),
    [anon_sym_alignof] = ACTIONS(2812),
    [anon_sym__Alignof] = ACTIONS(2812),
    [anon_sym_offsetof] = ACTIONS(2812),
    [anon_sym__Generic] = ACTIONS(2812),
    [anon_sym_asm] = ACTIONS(2812),
    [anon_sym___asm__] = ACTIONS(2812),
    [anon_sym___asm] = ACTIONS(2812),
    [sym_number_literal] = ACTIONS(2814),
    [anon_sym_L_SQUOTE] = ACTIONS(2814),
    [anon_sym_u_SQUOTE] = ACTIONS(2814),
    [anon_sym_U_SQUOTE] = ACTIONS(2814),
    [anon_sym_u8_SQUOTE] = ACTIONS(2814),
    [anon_sym_SQUOTE] = ACTIONS(2814),
    [anon_sym_AT] = ACTIONS(2812),
    [anon_sym_DQUOTE] = ACTIONS(2814),
    [anon_sym_L_DQUOTE] = ACTIONS(2814),
    [anon_sym_u_DQUOTE] = ACTIONS(2814),
    [anon_sym_U_DQUOTE] = ACTIONS(2814),
    [anon_sym_u8_DQUOTE] = ACTIONS(2814),
    [sym_true] = ACTIONS(2812),
    [sym_false] = ACTIONS(2812),
    [anon_sym_NULL] = ACTIONS(2812),
    [anon_sym_nullptr] = ACTIONS(2812),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2812),
    [anon_sym___typeof] = ACTIONS(2812),
    [anon_sym_typeof] = ACTIONS(2812),
    [anon_sym_ATimport] = ACTIONS(2814),
    [aux_sym_preproc_undef_token1] = ACTIONS(2812),
    [anon_sym_POUND] = ACTIONS(2812),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2812),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2812),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2812),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2812),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2812),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2812),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2812),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_API_AVAILABLE] = ACTIONS(2812),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_API_DEPRECATED] = ACTIONS(2812),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2812),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym___deprecated_msg] = ACTIONS(2812),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2812),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2812),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2812),
    [anon_sym_ATprotocol] = ACTIONS(2814),
    [anon_sym_ATinterface] = ACTIONS(2814),
    [anon_sym_ATimplementation] = ACTIONS(2814),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2814),
    [anon_sym_ATtry] = ACTIONS(2814),
    [anon_sym___try] = ACTIONS(2812),
    [anon_sym_ATthrow] = ACTIONS(2814),
    [anon_sym_ATselector] = ACTIONS(2814),
    [anon_sym_ATavailable] = ACTIONS(2814),
    [anon_sym___builtin_available] = ACTIONS(2812),
    [anon_sym_va_arg] = ACTIONS(2812),
    [anon_sym_ATencode] = ACTIONS(2814),
    [anon_sym_ATsynchronized] = ACTIONS(2814),
    [anon_sym_BOOL] = ACTIONS(2812),
    [anon_sym_IMP] = ACTIONS(2812),
    [anon_sym_SEL] = ACTIONS(2812),
    [anon_sym_Class] = ACTIONS(2812),
    [anon_sym_id] = ACTIONS(2812),
  },
  [STATE(887)] = {
    [sym_identifier] = ACTIONS(2816),
    [aux_sym_preproc_include_token1] = ACTIONS(2816),
    [aux_sym_preproc_include_token2] = ACTIONS(2816),
    [aux_sym_preproc_def_token1] = ACTIONS(2816),
    [aux_sym_preproc_if_token1] = ACTIONS(2816),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2816),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2816),
    [sym_preproc_directive] = ACTIONS(2816),
    [anon_sym_LPAREN2] = ACTIONS(2818),
    [anon_sym_BANG] = ACTIONS(2818),
    [anon_sym_TILDE] = ACTIONS(2818),
    [anon_sym_DASH] = ACTIONS(2816),
    [anon_sym_PLUS] = ACTIONS(2816),
    [anon_sym_STAR] = ACTIONS(2818),
    [anon_sym_CARET] = ACTIONS(2818),
    [anon_sym_AMP] = ACTIONS(2818),
    [anon_sym_SEMI] = ACTIONS(2818),
    [anon_sym___extension__] = ACTIONS(2816),
    [anon_sym_typedef] = ACTIONS(2816),
    [anon_sym_extern] = ACTIONS(2816),
    [anon_sym___attribute__] = ACTIONS(2816),
    [anon_sym___attribute] = ACTIONS(2816),
    [anon_sym_noreturn] = ACTIONS(2816),
    [anon_sym_LBRACK] = ACTIONS(2818),
    [anon_sym___declspec] = ACTIONS(2816),
    [anon_sym___cdecl] = ACTIONS(2816),
    [anon_sym___clrcall] = ACTIONS(2816),
    [anon_sym___stdcall] = ACTIONS(2816),
    [anon_sym___fastcall] = ACTIONS(2816),
    [anon_sym___thiscall] = ACTIONS(2816),
    [anon_sym___vectorcall] = ACTIONS(2816),
    [anon_sym_LBRACE] = ACTIONS(2818),
    [anon_sym_RBRACE] = ACTIONS(2818),
    [anon_sym_signed] = ACTIONS(2816),
    [anon_sym_unsigned] = ACTIONS(2816),
    [anon_sym_long] = ACTIONS(2816),
    [anon_sym_short] = ACTIONS(2816),
    [anon_sym_static] = ACTIONS(2816),
    [anon_sym_ATautoreleasepool] = ACTIONS(2818),
    [anon_sym_auto] = ACTIONS(2816),
    [anon_sym_register] = ACTIONS(2816),
    [anon_sym_inline] = ACTIONS(2816),
    [anon_sym___inline] = ACTIONS(2816),
    [anon_sym___inline__] = ACTIONS(2816),
    [anon_sym___forceinline] = ACTIONS(2816),
    [anon_sym_thread_local] = ACTIONS(2816),
    [anon_sym___thread] = ACTIONS(2816),
    [anon_sym_CG_EXTERN] = ACTIONS(2816),
    [anon_sym_CG_INLINE] = ACTIONS(2816),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2816),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2816),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2816),
    [anon_sym_IBOutlet] = ACTIONS(2816),
    [anon_sym_IBInspectable] = ACTIONS(2816),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2816),
    [anon_sym_NS_INLINE] = ACTIONS(2816),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2816),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2816),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2816),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2816),
    [anon_sym_const] = ACTIONS(2816),
    [anon_sym_constexpr] = ACTIONS(2816),
    [anon_sym_volatile] = ACTIONS(2816),
    [anon_sym_restrict] = ACTIONS(2816),
    [anon_sym___restrict__] = ACTIONS(2816),
    [anon_sym__Atomic] = ACTIONS(2816),
    [anon_sym__Noreturn] = ACTIONS(2816),
    [anon_sym__Nonnull] = ACTIONS(2816),
    [anon_sym_nullable] = ACTIONS(2816),
    [anon_sym__Complex] = ACTIONS(2816),
    [anon_sym__Nullable] = ACTIONS(2816),
    [anon_sym__Nullable_result] = ACTIONS(2816),
    [anon_sym__Null_unspecified] = ACTIONS(2816),
    [anon_sym___autoreleasing] = ACTIONS(2816),
    [anon_sym___block] = ACTIONS(2816),
    [anon_sym___bridge] = ACTIONS(2816),
    [anon_sym___bridge_retained] = ACTIONS(2816),
    [anon_sym___bridge_transfer] = ACTIONS(2816),
    [anon_sym___complex] = ACTIONS(2816),
    [anon_sym___const] = ACTIONS(2816),
    [anon_sym___imag] = ACTIONS(2816),
    [anon_sym___kindof] = ACTIONS(2816),
    [anon_sym___nonnull] = ACTIONS(2816),
    [anon_sym___nullable] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2816),
    [anon_sym___real] = ACTIONS(2816),
    [anon_sym___strong] = ACTIONS(2816),
    [anon_sym___unsafe_unretained] = ACTIONS(2816),
    [anon_sym___unused] = ACTIONS(2816),
    [anon_sym___weak] = ACTIONS(2816),
    [anon_sym_alignas] = ACTIONS(2816),
    [anon_sym__Alignas] = ACTIONS(2816),
    [sym_primitive_type] = ACTIONS(2816),
    [anon_sym_enum] = ACTIONS(2816),
    [anon_sym_struct] = ACTIONS(2816),
    [anon_sym_union] = ACTIONS(2816),
    [anon_sym_if] = ACTIONS(2816),
    [anon_sym_switch] = ACTIONS(2816),
    [anon_sym_case] = ACTIONS(2816),
    [anon_sym_default] = ACTIONS(2816),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_do] = ACTIONS(2816),
    [anon_sym_for] = ACTIONS(2816),
    [anon_sym_in] = ACTIONS(2816),
    [anon_sym_return] = ACTIONS(2816),
    [anon_sym_break] = ACTIONS(2816),
    [anon_sym_continue] = ACTIONS(2816),
    [anon_sym_goto] = ACTIONS(2816),
    [anon_sym_DASH_DASH] = ACTIONS(2818),
    [anon_sym_PLUS_PLUS] = ACTIONS(2818),
    [anon_sym_sizeof] = ACTIONS(2816),
    [anon_sym___alignof__] = ACTIONS(2816),
    [anon_sym___alignof] = ACTIONS(2816),
    [anon_sym__alignof] = ACTIONS(2816),
    [anon_sym_alignof] = ACTIONS(2816),
    [anon_sym__Alignof] = ACTIONS(2816),
    [anon_sym_offsetof] = ACTIONS(2816),
    [anon_sym__Generic] = ACTIONS(2816),
    [anon_sym_asm] = ACTIONS(2816),
    [anon_sym___asm__] = ACTIONS(2816),
    [anon_sym___asm] = ACTIONS(2816),
    [sym_number_literal] = ACTIONS(2818),
    [anon_sym_L_SQUOTE] = ACTIONS(2818),
    [anon_sym_u_SQUOTE] = ACTIONS(2818),
    [anon_sym_U_SQUOTE] = ACTIONS(2818),
    [anon_sym_u8_SQUOTE] = ACTIONS(2818),
    [anon_sym_SQUOTE] = ACTIONS(2818),
    [anon_sym_AT] = ACTIONS(2816),
    [anon_sym_DQUOTE] = ACTIONS(2818),
    [anon_sym_L_DQUOTE] = ACTIONS(2818),
    [anon_sym_u_DQUOTE] = ACTIONS(2818),
    [anon_sym_U_DQUOTE] = ACTIONS(2818),
    [anon_sym_u8_DQUOTE] = ACTIONS(2818),
    [sym_true] = ACTIONS(2816),
    [sym_false] = ACTIONS(2816),
    [anon_sym_NULL] = ACTIONS(2816),
    [anon_sym_nullptr] = ACTIONS(2816),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2816),
    [anon_sym___typeof] = ACTIONS(2816),
    [anon_sym_typeof] = ACTIONS(2816),
    [anon_sym_ATimport] = ACTIONS(2818),
    [aux_sym_preproc_undef_token1] = ACTIONS(2816),
    [anon_sym_POUND] = ACTIONS(2816),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2816),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2816),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2816),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2816),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2816),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2816),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2816),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_API_AVAILABLE] = ACTIONS(2816),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_API_DEPRECATED] = ACTIONS(2816),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2816),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym___deprecated_msg] = ACTIONS(2816),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2816),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2816),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2816),
    [anon_sym_ATprotocol] = ACTIONS(2818),
    [anon_sym_ATinterface] = ACTIONS(2818),
    [anon_sym_ATimplementation] = ACTIONS(2818),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2818),
    [anon_sym_ATtry] = ACTIONS(2818),
    [anon_sym___try] = ACTIONS(2816),
    [anon_sym_ATthrow] = ACTIONS(2818),
    [anon_sym_ATselector] = ACTIONS(2818),
    [anon_sym_ATavailable] = ACTIONS(2818),
    [anon_sym___builtin_available] = ACTIONS(2816),
    [anon_sym_va_arg] = ACTIONS(2816),
    [anon_sym_ATencode] = ACTIONS(2818),
    [anon_sym_ATsynchronized] = ACTIONS(2818),
    [anon_sym_BOOL] = ACTIONS(2816),
    [anon_sym_IMP] = ACTIONS(2816),
    [anon_sym_SEL] = ACTIONS(2816),
    [anon_sym_Class] = ACTIONS(2816),
    [anon_sym_id] = ACTIONS(2816),
  },
  [STATE(888)] = {
    [sym_identifier] = ACTIONS(2820),
    [aux_sym_preproc_include_token1] = ACTIONS(2820),
    [aux_sym_preproc_include_token2] = ACTIONS(2820),
    [aux_sym_preproc_def_token1] = ACTIONS(2820),
    [aux_sym_preproc_if_token1] = ACTIONS(2820),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2820),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2820),
    [sym_preproc_directive] = ACTIONS(2820),
    [anon_sym_LPAREN2] = ACTIONS(2822),
    [anon_sym_BANG] = ACTIONS(2822),
    [anon_sym_TILDE] = ACTIONS(2822),
    [anon_sym_DASH] = ACTIONS(2820),
    [anon_sym_PLUS] = ACTIONS(2820),
    [anon_sym_STAR] = ACTIONS(2822),
    [anon_sym_CARET] = ACTIONS(2822),
    [anon_sym_AMP] = ACTIONS(2822),
    [anon_sym_SEMI] = ACTIONS(2822),
    [anon_sym___extension__] = ACTIONS(2820),
    [anon_sym_typedef] = ACTIONS(2820),
    [anon_sym_extern] = ACTIONS(2820),
    [anon_sym___attribute__] = ACTIONS(2820),
    [anon_sym___attribute] = ACTIONS(2820),
    [anon_sym_noreturn] = ACTIONS(2820),
    [anon_sym_LBRACK] = ACTIONS(2822),
    [anon_sym___declspec] = ACTIONS(2820),
    [anon_sym___cdecl] = ACTIONS(2820),
    [anon_sym___clrcall] = ACTIONS(2820),
    [anon_sym___stdcall] = ACTIONS(2820),
    [anon_sym___fastcall] = ACTIONS(2820),
    [anon_sym___thiscall] = ACTIONS(2820),
    [anon_sym___vectorcall] = ACTIONS(2820),
    [anon_sym_LBRACE] = ACTIONS(2822),
    [anon_sym_RBRACE] = ACTIONS(2822),
    [anon_sym_signed] = ACTIONS(2820),
    [anon_sym_unsigned] = ACTIONS(2820),
    [anon_sym_long] = ACTIONS(2820),
    [anon_sym_short] = ACTIONS(2820),
    [anon_sym_static] = ACTIONS(2820),
    [anon_sym_ATautoreleasepool] = ACTIONS(2822),
    [anon_sym_auto] = ACTIONS(2820),
    [anon_sym_register] = ACTIONS(2820),
    [anon_sym_inline] = ACTIONS(2820),
    [anon_sym___inline] = ACTIONS(2820),
    [anon_sym___inline__] = ACTIONS(2820),
    [anon_sym___forceinline] = ACTIONS(2820),
    [anon_sym_thread_local] = ACTIONS(2820),
    [anon_sym___thread] = ACTIONS(2820),
    [anon_sym_CG_EXTERN] = ACTIONS(2820),
    [anon_sym_CG_INLINE] = ACTIONS(2820),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2820),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2820),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2820),
    [anon_sym_IBOutlet] = ACTIONS(2820),
    [anon_sym_IBInspectable] = ACTIONS(2820),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2820),
    [anon_sym_NS_INLINE] = ACTIONS(2820),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2820),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2820),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2820),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2820),
    [anon_sym_const] = ACTIONS(2820),
    [anon_sym_constexpr] = ACTIONS(2820),
    [anon_sym_volatile] = ACTIONS(2820),
    [anon_sym_restrict] = ACTIONS(2820),
    [anon_sym___restrict__] = ACTIONS(2820),
    [anon_sym__Atomic] = ACTIONS(2820),
    [anon_sym__Noreturn] = ACTIONS(2820),
    [anon_sym__Nonnull] = ACTIONS(2820),
    [anon_sym_nullable] = ACTIONS(2820),
    [anon_sym__Complex] = ACTIONS(2820),
    [anon_sym__Nullable] = ACTIONS(2820),
    [anon_sym__Nullable_result] = ACTIONS(2820),
    [anon_sym__Null_unspecified] = ACTIONS(2820),
    [anon_sym___autoreleasing] = ACTIONS(2820),
    [anon_sym___block] = ACTIONS(2820),
    [anon_sym___bridge] = ACTIONS(2820),
    [anon_sym___bridge_retained] = ACTIONS(2820),
    [anon_sym___bridge_transfer] = ACTIONS(2820),
    [anon_sym___complex] = ACTIONS(2820),
    [anon_sym___const] = ACTIONS(2820),
    [anon_sym___imag] = ACTIONS(2820),
    [anon_sym___kindof] = ACTIONS(2820),
    [anon_sym___nonnull] = ACTIONS(2820),
    [anon_sym___nullable] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2820),
    [anon_sym___real] = ACTIONS(2820),
    [anon_sym___strong] = ACTIONS(2820),
    [anon_sym___unsafe_unretained] = ACTIONS(2820),
    [anon_sym___unused] = ACTIONS(2820),
    [anon_sym___weak] = ACTIONS(2820),
    [anon_sym_alignas] = ACTIONS(2820),
    [anon_sym__Alignas] = ACTIONS(2820),
    [sym_primitive_type] = ACTIONS(2820),
    [anon_sym_enum] = ACTIONS(2820),
    [anon_sym_struct] = ACTIONS(2820),
    [anon_sym_union] = ACTIONS(2820),
    [anon_sym_if] = ACTIONS(2820),
    [anon_sym_switch] = ACTIONS(2820),
    [anon_sym_case] = ACTIONS(2820),
    [anon_sym_default] = ACTIONS(2820),
    [anon_sym_while] = ACTIONS(2820),
    [anon_sym_do] = ACTIONS(2820),
    [anon_sym_for] = ACTIONS(2820),
    [anon_sym_in] = ACTIONS(2820),
    [anon_sym_return] = ACTIONS(2820),
    [anon_sym_break] = ACTIONS(2820),
    [anon_sym_continue] = ACTIONS(2820),
    [anon_sym_goto] = ACTIONS(2820),
    [anon_sym_DASH_DASH] = ACTIONS(2822),
    [anon_sym_PLUS_PLUS] = ACTIONS(2822),
    [anon_sym_sizeof] = ACTIONS(2820),
    [anon_sym___alignof__] = ACTIONS(2820),
    [anon_sym___alignof] = ACTIONS(2820),
    [anon_sym__alignof] = ACTIONS(2820),
    [anon_sym_alignof] = ACTIONS(2820),
    [anon_sym__Alignof] = ACTIONS(2820),
    [anon_sym_offsetof] = ACTIONS(2820),
    [anon_sym__Generic] = ACTIONS(2820),
    [anon_sym_asm] = ACTIONS(2820),
    [anon_sym___asm__] = ACTIONS(2820),
    [anon_sym___asm] = ACTIONS(2820),
    [sym_number_literal] = ACTIONS(2822),
    [anon_sym_L_SQUOTE] = ACTIONS(2822),
    [anon_sym_u_SQUOTE] = ACTIONS(2822),
    [anon_sym_U_SQUOTE] = ACTIONS(2822),
    [anon_sym_u8_SQUOTE] = ACTIONS(2822),
    [anon_sym_SQUOTE] = ACTIONS(2822),
    [anon_sym_AT] = ACTIONS(2820),
    [anon_sym_DQUOTE] = ACTIONS(2822),
    [anon_sym_L_DQUOTE] = ACTIONS(2822),
    [anon_sym_u_DQUOTE] = ACTIONS(2822),
    [anon_sym_U_DQUOTE] = ACTIONS(2822),
    [anon_sym_u8_DQUOTE] = ACTIONS(2822),
    [sym_true] = ACTIONS(2820),
    [sym_false] = ACTIONS(2820),
    [anon_sym_NULL] = ACTIONS(2820),
    [anon_sym_nullptr] = ACTIONS(2820),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2820),
    [anon_sym___typeof] = ACTIONS(2820),
    [anon_sym_typeof] = ACTIONS(2820),
    [anon_sym_ATimport] = ACTIONS(2822),
    [aux_sym_preproc_undef_token1] = ACTIONS(2820),
    [anon_sym_POUND] = ACTIONS(2820),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2820),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2820),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2820),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2820),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2820),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2820),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2820),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_API_AVAILABLE] = ACTIONS(2820),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_API_DEPRECATED] = ACTIONS(2820),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2820),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym___deprecated_msg] = ACTIONS(2820),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2820),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2820),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2820),
    [anon_sym_ATprotocol] = ACTIONS(2822),
    [anon_sym_ATinterface] = ACTIONS(2822),
    [anon_sym_ATimplementation] = ACTIONS(2822),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2822),
    [anon_sym_ATtry] = ACTIONS(2822),
    [anon_sym___try] = ACTIONS(2820),
    [anon_sym_ATthrow] = ACTIONS(2822),
    [anon_sym_ATselector] = ACTIONS(2822),
    [anon_sym_ATavailable] = ACTIONS(2822),
    [anon_sym___builtin_available] = ACTIONS(2820),
    [anon_sym_va_arg] = ACTIONS(2820),
    [anon_sym_ATencode] = ACTIONS(2822),
    [anon_sym_ATsynchronized] = ACTIONS(2822),
    [anon_sym_BOOL] = ACTIONS(2820),
    [anon_sym_IMP] = ACTIONS(2820),
    [anon_sym_SEL] = ACTIONS(2820),
    [anon_sym_Class] = ACTIONS(2820),
    [anon_sym_id] = ACTIONS(2820),
  },
  [STATE(889)] = {
    [sym_identifier] = ACTIONS(2824),
    [aux_sym_preproc_include_token1] = ACTIONS(2824),
    [aux_sym_preproc_include_token2] = ACTIONS(2824),
    [aux_sym_preproc_def_token1] = ACTIONS(2824),
    [aux_sym_preproc_if_token1] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2824),
    [sym_preproc_directive] = ACTIONS(2824),
    [anon_sym_LPAREN2] = ACTIONS(2826),
    [anon_sym_BANG] = ACTIONS(2826),
    [anon_sym_TILDE] = ACTIONS(2826),
    [anon_sym_DASH] = ACTIONS(2824),
    [anon_sym_PLUS] = ACTIONS(2824),
    [anon_sym_STAR] = ACTIONS(2826),
    [anon_sym_CARET] = ACTIONS(2826),
    [anon_sym_AMP] = ACTIONS(2826),
    [anon_sym_SEMI] = ACTIONS(2826),
    [anon_sym___extension__] = ACTIONS(2824),
    [anon_sym_typedef] = ACTIONS(2824),
    [anon_sym_extern] = ACTIONS(2824),
    [anon_sym___attribute__] = ACTIONS(2824),
    [anon_sym___attribute] = ACTIONS(2824),
    [anon_sym_noreturn] = ACTIONS(2824),
    [anon_sym_LBRACK] = ACTIONS(2826),
    [anon_sym___declspec] = ACTIONS(2824),
    [anon_sym___cdecl] = ACTIONS(2824),
    [anon_sym___clrcall] = ACTIONS(2824),
    [anon_sym___stdcall] = ACTIONS(2824),
    [anon_sym___fastcall] = ACTIONS(2824),
    [anon_sym___thiscall] = ACTIONS(2824),
    [anon_sym___vectorcall] = ACTIONS(2824),
    [anon_sym_LBRACE] = ACTIONS(2826),
    [anon_sym_RBRACE] = ACTIONS(2826),
    [anon_sym_signed] = ACTIONS(2824),
    [anon_sym_unsigned] = ACTIONS(2824),
    [anon_sym_long] = ACTIONS(2824),
    [anon_sym_short] = ACTIONS(2824),
    [anon_sym_static] = ACTIONS(2824),
    [anon_sym_ATautoreleasepool] = ACTIONS(2826),
    [anon_sym_auto] = ACTIONS(2824),
    [anon_sym_register] = ACTIONS(2824),
    [anon_sym_inline] = ACTIONS(2824),
    [anon_sym___inline] = ACTIONS(2824),
    [anon_sym___inline__] = ACTIONS(2824),
    [anon_sym___forceinline] = ACTIONS(2824),
    [anon_sym_thread_local] = ACTIONS(2824),
    [anon_sym___thread] = ACTIONS(2824),
    [anon_sym_CG_EXTERN] = ACTIONS(2824),
    [anon_sym_CG_INLINE] = ACTIONS(2824),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2824),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2824),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2824),
    [anon_sym_IBOutlet] = ACTIONS(2824),
    [anon_sym_IBInspectable] = ACTIONS(2824),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2824),
    [anon_sym_NS_INLINE] = ACTIONS(2824),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2824),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2824),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2824),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2824),
    [anon_sym_const] = ACTIONS(2824),
    [anon_sym_constexpr] = ACTIONS(2824),
    [anon_sym_volatile] = ACTIONS(2824),
    [anon_sym_restrict] = ACTIONS(2824),
    [anon_sym___restrict__] = ACTIONS(2824),
    [anon_sym__Atomic] = ACTIONS(2824),
    [anon_sym__Noreturn] = ACTIONS(2824),
    [anon_sym__Nonnull] = ACTIONS(2824),
    [anon_sym_nullable] = ACTIONS(2824),
    [anon_sym__Complex] = ACTIONS(2824),
    [anon_sym__Nullable] = ACTIONS(2824),
    [anon_sym__Nullable_result] = ACTIONS(2824),
    [anon_sym__Null_unspecified] = ACTIONS(2824),
    [anon_sym___autoreleasing] = ACTIONS(2824),
    [anon_sym___block] = ACTIONS(2824),
    [anon_sym___bridge] = ACTIONS(2824),
    [anon_sym___bridge_retained] = ACTIONS(2824),
    [anon_sym___bridge_transfer] = ACTIONS(2824),
    [anon_sym___complex] = ACTIONS(2824),
    [anon_sym___const] = ACTIONS(2824),
    [anon_sym___imag] = ACTIONS(2824),
    [anon_sym___kindof] = ACTIONS(2824),
    [anon_sym___nonnull] = ACTIONS(2824),
    [anon_sym___nullable] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2824),
    [anon_sym___real] = ACTIONS(2824),
    [anon_sym___strong] = ACTIONS(2824),
    [anon_sym___unsafe_unretained] = ACTIONS(2824),
    [anon_sym___unused] = ACTIONS(2824),
    [anon_sym___weak] = ACTIONS(2824),
    [anon_sym_alignas] = ACTIONS(2824),
    [anon_sym__Alignas] = ACTIONS(2824),
    [sym_primitive_type] = ACTIONS(2824),
    [anon_sym_enum] = ACTIONS(2824),
    [anon_sym_struct] = ACTIONS(2824),
    [anon_sym_union] = ACTIONS(2824),
    [anon_sym_if] = ACTIONS(2824),
    [anon_sym_switch] = ACTIONS(2824),
    [anon_sym_case] = ACTIONS(2824),
    [anon_sym_default] = ACTIONS(2824),
    [anon_sym_while] = ACTIONS(2824),
    [anon_sym_do] = ACTIONS(2824),
    [anon_sym_for] = ACTIONS(2824),
    [anon_sym_in] = ACTIONS(2824),
    [anon_sym_return] = ACTIONS(2824),
    [anon_sym_break] = ACTIONS(2824),
    [anon_sym_continue] = ACTIONS(2824),
    [anon_sym_goto] = ACTIONS(2824),
    [anon_sym_DASH_DASH] = ACTIONS(2826),
    [anon_sym_PLUS_PLUS] = ACTIONS(2826),
    [anon_sym_sizeof] = ACTIONS(2824),
    [anon_sym___alignof__] = ACTIONS(2824),
    [anon_sym___alignof] = ACTIONS(2824),
    [anon_sym__alignof] = ACTIONS(2824),
    [anon_sym_alignof] = ACTIONS(2824),
    [anon_sym__Alignof] = ACTIONS(2824),
    [anon_sym_offsetof] = ACTIONS(2824),
    [anon_sym__Generic] = ACTIONS(2824),
    [anon_sym_asm] = ACTIONS(2824),
    [anon_sym___asm__] = ACTIONS(2824),
    [anon_sym___asm] = ACTIONS(2824),
    [sym_number_literal] = ACTIONS(2826),
    [anon_sym_L_SQUOTE] = ACTIONS(2826),
    [anon_sym_u_SQUOTE] = ACTIONS(2826),
    [anon_sym_U_SQUOTE] = ACTIONS(2826),
    [anon_sym_u8_SQUOTE] = ACTIONS(2826),
    [anon_sym_SQUOTE] = ACTIONS(2826),
    [anon_sym_AT] = ACTIONS(2824),
    [anon_sym_DQUOTE] = ACTIONS(2826),
    [anon_sym_L_DQUOTE] = ACTIONS(2826),
    [anon_sym_u_DQUOTE] = ACTIONS(2826),
    [anon_sym_U_DQUOTE] = ACTIONS(2826),
    [anon_sym_u8_DQUOTE] = ACTIONS(2826),
    [sym_true] = ACTIONS(2824),
    [sym_false] = ACTIONS(2824),
    [anon_sym_NULL] = ACTIONS(2824),
    [anon_sym_nullptr] = ACTIONS(2824),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2824),
    [anon_sym___typeof] = ACTIONS(2824),
    [anon_sym_typeof] = ACTIONS(2824),
    [anon_sym_ATimport] = ACTIONS(2826),
    [aux_sym_preproc_undef_token1] = ACTIONS(2824),
    [anon_sym_POUND] = ACTIONS(2824),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2824),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2824),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2824),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2824),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2824),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2824),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2824),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_API_AVAILABLE] = ACTIONS(2824),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_API_DEPRECATED] = ACTIONS(2824),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2824),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym___deprecated_msg] = ACTIONS(2824),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2824),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2824),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2824),
    [anon_sym_ATprotocol] = ACTIONS(2826),
    [anon_sym_ATinterface] = ACTIONS(2826),
    [anon_sym_ATimplementation] = ACTIONS(2826),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2826),
    [anon_sym_ATtry] = ACTIONS(2826),
    [anon_sym___try] = ACTIONS(2824),
    [anon_sym_ATthrow] = ACTIONS(2826),
    [anon_sym_ATselector] = ACTIONS(2826),
    [anon_sym_ATavailable] = ACTIONS(2826),
    [anon_sym___builtin_available] = ACTIONS(2824),
    [anon_sym_va_arg] = ACTIONS(2824),
    [anon_sym_ATencode] = ACTIONS(2826),
    [anon_sym_ATsynchronized] = ACTIONS(2826),
    [anon_sym_BOOL] = ACTIONS(2824),
    [anon_sym_IMP] = ACTIONS(2824),
    [anon_sym_SEL] = ACTIONS(2824),
    [anon_sym_Class] = ACTIONS(2824),
    [anon_sym_id] = ACTIONS(2824),
  },
  [STATE(890)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_RBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(891)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_RBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(892)] = {
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_RBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATtry] = ACTIONS(2754),
    [anon_sym___try] = ACTIONS(2752),
    [anon_sym_ATthrow] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_ATsynchronized] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(893)] = {
    [sym_identifier] = ACTIONS(2832),
    [aux_sym_preproc_include_token1] = ACTIONS(2832),
    [aux_sym_preproc_include_token2] = ACTIONS(2832),
    [aux_sym_preproc_def_token1] = ACTIONS(2832),
    [aux_sym_preproc_if_token1] = ACTIONS(2832),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2832),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2832),
    [sym_preproc_directive] = ACTIONS(2832),
    [anon_sym_LPAREN2] = ACTIONS(2834),
    [anon_sym_BANG] = ACTIONS(2834),
    [anon_sym_TILDE] = ACTIONS(2834),
    [anon_sym_DASH] = ACTIONS(2832),
    [anon_sym_PLUS] = ACTIONS(2832),
    [anon_sym_STAR] = ACTIONS(2834),
    [anon_sym_CARET] = ACTIONS(2834),
    [anon_sym_AMP] = ACTIONS(2834),
    [anon_sym_SEMI] = ACTIONS(2834),
    [anon_sym___extension__] = ACTIONS(2832),
    [anon_sym_typedef] = ACTIONS(2832),
    [anon_sym_extern] = ACTIONS(2832),
    [anon_sym___attribute__] = ACTIONS(2832),
    [anon_sym___attribute] = ACTIONS(2832),
    [anon_sym_noreturn] = ACTIONS(2832),
    [anon_sym_LBRACK] = ACTIONS(2834),
    [anon_sym___declspec] = ACTIONS(2832),
    [anon_sym___cdecl] = ACTIONS(2832),
    [anon_sym___clrcall] = ACTIONS(2832),
    [anon_sym___stdcall] = ACTIONS(2832),
    [anon_sym___fastcall] = ACTIONS(2832),
    [anon_sym___thiscall] = ACTIONS(2832),
    [anon_sym___vectorcall] = ACTIONS(2832),
    [anon_sym_LBRACE] = ACTIONS(2834),
    [anon_sym_RBRACE] = ACTIONS(2834),
    [anon_sym_signed] = ACTIONS(2832),
    [anon_sym_unsigned] = ACTIONS(2832),
    [anon_sym_long] = ACTIONS(2832),
    [anon_sym_short] = ACTIONS(2832),
    [anon_sym_static] = ACTIONS(2832),
    [anon_sym_ATautoreleasepool] = ACTIONS(2834),
    [anon_sym_auto] = ACTIONS(2832),
    [anon_sym_register] = ACTIONS(2832),
    [anon_sym_inline] = ACTIONS(2832),
    [anon_sym___inline] = ACTIONS(2832),
    [anon_sym___inline__] = ACTIONS(2832),
    [anon_sym___forceinline] = ACTIONS(2832),
    [anon_sym_thread_local] = ACTIONS(2832),
    [anon_sym___thread] = ACTIONS(2832),
    [anon_sym_CG_EXTERN] = ACTIONS(2832),
    [anon_sym_CG_INLINE] = ACTIONS(2832),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2832),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2832),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2832),
    [anon_sym_IBOutlet] = ACTIONS(2832),
    [anon_sym_IBInspectable] = ACTIONS(2832),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2832),
    [anon_sym_NS_INLINE] = ACTIONS(2832),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2832),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2832),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2832),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2832),
    [anon_sym_const] = ACTIONS(2832),
    [anon_sym_constexpr] = ACTIONS(2832),
    [anon_sym_volatile] = ACTIONS(2832),
    [anon_sym_restrict] = ACTIONS(2832),
    [anon_sym___restrict__] = ACTIONS(2832),
    [anon_sym__Atomic] = ACTIONS(2832),
    [anon_sym__Noreturn] = ACTIONS(2832),
    [anon_sym__Nonnull] = ACTIONS(2832),
    [anon_sym_nullable] = ACTIONS(2832),
    [anon_sym__Complex] = ACTIONS(2832),
    [anon_sym__Nullable] = ACTIONS(2832),
    [anon_sym__Nullable_result] = ACTIONS(2832),
    [anon_sym__Null_unspecified] = ACTIONS(2832),
    [anon_sym___autoreleasing] = ACTIONS(2832),
    [anon_sym___block] = ACTIONS(2832),
    [anon_sym___bridge] = ACTIONS(2832),
    [anon_sym___bridge_retained] = ACTIONS(2832),
    [anon_sym___bridge_transfer] = ACTIONS(2832),
    [anon_sym___complex] = ACTIONS(2832),
    [anon_sym___const] = ACTIONS(2832),
    [anon_sym___imag] = ACTIONS(2832),
    [anon_sym___kindof] = ACTIONS(2832),
    [anon_sym___nonnull] = ACTIONS(2832),
    [anon_sym___nullable] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2832),
    [anon_sym___real] = ACTIONS(2832),
    [anon_sym___strong] = ACTIONS(2832),
    [anon_sym___unsafe_unretained] = ACTIONS(2832),
    [anon_sym___unused] = ACTIONS(2832),
    [anon_sym___weak] = ACTIONS(2832),
    [anon_sym_alignas] = ACTIONS(2832),
    [anon_sym__Alignas] = ACTIONS(2832),
    [sym_primitive_type] = ACTIONS(2832),
    [anon_sym_enum] = ACTIONS(2832),
    [anon_sym_struct] = ACTIONS(2832),
    [anon_sym_union] = ACTIONS(2832),
    [anon_sym_if] = ACTIONS(2832),
    [anon_sym_switch] = ACTIONS(2832),
    [anon_sym_case] = ACTIONS(2832),
    [anon_sym_default] = ACTIONS(2832),
    [anon_sym_while] = ACTIONS(2832),
    [anon_sym_do] = ACTIONS(2832),
    [anon_sym_for] = ACTIONS(2832),
    [anon_sym_in] = ACTIONS(2832),
    [anon_sym_return] = ACTIONS(2832),
    [anon_sym_break] = ACTIONS(2832),
    [anon_sym_continue] = ACTIONS(2832),
    [anon_sym_goto] = ACTIONS(2832),
    [anon_sym_DASH_DASH] = ACTIONS(2834),
    [anon_sym_PLUS_PLUS] = ACTIONS(2834),
    [anon_sym_sizeof] = ACTIONS(2832),
    [anon_sym___alignof__] = ACTIONS(2832),
    [anon_sym___alignof] = ACTIONS(2832),
    [anon_sym__alignof] = ACTIONS(2832),
    [anon_sym_alignof] = ACTIONS(2832),
    [anon_sym__Alignof] = ACTIONS(2832),
    [anon_sym_offsetof] = ACTIONS(2832),
    [anon_sym__Generic] = ACTIONS(2832),
    [anon_sym_asm] = ACTIONS(2832),
    [anon_sym___asm__] = ACTIONS(2832),
    [anon_sym___asm] = ACTIONS(2832),
    [sym_number_literal] = ACTIONS(2834),
    [anon_sym_L_SQUOTE] = ACTIONS(2834),
    [anon_sym_u_SQUOTE] = ACTIONS(2834),
    [anon_sym_U_SQUOTE] = ACTIONS(2834),
    [anon_sym_u8_SQUOTE] = ACTIONS(2834),
    [anon_sym_SQUOTE] = ACTIONS(2834),
    [anon_sym_AT] = ACTIONS(2832),
    [anon_sym_DQUOTE] = ACTIONS(2834),
    [anon_sym_L_DQUOTE] = ACTIONS(2834),
    [anon_sym_u_DQUOTE] = ACTIONS(2834),
    [anon_sym_U_DQUOTE] = ACTIONS(2834),
    [anon_sym_u8_DQUOTE] = ACTIONS(2834),
    [sym_true] = ACTIONS(2832),
    [sym_false] = ACTIONS(2832),
    [anon_sym_NULL] = ACTIONS(2832),
    [anon_sym_nullptr] = ACTIONS(2832),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2832),
    [anon_sym___typeof] = ACTIONS(2832),
    [anon_sym_typeof] = ACTIONS(2832),
    [anon_sym_ATimport] = ACTIONS(2834),
    [aux_sym_preproc_undef_token1] = ACTIONS(2832),
    [anon_sym_POUND] = ACTIONS(2832),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2832),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2832),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2832),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2832),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2832),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2832),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2832),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_API_AVAILABLE] = ACTIONS(2832),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_API_DEPRECATED] = ACTIONS(2832),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2832),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym___deprecated_msg] = ACTIONS(2832),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2832),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2832),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2832),
    [anon_sym_ATprotocol] = ACTIONS(2834),
    [anon_sym_ATinterface] = ACTIONS(2834),
    [anon_sym_ATimplementation] = ACTIONS(2834),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2834),
    [anon_sym_ATtry] = ACTIONS(2834),
    [anon_sym___try] = ACTIONS(2832),
    [anon_sym_ATthrow] = ACTIONS(2834),
    [anon_sym_ATselector] = ACTIONS(2834),
    [anon_sym_ATavailable] = ACTIONS(2834),
    [anon_sym___builtin_available] = ACTIONS(2832),
    [anon_sym_va_arg] = ACTIONS(2832),
    [anon_sym_ATencode] = ACTIONS(2834),
    [anon_sym_ATsynchronized] = ACTIONS(2834),
    [anon_sym_BOOL] = ACTIONS(2832),
    [anon_sym_IMP] = ACTIONS(2832),
    [anon_sym_SEL] = ACTIONS(2832),
    [anon_sym_Class] = ACTIONS(2832),
    [anon_sym_id] = ACTIONS(2832),
  },
  [STATE(894)] = {
    [sym_identifier] = ACTIONS(2836),
    [aux_sym_preproc_include_token1] = ACTIONS(2836),
    [aux_sym_preproc_include_token2] = ACTIONS(2836),
    [aux_sym_preproc_def_token1] = ACTIONS(2836),
    [aux_sym_preproc_if_token1] = ACTIONS(2836),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2836),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2836),
    [sym_preproc_directive] = ACTIONS(2836),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_BANG] = ACTIONS(2838),
    [anon_sym_TILDE] = ACTIONS(2838),
    [anon_sym_DASH] = ACTIONS(2836),
    [anon_sym_PLUS] = ACTIONS(2836),
    [anon_sym_STAR] = ACTIONS(2838),
    [anon_sym_CARET] = ACTIONS(2838),
    [anon_sym_AMP] = ACTIONS(2838),
    [anon_sym_SEMI] = ACTIONS(2838),
    [anon_sym___extension__] = ACTIONS(2836),
    [anon_sym_typedef] = ACTIONS(2836),
    [anon_sym_extern] = ACTIONS(2836),
    [anon_sym___attribute__] = ACTIONS(2836),
    [anon_sym___attribute] = ACTIONS(2836),
    [anon_sym_noreturn] = ACTIONS(2836),
    [anon_sym_LBRACK] = ACTIONS(2838),
    [anon_sym___declspec] = ACTIONS(2836),
    [anon_sym___cdecl] = ACTIONS(2836),
    [anon_sym___clrcall] = ACTIONS(2836),
    [anon_sym___stdcall] = ACTIONS(2836),
    [anon_sym___fastcall] = ACTIONS(2836),
    [anon_sym___thiscall] = ACTIONS(2836),
    [anon_sym___vectorcall] = ACTIONS(2836),
    [anon_sym_LBRACE] = ACTIONS(2838),
    [anon_sym_RBRACE] = ACTIONS(2838),
    [anon_sym_signed] = ACTIONS(2836),
    [anon_sym_unsigned] = ACTIONS(2836),
    [anon_sym_long] = ACTIONS(2836),
    [anon_sym_short] = ACTIONS(2836),
    [anon_sym_static] = ACTIONS(2836),
    [anon_sym_ATautoreleasepool] = ACTIONS(2838),
    [anon_sym_auto] = ACTIONS(2836),
    [anon_sym_register] = ACTIONS(2836),
    [anon_sym_inline] = ACTIONS(2836),
    [anon_sym___inline] = ACTIONS(2836),
    [anon_sym___inline__] = ACTIONS(2836),
    [anon_sym___forceinline] = ACTIONS(2836),
    [anon_sym_thread_local] = ACTIONS(2836),
    [anon_sym___thread] = ACTIONS(2836),
    [anon_sym_CG_EXTERN] = ACTIONS(2836),
    [anon_sym_CG_INLINE] = ACTIONS(2836),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2836),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2836),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2836),
    [anon_sym_IBOutlet] = ACTIONS(2836),
    [anon_sym_IBInspectable] = ACTIONS(2836),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2836),
    [anon_sym_NS_INLINE] = ACTIONS(2836),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2836),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2836),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2836),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2836),
    [anon_sym_const] = ACTIONS(2836),
    [anon_sym_constexpr] = ACTIONS(2836),
    [anon_sym_volatile] = ACTIONS(2836),
    [anon_sym_restrict] = ACTIONS(2836),
    [anon_sym___restrict__] = ACTIONS(2836),
    [anon_sym__Atomic] = ACTIONS(2836),
    [anon_sym__Noreturn] = ACTIONS(2836),
    [anon_sym__Nonnull] = ACTIONS(2836),
    [anon_sym_nullable] = ACTIONS(2836),
    [anon_sym__Complex] = ACTIONS(2836),
    [anon_sym__Nullable] = ACTIONS(2836),
    [anon_sym__Nullable_result] = ACTIONS(2836),
    [anon_sym__Null_unspecified] = ACTIONS(2836),
    [anon_sym___autoreleasing] = ACTIONS(2836),
    [anon_sym___block] = ACTIONS(2836),
    [anon_sym___bridge] = ACTIONS(2836),
    [anon_sym___bridge_retained] = ACTIONS(2836),
    [anon_sym___bridge_transfer] = ACTIONS(2836),
    [anon_sym___complex] = ACTIONS(2836),
    [anon_sym___const] = ACTIONS(2836),
    [anon_sym___imag] = ACTIONS(2836),
    [anon_sym___kindof] = ACTIONS(2836),
    [anon_sym___nonnull] = ACTIONS(2836),
    [anon_sym___nullable] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2836),
    [anon_sym___real] = ACTIONS(2836),
    [anon_sym___strong] = ACTIONS(2836),
    [anon_sym___unsafe_unretained] = ACTIONS(2836),
    [anon_sym___unused] = ACTIONS(2836),
    [anon_sym___weak] = ACTIONS(2836),
    [anon_sym_alignas] = ACTIONS(2836),
    [anon_sym__Alignas] = ACTIONS(2836),
    [sym_primitive_type] = ACTIONS(2836),
    [anon_sym_enum] = ACTIONS(2836),
    [anon_sym_struct] = ACTIONS(2836),
    [anon_sym_union] = ACTIONS(2836),
    [anon_sym_if] = ACTIONS(2836),
    [anon_sym_switch] = ACTIONS(2836),
    [anon_sym_case] = ACTIONS(2836),
    [anon_sym_default] = ACTIONS(2836),
    [anon_sym_while] = ACTIONS(2836),
    [anon_sym_do] = ACTIONS(2836),
    [anon_sym_for] = ACTIONS(2836),
    [anon_sym_in] = ACTIONS(2836),
    [anon_sym_return] = ACTIONS(2836),
    [anon_sym_break] = ACTIONS(2836),
    [anon_sym_continue] = ACTIONS(2836),
    [anon_sym_goto] = ACTIONS(2836),
    [anon_sym_DASH_DASH] = ACTIONS(2838),
    [anon_sym_PLUS_PLUS] = ACTIONS(2838),
    [anon_sym_sizeof] = ACTIONS(2836),
    [anon_sym___alignof__] = ACTIONS(2836),
    [anon_sym___alignof] = ACTIONS(2836),
    [anon_sym__alignof] = ACTIONS(2836),
    [anon_sym_alignof] = ACTIONS(2836),
    [anon_sym__Alignof] = ACTIONS(2836),
    [anon_sym_offsetof] = ACTIONS(2836),
    [anon_sym__Generic] = ACTIONS(2836),
    [anon_sym_asm] = ACTIONS(2836),
    [anon_sym___asm__] = ACTIONS(2836),
    [anon_sym___asm] = ACTIONS(2836),
    [sym_number_literal] = ACTIONS(2838),
    [anon_sym_L_SQUOTE] = ACTIONS(2838),
    [anon_sym_u_SQUOTE] = ACTIONS(2838),
    [anon_sym_U_SQUOTE] = ACTIONS(2838),
    [anon_sym_u8_SQUOTE] = ACTIONS(2838),
    [anon_sym_SQUOTE] = ACTIONS(2838),
    [anon_sym_AT] = ACTIONS(2836),
    [anon_sym_DQUOTE] = ACTIONS(2838),
    [anon_sym_L_DQUOTE] = ACTIONS(2838),
    [anon_sym_u_DQUOTE] = ACTIONS(2838),
    [anon_sym_U_DQUOTE] = ACTIONS(2838),
    [anon_sym_u8_DQUOTE] = ACTIONS(2838),
    [sym_true] = ACTIONS(2836),
    [sym_false] = ACTIONS(2836),
    [anon_sym_NULL] = ACTIONS(2836),
    [anon_sym_nullptr] = ACTIONS(2836),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2836),
    [anon_sym___typeof] = ACTIONS(2836),
    [anon_sym_typeof] = ACTIONS(2836),
    [anon_sym_ATimport] = ACTIONS(2838),
    [aux_sym_preproc_undef_token1] = ACTIONS(2836),
    [anon_sym_POUND] = ACTIONS(2836),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2836),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2836),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2836),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2836),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2836),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2836),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2836),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_API_AVAILABLE] = ACTIONS(2836),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_API_DEPRECATED] = ACTIONS(2836),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2836),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym___deprecated_msg] = ACTIONS(2836),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2836),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2836),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2836),
    [anon_sym_ATprotocol] = ACTIONS(2838),
    [anon_sym_ATinterface] = ACTIONS(2838),
    [anon_sym_ATimplementation] = ACTIONS(2838),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2838),
    [anon_sym_ATtry] = ACTIONS(2838),
    [anon_sym___try] = ACTIONS(2836),
    [anon_sym_ATthrow] = ACTIONS(2838),
    [anon_sym_ATselector] = ACTIONS(2838),
    [anon_sym_ATavailable] = ACTIONS(2838),
    [anon_sym___builtin_available] = ACTIONS(2836),
    [anon_sym_va_arg] = ACTIONS(2836),
    [anon_sym_ATencode] = ACTIONS(2838),
    [anon_sym_ATsynchronized] = ACTIONS(2838),
    [anon_sym_BOOL] = ACTIONS(2836),
    [anon_sym_IMP] = ACTIONS(2836),
    [anon_sym_SEL] = ACTIONS(2836),
    [anon_sym_Class] = ACTIONS(2836),
    [anon_sym_id] = ACTIONS(2836),
  },
  [STATE(895)] = {
    [sym_identifier] = ACTIONS(2852),
    [aux_sym_preproc_include_token1] = ACTIONS(2852),
    [aux_sym_preproc_include_token2] = ACTIONS(2852),
    [aux_sym_preproc_def_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2852),
    [sym_preproc_directive] = ACTIONS(2852),
    [anon_sym_LPAREN2] = ACTIONS(2854),
    [anon_sym_BANG] = ACTIONS(2854),
    [anon_sym_TILDE] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2852),
    [anon_sym_PLUS] = ACTIONS(2852),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_SEMI] = ACTIONS(2854),
    [anon_sym___extension__] = ACTIONS(2852),
    [anon_sym_typedef] = ACTIONS(2852),
    [anon_sym_extern] = ACTIONS(2852),
    [anon_sym___attribute__] = ACTIONS(2852),
    [anon_sym___attribute] = ACTIONS(2852),
    [anon_sym_noreturn] = ACTIONS(2852),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym___declspec] = ACTIONS(2852),
    [anon_sym___cdecl] = ACTIONS(2852),
    [anon_sym___clrcall] = ACTIONS(2852),
    [anon_sym___stdcall] = ACTIONS(2852),
    [anon_sym___fastcall] = ACTIONS(2852),
    [anon_sym___thiscall] = ACTIONS(2852),
    [anon_sym___vectorcall] = ACTIONS(2852),
    [anon_sym_LBRACE] = ACTIONS(2854),
    [anon_sym_RBRACE] = ACTIONS(2854),
    [anon_sym_signed] = ACTIONS(2852),
    [anon_sym_unsigned] = ACTIONS(2852),
    [anon_sym_long] = ACTIONS(2852),
    [anon_sym_short] = ACTIONS(2852),
    [anon_sym_static] = ACTIONS(2852),
    [anon_sym_ATautoreleasepool] = ACTIONS(2854),
    [anon_sym_auto] = ACTIONS(2852),
    [anon_sym_register] = ACTIONS(2852),
    [anon_sym_inline] = ACTIONS(2852),
    [anon_sym___inline] = ACTIONS(2852),
    [anon_sym___inline__] = ACTIONS(2852),
    [anon_sym___forceinline] = ACTIONS(2852),
    [anon_sym_thread_local] = ACTIONS(2852),
    [anon_sym___thread] = ACTIONS(2852),
    [anon_sym_CG_EXTERN] = ACTIONS(2852),
    [anon_sym_CG_INLINE] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2852),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2852),
    [anon_sym_IBOutlet] = ACTIONS(2852),
    [anon_sym_IBInspectable] = ACTIONS(2852),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2852),
    [anon_sym_NS_INLINE] = ACTIONS(2852),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2852),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2852),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2852),
    [anon_sym_const] = ACTIONS(2852),
    [anon_sym_constexpr] = ACTIONS(2852),
    [anon_sym_volatile] = ACTIONS(2852),
    [anon_sym_restrict] = ACTIONS(2852),
    [anon_sym___restrict__] = ACTIONS(2852),
    [anon_sym__Atomic] = ACTIONS(2852),
    [anon_sym__Noreturn] = ACTIONS(2852),
    [anon_sym__Nonnull] = ACTIONS(2852),
    [anon_sym_nullable] = ACTIONS(2852),
    [anon_sym__Complex] = ACTIONS(2852),
    [anon_sym__Nullable] = ACTIONS(2852),
    [anon_sym__Nullable_result] = ACTIONS(2852),
    [anon_sym__Null_unspecified] = ACTIONS(2852),
    [anon_sym___autoreleasing] = ACTIONS(2852),
    [anon_sym___block] = ACTIONS(2852),
    [anon_sym___bridge] = ACTIONS(2852),
    [anon_sym___bridge_retained] = ACTIONS(2852),
    [anon_sym___bridge_transfer] = ACTIONS(2852),
    [anon_sym___complex] = ACTIONS(2852),
    [anon_sym___const] = ACTIONS(2852),
    [anon_sym___imag] = ACTIONS(2852),
    [anon_sym___kindof] = ACTIONS(2852),
    [anon_sym___nonnull] = ACTIONS(2852),
    [anon_sym___nullable] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2852),
    [anon_sym___real] = ACTIONS(2852),
    [anon_sym___strong] = ACTIONS(2852),
    [anon_sym___unsafe_unretained] = ACTIONS(2852),
    [anon_sym___unused] = ACTIONS(2852),
    [anon_sym___weak] = ACTIONS(2852),
    [anon_sym_alignas] = ACTIONS(2852),
    [anon_sym__Alignas] = ACTIONS(2852),
    [sym_primitive_type] = ACTIONS(2852),
    [anon_sym_enum] = ACTIONS(2852),
    [anon_sym_struct] = ACTIONS(2852),
    [anon_sym_union] = ACTIONS(2852),
    [anon_sym_if] = ACTIONS(2852),
    [anon_sym_switch] = ACTIONS(2852),
    [anon_sym_case] = ACTIONS(2852),
    [anon_sym_default] = ACTIONS(2852),
    [anon_sym_while] = ACTIONS(2852),
    [anon_sym_do] = ACTIONS(2852),
    [anon_sym_for] = ACTIONS(2852),
    [anon_sym_in] = ACTIONS(2852),
    [anon_sym_return] = ACTIONS(2852),
    [anon_sym_break] = ACTIONS(2852),
    [anon_sym_continue] = ACTIONS(2852),
    [anon_sym_goto] = ACTIONS(2852),
    [anon_sym_DASH_DASH] = ACTIONS(2854),
    [anon_sym_PLUS_PLUS] = ACTIONS(2854),
    [anon_sym_sizeof] = ACTIONS(2852),
    [anon_sym___alignof__] = ACTIONS(2852),
    [anon_sym___alignof] = ACTIONS(2852),
    [anon_sym__alignof] = ACTIONS(2852),
    [anon_sym_alignof] = ACTIONS(2852),
    [anon_sym__Alignof] = ACTIONS(2852),
    [anon_sym_offsetof] = ACTIONS(2852),
    [anon_sym__Generic] = ACTIONS(2852),
    [anon_sym_asm] = ACTIONS(2852),
    [anon_sym___asm__] = ACTIONS(2852),
    [anon_sym___asm] = ACTIONS(2852),
    [sym_number_literal] = ACTIONS(2854),
    [anon_sym_L_SQUOTE] = ACTIONS(2854),
    [anon_sym_u_SQUOTE] = ACTIONS(2854),
    [anon_sym_U_SQUOTE] = ACTIONS(2854),
    [anon_sym_u8_SQUOTE] = ACTIONS(2854),
    [anon_sym_SQUOTE] = ACTIONS(2854),
    [anon_sym_AT] = ACTIONS(2852),
    [anon_sym_DQUOTE] = ACTIONS(2854),
    [anon_sym_L_DQUOTE] = ACTIONS(2854),
    [anon_sym_u_DQUOTE] = ACTIONS(2854),
    [anon_sym_U_DQUOTE] = ACTIONS(2854),
    [anon_sym_u8_DQUOTE] = ACTIONS(2854),
    [sym_true] = ACTIONS(2852),
    [sym_false] = ACTIONS(2852),
    [anon_sym_NULL] = ACTIONS(2852),
    [anon_sym_nullptr] = ACTIONS(2852),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2852),
    [anon_sym___typeof] = ACTIONS(2852),
    [anon_sym_typeof] = ACTIONS(2852),
    [anon_sym_ATimport] = ACTIONS(2854),
    [aux_sym_preproc_undef_token1] = ACTIONS(2852),
    [anon_sym_POUND] = ACTIONS(2852),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2852),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2852),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2852),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_API_AVAILABLE] = ACTIONS(2852),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_API_DEPRECATED] = ACTIONS(2852),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym___deprecated_msg] = ACTIONS(2852),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2852),
    [anon_sym_ATprotocol] = ACTIONS(2854),
    [anon_sym_ATinterface] = ACTIONS(2854),
    [anon_sym_ATimplementation] = ACTIONS(2854),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2854),
    [anon_sym_ATtry] = ACTIONS(2854),
    [anon_sym___try] = ACTIONS(2852),
    [anon_sym_ATthrow] = ACTIONS(2854),
    [anon_sym_ATselector] = ACTIONS(2854),
    [anon_sym_ATavailable] = ACTIONS(2854),
    [anon_sym___builtin_available] = ACTIONS(2852),
    [anon_sym_va_arg] = ACTIONS(2852),
    [anon_sym_ATencode] = ACTIONS(2854),
    [anon_sym_ATsynchronized] = ACTIONS(2854),
    [anon_sym_BOOL] = ACTIONS(2852),
    [anon_sym_IMP] = ACTIONS(2852),
    [anon_sym_SEL] = ACTIONS(2852),
    [anon_sym_Class] = ACTIONS(2852),
    [anon_sym_id] = ACTIONS(2852),
  },
  [STATE(896)] = {
    [sym_identifier] = ACTIONS(2654),
    [aux_sym_preproc_include_token1] = ACTIONS(2654),
    [aux_sym_preproc_include_token2] = ACTIONS(2654),
    [aux_sym_preproc_def_token1] = ACTIONS(2654),
    [aux_sym_preproc_if_token1] = ACTIONS(2654),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2654),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2654),
    [sym_preproc_directive] = ACTIONS(2654),
    [anon_sym_LPAREN2] = ACTIONS(2656),
    [anon_sym_BANG] = ACTIONS(2656),
    [anon_sym_TILDE] = ACTIONS(2656),
    [anon_sym_DASH] = ACTIONS(2654),
    [anon_sym_PLUS] = ACTIONS(2654),
    [anon_sym_STAR] = ACTIONS(2656),
    [anon_sym_CARET] = ACTIONS(2656),
    [anon_sym_AMP] = ACTIONS(2656),
    [anon_sym_SEMI] = ACTIONS(2656),
    [anon_sym___extension__] = ACTIONS(2654),
    [anon_sym_typedef] = ACTIONS(2654),
    [anon_sym_extern] = ACTIONS(2654),
    [anon_sym___attribute__] = ACTIONS(2654),
    [anon_sym___attribute] = ACTIONS(2654),
    [anon_sym_noreturn] = ACTIONS(2654),
    [anon_sym_LBRACK] = ACTIONS(2656),
    [anon_sym___declspec] = ACTIONS(2654),
    [anon_sym___cdecl] = ACTIONS(2654),
    [anon_sym___clrcall] = ACTIONS(2654),
    [anon_sym___stdcall] = ACTIONS(2654),
    [anon_sym___fastcall] = ACTIONS(2654),
    [anon_sym___thiscall] = ACTIONS(2654),
    [anon_sym___vectorcall] = ACTIONS(2654),
    [anon_sym_LBRACE] = ACTIONS(2656),
    [anon_sym_RBRACE] = ACTIONS(2656),
    [anon_sym_signed] = ACTIONS(2654),
    [anon_sym_unsigned] = ACTIONS(2654),
    [anon_sym_long] = ACTIONS(2654),
    [anon_sym_short] = ACTIONS(2654),
    [anon_sym_static] = ACTIONS(2654),
    [anon_sym_ATautoreleasepool] = ACTIONS(2656),
    [anon_sym_auto] = ACTIONS(2654),
    [anon_sym_register] = ACTIONS(2654),
    [anon_sym_inline] = ACTIONS(2654),
    [anon_sym___inline] = ACTIONS(2654),
    [anon_sym___inline__] = ACTIONS(2654),
    [anon_sym___forceinline] = ACTIONS(2654),
    [anon_sym_thread_local] = ACTIONS(2654),
    [anon_sym___thread] = ACTIONS(2654),
    [anon_sym_CG_EXTERN] = ACTIONS(2654),
    [anon_sym_CG_INLINE] = ACTIONS(2654),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2654),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2654),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2654),
    [anon_sym_IBOutlet] = ACTIONS(2654),
    [anon_sym_IBInspectable] = ACTIONS(2654),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2654),
    [anon_sym_NS_INLINE] = ACTIONS(2654),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2654),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2654),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2654),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2654),
    [anon_sym_const] = ACTIONS(2654),
    [anon_sym_constexpr] = ACTIONS(2654),
    [anon_sym_volatile] = ACTIONS(2654),
    [anon_sym_restrict] = ACTIONS(2654),
    [anon_sym___restrict__] = ACTIONS(2654),
    [anon_sym__Atomic] = ACTIONS(2654),
    [anon_sym__Noreturn] = ACTIONS(2654),
    [anon_sym__Nonnull] = ACTIONS(2654),
    [anon_sym_nullable] = ACTIONS(2654),
    [anon_sym__Complex] = ACTIONS(2654),
    [anon_sym__Nullable] = ACTIONS(2654),
    [anon_sym__Nullable_result] = ACTIONS(2654),
    [anon_sym__Null_unspecified] = ACTIONS(2654),
    [anon_sym___autoreleasing] = ACTIONS(2654),
    [anon_sym___block] = ACTIONS(2654),
    [anon_sym___bridge] = ACTIONS(2654),
    [anon_sym___bridge_retained] = ACTIONS(2654),
    [anon_sym___bridge_transfer] = ACTIONS(2654),
    [anon_sym___complex] = ACTIONS(2654),
    [anon_sym___const] = ACTIONS(2654),
    [anon_sym___imag] = ACTIONS(2654),
    [anon_sym___kindof] = ACTIONS(2654),
    [anon_sym___nonnull] = ACTIONS(2654),
    [anon_sym___nullable] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2654),
    [anon_sym___real] = ACTIONS(2654),
    [anon_sym___strong] = ACTIONS(2654),
    [anon_sym___unsafe_unretained] = ACTIONS(2654),
    [anon_sym___unused] = ACTIONS(2654),
    [anon_sym___weak] = ACTIONS(2654),
    [anon_sym_alignas] = ACTIONS(2654),
    [anon_sym__Alignas] = ACTIONS(2654),
    [sym_primitive_type] = ACTIONS(2654),
    [anon_sym_enum] = ACTIONS(2654),
    [anon_sym_struct] = ACTIONS(2654),
    [anon_sym_union] = ACTIONS(2654),
    [anon_sym_if] = ACTIONS(2654),
    [anon_sym_switch] = ACTIONS(2654),
    [anon_sym_case] = ACTIONS(2654),
    [anon_sym_default] = ACTIONS(2654),
    [anon_sym_while] = ACTIONS(2654),
    [anon_sym_do] = ACTIONS(2654),
    [anon_sym_for] = ACTIONS(2654),
    [anon_sym_in] = ACTIONS(2654),
    [anon_sym_return] = ACTIONS(2654),
    [anon_sym_break] = ACTIONS(2654),
    [anon_sym_continue] = ACTIONS(2654),
    [anon_sym_goto] = ACTIONS(2654),
    [anon_sym_DASH_DASH] = ACTIONS(2656),
    [anon_sym_PLUS_PLUS] = ACTIONS(2656),
    [anon_sym_sizeof] = ACTIONS(2654),
    [anon_sym___alignof__] = ACTIONS(2654),
    [anon_sym___alignof] = ACTIONS(2654),
    [anon_sym__alignof] = ACTIONS(2654),
    [anon_sym_alignof] = ACTIONS(2654),
    [anon_sym__Alignof] = ACTIONS(2654),
    [anon_sym_offsetof] = ACTIONS(2654),
    [anon_sym__Generic] = ACTIONS(2654),
    [anon_sym_asm] = ACTIONS(2654),
    [anon_sym___asm__] = ACTIONS(2654),
    [anon_sym___asm] = ACTIONS(2654),
    [sym_number_literal] = ACTIONS(2656),
    [anon_sym_L_SQUOTE] = ACTIONS(2656),
    [anon_sym_u_SQUOTE] = ACTIONS(2656),
    [anon_sym_U_SQUOTE] = ACTIONS(2656),
    [anon_sym_u8_SQUOTE] = ACTIONS(2656),
    [anon_sym_SQUOTE] = ACTIONS(2656),
    [anon_sym_AT] = ACTIONS(2654),
    [anon_sym_DQUOTE] = ACTIONS(2656),
    [anon_sym_L_DQUOTE] = ACTIONS(2656),
    [anon_sym_u_DQUOTE] = ACTIONS(2656),
    [anon_sym_U_DQUOTE] = ACTIONS(2656),
    [anon_sym_u8_DQUOTE] = ACTIONS(2656),
    [sym_true] = ACTIONS(2654),
    [sym_false] = ACTIONS(2654),
    [anon_sym_NULL] = ACTIONS(2654),
    [anon_sym_nullptr] = ACTIONS(2654),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2654),
    [anon_sym___typeof] = ACTIONS(2654),
    [anon_sym_typeof] = ACTIONS(2654),
    [anon_sym_ATimport] = ACTIONS(2656),
    [aux_sym_preproc_undef_token1] = ACTIONS(2654),
    [anon_sym_POUND] = ACTIONS(2654),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2654),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2654),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2654),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2654),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2654),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2654),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2654),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_API_AVAILABLE] = ACTIONS(2654),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_API_DEPRECATED] = ACTIONS(2654),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2654),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym___deprecated_msg] = ACTIONS(2654),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2654),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2654),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2654),
    [anon_sym_ATprotocol] = ACTIONS(2656),
    [anon_sym_ATinterface] = ACTIONS(2656),
    [anon_sym_ATimplementation] = ACTIONS(2656),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2656),
    [anon_sym_ATtry] = ACTIONS(2656),
    [anon_sym___try] = ACTIONS(2654),
    [anon_sym_ATthrow] = ACTIONS(2656),
    [anon_sym_ATselector] = ACTIONS(2656),
    [anon_sym_ATavailable] = ACTIONS(2656),
    [anon_sym___builtin_available] = ACTIONS(2654),
    [anon_sym_va_arg] = ACTIONS(2654),
    [anon_sym_ATencode] = ACTIONS(2656),
    [anon_sym_ATsynchronized] = ACTIONS(2656),
    [anon_sym_BOOL] = ACTIONS(2654),
    [anon_sym_IMP] = ACTIONS(2654),
    [anon_sym_SEL] = ACTIONS(2654),
    [anon_sym_Class] = ACTIONS(2654),
    [anon_sym_id] = ACTIONS(2654),
  },
  [STATE(897)] = {
    [sym_identifier] = ACTIONS(2840),
    [aux_sym_preproc_include_token1] = ACTIONS(2840),
    [aux_sym_preproc_include_token2] = ACTIONS(2840),
    [aux_sym_preproc_def_token1] = ACTIONS(2840),
    [aux_sym_preproc_if_token1] = ACTIONS(2840),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2840),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2840),
    [sym_preproc_directive] = ACTIONS(2840),
    [anon_sym_LPAREN2] = ACTIONS(2842),
    [anon_sym_BANG] = ACTIONS(2842),
    [anon_sym_TILDE] = ACTIONS(2842),
    [anon_sym_DASH] = ACTIONS(2840),
    [anon_sym_PLUS] = ACTIONS(2840),
    [anon_sym_STAR] = ACTIONS(2842),
    [anon_sym_CARET] = ACTIONS(2842),
    [anon_sym_AMP] = ACTIONS(2842),
    [anon_sym_SEMI] = ACTIONS(2842),
    [anon_sym___extension__] = ACTIONS(2840),
    [anon_sym_typedef] = ACTIONS(2840),
    [anon_sym_extern] = ACTIONS(2840),
    [anon_sym___attribute__] = ACTIONS(2840),
    [anon_sym___attribute] = ACTIONS(2840),
    [anon_sym_noreturn] = ACTIONS(2840),
    [anon_sym_LBRACK] = ACTIONS(2842),
    [anon_sym___declspec] = ACTIONS(2840),
    [anon_sym___cdecl] = ACTIONS(2840),
    [anon_sym___clrcall] = ACTIONS(2840),
    [anon_sym___stdcall] = ACTIONS(2840),
    [anon_sym___fastcall] = ACTIONS(2840),
    [anon_sym___thiscall] = ACTIONS(2840),
    [anon_sym___vectorcall] = ACTIONS(2840),
    [anon_sym_LBRACE] = ACTIONS(2842),
    [anon_sym_RBRACE] = ACTIONS(2842),
    [anon_sym_signed] = ACTIONS(2840),
    [anon_sym_unsigned] = ACTIONS(2840),
    [anon_sym_long] = ACTIONS(2840),
    [anon_sym_short] = ACTIONS(2840),
    [anon_sym_static] = ACTIONS(2840),
    [anon_sym_ATautoreleasepool] = ACTIONS(2842),
    [anon_sym_auto] = ACTIONS(2840),
    [anon_sym_register] = ACTIONS(2840),
    [anon_sym_inline] = ACTIONS(2840),
    [anon_sym___inline] = ACTIONS(2840),
    [anon_sym___inline__] = ACTIONS(2840),
    [anon_sym___forceinline] = ACTIONS(2840),
    [anon_sym_thread_local] = ACTIONS(2840),
    [anon_sym___thread] = ACTIONS(2840),
    [anon_sym_CG_EXTERN] = ACTIONS(2840),
    [anon_sym_CG_INLINE] = ACTIONS(2840),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2840),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2840),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2840),
    [anon_sym_IBOutlet] = ACTIONS(2840),
    [anon_sym_IBInspectable] = ACTIONS(2840),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2840),
    [anon_sym_NS_INLINE] = ACTIONS(2840),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2840),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2840),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2840),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2840),
    [anon_sym_const] = ACTIONS(2840),
    [anon_sym_constexpr] = ACTIONS(2840),
    [anon_sym_volatile] = ACTIONS(2840),
    [anon_sym_restrict] = ACTIONS(2840),
    [anon_sym___restrict__] = ACTIONS(2840),
    [anon_sym__Atomic] = ACTIONS(2840),
    [anon_sym__Noreturn] = ACTIONS(2840),
    [anon_sym__Nonnull] = ACTIONS(2840),
    [anon_sym_nullable] = ACTIONS(2840),
    [anon_sym__Complex] = ACTIONS(2840),
    [anon_sym__Nullable] = ACTIONS(2840),
    [anon_sym__Nullable_result] = ACTIONS(2840),
    [anon_sym__Null_unspecified] = ACTIONS(2840),
    [anon_sym___autoreleasing] = ACTIONS(2840),
    [anon_sym___block] = ACTIONS(2840),
    [anon_sym___bridge] = ACTIONS(2840),
    [anon_sym___bridge_retained] = ACTIONS(2840),
    [anon_sym___bridge_transfer] = ACTIONS(2840),
    [anon_sym___complex] = ACTIONS(2840),
    [anon_sym___const] = ACTIONS(2840),
    [anon_sym___imag] = ACTIONS(2840),
    [anon_sym___kindof] = ACTIONS(2840),
    [anon_sym___nonnull] = ACTIONS(2840),
    [anon_sym___nullable] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2840),
    [anon_sym___real] = ACTIONS(2840),
    [anon_sym___strong] = ACTIONS(2840),
    [anon_sym___unsafe_unretained] = ACTIONS(2840),
    [anon_sym___unused] = ACTIONS(2840),
    [anon_sym___weak] = ACTIONS(2840),
    [anon_sym_alignas] = ACTIONS(2840),
    [anon_sym__Alignas] = ACTIONS(2840),
    [sym_primitive_type] = ACTIONS(2840),
    [anon_sym_enum] = ACTIONS(2840),
    [anon_sym_struct] = ACTIONS(2840),
    [anon_sym_union] = ACTIONS(2840),
    [anon_sym_if] = ACTIONS(2840),
    [anon_sym_switch] = ACTIONS(2840),
    [anon_sym_case] = ACTIONS(2840),
    [anon_sym_default] = ACTIONS(2840),
    [anon_sym_while] = ACTIONS(2840),
    [anon_sym_do] = ACTIONS(2840),
    [anon_sym_for] = ACTIONS(2840),
    [anon_sym_in] = ACTIONS(2840),
    [anon_sym_return] = ACTIONS(2840),
    [anon_sym_break] = ACTIONS(2840),
    [anon_sym_continue] = ACTIONS(2840),
    [anon_sym_goto] = ACTIONS(2840),
    [anon_sym_DASH_DASH] = ACTIONS(2842),
    [anon_sym_PLUS_PLUS] = ACTIONS(2842),
    [anon_sym_sizeof] = ACTIONS(2840),
    [anon_sym___alignof__] = ACTIONS(2840),
    [anon_sym___alignof] = ACTIONS(2840),
    [anon_sym__alignof] = ACTIONS(2840),
    [anon_sym_alignof] = ACTIONS(2840),
    [anon_sym__Alignof] = ACTIONS(2840),
    [anon_sym_offsetof] = ACTIONS(2840),
    [anon_sym__Generic] = ACTIONS(2840),
    [anon_sym_asm] = ACTIONS(2840),
    [anon_sym___asm__] = ACTIONS(2840),
    [anon_sym___asm] = ACTIONS(2840),
    [sym_number_literal] = ACTIONS(2842),
    [anon_sym_L_SQUOTE] = ACTIONS(2842),
    [anon_sym_u_SQUOTE] = ACTIONS(2842),
    [anon_sym_U_SQUOTE] = ACTIONS(2842),
    [anon_sym_u8_SQUOTE] = ACTIONS(2842),
    [anon_sym_SQUOTE] = ACTIONS(2842),
    [anon_sym_AT] = ACTIONS(2840),
    [anon_sym_DQUOTE] = ACTIONS(2842),
    [anon_sym_L_DQUOTE] = ACTIONS(2842),
    [anon_sym_u_DQUOTE] = ACTIONS(2842),
    [anon_sym_U_DQUOTE] = ACTIONS(2842),
    [anon_sym_u8_DQUOTE] = ACTIONS(2842),
    [sym_true] = ACTIONS(2840),
    [sym_false] = ACTIONS(2840),
    [anon_sym_NULL] = ACTIONS(2840),
    [anon_sym_nullptr] = ACTIONS(2840),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2840),
    [anon_sym___typeof] = ACTIONS(2840),
    [anon_sym_typeof] = ACTIONS(2840),
    [anon_sym_ATimport] = ACTIONS(2842),
    [aux_sym_preproc_undef_token1] = ACTIONS(2840),
    [anon_sym_POUND] = ACTIONS(2840),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2840),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2840),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2840),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2840),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2840),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2840),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2840),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_API_AVAILABLE] = ACTIONS(2840),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_API_DEPRECATED] = ACTIONS(2840),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2840),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym___deprecated_msg] = ACTIONS(2840),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2840),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2840),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2840),
    [anon_sym_ATprotocol] = ACTIONS(2842),
    [anon_sym_ATinterface] = ACTIONS(2842),
    [anon_sym_ATimplementation] = ACTIONS(2842),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2842),
    [anon_sym_ATtry] = ACTIONS(2842),
    [anon_sym___try] = ACTIONS(2840),
    [anon_sym_ATthrow] = ACTIONS(2842),
    [anon_sym_ATselector] = ACTIONS(2842),
    [anon_sym_ATavailable] = ACTIONS(2842),
    [anon_sym___builtin_available] = ACTIONS(2840),
    [anon_sym_va_arg] = ACTIONS(2840),
    [anon_sym_ATencode] = ACTIONS(2842),
    [anon_sym_ATsynchronized] = ACTIONS(2842),
    [anon_sym_BOOL] = ACTIONS(2840),
    [anon_sym_IMP] = ACTIONS(2840),
    [anon_sym_SEL] = ACTIONS(2840),
    [anon_sym_Class] = ACTIONS(2840),
    [anon_sym_id] = ACTIONS(2840),
  },
  [STATE(898)] = {
    [sym_identifier] = ACTIONS(2676),
    [aux_sym_preproc_include_token1] = ACTIONS(2676),
    [aux_sym_preproc_include_token2] = ACTIONS(2676),
    [aux_sym_preproc_def_token1] = ACTIONS(2676),
    [aux_sym_preproc_if_token1] = ACTIONS(2676),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2676),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2676),
    [sym_preproc_directive] = ACTIONS(2676),
    [anon_sym_LPAREN2] = ACTIONS(2678),
    [anon_sym_BANG] = ACTIONS(2678),
    [anon_sym_TILDE] = ACTIONS(2678),
    [anon_sym_DASH] = ACTIONS(2676),
    [anon_sym_PLUS] = ACTIONS(2676),
    [anon_sym_STAR] = ACTIONS(2678),
    [anon_sym_CARET] = ACTIONS(2678),
    [anon_sym_AMP] = ACTIONS(2678),
    [anon_sym_SEMI] = ACTIONS(2678),
    [anon_sym___extension__] = ACTIONS(2676),
    [anon_sym_typedef] = ACTIONS(2676),
    [anon_sym_extern] = ACTIONS(2676),
    [anon_sym___attribute__] = ACTIONS(2676),
    [anon_sym___attribute] = ACTIONS(2676),
    [anon_sym_noreturn] = ACTIONS(2676),
    [anon_sym_LBRACK] = ACTIONS(2678),
    [anon_sym___declspec] = ACTIONS(2676),
    [anon_sym___cdecl] = ACTIONS(2676),
    [anon_sym___clrcall] = ACTIONS(2676),
    [anon_sym___stdcall] = ACTIONS(2676),
    [anon_sym___fastcall] = ACTIONS(2676),
    [anon_sym___thiscall] = ACTIONS(2676),
    [anon_sym___vectorcall] = ACTIONS(2676),
    [anon_sym_LBRACE] = ACTIONS(2678),
    [anon_sym_RBRACE] = ACTIONS(2678),
    [anon_sym_signed] = ACTIONS(2676),
    [anon_sym_unsigned] = ACTIONS(2676),
    [anon_sym_long] = ACTIONS(2676),
    [anon_sym_short] = ACTIONS(2676),
    [anon_sym_static] = ACTIONS(2676),
    [anon_sym_ATautoreleasepool] = ACTIONS(2678),
    [anon_sym_auto] = ACTIONS(2676),
    [anon_sym_register] = ACTIONS(2676),
    [anon_sym_inline] = ACTIONS(2676),
    [anon_sym___inline] = ACTIONS(2676),
    [anon_sym___inline__] = ACTIONS(2676),
    [anon_sym___forceinline] = ACTIONS(2676),
    [anon_sym_thread_local] = ACTIONS(2676),
    [anon_sym___thread] = ACTIONS(2676),
    [anon_sym_CG_EXTERN] = ACTIONS(2676),
    [anon_sym_CG_INLINE] = ACTIONS(2676),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2676),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2676),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2676),
    [anon_sym_IBOutlet] = ACTIONS(2676),
    [anon_sym_IBInspectable] = ACTIONS(2676),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2676),
    [anon_sym_NS_INLINE] = ACTIONS(2676),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2676),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2676),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2676),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2676),
    [anon_sym_const] = ACTIONS(2676),
    [anon_sym_constexpr] = ACTIONS(2676),
    [anon_sym_volatile] = ACTIONS(2676),
    [anon_sym_restrict] = ACTIONS(2676),
    [anon_sym___restrict__] = ACTIONS(2676),
    [anon_sym__Atomic] = ACTIONS(2676),
    [anon_sym__Noreturn] = ACTIONS(2676),
    [anon_sym__Nonnull] = ACTIONS(2676),
    [anon_sym_nullable] = ACTIONS(2676),
    [anon_sym__Complex] = ACTIONS(2676),
    [anon_sym__Nullable] = ACTIONS(2676),
    [anon_sym__Nullable_result] = ACTIONS(2676),
    [anon_sym__Null_unspecified] = ACTIONS(2676),
    [anon_sym___autoreleasing] = ACTIONS(2676),
    [anon_sym___block] = ACTIONS(2676),
    [anon_sym___bridge] = ACTIONS(2676),
    [anon_sym___bridge_retained] = ACTIONS(2676),
    [anon_sym___bridge_transfer] = ACTIONS(2676),
    [anon_sym___complex] = ACTIONS(2676),
    [anon_sym___const] = ACTIONS(2676),
    [anon_sym___imag] = ACTIONS(2676),
    [anon_sym___kindof] = ACTIONS(2676),
    [anon_sym___nonnull] = ACTIONS(2676),
    [anon_sym___nullable] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2676),
    [anon_sym___real] = ACTIONS(2676),
    [anon_sym___strong] = ACTIONS(2676),
    [anon_sym___unsafe_unretained] = ACTIONS(2676),
    [anon_sym___unused] = ACTIONS(2676),
    [anon_sym___weak] = ACTIONS(2676),
    [anon_sym_alignas] = ACTIONS(2676),
    [anon_sym__Alignas] = ACTIONS(2676),
    [sym_primitive_type] = ACTIONS(2676),
    [anon_sym_enum] = ACTIONS(2676),
    [anon_sym_struct] = ACTIONS(2676),
    [anon_sym_union] = ACTIONS(2676),
    [anon_sym_if] = ACTIONS(2676),
    [anon_sym_switch] = ACTIONS(2676),
    [anon_sym_case] = ACTIONS(2676),
    [anon_sym_default] = ACTIONS(2676),
    [anon_sym_while] = ACTIONS(2676),
    [anon_sym_do] = ACTIONS(2676),
    [anon_sym_for] = ACTIONS(2676),
    [anon_sym_in] = ACTIONS(2676),
    [anon_sym_return] = ACTIONS(2676),
    [anon_sym_break] = ACTIONS(2676),
    [anon_sym_continue] = ACTIONS(2676),
    [anon_sym_goto] = ACTIONS(2676),
    [anon_sym_DASH_DASH] = ACTIONS(2678),
    [anon_sym_PLUS_PLUS] = ACTIONS(2678),
    [anon_sym_sizeof] = ACTIONS(2676),
    [anon_sym___alignof__] = ACTIONS(2676),
    [anon_sym___alignof] = ACTIONS(2676),
    [anon_sym__alignof] = ACTIONS(2676),
    [anon_sym_alignof] = ACTIONS(2676),
    [anon_sym__Alignof] = ACTIONS(2676),
    [anon_sym_offsetof] = ACTIONS(2676),
    [anon_sym__Generic] = ACTIONS(2676),
    [anon_sym_asm] = ACTIONS(2676),
    [anon_sym___asm__] = ACTIONS(2676),
    [anon_sym___asm] = ACTIONS(2676),
    [sym_number_literal] = ACTIONS(2678),
    [anon_sym_L_SQUOTE] = ACTIONS(2678),
    [anon_sym_u_SQUOTE] = ACTIONS(2678),
    [anon_sym_U_SQUOTE] = ACTIONS(2678),
    [anon_sym_u8_SQUOTE] = ACTIONS(2678),
    [anon_sym_SQUOTE] = ACTIONS(2678),
    [anon_sym_AT] = ACTIONS(2676),
    [anon_sym_DQUOTE] = ACTIONS(2678),
    [anon_sym_L_DQUOTE] = ACTIONS(2678),
    [anon_sym_u_DQUOTE] = ACTIONS(2678),
    [anon_sym_U_DQUOTE] = ACTIONS(2678),
    [anon_sym_u8_DQUOTE] = ACTIONS(2678),
    [sym_true] = ACTIONS(2676),
    [sym_false] = ACTIONS(2676),
    [anon_sym_NULL] = ACTIONS(2676),
    [anon_sym_nullptr] = ACTIONS(2676),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2676),
    [anon_sym___typeof] = ACTIONS(2676),
    [anon_sym_typeof] = ACTIONS(2676),
    [anon_sym_ATimport] = ACTIONS(2678),
    [aux_sym_preproc_undef_token1] = ACTIONS(2676),
    [anon_sym_POUND] = ACTIONS(2676),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2676),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2676),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2676),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2676),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2676),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2676),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2676),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_API_AVAILABLE] = ACTIONS(2676),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_API_DEPRECATED] = ACTIONS(2676),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2676),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym___deprecated_msg] = ACTIONS(2676),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2676),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2676),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2676),
    [anon_sym_ATprotocol] = ACTIONS(2678),
    [anon_sym_ATinterface] = ACTIONS(2678),
    [anon_sym_ATimplementation] = ACTIONS(2678),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2678),
    [anon_sym_ATtry] = ACTIONS(2678),
    [anon_sym___try] = ACTIONS(2676),
    [anon_sym_ATthrow] = ACTIONS(2678),
    [anon_sym_ATselector] = ACTIONS(2678),
    [anon_sym_ATavailable] = ACTIONS(2678),
    [anon_sym___builtin_available] = ACTIONS(2676),
    [anon_sym_va_arg] = ACTIONS(2676),
    [anon_sym_ATencode] = ACTIONS(2678),
    [anon_sym_ATsynchronized] = ACTIONS(2678),
    [anon_sym_BOOL] = ACTIONS(2676),
    [anon_sym_IMP] = ACTIONS(2676),
    [anon_sym_SEL] = ACTIONS(2676),
    [anon_sym_Class] = ACTIONS(2676),
    [anon_sym_id] = ACTIONS(2676),
  },
  [STATE(899)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_RBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(900)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_RBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(901)] = {
    [sym_identifier] = ACTIONS(2856),
    [aux_sym_preproc_include_token1] = ACTIONS(2856),
    [aux_sym_preproc_include_token2] = ACTIONS(2856),
    [aux_sym_preproc_def_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2856),
    [sym_preproc_directive] = ACTIONS(2856),
    [anon_sym_LPAREN2] = ACTIONS(2858),
    [anon_sym_BANG] = ACTIONS(2858),
    [anon_sym_TILDE] = ACTIONS(2858),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_STAR] = ACTIONS(2858),
    [anon_sym_CARET] = ACTIONS(2858),
    [anon_sym_AMP] = ACTIONS(2858),
    [anon_sym_SEMI] = ACTIONS(2858),
    [anon_sym___extension__] = ACTIONS(2856),
    [anon_sym_typedef] = ACTIONS(2856),
    [anon_sym_extern] = ACTIONS(2856),
    [anon_sym___attribute__] = ACTIONS(2856),
    [anon_sym___attribute] = ACTIONS(2856),
    [anon_sym_noreturn] = ACTIONS(2856),
    [anon_sym_LBRACK] = ACTIONS(2858),
    [anon_sym___declspec] = ACTIONS(2856),
    [anon_sym___cdecl] = ACTIONS(2856),
    [anon_sym___clrcall] = ACTIONS(2856),
    [anon_sym___stdcall] = ACTIONS(2856),
    [anon_sym___fastcall] = ACTIONS(2856),
    [anon_sym___thiscall] = ACTIONS(2856),
    [anon_sym___vectorcall] = ACTIONS(2856),
    [anon_sym_LBRACE] = ACTIONS(2858),
    [anon_sym_RBRACE] = ACTIONS(2858),
    [anon_sym_signed] = ACTIONS(2856),
    [anon_sym_unsigned] = ACTIONS(2856),
    [anon_sym_long] = ACTIONS(2856),
    [anon_sym_short] = ACTIONS(2856),
    [anon_sym_static] = ACTIONS(2856),
    [anon_sym_ATautoreleasepool] = ACTIONS(2858),
    [anon_sym_auto] = ACTIONS(2856),
    [anon_sym_register] = ACTIONS(2856),
    [anon_sym_inline] = ACTIONS(2856),
    [anon_sym___inline] = ACTIONS(2856),
    [anon_sym___inline__] = ACTIONS(2856),
    [anon_sym___forceinline] = ACTIONS(2856),
    [anon_sym_thread_local] = ACTIONS(2856),
    [anon_sym___thread] = ACTIONS(2856),
    [anon_sym_CG_EXTERN] = ACTIONS(2856),
    [anon_sym_CG_INLINE] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2856),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2856),
    [anon_sym_IBOutlet] = ACTIONS(2856),
    [anon_sym_IBInspectable] = ACTIONS(2856),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2856),
    [anon_sym_NS_INLINE] = ACTIONS(2856),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2856),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2856),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2856),
    [anon_sym_const] = ACTIONS(2856),
    [anon_sym_constexpr] = ACTIONS(2856),
    [anon_sym_volatile] = ACTIONS(2856),
    [anon_sym_restrict] = ACTIONS(2856),
    [anon_sym___restrict__] = ACTIONS(2856),
    [anon_sym__Atomic] = ACTIONS(2856),
    [anon_sym__Noreturn] = ACTIONS(2856),
    [anon_sym__Nonnull] = ACTIONS(2856),
    [anon_sym_nullable] = ACTIONS(2856),
    [anon_sym__Complex] = ACTIONS(2856),
    [anon_sym__Nullable] = ACTIONS(2856),
    [anon_sym__Nullable_result] = ACTIONS(2856),
    [anon_sym__Null_unspecified] = ACTIONS(2856),
    [anon_sym___autoreleasing] = ACTIONS(2856),
    [anon_sym___block] = ACTIONS(2856),
    [anon_sym___bridge] = ACTIONS(2856),
    [anon_sym___bridge_retained] = ACTIONS(2856),
    [anon_sym___bridge_transfer] = ACTIONS(2856),
    [anon_sym___complex] = ACTIONS(2856),
    [anon_sym___const] = ACTIONS(2856),
    [anon_sym___imag] = ACTIONS(2856),
    [anon_sym___kindof] = ACTIONS(2856),
    [anon_sym___nonnull] = ACTIONS(2856),
    [anon_sym___nullable] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2856),
    [anon_sym___real] = ACTIONS(2856),
    [anon_sym___strong] = ACTIONS(2856),
    [anon_sym___unsafe_unretained] = ACTIONS(2856),
    [anon_sym___unused] = ACTIONS(2856),
    [anon_sym___weak] = ACTIONS(2856),
    [anon_sym_alignas] = ACTIONS(2856),
    [anon_sym__Alignas] = ACTIONS(2856),
    [sym_primitive_type] = ACTIONS(2856),
    [anon_sym_enum] = ACTIONS(2856),
    [anon_sym_struct] = ACTIONS(2856),
    [anon_sym_union] = ACTIONS(2856),
    [anon_sym_if] = ACTIONS(2856),
    [anon_sym_switch] = ACTIONS(2856),
    [anon_sym_case] = ACTIONS(2856),
    [anon_sym_default] = ACTIONS(2856),
    [anon_sym_while] = ACTIONS(2856),
    [anon_sym_do] = ACTIONS(2856),
    [anon_sym_for] = ACTIONS(2856),
    [anon_sym_in] = ACTIONS(2856),
    [anon_sym_return] = ACTIONS(2856),
    [anon_sym_break] = ACTIONS(2856),
    [anon_sym_continue] = ACTIONS(2856),
    [anon_sym_goto] = ACTIONS(2856),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_sizeof] = ACTIONS(2856),
    [anon_sym___alignof__] = ACTIONS(2856),
    [anon_sym___alignof] = ACTIONS(2856),
    [anon_sym__alignof] = ACTIONS(2856),
    [anon_sym_alignof] = ACTIONS(2856),
    [anon_sym__Alignof] = ACTIONS(2856),
    [anon_sym_offsetof] = ACTIONS(2856),
    [anon_sym__Generic] = ACTIONS(2856),
    [anon_sym_asm] = ACTIONS(2856),
    [anon_sym___asm__] = ACTIONS(2856),
    [anon_sym___asm] = ACTIONS(2856),
    [sym_number_literal] = ACTIONS(2858),
    [anon_sym_L_SQUOTE] = ACTIONS(2858),
    [anon_sym_u_SQUOTE] = ACTIONS(2858),
    [anon_sym_U_SQUOTE] = ACTIONS(2858),
    [anon_sym_u8_SQUOTE] = ACTIONS(2858),
    [anon_sym_SQUOTE] = ACTIONS(2858),
    [anon_sym_AT] = ACTIONS(2856),
    [anon_sym_DQUOTE] = ACTIONS(2858),
    [anon_sym_L_DQUOTE] = ACTIONS(2858),
    [anon_sym_u_DQUOTE] = ACTIONS(2858),
    [anon_sym_U_DQUOTE] = ACTIONS(2858),
    [anon_sym_u8_DQUOTE] = ACTIONS(2858),
    [sym_true] = ACTIONS(2856),
    [sym_false] = ACTIONS(2856),
    [anon_sym_NULL] = ACTIONS(2856),
    [anon_sym_nullptr] = ACTIONS(2856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2856),
    [anon_sym___typeof] = ACTIONS(2856),
    [anon_sym_typeof] = ACTIONS(2856),
    [anon_sym_ATimport] = ACTIONS(2858),
    [aux_sym_preproc_undef_token1] = ACTIONS(2856),
    [anon_sym_POUND] = ACTIONS(2856),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2856),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2856),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2856),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_API_AVAILABLE] = ACTIONS(2856),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_API_DEPRECATED] = ACTIONS(2856),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym___deprecated_msg] = ACTIONS(2856),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2856),
    [anon_sym_ATprotocol] = ACTIONS(2858),
    [anon_sym_ATinterface] = ACTIONS(2858),
    [anon_sym_ATimplementation] = ACTIONS(2858),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2858),
    [anon_sym_ATtry] = ACTIONS(2858),
    [anon_sym___try] = ACTIONS(2856),
    [anon_sym_ATthrow] = ACTIONS(2858),
    [anon_sym_ATselector] = ACTIONS(2858),
    [anon_sym_ATavailable] = ACTIONS(2858),
    [anon_sym___builtin_available] = ACTIONS(2856),
    [anon_sym_va_arg] = ACTIONS(2856),
    [anon_sym_ATencode] = ACTIONS(2858),
    [anon_sym_ATsynchronized] = ACTIONS(2858),
    [anon_sym_BOOL] = ACTIONS(2856),
    [anon_sym_IMP] = ACTIONS(2856),
    [anon_sym_SEL] = ACTIONS(2856),
    [anon_sym_Class] = ACTIONS(2856),
    [anon_sym_id] = ACTIONS(2856),
  },
  [STATE(902)] = {
    [sym_identifier] = ACTIONS(2756),
    [aux_sym_preproc_include_token1] = ACTIONS(2756),
    [aux_sym_preproc_include_token2] = ACTIONS(2756),
    [aux_sym_preproc_def_token1] = ACTIONS(2756),
    [aux_sym_preproc_if_token1] = ACTIONS(2756),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2756),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2756),
    [sym_preproc_directive] = ACTIONS(2756),
    [anon_sym_LPAREN2] = ACTIONS(2758),
    [anon_sym_BANG] = ACTIONS(2758),
    [anon_sym_TILDE] = ACTIONS(2758),
    [anon_sym_DASH] = ACTIONS(2756),
    [anon_sym_PLUS] = ACTIONS(2756),
    [anon_sym_STAR] = ACTIONS(2758),
    [anon_sym_CARET] = ACTIONS(2758),
    [anon_sym_AMP] = ACTIONS(2758),
    [anon_sym_SEMI] = ACTIONS(2758),
    [anon_sym___extension__] = ACTIONS(2756),
    [anon_sym_typedef] = ACTIONS(2756),
    [anon_sym_extern] = ACTIONS(2756),
    [anon_sym___attribute__] = ACTIONS(2756),
    [anon_sym___attribute] = ACTIONS(2756),
    [anon_sym_noreturn] = ACTIONS(2756),
    [anon_sym_LBRACK] = ACTIONS(2758),
    [anon_sym___declspec] = ACTIONS(2756),
    [anon_sym___cdecl] = ACTIONS(2756),
    [anon_sym___clrcall] = ACTIONS(2756),
    [anon_sym___stdcall] = ACTIONS(2756),
    [anon_sym___fastcall] = ACTIONS(2756),
    [anon_sym___thiscall] = ACTIONS(2756),
    [anon_sym___vectorcall] = ACTIONS(2756),
    [anon_sym_LBRACE] = ACTIONS(2758),
    [anon_sym_RBRACE] = ACTIONS(2758),
    [anon_sym_signed] = ACTIONS(2756),
    [anon_sym_unsigned] = ACTIONS(2756),
    [anon_sym_long] = ACTIONS(2756),
    [anon_sym_short] = ACTIONS(2756),
    [anon_sym_static] = ACTIONS(2756),
    [anon_sym_ATautoreleasepool] = ACTIONS(2758),
    [anon_sym_auto] = ACTIONS(2756),
    [anon_sym_register] = ACTIONS(2756),
    [anon_sym_inline] = ACTIONS(2756),
    [anon_sym___inline] = ACTIONS(2756),
    [anon_sym___inline__] = ACTIONS(2756),
    [anon_sym___forceinline] = ACTIONS(2756),
    [anon_sym_thread_local] = ACTIONS(2756),
    [anon_sym___thread] = ACTIONS(2756),
    [anon_sym_CG_EXTERN] = ACTIONS(2756),
    [anon_sym_CG_INLINE] = ACTIONS(2756),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2756),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2756),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2756),
    [anon_sym_IBOutlet] = ACTIONS(2756),
    [anon_sym_IBInspectable] = ACTIONS(2756),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2756),
    [anon_sym_NS_INLINE] = ACTIONS(2756),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2756),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2756),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2756),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2756),
    [anon_sym_const] = ACTIONS(2756),
    [anon_sym_constexpr] = ACTIONS(2756),
    [anon_sym_volatile] = ACTIONS(2756),
    [anon_sym_restrict] = ACTIONS(2756),
    [anon_sym___restrict__] = ACTIONS(2756),
    [anon_sym__Atomic] = ACTIONS(2756),
    [anon_sym__Noreturn] = ACTIONS(2756),
    [anon_sym__Nonnull] = ACTIONS(2756),
    [anon_sym_nullable] = ACTIONS(2756),
    [anon_sym__Complex] = ACTIONS(2756),
    [anon_sym__Nullable] = ACTIONS(2756),
    [anon_sym__Nullable_result] = ACTIONS(2756),
    [anon_sym__Null_unspecified] = ACTIONS(2756),
    [anon_sym___autoreleasing] = ACTIONS(2756),
    [anon_sym___block] = ACTIONS(2756),
    [anon_sym___bridge] = ACTIONS(2756),
    [anon_sym___bridge_retained] = ACTIONS(2756),
    [anon_sym___bridge_transfer] = ACTIONS(2756),
    [anon_sym___complex] = ACTIONS(2756),
    [anon_sym___const] = ACTIONS(2756),
    [anon_sym___imag] = ACTIONS(2756),
    [anon_sym___kindof] = ACTIONS(2756),
    [anon_sym___nonnull] = ACTIONS(2756),
    [anon_sym___nullable] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2756),
    [anon_sym___real] = ACTIONS(2756),
    [anon_sym___strong] = ACTIONS(2756),
    [anon_sym___unsafe_unretained] = ACTIONS(2756),
    [anon_sym___unused] = ACTIONS(2756),
    [anon_sym___weak] = ACTIONS(2756),
    [anon_sym_alignas] = ACTIONS(2756),
    [anon_sym__Alignas] = ACTIONS(2756),
    [sym_primitive_type] = ACTIONS(2756),
    [anon_sym_enum] = ACTIONS(2756),
    [anon_sym_struct] = ACTIONS(2756),
    [anon_sym_union] = ACTIONS(2756),
    [anon_sym_if] = ACTIONS(2756),
    [anon_sym_switch] = ACTIONS(2756),
    [anon_sym_case] = ACTIONS(2756),
    [anon_sym_default] = ACTIONS(2756),
    [anon_sym_while] = ACTIONS(2756),
    [anon_sym_do] = ACTIONS(2756),
    [anon_sym_for] = ACTIONS(2756),
    [anon_sym_in] = ACTIONS(2756),
    [anon_sym_return] = ACTIONS(2756),
    [anon_sym_break] = ACTIONS(2756),
    [anon_sym_continue] = ACTIONS(2756),
    [anon_sym_goto] = ACTIONS(2756),
    [anon_sym_DASH_DASH] = ACTIONS(2758),
    [anon_sym_PLUS_PLUS] = ACTIONS(2758),
    [anon_sym_sizeof] = ACTIONS(2756),
    [anon_sym___alignof__] = ACTIONS(2756),
    [anon_sym___alignof] = ACTIONS(2756),
    [anon_sym__alignof] = ACTIONS(2756),
    [anon_sym_alignof] = ACTIONS(2756),
    [anon_sym__Alignof] = ACTIONS(2756),
    [anon_sym_offsetof] = ACTIONS(2756),
    [anon_sym__Generic] = ACTIONS(2756),
    [anon_sym_asm] = ACTIONS(2756),
    [anon_sym___asm__] = ACTIONS(2756),
    [anon_sym___asm] = ACTIONS(2756),
    [sym_number_literal] = ACTIONS(2758),
    [anon_sym_L_SQUOTE] = ACTIONS(2758),
    [anon_sym_u_SQUOTE] = ACTIONS(2758),
    [anon_sym_U_SQUOTE] = ACTIONS(2758),
    [anon_sym_u8_SQUOTE] = ACTIONS(2758),
    [anon_sym_SQUOTE] = ACTIONS(2758),
    [anon_sym_AT] = ACTIONS(2756),
    [anon_sym_DQUOTE] = ACTIONS(2758),
    [anon_sym_L_DQUOTE] = ACTIONS(2758),
    [anon_sym_u_DQUOTE] = ACTIONS(2758),
    [anon_sym_U_DQUOTE] = ACTIONS(2758),
    [anon_sym_u8_DQUOTE] = ACTIONS(2758),
    [sym_true] = ACTIONS(2756),
    [sym_false] = ACTIONS(2756),
    [anon_sym_NULL] = ACTIONS(2756),
    [anon_sym_nullptr] = ACTIONS(2756),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2756),
    [anon_sym___typeof] = ACTIONS(2756),
    [anon_sym_typeof] = ACTIONS(2756),
    [anon_sym_ATimport] = ACTIONS(2758),
    [aux_sym_preproc_undef_token1] = ACTIONS(2756),
    [anon_sym_POUND] = ACTIONS(2756),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2756),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2756),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2756),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2756),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2756),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2756),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2756),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_API_AVAILABLE] = ACTIONS(2756),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_API_DEPRECATED] = ACTIONS(2756),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2756),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym___deprecated_msg] = ACTIONS(2756),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2756),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2756),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2756),
    [anon_sym_ATprotocol] = ACTIONS(2758),
    [anon_sym_ATinterface] = ACTIONS(2758),
    [anon_sym_ATimplementation] = ACTIONS(2758),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2758),
    [anon_sym_ATtry] = ACTIONS(2758),
    [anon_sym___try] = ACTIONS(2756),
    [anon_sym_ATthrow] = ACTIONS(2758),
    [anon_sym_ATselector] = ACTIONS(2758),
    [anon_sym_ATavailable] = ACTIONS(2758),
    [anon_sym___builtin_available] = ACTIONS(2756),
    [anon_sym_va_arg] = ACTIONS(2756),
    [anon_sym_ATencode] = ACTIONS(2758),
    [anon_sym_ATsynchronized] = ACTIONS(2758),
    [anon_sym_BOOL] = ACTIONS(2756),
    [anon_sym_IMP] = ACTIONS(2756),
    [anon_sym_SEL] = ACTIONS(2756),
    [anon_sym_Class] = ACTIONS(2756),
    [anon_sym_id] = ACTIONS(2756),
  },
  [STATE(903)] = {
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_RBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATtry] = ACTIONS(2846),
    [anon_sym___try] = ACTIONS(2844),
    [anon_sym_ATthrow] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_ATsynchronized] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(904)] = {
    [sym_identifier] = ACTIONS(2848),
    [aux_sym_preproc_include_token1] = ACTIONS(2848),
    [aux_sym_preproc_include_token2] = ACTIONS(2848),
    [aux_sym_preproc_def_token1] = ACTIONS(2848),
    [aux_sym_preproc_if_token1] = ACTIONS(2848),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2848),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2848),
    [sym_preproc_directive] = ACTIONS(2848),
    [anon_sym_LPAREN2] = ACTIONS(2850),
    [anon_sym_BANG] = ACTIONS(2850),
    [anon_sym_TILDE] = ACTIONS(2850),
    [anon_sym_DASH] = ACTIONS(2848),
    [anon_sym_PLUS] = ACTIONS(2848),
    [anon_sym_STAR] = ACTIONS(2850),
    [anon_sym_CARET] = ACTIONS(2850),
    [anon_sym_AMP] = ACTIONS(2850),
    [anon_sym_SEMI] = ACTIONS(2850),
    [anon_sym___extension__] = ACTIONS(2848),
    [anon_sym_typedef] = ACTIONS(2848),
    [anon_sym_extern] = ACTIONS(2848),
    [anon_sym___attribute__] = ACTIONS(2848),
    [anon_sym___attribute] = ACTIONS(2848),
    [anon_sym_noreturn] = ACTIONS(2848),
    [anon_sym_LBRACK] = ACTIONS(2850),
    [anon_sym___declspec] = ACTIONS(2848),
    [anon_sym___cdecl] = ACTIONS(2848),
    [anon_sym___clrcall] = ACTIONS(2848),
    [anon_sym___stdcall] = ACTIONS(2848),
    [anon_sym___fastcall] = ACTIONS(2848),
    [anon_sym___thiscall] = ACTIONS(2848),
    [anon_sym___vectorcall] = ACTIONS(2848),
    [anon_sym_LBRACE] = ACTIONS(2850),
    [anon_sym_RBRACE] = ACTIONS(2850),
    [anon_sym_signed] = ACTIONS(2848),
    [anon_sym_unsigned] = ACTIONS(2848),
    [anon_sym_long] = ACTIONS(2848),
    [anon_sym_short] = ACTIONS(2848),
    [anon_sym_static] = ACTIONS(2848),
    [anon_sym_ATautoreleasepool] = ACTIONS(2850),
    [anon_sym_auto] = ACTIONS(2848),
    [anon_sym_register] = ACTIONS(2848),
    [anon_sym_inline] = ACTIONS(2848),
    [anon_sym___inline] = ACTIONS(2848),
    [anon_sym___inline__] = ACTIONS(2848),
    [anon_sym___forceinline] = ACTIONS(2848),
    [anon_sym_thread_local] = ACTIONS(2848),
    [anon_sym___thread] = ACTIONS(2848),
    [anon_sym_CG_EXTERN] = ACTIONS(2848),
    [anon_sym_CG_INLINE] = ACTIONS(2848),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2848),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2848),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2848),
    [anon_sym_IBOutlet] = ACTIONS(2848),
    [anon_sym_IBInspectable] = ACTIONS(2848),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2848),
    [anon_sym_NS_INLINE] = ACTIONS(2848),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2848),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2848),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2848),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2848),
    [anon_sym_const] = ACTIONS(2848),
    [anon_sym_constexpr] = ACTIONS(2848),
    [anon_sym_volatile] = ACTIONS(2848),
    [anon_sym_restrict] = ACTIONS(2848),
    [anon_sym___restrict__] = ACTIONS(2848),
    [anon_sym__Atomic] = ACTIONS(2848),
    [anon_sym__Noreturn] = ACTIONS(2848),
    [anon_sym__Nonnull] = ACTIONS(2848),
    [anon_sym_nullable] = ACTIONS(2848),
    [anon_sym__Complex] = ACTIONS(2848),
    [anon_sym__Nullable] = ACTIONS(2848),
    [anon_sym__Nullable_result] = ACTIONS(2848),
    [anon_sym__Null_unspecified] = ACTIONS(2848),
    [anon_sym___autoreleasing] = ACTIONS(2848),
    [anon_sym___block] = ACTIONS(2848),
    [anon_sym___bridge] = ACTIONS(2848),
    [anon_sym___bridge_retained] = ACTIONS(2848),
    [anon_sym___bridge_transfer] = ACTIONS(2848),
    [anon_sym___complex] = ACTIONS(2848),
    [anon_sym___const] = ACTIONS(2848),
    [anon_sym___imag] = ACTIONS(2848),
    [anon_sym___kindof] = ACTIONS(2848),
    [anon_sym___nonnull] = ACTIONS(2848),
    [anon_sym___nullable] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2848),
    [anon_sym___real] = ACTIONS(2848),
    [anon_sym___strong] = ACTIONS(2848),
    [anon_sym___unsafe_unretained] = ACTIONS(2848),
    [anon_sym___unused] = ACTIONS(2848),
    [anon_sym___weak] = ACTIONS(2848),
    [anon_sym_alignas] = ACTIONS(2848),
    [anon_sym__Alignas] = ACTIONS(2848),
    [sym_primitive_type] = ACTIONS(2848),
    [anon_sym_enum] = ACTIONS(2848),
    [anon_sym_struct] = ACTIONS(2848),
    [anon_sym_union] = ACTIONS(2848),
    [anon_sym_if] = ACTIONS(2848),
    [anon_sym_switch] = ACTIONS(2848),
    [anon_sym_case] = ACTIONS(2848),
    [anon_sym_default] = ACTIONS(2848),
    [anon_sym_while] = ACTIONS(2848),
    [anon_sym_do] = ACTIONS(2848),
    [anon_sym_for] = ACTIONS(2848),
    [anon_sym_in] = ACTIONS(2848),
    [anon_sym_return] = ACTIONS(2848),
    [anon_sym_break] = ACTIONS(2848),
    [anon_sym_continue] = ACTIONS(2848),
    [anon_sym_goto] = ACTIONS(2848),
    [anon_sym_DASH_DASH] = ACTIONS(2850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2850),
    [anon_sym_sizeof] = ACTIONS(2848),
    [anon_sym___alignof__] = ACTIONS(2848),
    [anon_sym___alignof] = ACTIONS(2848),
    [anon_sym__alignof] = ACTIONS(2848),
    [anon_sym_alignof] = ACTIONS(2848),
    [anon_sym__Alignof] = ACTIONS(2848),
    [anon_sym_offsetof] = ACTIONS(2848),
    [anon_sym__Generic] = ACTIONS(2848),
    [anon_sym_asm] = ACTIONS(2848),
    [anon_sym___asm__] = ACTIONS(2848),
    [anon_sym___asm] = ACTIONS(2848),
    [sym_number_literal] = ACTIONS(2850),
    [anon_sym_L_SQUOTE] = ACTIONS(2850),
    [anon_sym_u_SQUOTE] = ACTIONS(2850),
    [anon_sym_U_SQUOTE] = ACTIONS(2850),
    [anon_sym_u8_SQUOTE] = ACTIONS(2850),
    [anon_sym_SQUOTE] = ACTIONS(2850),
    [anon_sym_AT] = ACTIONS(2848),
    [anon_sym_DQUOTE] = ACTIONS(2850),
    [anon_sym_L_DQUOTE] = ACTIONS(2850),
    [anon_sym_u_DQUOTE] = ACTIONS(2850),
    [anon_sym_U_DQUOTE] = ACTIONS(2850),
    [anon_sym_u8_DQUOTE] = ACTIONS(2850),
    [sym_true] = ACTIONS(2848),
    [sym_false] = ACTIONS(2848),
    [anon_sym_NULL] = ACTIONS(2848),
    [anon_sym_nullptr] = ACTIONS(2848),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2848),
    [anon_sym___typeof] = ACTIONS(2848),
    [anon_sym_typeof] = ACTIONS(2848),
    [anon_sym_ATimport] = ACTIONS(2850),
    [aux_sym_preproc_undef_token1] = ACTIONS(2848),
    [anon_sym_POUND] = ACTIONS(2848),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2848),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2848),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2848),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2848),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2848),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2848),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2848),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_API_AVAILABLE] = ACTIONS(2848),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_API_DEPRECATED] = ACTIONS(2848),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2848),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym___deprecated_msg] = ACTIONS(2848),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2848),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2848),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2848),
    [anon_sym_ATprotocol] = ACTIONS(2850),
    [anon_sym_ATinterface] = ACTIONS(2850),
    [anon_sym_ATimplementation] = ACTIONS(2850),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2850),
    [anon_sym_ATtry] = ACTIONS(2850),
    [anon_sym___try] = ACTIONS(2848),
    [anon_sym_ATthrow] = ACTIONS(2850),
    [anon_sym_ATselector] = ACTIONS(2850),
    [anon_sym_ATavailable] = ACTIONS(2850),
    [anon_sym___builtin_available] = ACTIONS(2848),
    [anon_sym_va_arg] = ACTIONS(2848),
    [anon_sym_ATencode] = ACTIONS(2850),
    [anon_sym_ATsynchronized] = ACTIONS(2850),
    [anon_sym_BOOL] = ACTIONS(2848),
    [anon_sym_IMP] = ACTIONS(2848),
    [anon_sym_SEL] = ACTIONS(2848),
    [anon_sym_Class] = ACTIONS(2848),
    [anon_sym_id] = ACTIONS(2848),
  },
  [STATE(905)] = {
    [sym_identifier] = ACTIONS(2572),
    [aux_sym_preproc_include_token1] = ACTIONS(2572),
    [aux_sym_preproc_include_token2] = ACTIONS(2572),
    [aux_sym_preproc_def_token1] = ACTIONS(2572),
    [aux_sym_preproc_if_token1] = ACTIONS(2572),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2572),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2572),
    [sym_preproc_directive] = ACTIONS(2572),
    [anon_sym_LPAREN2] = ACTIONS(2574),
    [anon_sym_BANG] = ACTIONS(2574),
    [anon_sym_TILDE] = ACTIONS(2574),
    [anon_sym_DASH] = ACTIONS(2572),
    [anon_sym_PLUS] = ACTIONS(2572),
    [anon_sym_STAR] = ACTIONS(2574),
    [anon_sym_CARET] = ACTIONS(2574),
    [anon_sym_AMP] = ACTIONS(2574),
    [anon_sym_SEMI] = ACTIONS(2574),
    [anon_sym___extension__] = ACTIONS(2572),
    [anon_sym_typedef] = ACTIONS(2572),
    [anon_sym_extern] = ACTIONS(2572),
    [anon_sym___attribute__] = ACTIONS(2572),
    [anon_sym___attribute] = ACTIONS(2572),
    [anon_sym_noreturn] = ACTIONS(2572),
    [anon_sym_LBRACK] = ACTIONS(2574),
    [anon_sym___declspec] = ACTIONS(2572),
    [anon_sym___cdecl] = ACTIONS(2572),
    [anon_sym___clrcall] = ACTIONS(2572),
    [anon_sym___stdcall] = ACTIONS(2572),
    [anon_sym___fastcall] = ACTIONS(2572),
    [anon_sym___thiscall] = ACTIONS(2572),
    [anon_sym___vectorcall] = ACTIONS(2572),
    [anon_sym_LBRACE] = ACTIONS(2574),
    [anon_sym_RBRACE] = ACTIONS(2574),
    [anon_sym_signed] = ACTIONS(2572),
    [anon_sym_unsigned] = ACTIONS(2572),
    [anon_sym_long] = ACTIONS(2572),
    [anon_sym_short] = ACTIONS(2572),
    [anon_sym_static] = ACTIONS(2572),
    [anon_sym_ATautoreleasepool] = ACTIONS(2574),
    [anon_sym_auto] = ACTIONS(2572),
    [anon_sym_register] = ACTIONS(2572),
    [anon_sym_inline] = ACTIONS(2572),
    [anon_sym___inline] = ACTIONS(2572),
    [anon_sym___inline__] = ACTIONS(2572),
    [anon_sym___forceinline] = ACTIONS(2572),
    [anon_sym_thread_local] = ACTIONS(2572),
    [anon_sym___thread] = ACTIONS(2572),
    [anon_sym_CG_EXTERN] = ACTIONS(2572),
    [anon_sym_CG_INLINE] = ACTIONS(2572),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2572),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2572),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2572),
    [anon_sym_IBOutlet] = ACTIONS(2572),
    [anon_sym_IBInspectable] = ACTIONS(2572),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2572),
    [anon_sym_NS_INLINE] = ACTIONS(2572),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2572),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2572),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2572),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2572),
    [anon_sym_const] = ACTIONS(2572),
    [anon_sym_constexpr] = ACTIONS(2572),
    [anon_sym_volatile] = ACTIONS(2572),
    [anon_sym_restrict] = ACTIONS(2572),
    [anon_sym___restrict__] = ACTIONS(2572),
    [anon_sym__Atomic] = ACTIONS(2572),
    [anon_sym__Noreturn] = ACTIONS(2572),
    [anon_sym__Nonnull] = ACTIONS(2572),
    [anon_sym_nullable] = ACTIONS(2572),
    [anon_sym__Complex] = ACTIONS(2572),
    [anon_sym__Nullable] = ACTIONS(2572),
    [anon_sym__Nullable_result] = ACTIONS(2572),
    [anon_sym__Null_unspecified] = ACTIONS(2572),
    [anon_sym___autoreleasing] = ACTIONS(2572),
    [anon_sym___block] = ACTIONS(2572),
    [anon_sym___bridge] = ACTIONS(2572),
    [anon_sym___bridge_retained] = ACTIONS(2572),
    [anon_sym___bridge_transfer] = ACTIONS(2572),
    [anon_sym___complex] = ACTIONS(2572),
    [anon_sym___const] = ACTIONS(2572),
    [anon_sym___imag] = ACTIONS(2572),
    [anon_sym___kindof] = ACTIONS(2572),
    [anon_sym___nonnull] = ACTIONS(2572),
    [anon_sym___nullable] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2572),
    [anon_sym___real] = ACTIONS(2572),
    [anon_sym___strong] = ACTIONS(2572),
    [anon_sym___unsafe_unretained] = ACTIONS(2572),
    [anon_sym___unused] = ACTIONS(2572),
    [anon_sym___weak] = ACTIONS(2572),
    [anon_sym_alignas] = ACTIONS(2572),
    [anon_sym__Alignas] = ACTIONS(2572),
    [sym_primitive_type] = ACTIONS(2572),
    [anon_sym_enum] = ACTIONS(2572),
    [anon_sym_struct] = ACTIONS(2572),
    [anon_sym_union] = ACTIONS(2572),
    [anon_sym_if] = ACTIONS(2572),
    [anon_sym_switch] = ACTIONS(2572),
    [anon_sym_case] = ACTIONS(2572),
    [anon_sym_default] = ACTIONS(2572),
    [anon_sym_while] = ACTIONS(2572),
    [anon_sym_do] = ACTIONS(2572),
    [anon_sym_for] = ACTIONS(2572),
    [anon_sym_in] = ACTIONS(2572),
    [anon_sym_return] = ACTIONS(2572),
    [anon_sym_break] = ACTIONS(2572),
    [anon_sym_continue] = ACTIONS(2572),
    [anon_sym_goto] = ACTIONS(2572),
    [anon_sym_DASH_DASH] = ACTIONS(2574),
    [anon_sym_PLUS_PLUS] = ACTIONS(2574),
    [anon_sym_sizeof] = ACTIONS(2572),
    [anon_sym___alignof__] = ACTIONS(2572),
    [anon_sym___alignof] = ACTIONS(2572),
    [anon_sym__alignof] = ACTIONS(2572),
    [anon_sym_alignof] = ACTIONS(2572),
    [anon_sym__Alignof] = ACTIONS(2572),
    [anon_sym_offsetof] = ACTIONS(2572),
    [anon_sym__Generic] = ACTIONS(2572),
    [anon_sym_asm] = ACTIONS(2572),
    [anon_sym___asm__] = ACTIONS(2572),
    [anon_sym___asm] = ACTIONS(2572),
    [sym_number_literal] = ACTIONS(2574),
    [anon_sym_L_SQUOTE] = ACTIONS(2574),
    [anon_sym_u_SQUOTE] = ACTIONS(2574),
    [anon_sym_U_SQUOTE] = ACTIONS(2574),
    [anon_sym_u8_SQUOTE] = ACTIONS(2574),
    [anon_sym_SQUOTE] = ACTIONS(2574),
    [anon_sym_AT] = ACTIONS(2572),
    [anon_sym_DQUOTE] = ACTIONS(2574),
    [anon_sym_L_DQUOTE] = ACTIONS(2574),
    [anon_sym_u_DQUOTE] = ACTIONS(2574),
    [anon_sym_U_DQUOTE] = ACTIONS(2574),
    [anon_sym_u8_DQUOTE] = ACTIONS(2574),
    [sym_true] = ACTIONS(2572),
    [sym_false] = ACTIONS(2572),
    [anon_sym_NULL] = ACTIONS(2572),
    [anon_sym_nullptr] = ACTIONS(2572),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2572),
    [anon_sym___typeof] = ACTIONS(2572),
    [anon_sym_typeof] = ACTIONS(2572),
    [anon_sym_ATimport] = ACTIONS(2574),
    [aux_sym_preproc_undef_token1] = ACTIONS(2572),
    [anon_sym_POUND] = ACTIONS(2572),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2572),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2572),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2572),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2572),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2572),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2572),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2572),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_API_AVAILABLE] = ACTIONS(2572),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_API_DEPRECATED] = ACTIONS(2572),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2572),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym___deprecated_msg] = ACTIONS(2572),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2572),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2572),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2572),
    [anon_sym_ATprotocol] = ACTIONS(2574),
    [anon_sym_ATinterface] = ACTIONS(2574),
    [anon_sym_ATimplementation] = ACTIONS(2574),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2574),
    [anon_sym_ATtry] = ACTIONS(2574),
    [anon_sym___try] = ACTIONS(2572),
    [anon_sym_ATthrow] = ACTIONS(2574),
    [anon_sym_ATselector] = ACTIONS(2574),
    [anon_sym_ATavailable] = ACTIONS(2574),
    [anon_sym___builtin_available] = ACTIONS(2572),
    [anon_sym_va_arg] = ACTIONS(2572),
    [anon_sym_ATencode] = ACTIONS(2574),
    [anon_sym_ATsynchronized] = ACTIONS(2574),
    [anon_sym_BOOL] = ACTIONS(2572),
    [anon_sym_IMP] = ACTIONS(2572),
    [anon_sym_SEL] = ACTIONS(2572),
    [anon_sym_Class] = ACTIONS(2572),
    [anon_sym_id] = ACTIONS(2572),
  },
  [STATE(906)] = {
    [sym_identifier] = ACTIONS(2658),
    [aux_sym_preproc_include_token1] = ACTIONS(2658),
    [aux_sym_preproc_include_token2] = ACTIONS(2658),
    [aux_sym_preproc_def_token1] = ACTIONS(2658),
    [aux_sym_preproc_if_token1] = ACTIONS(2658),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2658),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2658),
    [sym_preproc_directive] = ACTIONS(2658),
    [anon_sym_LPAREN2] = ACTIONS(2660),
    [anon_sym_BANG] = ACTIONS(2660),
    [anon_sym_TILDE] = ACTIONS(2660),
    [anon_sym_DASH] = ACTIONS(2658),
    [anon_sym_PLUS] = ACTIONS(2658),
    [anon_sym_STAR] = ACTIONS(2660),
    [anon_sym_CARET] = ACTIONS(2660),
    [anon_sym_AMP] = ACTIONS(2660),
    [anon_sym_SEMI] = ACTIONS(2660),
    [anon_sym___extension__] = ACTIONS(2658),
    [anon_sym_typedef] = ACTIONS(2658),
    [anon_sym_extern] = ACTIONS(2658),
    [anon_sym___attribute__] = ACTIONS(2658),
    [anon_sym___attribute] = ACTIONS(2658),
    [anon_sym_noreturn] = ACTIONS(2658),
    [anon_sym_LBRACK] = ACTIONS(2660),
    [anon_sym___declspec] = ACTIONS(2658),
    [anon_sym___cdecl] = ACTIONS(2658),
    [anon_sym___clrcall] = ACTIONS(2658),
    [anon_sym___stdcall] = ACTIONS(2658),
    [anon_sym___fastcall] = ACTIONS(2658),
    [anon_sym___thiscall] = ACTIONS(2658),
    [anon_sym___vectorcall] = ACTIONS(2658),
    [anon_sym_LBRACE] = ACTIONS(2660),
    [anon_sym_RBRACE] = ACTIONS(2660),
    [anon_sym_signed] = ACTIONS(2658),
    [anon_sym_unsigned] = ACTIONS(2658),
    [anon_sym_long] = ACTIONS(2658),
    [anon_sym_short] = ACTIONS(2658),
    [anon_sym_static] = ACTIONS(2658),
    [anon_sym_ATautoreleasepool] = ACTIONS(2660),
    [anon_sym_auto] = ACTIONS(2658),
    [anon_sym_register] = ACTIONS(2658),
    [anon_sym_inline] = ACTIONS(2658),
    [anon_sym___inline] = ACTIONS(2658),
    [anon_sym___inline__] = ACTIONS(2658),
    [anon_sym___forceinline] = ACTIONS(2658),
    [anon_sym_thread_local] = ACTIONS(2658),
    [anon_sym___thread] = ACTIONS(2658),
    [anon_sym_CG_EXTERN] = ACTIONS(2658),
    [anon_sym_CG_INLINE] = ACTIONS(2658),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2658),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2658),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2658),
    [anon_sym_IBOutlet] = ACTIONS(2658),
    [anon_sym_IBInspectable] = ACTIONS(2658),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2658),
    [anon_sym_NS_INLINE] = ACTIONS(2658),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2658),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2658),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2658),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2658),
    [anon_sym_const] = ACTIONS(2658),
    [anon_sym_constexpr] = ACTIONS(2658),
    [anon_sym_volatile] = ACTIONS(2658),
    [anon_sym_restrict] = ACTIONS(2658),
    [anon_sym___restrict__] = ACTIONS(2658),
    [anon_sym__Atomic] = ACTIONS(2658),
    [anon_sym__Noreturn] = ACTIONS(2658),
    [anon_sym__Nonnull] = ACTIONS(2658),
    [anon_sym_nullable] = ACTIONS(2658),
    [anon_sym__Complex] = ACTIONS(2658),
    [anon_sym__Nullable] = ACTIONS(2658),
    [anon_sym__Nullable_result] = ACTIONS(2658),
    [anon_sym__Null_unspecified] = ACTIONS(2658),
    [anon_sym___autoreleasing] = ACTIONS(2658),
    [anon_sym___block] = ACTIONS(2658),
    [anon_sym___bridge] = ACTIONS(2658),
    [anon_sym___bridge_retained] = ACTIONS(2658),
    [anon_sym___bridge_transfer] = ACTIONS(2658),
    [anon_sym___complex] = ACTIONS(2658),
    [anon_sym___const] = ACTIONS(2658),
    [anon_sym___imag] = ACTIONS(2658),
    [anon_sym___kindof] = ACTIONS(2658),
    [anon_sym___nonnull] = ACTIONS(2658),
    [anon_sym___nullable] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2658),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2658),
    [anon_sym___real] = ACTIONS(2658),
    [anon_sym___strong] = ACTIONS(2658),
    [anon_sym___unsafe_unretained] = ACTIONS(2658),
    [anon_sym___unused] = ACTIONS(2658),
    [anon_sym___weak] = ACTIONS(2658),
    [anon_sym_alignas] = ACTIONS(2658),
    [anon_sym__Alignas] = ACTIONS(2658),
    [sym_primitive_type] = ACTIONS(2658),
    [anon_sym_enum] = ACTIONS(2658),
    [anon_sym_struct] = ACTIONS(2658),
    [anon_sym_union] = ACTIONS(2658),
    [anon_sym_if] = ACTIONS(2658),
    [anon_sym_switch] = ACTIONS(2658),
    [anon_sym_case] = ACTIONS(2658),
    [anon_sym_default] = ACTIONS(2658),
    [anon_sym_while] = ACTIONS(2658),
    [anon_sym_do] = ACTIONS(2658),
    [anon_sym_for] = ACTIONS(2658),
    [anon_sym_in] = ACTIONS(2658),
    [anon_sym_return] = ACTIONS(2658),
    [anon_sym_break] = ACTIONS(2658),
    [anon_sym_continue] = ACTIONS(2658),
    [anon_sym_goto] = ACTIONS(2658),
    [anon_sym_DASH_DASH] = ACTIONS(2660),
    [anon_sym_PLUS_PLUS] = ACTIONS(2660),
    [anon_sym_sizeof] = ACTIONS(2658),
    [anon_sym___alignof__] = ACTIONS(2658),
    [anon_sym___alignof] = ACTIONS(2658),
    [anon_sym__alignof] = ACTIONS(2658),
    [anon_sym_alignof] = ACTIONS(2658),
    [anon_sym__Alignof] = ACTIONS(2658),
    [anon_sym_offsetof] = ACTIONS(2658),
    [anon_sym__Generic] = ACTIONS(2658),
    [anon_sym_asm] = ACTIONS(2658),
    [anon_sym___asm__] = ACTIONS(2658),
    [anon_sym___asm] = ACTIONS(2658),
    [sym_number_literal] = ACTIONS(2660),
    [anon_sym_L_SQUOTE] = ACTIONS(2660),
    [anon_sym_u_SQUOTE] = ACTIONS(2660),
    [anon_sym_U_SQUOTE] = ACTIONS(2660),
    [anon_sym_u8_SQUOTE] = ACTIONS(2660),
    [anon_sym_SQUOTE] = ACTIONS(2660),
    [anon_sym_AT] = ACTIONS(2658),
    [anon_sym_DQUOTE] = ACTIONS(2660),
    [anon_sym_L_DQUOTE] = ACTIONS(2660),
    [anon_sym_u_DQUOTE] = ACTIONS(2660),
    [anon_sym_U_DQUOTE] = ACTIONS(2660),
    [anon_sym_u8_DQUOTE] = ACTIONS(2660),
    [sym_true] = ACTIONS(2658),
    [sym_false] = ACTIONS(2658),
    [anon_sym_NULL] = ACTIONS(2658),
    [anon_sym_nullptr] = ACTIONS(2658),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2658),
    [anon_sym___typeof] = ACTIONS(2658),
    [anon_sym_typeof] = ACTIONS(2658),
    [anon_sym_ATimport] = ACTIONS(2660),
    [aux_sym_preproc_undef_token1] = ACTIONS(2658),
    [anon_sym_POUND] = ACTIONS(2658),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2658),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2658),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2658),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2658),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2658),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2658),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2658),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_API_AVAILABLE] = ACTIONS(2658),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_API_DEPRECATED] = ACTIONS(2658),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2658),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2658),
    [anon_sym___deprecated_msg] = ACTIONS(2658),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2658),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2658),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2658),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2658),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2658),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2658),
    [anon_sym_ATprotocol] = ACTIONS(2660),
    [anon_sym_ATinterface] = ACTIONS(2660),
    [anon_sym_ATimplementation] = ACTIONS(2660),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2660),
    [anon_sym_ATtry] = ACTIONS(2660),
    [anon_sym___try] = ACTIONS(2658),
    [anon_sym_ATthrow] = ACTIONS(2660),
    [anon_sym_ATselector] = ACTIONS(2660),
    [anon_sym_ATavailable] = ACTIONS(2660),
    [anon_sym___builtin_available] = ACTIONS(2658),
    [anon_sym_va_arg] = ACTIONS(2658),
    [anon_sym_ATencode] = ACTIONS(2660),
    [anon_sym_ATsynchronized] = ACTIONS(2660),
    [anon_sym_BOOL] = ACTIONS(2658),
    [anon_sym_IMP] = ACTIONS(2658),
    [anon_sym_SEL] = ACTIONS(2658),
    [anon_sym_Class] = ACTIONS(2658),
    [anon_sym_id] = ACTIONS(2658),
  },
  [STATE(907)] = {
    [sym_identifier] = ACTIONS(2576),
    [aux_sym_preproc_include_token1] = ACTIONS(2576),
    [aux_sym_preproc_include_token2] = ACTIONS(2576),
    [aux_sym_preproc_def_token1] = ACTIONS(2576),
    [aux_sym_preproc_if_token1] = ACTIONS(2576),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2576),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2576),
    [sym_preproc_directive] = ACTIONS(2576),
    [anon_sym_LPAREN2] = ACTIONS(2578),
    [anon_sym_BANG] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2578),
    [anon_sym_DASH] = ACTIONS(2576),
    [anon_sym_PLUS] = ACTIONS(2576),
    [anon_sym_STAR] = ACTIONS(2578),
    [anon_sym_CARET] = ACTIONS(2578),
    [anon_sym_AMP] = ACTIONS(2578),
    [anon_sym_SEMI] = ACTIONS(2578),
    [anon_sym___extension__] = ACTIONS(2576),
    [anon_sym_typedef] = ACTIONS(2576),
    [anon_sym_extern] = ACTIONS(2576),
    [anon_sym___attribute__] = ACTIONS(2576),
    [anon_sym___attribute] = ACTIONS(2576),
    [anon_sym_noreturn] = ACTIONS(2576),
    [anon_sym_LBRACK] = ACTIONS(2578),
    [anon_sym___declspec] = ACTIONS(2576),
    [anon_sym___cdecl] = ACTIONS(2576),
    [anon_sym___clrcall] = ACTIONS(2576),
    [anon_sym___stdcall] = ACTIONS(2576),
    [anon_sym___fastcall] = ACTIONS(2576),
    [anon_sym___thiscall] = ACTIONS(2576),
    [anon_sym___vectorcall] = ACTIONS(2576),
    [anon_sym_LBRACE] = ACTIONS(2578),
    [anon_sym_RBRACE] = ACTIONS(2578),
    [anon_sym_signed] = ACTIONS(2576),
    [anon_sym_unsigned] = ACTIONS(2576),
    [anon_sym_long] = ACTIONS(2576),
    [anon_sym_short] = ACTIONS(2576),
    [anon_sym_static] = ACTIONS(2576),
    [anon_sym_ATautoreleasepool] = ACTIONS(2578),
    [anon_sym_auto] = ACTIONS(2576),
    [anon_sym_register] = ACTIONS(2576),
    [anon_sym_inline] = ACTIONS(2576),
    [anon_sym___inline] = ACTIONS(2576),
    [anon_sym___inline__] = ACTIONS(2576),
    [anon_sym___forceinline] = ACTIONS(2576),
    [anon_sym_thread_local] = ACTIONS(2576),
    [anon_sym___thread] = ACTIONS(2576),
    [anon_sym_CG_EXTERN] = ACTIONS(2576),
    [anon_sym_CG_INLINE] = ACTIONS(2576),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2576),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2576),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2576),
    [anon_sym_IBOutlet] = ACTIONS(2576),
    [anon_sym_IBInspectable] = ACTIONS(2576),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2576),
    [anon_sym_NS_INLINE] = ACTIONS(2576),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2576),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2576),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2576),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2576),
    [anon_sym_const] = ACTIONS(2576),
    [anon_sym_constexpr] = ACTIONS(2576),
    [anon_sym_volatile] = ACTIONS(2576),
    [anon_sym_restrict] = ACTIONS(2576),
    [anon_sym___restrict__] = ACTIONS(2576),
    [anon_sym__Atomic] = ACTIONS(2576),
    [anon_sym__Noreturn] = ACTIONS(2576),
    [anon_sym__Nonnull] = ACTIONS(2576),
    [anon_sym_nullable] = ACTIONS(2576),
    [anon_sym__Complex] = ACTIONS(2576),
    [anon_sym__Nullable] = ACTIONS(2576),
    [anon_sym__Nullable_result] = ACTIONS(2576),
    [anon_sym__Null_unspecified] = ACTIONS(2576),
    [anon_sym___autoreleasing] = ACTIONS(2576),
    [anon_sym___block] = ACTIONS(2576),
    [anon_sym___bridge] = ACTIONS(2576),
    [anon_sym___bridge_retained] = ACTIONS(2576),
    [anon_sym___bridge_transfer] = ACTIONS(2576),
    [anon_sym___complex] = ACTIONS(2576),
    [anon_sym___const] = ACTIONS(2576),
    [anon_sym___imag] = ACTIONS(2576),
    [anon_sym___kindof] = ACTIONS(2576),
    [anon_sym___nonnull] = ACTIONS(2576),
    [anon_sym___nullable] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2576),
    [anon_sym___real] = ACTIONS(2576),
    [anon_sym___strong] = ACTIONS(2576),
    [anon_sym___unsafe_unretained] = ACTIONS(2576),
    [anon_sym___unused] = ACTIONS(2576),
    [anon_sym___weak] = ACTIONS(2576),
    [anon_sym_alignas] = ACTIONS(2576),
    [anon_sym__Alignas] = ACTIONS(2576),
    [sym_primitive_type] = ACTIONS(2576),
    [anon_sym_enum] = ACTIONS(2576),
    [anon_sym_struct] = ACTIONS(2576),
    [anon_sym_union] = ACTIONS(2576),
    [anon_sym_if] = ACTIONS(2576),
    [anon_sym_switch] = ACTIONS(2576),
    [anon_sym_case] = ACTIONS(2576),
    [anon_sym_default] = ACTIONS(2576),
    [anon_sym_while] = ACTIONS(2576),
    [anon_sym_do] = ACTIONS(2576),
    [anon_sym_for] = ACTIONS(2576),
    [anon_sym_in] = ACTIONS(2576),
    [anon_sym_return] = ACTIONS(2576),
    [anon_sym_break] = ACTIONS(2576),
    [anon_sym_continue] = ACTIONS(2576),
    [anon_sym_goto] = ACTIONS(2576),
    [anon_sym_DASH_DASH] = ACTIONS(2578),
    [anon_sym_PLUS_PLUS] = ACTIONS(2578),
    [anon_sym_sizeof] = ACTIONS(2576),
    [anon_sym___alignof__] = ACTIONS(2576),
    [anon_sym___alignof] = ACTIONS(2576),
    [anon_sym__alignof] = ACTIONS(2576),
    [anon_sym_alignof] = ACTIONS(2576),
    [anon_sym__Alignof] = ACTIONS(2576),
    [anon_sym_offsetof] = ACTIONS(2576),
    [anon_sym__Generic] = ACTIONS(2576),
    [anon_sym_asm] = ACTIONS(2576),
    [anon_sym___asm__] = ACTIONS(2576),
    [anon_sym___asm] = ACTIONS(2576),
    [sym_number_literal] = ACTIONS(2578),
    [anon_sym_L_SQUOTE] = ACTIONS(2578),
    [anon_sym_u_SQUOTE] = ACTIONS(2578),
    [anon_sym_U_SQUOTE] = ACTIONS(2578),
    [anon_sym_u8_SQUOTE] = ACTIONS(2578),
    [anon_sym_SQUOTE] = ACTIONS(2578),
    [anon_sym_AT] = ACTIONS(2576),
    [anon_sym_DQUOTE] = ACTIONS(2578),
    [anon_sym_L_DQUOTE] = ACTIONS(2578),
    [anon_sym_u_DQUOTE] = ACTIONS(2578),
    [anon_sym_U_DQUOTE] = ACTIONS(2578),
    [anon_sym_u8_DQUOTE] = ACTIONS(2578),
    [sym_true] = ACTIONS(2576),
    [sym_false] = ACTIONS(2576),
    [anon_sym_NULL] = ACTIONS(2576),
    [anon_sym_nullptr] = ACTIONS(2576),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2576),
    [anon_sym___typeof] = ACTIONS(2576),
    [anon_sym_typeof] = ACTIONS(2576),
    [anon_sym_ATimport] = ACTIONS(2578),
    [aux_sym_preproc_undef_token1] = ACTIONS(2576),
    [anon_sym_POUND] = ACTIONS(2576),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2576),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2576),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2576),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2576),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2576),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2576),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2576),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_API_AVAILABLE] = ACTIONS(2576),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_API_DEPRECATED] = ACTIONS(2576),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2576),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym___deprecated_msg] = ACTIONS(2576),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2576),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2576),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2576),
    [anon_sym_ATprotocol] = ACTIONS(2578),
    [anon_sym_ATinterface] = ACTIONS(2578),
    [anon_sym_ATimplementation] = ACTIONS(2578),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2578),
    [anon_sym_ATtry] = ACTIONS(2578),
    [anon_sym___try] = ACTIONS(2576),
    [anon_sym_ATthrow] = ACTIONS(2578),
    [anon_sym_ATselector] = ACTIONS(2578),
    [anon_sym_ATavailable] = ACTIONS(2578),
    [anon_sym___builtin_available] = ACTIONS(2576),
    [anon_sym_va_arg] = ACTIONS(2576),
    [anon_sym_ATencode] = ACTIONS(2578),
    [anon_sym_ATsynchronized] = ACTIONS(2578),
    [anon_sym_BOOL] = ACTIONS(2576),
    [anon_sym_IMP] = ACTIONS(2576),
    [anon_sym_SEL] = ACTIONS(2576),
    [anon_sym_Class] = ACTIONS(2576),
    [anon_sym_id] = ACTIONS(2576),
  },
  [STATE(908)] = {
    [sym_identifier] = ACTIONS(2680),
    [aux_sym_preproc_include_token1] = ACTIONS(2680),
    [aux_sym_preproc_include_token2] = ACTIONS(2680),
    [aux_sym_preproc_def_token1] = ACTIONS(2680),
    [aux_sym_preproc_if_token1] = ACTIONS(2680),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2680),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2680),
    [sym_preproc_directive] = ACTIONS(2680),
    [anon_sym_LPAREN2] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2680),
    [anon_sym_PLUS] = ACTIONS(2680),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym___extension__] = ACTIONS(2680),
    [anon_sym_typedef] = ACTIONS(2680),
    [anon_sym_extern] = ACTIONS(2680),
    [anon_sym___attribute__] = ACTIONS(2680),
    [anon_sym___attribute] = ACTIONS(2680),
    [anon_sym_noreturn] = ACTIONS(2680),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym___declspec] = ACTIONS(2680),
    [anon_sym___cdecl] = ACTIONS(2680),
    [anon_sym___clrcall] = ACTIONS(2680),
    [anon_sym___stdcall] = ACTIONS(2680),
    [anon_sym___fastcall] = ACTIONS(2680),
    [anon_sym___thiscall] = ACTIONS(2680),
    [anon_sym___vectorcall] = ACTIONS(2680),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_signed] = ACTIONS(2680),
    [anon_sym_unsigned] = ACTIONS(2680),
    [anon_sym_long] = ACTIONS(2680),
    [anon_sym_short] = ACTIONS(2680),
    [anon_sym_static] = ACTIONS(2680),
    [anon_sym_ATautoreleasepool] = ACTIONS(2682),
    [anon_sym_auto] = ACTIONS(2680),
    [anon_sym_register] = ACTIONS(2680),
    [anon_sym_inline] = ACTIONS(2680),
    [anon_sym___inline] = ACTIONS(2680),
    [anon_sym___inline__] = ACTIONS(2680),
    [anon_sym___forceinline] = ACTIONS(2680),
    [anon_sym_thread_local] = ACTIONS(2680),
    [anon_sym___thread] = ACTIONS(2680),
    [anon_sym_CG_EXTERN] = ACTIONS(2680),
    [anon_sym_CG_INLINE] = ACTIONS(2680),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2680),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2680),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2680),
    [anon_sym_IBOutlet] = ACTIONS(2680),
    [anon_sym_IBInspectable] = ACTIONS(2680),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2680),
    [anon_sym_NS_INLINE] = ACTIONS(2680),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2680),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2680),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2680),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2680),
    [anon_sym_const] = ACTIONS(2680),
    [anon_sym_constexpr] = ACTIONS(2680),
    [anon_sym_volatile] = ACTIONS(2680),
    [anon_sym_restrict] = ACTIONS(2680),
    [anon_sym___restrict__] = ACTIONS(2680),
    [anon_sym__Atomic] = ACTIONS(2680),
    [anon_sym__Noreturn] = ACTIONS(2680),
    [anon_sym__Nonnull] = ACTIONS(2680),
    [anon_sym_nullable] = ACTIONS(2680),
    [anon_sym__Complex] = ACTIONS(2680),
    [anon_sym__Nullable] = ACTIONS(2680),
    [anon_sym__Nullable_result] = ACTIONS(2680),
    [anon_sym__Null_unspecified] = ACTIONS(2680),
    [anon_sym___autoreleasing] = ACTIONS(2680),
    [anon_sym___block] = ACTIONS(2680),
    [anon_sym___bridge] = ACTIONS(2680),
    [anon_sym___bridge_retained] = ACTIONS(2680),
    [anon_sym___bridge_transfer] = ACTIONS(2680),
    [anon_sym___complex] = ACTIONS(2680),
    [anon_sym___const] = ACTIONS(2680),
    [anon_sym___imag] = ACTIONS(2680),
    [anon_sym___kindof] = ACTIONS(2680),
    [anon_sym___nonnull] = ACTIONS(2680),
    [anon_sym___nullable] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2680),
    [anon_sym___real] = ACTIONS(2680),
    [anon_sym___strong] = ACTIONS(2680),
    [anon_sym___unsafe_unretained] = ACTIONS(2680),
    [anon_sym___unused] = ACTIONS(2680),
    [anon_sym___weak] = ACTIONS(2680),
    [anon_sym_alignas] = ACTIONS(2680),
    [anon_sym__Alignas] = ACTIONS(2680),
    [sym_primitive_type] = ACTIONS(2680),
    [anon_sym_enum] = ACTIONS(2680),
    [anon_sym_struct] = ACTIONS(2680),
    [anon_sym_union] = ACTIONS(2680),
    [anon_sym_if] = ACTIONS(2680),
    [anon_sym_switch] = ACTIONS(2680),
    [anon_sym_case] = ACTIONS(2680),
    [anon_sym_default] = ACTIONS(2680),
    [anon_sym_while] = ACTIONS(2680),
    [anon_sym_do] = ACTIONS(2680),
    [anon_sym_for] = ACTIONS(2680),
    [anon_sym_in] = ACTIONS(2680),
    [anon_sym_return] = ACTIONS(2680),
    [anon_sym_break] = ACTIONS(2680),
    [anon_sym_continue] = ACTIONS(2680),
    [anon_sym_goto] = ACTIONS(2680),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_sizeof] = ACTIONS(2680),
    [anon_sym___alignof__] = ACTIONS(2680),
    [anon_sym___alignof] = ACTIONS(2680),
    [anon_sym__alignof] = ACTIONS(2680),
    [anon_sym_alignof] = ACTIONS(2680),
    [anon_sym__Alignof] = ACTIONS(2680),
    [anon_sym_offsetof] = ACTIONS(2680),
    [anon_sym__Generic] = ACTIONS(2680),
    [anon_sym_asm] = ACTIONS(2680),
    [anon_sym___asm__] = ACTIONS(2680),
    [anon_sym___asm] = ACTIONS(2680),
    [sym_number_literal] = ACTIONS(2682),
    [anon_sym_L_SQUOTE] = ACTIONS(2682),
    [anon_sym_u_SQUOTE] = ACTIONS(2682),
    [anon_sym_U_SQUOTE] = ACTIONS(2682),
    [anon_sym_u8_SQUOTE] = ACTIONS(2682),
    [anon_sym_SQUOTE] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2680),
    [anon_sym_DQUOTE] = ACTIONS(2682),
    [anon_sym_L_DQUOTE] = ACTIONS(2682),
    [anon_sym_u_DQUOTE] = ACTIONS(2682),
    [anon_sym_U_DQUOTE] = ACTIONS(2682),
    [anon_sym_u8_DQUOTE] = ACTIONS(2682),
    [sym_true] = ACTIONS(2680),
    [sym_false] = ACTIONS(2680),
    [anon_sym_NULL] = ACTIONS(2680),
    [anon_sym_nullptr] = ACTIONS(2680),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2680),
    [anon_sym___typeof] = ACTIONS(2680),
    [anon_sym_typeof] = ACTIONS(2680),
    [anon_sym_ATimport] = ACTIONS(2682),
    [aux_sym_preproc_undef_token1] = ACTIONS(2680),
    [anon_sym_POUND] = ACTIONS(2680),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2680),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2680),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2680),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2680),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2680),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2680),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2680),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_API_AVAILABLE] = ACTIONS(2680),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_API_DEPRECATED] = ACTIONS(2680),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2680),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym___deprecated_msg] = ACTIONS(2680),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2680),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2680),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2680),
    [anon_sym_ATprotocol] = ACTIONS(2682),
    [anon_sym_ATinterface] = ACTIONS(2682),
    [anon_sym_ATimplementation] = ACTIONS(2682),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2682),
    [anon_sym_ATtry] = ACTIONS(2682),
    [anon_sym___try] = ACTIONS(2680),
    [anon_sym_ATthrow] = ACTIONS(2682),
    [anon_sym_ATselector] = ACTIONS(2682),
    [anon_sym_ATavailable] = ACTIONS(2682),
    [anon_sym___builtin_available] = ACTIONS(2680),
    [anon_sym_va_arg] = ACTIONS(2680),
    [anon_sym_ATencode] = ACTIONS(2682),
    [anon_sym_ATsynchronized] = ACTIONS(2682),
    [anon_sym_BOOL] = ACTIONS(2680),
    [anon_sym_IMP] = ACTIONS(2680),
    [anon_sym_SEL] = ACTIONS(2680),
    [anon_sym_Class] = ACTIONS(2680),
    [anon_sym_id] = ACTIONS(2680),
  },
  [STATE(909)] = {
    [ts_builtin_sym_end] = ACTIONS(2352),
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_include_token1] = ACTIONS(2350),
    [aux_sym_preproc_include_token2] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [anon_sym_COMMA] = ACTIONS(2352),
    [anon_sym_RPAREN] = ACTIONS(2352),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_case] = ACTIONS(2350),
    [anon_sym_default] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_ATimport] = ACTIONS(2352),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_POUND] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATprotocol] = ACTIONS(2352),
    [anon_sym_ATinterface] = ACTIONS(2352),
    [anon_sym_ATimplementation] = ACTIONS(2352),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(910)] = {
    [ts_builtin_sym_end] = ACTIONS(2348),
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_include_token1] = ACTIONS(2346),
    [aux_sym_preproc_include_token2] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [anon_sym_COMMA] = ACTIONS(2348),
    [anon_sym_RPAREN] = ACTIONS(2348),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_case] = ACTIONS(2346),
    [anon_sym_default] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_ATimport] = ACTIONS(2348),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_POUND] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATprotocol] = ACTIONS(2348),
    [anon_sym_ATinterface] = ACTIONS(2348),
    [anon_sym_ATimplementation] = ACTIONS(2348),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(911)] = {
    [sym_ms_declspec_modifier] = STATE(6585),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2912),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATprotocol] = ACTIONS(2920),
    [anon_sym_ATinterface] = ACTIONS(2920),
    [anon_sym_ATimplementation] = ACTIONS(2920),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(912)] = {
    [sym_ms_declspec_modifier] = STATE(6393),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2922),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATprotocol] = ACTIONS(2920),
    [anon_sym_ATinterface] = ACTIONS(2920),
    [anon_sym_ATimplementation] = ACTIONS(2920),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(913)] = {
    [ts_builtin_sym_end] = ACTIONS(2362),
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_include_token1] = ACTIONS(2360),
    [aux_sym_preproc_include_token2] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [anon_sym_COMMA] = ACTIONS(2362),
    [anon_sym_RPAREN] = ACTIONS(2362),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_case] = ACTIONS(2360),
    [anon_sym_default] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_ATimport] = ACTIONS(2362),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATprotocol] = ACTIONS(2362),
    [anon_sym_ATinterface] = ACTIONS(2362),
    [anon_sym_ATimplementation] = ACTIONS(2362),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(914)] = {
    [sym_ms_declspec_modifier] = STATE(6735),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2924),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATprotocol] = ACTIONS(2920),
    [anon_sym_ATinterface] = ACTIONS(2920),
    [anon_sym_ATimplementation] = ACTIONS(2920),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(915)] = {
    [sym_ms_declspec_modifier] = STATE(6760),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2926),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATprotocol] = ACTIONS(2920),
    [anon_sym_ATinterface] = ACTIONS(2920),
    [anon_sym_ATimplementation] = ACTIONS(2920),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(916)] = {
    [sym_ms_declspec_modifier] = STATE(6753),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2928),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATprotocol] = ACTIONS(2920),
    [anon_sym_ATinterface] = ACTIONS(2920),
    [anon_sym_ATimplementation] = ACTIONS(2920),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(917)] = {
    [sym_else_clause] = STATE(947),
    [ts_builtin_sym_end] = ACTIONS(2396),
    [sym_identifier] = ACTIONS(2394),
    [aux_sym_preproc_include_token1] = ACTIONS(2394),
    [aux_sym_preproc_include_token2] = ACTIONS(2394),
    [aux_sym_preproc_def_token1] = ACTIONS(2394),
    [aux_sym_preproc_if_token1] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2394),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2394),
    [sym_preproc_directive] = ACTIONS(2394),
    [anon_sym_LPAREN2] = ACTIONS(2396),
    [anon_sym_BANG] = ACTIONS(2396),
    [anon_sym_TILDE] = ACTIONS(2396),
    [anon_sym_DASH] = ACTIONS(2394),
    [anon_sym_PLUS] = ACTIONS(2394),
    [anon_sym_STAR] = ACTIONS(2396),
    [anon_sym_CARET] = ACTIONS(2396),
    [anon_sym_AMP] = ACTIONS(2396),
    [anon_sym_SEMI] = ACTIONS(2396),
    [anon_sym___extension__] = ACTIONS(2394),
    [anon_sym_typedef] = ACTIONS(2394),
    [anon_sym_extern] = ACTIONS(2394),
    [anon_sym___attribute__] = ACTIONS(2394),
    [anon_sym___attribute] = ACTIONS(2394),
    [anon_sym_noreturn] = ACTIONS(2394),
    [anon_sym_LBRACK] = ACTIONS(2396),
    [anon_sym___declspec] = ACTIONS(2394),
    [anon_sym___cdecl] = ACTIONS(2394),
    [anon_sym___clrcall] = ACTIONS(2394),
    [anon_sym___stdcall] = ACTIONS(2394),
    [anon_sym___fastcall] = ACTIONS(2394),
    [anon_sym___thiscall] = ACTIONS(2394),
    [anon_sym___vectorcall] = ACTIONS(2394),
    [anon_sym_LBRACE] = ACTIONS(2396),
    [anon_sym_signed] = ACTIONS(2394),
    [anon_sym_unsigned] = ACTIONS(2394),
    [anon_sym_long] = ACTIONS(2394),
    [anon_sym_short] = ACTIONS(2394),
    [anon_sym_static] = ACTIONS(2394),
    [anon_sym_ATautoreleasepool] = ACTIONS(2396),
    [anon_sym_auto] = ACTIONS(2394),
    [anon_sym_register] = ACTIONS(2394),
    [anon_sym_inline] = ACTIONS(2394),
    [anon_sym___inline] = ACTIONS(2394),
    [anon_sym___inline__] = ACTIONS(2394),
    [anon_sym___forceinline] = ACTIONS(2394),
    [anon_sym_thread_local] = ACTIONS(2394),
    [anon_sym___thread] = ACTIONS(2394),
    [anon_sym_CG_EXTERN] = ACTIONS(2394),
    [anon_sym_CG_INLINE] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2394),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2394),
    [anon_sym_IBOutlet] = ACTIONS(2394),
    [anon_sym_IBInspectable] = ACTIONS(2394),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2394),
    [anon_sym_NS_INLINE] = ACTIONS(2394),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2394),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2394),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2394),
    [anon_sym_const] = ACTIONS(2394),
    [anon_sym_constexpr] = ACTIONS(2394),
    [anon_sym_volatile] = ACTIONS(2394),
    [anon_sym_restrict] = ACTIONS(2394),
    [anon_sym___restrict__] = ACTIONS(2394),
    [anon_sym__Atomic] = ACTIONS(2394),
    [anon_sym__Noreturn] = ACTIONS(2394),
    [anon_sym__Nonnull] = ACTIONS(2394),
    [anon_sym_nullable] = ACTIONS(2394),
    [anon_sym__Complex] = ACTIONS(2394),
    [anon_sym__Nullable] = ACTIONS(2394),
    [anon_sym__Nullable_result] = ACTIONS(2394),
    [anon_sym__Null_unspecified] = ACTIONS(2394),
    [anon_sym___autoreleasing] = ACTIONS(2394),
    [anon_sym___block] = ACTIONS(2394),
    [anon_sym___bridge] = ACTIONS(2394),
    [anon_sym___bridge_retained] = ACTIONS(2394),
    [anon_sym___bridge_transfer] = ACTIONS(2394),
    [anon_sym___complex] = ACTIONS(2394),
    [anon_sym___const] = ACTIONS(2394),
    [anon_sym___imag] = ACTIONS(2394),
    [anon_sym___kindof] = ACTIONS(2394),
    [anon_sym___nonnull] = ACTIONS(2394),
    [anon_sym___nullable] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2394),
    [anon_sym___real] = ACTIONS(2394),
    [anon_sym___strong] = ACTIONS(2394),
    [anon_sym___unsafe_unretained] = ACTIONS(2394),
    [anon_sym___unused] = ACTIONS(2394),
    [anon_sym___weak] = ACTIONS(2394),
    [anon_sym_alignas] = ACTIONS(2394),
    [anon_sym__Alignas] = ACTIONS(2394),
    [sym_primitive_type] = ACTIONS(2394),
    [anon_sym_enum] = ACTIONS(2394),
    [anon_sym_struct] = ACTIONS(2394),
    [anon_sym_union] = ACTIONS(2394),
    [anon_sym_if] = ACTIONS(2394),
    [anon_sym_else] = ACTIONS(2930),
    [anon_sym_switch] = ACTIONS(2394),
    [anon_sym_case] = ACTIONS(2394),
    [anon_sym_default] = ACTIONS(2394),
    [anon_sym_while] = ACTIONS(2394),
    [anon_sym_do] = ACTIONS(2394),
    [anon_sym_for] = ACTIONS(2394),
    [anon_sym_in] = ACTIONS(2394),
    [anon_sym_return] = ACTIONS(2394),
    [anon_sym_break] = ACTIONS(2394),
    [anon_sym_continue] = ACTIONS(2394),
    [anon_sym_goto] = ACTIONS(2394),
    [anon_sym_DASH_DASH] = ACTIONS(2396),
    [anon_sym_PLUS_PLUS] = ACTIONS(2396),
    [anon_sym_sizeof] = ACTIONS(2394),
    [anon_sym___alignof__] = ACTIONS(2394),
    [anon_sym___alignof] = ACTIONS(2394),
    [anon_sym__alignof] = ACTIONS(2394),
    [anon_sym_alignof] = ACTIONS(2394),
    [anon_sym__Alignof] = ACTIONS(2394),
    [anon_sym_offsetof] = ACTIONS(2394),
    [anon_sym__Generic] = ACTIONS(2394),
    [anon_sym_asm] = ACTIONS(2394),
    [anon_sym___asm__] = ACTIONS(2394),
    [anon_sym___asm] = ACTIONS(2394),
    [sym_number_literal] = ACTIONS(2396),
    [anon_sym_L_SQUOTE] = ACTIONS(2396),
    [anon_sym_u_SQUOTE] = ACTIONS(2396),
    [anon_sym_U_SQUOTE] = ACTIONS(2396),
    [anon_sym_u8_SQUOTE] = ACTIONS(2396),
    [anon_sym_SQUOTE] = ACTIONS(2396),
    [anon_sym_AT] = ACTIONS(2394),
    [anon_sym_DQUOTE] = ACTIONS(2396),
    [anon_sym_L_DQUOTE] = ACTIONS(2396),
    [anon_sym_u_DQUOTE] = ACTIONS(2396),
    [anon_sym_U_DQUOTE] = ACTIONS(2396),
    [anon_sym_u8_DQUOTE] = ACTIONS(2396),
    [sym_true] = ACTIONS(2394),
    [sym_false] = ACTIONS(2394),
    [anon_sym_NULL] = ACTIONS(2394),
    [anon_sym_nullptr] = ACTIONS(2394),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2394),
    [anon_sym___typeof] = ACTIONS(2394),
    [anon_sym_typeof] = ACTIONS(2394),
    [anon_sym_ATimport] = ACTIONS(2396),
    [aux_sym_preproc_undef_token1] = ACTIONS(2394),
    [anon_sym_POUND] = ACTIONS(2394),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2394),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2394),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2394),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_API_AVAILABLE] = ACTIONS(2394),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_API_DEPRECATED] = ACTIONS(2394),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym___deprecated_msg] = ACTIONS(2394),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2394),
    [anon_sym_ATprotocol] = ACTIONS(2396),
    [anon_sym_ATinterface] = ACTIONS(2396),
    [anon_sym_ATimplementation] = ACTIONS(2396),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2396),
    [anon_sym_ATselector] = ACTIONS(2396),
    [anon_sym_ATavailable] = ACTIONS(2396),
    [anon_sym___builtin_available] = ACTIONS(2394),
    [anon_sym_va_arg] = ACTIONS(2394),
    [anon_sym_ATencode] = ACTIONS(2396),
    [anon_sym_BOOL] = ACTIONS(2394),
    [anon_sym_IMP] = ACTIONS(2394),
    [anon_sym_SEL] = ACTIONS(2394),
    [anon_sym_Class] = ACTIONS(2394),
    [anon_sym_id] = ACTIONS(2394),
  },
  [STATE(918)] = {
    [ts_builtin_sym_end] = ACTIONS(2444),
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(919)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(2914),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATprotocol] = ACTIONS(2920),
    [anon_sym_ATinterface] = ACTIONS(2920),
    [anon_sym_ATimplementation] = ACTIONS(2920),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(920)] = {
    [ts_builtin_sym_end] = ACTIONS(2548),
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(921)] = {
    [ts_builtin_sym_end] = ACTIONS(2548),
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_include_token1] = ACTIONS(2546),
    [aux_sym_preproc_include_token2] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_case] = ACTIONS(2546),
    [anon_sym_default] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_ATimport] = ACTIONS(2548),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_POUND] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATprotocol] = ACTIONS(2548),
    [anon_sym_ATinterface] = ACTIONS(2548),
    [anon_sym_ATimplementation] = ACTIONS(2548),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(922)] = {
    [ts_builtin_sym_end] = ACTIONS(2464),
    [sym_identifier] = ACTIONS(2462),
    [aux_sym_preproc_include_token1] = ACTIONS(2462),
    [aux_sym_preproc_include_token2] = ACTIONS(2462),
    [aux_sym_preproc_def_token1] = ACTIONS(2462),
    [aux_sym_preproc_if_token1] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2462),
    [sym_preproc_directive] = ACTIONS(2462),
    [anon_sym_LPAREN2] = ACTIONS(2464),
    [anon_sym_BANG] = ACTIONS(2464),
    [anon_sym_TILDE] = ACTIONS(2464),
    [anon_sym_DASH] = ACTIONS(2462),
    [anon_sym_PLUS] = ACTIONS(2462),
    [anon_sym_STAR] = ACTIONS(2464),
    [anon_sym_CARET] = ACTIONS(2464),
    [anon_sym_AMP] = ACTIONS(2464),
    [anon_sym_SEMI] = ACTIONS(2464),
    [anon_sym___extension__] = ACTIONS(2462),
    [anon_sym_typedef] = ACTIONS(2462),
    [anon_sym_extern] = ACTIONS(2462),
    [anon_sym___attribute__] = ACTIONS(2462),
    [anon_sym___attribute] = ACTIONS(2462),
    [anon_sym_noreturn] = ACTIONS(2462),
    [anon_sym_LBRACK] = ACTIONS(2464),
    [anon_sym___declspec] = ACTIONS(2462),
    [anon_sym___cdecl] = ACTIONS(2462),
    [anon_sym___clrcall] = ACTIONS(2462),
    [anon_sym___stdcall] = ACTIONS(2462),
    [anon_sym___fastcall] = ACTIONS(2462),
    [anon_sym___thiscall] = ACTIONS(2462),
    [anon_sym___vectorcall] = ACTIONS(2462),
    [anon_sym_LBRACE] = ACTIONS(2464),
    [anon_sym_signed] = ACTIONS(2462),
    [anon_sym_unsigned] = ACTIONS(2462),
    [anon_sym_long] = ACTIONS(2462),
    [anon_sym_short] = ACTIONS(2462),
    [anon_sym_static] = ACTIONS(2462),
    [anon_sym_ATautoreleasepool] = ACTIONS(2464),
    [anon_sym_auto] = ACTIONS(2462),
    [anon_sym_register] = ACTIONS(2462),
    [anon_sym_inline] = ACTIONS(2462),
    [anon_sym___inline] = ACTIONS(2462),
    [anon_sym___inline__] = ACTIONS(2462),
    [anon_sym___forceinline] = ACTIONS(2462),
    [anon_sym_thread_local] = ACTIONS(2462),
    [anon_sym___thread] = ACTIONS(2462),
    [anon_sym_CG_EXTERN] = ACTIONS(2462),
    [anon_sym_CG_INLINE] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2462),
    [anon_sym_IBOutlet] = ACTIONS(2462),
    [anon_sym_IBInspectable] = ACTIONS(2462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2462),
    [anon_sym_NS_INLINE] = ACTIONS(2462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2462),
    [anon_sym_const] = ACTIONS(2462),
    [anon_sym_constexpr] = ACTIONS(2462),
    [anon_sym_volatile] = ACTIONS(2462),
    [anon_sym_restrict] = ACTIONS(2462),
    [anon_sym___restrict__] = ACTIONS(2462),
    [anon_sym__Atomic] = ACTIONS(2462),
    [anon_sym__Noreturn] = ACTIONS(2462),
    [anon_sym__Nonnull] = ACTIONS(2462),
    [anon_sym_nullable] = ACTIONS(2462),
    [anon_sym__Complex] = ACTIONS(2462),
    [anon_sym__Nullable] = ACTIONS(2462),
    [anon_sym__Nullable_result] = ACTIONS(2462),
    [anon_sym__Null_unspecified] = ACTIONS(2462),
    [anon_sym___autoreleasing] = ACTIONS(2462),
    [anon_sym___block] = ACTIONS(2462),
    [anon_sym___bridge] = ACTIONS(2462),
    [anon_sym___bridge_retained] = ACTIONS(2462),
    [anon_sym___bridge_transfer] = ACTIONS(2462),
    [anon_sym___complex] = ACTIONS(2462),
    [anon_sym___const] = ACTIONS(2462),
    [anon_sym___imag] = ACTIONS(2462),
    [anon_sym___kindof] = ACTIONS(2462),
    [anon_sym___nonnull] = ACTIONS(2462),
    [anon_sym___nullable] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2462),
    [anon_sym___real] = ACTIONS(2462),
    [anon_sym___strong] = ACTIONS(2462),
    [anon_sym___unsafe_unretained] = ACTIONS(2462),
    [anon_sym___unused] = ACTIONS(2462),
    [anon_sym___weak] = ACTIONS(2462),
    [anon_sym_alignas] = ACTIONS(2462),
    [anon_sym__Alignas] = ACTIONS(2462),
    [sym_primitive_type] = ACTIONS(2462),
    [anon_sym_enum] = ACTIONS(2462),
    [anon_sym_struct] = ACTIONS(2462),
    [anon_sym_union] = ACTIONS(2462),
    [anon_sym_if] = ACTIONS(2462),
    [anon_sym_else] = ACTIONS(2462),
    [anon_sym_switch] = ACTIONS(2462),
    [anon_sym_case] = ACTIONS(2462),
    [anon_sym_default] = ACTIONS(2462),
    [anon_sym_while] = ACTIONS(2462),
    [anon_sym_do] = ACTIONS(2462),
    [anon_sym_for] = ACTIONS(2462),
    [anon_sym_in] = ACTIONS(2462),
    [anon_sym_return] = ACTIONS(2462),
    [anon_sym_break] = ACTIONS(2462),
    [anon_sym_continue] = ACTIONS(2462),
    [anon_sym_goto] = ACTIONS(2462),
    [anon_sym_DASH_DASH] = ACTIONS(2464),
    [anon_sym_PLUS_PLUS] = ACTIONS(2464),
    [anon_sym_sizeof] = ACTIONS(2462),
    [anon_sym___alignof__] = ACTIONS(2462),
    [anon_sym___alignof] = ACTIONS(2462),
    [anon_sym__alignof] = ACTIONS(2462),
    [anon_sym_alignof] = ACTIONS(2462),
    [anon_sym__Alignof] = ACTIONS(2462),
    [anon_sym_offsetof] = ACTIONS(2462),
    [anon_sym__Generic] = ACTIONS(2462),
    [anon_sym_asm] = ACTIONS(2462),
    [anon_sym___asm__] = ACTIONS(2462),
    [anon_sym___asm] = ACTIONS(2462),
    [sym_number_literal] = ACTIONS(2464),
    [anon_sym_L_SQUOTE] = ACTIONS(2464),
    [anon_sym_u_SQUOTE] = ACTIONS(2464),
    [anon_sym_U_SQUOTE] = ACTIONS(2464),
    [anon_sym_u8_SQUOTE] = ACTIONS(2464),
    [anon_sym_SQUOTE] = ACTIONS(2464),
    [anon_sym_AT] = ACTIONS(2462),
    [anon_sym_DQUOTE] = ACTIONS(2464),
    [anon_sym_L_DQUOTE] = ACTIONS(2464),
    [anon_sym_u_DQUOTE] = ACTIONS(2464),
    [anon_sym_U_DQUOTE] = ACTIONS(2464),
    [anon_sym_u8_DQUOTE] = ACTIONS(2464),
    [sym_true] = ACTIONS(2462),
    [sym_false] = ACTIONS(2462),
    [anon_sym_NULL] = ACTIONS(2462),
    [anon_sym_nullptr] = ACTIONS(2462),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2462),
    [anon_sym___typeof] = ACTIONS(2462),
    [anon_sym_typeof] = ACTIONS(2462),
    [anon_sym_ATimport] = ACTIONS(2464),
    [aux_sym_preproc_undef_token1] = ACTIONS(2462),
    [anon_sym_POUND] = ACTIONS(2462),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2462),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_API_AVAILABLE] = ACTIONS(2462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_API_DEPRECATED] = ACTIONS(2462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym___deprecated_msg] = ACTIONS(2462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2462),
    [anon_sym_ATprotocol] = ACTIONS(2464),
    [anon_sym_ATinterface] = ACTIONS(2464),
    [anon_sym_ATimplementation] = ACTIONS(2464),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2464),
    [anon_sym_ATselector] = ACTIONS(2464),
    [anon_sym_ATavailable] = ACTIONS(2464),
    [anon_sym___builtin_available] = ACTIONS(2462),
    [anon_sym_va_arg] = ACTIONS(2462),
    [anon_sym_ATencode] = ACTIONS(2464),
    [anon_sym_BOOL] = ACTIONS(2462),
    [anon_sym_IMP] = ACTIONS(2462),
    [anon_sym_SEL] = ACTIONS(2462),
    [anon_sym_Class] = ACTIONS(2462),
    [anon_sym_id] = ACTIONS(2462),
  },
  [STATE(923)] = {
    [ts_builtin_sym_end] = ACTIONS(2528),
    [sym_identifier] = ACTIONS(2526),
    [aux_sym_preproc_include_token1] = ACTIONS(2526),
    [aux_sym_preproc_include_token2] = ACTIONS(2526),
    [aux_sym_preproc_def_token1] = ACTIONS(2526),
    [aux_sym_preproc_if_token1] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2526),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2526),
    [sym_preproc_directive] = ACTIONS(2526),
    [anon_sym_LPAREN2] = ACTIONS(2528),
    [anon_sym_BANG] = ACTIONS(2528),
    [anon_sym_TILDE] = ACTIONS(2528),
    [anon_sym_DASH] = ACTIONS(2526),
    [anon_sym_PLUS] = ACTIONS(2526),
    [anon_sym_STAR] = ACTIONS(2528),
    [anon_sym_CARET] = ACTIONS(2528),
    [anon_sym_AMP] = ACTIONS(2528),
    [anon_sym_SEMI] = ACTIONS(2528),
    [anon_sym___extension__] = ACTIONS(2526),
    [anon_sym_typedef] = ACTIONS(2526),
    [anon_sym_extern] = ACTIONS(2526),
    [anon_sym___attribute__] = ACTIONS(2526),
    [anon_sym___attribute] = ACTIONS(2526),
    [anon_sym_noreturn] = ACTIONS(2526),
    [anon_sym_LBRACK] = ACTIONS(2528),
    [anon_sym___declspec] = ACTIONS(2526),
    [anon_sym___cdecl] = ACTIONS(2526),
    [anon_sym___clrcall] = ACTIONS(2526),
    [anon_sym___stdcall] = ACTIONS(2526),
    [anon_sym___fastcall] = ACTIONS(2526),
    [anon_sym___thiscall] = ACTIONS(2526),
    [anon_sym___vectorcall] = ACTIONS(2526),
    [anon_sym_LBRACE] = ACTIONS(2528),
    [anon_sym_signed] = ACTIONS(2526),
    [anon_sym_unsigned] = ACTIONS(2526),
    [anon_sym_long] = ACTIONS(2526),
    [anon_sym_short] = ACTIONS(2526),
    [anon_sym_static] = ACTIONS(2526),
    [anon_sym_ATautoreleasepool] = ACTIONS(2528),
    [anon_sym_auto] = ACTIONS(2526),
    [anon_sym_register] = ACTIONS(2526),
    [anon_sym_inline] = ACTIONS(2526),
    [anon_sym___inline] = ACTIONS(2526),
    [anon_sym___inline__] = ACTIONS(2526),
    [anon_sym___forceinline] = ACTIONS(2526),
    [anon_sym_thread_local] = ACTIONS(2526),
    [anon_sym___thread] = ACTIONS(2526),
    [anon_sym_CG_EXTERN] = ACTIONS(2526),
    [anon_sym_CG_INLINE] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2526),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2526),
    [anon_sym_IBOutlet] = ACTIONS(2526),
    [anon_sym_IBInspectable] = ACTIONS(2526),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2526),
    [anon_sym_NS_INLINE] = ACTIONS(2526),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2526),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2526),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2526),
    [anon_sym_const] = ACTIONS(2526),
    [anon_sym_constexpr] = ACTIONS(2526),
    [anon_sym_volatile] = ACTIONS(2526),
    [anon_sym_restrict] = ACTIONS(2526),
    [anon_sym___restrict__] = ACTIONS(2526),
    [anon_sym__Atomic] = ACTIONS(2526),
    [anon_sym__Noreturn] = ACTIONS(2526),
    [anon_sym__Nonnull] = ACTIONS(2526),
    [anon_sym_nullable] = ACTIONS(2526),
    [anon_sym__Complex] = ACTIONS(2526),
    [anon_sym__Nullable] = ACTIONS(2526),
    [anon_sym__Nullable_result] = ACTIONS(2526),
    [anon_sym__Null_unspecified] = ACTIONS(2526),
    [anon_sym___autoreleasing] = ACTIONS(2526),
    [anon_sym___block] = ACTIONS(2526),
    [anon_sym___bridge] = ACTIONS(2526),
    [anon_sym___bridge_retained] = ACTIONS(2526),
    [anon_sym___bridge_transfer] = ACTIONS(2526),
    [anon_sym___complex] = ACTIONS(2526),
    [anon_sym___const] = ACTIONS(2526),
    [anon_sym___imag] = ACTIONS(2526),
    [anon_sym___kindof] = ACTIONS(2526),
    [anon_sym___nonnull] = ACTIONS(2526),
    [anon_sym___nullable] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2526),
    [anon_sym___real] = ACTIONS(2526),
    [anon_sym___strong] = ACTIONS(2526),
    [anon_sym___unsafe_unretained] = ACTIONS(2526),
    [anon_sym___unused] = ACTIONS(2526),
    [anon_sym___weak] = ACTIONS(2526),
    [anon_sym_alignas] = ACTIONS(2526),
    [anon_sym__Alignas] = ACTIONS(2526),
    [sym_primitive_type] = ACTIONS(2526),
    [anon_sym_enum] = ACTIONS(2526),
    [anon_sym_struct] = ACTIONS(2526),
    [anon_sym_union] = ACTIONS(2526),
    [anon_sym_if] = ACTIONS(2526),
    [anon_sym_else] = ACTIONS(2526),
    [anon_sym_switch] = ACTIONS(2526),
    [anon_sym_case] = ACTIONS(2526),
    [anon_sym_default] = ACTIONS(2526),
    [anon_sym_while] = ACTIONS(2526),
    [anon_sym_do] = ACTIONS(2526),
    [anon_sym_for] = ACTIONS(2526),
    [anon_sym_in] = ACTIONS(2526),
    [anon_sym_return] = ACTIONS(2526),
    [anon_sym_break] = ACTIONS(2526),
    [anon_sym_continue] = ACTIONS(2526),
    [anon_sym_goto] = ACTIONS(2526),
    [anon_sym_DASH_DASH] = ACTIONS(2528),
    [anon_sym_PLUS_PLUS] = ACTIONS(2528),
    [anon_sym_sizeof] = ACTIONS(2526),
    [anon_sym___alignof__] = ACTIONS(2526),
    [anon_sym___alignof] = ACTIONS(2526),
    [anon_sym__alignof] = ACTIONS(2526),
    [anon_sym_alignof] = ACTIONS(2526),
    [anon_sym__Alignof] = ACTIONS(2526),
    [anon_sym_offsetof] = ACTIONS(2526),
    [anon_sym__Generic] = ACTIONS(2526),
    [anon_sym_asm] = ACTIONS(2526),
    [anon_sym___asm__] = ACTIONS(2526),
    [anon_sym___asm] = ACTIONS(2526),
    [sym_number_literal] = ACTIONS(2528),
    [anon_sym_L_SQUOTE] = ACTIONS(2528),
    [anon_sym_u_SQUOTE] = ACTIONS(2528),
    [anon_sym_U_SQUOTE] = ACTIONS(2528),
    [anon_sym_u8_SQUOTE] = ACTIONS(2528),
    [anon_sym_SQUOTE] = ACTIONS(2528),
    [anon_sym_AT] = ACTIONS(2526),
    [anon_sym_DQUOTE] = ACTIONS(2528),
    [anon_sym_L_DQUOTE] = ACTIONS(2528),
    [anon_sym_u_DQUOTE] = ACTIONS(2528),
    [anon_sym_U_DQUOTE] = ACTIONS(2528),
    [anon_sym_u8_DQUOTE] = ACTIONS(2528),
    [sym_true] = ACTIONS(2526),
    [sym_false] = ACTIONS(2526),
    [anon_sym_NULL] = ACTIONS(2526),
    [anon_sym_nullptr] = ACTIONS(2526),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2526),
    [anon_sym___typeof] = ACTIONS(2526),
    [anon_sym_typeof] = ACTIONS(2526),
    [anon_sym_ATimport] = ACTIONS(2528),
    [aux_sym_preproc_undef_token1] = ACTIONS(2526),
    [anon_sym_POUND] = ACTIONS(2526),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2526),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2526),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2526),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_API_AVAILABLE] = ACTIONS(2526),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_API_DEPRECATED] = ACTIONS(2526),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym___deprecated_msg] = ACTIONS(2526),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2526),
    [anon_sym_ATprotocol] = ACTIONS(2528),
    [anon_sym_ATinterface] = ACTIONS(2528),
    [anon_sym_ATimplementation] = ACTIONS(2528),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2528),
    [anon_sym_ATselector] = ACTIONS(2528),
    [anon_sym_ATavailable] = ACTIONS(2528),
    [anon_sym___builtin_available] = ACTIONS(2526),
    [anon_sym_va_arg] = ACTIONS(2526),
    [anon_sym_ATencode] = ACTIONS(2528),
    [anon_sym_BOOL] = ACTIONS(2526),
    [anon_sym_IMP] = ACTIONS(2526),
    [anon_sym_SEL] = ACTIONS(2526),
    [anon_sym_Class] = ACTIONS(2526),
    [anon_sym_id] = ACTIONS(2526),
  },
  [STATE(924)] = {
    [ts_builtin_sym_end] = ACTIONS(2532),
    [sym_identifier] = ACTIONS(2530),
    [aux_sym_preproc_include_token1] = ACTIONS(2530),
    [aux_sym_preproc_include_token2] = ACTIONS(2530),
    [aux_sym_preproc_def_token1] = ACTIONS(2530),
    [aux_sym_preproc_if_token1] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2530),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2530),
    [sym_preproc_directive] = ACTIONS(2530),
    [anon_sym_LPAREN2] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2530),
    [anon_sym_PLUS] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym___extension__] = ACTIONS(2530),
    [anon_sym_typedef] = ACTIONS(2530),
    [anon_sym_extern] = ACTIONS(2530),
    [anon_sym___attribute__] = ACTIONS(2530),
    [anon_sym___attribute] = ACTIONS(2530),
    [anon_sym_noreturn] = ACTIONS(2530),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym___declspec] = ACTIONS(2530),
    [anon_sym___cdecl] = ACTIONS(2530),
    [anon_sym___clrcall] = ACTIONS(2530),
    [anon_sym___stdcall] = ACTIONS(2530),
    [anon_sym___fastcall] = ACTIONS(2530),
    [anon_sym___thiscall] = ACTIONS(2530),
    [anon_sym___vectorcall] = ACTIONS(2530),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_signed] = ACTIONS(2530),
    [anon_sym_unsigned] = ACTIONS(2530),
    [anon_sym_long] = ACTIONS(2530),
    [anon_sym_short] = ACTIONS(2530),
    [anon_sym_static] = ACTIONS(2530),
    [anon_sym_ATautoreleasepool] = ACTIONS(2532),
    [anon_sym_auto] = ACTIONS(2530),
    [anon_sym_register] = ACTIONS(2530),
    [anon_sym_inline] = ACTIONS(2530),
    [anon_sym___inline] = ACTIONS(2530),
    [anon_sym___inline__] = ACTIONS(2530),
    [anon_sym___forceinline] = ACTIONS(2530),
    [anon_sym_thread_local] = ACTIONS(2530),
    [anon_sym___thread] = ACTIONS(2530),
    [anon_sym_CG_EXTERN] = ACTIONS(2530),
    [anon_sym_CG_INLINE] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2530),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2530),
    [anon_sym_IBOutlet] = ACTIONS(2530),
    [anon_sym_IBInspectable] = ACTIONS(2530),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2530),
    [anon_sym_NS_INLINE] = ACTIONS(2530),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2530),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2530),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2530),
    [anon_sym_const] = ACTIONS(2530),
    [anon_sym_constexpr] = ACTIONS(2530),
    [anon_sym_volatile] = ACTIONS(2530),
    [anon_sym_restrict] = ACTIONS(2530),
    [anon_sym___restrict__] = ACTIONS(2530),
    [anon_sym__Atomic] = ACTIONS(2530),
    [anon_sym__Noreturn] = ACTIONS(2530),
    [anon_sym__Nonnull] = ACTIONS(2530),
    [anon_sym_nullable] = ACTIONS(2530),
    [anon_sym__Complex] = ACTIONS(2530),
    [anon_sym__Nullable] = ACTIONS(2530),
    [anon_sym__Nullable_result] = ACTIONS(2530),
    [anon_sym__Null_unspecified] = ACTIONS(2530),
    [anon_sym___autoreleasing] = ACTIONS(2530),
    [anon_sym___block] = ACTIONS(2530),
    [anon_sym___bridge] = ACTIONS(2530),
    [anon_sym___bridge_retained] = ACTIONS(2530),
    [anon_sym___bridge_transfer] = ACTIONS(2530),
    [anon_sym___complex] = ACTIONS(2530),
    [anon_sym___const] = ACTIONS(2530),
    [anon_sym___imag] = ACTIONS(2530),
    [anon_sym___kindof] = ACTIONS(2530),
    [anon_sym___nonnull] = ACTIONS(2530),
    [anon_sym___nullable] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2530),
    [anon_sym___real] = ACTIONS(2530),
    [anon_sym___strong] = ACTIONS(2530),
    [anon_sym___unsafe_unretained] = ACTIONS(2530),
    [anon_sym___unused] = ACTIONS(2530),
    [anon_sym___weak] = ACTIONS(2530),
    [anon_sym_alignas] = ACTIONS(2530),
    [anon_sym__Alignas] = ACTIONS(2530),
    [sym_primitive_type] = ACTIONS(2530),
    [anon_sym_enum] = ACTIONS(2530),
    [anon_sym_struct] = ACTIONS(2530),
    [anon_sym_union] = ACTIONS(2530),
    [anon_sym_if] = ACTIONS(2530),
    [anon_sym_else] = ACTIONS(2530),
    [anon_sym_switch] = ACTIONS(2530),
    [anon_sym_case] = ACTIONS(2530),
    [anon_sym_default] = ACTIONS(2530),
    [anon_sym_while] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2530),
    [anon_sym_for] = ACTIONS(2530),
    [anon_sym_in] = ACTIONS(2530),
    [anon_sym_return] = ACTIONS(2530),
    [anon_sym_break] = ACTIONS(2530),
    [anon_sym_continue] = ACTIONS(2530),
    [anon_sym_goto] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_sizeof] = ACTIONS(2530),
    [anon_sym___alignof__] = ACTIONS(2530),
    [anon_sym___alignof] = ACTIONS(2530),
    [anon_sym__alignof] = ACTIONS(2530),
    [anon_sym_alignof] = ACTIONS(2530),
    [anon_sym__Alignof] = ACTIONS(2530),
    [anon_sym_offsetof] = ACTIONS(2530),
    [anon_sym__Generic] = ACTIONS(2530),
    [anon_sym_asm] = ACTIONS(2530),
    [anon_sym___asm__] = ACTIONS(2530),
    [anon_sym___asm] = ACTIONS(2530),
    [sym_number_literal] = ACTIONS(2532),
    [anon_sym_L_SQUOTE] = ACTIONS(2532),
    [anon_sym_u_SQUOTE] = ACTIONS(2532),
    [anon_sym_U_SQUOTE] = ACTIONS(2532),
    [anon_sym_u8_SQUOTE] = ACTIONS(2532),
    [anon_sym_SQUOTE] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2530),
    [anon_sym_DQUOTE] = ACTIONS(2532),
    [anon_sym_L_DQUOTE] = ACTIONS(2532),
    [anon_sym_u_DQUOTE] = ACTIONS(2532),
    [anon_sym_U_DQUOTE] = ACTIONS(2532),
    [anon_sym_u8_DQUOTE] = ACTIONS(2532),
    [sym_true] = ACTIONS(2530),
    [sym_false] = ACTIONS(2530),
    [anon_sym_NULL] = ACTIONS(2530),
    [anon_sym_nullptr] = ACTIONS(2530),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2530),
    [anon_sym___typeof] = ACTIONS(2530),
    [anon_sym_typeof] = ACTIONS(2530),
    [anon_sym_ATimport] = ACTIONS(2532),
    [aux_sym_preproc_undef_token1] = ACTIONS(2530),
    [anon_sym_POUND] = ACTIONS(2530),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2530),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2530),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2530),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_API_AVAILABLE] = ACTIONS(2530),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_API_DEPRECATED] = ACTIONS(2530),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym___deprecated_msg] = ACTIONS(2530),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2530),
    [anon_sym_ATprotocol] = ACTIONS(2532),
    [anon_sym_ATinterface] = ACTIONS(2532),
    [anon_sym_ATimplementation] = ACTIONS(2532),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2532),
    [anon_sym_ATselector] = ACTIONS(2532),
    [anon_sym_ATavailable] = ACTIONS(2532),
    [anon_sym___builtin_available] = ACTIONS(2530),
    [anon_sym_va_arg] = ACTIONS(2530),
    [anon_sym_ATencode] = ACTIONS(2532),
    [anon_sym_BOOL] = ACTIONS(2530),
    [anon_sym_IMP] = ACTIONS(2530),
    [anon_sym_SEL] = ACTIONS(2530),
    [anon_sym_Class] = ACTIONS(2530),
    [anon_sym_id] = ACTIONS(2530),
  },
  [STATE(925)] = {
    [ts_builtin_sym_end] = ACTIONS(2496),
    [sym_identifier] = ACTIONS(2494),
    [aux_sym_preproc_include_token1] = ACTIONS(2494),
    [aux_sym_preproc_include_token2] = ACTIONS(2494),
    [aux_sym_preproc_def_token1] = ACTIONS(2494),
    [aux_sym_preproc_if_token1] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2494),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2494),
    [sym_preproc_directive] = ACTIONS(2494),
    [anon_sym_LPAREN2] = ACTIONS(2496),
    [anon_sym_BANG] = ACTIONS(2496),
    [anon_sym_TILDE] = ACTIONS(2496),
    [anon_sym_DASH] = ACTIONS(2494),
    [anon_sym_PLUS] = ACTIONS(2494),
    [anon_sym_STAR] = ACTIONS(2496),
    [anon_sym_CARET] = ACTIONS(2496),
    [anon_sym_AMP] = ACTIONS(2496),
    [anon_sym_SEMI] = ACTIONS(2496),
    [anon_sym___extension__] = ACTIONS(2494),
    [anon_sym_typedef] = ACTIONS(2494),
    [anon_sym_extern] = ACTIONS(2494),
    [anon_sym___attribute__] = ACTIONS(2494),
    [anon_sym___attribute] = ACTIONS(2494),
    [anon_sym_noreturn] = ACTIONS(2494),
    [anon_sym_LBRACK] = ACTIONS(2496),
    [anon_sym___declspec] = ACTIONS(2494),
    [anon_sym___cdecl] = ACTIONS(2494),
    [anon_sym___clrcall] = ACTIONS(2494),
    [anon_sym___stdcall] = ACTIONS(2494),
    [anon_sym___fastcall] = ACTIONS(2494),
    [anon_sym___thiscall] = ACTIONS(2494),
    [anon_sym___vectorcall] = ACTIONS(2494),
    [anon_sym_LBRACE] = ACTIONS(2496),
    [anon_sym_signed] = ACTIONS(2494),
    [anon_sym_unsigned] = ACTIONS(2494),
    [anon_sym_long] = ACTIONS(2494),
    [anon_sym_short] = ACTIONS(2494),
    [anon_sym_static] = ACTIONS(2494),
    [anon_sym_ATautoreleasepool] = ACTIONS(2496),
    [anon_sym_auto] = ACTIONS(2494),
    [anon_sym_register] = ACTIONS(2494),
    [anon_sym_inline] = ACTIONS(2494),
    [anon_sym___inline] = ACTIONS(2494),
    [anon_sym___inline__] = ACTIONS(2494),
    [anon_sym___forceinline] = ACTIONS(2494),
    [anon_sym_thread_local] = ACTIONS(2494),
    [anon_sym___thread] = ACTIONS(2494),
    [anon_sym_CG_EXTERN] = ACTIONS(2494),
    [anon_sym_CG_INLINE] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2494),
    [anon_sym_IBOutlet] = ACTIONS(2494),
    [anon_sym_IBInspectable] = ACTIONS(2494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2494),
    [anon_sym_NS_INLINE] = ACTIONS(2494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2494),
    [anon_sym_const] = ACTIONS(2494),
    [anon_sym_constexpr] = ACTIONS(2494),
    [anon_sym_volatile] = ACTIONS(2494),
    [anon_sym_restrict] = ACTIONS(2494),
    [anon_sym___restrict__] = ACTIONS(2494),
    [anon_sym__Atomic] = ACTIONS(2494),
    [anon_sym__Noreturn] = ACTIONS(2494),
    [anon_sym__Nonnull] = ACTIONS(2494),
    [anon_sym_nullable] = ACTIONS(2494),
    [anon_sym__Complex] = ACTIONS(2494),
    [anon_sym__Nullable] = ACTIONS(2494),
    [anon_sym__Nullable_result] = ACTIONS(2494),
    [anon_sym__Null_unspecified] = ACTIONS(2494),
    [anon_sym___autoreleasing] = ACTIONS(2494),
    [anon_sym___block] = ACTIONS(2494),
    [anon_sym___bridge] = ACTIONS(2494),
    [anon_sym___bridge_retained] = ACTIONS(2494),
    [anon_sym___bridge_transfer] = ACTIONS(2494),
    [anon_sym___complex] = ACTIONS(2494),
    [anon_sym___const] = ACTIONS(2494),
    [anon_sym___imag] = ACTIONS(2494),
    [anon_sym___kindof] = ACTIONS(2494),
    [anon_sym___nonnull] = ACTIONS(2494),
    [anon_sym___nullable] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2494),
    [anon_sym___real] = ACTIONS(2494),
    [anon_sym___strong] = ACTIONS(2494),
    [anon_sym___unsafe_unretained] = ACTIONS(2494),
    [anon_sym___unused] = ACTIONS(2494),
    [anon_sym___weak] = ACTIONS(2494),
    [anon_sym_alignas] = ACTIONS(2494),
    [anon_sym__Alignas] = ACTIONS(2494),
    [sym_primitive_type] = ACTIONS(2494),
    [anon_sym_enum] = ACTIONS(2494),
    [anon_sym_struct] = ACTIONS(2494),
    [anon_sym_union] = ACTIONS(2494),
    [anon_sym_if] = ACTIONS(2494),
    [anon_sym_else] = ACTIONS(2494),
    [anon_sym_switch] = ACTIONS(2494),
    [anon_sym_case] = ACTIONS(2494),
    [anon_sym_default] = ACTIONS(2494),
    [anon_sym_while] = ACTIONS(2494),
    [anon_sym_do] = ACTIONS(2494),
    [anon_sym_for] = ACTIONS(2494),
    [anon_sym_in] = ACTIONS(2494),
    [anon_sym_return] = ACTIONS(2494),
    [anon_sym_break] = ACTIONS(2494),
    [anon_sym_continue] = ACTIONS(2494),
    [anon_sym_goto] = ACTIONS(2494),
    [anon_sym_DASH_DASH] = ACTIONS(2496),
    [anon_sym_PLUS_PLUS] = ACTIONS(2496),
    [anon_sym_sizeof] = ACTIONS(2494),
    [anon_sym___alignof__] = ACTIONS(2494),
    [anon_sym___alignof] = ACTIONS(2494),
    [anon_sym__alignof] = ACTIONS(2494),
    [anon_sym_alignof] = ACTIONS(2494),
    [anon_sym__Alignof] = ACTIONS(2494),
    [anon_sym_offsetof] = ACTIONS(2494),
    [anon_sym__Generic] = ACTIONS(2494),
    [anon_sym_asm] = ACTIONS(2494),
    [anon_sym___asm__] = ACTIONS(2494),
    [anon_sym___asm] = ACTIONS(2494),
    [sym_number_literal] = ACTIONS(2496),
    [anon_sym_L_SQUOTE] = ACTIONS(2496),
    [anon_sym_u_SQUOTE] = ACTIONS(2496),
    [anon_sym_U_SQUOTE] = ACTIONS(2496),
    [anon_sym_u8_SQUOTE] = ACTIONS(2496),
    [anon_sym_SQUOTE] = ACTIONS(2496),
    [anon_sym_AT] = ACTIONS(2494),
    [anon_sym_DQUOTE] = ACTIONS(2496),
    [anon_sym_L_DQUOTE] = ACTIONS(2496),
    [anon_sym_u_DQUOTE] = ACTIONS(2496),
    [anon_sym_U_DQUOTE] = ACTIONS(2496),
    [anon_sym_u8_DQUOTE] = ACTIONS(2496),
    [sym_true] = ACTIONS(2494),
    [sym_false] = ACTIONS(2494),
    [anon_sym_NULL] = ACTIONS(2494),
    [anon_sym_nullptr] = ACTIONS(2494),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2494),
    [anon_sym___typeof] = ACTIONS(2494),
    [anon_sym_typeof] = ACTIONS(2494),
    [anon_sym_ATimport] = ACTIONS(2496),
    [aux_sym_preproc_undef_token1] = ACTIONS(2494),
    [anon_sym_POUND] = ACTIONS(2494),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_API_AVAILABLE] = ACTIONS(2494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_API_DEPRECATED] = ACTIONS(2494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym___deprecated_msg] = ACTIONS(2494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2494),
    [anon_sym_ATprotocol] = ACTIONS(2496),
    [anon_sym_ATinterface] = ACTIONS(2496),
    [anon_sym_ATimplementation] = ACTIONS(2496),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2496),
    [anon_sym_ATselector] = ACTIONS(2496),
    [anon_sym_ATavailable] = ACTIONS(2496),
    [anon_sym___builtin_available] = ACTIONS(2494),
    [anon_sym_va_arg] = ACTIONS(2494),
    [anon_sym_ATencode] = ACTIONS(2496),
    [anon_sym_BOOL] = ACTIONS(2494),
    [anon_sym_IMP] = ACTIONS(2494),
    [anon_sym_SEL] = ACTIONS(2494),
    [anon_sym_Class] = ACTIONS(2494),
    [anon_sym_id] = ACTIONS(2494),
  },
  [STATE(926)] = {
    [ts_builtin_sym_end] = ACTIONS(2520),
    [sym_identifier] = ACTIONS(2518),
    [aux_sym_preproc_include_token1] = ACTIONS(2518),
    [aux_sym_preproc_include_token2] = ACTIONS(2518),
    [aux_sym_preproc_def_token1] = ACTIONS(2518),
    [aux_sym_preproc_if_token1] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2518),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2518),
    [sym_preproc_directive] = ACTIONS(2518),
    [anon_sym_LPAREN2] = ACTIONS(2520),
    [anon_sym_BANG] = ACTIONS(2520),
    [anon_sym_TILDE] = ACTIONS(2520),
    [anon_sym_DASH] = ACTIONS(2518),
    [anon_sym_PLUS] = ACTIONS(2518),
    [anon_sym_STAR] = ACTIONS(2520),
    [anon_sym_CARET] = ACTIONS(2520),
    [anon_sym_AMP] = ACTIONS(2520),
    [anon_sym_SEMI] = ACTIONS(2520),
    [anon_sym___extension__] = ACTIONS(2518),
    [anon_sym_typedef] = ACTIONS(2518),
    [anon_sym_extern] = ACTIONS(2518),
    [anon_sym___attribute__] = ACTIONS(2518),
    [anon_sym___attribute] = ACTIONS(2518),
    [anon_sym_noreturn] = ACTIONS(2518),
    [anon_sym_LBRACK] = ACTIONS(2520),
    [anon_sym___declspec] = ACTIONS(2518),
    [anon_sym___cdecl] = ACTIONS(2518),
    [anon_sym___clrcall] = ACTIONS(2518),
    [anon_sym___stdcall] = ACTIONS(2518),
    [anon_sym___fastcall] = ACTIONS(2518),
    [anon_sym___thiscall] = ACTIONS(2518),
    [anon_sym___vectorcall] = ACTIONS(2518),
    [anon_sym_LBRACE] = ACTIONS(2520),
    [anon_sym_signed] = ACTIONS(2518),
    [anon_sym_unsigned] = ACTIONS(2518),
    [anon_sym_long] = ACTIONS(2518),
    [anon_sym_short] = ACTIONS(2518),
    [anon_sym_static] = ACTIONS(2518),
    [anon_sym_ATautoreleasepool] = ACTIONS(2520),
    [anon_sym_auto] = ACTIONS(2518),
    [anon_sym_register] = ACTIONS(2518),
    [anon_sym_inline] = ACTIONS(2518),
    [anon_sym___inline] = ACTIONS(2518),
    [anon_sym___inline__] = ACTIONS(2518),
    [anon_sym___forceinline] = ACTIONS(2518),
    [anon_sym_thread_local] = ACTIONS(2518),
    [anon_sym___thread] = ACTIONS(2518),
    [anon_sym_CG_EXTERN] = ACTIONS(2518),
    [anon_sym_CG_INLINE] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2518),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2518),
    [anon_sym_IBOutlet] = ACTIONS(2518),
    [anon_sym_IBInspectable] = ACTIONS(2518),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2518),
    [anon_sym_NS_INLINE] = ACTIONS(2518),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2518),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2518),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2518),
    [anon_sym_const] = ACTIONS(2518),
    [anon_sym_constexpr] = ACTIONS(2518),
    [anon_sym_volatile] = ACTIONS(2518),
    [anon_sym_restrict] = ACTIONS(2518),
    [anon_sym___restrict__] = ACTIONS(2518),
    [anon_sym__Atomic] = ACTIONS(2518),
    [anon_sym__Noreturn] = ACTIONS(2518),
    [anon_sym__Nonnull] = ACTIONS(2518),
    [anon_sym_nullable] = ACTIONS(2518),
    [anon_sym__Complex] = ACTIONS(2518),
    [anon_sym__Nullable] = ACTIONS(2518),
    [anon_sym__Nullable_result] = ACTIONS(2518),
    [anon_sym__Null_unspecified] = ACTIONS(2518),
    [anon_sym___autoreleasing] = ACTIONS(2518),
    [anon_sym___block] = ACTIONS(2518),
    [anon_sym___bridge] = ACTIONS(2518),
    [anon_sym___bridge_retained] = ACTIONS(2518),
    [anon_sym___bridge_transfer] = ACTIONS(2518),
    [anon_sym___complex] = ACTIONS(2518),
    [anon_sym___const] = ACTIONS(2518),
    [anon_sym___imag] = ACTIONS(2518),
    [anon_sym___kindof] = ACTIONS(2518),
    [anon_sym___nonnull] = ACTIONS(2518),
    [anon_sym___nullable] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2518),
    [anon_sym___real] = ACTIONS(2518),
    [anon_sym___strong] = ACTIONS(2518),
    [anon_sym___unsafe_unretained] = ACTIONS(2518),
    [anon_sym___unused] = ACTIONS(2518),
    [anon_sym___weak] = ACTIONS(2518),
    [anon_sym_alignas] = ACTIONS(2518),
    [anon_sym__Alignas] = ACTIONS(2518),
    [sym_primitive_type] = ACTIONS(2518),
    [anon_sym_enum] = ACTIONS(2518),
    [anon_sym_struct] = ACTIONS(2518),
    [anon_sym_union] = ACTIONS(2518),
    [anon_sym_if] = ACTIONS(2518),
    [anon_sym_else] = ACTIONS(2518),
    [anon_sym_switch] = ACTIONS(2518),
    [anon_sym_case] = ACTIONS(2518),
    [anon_sym_default] = ACTIONS(2518),
    [anon_sym_while] = ACTIONS(2518),
    [anon_sym_do] = ACTIONS(2518),
    [anon_sym_for] = ACTIONS(2518),
    [anon_sym_in] = ACTIONS(2518),
    [anon_sym_return] = ACTIONS(2518),
    [anon_sym_break] = ACTIONS(2518),
    [anon_sym_continue] = ACTIONS(2518),
    [anon_sym_goto] = ACTIONS(2518),
    [anon_sym_DASH_DASH] = ACTIONS(2520),
    [anon_sym_PLUS_PLUS] = ACTIONS(2520),
    [anon_sym_sizeof] = ACTIONS(2518),
    [anon_sym___alignof__] = ACTIONS(2518),
    [anon_sym___alignof] = ACTIONS(2518),
    [anon_sym__alignof] = ACTIONS(2518),
    [anon_sym_alignof] = ACTIONS(2518),
    [anon_sym__Alignof] = ACTIONS(2518),
    [anon_sym_offsetof] = ACTIONS(2518),
    [anon_sym__Generic] = ACTIONS(2518),
    [anon_sym_asm] = ACTIONS(2518),
    [anon_sym___asm__] = ACTIONS(2518),
    [anon_sym___asm] = ACTIONS(2518),
    [sym_number_literal] = ACTIONS(2520),
    [anon_sym_L_SQUOTE] = ACTIONS(2520),
    [anon_sym_u_SQUOTE] = ACTIONS(2520),
    [anon_sym_U_SQUOTE] = ACTIONS(2520),
    [anon_sym_u8_SQUOTE] = ACTIONS(2520),
    [anon_sym_SQUOTE] = ACTIONS(2520),
    [anon_sym_AT] = ACTIONS(2518),
    [anon_sym_DQUOTE] = ACTIONS(2520),
    [anon_sym_L_DQUOTE] = ACTIONS(2520),
    [anon_sym_u_DQUOTE] = ACTIONS(2520),
    [anon_sym_U_DQUOTE] = ACTIONS(2520),
    [anon_sym_u8_DQUOTE] = ACTIONS(2520),
    [sym_true] = ACTIONS(2518),
    [sym_false] = ACTIONS(2518),
    [anon_sym_NULL] = ACTIONS(2518),
    [anon_sym_nullptr] = ACTIONS(2518),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2518),
    [anon_sym___typeof] = ACTIONS(2518),
    [anon_sym_typeof] = ACTIONS(2518),
    [anon_sym_ATimport] = ACTIONS(2520),
    [aux_sym_preproc_undef_token1] = ACTIONS(2518),
    [anon_sym_POUND] = ACTIONS(2518),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2518),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2518),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2518),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_API_AVAILABLE] = ACTIONS(2518),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_API_DEPRECATED] = ACTIONS(2518),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym___deprecated_msg] = ACTIONS(2518),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2518),
    [anon_sym_ATprotocol] = ACTIONS(2520),
    [anon_sym_ATinterface] = ACTIONS(2520),
    [anon_sym_ATimplementation] = ACTIONS(2520),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2520),
    [anon_sym_ATselector] = ACTIONS(2520),
    [anon_sym_ATavailable] = ACTIONS(2520),
    [anon_sym___builtin_available] = ACTIONS(2518),
    [anon_sym_va_arg] = ACTIONS(2518),
    [anon_sym_ATencode] = ACTIONS(2520),
    [anon_sym_BOOL] = ACTIONS(2518),
    [anon_sym_IMP] = ACTIONS(2518),
    [anon_sym_SEL] = ACTIONS(2518),
    [anon_sym_Class] = ACTIONS(2518),
    [anon_sym_id] = ACTIONS(2518),
  },
  [STATE(927)] = {
    [ts_builtin_sym_end] = ACTIONS(2552),
    [sym_identifier] = ACTIONS(2550),
    [aux_sym_preproc_include_token1] = ACTIONS(2550),
    [aux_sym_preproc_include_token2] = ACTIONS(2550),
    [aux_sym_preproc_def_token1] = ACTIONS(2550),
    [aux_sym_preproc_if_token1] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2550),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2550),
    [sym_preproc_directive] = ACTIONS(2550),
    [anon_sym_LPAREN2] = ACTIONS(2552),
    [anon_sym_BANG] = ACTIONS(2552),
    [anon_sym_TILDE] = ACTIONS(2552),
    [anon_sym_DASH] = ACTIONS(2550),
    [anon_sym_PLUS] = ACTIONS(2550),
    [anon_sym_STAR] = ACTIONS(2552),
    [anon_sym_CARET] = ACTIONS(2552),
    [anon_sym_AMP] = ACTIONS(2552),
    [anon_sym_SEMI] = ACTIONS(2552),
    [anon_sym___extension__] = ACTIONS(2550),
    [anon_sym_typedef] = ACTIONS(2550),
    [anon_sym_extern] = ACTIONS(2550),
    [anon_sym___attribute__] = ACTIONS(2550),
    [anon_sym___attribute] = ACTIONS(2550),
    [anon_sym_noreturn] = ACTIONS(2550),
    [anon_sym_LBRACK] = ACTIONS(2552),
    [anon_sym___declspec] = ACTIONS(2550),
    [anon_sym___cdecl] = ACTIONS(2550),
    [anon_sym___clrcall] = ACTIONS(2550),
    [anon_sym___stdcall] = ACTIONS(2550),
    [anon_sym___fastcall] = ACTIONS(2550),
    [anon_sym___thiscall] = ACTIONS(2550),
    [anon_sym___vectorcall] = ACTIONS(2550),
    [anon_sym_LBRACE] = ACTIONS(2552),
    [anon_sym_signed] = ACTIONS(2550),
    [anon_sym_unsigned] = ACTIONS(2550),
    [anon_sym_long] = ACTIONS(2550),
    [anon_sym_short] = ACTIONS(2550),
    [anon_sym_static] = ACTIONS(2550),
    [anon_sym_ATautoreleasepool] = ACTIONS(2552),
    [anon_sym_auto] = ACTIONS(2550),
    [anon_sym_register] = ACTIONS(2550),
    [anon_sym_inline] = ACTIONS(2550),
    [anon_sym___inline] = ACTIONS(2550),
    [anon_sym___inline__] = ACTIONS(2550),
    [anon_sym___forceinline] = ACTIONS(2550),
    [anon_sym_thread_local] = ACTIONS(2550),
    [anon_sym___thread] = ACTIONS(2550),
    [anon_sym_CG_EXTERN] = ACTIONS(2550),
    [anon_sym_CG_INLINE] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2550),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2550),
    [anon_sym_IBOutlet] = ACTIONS(2550),
    [anon_sym_IBInspectable] = ACTIONS(2550),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2550),
    [anon_sym_NS_INLINE] = ACTIONS(2550),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2550),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2550),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2550),
    [anon_sym_const] = ACTIONS(2550),
    [anon_sym_constexpr] = ACTIONS(2550),
    [anon_sym_volatile] = ACTIONS(2550),
    [anon_sym_restrict] = ACTIONS(2550),
    [anon_sym___restrict__] = ACTIONS(2550),
    [anon_sym__Atomic] = ACTIONS(2550),
    [anon_sym__Noreturn] = ACTIONS(2550),
    [anon_sym__Nonnull] = ACTIONS(2550),
    [anon_sym_nullable] = ACTIONS(2550),
    [anon_sym__Complex] = ACTIONS(2550),
    [anon_sym__Nullable] = ACTIONS(2550),
    [anon_sym__Nullable_result] = ACTIONS(2550),
    [anon_sym__Null_unspecified] = ACTIONS(2550),
    [anon_sym___autoreleasing] = ACTIONS(2550),
    [anon_sym___block] = ACTIONS(2550),
    [anon_sym___bridge] = ACTIONS(2550),
    [anon_sym___bridge_retained] = ACTIONS(2550),
    [anon_sym___bridge_transfer] = ACTIONS(2550),
    [anon_sym___complex] = ACTIONS(2550),
    [anon_sym___const] = ACTIONS(2550),
    [anon_sym___imag] = ACTIONS(2550),
    [anon_sym___kindof] = ACTIONS(2550),
    [anon_sym___nonnull] = ACTIONS(2550),
    [anon_sym___nullable] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2550),
    [anon_sym___real] = ACTIONS(2550),
    [anon_sym___strong] = ACTIONS(2550),
    [anon_sym___unsafe_unretained] = ACTIONS(2550),
    [anon_sym___unused] = ACTIONS(2550),
    [anon_sym___weak] = ACTIONS(2550),
    [anon_sym_alignas] = ACTIONS(2550),
    [anon_sym__Alignas] = ACTIONS(2550),
    [sym_primitive_type] = ACTIONS(2550),
    [anon_sym_enum] = ACTIONS(2550),
    [anon_sym_struct] = ACTIONS(2550),
    [anon_sym_union] = ACTIONS(2550),
    [anon_sym_if] = ACTIONS(2550),
    [anon_sym_else] = ACTIONS(2550),
    [anon_sym_switch] = ACTIONS(2550),
    [anon_sym_case] = ACTIONS(2550),
    [anon_sym_default] = ACTIONS(2550),
    [anon_sym_while] = ACTIONS(2550),
    [anon_sym_do] = ACTIONS(2550),
    [anon_sym_for] = ACTIONS(2550),
    [anon_sym_in] = ACTIONS(2550),
    [anon_sym_return] = ACTIONS(2550),
    [anon_sym_break] = ACTIONS(2550),
    [anon_sym_continue] = ACTIONS(2550),
    [anon_sym_goto] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2552),
    [anon_sym_PLUS_PLUS] = ACTIONS(2552),
    [anon_sym_sizeof] = ACTIONS(2550),
    [anon_sym___alignof__] = ACTIONS(2550),
    [anon_sym___alignof] = ACTIONS(2550),
    [anon_sym__alignof] = ACTIONS(2550),
    [anon_sym_alignof] = ACTIONS(2550),
    [anon_sym__Alignof] = ACTIONS(2550),
    [anon_sym_offsetof] = ACTIONS(2550),
    [anon_sym__Generic] = ACTIONS(2550),
    [anon_sym_asm] = ACTIONS(2550),
    [anon_sym___asm__] = ACTIONS(2550),
    [anon_sym___asm] = ACTIONS(2550),
    [sym_number_literal] = ACTIONS(2552),
    [anon_sym_L_SQUOTE] = ACTIONS(2552),
    [anon_sym_u_SQUOTE] = ACTIONS(2552),
    [anon_sym_U_SQUOTE] = ACTIONS(2552),
    [anon_sym_u8_SQUOTE] = ACTIONS(2552),
    [anon_sym_SQUOTE] = ACTIONS(2552),
    [anon_sym_AT] = ACTIONS(2550),
    [anon_sym_DQUOTE] = ACTIONS(2552),
    [anon_sym_L_DQUOTE] = ACTIONS(2552),
    [anon_sym_u_DQUOTE] = ACTIONS(2552),
    [anon_sym_U_DQUOTE] = ACTIONS(2552),
    [anon_sym_u8_DQUOTE] = ACTIONS(2552),
    [sym_true] = ACTIONS(2550),
    [sym_false] = ACTIONS(2550),
    [anon_sym_NULL] = ACTIONS(2550),
    [anon_sym_nullptr] = ACTIONS(2550),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2550),
    [anon_sym___typeof] = ACTIONS(2550),
    [anon_sym_typeof] = ACTIONS(2550),
    [anon_sym_ATimport] = ACTIONS(2552),
    [aux_sym_preproc_undef_token1] = ACTIONS(2550),
    [anon_sym_POUND] = ACTIONS(2550),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2550),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2550),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2550),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_API_AVAILABLE] = ACTIONS(2550),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_API_DEPRECATED] = ACTIONS(2550),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym___deprecated_msg] = ACTIONS(2550),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2550),
    [anon_sym_ATprotocol] = ACTIONS(2552),
    [anon_sym_ATinterface] = ACTIONS(2552),
    [anon_sym_ATimplementation] = ACTIONS(2552),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2552),
    [anon_sym_ATselector] = ACTIONS(2552),
    [anon_sym_ATavailable] = ACTIONS(2552),
    [anon_sym___builtin_available] = ACTIONS(2550),
    [anon_sym_va_arg] = ACTIONS(2550),
    [anon_sym_ATencode] = ACTIONS(2552),
    [anon_sym_BOOL] = ACTIONS(2550),
    [anon_sym_IMP] = ACTIONS(2550),
    [anon_sym_SEL] = ACTIONS(2550),
    [anon_sym_Class] = ACTIONS(2550),
    [anon_sym_id] = ACTIONS(2550),
  },
  [STATE(928)] = {
    [ts_builtin_sym_end] = ACTIONS(2468),
    [sym_identifier] = ACTIONS(2466),
    [aux_sym_preproc_include_token1] = ACTIONS(2466),
    [aux_sym_preproc_include_token2] = ACTIONS(2466),
    [aux_sym_preproc_def_token1] = ACTIONS(2466),
    [aux_sym_preproc_if_token1] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2466),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2466),
    [sym_preproc_directive] = ACTIONS(2466),
    [anon_sym_LPAREN2] = ACTIONS(2468),
    [anon_sym_BANG] = ACTIONS(2468),
    [anon_sym_TILDE] = ACTIONS(2468),
    [anon_sym_DASH] = ACTIONS(2466),
    [anon_sym_PLUS] = ACTIONS(2466),
    [anon_sym_STAR] = ACTIONS(2468),
    [anon_sym_CARET] = ACTIONS(2468),
    [anon_sym_AMP] = ACTIONS(2468),
    [anon_sym_SEMI] = ACTIONS(2468),
    [anon_sym___extension__] = ACTIONS(2466),
    [anon_sym_typedef] = ACTIONS(2466),
    [anon_sym_extern] = ACTIONS(2466),
    [anon_sym___attribute__] = ACTIONS(2466),
    [anon_sym___attribute] = ACTIONS(2466),
    [anon_sym_noreturn] = ACTIONS(2466),
    [anon_sym_LBRACK] = ACTIONS(2468),
    [anon_sym___declspec] = ACTIONS(2466),
    [anon_sym___cdecl] = ACTIONS(2466),
    [anon_sym___clrcall] = ACTIONS(2466),
    [anon_sym___stdcall] = ACTIONS(2466),
    [anon_sym___fastcall] = ACTIONS(2466),
    [anon_sym___thiscall] = ACTIONS(2466),
    [anon_sym___vectorcall] = ACTIONS(2466),
    [anon_sym_LBRACE] = ACTIONS(2468),
    [anon_sym_signed] = ACTIONS(2466),
    [anon_sym_unsigned] = ACTIONS(2466),
    [anon_sym_long] = ACTIONS(2466),
    [anon_sym_short] = ACTIONS(2466),
    [anon_sym_static] = ACTIONS(2466),
    [anon_sym_ATautoreleasepool] = ACTIONS(2468),
    [anon_sym_auto] = ACTIONS(2466),
    [anon_sym_register] = ACTIONS(2466),
    [anon_sym_inline] = ACTIONS(2466),
    [anon_sym___inline] = ACTIONS(2466),
    [anon_sym___inline__] = ACTIONS(2466),
    [anon_sym___forceinline] = ACTIONS(2466),
    [anon_sym_thread_local] = ACTIONS(2466),
    [anon_sym___thread] = ACTIONS(2466),
    [anon_sym_CG_EXTERN] = ACTIONS(2466),
    [anon_sym_CG_INLINE] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2466),
    [anon_sym_IBOutlet] = ACTIONS(2466),
    [anon_sym_IBInspectable] = ACTIONS(2466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2466),
    [anon_sym_NS_INLINE] = ACTIONS(2466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2466),
    [anon_sym_const] = ACTIONS(2466),
    [anon_sym_constexpr] = ACTIONS(2466),
    [anon_sym_volatile] = ACTIONS(2466),
    [anon_sym_restrict] = ACTIONS(2466),
    [anon_sym___restrict__] = ACTIONS(2466),
    [anon_sym__Atomic] = ACTIONS(2466),
    [anon_sym__Noreturn] = ACTIONS(2466),
    [anon_sym__Nonnull] = ACTIONS(2466),
    [anon_sym_nullable] = ACTIONS(2466),
    [anon_sym__Complex] = ACTIONS(2466),
    [anon_sym__Nullable] = ACTIONS(2466),
    [anon_sym__Nullable_result] = ACTIONS(2466),
    [anon_sym__Null_unspecified] = ACTIONS(2466),
    [anon_sym___autoreleasing] = ACTIONS(2466),
    [anon_sym___block] = ACTIONS(2466),
    [anon_sym___bridge] = ACTIONS(2466),
    [anon_sym___bridge_retained] = ACTIONS(2466),
    [anon_sym___bridge_transfer] = ACTIONS(2466),
    [anon_sym___complex] = ACTIONS(2466),
    [anon_sym___const] = ACTIONS(2466),
    [anon_sym___imag] = ACTIONS(2466),
    [anon_sym___kindof] = ACTIONS(2466),
    [anon_sym___nonnull] = ACTIONS(2466),
    [anon_sym___nullable] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2466),
    [anon_sym___real] = ACTIONS(2466),
    [anon_sym___strong] = ACTIONS(2466),
    [anon_sym___unsafe_unretained] = ACTIONS(2466),
    [anon_sym___unused] = ACTIONS(2466),
    [anon_sym___weak] = ACTIONS(2466),
    [anon_sym_alignas] = ACTIONS(2466),
    [anon_sym__Alignas] = ACTIONS(2466),
    [sym_primitive_type] = ACTIONS(2466),
    [anon_sym_enum] = ACTIONS(2466),
    [anon_sym_struct] = ACTIONS(2466),
    [anon_sym_union] = ACTIONS(2466),
    [anon_sym_if] = ACTIONS(2466),
    [anon_sym_else] = ACTIONS(2466),
    [anon_sym_switch] = ACTIONS(2466),
    [anon_sym_case] = ACTIONS(2466),
    [anon_sym_default] = ACTIONS(2466),
    [anon_sym_while] = ACTIONS(2466),
    [anon_sym_do] = ACTIONS(2466),
    [anon_sym_for] = ACTIONS(2466),
    [anon_sym_in] = ACTIONS(2466),
    [anon_sym_return] = ACTIONS(2466),
    [anon_sym_break] = ACTIONS(2466),
    [anon_sym_continue] = ACTIONS(2466),
    [anon_sym_goto] = ACTIONS(2466),
    [anon_sym_DASH_DASH] = ACTIONS(2468),
    [anon_sym_PLUS_PLUS] = ACTIONS(2468),
    [anon_sym_sizeof] = ACTIONS(2466),
    [anon_sym___alignof__] = ACTIONS(2466),
    [anon_sym___alignof] = ACTIONS(2466),
    [anon_sym__alignof] = ACTIONS(2466),
    [anon_sym_alignof] = ACTIONS(2466),
    [anon_sym__Alignof] = ACTIONS(2466),
    [anon_sym_offsetof] = ACTIONS(2466),
    [anon_sym__Generic] = ACTIONS(2466),
    [anon_sym_asm] = ACTIONS(2466),
    [anon_sym___asm__] = ACTIONS(2466),
    [anon_sym___asm] = ACTIONS(2466),
    [sym_number_literal] = ACTIONS(2468),
    [anon_sym_L_SQUOTE] = ACTIONS(2468),
    [anon_sym_u_SQUOTE] = ACTIONS(2468),
    [anon_sym_U_SQUOTE] = ACTIONS(2468),
    [anon_sym_u8_SQUOTE] = ACTIONS(2468),
    [anon_sym_SQUOTE] = ACTIONS(2468),
    [anon_sym_AT] = ACTIONS(2466),
    [anon_sym_DQUOTE] = ACTIONS(2468),
    [anon_sym_L_DQUOTE] = ACTIONS(2468),
    [anon_sym_u_DQUOTE] = ACTIONS(2468),
    [anon_sym_U_DQUOTE] = ACTIONS(2468),
    [anon_sym_u8_DQUOTE] = ACTIONS(2468),
    [sym_true] = ACTIONS(2466),
    [sym_false] = ACTIONS(2466),
    [anon_sym_NULL] = ACTIONS(2466),
    [anon_sym_nullptr] = ACTIONS(2466),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2466),
    [anon_sym___typeof] = ACTIONS(2466),
    [anon_sym_typeof] = ACTIONS(2466),
    [anon_sym_ATimport] = ACTIONS(2468),
    [aux_sym_preproc_undef_token1] = ACTIONS(2466),
    [anon_sym_POUND] = ACTIONS(2466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_API_AVAILABLE] = ACTIONS(2466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_API_DEPRECATED] = ACTIONS(2466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym___deprecated_msg] = ACTIONS(2466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2466),
    [anon_sym_ATprotocol] = ACTIONS(2468),
    [anon_sym_ATinterface] = ACTIONS(2468),
    [anon_sym_ATimplementation] = ACTIONS(2468),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2468),
    [anon_sym_ATselector] = ACTIONS(2468),
    [anon_sym_ATavailable] = ACTIONS(2468),
    [anon_sym___builtin_available] = ACTIONS(2466),
    [anon_sym_va_arg] = ACTIONS(2466),
    [anon_sym_ATencode] = ACTIONS(2468),
    [anon_sym_BOOL] = ACTIONS(2466),
    [anon_sym_IMP] = ACTIONS(2466),
    [anon_sym_SEL] = ACTIONS(2466),
    [anon_sym_Class] = ACTIONS(2466),
    [anon_sym_id] = ACTIONS(2466),
  },
  [STATE(929)] = {
    [ts_builtin_sym_end] = ACTIONS(2516),
    [sym_identifier] = ACTIONS(2514),
    [aux_sym_preproc_include_token1] = ACTIONS(2514),
    [aux_sym_preproc_include_token2] = ACTIONS(2514),
    [aux_sym_preproc_def_token1] = ACTIONS(2514),
    [aux_sym_preproc_if_token1] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2514),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2514),
    [sym_preproc_directive] = ACTIONS(2514),
    [anon_sym_LPAREN2] = ACTIONS(2516),
    [anon_sym_BANG] = ACTIONS(2516),
    [anon_sym_TILDE] = ACTIONS(2516),
    [anon_sym_DASH] = ACTIONS(2514),
    [anon_sym_PLUS] = ACTIONS(2514),
    [anon_sym_STAR] = ACTIONS(2516),
    [anon_sym_CARET] = ACTIONS(2516),
    [anon_sym_AMP] = ACTIONS(2516),
    [anon_sym_SEMI] = ACTIONS(2516),
    [anon_sym___extension__] = ACTIONS(2514),
    [anon_sym_typedef] = ACTIONS(2514),
    [anon_sym_extern] = ACTIONS(2514),
    [anon_sym___attribute__] = ACTIONS(2514),
    [anon_sym___attribute] = ACTIONS(2514),
    [anon_sym_noreturn] = ACTIONS(2514),
    [anon_sym_LBRACK] = ACTIONS(2516),
    [anon_sym___declspec] = ACTIONS(2514),
    [anon_sym___cdecl] = ACTIONS(2514),
    [anon_sym___clrcall] = ACTIONS(2514),
    [anon_sym___stdcall] = ACTIONS(2514),
    [anon_sym___fastcall] = ACTIONS(2514),
    [anon_sym___thiscall] = ACTIONS(2514),
    [anon_sym___vectorcall] = ACTIONS(2514),
    [anon_sym_LBRACE] = ACTIONS(2516),
    [anon_sym_signed] = ACTIONS(2514),
    [anon_sym_unsigned] = ACTIONS(2514),
    [anon_sym_long] = ACTIONS(2514),
    [anon_sym_short] = ACTIONS(2514),
    [anon_sym_static] = ACTIONS(2514),
    [anon_sym_ATautoreleasepool] = ACTIONS(2516),
    [anon_sym_auto] = ACTIONS(2514),
    [anon_sym_register] = ACTIONS(2514),
    [anon_sym_inline] = ACTIONS(2514),
    [anon_sym___inline] = ACTIONS(2514),
    [anon_sym___inline__] = ACTIONS(2514),
    [anon_sym___forceinline] = ACTIONS(2514),
    [anon_sym_thread_local] = ACTIONS(2514),
    [anon_sym___thread] = ACTIONS(2514),
    [anon_sym_CG_EXTERN] = ACTIONS(2514),
    [anon_sym_CG_INLINE] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2514),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2514),
    [anon_sym_IBOutlet] = ACTIONS(2514),
    [anon_sym_IBInspectable] = ACTIONS(2514),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2514),
    [anon_sym_NS_INLINE] = ACTIONS(2514),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2514),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2514),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2514),
    [anon_sym_const] = ACTIONS(2514),
    [anon_sym_constexpr] = ACTIONS(2514),
    [anon_sym_volatile] = ACTIONS(2514),
    [anon_sym_restrict] = ACTIONS(2514),
    [anon_sym___restrict__] = ACTIONS(2514),
    [anon_sym__Atomic] = ACTIONS(2514),
    [anon_sym__Noreturn] = ACTIONS(2514),
    [anon_sym__Nonnull] = ACTIONS(2514),
    [anon_sym_nullable] = ACTIONS(2514),
    [anon_sym__Complex] = ACTIONS(2514),
    [anon_sym__Nullable] = ACTIONS(2514),
    [anon_sym__Nullable_result] = ACTIONS(2514),
    [anon_sym__Null_unspecified] = ACTIONS(2514),
    [anon_sym___autoreleasing] = ACTIONS(2514),
    [anon_sym___block] = ACTIONS(2514),
    [anon_sym___bridge] = ACTIONS(2514),
    [anon_sym___bridge_retained] = ACTIONS(2514),
    [anon_sym___bridge_transfer] = ACTIONS(2514),
    [anon_sym___complex] = ACTIONS(2514),
    [anon_sym___const] = ACTIONS(2514),
    [anon_sym___imag] = ACTIONS(2514),
    [anon_sym___kindof] = ACTIONS(2514),
    [anon_sym___nonnull] = ACTIONS(2514),
    [anon_sym___nullable] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2514),
    [anon_sym___real] = ACTIONS(2514),
    [anon_sym___strong] = ACTIONS(2514),
    [anon_sym___unsafe_unretained] = ACTIONS(2514),
    [anon_sym___unused] = ACTIONS(2514),
    [anon_sym___weak] = ACTIONS(2514),
    [anon_sym_alignas] = ACTIONS(2514),
    [anon_sym__Alignas] = ACTIONS(2514),
    [sym_primitive_type] = ACTIONS(2514),
    [anon_sym_enum] = ACTIONS(2514),
    [anon_sym_struct] = ACTIONS(2514),
    [anon_sym_union] = ACTIONS(2514),
    [anon_sym_if] = ACTIONS(2514),
    [anon_sym_else] = ACTIONS(2514),
    [anon_sym_switch] = ACTIONS(2514),
    [anon_sym_case] = ACTIONS(2514),
    [anon_sym_default] = ACTIONS(2514),
    [anon_sym_while] = ACTIONS(2514),
    [anon_sym_do] = ACTIONS(2514),
    [anon_sym_for] = ACTIONS(2514),
    [anon_sym_in] = ACTIONS(2514),
    [anon_sym_return] = ACTIONS(2514),
    [anon_sym_break] = ACTIONS(2514),
    [anon_sym_continue] = ACTIONS(2514),
    [anon_sym_goto] = ACTIONS(2514),
    [anon_sym_DASH_DASH] = ACTIONS(2516),
    [anon_sym_PLUS_PLUS] = ACTIONS(2516),
    [anon_sym_sizeof] = ACTIONS(2514),
    [anon_sym___alignof__] = ACTIONS(2514),
    [anon_sym___alignof] = ACTIONS(2514),
    [anon_sym__alignof] = ACTIONS(2514),
    [anon_sym_alignof] = ACTIONS(2514),
    [anon_sym__Alignof] = ACTIONS(2514),
    [anon_sym_offsetof] = ACTIONS(2514),
    [anon_sym__Generic] = ACTIONS(2514),
    [anon_sym_asm] = ACTIONS(2514),
    [anon_sym___asm__] = ACTIONS(2514),
    [anon_sym___asm] = ACTIONS(2514),
    [sym_number_literal] = ACTIONS(2516),
    [anon_sym_L_SQUOTE] = ACTIONS(2516),
    [anon_sym_u_SQUOTE] = ACTIONS(2516),
    [anon_sym_U_SQUOTE] = ACTIONS(2516),
    [anon_sym_u8_SQUOTE] = ACTIONS(2516),
    [anon_sym_SQUOTE] = ACTIONS(2516),
    [anon_sym_AT] = ACTIONS(2514),
    [anon_sym_DQUOTE] = ACTIONS(2516),
    [anon_sym_L_DQUOTE] = ACTIONS(2516),
    [anon_sym_u_DQUOTE] = ACTIONS(2516),
    [anon_sym_U_DQUOTE] = ACTIONS(2516),
    [anon_sym_u8_DQUOTE] = ACTIONS(2516),
    [sym_true] = ACTIONS(2514),
    [sym_false] = ACTIONS(2514),
    [anon_sym_NULL] = ACTIONS(2514),
    [anon_sym_nullptr] = ACTIONS(2514),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2514),
    [anon_sym___typeof] = ACTIONS(2514),
    [anon_sym_typeof] = ACTIONS(2514),
    [anon_sym_ATimport] = ACTIONS(2516),
    [aux_sym_preproc_undef_token1] = ACTIONS(2514),
    [anon_sym_POUND] = ACTIONS(2514),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2514),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2514),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2514),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_API_AVAILABLE] = ACTIONS(2514),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_API_DEPRECATED] = ACTIONS(2514),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym___deprecated_msg] = ACTIONS(2514),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2514),
    [anon_sym_ATprotocol] = ACTIONS(2516),
    [anon_sym_ATinterface] = ACTIONS(2516),
    [anon_sym_ATimplementation] = ACTIONS(2516),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2516),
    [anon_sym_ATselector] = ACTIONS(2516),
    [anon_sym_ATavailable] = ACTIONS(2516),
    [anon_sym___builtin_available] = ACTIONS(2514),
    [anon_sym_va_arg] = ACTIONS(2514),
    [anon_sym_ATencode] = ACTIONS(2516),
    [anon_sym_BOOL] = ACTIONS(2514),
    [anon_sym_IMP] = ACTIONS(2514),
    [anon_sym_SEL] = ACTIONS(2514),
    [anon_sym_Class] = ACTIONS(2514),
    [anon_sym_id] = ACTIONS(2514),
  },
  [STATE(930)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(931)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(932)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(933)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(934)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(935)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(936)] = {
    [ts_builtin_sym_end] = ACTIONS(2540),
    [sym_identifier] = ACTIONS(2538),
    [aux_sym_preproc_include_token1] = ACTIONS(2538),
    [aux_sym_preproc_include_token2] = ACTIONS(2538),
    [aux_sym_preproc_def_token1] = ACTIONS(2538),
    [aux_sym_preproc_if_token1] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2538),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2538),
    [sym_preproc_directive] = ACTIONS(2538),
    [anon_sym_LPAREN2] = ACTIONS(2540),
    [anon_sym_BANG] = ACTIONS(2540),
    [anon_sym_TILDE] = ACTIONS(2540),
    [anon_sym_DASH] = ACTIONS(2538),
    [anon_sym_PLUS] = ACTIONS(2538),
    [anon_sym_STAR] = ACTIONS(2540),
    [anon_sym_CARET] = ACTIONS(2540),
    [anon_sym_AMP] = ACTIONS(2540),
    [anon_sym_SEMI] = ACTIONS(2540),
    [anon_sym___extension__] = ACTIONS(2538),
    [anon_sym_typedef] = ACTIONS(2538),
    [anon_sym_extern] = ACTIONS(2538),
    [anon_sym___attribute__] = ACTIONS(2538),
    [anon_sym___attribute] = ACTIONS(2538),
    [anon_sym_noreturn] = ACTIONS(2538),
    [anon_sym_LBRACK] = ACTIONS(2540),
    [anon_sym___declspec] = ACTIONS(2538),
    [anon_sym___cdecl] = ACTIONS(2538),
    [anon_sym___clrcall] = ACTIONS(2538),
    [anon_sym___stdcall] = ACTIONS(2538),
    [anon_sym___fastcall] = ACTIONS(2538),
    [anon_sym___thiscall] = ACTIONS(2538),
    [anon_sym___vectorcall] = ACTIONS(2538),
    [anon_sym_LBRACE] = ACTIONS(2540),
    [anon_sym_signed] = ACTIONS(2538),
    [anon_sym_unsigned] = ACTIONS(2538),
    [anon_sym_long] = ACTIONS(2538),
    [anon_sym_short] = ACTIONS(2538),
    [anon_sym_static] = ACTIONS(2538),
    [anon_sym_ATautoreleasepool] = ACTIONS(2540),
    [anon_sym_auto] = ACTIONS(2538),
    [anon_sym_register] = ACTIONS(2538),
    [anon_sym_inline] = ACTIONS(2538),
    [anon_sym___inline] = ACTIONS(2538),
    [anon_sym___inline__] = ACTIONS(2538),
    [anon_sym___forceinline] = ACTIONS(2538),
    [anon_sym_thread_local] = ACTIONS(2538),
    [anon_sym___thread] = ACTIONS(2538),
    [anon_sym_CG_EXTERN] = ACTIONS(2538),
    [anon_sym_CG_INLINE] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2538),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2538),
    [anon_sym_IBOutlet] = ACTIONS(2538),
    [anon_sym_IBInspectable] = ACTIONS(2538),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2538),
    [anon_sym_NS_INLINE] = ACTIONS(2538),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2538),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2538),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2538),
    [anon_sym_const] = ACTIONS(2538),
    [anon_sym_constexpr] = ACTIONS(2538),
    [anon_sym_volatile] = ACTIONS(2538),
    [anon_sym_restrict] = ACTIONS(2538),
    [anon_sym___restrict__] = ACTIONS(2538),
    [anon_sym__Atomic] = ACTIONS(2538),
    [anon_sym__Noreturn] = ACTIONS(2538),
    [anon_sym__Nonnull] = ACTIONS(2538),
    [anon_sym_nullable] = ACTIONS(2538),
    [anon_sym__Complex] = ACTIONS(2538),
    [anon_sym__Nullable] = ACTIONS(2538),
    [anon_sym__Nullable_result] = ACTIONS(2538),
    [anon_sym__Null_unspecified] = ACTIONS(2538),
    [anon_sym___autoreleasing] = ACTIONS(2538),
    [anon_sym___block] = ACTIONS(2538),
    [anon_sym___bridge] = ACTIONS(2538),
    [anon_sym___bridge_retained] = ACTIONS(2538),
    [anon_sym___bridge_transfer] = ACTIONS(2538),
    [anon_sym___complex] = ACTIONS(2538),
    [anon_sym___const] = ACTIONS(2538),
    [anon_sym___imag] = ACTIONS(2538),
    [anon_sym___kindof] = ACTIONS(2538),
    [anon_sym___nonnull] = ACTIONS(2538),
    [anon_sym___nullable] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2538),
    [anon_sym___real] = ACTIONS(2538),
    [anon_sym___strong] = ACTIONS(2538),
    [anon_sym___unsafe_unretained] = ACTIONS(2538),
    [anon_sym___unused] = ACTIONS(2538),
    [anon_sym___weak] = ACTIONS(2538),
    [anon_sym_alignas] = ACTIONS(2538),
    [anon_sym__Alignas] = ACTIONS(2538),
    [sym_primitive_type] = ACTIONS(2538),
    [anon_sym_enum] = ACTIONS(2538),
    [anon_sym_struct] = ACTIONS(2538),
    [anon_sym_union] = ACTIONS(2538),
    [anon_sym_if] = ACTIONS(2538),
    [anon_sym_else] = ACTIONS(2538),
    [anon_sym_switch] = ACTIONS(2538),
    [anon_sym_case] = ACTIONS(2538),
    [anon_sym_default] = ACTIONS(2538),
    [anon_sym_while] = ACTIONS(2538),
    [anon_sym_do] = ACTIONS(2538),
    [anon_sym_for] = ACTIONS(2538),
    [anon_sym_in] = ACTIONS(2538),
    [anon_sym_return] = ACTIONS(2538),
    [anon_sym_break] = ACTIONS(2538),
    [anon_sym_continue] = ACTIONS(2538),
    [anon_sym_goto] = ACTIONS(2538),
    [anon_sym_DASH_DASH] = ACTIONS(2540),
    [anon_sym_PLUS_PLUS] = ACTIONS(2540),
    [anon_sym_sizeof] = ACTIONS(2538),
    [anon_sym___alignof__] = ACTIONS(2538),
    [anon_sym___alignof] = ACTIONS(2538),
    [anon_sym__alignof] = ACTIONS(2538),
    [anon_sym_alignof] = ACTIONS(2538),
    [anon_sym__Alignof] = ACTIONS(2538),
    [anon_sym_offsetof] = ACTIONS(2538),
    [anon_sym__Generic] = ACTIONS(2538),
    [anon_sym_asm] = ACTIONS(2538),
    [anon_sym___asm__] = ACTIONS(2538),
    [anon_sym___asm] = ACTIONS(2538),
    [sym_number_literal] = ACTIONS(2540),
    [anon_sym_L_SQUOTE] = ACTIONS(2540),
    [anon_sym_u_SQUOTE] = ACTIONS(2540),
    [anon_sym_U_SQUOTE] = ACTIONS(2540),
    [anon_sym_u8_SQUOTE] = ACTIONS(2540),
    [anon_sym_SQUOTE] = ACTIONS(2540),
    [anon_sym_AT] = ACTIONS(2538),
    [anon_sym_DQUOTE] = ACTIONS(2540),
    [anon_sym_L_DQUOTE] = ACTIONS(2540),
    [anon_sym_u_DQUOTE] = ACTIONS(2540),
    [anon_sym_U_DQUOTE] = ACTIONS(2540),
    [anon_sym_u8_DQUOTE] = ACTIONS(2540),
    [sym_true] = ACTIONS(2538),
    [sym_false] = ACTIONS(2538),
    [anon_sym_NULL] = ACTIONS(2538),
    [anon_sym_nullptr] = ACTIONS(2538),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2538),
    [anon_sym___typeof] = ACTIONS(2538),
    [anon_sym_typeof] = ACTIONS(2538),
    [anon_sym_ATimport] = ACTIONS(2540),
    [aux_sym_preproc_undef_token1] = ACTIONS(2538),
    [anon_sym_POUND] = ACTIONS(2538),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2538),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2538),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2538),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_API_AVAILABLE] = ACTIONS(2538),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_API_DEPRECATED] = ACTIONS(2538),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym___deprecated_msg] = ACTIONS(2538),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2538),
    [anon_sym_ATprotocol] = ACTIONS(2540),
    [anon_sym_ATinterface] = ACTIONS(2540),
    [anon_sym_ATimplementation] = ACTIONS(2540),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2540),
    [anon_sym_ATselector] = ACTIONS(2540),
    [anon_sym_ATavailable] = ACTIONS(2540),
    [anon_sym___builtin_available] = ACTIONS(2538),
    [anon_sym_va_arg] = ACTIONS(2538),
    [anon_sym_ATencode] = ACTIONS(2540),
    [anon_sym_BOOL] = ACTIONS(2538),
    [anon_sym_IMP] = ACTIONS(2538),
    [anon_sym_SEL] = ACTIONS(2538),
    [anon_sym_Class] = ACTIONS(2538),
    [anon_sym_id] = ACTIONS(2538),
  },
  [STATE(937)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(938)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(939)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(940)] = {
    [ts_builtin_sym_end] = ACTIONS(2536),
    [sym_identifier] = ACTIONS(2534),
    [aux_sym_preproc_include_token1] = ACTIONS(2534),
    [aux_sym_preproc_include_token2] = ACTIONS(2534),
    [aux_sym_preproc_def_token1] = ACTIONS(2534),
    [aux_sym_preproc_if_token1] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2534),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2534),
    [sym_preproc_directive] = ACTIONS(2534),
    [anon_sym_LPAREN2] = ACTIONS(2536),
    [anon_sym_BANG] = ACTIONS(2536),
    [anon_sym_TILDE] = ACTIONS(2536),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2536),
    [anon_sym_CARET] = ACTIONS(2536),
    [anon_sym_AMP] = ACTIONS(2536),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym___extension__] = ACTIONS(2534),
    [anon_sym_typedef] = ACTIONS(2534),
    [anon_sym_extern] = ACTIONS(2534),
    [anon_sym___attribute__] = ACTIONS(2534),
    [anon_sym___attribute] = ACTIONS(2534),
    [anon_sym_noreturn] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2536),
    [anon_sym___declspec] = ACTIONS(2534),
    [anon_sym___cdecl] = ACTIONS(2534),
    [anon_sym___clrcall] = ACTIONS(2534),
    [anon_sym___stdcall] = ACTIONS(2534),
    [anon_sym___fastcall] = ACTIONS(2534),
    [anon_sym___thiscall] = ACTIONS(2534),
    [anon_sym___vectorcall] = ACTIONS(2534),
    [anon_sym_LBRACE] = ACTIONS(2536),
    [anon_sym_signed] = ACTIONS(2534),
    [anon_sym_unsigned] = ACTIONS(2534),
    [anon_sym_long] = ACTIONS(2534),
    [anon_sym_short] = ACTIONS(2534),
    [anon_sym_static] = ACTIONS(2534),
    [anon_sym_ATautoreleasepool] = ACTIONS(2536),
    [anon_sym_auto] = ACTIONS(2534),
    [anon_sym_register] = ACTIONS(2534),
    [anon_sym_inline] = ACTIONS(2534),
    [anon_sym___inline] = ACTIONS(2534),
    [anon_sym___inline__] = ACTIONS(2534),
    [anon_sym___forceinline] = ACTIONS(2534),
    [anon_sym_thread_local] = ACTIONS(2534),
    [anon_sym___thread] = ACTIONS(2534),
    [anon_sym_CG_EXTERN] = ACTIONS(2534),
    [anon_sym_CG_INLINE] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2534),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2534),
    [anon_sym_IBOutlet] = ACTIONS(2534),
    [anon_sym_IBInspectable] = ACTIONS(2534),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2534),
    [anon_sym_NS_INLINE] = ACTIONS(2534),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2534),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2534),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2534),
    [anon_sym_const] = ACTIONS(2534),
    [anon_sym_constexpr] = ACTIONS(2534),
    [anon_sym_volatile] = ACTIONS(2534),
    [anon_sym_restrict] = ACTIONS(2534),
    [anon_sym___restrict__] = ACTIONS(2534),
    [anon_sym__Atomic] = ACTIONS(2534),
    [anon_sym__Noreturn] = ACTIONS(2534),
    [anon_sym__Nonnull] = ACTIONS(2534),
    [anon_sym_nullable] = ACTIONS(2534),
    [anon_sym__Complex] = ACTIONS(2534),
    [anon_sym__Nullable] = ACTIONS(2534),
    [anon_sym__Nullable_result] = ACTIONS(2534),
    [anon_sym__Null_unspecified] = ACTIONS(2534),
    [anon_sym___autoreleasing] = ACTIONS(2534),
    [anon_sym___block] = ACTIONS(2534),
    [anon_sym___bridge] = ACTIONS(2534),
    [anon_sym___bridge_retained] = ACTIONS(2534),
    [anon_sym___bridge_transfer] = ACTIONS(2534),
    [anon_sym___complex] = ACTIONS(2534),
    [anon_sym___const] = ACTIONS(2534),
    [anon_sym___imag] = ACTIONS(2534),
    [anon_sym___kindof] = ACTIONS(2534),
    [anon_sym___nonnull] = ACTIONS(2534),
    [anon_sym___nullable] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2534),
    [anon_sym___real] = ACTIONS(2534),
    [anon_sym___strong] = ACTIONS(2534),
    [anon_sym___unsafe_unretained] = ACTIONS(2534),
    [anon_sym___unused] = ACTIONS(2534),
    [anon_sym___weak] = ACTIONS(2534),
    [anon_sym_alignas] = ACTIONS(2534),
    [anon_sym__Alignas] = ACTIONS(2534),
    [sym_primitive_type] = ACTIONS(2534),
    [anon_sym_enum] = ACTIONS(2534),
    [anon_sym_struct] = ACTIONS(2534),
    [anon_sym_union] = ACTIONS(2534),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_else] = ACTIONS(2534),
    [anon_sym_switch] = ACTIONS(2534),
    [anon_sym_case] = ACTIONS(2534),
    [anon_sym_default] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_do] = ACTIONS(2534),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_in] = ACTIONS(2534),
    [anon_sym_return] = ACTIONS(2534),
    [anon_sym_break] = ACTIONS(2534),
    [anon_sym_continue] = ACTIONS(2534),
    [anon_sym_goto] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_sizeof] = ACTIONS(2534),
    [anon_sym___alignof__] = ACTIONS(2534),
    [anon_sym___alignof] = ACTIONS(2534),
    [anon_sym__alignof] = ACTIONS(2534),
    [anon_sym_alignof] = ACTIONS(2534),
    [anon_sym__Alignof] = ACTIONS(2534),
    [anon_sym_offsetof] = ACTIONS(2534),
    [anon_sym__Generic] = ACTIONS(2534),
    [anon_sym_asm] = ACTIONS(2534),
    [anon_sym___asm__] = ACTIONS(2534),
    [anon_sym___asm] = ACTIONS(2534),
    [sym_number_literal] = ACTIONS(2536),
    [anon_sym_L_SQUOTE] = ACTIONS(2536),
    [anon_sym_u_SQUOTE] = ACTIONS(2536),
    [anon_sym_U_SQUOTE] = ACTIONS(2536),
    [anon_sym_u8_SQUOTE] = ACTIONS(2536),
    [anon_sym_SQUOTE] = ACTIONS(2536),
    [anon_sym_AT] = ACTIONS(2534),
    [anon_sym_DQUOTE] = ACTIONS(2536),
    [anon_sym_L_DQUOTE] = ACTIONS(2536),
    [anon_sym_u_DQUOTE] = ACTIONS(2536),
    [anon_sym_U_DQUOTE] = ACTIONS(2536),
    [anon_sym_u8_DQUOTE] = ACTIONS(2536),
    [sym_true] = ACTIONS(2534),
    [sym_false] = ACTIONS(2534),
    [anon_sym_NULL] = ACTIONS(2534),
    [anon_sym_nullptr] = ACTIONS(2534),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2534),
    [anon_sym___typeof] = ACTIONS(2534),
    [anon_sym_typeof] = ACTIONS(2534),
    [anon_sym_ATimport] = ACTIONS(2536),
    [aux_sym_preproc_undef_token1] = ACTIONS(2534),
    [anon_sym_POUND] = ACTIONS(2534),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2534),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2534),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2534),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_API_AVAILABLE] = ACTIONS(2534),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_API_DEPRECATED] = ACTIONS(2534),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym___deprecated_msg] = ACTIONS(2534),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2534),
    [anon_sym_ATprotocol] = ACTIONS(2536),
    [anon_sym_ATinterface] = ACTIONS(2536),
    [anon_sym_ATimplementation] = ACTIONS(2536),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2536),
    [anon_sym_ATselector] = ACTIONS(2536),
    [anon_sym_ATavailable] = ACTIONS(2536),
    [anon_sym___builtin_available] = ACTIONS(2534),
    [anon_sym_va_arg] = ACTIONS(2534),
    [anon_sym_ATencode] = ACTIONS(2536),
    [anon_sym_BOOL] = ACTIONS(2534),
    [anon_sym_IMP] = ACTIONS(2534),
    [anon_sym_SEL] = ACTIONS(2534),
    [anon_sym_Class] = ACTIONS(2534),
    [anon_sym_id] = ACTIONS(2534),
  },
  [STATE(941)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(942)] = {
    [ts_builtin_sym_end] = ACTIONS(2472),
    [sym_identifier] = ACTIONS(2470),
    [aux_sym_preproc_include_token1] = ACTIONS(2470),
    [aux_sym_preproc_include_token2] = ACTIONS(2470),
    [aux_sym_preproc_def_token1] = ACTIONS(2470),
    [aux_sym_preproc_if_token1] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2470),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2470),
    [sym_preproc_directive] = ACTIONS(2470),
    [anon_sym_LPAREN2] = ACTIONS(2472),
    [anon_sym_BANG] = ACTIONS(2472),
    [anon_sym_TILDE] = ACTIONS(2472),
    [anon_sym_DASH] = ACTIONS(2470),
    [anon_sym_PLUS] = ACTIONS(2470),
    [anon_sym_STAR] = ACTIONS(2472),
    [anon_sym_CARET] = ACTIONS(2472),
    [anon_sym_AMP] = ACTIONS(2472),
    [anon_sym_SEMI] = ACTIONS(2472),
    [anon_sym___extension__] = ACTIONS(2470),
    [anon_sym_typedef] = ACTIONS(2470),
    [anon_sym_extern] = ACTIONS(2470),
    [anon_sym___attribute__] = ACTIONS(2470),
    [anon_sym___attribute] = ACTIONS(2470),
    [anon_sym_noreturn] = ACTIONS(2470),
    [anon_sym_LBRACK] = ACTIONS(2472),
    [anon_sym___declspec] = ACTIONS(2470),
    [anon_sym___cdecl] = ACTIONS(2470),
    [anon_sym___clrcall] = ACTIONS(2470),
    [anon_sym___stdcall] = ACTIONS(2470),
    [anon_sym___fastcall] = ACTIONS(2470),
    [anon_sym___thiscall] = ACTIONS(2470),
    [anon_sym___vectorcall] = ACTIONS(2470),
    [anon_sym_LBRACE] = ACTIONS(2472),
    [anon_sym_signed] = ACTIONS(2470),
    [anon_sym_unsigned] = ACTIONS(2470),
    [anon_sym_long] = ACTIONS(2470),
    [anon_sym_short] = ACTIONS(2470),
    [anon_sym_static] = ACTIONS(2470),
    [anon_sym_ATautoreleasepool] = ACTIONS(2472),
    [anon_sym_auto] = ACTIONS(2470),
    [anon_sym_register] = ACTIONS(2470),
    [anon_sym_inline] = ACTIONS(2470),
    [anon_sym___inline] = ACTIONS(2470),
    [anon_sym___inline__] = ACTIONS(2470),
    [anon_sym___forceinline] = ACTIONS(2470),
    [anon_sym_thread_local] = ACTIONS(2470),
    [anon_sym___thread] = ACTIONS(2470),
    [anon_sym_CG_EXTERN] = ACTIONS(2470),
    [anon_sym_CG_INLINE] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2470),
    [anon_sym_IBOutlet] = ACTIONS(2470),
    [anon_sym_IBInspectable] = ACTIONS(2470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2470),
    [anon_sym_NS_INLINE] = ACTIONS(2470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2470),
    [anon_sym_const] = ACTIONS(2470),
    [anon_sym_constexpr] = ACTIONS(2470),
    [anon_sym_volatile] = ACTIONS(2470),
    [anon_sym_restrict] = ACTIONS(2470),
    [anon_sym___restrict__] = ACTIONS(2470),
    [anon_sym__Atomic] = ACTIONS(2470),
    [anon_sym__Noreturn] = ACTIONS(2470),
    [anon_sym__Nonnull] = ACTIONS(2470),
    [anon_sym_nullable] = ACTIONS(2470),
    [anon_sym__Complex] = ACTIONS(2470),
    [anon_sym__Nullable] = ACTIONS(2470),
    [anon_sym__Nullable_result] = ACTIONS(2470),
    [anon_sym__Null_unspecified] = ACTIONS(2470),
    [anon_sym___autoreleasing] = ACTIONS(2470),
    [anon_sym___block] = ACTIONS(2470),
    [anon_sym___bridge] = ACTIONS(2470),
    [anon_sym___bridge_retained] = ACTIONS(2470),
    [anon_sym___bridge_transfer] = ACTIONS(2470),
    [anon_sym___complex] = ACTIONS(2470),
    [anon_sym___const] = ACTIONS(2470),
    [anon_sym___imag] = ACTIONS(2470),
    [anon_sym___kindof] = ACTIONS(2470),
    [anon_sym___nonnull] = ACTIONS(2470),
    [anon_sym___nullable] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2470),
    [anon_sym___real] = ACTIONS(2470),
    [anon_sym___strong] = ACTIONS(2470),
    [anon_sym___unsafe_unretained] = ACTIONS(2470),
    [anon_sym___unused] = ACTIONS(2470),
    [anon_sym___weak] = ACTIONS(2470),
    [anon_sym_alignas] = ACTIONS(2470),
    [anon_sym__Alignas] = ACTIONS(2470),
    [sym_primitive_type] = ACTIONS(2470),
    [anon_sym_enum] = ACTIONS(2470),
    [anon_sym_struct] = ACTIONS(2470),
    [anon_sym_union] = ACTIONS(2470),
    [anon_sym_if] = ACTIONS(2470),
    [anon_sym_else] = ACTIONS(2470),
    [anon_sym_switch] = ACTIONS(2470),
    [anon_sym_case] = ACTIONS(2470),
    [anon_sym_default] = ACTIONS(2470),
    [anon_sym_while] = ACTIONS(2470),
    [anon_sym_do] = ACTIONS(2470),
    [anon_sym_for] = ACTIONS(2470),
    [anon_sym_in] = ACTIONS(2470),
    [anon_sym_return] = ACTIONS(2470),
    [anon_sym_break] = ACTIONS(2470),
    [anon_sym_continue] = ACTIONS(2470),
    [anon_sym_goto] = ACTIONS(2470),
    [anon_sym_DASH_DASH] = ACTIONS(2472),
    [anon_sym_PLUS_PLUS] = ACTIONS(2472),
    [anon_sym_sizeof] = ACTIONS(2470),
    [anon_sym___alignof__] = ACTIONS(2470),
    [anon_sym___alignof] = ACTIONS(2470),
    [anon_sym__alignof] = ACTIONS(2470),
    [anon_sym_alignof] = ACTIONS(2470),
    [anon_sym__Alignof] = ACTIONS(2470),
    [anon_sym_offsetof] = ACTIONS(2470),
    [anon_sym__Generic] = ACTIONS(2470),
    [anon_sym_asm] = ACTIONS(2470),
    [anon_sym___asm__] = ACTIONS(2470),
    [anon_sym___asm] = ACTIONS(2470),
    [sym_number_literal] = ACTIONS(2472),
    [anon_sym_L_SQUOTE] = ACTIONS(2472),
    [anon_sym_u_SQUOTE] = ACTIONS(2472),
    [anon_sym_U_SQUOTE] = ACTIONS(2472),
    [anon_sym_u8_SQUOTE] = ACTIONS(2472),
    [anon_sym_SQUOTE] = ACTIONS(2472),
    [anon_sym_AT] = ACTIONS(2470),
    [anon_sym_DQUOTE] = ACTIONS(2472),
    [anon_sym_L_DQUOTE] = ACTIONS(2472),
    [anon_sym_u_DQUOTE] = ACTIONS(2472),
    [anon_sym_U_DQUOTE] = ACTIONS(2472),
    [anon_sym_u8_DQUOTE] = ACTIONS(2472),
    [sym_true] = ACTIONS(2470),
    [sym_false] = ACTIONS(2470),
    [anon_sym_NULL] = ACTIONS(2470),
    [anon_sym_nullptr] = ACTIONS(2470),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2470),
    [anon_sym___typeof] = ACTIONS(2470),
    [anon_sym_typeof] = ACTIONS(2470),
    [anon_sym_ATimport] = ACTIONS(2472),
    [aux_sym_preproc_undef_token1] = ACTIONS(2470),
    [anon_sym_POUND] = ACTIONS(2470),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_API_AVAILABLE] = ACTIONS(2470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_API_DEPRECATED] = ACTIONS(2470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym___deprecated_msg] = ACTIONS(2470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2470),
    [anon_sym_ATprotocol] = ACTIONS(2472),
    [anon_sym_ATinterface] = ACTIONS(2472),
    [anon_sym_ATimplementation] = ACTIONS(2472),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2472),
    [anon_sym_ATselector] = ACTIONS(2472),
    [anon_sym_ATavailable] = ACTIONS(2472),
    [anon_sym___builtin_available] = ACTIONS(2470),
    [anon_sym_va_arg] = ACTIONS(2470),
    [anon_sym_ATencode] = ACTIONS(2472),
    [anon_sym_BOOL] = ACTIONS(2470),
    [anon_sym_IMP] = ACTIONS(2470),
    [anon_sym_SEL] = ACTIONS(2470),
    [anon_sym_Class] = ACTIONS(2470),
    [anon_sym_id] = ACTIONS(2470),
  },
  [STATE(943)] = {
    [ts_builtin_sym_end] = ACTIONS(2440),
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(944)] = {
    [ts_builtin_sym_end] = ACTIONS(2544),
    [sym_identifier] = ACTIONS(2542),
    [aux_sym_preproc_include_token1] = ACTIONS(2542),
    [aux_sym_preproc_include_token2] = ACTIONS(2542),
    [aux_sym_preproc_def_token1] = ACTIONS(2542),
    [aux_sym_preproc_if_token1] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2542),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2542),
    [sym_preproc_directive] = ACTIONS(2542),
    [anon_sym_LPAREN2] = ACTIONS(2544),
    [anon_sym_BANG] = ACTIONS(2544),
    [anon_sym_TILDE] = ACTIONS(2544),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_PLUS] = ACTIONS(2542),
    [anon_sym_STAR] = ACTIONS(2544),
    [anon_sym_CARET] = ACTIONS(2544),
    [anon_sym_AMP] = ACTIONS(2544),
    [anon_sym_SEMI] = ACTIONS(2544),
    [anon_sym___extension__] = ACTIONS(2542),
    [anon_sym_typedef] = ACTIONS(2542),
    [anon_sym_extern] = ACTIONS(2542),
    [anon_sym___attribute__] = ACTIONS(2542),
    [anon_sym___attribute] = ACTIONS(2542),
    [anon_sym_noreturn] = ACTIONS(2542),
    [anon_sym_LBRACK] = ACTIONS(2544),
    [anon_sym___declspec] = ACTIONS(2542),
    [anon_sym___cdecl] = ACTIONS(2542),
    [anon_sym___clrcall] = ACTIONS(2542),
    [anon_sym___stdcall] = ACTIONS(2542),
    [anon_sym___fastcall] = ACTIONS(2542),
    [anon_sym___thiscall] = ACTIONS(2542),
    [anon_sym___vectorcall] = ACTIONS(2542),
    [anon_sym_LBRACE] = ACTIONS(2544),
    [anon_sym_signed] = ACTIONS(2542),
    [anon_sym_unsigned] = ACTIONS(2542),
    [anon_sym_long] = ACTIONS(2542),
    [anon_sym_short] = ACTIONS(2542),
    [anon_sym_static] = ACTIONS(2542),
    [anon_sym_ATautoreleasepool] = ACTIONS(2544),
    [anon_sym_auto] = ACTIONS(2542),
    [anon_sym_register] = ACTIONS(2542),
    [anon_sym_inline] = ACTIONS(2542),
    [anon_sym___inline] = ACTIONS(2542),
    [anon_sym___inline__] = ACTIONS(2542),
    [anon_sym___forceinline] = ACTIONS(2542),
    [anon_sym_thread_local] = ACTIONS(2542),
    [anon_sym___thread] = ACTIONS(2542),
    [anon_sym_CG_EXTERN] = ACTIONS(2542),
    [anon_sym_CG_INLINE] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2542),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2542),
    [anon_sym_IBOutlet] = ACTIONS(2542),
    [anon_sym_IBInspectable] = ACTIONS(2542),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2542),
    [anon_sym_NS_INLINE] = ACTIONS(2542),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2542),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2542),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2542),
    [anon_sym_const] = ACTIONS(2542),
    [anon_sym_constexpr] = ACTIONS(2542),
    [anon_sym_volatile] = ACTIONS(2542),
    [anon_sym_restrict] = ACTIONS(2542),
    [anon_sym___restrict__] = ACTIONS(2542),
    [anon_sym__Atomic] = ACTIONS(2542),
    [anon_sym__Noreturn] = ACTIONS(2542),
    [anon_sym__Nonnull] = ACTIONS(2542),
    [anon_sym_nullable] = ACTIONS(2542),
    [anon_sym__Complex] = ACTIONS(2542),
    [anon_sym__Nullable] = ACTIONS(2542),
    [anon_sym__Nullable_result] = ACTIONS(2542),
    [anon_sym__Null_unspecified] = ACTIONS(2542),
    [anon_sym___autoreleasing] = ACTIONS(2542),
    [anon_sym___block] = ACTIONS(2542),
    [anon_sym___bridge] = ACTIONS(2542),
    [anon_sym___bridge_retained] = ACTIONS(2542),
    [anon_sym___bridge_transfer] = ACTIONS(2542),
    [anon_sym___complex] = ACTIONS(2542),
    [anon_sym___const] = ACTIONS(2542),
    [anon_sym___imag] = ACTIONS(2542),
    [anon_sym___kindof] = ACTIONS(2542),
    [anon_sym___nonnull] = ACTIONS(2542),
    [anon_sym___nullable] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2542),
    [anon_sym___real] = ACTIONS(2542),
    [anon_sym___strong] = ACTIONS(2542),
    [anon_sym___unsafe_unretained] = ACTIONS(2542),
    [anon_sym___unused] = ACTIONS(2542),
    [anon_sym___weak] = ACTIONS(2542),
    [anon_sym_alignas] = ACTIONS(2542),
    [anon_sym__Alignas] = ACTIONS(2542),
    [sym_primitive_type] = ACTIONS(2542),
    [anon_sym_enum] = ACTIONS(2542),
    [anon_sym_struct] = ACTIONS(2542),
    [anon_sym_union] = ACTIONS(2542),
    [anon_sym_if] = ACTIONS(2542),
    [anon_sym_else] = ACTIONS(2542),
    [anon_sym_switch] = ACTIONS(2542),
    [anon_sym_case] = ACTIONS(2542),
    [anon_sym_default] = ACTIONS(2542),
    [anon_sym_while] = ACTIONS(2542),
    [anon_sym_do] = ACTIONS(2542),
    [anon_sym_for] = ACTIONS(2542),
    [anon_sym_in] = ACTIONS(2542),
    [anon_sym_return] = ACTIONS(2542),
    [anon_sym_break] = ACTIONS(2542),
    [anon_sym_continue] = ACTIONS(2542),
    [anon_sym_goto] = ACTIONS(2542),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_sizeof] = ACTIONS(2542),
    [anon_sym___alignof__] = ACTIONS(2542),
    [anon_sym___alignof] = ACTIONS(2542),
    [anon_sym__alignof] = ACTIONS(2542),
    [anon_sym_alignof] = ACTIONS(2542),
    [anon_sym__Alignof] = ACTIONS(2542),
    [anon_sym_offsetof] = ACTIONS(2542),
    [anon_sym__Generic] = ACTIONS(2542),
    [anon_sym_asm] = ACTIONS(2542),
    [anon_sym___asm__] = ACTIONS(2542),
    [anon_sym___asm] = ACTIONS(2542),
    [sym_number_literal] = ACTIONS(2544),
    [anon_sym_L_SQUOTE] = ACTIONS(2544),
    [anon_sym_u_SQUOTE] = ACTIONS(2544),
    [anon_sym_U_SQUOTE] = ACTIONS(2544),
    [anon_sym_u8_SQUOTE] = ACTIONS(2544),
    [anon_sym_SQUOTE] = ACTIONS(2544),
    [anon_sym_AT] = ACTIONS(2542),
    [anon_sym_DQUOTE] = ACTIONS(2544),
    [anon_sym_L_DQUOTE] = ACTIONS(2544),
    [anon_sym_u_DQUOTE] = ACTIONS(2544),
    [anon_sym_U_DQUOTE] = ACTIONS(2544),
    [anon_sym_u8_DQUOTE] = ACTIONS(2544),
    [sym_true] = ACTIONS(2542),
    [sym_false] = ACTIONS(2542),
    [anon_sym_NULL] = ACTIONS(2542),
    [anon_sym_nullptr] = ACTIONS(2542),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2542),
    [anon_sym___typeof] = ACTIONS(2542),
    [anon_sym_typeof] = ACTIONS(2542),
    [anon_sym_ATimport] = ACTIONS(2544),
    [aux_sym_preproc_undef_token1] = ACTIONS(2542),
    [anon_sym_POUND] = ACTIONS(2542),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2542),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2542),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2542),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_API_AVAILABLE] = ACTIONS(2542),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_API_DEPRECATED] = ACTIONS(2542),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym___deprecated_msg] = ACTIONS(2542),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2542),
    [anon_sym_ATprotocol] = ACTIONS(2544),
    [anon_sym_ATinterface] = ACTIONS(2544),
    [anon_sym_ATimplementation] = ACTIONS(2544),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2544),
    [anon_sym_ATselector] = ACTIONS(2544),
    [anon_sym_ATavailable] = ACTIONS(2544),
    [anon_sym___builtin_available] = ACTIONS(2542),
    [anon_sym_va_arg] = ACTIONS(2542),
    [anon_sym_ATencode] = ACTIONS(2544),
    [anon_sym_BOOL] = ACTIONS(2542),
    [anon_sym_IMP] = ACTIONS(2542),
    [anon_sym_SEL] = ACTIONS(2542),
    [anon_sym_Class] = ACTIONS(2542),
    [anon_sym_id] = ACTIONS(2542),
  },
  [STATE(945)] = {
    [ts_builtin_sym_end] = ACTIONS(2448),
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(946)] = {
    [ts_builtin_sym_end] = ACTIONS(2460),
    [sym_identifier] = ACTIONS(2458),
    [aux_sym_preproc_include_token1] = ACTIONS(2458),
    [aux_sym_preproc_include_token2] = ACTIONS(2458),
    [aux_sym_preproc_def_token1] = ACTIONS(2458),
    [aux_sym_preproc_if_token1] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2458),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2458),
    [sym_preproc_directive] = ACTIONS(2458),
    [anon_sym_LPAREN2] = ACTIONS(2460),
    [anon_sym_BANG] = ACTIONS(2460),
    [anon_sym_TILDE] = ACTIONS(2460),
    [anon_sym_DASH] = ACTIONS(2458),
    [anon_sym_PLUS] = ACTIONS(2458),
    [anon_sym_STAR] = ACTIONS(2460),
    [anon_sym_CARET] = ACTIONS(2460),
    [anon_sym_AMP] = ACTIONS(2460),
    [anon_sym_SEMI] = ACTIONS(2460),
    [anon_sym___extension__] = ACTIONS(2458),
    [anon_sym_typedef] = ACTIONS(2458),
    [anon_sym_extern] = ACTIONS(2458),
    [anon_sym___attribute__] = ACTIONS(2458),
    [anon_sym___attribute] = ACTIONS(2458),
    [anon_sym_noreturn] = ACTIONS(2458),
    [anon_sym_LBRACK] = ACTIONS(2460),
    [anon_sym___declspec] = ACTIONS(2458),
    [anon_sym___cdecl] = ACTIONS(2458),
    [anon_sym___clrcall] = ACTIONS(2458),
    [anon_sym___stdcall] = ACTIONS(2458),
    [anon_sym___fastcall] = ACTIONS(2458),
    [anon_sym___thiscall] = ACTIONS(2458),
    [anon_sym___vectorcall] = ACTIONS(2458),
    [anon_sym_LBRACE] = ACTIONS(2460),
    [anon_sym_signed] = ACTIONS(2458),
    [anon_sym_unsigned] = ACTIONS(2458),
    [anon_sym_long] = ACTIONS(2458),
    [anon_sym_short] = ACTIONS(2458),
    [anon_sym_static] = ACTIONS(2458),
    [anon_sym_ATautoreleasepool] = ACTIONS(2460),
    [anon_sym_auto] = ACTIONS(2458),
    [anon_sym_register] = ACTIONS(2458),
    [anon_sym_inline] = ACTIONS(2458),
    [anon_sym___inline] = ACTIONS(2458),
    [anon_sym___inline__] = ACTIONS(2458),
    [anon_sym___forceinline] = ACTIONS(2458),
    [anon_sym_thread_local] = ACTIONS(2458),
    [anon_sym___thread] = ACTIONS(2458),
    [anon_sym_CG_EXTERN] = ACTIONS(2458),
    [anon_sym_CG_INLINE] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2458),
    [anon_sym_IBOutlet] = ACTIONS(2458),
    [anon_sym_IBInspectable] = ACTIONS(2458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2458),
    [anon_sym_NS_INLINE] = ACTIONS(2458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2458),
    [anon_sym_const] = ACTIONS(2458),
    [anon_sym_constexpr] = ACTIONS(2458),
    [anon_sym_volatile] = ACTIONS(2458),
    [anon_sym_restrict] = ACTIONS(2458),
    [anon_sym___restrict__] = ACTIONS(2458),
    [anon_sym__Atomic] = ACTIONS(2458),
    [anon_sym__Noreturn] = ACTIONS(2458),
    [anon_sym__Nonnull] = ACTIONS(2458),
    [anon_sym_nullable] = ACTIONS(2458),
    [anon_sym__Complex] = ACTIONS(2458),
    [anon_sym__Nullable] = ACTIONS(2458),
    [anon_sym__Nullable_result] = ACTIONS(2458),
    [anon_sym__Null_unspecified] = ACTIONS(2458),
    [anon_sym___autoreleasing] = ACTIONS(2458),
    [anon_sym___block] = ACTIONS(2458),
    [anon_sym___bridge] = ACTIONS(2458),
    [anon_sym___bridge_retained] = ACTIONS(2458),
    [anon_sym___bridge_transfer] = ACTIONS(2458),
    [anon_sym___complex] = ACTIONS(2458),
    [anon_sym___const] = ACTIONS(2458),
    [anon_sym___imag] = ACTIONS(2458),
    [anon_sym___kindof] = ACTIONS(2458),
    [anon_sym___nonnull] = ACTIONS(2458),
    [anon_sym___nullable] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2458),
    [anon_sym___real] = ACTIONS(2458),
    [anon_sym___strong] = ACTIONS(2458),
    [anon_sym___unsafe_unretained] = ACTIONS(2458),
    [anon_sym___unused] = ACTIONS(2458),
    [anon_sym___weak] = ACTIONS(2458),
    [anon_sym_alignas] = ACTIONS(2458),
    [anon_sym__Alignas] = ACTIONS(2458),
    [sym_primitive_type] = ACTIONS(2458),
    [anon_sym_enum] = ACTIONS(2458),
    [anon_sym_struct] = ACTIONS(2458),
    [anon_sym_union] = ACTIONS(2458),
    [anon_sym_if] = ACTIONS(2458),
    [anon_sym_else] = ACTIONS(2458),
    [anon_sym_switch] = ACTIONS(2458),
    [anon_sym_case] = ACTIONS(2458),
    [anon_sym_default] = ACTIONS(2458),
    [anon_sym_while] = ACTIONS(2458),
    [anon_sym_do] = ACTIONS(2458),
    [anon_sym_for] = ACTIONS(2458),
    [anon_sym_in] = ACTIONS(2458),
    [anon_sym_return] = ACTIONS(2458),
    [anon_sym_break] = ACTIONS(2458),
    [anon_sym_continue] = ACTIONS(2458),
    [anon_sym_goto] = ACTIONS(2458),
    [anon_sym_DASH_DASH] = ACTIONS(2460),
    [anon_sym_PLUS_PLUS] = ACTIONS(2460),
    [anon_sym_sizeof] = ACTIONS(2458),
    [anon_sym___alignof__] = ACTIONS(2458),
    [anon_sym___alignof] = ACTIONS(2458),
    [anon_sym__alignof] = ACTIONS(2458),
    [anon_sym_alignof] = ACTIONS(2458),
    [anon_sym__Alignof] = ACTIONS(2458),
    [anon_sym_offsetof] = ACTIONS(2458),
    [anon_sym__Generic] = ACTIONS(2458),
    [anon_sym_asm] = ACTIONS(2458),
    [anon_sym___asm__] = ACTIONS(2458),
    [anon_sym___asm] = ACTIONS(2458),
    [sym_number_literal] = ACTIONS(2460),
    [anon_sym_L_SQUOTE] = ACTIONS(2460),
    [anon_sym_u_SQUOTE] = ACTIONS(2460),
    [anon_sym_U_SQUOTE] = ACTIONS(2460),
    [anon_sym_u8_SQUOTE] = ACTIONS(2460),
    [anon_sym_SQUOTE] = ACTIONS(2460),
    [anon_sym_AT] = ACTIONS(2458),
    [anon_sym_DQUOTE] = ACTIONS(2460),
    [anon_sym_L_DQUOTE] = ACTIONS(2460),
    [anon_sym_u_DQUOTE] = ACTIONS(2460),
    [anon_sym_U_DQUOTE] = ACTIONS(2460),
    [anon_sym_u8_DQUOTE] = ACTIONS(2460),
    [sym_true] = ACTIONS(2458),
    [sym_false] = ACTIONS(2458),
    [anon_sym_NULL] = ACTIONS(2458),
    [anon_sym_nullptr] = ACTIONS(2458),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2458),
    [anon_sym___typeof] = ACTIONS(2458),
    [anon_sym_typeof] = ACTIONS(2458),
    [anon_sym_ATimport] = ACTIONS(2460),
    [aux_sym_preproc_undef_token1] = ACTIONS(2458),
    [anon_sym_POUND] = ACTIONS(2458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_API_AVAILABLE] = ACTIONS(2458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_API_DEPRECATED] = ACTIONS(2458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym___deprecated_msg] = ACTIONS(2458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2458),
    [anon_sym_ATprotocol] = ACTIONS(2460),
    [anon_sym_ATinterface] = ACTIONS(2460),
    [anon_sym_ATimplementation] = ACTIONS(2460),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2460),
    [anon_sym_ATselector] = ACTIONS(2460),
    [anon_sym_ATavailable] = ACTIONS(2460),
    [anon_sym___builtin_available] = ACTIONS(2458),
    [anon_sym_va_arg] = ACTIONS(2458),
    [anon_sym_ATencode] = ACTIONS(2460),
    [anon_sym_BOOL] = ACTIONS(2458),
    [anon_sym_IMP] = ACTIONS(2458),
    [anon_sym_SEL] = ACTIONS(2458),
    [anon_sym_Class] = ACTIONS(2458),
    [anon_sym_id] = ACTIONS(2458),
  },
  [STATE(947)] = {
    [ts_builtin_sym_end] = ACTIONS(2508),
    [sym_identifier] = ACTIONS(2506),
    [aux_sym_preproc_include_token1] = ACTIONS(2506),
    [aux_sym_preproc_include_token2] = ACTIONS(2506),
    [aux_sym_preproc_def_token1] = ACTIONS(2506),
    [aux_sym_preproc_if_token1] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2506),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2506),
    [sym_preproc_directive] = ACTIONS(2506),
    [anon_sym_LPAREN2] = ACTIONS(2508),
    [anon_sym_BANG] = ACTIONS(2508),
    [anon_sym_TILDE] = ACTIONS(2508),
    [anon_sym_DASH] = ACTIONS(2506),
    [anon_sym_PLUS] = ACTIONS(2506),
    [anon_sym_STAR] = ACTIONS(2508),
    [anon_sym_CARET] = ACTIONS(2508),
    [anon_sym_AMP] = ACTIONS(2508),
    [anon_sym_SEMI] = ACTIONS(2508),
    [anon_sym___extension__] = ACTIONS(2506),
    [anon_sym_typedef] = ACTIONS(2506),
    [anon_sym_extern] = ACTIONS(2506),
    [anon_sym___attribute__] = ACTIONS(2506),
    [anon_sym___attribute] = ACTIONS(2506),
    [anon_sym_noreturn] = ACTIONS(2506),
    [anon_sym_LBRACK] = ACTIONS(2508),
    [anon_sym___declspec] = ACTIONS(2506),
    [anon_sym___cdecl] = ACTIONS(2506),
    [anon_sym___clrcall] = ACTIONS(2506),
    [anon_sym___stdcall] = ACTIONS(2506),
    [anon_sym___fastcall] = ACTIONS(2506),
    [anon_sym___thiscall] = ACTIONS(2506),
    [anon_sym___vectorcall] = ACTIONS(2506),
    [anon_sym_LBRACE] = ACTIONS(2508),
    [anon_sym_signed] = ACTIONS(2506),
    [anon_sym_unsigned] = ACTIONS(2506),
    [anon_sym_long] = ACTIONS(2506),
    [anon_sym_short] = ACTIONS(2506),
    [anon_sym_static] = ACTIONS(2506),
    [anon_sym_ATautoreleasepool] = ACTIONS(2508),
    [anon_sym_auto] = ACTIONS(2506),
    [anon_sym_register] = ACTIONS(2506),
    [anon_sym_inline] = ACTIONS(2506),
    [anon_sym___inline] = ACTIONS(2506),
    [anon_sym___inline__] = ACTIONS(2506),
    [anon_sym___forceinline] = ACTIONS(2506),
    [anon_sym_thread_local] = ACTIONS(2506),
    [anon_sym___thread] = ACTIONS(2506),
    [anon_sym_CG_EXTERN] = ACTIONS(2506),
    [anon_sym_CG_INLINE] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2506),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2506),
    [anon_sym_IBOutlet] = ACTIONS(2506),
    [anon_sym_IBInspectable] = ACTIONS(2506),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2506),
    [anon_sym_NS_INLINE] = ACTIONS(2506),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2506),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2506),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2506),
    [anon_sym_const] = ACTIONS(2506),
    [anon_sym_constexpr] = ACTIONS(2506),
    [anon_sym_volatile] = ACTIONS(2506),
    [anon_sym_restrict] = ACTIONS(2506),
    [anon_sym___restrict__] = ACTIONS(2506),
    [anon_sym__Atomic] = ACTIONS(2506),
    [anon_sym__Noreturn] = ACTIONS(2506),
    [anon_sym__Nonnull] = ACTIONS(2506),
    [anon_sym_nullable] = ACTIONS(2506),
    [anon_sym__Complex] = ACTIONS(2506),
    [anon_sym__Nullable] = ACTIONS(2506),
    [anon_sym__Nullable_result] = ACTIONS(2506),
    [anon_sym__Null_unspecified] = ACTIONS(2506),
    [anon_sym___autoreleasing] = ACTIONS(2506),
    [anon_sym___block] = ACTIONS(2506),
    [anon_sym___bridge] = ACTIONS(2506),
    [anon_sym___bridge_retained] = ACTIONS(2506),
    [anon_sym___bridge_transfer] = ACTIONS(2506),
    [anon_sym___complex] = ACTIONS(2506),
    [anon_sym___const] = ACTIONS(2506),
    [anon_sym___imag] = ACTIONS(2506),
    [anon_sym___kindof] = ACTIONS(2506),
    [anon_sym___nonnull] = ACTIONS(2506),
    [anon_sym___nullable] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2506),
    [anon_sym___real] = ACTIONS(2506),
    [anon_sym___strong] = ACTIONS(2506),
    [anon_sym___unsafe_unretained] = ACTIONS(2506),
    [anon_sym___unused] = ACTIONS(2506),
    [anon_sym___weak] = ACTIONS(2506),
    [anon_sym_alignas] = ACTIONS(2506),
    [anon_sym__Alignas] = ACTIONS(2506),
    [sym_primitive_type] = ACTIONS(2506),
    [anon_sym_enum] = ACTIONS(2506),
    [anon_sym_struct] = ACTIONS(2506),
    [anon_sym_union] = ACTIONS(2506),
    [anon_sym_if] = ACTIONS(2506),
    [anon_sym_else] = ACTIONS(2506),
    [anon_sym_switch] = ACTIONS(2506),
    [anon_sym_case] = ACTIONS(2506),
    [anon_sym_default] = ACTIONS(2506),
    [anon_sym_while] = ACTIONS(2506),
    [anon_sym_do] = ACTIONS(2506),
    [anon_sym_for] = ACTIONS(2506),
    [anon_sym_in] = ACTIONS(2506),
    [anon_sym_return] = ACTIONS(2506),
    [anon_sym_break] = ACTIONS(2506),
    [anon_sym_continue] = ACTIONS(2506),
    [anon_sym_goto] = ACTIONS(2506),
    [anon_sym_DASH_DASH] = ACTIONS(2508),
    [anon_sym_PLUS_PLUS] = ACTIONS(2508),
    [anon_sym_sizeof] = ACTIONS(2506),
    [anon_sym___alignof__] = ACTIONS(2506),
    [anon_sym___alignof] = ACTIONS(2506),
    [anon_sym__alignof] = ACTIONS(2506),
    [anon_sym_alignof] = ACTIONS(2506),
    [anon_sym__Alignof] = ACTIONS(2506),
    [anon_sym_offsetof] = ACTIONS(2506),
    [anon_sym__Generic] = ACTIONS(2506),
    [anon_sym_asm] = ACTIONS(2506),
    [anon_sym___asm__] = ACTIONS(2506),
    [anon_sym___asm] = ACTIONS(2506),
    [sym_number_literal] = ACTIONS(2508),
    [anon_sym_L_SQUOTE] = ACTIONS(2508),
    [anon_sym_u_SQUOTE] = ACTIONS(2508),
    [anon_sym_U_SQUOTE] = ACTIONS(2508),
    [anon_sym_u8_SQUOTE] = ACTIONS(2508),
    [anon_sym_SQUOTE] = ACTIONS(2508),
    [anon_sym_AT] = ACTIONS(2506),
    [anon_sym_DQUOTE] = ACTIONS(2508),
    [anon_sym_L_DQUOTE] = ACTIONS(2508),
    [anon_sym_u_DQUOTE] = ACTIONS(2508),
    [anon_sym_U_DQUOTE] = ACTIONS(2508),
    [anon_sym_u8_DQUOTE] = ACTIONS(2508),
    [sym_true] = ACTIONS(2506),
    [sym_false] = ACTIONS(2506),
    [anon_sym_NULL] = ACTIONS(2506),
    [anon_sym_nullptr] = ACTIONS(2506),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2506),
    [anon_sym___typeof] = ACTIONS(2506),
    [anon_sym_typeof] = ACTIONS(2506),
    [anon_sym_ATimport] = ACTIONS(2508),
    [aux_sym_preproc_undef_token1] = ACTIONS(2506),
    [anon_sym_POUND] = ACTIONS(2506),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2506),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2506),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2506),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_API_AVAILABLE] = ACTIONS(2506),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_API_DEPRECATED] = ACTIONS(2506),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym___deprecated_msg] = ACTIONS(2506),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2506),
    [anon_sym_ATprotocol] = ACTIONS(2508),
    [anon_sym_ATinterface] = ACTIONS(2508),
    [anon_sym_ATimplementation] = ACTIONS(2508),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2508),
    [anon_sym_ATselector] = ACTIONS(2508),
    [anon_sym_ATavailable] = ACTIONS(2508),
    [anon_sym___builtin_available] = ACTIONS(2506),
    [anon_sym_va_arg] = ACTIONS(2506),
    [anon_sym_ATencode] = ACTIONS(2508),
    [anon_sym_BOOL] = ACTIONS(2506),
    [anon_sym_IMP] = ACTIONS(2506),
    [anon_sym_SEL] = ACTIONS(2506),
    [anon_sym_Class] = ACTIONS(2506),
    [anon_sym_id] = ACTIONS(2506),
  },
  [STATE(948)] = {
    [ts_builtin_sym_end] = ACTIONS(2488),
    [sym_identifier] = ACTIONS(2486),
    [aux_sym_preproc_include_token1] = ACTIONS(2486),
    [aux_sym_preproc_include_token2] = ACTIONS(2486),
    [aux_sym_preproc_def_token1] = ACTIONS(2486),
    [aux_sym_preproc_if_token1] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2486),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2486),
    [sym_preproc_directive] = ACTIONS(2486),
    [anon_sym_LPAREN2] = ACTIONS(2488),
    [anon_sym_BANG] = ACTIONS(2488),
    [anon_sym_TILDE] = ACTIONS(2488),
    [anon_sym_DASH] = ACTIONS(2486),
    [anon_sym_PLUS] = ACTIONS(2486),
    [anon_sym_STAR] = ACTIONS(2488),
    [anon_sym_CARET] = ACTIONS(2488),
    [anon_sym_AMP] = ACTIONS(2488),
    [anon_sym_SEMI] = ACTIONS(2488),
    [anon_sym___extension__] = ACTIONS(2486),
    [anon_sym_typedef] = ACTIONS(2486),
    [anon_sym_extern] = ACTIONS(2486),
    [anon_sym___attribute__] = ACTIONS(2486),
    [anon_sym___attribute] = ACTIONS(2486),
    [anon_sym_noreturn] = ACTIONS(2486),
    [anon_sym_LBRACK] = ACTIONS(2488),
    [anon_sym___declspec] = ACTIONS(2486),
    [anon_sym___cdecl] = ACTIONS(2486),
    [anon_sym___clrcall] = ACTIONS(2486),
    [anon_sym___stdcall] = ACTIONS(2486),
    [anon_sym___fastcall] = ACTIONS(2486),
    [anon_sym___thiscall] = ACTIONS(2486),
    [anon_sym___vectorcall] = ACTIONS(2486),
    [anon_sym_LBRACE] = ACTIONS(2488),
    [anon_sym_signed] = ACTIONS(2486),
    [anon_sym_unsigned] = ACTIONS(2486),
    [anon_sym_long] = ACTIONS(2486),
    [anon_sym_short] = ACTIONS(2486),
    [anon_sym_static] = ACTIONS(2486),
    [anon_sym_ATautoreleasepool] = ACTIONS(2488),
    [anon_sym_auto] = ACTIONS(2486),
    [anon_sym_register] = ACTIONS(2486),
    [anon_sym_inline] = ACTIONS(2486),
    [anon_sym___inline] = ACTIONS(2486),
    [anon_sym___inline__] = ACTIONS(2486),
    [anon_sym___forceinline] = ACTIONS(2486),
    [anon_sym_thread_local] = ACTIONS(2486),
    [anon_sym___thread] = ACTIONS(2486),
    [anon_sym_CG_EXTERN] = ACTIONS(2486),
    [anon_sym_CG_INLINE] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2486),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2486),
    [anon_sym_IBOutlet] = ACTIONS(2486),
    [anon_sym_IBInspectable] = ACTIONS(2486),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2486),
    [anon_sym_NS_INLINE] = ACTIONS(2486),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2486),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2486),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2486),
    [anon_sym_const] = ACTIONS(2486),
    [anon_sym_constexpr] = ACTIONS(2486),
    [anon_sym_volatile] = ACTIONS(2486),
    [anon_sym_restrict] = ACTIONS(2486),
    [anon_sym___restrict__] = ACTIONS(2486),
    [anon_sym__Atomic] = ACTIONS(2486),
    [anon_sym__Noreturn] = ACTIONS(2486),
    [anon_sym__Nonnull] = ACTIONS(2486),
    [anon_sym_nullable] = ACTIONS(2486),
    [anon_sym__Complex] = ACTIONS(2486),
    [anon_sym__Nullable] = ACTIONS(2486),
    [anon_sym__Nullable_result] = ACTIONS(2486),
    [anon_sym__Null_unspecified] = ACTIONS(2486),
    [anon_sym___autoreleasing] = ACTIONS(2486),
    [anon_sym___block] = ACTIONS(2486),
    [anon_sym___bridge] = ACTIONS(2486),
    [anon_sym___bridge_retained] = ACTIONS(2486),
    [anon_sym___bridge_transfer] = ACTIONS(2486),
    [anon_sym___complex] = ACTIONS(2486),
    [anon_sym___const] = ACTIONS(2486),
    [anon_sym___imag] = ACTIONS(2486),
    [anon_sym___kindof] = ACTIONS(2486),
    [anon_sym___nonnull] = ACTIONS(2486),
    [anon_sym___nullable] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2486),
    [anon_sym___real] = ACTIONS(2486),
    [anon_sym___strong] = ACTIONS(2486),
    [anon_sym___unsafe_unretained] = ACTIONS(2486),
    [anon_sym___unused] = ACTIONS(2486),
    [anon_sym___weak] = ACTIONS(2486),
    [anon_sym_alignas] = ACTIONS(2486),
    [anon_sym__Alignas] = ACTIONS(2486),
    [sym_primitive_type] = ACTIONS(2486),
    [anon_sym_enum] = ACTIONS(2486),
    [anon_sym_struct] = ACTIONS(2486),
    [anon_sym_union] = ACTIONS(2486),
    [anon_sym_if] = ACTIONS(2486),
    [anon_sym_else] = ACTIONS(2486),
    [anon_sym_switch] = ACTIONS(2486),
    [anon_sym_case] = ACTIONS(2486),
    [anon_sym_default] = ACTIONS(2486),
    [anon_sym_while] = ACTIONS(2486),
    [anon_sym_do] = ACTIONS(2486),
    [anon_sym_for] = ACTIONS(2486),
    [anon_sym_in] = ACTIONS(2486),
    [anon_sym_return] = ACTIONS(2486),
    [anon_sym_break] = ACTIONS(2486),
    [anon_sym_continue] = ACTIONS(2486),
    [anon_sym_goto] = ACTIONS(2486),
    [anon_sym_DASH_DASH] = ACTIONS(2488),
    [anon_sym_PLUS_PLUS] = ACTIONS(2488),
    [anon_sym_sizeof] = ACTIONS(2486),
    [anon_sym___alignof__] = ACTIONS(2486),
    [anon_sym___alignof] = ACTIONS(2486),
    [anon_sym__alignof] = ACTIONS(2486),
    [anon_sym_alignof] = ACTIONS(2486),
    [anon_sym__Alignof] = ACTIONS(2486),
    [anon_sym_offsetof] = ACTIONS(2486),
    [anon_sym__Generic] = ACTIONS(2486),
    [anon_sym_asm] = ACTIONS(2486),
    [anon_sym___asm__] = ACTIONS(2486),
    [anon_sym___asm] = ACTIONS(2486),
    [sym_number_literal] = ACTIONS(2488),
    [anon_sym_L_SQUOTE] = ACTIONS(2488),
    [anon_sym_u_SQUOTE] = ACTIONS(2488),
    [anon_sym_U_SQUOTE] = ACTIONS(2488),
    [anon_sym_u8_SQUOTE] = ACTIONS(2488),
    [anon_sym_SQUOTE] = ACTIONS(2488),
    [anon_sym_AT] = ACTIONS(2486),
    [anon_sym_DQUOTE] = ACTIONS(2488),
    [anon_sym_L_DQUOTE] = ACTIONS(2488),
    [anon_sym_u_DQUOTE] = ACTIONS(2488),
    [anon_sym_U_DQUOTE] = ACTIONS(2488),
    [anon_sym_u8_DQUOTE] = ACTIONS(2488),
    [sym_true] = ACTIONS(2486),
    [sym_false] = ACTIONS(2486),
    [anon_sym_NULL] = ACTIONS(2486),
    [anon_sym_nullptr] = ACTIONS(2486),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2486),
    [anon_sym___typeof] = ACTIONS(2486),
    [anon_sym_typeof] = ACTIONS(2486),
    [anon_sym_ATimport] = ACTIONS(2488),
    [aux_sym_preproc_undef_token1] = ACTIONS(2486),
    [anon_sym_POUND] = ACTIONS(2486),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2486),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2486),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2486),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_API_AVAILABLE] = ACTIONS(2486),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_API_DEPRECATED] = ACTIONS(2486),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym___deprecated_msg] = ACTIONS(2486),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2486),
    [anon_sym_ATprotocol] = ACTIONS(2488),
    [anon_sym_ATinterface] = ACTIONS(2488),
    [anon_sym_ATimplementation] = ACTIONS(2488),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2488),
    [anon_sym_ATselector] = ACTIONS(2488),
    [anon_sym_ATavailable] = ACTIONS(2488),
    [anon_sym___builtin_available] = ACTIONS(2486),
    [anon_sym_va_arg] = ACTIONS(2486),
    [anon_sym_ATencode] = ACTIONS(2488),
    [anon_sym_BOOL] = ACTIONS(2486),
    [anon_sym_IMP] = ACTIONS(2486),
    [anon_sym_SEL] = ACTIONS(2486),
    [anon_sym_Class] = ACTIONS(2486),
    [anon_sym_id] = ACTIONS(2486),
  },
  [STATE(949)] = {
    [ts_builtin_sym_end] = ACTIONS(2492),
    [sym_identifier] = ACTIONS(2490),
    [aux_sym_preproc_include_token1] = ACTIONS(2490),
    [aux_sym_preproc_include_token2] = ACTIONS(2490),
    [aux_sym_preproc_def_token1] = ACTIONS(2490),
    [aux_sym_preproc_if_token1] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2490),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2490),
    [sym_preproc_directive] = ACTIONS(2490),
    [anon_sym_LPAREN2] = ACTIONS(2492),
    [anon_sym_BANG] = ACTIONS(2492),
    [anon_sym_TILDE] = ACTIONS(2492),
    [anon_sym_DASH] = ACTIONS(2490),
    [anon_sym_PLUS] = ACTIONS(2490),
    [anon_sym_STAR] = ACTIONS(2492),
    [anon_sym_CARET] = ACTIONS(2492),
    [anon_sym_AMP] = ACTIONS(2492),
    [anon_sym_SEMI] = ACTIONS(2492),
    [anon_sym___extension__] = ACTIONS(2490),
    [anon_sym_typedef] = ACTIONS(2490),
    [anon_sym_extern] = ACTIONS(2490),
    [anon_sym___attribute__] = ACTIONS(2490),
    [anon_sym___attribute] = ACTIONS(2490),
    [anon_sym_noreturn] = ACTIONS(2490),
    [anon_sym_LBRACK] = ACTIONS(2492),
    [anon_sym___declspec] = ACTIONS(2490),
    [anon_sym___cdecl] = ACTIONS(2490),
    [anon_sym___clrcall] = ACTIONS(2490),
    [anon_sym___stdcall] = ACTIONS(2490),
    [anon_sym___fastcall] = ACTIONS(2490),
    [anon_sym___thiscall] = ACTIONS(2490),
    [anon_sym___vectorcall] = ACTIONS(2490),
    [anon_sym_LBRACE] = ACTIONS(2492),
    [anon_sym_signed] = ACTIONS(2490),
    [anon_sym_unsigned] = ACTIONS(2490),
    [anon_sym_long] = ACTIONS(2490),
    [anon_sym_short] = ACTIONS(2490),
    [anon_sym_static] = ACTIONS(2490),
    [anon_sym_ATautoreleasepool] = ACTIONS(2492),
    [anon_sym_auto] = ACTIONS(2490),
    [anon_sym_register] = ACTIONS(2490),
    [anon_sym_inline] = ACTIONS(2490),
    [anon_sym___inline] = ACTIONS(2490),
    [anon_sym___inline__] = ACTIONS(2490),
    [anon_sym___forceinline] = ACTIONS(2490),
    [anon_sym_thread_local] = ACTIONS(2490),
    [anon_sym___thread] = ACTIONS(2490),
    [anon_sym_CG_EXTERN] = ACTIONS(2490),
    [anon_sym_CG_INLINE] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2490),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2490),
    [anon_sym_IBOutlet] = ACTIONS(2490),
    [anon_sym_IBInspectable] = ACTIONS(2490),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2490),
    [anon_sym_NS_INLINE] = ACTIONS(2490),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2490),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2490),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2490),
    [anon_sym_const] = ACTIONS(2490),
    [anon_sym_constexpr] = ACTIONS(2490),
    [anon_sym_volatile] = ACTIONS(2490),
    [anon_sym_restrict] = ACTIONS(2490),
    [anon_sym___restrict__] = ACTIONS(2490),
    [anon_sym__Atomic] = ACTIONS(2490),
    [anon_sym__Noreturn] = ACTIONS(2490),
    [anon_sym__Nonnull] = ACTIONS(2490),
    [anon_sym_nullable] = ACTIONS(2490),
    [anon_sym__Complex] = ACTIONS(2490),
    [anon_sym__Nullable] = ACTIONS(2490),
    [anon_sym__Nullable_result] = ACTIONS(2490),
    [anon_sym__Null_unspecified] = ACTIONS(2490),
    [anon_sym___autoreleasing] = ACTIONS(2490),
    [anon_sym___block] = ACTIONS(2490),
    [anon_sym___bridge] = ACTIONS(2490),
    [anon_sym___bridge_retained] = ACTIONS(2490),
    [anon_sym___bridge_transfer] = ACTIONS(2490),
    [anon_sym___complex] = ACTIONS(2490),
    [anon_sym___const] = ACTIONS(2490),
    [anon_sym___imag] = ACTIONS(2490),
    [anon_sym___kindof] = ACTIONS(2490),
    [anon_sym___nonnull] = ACTIONS(2490),
    [anon_sym___nullable] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2490),
    [anon_sym___real] = ACTIONS(2490),
    [anon_sym___strong] = ACTIONS(2490),
    [anon_sym___unsafe_unretained] = ACTIONS(2490),
    [anon_sym___unused] = ACTIONS(2490),
    [anon_sym___weak] = ACTIONS(2490),
    [anon_sym_alignas] = ACTIONS(2490),
    [anon_sym__Alignas] = ACTIONS(2490),
    [sym_primitive_type] = ACTIONS(2490),
    [anon_sym_enum] = ACTIONS(2490),
    [anon_sym_struct] = ACTIONS(2490),
    [anon_sym_union] = ACTIONS(2490),
    [anon_sym_if] = ACTIONS(2490),
    [anon_sym_else] = ACTIONS(2490),
    [anon_sym_switch] = ACTIONS(2490),
    [anon_sym_case] = ACTIONS(2490),
    [anon_sym_default] = ACTIONS(2490),
    [anon_sym_while] = ACTIONS(2490),
    [anon_sym_do] = ACTIONS(2490),
    [anon_sym_for] = ACTIONS(2490),
    [anon_sym_in] = ACTIONS(2490),
    [anon_sym_return] = ACTIONS(2490),
    [anon_sym_break] = ACTIONS(2490),
    [anon_sym_continue] = ACTIONS(2490),
    [anon_sym_goto] = ACTIONS(2490),
    [anon_sym_DASH_DASH] = ACTIONS(2492),
    [anon_sym_PLUS_PLUS] = ACTIONS(2492),
    [anon_sym_sizeof] = ACTIONS(2490),
    [anon_sym___alignof__] = ACTIONS(2490),
    [anon_sym___alignof] = ACTIONS(2490),
    [anon_sym__alignof] = ACTIONS(2490),
    [anon_sym_alignof] = ACTIONS(2490),
    [anon_sym__Alignof] = ACTIONS(2490),
    [anon_sym_offsetof] = ACTIONS(2490),
    [anon_sym__Generic] = ACTIONS(2490),
    [anon_sym_asm] = ACTIONS(2490),
    [anon_sym___asm__] = ACTIONS(2490),
    [anon_sym___asm] = ACTIONS(2490),
    [sym_number_literal] = ACTIONS(2492),
    [anon_sym_L_SQUOTE] = ACTIONS(2492),
    [anon_sym_u_SQUOTE] = ACTIONS(2492),
    [anon_sym_U_SQUOTE] = ACTIONS(2492),
    [anon_sym_u8_SQUOTE] = ACTIONS(2492),
    [anon_sym_SQUOTE] = ACTIONS(2492),
    [anon_sym_AT] = ACTIONS(2490),
    [anon_sym_DQUOTE] = ACTIONS(2492),
    [anon_sym_L_DQUOTE] = ACTIONS(2492),
    [anon_sym_u_DQUOTE] = ACTIONS(2492),
    [anon_sym_U_DQUOTE] = ACTIONS(2492),
    [anon_sym_u8_DQUOTE] = ACTIONS(2492),
    [sym_true] = ACTIONS(2490),
    [sym_false] = ACTIONS(2490),
    [anon_sym_NULL] = ACTIONS(2490),
    [anon_sym_nullptr] = ACTIONS(2490),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2490),
    [anon_sym___typeof] = ACTIONS(2490),
    [anon_sym_typeof] = ACTIONS(2490),
    [anon_sym_ATimport] = ACTIONS(2492),
    [aux_sym_preproc_undef_token1] = ACTIONS(2490),
    [anon_sym_POUND] = ACTIONS(2490),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2490),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2490),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2490),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_API_AVAILABLE] = ACTIONS(2490),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_API_DEPRECATED] = ACTIONS(2490),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym___deprecated_msg] = ACTIONS(2490),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2490),
    [anon_sym_ATprotocol] = ACTIONS(2492),
    [anon_sym_ATinterface] = ACTIONS(2492),
    [anon_sym_ATimplementation] = ACTIONS(2492),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2492),
    [anon_sym_ATselector] = ACTIONS(2492),
    [anon_sym_ATavailable] = ACTIONS(2492),
    [anon_sym___builtin_available] = ACTIONS(2490),
    [anon_sym_va_arg] = ACTIONS(2490),
    [anon_sym_ATencode] = ACTIONS(2492),
    [anon_sym_BOOL] = ACTIONS(2490),
    [anon_sym_IMP] = ACTIONS(2490),
    [anon_sym_SEL] = ACTIONS(2490),
    [anon_sym_Class] = ACTIONS(2490),
    [anon_sym_id] = ACTIONS(2490),
  },
  [STATE(950)] = {
    [ts_builtin_sym_end] = ACTIONS(2476),
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(951)] = {
    [ts_builtin_sym_end] = ACTIONS(2570),
    [sym_identifier] = ACTIONS(2568),
    [aux_sym_preproc_include_token1] = ACTIONS(2568),
    [aux_sym_preproc_include_token2] = ACTIONS(2568),
    [aux_sym_preproc_def_token1] = ACTIONS(2568),
    [aux_sym_preproc_if_token1] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2568),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2568),
    [sym_preproc_directive] = ACTIONS(2568),
    [anon_sym_LPAREN2] = ACTIONS(2570),
    [anon_sym_BANG] = ACTIONS(2570),
    [anon_sym_TILDE] = ACTIONS(2570),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_PLUS] = ACTIONS(2568),
    [anon_sym_STAR] = ACTIONS(2570),
    [anon_sym_CARET] = ACTIONS(2570),
    [anon_sym_AMP] = ACTIONS(2570),
    [anon_sym_SEMI] = ACTIONS(2570),
    [anon_sym___extension__] = ACTIONS(2568),
    [anon_sym_typedef] = ACTIONS(2568),
    [anon_sym_extern] = ACTIONS(2568),
    [anon_sym___attribute__] = ACTIONS(2568),
    [anon_sym___attribute] = ACTIONS(2568),
    [anon_sym_noreturn] = ACTIONS(2568),
    [anon_sym_LBRACK] = ACTIONS(2570),
    [anon_sym___declspec] = ACTIONS(2568),
    [anon_sym___cdecl] = ACTIONS(2568),
    [anon_sym___clrcall] = ACTIONS(2568),
    [anon_sym___stdcall] = ACTIONS(2568),
    [anon_sym___fastcall] = ACTIONS(2568),
    [anon_sym___thiscall] = ACTIONS(2568),
    [anon_sym___vectorcall] = ACTIONS(2568),
    [anon_sym_LBRACE] = ACTIONS(2570),
    [anon_sym_signed] = ACTIONS(2568),
    [anon_sym_unsigned] = ACTIONS(2568),
    [anon_sym_long] = ACTIONS(2568),
    [anon_sym_short] = ACTIONS(2568),
    [anon_sym_static] = ACTIONS(2568),
    [anon_sym_ATautoreleasepool] = ACTIONS(2570),
    [anon_sym_auto] = ACTIONS(2568),
    [anon_sym_register] = ACTIONS(2568),
    [anon_sym_inline] = ACTIONS(2568),
    [anon_sym___inline] = ACTIONS(2568),
    [anon_sym___inline__] = ACTIONS(2568),
    [anon_sym___forceinline] = ACTIONS(2568),
    [anon_sym_thread_local] = ACTIONS(2568),
    [anon_sym___thread] = ACTIONS(2568),
    [anon_sym_CG_EXTERN] = ACTIONS(2568),
    [anon_sym_CG_INLINE] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2568),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2568),
    [anon_sym_IBOutlet] = ACTIONS(2568),
    [anon_sym_IBInspectable] = ACTIONS(2568),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2568),
    [anon_sym_NS_INLINE] = ACTIONS(2568),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2568),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2568),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2568),
    [anon_sym_const] = ACTIONS(2568),
    [anon_sym_constexpr] = ACTIONS(2568),
    [anon_sym_volatile] = ACTIONS(2568),
    [anon_sym_restrict] = ACTIONS(2568),
    [anon_sym___restrict__] = ACTIONS(2568),
    [anon_sym__Atomic] = ACTIONS(2568),
    [anon_sym__Noreturn] = ACTIONS(2568),
    [anon_sym__Nonnull] = ACTIONS(2568),
    [anon_sym_nullable] = ACTIONS(2568),
    [anon_sym__Complex] = ACTIONS(2568),
    [anon_sym__Nullable] = ACTIONS(2568),
    [anon_sym__Nullable_result] = ACTIONS(2568),
    [anon_sym__Null_unspecified] = ACTIONS(2568),
    [anon_sym___autoreleasing] = ACTIONS(2568),
    [anon_sym___block] = ACTIONS(2568),
    [anon_sym___bridge] = ACTIONS(2568),
    [anon_sym___bridge_retained] = ACTIONS(2568),
    [anon_sym___bridge_transfer] = ACTIONS(2568),
    [anon_sym___complex] = ACTIONS(2568),
    [anon_sym___const] = ACTIONS(2568),
    [anon_sym___imag] = ACTIONS(2568),
    [anon_sym___kindof] = ACTIONS(2568),
    [anon_sym___nonnull] = ACTIONS(2568),
    [anon_sym___nullable] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2568),
    [anon_sym___real] = ACTIONS(2568),
    [anon_sym___strong] = ACTIONS(2568),
    [anon_sym___unsafe_unretained] = ACTIONS(2568),
    [anon_sym___unused] = ACTIONS(2568),
    [anon_sym___weak] = ACTIONS(2568),
    [anon_sym_alignas] = ACTIONS(2568),
    [anon_sym__Alignas] = ACTIONS(2568),
    [sym_primitive_type] = ACTIONS(2568),
    [anon_sym_enum] = ACTIONS(2568),
    [anon_sym_struct] = ACTIONS(2568),
    [anon_sym_union] = ACTIONS(2568),
    [anon_sym_if] = ACTIONS(2568),
    [anon_sym_else] = ACTIONS(2568),
    [anon_sym_switch] = ACTIONS(2568),
    [anon_sym_case] = ACTIONS(2568),
    [anon_sym_default] = ACTIONS(2568),
    [anon_sym_while] = ACTIONS(2568),
    [anon_sym_do] = ACTIONS(2568),
    [anon_sym_for] = ACTIONS(2568),
    [anon_sym_in] = ACTIONS(2568),
    [anon_sym_return] = ACTIONS(2568),
    [anon_sym_break] = ACTIONS(2568),
    [anon_sym_continue] = ACTIONS(2568),
    [anon_sym_goto] = ACTIONS(2568),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_sizeof] = ACTIONS(2568),
    [anon_sym___alignof__] = ACTIONS(2568),
    [anon_sym___alignof] = ACTIONS(2568),
    [anon_sym__alignof] = ACTIONS(2568),
    [anon_sym_alignof] = ACTIONS(2568),
    [anon_sym__Alignof] = ACTIONS(2568),
    [anon_sym_offsetof] = ACTIONS(2568),
    [anon_sym__Generic] = ACTIONS(2568),
    [anon_sym_asm] = ACTIONS(2568),
    [anon_sym___asm__] = ACTIONS(2568),
    [anon_sym___asm] = ACTIONS(2568),
    [sym_number_literal] = ACTIONS(2570),
    [anon_sym_L_SQUOTE] = ACTIONS(2570),
    [anon_sym_u_SQUOTE] = ACTIONS(2570),
    [anon_sym_U_SQUOTE] = ACTIONS(2570),
    [anon_sym_u8_SQUOTE] = ACTIONS(2570),
    [anon_sym_SQUOTE] = ACTIONS(2570),
    [anon_sym_AT] = ACTIONS(2568),
    [anon_sym_DQUOTE] = ACTIONS(2570),
    [anon_sym_L_DQUOTE] = ACTIONS(2570),
    [anon_sym_u_DQUOTE] = ACTIONS(2570),
    [anon_sym_U_DQUOTE] = ACTIONS(2570),
    [anon_sym_u8_DQUOTE] = ACTIONS(2570),
    [sym_true] = ACTIONS(2568),
    [sym_false] = ACTIONS(2568),
    [anon_sym_NULL] = ACTIONS(2568),
    [anon_sym_nullptr] = ACTIONS(2568),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2568),
    [anon_sym___typeof] = ACTIONS(2568),
    [anon_sym_typeof] = ACTIONS(2568),
    [anon_sym_ATimport] = ACTIONS(2570),
    [aux_sym_preproc_undef_token1] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2568),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2568),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2568),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2568),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_API_AVAILABLE] = ACTIONS(2568),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_API_DEPRECATED] = ACTIONS(2568),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym___deprecated_msg] = ACTIONS(2568),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2568),
    [anon_sym_ATprotocol] = ACTIONS(2570),
    [anon_sym_ATinterface] = ACTIONS(2570),
    [anon_sym_ATimplementation] = ACTIONS(2570),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2570),
    [anon_sym_ATselector] = ACTIONS(2570),
    [anon_sym_ATavailable] = ACTIONS(2570),
    [anon_sym___builtin_available] = ACTIONS(2568),
    [anon_sym_va_arg] = ACTIONS(2568),
    [anon_sym_ATencode] = ACTIONS(2570),
    [anon_sym_BOOL] = ACTIONS(2568),
    [anon_sym_IMP] = ACTIONS(2568),
    [anon_sym_SEL] = ACTIONS(2568),
    [anon_sym_Class] = ACTIONS(2568),
    [anon_sym_id] = ACTIONS(2568),
  },
  [STATE(952)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(953)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(954)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(955)] = {
    [ts_builtin_sym_end] = ACTIONS(2456),
    [sym_identifier] = ACTIONS(2454),
    [aux_sym_preproc_include_token1] = ACTIONS(2454),
    [aux_sym_preproc_include_token2] = ACTIONS(2454),
    [aux_sym_preproc_def_token1] = ACTIONS(2454),
    [aux_sym_preproc_if_token1] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2454),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2454),
    [sym_preproc_directive] = ACTIONS(2454),
    [anon_sym_LPAREN2] = ACTIONS(2456),
    [anon_sym_BANG] = ACTIONS(2456),
    [anon_sym_TILDE] = ACTIONS(2456),
    [anon_sym_DASH] = ACTIONS(2454),
    [anon_sym_PLUS] = ACTIONS(2454),
    [anon_sym_STAR] = ACTIONS(2456),
    [anon_sym_CARET] = ACTIONS(2456),
    [anon_sym_AMP] = ACTIONS(2456),
    [anon_sym_SEMI] = ACTIONS(2456),
    [anon_sym___extension__] = ACTIONS(2454),
    [anon_sym_typedef] = ACTIONS(2454),
    [anon_sym_extern] = ACTIONS(2454),
    [anon_sym___attribute__] = ACTIONS(2454),
    [anon_sym___attribute] = ACTIONS(2454),
    [anon_sym_noreturn] = ACTIONS(2454),
    [anon_sym_LBRACK] = ACTIONS(2456),
    [anon_sym___declspec] = ACTIONS(2454),
    [anon_sym___cdecl] = ACTIONS(2454),
    [anon_sym___clrcall] = ACTIONS(2454),
    [anon_sym___stdcall] = ACTIONS(2454),
    [anon_sym___fastcall] = ACTIONS(2454),
    [anon_sym___thiscall] = ACTIONS(2454),
    [anon_sym___vectorcall] = ACTIONS(2454),
    [anon_sym_LBRACE] = ACTIONS(2456),
    [anon_sym_signed] = ACTIONS(2454),
    [anon_sym_unsigned] = ACTIONS(2454),
    [anon_sym_long] = ACTIONS(2454),
    [anon_sym_short] = ACTIONS(2454),
    [anon_sym_static] = ACTIONS(2454),
    [anon_sym_ATautoreleasepool] = ACTIONS(2456),
    [anon_sym_auto] = ACTIONS(2454),
    [anon_sym_register] = ACTIONS(2454),
    [anon_sym_inline] = ACTIONS(2454),
    [anon_sym___inline] = ACTIONS(2454),
    [anon_sym___inline__] = ACTIONS(2454),
    [anon_sym___forceinline] = ACTIONS(2454),
    [anon_sym_thread_local] = ACTIONS(2454),
    [anon_sym___thread] = ACTIONS(2454),
    [anon_sym_CG_EXTERN] = ACTIONS(2454),
    [anon_sym_CG_INLINE] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2454),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2454),
    [anon_sym_IBOutlet] = ACTIONS(2454),
    [anon_sym_IBInspectable] = ACTIONS(2454),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2454),
    [anon_sym_NS_INLINE] = ACTIONS(2454),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2454),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2454),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2454),
    [anon_sym_const] = ACTIONS(2454),
    [anon_sym_constexpr] = ACTIONS(2454),
    [anon_sym_volatile] = ACTIONS(2454),
    [anon_sym_restrict] = ACTIONS(2454),
    [anon_sym___restrict__] = ACTIONS(2454),
    [anon_sym__Atomic] = ACTIONS(2454),
    [anon_sym__Noreturn] = ACTIONS(2454),
    [anon_sym__Nonnull] = ACTIONS(2454),
    [anon_sym_nullable] = ACTIONS(2454),
    [anon_sym__Complex] = ACTIONS(2454),
    [anon_sym__Nullable] = ACTIONS(2454),
    [anon_sym__Nullable_result] = ACTIONS(2454),
    [anon_sym__Null_unspecified] = ACTIONS(2454),
    [anon_sym___autoreleasing] = ACTIONS(2454),
    [anon_sym___block] = ACTIONS(2454),
    [anon_sym___bridge] = ACTIONS(2454),
    [anon_sym___bridge_retained] = ACTIONS(2454),
    [anon_sym___bridge_transfer] = ACTIONS(2454),
    [anon_sym___complex] = ACTIONS(2454),
    [anon_sym___const] = ACTIONS(2454),
    [anon_sym___imag] = ACTIONS(2454),
    [anon_sym___kindof] = ACTIONS(2454),
    [anon_sym___nonnull] = ACTIONS(2454),
    [anon_sym___nullable] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2454),
    [anon_sym___real] = ACTIONS(2454),
    [anon_sym___strong] = ACTIONS(2454),
    [anon_sym___unsafe_unretained] = ACTIONS(2454),
    [anon_sym___unused] = ACTIONS(2454),
    [anon_sym___weak] = ACTIONS(2454),
    [anon_sym_alignas] = ACTIONS(2454),
    [anon_sym__Alignas] = ACTIONS(2454),
    [sym_primitive_type] = ACTIONS(2454),
    [anon_sym_enum] = ACTIONS(2454),
    [anon_sym_struct] = ACTIONS(2454),
    [anon_sym_union] = ACTIONS(2454),
    [anon_sym_if] = ACTIONS(2454),
    [anon_sym_else] = ACTIONS(2454),
    [anon_sym_switch] = ACTIONS(2454),
    [anon_sym_case] = ACTIONS(2454),
    [anon_sym_default] = ACTIONS(2454),
    [anon_sym_while] = ACTIONS(2454),
    [anon_sym_do] = ACTIONS(2454),
    [anon_sym_for] = ACTIONS(2454),
    [anon_sym_in] = ACTIONS(2454),
    [anon_sym_return] = ACTIONS(2454),
    [anon_sym_break] = ACTIONS(2454),
    [anon_sym_continue] = ACTIONS(2454),
    [anon_sym_goto] = ACTIONS(2454),
    [anon_sym_DASH_DASH] = ACTIONS(2456),
    [anon_sym_PLUS_PLUS] = ACTIONS(2456),
    [anon_sym_sizeof] = ACTIONS(2454),
    [anon_sym___alignof__] = ACTIONS(2454),
    [anon_sym___alignof] = ACTIONS(2454),
    [anon_sym__alignof] = ACTIONS(2454),
    [anon_sym_alignof] = ACTIONS(2454),
    [anon_sym__Alignof] = ACTIONS(2454),
    [anon_sym_offsetof] = ACTIONS(2454),
    [anon_sym__Generic] = ACTIONS(2454),
    [anon_sym_asm] = ACTIONS(2454),
    [anon_sym___asm__] = ACTIONS(2454),
    [anon_sym___asm] = ACTIONS(2454),
    [sym_number_literal] = ACTIONS(2456),
    [anon_sym_L_SQUOTE] = ACTIONS(2456),
    [anon_sym_u_SQUOTE] = ACTIONS(2456),
    [anon_sym_U_SQUOTE] = ACTIONS(2456),
    [anon_sym_u8_SQUOTE] = ACTIONS(2456),
    [anon_sym_SQUOTE] = ACTIONS(2456),
    [anon_sym_AT] = ACTIONS(2454),
    [anon_sym_DQUOTE] = ACTIONS(2456),
    [anon_sym_L_DQUOTE] = ACTIONS(2456),
    [anon_sym_u_DQUOTE] = ACTIONS(2456),
    [anon_sym_U_DQUOTE] = ACTIONS(2456),
    [anon_sym_u8_DQUOTE] = ACTIONS(2456),
    [sym_true] = ACTIONS(2454),
    [sym_false] = ACTIONS(2454),
    [anon_sym_NULL] = ACTIONS(2454),
    [anon_sym_nullptr] = ACTIONS(2454),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2454),
    [anon_sym___typeof] = ACTIONS(2454),
    [anon_sym_typeof] = ACTIONS(2454),
    [anon_sym_ATimport] = ACTIONS(2456),
    [aux_sym_preproc_undef_token1] = ACTIONS(2454),
    [anon_sym_POUND] = ACTIONS(2454),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2454),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2454),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2454),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_API_AVAILABLE] = ACTIONS(2454),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_API_DEPRECATED] = ACTIONS(2454),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym___deprecated_msg] = ACTIONS(2454),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2454),
    [anon_sym_ATprotocol] = ACTIONS(2456),
    [anon_sym_ATinterface] = ACTIONS(2456),
    [anon_sym_ATimplementation] = ACTIONS(2456),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2456),
    [anon_sym_ATselector] = ACTIONS(2456),
    [anon_sym_ATavailable] = ACTIONS(2456),
    [anon_sym___builtin_available] = ACTIONS(2454),
    [anon_sym_va_arg] = ACTIONS(2454),
    [anon_sym_ATencode] = ACTIONS(2456),
    [anon_sym_BOOL] = ACTIONS(2454),
    [anon_sym_IMP] = ACTIONS(2454),
    [anon_sym_SEL] = ACTIONS(2454),
    [anon_sym_Class] = ACTIONS(2454),
    [anon_sym_id] = ACTIONS(2454),
  },
  [STATE(956)] = {
    [ts_builtin_sym_end] = ACTIONS(2480),
    [sym_identifier] = ACTIONS(2478),
    [aux_sym_preproc_include_token1] = ACTIONS(2478),
    [aux_sym_preproc_include_token2] = ACTIONS(2478),
    [aux_sym_preproc_def_token1] = ACTIONS(2478),
    [aux_sym_preproc_if_token1] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2478),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2478),
    [sym_preproc_directive] = ACTIONS(2478),
    [anon_sym_LPAREN2] = ACTIONS(2480),
    [anon_sym_BANG] = ACTIONS(2480),
    [anon_sym_TILDE] = ACTIONS(2480),
    [anon_sym_DASH] = ACTIONS(2478),
    [anon_sym_PLUS] = ACTIONS(2478),
    [anon_sym_STAR] = ACTIONS(2480),
    [anon_sym_CARET] = ACTIONS(2480),
    [anon_sym_AMP] = ACTIONS(2480),
    [anon_sym_SEMI] = ACTIONS(2480),
    [anon_sym___extension__] = ACTIONS(2478),
    [anon_sym_typedef] = ACTIONS(2478),
    [anon_sym_extern] = ACTIONS(2478),
    [anon_sym___attribute__] = ACTIONS(2478),
    [anon_sym___attribute] = ACTIONS(2478),
    [anon_sym_noreturn] = ACTIONS(2478),
    [anon_sym_LBRACK] = ACTIONS(2480),
    [anon_sym___declspec] = ACTIONS(2478),
    [anon_sym___cdecl] = ACTIONS(2478),
    [anon_sym___clrcall] = ACTIONS(2478),
    [anon_sym___stdcall] = ACTIONS(2478),
    [anon_sym___fastcall] = ACTIONS(2478),
    [anon_sym___thiscall] = ACTIONS(2478),
    [anon_sym___vectorcall] = ACTIONS(2478),
    [anon_sym_LBRACE] = ACTIONS(2480),
    [anon_sym_signed] = ACTIONS(2478),
    [anon_sym_unsigned] = ACTIONS(2478),
    [anon_sym_long] = ACTIONS(2478),
    [anon_sym_short] = ACTIONS(2478),
    [anon_sym_static] = ACTIONS(2478),
    [anon_sym_ATautoreleasepool] = ACTIONS(2480),
    [anon_sym_auto] = ACTIONS(2478),
    [anon_sym_register] = ACTIONS(2478),
    [anon_sym_inline] = ACTIONS(2478),
    [anon_sym___inline] = ACTIONS(2478),
    [anon_sym___inline__] = ACTIONS(2478),
    [anon_sym___forceinline] = ACTIONS(2478),
    [anon_sym_thread_local] = ACTIONS(2478),
    [anon_sym___thread] = ACTIONS(2478),
    [anon_sym_CG_EXTERN] = ACTIONS(2478),
    [anon_sym_CG_INLINE] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2478),
    [anon_sym_IBOutlet] = ACTIONS(2478),
    [anon_sym_IBInspectable] = ACTIONS(2478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2478),
    [anon_sym_NS_INLINE] = ACTIONS(2478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2478),
    [anon_sym_const] = ACTIONS(2478),
    [anon_sym_constexpr] = ACTIONS(2478),
    [anon_sym_volatile] = ACTIONS(2478),
    [anon_sym_restrict] = ACTIONS(2478),
    [anon_sym___restrict__] = ACTIONS(2478),
    [anon_sym__Atomic] = ACTIONS(2478),
    [anon_sym__Noreturn] = ACTIONS(2478),
    [anon_sym__Nonnull] = ACTIONS(2478),
    [anon_sym_nullable] = ACTIONS(2478),
    [anon_sym__Complex] = ACTIONS(2478),
    [anon_sym__Nullable] = ACTIONS(2478),
    [anon_sym__Nullable_result] = ACTIONS(2478),
    [anon_sym__Null_unspecified] = ACTIONS(2478),
    [anon_sym___autoreleasing] = ACTIONS(2478),
    [anon_sym___block] = ACTIONS(2478),
    [anon_sym___bridge] = ACTIONS(2478),
    [anon_sym___bridge_retained] = ACTIONS(2478),
    [anon_sym___bridge_transfer] = ACTIONS(2478),
    [anon_sym___complex] = ACTIONS(2478),
    [anon_sym___const] = ACTIONS(2478),
    [anon_sym___imag] = ACTIONS(2478),
    [anon_sym___kindof] = ACTIONS(2478),
    [anon_sym___nonnull] = ACTIONS(2478),
    [anon_sym___nullable] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2478),
    [anon_sym___real] = ACTIONS(2478),
    [anon_sym___strong] = ACTIONS(2478),
    [anon_sym___unsafe_unretained] = ACTIONS(2478),
    [anon_sym___unused] = ACTIONS(2478),
    [anon_sym___weak] = ACTIONS(2478),
    [anon_sym_alignas] = ACTIONS(2478),
    [anon_sym__Alignas] = ACTIONS(2478),
    [sym_primitive_type] = ACTIONS(2478),
    [anon_sym_enum] = ACTIONS(2478),
    [anon_sym_struct] = ACTIONS(2478),
    [anon_sym_union] = ACTIONS(2478),
    [anon_sym_if] = ACTIONS(2478),
    [anon_sym_else] = ACTIONS(2478),
    [anon_sym_switch] = ACTIONS(2478),
    [anon_sym_case] = ACTIONS(2478),
    [anon_sym_default] = ACTIONS(2478),
    [anon_sym_while] = ACTIONS(2478),
    [anon_sym_do] = ACTIONS(2478),
    [anon_sym_for] = ACTIONS(2478),
    [anon_sym_in] = ACTIONS(2478),
    [anon_sym_return] = ACTIONS(2478),
    [anon_sym_break] = ACTIONS(2478),
    [anon_sym_continue] = ACTIONS(2478),
    [anon_sym_goto] = ACTIONS(2478),
    [anon_sym_DASH_DASH] = ACTIONS(2480),
    [anon_sym_PLUS_PLUS] = ACTIONS(2480),
    [anon_sym_sizeof] = ACTIONS(2478),
    [anon_sym___alignof__] = ACTIONS(2478),
    [anon_sym___alignof] = ACTIONS(2478),
    [anon_sym__alignof] = ACTIONS(2478),
    [anon_sym_alignof] = ACTIONS(2478),
    [anon_sym__Alignof] = ACTIONS(2478),
    [anon_sym_offsetof] = ACTIONS(2478),
    [anon_sym__Generic] = ACTIONS(2478),
    [anon_sym_asm] = ACTIONS(2478),
    [anon_sym___asm__] = ACTIONS(2478),
    [anon_sym___asm] = ACTIONS(2478),
    [sym_number_literal] = ACTIONS(2480),
    [anon_sym_L_SQUOTE] = ACTIONS(2480),
    [anon_sym_u_SQUOTE] = ACTIONS(2480),
    [anon_sym_U_SQUOTE] = ACTIONS(2480),
    [anon_sym_u8_SQUOTE] = ACTIONS(2480),
    [anon_sym_SQUOTE] = ACTIONS(2480),
    [anon_sym_AT] = ACTIONS(2478),
    [anon_sym_DQUOTE] = ACTIONS(2480),
    [anon_sym_L_DQUOTE] = ACTIONS(2480),
    [anon_sym_u_DQUOTE] = ACTIONS(2480),
    [anon_sym_U_DQUOTE] = ACTIONS(2480),
    [anon_sym_u8_DQUOTE] = ACTIONS(2480),
    [sym_true] = ACTIONS(2478),
    [sym_false] = ACTIONS(2478),
    [anon_sym_NULL] = ACTIONS(2478),
    [anon_sym_nullptr] = ACTIONS(2478),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2478),
    [anon_sym___typeof] = ACTIONS(2478),
    [anon_sym_typeof] = ACTIONS(2478),
    [anon_sym_ATimport] = ACTIONS(2480),
    [aux_sym_preproc_undef_token1] = ACTIONS(2478),
    [anon_sym_POUND] = ACTIONS(2478),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_API_AVAILABLE] = ACTIONS(2478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_API_DEPRECATED] = ACTIONS(2478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym___deprecated_msg] = ACTIONS(2478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2478),
    [anon_sym_ATprotocol] = ACTIONS(2480),
    [anon_sym_ATinterface] = ACTIONS(2480),
    [anon_sym_ATimplementation] = ACTIONS(2480),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2480),
    [anon_sym_ATselector] = ACTIONS(2480),
    [anon_sym_ATavailable] = ACTIONS(2480),
    [anon_sym___builtin_available] = ACTIONS(2478),
    [anon_sym_va_arg] = ACTIONS(2478),
    [anon_sym_ATencode] = ACTIONS(2480),
    [anon_sym_BOOL] = ACTIONS(2478),
    [anon_sym_IMP] = ACTIONS(2478),
    [anon_sym_SEL] = ACTIONS(2478),
    [anon_sym_Class] = ACTIONS(2478),
    [anon_sym_id] = ACTIONS(2478),
  },
  [STATE(957)] = {
    [ts_builtin_sym_end] = ACTIONS(2504),
    [sym_identifier] = ACTIONS(2502),
    [aux_sym_preproc_include_token1] = ACTIONS(2502),
    [aux_sym_preproc_include_token2] = ACTIONS(2502),
    [aux_sym_preproc_def_token1] = ACTIONS(2502),
    [aux_sym_preproc_if_token1] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2502),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2502),
    [sym_preproc_directive] = ACTIONS(2502),
    [anon_sym_LPAREN2] = ACTIONS(2504),
    [anon_sym_BANG] = ACTIONS(2504),
    [anon_sym_TILDE] = ACTIONS(2504),
    [anon_sym_DASH] = ACTIONS(2502),
    [anon_sym_PLUS] = ACTIONS(2502),
    [anon_sym_STAR] = ACTIONS(2504),
    [anon_sym_CARET] = ACTIONS(2504),
    [anon_sym_AMP] = ACTIONS(2504),
    [anon_sym_SEMI] = ACTIONS(2504),
    [anon_sym___extension__] = ACTIONS(2502),
    [anon_sym_typedef] = ACTIONS(2502),
    [anon_sym_extern] = ACTIONS(2502),
    [anon_sym___attribute__] = ACTIONS(2502),
    [anon_sym___attribute] = ACTIONS(2502),
    [anon_sym_noreturn] = ACTIONS(2502),
    [anon_sym_LBRACK] = ACTIONS(2504),
    [anon_sym___declspec] = ACTIONS(2502),
    [anon_sym___cdecl] = ACTIONS(2502),
    [anon_sym___clrcall] = ACTIONS(2502),
    [anon_sym___stdcall] = ACTIONS(2502),
    [anon_sym___fastcall] = ACTIONS(2502),
    [anon_sym___thiscall] = ACTIONS(2502),
    [anon_sym___vectorcall] = ACTIONS(2502),
    [anon_sym_LBRACE] = ACTIONS(2504),
    [anon_sym_signed] = ACTIONS(2502),
    [anon_sym_unsigned] = ACTIONS(2502),
    [anon_sym_long] = ACTIONS(2502),
    [anon_sym_short] = ACTIONS(2502),
    [anon_sym_static] = ACTIONS(2502),
    [anon_sym_ATautoreleasepool] = ACTIONS(2504),
    [anon_sym_auto] = ACTIONS(2502),
    [anon_sym_register] = ACTIONS(2502),
    [anon_sym_inline] = ACTIONS(2502),
    [anon_sym___inline] = ACTIONS(2502),
    [anon_sym___inline__] = ACTIONS(2502),
    [anon_sym___forceinline] = ACTIONS(2502),
    [anon_sym_thread_local] = ACTIONS(2502),
    [anon_sym___thread] = ACTIONS(2502),
    [anon_sym_CG_EXTERN] = ACTIONS(2502),
    [anon_sym_CG_INLINE] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2502),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2502),
    [anon_sym_IBOutlet] = ACTIONS(2502),
    [anon_sym_IBInspectable] = ACTIONS(2502),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2502),
    [anon_sym_NS_INLINE] = ACTIONS(2502),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2502),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2502),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2502),
    [anon_sym_const] = ACTIONS(2502),
    [anon_sym_constexpr] = ACTIONS(2502),
    [anon_sym_volatile] = ACTIONS(2502),
    [anon_sym_restrict] = ACTIONS(2502),
    [anon_sym___restrict__] = ACTIONS(2502),
    [anon_sym__Atomic] = ACTIONS(2502),
    [anon_sym__Noreturn] = ACTIONS(2502),
    [anon_sym__Nonnull] = ACTIONS(2502),
    [anon_sym_nullable] = ACTIONS(2502),
    [anon_sym__Complex] = ACTIONS(2502),
    [anon_sym__Nullable] = ACTIONS(2502),
    [anon_sym__Nullable_result] = ACTIONS(2502),
    [anon_sym__Null_unspecified] = ACTIONS(2502),
    [anon_sym___autoreleasing] = ACTIONS(2502),
    [anon_sym___block] = ACTIONS(2502),
    [anon_sym___bridge] = ACTIONS(2502),
    [anon_sym___bridge_retained] = ACTIONS(2502),
    [anon_sym___bridge_transfer] = ACTIONS(2502),
    [anon_sym___complex] = ACTIONS(2502),
    [anon_sym___const] = ACTIONS(2502),
    [anon_sym___imag] = ACTIONS(2502),
    [anon_sym___kindof] = ACTIONS(2502),
    [anon_sym___nonnull] = ACTIONS(2502),
    [anon_sym___nullable] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2502),
    [anon_sym___real] = ACTIONS(2502),
    [anon_sym___strong] = ACTIONS(2502),
    [anon_sym___unsafe_unretained] = ACTIONS(2502),
    [anon_sym___unused] = ACTIONS(2502),
    [anon_sym___weak] = ACTIONS(2502),
    [anon_sym_alignas] = ACTIONS(2502),
    [anon_sym__Alignas] = ACTIONS(2502),
    [sym_primitive_type] = ACTIONS(2502),
    [anon_sym_enum] = ACTIONS(2502),
    [anon_sym_struct] = ACTIONS(2502),
    [anon_sym_union] = ACTIONS(2502),
    [anon_sym_if] = ACTIONS(2502),
    [anon_sym_else] = ACTIONS(2502),
    [anon_sym_switch] = ACTIONS(2502),
    [anon_sym_case] = ACTIONS(2502),
    [anon_sym_default] = ACTIONS(2502),
    [anon_sym_while] = ACTIONS(2502),
    [anon_sym_do] = ACTIONS(2502),
    [anon_sym_for] = ACTIONS(2502),
    [anon_sym_in] = ACTIONS(2502),
    [anon_sym_return] = ACTIONS(2502),
    [anon_sym_break] = ACTIONS(2502),
    [anon_sym_continue] = ACTIONS(2502),
    [anon_sym_goto] = ACTIONS(2502),
    [anon_sym_DASH_DASH] = ACTIONS(2504),
    [anon_sym_PLUS_PLUS] = ACTIONS(2504),
    [anon_sym_sizeof] = ACTIONS(2502),
    [anon_sym___alignof__] = ACTIONS(2502),
    [anon_sym___alignof] = ACTIONS(2502),
    [anon_sym__alignof] = ACTIONS(2502),
    [anon_sym_alignof] = ACTIONS(2502),
    [anon_sym__Alignof] = ACTIONS(2502),
    [anon_sym_offsetof] = ACTIONS(2502),
    [anon_sym__Generic] = ACTIONS(2502),
    [anon_sym_asm] = ACTIONS(2502),
    [anon_sym___asm__] = ACTIONS(2502),
    [anon_sym___asm] = ACTIONS(2502),
    [sym_number_literal] = ACTIONS(2504),
    [anon_sym_L_SQUOTE] = ACTIONS(2504),
    [anon_sym_u_SQUOTE] = ACTIONS(2504),
    [anon_sym_U_SQUOTE] = ACTIONS(2504),
    [anon_sym_u8_SQUOTE] = ACTIONS(2504),
    [anon_sym_SQUOTE] = ACTIONS(2504),
    [anon_sym_AT] = ACTIONS(2502),
    [anon_sym_DQUOTE] = ACTIONS(2504),
    [anon_sym_L_DQUOTE] = ACTIONS(2504),
    [anon_sym_u_DQUOTE] = ACTIONS(2504),
    [anon_sym_U_DQUOTE] = ACTIONS(2504),
    [anon_sym_u8_DQUOTE] = ACTIONS(2504),
    [sym_true] = ACTIONS(2502),
    [sym_false] = ACTIONS(2502),
    [anon_sym_NULL] = ACTIONS(2502),
    [anon_sym_nullptr] = ACTIONS(2502),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2502),
    [anon_sym___typeof] = ACTIONS(2502),
    [anon_sym_typeof] = ACTIONS(2502),
    [anon_sym_ATimport] = ACTIONS(2504),
    [aux_sym_preproc_undef_token1] = ACTIONS(2502),
    [anon_sym_POUND] = ACTIONS(2502),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2502),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2502),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2502),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_API_AVAILABLE] = ACTIONS(2502),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_API_DEPRECATED] = ACTIONS(2502),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym___deprecated_msg] = ACTIONS(2502),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2502),
    [anon_sym_ATprotocol] = ACTIONS(2504),
    [anon_sym_ATinterface] = ACTIONS(2504),
    [anon_sym_ATimplementation] = ACTIONS(2504),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2504),
    [anon_sym_ATselector] = ACTIONS(2504),
    [anon_sym_ATavailable] = ACTIONS(2504),
    [anon_sym___builtin_available] = ACTIONS(2502),
    [anon_sym_va_arg] = ACTIONS(2502),
    [anon_sym_ATencode] = ACTIONS(2504),
    [anon_sym_BOOL] = ACTIONS(2502),
    [anon_sym_IMP] = ACTIONS(2502),
    [anon_sym_SEL] = ACTIONS(2502),
    [anon_sym_Class] = ACTIONS(2502),
    [anon_sym_id] = ACTIONS(2502),
  },
  [STATE(958)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(959)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(960)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(961)] = {
    [ts_builtin_sym_end] = ACTIONS(2418),
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(962)] = {
    [ts_builtin_sym_end] = ACTIONS(2452),
    [sym_identifier] = ACTIONS(2450),
    [aux_sym_preproc_include_token1] = ACTIONS(2450),
    [aux_sym_preproc_include_token2] = ACTIONS(2450),
    [aux_sym_preproc_def_token1] = ACTIONS(2450),
    [aux_sym_preproc_if_token1] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2450),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2450),
    [sym_preproc_directive] = ACTIONS(2450),
    [anon_sym_LPAREN2] = ACTIONS(2452),
    [anon_sym_BANG] = ACTIONS(2452),
    [anon_sym_TILDE] = ACTIONS(2452),
    [anon_sym_DASH] = ACTIONS(2450),
    [anon_sym_PLUS] = ACTIONS(2450),
    [anon_sym_STAR] = ACTIONS(2452),
    [anon_sym_CARET] = ACTIONS(2452),
    [anon_sym_AMP] = ACTIONS(2452),
    [anon_sym_SEMI] = ACTIONS(2452),
    [anon_sym___extension__] = ACTIONS(2450),
    [anon_sym_typedef] = ACTIONS(2450),
    [anon_sym_extern] = ACTIONS(2450),
    [anon_sym___attribute__] = ACTIONS(2450),
    [anon_sym___attribute] = ACTIONS(2450),
    [anon_sym_noreturn] = ACTIONS(2450),
    [anon_sym_LBRACK] = ACTIONS(2452),
    [anon_sym___declspec] = ACTIONS(2450),
    [anon_sym___cdecl] = ACTIONS(2450),
    [anon_sym___clrcall] = ACTIONS(2450),
    [anon_sym___stdcall] = ACTIONS(2450),
    [anon_sym___fastcall] = ACTIONS(2450),
    [anon_sym___thiscall] = ACTIONS(2450),
    [anon_sym___vectorcall] = ACTIONS(2450),
    [anon_sym_LBRACE] = ACTIONS(2452),
    [anon_sym_signed] = ACTIONS(2450),
    [anon_sym_unsigned] = ACTIONS(2450),
    [anon_sym_long] = ACTIONS(2450),
    [anon_sym_short] = ACTIONS(2450),
    [anon_sym_static] = ACTIONS(2450),
    [anon_sym_ATautoreleasepool] = ACTIONS(2452),
    [anon_sym_auto] = ACTIONS(2450),
    [anon_sym_register] = ACTIONS(2450),
    [anon_sym_inline] = ACTIONS(2450),
    [anon_sym___inline] = ACTIONS(2450),
    [anon_sym___inline__] = ACTIONS(2450),
    [anon_sym___forceinline] = ACTIONS(2450),
    [anon_sym_thread_local] = ACTIONS(2450),
    [anon_sym___thread] = ACTIONS(2450),
    [anon_sym_CG_EXTERN] = ACTIONS(2450),
    [anon_sym_CG_INLINE] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2450),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2450),
    [anon_sym_IBOutlet] = ACTIONS(2450),
    [anon_sym_IBInspectable] = ACTIONS(2450),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2450),
    [anon_sym_NS_INLINE] = ACTIONS(2450),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2450),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2450),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2450),
    [anon_sym_const] = ACTIONS(2450),
    [anon_sym_constexpr] = ACTIONS(2450),
    [anon_sym_volatile] = ACTIONS(2450),
    [anon_sym_restrict] = ACTIONS(2450),
    [anon_sym___restrict__] = ACTIONS(2450),
    [anon_sym__Atomic] = ACTIONS(2450),
    [anon_sym__Noreturn] = ACTIONS(2450),
    [anon_sym__Nonnull] = ACTIONS(2450),
    [anon_sym_nullable] = ACTIONS(2450),
    [anon_sym__Complex] = ACTIONS(2450),
    [anon_sym__Nullable] = ACTIONS(2450),
    [anon_sym__Nullable_result] = ACTIONS(2450),
    [anon_sym__Null_unspecified] = ACTIONS(2450),
    [anon_sym___autoreleasing] = ACTIONS(2450),
    [anon_sym___block] = ACTIONS(2450),
    [anon_sym___bridge] = ACTIONS(2450),
    [anon_sym___bridge_retained] = ACTIONS(2450),
    [anon_sym___bridge_transfer] = ACTIONS(2450),
    [anon_sym___complex] = ACTIONS(2450),
    [anon_sym___const] = ACTIONS(2450),
    [anon_sym___imag] = ACTIONS(2450),
    [anon_sym___kindof] = ACTIONS(2450),
    [anon_sym___nonnull] = ACTIONS(2450),
    [anon_sym___nullable] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2450),
    [anon_sym___real] = ACTIONS(2450),
    [anon_sym___strong] = ACTIONS(2450),
    [anon_sym___unsafe_unretained] = ACTIONS(2450),
    [anon_sym___unused] = ACTIONS(2450),
    [anon_sym___weak] = ACTIONS(2450),
    [anon_sym_alignas] = ACTIONS(2450),
    [anon_sym__Alignas] = ACTIONS(2450),
    [sym_primitive_type] = ACTIONS(2450),
    [anon_sym_enum] = ACTIONS(2450),
    [anon_sym_struct] = ACTIONS(2450),
    [anon_sym_union] = ACTIONS(2450),
    [anon_sym_if] = ACTIONS(2450),
    [anon_sym_else] = ACTIONS(2450),
    [anon_sym_switch] = ACTIONS(2450),
    [anon_sym_case] = ACTIONS(2450),
    [anon_sym_default] = ACTIONS(2450),
    [anon_sym_while] = ACTIONS(2450),
    [anon_sym_do] = ACTIONS(2450),
    [anon_sym_for] = ACTIONS(2450),
    [anon_sym_in] = ACTIONS(2450),
    [anon_sym_return] = ACTIONS(2450),
    [anon_sym_break] = ACTIONS(2450),
    [anon_sym_continue] = ACTIONS(2450),
    [anon_sym_goto] = ACTIONS(2450),
    [anon_sym_DASH_DASH] = ACTIONS(2452),
    [anon_sym_PLUS_PLUS] = ACTIONS(2452),
    [anon_sym_sizeof] = ACTIONS(2450),
    [anon_sym___alignof__] = ACTIONS(2450),
    [anon_sym___alignof] = ACTIONS(2450),
    [anon_sym__alignof] = ACTIONS(2450),
    [anon_sym_alignof] = ACTIONS(2450),
    [anon_sym__Alignof] = ACTIONS(2450),
    [anon_sym_offsetof] = ACTIONS(2450),
    [anon_sym__Generic] = ACTIONS(2450),
    [anon_sym_asm] = ACTIONS(2450),
    [anon_sym___asm__] = ACTIONS(2450),
    [anon_sym___asm] = ACTIONS(2450),
    [sym_number_literal] = ACTIONS(2452),
    [anon_sym_L_SQUOTE] = ACTIONS(2452),
    [anon_sym_u_SQUOTE] = ACTIONS(2452),
    [anon_sym_U_SQUOTE] = ACTIONS(2452),
    [anon_sym_u8_SQUOTE] = ACTIONS(2452),
    [anon_sym_SQUOTE] = ACTIONS(2452),
    [anon_sym_AT] = ACTIONS(2450),
    [anon_sym_DQUOTE] = ACTIONS(2452),
    [anon_sym_L_DQUOTE] = ACTIONS(2452),
    [anon_sym_u_DQUOTE] = ACTIONS(2452),
    [anon_sym_U_DQUOTE] = ACTIONS(2452),
    [anon_sym_u8_DQUOTE] = ACTIONS(2452),
    [sym_true] = ACTIONS(2450),
    [sym_false] = ACTIONS(2450),
    [anon_sym_NULL] = ACTIONS(2450),
    [anon_sym_nullptr] = ACTIONS(2450),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2450),
    [anon_sym___typeof] = ACTIONS(2450),
    [anon_sym_typeof] = ACTIONS(2450),
    [anon_sym_ATimport] = ACTIONS(2452),
    [aux_sym_preproc_undef_token1] = ACTIONS(2450),
    [anon_sym_POUND] = ACTIONS(2450),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2450),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2450),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2450),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_API_AVAILABLE] = ACTIONS(2450),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_API_DEPRECATED] = ACTIONS(2450),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym___deprecated_msg] = ACTIONS(2450),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2450),
    [anon_sym_ATprotocol] = ACTIONS(2452),
    [anon_sym_ATinterface] = ACTIONS(2452),
    [anon_sym_ATimplementation] = ACTIONS(2452),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2452),
    [anon_sym_ATselector] = ACTIONS(2452),
    [anon_sym_ATavailable] = ACTIONS(2452),
    [anon_sym___builtin_available] = ACTIONS(2450),
    [anon_sym_va_arg] = ACTIONS(2450),
    [anon_sym_ATencode] = ACTIONS(2452),
    [anon_sym_BOOL] = ACTIONS(2450),
    [anon_sym_IMP] = ACTIONS(2450),
    [anon_sym_SEL] = ACTIONS(2450),
    [anon_sym_Class] = ACTIONS(2450),
    [anon_sym_id] = ACTIONS(2450),
  },
  [STATE(963)] = {
    [ts_builtin_sym_end] = ACTIONS(2422),
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(964)] = {
    [ts_builtin_sym_end] = ACTIONS(2308),
    [sym_identifier] = ACTIONS(2306),
    [aux_sym_preproc_include_token1] = ACTIONS(2306),
    [aux_sym_preproc_include_token2] = ACTIONS(2306),
    [aux_sym_preproc_def_token1] = ACTIONS(2306),
    [aux_sym_preproc_if_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2306),
    [sym_preproc_directive] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym___cdecl] = ACTIONS(2306),
    [anon_sym___clrcall] = ACTIONS(2306),
    [anon_sym___stdcall] = ACTIONS(2306),
    [anon_sym___fastcall] = ACTIONS(2306),
    [anon_sym___thiscall] = ACTIONS(2306),
    [anon_sym___vectorcall] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_case] = ACTIONS(2306),
    [anon_sym_default] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_ATimport] = ACTIONS(2308),
    [aux_sym_preproc_undef_token1] = ACTIONS(2306),
    [anon_sym_POUND] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATprotocol] = ACTIONS(2308),
    [anon_sym_ATinterface] = ACTIONS(2308),
    [anon_sym_ATimplementation] = ACTIONS(2308),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(965)] = {
    [ts_builtin_sym_end] = ACTIONS(2426),
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(966)] = {
    [ts_builtin_sym_end] = ACTIONS(2500),
    [sym_identifier] = ACTIONS(2498),
    [aux_sym_preproc_include_token1] = ACTIONS(2498),
    [aux_sym_preproc_include_token2] = ACTIONS(2498),
    [aux_sym_preproc_def_token1] = ACTIONS(2498),
    [aux_sym_preproc_if_token1] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2498),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2498),
    [sym_preproc_directive] = ACTIONS(2498),
    [anon_sym_LPAREN2] = ACTIONS(2500),
    [anon_sym_BANG] = ACTIONS(2500),
    [anon_sym_TILDE] = ACTIONS(2500),
    [anon_sym_DASH] = ACTIONS(2498),
    [anon_sym_PLUS] = ACTIONS(2498),
    [anon_sym_STAR] = ACTIONS(2500),
    [anon_sym_CARET] = ACTIONS(2500),
    [anon_sym_AMP] = ACTIONS(2500),
    [anon_sym_SEMI] = ACTIONS(2500),
    [anon_sym___extension__] = ACTIONS(2498),
    [anon_sym_typedef] = ACTIONS(2498),
    [anon_sym_extern] = ACTIONS(2498),
    [anon_sym___attribute__] = ACTIONS(2498),
    [anon_sym___attribute] = ACTIONS(2498),
    [anon_sym_noreturn] = ACTIONS(2498),
    [anon_sym_LBRACK] = ACTIONS(2500),
    [anon_sym___declspec] = ACTIONS(2498),
    [anon_sym___cdecl] = ACTIONS(2498),
    [anon_sym___clrcall] = ACTIONS(2498),
    [anon_sym___stdcall] = ACTIONS(2498),
    [anon_sym___fastcall] = ACTIONS(2498),
    [anon_sym___thiscall] = ACTIONS(2498),
    [anon_sym___vectorcall] = ACTIONS(2498),
    [anon_sym_LBRACE] = ACTIONS(2500),
    [anon_sym_signed] = ACTIONS(2498),
    [anon_sym_unsigned] = ACTIONS(2498),
    [anon_sym_long] = ACTIONS(2498),
    [anon_sym_short] = ACTIONS(2498),
    [anon_sym_static] = ACTIONS(2498),
    [anon_sym_ATautoreleasepool] = ACTIONS(2500),
    [anon_sym_auto] = ACTIONS(2498),
    [anon_sym_register] = ACTIONS(2498),
    [anon_sym_inline] = ACTIONS(2498),
    [anon_sym___inline] = ACTIONS(2498),
    [anon_sym___inline__] = ACTIONS(2498),
    [anon_sym___forceinline] = ACTIONS(2498),
    [anon_sym_thread_local] = ACTIONS(2498),
    [anon_sym___thread] = ACTIONS(2498),
    [anon_sym_CG_EXTERN] = ACTIONS(2498),
    [anon_sym_CG_INLINE] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2498),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2498),
    [anon_sym_IBOutlet] = ACTIONS(2498),
    [anon_sym_IBInspectable] = ACTIONS(2498),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2498),
    [anon_sym_NS_INLINE] = ACTIONS(2498),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2498),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2498),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2498),
    [anon_sym_const] = ACTIONS(2498),
    [anon_sym_constexpr] = ACTIONS(2498),
    [anon_sym_volatile] = ACTIONS(2498),
    [anon_sym_restrict] = ACTIONS(2498),
    [anon_sym___restrict__] = ACTIONS(2498),
    [anon_sym__Atomic] = ACTIONS(2498),
    [anon_sym__Noreturn] = ACTIONS(2498),
    [anon_sym__Nonnull] = ACTIONS(2498),
    [anon_sym_nullable] = ACTIONS(2498),
    [anon_sym__Complex] = ACTIONS(2498),
    [anon_sym__Nullable] = ACTIONS(2498),
    [anon_sym__Nullable_result] = ACTIONS(2498),
    [anon_sym__Null_unspecified] = ACTIONS(2498),
    [anon_sym___autoreleasing] = ACTIONS(2498),
    [anon_sym___block] = ACTIONS(2498),
    [anon_sym___bridge] = ACTIONS(2498),
    [anon_sym___bridge_retained] = ACTIONS(2498),
    [anon_sym___bridge_transfer] = ACTIONS(2498),
    [anon_sym___complex] = ACTIONS(2498),
    [anon_sym___const] = ACTIONS(2498),
    [anon_sym___imag] = ACTIONS(2498),
    [anon_sym___kindof] = ACTIONS(2498),
    [anon_sym___nonnull] = ACTIONS(2498),
    [anon_sym___nullable] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2498),
    [anon_sym___real] = ACTIONS(2498),
    [anon_sym___strong] = ACTIONS(2498),
    [anon_sym___unsafe_unretained] = ACTIONS(2498),
    [anon_sym___unused] = ACTIONS(2498),
    [anon_sym___weak] = ACTIONS(2498),
    [anon_sym_alignas] = ACTIONS(2498),
    [anon_sym__Alignas] = ACTIONS(2498),
    [sym_primitive_type] = ACTIONS(2498),
    [anon_sym_enum] = ACTIONS(2498),
    [anon_sym_struct] = ACTIONS(2498),
    [anon_sym_union] = ACTIONS(2498),
    [anon_sym_if] = ACTIONS(2498),
    [anon_sym_else] = ACTIONS(2498),
    [anon_sym_switch] = ACTIONS(2498),
    [anon_sym_case] = ACTIONS(2498),
    [anon_sym_default] = ACTIONS(2498),
    [anon_sym_while] = ACTIONS(2498),
    [anon_sym_do] = ACTIONS(2498),
    [anon_sym_for] = ACTIONS(2498),
    [anon_sym_in] = ACTIONS(2498),
    [anon_sym_return] = ACTIONS(2498),
    [anon_sym_break] = ACTIONS(2498),
    [anon_sym_continue] = ACTIONS(2498),
    [anon_sym_goto] = ACTIONS(2498),
    [anon_sym_DASH_DASH] = ACTIONS(2500),
    [anon_sym_PLUS_PLUS] = ACTIONS(2500),
    [anon_sym_sizeof] = ACTIONS(2498),
    [anon_sym___alignof__] = ACTIONS(2498),
    [anon_sym___alignof] = ACTIONS(2498),
    [anon_sym__alignof] = ACTIONS(2498),
    [anon_sym_alignof] = ACTIONS(2498),
    [anon_sym__Alignof] = ACTIONS(2498),
    [anon_sym_offsetof] = ACTIONS(2498),
    [anon_sym__Generic] = ACTIONS(2498),
    [anon_sym_asm] = ACTIONS(2498),
    [anon_sym___asm__] = ACTIONS(2498),
    [anon_sym___asm] = ACTIONS(2498),
    [sym_number_literal] = ACTIONS(2500),
    [anon_sym_L_SQUOTE] = ACTIONS(2500),
    [anon_sym_u_SQUOTE] = ACTIONS(2500),
    [anon_sym_U_SQUOTE] = ACTIONS(2500),
    [anon_sym_u8_SQUOTE] = ACTIONS(2500),
    [anon_sym_SQUOTE] = ACTIONS(2500),
    [anon_sym_AT] = ACTIONS(2498),
    [anon_sym_DQUOTE] = ACTIONS(2500),
    [anon_sym_L_DQUOTE] = ACTIONS(2500),
    [anon_sym_u_DQUOTE] = ACTIONS(2500),
    [anon_sym_U_DQUOTE] = ACTIONS(2500),
    [anon_sym_u8_DQUOTE] = ACTIONS(2500),
    [sym_true] = ACTIONS(2498),
    [sym_false] = ACTIONS(2498),
    [anon_sym_NULL] = ACTIONS(2498),
    [anon_sym_nullptr] = ACTIONS(2498),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2498),
    [anon_sym___typeof] = ACTIONS(2498),
    [anon_sym_typeof] = ACTIONS(2498),
    [anon_sym_ATimport] = ACTIONS(2500),
    [aux_sym_preproc_undef_token1] = ACTIONS(2498),
    [anon_sym_POUND] = ACTIONS(2498),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2498),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2498),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2498),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_API_AVAILABLE] = ACTIONS(2498),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_API_DEPRECATED] = ACTIONS(2498),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym___deprecated_msg] = ACTIONS(2498),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2498),
    [anon_sym_ATprotocol] = ACTIONS(2500),
    [anon_sym_ATinterface] = ACTIONS(2500),
    [anon_sym_ATimplementation] = ACTIONS(2500),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2500),
    [anon_sym_ATselector] = ACTIONS(2500),
    [anon_sym_ATavailable] = ACTIONS(2500),
    [anon_sym___builtin_available] = ACTIONS(2498),
    [anon_sym_va_arg] = ACTIONS(2498),
    [anon_sym_ATencode] = ACTIONS(2500),
    [anon_sym_BOOL] = ACTIONS(2498),
    [anon_sym_IMP] = ACTIONS(2498),
    [anon_sym_SEL] = ACTIONS(2498),
    [anon_sym_Class] = ACTIONS(2498),
    [anon_sym_id] = ACTIONS(2498),
  },
  [STATE(967)] = {
    [ts_builtin_sym_end] = ACTIONS(2476),
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_include_token1] = ACTIONS(2474),
    [aux_sym_preproc_include_token2] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_case] = ACTIONS(2474),
    [anon_sym_default] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_ATimport] = ACTIONS(2476),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_POUND] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATprotocol] = ACTIONS(2476),
    [anon_sym_ATinterface] = ACTIONS(2476),
    [anon_sym_ATimplementation] = ACTIONS(2476),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(968)] = {
    [ts_builtin_sym_end] = ACTIONS(2418),
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_include_token1] = ACTIONS(2416),
    [aux_sym_preproc_include_token2] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_case] = ACTIONS(2416),
    [anon_sym_default] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_ATimport] = ACTIONS(2418),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_POUND] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATprotocol] = ACTIONS(2418),
    [anon_sym_ATinterface] = ACTIONS(2418),
    [anon_sym_ATimplementation] = ACTIONS(2418),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(969)] = {
    [ts_builtin_sym_end] = ACTIONS(2426),
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_include_token1] = ACTIONS(2424),
    [aux_sym_preproc_include_token2] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_case] = ACTIONS(2424),
    [anon_sym_default] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_ATimport] = ACTIONS(2426),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_POUND] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATprotocol] = ACTIONS(2426),
    [anon_sym_ATinterface] = ACTIONS(2426),
    [anon_sym_ATimplementation] = ACTIONS(2426),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(970)] = {
    [ts_builtin_sym_end] = ACTIONS(2422),
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_include_token1] = ACTIONS(2420),
    [aux_sym_preproc_include_token2] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_case] = ACTIONS(2420),
    [anon_sym_default] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_ATimport] = ACTIONS(2422),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_POUND] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATprotocol] = ACTIONS(2422),
    [anon_sym_ATinterface] = ACTIONS(2422),
    [anon_sym_ATimplementation] = ACTIONS(2422),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(971)] = {
    [ts_builtin_sym_end] = ACTIONS(2524),
    [sym_identifier] = ACTIONS(2522),
    [aux_sym_preproc_include_token1] = ACTIONS(2522),
    [aux_sym_preproc_include_token2] = ACTIONS(2522),
    [aux_sym_preproc_def_token1] = ACTIONS(2522),
    [aux_sym_preproc_if_token1] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2522),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2522),
    [sym_preproc_directive] = ACTIONS(2522),
    [anon_sym_LPAREN2] = ACTIONS(2524),
    [anon_sym_BANG] = ACTIONS(2524),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_DASH] = ACTIONS(2522),
    [anon_sym_PLUS] = ACTIONS(2522),
    [anon_sym_STAR] = ACTIONS(2524),
    [anon_sym_CARET] = ACTIONS(2524),
    [anon_sym_AMP] = ACTIONS(2524),
    [anon_sym_SEMI] = ACTIONS(2524),
    [anon_sym___extension__] = ACTIONS(2522),
    [anon_sym_typedef] = ACTIONS(2522),
    [anon_sym_extern] = ACTIONS(2522),
    [anon_sym___attribute__] = ACTIONS(2522),
    [anon_sym___attribute] = ACTIONS(2522),
    [anon_sym_noreturn] = ACTIONS(2522),
    [anon_sym_LBRACK] = ACTIONS(2524),
    [anon_sym___declspec] = ACTIONS(2522),
    [anon_sym___cdecl] = ACTIONS(2522),
    [anon_sym___clrcall] = ACTIONS(2522),
    [anon_sym___stdcall] = ACTIONS(2522),
    [anon_sym___fastcall] = ACTIONS(2522),
    [anon_sym___thiscall] = ACTIONS(2522),
    [anon_sym___vectorcall] = ACTIONS(2522),
    [anon_sym_LBRACE] = ACTIONS(2524),
    [anon_sym_signed] = ACTIONS(2522),
    [anon_sym_unsigned] = ACTIONS(2522),
    [anon_sym_long] = ACTIONS(2522),
    [anon_sym_short] = ACTIONS(2522),
    [anon_sym_static] = ACTIONS(2522),
    [anon_sym_ATautoreleasepool] = ACTIONS(2524),
    [anon_sym_auto] = ACTIONS(2522),
    [anon_sym_register] = ACTIONS(2522),
    [anon_sym_inline] = ACTIONS(2522),
    [anon_sym___inline] = ACTIONS(2522),
    [anon_sym___inline__] = ACTIONS(2522),
    [anon_sym___forceinline] = ACTIONS(2522),
    [anon_sym_thread_local] = ACTIONS(2522),
    [anon_sym___thread] = ACTIONS(2522),
    [anon_sym_CG_EXTERN] = ACTIONS(2522),
    [anon_sym_CG_INLINE] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2522),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2522),
    [anon_sym_IBOutlet] = ACTIONS(2522),
    [anon_sym_IBInspectable] = ACTIONS(2522),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2522),
    [anon_sym_NS_INLINE] = ACTIONS(2522),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2522),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2522),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2522),
    [anon_sym_const] = ACTIONS(2522),
    [anon_sym_constexpr] = ACTIONS(2522),
    [anon_sym_volatile] = ACTIONS(2522),
    [anon_sym_restrict] = ACTIONS(2522),
    [anon_sym___restrict__] = ACTIONS(2522),
    [anon_sym__Atomic] = ACTIONS(2522),
    [anon_sym__Noreturn] = ACTIONS(2522),
    [anon_sym__Nonnull] = ACTIONS(2522),
    [anon_sym_nullable] = ACTIONS(2522),
    [anon_sym__Complex] = ACTIONS(2522),
    [anon_sym__Nullable] = ACTIONS(2522),
    [anon_sym__Nullable_result] = ACTIONS(2522),
    [anon_sym__Null_unspecified] = ACTIONS(2522),
    [anon_sym___autoreleasing] = ACTIONS(2522),
    [anon_sym___block] = ACTIONS(2522),
    [anon_sym___bridge] = ACTIONS(2522),
    [anon_sym___bridge_retained] = ACTIONS(2522),
    [anon_sym___bridge_transfer] = ACTIONS(2522),
    [anon_sym___complex] = ACTIONS(2522),
    [anon_sym___const] = ACTIONS(2522),
    [anon_sym___imag] = ACTIONS(2522),
    [anon_sym___kindof] = ACTIONS(2522),
    [anon_sym___nonnull] = ACTIONS(2522),
    [anon_sym___nullable] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2522),
    [anon_sym___real] = ACTIONS(2522),
    [anon_sym___strong] = ACTIONS(2522),
    [anon_sym___unsafe_unretained] = ACTIONS(2522),
    [anon_sym___unused] = ACTIONS(2522),
    [anon_sym___weak] = ACTIONS(2522),
    [anon_sym_alignas] = ACTIONS(2522),
    [anon_sym__Alignas] = ACTIONS(2522),
    [sym_primitive_type] = ACTIONS(2522),
    [anon_sym_enum] = ACTIONS(2522),
    [anon_sym_struct] = ACTIONS(2522),
    [anon_sym_union] = ACTIONS(2522),
    [anon_sym_if] = ACTIONS(2522),
    [anon_sym_else] = ACTIONS(2522),
    [anon_sym_switch] = ACTIONS(2522),
    [anon_sym_case] = ACTIONS(2522),
    [anon_sym_default] = ACTIONS(2522),
    [anon_sym_while] = ACTIONS(2522),
    [anon_sym_do] = ACTIONS(2522),
    [anon_sym_for] = ACTIONS(2522),
    [anon_sym_in] = ACTIONS(2522),
    [anon_sym_return] = ACTIONS(2522),
    [anon_sym_break] = ACTIONS(2522),
    [anon_sym_continue] = ACTIONS(2522),
    [anon_sym_goto] = ACTIONS(2522),
    [anon_sym_DASH_DASH] = ACTIONS(2524),
    [anon_sym_PLUS_PLUS] = ACTIONS(2524),
    [anon_sym_sizeof] = ACTIONS(2522),
    [anon_sym___alignof__] = ACTIONS(2522),
    [anon_sym___alignof] = ACTIONS(2522),
    [anon_sym__alignof] = ACTIONS(2522),
    [anon_sym_alignof] = ACTIONS(2522),
    [anon_sym__Alignof] = ACTIONS(2522),
    [anon_sym_offsetof] = ACTIONS(2522),
    [anon_sym__Generic] = ACTIONS(2522),
    [anon_sym_asm] = ACTIONS(2522),
    [anon_sym___asm__] = ACTIONS(2522),
    [anon_sym___asm] = ACTIONS(2522),
    [sym_number_literal] = ACTIONS(2524),
    [anon_sym_L_SQUOTE] = ACTIONS(2524),
    [anon_sym_u_SQUOTE] = ACTIONS(2524),
    [anon_sym_U_SQUOTE] = ACTIONS(2524),
    [anon_sym_u8_SQUOTE] = ACTIONS(2524),
    [anon_sym_SQUOTE] = ACTIONS(2524),
    [anon_sym_AT] = ACTIONS(2522),
    [anon_sym_DQUOTE] = ACTIONS(2524),
    [anon_sym_L_DQUOTE] = ACTIONS(2524),
    [anon_sym_u_DQUOTE] = ACTIONS(2524),
    [anon_sym_U_DQUOTE] = ACTIONS(2524),
    [anon_sym_u8_DQUOTE] = ACTIONS(2524),
    [sym_true] = ACTIONS(2522),
    [sym_false] = ACTIONS(2522),
    [anon_sym_NULL] = ACTIONS(2522),
    [anon_sym_nullptr] = ACTIONS(2522),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2522),
    [anon_sym___typeof] = ACTIONS(2522),
    [anon_sym_typeof] = ACTIONS(2522),
    [anon_sym_ATimport] = ACTIONS(2524),
    [aux_sym_preproc_undef_token1] = ACTIONS(2522),
    [anon_sym_POUND] = ACTIONS(2522),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2522),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2522),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2522),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_API_AVAILABLE] = ACTIONS(2522),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_API_DEPRECATED] = ACTIONS(2522),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym___deprecated_msg] = ACTIONS(2522),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2522),
    [anon_sym_ATprotocol] = ACTIONS(2524),
    [anon_sym_ATinterface] = ACTIONS(2524),
    [anon_sym_ATimplementation] = ACTIONS(2524),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2524),
    [anon_sym_ATselector] = ACTIONS(2524),
    [anon_sym_ATavailable] = ACTIONS(2524),
    [anon_sym___builtin_available] = ACTIONS(2522),
    [anon_sym_va_arg] = ACTIONS(2522),
    [anon_sym_ATencode] = ACTIONS(2524),
    [anon_sym_BOOL] = ACTIONS(2522),
    [anon_sym_IMP] = ACTIONS(2522),
    [anon_sym_SEL] = ACTIONS(2522),
    [anon_sym_Class] = ACTIONS(2522),
    [anon_sym_id] = ACTIONS(2522),
  },
  [STATE(972)] = {
    [ts_builtin_sym_end] = ACTIONS(2448),
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_include_token1] = ACTIONS(2446),
    [aux_sym_preproc_include_token2] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_case] = ACTIONS(2446),
    [anon_sym_default] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_ATimport] = ACTIONS(2448),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_POUND] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATprotocol] = ACTIONS(2448),
    [anon_sym_ATinterface] = ACTIONS(2448),
    [anon_sym_ATimplementation] = ACTIONS(2448),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(973)] = {
    [ts_builtin_sym_end] = ACTIONS(2440),
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_include_token1] = ACTIONS(2438),
    [aux_sym_preproc_include_token2] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_case] = ACTIONS(2438),
    [anon_sym_default] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_ATimport] = ACTIONS(2440),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_POUND] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATprotocol] = ACTIONS(2440),
    [anon_sym_ATinterface] = ACTIONS(2440),
    [anon_sym_ATimplementation] = ACTIONS(2440),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(974)] = {
    [ts_builtin_sym_end] = ACTIONS(2444),
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_include_token1] = ACTIONS(2442),
    [aux_sym_preproc_include_token2] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_case] = ACTIONS(2442),
    [anon_sym_default] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_ATimport] = ACTIONS(2444),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_POUND] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATprotocol] = ACTIONS(2444),
    [anon_sym_ATinterface] = ACTIONS(2444),
    [anon_sym_ATimplementation] = ACTIONS(2444),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(975)] = {
    [ts_builtin_sym_end] = ACTIONS(2484),
    [sym_identifier] = ACTIONS(2482),
    [aux_sym_preproc_include_token1] = ACTIONS(2482),
    [aux_sym_preproc_include_token2] = ACTIONS(2482),
    [aux_sym_preproc_def_token1] = ACTIONS(2482),
    [aux_sym_preproc_if_token1] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2482),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2482),
    [sym_preproc_directive] = ACTIONS(2482),
    [anon_sym_LPAREN2] = ACTIONS(2484),
    [anon_sym_BANG] = ACTIONS(2484),
    [anon_sym_TILDE] = ACTIONS(2484),
    [anon_sym_DASH] = ACTIONS(2482),
    [anon_sym_PLUS] = ACTIONS(2482),
    [anon_sym_STAR] = ACTIONS(2484),
    [anon_sym_CARET] = ACTIONS(2484),
    [anon_sym_AMP] = ACTIONS(2484),
    [anon_sym_SEMI] = ACTIONS(2484),
    [anon_sym___extension__] = ACTIONS(2482),
    [anon_sym_typedef] = ACTIONS(2482),
    [anon_sym_extern] = ACTIONS(2482),
    [anon_sym___attribute__] = ACTIONS(2482),
    [anon_sym___attribute] = ACTIONS(2482),
    [anon_sym_noreturn] = ACTIONS(2482),
    [anon_sym_LBRACK] = ACTIONS(2484),
    [anon_sym___declspec] = ACTIONS(2482),
    [anon_sym___cdecl] = ACTIONS(2482),
    [anon_sym___clrcall] = ACTIONS(2482),
    [anon_sym___stdcall] = ACTIONS(2482),
    [anon_sym___fastcall] = ACTIONS(2482),
    [anon_sym___thiscall] = ACTIONS(2482),
    [anon_sym___vectorcall] = ACTIONS(2482),
    [anon_sym_LBRACE] = ACTIONS(2484),
    [anon_sym_signed] = ACTIONS(2482),
    [anon_sym_unsigned] = ACTIONS(2482),
    [anon_sym_long] = ACTIONS(2482),
    [anon_sym_short] = ACTIONS(2482),
    [anon_sym_static] = ACTIONS(2482),
    [anon_sym_ATautoreleasepool] = ACTIONS(2484),
    [anon_sym_auto] = ACTIONS(2482),
    [anon_sym_register] = ACTIONS(2482),
    [anon_sym_inline] = ACTIONS(2482),
    [anon_sym___inline] = ACTIONS(2482),
    [anon_sym___inline__] = ACTIONS(2482),
    [anon_sym___forceinline] = ACTIONS(2482),
    [anon_sym_thread_local] = ACTIONS(2482),
    [anon_sym___thread] = ACTIONS(2482),
    [anon_sym_CG_EXTERN] = ACTIONS(2482),
    [anon_sym_CG_INLINE] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2482),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2482),
    [anon_sym_IBOutlet] = ACTIONS(2482),
    [anon_sym_IBInspectable] = ACTIONS(2482),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2482),
    [anon_sym_NS_INLINE] = ACTIONS(2482),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2482),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2482),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2482),
    [anon_sym_const] = ACTIONS(2482),
    [anon_sym_constexpr] = ACTIONS(2482),
    [anon_sym_volatile] = ACTIONS(2482),
    [anon_sym_restrict] = ACTIONS(2482),
    [anon_sym___restrict__] = ACTIONS(2482),
    [anon_sym__Atomic] = ACTIONS(2482),
    [anon_sym__Noreturn] = ACTIONS(2482),
    [anon_sym__Nonnull] = ACTIONS(2482),
    [anon_sym_nullable] = ACTIONS(2482),
    [anon_sym__Complex] = ACTIONS(2482),
    [anon_sym__Nullable] = ACTIONS(2482),
    [anon_sym__Nullable_result] = ACTIONS(2482),
    [anon_sym__Null_unspecified] = ACTIONS(2482),
    [anon_sym___autoreleasing] = ACTIONS(2482),
    [anon_sym___block] = ACTIONS(2482),
    [anon_sym___bridge] = ACTIONS(2482),
    [anon_sym___bridge_retained] = ACTIONS(2482),
    [anon_sym___bridge_transfer] = ACTIONS(2482),
    [anon_sym___complex] = ACTIONS(2482),
    [anon_sym___const] = ACTIONS(2482),
    [anon_sym___imag] = ACTIONS(2482),
    [anon_sym___kindof] = ACTIONS(2482),
    [anon_sym___nonnull] = ACTIONS(2482),
    [anon_sym___nullable] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2482),
    [anon_sym___real] = ACTIONS(2482),
    [anon_sym___strong] = ACTIONS(2482),
    [anon_sym___unsafe_unretained] = ACTIONS(2482),
    [anon_sym___unused] = ACTIONS(2482),
    [anon_sym___weak] = ACTIONS(2482),
    [anon_sym_alignas] = ACTIONS(2482),
    [anon_sym__Alignas] = ACTIONS(2482),
    [sym_primitive_type] = ACTIONS(2482),
    [anon_sym_enum] = ACTIONS(2482),
    [anon_sym_struct] = ACTIONS(2482),
    [anon_sym_union] = ACTIONS(2482),
    [anon_sym_if] = ACTIONS(2482),
    [anon_sym_else] = ACTIONS(2482),
    [anon_sym_switch] = ACTIONS(2482),
    [anon_sym_case] = ACTIONS(2482),
    [anon_sym_default] = ACTIONS(2482),
    [anon_sym_while] = ACTIONS(2482),
    [anon_sym_do] = ACTIONS(2482),
    [anon_sym_for] = ACTIONS(2482),
    [anon_sym_in] = ACTIONS(2482),
    [anon_sym_return] = ACTIONS(2482),
    [anon_sym_break] = ACTIONS(2482),
    [anon_sym_continue] = ACTIONS(2482),
    [anon_sym_goto] = ACTIONS(2482),
    [anon_sym_DASH_DASH] = ACTIONS(2484),
    [anon_sym_PLUS_PLUS] = ACTIONS(2484),
    [anon_sym_sizeof] = ACTIONS(2482),
    [anon_sym___alignof__] = ACTIONS(2482),
    [anon_sym___alignof] = ACTIONS(2482),
    [anon_sym__alignof] = ACTIONS(2482),
    [anon_sym_alignof] = ACTIONS(2482),
    [anon_sym__Alignof] = ACTIONS(2482),
    [anon_sym_offsetof] = ACTIONS(2482),
    [anon_sym__Generic] = ACTIONS(2482),
    [anon_sym_asm] = ACTIONS(2482),
    [anon_sym___asm__] = ACTIONS(2482),
    [anon_sym___asm] = ACTIONS(2482),
    [sym_number_literal] = ACTIONS(2484),
    [anon_sym_L_SQUOTE] = ACTIONS(2484),
    [anon_sym_u_SQUOTE] = ACTIONS(2484),
    [anon_sym_U_SQUOTE] = ACTIONS(2484),
    [anon_sym_u8_SQUOTE] = ACTIONS(2484),
    [anon_sym_SQUOTE] = ACTIONS(2484),
    [anon_sym_AT] = ACTIONS(2482),
    [anon_sym_DQUOTE] = ACTIONS(2484),
    [anon_sym_L_DQUOTE] = ACTIONS(2484),
    [anon_sym_u_DQUOTE] = ACTIONS(2484),
    [anon_sym_U_DQUOTE] = ACTIONS(2484),
    [anon_sym_u8_DQUOTE] = ACTIONS(2484),
    [sym_true] = ACTIONS(2482),
    [sym_false] = ACTIONS(2482),
    [anon_sym_NULL] = ACTIONS(2482),
    [anon_sym_nullptr] = ACTIONS(2482),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2482),
    [anon_sym___typeof] = ACTIONS(2482),
    [anon_sym_typeof] = ACTIONS(2482),
    [anon_sym_ATimport] = ACTIONS(2484),
    [aux_sym_preproc_undef_token1] = ACTIONS(2482),
    [anon_sym_POUND] = ACTIONS(2482),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2482),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2482),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2482),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_API_AVAILABLE] = ACTIONS(2482),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_API_DEPRECATED] = ACTIONS(2482),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym___deprecated_msg] = ACTIONS(2482),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2482),
    [anon_sym_ATprotocol] = ACTIONS(2484),
    [anon_sym_ATinterface] = ACTIONS(2484),
    [anon_sym_ATimplementation] = ACTIONS(2484),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2484),
    [anon_sym_ATselector] = ACTIONS(2484),
    [anon_sym_ATavailable] = ACTIONS(2484),
    [anon_sym___builtin_available] = ACTIONS(2482),
    [anon_sym_va_arg] = ACTIONS(2482),
    [anon_sym_ATencode] = ACTIONS(2484),
    [anon_sym_BOOL] = ACTIONS(2482),
    [anon_sym_IMP] = ACTIONS(2482),
    [anon_sym_SEL] = ACTIONS(2482),
    [anon_sym_Class] = ACTIONS(2482),
    [anon_sym_id] = ACTIONS(2482),
  },
  [STATE(976)] = {
    [ts_builtin_sym_end] = ACTIONS(2410),
    [sym_identifier] = ACTIONS(2408),
    [aux_sym_preproc_include_token1] = ACTIONS(2408),
    [aux_sym_preproc_include_token2] = ACTIONS(2408),
    [aux_sym_preproc_def_token1] = ACTIONS(2408),
    [aux_sym_preproc_if_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2408),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2408),
    [sym_preproc_directive] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym___cdecl] = ACTIONS(2408),
    [anon_sym___clrcall] = ACTIONS(2408),
    [anon_sym___stdcall] = ACTIONS(2408),
    [anon_sym___fastcall] = ACTIONS(2408),
    [anon_sym___thiscall] = ACTIONS(2408),
    [anon_sym___vectorcall] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_case] = ACTIONS(2408),
    [anon_sym_default] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_ATimport] = ACTIONS(2410),
    [aux_sym_preproc_undef_token1] = ACTIONS(2408),
    [anon_sym_POUND] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATprotocol] = ACTIONS(2410),
    [anon_sym_ATinterface] = ACTIONS(2410),
    [anon_sym_ATimplementation] = ACTIONS(2410),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(977)] = {
    [ts_builtin_sym_end] = ACTIONS(2620),
    [sym_identifier] = ACTIONS(2618),
    [aux_sym_preproc_include_token1] = ACTIONS(2618),
    [aux_sym_preproc_include_token2] = ACTIONS(2618),
    [aux_sym_preproc_def_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2618),
    [sym_preproc_directive] = ACTIONS(2618),
    [anon_sym_LPAREN2] = ACTIONS(2620),
    [anon_sym_BANG] = ACTIONS(2620),
    [anon_sym_TILDE] = ACTIONS(2620),
    [anon_sym_DASH] = ACTIONS(2618),
    [anon_sym_PLUS] = ACTIONS(2618),
    [anon_sym_STAR] = ACTIONS(2620),
    [anon_sym_CARET] = ACTIONS(2620),
    [anon_sym_AMP] = ACTIONS(2620),
    [anon_sym_SEMI] = ACTIONS(2620),
    [anon_sym___extension__] = ACTIONS(2618),
    [anon_sym_typedef] = ACTIONS(2618),
    [anon_sym_extern] = ACTIONS(2618),
    [anon_sym___attribute__] = ACTIONS(2618),
    [anon_sym___attribute] = ACTIONS(2618),
    [anon_sym_noreturn] = ACTIONS(2618),
    [anon_sym_LBRACK] = ACTIONS(2620),
    [anon_sym___declspec] = ACTIONS(2618),
    [anon_sym___cdecl] = ACTIONS(2618),
    [anon_sym___clrcall] = ACTIONS(2618),
    [anon_sym___stdcall] = ACTIONS(2618),
    [anon_sym___fastcall] = ACTIONS(2618),
    [anon_sym___thiscall] = ACTIONS(2618),
    [anon_sym___vectorcall] = ACTIONS(2618),
    [anon_sym_LBRACE] = ACTIONS(2620),
    [anon_sym_signed] = ACTIONS(2618),
    [anon_sym_unsigned] = ACTIONS(2618),
    [anon_sym_long] = ACTIONS(2618),
    [anon_sym_short] = ACTIONS(2618),
    [anon_sym_static] = ACTIONS(2618),
    [anon_sym_ATautoreleasepool] = ACTIONS(2620),
    [anon_sym_auto] = ACTIONS(2618),
    [anon_sym_register] = ACTIONS(2618),
    [anon_sym_inline] = ACTIONS(2618),
    [anon_sym___inline] = ACTIONS(2618),
    [anon_sym___inline__] = ACTIONS(2618),
    [anon_sym___forceinline] = ACTIONS(2618),
    [anon_sym_thread_local] = ACTIONS(2618),
    [anon_sym___thread] = ACTIONS(2618),
    [anon_sym_CG_EXTERN] = ACTIONS(2618),
    [anon_sym_CG_INLINE] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2618),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2618),
    [anon_sym_IBOutlet] = ACTIONS(2618),
    [anon_sym_IBInspectable] = ACTIONS(2618),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2618),
    [anon_sym_NS_INLINE] = ACTIONS(2618),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2618),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2618),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2618),
    [anon_sym_const] = ACTIONS(2618),
    [anon_sym_constexpr] = ACTIONS(2618),
    [anon_sym_volatile] = ACTIONS(2618),
    [anon_sym_restrict] = ACTIONS(2618),
    [anon_sym___restrict__] = ACTIONS(2618),
    [anon_sym__Atomic] = ACTIONS(2618),
    [anon_sym__Noreturn] = ACTIONS(2618),
    [anon_sym__Nonnull] = ACTIONS(2618),
    [anon_sym_nullable] = ACTIONS(2618),
    [anon_sym__Complex] = ACTIONS(2618),
    [anon_sym__Nullable] = ACTIONS(2618),
    [anon_sym__Nullable_result] = ACTIONS(2618),
    [anon_sym__Null_unspecified] = ACTIONS(2618),
    [anon_sym___autoreleasing] = ACTIONS(2618),
    [anon_sym___block] = ACTIONS(2618),
    [anon_sym___bridge] = ACTIONS(2618),
    [anon_sym___bridge_retained] = ACTIONS(2618),
    [anon_sym___bridge_transfer] = ACTIONS(2618),
    [anon_sym___complex] = ACTIONS(2618),
    [anon_sym___const] = ACTIONS(2618),
    [anon_sym___imag] = ACTIONS(2618),
    [anon_sym___kindof] = ACTIONS(2618),
    [anon_sym___nonnull] = ACTIONS(2618),
    [anon_sym___nullable] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2618),
    [anon_sym___real] = ACTIONS(2618),
    [anon_sym___strong] = ACTIONS(2618),
    [anon_sym___unsafe_unretained] = ACTIONS(2618),
    [anon_sym___unused] = ACTIONS(2618),
    [anon_sym___weak] = ACTIONS(2618),
    [anon_sym_alignas] = ACTIONS(2618),
    [anon_sym__Alignas] = ACTIONS(2618),
    [sym_primitive_type] = ACTIONS(2618),
    [anon_sym_enum] = ACTIONS(2618),
    [anon_sym_struct] = ACTIONS(2618),
    [anon_sym_union] = ACTIONS(2618),
    [anon_sym_if] = ACTIONS(2618),
    [anon_sym_switch] = ACTIONS(2618),
    [anon_sym_case] = ACTIONS(2618),
    [anon_sym_default] = ACTIONS(2618),
    [anon_sym_while] = ACTIONS(2618),
    [anon_sym_do] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2618),
    [anon_sym_in] = ACTIONS(2618),
    [anon_sym_return] = ACTIONS(2618),
    [anon_sym_break] = ACTIONS(2618),
    [anon_sym_continue] = ACTIONS(2618),
    [anon_sym_goto] = ACTIONS(2618),
    [anon_sym_DASH_DASH] = ACTIONS(2620),
    [anon_sym_PLUS_PLUS] = ACTIONS(2620),
    [anon_sym_sizeof] = ACTIONS(2618),
    [anon_sym___alignof__] = ACTIONS(2618),
    [anon_sym___alignof] = ACTIONS(2618),
    [anon_sym__alignof] = ACTIONS(2618),
    [anon_sym_alignof] = ACTIONS(2618),
    [anon_sym__Alignof] = ACTIONS(2618),
    [anon_sym_offsetof] = ACTIONS(2618),
    [anon_sym__Generic] = ACTIONS(2618),
    [anon_sym_asm] = ACTIONS(2618),
    [anon_sym___asm__] = ACTIONS(2618),
    [anon_sym___asm] = ACTIONS(2618),
    [sym_number_literal] = ACTIONS(2620),
    [anon_sym_L_SQUOTE] = ACTIONS(2620),
    [anon_sym_u_SQUOTE] = ACTIONS(2620),
    [anon_sym_U_SQUOTE] = ACTIONS(2620),
    [anon_sym_u8_SQUOTE] = ACTIONS(2620),
    [anon_sym_SQUOTE] = ACTIONS(2620),
    [anon_sym_AT] = ACTIONS(2618),
    [anon_sym_DQUOTE] = ACTIONS(2620),
    [anon_sym_L_DQUOTE] = ACTIONS(2620),
    [anon_sym_u_DQUOTE] = ACTIONS(2620),
    [anon_sym_U_DQUOTE] = ACTIONS(2620),
    [anon_sym_u8_DQUOTE] = ACTIONS(2620),
    [sym_true] = ACTIONS(2618),
    [sym_false] = ACTIONS(2618),
    [anon_sym_NULL] = ACTIONS(2618),
    [anon_sym_nullptr] = ACTIONS(2618),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2618),
    [anon_sym___typeof] = ACTIONS(2618),
    [anon_sym_typeof] = ACTIONS(2618),
    [anon_sym_ATimport] = ACTIONS(2620),
    [aux_sym_preproc_undef_token1] = ACTIONS(2618),
    [anon_sym_POUND] = ACTIONS(2618),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2618),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2618),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2618),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_API_AVAILABLE] = ACTIONS(2618),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_API_DEPRECATED] = ACTIONS(2618),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym___deprecated_msg] = ACTIONS(2618),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2618),
    [anon_sym_ATprotocol] = ACTIONS(2620),
    [anon_sym_ATinterface] = ACTIONS(2620),
    [anon_sym_ATimplementation] = ACTIONS(2620),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2620),
    [anon_sym_ATselector] = ACTIONS(2620),
    [anon_sym_ATavailable] = ACTIONS(2620),
    [anon_sym___builtin_available] = ACTIONS(2618),
    [anon_sym_va_arg] = ACTIONS(2618),
    [anon_sym_ATencode] = ACTIONS(2620),
    [anon_sym_BOOL] = ACTIONS(2618),
    [anon_sym_IMP] = ACTIONS(2618),
    [anon_sym_SEL] = ACTIONS(2618),
    [anon_sym_Class] = ACTIONS(2618),
    [anon_sym_id] = ACTIONS(2618),
  },
  [STATE(978)] = {
    [ts_builtin_sym_end] = ACTIONS(2932),
    [sym_identifier] = ACTIONS(2935),
    [aux_sym_preproc_include_token1] = ACTIONS(2935),
    [aux_sym_preproc_include_token2] = ACTIONS(2935),
    [aux_sym_preproc_def_token1] = ACTIONS(2935),
    [aux_sym_preproc_if_token1] = ACTIONS(2935),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2935),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2935),
    [sym_preproc_directive] = ACTIONS(2935),
    [anon_sym_LPAREN2] = ACTIONS(2932),
    [anon_sym_BANG] = ACTIONS(2932),
    [anon_sym_TILDE] = ACTIONS(2932),
    [anon_sym_DASH] = ACTIONS(2935),
    [anon_sym_PLUS] = ACTIONS(2935),
    [anon_sym_STAR] = ACTIONS(2932),
    [anon_sym_CARET] = ACTIONS(2932),
    [anon_sym_AMP] = ACTIONS(2932),
    [anon_sym_SEMI] = ACTIONS(2932),
    [anon_sym___extension__] = ACTIONS(2935),
    [anon_sym_typedef] = ACTIONS(2935),
    [anon_sym_extern] = ACTIONS(2935),
    [anon_sym___attribute__] = ACTIONS(2935),
    [anon_sym___attribute] = ACTIONS(2935),
    [anon_sym_noreturn] = ACTIONS(2935),
    [anon_sym_LBRACK] = ACTIONS(2932),
    [anon_sym___declspec] = ACTIONS(2935),
    [anon_sym___cdecl] = ACTIONS(2935),
    [anon_sym___clrcall] = ACTIONS(2935),
    [anon_sym___stdcall] = ACTIONS(2935),
    [anon_sym___fastcall] = ACTIONS(2935),
    [anon_sym___thiscall] = ACTIONS(2935),
    [anon_sym___vectorcall] = ACTIONS(2935),
    [anon_sym_LBRACE] = ACTIONS(2932),
    [anon_sym_signed] = ACTIONS(2935),
    [anon_sym_unsigned] = ACTIONS(2935),
    [anon_sym_long] = ACTIONS(2935),
    [anon_sym_short] = ACTIONS(2935),
    [anon_sym_static] = ACTIONS(2935),
    [anon_sym_ATautoreleasepool] = ACTIONS(2932),
    [anon_sym_auto] = ACTIONS(2935),
    [anon_sym_register] = ACTIONS(2935),
    [anon_sym_inline] = ACTIONS(2935),
    [anon_sym___inline] = ACTIONS(2935),
    [anon_sym___inline__] = ACTIONS(2935),
    [anon_sym___forceinline] = ACTIONS(2935),
    [anon_sym_thread_local] = ACTIONS(2935),
    [anon_sym___thread] = ACTIONS(2935),
    [anon_sym_CG_EXTERN] = ACTIONS(2935),
    [anon_sym_CG_INLINE] = ACTIONS(2935),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2935),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2935),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2935),
    [anon_sym_IBOutlet] = ACTIONS(2935),
    [anon_sym_IBInspectable] = ACTIONS(2935),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2935),
    [anon_sym_NS_INLINE] = ACTIONS(2935),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2935),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2935),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2935),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2935),
    [anon_sym_const] = ACTIONS(2935),
    [anon_sym_constexpr] = ACTIONS(2935),
    [anon_sym_volatile] = ACTIONS(2935),
    [anon_sym_restrict] = ACTIONS(2935),
    [anon_sym___restrict__] = ACTIONS(2935),
    [anon_sym__Atomic] = ACTIONS(2935),
    [anon_sym__Noreturn] = ACTIONS(2935),
    [anon_sym__Nonnull] = ACTIONS(2935),
    [anon_sym_nullable] = ACTIONS(2935),
    [anon_sym__Complex] = ACTIONS(2935),
    [anon_sym__Nullable] = ACTIONS(2935),
    [anon_sym__Nullable_result] = ACTIONS(2935),
    [anon_sym__Null_unspecified] = ACTIONS(2935),
    [anon_sym___autoreleasing] = ACTIONS(2935),
    [anon_sym___block] = ACTIONS(2935),
    [anon_sym___bridge] = ACTIONS(2935),
    [anon_sym___bridge_retained] = ACTIONS(2935),
    [anon_sym___bridge_transfer] = ACTIONS(2935),
    [anon_sym___complex] = ACTIONS(2935),
    [anon_sym___const] = ACTIONS(2935),
    [anon_sym___imag] = ACTIONS(2935),
    [anon_sym___kindof] = ACTIONS(2935),
    [anon_sym___nonnull] = ACTIONS(2935),
    [anon_sym___nullable] = ACTIONS(2935),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2935),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2935),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2935),
    [anon_sym___real] = ACTIONS(2935),
    [anon_sym___strong] = ACTIONS(2935),
    [anon_sym___unsafe_unretained] = ACTIONS(2935),
    [anon_sym___unused] = ACTIONS(2935),
    [anon_sym___weak] = ACTIONS(2935),
    [anon_sym_alignas] = ACTIONS(2935),
    [anon_sym__Alignas] = ACTIONS(2935),
    [sym_primitive_type] = ACTIONS(2935),
    [anon_sym_enum] = ACTIONS(2935),
    [anon_sym_struct] = ACTIONS(2935),
    [anon_sym_union] = ACTIONS(2935),
    [anon_sym_if] = ACTIONS(2935),
    [anon_sym_switch] = ACTIONS(2935),
    [anon_sym_case] = ACTIONS(2935),
    [anon_sym_default] = ACTIONS(2935),
    [anon_sym_while] = ACTIONS(2935),
    [anon_sym_do] = ACTIONS(2935),
    [anon_sym_for] = ACTIONS(2935),
    [anon_sym_in] = ACTIONS(2935),
    [anon_sym_return] = ACTIONS(2935),
    [anon_sym_break] = ACTIONS(2935),
    [anon_sym_continue] = ACTIONS(2935),
    [anon_sym_goto] = ACTIONS(2935),
    [anon_sym_DASH_DASH] = ACTIONS(2932),
    [anon_sym_PLUS_PLUS] = ACTIONS(2932),
    [anon_sym_sizeof] = ACTIONS(2935),
    [anon_sym___alignof__] = ACTIONS(2935),
    [anon_sym___alignof] = ACTIONS(2935),
    [anon_sym__alignof] = ACTIONS(2935),
    [anon_sym_alignof] = ACTIONS(2935),
    [anon_sym__Alignof] = ACTIONS(2935),
    [anon_sym_offsetof] = ACTIONS(2935),
    [anon_sym__Generic] = ACTIONS(2935),
    [anon_sym_asm] = ACTIONS(2935),
    [anon_sym___asm__] = ACTIONS(2935),
    [anon_sym___asm] = ACTIONS(2935),
    [sym_number_literal] = ACTIONS(2932),
    [anon_sym_L_SQUOTE] = ACTIONS(2932),
    [anon_sym_u_SQUOTE] = ACTIONS(2932),
    [anon_sym_U_SQUOTE] = ACTIONS(2932),
    [anon_sym_u8_SQUOTE] = ACTIONS(2932),
    [anon_sym_SQUOTE] = ACTIONS(2932),
    [anon_sym_AT] = ACTIONS(2935),
    [anon_sym_DQUOTE] = ACTIONS(2932),
    [anon_sym_L_DQUOTE] = ACTIONS(2932),
    [anon_sym_u_DQUOTE] = ACTIONS(2932),
    [anon_sym_U_DQUOTE] = ACTIONS(2932),
    [anon_sym_u8_DQUOTE] = ACTIONS(2932),
    [sym_true] = ACTIONS(2935),
    [sym_false] = ACTIONS(2935),
    [anon_sym_NULL] = ACTIONS(2935),
    [anon_sym_nullptr] = ACTIONS(2935),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2935),
    [anon_sym___typeof] = ACTIONS(2935),
    [anon_sym_typeof] = ACTIONS(2935),
    [anon_sym_ATimport] = ACTIONS(2932),
    [aux_sym_preproc_undef_token1] = ACTIONS(2935),
    [anon_sym_POUND] = ACTIONS(2935),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2935),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2935),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2935),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2935),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2935),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2935),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2935),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2935),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2935),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2935),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2935),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2935),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2935),
    [anon_sym_API_AVAILABLE] = ACTIONS(2935),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2935),
    [anon_sym_API_DEPRECATED] = ACTIONS(2935),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2935),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2935),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2935),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2935),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2935),
    [anon_sym___deprecated_msg] = ACTIONS(2935),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2935),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2935),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2935),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2935),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2935),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2935),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2935),
    [anon_sym_ATprotocol] = ACTIONS(2932),
    [anon_sym_ATinterface] = ACTIONS(2932),
    [anon_sym_ATimplementation] = ACTIONS(2932),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2932),
    [anon_sym_ATselector] = ACTIONS(2932),
    [anon_sym_ATavailable] = ACTIONS(2932),
    [anon_sym___builtin_available] = ACTIONS(2935),
    [anon_sym_va_arg] = ACTIONS(2935),
    [anon_sym_ATencode] = ACTIONS(2932),
    [anon_sym_BOOL] = ACTIONS(2935),
    [anon_sym_IMP] = ACTIONS(2935),
    [anon_sym_SEL] = ACTIONS(2935),
    [anon_sym_Class] = ACTIONS(2935),
    [anon_sym_id] = ACTIONS(2935),
  },
  [STATE(979)] = {
    [ts_builtin_sym_end] = ACTIONS(2782),
    [sym_identifier] = ACTIONS(2780),
    [aux_sym_preproc_include_token1] = ACTIONS(2780),
    [aux_sym_preproc_include_token2] = ACTIONS(2780),
    [aux_sym_preproc_def_token1] = ACTIONS(2780),
    [aux_sym_preproc_if_token1] = ACTIONS(2780),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2780),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2780),
    [sym_preproc_directive] = ACTIONS(2780),
    [anon_sym_LPAREN2] = ACTIONS(2782),
    [anon_sym_BANG] = ACTIONS(2782),
    [anon_sym_TILDE] = ACTIONS(2782),
    [anon_sym_DASH] = ACTIONS(2780),
    [anon_sym_PLUS] = ACTIONS(2780),
    [anon_sym_STAR] = ACTIONS(2782),
    [anon_sym_CARET] = ACTIONS(2782),
    [anon_sym_AMP] = ACTIONS(2782),
    [anon_sym_SEMI] = ACTIONS(2782),
    [anon_sym___extension__] = ACTIONS(2780),
    [anon_sym_typedef] = ACTIONS(2780),
    [anon_sym_extern] = ACTIONS(2780),
    [anon_sym___attribute__] = ACTIONS(2780),
    [anon_sym___attribute] = ACTIONS(2780),
    [anon_sym_noreturn] = ACTIONS(2780),
    [anon_sym_LBRACK] = ACTIONS(2782),
    [anon_sym___declspec] = ACTIONS(2780),
    [anon_sym___cdecl] = ACTIONS(2780),
    [anon_sym___clrcall] = ACTIONS(2780),
    [anon_sym___stdcall] = ACTIONS(2780),
    [anon_sym___fastcall] = ACTIONS(2780),
    [anon_sym___thiscall] = ACTIONS(2780),
    [anon_sym___vectorcall] = ACTIONS(2780),
    [anon_sym_LBRACE] = ACTIONS(2782),
    [anon_sym_signed] = ACTIONS(2780),
    [anon_sym_unsigned] = ACTIONS(2780),
    [anon_sym_long] = ACTIONS(2780),
    [anon_sym_short] = ACTIONS(2780),
    [anon_sym_static] = ACTIONS(2780),
    [anon_sym_ATautoreleasepool] = ACTIONS(2782),
    [anon_sym_auto] = ACTIONS(2780),
    [anon_sym_register] = ACTIONS(2780),
    [anon_sym_inline] = ACTIONS(2780),
    [anon_sym___inline] = ACTIONS(2780),
    [anon_sym___inline__] = ACTIONS(2780),
    [anon_sym___forceinline] = ACTIONS(2780),
    [anon_sym_thread_local] = ACTIONS(2780),
    [anon_sym___thread] = ACTIONS(2780),
    [anon_sym_CG_EXTERN] = ACTIONS(2780),
    [anon_sym_CG_INLINE] = ACTIONS(2780),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2780),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2780),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2780),
    [anon_sym_IBOutlet] = ACTIONS(2780),
    [anon_sym_IBInspectable] = ACTIONS(2780),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2780),
    [anon_sym_NS_INLINE] = ACTIONS(2780),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2780),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2780),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2780),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2780),
    [anon_sym_const] = ACTIONS(2780),
    [anon_sym_constexpr] = ACTIONS(2780),
    [anon_sym_volatile] = ACTIONS(2780),
    [anon_sym_restrict] = ACTIONS(2780),
    [anon_sym___restrict__] = ACTIONS(2780),
    [anon_sym__Atomic] = ACTIONS(2780),
    [anon_sym__Noreturn] = ACTIONS(2780),
    [anon_sym__Nonnull] = ACTIONS(2780),
    [anon_sym_nullable] = ACTIONS(2780),
    [anon_sym__Complex] = ACTIONS(2780),
    [anon_sym__Nullable] = ACTIONS(2780),
    [anon_sym__Nullable_result] = ACTIONS(2780),
    [anon_sym__Null_unspecified] = ACTIONS(2780),
    [anon_sym___autoreleasing] = ACTIONS(2780),
    [anon_sym___block] = ACTIONS(2780),
    [anon_sym___bridge] = ACTIONS(2780),
    [anon_sym___bridge_retained] = ACTIONS(2780),
    [anon_sym___bridge_transfer] = ACTIONS(2780),
    [anon_sym___complex] = ACTIONS(2780),
    [anon_sym___const] = ACTIONS(2780),
    [anon_sym___imag] = ACTIONS(2780),
    [anon_sym___kindof] = ACTIONS(2780),
    [anon_sym___nonnull] = ACTIONS(2780),
    [anon_sym___nullable] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2780),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2780),
    [anon_sym___real] = ACTIONS(2780),
    [anon_sym___strong] = ACTIONS(2780),
    [anon_sym___unsafe_unretained] = ACTIONS(2780),
    [anon_sym___unused] = ACTIONS(2780),
    [anon_sym___weak] = ACTIONS(2780),
    [anon_sym_alignas] = ACTIONS(2780),
    [anon_sym__Alignas] = ACTIONS(2780),
    [sym_primitive_type] = ACTIONS(2780),
    [anon_sym_enum] = ACTIONS(2780),
    [anon_sym_struct] = ACTIONS(2780),
    [anon_sym_union] = ACTIONS(2780),
    [anon_sym_if] = ACTIONS(2780),
    [anon_sym_switch] = ACTIONS(2780),
    [anon_sym_case] = ACTIONS(2780),
    [anon_sym_default] = ACTIONS(2780),
    [anon_sym_while] = ACTIONS(2780),
    [anon_sym_do] = ACTIONS(2780),
    [anon_sym_for] = ACTIONS(2780),
    [anon_sym_in] = ACTIONS(2780),
    [anon_sym_return] = ACTIONS(2780),
    [anon_sym_break] = ACTIONS(2780),
    [anon_sym_continue] = ACTIONS(2780),
    [anon_sym_goto] = ACTIONS(2780),
    [anon_sym_DASH_DASH] = ACTIONS(2782),
    [anon_sym_PLUS_PLUS] = ACTIONS(2782),
    [anon_sym_sizeof] = ACTIONS(2780),
    [anon_sym___alignof__] = ACTIONS(2780),
    [anon_sym___alignof] = ACTIONS(2780),
    [anon_sym__alignof] = ACTIONS(2780),
    [anon_sym_alignof] = ACTIONS(2780),
    [anon_sym__Alignof] = ACTIONS(2780),
    [anon_sym_offsetof] = ACTIONS(2780),
    [anon_sym__Generic] = ACTIONS(2780),
    [anon_sym_asm] = ACTIONS(2780),
    [anon_sym___asm__] = ACTIONS(2780),
    [anon_sym___asm] = ACTIONS(2780),
    [sym_number_literal] = ACTIONS(2782),
    [anon_sym_L_SQUOTE] = ACTIONS(2782),
    [anon_sym_u_SQUOTE] = ACTIONS(2782),
    [anon_sym_U_SQUOTE] = ACTIONS(2782),
    [anon_sym_u8_SQUOTE] = ACTIONS(2782),
    [anon_sym_SQUOTE] = ACTIONS(2782),
    [anon_sym_AT] = ACTIONS(2780),
    [anon_sym_DQUOTE] = ACTIONS(2782),
    [anon_sym_L_DQUOTE] = ACTIONS(2782),
    [anon_sym_u_DQUOTE] = ACTIONS(2782),
    [anon_sym_U_DQUOTE] = ACTIONS(2782),
    [anon_sym_u8_DQUOTE] = ACTIONS(2782),
    [sym_true] = ACTIONS(2780),
    [sym_false] = ACTIONS(2780),
    [anon_sym_NULL] = ACTIONS(2780),
    [anon_sym_nullptr] = ACTIONS(2780),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2780),
    [anon_sym___typeof] = ACTIONS(2780),
    [anon_sym_typeof] = ACTIONS(2780),
    [anon_sym_ATimport] = ACTIONS(2782),
    [aux_sym_preproc_undef_token1] = ACTIONS(2780),
    [anon_sym_POUND] = ACTIONS(2780),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2780),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2780),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2780),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2780),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2780),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2780),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2780),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_API_AVAILABLE] = ACTIONS(2780),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_API_DEPRECATED] = ACTIONS(2780),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2780),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2780),
    [anon_sym___deprecated_msg] = ACTIONS(2780),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2780),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2780),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2780),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2780),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2780),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2780),
    [anon_sym_ATprotocol] = ACTIONS(2782),
    [anon_sym_ATinterface] = ACTIONS(2782),
    [anon_sym_ATimplementation] = ACTIONS(2782),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2782),
    [anon_sym_ATselector] = ACTIONS(2782),
    [anon_sym_ATavailable] = ACTIONS(2782),
    [anon_sym___builtin_available] = ACTIONS(2780),
    [anon_sym_va_arg] = ACTIONS(2780),
    [anon_sym_ATencode] = ACTIONS(2782),
    [anon_sym_BOOL] = ACTIONS(2780),
    [anon_sym_IMP] = ACTIONS(2780),
    [anon_sym_SEL] = ACTIONS(2780),
    [anon_sym_Class] = ACTIONS(2780),
    [anon_sym_id] = ACTIONS(2780),
  },
  [STATE(980)] = {
    [ts_builtin_sym_end] = ACTIONS(2786),
    [sym_identifier] = ACTIONS(2784),
    [aux_sym_preproc_include_token1] = ACTIONS(2784),
    [aux_sym_preproc_include_token2] = ACTIONS(2784),
    [aux_sym_preproc_def_token1] = ACTIONS(2784),
    [aux_sym_preproc_if_token1] = ACTIONS(2784),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2784),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2784),
    [sym_preproc_directive] = ACTIONS(2784),
    [anon_sym_LPAREN2] = ACTIONS(2786),
    [anon_sym_BANG] = ACTIONS(2786),
    [anon_sym_TILDE] = ACTIONS(2786),
    [anon_sym_DASH] = ACTIONS(2784),
    [anon_sym_PLUS] = ACTIONS(2784),
    [anon_sym_STAR] = ACTIONS(2786),
    [anon_sym_CARET] = ACTIONS(2786),
    [anon_sym_AMP] = ACTIONS(2786),
    [anon_sym_SEMI] = ACTIONS(2786),
    [anon_sym___extension__] = ACTIONS(2784),
    [anon_sym_typedef] = ACTIONS(2784),
    [anon_sym_extern] = ACTIONS(2784),
    [anon_sym___attribute__] = ACTIONS(2784),
    [anon_sym___attribute] = ACTIONS(2784),
    [anon_sym_noreturn] = ACTIONS(2784),
    [anon_sym_LBRACK] = ACTIONS(2786),
    [anon_sym___declspec] = ACTIONS(2784),
    [anon_sym___cdecl] = ACTIONS(2784),
    [anon_sym___clrcall] = ACTIONS(2784),
    [anon_sym___stdcall] = ACTIONS(2784),
    [anon_sym___fastcall] = ACTIONS(2784),
    [anon_sym___thiscall] = ACTIONS(2784),
    [anon_sym___vectorcall] = ACTIONS(2784),
    [anon_sym_LBRACE] = ACTIONS(2786),
    [anon_sym_signed] = ACTIONS(2784),
    [anon_sym_unsigned] = ACTIONS(2784),
    [anon_sym_long] = ACTIONS(2784),
    [anon_sym_short] = ACTIONS(2784),
    [anon_sym_static] = ACTIONS(2784),
    [anon_sym_ATautoreleasepool] = ACTIONS(2786),
    [anon_sym_auto] = ACTIONS(2784),
    [anon_sym_register] = ACTIONS(2784),
    [anon_sym_inline] = ACTIONS(2784),
    [anon_sym___inline] = ACTIONS(2784),
    [anon_sym___inline__] = ACTIONS(2784),
    [anon_sym___forceinline] = ACTIONS(2784),
    [anon_sym_thread_local] = ACTIONS(2784),
    [anon_sym___thread] = ACTIONS(2784),
    [anon_sym_CG_EXTERN] = ACTIONS(2784),
    [anon_sym_CG_INLINE] = ACTIONS(2784),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2784),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2784),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2784),
    [anon_sym_IBOutlet] = ACTIONS(2784),
    [anon_sym_IBInspectable] = ACTIONS(2784),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2784),
    [anon_sym_NS_INLINE] = ACTIONS(2784),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2784),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2784),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2784),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2784),
    [anon_sym_const] = ACTIONS(2784),
    [anon_sym_constexpr] = ACTIONS(2784),
    [anon_sym_volatile] = ACTIONS(2784),
    [anon_sym_restrict] = ACTIONS(2784),
    [anon_sym___restrict__] = ACTIONS(2784),
    [anon_sym__Atomic] = ACTIONS(2784),
    [anon_sym__Noreturn] = ACTIONS(2784),
    [anon_sym__Nonnull] = ACTIONS(2784),
    [anon_sym_nullable] = ACTIONS(2784),
    [anon_sym__Complex] = ACTIONS(2784),
    [anon_sym__Nullable] = ACTIONS(2784),
    [anon_sym__Nullable_result] = ACTIONS(2784),
    [anon_sym__Null_unspecified] = ACTIONS(2784),
    [anon_sym___autoreleasing] = ACTIONS(2784),
    [anon_sym___block] = ACTIONS(2784),
    [anon_sym___bridge] = ACTIONS(2784),
    [anon_sym___bridge_retained] = ACTIONS(2784),
    [anon_sym___bridge_transfer] = ACTIONS(2784),
    [anon_sym___complex] = ACTIONS(2784),
    [anon_sym___const] = ACTIONS(2784),
    [anon_sym___imag] = ACTIONS(2784),
    [anon_sym___kindof] = ACTIONS(2784),
    [anon_sym___nonnull] = ACTIONS(2784),
    [anon_sym___nullable] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2784),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2784),
    [anon_sym___real] = ACTIONS(2784),
    [anon_sym___strong] = ACTIONS(2784),
    [anon_sym___unsafe_unretained] = ACTIONS(2784),
    [anon_sym___unused] = ACTIONS(2784),
    [anon_sym___weak] = ACTIONS(2784),
    [anon_sym_alignas] = ACTIONS(2784),
    [anon_sym__Alignas] = ACTIONS(2784),
    [sym_primitive_type] = ACTIONS(2784),
    [anon_sym_enum] = ACTIONS(2784),
    [anon_sym_struct] = ACTIONS(2784),
    [anon_sym_union] = ACTIONS(2784),
    [anon_sym_if] = ACTIONS(2784),
    [anon_sym_switch] = ACTIONS(2784),
    [anon_sym_case] = ACTIONS(2784),
    [anon_sym_default] = ACTIONS(2784),
    [anon_sym_while] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2784),
    [anon_sym_for] = ACTIONS(2784),
    [anon_sym_in] = ACTIONS(2784),
    [anon_sym_return] = ACTIONS(2784),
    [anon_sym_break] = ACTIONS(2784),
    [anon_sym_continue] = ACTIONS(2784),
    [anon_sym_goto] = ACTIONS(2784),
    [anon_sym_DASH_DASH] = ACTIONS(2786),
    [anon_sym_PLUS_PLUS] = ACTIONS(2786),
    [anon_sym_sizeof] = ACTIONS(2784),
    [anon_sym___alignof__] = ACTIONS(2784),
    [anon_sym___alignof] = ACTIONS(2784),
    [anon_sym__alignof] = ACTIONS(2784),
    [anon_sym_alignof] = ACTIONS(2784),
    [anon_sym__Alignof] = ACTIONS(2784),
    [anon_sym_offsetof] = ACTIONS(2784),
    [anon_sym__Generic] = ACTIONS(2784),
    [anon_sym_asm] = ACTIONS(2784),
    [anon_sym___asm__] = ACTIONS(2784),
    [anon_sym___asm] = ACTIONS(2784),
    [sym_number_literal] = ACTIONS(2786),
    [anon_sym_L_SQUOTE] = ACTIONS(2786),
    [anon_sym_u_SQUOTE] = ACTIONS(2786),
    [anon_sym_U_SQUOTE] = ACTIONS(2786),
    [anon_sym_u8_SQUOTE] = ACTIONS(2786),
    [anon_sym_SQUOTE] = ACTIONS(2786),
    [anon_sym_AT] = ACTIONS(2784),
    [anon_sym_DQUOTE] = ACTIONS(2786),
    [anon_sym_L_DQUOTE] = ACTIONS(2786),
    [anon_sym_u_DQUOTE] = ACTIONS(2786),
    [anon_sym_U_DQUOTE] = ACTIONS(2786),
    [anon_sym_u8_DQUOTE] = ACTIONS(2786),
    [sym_true] = ACTIONS(2784),
    [sym_false] = ACTIONS(2784),
    [anon_sym_NULL] = ACTIONS(2784),
    [anon_sym_nullptr] = ACTIONS(2784),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2784),
    [anon_sym___typeof] = ACTIONS(2784),
    [anon_sym_typeof] = ACTIONS(2784),
    [anon_sym_ATimport] = ACTIONS(2786),
    [aux_sym_preproc_undef_token1] = ACTIONS(2784),
    [anon_sym_POUND] = ACTIONS(2784),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2784),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2784),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2784),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2784),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2784),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2784),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2784),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_API_AVAILABLE] = ACTIONS(2784),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_API_DEPRECATED] = ACTIONS(2784),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2784),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2784),
    [anon_sym___deprecated_msg] = ACTIONS(2784),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2784),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2784),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2784),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2784),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2784),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2784),
    [anon_sym_ATprotocol] = ACTIONS(2786),
    [anon_sym_ATinterface] = ACTIONS(2786),
    [anon_sym_ATimplementation] = ACTIONS(2786),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2786),
    [anon_sym_ATselector] = ACTIONS(2786),
    [anon_sym_ATavailable] = ACTIONS(2786),
    [anon_sym___builtin_available] = ACTIONS(2784),
    [anon_sym_va_arg] = ACTIONS(2784),
    [anon_sym_ATencode] = ACTIONS(2786),
    [anon_sym_BOOL] = ACTIONS(2784),
    [anon_sym_IMP] = ACTIONS(2784),
    [anon_sym_SEL] = ACTIONS(2784),
    [anon_sym_Class] = ACTIONS(2784),
    [anon_sym_id] = ACTIONS(2784),
  },
  [STATE(981)] = {
    [ts_builtin_sym_end] = ACTIONS(2790),
    [sym_identifier] = ACTIONS(2788),
    [aux_sym_preproc_include_token1] = ACTIONS(2788),
    [aux_sym_preproc_include_token2] = ACTIONS(2788),
    [aux_sym_preproc_def_token1] = ACTIONS(2788),
    [aux_sym_preproc_if_token1] = ACTIONS(2788),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2788),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2788),
    [sym_preproc_directive] = ACTIONS(2788),
    [anon_sym_LPAREN2] = ACTIONS(2790),
    [anon_sym_BANG] = ACTIONS(2790),
    [anon_sym_TILDE] = ACTIONS(2790),
    [anon_sym_DASH] = ACTIONS(2788),
    [anon_sym_PLUS] = ACTIONS(2788),
    [anon_sym_STAR] = ACTIONS(2790),
    [anon_sym_CARET] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(2790),
    [anon_sym_SEMI] = ACTIONS(2790),
    [anon_sym___extension__] = ACTIONS(2788),
    [anon_sym_typedef] = ACTIONS(2788),
    [anon_sym_extern] = ACTIONS(2788),
    [anon_sym___attribute__] = ACTIONS(2788),
    [anon_sym___attribute] = ACTIONS(2788),
    [anon_sym_noreturn] = ACTIONS(2788),
    [anon_sym_LBRACK] = ACTIONS(2790),
    [anon_sym___declspec] = ACTIONS(2788),
    [anon_sym___cdecl] = ACTIONS(2788),
    [anon_sym___clrcall] = ACTIONS(2788),
    [anon_sym___stdcall] = ACTIONS(2788),
    [anon_sym___fastcall] = ACTIONS(2788),
    [anon_sym___thiscall] = ACTIONS(2788),
    [anon_sym___vectorcall] = ACTIONS(2788),
    [anon_sym_LBRACE] = ACTIONS(2790),
    [anon_sym_signed] = ACTIONS(2788),
    [anon_sym_unsigned] = ACTIONS(2788),
    [anon_sym_long] = ACTIONS(2788),
    [anon_sym_short] = ACTIONS(2788),
    [anon_sym_static] = ACTIONS(2788),
    [anon_sym_ATautoreleasepool] = ACTIONS(2790),
    [anon_sym_auto] = ACTIONS(2788),
    [anon_sym_register] = ACTIONS(2788),
    [anon_sym_inline] = ACTIONS(2788),
    [anon_sym___inline] = ACTIONS(2788),
    [anon_sym___inline__] = ACTIONS(2788),
    [anon_sym___forceinline] = ACTIONS(2788),
    [anon_sym_thread_local] = ACTIONS(2788),
    [anon_sym___thread] = ACTIONS(2788),
    [anon_sym_CG_EXTERN] = ACTIONS(2788),
    [anon_sym_CG_INLINE] = ACTIONS(2788),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2788),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2788),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2788),
    [anon_sym_IBOutlet] = ACTIONS(2788),
    [anon_sym_IBInspectable] = ACTIONS(2788),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2788),
    [anon_sym_NS_INLINE] = ACTIONS(2788),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2788),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2788),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2788),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2788),
    [anon_sym_const] = ACTIONS(2788),
    [anon_sym_constexpr] = ACTIONS(2788),
    [anon_sym_volatile] = ACTIONS(2788),
    [anon_sym_restrict] = ACTIONS(2788),
    [anon_sym___restrict__] = ACTIONS(2788),
    [anon_sym__Atomic] = ACTIONS(2788),
    [anon_sym__Noreturn] = ACTIONS(2788),
    [anon_sym__Nonnull] = ACTIONS(2788),
    [anon_sym_nullable] = ACTIONS(2788),
    [anon_sym__Complex] = ACTIONS(2788),
    [anon_sym__Nullable] = ACTIONS(2788),
    [anon_sym__Nullable_result] = ACTIONS(2788),
    [anon_sym__Null_unspecified] = ACTIONS(2788),
    [anon_sym___autoreleasing] = ACTIONS(2788),
    [anon_sym___block] = ACTIONS(2788),
    [anon_sym___bridge] = ACTIONS(2788),
    [anon_sym___bridge_retained] = ACTIONS(2788),
    [anon_sym___bridge_transfer] = ACTIONS(2788),
    [anon_sym___complex] = ACTIONS(2788),
    [anon_sym___const] = ACTIONS(2788),
    [anon_sym___imag] = ACTIONS(2788),
    [anon_sym___kindof] = ACTIONS(2788),
    [anon_sym___nonnull] = ACTIONS(2788),
    [anon_sym___nullable] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2788),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2788),
    [anon_sym___real] = ACTIONS(2788),
    [anon_sym___strong] = ACTIONS(2788),
    [anon_sym___unsafe_unretained] = ACTIONS(2788),
    [anon_sym___unused] = ACTIONS(2788),
    [anon_sym___weak] = ACTIONS(2788),
    [anon_sym_alignas] = ACTIONS(2788),
    [anon_sym__Alignas] = ACTIONS(2788),
    [sym_primitive_type] = ACTIONS(2788),
    [anon_sym_enum] = ACTIONS(2788),
    [anon_sym_struct] = ACTIONS(2788),
    [anon_sym_union] = ACTIONS(2788),
    [anon_sym_if] = ACTIONS(2788),
    [anon_sym_switch] = ACTIONS(2788),
    [anon_sym_case] = ACTIONS(2788),
    [anon_sym_default] = ACTIONS(2788),
    [anon_sym_while] = ACTIONS(2788),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_for] = ACTIONS(2788),
    [anon_sym_in] = ACTIONS(2788),
    [anon_sym_return] = ACTIONS(2788),
    [anon_sym_break] = ACTIONS(2788),
    [anon_sym_continue] = ACTIONS(2788),
    [anon_sym_goto] = ACTIONS(2788),
    [anon_sym_DASH_DASH] = ACTIONS(2790),
    [anon_sym_PLUS_PLUS] = ACTIONS(2790),
    [anon_sym_sizeof] = ACTIONS(2788),
    [anon_sym___alignof__] = ACTIONS(2788),
    [anon_sym___alignof] = ACTIONS(2788),
    [anon_sym__alignof] = ACTIONS(2788),
    [anon_sym_alignof] = ACTIONS(2788),
    [anon_sym__Alignof] = ACTIONS(2788),
    [anon_sym_offsetof] = ACTIONS(2788),
    [anon_sym__Generic] = ACTIONS(2788),
    [anon_sym_asm] = ACTIONS(2788),
    [anon_sym___asm__] = ACTIONS(2788),
    [anon_sym___asm] = ACTIONS(2788),
    [sym_number_literal] = ACTIONS(2790),
    [anon_sym_L_SQUOTE] = ACTIONS(2790),
    [anon_sym_u_SQUOTE] = ACTIONS(2790),
    [anon_sym_U_SQUOTE] = ACTIONS(2790),
    [anon_sym_u8_SQUOTE] = ACTIONS(2790),
    [anon_sym_SQUOTE] = ACTIONS(2790),
    [anon_sym_AT] = ACTIONS(2788),
    [anon_sym_DQUOTE] = ACTIONS(2790),
    [anon_sym_L_DQUOTE] = ACTIONS(2790),
    [anon_sym_u_DQUOTE] = ACTIONS(2790),
    [anon_sym_U_DQUOTE] = ACTIONS(2790),
    [anon_sym_u8_DQUOTE] = ACTIONS(2790),
    [sym_true] = ACTIONS(2788),
    [sym_false] = ACTIONS(2788),
    [anon_sym_NULL] = ACTIONS(2788),
    [anon_sym_nullptr] = ACTIONS(2788),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2788),
    [anon_sym___typeof] = ACTIONS(2788),
    [anon_sym_typeof] = ACTIONS(2788),
    [anon_sym_ATimport] = ACTIONS(2790),
    [aux_sym_preproc_undef_token1] = ACTIONS(2788),
    [anon_sym_POUND] = ACTIONS(2788),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2788),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2788),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2788),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2788),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2788),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2788),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2788),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_API_AVAILABLE] = ACTIONS(2788),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_API_DEPRECATED] = ACTIONS(2788),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2788),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2788),
    [anon_sym___deprecated_msg] = ACTIONS(2788),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2788),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2788),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2788),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2788),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2788),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2788),
    [anon_sym_ATprotocol] = ACTIONS(2790),
    [anon_sym_ATinterface] = ACTIONS(2790),
    [anon_sym_ATimplementation] = ACTIONS(2790),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2790),
    [anon_sym_ATselector] = ACTIONS(2790),
    [anon_sym_ATavailable] = ACTIONS(2790),
    [anon_sym___builtin_available] = ACTIONS(2788),
    [anon_sym_va_arg] = ACTIONS(2788),
    [anon_sym_ATencode] = ACTIONS(2790),
    [anon_sym_BOOL] = ACTIONS(2788),
    [anon_sym_IMP] = ACTIONS(2788),
    [anon_sym_SEL] = ACTIONS(2788),
    [anon_sym_Class] = ACTIONS(2788),
    [anon_sym_id] = ACTIONS(2788),
  },
  [STATE(982)] = {
    [ts_builtin_sym_end] = ACTIONS(2938),
    [sym_identifier] = ACTIONS(2940),
    [aux_sym_preproc_include_token1] = ACTIONS(2940),
    [aux_sym_preproc_include_token2] = ACTIONS(2940),
    [aux_sym_preproc_def_token1] = ACTIONS(2940),
    [aux_sym_preproc_if_token1] = ACTIONS(2940),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2940),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2940),
    [sym_preproc_directive] = ACTIONS(2940),
    [anon_sym_LPAREN2] = ACTIONS(2938),
    [anon_sym_BANG] = ACTIONS(2938),
    [anon_sym_TILDE] = ACTIONS(2938),
    [anon_sym_DASH] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2938),
    [anon_sym_CARET] = ACTIONS(2938),
    [anon_sym_AMP] = ACTIONS(2938),
    [anon_sym_SEMI] = ACTIONS(2938),
    [anon_sym___extension__] = ACTIONS(2940),
    [anon_sym_typedef] = ACTIONS(2940),
    [anon_sym_extern] = ACTIONS(2940),
    [anon_sym___attribute__] = ACTIONS(2940),
    [anon_sym___attribute] = ACTIONS(2940),
    [anon_sym_noreturn] = ACTIONS(2940),
    [anon_sym_LBRACK] = ACTIONS(2938),
    [anon_sym___declspec] = ACTIONS(2940),
    [anon_sym___cdecl] = ACTIONS(2940),
    [anon_sym___clrcall] = ACTIONS(2940),
    [anon_sym___stdcall] = ACTIONS(2940),
    [anon_sym___fastcall] = ACTIONS(2940),
    [anon_sym___thiscall] = ACTIONS(2940),
    [anon_sym___vectorcall] = ACTIONS(2940),
    [anon_sym_LBRACE] = ACTIONS(2938),
    [anon_sym_signed] = ACTIONS(2940),
    [anon_sym_unsigned] = ACTIONS(2940),
    [anon_sym_long] = ACTIONS(2940),
    [anon_sym_short] = ACTIONS(2940),
    [anon_sym_static] = ACTIONS(2940),
    [anon_sym_ATautoreleasepool] = ACTIONS(2938),
    [anon_sym_auto] = ACTIONS(2940),
    [anon_sym_register] = ACTIONS(2940),
    [anon_sym_inline] = ACTIONS(2940),
    [anon_sym___inline] = ACTIONS(2940),
    [anon_sym___inline__] = ACTIONS(2940),
    [anon_sym___forceinline] = ACTIONS(2940),
    [anon_sym_thread_local] = ACTIONS(2940),
    [anon_sym___thread] = ACTIONS(2940),
    [anon_sym_CG_EXTERN] = ACTIONS(2940),
    [anon_sym_CG_INLINE] = ACTIONS(2940),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2940),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2940),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2940),
    [anon_sym_IBOutlet] = ACTIONS(2940),
    [anon_sym_IBInspectable] = ACTIONS(2940),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2940),
    [anon_sym_NS_INLINE] = ACTIONS(2940),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2940),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2940),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2940),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2940),
    [anon_sym_const] = ACTIONS(2940),
    [anon_sym_constexpr] = ACTIONS(2940),
    [anon_sym_volatile] = ACTIONS(2940),
    [anon_sym_restrict] = ACTIONS(2940),
    [anon_sym___restrict__] = ACTIONS(2940),
    [anon_sym__Atomic] = ACTIONS(2940),
    [anon_sym__Noreturn] = ACTIONS(2940),
    [anon_sym__Nonnull] = ACTIONS(2940),
    [anon_sym_nullable] = ACTIONS(2940),
    [anon_sym__Complex] = ACTIONS(2940),
    [anon_sym__Nullable] = ACTIONS(2940),
    [anon_sym__Nullable_result] = ACTIONS(2940),
    [anon_sym__Null_unspecified] = ACTIONS(2940),
    [anon_sym___autoreleasing] = ACTIONS(2940),
    [anon_sym___block] = ACTIONS(2940),
    [anon_sym___bridge] = ACTIONS(2940),
    [anon_sym___bridge_retained] = ACTIONS(2940),
    [anon_sym___bridge_transfer] = ACTIONS(2940),
    [anon_sym___complex] = ACTIONS(2940),
    [anon_sym___const] = ACTIONS(2940),
    [anon_sym___imag] = ACTIONS(2940),
    [anon_sym___kindof] = ACTIONS(2940),
    [anon_sym___nonnull] = ACTIONS(2940),
    [anon_sym___nullable] = ACTIONS(2940),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2940),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2940),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2940),
    [anon_sym___real] = ACTIONS(2940),
    [anon_sym___strong] = ACTIONS(2940),
    [anon_sym___unsafe_unretained] = ACTIONS(2940),
    [anon_sym___unused] = ACTIONS(2940),
    [anon_sym___weak] = ACTIONS(2940),
    [anon_sym_alignas] = ACTIONS(2940),
    [anon_sym__Alignas] = ACTIONS(2940),
    [sym_primitive_type] = ACTIONS(2940),
    [anon_sym_enum] = ACTIONS(2940),
    [anon_sym_struct] = ACTIONS(2940),
    [anon_sym_union] = ACTIONS(2940),
    [anon_sym_if] = ACTIONS(2940),
    [anon_sym_switch] = ACTIONS(2940),
    [anon_sym_case] = ACTIONS(2940),
    [anon_sym_default] = ACTIONS(2940),
    [anon_sym_while] = ACTIONS(2940),
    [anon_sym_do] = ACTIONS(2940),
    [anon_sym_for] = ACTIONS(2940),
    [anon_sym_in] = ACTIONS(2940),
    [anon_sym_return] = ACTIONS(2940),
    [anon_sym_break] = ACTIONS(2940),
    [anon_sym_continue] = ACTIONS(2940),
    [anon_sym_goto] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2938),
    [anon_sym_PLUS_PLUS] = ACTIONS(2938),
    [anon_sym_sizeof] = ACTIONS(2940),
    [anon_sym___alignof__] = ACTIONS(2940),
    [anon_sym___alignof] = ACTIONS(2940),
    [anon_sym__alignof] = ACTIONS(2940),
    [anon_sym_alignof] = ACTIONS(2940),
    [anon_sym__Alignof] = ACTIONS(2940),
    [anon_sym_offsetof] = ACTIONS(2940),
    [anon_sym__Generic] = ACTIONS(2940),
    [anon_sym_asm] = ACTIONS(2940),
    [anon_sym___asm__] = ACTIONS(2940),
    [anon_sym___asm] = ACTIONS(2940),
    [sym_number_literal] = ACTIONS(2938),
    [anon_sym_L_SQUOTE] = ACTIONS(2938),
    [anon_sym_u_SQUOTE] = ACTIONS(2938),
    [anon_sym_U_SQUOTE] = ACTIONS(2938),
    [anon_sym_u8_SQUOTE] = ACTIONS(2938),
    [anon_sym_SQUOTE] = ACTIONS(2938),
    [anon_sym_AT] = ACTIONS(2940),
    [anon_sym_DQUOTE] = ACTIONS(2938),
    [anon_sym_L_DQUOTE] = ACTIONS(2938),
    [anon_sym_u_DQUOTE] = ACTIONS(2938),
    [anon_sym_U_DQUOTE] = ACTIONS(2938),
    [anon_sym_u8_DQUOTE] = ACTIONS(2938),
    [sym_true] = ACTIONS(2940),
    [sym_false] = ACTIONS(2940),
    [anon_sym_NULL] = ACTIONS(2940),
    [anon_sym_nullptr] = ACTIONS(2940),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2940),
    [anon_sym___typeof] = ACTIONS(2940),
    [anon_sym_typeof] = ACTIONS(2940),
    [anon_sym_ATimport] = ACTIONS(2938),
    [aux_sym_preproc_undef_token1] = ACTIONS(2940),
    [anon_sym_POUND] = ACTIONS(2940),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2940),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2940),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2940),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2940),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2940),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2940),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2940),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2940),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2940),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2940),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2940),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2940),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2940),
    [anon_sym_API_AVAILABLE] = ACTIONS(2940),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2940),
    [anon_sym_API_DEPRECATED] = ACTIONS(2940),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2940),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2940),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2940),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2940),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2940),
    [anon_sym___deprecated_msg] = ACTIONS(2940),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2940),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2940),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2940),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2940),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2940),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2940),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2940),
    [anon_sym_ATprotocol] = ACTIONS(2938),
    [anon_sym_ATinterface] = ACTIONS(2938),
    [anon_sym_ATimplementation] = ACTIONS(2938),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2938),
    [anon_sym_ATselector] = ACTIONS(2938),
    [anon_sym_ATavailable] = ACTIONS(2938),
    [anon_sym___builtin_available] = ACTIONS(2940),
    [anon_sym_va_arg] = ACTIONS(2940),
    [anon_sym_ATencode] = ACTIONS(2938),
    [anon_sym_BOOL] = ACTIONS(2940),
    [anon_sym_IMP] = ACTIONS(2940),
    [anon_sym_SEL] = ACTIONS(2940),
    [anon_sym_Class] = ACTIONS(2940),
    [anon_sym_id] = ACTIONS(2940),
  },
  [STATE(983)] = {
    [ts_builtin_sym_end] = ACTIONS(2942),
    [sym_identifier] = ACTIONS(2944),
    [aux_sym_preproc_include_token1] = ACTIONS(2944),
    [aux_sym_preproc_include_token2] = ACTIONS(2944),
    [aux_sym_preproc_def_token1] = ACTIONS(2944),
    [aux_sym_preproc_if_token1] = ACTIONS(2944),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2944),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2944),
    [sym_preproc_directive] = ACTIONS(2944),
    [anon_sym_LPAREN2] = ACTIONS(2942),
    [anon_sym_BANG] = ACTIONS(2942),
    [anon_sym_TILDE] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2944),
    [anon_sym_PLUS] = ACTIONS(2944),
    [anon_sym_STAR] = ACTIONS(2942),
    [anon_sym_CARET] = ACTIONS(2942),
    [anon_sym_AMP] = ACTIONS(2942),
    [anon_sym_SEMI] = ACTIONS(2942),
    [anon_sym___extension__] = ACTIONS(2944),
    [anon_sym_typedef] = ACTIONS(2944),
    [anon_sym_extern] = ACTIONS(2944),
    [anon_sym___attribute__] = ACTIONS(2944),
    [anon_sym___attribute] = ACTIONS(2944),
    [anon_sym_noreturn] = ACTIONS(2944),
    [anon_sym_LBRACK] = ACTIONS(2942),
    [anon_sym___declspec] = ACTIONS(2944),
    [anon_sym___cdecl] = ACTIONS(2944),
    [anon_sym___clrcall] = ACTIONS(2944),
    [anon_sym___stdcall] = ACTIONS(2944),
    [anon_sym___fastcall] = ACTIONS(2944),
    [anon_sym___thiscall] = ACTIONS(2944),
    [anon_sym___vectorcall] = ACTIONS(2944),
    [anon_sym_LBRACE] = ACTIONS(2942),
    [anon_sym_signed] = ACTIONS(2944),
    [anon_sym_unsigned] = ACTIONS(2944),
    [anon_sym_long] = ACTIONS(2944),
    [anon_sym_short] = ACTIONS(2944),
    [anon_sym_static] = ACTIONS(2944),
    [anon_sym_ATautoreleasepool] = ACTIONS(2942),
    [anon_sym_auto] = ACTIONS(2944),
    [anon_sym_register] = ACTIONS(2944),
    [anon_sym_inline] = ACTIONS(2944),
    [anon_sym___inline] = ACTIONS(2944),
    [anon_sym___inline__] = ACTIONS(2944),
    [anon_sym___forceinline] = ACTIONS(2944),
    [anon_sym_thread_local] = ACTIONS(2944),
    [anon_sym___thread] = ACTIONS(2944),
    [anon_sym_CG_EXTERN] = ACTIONS(2944),
    [anon_sym_CG_INLINE] = ACTIONS(2944),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2944),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2944),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2944),
    [anon_sym_IBOutlet] = ACTIONS(2944),
    [anon_sym_IBInspectable] = ACTIONS(2944),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2944),
    [anon_sym_NS_INLINE] = ACTIONS(2944),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2944),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2944),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2944),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2944),
    [anon_sym_const] = ACTIONS(2944),
    [anon_sym_constexpr] = ACTIONS(2944),
    [anon_sym_volatile] = ACTIONS(2944),
    [anon_sym_restrict] = ACTIONS(2944),
    [anon_sym___restrict__] = ACTIONS(2944),
    [anon_sym__Atomic] = ACTIONS(2944),
    [anon_sym__Noreturn] = ACTIONS(2944),
    [anon_sym__Nonnull] = ACTIONS(2944),
    [anon_sym_nullable] = ACTIONS(2944),
    [anon_sym__Complex] = ACTIONS(2944),
    [anon_sym__Nullable] = ACTIONS(2944),
    [anon_sym__Nullable_result] = ACTIONS(2944),
    [anon_sym__Null_unspecified] = ACTIONS(2944),
    [anon_sym___autoreleasing] = ACTIONS(2944),
    [anon_sym___block] = ACTIONS(2944),
    [anon_sym___bridge] = ACTIONS(2944),
    [anon_sym___bridge_retained] = ACTIONS(2944),
    [anon_sym___bridge_transfer] = ACTIONS(2944),
    [anon_sym___complex] = ACTIONS(2944),
    [anon_sym___const] = ACTIONS(2944),
    [anon_sym___imag] = ACTIONS(2944),
    [anon_sym___kindof] = ACTIONS(2944),
    [anon_sym___nonnull] = ACTIONS(2944),
    [anon_sym___nullable] = ACTIONS(2944),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2944),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2944),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2944),
    [anon_sym___real] = ACTIONS(2944),
    [anon_sym___strong] = ACTIONS(2944),
    [anon_sym___unsafe_unretained] = ACTIONS(2944),
    [anon_sym___unused] = ACTIONS(2944),
    [anon_sym___weak] = ACTIONS(2944),
    [anon_sym_alignas] = ACTIONS(2944),
    [anon_sym__Alignas] = ACTIONS(2944),
    [sym_primitive_type] = ACTIONS(2944),
    [anon_sym_enum] = ACTIONS(2944),
    [anon_sym_struct] = ACTIONS(2944),
    [anon_sym_union] = ACTIONS(2944),
    [anon_sym_if] = ACTIONS(2944),
    [anon_sym_switch] = ACTIONS(2944),
    [anon_sym_case] = ACTIONS(2944),
    [anon_sym_default] = ACTIONS(2944),
    [anon_sym_while] = ACTIONS(2944),
    [anon_sym_do] = ACTIONS(2944),
    [anon_sym_for] = ACTIONS(2944),
    [anon_sym_in] = ACTIONS(2944),
    [anon_sym_return] = ACTIONS(2944),
    [anon_sym_break] = ACTIONS(2944),
    [anon_sym_continue] = ACTIONS(2944),
    [anon_sym_goto] = ACTIONS(2944),
    [anon_sym_DASH_DASH] = ACTIONS(2942),
    [anon_sym_PLUS_PLUS] = ACTIONS(2942),
    [anon_sym_sizeof] = ACTIONS(2944),
    [anon_sym___alignof__] = ACTIONS(2944),
    [anon_sym___alignof] = ACTIONS(2944),
    [anon_sym__alignof] = ACTIONS(2944),
    [anon_sym_alignof] = ACTIONS(2944),
    [anon_sym__Alignof] = ACTIONS(2944),
    [anon_sym_offsetof] = ACTIONS(2944),
    [anon_sym__Generic] = ACTIONS(2944),
    [anon_sym_asm] = ACTIONS(2944),
    [anon_sym___asm__] = ACTIONS(2944),
    [anon_sym___asm] = ACTIONS(2944),
    [sym_number_literal] = ACTIONS(2942),
    [anon_sym_L_SQUOTE] = ACTIONS(2942),
    [anon_sym_u_SQUOTE] = ACTIONS(2942),
    [anon_sym_U_SQUOTE] = ACTIONS(2942),
    [anon_sym_u8_SQUOTE] = ACTIONS(2942),
    [anon_sym_SQUOTE] = ACTIONS(2942),
    [anon_sym_AT] = ACTIONS(2944),
    [anon_sym_DQUOTE] = ACTIONS(2942),
    [anon_sym_L_DQUOTE] = ACTIONS(2942),
    [anon_sym_u_DQUOTE] = ACTIONS(2942),
    [anon_sym_U_DQUOTE] = ACTIONS(2942),
    [anon_sym_u8_DQUOTE] = ACTIONS(2942),
    [sym_true] = ACTIONS(2944),
    [sym_false] = ACTIONS(2944),
    [anon_sym_NULL] = ACTIONS(2944),
    [anon_sym_nullptr] = ACTIONS(2944),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2944),
    [anon_sym___typeof] = ACTIONS(2944),
    [anon_sym_typeof] = ACTIONS(2944),
    [anon_sym_ATimport] = ACTIONS(2942),
    [aux_sym_preproc_undef_token1] = ACTIONS(2944),
    [anon_sym_POUND] = ACTIONS(2944),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2944),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2944),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2944),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2944),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2944),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2944),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2944),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2944),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2944),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2944),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2944),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2944),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2944),
    [anon_sym_API_AVAILABLE] = ACTIONS(2944),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2944),
    [anon_sym_API_DEPRECATED] = ACTIONS(2944),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2944),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2944),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2944),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2944),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2944),
    [anon_sym___deprecated_msg] = ACTIONS(2944),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2944),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2944),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2944),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2944),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2944),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2944),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2944),
    [anon_sym_ATprotocol] = ACTIONS(2942),
    [anon_sym_ATinterface] = ACTIONS(2942),
    [anon_sym_ATimplementation] = ACTIONS(2942),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2942),
    [anon_sym_ATselector] = ACTIONS(2942),
    [anon_sym_ATavailable] = ACTIONS(2942),
    [anon_sym___builtin_available] = ACTIONS(2944),
    [anon_sym_va_arg] = ACTIONS(2944),
    [anon_sym_ATencode] = ACTIONS(2942),
    [anon_sym_BOOL] = ACTIONS(2944),
    [anon_sym_IMP] = ACTIONS(2944),
    [anon_sym_SEL] = ACTIONS(2944),
    [anon_sym_Class] = ACTIONS(2944),
    [anon_sym_id] = ACTIONS(2944),
  },
  [STATE(984)] = {
    [ts_builtin_sym_end] = ACTIONS(2794),
    [sym_identifier] = ACTIONS(2792),
    [aux_sym_preproc_include_token1] = ACTIONS(2792),
    [aux_sym_preproc_include_token2] = ACTIONS(2792),
    [aux_sym_preproc_def_token1] = ACTIONS(2792),
    [aux_sym_preproc_if_token1] = ACTIONS(2792),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2792),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2792),
    [sym_preproc_directive] = ACTIONS(2792),
    [anon_sym_LPAREN2] = ACTIONS(2794),
    [anon_sym_BANG] = ACTIONS(2794),
    [anon_sym_TILDE] = ACTIONS(2794),
    [anon_sym_DASH] = ACTIONS(2792),
    [anon_sym_PLUS] = ACTIONS(2792),
    [anon_sym_STAR] = ACTIONS(2794),
    [anon_sym_CARET] = ACTIONS(2794),
    [anon_sym_AMP] = ACTIONS(2794),
    [anon_sym_SEMI] = ACTIONS(2794),
    [anon_sym___extension__] = ACTIONS(2792),
    [anon_sym_typedef] = ACTIONS(2792),
    [anon_sym_extern] = ACTIONS(2792),
    [anon_sym___attribute__] = ACTIONS(2792),
    [anon_sym___attribute] = ACTIONS(2792),
    [anon_sym_noreturn] = ACTIONS(2792),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym___declspec] = ACTIONS(2792),
    [anon_sym___cdecl] = ACTIONS(2792),
    [anon_sym___clrcall] = ACTIONS(2792),
    [anon_sym___stdcall] = ACTIONS(2792),
    [anon_sym___fastcall] = ACTIONS(2792),
    [anon_sym___thiscall] = ACTIONS(2792),
    [anon_sym___vectorcall] = ACTIONS(2792),
    [anon_sym_LBRACE] = ACTIONS(2794),
    [anon_sym_signed] = ACTIONS(2792),
    [anon_sym_unsigned] = ACTIONS(2792),
    [anon_sym_long] = ACTIONS(2792),
    [anon_sym_short] = ACTIONS(2792),
    [anon_sym_static] = ACTIONS(2792),
    [anon_sym_ATautoreleasepool] = ACTIONS(2794),
    [anon_sym_auto] = ACTIONS(2792),
    [anon_sym_register] = ACTIONS(2792),
    [anon_sym_inline] = ACTIONS(2792),
    [anon_sym___inline] = ACTIONS(2792),
    [anon_sym___inline__] = ACTIONS(2792),
    [anon_sym___forceinline] = ACTIONS(2792),
    [anon_sym_thread_local] = ACTIONS(2792),
    [anon_sym___thread] = ACTIONS(2792),
    [anon_sym_CG_EXTERN] = ACTIONS(2792),
    [anon_sym_CG_INLINE] = ACTIONS(2792),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2792),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2792),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2792),
    [anon_sym_IBOutlet] = ACTIONS(2792),
    [anon_sym_IBInspectable] = ACTIONS(2792),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2792),
    [anon_sym_NS_INLINE] = ACTIONS(2792),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2792),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2792),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2792),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2792),
    [anon_sym_const] = ACTIONS(2792),
    [anon_sym_constexpr] = ACTIONS(2792),
    [anon_sym_volatile] = ACTIONS(2792),
    [anon_sym_restrict] = ACTIONS(2792),
    [anon_sym___restrict__] = ACTIONS(2792),
    [anon_sym__Atomic] = ACTIONS(2792),
    [anon_sym__Noreturn] = ACTIONS(2792),
    [anon_sym__Nonnull] = ACTIONS(2792),
    [anon_sym_nullable] = ACTIONS(2792),
    [anon_sym__Complex] = ACTIONS(2792),
    [anon_sym__Nullable] = ACTIONS(2792),
    [anon_sym__Nullable_result] = ACTIONS(2792),
    [anon_sym__Null_unspecified] = ACTIONS(2792),
    [anon_sym___autoreleasing] = ACTIONS(2792),
    [anon_sym___block] = ACTIONS(2792),
    [anon_sym___bridge] = ACTIONS(2792),
    [anon_sym___bridge_retained] = ACTIONS(2792),
    [anon_sym___bridge_transfer] = ACTIONS(2792),
    [anon_sym___complex] = ACTIONS(2792),
    [anon_sym___const] = ACTIONS(2792),
    [anon_sym___imag] = ACTIONS(2792),
    [anon_sym___kindof] = ACTIONS(2792),
    [anon_sym___nonnull] = ACTIONS(2792),
    [anon_sym___nullable] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2792),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2792),
    [anon_sym___real] = ACTIONS(2792),
    [anon_sym___strong] = ACTIONS(2792),
    [anon_sym___unsafe_unretained] = ACTIONS(2792),
    [anon_sym___unused] = ACTIONS(2792),
    [anon_sym___weak] = ACTIONS(2792),
    [anon_sym_alignas] = ACTIONS(2792),
    [anon_sym__Alignas] = ACTIONS(2792),
    [sym_primitive_type] = ACTIONS(2792),
    [anon_sym_enum] = ACTIONS(2792),
    [anon_sym_struct] = ACTIONS(2792),
    [anon_sym_union] = ACTIONS(2792),
    [anon_sym_if] = ACTIONS(2792),
    [anon_sym_switch] = ACTIONS(2792),
    [anon_sym_case] = ACTIONS(2792),
    [anon_sym_default] = ACTIONS(2792),
    [anon_sym_while] = ACTIONS(2792),
    [anon_sym_do] = ACTIONS(2792),
    [anon_sym_for] = ACTIONS(2792),
    [anon_sym_in] = ACTIONS(2792),
    [anon_sym_return] = ACTIONS(2792),
    [anon_sym_break] = ACTIONS(2792),
    [anon_sym_continue] = ACTIONS(2792),
    [anon_sym_goto] = ACTIONS(2792),
    [anon_sym_DASH_DASH] = ACTIONS(2794),
    [anon_sym_PLUS_PLUS] = ACTIONS(2794),
    [anon_sym_sizeof] = ACTIONS(2792),
    [anon_sym___alignof__] = ACTIONS(2792),
    [anon_sym___alignof] = ACTIONS(2792),
    [anon_sym__alignof] = ACTIONS(2792),
    [anon_sym_alignof] = ACTIONS(2792),
    [anon_sym__Alignof] = ACTIONS(2792),
    [anon_sym_offsetof] = ACTIONS(2792),
    [anon_sym__Generic] = ACTIONS(2792),
    [anon_sym_asm] = ACTIONS(2792),
    [anon_sym___asm__] = ACTIONS(2792),
    [anon_sym___asm] = ACTIONS(2792),
    [sym_number_literal] = ACTIONS(2794),
    [anon_sym_L_SQUOTE] = ACTIONS(2794),
    [anon_sym_u_SQUOTE] = ACTIONS(2794),
    [anon_sym_U_SQUOTE] = ACTIONS(2794),
    [anon_sym_u8_SQUOTE] = ACTIONS(2794),
    [anon_sym_SQUOTE] = ACTIONS(2794),
    [anon_sym_AT] = ACTIONS(2792),
    [anon_sym_DQUOTE] = ACTIONS(2794),
    [anon_sym_L_DQUOTE] = ACTIONS(2794),
    [anon_sym_u_DQUOTE] = ACTIONS(2794),
    [anon_sym_U_DQUOTE] = ACTIONS(2794),
    [anon_sym_u8_DQUOTE] = ACTIONS(2794),
    [sym_true] = ACTIONS(2792),
    [sym_false] = ACTIONS(2792),
    [anon_sym_NULL] = ACTIONS(2792),
    [anon_sym_nullptr] = ACTIONS(2792),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2792),
    [anon_sym___typeof] = ACTIONS(2792),
    [anon_sym_typeof] = ACTIONS(2792),
    [anon_sym_ATimport] = ACTIONS(2794),
    [aux_sym_preproc_undef_token1] = ACTIONS(2792),
    [anon_sym_POUND] = ACTIONS(2792),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2792),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2792),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2792),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2792),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2792),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2792),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2792),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_API_AVAILABLE] = ACTIONS(2792),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_API_DEPRECATED] = ACTIONS(2792),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2792),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2792),
    [anon_sym___deprecated_msg] = ACTIONS(2792),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2792),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2792),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2792),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2792),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2792),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2792),
    [anon_sym_ATprotocol] = ACTIONS(2794),
    [anon_sym_ATinterface] = ACTIONS(2794),
    [anon_sym_ATimplementation] = ACTIONS(2794),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2794),
    [anon_sym_ATselector] = ACTIONS(2794),
    [anon_sym_ATavailable] = ACTIONS(2794),
    [anon_sym___builtin_available] = ACTIONS(2792),
    [anon_sym_va_arg] = ACTIONS(2792),
    [anon_sym_ATencode] = ACTIONS(2794),
    [anon_sym_BOOL] = ACTIONS(2792),
    [anon_sym_IMP] = ACTIONS(2792),
    [anon_sym_SEL] = ACTIONS(2792),
    [anon_sym_Class] = ACTIONS(2792),
    [anon_sym_id] = ACTIONS(2792),
  },
  [STATE(985)] = {
    [ts_builtin_sym_end] = ACTIONS(2798),
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(986)] = {
    [ts_builtin_sym_end] = ACTIONS(2798),
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(987)] = {
    [ts_builtin_sym_end] = ACTIONS(2798),
    [sym_identifier] = ACTIONS(2796),
    [aux_sym_preproc_include_token1] = ACTIONS(2796),
    [aux_sym_preproc_include_token2] = ACTIONS(2796),
    [aux_sym_preproc_def_token1] = ACTIONS(2796),
    [aux_sym_preproc_if_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2796),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2796),
    [sym_preproc_directive] = ACTIONS(2796),
    [anon_sym_LPAREN2] = ACTIONS(2798),
    [anon_sym_BANG] = ACTIONS(2798),
    [anon_sym_TILDE] = ACTIONS(2798),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_STAR] = ACTIONS(2798),
    [anon_sym_CARET] = ACTIONS(2798),
    [anon_sym_AMP] = ACTIONS(2798),
    [anon_sym_SEMI] = ACTIONS(2798),
    [anon_sym___extension__] = ACTIONS(2796),
    [anon_sym_typedef] = ACTIONS(2796),
    [anon_sym_extern] = ACTIONS(2796),
    [anon_sym___attribute__] = ACTIONS(2796),
    [anon_sym___attribute] = ACTIONS(2796),
    [anon_sym_noreturn] = ACTIONS(2796),
    [anon_sym_LBRACK] = ACTIONS(2798),
    [anon_sym___declspec] = ACTIONS(2796),
    [anon_sym___cdecl] = ACTIONS(2796),
    [anon_sym___clrcall] = ACTIONS(2796),
    [anon_sym___stdcall] = ACTIONS(2796),
    [anon_sym___fastcall] = ACTIONS(2796),
    [anon_sym___thiscall] = ACTIONS(2796),
    [anon_sym___vectorcall] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_signed] = ACTIONS(2796),
    [anon_sym_unsigned] = ACTIONS(2796),
    [anon_sym_long] = ACTIONS(2796),
    [anon_sym_short] = ACTIONS(2796),
    [anon_sym_static] = ACTIONS(2796),
    [anon_sym_ATautoreleasepool] = ACTIONS(2798),
    [anon_sym_auto] = ACTIONS(2796),
    [anon_sym_register] = ACTIONS(2796),
    [anon_sym_inline] = ACTIONS(2796),
    [anon_sym___inline] = ACTIONS(2796),
    [anon_sym___inline__] = ACTIONS(2796),
    [anon_sym___forceinline] = ACTIONS(2796),
    [anon_sym_thread_local] = ACTIONS(2796),
    [anon_sym___thread] = ACTIONS(2796),
    [anon_sym_CG_EXTERN] = ACTIONS(2796),
    [anon_sym_CG_INLINE] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2796),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2796),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2796),
    [anon_sym_IBOutlet] = ACTIONS(2796),
    [anon_sym_IBInspectable] = ACTIONS(2796),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2796),
    [anon_sym_NS_INLINE] = ACTIONS(2796),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2796),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2796),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2796),
    [anon_sym_const] = ACTIONS(2796),
    [anon_sym_constexpr] = ACTIONS(2796),
    [anon_sym_volatile] = ACTIONS(2796),
    [anon_sym_restrict] = ACTIONS(2796),
    [anon_sym___restrict__] = ACTIONS(2796),
    [anon_sym__Atomic] = ACTIONS(2796),
    [anon_sym__Noreturn] = ACTIONS(2796),
    [anon_sym__Nonnull] = ACTIONS(2796),
    [anon_sym_nullable] = ACTIONS(2796),
    [anon_sym__Complex] = ACTIONS(2796),
    [anon_sym__Nullable] = ACTIONS(2796),
    [anon_sym__Nullable_result] = ACTIONS(2796),
    [anon_sym__Null_unspecified] = ACTIONS(2796),
    [anon_sym___autoreleasing] = ACTIONS(2796),
    [anon_sym___block] = ACTIONS(2796),
    [anon_sym___bridge] = ACTIONS(2796),
    [anon_sym___bridge_retained] = ACTIONS(2796),
    [anon_sym___bridge_transfer] = ACTIONS(2796),
    [anon_sym___complex] = ACTIONS(2796),
    [anon_sym___const] = ACTIONS(2796),
    [anon_sym___imag] = ACTIONS(2796),
    [anon_sym___kindof] = ACTIONS(2796),
    [anon_sym___nonnull] = ACTIONS(2796),
    [anon_sym___nullable] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2796),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2796),
    [anon_sym___real] = ACTIONS(2796),
    [anon_sym___strong] = ACTIONS(2796),
    [anon_sym___unsafe_unretained] = ACTIONS(2796),
    [anon_sym___unused] = ACTIONS(2796),
    [anon_sym___weak] = ACTIONS(2796),
    [anon_sym_alignas] = ACTIONS(2796),
    [anon_sym__Alignas] = ACTIONS(2796),
    [sym_primitive_type] = ACTIONS(2796),
    [anon_sym_enum] = ACTIONS(2796),
    [anon_sym_struct] = ACTIONS(2796),
    [anon_sym_union] = ACTIONS(2796),
    [anon_sym_if] = ACTIONS(2796),
    [anon_sym_switch] = ACTIONS(2796),
    [anon_sym_case] = ACTIONS(2796),
    [anon_sym_default] = ACTIONS(2796),
    [anon_sym_while] = ACTIONS(2796),
    [anon_sym_do] = ACTIONS(2796),
    [anon_sym_for] = ACTIONS(2796),
    [anon_sym_in] = ACTIONS(2796),
    [anon_sym_return] = ACTIONS(2796),
    [anon_sym_break] = ACTIONS(2796),
    [anon_sym_continue] = ACTIONS(2796),
    [anon_sym_goto] = ACTIONS(2796),
    [anon_sym_DASH_DASH] = ACTIONS(2798),
    [anon_sym_PLUS_PLUS] = ACTIONS(2798),
    [anon_sym_sizeof] = ACTIONS(2796),
    [anon_sym___alignof__] = ACTIONS(2796),
    [anon_sym___alignof] = ACTIONS(2796),
    [anon_sym__alignof] = ACTIONS(2796),
    [anon_sym_alignof] = ACTIONS(2796),
    [anon_sym__Alignof] = ACTIONS(2796),
    [anon_sym_offsetof] = ACTIONS(2796),
    [anon_sym__Generic] = ACTIONS(2796),
    [anon_sym_asm] = ACTIONS(2796),
    [anon_sym___asm__] = ACTIONS(2796),
    [anon_sym___asm] = ACTIONS(2796),
    [sym_number_literal] = ACTIONS(2798),
    [anon_sym_L_SQUOTE] = ACTIONS(2798),
    [anon_sym_u_SQUOTE] = ACTIONS(2798),
    [anon_sym_U_SQUOTE] = ACTIONS(2798),
    [anon_sym_u8_SQUOTE] = ACTIONS(2798),
    [anon_sym_SQUOTE] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2796),
    [anon_sym_DQUOTE] = ACTIONS(2798),
    [anon_sym_L_DQUOTE] = ACTIONS(2798),
    [anon_sym_u_DQUOTE] = ACTIONS(2798),
    [anon_sym_U_DQUOTE] = ACTIONS(2798),
    [anon_sym_u8_DQUOTE] = ACTIONS(2798),
    [sym_true] = ACTIONS(2796),
    [sym_false] = ACTIONS(2796),
    [anon_sym_NULL] = ACTIONS(2796),
    [anon_sym_nullptr] = ACTIONS(2796),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2796),
    [anon_sym___typeof] = ACTIONS(2796),
    [anon_sym_typeof] = ACTIONS(2796),
    [anon_sym_ATimport] = ACTIONS(2798),
    [aux_sym_preproc_undef_token1] = ACTIONS(2796),
    [anon_sym_POUND] = ACTIONS(2796),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2796),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2796),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2796),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2796),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2796),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_API_AVAILABLE] = ACTIONS(2796),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_API_DEPRECATED] = ACTIONS(2796),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2796),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2796),
    [anon_sym___deprecated_msg] = ACTIONS(2796),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2796),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2796),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2796),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2796),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2796),
    [anon_sym_ATprotocol] = ACTIONS(2798),
    [anon_sym_ATinterface] = ACTIONS(2798),
    [anon_sym_ATimplementation] = ACTIONS(2798),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2798),
    [anon_sym_ATselector] = ACTIONS(2798),
    [anon_sym_ATavailable] = ACTIONS(2798),
    [anon_sym___builtin_available] = ACTIONS(2796),
    [anon_sym_va_arg] = ACTIONS(2796),
    [anon_sym_ATencode] = ACTIONS(2798),
    [anon_sym_BOOL] = ACTIONS(2796),
    [anon_sym_IMP] = ACTIONS(2796),
    [anon_sym_SEL] = ACTIONS(2796),
    [anon_sym_Class] = ACTIONS(2796),
    [anon_sym_id] = ACTIONS(2796),
  },
  [STATE(988)] = {
    [ts_builtin_sym_end] = ACTIONS(2802),
    [sym_identifier] = ACTIONS(2800),
    [aux_sym_preproc_include_token1] = ACTIONS(2800),
    [aux_sym_preproc_include_token2] = ACTIONS(2800),
    [aux_sym_preproc_def_token1] = ACTIONS(2800),
    [aux_sym_preproc_if_token1] = ACTIONS(2800),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2800),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2800),
    [sym_preproc_directive] = ACTIONS(2800),
    [anon_sym_LPAREN2] = ACTIONS(2802),
    [anon_sym_BANG] = ACTIONS(2802),
    [anon_sym_TILDE] = ACTIONS(2802),
    [anon_sym_DASH] = ACTIONS(2800),
    [anon_sym_PLUS] = ACTIONS(2800),
    [anon_sym_STAR] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(2802),
    [anon_sym_AMP] = ACTIONS(2802),
    [anon_sym_SEMI] = ACTIONS(2802),
    [anon_sym___extension__] = ACTIONS(2800),
    [anon_sym_typedef] = ACTIONS(2800),
    [anon_sym_extern] = ACTIONS(2800),
    [anon_sym___attribute__] = ACTIONS(2800),
    [anon_sym___attribute] = ACTIONS(2800),
    [anon_sym_noreturn] = ACTIONS(2800),
    [anon_sym_LBRACK] = ACTIONS(2802),
    [anon_sym___declspec] = ACTIONS(2800),
    [anon_sym___cdecl] = ACTIONS(2800),
    [anon_sym___clrcall] = ACTIONS(2800),
    [anon_sym___stdcall] = ACTIONS(2800),
    [anon_sym___fastcall] = ACTIONS(2800),
    [anon_sym___thiscall] = ACTIONS(2800),
    [anon_sym___vectorcall] = ACTIONS(2800),
    [anon_sym_LBRACE] = ACTIONS(2802),
    [anon_sym_signed] = ACTIONS(2800),
    [anon_sym_unsigned] = ACTIONS(2800),
    [anon_sym_long] = ACTIONS(2800),
    [anon_sym_short] = ACTIONS(2800),
    [anon_sym_static] = ACTIONS(2800),
    [anon_sym_ATautoreleasepool] = ACTIONS(2802),
    [anon_sym_auto] = ACTIONS(2800),
    [anon_sym_register] = ACTIONS(2800),
    [anon_sym_inline] = ACTIONS(2800),
    [anon_sym___inline] = ACTIONS(2800),
    [anon_sym___inline__] = ACTIONS(2800),
    [anon_sym___forceinline] = ACTIONS(2800),
    [anon_sym_thread_local] = ACTIONS(2800),
    [anon_sym___thread] = ACTIONS(2800),
    [anon_sym_CG_EXTERN] = ACTIONS(2800),
    [anon_sym_CG_INLINE] = ACTIONS(2800),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2800),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2800),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2800),
    [anon_sym_IBOutlet] = ACTIONS(2800),
    [anon_sym_IBInspectable] = ACTIONS(2800),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2800),
    [anon_sym_NS_INLINE] = ACTIONS(2800),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2800),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2800),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2800),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2800),
    [anon_sym_const] = ACTIONS(2800),
    [anon_sym_constexpr] = ACTIONS(2800),
    [anon_sym_volatile] = ACTIONS(2800),
    [anon_sym_restrict] = ACTIONS(2800),
    [anon_sym___restrict__] = ACTIONS(2800),
    [anon_sym__Atomic] = ACTIONS(2800),
    [anon_sym__Noreturn] = ACTIONS(2800),
    [anon_sym__Nonnull] = ACTIONS(2800),
    [anon_sym_nullable] = ACTIONS(2800),
    [anon_sym__Complex] = ACTIONS(2800),
    [anon_sym__Nullable] = ACTIONS(2800),
    [anon_sym__Nullable_result] = ACTIONS(2800),
    [anon_sym__Null_unspecified] = ACTIONS(2800),
    [anon_sym___autoreleasing] = ACTIONS(2800),
    [anon_sym___block] = ACTIONS(2800),
    [anon_sym___bridge] = ACTIONS(2800),
    [anon_sym___bridge_retained] = ACTIONS(2800),
    [anon_sym___bridge_transfer] = ACTIONS(2800),
    [anon_sym___complex] = ACTIONS(2800),
    [anon_sym___const] = ACTIONS(2800),
    [anon_sym___imag] = ACTIONS(2800),
    [anon_sym___kindof] = ACTIONS(2800),
    [anon_sym___nonnull] = ACTIONS(2800),
    [anon_sym___nullable] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2800),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2800),
    [anon_sym___real] = ACTIONS(2800),
    [anon_sym___strong] = ACTIONS(2800),
    [anon_sym___unsafe_unretained] = ACTIONS(2800),
    [anon_sym___unused] = ACTIONS(2800),
    [anon_sym___weak] = ACTIONS(2800),
    [anon_sym_alignas] = ACTIONS(2800),
    [anon_sym__Alignas] = ACTIONS(2800),
    [sym_primitive_type] = ACTIONS(2800),
    [anon_sym_enum] = ACTIONS(2800),
    [anon_sym_struct] = ACTIONS(2800),
    [anon_sym_union] = ACTIONS(2800),
    [anon_sym_if] = ACTIONS(2800),
    [anon_sym_switch] = ACTIONS(2800),
    [anon_sym_case] = ACTIONS(2800),
    [anon_sym_default] = ACTIONS(2800),
    [anon_sym_while] = ACTIONS(2800),
    [anon_sym_do] = ACTIONS(2800),
    [anon_sym_for] = ACTIONS(2800),
    [anon_sym_in] = ACTIONS(2800),
    [anon_sym_return] = ACTIONS(2800),
    [anon_sym_break] = ACTIONS(2800),
    [anon_sym_continue] = ACTIONS(2800),
    [anon_sym_goto] = ACTIONS(2800),
    [anon_sym_DASH_DASH] = ACTIONS(2802),
    [anon_sym_PLUS_PLUS] = ACTIONS(2802),
    [anon_sym_sizeof] = ACTIONS(2800),
    [anon_sym___alignof__] = ACTIONS(2800),
    [anon_sym___alignof] = ACTIONS(2800),
    [anon_sym__alignof] = ACTIONS(2800),
    [anon_sym_alignof] = ACTIONS(2800),
    [anon_sym__Alignof] = ACTIONS(2800),
    [anon_sym_offsetof] = ACTIONS(2800),
    [anon_sym__Generic] = ACTIONS(2800),
    [anon_sym_asm] = ACTIONS(2800),
    [anon_sym___asm__] = ACTIONS(2800),
    [anon_sym___asm] = ACTIONS(2800),
    [sym_number_literal] = ACTIONS(2802),
    [anon_sym_L_SQUOTE] = ACTIONS(2802),
    [anon_sym_u_SQUOTE] = ACTIONS(2802),
    [anon_sym_U_SQUOTE] = ACTIONS(2802),
    [anon_sym_u8_SQUOTE] = ACTIONS(2802),
    [anon_sym_SQUOTE] = ACTIONS(2802),
    [anon_sym_AT] = ACTIONS(2800),
    [anon_sym_DQUOTE] = ACTIONS(2802),
    [anon_sym_L_DQUOTE] = ACTIONS(2802),
    [anon_sym_u_DQUOTE] = ACTIONS(2802),
    [anon_sym_U_DQUOTE] = ACTIONS(2802),
    [anon_sym_u8_DQUOTE] = ACTIONS(2802),
    [sym_true] = ACTIONS(2800),
    [sym_false] = ACTIONS(2800),
    [anon_sym_NULL] = ACTIONS(2800),
    [anon_sym_nullptr] = ACTIONS(2800),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2800),
    [anon_sym___typeof] = ACTIONS(2800),
    [anon_sym_typeof] = ACTIONS(2800),
    [anon_sym_ATimport] = ACTIONS(2802),
    [aux_sym_preproc_undef_token1] = ACTIONS(2800),
    [anon_sym_POUND] = ACTIONS(2800),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2800),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2800),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2800),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2800),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2800),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2800),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2800),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_API_AVAILABLE] = ACTIONS(2800),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_API_DEPRECATED] = ACTIONS(2800),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2800),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2800),
    [anon_sym___deprecated_msg] = ACTIONS(2800),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2800),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2800),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2800),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2800),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2800),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2800),
    [anon_sym_ATprotocol] = ACTIONS(2802),
    [anon_sym_ATinterface] = ACTIONS(2802),
    [anon_sym_ATimplementation] = ACTIONS(2802),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2802),
    [anon_sym_ATselector] = ACTIONS(2802),
    [anon_sym_ATavailable] = ACTIONS(2802),
    [anon_sym___builtin_available] = ACTIONS(2800),
    [anon_sym_va_arg] = ACTIONS(2800),
    [anon_sym_ATencode] = ACTIONS(2802),
    [anon_sym_BOOL] = ACTIONS(2800),
    [anon_sym_IMP] = ACTIONS(2800),
    [anon_sym_SEL] = ACTIONS(2800),
    [anon_sym_Class] = ACTIONS(2800),
    [anon_sym_id] = ACTIONS(2800),
  },
  [STATE(989)] = {
    [ts_builtin_sym_end] = ACTIONS(2806),
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(990)] = {
    [ts_builtin_sym_end] = ACTIONS(2806),
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(991)] = {
    [ts_builtin_sym_end] = ACTIONS(2806),
    [sym_identifier] = ACTIONS(2804),
    [aux_sym_preproc_include_token1] = ACTIONS(2804),
    [aux_sym_preproc_include_token2] = ACTIONS(2804),
    [aux_sym_preproc_def_token1] = ACTIONS(2804),
    [aux_sym_preproc_if_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2804),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2804),
    [sym_preproc_directive] = ACTIONS(2804),
    [anon_sym_LPAREN2] = ACTIONS(2806),
    [anon_sym_BANG] = ACTIONS(2806),
    [anon_sym_TILDE] = ACTIONS(2806),
    [anon_sym_DASH] = ACTIONS(2804),
    [anon_sym_PLUS] = ACTIONS(2804),
    [anon_sym_STAR] = ACTIONS(2806),
    [anon_sym_CARET] = ACTIONS(2806),
    [anon_sym_AMP] = ACTIONS(2806),
    [anon_sym_SEMI] = ACTIONS(2806),
    [anon_sym___extension__] = ACTIONS(2804),
    [anon_sym_typedef] = ACTIONS(2804),
    [anon_sym_extern] = ACTIONS(2804),
    [anon_sym___attribute__] = ACTIONS(2804),
    [anon_sym___attribute] = ACTIONS(2804),
    [anon_sym_noreturn] = ACTIONS(2804),
    [anon_sym_LBRACK] = ACTIONS(2806),
    [anon_sym___declspec] = ACTIONS(2804),
    [anon_sym___cdecl] = ACTIONS(2804),
    [anon_sym___clrcall] = ACTIONS(2804),
    [anon_sym___stdcall] = ACTIONS(2804),
    [anon_sym___fastcall] = ACTIONS(2804),
    [anon_sym___thiscall] = ACTIONS(2804),
    [anon_sym___vectorcall] = ACTIONS(2804),
    [anon_sym_LBRACE] = ACTIONS(2806),
    [anon_sym_signed] = ACTIONS(2804),
    [anon_sym_unsigned] = ACTIONS(2804),
    [anon_sym_long] = ACTIONS(2804),
    [anon_sym_short] = ACTIONS(2804),
    [anon_sym_static] = ACTIONS(2804),
    [anon_sym_ATautoreleasepool] = ACTIONS(2806),
    [anon_sym_auto] = ACTIONS(2804),
    [anon_sym_register] = ACTIONS(2804),
    [anon_sym_inline] = ACTIONS(2804),
    [anon_sym___inline] = ACTIONS(2804),
    [anon_sym___inline__] = ACTIONS(2804),
    [anon_sym___forceinline] = ACTIONS(2804),
    [anon_sym_thread_local] = ACTIONS(2804),
    [anon_sym___thread] = ACTIONS(2804),
    [anon_sym_CG_EXTERN] = ACTIONS(2804),
    [anon_sym_CG_INLINE] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2804),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2804),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2804),
    [anon_sym_IBOutlet] = ACTIONS(2804),
    [anon_sym_IBInspectable] = ACTIONS(2804),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2804),
    [anon_sym_NS_INLINE] = ACTIONS(2804),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2804),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2804),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2804),
    [anon_sym_const] = ACTIONS(2804),
    [anon_sym_constexpr] = ACTIONS(2804),
    [anon_sym_volatile] = ACTIONS(2804),
    [anon_sym_restrict] = ACTIONS(2804),
    [anon_sym___restrict__] = ACTIONS(2804),
    [anon_sym__Atomic] = ACTIONS(2804),
    [anon_sym__Noreturn] = ACTIONS(2804),
    [anon_sym__Nonnull] = ACTIONS(2804),
    [anon_sym_nullable] = ACTIONS(2804),
    [anon_sym__Complex] = ACTIONS(2804),
    [anon_sym__Nullable] = ACTIONS(2804),
    [anon_sym__Nullable_result] = ACTIONS(2804),
    [anon_sym__Null_unspecified] = ACTIONS(2804),
    [anon_sym___autoreleasing] = ACTIONS(2804),
    [anon_sym___block] = ACTIONS(2804),
    [anon_sym___bridge] = ACTIONS(2804),
    [anon_sym___bridge_retained] = ACTIONS(2804),
    [anon_sym___bridge_transfer] = ACTIONS(2804),
    [anon_sym___complex] = ACTIONS(2804),
    [anon_sym___const] = ACTIONS(2804),
    [anon_sym___imag] = ACTIONS(2804),
    [anon_sym___kindof] = ACTIONS(2804),
    [anon_sym___nonnull] = ACTIONS(2804),
    [anon_sym___nullable] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2804),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2804),
    [anon_sym___real] = ACTIONS(2804),
    [anon_sym___strong] = ACTIONS(2804),
    [anon_sym___unsafe_unretained] = ACTIONS(2804),
    [anon_sym___unused] = ACTIONS(2804),
    [anon_sym___weak] = ACTIONS(2804),
    [anon_sym_alignas] = ACTIONS(2804),
    [anon_sym__Alignas] = ACTIONS(2804),
    [sym_primitive_type] = ACTIONS(2804),
    [anon_sym_enum] = ACTIONS(2804),
    [anon_sym_struct] = ACTIONS(2804),
    [anon_sym_union] = ACTIONS(2804),
    [anon_sym_if] = ACTIONS(2804),
    [anon_sym_switch] = ACTIONS(2804),
    [anon_sym_case] = ACTIONS(2804),
    [anon_sym_default] = ACTIONS(2804),
    [anon_sym_while] = ACTIONS(2804),
    [anon_sym_do] = ACTIONS(2804),
    [anon_sym_for] = ACTIONS(2804),
    [anon_sym_in] = ACTIONS(2804),
    [anon_sym_return] = ACTIONS(2804),
    [anon_sym_break] = ACTIONS(2804),
    [anon_sym_continue] = ACTIONS(2804),
    [anon_sym_goto] = ACTIONS(2804),
    [anon_sym_DASH_DASH] = ACTIONS(2806),
    [anon_sym_PLUS_PLUS] = ACTIONS(2806),
    [anon_sym_sizeof] = ACTIONS(2804),
    [anon_sym___alignof__] = ACTIONS(2804),
    [anon_sym___alignof] = ACTIONS(2804),
    [anon_sym__alignof] = ACTIONS(2804),
    [anon_sym_alignof] = ACTIONS(2804),
    [anon_sym__Alignof] = ACTIONS(2804),
    [anon_sym_offsetof] = ACTIONS(2804),
    [anon_sym__Generic] = ACTIONS(2804),
    [anon_sym_asm] = ACTIONS(2804),
    [anon_sym___asm__] = ACTIONS(2804),
    [anon_sym___asm] = ACTIONS(2804),
    [sym_number_literal] = ACTIONS(2806),
    [anon_sym_L_SQUOTE] = ACTIONS(2806),
    [anon_sym_u_SQUOTE] = ACTIONS(2806),
    [anon_sym_U_SQUOTE] = ACTIONS(2806),
    [anon_sym_u8_SQUOTE] = ACTIONS(2806),
    [anon_sym_SQUOTE] = ACTIONS(2806),
    [anon_sym_AT] = ACTIONS(2804),
    [anon_sym_DQUOTE] = ACTIONS(2806),
    [anon_sym_L_DQUOTE] = ACTIONS(2806),
    [anon_sym_u_DQUOTE] = ACTIONS(2806),
    [anon_sym_U_DQUOTE] = ACTIONS(2806),
    [anon_sym_u8_DQUOTE] = ACTIONS(2806),
    [sym_true] = ACTIONS(2804),
    [sym_false] = ACTIONS(2804),
    [anon_sym_NULL] = ACTIONS(2804),
    [anon_sym_nullptr] = ACTIONS(2804),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2804),
    [anon_sym___typeof] = ACTIONS(2804),
    [anon_sym_typeof] = ACTIONS(2804),
    [anon_sym_ATimport] = ACTIONS(2806),
    [aux_sym_preproc_undef_token1] = ACTIONS(2804),
    [anon_sym_POUND] = ACTIONS(2804),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2804),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2804),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2804),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2804),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2804),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_API_AVAILABLE] = ACTIONS(2804),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_API_DEPRECATED] = ACTIONS(2804),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2804),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2804),
    [anon_sym___deprecated_msg] = ACTIONS(2804),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2804),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2804),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2804),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2804),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2804),
    [anon_sym_ATprotocol] = ACTIONS(2806),
    [anon_sym_ATinterface] = ACTIONS(2806),
    [anon_sym_ATimplementation] = ACTIONS(2806),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2806),
    [anon_sym_ATselector] = ACTIONS(2806),
    [anon_sym_ATavailable] = ACTIONS(2806),
    [anon_sym___builtin_available] = ACTIONS(2804),
    [anon_sym_va_arg] = ACTIONS(2804),
    [anon_sym_ATencode] = ACTIONS(2806),
    [anon_sym_BOOL] = ACTIONS(2804),
    [anon_sym_IMP] = ACTIONS(2804),
    [anon_sym_SEL] = ACTIONS(2804),
    [anon_sym_Class] = ACTIONS(2804),
    [anon_sym_id] = ACTIONS(2804),
  },
  [STATE(992)] = {
    [sym_ms_declspec_modifier] = STATE(6760),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2926),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(993)] = {
    [ts_builtin_sym_end] = ACTIONS(2810),
    [sym_identifier] = ACTIONS(2808),
    [aux_sym_preproc_include_token1] = ACTIONS(2808),
    [aux_sym_preproc_include_token2] = ACTIONS(2808),
    [aux_sym_preproc_def_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2808),
    [sym_preproc_directive] = ACTIONS(2808),
    [anon_sym_LPAREN2] = ACTIONS(2810),
    [anon_sym_BANG] = ACTIONS(2810),
    [anon_sym_TILDE] = ACTIONS(2810),
    [anon_sym_DASH] = ACTIONS(2808),
    [anon_sym_PLUS] = ACTIONS(2808),
    [anon_sym_STAR] = ACTIONS(2810),
    [anon_sym_CARET] = ACTIONS(2810),
    [anon_sym_AMP] = ACTIONS(2810),
    [anon_sym_SEMI] = ACTIONS(2810),
    [anon_sym___extension__] = ACTIONS(2808),
    [anon_sym_typedef] = ACTIONS(2808),
    [anon_sym_extern] = ACTIONS(2808),
    [anon_sym___attribute__] = ACTIONS(2808),
    [anon_sym___attribute] = ACTIONS(2808),
    [anon_sym_noreturn] = ACTIONS(2808),
    [anon_sym_LBRACK] = ACTIONS(2810),
    [anon_sym___declspec] = ACTIONS(2808),
    [anon_sym___cdecl] = ACTIONS(2808),
    [anon_sym___clrcall] = ACTIONS(2808),
    [anon_sym___stdcall] = ACTIONS(2808),
    [anon_sym___fastcall] = ACTIONS(2808),
    [anon_sym___thiscall] = ACTIONS(2808),
    [anon_sym___vectorcall] = ACTIONS(2808),
    [anon_sym_LBRACE] = ACTIONS(2810),
    [anon_sym_signed] = ACTIONS(2808),
    [anon_sym_unsigned] = ACTIONS(2808),
    [anon_sym_long] = ACTIONS(2808),
    [anon_sym_short] = ACTIONS(2808),
    [anon_sym_static] = ACTIONS(2808),
    [anon_sym_ATautoreleasepool] = ACTIONS(2810),
    [anon_sym_auto] = ACTIONS(2808),
    [anon_sym_register] = ACTIONS(2808),
    [anon_sym_inline] = ACTIONS(2808),
    [anon_sym___inline] = ACTIONS(2808),
    [anon_sym___inline__] = ACTIONS(2808),
    [anon_sym___forceinline] = ACTIONS(2808),
    [anon_sym_thread_local] = ACTIONS(2808),
    [anon_sym___thread] = ACTIONS(2808),
    [anon_sym_CG_EXTERN] = ACTIONS(2808),
    [anon_sym_CG_INLINE] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2808),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2808),
    [anon_sym_IBOutlet] = ACTIONS(2808),
    [anon_sym_IBInspectable] = ACTIONS(2808),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2808),
    [anon_sym_NS_INLINE] = ACTIONS(2808),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2808),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2808),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2808),
    [anon_sym_const] = ACTIONS(2808),
    [anon_sym_constexpr] = ACTIONS(2808),
    [anon_sym_volatile] = ACTIONS(2808),
    [anon_sym_restrict] = ACTIONS(2808),
    [anon_sym___restrict__] = ACTIONS(2808),
    [anon_sym__Atomic] = ACTIONS(2808),
    [anon_sym__Noreturn] = ACTIONS(2808),
    [anon_sym__Nonnull] = ACTIONS(2808),
    [anon_sym_nullable] = ACTIONS(2808),
    [anon_sym__Complex] = ACTIONS(2808),
    [anon_sym__Nullable] = ACTIONS(2808),
    [anon_sym__Nullable_result] = ACTIONS(2808),
    [anon_sym__Null_unspecified] = ACTIONS(2808),
    [anon_sym___autoreleasing] = ACTIONS(2808),
    [anon_sym___block] = ACTIONS(2808),
    [anon_sym___bridge] = ACTIONS(2808),
    [anon_sym___bridge_retained] = ACTIONS(2808),
    [anon_sym___bridge_transfer] = ACTIONS(2808),
    [anon_sym___complex] = ACTIONS(2808),
    [anon_sym___const] = ACTIONS(2808),
    [anon_sym___imag] = ACTIONS(2808),
    [anon_sym___kindof] = ACTIONS(2808),
    [anon_sym___nonnull] = ACTIONS(2808),
    [anon_sym___nullable] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2808),
    [anon_sym___real] = ACTIONS(2808),
    [anon_sym___strong] = ACTIONS(2808),
    [anon_sym___unsafe_unretained] = ACTIONS(2808),
    [anon_sym___unused] = ACTIONS(2808),
    [anon_sym___weak] = ACTIONS(2808),
    [anon_sym_alignas] = ACTIONS(2808),
    [anon_sym__Alignas] = ACTIONS(2808),
    [sym_primitive_type] = ACTIONS(2808),
    [anon_sym_enum] = ACTIONS(2808),
    [anon_sym_struct] = ACTIONS(2808),
    [anon_sym_union] = ACTIONS(2808),
    [anon_sym_if] = ACTIONS(2808),
    [anon_sym_switch] = ACTIONS(2808),
    [anon_sym_case] = ACTIONS(2808),
    [anon_sym_default] = ACTIONS(2808),
    [anon_sym_while] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2808),
    [anon_sym_in] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_break] = ACTIONS(2808),
    [anon_sym_continue] = ACTIONS(2808),
    [anon_sym_goto] = ACTIONS(2808),
    [anon_sym_DASH_DASH] = ACTIONS(2810),
    [anon_sym_PLUS_PLUS] = ACTIONS(2810),
    [anon_sym_sizeof] = ACTIONS(2808),
    [anon_sym___alignof__] = ACTIONS(2808),
    [anon_sym___alignof] = ACTIONS(2808),
    [anon_sym__alignof] = ACTIONS(2808),
    [anon_sym_alignof] = ACTIONS(2808),
    [anon_sym__Alignof] = ACTIONS(2808),
    [anon_sym_offsetof] = ACTIONS(2808),
    [anon_sym__Generic] = ACTIONS(2808),
    [anon_sym_asm] = ACTIONS(2808),
    [anon_sym___asm__] = ACTIONS(2808),
    [anon_sym___asm] = ACTIONS(2808),
    [sym_number_literal] = ACTIONS(2810),
    [anon_sym_L_SQUOTE] = ACTIONS(2810),
    [anon_sym_u_SQUOTE] = ACTIONS(2810),
    [anon_sym_U_SQUOTE] = ACTIONS(2810),
    [anon_sym_u8_SQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2810),
    [anon_sym_AT] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_L_DQUOTE] = ACTIONS(2810),
    [anon_sym_u_DQUOTE] = ACTIONS(2810),
    [anon_sym_U_DQUOTE] = ACTIONS(2810),
    [anon_sym_u8_DQUOTE] = ACTIONS(2810),
    [sym_true] = ACTIONS(2808),
    [sym_false] = ACTIONS(2808),
    [anon_sym_NULL] = ACTIONS(2808),
    [anon_sym_nullptr] = ACTIONS(2808),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2808),
    [anon_sym___typeof] = ACTIONS(2808),
    [anon_sym_typeof] = ACTIONS(2808),
    [anon_sym_ATimport] = ACTIONS(2810),
    [aux_sym_preproc_undef_token1] = ACTIONS(2808),
    [anon_sym_POUND] = ACTIONS(2808),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2808),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2808),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2808),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_API_AVAILABLE] = ACTIONS(2808),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_API_DEPRECATED] = ACTIONS(2808),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym___deprecated_msg] = ACTIONS(2808),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2808),
    [anon_sym_ATprotocol] = ACTIONS(2810),
    [anon_sym_ATinterface] = ACTIONS(2810),
    [anon_sym_ATimplementation] = ACTIONS(2810),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2810),
    [anon_sym_ATselector] = ACTIONS(2810),
    [anon_sym_ATavailable] = ACTIONS(2810),
    [anon_sym___builtin_available] = ACTIONS(2808),
    [anon_sym_va_arg] = ACTIONS(2808),
    [anon_sym_ATencode] = ACTIONS(2810),
    [anon_sym_BOOL] = ACTIONS(2808),
    [anon_sym_IMP] = ACTIONS(2808),
    [anon_sym_SEL] = ACTIONS(2808),
    [anon_sym_Class] = ACTIONS(2808),
    [anon_sym_id] = ACTIONS(2808),
  },
  [STATE(994)] = {
    [ts_builtin_sym_end] = ACTIONS(2814),
    [sym_identifier] = ACTIONS(2812),
    [aux_sym_preproc_include_token1] = ACTIONS(2812),
    [aux_sym_preproc_include_token2] = ACTIONS(2812),
    [aux_sym_preproc_def_token1] = ACTIONS(2812),
    [aux_sym_preproc_if_token1] = ACTIONS(2812),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2812),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2812),
    [sym_preproc_directive] = ACTIONS(2812),
    [anon_sym_LPAREN2] = ACTIONS(2814),
    [anon_sym_BANG] = ACTIONS(2814),
    [anon_sym_TILDE] = ACTIONS(2814),
    [anon_sym_DASH] = ACTIONS(2812),
    [anon_sym_PLUS] = ACTIONS(2812),
    [anon_sym_STAR] = ACTIONS(2814),
    [anon_sym_CARET] = ACTIONS(2814),
    [anon_sym_AMP] = ACTIONS(2814),
    [anon_sym_SEMI] = ACTIONS(2814),
    [anon_sym___extension__] = ACTIONS(2812),
    [anon_sym_typedef] = ACTIONS(2812),
    [anon_sym_extern] = ACTIONS(2812),
    [anon_sym___attribute__] = ACTIONS(2812),
    [anon_sym___attribute] = ACTIONS(2812),
    [anon_sym_noreturn] = ACTIONS(2812),
    [anon_sym_LBRACK] = ACTIONS(2814),
    [anon_sym___declspec] = ACTIONS(2812),
    [anon_sym___cdecl] = ACTIONS(2812),
    [anon_sym___clrcall] = ACTIONS(2812),
    [anon_sym___stdcall] = ACTIONS(2812),
    [anon_sym___fastcall] = ACTIONS(2812),
    [anon_sym___thiscall] = ACTIONS(2812),
    [anon_sym___vectorcall] = ACTIONS(2812),
    [anon_sym_LBRACE] = ACTIONS(2814),
    [anon_sym_signed] = ACTIONS(2812),
    [anon_sym_unsigned] = ACTIONS(2812),
    [anon_sym_long] = ACTIONS(2812),
    [anon_sym_short] = ACTIONS(2812),
    [anon_sym_static] = ACTIONS(2812),
    [anon_sym_ATautoreleasepool] = ACTIONS(2814),
    [anon_sym_auto] = ACTIONS(2812),
    [anon_sym_register] = ACTIONS(2812),
    [anon_sym_inline] = ACTIONS(2812),
    [anon_sym___inline] = ACTIONS(2812),
    [anon_sym___inline__] = ACTIONS(2812),
    [anon_sym___forceinline] = ACTIONS(2812),
    [anon_sym_thread_local] = ACTIONS(2812),
    [anon_sym___thread] = ACTIONS(2812),
    [anon_sym_CG_EXTERN] = ACTIONS(2812),
    [anon_sym_CG_INLINE] = ACTIONS(2812),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2812),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2812),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2812),
    [anon_sym_IBOutlet] = ACTIONS(2812),
    [anon_sym_IBInspectable] = ACTIONS(2812),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2812),
    [anon_sym_NS_INLINE] = ACTIONS(2812),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2812),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2812),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2812),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2812),
    [anon_sym_const] = ACTIONS(2812),
    [anon_sym_constexpr] = ACTIONS(2812),
    [anon_sym_volatile] = ACTIONS(2812),
    [anon_sym_restrict] = ACTIONS(2812),
    [anon_sym___restrict__] = ACTIONS(2812),
    [anon_sym__Atomic] = ACTIONS(2812),
    [anon_sym__Noreturn] = ACTIONS(2812),
    [anon_sym__Nonnull] = ACTIONS(2812),
    [anon_sym_nullable] = ACTIONS(2812),
    [anon_sym__Complex] = ACTIONS(2812),
    [anon_sym__Nullable] = ACTIONS(2812),
    [anon_sym__Nullable_result] = ACTIONS(2812),
    [anon_sym__Null_unspecified] = ACTIONS(2812),
    [anon_sym___autoreleasing] = ACTIONS(2812),
    [anon_sym___block] = ACTIONS(2812),
    [anon_sym___bridge] = ACTIONS(2812),
    [anon_sym___bridge_retained] = ACTIONS(2812),
    [anon_sym___bridge_transfer] = ACTIONS(2812),
    [anon_sym___complex] = ACTIONS(2812),
    [anon_sym___const] = ACTIONS(2812),
    [anon_sym___imag] = ACTIONS(2812),
    [anon_sym___kindof] = ACTIONS(2812),
    [anon_sym___nonnull] = ACTIONS(2812),
    [anon_sym___nullable] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2812),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2812),
    [anon_sym___real] = ACTIONS(2812),
    [anon_sym___strong] = ACTIONS(2812),
    [anon_sym___unsafe_unretained] = ACTIONS(2812),
    [anon_sym___unused] = ACTIONS(2812),
    [anon_sym___weak] = ACTIONS(2812),
    [anon_sym_alignas] = ACTIONS(2812),
    [anon_sym__Alignas] = ACTIONS(2812),
    [sym_primitive_type] = ACTIONS(2812),
    [anon_sym_enum] = ACTIONS(2812),
    [anon_sym_struct] = ACTIONS(2812),
    [anon_sym_union] = ACTIONS(2812),
    [anon_sym_if] = ACTIONS(2812),
    [anon_sym_switch] = ACTIONS(2812),
    [anon_sym_case] = ACTIONS(2812),
    [anon_sym_default] = ACTIONS(2812),
    [anon_sym_while] = ACTIONS(2812),
    [anon_sym_do] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2812),
    [anon_sym_in] = ACTIONS(2812),
    [anon_sym_return] = ACTIONS(2812),
    [anon_sym_break] = ACTIONS(2812),
    [anon_sym_continue] = ACTIONS(2812),
    [anon_sym_goto] = ACTIONS(2812),
    [anon_sym_DASH_DASH] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_sizeof] = ACTIONS(2812),
    [anon_sym___alignof__] = ACTIONS(2812),
    [anon_sym___alignof] = ACTIONS(2812),
    [anon_sym__alignof] = ACTIONS(2812),
    [anon_sym_alignof] = ACTIONS(2812),
    [anon_sym__Alignof] = ACTIONS(2812),
    [anon_sym_offsetof] = ACTIONS(2812),
    [anon_sym__Generic] = ACTIONS(2812),
    [anon_sym_asm] = ACTIONS(2812),
    [anon_sym___asm__] = ACTIONS(2812),
    [anon_sym___asm] = ACTIONS(2812),
    [sym_number_literal] = ACTIONS(2814),
    [anon_sym_L_SQUOTE] = ACTIONS(2814),
    [anon_sym_u_SQUOTE] = ACTIONS(2814),
    [anon_sym_U_SQUOTE] = ACTIONS(2814),
    [anon_sym_u8_SQUOTE] = ACTIONS(2814),
    [anon_sym_SQUOTE] = ACTIONS(2814),
    [anon_sym_AT] = ACTIONS(2812),
    [anon_sym_DQUOTE] = ACTIONS(2814),
    [anon_sym_L_DQUOTE] = ACTIONS(2814),
    [anon_sym_u_DQUOTE] = ACTIONS(2814),
    [anon_sym_U_DQUOTE] = ACTIONS(2814),
    [anon_sym_u8_DQUOTE] = ACTIONS(2814),
    [sym_true] = ACTIONS(2812),
    [sym_false] = ACTIONS(2812),
    [anon_sym_NULL] = ACTIONS(2812),
    [anon_sym_nullptr] = ACTIONS(2812),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2812),
    [anon_sym___typeof] = ACTIONS(2812),
    [anon_sym_typeof] = ACTIONS(2812),
    [anon_sym_ATimport] = ACTIONS(2814),
    [aux_sym_preproc_undef_token1] = ACTIONS(2812),
    [anon_sym_POUND] = ACTIONS(2812),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2812),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2812),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2812),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2812),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2812),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2812),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2812),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_API_AVAILABLE] = ACTIONS(2812),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_API_DEPRECATED] = ACTIONS(2812),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2812),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2812),
    [anon_sym___deprecated_msg] = ACTIONS(2812),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2812),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2812),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2812),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2812),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2812),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2812),
    [anon_sym_ATprotocol] = ACTIONS(2814),
    [anon_sym_ATinterface] = ACTIONS(2814),
    [anon_sym_ATimplementation] = ACTIONS(2814),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2814),
    [anon_sym_ATselector] = ACTIONS(2814),
    [anon_sym_ATavailable] = ACTIONS(2814),
    [anon_sym___builtin_available] = ACTIONS(2812),
    [anon_sym_va_arg] = ACTIONS(2812),
    [anon_sym_ATencode] = ACTIONS(2814),
    [anon_sym_BOOL] = ACTIONS(2812),
    [anon_sym_IMP] = ACTIONS(2812),
    [anon_sym_SEL] = ACTIONS(2812),
    [anon_sym_Class] = ACTIONS(2812),
    [anon_sym_id] = ACTIONS(2812),
  },
  [STATE(995)] = {
    [ts_builtin_sym_end] = ACTIONS(2818),
    [sym_identifier] = ACTIONS(2816),
    [aux_sym_preproc_include_token1] = ACTIONS(2816),
    [aux_sym_preproc_include_token2] = ACTIONS(2816),
    [aux_sym_preproc_def_token1] = ACTIONS(2816),
    [aux_sym_preproc_if_token1] = ACTIONS(2816),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2816),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2816),
    [sym_preproc_directive] = ACTIONS(2816),
    [anon_sym_LPAREN2] = ACTIONS(2818),
    [anon_sym_BANG] = ACTIONS(2818),
    [anon_sym_TILDE] = ACTIONS(2818),
    [anon_sym_DASH] = ACTIONS(2816),
    [anon_sym_PLUS] = ACTIONS(2816),
    [anon_sym_STAR] = ACTIONS(2818),
    [anon_sym_CARET] = ACTIONS(2818),
    [anon_sym_AMP] = ACTIONS(2818),
    [anon_sym_SEMI] = ACTIONS(2818),
    [anon_sym___extension__] = ACTIONS(2816),
    [anon_sym_typedef] = ACTIONS(2816),
    [anon_sym_extern] = ACTIONS(2816),
    [anon_sym___attribute__] = ACTIONS(2816),
    [anon_sym___attribute] = ACTIONS(2816),
    [anon_sym_noreturn] = ACTIONS(2816),
    [anon_sym_LBRACK] = ACTIONS(2818),
    [anon_sym___declspec] = ACTIONS(2816),
    [anon_sym___cdecl] = ACTIONS(2816),
    [anon_sym___clrcall] = ACTIONS(2816),
    [anon_sym___stdcall] = ACTIONS(2816),
    [anon_sym___fastcall] = ACTIONS(2816),
    [anon_sym___thiscall] = ACTIONS(2816),
    [anon_sym___vectorcall] = ACTIONS(2816),
    [anon_sym_LBRACE] = ACTIONS(2818),
    [anon_sym_signed] = ACTIONS(2816),
    [anon_sym_unsigned] = ACTIONS(2816),
    [anon_sym_long] = ACTIONS(2816),
    [anon_sym_short] = ACTIONS(2816),
    [anon_sym_static] = ACTIONS(2816),
    [anon_sym_ATautoreleasepool] = ACTIONS(2818),
    [anon_sym_auto] = ACTIONS(2816),
    [anon_sym_register] = ACTIONS(2816),
    [anon_sym_inline] = ACTIONS(2816),
    [anon_sym___inline] = ACTIONS(2816),
    [anon_sym___inline__] = ACTIONS(2816),
    [anon_sym___forceinline] = ACTIONS(2816),
    [anon_sym_thread_local] = ACTIONS(2816),
    [anon_sym___thread] = ACTIONS(2816),
    [anon_sym_CG_EXTERN] = ACTIONS(2816),
    [anon_sym_CG_INLINE] = ACTIONS(2816),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2816),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2816),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2816),
    [anon_sym_IBOutlet] = ACTIONS(2816),
    [anon_sym_IBInspectable] = ACTIONS(2816),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2816),
    [anon_sym_NS_INLINE] = ACTIONS(2816),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2816),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2816),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2816),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2816),
    [anon_sym_const] = ACTIONS(2816),
    [anon_sym_constexpr] = ACTIONS(2816),
    [anon_sym_volatile] = ACTIONS(2816),
    [anon_sym_restrict] = ACTIONS(2816),
    [anon_sym___restrict__] = ACTIONS(2816),
    [anon_sym__Atomic] = ACTIONS(2816),
    [anon_sym__Noreturn] = ACTIONS(2816),
    [anon_sym__Nonnull] = ACTIONS(2816),
    [anon_sym_nullable] = ACTIONS(2816),
    [anon_sym__Complex] = ACTIONS(2816),
    [anon_sym__Nullable] = ACTIONS(2816),
    [anon_sym__Nullable_result] = ACTIONS(2816),
    [anon_sym__Null_unspecified] = ACTIONS(2816),
    [anon_sym___autoreleasing] = ACTIONS(2816),
    [anon_sym___block] = ACTIONS(2816),
    [anon_sym___bridge] = ACTIONS(2816),
    [anon_sym___bridge_retained] = ACTIONS(2816),
    [anon_sym___bridge_transfer] = ACTIONS(2816),
    [anon_sym___complex] = ACTIONS(2816),
    [anon_sym___const] = ACTIONS(2816),
    [anon_sym___imag] = ACTIONS(2816),
    [anon_sym___kindof] = ACTIONS(2816),
    [anon_sym___nonnull] = ACTIONS(2816),
    [anon_sym___nullable] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2816),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2816),
    [anon_sym___real] = ACTIONS(2816),
    [anon_sym___strong] = ACTIONS(2816),
    [anon_sym___unsafe_unretained] = ACTIONS(2816),
    [anon_sym___unused] = ACTIONS(2816),
    [anon_sym___weak] = ACTIONS(2816),
    [anon_sym_alignas] = ACTIONS(2816),
    [anon_sym__Alignas] = ACTIONS(2816),
    [sym_primitive_type] = ACTIONS(2816),
    [anon_sym_enum] = ACTIONS(2816),
    [anon_sym_struct] = ACTIONS(2816),
    [anon_sym_union] = ACTIONS(2816),
    [anon_sym_if] = ACTIONS(2816),
    [anon_sym_switch] = ACTIONS(2816),
    [anon_sym_case] = ACTIONS(2816),
    [anon_sym_default] = ACTIONS(2816),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_do] = ACTIONS(2816),
    [anon_sym_for] = ACTIONS(2816),
    [anon_sym_in] = ACTIONS(2816),
    [anon_sym_return] = ACTIONS(2816),
    [anon_sym_break] = ACTIONS(2816),
    [anon_sym_continue] = ACTIONS(2816),
    [anon_sym_goto] = ACTIONS(2816),
    [anon_sym_DASH_DASH] = ACTIONS(2818),
    [anon_sym_PLUS_PLUS] = ACTIONS(2818),
    [anon_sym_sizeof] = ACTIONS(2816),
    [anon_sym___alignof__] = ACTIONS(2816),
    [anon_sym___alignof] = ACTIONS(2816),
    [anon_sym__alignof] = ACTIONS(2816),
    [anon_sym_alignof] = ACTIONS(2816),
    [anon_sym__Alignof] = ACTIONS(2816),
    [anon_sym_offsetof] = ACTIONS(2816),
    [anon_sym__Generic] = ACTIONS(2816),
    [anon_sym_asm] = ACTIONS(2816),
    [anon_sym___asm__] = ACTIONS(2816),
    [anon_sym___asm] = ACTIONS(2816),
    [sym_number_literal] = ACTIONS(2818),
    [anon_sym_L_SQUOTE] = ACTIONS(2818),
    [anon_sym_u_SQUOTE] = ACTIONS(2818),
    [anon_sym_U_SQUOTE] = ACTIONS(2818),
    [anon_sym_u8_SQUOTE] = ACTIONS(2818),
    [anon_sym_SQUOTE] = ACTIONS(2818),
    [anon_sym_AT] = ACTIONS(2816),
    [anon_sym_DQUOTE] = ACTIONS(2818),
    [anon_sym_L_DQUOTE] = ACTIONS(2818),
    [anon_sym_u_DQUOTE] = ACTIONS(2818),
    [anon_sym_U_DQUOTE] = ACTIONS(2818),
    [anon_sym_u8_DQUOTE] = ACTIONS(2818),
    [sym_true] = ACTIONS(2816),
    [sym_false] = ACTIONS(2816),
    [anon_sym_NULL] = ACTIONS(2816),
    [anon_sym_nullptr] = ACTIONS(2816),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2816),
    [anon_sym___typeof] = ACTIONS(2816),
    [anon_sym_typeof] = ACTIONS(2816),
    [anon_sym_ATimport] = ACTIONS(2818),
    [aux_sym_preproc_undef_token1] = ACTIONS(2816),
    [anon_sym_POUND] = ACTIONS(2816),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2816),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2816),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2816),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2816),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2816),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2816),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2816),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_API_AVAILABLE] = ACTIONS(2816),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_API_DEPRECATED] = ACTIONS(2816),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2816),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2816),
    [anon_sym___deprecated_msg] = ACTIONS(2816),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2816),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2816),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2816),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2816),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2816),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2816),
    [anon_sym_ATprotocol] = ACTIONS(2818),
    [anon_sym_ATinterface] = ACTIONS(2818),
    [anon_sym_ATimplementation] = ACTIONS(2818),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2818),
    [anon_sym_ATselector] = ACTIONS(2818),
    [anon_sym_ATavailable] = ACTIONS(2818),
    [anon_sym___builtin_available] = ACTIONS(2816),
    [anon_sym_va_arg] = ACTIONS(2816),
    [anon_sym_ATencode] = ACTIONS(2818),
    [anon_sym_BOOL] = ACTIONS(2816),
    [anon_sym_IMP] = ACTIONS(2816),
    [anon_sym_SEL] = ACTIONS(2816),
    [anon_sym_Class] = ACTIONS(2816),
    [anon_sym_id] = ACTIONS(2816),
  },
  [STATE(996)] = {
    [ts_builtin_sym_end] = ACTIONS(2822),
    [sym_identifier] = ACTIONS(2820),
    [aux_sym_preproc_include_token1] = ACTIONS(2820),
    [aux_sym_preproc_include_token2] = ACTIONS(2820),
    [aux_sym_preproc_def_token1] = ACTIONS(2820),
    [aux_sym_preproc_if_token1] = ACTIONS(2820),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2820),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2820),
    [sym_preproc_directive] = ACTIONS(2820),
    [anon_sym_LPAREN2] = ACTIONS(2822),
    [anon_sym_BANG] = ACTIONS(2822),
    [anon_sym_TILDE] = ACTIONS(2822),
    [anon_sym_DASH] = ACTIONS(2820),
    [anon_sym_PLUS] = ACTIONS(2820),
    [anon_sym_STAR] = ACTIONS(2822),
    [anon_sym_CARET] = ACTIONS(2822),
    [anon_sym_AMP] = ACTIONS(2822),
    [anon_sym_SEMI] = ACTIONS(2822),
    [anon_sym___extension__] = ACTIONS(2820),
    [anon_sym_typedef] = ACTIONS(2820),
    [anon_sym_extern] = ACTIONS(2820),
    [anon_sym___attribute__] = ACTIONS(2820),
    [anon_sym___attribute] = ACTIONS(2820),
    [anon_sym_noreturn] = ACTIONS(2820),
    [anon_sym_LBRACK] = ACTIONS(2822),
    [anon_sym___declspec] = ACTIONS(2820),
    [anon_sym___cdecl] = ACTIONS(2820),
    [anon_sym___clrcall] = ACTIONS(2820),
    [anon_sym___stdcall] = ACTIONS(2820),
    [anon_sym___fastcall] = ACTIONS(2820),
    [anon_sym___thiscall] = ACTIONS(2820),
    [anon_sym___vectorcall] = ACTIONS(2820),
    [anon_sym_LBRACE] = ACTIONS(2822),
    [anon_sym_signed] = ACTIONS(2820),
    [anon_sym_unsigned] = ACTIONS(2820),
    [anon_sym_long] = ACTIONS(2820),
    [anon_sym_short] = ACTIONS(2820),
    [anon_sym_static] = ACTIONS(2820),
    [anon_sym_ATautoreleasepool] = ACTIONS(2822),
    [anon_sym_auto] = ACTIONS(2820),
    [anon_sym_register] = ACTIONS(2820),
    [anon_sym_inline] = ACTIONS(2820),
    [anon_sym___inline] = ACTIONS(2820),
    [anon_sym___inline__] = ACTIONS(2820),
    [anon_sym___forceinline] = ACTIONS(2820),
    [anon_sym_thread_local] = ACTIONS(2820),
    [anon_sym___thread] = ACTIONS(2820),
    [anon_sym_CG_EXTERN] = ACTIONS(2820),
    [anon_sym_CG_INLINE] = ACTIONS(2820),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2820),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2820),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2820),
    [anon_sym_IBOutlet] = ACTIONS(2820),
    [anon_sym_IBInspectable] = ACTIONS(2820),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2820),
    [anon_sym_NS_INLINE] = ACTIONS(2820),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2820),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2820),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2820),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2820),
    [anon_sym_const] = ACTIONS(2820),
    [anon_sym_constexpr] = ACTIONS(2820),
    [anon_sym_volatile] = ACTIONS(2820),
    [anon_sym_restrict] = ACTIONS(2820),
    [anon_sym___restrict__] = ACTIONS(2820),
    [anon_sym__Atomic] = ACTIONS(2820),
    [anon_sym__Noreturn] = ACTIONS(2820),
    [anon_sym__Nonnull] = ACTIONS(2820),
    [anon_sym_nullable] = ACTIONS(2820),
    [anon_sym__Complex] = ACTIONS(2820),
    [anon_sym__Nullable] = ACTIONS(2820),
    [anon_sym__Nullable_result] = ACTIONS(2820),
    [anon_sym__Null_unspecified] = ACTIONS(2820),
    [anon_sym___autoreleasing] = ACTIONS(2820),
    [anon_sym___block] = ACTIONS(2820),
    [anon_sym___bridge] = ACTIONS(2820),
    [anon_sym___bridge_retained] = ACTIONS(2820),
    [anon_sym___bridge_transfer] = ACTIONS(2820),
    [anon_sym___complex] = ACTIONS(2820),
    [anon_sym___const] = ACTIONS(2820),
    [anon_sym___imag] = ACTIONS(2820),
    [anon_sym___kindof] = ACTIONS(2820),
    [anon_sym___nonnull] = ACTIONS(2820),
    [anon_sym___nullable] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2820),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2820),
    [anon_sym___real] = ACTIONS(2820),
    [anon_sym___strong] = ACTIONS(2820),
    [anon_sym___unsafe_unretained] = ACTIONS(2820),
    [anon_sym___unused] = ACTIONS(2820),
    [anon_sym___weak] = ACTIONS(2820),
    [anon_sym_alignas] = ACTIONS(2820),
    [anon_sym__Alignas] = ACTIONS(2820),
    [sym_primitive_type] = ACTIONS(2820),
    [anon_sym_enum] = ACTIONS(2820),
    [anon_sym_struct] = ACTIONS(2820),
    [anon_sym_union] = ACTIONS(2820),
    [anon_sym_if] = ACTIONS(2820),
    [anon_sym_switch] = ACTIONS(2820),
    [anon_sym_case] = ACTIONS(2820),
    [anon_sym_default] = ACTIONS(2820),
    [anon_sym_while] = ACTIONS(2820),
    [anon_sym_do] = ACTIONS(2820),
    [anon_sym_for] = ACTIONS(2820),
    [anon_sym_in] = ACTIONS(2820),
    [anon_sym_return] = ACTIONS(2820),
    [anon_sym_break] = ACTIONS(2820),
    [anon_sym_continue] = ACTIONS(2820),
    [anon_sym_goto] = ACTIONS(2820),
    [anon_sym_DASH_DASH] = ACTIONS(2822),
    [anon_sym_PLUS_PLUS] = ACTIONS(2822),
    [anon_sym_sizeof] = ACTIONS(2820),
    [anon_sym___alignof__] = ACTIONS(2820),
    [anon_sym___alignof] = ACTIONS(2820),
    [anon_sym__alignof] = ACTIONS(2820),
    [anon_sym_alignof] = ACTIONS(2820),
    [anon_sym__Alignof] = ACTIONS(2820),
    [anon_sym_offsetof] = ACTIONS(2820),
    [anon_sym__Generic] = ACTIONS(2820),
    [anon_sym_asm] = ACTIONS(2820),
    [anon_sym___asm__] = ACTIONS(2820),
    [anon_sym___asm] = ACTIONS(2820),
    [sym_number_literal] = ACTIONS(2822),
    [anon_sym_L_SQUOTE] = ACTIONS(2822),
    [anon_sym_u_SQUOTE] = ACTIONS(2822),
    [anon_sym_U_SQUOTE] = ACTIONS(2822),
    [anon_sym_u8_SQUOTE] = ACTIONS(2822),
    [anon_sym_SQUOTE] = ACTIONS(2822),
    [anon_sym_AT] = ACTIONS(2820),
    [anon_sym_DQUOTE] = ACTIONS(2822),
    [anon_sym_L_DQUOTE] = ACTIONS(2822),
    [anon_sym_u_DQUOTE] = ACTIONS(2822),
    [anon_sym_U_DQUOTE] = ACTIONS(2822),
    [anon_sym_u8_DQUOTE] = ACTIONS(2822),
    [sym_true] = ACTIONS(2820),
    [sym_false] = ACTIONS(2820),
    [anon_sym_NULL] = ACTIONS(2820),
    [anon_sym_nullptr] = ACTIONS(2820),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2820),
    [anon_sym___typeof] = ACTIONS(2820),
    [anon_sym_typeof] = ACTIONS(2820),
    [anon_sym_ATimport] = ACTIONS(2822),
    [aux_sym_preproc_undef_token1] = ACTIONS(2820),
    [anon_sym_POUND] = ACTIONS(2820),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2820),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2820),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2820),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2820),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2820),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2820),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2820),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_API_AVAILABLE] = ACTIONS(2820),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_API_DEPRECATED] = ACTIONS(2820),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2820),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2820),
    [anon_sym___deprecated_msg] = ACTIONS(2820),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2820),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2820),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2820),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2820),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2820),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2820),
    [anon_sym_ATprotocol] = ACTIONS(2822),
    [anon_sym_ATinterface] = ACTIONS(2822),
    [anon_sym_ATimplementation] = ACTIONS(2822),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2822),
    [anon_sym_ATselector] = ACTIONS(2822),
    [anon_sym_ATavailable] = ACTIONS(2822),
    [anon_sym___builtin_available] = ACTIONS(2820),
    [anon_sym_va_arg] = ACTIONS(2820),
    [anon_sym_ATencode] = ACTIONS(2822),
    [anon_sym_BOOL] = ACTIONS(2820),
    [anon_sym_IMP] = ACTIONS(2820),
    [anon_sym_SEL] = ACTIONS(2820),
    [anon_sym_Class] = ACTIONS(2820),
    [anon_sym_id] = ACTIONS(2820),
  },
  [STATE(997)] = {
    [ts_builtin_sym_end] = ACTIONS(2826),
    [sym_identifier] = ACTIONS(2824),
    [aux_sym_preproc_include_token1] = ACTIONS(2824),
    [aux_sym_preproc_include_token2] = ACTIONS(2824),
    [aux_sym_preproc_def_token1] = ACTIONS(2824),
    [aux_sym_preproc_if_token1] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2824),
    [sym_preproc_directive] = ACTIONS(2824),
    [anon_sym_LPAREN2] = ACTIONS(2826),
    [anon_sym_BANG] = ACTIONS(2826),
    [anon_sym_TILDE] = ACTIONS(2826),
    [anon_sym_DASH] = ACTIONS(2824),
    [anon_sym_PLUS] = ACTIONS(2824),
    [anon_sym_STAR] = ACTIONS(2826),
    [anon_sym_CARET] = ACTIONS(2826),
    [anon_sym_AMP] = ACTIONS(2826),
    [anon_sym_SEMI] = ACTIONS(2826),
    [anon_sym___extension__] = ACTIONS(2824),
    [anon_sym_typedef] = ACTIONS(2824),
    [anon_sym_extern] = ACTIONS(2824),
    [anon_sym___attribute__] = ACTIONS(2824),
    [anon_sym___attribute] = ACTIONS(2824),
    [anon_sym_noreturn] = ACTIONS(2824),
    [anon_sym_LBRACK] = ACTIONS(2826),
    [anon_sym___declspec] = ACTIONS(2824),
    [anon_sym___cdecl] = ACTIONS(2824),
    [anon_sym___clrcall] = ACTIONS(2824),
    [anon_sym___stdcall] = ACTIONS(2824),
    [anon_sym___fastcall] = ACTIONS(2824),
    [anon_sym___thiscall] = ACTIONS(2824),
    [anon_sym___vectorcall] = ACTIONS(2824),
    [anon_sym_LBRACE] = ACTIONS(2826),
    [anon_sym_signed] = ACTIONS(2824),
    [anon_sym_unsigned] = ACTIONS(2824),
    [anon_sym_long] = ACTIONS(2824),
    [anon_sym_short] = ACTIONS(2824),
    [anon_sym_static] = ACTIONS(2824),
    [anon_sym_ATautoreleasepool] = ACTIONS(2826),
    [anon_sym_auto] = ACTIONS(2824),
    [anon_sym_register] = ACTIONS(2824),
    [anon_sym_inline] = ACTIONS(2824),
    [anon_sym___inline] = ACTIONS(2824),
    [anon_sym___inline__] = ACTIONS(2824),
    [anon_sym___forceinline] = ACTIONS(2824),
    [anon_sym_thread_local] = ACTIONS(2824),
    [anon_sym___thread] = ACTIONS(2824),
    [anon_sym_CG_EXTERN] = ACTIONS(2824),
    [anon_sym_CG_INLINE] = ACTIONS(2824),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2824),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2824),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2824),
    [anon_sym_IBOutlet] = ACTIONS(2824),
    [anon_sym_IBInspectable] = ACTIONS(2824),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2824),
    [anon_sym_NS_INLINE] = ACTIONS(2824),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2824),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2824),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2824),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2824),
    [anon_sym_const] = ACTIONS(2824),
    [anon_sym_constexpr] = ACTIONS(2824),
    [anon_sym_volatile] = ACTIONS(2824),
    [anon_sym_restrict] = ACTIONS(2824),
    [anon_sym___restrict__] = ACTIONS(2824),
    [anon_sym__Atomic] = ACTIONS(2824),
    [anon_sym__Noreturn] = ACTIONS(2824),
    [anon_sym__Nonnull] = ACTIONS(2824),
    [anon_sym_nullable] = ACTIONS(2824),
    [anon_sym__Complex] = ACTIONS(2824),
    [anon_sym__Nullable] = ACTIONS(2824),
    [anon_sym__Nullable_result] = ACTIONS(2824),
    [anon_sym__Null_unspecified] = ACTIONS(2824),
    [anon_sym___autoreleasing] = ACTIONS(2824),
    [anon_sym___block] = ACTIONS(2824),
    [anon_sym___bridge] = ACTIONS(2824),
    [anon_sym___bridge_retained] = ACTIONS(2824),
    [anon_sym___bridge_transfer] = ACTIONS(2824),
    [anon_sym___complex] = ACTIONS(2824),
    [anon_sym___const] = ACTIONS(2824),
    [anon_sym___imag] = ACTIONS(2824),
    [anon_sym___kindof] = ACTIONS(2824),
    [anon_sym___nonnull] = ACTIONS(2824),
    [anon_sym___nullable] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2824),
    [anon_sym___real] = ACTIONS(2824),
    [anon_sym___strong] = ACTIONS(2824),
    [anon_sym___unsafe_unretained] = ACTIONS(2824),
    [anon_sym___unused] = ACTIONS(2824),
    [anon_sym___weak] = ACTIONS(2824),
    [anon_sym_alignas] = ACTIONS(2824),
    [anon_sym__Alignas] = ACTIONS(2824),
    [sym_primitive_type] = ACTIONS(2824),
    [anon_sym_enum] = ACTIONS(2824),
    [anon_sym_struct] = ACTIONS(2824),
    [anon_sym_union] = ACTIONS(2824),
    [anon_sym_if] = ACTIONS(2824),
    [anon_sym_switch] = ACTIONS(2824),
    [anon_sym_case] = ACTIONS(2824),
    [anon_sym_default] = ACTIONS(2824),
    [anon_sym_while] = ACTIONS(2824),
    [anon_sym_do] = ACTIONS(2824),
    [anon_sym_for] = ACTIONS(2824),
    [anon_sym_in] = ACTIONS(2824),
    [anon_sym_return] = ACTIONS(2824),
    [anon_sym_break] = ACTIONS(2824),
    [anon_sym_continue] = ACTIONS(2824),
    [anon_sym_goto] = ACTIONS(2824),
    [anon_sym_DASH_DASH] = ACTIONS(2826),
    [anon_sym_PLUS_PLUS] = ACTIONS(2826),
    [anon_sym_sizeof] = ACTIONS(2824),
    [anon_sym___alignof__] = ACTIONS(2824),
    [anon_sym___alignof] = ACTIONS(2824),
    [anon_sym__alignof] = ACTIONS(2824),
    [anon_sym_alignof] = ACTIONS(2824),
    [anon_sym__Alignof] = ACTIONS(2824),
    [anon_sym_offsetof] = ACTIONS(2824),
    [anon_sym__Generic] = ACTIONS(2824),
    [anon_sym_asm] = ACTIONS(2824),
    [anon_sym___asm__] = ACTIONS(2824),
    [anon_sym___asm] = ACTIONS(2824),
    [sym_number_literal] = ACTIONS(2826),
    [anon_sym_L_SQUOTE] = ACTIONS(2826),
    [anon_sym_u_SQUOTE] = ACTIONS(2826),
    [anon_sym_U_SQUOTE] = ACTIONS(2826),
    [anon_sym_u8_SQUOTE] = ACTIONS(2826),
    [anon_sym_SQUOTE] = ACTIONS(2826),
    [anon_sym_AT] = ACTIONS(2824),
    [anon_sym_DQUOTE] = ACTIONS(2826),
    [anon_sym_L_DQUOTE] = ACTIONS(2826),
    [anon_sym_u_DQUOTE] = ACTIONS(2826),
    [anon_sym_U_DQUOTE] = ACTIONS(2826),
    [anon_sym_u8_DQUOTE] = ACTIONS(2826),
    [sym_true] = ACTIONS(2824),
    [sym_false] = ACTIONS(2824),
    [anon_sym_NULL] = ACTIONS(2824),
    [anon_sym_nullptr] = ACTIONS(2824),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2824),
    [anon_sym___typeof] = ACTIONS(2824),
    [anon_sym_typeof] = ACTIONS(2824),
    [anon_sym_ATimport] = ACTIONS(2826),
    [aux_sym_preproc_undef_token1] = ACTIONS(2824),
    [anon_sym_POUND] = ACTIONS(2824),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2824),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2824),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2824),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2824),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2824),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2824),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2824),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_API_AVAILABLE] = ACTIONS(2824),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_API_DEPRECATED] = ACTIONS(2824),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2824),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2824),
    [anon_sym___deprecated_msg] = ACTIONS(2824),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2824),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2824),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2824),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2824),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2824),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2824),
    [anon_sym_ATprotocol] = ACTIONS(2826),
    [anon_sym_ATinterface] = ACTIONS(2826),
    [anon_sym_ATimplementation] = ACTIONS(2826),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2826),
    [anon_sym_ATselector] = ACTIONS(2826),
    [anon_sym_ATavailable] = ACTIONS(2826),
    [anon_sym___builtin_available] = ACTIONS(2824),
    [anon_sym_va_arg] = ACTIONS(2824),
    [anon_sym_ATencode] = ACTIONS(2826),
    [anon_sym_BOOL] = ACTIONS(2824),
    [anon_sym_IMP] = ACTIONS(2824),
    [anon_sym_SEL] = ACTIONS(2824),
    [anon_sym_Class] = ACTIONS(2824),
    [anon_sym_id] = ACTIONS(2824),
  },
  [STATE(998)] = {
    [ts_builtin_sym_end] = ACTIONS(2512),
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(999)] = {
    [ts_builtin_sym_end] = ACTIONS(2512),
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(1000)] = {
    [ts_builtin_sym_end] = ACTIONS(2512),
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_include_token1] = ACTIONS(2510),
    [aux_sym_preproc_include_token2] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_case] = ACTIONS(2510),
    [anon_sym_default] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_ATimport] = ACTIONS(2512),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_POUND] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATprotocol] = ACTIONS(2512),
    [anon_sym_ATinterface] = ACTIONS(2512),
    [anon_sym_ATimplementation] = ACTIONS(2512),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(1001)] = {
    [ts_builtin_sym_end] = ACTIONS(2834),
    [sym_identifier] = ACTIONS(2832),
    [aux_sym_preproc_include_token1] = ACTIONS(2832),
    [aux_sym_preproc_include_token2] = ACTIONS(2832),
    [aux_sym_preproc_def_token1] = ACTIONS(2832),
    [aux_sym_preproc_if_token1] = ACTIONS(2832),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2832),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2832),
    [sym_preproc_directive] = ACTIONS(2832),
    [anon_sym_LPAREN2] = ACTIONS(2834),
    [anon_sym_BANG] = ACTIONS(2834),
    [anon_sym_TILDE] = ACTIONS(2834),
    [anon_sym_DASH] = ACTIONS(2832),
    [anon_sym_PLUS] = ACTIONS(2832),
    [anon_sym_STAR] = ACTIONS(2834),
    [anon_sym_CARET] = ACTIONS(2834),
    [anon_sym_AMP] = ACTIONS(2834),
    [anon_sym_SEMI] = ACTIONS(2834),
    [anon_sym___extension__] = ACTIONS(2832),
    [anon_sym_typedef] = ACTIONS(2832),
    [anon_sym_extern] = ACTIONS(2832),
    [anon_sym___attribute__] = ACTIONS(2832),
    [anon_sym___attribute] = ACTIONS(2832),
    [anon_sym_noreturn] = ACTIONS(2832),
    [anon_sym_LBRACK] = ACTIONS(2834),
    [anon_sym___declspec] = ACTIONS(2832),
    [anon_sym___cdecl] = ACTIONS(2832),
    [anon_sym___clrcall] = ACTIONS(2832),
    [anon_sym___stdcall] = ACTIONS(2832),
    [anon_sym___fastcall] = ACTIONS(2832),
    [anon_sym___thiscall] = ACTIONS(2832),
    [anon_sym___vectorcall] = ACTIONS(2832),
    [anon_sym_LBRACE] = ACTIONS(2834),
    [anon_sym_signed] = ACTIONS(2832),
    [anon_sym_unsigned] = ACTIONS(2832),
    [anon_sym_long] = ACTIONS(2832),
    [anon_sym_short] = ACTIONS(2832),
    [anon_sym_static] = ACTIONS(2832),
    [anon_sym_ATautoreleasepool] = ACTIONS(2834),
    [anon_sym_auto] = ACTIONS(2832),
    [anon_sym_register] = ACTIONS(2832),
    [anon_sym_inline] = ACTIONS(2832),
    [anon_sym___inline] = ACTIONS(2832),
    [anon_sym___inline__] = ACTIONS(2832),
    [anon_sym___forceinline] = ACTIONS(2832),
    [anon_sym_thread_local] = ACTIONS(2832),
    [anon_sym___thread] = ACTIONS(2832),
    [anon_sym_CG_EXTERN] = ACTIONS(2832),
    [anon_sym_CG_INLINE] = ACTIONS(2832),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2832),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2832),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2832),
    [anon_sym_IBOutlet] = ACTIONS(2832),
    [anon_sym_IBInspectable] = ACTIONS(2832),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2832),
    [anon_sym_NS_INLINE] = ACTIONS(2832),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2832),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2832),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2832),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2832),
    [anon_sym_const] = ACTIONS(2832),
    [anon_sym_constexpr] = ACTIONS(2832),
    [anon_sym_volatile] = ACTIONS(2832),
    [anon_sym_restrict] = ACTIONS(2832),
    [anon_sym___restrict__] = ACTIONS(2832),
    [anon_sym__Atomic] = ACTIONS(2832),
    [anon_sym__Noreturn] = ACTIONS(2832),
    [anon_sym__Nonnull] = ACTIONS(2832),
    [anon_sym_nullable] = ACTIONS(2832),
    [anon_sym__Complex] = ACTIONS(2832),
    [anon_sym__Nullable] = ACTIONS(2832),
    [anon_sym__Nullable_result] = ACTIONS(2832),
    [anon_sym__Null_unspecified] = ACTIONS(2832),
    [anon_sym___autoreleasing] = ACTIONS(2832),
    [anon_sym___block] = ACTIONS(2832),
    [anon_sym___bridge] = ACTIONS(2832),
    [anon_sym___bridge_retained] = ACTIONS(2832),
    [anon_sym___bridge_transfer] = ACTIONS(2832),
    [anon_sym___complex] = ACTIONS(2832),
    [anon_sym___const] = ACTIONS(2832),
    [anon_sym___imag] = ACTIONS(2832),
    [anon_sym___kindof] = ACTIONS(2832),
    [anon_sym___nonnull] = ACTIONS(2832),
    [anon_sym___nullable] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2832),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2832),
    [anon_sym___real] = ACTIONS(2832),
    [anon_sym___strong] = ACTIONS(2832),
    [anon_sym___unsafe_unretained] = ACTIONS(2832),
    [anon_sym___unused] = ACTIONS(2832),
    [anon_sym___weak] = ACTIONS(2832),
    [anon_sym_alignas] = ACTIONS(2832),
    [anon_sym__Alignas] = ACTIONS(2832),
    [sym_primitive_type] = ACTIONS(2832),
    [anon_sym_enum] = ACTIONS(2832),
    [anon_sym_struct] = ACTIONS(2832),
    [anon_sym_union] = ACTIONS(2832),
    [anon_sym_if] = ACTIONS(2832),
    [anon_sym_switch] = ACTIONS(2832),
    [anon_sym_case] = ACTIONS(2832),
    [anon_sym_default] = ACTIONS(2832),
    [anon_sym_while] = ACTIONS(2832),
    [anon_sym_do] = ACTIONS(2832),
    [anon_sym_for] = ACTIONS(2832),
    [anon_sym_in] = ACTIONS(2832),
    [anon_sym_return] = ACTIONS(2832),
    [anon_sym_break] = ACTIONS(2832),
    [anon_sym_continue] = ACTIONS(2832),
    [anon_sym_goto] = ACTIONS(2832),
    [anon_sym_DASH_DASH] = ACTIONS(2834),
    [anon_sym_PLUS_PLUS] = ACTIONS(2834),
    [anon_sym_sizeof] = ACTIONS(2832),
    [anon_sym___alignof__] = ACTIONS(2832),
    [anon_sym___alignof] = ACTIONS(2832),
    [anon_sym__alignof] = ACTIONS(2832),
    [anon_sym_alignof] = ACTIONS(2832),
    [anon_sym__Alignof] = ACTIONS(2832),
    [anon_sym_offsetof] = ACTIONS(2832),
    [anon_sym__Generic] = ACTIONS(2832),
    [anon_sym_asm] = ACTIONS(2832),
    [anon_sym___asm__] = ACTIONS(2832),
    [anon_sym___asm] = ACTIONS(2832),
    [sym_number_literal] = ACTIONS(2834),
    [anon_sym_L_SQUOTE] = ACTIONS(2834),
    [anon_sym_u_SQUOTE] = ACTIONS(2834),
    [anon_sym_U_SQUOTE] = ACTIONS(2834),
    [anon_sym_u8_SQUOTE] = ACTIONS(2834),
    [anon_sym_SQUOTE] = ACTIONS(2834),
    [anon_sym_AT] = ACTIONS(2832),
    [anon_sym_DQUOTE] = ACTIONS(2834),
    [anon_sym_L_DQUOTE] = ACTIONS(2834),
    [anon_sym_u_DQUOTE] = ACTIONS(2834),
    [anon_sym_U_DQUOTE] = ACTIONS(2834),
    [anon_sym_u8_DQUOTE] = ACTIONS(2834),
    [sym_true] = ACTIONS(2832),
    [sym_false] = ACTIONS(2832),
    [anon_sym_NULL] = ACTIONS(2832),
    [anon_sym_nullptr] = ACTIONS(2832),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2832),
    [anon_sym___typeof] = ACTIONS(2832),
    [anon_sym_typeof] = ACTIONS(2832),
    [anon_sym_ATimport] = ACTIONS(2834),
    [aux_sym_preproc_undef_token1] = ACTIONS(2832),
    [anon_sym_POUND] = ACTIONS(2832),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2832),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2832),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2832),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2832),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2832),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2832),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2832),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_API_AVAILABLE] = ACTIONS(2832),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_API_DEPRECATED] = ACTIONS(2832),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2832),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2832),
    [anon_sym___deprecated_msg] = ACTIONS(2832),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2832),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2832),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2832),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2832),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2832),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2832),
    [anon_sym_ATprotocol] = ACTIONS(2834),
    [anon_sym_ATinterface] = ACTIONS(2834),
    [anon_sym_ATimplementation] = ACTIONS(2834),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2834),
    [anon_sym_ATselector] = ACTIONS(2834),
    [anon_sym_ATavailable] = ACTIONS(2834),
    [anon_sym___builtin_available] = ACTIONS(2832),
    [anon_sym_va_arg] = ACTIONS(2832),
    [anon_sym_ATencode] = ACTIONS(2834),
    [anon_sym_BOOL] = ACTIONS(2832),
    [anon_sym_IMP] = ACTIONS(2832),
    [anon_sym_SEL] = ACTIONS(2832),
    [anon_sym_Class] = ACTIONS(2832),
    [anon_sym_id] = ACTIONS(2832),
  },
  [STATE(1002)] = {
    [ts_builtin_sym_end] = ACTIONS(2838),
    [sym_identifier] = ACTIONS(2836),
    [aux_sym_preproc_include_token1] = ACTIONS(2836),
    [aux_sym_preproc_include_token2] = ACTIONS(2836),
    [aux_sym_preproc_def_token1] = ACTIONS(2836),
    [aux_sym_preproc_if_token1] = ACTIONS(2836),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2836),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2836),
    [sym_preproc_directive] = ACTIONS(2836),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_BANG] = ACTIONS(2838),
    [anon_sym_TILDE] = ACTIONS(2838),
    [anon_sym_DASH] = ACTIONS(2836),
    [anon_sym_PLUS] = ACTIONS(2836),
    [anon_sym_STAR] = ACTIONS(2838),
    [anon_sym_CARET] = ACTIONS(2838),
    [anon_sym_AMP] = ACTIONS(2838),
    [anon_sym_SEMI] = ACTIONS(2838),
    [anon_sym___extension__] = ACTIONS(2836),
    [anon_sym_typedef] = ACTIONS(2836),
    [anon_sym_extern] = ACTIONS(2836),
    [anon_sym___attribute__] = ACTIONS(2836),
    [anon_sym___attribute] = ACTIONS(2836),
    [anon_sym_noreturn] = ACTIONS(2836),
    [anon_sym_LBRACK] = ACTIONS(2838),
    [anon_sym___declspec] = ACTIONS(2836),
    [anon_sym___cdecl] = ACTIONS(2836),
    [anon_sym___clrcall] = ACTIONS(2836),
    [anon_sym___stdcall] = ACTIONS(2836),
    [anon_sym___fastcall] = ACTIONS(2836),
    [anon_sym___thiscall] = ACTIONS(2836),
    [anon_sym___vectorcall] = ACTIONS(2836),
    [anon_sym_LBRACE] = ACTIONS(2838),
    [anon_sym_signed] = ACTIONS(2836),
    [anon_sym_unsigned] = ACTIONS(2836),
    [anon_sym_long] = ACTIONS(2836),
    [anon_sym_short] = ACTIONS(2836),
    [anon_sym_static] = ACTIONS(2836),
    [anon_sym_ATautoreleasepool] = ACTIONS(2838),
    [anon_sym_auto] = ACTIONS(2836),
    [anon_sym_register] = ACTIONS(2836),
    [anon_sym_inline] = ACTIONS(2836),
    [anon_sym___inline] = ACTIONS(2836),
    [anon_sym___inline__] = ACTIONS(2836),
    [anon_sym___forceinline] = ACTIONS(2836),
    [anon_sym_thread_local] = ACTIONS(2836),
    [anon_sym___thread] = ACTIONS(2836),
    [anon_sym_CG_EXTERN] = ACTIONS(2836),
    [anon_sym_CG_INLINE] = ACTIONS(2836),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2836),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2836),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2836),
    [anon_sym_IBOutlet] = ACTIONS(2836),
    [anon_sym_IBInspectable] = ACTIONS(2836),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2836),
    [anon_sym_NS_INLINE] = ACTIONS(2836),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2836),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2836),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2836),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2836),
    [anon_sym_const] = ACTIONS(2836),
    [anon_sym_constexpr] = ACTIONS(2836),
    [anon_sym_volatile] = ACTIONS(2836),
    [anon_sym_restrict] = ACTIONS(2836),
    [anon_sym___restrict__] = ACTIONS(2836),
    [anon_sym__Atomic] = ACTIONS(2836),
    [anon_sym__Noreturn] = ACTIONS(2836),
    [anon_sym__Nonnull] = ACTIONS(2836),
    [anon_sym_nullable] = ACTIONS(2836),
    [anon_sym__Complex] = ACTIONS(2836),
    [anon_sym__Nullable] = ACTIONS(2836),
    [anon_sym__Nullable_result] = ACTIONS(2836),
    [anon_sym__Null_unspecified] = ACTIONS(2836),
    [anon_sym___autoreleasing] = ACTIONS(2836),
    [anon_sym___block] = ACTIONS(2836),
    [anon_sym___bridge] = ACTIONS(2836),
    [anon_sym___bridge_retained] = ACTIONS(2836),
    [anon_sym___bridge_transfer] = ACTIONS(2836),
    [anon_sym___complex] = ACTIONS(2836),
    [anon_sym___const] = ACTIONS(2836),
    [anon_sym___imag] = ACTIONS(2836),
    [anon_sym___kindof] = ACTIONS(2836),
    [anon_sym___nonnull] = ACTIONS(2836),
    [anon_sym___nullable] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2836),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2836),
    [anon_sym___real] = ACTIONS(2836),
    [anon_sym___strong] = ACTIONS(2836),
    [anon_sym___unsafe_unretained] = ACTIONS(2836),
    [anon_sym___unused] = ACTIONS(2836),
    [anon_sym___weak] = ACTIONS(2836),
    [anon_sym_alignas] = ACTIONS(2836),
    [anon_sym__Alignas] = ACTIONS(2836),
    [sym_primitive_type] = ACTIONS(2836),
    [anon_sym_enum] = ACTIONS(2836),
    [anon_sym_struct] = ACTIONS(2836),
    [anon_sym_union] = ACTIONS(2836),
    [anon_sym_if] = ACTIONS(2836),
    [anon_sym_switch] = ACTIONS(2836),
    [anon_sym_case] = ACTIONS(2836),
    [anon_sym_default] = ACTIONS(2836),
    [anon_sym_while] = ACTIONS(2836),
    [anon_sym_do] = ACTIONS(2836),
    [anon_sym_for] = ACTIONS(2836),
    [anon_sym_in] = ACTIONS(2836),
    [anon_sym_return] = ACTIONS(2836),
    [anon_sym_break] = ACTIONS(2836),
    [anon_sym_continue] = ACTIONS(2836),
    [anon_sym_goto] = ACTIONS(2836),
    [anon_sym_DASH_DASH] = ACTIONS(2838),
    [anon_sym_PLUS_PLUS] = ACTIONS(2838),
    [anon_sym_sizeof] = ACTIONS(2836),
    [anon_sym___alignof__] = ACTIONS(2836),
    [anon_sym___alignof] = ACTIONS(2836),
    [anon_sym__alignof] = ACTIONS(2836),
    [anon_sym_alignof] = ACTIONS(2836),
    [anon_sym__Alignof] = ACTIONS(2836),
    [anon_sym_offsetof] = ACTIONS(2836),
    [anon_sym__Generic] = ACTIONS(2836),
    [anon_sym_asm] = ACTIONS(2836),
    [anon_sym___asm__] = ACTIONS(2836),
    [anon_sym___asm] = ACTIONS(2836),
    [sym_number_literal] = ACTIONS(2838),
    [anon_sym_L_SQUOTE] = ACTIONS(2838),
    [anon_sym_u_SQUOTE] = ACTIONS(2838),
    [anon_sym_U_SQUOTE] = ACTIONS(2838),
    [anon_sym_u8_SQUOTE] = ACTIONS(2838),
    [anon_sym_SQUOTE] = ACTIONS(2838),
    [anon_sym_AT] = ACTIONS(2836),
    [anon_sym_DQUOTE] = ACTIONS(2838),
    [anon_sym_L_DQUOTE] = ACTIONS(2838),
    [anon_sym_u_DQUOTE] = ACTIONS(2838),
    [anon_sym_U_DQUOTE] = ACTIONS(2838),
    [anon_sym_u8_DQUOTE] = ACTIONS(2838),
    [sym_true] = ACTIONS(2836),
    [sym_false] = ACTIONS(2836),
    [anon_sym_NULL] = ACTIONS(2836),
    [anon_sym_nullptr] = ACTIONS(2836),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2836),
    [anon_sym___typeof] = ACTIONS(2836),
    [anon_sym_typeof] = ACTIONS(2836),
    [anon_sym_ATimport] = ACTIONS(2838),
    [aux_sym_preproc_undef_token1] = ACTIONS(2836),
    [anon_sym_POUND] = ACTIONS(2836),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2836),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2836),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2836),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2836),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2836),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2836),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2836),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_API_AVAILABLE] = ACTIONS(2836),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_API_DEPRECATED] = ACTIONS(2836),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2836),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2836),
    [anon_sym___deprecated_msg] = ACTIONS(2836),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2836),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2836),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2836),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2836),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2836),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2836),
    [anon_sym_ATprotocol] = ACTIONS(2838),
    [anon_sym_ATinterface] = ACTIONS(2838),
    [anon_sym_ATimplementation] = ACTIONS(2838),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2838),
    [anon_sym_ATselector] = ACTIONS(2838),
    [anon_sym_ATavailable] = ACTIONS(2838),
    [anon_sym___builtin_available] = ACTIONS(2836),
    [anon_sym_va_arg] = ACTIONS(2836),
    [anon_sym_ATencode] = ACTIONS(2838),
    [anon_sym_BOOL] = ACTIONS(2836),
    [anon_sym_IMP] = ACTIONS(2836),
    [anon_sym_SEL] = ACTIONS(2836),
    [anon_sym_Class] = ACTIONS(2836),
    [anon_sym_id] = ACTIONS(2836),
  },
  [STATE(1003)] = {
    [ts_builtin_sym_end] = ACTIONS(2842),
    [sym_identifier] = ACTIONS(2840),
    [aux_sym_preproc_include_token1] = ACTIONS(2840),
    [aux_sym_preproc_include_token2] = ACTIONS(2840),
    [aux_sym_preproc_def_token1] = ACTIONS(2840),
    [aux_sym_preproc_if_token1] = ACTIONS(2840),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2840),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2840),
    [sym_preproc_directive] = ACTIONS(2840),
    [anon_sym_LPAREN2] = ACTIONS(2842),
    [anon_sym_BANG] = ACTIONS(2842),
    [anon_sym_TILDE] = ACTIONS(2842),
    [anon_sym_DASH] = ACTIONS(2840),
    [anon_sym_PLUS] = ACTIONS(2840),
    [anon_sym_STAR] = ACTIONS(2842),
    [anon_sym_CARET] = ACTIONS(2842),
    [anon_sym_AMP] = ACTIONS(2842),
    [anon_sym_SEMI] = ACTIONS(2842),
    [anon_sym___extension__] = ACTIONS(2840),
    [anon_sym_typedef] = ACTIONS(2840),
    [anon_sym_extern] = ACTIONS(2840),
    [anon_sym___attribute__] = ACTIONS(2840),
    [anon_sym___attribute] = ACTIONS(2840),
    [anon_sym_noreturn] = ACTIONS(2840),
    [anon_sym_LBRACK] = ACTIONS(2842),
    [anon_sym___declspec] = ACTIONS(2840),
    [anon_sym___cdecl] = ACTIONS(2840),
    [anon_sym___clrcall] = ACTIONS(2840),
    [anon_sym___stdcall] = ACTIONS(2840),
    [anon_sym___fastcall] = ACTIONS(2840),
    [anon_sym___thiscall] = ACTIONS(2840),
    [anon_sym___vectorcall] = ACTIONS(2840),
    [anon_sym_LBRACE] = ACTIONS(2842),
    [anon_sym_signed] = ACTIONS(2840),
    [anon_sym_unsigned] = ACTIONS(2840),
    [anon_sym_long] = ACTIONS(2840),
    [anon_sym_short] = ACTIONS(2840),
    [anon_sym_static] = ACTIONS(2840),
    [anon_sym_ATautoreleasepool] = ACTIONS(2842),
    [anon_sym_auto] = ACTIONS(2840),
    [anon_sym_register] = ACTIONS(2840),
    [anon_sym_inline] = ACTIONS(2840),
    [anon_sym___inline] = ACTIONS(2840),
    [anon_sym___inline__] = ACTIONS(2840),
    [anon_sym___forceinline] = ACTIONS(2840),
    [anon_sym_thread_local] = ACTIONS(2840),
    [anon_sym___thread] = ACTIONS(2840),
    [anon_sym_CG_EXTERN] = ACTIONS(2840),
    [anon_sym_CG_INLINE] = ACTIONS(2840),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2840),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2840),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2840),
    [anon_sym_IBOutlet] = ACTIONS(2840),
    [anon_sym_IBInspectable] = ACTIONS(2840),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2840),
    [anon_sym_NS_INLINE] = ACTIONS(2840),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2840),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2840),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2840),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2840),
    [anon_sym_const] = ACTIONS(2840),
    [anon_sym_constexpr] = ACTIONS(2840),
    [anon_sym_volatile] = ACTIONS(2840),
    [anon_sym_restrict] = ACTIONS(2840),
    [anon_sym___restrict__] = ACTIONS(2840),
    [anon_sym__Atomic] = ACTIONS(2840),
    [anon_sym__Noreturn] = ACTIONS(2840),
    [anon_sym__Nonnull] = ACTIONS(2840),
    [anon_sym_nullable] = ACTIONS(2840),
    [anon_sym__Complex] = ACTIONS(2840),
    [anon_sym__Nullable] = ACTIONS(2840),
    [anon_sym__Nullable_result] = ACTIONS(2840),
    [anon_sym__Null_unspecified] = ACTIONS(2840),
    [anon_sym___autoreleasing] = ACTIONS(2840),
    [anon_sym___block] = ACTIONS(2840),
    [anon_sym___bridge] = ACTIONS(2840),
    [anon_sym___bridge_retained] = ACTIONS(2840),
    [anon_sym___bridge_transfer] = ACTIONS(2840),
    [anon_sym___complex] = ACTIONS(2840),
    [anon_sym___const] = ACTIONS(2840),
    [anon_sym___imag] = ACTIONS(2840),
    [anon_sym___kindof] = ACTIONS(2840),
    [anon_sym___nonnull] = ACTIONS(2840),
    [anon_sym___nullable] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2840),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2840),
    [anon_sym___real] = ACTIONS(2840),
    [anon_sym___strong] = ACTIONS(2840),
    [anon_sym___unsafe_unretained] = ACTIONS(2840),
    [anon_sym___unused] = ACTIONS(2840),
    [anon_sym___weak] = ACTIONS(2840),
    [anon_sym_alignas] = ACTIONS(2840),
    [anon_sym__Alignas] = ACTIONS(2840),
    [sym_primitive_type] = ACTIONS(2840),
    [anon_sym_enum] = ACTIONS(2840),
    [anon_sym_struct] = ACTIONS(2840),
    [anon_sym_union] = ACTIONS(2840),
    [anon_sym_if] = ACTIONS(2840),
    [anon_sym_switch] = ACTIONS(2840),
    [anon_sym_case] = ACTIONS(2840),
    [anon_sym_default] = ACTIONS(2840),
    [anon_sym_while] = ACTIONS(2840),
    [anon_sym_do] = ACTIONS(2840),
    [anon_sym_for] = ACTIONS(2840),
    [anon_sym_in] = ACTIONS(2840),
    [anon_sym_return] = ACTIONS(2840),
    [anon_sym_break] = ACTIONS(2840),
    [anon_sym_continue] = ACTIONS(2840),
    [anon_sym_goto] = ACTIONS(2840),
    [anon_sym_DASH_DASH] = ACTIONS(2842),
    [anon_sym_PLUS_PLUS] = ACTIONS(2842),
    [anon_sym_sizeof] = ACTIONS(2840),
    [anon_sym___alignof__] = ACTIONS(2840),
    [anon_sym___alignof] = ACTIONS(2840),
    [anon_sym__alignof] = ACTIONS(2840),
    [anon_sym_alignof] = ACTIONS(2840),
    [anon_sym__Alignof] = ACTIONS(2840),
    [anon_sym_offsetof] = ACTIONS(2840),
    [anon_sym__Generic] = ACTIONS(2840),
    [anon_sym_asm] = ACTIONS(2840),
    [anon_sym___asm__] = ACTIONS(2840),
    [anon_sym___asm] = ACTIONS(2840),
    [sym_number_literal] = ACTIONS(2842),
    [anon_sym_L_SQUOTE] = ACTIONS(2842),
    [anon_sym_u_SQUOTE] = ACTIONS(2842),
    [anon_sym_U_SQUOTE] = ACTIONS(2842),
    [anon_sym_u8_SQUOTE] = ACTIONS(2842),
    [anon_sym_SQUOTE] = ACTIONS(2842),
    [anon_sym_AT] = ACTIONS(2840),
    [anon_sym_DQUOTE] = ACTIONS(2842),
    [anon_sym_L_DQUOTE] = ACTIONS(2842),
    [anon_sym_u_DQUOTE] = ACTIONS(2842),
    [anon_sym_U_DQUOTE] = ACTIONS(2842),
    [anon_sym_u8_DQUOTE] = ACTIONS(2842),
    [sym_true] = ACTIONS(2840),
    [sym_false] = ACTIONS(2840),
    [anon_sym_NULL] = ACTIONS(2840),
    [anon_sym_nullptr] = ACTIONS(2840),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2840),
    [anon_sym___typeof] = ACTIONS(2840),
    [anon_sym_typeof] = ACTIONS(2840),
    [anon_sym_ATimport] = ACTIONS(2842),
    [aux_sym_preproc_undef_token1] = ACTIONS(2840),
    [anon_sym_POUND] = ACTIONS(2840),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2840),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2840),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2840),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2840),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2840),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2840),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2840),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_API_AVAILABLE] = ACTIONS(2840),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_API_DEPRECATED] = ACTIONS(2840),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2840),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2840),
    [anon_sym___deprecated_msg] = ACTIONS(2840),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2840),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2840),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2840),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2840),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2840),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2840),
    [anon_sym_ATprotocol] = ACTIONS(2842),
    [anon_sym_ATinterface] = ACTIONS(2842),
    [anon_sym_ATimplementation] = ACTIONS(2842),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2842),
    [anon_sym_ATselector] = ACTIONS(2842),
    [anon_sym_ATavailable] = ACTIONS(2842),
    [anon_sym___builtin_available] = ACTIONS(2840),
    [anon_sym_va_arg] = ACTIONS(2840),
    [anon_sym_ATencode] = ACTIONS(2842),
    [anon_sym_BOOL] = ACTIONS(2840),
    [anon_sym_IMP] = ACTIONS(2840),
    [anon_sym_SEL] = ACTIONS(2840),
    [anon_sym_Class] = ACTIONS(2840),
    [anon_sym_id] = ACTIONS(2840),
  },
  [STATE(1004)] = {
    [ts_builtin_sym_end] = ACTIONS(2640),
    [sym_identifier] = ACTIONS(2638),
    [aux_sym_preproc_include_token1] = ACTIONS(2638),
    [aux_sym_preproc_include_token2] = ACTIONS(2638),
    [aux_sym_preproc_def_token1] = ACTIONS(2638),
    [aux_sym_preproc_if_token1] = ACTIONS(2638),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2638),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2638),
    [sym_preproc_directive] = ACTIONS(2638),
    [anon_sym_LPAREN2] = ACTIONS(2640),
    [anon_sym_BANG] = ACTIONS(2640),
    [anon_sym_TILDE] = ACTIONS(2640),
    [anon_sym_DASH] = ACTIONS(2638),
    [anon_sym_PLUS] = ACTIONS(2638),
    [anon_sym_STAR] = ACTIONS(2640),
    [anon_sym_CARET] = ACTIONS(2640),
    [anon_sym_AMP] = ACTIONS(2640),
    [anon_sym_SEMI] = ACTIONS(2640),
    [anon_sym___extension__] = ACTIONS(2638),
    [anon_sym_typedef] = ACTIONS(2638),
    [anon_sym_extern] = ACTIONS(2638),
    [anon_sym___attribute__] = ACTIONS(2638),
    [anon_sym___attribute] = ACTIONS(2638),
    [anon_sym_noreturn] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2640),
    [anon_sym___declspec] = ACTIONS(2638),
    [anon_sym___cdecl] = ACTIONS(2638),
    [anon_sym___clrcall] = ACTIONS(2638),
    [anon_sym___stdcall] = ACTIONS(2638),
    [anon_sym___fastcall] = ACTIONS(2638),
    [anon_sym___thiscall] = ACTIONS(2638),
    [anon_sym___vectorcall] = ACTIONS(2638),
    [anon_sym_LBRACE] = ACTIONS(2640),
    [anon_sym_signed] = ACTIONS(2638),
    [anon_sym_unsigned] = ACTIONS(2638),
    [anon_sym_long] = ACTIONS(2638),
    [anon_sym_short] = ACTIONS(2638),
    [anon_sym_static] = ACTIONS(2638),
    [anon_sym_ATautoreleasepool] = ACTIONS(2640),
    [anon_sym_auto] = ACTIONS(2638),
    [anon_sym_register] = ACTIONS(2638),
    [anon_sym_inline] = ACTIONS(2638),
    [anon_sym___inline] = ACTIONS(2638),
    [anon_sym___inline__] = ACTIONS(2638),
    [anon_sym___forceinline] = ACTIONS(2638),
    [anon_sym_thread_local] = ACTIONS(2638),
    [anon_sym___thread] = ACTIONS(2638),
    [anon_sym_CG_EXTERN] = ACTIONS(2638),
    [anon_sym_CG_INLINE] = ACTIONS(2638),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2638),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2638),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2638),
    [anon_sym_IBOutlet] = ACTIONS(2638),
    [anon_sym_IBInspectable] = ACTIONS(2638),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2638),
    [anon_sym_NS_INLINE] = ACTIONS(2638),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2638),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2638),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2638),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2638),
    [anon_sym_const] = ACTIONS(2638),
    [anon_sym_constexpr] = ACTIONS(2638),
    [anon_sym_volatile] = ACTIONS(2638),
    [anon_sym_restrict] = ACTIONS(2638),
    [anon_sym___restrict__] = ACTIONS(2638),
    [anon_sym__Atomic] = ACTIONS(2638),
    [anon_sym__Noreturn] = ACTIONS(2638),
    [anon_sym__Nonnull] = ACTIONS(2638),
    [anon_sym_nullable] = ACTIONS(2638),
    [anon_sym__Complex] = ACTIONS(2638),
    [anon_sym__Nullable] = ACTIONS(2638),
    [anon_sym__Nullable_result] = ACTIONS(2638),
    [anon_sym__Null_unspecified] = ACTIONS(2638),
    [anon_sym___autoreleasing] = ACTIONS(2638),
    [anon_sym___block] = ACTIONS(2638),
    [anon_sym___bridge] = ACTIONS(2638),
    [anon_sym___bridge_retained] = ACTIONS(2638),
    [anon_sym___bridge_transfer] = ACTIONS(2638),
    [anon_sym___complex] = ACTIONS(2638),
    [anon_sym___const] = ACTIONS(2638),
    [anon_sym___imag] = ACTIONS(2638),
    [anon_sym___kindof] = ACTIONS(2638),
    [anon_sym___nonnull] = ACTIONS(2638),
    [anon_sym___nullable] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2638),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2638),
    [anon_sym___real] = ACTIONS(2638),
    [anon_sym___strong] = ACTIONS(2638),
    [anon_sym___unsafe_unretained] = ACTIONS(2638),
    [anon_sym___unused] = ACTIONS(2638),
    [anon_sym___weak] = ACTIONS(2638),
    [anon_sym_alignas] = ACTIONS(2638),
    [anon_sym__Alignas] = ACTIONS(2638),
    [sym_primitive_type] = ACTIONS(2638),
    [anon_sym_enum] = ACTIONS(2638),
    [anon_sym_struct] = ACTIONS(2638),
    [anon_sym_union] = ACTIONS(2638),
    [anon_sym_if] = ACTIONS(2638),
    [anon_sym_switch] = ACTIONS(2638),
    [anon_sym_case] = ACTIONS(2638),
    [anon_sym_default] = ACTIONS(2638),
    [anon_sym_while] = ACTIONS(2638),
    [anon_sym_do] = ACTIONS(2638),
    [anon_sym_for] = ACTIONS(2638),
    [anon_sym_in] = ACTIONS(2638),
    [anon_sym_return] = ACTIONS(2638),
    [anon_sym_break] = ACTIONS(2638),
    [anon_sym_continue] = ACTIONS(2638),
    [anon_sym_goto] = ACTIONS(2638),
    [anon_sym_DASH_DASH] = ACTIONS(2640),
    [anon_sym_PLUS_PLUS] = ACTIONS(2640),
    [anon_sym_sizeof] = ACTIONS(2638),
    [anon_sym___alignof__] = ACTIONS(2638),
    [anon_sym___alignof] = ACTIONS(2638),
    [anon_sym__alignof] = ACTIONS(2638),
    [anon_sym_alignof] = ACTIONS(2638),
    [anon_sym__Alignof] = ACTIONS(2638),
    [anon_sym_offsetof] = ACTIONS(2638),
    [anon_sym__Generic] = ACTIONS(2638),
    [anon_sym_asm] = ACTIONS(2638),
    [anon_sym___asm__] = ACTIONS(2638),
    [anon_sym___asm] = ACTIONS(2638),
    [sym_number_literal] = ACTIONS(2640),
    [anon_sym_L_SQUOTE] = ACTIONS(2640),
    [anon_sym_u_SQUOTE] = ACTIONS(2640),
    [anon_sym_U_SQUOTE] = ACTIONS(2640),
    [anon_sym_u8_SQUOTE] = ACTIONS(2640),
    [anon_sym_SQUOTE] = ACTIONS(2640),
    [anon_sym_AT] = ACTIONS(2638),
    [anon_sym_DQUOTE] = ACTIONS(2640),
    [anon_sym_L_DQUOTE] = ACTIONS(2640),
    [anon_sym_u_DQUOTE] = ACTIONS(2640),
    [anon_sym_U_DQUOTE] = ACTIONS(2640),
    [anon_sym_u8_DQUOTE] = ACTIONS(2640),
    [sym_true] = ACTIONS(2638),
    [sym_false] = ACTIONS(2638),
    [anon_sym_NULL] = ACTIONS(2638),
    [anon_sym_nullptr] = ACTIONS(2638),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2638),
    [anon_sym___typeof] = ACTIONS(2638),
    [anon_sym_typeof] = ACTIONS(2638),
    [anon_sym_ATimport] = ACTIONS(2640),
    [aux_sym_preproc_undef_token1] = ACTIONS(2638),
    [anon_sym_POUND] = ACTIONS(2638),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2638),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2638),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2638),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2638),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2638),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2638),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2638),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_API_AVAILABLE] = ACTIONS(2638),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_API_DEPRECATED] = ACTIONS(2638),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2638),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2638),
    [anon_sym___deprecated_msg] = ACTIONS(2638),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2638),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2638),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2638),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2638),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2638),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2638),
    [anon_sym_ATprotocol] = ACTIONS(2640),
    [anon_sym_ATinterface] = ACTIONS(2640),
    [anon_sym_ATimplementation] = ACTIONS(2640),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2640),
    [anon_sym_ATselector] = ACTIONS(2640),
    [anon_sym_ATavailable] = ACTIONS(2640),
    [anon_sym___builtin_available] = ACTIONS(2638),
    [anon_sym_va_arg] = ACTIONS(2638),
    [anon_sym_ATencode] = ACTIONS(2640),
    [anon_sym_BOOL] = ACTIONS(2638),
    [anon_sym_IMP] = ACTIONS(2638),
    [anon_sym_SEL] = ACTIONS(2638),
    [anon_sym_Class] = ACTIONS(2638),
    [anon_sym_id] = ACTIONS(2638),
  },
  [STATE(1005)] = {
    [ts_builtin_sym_end] = ACTIONS(2846),
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(1006)] = {
    [ts_builtin_sym_end] = ACTIONS(2846),
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(1007)] = {
    [ts_builtin_sym_end] = ACTIONS(2846),
    [sym_identifier] = ACTIONS(2844),
    [aux_sym_preproc_include_token1] = ACTIONS(2844),
    [aux_sym_preproc_include_token2] = ACTIONS(2844),
    [aux_sym_preproc_def_token1] = ACTIONS(2844),
    [aux_sym_preproc_if_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2844),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2844),
    [sym_preproc_directive] = ACTIONS(2844),
    [anon_sym_LPAREN2] = ACTIONS(2846),
    [anon_sym_BANG] = ACTIONS(2846),
    [anon_sym_TILDE] = ACTIONS(2846),
    [anon_sym_DASH] = ACTIONS(2844),
    [anon_sym_PLUS] = ACTIONS(2844),
    [anon_sym_STAR] = ACTIONS(2846),
    [anon_sym_CARET] = ACTIONS(2846),
    [anon_sym_AMP] = ACTIONS(2846),
    [anon_sym_SEMI] = ACTIONS(2846),
    [anon_sym___extension__] = ACTIONS(2844),
    [anon_sym_typedef] = ACTIONS(2844),
    [anon_sym_extern] = ACTIONS(2844),
    [anon_sym___attribute__] = ACTIONS(2844),
    [anon_sym___attribute] = ACTIONS(2844),
    [anon_sym_noreturn] = ACTIONS(2844),
    [anon_sym_LBRACK] = ACTIONS(2846),
    [anon_sym___declspec] = ACTIONS(2844),
    [anon_sym___cdecl] = ACTIONS(2844),
    [anon_sym___clrcall] = ACTIONS(2844),
    [anon_sym___stdcall] = ACTIONS(2844),
    [anon_sym___fastcall] = ACTIONS(2844),
    [anon_sym___thiscall] = ACTIONS(2844),
    [anon_sym___vectorcall] = ACTIONS(2844),
    [anon_sym_LBRACE] = ACTIONS(2846),
    [anon_sym_signed] = ACTIONS(2844),
    [anon_sym_unsigned] = ACTIONS(2844),
    [anon_sym_long] = ACTIONS(2844),
    [anon_sym_short] = ACTIONS(2844),
    [anon_sym_static] = ACTIONS(2844),
    [anon_sym_ATautoreleasepool] = ACTIONS(2846),
    [anon_sym_auto] = ACTIONS(2844),
    [anon_sym_register] = ACTIONS(2844),
    [anon_sym_inline] = ACTIONS(2844),
    [anon_sym___inline] = ACTIONS(2844),
    [anon_sym___inline__] = ACTIONS(2844),
    [anon_sym___forceinline] = ACTIONS(2844),
    [anon_sym_thread_local] = ACTIONS(2844),
    [anon_sym___thread] = ACTIONS(2844),
    [anon_sym_CG_EXTERN] = ACTIONS(2844),
    [anon_sym_CG_INLINE] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2844),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2844),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2844),
    [anon_sym_IBOutlet] = ACTIONS(2844),
    [anon_sym_IBInspectable] = ACTIONS(2844),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2844),
    [anon_sym_NS_INLINE] = ACTIONS(2844),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2844),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2844),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2844),
    [anon_sym_const] = ACTIONS(2844),
    [anon_sym_constexpr] = ACTIONS(2844),
    [anon_sym_volatile] = ACTIONS(2844),
    [anon_sym_restrict] = ACTIONS(2844),
    [anon_sym___restrict__] = ACTIONS(2844),
    [anon_sym__Atomic] = ACTIONS(2844),
    [anon_sym__Noreturn] = ACTIONS(2844),
    [anon_sym__Nonnull] = ACTIONS(2844),
    [anon_sym_nullable] = ACTIONS(2844),
    [anon_sym__Complex] = ACTIONS(2844),
    [anon_sym__Nullable] = ACTIONS(2844),
    [anon_sym__Nullable_result] = ACTIONS(2844),
    [anon_sym__Null_unspecified] = ACTIONS(2844),
    [anon_sym___autoreleasing] = ACTIONS(2844),
    [anon_sym___block] = ACTIONS(2844),
    [anon_sym___bridge] = ACTIONS(2844),
    [anon_sym___bridge_retained] = ACTIONS(2844),
    [anon_sym___bridge_transfer] = ACTIONS(2844),
    [anon_sym___complex] = ACTIONS(2844),
    [anon_sym___const] = ACTIONS(2844),
    [anon_sym___imag] = ACTIONS(2844),
    [anon_sym___kindof] = ACTIONS(2844),
    [anon_sym___nonnull] = ACTIONS(2844),
    [anon_sym___nullable] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2844),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2844),
    [anon_sym___real] = ACTIONS(2844),
    [anon_sym___strong] = ACTIONS(2844),
    [anon_sym___unsafe_unretained] = ACTIONS(2844),
    [anon_sym___unused] = ACTIONS(2844),
    [anon_sym___weak] = ACTIONS(2844),
    [anon_sym_alignas] = ACTIONS(2844),
    [anon_sym__Alignas] = ACTIONS(2844),
    [sym_primitive_type] = ACTIONS(2844),
    [anon_sym_enum] = ACTIONS(2844),
    [anon_sym_struct] = ACTIONS(2844),
    [anon_sym_union] = ACTIONS(2844),
    [anon_sym_if] = ACTIONS(2844),
    [anon_sym_switch] = ACTIONS(2844),
    [anon_sym_case] = ACTIONS(2844),
    [anon_sym_default] = ACTIONS(2844),
    [anon_sym_while] = ACTIONS(2844),
    [anon_sym_do] = ACTIONS(2844),
    [anon_sym_for] = ACTIONS(2844),
    [anon_sym_in] = ACTIONS(2844),
    [anon_sym_return] = ACTIONS(2844),
    [anon_sym_break] = ACTIONS(2844),
    [anon_sym_continue] = ACTIONS(2844),
    [anon_sym_goto] = ACTIONS(2844),
    [anon_sym_DASH_DASH] = ACTIONS(2846),
    [anon_sym_PLUS_PLUS] = ACTIONS(2846),
    [anon_sym_sizeof] = ACTIONS(2844),
    [anon_sym___alignof__] = ACTIONS(2844),
    [anon_sym___alignof] = ACTIONS(2844),
    [anon_sym__alignof] = ACTIONS(2844),
    [anon_sym_alignof] = ACTIONS(2844),
    [anon_sym__Alignof] = ACTIONS(2844),
    [anon_sym_offsetof] = ACTIONS(2844),
    [anon_sym__Generic] = ACTIONS(2844),
    [anon_sym_asm] = ACTIONS(2844),
    [anon_sym___asm__] = ACTIONS(2844),
    [anon_sym___asm] = ACTIONS(2844),
    [sym_number_literal] = ACTIONS(2846),
    [anon_sym_L_SQUOTE] = ACTIONS(2846),
    [anon_sym_u_SQUOTE] = ACTIONS(2846),
    [anon_sym_U_SQUOTE] = ACTIONS(2846),
    [anon_sym_u8_SQUOTE] = ACTIONS(2846),
    [anon_sym_SQUOTE] = ACTIONS(2846),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_DQUOTE] = ACTIONS(2846),
    [anon_sym_L_DQUOTE] = ACTIONS(2846),
    [anon_sym_u_DQUOTE] = ACTIONS(2846),
    [anon_sym_U_DQUOTE] = ACTIONS(2846),
    [anon_sym_u8_DQUOTE] = ACTIONS(2846),
    [sym_true] = ACTIONS(2844),
    [sym_false] = ACTIONS(2844),
    [anon_sym_NULL] = ACTIONS(2844),
    [anon_sym_nullptr] = ACTIONS(2844),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2844),
    [anon_sym___typeof] = ACTIONS(2844),
    [anon_sym_typeof] = ACTIONS(2844),
    [anon_sym_ATimport] = ACTIONS(2846),
    [aux_sym_preproc_undef_token1] = ACTIONS(2844),
    [anon_sym_POUND] = ACTIONS(2844),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2844),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2844),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2844),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2844),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2844),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_API_AVAILABLE] = ACTIONS(2844),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_API_DEPRECATED] = ACTIONS(2844),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2844),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2844),
    [anon_sym___deprecated_msg] = ACTIONS(2844),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2844),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2844),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2844),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2844),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2844),
    [anon_sym_ATprotocol] = ACTIONS(2846),
    [anon_sym_ATinterface] = ACTIONS(2846),
    [anon_sym_ATimplementation] = ACTIONS(2846),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2846),
    [anon_sym_ATselector] = ACTIONS(2846),
    [anon_sym_ATavailable] = ACTIONS(2846),
    [anon_sym___builtin_available] = ACTIONS(2844),
    [anon_sym_va_arg] = ACTIONS(2844),
    [anon_sym_ATencode] = ACTIONS(2846),
    [anon_sym_BOOL] = ACTIONS(2844),
    [anon_sym_IMP] = ACTIONS(2844),
    [anon_sym_SEL] = ACTIONS(2844),
    [anon_sym_Class] = ACTIONS(2844),
    [anon_sym_id] = ACTIONS(2844),
  },
  [STATE(1008)] = {
    [ts_builtin_sym_end] = ACTIONS(2850),
    [sym_identifier] = ACTIONS(2848),
    [aux_sym_preproc_include_token1] = ACTIONS(2848),
    [aux_sym_preproc_include_token2] = ACTIONS(2848),
    [aux_sym_preproc_def_token1] = ACTIONS(2848),
    [aux_sym_preproc_if_token1] = ACTIONS(2848),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2848),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2848),
    [sym_preproc_directive] = ACTIONS(2848),
    [anon_sym_LPAREN2] = ACTIONS(2850),
    [anon_sym_BANG] = ACTIONS(2850),
    [anon_sym_TILDE] = ACTIONS(2850),
    [anon_sym_DASH] = ACTIONS(2848),
    [anon_sym_PLUS] = ACTIONS(2848),
    [anon_sym_STAR] = ACTIONS(2850),
    [anon_sym_CARET] = ACTIONS(2850),
    [anon_sym_AMP] = ACTIONS(2850),
    [anon_sym_SEMI] = ACTIONS(2850),
    [anon_sym___extension__] = ACTIONS(2848),
    [anon_sym_typedef] = ACTIONS(2848),
    [anon_sym_extern] = ACTIONS(2848),
    [anon_sym___attribute__] = ACTIONS(2848),
    [anon_sym___attribute] = ACTIONS(2848),
    [anon_sym_noreturn] = ACTIONS(2848),
    [anon_sym_LBRACK] = ACTIONS(2850),
    [anon_sym___declspec] = ACTIONS(2848),
    [anon_sym___cdecl] = ACTIONS(2848),
    [anon_sym___clrcall] = ACTIONS(2848),
    [anon_sym___stdcall] = ACTIONS(2848),
    [anon_sym___fastcall] = ACTIONS(2848),
    [anon_sym___thiscall] = ACTIONS(2848),
    [anon_sym___vectorcall] = ACTIONS(2848),
    [anon_sym_LBRACE] = ACTIONS(2850),
    [anon_sym_signed] = ACTIONS(2848),
    [anon_sym_unsigned] = ACTIONS(2848),
    [anon_sym_long] = ACTIONS(2848),
    [anon_sym_short] = ACTIONS(2848),
    [anon_sym_static] = ACTIONS(2848),
    [anon_sym_ATautoreleasepool] = ACTIONS(2850),
    [anon_sym_auto] = ACTIONS(2848),
    [anon_sym_register] = ACTIONS(2848),
    [anon_sym_inline] = ACTIONS(2848),
    [anon_sym___inline] = ACTIONS(2848),
    [anon_sym___inline__] = ACTIONS(2848),
    [anon_sym___forceinline] = ACTIONS(2848),
    [anon_sym_thread_local] = ACTIONS(2848),
    [anon_sym___thread] = ACTIONS(2848),
    [anon_sym_CG_EXTERN] = ACTIONS(2848),
    [anon_sym_CG_INLINE] = ACTIONS(2848),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2848),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2848),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2848),
    [anon_sym_IBOutlet] = ACTIONS(2848),
    [anon_sym_IBInspectable] = ACTIONS(2848),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2848),
    [anon_sym_NS_INLINE] = ACTIONS(2848),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2848),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2848),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2848),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2848),
    [anon_sym_const] = ACTIONS(2848),
    [anon_sym_constexpr] = ACTIONS(2848),
    [anon_sym_volatile] = ACTIONS(2848),
    [anon_sym_restrict] = ACTIONS(2848),
    [anon_sym___restrict__] = ACTIONS(2848),
    [anon_sym__Atomic] = ACTIONS(2848),
    [anon_sym__Noreturn] = ACTIONS(2848),
    [anon_sym__Nonnull] = ACTIONS(2848),
    [anon_sym_nullable] = ACTIONS(2848),
    [anon_sym__Complex] = ACTIONS(2848),
    [anon_sym__Nullable] = ACTIONS(2848),
    [anon_sym__Nullable_result] = ACTIONS(2848),
    [anon_sym__Null_unspecified] = ACTIONS(2848),
    [anon_sym___autoreleasing] = ACTIONS(2848),
    [anon_sym___block] = ACTIONS(2848),
    [anon_sym___bridge] = ACTIONS(2848),
    [anon_sym___bridge_retained] = ACTIONS(2848),
    [anon_sym___bridge_transfer] = ACTIONS(2848),
    [anon_sym___complex] = ACTIONS(2848),
    [anon_sym___const] = ACTIONS(2848),
    [anon_sym___imag] = ACTIONS(2848),
    [anon_sym___kindof] = ACTIONS(2848),
    [anon_sym___nonnull] = ACTIONS(2848),
    [anon_sym___nullable] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2848),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2848),
    [anon_sym___real] = ACTIONS(2848),
    [anon_sym___strong] = ACTIONS(2848),
    [anon_sym___unsafe_unretained] = ACTIONS(2848),
    [anon_sym___unused] = ACTIONS(2848),
    [anon_sym___weak] = ACTIONS(2848),
    [anon_sym_alignas] = ACTIONS(2848),
    [anon_sym__Alignas] = ACTIONS(2848),
    [sym_primitive_type] = ACTIONS(2848),
    [anon_sym_enum] = ACTIONS(2848),
    [anon_sym_struct] = ACTIONS(2848),
    [anon_sym_union] = ACTIONS(2848),
    [anon_sym_if] = ACTIONS(2848),
    [anon_sym_switch] = ACTIONS(2848),
    [anon_sym_case] = ACTIONS(2848),
    [anon_sym_default] = ACTIONS(2848),
    [anon_sym_while] = ACTIONS(2848),
    [anon_sym_do] = ACTIONS(2848),
    [anon_sym_for] = ACTIONS(2848),
    [anon_sym_in] = ACTIONS(2848),
    [anon_sym_return] = ACTIONS(2848),
    [anon_sym_break] = ACTIONS(2848),
    [anon_sym_continue] = ACTIONS(2848),
    [anon_sym_goto] = ACTIONS(2848),
    [anon_sym_DASH_DASH] = ACTIONS(2850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2850),
    [anon_sym_sizeof] = ACTIONS(2848),
    [anon_sym___alignof__] = ACTIONS(2848),
    [anon_sym___alignof] = ACTIONS(2848),
    [anon_sym__alignof] = ACTIONS(2848),
    [anon_sym_alignof] = ACTIONS(2848),
    [anon_sym__Alignof] = ACTIONS(2848),
    [anon_sym_offsetof] = ACTIONS(2848),
    [anon_sym__Generic] = ACTIONS(2848),
    [anon_sym_asm] = ACTIONS(2848),
    [anon_sym___asm__] = ACTIONS(2848),
    [anon_sym___asm] = ACTIONS(2848),
    [sym_number_literal] = ACTIONS(2850),
    [anon_sym_L_SQUOTE] = ACTIONS(2850),
    [anon_sym_u_SQUOTE] = ACTIONS(2850),
    [anon_sym_U_SQUOTE] = ACTIONS(2850),
    [anon_sym_u8_SQUOTE] = ACTIONS(2850),
    [anon_sym_SQUOTE] = ACTIONS(2850),
    [anon_sym_AT] = ACTIONS(2848),
    [anon_sym_DQUOTE] = ACTIONS(2850),
    [anon_sym_L_DQUOTE] = ACTIONS(2850),
    [anon_sym_u_DQUOTE] = ACTIONS(2850),
    [anon_sym_U_DQUOTE] = ACTIONS(2850),
    [anon_sym_u8_DQUOTE] = ACTIONS(2850),
    [sym_true] = ACTIONS(2848),
    [sym_false] = ACTIONS(2848),
    [anon_sym_NULL] = ACTIONS(2848),
    [anon_sym_nullptr] = ACTIONS(2848),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2848),
    [anon_sym___typeof] = ACTIONS(2848),
    [anon_sym_typeof] = ACTIONS(2848),
    [anon_sym_ATimport] = ACTIONS(2850),
    [aux_sym_preproc_undef_token1] = ACTIONS(2848),
    [anon_sym_POUND] = ACTIONS(2848),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2848),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2848),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2848),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2848),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2848),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2848),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2848),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_API_AVAILABLE] = ACTIONS(2848),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_API_DEPRECATED] = ACTIONS(2848),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2848),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2848),
    [anon_sym___deprecated_msg] = ACTIONS(2848),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2848),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2848),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2848),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2848),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2848),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2848),
    [anon_sym_ATprotocol] = ACTIONS(2850),
    [anon_sym_ATinterface] = ACTIONS(2850),
    [anon_sym_ATimplementation] = ACTIONS(2850),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2850),
    [anon_sym_ATselector] = ACTIONS(2850),
    [anon_sym_ATavailable] = ACTIONS(2850),
    [anon_sym___builtin_available] = ACTIONS(2848),
    [anon_sym_va_arg] = ACTIONS(2848),
    [anon_sym_ATencode] = ACTIONS(2850),
    [anon_sym_BOOL] = ACTIONS(2848),
    [anon_sym_IMP] = ACTIONS(2848),
    [anon_sym_SEL] = ACTIONS(2848),
    [anon_sym_Class] = ACTIONS(2848),
    [anon_sym_id] = ACTIONS(2848),
  },
  [STATE(1009)] = {
    [ts_builtin_sym_end] = ACTIONS(2574),
    [sym_identifier] = ACTIONS(2572),
    [aux_sym_preproc_include_token1] = ACTIONS(2572),
    [aux_sym_preproc_include_token2] = ACTIONS(2572),
    [aux_sym_preproc_def_token1] = ACTIONS(2572),
    [aux_sym_preproc_if_token1] = ACTIONS(2572),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2572),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2572),
    [sym_preproc_directive] = ACTIONS(2572),
    [anon_sym_LPAREN2] = ACTIONS(2574),
    [anon_sym_BANG] = ACTIONS(2574),
    [anon_sym_TILDE] = ACTIONS(2574),
    [anon_sym_DASH] = ACTIONS(2572),
    [anon_sym_PLUS] = ACTIONS(2572),
    [anon_sym_STAR] = ACTIONS(2574),
    [anon_sym_CARET] = ACTIONS(2574),
    [anon_sym_AMP] = ACTIONS(2574),
    [anon_sym_SEMI] = ACTIONS(2574),
    [anon_sym___extension__] = ACTIONS(2572),
    [anon_sym_typedef] = ACTIONS(2572),
    [anon_sym_extern] = ACTIONS(2572),
    [anon_sym___attribute__] = ACTIONS(2572),
    [anon_sym___attribute] = ACTIONS(2572),
    [anon_sym_noreturn] = ACTIONS(2572),
    [anon_sym_LBRACK] = ACTIONS(2574),
    [anon_sym___declspec] = ACTIONS(2572),
    [anon_sym___cdecl] = ACTIONS(2572),
    [anon_sym___clrcall] = ACTIONS(2572),
    [anon_sym___stdcall] = ACTIONS(2572),
    [anon_sym___fastcall] = ACTIONS(2572),
    [anon_sym___thiscall] = ACTIONS(2572),
    [anon_sym___vectorcall] = ACTIONS(2572),
    [anon_sym_LBRACE] = ACTIONS(2574),
    [anon_sym_signed] = ACTIONS(2572),
    [anon_sym_unsigned] = ACTIONS(2572),
    [anon_sym_long] = ACTIONS(2572),
    [anon_sym_short] = ACTIONS(2572),
    [anon_sym_static] = ACTIONS(2572),
    [anon_sym_ATautoreleasepool] = ACTIONS(2574),
    [anon_sym_auto] = ACTIONS(2572),
    [anon_sym_register] = ACTIONS(2572),
    [anon_sym_inline] = ACTIONS(2572),
    [anon_sym___inline] = ACTIONS(2572),
    [anon_sym___inline__] = ACTIONS(2572),
    [anon_sym___forceinline] = ACTIONS(2572),
    [anon_sym_thread_local] = ACTIONS(2572),
    [anon_sym___thread] = ACTIONS(2572),
    [anon_sym_CG_EXTERN] = ACTIONS(2572),
    [anon_sym_CG_INLINE] = ACTIONS(2572),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2572),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2572),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2572),
    [anon_sym_IBOutlet] = ACTIONS(2572),
    [anon_sym_IBInspectable] = ACTIONS(2572),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2572),
    [anon_sym_NS_INLINE] = ACTIONS(2572),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2572),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2572),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2572),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2572),
    [anon_sym_const] = ACTIONS(2572),
    [anon_sym_constexpr] = ACTIONS(2572),
    [anon_sym_volatile] = ACTIONS(2572),
    [anon_sym_restrict] = ACTIONS(2572),
    [anon_sym___restrict__] = ACTIONS(2572),
    [anon_sym__Atomic] = ACTIONS(2572),
    [anon_sym__Noreturn] = ACTIONS(2572),
    [anon_sym__Nonnull] = ACTIONS(2572),
    [anon_sym_nullable] = ACTIONS(2572),
    [anon_sym__Complex] = ACTIONS(2572),
    [anon_sym__Nullable] = ACTIONS(2572),
    [anon_sym__Nullable_result] = ACTIONS(2572),
    [anon_sym__Null_unspecified] = ACTIONS(2572),
    [anon_sym___autoreleasing] = ACTIONS(2572),
    [anon_sym___block] = ACTIONS(2572),
    [anon_sym___bridge] = ACTIONS(2572),
    [anon_sym___bridge_retained] = ACTIONS(2572),
    [anon_sym___bridge_transfer] = ACTIONS(2572),
    [anon_sym___complex] = ACTIONS(2572),
    [anon_sym___const] = ACTIONS(2572),
    [anon_sym___imag] = ACTIONS(2572),
    [anon_sym___kindof] = ACTIONS(2572),
    [anon_sym___nonnull] = ACTIONS(2572),
    [anon_sym___nullable] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2572),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2572),
    [anon_sym___real] = ACTIONS(2572),
    [anon_sym___strong] = ACTIONS(2572),
    [anon_sym___unsafe_unretained] = ACTIONS(2572),
    [anon_sym___unused] = ACTIONS(2572),
    [anon_sym___weak] = ACTIONS(2572),
    [anon_sym_alignas] = ACTIONS(2572),
    [anon_sym__Alignas] = ACTIONS(2572),
    [sym_primitive_type] = ACTIONS(2572),
    [anon_sym_enum] = ACTIONS(2572),
    [anon_sym_struct] = ACTIONS(2572),
    [anon_sym_union] = ACTIONS(2572),
    [anon_sym_if] = ACTIONS(2572),
    [anon_sym_switch] = ACTIONS(2572),
    [anon_sym_case] = ACTIONS(2572),
    [anon_sym_default] = ACTIONS(2572),
    [anon_sym_while] = ACTIONS(2572),
    [anon_sym_do] = ACTIONS(2572),
    [anon_sym_for] = ACTIONS(2572),
    [anon_sym_in] = ACTIONS(2572),
    [anon_sym_return] = ACTIONS(2572),
    [anon_sym_break] = ACTIONS(2572),
    [anon_sym_continue] = ACTIONS(2572),
    [anon_sym_goto] = ACTIONS(2572),
    [anon_sym_DASH_DASH] = ACTIONS(2574),
    [anon_sym_PLUS_PLUS] = ACTIONS(2574),
    [anon_sym_sizeof] = ACTIONS(2572),
    [anon_sym___alignof__] = ACTIONS(2572),
    [anon_sym___alignof] = ACTIONS(2572),
    [anon_sym__alignof] = ACTIONS(2572),
    [anon_sym_alignof] = ACTIONS(2572),
    [anon_sym__Alignof] = ACTIONS(2572),
    [anon_sym_offsetof] = ACTIONS(2572),
    [anon_sym__Generic] = ACTIONS(2572),
    [anon_sym_asm] = ACTIONS(2572),
    [anon_sym___asm__] = ACTIONS(2572),
    [anon_sym___asm] = ACTIONS(2572),
    [sym_number_literal] = ACTIONS(2574),
    [anon_sym_L_SQUOTE] = ACTIONS(2574),
    [anon_sym_u_SQUOTE] = ACTIONS(2574),
    [anon_sym_U_SQUOTE] = ACTIONS(2574),
    [anon_sym_u8_SQUOTE] = ACTIONS(2574),
    [anon_sym_SQUOTE] = ACTIONS(2574),
    [anon_sym_AT] = ACTIONS(2572),
    [anon_sym_DQUOTE] = ACTIONS(2574),
    [anon_sym_L_DQUOTE] = ACTIONS(2574),
    [anon_sym_u_DQUOTE] = ACTIONS(2574),
    [anon_sym_U_DQUOTE] = ACTIONS(2574),
    [anon_sym_u8_DQUOTE] = ACTIONS(2574),
    [sym_true] = ACTIONS(2572),
    [sym_false] = ACTIONS(2572),
    [anon_sym_NULL] = ACTIONS(2572),
    [anon_sym_nullptr] = ACTIONS(2572),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2572),
    [anon_sym___typeof] = ACTIONS(2572),
    [anon_sym_typeof] = ACTIONS(2572),
    [anon_sym_ATimport] = ACTIONS(2574),
    [aux_sym_preproc_undef_token1] = ACTIONS(2572),
    [anon_sym_POUND] = ACTIONS(2572),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2572),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2572),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2572),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2572),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2572),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2572),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2572),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_API_AVAILABLE] = ACTIONS(2572),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_API_DEPRECATED] = ACTIONS(2572),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2572),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2572),
    [anon_sym___deprecated_msg] = ACTIONS(2572),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2572),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2572),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2572),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2572),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2572),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2572),
    [anon_sym_ATprotocol] = ACTIONS(2574),
    [anon_sym_ATinterface] = ACTIONS(2574),
    [anon_sym_ATimplementation] = ACTIONS(2574),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2574),
    [anon_sym_ATselector] = ACTIONS(2574),
    [anon_sym_ATavailable] = ACTIONS(2574),
    [anon_sym___builtin_available] = ACTIONS(2572),
    [anon_sym_va_arg] = ACTIONS(2572),
    [anon_sym_ATencode] = ACTIONS(2574),
    [anon_sym_BOOL] = ACTIONS(2572),
    [anon_sym_IMP] = ACTIONS(2572),
    [anon_sym_SEL] = ACTIONS(2572),
    [anon_sym_Class] = ACTIONS(2572),
    [anon_sym_id] = ACTIONS(2572),
  },
  [STATE(1010)] = {
    [ts_builtin_sym_end] = ACTIONS(2578),
    [sym_identifier] = ACTIONS(2576),
    [aux_sym_preproc_include_token1] = ACTIONS(2576),
    [aux_sym_preproc_include_token2] = ACTIONS(2576),
    [aux_sym_preproc_def_token1] = ACTIONS(2576),
    [aux_sym_preproc_if_token1] = ACTIONS(2576),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2576),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2576),
    [sym_preproc_directive] = ACTIONS(2576),
    [anon_sym_LPAREN2] = ACTIONS(2578),
    [anon_sym_BANG] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2578),
    [anon_sym_DASH] = ACTIONS(2576),
    [anon_sym_PLUS] = ACTIONS(2576),
    [anon_sym_STAR] = ACTIONS(2578),
    [anon_sym_CARET] = ACTIONS(2578),
    [anon_sym_AMP] = ACTIONS(2578),
    [anon_sym_SEMI] = ACTIONS(2578),
    [anon_sym___extension__] = ACTIONS(2576),
    [anon_sym_typedef] = ACTIONS(2576),
    [anon_sym_extern] = ACTIONS(2576),
    [anon_sym___attribute__] = ACTIONS(2576),
    [anon_sym___attribute] = ACTIONS(2576),
    [anon_sym_noreturn] = ACTIONS(2576),
    [anon_sym_LBRACK] = ACTIONS(2578),
    [anon_sym___declspec] = ACTIONS(2576),
    [anon_sym___cdecl] = ACTIONS(2576),
    [anon_sym___clrcall] = ACTIONS(2576),
    [anon_sym___stdcall] = ACTIONS(2576),
    [anon_sym___fastcall] = ACTIONS(2576),
    [anon_sym___thiscall] = ACTIONS(2576),
    [anon_sym___vectorcall] = ACTIONS(2576),
    [anon_sym_LBRACE] = ACTIONS(2578),
    [anon_sym_signed] = ACTIONS(2576),
    [anon_sym_unsigned] = ACTIONS(2576),
    [anon_sym_long] = ACTIONS(2576),
    [anon_sym_short] = ACTIONS(2576),
    [anon_sym_static] = ACTIONS(2576),
    [anon_sym_ATautoreleasepool] = ACTIONS(2578),
    [anon_sym_auto] = ACTIONS(2576),
    [anon_sym_register] = ACTIONS(2576),
    [anon_sym_inline] = ACTIONS(2576),
    [anon_sym___inline] = ACTIONS(2576),
    [anon_sym___inline__] = ACTIONS(2576),
    [anon_sym___forceinline] = ACTIONS(2576),
    [anon_sym_thread_local] = ACTIONS(2576),
    [anon_sym___thread] = ACTIONS(2576),
    [anon_sym_CG_EXTERN] = ACTIONS(2576),
    [anon_sym_CG_INLINE] = ACTIONS(2576),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2576),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2576),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2576),
    [anon_sym_IBOutlet] = ACTIONS(2576),
    [anon_sym_IBInspectable] = ACTIONS(2576),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2576),
    [anon_sym_NS_INLINE] = ACTIONS(2576),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2576),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2576),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2576),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2576),
    [anon_sym_const] = ACTIONS(2576),
    [anon_sym_constexpr] = ACTIONS(2576),
    [anon_sym_volatile] = ACTIONS(2576),
    [anon_sym_restrict] = ACTIONS(2576),
    [anon_sym___restrict__] = ACTIONS(2576),
    [anon_sym__Atomic] = ACTIONS(2576),
    [anon_sym__Noreturn] = ACTIONS(2576),
    [anon_sym__Nonnull] = ACTIONS(2576),
    [anon_sym_nullable] = ACTIONS(2576),
    [anon_sym__Complex] = ACTIONS(2576),
    [anon_sym__Nullable] = ACTIONS(2576),
    [anon_sym__Nullable_result] = ACTIONS(2576),
    [anon_sym__Null_unspecified] = ACTIONS(2576),
    [anon_sym___autoreleasing] = ACTIONS(2576),
    [anon_sym___block] = ACTIONS(2576),
    [anon_sym___bridge] = ACTIONS(2576),
    [anon_sym___bridge_retained] = ACTIONS(2576),
    [anon_sym___bridge_transfer] = ACTIONS(2576),
    [anon_sym___complex] = ACTIONS(2576),
    [anon_sym___const] = ACTIONS(2576),
    [anon_sym___imag] = ACTIONS(2576),
    [anon_sym___kindof] = ACTIONS(2576),
    [anon_sym___nonnull] = ACTIONS(2576),
    [anon_sym___nullable] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2576),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2576),
    [anon_sym___real] = ACTIONS(2576),
    [anon_sym___strong] = ACTIONS(2576),
    [anon_sym___unsafe_unretained] = ACTIONS(2576),
    [anon_sym___unused] = ACTIONS(2576),
    [anon_sym___weak] = ACTIONS(2576),
    [anon_sym_alignas] = ACTIONS(2576),
    [anon_sym__Alignas] = ACTIONS(2576),
    [sym_primitive_type] = ACTIONS(2576),
    [anon_sym_enum] = ACTIONS(2576),
    [anon_sym_struct] = ACTIONS(2576),
    [anon_sym_union] = ACTIONS(2576),
    [anon_sym_if] = ACTIONS(2576),
    [anon_sym_switch] = ACTIONS(2576),
    [anon_sym_case] = ACTIONS(2576),
    [anon_sym_default] = ACTIONS(2576),
    [anon_sym_while] = ACTIONS(2576),
    [anon_sym_do] = ACTIONS(2576),
    [anon_sym_for] = ACTIONS(2576),
    [anon_sym_in] = ACTIONS(2576),
    [anon_sym_return] = ACTIONS(2576),
    [anon_sym_break] = ACTIONS(2576),
    [anon_sym_continue] = ACTIONS(2576),
    [anon_sym_goto] = ACTIONS(2576),
    [anon_sym_DASH_DASH] = ACTIONS(2578),
    [anon_sym_PLUS_PLUS] = ACTIONS(2578),
    [anon_sym_sizeof] = ACTIONS(2576),
    [anon_sym___alignof__] = ACTIONS(2576),
    [anon_sym___alignof] = ACTIONS(2576),
    [anon_sym__alignof] = ACTIONS(2576),
    [anon_sym_alignof] = ACTIONS(2576),
    [anon_sym__Alignof] = ACTIONS(2576),
    [anon_sym_offsetof] = ACTIONS(2576),
    [anon_sym__Generic] = ACTIONS(2576),
    [anon_sym_asm] = ACTIONS(2576),
    [anon_sym___asm__] = ACTIONS(2576),
    [anon_sym___asm] = ACTIONS(2576),
    [sym_number_literal] = ACTIONS(2578),
    [anon_sym_L_SQUOTE] = ACTIONS(2578),
    [anon_sym_u_SQUOTE] = ACTIONS(2578),
    [anon_sym_U_SQUOTE] = ACTIONS(2578),
    [anon_sym_u8_SQUOTE] = ACTIONS(2578),
    [anon_sym_SQUOTE] = ACTIONS(2578),
    [anon_sym_AT] = ACTIONS(2576),
    [anon_sym_DQUOTE] = ACTIONS(2578),
    [anon_sym_L_DQUOTE] = ACTIONS(2578),
    [anon_sym_u_DQUOTE] = ACTIONS(2578),
    [anon_sym_U_DQUOTE] = ACTIONS(2578),
    [anon_sym_u8_DQUOTE] = ACTIONS(2578),
    [sym_true] = ACTIONS(2576),
    [sym_false] = ACTIONS(2576),
    [anon_sym_NULL] = ACTIONS(2576),
    [anon_sym_nullptr] = ACTIONS(2576),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2576),
    [anon_sym___typeof] = ACTIONS(2576),
    [anon_sym_typeof] = ACTIONS(2576),
    [anon_sym_ATimport] = ACTIONS(2578),
    [aux_sym_preproc_undef_token1] = ACTIONS(2576),
    [anon_sym_POUND] = ACTIONS(2576),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2576),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2576),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2576),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2576),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2576),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2576),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2576),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_API_AVAILABLE] = ACTIONS(2576),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_API_DEPRECATED] = ACTIONS(2576),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2576),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2576),
    [anon_sym___deprecated_msg] = ACTIONS(2576),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2576),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2576),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2576),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2576),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2576),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2576),
    [anon_sym_ATprotocol] = ACTIONS(2578),
    [anon_sym_ATinterface] = ACTIONS(2578),
    [anon_sym_ATimplementation] = ACTIONS(2578),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2578),
    [anon_sym_ATselector] = ACTIONS(2578),
    [anon_sym_ATavailable] = ACTIONS(2578),
    [anon_sym___builtin_available] = ACTIONS(2576),
    [anon_sym_va_arg] = ACTIONS(2576),
    [anon_sym_ATencode] = ACTIONS(2578),
    [anon_sym_BOOL] = ACTIONS(2576),
    [anon_sym_IMP] = ACTIONS(2576),
    [anon_sym_SEL] = ACTIONS(2576),
    [anon_sym_Class] = ACTIONS(2576),
    [anon_sym_id] = ACTIONS(2576),
  },
  [STATE(1011)] = {
    [ts_builtin_sym_end] = ACTIONS(2556),
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(1012)] = {
    [ts_builtin_sym_end] = ACTIONS(2556),
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_include_token1] = ACTIONS(2554),
    [aux_sym_preproc_include_token2] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_case] = ACTIONS(2554),
    [anon_sym_default] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_ATimport] = ACTIONS(2556),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_POUND] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATprotocol] = ACTIONS(2556),
    [anon_sym_ATinterface] = ACTIONS(2556),
    [anon_sym_ATimplementation] = ACTIONS(2556),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(1013)] = {
    [ts_builtin_sym_end] = ACTIONS(2674),
    [sym_identifier] = ACTIONS(2672),
    [aux_sym_preproc_include_token1] = ACTIONS(2672),
    [aux_sym_preproc_include_token2] = ACTIONS(2672),
    [aux_sym_preproc_def_token1] = ACTIONS(2672),
    [aux_sym_preproc_if_token1] = ACTIONS(2672),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2672),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2672),
    [sym_preproc_directive] = ACTIONS(2672),
    [anon_sym_LPAREN2] = ACTIONS(2674),
    [anon_sym_BANG] = ACTIONS(2674),
    [anon_sym_TILDE] = ACTIONS(2674),
    [anon_sym_DASH] = ACTIONS(2672),
    [anon_sym_PLUS] = ACTIONS(2672),
    [anon_sym_STAR] = ACTIONS(2674),
    [anon_sym_CARET] = ACTIONS(2674),
    [anon_sym_AMP] = ACTIONS(2674),
    [anon_sym_SEMI] = ACTIONS(2674),
    [anon_sym___extension__] = ACTIONS(2672),
    [anon_sym_typedef] = ACTIONS(2672),
    [anon_sym_extern] = ACTIONS(2672),
    [anon_sym___attribute__] = ACTIONS(2672),
    [anon_sym___attribute] = ACTIONS(2672),
    [anon_sym_noreturn] = ACTIONS(2672),
    [anon_sym_LBRACK] = ACTIONS(2674),
    [anon_sym___declspec] = ACTIONS(2672),
    [anon_sym___cdecl] = ACTIONS(2672),
    [anon_sym___clrcall] = ACTIONS(2672),
    [anon_sym___stdcall] = ACTIONS(2672),
    [anon_sym___fastcall] = ACTIONS(2672),
    [anon_sym___thiscall] = ACTIONS(2672),
    [anon_sym___vectorcall] = ACTIONS(2672),
    [anon_sym_LBRACE] = ACTIONS(2674),
    [anon_sym_signed] = ACTIONS(2672),
    [anon_sym_unsigned] = ACTIONS(2672),
    [anon_sym_long] = ACTIONS(2672),
    [anon_sym_short] = ACTIONS(2672),
    [anon_sym_static] = ACTIONS(2672),
    [anon_sym_ATautoreleasepool] = ACTIONS(2674),
    [anon_sym_auto] = ACTIONS(2672),
    [anon_sym_register] = ACTIONS(2672),
    [anon_sym_inline] = ACTIONS(2672),
    [anon_sym___inline] = ACTIONS(2672),
    [anon_sym___inline__] = ACTIONS(2672),
    [anon_sym___forceinline] = ACTIONS(2672),
    [anon_sym_thread_local] = ACTIONS(2672),
    [anon_sym___thread] = ACTIONS(2672),
    [anon_sym_CG_EXTERN] = ACTIONS(2672),
    [anon_sym_CG_INLINE] = ACTIONS(2672),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2672),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2672),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2672),
    [anon_sym_IBOutlet] = ACTIONS(2672),
    [anon_sym_IBInspectable] = ACTIONS(2672),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2672),
    [anon_sym_NS_INLINE] = ACTIONS(2672),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2672),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2672),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2672),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2672),
    [anon_sym_const] = ACTIONS(2672),
    [anon_sym_constexpr] = ACTIONS(2672),
    [anon_sym_volatile] = ACTIONS(2672),
    [anon_sym_restrict] = ACTIONS(2672),
    [anon_sym___restrict__] = ACTIONS(2672),
    [anon_sym__Atomic] = ACTIONS(2672),
    [anon_sym__Noreturn] = ACTIONS(2672),
    [anon_sym__Nonnull] = ACTIONS(2672),
    [anon_sym_nullable] = ACTIONS(2672),
    [anon_sym__Complex] = ACTIONS(2672),
    [anon_sym__Nullable] = ACTIONS(2672),
    [anon_sym__Nullable_result] = ACTIONS(2672),
    [anon_sym__Null_unspecified] = ACTIONS(2672),
    [anon_sym___autoreleasing] = ACTIONS(2672),
    [anon_sym___block] = ACTIONS(2672),
    [anon_sym___bridge] = ACTIONS(2672),
    [anon_sym___bridge_retained] = ACTIONS(2672),
    [anon_sym___bridge_transfer] = ACTIONS(2672),
    [anon_sym___complex] = ACTIONS(2672),
    [anon_sym___const] = ACTIONS(2672),
    [anon_sym___imag] = ACTIONS(2672),
    [anon_sym___kindof] = ACTIONS(2672),
    [anon_sym___nonnull] = ACTIONS(2672),
    [anon_sym___nullable] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2672),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2672),
    [anon_sym___real] = ACTIONS(2672),
    [anon_sym___strong] = ACTIONS(2672),
    [anon_sym___unsafe_unretained] = ACTIONS(2672),
    [anon_sym___unused] = ACTIONS(2672),
    [anon_sym___weak] = ACTIONS(2672),
    [anon_sym_alignas] = ACTIONS(2672),
    [anon_sym__Alignas] = ACTIONS(2672),
    [sym_primitive_type] = ACTIONS(2672),
    [anon_sym_enum] = ACTIONS(2672),
    [anon_sym_struct] = ACTIONS(2672),
    [anon_sym_union] = ACTIONS(2672),
    [anon_sym_if] = ACTIONS(2672),
    [anon_sym_switch] = ACTIONS(2672),
    [anon_sym_case] = ACTIONS(2672),
    [anon_sym_default] = ACTIONS(2672),
    [anon_sym_while] = ACTIONS(2672),
    [anon_sym_do] = ACTIONS(2672),
    [anon_sym_for] = ACTIONS(2672),
    [anon_sym_in] = ACTIONS(2672),
    [anon_sym_return] = ACTIONS(2672),
    [anon_sym_break] = ACTIONS(2672),
    [anon_sym_continue] = ACTIONS(2672),
    [anon_sym_goto] = ACTIONS(2672),
    [anon_sym_DASH_DASH] = ACTIONS(2674),
    [anon_sym_PLUS_PLUS] = ACTIONS(2674),
    [anon_sym_sizeof] = ACTIONS(2672),
    [anon_sym___alignof__] = ACTIONS(2672),
    [anon_sym___alignof] = ACTIONS(2672),
    [anon_sym__alignof] = ACTIONS(2672),
    [anon_sym_alignof] = ACTIONS(2672),
    [anon_sym__Alignof] = ACTIONS(2672),
    [anon_sym_offsetof] = ACTIONS(2672),
    [anon_sym__Generic] = ACTIONS(2672),
    [anon_sym_asm] = ACTIONS(2672),
    [anon_sym___asm__] = ACTIONS(2672),
    [anon_sym___asm] = ACTIONS(2672),
    [sym_number_literal] = ACTIONS(2674),
    [anon_sym_L_SQUOTE] = ACTIONS(2674),
    [anon_sym_u_SQUOTE] = ACTIONS(2674),
    [anon_sym_U_SQUOTE] = ACTIONS(2674),
    [anon_sym_u8_SQUOTE] = ACTIONS(2674),
    [anon_sym_SQUOTE] = ACTIONS(2674),
    [anon_sym_AT] = ACTIONS(2672),
    [anon_sym_DQUOTE] = ACTIONS(2674),
    [anon_sym_L_DQUOTE] = ACTIONS(2674),
    [anon_sym_u_DQUOTE] = ACTIONS(2674),
    [anon_sym_U_DQUOTE] = ACTIONS(2674),
    [anon_sym_u8_DQUOTE] = ACTIONS(2674),
    [sym_true] = ACTIONS(2672),
    [sym_false] = ACTIONS(2672),
    [anon_sym_NULL] = ACTIONS(2672),
    [anon_sym_nullptr] = ACTIONS(2672),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2672),
    [anon_sym___typeof] = ACTIONS(2672),
    [anon_sym_typeof] = ACTIONS(2672),
    [anon_sym_ATimport] = ACTIONS(2674),
    [aux_sym_preproc_undef_token1] = ACTIONS(2672),
    [anon_sym_POUND] = ACTIONS(2672),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2672),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2672),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2672),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2672),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2672),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2672),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2672),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_API_AVAILABLE] = ACTIONS(2672),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_API_DEPRECATED] = ACTIONS(2672),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2672),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2672),
    [anon_sym___deprecated_msg] = ACTIONS(2672),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2672),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2672),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2672),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2672),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2672),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2672),
    [anon_sym_ATprotocol] = ACTIONS(2674),
    [anon_sym_ATinterface] = ACTIONS(2674),
    [anon_sym_ATimplementation] = ACTIONS(2674),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2674),
    [anon_sym_ATselector] = ACTIONS(2674),
    [anon_sym_ATavailable] = ACTIONS(2674),
    [anon_sym___builtin_available] = ACTIONS(2672),
    [anon_sym_va_arg] = ACTIONS(2672),
    [anon_sym_ATencode] = ACTIONS(2674),
    [anon_sym_BOOL] = ACTIONS(2672),
    [anon_sym_IMP] = ACTIONS(2672),
    [anon_sym_SEL] = ACTIONS(2672),
    [anon_sym_Class] = ACTIONS(2672),
    [anon_sym_id] = ACTIONS(2672),
  },
  [STATE(1014)] = {
    [ts_builtin_sym_end] = ACTIONS(2690),
    [sym_identifier] = ACTIONS(2688),
    [aux_sym_preproc_include_token1] = ACTIONS(2688),
    [aux_sym_preproc_include_token2] = ACTIONS(2688),
    [aux_sym_preproc_def_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2688),
    [sym_preproc_directive] = ACTIONS(2688),
    [anon_sym_LPAREN2] = ACTIONS(2690),
    [anon_sym_BANG] = ACTIONS(2690),
    [anon_sym_TILDE] = ACTIONS(2690),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2690),
    [anon_sym_CARET] = ACTIONS(2690),
    [anon_sym_AMP] = ACTIONS(2690),
    [anon_sym_SEMI] = ACTIONS(2690),
    [anon_sym___extension__] = ACTIONS(2688),
    [anon_sym_typedef] = ACTIONS(2688),
    [anon_sym_extern] = ACTIONS(2688),
    [anon_sym___attribute__] = ACTIONS(2688),
    [anon_sym___attribute] = ACTIONS(2688),
    [anon_sym_noreturn] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym___declspec] = ACTIONS(2688),
    [anon_sym___cdecl] = ACTIONS(2688),
    [anon_sym___clrcall] = ACTIONS(2688),
    [anon_sym___stdcall] = ACTIONS(2688),
    [anon_sym___fastcall] = ACTIONS(2688),
    [anon_sym___thiscall] = ACTIONS(2688),
    [anon_sym___vectorcall] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_signed] = ACTIONS(2688),
    [anon_sym_unsigned] = ACTIONS(2688),
    [anon_sym_long] = ACTIONS(2688),
    [anon_sym_short] = ACTIONS(2688),
    [anon_sym_static] = ACTIONS(2688),
    [anon_sym_ATautoreleasepool] = ACTIONS(2690),
    [anon_sym_auto] = ACTIONS(2688),
    [anon_sym_register] = ACTIONS(2688),
    [anon_sym_inline] = ACTIONS(2688),
    [anon_sym___inline] = ACTIONS(2688),
    [anon_sym___inline__] = ACTIONS(2688),
    [anon_sym___forceinline] = ACTIONS(2688),
    [anon_sym_thread_local] = ACTIONS(2688),
    [anon_sym___thread] = ACTIONS(2688),
    [anon_sym_CG_EXTERN] = ACTIONS(2688),
    [anon_sym_CG_INLINE] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2688),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2688),
    [anon_sym_IBOutlet] = ACTIONS(2688),
    [anon_sym_IBInspectable] = ACTIONS(2688),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2688),
    [anon_sym_NS_INLINE] = ACTIONS(2688),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2688),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2688),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2688),
    [anon_sym_const] = ACTIONS(2688),
    [anon_sym_constexpr] = ACTIONS(2688),
    [anon_sym_volatile] = ACTIONS(2688),
    [anon_sym_restrict] = ACTIONS(2688),
    [anon_sym___restrict__] = ACTIONS(2688),
    [anon_sym__Atomic] = ACTIONS(2688),
    [anon_sym__Noreturn] = ACTIONS(2688),
    [anon_sym__Nonnull] = ACTIONS(2688),
    [anon_sym_nullable] = ACTIONS(2688),
    [anon_sym__Complex] = ACTIONS(2688),
    [anon_sym__Nullable] = ACTIONS(2688),
    [anon_sym__Nullable_result] = ACTIONS(2688),
    [anon_sym__Null_unspecified] = ACTIONS(2688),
    [anon_sym___autoreleasing] = ACTIONS(2688),
    [anon_sym___block] = ACTIONS(2688),
    [anon_sym___bridge] = ACTIONS(2688),
    [anon_sym___bridge_retained] = ACTIONS(2688),
    [anon_sym___bridge_transfer] = ACTIONS(2688),
    [anon_sym___complex] = ACTIONS(2688),
    [anon_sym___const] = ACTIONS(2688),
    [anon_sym___imag] = ACTIONS(2688),
    [anon_sym___kindof] = ACTIONS(2688),
    [anon_sym___nonnull] = ACTIONS(2688),
    [anon_sym___nullable] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2688),
    [anon_sym___real] = ACTIONS(2688),
    [anon_sym___strong] = ACTIONS(2688),
    [anon_sym___unsafe_unretained] = ACTIONS(2688),
    [anon_sym___unused] = ACTIONS(2688),
    [anon_sym___weak] = ACTIONS(2688),
    [anon_sym_alignas] = ACTIONS(2688),
    [anon_sym__Alignas] = ACTIONS(2688),
    [sym_primitive_type] = ACTIONS(2688),
    [anon_sym_enum] = ACTIONS(2688),
    [anon_sym_struct] = ACTIONS(2688),
    [anon_sym_union] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_switch] = ACTIONS(2688),
    [anon_sym_case] = ACTIONS(2688),
    [anon_sym_default] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_in] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_sizeof] = ACTIONS(2688),
    [anon_sym___alignof__] = ACTIONS(2688),
    [anon_sym___alignof] = ACTIONS(2688),
    [anon_sym__alignof] = ACTIONS(2688),
    [anon_sym_alignof] = ACTIONS(2688),
    [anon_sym__Alignof] = ACTIONS(2688),
    [anon_sym_offsetof] = ACTIONS(2688),
    [anon_sym__Generic] = ACTIONS(2688),
    [anon_sym_asm] = ACTIONS(2688),
    [anon_sym___asm__] = ACTIONS(2688),
    [anon_sym___asm] = ACTIONS(2688),
    [sym_number_literal] = ACTIONS(2690),
    [anon_sym_L_SQUOTE] = ACTIONS(2690),
    [anon_sym_u_SQUOTE] = ACTIONS(2690),
    [anon_sym_U_SQUOTE] = ACTIONS(2690),
    [anon_sym_u8_SQUOTE] = ACTIONS(2690),
    [anon_sym_SQUOTE] = ACTIONS(2690),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_DQUOTE] = ACTIONS(2690),
    [anon_sym_L_DQUOTE] = ACTIONS(2690),
    [anon_sym_u_DQUOTE] = ACTIONS(2690),
    [anon_sym_U_DQUOTE] = ACTIONS(2690),
    [anon_sym_u8_DQUOTE] = ACTIONS(2690),
    [sym_true] = ACTIONS(2688),
    [sym_false] = ACTIONS(2688),
    [anon_sym_NULL] = ACTIONS(2688),
    [anon_sym_nullptr] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2688),
    [anon_sym___typeof] = ACTIONS(2688),
    [anon_sym_typeof] = ACTIONS(2688),
    [anon_sym_ATimport] = ACTIONS(2690),
    [aux_sym_preproc_undef_token1] = ACTIONS(2688),
    [anon_sym_POUND] = ACTIONS(2688),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2688),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2688),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2688),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_API_AVAILABLE] = ACTIONS(2688),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_API_DEPRECATED] = ACTIONS(2688),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym___deprecated_msg] = ACTIONS(2688),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2688),
    [anon_sym_ATprotocol] = ACTIONS(2690),
    [anon_sym_ATinterface] = ACTIONS(2690),
    [anon_sym_ATimplementation] = ACTIONS(2690),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2690),
    [anon_sym_ATselector] = ACTIONS(2690),
    [anon_sym_ATavailable] = ACTIONS(2690),
    [anon_sym___builtin_available] = ACTIONS(2688),
    [anon_sym_va_arg] = ACTIONS(2688),
    [anon_sym_ATencode] = ACTIONS(2690),
    [anon_sym_BOOL] = ACTIONS(2688),
    [anon_sym_IMP] = ACTIONS(2688),
    [anon_sym_SEL] = ACTIONS(2688),
    [anon_sym_Class] = ACTIONS(2688),
    [anon_sym_id] = ACTIONS(2688),
  },
  [STATE(1015)] = {
    [ts_builtin_sym_end] = ACTIONS(2862),
    [sym_identifier] = ACTIONS(2860),
    [aux_sym_preproc_include_token1] = ACTIONS(2860),
    [aux_sym_preproc_include_token2] = ACTIONS(2860),
    [aux_sym_preproc_def_token1] = ACTIONS(2860),
    [aux_sym_preproc_if_token1] = ACTIONS(2860),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2860),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2860),
    [sym_preproc_directive] = ACTIONS(2860),
    [anon_sym_LPAREN2] = ACTIONS(2862),
    [anon_sym_BANG] = ACTIONS(2862),
    [anon_sym_TILDE] = ACTIONS(2862),
    [anon_sym_DASH] = ACTIONS(2860),
    [anon_sym_PLUS] = ACTIONS(2860),
    [anon_sym_STAR] = ACTIONS(2862),
    [anon_sym_CARET] = ACTIONS(2862),
    [anon_sym_AMP] = ACTIONS(2862),
    [anon_sym_SEMI] = ACTIONS(2862),
    [anon_sym___extension__] = ACTIONS(2860),
    [anon_sym_typedef] = ACTIONS(2860),
    [anon_sym_extern] = ACTIONS(2860),
    [anon_sym___attribute__] = ACTIONS(2860),
    [anon_sym___attribute] = ACTIONS(2860),
    [anon_sym_noreturn] = ACTIONS(2860),
    [anon_sym_LBRACK] = ACTIONS(2862),
    [anon_sym___declspec] = ACTIONS(2860),
    [anon_sym___cdecl] = ACTIONS(2860),
    [anon_sym___clrcall] = ACTIONS(2860),
    [anon_sym___stdcall] = ACTIONS(2860),
    [anon_sym___fastcall] = ACTIONS(2860),
    [anon_sym___thiscall] = ACTIONS(2860),
    [anon_sym___vectorcall] = ACTIONS(2860),
    [anon_sym_LBRACE] = ACTIONS(2862),
    [anon_sym_signed] = ACTIONS(2860),
    [anon_sym_unsigned] = ACTIONS(2860),
    [anon_sym_long] = ACTIONS(2860),
    [anon_sym_short] = ACTIONS(2860),
    [anon_sym_static] = ACTIONS(2860),
    [anon_sym_ATautoreleasepool] = ACTIONS(2862),
    [anon_sym_auto] = ACTIONS(2860),
    [anon_sym_register] = ACTIONS(2860),
    [anon_sym_inline] = ACTIONS(2860),
    [anon_sym___inline] = ACTIONS(2860),
    [anon_sym___inline__] = ACTIONS(2860),
    [anon_sym___forceinline] = ACTIONS(2860),
    [anon_sym_thread_local] = ACTIONS(2860),
    [anon_sym___thread] = ACTIONS(2860),
    [anon_sym_CG_EXTERN] = ACTIONS(2860),
    [anon_sym_CG_INLINE] = ACTIONS(2860),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2860),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2860),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2860),
    [anon_sym_IBOutlet] = ACTIONS(2860),
    [anon_sym_IBInspectable] = ACTIONS(2860),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2860),
    [anon_sym_NS_INLINE] = ACTIONS(2860),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2860),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2860),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2860),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2860),
    [anon_sym_const] = ACTIONS(2860),
    [anon_sym_constexpr] = ACTIONS(2860),
    [anon_sym_volatile] = ACTIONS(2860),
    [anon_sym_restrict] = ACTIONS(2860),
    [anon_sym___restrict__] = ACTIONS(2860),
    [anon_sym__Atomic] = ACTIONS(2860),
    [anon_sym__Noreturn] = ACTIONS(2860),
    [anon_sym__Nonnull] = ACTIONS(2860),
    [anon_sym_nullable] = ACTIONS(2860),
    [anon_sym__Complex] = ACTIONS(2860),
    [anon_sym__Nullable] = ACTIONS(2860),
    [anon_sym__Nullable_result] = ACTIONS(2860),
    [anon_sym__Null_unspecified] = ACTIONS(2860),
    [anon_sym___autoreleasing] = ACTIONS(2860),
    [anon_sym___block] = ACTIONS(2860),
    [anon_sym___bridge] = ACTIONS(2860),
    [anon_sym___bridge_retained] = ACTIONS(2860),
    [anon_sym___bridge_transfer] = ACTIONS(2860),
    [anon_sym___complex] = ACTIONS(2860),
    [anon_sym___const] = ACTIONS(2860),
    [anon_sym___imag] = ACTIONS(2860),
    [anon_sym___kindof] = ACTIONS(2860),
    [anon_sym___nonnull] = ACTIONS(2860),
    [anon_sym___nullable] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2860),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2860),
    [anon_sym___real] = ACTIONS(2860),
    [anon_sym___strong] = ACTIONS(2860),
    [anon_sym___unsafe_unretained] = ACTIONS(2860),
    [anon_sym___unused] = ACTIONS(2860),
    [anon_sym___weak] = ACTIONS(2860),
    [anon_sym_alignas] = ACTIONS(2860),
    [anon_sym__Alignas] = ACTIONS(2860),
    [sym_primitive_type] = ACTIONS(2860),
    [anon_sym_enum] = ACTIONS(2860),
    [anon_sym_struct] = ACTIONS(2860),
    [anon_sym_union] = ACTIONS(2860),
    [anon_sym_if] = ACTIONS(2860),
    [anon_sym_switch] = ACTIONS(2860),
    [anon_sym_case] = ACTIONS(2860),
    [anon_sym_default] = ACTIONS(2860),
    [anon_sym_while] = ACTIONS(2860),
    [anon_sym_do] = ACTIONS(2860),
    [anon_sym_for] = ACTIONS(2860),
    [anon_sym_in] = ACTIONS(2860),
    [anon_sym_return] = ACTIONS(2860),
    [anon_sym_break] = ACTIONS(2860),
    [anon_sym_continue] = ACTIONS(2860),
    [anon_sym_goto] = ACTIONS(2860),
    [anon_sym_DASH_DASH] = ACTIONS(2862),
    [anon_sym_PLUS_PLUS] = ACTIONS(2862),
    [anon_sym_sizeof] = ACTIONS(2860),
    [anon_sym___alignof__] = ACTIONS(2860),
    [anon_sym___alignof] = ACTIONS(2860),
    [anon_sym__alignof] = ACTIONS(2860),
    [anon_sym_alignof] = ACTIONS(2860),
    [anon_sym__Alignof] = ACTIONS(2860),
    [anon_sym_offsetof] = ACTIONS(2860),
    [anon_sym__Generic] = ACTIONS(2860),
    [anon_sym_asm] = ACTIONS(2860),
    [anon_sym___asm__] = ACTIONS(2860),
    [anon_sym___asm] = ACTIONS(2860),
    [sym_number_literal] = ACTIONS(2862),
    [anon_sym_L_SQUOTE] = ACTIONS(2862),
    [anon_sym_u_SQUOTE] = ACTIONS(2862),
    [anon_sym_U_SQUOTE] = ACTIONS(2862),
    [anon_sym_u8_SQUOTE] = ACTIONS(2862),
    [anon_sym_SQUOTE] = ACTIONS(2862),
    [anon_sym_AT] = ACTIONS(2860),
    [anon_sym_DQUOTE] = ACTIONS(2862),
    [anon_sym_L_DQUOTE] = ACTIONS(2862),
    [anon_sym_u_DQUOTE] = ACTIONS(2862),
    [anon_sym_U_DQUOTE] = ACTIONS(2862),
    [anon_sym_u8_DQUOTE] = ACTIONS(2862),
    [sym_true] = ACTIONS(2860),
    [sym_false] = ACTIONS(2860),
    [anon_sym_NULL] = ACTIONS(2860),
    [anon_sym_nullptr] = ACTIONS(2860),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2860),
    [anon_sym___typeof] = ACTIONS(2860),
    [anon_sym_typeof] = ACTIONS(2860),
    [anon_sym_ATimport] = ACTIONS(2862),
    [aux_sym_preproc_undef_token1] = ACTIONS(2860),
    [anon_sym_POUND] = ACTIONS(2860),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2860),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2860),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2860),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2860),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2860),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2860),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2860),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_API_AVAILABLE] = ACTIONS(2860),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_API_DEPRECATED] = ACTIONS(2860),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2860),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2860),
    [anon_sym___deprecated_msg] = ACTIONS(2860),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2860),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2860),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2860),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2860),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2860),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2860),
    [anon_sym_ATprotocol] = ACTIONS(2862),
    [anon_sym_ATinterface] = ACTIONS(2862),
    [anon_sym_ATimplementation] = ACTIONS(2862),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2862),
    [anon_sym_ATselector] = ACTIONS(2862),
    [anon_sym_ATavailable] = ACTIONS(2862),
    [anon_sym___builtin_available] = ACTIONS(2860),
    [anon_sym_va_arg] = ACTIONS(2860),
    [anon_sym_ATencode] = ACTIONS(2862),
    [anon_sym_BOOL] = ACTIONS(2860),
    [anon_sym_IMP] = ACTIONS(2860),
    [anon_sym_SEL] = ACTIONS(2860),
    [anon_sym_Class] = ACTIONS(2860),
    [anon_sym_id] = ACTIONS(2860),
  },
  [STATE(1016)] = {
    [ts_builtin_sym_end] = ACTIONS(2582),
    [sym_identifier] = ACTIONS(2580),
    [aux_sym_preproc_include_token1] = ACTIONS(2580),
    [aux_sym_preproc_include_token2] = ACTIONS(2580),
    [aux_sym_preproc_def_token1] = ACTIONS(2580),
    [aux_sym_preproc_if_token1] = ACTIONS(2580),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2580),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2580),
    [sym_preproc_directive] = ACTIONS(2580),
    [anon_sym_LPAREN2] = ACTIONS(2582),
    [anon_sym_BANG] = ACTIONS(2582),
    [anon_sym_TILDE] = ACTIONS(2582),
    [anon_sym_DASH] = ACTIONS(2580),
    [anon_sym_PLUS] = ACTIONS(2580),
    [anon_sym_STAR] = ACTIONS(2582),
    [anon_sym_CARET] = ACTIONS(2582),
    [anon_sym_AMP] = ACTIONS(2582),
    [anon_sym_SEMI] = ACTIONS(2582),
    [anon_sym___extension__] = ACTIONS(2580),
    [anon_sym_typedef] = ACTIONS(2580),
    [anon_sym_extern] = ACTIONS(2580),
    [anon_sym___attribute__] = ACTIONS(2580),
    [anon_sym___attribute] = ACTIONS(2580),
    [anon_sym_noreturn] = ACTIONS(2580),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym___declspec] = ACTIONS(2580),
    [anon_sym___cdecl] = ACTIONS(2580),
    [anon_sym___clrcall] = ACTIONS(2580),
    [anon_sym___stdcall] = ACTIONS(2580),
    [anon_sym___fastcall] = ACTIONS(2580),
    [anon_sym___thiscall] = ACTIONS(2580),
    [anon_sym___vectorcall] = ACTIONS(2580),
    [anon_sym_LBRACE] = ACTIONS(2582),
    [anon_sym_signed] = ACTIONS(2580),
    [anon_sym_unsigned] = ACTIONS(2580),
    [anon_sym_long] = ACTIONS(2580),
    [anon_sym_short] = ACTIONS(2580),
    [anon_sym_static] = ACTIONS(2580),
    [anon_sym_ATautoreleasepool] = ACTIONS(2582),
    [anon_sym_auto] = ACTIONS(2580),
    [anon_sym_register] = ACTIONS(2580),
    [anon_sym_inline] = ACTIONS(2580),
    [anon_sym___inline] = ACTIONS(2580),
    [anon_sym___inline__] = ACTIONS(2580),
    [anon_sym___forceinline] = ACTIONS(2580),
    [anon_sym_thread_local] = ACTIONS(2580),
    [anon_sym___thread] = ACTIONS(2580),
    [anon_sym_CG_EXTERN] = ACTIONS(2580),
    [anon_sym_CG_INLINE] = ACTIONS(2580),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2580),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2580),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2580),
    [anon_sym_IBOutlet] = ACTIONS(2580),
    [anon_sym_IBInspectable] = ACTIONS(2580),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2580),
    [anon_sym_NS_INLINE] = ACTIONS(2580),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2580),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2580),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2580),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2580),
    [anon_sym_const] = ACTIONS(2580),
    [anon_sym_constexpr] = ACTIONS(2580),
    [anon_sym_volatile] = ACTIONS(2580),
    [anon_sym_restrict] = ACTIONS(2580),
    [anon_sym___restrict__] = ACTIONS(2580),
    [anon_sym__Atomic] = ACTIONS(2580),
    [anon_sym__Noreturn] = ACTIONS(2580),
    [anon_sym__Nonnull] = ACTIONS(2580),
    [anon_sym_nullable] = ACTIONS(2580),
    [anon_sym__Complex] = ACTIONS(2580),
    [anon_sym__Nullable] = ACTIONS(2580),
    [anon_sym__Nullable_result] = ACTIONS(2580),
    [anon_sym__Null_unspecified] = ACTIONS(2580),
    [anon_sym___autoreleasing] = ACTIONS(2580),
    [anon_sym___block] = ACTIONS(2580),
    [anon_sym___bridge] = ACTIONS(2580),
    [anon_sym___bridge_retained] = ACTIONS(2580),
    [anon_sym___bridge_transfer] = ACTIONS(2580),
    [anon_sym___complex] = ACTIONS(2580),
    [anon_sym___const] = ACTIONS(2580),
    [anon_sym___imag] = ACTIONS(2580),
    [anon_sym___kindof] = ACTIONS(2580),
    [anon_sym___nonnull] = ACTIONS(2580),
    [anon_sym___nullable] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2580),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2580),
    [anon_sym___real] = ACTIONS(2580),
    [anon_sym___strong] = ACTIONS(2580),
    [anon_sym___unsafe_unretained] = ACTIONS(2580),
    [anon_sym___unused] = ACTIONS(2580),
    [anon_sym___weak] = ACTIONS(2580),
    [anon_sym_alignas] = ACTIONS(2580),
    [anon_sym__Alignas] = ACTIONS(2580),
    [sym_primitive_type] = ACTIONS(2580),
    [anon_sym_enum] = ACTIONS(2580),
    [anon_sym_struct] = ACTIONS(2580),
    [anon_sym_union] = ACTIONS(2580),
    [anon_sym_if] = ACTIONS(2580),
    [anon_sym_switch] = ACTIONS(2580),
    [anon_sym_case] = ACTIONS(2580),
    [anon_sym_default] = ACTIONS(2580),
    [anon_sym_while] = ACTIONS(2580),
    [anon_sym_do] = ACTIONS(2580),
    [anon_sym_for] = ACTIONS(2580),
    [anon_sym_in] = ACTIONS(2580),
    [anon_sym_return] = ACTIONS(2580),
    [anon_sym_break] = ACTIONS(2580),
    [anon_sym_continue] = ACTIONS(2580),
    [anon_sym_goto] = ACTIONS(2580),
    [anon_sym_DASH_DASH] = ACTIONS(2582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2582),
    [anon_sym_sizeof] = ACTIONS(2580),
    [anon_sym___alignof__] = ACTIONS(2580),
    [anon_sym___alignof] = ACTIONS(2580),
    [anon_sym__alignof] = ACTIONS(2580),
    [anon_sym_alignof] = ACTIONS(2580),
    [anon_sym__Alignof] = ACTIONS(2580),
    [anon_sym_offsetof] = ACTIONS(2580),
    [anon_sym__Generic] = ACTIONS(2580),
    [anon_sym_asm] = ACTIONS(2580),
    [anon_sym___asm__] = ACTIONS(2580),
    [anon_sym___asm] = ACTIONS(2580),
    [sym_number_literal] = ACTIONS(2582),
    [anon_sym_L_SQUOTE] = ACTIONS(2582),
    [anon_sym_u_SQUOTE] = ACTIONS(2582),
    [anon_sym_U_SQUOTE] = ACTIONS(2582),
    [anon_sym_u8_SQUOTE] = ACTIONS(2582),
    [anon_sym_SQUOTE] = ACTIONS(2582),
    [anon_sym_AT] = ACTIONS(2580),
    [anon_sym_DQUOTE] = ACTIONS(2582),
    [anon_sym_L_DQUOTE] = ACTIONS(2582),
    [anon_sym_u_DQUOTE] = ACTIONS(2582),
    [anon_sym_U_DQUOTE] = ACTIONS(2582),
    [anon_sym_u8_DQUOTE] = ACTIONS(2582),
    [sym_true] = ACTIONS(2580),
    [sym_false] = ACTIONS(2580),
    [anon_sym_NULL] = ACTIONS(2580),
    [anon_sym_nullptr] = ACTIONS(2580),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2580),
    [anon_sym___typeof] = ACTIONS(2580),
    [anon_sym_typeof] = ACTIONS(2580),
    [anon_sym_ATimport] = ACTIONS(2582),
    [aux_sym_preproc_undef_token1] = ACTIONS(2580),
    [anon_sym_POUND] = ACTIONS(2580),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2580),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2580),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2580),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2580),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2580),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2580),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2580),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_API_AVAILABLE] = ACTIONS(2580),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_API_DEPRECATED] = ACTIONS(2580),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2580),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2580),
    [anon_sym___deprecated_msg] = ACTIONS(2580),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2580),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2580),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2580),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2580),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2580),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2580),
    [anon_sym_ATprotocol] = ACTIONS(2582),
    [anon_sym_ATinterface] = ACTIONS(2582),
    [anon_sym_ATimplementation] = ACTIONS(2582),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2582),
    [anon_sym_ATselector] = ACTIONS(2582),
    [anon_sym_ATavailable] = ACTIONS(2582),
    [anon_sym___builtin_available] = ACTIONS(2580),
    [anon_sym_va_arg] = ACTIONS(2580),
    [anon_sym_ATencode] = ACTIONS(2582),
    [anon_sym_BOOL] = ACTIONS(2580),
    [anon_sym_IMP] = ACTIONS(2580),
    [anon_sym_SEL] = ACTIONS(2580),
    [anon_sym_Class] = ACTIONS(2580),
    [anon_sym_id] = ACTIONS(2580),
  },
  [STATE(1017)] = {
    [ts_builtin_sym_end] = ACTIONS(2560),
    [sym_identifier] = ACTIONS(2558),
    [aux_sym_preproc_include_token1] = ACTIONS(2558),
    [aux_sym_preproc_include_token2] = ACTIONS(2558),
    [aux_sym_preproc_def_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2558),
    [sym_preproc_directive] = ACTIONS(2558),
    [anon_sym_LPAREN2] = ACTIONS(2560),
    [anon_sym_BANG] = ACTIONS(2560),
    [anon_sym_TILDE] = ACTIONS(2560),
    [anon_sym_DASH] = ACTIONS(2558),
    [anon_sym_PLUS] = ACTIONS(2558),
    [anon_sym_STAR] = ACTIONS(2560),
    [anon_sym_CARET] = ACTIONS(2560),
    [anon_sym_AMP] = ACTIONS(2560),
    [anon_sym_SEMI] = ACTIONS(2560),
    [anon_sym___extension__] = ACTIONS(2558),
    [anon_sym_typedef] = ACTIONS(2558),
    [anon_sym_extern] = ACTIONS(2558),
    [anon_sym___attribute__] = ACTIONS(2558),
    [anon_sym___attribute] = ACTIONS(2558),
    [anon_sym_noreturn] = ACTIONS(2558),
    [anon_sym_LBRACK] = ACTIONS(2560),
    [anon_sym___declspec] = ACTIONS(2558),
    [anon_sym___cdecl] = ACTIONS(2558),
    [anon_sym___clrcall] = ACTIONS(2558),
    [anon_sym___stdcall] = ACTIONS(2558),
    [anon_sym___fastcall] = ACTIONS(2558),
    [anon_sym___thiscall] = ACTIONS(2558),
    [anon_sym___vectorcall] = ACTIONS(2558),
    [anon_sym_LBRACE] = ACTIONS(2560),
    [anon_sym_signed] = ACTIONS(2558),
    [anon_sym_unsigned] = ACTIONS(2558),
    [anon_sym_long] = ACTIONS(2558),
    [anon_sym_short] = ACTIONS(2558),
    [anon_sym_static] = ACTIONS(2558),
    [anon_sym_ATautoreleasepool] = ACTIONS(2560),
    [anon_sym_auto] = ACTIONS(2558),
    [anon_sym_register] = ACTIONS(2558),
    [anon_sym_inline] = ACTIONS(2558),
    [anon_sym___inline] = ACTIONS(2558),
    [anon_sym___inline__] = ACTIONS(2558),
    [anon_sym___forceinline] = ACTIONS(2558),
    [anon_sym_thread_local] = ACTIONS(2558),
    [anon_sym___thread] = ACTIONS(2558),
    [anon_sym_CG_EXTERN] = ACTIONS(2558),
    [anon_sym_CG_INLINE] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2558),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2558),
    [anon_sym_IBOutlet] = ACTIONS(2558),
    [anon_sym_IBInspectable] = ACTIONS(2558),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2558),
    [anon_sym_NS_INLINE] = ACTIONS(2558),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2558),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2558),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2558),
    [anon_sym_const] = ACTIONS(2558),
    [anon_sym_constexpr] = ACTIONS(2558),
    [anon_sym_volatile] = ACTIONS(2558),
    [anon_sym_restrict] = ACTIONS(2558),
    [anon_sym___restrict__] = ACTIONS(2558),
    [anon_sym__Atomic] = ACTIONS(2558),
    [anon_sym__Noreturn] = ACTIONS(2558),
    [anon_sym__Nonnull] = ACTIONS(2558),
    [anon_sym_nullable] = ACTIONS(2558),
    [anon_sym__Complex] = ACTIONS(2558),
    [anon_sym__Nullable] = ACTIONS(2558),
    [anon_sym__Nullable_result] = ACTIONS(2558),
    [anon_sym__Null_unspecified] = ACTIONS(2558),
    [anon_sym___autoreleasing] = ACTIONS(2558),
    [anon_sym___block] = ACTIONS(2558),
    [anon_sym___bridge] = ACTIONS(2558),
    [anon_sym___bridge_retained] = ACTIONS(2558),
    [anon_sym___bridge_transfer] = ACTIONS(2558),
    [anon_sym___complex] = ACTIONS(2558),
    [anon_sym___const] = ACTIONS(2558),
    [anon_sym___imag] = ACTIONS(2558),
    [anon_sym___kindof] = ACTIONS(2558),
    [anon_sym___nonnull] = ACTIONS(2558),
    [anon_sym___nullable] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2558),
    [anon_sym___real] = ACTIONS(2558),
    [anon_sym___strong] = ACTIONS(2558),
    [anon_sym___unsafe_unretained] = ACTIONS(2558),
    [anon_sym___unused] = ACTIONS(2558),
    [anon_sym___weak] = ACTIONS(2558),
    [anon_sym_alignas] = ACTIONS(2558),
    [anon_sym__Alignas] = ACTIONS(2558),
    [sym_primitive_type] = ACTIONS(2558),
    [anon_sym_enum] = ACTIONS(2558),
    [anon_sym_struct] = ACTIONS(2558),
    [anon_sym_union] = ACTIONS(2558),
    [anon_sym_if] = ACTIONS(2558),
    [anon_sym_switch] = ACTIONS(2558),
    [anon_sym_case] = ACTIONS(2558),
    [anon_sym_default] = ACTIONS(2558),
    [anon_sym_while] = ACTIONS(2558),
    [anon_sym_do] = ACTIONS(2558),
    [anon_sym_for] = ACTIONS(2558),
    [anon_sym_in] = ACTIONS(2558),
    [anon_sym_return] = ACTIONS(2558),
    [anon_sym_break] = ACTIONS(2558),
    [anon_sym_continue] = ACTIONS(2558),
    [anon_sym_goto] = ACTIONS(2558),
    [anon_sym_DASH_DASH] = ACTIONS(2560),
    [anon_sym_PLUS_PLUS] = ACTIONS(2560),
    [anon_sym_sizeof] = ACTIONS(2558),
    [anon_sym___alignof__] = ACTIONS(2558),
    [anon_sym___alignof] = ACTIONS(2558),
    [anon_sym__alignof] = ACTIONS(2558),
    [anon_sym_alignof] = ACTIONS(2558),
    [anon_sym__Alignof] = ACTIONS(2558),
    [anon_sym_offsetof] = ACTIONS(2558),
    [anon_sym__Generic] = ACTIONS(2558),
    [anon_sym_asm] = ACTIONS(2558),
    [anon_sym___asm__] = ACTIONS(2558),
    [anon_sym___asm] = ACTIONS(2558),
    [sym_number_literal] = ACTIONS(2560),
    [anon_sym_L_SQUOTE] = ACTIONS(2560),
    [anon_sym_u_SQUOTE] = ACTIONS(2560),
    [anon_sym_U_SQUOTE] = ACTIONS(2560),
    [anon_sym_u8_SQUOTE] = ACTIONS(2560),
    [anon_sym_SQUOTE] = ACTIONS(2560),
    [anon_sym_AT] = ACTIONS(2558),
    [anon_sym_DQUOTE] = ACTIONS(2560),
    [anon_sym_L_DQUOTE] = ACTIONS(2560),
    [anon_sym_u_DQUOTE] = ACTIONS(2560),
    [anon_sym_U_DQUOTE] = ACTIONS(2560),
    [anon_sym_u8_DQUOTE] = ACTIONS(2560),
    [sym_true] = ACTIONS(2558),
    [sym_false] = ACTIONS(2558),
    [anon_sym_NULL] = ACTIONS(2558),
    [anon_sym_nullptr] = ACTIONS(2558),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2558),
    [anon_sym___typeof] = ACTIONS(2558),
    [anon_sym_typeof] = ACTIONS(2558),
    [anon_sym_ATimport] = ACTIONS(2560),
    [aux_sym_preproc_undef_token1] = ACTIONS(2558),
    [anon_sym_POUND] = ACTIONS(2558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2558),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2558),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2558),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_API_AVAILABLE] = ACTIONS(2558),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_API_DEPRECATED] = ACTIONS(2558),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym___deprecated_msg] = ACTIONS(2558),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2558),
    [anon_sym_ATprotocol] = ACTIONS(2560),
    [anon_sym_ATinterface] = ACTIONS(2560),
    [anon_sym_ATimplementation] = ACTIONS(2560),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2560),
    [anon_sym_ATselector] = ACTIONS(2560),
    [anon_sym_ATavailable] = ACTIONS(2560),
    [anon_sym___builtin_available] = ACTIONS(2558),
    [anon_sym_va_arg] = ACTIONS(2558),
    [anon_sym_ATencode] = ACTIONS(2560),
    [anon_sym_BOOL] = ACTIONS(2558),
    [anon_sym_IMP] = ACTIONS(2558),
    [anon_sym_SEL] = ACTIONS(2558),
    [anon_sym_Class] = ACTIONS(2558),
    [anon_sym_id] = ACTIONS(2558),
  },
  [STATE(1018)] = {
    [ts_builtin_sym_end] = ACTIONS(2678),
    [sym_identifier] = ACTIONS(2676),
    [aux_sym_preproc_include_token1] = ACTIONS(2676),
    [aux_sym_preproc_include_token2] = ACTIONS(2676),
    [aux_sym_preproc_def_token1] = ACTIONS(2676),
    [aux_sym_preproc_if_token1] = ACTIONS(2676),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2676),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2676),
    [sym_preproc_directive] = ACTIONS(2676),
    [anon_sym_LPAREN2] = ACTIONS(2678),
    [anon_sym_BANG] = ACTIONS(2678),
    [anon_sym_TILDE] = ACTIONS(2678),
    [anon_sym_DASH] = ACTIONS(2676),
    [anon_sym_PLUS] = ACTIONS(2676),
    [anon_sym_STAR] = ACTIONS(2678),
    [anon_sym_CARET] = ACTIONS(2678),
    [anon_sym_AMP] = ACTIONS(2678),
    [anon_sym_SEMI] = ACTIONS(2678),
    [anon_sym___extension__] = ACTIONS(2676),
    [anon_sym_typedef] = ACTIONS(2676),
    [anon_sym_extern] = ACTIONS(2676),
    [anon_sym___attribute__] = ACTIONS(2676),
    [anon_sym___attribute] = ACTIONS(2676),
    [anon_sym_noreturn] = ACTIONS(2676),
    [anon_sym_LBRACK] = ACTIONS(2678),
    [anon_sym___declspec] = ACTIONS(2676),
    [anon_sym___cdecl] = ACTIONS(2676),
    [anon_sym___clrcall] = ACTIONS(2676),
    [anon_sym___stdcall] = ACTIONS(2676),
    [anon_sym___fastcall] = ACTIONS(2676),
    [anon_sym___thiscall] = ACTIONS(2676),
    [anon_sym___vectorcall] = ACTIONS(2676),
    [anon_sym_LBRACE] = ACTIONS(2678),
    [anon_sym_signed] = ACTIONS(2676),
    [anon_sym_unsigned] = ACTIONS(2676),
    [anon_sym_long] = ACTIONS(2676),
    [anon_sym_short] = ACTIONS(2676),
    [anon_sym_static] = ACTIONS(2676),
    [anon_sym_ATautoreleasepool] = ACTIONS(2678),
    [anon_sym_auto] = ACTIONS(2676),
    [anon_sym_register] = ACTIONS(2676),
    [anon_sym_inline] = ACTIONS(2676),
    [anon_sym___inline] = ACTIONS(2676),
    [anon_sym___inline__] = ACTIONS(2676),
    [anon_sym___forceinline] = ACTIONS(2676),
    [anon_sym_thread_local] = ACTIONS(2676),
    [anon_sym___thread] = ACTIONS(2676),
    [anon_sym_CG_EXTERN] = ACTIONS(2676),
    [anon_sym_CG_INLINE] = ACTIONS(2676),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2676),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2676),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2676),
    [anon_sym_IBOutlet] = ACTIONS(2676),
    [anon_sym_IBInspectable] = ACTIONS(2676),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2676),
    [anon_sym_NS_INLINE] = ACTIONS(2676),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2676),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2676),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2676),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2676),
    [anon_sym_const] = ACTIONS(2676),
    [anon_sym_constexpr] = ACTIONS(2676),
    [anon_sym_volatile] = ACTIONS(2676),
    [anon_sym_restrict] = ACTIONS(2676),
    [anon_sym___restrict__] = ACTIONS(2676),
    [anon_sym__Atomic] = ACTIONS(2676),
    [anon_sym__Noreturn] = ACTIONS(2676),
    [anon_sym__Nonnull] = ACTIONS(2676),
    [anon_sym_nullable] = ACTIONS(2676),
    [anon_sym__Complex] = ACTIONS(2676),
    [anon_sym__Nullable] = ACTIONS(2676),
    [anon_sym__Nullable_result] = ACTIONS(2676),
    [anon_sym__Null_unspecified] = ACTIONS(2676),
    [anon_sym___autoreleasing] = ACTIONS(2676),
    [anon_sym___block] = ACTIONS(2676),
    [anon_sym___bridge] = ACTIONS(2676),
    [anon_sym___bridge_retained] = ACTIONS(2676),
    [anon_sym___bridge_transfer] = ACTIONS(2676),
    [anon_sym___complex] = ACTIONS(2676),
    [anon_sym___const] = ACTIONS(2676),
    [anon_sym___imag] = ACTIONS(2676),
    [anon_sym___kindof] = ACTIONS(2676),
    [anon_sym___nonnull] = ACTIONS(2676),
    [anon_sym___nullable] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2676),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2676),
    [anon_sym___real] = ACTIONS(2676),
    [anon_sym___strong] = ACTIONS(2676),
    [anon_sym___unsafe_unretained] = ACTIONS(2676),
    [anon_sym___unused] = ACTIONS(2676),
    [anon_sym___weak] = ACTIONS(2676),
    [anon_sym_alignas] = ACTIONS(2676),
    [anon_sym__Alignas] = ACTIONS(2676),
    [sym_primitive_type] = ACTIONS(2676),
    [anon_sym_enum] = ACTIONS(2676),
    [anon_sym_struct] = ACTIONS(2676),
    [anon_sym_union] = ACTIONS(2676),
    [anon_sym_if] = ACTIONS(2676),
    [anon_sym_switch] = ACTIONS(2676),
    [anon_sym_case] = ACTIONS(2676),
    [anon_sym_default] = ACTIONS(2676),
    [anon_sym_while] = ACTIONS(2676),
    [anon_sym_do] = ACTIONS(2676),
    [anon_sym_for] = ACTIONS(2676),
    [anon_sym_in] = ACTIONS(2676),
    [anon_sym_return] = ACTIONS(2676),
    [anon_sym_break] = ACTIONS(2676),
    [anon_sym_continue] = ACTIONS(2676),
    [anon_sym_goto] = ACTIONS(2676),
    [anon_sym_DASH_DASH] = ACTIONS(2678),
    [anon_sym_PLUS_PLUS] = ACTIONS(2678),
    [anon_sym_sizeof] = ACTIONS(2676),
    [anon_sym___alignof__] = ACTIONS(2676),
    [anon_sym___alignof] = ACTIONS(2676),
    [anon_sym__alignof] = ACTIONS(2676),
    [anon_sym_alignof] = ACTIONS(2676),
    [anon_sym__Alignof] = ACTIONS(2676),
    [anon_sym_offsetof] = ACTIONS(2676),
    [anon_sym__Generic] = ACTIONS(2676),
    [anon_sym_asm] = ACTIONS(2676),
    [anon_sym___asm__] = ACTIONS(2676),
    [anon_sym___asm] = ACTIONS(2676),
    [sym_number_literal] = ACTIONS(2678),
    [anon_sym_L_SQUOTE] = ACTIONS(2678),
    [anon_sym_u_SQUOTE] = ACTIONS(2678),
    [anon_sym_U_SQUOTE] = ACTIONS(2678),
    [anon_sym_u8_SQUOTE] = ACTIONS(2678),
    [anon_sym_SQUOTE] = ACTIONS(2678),
    [anon_sym_AT] = ACTIONS(2676),
    [anon_sym_DQUOTE] = ACTIONS(2678),
    [anon_sym_L_DQUOTE] = ACTIONS(2678),
    [anon_sym_u_DQUOTE] = ACTIONS(2678),
    [anon_sym_U_DQUOTE] = ACTIONS(2678),
    [anon_sym_u8_DQUOTE] = ACTIONS(2678),
    [sym_true] = ACTIONS(2676),
    [sym_false] = ACTIONS(2676),
    [anon_sym_NULL] = ACTIONS(2676),
    [anon_sym_nullptr] = ACTIONS(2676),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2676),
    [anon_sym___typeof] = ACTIONS(2676),
    [anon_sym_typeof] = ACTIONS(2676),
    [anon_sym_ATimport] = ACTIONS(2678),
    [aux_sym_preproc_undef_token1] = ACTIONS(2676),
    [anon_sym_POUND] = ACTIONS(2676),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2676),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2676),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2676),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2676),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2676),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2676),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2676),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_API_AVAILABLE] = ACTIONS(2676),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_API_DEPRECATED] = ACTIONS(2676),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2676),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2676),
    [anon_sym___deprecated_msg] = ACTIONS(2676),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2676),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2676),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2676),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2676),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2676),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2676),
    [anon_sym_ATprotocol] = ACTIONS(2678),
    [anon_sym_ATinterface] = ACTIONS(2678),
    [anon_sym_ATimplementation] = ACTIONS(2678),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2678),
    [anon_sym_ATselector] = ACTIONS(2678),
    [anon_sym_ATavailable] = ACTIONS(2678),
    [anon_sym___builtin_available] = ACTIONS(2676),
    [anon_sym_va_arg] = ACTIONS(2676),
    [anon_sym_ATencode] = ACTIONS(2678),
    [anon_sym_BOOL] = ACTIONS(2676),
    [anon_sym_IMP] = ACTIONS(2676),
    [anon_sym_SEL] = ACTIONS(2676),
    [anon_sym_Class] = ACTIONS(2676),
    [anon_sym_id] = ACTIONS(2676),
  },
  [STATE(1019)] = {
    [ts_builtin_sym_end] = ACTIONS(2682),
    [sym_identifier] = ACTIONS(2680),
    [aux_sym_preproc_include_token1] = ACTIONS(2680),
    [aux_sym_preproc_include_token2] = ACTIONS(2680),
    [aux_sym_preproc_def_token1] = ACTIONS(2680),
    [aux_sym_preproc_if_token1] = ACTIONS(2680),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2680),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2680),
    [sym_preproc_directive] = ACTIONS(2680),
    [anon_sym_LPAREN2] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2680),
    [anon_sym_PLUS] = ACTIONS(2680),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym___extension__] = ACTIONS(2680),
    [anon_sym_typedef] = ACTIONS(2680),
    [anon_sym_extern] = ACTIONS(2680),
    [anon_sym___attribute__] = ACTIONS(2680),
    [anon_sym___attribute] = ACTIONS(2680),
    [anon_sym_noreturn] = ACTIONS(2680),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym___declspec] = ACTIONS(2680),
    [anon_sym___cdecl] = ACTIONS(2680),
    [anon_sym___clrcall] = ACTIONS(2680),
    [anon_sym___stdcall] = ACTIONS(2680),
    [anon_sym___fastcall] = ACTIONS(2680),
    [anon_sym___thiscall] = ACTIONS(2680),
    [anon_sym___vectorcall] = ACTIONS(2680),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_signed] = ACTIONS(2680),
    [anon_sym_unsigned] = ACTIONS(2680),
    [anon_sym_long] = ACTIONS(2680),
    [anon_sym_short] = ACTIONS(2680),
    [anon_sym_static] = ACTIONS(2680),
    [anon_sym_ATautoreleasepool] = ACTIONS(2682),
    [anon_sym_auto] = ACTIONS(2680),
    [anon_sym_register] = ACTIONS(2680),
    [anon_sym_inline] = ACTIONS(2680),
    [anon_sym___inline] = ACTIONS(2680),
    [anon_sym___inline__] = ACTIONS(2680),
    [anon_sym___forceinline] = ACTIONS(2680),
    [anon_sym_thread_local] = ACTIONS(2680),
    [anon_sym___thread] = ACTIONS(2680),
    [anon_sym_CG_EXTERN] = ACTIONS(2680),
    [anon_sym_CG_INLINE] = ACTIONS(2680),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2680),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2680),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2680),
    [anon_sym_IBOutlet] = ACTIONS(2680),
    [anon_sym_IBInspectable] = ACTIONS(2680),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2680),
    [anon_sym_NS_INLINE] = ACTIONS(2680),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2680),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2680),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2680),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2680),
    [anon_sym_const] = ACTIONS(2680),
    [anon_sym_constexpr] = ACTIONS(2680),
    [anon_sym_volatile] = ACTIONS(2680),
    [anon_sym_restrict] = ACTIONS(2680),
    [anon_sym___restrict__] = ACTIONS(2680),
    [anon_sym__Atomic] = ACTIONS(2680),
    [anon_sym__Noreturn] = ACTIONS(2680),
    [anon_sym__Nonnull] = ACTIONS(2680),
    [anon_sym_nullable] = ACTIONS(2680),
    [anon_sym__Complex] = ACTIONS(2680),
    [anon_sym__Nullable] = ACTIONS(2680),
    [anon_sym__Nullable_result] = ACTIONS(2680),
    [anon_sym__Null_unspecified] = ACTIONS(2680),
    [anon_sym___autoreleasing] = ACTIONS(2680),
    [anon_sym___block] = ACTIONS(2680),
    [anon_sym___bridge] = ACTIONS(2680),
    [anon_sym___bridge_retained] = ACTIONS(2680),
    [anon_sym___bridge_transfer] = ACTIONS(2680),
    [anon_sym___complex] = ACTIONS(2680),
    [anon_sym___const] = ACTIONS(2680),
    [anon_sym___imag] = ACTIONS(2680),
    [anon_sym___kindof] = ACTIONS(2680),
    [anon_sym___nonnull] = ACTIONS(2680),
    [anon_sym___nullable] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2680),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2680),
    [anon_sym___real] = ACTIONS(2680),
    [anon_sym___strong] = ACTIONS(2680),
    [anon_sym___unsafe_unretained] = ACTIONS(2680),
    [anon_sym___unused] = ACTIONS(2680),
    [anon_sym___weak] = ACTIONS(2680),
    [anon_sym_alignas] = ACTIONS(2680),
    [anon_sym__Alignas] = ACTIONS(2680),
    [sym_primitive_type] = ACTIONS(2680),
    [anon_sym_enum] = ACTIONS(2680),
    [anon_sym_struct] = ACTIONS(2680),
    [anon_sym_union] = ACTIONS(2680),
    [anon_sym_if] = ACTIONS(2680),
    [anon_sym_switch] = ACTIONS(2680),
    [anon_sym_case] = ACTIONS(2680),
    [anon_sym_default] = ACTIONS(2680),
    [anon_sym_while] = ACTIONS(2680),
    [anon_sym_do] = ACTIONS(2680),
    [anon_sym_for] = ACTIONS(2680),
    [anon_sym_in] = ACTIONS(2680),
    [anon_sym_return] = ACTIONS(2680),
    [anon_sym_break] = ACTIONS(2680),
    [anon_sym_continue] = ACTIONS(2680),
    [anon_sym_goto] = ACTIONS(2680),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_sizeof] = ACTIONS(2680),
    [anon_sym___alignof__] = ACTIONS(2680),
    [anon_sym___alignof] = ACTIONS(2680),
    [anon_sym__alignof] = ACTIONS(2680),
    [anon_sym_alignof] = ACTIONS(2680),
    [anon_sym__Alignof] = ACTIONS(2680),
    [anon_sym_offsetof] = ACTIONS(2680),
    [anon_sym__Generic] = ACTIONS(2680),
    [anon_sym_asm] = ACTIONS(2680),
    [anon_sym___asm__] = ACTIONS(2680),
    [anon_sym___asm] = ACTIONS(2680),
    [sym_number_literal] = ACTIONS(2682),
    [anon_sym_L_SQUOTE] = ACTIONS(2682),
    [anon_sym_u_SQUOTE] = ACTIONS(2682),
    [anon_sym_U_SQUOTE] = ACTIONS(2682),
    [anon_sym_u8_SQUOTE] = ACTIONS(2682),
    [anon_sym_SQUOTE] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2680),
    [anon_sym_DQUOTE] = ACTIONS(2682),
    [anon_sym_L_DQUOTE] = ACTIONS(2682),
    [anon_sym_u_DQUOTE] = ACTIONS(2682),
    [anon_sym_U_DQUOTE] = ACTIONS(2682),
    [anon_sym_u8_DQUOTE] = ACTIONS(2682),
    [sym_true] = ACTIONS(2680),
    [sym_false] = ACTIONS(2680),
    [anon_sym_NULL] = ACTIONS(2680),
    [anon_sym_nullptr] = ACTIONS(2680),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2680),
    [anon_sym___typeof] = ACTIONS(2680),
    [anon_sym_typeof] = ACTIONS(2680),
    [anon_sym_ATimport] = ACTIONS(2682),
    [aux_sym_preproc_undef_token1] = ACTIONS(2680),
    [anon_sym_POUND] = ACTIONS(2680),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2680),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2680),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2680),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2680),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2680),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2680),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2680),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_API_AVAILABLE] = ACTIONS(2680),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_API_DEPRECATED] = ACTIONS(2680),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2680),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2680),
    [anon_sym___deprecated_msg] = ACTIONS(2680),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2680),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2680),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2680),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2680),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2680),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2680),
    [anon_sym_ATprotocol] = ACTIONS(2682),
    [anon_sym_ATinterface] = ACTIONS(2682),
    [anon_sym_ATimplementation] = ACTIONS(2682),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2682),
    [anon_sym_ATselector] = ACTIONS(2682),
    [anon_sym_ATavailable] = ACTIONS(2682),
    [anon_sym___builtin_available] = ACTIONS(2680),
    [anon_sym_va_arg] = ACTIONS(2680),
    [anon_sym_ATencode] = ACTIONS(2682),
    [anon_sym_BOOL] = ACTIONS(2680),
    [anon_sym_IMP] = ACTIONS(2680),
    [anon_sym_SEL] = ACTIONS(2680),
    [anon_sym_Class] = ACTIONS(2680),
    [anon_sym_id] = ACTIONS(2680),
  },
  [STATE(1020)] = {
    [ts_builtin_sym_end] = ACTIONS(2686),
    [sym_identifier] = ACTIONS(2684),
    [aux_sym_preproc_include_token1] = ACTIONS(2684),
    [aux_sym_preproc_include_token2] = ACTIONS(2684),
    [aux_sym_preproc_def_token1] = ACTIONS(2684),
    [aux_sym_preproc_if_token1] = ACTIONS(2684),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2684),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2684),
    [sym_preproc_directive] = ACTIONS(2684),
    [anon_sym_LPAREN2] = ACTIONS(2686),
    [anon_sym_BANG] = ACTIONS(2686),
    [anon_sym_TILDE] = ACTIONS(2686),
    [anon_sym_DASH] = ACTIONS(2684),
    [anon_sym_PLUS] = ACTIONS(2684),
    [anon_sym_STAR] = ACTIONS(2686),
    [anon_sym_CARET] = ACTIONS(2686),
    [anon_sym_AMP] = ACTIONS(2686),
    [anon_sym_SEMI] = ACTIONS(2686),
    [anon_sym___extension__] = ACTIONS(2684),
    [anon_sym_typedef] = ACTIONS(2684),
    [anon_sym_extern] = ACTIONS(2684),
    [anon_sym___attribute__] = ACTIONS(2684),
    [anon_sym___attribute] = ACTIONS(2684),
    [anon_sym_noreturn] = ACTIONS(2684),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym___declspec] = ACTIONS(2684),
    [anon_sym___cdecl] = ACTIONS(2684),
    [anon_sym___clrcall] = ACTIONS(2684),
    [anon_sym___stdcall] = ACTIONS(2684),
    [anon_sym___fastcall] = ACTIONS(2684),
    [anon_sym___thiscall] = ACTIONS(2684),
    [anon_sym___vectorcall] = ACTIONS(2684),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_signed] = ACTIONS(2684),
    [anon_sym_unsigned] = ACTIONS(2684),
    [anon_sym_long] = ACTIONS(2684),
    [anon_sym_short] = ACTIONS(2684),
    [anon_sym_static] = ACTIONS(2684),
    [anon_sym_ATautoreleasepool] = ACTIONS(2686),
    [anon_sym_auto] = ACTIONS(2684),
    [anon_sym_register] = ACTIONS(2684),
    [anon_sym_inline] = ACTIONS(2684),
    [anon_sym___inline] = ACTIONS(2684),
    [anon_sym___inline__] = ACTIONS(2684),
    [anon_sym___forceinline] = ACTIONS(2684),
    [anon_sym_thread_local] = ACTIONS(2684),
    [anon_sym___thread] = ACTIONS(2684),
    [anon_sym_CG_EXTERN] = ACTIONS(2684),
    [anon_sym_CG_INLINE] = ACTIONS(2684),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2684),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2684),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2684),
    [anon_sym_IBOutlet] = ACTIONS(2684),
    [anon_sym_IBInspectable] = ACTIONS(2684),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2684),
    [anon_sym_NS_INLINE] = ACTIONS(2684),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2684),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2684),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2684),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2684),
    [anon_sym_const] = ACTIONS(2684),
    [anon_sym_constexpr] = ACTIONS(2684),
    [anon_sym_volatile] = ACTIONS(2684),
    [anon_sym_restrict] = ACTIONS(2684),
    [anon_sym___restrict__] = ACTIONS(2684),
    [anon_sym__Atomic] = ACTIONS(2684),
    [anon_sym__Noreturn] = ACTIONS(2684),
    [anon_sym__Nonnull] = ACTIONS(2684),
    [anon_sym_nullable] = ACTIONS(2684),
    [anon_sym__Complex] = ACTIONS(2684),
    [anon_sym__Nullable] = ACTIONS(2684),
    [anon_sym__Nullable_result] = ACTIONS(2684),
    [anon_sym__Null_unspecified] = ACTIONS(2684),
    [anon_sym___autoreleasing] = ACTIONS(2684),
    [anon_sym___block] = ACTIONS(2684),
    [anon_sym___bridge] = ACTIONS(2684),
    [anon_sym___bridge_retained] = ACTIONS(2684),
    [anon_sym___bridge_transfer] = ACTIONS(2684),
    [anon_sym___complex] = ACTIONS(2684),
    [anon_sym___const] = ACTIONS(2684),
    [anon_sym___imag] = ACTIONS(2684),
    [anon_sym___kindof] = ACTIONS(2684),
    [anon_sym___nonnull] = ACTIONS(2684),
    [anon_sym___nullable] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2684),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2684),
    [anon_sym___real] = ACTIONS(2684),
    [anon_sym___strong] = ACTIONS(2684),
    [anon_sym___unsafe_unretained] = ACTIONS(2684),
    [anon_sym___unused] = ACTIONS(2684),
    [anon_sym___weak] = ACTIONS(2684),
    [anon_sym_alignas] = ACTIONS(2684),
    [anon_sym__Alignas] = ACTIONS(2684),
    [sym_primitive_type] = ACTIONS(2684),
    [anon_sym_enum] = ACTIONS(2684),
    [anon_sym_struct] = ACTIONS(2684),
    [anon_sym_union] = ACTIONS(2684),
    [anon_sym_if] = ACTIONS(2684),
    [anon_sym_switch] = ACTIONS(2684),
    [anon_sym_case] = ACTIONS(2684),
    [anon_sym_default] = ACTIONS(2684),
    [anon_sym_while] = ACTIONS(2684),
    [anon_sym_do] = ACTIONS(2684),
    [anon_sym_for] = ACTIONS(2684),
    [anon_sym_in] = ACTIONS(2684),
    [anon_sym_return] = ACTIONS(2684),
    [anon_sym_break] = ACTIONS(2684),
    [anon_sym_continue] = ACTIONS(2684),
    [anon_sym_goto] = ACTIONS(2684),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_sizeof] = ACTIONS(2684),
    [anon_sym___alignof__] = ACTIONS(2684),
    [anon_sym___alignof] = ACTIONS(2684),
    [anon_sym__alignof] = ACTIONS(2684),
    [anon_sym_alignof] = ACTIONS(2684),
    [anon_sym__Alignof] = ACTIONS(2684),
    [anon_sym_offsetof] = ACTIONS(2684),
    [anon_sym__Generic] = ACTIONS(2684),
    [anon_sym_asm] = ACTIONS(2684),
    [anon_sym___asm__] = ACTIONS(2684),
    [anon_sym___asm] = ACTIONS(2684),
    [sym_number_literal] = ACTIONS(2686),
    [anon_sym_L_SQUOTE] = ACTIONS(2686),
    [anon_sym_u_SQUOTE] = ACTIONS(2686),
    [anon_sym_U_SQUOTE] = ACTIONS(2686),
    [anon_sym_u8_SQUOTE] = ACTIONS(2686),
    [anon_sym_SQUOTE] = ACTIONS(2686),
    [anon_sym_AT] = ACTIONS(2684),
    [anon_sym_DQUOTE] = ACTIONS(2686),
    [anon_sym_L_DQUOTE] = ACTIONS(2686),
    [anon_sym_u_DQUOTE] = ACTIONS(2686),
    [anon_sym_U_DQUOTE] = ACTIONS(2686),
    [anon_sym_u8_DQUOTE] = ACTIONS(2686),
    [sym_true] = ACTIONS(2684),
    [sym_false] = ACTIONS(2684),
    [anon_sym_NULL] = ACTIONS(2684),
    [anon_sym_nullptr] = ACTIONS(2684),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2684),
    [anon_sym___typeof] = ACTIONS(2684),
    [anon_sym_typeof] = ACTIONS(2684),
    [anon_sym_ATimport] = ACTIONS(2686),
    [aux_sym_preproc_undef_token1] = ACTIONS(2684),
    [anon_sym_POUND] = ACTIONS(2684),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2684),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2684),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2684),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2684),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2684),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2684),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2684),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_API_AVAILABLE] = ACTIONS(2684),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_API_DEPRECATED] = ACTIONS(2684),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2684),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2684),
    [anon_sym___deprecated_msg] = ACTIONS(2684),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2684),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2684),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2684),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2684),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2684),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2684),
    [anon_sym_ATprotocol] = ACTIONS(2686),
    [anon_sym_ATinterface] = ACTIONS(2686),
    [anon_sym_ATimplementation] = ACTIONS(2686),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2686),
    [anon_sym_ATselector] = ACTIONS(2686),
    [anon_sym_ATavailable] = ACTIONS(2686),
    [anon_sym___builtin_available] = ACTIONS(2684),
    [anon_sym_va_arg] = ACTIONS(2684),
    [anon_sym_ATencode] = ACTIONS(2686),
    [anon_sym_BOOL] = ACTIONS(2684),
    [anon_sym_IMP] = ACTIONS(2684),
    [anon_sym_SEL] = ACTIONS(2684),
    [anon_sym_Class] = ACTIONS(2684),
    [anon_sym_id] = ACTIONS(2684),
  },
  [STATE(1021)] = {
    [ts_builtin_sym_end] = ACTIONS(2644),
    [sym_identifier] = ACTIONS(2642),
    [aux_sym_preproc_include_token1] = ACTIONS(2642),
    [aux_sym_preproc_include_token2] = ACTIONS(2642),
    [aux_sym_preproc_def_token1] = ACTIONS(2642),
    [aux_sym_preproc_if_token1] = ACTIONS(2642),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2642),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2642),
    [sym_preproc_directive] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_BANG] = ACTIONS(2644),
    [anon_sym_TILDE] = ACTIONS(2644),
    [anon_sym_DASH] = ACTIONS(2642),
    [anon_sym_PLUS] = ACTIONS(2642),
    [anon_sym_STAR] = ACTIONS(2644),
    [anon_sym_CARET] = ACTIONS(2644),
    [anon_sym_AMP] = ACTIONS(2644),
    [anon_sym_SEMI] = ACTIONS(2644),
    [anon_sym___extension__] = ACTIONS(2642),
    [anon_sym_typedef] = ACTIONS(2642),
    [anon_sym_extern] = ACTIONS(2642),
    [anon_sym___attribute__] = ACTIONS(2642),
    [anon_sym___attribute] = ACTIONS(2642),
    [anon_sym_noreturn] = ACTIONS(2642),
    [anon_sym_LBRACK] = ACTIONS(2644),
    [anon_sym___declspec] = ACTIONS(2642),
    [anon_sym___cdecl] = ACTIONS(2642),
    [anon_sym___clrcall] = ACTIONS(2642),
    [anon_sym___stdcall] = ACTIONS(2642),
    [anon_sym___fastcall] = ACTIONS(2642),
    [anon_sym___thiscall] = ACTIONS(2642),
    [anon_sym___vectorcall] = ACTIONS(2642),
    [anon_sym_LBRACE] = ACTIONS(2644),
    [anon_sym_signed] = ACTIONS(2642),
    [anon_sym_unsigned] = ACTIONS(2642),
    [anon_sym_long] = ACTIONS(2642),
    [anon_sym_short] = ACTIONS(2642),
    [anon_sym_static] = ACTIONS(2642),
    [anon_sym_ATautoreleasepool] = ACTIONS(2644),
    [anon_sym_auto] = ACTIONS(2642),
    [anon_sym_register] = ACTIONS(2642),
    [anon_sym_inline] = ACTIONS(2642),
    [anon_sym___inline] = ACTIONS(2642),
    [anon_sym___inline__] = ACTIONS(2642),
    [anon_sym___forceinline] = ACTIONS(2642),
    [anon_sym_thread_local] = ACTIONS(2642),
    [anon_sym___thread] = ACTIONS(2642),
    [anon_sym_CG_EXTERN] = ACTIONS(2642),
    [anon_sym_CG_INLINE] = ACTIONS(2642),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2642),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2642),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2642),
    [anon_sym_IBOutlet] = ACTIONS(2642),
    [anon_sym_IBInspectable] = ACTIONS(2642),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2642),
    [anon_sym_NS_INLINE] = ACTIONS(2642),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2642),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2642),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2642),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2642),
    [anon_sym_const] = ACTIONS(2642),
    [anon_sym_constexpr] = ACTIONS(2642),
    [anon_sym_volatile] = ACTIONS(2642),
    [anon_sym_restrict] = ACTIONS(2642),
    [anon_sym___restrict__] = ACTIONS(2642),
    [anon_sym__Atomic] = ACTIONS(2642),
    [anon_sym__Noreturn] = ACTIONS(2642),
    [anon_sym__Nonnull] = ACTIONS(2642),
    [anon_sym_nullable] = ACTIONS(2642),
    [anon_sym__Complex] = ACTIONS(2642),
    [anon_sym__Nullable] = ACTIONS(2642),
    [anon_sym__Nullable_result] = ACTIONS(2642),
    [anon_sym__Null_unspecified] = ACTIONS(2642),
    [anon_sym___autoreleasing] = ACTIONS(2642),
    [anon_sym___block] = ACTIONS(2642),
    [anon_sym___bridge] = ACTIONS(2642),
    [anon_sym___bridge_retained] = ACTIONS(2642),
    [anon_sym___bridge_transfer] = ACTIONS(2642),
    [anon_sym___complex] = ACTIONS(2642),
    [anon_sym___const] = ACTIONS(2642),
    [anon_sym___imag] = ACTIONS(2642),
    [anon_sym___kindof] = ACTIONS(2642),
    [anon_sym___nonnull] = ACTIONS(2642),
    [anon_sym___nullable] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2642),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2642),
    [anon_sym___real] = ACTIONS(2642),
    [anon_sym___strong] = ACTIONS(2642),
    [anon_sym___unsafe_unretained] = ACTIONS(2642),
    [anon_sym___unused] = ACTIONS(2642),
    [anon_sym___weak] = ACTIONS(2642),
    [anon_sym_alignas] = ACTIONS(2642),
    [anon_sym__Alignas] = ACTIONS(2642),
    [sym_primitive_type] = ACTIONS(2642),
    [anon_sym_enum] = ACTIONS(2642),
    [anon_sym_struct] = ACTIONS(2642),
    [anon_sym_union] = ACTIONS(2642),
    [anon_sym_if] = ACTIONS(2642),
    [anon_sym_switch] = ACTIONS(2642),
    [anon_sym_case] = ACTIONS(2642),
    [anon_sym_default] = ACTIONS(2642),
    [anon_sym_while] = ACTIONS(2642),
    [anon_sym_do] = ACTIONS(2642),
    [anon_sym_for] = ACTIONS(2642),
    [anon_sym_in] = ACTIONS(2642),
    [anon_sym_return] = ACTIONS(2642),
    [anon_sym_break] = ACTIONS(2642),
    [anon_sym_continue] = ACTIONS(2642),
    [anon_sym_goto] = ACTIONS(2642),
    [anon_sym_DASH_DASH] = ACTIONS(2644),
    [anon_sym_PLUS_PLUS] = ACTIONS(2644),
    [anon_sym_sizeof] = ACTIONS(2642),
    [anon_sym___alignof__] = ACTIONS(2642),
    [anon_sym___alignof] = ACTIONS(2642),
    [anon_sym__alignof] = ACTIONS(2642),
    [anon_sym_alignof] = ACTIONS(2642),
    [anon_sym__Alignof] = ACTIONS(2642),
    [anon_sym_offsetof] = ACTIONS(2642),
    [anon_sym__Generic] = ACTIONS(2642),
    [anon_sym_asm] = ACTIONS(2642),
    [anon_sym___asm__] = ACTIONS(2642),
    [anon_sym___asm] = ACTIONS(2642),
    [sym_number_literal] = ACTIONS(2644),
    [anon_sym_L_SQUOTE] = ACTIONS(2644),
    [anon_sym_u_SQUOTE] = ACTIONS(2644),
    [anon_sym_U_SQUOTE] = ACTIONS(2644),
    [anon_sym_u8_SQUOTE] = ACTIONS(2644),
    [anon_sym_SQUOTE] = ACTIONS(2644),
    [anon_sym_AT] = ACTIONS(2642),
    [anon_sym_DQUOTE] = ACTIONS(2644),
    [anon_sym_L_DQUOTE] = ACTIONS(2644),
    [anon_sym_u_DQUOTE] = ACTIONS(2644),
    [anon_sym_U_DQUOTE] = ACTIONS(2644),
    [anon_sym_u8_DQUOTE] = ACTIONS(2644),
    [sym_true] = ACTIONS(2642),
    [sym_false] = ACTIONS(2642),
    [anon_sym_NULL] = ACTIONS(2642),
    [anon_sym_nullptr] = ACTIONS(2642),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2642),
    [anon_sym___typeof] = ACTIONS(2642),
    [anon_sym_typeof] = ACTIONS(2642),
    [anon_sym_ATimport] = ACTIONS(2644),
    [aux_sym_preproc_undef_token1] = ACTIONS(2642),
    [anon_sym_POUND] = ACTIONS(2642),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2642),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2642),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2642),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2642),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2642),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2642),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2642),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_API_AVAILABLE] = ACTIONS(2642),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_API_DEPRECATED] = ACTIONS(2642),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2642),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2642),
    [anon_sym___deprecated_msg] = ACTIONS(2642),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2642),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2642),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2642),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2642),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2642),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2642),
    [anon_sym_ATprotocol] = ACTIONS(2644),
    [anon_sym_ATinterface] = ACTIONS(2644),
    [anon_sym_ATimplementation] = ACTIONS(2644),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2644),
    [anon_sym_ATselector] = ACTIONS(2644),
    [anon_sym_ATavailable] = ACTIONS(2644),
    [anon_sym___builtin_available] = ACTIONS(2642),
    [anon_sym_va_arg] = ACTIONS(2642),
    [anon_sym_ATencode] = ACTIONS(2644),
    [anon_sym_BOOL] = ACTIONS(2642),
    [anon_sym_IMP] = ACTIONS(2642),
    [anon_sym_SEL] = ACTIONS(2642),
    [anon_sym_Class] = ACTIONS(2642),
    [anon_sym_id] = ACTIONS(2642),
  },
  [STATE(1022)] = {
    [ts_builtin_sym_end] = ACTIONS(2586),
    [sym_identifier] = ACTIONS(2584),
    [aux_sym_preproc_include_token1] = ACTIONS(2584),
    [aux_sym_preproc_include_token2] = ACTIONS(2584),
    [aux_sym_preproc_def_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2584),
    [sym_preproc_directive] = ACTIONS(2584),
    [anon_sym_LPAREN2] = ACTIONS(2586),
    [anon_sym_BANG] = ACTIONS(2586),
    [anon_sym_TILDE] = ACTIONS(2586),
    [anon_sym_DASH] = ACTIONS(2584),
    [anon_sym_PLUS] = ACTIONS(2584),
    [anon_sym_STAR] = ACTIONS(2586),
    [anon_sym_CARET] = ACTIONS(2586),
    [anon_sym_AMP] = ACTIONS(2586),
    [anon_sym_SEMI] = ACTIONS(2586),
    [anon_sym___extension__] = ACTIONS(2584),
    [anon_sym_typedef] = ACTIONS(2584),
    [anon_sym_extern] = ACTIONS(2584),
    [anon_sym___attribute__] = ACTIONS(2584),
    [anon_sym___attribute] = ACTIONS(2584),
    [anon_sym_noreturn] = ACTIONS(2584),
    [anon_sym_LBRACK] = ACTIONS(2586),
    [anon_sym___declspec] = ACTIONS(2584),
    [anon_sym___cdecl] = ACTIONS(2584),
    [anon_sym___clrcall] = ACTIONS(2584),
    [anon_sym___stdcall] = ACTIONS(2584),
    [anon_sym___fastcall] = ACTIONS(2584),
    [anon_sym___thiscall] = ACTIONS(2584),
    [anon_sym___vectorcall] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_signed] = ACTIONS(2584),
    [anon_sym_unsigned] = ACTIONS(2584),
    [anon_sym_long] = ACTIONS(2584),
    [anon_sym_short] = ACTIONS(2584),
    [anon_sym_static] = ACTIONS(2584),
    [anon_sym_ATautoreleasepool] = ACTIONS(2586),
    [anon_sym_auto] = ACTIONS(2584),
    [anon_sym_register] = ACTIONS(2584),
    [anon_sym_inline] = ACTIONS(2584),
    [anon_sym___inline] = ACTIONS(2584),
    [anon_sym___inline__] = ACTIONS(2584),
    [anon_sym___forceinline] = ACTIONS(2584),
    [anon_sym_thread_local] = ACTIONS(2584),
    [anon_sym___thread] = ACTIONS(2584),
    [anon_sym_CG_EXTERN] = ACTIONS(2584),
    [anon_sym_CG_INLINE] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2584),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2584),
    [anon_sym_IBOutlet] = ACTIONS(2584),
    [anon_sym_IBInspectable] = ACTIONS(2584),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2584),
    [anon_sym_NS_INLINE] = ACTIONS(2584),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2584),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2584),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2584),
    [anon_sym_const] = ACTIONS(2584),
    [anon_sym_constexpr] = ACTIONS(2584),
    [anon_sym_volatile] = ACTIONS(2584),
    [anon_sym_restrict] = ACTIONS(2584),
    [anon_sym___restrict__] = ACTIONS(2584),
    [anon_sym__Atomic] = ACTIONS(2584),
    [anon_sym__Noreturn] = ACTIONS(2584),
    [anon_sym__Nonnull] = ACTIONS(2584),
    [anon_sym_nullable] = ACTIONS(2584),
    [anon_sym__Complex] = ACTIONS(2584),
    [anon_sym__Nullable] = ACTIONS(2584),
    [anon_sym__Nullable_result] = ACTIONS(2584),
    [anon_sym__Null_unspecified] = ACTIONS(2584),
    [anon_sym___autoreleasing] = ACTIONS(2584),
    [anon_sym___block] = ACTIONS(2584),
    [anon_sym___bridge] = ACTIONS(2584),
    [anon_sym___bridge_retained] = ACTIONS(2584),
    [anon_sym___bridge_transfer] = ACTIONS(2584),
    [anon_sym___complex] = ACTIONS(2584),
    [anon_sym___const] = ACTIONS(2584),
    [anon_sym___imag] = ACTIONS(2584),
    [anon_sym___kindof] = ACTIONS(2584),
    [anon_sym___nonnull] = ACTIONS(2584),
    [anon_sym___nullable] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2584),
    [anon_sym___real] = ACTIONS(2584),
    [anon_sym___strong] = ACTIONS(2584),
    [anon_sym___unsafe_unretained] = ACTIONS(2584),
    [anon_sym___unused] = ACTIONS(2584),
    [anon_sym___weak] = ACTIONS(2584),
    [anon_sym_alignas] = ACTIONS(2584),
    [anon_sym__Alignas] = ACTIONS(2584),
    [sym_primitive_type] = ACTIONS(2584),
    [anon_sym_enum] = ACTIONS(2584),
    [anon_sym_struct] = ACTIONS(2584),
    [anon_sym_union] = ACTIONS(2584),
    [anon_sym_if] = ACTIONS(2584),
    [anon_sym_switch] = ACTIONS(2584),
    [anon_sym_case] = ACTIONS(2584),
    [anon_sym_default] = ACTIONS(2584),
    [anon_sym_while] = ACTIONS(2584),
    [anon_sym_do] = ACTIONS(2584),
    [anon_sym_for] = ACTIONS(2584),
    [anon_sym_in] = ACTIONS(2584),
    [anon_sym_return] = ACTIONS(2584),
    [anon_sym_break] = ACTIONS(2584),
    [anon_sym_continue] = ACTIONS(2584),
    [anon_sym_goto] = ACTIONS(2584),
    [anon_sym_DASH_DASH] = ACTIONS(2586),
    [anon_sym_PLUS_PLUS] = ACTIONS(2586),
    [anon_sym_sizeof] = ACTIONS(2584),
    [anon_sym___alignof__] = ACTIONS(2584),
    [anon_sym___alignof] = ACTIONS(2584),
    [anon_sym__alignof] = ACTIONS(2584),
    [anon_sym_alignof] = ACTIONS(2584),
    [anon_sym__Alignof] = ACTIONS(2584),
    [anon_sym_offsetof] = ACTIONS(2584),
    [anon_sym__Generic] = ACTIONS(2584),
    [anon_sym_asm] = ACTIONS(2584),
    [anon_sym___asm__] = ACTIONS(2584),
    [anon_sym___asm] = ACTIONS(2584),
    [sym_number_literal] = ACTIONS(2586),
    [anon_sym_L_SQUOTE] = ACTIONS(2586),
    [anon_sym_u_SQUOTE] = ACTIONS(2586),
    [anon_sym_U_SQUOTE] = ACTIONS(2586),
    [anon_sym_u8_SQUOTE] = ACTIONS(2586),
    [anon_sym_SQUOTE] = ACTIONS(2586),
    [anon_sym_AT] = ACTIONS(2584),
    [anon_sym_DQUOTE] = ACTIONS(2586),
    [anon_sym_L_DQUOTE] = ACTIONS(2586),
    [anon_sym_u_DQUOTE] = ACTIONS(2586),
    [anon_sym_U_DQUOTE] = ACTIONS(2586),
    [anon_sym_u8_DQUOTE] = ACTIONS(2586),
    [sym_true] = ACTIONS(2584),
    [sym_false] = ACTIONS(2584),
    [anon_sym_NULL] = ACTIONS(2584),
    [anon_sym_nullptr] = ACTIONS(2584),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2584),
    [anon_sym___typeof] = ACTIONS(2584),
    [anon_sym_typeof] = ACTIONS(2584),
    [anon_sym_ATimport] = ACTIONS(2586),
    [aux_sym_preproc_undef_token1] = ACTIONS(2584),
    [anon_sym_POUND] = ACTIONS(2584),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2584),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2584),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2584),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_API_AVAILABLE] = ACTIONS(2584),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_API_DEPRECATED] = ACTIONS(2584),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym___deprecated_msg] = ACTIONS(2584),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2584),
    [anon_sym_ATprotocol] = ACTIONS(2586),
    [anon_sym_ATinterface] = ACTIONS(2586),
    [anon_sym_ATimplementation] = ACTIONS(2586),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2586),
    [anon_sym_ATselector] = ACTIONS(2586),
    [anon_sym_ATavailable] = ACTIONS(2586),
    [anon_sym___builtin_available] = ACTIONS(2584),
    [anon_sym_va_arg] = ACTIONS(2584),
    [anon_sym_ATencode] = ACTIONS(2586),
    [anon_sym_BOOL] = ACTIONS(2584),
    [anon_sym_IMP] = ACTIONS(2584),
    [anon_sym_SEL] = ACTIONS(2584),
    [anon_sym_Class] = ACTIONS(2584),
    [anon_sym_id] = ACTIONS(2584),
  },
  [STATE(1023)] = {
    [ts_builtin_sym_end] = ACTIONS(2866),
    [sym_identifier] = ACTIONS(2864),
    [aux_sym_preproc_include_token1] = ACTIONS(2864),
    [aux_sym_preproc_include_token2] = ACTIONS(2864),
    [aux_sym_preproc_def_token1] = ACTIONS(2864),
    [aux_sym_preproc_if_token1] = ACTIONS(2864),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2864),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2864),
    [sym_preproc_directive] = ACTIONS(2864),
    [anon_sym_LPAREN2] = ACTIONS(2866),
    [anon_sym_BANG] = ACTIONS(2866),
    [anon_sym_TILDE] = ACTIONS(2866),
    [anon_sym_DASH] = ACTIONS(2864),
    [anon_sym_PLUS] = ACTIONS(2864),
    [anon_sym_STAR] = ACTIONS(2866),
    [anon_sym_CARET] = ACTIONS(2866),
    [anon_sym_AMP] = ACTIONS(2866),
    [anon_sym_SEMI] = ACTIONS(2866),
    [anon_sym___extension__] = ACTIONS(2864),
    [anon_sym_typedef] = ACTIONS(2864),
    [anon_sym_extern] = ACTIONS(2864),
    [anon_sym___attribute__] = ACTIONS(2864),
    [anon_sym___attribute] = ACTIONS(2864),
    [anon_sym_noreturn] = ACTIONS(2864),
    [anon_sym_LBRACK] = ACTIONS(2866),
    [anon_sym___declspec] = ACTIONS(2864),
    [anon_sym___cdecl] = ACTIONS(2864),
    [anon_sym___clrcall] = ACTIONS(2864),
    [anon_sym___stdcall] = ACTIONS(2864),
    [anon_sym___fastcall] = ACTIONS(2864),
    [anon_sym___thiscall] = ACTIONS(2864),
    [anon_sym___vectorcall] = ACTIONS(2864),
    [anon_sym_LBRACE] = ACTIONS(2866),
    [anon_sym_signed] = ACTIONS(2864),
    [anon_sym_unsigned] = ACTIONS(2864),
    [anon_sym_long] = ACTIONS(2864),
    [anon_sym_short] = ACTIONS(2864),
    [anon_sym_static] = ACTIONS(2864),
    [anon_sym_ATautoreleasepool] = ACTIONS(2866),
    [anon_sym_auto] = ACTIONS(2864),
    [anon_sym_register] = ACTIONS(2864),
    [anon_sym_inline] = ACTIONS(2864),
    [anon_sym___inline] = ACTIONS(2864),
    [anon_sym___inline__] = ACTIONS(2864),
    [anon_sym___forceinline] = ACTIONS(2864),
    [anon_sym_thread_local] = ACTIONS(2864),
    [anon_sym___thread] = ACTIONS(2864),
    [anon_sym_CG_EXTERN] = ACTIONS(2864),
    [anon_sym_CG_INLINE] = ACTIONS(2864),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2864),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2864),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2864),
    [anon_sym_IBOutlet] = ACTIONS(2864),
    [anon_sym_IBInspectable] = ACTIONS(2864),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2864),
    [anon_sym_NS_INLINE] = ACTIONS(2864),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2864),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2864),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2864),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2864),
    [anon_sym_const] = ACTIONS(2864),
    [anon_sym_constexpr] = ACTIONS(2864),
    [anon_sym_volatile] = ACTIONS(2864),
    [anon_sym_restrict] = ACTIONS(2864),
    [anon_sym___restrict__] = ACTIONS(2864),
    [anon_sym__Atomic] = ACTIONS(2864),
    [anon_sym__Noreturn] = ACTIONS(2864),
    [anon_sym__Nonnull] = ACTIONS(2864),
    [anon_sym_nullable] = ACTIONS(2864),
    [anon_sym__Complex] = ACTIONS(2864),
    [anon_sym__Nullable] = ACTIONS(2864),
    [anon_sym__Nullable_result] = ACTIONS(2864),
    [anon_sym__Null_unspecified] = ACTIONS(2864),
    [anon_sym___autoreleasing] = ACTIONS(2864),
    [anon_sym___block] = ACTIONS(2864),
    [anon_sym___bridge] = ACTIONS(2864),
    [anon_sym___bridge_retained] = ACTIONS(2864),
    [anon_sym___bridge_transfer] = ACTIONS(2864),
    [anon_sym___complex] = ACTIONS(2864),
    [anon_sym___const] = ACTIONS(2864),
    [anon_sym___imag] = ACTIONS(2864),
    [anon_sym___kindof] = ACTIONS(2864),
    [anon_sym___nonnull] = ACTIONS(2864),
    [anon_sym___nullable] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2864),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2864),
    [anon_sym___real] = ACTIONS(2864),
    [anon_sym___strong] = ACTIONS(2864),
    [anon_sym___unsafe_unretained] = ACTIONS(2864),
    [anon_sym___unused] = ACTIONS(2864),
    [anon_sym___weak] = ACTIONS(2864),
    [anon_sym_alignas] = ACTIONS(2864),
    [anon_sym__Alignas] = ACTIONS(2864),
    [sym_primitive_type] = ACTIONS(2864),
    [anon_sym_enum] = ACTIONS(2864),
    [anon_sym_struct] = ACTIONS(2864),
    [anon_sym_union] = ACTIONS(2864),
    [anon_sym_if] = ACTIONS(2864),
    [anon_sym_switch] = ACTIONS(2864),
    [anon_sym_case] = ACTIONS(2864),
    [anon_sym_default] = ACTIONS(2864),
    [anon_sym_while] = ACTIONS(2864),
    [anon_sym_do] = ACTIONS(2864),
    [anon_sym_for] = ACTIONS(2864),
    [anon_sym_in] = ACTIONS(2864),
    [anon_sym_return] = ACTIONS(2864),
    [anon_sym_break] = ACTIONS(2864),
    [anon_sym_continue] = ACTIONS(2864),
    [anon_sym_goto] = ACTIONS(2864),
    [anon_sym_DASH_DASH] = ACTIONS(2866),
    [anon_sym_PLUS_PLUS] = ACTIONS(2866),
    [anon_sym_sizeof] = ACTIONS(2864),
    [anon_sym___alignof__] = ACTIONS(2864),
    [anon_sym___alignof] = ACTIONS(2864),
    [anon_sym__alignof] = ACTIONS(2864),
    [anon_sym_alignof] = ACTIONS(2864),
    [anon_sym__Alignof] = ACTIONS(2864),
    [anon_sym_offsetof] = ACTIONS(2864),
    [anon_sym__Generic] = ACTIONS(2864),
    [anon_sym_asm] = ACTIONS(2864),
    [anon_sym___asm__] = ACTIONS(2864),
    [anon_sym___asm] = ACTIONS(2864),
    [sym_number_literal] = ACTIONS(2866),
    [anon_sym_L_SQUOTE] = ACTIONS(2866),
    [anon_sym_u_SQUOTE] = ACTIONS(2866),
    [anon_sym_U_SQUOTE] = ACTIONS(2866),
    [anon_sym_u8_SQUOTE] = ACTIONS(2866),
    [anon_sym_SQUOTE] = ACTIONS(2866),
    [anon_sym_AT] = ACTIONS(2864),
    [anon_sym_DQUOTE] = ACTIONS(2866),
    [anon_sym_L_DQUOTE] = ACTIONS(2866),
    [anon_sym_u_DQUOTE] = ACTIONS(2866),
    [anon_sym_U_DQUOTE] = ACTIONS(2866),
    [anon_sym_u8_DQUOTE] = ACTIONS(2866),
    [sym_true] = ACTIONS(2864),
    [sym_false] = ACTIONS(2864),
    [anon_sym_NULL] = ACTIONS(2864),
    [anon_sym_nullptr] = ACTIONS(2864),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2864),
    [anon_sym___typeof] = ACTIONS(2864),
    [anon_sym_typeof] = ACTIONS(2864),
    [anon_sym_ATimport] = ACTIONS(2866),
    [aux_sym_preproc_undef_token1] = ACTIONS(2864),
    [anon_sym_POUND] = ACTIONS(2864),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2864),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2864),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2864),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2864),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2864),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2864),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2864),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_API_AVAILABLE] = ACTIONS(2864),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_API_DEPRECATED] = ACTIONS(2864),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2864),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2864),
    [anon_sym___deprecated_msg] = ACTIONS(2864),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2864),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2864),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2864),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2864),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2864),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2864),
    [anon_sym_ATprotocol] = ACTIONS(2866),
    [anon_sym_ATinterface] = ACTIONS(2866),
    [anon_sym_ATimplementation] = ACTIONS(2866),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2866),
    [anon_sym_ATselector] = ACTIONS(2866),
    [anon_sym_ATavailable] = ACTIONS(2866),
    [anon_sym___builtin_available] = ACTIONS(2864),
    [anon_sym_va_arg] = ACTIONS(2864),
    [anon_sym_ATencode] = ACTIONS(2866),
    [anon_sym_BOOL] = ACTIONS(2864),
    [anon_sym_IMP] = ACTIONS(2864),
    [anon_sym_SEL] = ACTIONS(2864),
    [anon_sym_Class] = ACTIONS(2864),
    [anon_sym_id] = ACTIONS(2864),
  },
  [STATE(1024)] = {
    [ts_builtin_sym_end] = ACTIONS(2694),
    [sym_identifier] = ACTIONS(2692),
    [aux_sym_preproc_include_token1] = ACTIONS(2692),
    [aux_sym_preproc_include_token2] = ACTIONS(2692),
    [aux_sym_preproc_def_token1] = ACTIONS(2692),
    [aux_sym_preproc_if_token1] = ACTIONS(2692),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2692),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2692),
    [sym_preproc_directive] = ACTIONS(2692),
    [anon_sym_LPAREN2] = ACTIONS(2694),
    [anon_sym_BANG] = ACTIONS(2694),
    [anon_sym_TILDE] = ACTIONS(2694),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2694),
    [anon_sym_CARET] = ACTIONS(2694),
    [anon_sym_AMP] = ACTIONS(2694),
    [anon_sym_SEMI] = ACTIONS(2694),
    [anon_sym___extension__] = ACTIONS(2692),
    [anon_sym_typedef] = ACTIONS(2692),
    [anon_sym_extern] = ACTIONS(2692),
    [anon_sym___attribute__] = ACTIONS(2692),
    [anon_sym___attribute] = ACTIONS(2692),
    [anon_sym_noreturn] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym___declspec] = ACTIONS(2692),
    [anon_sym___cdecl] = ACTIONS(2692),
    [anon_sym___clrcall] = ACTIONS(2692),
    [anon_sym___stdcall] = ACTIONS(2692),
    [anon_sym___fastcall] = ACTIONS(2692),
    [anon_sym___thiscall] = ACTIONS(2692),
    [anon_sym___vectorcall] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_signed] = ACTIONS(2692),
    [anon_sym_unsigned] = ACTIONS(2692),
    [anon_sym_long] = ACTIONS(2692),
    [anon_sym_short] = ACTIONS(2692),
    [anon_sym_static] = ACTIONS(2692),
    [anon_sym_ATautoreleasepool] = ACTIONS(2694),
    [anon_sym_auto] = ACTIONS(2692),
    [anon_sym_register] = ACTIONS(2692),
    [anon_sym_inline] = ACTIONS(2692),
    [anon_sym___inline] = ACTIONS(2692),
    [anon_sym___inline__] = ACTIONS(2692),
    [anon_sym___forceinline] = ACTIONS(2692),
    [anon_sym_thread_local] = ACTIONS(2692),
    [anon_sym___thread] = ACTIONS(2692),
    [anon_sym_CG_EXTERN] = ACTIONS(2692),
    [anon_sym_CG_INLINE] = ACTIONS(2692),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2692),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2692),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2692),
    [anon_sym_IBOutlet] = ACTIONS(2692),
    [anon_sym_IBInspectable] = ACTIONS(2692),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2692),
    [anon_sym_NS_INLINE] = ACTIONS(2692),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2692),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2692),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2692),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2692),
    [anon_sym_const] = ACTIONS(2692),
    [anon_sym_constexpr] = ACTIONS(2692),
    [anon_sym_volatile] = ACTIONS(2692),
    [anon_sym_restrict] = ACTIONS(2692),
    [anon_sym___restrict__] = ACTIONS(2692),
    [anon_sym__Atomic] = ACTIONS(2692),
    [anon_sym__Noreturn] = ACTIONS(2692),
    [anon_sym__Nonnull] = ACTIONS(2692),
    [anon_sym_nullable] = ACTIONS(2692),
    [anon_sym__Complex] = ACTIONS(2692),
    [anon_sym__Nullable] = ACTIONS(2692),
    [anon_sym__Nullable_result] = ACTIONS(2692),
    [anon_sym__Null_unspecified] = ACTIONS(2692),
    [anon_sym___autoreleasing] = ACTIONS(2692),
    [anon_sym___block] = ACTIONS(2692),
    [anon_sym___bridge] = ACTIONS(2692),
    [anon_sym___bridge_retained] = ACTIONS(2692),
    [anon_sym___bridge_transfer] = ACTIONS(2692),
    [anon_sym___complex] = ACTIONS(2692),
    [anon_sym___const] = ACTIONS(2692),
    [anon_sym___imag] = ACTIONS(2692),
    [anon_sym___kindof] = ACTIONS(2692),
    [anon_sym___nonnull] = ACTIONS(2692),
    [anon_sym___nullable] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2692),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2692),
    [anon_sym___real] = ACTIONS(2692),
    [anon_sym___strong] = ACTIONS(2692),
    [anon_sym___unsafe_unretained] = ACTIONS(2692),
    [anon_sym___unused] = ACTIONS(2692),
    [anon_sym___weak] = ACTIONS(2692),
    [anon_sym_alignas] = ACTIONS(2692),
    [anon_sym__Alignas] = ACTIONS(2692),
    [sym_primitive_type] = ACTIONS(2692),
    [anon_sym_enum] = ACTIONS(2692),
    [anon_sym_struct] = ACTIONS(2692),
    [anon_sym_union] = ACTIONS(2692),
    [anon_sym_if] = ACTIONS(2692),
    [anon_sym_switch] = ACTIONS(2692),
    [anon_sym_case] = ACTIONS(2692),
    [anon_sym_default] = ACTIONS(2692),
    [anon_sym_while] = ACTIONS(2692),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_for] = ACTIONS(2692),
    [anon_sym_in] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_sizeof] = ACTIONS(2692),
    [anon_sym___alignof__] = ACTIONS(2692),
    [anon_sym___alignof] = ACTIONS(2692),
    [anon_sym__alignof] = ACTIONS(2692),
    [anon_sym_alignof] = ACTIONS(2692),
    [anon_sym__Alignof] = ACTIONS(2692),
    [anon_sym_offsetof] = ACTIONS(2692),
    [anon_sym__Generic] = ACTIONS(2692),
    [anon_sym_asm] = ACTIONS(2692),
    [anon_sym___asm__] = ACTIONS(2692),
    [anon_sym___asm] = ACTIONS(2692),
    [sym_number_literal] = ACTIONS(2694),
    [anon_sym_L_SQUOTE] = ACTIONS(2694),
    [anon_sym_u_SQUOTE] = ACTIONS(2694),
    [anon_sym_U_SQUOTE] = ACTIONS(2694),
    [anon_sym_u8_SQUOTE] = ACTIONS(2694),
    [anon_sym_SQUOTE] = ACTIONS(2694),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_DQUOTE] = ACTIONS(2694),
    [anon_sym_L_DQUOTE] = ACTIONS(2694),
    [anon_sym_u_DQUOTE] = ACTIONS(2694),
    [anon_sym_U_DQUOTE] = ACTIONS(2694),
    [anon_sym_u8_DQUOTE] = ACTIONS(2694),
    [sym_true] = ACTIONS(2692),
    [sym_false] = ACTIONS(2692),
    [anon_sym_NULL] = ACTIONS(2692),
    [anon_sym_nullptr] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2692),
    [anon_sym___typeof] = ACTIONS(2692),
    [anon_sym_typeof] = ACTIONS(2692),
    [anon_sym_ATimport] = ACTIONS(2694),
    [aux_sym_preproc_undef_token1] = ACTIONS(2692),
    [anon_sym_POUND] = ACTIONS(2692),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2692),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2692),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2692),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2692),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2692),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2692),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2692),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_API_AVAILABLE] = ACTIONS(2692),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_API_DEPRECATED] = ACTIONS(2692),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2692),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2692),
    [anon_sym___deprecated_msg] = ACTIONS(2692),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2692),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2692),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2692),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2692),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2692),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2692),
    [anon_sym_ATprotocol] = ACTIONS(2694),
    [anon_sym_ATinterface] = ACTIONS(2694),
    [anon_sym_ATimplementation] = ACTIONS(2694),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2694),
    [anon_sym_ATselector] = ACTIONS(2694),
    [anon_sym_ATavailable] = ACTIONS(2694),
    [anon_sym___builtin_available] = ACTIONS(2692),
    [anon_sym_va_arg] = ACTIONS(2692),
    [anon_sym_ATencode] = ACTIONS(2694),
    [anon_sym_BOOL] = ACTIONS(2692),
    [anon_sym_IMP] = ACTIONS(2692),
    [anon_sym_SEL] = ACTIONS(2692),
    [anon_sym_Class] = ACTIONS(2692),
    [anon_sym_id] = ACTIONS(2692),
  },
  [STATE(1025)] = {
    [sym_ms_declspec_modifier] = STATE(6745),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2946),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(1026)] = {
    [ts_builtin_sym_end] = ACTIONS(2698),
    [sym_identifier] = ACTIONS(2696),
    [aux_sym_preproc_include_token1] = ACTIONS(2696),
    [aux_sym_preproc_include_token2] = ACTIONS(2696),
    [aux_sym_preproc_def_token1] = ACTIONS(2696),
    [aux_sym_preproc_if_token1] = ACTIONS(2696),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2696),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2696),
    [sym_preproc_directive] = ACTIONS(2696),
    [anon_sym_LPAREN2] = ACTIONS(2698),
    [anon_sym_BANG] = ACTIONS(2698),
    [anon_sym_TILDE] = ACTIONS(2698),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2698),
    [anon_sym_CARET] = ACTIONS(2698),
    [anon_sym_AMP] = ACTIONS(2698),
    [anon_sym_SEMI] = ACTIONS(2698),
    [anon_sym___extension__] = ACTIONS(2696),
    [anon_sym_typedef] = ACTIONS(2696),
    [anon_sym_extern] = ACTIONS(2696),
    [anon_sym___attribute__] = ACTIONS(2696),
    [anon_sym___attribute] = ACTIONS(2696),
    [anon_sym_noreturn] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym___declspec] = ACTIONS(2696),
    [anon_sym___cdecl] = ACTIONS(2696),
    [anon_sym___clrcall] = ACTIONS(2696),
    [anon_sym___stdcall] = ACTIONS(2696),
    [anon_sym___fastcall] = ACTIONS(2696),
    [anon_sym___thiscall] = ACTIONS(2696),
    [anon_sym___vectorcall] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_signed] = ACTIONS(2696),
    [anon_sym_unsigned] = ACTIONS(2696),
    [anon_sym_long] = ACTIONS(2696),
    [anon_sym_short] = ACTIONS(2696),
    [anon_sym_static] = ACTIONS(2696),
    [anon_sym_ATautoreleasepool] = ACTIONS(2698),
    [anon_sym_auto] = ACTIONS(2696),
    [anon_sym_register] = ACTIONS(2696),
    [anon_sym_inline] = ACTIONS(2696),
    [anon_sym___inline] = ACTIONS(2696),
    [anon_sym___inline__] = ACTIONS(2696),
    [anon_sym___forceinline] = ACTIONS(2696),
    [anon_sym_thread_local] = ACTIONS(2696),
    [anon_sym___thread] = ACTIONS(2696),
    [anon_sym_CG_EXTERN] = ACTIONS(2696),
    [anon_sym_CG_INLINE] = ACTIONS(2696),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2696),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2696),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2696),
    [anon_sym_IBOutlet] = ACTIONS(2696),
    [anon_sym_IBInspectable] = ACTIONS(2696),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2696),
    [anon_sym_NS_INLINE] = ACTIONS(2696),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2696),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2696),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2696),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2696),
    [anon_sym_const] = ACTIONS(2696),
    [anon_sym_constexpr] = ACTIONS(2696),
    [anon_sym_volatile] = ACTIONS(2696),
    [anon_sym_restrict] = ACTIONS(2696),
    [anon_sym___restrict__] = ACTIONS(2696),
    [anon_sym__Atomic] = ACTIONS(2696),
    [anon_sym__Noreturn] = ACTIONS(2696),
    [anon_sym__Nonnull] = ACTIONS(2696),
    [anon_sym_nullable] = ACTIONS(2696),
    [anon_sym__Complex] = ACTIONS(2696),
    [anon_sym__Nullable] = ACTIONS(2696),
    [anon_sym__Nullable_result] = ACTIONS(2696),
    [anon_sym__Null_unspecified] = ACTIONS(2696),
    [anon_sym___autoreleasing] = ACTIONS(2696),
    [anon_sym___block] = ACTIONS(2696),
    [anon_sym___bridge] = ACTIONS(2696),
    [anon_sym___bridge_retained] = ACTIONS(2696),
    [anon_sym___bridge_transfer] = ACTIONS(2696),
    [anon_sym___complex] = ACTIONS(2696),
    [anon_sym___const] = ACTIONS(2696),
    [anon_sym___imag] = ACTIONS(2696),
    [anon_sym___kindof] = ACTIONS(2696),
    [anon_sym___nonnull] = ACTIONS(2696),
    [anon_sym___nullable] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2696),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2696),
    [anon_sym___real] = ACTIONS(2696),
    [anon_sym___strong] = ACTIONS(2696),
    [anon_sym___unsafe_unretained] = ACTIONS(2696),
    [anon_sym___unused] = ACTIONS(2696),
    [anon_sym___weak] = ACTIONS(2696),
    [anon_sym_alignas] = ACTIONS(2696),
    [anon_sym__Alignas] = ACTIONS(2696),
    [sym_primitive_type] = ACTIONS(2696),
    [anon_sym_enum] = ACTIONS(2696),
    [anon_sym_struct] = ACTIONS(2696),
    [anon_sym_union] = ACTIONS(2696),
    [anon_sym_if] = ACTIONS(2696),
    [anon_sym_switch] = ACTIONS(2696),
    [anon_sym_case] = ACTIONS(2696),
    [anon_sym_default] = ACTIONS(2696),
    [anon_sym_while] = ACTIONS(2696),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_for] = ACTIONS(2696),
    [anon_sym_in] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_sizeof] = ACTIONS(2696),
    [anon_sym___alignof__] = ACTIONS(2696),
    [anon_sym___alignof] = ACTIONS(2696),
    [anon_sym__alignof] = ACTIONS(2696),
    [anon_sym_alignof] = ACTIONS(2696),
    [anon_sym__Alignof] = ACTIONS(2696),
    [anon_sym_offsetof] = ACTIONS(2696),
    [anon_sym__Generic] = ACTIONS(2696),
    [anon_sym_asm] = ACTIONS(2696),
    [anon_sym___asm__] = ACTIONS(2696),
    [anon_sym___asm] = ACTIONS(2696),
    [sym_number_literal] = ACTIONS(2698),
    [anon_sym_L_SQUOTE] = ACTIONS(2698),
    [anon_sym_u_SQUOTE] = ACTIONS(2698),
    [anon_sym_U_SQUOTE] = ACTIONS(2698),
    [anon_sym_u8_SQUOTE] = ACTIONS(2698),
    [anon_sym_SQUOTE] = ACTIONS(2698),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_DQUOTE] = ACTIONS(2698),
    [anon_sym_L_DQUOTE] = ACTIONS(2698),
    [anon_sym_u_DQUOTE] = ACTIONS(2698),
    [anon_sym_U_DQUOTE] = ACTIONS(2698),
    [anon_sym_u8_DQUOTE] = ACTIONS(2698),
    [sym_true] = ACTIONS(2696),
    [sym_false] = ACTIONS(2696),
    [anon_sym_NULL] = ACTIONS(2696),
    [anon_sym_nullptr] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2696),
    [anon_sym___typeof] = ACTIONS(2696),
    [anon_sym_typeof] = ACTIONS(2696),
    [anon_sym_ATimport] = ACTIONS(2698),
    [aux_sym_preproc_undef_token1] = ACTIONS(2696),
    [anon_sym_POUND] = ACTIONS(2696),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2696),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2696),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2696),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2696),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2696),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2696),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2696),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_API_AVAILABLE] = ACTIONS(2696),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_API_DEPRECATED] = ACTIONS(2696),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2696),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2696),
    [anon_sym___deprecated_msg] = ACTIONS(2696),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2696),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2696),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2696),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2696),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2696),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2696),
    [anon_sym_ATprotocol] = ACTIONS(2698),
    [anon_sym_ATinterface] = ACTIONS(2698),
    [anon_sym_ATimplementation] = ACTIONS(2698),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2698),
    [anon_sym_ATselector] = ACTIONS(2698),
    [anon_sym_ATavailable] = ACTIONS(2698),
    [anon_sym___builtin_available] = ACTIONS(2696),
    [anon_sym_va_arg] = ACTIONS(2696),
    [anon_sym_ATencode] = ACTIONS(2698),
    [anon_sym_BOOL] = ACTIONS(2696),
    [anon_sym_IMP] = ACTIONS(2696),
    [anon_sym_SEL] = ACTIONS(2696),
    [anon_sym_Class] = ACTIONS(2696),
    [anon_sym_id] = ACTIONS(2696),
  },
  [STATE(1027)] = {
    [ts_builtin_sym_end] = ACTIONS(2702),
    [sym_identifier] = ACTIONS(2700),
    [aux_sym_preproc_include_token1] = ACTIONS(2700),
    [aux_sym_preproc_include_token2] = ACTIONS(2700),
    [aux_sym_preproc_def_token1] = ACTIONS(2700),
    [aux_sym_preproc_if_token1] = ACTIONS(2700),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2700),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2700),
    [sym_preproc_directive] = ACTIONS(2700),
    [anon_sym_LPAREN2] = ACTIONS(2702),
    [anon_sym_BANG] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2700),
    [anon_sym_PLUS] = ACTIONS(2700),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym___extension__] = ACTIONS(2700),
    [anon_sym_typedef] = ACTIONS(2700),
    [anon_sym_extern] = ACTIONS(2700),
    [anon_sym___attribute__] = ACTIONS(2700),
    [anon_sym___attribute] = ACTIONS(2700),
    [anon_sym_noreturn] = ACTIONS(2700),
    [anon_sym_LBRACK] = ACTIONS(2702),
    [anon_sym___declspec] = ACTIONS(2700),
    [anon_sym___cdecl] = ACTIONS(2700),
    [anon_sym___clrcall] = ACTIONS(2700),
    [anon_sym___stdcall] = ACTIONS(2700),
    [anon_sym___fastcall] = ACTIONS(2700),
    [anon_sym___thiscall] = ACTIONS(2700),
    [anon_sym___vectorcall] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2702),
    [anon_sym_signed] = ACTIONS(2700),
    [anon_sym_unsigned] = ACTIONS(2700),
    [anon_sym_long] = ACTIONS(2700),
    [anon_sym_short] = ACTIONS(2700),
    [anon_sym_static] = ACTIONS(2700),
    [anon_sym_ATautoreleasepool] = ACTIONS(2702),
    [anon_sym_auto] = ACTIONS(2700),
    [anon_sym_register] = ACTIONS(2700),
    [anon_sym_inline] = ACTIONS(2700),
    [anon_sym___inline] = ACTIONS(2700),
    [anon_sym___inline__] = ACTIONS(2700),
    [anon_sym___forceinline] = ACTIONS(2700),
    [anon_sym_thread_local] = ACTIONS(2700),
    [anon_sym___thread] = ACTIONS(2700),
    [anon_sym_CG_EXTERN] = ACTIONS(2700),
    [anon_sym_CG_INLINE] = ACTIONS(2700),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2700),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2700),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2700),
    [anon_sym_IBOutlet] = ACTIONS(2700),
    [anon_sym_IBInspectable] = ACTIONS(2700),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2700),
    [anon_sym_NS_INLINE] = ACTIONS(2700),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2700),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2700),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2700),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2700),
    [anon_sym_const] = ACTIONS(2700),
    [anon_sym_constexpr] = ACTIONS(2700),
    [anon_sym_volatile] = ACTIONS(2700),
    [anon_sym_restrict] = ACTIONS(2700),
    [anon_sym___restrict__] = ACTIONS(2700),
    [anon_sym__Atomic] = ACTIONS(2700),
    [anon_sym__Noreturn] = ACTIONS(2700),
    [anon_sym__Nonnull] = ACTIONS(2700),
    [anon_sym_nullable] = ACTIONS(2700),
    [anon_sym__Complex] = ACTIONS(2700),
    [anon_sym__Nullable] = ACTIONS(2700),
    [anon_sym__Nullable_result] = ACTIONS(2700),
    [anon_sym__Null_unspecified] = ACTIONS(2700),
    [anon_sym___autoreleasing] = ACTIONS(2700),
    [anon_sym___block] = ACTIONS(2700),
    [anon_sym___bridge] = ACTIONS(2700),
    [anon_sym___bridge_retained] = ACTIONS(2700),
    [anon_sym___bridge_transfer] = ACTIONS(2700),
    [anon_sym___complex] = ACTIONS(2700),
    [anon_sym___const] = ACTIONS(2700),
    [anon_sym___imag] = ACTIONS(2700),
    [anon_sym___kindof] = ACTIONS(2700),
    [anon_sym___nonnull] = ACTIONS(2700),
    [anon_sym___nullable] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2700),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2700),
    [anon_sym___real] = ACTIONS(2700),
    [anon_sym___strong] = ACTIONS(2700),
    [anon_sym___unsafe_unretained] = ACTIONS(2700),
    [anon_sym___unused] = ACTIONS(2700),
    [anon_sym___weak] = ACTIONS(2700),
    [anon_sym_alignas] = ACTIONS(2700),
    [anon_sym__Alignas] = ACTIONS(2700),
    [sym_primitive_type] = ACTIONS(2700),
    [anon_sym_enum] = ACTIONS(2700),
    [anon_sym_struct] = ACTIONS(2700),
    [anon_sym_union] = ACTIONS(2700),
    [anon_sym_if] = ACTIONS(2700),
    [anon_sym_switch] = ACTIONS(2700),
    [anon_sym_case] = ACTIONS(2700),
    [anon_sym_default] = ACTIONS(2700),
    [anon_sym_while] = ACTIONS(2700),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_for] = ACTIONS(2700),
    [anon_sym_in] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_DASH_DASH] = ACTIONS(2702),
    [anon_sym_PLUS_PLUS] = ACTIONS(2702),
    [anon_sym_sizeof] = ACTIONS(2700),
    [anon_sym___alignof__] = ACTIONS(2700),
    [anon_sym___alignof] = ACTIONS(2700),
    [anon_sym__alignof] = ACTIONS(2700),
    [anon_sym_alignof] = ACTIONS(2700),
    [anon_sym__Alignof] = ACTIONS(2700),
    [anon_sym_offsetof] = ACTIONS(2700),
    [anon_sym__Generic] = ACTIONS(2700),
    [anon_sym_asm] = ACTIONS(2700),
    [anon_sym___asm__] = ACTIONS(2700),
    [anon_sym___asm] = ACTIONS(2700),
    [sym_number_literal] = ACTIONS(2702),
    [anon_sym_L_SQUOTE] = ACTIONS(2702),
    [anon_sym_u_SQUOTE] = ACTIONS(2702),
    [anon_sym_U_SQUOTE] = ACTIONS(2702),
    [anon_sym_u8_SQUOTE] = ACTIONS(2702),
    [anon_sym_SQUOTE] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_DQUOTE] = ACTIONS(2702),
    [anon_sym_L_DQUOTE] = ACTIONS(2702),
    [anon_sym_u_DQUOTE] = ACTIONS(2702),
    [anon_sym_U_DQUOTE] = ACTIONS(2702),
    [anon_sym_u8_DQUOTE] = ACTIONS(2702),
    [sym_true] = ACTIONS(2700),
    [sym_false] = ACTIONS(2700),
    [anon_sym_NULL] = ACTIONS(2700),
    [anon_sym_nullptr] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2700),
    [anon_sym___typeof] = ACTIONS(2700),
    [anon_sym_typeof] = ACTIONS(2700),
    [anon_sym_ATimport] = ACTIONS(2702),
    [aux_sym_preproc_undef_token1] = ACTIONS(2700),
    [anon_sym_POUND] = ACTIONS(2700),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2700),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2700),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2700),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2700),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2700),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2700),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2700),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_API_AVAILABLE] = ACTIONS(2700),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_API_DEPRECATED] = ACTIONS(2700),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2700),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2700),
    [anon_sym___deprecated_msg] = ACTIONS(2700),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2700),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2700),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2700),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2700),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2700),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2700),
    [anon_sym_ATprotocol] = ACTIONS(2702),
    [anon_sym_ATinterface] = ACTIONS(2702),
    [anon_sym_ATimplementation] = ACTIONS(2702),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2702),
    [anon_sym_ATselector] = ACTIONS(2702),
    [anon_sym_ATavailable] = ACTIONS(2702),
    [anon_sym___builtin_available] = ACTIONS(2700),
    [anon_sym_va_arg] = ACTIONS(2700),
    [anon_sym_ATencode] = ACTIONS(2702),
    [anon_sym_BOOL] = ACTIONS(2700),
    [anon_sym_IMP] = ACTIONS(2700),
    [anon_sym_SEL] = ACTIONS(2700),
    [anon_sym_Class] = ACTIONS(2700),
    [anon_sym_id] = ACTIONS(2700),
  },
  [STATE(1028)] = {
    [ts_builtin_sym_end] = ACTIONS(2610),
    [sym_identifier] = ACTIONS(2608),
    [aux_sym_preproc_include_token1] = ACTIONS(2608),
    [aux_sym_preproc_include_token2] = ACTIONS(2608),
    [aux_sym_preproc_def_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2608),
    [sym_preproc_directive] = ACTIONS(2608),
    [anon_sym_LPAREN2] = ACTIONS(2610),
    [anon_sym_BANG] = ACTIONS(2610),
    [anon_sym_TILDE] = ACTIONS(2610),
    [anon_sym_DASH] = ACTIONS(2608),
    [anon_sym_PLUS] = ACTIONS(2608),
    [anon_sym_STAR] = ACTIONS(2610),
    [anon_sym_CARET] = ACTIONS(2610),
    [anon_sym_AMP] = ACTIONS(2610),
    [anon_sym_SEMI] = ACTIONS(2610),
    [anon_sym___extension__] = ACTIONS(2608),
    [anon_sym_typedef] = ACTIONS(2608),
    [anon_sym_extern] = ACTIONS(2608),
    [anon_sym___attribute__] = ACTIONS(2608),
    [anon_sym___attribute] = ACTIONS(2608),
    [anon_sym_noreturn] = ACTIONS(2608),
    [anon_sym_LBRACK] = ACTIONS(2610),
    [anon_sym___declspec] = ACTIONS(2608),
    [anon_sym___cdecl] = ACTIONS(2608),
    [anon_sym___clrcall] = ACTIONS(2608),
    [anon_sym___stdcall] = ACTIONS(2608),
    [anon_sym___fastcall] = ACTIONS(2608),
    [anon_sym___thiscall] = ACTIONS(2608),
    [anon_sym___vectorcall] = ACTIONS(2608),
    [anon_sym_LBRACE] = ACTIONS(2610),
    [anon_sym_signed] = ACTIONS(2608),
    [anon_sym_unsigned] = ACTIONS(2608),
    [anon_sym_long] = ACTIONS(2608),
    [anon_sym_short] = ACTIONS(2608),
    [anon_sym_static] = ACTIONS(2608),
    [anon_sym_ATautoreleasepool] = ACTIONS(2610),
    [anon_sym_auto] = ACTIONS(2608),
    [anon_sym_register] = ACTIONS(2608),
    [anon_sym_inline] = ACTIONS(2608),
    [anon_sym___inline] = ACTIONS(2608),
    [anon_sym___inline__] = ACTIONS(2608),
    [anon_sym___forceinline] = ACTIONS(2608),
    [anon_sym_thread_local] = ACTIONS(2608),
    [anon_sym___thread] = ACTIONS(2608),
    [anon_sym_CG_EXTERN] = ACTIONS(2608),
    [anon_sym_CG_INLINE] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2608),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2608),
    [anon_sym_IBOutlet] = ACTIONS(2608),
    [anon_sym_IBInspectable] = ACTIONS(2608),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2608),
    [anon_sym_NS_INLINE] = ACTIONS(2608),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2608),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2608),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2608),
    [anon_sym_const] = ACTIONS(2608),
    [anon_sym_constexpr] = ACTIONS(2608),
    [anon_sym_volatile] = ACTIONS(2608),
    [anon_sym_restrict] = ACTIONS(2608),
    [anon_sym___restrict__] = ACTIONS(2608),
    [anon_sym__Atomic] = ACTIONS(2608),
    [anon_sym__Noreturn] = ACTIONS(2608),
    [anon_sym__Nonnull] = ACTIONS(2608),
    [anon_sym_nullable] = ACTIONS(2608),
    [anon_sym__Complex] = ACTIONS(2608),
    [anon_sym__Nullable] = ACTIONS(2608),
    [anon_sym__Nullable_result] = ACTIONS(2608),
    [anon_sym__Null_unspecified] = ACTIONS(2608),
    [anon_sym___autoreleasing] = ACTIONS(2608),
    [anon_sym___block] = ACTIONS(2608),
    [anon_sym___bridge] = ACTIONS(2608),
    [anon_sym___bridge_retained] = ACTIONS(2608),
    [anon_sym___bridge_transfer] = ACTIONS(2608),
    [anon_sym___complex] = ACTIONS(2608),
    [anon_sym___const] = ACTIONS(2608),
    [anon_sym___imag] = ACTIONS(2608),
    [anon_sym___kindof] = ACTIONS(2608),
    [anon_sym___nonnull] = ACTIONS(2608),
    [anon_sym___nullable] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2608),
    [anon_sym___real] = ACTIONS(2608),
    [anon_sym___strong] = ACTIONS(2608),
    [anon_sym___unsafe_unretained] = ACTIONS(2608),
    [anon_sym___unused] = ACTIONS(2608),
    [anon_sym___weak] = ACTIONS(2608),
    [anon_sym_alignas] = ACTIONS(2608),
    [anon_sym__Alignas] = ACTIONS(2608),
    [sym_primitive_type] = ACTIONS(2608),
    [anon_sym_enum] = ACTIONS(2608),
    [anon_sym_struct] = ACTIONS(2608),
    [anon_sym_union] = ACTIONS(2608),
    [anon_sym_if] = ACTIONS(2608),
    [anon_sym_switch] = ACTIONS(2608),
    [anon_sym_case] = ACTIONS(2608),
    [anon_sym_default] = ACTIONS(2608),
    [anon_sym_while] = ACTIONS(2608),
    [anon_sym_do] = ACTIONS(2608),
    [anon_sym_for] = ACTIONS(2608),
    [anon_sym_in] = ACTIONS(2608),
    [anon_sym_return] = ACTIONS(2608),
    [anon_sym_break] = ACTIONS(2608),
    [anon_sym_continue] = ACTIONS(2608),
    [anon_sym_goto] = ACTIONS(2608),
    [anon_sym_DASH_DASH] = ACTIONS(2610),
    [anon_sym_PLUS_PLUS] = ACTIONS(2610),
    [anon_sym_sizeof] = ACTIONS(2608),
    [anon_sym___alignof__] = ACTIONS(2608),
    [anon_sym___alignof] = ACTIONS(2608),
    [anon_sym__alignof] = ACTIONS(2608),
    [anon_sym_alignof] = ACTIONS(2608),
    [anon_sym__Alignof] = ACTIONS(2608),
    [anon_sym_offsetof] = ACTIONS(2608),
    [anon_sym__Generic] = ACTIONS(2608),
    [anon_sym_asm] = ACTIONS(2608),
    [anon_sym___asm__] = ACTIONS(2608),
    [anon_sym___asm] = ACTIONS(2608),
    [sym_number_literal] = ACTIONS(2610),
    [anon_sym_L_SQUOTE] = ACTIONS(2610),
    [anon_sym_u_SQUOTE] = ACTIONS(2610),
    [anon_sym_U_SQUOTE] = ACTIONS(2610),
    [anon_sym_u8_SQUOTE] = ACTIONS(2610),
    [anon_sym_SQUOTE] = ACTIONS(2610),
    [anon_sym_AT] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [anon_sym_L_DQUOTE] = ACTIONS(2610),
    [anon_sym_u_DQUOTE] = ACTIONS(2610),
    [anon_sym_U_DQUOTE] = ACTIONS(2610),
    [anon_sym_u8_DQUOTE] = ACTIONS(2610),
    [sym_true] = ACTIONS(2608),
    [sym_false] = ACTIONS(2608),
    [anon_sym_NULL] = ACTIONS(2608),
    [anon_sym_nullptr] = ACTIONS(2608),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2608),
    [anon_sym___typeof] = ACTIONS(2608),
    [anon_sym_typeof] = ACTIONS(2608),
    [anon_sym_ATimport] = ACTIONS(2610),
    [aux_sym_preproc_undef_token1] = ACTIONS(2608),
    [anon_sym_POUND] = ACTIONS(2608),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2608),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2608),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2608),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_API_AVAILABLE] = ACTIONS(2608),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_API_DEPRECATED] = ACTIONS(2608),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym___deprecated_msg] = ACTIONS(2608),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2608),
    [anon_sym_ATprotocol] = ACTIONS(2610),
    [anon_sym_ATinterface] = ACTIONS(2610),
    [anon_sym_ATimplementation] = ACTIONS(2610),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2610),
    [anon_sym_ATselector] = ACTIONS(2610),
    [anon_sym_ATavailable] = ACTIONS(2610),
    [anon_sym___builtin_available] = ACTIONS(2608),
    [anon_sym_va_arg] = ACTIONS(2608),
    [anon_sym_ATencode] = ACTIONS(2610),
    [anon_sym_BOOL] = ACTIONS(2608),
    [anon_sym_IMP] = ACTIONS(2608),
    [anon_sym_SEL] = ACTIONS(2608),
    [anon_sym_Class] = ACTIONS(2608),
    [anon_sym_id] = ACTIONS(2608),
  },
  [STATE(1029)] = {
    [ts_builtin_sym_end] = ACTIONS(2948),
    [sym_identifier] = ACTIONS(2950),
    [aux_sym_preproc_include_token1] = ACTIONS(2950),
    [aux_sym_preproc_include_token2] = ACTIONS(2950),
    [aux_sym_preproc_def_token1] = ACTIONS(2950),
    [aux_sym_preproc_if_token1] = ACTIONS(2950),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2950),
    [sym_preproc_directive] = ACTIONS(2950),
    [anon_sym_LPAREN2] = ACTIONS(2948),
    [anon_sym_BANG] = ACTIONS(2948),
    [anon_sym_TILDE] = ACTIONS(2948),
    [anon_sym_DASH] = ACTIONS(2950),
    [anon_sym_PLUS] = ACTIONS(2950),
    [anon_sym_STAR] = ACTIONS(2948),
    [anon_sym_CARET] = ACTIONS(2948),
    [anon_sym_AMP] = ACTIONS(2948),
    [anon_sym_SEMI] = ACTIONS(2948),
    [anon_sym___extension__] = ACTIONS(2950),
    [anon_sym_typedef] = ACTIONS(2950),
    [anon_sym_extern] = ACTIONS(2950),
    [anon_sym___attribute__] = ACTIONS(2950),
    [anon_sym___attribute] = ACTIONS(2950),
    [anon_sym_noreturn] = ACTIONS(2950),
    [anon_sym_LBRACK] = ACTIONS(2948),
    [anon_sym___declspec] = ACTIONS(2950),
    [anon_sym___cdecl] = ACTIONS(2950),
    [anon_sym___clrcall] = ACTIONS(2950),
    [anon_sym___stdcall] = ACTIONS(2950),
    [anon_sym___fastcall] = ACTIONS(2950),
    [anon_sym___thiscall] = ACTIONS(2950),
    [anon_sym___vectorcall] = ACTIONS(2950),
    [anon_sym_LBRACE] = ACTIONS(2948),
    [anon_sym_signed] = ACTIONS(2950),
    [anon_sym_unsigned] = ACTIONS(2950),
    [anon_sym_long] = ACTIONS(2950),
    [anon_sym_short] = ACTIONS(2950),
    [anon_sym_static] = ACTIONS(2950),
    [anon_sym_ATautoreleasepool] = ACTIONS(2948),
    [anon_sym_auto] = ACTIONS(2950),
    [anon_sym_register] = ACTIONS(2950),
    [anon_sym_inline] = ACTIONS(2950),
    [anon_sym___inline] = ACTIONS(2950),
    [anon_sym___inline__] = ACTIONS(2950),
    [anon_sym___forceinline] = ACTIONS(2950),
    [anon_sym_thread_local] = ACTIONS(2950),
    [anon_sym___thread] = ACTIONS(2950),
    [anon_sym_CG_EXTERN] = ACTIONS(2950),
    [anon_sym_CG_INLINE] = ACTIONS(2950),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2950),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2950),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2950),
    [anon_sym_IBOutlet] = ACTIONS(2950),
    [anon_sym_IBInspectable] = ACTIONS(2950),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2950),
    [anon_sym_NS_INLINE] = ACTIONS(2950),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2950),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2950),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2950),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2950),
    [anon_sym_const] = ACTIONS(2950),
    [anon_sym_constexpr] = ACTIONS(2950),
    [anon_sym_volatile] = ACTIONS(2950),
    [anon_sym_restrict] = ACTIONS(2950),
    [anon_sym___restrict__] = ACTIONS(2950),
    [anon_sym__Atomic] = ACTIONS(2950),
    [anon_sym__Noreturn] = ACTIONS(2950),
    [anon_sym__Nonnull] = ACTIONS(2950),
    [anon_sym_nullable] = ACTIONS(2950),
    [anon_sym__Complex] = ACTIONS(2950),
    [anon_sym__Nullable] = ACTIONS(2950),
    [anon_sym__Nullable_result] = ACTIONS(2950),
    [anon_sym__Null_unspecified] = ACTIONS(2950),
    [anon_sym___autoreleasing] = ACTIONS(2950),
    [anon_sym___block] = ACTIONS(2950),
    [anon_sym___bridge] = ACTIONS(2950),
    [anon_sym___bridge_retained] = ACTIONS(2950),
    [anon_sym___bridge_transfer] = ACTIONS(2950),
    [anon_sym___complex] = ACTIONS(2950),
    [anon_sym___const] = ACTIONS(2950),
    [anon_sym___imag] = ACTIONS(2950),
    [anon_sym___kindof] = ACTIONS(2950),
    [anon_sym___nonnull] = ACTIONS(2950),
    [anon_sym___nullable] = ACTIONS(2950),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2950),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2950),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2950),
    [anon_sym___real] = ACTIONS(2950),
    [anon_sym___strong] = ACTIONS(2950),
    [anon_sym___unsafe_unretained] = ACTIONS(2950),
    [anon_sym___unused] = ACTIONS(2950),
    [anon_sym___weak] = ACTIONS(2950),
    [anon_sym_alignas] = ACTIONS(2950),
    [anon_sym__Alignas] = ACTIONS(2950),
    [sym_primitive_type] = ACTIONS(2950),
    [anon_sym_enum] = ACTIONS(2950),
    [anon_sym_struct] = ACTIONS(2950),
    [anon_sym_union] = ACTIONS(2950),
    [anon_sym_if] = ACTIONS(2950),
    [anon_sym_switch] = ACTIONS(2950),
    [anon_sym_case] = ACTIONS(2950),
    [anon_sym_default] = ACTIONS(2950),
    [anon_sym_while] = ACTIONS(2950),
    [anon_sym_do] = ACTIONS(2950),
    [anon_sym_for] = ACTIONS(2950),
    [anon_sym_in] = ACTIONS(2950),
    [anon_sym_return] = ACTIONS(2950),
    [anon_sym_break] = ACTIONS(2950),
    [anon_sym_continue] = ACTIONS(2950),
    [anon_sym_goto] = ACTIONS(2950),
    [anon_sym_DASH_DASH] = ACTIONS(2948),
    [anon_sym_PLUS_PLUS] = ACTIONS(2948),
    [anon_sym_sizeof] = ACTIONS(2950),
    [anon_sym___alignof__] = ACTIONS(2950),
    [anon_sym___alignof] = ACTIONS(2950),
    [anon_sym__alignof] = ACTIONS(2950),
    [anon_sym_alignof] = ACTIONS(2950),
    [anon_sym__Alignof] = ACTIONS(2950),
    [anon_sym_offsetof] = ACTIONS(2950),
    [anon_sym__Generic] = ACTIONS(2950),
    [anon_sym_asm] = ACTIONS(2950),
    [anon_sym___asm__] = ACTIONS(2950),
    [anon_sym___asm] = ACTIONS(2950),
    [sym_number_literal] = ACTIONS(2948),
    [anon_sym_L_SQUOTE] = ACTIONS(2948),
    [anon_sym_u_SQUOTE] = ACTIONS(2948),
    [anon_sym_U_SQUOTE] = ACTIONS(2948),
    [anon_sym_u8_SQUOTE] = ACTIONS(2948),
    [anon_sym_SQUOTE] = ACTIONS(2948),
    [anon_sym_AT] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2948),
    [anon_sym_L_DQUOTE] = ACTIONS(2948),
    [anon_sym_u_DQUOTE] = ACTIONS(2948),
    [anon_sym_U_DQUOTE] = ACTIONS(2948),
    [anon_sym_u8_DQUOTE] = ACTIONS(2948),
    [sym_true] = ACTIONS(2950),
    [sym_false] = ACTIONS(2950),
    [anon_sym_NULL] = ACTIONS(2950),
    [anon_sym_nullptr] = ACTIONS(2950),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2950),
    [anon_sym___typeof] = ACTIONS(2950),
    [anon_sym_typeof] = ACTIONS(2950),
    [anon_sym_ATimport] = ACTIONS(2948),
    [aux_sym_preproc_undef_token1] = ACTIONS(2950),
    [anon_sym_POUND] = ACTIONS(2950),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2950),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2950),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2950),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2950),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2950),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2950),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2950),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2950),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2950),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2950),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2950),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2950),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2950),
    [anon_sym_API_AVAILABLE] = ACTIONS(2950),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2950),
    [anon_sym_API_DEPRECATED] = ACTIONS(2950),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2950),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2950),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2950),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2950),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2950),
    [anon_sym___deprecated_msg] = ACTIONS(2950),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2950),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2950),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2950),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2950),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2950),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2950),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2950),
    [anon_sym_ATprotocol] = ACTIONS(2948),
    [anon_sym_ATinterface] = ACTIONS(2948),
    [anon_sym_ATimplementation] = ACTIONS(2948),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2948),
    [anon_sym_ATselector] = ACTIONS(2948),
    [anon_sym_ATavailable] = ACTIONS(2948),
    [anon_sym___builtin_available] = ACTIONS(2950),
    [anon_sym_va_arg] = ACTIONS(2950),
    [anon_sym_ATencode] = ACTIONS(2948),
    [anon_sym_BOOL] = ACTIONS(2950),
    [anon_sym_IMP] = ACTIONS(2950),
    [anon_sym_SEL] = ACTIONS(2950),
    [anon_sym_Class] = ACTIONS(2950),
    [anon_sym_id] = ACTIONS(2950),
  },
  [STATE(1030)] = {
    [ts_builtin_sym_end] = ACTIONS(2706),
    [sym_identifier] = ACTIONS(2704),
    [aux_sym_preproc_include_token1] = ACTIONS(2704),
    [aux_sym_preproc_include_token2] = ACTIONS(2704),
    [aux_sym_preproc_def_token1] = ACTIONS(2704),
    [aux_sym_preproc_if_token1] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2704),
    [sym_preproc_directive] = ACTIONS(2704),
    [anon_sym_LPAREN2] = ACTIONS(2706),
    [anon_sym_BANG] = ACTIONS(2706),
    [anon_sym_TILDE] = ACTIONS(2706),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_STAR] = ACTIONS(2706),
    [anon_sym_CARET] = ACTIONS(2706),
    [anon_sym_AMP] = ACTIONS(2706),
    [anon_sym_SEMI] = ACTIONS(2706),
    [anon_sym___extension__] = ACTIONS(2704),
    [anon_sym_typedef] = ACTIONS(2704),
    [anon_sym_extern] = ACTIONS(2704),
    [anon_sym___attribute__] = ACTIONS(2704),
    [anon_sym___attribute] = ACTIONS(2704),
    [anon_sym_noreturn] = ACTIONS(2704),
    [anon_sym_LBRACK] = ACTIONS(2706),
    [anon_sym___declspec] = ACTIONS(2704),
    [anon_sym___cdecl] = ACTIONS(2704),
    [anon_sym___clrcall] = ACTIONS(2704),
    [anon_sym___stdcall] = ACTIONS(2704),
    [anon_sym___fastcall] = ACTIONS(2704),
    [anon_sym___thiscall] = ACTIONS(2704),
    [anon_sym___vectorcall] = ACTIONS(2704),
    [anon_sym_LBRACE] = ACTIONS(2706),
    [anon_sym_signed] = ACTIONS(2704),
    [anon_sym_unsigned] = ACTIONS(2704),
    [anon_sym_long] = ACTIONS(2704),
    [anon_sym_short] = ACTIONS(2704),
    [anon_sym_static] = ACTIONS(2704),
    [anon_sym_ATautoreleasepool] = ACTIONS(2706),
    [anon_sym_auto] = ACTIONS(2704),
    [anon_sym_register] = ACTIONS(2704),
    [anon_sym_inline] = ACTIONS(2704),
    [anon_sym___inline] = ACTIONS(2704),
    [anon_sym___inline__] = ACTIONS(2704),
    [anon_sym___forceinline] = ACTIONS(2704),
    [anon_sym_thread_local] = ACTIONS(2704),
    [anon_sym___thread] = ACTIONS(2704),
    [anon_sym_CG_EXTERN] = ACTIONS(2704),
    [anon_sym_CG_INLINE] = ACTIONS(2704),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2704),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2704),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2704),
    [anon_sym_IBOutlet] = ACTIONS(2704),
    [anon_sym_IBInspectable] = ACTIONS(2704),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2704),
    [anon_sym_NS_INLINE] = ACTIONS(2704),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2704),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2704),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2704),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2704),
    [anon_sym_const] = ACTIONS(2704),
    [anon_sym_constexpr] = ACTIONS(2704),
    [anon_sym_volatile] = ACTIONS(2704),
    [anon_sym_restrict] = ACTIONS(2704),
    [anon_sym___restrict__] = ACTIONS(2704),
    [anon_sym__Atomic] = ACTIONS(2704),
    [anon_sym__Noreturn] = ACTIONS(2704),
    [anon_sym__Nonnull] = ACTIONS(2704),
    [anon_sym_nullable] = ACTIONS(2704),
    [anon_sym__Complex] = ACTIONS(2704),
    [anon_sym__Nullable] = ACTIONS(2704),
    [anon_sym__Nullable_result] = ACTIONS(2704),
    [anon_sym__Null_unspecified] = ACTIONS(2704),
    [anon_sym___autoreleasing] = ACTIONS(2704),
    [anon_sym___block] = ACTIONS(2704),
    [anon_sym___bridge] = ACTIONS(2704),
    [anon_sym___bridge_retained] = ACTIONS(2704),
    [anon_sym___bridge_transfer] = ACTIONS(2704),
    [anon_sym___complex] = ACTIONS(2704),
    [anon_sym___const] = ACTIONS(2704),
    [anon_sym___imag] = ACTIONS(2704),
    [anon_sym___kindof] = ACTIONS(2704),
    [anon_sym___nonnull] = ACTIONS(2704),
    [anon_sym___nullable] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2704),
    [anon_sym___real] = ACTIONS(2704),
    [anon_sym___strong] = ACTIONS(2704),
    [anon_sym___unsafe_unretained] = ACTIONS(2704),
    [anon_sym___unused] = ACTIONS(2704),
    [anon_sym___weak] = ACTIONS(2704),
    [anon_sym_alignas] = ACTIONS(2704),
    [anon_sym__Alignas] = ACTIONS(2704),
    [sym_primitive_type] = ACTIONS(2704),
    [anon_sym_enum] = ACTIONS(2704),
    [anon_sym_struct] = ACTIONS(2704),
    [anon_sym_union] = ACTIONS(2704),
    [anon_sym_if] = ACTIONS(2704),
    [anon_sym_switch] = ACTIONS(2704),
    [anon_sym_case] = ACTIONS(2704),
    [anon_sym_default] = ACTIONS(2704),
    [anon_sym_while] = ACTIONS(2704),
    [anon_sym_do] = ACTIONS(2704),
    [anon_sym_for] = ACTIONS(2704),
    [anon_sym_in] = ACTIONS(2704),
    [anon_sym_return] = ACTIONS(2704),
    [anon_sym_break] = ACTIONS(2704),
    [anon_sym_continue] = ACTIONS(2704),
    [anon_sym_goto] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2706),
    [anon_sym_PLUS_PLUS] = ACTIONS(2706),
    [anon_sym_sizeof] = ACTIONS(2704),
    [anon_sym___alignof__] = ACTIONS(2704),
    [anon_sym___alignof] = ACTIONS(2704),
    [anon_sym__alignof] = ACTIONS(2704),
    [anon_sym_alignof] = ACTIONS(2704),
    [anon_sym__Alignof] = ACTIONS(2704),
    [anon_sym_offsetof] = ACTIONS(2704),
    [anon_sym__Generic] = ACTIONS(2704),
    [anon_sym_asm] = ACTIONS(2704),
    [anon_sym___asm__] = ACTIONS(2704),
    [anon_sym___asm] = ACTIONS(2704),
    [sym_number_literal] = ACTIONS(2706),
    [anon_sym_L_SQUOTE] = ACTIONS(2706),
    [anon_sym_u_SQUOTE] = ACTIONS(2706),
    [anon_sym_U_SQUOTE] = ACTIONS(2706),
    [anon_sym_u8_SQUOTE] = ACTIONS(2706),
    [anon_sym_SQUOTE] = ACTIONS(2706),
    [anon_sym_AT] = ACTIONS(2704),
    [anon_sym_DQUOTE] = ACTIONS(2706),
    [anon_sym_L_DQUOTE] = ACTIONS(2706),
    [anon_sym_u_DQUOTE] = ACTIONS(2706),
    [anon_sym_U_DQUOTE] = ACTIONS(2706),
    [anon_sym_u8_DQUOTE] = ACTIONS(2706),
    [sym_true] = ACTIONS(2704),
    [sym_false] = ACTIONS(2704),
    [anon_sym_NULL] = ACTIONS(2704),
    [anon_sym_nullptr] = ACTIONS(2704),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2704),
    [anon_sym___typeof] = ACTIONS(2704),
    [anon_sym_typeof] = ACTIONS(2704),
    [anon_sym_ATimport] = ACTIONS(2706),
    [aux_sym_preproc_undef_token1] = ACTIONS(2704),
    [anon_sym_POUND] = ACTIONS(2704),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2704),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2704),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2704),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2704),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2704),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2704),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2704),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_API_AVAILABLE] = ACTIONS(2704),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_API_DEPRECATED] = ACTIONS(2704),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2704),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2704),
    [anon_sym___deprecated_msg] = ACTIONS(2704),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2704),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2704),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2704),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2704),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2704),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2704),
    [anon_sym_ATprotocol] = ACTIONS(2706),
    [anon_sym_ATinterface] = ACTIONS(2706),
    [anon_sym_ATimplementation] = ACTIONS(2706),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2706),
    [anon_sym_ATselector] = ACTIONS(2706),
    [anon_sym_ATavailable] = ACTIONS(2706),
    [anon_sym___builtin_available] = ACTIONS(2704),
    [anon_sym_va_arg] = ACTIONS(2704),
    [anon_sym_ATencode] = ACTIONS(2706),
    [anon_sym_BOOL] = ACTIONS(2704),
    [anon_sym_IMP] = ACTIONS(2704),
    [anon_sym_SEL] = ACTIONS(2704),
    [anon_sym_Class] = ACTIONS(2704),
    [anon_sym_id] = ACTIONS(2704),
  },
  [STATE(1031)] = {
    [ts_builtin_sym_end] = ACTIONS(2636),
    [sym_identifier] = ACTIONS(2634),
    [aux_sym_preproc_include_token1] = ACTIONS(2634),
    [aux_sym_preproc_include_token2] = ACTIONS(2634),
    [aux_sym_preproc_def_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2634),
    [sym_preproc_directive] = ACTIONS(2634),
    [anon_sym_LPAREN2] = ACTIONS(2636),
    [anon_sym_BANG] = ACTIONS(2636),
    [anon_sym_TILDE] = ACTIONS(2636),
    [anon_sym_DASH] = ACTIONS(2634),
    [anon_sym_PLUS] = ACTIONS(2634),
    [anon_sym_STAR] = ACTIONS(2636),
    [anon_sym_CARET] = ACTIONS(2636),
    [anon_sym_AMP] = ACTIONS(2636),
    [anon_sym_SEMI] = ACTIONS(2636),
    [anon_sym___extension__] = ACTIONS(2634),
    [anon_sym_typedef] = ACTIONS(2634),
    [anon_sym_extern] = ACTIONS(2634),
    [anon_sym___attribute__] = ACTIONS(2634),
    [anon_sym___attribute] = ACTIONS(2634),
    [anon_sym_noreturn] = ACTIONS(2634),
    [anon_sym_LBRACK] = ACTIONS(2636),
    [anon_sym___declspec] = ACTIONS(2634),
    [anon_sym___cdecl] = ACTIONS(2634),
    [anon_sym___clrcall] = ACTIONS(2634),
    [anon_sym___stdcall] = ACTIONS(2634),
    [anon_sym___fastcall] = ACTIONS(2634),
    [anon_sym___thiscall] = ACTIONS(2634),
    [anon_sym___vectorcall] = ACTIONS(2634),
    [anon_sym_LBRACE] = ACTIONS(2636),
    [anon_sym_signed] = ACTIONS(2634),
    [anon_sym_unsigned] = ACTIONS(2634),
    [anon_sym_long] = ACTIONS(2634),
    [anon_sym_short] = ACTIONS(2634),
    [anon_sym_static] = ACTIONS(2634),
    [anon_sym_ATautoreleasepool] = ACTIONS(2636),
    [anon_sym_auto] = ACTIONS(2634),
    [anon_sym_register] = ACTIONS(2634),
    [anon_sym_inline] = ACTIONS(2634),
    [anon_sym___inline] = ACTIONS(2634),
    [anon_sym___inline__] = ACTIONS(2634),
    [anon_sym___forceinline] = ACTIONS(2634),
    [anon_sym_thread_local] = ACTIONS(2634),
    [anon_sym___thread] = ACTIONS(2634),
    [anon_sym_CG_EXTERN] = ACTIONS(2634),
    [anon_sym_CG_INLINE] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2634),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2634),
    [anon_sym_IBOutlet] = ACTIONS(2634),
    [anon_sym_IBInspectable] = ACTIONS(2634),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2634),
    [anon_sym_NS_INLINE] = ACTIONS(2634),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2634),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2634),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2634),
    [anon_sym_const] = ACTIONS(2634),
    [anon_sym_constexpr] = ACTIONS(2634),
    [anon_sym_volatile] = ACTIONS(2634),
    [anon_sym_restrict] = ACTIONS(2634),
    [anon_sym___restrict__] = ACTIONS(2634),
    [anon_sym__Atomic] = ACTIONS(2634),
    [anon_sym__Noreturn] = ACTIONS(2634),
    [anon_sym__Nonnull] = ACTIONS(2634),
    [anon_sym_nullable] = ACTIONS(2634),
    [anon_sym__Complex] = ACTIONS(2634),
    [anon_sym__Nullable] = ACTIONS(2634),
    [anon_sym__Nullable_result] = ACTIONS(2634),
    [anon_sym__Null_unspecified] = ACTIONS(2634),
    [anon_sym___autoreleasing] = ACTIONS(2634),
    [anon_sym___block] = ACTIONS(2634),
    [anon_sym___bridge] = ACTIONS(2634),
    [anon_sym___bridge_retained] = ACTIONS(2634),
    [anon_sym___bridge_transfer] = ACTIONS(2634),
    [anon_sym___complex] = ACTIONS(2634),
    [anon_sym___const] = ACTIONS(2634),
    [anon_sym___imag] = ACTIONS(2634),
    [anon_sym___kindof] = ACTIONS(2634),
    [anon_sym___nonnull] = ACTIONS(2634),
    [anon_sym___nullable] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2634),
    [anon_sym___real] = ACTIONS(2634),
    [anon_sym___strong] = ACTIONS(2634),
    [anon_sym___unsafe_unretained] = ACTIONS(2634),
    [anon_sym___unused] = ACTIONS(2634),
    [anon_sym___weak] = ACTIONS(2634),
    [anon_sym_alignas] = ACTIONS(2634),
    [anon_sym__Alignas] = ACTIONS(2634),
    [sym_primitive_type] = ACTIONS(2634),
    [anon_sym_enum] = ACTIONS(2634),
    [anon_sym_struct] = ACTIONS(2634),
    [anon_sym_union] = ACTIONS(2634),
    [anon_sym_if] = ACTIONS(2634),
    [anon_sym_switch] = ACTIONS(2634),
    [anon_sym_case] = ACTIONS(2634),
    [anon_sym_default] = ACTIONS(2634),
    [anon_sym_while] = ACTIONS(2634),
    [anon_sym_do] = ACTIONS(2634),
    [anon_sym_for] = ACTIONS(2634),
    [anon_sym_in] = ACTIONS(2634),
    [anon_sym_return] = ACTIONS(2634),
    [anon_sym_break] = ACTIONS(2634),
    [anon_sym_continue] = ACTIONS(2634),
    [anon_sym_goto] = ACTIONS(2634),
    [anon_sym_DASH_DASH] = ACTIONS(2636),
    [anon_sym_PLUS_PLUS] = ACTIONS(2636),
    [anon_sym_sizeof] = ACTIONS(2634),
    [anon_sym___alignof__] = ACTIONS(2634),
    [anon_sym___alignof] = ACTIONS(2634),
    [anon_sym__alignof] = ACTIONS(2634),
    [anon_sym_alignof] = ACTIONS(2634),
    [anon_sym__Alignof] = ACTIONS(2634),
    [anon_sym_offsetof] = ACTIONS(2634),
    [anon_sym__Generic] = ACTIONS(2634),
    [anon_sym_asm] = ACTIONS(2634),
    [anon_sym___asm__] = ACTIONS(2634),
    [anon_sym___asm] = ACTIONS(2634),
    [sym_number_literal] = ACTIONS(2636),
    [anon_sym_L_SQUOTE] = ACTIONS(2636),
    [anon_sym_u_SQUOTE] = ACTIONS(2636),
    [anon_sym_U_SQUOTE] = ACTIONS(2636),
    [anon_sym_u8_SQUOTE] = ACTIONS(2636),
    [anon_sym_SQUOTE] = ACTIONS(2636),
    [anon_sym_AT] = ACTIONS(2634),
    [anon_sym_DQUOTE] = ACTIONS(2636),
    [anon_sym_L_DQUOTE] = ACTIONS(2636),
    [anon_sym_u_DQUOTE] = ACTIONS(2636),
    [anon_sym_U_DQUOTE] = ACTIONS(2636),
    [anon_sym_u8_DQUOTE] = ACTIONS(2636),
    [sym_true] = ACTIONS(2634),
    [sym_false] = ACTIONS(2634),
    [anon_sym_NULL] = ACTIONS(2634),
    [anon_sym_nullptr] = ACTIONS(2634),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2634),
    [anon_sym___typeof] = ACTIONS(2634),
    [anon_sym_typeof] = ACTIONS(2634),
    [anon_sym_ATimport] = ACTIONS(2636),
    [aux_sym_preproc_undef_token1] = ACTIONS(2634),
    [anon_sym_POUND] = ACTIONS(2634),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2634),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2634),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2634),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_API_AVAILABLE] = ACTIONS(2634),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_API_DEPRECATED] = ACTIONS(2634),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym___deprecated_msg] = ACTIONS(2634),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2634),
    [anon_sym_ATprotocol] = ACTIONS(2636),
    [anon_sym_ATinterface] = ACTIONS(2636),
    [anon_sym_ATimplementation] = ACTIONS(2636),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2636),
    [anon_sym_ATselector] = ACTIONS(2636),
    [anon_sym_ATavailable] = ACTIONS(2636),
    [anon_sym___builtin_available] = ACTIONS(2634),
    [anon_sym_va_arg] = ACTIONS(2634),
    [anon_sym_ATencode] = ACTIONS(2636),
    [anon_sym_BOOL] = ACTIONS(2634),
    [anon_sym_IMP] = ACTIONS(2634),
    [anon_sym_SEL] = ACTIONS(2634),
    [anon_sym_Class] = ACTIONS(2634),
    [anon_sym_id] = ACTIONS(2634),
  },
  [STATE(1032)] = {
    [ts_builtin_sym_end] = ACTIONS(2670),
    [sym_identifier] = ACTIONS(2668),
    [aux_sym_preproc_include_token1] = ACTIONS(2668),
    [aux_sym_preproc_include_token2] = ACTIONS(2668),
    [aux_sym_preproc_def_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2668),
    [sym_preproc_directive] = ACTIONS(2668),
    [anon_sym_LPAREN2] = ACTIONS(2670),
    [anon_sym_BANG] = ACTIONS(2670),
    [anon_sym_TILDE] = ACTIONS(2670),
    [anon_sym_DASH] = ACTIONS(2668),
    [anon_sym_PLUS] = ACTIONS(2668),
    [anon_sym_STAR] = ACTIONS(2670),
    [anon_sym_CARET] = ACTIONS(2670),
    [anon_sym_AMP] = ACTIONS(2670),
    [anon_sym_SEMI] = ACTIONS(2670),
    [anon_sym___extension__] = ACTIONS(2668),
    [anon_sym_typedef] = ACTIONS(2668),
    [anon_sym_extern] = ACTIONS(2668),
    [anon_sym___attribute__] = ACTIONS(2668),
    [anon_sym___attribute] = ACTIONS(2668),
    [anon_sym_noreturn] = ACTIONS(2668),
    [anon_sym_LBRACK] = ACTIONS(2670),
    [anon_sym___declspec] = ACTIONS(2668),
    [anon_sym___cdecl] = ACTIONS(2668),
    [anon_sym___clrcall] = ACTIONS(2668),
    [anon_sym___stdcall] = ACTIONS(2668),
    [anon_sym___fastcall] = ACTIONS(2668),
    [anon_sym___thiscall] = ACTIONS(2668),
    [anon_sym___vectorcall] = ACTIONS(2668),
    [anon_sym_LBRACE] = ACTIONS(2670),
    [anon_sym_signed] = ACTIONS(2668),
    [anon_sym_unsigned] = ACTIONS(2668),
    [anon_sym_long] = ACTIONS(2668),
    [anon_sym_short] = ACTIONS(2668),
    [anon_sym_static] = ACTIONS(2668),
    [anon_sym_ATautoreleasepool] = ACTIONS(2670),
    [anon_sym_auto] = ACTIONS(2668),
    [anon_sym_register] = ACTIONS(2668),
    [anon_sym_inline] = ACTIONS(2668),
    [anon_sym___inline] = ACTIONS(2668),
    [anon_sym___inline__] = ACTIONS(2668),
    [anon_sym___forceinline] = ACTIONS(2668),
    [anon_sym_thread_local] = ACTIONS(2668),
    [anon_sym___thread] = ACTIONS(2668),
    [anon_sym_CG_EXTERN] = ACTIONS(2668),
    [anon_sym_CG_INLINE] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2668),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2668),
    [anon_sym_IBOutlet] = ACTIONS(2668),
    [anon_sym_IBInspectable] = ACTIONS(2668),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2668),
    [anon_sym_NS_INLINE] = ACTIONS(2668),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2668),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2668),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2668),
    [anon_sym_const] = ACTIONS(2668),
    [anon_sym_constexpr] = ACTIONS(2668),
    [anon_sym_volatile] = ACTIONS(2668),
    [anon_sym_restrict] = ACTIONS(2668),
    [anon_sym___restrict__] = ACTIONS(2668),
    [anon_sym__Atomic] = ACTIONS(2668),
    [anon_sym__Noreturn] = ACTIONS(2668),
    [anon_sym__Nonnull] = ACTIONS(2668),
    [anon_sym_nullable] = ACTIONS(2668),
    [anon_sym__Complex] = ACTIONS(2668),
    [anon_sym__Nullable] = ACTIONS(2668),
    [anon_sym__Nullable_result] = ACTIONS(2668),
    [anon_sym__Null_unspecified] = ACTIONS(2668),
    [anon_sym___autoreleasing] = ACTIONS(2668),
    [anon_sym___block] = ACTIONS(2668),
    [anon_sym___bridge] = ACTIONS(2668),
    [anon_sym___bridge_retained] = ACTIONS(2668),
    [anon_sym___bridge_transfer] = ACTIONS(2668),
    [anon_sym___complex] = ACTIONS(2668),
    [anon_sym___const] = ACTIONS(2668),
    [anon_sym___imag] = ACTIONS(2668),
    [anon_sym___kindof] = ACTIONS(2668),
    [anon_sym___nonnull] = ACTIONS(2668),
    [anon_sym___nullable] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2668),
    [anon_sym___real] = ACTIONS(2668),
    [anon_sym___strong] = ACTIONS(2668),
    [anon_sym___unsafe_unretained] = ACTIONS(2668),
    [anon_sym___unused] = ACTIONS(2668),
    [anon_sym___weak] = ACTIONS(2668),
    [anon_sym_alignas] = ACTIONS(2668),
    [anon_sym__Alignas] = ACTIONS(2668),
    [sym_primitive_type] = ACTIONS(2668),
    [anon_sym_enum] = ACTIONS(2668),
    [anon_sym_struct] = ACTIONS(2668),
    [anon_sym_union] = ACTIONS(2668),
    [anon_sym_if] = ACTIONS(2668),
    [anon_sym_switch] = ACTIONS(2668),
    [anon_sym_case] = ACTIONS(2668),
    [anon_sym_default] = ACTIONS(2668),
    [anon_sym_while] = ACTIONS(2668),
    [anon_sym_do] = ACTIONS(2668),
    [anon_sym_for] = ACTIONS(2668),
    [anon_sym_in] = ACTIONS(2668),
    [anon_sym_return] = ACTIONS(2668),
    [anon_sym_break] = ACTIONS(2668),
    [anon_sym_continue] = ACTIONS(2668),
    [anon_sym_goto] = ACTIONS(2668),
    [anon_sym_DASH_DASH] = ACTIONS(2670),
    [anon_sym_PLUS_PLUS] = ACTIONS(2670),
    [anon_sym_sizeof] = ACTIONS(2668),
    [anon_sym___alignof__] = ACTIONS(2668),
    [anon_sym___alignof] = ACTIONS(2668),
    [anon_sym__alignof] = ACTIONS(2668),
    [anon_sym_alignof] = ACTIONS(2668),
    [anon_sym__Alignof] = ACTIONS(2668),
    [anon_sym_offsetof] = ACTIONS(2668),
    [anon_sym__Generic] = ACTIONS(2668),
    [anon_sym_asm] = ACTIONS(2668),
    [anon_sym___asm__] = ACTIONS(2668),
    [anon_sym___asm] = ACTIONS(2668),
    [sym_number_literal] = ACTIONS(2670),
    [anon_sym_L_SQUOTE] = ACTIONS(2670),
    [anon_sym_u_SQUOTE] = ACTIONS(2670),
    [anon_sym_U_SQUOTE] = ACTIONS(2670),
    [anon_sym_u8_SQUOTE] = ACTIONS(2670),
    [anon_sym_SQUOTE] = ACTIONS(2670),
    [anon_sym_AT] = ACTIONS(2668),
    [anon_sym_DQUOTE] = ACTIONS(2670),
    [anon_sym_L_DQUOTE] = ACTIONS(2670),
    [anon_sym_u_DQUOTE] = ACTIONS(2670),
    [anon_sym_U_DQUOTE] = ACTIONS(2670),
    [anon_sym_u8_DQUOTE] = ACTIONS(2670),
    [sym_true] = ACTIONS(2668),
    [sym_false] = ACTIONS(2668),
    [anon_sym_NULL] = ACTIONS(2668),
    [anon_sym_nullptr] = ACTIONS(2668),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2668),
    [anon_sym___typeof] = ACTIONS(2668),
    [anon_sym_typeof] = ACTIONS(2668),
    [anon_sym_ATimport] = ACTIONS(2670),
    [aux_sym_preproc_undef_token1] = ACTIONS(2668),
    [anon_sym_POUND] = ACTIONS(2668),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2668),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2668),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2668),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_API_AVAILABLE] = ACTIONS(2668),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_API_DEPRECATED] = ACTIONS(2668),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym___deprecated_msg] = ACTIONS(2668),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2668),
    [anon_sym_ATprotocol] = ACTIONS(2670),
    [anon_sym_ATinterface] = ACTIONS(2670),
    [anon_sym_ATimplementation] = ACTIONS(2670),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2670),
    [anon_sym_ATselector] = ACTIONS(2670),
    [anon_sym_ATavailable] = ACTIONS(2670),
    [anon_sym___builtin_available] = ACTIONS(2668),
    [anon_sym_va_arg] = ACTIONS(2668),
    [anon_sym_ATencode] = ACTIONS(2670),
    [anon_sym_BOOL] = ACTIONS(2668),
    [anon_sym_IMP] = ACTIONS(2668),
    [anon_sym_SEL] = ACTIONS(2668),
    [anon_sym_Class] = ACTIONS(2668),
    [anon_sym_id] = ACTIONS(2668),
  },
  [STATE(1033)] = {
    [ts_builtin_sym_end] = ACTIONS(2414),
    [sym_identifier] = ACTIONS(2412),
    [aux_sym_preproc_include_token1] = ACTIONS(2412),
    [aux_sym_preproc_include_token2] = ACTIONS(2412),
    [aux_sym_preproc_def_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2412),
    [sym_preproc_directive] = ACTIONS(2412),
    [anon_sym_LPAREN2] = ACTIONS(2414),
    [anon_sym_BANG] = ACTIONS(2414),
    [anon_sym_TILDE] = ACTIONS(2414),
    [anon_sym_DASH] = ACTIONS(2412),
    [anon_sym_PLUS] = ACTIONS(2412),
    [anon_sym_STAR] = ACTIONS(2414),
    [anon_sym_CARET] = ACTIONS(2414),
    [anon_sym_AMP] = ACTIONS(2414),
    [anon_sym_SEMI] = ACTIONS(2414),
    [anon_sym___extension__] = ACTIONS(2412),
    [anon_sym_typedef] = ACTIONS(2412),
    [anon_sym_extern] = ACTIONS(2412),
    [anon_sym___attribute__] = ACTIONS(2412),
    [anon_sym___attribute] = ACTIONS(2412),
    [anon_sym_noreturn] = ACTIONS(2412),
    [anon_sym_LBRACK] = ACTIONS(2414),
    [anon_sym___declspec] = ACTIONS(2412),
    [anon_sym___cdecl] = ACTIONS(2412),
    [anon_sym___clrcall] = ACTIONS(2412),
    [anon_sym___stdcall] = ACTIONS(2412),
    [anon_sym___fastcall] = ACTIONS(2412),
    [anon_sym___thiscall] = ACTIONS(2412),
    [anon_sym___vectorcall] = ACTIONS(2412),
    [anon_sym_LBRACE] = ACTIONS(2414),
    [anon_sym_signed] = ACTIONS(2412),
    [anon_sym_unsigned] = ACTIONS(2412),
    [anon_sym_long] = ACTIONS(2412),
    [anon_sym_short] = ACTIONS(2412),
    [anon_sym_static] = ACTIONS(2412),
    [anon_sym_ATautoreleasepool] = ACTIONS(2414),
    [anon_sym_auto] = ACTIONS(2412),
    [anon_sym_register] = ACTIONS(2412),
    [anon_sym_inline] = ACTIONS(2412),
    [anon_sym___inline] = ACTIONS(2412),
    [anon_sym___inline__] = ACTIONS(2412),
    [anon_sym___forceinline] = ACTIONS(2412),
    [anon_sym_thread_local] = ACTIONS(2412),
    [anon_sym___thread] = ACTIONS(2412),
    [anon_sym_CG_EXTERN] = ACTIONS(2412),
    [anon_sym_CG_INLINE] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2412),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2412),
    [anon_sym_IBOutlet] = ACTIONS(2412),
    [anon_sym_IBInspectable] = ACTIONS(2412),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2412),
    [anon_sym_NS_INLINE] = ACTIONS(2412),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2412),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2412),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2412),
    [anon_sym_const] = ACTIONS(2412),
    [anon_sym_constexpr] = ACTIONS(2412),
    [anon_sym_volatile] = ACTIONS(2412),
    [anon_sym_restrict] = ACTIONS(2412),
    [anon_sym___restrict__] = ACTIONS(2412),
    [anon_sym__Atomic] = ACTIONS(2412),
    [anon_sym__Noreturn] = ACTIONS(2412),
    [anon_sym__Nonnull] = ACTIONS(2412),
    [anon_sym_nullable] = ACTIONS(2412),
    [anon_sym__Complex] = ACTIONS(2412),
    [anon_sym__Nullable] = ACTIONS(2412),
    [anon_sym__Nullable_result] = ACTIONS(2412),
    [anon_sym__Null_unspecified] = ACTIONS(2412),
    [anon_sym___autoreleasing] = ACTIONS(2412),
    [anon_sym___block] = ACTIONS(2412),
    [anon_sym___bridge] = ACTIONS(2412),
    [anon_sym___bridge_retained] = ACTIONS(2412),
    [anon_sym___bridge_transfer] = ACTIONS(2412),
    [anon_sym___complex] = ACTIONS(2412),
    [anon_sym___const] = ACTIONS(2412),
    [anon_sym___imag] = ACTIONS(2412),
    [anon_sym___kindof] = ACTIONS(2412),
    [anon_sym___nonnull] = ACTIONS(2412),
    [anon_sym___nullable] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2412),
    [anon_sym___real] = ACTIONS(2412),
    [anon_sym___strong] = ACTIONS(2412),
    [anon_sym___unsafe_unretained] = ACTIONS(2412),
    [anon_sym___unused] = ACTIONS(2412),
    [anon_sym___weak] = ACTIONS(2412),
    [anon_sym_alignas] = ACTIONS(2412),
    [anon_sym__Alignas] = ACTIONS(2412),
    [sym_primitive_type] = ACTIONS(2412),
    [anon_sym_enum] = ACTIONS(2412),
    [anon_sym_struct] = ACTIONS(2412),
    [anon_sym_union] = ACTIONS(2412),
    [anon_sym_if] = ACTIONS(2412),
    [anon_sym_switch] = ACTIONS(2412),
    [anon_sym_case] = ACTIONS(2412),
    [anon_sym_default] = ACTIONS(2412),
    [anon_sym_while] = ACTIONS(2412),
    [anon_sym_do] = ACTIONS(2412),
    [anon_sym_for] = ACTIONS(2412),
    [anon_sym_in] = ACTIONS(2412),
    [anon_sym_return] = ACTIONS(2412),
    [anon_sym_break] = ACTIONS(2412),
    [anon_sym_continue] = ACTIONS(2412),
    [anon_sym_goto] = ACTIONS(2412),
    [anon_sym_DASH_DASH] = ACTIONS(2414),
    [anon_sym_PLUS_PLUS] = ACTIONS(2414),
    [anon_sym_sizeof] = ACTIONS(2412),
    [anon_sym___alignof__] = ACTIONS(2412),
    [anon_sym___alignof] = ACTIONS(2412),
    [anon_sym__alignof] = ACTIONS(2412),
    [anon_sym_alignof] = ACTIONS(2412),
    [anon_sym__Alignof] = ACTIONS(2412),
    [anon_sym_offsetof] = ACTIONS(2412),
    [anon_sym__Generic] = ACTIONS(2412),
    [anon_sym_asm] = ACTIONS(2412),
    [anon_sym___asm__] = ACTIONS(2412),
    [anon_sym___asm] = ACTIONS(2412),
    [sym_number_literal] = ACTIONS(2414),
    [anon_sym_L_SQUOTE] = ACTIONS(2414),
    [anon_sym_u_SQUOTE] = ACTIONS(2414),
    [anon_sym_U_SQUOTE] = ACTIONS(2414),
    [anon_sym_u8_SQUOTE] = ACTIONS(2414),
    [anon_sym_SQUOTE] = ACTIONS(2414),
    [anon_sym_AT] = ACTIONS(2412),
    [anon_sym_DQUOTE] = ACTIONS(2414),
    [anon_sym_L_DQUOTE] = ACTIONS(2414),
    [anon_sym_u_DQUOTE] = ACTIONS(2414),
    [anon_sym_U_DQUOTE] = ACTIONS(2414),
    [anon_sym_u8_DQUOTE] = ACTIONS(2414),
    [sym_true] = ACTIONS(2412),
    [sym_false] = ACTIONS(2412),
    [anon_sym_NULL] = ACTIONS(2412),
    [anon_sym_nullptr] = ACTIONS(2412),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2412),
    [anon_sym___typeof] = ACTIONS(2412),
    [anon_sym_typeof] = ACTIONS(2412),
    [anon_sym_ATimport] = ACTIONS(2414),
    [aux_sym_preproc_undef_token1] = ACTIONS(2412),
    [anon_sym_POUND] = ACTIONS(2412),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2412),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2412),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2412),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_API_AVAILABLE] = ACTIONS(2412),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_API_DEPRECATED] = ACTIONS(2412),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym___deprecated_msg] = ACTIONS(2412),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2412),
    [anon_sym_ATprotocol] = ACTIONS(2414),
    [anon_sym_ATinterface] = ACTIONS(2414),
    [anon_sym_ATimplementation] = ACTIONS(2414),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2414),
    [anon_sym_ATselector] = ACTIONS(2414),
    [anon_sym_ATavailable] = ACTIONS(2414),
    [anon_sym___builtin_available] = ACTIONS(2412),
    [anon_sym_va_arg] = ACTIONS(2412),
    [anon_sym_ATencode] = ACTIONS(2414),
    [anon_sym_BOOL] = ACTIONS(2412),
    [anon_sym_IMP] = ACTIONS(2412),
    [anon_sym_SEL] = ACTIONS(2412),
    [anon_sym_Class] = ACTIONS(2412),
    [anon_sym_id] = ACTIONS(2412),
  },
  [STATE(1034)] = {
    [ts_builtin_sym_end] = ACTIONS(2710),
    [sym_identifier] = ACTIONS(2708),
    [aux_sym_preproc_include_token1] = ACTIONS(2708),
    [aux_sym_preproc_include_token2] = ACTIONS(2708),
    [aux_sym_preproc_def_token1] = ACTIONS(2708),
    [aux_sym_preproc_if_token1] = ACTIONS(2708),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2708),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2708),
    [sym_preproc_directive] = ACTIONS(2708),
    [anon_sym_LPAREN2] = ACTIONS(2710),
    [anon_sym_BANG] = ACTIONS(2710),
    [anon_sym_TILDE] = ACTIONS(2710),
    [anon_sym_DASH] = ACTIONS(2708),
    [anon_sym_PLUS] = ACTIONS(2708),
    [anon_sym_STAR] = ACTIONS(2710),
    [anon_sym_CARET] = ACTIONS(2710),
    [anon_sym_AMP] = ACTIONS(2710),
    [anon_sym_SEMI] = ACTIONS(2710),
    [anon_sym___extension__] = ACTIONS(2708),
    [anon_sym_typedef] = ACTIONS(2708),
    [anon_sym_extern] = ACTIONS(2708),
    [anon_sym___attribute__] = ACTIONS(2708),
    [anon_sym___attribute] = ACTIONS(2708),
    [anon_sym_noreturn] = ACTIONS(2708),
    [anon_sym_LBRACK] = ACTIONS(2710),
    [anon_sym___declspec] = ACTIONS(2708),
    [anon_sym___cdecl] = ACTIONS(2708),
    [anon_sym___clrcall] = ACTIONS(2708),
    [anon_sym___stdcall] = ACTIONS(2708),
    [anon_sym___fastcall] = ACTIONS(2708),
    [anon_sym___thiscall] = ACTIONS(2708),
    [anon_sym___vectorcall] = ACTIONS(2708),
    [anon_sym_LBRACE] = ACTIONS(2710),
    [anon_sym_signed] = ACTIONS(2708),
    [anon_sym_unsigned] = ACTIONS(2708),
    [anon_sym_long] = ACTIONS(2708),
    [anon_sym_short] = ACTIONS(2708),
    [anon_sym_static] = ACTIONS(2708),
    [anon_sym_ATautoreleasepool] = ACTIONS(2710),
    [anon_sym_auto] = ACTIONS(2708),
    [anon_sym_register] = ACTIONS(2708),
    [anon_sym_inline] = ACTIONS(2708),
    [anon_sym___inline] = ACTIONS(2708),
    [anon_sym___inline__] = ACTIONS(2708),
    [anon_sym___forceinline] = ACTIONS(2708),
    [anon_sym_thread_local] = ACTIONS(2708),
    [anon_sym___thread] = ACTIONS(2708),
    [anon_sym_CG_EXTERN] = ACTIONS(2708),
    [anon_sym_CG_INLINE] = ACTIONS(2708),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2708),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2708),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2708),
    [anon_sym_IBOutlet] = ACTIONS(2708),
    [anon_sym_IBInspectable] = ACTIONS(2708),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2708),
    [anon_sym_NS_INLINE] = ACTIONS(2708),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2708),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2708),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2708),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2708),
    [anon_sym_const] = ACTIONS(2708),
    [anon_sym_constexpr] = ACTIONS(2708),
    [anon_sym_volatile] = ACTIONS(2708),
    [anon_sym_restrict] = ACTIONS(2708),
    [anon_sym___restrict__] = ACTIONS(2708),
    [anon_sym__Atomic] = ACTIONS(2708),
    [anon_sym__Noreturn] = ACTIONS(2708),
    [anon_sym__Nonnull] = ACTIONS(2708),
    [anon_sym_nullable] = ACTIONS(2708),
    [anon_sym__Complex] = ACTIONS(2708),
    [anon_sym__Nullable] = ACTIONS(2708),
    [anon_sym__Nullable_result] = ACTIONS(2708),
    [anon_sym__Null_unspecified] = ACTIONS(2708),
    [anon_sym___autoreleasing] = ACTIONS(2708),
    [anon_sym___block] = ACTIONS(2708),
    [anon_sym___bridge] = ACTIONS(2708),
    [anon_sym___bridge_retained] = ACTIONS(2708),
    [anon_sym___bridge_transfer] = ACTIONS(2708),
    [anon_sym___complex] = ACTIONS(2708),
    [anon_sym___const] = ACTIONS(2708),
    [anon_sym___imag] = ACTIONS(2708),
    [anon_sym___kindof] = ACTIONS(2708),
    [anon_sym___nonnull] = ACTIONS(2708),
    [anon_sym___nullable] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2708),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2708),
    [anon_sym___real] = ACTIONS(2708),
    [anon_sym___strong] = ACTIONS(2708),
    [anon_sym___unsafe_unretained] = ACTIONS(2708),
    [anon_sym___unused] = ACTIONS(2708),
    [anon_sym___weak] = ACTIONS(2708),
    [anon_sym_alignas] = ACTIONS(2708),
    [anon_sym__Alignas] = ACTIONS(2708),
    [sym_primitive_type] = ACTIONS(2708),
    [anon_sym_enum] = ACTIONS(2708),
    [anon_sym_struct] = ACTIONS(2708),
    [anon_sym_union] = ACTIONS(2708),
    [anon_sym_if] = ACTIONS(2708),
    [anon_sym_switch] = ACTIONS(2708),
    [anon_sym_case] = ACTIONS(2708),
    [anon_sym_default] = ACTIONS(2708),
    [anon_sym_while] = ACTIONS(2708),
    [anon_sym_do] = ACTIONS(2708),
    [anon_sym_for] = ACTIONS(2708),
    [anon_sym_in] = ACTIONS(2708),
    [anon_sym_return] = ACTIONS(2708),
    [anon_sym_break] = ACTIONS(2708),
    [anon_sym_continue] = ACTIONS(2708),
    [anon_sym_goto] = ACTIONS(2708),
    [anon_sym_DASH_DASH] = ACTIONS(2710),
    [anon_sym_PLUS_PLUS] = ACTIONS(2710),
    [anon_sym_sizeof] = ACTIONS(2708),
    [anon_sym___alignof__] = ACTIONS(2708),
    [anon_sym___alignof] = ACTIONS(2708),
    [anon_sym__alignof] = ACTIONS(2708),
    [anon_sym_alignof] = ACTIONS(2708),
    [anon_sym__Alignof] = ACTIONS(2708),
    [anon_sym_offsetof] = ACTIONS(2708),
    [anon_sym__Generic] = ACTIONS(2708),
    [anon_sym_asm] = ACTIONS(2708),
    [anon_sym___asm__] = ACTIONS(2708),
    [anon_sym___asm] = ACTIONS(2708),
    [sym_number_literal] = ACTIONS(2710),
    [anon_sym_L_SQUOTE] = ACTIONS(2710),
    [anon_sym_u_SQUOTE] = ACTIONS(2710),
    [anon_sym_U_SQUOTE] = ACTIONS(2710),
    [anon_sym_u8_SQUOTE] = ACTIONS(2710),
    [anon_sym_SQUOTE] = ACTIONS(2710),
    [anon_sym_AT] = ACTIONS(2708),
    [anon_sym_DQUOTE] = ACTIONS(2710),
    [anon_sym_L_DQUOTE] = ACTIONS(2710),
    [anon_sym_u_DQUOTE] = ACTIONS(2710),
    [anon_sym_U_DQUOTE] = ACTIONS(2710),
    [anon_sym_u8_DQUOTE] = ACTIONS(2710),
    [sym_true] = ACTIONS(2708),
    [sym_false] = ACTIONS(2708),
    [anon_sym_NULL] = ACTIONS(2708),
    [anon_sym_nullptr] = ACTIONS(2708),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2708),
    [anon_sym___typeof] = ACTIONS(2708),
    [anon_sym_typeof] = ACTIONS(2708),
    [anon_sym_ATimport] = ACTIONS(2710),
    [aux_sym_preproc_undef_token1] = ACTIONS(2708),
    [anon_sym_POUND] = ACTIONS(2708),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2708),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2708),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2708),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2708),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2708),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2708),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2708),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_API_AVAILABLE] = ACTIONS(2708),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_API_DEPRECATED] = ACTIONS(2708),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2708),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2708),
    [anon_sym___deprecated_msg] = ACTIONS(2708),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2708),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2708),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2708),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2708),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2708),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2708),
    [anon_sym_ATprotocol] = ACTIONS(2710),
    [anon_sym_ATinterface] = ACTIONS(2710),
    [anon_sym_ATimplementation] = ACTIONS(2710),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2710),
    [anon_sym_ATselector] = ACTIONS(2710),
    [anon_sym_ATavailable] = ACTIONS(2710),
    [anon_sym___builtin_available] = ACTIONS(2708),
    [anon_sym_va_arg] = ACTIONS(2708),
    [anon_sym_ATencode] = ACTIONS(2710),
    [anon_sym_BOOL] = ACTIONS(2708),
    [anon_sym_IMP] = ACTIONS(2708),
    [anon_sym_SEL] = ACTIONS(2708),
    [anon_sym_Class] = ACTIONS(2708),
    [anon_sym_id] = ACTIONS(2708),
  },
  [STATE(1035)] = {
    [ts_builtin_sym_end] = ACTIONS(2648),
    [sym_identifier] = ACTIONS(2646),
    [aux_sym_preproc_include_token1] = ACTIONS(2646),
    [aux_sym_preproc_include_token2] = ACTIONS(2646),
    [aux_sym_preproc_def_token1] = ACTIONS(2646),
    [aux_sym_preproc_if_token1] = ACTIONS(2646),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2646),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2646),
    [sym_preproc_directive] = ACTIONS(2646),
    [anon_sym_LPAREN2] = ACTIONS(2648),
    [anon_sym_BANG] = ACTIONS(2648),
    [anon_sym_TILDE] = ACTIONS(2648),
    [anon_sym_DASH] = ACTIONS(2646),
    [anon_sym_PLUS] = ACTIONS(2646),
    [anon_sym_STAR] = ACTIONS(2648),
    [anon_sym_CARET] = ACTIONS(2648),
    [anon_sym_AMP] = ACTIONS(2648),
    [anon_sym_SEMI] = ACTIONS(2648),
    [anon_sym___extension__] = ACTIONS(2646),
    [anon_sym_typedef] = ACTIONS(2646),
    [anon_sym_extern] = ACTIONS(2646),
    [anon_sym___attribute__] = ACTIONS(2646),
    [anon_sym___attribute] = ACTIONS(2646),
    [anon_sym_noreturn] = ACTIONS(2646),
    [anon_sym_LBRACK] = ACTIONS(2648),
    [anon_sym___declspec] = ACTIONS(2646),
    [anon_sym___cdecl] = ACTIONS(2646),
    [anon_sym___clrcall] = ACTIONS(2646),
    [anon_sym___stdcall] = ACTIONS(2646),
    [anon_sym___fastcall] = ACTIONS(2646),
    [anon_sym___thiscall] = ACTIONS(2646),
    [anon_sym___vectorcall] = ACTIONS(2646),
    [anon_sym_LBRACE] = ACTIONS(2648),
    [anon_sym_signed] = ACTIONS(2646),
    [anon_sym_unsigned] = ACTIONS(2646),
    [anon_sym_long] = ACTIONS(2646),
    [anon_sym_short] = ACTIONS(2646),
    [anon_sym_static] = ACTIONS(2646),
    [anon_sym_ATautoreleasepool] = ACTIONS(2648),
    [anon_sym_auto] = ACTIONS(2646),
    [anon_sym_register] = ACTIONS(2646),
    [anon_sym_inline] = ACTIONS(2646),
    [anon_sym___inline] = ACTIONS(2646),
    [anon_sym___inline__] = ACTIONS(2646),
    [anon_sym___forceinline] = ACTIONS(2646),
    [anon_sym_thread_local] = ACTIONS(2646),
    [anon_sym___thread] = ACTIONS(2646),
    [anon_sym_CG_EXTERN] = ACTIONS(2646),
    [anon_sym_CG_INLINE] = ACTIONS(2646),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2646),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2646),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2646),
    [anon_sym_IBOutlet] = ACTIONS(2646),
    [anon_sym_IBInspectable] = ACTIONS(2646),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2646),
    [anon_sym_NS_INLINE] = ACTIONS(2646),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2646),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2646),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2646),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2646),
    [anon_sym_const] = ACTIONS(2646),
    [anon_sym_constexpr] = ACTIONS(2646),
    [anon_sym_volatile] = ACTIONS(2646),
    [anon_sym_restrict] = ACTIONS(2646),
    [anon_sym___restrict__] = ACTIONS(2646),
    [anon_sym__Atomic] = ACTIONS(2646),
    [anon_sym__Noreturn] = ACTIONS(2646),
    [anon_sym__Nonnull] = ACTIONS(2646),
    [anon_sym_nullable] = ACTIONS(2646),
    [anon_sym__Complex] = ACTIONS(2646),
    [anon_sym__Nullable] = ACTIONS(2646),
    [anon_sym__Nullable_result] = ACTIONS(2646),
    [anon_sym__Null_unspecified] = ACTIONS(2646),
    [anon_sym___autoreleasing] = ACTIONS(2646),
    [anon_sym___block] = ACTIONS(2646),
    [anon_sym___bridge] = ACTIONS(2646),
    [anon_sym___bridge_retained] = ACTIONS(2646),
    [anon_sym___bridge_transfer] = ACTIONS(2646),
    [anon_sym___complex] = ACTIONS(2646),
    [anon_sym___const] = ACTIONS(2646),
    [anon_sym___imag] = ACTIONS(2646),
    [anon_sym___kindof] = ACTIONS(2646),
    [anon_sym___nonnull] = ACTIONS(2646),
    [anon_sym___nullable] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2646),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2646),
    [anon_sym___real] = ACTIONS(2646),
    [anon_sym___strong] = ACTIONS(2646),
    [anon_sym___unsafe_unretained] = ACTIONS(2646),
    [anon_sym___unused] = ACTIONS(2646),
    [anon_sym___weak] = ACTIONS(2646),
    [anon_sym_alignas] = ACTIONS(2646),
    [anon_sym__Alignas] = ACTIONS(2646),
    [sym_primitive_type] = ACTIONS(2646),
    [anon_sym_enum] = ACTIONS(2646),
    [anon_sym_struct] = ACTIONS(2646),
    [anon_sym_union] = ACTIONS(2646),
    [anon_sym_if] = ACTIONS(2646),
    [anon_sym_switch] = ACTIONS(2646),
    [anon_sym_case] = ACTIONS(2646),
    [anon_sym_default] = ACTIONS(2646),
    [anon_sym_while] = ACTIONS(2646),
    [anon_sym_do] = ACTIONS(2646),
    [anon_sym_for] = ACTIONS(2646),
    [anon_sym_in] = ACTIONS(2646),
    [anon_sym_return] = ACTIONS(2646),
    [anon_sym_break] = ACTIONS(2646),
    [anon_sym_continue] = ACTIONS(2646),
    [anon_sym_goto] = ACTIONS(2646),
    [anon_sym_DASH_DASH] = ACTIONS(2648),
    [anon_sym_PLUS_PLUS] = ACTIONS(2648),
    [anon_sym_sizeof] = ACTIONS(2646),
    [anon_sym___alignof__] = ACTIONS(2646),
    [anon_sym___alignof] = ACTIONS(2646),
    [anon_sym__alignof] = ACTIONS(2646),
    [anon_sym_alignof] = ACTIONS(2646),
    [anon_sym__Alignof] = ACTIONS(2646),
    [anon_sym_offsetof] = ACTIONS(2646),
    [anon_sym__Generic] = ACTIONS(2646),
    [anon_sym_asm] = ACTIONS(2646),
    [anon_sym___asm__] = ACTIONS(2646),
    [anon_sym___asm] = ACTIONS(2646),
    [sym_number_literal] = ACTIONS(2648),
    [anon_sym_L_SQUOTE] = ACTIONS(2648),
    [anon_sym_u_SQUOTE] = ACTIONS(2648),
    [anon_sym_U_SQUOTE] = ACTIONS(2648),
    [anon_sym_u8_SQUOTE] = ACTIONS(2648),
    [anon_sym_SQUOTE] = ACTIONS(2648),
    [anon_sym_AT] = ACTIONS(2646),
    [anon_sym_DQUOTE] = ACTIONS(2648),
    [anon_sym_L_DQUOTE] = ACTIONS(2648),
    [anon_sym_u_DQUOTE] = ACTIONS(2648),
    [anon_sym_U_DQUOTE] = ACTIONS(2648),
    [anon_sym_u8_DQUOTE] = ACTIONS(2648),
    [sym_true] = ACTIONS(2646),
    [sym_false] = ACTIONS(2646),
    [anon_sym_NULL] = ACTIONS(2646),
    [anon_sym_nullptr] = ACTIONS(2646),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2646),
    [anon_sym___typeof] = ACTIONS(2646),
    [anon_sym_typeof] = ACTIONS(2646),
    [anon_sym_ATimport] = ACTIONS(2648),
    [aux_sym_preproc_undef_token1] = ACTIONS(2646),
    [anon_sym_POUND] = ACTIONS(2646),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2646),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2646),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2646),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2646),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2646),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2646),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2646),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_API_AVAILABLE] = ACTIONS(2646),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_API_DEPRECATED] = ACTIONS(2646),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2646),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2646),
    [anon_sym___deprecated_msg] = ACTIONS(2646),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2646),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2646),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2646),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2646),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2646),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2646),
    [anon_sym_ATprotocol] = ACTIONS(2648),
    [anon_sym_ATinterface] = ACTIONS(2648),
    [anon_sym_ATimplementation] = ACTIONS(2648),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2648),
    [anon_sym_ATselector] = ACTIONS(2648),
    [anon_sym_ATavailable] = ACTIONS(2648),
    [anon_sym___builtin_available] = ACTIONS(2646),
    [anon_sym_va_arg] = ACTIONS(2646),
    [anon_sym_ATencode] = ACTIONS(2648),
    [anon_sym_BOOL] = ACTIONS(2646),
    [anon_sym_IMP] = ACTIONS(2646),
    [anon_sym_SEL] = ACTIONS(2646),
    [anon_sym_Class] = ACTIONS(2646),
    [anon_sym_id] = ACTIONS(2646),
  },
  [STATE(1036)] = {
    [ts_builtin_sym_end] = ACTIONS(2858),
    [sym_identifier] = ACTIONS(2856),
    [aux_sym_preproc_include_token1] = ACTIONS(2856),
    [aux_sym_preproc_include_token2] = ACTIONS(2856),
    [aux_sym_preproc_def_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2856),
    [sym_preproc_directive] = ACTIONS(2856),
    [anon_sym_LPAREN2] = ACTIONS(2858),
    [anon_sym_BANG] = ACTIONS(2858),
    [anon_sym_TILDE] = ACTIONS(2858),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_STAR] = ACTIONS(2858),
    [anon_sym_CARET] = ACTIONS(2858),
    [anon_sym_AMP] = ACTIONS(2858),
    [anon_sym_SEMI] = ACTIONS(2858),
    [anon_sym___extension__] = ACTIONS(2856),
    [anon_sym_typedef] = ACTIONS(2856),
    [anon_sym_extern] = ACTIONS(2856),
    [anon_sym___attribute__] = ACTIONS(2856),
    [anon_sym___attribute] = ACTIONS(2856),
    [anon_sym_noreturn] = ACTIONS(2856),
    [anon_sym_LBRACK] = ACTIONS(2858),
    [anon_sym___declspec] = ACTIONS(2856),
    [anon_sym___cdecl] = ACTIONS(2856),
    [anon_sym___clrcall] = ACTIONS(2856),
    [anon_sym___stdcall] = ACTIONS(2856),
    [anon_sym___fastcall] = ACTIONS(2856),
    [anon_sym___thiscall] = ACTIONS(2856),
    [anon_sym___vectorcall] = ACTIONS(2856),
    [anon_sym_LBRACE] = ACTIONS(2858),
    [anon_sym_signed] = ACTIONS(2856),
    [anon_sym_unsigned] = ACTIONS(2856),
    [anon_sym_long] = ACTIONS(2856),
    [anon_sym_short] = ACTIONS(2856),
    [anon_sym_static] = ACTIONS(2856),
    [anon_sym_ATautoreleasepool] = ACTIONS(2858),
    [anon_sym_auto] = ACTIONS(2856),
    [anon_sym_register] = ACTIONS(2856),
    [anon_sym_inline] = ACTIONS(2856),
    [anon_sym___inline] = ACTIONS(2856),
    [anon_sym___inline__] = ACTIONS(2856),
    [anon_sym___forceinline] = ACTIONS(2856),
    [anon_sym_thread_local] = ACTIONS(2856),
    [anon_sym___thread] = ACTIONS(2856),
    [anon_sym_CG_EXTERN] = ACTIONS(2856),
    [anon_sym_CG_INLINE] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2856),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2856),
    [anon_sym_IBOutlet] = ACTIONS(2856),
    [anon_sym_IBInspectable] = ACTIONS(2856),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2856),
    [anon_sym_NS_INLINE] = ACTIONS(2856),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2856),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2856),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2856),
    [anon_sym_const] = ACTIONS(2856),
    [anon_sym_constexpr] = ACTIONS(2856),
    [anon_sym_volatile] = ACTIONS(2856),
    [anon_sym_restrict] = ACTIONS(2856),
    [anon_sym___restrict__] = ACTIONS(2856),
    [anon_sym__Atomic] = ACTIONS(2856),
    [anon_sym__Noreturn] = ACTIONS(2856),
    [anon_sym__Nonnull] = ACTIONS(2856),
    [anon_sym_nullable] = ACTIONS(2856),
    [anon_sym__Complex] = ACTIONS(2856),
    [anon_sym__Nullable] = ACTIONS(2856),
    [anon_sym__Nullable_result] = ACTIONS(2856),
    [anon_sym__Null_unspecified] = ACTIONS(2856),
    [anon_sym___autoreleasing] = ACTIONS(2856),
    [anon_sym___block] = ACTIONS(2856),
    [anon_sym___bridge] = ACTIONS(2856),
    [anon_sym___bridge_retained] = ACTIONS(2856),
    [anon_sym___bridge_transfer] = ACTIONS(2856),
    [anon_sym___complex] = ACTIONS(2856),
    [anon_sym___const] = ACTIONS(2856),
    [anon_sym___imag] = ACTIONS(2856),
    [anon_sym___kindof] = ACTIONS(2856),
    [anon_sym___nonnull] = ACTIONS(2856),
    [anon_sym___nullable] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2856),
    [anon_sym___real] = ACTIONS(2856),
    [anon_sym___strong] = ACTIONS(2856),
    [anon_sym___unsafe_unretained] = ACTIONS(2856),
    [anon_sym___unused] = ACTIONS(2856),
    [anon_sym___weak] = ACTIONS(2856),
    [anon_sym_alignas] = ACTIONS(2856),
    [anon_sym__Alignas] = ACTIONS(2856),
    [sym_primitive_type] = ACTIONS(2856),
    [anon_sym_enum] = ACTIONS(2856),
    [anon_sym_struct] = ACTIONS(2856),
    [anon_sym_union] = ACTIONS(2856),
    [anon_sym_if] = ACTIONS(2856),
    [anon_sym_switch] = ACTIONS(2856),
    [anon_sym_case] = ACTIONS(2856),
    [anon_sym_default] = ACTIONS(2856),
    [anon_sym_while] = ACTIONS(2856),
    [anon_sym_do] = ACTIONS(2856),
    [anon_sym_for] = ACTIONS(2856),
    [anon_sym_in] = ACTIONS(2856),
    [anon_sym_return] = ACTIONS(2856),
    [anon_sym_break] = ACTIONS(2856),
    [anon_sym_continue] = ACTIONS(2856),
    [anon_sym_goto] = ACTIONS(2856),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_sizeof] = ACTIONS(2856),
    [anon_sym___alignof__] = ACTIONS(2856),
    [anon_sym___alignof] = ACTIONS(2856),
    [anon_sym__alignof] = ACTIONS(2856),
    [anon_sym_alignof] = ACTIONS(2856),
    [anon_sym__Alignof] = ACTIONS(2856),
    [anon_sym_offsetof] = ACTIONS(2856),
    [anon_sym__Generic] = ACTIONS(2856),
    [anon_sym_asm] = ACTIONS(2856),
    [anon_sym___asm__] = ACTIONS(2856),
    [anon_sym___asm] = ACTIONS(2856),
    [sym_number_literal] = ACTIONS(2858),
    [anon_sym_L_SQUOTE] = ACTIONS(2858),
    [anon_sym_u_SQUOTE] = ACTIONS(2858),
    [anon_sym_U_SQUOTE] = ACTIONS(2858),
    [anon_sym_u8_SQUOTE] = ACTIONS(2858),
    [anon_sym_SQUOTE] = ACTIONS(2858),
    [anon_sym_AT] = ACTIONS(2856),
    [anon_sym_DQUOTE] = ACTIONS(2858),
    [anon_sym_L_DQUOTE] = ACTIONS(2858),
    [anon_sym_u_DQUOTE] = ACTIONS(2858),
    [anon_sym_U_DQUOTE] = ACTIONS(2858),
    [anon_sym_u8_DQUOTE] = ACTIONS(2858),
    [sym_true] = ACTIONS(2856),
    [sym_false] = ACTIONS(2856),
    [anon_sym_NULL] = ACTIONS(2856),
    [anon_sym_nullptr] = ACTIONS(2856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2856),
    [anon_sym___typeof] = ACTIONS(2856),
    [anon_sym_typeof] = ACTIONS(2856),
    [anon_sym_ATimport] = ACTIONS(2858),
    [aux_sym_preproc_undef_token1] = ACTIONS(2856),
    [anon_sym_POUND] = ACTIONS(2856),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2856),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2856),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2856),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_API_AVAILABLE] = ACTIONS(2856),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_API_DEPRECATED] = ACTIONS(2856),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym___deprecated_msg] = ACTIONS(2856),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2856),
    [anon_sym_ATprotocol] = ACTIONS(2858),
    [anon_sym_ATinterface] = ACTIONS(2858),
    [anon_sym_ATimplementation] = ACTIONS(2858),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2858),
    [anon_sym_ATselector] = ACTIONS(2858),
    [anon_sym_ATavailable] = ACTIONS(2858),
    [anon_sym___builtin_available] = ACTIONS(2856),
    [anon_sym_va_arg] = ACTIONS(2856),
    [anon_sym_ATencode] = ACTIONS(2858),
    [anon_sym_BOOL] = ACTIONS(2856),
    [anon_sym_IMP] = ACTIONS(2856),
    [anon_sym_SEL] = ACTIONS(2856),
    [anon_sym_Class] = ACTIONS(2856),
    [anon_sym_id] = ACTIONS(2856),
  },
  [STATE(1037)] = {
    [ts_builtin_sym_end] = ACTIONS(2652),
    [sym_identifier] = ACTIONS(2650),
    [aux_sym_preproc_include_token1] = ACTIONS(2650),
    [aux_sym_preproc_include_token2] = ACTIONS(2650),
    [aux_sym_preproc_def_token1] = ACTIONS(2650),
    [aux_sym_preproc_if_token1] = ACTIONS(2650),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2650),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2650),
    [sym_preproc_directive] = ACTIONS(2650),
    [anon_sym_LPAREN2] = ACTIONS(2652),
    [anon_sym_BANG] = ACTIONS(2652),
    [anon_sym_TILDE] = ACTIONS(2652),
    [anon_sym_DASH] = ACTIONS(2650),
    [anon_sym_PLUS] = ACTIONS(2650),
    [anon_sym_STAR] = ACTIONS(2652),
    [anon_sym_CARET] = ACTIONS(2652),
    [anon_sym_AMP] = ACTIONS(2652),
    [anon_sym_SEMI] = ACTIONS(2652),
    [anon_sym___extension__] = ACTIONS(2650),
    [anon_sym_typedef] = ACTIONS(2650),
    [anon_sym_extern] = ACTIONS(2650),
    [anon_sym___attribute__] = ACTIONS(2650),
    [anon_sym___attribute] = ACTIONS(2650),
    [anon_sym_noreturn] = ACTIONS(2650),
    [anon_sym_LBRACK] = ACTIONS(2652),
    [anon_sym___declspec] = ACTIONS(2650),
    [anon_sym___cdecl] = ACTIONS(2650),
    [anon_sym___clrcall] = ACTIONS(2650),
    [anon_sym___stdcall] = ACTIONS(2650),
    [anon_sym___fastcall] = ACTIONS(2650),
    [anon_sym___thiscall] = ACTIONS(2650),
    [anon_sym___vectorcall] = ACTIONS(2650),
    [anon_sym_LBRACE] = ACTIONS(2652),
    [anon_sym_signed] = ACTIONS(2650),
    [anon_sym_unsigned] = ACTIONS(2650),
    [anon_sym_long] = ACTIONS(2650),
    [anon_sym_short] = ACTIONS(2650),
    [anon_sym_static] = ACTIONS(2650),
    [anon_sym_ATautoreleasepool] = ACTIONS(2652),
    [anon_sym_auto] = ACTIONS(2650),
    [anon_sym_register] = ACTIONS(2650),
    [anon_sym_inline] = ACTIONS(2650),
    [anon_sym___inline] = ACTIONS(2650),
    [anon_sym___inline__] = ACTIONS(2650),
    [anon_sym___forceinline] = ACTIONS(2650),
    [anon_sym_thread_local] = ACTIONS(2650),
    [anon_sym___thread] = ACTIONS(2650),
    [anon_sym_CG_EXTERN] = ACTIONS(2650),
    [anon_sym_CG_INLINE] = ACTIONS(2650),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2650),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2650),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2650),
    [anon_sym_IBOutlet] = ACTIONS(2650),
    [anon_sym_IBInspectable] = ACTIONS(2650),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2650),
    [anon_sym_NS_INLINE] = ACTIONS(2650),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2650),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2650),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2650),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2650),
    [anon_sym_const] = ACTIONS(2650),
    [anon_sym_constexpr] = ACTIONS(2650),
    [anon_sym_volatile] = ACTIONS(2650),
    [anon_sym_restrict] = ACTIONS(2650),
    [anon_sym___restrict__] = ACTIONS(2650),
    [anon_sym__Atomic] = ACTIONS(2650),
    [anon_sym__Noreturn] = ACTIONS(2650),
    [anon_sym__Nonnull] = ACTIONS(2650),
    [anon_sym_nullable] = ACTIONS(2650),
    [anon_sym__Complex] = ACTIONS(2650),
    [anon_sym__Nullable] = ACTIONS(2650),
    [anon_sym__Nullable_result] = ACTIONS(2650),
    [anon_sym__Null_unspecified] = ACTIONS(2650),
    [anon_sym___autoreleasing] = ACTIONS(2650),
    [anon_sym___block] = ACTIONS(2650),
    [anon_sym___bridge] = ACTIONS(2650),
    [anon_sym___bridge_retained] = ACTIONS(2650),
    [anon_sym___bridge_transfer] = ACTIONS(2650),
    [anon_sym___complex] = ACTIONS(2650),
    [anon_sym___const] = ACTIONS(2650),
    [anon_sym___imag] = ACTIONS(2650),
    [anon_sym___kindof] = ACTIONS(2650),
    [anon_sym___nonnull] = ACTIONS(2650),
    [anon_sym___nullable] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2650),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2650),
    [anon_sym___real] = ACTIONS(2650),
    [anon_sym___strong] = ACTIONS(2650),
    [anon_sym___unsafe_unretained] = ACTIONS(2650),
    [anon_sym___unused] = ACTIONS(2650),
    [anon_sym___weak] = ACTIONS(2650),
    [anon_sym_alignas] = ACTIONS(2650),
    [anon_sym__Alignas] = ACTIONS(2650),
    [sym_primitive_type] = ACTIONS(2650),
    [anon_sym_enum] = ACTIONS(2650),
    [anon_sym_struct] = ACTIONS(2650),
    [anon_sym_union] = ACTIONS(2650),
    [anon_sym_if] = ACTIONS(2650),
    [anon_sym_switch] = ACTIONS(2650),
    [anon_sym_case] = ACTIONS(2650),
    [anon_sym_default] = ACTIONS(2650),
    [anon_sym_while] = ACTIONS(2650),
    [anon_sym_do] = ACTIONS(2650),
    [anon_sym_for] = ACTIONS(2650),
    [anon_sym_in] = ACTIONS(2650),
    [anon_sym_return] = ACTIONS(2650),
    [anon_sym_break] = ACTIONS(2650),
    [anon_sym_continue] = ACTIONS(2650),
    [anon_sym_goto] = ACTIONS(2650),
    [anon_sym_DASH_DASH] = ACTIONS(2652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2652),
    [anon_sym_sizeof] = ACTIONS(2650),
    [anon_sym___alignof__] = ACTIONS(2650),
    [anon_sym___alignof] = ACTIONS(2650),
    [anon_sym__alignof] = ACTIONS(2650),
    [anon_sym_alignof] = ACTIONS(2650),
    [anon_sym__Alignof] = ACTIONS(2650),
    [anon_sym_offsetof] = ACTIONS(2650),
    [anon_sym__Generic] = ACTIONS(2650),
    [anon_sym_asm] = ACTIONS(2650),
    [anon_sym___asm__] = ACTIONS(2650),
    [anon_sym___asm] = ACTIONS(2650),
    [sym_number_literal] = ACTIONS(2652),
    [anon_sym_L_SQUOTE] = ACTIONS(2652),
    [anon_sym_u_SQUOTE] = ACTIONS(2652),
    [anon_sym_U_SQUOTE] = ACTIONS(2652),
    [anon_sym_u8_SQUOTE] = ACTIONS(2652),
    [anon_sym_SQUOTE] = ACTIONS(2652),
    [anon_sym_AT] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_L_DQUOTE] = ACTIONS(2652),
    [anon_sym_u_DQUOTE] = ACTIONS(2652),
    [anon_sym_U_DQUOTE] = ACTIONS(2652),
    [anon_sym_u8_DQUOTE] = ACTIONS(2652),
    [sym_true] = ACTIONS(2650),
    [sym_false] = ACTIONS(2650),
    [anon_sym_NULL] = ACTIONS(2650),
    [anon_sym_nullptr] = ACTIONS(2650),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2650),
    [anon_sym___typeof] = ACTIONS(2650),
    [anon_sym_typeof] = ACTIONS(2650),
    [anon_sym_ATimport] = ACTIONS(2652),
    [aux_sym_preproc_undef_token1] = ACTIONS(2650),
    [anon_sym_POUND] = ACTIONS(2650),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2650),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2650),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2650),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2650),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2650),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2650),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2650),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_API_AVAILABLE] = ACTIONS(2650),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_API_DEPRECATED] = ACTIONS(2650),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2650),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2650),
    [anon_sym___deprecated_msg] = ACTIONS(2650),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2650),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2650),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2650),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2650),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2650),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2650),
    [anon_sym_ATprotocol] = ACTIONS(2652),
    [anon_sym_ATinterface] = ACTIONS(2652),
    [anon_sym_ATimplementation] = ACTIONS(2652),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2652),
    [anon_sym_ATselector] = ACTIONS(2652),
    [anon_sym_ATavailable] = ACTIONS(2652),
    [anon_sym___builtin_available] = ACTIONS(2650),
    [anon_sym_va_arg] = ACTIONS(2650),
    [anon_sym_ATencode] = ACTIONS(2652),
    [anon_sym_BOOL] = ACTIONS(2650),
    [anon_sym_IMP] = ACTIONS(2650),
    [anon_sym_SEL] = ACTIONS(2650),
    [anon_sym_Class] = ACTIONS(2650),
    [anon_sym_id] = ACTIONS(2650),
  },
  [STATE(1038)] = {
    [ts_builtin_sym_end] = ACTIONS(2722),
    [sym_identifier] = ACTIONS(2720),
    [aux_sym_preproc_include_token1] = ACTIONS(2720),
    [aux_sym_preproc_include_token2] = ACTIONS(2720),
    [aux_sym_preproc_def_token1] = ACTIONS(2720),
    [aux_sym_preproc_if_token1] = ACTIONS(2720),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2720),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2720),
    [sym_preproc_directive] = ACTIONS(2720),
    [anon_sym_LPAREN2] = ACTIONS(2722),
    [anon_sym_BANG] = ACTIONS(2722),
    [anon_sym_TILDE] = ACTIONS(2722),
    [anon_sym_DASH] = ACTIONS(2720),
    [anon_sym_PLUS] = ACTIONS(2720),
    [anon_sym_STAR] = ACTIONS(2722),
    [anon_sym_CARET] = ACTIONS(2722),
    [anon_sym_AMP] = ACTIONS(2722),
    [anon_sym_SEMI] = ACTIONS(2722),
    [anon_sym___extension__] = ACTIONS(2720),
    [anon_sym_typedef] = ACTIONS(2720),
    [anon_sym_extern] = ACTIONS(2720),
    [anon_sym___attribute__] = ACTIONS(2720),
    [anon_sym___attribute] = ACTIONS(2720),
    [anon_sym_noreturn] = ACTIONS(2720),
    [anon_sym_LBRACK] = ACTIONS(2722),
    [anon_sym___declspec] = ACTIONS(2720),
    [anon_sym___cdecl] = ACTIONS(2720),
    [anon_sym___clrcall] = ACTIONS(2720),
    [anon_sym___stdcall] = ACTIONS(2720),
    [anon_sym___fastcall] = ACTIONS(2720),
    [anon_sym___thiscall] = ACTIONS(2720),
    [anon_sym___vectorcall] = ACTIONS(2720),
    [anon_sym_LBRACE] = ACTIONS(2722),
    [anon_sym_signed] = ACTIONS(2720),
    [anon_sym_unsigned] = ACTIONS(2720),
    [anon_sym_long] = ACTIONS(2720),
    [anon_sym_short] = ACTIONS(2720),
    [anon_sym_static] = ACTIONS(2720),
    [anon_sym_ATautoreleasepool] = ACTIONS(2722),
    [anon_sym_auto] = ACTIONS(2720),
    [anon_sym_register] = ACTIONS(2720),
    [anon_sym_inline] = ACTIONS(2720),
    [anon_sym___inline] = ACTIONS(2720),
    [anon_sym___inline__] = ACTIONS(2720),
    [anon_sym___forceinline] = ACTIONS(2720),
    [anon_sym_thread_local] = ACTIONS(2720),
    [anon_sym___thread] = ACTIONS(2720),
    [anon_sym_CG_EXTERN] = ACTIONS(2720),
    [anon_sym_CG_INLINE] = ACTIONS(2720),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2720),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2720),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2720),
    [anon_sym_IBOutlet] = ACTIONS(2720),
    [anon_sym_IBInspectable] = ACTIONS(2720),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2720),
    [anon_sym_NS_INLINE] = ACTIONS(2720),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2720),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2720),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2720),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2720),
    [anon_sym_const] = ACTIONS(2720),
    [anon_sym_constexpr] = ACTIONS(2720),
    [anon_sym_volatile] = ACTIONS(2720),
    [anon_sym_restrict] = ACTIONS(2720),
    [anon_sym___restrict__] = ACTIONS(2720),
    [anon_sym__Atomic] = ACTIONS(2720),
    [anon_sym__Noreturn] = ACTIONS(2720),
    [anon_sym__Nonnull] = ACTIONS(2720),
    [anon_sym_nullable] = ACTIONS(2720),
    [anon_sym__Complex] = ACTIONS(2720),
    [anon_sym__Nullable] = ACTIONS(2720),
    [anon_sym__Nullable_result] = ACTIONS(2720),
    [anon_sym__Null_unspecified] = ACTIONS(2720),
    [anon_sym___autoreleasing] = ACTIONS(2720),
    [anon_sym___block] = ACTIONS(2720),
    [anon_sym___bridge] = ACTIONS(2720),
    [anon_sym___bridge_retained] = ACTIONS(2720),
    [anon_sym___bridge_transfer] = ACTIONS(2720),
    [anon_sym___complex] = ACTIONS(2720),
    [anon_sym___const] = ACTIONS(2720),
    [anon_sym___imag] = ACTIONS(2720),
    [anon_sym___kindof] = ACTIONS(2720),
    [anon_sym___nonnull] = ACTIONS(2720),
    [anon_sym___nullable] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2720),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2720),
    [anon_sym___real] = ACTIONS(2720),
    [anon_sym___strong] = ACTIONS(2720),
    [anon_sym___unsafe_unretained] = ACTIONS(2720),
    [anon_sym___unused] = ACTIONS(2720),
    [anon_sym___weak] = ACTIONS(2720),
    [anon_sym_alignas] = ACTIONS(2720),
    [anon_sym__Alignas] = ACTIONS(2720),
    [sym_primitive_type] = ACTIONS(2720),
    [anon_sym_enum] = ACTIONS(2720),
    [anon_sym_struct] = ACTIONS(2720),
    [anon_sym_union] = ACTIONS(2720),
    [anon_sym_if] = ACTIONS(2720),
    [anon_sym_switch] = ACTIONS(2720),
    [anon_sym_case] = ACTIONS(2720),
    [anon_sym_default] = ACTIONS(2720),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_do] = ACTIONS(2720),
    [anon_sym_for] = ACTIONS(2720),
    [anon_sym_in] = ACTIONS(2720),
    [anon_sym_return] = ACTIONS(2720),
    [anon_sym_break] = ACTIONS(2720),
    [anon_sym_continue] = ACTIONS(2720),
    [anon_sym_goto] = ACTIONS(2720),
    [anon_sym_DASH_DASH] = ACTIONS(2722),
    [anon_sym_PLUS_PLUS] = ACTIONS(2722),
    [anon_sym_sizeof] = ACTIONS(2720),
    [anon_sym___alignof__] = ACTIONS(2720),
    [anon_sym___alignof] = ACTIONS(2720),
    [anon_sym__alignof] = ACTIONS(2720),
    [anon_sym_alignof] = ACTIONS(2720),
    [anon_sym__Alignof] = ACTIONS(2720),
    [anon_sym_offsetof] = ACTIONS(2720),
    [anon_sym__Generic] = ACTIONS(2720),
    [anon_sym_asm] = ACTIONS(2720),
    [anon_sym___asm__] = ACTIONS(2720),
    [anon_sym___asm] = ACTIONS(2720),
    [sym_number_literal] = ACTIONS(2722),
    [anon_sym_L_SQUOTE] = ACTIONS(2722),
    [anon_sym_u_SQUOTE] = ACTIONS(2722),
    [anon_sym_U_SQUOTE] = ACTIONS(2722),
    [anon_sym_u8_SQUOTE] = ACTIONS(2722),
    [anon_sym_SQUOTE] = ACTIONS(2722),
    [anon_sym_AT] = ACTIONS(2720),
    [anon_sym_DQUOTE] = ACTIONS(2722),
    [anon_sym_L_DQUOTE] = ACTIONS(2722),
    [anon_sym_u_DQUOTE] = ACTIONS(2722),
    [anon_sym_U_DQUOTE] = ACTIONS(2722),
    [anon_sym_u8_DQUOTE] = ACTIONS(2722),
    [sym_true] = ACTIONS(2720),
    [sym_false] = ACTIONS(2720),
    [anon_sym_NULL] = ACTIONS(2720),
    [anon_sym_nullptr] = ACTIONS(2720),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2720),
    [anon_sym___typeof] = ACTIONS(2720),
    [anon_sym_typeof] = ACTIONS(2720),
    [anon_sym_ATimport] = ACTIONS(2722),
    [aux_sym_preproc_undef_token1] = ACTIONS(2720),
    [anon_sym_POUND] = ACTIONS(2720),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2720),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2720),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2720),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2720),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2720),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2720),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2720),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_API_AVAILABLE] = ACTIONS(2720),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_API_DEPRECATED] = ACTIONS(2720),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2720),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2720),
    [anon_sym___deprecated_msg] = ACTIONS(2720),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2720),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2720),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2720),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2720),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2720),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2720),
    [anon_sym_ATprotocol] = ACTIONS(2722),
    [anon_sym_ATinterface] = ACTIONS(2722),
    [anon_sym_ATimplementation] = ACTIONS(2722),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2722),
    [anon_sym_ATselector] = ACTIONS(2722),
    [anon_sym_ATavailable] = ACTIONS(2722),
    [anon_sym___builtin_available] = ACTIONS(2720),
    [anon_sym_va_arg] = ACTIONS(2720),
    [anon_sym_ATencode] = ACTIONS(2722),
    [anon_sym_BOOL] = ACTIONS(2720),
    [anon_sym_IMP] = ACTIONS(2720),
    [anon_sym_SEL] = ACTIONS(2720),
    [anon_sym_Class] = ACTIONS(2720),
    [anon_sym_id] = ACTIONS(2720),
  },
  [STATE(1039)] = {
    [ts_builtin_sym_end] = ACTIONS(2726),
    [sym_identifier] = ACTIONS(2724),
    [aux_sym_preproc_include_token1] = ACTIONS(2724),
    [aux_sym_preproc_include_token2] = ACTIONS(2724),
    [aux_sym_preproc_def_token1] = ACTIONS(2724),
    [aux_sym_preproc_if_token1] = ACTIONS(2724),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2724),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2724),
    [sym_preproc_directive] = ACTIONS(2724),
    [anon_sym_LPAREN2] = ACTIONS(2726),
    [anon_sym_BANG] = ACTIONS(2726),
    [anon_sym_TILDE] = ACTIONS(2726),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2726),
    [anon_sym_CARET] = ACTIONS(2726),
    [anon_sym_AMP] = ACTIONS(2726),
    [anon_sym_SEMI] = ACTIONS(2726),
    [anon_sym___extension__] = ACTIONS(2724),
    [anon_sym_typedef] = ACTIONS(2724),
    [anon_sym_extern] = ACTIONS(2724),
    [anon_sym___attribute__] = ACTIONS(2724),
    [anon_sym___attribute] = ACTIONS(2724),
    [anon_sym_noreturn] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2726),
    [anon_sym___declspec] = ACTIONS(2724),
    [anon_sym___cdecl] = ACTIONS(2724),
    [anon_sym___clrcall] = ACTIONS(2724),
    [anon_sym___stdcall] = ACTIONS(2724),
    [anon_sym___fastcall] = ACTIONS(2724),
    [anon_sym___thiscall] = ACTIONS(2724),
    [anon_sym___vectorcall] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2726),
    [anon_sym_signed] = ACTIONS(2724),
    [anon_sym_unsigned] = ACTIONS(2724),
    [anon_sym_long] = ACTIONS(2724),
    [anon_sym_short] = ACTIONS(2724),
    [anon_sym_static] = ACTIONS(2724),
    [anon_sym_ATautoreleasepool] = ACTIONS(2726),
    [anon_sym_auto] = ACTIONS(2724),
    [anon_sym_register] = ACTIONS(2724),
    [anon_sym_inline] = ACTIONS(2724),
    [anon_sym___inline] = ACTIONS(2724),
    [anon_sym___inline__] = ACTIONS(2724),
    [anon_sym___forceinline] = ACTIONS(2724),
    [anon_sym_thread_local] = ACTIONS(2724),
    [anon_sym___thread] = ACTIONS(2724),
    [anon_sym_CG_EXTERN] = ACTIONS(2724),
    [anon_sym_CG_INLINE] = ACTIONS(2724),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2724),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2724),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2724),
    [anon_sym_IBOutlet] = ACTIONS(2724),
    [anon_sym_IBInspectable] = ACTIONS(2724),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2724),
    [anon_sym_NS_INLINE] = ACTIONS(2724),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2724),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2724),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2724),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2724),
    [anon_sym_const] = ACTIONS(2724),
    [anon_sym_constexpr] = ACTIONS(2724),
    [anon_sym_volatile] = ACTIONS(2724),
    [anon_sym_restrict] = ACTIONS(2724),
    [anon_sym___restrict__] = ACTIONS(2724),
    [anon_sym__Atomic] = ACTIONS(2724),
    [anon_sym__Noreturn] = ACTIONS(2724),
    [anon_sym__Nonnull] = ACTIONS(2724),
    [anon_sym_nullable] = ACTIONS(2724),
    [anon_sym__Complex] = ACTIONS(2724),
    [anon_sym__Nullable] = ACTIONS(2724),
    [anon_sym__Nullable_result] = ACTIONS(2724),
    [anon_sym__Null_unspecified] = ACTIONS(2724),
    [anon_sym___autoreleasing] = ACTIONS(2724),
    [anon_sym___block] = ACTIONS(2724),
    [anon_sym___bridge] = ACTIONS(2724),
    [anon_sym___bridge_retained] = ACTIONS(2724),
    [anon_sym___bridge_transfer] = ACTIONS(2724),
    [anon_sym___complex] = ACTIONS(2724),
    [anon_sym___const] = ACTIONS(2724),
    [anon_sym___imag] = ACTIONS(2724),
    [anon_sym___kindof] = ACTIONS(2724),
    [anon_sym___nonnull] = ACTIONS(2724),
    [anon_sym___nullable] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2724),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2724),
    [anon_sym___real] = ACTIONS(2724),
    [anon_sym___strong] = ACTIONS(2724),
    [anon_sym___unsafe_unretained] = ACTIONS(2724),
    [anon_sym___unused] = ACTIONS(2724),
    [anon_sym___weak] = ACTIONS(2724),
    [anon_sym_alignas] = ACTIONS(2724),
    [anon_sym__Alignas] = ACTIONS(2724),
    [sym_primitive_type] = ACTIONS(2724),
    [anon_sym_enum] = ACTIONS(2724),
    [anon_sym_struct] = ACTIONS(2724),
    [anon_sym_union] = ACTIONS(2724),
    [anon_sym_if] = ACTIONS(2724),
    [anon_sym_switch] = ACTIONS(2724),
    [anon_sym_case] = ACTIONS(2724),
    [anon_sym_default] = ACTIONS(2724),
    [anon_sym_while] = ACTIONS(2724),
    [anon_sym_do] = ACTIONS(2724),
    [anon_sym_for] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_return] = ACTIONS(2724),
    [anon_sym_break] = ACTIONS(2724),
    [anon_sym_continue] = ACTIONS(2724),
    [anon_sym_goto] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2726),
    [anon_sym_PLUS_PLUS] = ACTIONS(2726),
    [anon_sym_sizeof] = ACTIONS(2724),
    [anon_sym___alignof__] = ACTIONS(2724),
    [anon_sym___alignof] = ACTIONS(2724),
    [anon_sym__alignof] = ACTIONS(2724),
    [anon_sym_alignof] = ACTIONS(2724),
    [anon_sym__Alignof] = ACTIONS(2724),
    [anon_sym_offsetof] = ACTIONS(2724),
    [anon_sym__Generic] = ACTIONS(2724),
    [anon_sym_asm] = ACTIONS(2724),
    [anon_sym___asm__] = ACTIONS(2724),
    [anon_sym___asm] = ACTIONS(2724),
    [sym_number_literal] = ACTIONS(2726),
    [anon_sym_L_SQUOTE] = ACTIONS(2726),
    [anon_sym_u_SQUOTE] = ACTIONS(2726),
    [anon_sym_U_SQUOTE] = ACTIONS(2726),
    [anon_sym_u8_SQUOTE] = ACTIONS(2726),
    [anon_sym_SQUOTE] = ACTIONS(2726),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2726),
    [anon_sym_L_DQUOTE] = ACTIONS(2726),
    [anon_sym_u_DQUOTE] = ACTIONS(2726),
    [anon_sym_U_DQUOTE] = ACTIONS(2726),
    [anon_sym_u8_DQUOTE] = ACTIONS(2726),
    [sym_true] = ACTIONS(2724),
    [sym_false] = ACTIONS(2724),
    [anon_sym_NULL] = ACTIONS(2724),
    [anon_sym_nullptr] = ACTIONS(2724),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2724),
    [anon_sym___typeof] = ACTIONS(2724),
    [anon_sym_typeof] = ACTIONS(2724),
    [anon_sym_ATimport] = ACTIONS(2726),
    [aux_sym_preproc_undef_token1] = ACTIONS(2724),
    [anon_sym_POUND] = ACTIONS(2724),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2724),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2724),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2724),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2724),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2724),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2724),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2724),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_API_AVAILABLE] = ACTIONS(2724),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_API_DEPRECATED] = ACTIONS(2724),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2724),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2724),
    [anon_sym___deprecated_msg] = ACTIONS(2724),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2724),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2724),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2724),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2724),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2724),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2724),
    [anon_sym_ATprotocol] = ACTIONS(2726),
    [anon_sym_ATinterface] = ACTIONS(2726),
    [anon_sym_ATimplementation] = ACTIONS(2726),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2726),
    [anon_sym_ATselector] = ACTIONS(2726),
    [anon_sym_ATavailable] = ACTIONS(2726),
    [anon_sym___builtin_available] = ACTIONS(2724),
    [anon_sym_va_arg] = ACTIONS(2724),
    [anon_sym_ATencode] = ACTIONS(2726),
    [anon_sym_BOOL] = ACTIONS(2724),
    [anon_sym_IMP] = ACTIONS(2724),
    [anon_sym_SEL] = ACTIONS(2724),
    [anon_sym_Class] = ACTIONS(2724),
    [anon_sym_id] = ACTIONS(2724),
  },
  [STATE(1040)] = {
    [ts_builtin_sym_end] = ACTIONS(2730),
    [sym_identifier] = ACTIONS(2728),
    [aux_sym_preproc_include_token1] = ACTIONS(2728),
    [aux_sym_preproc_include_token2] = ACTIONS(2728),
    [aux_sym_preproc_def_token1] = ACTIONS(2728),
    [aux_sym_preproc_if_token1] = ACTIONS(2728),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2728),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2728),
    [sym_preproc_directive] = ACTIONS(2728),
    [anon_sym_LPAREN2] = ACTIONS(2730),
    [anon_sym_BANG] = ACTIONS(2730),
    [anon_sym_TILDE] = ACTIONS(2730),
    [anon_sym_DASH] = ACTIONS(2728),
    [anon_sym_PLUS] = ACTIONS(2728),
    [anon_sym_STAR] = ACTIONS(2730),
    [anon_sym_CARET] = ACTIONS(2730),
    [anon_sym_AMP] = ACTIONS(2730),
    [anon_sym_SEMI] = ACTIONS(2730),
    [anon_sym___extension__] = ACTIONS(2728),
    [anon_sym_typedef] = ACTIONS(2728),
    [anon_sym_extern] = ACTIONS(2728),
    [anon_sym___attribute__] = ACTIONS(2728),
    [anon_sym___attribute] = ACTIONS(2728),
    [anon_sym_noreturn] = ACTIONS(2728),
    [anon_sym_LBRACK] = ACTIONS(2730),
    [anon_sym___declspec] = ACTIONS(2728),
    [anon_sym___cdecl] = ACTIONS(2728),
    [anon_sym___clrcall] = ACTIONS(2728),
    [anon_sym___stdcall] = ACTIONS(2728),
    [anon_sym___fastcall] = ACTIONS(2728),
    [anon_sym___thiscall] = ACTIONS(2728),
    [anon_sym___vectorcall] = ACTIONS(2728),
    [anon_sym_LBRACE] = ACTIONS(2730),
    [anon_sym_signed] = ACTIONS(2728),
    [anon_sym_unsigned] = ACTIONS(2728),
    [anon_sym_long] = ACTIONS(2728),
    [anon_sym_short] = ACTIONS(2728),
    [anon_sym_static] = ACTIONS(2728),
    [anon_sym_ATautoreleasepool] = ACTIONS(2730),
    [anon_sym_auto] = ACTIONS(2728),
    [anon_sym_register] = ACTIONS(2728),
    [anon_sym_inline] = ACTIONS(2728),
    [anon_sym___inline] = ACTIONS(2728),
    [anon_sym___inline__] = ACTIONS(2728),
    [anon_sym___forceinline] = ACTIONS(2728),
    [anon_sym_thread_local] = ACTIONS(2728),
    [anon_sym___thread] = ACTIONS(2728),
    [anon_sym_CG_EXTERN] = ACTIONS(2728),
    [anon_sym_CG_INLINE] = ACTIONS(2728),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2728),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2728),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2728),
    [anon_sym_IBOutlet] = ACTIONS(2728),
    [anon_sym_IBInspectable] = ACTIONS(2728),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2728),
    [anon_sym_NS_INLINE] = ACTIONS(2728),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2728),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2728),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2728),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2728),
    [anon_sym_const] = ACTIONS(2728),
    [anon_sym_constexpr] = ACTIONS(2728),
    [anon_sym_volatile] = ACTIONS(2728),
    [anon_sym_restrict] = ACTIONS(2728),
    [anon_sym___restrict__] = ACTIONS(2728),
    [anon_sym__Atomic] = ACTIONS(2728),
    [anon_sym__Noreturn] = ACTIONS(2728),
    [anon_sym__Nonnull] = ACTIONS(2728),
    [anon_sym_nullable] = ACTIONS(2728),
    [anon_sym__Complex] = ACTIONS(2728),
    [anon_sym__Nullable] = ACTIONS(2728),
    [anon_sym__Nullable_result] = ACTIONS(2728),
    [anon_sym__Null_unspecified] = ACTIONS(2728),
    [anon_sym___autoreleasing] = ACTIONS(2728),
    [anon_sym___block] = ACTIONS(2728),
    [anon_sym___bridge] = ACTIONS(2728),
    [anon_sym___bridge_retained] = ACTIONS(2728),
    [anon_sym___bridge_transfer] = ACTIONS(2728),
    [anon_sym___complex] = ACTIONS(2728),
    [anon_sym___const] = ACTIONS(2728),
    [anon_sym___imag] = ACTIONS(2728),
    [anon_sym___kindof] = ACTIONS(2728),
    [anon_sym___nonnull] = ACTIONS(2728),
    [anon_sym___nullable] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2728),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2728),
    [anon_sym___real] = ACTIONS(2728),
    [anon_sym___strong] = ACTIONS(2728),
    [anon_sym___unsafe_unretained] = ACTIONS(2728),
    [anon_sym___unused] = ACTIONS(2728),
    [anon_sym___weak] = ACTIONS(2728),
    [anon_sym_alignas] = ACTIONS(2728),
    [anon_sym__Alignas] = ACTIONS(2728),
    [sym_primitive_type] = ACTIONS(2728),
    [anon_sym_enum] = ACTIONS(2728),
    [anon_sym_struct] = ACTIONS(2728),
    [anon_sym_union] = ACTIONS(2728),
    [anon_sym_if] = ACTIONS(2728),
    [anon_sym_switch] = ACTIONS(2728),
    [anon_sym_case] = ACTIONS(2728),
    [anon_sym_default] = ACTIONS(2728),
    [anon_sym_while] = ACTIONS(2728),
    [anon_sym_do] = ACTIONS(2728),
    [anon_sym_for] = ACTIONS(2728),
    [anon_sym_in] = ACTIONS(2728),
    [anon_sym_return] = ACTIONS(2728),
    [anon_sym_break] = ACTIONS(2728),
    [anon_sym_continue] = ACTIONS(2728),
    [anon_sym_goto] = ACTIONS(2728),
    [anon_sym_DASH_DASH] = ACTIONS(2730),
    [anon_sym_PLUS_PLUS] = ACTIONS(2730),
    [anon_sym_sizeof] = ACTIONS(2728),
    [anon_sym___alignof__] = ACTIONS(2728),
    [anon_sym___alignof] = ACTIONS(2728),
    [anon_sym__alignof] = ACTIONS(2728),
    [anon_sym_alignof] = ACTIONS(2728),
    [anon_sym__Alignof] = ACTIONS(2728),
    [anon_sym_offsetof] = ACTIONS(2728),
    [anon_sym__Generic] = ACTIONS(2728),
    [anon_sym_asm] = ACTIONS(2728),
    [anon_sym___asm__] = ACTIONS(2728),
    [anon_sym___asm] = ACTIONS(2728),
    [sym_number_literal] = ACTIONS(2730),
    [anon_sym_L_SQUOTE] = ACTIONS(2730),
    [anon_sym_u_SQUOTE] = ACTIONS(2730),
    [anon_sym_U_SQUOTE] = ACTIONS(2730),
    [anon_sym_u8_SQUOTE] = ACTIONS(2730),
    [anon_sym_SQUOTE] = ACTIONS(2730),
    [anon_sym_AT] = ACTIONS(2728),
    [anon_sym_DQUOTE] = ACTIONS(2730),
    [anon_sym_L_DQUOTE] = ACTIONS(2730),
    [anon_sym_u_DQUOTE] = ACTIONS(2730),
    [anon_sym_U_DQUOTE] = ACTIONS(2730),
    [anon_sym_u8_DQUOTE] = ACTIONS(2730),
    [sym_true] = ACTIONS(2728),
    [sym_false] = ACTIONS(2728),
    [anon_sym_NULL] = ACTIONS(2728),
    [anon_sym_nullptr] = ACTIONS(2728),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2728),
    [anon_sym___typeof] = ACTIONS(2728),
    [anon_sym_typeof] = ACTIONS(2728),
    [anon_sym_ATimport] = ACTIONS(2730),
    [aux_sym_preproc_undef_token1] = ACTIONS(2728),
    [anon_sym_POUND] = ACTIONS(2728),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2728),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2728),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2728),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2728),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2728),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2728),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2728),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_API_AVAILABLE] = ACTIONS(2728),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_API_DEPRECATED] = ACTIONS(2728),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2728),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2728),
    [anon_sym___deprecated_msg] = ACTIONS(2728),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2728),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2728),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2728),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2728),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2728),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2728),
    [anon_sym_ATprotocol] = ACTIONS(2730),
    [anon_sym_ATinterface] = ACTIONS(2730),
    [anon_sym_ATimplementation] = ACTIONS(2730),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2730),
    [anon_sym_ATselector] = ACTIONS(2730),
    [anon_sym_ATavailable] = ACTIONS(2730),
    [anon_sym___builtin_available] = ACTIONS(2728),
    [anon_sym_va_arg] = ACTIONS(2728),
    [anon_sym_ATencode] = ACTIONS(2730),
    [anon_sym_BOOL] = ACTIONS(2728),
    [anon_sym_IMP] = ACTIONS(2728),
    [anon_sym_SEL] = ACTIONS(2728),
    [anon_sym_Class] = ACTIONS(2728),
    [anon_sym_id] = ACTIONS(2728),
  },
  [STATE(1041)] = {
    [ts_builtin_sym_end] = ACTIONS(2734),
    [sym_identifier] = ACTIONS(2732),
    [aux_sym_preproc_include_token1] = ACTIONS(2732),
    [aux_sym_preproc_include_token2] = ACTIONS(2732),
    [aux_sym_preproc_def_token1] = ACTIONS(2732),
    [aux_sym_preproc_if_token1] = ACTIONS(2732),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2732),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2732),
    [sym_preproc_directive] = ACTIONS(2732),
    [anon_sym_LPAREN2] = ACTIONS(2734),
    [anon_sym_BANG] = ACTIONS(2734),
    [anon_sym_TILDE] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2732),
    [anon_sym_PLUS] = ACTIONS(2732),
    [anon_sym_STAR] = ACTIONS(2734),
    [anon_sym_CARET] = ACTIONS(2734),
    [anon_sym_AMP] = ACTIONS(2734),
    [anon_sym_SEMI] = ACTIONS(2734),
    [anon_sym___extension__] = ACTIONS(2732),
    [anon_sym_typedef] = ACTIONS(2732),
    [anon_sym_extern] = ACTIONS(2732),
    [anon_sym___attribute__] = ACTIONS(2732),
    [anon_sym___attribute] = ACTIONS(2732),
    [anon_sym_noreturn] = ACTIONS(2732),
    [anon_sym_LBRACK] = ACTIONS(2734),
    [anon_sym___declspec] = ACTIONS(2732),
    [anon_sym___cdecl] = ACTIONS(2732),
    [anon_sym___clrcall] = ACTIONS(2732),
    [anon_sym___stdcall] = ACTIONS(2732),
    [anon_sym___fastcall] = ACTIONS(2732),
    [anon_sym___thiscall] = ACTIONS(2732),
    [anon_sym___vectorcall] = ACTIONS(2732),
    [anon_sym_LBRACE] = ACTIONS(2734),
    [anon_sym_signed] = ACTIONS(2732),
    [anon_sym_unsigned] = ACTIONS(2732),
    [anon_sym_long] = ACTIONS(2732),
    [anon_sym_short] = ACTIONS(2732),
    [anon_sym_static] = ACTIONS(2732),
    [anon_sym_ATautoreleasepool] = ACTIONS(2734),
    [anon_sym_auto] = ACTIONS(2732),
    [anon_sym_register] = ACTIONS(2732),
    [anon_sym_inline] = ACTIONS(2732),
    [anon_sym___inline] = ACTIONS(2732),
    [anon_sym___inline__] = ACTIONS(2732),
    [anon_sym___forceinline] = ACTIONS(2732),
    [anon_sym_thread_local] = ACTIONS(2732),
    [anon_sym___thread] = ACTIONS(2732),
    [anon_sym_CG_EXTERN] = ACTIONS(2732),
    [anon_sym_CG_INLINE] = ACTIONS(2732),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2732),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2732),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2732),
    [anon_sym_IBOutlet] = ACTIONS(2732),
    [anon_sym_IBInspectable] = ACTIONS(2732),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2732),
    [anon_sym_NS_INLINE] = ACTIONS(2732),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2732),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2732),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2732),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2732),
    [anon_sym_const] = ACTIONS(2732),
    [anon_sym_constexpr] = ACTIONS(2732),
    [anon_sym_volatile] = ACTIONS(2732),
    [anon_sym_restrict] = ACTIONS(2732),
    [anon_sym___restrict__] = ACTIONS(2732),
    [anon_sym__Atomic] = ACTIONS(2732),
    [anon_sym__Noreturn] = ACTIONS(2732),
    [anon_sym__Nonnull] = ACTIONS(2732),
    [anon_sym_nullable] = ACTIONS(2732),
    [anon_sym__Complex] = ACTIONS(2732),
    [anon_sym__Nullable] = ACTIONS(2732),
    [anon_sym__Nullable_result] = ACTIONS(2732),
    [anon_sym__Null_unspecified] = ACTIONS(2732),
    [anon_sym___autoreleasing] = ACTIONS(2732),
    [anon_sym___block] = ACTIONS(2732),
    [anon_sym___bridge] = ACTIONS(2732),
    [anon_sym___bridge_retained] = ACTIONS(2732),
    [anon_sym___bridge_transfer] = ACTIONS(2732),
    [anon_sym___complex] = ACTIONS(2732),
    [anon_sym___const] = ACTIONS(2732),
    [anon_sym___imag] = ACTIONS(2732),
    [anon_sym___kindof] = ACTIONS(2732),
    [anon_sym___nonnull] = ACTIONS(2732),
    [anon_sym___nullable] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2732),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2732),
    [anon_sym___real] = ACTIONS(2732),
    [anon_sym___strong] = ACTIONS(2732),
    [anon_sym___unsafe_unretained] = ACTIONS(2732),
    [anon_sym___unused] = ACTIONS(2732),
    [anon_sym___weak] = ACTIONS(2732),
    [anon_sym_alignas] = ACTIONS(2732),
    [anon_sym__Alignas] = ACTIONS(2732),
    [sym_primitive_type] = ACTIONS(2732),
    [anon_sym_enum] = ACTIONS(2732),
    [anon_sym_struct] = ACTIONS(2732),
    [anon_sym_union] = ACTIONS(2732),
    [anon_sym_if] = ACTIONS(2732),
    [anon_sym_switch] = ACTIONS(2732),
    [anon_sym_case] = ACTIONS(2732),
    [anon_sym_default] = ACTIONS(2732),
    [anon_sym_while] = ACTIONS(2732),
    [anon_sym_do] = ACTIONS(2732),
    [anon_sym_for] = ACTIONS(2732),
    [anon_sym_in] = ACTIONS(2732),
    [anon_sym_return] = ACTIONS(2732),
    [anon_sym_break] = ACTIONS(2732),
    [anon_sym_continue] = ACTIONS(2732),
    [anon_sym_goto] = ACTIONS(2732),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_sizeof] = ACTIONS(2732),
    [anon_sym___alignof__] = ACTIONS(2732),
    [anon_sym___alignof] = ACTIONS(2732),
    [anon_sym__alignof] = ACTIONS(2732),
    [anon_sym_alignof] = ACTIONS(2732),
    [anon_sym__Alignof] = ACTIONS(2732),
    [anon_sym_offsetof] = ACTIONS(2732),
    [anon_sym__Generic] = ACTIONS(2732),
    [anon_sym_asm] = ACTIONS(2732),
    [anon_sym___asm__] = ACTIONS(2732),
    [anon_sym___asm] = ACTIONS(2732),
    [sym_number_literal] = ACTIONS(2734),
    [anon_sym_L_SQUOTE] = ACTIONS(2734),
    [anon_sym_u_SQUOTE] = ACTIONS(2734),
    [anon_sym_U_SQUOTE] = ACTIONS(2734),
    [anon_sym_u8_SQUOTE] = ACTIONS(2734),
    [anon_sym_SQUOTE] = ACTIONS(2734),
    [anon_sym_AT] = ACTIONS(2732),
    [anon_sym_DQUOTE] = ACTIONS(2734),
    [anon_sym_L_DQUOTE] = ACTIONS(2734),
    [anon_sym_u_DQUOTE] = ACTIONS(2734),
    [anon_sym_U_DQUOTE] = ACTIONS(2734),
    [anon_sym_u8_DQUOTE] = ACTIONS(2734),
    [sym_true] = ACTIONS(2732),
    [sym_false] = ACTIONS(2732),
    [anon_sym_NULL] = ACTIONS(2732),
    [anon_sym_nullptr] = ACTIONS(2732),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2732),
    [anon_sym___typeof] = ACTIONS(2732),
    [anon_sym_typeof] = ACTIONS(2732),
    [anon_sym_ATimport] = ACTIONS(2734),
    [aux_sym_preproc_undef_token1] = ACTIONS(2732),
    [anon_sym_POUND] = ACTIONS(2732),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2732),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2732),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2732),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2732),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2732),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2732),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2732),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_API_AVAILABLE] = ACTIONS(2732),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_API_DEPRECATED] = ACTIONS(2732),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2732),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2732),
    [anon_sym___deprecated_msg] = ACTIONS(2732),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2732),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2732),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2732),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2732),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2732),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2732),
    [anon_sym_ATprotocol] = ACTIONS(2734),
    [anon_sym_ATinterface] = ACTIONS(2734),
    [anon_sym_ATimplementation] = ACTIONS(2734),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2734),
    [anon_sym_ATselector] = ACTIONS(2734),
    [anon_sym_ATavailable] = ACTIONS(2734),
    [anon_sym___builtin_available] = ACTIONS(2732),
    [anon_sym_va_arg] = ACTIONS(2732),
    [anon_sym_ATencode] = ACTIONS(2734),
    [anon_sym_BOOL] = ACTIONS(2732),
    [anon_sym_IMP] = ACTIONS(2732),
    [anon_sym_SEL] = ACTIONS(2732),
    [anon_sym_Class] = ACTIONS(2732),
    [anon_sym_id] = ACTIONS(2732),
  },
  [STATE(1042)] = {
    [ts_builtin_sym_end] = ACTIONS(2738),
    [sym_identifier] = ACTIONS(2736),
    [aux_sym_preproc_include_token1] = ACTIONS(2736),
    [aux_sym_preproc_include_token2] = ACTIONS(2736),
    [aux_sym_preproc_def_token1] = ACTIONS(2736),
    [aux_sym_preproc_if_token1] = ACTIONS(2736),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2736),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2736),
    [sym_preproc_directive] = ACTIONS(2736),
    [anon_sym_LPAREN2] = ACTIONS(2738),
    [anon_sym_BANG] = ACTIONS(2738),
    [anon_sym_TILDE] = ACTIONS(2738),
    [anon_sym_DASH] = ACTIONS(2736),
    [anon_sym_PLUS] = ACTIONS(2736),
    [anon_sym_STAR] = ACTIONS(2738),
    [anon_sym_CARET] = ACTIONS(2738),
    [anon_sym_AMP] = ACTIONS(2738),
    [anon_sym_SEMI] = ACTIONS(2738),
    [anon_sym___extension__] = ACTIONS(2736),
    [anon_sym_typedef] = ACTIONS(2736),
    [anon_sym_extern] = ACTIONS(2736),
    [anon_sym___attribute__] = ACTIONS(2736),
    [anon_sym___attribute] = ACTIONS(2736),
    [anon_sym_noreturn] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(2738),
    [anon_sym___declspec] = ACTIONS(2736),
    [anon_sym___cdecl] = ACTIONS(2736),
    [anon_sym___clrcall] = ACTIONS(2736),
    [anon_sym___stdcall] = ACTIONS(2736),
    [anon_sym___fastcall] = ACTIONS(2736),
    [anon_sym___thiscall] = ACTIONS(2736),
    [anon_sym___vectorcall] = ACTIONS(2736),
    [anon_sym_LBRACE] = ACTIONS(2738),
    [anon_sym_signed] = ACTIONS(2736),
    [anon_sym_unsigned] = ACTIONS(2736),
    [anon_sym_long] = ACTIONS(2736),
    [anon_sym_short] = ACTIONS(2736),
    [anon_sym_static] = ACTIONS(2736),
    [anon_sym_ATautoreleasepool] = ACTIONS(2738),
    [anon_sym_auto] = ACTIONS(2736),
    [anon_sym_register] = ACTIONS(2736),
    [anon_sym_inline] = ACTIONS(2736),
    [anon_sym___inline] = ACTIONS(2736),
    [anon_sym___inline__] = ACTIONS(2736),
    [anon_sym___forceinline] = ACTIONS(2736),
    [anon_sym_thread_local] = ACTIONS(2736),
    [anon_sym___thread] = ACTIONS(2736),
    [anon_sym_CG_EXTERN] = ACTIONS(2736),
    [anon_sym_CG_INLINE] = ACTIONS(2736),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2736),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2736),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2736),
    [anon_sym_IBOutlet] = ACTIONS(2736),
    [anon_sym_IBInspectable] = ACTIONS(2736),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2736),
    [anon_sym_NS_INLINE] = ACTIONS(2736),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2736),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2736),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2736),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2736),
    [anon_sym_const] = ACTIONS(2736),
    [anon_sym_constexpr] = ACTIONS(2736),
    [anon_sym_volatile] = ACTIONS(2736),
    [anon_sym_restrict] = ACTIONS(2736),
    [anon_sym___restrict__] = ACTIONS(2736),
    [anon_sym__Atomic] = ACTIONS(2736),
    [anon_sym__Noreturn] = ACTIONS(2736),
    [anon_sym__Nonnull] = ACTIONS(2736),
    [anon_sym_nullable] = ACTIONS(2736),
    [anon_sym__Complex] = ACTIONS(2736),
    [anon_sym__Nullable] = ACTIONS(2736),
    [anon_sym__Nullable_result] = ACTIONS(2736),
    [anon_sym__Null_unspecified] = ACTIONS(2736),
    [anon_sym___autoreleasing] = ACTIONS(2736),
    [anon_sym___block] = ACTIONS(2736),
    [anon_sym___bridge] = ACTIONS(2736),
    [anon_sym___bridge_retained] = ACTIONS(2736),
    [anon_sym___bridge_transfer] = ACTIONS(2736),
    [anon_sym___complex] = ACTIONS(2736),
    [anon_sym___const] = ACTIONS(2736),
    [anon_sym___imag] = ACTIONS(2736),
    [anon_sym___kindof] = ACTIONS(2736),
    [anon_sym___nonnull] = ACTIONS(2736),
    [anon_sym___nullable] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2736),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2736),
    [anon_sym___real] = ACTIONS(2736),
    [anon_sym___strong] = ACTIONS(2736),
    [anon_sym___unsafe_unretained] = ACTIONS(2736),
    [anon_sym___unused] = ACTIONS(2736),
    [anon_sym___weak] = ACTIONS(2736),
    [anon_sym_alignas] = ACTIONS(2736),
    [anon_sym__Alignas] = ACTIONS(2736),
    [sym_primitive_type] = ACTIONS(2736),
    [anon_sym_enum] = ACTIONS(2736),
    [anon_sym_struct] = ACTIONS(2736),
    [anon_sym_union] = ACTIONS(2736),
    [anon_sym_if] = ACTIONS(2736),
    [anon_sym_switch] = ACTIONS(2736),
    [anon_sym_case] = ACTIONS(2736),
    [anon_sym_default] = ACTIONS(2736),
    [anon_sym_while] = ACTIONS(2736),
    [anon_sym_do] = ACTIONS(2736),
    [anon_sym_for] = ACTIONS(2736),
    [anon_sym_in] = ACTIONS(2736),
    [anon_sym_return] = ACTIONS(2736),
    [anon_sym_break] = ACTIONS(2736),
    [anon_sym_continue] = ACTIONS(2736),
    [anon_sym_goto] = ACTIONS(2736),
    [anon_sym_DASH_DASH] = ACTIONS(2738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2738),
    [anon_sym_sizeof] = ACTIONS(2736),
    [anon_sym___alignof__] = ACTIONS(2736),
    [anon_sym___alignof] = ACTIONS(2736),
    [anon_sym__alignof] = ACTIONS(2736),
    [anon_sym_alignof] = ACTIONS(2736),
    [anon_sym__Alignof] = ACTIONS(2736),
    [anon_sym_offsetof] = ACTIONS(2736),
    [anon_sym__Generic] = ACTIONS(2736),
    [anon_sym_asm] = ACTIONS(2736),
    [anon_sym___asm__] = ACTIONS(2736),
    [anon_sym___asm] = ACTIONS(2736),
    [sym_number_literal] = ACTIONS(2738),
    [anon_sym_L_SQUOTE] = ACTIONS(2738),
    [anon_sym_u_SQUOTE] = ACTIONS(2738),
    [anon_sym_U_SQUOTE] = ACTIONS(2738),
    [anon_sym_u8_SQUOTE] = ACTIONS(2738),
    [anon_sym_SQUOTE] = ACTIONS(2738),
    [anon_sym_AT] = ACTIONS(2736),
    [anon_sym_DQUOTE] = ACTIONS(2738),
    [anon_sym_L_DQUOTE] = ACTIONS(2738),
    [anon_sym_u_DQUOTE] = ACTIONS(2738),
    [anon_sym_U_DQUOTE] = ACTIONS(2738),
    [anon_sym_u8_DQUOTE] = ACTIONS(2738),
    [sym_true] = ACTIONS(2736),
    [sym_false] = ACTIONS(2736),
    [anon_sym_NULL] = ACTIONS(2736),
    [anon_sym_nullptr] = ACTIONS(2736),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2736),
    [anon_sym___typeof] = ACTIONS(2736),
    [anon_sym_typeof] = ACTIONS(2736),
    [anon_sym_ATimport] = ACTIONS(2738),
    [aux_sym_preproc_undef_token1] = ACTIONS(2736),
    [anon_sym_POUND] = ACTIONS(2736),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2736),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2736),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2736),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2736),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2736),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2736),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2736),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_API_AVAILABLE] = ACTIONS(2736),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_API_DEPRECATED] = ACTIONS(2736),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2736),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2736),
    [anon_sym___deprecated_msg] = ACTIONS(2736),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2736),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2736),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2736),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2736),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2736),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2736),
    [anon_sym_ATprotocol] = ACTIONS(2738),
    [anon_sym_ATinterface] = ACTIONS(2738),
    [anon_sym_ATimplementation] = ACTIONS(2738),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2738),
    [anon_sym_ATselector] = ACTIONS(2738),
    [anon_sym_ATavailable] = ACTIONS(2738),
    [anon_sym___builtin_available] = ACTIONS(2736),
    [anon_sym_va_arg] = ACTIONS(2736),
    [anon_sym_ATencode] = ACTIONS(2738),
    [anon_sym_BOOL] = ACTIONS(2736),
    [anon_sym_IMP] = ACTIONS(2736),
    [anon_sym_SEL] = ACTIONS(2736),
    [anon_sym_Class] = ACTIONS(2736),
    [anon_sym_id] = ACTIONS(2736),
  },
  [STATE(1043)] = {
    [ts_builtin_sym_end] = ACTIONS(2590),
    [sym_identifier] = ACTIONS(2588),
    [aux_sym_preproc_include_token1] = ACTIONS(2588),
    [aux_sym_preproc_include_token2] = ACTIONS(2588),
    [aux_sym_preproc_def_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2588),
    [sym_preproc_directive] = ACTIONS(2588),
    [anon_sym_LPAREN2] = ACTIONS(2590),
    [anon_sym_BANG] = ACTIONS(2590),
    [anon_sym_TILDE] = ACTIONS(2590),
    [anon_sym_DASH] = ACTIONS(2588),
    [anon_sym_PLUS] = ACTIONS(2588),
    [anon_sym_STAR] = ACTIONS(2590),
    [anon_sym_CARET] = ACTIONS(2590),
    [anon_sym_AMP] = ACTIONS(2590),
    [anon_sym_SEMI] = ACTIONS(2590),
    [anon_sym___extension__] = ACTIONS(2588),
    [anon_sym_typedef] = ACTIONS(2588),
    [anon_sym_extern] = ACTIONS(2588),
    [anon_sym___attribute__] = ACTIONS(2588),
    [anon_sym___attribute] = ACTIONS(2588),
    [anon_sym_noreturn] = ACTIONS(2588),
    [anon_sym_LBRACK] = ACTIONS(2590),
    [anon_sym___declspec] = ACTIONS(2588),
    [anon_sym___cdecl] = ACTIONS(2588),
    [anon_sym___clrcall] = ACTIONS(2588),
    [anon_sym___stdcall] = ACTIONS(2588),
    [anon_sym___fastcall] = ACTIONS(2588),
    [anon_sym___thiscall] = ACTIONS(2588),
    [anon_sym___vectorcall] = ACTIONS(2588),
    [anon_sym_LBRACE] = ACTIONS(2590),
    [anon_sym_signed] = ACTIONS(2588),
    [anon_sym_unsigned] = ACTIONS(2588),
    [anon_sym_long] = ACTIONS(2588),
    [anon_sym_short] = ACTIONS(2588),
    [anon_sym_static] = ACTIONS(2588),
    [anon_sym_ATautoreleasepool] = ACTIONS(2590),
    [anon_sym_auto] = ACTIONS(2588),
    [anon_sym_register] = ACTIONS(2588),
    [anon_sym_inline] = ACTIONS(2588),
    [anon_sym___inline] = ACTIONS(2588),
    [anon_sym___inline__] = ACTIONS(2588),
    [anon_sym___forceinline] = ACTIONS(2588),
    [anon_sym_thread_local] = ACTIONS(2588),
    [anon_sym___thread] = ACTIONS(2588),
    [anon_sym_CG_EXTERN] = ACTIONS(2588),
    [anon_sym_CG_INLINE] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2588),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2588),
    [anon_sym_IBOutlet] = ACTIONS(2588),
    [anon_sym_IBInspectable] = ACTIONS(2588),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2588),
    [anon_sym_NS_INLINE] = ACTIONS(2588),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2588),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2588),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2588),
    [anon_sym_const] = ACTIONS(2588),
    [anon_sym_constexpr] = ACTIONS(2588),
    [anon_sym_volatile] = ACTIONS(2588),
    [anon_sym_restrict] = ACTIONS(2588),
    [anon_sym___restrict__] = ACTIONS(2588),
    [anon_sym__Atomic] = ACTIONS(2588),
    [anon_sym__Noreturn] = ACTIONS(2588),
    [anon_sym__Nonnull] = ACTIONS(2588),
    [anon_sym_nullable] = ACTIONS(2588),
    [anon_sym__Complex] = ACTIONS(2588),
    [anon_sym__Nullable] = ACTIONS(2588),
    [anon_sym__Nullable_result] = ACTIONS(2588),
    [anon_sym__Null_unspecified] = ACTIONS(2588),
    [anon_sym___autoreleasing] = ACTIONS(2588),
    [anon_sym___block] = ACTIONS(2588),
    [anon_sym___bridge] = ACTIONS(2588),
    [anon_sym___bridge_retained] = ACTIONS(2588),
    [anon_sym___bridge_transfer] = ACTIONS(2588),
    [anon_sym___complex] = ACTIONS(2588),
    [anon_sym___const] = ACTIONS(2588),
    [anon_sym___imag] = ACTIONS(2588),
    [anon_sym___kindof] = ACTIONS(2588),
    [anon_sym___nonnull] = ACTIONS(2588),
    [anon_sym___nullable] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2588),
    [anon_sym___real] = ACTIONS(2588),
    [anon_sym___strong] = ACTIONS(2588),
    [anon_sym___unsafe_unretained] = ACTIONS(2588),
    [anon_sym___unused] = ACTIONS(2588),
    [anon_sym___weak] = ACTIONS(2588),
    [anon_sym_alignas] = ACTIONS(2588),
    [anon_sym__Alignas] = ACTIONS(2588),
    [sym_primitive_type] = ACTIONS(2588),
    [anon_sym_enum] = ACTIONS(2588),
    [anon_sym_struct] = ACTIONS(2588),
    [anon_sym_union] = ACTIONS(2588),
    [anon_sym_if] = ACTIONS(2588),
    [anon_sym_switch] = ACTIONS(2588),
    [anon_sym_case] = ACTIONS(2588),
    [anon_sym_default] = ACTIONS(2588),
    [anon_sym_while] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2588),
    [anon_sym_for] = ACTIONS(2588),
    [anon_sym_in] = ACTIONS(2588),
    [anon_sym_return] = ACTIONS(2588),
    [anon_sym_break] = ACTIONS(2588),
    [anon_sym_continue] = ACTIONS(2588),
    [anon_sym_goto] = ACTIONS(2588),
    [anon_sym_DASH_DASH] = ACTIONS(2590),
    [anon_sym_PLUS_PLUS] = ACTIONS(2590),
    [anon_sym_sizeof] = ACTIONS(2588),
    [anon_sym___alignof__] = ACTIONS(2588),
    [anon_sym___alignof] = ACTIONS(2588),
    [anon_sym__alignof] = ACTIONS(2588),
    [anon_sym_alignof] = ACTIONS(2588),
    [anon_sym__Alignof] = ACTIONS(2588),
    [anon_sym_offsetof] = ACTIONS(2588),
    [anon_sym__Generic] = ACTIONS(2588),
    [anon_sym_asm] = ACTIONS(2588),
    [anon_sym___asm__] = ACTIONS(2588),
    [anon_sym___asm] = ACTIONS(2588),
    [sym_number_literal] = ACTIONS(2590),
    [anon_sym_L_SQUOTE] = ACTIONS(2590),
    [anon_sym_u_SQUOTE] = ACTIONS(2590),
    [anon_sym_U_SQUOTE] = ACTIONS(2590),
    [anon_sym_u8_SQUOTE] = ACTIONS(2590),
    [anon_sym_SQUOTE] = ACTIONS(2590),
    [anon_sym_AT] = ACTIONS(2588),
    [anon_sym_DQUOTE] = ACTIONS(2590),
    [anon_sym_L_DQUOTE] = ACTIONS(2590),
    [anon_sym_u_DQUOTE] = ACTIONS(2590),
    [anon_sym_U_DQUOTE] = ACTIONS(2590),
    [anon_sym_u8_DQUOTE] = ACTIONS(2590),
    [sym_true] = ACTIONS(2588),
    [sym_false] = ACTIONS(2588),
    [anon_sym_NULL] = ACTIONS(2588),
    [anon_sym_nullptr] = ACTIONS(2588),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2588),
    [anon_sym___typeof] = ACTIONS(2588),
    [anon_sym_typeof] = ACTIONS(2588),
    [anon_sym_ATimport] = ACTIONS(2590),
    [aux_sym_preproc_undef_token1] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2588),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2588),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2588),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2588),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_API_AVAILABLE] = ACTIONS(2588),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_API_DEPRECATED] = ACTIONS(2588),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym___deprecated_msg] = ACTIONS(2588),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2588),
    [anon_sym_ATprotocol] = ACTIONS(2590),
    [anon_sym_ATinterface] = ACTIONS(2590),
    [anon_sym_ATimplementation] = ACTIONS(2590),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2590),
    [anon_sym_ATselector] = ACTIONS(2590),
    [anon_sym_ATavailable] = ACTIONS(2590),
    [anon_sym___builtin_available] = ACTIONS(2588),
    [anon_sym_va_arg] = ACTIONS(2588),
    [anon_sym_ATencode] = ACTIONS(2590),
    [anon_sym_BOOL] = ACTIONS(2588),
    [anon_sym_IMP] = ACTIONS(2588),
    [anon_sym_SEL] = ACTIONS(2588),
    [anon_sym_Class] = ACTIONS(2588),
    [anon_sym_id] = ACTIONS(2588),
  },
  [STATE(1044)] = {
    [ts_builtin_sym_end] = ACTIONS(2718),
    [sym_identifier] = ACTIONS(2716),
    [aux_sym_preproc_include_token1] = ACTIONS(2716),
    [aux_sym_preproc_include_token2] = ACTIONS(2716),
    [aux_sym_preproc_def_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2716),
    [sym_preproc_directive] = ACTIONS(2716),
    [anon_sym_LPAREN2] = ACTIONS(2718),
    [anon_sym_BANG] = ACTIONS(2718),
    [anon_sym_TILDE] = ACTIONS(2718),
    [anon_sym_DASH] = ACTIONS(2716),
    [anon_sym_PLUS] = ACTIONS(2716),
    [anon_sym_STAR] = ACTIONS(2718),
    [anon_sym_CARET] = ACTIONS(2718),
    [anon_sym_AMP] = ACTIONS(2718),
    [anon_sym_SEMI] = ACTIONS(2718),
    [anon_sym___extension__] = ACTIONS(2716),
    [anon_sym_typedef] = ACTIONS(2716),
    [anon_sym_extern] = ACTIONS(2716),
    [anon_sym___attribute__] = ACTIONS(2716),
    [anon_sym___attribute] = ACTIONS(2716),
    [anon_sym_noreturn] = ACTIONS(2716),
    [anon_sym_LBRACK] = ACTIONS(2718),
    [anon_sym___declspec] = ACTIONS(2716),
    [anon_sym___cdecl] = ACTIONS(2716),
    [anon_sym___clrcall] = ACTIONS(2716),
    [anon_sym___stdcall] = ACTIONS(2716),
    [anon_sym___fastcall] = ACTIONS(2716),
    [anon_sym___thiscall] = ACTIONS(2716),
    [anon_sym___vectorcall] = ACTIONS(2716),
    [anon_sym_LBRACE] = ACTIONS(2718),
    [anon_sym_signed] = ACTIONS(2716),
    [anon_sym_unsigned] = ACTIONS(2716),
    [anon_sym_long] = ACTIONS(2716),
    [anon_sym_short] = ACTIONS(2716),
    [anon_sym_static] = ACTIONS(2716),
    [anon_sym_ATautoreleasepool] = ACTIONS(2718),
    [anon_sym_auto] = ACTIONS(2716),
    [anon_sym_register] = ACTIONS(2716),
    [anon_sym_inline] = ACTIONS(2716),
    [anon_sym___inline] = ACTIONS(2716),
    [anon_sym___inline__] = ACTIONS(2716),
    [anon_sym___forceinline] = ACTIONS(2716),
    [anon_sym_thread_local] = ACTIONS(2716),
    [anon_sym___thread] = ACTIONS(2716),
    [anon_sym_CG_EXTERN] = ACTIONS(2716),
    [anon_sym_CG_INLINE] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2716),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2716),
    [anon_sym_IBOutlet] = ACTIONS(2716),
    [anon_sym_IBInspectable] = ACTIONS(2716),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2716),
    [anon_sym_NS_INLINE] = ACTIONS(2716),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2716),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2716),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2716),
    [anon_sym_const] = ACTIONS(2716),
    [anon_sym_constexpr] = ACTIONS(2716),
    [anon_sym_volatile] = ACTIONS(2716),
    [anon_sym_restrict] = ACTIONS(2716),
    [anon_sym___restrict__] = ACTIONS(2716),
    [anon_sym__Atomic] = ACTIONS(2716),
    [anon_sym__Noreturn] = ACTIONS(2716),
    [anon_sym__Nonnull] = ACTIONS(2716),
    [anon_sym_nullable] = ACTIONS(2716),
    [anon_sym__Complex] = ACTIONS(2716),
    [anon_sym__Nullable] = ACTIONS(2716),
    [anon_sym__Nullable_result] = ACTIONS(2716),
    [anon_sym__Null_unspecified] = ACTIONS(2716),
    [anon_sym___autoreleasing] = ACTIONS(2716),
    [anon_sym___block] = ACTIONS(2716),
    [anon_sym___bridge] = ACTIONS(2716),
    [anon_sym___bridge_retained] = ACTIONS(2716),
    [anon_sym___bridge_transfer] = ACTIONS(2716),
    [anon_sym___complex] = ACTIONS(2716),
    [anon_sym___const] = ACTIONS(2716),
    [anon_sym___imag] = ACTIONS(2716),
    [anon_sym___kindof] = ACTIONS(2716),
    [anon_sym___nonnull] = ACTIONS(2716),
    [anon_sym___nullable] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2716),
    [anon_sym___real] = ACTIONS(2716),
    [anon_sym___strong] = ACTIONS(2716),
    [anon_sym___unsafe_unretained] = ACTIONS(2716),
    [anon_sym___unused] = ACTIONS(2716),
    [anon_sym___weak] = ACTIONS(2716),
    [anon_sym_alignas] = ACTIONS(2716),
    [anon_sym__Alignas] = ACTIONS(2716),
    [sym_primitive_type] = ACTIONS(2716),
    [anon_sym_enum] = ACTIONS(2716),
    [anon_sym_struct] = ACTIONS(2716),
    [anon_sym_union] = ACTIONS(2716),
    [anon_sym_if] = ACTIONS(2716),
    [anon_sym_switch] = ACTIONS(2716),
    [anon_sym_case] = ACTIONS(2716),
    [anon_sym_default] = ACTIONS(2716),
    [anon_sym_while] = ACTIONS(2716),
    [anon_sym_do] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2716),
    [anon_sym_in] = ACTIONS(2716),
    [anon_sym_return] = ACTIONS(2716),
    [anon_sym_break] = ACTIONS(2716),
    [anon_sym_continue] = ACTIONS(2716),
    [anon_sym_goto] = ACTIONS(2716),
    [anon_sym_DASH_DASH] = ACTIONS(2718),
    [anon_sym_PLUS_PLUS] = ACTIONS(2718),
    [anon_sym_sizeof] = ACTIONS(2716),
    [anon_sym___alignof__] = ACTIONS(2716),
    [anon_sym___alignof] = ACTIONS(2716),
    [anon_sym__alignof] = ACTIONS(2716),
    [anon_sym_alignof] = ACTIONS(2716),
    [anon_sym__Alignof] = ACTIONS(2716),
    [anon_sym_offsetof] = ACTIONS(2716),
    [anon_sym__Generic] = ACTIONS(2716),
    [anon_sym_asm] = ACTIONS(2716),
    [anon_sym___asm__] = ACTIONS(2716),
    [anon_sym___asm] = ACTIONS(2716),
    [sym_number_literal] = ACTIONS(2718),
    [anon_sym_L_SQUOTE] = ACTIONS(2718),
    [anon_sym_u_SQUOTE] = ACTIONS(2718),
    [anon_sym_U_SQUOTE] = ACTIONS(2718),
    [anon_sym_u8_SQUOTE] = ACTIONS(2718),
    [anon_sym_SQUOTE] = ACTIONS(2718),
    [anon_sym_AT] = ACTIONS(2716),
    [anon_sym_DQUOTE] = ACTIONS(2718),
    [anon_sym_L_DQUOTE] = ACTIONS(2718),
    [anon_sym_u_DQUOTE] = ACTIONS(2718),
    [anon_sym_U_DQUOTE] = ACTIONS(2718),
    [anon_sym_u8_DQUOTE] = ACTIONS(2718),
    [sym_true] = ACTIONS(2716),
    [sym_false] = ACTIONS(2716),
    [anon_sym_NULL] = ACTIONS(2716),
    [anon_sym_nullptr] = ACTIONS(2716),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2716),
    [anon_sym___typeof] = ACTIONS(2716),
    [anon_sym_typeof] = ACTIONS(2716),
    [anon_sym_ATimport] = ACTIONS(2718),
    [aux_sym_preproc_undef_token1] = ACTIONS(2716),
    [anon_sym_POUND] = ACTIONS(2716),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2716),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2716),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2716),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_API_AVAILABLE] = ACTIONS(2716),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_API_DEPRECATED] = ACTIONS(2716),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym___deprecated_msg] = ACTIONS(2716),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2716),
    [anon_sym_ATprotocol] = ACTIONS(2718),
    [anon_sym_ATinterface] = ACTIONS(2718),
    [anon_sym_ATimplementation] = ACTIONS(2718),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2718),
    [anon_sym_ATselector] = ACTIONS(2718),
    [anon_sym_ATavailable] = ACTIONS(2718),
    [anon_sym___builtin_available] = ACTIONS(2716),
    [anon_sym_va_arg] = ACTIONS(2716),
    [anon_sym_ATencode] = ACTIONS(2718),
    [anon_sym_BOOL] = ACTIONS(2716),
    [anon_sym_IMP] = ACTIONS(2716),
    [anon_sym_SEL] = ACTIONS(2716),
    [anon_sym_Class] = ACTIONS(2716),
    [anon_sym_id] = ACTIONS(2716),
  },
  [STATE(1045)] = {
    [ts_builtin_sym_end] = ACTIONS(2742),
    [sym_identifier] = ACTIONS(2740),
    [aux_sym_preproc_include_token1] = ACTIONS(2740),
    [aux_sym_preproc_include_token2] = ACTIONS(2740),
    [aux_sym_preproc_def_token1] = ACTIONS(2740),
    [aux_sym_preproc_if_token1] = ACTIONS(2740),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2740),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2740),
    [sym_preproc_directive] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_BANG] = ACTIONS(2742),
    [anon_sym_TILDE] = ACTIONS(2742),
    [anon_sym_DASH] = ACTIONS(2740),
    [anon_sym_PLUS] = ACTIONS(2740),
    [anon_sym_STAR] = ACTIONS(2742),
    [anon_sym_CARET] = ACTIONS(2742),
    [anon_sym_AMP] = ACTIONS(2742),
    [anon_sym_SEMI] = ACTIONS(2742),
    [anon_sym___extension__] = ACTIONS(2740),
    [anon_sym_typedef] = ACTIONS(2740),
    [anon_sym_extern] = ACTIONS(2740),
    [anon_sym___attribute__] = ACTIONS(2740),
    [anon_sym___attribute] = ACTIONS(2740),
    [anon_sym_noreturn] = ACTIONS(2740),
    [anon_sym_LBRACK] = ACTIONS(2742),
    [anon_sym___declspec] = ACTIONS(2740),
    [anon_sym___cdecl] = ACTIONS(2740),
    [anon_sym___clrcall] = ACTIONS(2740),
    [anon_sym___stdcall] = ACTIONS(2740),
    [anon_sym___fastcall] = ACTIONS(2740),
    [anon_sym___thiscall] = ACTIONS(2740),
    [anon_sym___vectorcall] = ACTIONS(2740),
    [anon_sym_LBRACE] = ACTIONS(2742),
    [anon_sym_signed] = ACTIONS(2740),
    [anon_sym_unsigned] = ACTIONS(2740),
    [anon_sym_long] = ACTIONS(2740),
    [anon_sym_short] = ACTIONS(2740),
    [anon_sym_static] = ACTIONS(2740),
    [anon_sym_ATautoreleasepool] = ACTIONS(2742),
    [anon_sym_auto] = ACTIONS(2740),
    [anon_sym_register] = ACTIONS(2740),
    [anon_sym_inline] = ACTIONS(2740),
    [anon_sym___inline] = ACTIONS(2740),
    [anon_sym___inline__] = ACTIONS(2740),
    [anon_sym___forceinline] = ACTIONS(2740),
    [anon_sym_thread_local] = ACTIONS(2740),
    [anon_sym___thread] = ACTIONS(2740),
    [anon_sym_CG_EXTERN] = ACTIONS(2740),
    [anon_sym_CG_INLINE] = ACTIONS(2740),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2740),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2740),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2740),
    [anon_sym_IBOutlet] = ACTIONS(2740),
    [anon_sym_IBInspectable] = ACTIONS(2740),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2740),
    [anon_sym_NS_INLINE] = ACTIONS(2740),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2740),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2740),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2740),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2740),
    [anon_sym_const] = ACTIONS(2740),
    [anon_sym_constexpr] = ACTIONS(2740),
    [anon_sym_volatile] = ACTIONS(2740),
    [anon_sym_restrict] = ACTIONS(2740),
    [anon_sym___restrict__] = ACTIONS(2740),
    [anon_sym__Atomic] = ACTIONS(2740),
    [anon_sym__Noreturn] = ACTIONS(2740),
    [anon_sym__Nonnull] = ACTIONS(2740),
    [anon_sym_nullable] = ACTIONS(2740),
    [anon_sym__Complex] = ACTIONS(2740),
    [anon_sym__Nullable] = ACTIONS(2740),
    [anon_sym__Nullable_result] = ACTIONS(2740),
    [anon_sym__Null_unspecified] = ACTIONS(2740),
    [anon_sym___autoreleasing] = ACTIONS(2740),
    [anon_sym___block] = ACTIONS(2740),
    [anon_sym___bridge] = ACTIONS(2740),
    [anon_sym___bridge_retained] = ACTIONS(2740),
    [anon_sym___bridge_transfer] = ACTIONS(2740),
    [anon_sym___complex] = ACTIONS(2740),
    [anon_sym___const] = ACTIONS(2740),
    [anon_sym___imag] = ACTIONS(2740),
    [anon_sym___kindof] = ACTIONS(2740),
    [anon_sym___nonnull] = ACTIONS(2740),
    [anon_sym___nullable] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2740),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2740),
    [anon_sym___real] = ACTIONS(2740),
    [anon_sym___strong] = ACTIONS(2740),
    [anon_sym___unsafe_unretained] = ACTIONS(2740),
    [anon_sym___unused] = ACTIONS(2740),
    [anon_sym___weak] = ACTIONS(2740),
    [anon_sym_alignas] = ACTIONS(2740),
    [anon_sym__Alignas] = ACTIONS(2740),
    [sym_primitive_type] = ACTIONS(2740),
    [anon_sym_enum] = ACTIONS(2740),
    [anon_sym_struct] = ACTIONS(2740),
    [anon_sym_union] = ACTIONS(2740),
    [anon_sym_if] = ACTIONS(2740),
    [anon_sym_switch] = ACTIONS(2740),
    [anon_sym_case] = ACTIONS(2740),
    [anon_sym_default] = ACTIONS(2740),
    [anon_sym_while] = ACTIONS(2740),
    [anon_sym_do] = ACTIONS(2740),
    [anon_sym_for] = ACTIONS(2740),
    [anon_sym_in] = ACTIONS(2740),
    [anon_sym_return] = ACTIONS(2740),
    [anon_sym_break] = ACTIONS(2740),
    [anon_sym_continue] = ACTIONS(2740),
    [anon_sym_goto] = ACTIONS(2740),
    [anon_sym_DASH_DASH] = ACTIONS(2742),
    [anon_sym_PLUS_PLUS] = ACTIONS(2742),
    [anon_sym_sizeof] = ACTIONS(2740),
    [anon_sym___alignof__] = ACTIONS(2740),
    [anon_sym___alignof] = ACTIONS(2740),
    [anon_sym__alignof] = ACTIONS(2740),
    [anon_sym_alignof] = ACTIONS(2740),
    [anon_sym__Alignof] = ACTIONS(2740),
    [anon_sym_offsetof] = ACTIONS(2740),
    [anon_sym__Generic] = ACTIONS(2740),
    [anon_sym_asm] = ACTIONS(2740),
    [anon_sym___asm__] = ACTIONS(2740),
    [anon_sym___asm] = ACTIONS(2740),
    [sym_number_literal] = ACTIONS(2742),
    [anon_sym_L_SQUOTE] = ACTIONS(2742),
    [anon_sym_u_SQUOTE] = ACTIONS(2742),
    [anon_sym_U_SQUOTE] = ACTIONS(2742),
    [anon_sym_u8_SQUOTE] = ACTIONS(2742),
    [anon_sym_SQUOTE] = ACTIONS(2742),
    [anon_sym_AT] = ACTIONS(2740),
    [anon_sym_DQUOTE] = ACTIONS(2742),
    [anon_sym_L_DQUOTE] = ACTIONS(2742),
    [anon_sym_u_DQUOTE] = ACTIONS(2742),
    [anon_sym_U_DQUOTE] = ACTIONS(2742),
    [anon_sym_u8_DQUOTE] = ACTIONS(2742),
    [sym_true] = ACTIONS(2740),
    [sym_false] = ACTIONS(2740),
    [anon_sym_NULL] = ACTIONS(2740),
    [anon_sym_nullptr] = ACTIONS(2740),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2740),
    [anon_sym___typeof] = ACTIONS(2740),
    [anon_sym_typeof] = ACTIONS(2740),
    [anon_sym_ATimport] = ACTIONS(2742),
    [aux_sym_preproc_undef_token1] = ACTIONS(2740),
    [anon_sym_POUND] = ACTIONS(2740),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2740),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2740),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2740),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2740),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2740),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2740),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2740),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_API_AVAILABLE] = ACTIONS(2740),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_API_DEPRECATED] = ACTIONS(2740),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2740),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2740),
    [anon_sym___deprecated_msg] = ACTIONS(2740),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2740),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2740),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2740),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2740),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2740),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2740),
    [anon_sym_ATprotocol] = ACTIONS(2742),
    [anon_sym_ATinterface] = ACTIONS(2742),
    [anon_sym_ATimplementation] = ACTIONS(2742),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2742),
    [anon_sym_ATselector] = ACTIONS(2742),
    [anon_sym_ATavailable] = ACTIONS(2742),
    [anon_sym___builtin_available] = ACTIONS(2740),
    [anon_sym_va_arg] = ACTIONS(2740),
    [anon_sym_ATencode] = ACTIONS(2742),
    [anon_sym_BOOL] = ACTIONS(2740),
    [anon_sym_IMP] = ACTIONS(2740),
    [anon_sym_SEL] = ACTIONS(2740),
    [anon_sym_Class] = ACTIONS(2740),
    [anon_sym_id] = ACTIONS(2740),
  },
  [STATE(1046)] = {
    [ts_builtin_sym_end] = ACTIONS(2952),
    [sym_identifier] = ACTIONS(2954),
    [aux_sym_preproc_include_token1] = ACTIONS(2954),
    [aux_sym_preproc_include_token2] = ACTIONS(2954),
    [aux_sym_preproc_def_token1] = ACTIONS(2954),
    [aux_sym_preproc_if_token1] = ACTIONS(2954),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2954),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2954),
    [sym_preproc_directive] = ACTIONS(2954),
    [anon_sym_LPAREN2] = ACTIONS(2952),
    [anon_sym_BANG] = ACTIONS(2952),
    [anon_sym_TILDE] = ACTIONS(2952),
    [anon_sym_DASH] = ACTIONS(2954),
    [anon_sym_PLUS] = ACTIONS(2954),
    [anon_sym_STAR] = ACTIONS(2952),
    [anon_sym_CARET] = ACTIONS(2952),
    [anon_sym_AMP] = ACTIONS(2952),
    [anon_sym_SEMI] = ACTIONS(2952),
    [anon_sym___extension__] = ACTIONS(2954),
    [anon_sym_typedef] = ACTIONS(2954),
    [anon_sym_extern] = ACTIONS(2954),
    [anon_sym___attribute__] = ACTIONS(2954),
    [anon_sym___attribute] = ACTIONS(2954),
    [anon_sym_noreturn] = ACTIONS(2954),
    [anon_sym_LBRACK] = ACTIONS(2952),
    [anon_sym___declspec] = ACTIONS(2954),
    [anon_sym___cdecl] = ACTIONS(2954),
    [anon_sym___clrcall] = ACTIONS(2954),
    [anon_sym___stdcall] = ACTIONS(2954),
    [anon_sym___fastcall] = ACTIONS(2954),
    [anon_sym___thiscall] = ACTIONS(2954),
    [anon_sym___vectorcall] = ACTIONS(2954),
    [anon_sym_LBRACE] = ACTIONS(2952),
    [anon_sym_signed] = ACTIONS(2954),
    [anon_sym_unsigned] = ACTIONS(2954),
    [anon_sym_long] = ACTIONS(2954),
    [anon_sym_short] = ACTIONS(2954),
    [anon_sym_static] = ACTIONS(2954),
    [anon_sym_ATautoreleasepool] = ACTIONS(2952),
    [anon_sym_auto] = ACTIONS(2954),
    [anon_sym_register] = ACTIONS(2954),
    [anon_sym_inline] = ACTIONS(2954),
    [anon_sym___inline] = ACTIONS(2954),
    [anon_sym___inline__] = ACTIONS(2954),
    [anon_sym___forceinline] = ACTIONS(2954),
    [anon_sym_thread_local] = ACTIONS(2954),
    [anon_sym___thread] = ACTIONS(2954),
    [anon_sym_CG_EXTERN] = ACTIONS(2954),
    [anon_sym_CG_INLINE] = ACTIONS(2954),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2954),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2954),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2954),
    [anon_sym_IBOutlet] = ACTIONS(2954),
    [anon_sym_IBInspectable] = ACTIONS(2954),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2954),
    [anon_sym_NS_INLINE] = ACTIONS(2954),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2954),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2954),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2954),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2954),
    [anon_sym_const] = ACTIONS(2954),
    [anon_sym_constexpr] = ACTIONS(2954),
    [anon_sym_volatile] = ACTIONS(2954),
    [anon_sym_restrict] = ACTIONS(2954),
    [anon_sym___restrict__] = ACTIONS(2954),
    [anon_sym__Atomic] = ACTIONS(2954),
    [anon_sym__Noreturn] = ACTIONS(2954),
    [anon_sym__Nonnull] = ACTIONS(2954),
    [anon_sym_nullable] = ACTIONS(2954),
    [anon_sym__Complex] = ACTIONS(2954),
    [anon_sym__Nullable] = ACTIONS(2954),
    [anon_sym__Nullable_result] = ACTIONS(2954),
    [anon_sym__Null_unspecified] = ACTIONS(2954),
    [anon_sym___autoreleasing] = ACTIONS(2954),
    [anon_sym___block] = ACTIONS(2954),
    [anon_sym___bridge] = ACTIONS(2954),
    [anon_sym___bridge_retained] = ACTIONS(2954),
    [anon_sym___bridge_transfer] = ACTIONS(2954),
    [anon_sym___complex] = ACTIONS(2954),
    [anon_sym___const] = ACTIONS(2954),
    [anon_sym___imag] = ACTIONS(2954),
    [anon_sym___kindof] = ACTIONS(2954),
    [anon_sym___nonnull] = ACTIONS(2954),
    [anon_sym___nullable] = ACTIONS(2954),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2954),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2954),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2954),
    [anon_sym___real] = ACTIONS(2954),
    [anon_sym___strong] = ACTIONS(2954),
    [anon_sym___unsafe_unretained] = ACTIONS(2954),
    [anon_sym___unused] = ACTIONS(2954),
    [anon_sym___weak] = ACTIONS(2954),
    [anon_sym_alignas] = ACTIONS(2954),
    [anon_sym__Alignas] = ACTIONS(2954),
    [sym_primitive_type] = ACTIONS(2954),
    [anon_sym_enum] = ACTIONS(2954),
    [anon_sym_struct] = ACTIONS(2954),
    [anon_sym_union] = ACTIONS(2954),
    [anon_sym_if] = ACTIONS(2954),
    [anon_sym_switch] = ACTIONS(2954),
    [anon_sym_case] = ACTIONS(2954),
    [anon_sym_default] = ACTIONS(2954),
    [anon_sym_while] = ACTIONS(2954),
    [anon_sym_do] = ACTIONS(2954),
    [anon_sym_for] = ACTIONS(2954),
    [anon_sym_in] = ACTIONS(2954),
    [anon_sym_return] = ACTIONS(2954),
    [anon_sym_break] = ACTIONS(2954),
    [anon_sym_continue] = ACTIONS(2954),
    [anon_sym_goto] = ACTIONS(2954),
    [anon_sym_DASH_DASH] = ACTIONS(2952),
    [anon_sym_PLUS_PLUS] = ACTIONS(2952),
    [anon_sym_sizeof] = ACTIONS(2954),
    [anon_sym___alignof__] = ACTIONS(2954),
    [anon_sym___alignof] = ACTIONS(2954),
    [anon_sym__alignof] = ACTIONS(2954),
    [anon_sym_alignof] = ACTIONS(2954),
    [anon_sym__Alignof] = ACTIONS(2954),
    [anon_sym_offsetof] = ACTIONS(2954),
    [anon_sym__Generic] = ACTIONS(2954),
    [anon_sym_asm] = ACTIONS(2954),
    [anon_sym___asm__] = ACTIONS(2954),
    [anon_sym___asm] = ACTIONS(2954),
    [sym_number_literal] = ACTIONS(2952),
    [anon_sym_L_SQUOTE] = ACTIONS(2952),
    [anon_sym_u_SQUOTE] = ACTIONS(2952),
    [anon_sym_U_SQUOTE] = ACTIONS(2952),
    [anon_sym_u8_SQUOTE] = ACTIONS(2952),
    [anon_sym_SQUOTE] = ACTIONS(2952),
    [anon_sym_AT] = ACTIONS(2954),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_L_DQUOTE] = ACTIONS(2952),
    [anon_sym_u_DQUOTE] = ACTIONS(2952),
    [anon_sym_U_DQUOTE] = ACTIONS(2952),
    [anon_sym_u8_DQUOTE] = ACTIONS(2952),
    [sym_true] = ACTIONS(2954),
    [sym_false] = ACTIONS(2954),
    [anon_sym_NULL] = ACTIONS(2954),
    [anon_sym_nullptr] = ACTIONS(2954),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2954),
    [anon_sym___typeof] = ACTIONS(2954),
    [anon_sym_typeof] = ACTIONS(2954),
    [anon_sym_ATimport] = ACTIONS(2952),
    [aux_sym_preproc_undef_token1] = ACTIONS(2954),
    [anon_sym_POUND] = ACTIONS(2954),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2954),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2954),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2954),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2954),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2954),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2954),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2954),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2954),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2954),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2954),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2954),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2954),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2954),
    [anon_sym_API_AVAILABLE] = ACTIONS(2954),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2954),
    [anon_sym_API_DEPRECATED] = ACTIONS(2954),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2954),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2954),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2954),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2954),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2954),
    [anon_sym___deprecated_msg] = ACTIONS(2954),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2954),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2954),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2954),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2954),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2954),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2954),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2954),
    [anon_sym_ATprotocol] = ACTIONS(2952),
    [anon_sym_ATinterface] = ACTIONS(2952),
    [anon_sym_ATimplementation] = ACTIONS(2952),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2952),
    [anon_sym_ATselector] = ACTIONS(2952),
    [anon_sym_ATavailable] = ACTIONS(2952),
    [anon_sym___builtin_available] = ACTIONS(2954),
    [anon_sym_va_arg] = ACTIONS(2954),
    [anon_sym_ATencode] = ACTIONS(2952),
    [anon_sym_BOOL] = ACTIONS(2954),
    [anon_sym_IMP] = ACTIONS(2954),
    [anon_sym_SEL] = ACTIONS(2954),
    [anon_sym_Class] = ACTIONS(2954),
    [anon_sym_id] = ACTIONS(2954),
  },
  [STATE(1047)] = {
    [ts_builtin_sym_end] = ACTIONS(2854),
    [sym_identifier] = ACTIONS(2852),
    [aux_sym_preproc_include_token1] = ACTIONS(2852),
    [aux_sym_preproc_include_token2] = ACTIONS(2852),
    [aux_sym_preproc_def_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2852),
    [sym_preproc_directive] = ACTIONS(2852),
    [anon_sym_LPAREN2] = ACTIONS(2854),
    [anon_sym_BANG] = ACTIONS(2854),
    [anon_sym_TILDE] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2852),
    [anon_sym_PLUS] = ACTIONS(2852),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_SEMI] = ACTIONS(2854),
    [anon_sym___extension__] = ACTIONS(2852),
    [anon_sym_typedef] = ACTIONS(2852),
    [anon_sym_extern] = ACTIONS(2852),
    [anon_sym___attribute__] = ACTIONS(2852),
    [anon_sym___attribute] = ACTIONS(2852),
    [anon_sym_noreturn] = ACTIONS(2852),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym___declspec] = ACTIONS(2852),
    [anon_sym___cdecl] = ACTIONS(2852),
    [anon_sym___clrcall] = ACTIONS(2852),
    [anon_sym___stdcall] = ACTIONS(2852),
    [anon_sym___fastcall] = ACTIONS(2852),
    [anon_sym___thiscall] = ACTIONS(2852),
    [anon_sym___vectorcall] = ACTIONS(2852),
    [anon_sym_LBRACE] = ACTIONS(2854),
    [anon_sym_signed] = ACTIONS(2852),
    [anon_sym_unsigned] = ACTIONS(2852),
    [anon_sym_long] = ACTIONS(2852),
    [anon_sym_short] = ACTIONS(2852),
    [anon_sym_static] = ACTIONS(2852),
    [anon_sym_ATautoreleasepool] = ACTIONS(2854),
    [anon_sym_auto] = ACTIONS(2852),
    [anon_sym_register] = ACTIONS(2852),
    [anon_sym_inline] = ACTIONS(2852),
    [anon_sym___inline] = ACTIONS(2852),
    [anon_sym___inline__] = ACTIONS(2852),
    [anon_sym___forceinline] = ACTIONS(2852),
    [anon_sym_thread_local] = ACTIONS(2852),
    [anon_sym___thread] = ACTIONS(2852),
    [anon_sym_CG_EXTERN] = ACTIONS(2852),
    [anon_sym_CG_INLINE] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2852),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2852),
    [anon_sym_IBOutlet] = ACTIONS(2852),
    [anon_sym_IBInspectable] = ACTIONS(2852),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2852),
    [anon_sym_NS_INLINE] = ACTIONS(2852),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2852),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2852),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2852),
    [anon_sym_const] = ACTIONS(2852),
    [anon_sym_constexpr] = ACTIONS(2852),
    [anon_sym_volatile] = ACTIONS(2852),
    [anon_sym_restrict] = ACTIONS(2852),
    [anon_sym___restrict__] = ACTIONS(2852),
    [anon_sym__Atomic] = ACTIONS(2852),
    [anon_sym__Noreturn] = ACTIONS(2852),
    [anon_sym__Nonnull] = ACTIONS(2852),
    [anon_sym_nullable] = ACTIONS(2852),
    [anon_sym__Complex] = ACTIONS(2852),
    [anon_sym__Nullable] = ACTIONS(2852),
    [anon_sym__Nullable_result] = ACTIONS(2852),
    [anon_sym__Null_unspecified] = ACTIONS(2852),
    [anon_sym___autoreleasing] = ACTIONS(2852),
    [anon_sym___block] = ACTIONS(2852),
    [anon_sym___bridge] = ACTIONS(2852),
    [anon_sym___bridge_retained] = ACTIONS(2852),
    [anon_sym___bridge_transfer] = ACTIONS(2852),
    [anon_sym___complex] = ACTIONS(2852),
    [anon_sym___const] = ACTIONS(2852),
    [anon_sym___imag] = ACTIONS(2852),
    [anon_sym___kindof] = ACTIONS(2852),
    [anon_sym___nonnull] = ACTIONS(2852),
    [anon_sym___nullable] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2852),
    [anon_sym___real] = ACTIONS(2852),
    [anon_sym___strong] = ACTIONS(2852),
    [anon_sym___unsafe_unretained] = ACTIONS(2852),
    [anon_sym___unused] = ACTIONS(2852),
    [anon_sym___weak] = ACTIONS(2852),
    [anon_sym_alignas] = ACTIONS(2852),
    [anon_sym__Alignas] = ACTIONS(2852),
    [sym_primitive_type] = ACTIONS(2852),
    [anon_sym_enum] = ACTIONS(2852),
    [anon_sym_struct] = ACTIONS(2852),
    [anon_sym_union] = ACTIONS(2852),
    [anon_sym_if] = ACTIONS(2852),
    [anon_sym_switch] = ACTIONS(2852),
    [anon_sym_case] = ACTIONS(2852),
    [anon_sym_default] = ACTIONS(2852),
    [anon_sym_while] = ACTIONS(2852),
    [anon_sym_do] = ACTIONS(2852),
    [anon_sym_for] = ACTIONS(2852),
    [anon_sym_in] = ACTIONS(2852),
    [anon_sym_return] = ACTIONS(2852),
    [anon_sym_break] = ACTIONS(2852),
    [anon_sym_continue] = ACTIONS(2852),
    [anon_sym_goto] = ACTIONS(2852),
    [anon_sym_DASH_DASH] = ACTIONS(2854),
    [anon_sym_PLUS_PLUS] = ACTIONS(2854),
    [anon_sym_sizeof] = ACTIONS(2852),
    [anon_sym___alignof__] = ACTIONS(2852),
    [anon_sym___alignof] = ACTIONS(2852),
    [anon_sym__alignof] = ACTIONS(2852),
    [anon_sym_alignof] = ACTIONS(2852),
    [anon_sym__Alignof] = ACTIONS(2852),
    [anon_sym_offsetof] = ACTIONS(2852),
    [anon_sym__Generic] = ACTIONS(2852),
    [anon_sym_asm] = ACTIONS(2852),
    [anon_sym___asm__] = ACTIONS(2852),
    [anon_sym___asm] = ACTIONS(2852),
    [sym_number_literal] = ACTIONS(2854),
    [anon_sym_L_SQUOTE] = ACTIONS(2854),
    [anon_sym_u_SQUOTE] = ACTIONS(2854),
    [anon_sym_U_SQUOTE] = ACTIONS(2854),
    [anon_sym_u8_SQUOTE] = ACTIONS(2854),
    [anon_sym_SQUOTE] = ACTIONS(2854),
    [anon_sym_AT] = ACTIONS(2852),
    [anon_sym_DQUOTE] = ACTIONS(2854),
    [anon_sym_L_DQUOTE] = ACTIONS(2854),
    [anon_sym_u_DQUOTE] = ACTIONS(2854),
    [anon_sym_U_DQUOTE] = ACTIONS(2854),
    [anon_sym_u8_DQUOTE] = ACTIONS(2854),
    [sym_true] = ACTIONS(2852),
    [sym_false] = ACTIONS(2852),
    [anon_sym_NULL] = ACTIONS(2852),
    [anon_sym_nullptr] = ACTIONS(2852),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2852),
    [anon_sym___typeof] = ACTIONS(2852),
    [anon_sym_typeof] = ACTIONS(2852),
    [anon_sym_ATimport] = ACTIONS(2854),
    [aux_sym_preproc_undef_token1] = ACTIONS(2852),
    [anon_sym_POUND] = ACTIONS(2852),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2852),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2852),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2852),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_API_AVAILABLE] = ACTIONS(2852),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_API_DEPRECATED] = ACTIONS(2852),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym___deprecated_msg] = ACTIONS(2852),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2852),
    [anon_sym_ATprotocol] = ACTIONS(2854),
    [anon_sym_ATinterface] = ACTIONS(2854),
    [anon_sym_ATimplementation] = ACTIONS(2854),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2854),
    [anon_sym_ATselector] = ACTIONS(2854),
    [anon_sym_ATavailable] = ACTIONS(2854),
    [anon_sym___builtin_available] = ACTIONS(2852),
    [anon_sym_va_arg] = ACTIONS(2852),
    [anon_sym_ATencode] = ACTIONS(2854),
    [anon_sym_BOOL] = ACTIONS(2852),
    [anon_sym_IMP] = ACTIONS(2852),
    [anon_sym_SEL] = ACTIONS(2852),
    [anon_sym_Class] = ACTIONS(2852),
    [anon_sym_id] = ACTIONS(2852),
  },
  [STATE(1048)] = {
    [ts_builtin_sym_end] = ACTIONS(2598),
    [sym_identifier] = ACTIONS(2596),
    [aux_sym_preproc_include_token1] = ACTIONS(2596),
    [aux_sym_preproc_include_token2] = ACTIONS(2596),
    [aux_sym_preproc_def_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2596),
    [sym_preproc_directive] = ACTIONS(2596),
    [anon_sym_LPAREN2] = ACTIONS(2598),
    [anon_sym_BANG] = ACTIONS(2598),
    [anon_sym_TILDE] = ACTIONS(2598),
    [anon_sym_DASH] = ACTIONS(2596),
    [anon_sym_PLUS] = ACTIONS(2596),
    [anon_sym_STAR] = ACTIONS(2598),
    [anon_sym_CARET] = ACTIONS(2598),
    [anon_sym_AMP] = ACTIONS(2598),
    [anon_sym_SEMI] = ACTIONS(2598),
    [anon_sym___extension__] = ACTIONS(2596),
    [anon_sym_typedef] = ACTIONS(2596),
    [anon_sym_extern] = ACTIONS(2596),
    [anon_sym___attribute__] = ACTIONS(2596),
    [anon_sym___attribute] = ACTIONS(2596),
    [anon_sym_noreturn] = ACTIONS(2596),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym___declspec] = ACTIONS(2596),
    [anon_sym___cdecl] = ACTIONS(2596),
    [anon_sym___clrcall] = ACTIONS(2596),
    [anon_sym___stdcall] = ACTIONS(2596),
    [anon_sym___fastcall] = ACTIONS(2596),
    [anon_sym___thiscall] = ACTIONS(2596),
    [anon_sym___vectorcall] = ACTIONS(2596),
    [anon_sym_LBRACE] = ACTIONS(2598),
    [anon_sym_signed] = ACTIONS(2596),
    [anon_sym_unsigned] = ACTIONS(2596),
    [anon_sym_long] = ACTIONS(2596),
    [anon_sym_short] = ACTIONS(2596),
    [anon_sym_static] = ACTIONS(2596),
    [anon_sym_ATautoreleasepool] = ACTIONS(2598),
    [anon_sym_auto] = ACTIONS(2596),
    [anon_sym_register] = ACTIONS(2596),
    [anon_sym_inline] = ACTIONS(2596),
    [anon_sym___inline] = ACTIONS(2596),
    [anon_sym___inline__] = ACTIONS(2596),
    [anon_sym___forceinline] = ACTIONS(2596),
    [anon_sym_thread_local] = ACTIONS(2596),
    [anon_sym___thread] = ACTIONS(2596),
    [anon_sym_CG_EXTERN] = ACTIONS(2596),
    [anon_sym_CG_INLINE] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2596),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2596),
    [anon_sym_IBOutlet] = ACTIONS(2596),
    [anon_sym_IBInspectable] = ACTIONS(2596),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2596),
    [anon_sym_NS_INLINE] = ACTIONS(2596),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2596),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2596),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2596),
    [anon_sym_const] = ACTIONS(2596),
    [anon_sym_constexpr] = ACTIONS(2596),
    [anon_sym_volatile] = ACTIONS(2596),
    [anon_sym_restrict] = ACTIONS(2596),
    [anon_sym___restrict__] = ACTIONS(2596),
    [anon_sym__Atomic] = ACTIONS(2596),
    [anon_sym__Noreturn] = ACTIONS(2596),
    [anon_sym__Nonnull] = ACTIONS(2596),
    [anon_sym_nullable] = ACTIONS(2596),
    [anon_sym__Complex] = ACTIONS(2596),
    [anon_sym__Nullable] = ACTIONS(2596),
    [anon_sym__Nullable_result] = ACTIONS(2596),
    [anon_sym__Null_unspecified] = ACTIONS(2596),
    [anon_sym___autoreleasing] = ACTIONS(2596),
    [anon_sym___block] = ACTIONS(2596),
    [anon_sym___bridge] = ACTIONS(2596),
    [anon_sym___bridge_retained] = ACTIONS(2596),
    [anon_sym___bridge_transfer] = ACTIONS(2596),
    [anon_sym___complex] = ACTIONS(2596),
    [anon_sym___const] = ACTIONS(2596),
    [anon_sym___imag] = ACTIONS(2596),
    [anon_sym___kindof] = ACTIONS(2596),
    [anon_sym___nonnull] = ACTIONS(2596),
    [anon_sym___nullable] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2596),
    [anon_sym___real] = ACTIONS(2596),
    [anon_sym___strong] = ACTIONS(2596),
    [anon_sym___unsafe_unretained] = ACTIONS(2596),
    [anon_sym___unused] = ACTIONS(2596),
    [anon_sym___weak] = ACTIONS(2596),
    [anon_sym_alignas] = ACTIONS(2596),
    [anon_sym__Alignas] = ACTIONS(2596),
    [sym_primitive_type] = ACTIONS(2596),
    [anon_sym_enum] = ACTIONS(2596),
    [anon_sym_struct] = ACTIONS(2596),
    [anon_sym_union] = ACTIONS(2596),
    [anon_sym_if] = ACTIONS(2596),
    [anon_sym_switch] = ACTIONS(2596),
    [anon_sym_case] = ACTIONS(2596),
    [anon_sym_default] = ACTIONS(2596),
    [anon_sym_while] = ACTIONS(2596),
    [anon_sym_do] = ACTIONS(2596),
    [anon_sym_for] = ACTIONS(2596),
    [anon_sym_in] = ACTIONS(2596),
    [anon_sym_return] = ACTIONS(2596),
    [anon_sym_break] = ACTIONS(2596),
    [anon_sym_continue] = ACTIONS(2596),
    [anon_sym_goto] = ACTIONS(2596),
    [anon_sym_DASH_DASH] = ACTIONS(2598),
    [anon_sym_PLUS_PLUS] = ACTIONS(2598),
    [anon_sym_sizeof] = ACTIONS(2596),
    [anon_sym___alignof__] = ACTIONS(2596),
    [anon_sym___alignof] = ACTIONS(2596),
    [anon_sym__alignof] = ACTIONS(2596),
    [anon_sym_alignof] = ACTIONS(2596),
    [anon_sym__Alignof] = ACTIONS(2596),
    [anon_sym_offsetof] = ACTIONS(2596),
    [anon_sym__Generic] = ACTIONS(2596),
    [anon_sym_asm] = ACTIONS(2596),
    [anon_sym___asm__] = ACTIONS(2596),
    [anon_sym___asm] = ACTIONS(2596),
    [sym_number_literal] = ACTIONS(2598),
    [anon_sym_L_SQUOTE] = ACTIONS(2598),
    [anon_sym_u_SQUOTE] = ACTIONS(2598),
    [anon_sym_U_SQUOTE] = ACTIONS(2598),
    [anon_sym_u8_SQUOTE] = ACTIONS(2598),
    [anon_sym_SQUOTE] = ACTIONS(2598),
    [anon_sym_AT] = ACTIONS(2596),
    [anon_sym_DQUOTE] = ACTIONS(2598),
    [anon_sym_L_DQUOTE] = ACTIONS(2598),
    [anon_sym_u_DQUOTE] = ACTIONS(2598),
    [anon_sym_U_DQUOTE] = ACTIONS(2598),
    [anon_sym_u8_DQUOTE] = ACTIONS(2598),
    [sym_true] = ACTIONS(2596),
    [sym_false] = ACTIONS(2596),
    [anon_sym_NULL] = ACTIONS(2596),
    [anon_sym_nullptr] = ACTIONS(2596),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2596),
    [anon_sym___typeof] = ACTIONS(2596),
    [anon_sym_typeof] = ACTIONS(2596),
    [anon_sym_ATimport] = ACTIONS(2598),
    [aux_sym_preproc_undef_token1] = ACTIONS(2596),
    [anon_sym_POUND] = ACTIONS(2596),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2596),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2596),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2596),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_API_AVAILABLE] = ACTIONS(2596),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_API_DEPRECATED] = ACTIONS(2596),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym___deprecated_msg] = ACTIONS(2596),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2596),
    [anon_sym_ATprotocol] = ACTIONS(2598),
    [anon_sym_ATinterface] = ACTIONS(2598),
    [anon_sym_ATimplementation] = ACTIONS(2598),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2598),
    [anon_sym_ATselector] = ACTIONS(2598),
    [anon_sym_ATavailable] = ACTIONS(2598),
    [anon_sym___builtin_available] = ACTIONS(2596),
    [anon_sym_va_arg] = ACTIONS(2596),
    [anon_sym_ATencode] = ACTIONS(2598),
    [anon_sym_BOOL] = ACTIONS(2596),
    [anon_sym_IMP] = ACTIONS(2596),
    [anon_sym_SEL] = ACTIONS(2596),
    [anon_sym_Class] = ACTIONS(2596),
    [anon_sym_id] = ACTIONS(2596),
  },
  [STATE(1049)] = {
    [ts_builtin_sym_end] = ACTIONS(2746),
    [sym_identifier] = ACTIONS(2744),
    [aux_sym_preproc_include_token1] = ACTIONS(2744),
    [aux_sym_preproc_include_token2] = ACTIONS(2744),
    [aux_sym_preproc_def_token1] = ACTIONS(2744),
    [aux_sym_preproc_if_token1] = ACTIONS(2744),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2744),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2744),
    [sym_preproc_directive] = ACTIONS(2744),
    [anon_sym_LPAREN2] = ACTIONS(2746),
    [anon_sym_BANG] = ACTIONS(2746),
    [anon_sym_TILDE] = ACTIONS(2746),
    [anon_sym_DASH] = ACTIONS(2744),
    [anon_sym_PLUS] = ACTIONS(2744),
    [anon_sym_STAR] = ACTIONS(2746),
    [anon_sym_CARET] = ACTIONS(2746),
    [anon_sym_AMP] = ACTIONS(2746),
    [anon_sym_SEMI] = ACTIONS(2746),
    [anon_sym___extension__] = ACTIONS(2744),
    [anon_sym_typedef] = ACTIONS(2744),
    [anon_sym_extern] = ACTIONS(2744),
    [anon_sym___attribute__] = ACTIONS(2744),
    [anon_sym___attribute] = ACTIONS(2744),
    [anon_sym_noreturn] = ACTIONS(2744),
    [anon_sym_LBRACK] = ACTIONS(2746),
    [anon_sym___declspec] = ACTIONS(2744),
    [anon_sym___cdecl] = ACTIONS(2744),
    [anon_sym___clrcall] = ACTIONS(2744),
    [anon_sym___stdcall] = ACTIONS(2744),
    [anon_sym___fastcall] = ACTIONS(2744),
    [anon_sym___thiscall] = ACTIONS(2744),
    [anon_sym___vectorcall] = ACTIONS(2744),
    [anon_sym_LBRACE] = ACTIONS(2746),
    [anon_sym_signed] = ACTIONS(2744),
    [anon_sym_unsigned] = ACTIONS(2744),
    [anon_sym_long] = ACTIONS(2744),
    [anon_sym_short] = ACTIONS(2744),
    [anon_sym_static] = ACTIONS(2744),
    [anon_sym_ATautoreleasepool] = ACTIONS(2746),
    [anon_sym_auto] = ACTIONS(2744),
    [anon_sym_register] = ACTIONS(2744),
    [anon_sym_inline] = ACTIONS(2744),
    [anon_sym___inline] = ACTIONS(2744),
    [anon_sym___inline__] = ACTIONS(2744),
    [anon_sym___forceinline] = ACTIONS(2744),
    [anon_sym_thread_local] = ACTIONS(2744),
    [anon_sym___thread] = ACTIONS(2744),
    [anon_sym_CG_EXTERN] = ACTIONS(2744),
    [anon_sym_CG_INLINE] = ACTIONS(2744),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2744),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2744),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2744),
    [anon_sym_IBOutlet] = ACTIONS(2744),
    [anon_sym_IBInspectable] = ACTIONS(2744),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2744),
    [anon_sym_NS_INLINE] = ACTIONS(2744),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2744),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2744),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2744),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2744),
    [anon_sym_const] = ACTIONS(2744),
    [anon_sym_constexpr] = ACTIONS(2744),
    [anon_sym_volatile] = ACTIONS(2744),
    [anon_sym_restrict] = ACTIONS(2744),
    [anon_sym___restrict__] = ACTIONS(2744),
    [anon_sym__Atomic] = ACTIONS(2744),
    [anon_sym__Noreturn] = ACTIONS(2744),
    [anon_sym__Nonnull] = ACTIONS(2744),
    [anon_sym_nullable] = ACTIONS(2744),
    [anon_sym__Complex] = ACTIONS(2744),
    [anon_sym__Nullable] = ACTIONS(2744),
    [anon_sym__Nullable_result] = ACTIONS(2744),
    [anon_sym__Null_unspecified] = ACTIONS(2744),
    [anon_sym___autoreleasing] = ACTIONS(2744),
    [anon_sym___block] = ACTIONS(2744),
    [anon_sym___bridge] = ACTIONS(2744),
    [anon_sym___bridge_retained] = ACTIONS(2744),
    [anon_sym___bridge_transfer] = ACTIONS(2744),
    [anon_sym___complex] = ACTIONS(2744),
    [anon_sym___const] = ACTIONS(2744),
    [anon_sym___imag] = ACTIONS(2744),
    [anon_sym___kindof] = ACTIONS(2744),
    [anon_sym___nonnull] = ACTIONS(2744),
    [anon_sym___nullable] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2744),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2744),
    [anon_sym___real] = ACTIONS(2744),
    [anon_sym___strong] = ACTIONS(2744),
    [anon_sym___unsafe_unretained] = ACTIONS(2744),
    [anon_sym___unused] = ACTIONS(2744),
    [anon_sym___weak] = ACTIONS(2744),
    [anon_sym_alignas] = ACTIONS(2744),
    [anon_sym__Alignas] = ACTIONS(2744),
    [sym_primitive_type] = ACTIONS(2744),
    [anon_sym_enum] = ACTIONS(2744),
    [anon_sym_struct] = ACTIONS(2744),
    [anon_sym_union] = ACTIONS(2744),
    [anon_sym_if] = ACTIONS(2744),
    [anon_sym_switch] = ACTIONS(2744),
    [anon_sym_case] = ACTIONS(2744),
    [anon_sym_default] = ACTIONS(2744),
    [anon_sym_while] = ACTIONS(2744),
    [anon_sym_do] = ACTIONS(2744),
    [anon_sym_for] = ACTIONS(2744),
    [anon_sym_in] = ACTIONS(2744),
    [anon_sym_return] = ACTIONS(2744),
    [anon_sym_break] = ACTIONS(2744),
    [anon_sym_continue] = ACTIONS(2744),
    [anon_sym_goto] = ACTIONS(2744),
    [anon_sym_DASH_DASH] = ACTIONS(2746),
    [anon_sym_PLUS_PLUS] = ACTIONS(2746),
    [anon_sym_sizeof] = ACTIONS(2744),
    [anon_sym___alignof__] = ACTIONS(2744),
    [anon_sym___alignof] = ACTIONS(2744),
    [anon_sym__alignof] = ACTIONS(2744),
    [anon_sym_alignof] = ACTIONS(2744),
    [anon_sym__Alignof] = ACTIONS(2744),
    [anon_sym_offsetof] = ACTIONS(2744),
    [anon_sym__Generic] = ACTIONS(2744),
    [anon_sym_asm] = ACTIONS(2744),
    [anon_sym___asm__] = ACTIONS(2744),
    [anon_sym___asm] = ACTIONS(2744),
    [sym_number_literal] = ACTIONS(2746),
    [anon_sym_L_SQUOTE] = ACTIONS(2746),
    [anon_sym_u_SQUOTE] = ACTIONS(2746),
    [anon_sym_U_SQUOTE] = ACTIONS(2746),
    [anon_sym_u8_SQUOTE] = ACTIONS(2746),
    [anon_sym_SQUOTE] = ACTIONS(2746),
    [anon_sym_AT] = ACTIONS(2744),
    [anon_sym_DQUOTE] = ACTIONS(2746),
    [anon_sym_L_DQUOTE] = ACTIONS(2746),
    [anon_sym_u_DQUOTE] = ACTIONS(2746),
    [anon_sym_U_DQUOTE] = ACTIONS(2746),
    [anon_sym_u8_DQUOTE] = ACTIONS(2746),
    [sym_true] = ACTIONS(2744),
    [sym_false] = ACTIONS(2744),
    [anon_sym_NULL] = ACTIONS(2744),
    [anon_sym_nullptr] = ACTIONS(2744),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2744),
    [anon_sym___typeof] = ACTIONS(2744),
    [anon_sym_typeof] = ACTIONS(2744),
    [anon_sym_ATimport] = ACTIONS(2746),
    [aux_sym_preproc_undef_token1] = ACTIONS(2744),
    [anon_sym_POUND] = ACTIONS(2744),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2744),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2744),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2744),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2744),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2744),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2744),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2744),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_API_AVAILABLE] = ACTIONS(2744),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_API_DEPRECATED] = ACTIONS(2744),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2744),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2744),
    [anon_sym___deprecated_msg] = ACTIONS(2744),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2744),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2744),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2744),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2744),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2744),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2744),
    [anon_sym_ATprotocol] = ACTIONS(2746),
    [anon_sym_ATinterface] = ACTIONS(2746),
    [anon_sym_ATimplementation] = ACTIONS(2746),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2746),
    [anon_sym_ATselector] = ACTIONS(2746),
    [anon_sym_ATavailable] = ACTIONS(2746),
    [anon_sym___builtin_available] = ACTIONS(2744),
    [anon_sym_va_arg] = ACTIONS(2744),
    [anon_sym_ATencode] = ACTIONS(2746),
    [anon_sym_BOOL] = ACTIONS(2744),
    [anon_sym_IMP] = ACTIONS(2744),
    [anon_sym_SEL] = ACTIONS(2744),
    [anon_sym_Class] = ACTIONS(2744),
    [anon_sym_id] = ACTIONS(2744),
  },
  [STATE(1050)] = {
    [ts_builtin_sym_end] = ACTIONS(2750),
    [sym_identifier] = ACTIONS(2748),
    [aux_sym_preproc_include_token1] = ACTIONS(2748),
    [aux_sym_preproc_include_token2] = ACTIONS(2748),
    [aux_sym_preproc_def_token1] = ACTIONS(2748),
    [aux_sym_preproc_if_token1] = ACTIONS(2748),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2748),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2748),
    [sym_preproc_directive] = ACTIONS(2748),
    [anon_sym_LPAREN2] = ACTIONS(2750),
    [anon_sym_BANG] = ACTIONS(2750),
    [anon_sym_TILDE] = ACTIONS(2750),
    [anon_sym_DASH] = ACTIONS(2748),
    [anon_sym_PLUS] = ACTIONS(2748),
    [anon_sym_STAR] = ACTIONS(2750),
    [anon_sym_CARET] = ACTIONS(2750),
    [anon_sym_AMP] = ACTIONS(2750),
    [anon_sym_SEMI] = ACTIONS(2750),
    [anon_sym___extension__] = ACTIONS(2748),
    [anon_sym_typedef] = ACTIONS(2748),
    [anon_sym_extern] = ACTIONS(2748),
    [anon_sym___attribute__] = ACTIONS(2748),
    [anon_sym___attribute] = ACTIONS(2748),
    [anon_sym_noreturn] = ACTIONS(2748),
    [anon_sym_LBRACK] = ACTIONS(2750),
    [anon_sym___declspec] = ACTIONS(2748),
    [anon_sym___cdecl] = ACTIONS(2748),
    [anon_sym___clrcall] = ACTIONS(2748),
    [anon_sym___stdcall] = ACTIONS(2748),
    [anon_sym___fastcall] = ACTIONS(2748),
    [anon_sym___thiscall] = ACTIONS(2748),
    [anon_sym___vectorcall] = ACTIONS(2748),
    [anon_sym_LBRACE] = ACTIONS(2750),
    [anon_sym_signed] = ACTIONS(2748),
    [anon_sym_unsigned] = ACTIONS(2748),
    [anon_sym_long] = ACTIONS(2748),
    [anon_sym_short] = ACTIONS(2748),
    [anon_sym_static] = ACTIONS(2748),
    [anon_sym_ATautoreleasepool] = ACTIONS(2750),
    [anon_sym_auto] = ACTIONS(2748),
    [anon_sym_register] = ACTIONS(2748),
    [anon_sym_inline] = ACTIONS(2748),
    [anon_sym___inline] = ACTIONS(2748),
    [anon_sym___inline__] = ACTIONS(2748),
    [anon_sym___forceinline] = ACTIONS(2748),
    [anon_sym_thread_local] = ACTIONS(2748),
    [anon_sym___thread] = ACTIONS(2748),
    [anon_sym_CG_EXTERN] = ACTIONS(2748),
    [anon_sym_CG_INLINE] = ACTIONS(2748),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2748),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2748),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2748),
    [anon_sym_IBOutlet] = ACTIONS(2748),
    [anon_sym_IBInspectable] = ACTIONS(2748),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2748),
    [anon_sym_NS_INLINE] = ACTIONS(2748),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2748),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2748),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2748),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2748),
    [anon_sym_const] = ACTIONS(2748),
    [anon_sym_constexpr] = ACTIONS(2748),
    [anon_sym_volatile] = ACTIONS(2748),
    [anon_sym_restrict] = ACTIONS(2748),
    [anon_sym___restrict__] = ACTIONS(2748),
    [anon_sym__Atomic] = ACTIONS(2748),
    [anon_sym__Noreturn] = ACTIONS(2748),
    [anon_sym__Nonnull] = ACTIONS(2748),
    [anon_sym_nullable] = ACTIONS(2748),
    [anon_sym__Complex] = ACTIONS(2748),
    [anon_sym__Nullable] = ACTIONS(2748),
    [anon_sym__Nullable_result] = ACTIONS(2748),
    [anon_sym__Null_unspecified] = ACTIONS(2748),
    [anon_sym___autoreleasing] = ACTIONS(2748),
    [anon_sym___block] = ACTIONS(2748),
    [anon_sym___bridge] = ACTIONS(2748),
    [anon_sym___bridge_retained] = ACTIONS(2748),
    [anon_sym___bridge_transfer] = ACTIONS(2748),
    [anon_sym___complex] = ACTIONS(2748),
    [anon_sym___const] = ACTIONS(2748),
    [anon_sym___imag] = ACTIONS(2748),
    [anon_sym___kindof] = ACTIONS(2748),
    [anon_sym___nonnull] = ACTIONS(2748),
    [anon_sym___nullable] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2748),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2748),
    [anon_sym___real] = ACTIONS(2748),
    [anon_sym___strong] = ACTIONS(2748),
    [anon_sym___unsafe_unretained] = ACTIONS(2748),
    [anon_sym___unused] = ACTIONS(2748),
    [anon_sym___weak] = ACTIONS(2748),
    [anon_sym_alignas] = ACTIONS(2748),
    [anon_sym__Alignas] = ACTIONS(2748),
    [sym_primitive_type] = ACTIONS(2748),
    [anon_sym_enum] = ACTIONS(2748),
    [anon_sym_struct] = ACTIONS(2748),
    [anon_sym_union] = ACTIONS(2748),
    [anon_sym_if] = ACTIONS(2748),
    [anon_sym_switch] = ACTIONS(2748),
    [anon_sym_case] = ACTIONS(2748),
    [anon_sym_default] = ACTIONS(2748),
    [anon_sym_while] = ACTIONS(2748),
    [anon_sym_do] = ACTIONS(2748),
    [anon_sym_for] = ACTIONS(2748),
    [anon_sym_in] = ACTIONS(2748),
    [anon_sym_return] = ACTIONS(2748),
    [anon_sym_break] = ACTIONS(2748),
    [anon_sym_continue] = ACTIONS(2748),
    [anon_sym_goto] = ACTIONS(2748),
    [anon_sym_DASH_DASH] = ACTIONS(2750),
    [anon_sym_PLUS_PLUS] = ACTIONS(2750),
    [anon_sym_sizeof] = ACTIONS(2748),
    [anon_sym___alignof__] = ACTIONS(2748),
    [anon_sym___alignof] = ACTIONS(2748),
    [anon_sym__alignof] = ACTIONS(2748),
    [anon_sym_alignof] = ACTIONS(2748),
    [anon_sym__Alignof] = ACTIONS(2748),
    [anon_sym_offsetof] = ACTIONS(2748),
    [anon_sym__Generic] = ACTIONS(2748),
    [anon_sym_asm] = ACTIONS(2748),
    [anon_sym___asm__] = ACTIONS(2748),
    [anon_sym___asm] = ACTIONS(2748),
    [sym_number_literal] = ACTIONS(2750),
    [anon_sym_L_SQUOTE] = ACTIONS(2750),
    [anon_sym_u_SQUOTE] = ACTIONS(2750),
    [anon_sym_U_SQUOTE] = ACTIONS(2750),
    [anon_sym_u8_SQUOTE] = ACTIONS(2750),
    [anon_sym_SQUOTE] = ACTIONS(2750),
    [anon_sym_AT] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_L_DQUOTE] = ACTIONS(2750),
    [anon_sym_u_DQUOTE] = ACTIONS(2750),
    [anon_sym_U_DQUOTE] = ACTIONS(2750),
    [anon_sym_u8_DQUOTE] = ACTIONS(2750),
    [sym_true] = ACTIONS(2748),
    [sym_false] = ACTIONS(2748),
    [anon_sym_NULL] = ACTIONS(2748),
    [anon_sym_nullptr] = ACTIONS(2748),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2748),
    [anon_sym___typeof] = ACTIONS(2748),
    [anon_sym_typeof] = ACTIONS(2748),
    [anon_sym_ATimport] = ACTIONS(2750),
    [aux_sym_preproc_undef_token1] = ACTIONS(2748),
    [anon_sym_POUND] = ACTIONS(2748),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2748),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2748),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2748),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2748),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2748),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2748),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2748),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_API_AVAILABLE] = ACTIONS(2748),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_API_DEPRECATED] = ACTIONS(2748),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2748),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2748),
    [anon_sym___deprecated_msg] = ACTIONS(2748),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2748),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2748),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2748),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2748),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2748),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2748),
    [anon_sym_ATprotocol] = ACTIONS(2750),
    [anon_sym_ATinterface] = ACTIONS(2750),
    [anon_sym_ATimplementation] = ACTIONS(2750),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2750),
    [anon_sym_ATselector] = ACTIONS(2750),
    [anon_sym_ATavailable] = ACTIONS(2750),
    [anon_sym___builtin_available] = ACTIONS(2748),
    [anon_sym_va_arg] = ACTIONS(2748),
    [anon_sym_ATencode] = ACTIONS(2750),
    [anon_sym_BOOL] = ACTIONS(2748),
    [anon_sym_IMP] = ACTIONS(2748),
    [anon_sym_SEL] = ACTIONS(2748),
    [anon_sym_Class] = ACTIONS(2748),
    [anon_sym_id] = ACTIONS(2748),
  },
  [STATE(1051)] = {
    [ts_builtin_sym_end] = ACTIONS(2754),
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(1052)] = {
    [ts_builtin_sym_end] = ACTIONS(2754),
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(1053)] = {
    [ts_builtin_sym_end] = ACTIONS(2754),
    [sym_identifier] = ACTIONS(2752),
    [aux_sym_preproc_include_token1] = ACTIONS(2752),
    [aux_sym_preproc_include_token2] = ACTIONS(2752),
    [aux_sym_preproc_def_token1] = ACTIONS(2752),
    [aux_sym_preproc_if_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2752),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2752),
    [sym_preproc_directive] = ACTIONS(2752),
    [anon_sym_LPAREN2] = ACTIONS(2754),
    [anon_sym_BANG] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(2754),
    [anon_sym_DASH] = ACTIONS(2752),
    [anon_sym_PLUS] = ACTIONS(2752),
    [anon_sym_STAR] = ACTIONS(2754),
    [anon_sym_CARET] = ACTIONS(2754),
    [anon_sym_AMP] = ACTIONS(2754),
    [anon_sym_SEMI] = ACTIONS(2754),
    [anon_sym___extension__] = ACTIONS(2752),
    [anon_sym_typedef] = ACTIONS(2752),
    [anon_sym_extern] = ACTIONS(2752),
    [anon_sym___attribute__] = ACTIONS(2752),
    [anon_sym___attribute] = ACTIONS(2752),
    [anon_sym_noreturn] = ACTIONS(2752),
    [anon_sym_LBRACK] = ACTIONS(2754),
    [anon_sym___declspec] = ACTIONS(2752),
    [anon_sym___cdecl] = ACTIONS(2752),
    [anon_sym___clrcall] = ACTIONS(2752),
    [anon_sym___stdcall] = ACTIONS(2752),
    [anon_sym___fastcall] = ACTIONS(2752),
    [anon_sym___thiscall] = ACTIONS(2752),
    [anon_sym___vectorcall] = ACTIONS(2752),
    [anon_sym_LBRACE] = ACTIONS(2754),
    [anon_sym_signed] = ACTIONS(2752),
    [anon_sym_unsigned] = ACTIONS(2752),
    [anon_sym_long] = ACTIONS(2752),
    [anon_sym_short] = ACTIONS(2752),
    [anon_sym_static] = ACTIONS(2752),
    [anon_sym_ATautoreleasepool] = ACTIONS(2754),
    [anon_sym_auto] = ACTIONS(2752),
    [anon_sym_register] = ACTIONS(2752),
    [anon_sym_inline] = ACTIONS(2752),
    [anon_sym___inline] = ACTIONS(2752),
    [anon_sym___inline__] = ACTIONS(2752),
    [anon_sym___forceinline] = ACTIONS(2752),
    [anon_sym_thread_local] = ACTIONS(2752),
    [anon_sym___thread] = ACTIONS(2752),
    [anon_sym_CG_EXTERN] = ACTIONS(2752),
    [anon_sym_CG_INLINE] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2752),
    [anon_sym_IBOutlet] = ACTIONS(2752),
    [anon_sym_IBInspectable] = ACTIONS(2752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2752),
    [anon_sym_NS_INLINE] = ACTIONS(2752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2752),
    [anon_sym_const] = ACTIONS(2752),
    [anon_sym_constexpr] = ACTIONS(2752),
    [anon_sym_volatile] = ACTIONS(2752),
    [anon_sym_restrict] = ACTIONS(2752),
    [anon_sym___restrict__] = ACTIONS(2752),
    [anon_sym__Atomic] = ACTIONS(2752),
    [anon_sym__Noreturn] = ACTIONS(2752),
    [anon_sym__Nonnull] = ACTIONS(2752),
    [anon_sym_nullable] = ACTIONS(2752),
    [anon_sym__Complex] = ACTIONS(2752),
    [anon_sym__Nullable] = ACTIONS(2752),
    [anon_sym__Nullable_result] = ACTIONS(2752),
    [anon_sym__Null_unspecified] = ACTIONS(2752),
    [anon_sym___autoreleasing] = ACTIONS(2752),
    [anon_sym___block] = ACTIONS(2752),
    [anon_sym___bridge] = ACTIONS(2752),
    [anon_sym___bridge_retained] = ACTIONS(2752),
    [anon_sym___bridge_transfer] = ACTIONS(2752),
    [anon_sym___complex] = ACTIONS(2752),
    [anon_sym___const] = ACTIONS(2752),
    [anon_sym___imag] = ACTIONS(2752),
    [anon_sym___kindof] = ACTIONS(2752),
    [anon_sym___nonnull] = ACTIONS(2752),
    [anon_sym___nullable] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2752),
    [anon_sym___real] = ACTIONS(2752),
    [anon_sym___strong] = ACTIONS(2752),
    [anon_sym___unsafe_unretained] = ACTIONS(2752),
    [anon_sym___unused] = ACTIONS(2752),
    [anon_sym___weak] = ACTIONS(2752),
    [anon_sym_alignas] = ACTIONS(2752),
    [anon_sym__Alignas] = ACTIONS(2752),
    [sym_primitive_type] = ACTIONS(2752),
    [anon_sym_enum] = ACTIONS(2752),
    [anon_sym_struct] = ACTIONS(2752),
    [anon_sym_union] = ACTIONS(2752),
    [anon_sym_if] = ACTIONS(2752),
    [anon_sym_switch] = ACTIONS(2752),
    [anon_sym_case] = ACTIONS(2752),
    [anon_sym_default] = ACTIONS(2752),
    [anon_sym_while] = ACTIONS(2752),
    [anon_sym_do] = ACTIONS(2752),
    [anon_sym_for] = ACTIONS(2752),
    [anon_sym_in] = ACTIONS(2752),
    [anon_sym_return] = ACTIONS(2752),
    [anon_sym_break] = ACTIONS(2752),
    [anon_sym_continue] = ACTIONS(2752),
    [anon_sym_goto] = ACTIONS(2752),
    [anon_sym_DASH_DASH] = ACTIONS(2754),
    [anon_sym_PLUS_PLUS] = ACTIONS(2754),
    [anon_sym_sizeof] = ACTIONS(2752),
    [anon_sym___alignof__] = ACTIONS(2752),
    [anon_sym___alignof] = ACTIONS(2752),
    [anon_sym__alignof] = ACTIONS(2752),
    [anon_sym_alignof] = ACTIONS(2752),
    [anon_sym__Alignof] = ACTIONS(2752),
    [anon_sym_offsetof] = ACTIONS(2752),
    [anon_sym__Generic] = ACTIONS(2752),
    [anon_sym_asm] = ACTIONS(2752),
    [anon_sym___asm__] = ACTIONS(2752),
    [anon_sym___asm] = ACTIONS(2752),
    [sym_number_literal] = ACTIONS(2754),
    [anon_sym_L_SQUOTE] = ACTIONS(2754),
    [anon_sym_u_SQUOTE] = ACTIONS(2754),
    [anon_sym_U_SQUOTE] = ACTIONS(2754),
    [anon_sym_u8_SQUOTE] = ACTIONS(2754),
    [anon_sym_SQUOTE] = ACTIONS(2754),
    [anon_sym_AT] = ACTIONS(2752),
    [anon_sym_DQUOTE] = ACTIONS(2754),
    [anon_sym_L_DQUOTE] = ACTIONS(2754),
    [anon_sym_u_DQUOTE] = ACTIONS(2754),
    [anon_sym_U_DQUOTE] = ACTIONS(2754),
    [anon_sym_u8_DQUOTE] = ACTIONS(2754),
    [sym_true] = ACTIONS(2752),
    [sym_false] = ACTIONS(2752),
    [anon_sym_NULL] = ACTIONS(2752),
    [anon_sym_nullptr] = ACTIONS(2752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2752),
    [anon_sym___typeof] = ACTIONS(2752),
    [anon_sym_typeof] = ACTIONS(2752),
    [anon_sym_ATimport] = ACTIONS(2754),
    [aux_sym_preproc_undef_token1] = ACTIONS(2752),
    [anon_sym_POUND] = ACTIONS(2752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_API_AVAILABLE] = ACTIONS(2752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_API_DEPRECATED] = ACTIONS(2752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2752),
    [anon_sym___deprecated_msg] = ACTIONS(2752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2752),
    [anon_sym_ATprotocol] = ACTIONS(2754),
    [anon_sym_ATinterface] = ACTIONS(2754),
    [anon_sym_ATimplementation] = ACTIONS(2754),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2754),
    [anon_sym_ATselector] = ACTIONS(2754),
    [anon_sym_ATavailable] = ACTIONS(2754),
    [anon_sym___builtin_available] = ACTIONS(2752),
    [anon_sym_va_arg] = ACTIONS(2752),
    [anon_sym_ATencode] = ACTIONS(2754),
    [anon_sym_BOOL] = ACTIONS(2752),
    [anon_sym_IMP] = ACTIONS(2752),
    [anon_sym_SEL] = ACTIONS(2752),
    [anon_sym_Class] = ACTIONS(2752),
    [anon_sym_id] = ACTIONS(2752),
  },
  [STATE(1054)] = {
    [ts_builtin_sym_end] = ACTIONS(2758),
    [sym_identifier] = ACTIONS(2756),
    [aux_sym_preproc_include_token1] = ACTIONS(2756),
    [aux_sym_preproc_include_token2] = ACTIONS(2756),
    [aux_sym_preproc_def_token1] = ACTIONS(2756),
    [aux_sym_preproc_if_token1] = ACTIONS(2756),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2756),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2756),
    [sym_preproc_directive] = ACTIONS(2756),
    [anon_sym_LPAREN2] = ACTIONS(2758),
    [anon_sym_BANG] = ACTIONS(2758),
    [anon_sym_TILDE] = ACTIONS(2758),
    [anon_sym_DASH] = ACTIONS(2756),
    [anon_sym_PLUS] = ACTIONS(2756),
    [anon_sym_STAR] = ACTIONS(2758),
    [anon_sym_CARET] = ACTIONS(2758),
    [anon_sym_AMP] = ACTIONS(2758),
    [anon_sym_SEMI] = ACTIONS(2758),
    [anon_sym___extension__] = ACTIONS(2756),
    [anon_sym_typedef] = ACTIONS(2756),
    [anon_sym_extern] = ACTIONS(2756),
    [anon_sym___attribute__] = ACTIONS(2756),
    [anon_sym___attribute] = ACTIONS(2756),
    [anon_sym_noreturn] = ACTIONS(2756),
    [anon_sym_LBRACK] = ACTIONS(2758),
    [anon_sym___declspec] = ACTIONS(2756),
    [anon_sym___cdecl] = ACTIONS(2756),
    [anon_sym___clrcall] = ACTIONS(2756),
    [anon_sym___stdcall] = ACTIONS(2756),
    [anon_sym___fastcall] = ACTIONS(2756),
    [anon_sym___thiscall] = ACTIONS(2756),
    [anon_sym___vectorcall] = ACTIONS(2756),
    [anon_sym_LBRACE] = ACTIONS(2758),
    [anon_sym_signed] = ACTIONS(2756),
    [anon_sym_unsigned] = ACTIONS(2756),
    [anon_sym_long] = ACTIONS(2756),
    [anon_sym_short] = ACTIONS(2756),
    [anon_sym_static] = ACTIONS(2756),
    [anon_sym_ATautoreleasepool] = ACTIONS(2758),
    [anon_sym_auto] = ACTIONS(2756),
    [anon_sym_register] = ACTIONS(2756),
    [anon_sym_inline] = ACTIONS(2756),
    [anon_sym___inline] = ACTIONS(2756),
    [anon_sym___inline__] = ACTIONS(2756),
    [anon_sym___forceinline] = ACTIONS(2756),
    [anon_sym_thread_local] = ACTIONS(2756),
    [anon_sym___thread] = ACTIONS(2756),
    [anon_sym_CG_EXTERN] = ACTIONS(2756),
    [anon_sym_CG_INLINE] = ACTIONS(2756),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2756),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2756),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2756),
    [anon_sym_IBOutlet] = ACTIONS(2756),
    [anon_sym_IBInspectable] = ACTIONS(2756),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2756),
    [anon_sym_NS_INLINE] = ACTIONS(2756),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2756),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2756),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2756),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2756),
    [anon_sym_const] = ACTIONS(2756),
    [anon_sym_constexpr] = ACTIONS(2756),
    [anon_sym_volatile] = ACTIONS(2756),
    [anon_sym_restrict] = ACTIONS(2756),
    [anon_sym___restrict__] = ACTIONS(2756),
    [anon_sym__Atomic] = ACTIONS(2756),
    [anon_sym__Noreturn] = ACTIONS(2756),
    [anon_sym__Nonnull] = ACTIONS(2756),
    [anon_sym_nullable] = ACTIONS(2756),
    [anon_sym__Complex] = ACTIONS(2756),
    [anon_sym__Nullable] = ACTIONS(2756),
    [anon_sym__Nullable_result] = ACTIONS(2756),
    [anon_sym__Null_unspecified] = ACTIONS(2756),
    [anon_sym___autoreleasing] = ACTIONS(2756),
    [anon_sym___block] = ACTIONS(2756),
    [anon_sym___bridge] = ACTIONS(2756),
    [anon_sym___bridge_retained] = ACTIONS(2756),
    [anon_sym___bridge_transfer] = ACTIONS(2756),
    [anon_sym___complex] = ACTIONS(2756),
    [anon_sym___const] = ACTIONS(2756),
    [anon_sym___imag] = ACTIONS(2756),
    [anon_sym___kindof] = ACTIONS(2756),
    [anon_sym___nonnull] = ACTIONS(2756),
    [anon_sym___nullable] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2756),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2756),
    [anon_sym___real] = ACTIONS(2756),
    [anon_sym___strong] = ACTIONS(2756),
    [anon_sym___unsafe_unretained] = ACTIONS(2756),
    [anon_sym___unused] = ACTIONS(2756),
    [anon_sym___weak] = ACTIONS(2756),
    [anon_sym_alignas] = ACTIONS(2756),
    [anon_sym__Alignas] = ACTIONS(2756),
    [sym_primitive_type] = ACTIONS(2756),
    [anon_sym_enum] = ACTIONS(2756),
    [anon_sym_struct] = ACTIONS(2756),
    [anon_sym_union] = ACTIONS(2756),
    [anon_sym_if] = ACTIONS(2756),
    [anon_sym_switch] = ACTIONS(2756),
    [anon_sym_case] = ACTIONS(2756),
    [anon_sym_default] = ACTIONS(2756),
    [anon_sym_while] = ACTIONS(2756),
    [anon_sym_do] = ACTIONS(2756),
    [anon_sym_for] = ACTIONS(2756),
    [anon_sym_in] = ACTIONS(2756),
    [anon_sym_return] = ACTIONS(2756),
    [anon_sym_break] = ACTIONS(2756),
    [anon_sym_continue] = ACTIONS(2756),
    [anon_sym_goto] = ACTIONS(2756),
    [anon_sym_DASH_DASH] = ACTIONS(2758),
    [anon_sym_PLUS_PLUS] = ACTIONS(2758),
    [anon_sym_sizeof] = ACTIONS(2756),
    [anon_sym___alignof__] = ACTIONS(2756),
    [anon_sym___alignof] = ACTIONS(2756),
    [anon_sym__alignof] = ACTIONS(2756),
    [anon_sym_alignof] = ACTIONS(2756),
    [anon_sym__Alignof] = ACTIONS(2756),
    [anon_sym_offsetof] = ACTIONS(2756),
    [anon_sym__Generic] = ACTIONS(2756),
    [anon_sym_asm] = ACTIONS(2756),
    [anon_sym___asm__] = ACTIONS(2756),
    [anon_sym___asm] = ACTIONS(2756),
    [sym_number_literal] = ACTIONS(2758),
    [anon_sym_L_SQUOTE] = ACTIONS(2758),
    [anon_sym_u_SQUOTE] = ACTIONS(2758),
    [anon_sym_U_SQUOTE] = ACTIONS(2758),
    [anon_sym_u8_SQUOTE] = ACTIONS(2758),
    [anon_sym_SQUOTE] = ACTIONS(2758),
    [anon_sym_AT] = ACTIONS(2756),
    [anon_sym_DQUOTE] = ACTIONS(2758),
    [anon_sym_L_DQUOTE] = ACTIONS(2758),
    [anon_sym_u_DQUOTE] = ACTIONS(2758),
    [anon_sym_U_DQUOTE] = ACTIONS(2758),
    [anon_sym_u8_DQUOTE] = ACTIONS(2758),
    [sym_true] = ACTIONS(2756),
    [sym_false] = ACTIONS(2756),
    [anon_sym_NULL] = ACTIONS(2756),
    [anon_sym_nullptr] = ACTIONS(2756),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2756),
    [anon_sym___typeof] = ACTIONS(2756),
    [anon_sym_typeof] = ACTIONS(2756),
    [anon_sym_ATimport] = ACTIONS(2758),
    [aux_sym_preproc_undef_token1] = ACTIONS(2756),
    [anon_sym_POUND] = ACTIONS(2756),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2756),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2756),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2756),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2756),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2756),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2756),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2756),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_API_AVAILABLE] = ACTIONS(2756),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_API_DEPRECATED] = ACTIONS(2756),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2756),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2756),
    [anon_sym___deprecated_msg] = ACTIONS(2756),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2756),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2756),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2756),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2756),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2756),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2756),
    [anon_sym_ATprotocol] = ACTIONS(2758),
    [anon_sym_ATinterface] = ACTIONS(2758),
    [anon_sym_ATimplementation] = ACTIONS(2758),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2758),
    [anon_sym_ATselector] = ACTIONS(2758),
    [anon_sym_ATavailable] = ACTIONS(2758),
    [anon_sym___builtin_available] = ACTIONS(2756),
    [anon_sym_va_arg] = ACTIONS(2756),
    [anon_sym_ATencode] = ACTIONS(2758),
    [anon_sym_BOOL] = ACTIONS(2756),
    [anon_sym_IMP] = ACTIONS(2756),
    [anon_sym_SEL] = ACTIONS(2756),
    [anon_sym_Class] = ACTIONS(2756),
    [anon_sym_id] = ACTIONS(2756),
  },
  [STATE(1055)] = {
    [ts_builtin_sym_end] = ACTIONS(2762),
    [sym_identifier] = ACTIONS(2760),
    [aux_sym_preproc_include_token1] = ACTIONS(2760),
    [aux_sym_preproc_include_token2] = ACTIONS(2760),
    [aux_sym_preproc_def_token1] = ACTIONS(2760),
    [aux_sym_preproc_if_token1] = ACTIONS(2760),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2760),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2760),
    [sym_preproc_directive] = ACTIONS(2760),
    [anon_sym_LPAREN2] = ACTIONS(2762),
    [anon_sym_BANG] = ACTIONS(2762),
    [anon_sym_TILDE] = ACTIONS(2762),
    [anon_sym_DASH] = ACTIONS(2760),
    [anon_sym_PLUS] = ACTIONS(2760),
    [anon_sym_STAR] = ACTIONS(2762),
    [anon_sym_CARET] = ACTIONS(2762),
    [anon_sym_AMP] = ACTIONS(2762),
    [anon_sym_SEMI] = ACTIONS(2762),
    [anon_sym___extension__] = ACTIONS(2760),
    [anon_sym_typedef] = ACTIONS(2760),
    [anon_sym_extern] = ACTIONS(2760),
    [anon_sym___attribute__] = ACTIONS(2760),
    [anon_sym___attribute] = ACTIONS(2760),
    [anon_sym_noreturn] = ACTIONS(2760),
    [anon_sym_LBRACK] = ACTIONS(2762),
    [anon_sym___declspec] = ACTIONS(2760),
    [anon_sym___cdecl] = ACTIONS(2760),
    [anon_sym___clrcall] = ACTIONS(2760),
    [anon_sym___stdcall] = ACTIONS(2760),
    [anon_sym___fastcall] = ACTIONS(2760),
    [anon_sym___thiscall] = ACTIONS(2760),
    [anon_sym___vectorcall] = ACTIONS(2760),
    [anon_sym_LBRACE] = ACTIONS(2762),
    [anon_sym_signed] = ACTIONS(2760),
    [anon_sym_unsigned] = ACTIONS(2760),
    [anon_sym_long] = ACTIONS(2760),
    [anon_sym_short] = ACTIONS(2760),
    [anon_sym_static] = ACTIONS(2760),
    [anon_sym_ATautoreleasepool] = ACTIONS(2762),
    [anon_sym_auto] = ACTIONS(2760),
    [anon_sym_register] = ACTIONS(2760),
    [anon_sym_inline] = ACTIONS(2760),
    [anon_sym___inline] = ACTIONS(2760),
    [anon_sym___inline__] = ACTIONS(2760),
    [anon_sym___forceinline] = ACTIONS(2760),
    [anon_sym_thread_local] = ACTIONS(2760),
    [anon_sym___thread] = ACTIONS(2760),
    [anon_sym_CG_EXTERN] = ACTIONS(2760),
    [anon_sym_CG_INLINE] = ACTIONS(2760),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2760),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2760),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2760),
    [anon_sym_IBOutlet] = ACTIONS(2760),
    [anon_sym_IBInspectable] = ACTIONS(2760),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2760),
    [anon_sym_NS_INLINE] = ACTIONS(2760),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2760),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2760),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2760),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2760),
    [anon_sym_const] = ACTIONS(2760),
    [anon_sym_constexpr] = ACTIONS(2760),
    [anon_sym_volatile] = ACTIONS(2760),
    [anon_sym_restrict] = ACTIONS(2760),
    [anon_sym___restrict__] = ACTIONS(2760),
    [anon_sym__Atomic] = ACTIONS(2760),
    [anon_sym__Noreturn] = ACTIONS(2760),
    [anon_sym__Nonnull] = ACTIONS(2760),
    [anon_sym_nullable] = ACTIONS(2760),
    [anon_sym__Complex] = ACTIONS(2760),
    [anon_sym__Nullable] = ACTIONS(2760),
    [anon_sym__Nullable_result] = ACTIONS(2760),
    [anon_sym__Null_unspecified] = ACTIONS(2760),
    [anon_sym___autoreleasing] = ACTIONS(2760),
    [anon_sym___block] = ACTIONS(2760),
    [anon_sym___bridge] = ACTIONS(2760),
    [anon_sym___bridge_retained] = ACTIONS(2760),
    [anon_sym___bridge_transfer] = ACTIONS(2760),
    [anon_sym___complex] = ACTIONS(2760),
    [anon_sym___const] = ACTIONS(2760),
    [anon_sym___imag] = ACTIONS(2760),
    [anon_sym___kindof] = ACTIONS(2760),
    [anon_sym___nonnull] = ACTIONS(2760),
    [anon_sym___nullable] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2760),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2760),
    [anon_sym___real] = ACTIONS(2760),
    [anon_sym___strong] = ACTIONS(2760),
    [anon_sym___unsafe_unretained] = ACTIONS(2760),
    [anon_sym___unused] = ACTIONS(2760),
    [anon_sym___weak] = ACTIONS(2760),
    [anon_sym_alignas] = ACTIONS(2760),
    [anon_sym__Alignas] = ACTIONS(2760),
    [sym_primitive_type] = ACTIONS(2760),
    [anon_sym_enum] = ACTIONS(2760),
    [anon_sym_struct] = ACTIONS(2760),
    [anon_sym_union] = ACTIONS(2760),
    [anon_sym_if] = ACTIONS(2760),
    [anon_sym_switch] = ACTIONS(2760),
    [anon_sym_case] = ACTIONS(2760),
    [anon_sym_default] = ACTIONS(2760),
    [anon_sym_while] = ACTIONS(2760),
    [anon_sym_do] = ACTIONS(2760),
    [anon_sym_for] = ACTIONS(2760),
    [anon_sym_in] = ACTIONS(2760),
    [anon_sym_return] = ACTIONS(2760),
    [anon_sym_break] = ACTIONS(2760),
    [anon_sym_continue] = ACTIONS(2760),
    [anon_sym_goto] = ACTIONS(2760),
    [anon_sym_DASH_DASH] = ACTIONS(2762),
    [anon_sym_PLUS_PLUS] = ACTIONS(2762),
    [anon_sym_sizeof] = ACTIONS(2760),
    [anon_sym___alignof__] = ACTIONS(2760),
    [anon_sym___alignof] = ACTIONS(2760),
    [anon_sym__alignof] = ACTIONS(2760),
    [anon_sym_alignof] = ACTIONS(2760),
    [anon_sym__Alignof] = ACTIONS(2760),
    [anon_sym_offsetof] = ACTIONS(2760),
    [anon_sym__Generic] = ACTIONS(2760),
    [anon_sym_asm] = ACTIONS(2760),
    [anon_sym___asm__] = ACTIONS(2760),
    [anon_sym___asm] = ACTIONS(2760),
    [sym_number_literal] = ACTIONS(2762),
    [anon_sym_L_SQUOTE] = ACTIONS(2762),
    [anon_sym_u_SQUOTE] = ACTIONS(2762),
    [anon_sym_U_SQUOTE] = ACTIONS(2762),
    [anon_sym_u8_SQUOTE] = ACTIONS(2762),
    [anon_sym_SQUOTE] = ACTIONS(2762),
    [anon_sym_AT] = ACTIONS(2760),
    [anon_sym_DQUOTE] = ACTIONS(2762),
    [anon_sym_L_DQUOTE] = ACTIONS(2762),
    [anon_sym_u_DQUOTE] = ACTIONS(2762),
    [anon_sym_U_DQUOTE] = ACTIONS(2762),
    [anon_sym_u8_DQUOTE] = ACTIONS(2762),
    [sym_true] = ACTIONS(2760),
    [sym_false] = ACTIONS(2760),
    [anon_sym_NULL] = ACTIONS(2760),
    [anon_sym_nullptr] = ACTIONS(2760),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2760),
    [anon_sym___typeof] = ACTIONS(2760),
    [anon_sym_typeof] = ACTIONS(2760),
    [anon_sym_ATimport] = ACTIONS(2762),
    [aux_sym_preproc_undef_token1] = ACTIONS(2760),
    [anon_sym_POUND] = ACTIONS(2760),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2760),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2760),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2760),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2760),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2760),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2760),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2760),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_API_AVAILABLE] = ACTIONS(2760),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_API_DEPRECATED] = ACTIONS(2760),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2760),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2760),
    [anon_sym___deprecated_msg] = ACTIONS(2760),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2760),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2760),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2760),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2760),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2760),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2760),
    [anon_sym_ATprotocol] = ACTIONS(2762),
    [anon_sym_ATinterface] = ACTIONS(2762),
    [anon_sym_ATimplementation] = ACTIONS(2762),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2762),
    [anon_sym_ATselector] = ACTIONS(2762),
    [anon_sym_ATavailable] = ACTIONS(2762),
    [anon_sym___builtin_available] = ACTIONS(2760),
    [anon_sym_va_arg] = ACTIONS(2760),
    [anon_sym_ATencode] = ACTIONS(2762),
    [anon_sym_BOOL] = ACTIONS(2760),
    [anon_sym_IMP] = ACTIONS(2760),
    [anon_sym_SEL] = ACTIONS(2760),
    [anon_sym_Class] = ACTIONS(2760),
    [anon_sym_id] = ACTIONS(2760),
  },
  [STATE(1056)] = {
    [ts_builtin_sym_end] = ACTIONS(2766),
    [sym_identifier] = ACTIONS(2764),
    [aux_sym_preproc_include_token1] = ACTIONS(2764),
    [aux_sym_preproc_include_token2] = ACTIONS(2764),
    [aux_sym_preproc_def_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2764),
    [sym_preproc_directive] = ACTIONS(2764),
    [anon_sym_LPAREN2] = ACTIONS(2766),
    [anon_sym_BANG] = ACTIONS(2766),
    [anon_sym_TILDE] = ACTIONS(2766),
    [anon_sym_DASH] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(2764),
    [anon_sym_STAR] = ACTIONS(2766),
    [anon_sym_CARET] = ACTIONS(2766),
    [anon_sym_AMP] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2766),
    [anon_sym___extension__] = ACTIONS(2764),
    [anon_sym_typedef] = ACTIONS(2764),
    [anon_sym_extern] = ACTIONS(2764),
    [anon_sym___attribute__] = ACTIONS(2764),
    [anon_sym___attribute] = ACTIONS(2764),
    [anon_sym_noreturn] = ACTIONS(2764),
    [anon_sym_LBRACK] = ACTIONS(2766),
    [anon_sym___declspec] = ACTIONS(2764),
    [anon_sym___cdecl] = ACTIONS(2764),
    [anon_sym___clrcall] = ACTIONS(2764),
    [anon_sym___stdcall] = ACTIONS(2764),
    [anon_sym___fastcall] = ACTIONS(2764),
    [anon_sym___thiscall] = ACTIONS(2764),
    [anon_sym___vectorcall] = ACTIONS(2764),
    [anon_sym_LBRACE] = ACTIONS(2766),
    [anon_sym_signed] = ACTIONS(2764),
    [anon_sym_unsigned] = ACTIONS(2764),
    [anon_sym_long] = ACTIONS(2764),
    [anon_sym_short] = ACTIONS(2764),
    [anon_sym_static] = ACTIONS(2764),
    [anon_sym_ATautoreleasepool] = ACTIONS(2766),
    [anon_sym_auto] = ACTIONS(2764),
    [anon_sym_register] = ACTIONS(2764),
    [anon_sym_inline] = ACTIONS(2764),
    [anon_sym___inline] = ACTIONS(2764),
    [anon_sym___inline__] = ACTIONS(2764),
    [anon_sym___forceinline] = ACTIONS(2764),
    [anon_sym_thread_local] = ACTIONS(2764),
    [anon_sym___thread] = ACTIONS(2764),
    [anon_sym_CG_EXTERN] = ACTIONS(2764),
    [anon_sym_CG_INLINE] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2764),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2764),
    [anon_sym_IBOutlet] = ACTIONS(2764),
    [anon_sym_IBInspectable] = ACTIONS(2764),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2764),
    [anon_sym_NS_INLINE] = ACTIONS(2764),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2764),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2764),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2764),
    [anon_sym_const] = ACTIONS(2764),
    [anon_sym_constexpr] = ACTIONS(2764),
    [anon_sym_volatile] = ACTIONS(2764),
    [anon_sym_restrict] = ACTIONS(2764),
    [anon_sym___restrict__] = ACTIONS(2764),
    [anon_sym__Atomic] = ACTIONS(2764),
    [anon_sym__Noreturn] = ACTIONS(2764),
    [anon_sym__Nonnull] = ACTIONS(2764),
    [anon_sym_nullable] = ACTIONS(2764),
    [anon_sym__Complex] = ACTIONS(2764),
    [anon_sym__Nullable] = ACTIONS(2764),
    [anon_sym__Nullable_result] = ACTIONS(2764),
    [anon_sym__Null_unspecified] = ACTIONS(2764),
    [anon_sym___autoreleasing] = ACTIONS(2764),
    [anon_sym___block] = ACTIONS(2764),
    [anon_sym___bridge] = ACTIONS(2764),
    [anon_sym___bridge_retained] = ACTIONS(2764),
    [anon_sym___bridge_transfer] = ACTIONS(2764),
    [anon_sym___complex] = ACTIONS(2764),
    [anon_sym___const] = ACTIONS(2764),
    [anon_sym___imag] = ACTIONS(2764),
    [anon_sym___kindof] = ACTIONS(2764),
    [anon_sym___nonnull] = ACTIONS(2764),
    [anon_sym___nullable] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2764),
    [anon_sym___real] = ACTIONS(2764),
    [anon_sym___strong] = ACTIONS(2764),
    [anon_sym___unsafe_unretained] = ACTIONS(2764),
    [anon_sym___unused] = ACTIONS(2764),
    [anon_sym___weak] = ACTIONS(2764),
    [anon_sym_alignas] = ACTIONS(2764),
    [anon_sym__Alignas] = ACTIONS(2764),
    [sym_primitive_type] = ACTIONS(2764),
    [anon_sym_enum] = ACTIONS(2764),
    [anon_sym_struct] = ACTIONS(2764),
    [anon_sym_union] = ACTIONS(2764),
    [anon_sym_if] = ACTIONS(2764),
    [anon_sym_switch] = ACTIONS(2764),
    [anon_sym_case] = ACTIONS(2764),
    [anon_sym_default] = ACTIONS(2764),
    [anon_sym_while] = ACTIONS(2764),
    [anon_sym_do] = ACTIONS(2764),
    [anon_sym_for] = ACTIONS(2764),
    [anon_sym_in] = ACTIONS(2764),
    [anon_sym_return] = ACTIONS(2764),
    [anon_sym_break] = ACTIONS(2764),
    [anon_sym_continue] = ACTIONS(2764),
    [anon_sym_goto] = ACTIONS(2764),
    [anon_sym_DASH_DASH] = ACTIONS(2766),
    [anon_sym_PLUS_PLUS] = ACTIONS(2766),
    [anon_sym_sizeof] = ACTIONS(2764),
    [anon_sym___alignof__] = ACTIONS(2764),
    [anon_sym___alignof] = ACTIONS(2764),
    [anon_sym__alignof] = ACTIONS(2764),
    [anon_sym_alignof] = ACTIONS(2764),
    [anon_sym__Alignof] = ACTIONS(2764),
    [anon_sym_offsetof] = ACTIONS(2764),
    [anon_sym__Generic] = ACTIONS(2764),
    [anon_sym_asm] = ACTIONS(2764),
    [anon_sym___asm__] = ACTIONS(2764),
    [anon_sym___asm] = ACTIONS(2764),
    [sym_number_literal] = ACTIONS(2766),
    [anon_sym_L_SQUOTE] = ACTIONS(2766),
    [anon_sym_u_SQUOTE] = ACTIONS(2766),
    [anon_sym_U_SQUOTE] = ACTIONS(2766),
    [anon_sym_u8_SQUOTE] = ACTIONS(2766),
    [anon_sym_SQUOTE] = ACTIONS(2766),
    [anon_sym_AT] = ACTIONS(2764),
    [anon_sym_DQUOTE] = ACTIONS(2766),
    [anon_sym_L_DQUOTE] = ACTIONS(2766),
    [anon_sym_u_DQUOTE] = ACTIONS(2766),
    [anon_sym_U_DQUOTE] = ACTIONS(2766),
    [anon_sym_u8_DQUOTE] = ACTIONS(2766),
    [sym_true] = ACTIONS(2764),
    [sym_false] = ACTIONS(2764),
    [anon_sym_NULL] = ACTIONS(2764),
    [anon_sym_nullptr] = ACTIONS(2764),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2764),
    [anon_sym___typeof] = ACTIONS(2764),
    [anon_sym_typeof] = ACTIONS(2764),
    [anon_sym_ATimport] = ACTIONS(2766),
    [aux_sym_preproc_undef_token1] = ACTIONS(2764),
    [anon_sym_POUND] = ACTIONS(2764),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2764),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2764),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2764),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_API_AVAILABLE] = ACTIONS(2764),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_API_DEPRECATED] = ACTIONS(2764),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym___deprecated_msg] = ACTIONS(2764),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2764),
    [anon_sym_ATprotocol] = ACTIONS(2766),
    [anon_sym_ATinterface] = ACTIONS(2766),
    [anon_sym_ATimplementation] = ACTIONS(2766),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2766),
    [anon_sym_ATselector] = ACTIONS(2766),
    [anon_sym_ATavailable] = ACTIONS(2766),
    [anon_sym___builtin_available] = ACTIONS(2764),
    [anon_sym_va_arg] = ACTIONS(2764),
    [anon_sym_ATencode] = ACTIONS(2766),
    [anon_sym_BOOL] = ACTIONS(2764),
    [anon_sym_IMP] = ACTIONS(2764),
    [anon_sym_SEL] = ACTIONS(2764),
    [anon_sym_Class] = ACTIONS(2764),
    [anon_sym_id] = ACTIONS(2764),
  },
  [STATE(1057)] = {
    [ts_builtin_sym_end] = ACTIONS(2766),
    [sym_identifier] = ACTIONS(2764),
    [aux_sym_preproc_include_token1] = ACTIONS(2764),
    [aux_sym_preproc_include_token2] = ACTIONS(2764),
    [aux_sym_preproc_def_token1] = ACTIONS(2764),
    [aux_sym_preproc_if_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2764),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2764),
    [sym_preproc_directive] = ACTIONS(2764),
    [anon_sym_LPAREN2] = ACTIONS(2766),
    [anon_sym_BANG] = ACTIONS(2766),
    [anon_sym_TILDE] = ACTIONS(2766),
    [anon_sym_DASH] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(2764),
    [anon_sym_STAR] = ACTIONS(2766),
    [anon_sym_CARET] = ACTIONS(2766),
    [anon_sym_AMP] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2766),
    [anon_sym___extension__] = ACTIONS(2764),
    [anon_sym_typedef] = ACTIONS(2764),
    [anon_sym_extern] = ACTIONS(2764),
    [anon_sym___attribute__] = ACTIONS(2764),
    [anon_sym___attribute] = ACTIONS(2764),
    [anon_sym_noreturn] = ACTIONS(2764),
    [anon_sym_LBRACK] = ACTIONS(2766),
    [anon_sym___declspec] = ACTIONS(2764),
    [anon_sym___cdecl] = ACTIONS(2764),
    [anon_sym___clrcall] = ACTIONS(2764),
    [anon_sym___stdcall] = ACTIONS(2764),
    [anon_sym___fastcall] = ACTIONS(2764),
    [anon_sym___thiscall] = ACTIONS(2764),
    [anon_sym___vectorcall] = ACTIONS(2764),
    [anon_sym_LBRACE] = ACTIONS(2766),
    [anon_sym_signed] = ACTIONS(2764),
    [anon_sym_unsigned] = ACTIONS(2764),
    [anon_sym_long] = ACTIONS(2764),
    [anon_sym_short] = ACTIONS(2764),
    [anon_sym_static] = ACTIONS(2764),
    [anon_sym_ATautoreleasepool] = ACTIONS(2766),
    [anon_sym_auto] = ACTIONS(2764),
    [anon_sym_register] = ACTIONS(2764),
    [anon_sym_inline] = ACTIONS(2764),
    [anon_sym___inline] = ACTIONS(2764),
    [anon_sym___inline__] = ACTIONS(2764),
    [anon_sym___forceinline] = ACTIONS(2764),
    [anon_sym_thread_local] = ACTIONS(2764),
    [anon_sym___thread] = ACTIONS(2764),
    [anon_sym_CG_EXTERN] = ACTIONS(2764),
    [anon_sym_CG_INLINE] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2764),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2764),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2764),
    [anon_sym_IBOutlet] = ACTIONS(2764),
    [anon_sym_IBInspectable] = ACTIONS(2764),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2764),
    [anon_sym_NS_INLINE] = ACTIONS(2764),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2764),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2764),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2764),
    [anon_sym_const] = ACTIONS(2764),
    [anon_sym_constexpr] = ACTIONS(2764),
    [anon_sym_volatile] = ACTIONS(2764),
    [anon_sym_restrict] = ACTIONS(2764),
    [anon_sym___restrict__] = ACTIONS(2764),
    [anon_sym__Atomic] = ACTIONS(2764),
    [anon_sym__Noreturn] = ACTIONS(2764),
    [anon_sym__Nonnull] = ACTIONS(2764),
    [anon_sym_nullable] = ACTIONS(2764),
    [anon_sym__Complex] = ACTIONS(2764),
    [anon_sym__Nullable] = ACTIONS(2764),
    [anon_sym__Nullable_result] = ACTIONS(2764),
    [anon_sym__Null_unspecified] = ACTIONS(2764),
    [anon_sym___autoreleasing] = ACTIONS(2764),
    [anon_sym___block] = ACTIONS(2764),
    [anon_sym___bridge] = ACTIONS(2764),
    [anon_sym___bridge_retained] = ACTIONS(2764),
    [anon_sym___bridge_transfer] = ACTIONS(2764),
    [anon_sym___complex] = ACTIONS(2764),
    [anon_sym___const] = ACTIONS(2764),
    [anon_sym___imag] = ACTIONS(2764),
    [anon_sym___kindof] = ACTIONS(2764),
    [anon_sym___nonnull] = ACTIONS(2764),
    [anon_sym___nullable] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2764),
    [anon_sym___real] = ACTIONS(2764),
    [anon_sym___strong] = ACTIONS(2764),
    [anon_sym___unsafe_unretained] = ACTIONS(2764),
    [anon_sym___unused] = ACTIONS(2764),
    [anon_sym___weak] = ACTIONS(2764),
    [anon_sym_alignas] = ACTIONS(2764),
    [anon_sym__Alignas] = ACTIONS(2764),
    [sym_primitive_type] = ACTIONS(2764),
    [anon_sym_enum] = ACTIONS(2764),
    [anon_sym_struct] = ACTIONS(2764),
    [anon_sym_union] = ACTIONS(2764),
    [anon_sym_if] = ACTIONS(2764),
    [anon_sym_switch] = ACTIONS(2764),
    [anon_sym_case] = ACTIONS(2764),
    [anon_sym_default] = ACTIONS(2764),
    [anon_sym_while] = ACTIONS(2764),
    [anon_sym_do] = ACTIONS(2764),
    [anon_sym_for] = ACTIONS(2764),
    [anon_sym_in] = ACTIONS(2764),
    [anon_sym_return] = ACTIONS(2764),
    [anon_sym_break] = ACTIONS(2764),
    [anon_sym_continue] = ACTIONS(2764),
    [anon_sym_goto] = ACTIONS(2764),
    [anon_sym_DASH_DASH] = ACTIONS(2766),
    [anon_sym_PLUS_PLUS] = ACTIONS(2766),
    [anon_sym_sizeof] = ACTIONS(2764),
    [anon_sym___alignof__] = ACTIONS(2764),
    [anon_sym___alignof] = ACTIONS(2764),
    [anon_sym__alignof] = ACTIONS(2764),
    [anon_sym_alignof] = ACTIONS(2764),
    [anon_sym__Alignof] = ACTIONS(2764),
    [anon_sym_offsetof] = ACTIONS(2764),
    [anon_sym__Generic] = ACTIONS(2764),
    [anon_sym_asm] = ACTIONS(2764),
    [anon_sym___asm__] = ACTIONS(2764),
    [anon_sym___asm] = ACTIONS(2764),
    [sym_number_literal] = ACTIONS(2766),
    [anon_sym_L_SQUOTE] = ACTIONS(2766),
    [anon_sym_u_SQUOTE] = ACTIONS(2766),
    [anon_sym_U_SQUOTE] = ACTIONS(2766),
    [anon_sym_u8_SQUOTE] = ACTIONS(2766),
    [anon_sym_SQUOTE] = ACTIONS(2766),
    [anon_sym_AT] = ACTIONS(2764),
    [anon_sym_DQUOTE] = ACTIONS(2766),
    [anon_sym_L_DQUOTE] = ACTIONS(2766),
    [anon_sym_u_DQUOTE] = ACTIONS(2766),
    [anon_sym_U_DQUOTE] = ACTIONS(2766),
    [anon_sym_u8_DQUOTE] = ACTIONS(2766),
    [sym_true] = ACTIONS(2764),
    [sym_false] = ACTIONS(2764),
    [anon_sym_NULL] = ACTIONS(2764),
    [anon_sym_nullptr] = ACTIONS(2764),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2764),
    [anon_sym___typeof] = ACTIONS(2764),
    [anon_sym_typeof] = ACTIONS(2764),
    [anon_sym_ATimport] = ACTIONS(2766),
    [aux_sym_preproc_undef_token1] = ACTIONS(2764),
    [anon_sym_POUND] = ACTIONS(2764),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2764),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2764),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2764),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2764),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2764),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_API_AVAILABLE] = ACTIONS(2764),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_API_DEPRECATED] = ACTIONS(2764),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2764),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2764),
    [anon_sym___deprecated_msg] = ACTIONS(2764),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2764),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2764),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2764),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2764),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2764),
    [anon_sym_ATprotocol] = ACTIONS(2766),
    [anon_sym_ATinterface] = ACTIONS(2766),
    [anon_sym_ATimplementation] = ACTIONS(2766),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2766),
    [anon_sym_ATselector] = ACTIONS(2766),
    [anon_sym_ATavailable] = ACTIONS(2766),
    [anon_sym___builtin_available] = ACTIONS(2764),
    [anon_sym_va_arg] = ACTIONS(2764),
    [anon_sym_ATencode] = ACTIONS(2766),
    [anon_sym_BOOL] = ACTIONS(2764),
    [anon_sym_IMP] = ACTIONS(2764),
    [anon_sym_SEL] = ACTIONS(2764),
    [anon_sym_Class] = ACTIONS(2764),
    [anon_sym_id] = ACTIONS(2764),
  },
  [STATE(1058)] = {
    [ts_builtin_sym_end] = ACTIONS(2602),
    [sym_identifier] = ACTIONS(2600),
    [aux_sym_preproc_include_token1] = ACTIONS(2600),
    [aux_sym_preproc_include_token2] = ACTIONS(2600),
    [aux_sym_preproc_def_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2600),
    [sym_preproc_directive] = ACTIONS(2600),
    [anon_sym_LPAREN2] = ACTIONS(2602),
    [anon_sym_BANG] = ACTIONS(2602),
    [anon_sym_TILDE] = ACTIONS(2602),
    [anon_sym_DASH] = ACTIONS(2600),
    [anon_sym_PLUS] = ACTIONS(2600),
    [anon_sym_STAR] = ACTIONS(2602),
    [anon_sym_CARET] = ACTIONS(2602),
    [anon_sym_AMP] = ACTIONS(2602),
    [anon_sym_SEMI] = ACTIONS(2602),
    [anon_sym___extension__] = ACTIONS(2600),
    [anon_sym_typedef] = ACTIONS(2600),
    [anon_sym_extern] = ACTIONS(2600),
    [anon_sym___attribute__] = ACTIONS(2600),
    [anon_sym___attribute] = ACTIONS(2600),
    [anon_sym_noreturn] = ACTIONS(2600),
    [anon_sym_LBRACK] = ACTIONS(2602),
    [anon_sym___declspec] = ACTIONS(2600),
    [anon_sym___cdecl] = ACTIONS(2600),
    [anon_sym___clrcall] = ACTIONS(2600),
    [anon_sym___stdcall] = ACTIONS(2600),
    [anon_sym___fastcall] = ACTIONS(2600),
    [anon_sym___thiscall] = ACTIONS(2600),
    [anon_sym___vectorcall] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_signed] = ACTIONS(2600),
    [anon_sym_unsigned] = ACTIONS(2600),
    [anon_sym_long] = ACTIONS(2600),
    [anon_sym_short] = ACTIONS(2600),
    [anon_sym_static] = ACTIONS(2600),
    [anon_sym_ATautoreleasepool] = ACTIONS(2602),
    [anon_sym_auto] = ACTIONS(2600),
    [anon_sym_register] = ACTIONS(2600),
    [anon_sym_inline] = ACTIONS(2600),
    [anon_sym___inline] = ACTIONS(2600),
    [anon_sym___inline__] = ACTIONS(2600),
    [anon_sym___forceinline] = ACTIONS(2600),
    [anon_sym_thread_local] = ACTIONS(2600),
    [anon_sym___thread] = ACTIONS(2600),
    [anon_sym_CG_EXTERN] = ACTIONS(2600),
    [anon_sym_CG_INLINE] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2600),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2600),
    [anon_sym_IBOutlet] = ACTIONS(2600),
    [anon_sym_IBInspectable] = ACTIONS(2600),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2600),
    [anon_sym_NS_INLINE] = ACTIONS(2600),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2600),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2600),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2600),
    [anon_sym_const] = ACTIONS(2600),
    [anon_sym_constexpr] = ACTIONS(2600),
    [anon_sym_volatile] = ACTIONS(2600),
    [anon_sym_restrict] = ACTIONS(2600),
    [anon_sym___restrict__] = ACTIONS(2600),
    [anon_sym__Atomic] = ACTIONS(2600),
    [anon_sym__Noreturn] = ACTIONS(2600),
    [anon_sym__Nonnull] = ACTIONS(2600),
    [anon_sym_nullable] = ACTIONS(2600),
    [anon_sym__Complex] = ACTIONS(2600),
    [anon_sym__Nullable] = ACTIONS(2600),
    [anon_sym__Nullable_result] = ACTIONS(2600),
    [anon_sym__Null_unspecified] = ACTIONS(2600),
    [anon_sym___autoreleasing] = ACTIONS(2600),
    [anon_sym___block] = ACTIONS(2600),
    [anon_sym___bridge] = ACTIONS(2600),
    [anon_sym___bridge_retained] = ACTIONS(2600),
    [anon_sym___bridge_transfer] = ACTIONS(2600),
    [anon_sym___complex] = ACTIONS(2600),
    [anon_sym___const] = ACTIONS(2600),
    [anon_sym___imag] = ACTIONS(2600),
    [anon_sym___kindof] = ACTIONS(2600),
    [anon_sym___nonnull] = ACTIONS(2600),
    [anon_sym___nullable] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2600),
    [anon_sym___real] = ACTIONS(2600),
    [anon_sym___strong] = ACTIONS(2600),
    [anon_sym___unsafe_unretained] = ACTIONS(2600),
    [anon_sym___unused] = ACTIONS(2600),
    [anon_sym___weak] = ACTIONS(2600),
    [anon_sym_alignas] = ACTIONS(2600),
    [anon_sym__Alignas] = ACTIONS(2600),
    [sym_primitive_type] = ACTIONS(2600),
    [anon_sym_enum] = ACTIONS(2600),
    [anon_sym_struct] = ACTIONS(2600),
    [anon_sym_union] = ACTIONS(2600),
    [anon_sym_if] = ACTIONS(2600),
    [anon_sym_switch] = ACTIONS(2600),
    [anon_sym_case] = ACTIONS(2600),
    [anon_sym_default] = ACTIONS(2600),
    [anon_sym_while] = ACTIONS(2600),
    [anon_sym_do] = ACTIONS(2600),
    [anon_sym_for] = ACTIONS(2600),
    [anon_sym_in] = ACTIONS(2600),
    [anon_sym_return] = ACTIONS(2600),
    [anon_sym_break] = ACTIONS(2600),
    [anon_sym_continue] = ACTIONS(2600),
    [anon_sym_goto] = ACTIONS(2600),
    [anon_sym_DASH_DASH] = ACTIONS(2602),
    [anon_sym_PLUS_PLUS] = ACTIONS(2602),
    [anon_sym_sizeof] = ACTIONS(2600),
    [anon_sym___alignof__] = ACTIONS(2600),
    [anon_sym___alignof] = ACTIONS(2600),
    [anon_sym__alignof] = ACTIONS(2600),
    [anon_sym_alignof] = ACTIONS(2600),
    [anon_sym__Alignof] = ACTIONS(2600),
    [anon_sym_offsetof] = ACTIONS(2600),
    [anon_sym__Generic] = ACTIONS(2600),
    [anon_sym_asm] = ACTIONS(2600),
    [anon_sym___asm__] = ACTIONS(2600),
    [anon_sym___asm] = ACTIONS(2600),
    [sym_number_literal] = ACTIONS(2602),
    [anon_sym_L_SQUOTE] = ACTIONS(2602),
    [anon_sym_u_SQUOTE] = ACTIONS(2602),
    [anon_sym_U_SQUOTE] = ACTIONS(2602),
    [anon_sym_u8_SQUOTE] = ACTIONS(2602),
    [anon_sym_SQUOTE] = ACTIONS(2602),
    [anon_sym_AT] = ACTIONS(2600),
    [anon_sym_DQUOTE] = ACTIONS(2602),
    [anon_sym_L_DQUOTE] = ACTIONS(2602),
    [anon_sym_u_DQUOTE] = ACTIONS(2602),
    [anon_sym_U_DQUOTE] = ACTIONS(2602),
    [anon_sym_u8_DQUOTE] = ACTIONS(2602),
    [sym_true] = ACTIONS(2600),
    [sym_false] = ACTIONS(2600),
    [anon_sym_NULL] = ACTIONS(2600),
    [anon_sym_nullptr] = ACTIONS(2600),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2600),
    [anon_sym___typeof] = ACTIONS(2600),
    [anon_sym_typeof] = ACTIONS(2600),
    [anon_sym_ATimport] = ACTIONS(2602),
    [aux_sym_preproc_undef_token1] = ACTIONS(2600),
    [anon_sym_POUND] = ACTIONS(2600),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2600),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2600),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2600),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_API_AVAILABLE] = ACTIONS(2600),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_API_DEPRECATED] = ACTIONS(2600),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym___deprecated_msg] = ACTIONS(2600),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2600),
    [anon_sym_ATprotocol] = ACTIONS(2602),
    [anon_sym_ATinterface] = ACTIONS(2602),
    [anon_sym_ATimplementation] = ACTIONS(2602),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2602),
    [anon_sym_ATselector] = ACTIONS(2602),
    [anon_sym_ATavailable] = ACTIONS(2602),
    [anon_sym___builtin_available] = ACTIONS(2600),
    [anon_sym_va_arg] = ACTIONS(2600),
    [anon_sym_ATencode] = ACTIONS(2602),
    [anon_sym_BOOL] = ACTIONS(2600),
    [anon_sym_IMP] = ACTIONS(2600),
    [anon_sym_SEL] = ACTIONS(2600),
    [anon_sym_Class] = ACTIONS(2600),
    [anon_sym_id] = ACTIONS(2600),
  },
  [STATE(1059)] = {
    [ts_builtin_sym_end] = ACTIONS(2770),
    [sym_identifier] = ACTIONS(2768),
    [aux_sym_preproc_include_token1] = ACTIONS(2768),
    [aux_sym_preproc_include_token2] = ACTIONS(2768),
    [aux_sym_preproc_def_token1] = ACTIONS(2768),
    [aux_sym_preproc_if_token1] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2768),
    [sym_preproc_directive] = ACTIONS(2768),
    [anon_sym_LPAREN2] = ACTIONS(2770),
    [anon_sym_BANG] = ACTIONS(2770),
    [anon_sym_TILDE] = ACTIONS(2770),
    [anon_sym_DASH] = ACTIONS(2768),
    [anon_sym_PLUS] = ACTIONS(2768),
    [anon_sym_STAR] = ACTIONS(2770),
    [anon_sym_CARET] = ACTIONS(2770),
    [anon_sym_AMP] = ACTIONS(2770),
    [anon_sym_SEMI] = ACTIONS(2770),
    [anon_sym___extension__] = ACTIONS(2768),
    [anon_sym_typedef] = ACTIONS(2768),
    [anon_sym_extern] = ACTIONS(2768),
    [anon_sym___attribute__] = ACTIONS(2768),
    [anon_sym___attribute] = ACTIONS(2768),
    [anon_sym_noreturn] = ACTIONS(2768),
    [anon_sym_LBRACK] = ACTIONS(2770),
    [anon_sym___declspec] = ACTIONS(2768),
    [anon_sym___cdecl] = ACTIONS(2768),
    [anon_sym___clrcall] = ACTIONS(2768),
    [anon_sym___stdcall] = ACTIONS(2768),
    [anon_sym___fastcall] = ACTIONS(2768),
    [anon_sym___thiscall] = ACTIONS(2768),
    [anon_sym___vectorcall] = ACTIONS(2768),
    [anon_sym_LBRACE] = ACTIONS(2770),
    [anon_sym_signed] = ACTIONS(2768),
    [anon_sym_unsigned] = ACTIONS(2768),
    [anon_sym_long] = ACTIONS(2768),
    [anon_sym_short] = ACTIONS(2768),
    [anon_sym_static] = ACTIONS(2768),
    [anon_sym_ATautoreleasepool] = ACTIONS(2770),
    [anon_sym_auto] = ACTIONS(2768),
    [anon_sym_register] = ACTIONS(2768),
    [anon_sym_inline] = ACTIONS(2768),
    [anon_sym___inline] = ACTIONS(2768),
    [anon_sym___inline__] = ACTIONS(2768),
    [anon_sym___forceinline] = ACTIONS(2768),
    [anon_sym_thread_local] = ACTIONS(2768),
    [anon_sym___thread] = ACTIONS(2768),
    [anon_sym_CG_EXTERN] = ACTIONS(2768),
    [anon_sym_CG_INLINE] = ACTIONS(2768),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2768),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2768),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2768),
    [anon_sym_IBOutlet] = ACTIONS(2768),
    [anon_sym_IBInspectable] = ACTIONS(2768),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2768),
    [anon_sym_NS_INLINE] = ACTIONS(2768),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2768),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2768),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2768),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2768),
    [anon_sym_const] = ACTIONS(2768),
    [anon_sym_constexpr] = ACTIONS(2768),
    [anon_sym_volatile] = ACTIONS(2768),
    [anon_sym_restrict] = ACTIONS(2768),
    [anon_sym___restrict__] = ACTIONS(2768),
    [anon_sym__Atomic] = ACTIONS(2768),
    [anon_sym__Noreturn] = ACTIONS(2768),
    [anon_sym__Nonnull] = ACTIONS(2768),
    [anon_sym_nullable] = ACTIONS(2768),
    [anon_sym__Complex] = ACTIONS(2768),
    [anon_sym__Nullable] = ACTIONS(2768),
    [anon_sym__Nullable_result] = ACTIONS(2768),
    [anon_sym__Null_unspecified] = ACTIONS(2768),
    [anon_sym___autoreleasing] = ACTIONS(2768),
    [anon_sym___block] = ACTIONS(2768),
    [anon_sym___bridge] = ACTIONS(2768),
    [anon_sym___bridge_retained] = ACTIONS(2768),
    [anon_sym___bridge_transfer] = ACTIONS(2768),
    [anon_sym___complex] = ACTIONS(2768),
    [anon_sym___const] = ACTIONS(2768),
    [anon_sym___imag] = ACTIONS(2768),
    [anon_sym___kindof] = ACTIONS(2768),
    [anon_sym___nonnull] = ACTIONS(2768),
    [anon_sym___nullable] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2768),
    [anon_sym___real] = ACTIONS(2768),
    [anon_sym___strong] = ACTIONS(2768),
    [anon_sym___unsafe_unretained] = ACTIONS(2768),
    [anon_sym___unused] = ACTIONS(2768),
    [anon_sym___weak] = ACTIONS(2768),
    [anon_sym_alignas] = ACTIONS(2768),
    [anon_sym__Alignas] = ACTIONS(2768),
    [sym_primitive_type] = ACTIONS(2768),
    [anon_sym_enum] = ACTIONS(2768),
    [anon_sym_struct] = ACTIONS(2768),
    [anon_sym_union] = ACTIONS(2768),
    [anon_sym_if] = ACTIONS(2768),
    [anon_sym_switch] = ACTIONS(2768),
    [anon_sym_case] = ACTIONS(2768),
    [anon_sym_default] = ACTIONS(2768),
    [anon_sym_while] = ACTIONS(2768),
    [anon_sym_do] = ACTIONS(2768),
    [anon_sym_for] = ACTIONS(2768),
    [anon_sym_in] = ACTIONS(2768),
    [anon_sym_return] = ACTIONS(2768),
    [anon_sym_break] = ACTIONS(2768),
    [anon_sym_continue] = ACTIONS(2768),
    [anon_sym_goto] = ACTIONS(2768),
    [anon_sym_DASH_DASH] = ACTIONS(2770),
    [anon_sym_PLUS_PLUS] = ACTIONS(2770),
    [anon_sym_sizeof] = ACTIONS(2768),
    [anon_sym___alignof__] = ACTIONS(2768),
    [anon_sym___alignof] = ACTIONS(2768),
    [anon_sym__alignof] = ACTIONS(2768),
    [anon_sym_alignof] = ACTIONS(2768),
    [anon_sym__Alignof] = ACTIONS(2768),
    [anon_sym_offsetof] = ACTIONS(2768),
    [anon_sym__Generic] = ACTIONS(2768),
    [anon_sym_asm] = ACTIONS(2768),
    [anon_sym___asm__] = ACTIONS(2768),
    [anon_sym___asm] = ACTIONS(2768),
    [sym_number_literal] = ACTIONS(2770),
    [anon_sym_L_SQUOTE] = ACTIONS(2770),
    [anon_sym_u_SQUOTE] = ACTIONS(2770),
    [anon_sym_U_SQUOTE] = ACTIONS(2770),
    [anon_sym_u8_SQUOTE] = ACTIONS(2770),
    [anon_sym_SQUOTE] = ACTIONS(2770),
    [anon_sym_AT] = ACTIONS(2768),
    [anon_sym_DQUOTE] = ACTIONS(2770),
    [anon_sym_L_DQUOTE] = ACTIONS(2770),
    [anon_sym_u_DQUOTE] = ACTIONS(2770),
    [anon_sym_U_DQUOTE] = ACTIONS(2770),
    [anon_sym_u8_DQUOTE] = ACTIONS(2770),
    [sym_true] = ACTIONS(2768),
    [sym_false] = ACTIONS(2768),
    [anon_sym_NULL] = ACTIONS(2768),
    [anon_sym_nullptr] = ACTIONS(2768),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2768),
    [anon_sym___typeof] = ACTIONS(2768),
    [anon_sym_typeof] = ACTIONS(2768),
    [anon_sym_ATimport] = ACTIONS(2770),
    [aux_sym_preproc_undef_token1] = ACTIONS(2768),
    [anon_sym_POUND] = ACTIONS(2768),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2768),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2768),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2768),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2768),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2768),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2768),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2768),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_API_AVAILABLE] = ACTIONS(2768),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_API_DEPRECATED] = ACTIONS(2768),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2768),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2768),
    [anon_sym___deprecated_msg] = ACTIONS(2768),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2768),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2768),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2768),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2768),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2768),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2768),
    [anon_sym_ATprotocol] = ACTIONS(2770),
    [anon_sym_ATinterface] = ACTIONS(2770),
    [anon_sym_ATimplementation] = ACTIONS(2770),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2770),
    [anon_sym_ATselector] = ACTIONS(2770),
    [anon_sym_ATavailable] = ACTIONS(2770),
    [anon_sym___builtin_available] = ACTIONS(2768),
    [anon_sym_va_arg] = ACTIONS(2768),
    [anon_sym_ATencode] = ACTIONS(2770),
    [anon_sym_BOOL] = ACTIONS(2768),
    [anon_sym_IMP] = ACTIONS(2768),
    [anon_sym_SEL] = ACTIONS(2768),
    [anon_sym_Class] = ACTIONS(2768),
    [anon_sym_id] = ACTIONS(2768),
  },
  [STATE(1060)] = {
    [ts_builtin_sym_end] = ACTIONS(2774),
    [sym_identifier] = ACTIONS(2772),
    [aux_sym_preproc_include_token1] = ACTIONS(2772),
    [aux_sym_preproc_include_token2] = ACTIONS(2772),
    [aux_sym_preproc_def_token1] = ACTIONS(2772),
    [aux_sym_preproc_if_token1] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2772),
    [sym_preproc_directive] = ACTIONS(2772),
    [anon_sym_LPAREN2] = ACTIONS(2774),
    [anon_sym_BANG] = ACTIONS(2774),
    [anon_sym_TILDE] = ACTIONS(2774),
    [anon_sym_DASH] = ACTIONS(2772),
    [anon_sym_PLUS] = ACTIONS(2772),
    [anon_sym_STAR] = ACTIONS(2774),
    [anon_sym_CARET] = ACTIONS(2774),
    [anon_sym_AMP] = ACTIONS(2774),
    [anon_sym_SEMI] = ACTIONS(2774),
    [anon_sym___extension__] = ACTIONS(2772),
    [anon_sym_typedef] = ACTIONS(2772),
    [anon_sym_extern] = ACTIONS(2772),
    [anon_sym___attribute__] = ACTIONS(2772),
    [anon_sym___attribute] = ACTIONS(2772),
    [anon_sym_noreturn] = ACTIONS(2772),
    [anon_sym_LBRACK] = ACTIONS(2774),
    [anon_sym___declspec] = ACTIONS(2772),
    [anon_sym___cdecl] = ACTIONS(2772),
    [anon_sym___clrcall] = ACTIONS(2772),
    [anon_sym___stdcall] = ACTIONS(2772),
    [anon_sym___fastcall] = ACTIONS(2772),
    [anon_sym___thiscall] = ACTIONS(2772),
    [anon_sym___vectorcall] = ACTIONS(2772),
    [anon_sym_LBRACE] = ACTIONS(2774),
    [anon_sym_signed] = ACTIONS(2772),
    [anon_sym_unsigned] = ACTIONS(2772),
    [anon_sym_long] = ACTIONS(2772),
    [anon_sym_short] = ACTIONS(2772),
    [anon_sym_static] = ACTIONS(2772),
    [anon_sym_ATautoreleasepool] = ACTIONS(2774),
    [anon_sym_auto] = ACTIONS(2772),
    [anon_sym_register] = ACTIONS(2772),
    [anon_sym_inline] = ACTIONS(2772),
    [anon_sym___inline] = ACTIONS(2772),
    [anon_sym___inline__] = ACTIONS(2772),
    [anon_sym___forceinline] = ACTIONS(2772),
    [anon_sym_thread_local] = ACTIONS(2772),
    [anon_sym___thread] = ACTIONS(2772),
    [anon_sym_CG_EXTERN] = ACTIONS(2772),
    [anon_sym_CG_INLINE] = ACTIONS(2772),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2772),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2772),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2772),
    [anon_sym_IBOutlet] = ACTIONS(2772),
    [anon_sym_IBInspectable] = ACTIONS(2772),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2772),
    [anon_sym_NS_INLINE] = ACTIONS(2772),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2772),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2772),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2772),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2772),
    [anon_sym_const] = ACTIONS(2772),
    [anon_sym_constexpr] = ACTIONS(2772),
    [anon_sym_volatile] = ACTIONS(2772),
    [anon_sym_restrict] = ACTIONS(2772),
    [anon_sym___restrict__] = ACTIONS(2772),
    [anon_sym__Atomic] = ACTIONS(2772),
    [anon_sym__Noreturn] = ACTIONS(2772),
    [anon_sym__Nonnull] = ACTIONS(2772),
    [anon_sym_nullable] = ACTIONS(2772),
    [anon_sym__Complex] = ACTIONS(2772),
    [anon_sym__Nullable] = ACTIONS(2772),
    [anon_sym__Nullable_result] = ACTIONS(2772),
    [anon_sym__Null_unspecified] = ACTIONS(2772),
    [anon_sym___autoreleasing] = ACTIONS(2772),
    [anon_sym___block] = ACTIONS(2772),
    [anon_sym___bridge] = ACTIONS(2772),
    [anon_sym___bridge_retained] = ACTIONS(2772),
    [anon_sym___bridge_transfer] = ACTIONS(2772),
    [anon_sym___complex] = ACTIONS(2772),
    [anon_sym___const] = ACTIONS(2772),
    [anon_sym___imag] = ACTIONS(2772),
    [anon_sym___kindof] = ACTIONS(2772),
    [anon_sym___nonnull] = ACTIONS(2772),
    [anon_sym___nullable] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2772),
    [anon_sym___real] = ACTIONS(2772),
    [anon_sym___strong] = ACTIONS(2772),
    [anon_sym___unsafe_unretained] = ACTIONS(2772),
    [anon_sym___unused] = ACTIONS(2772),
    [anon_sym___weak] = ACTIONS(2772),
    [anon_sym_alignas] = ACTIONS(2772),
    [anon_sym__Alignas] = ACTIONS(2772),
    [sym_primitive_type] = ACTIONS(2772),
    [anon_sym_enum] = ACTIONS(2772),
    [anon_sym_struct] = ACTIONS(2772),
    [anon_sym_union] = ACTIONS(2772),
    [anon_sym_if] = ACTIONS(2772),
    [anon_sym_switch] = ACTIONS(2772),
    [anon_sym_case] = ACTIONS(2772),
    [anon_sym_default] = ACTIONS(2772),
    [anon_sym_while] = ACTIONS(2772),
    [anon_sym_do] = ACTIONS(2772),
    [anon_sym_for] = ACTIONS(2772),
    [anon_sym_in] = ACTIONS(2772),
    [anon_sym_return] = ACTIONS(2772),
    [anon_sym_break] = ACTIONS(2772),
    [anon_sym_continue] = ACTIONS(2772),
    [anon_sym_goto] = ACTIONS(2772),
    [anon_sym_DASH_DASH] = ACTIONS(2774),
    [anon_sym_PLUS_PLUS] = ACTIONS(2774),
    [anon_sym_sizeof] = ACTIONS(2772),
    [anon_sym___alignof__] = ACTIONS(2772),
    [anon_sym___alignof] = ACTIONS(2772),
    [anon_sym__alignof] = ACTIONS(2772),
    [anon_sym_alignof] = ACTIONS(2772),
    [anon_sym__Alignof] = ACTIONS(2772),
    [anon_sym_offsetof] = ACTIONS(2772),
    [anon_sym__Generic] = ACTIONS(2772),
    [anon_sym_asm] = ACTIONS(2772),
    [anon_sym___asm__] = ACTIONS(2772),
    [anon_sym___asm] = ACTIONS(2772),
    [sym_number_literal] = ACTIONS(2774),
    [anon_sym_L_SQUOTE] = ACTIONS(2774),
    [anon_sym_u_SQUOTE] = ACTIONS(2774),
    [anon_sym_U_SQUOTE] = ACTIONS(2774),
    [anon_sym_u8_SQUOTE] = ACTIONS(2774),
    [anon_sym_SQUOTE] = ACTIONS(2774),
    [anon_sym_AT] = ACTIONS(2772),
    [anon_sym_DQUOTE] = ACTIONS(2774),
    [anon_sym_L_DQUOTE] = ACTIONS(2774),
    [anon_sym_u_DQUOTE] = ACTIONS(2774),
    [anon_sym_U_DQUOTE] = ACTIONS(2774),
    [anon_sym_u8_DQUOTE] = ACTIONS(2774),
    [sym_true] = ACTIONS(2772),
    [sym_false] = ACTIONS(2772),
    [anon_sym_NULL] = ACTIONS(2772),
    [anon_sym_nullptr] = ACTIONS(2772),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2772),
    [anon_sym___typeof] = ACTIONS(2772),
    [anon_sym_typeof] = ACTIONS(2772),
    [anon_sym_ATimport] = ACTIONS(2774),
    [aux_sym_preproc_undef_token1] = ACTIONS(2772),
    [anon_sym_POUND] = ACTIONS(2772),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2772),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2772),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2772),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2772),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2772),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2772),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2772),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_API_AVAILABLE] = ACTIONS(2772),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_API_DEPRECATED] = ACTIONS(2772),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2772),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2772),
    [anon_sym___deprecated_msg] = ACTIONS(2772),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2772),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2772),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2772),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2772),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2772),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2772),
    [anon_sym_ATprotocol] = ACTIONS(2774),
    [anon_sym_ATinterface] = ACTIONS(2774),
    [anon_sym_ATimplementation] = ACTIONS(2774),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2774),
    [anon_sym_ATselector] = ACTIONS(2774),
    [anon_sym_ATavailable] = ACTIONS(2774),
    [anon_sym___builtin_available] = ACTIONS(2772),
    [anon_sym_va_arg] = ACTIONS(2772),
    [anon_sym_ATencode] = ACTIONS(2774),
    [anon_sym_BOOL] = ACTIONS(2772),
    [anon_sym_IMP] = ACTIONS(2772),
    [anon_sym_SEL] = ACTIONS(2772),
    [anon_sym_Class] = ACTIONS(2772),
    [anon_sym_id] = ACTIONS(2772),
  },
  [STATE(1061)] = {
    [ts_builtin_sym_end] = ACTIONS(2606),
    [sym_identifier] = ACTIONS(2604),
    [aux_sym_preproc_include_token1] = ACTIONS(2604),
    [aux_sym_preproc_include_token2] = ACTIONS(2604),
    [aux_sym_preproc_def_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2604),
    [sym_preproc_directive] = ACTIONS(2604),
    [anon_sym_LPAREN2] = ACTIONS(2606),
    [anon_sym_BANG] = ACTIONS(2606),
    [anon_sym_TILDE] = ACTIONS(2606),
    [anon_sym_DASH] = ACTIONS(2604),
    [anon_sym_PLUS] = ACTIONS(2604),
    [anon_sym_STAR] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(2606),
    [anon_sym_AMP] = ACTIONS(2606),
    [anon_sym_SEMI] = ACTIONS(2606),
    [anon_sym___extension__] = ACTIONS(2604),
    [anon_sym_typedef] = ACTIONS(2604),
    [anon_sym_extern] = ACTIONS(2604),
    [anon_sym___attribute__] = ACTIONS(2604),
    [anon_sym___attribute] = ACTIONS(2604),
    [anon_sym_noreturn] = ACTIONS(2604),
    [anon_sym_LBRACK] = ACTIONS(2606),
    [anon_sym___declspec] = ACTIONS(2604),
    [anon_sym___cdecl] = ACTIONS(2604),
    [anon_sym___clrcall] = ACTIONS(2604),
    [anon_sym___stdcall] = ACTIONS(2604),
    [anon_sym___fastcall] = ACTIONS(2604),
    [anon_sym___thiscall] = ACTIONS(2604),
    [anon_sym___vectorcall] = ACTIONS(2604),
    [anon_sym_LBRACE] = ACTIONS(2606),
    [anon_sym_signed] = ACTIONS(2604),
    [anon_sym_unsigned] = ACTIONS(2604),
    [anon_sym_long] = ACTIONS(2604),
    [anon_sym_short] = ACTIONS(2604),
    [anon_sym_static] = ACTIONS(2604),
    [anon_sym_ATautoreleasepool] = ACTIONS(2606),
    [anon_sym_auto] = ACTIONS(2604),
    [anon_sym_register] = ACTIONS(2604),
    [anon_sym_inline] = ACTIONS(2604),
    [anon_sym___inline] = ACTIONS(2604),
    [anon_sym___inline__] = ACTIONS(2604),
    [anon_sym___forceinline] = ACTIONS(2604),
    [anon_sym_thread_local] = ACTIONS(2604),
    [anon_sym___thread] = ACTIONS(2604),
    [anon_sym_CG_EXTERN] = ACTIONS(2604),
    [anon_sym_CG_INLINE] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2604),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2604),
    [anon_sym_IBOutlet] = ACTIONS(2604),
    [anon_sym_IBInspectable] = ACTIONS(2604),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2604),
    [anon_sym_NS_INLINE] = ACTIONS(2604),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2604),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2604),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2604),
    [anon_sym_const] = ACTIONS(2604),
    [anon_sym_constexpr] = ACTIONS(2604),
    [anon_sym_volatile] = ACTIONS(2604),
    [anon_sym_restrict] = ACTIONS(2604),
    [anon_sym___restrict__] = ACTIONS(2604),
    [anon_sym__Atomic] = ACTIONS(2604),
    [anon_sym__Noreturn] = ACTIONS(2604),
    [anon_sym__Nonnull] = ACTIONS(2604),
    [anon_sym_nullable] = ACTIONS(2604),
    [anon_sym__Complex] = ACTIONS(2604),
    [anon_sym__Nullable] = ACTIONS(2604),
    [anon_sym__Nullable_result] = ACTIONS(2604),
    [anon_sym__Null_unspecified] = ACTIONS(2604),
    [anon_sym___autoreleasing] = ACTIONS(2604),
    [anon_sym___block] = ACTIONS(2604),
    [anon_sym___bridge] = ACTIONS(2604),
    [anon_sym___bridge_retained] = ACTIONS(2604),
    [anon_sym___bridge_transfer] = ACTIONS(2604),
    [anon_sym___complex] = ACTIONS(2604),
    [anon_sym___const] = ACTIONS(2604),
    [anon_sym___imag] = ACTIONS(2604),
    [anon_sym___kindof] = ACTIONS(2604),
    [anon_sym___nonnull] = ACTIONS(2604),
    [anon_sym___nullable] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2604),
    [anon_sym___real] = ACTIONS(2604),
    [anon_sym___strong] = ACTIONS(2604),
    [anon_sym___unsafe_unretained] = ACTIONS(2604),
    [anon_sym___unused] = ACTIONS(2604),
    [anon_sym___weak] = ACTIONS(2604),
    [anon_sym_alignas] = ACTIONS(2604),
    [anon_sym__Alignas] = ACTIONS(2604),
    [sym_primitive_type] = ACTIONS(2604),
    [anon_sym_enum] = ACTIONS(2604),
    [anon_sym_struct] = ACTIONS(2604),
    [anon_sym_union] = ACTIONS(2604),
    [anon_sym_if] = ACTIONS(2604),
    [anon_sym_switch] = ACTIONS(2604),
    [anon_sym_case] = ACTIONS(2604),
    [anon_sym_default] = ACTIONS(2604),
    [anon_sym_while] = ACTIONS(2604),
    [anon_sym_do] = ACTIONS(2604),
    [anon_sym_for] = ACTIONS(2604),
    [anon_sym_in] = ACTIONS(2604),
    [anon_sym_return] = ACTIONS(2604),
    [anon_sym_break] = ACTIONS(2604),
    [anon_sym_continue] = ACTIONS(2604),
    [anon_sym_goto] = ACTIONS(2604),
    [anon_sym_DASH_DASH] = ACTIONS(2606),
    [anon_sym_PLUS_PLUS] = ACTIONS(2606),
    [anon_sym_sizeof] = ACTIONS(2604),
    [anon_sym___alignof__] = ACTIONS(2604),
    [anon_sym___alignof] = ACTIONS(2604),
    [anon_sym__alignof] = ACTIONS(2604),
    [anon_sym_alignof] = ACTIONS(2604),
    [anon_sym__Alignof] = ACTIONS(2604),
    [anon_sym_offsetof] = ACTIONS(2604),
    [anon_sym__Generic] = ACTIONS(2604),
    [anon_sym_asm] = ACTIONS(2604),
    [anon_sym___asm__] = ACTIONS(2604),
    [anon_sym___asm] = ACTIONS(2604),
    [sym_number_literal] = ACTIONS(2606),
    [anon_sym_L_SQUOTE] = ACTIONS(2606),
    [anon_sym_u_SQUOTE] = ACTIONS(2606),
    [anon_sym_U_SQUOTE] = ACTIONS(2606),
    [anon_sym_u8_SQUOTE] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2606),
    [anon_sym_AT] = ACTIONS(2604),
    [anon_sym_DQUOTE] = ACTIONS(2606),
    [anon_sym_L_DQUOTE] = ACTIONS(2606),
    [anon_sym_u_DQUOTE] = ACTIONS(2606),
    [anon_sym_U_DQUOTE] = ACTIONS(2606),
    [anon_sym_u8_DQUOTE] = ACTIONS(2606),
    [sym_true] = ACTIONS(2604),
    [sym_false] = ACTIONS(2604),
    [anon_sym_NULL] = ACTIONS(2604),
    [anon_sym_nullptr] = ACTIONS(2604),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2604),
    [anon_sym___typeof] = ACTIONS(2604),
    [anon_sym_typeof] = ACTIONS(2604),
    [anon_sym_ATimport] = ACTIONS(2606),
    [aux_sym_preproc_undef_token1] = ACTIONS(2604),
    [anon_sym_POUND] = ACTIONS(2604),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2604),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2604),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2604),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_API_AVAILABLE] = ACTIONS(2604),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_API_DEPRECATED] = ACTIONS(2604),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym___deprecated_msg] = ACTIONS(2604),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2604),
    [anon_sym_ATprotocol] = ACTIONS(2606),
    [anon_sym_ATinterface] = ACTIONS(2606),
    [anon_sym_ATimplementation] = ACTIONS(2606),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2606),
    [anon_sym_ATselector] = ACTIONS(2606),
    [anon_sym_ATavailable] = ACTIONS(2606),
    [anon_sym___builtin_available] = ACTIONS(2604),
    [anon_sym_va_arg] = ACTIONS(2604),
    [anon_sym_ATencode] = ACTIONS(2606),
    [anon_sym_BOOL] = ACTIONS(2604),
    [anon_sym_IMP] = ACTIONS(2604),
    [anon_sym_SEL] = ACTIONS(2604),
    [anon_sym_Class] = ACTIONS(2604),
    [anon_sym_id] = ACTIONS(2604),
  },
  [STATE(1062)] = {
    [ts_builtin_sym_end] = ACTIONS(2430),
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(1063)] = {
    [ts_builtin_sym_end] = ACTIONS(2430),
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(1064)] = {
    [ts_builtin_sym_end] = ACTIONS(2430),
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_include_token1] = ACTIONS(2428),
    [aux_sym_preproc_include_token2] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_case] = ACTIONS(2428),
    [anon_sym_default] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_ATimport] = ACTIONS(2430),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_POUND] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATprotocol] = ACTIONS(2430),
    [anon_sym_ATinterface] = ACTIONS(2430),
    [anon_sym_ATimplementation] = ACTIONS(2430),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(1065)] = {
    [ts_builtin_sym_end] = ACTIONS(2778),
    [sym_identifier] = ACTIONS(2776),
    [aux_sym_preproc_include_token1] = ACTIONS(2776),
    [aux_sym_preproc_include_token2] = ACTIONS(2776),
    [aux_sym_preproc_def_token1] = ACTIONS(2776),
    [aux_sym_preproc_if_token1] = ACTIONS(2776),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2776),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2776),
    [sym_preproc_directive] = ACTIONS(2776),
    [anon_sym_LPAREN2] = ACTIONS(2778),
    [anon_sym_BANG] = ACTIONS(2778),
    [anon_sym_TILDE] = ACTIONS(2778),
    [anon_sym_DASH] = ACTIONS(2776),
    [anon_sym_PLUS] = ACTIONS(2776),
    [anon_sym_STAR] = ACTIONS(2778),
    [anon_sym_CARET] = ACTIONS(2778),
    [anon_sym_AMP] = ACTIONS(2778),
    [anon_sym_SEMI] = ACTIONS(2778),
    [anon_sym___extension__] = ACTIONS(2776),
    [anon_sym_typedef] = ACTIONS(2776),
    [anon_sym_extern] = ACTIONS(2776),
    [anon_sym___attribute__] = ACTIONS(2776),
    [anon_sym___attribute] = ACTIONS(2776),
    [anon_sym_noreturn] = ACTIONS(2776),
    [anon_sym_LBRACK] = ACTIONS(2778),
    [anon_sym___declspec] = ACTIONS(2776),
    [anon_sym___cdecl] = ACTIONS(2776),
    [anon_sym___clrcall] = ACTIONS(2776),
    [anon_sym___stdcall] = ACTIONS(2776),
    [anon_sym___fastcall] = ACTIONS(2776),
    [anon_sym___thiscall] = ACTIONS(2776),
    [anon_sym___vectorcall] = ACTIONS(2776),
    [anon_sym_LBRACE] = ACTIONS(2778),
    [anon_sym_signed] = ACTIONS(2776),
    [anon_sym_unsigned] = ACTIONS(2776),
    [anon_sym_long] = ACTIONS(2776),
    [anon_sym_short] = ACTIONS(2776),
    [anon_sym_static] = ACTIONS(2776),
    [anon_sym_ATautoreleasepool] = ACTIONS(2778),
    [anon_sym_auto] = ACTIONS(2776),
    [anon_sym_register] = ACTIONS(2776),
    [anon_sym_inline] = ACTIONS(2776),
    [anon_sym___inline] = ACTIONS(2776),
    [anon_sym___inline__] = ACTIONS(2776),
    [anon_sym___forceinline] = ACTIONS(2776),
    [anon_sym_thread_local] = ACTIONS(2776),
    [anon_sym___thread] = ACTIONS(2776),
    [anon_sym_CG_EXTERN] = ACTIONS(2776),
    [anon_sym_CG_INLINE] = ACTIONS(2776),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2776),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2776),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2776),
    [anon_sym_IBOutlet] = ACTIONS(2776),
    [anon_sym_IBInspectable] = ACTIONS(2776),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2776),
    [anon_sym_NS_INLINE] = ACTIONS(2776),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2776),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2776),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2776),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2776),
    [anon_sym_const] = ACTIONS(2776),
    [anon_sym_constexpr] = ACTIONS(2776),
    [anon_sym_volatile] = ACTIONS(2776),
    [anon_sym_restrict] = ACTIONS(2776),
    [anon_sym___restrict__] = ACTIONS(2776),
    [anon_sym__Atomic] = ACTIONS(2776),
    [anon_sym__Noreturn] = ACTIONS(2776),
    [anon_sym__Nonnull] = ACTIONS(2776),
    [anon_sym_nullable] = ACTIONS(2776),
    [anon_sym__Complex] = ACTIONS(2776),
    [anon_sym__Nullable] = ACTIONS(2776),
    [anon_sym__Nullable_result] = ACTIONS(2776),
    [anon_sym__Null_unspecified] = ACTIONS(2776),
    [anon_sym___autoreleasing] = ACTIONS(2776),
    [anon_sym___block] = ACTIONS(2776),
    [anon_sym___bridge] = ACTIONS(2776),
    [anon_sym___bridge_retained] = ACTIONS(2776),
    [anon_sym___bridge_transfer] = ACTIONS(2776),
    [anon_sym___complex] = ACTIONS(2776),
    [anon_sym___const] = ACTIONS(2776),
    [anon_sym___imag] = ACTIONS(2776),
    [anon_sym___kindof] = ACTIONS(2776),
    [anon_sym___nonnull] = ACTIONS(2776),
    [anon_sym___nullable] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2776),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2776),
    [anon_sym___real] = ACTIONS(2776),
    [anon_sym___strong] = ACTIONS(2776),
    [anon_sym___unsafe_unretained] = ACTIONS(2776),
    [anon_sym___unused] = ACTIONS(2776),
    [anon_sym___weak] = ACTIONS(2776),
    [anon_sym_alignas] = ACTIONS(2776),
    [anon_sym__Alignas] = ACTIONS(2776),
    [sym_primitive_type] = ACTIONS(2776),
    [anon_sym_enum] = ACTIONS(2776),
    [anon_sym_struct] = ACTIONS(2776),
    [anon_sym_union] = ACTIONS(2776),
    [anon_sym_if] = ACTIONS(2776),
    [anon_sym_switch] = ACTIONS(2776),
    [anon_sym_case] = ACTIONS(2776),
    [anon_sym_default] = ACTIONS(2776),
    [anon_sym_while] = ACTIONS(2776),
    [anon_sym_do] = ACTIONS(2776),
    [anon_sym_for] = ACTIONS(2776),
    [anon_sym_in] = ACTIONS(2776),
    [anon_sym_return] = ACTIONS(2776),
    [anon_sym_break] = ACTIONS(2776),
    [anon_sym_continue] = ACTIONS(2776),
    [anon_sym_goto] = ACTIONS(2776),
    [anon_sym_DASH_DASH] = ACTIONS(2778),
    [anon_sym_PLUS_PLUS] = ACTIONS(2778),
    [anon_sym_sizeof] = ACTIONS(2776),
    [anon_sym___alignof__] = ACTIONS(2776),
    [anon_sym___alignof] = ACTIONS(2776),
    [anon_sym__alignof] = ACTIONS(2776),
    [anon_sym_alignof] = ACTIONS(2776),
    [anon_sym__Alignof] = ACTIONS(2776),
    [anon_sym_offsetof] = ACTIONS(2776),
    [anon_sym__Generic] = ACTIONS(2776),
    [anon_sym_asm] = ACTIONS(2776),
    [anon_sym___asm__] = ACTIONS(2776),
    [anon_sym___asm] = ACTIONS(2776),
    [sym_number_literal] = ACTIONS(2778),
    [anon_sym_L_SQUOTE] = ACTIONS(2778),
    [anon_sym_u_SQUOTE] = ACTIONS(2778),
    [anon_sym_U_SQUOTE] = ACTIONS(2778),
    [anon_sym_u8_SQUOTE] = ACTIONS(2778),
    [anon_sym_SQUOTE] = ACTIONS(2778),
    [anon_sym_AT] = ACTIONS(2776),
    [anon_sym_DQUOTE] = ACTIONS(2778),
    [anon_sym_L_DQUOTE] = ACTIONS(2778),
    [anon_sym_u_DQUOTE] = ACTIONS(2778),
    [anon_sym_U_DQUOTE] = ACTIONS(2778),
    [anon_sym_u8_DQUOTE] = ACTIONS(2778),
    [sym_true] = ACTIONS(2776),
    [sym_false] = ACTIONS(2776),
    [anon_sym_NULL] = ACTIONS(2776),
    [anon_sym_nullptr] = ACTIONS(2776),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2776),
    [anon_sym___typeof] = ACTIONS(2776),
    [anon_sym_typeof] = ACTIONS(2776),
    [anon_sym_ATimport] = ACTIONS(2778),
    [aux_sym_preproc_undef_token1] = ACTIONS(2776),
    [anon_sym_POUND] = ACTIONS(2776),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2776),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2776),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_API_AVAILABLE] = ACTIONS(2776),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_API_DEPRECATED] = ACTIONS(2776),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2776),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2776),
    [anon_sym___deprecated_msg] = ACTIONS(2776),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2776),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2776),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2776),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2776),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2776),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2776),
    [anon_sym_ATprotocol] = ACTIONS(2778),
    [anon_sym_ATinterface] = ACTIONS(2778),
    [anon_sym_ATimplementation] = ACTIONS(2778),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2778),
    [anon_sym_ATselector] = ACTIONS(2778),
    [anon_sym_ATavailable] = ACTIONS(2778),
    [anon_sym___builtin_available] = ACTIONS(2776),
    [anon_sym_va_arg] = ACTIONS(2776),
    [anon_sym_ATencode] = ACTIONS(2778),
    [anon_sym_BOOL] = ACTIONS(2776),
    [anon_sym_IMP] = ACTIONS(2776),
    [anon_sym_SEL] = ACTIONS(2776),
    [anon_sym_Class] = ACTIONS(2776),
    [anon_sym_id] = ACTIONS(2776),
  },
  [STATE(1066)] = {
    [ts_builtin_sym_end] = ACTIONS(2656),
    [sym_identifier] = ACTIONS(2654),
    [aux_sym_preproc_include_token1] = ACTIONS(2654),
    [aux_sym_preproc_include_token2] = ACTIONS(2654),
    [aux_sym_preproc_def_token1] = ACTIONS(2654),
    [aux_sym_preproc_if_token1] = ACTIONS(2654),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2654),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2654),
    [sym_preproc_directive] = ACTIONS(2654),
    [anon_sym_LPAREN2] = ACTIONS(2656),
    [anon_sym_BANG] = ACTIONS(2656),
    [anon_sym_TILDE] = ACTIONS(2656),
    [anon_sym_DASH] = ACTIONS(2654),
    [anon_sym_PLUS] = ACTIONS(2654),
    [anon_sym_STAR] = ACTIONS(2656),
    [anon_sym_CARET] = ACTIONS(2656),
    [anon_sym_AMP] = ACTIONS(2656),
    [anon_sym_SEMI] = ACTIONS(2656),
    [anon_sym___extension__] = ACTIONS(2654),
    [anon_sym_typedef] = ACTIONS(2654),
    [anon_sym_extern] = ACTIONS(2654),
    [anon_sym___attribute__] = ACTIONS(2654),
    [anon_sym___attribute] = ACTIONS(2654),
    [anon_sym_noreturn] = ACTIONS(2654),
    [anon_sym_LBRACK] = ACTIONS(2656),
    [anon_sym___declspec] = ACTIONS(2654),
    [anon_sym___cdecl] = ACTIONS(2654),
    [anon_sym___clrcall] = ACTIONS(2654),
    [anon_sym___stdcall] = ACTIONS(2654),
    [anon_sym___fastcall] = ACTIONS(2654),
    [anon_sym___thiscall] = ACTIONS(2654),
    [anon_sym___vectorcall] = ACTIONS(2654),
    [anon_sym_LBRACE] = ACTIONS(2656),
    [anon_sym_signed] = ACTIONS(2654),
    [anon_sym_unsigned] = ACTIONS(2654),
    [anon_sym_long] = ACTIONS(2654),
    [anon_sym_short] = ACTIONS(2654),
    [anon_sym_static] = ACTIONS(2654),
    [anon_sym_ATautoreleasepool] = ACTIONS(2656),
    [anon_sym_auto] = ACTIONS(2654),
    [anon_sym_register] = ACTIONS(2654),
    [anon_sym_inline] = ACTIONS(2654),
    [anon_sym___inline] = ACTIONS(2654),
    [anon_sym___inline__] = ACTIONS(2654),
    [anon_sym___forceinline] = ACTIONS(2654),
    [anon_sym_thread_local] = ACTIONS(2654),
    [anon_sym___thread] = ACTIONS(2654),
    [anon_sym_CG_EXTERN] = ACTIONS(2654),
    [anon_sym_CG_INLINE] = ACTIONS(2654),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2654),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2654),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2654),
    [anon_sym_IBOutlet] = ACTIONS(2654),
    [anon_sym_IBInspectable] = ACTIONS(2654),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2654),
    [anon_sym_NS_INLINE] = ACTIONS(2654),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2654),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2654),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2654),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2654),
    [anon_sym_const] = ACTIONS(2654),
    [anon_sym_constexpr] = ACTIONS(2654),
    [anon_sym_volatile] = ACTIONS(2654),
    [anon_sym_restrict] = ACTIONS(2654),
    [anon_sym___restrict__] = ACTIONS(2654),
    [anon_sym__Atomic] = ACTIONS(2654),
    [anon_sym__Noreturn] = ACTIONS(2654),
    [anon_sym__Nonnull] = ACTIONS(2654),
    [anon_sym_nullable] = ACTIONS(2654),
    [anon_sym__Complex] = ACTIONS(2654),
    [anon_sym__Nullable] = ACTIONS(2654),
    [anon_sym__Nullable_result] = ACTIONS(2654),
    [anon_sym__Null_unspecified] = ACTIONS(2654),
    [anon_sym___autoreleasing] = ACTIONS(2654),
    [anon_sym___block] = ACTIONS(2654),
    [anon_sym___bridge] = ACTIONS(2654),
    [anon_sym___bridge_retained] = ACTIONS(2654),
    [anon_sym___bridge_transfer] = ACTIONS(2654),
    [anon_sym___complex] = ACTIONS(2654),
    [anon_sym___const] = ACTIONS(2654),
    [anon_sym___imag] = ACTIONS(2654),
    [anon_sym___kindof] = ACTIONS(2654),
    [anon_sym___nonnull] = ACTIONS(2654),
    [anon_sym___nullable] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2654),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2654),
    [anon_sym___real] = ACTIONS(2654),
    [anon_sym___strong] = ACTIONS(2654),
    [anon_sym___unsafe_unretained] = ACTIONS(2654),
    [anon_sym___unused] = ACTIONS(2654),
    [anon_sym___weak] = ACTIONS(2654),
    [anon_sym_alignas] = ACTIONS(2654),
    [anon_sym__Alignas] = ACTIONS(2654),
    [sym_primitive_type] = ACTIONS(2654),
    [anon_sym_enum] = ACTIONS(2654),
    [anon_sym_struct] = ACTIONS(2654),
    [anon_sym_union] = ACTIONS(2654),
    [anon_sym_if] = ACTIONS(2654),
    [anon_sym_switch] = ACTIONS(2654),
    [anon_sym_case] = ACTIONS(2654),
    [anon_sym_default] = ACTIONS(2654),
    [anon_sym_while] = ACTIONS(2654),
    [anon_sym_do] = ACTIONS(2654),
    [anon_sym_for] = ACTIONS(2654),
    [anon_sym_in] = ACTIONS(2654),
    [anon_sym_return] = ACTIONS(2654),
    [anon_sym_break] = ACTIONS(2654),
    [anon_sym_continue] = ACTIONS(2654),
    [anon_sym_goto] = ACTIONS(2654),
    [anon_sym_DASH_DASH] = ACTIONS(2656),
    [anon_sym_PLUS_PLUS] = ACTIONS(2656),
    [anon_sym_sizeof] = ACTIONS(2654),
    [anon_sym___alignof__] = ACTIONS(2654),
    [anon_sym___alignof] = ACTIONS(2654),
    [anon_sym__alignof] = ACTIONS(2654),
    [anon_sym_alignof] = ACTIONS(2654),
    [anon_sym__Alignof] = ACTIONS(2654),
    [anon_sym_offsetof] = ACTIONS(2654),
    [anon_sym__Generic] = ACTIONS(2654),
    [anon_sym_asm] = ACTIONS(2654),
    [anon_sym___asm__] = ACTIONS(2654),
    [anon_sym___asm] = ACTIONS(2654),
    [sym_number_literal] = ACTIONS(2656),
    [anon_sym_L_SQUOTE] = ACTIONS(2656),
    [anon_sym_u_SQUOTE] = ACTIONS(2656),
    [anon_sym_U_SQUOTE] = ACTIONS(2656),
    [anon_sym_u8_SQUOTE] = ACTIONS(2656),
    [anon_sym_SQUOTE] = ACTIONS(2656),
    [anon_sym_AT] = ACTIONS(2654),
    [anon_sym_DQUOTE] = ACTIONS(2656),
    [anon_sym_L_DQUOTE] = ACTIONS(2656),
    [anon_sym_u_DQUOTE] = ACTIONS(2656),
    [anon_sym_U_DQUOTE] = ACTIONS(2656),
    [anon_sym_u8_DQUOTE] = ACTIONS(2656),
    [sym_true] = ACTIONS(2654),
    [sym_false] = ACTIONS(2654),
    [anon_sym_NULL] = ACTIONS(2654),
    [anon_sym_nullptr] = ACTIONS(2654),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2654),
    [anon_sym___typeof] = ACTIONS(2654),
    [anon_sym_typeof] = ACTIONS(2654),
    [anon_sym_ATimport] = ACTIONS(2656),
    [aux_sym_preproc_undef_token1] = ACTIONS(2654),
    [anon_sym_POUND] = ACTIONS(2654),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2654),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2654),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2654),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2654),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2654),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2654),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2654),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_API_AVAILABLE] = ACTIONS(2654),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_API_DEPRECATED] = ACTIONS(2654),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2654),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2654),
    [anon_sym___deprecated_msg] = ACTIONS(2654),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2654),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2654),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2654),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2654),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2654),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2654),
    [anon_sym_ATprotocol] = ACTIONS(2656),
    [anon_sym_ATinterface] = ACTIONS(2656),
    [anon_sym_ATimplementation] = ACTIONS(2656),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2656),
    [anon_sym_ATselector] = ACTIONS(2656),
    [anon_sym_ATavailable] = ACTIONS(2656),
    [anon_sym___builtin_available] = ACTIONS(2654),
    [anon_sym_va_arg] = ACTIONS(2654),
    [anon_sym_ATencode] = ACTIONS(2656),
    [anon_sym_BOOL] = ACTIONS(2654),
    [anon_sym_IMP] = ACTIONS(2654),
    [anon_sym_SEL] = ACTIONS(2654),
    [anon_sym_Class] = ACTIONS(2654),
    [anon_sym_id] = ACTIONS(2654),
  },
  [STATE(1067)] = {
    [sym_ms_declspec_modifier] = STATE(6585),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2912),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(1068)] = {
    [sym_ms_declspec_modifier] = STATE(6763),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2956),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(1069)] = {
    [sym_ms_declspec_modifier] = STATE(6735),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2924),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(1070)] = {
    [sym_ms_declspec_modifier] = STATE(6753),
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_typedef] = ACTIONS(2928),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(1071)] = {
    [ts_builtin_sym_end] = ACTIONS(2810),
    [sym_identifier] = ACTIONS(2808),
    [aux_sym_preproc_include_token1] = ACTIONS(2808),
    [aux_sym_preproc_include_token2] = ACTIONS(2808),
    [aux_sym_preproc_def_token1] = ACTIONS(2808),
    [aux_sym_preproc_if_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2808),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2808),
    [sym_preproc_directive] = ACTIONS(2808),
    [anon_sym_LPAREN2] = ACTIONS(2810),
    [anon_sym_BANG] = ACTIONS(2810),
    [anon_sym_TILDE] = ACTIONS(2810),
    [anon_sym_DASH] = ACTIONS(2808),
    [anon_sym_PLUS] = ACTIONS(2808),
    [anon_sym_STAR] = ACTIONS(2810),
    [anon_sym_CARET] = ACTIONS(2810),
    [anon_sym_AMP] = ACTIONS(2810),
    [anon_sym_SEMI] = ACTIONS(2810),
    [anon_sym___extension__] = ACTIONS(2808),
    [anon_sym_typedef] = ACTIONS(2808),
    [anon_sym_extern] = ACTIONS(2808),
    [anon_sym___attribute__] = ACTIONS(2808),
    [anon_sym___attribute] = ACTIONS(2808),
    [anon_sym_noreturn] = ACTIONS(2808),
    [anon_sym_LBRACK] = ACTIONS(2810),
    [anon_sym___declspec] = ACTIONS(2808),
    [anon_sym___cdecl] = ACTIONS(2808),
    [anon_sym___clrcall] = ACTIONS(2808),
    [anon_sym___stdcall] = ACTIONS(2808),
    [anon_sym___fastcall] = ACTIONS(2808),
    [anon_sym___thiscall] = ACTIONS(2808),
    [anon_sym___vectorcall] = ACTIONS(2808),
    [anon_sym_LBRACE] = ACTIONS(2810),
    [anon_sym_signed] = ACTIONS(2808),
    [anon_sym_unsigned] = ACTIONS(2808),
    [anon_sym_long] = ACTIONS(2808),
    [anon_sym_short] = ACTIONS(2808),
    [anon_sym_static] = ACTIONS(2808),
    [anon_sym_ATautoreleasepool] = ACTIONS(2810),
    [anon_sym_auto] = ACTIONS(2808),
    [anon_sym_register] = ACTIONS(2808),
    [anon_sym_inline] = ACTIONS(2808),
    [anon_sym___inline] = ACTIONS(2808),
    [anon_sym___inline__] = ACTIONS(2808),
    [anon_sym___forceinline] = ACTIONS(2808),
    [anon_sym_thread_local] = ACTIONS(2808),
    [anon_sym___thread] = ACTIONS(2808),
    [anon_sym_CG_EXTERN] = ACTIONS(2808),
    [anon_sym_CG_INLINE] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2808),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2808),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2808),
    [anon_sym_IBOutlet] = ACTIONS(2808),
    [anon_sym_IBInspectable] = ACTIONS(2808),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2808),
    [anon_sym_NS_INLINE] = ACTIONS(2808),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2808),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2808),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2808),
    [anon_sym_const] = ACTIONS(2808),
    [anon_sym_constexpr] = ACTIONS(2808),
    [anon_sym_volatile] = ACTIONS(2808),
    [anon_sym_restrict] = ACTIONS(2808),
    [anon_sym___restrict__] = ACTIONS(2808),
    [anon_sym__Atomic] = ACTIONS(2808),
    [anon_sym__Noreturn] = ACTIONS(2808),
    [anon_sym__Nonnull] = ACTIONS(2808),
    [anon_sym_nullable] = ACTIONS(2808),
    [anon_sym__Complex] = ACTIONS(2808),
    [anon_sym__Nullable] = ACTIONS(2808),
    [anon_sym__Nullable_result] = ACTIONS(2808),
    [anon_sym__Null_unspecified] = ACTIONS(2808),
    [anon_sym___autoreleasing] = ACTIONS(2808),
    [anon_sym___block] = ACTIONS(2808),
    [anon_sym___bridge] = ACTIONS(2808),
    [anon_sym___bridge_retained] = ACTIONS(2808),
    [anon_sym___bridge_transfer] = ACTIONS(2808),
    [anon_sym___complex] = ACTIONS(2808),
    [anon_sym___const] = ACTIONS(2808),
    [anon_sym___imag] = ACTIONS(2808),
    [anon_sym___kindof] = ACTIONS(2808),
    [anon_sym___nonnull] = ACTIONS(2808),
    [anon_sym___nullable] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2808),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2808),
    [anon_sym___real] = ACTIONS(2808),
    [anon_sym___strong] = ACTIONS(2808),
    [anon_sym___unsafe_unretained] = ACTIONS(2808),
    [anon_sym___unused] = ACTIONS(2808),
    [anon_sym___weak] = ACTIONS(2808),
    [anon_sym_alignas] = ACTIONS(2808),
    [anon_sym__Alignas] = ACTIONS(2808),
    [sym_primitive_type] = ACTIONS(2808),
    [anon_sym_enum] = ACTIONS(2808),
    [anon_sym_struct] = ACTIONS(2808),
    [anon_sym_union] = ACTIONS(2808),
    [anon_sym_if] = ACTIONS(2808),
    [anon_sym_switch] = ACTIONS(2808),
    [anon_sym_case] = ACTIONS(2808),
    [anon_sym_default] = ACTIONS(2808),
    [anon_sym_while] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2808),
    [anon_sym_in] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_break] = ACTIONS(2808),
    [anon_sym_continue] = ACTIONS(2808),
    [anon_sym_goto] = ACTIONS(2808),
    [anon_sym_DASH_DASH] = ACTIONS(2810),
    [anon_sym_PLUS_PLUS] = ACTIONS(2810),
    [anon_sym_sizeof] = ACTIONS(2808),
    [anon_sym___alignof__] = ACTIONS(2808),
    [anon_sym___alignof] = ACTIONS(2808),
    [anon_sym__alignof] = ACTIONS(2808),
    [anon_sym_alignof] = ACTIONS(2808),
    [anon_sym__Alignof] = ACTIONS(2808),
    [anon_sym_offsetof] = ACTIONS(2808),
    [anon_sym__Generic] = ACTIONS(2808),
    [anon_sym_asm] = ACTIONS(2808),
    [anon_sym___asm__] = ACTIONS(2808),
    [anon_sym___asm] = ACTIONS(2808),
    [sym_number_literal] = ACTIONS(2810),
    [anon_sym_L_SQUOTE] = ACTIONS(2810),
    [anon_sym_u_SQUOTE] = ACTIONS(2810),
    [anon_sym_U_SQUOTE] = ACTIONS(2810),
    [anon_sym_u8_SQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2810),
    [anon_sym_AT] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_L_DQUOTE] = ACTIONS(2810),
    [anon_sym_u_DQUOTE] = ACTIONS(2810),
    [anon_sym_U_DQUOTE] = ACTIONS(2810),
    [anon_sym_u8_DQUOTE] = ACTIONS(2810),
    [sym_true] = ACTIONS(2808),
    [sym_false] = ACTIONS(2808),
    [anon_sym_NULL] = ACTIONS(2808),
    [anon_sym_nullptr] = ACTIONS(2808),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2808),
    [anon_sym___typeof] = ACTIONS(2808),
    [anon_sym_typeof] = ACTIONS(2808),
    [anon_sym_ATimport] = ACTIONS(2810),
    [aux_sym_preproc_undef_token1] = ACTIONS(2808),
    [anon_sym_POUND] = ACTIONS(2808),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2808),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2808),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2808),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2808),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2808),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_API_AVAILABLE] = ACTIONS(2808),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_API_DEPRECATED] = ACTIONS(2808),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2808),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2808),
    [anon_sym___deprecated_msg] = ACTIONS(2808),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2808),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2808),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2808),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2808),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2808),
    [anon_sym_ATprotocol] = ACTIONS(2810),
    [anon_sym_ATinterface] = ACTIONS(2810),
    [anon_sym_ATimplementation] = ACTIONS(2810),
    [anon_sym_ATcompatibility_alias] = ACTIONS(2810),
    [anon_sym_ATselector] = ACTIONS(2810),
    [anon_sym_ATavailable] = ACTIONS(2810),
    [anon_sym___builtin_available] = ACTIONS(2808),
    [anon_sym_va_arg] = ACTIONS(2808),
    [anon_sym_ATencode] = ACTIONS(2810),
    [anon_sym_BOOL] = ACTIONS(2808),
    [anon_sym_IMP] = ACTIONS(2808),
    [anon_sym_SEL] = ACTIONS(2808),
    [anon_sym_Class] = ACTIONS(2808),
    [anon_sym_id] = ACTIONS(2808),
  },
  [STATE(1072)] = {
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(2914),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(1073)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym___declspec] = ACTIONS(2914),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(1074)] = {
    [sym_identifier] = ACTIONS(2958),
    [anon_sym_COMMA] = ACTIONS(2960),
    [anon_sym_RPAREN] = ACTIONS(2960),
    [anon_sym_LPAREN2] = ACTIONS(2960),
    [anon_sym_BANG] = ACTIONS(2960),
    [anon_sym_TILDE] = ACTIONS(2960),
    [anon_sym_DASH] = ACTIONS(2958),
    [anon_sym_PLUS] = ACTIONS(2958),
    [anon_sym_STAR] = ACTIONS(2960),
    [anon_sym_CARET] = ACTIONS(2960),
    [anon_sym_AMP] = ACTIONS(2960),
    [anon_sym_GT] = ACTIONS(2960),
    [anon_sym_SEMI] = ACTIONS(2960),
    [anon_sym___extension__] = ACTIONS(2958),
    [anon_sym_extern] = ACTIONS(2958),
    [anon_sym___attribute__] = ACTIONS(2958),
    [anon_sym___attribute] = ACTIONS(2958),
    [anon_sym_noreturn] = ACTIONS(2958),
    [anon_sym_LBRACK] = ACTIONS(2960),
    [anon_sym___declspec] = ACTIONS(2958),
    [anon_sym_LBRACE] = ACTIONS(2960),
    [anon_sym_signed] = ACTIONS(2958),
    [anon_sym_unsigned] = ACTIONS(2958),
    [anon_sym_long] = ACTIONS(2958),
    [anon_sym_short] = ACTIONS(2958),
    [anon_sym_static] = ACTIONS(2958),
    [anon_sym_ATautoreleasepool] = ACTIONS(2960),
    [anon_sym_auto] = ACTIONS(2958),
    [anon_sym_register] = ACTIONS(2958),
    [anon_sym_inline] = ACTIONS(2958),
    [anon_sym___inline] = ACTIONS(2958),
    [anon_sym___inline__] = ACTIONS(2958),
    [anon_sym___forceinline] = ACTIONS(2958),
    [anon_sym_thread_local] = ACTIONS(2958),
    [anon_sym___thread] = ACTIONS(2958),
    [anon_sym_CG_EXTERN] = ACTIONS(2958),
    [anon_sym_CG_INLINE] = ACTIONS(2958),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2958),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2958),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2958),
    [anon_sym_IBOutlet] = ACTIONS(2958),
    [anon_sym_IBInspectable] = ACTIONS(2958),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2958),
    [anon_sym_NS_INLINE] = ACTIONS(2958),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2958),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2958),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2958),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2958),
    [anon_sym_const] = ACTIONS(2958),
    [anon_sym_constexpr] = ACTIONS(2958),
    [anon_sym_volatile] = ACTIONS(2958),
    [anon_sym_restrict] = ACTIONS(2958),
    [anon_sym___restrict__] = ACTIONS(2958),
    [anon_sym__Atomic] = ACTIONS(2958),
    [anon_sym__Noreturn] = ACTIONS(2958),
    [anon_sym__Nonnull] = ACTIONS(2958),
    [anon_sym_nullable] = ACTIONS(2958),
    [anon_sym__Complex] = ACTIONS(2958),
    [anon_sym__Nullable] = ACTIONS(2958),
    [anon_sym__Nullable_result] = ACTIONS(2958),
    [anon_sym__Null_unspecified] = ACTIONS(2958),
    [anon_sym___autoreleasing] = ACTIONS(2958),
    [anon_sym___block] = ACTIONS(2958),
    [anon_sym___bridge] = ACTIONS(2958),
    [anon_sym___bridge_retained] = ACTIONS(2958),
    [anon_sym___bridge_transfer] = ACTIONS(2958),
    [anon_sym___complex] = ACTIONS(2958),
    [anon_sym___const] = ACTIONS(2958),
    [anon_sym___imag] = ACTIONS(2958),
    [anon_sym___kindof] = ACTIONS(2958),
    [anon_sym___nonnull] = ACTIONS(2958),
    [anon_sym___nullable] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2958),
    [anon_sym___real] = ACTIONS(2958),
    [anon_sym___strong] = ACTIONS(2958),
    [anon_sym___unsafe_unretained] = ACTIONS(2958),
    [anon_sym___unused] = ACTIONS(2958),
    [anon_sym___weak] = ACTIONS(2958),
    [anon_sym_alignas] = ACTIONS(2958),
    [anon_sym__Alignas] = ACTIONS(2958),
    [sym_primitive_type] = ACTIONS(2958),
    [anon_sym_enum] = ACTIONS(2958),
    [anon_sym_COLON] = ACTIONS(2960),
    [anon_sym_struct] = ACTIONS(2958),
    [anon_sym_union] = ACTIONS(2958),
    [anon_sym_if] = ACTIONS(2958),
    [anon_sym_switch] = ACTIONS(2958),
    [anon_sym_case] = ACTIONS(2958),
    [anon_sym_default] = ACTIONS(2958),
    [anon_sym_while] = ACTIONS(2958),
    [anon_sym_do] = ACTIONS(2958),
    [anon_sym_for] = ACTIONS(2958),
    [anon_sym_in] = ACTIONS(2958),
    [anon_sym_return] = ACTIONS(2958),
    [anon_sym_break] = ACTIONS(2958),
    [anon_sym_continue] = ACTIONS(2958),
    [anon_sym_goto] = ACTIONS(2958),
    [anon_sym_DASH_DASH] = ACTIONS(2960),
    [anon_sym_PLUS_PLUS] = ACTIONS(2960),
    [anon_sym_sizeof] = ACTIONS(2958),
    [anon_sym___alignof__] = ACTIONS(2958),
    [anon_sym___alignof] = ACTIONS(2958),
    [anon_sym__alignof] = ACTIONS(2958),
    [anon_sym_alignof] = ACTIONS(2958),
    [anon_sym__Alignof] = ACTIONS(2958),
    [anon_sym_offsetof] = ACTIONS(2958),
    [anon_sym__Generic] = ACTIONS(2958),
    [anon_sym_asm] = ACTIONS(2958),
    [anon_sym___asm__] = ACTIONS(2958),
    [anon_sym___asm] = ACTIONS(2958),
    [sym_number_literal] = ACTIONS(2960),
    [anon_sym_L_SQUOTE] = ACTIONS(2960),
    [anon_sym_u_SQUOTE] = ACTIONS(2960),
    [anon_sym_U_SQUOTE] = ACTIONS(2960),
    [anon_sym_u8_SQUOTE] = ACTIONS(2960),
    [anon_sym_SQUOTE] = ACTIONS(2960),
    [anon_sym_AT] = ACTIONS(2958),
    [anon_sym_DQUOTE] = ACTIONS(2960),
    [anon_sym_L_DQUOTE] = ACTIONS(2960),
    [anon_sym_u_DQUOTE] = ACTIONS(2960),
    [anon_sym_U_DQUOTE] = ACTIONS(2960),
    [anon_sym_u8_DQUOTE] = ACTIONS(2960),
    [sym_true] = ACTIONS(2958),
    [sym_false] = ACTIONS(2958),
    [anon_sym_NULL] = ACTIONS(2958),
    [anon_sym_nullptr] = ACTIONS(2958),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2958),
    [anon_sym___typeof] = ACTIONS(2958),
    [anon_sym_typeof] = ACTIONS(2958),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2958),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2958),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2958),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2958),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2958),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2958),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2958),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2958),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2958),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_API_AVAILABLE] = ACTIONS(2958),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_API_DEPRECATED] = ACTIONS(2958),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2958),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2958),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2958),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2958),
    [anon_sym___deprecated_msg] = ACTIONS(2958),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2958),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2958),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2958),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2958),
    [anon_sym_ATprotocol] = ACTIONS(2960),
    [anon_sym_ATinterface] = ACTIONS(2960),
    [anon_sym_ATimplementation] = ACTIONS(2960),
    [anon_sym_ATtry] = ACTIONS(2960),
    [anon_sym___try] = ACTIONS(2958),
    [anon_sym_ATthrow] = ACTIONS(2960),
    [anon_sym_ATselector] = ACTIONS(2960),
    [anon_sym_ATavailable] = ACTIONS(2960),
    [anon_sym___builtin_available] = ACTIONS(2958),
    [anon_sym_va_arg] = ACTIONS(2958),
    [anon_sym_ATencode] = ACTIONS(2960),
    [anon_sym_ATsynchronized] = ACTIONS(2960),
    [anon_sym_BOOL] = ACTIONS(2958),
    [anon_sym_IMP] = ACTIONS(2958),
    [anon_sym_SEL] = ACTIONS(2958),
    [anon_sym_Class] = ACTIONS(2958),
    [anon_sym_id] = ACTIONS(2958),
  },
  [STATE(1075)] = {
    [sym_identifier] = ACTIONS(2962),
    [anon_sym_COMMA] = ACTIONS(2964),
    [anon_sym_RPAREN] = ACTIONS(2964),
    [anon_sym_LPAREN2] = ACTIONS(2964),
    [anon_sym_BANG] = ACTIONS(2964),
    [anon_sym_TILDE] = ACTIONS(2964),
    [anon_sym_DASH] = ACTIONS(2962),
    [anon_sym_PLUS] = ACTIONS(2962),
    [anon_sym_STAR] = ACTIONS(2964),
    [anon_sym_CARET] = ACTIONS(2964),
    [anon_sym_AMP] = ACTIONS(2964),
    [anon_sym_GT] = ACTIONS(2964),
    [anon_sym_SEMI] = ACTIONS(2964),
    [anon_sym___extension__] = ACTIONS(2962),
    [anon_sym_extern] = ACTIONS(2962),
    [anon_sym___attribute__] = ACTIONS(2962),
    [anon_sym___attribute] = ACTIONS(2962),
    [anon_sym_noreturn] = ACTIONS(2962),
    [anon_sym_LBRACK] = ACTIONS(2964),
    [anon_sym___declspec] = ACTIONS(2962),
    [anon_sym_LBRACE] = ACTIONS(2964),
    [anon_sym_signed] = ACTIONS(2962),
    [anon_sym_unsigned] = ACTIONS(2962),
    [anon_sym_long] = ACTIONS(2962),
    [anon_sym_short] = ACTIONS(2962),
    [anon_sym_static] = ACTIONS(2962),
    [anon_sym_ATautoreleasepool] = ACTIONS(2964),
    [anon_sym_auto] = ACTIONS(2962),
    [anon_sym_register] = ACTIONS(2962),
    [anon_sym_inline] = ACTIONS(2962),
    [anon_sym___inline] = ACTIONS(2962),
    [anon_sym___inline__] = ACTIONS(2962),
    [anon_sym___forceinline] = ACTIONS(2962),
    [anon_sym_thread_local] = ACTIONS(2962),
    [anon_sym___thread] = ACTIONS(2962),
    [anon_sym_CG_EXTERN] = ACTIONS(2962),
    [anon_sym_CG_INLINE] = ACTIONS(2962),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2962),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2962),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2962),
    [anon_sym_IBOutlet] = ACTIONS(2962),
    [anon_sym_IBInspectable] = ACTIONS(2962),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2962),
    [anon_sym_NS_INLINE] = ACTIONS(2962),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2962),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2962),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2962),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2962),
    [anon_sym_const] = ACTIONS(2962),
    [anon_sym_constexpr] = ACTIONS(2962),
    [anon_sym_volatile] = ACTIONS(2962),
    [anon_sym_restrict] = ACTIONS(2962),
    [anon_sym___restrict__] = ACTIONS(2962),
    [anon_sym__Atomic] = ACTIONS(2962),
    [anon_sym__Noreturn] = ACTIONS(2962),
    [anon_sym__Nonnull] = ACTIONS(2962),
    [anon_sym_nullable] = ACTIONS(2962),
    [anon_sym__Complex] = ACTIONS(2962),
    [anon_sym__Nullable] = ACTIONS(2962),
    [anon_sym__Nullable_result] = ACTIONS(2962),
    [anon_sym__Null_unspecified] = ACTIONS(2962),
    [anon_sym___autoreleasing] = ACTIONS(2962),
    [anon_sym___block] = ACTIONS(2962),
    [anon_sym___bridge] = ACTIONS(2962),
    [anon_sym___bridge_retained] = ACTIONS(2962),
    [anon_sym___bridge_transfer] = ACTIONS(2962),
    [anon_sym___complex] = ACTIONS(2962),
    [anon_sym___const] = ACTIONS(2962),
    [anon_sym___imag] = ACTIONS(2962),
    [anon_sym___kindof] = ACTIONS(2962),
    [anon_sym___nonnull] = ACTIONS(2962),
    [anon_sym___nullable] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2962),
    [anon_sym___real] = ACTIONS(2962),
    [anon_sym___strong] = ACTIONS(2962),
    [anon_sym___unsafe_unretained] = ACTIONS(2962),
    [anon_sym___unused] = ACTIONS(2962),
    [anon_sym___weak] = ACTIONS(2962),
    [anon_sym_alignas] = ACTIONS(2962),
    [anon_sym__Alignas] = ACTIONS(2962),
    [sym_primitive_type] = ACTIONS(2962),
    [anon_sym_enum] = ACTIONS(2962),
    [anon_sym_COLON] = ACTIONS(2964),
    [anon_sym_struct] = ACTIONS(2962),
    [anon_sym_union] = ACTIONS(2962),
    [anon_sym_if] = ACTIONS(2962),
    [anon_sym_switch] = ACTIONS(2962),
    [anon_sym_case] = ACTIONS(2962),
    [anon_sym_default] = ACTIONS(2962),
    [anon_sym_while] = ACTIONS(2962),
    [anon_sym_do] = ACTIONS(2962),
    [anon_sym_for] = ACTIONS(2962),
    [anon_sym_in] = ACTIONS(2962),
    [anon_sym_return] = ACTIONS(2962),
    [anon_sym_break] = ACTIONS(2962),
    [anon_sym_continue] = ACTIONS(2962),
    [anon_sym_goto] = ACTIONS(2962),
    [anon_sym_DASH_DASH] = ACTIONS(2964),
    [anon_sym_PLUS_PLUS] = ACTIONS(2964),
    [anon_sym_sizeof] = ACTIONS(2962),
    [anon_sym___alignof__] = ACTIONS(2962),
    [anon_sym___alignof] = ACTIONS(2962),
    [anon_sym__alignof] = ACTIONS(2962),
    [anon_sym_alignof] = ACTIONS(2962),
    [anon_sym__Alignof] = ACTIONS(2962),
    [anon_sym_offsetof] = ACTIONS(2962),
    [anon_sym__Generic] = ACTIONS(2962),
    [anon_sym_asm] = ACTIONS(2962),
    [anon_sym___asm__] = ACTIONS(2962),
    [anon_sym___asm] = ACTIONS(2962),
    [sym_number_literal] = ACTIONS(2964),
    [anon_sym_L_SQUOTE] = ACTIONS(2964),
    [anon_sym_u_SQUOTE] = ACTIONS(2964),
    [anon_sym_U_SQUOTE] = ACTIONS(2964),
    [anon_sym_u8_SQUOTE] = ACTIONS(2964),
    [anon_sym_SQUOTE] = ACTIONS(2964),
    [anon_sym_AT] = ACTIONS(2962),
    [anon_sym_DQUOTE] = ACTIONS(2964),
    [anon_sym_L_DQUOTE] = ACTIONS(2964),
    [anon_sym_u_DQUOTE] = ACTIONS(2964),
    [anon_sym_U_DQUOTE] = ACTIONS(2964),
    [anon_sym_u8_DQUOTE] = ACTIONS(2964),
    [sym_true] = ACTIONS(2962),
    [sym_false] = ACTIONS(2962),
    [anon_sym_NULL] = ACTIONS(2962),
    [anon_sym_nullptr] = ACTIONS(2962),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2962),
    [anon_sym___typeof] = ACTIONS(2962),
    [anon_sym_typeof] = ACTIONS(2962),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2962),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2962),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2962),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2962),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2962),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2962),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2962),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2962),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2962),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_API_AVAILABLE] = ACTIONS(2962),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_API_DEPRECATED] = ACTIONS(2962),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2962),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2962),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2962),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2962),
    [anon_sym___deprecated_msg] = ACTIONS(2962),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2962),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2962),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2962),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2962),
    [anon_sym_ATprotocol] = ACTIONS(2964),
    [anon_sym_ATinterface] = ACTIONS(2964),
    [anon_sym_ATimplementation] = ACTIONS(2964),
    [anon_sym_ATtry] = ACTIONS(2964),
    [anon_sym___try] = ACTIONS(2962),
    [anon_sym_ATthrow] = ACTIONS(2964),
    [anon_sym_ATselector] = ACTIONS(2964),
    [anon_sym_ATavailable] = ACTIONS(2964),
    [anon_sym___builtin_available] = ACTIONS(2962),
    [anon_sym_va_arg] = ACTIONS(2962),
    [anon_sym_ATencode] = ACTIONS(2964),
    [anon_sym_ATsynchronized] = ACTIONS(2964),
    [anon_sym_BOOL] = ACTIONS(2962),
    [anon_sym_IMP] = ACTIONS(2962),
    [anon_sym_SEL] = ACTIONS(2962),
    [anon_sym_Class] = ACTIONS(2962),
    [anon_sym_id] = ACTIONS(2962),
  },
  [STATE(1076)] = {
    [sym_catch_clause] = STATE(1077),
    [sym_finally_clause] = STATE(1149),
    [aux_sym_try_statement_repeat1] = STATE(1077),
    [sym_identifier] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATcatch] = ACTIONS(2966),
    [anon_sym___catch] = ACTIONS(2968),
    [anon_sym_ATfinally] = ACTIONS(2970),
    [anon_sym___finally] = ACTIONS(2972),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(1077)] = {
    [sym_catch_clause] = STATE(1077),
    [aux_sym_try_statement_repeat1] = STATE(1077),
    [sym_identifier] = ACTIONS(2318),
    [anon_sym_LPAREN2] = ACTIONS(2320),
    [anon_sym_BANG] = ACTIONS(2320),
    [anon_sym_TILDE] = ACTIONS(2320),
    [anon_sym_DASH] = ACTIONS(2318),
    [anon_sym_PLUS] = ACTIONS(2318),
    [anon_sym_STAR] = ACTIONS(2320),
    [anon_sym_CARET] = ACTIONS(2320),
    [anon_sym_AMP] = ACTIONS(2320),
    [anon_sym_SEMI] = ACTIONS(2320),
    [anon_sym___extension__] = ACTIONS(2318),
    [anon_sym_typedef] = ACTIONS(2318),
    [anon_sym_extern] = ACTIONS(2318),
    [anon_sym___attribute__] = ACTIONS(2318),
    [anon_sym___attribute] = ACTIONS(2318),
    [anon_sym_noreturn] = ACTIONS(2318),
    [anon_sym_LBRACK] = ACTIONS(2320),
    [anon_sym___declspec] = ACTIONS(2318),
    [anon_sym_LBRACE] = ACTIONS(2320),
    [anon_sym_signed] = ACTIONS(2318),
    [anon_sym_unsigned] = ACTIONS(2318),
    [anon_sym_long] = ACTIONS(2318),
    [anon_sym_short] = ACTIONS(2318),
    [anon_sym_static] = ACTIONS(2318),
    [anon_sym_ATautoreleasepool] = ACTIONS(2320),
    [anon_sym_auto] = ACTIONS(2318),
    [anon_sym_register] = ACTIONS(2318),
    [anon_sym_inline] = ACTIONS(2318),
    [anon_sym___inline] = ACTIONS(2318),
    [anon_sym___inline__] = ACTIONS(2318),
    [anon_sym___forceinline] = ACTIONS(2318),
    [anon_sym_thread_local] = ACTIONS(2318),
    [anon_sym___thread] = ACTIONS(2318),
    [anon_sym_CG_EXTERN] = ACTIONS(2318),
    [anon_sym_CG_INLINE] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2318),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2318),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2318),
    [anon_sym_IBOutlet] = ACTIONS(2318),
    [anon_sym_IBInspectable] = ACTIONS(2318),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2318),
    [anon_sym_NS_INLINE] = ACTIONS(2318),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2318),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2318),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2318),
    [anon_sym_const] = ACTIONS(2318),
    [anon_sym_constexpr] = ACTIONS(2318),
    [anon_sym_volatile] = ACTIONS(2318),
    [anon_sym_restrict] = ACTIONS(2318),
    [anon_sym___restrict__] = ACTIONS(2318),
    [anon_sym__Atomic] = ACTIONS(2318),
    [anon_sym__Noreturn] = ACTIONS(2318),
    [anon_sym__Nonnull] = ACTIONS(2318),
    [anon_sym_nullable] = ACTIONS(2318),
    [anon_sym__Complex] = ACTIONS(2318),
    [anon_sym__Nullable] = ACTIONS(2318),
    [anon_sym__Nullable_result] = ACTIONS(2318),
    [anon_sym__Null_unspecified] = ACTIONS(2318),
    [anon_sym___autoreleasing] = ACTIONS(2318),
    [anon_sym___block] = ACTIONS(2318),
    [anon_sym___bridge] = ACTIONS(2318),
    [anon_sym___bridge_retained] = ACTIONS(2318),
    [anon_sym___bridge_transfer] = ACTIONS(2318),
    [anon_sym___complex] = ACTIONS(2318),
    [anon_sym___const] = ACTIONS(2318),
    [anon_sym___imag] = ACTIONS(2318),
    [anon_sym___kindof] = ACTIONS(2318),
    [anon_sym___nonnull] = ACTIONS(2318),
    [anon_sym___nullable] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2318),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2318),
    [anon_sym___real] = ACTIONS(2318),
    [anon_sym___strong] = ACTIONS(2318),
    [anon_sym___unsafe_unretained] = ACTIONS(2318),
    [anon_sym___unused] = ACTIONS(2318),
    [anon_sym___weak] = ACTIONS(2318),
    [anon_sym_alignas] = ACTIONS(2318),
    [anon_sym__Alignas] = ACTIONS(2318),
    [sym_primitive_type] = ACTIONS(2318),
    [anon_sym_enum] = ACTIONS(2318),
    [anon_sym_struct] = ACTIONS(2318),
    [anon_sym_union] = ACTIONS(2318),
    [anon_sym_if] = ACTIONS(2318),
    [anon_sym_else] = ACTIONS(2318),
    [anon_sym_switch] = ACTIONS(2318),
    [anon_sym_while] = ACTIONS(2318),
    [anon_sym_do] = ACTIONS(2318),
    [anon_sym_for] = ACTIONS(2318),
    [anon_sym_in] = ACTIONS(2318),
    [anon_sym_return] = ACTIONS(2318),
    [anon_sym_break] = ACTIONS(2318),
    [anon_sym_continue] = ACTIONS(2318),
    [anon_sym_goto] = ACTIONS(2318),
    [anon_sym_DASH_DASH] = ACTIONS(2320),
    [anon_sym_PLUS_PLUS] = ACTIONS(2320),
    [anon_sym_sizeof] = ACTIONS(2318),
    [anon_sym___alignof__] = ACTIONS(2318),
    [anon_sym___alignof] = ACTIONS(2318),
    [anon_sym__alignof] = ACTIONS(2318),
    [anon_sym_alignof] = ACTIONS(2318),
    [anon_sym__Alignof] = ACTIONS(2318),
    [anon_sym_offsetof] = ACTIONS(2318),
    [anon_sym__Generic] = ACTIONS(2318),
    [anon_sym_asm] = ACTIONS(2318),
    [anon_sym___asm__] = ACTIONS(2318),
    [anon_sym___asm] = ACTIONS(2318),
    [sym_number_literal] = ACTIONS(2320),
    [anon_sym_L_SQUOTE] = ACTIONS(2320),
    [anon_sym_u_SQUOTE] = ACTIONS(2320),
    [anon_sym_U_SQUOTE] = ACTIONS(2320),
    [anon_sym_u8_SQUOTE] = ACTIONS(2320),
    [anon_sym_SQUOTE] = ACTIONS(2320),
    [anon_sym_AT] = ACTIONS(2318),
    [anon_sym_DQUOTE] = ACTIONS(2320),
    [anon_sym_L_DQUOTE] = ACTIONS(2320),
    [anon_sym_u_DQUOTE] = ACTIONS(2320),
    [anon_sym_U_DQUOTE] = ACTIONS(2320),
    [anon_sym_u8_DQUOTE] = ACTIONS(2320),
    [sym_true] = ACTIONS(2318),
    [sym_false] = ACTIONS(2318),
    [anon_sym_NULL] = ACTIONS(2318),
    [anon_sym_nullptr] = ACTIONS(2318),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2318),
    [anon_sym___typeof] = ACTIONS(2318),
    [anon_sym_typeof] = ACTIONS(2318),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2318),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2318),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2318),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2318),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2318),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_API_AVAILABLE] = ACTIONS(2318),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_API_DEPRECATED] = ACTIONS(2318),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2318),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2318),
    [anon_sym___deprecated_msg] = ACTIONS(2318),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2318),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2318),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2318),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2318),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2318),
    [anon_sym_ATtry] = ACTIONS(2320),
    [anon_sym___try] = ACTIONS(2318),
    [anon_sym_ATcatch] = ACTIONS(2974),
    [anon_sym___catch] = ACTIONS(2977),
    [anon_sym_ATfinally] = ACTIONS(2320),
    [anon_sym___finally] = ACTIONS(2318),
    [anon_sym_ATthrow] = ACTIONS(2320),
    [anon_sym_ATselector] = ACTIONS(2320),
    [anon_sym_ATavailable] = ACTIONS(2320),
    [anon_sym___builtin_available] = ACTIONS(2318),
    [anon_sym_va_arg] = ACTIONS(2318),
    [anon_sym_ATencode] = ACTIONS(2320),
    [anon_sym_ATsynchronized] = ACTIONS(2320),
    [anon_sym_BOOL] = ACTIONS(2318),
    [anon_sym_IMP] = ACTIONS(2318),
    [anon_sym_SEL] = ACTIONS(2318),
    [anon_sym_Class] = ACTIONS(2318),
    [anon_sym_id] = ACTIONS(2318),
  },
  [STATE(1078)] = {
    [sym_identifier] = ACTIONS(2368),
    [anon_sym_LPAREN2] = ACTIONS(2370),
    [anon_sym_BANG] = ACTIONS(2370),
    [anon_sym_TILDE] = ACTIONS(2370),
    [anon_sym_DASH] = ACTIONS(2368),
    [anon_sym_PLUS] = ACTIONS(2368),
    [anon_sym_STAR] = ACTIONS(2370),
    [anon_sym_CARET] = ACTIONS(2370),
    [anon_sym_AMP] = ACTIONS(2370),
    [anon_sym_SEMI] = ACTIONS(2370),
    [anon_sym___extension__] = ACTIONS(2368),
    [anon_sym_typedef] = ACTIONS(2368),
    [anon_sym_extern] = ACTIONS(2368),
    [anon_sym___attribute__] = ACTIONS(2368),
    [anon_sym___attribute] = ACTIONS(2368),
    [anon_sym_noreturn] = ACTIONS(2368),
    [anon_sym_LBRACK] = ACTIONS(2370),
    [anon_sym___declspec] = ACTIONS(2368),
    [anon_sym_LBRACE] = ACTIONS(2370),
    [anon_sym_signed] = ACTIONS(2368),
    [anon_sym_unsigned] = ACTIONS(2368),
    [anon_sym_long] = ACTIONS(2368),
    [anon_sym_short] = ACTIONS(2368),
    [anon_sym_static] = ACTIONS(2368),
    [anon_sym_ATautoreleasepool] = ACTIONS(2370),
    [anon_sym_auto] = ACTIONS(2368),
    [anon_sym_register] = ACTIONS(2368),
    [anon_sym_inline] = ACTIONS(2368),
    [anon_sym___inline] = ACTIONS(2368),
    [anon_sym___inline__] = ACTIONS(2368),
    [anon_sym___forceinline] = ACTIONS(2368),
    [anon_sym_thread_local] = ACTIONS(2368),
    [anon_sym___thread] = ACTIONS(2368),
    [anon_sym_CG_EXTERN] = ACTIONS(2368),
    [anon_sym_CG_INLINE] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2368),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2368),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2368),
    [anon_sym_IBOutlet] = ACTIONS(2368),
    [anon_sym_IBInspectable] = ACTIONS(2368),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2368),
    [anon_sym_NS_INLINE] = ACTIONS(2368),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2368),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2368),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2368),
    [anon_sym_const] = ACTIONS(2368),
    [anon_sym_constexpr] = ACTIONS(2368),
    [anon_sym_volatile] = ACTIONS(2368),
    [anon_sym_restrict] = ACTIONS(2368),
    [anon_sym___restrict__] = ACTIONS(2368),
    [anon_sym__Atomic] = ACTIONS(2368),
    [anon_sym__Noreturn] = ACTIONS(2368),
    [anon_sym__Nonnull] = ACTIONS(2368),
    [anon_sym_nullable] = ACTIONS(2368),
    [anon_sym__Complex] = ACTIONS(2368),
    [anon_sym__Nullable] = ACTIONS(2368),
    [anon_sym__Nullable_result] = ACTIONS(2368),
    [anon_sym__Null_unspecified] = ACTIONS(2368),
    [anon_sym___autoreleasing] = ACTIONS(2368),
    [anon_sym___block] = ACTIONS(2368),
    [anon_sym___bridge] = ACTIONS(2368),
    [anon_sym___bridge_retained] = ACTIONS(2368),
    [anon_sym___bridge_transfer] = ACTIONS(2368),
    [anon_sym___complex] = ACTIONS(2368),
    [anon_sym___const] = ACTIONS(2368),
    [anon_sym___imag] = ACTIONS(2368),
    [anon_sym___kindof] = ACTIONS(2368),
    [anon_sym___nonnull] = ACTIONS(2368),
    [anon_sym___nullable] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2368),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2368),
    [anon_sym___real] = ACTIONS(2368),
    [anon_sym___strong] = ACTIONS(2368),
    [anon_sym___unsafe_unretained] = ACTIONS(2368),
    [anon_sym___unused] = ACTIONS(2368),
    [anon_sym___weak] = ACTIONS(2368),
    [anon_sym_alignas] = ACTIONS(2368),
    [anon_sym__Alignas] = ACTIONS(2368),
    [sym_primitive_type] = ACTIONS(2368),
    [anon_sym_enum] = ACTIONS(2368),
    [anon_sym_struct] = ACTIONS(2368),
    [anon_sym_union] = ACTIONS(2368),
    [anon_sym_if] = ACTIONS(2368),
    [anon_sym_else] = ACTIONS(2368),
    [anon_sym_switch] = ACTIONS(2368),
    [anon_sym_while] = ACTIONS(2368),
    [anon_sym_do] = ACTIONS(2368),
    [anon_sym_for] = ACTIONS(2368),
    [anon_sym_in] = ACTIONS(2368),
    [anon_sym_return] = ACTIONS(2368),
    [anon_sym_break] = ACTIONS(2368),
    [anon_sym_continue] = ACTIONS(2368),
    [anon_sym_goto] = ACTIONS(2368),
    [anon_sym_DASH_DASH] = ACTIONS(2370),
    [anon_sym_PLUS_PLUS] = ACTIONS(2370),
    [anon_sym_sizeof] = ACTIONS(2368),
    [anon_sym___alignof__] = ACTIONS(2368),
    [anon_sym___alignof] = ACTIONS(2368),
    [anon_sym__alignof] = ACTIONS(2368),
    [anon_sym_alignof] = ACTIONS(2368),
    [anon_sym__Alignof] = ACTIONS(2368),
    [anon_sym_offsetof] = ACTIONS(2368),
    [anon_sym__Generic] = ACTIONS(2368),
    [anon_sym_asm] = ACTIONS(2368),
    [anon_sym___asm__] = ACTIONS(2368),
    [anon_sym___asm] = ACTIONS(2368),
    [sym_number_literal] = ACTIONS(2370),
    [anon_sym_L_SQUOTE] = ACTIONS(2370),
    [anon_sym_u_SQUOTE] = ACTIONS(2370),
    [anon_sym_U_SQUOTE] = ACTIONS(2370),
    [anon_sym_u8_SQUOTE] = ACTIONS(2370),
    [anon_sym_SQUOTE] = ACTIONS(2370),
    [anon_sym_AT] = ACTIONS(2368),
    [anon_sym_DQUOTE] = ACTIONS(2370),
    [anon_sym_L_DQUOTE] = ACTIONS(2370),
    [anon_sym_u_DQUOTE] = ACTIONS(2370),
    [anon_sym_U_DQUOTE] = ACTIONS(2370),
    [anon_sym_u8_DQUOTE] = ACTIONS(2370),
    [sym_true] = ACTIONS(2368),
    [sym_false] = ACTIONS(2368),
    [anon_sym_NULL] = ACTIONS(2368),
    [anon_sym_nullptr] = ACTIONS(2368),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2368),
    [anon_sym___typeof] = ACTIONS(2368),
    [anon_sym_typeof] = ACTIONS(2368),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2368),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2368),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2368),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2368),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2368),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_API_AVAILABLE] = ACTIONS(2368),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_API_DEPRECATED] = ACTIONS(2368),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2368),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2368),
    [anon_sym___deprecated_msg] = ACTIONS(2368),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2368),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2368),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2368),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2368),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2368),
    [anon_sym_ATtry] = ACTIONS(2370),
    [anon_sym___try] = ACTIONS(2368),
    [anon_sym_ATcatch] = ACTIONS(2370),
    [anon_sym___catch] = ACTIONS(2368),
    [anon_sym_ATfinally] = ACTIONS(2370),
    [anon_sym___finally] = ACTIONS(2368),
    [anon_sym_ATthrow] = ACTIONS(2370),
    [anon_sym_ATselector] = ACTIONS(2370),
    [anon_sym_ATavailable] = ACTIONS(2370),
    [anon_sym___builtin_available] = ACTIONS(2368),
    [anon_sym_va_arg] = ACTIONS(2368),
    [anon_sym_ATencode] = ACTIONS(2370),
    [anon_sym_ATsynchronized] = ACTIONS(2370),
    [anon_sym_BOOL] = ACTIONS(2368),
    [anon_sym_IMP] = ACTIONS(2368),
    [anon_sym_SEL] = ACTIONS(2368),
    [anon_sym_Class] = ACTIONS(2368),
    [anon_sym_id] = ACTIONS(2368),
  },
  [STATE(1079)] = {
    [sym_identifier] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATcatch] = ACTIONS(2348),
    [anon_sym___catch] = ACTIONS(2346),
    [anon_sym_ATfinally] = ACTIONS(2348),
    [anon_sym___finally] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(1080)] = {
    [sym_identifier] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATcatch] = ACTIONS(2352),
    [anon_sym___catch] = ACTIONS(2350),
    [anon_sym_ATfinally] = ACTIONS(2352),
    [anon_sym___finally] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(1081)] = {
    [sym_identifier] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATcatch] = ACTIONS(2362),
    [anon_sym___catch] = ACTIONS(2360),
    [anon_sym_ATfinally] = ACTIONS(2362),
    [anon_sym___finally] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(1082)] = {
    [sym_identifier] = ACTIONS(2364),
    [anon_sym_LPAREN2] = ACTIONS(2366),
    [anon_sym_BANG] = ACTIONS(2366),
    [anon_sym_TILDE] = ACTIONS(2366),
    [anon_sym_DASH] = ACTIONS(2364),
    [anon_sym_PLUS] = ACTIONS(2364),
    [anon_sym_STAR] = ACTIONS(2366),
    [anon_sym_CARET] = ACTIONS(2366),
    [anon_sym_AMP] = ACTIONS(2366),
    [anon_sym_SEMI] = ACTIONS(2366),
    [anon_sym___extension__] = ACTIONS(2364),
    [anon_sym_typedef] = ACTIONS(2364),
    [anon_sym_extern] = ACTIONS(2364),
    [anon_sym___attribute__] = ACTIONS(2364),
    [anon_sym___attribute] = ACTIONS(2364),
    [anon_sym_noreturn] = ACTIONS(2364),
    [anon_sym_LBRACK] = ACTIONS(2366),
    [anon_sym___declspec] = ACTIONS(2364),
    [anon_sym_LBRACE] = ACTIONS(2366),
    [anon_sym_signed] = ACTIONS(2364),
    [anon_sym_unsigned] = ACTIONS(2364),
    [anon_sym_long] = ACTIONS(2364),
    [anon_sym_short] = ACTIONS(2364),
    [anon_sym_static] = ACTIONS(2364),
    [anon_sym_ATautoreleasepool] = ACTIONS(2366),
    [anon_sym_auto] = ACTIONS(2364),
    [anon_sym_register] = ACTIONS(2364),
    [anon_sym_inline] = ACTIONS(2364),
    [anon_sym___inline] = ACTIONS(2364),
    [anon_sym___inline__] = ACTIONS(2364),
    [anon_sym___forceinline] = ACTIONS(2364),
    [anon_sym_thread_local] = ACTIONS(2364),
    [anon_sym___thread] = ACTIONS(2364),
    [anon_sym_CG_EXTERN] = ACTIONS(2364),
    [anon_sym_CG_INLINE] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2364),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2364),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2364),
    [anon_sym_IBOutlet] = ACTIONS(2364),
    [anon_sym_IBInspectable] = ACTIONS(2364),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2364),
    [anon_sym_NS_INLINE] = ACTIONS(2364),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2364),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2364),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2364),
    [anon_sym_const] = ACTIONS(2364),
    [anon_sym_constexpr] = ACTIONS(2364),
    [anon_sym_volatile] = ACTIONS(2364),
    [anon_sym_restrict] = ACTIONS(2364),
    [anon_sym___restrict__] = ACTIONS(2364),
    [anon_sym__Atomic] = ACTIONS(2364),
    [anon_sym__Noreturn] = ACTIONS(2364),
    [anon_sym__Nonnull] = ACTIONS(2364),
    [anon_sym_nullable] = ACTIONS(2364),
    [anon_sym__Complex] = ACTIONS(2364),
    [anon_sym__Nullable] = ACTIONS(2364),
    [anon_sym__Nullable_result] = ACTIONS(2364),
    [anon_sym__Null_unspecified] = ACTIONS(2364),
    [anon_sym___autoreleasing] = ACTIONS(2364),
    [anon_sym___block] = ACTIONS(2364),
    [anon_sym___bridge] = ACTIONS(2364),
    [anon_sym___bridge_retained] = ACTIONS(2364),
    [anon_sym___bridge_transfer] = ACTIONS(2364),
    [anon_sym___complex] = ACTIONS(2364),
    [anon_sym___const] = ACTIONS(2364),
    [anon_sym___imag] = ACTIONS(2364),
    [anon_sym___kindof] = ACTIONS(2364),
    [anon_sym___nonnull] = ACTIONS(2364),
    [anon_sym___nullable] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2364),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2364),
    [anon_sym___real] = ACTIONS(2364),
    [anon_sym___strong] = ACTIONS(2364),
    [anon_sym___unsafe_unretained] = ACTIONS(2364),
    [anon_sym___unused] = ACTIONS(2364),
    [anon_sym___weak] = ACTIONS(2364),
    [anon_sym_alignas] = ACTIONS(2364),
    [anon_sym__Alignas] = ACTIONS(2364),
    [sym_primitive_type] = ACTIONS(2364),
    [anon_sym_enum] = ACTIONS(2364),
    [anon_sym_struct] = ACTIONS(2364),
    [anon_sym_union] = ACTIONS(2364),
    [anon_sym_if] = ACTIONS(2364),
    [anon_sym_else] = ACTIONS(2364),
    [anon_sym_switch] = ACTIONS(2364),
    [anon_sym_while] = ACTIONS(2364),
    [anon_sym_do] = ACTIONS(2364),
    [anon_sym_for] = ACTIONS(2364),
    [anon_sym_in] = ACTIONS(2364),
    [anon_sym_return] = ACTIONS(2364),
    [anon_sym_break] = ACTIONS(2364),
    [anon_sym_continue] = ACTIONS(2364),
    [anon_sym_goto] = ACTIONS(2364),
    [anon_sym_DASH_DASH] = ACTIONS(2366),
    [anon_sym_PLUS_PLUS] = ACTIONS(2366),
    [anon_sym_sizeof] = ACTIONS(2364),
    [anon_sym___alignof__] = ACTIONS(2364),
    [anon_sym___alignof] = ACTIONS(2364),
    [anon_sym__alignof] = ACTIONS(2364),
    [anon_sym_alignof] = ACTIONS(2364),
    [anon_sym__Alignof] = ACTIONS(2364),
    [anon_sym_offsetof] = ACTIONS(2364),
    [anon_sym__Generic] = ACTIONS(2364),
    [anon_sym_asm] = ACTIONS(2364),
    [anon_sym___asm__] = ACTIONS(2364),
    [anon_sym___asm] = ACTIONS(2364),
    [sym_number_literal] = ACTIONS(2366),
    [anon_sym_L_SQUOTE] = ACTIONS(2366),
    [anon_sym_u_SQUOTE] = ACTIONS(2366),
    [anon_sym_U_SQUOTE] = ACTIONS(2366),
    [anon_sym_u8_SQUOTE] = ACTIONS(2366),
    [anon_sym_SQUOTE] = ACTIONS(2366),
    [anon_sym_AT] = ACTIONS(2364),
    [anon_sym_DQUOTE] = ACTIONS(2366),
    [anon_sym_L_DQUOTE] = ACTIONS(2366),
    [anon_sym_u_DQUOTE] = ACTIONS(2366),
    [anon_sym_U_DQUOTE] = ACTIONS(2366),
    [anon_sym_u8_DQUOTE] = ACTIONS(2366),
    [sym_true] = ACTIONS(2364),
    [sym_false] = ACTIONS(2364),
    [anon_sym_NULL] = ACTIONS(2364),
    [anon_sym_nullptr] = ACTIONS(2364),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2364),
    [anon_sym___typeof] = ACTIONS(2364),
    [anon_sym_typeof] = ACTIONS(2364),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2364),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2364),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2364),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2364),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2364),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_API_AVAILABLE] = ACTIONS(2364),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_API_DEPRECATED] = ACTIONS(2364),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2364),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2364),
    [anon_sym___deprecated_msg] = ACTIONS(2364),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2364),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2364),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2364),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2364),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2364),
    [anon_sym_ATtry] = ACTIONS(2366),
    [anon_sym___try] = ACTIONS(2364),
    [anon_sym_ATcatch] = ACTIONS(2366),
    [anon_sym___catch] = ACTIONS(2364),
    [anon_sym_ATfinally] = ACTIONS(2366),
    [anon_sym___finally] = ACTIONS(2364),
    [anon_sym_ATthrow] = ACTIONS(2366),
    [anon_sym_ATselector] = ACTIONS(2366),
    [anon_sym_ATavailable] = ACTIONS(2366),
    [anon_sym___builtin_available] = ACTIONS(2364),
    [anon_sym_va_arg] = ACTIONS(2364),
    [anon_sym_ATencode] = ACTIONS(2366),
    [anon_sym_ATsynchronized] = ACTIONS(2366),
    [anon_sym_BOOL] = ACTIONS(2364),
    [anon_sym_IMP] = ACTIONS(2364),
    [anon_sym_SEL] = ACTIONS(2364),
    [anon_sym_Class] = ACTIONS(2364),
    [anon_sym_id] = ACTIONS(2364),
  },
  [STATE(1083)] = {
    [sym_identifier] = ACTIONS(2980),
    [anon_sym_LPAREN2] = ACTIONS(2983),
    [anon_sym_BANG] = ACTIONS(2983),
    [anon_sym_TILDE] = ACTIONS(2983),
    [anon_sym_DASH] = ACTIONS(2985),
    [anon_sym_PLUS] = ACTIONS(2985),
    [anon_sym_STAR] = ACTIONS(2983),
    [anon_sym_CARET] = ACTIONS(2983),
    [anon_sym_AMP] = ACTIONS(2983),
    [anon_sym_SEMI] = ACTIONS(2983),
    [anon_sym___extension__] = ACTIONS(2980),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2989),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_LBRACE] = ACTIONS(2983),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_ATautoreleasepool] = ACTIONS(2983),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2980),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2980),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2980),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2980),
    [anon_sym_union] = ACTIONS(2987),
    [anon_sym_if] = ACTIONS(2985),
    [anon_sym_switch] = ACTIONS(2985),
    [anon_sym_case] = ACTIONS(2985),
    [anon_sym_default] = ACTIONS(2985),
    [anon_sym_while] = ACTIONS(2985),
    [anon_sym_do] = ACTIONS(2985),
    [anon_sym_for] = ACTIONS(2985),
    [anon_sym_in] = ACTIONS(2985),
    [anon_sym_return] = ACTIONS(2985),
    [anon_sym_break] = ACTIONS(2985),
    [anon_sym_continue] = ACTIONS(2985),
    [anon_sym_goto] = ACTIONS(2985),
    [anon_sym_DASH_DASH] = ACTIONS(2983),
    [anon_sym_PLUS_PLUS] = ACTIONS(2983),
    [anon_sym_sizeof] = ACTIONS(2985),
    [anon_sym___alignof__] = ACTIONS(2985),
    [anon_sym___alignof] = ACTIONS(2985),
    [anon_sym__alignof] = ACTIONS(2985),
    [anon_sym_alignof] = ACTIONS(2985),
    [anon_sym__Alignof] = ACTIONS(2985),
    [anon_sym_offsetof] = ACTIONS(2985),
    [anon_sym__Generic] = ACTIONS(2985),
    [anon_sym_asm] = ACTIONS(2985),
    [anon_sym___asm__] = ACTIONS(2985),
    [anon_sym___asm] = ACTIONS(2985),
    [sym_number_literal] = ACTIONS(2983),
    [anon_sym_L_SQUOTE] = ACTIONS(2983),
    [anon_sym_u_SQUOTE] = ACTIONS(2983),
    [anon_sym_U_SQUOTE] = ACTIONS(2983),
    [anon_sym_u8_SQUOTE] = ACTIONS(2983),
    [anon_sym_SQUOTE] = ACTIONS(2983),
    [anon_sym_AT] = ACTIONS(2985),
    [anon_sym_DQUOTE] = ACTIONS(2983),
    [anon_sym_L_DQUOTE] = ACTIONS(2983),
    [anon_sym_u_DQUOTE] = ACTIONS(2983),
    [anon_sym_U_DQUOTE] = ACTIONS(2983),
    [anon_sym_u8_DQUOTE] = ACTIONS(2983),
    [sym_true] = ACTIONS(2985),
    [sym_false] = ACTIONS(2985),
    [anon_sym_NULL] = ACTIONS(2985),
    [anon_sym_nullptr] = ACTIONS(2985),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_ATprotocol] = ACTIONS(2992),
    [anon_sym_ATinterface] = ACTIONS(2992),
    [anon_sym_ATimplementation] = ACTIONS(2992),
    [anon_sym_ATtry] = ACTIONS(2983),
    [anon_sym___try] = ACTIONS(2985),
    [anon_sym_ATthrow] = ACTIONS(2983),
    [anon_sym_ATselector] = ACTIONS(2983),
    [anon_sym_ATavailable] = ACTIONS(2983),
    [anon_sym___builtin_available] = ACTIONS(2985),
    [anon_sym_va_arg] = ACTIONS(2985),
    [anon_sym_ATencode] = ACTIONS(2983),
    [anon_sym_ATsynchronized] = ACTIONS(2983),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2980),
  },
  [STATE(1084)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym__type_params] = STATE(1179),
    [sym__class_implementation_inheritance] = STATE(1210),
    [sym_parameterized_arguments] = STATE(1179),
    [sym_generic_arguments] = STATE(1970),
    [sym_instance_variables] = STATE(1262),
    [sym_implementation_definition] = STATE(1265),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1265),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3004),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3018),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3024),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1085)] = {
    [sym_else_clause] = STATE(1144),
    [sym_identifier] = ACTIONS(2394),
    [anon_sym_LPAREN2] = ACTIONS(2396),
    [anon_sym_BANG] = ACTIONS(2396),
    [anon_sym_TILDE] = ACTIONS(2396),
    [anon_sym_DASH] = ACTIONS(2394),
    [anon_sym_PLUS] = ACTIONS(2394),
    [anon_sym_STAR] = ACTIONS(2396),
    [anon_sym_CARET] = ACTIONS(2396),
    [anon_sym_AMP] = ACTIONS(2396),
    [anon_sym_SEMI] = ACTIONS(2396),
    [anon_sym___extension__] = ACTIONS(2394),
    [anon_sym_typedef] = ACTIONS(2394),
    [anon_sym_extern] = ACTIONS(2394),
    [anon_sym___attribute__] = ACTIONS(2394),
    [anon_sym___attribute] = ACTIONS(2394),
    [anon_sym_noreturn] = ACTIONS(2394),
    [anon_sym_LBRACK] = ACTIONS(2396),
    [anon_sym___declspec] = ACTIONS(2394),
    [anon_sym_LBRACE] = ACTIONS(2396),
    [anon_sym_signed] = ACTIONS(2394),
    [anon_sym_unsigned] = ACTIONS(2394),
    [anon_sym_long] = ACTIONS(2394),
    [anon_sym_short] = ACTIONS(2394),
    [anon_sym_static] = ACTIONS(2394),
    [anon_sym_ATautoreleasepool] = ACTIONS(2396),
    [anon_sym_auto] = ACTIONS(2394),
    [anon_sym_register] = ACTIONS(2394),
    [anon_sym_inline] = ACTIONS(2394),
    [anon_sym___inline] = ACTIONS(2394),
    [anon_sym___inline__] = ACTIONS(2394),
    [anon_sym___forceinline] = ACTIONS(2394),
    [anon_sym_thread_local] = ACTIONS(2394),
    [anon_sym___thread] = ACTIONS(2394),
    [anon_sym_CG_EXTERN] = ACTIONS(2394),
    [anon_sym_CG_INLINE] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2394),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2394),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2394),
    [anon_sym_IBOutlet] = ACTIONS(2394),
    [anon_sym_IBInspectable] = ACTIONS(2394),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2394),
    [anon_sym_NS_INLINE] = ACTIONS(2394),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2394),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2394),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2394),
    [anon_sym_const] = ACTIONS(2394),
    [anon_sym_constexpr] = ACTIONS(2394),
    [anon_sym_volatile] = ACTIONS(2394),
    [anon_sym_restrict] = ACTIONS(2394),
    [anon_sym___restrict__] = ACTIONS(2394),
    [anon_sym__Atomic] = ACTIONS(2394),
    [anon_sym__Noreturn] = ACTIONS(2394),
    [anon_sym__Nonnull] = ACTIONS(2394),
    [anon_sym_nullable] = ACTIONS(2394),
    [anon_sym__Complex] = ACTIONS(2394),
    [anon_sym__Nullable] = ACTIONS(2394),
    [anon_sym__Nullable_result] = ACTIONS(2394),
    [anon_sym__Null_unspecified] = ACTIONS(2394),
    [anon_sym___autoreleasing] = ACTIONS(2394),
    [anon_sym___block] = ACTIONS(2394),
    [anon_sym___bridge] = ACTIONS(2394),
    [anon_sym___bridge_retained] = ACTIONS(2394),
    [anon_sym___bridge_transfer] = ACTIONS(2394),
    [anon_sym___complex] = ACTIONS(2394),
    [anon_sym___const] = ACTIONS(2394),
    [anon_sym___imag] = ACTIONS(2394),
    [anon_sym___kindof] = ACTIONS(2394),
    [anon_sym___nonnull] = ACTIONS(2394),
    [anon_sym___nullable] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2394),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2394),
    [anon_sym___real] = ACTIONS(2394),
    [anon_sym___strong] = ACTIONS(2394),
    [anon_sym___unsafe_unretained] = ACTIONS(2394),
    [anon_sym___unused] = ACTIONS(2394),
    [anon_sym___weak] = ACTIONS(2394),
    [anon_sym_alignas] = ACTIONS(2394),
    [anon_sym__Alignas] = ACTIONS(2394),
    [sym_primitive_type] = ACTIONS(2394),
    [anon_sym_enum] = ACTIONS(2394),
    [anon_sym_struct] = ACTIONS(2394),
    [anon_sym_union] = ACTIONS(2394),
    [anon_sym_if] = ACTIONS(2394),
    [anon_sym_else] = ACTIONS(3030),
    [anon_sym_switch] = ACTIONS(2394),
    [anon_sym_while] = ACTIONS(2394),
    [anon_sym_do] = ACTIONS(2394),
    [anon_sym_for] = ACTIONS(2394),
    [anon_sym_in] = ACTIONS(2394),
    [anon_sym_return] = ACTIONS(2394),
    [anon_sym_break] = ACTIONS(2394),
    [anon_sym_continue] = ACTIONS(2394),
    [anon_sym_goto] = ACTIONS(2394),
    [anon_sym_DASH_DASH] = ACTIONS(2396),
    [anon_sym_PLUS_PLUS] = ACTIONS(2396),
    [anon_sym_sizeof] = ACTIONS(2394),
    [anon_sym___alignof__] = ACTIONS(2394),
    [anon_sym___alignof] = ACTIONS(2394),
    [anon_sym__alignof] = ACTIONS(2394),
    [anon_sym_alignof] = ACTIONS(2394),
    [anon_sym__Alignof] = ACTIONS(2394),
    [anon_sym_offsetof] = ACTIONS(2394),
    [anon_sym__Generic] = ACTIONS(2394),
    [anon_sym_asm] = ACTIONS(2394),
    [anon_sym___asm__] = ACTIONS(2394),
    [anon_sym___asm] = ACTIONS(2394),
    [sym_number_literal] = ACTIONS(2396),
    [anon_sym_L_SQUOTE] = ACTIONS(2396),
    [anon_sym_u_SQUOTE] = ACTIONS(2396),
    [anon_sym_U_SQUOTE] = ACTIONS(2396),
    [anon_sym_u8_SQUOTE] = ACTIONS(2396),
    [anon_sym_SQUOTE] = ACTIONS(2396),
    [anon_sym_AT] = ACTIONS(2394),
    [anon_sym_DQUOTE] = ACTIONS(2396),
    [anon_sym_L_DQUOTE] = ACTIONS(2396),
    [anon_sym_u_DQUOTE] = ACTIONS(2396),
    [anon_sym_U_DQUOTE] = ACTIONS(2396),
    [anon_sym_u8_DQUOTE] = ACTIONS(2396),
    [sym_true] = ACTIONS(2394),
    [sym_false] = ACTIONS(2394),
    [anon_sym_NULL] = ACTIONS(2394),
    [anon_sym_nullptr] = ACTIONS(2394),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2394),
    [anon_sym___typeof] = ACTIONS(2394),
    [anon_sym_typeof] = ACTIONS(2394),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2394),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2394),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2394),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2394),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2394),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_API_AVAILABLE] = ACTIONS(2394),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_API_DEPRECATED] = ACTIONS(2394),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2394),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2394),
    [anon_sym___deprecated_msg] = ACTIONS(2394),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2394),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2394),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2394),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2394),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2394),
    [anon_sym_ATtry] = ACTIONS(2396),
    [anon_sym___try] = ACTIONS(2394),
    [anon_sym_ATthrow] = ACTIONS(2396),
    [anon_sym_ATselector] = ACTIONS(2396),
    [anon_sym_ATavailable] = ACTIONS(2396),
    [anon_sym___builtin_available] = ACTIONS(2394),
    [anon_sym_va_arg] = ACTIONS(2394),
    [anon_sym_ATencode] = ACTIONS(2396),
    [anon_sym_ATsynchronized] = ACTIONS(2396),
    [anon_sym_BOOL] = ACTIONS(2394),
    [anon_sym_IMP] = ACTIONS(2394),
    [anon_sym_SEL] = ACTIONS(2394),
    [anon_sym_Class] = ACTIONS(2394),
    [anon_sym_id] = ACTIONS(2394),
  },
  [STATE(1086)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym__type_params] = STATE(1183),
    [sym__class_implementation_inheritance] = STATE(1205),
    [sym_parameterized_arguments] = STATE(1183),
    [sym_generic_arguments] = STATE(1970),
    [sym_instance_variables] = STATE(1287),
    [sym_implementation_definition] = STATE(1288),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1288),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3004),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3018),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3032),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1087)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym__type_params] = STATE(1181),
    [sym__class_implementation_inheritance] = STATE(1197),
    [sym_parameterized_arguments] = STATE(1181),
    [sym_generic_arguments] = STATE(1970),
    [sym_instance_variables] = STATE(1283),
    [sym_implementation_definition] = STATE(1258),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1258),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3004),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3018),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3034),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1088)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym__type_params] = STATE(1182),
    [sym__class_implementation_inheritance] = STATE(1212),
    [sym_parameterized_arguments] = STATE(1182),
    [sym_generic_arguments] = STATE(1970),
    [sym_instance_variables] = STATE(1269),
    [sym_implementation_definition] = STATE(1309),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1309),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3004),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3018),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3036),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1089)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym__type_params] = STATE(1180),
    [sym__class_implementation_inheritance] = STATE(1204),
    [sym_parameterized_arguments] = STATE(1180),
    [sym_generic_arguments] = STATE(1970),
    [sym_instance_variables] = STATE(1263),
    [sym_implementation_definition] = STATE(1264),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1264),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3004),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3018),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3038),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1090)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1091)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1092)] = {
    [sym_identifier] = ACTIONS(2470),
    [anon_sym_LPAREN2] = ACTIONS(2472),
    [anon_sym_BANG] = ACTIONS(2472),
    [anon_sym_TILDE] = ACTIONS(2472),
    [anon_sym_DASH] = ACTIONS(2470),
    [anon_sym_PLUS] = ACTIONS(2470),
    [anon_sym_STAR] = ACTIONS(2472),
    [anon_sym_CARET] = ACTIONS(2472),
    [anon_sym_AMP] = ACTIONS(2472),
    [anon_sym_SEMI] = ACTIONS(2472),
    [anon_sym___extension__] = ACTIONS(2470),
    [anon_sym_typedef] = ACTIONS(2470),
    [anon_sym_extern] = ACTIONS(2470),
    [anon_sym___attribute__] = ACTIONS(2470),
    [anon_sym___attribute] = ACTIONS(2470),
    [anon_sym_noreturn] = ACTIONS(2470),
    [anon_sym_LBRACK] = ACTIONS(2472),
    [anon_sym___declspec] = ACTIONS(2470),
    [anon_sym_LBRACE] = ACTIONS(2472),
    [anon_sym_signed] = ACTIONS(2470),
    [anon_sym_unsigned] = ACTIONS(2470),
    [anon_sym_long] = ACTIONS(2470),
    [anon_sym_short] = ACTIONS(2470),
    [anon_sym_static] = ACTIONS(2470),
    [anon_sym_ATautoreleasepool] = ACTIONS(2472),
    [anon_sym_auto] = ACTIONS(2470),
    [anon_sym_register] = ACTIONS(2470),
    [anon_sym_inline] = ACTIONS(2470),
    [anon_sym___inline] = ACTIONS(2470),
    [anon_sym___inline__] = ACTIONS(2470),
    [anon_sym___forceinline] = ACTIONS(2470),
    [anon_sym_thread_local] = ACTIONS(2470),
    [anon_sym___thread] = ACTIONS(2470),
    [anon_sym_CG_EXTERN] = ACTIONS(2470),
    [anon_sym_CG_INLINE] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2470),
    [anon_sym_IBOutlet] = ACTIONS(2470),
    [anon_sym_IBInspectable] = ACTIONS(2470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2470),
    [anon_sym_NS_INLINE] = ACTIONS(2470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2470),
    [anon_sym_const] = ACTIONS(2470),
    [anon_sym_constexpr] = ACTIONS(2470),
    [anon_sym_volatile] = ACTIONS(2470),
    [anon_sym_restrict] = ACTIONS(2470),
    [anon_sym___restrict__] = ACTIONS(2470),
    [anon_sym__Atomic] = ACTIONS(2470),
    [anon_sym__Noreturn] = ACTIONS(2470),
    [anon_sym__Nonnull] = ACTIONS(2470),
    [anon_sym_nullable] = ACTIONS(2470),
    [anon_sym__Complex] = ACTIONS(2470),
    [anon_sym__Nullable] = ACTIONS(2470),
    [anon_sym__Nullable_result] = ACTIONS(2470),
    [anon_sym__Null_unspecified] = ACTIONS(2470),
    [anon_sym___autoreleasing] = ACTIONS(2470),
    [anon_sym___block] = ACTIONS(2470),
    [anon_sym___bridge] = ACTIONS(2470),
    [anon_sym___bridge_retained] = ACTIONS(2470),
    [anon_sym___bridge_transfer] = ACTIONS(2470),
    [anon_sym___complex] = ACTIONS(2470),
    [anon_sym___const] = ACTIONS(2470),
    [anon_sym___imag] = ACTIONS(2470),
    [anon_sym___kindof] = ACTIONS(2470),
    [anon_sym___nonnull] = ACTIONS(2470),
    [anon_sym___nullable] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2470),
    [anon_sym___real] = ACTIONS(2470),
    [anon_sym___strong] = ACTIONS(2470),
    [anon_sym___unsafe_unretained] = ACTIONS(2470),
    [anon_sym___unused] = ACTIONS(2470),
    [anon_sym___weak] = ACTIONS(2470),
    [anon_sym_alignas] = ACTIONS(2470),
    [anon_sym__Alignas] = ACTIONS(2470),
    [sym_primitive_type] = ACTIONS(2470),
    [anon_sym_enum] = ACTIONS(2470),
    [anon_sym_struct] = ACTIONS(2470),
    [anon_sym_union] = ACTIONS(2470),
    [anon_sym_if] = ACTIONS(2470),
    [anon_sym_else] = ACTIONS(2470),
    [anon_sym_switch] = ACTIONS(2470),
    [anon_sym_while] = ACTIONS(2470),
    [anon_sym_do] = ACTIONS(2470),
    [anon_sym_for] = ACTIONS(2470),
    [anon_sym_in] = ACTIONS(2470),
    [anon_sym_return] = ACTIONS(2470),
    [anon_sym_break] = ACTIONS(2470),
    [anon_sym_continue] = ACTIONS(2470),
    [anon_sym_goto] = ACTIONS(2470),
    [anon_sym_DASH_DASH] = ACTIONS(2472),
    [anon_sym_PLUS_PLUS] = ACTIONS(2472),
    [anon_sym_sizeof] = ACTIONS(2470),
    [anon_sym___alignof__] = ACTIONS(2470),
    [anon_sym___alignof] = ACTIONS(2470),
    [anon_sym__alignof] = ACTIONS(2470),
    [anon_sym_alignof] = ACTIONS(2470),
    [anon_sym__Alignof] = ACTIONS(2470),
    [anon_sym_offsetof] = ACTIONS(2470),
    [anon_sym__Generic] = ACTIONS(2470),
    [anon_sym_asm] = ACTIONS(2470),
    [anon_sym___asm__] = ACTIONS(2470),
    [anon_sym___asm] = ACTIONS(2470),
    [sym_number_literal] = ACTIONS(2472),
    [anon_sym_L_SQUOTE] = ACTIONS(2472),
    [anon_sym_u_SQUOTE] = ACTIONS(2472),
    [anon_sym_U_SQUOTE] = ACTIONS(2472),
    [anon_sym_u8_SQUOTE] = ACTIONS(2472),
    [anon_sym_SQUOTE] = ACTIONS(2472),
    [anon_sym_AT] = ACTIONS(2470),
    [anon_sym_DQUOTE] = ACTIONS(2472),
    [anon_sym_L_DQUOTE] = ACTIONS(2472),
    [anon_sym_u_DQUOTE] = ACTIONS(2472),
    [anon_sym_U_DQUOTE] = ACTIONS(2472),
    [anon_sym_u8_DQUOTE] = ACTIONS(2472),
    [sym_true] = ACTIONS(2470),
    [sym_false] = ACTIONS(2470),
    [anon_sym_NULL] = ACTIONS(2470),
    [anon_sym_nullptr] = ACTIONS(2470),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2470),
    [anon_sym___typeof] = ACTIONS(2470),
    [anon_sym_typeof] = ACTIONS(2470),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_API_AVAILABLE] = ACTIONS(2470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_API_DEPRECATED] = ACTIONS(2470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2470),
    [anon_sym___deprecated_msg] = ACTIONS(2470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2470),
    [anon_sym_ATtry] = ACTIONS(2472),
    [anon_sym___try] = ACTIONS(2470),
    [anon_sym_ATthrow] = ACTIONS(2472),
    [anon_sym_ATselector] = ACTIONS(2472),
    [anon_sym_ATavailable] = ACTIONS(2472),
    [anon_sym___builtin_available] = ACTIONS(2470),
    [anon_sym_va_arg] = ACTIONS(2470),
    [anon_sym_ATencode] = ACTIONS(2472),
    [anon_sym_ATsynchronized] = ACTIONS(2472),
    [anon_sym_BOOL] = ACTIONS(2470),
    [anon_sym_IMP] = ACTIONS(2470),
    [anon_sym_SEL] = ACTIONS(2470),
    [anon_sym_Class] = ACTIONS(2470),
    [anon_sym_id] = ACTIONS(2470),
  },
  [STATE(1093)] = {
    [sym_identifier] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(1094)] = {
    [sym_identifier] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(1095)] = {
    [sym_identifier] = ACTIONS(2428),
    [anon_sym_LPAREN2] = ACTIONS(2430),
    [anon_sym_BANG] = ACTIONS(2430),
    [anon_sym_TILDE] = ACTIONS(2430),
    [anon_sym_DASH] = ACTIONS(2428),
    [anon_sym_PLUS] = ACTIONS(2428),
    [anon_sym_STAR] = ACTIONS(2430),
    [anon_sym_CARET] = ACTIONS(2430),
    [anon_sym_AMP] = ACTIONS(2430),
    [anon_sym_SEMI] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym_LBRACE] = ACTIONS(2430),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_ATautoreleasepool] = ACTIONS(2430),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [anon_sym_if] = ACTIONS(2428),
    [anon_sym_else] = ACTIONS(2428),
    [anon_sym_switch] = ACTIONS(2428),
    [anon_sym_while] = ACTIONS(2428),
    [anon_sym_do] = ACTIONS(2428),
    [anon_sym_for] = ACTIONS(2428),
    [anon_sym_in] = ACTIONS(2428),
    [anon_sym_return] = ACTIONS(2428),
    [anon_sym_break] = ACTIONS(2428),
    [anon_sym_continue] = ACTIONS(2428),
    [anon_sym_goto] = ACTIONS(2428),
    [anon_sym_DASH_DASH] = ACTIONS(2430),
    [anon_sym_PLUS_PLUS] = ACTIONS(2430),
    [anon_sym_sizeof] = ACTIONS(2428),
    [anon_sym___alignof__] = ACTIONS(2428),
    [anon_sym___alignof] = ACTIONS(2428),
    [anon_sym__alignof] = ACTIONS(2428),
    [anon_sym_alignof] = ACTIONS(2428),
    [anon_sym__Alignof] = ACTIONS(2428),
    [anon_sym_offsetof] = ACTIONS(2428),
    [anon_sym__Generic] = ACTIONS(2428),
    [anon_sym_asm] = ACTIONS(2428),
    [anon_sym___asm__] = ACTIONS(2428),
    [anon_sym___asm] = ACTIONS(2428),
    [sym_number_literal] = ACTIONS(2430),
    [anon_sym_L_SQUOTE] = ACTIONS(2430),
    [anon_sym_u_SQUOTE] = ACTIONS(2430),
    [anon_sym_U_SQUOTE] = ACTIONS(2430),
    [anon_sym_u8_SQUOTE] = ACTIONS(2430),
    [anon_sym_SQUOTE] = ACTIONS(2430),
    [anon_sym_AT] = ACTIONS(2428),
    [anon_sym_DQUOTE] = ACTIONS(2430),
    [anon_sym_L_DQUOTE] = ACTIONS(2430),
    [anon_sym_u_DQUOTE] = ACTIONS(2430),
    [anon_sym_U_DQUOTE] = ACTIONS(2430),
    [anon_sym_u8_DQUOTE] = ACTIONS(2430),
    [sym_true] = ACTIONS(2428),
    [sym_false] = ACTIONS(2428),
    [anon_sym_NULL] = ACTIONS(2428),
    [anon_sym_nullptr] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATtry] = ACTIONS(2430),
    [anon_sym___try] = ACTIONS(2428),
    [anon_sym_ATthrow] = ACTIONS(2430),
    [anon_sym_ATselector] = ACTIONS(2430),
    [anon_sym_ATavailable] = ACTIONS(2430),
    [anon_sym___builtin_available] = ACTIONS(2428),
    [anon_sym_va_arg] = ACTIONS(2428),
    [anon_sym_ATencode] = ACTIONS(2430),
    [anon_sym_ATsynchronized] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(1096)] = {
    [sym_identifier] = ACTIONS(2568),
    [anon_sym_LPAREN2] = ACTIONS(2570),
    [anon_sym_BANG] = ACTIONS(2570),
    [anon_sym_TILDE] = ACTIONS(2570),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_PLUS] = ACTIONS(2568),
    [anon_sym_STAR] = ACTIONS(2570),
    [anon_sym_CARET] = ACTIONS(2570),
    [anon_sym_AMP] = ACTIONS(2570),
    [anon_sym_SEMI] = ACTIONS(2570),
    [anon_sym___extension__] = ACTIONS(2568),
    [anon_sym_typedef] = ACTIONS(2568),
    [anon_sym_extern] = ACTIONS(2568),
    [anon_sym___attribute__] = ACTIONS(2568),
    [anon_sym___attribute] = ACTIONS(2568),
    [anon_sym_noreturn] = ACTIONS(2568),
    [anon_sym_LBRACK] = ACTIONS(2570),
    [anon_sym___declspec] = ACTIONS(2568),
    [anon_sym_LBRACE] = ACTIONS(2570),
    [anon_sym_signed] = ACTIONS(2568),
    [anon_sym_unsigned] = ACTIONS(2568),
    [anon_sym_long] = ACTIONS(2568),
    [anon_sym_short] = ACTIONS(2568),
    [anon_sym_static] = ACTIONS(2568),
    [anon_sym_ATautoreleasepool] = ACTIONS(2570),
    [anon_sym_auto] = ACTIONS(2568),
    [anon_sym_register] = ACTIONS(2568),
    [anon_sym_inline] = ACTIONS(2568),
    [anon_sym___inline] = ACTIONS(2568),
    [anon_sym___inline__] = ACTIONS(2568),
    [anon_sym___forceinline] = ACTIONS(2568),
    [anon_sym_thread_local] = ACTIONS(2568),
    [anon_sym___thread] = ACTIONS(2568),
    [anon_sym_CG_EXTERN] = ACTIONS(2568),
    [anon_sym_CG_INLINE] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2568),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2568),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2568),
    [anon_sym_IBOutlet] = ACTIONS(2568),
    [anon_sym_IBInspectable] = ACTIONS(2568),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2568),
    [anon_sym_NS_INLINE] = ACTIONS(2568),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2568),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2568),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2568),
    [anon_sym_const] = ACTIONS(2568),
    [anon_sym_constexpr] = ACTIONS(2568),
    [anon_sym_volatile] = ACTIONS(2568),
    [anon_sym_restrict] = ACTIONS(2568),
    [anon_sym___restrict__] = ACTIONS(2568),
    [anon_sym__Atomic] = ACTIONS(2568),
    [anon_sym__Noreturn] = ACTIONS(2568),
    [anon_sym__Nonnull] = ACTIONS(2568),
    [anon_sym_nullable] = ACTIONS(2568),
    [anon_sym__Complex] = ACTIONS(2568),
    [anon_sym__Nullable] = ACTIONS(2568),
    [anon_sym__Nullable_result] = ACTIONS(2568),
    [anon_sym__Null_unspecified] = ACTIONS(2568),
    [anon_sym___autoreleasing] = ACTIONS(2568),
    [anon_sym___block] = ACTIONS(2568),
    [anon_sym___bridge] = ACTIONS(2568),
    [anon_sym___bridge_retained] = ACTIONS(2568),
    [anon_sym___bridge_transfer] = ACTIONS(2568),
    [anon_sym___complex] = ACTIONS(2568),
    [anon_sym___const] = ACTIONS(2568),
    [anon_sym___imag] = ACTIONS(2568),
    [anon_sym___kindof] = ACTIONS(2568),
    [anon_sym___nonnull] = ACTIONS(2568),
    [anon_sym___nullable] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2568),
    [anon_sym___real] = ACTIONS(2568),
    [anon_sym___strong] = ACTIONS(2568),
    [anon_sym___unsafe_unretained] = ACTIONS(2568),
    [anon_sym___unused] = ACTIONS(2568),
    [anon_sym___weak] = ACTIONS(2568),
    [anon_sym_alignas] = ACTIONS(2568),
    [anon_sym__Alignas] = ACTIONS(2568),
    [sym_primitive_type] = ACTIONS(2568),
    [anon_sym_enum] = ACTIONS(2568),
    [anon_sym_struct] = ACTIONS(2568),
    [anon_sym_union] = ACTIONS(2568),
    [anon_sym_if] = ACTIONS(2568),
    [anon_sym_else] = ACTIONS(2568),
    [anon_sym_switch] = ACTIONS(2568),
    [anon_sym_while] = ACTIONS(2568),
    [anon_sym_do] = ACTIONS(2568),
    [anon_sym_for] = ACTIONS(2568),
    [anon_sym_in] = ACTIONS(2568),
    [anon_sym_return] = ACTIONS(2568),
    [anon_sym_break] = ACTIONS(2568),
    [anon_sym_continue] = ACTIONS(2568),
    [anon_sym_goto] = ACTIONS(2568),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_sizeof] = ACTIONS(2568),
    [anon_sym___alignof__] = ACTIONS(2568),
    [anon_sym___alignof] = ACTIONS(2568),
    [anon_sym__alignof] = ACTIONS(2568),
    [anon_sym_alignof] = ACTIONS(2568),
    [anon_sym__Alignof] = ACTIONS(2568),
    [anon_sym_offsetof] = ACTIONS(2568),
    [anon_sym__Generic] = ACTIONS(2568),
    [anon_sym_asm] = ACTIONS(2568),
    [anon_sym___asm__] = ACTIONS(2568),
    [anon_sym___asm] = ACTIONS(2568),
    [sym_number_literal] = ACTIONS(2570),
    [anon_sym_L_SQUOTE] = ACTIONS(2570),
    [anon_sym_u_SQUOTE] = ACTIONS(2570),
    [anon_sym_U_SQUOTE] = ACTIONS(2570),
    [anon_sym_u8_SQUOTE] = ACTIONS(2570),
    [anon_sym_SQUOTE] = ACTIONS(2570),
    [anon_sym_AT] = ACTIONS(2568),
    [anon_sym_DQUOTE] = ACTIONS(2570),
    [anon_sym_L_DQUOTE] = ACTIONS(2570),
    [anon_sym_u_DQUOTE] = ACTIONS(2570),
    [anon_sym_U_DQUOTE] = ACTIONS(2570),
    [anon_sym_u8_DQUOTE] = ACTIONS(2570),
    [sym_true] = ACTIONS(2568),
    [sym_false] = ACTIONS(2568),
    [anon_sym_NULL] = ACTIONS(2568),
    [anon_sym_nullptr] = ACTIONS(2568),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2568),
    [anon_sym___typeof] = ACTIONS(2568),
    [anon_sym_typeof] = ACTIONS(2568),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2568),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2568),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2568),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2568),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2568),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_API_AVAILABLE] = ACTIONS(2568),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_API_DEPRECATED] = ACTIONS(2568),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2568),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2568),
    [anon_sym___deprecated_msg] = ACTIONS(2568),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2568),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2568),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2568),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2568),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2568),
    [anon_sym_ATtry] = ACTIONS(2570),
    [anon_sym___try] = ACTIONS(2568),
    [anon_sym_ATthrow] = ACTIONS(2570),
    [anon_sym_ATselector] = ACTIONS(2570),
    [anon_sym_ATavailable] = ACTIONS(2570),
    [anon_sym___builtin_available] = ACTIONS(2568),
    [anon_sym_va_arg] = ACTIONS(2568),
    [anon_sym_ATencode] = ACTIONS(2570),
    [anon_sym_ATsynchronized] = ACTIONS(2570),
    [anon_sym_BOOL] = ACTIONS(2568),
    [anon_sym_IMP] = ACTIONS(2568),
    [anon_sym_SEL] = ACTIONS(2568),
    [anon_sym_Class] = ACTIONS(2568),
    [anon_sym_id] = ACTIONS(2568),
  },
  [STATE(1097)] = {
    [sym_identifier] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(1098)] = {
    [sym_identifier] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(1099)] = {
    [sym_identifier] = ACTIONS(2510),
    [anon_sym_LPAREN2] = ACTIONS(2512),
    [anon_sym_BANG] = ACTIONS(2512),
    [anon_sym_TILDE] = ACTIONS(2512),
    [anon_sym_DASH] = ACTIONS(2510),
    [anon_sym_PLUS] = ACTIONS(2510),
    [anon_sym_STAR] = ACTIONS(2512),
    [anon_sym_CARET] = ACTIONS(2512),
    [anon_sym_AMP] = ACTIONS(2512),
    [anon_sym_SEMI] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym_LBRACE] = ACTIONS(2512),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_ATautoreleasepool] = ACTIONS(2512),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [anon_sym_if] = ACTIONS(2510),
    [anon_sym_else] = ACTIONS(2510),
    [anon_sym_switch] = ACTIONS(2510),
    [anon_sym_while] = ACTIONS(2510),
    [anon_sym_do] = ACTIONS(2510),
    [anon_sym_for] = ACTIONS(2510),
    [anon_sym_in] = ACTIONS(2510),
    [anon_sym_return] = ACTIONS(2510),
    [anon_sym_break] = ACTIONS(2510),
    [anon_sym_continue] = ACTIONS(2510),
    [anon_sym_goto] = ACTIONS(2510),
    [anon_sym_DASH_DASH] = ACTIONS(2512),
    [anon_sym_PLUS_PLUS] = ACTIONS(2512),
    [anon_sym_sizeof] = ACTIONS(2510),
    [anon_sym___alignof__] = ACTIONS(2510),
    [anon_sym___alignof] = ACTIONS(2510),
    [anon_sym__alignof] = ACTIONS(2510),
    [anon_sym_alignof] = ACTIONS(2510),
    [anon_sym__Alignof] = ACTIONS(2510),
    [anon_sym_offsetof] = ACTIONS(2510),
    [anon_sym__Generic] = ACTIONS(2510),
    [anon_sym_asm] = ACTIONS(2510),
    [anon_sym___asm__] = ACTIONS(2510),
    [anon_sym___asm] = ACTIONS(2510),
    [sym_number_literal] = ACTIONS(2512),
    [anon_sym_L_SQUOTE] = ACTIONS(2512),
    [anon_sym_u_SQUOTE] = ACTIONS(2512),
    [anon_sym_U_SQUOTE] = ACTIONS(2512),
    [anon_sym_u8_SQUOTE] = ACTIONS(2512),
    [anon_sym_SQUOTE] = ACTIONS(2512),
    [anon_sym_AT] = ACTIONS(2510),
    [anon_sym_DQUOTE] = ACTIONS(2512),
    [anon_sym_L_DQUOTE] = ACTIONS(2512),
    [anon_sym_u_DQUOTE] = ACTIONS(2512),
    [anon_sym_U_DQUOTE] = ACTIONS(2512),
    [anon_sym_u8_DQUOTE] = ACTIONS(2512),
    [sym_true] = ACTIONS(2510),
    [sym_false] = ACTIONS(2510),
    [anon_sym_NULL] = ACTIONS(2510),
    [anon_sym_nullptr] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATtry] = ACTIONS(2512),
    [anon_sym___try] = ACTIONS(2510),
    [anon_sym_ATthrow] = ACTIONS(2512),
    [anon_sym_ATselector] = ACTIONS(2512),
    [anon_sym_ATavailable] = ACTIONS(2512),
    [anon_sym___builtin_available] = ACTIONS(2510),
    [anon_sym_va_arg] = ACTIONS(2510),
    [anon_sym_ATencode] = ACTIONS(2512),
    [anon_sym_ATsynchronized] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(1100)] = {
    [sym_identifier] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(1101)] = {
    [sym_identifier] = ACTIONS(2554),
    [anon_sym_LPAREN2] = ACTIONS(2556),
    [anon_sym_BANG] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2556),
    [anon_sym_DASH] = ACTIONS(2554),
    [anon_sym_PLUS] = ACTIONS(2554),
    [anon_sym_STAR] = ACTIONS(2556),
    [anon_sym_CARET] = ACTIONS(2556),
    [anon_sym_AMP] = ACTIONS(2556),
    [anon_sym_SEMI] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym_LBRACE] = ACTIONS(2556),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_ATautoreleasepool] = ACTIONS(2556),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [anon_sym_if] = ACTIONS(2554),
    [anon_sym_else] = ACTIONS(2554),
    [anon_sym_switch] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(2554),
    [anon_sym_do] = ACTIONS(2554),
    [anon_sym_for] = ACTIONS(2554),
    [anon_sym_in] = ACTIONS(2554),
    [anon_sym_return] = ACTIONS(2554),
    [anon_sym_break] = ACTIONS(2554),
    [anon_sym_continue] = ACTIONS(2554),
    [anon_sym_goto] = ACTIONS(2554),
    [anon_sym_DASH_DASH] = ACTIONS(2556),
    [anon_sym_PLUS_PLUS] = ACTIONS(2556),
    [anon_sym_sizeof] = ACTIONS(2554),
    [anon_sym___alignof__] = ACTIONS(2554),
    [anon_sym___alignof] = ACTIONS(2554),
    [anon_sym__alignof] = ACTIONS(2554),
    [anon_sym_alignof] = ACTIONS(2554),
    [anon_sym__Alignof] = ACTIONS(2554),
    [anon_sym_offsetof] = ACTIONS(2554),
    [anon_sym__Generic] = ACTIONS(2554),
    [anon_sym_asm] = ACTIONS(2554),
    [anon_sym___asm__] = ACTIONS(2554),
    [anon_sym___asm] = ACTIONS(2554),
    [sym_number_literal] = ACTIONS(2556),
    [anon_sym_L_SQUOTE] = ACTIONS(2556),
    [anon_sym_u_SQUOTE] = ACTIONS(2556),
    [anon_sym_U_SQUOTE] = ACTIONS(2556),
    [anon_sym_u8_SQUOTE] = ACTIONS(2556),
    [anon_sym_SQUOTE] = ACTIONS(2556),
    [anon_sym_AT] = ACTIONS(2554),
    [anon_sym_DQUOTE] = ACTIONS(2556),
    [anon_sym_L_DQUOTE] = ACTIONS(2556),
    [anon_sym_u_DQUOTE] = ACTIONS(2556),
    [anon_sym_U_DQUOTE] = ACTIONS(2556),
    [anon_sym_u8_DQUOTE] = ACTIONS(2556),
    [sym_true] = ACTIONS(2554),
    [sym_false] = ACTIONS(2554),
    [anon_sym_NULL] = ACTIONS(2554),
    [anon_sym_nullptr] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATtry] = ACTIONS(2556),
    [anon_sym___try] = ACTIONS(2554),
    [anon_sym_ATthrow] = ACTIONS(2556),
    [anon_sym_ATselector] = ACTIONS(2556),
    [anon_sym_ATavailable] = ACTIONS(2556),
    [anon_sym___builtin_available] = ACTIONS(2554),
    [anon_sym_va_arg] = ACTIONS(2554),
    [anon_sym_ATencode] = ACTIONS(2556),
    [anon_sym_ATsynchronized] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(1102)] = {
    [sym_identifier] = ACTIONS(2558),
    [anon_sym_LPAREN2] = ACTIONS(2560),
    [anon_sym_BANG] = ACTIONS(2560),
    [anon_sym_TILDE] = ACTIONS(2560),
    [anon_sym_DASH] = ACTIONS(2558),
    [anon_sym_PLUS] = ACTIONS(2558),
    [anon_sym_STAR] = ACTIONS(2560),
    [anon_sym_CARET] = ACTIONS(2560),
    [anon_sym_AMP] = ACTIONS(2560),
    [anon_sym_SEMI] = ACTIONS(2560),
    [anon_sym___extension__] = ACTIONS(2558),
    [anon_sym_typedef] = ACTIONS(2558),
    [anon_sym_extern] = ACTIONS(2558),
    [anon_sym___attribute__] = ACTIONS(2558),
    [anon_sym___attribute] = ACTIONS(2558),
    [anon_sym_noreturn] = ACTIONS(2558),
    [anon_sym_LBRACK] = ACTIONS(2560),
    [anon_sym___declspec] = ACTIONS(2558),
    [anon_sym_LBRACE] = ACTIONS(2560),
    [anon_sym_signed] = ACTIONS(2558),
    [anon_sym_unsigned] = ACTIONS(2558),
    [anon_sym_long] = ACTIONS(2558),
    [anon_sym_short] = ACTIONS(2558),
    [anon_sym_static] = ACTIONS(2558),
    [anon_sym_ATautoreleasepool] = ACTIONS(2560),
    [anon_sym_auto] = ACTIONS(2558),
    [anon_sym_register] = ACTIONS(2558),
    [anon_sym_inline] = ACTIONS(2558),
    [anon_sym___inline] = ACTIONS(2558),
    [anon_sym___inline__] = ACTIONS(2558),
    [anon_sym___forceinline] = ACTIONS(2558),
    [anon_sym_thread_local] = ACTIONS(2558),
    [anon_sym___thread] = ACTIONS(2558),
    [anon_sym_CG_EXTERN] = ACTIONS(2558),
    [anon_sym_CG_INLINE] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2558),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2558),
    [anon_sym_IBOutlet] = ACTIONS(2558),
    [anon_sym_IBInspectable] = ACTIONS(2558),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2558),
    [anon_sym_NS_INLINE] = ACTIONS(2558),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2558),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2558),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2558),
    [anon_sym_const] = ACTIONS(2558),
    [anon_sym_constexpr] = ACTIONS(2558),
    [anon_sym_volatile] = ACTIONS(2558),
    [anon_sym_restrict] = ACTIONS(2558),
    [anon_sym___restrict__] = ACTIONS(2558),
    [anon_sym__Atomic] = ACTIONS(2558),
    [anon_sym__Noreturn] = ACTIONS(2558),
    [anon_sym__Nonnull] = ACTIONS(2558),
    [anon_sym_nullable] = ACTIONS(2558),
    [anon_sym__Complex] = ACTIONS(2558),
    [anon_sym__Nullable] = ACTIONS(2558),
    [anon_sym__Nullable_result] = ACTIONS(2558),
    [anon_sym__Null_unspecified] = ACTIONS(2558),
    [anon_sym___autoreleasing] = ACTIONS(2558),
    [anon_sym___block] = ACTIONS(2558),
    [anon_sym___bridge] = ACTIONS(2558),
    [anon_sym___bridge_retained] = ACTIONS(2558),
    [anon_sym___bridge_transfer] = ACTIONS(2558),
    [anon_sym___complex] = ACTIONS(2558),
    [anon_sym___const] = ACTIONS(2558),
    [anon_sym___imag] = ACTIONS(2558),
    [anon_sym___kindof] = ACTIONS(2558),
    [anon_sym___nonnull] = ACTIONS(2558),
    [anon_sym___nullable] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2558),
    [anon_sym___real] = ACTIONS(2558),
    [anon_sym___strong] = ACTIONS(2558),
    [anon_sym___unsafe_unretained] = ACTIONS(2558),
    [anon_sym___unused] = ACTIONS(2558),
    [anon_sym___weak] = ACTIONS(2558),
    [anon_sym_alignas] = ACTIONS(2558),
    [anon_sym__Alignas] = ACTIONS(2558),
    [sym_primitive_type] = ACTIONS(2558),
    [anon_sym_enum] = ACTIONS(2558),
    [anon_sym_struct] = ACTIONS(2558),
    [anon_sym_union] = ACTIONS(2558),
    [anon_sym_if] = ACTIONS(2558),
    [anon_sym_else] = ACTIONS(2558),
    [anon_sym_switch] = ACTIONS(2558),
    [anon_sym_while] = ACTIONS(2558),
    [anon_sym_do] = ACTIONS(2558),
    [anon_sym_for] = ACTIONS(2558),
    [anon_sym_in] = ACTIONS(2558),
    [anon_sym_return] = ACTIONS(2558),
    [anon_sym_break] = ACTIONS(2558),
    [anon_sym_continue] = ACTIONS(2558),
    [anon_sym_goto] = ACTIONS(2558),
    [anon_sym_DASH_DASH] = ACTIONS(2560),
    [anon_sym_PLUS_PLUS] = ACTIONS(2560),
    [anon_sym_sizeof] = ACTIONS(2558),
    [anon_sym___alignof__] = ACTIONS(2558),
    [anon_sym___alignof] = ACTIONS(2558),
    [anon_sym__alignof] = ACTIONS(2558),
    [anon_sym_alignof] = ACTIONS(2558),
    [anon_sym__Alignof] = ACTIONS(2558),
    [anon_sym_offsetof] = ACTIONS(2558),
    [anon_sym__Generic] = ACTIONS(2558),
    [anon_sym_asm] = ACTIONS(2558),
    [anon_sym___asm__] = ACTIONS(2558),
    [anon_sym___asm] = ACTIONS(2558),
    [sym_number_literal] = ACTIONS(2560),
    [anon_sym_L_SQUOTE] = ACTIONS(2560),
    [anon_sym_u_SQUOTE] = ACTIONS(2560),
    [anon_sym_U_SQUOTE] = ACTIONS(2560),
    [anon_sym_u8_SQUOTE] = ACTIONS(2560),
    [anon_sym_SQUOTE] = ACTIONS(2560),
    [anon_sym_AT] = ACTIONS(2558),
    [anon_sym_DQUOTE] = ACTIONS(2560),
    [anon_sym_L_DQUOTE] = ACTIONS(2560),
    [anon_sym_u_DQUOTE] = ACTIONS(2560),
    [anon_sym_U_DQUOTE] = ACTIONS(2560),
    [anon_sym_u8_DQUOTE] = ACTIONS(2560),
    [sym_true] = ACTIONS(2558),
    [sym_false] = ACTIONS(2558),
    [anon_sym_NULL] = ACTIONS(2558),
    [anon_sym_nullptr] = ACTIONS(2558),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2558),
    [anon_sym___typeof] = ACTIONS(2558),
    [anon_sym_typeof] = ACTIONS(2558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2558),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2558),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2558),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_API_AVAILABLE] = ACTIONS(2558),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_API_DEPRECATED] = ACTIONS(2558),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym___deprecated_msg] = ACTIONS(2558),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2558),
    [anon_sym_ATtry] = ACTIONS(2560),
    [anon_sym___try] = ACTIONS(2558),
    [anon_sym_ATthrow] = ACTIONS(2560),
    [anon_sym_ATselector] = ACTIONS(2560),
    [anon_sym_ATavailable] = ACTIONS(2560),
    [anon_sym___builtin_available] = ACTIONS(2558),
    [anon_sym_va_arg] = ACTIONS(2558),
    [anon_sym_ATencode] = ACTIONS(2560),
    [anon_sym_ATsynchronized] = ACTIONS(2560),
    [anon_sym_BOOL] = ACTIONS(2558),
    [anon_sym_IMP] = ACTIONS(2558),
    [anon_sym_SEL] = ACTIONS(2558),
    [anon_sym_Class] = ACTIONS(2558),
    [anon_sym_id] = ACTIONS(2558),
  },
  [STATE(1103)] = {
    [sym_identifier] = ACTIONS(2346),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_BANG] = ACTIONS(2348),
    [anon_sym_TILDE] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_if] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_switch] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_do] = ACTIONS(2346),
    [anon_sym_for] = ACTIONS(2346),
    [anon_sym_in] = ACTIONS(2346),
    [anon_sym_return] = ACTIONS(2346),
    [anon_sym_break] = ACTIONS(2346),
    [anon_sym_continue] = ACTIONS(2346),
    [anon_sym_goto] = ACTIONS(2346),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_sizeof] = ACTIONS(2346),
    [anon_sym___alignof__] = ACTIONS(2346),
    [anon_sym___alignof] = ACTIONS(2346),
    [anon_sym__alignof] = ACTIONS(2346),
    [anon_sym_alignof] = ACTIONS(2346),
    [anon_sym__Alignof] = ACTIONS(2346),
    [anon_sym_offsetof] = ACTIONS(2346),
    [anon_sym__Generic] = ACTIONS(2346),
    [anon_sym_asm] = ACTIONS(2346),
    [anon_sym___asm__] = ACTIONS(2346),
    [anon_sym___asm] = ACTIONS(2346),
    [sym_number_literal] = ACTIONS(2348),
    [anon_sym_L_SQUOTE] = ACTIONS(2348),
    [anon_sym_u_SQUOTE] = ACTIONS(2348),
    [anon_sym_U_SQUOTE] = ACTIONS(2348),
    [anon_sym_u8_SQUOTE] = ACTIONS(2348),
    [anon_sym_SQUOTE] = ACTIONS(2348),
    [anon_sym_AT] = ACTIONS(2346),
    [anon_sym_DQUOTE] = ACTIONS(2348),
    [anon_sym_L_DQUOTE] = ACTIONS(2348),
    [anon_sym_u_DQUOTE] = ACTIONS(2348),
    [anon_sym_U_DQUOTE] = ACTIONS(2348),
    [anon_sym_u8_DQUOTE] = ACTIONS(2348),
    [sym_true] = ACTIONS(2346),
    [sym_false] = ACTIONS(2346),
    [anon_sym_NULL] = ACTIONS(2346),
    [anon_sym_nullptr] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATtry] = ACTIONS(2348),
    [anon_sym___try] = ACTIONS(2346),
    [anon_sym_ATthrow] = ACTIONS(2348),
    [anon_sym_ATselector] = ACTIONS(2348),
    [anon_sym_ATavailable] = ACTIONS(2348),
    [anon_sym___builtin_available] = ACTIONS(2346),
    [anon_sym_va_arg] = ACTIONS(2346),
    [anon_sym_ATencode] = ACTIONS(2348),
    [anon_sym_ATsynchronized] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(1104)] = {
    [sym_identifier] = ACTIONS(2350),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_BANG] = ACTIONS(2352),
    [anon_sym_TILDE] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_if] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_switch] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_do] = ACTIONS(2350),
    [anon_sym_for] = ACTIONS(2350),
    [anon_sym_in] = ACTIONS(2350),
    [anon_sym_return] = ACTIONS(2350),
    [anon_sym_break] = ACTIONS(2350),
    [anon_sym_continue] = ACTIONS(2350),
    [anon_sym_goto] = ACTIONS(2350),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_sizeof] = ACTIONS(2350),
    [anon_sym___alignof__] = ACTIONS(2350),
    [anon_sym___alignof] = ACTIONS(2350),
    [anon_sym__alignof] = ACTIONS(2350),
    [anon_sym_alignof] = ACTIONS(2350),
    [anon_sym__Alignof] = ACTIONS(2350),
    [anon_sym_offsetof] = ACTIONS(2350),
    [anon_sym__Generic] = ACTIONS(2350),
    [anon_sym_asm] = ACTIONS(2350),
    [anon_sym___asm__] = ACTIONS(2350),
    [anon_sym___asm] = ACTIONS(2350),
    [sym_number_literal] = ACTIONS(2352),
    [anon_sym_L_SQUOTE] = ACTIONS(2352),
    [anon_sym_u_SQUOTE] = ACTIONS(2352),
    [anon_sym_U_SQUOTE] = ACTIONS(2352),
    [anon_sym_u8_SQUOTE] = ACTIONS(2352),
    [anon_sym_SQUOTE] = ACTIONS(2352),
    [anon_sym_AT] = ACTIONS(2350),
    [anon_sym_DQUOTE] = ACTIONS(2352),
    [anon_sym_L_DQUOTE] = ACTIONS(2352),
    [anon_sym_u_DQUOTE] = ACTIONS(2352),
    [anon_sym_U_DQUOTE] = ACTIONS(2352),
    [anon_sym_u8_DQUOTE] = ACTIONS(2352),
    [sym_true] = ACTIONS(2350),
    [sym_false] = ACTIONS(2350),
    [anon_sym_NULL] = ACTIONS(2350),
    [anon_sym_nullptr] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATtry] = ACTIONS(2352),
    [anon_sym___try] = ACTIONS(2350),
    [anon_sym_ATthrow] = ACTIONS(2352),
    [anon_sym_ATselector] = ACTIONS(2352),
    [anon_sym_ATavailable] = ACTIONS(2352),
    [anon_sym___builtin_available] = ACTIONS(2350),
    [anon_sym_va_arg] = ACTIONS(2350),
    [anon_sym_ATencode] = ACTIONS(2352),
    [anon_sym_ATsynchronized] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(1105)] = {
    [sym_identifier] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(1106)] = {
    [sym_identifier] = ACTIONS(2474),
    [anon_sym_LPAREN2] = ACTIONS(2476),
    [anon_sym_BANG] = ACTIONS(2476),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_DASH] = ACTIONS(2474),
    [anon_sym_PLUS] = ACTIONS(2474),
    [anon_sym_STAR] = ACTIONS(2476),
    [anon_sym_CARET] = ACTIONS(2476),
    [anon_sym_AMP] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [anon_sym_if] = ACTIONS(2474),
    [anon_sym_else] = ACTIONS(2474),
    [anon_sym_switch] = ACTIONS(2474),
    [anon_sym_while] = ACTIONS(2474),
    [anon_sym_do] = ACTIONS(2474),
    [anon_sym_for] = ACTIONS(2474),
    [anon_sym_in] = ACTIONS(2474),
    [anon_sym_return] = ACTIONS(2474),
    [anon_sym_break] = ACTIONS(2474),
    [anon_sym_continue] = ACTIONS(2474),
    [anon_sym_goto] = ACTIONS(2474),
    [anon_sym_DASH_DASH] = ACTIONS(2476),
    [anon_sym_PLUS_PLUS] = ACTIONS(2476),
    [anon_sym_sizeof] = ACTIONS(2474),
    [anon_sym___alignof__] = ACTIONS(2474),
    [anon_sym___alignof] = ACTIONS(2474),
    [anon_sym__alignof] = ACTIONS(2474),
    [anon_sym_alignof] = ACTIONS(2474),
    [anon_sym__Alignof] = ACTIONS(2474),
    [anon_sym_offsetof] = ACTIONS(2474),
    [anon_sym__Generic] = ACTIONS(2474),
    [anon_sym_asm] = ACTIONS(2474),
    [anon_sym___asm__] = ACTIONS(2474),
    [anon_sym___asm] = ACTIONS(2474),
    [sym_number_literal] = ACTIONS(2476),
    [anon_sym_L_SQUOTE] = ACTIONS(2476),
    [anon_sym_u_SQUOTE] = ACTIONS(2476),
    [anon_sym_U_SQUOTE] = ACTIONS(2476),
    [anon_sym_u8_SQUOTE] = ACTIONS(2476),
    [anon_sym_SQUOTE] = ACTIONS(2476),
    [anon_sym_AT] = ACTIONS(2474),
    [anon_sym_DQUOTE] = ACTIONS(2476),
    [anon_sym_L_DQUOTE] = ACTIONS(2476),
    [anon_sym_u_DQUOTE] = ACTIONS(2476),
    [anon_sym_U_DQUOTE] = ACTIONS(2476),
    [anon_sym_u8_DQUOTE] = ACTIONS(2476),
    [sym_true] = ACTIONS(2474),
    [sym_false] = ACTIONS(2474),
    [anon_sym_NULL] = ACTIONS(2474),
    [anon_sym_nullptr] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATtry] = ACTIONS(2476),
    [anon_sym___try] = ACTIONS(2474),
    [anon_sym_ATthrow] = ACTIONS(2476),
    [anon_sym_ATselector] = ACTIONS(2476),
    [anon_sym_ATavailable] = ACTIONS(2476),
    [anon_sym___builtin_available] = ACTIONS(2474),
    [anon_sym_va_arg] = ACTIONS(2474),
    [anon_sym_ATencode] = ACTIONS(2476),
    [anon_sym_ATsynchronized] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(1107)] = {
    [sym_identifier] = ACTIONS(2360),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_BANG] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_if] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_switch] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_do] = ACTIONS(2360),
    [anon_sym_for] = ACTIONS(2360),
    [anon_sym_in] = ACTIONS(2360),
    [anon_sym_return] = ACTIONS(2360),
    [anon_sym_break] = ACTIONS(2360),
    [anon_sym_continue] = ACTIONS(2360),
    [anon_sym_goto] = ACTIONS(2360),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_sizeof] = ACTIONS(2360),
    [anon_sym___alignof__] = ACTIONS(2360),
    [anon_sym___alignof] = ACTIONS(2360),
    [anon_sym__alignof] = ACTIONS(2360),
    [anon_sym_alignof] = ACTIONS(2360),
    [anon_sym__Alignof] = ACTIONS(2360),
    [anon_sym_offsetof] = ACTIONS(2360),
    [anon_sym__Generic] = ACTIONS(2360),
    [anon_sym_asm] = ACTIONS(2360),
    [anon_sym___asm__] = ACTIONS(2360),
    [anon_sym___asm] = ACTIONS(2360),
    [sym_number_literal] = ACTIONS(2362),
    [anon_sym_L_SQUOTE] = ACTIONS(2362),
    [anon_sym_u_SQUOTE] = ACTIONS(2362),
    [anon_sym_U_SQUOTE] = ACTIONS(2362),
    [anon_sym_u8_SQUOTE] = ACTIONS(2362),
    [anon_sym_SQUOTE] = ACTIONS(2362),
    [anon_sym_AT] = ACTIONS(2360),
    [anon_sym_DQUOTE] = ACTIONS(2362),
    [anon_sym_L_DQUOTE] = ACTIONS(2362),
    [anon_sym_u_DQUOTE] = ACTIONS(2362),
    [anon_sym_U_DQUOTE] = ACTIONS(2362),
    [anon_sym_u8_DQUOTE] = ACTIONS(2362),
    [sym_true] = ACTIONS(2360),
    [sym_false] = ACTIONS(2360),
    [anon_sym_NULL] = ACTIONS(2360),
    [anon_sym_nullptr] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATtry] = ACTIONS(2362),
    [anon_sym___try] = ACTIONS(2360),
    [anon_sym_ATthrow] = ACTIONS(2362),
    [anon_sym_ATselector] = ACTIONS(2362),
    [anon_sym_ATavailable] = ACTIONS(2362),
    [anon_sym___builtin_available] = ACTIONS(2360),
    [anon_sym_va_arg] = ACTIONS(2360),
    [anon_sym_ATencode] = ACTIONS(2362),
    [anon_sym_ATsynchronized] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(1108)] = {
    [sym_identifier] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(1109)] = {
    [sym_identifier] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(1110)] = {
    [sym_identifier] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(1111)] = {
    [sym_identifier] = ACTIONS(2416),
    [anon_sym_LPAREN2] = ACTIONS(2418),
    [anon_sym_BANG] = ACTIONS(2418),
    [anon_sym_TILDE] = ACTIONS(2418),
    [anon_sym_DASH] = ACTIONS(2416),
    [anon_sym_PLUS] = ACTIONS(2416),
    [anon_sym_STAR] = ACTIONS(2418),
    [anon_sym_CARET] = ACTIONS(2418),
    [anon_sym_AMP] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [anon_sym_if] = ACTIONS(2416),
    [anon_sym_else] = ACTIONS(2416),
    [anon_sym_switch] = ACTIONS(2416),
    [anon_sym_while] = ACTIONS(2416),
    [anon_sym_do] = ACTIONS(2416),
    [anon_sym_for] = ACTIONS(2416),
    [anon_sym_in] = ACTIONS(2416),
    [anon_sym_return] = ACTIONS(2416),
    [anon_sym_break] = ACTIONS(2416),
    [anon_sym_continue] = ACTIONS(2416),
    [anon_sym_goto] = ACTIONS(2416),
    [anon_sym_DASH_DASH] = ACTIONS(2418),
    [anon_sym_PLUS_PLUS] = ACTIONS(2418),
    [anon_sym_sizeof] = ACTIONS(2416),
    [anon_sym___alignof__] = ACTIONS(2416),
    [anon_sym___alignof] = ACTIONS(2416),
    [anon_sym__alignof] = ACTIONS(2416),
    [anon_sym_alignof] = ACTIONS(2416),
    [anon_sym__Alignof] = ACTIONS(2416),
    [anon_sym_offsetof] = ACTIONS(2416),
    [anon_sym__Generic] = ACTIONS(2416),
    [anon_sym_asm] = ACTIONS(2416),
    [anon_sym___asm__] = ACTIONS(2416),
    [anon_sym___asm] = ACTIONS(2416),
    [sym_number_literal] = ACTIONS(2418),
    [anon_sym_L_SQUOTE] = ACTIONS(2418),
    [anon_sym_u_SQUOTE] = ACTIONS(2418),
    [anon_sym_U_SQUOTE] = ACTIONS(2418),
    [anon_sym_u8_SQUOTE] = ACTIONS(2418),
    [anon_sym_SQUOTE] = ACTIONS(2418),
    [anon_sym_AT] = ACTIONS(2416),
    [anon_sym_DQUOTE] = ACTIONS(2418),
    [anon_sym_L_DQUOTE] = ACTIONS(2418),
    [anon_sym_u_DQUOTE] = ACTIONS(2418),
    [anon_sym_U_DQUOTE] = ACTIONS(2418),
    [anon_sym_u8_DQUOTE] = ACTIONS(2418),
    [sym_true] = ACTIONS(2416),
    [sym_false] = ACTIONS(2416),
    [anon_sym_NULL] = ACTIONS(2416),
    [anon_sym_nullptr] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATtry] = ACTIONS(2418),
    [anon_sym___try] = ACTIONS(2416),
    [anon_sym_ATthrow] = ACTIONS(2418),
    [anon_sym_ATselector] = ACTIONS(2418),
    [anon_sym_ATavailable] = ACTIONS(2418),
    [anon_sym___builtin_available] = ACTIONS(2416),
    [anon_sym_va_arg] = ACTIONS(2416),
    [anon_sym_ATencode] = ACTIONS(2418),
    [anon_sym_ATsynchronized] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(1112)] = {
    [sym_identifier] = ACTIONS(2424),
    [anon_sym_LPAREN2] = ACTIONS(2426),
    [anon_sym_BANG] = ACTIONS(2426),
    [anon_sym_TILDE] = ACTIONS(2426),
    [anon_sym_DASH] = ACTIONS(2424),
    [anon_sym_PLUS] = ACTIONS(2424),
    [anon_sym_STAR] = ACTIONS(2426),
    [anon_sym_CARET] = ACTIONS(2426),
    [anon_sym_AMP] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [anon_sym_if] = ACTIONS(2424),
    [anon_sym_else] = ACTIONS(2424),
    [anon_sym_switch] = ACTIONS(2424),
    [anon_sym_while] = ACTIONS(2424),
    [anon_sym_do] = ACTIONS(2424),
    [anon_sym_for] = ACTIONS(2424),
    [anon_sym_in] = ACTIONS(2424),
    [anon_sym_return] = ACTIONS(2424),
    [anon_sym_break] = ACTIONS(2424),
    [anon_sym_continue] = ACTIONS(2424),
    [anon_sym_goto] = ACTIONS(2424),
    [anon_sym_DASH_DASH] = ACTIONS(2426),
    [anon_sym_PLUS_PLUS] = ACTIONS(2426),
    [anon_sym_sizeof] = ACTIONS(2424),
    [anon_sym___alignof__] = ACTIONS(2424),
    [anon_sym___alignof] = ACTIONS(2424),
    [anon_sym__alignof] = ACTIONS(2424),
    [anon_sym_alignof] = ACTIONS(2424),
    [anon_sym__Alignof] = ACTIONS(2424),
    [anon_sym_offsetof] = ACTIONS(2424),
    [anon_sym__Generic] = ACTIONS(2424),
    [anon_sym_asm] = ACTIONS(2424),
    [anon_sym___asm__] = ACTIONS(2424),
    [anon_sym___asm] = ACTIONS(2424),
    [sym_number_literal] = ACTIONS(2426),
    [anon_sym_L_SQUOTE] = ACTIONS(2426),
    [anon_sym_u_SQUOTE] = ACTIONS(2426),
    [anon_sym_U_SQUOTE] = ACTIONS(2426),
    [anon_sym_u8_SQUOTE] = ACTIONS(2426),
    [anon_sym_SQUOTE] = ACTIONS(2426),
    [anon_sym_AT] = ACTIONS(2424),
    [anon_sym_DQUOTE] = ACTIONS(2426),
    [anon_sym_L_DQUOTE] = ACTIONS(2426),
    [anon_sym_u_DQUOTE] = ACTIONS(2426),
    [anon_sym_U_DQUOTE] = ACTIONS(2426),
    [anon_sym_u8_DQUOTE] = ACTIONS(2426),
    [sym_true] = ACTIONS(2424),
    [sym_false] = ACTIONS(2424),
    [anon_sym_NULL] = ACTIONS(2424),
    [anon_sym_nullptr] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATtry] = ACTIONS(2426),
    [anon_sym___try] = ACTIONS(2424),
    [anon_sym_ATthrow] = ACTIONS(2426),
    [anon_sym_ATselector] = ACTIONS(2426),
    [anon_sym_ATavailable] = ACTIONS(2426),
    [anon_sym___builtin_available] = ACTIONS(2424),
    [anon_sym_va_arg] = ACTIONS(2424),
    [anon_sym_ATencode] = ACTIONS(2426),
    [anon_sym_ATsynchronized] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(1113)] = {
    [sym_identifier] = ACTIONS(2420),
    [anon_sym_LPAREN2] = ACTIONS(2422),
    [anon_sym_BANG] = ACTIONS(2422),
    [anon_sym_TILDE] = ACTIONS(2422),
    [anon_sym_DASH] = ACTIONS(2420),
    [anon_sym_PLUS] = ACTIONS(2420),
    [anon_sym_STAR] = ACTIONS(2422),
    [anon_sym_CARET] = ACTIONS(2422),
    [anon_sym_AMP] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [anon_sym_if] = ACTIONS(2420),
    [anon_sym_else] = ACTIONS(2420),
    [anon_sym_switch] = ACTIONS(2420),
    [anon_sym_while] = ACTIONS(2420),
    [anon_sym_do] = ACTIONS(2420),
    [anon_sym_for] = ACTIONS(2420),
    [anon_sym_in] = ACTIONS(2420),
    [anon_sym_return] = ACTIONS(2420),
    [anon_sym_break] = ACTIONS(2420),
    [anon_sym_continue] = ACTIONS(2420),
    [anon_sym_goto] = ACTIONS(2420),
    [anon_sym_DASH_DASH] = ACTIONS(2422),
    [anon_sym_PLUS_PLUS] = ACTIONS(2422),
    [anon_sym_sizeof] = ACTIONS(2420),
    [anon_sym___alignof__] = ACTIONS(2420),
    [anon_sym___alignof] = ACTIONS(2420),
    [anon_sym__alignof] = ACTIONS(2420),
    [anon_sym_alignof] = ACTIONS(2420),
    [anon_sym__Alignof] = ACTIONS(2420),
    [anon_sym_offsetof] = ACTIONS(2420),
    [anon_sym__Generic] = ACTIONS(2420),
    [anon_sym_asm] = ACTIONS(2420),
    [anon_sym___asm__] = ACTIONS(2420),
    [anon_sym___asm] = ACTIONS(2420),
    [sym_number_literal] = ACTIONS(2422),
    [anon_sym_L_SQUOTE] = ACTIONS(2422),
    [anon_sym_u_SQUOTE] = ACTIONS(2422),
    [anon_sym_U_SQUOTE] = ACTIONS(2422),
    [anon_sym_u8_SQUOTE] = ACTIONS(2422),
    [anon_sym_SQUOTE] = ACTIONS(2422),
    [anon_sym_AT] = ACTIONS(2420),
    [anon_sym_DQUOTE] = ACTIONS(2422),
    [anon_sym_L_DQUOTE] = ACTIONS(2422),
    [anon_sym_u_DQUOTE] = ACTIONS(2422),
    [anon_sym_U_DQUOTE] = ACTIONS(2422),
    [anon_sym_u8_DQUOTE] = ACTIONS(2422),
    [sym_true] = ACTIONS(2420),
    [sym_false] = ACTIONS(2420),
    [anon_sym_NULL] = ACTIONS(2420),
    [anon_sym_nullptr] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATtry] = ACTIONS(2422),
    [anon_sym___try] = ACTIONS(2420),
    [anon_sym_ATthrow] = ACTIONS(2422),
    [anon_sym_ATselector] = ACTIONS(2422),
    [anon_sym_ATavailable] = ACTIONS(2422),
    [anon_sym___builtin_available] = ACTIONS(2420),
    [anon_sym_va_arg] = ACTIONS(2420),
    [anon_sym_ATencode] = ACTIONS(2422),
    [anon_sym_ATsynchronized] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(1114)] = {
    [sym_identifier] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(1115)] = {
    [sym_identifier] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(1116)] = {
    [sym_identifier] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(1117)] = {
    [sym_identifier] = ACTIONS(2446),
    [anon_sym_LPAREN2] = ACTIONS(2448),
    [anon_sym_BANG] = ACTIONS(2448),
    [anon_sym_TILDE] = ACTIONS(2448),
    [anon_sym_DASH] = ACTIONS(2446),
    [anon_sym_PLUS] = ACTIONS(2446),
    [anon_sym_STAR] = ACTIONS(2448),
    [anon_sym_CARET] = ACTIONS(2448),
    [anon_sym_AMP] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [anon_sym_if] = ACTIONS(2446),
    [anon_sym_else] = ACTIONS(2446),
    [anon_sym_switch] = ACTIONS(2446),
    [anon_sym_while] = ACTIONS(2446),
    [anon_sym_do] = ACTIONS(2446),
    [anon_sym_for] = ACTIONS(2446),
    [anon_sym_in] = ACTIONS(2446),
    [anon_sym_return] = ACTIONS(2446),
    [anon_sym_break] = ACTIONS(2446),
    [anon_sym_continue] = ACTIONS(2446),
    [anon_sym_goto] = ACTIONS(2446),
    [anon_sym_DASH_DASH] = ACTIONS(2448),
    [anon_sym_PLUS_PLUS] = ACTIONS(2448),
    [anon_sym_sizeof] = ACTIONS(2446),
    [anon_sym___alignof__] = ACTIONS(2446),
    [anon_sym___alignof] = ACTIONS(2446),
    [anon_sym__alignof] = ACTIONS(2446),
    [anon_sym_alignof] = ACTIONS(2446),
    [anon_sym__Alignof] = ACTIONS(2446),
    [anon_sym_offsetof] = ACTIONS(2446),
    [anon_sym__Generic] = ACTIONS(2446),
    [anon_sym_asm] = ACTIONS(2446),
    [anon_sym___asm__] = ACTIONS(2446),
    [anon_sym___asm] = ACTIONS(2446),
    [sym_number_literal] = ACTIONS(2448),
    [anon_sym_L_SQUOTE] = ACTIONS(2448),
    [anon_sym_u_SQUOTE] = ACTIONS(2448),
    [anon_sym_U_SQUOTE] = ACTIONS(2448),
    [anon_sym_u8_SQUOTE] = ACTIONS(2448),
    [anon_sym_SQUOTE] = ACTIONS(2448),
    [anon_sym_AT] = ACTIONS(2446),
    [anon_sym_DQUOTE] = ACTIONS(2448),
    [anon_sym_L_DQUOTE] = ACTIONS(2448),
    [anon_sym_u_DQUOTE] = ACTIONS(2448),
    [anon_sym_U_DQUOTE] = ACTIONS(2448),
    [anon_sym_u8_DQUOTE] = ACTIONS(2448),
    [sym_true] = ACTIONS(2446),
    [sym_false] = ACTIONS(2446),
    [anon_sym_NULL] = ACTIONS(2446),
    [anon_sym_nullptr] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATtry] = ACTIONS(2448),
    [anon_sym___try] = ACTIONS(2446),
    [anon_sym_ATthrow] = ACTIONS(2448),
    [anon_sym_ATselector] = ACTIONS(2448),
    [anon_sym_ATavailable] = ACTIONS(2448),
    [anon_sym___builtin_available] = ACTIONS(2446),
    [anon_sym_va_arg] = ACTIONS(2446),
    [anon_sym_ATencode] = ACTIONS(2448),
    [anon_sym_ATsynchronized] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(1118)] = {
    [sym_identifier] = ACTIONS(2438),
    [anon_sym_LPAREN2] = ACTIONS(2440),
    [anon_sym_BANG] = ACTIONS(2440),
    [anon_sym_TILDE] = ACTIONS(2440),
    [anon_sym_DASH] = ACTIONS(2438),
    [anon_sym_PLUS] = ACTIONS(2438),
    [anon_sym_STAR] = ACTIONS(2440),
    [anon_sym_CARET] = ACTIONS(2440),
    [anon_sym_AMP] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [anon_sym_if] = ACTIONS(2438),
    [anon_sym_else] = ACTIONS(2438),
    [anon_sym_switch] = ACTIONS(2438),
    [anon_sym_while] = ACTIONS(2438),
    [anon_sym_do] = ACTIONS(2438),
    [anon_sym_for] = ACTIONS(2438),
    [anon_sym_in] = ACTIONS(2438),
    [anon_sym_return] = ACTIONS(2438),
    [anon_sym_break] = ACTIONS(2438),
    [anon_sym_continue] = ACTIONS(2438),
    [anon_sym_goto] = ACTIONS(2438),
    [anon_sym_DASH_DASH] = ACTIONS(2440),
    [anon_sym_PLUS_PLUS] = ACTIONS(2440),
    [anon_sym_sizeof] = ACTIONS(2438),
    [anon_sym___alignof__] = ACTIONS(2438),
    [anon_sym___alignof] = ACTIONS(2438),
    [anon_sym__alignof] = ACTIONS(2438),
    [anon_sym_alignof] = ACTIONS(2438),
    [anon_sym__Alignof] = ACTIONS(2438),
    [anon_sym_offsetof] = ACTIONS(2438),
    [anon_sym__Generic] = ACTIONS(2438),
    [anon_sym_asm] = ACTIONS(2438),
    [anon_sym___asm__] = ACTIONS(2438),
    [anon_sym___asm] = ACTIONS(2438),
    [sym_number_literal] = ACTIONS(2440),
    [anon_sym_L_SQUOTE] = ACTIONS(2440),
    [anon_sym_u_SQUOTE] = ACTIONS(2440),
    [anon_sym_U_SQUOTE] = ACTIONS(2440),
    [anon_sym_u8_SQUOTE] = ACTIONS(2440),
    [anon_sym_SQUOTE] = ACTIONS(2440),
    [anon_sym_AT] = ACTIONS(2438),
    [anon_sym_DQUOTE] = ACTIONS(2440),
    [anon_sym_L_DQUOTE] = ACTIONS(2440),
    [anon_sym_u_DQUOTE] = ACTIONS(2440),
    [anon_sym_U_DQUOTE] = ACTIONS(2440),
    [anon_sym_u8_DQUOTE] = ACTIONS(2440),
    [sym_true] = ACTIONS(2438),
    [sym_false] = ACTIONS(2438),
    [anon_sym_NULL] = ACTIONS(2438),
    [anon_sym_nullptr] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATtry] = ACTIONS(2440),
    [anon_sym___try] = ACTIONS(2438),
    [anon_sym_ATthrow] = ACTIONS(2440),
    [anon_sym_ATselector] = ACTIONS(2440),
    [anon_sym_ATavailable] = ACTIONS(2440),
    [anon_sym___builtin_available] = ACTIONS(2438),
    [anon_sym_va_arg] = ACTIONS(2438),
    [anon_sym_ATencode] = ACTIONS(2440),
    [anon_sym_ATsynchronized] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(1119)] = {
    [sym_identifier] = ACTIONS(2442),
    [anon_sym_LPAREN2] = ACTIONS(2444),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_TILDE] = ACTIONS(2444),
    [anon_sym_DASH] = ACTIONS(2442),
    [anon_sym_PLUS] = ACTIONS(2442),
    [anon_sym_STAR] = ACTIONS(2444),
    [anon_sym_CARET] = ACTIONS(2444),
    [anon_sym_AMP] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [anon_sym_if] = ACTIONS(2442),
    [anon_sym_else] = ACTIONS(2442),
    [anon_sym_switch] = ACTIONS(2442),
    [anon_sym_while] = ACTIONS(2442),
    [anon_sym_do] = ACTIONS(2442),
    [anon_sym_for] = ACTIONS(2442),
    [anon_sym_in] = ACTIONS(2442),
    [anon_sym_return] = ACTIONS(2442),
    [anon_sym_break] = ACTIONS(2442),
    [anon_sym_continue] = ACTIONS(2442),
    [anon_sym_goto] = ACTIONS(2442),
    [anon_sym_DASH_DASH] = ACTIONS(2444),
    [anon_sym_PLUS_PLUS] = ACTIONS(2444),
    [anon_sym_sizeof] = ACTIONS(2442),
    [anon_sym___alignof__] = ACTIONS(2442),
    [anon_sym___alignof] = ACTIONS(2442),
    [anon_sym__alignof] = ACTIONS(2442),
    [anon_sym_alignof] = ACTIONS(2442),
    [anon_sym__Alignof] = ACTIONS(2442),
    [anon_sym_offsetof] = ACTIONS(2442),
    [anon_sym__Generic] = ACTIONS(2442),
    [anon_sym_asm] = ACTIONS(2442),
    [anon_sym___asm__] = ACTIONS(2442),
    [anon_sym___asm] = ACTIONS(2442),
    [sym_number_literal] = ACTIONS(2444),
    [anon_sym_L_SQUOTE] = ACTIONS(2444),
    [anon_sym_u_SQUOTE] = ACTIONS(2444),
    [anon_sym_U_SQUOTE] = ACTIONS(2444),
    [anon_sym_u8_SQUOTE] = ACTIONS(2444),
    [anon_sym_SQUOTE] = ACTIONS(2444),
    [anon_sym_AT] = ACTIONS(2442),
    [anon_sym_DQUOTE] = ACTIONS(2444),
    [anon_sym_L_DQUOTE] = ACTIONS(2444),
    [anon_sym_u_DQUOTE] = ACTIONS(2444),
    [anon_sym_U_DQUOTE] = ACTIONS(2444),
    [anon_sym_u8_DQUOTE] = ACTIONS(2444),
    [sym_true] = ACTIONS(2442),
    [sym_false] = ACTIONS(2442),
    [anon_sym_NULL] = ACTIONS(2442),
    [anon_sym_nullptr] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATtry] = ACTIONS(2444),
    [anon_sym___try] = ACTIONS(2442),
    [anon_sym_ATthrow] = ACTIONS(2444),
    [anon_sym_ATselector] = ACTIONS(2444),
    [anon_sym_ATavailable] = ACTIONS(2444),
    [anon_sym___builtin_available] = ACTIONS(2442),
    [anon_sym_va_arg] = ACTIONS(2442),
    [anon_sym_ATencode] = ACTIONS(2444),
    [anon_sym_ATsynchronized] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(1120)] = {
    [sym_identifier] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(1121)] = {
    [sym_identifier] = ACTIONS(2546),
    [anon_sym_LPAREN2] = ACTIONS(2548),
    [anon_sym_BANG] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2548),
    [anon_sym_DASH] = ACTIONS(2546),
    [anon_sym_PLUS] = ACTIONS(2546),
    [anon_sym_STAR] = ACTIONS(2548),
    [anon_sym_CARET] = ACTIONS(2548),
    [anon_sym_AMP] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [anon_sym_if] = ACTIONS(2546),
    [anon_sym_else] = ACTIONS(2546),
    [anon_sym_switch] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(2546),
    [anon_sym_do] = ACTIONS(2546),
    [anon_sym_for] = ACTIONS(2546),
    [anon_sym_in] = ACTIONS(2546),
    [anon_sym_return] = ACTIONS(2546),
    [anon_sym_break] = ACTIONS(2546),
    [anon_sym_continue] = ACTIONS(2546),
    [anon_sym_goto] = ACTIONS(2546),
    [anon_sym_DASH_DASH] = ACTIONS(2548),
    [anon_sym_PLUS_PLUS] = ACTIONS(2548),
    [anon_sym_sizeof] = ACTIONS(2546),
    [anon_sym___alignof__] = ACTIONS(2546),
    [anon_sym___alignof] = ACTIONS(2546),
    [anon_sym__alignof] = ACTIONS(2546),
    [anon_sym_alignof] = ACTIONS(2546),
    [anon_sym__Alignof] = ACTIONS(2546),
    [anon_sym_offsetof] = ACTIONS(2546),
    [anon_sym__Generic] = ACTIONS(2546),
    [anon_sym_asm] = ACTIONS(2546),
    [anon_sym___asm__] = ACTIONS(2546),
    [anon_sym___asm] = ACTIONS(2546),
    [sym_number_literal] = ACTIONS(2548),
    [anon_sym_L_SQUOTE] = ACTIONS(2548),
    [anon_sym_u_SQUOTE] = ACTIONS(2548),
    [anon_sym_U_SQUOTE] = ACTIONS(2548),
    [anon_sym_u8_SQUOTE] = ACTIONS(2548),
    [anon_sym_SQUOTE] = ACTIONS(2548),
    [anon_sym_AT] = ACTIONS(2546),
    [anon_sym_DQUOTE] = ACTIONS(2548),
    [anon_sym_L_DQUOTE] = ACTIONS(2548),
    [anon_sym_u_DQUOTE] = ACTIONS(2548),
    [anon_sym_U_DQUOTE] = ACTIONS(2548),
    [anon_sym_u8_DQUOTE] = ACTIONS(2548),
    [sym_true] = ACTIONS(2546),
    [sym_false] = ACTIONS(2546),
    [anon_sym_NULL] = ACTIONS(2546),
    [anon_sym_nullptr] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATtry] = ACTIONS(2548),
    [anon_sym___try] = ACTIONS(2546),
    [anon_sym_ATthrow] = ACTIONS(2548),
    [anon_sym_ATselector] = ACTIONS(2548),
    [anon_sym_ATavailable] = ACTIONS(2548),
    [anon_sym___builtin_available] = ACTIONS(2546),
    [anon_sym_va_arg] = ACTIONS(2546),
    [anon_sym_ATencode] = ACTIONS(2548),
    [anon_sym_ATsynchronized] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(1122)] = {
    [sym_identifier] = ACTIONS(2478),
    [anon_sym_LPAREN2] = ACTIONS(2480),
    [anon_sym_BANG] = ACTIONS(2480),
    [anon_sym_TILDE] = ACTIONS(2480),
    [anon_sym_DASH] = ACTIONS(2478),
    [anon_sym_PLUS] = ACTIONS(2478),
    [anon_sym_STAR] = ACTIONS(2480),
    [anon_sym_CARET] = ACTIONS(2480),
    [anon_sym_AMP] = ACTIONS(2480),
    [anon_sym_SEMI] = ACTIONS(2480),
    [anon_sym___extension__] = ACTIONS(2478),
    [anon_sym_typedef] = ACTIONS(2478),
    [anon_sym_extern] = ACTIONS(2478),
    [anon_sym___attribute__] = ACTIONS(2478),
    [anon_sym___attribute] = ACTIONS(2478),
    [anon_sym_noreturn] = ACTIONS(2478),
    [anon_sym_LBRACK] = ACTIONS(2480),
    [anon_sym___declspec] = ACTIONS(2478),
    [anon_sym_LBRACE] = ACTIONS(2480),
    [anon_sym_signed] = ACTIONS(2478),
    [anon_sym_unsigned] = ACTIONS(2478),
    [anon_sym_long] = ACTIONS(2478),
    [anon_sym_short] = ACTIONS(2478),
    [anon_sym_static] = ACTIONS(2478),
    [anon_sym_ATautoreleasepool] = ACTIONS(2480),
    [anon_sym_auto] = ACTIONS(2478),
    [anon_sym_register] = ACTIONS(2478),
    [anon_sym_inline] = ACTIONS(2478),
    [anon_sym___inline] = ACTIONS(2478),
    [anon_sym___inline__] = ACTIONS(2478),
    [anon_sym___forceinline] = ACTIONS(2478),
    [anon_sym_thread_local] = ACTIONS(2478),
    [anon_sym___thread] = ACTIONS(2478),
    [anon_sym_CG_EXTERN] = ACTIONS(2478),
    [anon_sym_CG_INLINE] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2478),
    [anon_sym_IBOutlet] = ACTIONS(2478),
    [anon_sym_IBInspectable] = ACTIONS(2478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2478),
    [anon_sym_NS_INLINE] = ACTIONS(2478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2478),
    [anon_sym_const] = ACTIONS(2478),
    [anon_sym_constexpr] = ACTIONS(2478),
    [anon_sym_volatile] = ACTIONS(2478),
    [anon_sym_restrict] = ACTIONS(2478),
    [anon_sym___restrict__] = ACTIONS(2478),
    [anon_sym__Atomic] = ACTIONS(2478),
    [anon_sym__Noreturn] = ACTIONS(2478),
    [anon_sym__Nonnull] = ACTIONS(2478),
    [anon_sym_nullable] = ACTIONS(2478),
    [anon_sym__Complex] = ACTIONS(2478),
    [anon_sym__Nullable] = ACTIONS(2478),
    [anon_sym__Nullable_result] = ACTIONS(2478),
    [anon_sym__Null_unspecified] = ACTIONS(2478),
    [anon_sym___autoreleasing] = ACTIONS(2478),
    [anon_sym___block] = ACTIONS(2478),
    [anon_sym___bridge] = ACTIONS(2478),
    [anon_sym___bridge_retained] = ACTIONS(2478),
    [anon_sym___bridge_transfer] = ACTIONS(2478),
    [anon_sym___complex] = ACTIONS(2478),
    [anon_sym___const] = ACTIONS(2478),
    [anon_sym___imag] = ACTIONS(2478),
    [anon_sym___kindof] = ACTIONS(2478),
    [anon_sym___nonnull] = ACTIONS(2478),
    [anon_sym___nullable] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2478),
    [anon_sym___real] = ACTIONS(2478),
    [anon_sym___strong] = ACTIONS(2478),
    [anon_sym___unsafe_unretained] = ACTIONS(2478),
    [anon_sym___unused] = ACTIONS(2478),
    [anon_sym___weak] = ACTIONS(2478),
    [anon_sym_alignas] = ACTIONS(2478),
    [anon_sym__Alignas] = ACTIONS(2478),
    [sym_primitive_type] = ACTIONS(2478),
    [anon_sym_enum] = ACTIONS(2478),
    [anon_sym_struct] = ACTIONS(2478),
    [anon_sym_union] = ACTIONS(2478),
    [anon_sym_if] = ACTIONS(2478),
    [anon_sym_else] = ACTIONS(2478),
    [anon_sym_switch] = ACTIONS(2478),
    [anon_sym_while] = ACTIONS(2478),
    [anon_sym_do] = ACTIONS(2478),
    [anon_sym_for] = ACTIONS(2478),
    [anon_sym_in] = ACTIONS(2478),
    [anon_sym_return] = ACTIONS(2478),
    [anon_sym_break] = ACTIONS(2478),
    [anon_sym_continue] = ACTIONS(2478),
    [anon_sym_goto] = ACTIONS(2478),
    [anon_sym_DASH_DASH] = ACTIONS(2480),
    [anon_sym_PLUS_PLUS] = ACTIONS(2480),
    [anon_sym_sizeof] = ACTIONS(2478),
    [anon_sym___alignof__] = ACTIONS(2478),
    [anon_sym___alignof] = ACTIONS(2478),
    [anon_sym__alignof] = ACTIONS(2478),
    [anon_sym_alignof] = ACTIONS(2478),
    [anon_sym__Alignof] = ACTIONS(2478),
    [anon_sym_offsetof] = ACTIONS(2478),
    [anon_sym__Generic] = ACTIONS(2478),
    [anon_sym_asm] = ACTIONS(2478),
    [anon_sym___asm__] = ACTIONS(2478),
    [anon_sym___asm] = ACTIONS(2478),
    [sym_number_literal] = ACTIONS(2480),
    [anon_sym_L_SQUOTE] = ACTIONS(2480),
    [anon_sym_u_SQUOTE] = ACTIONS(2480),
    [anon_sym_U_SQUOTE] = ACTIONS(2480),
    [anon_sym_u8_SQUOTE] = ACTIONS(2480),
    [anon_sym_SQUOTE] = ACTIONS(2480),
    [anon_sym_AT] = ACTIONS(2478),
    [anon_sym_DQUOTE] = ACTIONS(2480),
    [anon_sym_L_DQUOTE] = ACTIONS(2480),
    [anon_sym_u_DQUOTE] = ACTIONS(2480),
    [anon_sym_U_DQUOTE] = ACTIONS(2480),
    [anon_sym_u8_DQUOTE] = ACTIONS(2480),
    [sym_true] = ACTIONS(2478),
    [sym_false] = ACTIONS(2478),
    [anon_sym_NULL] = ACTIONS(2478),
    [anon_sym_nullptr] = ACTIONS(2478),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2478),
    [anon_sym___typeof] = ACTIONS(2478),
    [anon_sym_typeof] = ACTIONS(2478),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_API_AVAILABLE] = ACTIONS(2478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_API_DEPRECATED] = ACTIONS(2478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2478),
    [anon_sym___deprecated_msg] = ACTIONS(2478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2478),
    [anon_sym_ATtry] = ACTIONS(2480),
    [anon_sym___try] = ACTIONS(2478),
    [anon_sym_ATthrow] = ACTIONS(2480),
    [anon_sym_ATselector] = ACTIONS(2480),
    [anon_sym_ATavailable] = ACTIONS(2480),
    [anon_sym___builtin_available] = ACTIONS(2478),
    [anon_sym_va_arg] = ACTIONS(2478),
    [anon_sym_ATencode] = ACTIONS(2480),
    [anon_sym_ATsynchronized] = ACTIONS(2480),
    [anon_sym_BOOL] = ACTIONS(2478),
    [anon_sym_IMP] = ACTIONS(2478),
    [anon_sym_SEL] = ACTIONS(2478),
    [anon_sym_Class] = ACTIONS(2478),
    [anon_sym_id] = ACTIONS(2478),
  },
  [STATE(1123)] = {
    [sym_identifier] = ACTIONS(2980),
    [anon_sym_LPAREN2] = ACTIONS(2983),
    [anon_sym_BANG] = ACTIONS(2983),
    [anon_sym_TILDE] = ACTIONS(2983),
    [anon_sym_DASH] = ACTIONS(2985),
    [anon_sym_PLUS] = ACTIONS(2985),
    [anon_sym_STAR] = ACTIONS(2983),
    [anon_sym_CARET] = ACTIONS(2983),
    [anon_sym_AMP] = ACTIONS(2983),
    [anon_sym_SEMI] = ACTIONS(2983),
    [anon_sym___extension__] = ACTIONS(2980),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2989),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_LBRACE] = ACTIONS(2983),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_ATautoreleasepool] = ACTIONS(2983),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2980),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2980),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2980),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2980),
    [anon_sym_union] = ACTIONS(2987),
    [anon_sym_if] = ACTIONS(2985),
    [anon_sym_switch] = ACTIONS(2985),
    [anon_sym_case] = ACTIONS(2985),
    [anon_sym_default] = ACTIONS(2985),
    [anon_sym_while] = ACTIONS(2985),
    [anon_sym_do] = ACTIONS(2985),
    [anon_sym_for] = ACTIONS(2985),
    [anon_sym_in] = ACTIONS(2985),
    [anon_sym_return] = ACTIONS(2985),
    [anon_sym_break] = ACTIONS(2985),
    [anon_sym_continue] = ACTIONS(2985),
    [anon_sym_goto] = ACTIONS(2985),
    [anon_sym_DASH_DASH] = ACTIONS(2983),
    [anon_sym_PLUS_PLUS] = ACTIONS(2983),
    [anon_sym_sizeof] = ACTIONS(2985),
    [anon_sym___alignof__] = ACTIONS(2985),
    [anon_sym___alignof] = ACTIONS(2985),
    [anon_sym__alignof] = ACTIONS(2985),
    [anon_sym_alignof] = ACTIONS(2985),
    [anon_sym__Alignof] = ACTIONS(2985),
    [anon_sym_offsetof] = ACTIONS(2985),
    [anon_sym__Generic] = ACTIONS(2985),
    [anon_sym_asm] = ACTIONS(2985),
    [anon_sym___asm__] = ACTIONS(2985),
    [anon_sym___asm] = ACTIONS(2985),
    [sym_number_literal] = ACTIONS(2983),
    [anon_sym_L_SQUOTE] = ACTIONS(2983),
    [anon_sym_u_SQUOTE] = ACTIONS(2983),
    [anon_sym_U_SQUOTE] = ACTIONS(2983),
    [anon_sym_u8_SQUOTE] = ACTIONS(2983),
    [anon_sym_SQUOTE] = ACTIONS(2983),
    [anon_sym_AT] = ACTIONS(2985),
    [anon_sym_DQUOTE] = ACTIONS(2983),
    [anon_sym_L_DQUOTE] = ACTIONS(2983),
    [anon_sym_u_DQUOTE] = ACTIONS(2983),
    [anon_sym_U_DQUOTE] = ACTIONS(2983),
    [anon_sym_u8_DQUOTE] = ACTIONS(2983),
    [sym_true] = ACTIONS(2985),
    [sym_false] = ACTIONS(2985),
    [anon_sym_NULL] = ACTIONS(2985),
    [anon_sym_nullptr] = ACTIONS(2985),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_ATtry] = ACTIONS(2983),
    [anon_sym___try] = ACTIONS(2985),
    [anon_sym_ATthrow] = ACTIONS(2983),
    [anon_sym_ATselector] = ACTIONS(2983),
    [anon_sym_ATavailable] = ACTIONS(2983),
    [anon_sym___builtin_available] = ACTIONS(2985),
    [anon_sym_va_arg] = ACTIONS(2985),
    [anon_sym_ATencode] = ACTIONS(2983),
    [anon_sym_ATsynchronized] = ACTIONS(2983),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2980),
  },
  [STATE(1124)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1125)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1126)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1127)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1128)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1129)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1130)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1131)] = {
    [sym_identifier] = ACTIONS(2482),
    [anon_sym_LPAREN2] = ACTIONS(2484),
    [anon_sym_BANG] = ACTIONS(2484),
    [anon_sym_TILDE] = ACTIONS(2484),
    [anon_sym_DASH] = ACTIONS(2482),
    [anon_sym_PLUS] = ACTIONS(2482),
    [anon_sym_STAR] = ACTIONS(2484),
    [anon_sym_CARET] = ACTIONS(2484),
    [anon_sym_AMP] = ACTIONS(2484),
    [anon_sym_SEMI] = ACTIONS(2484),
    [anon_sym___extension__] = ACTIONS(2482),
    [anon_sym_typedef] = ACTIONS(2482),
    [anon_sym_extern] = ACTIONS(2482),
    [anon_sym___attribute__] = ACTIONS(2482),
    [anon_sym___attribute] = ACTIONS(2482),
    [anon_sym_noreturn] = ACTIONS(2482),
    [anon_sym_LBRACK] = ACTIONS(2484),
    [anon_sym___declspec] = ACTIONS(2482),
    [anon_sym_LBRACE] = ACTIONS(2484),
    [anon_sym_signed] = ACTIONS(2482),
    [anon_sym_unsigned] = ACTIONS(2482),
    [anon_sym_long] = ACTIONS(2482),
    [anon_sym_short] = ACTIONS(2482),
    [anon_sym_static] = ACTIONS(2482),
    [anon_sym_ATautoreleasepool] = ACTIONS(2484),
    [anon_sym_auto] = ACTIONS(2482),
    [anon_sym_register] = ACTIONS(2482),
    [anon_sym_inline] = ACTIONS(2482),
    [anon_sym___inline] = ACTIONS(2482),
    [anon_sym___inline__] = ACTIONS(2482),
    [anon_sym___forceinline] = ACTIONS(2482),
    [anon_sym_thread_local] = ACTIONS(2482),
    [anon_sym___thread] = ACTIONS(2482),
    [anon_sym_CG_EXTERN] = ACTIONS(2482),
    [anon_sym_CG_INLINE] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2482),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2482),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2482),
    [anon_sym_IBOutlet] = ACTIONS(2482),
    [anon_sym_IBInspectable] = ACTIONS(2482),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2482),
    [anon_sym_NS_INLINE] = ACTIONS(2482),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2482),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2482),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2482),
    [anon_sym_const] = ACTIONS(2482),
    [anon_sym_constexpr] = ACTIONS(2482),
    [anon_sym_volatile] = ACTIONS(2482),
    [anon_sym_restrict] = ACTIONS(2482),
    [anon_sym___restrict__] = ACTIONS(2482),
    [anon_sym__Atomic] = ACTIONS(2482),
    [anon_sym__Noreturn] = ACTIONS(2482),
    [anon_sym__Nonnull] = ACTIONS(2482),
    [anon_sym_nullable] = ACTIONS(2482),
    [anon_sym__Complex] = ACTIONS(2482),
    [anon_sym__Nullable] = ACTIONS(2482),
    [anon_sym__Nullable_result] = ACTIONS(2482),
    [anon_sym__Null_unspecified] = ACTIONS(2482),
    [anon_sym___autoreleasing] = ACTIONS(2482),
    [anon_sym___block] = ACTIONS(2482),
    [anon_sym___bridge] = ACTIONS(2482),
    [anon_sym___bridge_retained] = ACTIONS(2482),
    [anon_sym___bridge_transfer] = ACTIONS(2482),
    [anon_sym___complex] = ACTIONS(2482),
    [anon_sym___const] = ACTIONS(2482),
    [anon_sym___imag] = ACTIONS(2482),
    [anon_sym___kindof] = ACTIONS(2482),
    [anon_sym___nonnull] = ACTIONS(2482),
    [anon_sym___nullable] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2482),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2482),
    [anon_sym___real] = ACTIONS(2482),
    [anon_sym___strong] = ACTIONS(2482),
    [anon_sym___unsafe_unretained] = ACTIONS(2482),
    [anon_sym___unused] = ACTIONS(2482),
    [anon_sym___weak] = ACTIONS(2482),
    [anon_sym_alignas] = ACTIONS(2482),
    [anon_sym__Alignas] = ACTIONS(2482),
    [sym_primitive_type] = ACTIONS(2482),
    [anon_sym_enum] = ACTIONS(2482),
    [anon_sym_struct] = ACTIONS(2482),
    [anon_sym_union] = ACTIONS(2482),
    [anon_sym_if] = ACTIONS(2482),
    [anon_sym_else] = ACTIONS(2482),
    [anon_sym_switch] = ACTIONS(2482),
    [anon_sym_while] = ACTIONS(2482),
    [anon_sym_do] = ACTIONS(2482),
    [anon_sym_for] = ACTIONS(2482),
    [anon_sym_in] = ACTIONS(2482),
    [anon_sym_return] = ACTIONS(2482),
    [anon_sym_break] = ACTIONS(2482),
    [anon_sym_continue] = ACTIONS(2482),
    [anon_sym_goto] = ACTIONS(2482),
    [anon_sym_DASH_DASH] = ACTIONS(2484),
    [anon_sym_PLUS_PLUS] = ACTIONS(2484),
    [anon_sym_sizeof] = ACTIONS(2482),
    [anon_sym___alignof__] = ACTIONS(2482),
    [anon_sym___alignof] = ACTIONS(2482),
    [anon_sym__alignof] = ACTIONS(2482),
    [anon_sym_alignof] = ACTIONS(2482),
    [anon_sym__Alignof] = ACTIONS(2482),
    [anon_sym_offsetof] = ACTIONS(2482),
    [anon_sym__Generic] = ACTIONS(2482),
    [anon_sym_asm] = ACTIONS(2482),
    [anon_sym___asm__] = ACTIONS(2482),
    [anon_sym___asm] = ACTIONS(2482),
    [sym_number_literal] = ACTIONS(2484),
    [anon_sym_L_SQUOTE] = ACTIONS(2484),
    [anon_sym_u_SQUOTE] = ACTIONS(2484),
    [anon_sym_U_SQUOTE] = ACTIONS(2484),
    [anon_sym_u8_SQUOTE] = ACTIONS(2484),
    [anon_sym_SQUOTE] = ACTIONS(2484),
    [anon_sym_AT] = ACTIONS(2482),
    [anon_sym_DQUOTE] = ACTIONS(2484),
    [anon_sym_L_DQUOTE] = ACTIONS(2484),
    [anon_sym_u_DQUOTE] = ACTIONS(2484),
    [anon_sym_U_DQUOTE] = ACTIONS(2484),
    [anon_sym_u8_DQUOTE] = ACTIONS(2484),
    [sym_true] = ACTIONS(2482),
    [sym_false] = ACTIONS(2482),
    [anon_sym_NULL] = ACTIONS(2482),
    [anon_sym_nullptr] = ACTIONS(2482),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2482),
    [anon_sym___typeof] = ACTIONS(2482),
    [anon_sym_typeof] = ACTIONS(2482),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2482),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2482),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2482),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2482),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2482),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_API_AVAILABLE] = ACTIONS(2482),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_API_DEPRECATED] = ACTIONS(2482),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2482),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2482),
    [anon_sym___deprecated_msg] = ACTIONS(2482),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2482),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2482),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2482),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2482),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2482),
    [anon_sym_ATtry] = ACTIONS(2484),
    [anon_sym___try] = ACTIONS(2482),
    [anon_sym_ATthrow] = ACTIONS(2484),
    [anon_sym_ATselector] = ACTIONS(2484),
    [anon_sym_ATavailable] = ACTIONS(2484),
    [anon_sym___builtin_available] = ACTIONS(2482),
    [anon_sym_va_arg] = ACTIONS(2482),
    [anon_sym_ATencode] = ACTIONS(2484),
    [anon_sym_ATsynchronized] = ACTIONS(2484),
    [anon_sym_BOOL] = ACTIONS(2482),
    [anon_sym_IMP] = ACTIONS(2482),
    [anon_sym_SEL] = ACTIONS(2482),
    [anon_sym_Class] = ACTIONS(2482),
    [anon_sym_id] = ACTIONS(2482),
  },
  [STATE(1132)] = {
    [sym_identifier] = ACTIONS(2486),
    [anon_sym_LPAREN2] = ACTIONS(2488),
    [anon_sym_BANG] = ACTIONS(2488),
    [anon_sym_TILDE] = ACTIONS(2488),
    [anon_sym_DASH] = ACTIONS(2486),
    [anon_sym_PLUS] = ACTIONS(2486),
    [anon_sym_STAR] = ACTIONS(2488),
    [anon_sym_CARET] = ACTIONS(2488),
    [anon_sym_AMP] = ACTIONS(2488),
    [anon_sym_SEMI] = ACTIONS(2488),
    [anon_sym___extension__] = ACTIONS(2486),
    [anon_sym_typedef] = ACTIONS(2486),
    [anon_sym_extern] = ACTIONS(2486),
    [anon_sym___attribute__] = ACTIONS(2486),
    [anon_sym___attribute] = ACTIONS(2486),
    [anon_sym_noreturn] = ACTIONS(2486),
    [anon_sym_LBRACK] = ACTIONS(2488),
    [anon_sym___declspec] = ACTIONS(2486),
    [anon_sym_LBRACE] = ACTIONS(2488),
    [anon_sym_signed] = ACTIONS(2486),
    [anon_sym_unsigned] = ACTIONS(2486),
    [anon_sym_long] = ACTIONS(2486),
    [anon_sym_short] = ACTIONS(2486),
    [anon_sym_static] = ACTIONS(2486),
    [anon_sym_ATautoreleasepool] = ACTIONS(2488),
    [anon_sym_auto] = ACTIONS(2486),
    [anon_sym_register] = ACTIONS(2486),
    [anon_sym_inline] = ACTIONS(2486),
    [anon_sym___inline] = ACTIONS(2486),
    [anon_sym___inline__] = ACTIONS(2486),
    [anon_sym___forceinline] = ACTIONS(2486),
    [anon_sym_thread_local] = ACTIONS(2486),
    [anon_sym___thread] = ACTIONS(2486),
    [anon_sym_CG_EXTERN] = ACTIONS(2486),
    [anon_sym_CG_INLINE] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2486),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2486),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2486),
    [anon_sym_IBOutlet] = ACTIONS(2486),
    [anon_sym_IBInspectable] = ACTIONS(2486),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2486),
    [anon_sym_NS_INLINE] = ACTIONS(2486),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2486),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2486),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2486),
    [anon_sym_const] = ACTIONS(2486),
    [anon_sym_constexpr] = ACTIONS(2486),
    [anon_sym_volatile] = ACTIONS(2486),
    [anon_sym_restrict] = ACTIONS(2486),
    [anon_sym___restrict__] = ACTIONS(2486),
    [anon_sym__Atomic] = ACTIONS(2486),
    [anon_sym__Noreturn] = ACTIONS(2486),
    [anon_sym__Nonnull] = ACTIONS(2486),
    [anon_sym_nullable] = ACTIONS(2486),
    [anon_sym__Complex] = ACTIONS(2486),
    [anon_sym__Nullable] = ACTIONS(2486),
    [anon_sym__Nullable_result] = ACTIONS(2486),
    [anon_sym__Null_unspecified] = ACTIONS(2486),
    [anon_sym___autoreleasing] = ACTIONS(2486),
    [anon_sym___block] = ACTIONS(2486),
    [anon_sym___bridge] = ACTIONS(2486),
    [anon_sym___bridge_retained] = ACTIONS(2486),
    [anon_sym___bridge_transfer] = ACTIONS(2486),
    [anon_sym___complex] = ACTIONS(2486),
    [anon_sym___const] = ACTIONS(2486),
    [anon_sym___imag] = ACTIONS(2486),
    [anon_sym___kindof] = ACTIONS(2486),
    [anon_sym___nonnull] = ACTIONS(2486),
    [anon_sym___nullable] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2486),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2486),
    [anon_sym___real] = ACTIONS(2486),
    [anon_sym___strong] = ACTIONS(2486),
    [anon_sym___unsafe_unretained] = ACTIONS(2486),
    [anon_sym___unused] = ACTIONS(2486),
    [anon_sym___weak] = ACTIONS(2486),
    [anon_sym_alignas] = ACTIONS(2486),
    [anon_sym__Alignas] = ACTIONS(2486),
    [sym_primitive_type] = ACTIONS(2486),
    [anon_sym_enum] = ACTIONS(2486),
    [anon_sym_struct] = ACTIONS(2486),
    [anon_sym_union] = ACTIONS(2486),
    [anon_sym_if] = ACTIONS(2486),
    [anon_sym_else] = ACTIONS(2486),
    [anon_sym_switch] = ACTIONS(2486),
    [anon_sym_while] = ACTIONS(2486),
    [anon_sym_do] = ACTIONS(2486),
    [anon_sym_for] = ACTIONS(2486),
    [anon_sym_in] = ACTIONS(2486),
    [anon_sym_return] = ACTIONS(2486),
    [anon_sym_break] = ACTIONS(2486),
    [anon_sym_continue] = ACTIONS(2486),
    [anon_sym_goto] = ACTIONS(2486),
    [anon_sym_DASH_DASH] = ACTIONS(2488),
    [anon_sym_PLUS_PLUS] = ACTIONS(2488),
    [anon_sym_sizeof] = ACTIONS(2486),
    [anon_sym___alignof__] = ACTIONS(2486),
    [anon_sym___alignof] = ACTIONS(2486),
    [anon_sym__alignof] = ACTIONS(2486),
    [anon_sym_alignof] = ACTIONS(2486),
    [anon_sym__Alignof] = ACTIONS(2486),
    [anon_sym_offsetof] = ACTIONS(2486),
    [anon_sym__Generic] = ACTIONS(2486),
    [anon_sym_asm] = ACTIONS(2486),
    [anon_sym___asm__] = ACTIONS(2486),
    [anon_sym___asm] = ACTIONS(2486),
    [sym_number_literal] = ACTIONS(2488),
    [anon_sym_L_SQUOTE] = ACTIONS(2488),
    [anon_sym_u_SQUOTE] = ACTIONS(2488),
    [anon_sym_U_SQUOTE] = ACTIONS(2488),
    [anon_sym_u8_SQUOTE] = ACTIONS(2488),
    [anon_sym_SQUOTE] = ACTIONS(2488),
    [anon_sym_AT] = ACTIONS(2486),
    [anon_sym_DQUOTE] = ACTIONS(2488),
    [anon_sym_L_DQUOTE] = ACTIONS(2488),
    [anon_sym_u_DQUOTE] = ACTIONS(2488),
    [anon_sym_U_DQUOTE] = ACTIONS(2488),
    [anon_sym_u8_DQUOTE] = ACTIONS(2488),
    [sym_true] = ACTIONS(2486),
    [sym_false] = ACTIONS(2486),
    [anon_sym_NULL] = ACTIONS(2486),
    [anon_sym_nullptr] = ACTIONS(2486),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2486),
    [anon_sym___typeof] = ACTIONS(2486),
    [anon_sym_typeof] = ACTIONS(2486),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2486),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2486),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2486),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2486),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2486),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_API_AVAILABLE] = ACTIONS(2486),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_API_DEPRECATED] = ACTIONS(2486),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2486),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2486),
    [anon_sym___deprecated_msg] = ACTIONS(2486),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2486),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2486),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2486),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2486),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2486),
    [anon_sym_ATtry] = ACTIONS(2488),
    [anon_sym___try] = ACTIONS(2486),
    [anon_sym_ATthrow] = ACTIONS(2488),
    [anon_sym_ATselector] = ACTIONS(2488),
    [anon_sym_ATavailable] = ACTIONS(2488),
    [anon_sym___builtin_available] = ACTIONS(2486),
    [anon_sym_va_arg] = ACTIONS(2486),
    [anon_sym_ATencode] = ACTIONS(2488),
    [anon_sym_ATsynchronized] = ACTIONS(2488),
    [anon_sym_BOOL] = ACTIONS(2486),
    [anon_sym_IMP] = ACTIONS(2486),
    [anon_sym_SEL] = ACTIONS(2486),
    [anon_sym_Class] = ACTIONS(2486),
    [anon_sym_id] = ACTIONS(2486),
  },
  [STATE(1133)] = {
    [sym_identifier] = ACTIONS(2490),
    [anon_sym_LPAREN2] = ACTIONS(2492),
    [anon_sym_BANG] = ACTIONS(2492),
    [anon_sym_TILDE] = ACTIONS(2492),
    [anon_sym_DASH] = ACTIONS(2490),
    [anon_sym_PLUS] = ACTIONS(2490),
    [anon_sym_STAR] = ACTIONS(2492),
    [anon_sym_CARET] = ACTIONS(2492),
    [anon_sym_AMP] = ACTIONS(2492),
    [anon_sym_SEMI] = ACTIONS(2492),
    [anon_sym___extension__] = ACTIONS(2490),
    [anon_sym_typedef] = ACTIONS(2490),
    [anon_sym_extern] = ACTIONS(2490),
    [anon_sym___attribute__] = ACTIONS(2490),
    [anon_sym___attribute] = ACTIONS(2490),
    [anon_sym_noreturn] = ACTIONS(2490),
    [anon_sym_LBRACK] = ACTIONS(2492),
    [anon_sym___declspec] = ACTIONS(2490),
    [anon_sym_LBRACE] = ACTIONS(2492),
    [anon_sym_signed] = ACTIONS(2490),
    [anon_sym_unsigned] = ACTIONS(2490),
    [anon_sym_long] = ACTIONS(2490),
    [anon_sym_short] = ACTIONS(2490),
    [anon_sym_static] = ACTIONS(2490),
    [anon_sym_ATautoreleasepool] = ACTIONS(2492),
    [anon_sym_auto] = ACTIONS(2490),
    [anon_sym_register] = ACTIONS(2490),
    [anon_sym_inline] = ACTIONS(2490),
    [anon_sym___inline] = ACTIONS(2490),
    [anon_sym___inline__] = ACTIONS(2490),
    [anon_sym___forceinline] = ACTIONS(2490),
    [anon_sym_thread_local] = ACTIONS(2490),
    [anon_sym___thread] = ACTIONS(2490),
    [anon_sym_CG_EXTERN] = ACTIONS(2490),
    [anon_sym_CG_INLINE] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2490),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2490),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2490),
    [anon_sym_IBOutlet] = ACTIONS(2490),
    [anon_sym_IBInspectable] = ACTIONS(2490),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2490),
    [anon_sym_NS_INLINE] = ACTIONS(2490),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2490),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2490),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2490),
    [anon_sym_const] = ACTIONS(2490),
    [anon_sym_constexpr] = ACTIONS(2490),
    [anon_sym_volatile] = ACTIONS(2490),
    [anon_sym_restrict] = ACTIONS(2490),
    [anon_sym___restrict__] = ACTIONS(2490),
    [anon_sym__Atomic] = ACTIONS(2490),
    [anon_sym__Noreturn] = ACTIONS(2490),
    [anon_sym__Nonnull] = ACTIONS(2490),
    [anon_sym_nullable] = ACTIONS(2490),
    [anon_sym__Complex] = ACTIONS(2490),
    [anon_sym__Nullable] = ACTIONS(2490),
    [anon_sym__Nullable_result] = ACTIONS(2490),
    [anon_sym__Null_unspecified] = ACTIONS(2490),
    [anon_sym___autoreleasing] = ACTIONS(2490),
    [anon_sym___block] = ACTIONS(2490),
    [anon_sym___bridge] = ACTIONS(2490),
    [anon_sym___bridge_retained] = ACTIONS(2490),
    [anon_sym___bridge_transfer] = ACTIONS(2490),
    [anon_sym___complex] = ACTIONS(2490),
    [anon_sym___const] = ACTIONS(2490),
    [anon_sym___imag] = ACTIONS(2490),
    [anon_sym___kindof] = ACTIONS(2490),
    [anon_sym___nonnull] = ACTIONS(2490),
    [anon_sym___nullable] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2490),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2490),
    [anon_sym___real] = ACTIONS(2490),
    [anon_sym___strong] = ACTIONS(2490),
    [anon_sym___unsafe_unretained] = ACTIONS(2490),
    [anon_sym___unused] = ACTIONS(2490),
    [anon_sym___weak] = ACTIONS(2490),
    [anon_sym_alignas] = ACTIONS(2490),
    [anon_sym__Alignas] = ACTIONS(2490),
    [sym_primitive_type] = ACTIONS(2490),
    [anon_sym_enum] = ACTIONS(2490),
    [anon_sym_struct] = ACTIONS(2490),
    [anon_sym_union] = ACTIONS(2490),
    [anon_sym_if] = ACTIONS(2490),
    [anon_sym_else] = ACTIONS(2490),
    [anon_sym_switch] = ACTIONS(2490),
    [anon_sym_while] = ACTIONS(2490),
    [anon_sym_do] = ACTIONS(2490),
    [anon_sym_for] = ACTIONS(2490),
    [anon_sym_in] = ACTIONS(2490),
    [anon_sym_return] = ACTIONS(2490),
    [anon_sym_break] = ACTIONS(2490),
    [anon_sym_continue] = ACTIONS(2490),
    [anon_sym_goto] = ACTIONS(2490),
    [anon_sym_DASH_DASH] = ACTIONS(2492),
    [anon_sym_PLUS_PLUS] = ACTIONS(2492),
    [anon_sym_sizeof] = ACTIONS(2490),
    [anon_sym___alignof__] = ACTIONS(2490),
    [anon_sym___alignof] = ACTIONS(2490),
    [anon_sym__alignof] = ACTIONS(2490),
    [anon_sym_alignof] = ACTIONS(2490),
    [anon_sym__Alignof] = ACTIONS(2490),
    [anon_sym_offsetof] = ACTIONS(2490),
    [anon_sym__Generic] = ACTIONS(2490),
    [anon_sym_asm] = ACTIONS(2490),
    [anon_sym___asm__] = ACTIONS(2490),
    [anon_sym___asm] = ACTIONS(2490),
    [sym_number_literal] = ACTIONS(2492),
    [anon_sym_L_SQUOTE] = ACTIONS(2492),
    [anon_sym_u_SQUOTE] = ACTIONS(2492),
    [anon_sym_U_SQUOTE] = ACTIONS(2492),
    [anon_sym_u8_SQUOTE] = ACTIONS(2492),
    [anon_sym_SQUOTE] = ACTIONS(2492),
    [anon_sym_AT] = ACTIONS(2490),
    [anon_sym_DQUOTE] = ACTIONS(2492),
    [anon_sym_L_DQUOTE] = ACTIONS(2492),
    [anon_sym_u_DQUOTE] = ACTIONS(2492),
    [anon_sym_U_DQUOTE] = ACTIONS(2492),
    [anon_sym_u8_DQUOTE] = ACTIONS(2492),
    [sym_true] = ACTIONS(2490),
    [sym_false] = ACTIONS(2490),
    [anon_sym_NULL] = ACTIONS(2490),
    [anon_sym_nullptr] = ACTIONS(2490),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2490),
    [anon_sym___typeof] = ACTIONS(2490),
    [anon_sym_typeof] = ACTIONS(2490),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2490),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2490),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2490),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2490),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2490),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_API_AVAILABLE] = ACTIONS(2490),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_API_DEPRECATED] = ACTIONS(2490),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2490),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2490),
    [anon_sym___deprecated_msg] = ACTIONS(2490),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2490),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2490),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2490),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2490),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2490),
    [anon_sym_ATtry] = ACTIONS(2492),
    [anon_sym___try] = ACTIONS(2490),
    [anon_sym_ATthrow] = ACTIONS(2492),
    [anon_sym_ATselector] = ACTIONS(2492),
    [anon_sym_ATavailable] = ACTIONS(2492),
    [anon_sym___builtin_available] = ACTIONS(2490),
    [anon_sym_va_arg] = ACTIONS(2490),
    [anon_sym_ATencode] = ACTIONS(2492),
    [anon_sym_ATsynchronized] = ACTIONS(2492),
    [anon_sym_BOOL] = ACTIONS(2490),
    [anon_sym_IMP] = ACTIONS(2490),
    [anon_sym_SEL] = ACTIONS(2490),
    [anon_sym_Class] = ACTIONS(2490),
    [anon_sym_id] = ACTIONS(2490),
  },
  [STATE(1134)] = {
    [sym_identifier] = ACTIONS(2466),
    [anon_sym_LPAREN2] = ACTIONS(2468),
    [anon_sym_BANG] = ACTIONS(2468),
    [anon_sym_TILDE] = ACTIONS(2468),
    [anon_sym_DASH] = ACTIONS(2466),
    [anon_sym_PLUS] = ACTIONS(2466),
    [anon_sym_STAR] = ACTIONS(2468),
    [anon_sym_CARET] = ACTIONS(2468),
    [anon_sym_AMP] = ACTIONS(2468),
    [anon_sym_SEMI] = ACTIONS(2468),
    [anon_sym___extension__] = ACTIONS(2466),
    [anon_sym_typedef] = ACTIONS(2466),
    [anon_sym_extern] = ACTIONS(2466),
    [anon_sym___attribute__] = ACTIONS(2466),
    [anon_sym___attribute] = ACTIONS(2466),
    [anon_sym_noreturn] = ACTIONS(2466),
    [anon_sym_LBRACK] = ACTIONS(2468),
    [anon_sym___declspec] = ACTIONS(2466),
    [anon_sym_LBRACE] = ACTIONS(2468),
    [anon_sym_signed] = ACTIONS(2466),
    [anon_sym_unsigned] = ACTIONS(2466),
    [anon_sym_long] = ACTIONS(2466),
    [anon_sym_short] = ACTIONS(2466),
    [anon_sym_static] = ACTIONS(2466),
    [anon_sym_ATautoreleasepool] = ACTIONS(2468),
    [anon_sym_auto] = ACTIONS(2466),
    [anon_sym_register] = ACTIONS(2466),
    [anon_sym_inline] = ACTIONS(2466),
    [anon_sym___inline] = ACTIONS(2466),
    [anon_sym___inline__] = ACTIONS(2466),
    [anon_sym___forceinline] = ACTIONS(2466),
    [anon_sym_thread_local] = ACTIONS(2466),
    [anon_sym___thread] = ACTIONS(2466),
    [anon_sym_CG_EXTERN] = ACTIONS(2466),
    [anon_sym_CG_INLINE] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2466),
    [anon_sym_IBOutlet] = ACTIONS(2466),
    [anon_sym_IBInspectable] = ACTIONS(2466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2466),
    [anon_sym_NS_INLINE] = ACTIONS(2466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2466),
    [anon_sym_const] = ACTIONS(2466),
    [anon_sym_constexpr] = ACTIONS(2466),
    [anon_sym_volatile] = ACTIONS(2466),
    [anon_sym_restrict] = ACTIONS(2466),
    [anon_sym___restrict__] = ACTIONS(2466),
    [anon_sym__Atomic] = ACTIONS(2466),
    [anon_sym__Noreturn] = ACTIONS(2466),
    [anon_sym__Nonnull] = ACTIONS(2466),
    [anon_sym_nullable] = ACTIONS(2466),
    [anon_sym__Complex] = ACTIONS(2466),
    [anon_sym__Nullable] = ACTIONS(2466),
    [anon_sym__Nullable_result] = ACTIONS(2466),
    [anon_sym__Null_unspecified] = ACTIONS(2466),
    [anon_sym___autoreleasing] = ACTIONS(2466),
    [anon_sym___block] = ACTIONS(2466),
    [anon_sym___bridge] = ACTIONS(2466),
    [anon_sym___bridge_retained] = ACTIONS(2466),
    [anon_sym___bridge_transfer] = ACTIONS(2466),
    [anon_sym___complex] = ACTIONS(2466),
    [anon_sym___const] = ACTIONS(2466),
    [anon_sym___imag] = ACTIONS(2466),
    [anon_sym___kindof] = ACTIONS(2466),
    [anon_sym___nonnull] = ACTIONS(2466),
    [anon_sym___nullable] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2466),
    [anon_sym___real] = ACTIONS(2466),
    [anon_sym___strong] = ACTIONS(2466),
    [anon_sym___unsafe_unretained] = ACTIONS(2466),
    [anon_sym___unused] = ACTIONS(2466),
    [anon_sym___weak] = ACTIONS(2466),
    [anon_sym_alignas] = ACTIONS(2466),
    [anon_sym__Alignas] = ACTIONS(2466),
    [sym_primitive_type] = ACTIONS(2466),
    [anon_sym_enum] = ACTIONS(2466),
    [anon_sym_struct] = ACTIONS(2466),
    [anon_sym_union] = ACTIONS(2466),
    [anon_sym_if] = ACTIONS(2466),
    [anon_sym_else] = ACTIONS(2466),
    [anon_sym_switch] = ACTIONS(2466),
    [anon_sym_while] = ACTIONS(2466),
    [anon_sym_do] = ACTIONS(2466),
    [anon_sym_for] = ACTIONS(2466),
    [anon_sym_in] = ACTIONS(2466),
    [anon_sym_return] = ACTIONS(2466),
    [anon_sym_break] = ACTIONS(2466),
    [anon_sym_continue] = ACTIONS(2466),
    [anon_sym_goto] = ACTIONS(2466),
    [anon_sym_DASH_DASH] = ACTIONS(2468),
    [anon_sym_PLUS_PLUS] = ACTIONS(2468),
    [anon_sym_sizeof] = ACTIONS(2466),
    [anon_sym___alignof__] = ACTIONS(2466),
    [anon_sym___alignof] = ACTIONS(2466),
    [anon_sym__alignof] = ACTIONS(2466),
    [anon_sym_alignof] = ACTIONS(2466),
    [anon_sym__Alignof] = ACTIONS(2466),
    [anon_sym_offsetof] = ACTIONS(2466),
    [anon_sym__Generic] = ACTIONS(2466),
    [anon_sym_asm] = ACTIONS(2466),
    [anon_sym___asm__] = ACTIONS(2466),
    [anon_sym___asm] = ACTIONS(2466),
    [sym_number_literal] = ACTIONS(2468),
    [anon_sym_L_SQUOTE] = ACTIONS(2468),
    [anon_sym_u_SQUOTE] = ACTIONS(2468),
    [anon_sym_U_SQUOTE] = ACTIONS(2468),
    [anon_sym_u8_SQUOTE] = ACTIONS(2468),
    [anon_sym_SQUOTE] = ACTIONS(2468),
    [anon_sym_AT] = ACTIONS(2466),
    [anon_sym_DQUOTE] = ACTIONS(2468),
    [anon_sym_L_DQUOTE] = ACTIONS(2468),
    [anon_sym_u_DQUOTE] = ACTIONS(2468),
    [anon_sym_U_DQUOTE] = ACTIONS(2468),
    [anon_sym_u8_DQUOTE] = ACTIONS(2468),
    [sym_true] = ACTIONS(2466),
    [sym_false] = ACTIONS(2466),
    [anon_sym_NULL] = ACTIONS(2466),
    [anon_sym_nullptr] = ACTIONS(2466),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2466),
    [anon_sym___typeof] = ACTIONS(2466),
    [anon_sym_typeof] = ACTIONS(2466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_API_AVAILABLE] = ACTIONS(2466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_API_DEPRECATED] = ACTIONS(2466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2466),
    [anon_sym___deprecated_msg] = ACTIONS(2466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2466),
    [anon_sym_ATtry] = ACTIONS(2468),
    [anon_sym___try] = ACTIONS(2466),
    [anon_sym_ATthrow] = ACTIONS(2468),
    [anon_sym_ATselector] = ACTIONS(2468),
    [anon_sym_ATavailable] = ACTIONS(2468),
    [anon_sym___builtin_available] = ACTIONS(2466),
    [anon_sym_va_arg] = ACTIONS(2466),
    [anon_sym_ATencode] = ACTIONS(2468),
    [anon_sym_ATsynchronized] = ACTIONS(2468),
    [anon_sym_BOOL] = ACTIONS(2466),
    [anon_sym_IMP] = ACTIONS(2466),
    [anon_sym_SEL] = ACTIONS(2466),
    [anon_sym_Class] = ACTIONS(2466),
    [anon_sym_id] = ACTIONS(2466),
  },
  [STATE(1135)] = {
    [sym_identifier] = ACTIONS(2498),
    [anon_sym_LPAREN2] = ACTIONS(2500),
    [anon_sym_BANG] = ACTIONS(2500),
    [anon_sym_TILDE] = ACTIONS(2500),
    [anon_sym_DASH] = ACTIONS(2498),
    [anon_sym_PLUS] = ACTIONS(2498),
    [anon_sym_STAR] = ACTIONS(2500),
    [anon_sym_CARET] = ACTIONS(2500),
    [anon_sym_AMP] = ACTIONS(2500),
    [anon_sym_SEMI] = ACTIONS(2500),
    [anon_sym___extension__] = ACTIONS(2498),
    [anon_sym_typedef] = ACTIONS(2498),
    [anon_sym_extern] = ACTIONS(2498),
    [anon_sym___attribute__] = ACTIONS(2498),
    [anon_sym___attribute] = ACTIONS(2498),
    [anon_sym_noreturn] = ACTIONS(2498),
    [anon_sym_LBRACK] = ACTIONS(2500),
    [anon_sym___declspec] = ACTIONS(2498),
    [anon_sym_LBRACE] = ACTIONS(2500),
    [anon_sym_signed] = ACTIONS(2498),
    [anon_sym_unsigned] = ACTIONS(2498),
    [anon_sym_long] = ACTIONS(2498),
    [anon_sym_short] = ACTIONS(2498),
    [anon_sym_static] = ACTIONS(2498),
    [anon_sym_ATautoreleasepool] = ACTIONS(2500),
    [anon_sym_auto] = ACTIONS(2498),
    [anon_sym_register] = ACTIONS(2498),
    [anon_sym_inline] = ACTIONS(2498),
    [anon_sym___inline] = ACTIONS(2498),
    [anon_sym___inline__] = ACTIONS(2498),
    [anon_sym___forceinline] = ACTIONS(2498),
    [anon_sym_thread_local] = ACTIONS(2498),
    [anon_sym___thread] = ACTIONS(2498),
    [anon_sym_CG_EXTERN] = ACTIONS(2498),
    [anon_sym_CG_INLINE] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2498),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2498),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2498),
    [anon_sym_IBOutlet] = ACTIONS(2498),
    [anon_sym_IBInspectable] = ACTIONS(2498),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2498),
    [anon_sym_NS_INLINE] = ACTIONS(2498),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2498),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2498),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2498),
    [anon_sym_const] = ACTIONS(2498),
    [anon_sym_constexpr] = ACTIONS(2498),
    [anon_sym_volatile] = ACTIONS(2498),
    [anon_sym_restrict] = ACTIONS(2498),
    [anon_sym___restrict__] = ACTIONS(2498),
    [anon_sym__Atomic] = ACTIONS(2498),
    [anon_sym__Noreturn] = ACTIONS(2498),
    [anon_sym__Nonnull] = ACTIONS(2498),
    [anon_sym_nullable] = ACTIONS(2498),
    [anon_sym__Complex] = ACTIONS(2498),
    [anon_sym__Nullable] = ACTIONS(2498),
    [anon_sym__Nullable_result] = ACTIONS(2498),
    [anon_sym__Null_unspecified] = ACTIONS(2498),
    [anon_sym___autoreleasing] = ACTIONS(2498),
    [anon_sym___block] = ACTIONS(2498),
    [anon_sym___bridge] = ACTIONS(2498),
    [anon_sym___bridge_retained] = ACTIONS(2498),
    [anon_sym___bridge_transfer] = ACTIONS(2498),
    [anon_sym___complex] = ACTIONS(2498),
    [anon_sym___const] = ACTIONS(2498),
    [anon_sym___imag] = ACTIONS(2498),
    [anon_sym___kindof] = ACTIONS(2498),
    [anon_sym___nonnull] = ACTIONS(2498),
    [anon_sym___nullable] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2498),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2498),
    [anon_sym___real] = ACTIONS(2498),
    [anon_sym___strong] = ACTIONS(2498),
    [anon_sym___unsafe_unretained] = ACTIONS(2498),
    [anon_sym___unused] = ACTIONS(2498),
    [anon_sym___weak] = ACTIONS(2498),
    [anon_sym_alignas] = ACTIONS(2498),
    [anon_sym__Alignas] = ACTIONS(2498),
    [sym_primitive_type] = ACTIONS(2498),
    [anon_sym_enum] = ACTIONS(2498),
    [anon_sym_struct] = ACTIONS(2498),
    [anon_sym_union] = ACTIONS(2498),
    [anon_sym_if] = ACTIONS(2498),
    [anon_sym_else] = ACTIONS(2498),
    [anon_sym_switch] = ACTIONS(2498),
    [anon_sym_while] = ACTIONS(2498),
    [anon_sym_do] = ACTIONS(2498),
    [anon_sym_for] = ACTIONS(2498),
    [anon_sym_in] = ACTIONS(2498),
    [anon_sym_return] = ACTIONS(2498),
    [anon_sym_break] = ACTIONS(2498),
    [anon_sym_continue] = ACTIONS(2498),
    [anon_sym_goto] = ACTIONS(2498),
    [anon_sym_DASH_DASH] = ACTIONS(2500),
    [anon_sym_PLUS_PLUS] = ACTIONS(2500),
    [anon_sym_sizeof] = ACTIONS(2498),
    [anon_sym___alignof__] = ACTIONS(2498),
    [anon_sym___alignof] = ACTIONS(2498),
    [anon_sym__alignof] = ACTIONS(2498),
    [anon_sym_alignof] = ACTIONS(2498),
    [anon_sym__Alignof] = ACTIONS(2498),
    [anon_sym_offsetof] = ACTIONS(2498),
    [anon_sym__Generic] = ACTIONS(2498),
    [anon_sym_asm] = ACTIONS(2498),
    [anon_sym___asm__] = ACTIONS(2498),
    [anon_sym___asm] = ACTIONS(2498),
    [sym_number_literal] = ACTIONS(2500),
    [anon_sym_L_SQUOTE] = ACTIONS(2500),
    [anon_sym_u_SQUOTE] = ACTIONS(2500),
    [anon_sym_U_SQUOTE] = ACTIONS(2500),
    [anon_sym_u8_SQUOTE] = ACTIONS(2500),
    [anon_sym_SQUOTE] = ACTIONS(2500),
    [anon_sym_AT] = ACTIONS(2498),
    [anon_sym_DQUOTE] = ACTIONS(2500),
    [anon_sym_L_DQUOTE] = ACTIONS(2500),
    [anon_sym_u_DQUOTE] = ACTIONS(2500),
    [anon_sym_U_DQUOTE] = ACTIONS(2500),
    [anon_sym_u8_DQUOTE] = ACTIONS(2500),
    [sym_true] = ACTIONS(2498),
    [sym_false] = ACTIONS(2498),
    [anon_sym_NULL] = ACTIONS(2498),
    [anon_sym_nullptr] = ACTIONS(2498),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2498),
    [anon_sym___typeof] = ACTIONS(2498),
    [anon_sym_typeof] = ACTIONS(2498),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2498),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2498),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2498),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2498),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2498),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_API_AVAILABLE] = ACTIONS(2498),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_API_DEPRECATED] = ACTIONS(2498),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2498),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2498),
    [anon_sym___deprecated_msg] = ACTIONS(2498),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2498),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2498),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2498),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2498),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2498),
    [anon_sym_ATtry] = ACTIONS(2500),
    [anon_sym___try] = ACTIONS(2498),
    [anon_sym_ATthrow] = ACTIONS(2500),
    [anon_sym_ATselector] = ACTIONS(2500),
    [anon_sym_ATavailable] = ACTIONS(2500),
    [anon_sym___builtin_available] = ACTIONS(2498),
    [anon_sym_va_arg] = ACTIONS(2498),
    [anon_sym_ATencode] = ACTIONS(2500),
    [anon_sym_ATsynchronized] = ACTIONS(2500),
    [anon_sym_BOOL] = ACTIONS(2498),
    [anon_sym_IMP] = ACTIONS(2498),
    [anon_sym_SEL] = ACTIONS(2498),
    [anon_sym_Class] = ACTIONS(2498),
    [anon_sym_id] = ACTIONS(2498),
  },
  [STATE(1136)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1137)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1138)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1139)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1140)] = {
    [sym_identifier] = ACTIONS(2306),
    [anon_sym_LPAREN2] = ACTIONS(2308),
    [anon_sym_BANG] = ACTIONS(2308),
    [anon_sym_TILDE] = ACTIONS(2308),
    [anon_sym_DASH] = ACTIONS(2306),
    [anon_sym_PLUS] = ACTIONS(2306),
    [anon_sym_STAR] = ACTIONS(2308),
    [anon_sym_CARET] = ACTIONS(2308),
    [anon_sym_AMP] = ACTIONS(2308),
    [anon_sym_SEMI] = ACTIONS(2308),
    [anon_sym___extension__] = ACTIONS(2306),
    [anon_sym_typedef] = ACTIONS(2306),
    [anon_sym_extern] = ACTIONS(2306),
    [anon_sym___attribute__] = ACTIONS(2306),
    [anon_sym___attribute] = ACTIONS(2306),
    [anon_sym_noreturn] = ACTIONS(2306),
    [anon_sym_LBRACK] = ACTIONS(2308),
    [anon_sym___declspec] = ACTIONS(2306),
    [anon_sym_LBRACE] = ACTIONS(2308),
    [anon_sym_signed] = ACTIONS(2306),
    [anon_sym_unsigned] = ACTIONS(2306),
    [anon_sym_long] = ACTIONS(2306),
    [anon_sym_short] = ACTIONS(2306),
    [anon_sym_static] = ACTIONS(2306),
    [anon_sym_ATautoreleasepool] = ACTIONS(2308),
    [anon_sym_auto] = ACTIONS(2306),
    [anon_sym_register] = ACTIONS(2306),
    [anon_sym_inline] = ACTIONS(2306),
    [anon_sym___inline] = ACTIONS(2306),
    [anon_sym___inline__] = ACTIONS(2306),
    [anon_sym___forceinline] = ACTIONS(2306),
    [anon_sym_thread_local] = ACTIONS(2306),
    [anon_sym___thread] = ACTIONS(2306),
    [anon_sym_CG_EXTERN] = ACTIONS(2306),
    [anon_sym_CG_INLINE] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2306),
    [anon_sym_IBOutlet] = ACTIONS(2306),
    [anon_sym_IBInspectable] = ACTIONS(2306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2306),
    [anon_sym_NS_INLINE] = ACTIONS(2306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2306),
    [anon_sym_const] = ACTIONS(2306),
    [anon_sym_constexpr] = ACTIONS(2306),
    [anon_sym_volatile] = ACTIONS(2306),
    [anon_sym_restrict] = ACTIONS(2306),
    [anon_sym___restrict__] = ACTIONS(2306),
    [anon_sym__Atomic] = ACTIONS(2306),
    [anon_sym__Noreturn] = ACTIONS(2306),
    [anon_sym__Nonnull] = ACTIONS(2306),
    [anon_sym_nullable] = ACTIONS(2306),
    [anon_sym__Complex] = ACTIONS(2306),
    [anon_sym__Nullable] = ACTIONS(2306),
    [anon_sym__Nullable_result] = ACTIONS(2306),
    [anon_sym__Null_unspecified] = ACTIONS(2306),
    [anon_sym___autoreleasing] = ACTIONS(2306),
    [anon_sym___block] = ACTIONS(2306),
    [anon_sym___bridge] = ACTIONS(2306),
    [anon_sym___bridge_retained] = ACTIONS(2306),
    [anon_sym___bridge_transfer] = ACTIONS(2306),
    [anon_sym___complex] = ACTIONS(2306),
    [anon_sym___const] = ACTIONS(2306),
    [anon_sym___imag] = ACTIONS(2306),
    [anon_sym___kindof] = ACTIONS(2306),
    [anon_sym___nonnull] = ACTIONS(2306),
    [anon_sym___nullable] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2306),
    [anon_sym___real] = ACTIONS(2306),
    [anon_sym___strong] = ACTIONS(2306),
    [anon_sym___unsafe_unretained] = ACTIONS(2306),
    [anon_sym___unused] = ACTIONS(2306),
    [anon_sym___weak] = ACTIONS(2306),
    [anon_sym_alignas] = ACTIONS(2306),
    [anon_sym__Alignas] = ACTIONS(2306),
    [sym_primitive_type] = ACTIONS(2306),
    [anon_sym_enum] = ACTIONS(2306),
    [anon_sym_struct] = ACTIONS(2306),
    [anon_sym_union] = ACTIONS(2306),
    [anon_sym_if] = ACTIONS(2306),
    [anon_sym_else] = ACTIONS(2306),
    [anon_sym_switch] = ACTIONS(2306),
    [anon_sym_while] = ACTIONS(2306),
    [anon_sym_do] = ACTIONS(2306),
    [anon_sym_for] = ACTIONS(2306),
    [anon_sym_in] = ACTIONS(2306),
    [anon_sym_return] = ACTIONS(2306),
    [anon_sym_break] = ACTIONS(2306),
    [anon_sym_continue] = ACTIONS(2306),
    [anon_sym_goto] = ACTIONS(2306),
    [anon_sym_DASH_DASH] = ACTIONS(2308),
    [anon_sym_PLUS_PLUS] = ACTIONS(2308),
    [anon_sym_sizeof] = ACTIONS(2306),
    [anon_sym___alignof__] = ACTIONS(2306),
    [anon_sym___alignof] = ACTIONS(2306),
    [anon_sym__alignof] = ACTIONS(2306),
    [anon_sym_alignof] = ACTIONS(2306),
    [anon_sym__Alignof] = ACTIONS(2306),
    [anon_sym_offsetof] = ACTIONS(2306),
    [anon_sym__Generic] = ACTIONS(2306),
    [anon_sym_asm] = ACTIONS(2306),
    [anon_sym___asm__] = ACTIONS(2306),
    [anon_sym___asm] = ACTIONS(2306),
    [sym_number_literal] = ACTIONS(2308),
    [anon_sym_L_SQUOTE] = ACTIONS(2308),
    [anon_sym_u_SQUOTE] = ACTIONS(2308),
    [anon_sym_U_SQUOTE] = ACTIONS(2308),
    [anon_sym_u8_SQUOTE] = ACTIONS(2308),
    [anon_sym_SQUOTE] = ACTIONS(2308),
    [anon_sym_AT] = ACTIONS(2306),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [anon_sym_L_DQUOTE] = ACTIONS(2308),
    [anon_sym_u_DQUOTE] = ACTIONS(2308),
    [anon_sym_U_DQUOTE] = ACTIONS(2308),
    [anon_sym_u8_DQUOTE] = ACTIONS(2308),
    [sym_true] = ACTIONS(2306),
    [sym_false] = ACTIONS(2306),
    [anon_sym_NULL] = ACTIONS(2306),
    [anon_sym_nullptr] = ACTIONS(2306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2306),
    [anon_sym___typeof] = ACTIONS(2306),
    [anon_sym_typeof] = ACTIONS(2306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_API_AVAILABLE] = ACTIONS(2306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_API_DEPRECATED] = ACTIONS(2306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2306),
    [anon_sym___deprecated_msg] = ACTIONS(2306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2306),
    [anon_sym_ATtry] = ACTIONS(2308),
    [anon_sym___try] = ACTIONS(2306),
    [anon_sym_ATthrow] = ACTIONS(2308),
    [anon_sym_ATselector] = ACTIONS(2308),
    [anon_sym_ATavailable] = ACTIONS(2308),
    [anon_sym___builtin_available] = ACTIONS(2306),
    [anon_sym_va_arg] = ACTIONS(2306),
    [anon_sym_ATencode] = ACTIONS(2308),
    [anon_sym_ATsynchronized] = ACTIONS(2308),
    [anon_sym_BOOL] = ACTIONS(2306),
    [anon_sym_IMP] = ACTIONS(2306),
    [anon_sym_SEL] = ACTIONS(2306),
    [anon_sym_Class] = ACTIONS(2306),
    [anon_sym_id] = ACTIONS(2306),
  },
  [STATE(1141)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1142)] = {
    [sym_identifier] = ACTIONS(2502),
    [anon_sym_LPAREN2] = ACTIONS(2504),
    [anon_sym_BANG] = ACTIONS(2504),
    [anon_sym_TILDE] = ACTIONS(2504),
    [anon_sym_DASH] = ACTIONS(2502),
    [anon_sym_PLUS] = ACTIONS(2502),
    [anon_sym_STAR] = ACTIONS(2504),
    [anon_sym_CARET] = ACTIONS(2504),
    [anon_sym_AMP] = ACTIONS(2504),
    [anon_sym_SEMI] = ACTIONS(2504),
    [anon_sym___extension__] = ACTIONS(2502),
    [anon_sym_typedef] = ACTIONS(2502),
    [anon_sym_extern] = ACTIONS(2502),
    [anon_sym___attribute__] = ACTIONS(2502),
    [anon_sym___attribute] = ACTIONS(2502),
    [anon_sym_noreturn] = ACTIONS(2502),
    [anon_sym_LBRACK] = ACTIONS(2504),
    [anon_sym___declspec] = ACTIONS(2502),
    [anon_sym_LBRACE] = ACTIONS(2504),
    [anon_sym_signed] = ACTIONS(2502),
    [anon_sym_unsigned] = ACTIONS(2502),
    [anon_sym_long] = ACTIONS(2502),
    [anon_sym_short] = ACTIONS(2502),
    [anon_sym_static] = ACTIONS(2502),
    [anon_sym_ATautoreleasepool] = ACTIONS(2504),
    [anon_sym_auto] = ACTIONS(2502),
    [anon_sym_register] = ACTIONS(2502),
    [anon_sym_inline] = ACTIONS(2502),
    [anon_sym___inline] = ACTIONS(2502),
    [anon_sym___inline__] = ACTIONS(2502),
    [anon_sym___forceinline] = ACTIONS(2502),
    [anon_sym_thread_local] = ACTIONS(2502),
    [anon_sym___thread] = ACTIONS(2502),
    [anon_sym_CG_EXTERN] = ACTIONS(2502),
    [anon_sym_CG_INLINE] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2502),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2502),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2502),
    [anon_sym_IBOutlet] = ACTIONS(2502),
    [anon_sym_IBInspectable] = ACTIONS(2502),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2502),
    [anon_sym_NS_INLINE] = ACTIONS(2502),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2502),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2502),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2502),
    [anon_sym_const] = ACTIONS(2502),
    [anon_sym_constexpr] = ACTIONS(2502),
    [anon_sym_volatile] = ACTIONS(2502),
    [anon_sym_restrict] = ACTIONS(2502),
    [anon_sym___restrict__] = ACTIONS(2502),
    [anon_sym__Atomic] = ACTIONS(2502),
    [anon_sym__Noreturn] = ACTIONS(2502),
    [anon_sym__Nonnull] = ACTIONS(2502),
    [anon_sym_nullable] = ACTIONS(2502),
    [anon_sym__Complex] = ACTIONS(2502),
    [anon_sym__Nullable] = ACTIONS(2502),
    [anon_sym__Nullable_result] = ACTIONS(2502),
    [anon_sym__Null_unspecified] = ACTIONS(2502),
    [anon_sym___autoreleasing] = ACTIONS(2502),
    [anon_sym___block] = ACTIONS(2502),
    [anon_sym___bridge] = ACTIONS(2502),
    [anon_sym___bridge_retained] = ACTIONS(2502),
    [anon_sym___bridge_transfer] = ACTIONS(2502),
    [anon_sym___complex] = ACTIONS(2502),
    [anon_sym___const] = ACTIONS(2502),
    [anon_sym___imag] = ACTIONS(2502),
    [anon_sym___kindof] = ACTIONS(2502),
    [anon_sym___nonnull] = ACTIONS(2502),
    [anon_sym___nullable] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2502),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2502),
    [anon_sym___real] = ACTIONS(2502),
    [anon_sym___strong] = ACTIONS(2502),
    [anon_sym___unsafe_unretained] = ACTIONS(2502),
    [anon_sym___unused] = ACTIONS(2502),
    [anon_sym___weak] = ACTIONS(2502),
    [anon_sym_alignas] = ACTIONS(2502),
    [anon_sym__Alignas] = ACTIONS(2502),
    [sym_primitive_type] = ACTIONS(2502),
    [anon_sym_enum] = ACTIONS(2502),
    [anon_sym_struct] = ACTIONS(2502),
    [anon_sym_union] = ACTIONS(2502),
    [anon_sym_if] = ACTIONS(2502),
    [anon_sym_else] = ACTIONS(2502),
    [anon_sym_switch] = ACTIONS(2502),
    [anon_sym_while] = ACTIONS(2502),
    [anon_sym_do] = ACTIONS(2502),
    [anon_sym_for] = ACTIONS(2502),
    [anon_sym_in] = ACTIONS(2502),
    [anon_sym_return] = ACTIONS(2502),
    [anon_sym_break] = ACTIONS(2502),
    [anon_sym_continue] = ACTIONS(2502),
    [anon_sym_goto] = ACTIONS(2502),
    [anon_sym_DASH_DASH] = ACTIONS(2504),
    [anon_sym_PLUS_PLUS] = ACTIONS(2504),
    [anon_sym_sizeof] = ACTIONS(2502),
    [anon_sym___alignof__] = ACTIONS(2502),
    [anon_sym___alignof] = ACTIONS(2502),
    [anon_sym__alignof] = ACTIONS(2502),
    [anon_sym_alignof] = ACTIONS(2502),
    [anon_sym__Alignof] = ACTIONS(2502),
    [anon_sym_offsetof] = ACTIONS(2502),
    [anon_sym__Generic] = ACTIONS(2502),
    [anon_sym_asm] = ACTIONS(2502),
    [anon_sym___asm__] = ACTIONS(2502),
    [anon_sym___asm] = ACTIONS(2502),
    [sym_number_literal] = ACTIONS(2504),
    [anon_sym_L_SQUOTE] = ACTIONS(2504),
    [anon_sym_u_SQUOTE] = ACTIONS(2504),
    [anon_sym_U_SQUOTE] = ACTIONS(2504),
    [anon_sym_u8_SQUOTE] = ACTIONS(2504),
    [anon_sym_SQUOTE] = ACTIONS(2504),
    [anon_sym_AT] = ACTIONS(2502),
    [anon_sym_DQUOTE] = ACTIONS(2504),
    [anon_sym_L_DQUOTE] = ACTIONS(2504),
    [anon_sym_u_DQUOTE] = ACTIONS(2504),
    [anon_sym_U_DQUOTE] = ACTIONS(2504),
    [anon_sym_u8_DQUOTE] = ACTIONS(2504),
    [sym_true] = ACTIONS(2502),
    [sym_false] = ACTIONS(2502),
    [anon_sym_NULL] = ACTIONS(2502),
    [anon_sym_nullptr] = ACTIONS(2502),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2502),
    [anon_sym___typeof] = ACTIONS(2502),
    [anon_sym_typeof] = ACTIONS(2502),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2502),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2502),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2502),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2502),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2502),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_API_AVAILABLE] = ACTIONS(2502),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_API_DEPRECATED] = ACTIONS(2502),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2502),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2502),
    [anon_sym___deprecated_msg] = ACTIONS(2502),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2502),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2502),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2502),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2502),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2502),
    [anon_sym_ATtry] = ACTIONS(2504),
    [anon_sym___try] = ACTIONS(2502),
    [anon_sym_ATthrow] = ACTIONS(2504),
    [anon_sym_ATselector] = ACTIONS(2504),
    [anon_sym_ATavailable] = ACTIONS(2504),
    [anon_sym___builtin_available] = ACTIONS(2502),
    [anon_sym_va_arg] = ACTIONS(2502),
    [anon_sym_ATencode] = ACTIONS(2504),
    [anon_sym_ATsynchronized] = ACTIONS(2504),
    [anon_sym_BOOL] = ACTIONS(2502),
    [anon_sym_IMP] = ACTIONS(2502),
    [anon_sym_SEL] = ACTIONS(2502),
    [anon_sym_Class] = ACTIONS(2502),
    [anon_sym_id] = ACTIONS(2502),
  },
  [STATE(1143)] = {
    [sym_identifier] = ACTIONS(2962),
    [anon_sym_LPAREN2] = ACTIONS(2964),
    [anon_sym_BANG] = ACTIONS(2964),
    [anon_sym_TILDE] = ACTIONS(2964),
    [anon_sym_DASH] = ACTIONS(2962),
    [anon_sym_PLUS] = ACTIONS(2962),
    [anon_sym_STAR] = ACTIONS(2964),
    [anon_sym_CARET] = ACTIONS(2964),
    [anon_sym_AMP] = ACTIONS(2964),
    [anon_sym_SEMI] = ACTIONS(2964),
    [anon_sym___extension__] = ACTIONS(2962),
    [anon_sym_extern] = ACTIONS(2962),
    [anon_sym___attribute__] = ACTIONS(2962),
    [anon_sym___attribute] = ACTIONS(2962),
    [anon_sym_noreturn] = ACTIONS(2962),
    [anon_sym_LBRACK] = ACTIONS(2964),
    [anon_sym___declspec] = ACTIONS(2962),
    [anon_sym_LBRACE] = ACTIONS(2964),
    [anon_sym_signed] = ACTIONS(2962),
    [anon_sym_unsigned] = ACTIONS(2962),
    [anon_sym_long] = ACTIONS(2962),
    [anon_sym_short] = ACTIONS(2962),
    [anon_sym_static] = ACTIONS(2962),
    [anon_sym_ATautoreleasepool] = ACTIONS(2964),
    [anon_sym_auto] = ACTIONS(2962),
    [anon_sym_register] = ACTIONS(2962),
    [anon_sym_inline] = ACTIONS(2962),
    [anon_sym___inline] = ACTIONS(2962),
    [anon_sym___inline__] = ACTIONS(2962),
    [anon_sym___forceinline] = ACTIONS(2962),
    [anon_sym_thread_local] = ACTIONS(2962),
    [anon_sym___thread] = ACTIONS(2962),
    [anon_sym_CG_EXTERN] = ACTIONS(2962),
    [anon_sym_CG_INLINE] = ACTIONS(2962),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2962),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2962),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2962),
    [anon_sym_IBOutlet] = ACTIONS(2962),
    [anon_sym_IBInspectable] = ACTIONS(2962),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2962),
    [anon_sym_NS_INLINE] = ACTIONS(2962),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2962),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2962),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2962),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2962),
    [anon_sym_const] = ACTIONS(2962),
    [anon_sym_constexpr] = ACTIONS(2962),
    [anon_sym_volatile] = ACTIONS(2962),
    [anon_sym_restrict] = ACTIONS(2962),
    [anon_sym___restrict__] = ACTIONS(2962),
    [anon_sym__Atomic] = ACTIONS(2962),
    [anon_sym__Noreturn] = ACTIONS(2962),
    [anon_sym__Nonnull] = ACTIONS(2962),
    [anon_sym_nullable] = ACTIONS(2962),
    [anon_sym__Complex] = ACTIONS(2962),
    [anon_sym__Nullable] = ACTIONS(2962),
    [anon_sym__Nullable_result] = ACTIONS(2962),
    [anon_sym__Null_unspecified] = ACTIONS(2962),
    [anon_sym___autoreleasing] = ACTIONS(2962),
    [anon_sym___block] = ACTIONS(2962),
    [anon_sym___bridge] = ACTIONS(2962),
    [anon_sym___bridge_retained] = ACTIONS(2962),
    [anon_sym___bridge_transfer] = ACTIONS(2962),
    [anon_sym___complex] = ACTIONS(2962),
    [anon_sym___const] = ACTIONS(2962),
    [anon_sym___imag] = ACTIONS(2962),
    [anon_sym___kindof] = ACTIONS(2962),
    [anon_sym___nonnull] = ACTIONS(2962),
    [anon_sym___nullable] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2962),
    [anon_sym___real] = ACTIONS(2962),
    [anon_sym___strong] = ACTIONS(2962),
    [anon_sym___unsafe_unretained] = ACTIONS(2962),
    [anon_sym___unused] = ACTIONS(2962),
    [anon_sym___weak] = ACTIONS(2962),
    [anon_sym_alignas] = ACTIONS(2962),
    [anon_sym__Alignas] = ACTIONS(2962),
    [sym_primitive_type] = ACTIONS(2962),
    [anon_sym_enum] = ACTIONS(2962),
    [anon_sym_struct] = ACTIONS(2962),
    [anon_sym_union] = ACTIONS(2962),
    [anon_sym_if] = ACTIONS(2962),
    [anon_sym_switch] = ACTIONS(2962),
    [anon_sym_case] = ACTIONS(2962),
    [anon_sym_default] = ACTIONS(2962),
    [anon_sym_while] = ACTIONS(2962),
    [anon_sym_do] = ACTIONS(2962),
    [anon_sym_for] = ACTIONS(2962),
    [anon_sym_in] = ACTIONS(2962),
    [anon_sym_return] = ACTIONS(2962),
    [anon_sym_break] = ACTIONS(2962),
    [anon_sym_continue] = ACTIONS(2962),
    [anon_sym_goto] = ACTIONS(2962),
    [anon_sym_DASH_DASH] = ACTIONS(2964),
    [anon_sym_PLUS_PLUS] = ACTIONS(2964),
    [anon_sym_sizeof] = ACTIONS(2962),
    [anon_sym___alignof__] = ACTIONS(2962),
    [anon_sym___alignof] = ACTIONS(2962),
    [anon_sym__alignof] = ACTIONS(2962),
    [anon_sym_alignof] = ACTIONS(2962),
    [anon_sym__Alignof] = ACTIONS(2962),
    [anon_sym_offsetof] = ACTIONS(2962),
    [anon_sym__Generic] = ACTIONS(2962),
    [anon_sym_asm] = ACTIONS(2962),
    [anon_sym___asm__] = ACTIONS(2962),
    [anon_sym___asm] = ACTIONS(2962),
    [sym_number_literal] = ACTIONS(2964),
    [anon_sym_L_SQUOTE] = ACTIONS(2964),
    [anon_sym_u_SQUOTE] = ACTIONS(2964),
    [anon_sym_U_SQUOTE] = ACTIONS(2964),
    [anon_sym_u8_SQUOTE] = ACTIONS(2964),
    [anon_sym_SQUOTE] = ACTIONS(2964),
    [anon_sym_AT] = ACTIONS(2962),
    [anon_sym_DQUOTE] = ACTIONS(2964),
    [anon_sym_L_DQUOTE] = ACTIONS(2964),
    [anon_sym_u_DQUOTE] = ACTIONS(2964),
    [anon_sym_U_DQUOTE] = ACTIONS(2964),
    [anon_sym_u8_DQUOTE] = ACTIONS(2964),
    [sym_true] = ACTIONS(2962),
    [sym_false] = ACTIONS(2962),
    [anon_sym_NULL] = ACTIONS(2962),
    [anon_sym_nullptr] = ACTIONS(2962),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2962),
    [anon_sym___typeof] = ACTIONS(2962),
    [anon_sym_typeof] = ACTIONS(2962),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2962),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2962),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2962),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2962),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2962),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2962),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2962),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2962),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2962),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_API_AVAILABLE] = ACTIONS(2962),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_API_DEPRECATED] = ACTIONS(2962),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2962),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2962),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2962),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2962),
    [anon_sym___deprecated_msg] = ACTIONS(2962),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2962),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2962),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2962),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2962),
    [anon_sym_ATtry] = ACTIONS(2964),
    [anon_sym___try] = ACTIONS(2962),
    [anon_sym_ATthrow] = ACTIONS(2964),
    [anon_sym_ATselector] = ACTIONS(2964),
    [anon_sym_ATavailable] = ACTIONS(2964),
    [anon_sym___builtin_available] = ACTIONS(2962),
    [anon_sym_va_arg] = ACTIONS(2962),
    [anon_sym_ATencode] = ACTIONS(2964),
    [anon_sym_ATsynchronized] = ACTIONS(2964),
    [anon_sym_BOOL] = ACTIONS(2962),
    [anon_sym_IMP] = ACTIONS(2962),
    [anon_sym_SEL] = ACTIONS(2962),
    [anon_sym_Class] = ACTIONS(2962),
    [anon_sym_id] = ACTIONS(2962),
  },
  [STATE(1144)] = {
    [sym_identifier] = ACTIONS(2506),
    [anon_sym_LPAREN2] = ACTIONS(2508),
    [anon_sym_BANG] = ACTIONS(2508),
    [anon_sym_TILDE] = ACTIONS(2508),
    [anon_sym_DASH] = ACTIONS(2506),
    [anon_sym_PLUS] = ACTIONS(2506),
    [anon_sym_STAR] = ACTIONS(2508),
    [anon_sym_CARET] = ACTIONS(2508),
    [anon_sym_AMP] = ACTIONS(2508),
    [anon_sym_SEMI] = ACTIONS(2508),
    [anon_sym___extension__] = ACTIONS(2506),
    [anon_sym_typedef] = ACTIONS(2506),
    [anon_sym_extern] = ACTIONS(2506),
    [anon_sym___attribute__] = ACTIONS(2506),
    [anon_sym___attribute] = ACTIONS(2506),
    [anon_sym_noreturn] = ACTIONS(2506),
    [anon_sym_LBRACK] = ACTIONS(2508),
    [anon_sym___declspec] = ACTIONS(2506),
    [anon_sym_LBRACE] = ACTIONS(2508),
    [anon_sym_signed] = ACTIONS(2506),
    [anon_sym_unsigned] = ACTIONS(2506),
    [anon_sym_long] = ACTIONS(2506),
    [anon_sym_short] = ACTIONS(2506),
    [anon_sym_static] = ACTIONS(2506),
    [anon_sym_ATautoreleasepool] = ACTIONS(2508),
    [anon_sym_auto] = ACTIONS(2506),
    [anon_sym_register] = ACTIONS(2506),
    [anon_sym_inline] = ACTIONS(2506),
    [anon_sym___inline] = ACTIONS(2506),
    [anon_sym___inline__] = ACTIONS(2506),
    [anon_sym___forceinline] = ACTIONS(2506),
    [anon_sym_thread_local] = ACTIONS(2506),
    [anon_sym___thread] = ACTIONS(2506),
    [anon_sym_CG_EXTERN] = ACTIONS(2506),
    [anon_sym_CG_INLINE] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2506),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2506),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2506),
    [anon_sym_IBOutlet] = ACTIONS(2506),
    [anon_sym_IBInspectable] = ACTIONS(2506),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2506),
    [anon_sym_NS_INLINE] = ACTIONS(2506),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2506),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2506),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2506),
    [anon_sym_const] = ACTIONS(2506),
    [anon_sym_constexpr] = ACTIONS(2506),
    [anon_sym_volatile] = ACTIONS(2506),
    [anon_sym_restrict] = ACTIONS(2506),
    [anon_sym___restrict__] = ACTIONS(2506),
    [anon_sym__Atomic] = ACTIONS(2506),
    [anon_sym__Noreturn] = ACTIONS(2506),
    [anon_sym__Nonnull] = ACTIONS(2506),
    [anon_sym_nullable] = ACTIONS(2506),
    [anon_sym__Complex] = ACTIONS(2506),
    [anon_sym__Nullable] = ACTIONS(2506),
    [anon_sym__Nullable_result] = ACTIONS(2506),
    [anon_sym__Null_unspecified] = ACTIONS(2506),
    [anon_sym___autoreleasing] = ACTIONS(2506),
    [anon_sym___block] = ACTIONS(2506),
    [anon_sym___bridge] = ACTIONS(2506),
    [anon_sym___bridge_retained] = ACTIONS(2506),
    [anon_sym___bridge_transfer] = ACTIONS(2506),
    [anon_sym___complex] = ACTIONS(2506),
    [anon_sym___const] = ACTIONS(2506),
    [anon_sym___imag] = ACTIONS(2506),
    [anon_sym___kindof] = ACTIONS(2506),
    [anon_sym___nonnull] = ACTIONS(2506),
    [anon_sym___nullable] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2506),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2506),
    [anon_sym___real] = ACTIONS(2506),
    [anon_sym___strong] = ACTIONS(2506),
    [anon_sym___unsafe_unretained] = ACTIONS(2506),
    [anon_sym___unused] = ACTIONS(2506),
    [anon_sym___weak] = ACTIONS(2506),
    [anon_sym_alignas] = ACTIONS(2506),
    [anon_sym__Alignas] = ACTIONS(2506),
    [sym_primitive_type] = ACTIONS(2506),
    [anon_sym_enum] = ACTIONS(2506),
    [anon_sym_struct] = ACTIONS(2506),
    [anon_sym_union] = ACTIONS(2506),
    [anon_sym_if] = ACTIONS(2506),
    [anon_sym_else] = ACTIONS(2506),
    [anon_sym_switch] = ACTIONS(2506),
    [anon_sym_while] = ACTIONS(2506),
    [anon_sym_do] = ACTIONS(2506),
    [anon_sym_for] = ACTIONS(2506),
    [anon_sym_in] = ACTIONS(2506),
    [anon_sym_return] = ACTIONS(2506),
    [anon_sym_break] = ACTIONS(2506),
    [anon_sym_continue] = ACTIONS(2506),
    [anon_sym_goto] = ACTIONS(2506),
    [anon_sym_DASH_DASH] = ACTIONS(2508),
    [anon_sym_PLUS_PLUS] = ACTIONS(2508),
    [anon_sym_sizeof] = ACTIONS(2506),
    [anon_sym___alignof__] = ACTIONS(2506),
    [anon_sym___alignof] = ACTIONS(2506),
    [anon_sym__alignof] = ACTIONS(2506),
    [anon_sym_alignof] = ACTIONS(2506),
    [anon_sym__Alignof] = ACTIONS(2506),
    [anon_sym_offsetof] = ACTIONS(2506),
    [anon_sym__Generic] = ACTIONS(2506),
    [anon_sym_asm] = ACTIONS(2506),
    [anon_sym___asm__] = ACTIONS(2506),
    [anon_sym___asm] = ACTIONS(2506),
    [sym_number_literal] = ACTIONS(2508),
    [anon_sym_L_SQUOTE] = ACTIONS(2508),
    [anon_sym_u_SQUOTE] = ACTIONS(2508),
    [anon_sym_U_SQUOTE] = ACTIONS(2508),
    [anon_sym_u8_SQUOTE] = ACTIONS(2508),
    [anon_sym_SQUOTE] = ACTIONS(2508),
    [anon_sym_AT] = ACTIONS(2506),
    [anon_sym_DQUOTE] = ACTIONS(2508),
    [anon_sym_L_DQUOTE] = ACTIONS(2508),
    [anon_sym_u_DQUOTE] = ACTIONS(2508),
    [anon_sym_U_DQUOTE] = ACTIONS(2508),
    [anon_sym_u8_DQUOTE] = ACTIONS(2508),
    [sym_true] = ACTIONS(2506),
    [sym_false] = ACTIONS(2506),
    [anon_sym_NULL] = ACTIONS(2506),
    [anon_sym_nullptr] = ACTIONS(2506),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2506),
    [anon_sym___typeof] = ACTIONS(2506),
    [anon_sym_typeof] = ACTIONS(2506),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2506),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2506),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2506),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2506),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2506),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_API_AVAILABLE] = ACTIONS(2506),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_API_DEPRECATED] = ACTIONS(2506),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2506),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2506),
    [anon_sym___deprecated_msg] = ACTIONS(2506),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2506),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2506),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2506),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2506),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2506),
    [anon_sym_ATtry] = ACTIONS(2508),
    [anon_sym___try] = ACTIONS(2506),
    [anon_sym_ATthrow] = ACTIONS(2508),
    [anon_sym_ATselector] = ACTIONS(2508),
    [anon_sym_ATavailable] = ACTIONS(2508),
    [anon_sym___builtin_available] = ACTIONS(2506),
    [anon_sym_va_arg] = ACTIONS(2506),
    [anon_sym_ATencode] = ACTIONS(2508),
    [anon_sym_ATsynchronized] = ACTIONS(2508),
    [anon_sym_BOOL] = ACTIONS(2506),
    [anon_sym_IMP] = ACTIONS(2506),
    [anon_sym_SEL] = ACTIONS(2506),
    [anon_sym_Class] = ACTIONS(2506),
    [anon_sym_id] = ACTIONS(2506),
  },
  [STATE(1145)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1146)] = {
    [sym_identifier] = ACTIONS(2958),
    [anon_sym_LPAREN2] = ACTIONS(2960),
    [anon_sym_BANG] = ACTIONS(2960),
    [anon_sym_TILDE] = ACTIONS(2960),
    [anon_sym_DASH] = ACTIONS(2958),
    [anon_sym_PLUS] = ACTIONS(2958),
    [anon_sym_STAR] = ACTIONS(2960),
    [anon_sym_CARET] = ACTIONS(2960),
    [anon_sym_AMP] = ACTIONS(2960),
    [anon_sym_SEMI] = ACTIONS(2960),
    [anon_sym___extension__] = ACTIONS(2958),
    [anon_sym_extern] = ACTIONS(2958),
    [anon_sym___attribute__] = ACTIONS(2958),
    [anon_sym___attribute] = ACTIONS(2958),
    [anon_sym_noreturn] = ACTIONS(2958),
    [anon_sym_LBRACK] = ACTIONS(2960),
    [anon_sym___declspec] = ACTIONS(2958),
    [anon_sym_LBRACE] = ACTIONS(2960),
    [anon_sym_signed] = ACTIONS(2958),
    [anon_sym_unsigned] = ACTIONS(2958),
    [anon_sym_long] = ACTIONS(2958),
    [anon_sym_short] = ACTIONS(2958),
    [anon_sym_static] = ACTIONS(2958),
    [anon_sym_ATautoreleasepool] = ACTIONS(2960),
    [anon_sym_auto] = ACTIONS(2958),
    [anon_sym_register] = ACTIONS(2958),
    [anon_sym_inline] = ACTIONS(2958),
    [anon_sym___inline] = ACTIONS(2958),
    [anon_sym___inline__] = ACTIONS(2958),
    [anon_sym___forceinline] = ACTIONS(2958),
    [anon_sym_thread_local] = ACTIONS(2958),
    [anon_sym___thread] = ACTIONS(2958),
    [anon_sym_CG_EXTERN] = ACTIONS(2958),
    [anon_sym_CG_INLINE] = ACTIONS(2958),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2958),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2958),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2958),
    [anon_sym_IBOutlet] = ACTIONS(2958),
    [anon_sym_IBInspectable] = ACTIONS(2958),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2958),
    [anon_sym_NS_INLINE] = ACTIONS(2958),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2958),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2958),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2958),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2958),
    [anon_sym_const] = ACTIONS(2958),
    [anon_sym_constexpr] = ACTIONS(2958),
    [anon_sym_volatile] = ACTIONS(2958),
    [anon_sym_restrict] = ACTIONS(2958),
    [anon_sym___restrict__] = ACTIONS(2958),
    [anon_sym__Atomic] = ACTIONS(2958),
    [anon_sym__Noreturn] = ACTIONS(2958),
    [anon_sym__Nonnull] = ACTIONS(2958),
    [anon_sym_nullable] = ACTIONS(2958),
    [anon_sym__Complex] = ACTIONS(2958),
    [anon_sym__Nullable] = ACTIONS(2958),
    [anon_sym__Nullable_result] = ACTIONS(2958),
    [anon_sym__Null_unspecified] = ACTIONS(2958),
    [anon_sym___autoreleasing] = ACTIONS(2958),
    [anon_sym___block] = ACTIONS(2958),
    [anon_sym___bridge] = ACTIONS(2958),
    [anon_sym___bridge_retained] = ACTIONS(2958),
    [anon_sym___bridge_transfer] = ACTIONS(2958),
    [anon_sym___complex] = ACTIONS(2958),
    [anon_sym___const] = ACTIONS(2958),
    [anon_sym___imag] = ACTIONS(2958),
    [anon_sym___kindof] = ACTIONS(2958),
    [anon_sym___nonnull] = ACTIONS(2958),
    [anon_sym___nullable] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2958),
    [anon_sym___real] = ACTIONS(2958),
    [anon_sym___strong] = ACTIONS(2958),
    [anon_sym___unsafe_unretained] = ACTIONS(2958),
    [anon_sym___unused] = ACTIONS(2958),
    [anon_sym___weak] = ACTIONS(2958),
    [anon_sym_alignas] = ACTIONS(2958),
    [anon_sym__Alignas] = ACTIONS(2958),
    [sym_primitive_type] = ACTIONS(2958),
    [anon_sym_enum] = ACTIONS(2958),
    [anon_sym_struct] = ACTIONS(2958),
    [anon_sym_union] = ACTIONS(2958),
    [anon_sym_if] = ACTIONS(2958),
    [anon_sym_switch] = ACTIONS(2958),
    [anon_sym_case] = ACTIONS(2958),
    [anon_sym_default] = ACTIONS(2958),
    [anon_sym_while] = ACTIONS(2958),
    [anon_sym_do] = ACTIONS(2958),
    [anon_sym_for] = ACTIONS(2958),
    [anon_sym_in] = ACTIONS(2958),
    [anon_sym_return] = ACTIONS(2958),
    [anon_sym_break] = ACTIONS(2958),
    [anon_sym_continue] = ACTIONS(2958),
    [anon_sym_goto] = ACTIONS(2958),
    [anon_sym_DASH_DASH] = ACTIONS(2960),
    [anon_sym_PLUS_PLUS] = ACTIONS(2960),
    [anon_sym_sizeof] = ACTIONS(2958),
    [anon_sym___alignof__] = ACTIONS(2958),
    [anon_sym___alignof] = ACTIONS(2958),
    [anon_sym__alignof] = ACTIONS(2958),
    [anon_sym_alignof] = ACTIONS(2958),
    [anon_sym__Alignof] = ACTIONS(2958),
    [anon_sym_offsetof] = ACTIONS(2958),
    [anon_sym__Generic] = ACTIONS(2958),
    [anon_sym_asm] = ACTIONS(2958),
    [anon_sym___asm__] = ACTIONS(2958),
    [anon_sym___asm] = ACTIONS(2958),
    [sym_number_literal] = ACTIONS(2960),
    [anon_sym_L_SQUOTE] = ACTIONS(2960),
    [anon_sym_u_SQUOTE] = ACTIONS(2960),
    [anon_sym_U_SQUOTE] = ACTIONS(2960),
    [anon_sym_u8_SQUOTE] = ACTIONS(2960),
    [anon_sym_SQUOTE] = ACTIONS(2960),
    [anon_sym_AT] = ACTIONS(2958),
    [anon_sym_DQUOTE] = ACTIONS(2960),
    [anon_sym_L_DQUOTE] = ACTIONS(2960),
    [anon_sym_u_DQUOTE] = ACTIONS(2960),
    [anon_sym_U_DQUOTE] = ACTIONS(2960),
    [anon_sym_u8_DQUOTE] = ACTIONS(2960),
    [sym_true] = ACTIONS(2958),
    [sym_false] = ACTIONS(2958),
    [anon_sym_NULL] = ACTIONS(2958),
    [anon_sym_nullptr] = ACTIONS(2958),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2958),
    [anon_sym___typeof] = ACTIONS(2958),
    [anon_sym_typeof] = ACTIONS(2958),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2958),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2958),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2958),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2958),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2958),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2958),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2958),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2958),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2958),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_API_AVAILABLE] = ACTIONS(2958),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_API_DEPRECATED] = ACTIONS(2958),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2958),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2958),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2958),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2958),
    [anon_sym___deprecated_msg] = ACTIONS(2958),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2958),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2958),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2958),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2958),
    [anon_sym_ATtry] = ACTIONS(2960),
    [anon_sym___try] = ACTIONS(2958),
    [anon_sym_ATthrow] = ACTIONS(2960),
    [anon_sym_ATselector] = ACTIONS(2960),
    [anon_sym_ATavailable] = ACTIONS(2960),
    [anon_sym___builtin_available] = ACTIONS(2958),
    [anon_sym_va_arg] = ACTIONS(2958),
    [anon_sym_ATencode] = ACTIONS(2960),
    [anon_sym_ATsynchronized] = ACTIONS(2960),
    [anon_sym_BOOL] = ACTIONS(2958),
    [anon_sym_IMP] = ACTIONS(2958),
    [anon_sym_SEL] = ACTIONS(2958),
    [anon_sym_Class] = ACTIONS(2958),
    [anon_sym_id] = ACTIONS(2958),
  },
  [STATE(1147)] = {
    [sym_identifier] = ACTIONS(2514),
    [anon_sym_LPAREN2] = ACTIONS(2516),
    [anon_sym_BANG] = ACTIONS(2516),
    [anon_sym_TILDE] = ACTIONS(2516),
    [anon_sym_DASH] = ACTIONS(2514),
    [anon_sym_PLUS] = ACTIONS(2514),
    [anon_sym_STAR] = ACTIONS(2516),
    [anon_sym_CARET] = ACTIONS(2516),
    [anon_sym_AMP] = ACTIONS(2516),
    [anon_sym_SEMI] = ACTIONS(2516),
    [anon_sym___extension__] = ACTIONS(2514),
    [anon_sym_typedef] = ACTIONS(2514),
    [anon_sym_extern] = ACTIONS(2514),
    [anon_sym___attribute__] = ACTIONS(2514),
    [anon_sym___attribute] = ACTIONS(2514),
    [anon_sym_noreturn] = ACTIONS(2514),
    [anon_sym_LBRACK] = ACTIONS(2516),
    [anon_sym___declspec] = ACTIONS(2514),
    [anon_sym_LBRACE] = ACTIONS(2516),
    [anon_sym_signed] = ACTIONS(2514),
    [anon_sym_unsigned] = ACTIONS(2514),
    [anon_sym_long] = ACTIONS(2514),
    [anon_sym_short] = ACTIONS(2514),
    [anon_sym_static] = ACTIONS(2514),
    [anon_sym_ATautoreleasepool] = ACTIONS(2516),
    [anon_sym_auto] = ACTIONS(2514),
    [anon_sym_register] = ACTIONS(2514),
    [anon_sym_inline] = ACTIONS(2514),
    [anon_sym___inline] = ACTIONS(2514),
    [anon_sym___inline__] = ACTIONS(2514),
    [anon_sym___forceinline] = ACTIONS(2514),
    [anon_sym_thread_local] = ACTIONS(2514),
    [anon_sym___thread] = ACTIONS(2514),
    [anon_sym_CG_EXTERN] = ACTIONS(2514),
    [anon_sym_CG_INLINE] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2514),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2514),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2514),
    [anon_sym_IBOutlet] = ACTIONS(2514),
    [anon_sym_IBInspectable] = ACTIONS(2514),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2514),
    [anon_sym_NS_INLINE] = ACTIONS(2514),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2514),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2514),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2514),
    [anon_sym_const] = ACTIONS(2514),
    [anon_sym_constexpr] = ACTIONS(2514),
    [anon_sym_volatile] = ACTIONS(2514),
    [anon_sym_restrict] = ACTIONS(2514),
    [anon_sym___restrict__] = ACTIONS(2514),
    [anon_sym__Atomic] = ACTIONS(2514),
    [anon_sym__Noreturn] = ACTIONS(2514),
    [anon_sym__Nonnull] = ACTIONS(2514),
    [anon_sym_nullable] = ACTIONS(2514),
    [anon_sym__Complex] = ACTIONS(2514),
    [anon_sym__Nullable] = ACTIONS(2514),
    [anon_sym__Nullable_result] = ACTIONS(2514),
    [anon_sym__Null_unspecified] = ACTIONS(2514),
    [anon_sym___autoreleasing] = ACTIONS(2514),
    [anon_sym___block] = ACTIONS(2514),
    [anon_sym___bridge] = ACTIONS(2514),
    [anon_sym___bridge_retained] = ACTIONS(2514),
    [anon_sym___bridge_transfer] = ACTIONS(2514),
    [anon_sym___complex] = ACTIONS(2514),
    [anon_sym___const] = ACTIONS(2514),
    [anon_sym___imag] = ACTIONS(2514),
    [anon_sym___kindof] = ACTIONS(2514),
    [anon_sym___nonnull] = ACTIONS(2514),
    [anon_sym___nullable] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2514),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2514),
    [anon_sym___real] = ACTIONS(2514),
    [anon_sym___strong] = ACTIONS(2514),
    [anon_sym___unsafe_unretained] = ACTIONS(2514),
    [anon_sym___unused] = ACTIONS(2514),
    [anon_sym___weak] = ACTIONS(2514),
    [anon_sym_alignas] = ACTIONS(2514),
    [anon_sym__Alignas] = ACTIONS(2514),
    [sym_primitive_type] = ACTIONS(2514),
    [anon_sym_enum] = ACTIONS(2514),
    [anon_sym_struct] = ACTIONS(2514),
    [anon_sym_union] = ACTIONS(2514),
    [anon_sym_if] = ACTIONS(2514),
    [anon_sym_else] = ACTIONS(2514),
    [anon_sym_switch] = ACTIONS(2514),
    [anon_sym_while] = ACTIONS(2514),
    [anon_sym_do] = ACTIONS(2514),
    [anon_sym_for] = ACTIONS(2514),
    [anon_sym_in] = ACTIONS(2514),
    [anon_sym_return] = ACTIONS(2514),
    [anon_sym_break] = ACTIONS(2514),
    [anon_sym_continue] = ACTIONS(2514),
    [anon_sym_goto] = ACTIONS(2514),
    [anon_sym_DASH_DASH] = ACTIONS(2516),
    [anon_sym_PLUS_PLUS] = ACTIONS(2516),
    [anon_sym_sizeof] = ACTIONS(2514),
    [anon_sym___alignof__] = ACTIONS(2514),
    [anon_sym___alignof] = ACTIONS(2514),
    [anon_sym__alignof] = ACTIONS(2514),
    [anon_sym_alignof] = ACTIONS(2514),
    [anon_sym__Alignof] = ACTIONS(2514),
    [anon_sym_offsetof] = ACTIONS(2514),
    [anon_sym__Generic] = ACTIONS(2514),
    [anon_sym_asm] = ACTIONS(2514),
    [anon_sym___asm__] = ACTIONS(2514),
    [anon_sym___asm] = ACTIONS(2514),
    [sym_number_literal] = ACTIONS(2516),
    [anon_sym_L_SQUOTE] = ACTIONS(2516),
    [anon_sym_u_SQUOTE] = ACTIONS(2516),
    [anon_sym_U_SQUOTE] = ACTIONS(2516),
    [anon_sym_u8_SQUOTE] = ACTIONS(2516),
    [anon_sym_SQUOTE] = ACTIONS(2516),
    [anon_sym_AT] = ACTIONS(2514),
    [anon_sym_DQUOTE] = ACTIONS(2516),
    [anon_sym_L_DQUOTE] = ACTIONS(2516),
    [anon_sym_u_DQUOTE] = ACTIONS(2516),
    [anon_sym_U_DQUOTE] = ACTIONS(2516),
    [anon_sym_u8_DQUOTE] = ACTIONS(2516),
    [sym_true] = ACTIONS(2514),
    [sym_false] = ACTIONS(2514),
    [anon_sym_NULL] = ACTIONS(2514),
    [anon_sym_nullptr] = ACTIONS(2514),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2514),
    [anon_sym___typeof] = ACTIONS(2514),
    [anon_sym_typeof] = ACTIONS(2514),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2514),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2514),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2514),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2514),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2514),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_API_AVAILABLE] = ACTIONS(2514),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_API_DEPRECATED] = ACTIONS(2514),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2514),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2514),
    [anon_sym___deprecated_msg] = ACTIONS(2514),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2514),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2514),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2514),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2514),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2514),
    [anon_sym_ATtry] = ACTIONS(2516),
    [anon_sym___try] = ACTIONS(2514),
    [anon_sym_ATthrow] = ACTIONS(2516),
    [anon_sym_ATselector] = ACTIONS(2516),
    [anon_sym_ATavailable] = ACTIONS(2516),
    [anon_sym___builtin_available] = ACTIONS(2514),
    [anon_sym_va_arg] = ACTIONS(2514),
    [anon_sym_ATencode] = ACTIONS(2516),
    [anon_sym_ATsynchronized] = ACTIONS(2516),
    [anon_sym_BOOL] = ACTIONS(2514),
    [anon_sym_IMP] = ACTIONS(2514),
    [anon_sym_SEL] = ACTIONS(2514),
    [anon_sym_Class] = ACTIONS(2514),
    [anon_sym_id] = ACTIONS(2514),
  },
  [STATE(1148)] = {
    [sym_identifier] = ACTIONS(2518),
    [anon_sym_LPAREN2] = ACTIONS(2520),
    [anon_sym_BANG] = ACTIONS(2520),
    [anon_sym_TILDE] = ACTIONS(2520),
    [anon_sym_DASH] = ACTIONS(2518),
    [anon_sym_PLUS] = ACTIONS(2518),
    [anon_sym_STAR] = ACTIONS(2520),
    [anon_sym_CARET] = ACTIONS(2520),
    [anon_sym_AMP] = ACTIONS(2520),
    [anon_sym_SEMI] = ACTIONS(2520),
    [anon_sym___extension__] = ACTIONS(2518),
    [anon_sym_typedef] = ACTIONS(2518),
    [anon_sym_extern] = ACTIONS(2518),
    [anon_sym___attribute__] = ACTIONS(2518),
    [anon_sym___attribute] = ACTIONS(2518),
    [anon_sym_noreturn] = ACTIONS(2518),
    [anon_sym_LBRACK] = ACTIONS(2520),
    [anon_sym___declspec] = ACTIONS(2518),
    [anon_sym_LBRACE] = ACTIONS(2520),
    [anon_sym_signed] = ACTIONS(2518),
    [anon_sym_unsigned] = ACTIONS(2518),
    [anon_sym_long] = ACTIONS(2518),
    [anon_sym_short] = ACTIONS(2518),
    [anon_sym_static] = ACTIONS(2518),
    [anon_sym_ATautoreleasepool] = ACTIONS(2520),
    [anon_sym_auto] = ACTIONS(2518),
    [anon_sym_register] = ACTIONS(2518),
    [anon_sym_inline] = ACTIONS(2518),
    [anon_sym___inline] = ACTIONS(2518),
    [anon_sym___inline__] = ACTIONS(2518),
    [anon_sym___forceinline] = ACTIONS(2518),
    [anon_sym_thread_local] = ACTIONS(2518),
    [anon_sym___thread] = ACTIONS(2518),
    [anon_sym_CG_EXTERN] = ACTIONS(2518),
    [anon_sym_CG_INLINE] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2518),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2518),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2518),
    [anon_sym_IBOutlet] = ACTIONS(2518),
    [anon_sym_IBInspectable] = ACTIONS(2518),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2518),
    [anon_sym_NS_INLINE] = ACTIONS(2518),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2518),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2518),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2518),
    [anon_sym_const] = ACTIONS(2518),
    [anon_sym_constexpr] = ACTIONS(2518),
    [anon_sym_volatile] = ACTIONS(2518),
    [anon_sym_restrict] = ACTIONS(2518),
    [anon_sym___restrict__] = ACTIONS(2518),
    [anon_sym__Atomic] = ACTIONS(2518),
    [anon_sym__Noreturn] = ACTIONS(2518),
    [anon_sym__Nonnull] = ACTIONS(2518),
    [anon_sym_nullable] = ACTIONS(2518),
    [anon_sym__Complex] = ACTIONS(2518),
    [anon_sym__Nullable] = ACTIONS(2518),
    [anon_sym__Nullable_result] = ACTIONS(2518),
    [anon_sym__Null_unspecified] = ACTIONS(2518),
    [anon_sym___autoreleasing] = ACTIONS(2518),
    [anon_sym___block] = ACTIONS(2518),
    [anon_sym___bridge] = ACTIONS(2518),
    [anon_sym___bridge_retained] = ACTIONS(2518),
    [anon_sym___bridge_transfer] = ACTIONS(2518),
    [anon_sym___complex] = ACTIONS(2518),
    [anon_sym___const] = ACTIONS(2518),
    [anon_sym___imag] = ACTIONS(2518),
    [anon_sym___kindof] = ACTIONS(2518),
    [anon_sym___nonnull] = ACTIONS(2518),
    [anon_sym___nullable] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2518),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2518),
    [anon_sym___real] = ACTIONS(2518),
    [anon_sym___strong] = ACTIONS(2518),
    [anon_sym___unsafe_unretained] = ACTIONS(2518),
    [anon_sym___unused] = ACTIONS(2518),
    [anon_sym___weak] = ACTIONS(2518),
    [anon_sym_alignas] = ACTIONS(2518),
    [anon_sym__Alignas] = ACTIONS(2518),
    [sym_primitive_type] = ACTIONS(2518),
    [anon_sym_enum] = ACTIONS(2518),
    [anon_sym_struct] = ACTIONS(2518),
    [anon_sym_union] = ACTIONS(2518),
    [anon_sym_if] = ACTIONS(2518),
    [anon_sym_else] = ACTIONS(2518),
    [anon_sym_switch] = ACTIONS(2518),
    [anon_sym_while] = ACTIONS(2518),
    [anon_sym_do] = ACTIONS(2518),
    [anon_sym_for] = ACTIONS(2518),
    [anon_sym_in] = ACTIONS(2518),
    [anon_sym_return] = ACTIONS(2518),
    [anon_sym_break] = ACTIONS(2518),
    [anon_sym_continue] = ACTIONS(2518),
    [anon_sym_goto] = ACTIONS(2518),
    [anon_sym_DASH_DASH] = ACTIONS(2520),
    [anon_sym_PLUS_PLUS] = ACTIONS(2520),
    [anon_sym_sizeof] = ACTIONS(2518),
    [anon_sym___alignof__] = ACTIONS(2518),
    [anon_sym___alignof] = ACTIONS(2518),
    [anon_sym__alignof] = ACTIONS(2518),
    [anon_sym_alignof] = ACTIONS(2518),
    [anon_sym__Alignof] = ACTIONS(2518),
    [anon_sym_offsetof] = ACTIONS(2518),
    [anon_sym__Generic] = ACTIONS(2518),
    [anon_sym_asm] = ACTIONS(2518),
    [anon_sym___asm__] = ACTIONS(2518),
    [anon_sym___asm] = ACTIONS(2518),
    [sym_number_literal] = ACTIONS(2520),
    [anon_sym_L_SQUOTE] = ACTIONS(2520),
    [anon_sym_u_SQUOTE] = ACTIONS(2520),
    [anon_sym_U_SQUOTE] = ACTIONS(2520),
    [anon_sym_u8_SQUOTE] = ACTIONS(2520),
    [anon_sym_SQUOTE] = ACTIONS(2520),
    [anon_sym_AT] = ACTIONS(2518),
    [anon_sym_DQUOTE] = ACTIONS(2520),
    [anon_sym_L_DQUOTE] = ACTIONS(2520),
    [anon_sym_u_DQUOTE] = ACTIONS(2520),
    [anon_sym_U_DQUOTE] = ACTIONS(2520),
    [anon_sym_u8_DQUOTE] = ACTIONS(2520),
    [sym_true] = ACTIONS(2518),
    [sym_false] = ACTIONS(2518),
    [anon_sym_NULL] = ACTIONS(2518),
    [anon_sym_nullptr] = ACTIONS(2518),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2518),
    [anon_sym___typeof] = ACTIONS(2518),
    [anon_sym_typeof] = ACTIONS(2518),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2518),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2518),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2518),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2518),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2518),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_API_AVAILABLE] = ACTIONS(2518),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_API_DEPRECATED] = ACTIONS(2518),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2518),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2518),
    [anon_sym___deprecated_msg] = ACTIONS(2518),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2518),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2518),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2518),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2518),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2518),
    [anon_sym_ATtry] = ACTIONS(2520),
    [anon_sym___try] = ACTIONS(2518),
    [anon_sym_ATthrow] = ACTIONS(2520),
    [anon_sym_ATselector] = ACTIONS(2520),
    [anon_sym_ATavailable] = ACTIONS(2520),
    [anon_sym___builtin_available] = ACTIONS(2518),
    [anon_sym_va_arg] = ACTIONS(2518),
    [anon_sym_ATencode] = ACTIONS(2520),
    [anon_sym_ATsynchronized] = ACTIONS(2520),
    [anon_sym_BOOL] = ACTIONS(2518),
    [anon_sym_IMP] = ACTIONS(2518),
    [anon_sym_SEL] = ACTIONS(2518),
    [anon_sym_Class] = ACTIONS(2518),
    [anon_sym_id] = ACTIONS(2518),
  },
  [STATE(1149)] = {
    [sym_identifier] = ACTIONS(2522),
    [anon_sym_LPAREN2] = ACTIONS(2524),
    [anon_sym_BANG] = ACTIONS(2524),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_DASH] = ACTIONS(2522),
    [anon_sym_PLUS] = ACTIONS(2522),
    [anon_sym_STAR] = ACTIONS(2524),
    [anon_sym_CARET] = ACTIONS(2524),
    [anon_sym_AMP] = ACTIONS(2524),
    [anon_sym_SEMI] = ACTIONS(2524),
    [anon_sym___extension__] = ACTIONS(2522),
    [anon_sym_typedef] = ACTIONS(2522),
    [anon_sym_extern] = ACTIONS(2522),
    [anon_sym___attribute__] = ACTIONS(2522),
    [anon_sym___attribute] = ACTIONS(2522),
    [anon_sym_noreturn] = ACTIONS(2522),
    [anon_sym_LBRACK] = ACTIONS(2524),
    [anon_sym___declspec] = ACTIONS(2522),
    [anon_sym_LBRACE] = ACTIONS(2524),
    [anon_sym_signed] = ACTIONS(2522),
    [anon_sym_unsigned] = ACTIONS(2522),
    [anon_sym_long] = ACTIONS(2522),
    [anon_sym_short] = ACTIONS(2522),
    [anon_sym_static] = ACTIONS(2522),
    [anon_sym_ATautoreleasepool] = ACTIONS(2524),
    [anon_sym_auto] = ACTIONS(2522),
    [anon_sym_register] = ACTIONS(2522),
    [anon_sym_inline] = ACTIONS(2522),
    [anon_sym___inline] = ACTIONS(2522),
    [anon_sym___inline__] = ACTIONS(2522),
    [anon_sym___forceinline] = ACTIONS(2522),
    [anon_sym_thread_local] = ACTIONS(2522),
    [anon_sym___thread] = ACTIONS(2522),
    [anon_sym_CG_EXTERN] = ACTIONS(2522),
    [anon_sym_CG_INLINE] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2522),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2522),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2522),
    [anon_sym_IBOutlet] = ACTIONS(2522),
    [anon_sym_IBInspectable] = ACTIONS(2522),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2522),
    [anon_sym_NS_INLINE] = ACTIONS(2522),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2522),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2522),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2522),
    [anon_sym_const] = ACTIONS(2522),
    [anon_sym_constexpr] = ACTIONS(2522),
    [anon_sym_volatile] = ACTIONS(2522),
    [anon_sym_restrict] = ACTIONS(2522),
    [anon_sym___restrict__] = ACTIONS(2522),
    [anon_sym__Atomic] = ACTIONS(2522),
    [anon_sym__Noreturn] = ACTIONS(2522),
    [anon_sym__Nonnull] = ACTIONS(2522),
    [anon_sym_nullable] = ACTIONS(2522),
    [anon_sym__Complex] = ACTIONS(2522),
    [anon_sym__Nullable] = ACTIONS(2522),
    [anon_sym__Nullable_result] = ACTIONS(2522),
    [anon_sym__Null_unspecified] = ACTIONS(2522),
    [anon_sym___autoreleasing] = ACTIONS(2522),
    [anon_sym___block] = ACTIONS(2522),
    [anon_sym___bridge] = ACTIONS(2522),
    [anon_sym___bridge_retained] = ACTIONS(2522),
    [anon_sym___bridge_transfer] = ACTIONS(2522),
    [anon_sym___complex] = ACTIONS(2522),
    [anon_sym___const] = ACTIONS(2522),
    [anon_sym___imag] = ACTIONS(2522),
    [anon_sym___kindof] = ACTIONS(2522),
    [anon_sym___nonnull] = ACTIONS(2522),
    [anon_sym___nullable] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2522),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2522),
    [anon_sym___real] = ACTIONS(2522),
    [anon_sym___strong] = ACTIONS(2522),
    [anon_sym___unsafe_unretained] = ACTIONS(2522),
    [anon_sym___unused] = ACTIONS(2522),
    [anon_sym___weak] = ACTIONS(2522),
    [anon_sym_alignas] = ACTIONS(2522),
    [anon_sym__Alignas] = ACTIONS(2522),
    [sym_primitive_type] = ACTIONS(2522),
    [anon_sym_enum] = ACTIONS(2522),
    [anon_sym_struct] = ACTIONS(2522),
    [anon_sym_union] = ACTIONS(2522),
    [anon_sym_if] = ACTIONS(2522),
    [anon_sym_else] = ACTIONS(2522),
    [anon_sym_switch] = ACTIONS(2522),
    [anon_sym_while] = ACTIONS(2522),
    [anon_sym_do] = ACTIONS(2522),
    [anon_sym_for] = ACTIONS(2522),
    [anon_sym_in] = ACTIONS(2522),
    [anon_sym_return] = ACTIONS(2522),
    [anon_sym_break] = ACTIONS(2522),
    [anon_sym_continue] = ACTIONS(2522),
    [anon_sym_goto] = ACTIONS(2522),
    [anon_sym_DASH_DASH] = ACTIONS(2524),
    [anon_sym_PLUS_PLUS] = ACTIONS(2524),
    [anon_sym_sizeof] = ACTIONS(2522),
    [anon_sym___alignof__] = ACTIONS(2522),
    [anon_sym___alignof] = ACTIONS(2522),
    [anon_sym__alignof] = ACTIONS(2522),
    [anon_sym_alignof] = ACTIONS(2522),
    [anon_sym__Alignof] = ACTIONS(2522),
    [anon_sym_offsetof] = ACTIONS(2522),
    [anon_sym__Generic] = ACTIONS(2522),
    [anon_sym_asm] = ACTIONS(2522),
    [anon_sym___asm__] = ACTIONS(2522),
    [anon_sym___asm] = ACTIONS(2522),
    [sym_number_literal] = ACTIONS(2524),
    [anon_sym_L_SQUOTE] = ACTIONS(2524),
    [anon_sym_u_SQUOTE] = ACTIONS(2524),
    [anon_sym_U_SQUOTE] = ACTIONS(2524),
    [anon_sym_u8_SQUOTE] = ACTIONS(2524),
    [anon_sym_SQUOTE] = ACTIONS(2524),
    [anon_sym_AT] = ACTIONS(2522),
    [anon_sym_DQUOTE] = ACTIONS(2524),
    [anon_sym_L_DQUOTE] = ACTIONS(2524),
    [anon_sym_u_DQUOTE] = ACTIONS(2524),
    [anon_sym_U_DQUOTE] = ACTIONS(2524),
    [anon_sym_u8_DQUOTE] = ACTIONS(2524),
    [sym_true] = ACTIONS(2522),
    [sym_false] = ACTIONS(2522),
    [anon_sym_NULL] = ACTIONS(2522),
    [anon_sym_nullptr] = ACTIONS(2522),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2522),
    [anon_sym___typeof] = ACTIONS(2522),
    [anon_sym_typeof] = ACTIONS(2522),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2522),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2522),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2522),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2522),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2522),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_API_AVAILABLE] = ACTIONS(2522),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_API_DEPRECATED] = ACTIONS(2522),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2522),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2522),
    [anon_sym___deprecated_msg] = ACTIONS(2522),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2522),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2522),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2522),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2522),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2522),
    [anon_sym_ATtry] = ACTIONS(2524),
    [anon_sym___try] = ACTIONS(2522),
    [anon_sym_ATthrow] = ACTIONS(2524),
    [anon_sym_ATselector] = ACTIONS(2524),
    [anon_sym_ATavailable] = ACTIONS(2524),
    [anon_sym___builtin_available] = ACTIONS(2522),
    [anon_sym_va_arg] = ACTIONS(2522),
    [anon_sym_ATencode] = ACTIONS(2524),
    [anon_sym_ATsynchronized] = ACTIONS(2524),
    [anon_sym_BOOL] = ACTIONS(2522),
    [anon_sym_IMP] = ACTIONS(2522),
    [anon_sym_SEL] = ACTIONS(2522),
    [anon_sym_Class] = ACTIONS(2522),
    [anon_sym_id] = ACTIONS(2522),
  },
  [STATE(1150)] = {
    [sym_identifier] = ACTIONS(2526),
    [anon_sym_LPAREN2] = ACTIONS(2528),
    [anon_sym_BANG] = ACTIONS(2528),
    [anon_sym_TILDE] = ACTIONS(2528),
    [anon_sym_DASH] = ACTIONS(2526),
    [anon_sym_PLUS] = ACTIONS(2526),
    [anon_sym_STAR] = ACTIONS(2528),
    [anon_sym_CARET] = ACTIONS(2528),
    [anon_sym_AMP] = ACTIONS(2528),
    [anon_sym_SEMI] = ACTIONS(2528),
    [anon_sym___extension__] = ACTIONS(2526),
    [anon_sym_typedef] = ACTIONS(2526),
    [anon_sym_extern] = ACTIONS(2526),
    [anon_sym___attribute__] = ACTIONS(2526),
    [anon_sym___attribute] = ACTIONS(2526),
    [anon_sym_noreturn] = ACTIONS(2526),
    [anon_sym_LBRACK] = ACTIONS(2528),
    [anon_sym___declspec] = ACTIONS(2526),
    [anon_sym_LBRACE] = ACTIONS(2528),
    [anon_sym_signed] = ACTIONS(2526),
    [anon_sym_unsigned] = ACTIONS(2526),
    [anon_sym_long] = ACTIONS(2526),
    [anon_sym_short] = ACTIONS(2526),
    [anon_sym_static] = ACTIONS(2526),
    [anon_sym_ATautoreleasepool] = ACTIONS(2528),
    [anon_sym_auto] = ACTIONS(2526),
    [anon_sym_register] = ACTIONS(2526),
    [anon_sym_inline] = ACTIONS(2526),
    [anon_sym___inline] = ACTIONS(2526),
    [anon_sym___inline__] = ACTIONS(2526),
    [anon_sym___forceinline] = ACTIONS(2526),
    [anon_sym_thread_local] = ACTIONS(2526),
    [anon_sym___thread] = ACTIONS(2526),
    [anon_sym_CG_EXTERN] = ACTIONS(2526),
    [anon_sym_CG_INLINE] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2526),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2526),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2526),
    [anon_sym_IBOutlet] = ACTIONS(2526),
    [anon_sym_IBInspectable] = ACTIONS(2526),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2526),
    [anon_sym_NS_INLINE] = ACTIONS(2526),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2526),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2526),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2526),
    [anon_sym_const] = ACTIONS(2526),
    [anon_sym_constexpr] = ACTIONS(2526),
    [anon_sym_volatile] = ACTIONS(2526),
    [anon_sym_restrict] = ACTIONS(2526),
    [anon_sym___restrict__] = ACTIONS(2526),
    [anon_sym__Atomic] = ACTIONS(2526),
    [anon_sym__Noreturn] = ACTIONS(2526),
    [anon_sym__Nonnull] = ACTIONS(2526),
    [anon_sym_nullable] = ACTIONS(2526),
    [anon_sym__Complex] = ACTIONS(2526),
    [anon_sym__Nullable] = ACTIONS(2526),
    [anon_sym__Nullable_result] = ACTIONS(2526),
    [anon_sym__Null_unspecified] = ACTIONS(2526),
    [anon_sym___autoreleasing] = ACTIONS(2526),
    [anon_sym___block] = ACTIONS(2526),
    [anon_sym___bridge] = ACTIONS(2526),
    [anon_sym___bridge_retained] = ACTIONS(2526),
    [anon_sym___bridge_transfer] = ACTIONS(2526),
    [anon_sym___complex] = ACTIONS(2526),
    [anon_sym___const] = ACTIONS(2526),
    [anon_sym___imag] = ACTIONS(2526),
    [anon_sym___kindof] = ACTIONS(2526),
    [anon_sym___nonnull] = ACTIONS(2526),
    [anon_sym___nullable] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2526),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2526),
    [anon_sym___real] = ACTIONS(2526),
    [anon_sym___strong] = ACTIONS(2526),
    [anon_sym___unsafe_unretained] = ACTIONS(2526),
    [anon_sym___unused] = ACTIONS(2526),
    [anon_sym___weak] = ACTIONS(2526),
    [anon_sym_alignas] = ACTIONS(2526),
    [anon_sym__Alignas] = ACTIONS(2526),
    [sym_primitive_type] = ACTIONS(2526),
    [anon_sym_enum] = ACTIONS(2526),
    [anon_sym_struct] = ACTIONS(2526),
    [anon_sym_union] = ACTIONS(2526),
    [anon_sym_if] = ACTIONS(2526),
    [anon_sym_else] = ACTIONS(2526),
    [anon_sym_switch] = ACTIONS(2526),
    [anon_sym_while] = ACTIONS(2526),
    [anon_sym_do] = ACTIONS(2526),
    [anon_sym_for] = ACTIONS(2526),
    [anon_sym_in] = ACTIONS(2526),
    [anon_sym_return] = ACTIONS(2526),
    [anon_sym_break] = ACTIONS(2526),
    [anon_sym_continue] = ACTIONS(2526),
    [anon_sym_goto] = ACTIONS(2526),
    [anon_sym_DASH_DASH] = ACTIONS(2528),
    [anon_sym_PLUS_PLUS] = ACTIONS(2528),
    [anon_sym_sizeof] = ACTIONS(2526),
    [anon_sym___alignof__] = ACTIONS(2526),
    [anon_sym___alignof] = ACTIONS(2526),
    [anon_sym__alignof] = ACTIONS(2526),
    [anon_sym_alignof] = ACTIONS(2526),
    [anon_sym__Alignof] = ACTIONS(2526),
    [anon_sym_offsetof] = ACTIONS(2526),
    [anon_sym__Generic] = ACTIONS(2526),
    [anon_sym_asm] = ACTIONS(2526),
    [anon_sym___asm__] = ACTIONS(2526),
    [anon_sym___asm] = ACTIONS(2526),
    [sym_number_literal] = ACTIONS(2528),
    [anon_sym_L_SQUOTE] = ACTIONS(2528),
    [anon_sym_u_SQUOTE] = ACTIONS(2528),
    [anon_sym_U_SQUOTE] = ACTIONS(2528),
    [anon_sym_u8_SQUOTE] = ACTIONS(2528),
    [anon_sym_SQUOTE] = ACTIONS(2528),
    [anon_sym_AT] = ACTIONS(2526),
    [anon_sym_DQUOTE] = ACTIONS(2528),
    [anon_sym_L_DQUOTE] = ACTIONS(2528),
    [anon_sym_u_DQUOTE] = ACTIONS(2528),
    [anon_sym_U_DQUOTE] = ACTIONS(2528),
    [anon_sym_u8_DQUOTE] = ACTIONS(2528),
    [sym_true] = ACTIONS(2526),
    [sym_false] = ACTIONS(2526),
    [anon_sym_NULL] = ACTIONS(2526),
    [anon_sym_nullptr] = ACTIONS(2526),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2526),
    [anon_sym___typeof] = ACTIONS(2526),
    [anon_sym_typeof] = ACTIONS(2526),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2526),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2526),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2526),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2526),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2526),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_API_AVAILABLE] = ACTIONS(2526),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_API_DEPRECATED] = ACTIONS(2526),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2526),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2526),
    [anon_sym___deprecated_msg] = ACTIONS(2526),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2526),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2526),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2526),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2526),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2526),
    [anon_sym_ATtry] = ACTIONS(2528),
    [anon_sym___try] = ACTIONS(2526),
    [anon_sym_ATthrow] = ACTIONS(2528),
    [anon_sym_ATselector] = ACTIONS(2528),
    [anon_sym_ATavailable] = ACTIONS(2528),
    [anon_sym___builtin_available] = ACTIONS(2526),
    [anon_sym_va_arg] = ACTIONS(2526),
    [anon_sym_ATencode] = ACTIONS(2528),
    [anon_sym_ATsynchronized] = ACTIONS(2528),
    [anon_sym_BOOL] = ACTIONS(2526),
    [anon_sym_IMP] = ACTIONS(2526),
    [anon_sym_SEL] = ACTIONS(2526),
    [anon_sym_Class] = ACTIONS(2526),
    [anon_sym_id] = ACTIONS(2526),
  },
  [STATE(1151)] = {
    [sym_identifier] = ACTIONS(2530),
    [anon_sym_LPAREN2] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2530),
    [anon_sym_PLUS] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym___extension__] = ACTIONS(2530),
    [anon_sym_typedef] = ACTIONS(2530),
    [anon_sym_extern] = ACTIONS(2530),
    [anon_sym___attribute__] = ACTIONS(2530),
    [anon_sym___attribute] = ACTIONS(2530),
    [anon_sym_noreturn] = ACTIONS(2530),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym___declspec] = ACTIONS(2530),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_signed] = ACTIONS(2530),
    [anon_sym_unsigned] = ACTIONS(2530),
    [anon_sym_long] = ACTIONS(2530),
    [anon_sym_short] = ACTIONS(2530),
    [anon_sym_static] = ACTIONS(2530),
    [anon_sym_ATautoreleasepool] = ACTIONS(2532),
    [anon_sym_auto] = ACTIONS(2530),
    [anon_sym_register] = ACTIONS(2530),
    [anon_sym_inline] = ACTIONS(2530),
    [anon_sym___inline] = ACTIONS(2530),
    [anon_sym___inline__] = ACTIONS(2530),
    [anon_sym___forceinline] = ACTIONS(2530),
    [anon_sym_thread_local] = ACTIONS(2530),
    [anon_sym___thread] = ACTIONS(2530),
    [anon_sym_CG_EXTERN] = ACTIONS(2530),
    [anon_sym_CG_INLINE] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2530),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2530),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2530),
    [anon_sym_IBOutlet] = ACTIONS(2530),
    [anon_sym_IBInspectable] = ACTIONS(2530),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2530),
    [anon_sym_NS_INLINE] = ACTIONS(2530),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2530),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2530),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2530),
    [anon_sym_const] = ACTIONS(2530),
    [anon_sym_constexpr] = ACTIONS(2530),
    [anon_sym_volatile] = ACTIONS(2530),
    [anon_sym_restrict] = ACTIONS(2530),
    [anon_sym___restrict__] = ACTIONS(2530),
    [anon_sym__Atomic] = ACTIONS(2530),
    [anon_sym__Noreturn] = ACTIONS(2530),
    [anon_sym__Nonnull] = ACTIONS(2530),
    [anon_sym_nullable] = ACTIONS(2530),
    [anon_sym__Complex] = ACTIONS(2530),
    [anon_sym__Nullable] = ACTIONS(2530),
    [anon_sym__Nullable_result] = ACTIONS(2530),
    [anon_sym__Null_unspecified] = ACTIONS(2530),
    [anon_sym___autoreleasing] = ACTIONS(2530),
    [anon_sym___block] = ACTIONS(2530),
    [anon_sym___bridge] = ACTIONS(2530),
    [anon_sym___bridge_retained] = ACTIONS(2530),
    [anon_sym___bridge_transfer] = ACTIONS(2530),
    [anon_sym___complex] = ACTIONS(2530),
    [anon_sym___const] = ACTIONS(2530),
    [anon_sym___imag] = ACTIONS(2530),
    [anon_sym___kindof] = ACTIONS(2530),
    [anon_sym___nonnull] = ACTIONS(2530),
    [anon_sym___nullable] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2530),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2530),
    [anon_sym___real] = ACTIONS(2530),
    [anon_sym___strong] = ACTIONS(2530),
    [anon_sym___unsafe_unretained] = ACTIONS(2530),
    [anon_sym___unused] = ACTIONS(2530),
    [anon_sym___weak] = ACTIONS(2530),
    [anon_sym_alignas] = ACTIONS(2530),
    [anon_sym__Alignas] = ACTIONS(2530),
    [sym_primitive_type] = ACTIONS(2530),
    [anon_sym_enum] = ACTIONS(2530),
    [anon_sym_struct] = ACTIONS(2530),
    [anon_sym_union] = ACTIONS(2530),
    [anon_sym_if] = ACTIONS(2530),
    [anon_sym_else] = ACTIONS(2530),
    [anon_sym_switch] = ACTIONS(2530),
    [anon_sym_while] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2530),
    [anon_sym_for] = ACTIONS(2530),
    [anon_sym_in] = ACTIONS(2530),
    [anon_sym_return] = ACTIONS(2530),
    [anon_sym_break] = ACTIONS(2530),
    [anon_sym_continue] = ACTIONS(2530),
    [anon_sym_goto] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_sizeof] = ACTIONS(2530),
    [anon_sym___alignof__] = ACTIONS(2530),
    [anon_sym___alignof] = ACTIONS(2530),
    [anon_sym__alignof] = ACTIONS(2530),
    [anon_sym_alignof] = ACTIONS(2530),
    [anon_sym__Alignof] = ACTIONS(2530),
    [anon_sym_offsetof] = ACTIONS(2530),
    [anon_sym__Generic] = ACTIONS(2530),
    [anon_sym_asm] = ACTIONS(2530),
    [anon_sym___asm__] = ACTIONS(2530),
    [anon_sym___asm] = ACTIONS(2530),
    [sym_number_literal] = ACTIONS(2532),
    [anon_sym_L_SQUOTE] = ACTIONS(2532),
    [anon_sym_u_SQUOTE] = ACTIONS(2532),
    [anon_sym_U_SQUOTE] = ACTIONS(2532),
    [anon_sym_u8_SQUOTE] = ACTIONS(2532),
    [anon_sym_SQUOTE] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2530),
    [anon_sym_DQUOTE] = ACTIONS(2532),
    [anon_sym_L_DQUOTE] = ACTIONS(2532),
    [anon_sym_u_DQUOTE] = ACTIONS(2532),
    [anon_sym_U_DQUOTE] = ACTIONS(2532),
    [anon_sym_u8_DQUOTE] = ACTIONS(2532),
    [sym_true] = ACTIONS(2530),
    [sym_false] = ACTIONS(2530),
    [anon_sym_NULL] = ACTIONS(2530),
    [anon_sym_nullptr] = ACTIONS(2530),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2530),
    [anon_sym___typeof] = ACTIONS(2530),
    [anon_sym_typeof] = ACTIONS(2530),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2530),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2530),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2530),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2530),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2530),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_API_AVAILABLE] = ACTIONS(2530),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_API_DEPRECATED] = ACTIONS(2530),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2530),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2530),
    [anon_sym___deprecated_msg] = ACTIONS(2530),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2530),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2530),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2530),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2530),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2530),
    [anon_sym_ATtry] = ACTIONS(2532),
    [anon_sym___try] = ACTIONS(2530),
    [anon_sym_ATthrow] = ACTIONS(2532),
    [anon_sym_ATselector] = ACTIONS(2532),
    [anon_sym_ATavailable] = ACTIONS(2532),
    [anon_sym___builtin_available] = ACTIONS(2530),
    [anon_sym_va_arg] = ACTIONS(2530),
    [anon_sym_ATencode] = ACTIONS(2532),
    [anon_sym_ATsynchronized] = ACTIONS(2532),
    [anon_sym_BOOL] = ACTIONS(2530),
    [anon_sym_IMP] = ACTIONS(2530),
    [anon_sym_SEL] = ACTIONS(2530),
    [anon_sym_Class] = ACTIONS(2530),
    [anon_sym_id] = ACTIONS(2530),
  },
  [STATE(1152)] = {
    [sym_identifier] = ACTIONS(2534),
    [anon_sym_LPAREN2] = ACTIONS(2536),
    [anon_sym_BANG] = ACTIONS(2536),
    [anon_sym_TILDE] = ACTIONS(2536),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2536),
    [anon_sym_CARET] = ACTIONS(2536),
    [anon_sym_AMP] = ACTIONS(2536),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym___extension__] = ACTIONS(2534),
    [anon_sym_typedef] = ACTIONS(2534),
    [anon_sym_extern] = ACTIONS(2534),
    [anon_sym___attribute__] = ACTIONS(2534),
    [anon_sym___attribute] = ACTIONS(2534),
    [anon_sym_noreturn] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2536),
    [anon_sym___declspec] = ACTIONS(2534),
    [anon_sym_LBRACE] = ACTIONS(2536),
    [anon_sym_signed] = ACTIONS(2534),
    [anon_sym_unsigned] = ACTIONS(2534),
    [anon_sym_long] = ACTIONS(2534),
    [anon_sym_short] = ACTIONS(2534),
    [anon_sym_static] = ACTIONS(2534),
    [anon_sym_ATautoreleasepool] = ACTIONS(2536),
    [anon_sym_auto] = ACTIONS(2534),
    [anon_sym_register] = ACTIONS(2534),
    [anon_sym_inline] = ACTIONS(2534),
    [anon_sym___inline] = ACTIONS(2534),
    [anon_sym___inline__] = ACTIONS(2534),
    [anon_sym___forceinline] = ACTIONS(2534),
    [anon_sym_thread_local] = ACTIONS(2534),
    [anon_sym___thread] = ACTIONS(2534),
    [anon_sym_CG_EXTERN] = ACTIONS(2534),
    [anon_sym_CG_INLINE] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2534),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2534),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2534),
    [anon_sym_IBOutlet] = ACTIONS(2534),
    [anon_sym_IBInspectable] = ACTIONS(2534),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2534),
    [anon_sym_NS_INLINE] = ACTIONS(2534),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2534),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2534),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2534),
    [anon_sym_const] = ACTIONS(2534),
    [anon_sym_constexpr] = ACTIONS(2534),
    [anon_sym_volatile] = ACTIONS(2534),
    [anon_sym_restrict] = ACTIONS(2534),
    [anon_sym___restrict__] = ACTIONS(2534),
    [anon_sym__Atomic] = ACTIONS(2534),
    [anon_sym__Noreturn] = ACTIONS(2534),
    [anon_sym__Nonnull] = ACTIONS(2534),
    [anon_sym_nullable] = ACTIONS(2534),
    [anon_sym__Complex] = ACTIONS(2534),
    [anon_sym__Nullable] = ACTIONS(2534),
    [anon_sym__Nullable_result] = ACTIONS(2534),
    [anon_sym__Null_unspecified] = ACTIONS(2534),
    [anon_sym___autoreleasing] = ACTIONS(2534),
    [anon_sym___block] = ACTIONS(2534),
    [anon_sym___bridge] = ACTIONS(2534),
    [anon_sym___bridge_retained] = ACTIONS(2534),
    [anon_sym___bridge_transfer] = ACTIONS(2534),
    [anon_sym___complex] = ACTIONS(2534),
    [anon_sym___const] = ACTIONS(2534),
    [anon_sym___imag] = ACTIONS(2534),
    [anon_sym___kindof] = ACTIONS(2534),
    [anon_sym___nonnull] = ACTIONS(2534),
    [anon_sym___nullable] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2534),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2534),
    [anon_sym___real] = ACTIONS(2534),
    [anon_sym___strong] = ACTIONS(2534),
    [anon_sym___unsafe_unretained] = ACTIONS(2534),
    [anon_sym___unused] = ACTIONS(2534),
    [anon_sym___weak] = ACTIONS(2534),
    [anon_sym_alignas] = ACTIONS(2534),
    [anon_sym__Alignas] = ACTIONS(2534),
    [sym_primitive_type] = ACTIONS(2534),
    [anon_sym_enum] = ACTIONS(2534),
    [anon_sym_struct] = ACTIONS(2534),
    [anon_sym_union] = ACTIONS(2534),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_else] = ACTIONS(2534),
    [anon_sym_switch] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_do] = ACTIONS(2534),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_in] = ACTIONS(2534),
    [anon_sym_return] = ACTIONS(2534),
    [anon_sym_break] = ACTIONS(2534),
    [anon_sym_continue] = ACTIONS(2534),
    [anon_sym_goto] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_sizeof] = ACTIONS(2534),
    [anon_sym___alignof__] = ACTIONS(2534),
    [anon_sym___alignof] = ACTIONS(2534),
    [anon_sym__alignof] = ACTIONS(2534),
    [anon_sym_alignof] = ACTIONS(2534),
    [anon_sym__Alignof] = ACTIONS(2534),
    [anon_sym_offsetof] = ACTIONS(2534),
    [anon_sym__Generic] = ACTIONS(2534),
    [anon_sym_asm] = ACTIONS(2534),
    [anon_sym___asm__] = ACTIONS(2534),
    [anon_sym___asm] = ACTIONS(2534),
    [sym_number_literal] = ACTIONS(2536),
    [anon_sym_L_SQUOTE] = ACTIONS(2536),
    [anon_sym_u_SQUOTE] = ACTIONS(2536),
    [anon_sym_U_SQUOTE] = ACTIONS(2536),
    [anon_sym_u8_SQUOTE] = ACTIONS(2536),
    [anon_sym_SQUOTE] = ACTIONS(2536),
    [anon_sym_AT] = ACTIONS(2534),
    [anon_sym_DQUOTE] = ACTIONS(2536),
    [anon_sym_L_DQUOTE] = ACTIONS(2536),
    [anon_sym_u_DQUOTE] = ACTIONS(2536),
    [anon_sym_U_DQUOTE] = ACTIONS(2536),
    [anon_sym_u8_DQUOTE] = ACTIONS(2536),
    [sym_true] = ACTIONS(2534),
    [sym_false] = ACTIONS(2534),
    [anon_sym_NULL] = ACTIONS(2534),
    [anon_sym_nullptr] = ACTIONS(2534),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2534),
    [anon_sym___typeof] = ACTIONS(2534),
    [anon_sym_typeof] = ACTIONS(2534),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2534),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2534),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2534),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2534),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2534),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_API_AVAILABLE] = ACTIONS(2534),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_API_DEPRECATED] = ACTIONS(2534),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2534),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2534),
    [anon_sym___deprecated_msg] = ACTIONS(2534),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2534),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2534),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2534),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2534),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2534),
    [anon_sym_ATtry] = ACTIONS(2536),
    [anon_sym___try] = ACTIONS(2534),
    [anon_sym_ATthrow] = ACTIONS(2536),
    [anon_sym_ATselector] = ACTIONS(2536),
    [anon_sym_ATavailable] = ACTIONS(2536),
    [anon_sym___builtin_available] = ACTIONS(2534),
    [anon_sym_va_arg] = ACTIONS(2534),
    [anon_sym_ATencode] = ACTIONS(2536),
    [anon_sym_ATsynchronized] = ACTIONS(2536),
    [anon_sym_BOOL] = ACTIONS(2534),
    [anon_sym_IMP] = ACTIONS(2534),
    [anon_sym_SEL] = ACTIONS(2534),
    [anon_sym_Class] = ACTIONS(2534),
    [anon_sym_id] = ACTIONS(2534),
  },
  [STATE(1153)] = {
    [sym_identifier] = ACTIONS(2538),
    [anon_sym_LPAREN2] = ACTIONS(2540),
    [anon_sym_BANG] = ACTIONS(2540),
    [anon_sym_TILDE] = ACTIONS(2540),
    [anon_sym_DASH] = ACTIONS(2538),
    [anon_sym_PLUS] = ACTIONS(2538),
    [anon_sym_STAR] = ACTIONS(2540),
    [anon_sym_CARET] = ACTIONS(2540),
    [anon_sym_AMP] = ACTIONS(2540),
    [anon_sym_SEMI] = ACTIONS(2540),
    [anon_sym___extension__] = ACTIONS(2538),
    [anon_sym_typedef] = ACTIONS(2538),
    [anon_sym_extern] = ACTIONS(2538),
    [anon_sym___attribute__] = ACTIONS(2538),
    [anon_sym___attribute] = ACTIONS(2538),
    [anon_sym_noreturn] = ACTIONS(2538),
    [anon_sym_LBRACK] = ACTIONS(2540),
    [anon_sym___declspec] = ACTIONS(2538),
    [anon_sym_LBRACE] = ACTIONS(2540),
    [anon_sym_signed] = ACTIONS(2538),
    [anon_sym_unsigned] = ACTIONS(2538),
    [anon_sym_long] = ACTIONS(2538),
    [anon_sym_short] = ACTIONS(2538),
    [anon_sym_static] = ACTIONS(2538),
    [anon_sym_ATautoreleasepool] = ACTIONS(2540),
    [anon_sym_auto] = ACTIONS(2538),
    [anon_sym_register] = ACTIONS(2538),
    [anon_sym_inline] = ACTIONS(2538),
    [anon_sym___inline] = ACTIONS(2538),
    [anon_sym___inline__] = ACTIONS(2538),
    [anon_sym___forceinline] = ACTIONS(2538),
    [anon_sym_thread_local] = ACTIONS(2538),
    [anon_sym___thread] = ACTIONS(2538),
    [anon_sym_CG_EXTERN] = ACTIONS(2538),
    [anon_sym_CG_INLINE] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2538),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2538),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2538),
    [anon_sym_IBOutlet] = ACTIONS(2538),
    [anon_sym_IBInspectable] = ACTIONS(2538),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2538),
    [anon_sym_NS_INLINE] = ACTIONS(2538),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2538),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2538),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2538),
    [anon_sym_const] = ACTIONS(2538),
    [anon_sym_constexpr] = ACTIONS(2538),
    [anon_sym_volatile] = ACTIONS(2538),
    [anon_sym_restrict] = ACTIONS(2538),
    [anon_sym___restrict__] = ACTIONS(2538),
    [anon_sym__Atomic] = ACTIONS(2538),
    [anon_sym__Noreturn] = ACTIONS(2538),
    [anon_sym__Nonnull] = ACTIONS(2538),
    [anon_sym_nullable] = ACTIONS(2538),
    [anon_sym__Complex] = ACTIONS(2538),
    [anon_sym__Nullable] = ACTIONS(2538),
    [anon_sym__Nullable_result] = ACTIONS(2538),
    [anon_sym__Null_unspecified] = ACTIONS(2538),
    [anon_sym___autoreleasing] = ACTIONS(2538),
    [anon_sym___block] = ACTIONS(2538),
    [anon_sym___bridge] = ACTIONS(2538),
    [anon_sym___bridge_retained] = ACTIONS(2538),
    [anon_sym___bridge_transfer] = ACTIONS(2538),
    [anon_sym___complex] = ACTIONS(2538),
    [anon_sym___const] = ACTIONS(2538),
    [anon_sym___imag] = ACTIONS(2538),
    [anon_sym___kindof] = ACTIONS(2538),
    [anon_sym___nonnull] = ACTIONS(2538),
    [anon_sym___nullable] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2538),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2538),
    [anon_sym___real] = ACTIONS(2538),
    [anon_sym___strong] = ACTIONS(2538),
    [anon_sym___unsafe_unretained] = ACTIONS(2538),
    [anon_sym___unused] = ACTIONS(2538),
    [anon_sym___weak] = ACTIONS(2538),
    [anon_sym_alignas] = ACTIONS(2538),
    [anon_sym__Alignas] = ACTIONS(2538),
    [sym_primitive_type] = ACTIONS(2538),
    [anon_sym_enum] = ACTIONS(2538),
    [anon_sym_struct] = ACTIONS(2538),
    [anon_sym_union] = ACTIONS(2538),
    [anon_sym_if] = ACTIONS(2538),
    [anon_sym_else] = ACTIONS(2538),
    [anon_sym_switch] = ACTIONS(2538),
    [anon_sym_while] = ACTIONS(2538),
    [anon_sym_do] = ACTIONS(2538),
    [anon_sym_for] = ACTIONS(2538),
    [anon_sym_in] = ACTIONS(2538),
    [anon_sym_return] = ACTIONS(2538),
    [anon_sym_break] = ACTIONS(2538),
    [anon_sym_continue] = ACTIONS(2538),
    [anon_sym_goto] = ACTIONS(2538),
    [anon_sym_DASH_DASH] = ACTIONS(2540),
    [anon_sym_PLUS_PLUS] = ACTIONS(2540),
    [anon_sym_sizeof] = ACTIONS(2538),
    [anon_sym___alignof__] = ACTIONS(2538),
    [anon_sym___alignof] = ACTIONS(2538),
    [anon_sym__alignof] = ACTIONS(2538),
    [anon_sym_alignof] = ACTIONS(2538),
    [anon_sym__Alignof] = ACTIONS(2538),
    [anon_sym_offsetof] = ACTIONS(2538),
    [anon_sym__Generic] = ACTIONS(2538),
    [anon_sym_asm] = ACTIONS(2538),
    [anon_sym___asm__] = ACTIONS(2538),
    [anon_sym___asm] = ACTIONS(2538),
    [sym_number_literal] = ACTIONS(2540),
    [anon_sym_L_SQUOTE] = ACTIONS(2540),
    [anon_sym_u_SQUOTE] = ACTIONS(2540),
    [anon_sym_U_SQUOTE] = ACTIONS(2540),
    [anon_sym_u8_SQUOTE] = ACTIONS(2540),
    [anon_sym_SQUOTE] = ACTIONS(2540),
    [anon_sym_AT] = ACTIONS(2538),
    [anon_sym_DQUOTE] = ACTIONS(2540),
    [anon_sym_L_DQUOTE] = ACTIONS(2540),
    [anon_sym_u_DQUOTE] = ACTIONS(2540),
    [anon_sym_U_DQUOTE] = ACTIONS(2540),
    [anon_sym_u8_DQUOTE] = ACTIONS(2540),
    [sym_true] = ACTIONS(2538),
    [sym_false] = ACTIONS(2538),
    [anon_sym_NULL] = ACTIONS(2538),
    [anon_sym_nullptr] = ACTIONS(2538),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2538),
    [anon_sym___typeof] = ACTIONS(2538),
    [anon_sym_typeof] = ACTIONS(2538),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2538),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2538),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2538),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2538),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2538),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_API_AVAILABLE] = ACTIONS(2538),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_API_DEPRECATED] = ACTIONS(2538),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2538),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2538),
    [anon_sym___deprecated_msg] = ACTIONS(2538),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2538),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2538),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2538),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2538),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2538),
    [anon_sym_ATtry] = ACTIONS(2540),
    [anon_sym___try] = ACTIONS(2538),
    [anon_sym_ATthrow] = ACTIONS(2540),
    [anon_sym_ATselector] = ACTIONS(2540),
    [anon_sym_ATavailable] = ACTIONS(2540),
    [anon_sym___builtin_available] = ACTIONS(2538),
    [anon_sym_va_arg] = ACTIONS(2538),
    [anon_sym_ATencode] = ACTIONS(2540),
    [anon_sym_ATsynchronized] = ACTIONS(2540),
    [anon_sym_BOOL] = ACTIONS(2538),
    [anon_sym_IMP] = ACTIONS(2538),
    [anon_sym_SEL] = ACTIONS(2538),
    [anon_sym_Class] = ACTIONS(2538),
    [anon_sym_id] = ACTIONS(2538),
  },
  [STATE(1154)] = {
    [sym_identifier] = ACTIONS(2450),
    [anon_sym_LPAREN2] = ACTIONS(2452),
    [anon_sym_BANG] = ACTIONS(2452),
    [anon_sym_TILDE] = ACTIONS(2452),
    [anon_sym_DASH] = ACTIONS(2450),
    [anon_sym_PLUS] = ACTIONS(2450),
    [anon_sym_STAR] = ACTIONS(2452),
    [anon_sym_CARET] = ACTIONS(2452),
    [anon_sym_AMP] = ACTIONS(2452),
    [anon_sym_SEMI] = ACTIONS(2452),
    [anon_sym___extension__] = ACTIONS(2450),
    [anon_sym_typedef] = ACTIONS(2450),
    [anon_sym_extern] = ACTIONS(2450),
    [anon_sym___attribute__] = ACTIONS(2450),
    [anon_sym___attribute] = ACTIONS(2450),
    [anon_sym_noreturn] = ACTIONS(2450),
    [anon_sym_LBRACK] = ACTIONS(2452),
    [anon_sym___declspec] = ACTIONS(2450),
    [anon_sym_LBRACE] = ACTIONS(2452),
    [anon_sym_signed] = ACTIONS(2450),
    [anon_sym_unsigned] = ACTIONS(2450),
    [anon_sym_long] = ACTIONS(2450),
    [anon_sym_short] = ACTIONS(2450),
    [anon_sym_static] = ACTIONS(2450),
    [anon_sym_ATautoreleasepool] = ACTIONS(2452),
    [anon_sym_auto] = ACTIONS(2450),
    [anon_sym_register] = ACTIONS(2450),
    [anon_sym_inline] = ACTIONS(2450),
    [anon_sym___inline] = ACTIONS(2450),
    [anon_sym___inline__] = ACTIONS(2450),
    [anon_sym___forceinline] = ACTIONS(2450),
    [anon_sym_thread_local] = ACTIONS(2450),
    [anon_sym___thread] = ACTIONS(2450),
    [anon_sym_CG_EXTERN] = ACTIONS(2450),
    [anon_sym_CG_INLINE] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2450),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2450),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2450),
    [anon_sym_IBOutlet] = ACTIONS(2450),
    [anon_sym_IBInspectable] = ACTIONS(2450),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2450),
    [anon_sym_NS_INLINE] = ACTIONS(2450),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2450),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2450),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2450),
    [anon_sym_const] = ACTIONS(2450),
    [anon_sym_constexpr] = ACTIONS(2450),
    [anon_sym_volatile] = ACTIONS(2450),
    [anon_sym_restrict] = ACTIONS(2450),
    [anon_sym___restrict__] = ACTIONS(2450),
    [anon_sym__Atomic] = ACTIONS(2450),
    [anon_sym__Noreturn] = ACTIONS(2450),
    [anon_sym__Nonnull] = ACTIONS(2450),
    [anon_sym_nullable] = ACTIONS(2450),
    [anon_sym__Complex] = ACTIONS(2450),
    [anon_sym__Nullable] = ACTIONS(2450),
    [anon_sym__Nullable_result] = ACTIONS(2450),
    [anon_sym__Null_unspecified] = ACTIONS(2450),
    [anon_sym___autoreleasing] = ACTIONS(2450),
    [anon_sym___block] = ACTIONS(2450),
    [anon_sym___bridge] = ACTIONS(2450),
    [anon_sym___bridge_retained] = ACTIONS(2450),
    [anon_sym___bridge_transfer] = ACTIONS(2450),
    [anon_sym___complex] = ACTIONS(2450),
    [anon_sym___const] = ACTIONS(2450),
    [anon_sym___imag] = ACTIONS(2450),
    [anon_sym___kindof] = ACTIONS(2450),
    [anon_sym___nonnull] = ACTIONS(2450),
    [anon_sym___nullable] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2450),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2450),
    [anon_sym___real] = ACTIONS(2450),
    [anon_sym___strong] = ACTIONS(2450),
    [anon_sym___unsafe_unretained] = ACTIONS(2450),
    [anon_sym___unused] = ACTIONS(2450),
    [anon_sym___weak] = ACTIONS(2450),
    [anon_sym_alignas] = ACTIONS(2450),
    [anon_sym__Alignas] = ACTIONS(2450),
    [sym_primitive_type] = ACTIONS(2450),
    [anon_sym_enum] = ACTIONS(2450),
    [anon_sym_struct] = ACTIONS(2450),
    [anon_sym_union] = ACTIONS(2450),
    [anon_sym_if] = ACTIONS(2450),
    [anon_sym_else] = ACTIONS(2450),
    [anon_sym_switch] = ACTIONS(2450),
    [anon_sym_while] = ACTIONS(2450),
    [anon_sym_do] = ACTIONS(2450),
    [anon_sym_for] = ACTIONS(2450),
    [anon_sym_in] = ACTIONS(2450),
    [anon_sym_return] = ACTIONS(2450),
    [anon_sym_break] = ACTIONS(2450),
    [anon_sym_continue] = ACTIONS(2450),
    [anon_sym_goto] = ACTIONS(2450),
    [anon_sym_DASH_DASH] = ACTIONS(2452),
    [anon_sym_PLUS_PLUS] = ACTIONS(2452),
    [anon_sym_sizeof] = ACTIONS(2450),
    [anon_sym___alignof__] = ACTIONS(2450),
    [anon_sym___alignof] = ACTIONS(2450),
    [anon_sym__alignof] = ACTIONS(2450),
    [anon_sym_alignof] = ACTIONS(2450),
    [anon_sym__Alignof] = ACTIONS(2450),
    [anon_sym_offsetof] = ACTIONS(2450),
    [anon_sym__Generic] = ACTIONS(2450),
    [anon_sym_asm] = ACTIONS(2450),
    [anon_sym___asm__] = ACTIONS(2450),
    [anon_sym___asm] = ACTIONS(2450),
    [sym_number_literal] = ACTIONS(2452),
    [anon_sym_L_SQUOTE] = ACTIONS(2452),
    [anon_sym_u_SQUOTE] = ACTIONS(2452),
    [anon_sym_U_SQUOTE] = ACTIONS(2452),
    [anon_sym_u8_SQUOTE] = ACTIONS(2452),
    [anon_sym_SQUOTE] = ACTIONS(2452),
    [anon_sym_AT] = ACTIONS(2450),
    [anon_sym_DQUOTE] = ACTIONS(2452),
    [anon_sym_L_DQUOTE] = ACTIONS(2452),
    [anon_sym_u_DQUOTE] = ACTIONS(2452),
    [anon_sym_U_DQUOTE] = ACTIONS(2452),
    [anon_sym_u8_DQUOTE] = ACTIONS(2452),
    [sym_true] = ACTIONS(2450),
    [sym_false] = ACTIONS(2450),
    [anon_sym_NULL] = ACTIONS(2450),
    [anon_sym_nullptr] = ACTIONS(2450),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2450),
    [anon_sym___typeof] = ACTIONS(2450),
    [anon_sym_typeof] = ACTIONS(2450),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2450),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2450),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2450),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2450),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2450),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_API_AVAILABLE] = ACTIONS(2450),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_API_DEPRECATED] = ACTIONS(2450),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2450),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2450),
    [anon_sym___deprecated_msg] = ACTIONS(2450),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2450),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2450),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2450),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2450),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2450),
    [anon_sym_ATtry] = ACTIONS(2452),
    [anon_sym___try] = ACTIONS(2450),
    [anon_sym_ATthrow] = ACTIONS(2452),
    [anon_sym_ATselector] = ACTIONS(2452),
    [anon_sym_ATavailable] = ACTIONS(2452),
    [anon_sym___builtin_available] = ACTIONS(2450),
    [anon_sym_va_arg] = ACTIONS(2450),
    [anon_sym_ATencode] = ACTIONS(2452),
    [anon_sym_ATsynchronized] = ACTIONS(2452),
    [anon_sym_BOOL] = ACTIONS(2450),
    [anon_sym_IMP] = ACTIONS(2450),
    [anon_sym_SEL] = ACTIONS(2450),
    [anon_sym_Class] = ACTIONS(2450),
    [anon_sym_id] = ACTIONS(2450),
  },
  [STATE(1155)] = {
    [sym_identifier] = ACTIONS(2458),
    [anon_sym_LPAREN2] = ACTIONS(2460),
    [anon_sym_BANG] = ACTIONS(2460),
    [anon_sym_TILDE] = ACTIONS(2460),
    [anon_sym_DASH] = ACTIONS(2458),
    [anon_sym_PLUS] = ACTIONS(2458),
    [anon_sym_STAR] = ACTIONS(2460),
    [anon_sym_CARET] = ACTIONS(2460),
    [anon_sym_AMP] = ACTIONS(2460),
    [anon_sym_SEMI] = ACTIONS(2460),
    [anon_sym___extension__] = ACTIONS(2458),
    [anon_sym_typedef] = ACTIONS(2458),
    [anon_sym_extern] = ACTIONS(2458),
    [anon_sym___attribute__] = ACTIONS(2458),
    [anon_sym___attribute] = ACTIONS(2458),
    [anon_sym_noreturn] = ACTIONS(2458),
    [anon_sym_LBRACK] = ACTIONS(2460),
    [anon_sym___declspec] = ACTIONS(2458),
    [anon_sym_LBRACE] = ACTIONS(2460),
    [anon_sym_signed] = ACTIONS(2458),
    [anon_sym_unsigned] = ACTIONS(2458),
    [anon_sym_long] = ACTIONS(2458),
    [anon_sym_short] = ACTIONS(2458),
    [anon_sym_static] = ACTIONS(2458),
    [anon_sym_ATautoreleasepool] = ACTIONS(2460),
    [anon_sym_auto] = ACTIONS(2458),
    [anon_sym_register] = ACTIONS(2458),
    [anon_sym_inline] = ACTIONS(2458),
    [anon_sym___inline] = ACTIONS(2458),
    [anon_sym___inline__] = ACTIONS(2458),
    [anon_sym___forceinline] = ACTIONS(2458),
    [anon_sym_thread_local] = ACTIONS(2458),
    [anon_sym___thread] = ACTIONS(2458),
    [anon_sym_CG_EXTERN] = ACTIONS(2458),
    [anon_sym_CG_INLINE] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2458),
    [anon_sym_IBOutlet] = ACTIONS(2458),
    [anon_sym_IBInspectable] = ACTIONS(2458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2458),
    [anon_sym_NS_INLINE] = ACTIONS(2458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2458),
    [anon_sym_const] = ACTIONS(2458),
    [anon_sym_constexpr] = ACTIONS(2458),
    [anon_sym_volatile] = ACTIONS(2458),
    [anon_sym_restrict] = ACTIONS(2458),
    [anon_sym___restrict__] = ACTIONS(2458),
    [anon_sym__Atomic] = ACTIONS(2458),
    [anon_sym__Noreturn] = ACTIONS(2458),
    [anon_sym__Nonnull] = ACTIONS(2458),
    [anon_sym_nullable] = ACTIONS(2458),
    [anon_sym__Complex] = ACTIONS(2458),
    [anon_sym__Nullable] = ACTIONS(2458),
    [anon_sym__Nullable_result] = ACTIONS(2458),
    [anon_sym__Null_unspecified] = ACTIONS(2458),
    [anon_sym___autoreleasing] = ACTIONS(2458),
    [anon_sym___block] = ACTIONS(2458),
    [anon_sym___bridge] = ACTIONS(2458),
    [anon_sym___bridge_retained] = ACTIONS(2458),
    [anon_sym___bridge_transfer] = ACTIONS(2458),
    [anon_sym___complex] = ACTIONS(2458),
    [anon_sym___const] = ACTIONS(2458),
    [anon_sym___imag] = ACTIONS(2458),
    [anon_sym___kindof] = ACTIONS(2458),
    [anon_sym___nonnull] = ACTIONS(2458),
    [anon_sym___nullable] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2458),
    [anon_sym___real] = ACTIONS(2458),
    [anon_sym___strong] = ACTIONS(2458),
    [anon_sym___unsafe_unretained] = ACTIONS(2458),
    [anon_sym___unused] = ACTIONS(2458),
    [anon_sym___weak] = ACTIONS(2458),
    [anon_sym_alignas] = ACTIONS(2458),
    [anon_sym__Alignas] = ACTIONS(2458),
    [sym_primitive_type] = ACTIONS(2458),
    [anon_sym_enum] = ACTIONS(2458),
    [anon_sym_struct] = ACTIONS(2458),
    [anon_sym_union] = ACTIONS(2458),
    [anon_sym_if] = ACTIONS(2458),
    [anon_sym_else] = ACTIONS(2458),
    [anon_sym_switch] = ACTIONS(2458),
    [anon_sym_while] = ACTIONS(2458),
    [anon_sym_do] = ACTIONS(2458),
    [anon_sym_for] = ACTIONS(2458),
    [anon_sym_in] = ACTIONS(2458),
    [anon_sym_return] = ACTIONS(2458),
    [anon_sym_break] = ACTIONS(2458),
    [anon_sym_continue] = ACTIONS(2458),
    [anon_sym_goto] = ACTIONS(2458),
    [anon_sym_DASH_DASH] = ACTIONS(2460),
    [anon_sym_PLUS_PLUS] = ACTIONS(2460),
    [anon_sym_sizeof] = ACTIONS(2458),
    [anon_sym___alignof__] = ACTIONS(2458),
    [anon_sym___alignof] = ACTIONS(2458),
    [anon_sym__alignof] = ACTIONS(2458),
    [anon_sym_alignof] = ACTIONS(2458),
    [anon_sym__Alignof] = ACTIONS(2458),
    [anon_sym_offsetof] = ACTIONS(2458),
    [anon_sym__Generic] = ACTIONS(2458),
    [anon_sym_asm] = ACTIONS(2458),
    [anon_sym___asm__] = ACTIONS(2458),
    [anon_sym___asm] = ACTIONS(2458),
    [sym_number_literal] = ACTIONS(2460),
    [anon_sym_L_SQUOTE] = ACTIONS(2460),
    [anon_sym_u_SQUOTE] = ACTIONS(2460),
    [anon_sym_U_SQUOTE] = ACTIONS(2460),
    [anon_sym_u8_SQUOTE] = ACTIONS(2460),
    [anon_sym_SQUOTE] = ACTIONS(2460),
    [anon_sym_AT] = ACTIONS(2458),
    [anon_sym_DQUOTE] = ACTIONS(2460),
    [anon_sym_L_DQUOTE] = ACTIONS(2460),
    [anon_sym_u_DQUOTE] = ACTIONS(2460),
    [anon_sym_U_DQUOTE] = ACTIONS(2460),
    [anon_sym_u8_DQUOTE] = ACTIONS(2460),
    [sym_true] = ACTIONS(2458),
    [sym_false] = ACTIONS(2458),
    [anon_sym_NULL] = ACTIONS(2458),
    [anon_sym_nullptr] = ACTIONS(2458),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2458),
    [anon_sym___typeof] = ACTIONS(2458),
    [anon_sym_typeof] = ACTIONS(2458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_API_AVAILABLE] = ACTIONS(2458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_API_DEPRECATED] = ACTIONS(2458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2458),
    [anon_sym___deprecated_msg] = ACTIONS(2458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2458),
    [anon_sym_ATtry] = ACTIONS(2460),
    [anon_sym___try] = ACTIONS(2458),
    [anon_sym_ATthrow] = ACTIONS(2460),
    [anon_sym_ATselector] = ACTIONS(2460),
    [anon_sym_ATavailable] = ACTIONS(2460),
    [anon_sym___builtin_available] = ACTIONS(2458),
    [anon_sym_va_arg] = ACTIONS(2458),
    [anon_sym_ATencode] = ACTIONS(2460),
    [anon_sym_ATsynchronized] = ACTIONS(2460),
    [anon_sym_BOOL] = ACTIONS(2458),
    [anon_sym_IMP] = ACTIONS(2458),
    [anon_sym_SEL] = ACTIONS(2458),
    [anon_sym_Class] = ACTIONS(2458),
    [anon_sym_id] = ACTIONS(2458),
  },
  [STATE(1156)] = {
    [sym_identifier] = ACTIONS(2542),
    [anon_sym_LPAREN2] = ACTIONS(2544),
    [anon_sym_BANG] = ACTIONS(2544),
    [anon_sym_TILDE] = ACTIONS(2544),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_PLUS] = ACTIONS(2542),
    [anon_sym_STAR] = ACTIONS(2544),
    [anon_sym_CARET] = ACTIONS(2544),
    [anon_sym_AMP] = ACTIONS(2544),
    [anon_sym_SEMI] = ACTIONS(2544),
    [anon_sym___extension__] = ACTIONS(2542),
    [anon_sym_typedef] = ACTIONS(2542),
    [anon_sym_extern] = ACTIONS(2542),
    [anon_sym___attribute__] = ACTIONS(2542),
    [anon_sym___attribute] = ACTIONS(2542),
    [anon_sym_noreturn] = ACTIONS(2542),
    [anon_sym_LBRACK] = ACTIONS(2544),
    [anon_sym___declspec] = ACTIONS(2542),
    [anon_sym_LBRACE] = ACTIONS(2544),
    [anon_sym_signed] = ACTIONS(2542),
    [anon_sym_unsigned] = ACTIONS(2542),
    [anon_sym_long] = ACTIONS(2542),
    [anon_sym_short] = ACTIONS(2542),
    [anon_sym_static] = ACTIONS(2542),
    [anon_sym_ATautoreleasepool] = ACTIONS(2544),
    [anon_sym_auto] = ACTIONS(2542),
    [anon_sym_register] = ACTIONS(2542),
    [anon_sym_inline] = ACTIONS(2542),
    [anon_sym___inline] = ACTIONS(2542),
    [anon_sym___inline__] = ACTIONS(2542),
    [anon_sym___forceinline] = ACTIONS(2542),
    [anon_sym_thread_local] = ACTIONS(2542),
    [anon_sym___thread] = ACTIONS(2542),
    [anon_sym_CG_EXTERN] = ACTIONS(2542),
    [anon_sym_CG_INLINE] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2542),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2542),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2542),
    [anon_sym_IBOutlet] = ACTIONS(2542),
    [anon_sym_IBInspectable] = ACTIONS(2542),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2542),
    [anon_sym_NS_INLINE] = ACTIONS(2542),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2542),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2542),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2542),
    [anon_sym_const] = ACTIONS(2542),
    [anon_sym_constexpr] = ACTIONS(2542),
    [anon_sym_volatile] = ACTIONS(2542),
    [anon_sym_restrict] = ACTIONS(2542),
    [anon_sym___restrict__] = ACTIONS(2542),
    [anon_sym__Atomic] = ACTIONS(2542),
    [anon_sym__Noreturn] = ACTIONS(2542),
    [anon_sym__Nonnull] = ACTIONS(2542),
    [anon_sym_nullable] = ACTIONS(2542),
    [anon_sym__Complex] = ACTIONS(2542),
    [anon_sym__Nullable] = ACTIONS(2542),
    [anon_sym__Nullable_result] = ACTIONS(2542),
    [anon_sym__Null_unspecified] = ACTIONS(2542),
    [anon_sym___autoreleasing] = ACTIONS(2542),
    [anon_sym___block] = ACTIONS(2542),
    [anon_sym___bridge] = ACTIONS(2542),
    [anon_sym___bridge_retained] = ACTIONS(2542),
    [anon_sym___bridge_transfer] = ACTIONS(2542),
    [anon_sym___complex] = ACTIONS(2542),
    [anon_sym___const] = ACTIONS(2542),
    [anon_sym___imag] = ACTIONS(2542),
    [anon_sym___kindof] = ACTIONS(2542),
    [anon_sym___nonnull] = ACTIONS(2542),
    [anon_sym___nullable] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2542),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2542),
    [anon_sym___real] = ACTIONS(2542),
    [anon_sym___strong] = ACTIONS(2542),
    [anon_sym___unsafe_unretained] = ACTIONS(2542),
    [anon_sym___unused] = ACTIONS(2542),
    [anon_sym___weak] = ACTIONS(2542),
    [anon_sym_alignas] = ACTIONS(2542),
    [anon_sym__Alignas] = ACTIONS(2542),
    [sym_primitive_type] = ACTIONS(2542),
    [anon_sym_enum] = ACTIONS(2542),
    [anon_sym_struct] = ACTIONS(2542),
    [anon_sym_union] = ACTIONS(2542),
    [anon_sym_if] = ACTIONS(2542),
    [anon_sym_else] = ACTIONS(2542),
    [anon_sym_switch] = ACTIONS(2542),
    [anon_sym_while] = ACTIONS(2542),
    [anon_sym_do] = ACTIONS(2542),
    [anon_sym_for] = ACTIONS(2542),
    [anon_sym_in] = ACTIONS(2542),
    [anon_sym_return] = ACTIONS(2542),
    [anon_sym_break] = ACTIONS(2542),
    [anon_sym_continue] = ACTIONS(2542),
    [anon_sym_goto] = ACTIONS(2542),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_sizeof] = ACTIONS(2542),
    [anon_sym___alignof__] = ACTIONS(2542),
    [anon_sym___alignof] = ACTIONS(2542),
    [anon_sym__alignof] = ACTIONS(2542),
    [anon_sym_alignof] = ACTIONS(2542),
    [anon_sym__Alignof] = ACTIONS(2542),
    [anon_sym_offsetof] = ACTIONS(2542),
    [anon_sym__Generic] = ACTIONS(2542),
    [anon_sym_asm] = ACTIONS(2542),
    [anon_sym___asm__] = ACTIONS(2542),
    [anon_sym___asm] = ACTIONS(2542),
    [sym_number_literal] = ACTIONS(2544),
    [anon_sym_L_SQUOTE] = ACTIONS(2544),
    [anon_sym_u_SQUOTE] = ACTIONS(2544),
    [anon_sym_U_SQUOTE] = ACTIONS(2544),
    [anon_sym_u8_SQUOTE] = ACTIONS(2544),
    [anon_sym_SQUOTE] = ACTIONS(2544),
    [anon_sym_AT] = ACTIONS(2542),
    [anon_sym_DQUOTE] = ACTIONS(2544),
    [anon_sym_L_DQUOTE] = ACTIONS(2544),
    [anon_sym_u_DQUOTE] = ACTIONS(2544),
    [anon_sym_U_DQUOTE] = ACTIONS(2544),
    [anon_sym_u8_DQUOTE] = ACTIONS(2544),
    [sym_true] = ACTIONS(2542),
    [sym_false] = ACTIONS(2542),
    [anon_sym_NULL] = ACTIONS(2542),
    [anon_sym_nullptr] = ACTIONS(2542),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2542),
    [anon_sym___typeof] = ACTIONS(2542),
    [anon_sym_typeof] = ACTIONS(2542),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2542),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2542),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2542),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2542),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2542),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_API_AVAILABLE] = ACTIONS(2542),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_API_DEPRECATED] = ACTIONS(2542),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2542),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2542),
    [anon_sym___deprecated_msg] = ACTIONS(2542),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2542),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2542),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2542),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2542),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2542),
    [anon_sym_ATtry] = ACTIONS(2544),
    [anon_sym___try] = ACTIONS(2542),
    [anon_sym_ATthrow] = ACTIONS(2544),
    [anon_sym_ATselector] = ACTIONS(2544),
    [anon_sym_ATavailable] = ACTIONS(2544),
    [anon_sym___builtin_available] = ACTIONS(2542),
    [anon_sym_va_arg] = ACTIONS(2542),
    [anon_sym_ATencode] = ACTIONS(2544),
    [anon_sym_ATsynchronized] = ACTIONS(2544),
    [anon_sym_BOOL] = ACTIONS(2542),
    [anon_sym_IMP] = ACTIONS(2542),
    [anon_sym_SEL] = ACTIONS(2542),
    [anon_sym_Class] = ACTIONS(2542),
    [anon_sym_id] = ACTIONS(2542),
  },
  [STATE(1157)] = {
    [sym_identifier] = ACTIONS(2550),
    [anon_sym_LPAREN2] = ACTIONS(2552),
    [anon_sym_BANG] = ACTIONS(2552),
    [anon_sym_TILDE] = ACTIONS(2552),
    [anon_sym_DASH] = ACTIONS(2550),
    [anon_sym_PLUS] = ACTIONS(2550),
    [anon_sym_STAR] = ACTIONS(2552),
    [anon_sym_CARET] = ACTIONS(2552),
    [anon_sym_AMP] = ACTIONS(2552),
    [anon_sym_SEMI] = ACTIONS(2552),
    [anon_sym___extension__] = ACTIONS(2550),
    [anon_sym_typedef] = ACTIONS(2550),
    [anon_sym_extern] = ACTIONS(2550),
    [anon_sym___attribute__] = ACTIONS(2550),
    [anon_sym___attribute] = ACTIONS(2550),
    [anon_sym_noreturn] = ACTIONS(2550),
    [anon_sym_LBRACK] = ACTIONS(2552),
    [anon_sym___declspec] = ACTIONS(2550),
    [anon_sym_LBRACE] = ACTIONS(2552),
    [anon_sym_signed] = ACTIONS(2550),
    [anon_sym_unsigned] = ACTIONS(2550),
    [anon_sym_long] = ACTIONS(2550),
    [anon_sym_short] = ACTIONS(2550),
    [anon_sym_static] = ACTIONS(2550),
    [anon_sym_ATautoreleasepool] = ACTIONS(2552),
    [anon_sym_auto] = ACTIONS(2550),
    [anon_sym_register] = ACTIONS(2550),
    [anon_sym_inline] = ACTIONS(2550),
    [anon_sym___inline] = ACTIONS(2550),
    [anon_sym___inline__] = ACTIONS(2550),
    [anon_sym___forceinline] = ACTIONS(2550),
    [anon_sym_thread_local] = ACTIONS(2550),
    [anon_sym___thread] = ACTIONS(2550),
    [anon_sym_CG_EXTERN] = ACTIONS(2550),
    [anon_sym_CG_INLINE] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2550),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2550),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2550),
    [anon_sym_IBOutlet] = ACTIONS(2550),
    [anon_sym_IBInspectable] = ACTIONS(2550),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2550),
    [anon_sym_NS_INLINE] = ACTIONS(2550),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2550),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2550),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2550),
    [anon_sym_const] = ACTIONS(2550),
    [anon_sym_constexpr] = ACTIONS(2550),
    [anon_sym_volatile] = ACTIONS(2550),
    [anon_sym_restrict] = ACTIONS(2550),
    [anon_sym___restrict__] = ACTIONS(2550),
    [anon_sym__Atomic] = ACTIONS(2550),
    [anon_sym__Noreturn] = ACTIONS(2550),
    [anon_sym__Nonnull] = ACTIONS(2550),
    [anon_sym_nullable] = ACTIONS(2550),
    [anon_sym__Complex] = ACTIONS(2550),
    [anon_sym__Nullable] = ACTIONS(2550),
    [anon_sym__Nullable_result] = ACTIONS(2550),
    [anon_sym__Null_unspecified] = ACTIONS(2550),
    [anon_sym___autoreleasing] = ACTIONS(2550),
    [anon_sym___block] = ACTIONS(2550),
    [anon_sym___bridge] = ACTIONS(2550),
    [anon_sym___bridge_retained] = ACTIONS(2550),
    [anon_sym___bridge_transfer] = ACTIONS(2550),
    [anon_sym___complex] = ACTIONS(2550),
    [anon_sym___const] = ACTIONS(2550),
    [anon_sym___imag] = ACTIONS(2550),
    [anon_sym___kindof] = ACTIONS(2550),
    [anon_sym___nonnull] = ACTIONS(2550),
    [anon_sym___nullable] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2550),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2550),
    [anon_sym___real] = ACTIONS(2550),
    [anon_sym___strong] = ACTIONS(2550),
    [anon_sym___unsafe_unretained] = ACTIONS(2550),
    [anon_sym___unused] = ACTIONS(2550),
    [anon_sym___weak] = ACTIONS(2550),
    [anon_sym_alignas] = ACTIONS(2550),
    [anon_sym__Alignas] = ACTIONS(2550),
    [sym_primitive_type] = ACTIONS(2550),
    [anon_sym_enum] = ACTIONS(2550),
    [anon_sym_struct] = ACTIONS(2550),
    [anon_sym_union] = ACTIONS(2550),
    [anon_sym_if] = ACTIONS(2550),
    [anon_sym_else] = ACTIONS(2550),
    [anon_sym_switch] = ACTIONS(2550),
    [anon_sym_while] = ACTIONS(2550),
    [anon_sym_do] = ACTIONS(2550),
    [anon_sym_for] = ACTIONS(2550),
    [anon_sym_in] = ACTIONS(2550),
    [anon_sym_return] = ACTIONS(2550),
    [anon_sym_break] = ACTIONS(2550),
    [anon_sym_continue] = ACTIONS(2550),
    [anon_sym_goto] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2552),
    [anon_sym_PLUS_PLUS] = ACTIONS(2552),
    [anon_sym_sizeof] = ACTIONS(2550),
    [anon_sym___alignof__] = ACTIONS(2550),
    [anon_sym___alignof] = ACTIONS(2550),
    [anon_sym__alignof] = ACTIONS(2550),
    [anon_sym_alignof] = ACTIONS(2550),
    [anon_sym__Alignof] = ACTIONS(2550),
    [anon_sym_offsetof] = ACTIONS(2550),
    [anon_sym__Generic] = ACTIONS(2550),
    [anon_sym_asm] = ACTIONS(2550),
    [anon_sym___asm__] = ACTIONS(2550),
    [anon_sym___asm] = ACTIONS(2550),
    [sym_number_literal] = ACTIONS(2552),
    [anon_sym_L_SQUOTE] = ACTIONS(2552),
    [anon_sym_u_SQUOTE] = ACTIONS(2552),
    [anon_sym_U_SQUOTE] = ACTIONS(2552),
    [anon_sym_u8_SQUOTE] = ACTIONS(2552),
    [anon_sym_SQUOTE] = ACTIONS(2552),
    [anon_sym_AT] = ACTIONS(2550),
    [anon_sym_DQUOTE] = ACTIONS(2552),
    [anon_sym_L_DQUOTE] = ACTIONS(2552),
    [anon_sym_u_DQUOTE] = ACTIONS(2552),
    [anon_sym_U_DQUOTE] = ACTIONS(2552),
    [anon_sym_u8_DQUOTE] = ACTIONS(2552),
    [sym_true] = ACTIONS(2550),
    [sym_false] = ACTIONS(2550),
    [anon_sym_NULL] = ACTIONS(2550),
    [anon_sym_nullptr] = ACTIONS(2550),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2550),
    [anon_sym___typeof] = ACTIONS(2550),
    [anon_sym_typeof] = ACTIONS(2550),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2550),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2550),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2550),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2550),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2550),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_API_AVAILABLE] = ACTIONS(2550),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_API_DEPRECATED] = ACTIONS(2550),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2550),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2550),
    [anon_sym___deprecated_msg] = ACTIONS(2550),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2550),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2550),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2550),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2550),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2550),
    [anon_sym_ATtry] = ACTIONS(2552),
    [anon_sym___try] = ACTIONS(2550),
    [anon_sym_ATthrow] = ACTIONS(2552),
    [anon_sym_ATselector] = ACTIONS(2552),
    [anon_sym_ATavailable] = ACTIONS(2552),
    [anon_sym___builtin_available] = ACTIONS(2550),
    [anon_sym_va_arg] = ACTIONS(2550),
    [anon_sym_ATencode] = ACTIONS(2552),
    [anon_sym_ATsynchronized] = ACTIONS(2552),
    [anon_sym_BOOL] = ACTIONS(2550),
    [anon_sym_IMP] = ACTIONS(2550),
    [anon_sym_SEL] = ACTIONS(2550),
    [anon_sym_Class] = ACTIONS(2550),
    [anon_sym_id] = ACTIONS(2550),
  },
  [STATE(1158)] = {
    [sym_identifier] = ACTIONS(2412),
    [anon_sym_LPAREN2] = ACTIONS(2414),
    [anon_sym_BANG] = ACTIONS(2414),
    [anon_sym_TILDE] = ACTIONS(2414),
    [anon_sym_DASH] = ACTIONS(2412),
    [anon_sym_PLUS] = ACTIONS(2412),
    [anon_sym_STAR] = ACTIONS(2414),
    [anon_sym_CARET] = ACTIONS(2414),
    [anon_sym_AMP] = ACTIONS(2414),
    [anon_sym_SEMI] = ACTIONS(2414),
    [anon_sym___extension__] = ACTIONS(2412),
    [anon_sym_typedef] = ACTIONS(2412),
    [anon_sym_extern] = ACTIONS(2412),
    [anon_sym___attribute__] = ACTIONS(2412),
    [anon_sym___attribute] = ACTIONS(2412),
    [anon_sym_noreturn] = ACTIONS(2412),
    [anon_sym_LBRACK] = ACTIONS(2414),
    [anon_sym___declspec] = ACTIONS(2412),
    [anon_sym_LBRACE] = ACTIONS(2414),
    [anon_sym_signed] = ACTIONS(2412),
    [anon_sym_unsigned] = ACTIONS(2412),
    [anon_sym_long] = ACTIONS(2412),
    [anon_sym_short] = ACTIONS(2412),
    [anon_sym_static] = ACTIONS(2412),
    [anon_sym_ATautoreleasepool] = ACTIONS(2414),
    [anon_sym_auto] = ACTIONS(2412),
    [anon_sym_register] = ACTIONS(2412),
    [anon_sym_inline] = ACTIONS(2412),
    [anon_sym___inline] = ACTIONS(2412),
    [anon_sym___inline__] = ACTIONS(2412),
    [anon_sym___forceinline] = ACTIONS(2412),
    [anon_sym_thread_local] = ACTIONS(2412),
    [anon_sym___thread] = ACTIONS(2412),
    [anon_sym_CG_EXTERN] = ACTIONS(2412),
    [anon_sym_CG_INLINE] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2412),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2412),
    [anon_sym_IBOutlet] = ACTIONS(2412),
    [anon_sym_IBInspectable] = ACTIONS(2412),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2412),
    [anon_sym_NS_INLINE] = ACTIONS(2412),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2412),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2412),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2412),
    [anon_sym_const] = ACTIONS(2412),
    [anon_sym_constexpr] = ACTIONS(2412),
    [anon_sym_volatile] = ACTIONS(2412),
    [anon_sym_restrict] = ACTIONS(2412),
    [anon_sym___restrict__] = ACTIONS(2412),
    [anon_sym__Atomic] = ACTIONS(2412),
    [anon_sym__Noreturn] = ACTIONS(2412),
    [anon_sym__Nonnull] = ACTIONS(2412),
    [anon_sym_nullable] = ACTIONS(2412),
    [anon_sym__Complex] = ACTIONS(2412),
    [anon_sym__Nullable] = ACTIONS(2412),
    [anon_sym__Nullable_result] = ACTIONS(2412),
    [anon_sym__Null_unspecified] = ACTIONS(2412),
    [anon_sym___autoreleasing] = ACTIONS(2412),
    [anon_sym___block] = ACTIONS(2412),
    [anon_sym___bridge] = ACTIONS(2412),
    [anon_sym___bridge_retained] = ACTIONS(2412),
    [anon_sym___bridge_transfer] = ACTIONS(2412),
    [anon_sym___complex] = ACTIONS(2412),
    [anon_sym___const] = ACTIONS(2412),
    [anon_sym___imag] = ACTIONS(2412),
    [anon_sym___kindof] = ACTIONS(2412),
    [anon_sym___nonnull] = ACTIONS(2412),
    [anon_sym___nullable] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2412),
    [anon_sym___real] = ACTIONS(2412),
    [anon_sym___strong] = ACTIONS(2412),
    [anon_sym___unsafe_unretained] = ACTIONS(2412),
    [anon_sym___unused] = ACTIONS(2412),
    [anon_sym___weak] = ACTIONS(2412),
    [anon_sym_alignas] = ACTIONS(2412),
    [anon_sym__Alignas] = ACTIONS(2412),
    [sym_primitive_type] = ACTIONS(2412),
    [anon_sym_enum] = ACTIONS(2412),
    [anon_sym_struct] = ACTIONS(2412),
    [anon_sym_union] = ACTIONS(2412),
    [anon_sym_if] = ACTIONS(2412),
    [anon_sym_else] = ACTIONS(2412),
    [anon_sym_switch] = ACTIONS(2412),
    [anon_sym_while] = ACTIONS(2412),
    [anon_sym_do] = ACTIONS(2412),
    [anon_sym_for] = ACTIONS(2412),
    [anon_sym_in] = ACTIONS(2412),
    [anon_sym_return] = ACTIONS(2412),
    [anon_sym_break] = ACTIONS(2412),
    [anon_sym_continue] = ACTIONS(2412),
    [anon_sym_goto] = ACTIONS(2412),
    [anon_sym_DASH_DASH] = ACTIONS(2414),
    [anon_sym_PLUS_PLUS] = ACTIONS(2414),
    [anon_sym_sizeof] = ACTIONS(2412),
    [anon_sym___alignof__] = ACTIONS(2412),
    [anon_sym___alignof] = ACTIONS(2412),
    [anon_sym__alignof] = ACTIONS(2412),
    [anon_sym_alignof] = ACTIONS(2412),
    [anon_sym__Alignof] = ACTIONS(2412),
    [anon_sym_offsetof] = ACTIONS(2412),
    [anon_sym__Generic] = ACTIONS(2412),
    [anon_sym_asm] = ACTIONS(2412),
    [anon_sym___asm__] = ACTIONS(2412),
    [anon_sym___asm] = ACTIONS(2412),
    [sym_number_literal] = ACTIONS(2414),
    [anon_sym_L_SQUOTE] = ACTIONS(2414),
    [anon_sym_u_SQUOTE] = ACTIONS(2414),
    [anon_sym_U_SQUOTE] = ACTIONS(2414),
    [anon_sym_u8_SQUOTE] = ACTIONS(2414),
    [anon_sym_SQUOTE] = ACTIONS(2414),
    [anon_sym_AT] = ACTIONS(2412),
    [anon_sym_DQUOTE] = ACTIONS(2414),
    [anon_sym_L_DQUOTE] = ACTIONS(2414),
    [anon_sym_u_DQUOTE] = ACTIONS(2414),
    [anon_sym_U_DQUOTE] = ACTIONS(2414),
    [anon_sym_u8_DQUOTE] = ACTIONS(2414),
    [sym_true] = ACTIONS(2412),
    [sym_false] = ACTIONS(2412),
    [anon_sym_NULL] = ACTIONS(2412),
    [anon_sym_nullptr] = ACTIONS(2412),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2412),
    [anon_sym___typeof] = ACTIONS(2412),
    [anon_sym_typeof] = ACTIONS(2412),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2412),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2412),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2412),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_API_AVAILABLE] = ACTIONS(2412),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_API_DEPRECATED] = ACTIONS(2412),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym___deprecated_msg] = ACTIONS(2412),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2412),
    [anon_sym_ATtry] = ACTIONS(2414),
    [anon_sym___try] = ACTIONS(2412),
    [anon_sym_ATthrow] = ACTIONS(2414),
    [anon_sym_ATselector] = ACTIONS(2414),
    [anon_sym_ATavailable] = ACTIONS(2414),
    [anon_sym___builtin_available] = ACTIONS(2412),
    [anon_sym_va_arg] = ACTIONS(2412),
    [anon_sym_ATencode] = ACTIONS(2414),
    [anon_sym_ATsynchronized] = ACTIONS(2414),
    [anon_sym_BOOL] = ACTIONS(2412),
    [anon_sym_IMP] = ACTIONS(2412),
    [anon_sym_SEL] = ACTIONS(2412),
    [anon_sym_Class] = ACTIONS(2412),
    [anon_sym_id] = ACTIONS(2412),
  },
  [STATE(1159)] = {
    [sym_identifier] = ACTIONS(2454),
    [anon_sym_LPAREN2] = ACTIONS(2456),
    [anon_sym_BANG] = ACTIONS(2456),
    [anon_sym_TILDE] = ACTIONS(2456),
    [anon_sym_DASH] = ACTIONS(2454),
    [anon_sym_PLUS] = ACTIONS(2454),
    [anon_sym_STAR] = ACTIONS(2456),
    [anon_sym_CARET] = ACTIONS(2456),
    [anon_sym_AMP] = ACTIONS(2456),
    [anon_sym_SEMI] = ACTIONS(2456),
    [anon_sym___extension__] = ACTIONS(2454),
    [anon_sym_typedef] = ACTIONS(2454),
    [anon_sym_extern] = ACTIONS(2454),
    [anon_sym___attribute__] = ACTIONS(2454),
    [anon_sym___attribute] = ACTIONS(2454),
    [anon_sym_noreturn] = ACTIONS(2454),
    [anon_sym_LBRACK] = ACTIONS(2456),
    [anon_sym___declspec] = ACTIONS(2454),
    [anon_sym_LBRACE] = ACTIONS(2456),
    [anon_sym_signed] = ACTIONS(2454),
    [anon_sym_unsigned] = ACTIONS(2454),
    [anon_sym_long] = ACTIONS(2454),
    [anon_sym_short] = ACTIONS(2454),
    [anon_sym_static] = ACTIONS(2454),
    [anon_sym_ATautoreleasepool] = ACTIONS(2456),
    [anon_sym_auto] = ACTIONS(2454),
    [anon_sym_register] = ACTIONS(2454),
    [anon_sym_inline] = ACTIONS(2454),
    [anon_sym___inline] = ACTIONS(2454),
    [anon_sym___inline__] = ACTIONS(2454),
    [anon_sym___forceinline] = ACTIONS(2454),
    [anon_sym_thread_local] = ACTIONS(2454),
    [anon_sym___thread] = ACTIONS(2454),
    [anon_sym_CG_EXTERN] = ACTIONS(2454),
    [anon_sym_CG_INLINE] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2454),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2454),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2454),
    [anon_sym_IBOutlet] = ACTIONS(2454),
    [anon_sym_IBInspectable] = ACTIONS(2454),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2454),
    [anon_sym_NS_INLINE] = ACTIONS(2454),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2454),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2454),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2454),
    [anon_sym_const] = ACTIONS(2454),
    [anon_sym_constexpr] = ACTIONS(2454),
    [anon_sym_volatile] = ACTIONS(2454),
    [anon_sym_restrict] = ACTIONS(2454),
    [anon_sym___restrict__] = ACTIONS(2454),
    [anon_sym__Atomic] = ACTIONS(2454),
    [anon_sym__Noreturn] = ACTIONS(2454),
    [anon_sym__Nonnull] = ACTIONS(2454),
    [anon_sym_nullable] = ACTIONS(2454),
    [anon_sym__Complex] = ACTIONS(2454),
    [anon_sym__Nullable] = ACTIONS(2454),
    [anon_sym__Nullable_result] = ACTIONS(2454),
    [anon_sym__Null_unspecified] = ACTIONS(2454),
    [anon_sym___autoreleasing] = ACTIONS(2454),
    [anon_sym___block] = ACTIONS(2454),
    [anon_sym___bridge] = ACTIONS(2454),
    [anon_sym___bridge_retained] = ACTIONS(2454),
    [anon_sym___bridge_transfer] = ACTIONS(2454),
    [anon_sym___complex] = ACTIONS(2454),
    [anon_sym___const] = ACTIONS(2454),
    [anon_sym___imag] = ACTIONS(2454),
    [anon_sym___kindof] = ACTIONS(2454),
    [anon_sym___nonnull] = ACTIONS(2454),
    [anon_sym___nullable] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2454),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2454),
    [anon_sym___real] = ACTIONS(2454),
    [anon_sym___strong] = ACTIONS(2454),
    [anon_sym___unsafe_unretained] = ACTIONS(2454),
    [anon_sym___unused] = ACTIONS(2454),
    [anon_sym___weak] = ACTIONS(2454),
    [anon_sym_alignas] = ACTIONS(2454),
    [anon_sym__Alignas] = ACTIONS(2454),
    [sym_primitive_type] = ACTIONS(2454),
    [anon_sym_enum] = ACTIONS(2454),
    [anon_sym_struct] = ACTIONS(2454),
    [anon_sym_union] = ACTIONS(2454),
    [anon_sym_if] = ACTIONS(2454),
    [anon_sym_else] = ACTIONS(2454),
    [anon_sym_switch] = ACTIONS(2454),
    [anon_sym_while] = ACTIONS(2454),
    [anon_sym_do] = ACTIONS(2454),
    [anon_sym_for] = ACTIONS(2454),
    [anon_sym_in] = ACTIONS(2454),
    [anon_sym_return] = ACTIONS(2454),
    [anon_sym_break] = ACTIONS(2454),
    [anon_sym_continue] = ACTIONS(2454),
    [anon_sym_goto] = ACTIONS(2454),
    [anon_sym_DASH_DASH] = ACTIONS(2456),
    [anon_sym_PLUS_PLUS] = ACTIONS(2456),
    [anon_sym_sizeof] = ACTIONS(2454),
    [anon_sym___alignof__] = ACTIONS(2454),
    [anon_sym___alignof] = ACTIONS(2454),
    [anon_sym__alignof] = ACTIONS(2454),
    [anon_sym_alignof] = ACTIONS(2454),
    [anon_sym__Alignof] = ACTIONS(2454),
    [anon_sym_offsetof] = ACTIONS(2454),
    [anon_sym__Generic] = ACTIONS(2454),
    [anon_sym_asm] = ACTIONS(2454),
    [anon_sym___asm__] = ACTIONS(2454),
    [anon_sym___asm] = ACTIONS(2454),
    [sym_number_literal] = ACTIONS(2456),
    [anon_sym_L_SQUOTE] = ACTIONS(2456),
    [anon_sym_u_SQUOTE] = ACTIONS(2456),
    [anon_sym_U_SQUOTE] = ACTIONS(2456),
    [anon_sym_u8_SQUOTE] = ACTIONS(2456),
    [anon_sym_SQUOTE] = ACTIONS(2456),
    [anon_sym_AT] = ACTIONS(2454),
    [anon_sym_DQUOTE] = ACTIONS(2456),
    [anon_sym_L_DQUOTE] = ACTIONS(2456),
    [anon_sym_u_DQUOTE] = ACTIONS(2456),
    [anon_sym_U_DQUOTE] = ACTIONS(2456),
    [anon_sym_u8_DQUOTE] = ACTIONS(2456),
    [sym_true] = ACTIONS(2454),
    [sym_false] = ACTIONS(2454),
    [anon_sym_NULL] = ACTIONS(2454),
    [anon_sym_nullptr] = ACTIONS(2454),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2454),
    [anon_sym___typeof] = ACTIONS(2454),
    [anon_sym_typeof] = ACTIONS(2454),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2454),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2454),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2454),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2454),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2454),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_API_AVAILABLE] = ACTIONS(2454),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_API_DEPRECATED] = ACTIONS(2454),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2454),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2454),
    [anon_sym___deprecated_msg] = ACTIONS(2454),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2454),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2454),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2454),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2454),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2454),
    [anon_sym_ATtry] = ACTIONS(2456),
    [anon_sym___try] = ACTIONS(2454),
    [anon_sym_ATthrow] = ACTIONS(2456),
    [anon_sym_ATselector] = ACTIONS(2456),
    [anon_sym_ATavailable] = ACTIONS(2456),
    [anon_sym___builtin_available] = ACTIONS(2454),
    [anon_sym_va_arg] = ACTIONS(2454),
    [anon_sym_ATencode] = ACTIONS(2456),
    [anon_sym_ATsynchronized] = ACTIONS(2456),
    [anon_sym_BOOL] = ACTIONS(2454),
    [anon_sym_IMP] = ACTIONS(2454),
    [anon_sym_SEL] = ACTIONS(2454),
    [anon_sym_Class] = ACTIONS(2454),
    [anon_sym_id] = ACTIONS(2454),
  },
  [STATE(1160)] = {
    [sym_identifier] = ACTIONS(2462),
    [anon_sym_LPAREN2] = ACTIONS(2464),
    [anon_sym_BANG] = ACTIONS(2464),
    [anon_sym_TILDE] = ACTIONS(2464),
    [anon_sym_DASH] = ACTIONS(2462),
    [anon_sym_PLUS] = ACTIONS(2462),
    [anon_sym_STAR] = ACTIONS(2464),
    [anon_sym_CARET] = ACTIONS(2464),
    [anon_sym_AMP] = ACTIONS(2464),
    [anon_sym_SEMI] = ACTIONS(2464),
    [anon_sym___extension__] = ACTIONS(2462),
    [anon_sym_typedef] = ACTIONS(2462),
    [anon_sym_extern] = ACTIONS(2462),
    [anon_sym___attribute__] = ACTIONS(2462),
    [anon_sym___attribute] = ACTIONS(2462),
    [anon_sym_noreturn] = ACTIONS(2462),
    [anon_sym_LBRACK] = ACTIONS(2464),
    [anon_sym___declspec] = ACTIONS(2462),
    [anon_sym_LBRACE] = ACTIONS(2464),
    [anon_sym_signed] = ACTIONS(2462),
    [anon_sym_unsigned] = ACTIONS(2462),
    [anon_sym_long] = ACTIONS(2462),
    [anon_sym_short] = ACTIONS(2462),
    [anon_sym_static] = ACTIONS(2462),
    [anon_sym_ATautoreleasepool] = ACTIONS(2464),
    [anon_sym_auto] = ACTIONS(2462),
    [anon_sym_register] = ACTIONS(2462),
    [anon_sym_inline] = ACTIONS(2462),
    [anon_sym___inline] = ACTIONS(2462),
    [anon_sym___inline__] = ACTIONS(2462),
    [anon_sym___forceinline] = ACTIONS(2462),
    [anon_sym_thread_local] = ACTIONS(2462),
    [anon_sym___thread] = ACTIONS(2462),
    [anon_sym_CG_EXTERN] = ACTIONS(2462),
    [anon_sym_CG_INLINE] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2462),
    [anon_sym_IBOutlet] = ACTIONS(2462),
    [anon_sym_IBInspectable] = ACTIONS(2462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2462),
    [anon_sym_NS_INLINE] = ACTIONS(2462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2462),
    [anon_sym_const] = ACTIONS(2462),
    [anon_sym_constexpr] = ACTIONS(2462),
    [anon_sym_volatile] = ACTIONS(2462),
    [anon_sym_restrict] = ACTIONS(2462),
    [anon_sym___restrict__] = ACTIONS(2462),
    [anon_sym__Atomic] = ACTIONS(2462),
    [anon_sym__Noreturn] = ACTIONS(2462),
    [anon_sym__Nonnull] = ACTIONS(2462),
    [anon_sym_nullable] = ACTIONS(2462),
    [anon_sym__Complex] = ACTIONS(2462),
    [anon_sym__Nullable] = ACTIONS(2462),
    [anon_sym__Nullable_result] = ACTIONS(2462),
    [anon_sym__Null_unspecified] = ACTIONS(2462),
    [anon_sym___autoreleasing] = ACTIONS(2462),
    [anon_sym___block] = ACTIONS(2462),
    [anon_sym___bridge] = ACTIONS(2462),
    [anon_sym___bridge_retained] = ACTIONS(2462),
    [anon_sym___bridge_transfer] = ACTIONS(2462),
    [anon_sym___complex] = ACTIONS(2462),
    [anon_sym___const] = ACTIONS(2462),
    [anon_sym___imag] = ACTIONS(2462),
    [anon_sym___kindof] = ACTIONS(2462),
    [anon_sym___nonnull] = ACTIONS(2462),
    [anon_sym___nullable] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2462),
    [anon_sym___real] = ACTIONS(2462),
    [anon_sym___strong] = ACTIONS(2462),
    [anon_sym___unsafe_unretained] = ACTIONS(2462),
    [anon_sym___unused] = ACTIONS(2462),
    [anon_sym___weak] = ACTIONS(2462),
    [anon_sym_alignas] = ACTIONS(2462),
    [anon_sym__Alignas] = ACTIONS(2462),
    [sym_primitive_type] = ACTIONS(2462),
    [anon_sym_enum] = ACTIONS(2462),
    [anon_sym_struct] = ACTIONS(2462),
    [anon_sym_union] = ACTIONS(2462),
    [anon_sym_if] = ACTIONS(2462),
    [anon_sym_else] = ACTIONS(2462),
    [anon_sym_switch] = ACTIONS(2462),
    [anon_sym_while] = ACTIONS(2462),
    [anon_sym_do] = ACTIONS(2462),
    [anon_sym_for] = ACTIONS(2462),
    [anon_sym_in] = ACTIONS(2462),
    [anon_sym_return] = ACTIONS(2462),
    [anon_sym_break] = ACTIONS(2462),
    [anon_sym_continue] = ACTIONS(2462),
    [anon_sym_goto] = ACTIONS(2462),
    [anon_sym_DASH_DASH] = ACTIONS(2464),
    [anon_sym_PLUS_PLUS] = ACTIONS(2464),
    [anon_sym_sizeof] = ACTIONS(2462),
    [anon_sym___alignof__] = ACTIONS(2462),
    [anon_sym___alignof] = ACTIONS(2462),
    [anon_sym__alignof] = ACTIONS(2462),
    [anon_sym_alignof] = ACTIONS(2462),
    [anon_sym__Alignof] = ACTIONS(2462),
    [anon_sym_offsetof] = ACTIONS(2462),
    [anon_sym__Generic] = ACTIONS(2462),
    [anon_sym_asm] = ACTIONS(2462),
    [anon_sym___asm__] = ACTIONS(2462),
    [anon_sym___asm] = ACTIONS(2462),
    [sym_number_literal] = ACTIONS(2464),
    [anon_sym_L_SQUOTE] = ACTIONS(2464),
    [anon_sym_u_SQUOTE] = ACTIONS(2464),
    [anon_sym_U_SQUOTE] = ACTIONS(2464),
    [anon_sym_u8_SQUOTE] = ACTIONS(2464),
    [anon_sym_SQUOTE] = ACTIONS(2464),
    [anon_sym_AT] = ACTIONS(2462),
    [anon_sym_DQUOTE] = ACTIONS(2464),
    [anon_sym_L_DQUOTE] = ACTIONS(2464),
    [anon_sym_u_DQUOTE] = ACTIONS(2464),
    [anon_sym_U_DQUOTE] = ACTIONS(2464),
    [anon_sym_u8_DQUOTE] = ACTIONS(2464),
    [sym_true] = ACTIONS(2462),
    [sym_false] = ACTIONS(2462),
    [anon_sym_NULL] = ACTIONS(2462),
    [anon_sym_nullptr] = ACTIONS(2462),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2462),
    [anon_sym___typeof] = ACTIONS(2462),
    [anon_sym_typeof] = ACTIONS(2462),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2462),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_API_AVAILABLE] = ACTIONS(2462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_API_DEPRECATED] = ACTIONS(2462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2462),
    [anon_sym___deprecated_msg] = ACTIONS(2462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2462),
    [anon_sym_ATtry] = ACTIONS(2464),
    [anon_sym___try] = ACTIONS(2462),
    [anon_sym_ATthrow] = ACTIONS(2464),
    [anon_sym_ATselector] = ACTIONS(2464),
    [anon_sym_ATavailable] = ACTIONS(2464),
    [anon_sym___builtin_available] = ACTIONS(2462),
    [anon_sym_va_arg] = ACTIONS(2462),
    [anon_sym_ATencode] = ACTIONS(2464),
    [anon_sym_ATsynchronized] = ACTIONS(2464),
    [anon_sym_BOOL] = ACTIONS(2462),
    [anon_sym_IMP] = ACTIONS(2462),
    [anon_sym_SEL] = ACTIONS(2462),
    [anon_sym_Class] = ACTIONS(2462),
    [anon_sym_id] = ACTIONS(2462),
  },
  [STATE(1161)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1162)] = {
    [sym_identifier] = ACTIONS(2408),
    [anon_sym_LPAREN2] = ACTIONS(2410),
    [anon_sym_BANG] = ACTIONS(2410),
    [anon_sym_TILDE] = ACTIONS(2410),
    [anon_sym_DASH] = ACTIONS(2408),
    [anon_sym_PLUS] = ACTIONS(2408),
    [anon_sym_STAR] = ACTIONS(2410),
    [anon_sym_CARET] = ACTIONS(2410),
    [anon_sym_AMP] = ACTIONS(2410),
    [anon_sym_SEMI] = ACTIONS(2410),
    [anon_sym___extension__] = ACTIONS(2408),
    [anon_sym_typedef] = ACTIONS(2408),
    [anon_sym_extern] = ACTIONS(2408),
    [anon_sym___attribute__] = ACTIONS(2408),
    [anon_sym___attribute] = ACTIONS(2408),
    [anon_sym_noreturn] = ACTIONS(2408),
    [anon_sym_LBRACK] = ACTIONS(2410),
    [anon_sym___declspec] = ACTIONS(2408),
    [anon_sym_LBRACE] = ACTIONS(2410),
    [anon_sym_signed] = ACTIONS(2408),
    [anon_sym_unsigned] = ACTIONS(2408),
    [anon_sym_long] = ACTIONS(2408),
    [anon_sym_short] = ACTIONS(2408),
    [anon_sym_static] = ACTIONS(2408),
    [anon_sym_ATautoreleasepool] = ACTIONS(2410),
    [anon_sym_auto] = ACTIONS(2408),
    [anon_sym_register] = ACTIONS(2408),
    [anon_sym_inline] = ACTIONS(2408),
    [anon_sym___inline] = ACTIONS(2408),
    [anon_sym___inline__] = ACTIONS(2408),
    [anon_sym___forceinline] = ACTIONS(2408),
    [anon_sym_thread_local] = ACTIONS(2408),
    [anon_sym___thread] = ACTIONS(2408),
    [anon_sym_CG_EXTERN] = ACTIONS(2408),
    [anon_sym_CG_INLINE] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2408),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2408),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2408),
    [anon_sym_IBOutlet] = ACTIONS(2408),
    [anon_sym_IBInspectable] = ACTIONS(2408),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2408),
    [anon_sym_NS_INLINE] = ACTIONS(2408),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2408),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2408),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2408),
    [anon_sym_const] = ACTIONS(2408),
    [anon_sym_constexpr] = ACTIONS(2408),
    [anon_sym_volatile] = ACTIONS(2408),
    [anon_sym_restrict] = ACTIONS(2408),
    [anon_sym___restrict__] = ACTIONS(2408),
    [anon_sym__Atomic] = ACTIONS(2408),
    [anon_sym__Noreturn] = ACTIONS(2408),
    [anon_sym__Nonnull] = ACTIONS(2408),
    [anon_sym_nullable] = ACTIONS(2408),
    [anon_sym__Complex] = ACTIONS(2408),
    [anon_sym__Nullable] = ACTIONS(2408),
    [anon_sym__Nullable_result] = ACTIONS(2408),
    [anon_sym__Null_unspecified] = ACTIONS(2408),
    [anon_sym___autoreleasing] = ACTIONS(2408),
    [anon_sym___block] = ACTIONS(2408),
    [anon_sym___bridge] = ACTIONS(2408),
    [anon_sym___bridge_retained] = ACTIONS(2408),
    [anon_sym___bridge_transfer] = ACTIONS(2408),
    [anon_sym___complex] = ACTIONS(2408),
    [anon_sym___const] = ACTIONS(2408),
    [anon_sym___imag] = ACTIONS(2408),
    [anon_sym___kindof] = ACTIONS(2408),
    [anon_sym___nonnull] = ACTIONS(2408),
    [anon_sym___nullable] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2408),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2408),
    [anon_sym___real] = ACTIONS(2408),
    [anon_sym___strong] = ACTIONS(2408),
    [anon_sym___unsafe_unretained] = ACTIONS(2408),
    [anon_sym___unused] = ACTIONS(2408),
    [anon_sym___weak] = ACTIONS(2408),
    [anon_sym_alignas] = ACTIONS(2408),
    [anon_sym__Alignas] = ACTIONS(2408),
    [sym_primitive_type] = ACTIONS(2408),
    [anon_sym_enum] = ACTIONS(2408),
    [anon_sym_struct] = ACTIONS(2408),
    [anon_sym_union] = ACTIONS(2408),
    [anon_sym_if] = ACTIONS(2408),
    [anon_sym_else] = ACTIONS(2408),
    [anon_sym_switch] = ACTIONS(2408),
    [anon_sym_while] = ACTIONS(2408),
    [anon_sym_do] = ACTIONS(2408),
    [anon_sym_for] = ACTIONS(2408),
    [anon_sym_in] = ACTIONS(2408),
    [anon_sym_return] = ACTIONS(2408),
    [anon_sym_break] = ACTIONS(2408),
    [anon_sym_continue] = ACTIONS(2408),
    [anon_sym_goto] = ACTIONS(2408),
    [anon_sym_DASH_DASH] = ACTIONS(2410),
    [anon_sym_PLUS_PLUS] = ACTIONS(2410),
    [anon_sym_sizeof] = ACTIONS(2408),
    [anon_sym___alignof__] = ACTIONS(2408),
    [anon_sym___alignof] = ACTIONS(2408),
    [anon_sym__alignof] = ACTIONS(2408),
    [anon_sym_alignof] = ACTIONS(2408),
    [anon_sym__Alignof] = ACTIONS(2408),
    [anon_sym_offsetof] = ACTIONS(2408),
    [anon_sym__Generic] = ACTIONS(2408),
    [anon_sym_asm] = ACTIONS(2408),
    [anon_sym___asm__] = ACTIONS(2408),
    [anon_sym___asm] = ACTIONS(2408),
    [sym_number_literal] = ACTIONS(2410),
    [anon_sym_L_SQUOTE] = ACTIONS(2410),
    [anon_sym_u_SQUOTE] = ACTIONS(2410),
    [anon_sym_U_SQUOTE] = ACTIONS(2410),
    [anon_sym_u8_SQUOTE] = ACTIONS(2410),
    [anon_sym_SQUOTE] = ACTIONS(2410),
    [anon_sym_AT] = ACTIONS(2408),
    [anon_sym_DQUOTE] = ACTIONS(2410),
    [anon_sym_L_DQUOTE] = ACTIONS(2410),
    [anon_sym_u_DQUOTE] = ACTIONS(2410),
    [anon_sym_U_DQUOTE] = ACTIONS(2410),
    [anon_sym_u8_DQUOTE] = ACTIONS(2410),
    [sym_true] = ACTIONS(2408),
    [sym_false] = ACTIONS(2408),
    [anon_sym_NULL] = ACTIONS(2408),
    [anon_sym_nullptr] = ACTIONS(2408),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2408),
    [anon_sym___typeof] = ACTIONS(2408),
    [anon_sym_typeof] = ACTIONS(2408),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2408),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2408),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2408),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2408),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2408),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_API_AVAILABLE] = ACTIONS(2408),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_API_DEPRECATED] = ACTIONS(2408),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2408),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2408),
    [anon_sym___deprecated_msg] = ACTIONS(2408),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2408),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2408),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2408),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2408),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2408),
    [anon_sym_ATtry] = ACTIONS(2410),
    [anon_sym___try] = ACTIONS(2408),
    [anon_sym_ATthrow] = ACTIONS(2410),
    [anon_sym_ATselector] = ACTIONS(2410),
    [anon_sym_ATavailable] = ACTIONS(2410),
    [anon_sym___builtin_available] = ACTIONS(2408),
    [anon_sym_va_arg] = ACTIONS(2408),
    [anon_sym_ATencode] = ACTIONS(2410),
    [anon_sym_ATsynchronized] = ACTIONS(2410),
    [anon_sym_BOOL] = ACTIONS(2408),
    [anon_sym_IMP] = ACTIONS(2408),
    [anon_sym_SEL] = ACTIONS(2408),
    [anon_sym_Class] = ACTIONS(2408),
    [anon_sym_id] = ACTIONS(2408),
  },
  [STATE(1163)] = {
    [sym_identifier] = ACTIONS(2494),
    [anon_sym_LPAREN2] = ACTIONS(2496),
    [anon_sym_BANG] = ACTIONS(2496),
    [anon_sym_TILDE] = ACTIONS(2496),
    [anon_sym_DASH] = ACTIONS(2494),
    [anon_sym_PLUS] = ACTIONS(2494),
    [anon_sym_STAR] = ACTIONS(2496),
    [anon_sym_CARET] = ACTIONS(2496),
    [anon_sym_AMP] = ACTIONS(2496),
    [anon_sym_SEMI] = ACTIONS(2496),
    [anon_sym___extension__] = ACTIONS(2494),
    [anon_sym_typedef] = ACTIONS(2494),
    [anon_sym_extern] = ACTIONS(2494),
    [anon_sym___attribute__] = ACTIONS(2494),
    [anon_sym___attribute] = ACTIONS(2494),
    [anon_sym_noreturn] = ACTIONS(2494),
    [anon_sym_LBRACK] = ACTIONS(2496),
    [anon_sym___declspec] = ACTIONS(2494),
    [anon_sym_LBRACE] = ACTIONS(2496),
    [anon_sym_signed] = ACTIONS(2494),
    [anon_sym_unsigned] = ACTIONS(2494),
    [anon_sym_long] = ACTIONS(2494),
    [anon_sym_short] = ACTIONS(2494),
    [anon_sym_static] = ACTIONS(2494),
    [anon_sym_ATautoreleasepool] = ACTIONS(2496),
    [anon_sym_auto] = ACTIONS(2494),
    [anon_sym_register] = ACTIONS(2494),
    [anon_sym_inline] = ACTIONS(2494),
    [anon_sym___inline] = ACTIONS(2494),
    [anon_sym___inline__] = ACTIONS(2494),
    [anon_sym___forceinline] = ACTIONS(2494),
    [anon_sym_thread_local] = ACTIONS(2494),
    [anon_sym___thread] = ACTIONS(2494),
    [anon_sym_CG_EXTERN] = ACTIONS(2494),
    [anon_sym_CG_INLINE] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2494),
    [anon_sym_IBOutlet] = ACTIONS(2494),
    [anon_sym_IBInspectable] = ACTIONS(2494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2494),
    [anon_sym_NS_INLINE] = ACTIONS(2494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2494),
    [anon_sym_const] = ACTIONS(2494),
    [anon_sym_constexpr] = ACTIONS(2494),
    [anon_sym_volatile] = ACTIONS(2494),
    [anon_sym_restrict] = ACTIONS(2494),
    [anon_sym___restrict__] = ACTIONS(2494),
    [anon_sym__Atomic] = ACTIONS(2494),
    [anon_sym__Noreturn] = ACTIONS(2494),
    [anon_sym__Nonnull] = ACTIONS(2494),
    [anon_sym_nullable] = ACTIONS(2494),
    [anon_sym__Complex] = ACTIONS(2494),
    [anon_sym__Nullable] = ACTIONS(2494),
    [anon_sym__Nullable_result] = ACTIONS(2494),
    [anon_sym__Null_unspecified] = ACTIONS(2494),
    [anon_sym___autoreleasing] = ACTIONS(2494),
    [anon_sym___block] = ACTIONS(2494),
    [anon_sym___bridge] = ACTIONS(2494),
    [anon_sym___bridge_retained] = ACTIONS(2494),
    [anon_sym___bridge_transfer] = ACTIONS(2494),
    [anon_sym___complex] = ACTIONS(2494),
    [anon_sym___const] = ACTIONS(2494),
    [anon_sym___imag] = ACTIONS(2494),
    [anon_sym___kindof] = ACTIONS(2494),
    [anon_sym___nonnull] = ACTIONS(2494),
    [anon_sym___nullable] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2494),
    [anon_sym___real] = ACTIONS(2494),
    [anon_sym___strong] = ACTIONS(2494),
    [anon_sym___unsafe_unretained] = ACTIONS(2494),
    [anon_sym___unused] = ACTIONS(2494),
    [anon_sym___weak] = ACTIONS(2494),
    [anon_sym_alignas] = ACTIONS(2494),
    [anon_sym__Alignas] = ACTIONS(2494),
    [sym_primitive_type] = ACTIONS(2494),
    [anon_sym_enum] = ACTIONS(2494),
    [anon_sym_struct] = ACTIONS(2494),
    [anon_sym_union] = ACTIONS(2494),
    [anon_sym_if] = ACTIONS(2494),
    [anon_sym_else] = ACTIONS(2494),
    [anon_sym_switch] = ACTIONS(2494),
    [anon_sym_while] = ACTIONS(2494),
    [anon_sym_do] = ACTIONS(2494),
    [anon_sym_for] = ACTIONS(2494),
    [anon_sym_in] = ACTIONS(2494),
    [anon_sym_return] = ACTIONS(2494),
    [anon_sym_break] = ACTIONS(2494),
    [anon_sym_continue] = ACTIONS(2494),
    [anon_sym_goto] = ACTIONS(2494),
    [anon_sym_DASH_DASH] = ACTIONS(2496),
    [anon_sym_PLUS_PLUS] = ACTIONS(2496),
    [anon_sym_sizeof] = ACTIONS(2494),
    [anon_sym___alignof__] = ACTIONS(2494),
    [anon_sym___alignof] = ACTIONS(2494),
    [anon_sym__alignof] = ACTIONS(2494),
    [anon_sym_alignof] = ACTIONS(2494),
    [anon_sym__Alignof] = ACTIONS(2494),
    [anon_sym_offsetof] = ACTIONS(2494),
    [anon_sym__Generic] = ACTIONS(2494),
    [anon_sym_asm] = ACTIONS(2494),
    [anon_sym___asm__] = ACTIONS(2494),
    [anon_sym___asm] = ACTIONS(2494),
    [sym_number_literal] = ACTIONS(2496),
    [anon_sym_L_SQUOTE] = ACTIONS(2496),
    [anon_sym_u_SQUOTE] = ACTIONS(2496),
    [anon_sym_U_SQUOTE] = ACTIONS(2496),
    [anon_sym_u8_SQUOTE] = ACTIONS(2496),
    [anon_sym_SQUOTE] = ACTIONS(2496),
    [anon_sym_AT] = ACTIONS(2494),
    [anon_sym_DQUOTE] = ACTIONS(2496),
    [anon_sym_L_DQUOTE] = ACTIONS(2496),
    [anon_sym_u_DQUOTE] = ACTIONS(2496),
    [anon_sym_U_DQUOTE] = ACTIONS(2496),
    [anon_sym_u8_DQUOTE] = ACTIONS(2496),
    [sym_true] = ACTIONS(2494),
    [sym_false] = ACTIONS(2494),
    [anon_sym_NULL] = ACTIONS(2494),
    [anon_sym_nullptr] = ACTIONS(2494),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2494),
    [anon_sym___typeof] = ACTIONS(2494),
    [anon_sym_typeof] = ACTIONS(2494),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_API_AVAILABLE] = ACTIONS(2494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_API_DEPRECATED] = ACTIONS(2494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2494),
    [anon_sym___deprecated_msg] = ACTIONS(2494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2494),
    [anon_sym_ATtry] = ACTIONS(2496),
    [anon_sym___try] = ACTIONS(2494),
    [anon_sym_ATthrow] = ACTIONS(2496),
    [anon_sym_ATselector] = ACTIONS(2496),
    [anon_sym_ATavailable] = ACTIONS(2496),
    [anon_sym___builtin_available] = ACTIONS(2494),
    [anon_sym_va_arg] = ACTIONS(2494),
    [anon_sym_ATencode] = ACTIONS(2496),
    [anon_sym_ATsynchronized] = ACTIONS(2496),
    [anon_sym_BOOL] = ACTIONS(2494),
    [anon_sym_IMP] = ACTIONS(2494),
    [anon_sym_SEL] = ACTIONS(2494),
    [anon_sym_Class] = ACTIONS(2494),
    [anon_sym_id] = ACTIONS(2494),
  },
  [STATE(1164)] = {
    [sym_preproc_def] = STATE(1221),
    [sym_preproc_call] = STATE(1221),
    [sym_preproc_ifdef] = STATE(1221),
    [sym_function_definition] = STATE(1221),
    [sym_declaration] = STATE(1221),
    [sym_type_definition] = STATE(1221),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1221),
    [sym_preproc_undef] = STATE(1221),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_protocol_reference_list] = STATE(1220),
    [sym_qualified_protocol_interface_declaration] = STATE(5136),
    [sym_property_declaration] = STATE(1221),
    [sym_method_declaration] = STATE(1221),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_protocol_forward_declaration_repeat1] = STATE(5504),
    [aux_sym_class_interface_repeat1] = STATE(1221),
    [aux_sym_protocol_declaration_repeat1] = STATE(5136),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [anon_sym_COMMA] = ACTIONS(3042),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(3050),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3052),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1165)] = {
    [sym_preproc_def] = STATE(1337),
    [sym_preproc_call] = STATE(1337),
    [sym_preproc_ifdef] = STATE(1337),
    [sym_function_definition] = STATE(1337),
    [sym_declaration] = STATE(1337),
    [sym_type_definition] = STATE(1337),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1337),
    [sym_preproc_undef] = STATE(1337),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym__type_params] = STATE(1192),
    [sym__class_interface_inheritance] = STATE(1232),
    [sym_parameterized_arguments] = STATE(1208),
    [sym_generic_arguments] = STATE(1970),
    [sym_instance_variables] = STATE(1426),
    [sym_property_declaration] = STATE(1337),
    [sym_method_declaration] = STATE(1337),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1337),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3058),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3060),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3062),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1166)] = {
    [sym_preproc_def] = STATE(1225),
    [sym_preproc_call] = STATE(1225),
    [sym_preproc_ifdef] = STATE(1225),
    [sym_function_definition] = STATE(1225),
    [sym_declaration] = STATE(1225),
    [sym_type_definition] = STATE(1225),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1225),
    [sym_preproc_undef] = STATE(1225),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_protocol_reference_list] = STATE(1226),
    [sym_qualified_protocol_interface_declaration] = STATE(5111),
    [sym_property_declaration] = STATE(1225),
    [sym_method_declaration] = STATE(1225),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_protocol_forward_declaration_repeat1] = STATE(5544),
    [aux_sym_class_interface_repeat1] = STATE(1225),
    [aux_sym_protocol_declaration_repeat1] = STATE(5111),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [anon_sym_COMMA] = ACTIONS(3042),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(3064),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3066),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1167)] = {
    [sym_preproc_def] = STATE(1382),
    [sym_preproc_call] = STATE(1382),
    [sym_preproc_ifdef] = STATE(1382),
    [sym_function_definition] = STATE(1382),
    [sym_declaration] = STATE(1382),
    [sym_type_definition] = STATE(1382),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1382),
    [sym_preproc_undef] = STATE(1382),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym__type_params] = STATE(1194),
    [sym__class_interface_inheritance] = STATE(1231),
    [sym_parameterized_arguments] = STATE(1209),
    [sym_generic_arguments] = STATE(1970),
    [sym_instance_variables] = STATE(1381),
    [sym_property_declaration] = STATE(1382),
    [sym_method_declaration] = STATE(1382),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1382),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3058),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3060),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3068),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1168)] = {
    [sym_preproc_def] = STATE(1216),
    [sym_preproc_call] = STATE(1216),
    [sym_preproc_ifdef] = STATE(1216),
    [sym_function_definition] = STATE(1216),
    [sym_declaration] = STATE(1216),
    [sym_type_definition] = STATE(1216),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1216),
    [sym_preproc_undef] = STATE(1216),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_protocol_reference_list] = STATE(1223),
    [sym_qualified_protocol_interface_declaration] = STATE(5134),
    [sym_property_declaration] = STATE(1216),
    [sym_method_declaration] = STATE(1216),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_protocol_forward_declaration_repeat1] = STATE(5659),
    [aux_sym_class_interface_repeat1] = STATE(1216),
    [aux_sym_protocol_declaration_repeat1] = STATE(5134),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [anon_sym_COMMA] = ACTIONS(3042),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(3070),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3072),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1169)] = {
    [sym_preproc_def] = STATE(1234),
    [sym_preproc_call] = STATE(1234),
    [sym_preproc_ifdef] = STATE(1234),
    [sym_function_definition] = STATE(1234),
    [sym_declaration] = STATE(1234),
    [sym_type_definition] = STATE(1234),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1234),
    [sym_preproc_undef] = STATE(1234),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_protocol_reference_list] = STATE(1255),
    [sym_qualified_protocol_interface_declaration] = STATE(5131),
    [sym_property_declaration] = STATE(1234),
    [sym_method_declaration] = STATE(1234),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_protocol_forward_declaration_repeat1] = STATE(5627),
    [aux_sym_class_interface_repeat1] = STATE(1234),
    [aux_sym_protocol_declaration_repeat1] = STATE(5131),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [anon_sym_COMMA] = ACTIONS(3042),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(3074),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3076),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1170)] = {
    [sym_preproc_def] = STATE(1242),
    [sym_preproc_call] = STATE(1242),
    [sym_preproc_ifdef] = STATE(1242),
    [sym_function_definition] = STATE(1242),
    [sym_declaration] = STATE(1242),
    [sym_type_definition] = STATE(1242),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1242),
    [sym_preproc_undef] = STATE(1242),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_protocol_reference_list] = STATE(1252),
    [sym_qualified_protocol_interface_declaration] = STATE(5129),
    [sym_property_declaration] = STATE(1242),
    [sym_method_declaration] = STATE(1242),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_protocol_forward_declaration_repeat1] = STATE(5382),
    [aux_sym_class_interface_repeat1] = STATE(1242),
    [aux_sym_protocol_declaration_repeat1] = STATE(5129),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [anon_sym_COMMA] = ACTIONS(3042),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(3078),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3080),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1171)] = {
    [sym_preproc_def] = STATE(1238),
    [sym_preproc_call] = STATE(1238),
    [sym_preproc_ifdef] = STATE(1238),
    [sym_function_definition] = STATE(1238),
    [sym_declaration] = STATE(1238),
    [sym_type_definition] = STATE(1238),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1238),
    [sym_preproc_undef] = STATE(1238),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_protocol_reference_list] = STATE(1239),
    [sym_qualified_protocol_interface_declaration] = STATE(5110),
    [sym_property_declaration] = STATE(1238),
    [sym_method_declaration] = STATE(1238),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_protocol_forward_declaration_repeat1] = STATE(5674),
    [aux_sym_class_interface_repeat1] = STATE(1238),
    [aux_sym_protocol_declaration_repeat1] = STATE(5110),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [anon_sym_COMMA] = ACTIONS(3042),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(3082),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3084),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1172)] = {
    [sym_preproc_def] = STATE(1425),
    [sym_preproc_call] = STATE(1425),
    [sym_preproc_ifdef] = STATE(1425),
    [sym_function_definition] = STATE(1425),
    [sym_declaration] = STATE(1425),
    [sym_type_definition] = STATE(1425),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1425),
    [sym_preproc_undef] = STATE(1425),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym__type_params] = STATE(1186),
    [sym__class_interface_inheritance] = STATE(1236),
    [sym_parameterized_arguments] = STATE(1203),
    [sym_generic_arguments] = STATE(1970),
    [sym_instance_variables] = STATE(1424),
    [sym_property_declaration] = STATE(1425),
    [sym_method_declaration] = STATE(1425),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1425),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3058),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3060),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3086),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1173)] = {
    [sym_preproc_def] = STATE(1358),
    [sym_preproc_call] = STATE(1358),
    [sym_preproc_ifdef] = STATE(1358),
    [sym_function_definition] = STATE(1358),
    [sym_declaration] = STATE(1358),
    [sym_type_definition] = STATE(1358),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1358),
    [sym_preproc_undef] = STATE(1358),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym__type_params] = STATE(1188),
    [sym__class_interface_inheritance] = STATE(1230),
    [sym_parameterized_arguments] = STATE(1207),
    [sym_generic_arguments] = STATE(1970),
    [sym_instance_variables] = STATE(1357),
    [sym_property_declaration] = STATE(1358),
    [sym_method_declaration] = STATE(1358),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1358),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3058),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3060),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3088),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1174)] = {
    [sym_preproc_def] = STATE(1248),
    [sym_preproc_call] = STATE(1248),
    [sym_preproc_ifdef] = STATE(1248),
    [sym_function_definition] = STATE(1248),
    [sym_declaration] = STATE(1248),
    [sym_type_definition] = STATE(1248),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1248),
    [sym_preproc_undef] = STATE(1248),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_protocol_reference_list] = STATE(1245),
    [sym_qualified_protocol_interface_declaration] = STATE(5128),
    [sym_property_declaration] = STATE(1248),
    [sym_method_declaration] = STATE(1248),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_protocol_forward_declaration_repeat1] = STATE(5557),
    [aux_sym_class_interface_repeat1] = STATE(1248),
    [aux_sym_protocol_declaration_repeat1] = STATE(5128),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [anon_sym_COMMA] = ACTIONS(3042),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(3090),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3092),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1175)] = {
    [sym_preproc_def] = STATE(1241),
    [sym_preproc_call] = STATE(1241),
    [sym_preproc_ifdef] = STATE(1241),
    [sym_function_definition] = STATE(1241),
    [sym_declaration] = STATE(1241),
    [sym_type_definition] = STATE(1241),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1241),
    [sym_preproc_undef] = STATE(1241),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_protocol_reference_list] = STATE(1237),
    [sym_qualified_protocol_interface_declaration] = STATE(5106),
    [sym_property_declaration] = STATE(1241),
    [sym_method_declaration] = STATE(1241),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_protocol_forward_declaration_repeat1] = STATE(5594),
    [aux_sym_class_interface_repeat1] = STATE(1241),
    [aux_sym_protocol_declaration_repeat1] = STATE(5106),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [anon_sym_COMMA] = ACTIONS(3042),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(3094),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3096),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1176)] = {
    [sym_preproc_def] = STATE(1247),
    [sym_preproc_call] = STATE(1247),
    [sym_preproc_ifdef] = STATE(1247),
    [sym_function_definition] = STATE(1247),
    [sym_declaration] = STATE(1247),
    [sym_type_definition] = STATE(1247),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1247),
    [sym_preproc_undef] = STATE(1247),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_protocol_reference_list] = STATE(1246),
    [sym_qualified_protocol_interface_declaration] = STATE(5150),
    [sym_property_declaration] = STATE(1247),
    [sym_method_declaration] = STATE(1247),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_protocol_forward_declaration_repeat1] = STATE(5481),
    [aux_sym_class_interface_repeat1] = STATE(1247),
    [aux_sym_protocol_declaration_repeat1] = STATE(5150),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [anon_sym_COMMA] = ACTIONS(3042),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(3098),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3100),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1177)] = {
    [sym_preproc_def] = STATE(1253),
    [sym_preproc_call] = STATE(1253),
    [sym_preproc_ifdef] = STATE(1253),
    [sym_function_definition] = STATE(1253),
    [sym_declaration] = STATE(1253),
    [sym_type_definition] = STATE(1253),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1253),
    [sym_preproc_undef] = STATE(1253),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_protocol_reference_list] = STATE(1233),
    [sym_qualified_protocol_interface_declaration] = STATE(5158),
    [sym_property_declaration] = STATE(1253),
    [sym_method_declaration] = STATE(1253),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_protocol_forward_declaration_repeat1] = STATE(5321),
    [aux_sym_class_interface_repeat1] = STATE(1253),
    [aux_sym_protocol_declaration_repeat1] = STATE(5158),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [anon_sym_COMMA] = ACTIONS(3042),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(3102),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3104),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1178)] = {
    [sym_preproc_def] = STATE(1355),
    [sym_preproc_call] = STATE(1355),
    [sym_preproc_ifdef] = STATE(1355),
    [sym_function_definition] = STATE(1355),
    [sym_declaration] = STATE(1355),
    [sym_type_definition] = STATE(1355),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1355),
    [sym_preproc_undef] = STATE(1355),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym__type_params] = STATE(1184),
    [sym__class_interface_inheritance] = STATE(1218),
    [sym_parameterized_arguments] = STATE(1199),
    [sym_generic_arguments] = STATE(1970),
    [sym_instance_variables] = STATE(1354),
    [sym_property_declaration] = STATE(1355),
    [sym_method_declaration] = STATE(1355),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1355),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3058),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3060),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3106),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1179)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym__class_implementation_inheritance] = STATE(1200),
    [sym_instance_variables] = STATE(1306),
    [sym_implementation_definition] = STATE(1256),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1256),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3108),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3018),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3110),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1180)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym__class_implementation_inheritance] = STATE(1202),
    [sym_instance_variables] = STATE(1273),
    [sym_implementation_definition] = STATE(1274),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1274),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3108),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3018),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3112),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1181)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym__class_implementation_inheritance] = STATE(1201),
    [sym_instance_variables] = STATE(1314),
    [sym_implementation_definition] = STATE(1284),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1284),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3108),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3018),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3114),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1182)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym__class_implementation_inheritance] = STATE(1211),
    [sym_instance_variables] = STATE(1296),
    [sym_implementation_definition] = STATE(1290),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1290),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3108),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3018),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3116),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1183)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym__class_implementation_inheritance] = STATE(1206),
    [sym_instance_variables] = STATE(1302),
    [sym_implementation_definition] = STATE(1303),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1303),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3108),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3018),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3118),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1184)] = {
    [sym_preproc_def] = STATE(1360),
    [sym_preproc_call] = STATE(1360),
    [sym_preproc_ifdef] = STATE(1360),
    [sym_function_definition] = STATE(1360),
    [sym_declaration] = STATE(1360),
    [sym_type_definition] = STATE(1360),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1360),
    [sym_preproc_undef] = STATE(1360),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym__class_interface_inheritance] = STATE(1222),
    [sym_parameterized_arguments] = STATE(1325),
    [sym_instance_variables] = STATE(1359),
    [sym_property_declaration] = STATE(1360),
    [sym_method_declaration] = STATE(1360),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1360),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3120),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3060),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3122),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1185)] = {
    [sym_preproc_def] = STATE(2235),
    [sym_preproc_function_def] = STATE(2235),
    [sym_preproc_call] = STATE(2235),
    [sym_preproc_ifdef] = STATE(2235),
    [sym_function_definition] = STATE(2235),
    [sym_declaration] = STATE(2235),
    [sym_type_definition] = STATE(2235),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3235),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(1952),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2235),
    [sym_preproc_else_in_implementation_definition] = STATE(6330),
    [sym_preproc_elif_in_implementation_definition] = STATE(6330),
    [sym_preproc_undef] = STATE(2235),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1191),
    [sym_property_implementation] = STATE(2235),
    [sym_method_definition] = STATE(2235),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1191),
    [sym_identifier] = ACTIONS(3124),
    [aux_sym_preproc_def_token1] = ACTIONS(3126),
    [aux_sym_preproc_if_token1] = ACTIONS(3128),
    [aux_sym_preproc_if_token2] = ACTIONS(3130),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3134),
    [aux_sym_preproc_elif_token1] = ACTIONS(3136),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3140),
    [anon_sym_PLUS] = ACTIONS(3140),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATsynthesize] = ACTIONS(3148),
    [anon_sym_ATdynamic] = ACTIONS(3150),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1186)] = {
    [sym_preproc_def] = STATE(1428),
    [sym_preproc_call] = STATE(1428),
    [sym_preproc_ifdef] = STATE(1428),
    [sym_function_definition] = STATE(1428),
    [sym_declaration] = STATE(1428),
    [sym_type_definition] = STATE(1428),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1428),
    [sym_preproc_undef] = STATE(1428),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym__class_interface_inheritance] = STATE(1249),
    [sym_parameterized_arguments] = STATE(1328),
    [sym_instance_variables] = STATE(1427),
    [sym_property_declaration] = STATE(1428),
    [sym_method_declaration] = STATE(1428),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1428),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3120),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3060),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3152),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1187)] = {
    [sym_preproc_def] = STATE(2235),
    [sym_preproc_function_def] = STATE(2235),
    [sym_preproc_call] = STATE(2235),
    [sym_preproc_ifdef] = STATE(2235),
    [sym_function_definition] = STATE(2235),
    [sym_declaration] = STATE(2235),
    [sym_type_definition] = STATE(2235),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3235),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(1952),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2235),
    [sym_preproc_else_in_implementation_definition] = STATE(6704),
    [sym_preproc_elif_in_implementation_definition] = STATE(6704),
    [sym_preproc_undef] = STATE(2235),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1190),
    [sym_property_implementation] = STATE(2235),
    [sym_method_definition] = STATE(2235),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1190),
    [sym_identifier] = ACTIONS(3124),
    [aux_sym_preproc_def_token1] = ACTIONS(3126),
    [aux_sym_preproc_if_token1] = ACTIONS(3128),
    [aux_sym_preproc_if_token2] = ACTIONS(3154),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3134),
    [aux_sym_preproc_elif_token1] = ACTIONS(3136),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3140),
    [anon_sym_PLUS] = ACTIONS(3140),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATsynthesize] = ACTIONS(3148),
    [anon_sym_ATdynamic] = ACTIONS(3150),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1188)] = {
    [sym_preproc_def] = STATE(1376),
    [sym_preproc_call] = STATE(1376),
    [sym_preproc_ifdef] = STATE(1376),
    [sym_function_definition] = STATE(1376),
    [sym_declaration] = STATE(1376),
    [sym_type_definition] = STATE(1376),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1376),
    [sym_preproc_undef] = STATE(1376),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym__class_interface_inheritance] = STATE(1243),
    [sym_parameterized_arguments] = STATE(1316),
    [sym_instance_variables] = STATE(1375),
    [sym_property_declaration] = STATE(1376),
    [sym_method_declaration] = STATE(1376),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1376),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3120),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3060),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3156),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1189)] = {
    [sym_preproc_def] = STATE(2235),
    [sym_preproc_function_def] = STATE(2235),
    [sym_preproc_call] = STATE(2235),
    [sym_preproc_ifdef] = STATE(2235),
    [sym_function_definition] = STATE(2235),
    [sym_declaration] = STATE(2235),
    [sym_type_definition] = STATE(2235),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3235),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(1952),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2235),
    [sym_preproc_else_in_implementation_definition] = STATE(6190),
    [sym_preproc_elif_in_implementation_definition] = STATE(6190),
    [sym_preproc_undef] = STATE(2235),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1198),
    [sym_property_implementation] = STATE(2235),
    [sym_method_definition] = STATE(2235),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1198),
    [sym_identifier] = ACTIONS(3124),
    [aux_sym_preproc_def_token1] = ACTIONS(3126),
    [aux_sym_preproc_if_token1] = ACTIONS(3128),
    [aux_sym_preproc_if_token2] = ACTIONS(3158),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3134),
    [aux_sym_preproc_elif_token1] = ACTIONS(3136),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3140),
    [anon_sym_PLUS] = ACTIONS(3140),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATsynthesize] = ACTIONS(3148),
    [anon_sym_ATdynamic] = ACTIONS(3150),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1190)] = {
    [sym_preproc_def] = STATE(2235),
    [sym_preproc_function_def] = STATE(2235),
    [sym_preproc_call] = STATE(2235),
    [sym_preproc_ifdef] = STATE(2235),
    [sym_function_definition] = STATE(2235),
    [sym_declaration] = STATE(2235),
    [sym_type_definition] = STATE(2235),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3235),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(1952),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2235),
    [sym_preproc_else_in_implementation_definition] = STATE(6782),
    [sym_preproc_elif_in_implementation_definition] = STATE(6782),
    [sym_preproc_undef] = STATE(2235),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1198),
    [sym_property_implementation] = STATE(2235),
    [sym_method_definition] = STATE(2235),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1198),
    [sym_identifier] = ACTIONS(3124),
    [aux_sym_preproc_def_token1] = ACTIONS(3126),
    [aux_sym_preproc_if_token1] = ACTIONS(3128),
    [aux_sym_preproc_if_token2] = ACTIONS(3160),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3134),
    [aux_sym_preproc_elif_token1] = ACTIONS(3136),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3140),
    [anon_sym_PLUS] = ACTIONS(3140),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATsynthesize] = ACTIONS(3148),
    [anon_sym_ATdynamic] = ACTIONS(3150),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1191)] = {
    [sym_preproc_def] = STATE(2235),
    [sym_preproc_function_def] = STATE(2235),
    [sym_preproc_call] = STATE(2235),
    [sym_preproc_ifdef] = STATE(2235),
    [sym_function_definition] = STATE(2235),
    [sym_declaration] = STATE(2235),
    [sym_type_definition] = STATE(2235),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3235),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(1952),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2235),
    [sym_preproc_else_in_implementation_definition] = STATE(6448),
    [sym_preproc_elif_in_implementation_definition] = STATE(6448),
    [sym_preproc_undef] = STATE(2235),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1198),
    [sym_property_implementation] = STATE(2235),
    [sym_method_definition] = STATE(2235),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1198),
    [sym_identifier] = ACTIONS(3124),
    [aux_sym_preproc_def_token1] = ACTIONS(3126),
    [aux_sym_preproc_if_token1] = ACTIONS(3128),
    [aux_sym_preproc_if_token2] = ACTIONS(3162),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3134),
    [aux_sym_preproc_elif_token1] = ACTIONS(3136),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3140),
    [anon_sym_PLUS] = ACTIONS(3140),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATsynthesize] = ACTIONS(3148),
    [anon_sym_ATdynamic] = ACTIONS(3150),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1192)] = {
    [sym_preproc_def] = STATE(1396),
    [sym_preproc_call] = STATE(1396),
    [sym_preproc_ifdef] = STATE(1396),
    [sym_function_definition] = STATE(1396),
    [sym_declaration] = STATE(1396),
    [sym_type_definition] = STATE(1396),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1396),
    [sym_preproc_undef] = STATE(1396),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym__class_interface_inheritance] = STATE(1219),
    [sym_parameterized_arguments] = STATE(1332),
    [sym_instance_variables] = STATE(1395),
    [sym_property_declaration] = STATE(1396),
    [sym_method_declaration] = STATE(1396),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1396),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3120),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3060),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3164),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1193)] = {
    [sym_preproc_def] = STATE(2235),
    [sym_preproc_function_def] = STATE(2235),
    [sym_preproc_call] = STATE(2235),
    [sym_preproc_ifdef] = STATE(2235),
    [sym_function_definition] = STATE(2235),
    [sym_declaration] = STATE(2235),
    [sym_type_definition] = STATE(2235),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3235),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(1952),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2235),
    [sym_preproc_else_in_implementation_definition] = STATE(6636),
    [sym_preproc_elif_in_implementation_definition] = STATE(6636),
    [sym_preproc_undef] = STATE(2235),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1189),
    [sym_property_implementation] = STATE(2235),
    [sym_method_definition] = STATE(2235),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1189),
    [sym_identifier] = ACTIONS(3124),
    [aux_sym_preproc_def_token1] = ACTIONS(3126),
    [aux_sym_preproc_if_token1] = ACTIONS(3128),
    [aux_sym_preproc_if_token2] = ACTIONS(3166),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3134),
    [aux_sym_preproc_elif_token1] = ACTIONS(3136),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3140),
    [anon_sym_PLUS] = ACTIONS(3140),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATsynthesize] = ACTIONS(3148),
    [anon_sym_ATdynamic] = ACTIONS(3150),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1194)] = {
    [sym_preproc_def] = STATE(1390),
    [sym_preproc_call] = STATE(1390),
    [sym_preproc_ifdef] = STATE(1390),
    [sym_function_definition] = STATE(1390),
    [sym_declaration] = STATE(1390),
    [sym_type_definition] = STATE(1390),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1390),
    [sym_preproc_undef] = STATE(1390),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym__class_interface_inheritance] = STATE(1235),
    [sym_parameterized_arguments] = STATE(1318),
    [sym_instance_variables] = STATE(1389),
    [sym_property_declaration] = STATE(1390),
    [sym_method_declaration] = STATE(1390),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1390),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3120),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3060),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3168),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1195)] = {
    [sym_preproc_def] = STATE(2235),
    [sym_preproc_function_def] = STATE(2235),
    [sym_preproc_call] = STATE(2235),
    [sym_preproc_ifdef] = STATE(2235),
    [sym_function_definition] = STATE(2235),
    [sym_declaration] = STATE(2235),
    [sym_type_definition] = STATE(2235),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3235),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(1952),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2235),
    [sym_preproc_else_in_implementation_definition] = STATE(6271),
    [sym_preproc_elif_in_implementation_definition] = STATE(6271),
    [sym_preproc_undef] = STATE(2235),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1198),
    [sym_property_implementation] = STATE(2235),
    [sym_method_definition] = STATE(2235),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1198),
    [sym_identifier] = ACTIONS(3124),
    [aux_sym_preproc_def_token1] = ACTIONS(3126),
    [aux_sym_preproc_if_token1] = ACTIONS(3128),
    [aux_sym_preproc_if_token2] = ACTIONS(3170),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3134),
    [aux_sym_preproc_elif_token1] = ACTIONS(3136),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3140),
    [anon_sym_PLUS] = ACTIONS(3140),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATsynthesize] = ACTIONS(3148),
    [anon_sym_ATdynamic] = ACTIONS(3150),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1196)] = {
    [sym_preproc_def] = STATE(2235),
    [sym_preproc_function_def] = STATE(2235),
    [sym_preproc_call] = STATE(2235),
    [sym_preproc_ifdef] = STATE(2235),
    [sym_function_definition] = STATE(2235),
    [sym_declaration] = STATE(2235),
    [sym_type_definition] = STATE(2235),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3235),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(1952),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2235),
    [sym_preproc_else_in_implementation_definition] = STATE(6610),
    [sym_preproc_elif_in_implementation_definition] = STATE(6610),
    [sym_preproc_undef] = STATE(2235),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1195),
    [sym_property_implementation] = STATE(2235),
    [sym_method_definition] = STATE(2235),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1195),
    [sym_identifier] = ACTIONS(3124),
    [aux_sym_preproc_def_token1] = ACTIONS(3126),
    [aux_sym_preproc_if_token1] = ACTIONS(3128),
    [aux_sym_preproc_if_token2] = ACTIONS(3172),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3134),
    [aux_sym_preproc_elif_token1] = ACTIONS(3136),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3140),
    [anon_sym_PLUS] = ACTIONS(3140),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATsynthesize] = ACTIONS(3148),
    [anon_sym_ATdynamic] = ACTIONS(3150),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1197)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_instance_variables] = STATE(1285),
    [sym_implementation_definition] = STATE(1286),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1286),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3174),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1198)] = {
    [sym_preproc_def] = STATE(2235),
    [sym_preproc_function_def] = STATE(2235),
    [sym_preproc_call] = STATE(2235),
    [sym_preproc_ifdef] = STATE(2235),
    [sym_function_definition] = STATE(2235),
    [sym_declaration] = STATE(2235),
    [sym_type_definition] = STATE(2235),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3235),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(1952),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2235),
    [sym_preproc_undef] = STATE(2235),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1198),
    [sym_property_implementation] = STATE(2235),
    [sym_method_definition] = STATE(2235),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1198),
    [sym_identifier] = ACTIONS(3176),
    [aux_sym_preproc_def_token1] = ACTIONS(3179),
    [aux_sym_preproc_if_token1] = ACTIONS(3182),
    [aux_sym_preproc_if_token2] = ACTIONS(3185),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3187),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3187),
    [aux_sym_preproc_else_token1] = ACTIONS(3185),
    [aux_sym_preproc_elif_token1] = ACTIONS(3185),
    [sym_preproc_directive] = ACTIONS(3190),
    [anon_sym_DASH] = ACTIONS(3193),
    [anon_sym_PLUS] = ACTIONS(3193),
    [anon_sym___extension__] = ACTIONS(3196),
    [anon_sym_typedef] = ACTIONS(3199),
    [anon_sym_extern] = ACTIONS(3202),
    [anon_sym___attribute__] = ACTIONS(3205),
    [anon_sym___attribute] = ACTIONS(3205),
    [anon_sym_noreturn] = ACTIONS(3208),
    [anon_sym_LBRACK] = ACTIONS(3211),
    [anon_sym___declspec] = ACTIONS(3214),
    [anon_sym___cdecl] = ACTIONS(3217),
    [anon_sym___clrcall] = ACTIONS(3217),
    [anon_sym___stdcall] = ACTIONS(3217),
    [anon_sym___fastcall] = ACTIONS(3217),
    [anon_sym___thiscall] = ACTIONS(3217),
    [anon_sym___vectorcall] = ACTIONS(3217),
    [anon_sym_signed] = ACTIONS(3220),
    [anon_sym_unsigned] = ACTIONS(3220),
    [anon_sym_long] = ACTIONS(3220),
    [anon_sym_short] = ACTIONS(3220),
    [anon_sym_static] = ACTIONS(3202),
    [anon_sym_auto] = ACTIONS(3202),
    [anon_sym_register] = ACTIONS(3202),
    [anon_sym_inline] = ACTIONS(3202),
    [anon_sym___inline] = ACTIONS(3202),
    [anon_sym___inline__] = ACTIONS(3202),
    [anon_sym___forceinline] = ACTIONS(3202),
    [anon_sym_thread_local] = ACTIONS(3202),
    [anon_sym___thread] = ACTIONS(3202),
    [anon_sym_CG_EXTERN] = ACTIONS(3202),
    [anon_sym_CG_INLINE] = ACTIONS(3202),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3202),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3202),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3202),
    [anon_sym_IBOutlet] = ACTIONS(3202),
    [anon_sym_IBInspectable] = ACTIONS(3202),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3202),
    [anon_sym_NS_INLINE] = ACTIONS(3202),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3202),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3202),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3202),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3202),
    [anon_sym_const] = ACTIONS(3208),
    [anon_sym_constexpr] = ACTIONS(3208),
    [anon_sym_volatile] = ACTIONS(3208),
    [anon_sym_restrict] = ACTIONS(3208),
    [anon_sym___restrict__] = ACTIONS(3208),
    [anon_sym__Atomic] = ACTIONS(3208),
    [anon_sym__Noreturn] = ACTIONS(3208),
    [anon_sym__Nonnull] = ACTIONS(3208),
    [anon_sym_nullable] = ACTIONS(3208),
    [anon_sym__Complex] = ACTIONS(3208),
    [anon_sym__Nullable] = ACTIONS(3208),
    [anon_sym__Nullable_result] = ACTIONS(3208),
    [anon_sym__Null_unspecified] = ACTIONS(3208),
    [anon_sym___autoreleasing] = ACTIONS(3208),
    [anon_sym___block] = ACTIONS(3208),
    [anon_sym___bridge] = ACTIONS(3208),
    [anon_sym___bridge_retained] = ACTIONS(3208),
    [anon_sym___bridge_transfer] = ACTIONS(3208),
    [anon_sym___complex] = ACTIONS(3208),
    [anon_sym___const] = ACTIONS(3208),
    [anon_sym___imag] = ACTIONS(3208),
    [anon_sym___kindof] = ACTIONS(3208),
    [anon_sym___nonnull] = ACTIONS(3208),
    [anon_sym___nullable] = ACTIONS(3208),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3208),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3208),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3208),
    [anon_sym___real] = ACTIONS(3208),
    [anon_sym___strong] = ACTIONS(3208),
    [anon_sym___unsafe_unretained] = ACTIONS(3208),
    [anon_sym___unused] = ACTIONS(3208),
    [anon_sym___weak] = ACTIONS(3208),
    [anon_sym_alignas] = ACTIONS(3223),
    [anon_sym__Alignas] = ACTIONS(3223),
    [sym_primitive_type] = ACTIONS(3226),
    [anon_sym_enum] = ACTIONS(3229),
    [anon_sym_struct] = ACTIONS(3232),
    [anon_sym_union] = ACTIONS(3235),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3238),
    [anon_sym___typeof] = ACTIONS(3238),
    [anon_sym_typeof] = ACTIONS(3238),
    [aux_sym_preproc_undef_token1] = ACTIONS(3241),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3244),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3244),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3244),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3244),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3244),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3244),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3244),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3244),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3244),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3247),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3247),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3247),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3247),
    [anon_sym_API_AVAILABLE] = ACTIONS(3247),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3247),
    [anon_sym_API_DEPRECATED] = ACTIONS(3247),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3247),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3247),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3247),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3247),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3247),
    [anon_sym___deprecated_msg] = ACTIONS(3247),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3247),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3247),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3247),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3247),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3247),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3247),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3247),
    [anon_sym_ATsynthesize] = ACTIONS(3250),
    [anon_sym_ATdynamic] = ACTIONS(3253),
    [anon_sym_BOOL] = ACTIONS(3256),
    [anon_sym_IMP] = ACTIONS(3256),
    [anon_sym_SEL] = ACTIONS(3256),
    [anon_sym_Class] = ACTIONS(3256),
    [anon_sym_id] = ACTIONS(3256),
  },
  [STATE(1199)] = {
    [sym_preproc_def] = STATE(1360),
    [sym_preproc_call] = STATE(1360),
    [sym_preproc_ifdef] = STATE(1360),
    [sym_function_definition] = STATE(1360),
    [sym_declaration] = STATE(1360),
    [sym_type_definition] = STATE(1360),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1360),
    [sym_preproc_undef] = STATE(1360),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1359),
    [sym_property_declaration] = STATE(1360),
    [sym_method_declaration] = STATE(1360),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1360),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3259),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3259),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3259),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3122),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1200)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_instance_variables] = STATE(1308),
    [sym_implementation_definition] = STATE(1260),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1260),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3261),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1201)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_instance_variables] = STATE(1291),
    [sym_implementation_definition] = STATE(1292),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1292),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3263),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1202)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_instance_variables] = STATE(1277),
    [sym_implementation_definition] = STATE(1278),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1278),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3265),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1203)] = {
    [sym_preproc_def] = STATE(1428),
    [sym_preproc_call] = STATE(1428),
    [sym_preproc_ifdef] = STATE(1428),
    [sym_function_definition] = STATE(1428),
    [sym_declaration] = STATE(1428),
    [sym_type_definition] = STATE(1428),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1428),
    [sym_preproc_undef] = STATE(1428),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1427),
    [sym_property_declaration] = STATE(1428),
    [sym_method_declaration] = STATE(1428),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1428),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3259),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3259),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3259),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3152),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1204)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_instance_variables] = STATE(1275),
    [sym_implementation_definition] = STATE(1276),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1276),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3267),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1205)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_instance_variables] = STATE(1304),
    [sym_implementation_definition] = STATE(1305),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1305),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3269),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1206)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_instance_variables] = STATE(1311),
    [sym_implementation_definition] = STATE(1312),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1312),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3271),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1207)] = {
    [sym_preproc_def] = STATE(1376),
    [sym_preproc_call] = STATE(1376),
    [sym_preproc_ifdef] = STATE(1376),
    [sym_function_definition] = STATE(1376),
    [sym_declaration] = STATE(1376),
    [sym_type_definition] = STATE(1376),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1376),
    [sym_preproc_undef] = STATE(1376),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1375),
    [sym_property_declaration] = STATE(1376),
    [sym_method_declaration] = STATE(1376),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1376),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3259),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3259),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3259),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3156),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1208)] = {
    [sym_preproc_def] = STATE(1396),
    [sym_preproc_call] = STATE(1396),
    [sym_preproc_ifdef] = STATE(1396),
    [sym_function_definition] = STATE(1396),
    [sym_declaration] = STATE(1396),
    [sym_type_definition] = STATE(1396),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1396),
    [sym_preproc_undef] = STATE(1396),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1395),
    [sym_property_declaration] = STATE(1396),
    [sym_method_declaration] = STATE(1396),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1396),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3259),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3259),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3259),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3164),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1209)] = {
    [sym_preproc_def] = STATE(1390),
    [sym_preproc_call] = STATE(1390),
    [sym_preproc_ifdef] = STATE(1390),
    [sym_function_definition] = STATE(1390),
    [sym_declaration] = STATE(1390),
    [sym_type_definition] = STATE(1390),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1390),
    [sym_preproc_undef] = STATE(1390),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1389),
    [sym_property_declaration] = STATE(1390),
    [sym_method_declaration] = STATE(1390),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1390),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_LPAREN2] = ACTIONS(3259),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3259),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(3259),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3168),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1210)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_instance_variables] = STATE(1310),
    [sym_implementation_definition] = STATE(1281),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1281),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3273),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1211)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_instance_variables] = STATE(1294),
    [sym_implementation_definition] = STATE(1295),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1295),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3275),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1212)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_instance_variables] = STATE(1268),
    [sym_implementation_definition] = STATE(1300),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1300),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3277),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1213)] = {
    [sym_preproc_def] = STATE(1322),
    [sym_preproc_call] = STATE(1322),
    [sym_preproc_ifdef] = STATE(1322),
    [sym_function_definition] = STATE(1322),
    [sym_declaration] = STATE(1322),
    [sym_type_definition] = STATE(1322),
    [sym__declaration_modifiers] = STATE(2901),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2901),
    [sym_attribute_declaration] = STATE(2901),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2901),
    [sym_type_qualifier] = STATE(2901),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3228),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1322),
    [sym_preproc_else_in_interface_declaration] = STATE(6237),
    [sym_preproc_elif_in_interface_declaration] = STATE(6237),
    [sym_preproc_undef] = STATE(1322),
    [sym_availability_attribute_specifier] = STATE(2901),
    [sym_property_declaration] = STATE(1322),
    [sym_method_declaration] = STATE(1322),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1322),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3279),
    [aux_sym_preproc_if_token1] = ACTIONS(3281),
    [aux_sym_preproc_if_token2] = ACTIONS(3283),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3285),
    [aux_sym_preproc_elif_token1] = ACTIONS(3287),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3289),
    [anon_sym_PLUS] = ACTIONS(3289),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATproperty] = ACTIONS(3291),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1214)] = {
    [sym_preproc_def] = STATE(1322),
    [sym_preproc_call] = STATE(1322),
    [sym_preproc_ifdef] = STATE(1322),
    [sym_function_definition] = STATE(1322),
    [sym_declaration] = STATE(1322),
    [sym_type_definition] = STATE(1322),
    [sym__declaration_modifiers] = STATE(2901),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2901),
    [sym_attribute_declaration] = STATE(2901),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2901),
    [sym_type_qualifier] = STATE(2901),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3228),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1322),
    [sym_preproc_else_in_interface_declaration] = STATE(6246),
    [sym_preproc_elif_in_interface_declaration] = STATE(6246),
    [sym_preproc_undef] = STATE(1322),
    [sym_availability_attribute_specifier] = STATE(2901),
    [sym_property_declaration] = STATE(1322),
    [sym_method_declaration] = STATE(1322),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1322),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3279),
    [aux_sym_preproc_if_token1] = ACTIONS(3281),
    [aux_sym_preproc_if_token2] = ACTIONS(3293),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3285),
    [aux_sym_preproc_elif_token1] = ACTIONS(3287),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3289),
    [anon_sym_PLUS] = ACTIONS(3289),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATproperty] = ACTIONS(3291),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1215)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5114),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5114),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3295),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1216)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5153),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5153),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3297),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1217)] = {
    [sym_preproc_def] = STATE(1224),
    [sym_preproc_call] = STATE(1224),
    [sym_preproc_ifdef] = STATE(1224),
    [sym_function_definition] = STATE(1224),
    [sym_declaration] = STATE(1224),
    [sym_type_definition] = STATE(1224),
    [sym__declaration_modifiers] = STATE(2901),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2901),
    [sym_attribute_declaration] = STATE(2901),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2901),
    [sym_type_qualifier] = STATE(2901),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3228),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1224),
    [sym_preproc_else_in_interface_declaration] = STATE(6689),
    [sym_preproc_elif_in_interface_declaration] = STATE(6689),
    [sym_preproc_undef] = STATE(1224),
    [sym_availability_attribute_specifier] = STATE(2901),
    [sym_property_declaration] = STATE(1224),
    [sym_method_declaration] = STATE(1224),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1224),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3279),
    [aux_sym_preproc_if_token1] = ACTIONS(3281),
    [aux_sym_preproc_if_token2] = ACTIONS(3299),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3285),
    [aux_sym_preproc_elif_token1] = ACTIONS(3287),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3289),
    [anon_sym_PLUS] = ACTIONS(3289),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATproperty] = ACTIONS(3291),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1218)] = {
    [sym_preproc_def] = STATE(1362),
    [sym_preproc_call] = STATE(1362),
    [sym_preproc_ifdef] = STATE(1362),
    [sym_function_definition] = STATE(1362),
    [sym_declaration] = STATE(1362),
    [sym_type_definition] = STATE(1362),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1362),
    [sym_preproc_undef] = STATE(1362),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_parameterized_arguments] = STATE(1327),
    [sym_instance_variables] = STATE(1361),
    [sym_property_declaration] = STATE(1362),
    [sym_method_declaration] = STATE(1362),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1362),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3301),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1219)] = {
    [sym_preproc_def] = STATE(1417),
    [sym_preproc_call] = STATE(1417),
    [sym_preproc_ifdef] = STATE(1417),
    [sym_function_definition] = STATE(1417),
    [sym_declaration] = STATE(1417),
    [sym_type_definition] = STATE(1417),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1417),
    [sym_preproc_undef] = STATE(1417),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_parameterized_arguments] = STATE(1323),
    [sym_instance_variables] = STATE(1416),
    [sym_property_declaration] = STATE(1417),
    [sym_method_declaration] = STATE(1417),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1417),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3303),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1220)] = {
    [sym_preproc_def] = STATE(1225),
    [sym_preproc_call] = STATE(1225),
    [sym_preproc_ifdef] = STATE(1225),
    [sym_function_definition] = STATE(1225),
    [sym_declaration] = STATE(1225),
    [sym_type_definition] = STATE(1225),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1225),
    [sym_preproc_undef] = STATE(1225),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5111),
    [sym_property_declaration] = STATE(1225),
    [sym_method_declaration] = STATE(1225),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1225),
    [aux_sym_protocol_declaration_repeat1] = STATE(5111),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3066),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1221)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5111),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5111),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3066),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1222)] = {
    [sym_preproc_def] = STATE(1364),
    [sym_preproc_call] = STATE(1364),
    [sym_preproc_ifdef] = STATE(1364),
    [sym_function_definition] = STATE(1364),
    [sym_declaration] = STATE(1364),
    [sym_type_definition] = STATE(1364),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1364),
    [sym_preproc_undef] = STATE(1364),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_parameterized_arguments] = STATE(1324),
    [sym_instance_variables] = STATE(1363),
    [sym_property_declaration] = STATE(1364),
    [sym_method_declaration] = STATE(1364),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1364),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3305),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1223)] = {
    [sym_preproc_def] = STATE(1229),
    [sym_preproc_call] = STATE(1229),
    [sym_preproc_ifdef] = STATE(1229),
    [sym_function_definition] = STATE(1229),
    [sym_declaration] = STATE(1229),
    [sym_type_definition] = STATE(1229),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1229),
    [sym_preproc_undef] = STATE(1229),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5153),
    [sym_property_declaration] = STATE(1229),
    [sym_method_declaration] = STATE(1229),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1229),
    [aux_sym_protocol_declaration_repeat1] = STATE(5153),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3297),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1224)] = {
    [sym_preproc_def] = STATE(1322),
    [sym_preproc_call] = STATE(1322),
    [sym_preproc_ifdef] = STATE(1322),
    [sym_function_definition] = STATE(1322),
    [sym_declaration] = STATE(1322),
    [sym_type_definition] = STATE(1322),
    [sym__declaration_modifiers] = STATE(2901),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2901),
    [sym_attribute_declaration] = STATE(2901),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2901),
    [sym_type_qualifier] = STATE(2901),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3228),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1322),
    [sym_preproc_else_in_interface_declaration] = STATE(6777),
    [sym_preproc_elif_in_interface_declaration] = STATE(6777),
    [sym_preproc_undef] = STATE(1322),
    [sym_availability_attribute_specifier] = STATE(2901),
    [sym_property_declaration] = STATE(1322),
    [sym_method_declaration] = STATE(1322),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1322),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3279),
    [aux_sym_preproc_if_token1] = ACTIONS(3281),
    [aux_sym_preproc_if_token2] = ACTIONS(3307),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3285),
    [aux_sym_preproc_elif_token1] = ACTIONS(3287),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3289),
    [anon_sym_PLUS] = ACTIONS(3289),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATproperty] = ACTIONS(3291),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1225)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5123),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5123),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3309),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1226)] = {
    [sym_preproc_def] = STATE(1227),
    [sym_preproc_call] = STATE(1227),
    [sym_preproc_ifdef] = STATE(1227),
    [sym_function_definition] = STATE(1227),
    [sym_declaration] = STATE(1227),
    [sym_type_definition] = STATE(1227),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1227),
    [sym_preproc_undef] = STATE(1227),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5123),
    [sym_property_declaration] = STATE(1227),
    [sym_method_declaration] = STATE(1227),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1227),
    [aux_sym_protocol_declaration_repeat1] = STATE(5123),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3309),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1227)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5130),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5130),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3311),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1228)] = {
    [sym_preproc_def] = STATE(1322),
    [sym_preproc_call] = STATE(1322),
    [sym_preproc_ifdef] = STATE(1322),
    [sym_function_definition] = STATE(1322),
    [sym_declaration] = STATE(1322),
    [sym_type_definition] = STATE(1322),
    [sym__declaration_modifiers] = STATE(2901),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2901),
    [sym_attribute_declaration] = STATE(2901),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2901),
    [sym_type_qualifier] = STATE(2901),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3228),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1322),
    [sym_preproc_else_in_interface_declaration] = STATE(6099),
    [sym_preproc_elif_in_interface_declaration] = STATE(6099),
    [sym_preproc_undef] = STATE(1322),
    [sym_availability_attribute_specifier] = STATE(2901),
    [sym_property_declaration] = STATE(1322),
    [sym_method_declaration] = STATE(1322),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1322),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3279),
    [aux_sym_preproc_if_token1] = ACTIONS(3281),
    [aux_sym_preproc_if_token2] = ACTIONS(3313),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3285),
    [aux_sym_preproc_elif_token1] = ACTIONS(3287),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3289),
    [anon_sym_PLUS] = ACTIONS(3289),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATproperty] = ACTIONS(3291),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1229)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5122),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5122),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3315),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1230)] = {
    [sym_preproc_def] = STATE(1379),
    [sym_preproc_call] = STATE(1379),
    [sym_preproc_ifdef] = STATE(1379),
    [sym_function_definition] = STATE(1379),
    [sym_declaration] = STATE(1379),
    [sym_type_definition] = STATE(1379),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1379),
    [sym_preproc_undef] = STATE(1379),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_parameterized_arguments] = STATE(1319),
    [sym_instance_variables] = STATE(1378),
    [sym_property_declaration] = STATE(1379),
    [sym_method_declaration] = STATE(1379),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1379),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3317),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1231)] = {
    [sym_preproc_def] = STATE(1393),
    [sym_preproc_call] = STATE(1393),
    [sym_preproc_ifdef] = STATE(1393),
    [sym_function_definition] = STATE(1393),
    [sym_declaration] = STATE(1393),
    [sym_type_definition] = STATE(1393),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1393),
    [sym_preproc_undef] = STATE(1393),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_parameterized_arguments] = STATE(1320),
    [sym_instance_variables] = STATE(1392),
    [sym_property_declaration] = STATE(1393),
    [sym_method_declaration] = STATE(1393),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1393),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3319),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1232)] = {
    [sym_preproc_def] = STATE(1344),
    [sym_preproc_call] = STATE(1344),
    [sym_preproc_ifdef] = STATE(1344),
    [sym_function_definition] = STATE(1344),
    [sym_declaration] = STATE(1344),
    [sym_type_definition] = STATE(1344),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1344),
    [sym_preproc_undef] = STATE(1344),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_parameterized_arguments] = STATE(1315),
    [sym_instance_variables] = STATE(1336),
    [sym_property_declaration] = STATE(1344),
    [sym_method_declaration] = STATE(1344),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1344),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3321),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1233)] = {
    [sym_preproc_def] = STATE(1215),
    [sym_preproc_call] = STATE(1215),
    [sym_preproc_ifdef] = STATE(1215),
    [sym_function_definition] = STATE(1215),
    [sym_declaration] = STATE(1215),
    [sym_type_definition] = STATE(1215),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1215),
    [sym_preproc_undef] = STATE(1215),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5141),
    [sym_property_declaration] = STATE(1215),
    [sym_method_declaration] = STATE(1215),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1215),
    [aux_sym_protocol_declaration_repeat1] = STATE(5141),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3323),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1234)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5110),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5110),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3084),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1235)] = {
    [sym_preproc_def] = STATE(1398),
    [sym_preproc_call] = STATE(1398),
    [sym_preproc_ifdef] = STATE(1398),
    [sym_function_definition] = STATE(1398),
    [sym_declaration] = STATE(1398),
    [sym_type_definition] = STATE(1398),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1398),
    [sym_preproc_undef] = STATE(1398),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_parameterized_arguments] = STATE(1331),
    [sym_instance_variables] = STATE(1397),
    [sym_property_declaration] = STATE(1398),
    [sym_method_declaration] = STATE(1398),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1398),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3325),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1236)] = {
    [sym_preproc_def] = STATE(1430),
    [sym_preproc_call] = STATE(1430),
    [sym_preproc_ifdef] = STATE(1430),
    [sym_function_definition] = STATE(1430),
    [sym_declaration] = STATE(1430),
    [sym_type_definition] = STATE(1430),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1430),
    [sym_preproc_undef] = STATE(1430),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_parameterized_arguments] = STATE(1330),
    [sym_instance_variables] = STATE(1429),
    [sym_property_declaration] = STATE(1430),
    [sym_method_declaration] = STATE(1430),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1430),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3327),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1237)] = {
    [sym_preproc_def] = STATE(1216),
    [sym_preproc_call] = STATE(1216),
    [sym_preproc_ifdef] = STATE(1216),
    [sym_function_definition] = STATE(1216),
    [sym_declaration] = STATE(1216),
    [sym_type_definition] = STATE(1216),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1216),
    [sym_preproc_undef] = STATE(1216),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5134),
    [sym_property_declaration] = STATE(1216),
    [sym_method_declaration] = STATE(1216),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1216),
    [aux_sym_protocol_declaration_repeat1] = STATE(5134),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3072),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1238)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5115),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5115),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3329),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1239)] = {
    [sym_preproc_def] = STATE(1240),
    [sym_preproc_call] = STATE(1240),
    [sym_preproc_ifdef] = STATE(1240),
    [sym_function_definition] = STATE(1240),
    [sym_declaration] = STATE(1240),
    [sym_type_definition] = STATE(1240),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1240),
    [sym_preproc_undef] = STATE(1240),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5115),
    [sym_property_declaration] = STATE(1240),
    [sym_method_declaration] = STATE(1240),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1240),
    [aux_sym_protocol_declaration_repeat1] = STATE(5115),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3329),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1240)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5119),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5119),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3331),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1241)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5134),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5134),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3072),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1242)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5116),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5116),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3333),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1243)] = {
    [sym_preproc_def] = STATE(1385),
    [sym_preproc_call] = STATE(1385),
    [sym_preproc_ifdef] = STATE(1385),
    [sym_function_definition] = STATE(1385),
    [sym_declaration] = STATE(1385),
    [sym_type_definition] = STATE(1385),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1385),
    [sym_preproc_undef] = STATE(1385),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_parameterized_arguments] = STATE(1321),
    [sym_instance_variables] = STATE(1384),
    [sym_property_declaration] = STATE(1385),
    [sym_method_declaration] = STATE(1385),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1385),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3335),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1244)] = {
    [sym_preproc_def] = STATE(1213),
    [sym_preproc_call] = STATE(1213),
    [sym_preproc_ifdef] = STATE(1213),
    [sym_function_definition] = STATE(1213),
    [sym_declaration] = STATE(1213),
    [sym_type_definition] = STATE(1213),
    [sym__declaration_modifiers] = STATE(2901),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2901),
    [sym_attribute_declaration] = STATE(2901),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2901),
    [sym_type_qualifier] = STATE(2901),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3228),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1213),
    [sym_preproc_else_in_interface_declaration] = STATE(6146),
    [sym_preproc_elif_in_interface_declaration] = STATE(6146),
    [sym_preproc_undef] = STATE(1213),
    [sym_availability_attribute_specifier] = STATE(2901),
    [sym_property_declaration] = STATE(1213),
    [sym_method_declaration] = STATE(1213),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1213),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3279),
    [aux_sym_preproc_if_token1] = ACTIONS(3281),
    [aux_sym_preproc_if_token2] = ACTIONS(3337),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3285),
    [aux_sym_preproc_elif_token1] = ACTIONS(3287),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3289),
    [anon_sym_PLUS] = ACTIONS(3289),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATproperty] = ACTIONS(3291),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1245)] = {
    [sym_preproc_def] = STATE(1242),
    [sym_preproc_call] = STATE(1242),
    [sym_preproc_ifdef] = STATE(1242),
    [sym_function_definition] = STATE(1242),
    [sym_declaration] = STATE(1242),
    [sym_type_definition] = STATE(1242),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1242),
    [sym_preproc_undef] = STATE(1242),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5129),
    [sym_property_declaration] = STATE(1242),
    [sym_method_declaration] = STATE(1242),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1242),
    [aux_sym_protocol_declaration_repeat1] = STATE(5129),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3080),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1246)] = {
    [sym_preproc_def] = STATE(1253),
    [sym_preproc_call] = STATE(1253),
    [sym_preproc_ifdef] = STATE(1253),
    [sym_function_definition] = STATE(1253),
    [sym_declaration] = STATE(1253),
    [sym_type_definition] = STATE(1253),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1253),
    [sym_preproc_undef] = STATE(1253),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5158),
    [sym_property_declaration] = STATE(1253),
    [sym_method_declaration] = STATE(1253),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1253),
    [aux_sym_protocol_declaration_repeat1] = STATE(5158),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3104),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1247)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5158),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5158),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3104),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1248)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5129),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5129),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3080),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1249)] = {
    [sym_preproc_def] = STATE(1339),
    [sym_preproc_call] = STATE(1339),
    [sym_preproc_ifdef] = STATE(1339),
    [sym_function_definition] = STATE(1339),
    [sym_declaration] = STATE(1339),
    [sym_type_definition] = STATE(1339),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1339),
    [sym_preproc_undef] = STATE(1339),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_parameterized_arguments] = STATE(1329),
    [sym_instance_variables] = STATE(1338),
    [sym_property_declaration] = STATE(1339),
    [sym_method_declaration] = STATE(1339),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1339),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3339),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1250)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5103),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5103),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3341),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1251)] = {
    [sym_preproc_def] = STATE(1228),
    [sym_preproc_call] = STATE(1228),
    [sym_preproc_ifdef] = STATE(1228),
    [sym_function_definition] = STATE(1228),
    [sym_declaration] = STATE(1228),
    [sym_type_definition] = STATE(1228),
    [sym__declaration_modifiers] = STATE(2901),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2901),
    [sym_attribute_declaration] = STATE(2901),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2901),
    [sym_type_qualifier] = STATE(2901),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3228),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1228),
    [sym_preproc_else_in_interface_declaration] = STATE(6561),
    [sym_preproc_elif_in_interface_declaration] = STATE(6561),
    [sym_preproc_undef] = STATE(1228),
    [sym_availability_attribute_specifier] = STATE(2901),
    [sym_property_declaration] = STATE(1228),
    [sym_method_declaration] = STATE(1228),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1228),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3279),
    [aux_sym_preproc_if_token1] = ACTIONS(3281),
    [aux_sym_preproc_if_token2] = ACTIONS(3343),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3285),
    [aux_sym_preproc_elif_token1] = ACTIONS(3287),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3289),
    [anon_sym_PLUS] = ACTIONS(3289),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATproperty] = ACTIONS(3291),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1252)] = {
    [sym_preproc_def] = STATE(1250),
    [sym_preproc_call] = STATE(1250),
    [sym_preproc_ifdef] = STATE(1250),
    [sym_function_definition] = STATE(1250),
    [sym_declaration] = STATE(1250),
    [sym_type_definition] = STATE(1250),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1250),
    [sym_preproc_undef] = STATE(1250),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5116),
    [sym_property_declaration] = STATE(1250),
    [sym_method_declaration] = STATE(1250),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1250),
    [aux_sym_protocol_declaration_repeat1] = STATE(5116),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3333),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1253)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5141),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [aux_sym_protocol_declaration_repeat1] = STATE(5141),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3323),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1254)] = {
    [sym_preproc_def] = STATE(1214),
    [sym_preproc_call] = STATE(1214),
    [sym_preproc_ifdef] = STATE(1214),
    [sym_function_definition] = STATE(1214),
    [sym_declaration] = STATE(1214),
    [sym_type_definition] = STATE(1214),
    [sym__declaration_modifiers] = STATE(2901),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2901),
    [sym_attribute_declaration] = STATE(2901),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2901),
    [sym_type_qualifier] = STATE(2901),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3228),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1214),
    [sym_preproc_else_in_interface_declaration] = STATE(6315),
    [sym_preproc_elif_in_interface_declaration] = STATE(6315),
    [sym_preproc_undef] = STATE(1214),
    [sym_availability_attribute_specifier] = STATE(2901),
    [sym_property_declaration] = STATE(1214),
    [sym_method_declaration] = STATE(1214),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1214),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3279),
    [aux_sym_preproc_if_token1] = ACTIONS(3281),
    [aux_sym_preproc_if_token2] = ACTIONS(3345),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3132),
    [aux_sym_preproc_else_token1] = ACTIONS(3285),
    [aux_sym_preproc_elif_token1] = ACTIONS(3287),
    [sym_preproc_directive] = ACTIONS(3138),
    [anon_sym_DASH] = ACTIONS(3289),
    [anon_sym_PLUS] = ACTIONS(3289),
    [anon_sym___extension__] = ACTIONS(3142),
    [anon_sym_typedef] = ACTIONS(3144),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATproperty] = ACTIONS(3291),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1255)] = {
    [sym_preproc_def] = STATE(1238),
    [sym_preproc_call] = STATE(1238),
    [sym_preproc_ifdef] = STATE(1238),
    [sym_function_definition] = STATE(1238),
    [sym_declaration] = STATE(1238),
    [sym_type_definition] = STATE(1238),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1238),
    [sym_preproc_undef] = STATE(1238),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_qualified_protocol_interface_declaration] = STATE(5110),
    [sym_property_declaration] = STATE(1238),
    [sym_method_declaration] = STATE(1238),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1238),
    [aux_sym_protocol_declaration_repeat1] = STATE(5110),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3084),
    [anon_sym_AToptional] = ACTIONS(3054),
    [anon_sym_ATrequired] = ACTIONS(3054),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1256)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3347),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1257)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3349),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1258)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3114),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1259)] = {
    [sym_preproc_def] = STATE(2386),
    [sym_preproc_function_def] = STATE(2386),
    [sym_preproc_call] = STATE(2386),
    [sym_preproc_ifdef] = STATE(2386),
    [sym_function_definition] = STATE(2386),
    [sym_declaration] = STATE(2386),
    [sym_type_definition] = STATE(2386),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4221),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2870),
    [sym_ms_call_modifier] = STATE(2179),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3237),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2073),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2386),
    [sym_preproc_undef] = STATE(2386),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1267),
    [sym_property_implementation] = STATE(2386),
    [sym_method_definition] = STATE(2386),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(3351),
    [aux_sym_preproc_def_token1] = ACTIONS(3353),
    [aux_sym_preproc_if_token1] = ACTIONS(3355),
    [aux_sym_preproc_if_token2] = ACTIONS(3357),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3359),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3359),
    [sym_preproc_directive] = ACTIONS(3361),
    [anon_sym_DASH] = ACTIONS(3363),
    [anon_sym_PLUS] = ACTIONS(3363),
    [anon_sym___extension__] = ACTIONS(3365),
    [anon_sym_typedef] = ACTIONS(3367),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3369),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATsynthesize] = ACTIONS(3371),
    [anon_sym_ATdynamic] = ACTIONS(3373),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1260)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3375),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1261)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3377),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1262)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1256),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1256),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3110),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1263)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1274),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1274),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3112),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1264)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3112),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1265)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3110),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1266)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3379),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1267)] = {
    [sym_preproc_def] = STATE(2386),
    [sym_preproc_function_def] = STATE(2386),
    [sym_preproc_call] = STATE(2386),
    [sym_preproc_ifdef] = STATE(2386),
    [sym_function_definition] = STATE(2386),
    [sym_declaration] = STATE(2386),
    [sym_type_definition] = STATE(2386),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4221),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2870),
    [sym_ms_call_modifier] = STATE(2179),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3237),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2073),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2386),
    [sym_preproc_undef] = STATE(2386),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1267),
    [sym_property_implementation] = STATE(2386),
    [sym_method_definition] = STATE(2386),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(3381),
    [aux_sym_preproc_def_token1] = ACTIONS(3384),
    [aux_sym_preproc_if_token1] = ACTIONS(3387),
    [aux_sym_preproc_if_token2] = ACTIONS(3185),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3390),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3390),
    [sym_preproc_directive] = ACTIONS(3393),
    [anon_sym_DASH] = ACTIONS(3396),
    [anon_sym_PLUS] = ACTIONS(3396),
    [anon_sym___extension__] = ACTIONS(3399),
    [anon_sym_typedef] = ACTIONS(3402),
    [anon_sym_extern] = ACTIONS(3202),
    [anon_sym___attribute__] = ACTIONS(3205),
    [anon_sym___attribute] = ACTIONS(3205),
    [anon_sym_noreturn] = ACTIONS(3208),
    [anon_sym_LBRACK] = ACTIONS(3211),
    [anon_sym___declspec] = ACTIONS(3214),
    [anon_sym___cdecl] = ACTIONS(3217),
    [anon_sym___clrcall] = ACTIONS(3217),
    [anon_sym___stdcall] = ACTIONS(3217),
    [anon_sym___fastcall] = ACTIONS(3217),
    [anon_sym___thiscall] = ACTIONS(3217),
    [anon_sym___vectorcall] = ACTIONS(3217),
    [anon_sym_signed] = ACTIONS(3220),
    [anon_sym_unsigned] = ACTIONS(3220),
    [anon_sym_long] = ACTIONS(3220),
    [anon_sym_short] = ACTIONS(3220),
    [anon_sym_static] = ACTIONS(3202),
    [anon_sym_auto] = ACTIONS(3202),
    [anon_sym_register] = ACTIONS(3202),
    [anon_sym_inline] = ACTIONS(3202),
    [anon_sym___inline] = ACTIONS(3202),
    [anon_sym___inline__] = ACTIONS(3202),
    [anon_sym___forceinline] = ACTIONS(3202),
    [anon_sym_thread_local] = ACTIONS(3202),
    [anon_sym___thread] = ACTIONS(3202),
    [anon_sym_CG_EXTERN] = ACTIONS(3202),
    [anon_sym_CG_INLINE] = ACTIONS(3202),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3202),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3202),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3202),
    [anon_sym_IBOutlet] = ACTIONS(3202),
    [anon_sym_IBInspectable] = ACTIONS(3202),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3202),
    [anon_sym_NS_INLINE] = ACTIONS(3202),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3202),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3202),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3202),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3202),
    [anon_sym_const] = ACTIONS(3208),
    [anon_sym_constexpr] = ACTIONS(3208),
    [anon_sym_volatile] = ACTIONS(3208),
    [anon_sym_restrict] = ACTIONS(3208),
    [anon_sym___restrict__] = ACTIONS(3208),
    [anon_sym__Atomic] = ACTIONS(3208),
    [anon_sym__Noreturn] = ACTIONS(3208),
    [anon_sym__Nonnull] = ACTIONS(3208),
    [anon_sym_nullable] = ACTIONS(3208),
    [anon_sym__Complex] = ACTIONS(3208),
    [anon_sym__Nullable] = ACTIONS(3208),
    [anon_sym__Nullable_result] = ACTIONS(3208),
    [anon_sym__Null_unspecified] = ACTIONS(3208),
    [anon_sym___autoreleasing] = ACTIONS(3208),
    [anon_sym___block] = ACTIONS(3208),
    [anon_sym___bridge] = ACTIONS(3208),
    [anon_sym___bridge_retained] = ACTIONS(3208),
    [anon_sym___bridge_transfer] = ACTIONS(3208),
    [anon_sym___complex] = ACTIONS(3208),
    [anon_sym___const] = ACTIONS(3208),
    [anon_sym___imag] = ACTIONS(3208),
    [anon_sym___kindof] = ACTIONS(3208),
    [anon_sym___nonnull] = ACTIONS(3208),
    [anon_sym___nullable] = ACTIONS(3208),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3208),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3208),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3208),
    [anon_sym___real] = ACTIONS(3208),
    [anon_sym___strong] = ACTIONS(3208),
    [anon_sym___unsafe_unretained] = ACTIONS(3208),
    [anon_sym___unused] = ACTIONS(3208),
    [anon_sym___weak] = ACTIONS(3208),
    [anon_sym_alignas] = ACTIONS(3223),
    [anon_sym__Alignas] = ACTIONS(3223),
    [sym_primitive_type] = ACTIONS(3226),
    [anon_sym_enum] = ACTIONS(3229),
    [anon_sym_struct] = ACTIONS(3232),
    [anon_sym_union] = ACTIONS(3235),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3238),
    [anon_sym___typeof] = ACTIONS(3238),
    [anon_sym_typeof] = ACTIONS(3238),
    [aux_sym_preproc_undef_token1] = ACTIONS(3405),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3244),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3244),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3244),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3244),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3244),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3244),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3244),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3244),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3244),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3247),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3247),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3247),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3247),
    [anon_sym_API_AVAILABLE] = ACTIONS(3247),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3247),
    [anon_sym_API_DEPRECATED] = ACTIONS(3247),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3247),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3247),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3247),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3247),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3247),
    [anon_sym___deprecated_msg] = ACTIONS(3247),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3247),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3247),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3247),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3247),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3247),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3247),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3247),
    [anon_sym_ATsynthesize] = ACTIONS(3408),
    [anon_sym_ATdynamic] = ACTIONS(3411),
    [anon_sym_BOOL] = ACTIONS(3256),
    [anon_sym_IMP] = ACTIONS(3256),
    [anon_sym_SEL] = ACTIONS(3256),
    [anon_sym_Class] = ACTIONS(3256),
    [anon_sym_id] = ACTIONS(3256),
  },
  [STATE(1268)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1299),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1299),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3414),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1269)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1290),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1290),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3116),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1270)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3416),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1271)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(3418),
    [aux_sym_preproc_def_token1] = ACTIONS(3421),
    [aux_sym_preproc_if_token1] = ACTIONS(3424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3427),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3427),
    [sym_preproc_directive] = ACTIONS(3430),
    [anon_sym_DASH] = ACTIONS(3433),
    [anon_sym_PLUS] = ACTIONS(3433),
    [anon_sym___extension__] = ACTIONS(3436),
    [anon_sym_typedef] = ACTIONS(3439),
    [anon_sym_extern] = ACTIONS(3442),
    [anon_sym___attribute__] = ACTIONS(3445),
    [anon_sym___attribute] = ACTIONS(3445),
    [anon_sym_noreturn] = ACTIONS(3448),
    [anon_sym_LBRACK] = ACTIONS(3451),
    [anon_sym___declspec] = ACTIONS(3454),
    [anon_sym___cdecl] = ACTIONS(3457),
    [anon_sym___clrcall] = ACTIONS(3457),
    [anon_sym___stdcall] = ACTIONS(3457),
    [anon_sym___fastcall] = ACTIONS(3457),
    [anon_sym___thiscall] = ACTIONS(3457),
    [anon_sym___vectorcall] = ACTIONS(3457),
    [anon_sym_signed] = ACTIONS(3460),
    [anon_sym_unsigned] = ACTIONS(3460),
    [anon_sym_long] = ACTIONS(3460),
    [anon_sym_short] = ACTIONS(3460),
    [anon_sym_static] = ACTIONS(3442),
    [anon_sym_auto] = ACTIONS(3442),
    [anon_sym_register] = ACTIONS(3442),
    [anon_sym_inline] = ACTIONS(3442),
    [anon_sym___inline] = ACTIONS(3442),
    [anon_sym___inline__] = ACTIONS(3442),
    [anon_sym___forceinline] = ACTIONS(3442),
    [anon_sym_thread_local] = ACTIONS(3442),
    [anon_sym___thread] = ACTIONS(3442),
    [anon_sym_CG_EXTERN] = ACTIONS(3442),
    [anon_sym_CG_INLINE] = ACTIONS(3442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3442),
    [anon_sym_IBOutlet] = ACTIONS(3442),
    [anon_sym_IBInspectable] = ACTIONS(3442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3442),
    [anon_sym_NS_INLINE] = ACTIONS(3442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3442),
    [anon_sym_const] = ACTIONS(3448),
    [anon_sym_constexpr] = ACTIONS(3448),
    [anon_sym_volatile] = ACTIONS(3448),
    [anon_sym_restrict] = ACTIONS(3448),
    [anon_sym___restrict__] = ACTIONS(3448),
    [anon_sym__Atomic] = ACTIONS(3448),
    [anon_sym__Noreturn] = ACTIONS(3448),
    [anon_sym__Nonnull] = ACTIONS(3448),
    [anon_sym_nullable] = ACTIONS(3448),
    [anon_sym__Complex] = ACTIONS(3448),
    [anon_sym__Nullable] = ACTIONS(3448),
    [anon_sym__Nullable_result] = ACTIONS(3448),
    [anon_sym__Null_unspecified] = ACTIONS(3448),
    [anon_sym___autoreleasing] = ACTIONS(3448),
    [anon_sym___block] = ACTIONS(3448),
    [anon_sym___bridge] = ACTIONS(3448),
    [anon_sym___bridge_retained] = ACTIONS(3448),
    [anon_sym___bridge_transfer] = ACTIONS(3448),
    [anon_sym___complex] = ACTIONS(3448),
    [anon_sym___const] = ACTIONS(3448),
    [anon_sym___imag] = ACTIONS(3448),
    [anon_sym___kindof] = ACTIONS(3448),
    [anon_sym___nonnull] = ACTIONS(3448),
    [anon_sym___nullable] = ACTIONS(3448),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3448),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3448),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3448),
    [anon_sym___real] = ACTIONS(3448),
    [anon_sym___strong] = ACTIONS(3448),
    [anon_sym___unsafe_unretained] = ACTIONS(3448),
    [anon_sym___unused] = ACTIONS(3448),
    [anon_sym___weak] = ACTIONS(3448),
    [anon_sym_alignas] = ACTIONS(3463),
    [anon_sym__Alignas] = ACTIONS(3463),
    [sym_primitive_type] = ACTIONS(3466),
    [anon_sym_enum] = ACTIONS(3469),
    [anon_sym_struct] = ACTIONS(3472),
    [anon_sym_union] = ACTIONS(3475),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3478),
    [anon_sym___typeof] = ACTIONS(3478),
    [anon_sym_typeof] = ACTIONS(3478),
    [aux_sym_preproc_undef_token1] = ACTIONS(3481),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3484),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3484),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3484),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3484),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3484),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3484),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3484),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3484),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3484),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3487),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3487),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3487),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3487),
    [anon_sym_API_AVAILABLE] = ACTIONS(3487),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3487),
    [anon_sym_API_DEPRECATED] = ACTIONS(3487),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3487),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3487),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3487),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3487),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3487),
    [anon_sym___deprecated_msg] = ACTIONS(3487),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3487),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3487),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3487),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3487),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3487),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3487),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3487),
    [anon_sym_ATend] = ACTIONS(3490),
    [anon_sym_ATsynthesize] = ACTIONS(3492),
    [anon_sym_ATdynamic] = ACTIONS(3495),
    [anon_sym_BOOL] = ACTIONS(3498),
    [anon_sym_IMP] = ACTIONS(3498),
    [anon_sym_SEL] = ACTIONS(3498),
    [anon_sym_Class] = ACTIONS(3498),
    [anon_sym_id] = ACTIONS(3498),
  },
  [STATE(1272)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3501),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1273)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1279),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1279),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3503),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1274)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3503),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1275)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1280),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1280),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3505),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1276)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3507),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1277)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1282),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1282),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3509),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1278)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3511),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1279)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3513),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1280)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3515),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1281)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3517),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1282)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3519),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1283)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1284),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1284),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3114),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1284)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3521),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1285)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1298),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1298),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3523),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1286)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3525),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1287)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1303),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1303),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3118),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1288)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3118),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1289)] = {
    [sym_preproc_def] = STATE(2386),
    [sym_preproc_function_def] = STATE(2386),
    [sym_preproc_call] = STATE(2386),
    [sym_preproc_ifdef] = STATE(2386),
    [sym_function_definition] = STATE(2386),
    [sym_declaration] = STATE(2386),
    [sym_type_definition] = STATE(2386),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4221),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2870),
    [sym_ms_call_modifier] = STATE(2179),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3237),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2073),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2386),
    [sym_preproc_undef] = STATE(2386),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1259),
    [sym_property_implementation] = STATE(2386),
    [sym_method_definition] = STATE(2386),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_implementation_definition_repeat1] = STATE(1259),
    [sym_identifier] = ACTIONS(3351),
    [aux_sym_preproc_def_token1] = ACTIONS(3353),
    [aux_sym_preproc_if_token1] = ACTIONS(3355),
    [aux_sym_preproc_if_token2] = ACTIONS(3527),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3359),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3359),
    [sym_preproc_directive] = ACTIONS(3361),
    [anon_sym_DASH] = ACTIONS(3363),
    [anon_sym_PLUS] = ACTIONS(3363),
    [anon_sym___extension__] = ACTIONS(3365),
    [anon_sym_typedef] = ACTIONS(3367),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3369),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATsynthesize] = ACTIONS(3371),
    [anon_sym_ATdynamic] = ACTIONS(3373),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1290)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3529),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1291)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1301),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1301),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3531),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1292)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3533),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1293)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3535),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1294)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1270),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1270),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3537),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1295)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3539),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1296)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1297),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1297),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3529),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1297)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3541),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1298)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3543),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1299)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3545),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1300)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3547),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1301)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3549),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1302)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1313),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1313),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3551),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1303)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3551),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1304)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1307),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1307),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3553),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1305)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3555),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1306)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1261),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1261),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3347),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1307)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3557),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1308)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1257),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1257),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3559),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1309)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3116),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1310)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1266),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1266),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3561),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1311)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1272),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1272),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3563),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1312)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3565),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1313)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1271),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1271),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3567),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1314)] = {
    [sym_preproc_def] = STATE(2371),
    [sym_preproc_function_def] = STATE(2371),
    [sym_preproc_call] = STATE(2371),
    [sym_preproc_ifdef] = STATE(2371),
    [sym_function_definition] = STATE(2371),
    [sym_declaration] = STATE(2371),
    [sym_type_definition] = STATE(2371),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3231),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2071),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_implementation_definition] = STATE(2371),
    [sym_preproc_undef] = STATE(2371),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_implementation_definition] = STATE(1293),
    [sym_property_implementation] = STATE(2371),
    [sym_method_definition] = STATE(2371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_implementation_repeat1] = STATE(1293),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(2998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3521),
    [anon_sym_ATsynthesize] = ACTIONS(3026),
    [anon_sym_ATdynamic] = ACTIONS(3028),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1315)] = {
    [sym_preproc_def] = STATE(1421),
    [sym_preproc_call] = STATE(1421),
    [sym_preproc_ifdef] = STATE(1421),
    [sym_function_definition] = STATE(1421),
    [sym_declaration] = STATE(1421),
    [sym_type_definition] = STATE(1421),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1421),
    [sym_preproc_undef] = STATE(1421),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1420),
    [sym_property_declaration] = STATE(1421),
    [sym_method_declaration] = STATE(1421),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1421),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3569),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1316)] = {
    [sym_preproc_def] = STATE(1387),
    [sym_preproc_call] = STATE(1387),
    [sym_preproc_ifdef] = STATE(1387),
    [sym_function_definition] = STATE(1387),
    [sym_declaration] = STATE(1387),
    [sym_type_definition] = STATE(1387),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1387),
    [sym_preproc_undef] = STATE(1387),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1386),
    [sym_property_declaration] = STATE(1387),
    [sym_method_declaration] = STATE(1387),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1387),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3571),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1317)] = {
    [sym_preproc_def] = STATE(1326),
    [sym_preproc_call] = STATE(1326),
    [sym_preproc_ifdef] = STATE(1326),
    [sym_function_definition] = STATE(1326),
    [sym_declaration] = STATE(1326),
    [sym_type_definition] = STATE(1326),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1326),
    [sym_preproc_undef] = STATE(1326),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1326),
    [sym_method_declaration] = STATE(1326),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1326),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3573),
    [anon_sym_AToptional] = ACTIONS(3573),
    [anon_sym_ATrequired] = ACTIONS(3573),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1318)] = {
    [sym_preproc_def] = STATE(1400),
    [sym_preproc_call] = STATE(1400),
    [sym_preproc_ifdef] = STATE(1400),
    [sym_function_definition] = STATE(1400),
    [sym_declaration] = STATE(1400),
    [sym_type_definition] = STATE(1400),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1400),
    [sym_preproc_undef] = STATE(1400),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1399),
    [sym_property_declaration] = STATE(1400),
    [sym_method_declaration] = STATE(1400),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1400),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3575),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1319)] = {
    [sym_preproc_def] = STATE(1391),
    [sym_preproc_call] = STATE(1391),
    [sym_preproc_ifdef] = STATE(1391),
    [sym_function_definition] = STATE(1391),
    [sym_declaration] = STATE(1391),
    [sym_type_definition] = STATE(1391),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1391),
    [sym_preproc_undef] = STATE(1391),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1388),
    [sym_property_declaration] = STATE(1391),
    [sym_method_declaration] = STATE(1391),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1391),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3577),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1320)] = {
    [sym_preproc_def] = STATE(1402),
    [sym_preproc_call] = STATE(1402),
    [sym_preproc_ifdef] = STATE(1402),
    [sym_function_definition] = STATE(1402),
    [sym_declaration] = STATE(1402),
    [sym_type_definition] = STATE(1402),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1402),
    [sym_preproc_undef] = STATE(1402),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1401),
    [sym_property_declaration] = STATE(1402),
    [sym_method_declaration] = STATE(1402),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1402),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3579),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1321)] = {
    [sym_preproc_def] = STATE(1405),
    [sym_preproc_call] = STATE(1405),
    [sym_preproc_ifdef] = STATE(1405),
    [sym_function_definition] = STATE(1405),
    [sym_declaration] = STATE(1405),
    [sym_type_definition] = STATE(1405),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1405),
    [sym_preproc_undef] = STATE(1405),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1404),
    [sym_property_declaration] = STATE(1405),
    [sym_method_declaration] = STATE(1405),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1405),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3581),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1322)] = {
    [sym_preproc_def] = STATE(1322),
    [sym_preproc_call] = STATE(1322),
    [sym_preproc_ifdef] = STATE(1322),
    [sym_function_definition] = STATE(1322),
    [sym_declaration] = STATE(1322),
    [sym_type_definition] = STATE(1322),
    [sym__declaration_modifiers] = STATE(2901),
    [sym__declaration_specifiers] = STATE(4229),
    [sym_attribute_specifier] = STATE(2901),
    [sym_attribute_declaration] = STATE(2901),
    [sym_ms_declspec_modifier] = STATE(2863),
    [sym_ms_call_modifier] = STATE(2174),
    [sym_storage_class_specifier] = STATE(2901),
    [sym_type_qualifier] = STATE(2901),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3228),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1322),
    [sym_preproc_undef] = STATE(1322),
    [sym_availability_attribute_specifier] = STATE(2901),
    [sym_property_declaration] = STATE(1322),
    [sym_method_declaration] = STATE(1322),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1322),
    [sym_identifier] = ACTIONS(3583),
    [aux_sym_preproc_def_token1] = ACTIONS(3586),
    [aux_sym_preproc_if_token1] = ACTIONS(3589),
    [aux_sym_preproc_if_token2] = ACTIONS(3592),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3594),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3594),
    [aux_sym_preproc_else_token1] = ACTIONS(3592),
    [aux_sym_preproc_elif_token1] = ACTIONS(3592),
    [sym_preproc_directive] = ACTIONS(3597),
    [anon_sym_DASH] = ACTIONS(3600),
    [anon_sym_PLUS] = ACTIONS(3600),
    [anon_sym___extension__] = ACTIONS(3603),
    [anon_sym_typedef] = ACTIONS(3606),
    [anon_sym_extern] = ACTIONS(3609),
    [anon_sym___attribute__] = ACTIONS(3612),
    [anon_sym___attribute] = ACTIONS(3612),
    [anon_sym_noreturn] = ACTIONS(3615),
    [anon_sym_LBRACK] = ACTIONS(3618),
    [anon_sym___declspec] = ACTIONS(3621),
    [anon_sym___cdecl] = ACTIONS(3624),
    [anon_sym___clrcall] = ACTIONS(3624),
    [anon_sym___stdcall] = ACTIONS(3624),
    [anon_sym___fastcall] = ACTIONS(3624),
    [anon_sym___thiscall] = ACTIONS(3624),
    [anon_sym___vectorcall] = ACTIONS(3624),
    [anon_sym_signed] = ACTIONS(3627),
    [anon_sym_unsigned] = ACTIONS(3627),
    [anon_sym_long] = ACTIONS(3627),
    [anon_sym_short] = ACTIONS(3627),
    [anon_sym_static] = ACTIONS(3609),
    [anon_sym_auto] = ACTIONS(3609),
    [anon_sym_register] = ACTIONS(3609),
    [anon_sym_inline] = ACTIONS(3609),
    [anon_sym___inline] = ACTIONS(3609),
    [anon_sym___inline__] = ACTIONS(3609),
    [anon_sym___forceinline] = ACTIONS(3609),
    [anon_sym_thread_local] = ACTIONS(3609),
    [anon_sym___thread] = ACTIONS(3609),
    [anon_sym_CG_EXTERN] = ACTIONS(3609),
    [anon_sym_CG_INLINE] = ACTIONS(3609),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3609),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3609),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3609),
    [anon_sym_IBOutlet] = ACTIONS(3609),
    [anon_sym_IBInspectable] = ACTIONS(3609),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3609),
    [anon_sym_NS_INLINE] = ACTIONS(3609),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3609),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3609),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3609),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3609),
    [anon_sym_const] = ACTIONS(3615),
    [anon_sym_constexpr] = ACTIONS(3615),
    [anon_sym_volatile] = ACTIONS(3615),
    [anon_sym_restrict] = ACTIONS(3615),
    [anon_sym___restrict__] = ACTIONS(3615),
    [anon_sym__Atomic] = ACTIONS(3615),
    [anon_sym__Noreturn] = ACTIONS(3615),
    [anon_sym__Nonnull] = ACTIONS(3615),
    [anon_sym_nullable] = ACTIONS(3615),
    [anon_sym__Complex] = ACTIONS(3615),
    [anon_sym__Nullable] = ACTIONS(3615),
    [anon_sym__Nullable_result] = ACTIONS(3615),
    [anon_sym__Null_unspecified] = ACTIONS(3615),
    [anon_sym___autoreleasing] = ACTIONS(3615),
    [anon_sym___block] = ACTIONS(3615),
    [anon_sym___bridge] = ACTIONS(3615),
    [anon_sym___bridge_retained] = ACTIONS(3615),
    [anon_sym___bridge_transfer] = ACTIONS(3615),
    [anon_sym___complex] = ACTIONS(3615),
    [anon_sym___const] = ACTIONS(3615),
    [anon_sym___imag] = ACTIONS(3615),
    [anon_sym___kindof] = ACTIONS(3615),
    [anon_sym___nonnull] = ACTIONS(3615),
    [anon_sym___nullable] = ACTIONS(3615),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3615),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3615),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3615),
    [anon_sym___real] = ACTIONS(3615),
    [anon_sym___strong] = ACTIONS(3615),
    [anon_sym___unsafe_unretained] = ACTIONS(3615),
    [anon_sym___unused] = ACTIONS(3615),
    [anon_sym___weak] = ACTIONS(3615),
    [anon_sym_alignas] = ACTIONS(3630),
    [anon_sym__Alignas] = ACTIONS(3630),
    [sym_primitive_type] = ACTIONS(3633),
    [anon_sym_enum] = ACTIONS(3636),
    [anon_sym_struct] = ACTIONS(3639),
    [anon_sym_union] = ACTIONS(3642),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3645),
    [anon_sym___typeof] = ACTIONS(3645),
    [anon_sym_typeof] = ACTIONS(3645),
    [aux_sym_preproc_undef_token1] = ACTIONS(3648),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3651),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3651),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3651),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3651),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3651),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3651),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3651),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3651),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3651),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3654),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3654),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3654),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3654),
    [anon_sym_API_AVAILABLE] = ACTIONS(3654),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3654),
    [anon_sym_API_DEPRECATED] = ACTIONS(3654),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3654),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3654),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3654),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3654),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3654),
    [anon_sym___deprecated_msg] = ACTIONS(3654),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3654),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3654),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3654),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3654),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3654),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3654),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3654),
    [anon_sym_ATproperty] = ACTIONS(3657),
    [anon_sym_BOOL] = ACTIONS(3660),
    [anon_sym_IMP] = ACTIONS(3660),
    [anon_sym_SEL] = ACTIONS(3660),
    [anon_sym_Class] = ACTIONS(3660),
    [anon_sym_id] = ACTIONS(3660),
  },
  [STATE(1323)] = {
    [sym_preproc_def] = STATE(1334),
    [sym_preproc_call] = STATE(1334),
    [sym_preproc_ifdef] = STATE(1334),
    [sym_function_definition] = STATE(1334),
    [sym_declaration] = STATE(1334),
    [sym_type_definition] = STATE(1334),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1334),
    [sym_preproc_undef] = STATE(1334),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1423),
    [sym_property_declaration] = STATE(1334),
    [sym_method_declaration] = STATE(1334),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1334),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3663),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1324)] = {
    [sym_preproc_def] = STATE(1371),
    [sym_preproc_call] = STATE(1371),
    [sym_preproc_ifdef] = STATE(1371),
    [sym_function_definition] = STATE(1371),
    [sym_declaration] = STATE(1371),
    [sym_type_definition] = STATE(1371),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1371),
    [sym_preproc_undef] = STATE(1371),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1370),
    [sym_property_declaration] = STATE(1371),
    [sym_method_declaration] = STATE(1371),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1371),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3665),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1325)] = {
    [sym_preproc_def] = STATE(1366),
    [sym_preproc_call] = STATE(1366),
    [sym_preproc_ifdef] = STATE(1366),
    [sym_function_definition] = STATE(1366),
    [sym_declaration] = STATE(1366),
    [sym_type_definition] = STATE(1366),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1366),
    [sym_preproc_undef] = STATE(1366),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1365),
    [sym_property_declaration] = STATE(1366),
    [sym_method_declaration] = STATE(1366),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1366),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3667),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1326)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3669),
    [anon_sym_AToptional] = ACTIONS(3669),
    [anon_sym_ATrequired] = ACTIONS(3669),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1327)] = {
    [sym_preproc_def] = STATE(1368),
    [sym_preproc_call] = STATE(1368),
    [sym_preproc_ifdef] = STATE(1368),
    [sym_function_definition] = STATE(1368),
    [sym_declaration] = STATE(1368),
    [sym_type_definition] = STATE(1368),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1368),
    [sym_preproc_undef] = STATE(1368),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1367),
    [sym_property_declaration] = STATE(1368),
    [sym_method_declaration] = STATE(1368),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1368),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3671),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1328)] = {
    [sym_preproc_def] = STATE(1341),
    [sym_preproc_call] = STATE(1341),
    [sym_preproc_ifdef] = STATE(1341),
    [sym_function_definition] = STATE(1341),
    [sym_declaration] = STATE(1341),
    [sym_type_definition] = STATE(1341),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1341),
    [sym_preproc_undef] = STATE(1341),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1340),
    [sym_property_declaration] = STATE(1341),
    [sym_method_declaration] = STATE(1341),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1341),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3673),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1329)] = {
    [sym_preproc_def] = STATE(1346),
    [sym_preproc_call] = STATE(1346),
    [sym_preproc_ifdef] = STATE(1346),
    [sym_function_definition] = STATE(1346),
    [sym_declaration] = STATE(1346),
    [sym_type_definition] = STATE(1346),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1346),
    [sym_preproc_undef] = STATE(1346),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1345),
    [sym_property_declaration] = STATE(1346),
    [sym_method_declaration] = STATE(1346),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1346),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3675),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1330)] = {
    [sym_preproc_def] = STATE(1342),
    [sym_preproc_call] = STATE(1342),
    [sym_preproc_ifdef] = STATE(1342),
    [sym_function_definition] = STATE(1342),
    [sym_declaration] = STATE(1342),
    [sym_type_definition] = STATE(1342),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1342),
    [sym_preproc_undef] = STATE(1342),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1431),
    [sym_property_declaration] = STATE(1342),
    [sym_method_declaration] = STATE(1342),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1342),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3677),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1331)] = {
    [sym_preproc_def] = STATE(1409),
    [sym_preproc_call] = STATE(1409),
    [sym_preproc_ifdef] = STATE(1409),
    [sym_function_definition] = STATE(1409),
    [sym_declaration] = STATE(1409),
    [sym_type_definition] = STATE(1409),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1409),
    [sym_preproc_undef] = STATE(1409),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1408),
    [sym_property_declaration] = STATE(1409),
    [sym_method_declaration] = STATE(1409),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1409),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3679),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1332)] = {
    [sym_preproc_def] = STATE(1419),
    [sym_preproc_call] = STATE(1419),
    [sym_preproc_ifdef] = STATE(1419),
    [sym_function_definition] = STATE(1419),
    [sym_declaration] = STATE(1419),
    [sym_type_definition] = STATE(1419),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1419),
    [sym_preproc_undef] = STATE(1419),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_instance_variables] = STATE(1418),
    [sym_property_declaration] = STATE(1419),
    [sym_method_declaration] = STATE(1419),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1419),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(3016),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3681),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1333)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(3683),
    [aux_sym_preproc_def_token1] = ACTIONS(3686),
    [aux_sym_preproc_if_token1] = ACTIONS(3689),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3692),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3692),
    [sym_preproc_directive] = ACTIONS(3695),
    [anon_sym_DASH] = ACTIONS(3698),
    [anon_sym_PLUS] = ACTIONS(3698),
    [anon_sym___extension__] = ACTIONS(3701),
    [anon_sym_typedef] = ACTIONS(3704),
    [anon_sym_extern] = ACTIONS(3707),
    [anon_sym___attribute__] = ACTIONS(3710),
    [anon_sym___attribute] = ACTIONS(3710),
    [anon_sym_noreturn] = ACTIONS(3713),
    [anon_sym_LBRACK] = ACTIONS(3716),
    [anon_sym___declspec] = ACTIONS(3719),
    [anon_sym___cdecl] = ACTIONS(3722),
    [anon_sym___clrcall] = ACTIONS(3722),
    [anon_sym___stdcall] = ACTIONS(3722),
    [anon_sym___fastcall] = ACTIONS(3722),
    [anon_sym___thiscall] = ACTIONS(3722),
    [anon_sym___vectorcall] = ACTIONS(3722),
    [anon_sym_signed] = ACTIONS(3725),
    [anon_sym_unsigned] = ACTIONS(3725),
    [anon_sym_long] = ACTIONS(3725),
    [anon_sym_short] = ACTIONS(3725),
    [anon_sym_static] = ACTIONS(3707),
    [anon_sym_auto] = ACTIONS(3707),
    [anon_sym_register] = ACTIONS(3707),
    [anon_sym_inline] = ACTIONS(3707),
    [anon_sym___inline] = ACTIONS(3707),
    [anon_sym___inline__] = ACTIONS(3707),
    [anon_sym___forceinline] = ACTIONS(3707),
    [anon_sym_thread_local] = ACTIONS(3707),
    [anon_sym___thread] = ACTIONS(3707),
    [anon_sym_CG_EXTERN] = ACTIONS(3707),
    [anon_sym_CG_INLINE] = ACTIONS(3707),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3707),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3707),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3707),
    [anon_sym_IBOutlet] = ACTIONS(3707),
    [anon_sym_IBInspectable] = ACTIONS(3707),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3707),
    [anon_sym_NS_INLINE] = ACTIONS(3707),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3707),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3707),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3707),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3707),
    [anon_sym_const] = ACTIONS(3713),
    [anon_sym_constexpr] = ACTIONS(3713),
    [anon_sym_volatile] = ACTIONS(3713),
    [anon_sym_restrict] = ACTIONS(3713),
    [anon_sym___restrict__] = ACTIONS(3713),
    [anon_sym__Atomic] = ACTIONS(3713),
    [anon_sym__Noreturn] = ACTIONS(3713),
    [anon_sym__Nonnull] = ACTIONS(3713),
    [anon_sym_nullable] = ACTIONS(3713),
    [anon_sym__Complex] = ACTIONS(3713),
    [anon_sym__Nullable] = ACTIONS(3713),
    [anon_sym__Nullable_result] = ACTIONS(3713),
    [anon_sym__Null_unspecified] = ACTIONS(3713),
    [anon_sym___autoreleasing] = ACTIONS(3713),
    [anon_sym___block] = ACTIONS(3713),
    [anon_sym___bridge] = ACTIONS(3713),
    [anon_sym___bridge_retained] = ACTIONS(3713),
    [anon_sym___bridge_transfer] = ACTIONS(3713),
    [anon_sym___complex] = ACTIONS(3713),
    [anon_sym___const] = ACTIONS(3713),
    [anon_sym___imag] = ACTIONS(3713),
    [anon_sym___kindof] = ACTIONS(3713),
    [anon_sym___nonnull] = ACTIONS(3713),
    [anon_sym___nullable] = ACTIONS(3713),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3713),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3713),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3713),
    [anon_sym___real] = ACTIONS(3713),
    [anon_sym___strong] = ACTIONS(3713),
    [anon_sym___unsafe_unretained] = ACTIONS(3713),
    [anon_sym___unused] = ACTIONS(3713),
    [anon_sym___weak] = ACTIONS(3713),
    [anon_sym_alignas] = ACTIONS(3728),
    [anon_sym__Alignas] = ACTIONS(3728),
    [sym_primitive_type] = ACTIONS(3731),
    [anon_sym_enum] = ACTIONS(3734),
    [anon_sym_struct] = ACTIONS(3737),
    [anon_sym_union] = ACTIONS(3740),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3743),
    [anon_sym___typeof] = ACTIONS(3743),
    [anon_sym_typeof] = ACTIONS(3743),
    [aux_sym_preproc_undef_token1] = ACTIONS(3746),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3749),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3749),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3749),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3749),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3749),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3749),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3749),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3749),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3749),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3752),
    [anon_sym_API_AVAILABLE] = ACTIONS(3752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3752),
    [anon_sym_API_DEPRECATED] = ACTIONS(3752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3752),
    [anon_sym___deprecated_msg] = ACTIONS(3752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3752),
    [anon_sym_ATend] = ACTIONS(3755),
    [anon_sym_AToptional] = ACTIONS(3755),
    [anon_sym_ATrequired] = ACTIONS(3755),
    [anon_sym_ATproperty] = ACTIONS(3757),
    [anon_sym_BOOL] = ACTIONS(3760),
    [anon_sym_IMP] = ACTIONS(3760),
    [anon_sym_SEL] = ACTIONS(3760),
    [anon_sym_Class] = ACTIONS(3760),
    [anon_sym_id] = ACTIONS(3760),
  },
  [STATE(1334)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3763),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1335)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3765),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1336)] = {
    [sym_preproc_def] = STATE(1422),
    [sym_preproc_call] = STATE(1422),
    [sym_preproc_ifdef] = STATE(1422),
    [sym_function_definition] = STATE(1422),
    [sym_declaration] = STATE(1422),
    [sym_type_definition] = STATE(1422),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1422),
    [sym_preproc_undef] = STATE(1422),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1422),
    [sym_method_declaration] = STATE(1422),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1422),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3767),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1337)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3164),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1338)] = {
    [sym_preproc_def] = STATE(1347),
    [sym_preproc_call] = STATE(1347),
    [sym_preproc_ifdef] = STATE(1347),
    [sym_function_definition] = STATE(1347),
    [sym_declaration] = STATE(1347),
    [sym_type_definition] = STATE(1347),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1347),
    [sym_preproc_undef] = STATE(1347),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1347),
    [sym_method_declaration] = STATE(1347),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1347),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3769),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1339)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3771),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1340)] = {
    [sym_preproc_def] = STATE(1348),
    [sym_preproc_call] = STATE(1348),
    [sym_preproc_ifdef] = STATE(1348),
    [sym_function_definition] = STATE(1348),
    [sym_declaration] = STATE(1348),
    [sym_type_definition] = STATE(1348),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1348),
    [sym_preproc_undef] = STATE(1348),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1348),
    [sym_method_declaration] = STATE(1348),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1348),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3773),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1341)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3773),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1342)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3775),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1343)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3777),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1344)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3779),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1345)] = {
    [sym_preproc_def] = STATE(1351),
    [sym_preproc_call] = STATE(1351),
    [sym_preproc_ifdef] = STATE(1351),
    [sym_function_definition] = STATE(1351),
    [sym_declaration] = STATE(1351),
    [sym_type_definition] = STATE(1351),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1351),
    [sym_preproc_undef] = STATE(1351),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1351),
    [sym_method_declaration] = STATE(1351),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1351),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3781),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1346)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3783),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1347)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3785),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1348)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3787),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1349)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3789),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1350)] = {
    [sym_preproc_def] = STATE(1380),
    [sym_preproc_call] = STATE(1380),
    [sym_preproc_ifdef] = STATE(1380),
    [sym_function_definition] = STATE(1380),
    [sym_declaration] = STATE(1380),
    [sym_type_definition] = STATE(1380),
    [sym__declaration_modifiers] = STATE(2887),
    [sym__declaration_specifiers] = STATE(4221),
    [sym_attribute_specifier] = STATE(2887),
    [sym_attribute_declaration] = STATE(2887),
    [sym_ms_declspec_modifier] = STATE(2870),
    [sym_ms_call_modifier] = STATE(2179),
    [sym_storage_class_specifier] = STATE(2887),
    [sym_type_qualifier] = STATE(2887),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3230),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1380),
    [sym_preproc_undef] = STATE(1380),
    [sym_availability_attribute_specifier] = STATE(2887),
    [sym_property_declaration] = STATE(1380),
    [sym_method_declaration] = STATE(1380),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1380),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3791),
    [aux_sym_preproc_if_token1] = ACTIONS(3793),
    [aux_sym_preproc_if_token2] = ACTIONS(3795),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3359),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3359),
    [sym_preproc_directive] = ACTIONS(3361),
    [anon_sym_DASH] = ACTIONS(3797),
    [anon_sym_PLUS] = ACTIONS(3797),
    [anon_sym___extension__] = ACTIONS(3365),
    [anon_sym_typedef] = ACTIONS(3367),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3369),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATproperty] = ACTIONS(3799),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1351)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3801),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1352)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3803),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1353)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3805),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1354)] = {
    [sym_preproc_def] = STATE(1360),
    [sym_preproc_call] = STATE(1360),
    [sym_preproc_ifdef] = STATE(1360),
    [sym_function_definition] = STATE(1360),
    [sym_declaration] = STATE(1360),
    [sym_type_definition] = STATE(1360),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1360),
    [sym_preproc_undef] = STATE(1360),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1360),
    [sym_method_declaration] = STATE(1360),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1360),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3122),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1355)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3122),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1356)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3807),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1357)] = {
    [sym_preproc_def] = STATE(1376),
    [sym_preproc_call] = STATE(1376),
    [sym_preproc_ifdef] = STATE(1376),
    [sym_function_definition] = STATE(1376),
    [sym_declaration] = STATE(1376),
    [sym_type_definition] = STATE(1376),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1376),
    [sym_preproc_undef] = STATE(1376),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1376),
    [sym_method_declaration] = STATE(1376),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1376),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3156),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1358)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3156),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1359)] = {
    [sym_preproc_def] = STATE(1366),
    [sym_preproc_call] = STATE(1366),
    [sym_preproc_ifdef] = STATE(1366),
    [sym_function_definition] = STATE(1366),
    [sym_declaration] = STATE(1366),
    [sym_type_definition] = STATE(1366),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1366),
    [sym_preproc_undef] = STATE(1366),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1366),
    [sym_method_declaration] = STATE(1366),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1366),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3667),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1360)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3667),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1361)] = {
    [sym_preproc_def] = STATE(1369),
    [sym_preproc_call] = STATE(1369),
    [sym_preproc_ifdef] = STATE(1369),
    [sym_function_definition] = STATE(1369),
    [sym_declaration] = STATE(1369),
    [sym_type_definition] = STATE(1369),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1369),
    [sym_preproc_undef] = STATE(1369),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1369),
    [sym_method_declaration] = STATE(1369),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1369),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3809),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1362)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3811),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1363)] = {
    [sym_preproc_def] = STATE(1372),
    [sym_preproc_call] = STATE(1372),
    [sym_preproc_ifdef] = STATE(1372),
    [sym_function_definition] = STATE(1372),
    [sym_declaration] = STATE(1372),
    [sym_type_definition] = STATE(1372),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1372),
    [sym_preproc_undef] = STATE(1372),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1372),
    [sym_method_declaration] = STATE(1372),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1372),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3813),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1364)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3815),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1365)] = {
    [sym_preproc_def] = STATE(1373),
    [sym_preproc_call] = STATE(1373),
    [sym_preproc_ifdef] = STATE(1373),
    [sym_function_definition] = STATE(1373),
    [sym_declaration] = STATE(1373),
    [sym_type_definition] = STATE(1373),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1373),
    [sym_preproc_undef] = STATE(1373),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1373),
    [sym_method_declaration] = STATE(1373),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1373),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3817),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1366)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3817),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1367)] = {
    [sym_preproc_def] = STATE(1374),
    [sym_preproc_call] = STATE(1374),
    [sym_preproc_ifdef] = STATE(1374),
    [sym_function_definition] = STATE(1374),
    [sym_declaration] = STATE(1374),
    [sym_type_definition] = STATE(1374),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1374),
    [sym_preproc_undef] = STATE(1374),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1374),
    [sym_method_declaration] = STATE(1374),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1374),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3819),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1368)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3821),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1369)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3823),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1370)] = {
    [sym_preproc_def] = STATE(1377),
    [sym_preproc_call] = STATE(1377),
    [sym_preproc_ifdef] = STATE(1377),
    [sym_function_definition] = STATE(1377),
    [sym_declaration] = STATE(1377),
    [sym_type_definition] = STATE(1377),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1377),
    [sym_preproc_undef] = STATE(1377),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1377),
    [sym_method_declaration] = STATE(1377),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1377),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3825),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1371)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3827),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1372)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3829),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1373)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3831),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1374)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3833),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1375)] = {
    [sym_preproc_def] = STATE(1387),
    [sym_preproc_call] = STATE(1387),
    [sym_preproc_ifdef] = STATE(1387),
    [sym_function_definition] = STATE(1387),
    [sym_declaration] = STATE(1387),
    [sym_type_definition] = STATE(1387),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1387),
    [sym_preproc_undef] = STATE(1387),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1387),
    [sym_method_declaration] = STATE(1387),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1387),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3571),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1376)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3571),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1377)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3835),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1378)] = {
    [sym_preproc_def] = STATE(1394),
    [sym_preproc_call] = STATE(1394),
    [sym_preproc_ifdef] = STATE(1394),
    [sym_function_definition] = STATE(1394),
    [sym_declaration] = STATE(1394),
    [sym_type_definition] = STATE(1394),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1394),
    [sym_preproc_undef] = STATE(1394),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1394),
    [sym_method_declaration] = STATE(1394),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1394),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3837),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1379)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3839),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1380)] = {
    [sym_preproc_def] = STATE(1380),
    [sym_preproc_call] = STATE(1380),
    [sym_preproc_ifdef] = STATE(1380),
    [sym_function_definition] = STATE(1380),
    [sym_declaration] = STATE(1380),
    [sym_type_definition] = STATE(1380),
    [sym__declaration_modifiers] = STATE(2887),
    [sym__declaration_specifiers] = STATE(4221),
    [sym_attribute_specifier] = STATE(2887),
    [sym_attribute_declaration] = STATE(2887),
    [sym_ms_declspec_modifier] = STATE(2870),
    [sym_ms_call_modifier] = STATE(2179),
    [sym_storage_class_specifier] = STATE(2887),
    [sym_type_qualifier] = STATE(2887),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3230),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1380),
    [sym_preproc_undef] = STATE(1380),
    [sym_availability_attribute_specifier] = STATE(2887),
    [sym_property_declaration] = STATE(1380),
    [sym_method_declaration] = STATE(1380),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1380),
    [sym_identifier] = ACTIONS(3583),
    [aux_sym_preproc_def_token1] = ACTIONS(3841),
    [aux_sym_preproc_if_token1] = ACTIONS(3844),
    [aux_sym_preproc_if_token2] = ACTIONS(3592),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3847),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3847),
    [sym_preproc_directive] = ACTIONS(3850),
    [anon_sym_DASH] = ACTIONS(3853),
    [anon_sym_PLUS] = ACTIONS(3853),
    [anon_sym___extension__] = ACTIONS(3856),
    [anon_sym_typedef] = ACTIONS(3859),
    [anon_sym_extern] = ACTIONS(3609),
    [anon_sym___attribute__] = ACTIONS(3612),
    [anon_sym___attribute] = ACTIONS(3612),
    [anon_sym_noreturn] = ACTIONS(3615),
    [anon_sym_LBRACK] = ACTIONS(3618),
    [anon_sym___declspec] = ACTIONS(3621),
    [anon_sym___cdecl] = ACTIONS(3624),
    [anon_sym___clrcall] = ACTIONS(3624),
    [anon_sym___stdcall] = ACTIONS(3624),
    [anon_sym___fastcall] = ACTIONS(3624),
    [anon_sym___thiscall] = ACTIONS(3624),
    [anon_sym___vectorcall] = ACTIONS(3624),
    [anon_sym_signed] = ACTIONS(3627),
    [anon_sym_unsigned] = ACTIONS(3627),
    [anon_sym_long] = ACTIONS(3627),
    [anon_sym_short] = ACTIONS(3627),
    [anon_sym_static] = ACTIONS(3609),
    [anon_sym_auto] = ACTIONS(3609),
    [anon_sym_register] = ACTIONS(3609),
    [anon_sym_inline] = ACTIONS(3609),
    [anon_sym___inline] = ACTIONS(3609),
    [anon_sym___inline__] = ACTIONS(3609),
    [anon_sym___forceinline] = ACTIONS(3609),
    [anon_sym_thread_local] = ACTIONS(3609),
    [anon_sym___thread] = ACTIONS(3609),
    [anon_sym_CG_EXTERN] = ACTIONS(3609),
    [anon_sym_CG_INLINE] = ACTIONS(3609),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3609),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3609),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3609),
    [anon_sym_IBOutlet] = ACTIONS(3609),
    [anon_sym_IBInspectable] = ACTIONS(3609),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3609),
    [anon_sym_NS_INLINE] = ACTIONS(3609),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3609),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3609),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3609),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3609),
    [anon_sym_const] = ACTIONS(3615),
    [anon_sym_constexpr] = ACTIONS(3615),
    [anon_sym_volatile] = ACTIONS(3615),
    [anon_sym_restrict] = ACTIONS(3615),
    [anon_sym___restrict__] = ACTIONS(3615),
    [anon_sym__Atomic] = ACTIONS(3615),
    [anon_sym__Noreturn] = ACTIONS(3615),
    [anon_sym__Nonnull] = ACTIONS(3615),
    [anon_sym_nullable] = ACTIONS(3615),
    [anon_sym__Complex] = ACTIONS(3615),
    [anon_sym__Nullable] = ACTIONS(3615),
    [anon_sym__Nullable_result] = ACTIONS(3615),
    [anon_sym__Null_unspecified] = ACTIONS(3615),
    [anon_sym___autoreleasing] = ACTIONS(3615),
    [anon_sym___block] = ACTIONS(3615),
    [anon_sym___bridge] = ACTIONS(3615),
    [anon_sym___bridge_retained] = ACTIONS(3615),
    [anon_sym___bridge_transfer] = ACTIONS(3615),
    [anon_sym___complex] = ACTIONS(3615),
    [anon_sym___const] = ACTIONS(3615),
    [anon_sym___imag] = ACTIONS(3615),
    [anon_sym___kindof] = ACTIONS(3615),
    [anon_sym___nonnull] = ACTIONS(3615),
    [anon_sym___nullable] = ACTIONS(3615),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3615),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3615),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3615),
    [anon_sym___real] = ACTIONS(3615),
    [anon_sym___strong] = ACTIONS(3615),
    [anon_sym___unsafe_unretained] = ACTIONS(3615),
    [anon_sym___unused] = ACTIONS(3615),
    [anon_sym___weak] = ACTIONS(3615),
    [anon_sym_alignas] = ACTIONS(3630),
    [anon_sym__Alignas] = ACTIONS(3630),
    [sym_primitive_type] = ACTIONS(3633),
    [anon_sym_enum] = ACTIONS(3636),
    [anon_sym_struct] = ACTIONS(3639),
    [anon_sym_union] = ACTIONS(3642),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3645),
    [anon_sym___typeof] = ACTIONS(3645),
    [anon_sym_typeof] = ACTIONS(3645),
    [aux_sym_preproc_undef_token1] = ACTIONS(3862),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3651),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3651),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3651),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3651),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3651),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3651),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3651),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3651),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3651),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3654),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3654),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3654),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3654),
    [anon_sym_API_AVAILABLE] = ACTIONS(3654),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3654),
    [anon_sym_API_DEPRECATED] = ACTIONS(3654),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3654),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3654),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3654),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3654),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3654),
    [anon_sym___deprecated_msg] = ACTIONS(3654),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3654),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3654),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3654),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3654),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3654),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3654),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3654),
    [anon_sym_ATproperty] = ACTIONS(3865),
    [anon_sym_BOOL] = ACTIONS(3660),
    [anon_sym_IMP] = ACTIONS(3660),
    [anon_sym_SEL] = ACTIONS(3660),
    [anon_sym_Class] = ACTIONS(3660),
    [anon_sym_id] = ACTIONS(3660),
  },
  [STATE(1381)] = {
    [sym_preproc_def] = STATE(1390),
    [sym_preproc_call] = STATE(1390),
    [sym_preproc_ifdef] = STATE(1390),
    [sym_function_definition] = STATE(1390),
    [sym_declaration] = STATE(1390),
    [sym_type_definition] = STATE(1390),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1390),
    [sym_preproc_undef] = STATE(1390),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1390),
    [sym_method_declaration] = STATE(1390),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1390),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3168),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1382)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3168),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1383)] = {
    [sym_preproc_def] = STATE(1350),
    [sym_preproc_call] = STATE(1350),
    [sym_preproc_ifdef] = STATE(1350),
    [sym_function_definition] = STATE(1350),
    [sym_declaration] = STATE(1350),
    [sym_type_definition] = STATE(1350),
    [sym__declaration_modifiers] = STATE(2887),
    [sym__declaration_specifiers] = STATE(4221),
    [sym_attribute_specifier] = STATE(2887),
    [sym_attribute_declaration] = STATE(2887),
    [sym_ms_declspec_modifier] = STATE(2870),
    [sym_ms_call_modifier] = STATE(2179),
    [sym_storage_class_specifier] = STATE(2887),
    [sym_type_qualifier] = STATE(2887),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3230),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1350),
    [sym_preproc_undef] = STATE(1350),
    [sym_availability_attribute_specifier] = STATE(2887),
    [sym_property_declaration] = STATE(1350),
    [sym_method_declaration] = STATE(1350),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_preproc_if_in_interface_declaration_repeat1] = STATE(1350),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3791),
    [aux_sym_preproc_if_token1] = ACTIONS(3793),
    [aux_sym_preproc_if_token2] = ACTIONS(3868),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3359),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3359),
    [sym_preproc_directive] = ACTIONS(3361),
    [anon_sym_DASH] = ACTIONS(3797),
    [anon_sym_PLUS] = ACTIONS(3797),
    [anon_sym___extension__] = ACTIONS(3365),
    [anon_sym_typedef] = ACTIONS(3367),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3369),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATproperty] = ACTIONS(3799),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1384)] = {
    [sym_preproc_def] = STATE(1406),
    [sym_preproc_call] = STATE(1406),
    [sym_preproc_ifdef] = STATE(1406),
    [sym_function_definition] = STATE(1406),
    [sym_declaration] = STATE(1406),
    [sym_type_definition] = STATE(1406),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1406),
    [sym_preproc_undef] = STATE(1406),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1406),
    [sym_method_declaration] = STATE(1406),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1406),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3870),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1385)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3872),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1386)] = {
    [sym_preproc_def] = STATE(1407),
    [sym_preproc_call] = STATE(1407),
    [sym_preproc_ifdef] = STATE(1407),
    [sym_function_definition] = STATE(1407),
    [sym_declaration] = STATE(1407),
    [sym_type_definition] = STATE(1407),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1407),
    [sym_preproc_undef] = STATE(1407),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1407),
    [sym_method_declaration] = STATE(1407),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1407),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3874),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1387)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3874),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1388)] = {
    [sym_preproc_def] = STATE(1413),
    [sym_preproc_call] = STATE(1413),
    [sym_preproc_ifdef] = STATE(1413),
    [sym_function_definition] = STATE(1413),
    [sym_declaration] = STATE(1413),
    [sym_type_definition] = STATE(1413),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1413),
    [sym_preproc_undef] = STATE(1413),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1413),
    [sym_method_declaration] = STATE(1413),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1413),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3876),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1389)] = {
    [sym_preproc_def] = STATE(1400),
    [sym_preproc_call] = STATE(1400),
    [sym_preproc_ifdef] = STATE(1400),
    [sym_function_definition] = STATE(1400),
    [sym_declaration] = STATE(1400),
    [sym_type_definition] = STATE(1400),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1400),
    [sym_preproc_undef] = STATE(1400),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1400),
    [sym_method_declaration] = STATE(1400),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1400),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3575),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1390)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3575),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1391)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3878),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1392)] = {
    [sym_preproc_def] = STATE(1403),
    [sym_preproc_call] = STATE(1403),
    [sym_preproc_ifdef] = STATE(1403),
    [sym_function_definition] = STATE(1403),
    [sym_declaration] = STATE(1403),
    [sym_type_definition] = STATE(1403),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1403),
    [sym_preproc_undef] = STATE(1403),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1403),
    [sym_method_declaration] = STATE(1403),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1403),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3880),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1393)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3882),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1394)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3884),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1395)] = {
    [sym_preproc_def] = STATE(1419),
    [sym_preproc_call] = STATE(1419),
    [sym_preproc_ifdef] = STATE(1419),
    [sym_function_definition] = STATE(1419),
    [sym_declaration] = STATE(1419),
    [sym_type_definition] = STATE(1419),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1419),
    [sym_preproc_undef] = STATE(1419),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1419),
    [sym_method_declaration] = STATE(1419),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1419),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3681),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1396)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3681),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1397)] = {
    [sym_preproc_def] = STATE(1410),
    [sym_preproc_call] = STATE(1410),
    [sym_preproc_ifdef] = STATE(1410),
    [sym_function_definition] = STATE(1410),
    [sym_declaration] = STATE(1410),
    [sym_type_definition] = STATE(1410),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1410),
    [sym_preproc_undef] = STATE(1410),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1410),
    [sym_method_declaration] = STATE(1410),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1410),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3886),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1398)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3888),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1399)] = {
    [sym_preproc_def] = STATE(1411),
    [sym_preproc_call] = STATE(1411),
    [sym_preproc_ifdef] = STATE(1411),
    [sym_function_definition] = STATE(1411),
    [sym_declaration] = STATE(1411),
    [sym_type_definition] = STATE(1411),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1411),
    [sym_preproc_undef] = STATE(1411),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1411),
    [sym_method_declaration] = STATE(1411),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1411),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3890),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1400)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3890),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1401)] = {
    [sym_preproc_def] = STATE(1412),
    [sym_preproc_call] = STATE(1412),
    [sym_preproc_ifdef] = STATE(1412),
    [sym_function_definition] = STATE(1412),
    [sym_declaration] = STATE(1412),
    [sym_type_definition] = STATE(1412),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1412),
    [sym_preproc_undef] = STATE(1412),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1412),
    [sym_method_declaration] = STATE(1412),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1412),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3892),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1402)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3894),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1403)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3896),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1404)] = {
    [sym_preproc_def] = STATE(1415),
    [sym_preproc_call] = STATE(1415),
    [sym_preproc_ifdef] = STATE(1415),
    [sym_function_definition] = STATE(1415),
    [sym_declaration] = STATE(1415),
    [sym_type_definition] = STATE(1415),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1415),
    [sym_preproc_undef] = STATE(1415),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1415),
    [sym_method_declaration] = STATE(1415),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1415),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3898),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1405)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3900),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1406)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3902),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1407)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3904),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1408)] = {
    [sym_preproc_def] = STATE(1414),
    [sym_preproc_call] = STATE(1414),
    [sym_preproc_ifdef] = STATE(1414),
    [sym_function_definition] = STATE(1414),
    [sym_declaration] = STATE(1414),
    [sym_type_definition] = STATE(1414),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1414),
    [sym_preproc_undef] = STATE(1414),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1414),
    [sym_method_declaration] = STATE(1414),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1414),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3906),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1409)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3908),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1410)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3910),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1411)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3912),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1412)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3914),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1413)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3916),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1414)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3918),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1415)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3920),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1416)] = {
    [sym_preproc_def] = STATE(1335),
    [sym_preproc_call] = STATE(1335),
    [sym_preproc_ifdef] = STATE(1335),
    [sym_function_definition] = STATE(1335),
    [sym_declaration] = STATE(1335),
    [sym_type_definition] = STATE(1335),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1335),
    [sym_preproc_undef] = STATE(1335),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1335),
    [sym_method_declaration] = STATE(1335),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1335),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3922),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1417)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3924),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1418)] = {
    [sym_preproc_def] = STATE(1352),
    [sym_preproc_call] = STATE(1352),
    [sym_preproc_ifdef] = STATE(1352),
    [sym_function_definition] = STATE(1352),
    [sym_declaration] = STATE(1352),
    [sym_type_definition] = STATE(1352),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1352),
    [sym_preproc_undef] = STATE(1352),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1352),
    [sym_method_declaration] = STATE(1352),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1352),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3926),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1419)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3926),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1420)] = {
    [sym_preproc_def] = STATE(1353),
    [sym_preproc_call] = STATE(1353),
    [sym_preproc_ifdef] = STATE(1353),
    [sym_function_definition] = STATE(1353),
    [sym_declaration] = STATE(1353),
    [sym_type_definition] = STATE(1353),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1353),
    [sym_preproc_undef] = STATE(1353),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1353),
    [sym_method_declaration] = STATE(1353),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1353),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3928),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1421)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3930),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1422)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3932),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1423)] = {
    [sym_preproc_def] = STATE(1356),
    [sym_preproc_call] = STATE(1356),
    [sym_preproc_ifdef] = STATE(1356),
    [sym_function_definition] = STATE(1356),
    [sym_declaration] = STATE(1356),
    [sym_type_definition] = STATE(1356),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1356),
    [sym_preproc_undef] = STATE(1356),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1356),
    [sym_method_declaration] = STATE(1356),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1356),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3934),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1424)] = {
    [sym_preproc_def] = STATE(1428),
    [sym_preproc_call] = STATE(1428),
    [sym_preproc_ifdef] = STATE(1428),
    [sym_function_definition] = STATE(1428),
    [sym_declaration] = STATE(1428),
    [sym_type_definition] = STATE(1428),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1428),
    [sym_preproc_undef] = STATE(1428),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1428),
    [sym_method_declaration] = STATE(1428),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1428),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3152),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1425)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3152),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1426)] = {
    [sym_preproc_def] = STATE(1396),
    [sym_preproc_call] = STATE(1396),
    [sym_preproc_ifdef] = STATE(1396),
    [sym_function_definition] = STATE(1396),
    [sym_declaration] = STATE(1396),
    [sym_type_definition] = STATE(1396),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1396),
    [sym_preproc_undef] = STATE(1396),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1396),
    [sym_method_declaration] = STATE(1396),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1396),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3164),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1427)] = {
    [sym_preproc_def] = STATE(1341),
    [sym_preproc_call] = STATE(1341),
    [sym_preproc_ifdef] = STATE(1341),
    [sym_function_definition] = STATE(1341),
    [sym_declaration] = STATE(1341),
    [sym_type_definition] = STATE(1341),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1341),
    [sym_preproc_undef] = STATE(1341),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1341),
    [sym_method_declaration] = STATE(1341),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1341),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3673),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1428)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3673),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1429)] = {
    [sym_preproc_def] = STATE(1343),
    [sym_preproc_call] = STATE(1343),
    [sym_preproc_ifdef] = STATE(1343),
    [sym_function_definition] = STATE(1343),
    [sym_declaration] = STATE(1343),
    [sym_type_definition] = STATE(1343),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1343),
    [sym_preproc_undef] = STATE(1343),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1343),
    [sym_method_declaration] = STATE(1343),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1343),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3936),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1430)] = {
    [sym_preproc_def] = STATE(1333),
    [sym_preproc_call] = STATE(1333),
    [sym_preproc_ifdef] = STATE(1333),
    [sym_function_definition] = STATE(1333),
    [sym_declaration] = STATE(1333),
    [sym_type_definition] = STATE(1333),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1333),
    [sym_preproc_undef] = STATE(1333),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1333),
    [sym_method_declaration] = STATE(1333),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1333),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3938),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1431)] = {
    [sym_preproc_def] = STATE(1349),
    [sym_preproc_call] = STATE(1349),
    [sym_preproc_ifdef] = STATE(1349),
    [sym_function_definition] = STATE(1349),
    [sym_declaration] = STATE(1349),
    [sym_type_definition] = STATE(1349),
    [sym__declaration_modifiers] = STATE(2894),
    [sym__declaration_specifiers] = STATE(4218),
    [sym_attribute_specifier] = STATE(2894),
    [sym_attribute_declaration] = STATE(2894),
    [sym_ms_declspec_modifier] = STATE(2861),
    [sym_ms_call_modifier] = STATE(2127),
    [sym_storage_class_specifier] = STATE(2894),
    [sym_type_qualifier] = STATE(2894),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(3227),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_preproc_if_in_interface_declaration] = STATE(1349),
    [sym_preproc_undef] = STATE(1349),
    [sym_availability_attribute_specifier] = STATE(2894),
    [sym_property_declaration] = STATE(1349),
    [sym_method_declaration] = STATE(1349),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_class_interface_repeat1] = STATE(1349),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3040),
    [aux_sym_preproc_if_token1] = ACTIONS(3044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3000),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3046),
    [anon_sym_PLUS] = ACTIONS(3046),
    [anon_sym___extension__] = ACTIONS(3010),
    [anon_sym_typedef] = ACTIONS(3012),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [aux_sym_preproc_undef_token1] = ACTIONS(3022),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATend] = ACTIONS(3940),
    [anon_sym_ATproperty] = ACTIONS(3056),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1432)] = {
    [sym_string_literal] = STATE(1433),
    [aux_sym_concatenated_string_repeat1] = STATE(1433),
    [sym_identifier] = ACTIONS(3942),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3944),
    [anon_sym_COMMA] = ACTIONS(3944),
    [anon_sym_LPAREN2] = ACTIONS(3944),
    [anon_sym_DASH] = ACTIONS(3946),
    [anon_sym_PLUS] = ACTIONS(3946),
    [anon_sym_STAR] = ACTIONS(3946),
    [anon_sym_SLASH] = ACTIONS(3946),
    [anon_sym_PERCENT] = ACTIONS(3946),
    [anon_sym_PIPE_PIPE] = ACTIONS(3944),
    [anon_sym_AMP_AMP] = ACTIONS(3944),
    [anon_sym_PIPE] = ACTIONS(3946),
    [anon_sym_CARET] = ACTIONS(3946),
    [anon_sym_AMP] = ACTIONS(3946),
    [anon_sym_EQ_EQ] = ACTIONS(3944),
    [anon_sym_BANG_EQ] = ACTIONS(3944),
    [anon_sym_GT] = ACTIONS(3946),
    [anon_sym_GT_EQ] = ACTIONS(3944),
    [anon_sym_LT_EQ] = ACTIONS(3944),
    [anon_sym_LT] = ACTIONS(3946),
    [anon_sym_LT_LT] = ACTIONS(3946),
    [anon_sym_GT_GT] = ACTIONS(3946),
    [anon_sym_SEMI] = ACTIONS(3944),
    [anon_sym___extension__] = ACTIONS(3946),
    [anon_sym_extern] = ACTIONS(3946),
    [anon_sym___attribute__] = ACTIONS(3946),
    [anon_sym___attribute] = ACTIONS(3946),
    [anon_sym_noreturn] = ACTIONS(3946),
    [anon_sym_LBRACK] = ACTIONS(3944),
    [anon_sym___declspec] = ACTIONS(3946),
    [anon_sym_LBRACE] = ACTIONS(3944),
    [anon_sym_signed] = ACTIONS(3946),
    [anon_sym_unsigned] = ACTIONS(3946),
    [anon_sym_long] = ACTIONS(3946),
    [anon_sym_short] = ACTIONS(3946),
    [anon_sym_static] = ACTIONS(3946),
    [anon_sym_EQ] = ACTIONS(3946),
    [anon_sym_ATautoreleasepool] = ACTIONS(3944),
    [anon_sym_auto] = ACTIONS(3946),
    [anon_sym_register] = ACTIONS(3946),
    [anon_sym_inline] = ACTIONS(3946),
    [anon_sym___inline] = ACTIONS(3946),
    [anon_sym___inline__] = ACTIONS(3946),
    [anon_sym___forceinline] = ACTIONS(3946),
    [anon_sym_thread_local] = ACTIONS(3946),
    [anon_sym___thread] = ACTIONS(3946),
    [anon_sym_CG_EXTERN] = ACTIONS(3946),
    [anon_sym_CG_INLINE] = ACTIONS(3946),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3946),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3946),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3946),
    [anon_sym_IBOutlet] = ACTIONS(3946),
    [anon_sym_IBInspectable] = ACTIONS(3946),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3946),
    [anon_sym_NS_INLINE] = ACTIONS(3946),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3946),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3946),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3946),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3946),
    [anon_sym_const] = ACTIONS(3946),
    [anon_sym_constexpr] = ACTIONS(3946),
    [anon_sym_volatile] = ACTIONS(3946),
    [anon_sym_restrict] = ACTIONS(3946),
    [anon_sym___restrict__] = ACTIONS(3946),
    [anon_sym__Atomic] = ACTIONS(3946),
    [anon_sym__Noreturn] = ACTIONS(3946),
    [anon_sym__Nonnull] = ACTIONS(3946),
    [anon_sym_nullable] = ACTIONS(3946),
    [anon_sym__Complex] = ACTIONS(3946),
    [anon_sym__Nullable] = ACTIONS(3946),
    [anon_sym__Nullable_result] = ACTIONS(3946),
    [anon_sym__Null_unspecified] = ACTIONS(3946),
    [anon_sym___autoreleasing] = ACTIONS(3946),
    [anon_sym___block] = ACTIONS(3946),
    [anon_sym___bridge] = ACTIONS(3946),
    [anon_sym___bridge_retained] = ACTIONS(3946),
    [anon_sym___bridge_transfer] = ACTIONS(3946),
    [anon_sym___complex] = ACTIONS(3946),
    [anon_sym___const] = ACTIONS(3946),
    [anon_sym___imag] = ACTIONS(3946),
    [anon_sym___kindof] = ACTIONS(3946),
    [anon_sym___nonnull] = ACTIONS(3946),
    [anon_sym___nullable] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3946),
    [anon_sym___real] = ACTIONS(3946),
    [anon_sym___strong] = ACTIONS(3946),
    [anon_sym___unsafe_unretained] = ACTIONS(3946),
    [anon_sym___unused] = ACTIONS(3946),
    [anon_sym___weak] = ACTIONS(3946),
    [anon_sym_alignas] = ACTIONS(3946),
    [anon_sym__Alignas] = ACTIONS(3946),
    [sym_primitive_type] = ACTIONS(3946),
    [anon_sym_enum] = ACTIONS(3946),
    [anon_sym_struct] = ACTIONS(3946),
    [anon_sym_union] = ACTIONS(3946),
    [anon_sym_QMARK] = ACTIONS(3944),
    [anon_sym_STAR_EQ] = ACTIONS(3944),
    [anon_sym_SLASH_EQ] = ACTIONS(3944),
    [anon_sym_PERCENT_EQ] = ACTIONS(3944),
    [anon_sym_PLUS_EQ] = ACTIONS(3944),
    [anon_sym_DASH_EQ] = ACTIONS(3944),
    [anon_sym_LT_LT_EQ] = ACTIONS(3944),
    [anon_sym_GT_GT_EQ] = ACTIONS(3944),
    [anon_sym_AMP_EQ] = ACTIONS(3944),
    [anon_sym_CARET_EQ] = ACTIONS(3944),
    [anon_sym_PIPE_EQ] = ACTIONS(3944),
    [anon_sym_DASH_DASH] = ACTIONS(3944),
    [anon_sym_PLUS_PLUS] = ACTIONS(3944),
    [anon_sym_DOT] = ACTIONS(3946),
    [anon_sym_DASH_GT] = ACTIONS(3944),
    [anon_sym_AT] = ACTIONS(3948),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3946),
    [anon_sym___typeof] = ACTIONS(3946),
    [anon_sym_typeof] = ACTIONS(3946),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3946),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3946),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3946),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3946),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3946),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3946),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3946),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_API_AVAILABLE] = ACTIONS(3946),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_API_DEPRECATED] = ACTIONS(3946),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3946),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym___deprecated_msg] = ACTIONS(3946),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3946),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3946),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3946),
    [anon_sym_BOOL] = ACTIONS(3946),
    [anon_sym_IMP] = ACTIONS(3946),
    [anon_sym_SEL] = ACTIONS(3946),
    [anon_sym_Class] = ACTIONS(3946),
    [anon_sym_id] = ACTIONS(3946),
  },
  [STATE(1433)] = {
    [sym_string_literal] = STATE(1433),
    [aux_sym_concatenated_string_repeat1] = STATE(1433),
    [sym_identifier] = ACTIONS(3950),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_LPAREN2] = ACTIONS(3953),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_STAR] = ACTIONS(3955),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_PERCENT] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3953),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3955),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_EQ_EQ] = ACTIONS(3953),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_LT_EQ] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3955),
    [anon_sym_GT_GT] = ACTIONS(3955),
    [anon_sym_SEMI] = ACTIONS(3953),
    [anon_sym___extension__] = ACTIONS(3955),
    [anon_sym_extern] = ACTIONS(3955),
    [anon_sym___attribute__] = ACTIONS(3955),
    [anon_sym___attribute] = ACTIONS(3955),
    [anon_sym_noreturn] = ACTIONS(3955),
    [anon_sym_LBRACK] = ACTIONS(3953),
    [anon_sym___declspec] = ACTIONS(3955),
    [anon_sym_LBRACE] = ACTIONS(3953),
    [anon_sym_signed] = ACTIONS(3955),
    [anon_sym_unsigned] = ACTIONS(3955),
    [anon_sym_long] = ACTIONS(3955),
    [anon_sym_short] = ACTIONS(3955),
    [anon_sym_static] = ACTIONS(3955),
    [anon_sym_EQ] = ACTIONS(3955),
    [anon_sym_ATautoreleasepool] = ACTIONS(3953),
    [anon_sym_auto] = ACTIONS(3955),
    [anon_sym_register] = ACTIONS(3955),
    [anon_sym_inline] = ACTIONS(3955),
    [anon_sym___inline] = ACTIONS(3955),
    [anon_sym___inline__] = ACTIONS(3955),
    [anon_sym___forceinline] = ACTIONS(3955),
    [anon_sym_thread_local] = ACTIONS(3955),
    [anon_sym___thread] = ACTIONS(3955),
    [anon_sym_CG_EXTERN] = ACTIONS(3955),
    [anon_sym_CG_INLINE] = ACTIONS(3955),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3955),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3955),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3955),
    [anon_sym_IBOutlet] = ACTIONS(3955),
    [anon_sym_IBInspectable] = ACTIONS(3955),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3955),
    [anon_sym_NS_INLINE] = ACTIONS(3955),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3955),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3955),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3955),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3955),
    [anon_sym_const] = ACTIONS(3955),
    [anon_sym_constexpr] = ACTIONS(3955),
    [anon_sym_volatile] = ACTIONS(3955),
    [anon_sym_restrict] = ACTIONS(3955),
    [anon_sym___restrict__] = ACTIONS(3955),
    [anon_sym__Atomic] = ACTIONS(3955),
    [anon_sym__Noreturn] = ACTIONS(3955),
    [anon_sym__Nonnull] = ACTIONS(3955),
    [anon_sym_nullable] = ACTIONS(3955),
    [anon_sym__Complex] = ACTIONS(3955),
    [anon_sym__Nullable] = ACTIONS(3955),
    [anon_sym__Nullable_result] = ACTIONS(3955),
    [anon_sym__Null_unspecified] = ACTIONS(3955),
    [anon_sym___autoreleasing] = ACTIONS(3955),
    [anon_sym___block] = ACTIONS(3955),
    [anon_sym___bridge] = ACTIONS(3955),
    [anon_sym___bridge_retained] = ACTIONS(3955),
    [anon_sym___bridge_transfer] = ACTIONS(3955),
    [anon_sym___complex] = ACTIONS(3955),
    [anon_sym___const] = ACTIONS(3955),
    [anon_sym___imag] = ACTIONS(3955),
    [anon_sym___kindof] = ACTIONS(3955),
    [anon_sym___nonnull] = ACTIONS(3955),
    [anon_sym___nullable] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3955),
    [anon_sym___real] = ACTIONS(3955),
    [anon_sym___strong] = ACTIONS(3955),
    [anon_sym___unsafe_unretained] = ACTIONS(3955),
    [anon_sym___unused] = ACTIONS(3955),
    [anon_sym___weak] = ACTIONS(3955),
    [anon_sym_alignas] = ACTIONS(3955),
    [anon_sym__Alignas] = ACTIONS(3955),
    [sym_primitive_type] = ACTIONS(3955),
    [anon_sym_enum] = ACTIONS(3955),
    [anon_sym_struct] = ACTIONS(3955),
    [anon_sym_union] = ACTIONS(3955),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_STAR_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_PERCENT_EQ] = ACTIONS(3953),
    [anon_sym_PLUS_EQ] = ACTIONS(3953),
    [anon_sym_DASH_EQ] = ACTIONS(3953),
    [anon_sym_LT_LT_EQ] = ACTIONS(3953),
    [anon_sym_GT_GT_EQ] = ACTIONS(3953),
    [anon_sym_AMP_EQ] = ACTIONS(3953),
    [anon_sym_CARET_EQ] = ACTIONS(3953),
    [anon_sym_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_AT] = ACTIONS(3957),
    [anon_sym_DQUOTE] = ACTIONS(3960),
    [anon_sym_L_DQUOTE] = ACTIONS(3960),
    [anon_sym_u_DQUOTE] = ACTIONS(3960),
    [anon_sym_U_DQUOTE] = ACTIONS(3960),
    [anon_sym_u8_DQUOTE] = ACTIONS(3960),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3955),
    [anon_sym___typeof] = ACTIONS(3955),
    [anon_sym_typeof] = ACTIONS(3955),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3955),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3955),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3955),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3955),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3955),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3955),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3955),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_API_AVAILABLE] = ACTIONS(3955),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_API_DEPRECATED] = ACTIONS(3955),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3955),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym___deprecated_msg] = ACTIONS(3955),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3955),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3955),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3955),
    [anon_sym_BOOL] = ACTIONS(3955),
    [anon_sym_IMP] = ACTIONS(3955),
    [anon_sym_SEL] = ACTIONS(3955),
    [anon_sym_Class] = ACTIONS(3955),
    [anon_sym_id] = ACTIONS(3955),
  },
  [STATE(1434)] = {
    [sym_string_literal] = STATE(1432),
    [aux_sym_concatenated_string_repeat1] = STATE(1432),
    [sym_identifier] = ACTIONS(3963),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_LPAREN2] = ACTIONS(3965),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_STAR] = ACTIONS(3967),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_PERCENT] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3965),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3967),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_EQ_EQ] = ACTIONS(3965),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_LT_EQ] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3967),
    [anon_sym_GT_GT] = ACTIONS(3967),
    [anon_sym_SEMI] = ACTIONS(3965),
    [anon_sym___extension__] = ACTIONS(3967),
    [anon_sym_extern] = ACTIONS(3967),
    [anon_sym___attribute__] = ACTIONS(3967),
    [anon_sym___attribute] = ACTIONS(3967),
    [anon_sym_noreturn] = ACTIONS(3967),
    [anon_sym_LBRACK] = ACTIONS(3965),
    [anon_sym___declspec] = ACTIONS(3967),
    [anon_sym_LBRACE] = ACTIONS(3965),
    [anon_sym_signed] = ACTIONS(3967),
    [anon_sym_unsigned] = ACTIONS(3967),
    [anon_sym_long] = ACTIONS(3967),
    [anon_sym_short] = ACTIONS(3967),
    [anon_sym_static] = ACTIONS(3967),
    [anon_sym_EQ] = ACTIONS(3967),
    [anon_sym_ATautoreleasepool] = ACTIONS(3965),
    [anon_sym_auto] = ACTIONS(3967),
    [anon_sym_register] = ACTIONS(3967),
    [anon_sym_inline] = ACTIONS(3967),
    [anon_sym___inline] = ACTIONS(3967),
    [anon_sym___inline__] = ACTIONS(3967),
    [anon_sym___forceinline] = ACTIONS(3967),
    [anon_sym_thread_local] = ACTIONS(3967),
    [anon_sym___thread] = ACTIONS(3967),
    [anon_sym_CG_EXTERN] = ACTIONS(3967),
    [anon_sym_CG_INLINE] = ACTIONS(3967),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3967),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3967),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3967),
    [anon_sym_IBOutlet] = ACTIONS(3967),
    [anon_sym_IBInspectable] = ACTIONS(3967),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3967),
    [anon_sym_NS_INLINE] = ACTIONS(3967),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3967),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3967),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3967),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3967),
    [anon_sym_const] = ACTIONS(3967),
    [anon_sym_constexpr] = ACTIONS(3967),
    [anon_sym_volatile] = ACTIONS(3967),
    [anon_sym_restrict] = ACTIONS(3967),
    [anon_sym___restrict__] = ACTIONS(3967),
    [anon_sym__Atomic] = ACTIONS(3967),
    [anon_sym__Noreturn] = ACTIONS(3967),
    [anon_sym__Nonnull] = ACTIONS(3967),
    [anon_sym_nullable] = ACTIONS(3967),
    [anon_sym__Complex] = ACTIONS(3967),
    [anon_sym__Nullable] = ACTIONS(3967),
    [anon_sym__Nullable_result] = ACTIONS(3967),
    [anon_sym__Null_unspecified] = ACTIONS(3967),
    [anon_sym___autoreleasing] = ACTIONS(3967),
    [anon_sym___block] = ACTIONS(3967),
    [anon_sym___bridge] = ACTIONS(3967),
    [anon_sym___bridge_retained] = ACTIONS(3967),
    [anon_sym___bridge_transfer] = ACTIONS(3967),
    [anon_sym___complex] = ACTIONS(3967),
    [anon_sym___const] = ACTIONS(3967),
    [anon_sym___imag] = ACTIONS(3967),
    [anon_sym___kindof] = ACTIONS(3967),
    [anon_sym___nonnull] = ACTIONS(3967),
    [anon_sym___nullable] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3967),
    [anon_sym___real] = ACTIONS(3967),
    [anon_sym___strong] = ACTIONS(3967),
    [anon_sym___unsafe_unretained] = ACTIONS(3967),
    [anon_sym___unused] = ACTIONS(3967),
    [anon_sym___weak] = ACTIONS(3967),
    [anon_sym_alignas] = ACTIONS(3967),
    [anon_sym__Alignas] = ACTIONS(3967),
    [sym_primitive_type] = ACTIONS(3967),
    [anon_sym_enum] = ACTIONS(3967),
    [anon_sym_struct] = ACTIONS(3967),
    [anon_sym_union] = ACTIONS(3967),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_STAR_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_PERCENT_EQ] = ACTIONS(3965),
    [anon_sym_PLUS_EQ] = ACTIONS(3965),
    [anon_sym_DASH_EQ] = ACTIONS(3965),
    [anon_sym_LT_LT_EQ] = ACTIONS(3965),
    [anon_sym_GT_GT_EQ] = ACTIONS(3965),
    [anon_sym_AMP_EQ] = ACTIONS(3965),
    [anon_sym_CARET_EQ] = ACTIONS(3965),
    [anon_sym_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_AT] = ACTIONS(3948),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3967),
    [anon_sym___typeof] = ACTIONS(3967),
    [anon_sym_typeof] = ACTIONS(3967),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3967),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3967),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3967),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3967),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3967),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3967),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3967),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_API_AVAILABLE] = ACTIONS(3967),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_API_DEPRECATED] = ACTIONS(3967),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3967),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym___deprecated_msg] = ACTIONS(3967),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3967),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3967),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3967),
    [anon_sym_BOOL] = ACTIONS(3967),
    [anon_sym_IMP] = ACTIONS(3967),
    [anon_sym_SEL] = ACTIONS(3967),
    [anon_sym_Class] = ACTIONS(3967),
    [anon_sym_id] = ACTIONS(3967),
  },
  [STATE(1435)] = {
    [aux_sym_selector_expression_repeat2] = STATE(1435),
    [sym_identifier] = ACTIONS(3969),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3971),
    [anon_sym_COMMA] = ACTIONS(3971),
    [anon_sym_RPAREN] = ACTIONS(3973),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3971),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3971),
    [anon_sym_LPAREN2] = ACTIONS(3971),
    [anon_sym_DASH] = ACTIONS(3969),
    [anon_sym_PLUS] = ACTIONS(3969),
    [anon_sym_STAR] = ACTIONS(3969),
    [anon_sym_SLASH] = ACTIONS(3969),
    [anon_sym_PERCENT] = ACTIONS(3969),
    [anon_sym_PIPE_PIPE] = ACTIONS(3971),
    [anon_sym_AMP_AMP] = ACTIONS(3971),
    [anon_sym_PIPE] = ACTIONS(3969),
    [anon_sym_CARET] = ACTIONS(3969),
    [anon_sym_AMP] = ACTIONS(3969),
    [anon_sym_EQ_EQ] = ACTIONS(3971),
    [anon_sym_BANG_EQ] = ACTIONS(3971),
    [anon_sym_GT] = ACTIONS(3969),
    [anon_sym_GT_EQ] = ACTIONS(3971),
    [anon_sym_LT_EQ] = ACTIONS(3971),
    [anon_sym_LT] = ACTIONS(3969),
    [anon_sym_LT_LT] = ACTIONS(3969),
    [anon_sym_GT_GT] = ACTIONS(3969),
    [anon_sym_SEMI] = ACTIONS(3971),
    [anon_sym___extension__] = ACTIONS(3969),
    [anon_sym_extern] = ACTIONS(3969),
    [anon_sym___attribute__] = ACTIONS(3969),
    [anon_sym___attribute] = ACTIONS(3969),
    [anon_sym_noreturn] = ACTIONS(3969),
    [anon_sym_LBRACK] = ACTIONS(3971),
    [anon_sym_RBRACK] = ACTIONS(3971),
    [anon_sym___declspec] = ACTIONS(3969),
    [anon_sym_LBRACE] = ACTIONS(3971),
    [anon_sym_RBRACE] = ACTIONS(3971),
    [anon_sym_signed] = ACTIONS(3969),
    [anon_sym_unsigned] = ACTIONS(3969),
    [anon_sym_long] = ACTIONS(3969),
    [anon_sym_short] = ACTIONS(3969),
    [anon_sym_static] = ACTIONS(3969),
    [anon_sym_EQ] = ACTIONS(3969),
    [anon_sym_ATautoreleasepool] = ACTIONS(3971),
    [anon_sym_auto] = ACTIONS(3969),
    [anon_sym_register] = ACTIONS(3969),
    [anon_sym_inline] = ACTIONS(3969),
    [anon_sym___inline] = ACTIONS(3969),
    [anon_sym___inline__] = ACTIONS(3969),
    [anon_sym___forceinline] = ACTIONS(3969),
    [anon_sym_thread_local] = ACTIONS(3969),
    [anon_sym___thread] = ACTIONS(3969),
    [anon_sym_CG_EXTERN] = ACTIONS(3969),
    [anon_sym_CG_INLINE] = ACTIONS(3969),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3969),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3969),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3969),
    [anon_sym_IBOutlet] = ACTIONS(3969),
    [anon_sym_IBInspectable] = ACTIONS(3969),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3969),
    [anon_sym_NS_INLINE] = ACTIONS(3969),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3969),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3969),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3969),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3969),
    [anon_sym_const] = ACTIONS(3969),
    [anon_sym_constexpr] = ACTIONS(3969),
    [anon_sym_volatile] = ACTIONS(3969),
    [anon_sym_restrict] = ACTIONS(3969),
    [anon_sym___restrict__] = ACTIONS(3969),
    [anon_sym__Atomic] = ACTIONS(3969),
    [anon_sym__Noreturn] = ACTIONS(3969),
    [anon_sym__Nonnull] = ACTIONS(3969),
    [anon_sym_nullable] = ACTIONS(3969),
    [anon_sym__Complex] = ACTIONS(3969),
    [anon_sym__Nullable] = ACTIONS(3969),
    [anon_sym__Nullable_result] = ACTIONS(3969),
    [anon_sym__Null_unspecified] = ACTIONS(3969),
    [anon_sym___autoreleasing] = ACTIONS(3969),
    [anon_sym___block] = ACTIONS(3969),
    [anon_sym___bridge] = ACTIONS(3969),
    [anon_sym___bridge_retained] = ACTIONS(3969),
    [anon_sym___bridge_transfer] = ACTIONS(3969),
    [anon_sym___complex] = ACTIONS(3969),
    [anon_sym___const] = ACTIONS(3969),
    [anon_sym___imag] = ACTIONS(3969),
    [anon_sym___kindof] = ACTIONS(3969),
    [anon_sym___nonnull] = ACTIONS(3969),
    [anon_sym___nullable] = ACTIONS(3969),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3969),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3969),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3969),
    [anon_sym___real] = ACTIONS(3969),
    [anon_sym___strong] = ACTIONS(3969),
    [anon_sym___unsafe_unretained] = ACTIONS(3969),
    [anon_sym___unused] = ACTIONS(3969),
    [anon_sym___weak] = ACTIONS(3969),
    [anon_sym_alignas] = ACTIONS(3969),
    [anon_sym__Alignas] = ACTIONS(3969),
    [sym_primitive_type] = ACTIONS(3969),
    [anon_sym_enum] = ACTIONS(3969),
    [anon_sym_COLON] = ACTIONS(3971),
    [anon_sym_struct] = ACTIONS(3969),
    [anon_sym_union] = ACTIONS(3969),
    [anon_sym_QMARK] = ACTIONS(3971),
    [anon_sym_STAR_EQ] = ACTIONS(3971),
    [anon_sym_SLASH_EQ] = ACTIONS(3971),
    [anon_sym_PERCENT_EQ] = ACTIONS(3971),
    [anon_sym_PLUS_EQ] = ACTIONS(3971),
    [anon_sym_DASH_EQ] = ACTIONS(3971),
    [anon_sym_LT_LT_EQ] = ACTIONS(3971),
    [anon_sym_GT_GT_EQ] = ACTIONS(3971),
    [anon_sym_AMP_EQ] = ACTIONS(3971),
    [anon_sym_CARET_EQ] = ACTIONS(3971),
    [anon_sym_PIPE_EQ] = ACTIONS(3971),
    [anon_sym_DASH_DASH] = ACTIONS(3971),
    [anon_sym_PLUS_PLUS] = ACTIONS(3971),
    [anon_sym_DOT] = ACTIONS(3969),
    [anon_sym_DASH_GT] = ACTIONS(3971),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3969),
    [anon_sym___typeof] = ACTIONS(3969),
    [anon_sym_typeof] = ACTIONS(3969),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3969),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3969),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3969),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3969),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3969),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3969),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3969),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3969),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3969),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3969),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3969),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3969),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3969),
    [anon_sym_API_AVAILABLE] = ACTIONS(3969),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3969),
    [anon_sym_API_DEPRECATED] = ACTIONS(3969),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3969),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3969),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3969),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3969),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3969),
    [anon_sym___deprecated_msg] = ACTIONS(3969),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3969),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3969),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3969),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3969),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3969),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3969),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3969),
    [anon_sym_BOOL] = ACTIONS(3969),
    [anon_sym_IMP] = ACTIONS(3969),
    [anon_sym_SEL] = ACTIONS(3969),
    [anon_sym_Class] = ACTIONS(3969),
    [anon_sym_id] = ACTIONS(3969),
  },
  [STATE(1436)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(3976),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3978),
    [anon_sym_COMMA] = ACTIONS(3978),
    [anon_sym_RPAREN] = ACTIONS(3978),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3978),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3978),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(3976),
    [anon_sym_PLUS] = ACTIONS(3976),
    [anon_sym_STAR] = ACTIONS(3976),
    [anon_sym_SLASH] = ACTIONS(3976),
    [anon_sym_PERCENT] = ACTIONS(3976),
    [anon_sym_PIPE_PIPE] = ACTIONS(3978),
    [anon_sym_AMP_AMP] = ACTIONS(3978),
    [anon_sym_PIPE] = ACTIONS(3976),
    [anon_sym_CARET] = ACTIONS(3976),
    [anon_sym_AMP] = ACTIONS(3976),
    [anon_sym_EQ_EQ] = ACTIONS(3978),
    [anon_sym_BANG_EQ] = ACTIONS(3978),
    [anon_sym_GT] = ACTIONS(3976),
    [anon_sym_GT_EQ] = ACTIONS(3978),
    [anon_sym_LT_EQ] = ACTIONS(3978),
    [anon_sym_LT] = ACTIONS(3976),
    [anon_sym_LT_LT] = ACTIONS(3976),
    [anon_sym_GT_GT] = ACTIONS(3976),
    [anon_sym_SEMI] = ACTIONS(3978),
    [anon_sym___extension__] = ACTIONS(3976),
    [anon_sym_extern] = ACTIONS(3976),
    [anon_sym___attribute__] = ACTIONS(3976),
    [anon_sym___attribute] = ACTIONS(3976),
    [anon_sym_noreturn] = ACTIONS(3976),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(3978),
    [anon_sym___declspec] = ACTIONS(3976),
    [anon_sym_LBRACE] = ACTIONS(3978),
    [anon_sym_RBRACE] = ACTIONS(3978),
    [anon_sym_signed] = ACTIONS(3976),
    [anon_sym_unsigned] = ACTIONS(3976),
    [anon_sym_long] = ACTIONS(3976),
    [anon_sym_short] = ACTIONS(3976),
    [anon_sym_static] = ACTIONS(3976),
    [anon_sym_EQ] = ACTIONS(3976),
    [anon_sym_ATautoreleasepool] = ACTIONS(3978),
    [anon_sym_auto] = ACTIONS(3976),
    [anon_sym_register] = ACTIONS(3976),
    [anon_sym_inline] = ACTIONS(3976),
    [anon_sym___inline] = ACTIONS(3976),
    [anon_sym___inline__] = ACTIONS(3976),
    [anon_sym___forceinline] = ACTIONS(3976),
    [anon_sym_thread_local] = ACTIONS(3976),
    [anon_sym___thread] = ACTIONS(3976),
    [anon_sym_CG_EXTERN] = ACTIONS(3976),
    [anon_sym_CG_INLINE] = ACTIONS(3976),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3976),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3976),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3976),
    [anon_sym_IBOutlet] = ACTIONS(3976),
    [anon_sym_IBInspectable] = ACTIONS(3976),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3976),
    [anon_sym_NS_INLINE] = ACTIONS(3976),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3976),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3976),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3976),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3976),
    [anon_sym_const] = ACTIONS(3976),
    [anon_sym_constexpr] = ACTIONS(3976),
    [anon_sym_volatile] = ACTIONS(3976),
    [anon_sym_restrict] = ACTIONS(3976),
    [anon_sym___restrict__] = ACTIONS(3976),
    [anon_sym__Atomic] = ACTIONS(3976),
    [anon_sym__Noreturn] = ACTIONS(3976),
    [anon_sym__Nonnull] = ACTIONS(3976),
    [anon_sym_nullable] = ACTIONS(3976),
    [anon_sym__Complex] = ACTIONS(3976),
    [anon_sym__Nullable] = ACTIONS(3976),
    [anon_sym__Nullable_result] = ACTIONS(3976),
    [anon_sym__Null_unspecified] = ACTIONS(3976),
    [anon_sym___autoreleasing] = ACTIONS(3976),
    [anon_sym___block] = ACTIONS(3976),
    [anon_sym___bridge] = ACTIONS(3976),
    [anon_sym___bridge_retained] = ACTIONS(3976),
    [anon_sym___bridge_transfer] = ACTIONS(3976),
    [anon_sym___complex] = ACTIONS(3976),
    [anon_sym___const] = ACTIONS(3976),
    [anon_sym___imag] = ACTIONS(3976),
    [anon_sym___kindof] = ACTIONS(3976),
    [anon_sym___nonnull] = ACTIONS(3976),
    [anon_sym___nullable] = ACTIONS(3976),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3976),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3976),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3976),
    [anon_sym___real] = ACTIONS(3976),
    [anon_sym___strong] = ACTIONS(3976),
    [anon_sym___unsafe_unretained] = ACTIONS(3976),
    [anon_sym___unused] = ACTIONS(3976),
    [anon_sym___weak] = ACTIONS(3976),
    [anon_sym_alignas] = ACTIONS(3976),
    [anon_sym__Alignas] = ACTIONS(3976),
    [sym_primitive_type] = ACTIONS(3976),
    [anon_sym_enum] = ACTIONS(3976),
    [anon_sym_COLON] = ACTIONS(3978),
    [anon_sym_struct] = ACTIONS(3976),
    [anon_sym_union] = ACTIONS(3976),
    [anon_sym_QMARK] = ACTIONS(3978),
    [anon_sym_STAR_EQ] = ACTIONS(3978),
    [anon_sym_SLASH_EQ] = ACTIONS(3978),
    [anon_sym_PERCENT_EQ] = ACTIONS(3978),
    [anon_sym_PLUS_EQ] = ACTIONS(3978),
    [anon_sym_DASH_EQ] = ACTIONS(3978),
    [anon_sym_LT_LT_EQ] = ACTIONS(3978),
    [anon_sym_GT_GT_EQ] = ACTIONS(3978),
    [anon_sym_AMP_EQ] = ACTIONS(3978),
    [anon_sym_CARET_EQ] = ACTIONS(3978),
    [anon_sym_PIPE_EQ] = ACTIONS(3978),
    [anon_sym_DASH_DASH] = ACTIONS(3978),
    [anon_sym_PLUS_PLUS] = ACTIONS(3978),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3976),
    [anon_sym___typeof] = ACTIONS(3976),
    [anon_sym_typeof] = ACTIONS(3976),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3976),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3976),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3976),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3976),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3976),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3976),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3976),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3976),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3976),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3976),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3976),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3976),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3976),
    [anon_sym_API_AVAILABLE] = ACTIONS(3976),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3976),
    [anon_sym_API_DEPRECATED] = ACTIONS(3976),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3976),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3976),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3976),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3976),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3976),
    [anon_sym___deprecated_msg] = ACTIONS(3976),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3976),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3976),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3976),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3976),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3976),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3976),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3976),
    [anon_sym_BOOL] = ACTIONS(3976),
    [anon_sym_IMP] = ACTIONS(3976),
    [anon_sym_SEL] = ACTIONS(3976),
    [anon_sym_Class] = ACTIONS(3976),
    [anon_sym_id] = ACTIONS(3976),
  },
  [STATE(1437)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(3988),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3990),
    [anon_sym_COMMA] = ACTIONS(3990),
    [anon_sym_RPAREN] = ACTIONS(3990),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3990),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3990),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(3988),
    [anon_sym_PLUS] = ACTIONS(3988),
    [anon_sym_STAR] = ACTIONS(3988),
    [anon_sym_SLASH] = ACTIONS(3988),
    [anon_sym_PERCENT] = ACTIONS(3988),
    [anon_sym_PIPE_PIPE] = ACTIONS(3990),
    [anon_sym_AMP_AMP] = ACTIONS(3990),
    [anon_sym_PIPE] = ACTIONS(3988),
    [anon_sym_CARET] = ACTIONS(3988),
    [anon_sym_AMP] = ACTIONS(3988),
    [anon_sym_EQ_EQ] = ACTIONS(3990),
    [anon_sym_BANG_EQ] = ACTIONS(3990),
    [anon_sym_GT] = ACTIONS(3988),
    [anon_sym_GT_EQ] = ACTIONS(3990),
    [anon_sym_LT_EQ] = ACTIONS(3990),
    [anon_sym_LT] = ACTIONS(3988),
    [anon_sym_LT_LT] = ACTIONS(3988),
    [anon_sym_GT_GT] = ACTIONS(3988),
    [anon_sym_SEMI] = ACTIONS(3990),
    [anon_sym___extension__] = ACTIONS(3988),
    [anon_sym_extern] = ACTIONS(3988),
    [anon_sym___attribute__] = ACTIONS(3988),
    [anon_sym___attribute] = ACTIONS(3988),
    [anon_sym_noreturn] = ACTIONS(3988),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(3990),
    [anon_sym___declspec] = ACTIONS(3988),
    [anon_sym_LBRACE] = ACTIONS(3990),
    [anon_sym_RBRACE] = ACTIONS(3990),
    [anon_sym_signed] = ACTIONS(3988),
    [anon_sym_unsigned] = ACTIONS(3988),
    [anon_sym_long] = ACTIONS(3988),
    [anon_sym_short] = ACTIONS(3988),
    [anon_sym_static] = ACTIONS(3988),
    [anon_sym_EQ] = ACTIONS(3988),
    [anon_sym_ATautoreleasepool] = ACTIONS(3990),
    [anon_sym_auto] = ACTIONS(3988),
    [anon_sym_register] = ACTIONS(3988),
    [anon_sym_inline] = ACTIONS(3988),
    [anon_sym___inline] = ACTIONS(3988),
    [anon_sym___inline__] = ACTIONS(3988),
    [anon_sym___forceinline] = ACTIONS(3988),
    [anon_sym_thread_local] = ACTIONS(3988),
    [anon_sym___thread] = ACTIONS(3988),
    [anon_sym_CG_EXTERN] = ACTIONS(3988),
    [anon_sym_CG_INLINE] = ACTIONS(3988),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3988),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3988),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3988),
    [anon_sym_IBOutlet] = ACTIONS(3988),
    [anon_sym_IBInspectable] = ACTIONS(3988),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3988),
    [anon_sym_NS_INLINE] = ACTIONS(3988),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3988),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3988),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3988),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3988),
    [anon_sym_const] = ACTIONS(3988),
    [anon_sym_constexpr] = ACTIONS(3988),
    [anon_sym_volatile] = ACTIONS(3988),
    [anon_sym_restrict] = ACTIONS(3988),
    [anon_sym___restrict__] = ACTIONS(3988),
    [anon_sym__Atomic] = ACTIONS(3988),
    [anon_sym__Noreturn] = ACTIONS(3988),
    [anon_sym__Nonnull] = ACTIONS(3988),
    [anon_sym_nullable] = ACTIONS(3988),
    [anon_sym__Complex] = ACTIONS(3988),
    [anon_sym__Nullable] = ACTIONS(3988),
    [anon_sym__Nullable_result] = ACTIONS(3988),
    [anon_sym__Null_unspecified] = ACTIONS(3988),
    [anon_sym___autoreleasing] = ACTIONS(3988),
    [anon_sym___block] = ACTIONS(3988),
    [anon_sym___bridge] = ACTIONS(3988),
    [anon_sym___bridge_retained] = ACTIONS(3988),
    [anon_sym___bridge_transfer] = ACTIONS(3988),
    [anon_sym___complex] = ACTIONS(3988),
    [anon_sym___const] = ACTIONS(3988),
    [anon_sym___imag] = ACTIONS(3988),
    [anon_sym___kindof] = ACTIONS(3988),
    [anon_sym___nonnull] = ACTIONS(3988),
    [anon_sym___nullable] = ACTIONS(3988),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3988),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3988),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3988),
    [anon_sym___real] = ACTIONS(3988),
    [anon_sym___strong] = ACTIONS(3988),
    [anon_sym___unsafe_unretained] = ACTIONS(3988),
    [anon_sym___unused] = ACTIONS(3988),
    [anon_sym___weak] = ACTIONS(3988),
    [anon_sym_alignas] = ACTIONS(3988),
    [anon_sym__Alignas] = ACTIONS(3988),
    [sym_primitive_type] = ACTIONS(3988),
    [anon_sym_enum] = ACTIONS(3988),
    [anon_sym_COLON] = ACTIONS(3990),
    [anon_sym_struct] = ACTIONS(3988),
    [anon_sym_union] = ACTIONS(3988),
    [anon_sym_QMARK] = ACTIONS(3990),
    [anon_sym_STAR_EQ] = ACTIONS(3990),
    [anon_sym_SLASH_EQ] = ACTIONS(3990),
    [anon_sym_PERCENT_EQ] = ACTIONS(3990),
    [anon_sym_PLUS_EQ] = ACTIONS(3990),
    [anon_sym_DASH_EQ] = ACTIONS(3990),
    [anon_sym_LT_LT_EQ] = ACTIONS(3990),
    [anon_sym_GT_GT_EQ] = ACTIONS(3990),
    [anon_sym_AMP_EQ] = ACTIONS(3990),
    [anon_sym_CARET_EQ] = ACTIONS(3990),
    [anon_sym_PIPE_EQ] = ACTIONS(3990),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3988),
    [anon_sym___typeof] = ACTIONS(3988),
    [anon_sym_typeof] = ACTIONS(3988),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3988),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3988),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3988),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3988),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3988),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3988),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3988),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3988),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3988),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3988),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3988),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3988),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3988),
    [anon_sym_API_AVAILABLE] = ACTIONS(3988),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3988),
    [anon_sym_API_DEPRECATED] = ACTIONS(3988),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3988),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3988),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3988),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3988),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3988),
    [anon_sym___deprecated_msg] = ACTIONS(3988),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3988),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3988),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3988),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3988),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3988),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3988),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3988),
    [anon_sym_BOOL] = ACTIONS(3988),
    [anon_sym_IMP] = ACTIONS(3988),
    [anon_sym_SEL] = ACTIONS(3988),
    [anon_sym_Class] = ACTIONS(3988),
    [anon_sym_id] = ACTIONS(3988),
  },
  [STATE(1438)] = {
    [aux_sym_selector_expression_repeat2] = STATE(1435),
    [sym_identifier] = ACTIONS(3994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3996),
    [anon_sym_COMMA] = ACTIONS(3996),
    [anon_sym_RPAREN] = ACTIONS(3998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3996),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3996),
    [anon_sym_LPAREN2] = ACTIONS(3996),
    [anon_sym_DASH] = ACTIONS(3994),
    [anon_sym_PLUS] = ACTIONS(3994),
    [anon_sym_STAR] = ACTIONS(3994),
    [anon_sym_SLASH] = ACTIONS(3994),
    [anon_sym_PERCENT] = ACTIONS(3994),
    [anon_sym_PIPE_PIPE] = ACTIONS(3996),
    [anon_sym_AMP_AMP] = ACTIONS(3996),
    [anon_sym_PIPE] = ACTIONS(3994),
    [anon_sym_CARET] = ACTIONS(3994),
    [anon_sym_AMP] = ACTIONS(3994),
    [anon_sym_EQ_EQ] = ACTIONS(3996),
    [anon_sym_BANG_EQ] = ACTIONS(3996),
    [anon_sym_GT] = ACTIONS(3994),
    [anon_sym_GT_EQ] = ACTIONS(3996),
    [anon_sym_LT_EQ] = ACTIONS(3996),
    [anon_sym_LT] = ACTIONS(3994),
    [anon_sym_LT_LT] = ACTIONS(3994),
    [anon_sym_GT_GT] = ACTIONS(3994),
    [anon_sym_SEMI] = ACTIONS(3996),
    [anon_sym___extension__] = ACTIONS(3994),
    [anon_sym_extern] = ACTIONS(3994),
    [anon_sym___attribute__] = ACTIONS(3994),
    [anon_sym___attribute] = ACTIONS(3994),
    [anon_sym_noreturn] = ACTIONS(3994),
    [anon_sym_LBRACK] = ACTIONS(3996),
    [anon_sym_RBRACK] = ACTIONS(3996),
    [anon_sym___declspec] = ACTIONS(3994),
    [anon_sym_LBRACE] = ACTIONS(3996),
    [anon_sym_RBRACE] = ACTIONS(3996),
    [anon_sym_signed] = ACTIONS(3994),
    [anon_sym_unsigned] = ACTIONS(3994),
    [anon_sym_long] = ACTIONS(3994),
    [anon_sym_short] = ACTIONS(3994),
    [anon_sym_static] = ACTIONS(3994),
    [anon_sym_EQ] = ACTIONS(3994),
    [anon_sym_ATautoreleasepool] = ACTIONS(3996),
    [anon_sym_auto] = ACTIONS(3994),
    [anon_sym_register] = ACTIONS(3994),
    [anon_sym_inline] = ACTIONS(3994),
    [anon_sym___inline] = ACTIONS(3994),
    [anon_sym___inline__] = ACTIONS(3994),
    [anon_sym___forceinline] = ACTIONS(3994),
    [anon_sym_thread_local] = ACTIONS(3994),
    [anon_sym___thread] = ACTIONS(3994),
    [anon_sym_CG_EXTERN] = ACTIONS(3994),
    [anon_sym_CG_INLINE] = ACTIONS(3994),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3994),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3994),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3994),
    [anon_sym_IBOutlet] = ACTIONS(3994),
    [anon_sym_IBInspectable] = ACTIONS(3994),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3994),
    [anon_sym_NS_INLINE] = ACTIONS(3994),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3994),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3994),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3994),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3994),
    [anon_sym_const] = ACTIONS(3994),
    [anon_sym_constexpr] = ACTIONS(3994),
    [anon_sym_volatile] = ACTIONS(3994),
    [anon_sym_restrict] = ACTIONS(3994),
    [anon_sym___restrict__] = ACTIONS(3994),
    [anon_sym__Atomic] = ACTIONS(3994),
    [anon_sym__Noreturn] = ACTIONS(3994),
    [anon_sym__Nonnull] = ACTIONS(3994),
    [anon_sym_nullable] = ACTIONS(3994),
    [anon_sym__Complex] = ACTIONS(3994),
    [anon_sym__Nullable] = ACTIONS(3994),
    [anon_sym__Nullable_result] = ACTIONS(3994),
    [anon_sym__Null_unspecified] = ACTIONS(3994),
    [anon_sym___autoreleasing] = ACTIONS(3994),
    [anon_sym___block] = ACTIONS(3994),
    [anon_sym___bridge] = ACTIONS(3994),
    [anon_sym___bridge_retained] = ACTIONS(3994),
    [anon_sym___bridge_transfer] = ACTIONS(3994),
    [anon_sym___complex] = ACTIONS(3994),
    [anon_sym___const] = ACTIONS(3994),
    [anon_sym___imag] = ACTIONS(3994),
    [anon_sym___kindof] = ACTIONS(3994),
    [anon_sym___nonnull] = ACTIONS(3994),
    [anon_sym___nullable] = ACTIONS(3994),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3994),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3994),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3994),
    [anon_sym___real] = ACTIONS(3994),
    [anon_sym___strong] = ACTIONS(3994),
    [anon_sym___unsafe_unretained] = ACTIONS(3994),
    [anon_sym___unused] = ACTIONS(3994),
    [anon_sym___weak] = ACTIONS(3994),
    [anon_sym_alignas] = ACTIONS(3994),
    [anon_sym__Alignas] = ACTIONS(3994),
    [sym_primitive_type] = ACTIONS(3994),
    [anon_sym_enum] = ACTIONS(3994),
    [anon_sym_COLON] = ACTIONS(3996),
    [anon_sym_struct] = ACTIONS(3994),
    [anon_sym_union] = ACTIONS(3994),
    [anon_sym_QMARK] = ACTIONS(3996),
    [anon_sym_STAR_EQ] = ACTIONS(3996),
    [anon_sym_SLASH_EQ] = ACTIONS(3996),
    [anon_sym_PERCENT_EQ] = ACTIONS(3996),
    [anon_sym_PLUS_EQ] = ACTIONS(3996),
    [anon_sym_DASH_EQ] = ACTIONS(3996),
    [anon_sym_LT_LT_EQ] = ACTIONS(3996),
    [anon_sym_GT_GT_EQ] = ACTIONS(3996),
    [anon_sym_AMP_EQ] = ACTIONS(3996),
    [anon_sym_CARET_EQ] = ACTIONS(3996),
    [anon_sym_PIPE_EQ] = ACTIONS(3996),
    [anon_sym_DASH_DASH] = ACTIONS(3996),
    [anon_sym_PLUS_PLUS] = ACTIONS(3996),
    [anon_sym_DOT] = ACTIONS(3994),
    [anon_sym_DASH_GT] = ACTIONS(3996),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3994),
    [anon_sym___typeof] = ACTIONS(3994),
    [anon_sym_typeof] = ACTIONS(3994),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3994),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3994),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3994),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3994),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3994),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3994),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3994),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3994),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3994),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3994),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3994),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3994),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3994),
    [anon_sym_API_AVAILABLE] = ACTIONS(3994),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3994),
    [anon_sym_API_DEPRECATED] = ACTIONS(3994),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3994),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3994),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3994),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3994),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3994),
    [anon_sym___deprecated_msg] = ACTIONS(3994),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3994),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3994),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3994),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3994),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3994),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3994),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3994),
    [anon_sym_BOOL] = ACTIONS(3994),
    [anon_sym_IMP] = ACTIONS(3994),
    [anon_sym_SEL] = ACTIONS(3994),
    [anon_sym_Class] = ACTIONS(3994),
    [anon_sym_id] = ACTIONS(3994),
  },
  [STATE(1439)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_RPAREN] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4000),
    [anon_sym_PLUS] = ACTIONS(4000),
    [anon_sym_STAR] = ACTIONS(4000),
    [anon_sym_SLASH] = ACTIONS(4000),
    [anon_sym_PERCENT] = ACTIONS(4000),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4000),
    [anon_sym_GT_GT] = ACTIONS(4000),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(4002),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_COLON] = ACTIONS(4002),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1440)] = {
    [sym_string_literal] = STATE(1434),
    [sym_identifier] = ACTIONS(4004),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4006),
    [anon_sym_COMMA] = ACTIONS(4006),
    [anon_sym_LPAREN2] = ACTIONS(4006),
    [anon_sym_DASH] = ACTIONS(4004),
    [anon_sym_PLUS] = ACTIONS(4004),
    [anon_sym_STAR] = ACTIONS(4004),
    [anon_sym_SLASH] = ACTIONS(4004),
    [anon_sym_PERCENT] = ACTIONS(4004),
    [anon_sym_PIPE_PIPE] = ACTIONS(4006),
    [anon_sym_AMP_AMP] = ACTIONS(4006),
    [anon_sym_PIPE] = ACTIONS(4004),
    [anon_sym_CARET] = ACTIONS(4004),
    [anon_sym_AMP] = ACTIONS(4004),
    [anon_sym_EQ_EQ] = ACTIONS(4006),
    [anon_sym_BANG_EQ] = ACTIONS(4006),
    [anon_sym_GT] = ACTIONS(4004),
    [anon_sym_GT_EQ] = ACTIONS(4006),
    [anon_sym_LT_EQ] = ACTIONS(4006),
    [anon_sym_LT] = ACTIONS(4004),
    [anon_sym_LT_LT] = ACTIONS(4004),
    [anon_sym_GT_GT] = ACTIONS(4004),
    [anon_sym_SEMI] = ACTIONS(4006),
    [anon_sym___extension__] = ACTIONS(4004),
    [anon_sym_extern] = ACTIONS(4004),
    [anon_sym___attribute__] = ACTIONS(4004),
    [anon_sym___attribute] = ACTIONS(4004),
    [anon_sym_noreturn] = ACTIONS(4004),
    [anon_sym_LBRACK] = ACTIONS(4006),
    [anon_sym___declspec] = ACTIONS(4004),
    [anon_sym_LBRACE] = ACTIONS(4006),
    [anon_sym_signed] = ACTIONS(4004),
    [anon_sym_unsigned] = ACTIONS(4004),
    [anon_sym_long] = ACTIONS(4004),
    [anon_sym_short] = ACTIONS(4004),
    [anon_sym_static] = ACTIONS(4004),
    [anon_sym_EQ] = ACTIONS(4004),
    [anon_sym_ATautoreleasepool] = ACTIONS(4006),
    [anon_sym_auto] = ACTIONS(4004),
    [anon_sym_register] = ACTIONS(4004),
    [anon_sym_inline] = ACTIONS(4004),
    [anon_sym___inline] = ACTIONS(4004),
    [anon_sym___inline__] = ACTIONS(4004),
    [anon_sym___forceinline] = ACTIONS(4004),
    [anon_sym_thread_local] = ACTIONS(4004),
    [anon_sym___thread] = ACTIONS(4004),
    [anon_sym_CG_EXTERN] = ACTIONS(4004),
    [anon_sym_CG_INLINE] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4004),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4004),
    [anon_sym_IBOutlet] = ACTIONS(4004),
    [anon_sym_IBInspectable] = ACTIONS(4004),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4004),
    [anon_sym_NS_INLINE] = ACTIONS(4004),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4004),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4004),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4004),
    [anon_sym_const] = ACTIONS(4004),
    [anon_sym_constexpr] = ACTIONS(4004),
    [anon_sym_volatile] = ACTIONS(4004),
    [anon_sym_restrict] = ACTIONS(4004),
    [anon_sym___restrict__] = ACTIONS(4004),
    [anon_sym__Atomic] = ACTIONS(4004),
    [anon_sym__Noreturn] = ACTIONS(4004),
    [anon_sym__Nonnull] = ACTIONS(4004),
    [anon_sym_nullable] = ACTIONS(4004),
    [anon_sym__Complex] = ACTIONS(4004),
    [anon_sym__Nullable] = ACTIONS(4004),
    [anon_sym__Nullable_result] = ACTIONS(4004),
    [anon_sym__Null_unspecified] = ACTIONS(4004),
    [anon_sym___autoreleasing] = ACTIONS(4004),
    [anon_sym___block] = ACTIONS(4004),
    [anon_sym___bridge] = ACTIONS(4004),
    [anon_sym___bridge_retained] = ACTIONS(4004),
    [anon_sym___bridge_transfer] = ACTIONS(4004),
    [anon_sym___complex] = ACTIONS(4004),
    [anon_sym___const] = ACTIONS(4004),
    [anon_sym___imag] = ACTIONS(4004),
    [anon_sym___kindof] = ACTIONS(4004),
    [anon_sym___nonnull] = ACTIONS(4004),
    [anon_sym___nullable] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4004),
    [anon_sym___real] = ACTIONS(4004),
    [anon_sym___strong] = ACTIONS(4004),
    [anon_sym___unsafe_unretained] = ACTIONS(4004),
    [anon_sym___unused] = ACTIONS(4004),
    [anon_sym___weak] = ACTIONS(4004),
    [anon_sym_alignas] = ACTIONS(4004),
    [anon_sym__Alignas] = ACTIONS(4004),
    [sym_primitive_type] = ACTIONS(4004),
    [anon_sym_enum] = ACTIONS(4004),
    [anon_sym_struct] = ACTIONS(4004),
    [anon_sym_union] = ACTIONS(4004),
    [anon_sym_QMARK] = ACTIONS(4006),
    [anon_sym_STAR_EQ] = ACTIONS(4006),
    [anon_sym_SLASH_EQ] = ACTIONS(4006),
    [anon_sym_PERCENT_EQ] = ACTIONS(4006),
    [anon_sym_PLUS_EQ] = ACTIONS(4006),
    [anon_sym_DASH_EQ] = ACTIONS(4006),
    [anon_sym_LT_LT_EQ] = ACTIONS(4006),
    [anon_sym_GT_GT_EQ] = ACTIONS(4006),
    [anon_sym_AMP_EQ] = ACTIONS(4006),
    [anon_sym_CARET_EQ] = ACTIONS(4006),
    [anon_sym_PIPE_EQ] = ACTIONS(4006),
    [anon_sym_DASH_DASH] = ACTIONS(4006),
    [anon_sym_PLUS_PLUS] = ACTIONS(4006),
    [anon_sym_DOT] = ACTIONS(4004),
    [anon_sym_DASH_GT] = ACTIONS(4006),
    [anon_sym_AT] = ACTIONS(3948),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4004),
    [anon_sym___typeof] = ACTIONS(4004),
    [anon_sym_typeof] = ACTIONS(4004),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4004),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4004),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4004),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_API_AVAILABLE] = ACTIONS(4004),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_API_DEPRECATED] = ACTIONS(4004),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym___deprecated_msg] = ACTIONS(4004),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4004),
    [anon_sym_BOOL] = ACTIONS(4004),
    [anon_sym_IMP] = ACTIONS(4004),
    [anon_sym_SEL] = ACTIONS(4004),
    [anon_sym_Class] = ACTIONS(4004),
    [anon_sym_id] = ACTIONS(4004),
  },
  [STATE(1441)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4008),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4010),
    [anon_sym_COMMA] = ACTIONS(4010),
    [anon_sym_RPAREN] = ACTIONS(4010),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4010),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4010),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4008),
    [anon_sym_PLUS] = ACTIONS(4008),
    [anon_sym_STAR] = ACTIONS(4008),
    [anon_sym_SLASH] = ACTIONS(4008),
    [anon_sym_PERCENT] = ACTIONS(4008),
    [anon_sym_PIPE_PIPE] = ACTIONS(4010),
    [anon_sym_AMP_AMP] = ACTIONS(4010),
    [anon_sym_PIPE] = ACTIONS(4008),
    [anon_sym_CARET] = ACTIONS(4008),
    [anon_sym_AMP] = ACTIONS(4008),
    [anon_sym_EQ_EQ] = ACTIONS(4010),
    [anon_sym_BANG_EQ] = ACTIONS(4010),
    [anon_sym_GT] = ACTIONS(4008),
    [anon_sym_GT_EQ] = ACTIONS(4010),
    [anon_sym_LT_EQ] = ACTIONS(4010),
    [anon_sym_LT] = ACTIONS(4008),
    [anon_sym_LT_LT] = ACTIONS(4008),
    [anon_sym_GT_GT] = ACTIONS(4008),
    [anon_sym_SEMI] = ACTIONS(4010),
    [anon_sym___extension__] = ACTIONS(4008),
    [anon_sym_extern] = ACTIONS(4008),
    [anon_sym___attribute__] = ACTIONS(4008),
    [anon_sym___attribute] = ACTIONS(4008),
    [anon_sym_noreturn] = ACTIONS(4008),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(4010),
    [anon_sym___declspec] = ACTIONS(4008),
    [anon_sym_LBRACE] = ACTIONS(4010),
    [anon_sym_RBRACE] = ACTIONS(4010),
    [anon_sym_signed] = ACTIONS(4008),
    [anon_sym_unsigned] = ACTIONS(4008),
    [anon_sym_long] = ACTIONS(4008),
    [anon_sym_short] = ACTIONS(4008),
    [anon_sym_static] = ACTIONS(4008),
    [anon_sym_EQ] = ACTIONS(4008),
    [anon_sym_ATautoreleasepool] = ACTIONS(4010),
    [anon_sym_auto] = ACTIONS(4008),
    [anon_sym_register] = ACTIONS(4008),
    [anon_sym_inline] = ACTIONS(4008),
    [anon_sym___inline] = ACTIONS(4008),
    [anon_sym___inline__] = ACTIONS(4008),
    [anon_sym___forceinline] = ACTIONS(4008),
    [anon_sym_thread_local] = ACTIONS(4008),
    [anon_sym___thread] = ACTIONS(4008),
    [anon_sym_CG_EXTERN] = ACTIONS(4008),
    [anon_sym_CG_INLINE] = ACTIONS(4008),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4008),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4008),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4008),
    [anon_sym_IBOutlet] = ACTIONS(4008),
    [anon_sym_IBInspectable] = ACTIONS(4008),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4008),
    [anon_sym_NS_INLINE] = ACTIONS(4008),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4008),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4008),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4008),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4008),
    [anon_sym_const] = ACTIONS(4008),
    [anon_sym_constexpr] = ACTIONS(4008),
    [anon_sym_volatile] = ACTIONS(4008),
    [anon_sym_restrict] = ACTIONS(4008),
    [anon_sym___restrict__] = ACTIONS(4008),
    [anon_sym__Atomic] = ACTIONS(4008),
    [anon_sym__Noreturn] = ACTIONS(4008),
    [anon_sym__Nonnull] = ACTIONS(4008),
    [anon_sym_nullable] = ACTIONS(4008),
    [anon_sym__Complex] = ACTIONS(4008),
    [anon_sym__Nullable] = ACTIONS(4008),
    [anon_sym__Nullable_result] = ACTIONS(4008),
    [anon_sym__Null_unspecified] = ACTIONS(4008),
    [anon_sym___autoreleasing] = ACTIONS(4008),
    [anon_sym___block] = ACTIONS(4008),
    [anon_sym___bridge] = ACTIONS(4008),
    [anon_sym___bridge_retained] = ACTIONS(4008),
    [anon_sym___bridge_transfer] = ACTIONS(4008),
    [anon_sym___complex] = ACTIONS(4008),
    [anon_sym___const] = ACTIONS(4008),
    [anon_sym___imag] = ACTIONS(4008),
    [anon_sym___kindof] = ACTIONS(4008),
    [anon_sym___nonnull] = ACTIONS(4008),
    [anon_sym___nullable] = ACTIONS(4008),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4008),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4008),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4008),
    [anon_sym___real] = ACTIONS(4008),
    [anon_sym___strong] = ACTIONS(4008),
    [anon_sym___unsafe_unretained] = ACTIONS(4008),
    [anon_sym___unused] = ACTIONS(4008),
    [anon_sym___weak] = ACTIONS(4008),
    [anon_sym_alignas] = ACTIONS(4008),
    [anon_sym__Alignas] = ACTIONS(4008),
    [sym_primitive_type] = ACTIONS(4008),
    [anon_sym_enum] = ACTIONS(4008),
    [anon_sym_COLON] = ACTIONS(4010),
    [anon_sym_struct] = ACTIONS(4008),
    [anon_sym_union] = ACTIONS(4008),
    [anon_sym_QMARK] = ACTIONS(4010),
    [anon_sym_STAR_EQ] = ACTIONS(4010),
    [anon_sym_SLASH_EQ] = ACTIONS(4010),
    [anon_sym_PERCENT_EQ] = ACTIONS(4010),
    [anon_sym_PLUS_EQ] = ACTIONS(4010),
    [anon_sym_DASH_EQ] = ACTIONS(4010),
    [anon_sym_LT_LT_EQ] = ACTIONS(4010),
    [anon_sym_GT_GT_EQ] = ACTIONS(4010),
    [anon_sym_AMP_EQ] = ACTIONS(4010),
    [anon_sym_CARET_EQ] = ACTIONS(4010),
    [anon_sym_PIPE_EQ] = ACTIONS(4010),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4008),
    [anon_sym___typeof] = ACTIONS(4008),
    [anon_sym_typeof] = ACTIONS(4008),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4008),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4008),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4008),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4008),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4008),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4008),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4008),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4008),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4008),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4008),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4008),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4008),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4008),
    [anon_sym_API_AVAILABLE] = ACTIONS(4008),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4008),
    [anon_sym_API_DEPRECATED] = ACTIONS(4008),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4008),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4008),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4008),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4008),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4008),
    [anon_sym___deprecated_msg] = ACTIONS(4008),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4008),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4008),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4008),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4008),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4008),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4008),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4008),
    [anon_sym_BOOL] = ACTIONS(4008),
    [anon_sym_IMP] = ACTIONS(4008),
    [anon_sym_SEL] = ACTIONS(4008),
    [anon_sym_Class] = ACTIONS(4008),
    [anon_sym_id] = ACTIONS(4008),
  },
  [STATE(1442)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4012),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4014),
    [anon_sym_COMMA] = ACTIONS(4014),
    [anon_sym_RPAREN] = ACTIONS(4014),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4014),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4014),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4012),
    [anon_sym_PLUS] = ACTIONS(4012),
    [anon_sym_STAR] = ACTIONS(4012),
    [anon_sym_SLASH] = ACTIONS(4012),
    [anon_sym_PERCENT] = ACTIONS(4012),
    [anon_sym_PIPE_PIPE] = ACTIONS(4014),
    [anon_sym_AMP_AMP] = ACTIONS(4014),
    [anon_sym_PIPE] = ACTIONS(4012),
    [anon_sym_CARET] = ACTIONS(4012),
    [anon_sym_AMP] = ACTIONS(4012),
    [anon_sym_EQ_EQ] = ACTIONS(4014),
    [anon_sym_BANG_EQ] = ACTIONS(4014),
    [anon_sym_GT] = ACTIONS(4012),
    [anon_sym_GT_EQ] = ACTIONS(4014),
    [anon_sym_LT_EQ] = ACTIONS(4014),
    [anon_sym_LT] = ACTIONS(4012),
    [anon_sym_LT_LT] = ACTIONS(4012),
    [anon_sym_GT_GT] = ACTIONS(4012),
    [anon_sym_SEMI] = ACTIONS(4014),
    [anon_sym___extension__] = ACTIONS(4012),
    [anon_sym_extern] = ACTIONS(4012),
    [anon_sym___attribute__] = ACTIONS(4012),
    [anon_sym___attribute] = ACTIONS(4012),
    [anon_sym_noreturn] = ACTIONS(4012),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(4014),
    [anon_sym___declspec] = ACTIONS(4012),
    [anon_sym_LBRACE] = ACTIONS(4014),
    [anon_sym_RBRACE] = ACTIONS(4014),
    [anon_sym_signed] = ACTIONS(4012),
    [anon_sym_unsigned] = ACTIONS(4012),
    [anon_sym_long] = ACTIONS(4012),
    [anon_sym_short] = ACTIONS(4012),
    [anon_sym_static] = ACTIONS(4012),
    [anon_sym_EQ] = ACTIONS(4012),
    [anon_sym_ATautoreleasepool] = ACTIONS(4014),
    [anon_sym_auto] = ACTIONS(4012),
    [anon_sym_register] = ACTIONS(4012),
    [anon_sym_inline] = ACTIONS(4012),
    [anon_sym___inline] = ACTIONS(4012),
    [anon_sym___inline__] = ACTIONS(4012),
    [anon_sym___forceinline] = ACTIONS(4012),
    [anon_sym_thread_local] = ACTIONS(4012),
    [anon_sym___thread] = ACTIONS(4012),
    [anon_sym_CG_EXTERN] = ACTIONS(4012),
    [anon_sym_CG_INLINE] = ACTIONS(4012),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4012),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4012),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4012),
    [anon_sym_IBOutlet] = ACTIONS(4012),
    [anon_sym_IBInspectable] = ACTIONS(4012),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4012),
    [anon_sym_NS_INLINE] = ACTIONS(4012),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4012),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4012),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4012),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4012),
    [anon_sym_const] = ACTIONS(4012),
    [anon_sym_constexpr] = ACTIONS(4012),
    [anon_sym_volatile] = ACTIONS(4012),
    [anon_sym_restrict] = ACTIONS(4012),
    [anon_sym___restrict__] = ACTIONS(4012),
    [anon_sym__Atomic] = ACTIONS(4012),
    [anon_sym__Noreturn] = ACTIONS(4012),
    [anon_sym__Nonnull] = ACTIONS(4012),
    [anon_sym_nullable] = ACTIONS(4012),
    [anon_sym__Complex] = ACTIONS(4012),
    [anon_sym__Nullable] = ACTIONS(4012),
    [anon_sym__Nullable_result] = ACTIONS(4012),
    [anon_sym__Null_unspecified] = ACTIONS(4012),
    [anon_sym___autoreleasing] = ACTIONS(4012),
    [anon_sym___block] = ACTIONS(4012),
    [anon_sym___bridge] = ACTIONS(4012),
    [anon_sym___bridge_retained] = ACTIONS(4012),
    [anon_sym___bridge_transfer] = ACTIONS(4012),
    [anon_sym___complex] = ACTIONS(4012),
    [anon_sym___const] = ACTIONS(4012),
    [anon_sym___imag] = ACTIONS(4012),
    [anon_sym___kindof] = ACTIONS(4012),
    [anon_sym___nonnull] = ACTIONS(4012),
    [anon_sym___nullable] = ACTIONS(4012),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4012),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4012),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4012),
    [anon_sym___real] = ACTIONS(4012),
    [anon_sym___strong] = ACTIONS(4012),
    [anon_sym___unsafe_unretained] = ACTIONS(4012),
    [anon_sym___unused] = ACTIONS(4012),
    [anon_sym___weak] = ACTIONS(4012),
    [anon_sym_alignas] = ACTIONS(4012),
    [anon_sym__Alignas] = ACTIONS(4012),
    [sym_primitive_type] = ACTIONS(4012),
    [anon_sym_enum] = ACTIONS(4012),
    [anon_sym_COLON] = ACTIONS(4014),
    [anon_sym_struct] = ACTIONS(4012),
    [anon_sym_union] = ACTIONS(4012),
    [anon_sym_QMARK] = ACTIONS(4014),
    [anon_sym_STAR_EQ] = ACTIONS(4014),
    [anon_sym_SLASH_EQ] = ACTIONS(4014),
    [anon_sym_PERCENT_EQ] = ACTIONS(4014),
    [anon_sym_PLUS_EQ] = ACTIONS(4014),
    [anon_sym_DASH_EQ] = ACTIONS(4014),
    [anon_sym_LT_LT_EQ] = ACTIONS(4014),
    [anon_sym_GT_GT_EQ] = ACTIONS(4014),
    [anon_sym_AMP_EQ] = ACTIONS(4014),
    [anon_sym_CARET_EQ] = ACTIONS(4014),
    [anon_sym_PIPE_EQ] = ACTIONS(4014),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4012),
    [anon_sym___typeof] = ACTIONS(4012),
    [anon_sym_typeof] = ACTIONS(4012),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4012),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4012),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4012),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4012),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4012),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4012),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4012),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4012),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4012),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4012),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4012),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4012),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4012),
    [anon_sym_API_AVAILABLE] = ACTIONS(4012),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4012),
    [anon_sym_API_DEPRECATED] = ACTIONS(4012),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4012),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4012),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4012),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4012),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4012),
    [anon_sym___deprecated_msg] = ACTIONS(4012),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4012),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4012),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4012),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4012),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4012),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4012),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4012),
    [anon_sym_BOOL] = ACTIONS(4012),
    [anon_sym_IMP] = ACTIONS(4012),
    [anon_sym_SEL] = ACTIONS(4012),
    [anon_sym_Class] = ACTIONS(4012),
    [anon_sym_id] = ACTIONS(4012),
  },
  [STATE(1443)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4016),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4018),
    [anon_sym_COMMA] = ACTIONS(4018),
    [anon_sym_RPAREN] = ACTIONS(4018),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4018),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4018),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4016),
    [anon_sym_PLUS] = ACTIONS(4016),
    [anon_sym_STAR] = ACTIONS(4016),
    [anon_sym_SLASH] = ACTIONS(4016),
    [anon_sym_PERCENT] = ACTIONS(4016),
    [anon_sym_PIPE_PIPE] = ACTIONS(4018),
    [anon_sym_AMP_AMP] = ACTIONS(4018),
    [anon_sym_PIPE] = ACTIONS(4016),
    [anon_sym_CARET] = ACTIONS(4016),
    [anon_sym_AMP] = ACTIONS(4016),
    [anon_sym_EQ_EQ] = ACTIONS(4018),
    [anon_sym_BANG_EQ] = ACTIONS(4018),
    [anon_sym_GT] = ACTIONS(4016),
    [anon_sym_GT_EQ] = ACTIONS(4018),
    [anon_sym_LT_EQ] = ACTIONS(4018),
    [anon_sym_LT] = ACTIONS(4016),
    [anon_sym_LT_LT] = ACTIONS(4016),
    [anon_sym_GT_GT] = ACTIONS(4016),
    [anon_sym_SEMI] = ACTIONS(4018),
    [anon_sym___extension__] = ACTIONS(4016),
    [anon_sym_extern] = ACTIONS(4016),
    [anon_sym___attribute__] = ACTIONS(4016),
    [anon_sym___attribute] = ACTIONS(4016),
    [anon_sym_noreturn] = ACTIONS(4016),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(4018),
    [anon_sym___declspec] = ACTIONS(4016),
    [anon_sym_LBRACE] = ACTIONS(4018),
    [anon_sym_RBRACE] = ACTIONS(4018),
    [anon_sym_signed] = ACTIONS(4016),
    [anon_sym_unsigned] = ACTIONS(4016),
    [anon_sym_long] = ACTIONS(4016),
    [anon_sym_short] = ACTIONS(4016),
    [anon_sym_static] = ACTIONS(4016),
    [anon_sym_EQ] = ACTIONS(4016),
    [anon_sym_ATautoreleasepool] = ACTIONS(4018),
    [anon_sym_auto] = ACTIONS(4016),
    [anon_sym_register] = ACTIONS(4016),
    [anon_sym_inline] = ACTIONS(4016),
    [anon_sym___inline] = ACTIONS(4016),
    [anon_sym___inline__] = ACTIONS(4016),
    [anon_sym___forceinline] = ACTIONS(4016),
    [anon_sym_thread_local] = ACTIONS(4016),
    [anon_sym___thread] = ACTIONS(4016),
    [anon_sym_CG_EXTERN] = ACTIONS(4016),
    [anon_sym_CG_INLINE] = ACTIONS(4016),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4016),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4016),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4016),
    [anon_sym_IBOutlet] = ACTIONS(4016),
    [anon_sym_IBInspectable] = ACTIONS(4016),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4016),
    [anon_sym_NS_INLINE] = ACTIONS(4016),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4016),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4016),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4016),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4016),
    [anon_sym_const] = ACTIONS(4016),
    [anon_sym_constexpr] = ACTIONS(4016),
    [anon_sym_volatile] = ACTIONS(4016),
    [anon_sym_restrict] = ACTIONS(4016),
    [anon_sym___restrict__] = ACTIONS(4016),
    [anon_sym__Atomic] = ACTIONS(4016),
    [anon_sym__Noreturn] = ACTIONS(4016),
    [anon_sym__Nonnull] = ACTIONS(4016),
    [anon_sym_nullable] = ACTIONS(4016),
    [anon_sym__Complex] = ACTIONS(4016),
    [anon_sym__Nullable] = ACTIONS(4016),
    [anon_sym__Nullable_result] = ACTIONS(4016),
    [anon_sym__Null_unspecified] = ACTIONS(4016),
    [anon_sym___autoreleasing] = ACTIONS(4016),
    [anon_sym___block] = ACTIONS(4016),
    [anon_sym___bridge] = ACTIONS(4016),
    [anon_sym___bridge_retained] = ACTIONS(4016),
    [anon_sym___bridge_transfer] = ACTIONS(4016),
    [anon_sym___complex] = ACTIONS(4016),
    [anon_sym___const] = ACTIONS(4016),
    [anon_sym___imag] = ACTIONS(4016),
    [anon_sym___kindof] = ACTIONS(4016),
    [anon_sym___nonnull] = ACTIONS(4016),
    [anon_sym___nullable] = ACTIONS(4016),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4016),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4016),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4016),
    [anon_sym___real] = ACTIONS(4016),
    [anon_sym___strong] = ACTIONS(4016),
    [anon_sym___unsafe_unretained] = ACTIONS(4016),
    [anon_sym___unused] = ACTIONS(4016),
    [anon_sym___weak] = ACTIONS(4016),
    [anon_sym_alignas] = ACTIONS(4016),
    [anon_sym__Alignas] = ACTIONS(4016),
    [sym_primitive_type] = ACTIONS(4016),
    [anon_sym_enum] = ACTIONS(4016),
    [anon_sym_COLON] = ACTIONS(4018),
    [anon_sym_struct] = ACTIONS(4016),
    [anon_sym_union] = ACTIONS(4016),
    [anon_sym_QMARK] = ACTIONS(4018),
    [anon_sym_STAR_EQ] = ACTIONS(4018),
    [anon_sym_SLASH_EQ] = ACTIONS(4018),
    [anon_sym_PERCENT_EQ] = ACTIONS(4018),
    [anon_sym_PLUS_EQ] = ACTIONS(4018),
    [anon_sym_DASH_EQ] = ACTIONS(4018),
    [anon_sym_LT_LT_EQ] = ACTIONS(4018),
    [anon_sym_GT_GT_EQ] = ACTIONS(4018),
    [anon_sym_AMP_EQ] = ACTIONS(4018),
    [anon_sym_CARET_EQ] = ACTIONS(4018),
    [anon_sym_PIPE_EQ] = ACTIONS(4018),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4016),
    [anon_sym___typeof] = ACTIONS(4016),
    [anon_sym_typeof] = ACTIONS(4016),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4016),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4016),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4016),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4016),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4016),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4016),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4016),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4016),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4016),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4016),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4016),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4016),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4016),
    [anon_sym_API_AVAILABLE] = ACTIONS(4016),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4016),
    [anon_sym_API_DEPRECATED] = ACTIONS(4016),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4016),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4016),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4016),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4016),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4016),
    [anon_sym___deprecated_msg] = ACTIONS(4016),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4016),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4016),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4016),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4016),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4016),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4016),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4016),
    [anon_sym_BOOL] = ACTIONS(4016),
    [anon_sym_IMP] = ACTIONS(4016),
    [anon_sym_SEL] = ACTIONS(4016),
    [anon_sym_Class] = ACTIONS(4016),
    [anon_sym_id] = ACTIONS(4016),
  },
  [STATE(1444)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4020),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4022),
    [anon_sym_COMMA] = ACTIONS(4022),
    [anon_sym_RPAREN] = ACTIONS(4022),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4022),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4022),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4020),
    [anon_sym_PLUS] = ACTIONS(4020),
    [anon_sym_STAR] = ACTIONS(4020),
    [anon_sym_SLASH] = ACTIONS(4020),
    [anon_sym_PERCENT] = ACTIONS(4020),
    [anon_sym_PIPE_PIPE] = ACTIONS(4022),
    [anon_sym_AMP_AMP] = ACTIONS(4022),
    [anon_sym_PIPE] = ACTIONS(4020),
    [anon_sym_CARET] = ACTIONS(4020),
    [anon_sym_AMP] = ACTIONS(4020),
    [anon_sym_EQ_EQ] = ACTIONS(4022),
    [anon_sym_BANG_EQ] = ACTIONS(4022),
    [anon_sym_GT] = ACTIONS(4020),
    [anon_sym_GT_EQ] = ACTIONS(4022),
    [anon_sym_LT_EQ] = ACTIONS(4022),
    [anon_sym_LT] = ACTIONS(4020),
    [anon_sym_LT_LT] = ACTIONS(4020),
    [anon_sym_GT_GT] = ACTIONS(4020),
    [anon_sym_SEMI] = ACTIONS(4022),
    [anon_sym___extension__] = ACTIONS(4020),
    [anon_sym_extern] = ACTIONS(4020),
    [anon_sym___attribute__] = ACTIONS(4020),
    [anon_sym___attribute] = ACTIONS(4020),
    [anon_sym_noreturn] = ACTIONS(4020),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(4022),
    [anon_sym___declspec] = ACTIONS(4020),
    [anon_sym_LBRACE] = ACTIONS(4022),
    [anon_sym_RBRACE] = ACTIONS(4022),
    [anon_sym_signed] = ACTIONS(4020),
    [anon_sym_unsigned] = ACTIONS(4020),
    [anon_sym_long] = ACTIONS(4020),
    [anon_sym_short] = ACTIONS(4020),
    [anon_sym_static] = ACTIONS(4020),
    [anon_sym_EQ] = ACTIONS(4020),
    [anon_sym_ATautoreleasepool] = ACTIONS(4022),
    [anon_sym_auto] = ACTIONS(4020),
    [anon_sym_register] = ACTIONS(4020),
    [anon_sym_inline] = ACTIONS(4020),
    [anon_sym___inline] = ACTIONS(4020),
    [anon_sym___inline__] = ACTIONS(4020),
    [anon_sym___forceinline] = ACTIONS(4020),
    [anon_sym_thread_local] = ACTIONS(4020),
    [anon_sym___thread] = ACTIONS(4020),
    [anon_sym_CG_EXTERN] = ACTIONS(4020),
    [anon_sym_CG_INLINE] = ACTIONS(4020),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4020),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4020),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4020),
    [anon_sym_IBOutlet] = ACTIONS(4020),
    [anon_sym_IBInspectable] = ACTIONS(4020),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4020),
    [anon_sym_NS_INLINE] = ACTIONS(4020),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4020),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4020),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4020),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4020),
    [anon_sym_const] = ACTIONS(4020),
    [anon_sym_constexpr] = ACTIONS(4020),
    [anon_sym_volatile] = ACTIONS(4020),
    [anon_sym_restrict] = ACTIONS(4020),
    [anon_sym___restrict__] = ACTIONS(4020),
    [anon_sym__Atomic] = ACTIONS(4020),
    [anon_sym__Noreturn] = ACTIONS(4020),
    [anon_sym__Nonnull] = ACTIONS(4020),
    [anon_sym_nullable] = ACTIONS(4020),
    [anon_sym__Complex] = ACTIONS(4020),
    [anon_sym__Nullable] = ACTIONS(4020),
    [anon_sym__Nullable_result] = ACTIONS(4020),
    [anon_sym__Null_unspecified] = ACTIONS(4020),
    [anon_sym___autoreleasing] = ACTIONS(4020),
    [anon_sym___block] = ACTIONS(4020),
    [anon_sym___bridge] = ACTIONS(4020),
    [anon_sym___bridge_retained] = ACTIONS(4020),
    [anon_sym___bridge_transfer] = ACTIONS(4020),
    [anon_sym___complex] = ACTIONS(4020),
    [anon_sym___const] = ACTIONS(4020),
    [anon_sym___imag] = ACTIONS(4020),
    [anon_sym___kindof] = ACTIONS(4020),
    [anon_sym___nonnull] = ACTIONS(4020),
    [anon_sym___nullable] = ACTIONS(4020),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4020),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4020),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4020),
    [anon_sym___real] = ACTIONS(4020),
    [anon_sym___strong] = ACTIONS(4020),
    [anon_sym___unsafe_unretained] = ACTIONS(4020),
    [anon_sym___unused] = ACTIONS(4020),
    [anon_sym___weak] = ACTIONS(4020),
    [anon_sym_alignas] = ACTIONS(4020),
    [anon_sym__Alignas] = ACTIONS(4020),
    [sym_primitive_type] = ACTIONS(4020),
    [anon_sym_enum] = ACTIONS(4020),
    [anon_sym_COLON] = ACTIONS(4022),
    [anon_sym_struct] = ACTIONS(4020),
    [anon_sym_union] = ACTIONS(4020),
    [anon_sym_QMARK] = ACTIONS(4022),
    [anon_sym_STAR_EQ] = ACTIONS(4022),
    [anon_sym_SLASH_EQ] = ACTIONS(4022),
    [anon_sym_PERCENT_EQ] = ACTIONS(4022),
    [anon_sym_PLUS_EQ] = ACTIONS(4022),
    [anon_sym_DASH_EQ] = ACTIONS(4022),
    [anon_sym_LT_LT_EQ] = ACTIONS(4022),
    [anon_sym_GT_GT_EQ] = ACTIONS(4022),
    [anon_sym_AMP_EQ] = ACTIONS(4022),
    [anon_sym_CARET_EQ] = ACTIONS(4022),
    [anon_sym_PIPE_EQ] = ACTIONS(4022),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4020),
    [anon_sym___typeof] = ACTIONS(4020),
    [anon_sym_typeof] = ACTIONS(4020),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4020),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4020),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4020),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4020),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4020),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4020),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4020),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4020),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4020),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4020),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4020),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4020),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4020),
    [anon_sym_API_AVAILABLE] = ACTIONS(4020),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4020),
    [anon_sym_API_DEPRECATED] = ACTIONS(4020),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4020),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4020),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4020),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4020),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4020),
    [anon_sym___deprecated_msg] = ACTIONS(4020),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4020),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4020),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4020),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4020),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4020),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4020),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4020),
    [anon_sym_BOOL] = ACTIONS(4020),
    [anon_sym_IMP] = ACTIONS(4020),
    [anon_sym_SEL] = ACTIONS(4020),
    [anon_sym_Class] = ACTIONS(4020),
    [anon_sym_id] = ACTIONS(4020),
  },
  [STATE(1445)] = {
    [sym_string_literal] = STATE(1601),
    [sym_identifier] = ACTIONS(4024),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4024),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4024),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4024),
    [anon_sym_extern] = ACTIONS(4024),
    [anon_sym___attribute__] = ACTIONS(4024),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4024),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4024),
    [anon_sym_LBRACE] = ACTIONS(4026),
    [anon_sym_signed] = ACTIONS(4024),
    [anon_sym_unsigned] = ACTIONS(4024),
    [anon_sym_long] = ACTIONS(4024),
    [anon_sym_short] = ACTIONS(4024),
    [anon_sym_static] = ACTIONS(4024),
    [anon_sym_EQ] = ACTIONS(4028),
    [anon_sym_ATautoreleasepool] = ACTIONS(4026),
    [anon_sym_auto] = ACTIONS(4024),
    [anon_sym_register] = ACTIONS(4024),
    [anon_sym_inline] = ACTIONS(4024),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4024),
    [anon_sym___forceinline] = ACTIONS(4024),
    [anon_sym_thread_local] = ACTIONS(4024),
    [anon_sym___thread] = ACTIONS(4024),
    [anon_sym_CG_EXTERN] = ACTIONS(4024),
    [anon_sym_CG_INLINE] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4024),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4024),
    [anon_sym_IBOutlet] = ACTIONS(4024),
    [anon_sym_IBInspectable] = ACTIONS(4024),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4024),
    [anon_sym_NS_INLINE] = ACTIONS(4024),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4024),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4024),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4024),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4024),
    [anon_sym_volatile] = ACTIONS(4024),
    [anon_sym_restrict] = ACTIONS(4024),
    [anon_sym___restrict__] = ACTIONS(4024),
    [anon_sym__Atomic] = ACTIONS(4024),
    [anon_sym__Noreturn] = ACTIONS(4024),
    [anon_sym__Nonnull] = ACTIONS(4024),
    [anon_sym_nullable] = ACTIONS(4024),
    [anon_sym__Complex] = ACTIONS(4024),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4024),
    [anon_sym__Null_unspecified] = ACTIONS(4024),
    [anon_sym___autoreleasing] = ACTIONS(4024),
    [anon_sym___block] = ACTIONS(4024),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4024),
    [anon_sym___bridge_transfer] = ACTIONS(4024),
    [anon_sym___complex] = ACTIONS(4024),
    [anon_sym___const] = ACTIONS(4024),
    [anon_sym___imag] = ACTIONS(4024),
    [anon_sym___kindof] = ACTIONS(4024),
    [anon_sym___nonnull] = ACTIONS(4024),
    [anon_sym___nullable] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4024),
    [anon_sym___real] = ACTIONS(4024),
    [anon_sym___strong] = ACTIONS(4024),
    [anon_sym___unsafe_unretained] = ACTIONS(4024),
    [anon_sym___unused] = ACTIONS(4024),
    [anon_sym___weak] = ACTIONS(4024),
    [anon_sym_alignas] = ACTIONS(4024),
    [anon_sym__Alignas] = ACTIONS(4024),
    [sym_primitive_type] = ACTIONS(4024),
    [anon_sym_enum] = ACTIONS(4024),
    [anon_sym_struct] = ACTIONS(4024),
    [anon_sym_union] = ACTIONS(4024),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4030),
    [anon_sym_SLASH_EQ] = ACTIONS(4030),
    [anon_sym_PERCENT_EQ] = ACTIONS(4030),
    [anon_sym_PLUS_EQ] = ACTIONS(4030),
    [anon_sym_DASH_EQ] = ACTIONS(4030),
    [anon_sym_LT_LT_EQ] = ACTIONS(4030),
    [anon_sym_GT_GT_EQ] = ACTIONS(4030),
    [anon_sym_AMP_EQ] = ACTIONS(4030),
    [anon_sym_CARET_EQ] = ACTIONS(4030),
    [anon_sym_PIPE_EQ] = ACTIONS(4030),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4032),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4024),
    [anon_sym___typeof] = ACTIONS(4024),
    [anon_sym_typeof] = ACTIONS(4024),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4024),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4024),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_API_AVAILABLE] = ACTIONS(4024),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_API_DEPRECATED] = ACTIONS(4024),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym___deprecated_msg] = ACTIONS(4024),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4024),
    [anon_sym_BOOL] = ACTIONS(4024),
    [anon_sym_IMP] = ACTIONS(4024),
    [anon_sym_SEL] = ACTIONS(4024),
    [anon_sym_Class] = ACTIONS(4024),
    [anon_sym_id] = ACTIONS(4024),
  },
  [STATE(1446)] = {
    [sym_string_literal] = STATE(1434),
    [sym_identifier] = ACTIONS(4024),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4024),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4024),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4024),
    [anon_sym_extern] = ACTIONS(4024),
    [anon_sym___attribute__] = ACTIONS(4024),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4024),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4024),
    [anon_sym_LBRACE] = ACTIONS(4026),
    [anon_sym_signed] = ACTIONS(4024),
    [anon_sym_unsigned] = ACTIONS(4024),
    [anon_sym_long] = ACTIONS(4024),
    [anon_sym_short] = ACTIONS(4024),
    [anon_sym_static] = ACTIONS(4024),
    [anon_sym_EQ] = ACTIONS(4024),
    [anon_sym_ATautoreleasepool] = ACTIONS(4026),
    [anon_sym_auto] = ACTIONS(4024),
    [anon_sym_register] = ACTIONS(4024),
    [anon_sym_inline] = ACTIONS(4024),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4024),
    [anon_sym___forceinline] = ACTIONS(4024),
    [anon_sym_thread_local] = ACTIONS(4024),
    [anon_sym___thread] = ACTIONS(4024),
    [anon_sym_CG_EXTERN] = ACTIONS(4024),
    [anon_sym_CG_INLINE] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4024),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4024),
    [anon_sym_IBOutlet] = ACTIONS(4024),
    [anon_sym_IBInspectable] = ACTIONS(4024),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4024),
    [anon_sym_NS_INLINE] = ACTIONS(4024),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4024),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4024),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4024),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4024),
    [anon_sym_volatile] = ACTIONS(4024),
    [anon_sym_restrict] = ACTIONS(4024),
    [anon_sym___restrict__] = ACTIONS(4024),
    [anon_sym__Atomic] = ACTIONS(4024),
    [anon_sym__Noreturn] = ACTIONS(4024),
    [anon_sym__Nonnull] = ACTIONS(4024),
    [anon_sym_nullable] = ACTIONS(4024),
    [anon_sym__Complex] = ACTIONS(4024),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4024),
    [anon_sym__Null_unspecified] = ACTIONS(4024),
    [anon_sym___autoreleasing] = ACTIONS(4024),
    [anon_sym___block] = ACTIONS(4024),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4024),
    [anon_sym___bridge_transfer] = ACTIONS(4024),
    [anon_sym___complex] = ACTIONS(4024),
    [anon_sym___const] = ACTIONS(4024),
    [anon_sym___imag] = ACTIONS(4024),
    [anon_sym___kindof] = ACTIONS(4024),
    [anon_sym___nonnull] = ACTIONS(4024),
    [anon_sym___nullable] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4024),
    [anon_sym___real] = ACTIONS(4024),
    [anon_sym___strong] = ACTIONS(4024),
    [anon_sym___unsafe_unretained] = ACTIONS(4024),
    [anon_sym___unused] = ACTIONS(4024),
    [anon_sym___weak] = ACTIONS(4024),
    [anon_sym_alignas] = ACTIONS(4024),
    [anon_sym__Alignas] = ACTIONS(4024),
    [sym_primitive_type] = ACTIONS(4024),
    [anon_sym_enum] = ACTIONS(4024),
    [anon_sym_struct] = ACTIONS(4024),
    [anon_sym_union] = ACTIONS(4024),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4026),
    [anon_sym_SLASH_EQ] = ACTIONS(4026),
    [anon_sym_PERCENT_EQ] = ACTIONS(4026),
    [anon_sym_PLUS_EQ] = ACTIONS(4026),
    [anon_sym_DASH_EQ] = ACTIONS(4026),
    [anon_sym_LT_LT_EQ] = ACTIONS(4026),
    [anon_sym_GT_GT_EQ] = ACTIONS(4026),
    [anon_sym_AMP_EQ] = ACTIONS(4026),
    [anon_sym_CARET_EQ] = ACTIONS(4026),
    [anon_sym_PIPE_EQ] = ACTIONS(4026),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(3948),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4024),
    [anon_sym___typeof] = ACTIONS(4024),
    [anon_sym_typeof] = ACTIONS(4024),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4024),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4024),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_API_AVAILABLE] = ACTIONS(4024),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_API_DEPRECATED] = ACTIONS(4024),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym___deprecated_msg] = ACTIONS(4024),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4024),
    [anon_sym_BOOL] = ACTIONS(4024),
    [anon_sym_IMP] = ACTIONS(4024),
    [anon_sym_SEL] = ACTIONS(4024),
    [anon_sym_Class] = ACTIONS(4024),
    [anon_sym_id] = ACTIONS(4024),
  },
  [STATE(1447)] = {
    [sym_identifier] = ACTIONS(4034),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4036),
    [anon_sym_COMMA] = ACTIONS(4036),
    [anon_sym_RPAREN] = ACTIONS(4036),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4036),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4036),
    [anon_sym_LPAREN2] = ACTIONS(4036),
    [anon_sym_DASH] = ACTIONS(4034),
    [anon_sym_PLUS] = ACTIONS(4034),
    [anon_sym_STAR] = ACTIONS(4034),
    [anon_sym_SLASH] = ACTIONS(4034),
    [anon_sym_PERCENT] = ACTIONS(4034),
    [anon_sym_PIPE_PIPE] = ACTIONS(4036),
    [anon_sym_AMP_AMP] = ACTIONS(4036),
    [anon_sym_PIPE] = ACTIONS(4034),
    [anon_sym_CARET] = ACTIONS(4034),
    [anon_sym_AMP] = ACTIONS(4034),
    [anon_sym_EQ_EQ] = ACTIONS(4036),
    [anon_sym_BANG_EQ] = ACTIONS(4036),
    [anon_sym_GT] = ACTIONS(4034),
    [anon_sym_GT_EQ] = ACTIONS(4036),
    [anon_sym_LT_EQ] = ACTIONS(4036),
    [anon_sym_LT] = ACTIONS(4034),
    [anon_sym_LT_LT] = ACTIONS(4034),
    [anon_sym_GT_GT] = ACTIONS(4034),
    [anon_sym_SEMI] = ACTIONS(4036),
    [anon_sym___extension__] = ACTIONS(4034),
    [anon_sym_extern] = ACTIONS(4034),
    [anon_sym___attribute__] = ACTIONS(4034),
    [anon_sym___attribute] = ACTIONS(4034),
    [anon_sym_noreturn] = ACTIONS(4034),
    [anon_sym_LBRACK] = ACTIONS(4036),
    [anon_sym_RBRACK] = ACTIONS(4036),
    [anon_sym___declspec] = ACTIONS(4034),
    [anon_sym_LBRACE] = ACTIONS(4036),
    [anon_sym_RBRACE] = ACTIONS(4036),
    [anon_sym_signed] = ACTIONS(4034),
    [anon_sym_unsigned] = ACTIONS(4034),
    [anon_sym_long] = ACTIONS(4034),
    [anon_sym_short] = ACTIONS(4034),
    [anon_sym_static] = ACTIONS(4034),
    [anon_sym_EQ] = ACTIONS(4034),
    [anon_sym_ATautoreleasepool] = ACTIONS(4036),
    [anon_sym_auto] = ACTIONS(4034),
    [anon_sym_register] = ACTIONS(4034),
    [anon_sym_inline] = ACTIONS(4034),
    [anon_sym___inline] = ACTIONS(4034),
    [anon_sym___inline__] = ACTIONS(4034),
    [anon_sym___forceinline] = ACTIONS(4034),
    [anon_sym_thread_local] = ACTIONS(4034),
    [anon_sym___thread] = ACTIONS(4034),
    [anon_sym_CG_EXTERN] = ACTIONS(4034),
    [anon_sym_CG_INLINE] = ACTIONS(4034),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4034),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4034),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4034),
    [anon_sym_IBOutlet] = ACTIONS(4034),
    [anon_sym_IBInspectable] = ACTIONS(4034),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4034),
    [anon_sym_NS_INLINE] = ACTIONS(4034),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4034),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4034),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4034),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4034),
    [anon_sym_const] = ACTIONS(4034),
    [anon_sym_constexpr] = ACTIONS(4034),
    [anon_sym_volatile] = ACTIONS(4034),
    [anon_sym_restrict] = ACTIONS(4034),
    [anon_sym___restrict__] = ACTIONS(4034),
    [anon_sym__Atomic] = ACTIONS(4034),
    [anon_sym__Noreturn] = ACTIONS(4034),
    [anon_sym__Nonnull] = ACTIONS(4034),
    [anon_sym_nullable] = ACTIONS(4034),
    [anon_sym__Complex] = ACTIONS(4034),
    [anon_sym__Nullable] = ACTIONS(4034),
    [anon_sym__Nullable_result] = ACTIONS(4034),
    [anon_sym__Null_unspecified] = ACTIONS(4034),
    [anon_sym___autoreleasing] = ACTIONS(4034),
    [anon_sym___block] = ACTIONS(4034),
    [anon_sym___bridge] = ACTIONS(4034),
    [anon_sym___bridge_retained] = ACTIONS(4034),
    [anon_sym___bridge_transfer] = ACTIONS(4034),
    [anon_sym___complex] = ACTIONS(4034),
    [anon_sym___const] = ACTIONS(4034),
    [anon_sym___imag] = ACTIONS(4034),
    [anon_sym___kindof] = ACTIONS(4034),
    [anon_sym___nonnull] = ACTIONS(4034),
    [anon_sym___nullable] = ACTIONS(4034),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4034),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4034),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4034),
    [anon_sym___real] = ACTIONS(4034),
    [anon_sym___strong] = ACTIONS(4034),
    [anon_sym___unsafe_unretained] = ACTIONS(4034),
    [anon_sym___unused] = ACTIONS(4034),
    [anon_sym___weak] = ACTIONS(4034),
    [anon_sym_alignas] = ACTIONS(4034),
    [anon_sym__Alignas] = ACTIONS(4034),
    [sym_primitive_type] = ACTIONS(4034),
    [anon_sym_enum] = ACTIONS(4034),
    [anon_sym_COLON] = ACTIONS(4036),
    [anon_sym_struct] = ACTIONS(4034),
    [anon_sym_union] = ACTIONS(4034),
    [anon_sym_QMARK] = ACTIONS(4036),
    [anon_sym_STAR_EQ] = ACTIONS(4036),
    [anon_sym_SLASH_EQ] = ACTIONS(4036),
    [anon_sym_PERCENT_EQ] = ACTIONS(4036),
    [anon_sym_PLUS_EQ] = ACTIONS(4036),
    [anon_sym_DASH_EQ] = ACTIONS(4036),
    [anon_sym_LT_LT_EQ] = ACTIONS(4036),
    [anon_sym_GT_GT_EQ] = ACTIONS(4036),
    [anon_sym_AMP_EQ] = ACTIONS(4036),
    [anon_sym_CARET_EQ] = ACTIONS(4036),
    [anon_sym_PIPE_EQ] = ACTIONS(4036),
    [anon_sym_DASH_DASH] = ACTIONS(4036),
    [anon_sym_PLUS_PLUS] = ACTIONS(4036),
    [anon_sym_DOT] = ACTIONS(4034),
    [anon_sym_DASH_GT] = ACTIONS(4036),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4034),
    [anon_sym___typeof] = ACTIONS(4034),
    [anon_sym_typeof] = ACTIONS(4034),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4034),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4034),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4034),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4034),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4034),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4034),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4034),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4034),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4034),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4034),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4034),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4034),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4034),
    [anon_sym_API_AVAILABLE] = ACTIONS(4034),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4034),
    [anon_sym_API_DEPRECATED] = ACTIONS(4034),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4034),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4034),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4034),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4034),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4034),
    [anon_sym___deprecated_msg] = ACTIONS(4034),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4034),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4034),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4034),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4034),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4034),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4034),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4034),
    [anon_sym_BOOL] = ACTIONS(4034),
    [anon_sym_IMP] = ACTIONS(4034),
    [anon_sym_SEL] = ACTIONS(4034),
    [anon_sym_Class] = ACTIONS(4034),
    [anon_sym_id] = ACTIONS(4034),
  },
  [STATE(1448)] = {
    [sym_identifier] = ACTIONS(4038),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4040),
    [anon_sym_COMMA] = ACTIONS(4040),
    [anon_sym_RPAREN] = ACTIONS(4040),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4040),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4040),
    [anon_sym_LPAREN2] = ACTIONS(4040),
    [anon_sym_DASH] = ACTIONS(4038),
    [anon_sym_PLUS] = ACTIONS(4038),
    [anon_sym_STAR] = ACTIONS(4038),
    [anon_sym_SLASH] = ACTIONS(4038),
    [anon_sym_PERCENT] = ACTIONS(4038),
    [anon_sym_PIPE_PIPE] = ACTIONS(4040),
    [anon_sym_AMP_AMP] = ACTIONS(4040),
    [anon_sym_PIPE] = ACTIONS(4038),
    [anon_sym_CARET] = ACTIONS(4038),
    [anon_sym_AMP] = ACTIONS(4038),
    [anon_sym_EQ_EQ] = ACTIONS(4040),
    [anon_sym_BANG_EQ] = ACTIONS(4040),
    [anon_sym_GT] = ACTIONS(4038),
    [anon_sym_GT_EQ] = ACTIONS(4040),
    [anon_sym_LT_EQ] = ACTIONS(4040),
    [anon_sym_LT] = ACTIONS(4038),
    [anon_sym_LT_LT] = ACTIONS(4038),
    [anon_sym_GT_GT] = ACTIONS(4038),
    [anon_sym_SEMI] = ACTIONS(4040),
    [anon_sym___extension__] = ACTIONS(4038),
    [anon_sym_extern] = ACTIONS(4038),
    [anon_sym___attribute__] = ACTIONS(4038),
    [anon_sym___attribute] = ACTIONS(4038),
    [anon_sym_noreturn] = ACTIONS(4038),
    [anon_sym_LBRACK] = ACTIONS(4040),
    [anon_sym_RBRACK] = ACTIONS(4040),
    [anon_sym___declspec] = ACTIONS(4038),
    [anon_sym_LBRACE] = ACTIONS(4040),
    [anon_sym_RBRACE] = ACTIONS(4040),
    [anon_sym_signed] = ACTIONS(4038),
    [anon_sym_unsigned] = ACTIONS(4038),
    [anon_sym_long] = ACTIONS(4038),
    [anon_sym_short] = ACTIONS(4038),
    [anon_sym_static] = ACTIONS(4038),
    [anon_sym_EQ] = ACTIONS(4038),
    [anon_sym_ATautoreleasepool] = ACTIONS(4040),
    [anon_sym_auto] = ACTIONS(4038),
    [anon_sym_register] = ACTIONS(4038),
    [anon_sym_inline] = ACTIONS(4038),
    [anon_sym___inline] = ACTIONS(4038),
    [anon_sym___inline__] = ACTIONS(4038),
    [anon_sym___forceinline] = ACTIONS(4038),
    [anon_sym_thread_local] = ACTIONS(4038),
    [anon_sym___thread] = ACTIONS(4038),
    [anon_sym_CG_EXTERN] = ACTIONS(4038),
    [anon_sym_CG_INLINE] = ACTIONS(4038),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4038),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4038),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4038),
    [anon_sym_IBOutlet] = ACTIONS(4038),
    [anon_sym_IBInspectable] = ACTIONS(4038),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4038),
    [anon_sym_NS_INLINE] = ACTIONS(4038),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4038),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4038),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4038),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4038),
    [anon_sym_const] = ACTIONS(4038),
    [anon_sym_constexpr] = ACTIONS(4038),
    [anon_sym_volatile] = ACTIONS(4038),
    [anon_sym_restrict] = ACTIONS(4038),
    [anon_sym___restrict__] = ACTIONS(4038),
    [anon_sym__Atomic] = ACTIONS(4038),
    [anon_sym__Noreturn] = ACTIONS(4038),
    [anon_sym__Nonnull] = ACTIONS(4038),
    [anon_sym_nullable] = ACTIONS(4038),
    [anon_sym__Complex] = ACTIONS(4038),
    [anon_sym__Nullable] = ACTIONS(4038),
    [anon_sym__Nullable_result] = ACTIONS(4038),
    [anon_sym__Null_unspecified] = ACTIONS(4038),
    [anon_sym___autoreleasing] = ACTIONS(4038),
    [anon_sym___block] = ACTIONS(4038),
    [anon_sym___bridge] = ACTIONS(4038),
    [anon_sym___bridge_retained] = ACTIONS(4038),
    [anon_sym___bridge_transfer] = ACTIONS(4038),
    [anon_sym___complex] = ACTIONS(4038),
    [anon_sym___const] = ACTIONS(4038),
    [anon_sym___imag] = ACTIONS(4038),
    [anon_sym___kindof] = ACTIONS(4038),
    [anon_sym___nonnull] = ACTIONS(4038),
    [anon_sym___nullable] = ACTIONS(4038),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4038),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4038),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4038),
    [anon_sym___real] = ACTIONS(4038),
    [anon_sym___strong] = ACTIONS(4038),
    [anon_sym___unsafe_unretained] = ACTIONS(4038),
    [anon_sym___unused] = ACTIONS(4038),
    [anon_sym___weak] = ACTIONS(4038),
    [anon_sym_alignas] = ACTIONS(4038),
    [anon_sym__Alignas] = ACTIONS(4038),
    [sym_primitive_type] = ACTIONS(4038),
    [anon_sym_enum] = ACTIONS(4038),
    [anon_sym_COLON] = ACTIONS(4040),
    [anon_sym_struct] = ACTIONS(4038),
    [anon_sym_union] = ACTIONS(4038),
    [anon_sym_QMARK] = ACTIONS(4040),
    [anon_sym_STAR_EQ] = ACTIONS(4040),
    [anon_sym_SLASH_EQ] = ACTIONS(4040),
    [anon_sym_PERCENT_EQ] = ACTIONS(4040),
    [anon_sym_PLUS_EQ] = ACTIONS(4040),
    [anon_sym_DASH_EQ] = ACTIONS(4040),
    [anon_sym_LT_LT_EQ] = ACTIONS(4040),
    [anon_sym_GT_GT_EQ] = ACTIONS(4040),
    [anon_sym_AMP_EQ] = ACTIONS(4040),
    [anon_sym_CARET_EQ] = ACTIONS(4040),
    [anon_sym_PIPE_EQ] = ACTIONS(4040),
    [anon_sym_DASH_DASH] = ACTIONS(4040),
    [anon_sym_PLUS_PLUS] = ACTIONS(4040),
    [anon_sym_DOT] = ACTIONS(4038),
    [anon_sym_DASH_GT] = ACTIONS(4040),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4038),
    [anon_sym___typeof] = ACTIONS(4038),
    [anon_sym_typeof] = ACTIONS(4038),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4038),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4038),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4038),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4038),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4038),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4038),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4038),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4038),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4038),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4038),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4038),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4038),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4038),
    [anon_sym_API_AVAILABLE] = ACTIONS(4038),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4038),
    [anon_sym_API_DEPRECATED] = ACTIONS(4038),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4038),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4038),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4038),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4038),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4038),
    [anon_sym___deprecated_msg] = ACTIONS(4038),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4038),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4038),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4038),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4038),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4038),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4038),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4038),
    [anon_sym_BOOL] = ACTIONS(4038),
    [anon_sym_IMP] = ACTIONS(4038),
    [anon_sym_SEL] = ACTIONS(4038),
    [anon_sym_Class] = ACTIONS(4038),
    [anon_sym_id] = ACTIONS(4038),
  },
  [STATE(1449)] = {
    [sym_identifier] = ACTIONS(4042),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_COMMA] = ACTIONS(4044),
    [anon_sym_RPAREN] = ACTIONS(4044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4044),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4044),
    [anon_sym_LPAREN2] = ACTIONS(4044),
    [anon_sym_DASH] = ACTIONS(4042),
    [anon_sym_PLUS] = ACTIONS(4042),
    [anon_sym_STAR] = ACTIONS(4042),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4042),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(4042),
    [anon_sym_AMP] = ACTIONS(4042),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(4042),
    [anon_sym_LT_LT] = ACTIONS(4042),
    [anon_sym_GT_GT] = ACTIONS(4042),
    [anon_sym_SEMI] = ACTIONS(4044),
    [anon_sym___extension__] = ACTIONS(4042),
    [anon_sym_extern] = ACTIONS(4042),
    [anon_sym___attribute__] = ACTIONS(4042),
    [anon_sym___attribute] = ACTIONS(4042),
    [anon_sym_noreturn] = ACTIONS(4042),
    [anon_sym_LBRACK] = ACTIONS(4044),
    [anon_sym_RBRACK] = ACTIONS(4044),
    [anon_sym___declspec] = ACTIONS(4042),
    [anon_sym_LBRACE] = ACTIONS(4044),
    [anon_sym_RBRACE] = ACTIONS(4044),
    [anon_sym_signed] = ACTIONS(4042),
    [anon_sym_unsigned] = ACTIONS(4042),
    [anon_sym_long] = ACTIONS(4042),
    [anon_sym_short] = ACTIONS(4042),
    [anon_sym_static] = ACTIONS(4042),
    [anon_sym_EQ] = ACTIONS(4042),
    [anon_sym_ATautoreleasepool] = ACTIONS(4044),
    [anon_sym_auto] = ACTIONS(4042),
    [anon_sym_register] = ACTIONS(4042),
    [anon_sym_inline] = ACTIONS(4042),
    [anon_sym___inline] = ACTIONS(4042),
    [anon_sym___inline__] = ACTIONS(4042),
    [anon_sym___forceinline] = ACTIONS(4042),
    [anon_sym_thread_local] = ACTIONS(4042),
    [anon_sym___thread] = ACTIONS(4042),
    [anon_sym_CG_EXTERN] = ACTIONS(4042),
    [anon_sym_CG_INLINE] = ACTIONS(4042),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4042),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4042),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4042),
    [anon_sym_IBOutlet] = ACTIONS(4042),
    [anon_sym_IBInspectable] = ACTIONS(4042),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4042),
    [anon_sym_NS_INLINE] = ACTIONS(4042),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4042),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4042),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4042),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4042),
    [anon_sym_const] = ACTIONS(4042),
    [anon_sym_constexpr] = ACTIONS(4042),
    [anon_sym_volatile] = ACTIONS(4042),
    [anon_sym_restrict] = ACTIONS(4042),
    [anon_sym___restrict__] = ACTIONS(4042),
    [anon_sym__Atomic] = ACTIONS(4042),
    [anon_sym__Noreturn] = ACTIONS(4042),
    [anon_sym__Nonnull] = ACTIONS(4042),
    [anon_sym_nullable] = ACTIONS(4042),
    [anon_sym__Complex] = ACTIONS(4042),
    [anon_sym__Nullable] = ACTIONS(4042),
    [anon_sym__Nullable_result] = ACTIONS(4042),
    [anon_sym__Null_unspecified] = ACTIONS(4042),
    [anon_sym___autoreleasing] = ACTIONS(4042),
    [anon_sym___block] = ACTIONS(4042),
    [anon_sym___bridge] = ACTIONS(4042),
    [anon_sym___bridge_retained] = ACTIONS(4042),
    [anon_sym___bridge_transfer] = ACTIONS(4042),
    [anon_sym___complex] = ACTIONS(4042),
    [anon_sym___const] = ACTIONS(4042),
    [anon_sym___imag] = ACTIONS(4042),
    [anon_sym___kindof] = ACTIONS(4042),
    [anon_sym___nonnull] = ACTIONS(4042),
    [anon_sym___nullable] = ACTIONS(4042),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4042),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4042),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4042),
    [anon_sym___real] = ACTIONS(4042),
    [anon_sym___strong] = ACTIONS(4042),
    [anon_sym___unsafe_unretained] = ACTIONS(4042),
    [anon_sym___unused] = ACTIONS(4042),
    [anon_sym___weak] = ACTIONS(4042),
    [anon_sym_alignas] = ACTIONS(4042),
    [anon_sym__Alignas] = ACTIONS(4042),
    [sym_primitive_type] = ACTIONS(4042),
    [anon_sym_enum] = ACTIONS(4042),
    [anon_sym_COLON] = ACTIONS(4044),
    [anon_sym_struct] = ACTIONS(4042),
    [anon_sym_union] = ACTIONS(4042),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_STAR_EQ] = ACTIONS(4044),
    [anon_sym_SLASH_EQ] = ACTIONS(4044),
    [anon_sym_PERCENT_EQ] = ACTIONS(4044),
    [anon_sym_PLUS_EQ] = ACTIONS(4044),
    [anon_sym_DASH_EQ] = ACTIONS(4044),
    [anon_sym_LT_LT_EQ] = ACTIONS(4044),
    [anon_sym_GT_GT_EQ] = ACTIONS(4044),
    [anon_sym_AMP_EQ] = ACTIONS(4044),
    [anon_sym_CARET_EQ] = ACTIONS(4044),
    [anon_sym_PIPE_EQ] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(4044),
    [anon_sym_PLUS_PLUS] = ACTIONS(4044),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4042),
    [anon_sym___typeof] = ACTIONS(4042),
    [anon_sym_typeof] = ACTIONS(4042),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4042),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4042),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4042),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4042),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4042),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4042),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4042),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4042),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4042),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4042),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4042),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4042),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4042),
    [anon_sym_API_AVAILABLE] = ACTIONS(4042),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4042),
    [anon_sym_API_DEPRECATED] = ACTIONS(4042),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4042),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4042),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4042),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4042),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4042),
    [anon_sym___deprecated_msg] = ACTIONS(4042),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4042),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4042),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4042),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4042),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4042),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4042),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4042),
    [anon_sym_BOOL] = ACTIONS(4042),
    [anon_sym_IMP] = ACTIONS(4042),
    [anon_sym_SEL] = ACTIONS(4042),
    [anon_sym_Class] = ACTIONS(4042),
    [anon_sym_id] = ACTIONS(4042),
  },
  [STATE(1450)] = {
    [sym_identifier] = ACTIONS(4046),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4048),
    [anon_sym_COMMA] = ACTIONS(4048),
    [anon_sym_RPAREN] = ACTIONS(4048),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4048),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4048),
    [anon_sym_LPAREN2] = ACTIONS(4048),
    [anon_sym_DASH] = ACTIONS(4046),
    [anon_sym_PLUS] = ACTIONS(4046),
    [anon_sym_STAR] = ACTIONS(4046),
    [anon_sym_SLASH] = ACTIONS(4046),
    [anon_sym_PERCENT] = ACTIONS(4046),
    [anon_sym_PIPE_PIPE] = ACTIONS(4048),
    [anon_sym_AMP_AMP] = ACTIONS(4048),
    [anon_sym_PIPE] = ACTIONS(4046),
    [anon_sym_CARET] = ACTIONS(4046),
    [anon_sym_AMP] = ACTIONS(4046),
    [anon_sym_EQ_EQ] = ACTIONS(4048),
    [anon_sym_BANG_EQ] = ACTIONS(4048),
    [anon_sym_GT] = ACTIONS(4046),
    [anon_sym_GT_EQ] = ACTIONS(4048),
    [anon_sym_LT_EQ] = ACTIONS(4048),
    [anon_sym_LT] = ACTIONS(4046),
    [anon_sym_LT_LT] = ACTIONS(4046),
    [anon_sym_GT_GT] = ACTIONS(4046),
    [anon_sym_SEMI] = ACTIONS(4048),
    [anon_sym___extension__] = ACTIONS(4046),
    [anon_sym_extern] = ACTIONS(4046),
    [anon_sym___attribute__] = ACTIONS(4046),
    [anon_sym___attribute] = ACTIONS(4046),
    [anon_sym_noreturn] = ACTIONS(4046),
    [anon_sym_LBRACK] = ACTIONS(4048),
    [anon_sym_RBRACK] = ACTIONS(4048),
    [anon_sym___declspec] = ACTIONS(4046),
    [anon_sym_LBRACE] = ACTIONS(4048),
    [anon_sym_RBRACE] = ACTIONS(4048),
    [anon_sym_signed] = ACTIONS(4046),
    [anon_sym_unsigned] = ACTIONS(4046),
    [anon_sym_long] = ACTIONS(4046),
    [anon_sym_short] = ACTIONS(4046),
    [anon_sym_static] = ACTIONS(4046),
    [anon_sym_EQ] = ACTIONS(4046),
    [anon_sym_ATautoreleasepool] = ACTIONS(4048),
    [anon_sym_auto] = ACTIONS(4046),
    [anon_sym_register] = ACTIONS(4046),
    [anon_sym_inline] = ACTIONS(4046),
    [anon_sym___inline] = ACTIONS(4046),
    [anon_sym___inline__] = ACTIONS(4046),
    [anon_sym___forceinline] = ACTIONS(4046),
    [anon_sym_thread_local] = ACTIONS(4046),
    [anon_sym___thread] = ACTIONS(4046),
    [anon_sym_CG_EXTERN] = ACTIONS(4046),
    [anon_sym_CG_INLINE] = ACTIONS(4046),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4046),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4046),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4046),
    [anon_sym_IBOutlet] = ACTIONS(4046),
    [anon_sym_IBInspectable] = ACTIONS(4046),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4046),
    [anon_sym_NS_INLINE] = ACTIONS(4046),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4046),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4046),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4046),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4046),
    [anon_sym_const] = ACTIONS(4046),
    [anon_sym_constexpr] = ACTIONS(4046),
    [anon_sym_volatile] = ACTIONS(4046),
    [anon_sym_restrict] = ACTIONS(4046),
    [anon_sym___restrict__] = ACTIONS(4046),
    [anon_sym__Atomic] = ACTIONS(4046),
    [anon_sym__Noreturn] = ACTIONS(4046),
    [anon_sym__Nonnull] = ACTIONS(4046),
    [anon_sym_nullable] = ACTIONS(4046),
    [anon_sym__Complex] = ACTIONS(4046),
    [anon_sym__Nullable] = ACTIONS(4046),
    [anon_sym__Nullable_result] = ACTIONS(4046),
    [anon_sym__Null_unspecified] = ACTIONS(4046),
    [anon_sym___autoreleasing] = ACTIONS(4046),
    [anon_sym___block] = ACTIONS(4046),
    [anon_sym___bridge] = ACTIONS(4046),
    [anon_sym___bridge_retained] = ACTIONS(4046),
    [anon_sym___bridge_transfer] = ACTIONS(4046),
    [anon_sym___complex] = ACTIONS(4046),
    [anon_sym___const] = ACTIONS(4046),
    [anon_sym___imag] = ACTIONS(4046),
    [anon_sym___kindof] = ACTIONS(4046),
    [anon_sym___nonnull] = ACTIONS(4046),
    [anon_sym___nullable] = ACTIONS(4046),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4046),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4046),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4046),
    [anon_sym___real] = ACTIONS(4046),
    [anon_sym___strong] = ACTIONS(4046),
    [anon_sym___unsafe_unretained] = ACTIONS(4046),
    [anon_sym___unused] = ACTIONS(4046),
    [anon_sym___weak] = ACTIONS(4046),
    [anon_sym_alignas] = ACTIONS(4046),
    [anon_sym__Alignas] = ACTIONS(4046),
    [sym_primitive_type] = ACTIONS(4046),
    [anon_sym_enum] = ACTIONS(4046),
    [anon_sym_COLON] = ACTIONS(4048),
    [anon_sym_struct] = ACTIONS(4046),
    [anon_sym_union] = ACTIONS(4046),
    [anon_sym_QMARK] = ACTIONS(4048),
    [anon_sym_STAR_EQ] = ACTIONS(4048),
    [anon_sym_SLASH_EQ] = ACTIONS(4048),
    [anon_sym_PERCENT_EQ] = ACTIONS(4048),
    [anon_sym_PLUS_EQ] = ACTIONS(4048),
    [anon_sym_DASH_EQ] = ACTIONS(4048),
    [anon_sym_LT_LT_EQ] = ACTIONS(4048),
    [anon_sym_GT_GT_EQ] = ACTIONS(4048),
    [anon_sym_AMP_EQ] = ACTIONS(4048),
    [anon_sym_CARET_EQ] = ACTIONS(4048),
    [anon_sym_PIPE_EQ] = ACTIONS(4048),
    [anon_sym_DASH_DASH] = ACTIONS(4048),
    [anon_sym_PLUS_PLUS] = ACTIONS(4048),
    [anon_sym_DOT] = ACTIONS(4046),
    [anon_sym_DASH_GT] = ACTIONS(4048),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4046),
    [anon_sym___typeof] = ACTIONS(4046),
    [anon_sym_typeof] = ACTIONS(4046),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4046),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4046),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4046),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4046),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4046),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4046),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4046),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4046),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4046),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4046),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4046),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4046),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4046),
    [anon_sym_API_AVAILABLE] = ACTIONS(4046),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4046),
    [anon_sym_API_DEPRECATED] = ACTIONS(4046),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4046),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4046),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4046),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4046),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4046),
    [anon_sym___deprecated_msg] = ACTIONS(4046),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4046),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4046),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4046),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4046),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4046),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4046),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4046),
    [anon_sym_BOOL] = ACTIONS(4046),
    [anon_sym_IMP] = ACTIONS(4046),
    [anon_sym_SEL] = ACTIONS(4046),
    [anon_sym_Class] = ACTIONS(4046),
    [anon_sym_id] = ACTIONS(4046),
  },
  [STATE(1451)] = {
    [sym_identifier] = ACTIONS(2346),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2348),
    [anon_sym_COMMA] = ACTIONS(2348),
    [anon_sym_RPAREN] = ACTIONS(2348),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2348),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2348),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2346),
    [anon_sym_SLASH] = ACTIONS(2346),
    [anon_sym_PERCENT] = ACTIONS(2346),
    [anon_sym_PIPE_PIPE] = ACTIONS(2348),
    [anon_sym_AMP_AMP] = ACTIONS(2348),
    [anon_sym_PIPE] = ACTIONS(2346),
    [anon_sym_CARET] = ACTIONS(2346),
    [anon_sym_AMP] = ACTIONS(2346),
    [anon_sym_EQ_EQ] = ACTIONS(2348),
    [anon_sym_BANG_EQ] = ACTIONS(2348),
    [anon_sym_GT] = ACTIONS(2346),
    [anon_sym_GT_EQ] = ACTIONS(2348),
    [anon_sym_LT_EQ] = ACTIONS(2348),
    [anon_sym_LT] = ACTIONS(2346),
    [anon_sym_LT_LT] = ACTIONS(2346),
    [anon_sym_GT_GT] = ACTIONS(2346),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym_RBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_RBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_EQ] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_COLON] = ACTIONS(2348),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_QMARK] = ACTIONS(2348),
    [anon_sym_STAR_EQ] = ACTIONS(2348),
    [anon_sym_SLASH_EQ] = ACTIONS(2348),
    [anon_sym_PERCENT_EQ] = ACTIONS(2348),
    [anon_sym_PLUS_EQ] = ACTIONS(2348),
    [anon_sym_DASH_EQ] = ACTIONS(2348),
    [anon_sym_LT_LT_EQ] = ACTIONS(2348),
    [anon_sym_GT_GT_EQ] = ACTIONS(2348),
    [anon_sym_AMP_EQ] = ACTIONS(2348),
    [anon_sym_CARET_EQ] = ACTIONS(2348),
    [anon_sym_PIPE_EQ] = ACTIONS(2348),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_DOT] = ACTIONS(2346),
    [anon_sym_DASH_GT] = ACTIONS(2348),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(1452)] = {
    [sym_identifier] = ACTIONS(4050),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4052),
    [anon_sym_COMMA] = ACTIONS(4052),
    [anon_sym_RPAREN] = ACTIONS(4052),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4052),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4052),
    [anon_sym_LPAREN2] = ACTIONS(4052),
    [anon_sym_DASH] = ACTIONS(4050),
    [anon_sym_PLUS] = ACTIONS(4050),
    [anon_sym_STAR] = ACTIONS(4050),
    [anon_sym_SLASH] = ACTIONS(4050),
    [anon_sym_PERCENT] = ACTIONS(4050),
    [anon_sym_PIPE_PIPE] = ACTIONS(4052),
    [anon_sym_AMP_AMP] = ACTIONS(4052),
    [anon_sym_PIPE] = ACTIONS(4050),
    [anon_sym_CARET] = ACTIONS(4050),
    [anon_sym_AMP] = ACTIONS(4050),
    [anon_sym_EQ_EQ] = ACTIONS(4052),
    [anon_sym_BANG_EQ] = ACTIONS(4052),
    [anon_sym_GT] = ACTIONS(4050),
    [anon_sym_GT_EQ] = ACTIONS(4052),
    [anon_sym_LT_EQ] = ACTIONS(4052),
    [anon_sym_LT] = ACTIONS(4050),
    [anon_sym_LT_LT] = ACTIONS(4050),
    [anon_sym_GT_GT] = ACTIONS(4050),
    [anon_sym_SEMI] = ACTIONS(4052),
    [anon_sym___extension__] = ACTIONS(4050),
    [anon_sym_extern] = ACTIONS(4050),
    [anon_sym___attribute__] = ACTIONS(4050),
    [anon_sym___attribute] = ACTIONS(4050),
    [anon_sym_noreturn] = ACTIONS(4050),
    [anon_sym_LBRACK] = ACTIONS(4052),
    [anon_sym_RBRACK] = ACTIONS(4052),
    [anon_sym___declspec] = ACTIONS(4050),
    [anon_sym_LBRACE] = ACTIONS(4052),
    [anon_sym_RBRACE] = ACTIONS(4052),
    [anon_sym_signed] = ACTIONS(4050),
    [anon_sym_unsigned] = ACTIONS(4050),
    [anon_sym_long] = ACTIONS(4050),
    [anon_sym_short] = ACTIONS(4050),
    [anon_sym_static] = ACTIONS(4050),
    [anon_sym_EQ] = ACTIONS(4050),
    [anon_sym_ATautoreleasepool] = ACTIONS(4052),
    [anon_sym_auto] = ACTIONS(4050),
    [anon_sym_register] = ACTIONS(4050),
    [anon_sym_inline] = ACTIONS(4050),
    [anon_sym___inline] = ACTIONS(4050),
    [anon_sym___inline__] = ACTIONS(4050),
    [anon_sym___forceinline] = ACTIONS(4050),
    [anon_sym_thread_local] = ACTIONS(4050),
    [anon_sym___thread] = ACTIONS(4050),
    [anon_sym_CG_EXTERN] = ACTIONS(4050),
    [anon_sym_CG_INLINE] = ACTIONS(4050),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4050),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4050),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4050),
    [anon_sym_IBOutlet] = ACTIONS(4050),
    [anon_sym_IBInspectable] = ACTIONS(4050),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4050),
    [anon_sym_NS_INLINE] = ACTIONS(4050),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4050),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4050),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4050),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4050),
    [anon_sym_const] = ACTIONS(4050),
    [anon_sym_constexpr] = ACTIONS(4050),
    [anon_sym_volatile] = ACTIONS(4050),
    [anon_sym_restrict] = ACTIONS(4050),
    [anon_sym___restrict__] = ACTIONS(4050),
    [anon_sym__Atomic] = ACTIONS(4050),
    [anon_sym__Noreturn] = ACTIONS(4050),
    [anon_sym__Nonnull] = ACTIONS(4050),
    [anon_sym_nullable] = ACTIONS(4050),
    [anon_sym__Complex] = ACTIONS(4050),
    [anon_sym__Nullable] = ACTIONS(4050),
    [anon_sym__Nullable_result] = ACTIONS(4050),
    [anon_sym__Null_unspecified] = ACTIONS(4050),
    [anon_sym___autoreleasing] = ACTIONS(4050),
    [anon_sym___block] = ACTIONS(4050),
    [anon_sym___bridge] = ACTIONS(4050),
    [anon_sym___bridge_retained] = ACTIONS(4050),
    [anon_sym___bridge_transfer] = ACTIONS(4050),
    [anon_sym___complex] = ACTIONS(4050),
    [anon_sym___const] = ACTIONS(4050),
    [anon_sym___imag] = ACTIONS(4050),
    [anon_sym___kindof] = ACTIONS(4050),
    [anon_sym___nonnull] = ACTIONS(4050),
    [anon_sym___nullable] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4050),
    [anon_sym___real] = ACTIONS(4050),
    [anon_sym___strong] = ACTIONS(4050),
    [anon_sym___unsafe_unretained] = ACTIONS(4050),
    [anon_sym___unused] = ACTIONS(4050),
    [anon_sym___weak] = ACTIONS(4050),
    [anon_sym_alignas] = ACTIONS(4050),
    [anon_sym__Alignas] = ACTIONS(4050),
    [sym_primitive_type] = ACTIONS(4050),
    [anon_sym_enum] = ACTIONS(4050),
    [anon_sym_COLON] = ACTIONS(4052),
    [anon_sym_struct] = ACTIONS(4050),
    [anon_sym_union] = ACTIONS(4050),
    [anon_sym_QMARK] = ACTIONS(4052),
    [anon_sym_STAR_EQ] = ACTIONS(4052),
    [anon_sym_SLASH_EQ] = ACTIONS(4052),
    [anon_sym_PERCENT_EQ] = ACTIONS(4052),
    [anon_sym_PLUS_EQ] = ACTIONS(4052),
    [anon_sym_DASH_EQ] = ACTIONS(4052),
    [anon_sym_LT_LT_EQ] = ACTIONS(4052),
    [anon_sym_GT_GT_EQ] = ACTIONS(4052),
    [anon_sym_AMP_EQ] = ACTIONS(4052),
    [anon_sym_CARET_EQ] = ACTIONS(4052),
    [anon_sym_PIPE_EQ] = ACTIONS(4052),
    [anon_sym_DASH_DASH] = ACTIONS(4052),
    [anon_sym_PLUS_PLUS] = ACTIONS(4052),
    [anon_sym_DOT] = ACTIONS(4050),
    [anon_sym_DASH_GT] = ACTIONS(4052),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4050),
    [anon_sym___typeof] = ACTIONS(4050),
    [anon_sym_typeof] = ACTIONS(4050),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4050),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4050),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4050),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4050),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4050),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4050),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4050),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4050),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4050),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_API_AVAILABLE] = ACTIONS(4050),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_API_DEPRECATED] = ACTIONS(4050),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4050),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4050),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4050),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4050),
    [anon_sym___deprecated_msg] = ACTIONS(4050),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4050),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4050),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4050),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4050),
    [anon_sym_BOOL] = ACTIONS(4050),
    [anon_sym_IMP] = ACTIONS(4050),
    [anon_sym_SEL] = ACTIONS(4050),
    [anon_sym_Class] = ACTIONS(4050),
    [anon_sym_id] = ACTIONS(4050),
  },
  [STATE(1453)] = {
    [sym_identifier] = ACTIONS(4054),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4056),
    [anon_sym_COMMA] = ACTIONS(4056),
    [anon_sym_RPAREN] = ACTIONS(4056),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4056),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4056),
    [anon_sym_LPAREN2] = ACTIONS(4056),
    [anon_sym_DASH] = ACTIONS(4054),
    [anon_sym_PLUS] = ACTIONS(4054),
    [anon_sym_STAR] = ACTIONS(4054),
    [anon_sym_SLASH] = ACTIONS(4054),
    [anon_sym_PERCENT] = ACTIONS(4054),
    [anon_sym_PIPE_PIPE] = ACTIONS(4056),
    [anon_sym_AMP_AMP] = ACTIONS(4056),
    [anon_sym_PIPE] = ACTIONS(4054),
    [anon_sym_CARET] = ACTIONS(4054),
    [anon_sym_AMP] = ACTIONS(4054),
    [anon_sym_EQ_EQ] = ACTIONS(4056),
    [anon_sym_BANG_EQ] = ACTIONS(4056),
    [anon_sym_GT] = ACTIONS(4054),
    [anon_sym_GT_EQ] = ACTIONS(4056),
    [anon_sym_LT_EQ] = ACTIONS(4056),
    [anon_sym_LT] = ACTIONS(4054),
    [anon_sym_LT_LT] = ACTIONS(4054),
    [anon_sym_GT_GT] = ACTIONS(4054),
    [anon_sym_SEMI] = ACTIONS(4056),
    [anon_sym___extension__] = ACTIONS(4054),
    [anon_sym_extern] = ACTIONS(4054),
    [anon_sym___attribute__] = ACTIONS(4054),
    [anon_sym___attribute] = ACTIONS(4054),
    [anon_sym_noreturn] = ACTIONS(4054),
    [anon_sym_LBRACK] = ACTIONS(4056),
    [anon_sym_RBRACK] = ACTIONS(4056),
    [anon_sym___declspec] = ACTIONS(4054),
    [anon_sym_LBRACE] = ACTIONS(4056),
    [anon_sym_RBRACE] = ACTIONS(4056),
    [anon_sym_signed] = ACTIONS(4054),
    [anon_sym_unsigned] = ACTIONS(4054),
    [anon_sym_long] = ACTIONS(4054),
    [anon_sym_short] = ACTIONS(4054),
    [anon_sym_static] = ACTIONS(4054),
    [anon_sym_EQ] = ACTIONS(4054),
    [anon_sym_ATautoreleasepool] = ACTIONS(4056),
    [anon_sym_auto] = ACTIONS(4054),
    [anon_sym_register] = ACTIONS(4054),
    [anon_sym_inline] = ACTIONS(4054),
    [anon_sym___inline] = ACTIONS(4054),
    [anon_sym___inline__] = ACTIONS(4054),
    [anon_sym___forceinline] = ACTIONS(4054),
    [anon_sym_thread_local] = ACTIONS(4054),
    [anon_sym___thread] = ACTIONS(4054),
    [anon_sym_CG_EXTERN] = ACTIONS(4054),
    [anon_sym_CG_INLINE] = ACTIONS(4054),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4054),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4054),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4054),
    [anon_sym_IBOutlet] = ACTIONS(4054),
    [anon_sym_IBInspectable] = ACTIONS(4054),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4054),
    [anon_sym_NS_INLINE] = ACTIONS(4054),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4054),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4054),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4054),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4054),
    [anon_sym_const] = ACTIONS(4054),
    [anon_sym_constexpr] = ACTIONS(4054),
    [anon_sym_volatile] = ACTIONS(4054),
    [anon_sym_restrict] = ACTIONS(4054),
    [anon_sym___restrict__] = ACTIONS(4054),
    [anon_sym__Atomic] = ACTIONS(4054),
    [anon_sym__Noreturn] = ACTIONS(4054),
    [anon_sym__Nonnull] = ACTIONS(4054),
    [anon_sym_nullable] = ACTIONS(4054),
    [anon_sym__Complex] = ACTIONS(4054),
    [anon_sym__Nullable] = ACTIONS(4054),
    [anon_sym__Nullable_result] = ACTIONS(4054),
    [anon_sym__Null_unspecified] = ACTIONS(4054),
    [anon_sym___autoreleasing] = ACTIONS(4054),
    [anon_sym___block] = ACTIONS(4054),
    [anon_sym___bridge] = ACTIONS(4054),
    [anon_sym___bridge_retained] = ACTIONS(4054),
    [anon_sym___bridge_transfer] = ACTIONS(4054),
    [anon_sym___complex] = ACTIONS(4054),
    [anon_sym___const] = ACTIONS(4054),
    [anon_sym___imag] = ACTIONS(4054),
    [anon_sym___kindof] = ACTIONS(4054),
    [anon_sym___nonnull] = ACTIONS(4054),
    [anon_sym___nullable] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4054),
    [anon_sym___real] = ACTIONS(4054),
    [anon_sym___strong] = ACTIONS(4054),
    [anon_sym___unsafe_unretained] = ACTIONS(4054),
    [anon_sym___unused] = ACTIONS(4054),
    [anon_sym___weak] = ACTIONS(4054),
    [anon_sym_alignas] = ACTIONS(4054),
    [anon_sym__Alignas] = ACTIONS(4054),
    [sym_primitive_type] = ACTIONS(4054),
    [anon_sym_enum] = ACTIONS(4054),
    [anon_sym_COLON] = ACTIONS(4056),
    [anon_sym_struct] = ACTIONS(4054),
    [anon_sym_union] = ACTIONS(4054),
    [anon_sym_QMARK] = ACTIONS(4056),
    [anon_sym_STAR_EQ] = ACTIONS(4056),
    [anon_sym_SLASH_EQ] = ACTIONS(4056),
    [anon_sym_PERCENT_EQ] = ACTIONS(4056),
    [anon_sym_PLUS_EQ] = ACTIONS(4056),
    [anon_sym_DASH_EQ] = ACTIONS(4056),
    [anon_sym_LT_LT_EQ] = ACTIONS(4056),
    [anon_sym_GT_GT_EQ] = ACTIONS(4056),
    [anon_sym_AMP_EQ] = ACTIONS(4056),
    [anon_sym_CARET_EQ] = ACTIONS(4056),
    [anon_sym_PIPE_EQ] = ACTIONS(4056),
    [anon_sym_DASH_DASH] = ACTIONS(4056),
    [anon_sym_PLUS_PLUS] = ACTIONS(4056),
    [anon_sym_DOT] = ACTIONS(4054),
    [anon_sym_DASH_GT] = ACTIONS(4056),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4054),
    [anon_sym___typeof] = ACTIONS(4054),
    [anon_sym_typeof] = ACTIONS(4054),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4054),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4054),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4054),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4054),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4054),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4054),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4054),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4054),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4054),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_API_AVAILABLE] = ACTIONS(4054),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_API_DEPRECATED] = ACTIONS(4054),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4054),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4054),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4054),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4054),
    [anon_sym___deprecated_msg] = ACTIONS(4054),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4054),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4054),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4054),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4054),
    [anon_sym_BOOL] = ACTIONS(4054),
    [anon_sym_IMP] = ACTIONS(4054),
    [anon_sym_SEL] = ACTIONS(4054),
    [anon_sym_Class] = ACTIONS(4054),
    [anon_sym_id] = ACTIONS(4054),
  },
  [STATE(1454)] = {
    [sym_identifier] = ACTIONS(4024),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_RPAREN] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4024),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4024),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4024),
    [anon_sym_extern] = ACTIONS(4024),
    [anon_sym___attribute__] = ACTIONS(4024),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4024),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym_RBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4024),
    [anon_sym_LBRACE] = ACTIONS(4026),
    [anon_sym_RBRACE] = ACTIONS(4026),
    [anon_sym_signed] = ACTIONS(4024),
    [anon_sym_unsigned] = ACTIONS(4024),
    [anon_sym_long] = ACTIONS(4024),
    [anon_sym_short] = ACTIONS(4024),
    [anon_sym_static] = ACTIONS(4024),
    [anon_sym_EQ] = ACTIONS(4024),
    [anon_sym_ATautoreleasepool] = ACTIONS(4026),
    [anon_sym_auto] = ACTIONS(4024),
    [anon_sym_register] = ACTIONS(4024),
    [anon_sym_inline] = ACTIONS(4024),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4024),
    [anon_sym___forceinline] = ACTIONS(4024),
    [anon_sym_thread_local] = ACTIONS(4024),
    [anon_sym___thread] = ACTIONS(4024),
    [anon_sym_CG_EXTERN] = ACTIONS(4024),
    [anon_sym_CG_INLINE] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4024),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4024),
    [anon_sym_IBOutlet] = ACTIONS(4024),
    [anon_sym_IBInspectable] = ACTIONS(4024),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4024),
    [anon_sym_NS_INLINE] = ACTIONS(4024),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4024),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4024),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4024),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4024),
    [anon_sym_volatile] = ACTIONS(4024),
    [anon_sym_restrict] = ACTIONS(4024),
    [anon_sym___restrict__] = ACTIONS(4024),
    [anon_sym__Atomic] = ACTIONS(4024),
    [anon_sym__Noreturn] = ACTIONS(4024),
    [anon_sym__Nonnull] = ACTIONS(4024),
    [anon_sym_nullable] = ACTIONS(4024),
    [anon_sym__Complex] = ACTIONS(4024),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4024),
    [anon_sym__Null_unspecified] = ACTIONS(4024),
    [anon_sym___autoreleasing] = ACTIONS(4024),
    [anon_sym___block] = ACTIONS(4024),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4024),
    [anon_sym___bridge_transfer] = ACTIONS(4024),
    [anon_sym___complex] = ACTIONS(4024),
    [anon_sym___const] = ACTIONS(4024),
    [anon_sym___imag] = ACTIONS(4024),
    [anon_sym___kindof] = ACTIONS(4024),
    [anon_sym___nonnull] = ACTIONS(4024),
    [anon_sym___nullable] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4024),
    [anon_sym___real] = ACTIONS(4024),
    [anon_sym___strong] = ACTIONS(4024),
    [anon_sym___unsafe_unretained] = ACTIONS(4024),
    [anon_sym___unused] = ACTIONS(4024),
    [anon_sym___weak] = ACTIONS(4024),
    [anon_sym_alignas] = ACTIONS(4024),
    [anon_sym__Alignas] = ACTIONS(4024),
    [sym_primitive_type] = ACTIONS(4024),
    [anon_sym_enum] = ACTIONS(4024),
    [anon_sym_COLON] = ACTIONS(4026),
    [anon_sym_struct] = ACTIONS(4024),
    [anon_sym_union] = ACTIONS(4024),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4026),
    [anon_sym_SLASH_EQ] = ACTIONS(4026),
    [anon_sym_PERCENT_EQ] = ACTIONS(4026),
    [anon_sym_PLUS_EQ] = ACTIONS(4026),
    [anon_sym_DASH_EQ] = ACTIONS(4026),
    [anon_sym_LT_LT_EQ] = ACTIONS(4026),
    [anon_sym_GT_GT_EQ] = ACTIONS(4026),
    [anon_sym_AMP_EQ] = ACTIONS(4026),
    [anon_sym_CARET_EQ] = ACTIONS(4026),
    [anon_sym_PIPE_EQ] = ACTIONS(4026),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4024),
    [anon_sym___typeof] = ACTIONS(4024),
    [anon_sym_typeof] = ACTIONS(4024),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4024),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4024),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_API_AVAILABLE] = ACTIONS(4024),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_API_DEPRECATED] = ACTIONS(4024),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym___deprecated_msg] = ACTIONS(4024),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4024),
    [anon_sym_BOOL] = ACTIONS(4024),
    [anon_sym_IMP] = ACTIONS(4024),
    [anon_sym_SEL] = ACTIONS(4024),
    [anon_sym_Class] = ACTIONS(4024),
    [anon_sym_id] = ACTIONS(4024),
  },
  [STATE(1455)] = {
    [sym_identifier] = ACTIONS(4058),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4060),
    [anon_sym_COMMA] = ACTIONS(4060),
    [anon_sym_RPAREN] = ACTIONS(4060),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4060),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4060),
    [anon_sym_LPAREN2] = ACTIONS(4060),
    [anon_sym_DASH] = ACTIONS(4058),
    [anon_sym_PLUS] = ACTIONS(4058),
    [anon_sym_STAR] = ACTIONS(4058),
    [anon_sym_SLASH] = ACTIONS(4058),
    [anon_sym_PERCENT] = ACTIONS(4058),
    [anon_sym_PIPE_PIPE] = ACTIONS(4060),
    [anon_sym_AMP_AMP] = ACTIONS(4060),
    [anon_sym_PIPE] = ACTIONS(4058),
    [anon_sym_CARET] = ACTIONS(4058),
    [anon_sym_AMP] = ACTIONS(4058),
    [anon_sym_EQ_EQ] = ACTIONS(4060),
    [anon_sym_BANG_EQ] = ACTIONS(4060),
    [anon_sym_GT] = ACTIONS(4058),
    [anon_sym_GT_EQ] = ACTIONS(4060),
    [anon_sym_LT_EQ] = ACTIONS(4060),
    [anon_sym_LT] = ACTIONS(4058),
    [anon_sym_LT_LT] = ACTIONS(4058),
    [anon_sym_GT_GT] = ACTIONS(4058),
    [anon_sym_SEMI] = ACTIONS(4060),
    [anon_sym___extension__] = ACTIONS(4058),
    [anon_sym_extern] = ACTIONS(4058),
    [anon_sym___attribute__] = ACTIONS(4058),
    [anon_sym___attribute] = ACTIONS(4058),
    [anon_sym_noreturn] = ACTIONS(4058),
    [anon_sym_LBRACK] = ACTIONS(4060),
    [anon_sym_RBRACK] = ACTIONS(4060),
    [anon_sym___declspec] = ACTIONS(4058),
    [anon_sym_LBRACE] = ACTIONS(4060),
    [anon_sym_RBRACE] = ACTIONS(4060),
    [anon_sym_signed] = ACTIONS(4058),
    [anon_sym_unsigned] = ACTIONS(4058),
    [anon_sym_long] = ACTIONS(4058),
    [anon_sym_short] = ACTIONS(4058),
    [anon_sym_static] = ACTIONS(4058),
    [anon_sym_EQ] = ACTIONS(4058),
    [anon_sym_ATautoreleasepool] = ACTIONS(4060),
    [anon_sym_auto] = ACTIONS(4058),
    [anon_sym_register] = ACTIONS(4058),
    [anon_sym_inline] = ACTIONS(4058),
    [anon_sym___inline] = ACTIONS(4058),
    [anon_sym___inline__] = ACTIONS(4058),
    [anon_sym___forceinline] = ACTIONS(4058),
    [anon_sym_thread_local] = ACTIONS(4058),
    [anon_sym___thread] = ACTIONS(4058),
    [anon_sym_CG_EXTERN] = ACTIONS(4058),
    [anon_sym_CG_INLINE] = ACTIONS(4058),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4058),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4058),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4058),
    [anon_sym_IBOutlet] = ACTIONS(4058),
    [anon_sym_IBInspectable] = ACTIONS(4058),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4058),
    [anon_sym_NS_INLINE] = ACTIONS(4058),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4058),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4058),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4058),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4058),
    [anon_sym_const] = ACTIONS(4058),
    [anon_sym_constexpr] = ACTIONS(4058),
    [anon_sym_volatile] = ACTIONS(4058),
    [anon_sym_restrict] = ACTIONS(4058),
    [anon_sym___restrict__] = ACTIONS(4058),
    [anon_sym__Atomic] = ACTIONS(4058),
    [anon_sym__Noreturn] = ACTIONS(4058),
    [anon_sym__Nonnull] = ACTIONS(4058),
    [anon_sym_nullable] = ACTIONS(4058),
    [anon_sym__Complex] = ACTIONS(4058),
    [anon_sym__Nullable] = ACTIONS(4058),
    [anon_sym__Nullable_result] = ACTIONS(4058),
    [anon_sym__Null_unspecified] = ACTIONS(4058),
    [anon_sym___autoreleasing] = ACTIONS(4058),
    [anon_sym___block] = ACTIONS(4058),
    [anon_sym___bridge] = ACTIONS(4058),
    [anon_sym___bridge_retained] = ACTIONS(4058),
    [anon_sym___bridge_transfer] = ACTIONS(4058),
    [anon_sym___complex] = ACTIONS(4058),
    [anon_sym___const] = ACTIONS(4058),
    [anon_sym___imag] = ACTIONS(4058),
    [anon_sym___kindof] = ACTIONS(4058),
    [anon_sym___nonnull] = ACTIONS(4058),
    [anon_sym___nullable] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4058),
    [anon_sym___real] = ACTIONS(4058),
    [anon_sym___strong] = ACTIONS(4058),
    [anon_sym___unsafe_unretained] = ACTIONS(4058),
    [anon_sym___unused] = ACTIONS(4058),
    [anon_sym___weak] = ACTIONS(4058),
    [anon_sym_alignas] = ACTIONS(4058),
    [anon_sym__Alignas] = ACTIONS(4058),
    [sym_primitive_type] = ACTIONS(4058),
    [anon_sym_enum] = ACTIONS(4058),
    [anon_sym_COLON] = ACTIONS(4060),
    [anon_sym_struct] = ACTIONS(4058),
    [anon_sym_union] = ACTIONS(4058),
    [anon_sym_QMARK] = ACTIONS(4060),
    [anon_sym_STAR_EQ] = ACTIONS(4060),
    [anon_sym_SLASH_EQ] = ACTIONS(4060),
    [anon_sym_PERCENT_EQ] = ACTIONS(4060),
    [anon_sym_PLUS_EQ] = ACTIONS(4060),
    [anon_sym_DASH_EQ] = ACTIONS(4060),
    [anon_sym_LT_LT_EQ] = ACTIONS(4060),
    [anon_sym_GT_GT_EQ] = ACTIONS(4060),
    [anon_sym_AMP_EQ] = ACTIONS(4060),
    [anon_sym_CARET_EQ] = ACTIONS(4060),
    [anon_sym_PIPE_EQ] = ACTIONS(4060),
    [anon_sym_DASH_DASH] = ACTIONS(4060),
    [anon_sym_PLUS_PLUS] = ACTIONS(4060),
    [anon_sym_DOT] = ACTIONS(4058),
    [anon_sym_DASH_GT] = ACTIONS(4060),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4058),
    [anon_sym___typeof] = ACTIONS(4058),
    [anon_sym_typeof] = ACTIONS(4058),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4058),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4058),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4058),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4058),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4058),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4058),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4058),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4058),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4058),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_API_AVAILABLE] = ACTIONS(4058),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_API_DEPRECATED] = ACTIONS(4058),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4058),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4058),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4058),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4058),
    [anon_sym___deprecated_msg] = ACTIONS(4058),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4058),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4058),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4058),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4058),
    [anon_sym_BOOL] = ACTIONS(4058),
    [anon_sym_IMP] = ACTIONS(4058),
    [anon_sym_SEL] = ACTIONS(4058),
    [anon_sym_Class] = ACTIONS(4058),
    [anon_sym_id] = ACTIONS(4058),
  },
  [STATE(1456)] = {
    [sym_identifier] = ACTIONS(4062),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4064),
    [anon_sym_COMMA] = ACTIONS(4064),
    [anon_sym_RPAREN] = ACTIONS(4064),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4064),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4064),
    [anon_sym_LPAREN2] = ACTIONS(4064),
    [anon_sym_DASH] = ACTIONS(4062),
    [anon_sym_PLUS] = ACTIONS(4062),
    [anon_sym_STAR] = ACTIONS(4062),
    [anon_sym_SLASH] = ACTIONS(4062),
    [anon_sym_PERCENT] = ACTIONS(4062),
    [anon_sym_PIPE_PIPE] = ACTIONS(4064),
    [anon_sym_AMP_AMP] = ACTIONS(4064),
    [anon_sym_PIPE] = ACTIONS(4062),
    [anon_sym_CARET] = ACTIONS(4062),
    [anon_sym_AMP] = ACTIONS(4062),
    [anon_sym_EQ_EQ] = ACTIONS(4064),
    [anon_sym_BANG_EQ] = ACTIONS(4064),
    [anon_sym_GT] = ACTIONS(4062),
    [anon_sym_GT_EQ] = ACTIONS(4064),
    [anon_sym_LT_EQ] = ACTIONS(4064),
    [anon_sym_LT] = ACTIONS(4062),
    [anon_sym_LT_LT] = ACTIONS(4062),
    [anon_sym_GT_GT] = ACTIONS(4062),
    [anon_sym_SEMI] = ACTIONS(4064),
    [anon_sym___extension__] = ACTIONS(4062),
    [anon_sym_extern] = ACTIONS(4062),
    [anon_sym___attribute__] = ACTIONS(4062),
    [anon_sym___attribute] = ACTIONS(4062),
    [anon_sym_noreturn] = ACTIONS(4062),
    [anon_sym_LBRACK] = ACTIONS(4064),
    [anon_sym_RBRACK] = ACTIONS(4064),
    [anon_sym___declspec] = ACTIONS(4062),
    [anon_sym_LBRACE] = ACTIONS(4064),
    [anon_sym_RBRACE] = ACTIONS(4064),
    [anon_sym_signed] = ACTIONS(4062),
    [anon_sym_unsigned] = ACTIONS(4062),
    [anon_sym_long] = ACTIONS(4062),
    [anon_sym_short] = ACTIONS(4062),
    [anon_sym_static] = ACTIONS(4062),
    [anon_sym_EQ] = ACTIONS(4062),
    [anon_sym_ATautoreleasepool] = ACTIONS(4064),
    [anon_sym_auto] = ACTIONS(4062),
    [anon_sym_register] = ACTIONS(4062),
    [anon_sym_inline] = ACTIONS(4062),
    [anon_sym___inline] = ACTIONS(4062),
    [anon_sym___inline__] = ACTIONS(4062),
    [anon_sym___forceinline] = ACTIONS(4062),
    [anon_sym_thread_local] = ACTIONS(4062),
    [anon_sym___thread] = ACTIONS(4062),
    [anon_sym_CG_EXTERN] = ACTIONS(4062),
    [anon_sym_CG_INLINE] = ACTIONS(4062),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4062),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4062),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4062),
    [anon_sym_IBOutlet] = ACTIONS(4062),
    [anon_sym_IBInspectable] = ACTIONS(4062),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4062),
    [anon_sym_NS_INLINE] = ACTIONS(4062),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4062),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4062),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4062),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4062),
    [anon_sym_const] = ACTIONS(4062),
    [anon_sym_constexpr] = ACTIONS(4062),
    [anon_sym_volatile] = ACTIONS(4062),
    [anon_sym_restrict] = ACTIONS(4062),
    [anon_sym___restrict__] = ACTIONS(4062),
    [anon_sym__Atomic] = ACTIONS(4062),
    [anon_sym__Noreturn] = ACTIONS(4062),
    [anon_sym__Nonnull] = ACTIONS(4062),
    [anon_sym_nullable] = ACTIONS(4062),
    [anon_sym__Complex] = ACTIONS(4062),
    [anon_sym__Nullable] = ACTIONS(4062),
    [anon_sym__Nullable_result] = ACTIONS(4062),
    [anon_sym__Null_unspecified] = ACTIONS(4062),
    [anon_sym___autoreleasing] = ACTIONS(4062),
    [anon_sym___block] = ACTIONS(4062),
    [anon_sym___bridge] = ACTIONS(4062),
    [anon_sym___bridge_retained] = ACTIONS(4062),
    [anon_sym___bridge_transfer] = ACTIONS(4062),
    [anon_sym___complex] = ACTIONS(4062),
    [anon_sym___const] = ACTIONS(4062),
    [anon_sym___imag] = ACTIONS(4062),
    [anon_sym___kindof] = ACTIONS(4062),
    [anon_sym___nonnull] = ACTIONS(4062),
    [anon_sym___nullable] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4062),
    [anon_sym___real] = ACTIONS(4062),
    [anon_sym___strong] = ACTIONS(4062),
    [anon_sym___unsafe_unretained] = ACTIONS(4062),
    [anon_sym___unused] = ACTIONS(4062),
    [anon_sym___weak] = ACTIONS(4062),
    [anon_sym_alignas] = ACTIONS(4062),
    [anon_sym__Alignas] = ACTIONS(4062),
    [sym_primitive_type] = ACTIONS(4062),
    [anon_sym_enum] = ACTIONS(4062),
    [anon_sym_COLON] = ACTIONS(4064),
    [anon_sym_struct] = ACTIONS(4062),
    [anon_sym_union] = ACTIONS(4062),
    [anon_sym_QMARK] = ACTIONS(4064),
    [anon_sym_STAR_EQ] = ACTIONS(4064),
    [anon_sym_SLASH_EQ] = ACTIONS(4064),
    [anon_sym_PERCENT_EQ] = ACTIONS(4064),
    [anon_sym_PLUS_EQ] = ACTIONS(4064),
    [anon_sym_DASH_EQ] = ACTIONS(4064),
    [anon_sym_LT_LT_EQ] = ACTIONS(4064),
    [anon_sym_GT_GT_EQ] = ACTIONS(4064),
    [anon_sym_AMP_EQ] = ACTIONS(4064),
    [anon_sym_CARET_EQ] = ACTIONS(4064),
    [anon_sym_PIPE_EQ] = ACTIONS(4064),
    [anon_sym_DASH_DASH] = ACTIONS(4064),
    [anon_sym_PLUS_PLUS] = ACTIONS(4064),
    [anon_sym_DOT] = ACTIONS(4062),
    [anon_sym_DASH_GT] = ACTIONS(4064),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4062),
    [anon_sym___typeof] = ACTIONS(4062),
    [anon_sym_typeof] = ACTIONS(4062),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4062),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4062),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4062),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4062),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4062),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4062),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4062),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4062),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4062),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_API_AVAILABLE] = ACTIONS(4062),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_API_DEPRECATED] = ACTIONS(4062),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4062),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4062),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4062),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4062),
    [anon_sym___deprecated_msg] = ACTIONS(4062),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4062),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4062),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4062),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4062),
    [anon_sym_BOOL] = ACTIONS(4062),
    [anon_sym_IMP] = ACTIONS(4062),
    [anon_sym_SEL] = ACTIONS(4062),
    [anon_sym_Class] = ACTIONS(4062),
    [anon_sym_id] = ACTIONS(4062),
  },
  [STATE(1457)] = {
    [sym_identifier] = ACTIONS(4066),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4068),
    [anon_sym_COMMA] = ACTIONS(4068),
    [anon_sym_RPAREN] = ACTIONS(4068),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4068),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4068),
    [anon_sym_LPAREN2] = ACTIONS(4068),
    [anon_sym_DASH] = ACTIONS(4066),
    [anon_sym_PLUS] = ACTIONS(4066),
    [anon_sym_STAR] = ACTIONS(4066),
    [anon_sym_SLASH] = ACTIONS(4066),
    [anon_sym_PERCENT] = ACTIONS(4066),
    [anon_sym_PIPE_PIPE] = ACTIONS(4068),
    [anon_sym_AMP_AMP] = ACTIONS(4068),
    [anon_sym_PIPE] = ACTIONS(4066),
    [anon_sym_CARET] = ACTIONS(4066),
    [anon_sym_AMP] = ACTIONS(4066),
    [anon_sym_EQ_EQ] = ACTIONS(4068),
    [anon_sym_BANG_EQ] = ACTIONS(4068),
    [anon_sym_GT] = ACTIONS(4066),
    [anon_sym_GT_EQ] = ACTIONS(4068),
    [anon_sym_LT_EQ] = ACTIONS(4068),
    [anon_sym_LT] = ACTIONS(4066),
    [anon_sym_LT_LT] = ACTIONS(4066),
    [anon_sym_GT_GT] = ACTIONS(4066),
    [anon_sym_SEMI] = ACTIONS(4068),
    [anon_sym___extension__] = ACTIONS(4066),
    [anon_sym_extern] = ACTIONS(4066),
    [anon_sym___attribute__] = ACTIONS(4066),
    [anon_sym___attribute] = ACTIONS(4066),
    [anon_sym_noreturn] = ACTIONS(4066),
    [anon_sym_LBRACK] = ACTIONS(4068),
    [anon_sym_RBRACK] = ACTIONS(4068),
    [anon_sym___declspec] = ACTIONS(4066),
    [anon_sym_LBRACE] = ACTIONS(4068),
    [anon_sym_RBRACE] = ACTIONS(4068),
    [anon_sym_signed] = ACTIONS(4066),
    [anon_sym_unsigned] = ACTIONS(4066),
    [anon_sym_long] = ACTIONS(4066),
    [anon_sym_short] = ACTIONS(4066),
    [anon_sym_static] = ACTIONS(4066),
    [anon_sym_EQ] = ACTIONS(4066),
    [anon_sym_ATautoreleasepool] = ACTIONS(4068),
    [anon_sym_auto] = ACTIONS(4066),
    [anon_sym_register] = ACTIONS(4066),
    [anon_sym_inline] = ACTIONS(4066),
    [anon_sym___inline] = ACTIONS(4066),
    [anon_sym___inline__] = ACTIONS(4066),
    [anon_sym___forceinline] = ACTIONS(4066),
    [anon_sym_thread_local] = ACTIONS(4066),
    [anon_sym___thread] = ACTIONS(4066),
    [anon_sym_CG_EXTERN] = ACTIONS(4066),
    [anon_sym_CG_INLINE] = ACTIONS(4066),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4066),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4066),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4066),
    [anon_sym_IBOutlet] = ACTIONS(4066),
    [anon_sym_IBInspectable] = ACTIONS(4066),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4066),
    [anon_sym_NS_INLINE] = ACTIONS(4066),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4066),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4066),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4066),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4066),
    [anon_sym_const] = ACTIONS(4066),
    [anon_sym_constexpr] = ACTIONS(4066),
    [anon_sym_volatile] = ACTIONS(4066),
    [anon_sym_restrict] = ACTIONS(4066),
    [anon_sym___restrict__] = ACTIONS(4066),
    [anon_sym__Atomic] = ACTIONS(4066),
    [anon_sym__Noreturn] = ACTIONS(4066),
    [anon_sym__Nonnull] = ACTIONS(4066),
    [anon_sym_nullable] = ACTIONS(4066),
    [anon_sym__Complex] = ACTIONS(4066),
    [anon_sym__Nullable] = ACTIONS(4066),
    [anon_sym__Nullable_result] = ACTIONS(4066),
    [anon_sym__Null_unspecified] = ACTIONS(4066),
    [anon_sym___autoreleasing] = ACTIONS(4066),
    [anon_sym___block] = ACTIONS(4066),
    [anon_sym___bridge] = ACTIONS(4066),
    [anon_sym___bridge_retained] = ACTIONS(4066),
    [anon_sym___bridge_transfer] = ACTIONS(4066),
    [anon_sym___complex] = ACTIONS(4066),
    [anon_sym___const] = ACTIONS(4066),
    [anon_sym___imag] = ACTIONS(4066),
    [anon_sym___kindof] = ACTIONS(4066),
    [anon_sym___nonnull] = ACTIONS(4066),
    [anon_sym___nullable] = ACTIONS(4066),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4066),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4066),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4066),
    [anon_sym___real] = ACTIONS(4066),
    [anon_sym___strong] = ACTIONS(4066),
    [anon_sym___unsafe_unretained] = ACTIONS(4066),
    [anon_sym___unused] = ACTIONS(4066),
    [anon_sym___weak] = ACTIONS(4066),
    [anon_sym_alignas] = ACTIONS(4066),
    [anon_sym__Alignas] = ACTIONS(4066),
    [sym_primitive_type] = ACTIONS(4066),
    [anon_sym_enum] = ACTIONS(4066),
    [anon_sym_COLON] = ACTIONS(4068),
    [anon_sym_struct] = ACTIONS(4066),
    [anon_sym_union] = ACTIONS(4066),
    [anon_sym_QMARK] = ACTIONS(4068),
    [anon_sym_STAR_EQ] = ACTIONS(4068),
    [anon_sym_SLASH_EQ] = ACTIONS(4068),
    [anon_sym_PERCENT_EQ] = ACTIONS(4068),
    [anon_sym_PLUS_EQ] = ACTIONS(4068),
    [anon_sym_DASH_EQ] = ACTIONS(4068),
    [anon_sym_LT_LT_EQ] = ACTIONS(4068),
    [anon_sym_GT_GT_EQ] = ACTIONS(4068),
    [anon_sym_AMP_EQ] = ACTIONS(4068),
    [anon_sym_CARET_EQ] = ACTIONS(4068),
    [anon_sym_PIPE_EQ] = ACTIONS(4068),
    [anon_sym_DASH_DASH] = ACTIONS(4068),
    [anon_sym_PLUS_PLUS] = ACTIONS(4068),
    [anon_sym_DOT] = ACTIONS(4066),
    [anon_sym_DASH_GT] = ACTIONS(4068),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4066),
    [anon_sym___typeof] = ACTIONS(4066),
    [anon_sym_typeof] = ACTIONS(4066),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4066),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4066),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4066),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4066),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4066),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4066),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4066),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4066),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4066),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4066),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4066),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4066),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4066),
    [anon_sym_API_AVAILABLE] = ACTIONS(4066),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4066),
    [anon_sym_API_DEPRECATED] = ACTIONS(4066),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4066),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4066),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4066),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4066),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4066),
    [anon_sym___deprecated_msg] = ACTIONS(4066),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4066),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4066),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4066),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4066),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4066),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4066),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4066),
    [anon_sym_BOOL] = ACTIONS(4066),
    [anon_sym_IMP] = ACTIONS(4066),
    [anon_sym_SEL] = ACTIONS(4066),
    [anon_sym_Class] = ACTIONS(4066),
    [anon_sym_id] = ACTIONS(4066),
  },
  [STATE(1458)] = {
    [sym_identifier] = ACTIONS(4070),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4072),
    [anon_sym_COMMA] = ACTIONS(4072),
    [anon_sym_RPAREN] = ACTIONS(4072),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4072),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4072),
    [anon_sym_LPAREN2] = ACTIONS(4072),
    [anon_sym_DASH] = ACTIONS(4070),
    [anon_sym_PLUS] = ACTIONS(4070),
    [anon_sym_STAR] = ACTIONS(4070),
    [anon_sym_SLASH] = ACTIONS(4070),
    [anon_sym_PERCENT] = ACTIONS(4070),
    [anon_sym_PIPE_PIPE] = ACTIONS(4072),
    [anon_sym_AMP_AMP] = ACTIONS(4072),
    [anon_sym_PIPE] = ACTIONS(4070),
    [anon_sym_CARET] = ACTIONS(4070),
    [anon_sym_AMP] = ACTIONS(4070),
    [anon_sym_EQ_EQ] = ACTIONS(4072),
    [anon_sym_BANG_EQ] = ACTIONS(4072),
    [anon_sym_GT] = ACTIONS(4070),
    [anon_sym_GT_EQ] = ACTIONS(4072),
    [anon_sym_LT_EQ] = ACTIONS(4072),
    [anon_sym_LT] = ACTIONS(4070),
    [anon_sym_LT_LT] = ACTIONS(4070),
    [anon_sym_GT_GT] = ACTIONS(4070),
    [anon_sym_SEMI] = ACTIONS(4072),
    [anon_sym___extension__] = ACTIONS(4070),
    [anon_sym_extern] = ACTIONS(4070),
    [anon_sym___attribute__] = ACTIONS(4070),
    [anon_sym___attribute] = ACTIONS(4070),
    [anon_sym_noreturn] = ACTIONS(4070),
    [anon_sym_LBRACK] = ACTIONS(4072),
    [anon_sym_RBRACK] = ACTIONS(4072),
    [anon_sym___declspec] = ACTIONS(4070),
    [anon_sym_LBRACE] = ACTIONS(4072),
    [anon_sym_RBRACE] = ACTIONS(4072),
    [anon_sym_signed] = ACTIONS(4070),
    [anon_sym_unsigned] = ACTIONS(4070),
    [anon_sym_long] = ACTIONS(4070),
    [anon_sym_short] = ACTIONS(4070),
    [anon_sym_static] = ACTIONS(4070),
    [anon_sym_EQ] = ACTIONS(4070),
    [anon_sym_ATautoreleasepool] = ACTIONS(4072),
    [anon_sym_auto] = ACTIONS(4070),
    [anon_sym_register] = ACTIONS(4070),
    [anon_sym_inline] = ACTIONS(4070),
    [anon_sym___inline] = ACTIONS(4070),
    [anon_sym___inline__] = ACTIONS(4070),
    [anon_sym___forceinline] = ACTIONS(4070),
    [anon_sym_thread_local] = ACTIONS(4070),
    [anon_sym___thread] = ACTIONS(4070),
    [anon_sym_CG_EXTERN] = ACTIONS(4070),
    [anon_sym_CG_INLINE] = ACTIONS(4070),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4070),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4070),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4070),
    [anon_sym_IBOutlet] = ACTIONS(4070),
    [anon_sym_IBInspectable] = ACTIONS(4070),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4070),
    [anon_sym_NS_INLINE] = ACTIONS(4070),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4070),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4070),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4070),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4070),
    [anon_sym_const] = ACTIONS(4070),
    [anon_sym_constexpr] = ACTIONS(4070),
    [anon_sym_volatile] = ACTIONS(4070),
    [anon_sym_restrict] = ACTIONS(4070),
    [anon_sym___restrict__] = ACTIONS(4070),
    [anon_sym__Atomic] = ACTIONS(4070),
    [anon_sym__Noreturn] = ACTIONS(4070),
    [anon_sym__Nonnull] = ACTIONS(4070),
    [anon_sym_nullable] = ACTIONS(4070),
    [anon_sym__Complex] = ACTIONS(4070),
    [anon_sym__Nullable] = ACTIONS(4070),
    [anon_sym__Nullable_result] = ACTIONS(4070),
    [anon_sym__Null_unspecified] = ACTIONS(4070),
    [anon_sym___autoreleasing] = ACTIONS(4070),
    [anon_sym___block] = ACTIONS(4070),
    [anon_sym___bridge] = ACTIONS(4070),
    [anon_sym___bridge_retained] = ACTIONS(4070),
    [anon_sym___bridge_transfer] = ACTIONS(4070),
    [anon_sym___complex] = ACTIONS(4070),
    [anon_sym___const] = ACTIONS(4070),
    [anon_sym___imag] = ACTIONS(4070),
    [anon_sym___kindof] = ACTIONS(4070),
    [anon_sym___nonnull] = ACTIONS(4070),
    [anon_sym___nullable] = ACTIONS(4070),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4070),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4070),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4070),
    [anon_sym___real] = ACTIONS(4070),
    [anon_sym___strong] = ACTIONS(4070),
    [anon_sym___unsafe_unretained] = ACTIONS(4070),
    [anon_sym___unused] = ACTIONS(4070),
    [anon_sym___weak] = ACTIONS(4070),
    [anon_sym_alignas] = ACTIONS(4070),
    [anon_sym__Alignas] = ACTIONS(4070),
    [sym_primitive_type] = ACTIONS(4070),
    [anon_sym_enum] = ACTIONS(4070),
    [anon_sym_COLON] = ACTIONS(4072),
    [anon_sym_struct] = ACTIONS(4070),
    [anon_sym_union] = ACTIONS(4070),
    [anon_sym_QMARK] = ACTIONS(4072),
    [anon_sym_STAR_EQ] = ACTIONS(4072),
    [anon_sym_SLASH_EQ] = ACTIONS(4072),
    [anon_sym_PERCENT_EQ] = ACTIONS(4072),
    [anon_sym_PLUS_EQ] = ACTIONS(4072),
    [anon_sym_DASH_EQ] = ACTIONS(4072),
    [anon_sym_LT_LT_EQ] = ACTIONS(4072),
    [anon_sym_GT_GT_EQ] = ACTIONS(4072),
    [anon_sym_AMP_EQ] = ACTIONS(4072),
    [anon_sym_CARET_EQ] = ACTIONS(4072),
    [anon_sym_PIPE_EQ] = ACTIONS(4072),
    [anon_sym_DASH_DASH] = ACTIONS(4072),
    [anon_sym_PLUS_PLUS] = ACTIONS(4072),
    [anon_sym_DOT] = ACTIONS(4070),
    [anon_sym_DASH_GT] = ACTIONS(4072),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4070),
    [anon_sym___typeof] = ACTIONS(4070),
    [anon_sym_typeof] = ACTIONS(4070),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4070),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4070),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4070),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4070),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4070),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4070),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4070),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4070),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4070),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4070),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4070),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4070),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4070),
    [anon_sym_API_AVAILABLE] = ACTIONS(4070),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4070),
    [anon_sym_API_DEPRECATED] = ACTIONS(4070),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4070),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4070),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4070),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4070),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4070),
    [anon_sym___deprecated_msg] = ACTIONS(4070),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4070),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4070),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4070),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4070),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4070),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4070),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4070),
    [anon_sym_BOOL] = ACTIONS(4070),
    [anon_sym_IMP] = ACTIONS(4070),
    [anon_sym_SEL] = ACTIONS(4070),
    [anon_sym_Class] = ACTIONS(4070),
    [anon_sym_id] = ACTIONS(4070),
  },
  [STATE(1459)] = {
    [sym_identifier] = ACTIONS(4074),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4076),
    [anon_sym_COMMA] = ACTIONS(4076),
    [anon_sym_RPAREN] = ACTIONS(4076),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4076),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4076),
    [anon_sym_LPAREN2] = ACTIONS(4076),
    [anon_sym_DASH] = ACTIONS(4074),
    [anon_sym_PLUS] = ACTIONS(4074),
    [anon_sym_STAR] = ACTIONS(4074),
    [anon_sym_SLASH] = ACTIONS(4074),
    [anon_sym_PERCENT] = ACTIONS(4074),
    [anon_sym_PIPE_PIPE] = ACTIONS(4076),
    [anon_sym_AMP_AMP] = ACTIONS(4076),
    [anon_sym_PIPE] = ACTIONS(4074),
    [anon_sym_CARET] = ACTIONS(4074),
    [anon_sym_AMP] = ACTIONS(4074),
    [anon_sym_EQ_EQ] = ACTIONS(4076),
    [anon_sym_BANG_EQ] = ACTIONS(4076),
    [anon_sym_GT] = ACTIONS(4074),
    [anon_sym_GT_EQ] = ACTIONS(4076),
    [anon_sym_LT_EQ] = ACTIONS(4076),
    [anon_sym_LT] = ACTIONS(4074),
    [anon_sym_LT_LT] = ACTIONS(4074),
    [anon_sym_GT_GT] = ACTIONS(4074),
    [anon_sym_SEMI] = ACTIONS(4076),
    [anon_sym___extension__] = ACTIONS(4074),
    [anon_sym_extern] = ACTIONS(4074),
    [anon_sym___attribute__] = ACTIONS(4074),
    [anon_sym___attribute] = ACTIONS(4074),
    [anon_sym_noreturn] = ACTIONS(4074),
    [anon_sym_LBRACK] = ACTIONS(4076),
    [anon_sym_RBRACK] = ACTIONS(4076),
    [anon_sym___declspec] = ACTIONS(4074),
    [anon_sym_LBRACE] = ACTIONS(4076),
    [anon_sym_RBRACE] = ACTIONS(4076),
    [anon_sym_signed] = ACTIONS(4074),
    [anon_sym_unsigned] = ACTIONS(4074),
    [anon_sym_long] = ACTIONS(4074),
    [anon_sym_short] = ACTIONS(4074),
    [anon_sym_static] = ACTIONS(4074),
    [anon_sym_EQ] = ACTIONS(4074),
    [anon_sym_ATautoreleasepool] = ACTIONS(4076),
    [anon_sym_auto] = ACTIONS(4074),
    [anon_sym_register] = ACTIONS(4074),
    [anon_sym_inline] = ACTIONS(4074),
    [anon_sym___inline] = ACTIONS(4074),
    [anon_sym___inline__] = ACTIONS(4074),
    [anon_sym___forceinline] = ACTIONS(4074),
    [anon_sym_thread_local] = ACTIONS(4074),
    [anon_sym___thread] = ACTIONS(4074),
    [anon_sym_CG_EXTERN] = ACTIONS(4074),
    [anon_sym_CG_INLINE] = ACTIONS(4074),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4074),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4074),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4074),
    [anon_sym_IBOutlet] = ACTIONS(4074),
    [anon_sym_IBInspectable] = ACTIONS(4074),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4074),
    [anon_sym_NS_INLINE] = ACTIONS(4074),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4074),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4074),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4074),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4074),
    [anon_sym_const] = ACTIONS(4074),
    [anon_sym_constexpr] = ACTIONS(4074),
    [anon_sym_volatile] = ACTIONS(4074),
    [anon_sym_restrict] = ACTIONS(4074),
    [anon_sym___restrict__] = ACTIONS(4074),
    [anon_sym__Atomic] = ACTIONS(4074),
    [anon_sym__Noreturn] = ACTIONS(4074),
    [anon_sym__Nonnull] = ACTIONS(4074),
    [anon_sym_nullable] = ACTIONS(4074),
    [anon_sym__Complex] = ACTIONS(4074),
    [anon_sym__Nullable] = ACTIONS(4074),
    [anon_sym__Nullable_result] = ACTIONS(4074),
    [anon_sym__Null_unspecified] = ACTIONS(4074),
    [anon_sym___autoreleasing] = ACTIONS(4074),
    [anon_sym___block] = ACTIONS(4074),
    [anon_sym___bridge] = ACTIONS(4074),
    [anon_sym___bridge_retained] = ACTIONS(4074),
    [anon_sym___bridge_transfer] = ACTIONS(4074),
    [anon_sym___complex] = ACTIONS(4074),
    [anon_sym___const] = ACTIONS(4074),
    [anon_sym___imag] = ACTIONS(4074),
    [anon_sym___kindof] = ACTIONS(4074),
    [anon_sym___nonnull] = ACTIONS(4074),
    [anon_sym___nullable] = ACTIONS(4074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4074),
    [anon_sym___real] = ACTIONS(4074),
    [anon_sym___strong] = ACTIONS(4074),
    [anon_sym___unsafe_unretained] = ACTIONS(4074),
    [anon_sym___unused] = ACTIONS(4074),
    [anon_sym___weak] = ACTIONS(4074),
    [anon_sym_alignas] = ACTIONS(4074),
    [anon_sym__Alignas] = ACTIONS(4074),
    [sym_primitive_type] = ACTIONS(4074),
    [anon_sym_enum] = ACTIONS(4074),
    [anon_sym_COLON] = ACTIONS(4076),
    [anon_sym_struct] = ACTIONS(4074),
    [anon_sym_union] = ACTIONS(4074),
    [anon_sym_QMARK] = ACTIONS(4076),
    [anon_sym_STAR_EQ] = ACTIONS(4076),
    [anon_sym_SLASH_EQ] = ACTIONS(4076),
    [anon_sym_PERCENT_EQ] = ACTIONS(4076),
    [anon_sym_PLUS_EQ] = ACTIONS(4076),
    [anon_sym_DASH_EQ] = ACTIONS(4076),
    [anon_sym_LT_LT_EQ] = ACTIONS(4076),
    [anon_sym_GT_GT_EQ] = ACTIONS(4076),
    [anon_sym_AMP_EQ] = ACTIONS(4076),
    [anon_sym_CARET_EQ] = ACTIONS(4076),
    [anon_sym_PIPE_EQ] = ACTIONS(4076),
    [anon_sym_DASH_DASH] = ACTIONS(4076),
    [anon_sym_PLUS_PLUS] = ACTIONS(4076),
    [anon_sym_DOT] = ACTIONS(4074),
    [anon_sym_DASH_GT] = ACTIONS(4076),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4074),
    [anon_sym___typeof] = ACTIONS(4074),
    [anon_sym_typeof] = ACTIONS(4074),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4074),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4074),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4074),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4074),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4074),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4074),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4074),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4074),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4074),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4074),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4074),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4074),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4074),
    [anon_sym_API_AVAILABLE] = ACTIONS(4074),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4074),
    [anon_sym_API_DEPRECATED] = ACTIONS(4074),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4074),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4074),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4074),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4074),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4074),
    [anon_sym___deprecated_msg] = ACTIONS(4074),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4074),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4074),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4074),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4074),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4074),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4074),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4074),
    [anon_sym_BOOL] = ACTIONS(4074),
    [anon_sym_IMP] = ACTIONS(4074),
    [anon_sym_SEL] = ACTIONS(4074),
    [anon_sym_Class] = ACTIONS(4074),
    [anon_sym_id] = ACTIONS(4074),
  },
  [STATE(1460)] = {
    [sym__declaration_modifiers] = STATE(1575),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(1575),
    [sym_attribute_declaration] = STATE(1575),
    [sym_ms_declspec_modifier] = STATE(1575),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5307),
    [sym__abstract_declarator] = STATE(5183),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_compound_statement] = STATE(6762),
    [sym_storage_class_specifier] = STATE(1575),
    [sym_type_qualifier] = STATE(1575),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5455),
    [sym_parameter_list] = STATE(4271),
    [sym_parameter_declaration] = STATE(5455),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1575),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1575),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4078),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym_RPAREN] = ACTIONS(4082),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(4086),
    [anon_sym_CARET] = ACTIONS(4088),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1461)] = {
    [sym_identifier] = ACTIONS(4094),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4096),
    [anon_sym_COMMA] = ACTIONS(4096),
    [anon_sym_RPAREN] = ACTIONS(4096),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4096),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4096),
    [anon_sym_LPAREN2] = ACTIONS(4096),
    [anon_sym_DASH] = ACTIONS(4094),
    [anon_sym_PLUS] = ACTIONS(4094),
    [anon_sym_STAR] = ACTIONS(4094),
    [anon_sym_SLASH] = ACTIONS(4094),
    [anon_sym_PERCENT] = ACTIONS(4094),
    [anon_sym_PIPE_PIPE] = ACTIONS(4096),
    [anon_sym_AMP_AMP] = ACTIONS(4096),
    [anon_sym_PIPE] = ACTIONS(4094),
    [anon_sym_CARET] = ACTIONS(4094),
    [anon_sym_AMP] = ACTIONS(4094),
    [anon_sym_EQ_EQ] = ACTIONS(4096),
    [anon_sym_BANG_EQ] = ACTIONS(4096),
    [anon_sym_GT] = ACTIONS(4094),
    [anon_sym_GT_EQ] = ACTIONS(4096),
    [anon_sym_LT_EQ] = ACTIONS(4096),
    [anon_sym_LT] = ACTIONS(4094),
    [anon_sym_LT_LT] = ACTIONS(4094),
    [anon_sym_GT_GT] = ACTIONS(4094),
    [anon_sym_SEMI] = ACTIONS(4096),
    [anon_sym___extension__] = ACTIONS(4094),
    [anon_sym_extern] = ACTIONS(4094),
    [anon_sym___attribute__] = ACTIONS(4094),
    [anon_sym___attribute] = ACTIONS(4094),
    [anon_sym_noreturn] = ACTIONS(4094),
    [anon_sym_LBRACK] = ACTIONS(4096),
    [anon_sym_RBRACK] = ACTIONS(4096),
    [anon_sym___declspec] = ACTIONS(4094),
    [anon_sym_LBRACE] = ACTIONS(4096),
    [anon_sym_RBRACE] = ACTIONS(4096),
    [anon_sym_signed] = ACTIONS(4094),
    [anon_sym_unsigned] = ACTIONS(4094),
    [anon_sym_long] = ACTIONS(4094),
    [anon_sym_short] = ACTIONS(4094),
    [anon_sym_static] = ACTIONS(4094),
    [anon_sym_EQ] = ACTIONS(4094),
    [anon_sym_ATautoreleasepool] = ACTIONS(4096),
    [anon_sym_auto] = ACTIONS(4094),
    [anon_sym_register] = ACTIONS(4094),
    [anon_sym_inline] = ACTIONS(4094),
    [anon_sym___inline] = ACTIONS(4094),
    [anon_sym___inline__] = ACTIONS(4094),
    [anon_sym___forceinline] = ACTIONS(4094),
    [anon_sym_thread_local] = ACTIONS(4094),
    [anon_sym___thread] = ACTIONS(4094),
    [anon_sym_CG_EXTERN] = ACTIONS(4094),
    [anon_sym_CG_INLINE] = ACTIONS(4094),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4094),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4094),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4094),
    [anon_sym_IBOutlet] = ACTIONS(4094),
    [anon_sym_IBInspectable] = ACTIONS(4094),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4094),
    [anon_sym_NS_INLINE] = ACTIONS(4094),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4094),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4094),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4094),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4094),
    [anon_sym_const] = ACTIONS(4094),
    [anon_sym_constexpr] = ACTIONS(4094),
    [anon_sym_volatile] = ACTIONS(4094),
    [anon_sym_restrict] = ACTIONS(4094),
    [anon_sym___restrict__] = ACTIONS(4094),
    [anon_sym__Atomic] = ACTIONS(4094),
    [anon_sym__Noreturn] = ACTIONS(4094),
    [anon_sym__Nonnull] = ACTIONS(4094),
    [anon_sym_nullable] = ACTIONS(4094),
    [anon_sym__Complex] = ACTIONS(4094),
    [anon_sym__Nullable] = ACTIONS(4094),
    [anon_sym__Nullable_result] = ACTIONS(4094),
    [anon_sym__Null_unspecified] = ACTIONS(4094),
    [anon_sym___autoreleasing] = ACTIONS(4094),
    [anon_sym___block] = ACTIONS(4094),
    [anon_sym___bridge] = ACTIONS(4094),
    [anon_sym___bridge_retained] = ACTIONS(4094),
    [anon_sym___bridge_transfer] = ACTIONS(4094),
    [anon_sym___complex] = ACTIONS(4094),
    [anon_sym___const] = ACTIONS(4094),
    [anon_sym___imag] = ACTIONS(4094),
    [anon_sym___kindof] = ACTIONS(4094),
    [anon_sym___nonnull] = ACTIONS(4094),
    [anon_sym___nullable] = ACTIONS(4094),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4094),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4094),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4094),
    [anon_sym___real] = ACTIONS(4094),
    [anon_sym___strong] = ACTIONS(4094),
    [anon_sym___unsafe_unretained] = ACTIONS(4094),
    [anon_sym___unused] = ACTIONS(4094),
    [anon_sym___weak] = ACTIONS(4094),
    [anon_sym_alignas] = ACTIONS(4094),
    [anon_sym__Alignas] = ACTIONS(4094),
    [sym_primitive_type] = ACTIONS(4094),
    [anon_sym_enum] = ACTIONS(4094),
    [anon_sym_COLON] = ACTIONS(4096),
    [anon_sym_struct] = ACTIONS(4094),
    [anon_sym_union] = ACTIONS(4094),
    [anon_sym_QMARK] = ACTIONS(4096),
    [anon_sym_STAR_EQ] = ACTIONS(4096),
    [anon_sym_SLASH_EQ] = ACTIONS(4096),
    [anon_sym_PERCENT_EQ] = ACTIONS(4096),
    [anon_sym_PLUS_EQ] = ACTIONS(4096),
    [anon_sym_DASH_EQ] = ACTIONS(4096),
    [anon_sym_LT_LT_EQ] = ACTIONS(4096),
    [anon_sym_GT_GT_EQ] = ACTIONS(4096),
    [anon_sym_AMP_EQ] = ACTIONS(4096),
    [anon_sym_CARET_EQ] = ACTIONS(4096),
    [anon_sym_PIPE_EQ] = ACTIONS(4096),
    [anon_sym_DASH_DASH] = ACTIONS(4096),
    [anon_sym_PLUS_PLUS] = ACTIONS(4096),
    [anon_sym_DOT] = ACTIONS(4094),
    [anon_sym_DASH_GT] = ACTIONS(4096),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4094),
    [anon_sym___typeof] = ACTIONS(4094),
    [anon_sym_typeof] = ACTIONS(4094),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4094),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4094),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4094),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4094),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4094),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4094),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4094),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4094),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4094),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4094),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4094),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4094),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4094),
    [anon_sym_API_AVAILABLE] = ACTIONS(4094),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4094),
    [anon_sym_API_DEPRECATED] = ACTIONS(4094),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4094),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4094),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4094),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4094),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4094),
    [anon_sym___deprecated_msg] = ACTIONS(4094),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4094),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4094),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4094),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4094),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4094),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4094),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4094),
    [anon_sym_BOOL] = ACTIONS(4094),
    [anon_sym_IMP] = ACTIONS(4094),
    [anon_sym_SEL] = ACTIONS(4094),
    [anon_sym_Class] = ACTIONS(4094),
    [anon_sym_id] = ACTIONS(4094),
  },
  [STATE(1462)] = {
    [sym_identifier] = ACTIONS(4098),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4100),
    [anon_sym_COMMA] = ACTIONS(4100),
    [anon_sym_RPAREN] = ACTIONS(4100),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4100),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4100),
    [anon_sym_LPAREN2] = ACTIONS(4100),
    [anon_sym_DASH] = ACTIONS(4098),
    [anon_sym_PLUS] = ACTIONS(4098),
    [anon_sym_STAR] = ACTIONS(4098),
    [anon_sym_SLASH] = ACTIONS(4098),
    [anon_sym_PERCENT] = ACTIONS(4098),
    [anon_sym_PIPE_PIPE] = ACTIONS(4100),
    [anon_sym_AMP_AMP] = ACTIONS(4100),
    [anon_sym_PIPE] = ACTIONS(4098),
    [anon_sym_CARET] = ACTIONS(4098),
    [anon_sym_AMP] = ACTIONS(4098),
    [anon_sym_EQ_EQ] = ACTIONS(4100),
    [anon_sym_BANG_EQ] = ACTIONS(4100),
    [anon_sym_GT] = ACTIONS(4098),
    [anon_sym_GT_EQ] = ACTIONS(4100),
    [anon_sym_LT_EQ] = ACTIONS(4100),
    [anon_sym_LT] = ACTIONS(4098),
    [anon_sym_LT_LT] = ACTIONS(4098),
    [anon_sym_GT_GT] = ACTIONS(4098),
    [anon_sym_SEMI] = ACTIONS(4100),
    [anon_sym___extension__] = ACTIONS(4098),
    [anon_sym_extern] = ACTIONS(4098),
    [anon_sym___attribute__] = ACTIONS(4098),
    [anon_sym___attribute] = ACTIONS(4098),
    [anon_sym_noreturn] = ACTIONS(4098),
    [anon_sym_LBRACK] = ACTIONS(4100),
    [anon_sym_RBRACK] = ACTIONS(4100),
    [anon_sym___declspec] = ACTIONS(4098),
    [anon_sym_LBRACE] = ACTIONS(4100),
    [anon_sym_RBRACE] = ACTIONS(4100),
    [anon_sym_signed] = ACTIONS(4098),
    [anon_sym_unsigned] = ACTIONS(4098),
    [anon_sym_long] = ACTIONS(4098),
    [anon_sym_short] = ACTIONS(4098),
    [anon_sym_static] = ACTIONS(4098),
    [anon_sym_EQ] = ACTIONS(4098),
    [anon_sym_ATautoreleasepool] = ACTIONS(4100),
    [anon_sym_auto] = ACTIONS(4098),
    [anon_sym_register] = ACTIONS(4098),
    [anon_sym_inline] = ACTIONS(4098),
    [anon_sym___inline] = ACTIONS(4098),
    [anon_sym___inline__] = ACTIONS(4098),
    [anon_sym___forceinline] = ACTIONS(4098),
    [anon_sym_thread_local] = ACTIONS(4098),
    [anon_sym___thread] = ACTIONS(4098),
    [anon_sym_CG_EXTERN] = ACTIONS(4098),
    [anon_sym_CG_INLINE] = ACTIONS(4098),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4098),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4098),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4098),
    [anon_sym_IBOutlet] = ACTIONS(4098),
    [anon_sym_IBInspectable] = ACTIONS(4098),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4098),
    [anon_sym_NS_INLINE] = ACTIONS(4098),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4098),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4098),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4098),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4098),
    [anon_sym_const] = ACTIONS(4098),
    [anon_sym_constexpr] = ACTIONS(4098),
    [anon_sym_volatile] = ACTIONS(4098),
    [anon_sym_restrict] = ACTIONS(4098),
    [anon_sym___restrict__] = ACTIONS(4098),
    [anon_sym__Atomic] = ACTIONS(4098),
    [anon_sym__Noreturn] = ACTIONS(4098),
    [anon_sym__Nonnull] = ACTIONS(4098),
    [anon_sym_nullable] = ACTIONS(4098),
    [anon_sym__Complex] = ACTIONS(4098),
    [anon_sym__Nullable] = ACTIONS(4098),
    [anon_sym__Nullable_result] = ACTIONS(4098),
    [anon_sym__Null_unspecified] = ACTIONS(4098),
    [anon_sym___autoreleasing] = ACTIONS(4098),
    [anon_sym___block] = ACTIONS(4098),
    [anon_sym___bridge] = ACTIONS(4098),
    [anon_sym___bridge_retained] = ACTIONS(4098),
    [anon_sym___bridge_transfer] = ACTIONS(4098),
    [anon_sym___complex] = ACTIONS(4098),
    [anon_sym___const] = ACTIONS(4098),
    [anon_sym___imag] = ACTIONS(4098),
    [anon_sym___kindof] = ACTIONS(4098),
    [anon_sym___nonnull] = ACTIONS(4098),
    [anon_sym___nullable] = ACTIONS(4098),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4098),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4098),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4098),
    [anon_sym___real] = ACTIONS(4098),
    [anon_sym___strong] = ACTIONS(4098),
    [anon_sym___unsafe_unretained] = ACTIONS(4098),
    [anon_sym___unused] = ACTIONS(4098),
    [anon_sym___weak] = ACTIONS(4098),
    [anon_sym_alignas] = ACTIONS(4098),
    [anon_sym__Alignas] = ACTIONS(4098),
    [sym_primitive_type] = ACTIONS(4098),
    [anon_sym_enum] = ACTIONS(4098),
    [anon_sym_COLON] = ACTIONS(4100),
    [anon_sym_struct] = ACTIONS(4098),
    [anon_sym_union] = ACTIONS(4098),
    [anon_sym_QMARK] = ACTIONS(4100),
    [anon_sym_STAR_EQ] = ACTIONS(4100),
    [anon_sym_SLASH_EQ] = ACTIONS(4100),
    [anon_sym_PERCENT_EQ] = ACTIONS(4100),
    [anon_sym_PLUS_EQ] = ACTIONS(4100),
    [anon_sym_DASH_EQ] = ACTIONS(4100),
    [anon_sym_LT_LT_EQ] = ACTIONS(4100),
    [anon_sym_GT_GT_EQ] = ACTIONS(4100),
    [anon_sym_AMP_EQ] = ACTIONS(4100),
    [anon_sym_CARET_EQ] = ACTIONS(4100),
    [anon_sym_PIPE_EQ] = ACTIONS(4100),
    [anon_sym_DASH_DASH] = ACTIONS(4100),
    [anon_sym_PLUS_PLUS] = ACTIONS(4100),
    [anon_sym_DOT] = ACTIONS(4098),
    [anon_sym_DASH_GT] = ACTIONS(4100),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4098),
    [anon_sym___typeof] = ACTIONS(4098),
    [anon_sym_typeof] = ACTIONS(4098),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4098),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4098),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4098),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4098),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4098),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4098),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4098),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4098),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4098),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4098),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4098),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4098),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4098),
    [anon_sym_API_AVAILABLE] = ACTIONS(4098),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4098),
    [anon_sym_API_DEPRECATED] = ACTIONS(4098),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4098),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4098),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4098),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4098),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4098),
    [anon_sym___deprecated_msg] = ACTIONS(4098),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4098),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4098),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4098),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4098),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4098),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4098),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4098),
    [anon_sym_BOOL] = ACTIONS(4098),
    [anon_sym_IMP] = ACTIONS(4098),
    [anon_sym_SEL] = ACTIONS(4098),
    [anon_sym_Class] = ACTIONS(4098),
    [anon_sym_id] = ACTIONS(4098),
  },
  [STATE(1463)] = {
    [sym_identifier] = ACTIONS(4102),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4104),
    [anon_sym_COMMA] = ACTIONS(4104),
    [anon_sym_RPAREN] = ACTIONS(4104),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4104),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4104),
    [anon_sym_LPAREN2] = ACTIONS(4104),
    [anon_sym_DASH] = ACTIONS(4102),
    [anon_sym_PLUS] = ACTIONS(4102),
    [anon_sym_STAR] = ACTIONS(4102),
    [anon_sym_SLASH] = ACTIONS(4102),
    [anon_sym_PERCENT] = ACTIONS(4102),
    [anon_sym_PIPE_PIPE] = ACTIONS(4104),
    [anon_sym_AMP_AMP] = ACTIONS(4104),
    [anon_sym_PIPE] = ACTIONS(4102),
    [anon_sym_CARET] = ACTIONS(4102),
    [anon_sym_AMP] = ACTIONS(4102),
    [anon_sym_EQ_EQ] = ACTIONS(4104),
    [anon_sym_BANG_EQ] = ACTIONS(4104),
    [anon_sym_GT] = ACTIONS(4102),
    [anon_sym_GT_EQ] = ACTIONS(4104),
    [anon_sym_LT_EQ] = ACTIONS(4104),
    [anon_sym_LT] = ACTIONS(4102),
    [anon_sym_LT_LT] = ACTIONS(4102),
    [anon_sym_GT_GT] = ACTIONS(4102),
    [anon_sym_SEMI] = ACTIONS(4104),
    [anon_sym___extension__] = ACTIONS(4102),
    [anon_sym_extern] = ACTIONS(4102),
    [anon_sym___attribute__] = ACTIONS(4102),
    [anon_sym___attribute] = ACTIONS(4102),
    [anon_sym_noreturn] = ACTIONS(4102),
    [anon_sym_LBRACK] = ACTIONS(4104),
    [anon_sym_RBRACK] = ACTIONS(4104),
    [anon_sym___declspec] = ACTIONS(4102),
    [anon_sym_LBRACE] = ACTIONS(4104),
    [anon_sym_RBRACE] = ACTIONS(4104),
    [anon_sym_signed] = ACTIONS(4102),
    [anon_sym_unsigned] = ACTIONS(4102),
    [anon_sym_long] = ACTIONS(4102),
    [anon_sym_short] = ACTIONS(4102),
    [anon_sym_static] = ACTIONS(4102),
    [anon_sym_EQ] = ACTIONS(4102),
    [anon_sym_ATautoreleasepool] = ACTIONS(4104),
    [anon_sym_auto] = ACTIONS(4102),
    [anon_sym_register] = ACTIONS(4102),
    [anon_sym_inline] = ACTIONS(4102),
    [anon_sym___inline] = ACTIONS(4102),
    [anon_sym___inline__] = ACTIONS(4102),
    [anon_sym___forceinline] = ACTIONS(4102),
    [anon_sym_thread_local] = ACTIONS(4102),
    [anon_sym___thread] = ACTIONS(4102),
    [anon_sym_CG_EXTERN] = ACTIONS(4102),
    [anon_sym_CG_INLINE] = ACTIONS(4102),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4102),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4102),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4102),
    [anon_sym_IBOutlet] = ACTIONS(4102),
    [anon_sym_IBInspectable] = ACTIONS(4102),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4102),
    [anon_sym_NS_INLINE] = ACTIONS(4102),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4102),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4102),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4102),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4102),
    [anon_sym_const] = ACTIONS(4102),
    [anon_sym_constexpr] = ACTIONS(4102),
    [anon_sym_volatile] = ACTIONS(4102),
    [anon_sym_restrict] = ACTIONS(4102),
    [anon_sym___restrict__] = ACTIONS(4102),
    [anon_sym__Atomic] = ACTIONS(4102),
    [anon_sym__Noreturn] = ACTIONS(4102),
    [anon_sym__Nonnull] = ACTIONS(4102),
    [anon_sym_nullable] = ACTIONS(4102),
    [anon_sym__Complex] = ACTIONS(4102),
    [anon_sym__Nullable] = ACTIONS(4102),
    [anon_sym__Nullable_result] = ACTIONS(4102),
    [anon_sym__Null_unspecified] = ACTIONS(4102),
    [anon_sym___autoreleasing] = ACTIONS(4102),
    [anon_sym___block] = ACTIONS(4102),
    [anon_sym___bridge] = ACTIONS(4102),
    [anon_sym___bridge_retained] = ACTIONS(4102),
    [anon_sym___bridge_transfer] = ACTIONS(4102),
    [anon_sym___complex] = ACTIONS(4102),
    [anon_sym___const] = ACTIONS(4102),
    [anon_sym___imag] = ACTIONS(4102),
    [anon_sym___kindof] = ACTIONS(4102),
    [anon_sym___nonnull] = ACTIONS(4102),
    [anon_sym___nullable] = ACTIONS(4102),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4102),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4102),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4102),
    [anon_sym___real] = ACTIONS(4102),
    [anon_sym___strong] = ACTIONS(4102),
    [anon_sym___unsafe_unretained] = ACTIONS(4102),
    [anon_sym___unused] = ACTIONS(4102),
    [anon_sym___weak] = ACTIONS(4102),
    [anon_sym_alignas] = ACTIONS(4102),
    [anon_sym__Alignas] = ACTIONS(4102),
    [sym_primitive_type] = ACTIONS(4102),
    [anon_sym_enum] = ACTIONS(4102),
    [anon_sym_COLON] = ACTIONS(4104),
    [anon_sym_struct] = ACTIONS(4102),
    [anon_sym_union] = ACTIONS(4102),
    [anon_sym_QMARK] = ACTIONS(4104),
    [anon_sym_STAR_EQ] = ACTIONS(4104),
    [anon_sym_SLASH_EQ] = ACTIONS(4104),
    [anon_sym_PERCENT_EQ] = ACTIONS(4104),
    [anon_sym_PLUS_EQ] = ACTIONS(4104),
    [anon_sym_DASH_EQ] = ACTIONS(4104),
    [anon_sym_LT_LT_EQ] = ACTIONS(4104),
    [anon_sym_GT_GT_EQ] = ACTIONS(4104),
    [anon_sym_AMP_EQ] = ACTIONS(4104),
    [anon_sym_CARET_EQ] = ACTIONS(4104),
    [anon_sym_PIPE_EQ] = ACTIONS(4104),
    [anon_sym_DASH_DASH] = ACTIONS(4104),
    [anon_sym_PLUS_PLUS] = ACTIONS(4104),
    [anon_sym_DOT] = ACTIONS(4102),
    [anon_sym_DASH_GT] = ACTIONS(4104),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4102),
    [anon_sym___typeof] = ACTIONS(4102),
    [anon_sym_typeof] = ACTIONS(4102),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4102),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4102),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4102),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4102),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4102),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4102),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4102),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4102),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4102),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4102),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4102),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4102),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4102),
    [anon_sym_API_AVAILABLE] = ACTIONS(4102),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4102),
    [anon_sym_API_DEPRECATED] = ACTIONS(4102),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4102),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4102),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4102),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4102),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4102),
    [anon_sym___deprecated_msg] = ACTIONS(4102),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4102),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4102),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4102),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4102),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4102),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4102),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4102),
    [anon_sym_BOOL] = ACTIONS(4102),
    [anon_sym_IMP] = ACTIONS(4102),
    [anon_sym_SEL] = ACTIONS(4102),
    [anon_sym_Class] = ACTIONS(4102),
    [anon_sym_id] = ACTIONS(4102),
  },
  [STATE(1464)] = {
    [sym_identifier] = ACTIONS(4106),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4108),
    [anon_sym_COMMA] = ACTIONS(4108),
    [anon_sym_RPAREN] = ACTIONS(4108),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4108),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4108),
    [anon_sym_LPAREN2] = ACTIONS(4108),
    [anon_sym_DASH] = ACTIONS(4106),
    [anon_sym_PLUS] = ACTIONS(4106),
    [anon_sym_STAR] = ACTIONS(4106),
    [anon_sym_SLASH] = ACTIONS(4106),
    [anon_sym_PERCENT] = ACTIONS(4106),
    [anon_sym_PIPE_PIPE] = ACTIONS(4108),
    [anon_sym_AMP_AMP] = ACTIONS(4108),
    [anon_sym_PIPE] = ACTIONS(4106),
    [anon_sym_CARET] = ACTIONS(4106),
    [anon_sym_AMP] = ACTIONS(4106),
    [anon_sym_EQ_EQ] = ACTIONS(4108),
    [anon_sym_BANG_EQ] = ACTIONS(4108),
    [anon_sym_GT] = ACTIONS(4106),
    [anon_sym_GT_EQ] = ACTIONS(4108),
    [anon_sym_LT_EQ] = ACTIONS(4108),
    [anon_sym_LT] = ACTIONS(4106),
    [anon_sym_LT_LT] = ACTIONS(4106),
    [anon_sym_GT_GT] = ACTIONS(4106),
    [anon_sym_SEMI] = ACTIONS(4108),
    [anon_sym___extension__] = ACTIONS(4106),
    [anon_sym_extern] = ACTIONS(4106),
    [anon_sym___attribute__] = ACTIONS(4106),
    [anon_sym___attribute] = ACTIONS(4106),
    [anon_sym_noreturn] = ACTIONS(4106),
    [anon_sym_LBRACK] = ACTIONS(4108),
    [anon_sym_RBRACK] = ACTIONS(4108),
    [anon_sym___declspec] = ACTIONS(4106),
    [anon_sym_LBRACE] = ACTIONS(4108),
    [anon_sym_RBRACE] = ACTIONS(4108),
    [anon_sym_signed] = ACTIONS(4106),
    [anon_sym_unsigned] = ACTIONS(4106),
    [anon_sym_long] = ACTIONS(4106),
    [anon_sym_short] = ACTIONS(4106),
    [anon_sym_static] = ACTIONS(4106),
    [anon_sym_EQ] = ACTIONS(4106),
    [anon_sym_ATautoreleasepool] = ACTIONS(4108),
    [anon_sym_auto] = ACTIONS(4106),
    [anon_sym_register] = ACTIONS(4106),
    [anon_sym_inline] = ACTIONS(4106),
    [anon_sym___inline] = ACTIONS(4106),
    [anon_sym___inline__] = ACTIONS(4106),
    [anon_sym___forceinline] = ACTIONS(4106),
    [anon_sym_thread_local] = ACTIONS(4106),
    [anon_sym___thread] = ACTIONS(4106),
    [anon_sym_CG_EXTERN] = ACTIONS(4106),
    [anon_sym_CG_INLINE] = ACTIONS(4106),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4106),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4106),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4106),
    [anon_sym_IBOutlet] = ACTIONS(4106),
    [anon_sym_IBInspectable] = ACTIONS(4106),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4106),
    [anon_sym_NS_INLINE] = ACTIONS(4106),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4106),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4106),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4106),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4106),
    [anon_sym_const] = ACTIONS(4106),
    [anon_sym_constexpr] = ACTIONS(4106),
    [anon_sym_volatile] = ACTIONS(4106),
    [anon_sym_restrict] = ACTIONS(4106),
    [anon_sym___restrict__] = ACTIONS(4106),
    [anon_sym__Atomic] = ACTIONS(4106),
    [anon_sym__Noreturn] = ACTIONS(4106),
    [anon_sym__Nonnull] = ACTIONS(4106),
    [anon_sym_nullable] = ACTIONS(4106),
    [anon_sym__Complex] = ACTIONS(4106),
    [anon_sym__Nullable] = ACTIONS(4106),
    [anon_sym__Nullable_result] = ACTIONS(4106),
    [anon_sym__Null_unspecified] = ACTIONS(4106),
    [anon_sym___autoreleasing] = ACTIONS(4106),
    [anon_sym___block] = ACTIONS(4106),
    [anon_sym___bridge] = ACTIONS(4106),
    [anon_sym___bridge_retained] = ACTIONS(4106),
    [anon_sym___bridge_transfer] = ACTIONS(4106),
    [anon_sym___complex] = ACTIONS(4106),
    [anon_sym___const] = ACTIONS(4106),
    [anon_sym___imag] = ACTIONS(4106),
    [anon_sym___kindof] = ACTIONS(4106),
    [anon_sym___nonnull] = ACTIONS(4106),
    [anon_sym___nullable] = ACTIONS(4106),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4106),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4106),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4106),
    [anon_sym___real] = ACTIONS(4106),
    [anon_sym___strong] = ACTIONS(4106),
    [anon_sym___unsafe_unretained] = ACTIONS(4106),
    [anon_sym___unused] = ACTIONS(4106),
    [anon_sym___weak] = ACTIONS(4106),
    [anon_sym_alignas] = ACTIONS(4106),
    [anon_sym__Alignas] = ACTIONS(4106),
    [sym_primitive_type] = ACTIONS(4106),
    [anon_sym_enum] = ACTIONS(4106),
    [anon_sym_COLON] = ACTIONS(4108),
    [anon_sym_struct] = ACTIONS(4106),
    [anon_sym_union] = ACTIONS(4106),
    [anon_sym_QMARK] = ACTIONS(4108),
    [anon_sym_STAR_EQ] = ACTIONS(4108),
    [anon_sym_SLASH_EQ] = ACTIONS(4108),
    [anon_sym_PERCENT_EQ] = ACTIONS(4108),
    [anon_sym_PLUS_EQ] = ACTIONS(4108),
    [anon_sym_DASH_EQ] = ACTIONS(4108),
    [anon_sym_LT_LT_EQ] = ACTIONS(4108),
    [anon_sym_GT_GT_EQ] = ACTIONS(4108),
    [anon_sym_AMP_EQ] = ACTIONS(4108),
    [anon_sym_CARET_EQ] = ACTIONS(4108),
    [anon_sym_PIPE_EQ] = ACTIONS(4108),
    [anon_sym_DASH_DASH] = ACTIONS(4108),
    [anon_sym_PLUS_PLUS] = ACTIONS(4108),
    [anon_sym_DOT] = ACTIONS(4106),
    [anon_sym_DASH_GT] = ACTIONS(4108),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4106),
    [anon_sym___typeof] = ACTIONS(4106),
    [anon_sym_typeof] = ACTIONS(4106),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4106),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4106),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4106),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4106),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4106),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4106),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4106),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4106),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4106),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4106),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4106),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4106),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4106),
    [anon_sym_API_AVAILABLE] = ACTIONS(4106),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4106),
    [anon_sym_API_DEPRECATED] = ACTIONS(4106),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4106),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4106),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4106),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4106),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4106),
    [anon_sym___deprecated_msg] = ACTIONS(4106),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4106),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4106),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4106),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4106),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4106),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4106),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4106),
    [anon_sym_BOOL] = ACTIONS(4106),
    [anon_sym_IMP] = ACTIONS(4106),
    [anon_sym_SEL] = ACTIONS(4106),
    [anon_sym_Class] = ACTIONS(4106),
    [anon_sym_id] = ACTIONS(4106),
  },
  [STATE(1465)] = {
    [sym_identifier] = ACTIONS(4024),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_RPAREN] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4024),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4024),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4024),
    [anon_sym_extern] = ACTIONS(4024),
    [anon_sym___attribute__] = ACTIONS(4024),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4024),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym_RBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4024),
    [anon_sym_LBRACE] = ACTIONS(4026),
    [anon_sym_RBRACE] = ACTIONS(4026),
    [anon_sym_signed] = ACTIONS(4024),
    [anon_sym_unsigned] = ACTIONS(4024),
    [anon_sym_long] = ACTIONS(4024),
    [anon_sym_short] = ACTIONS(4024),
    [anon_sym_static] = ACTIONS(4024),
    [anon_sym_EQ] = ACTIONS(4024),
    [anon_sym_ATautoreleasepool] = ACTIONS(4026),
    [anon_sym_auto] = ACTIONS(4024),
    [anon_sym_register] = ACTIONS(4024),
    [anon_sym_inline] = ACTIONS(4024),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4024),
    [anon_sym___forceinline] = ACTIONS(4024),
    [anon_sym_thread_local] = ACTIONS(4024),
    [anon_sym___thread] = ACTIONS(4024),
    [anon_sym_CG_EXTERN] = ACTIONS(4024),
    [anon_sym_CG_INLINE] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4024),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4024),
    [anon_sym_IBOutlet] = ACTIONS(4024),
    [anon_sym_IBInspectable] = ACTIONS(4024),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4024),
    [anon_sym_NS_INLINE] = ACTIONS(4024),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4024),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4024),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4024),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4024),
    [anon_sym_volatile] = ACTIONS(4024),
    [anon_sym_restrict] = ACTIONS(4024),
    [anon_sym___restrict__] = ACTIONS(4024),
    [anon_sym__Atomic] = ACTIONS(4024),
    [anon_sym__Noreturn] = ACTIONS(4024),
    [anon_sym__Nonnull] = ACTIONS(4024),
    [anon_sym_nullable] = ACTIONS(4024),
    [anon_sym__Complex] = ACTIONS(4024),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4024),
    [anon_sym__Null_unspecified] = ACTIONS(4024),
    [anon_sym___autoreleasing] = ACTIONS(4024),
    [anon_sym___block] = ACTIONS(4024),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4024),
    [anon_sym___bridge_transfer] = ACTIONS(4024),
    [anon_sym___complex] = ACTIONS(4024),
    [anon_sym___const] = ACTIONS(4024),
    [anon_sym___imag] = ACTIONS(4024),
    [anon_sym___kindof] = ACTIONS(4024),
    [anon_sym___nonnull] = ACTIONS(4024),
    [anon_sym___nullable] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4024),
    [anon_sym___real] = ACTIONS(4024),
    [anon_sym___strong] = ACTIONS(4024),
    [anon_sym___unsafe_unretained] = ACTIONS(4024),
    [anon_sym___unused] = ACTIONS(4024),
    [anon_sym___weak] = ACTIONS(4024),
    [anon_sym_alignas] = ACTIONS(4024),
    [anon_sym__Alignas] = ACTIONS(4024),
    [sym_primitive_type] = ACTIONS(4024),
    [anon_sym_enum] = ACTIONS(4024),
    [anon_sym_COLON] = ACTIONS(4026),
    [anon_sym_struct] = ACTIONS(4024),
    [anon_sym_union] = ACTIONS(4024),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4026),
    [anon_sym_SLASH_EQ] = ACTIONS(4026),
    [anon_sym_PERCENT_EQ] = ACTIONS(4026),
    [anon_sym_PLUS_EQ] = ACTIONS(4026),
    [anon_sym_DASH_EQ] = ACTIONS(4026),
    [anon_sym_LT_LT_EQ] = ACTIONS(4026),
    [anon_sym_GT_GT_EQ] = ACTIONS(4026),
    [anon_sym_AMP_EQ] = ACTIONS(4026),
    [anon_sym_CARET_EQ] = ACTIONS(4026),
    [anon_sym_PIPE_EQ] = ACTIONS(4026),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4024),
    [anon_sym___typeof] = ACTIONS(4024),
    [anon_sym_typeof] = ACTIONS(4024),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4024),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4024),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_API_AVAILABLE] = ACTIONS(4024),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_API_DEPRECATED] = ACTIONS(4024),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym___deprecated_msg] = ACTIONS(4024),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4024),
    [anon_sym_BOOL] = ACTIONS(4024),
    [anon_sym_IMP] = ACTIONS(4024),
    [anon_sym_SEL] = ACTIONS(4024),
    [anon_sym_Class] = ACTIONS(4024),
    [anon_sym_id] = ACTIONS(4024),
  },
  [STATE(1466)] = {
    [sym_identifier] = ACTIONS(4110),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4112),
    [anon_sym_COMMA] = ACTIONS(4112),
    [anon_sym_RPAREN] = ACTIONS(4112),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4112),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4112),
    [anon_sym_LPAREN2] = ACTIONS(4112),
    [anon_sym_DASH] = ACTIONS(4110),
    [anon_sym_PLUS] = ACTIONS(4110),
    [anon_sym_STAR] = ACTIONS(4110),
    [anon_sym_SLASH] = ACTIONS(4110),
    [anon_sym_PERCENT] = ACTIONS(4110),
    [anon_sym_PIPE_PIPE] = ACTIONS(4112),
    [anon_sym_AMP_AMP] = ACTIONS(4112),
    [anon_sym_PIPE] = ACTIONS(4110),
    [anon_sym_CARET] = ACTIONS(4110),
    [anon_sym_AMP] = ACTIONS(4110),
    [anon_sym_EQ_EQ] = ACTIONS(4112),
    [anon_sym_BANG_EQ] = ACTIONS(4112),
    [anon_sym_GT] = ACTIONS(4110),
    [anon_sym_GT_EQ] = ACTIONS(4112),
    [anon_sym_LT_EQ] = ACTIONS(4112),
    [anon_sym_LT] = ACTIONS(4110),
    [anon_sym_LT_LT] = ACTIONS(4110),
    [anon_sym_GT_GT] = ACTIONS(4110),
    [anon_sym_SEMI] = ACTIONS(4112),
    [anon_sym___extension__] = ACTIONS(4110),
    [anon_sym_extern] = ACTIONS(4110),
    [anon_sym___attribute__] = ACTIONS(4110),
    [anon_sym___attribute] = ACTIONS(4110),
    [anon_sym_noreturn] = ACTIONS(4110),
    [anon_sym_LBRACK] = ACTIONS(4112),
    [anon_sym_RBRACK] = ACTIONS(4112),
    [anon_sym___declspec] = ACTIONS(4110),
    [anon_sym_LBRACE] = ACTIONS(4112),
    [anon_sym_RBRACE] = ACTIONS(4112),
    [anon_sym_signed] = ACTIONS(4110),
    [anon_sym_unsigned] = ACTIONS(4110),
    [anon_sym_long] = ACTIONS(4110),
    [anon_sym_short] = ACTIONS(4110),
    [anon_sym_static] = ACTIONS(4110),
    [anon_sym_EQ] = ACTIONS(4110),
    [anon_sym_ATautoreleasepool] = ACTIONS(4112),
    [anon_sym_auto] = ACTIONS(4110),
    [anon_sym_register] = ACTIONS(4110),
    [anon_sym_inline] = ACTIONS(4110),
    [anon_sym___inline] = ACTIONS(4110),
    [anon_sym___inline__] = ACTIONS(4110),
    [anon_sym___forceinline] = ACTIONS(4110),
    [anon_sym_thread_local] = ACTIONS(4110),
    [anon_sym___thread] = ACTIONS(4110),
    [anon_sym_CG_EXTERN] = ACTIONS(4110),
    [anon_sym_CG_INLINE] = ACTIONS(4110),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4110),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4110),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4110),
    [anon_sym_IBOutlet] = ACTIONS(4110),
    [anon_sym_IBInspectable] = ACTIONS(4110),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4110),
    [anon_sym_NS_INLINE] = ACTIONS(4110),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4110),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4110),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4110),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4110),
    [anon_sym_const] = ACTIONS(4110),
    [anon_sym_constexpr] = ACTIONS(4110),
    [anon_sym_volatile] = ACTIONS(4110),
    [anon_sym_restrict] = ACTIONS(4110),
    [anon_sym___restrict__] = ACTIONS(4110),
    [anon_sym__Atomic] = ACTIONS(4110),
    [anon_sym__Noreturn] = ACTIONS(4110),
    [anon_sym__Nonnull] = ACTIONS(4110),
    [anon_sym_nullable] = ACTIONS(4110),
    [anon_sym__Complex] = ACTIONS(4110),
    [anon_sym__Nullable] = ACTIONS(4110),
    [anon_sym__Nullable_result] = ACTIONS(4110),
    [anon_sym__Null_unspecified] = ACTIONS(4110),
    [anon_sym___autoreleasing] = ACTIONS(4110),
    [anon_sym___block] = ACTIONS(4110),
    [anon_sym___bridge] = ACTIONS(4110),
    [anon_sym___bridge_retained] = ACTIONS(4110),
    [anon_sym___bridge_transfer] = ACTIONS(4110),
    [anon_sym___complex] = ACTIONS(4110),
    [anon_sym___const] = ACTIONS(4110),
    [anon_sym___imag] = ACTIONS(4110),
    [anon_sym___kindof] = ACTIONS(4110),
    [anon_sym___nonnull] = ACTIONS(4110),
    [anon_sym___nullable] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4110),
    [anon_sym___real] = ACTIONS(4110),
    [anon_sym___strong] = ACTIONS(4110),
    [anon_sym___unsafe_unretained] = ACTIONS(4110),
    [anon_sym___unused] = ACTIONS(4110),
    [anon_sym___weak] = ACTIONS(4110),
    [anon_sym_alignas] = ACTIONS(4110),
    [anon_sym__Alignas] = ACTIONS(4110),
    [sym_primitive_type] = ACTIONS(4110),
    [anon_sym_enum] = ACTIONS(4110),
    [anon_sym_COLON] = ACTIONS(4112),
    [anon_sym_struct] = ACTIONS(4110),
    [anon_sym_union] = ACTIONS(4110),
    [anon_sym_QMARK] = ACTIONS(4112),
    [anon_sym_STAR_EQ] = ACTIONS(4112),
    [anon_sym_SLASH_EQ] = ACTIONS(4112),
    [anon_sym_PERCENT_EQ] = ACTIONS(4112),
    [anon_sym_PLUS_EQ] = ACTIONS(4112),
    [anon_sym_DASH_EQ] = ACTIONS(4112),
    [anon_sym_LT_LT_EQ] = ACTIONS(4112),
    [anon_sym_GT_GT_EQ] = ACTIONS(4112),
    [anon_sym_AMP_EQ] = ACTIONS(4112),
    [anon_sym_CARET_EQ] = ACTIONS(4112),
    [anon_sym_PIPE_EQ] = ACTIONS(4112),
    [anon_sym_DASH_DASH] = ACTIONS(4112),
    [anon_sym_PLUS_PLUS] = ACTIONS(4112),
    [anon_sym_DOT] = ACTIONS(4110),
    [anon_sym_DASH_GT] = ACTIONS(4112),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4110),
    [anon_sym___typeof] = ACTIONS(4110),
    [anon_sym_typeof] = ACTIONS(4110),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4110),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4110),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4110),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4110),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4110),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4110),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4110),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4110),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4110),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_API_AVAILABLE] = ACTIONS(4110),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_API_DEPRECATED] = ACTIONS(4110),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4110),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4110),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4110),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4110),
    [anon_sym___deprecated_msg] = ACTIONS(4110),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4110),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4110),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4110),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4110),
    [anon_sym_BOOL] = ACTIONS(4110),
    [anon_sym_IMP] = ACTIONS(4110),
    [anon_sym_SEL] = ACTIONS(4110),
    [anon_sym_Class] = ACTIONS(4110),
    [anon_sym_id] = ACTIONS(4110),
  },
  [STATE(1467)] = {
    [sym_identifier] = ACTIONS(4114),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4116),
    [anon_sym_COMMA] = ACTIONS(4116),
    [anon_sym_RPAREN] = ACTIONS(4116),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4116),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4116),
    [anon_sym_LPAREN2] = ACTIONS(4116),
    [anon_sym_DASH] = ACTIONS(4114),
    [anon_sym_PLUS] = ACTIONS(4114),
    [anon_sym_STAR] = ACTIONS(4114),
    [anon_sym_SLASH] = ACTIONS(4114),
    [anon_sym_PERCENT] = ACTIONS(4114),
    [anon_sym_PIPE_PIPE] = ACTIONS(4116),
    [anon_sym_AMP_AMP] = ACTIONS(4116),
    [anon_sym_PIPE] = ACTIONS(4114),
    [anon_sym_CARET] = ACTIONS(4114),
    [anon_sym_AMP] = ACTIONS(4114),
    [anon_sym_EQ_EQ] = ACTIONS(4116),
    [anon_sym_BANG_EQ] = ACTIONS(4116),
    [anon_sym_GT] = ACTIONS(4114),
    [anon_sym_GT_EQ] = ACTIONS(4116),
    [anon_sym_LT_EQ] = ACTIONS(4116),
    [anon_sym_LT] = ACTIONS(4114),
    [anon_sym_LT_LT] = ACTIONS(4114),
    [anon_sym_GT_GT] = ACTIONS(4114),
    [anon_sym_SEMI] = ACTIONS(4116),
    [anon_sym___extension__] = ACTIONS(4114),
    [anon_sym_extern] = ACTIONS(4114),
    [anon_sym___attribute__] = ACTIONS(4114),
    [anon_sym___attribute] = ACTIONS(4114),
    [anon_sym_noreturn] = ACTIONS(4114),
    [anon_sym_LBRACK] = ACTIONS(4116),
    [anon_sym_RBRACK] = ACTIONS(4116),
    [anon_sym___declspec] = ACTIONS(4114),
    [anon_sym_LBRACE] = ACTIONS(4116),
    [anon_sym_RBRACE] = ACTIONS(4116),
    [anon_sym_signed] = ACTIONS(4114),
    [anon_sym_unsigned] = ACTIONS(4114),
    [anon_sym_long] = ACTIONS(4114),
    [anon_sym_short] = ACTIONS(4114),
    [anon_sym_static] = ACTIONS(4114),
    [anon_sym_EQ] = ACTIONS(4114),
    [anon_sym_ATautoreleasepool] = ACTIONS(4116),
    [anon_sym_auto] = ACTIONS(4114),
    [anon_sym_register] = ACTIONS(4114),
    [anon_sym_inline] = ACTIONS(4114),
    [anon_sym___inline] = ACTIONS(4114),
    [anon_sym___inline__] = ACTIONS(4114),
    [anon_sym___forceinline] = ACTIONS(4114),
    [anon_sym_thread_local] = ACTIONS(4114),
    [anon_sym___thread] = ACTIONS(4114),
    [anon_sym_CG_EXTERN] = ACTIONS(4114),
    [anon_sym_CG_INLINE] = ACTIONS(4114),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4114),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4114),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4114),
    [anon_sym_IBOutlet] = ACTIONS(4114),
    [anon_sym_IBInspectable] = ACTIONS(4114),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4114),
    [anon_sym_NS_INLINE] = ACTIONS(4114),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4114),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4114),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4114),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4114),
    [anon_sym_const] = ACTIONS(4114),
    [anon_sym_constexpr] = ACTIONS(4114),
    [anon_sym_volatile] = ACTIONS(4114),
    [anon_sym_restrict] = ACTIONS(4114),
    [anon_sym___restrict__] = ACTIONS(4114),
    [anon_sym__Atomic] = ACTIONS(4114),
    [anon_sym__Noreturn] = ACTIONS(4114),
    [anon_sym__Nonnull] = ACTIONS(4114),
    [anon_sym_nullable] = ACTIONS(4114),
    [anon_sym__Complex] = ACTIONS(4114),
    [anon_sym__Nullable] = ACTIONS(4114),
    [anon_sym__Nullable_result] = ACTIONS(4114),
    [anon_sym__Null_unspecified] = ACTIONS(4114),
    [anon_sym___autoreleasing] = ACTIONS(4114),
    [anon_sym___block] = ACTIONS(4114),
    [anon_sym___bridge] = ACTIONS(4114),
    [anon_sym___bridge_retained] = ACTIONS(4114),
    [anon_sym___bridge_transfer] = ACTIONS(4114),
    [anon_sym___complex] = ACTIONS(4114),
    [anon_sym___const] = ACTIONS(4114),
    [anon_sym___imag] = ACTIONS(4114),
    [anon_sym___kindof] = ACTIONS(4114),
    [anon_sym___nonnull] = ACTIONS(4114),
    [anon_sym___nullable] = ACTIONS(4114),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4114),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4114),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4114),
    [anon_sym___real] = ACTIONS(4114),
    [anon_sym___strong] = ACTIONS(4114),
    [anon_sym___unsafe_unretained] = ACTIONS(4114),
    [anon_sym___unused] = ACTIONS(4114),
    [anon_sym___weak] = ACTIONS(4114),
    [anon_sym_alignas] = ACTIONS(4114),
    [anon_sym__Alignas] = ACTIONS(4114),
    [sym_primitive_type] = ACTIONS(4114),
    [anon_sym_enum] = ACTIONS(4114),
    [anon_sym_COLON] = ACTIONS(4116),
    [anon_sym_struct] = ACTIONS(4114),
    [anon_sym_union] = ACTIONS(4114),
    [anon_sym_QMARK] = ACTIONS(4116),
    [anon_sym_STAR_EQ] = ACTIONS(4116),
    [anon_sym_SLASH_EQ] = ACTIONS(4116),
    [anon_sym_PERCENT_EQ] = ACTIONS(4116),
    [anon_sym_PLUS_EQ] = ACTIONS(4116),
    [anon_sym_DASH_EQ] = ACTIONS(4116),
    [anon_sym_LT_LT_EQ] = ACTIONS(4116),
    [anon_sym_GT_GT_EQ] = ACTIONS(4116),
    [anon_sym_AMP_EQ] = ACTIONS(4116),
    [anon_sym_CARET_EQ] = ACTIONS(4116),
    [anon_sym_PIPE_EQ] = ACTIONS(4116),
    [anon_sym_DASH_DASH] = ACTIONS(4116),
    [anon_sym_PLUS_PLUS] = ACTIONS(4116),
    [anon_sym_DOT] = ACTIONS(4114),
    [anon_sym_DASH_GT] = ACTIONS(4116),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4114),
    [anon_sym___typeof] = ACTIONS(4114),
    [anon_sym_typeof] = ACTIONS(4114),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4114),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4114),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4114),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4114),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4114),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4114),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4114),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4114),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4114),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4114),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4114),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4114),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4114),
    [anon_sym_API_AVAILABLE] = ACTIONS(4114),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4114),
    [anon_sym_API_DEPRECATED] = ACTIONS(4114),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4114),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4114),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4114),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4114),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4114),
    [anon_sym___deprecated_msg] = ACTIONS(4114),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4114),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4114),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4114),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4114),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4114),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4114),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4114),
    [anon_sym_BOOL] = ACTIONS(4114),
    [anon_sym_IMP] = ACTIONS(4114),
    [anon_sym_SEL] = ACTIONS(4114),
    [anon_sym_Class] = ACTIONS(4114),
    [anon_sym_id] = ACTIONS(4114),
  },
  [STATE(1468)] = {
    [sym_identifier] = ACTIONS(4118),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4120),
    [anon_sym_COMMA] = ACTIONS(4120),
    [anon_sym_LPAREN2] = ACTIONS(4120),
    [anon_sym_DASH] = ACTIONS(4118),
    [anon_sym_PLUS] = ACTIONS(4118),
    [anon_sym_STAR] = ACTIONS(4118),
    [anon_sym_SLASH] = ACTIONS(4118),
    [anon_sym_PERCENT] = ACTIONS(4118),
    [anon_sym_PIPE_PIPE] = ACTIONS(4120),
    [anon_sym_AMP_AMP] = ACTIONS(4120),
    [anon_sym_PIPE] = ACTIONS(4118),
    [anon_sym_CARET] = ACTIONS(4118),
    [anon_sym_AMP] = ACTIONS(4118),
    [anon_sym_EQ_EQ] = ACTIONS(4120),
    [anon_sym_BANG_EQ] = ACTIONS(4120),
    [anon_sym_GT] = ACTIONS(4118),
    [anon_sym_GT_EQ] = ACTIONS(4120),
    [anon_sym_LT_EQ] = ACTIONS(4120),
    [anon_sym_LT] = ACTIONS(4118),
    [anon_sym_LT_LT] = ACTIONS(4118),
    [anon_sym_GT_GT] = ACTIONS(4118),
    [anon_sym_SEMI] = ACTIONS(4120),
    [anon_sym___extension__] = ACTIONS(4118),
    [anon_sym_extern] = ACTIONS(4118),
    [anon_sym___attribute__] = ACTIONS(4118),
    [anon_sym___attribute] = ACTIONS(4118),
    [anon_sym_noreturn] = ACTIONS(4118),
    [anon_sym_LBRACK] = ACTIONS(4120),
    [anon_sym___declspec] = ACTIONS(4118),
    [anon_sym_LBRACE] = ACTIONS(4120),
    [anon_sym_signed] = ACTIONS(4118),
    [anon_sym_unsigned] = ACTIONS(4118),
    [anon_sym_long] = ACTIONS(4118),
    [anon_sym_short] = ACTIONS(4118),
    [anon_sym_static] = ACTIONS(4118),
    [anon_sym_EQ] = ACTIONS(4118),
    [anon_sym_ATautoreleasepool] = ACTIONS(4120),
    [anon_sym_auto] = ACTIONS(4118),
    [anon_sym_register] = ACTIONS(4118),
    [anon_sym_inline] = ACTIONS(4118),
    [anon_sym___inline] = ACTIONS(4118),
    [anon_sym___inline__] = ACTIONS(4118),
    [anon_sym___forceinline] = ACTIONS(4118),
    [anon_sym_thread_local] = ACTIONS(4118),
    [anon_sym___thread] = ACTIONS(4118),
    [anon_sym_CG_EXTERN] = ACTIONS(4118),
    [anon_sym_CG_INLINE] = ACTIONS(4118),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4118),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4118),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4118),
    [anon_sym_IBOutlet] = ACTIONS(4118),
    [anon_sym_IBInspectable] = ACTIONS(4118),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4118),
    [anon_sym_NS_INLINE] = ACTIONS(4118),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4118),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4118),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4118),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4118),
    [anon_sym_const] = ACTIONS(4118),
    [anon_sym_constexpr] = ACTIONS(4118),
    [anon_sym_volatile] = ACTIONS(4118),
    [anon_sym_restrict] = ACTIONS(4118),
    [anon_sym___restrict__] = ACTIONS(4118),
    [anon_sym__Atomic] = ACTIONS(4118),
    [anon_sym__Noreturn] = ACTIONS(4118),
    [anon_sym__Nonnull] = ACTIONS(4118),
    [anon_sym_nullable] = ACTIONS(4118),
    [anon_sym__Complex] = ACTIONS(4118),
    [anon_sym__Nullable] = ACTIONS(4118),
    [anon_sym__Nullable_result] = ACTIONS(4118),
    [anon_sym__Null_unspecified] = ACTIONS(4118),
    [anon_sym___autoreleasing] = ACTIONS(4118),
    [anon_sym___block] = ACTIONS(4118),
    [anon_sym___bridge] = ACTIONS(4118),
    [anon_sym___bridge_retained] = ACTIONS(4118),
    [anon_sym___bridge_transfer] = ACTIONS(4118),
    [anon_sym___complex] = ACTIONS(4118),
    [anon_sym___const] = ACTIONS(4118),
    [anon_sym___imag] = ACTIONS(4118),
    [anon_sym___kindof] = ACTIONS(4118),
    [anon_sym___nonnull] = ACTIONS(4118),
    [anon_sym___nullable] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4118),
    [anon_sym___real] = ACTIONS(4118),
    [anon_sym___strong] = ACTIONS(4118),
    [anon_sym___unsafe_unretained] = ACTIONS(4118),
    [anon_sym___unused] = ACTIONS(4118),
    [anon_sym___weak] = ACTIONS(4118),
    [anon_sym_alignas] = ACTIONS(4118),
    [anon_sym__Alignas] = ACTIONS(4118),
    [sym_primitive_type] = ACTIONS(4118),
    [anon_sym_enum] = ACTIONS(4118),
    [anon_sym_struct] = ACTIONS(4118),
    [anon_sym_union] = ACTIONS(4118),
    [anon_sym_QMARK] = ACTIONS(4120),
    [anon_sym_STAR_EQ] = ACTIONS(4120),
    [anon_sym_SLASH_EQ] = ACTIONS(4120),
    [anon_sym_PERCENT_EQ] = ACTIONS(4120),
    [anon_sym_PLUS_EQ] = ACTIONS(4120),
    [anon_sym_DASH_EQ] = ACTIONS(4120),
    [anon_sym_LT_LT_EQ] = ACTIONS(4120),
    [anon_sym_GT_GT_EQ] = ACTIONS(4120),
    [anon_sym_AMP_EQ] = ACTIONS(4120),
    [anon_sym_CARET_EQ] = ACTIONS(4120),
    [anon_sym_PIPE_EQ] = ACTIONS(4120),
    [anon_sym_DASH_DASH] = ACTIONS(4120),
    [anon_sym_PLUS_PLUS] = ACTIONS(4120),
    [anon_sym_DOT] = ACTIONS(4118),
    [anon_sym_DASH_GT] = ACTIONS(4120),
    [anon_sym_AT] = ACTIONS(4118),
    [anon_sym_DQUOTE] = ACTIONS(4120),
    [anon_sym_L_DQUOTE] = ACTIONS(4120),
    [anon_sym_u_DQUOTE] = ACTIONS(4120),
    [anon_sym_U_DQUOTE] = ACTIONS(4120),
    [anon_sym_u8_DQUOTE] = ACTIONS(4120),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4118),
    [anon_sym___typeof] = ACTIONS(4118),
    [anon_sym_typeof] = ACTIONS(4118),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4118),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4118),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4118),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4118),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4118),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4118),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4118),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_API_AVAILABLE] = ACTIONS(4118),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_API_DEPRECATED] = ACTIONS(4118),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4118),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym___deprecated_msg] = ACTIONS(4118),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4118),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4118),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4118),
    [anon_sym_BOOL] = ACTIONS(4118),
    [anon_sym_IMP] = ACTIONS(4118),
    [anon_sym_SEL] = ACTIONS(4118),
    [anon_sym_Class] = ACTIONS(4118),
    [anon_sym_id] = ACTIONS(4118),
  },
  [STATE(1469)] = {
    [sym_identifier] = ACTIONS(4122),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4124),
    [anon_sym_COMMA] = ACTIONS(4124),
    [anon_sym_RPAREN] = ACTIONS(4124),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4124),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4124),
    [anon_sym_LPAREN2] = ACTIONS(4124),
    [anon_sym_DASH] = ACTIONS(4122),
    [anon_sym_PLUS] = ACTIONS(4122),
    [anon_sym_STAR] = ACTIONS(4122),
    [anon_sym_SLASH] = ACTIONS(4122),
    [anon_sym_PERCENT] = ACTIONS(4122),
    [anon_sym_PIPE_PIPE] = ACTIONS(4124),
    [anon_sym_AMP_AMP] = ACTIONS(4124),
    [anon_sym_PIPE] = ACTIONS(4122),
    [anon_sym_CARET] = ACTIONS(4122),
    [anon_sym_AMP] = ACTIONS(4122),
    [anon_sym_EQ_EQ] = ACTIONS(4124),
    [anon_sym_BANG_EQ] = ACTIONS(4124),
    [anon_sym_GT] = ACTIONS(4122),
    [anon_sym_GT_EQ] = ACTIONS(4124),
    [anon_sym_LT_EQ] = ACTIONS(4124),
    [anon_sym_LT] = ACTIONS(4122),
    [anon_sym_LT_LT] = ACTIONS(4122),
    [anon_sym_GT_GT] = ACTIONS(4122),
    [anon_sym_SEMI] = ACTIONS(4124),
    [anon_sym___extension__] = ACTIONS(4122),
    [anon_sym_extern] = ACTIONS(4122),
    [anon_sym___attribute__] = ACTIONS(4122),
    [anon_sym___attribute] = ACTIONS(4122),
    [anon_sym_noreturn] = ACTIONS(4122),
    [anon_sym_LBRACK] = ACTIONS(4124),
    [anon_sym_RBRACK] = ACTIONS(4124),
    [anon_sym___declspec] = ACTIONS(4122),
    [anon_sym_LBRACE] = ACTIONS(4124),
    [anon_sym_RBRACE] = ACTIONS(4124),
    [anon_sym_signed] = ACTIONS(4122),
    [anon_sym_unsigned] = ACTIONS(4122),
    [anon_sym_long] = ACTIONS(4122),
    [anon_sym_short] = ACTIONS(4122),
    [anon_sym_static] = ACTIONS(4122),
    [anon_sym_EQ] = ACTIONS(4122),
    [anon_sym_ATautoreleasepool] = ACTIONS(4124),
    [anon_sym_auto] = ACTIONS(4122),
    [anon_sym_register] = ACTIONS(4122),
    [anon_sym_inline] = ACTIONS(4122),
    [anon_sym___inline] = ACTIONS(4122),
    [anon_sym___inline__] = ACTIONS(4122),
    [anon_sym___forceinline] = ACTIONS(4122),
    [anon_sym_thread_local] = ACTIONS(4122),
    [anon_sym___thread] = ACTIONS(4122),
    [anon_sym_CG_EXTERN] = ACTIONS(4122),
    [anon_sym_CG_INLINE] = ACTIONS(4122),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4122),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4122),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4122),
    [anon_sym_IBOutlet] = ACTIONS(4122),
    [anon_sym_IBInspectable] = ACTIONS(4122),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4122),
    [anon_sym_NS_INLINE] = ACTIONS(4122),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4122),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4122),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4122),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4122),
    [anon_sym_const] = ACTIONS(4122),
    [anon_sym_constexpr] = ACTIONS(4122),
    [anon_sym_volatile] = ACTIONS(4122),
    [anon_sym_restrict] = ACTIONS(4122),
    [anon_sym___restrict__] = ACTIONS(4122),
    [anon_sym__Atomic] = ACTIONS(4122),
    [anon_sym__Noreturn] = ACTIONS(4122),
    [anon_sym__Nonnull] = ACTIONS(4122),
    [anon_sym_nullable] = ACTIONS(4122),
    [anon_sym__Complex] = ACTIONS(4122),
    [anon_sym__Nullable] = ACTIONS(4122),
    [anon_sym__Nullable_result] = ACTIONS(4122),
    [anon_sym__Null_unspecified] = ACTIONS(4122),
    [anon_sym___autoreleasing] = ACTIONS(4122),
    [anon_sym___block] = ACTIONS(4122),
    [anon_sym___bridge] = ACTIONS(4122),
    [anon_sym___bridge_retained] = ACTIONS(4122),
    [anon_sym___bridge_transfer] = ACTIONS(4122),
    [anon_sym___complex] = ACTIONS(4122),
    [anon_sym___const] = ACTIONS(4122),
    [anon_sym___imag] = ACTIONS(4122),
    [anon_sym___kindof] = ACTIONS(4122),
    [anon_sym___nonnull] = ACTIONS(4122),
    [anon_sym___nullable] = ACTIONS(4122),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4122),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4122),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4122),
    [anon_sym___real] = ACTIONS(4122),
    [anon_sym___strong] = ACTIONS(4122),
    [anon_sym___unsafe_unretained] = ACTIONS(4122),
    [anon_sym___unused] = ACTIONS(4122),
    [anon_sym___weak] = ACTIONS(4122),
    [anon_sym_alignas] = ACTIONS(4122),
    [anon_sym__Alignas] = ACTIONS(4122),
    [sym_primitive_type] = ACTIONS(4122),
    [anon_sym_enum] = ACTIONS(4122),
    [anon_sym_COLON] = ACTIONS(4124),
    [anon_sym_struct] = ACTIONS(4122),
    [anon_sym_union] = ACTIONS(4122),
    [anon_sym_QMARK] = ACTIONS(4124),
    [anon_sym_STAR_EQ] = ACTIONS(4124),
    [anon_sym_SLASH_EQ] = ACTIONS(4124),
    [anon_sym_PERCENT_EQ] = ACTIONS(4124),
    [anon_sym_PLUS_EQ] = ACTIONS(4124),
    [anon_sym_DASH_EQ] = ACTIONS(4124),
    [anon_sym_LT_LT_EQ] = ACTIONS(4124),
    [anon_sym_GT_GT_EQ] = ACTIONS(4124),
    [anon_sym_AMP_EQ] = ACTIONS(4124),
    [anon_sym_CARET_EQ] = ACTIONS(4124),
    [anon_sym_PIPE_EQ] = ACTIONS(4124),
    [anon_sym_DASH_DASH] = ACTIONS(4124),
    [anon_sym_PLUS_PLUS] = ACTIONS(4124),
    [anon_sym_DOT] = ACTIONS(4122),
    [anon_sym_DASH_GT] = ACTIONS(4124),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4122),
    [anon_sym___typeof] = ACTIONS(4122),
    [anon_sym_typeof] = ACTIONS(4122),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4122),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4122),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4122),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4122),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4122),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4122),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4122),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4122),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4122),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4122),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4122),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4122),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4122),
    [anon_sym_API_AVAILABLE] = ACTIONS(4122),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4122),
    [anon_sym_API_DEPRECATED] = ACTIONS(4122),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4122),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4122),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4122),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4122),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4122),
    [anon_sym___deprecated_msg] = ACTIONS(4122),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4122),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4122),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4122),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4122),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4122),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4122),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4122),
    [anon_sym_BOOL] = ACTIONS(4122),
    [anon_sym_IMP] = ACTIONS(4122),
    [anon_sym_SEL] = ACTIONS(4122),
    [anon_sym_Class] = ACTIONS(4122),
    [anon_sym_id] = ACTIONS(4122),
  },
  [STATE(1470)] = {
    [sym_identifier] = ACTIONS(4126),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4128),
    [anon_sym_COMMA] = ACTIONS(4128),
    [anon_sym_RPAREN] = ACTIONS(4128),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4128),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4128),
    [anon_sym_LPAREN2] = ACTIONS(4128),
    [anon_sym_DASH] = ACTIONS(4126),
    [anon_sym_PLUS] = ACTIONS(4126),
    [anon_sym_STAR] = ACTIONS(4126),
    [anon_sym_SLASH] = ACTIONS(4126),
    [anon_sym_PERCENT] = ACTIONS(4126),
    [anon_sym_PIPE_PIPE] = ACTIONS(4128),
    [anon_sym_AMP_AMP] = ACTIONS(4128),
    [anon_sym_PIPE] = ACTIONS(4126),
    [anon_sym_CARET] = ACTIONS(4126),
    [anon_sym_AMP] = ACTIONS(4126),
    [anon_sym_EQ_EQ] = ACTIONS(4128),
    [anon_sym_BANG_EQ] = ACTIONS(4128),
    [anon_sym_GT] = ACTIONS(4126),
    [anon_sym_GT_EQ] = ACTIONS(4128),
    [anon_sym_LT_EQ] = ACTIONS(4128),
    [anon_sym_LT] = ACTIONS(4126),
    [anon_sym_LT_LT] = ACTIONS(4126),
    [anon_sym_GT_GT] = ACTIONS(4126),
    [anon_sym_SEMI] = ACTIONS(4128),
    [anon_sym___extension__] = ACTIONS(4126),
    [anon_sym_extern] = ACTIONS(4126),
    [anon_sym___attribute__] = ACTIONS(4126),
    [anon_sym___attribute] = ACTIONS(4126),
    [anon_sym_noreturn] = ACTIONS(4126),
    [anon_sym_LBRACK] = ACTIONS(4128),
    [anon_sym_RBRACK] = ACTIONS(4128),
    [anon_sym___declspec] = ACTIONS(4126),
    [anon_sym_LBRACE] = ACTIONS(4128),
    [anon_sym_RBRACE] = ACTIONS(4128),
    [anon_sym_signed] = ACTIONS(4126),
    [anon_sym_unsigned] = ACTIONS(4126),
    [anon_sym_long] = ACTIONS(4126),
    [anon_sym_short] = ACTIONS(4126),
    [anon_sym_static] = ACTIONS(4126),
    [anon_sym_EQ] = ACTIONS(4126),
    [anon_sym_ATautoreleasepool] = ACTIONS(4128),
    [anon_sym_auto] = ACTIONS(4126),
    [anon_sym_register] = ACTIONS(4126),
    [anon_sym_inline] = ACTIONS(4126),
    [anon_sym___inline] = ACTIONS(4126),
    [anon_sym___inline__] = ACTIONS(4126),
    [anon_sym___forceinline] = ACTIONS(4126),
    [anon_sym_thread_local] = ACTIONS(4126),
    [anon_sym___thread] = ACTIONS(4126),
    [anon_sym_CG_EXTERN] = ACTIONS(4126),
    [anon_sym_CG_INLINE] = ACTIONS(4126),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4126),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4126),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4126),
    [anon_sym_IBOutlet] = ACTIONS(4126),
    [anon_sym_IBInspectable] = ACTIONS(4126),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4126),
    [anon_sym_NS_INLINE] = ACTIONS(4126),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4126),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4126),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4126),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4126),
    [anon_sym_const] = ACTIONS(4126),
    [anon_sym_constexpr] = ACTIONS(4126),
    [anon_sym_volatile] = ACTIONS(4126),
    [anon_sym_restrict] = ACTIONS(4126),
    [anon_sym___restrict__] = ACTIONS(4126),
    [anon_sym__Atomic] = ACTIONS(4126),
    [anon_sym__Noreturn] = ACTIONS(4126),
    [anon_sym__Nonnull] = ACTIONS(4126),
    [anon_sym_nullable] = ACTIONS(4126),
    [anon_sym__Complex] = ACTIONS(4126),
    [anon_sym__Nullable] = ACTIONS(4126),
    [anon_sym__Nullable_result] = ACTIONS(4126),
    [anon_sym__Null_unspecified] = ACTIONS(4126),
    [anon_sym___autoreleasing] = ACTIONS(4126),
    [anon_sym___block] = ACTIONS(4126),
    [anon_sym___bridge] = ACTIONS(4126),
    [anon_sym___bridge_retained] = ACTIONS(4126),
    [anon_sym___bridge_transfer] = ACTIONS(4126),
    [anon_sym___complex] = ACTIONS(4126),
    [anon_sym___const] = ACTIONS(4126),
    [anon_sym___imag] = ACTIONS(4126),
    [anon_sym___kindof] = ACTIONS(4126),
    [anon_sym___nonnull] = ACTIONS(4126),
    [anon_sym___nullable] = ACTIONS(4126),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4126),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4126),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4126),
    [anon_sym___real] = ACTIONS(4126),
    [anon_sym___strong] = ACTIONS(4126),
    [anon_sym___unsafe_unretained] = ACTIONS(4126),
    [anon_sym___unused] = ACTIONS(4126),
    [anon_sym___weak] = ACTIONS(4126),
    [anon_sym_alignas] = ACTIONS(4126),
    [anon_sym__Alignas] = ACTIONS(4126),
    [sym_primitive_type] = ACTIONS(4126),
    [anon_sym_enum] = ACTIONS(4126),
    [anon_sym_COLON] = ACTIONS(4128),
    [anon_sym_struct] = ACTIONS(4126),
    [anon_sym_union] = ACTIONS(4126),
    [anon_sym_QMARK] = ACTIONS(4128),
    [anon_sym_STAR_EQ] = ACTIONS(4128),
    [anon_sym_SLASH_EQ] = ACTIONS(4128),
    [anon_sym_PERCENT_EQ] = ACTIONS(4128),
    [anon_sym_PLUS_EQ] = ACTIONS(4128),
    [anon_sym_DASH_EQ] = ACTIONS(4128),
    [anon_sym_LT_LT_EQ] = ACTIONS(4128),
    [anon_sym_GT_GT_EQ] = ACTIONS(4128),
    [anon_sym_AMP_EQ] = ACTIONS(4128),
    [anon_sym_CARET_EQ] = ACTIONS(4128),
    [anon_sym_PIPE_EQ] = ACTIONS(4128),
    [anon_sym_DASH_DASH] = ACTIONS(4128),
    [anon_sym_PLUS_PLUS] = ACTIONS(4128),
    [anon_sym_DOT] = ACTIONS(4126),
    [anon_sym_DASH_GT] = ACTIONS(4128),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4126),
    [anon_sym___typeof] = ACTIONS(4126),
    [anon_sym_typeof] = ACTIONS(4126),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4126),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4126),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4126),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4126),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4126),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4126),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4126),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4126),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4126),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4126),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4126),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4126),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4126),
    [anon_sym_API_AVAILABLE] = ACTIONS(4126),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4126),
    [anon_sym_API_DEPRECATED] = ACTIONS(4126),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4126),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4126),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4126),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4126),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4126),
    [anon_sym___deprecated_msg] = ACTIONS(4126),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4126),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4126),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4126),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4126),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4126),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4126),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4126),
    [anon_sym_BOOL] = ACTIONS(4126),
    [anon_sym_IMP] = ACTIONS(4126),
    [anon_sym_SEL] = ACTIONS(4126),
    [anon_sym_Class] = ACTIONS(4126),
    [anon_sym_id] = ACTIONS(4126),
  },
  [STATE(1471)] = {
    [sym_identifier] = ACTIONS(4130),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4132),
    [anon_sym_COMMA] = ACTIONS(4132),
    [anon_sym_RPAREN] = ACTIONS(4132),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4132),
    [anon_sym_LPAREN2] = ACTIONS(4132),
    [anon_sym_DASH] = ACTIONS(4130),
    [anon_sym_PLUS] = ACTIONS(4130),
    [anon_sym_STAR] = ACTIONS(4130),
    [anon_sym_SLASH] = ACTIONS(4130),
    [anon_sym_PERCENT] = ACTIONS(4130),
    [anon_sym_PIPE_PIPE] = ACTIONS(4132),
    [anon_sym_AMP_AMP] = ACTIONS(4132),
    [anon_sym_PIPE] = ACTIONS(4130),
    [anon_sym_CARET] = ACTIONS(4130),
    [anon_sym_AMP] = ACTIONS(4130),
    [anon_sym_EQ_EQ] = ACTIONS(4132),
    [anon_sym_BANG_EQ] = ACTIONS(4132),
    [anon_sym_GT] = ACTIONS(4130),
    [anon_sym_GT_EQ] = ACTIONS(4132),
    [anon_sym_LT_EQ] = ACTIONS(4132),
    [anon_sym_LT] = ACTIONS(4130),
    [anon_sym_LT_LT] = ACTIONS(4130),
    [anon_sym_GT_GT] = ACTIONS(4130),
    [anon_sym_SEMI] = ACTIONS(4132),
    [anon_sym___extension__] = ACTIONS(4130),
    [anon_sym_extern] = ACTIONS(4130),
    [anon_sym___attribute__] = ACTIONS(4130),
    [anon_sym___attribute] = ACTIONS(4130),
    [anon_sym_noreturn] = ACTIONS(4130),
    [anon_sym_LBRACK] = ACTIONS(4132),
    [anon_sym_RBRACK] = ACTIONS(4132),
    [anon_sym___declspec] = ACTIONS(4130),
    [anon_sym_LBRACE] = ACTIONS(4132),
    [anon_sym_RBRACE] = ACTIONS(4132),
    [anon_sym_signed] = ACTIONS(4130),
    [anon_sym_unsigned] = ACTIONS(4130),
    [anon_sym_long] = ACTIONS(4130),
    [anon_sym_short] = ACTIONS(4130),
    [anon_sym_static] = ACTIONS(4130),
    [anon_sym_EQ] = ACTIONS(4130),
    [anon_sym_ATautoreleasepool] = ACTIONS(4132),
    [anon_sym_auto] = ACTIONS(4130),
    [anon_sym_register] = ACTIONS(4130),
    [anon_sym_inline] = ACTIONS(4130),
    [anon_sym___inline] = ACTIONS(4130),
    [anon_sym___inline__] = ACTIONS(4130),
    [anon_sym___forceinline] = ACTIONS(4130),
    [anon_sym_thread_local] = ACTIONS(4130),
    [anon_sym___thread] = ACTIONS(4130),
    [anon_sym_CG_EXTERN] = ACTIONS(4130),
    [anon_sym_CG_INLINE] = ACTIONS(4130),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4130),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4130),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4130),
    [anon_sym_IBOutlet] = ACTIONS(4130),
    [anon_sym_IBInspectable] = ACTIONS(4130),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4130),
    [anon_sym_NS_INLINE] = ACTIONS(4130),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4130),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4130),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4130),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4130),
    [anon_sym_const] = ACTIONS(4130),
    [anon_sym_constexpr] = ACTIONS(4130),
    [anon_sym_volatile] = ACTIONS(4130),
    [anon_sym_restrict] = ACTIONS(4130),
    [anon_sym___restrict__] = ACTIONS(4130),
    [anon_sym__Atomic] = ACTIONS(4130),
    [anon_sym__Noreturn] = ACTIONS(4130),
    [anon_sym__Nonnull] = ACTIONS(4130),
    [anon_sym_nullable] = ACTIONS(4130),
    [anon_sym__Complex] = ACTIONS(4130),
    [anon_sym__Nullable] = ACTIONS(4130),
    [anon_sym__Nullable_result] = ACTIONS(4130),
    [anon_sym__Null_unspecified] = ACTIONS(4130),
    [anon_sym___autoreleasing] = ACTIONS(4130),
    [anon_sym___block] = ACTIONS(4130),
    [anon_sym___bridge] = ACTIONS(4130),
    [anon_sym___bridge_retained] = ACTIONS(4130),
    [anon_sym___bridge_transfer] = ACTIONS(4130),
    [anon_sym___complex] = ACTIONS(4130),
    [anon_sym___const] = ACTIONS(4130),
    [anon_sym___imag] = ACTIONS(4130),
    [anon_sym___kindof] = ACTIONS(4130),
    [anon_sym___nonnull] = ACTIONS(4130),
    [anon_sym___nullable] = ACTIONS(4130),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4130),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4130),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4130),
    [anon_sym___real] = ACTIONS(4130),
    [anon_sym___strong] = ACTIONS(4130),
    [anon_sym___unsafe_unretained] = ACTIONS(4130),
    [anon_sym___unused] = ACTIONS(4130),
    [anon_sym___weak] = ACTIONS(4130),
    [anon_sym_alignas] = ACTIONS(4130),
    [anon_sym__Alignas] = ACTIONS(4130),
    [sym_primitive_type] = ACTIONS(4130),
    [anon_sym_enum] = ACTIONS(4130),
    [anon_sym_COLON] = ACTIONS(4132),
    [anon_sym_struct] = ACTIONS(4130),
    [anon_sym_union] = ACTIONS(4130),
    [anon_sym_QMARK] = ACTIONS(4132),
    [anon_sym_STAR_EQ] = ACTIONS(4132),
    [anon_sym_SLASH_EQ] = ACTIONS(4132),
    [anon_sym_PERCENT_EQ] = ACTIONS(4132),
    [anon_sym_PLUS_EQ] = ACTIONS(4132),
    [anon_sym_DASH_EQ] = ACTIONS(4132),
    [anon_sym_LT_LT_EQ] = ACTIONS(4132),
    [anon_sym_GT_GT_EQ] = ACTIONS(4132),
    [anon_sym_AMP_EQ] = ACTIONS(4132),
    [anon_sym_CARET_EQ] = ACTIONS(4132),
    [anon_sym_PIPE_EQ] = ACTIONS(4132),
    [anon_sym_DASH_DASH] = ACTIONS(4132),
    [anon_sym_PLUS_PLUS] = ACTIONS(4132),
    [anon_sym_DOT] = ACTIONS(4130),
    [anon_sym_DASH_GT] = ACTIONS(4132),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4130),
    [anon_sym___typeof] = ACTIONS(4130),
    [anon_sym_typeof] = ACTIONS(4130),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4130),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4130),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4130),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4130),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4130),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4130),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4130),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4130),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4130),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4130),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4130),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4130),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4130),
    [anon_sym_API_AVAILABLE] = ACTIONS(4130),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4130),
    [anon_sym_API_DEPRECATED] = ACTIONS(4130),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4130),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4130),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4130),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4130),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4130),
    [anon_sym___deprecated_msg] = ACTIONS(4130),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4130),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4130),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4130),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4130),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4130),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4130),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4130),
    [anon_sym_BOOL] = ACTIONS(4130),
    [anon_sym_IMP] = ACTIONS(4130),
    [anon_sym_SEL] = ACTIONS(4130),
    [anon_sym_Class] = ACTIONS(4130),
    [anon_sym_id] = ACTIONS(4130),
  },
  [STATE(1472)] = {
    [sym_identifier] = ACTIONS(4134),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4136),
    [anon_sym_COMMA] = ACTIONS(4136),
    [anon_sym_RPAREN] = ACTIONS(4136),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4136),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4136),
    [anon_sym_LPAREN2] = ACTIONS(4136),
    [anon_sym_DASH] = ACTIONS(4134),
    [anon_sym_PLUS] = ACTIONS(4134),
    [anon_sym_STAR] = ACTIONS(4134),
    [anon_sym_SLASH] = ACTIONS(4134),
    [anon_sym_PERCENT] = ACTIONS(4134),
    [anon_sym_PIPE_PIPE] = ACTIONS(4136),
    [anon_sym_AMP_AMP] = ACTIONS(4136),
    [anon_sym_PIPE] = ACTIONS(4134),
    [anon_sym_CARET] = ACTIONS(4134),
    [anon_sym_AMP] = ACTIONS(4134),
    [anon_sym_EQ_EQ] = ACTIONS(4136),
    [anon_sym_BANG_EQ] = ACTIONS(4136),
    [anon_sym_GT] = ACTIONS(4134),
    [anon_sym_GT_EQ] = ACTIONS(4136),
    [anon_sym_LT_EQ] = ACTIONS(4136),
    [anon_sym_LT] = ACTIONS(4134),
    [anon_sym_LT_LT] = ACTIONS(4134),
    [anon_sym_GT_GT] = ACTIONS(4134),
    [anon_sym_SEMI] = ACTIONS(4136),
    [anon_sym___extension__] = ACTIONS(4134),
    [anon_sym_extern] = ACTIONS(4134),
    [anon_sym___attribute__] = ACTIONS(4134),
    [anon_sym___attribute] = ACTIONS(4134),
    [anon_sym_noreturn] = ACTIONS(4134),
    [anon_sym_LBRACK] = ACTIONS(4136),
    [anon_sym_RBRACK] = ACTIONS(4136),
    [anon_sym___declspec] = ACTIONS(4134),
    [anon_sym_LBRACE] = ACTIONS(4136),
    [anon_sym_RBRACE] = ACTIONS(4136),
    [anon_sym_signed] = ACTIONS(4134),
    [anon_sym_unsigned] = ACTIONS(4134),
    [anon_sym_long] = ACTIONS(4134),
    [anon_sym_short] = ACTIONS(4134),
    [anon_sym_static] = ACTIONS(4134),
    [anon_sym_EQ] = ACTIONS(4134),
    [anon_sym_ATautoreleasepool] = ACTIONS(4136),
    [anon_sym_auto] = ACTIONS(4134),
    [anon_sym_register] = ACTIONS(4134),
    [anon_sym_inline] = ACTIONS(4134),
    [anon_sym___inline] = ACTIONS(4134),
    [anon_sym___inline__] = ACTIONS(4134),
    [anon_sym___forceinline] = ACTIONS(4134),
    [anon_sym_thread_local] = ACTIONS(4134),
    [anon_sym___thread] = ACTIONS(4134),
    [anon_sym_CG_EXTERN] = ACTIONS(4134),
    [anon_sym_CG_INLINE] = ACTIONS(4134),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4134),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4134),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4134),
    [anon_sym_IBOutlet] = ACTIONS(4134),
    [anon_sym_IBInspectable] = ACTIONS(4134),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4134),
    [anon_sym_NS_INLINE] = ACTIONS(4134),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4134),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4134),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4134),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4134),
    [anon_sym_const] = ACTIONS(4134),
    [anon_sym_constexpr] = ACTIONS(4134),
    [anon_sym_volatile] = ACTIONS(4134),
    [anon_sym_restrict] = ACTIONS(4134),
    [anon_sym___restrict__] = ACTIONS(4134),
    [anon_sym__Atomic] = ACTIONS(4134),
    [anon_sym__Noreturn] = ACTIONS(4134),
    [anon_sym__Nonnull] = ACTIONS(4134),
    [anon_sym_nullable] = ACTIONS(4134),
    [anon_sym__Complex] = ACTIONS(4134),
    [anon_sym__Nullable] = ACTIONS(4134),
    [anon_sym__Nullable_result] = ACTIONS(4134),
    [anon_sym__Null_unspecified] = ACTIONS(4134),
    [anon_sym___autoreleasing] = ACTIONS(4134),
    [anon_sym___block] = ACTIONS(4134),
    [anon_sym___bridge] = ACTIONS(4134),
    [anon_sym___bridge_retained] = ACTIONS(4134),
    [anon_sym___bridge_transfer] = ACTIONS(4134),
    [anon_sym___complex] = ACTIONS(4134),
    [anon_sym___const] = ACTIONS(4134),
    [anon_sym___imag] = ACTIONS(4134),
    [anon_sym___kindof] = ACTIONS(4134),
    [anon_sym___nonnull] = ACTIONS(4134),
    [anon_sym___nullable] = ACTIONS(4134),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4134),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4134),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4134),
    [anon_sym___real] = ACTIONS(4134),
    [anon_sym___strong] = ACTIONS(4134),
    [anon_sym___unsafe_unretained] = ACTIONS(4134),
    [anon_sym___unused] = ACTIONS(4134),
    [anon_sym___weak] = ACTIONS(4134),
    [anon_sym_alignas] = ACTIONS(4134),
    [anon_sym__Alignas] = ACTIONS(4134),
    [sym_primitive_type] = ACTIONS(4134),
    [anon_sym_enum] = ACTIONS(4134),
    [anon_sym_COLON] = ACTIONS(4136),
    [anon_sym_struct] = ACTIONS(4134),
    [anon_sym_union] = ACTIONS(4134),
    [anon_sym_QMARK] = ACTIONS(4136),
    [anon_sym_STAR_EQ] = ACTIONS(4136),
    [anon_sym_SLASH_EQ] = ACTIONS(4136),
    [anon_sym_PERCENT_EQ] = ACTIONS(4136),
    [anon_sym_PLUS_EQ] = ACTIONS(4136),
    [anon_sym_DASH_EQ] = ACTIONS(4136),
    [anon_sym_LT_LT_EQ] = ACTIONS(4136),
    [anon_sym_GT_GT_EQ] = ACTIONS(4136),
    [anon_sym_AMP_EQ] = ACTIONS(4136),
    [anon_sym_CARET_EQ] = ACTIONS(4136),
    [anon_sym_PIPE_EQ] = ACTIONS(4136),
    [anon_sym_DASH_DASH] = ACTIONS(4136),
    [anon_sym_PLUS_PLUS] = ACTIONS(4136),
    [anon_sym_DOT] = ACTIONS(4134),
    [anon_sym_DASH_GT] = ACTIONS(4136),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4134),
    [anon_sym___typeof] = ACTIONS(4134),
    [anon_sym_typeof] = ACTIONS(4134),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4134),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4134),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4134),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4134),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4134),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4134),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4134),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4134),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4134),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4134),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4134),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4134),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4134),
    [anon_sym_API_AVAILABLE] = ACTIONS(4134),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4134),
    [anon_sym_API_DEPRECATED] = ACTIONS(4134),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4134),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4134),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4134),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4134),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4134),
    [anon_sym___deprecated_msg] = ACTIONS(4134),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4134),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4134),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4134),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4134),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4134),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4134),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4134),
    [anon_sym_BOOL] = ACTIONS(4134),
    [anon_sym_IMP] = ACTIONS(4134),
    [anon_sym_SEL] = ACTIONS(4134),
    [anon_sym_Class] = ACTIONS(4134),
    [anon_sym_id] = ACTIONS(4134),
  },
  [STATE(1473)] = {
    [sym_identifier] = ACTIONS(4138),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4140),
    [anon_sym_COMMA] = ACTIONS(4140),
    [anon_sym_RPAREN] = ACTIONS(4140),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4140),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4140),
    [anon_sym_LPAREN2] = ACTIONS(4140),
    [anon_sym_DASH] = ACTIONS(4138),
    [anon_sym_PLUS] = ACTIONS(4138),
    [anon_sym_STAR] = ACTIONS(4138),
    [anon_sym_SLASH] = ACTIONS(4138),
    [anon_sym_PERCENT] = ACTIONS(4138),
    [anon_sym_PIPE_PIPE] = ACTIONS(4140),
    [anon_sym_AMP_AMP] = ACTIONS(4140),
    [anon_sym_PIPE] = ACTIONS(4138),
    [anon_sym_CARET] = ACTIONS(4138),
    [anon_sym_AMP] = ACTIONS(4138),
    [anon_sym_EQ_EQ] = ACTIONS(4140),
    [anon_sym_BANG_EQ] = ACTIONS(4140),
    [anon_sym_GT] = ACTIONS(4138),
    [anon_sym_GT_EQ] = ACTIONS(4140),
    [anon_sym_LT_EQ] = ACTIONS(4140),
    [anon_sym_LT] = ACTIONS(4138),
    [anon_sym_LT_LT] = ACTIONS(4138),
    [anon_sym_GT_GT] = ACTIONS(4138),
    [anon_sym_SEMI] = ACTIONS(4140),
    [anon_sym___extension__] = ACTIONS(4138),
    [anon_sym_extern] = ACTIONS(4138),
    [anon_sym___attribute__] = ACTIONS(4138),
    [anon_sym___attribute] = ACTIONS(4138),
    [anon_sym_noreturn] = ACTIONS(4138),
    [anon_sym_LBRACK] = ACTIONS(4140),
    [anon_sym_RBRACK] = ACTIONS(4140),
    [anon_sym___declspec] = ACTIONS(4138),
    [anon_sym_LBRACE] = ACTIONS(4140),
    [anon_sym_RBRACE] = ACTIONS(4140),
    [anon_sym_signed] = ACTIONS(4138),
    [anon_sym_unsigned] = ACTIONS(4138),
    [anon_sym_long] = ACTIONS(4138),
    [anon_sym_short] = ACTIONS(4138),
    [anon_sym_static] = ACTIONS(4138),
    [anon_sym_EQ] = ACTIONS(4138),
    [anon_sym_ATautoreleasepool] = ACTIONS(4140),
    [anon_sym_auto] = ACTIONS(4138),
    [anon_sym_register] = ACTIONS(4138),
    [anon_sym_inline] = ACTIONS(4138),
    [anon_sym___inline] = ACTIONS(4138),
    [anon_sym___inline__] = ACTIONS(4138),
    [anon_sym___forceinline] = ACTIONS(4138),
    [anon_sym_thread_local] = ACTIONS(4138),
    [anon_sym___thread] = ACTIONS(4138),
    [anon_sym_CG_EXTERN] = ACTIONS(4138),
    [anon_sym_CG_INLINE] = ACTIONS(4138),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4138),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4138),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4138),
    [anon_sym_IBOutlet] = ACTIONS(4138),
    [anon_sym_IBInspectable] = ACTIONS(4138),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4138),
    [anon_sym_NS_INLINE] = ACTIONS(4138),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4138),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4138),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4138),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4138),
    [anon_sym_const] = ACTIONS(4138),
    [anon_sym_constexpr] = ACTIONS(4138),
    [anon_sym_volatile] = ACTIONS(4138),
    [anon_sym_restrict] = ACTIONS(4138),
    [anon_sym___restrict__] = ACTIONS(4138),
    [anon_sym__Atomic] = ACTIONS(4138),
    [anon_sym__Noreturn] = ACTIONS(4138),
    [anon_sym__Nonnull] = ACTIONS(4138),
    [anon_sym_nullable] = ACTIONS(4138),
    [anon_sym__Complex] = ACTIONS(4138),
    [anon_sym__Nullable] = ACTIONS(4138),
    [anon_sym__Nullable_result] = ACTIONS(4138),
    [anon_sym__Null_unspecified] = ACTIONS(4138),
    [anon_sym___autoreleasing] = ACTIONS(4138),
    [anon_sym___block] = ACTIONS(4138),
    [anon_sym___bridge] = ACTIONS(4138),
    [anon_sym___bridge_retained] = ACTIONS(4138),
    [anon_sym___bridge_transfer] = ACTIONS(4138),
    [anon_sym___complex] = ACTIONS(4138),
    [anon_sym___const] = ACTIONS(4138),
    [anon_sym___imag] = ACTIONS(4138),
    [anon_sym___kindof] = ACTIONS(4138),
    [anon_sym___nonnull] = ACTIONS(4138),
    [anon_sym___nullable] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4138),
    [anon_sym___real] = ACTIONS(4138),
    [anon_sym___strong] = ACTIONS(4138),
    [anon_sym___unsafe_unretained] = ACTIONS(4138),
    [anon_sym___unused] = ACTIONS(4138),
    [anon_sym___weak] = ACTIONS(4138),
    [anon_sym_alignas] = ACTIONS(4138),
    [anon_sym__Alignas] = ACTIONS(4138),
    [sym_primitive_type] = ACTIONS(4138),
    [anon_sym_enum] = ACTIONS(4138),
    [anon_sym_COLON] = ACTIONS(4140),
    [anon_sym_struct] = ACTIONS(4138),
    [anon_sym_union] = ACTIONS(4138),
    [anon_sym_QMARK] = ACTIONS(4140),
    [anon_sym_STAR_EQ] = ACTIONS(4140),
    [anon_sym_SLASH_EQ] = ACTIONS(4140),
    [anon_sym_PERCENT_EQ] = ACTIONS(4140),
    [anon_sym_PLUS_EQ] = ACTIONS(4140),
    [anon_sym_DASH_EQ] = ACTIONS(4140),
    [anon_sym_LT_LT_EQ] = ACTIONS(4140),
    [anon_sym_GT_GT_EQ] = ACTIONS(4140),
    [anon_sym_AMP_EQ] = ACTIONS(4140),
    [anon_sym_CARET_EQ] = ACTIONS(4140),
    [anon_sym_PIPE_EQ] = ACTIONS(4140),
    [anon_sym_DASH_DASH] = ACTIONS(4140),
    [anon_sym_PLUS_PLUS] = ACTIONS(4140),
    [anon_sym_DOT] = ACTIONS(4138),
    [anon_sym_DASH_GT] = ACTIONS(4140),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4138),
    [anon_sym___typeof] = ACTIONS(4138),
    [anon_sym_typeof] = ACTIONS(4138),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4138),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4138),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4138),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4138),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4138),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4138),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4138),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_API_AVAILABLE] = ACTIONS(4138),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_API_DEPRECATED] = ACTIONS(4138),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4138),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym___deprecated_msg] = ACTIONS(4138),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4138),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4138),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4138),
    [anon_sym_BOOL] = ACTIONS(4138),
    [anon_sym_IMP] = ACTIONS(4138),
    [anon_sym_SEL] = ACTIONS(4138),
    [anon_sym_Class] = ACTIONS(4138),
    [anon_sym_id] = ACTIONS(4138),
  },
  [STATE(1474)] = {
    [sym_identifier] = ACTIONS(4138),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4140),
    [anon_sym_COMMA] = ACTIONS(4140),
    [anon_sym_RPAREN] = ACTIONS(4140),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4140),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4140),
    [anon_sym_LPAREN2] = ACTIONS(4140),
    [anon_sym_DASH] = ACTIONS(4138),
    [anon_sym_PLUS] = ACTIONS(4138),
    [anon_sym_STAR] = ACTIONS(4138),
    [anon_sym_SLASH] = ACTIONS(4138),
    [anon_sym_PERCENT] = ACTIONS(4138),
    [anon_sym_PIPE_PIPE] = ACTIONS(4140),
    [anon_sym_AMP_AMP] = ACTIONS(4140),
    [anon_sym_PIPE] = ACTIONS(4138),
    [anon_sym_CARET] = ACTIONS(4138),
    [anon_sym_AMP] = ACTIONS(4138),
    [anon_sym_EQ_EQ] = ACTIONS(4140),
    [anon_sym_BANG_EQ] = ACTIONS(4140),
    [anon_sym_GT] = ACTIONS(4138),
    [anon_sym_GT_EQ] = ACTIONS(4140),
    [anon_sym_LT_EQ] = ACTIONS(4140),
    [anon_sym_LT] = ACTIONS(4138),
    [anon_sym_LT_LT] = ACTIONS(4138),
    [anon_sym_GT_GT] = ACTIONS(4138),
    [anon_sym_SEMI] = ACTIONS(4140),
    [anon_sym___extension__] = ACTIONS(4138),
    [anon_sym_extern] = ACTIONS(4138),
    [anon_sym___attribute__] = ACTIONS(4138),
    [anon_sym___attribute] = ACTIONS(4138),
    [anon_sym_noreturn] = ACTIONS(4138),
    [anon_sym_LBRACK] = ACTIONS(4140),
    [anon_sym_RBRACK] = ACTIONS(4140),
    [anon_sym___declspec] = ACTIONS(4138),
    [anon_sym_LBRACE] = ACTIONS(4140),
    [anon_sym_RBRACE] = ACTIONS(4140),
    [anon_sym_signed] = ACTIONS(4138),
    [anon_sym_unsigned] = ACTIONS(4138),
    [anon_sym_long] = ACTIONS(4138),
    [anon_sym_short] = ACTIONS(4138),
    [anon_sym_static] = ACTIONS(4138),
    [anon_sym_EQ] = ACTIONS(4138),
    [anon_sym_ATautoreleasepool] = ACTIONS(4140),
    [anon_sym_auto] = ACTIONS(4138),
    [anon_sym_register] = ACTIONS(4138),
    [anon_sym_inline] = ACTIONS(4138),
    [anon_sym___inline] = ACTIONS(4138),
    [anon_sym___inline__] = ACTIONS(4138),
    [anon_sym___forceinline] = ACTIONS(4138),
    [anon_sym_thread_local] = ACTIONS(4138),
    [anon_sym___thread] = ACTIONS(4138),
    [anon_sym_CG_EXTERN] = ACTIONS(4138),
    [anon_sym_CG_INLINE] = ACTIONS(4138),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4138),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4138),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4138),
    [anon_sym_IBOutlet] = ACTIONS(4138),
    [anon_sym_IBInspectable] = ACTIONS(4138),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4138),
    [anon_sym_NS_INLINE] = ACTIONS(4138),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4138),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4138),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4138),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4138),
    [anon_sym_const] = ACTIONS(4138),
    [anon_sym_constexpr] = ACTIONS(4138),
    [anon_sym_volatile] = ACTIONS(4138),
    [anon_sym_restrict] = ACTIONS(4138),
    [anon_sym___restrict__] = ACTIONS(4138),
    [anon_sym__Atomic] = ACTIONS(4138),
    [anon_sym__Noreturn] = ACTIONS(4138),
    [anon_sym__Nonnull] = ACTIONS(4138),
    [anon_sym_nullable] = ACTIONS(4138),
    [anon_sym__Complex] = ACTIONS(4138),
    [anon_sym__Nullable] = ACTIONS(4138),
    [anon_sym__Nullable_result] = ACTIONS(4138),
    [anon_sym__Null_unspecified] = ACTIONS(4138),
    [anon_sym___autoreleasing] = ACTIONS(4138),
    [anon_sym___block] = ACTIONS(4138),
    [anon_sym___bridge] = ACTIONS(4138),
    [anon_sym___bridge_retained] = ACTIONS(4138),
    [anon_sym___bridge_transfer] = ACTIONS(4138),
    [anon_sym___complex] = ACTIONS(4138),
    [anon_sym___const] = ACTIONS(4138),
    [anon_sym___imag] = ACTIONS(4138),
    [anon_sym___kindof] = ACTIONS(4138),
    [anon_sym___nonnull] = ACTIONS(4138),
    [anon_sym___nullable] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4138),
    [anon_sym___real] = ACTIONS(4138),
    [anon_sym___strong] = ACTIONS(4138),
    [anon_sym___unsafe_unretained] = ACTIONS(4138),
    [anon_sym___unused] = ACTIONS(4138),
    [anon_sym___weak] = ACTIONS(4138),
    [anon_sym_alignas] = ACTIONS(4138),
    [anon_sym__Alignas] = ACTIONS(4138),
    [sym_primitive_type] = ACTIONS(4138),
    [anon_sym_enum] = ACTIONS(4138),
    [anon_sym_COLON] = ACTIONS(4140),
    [anon_sym_struct] = ACTIONS(4138),
    [anon_sym_union] = ACTIONS(4138),
    [anon_sym_QMARK] = ACTIONS(4140),
    [anon_sym_STAR_EQ] = ACTIONS(4140),
    [anon_sym_SLASH_EQ] = ACTIONS(4140),
    [anon_sym_PERCENT_EQ] = ACTIONS(4140),
    [anon_sym_PLUS_EQ] = ACTIONS(4140),
    [anon_sym_DASH_EQ] = ACTIONS(4140),
    [anon_sym_LT_LT_EQ] = ACTIONS(4140),
    [anon_sym_GT_GT_EQ] = ACTIONS(4140),
    [anon_sym_AMP_EQ] = ACTIONS(4140),
    [anon_sym_CARET_EQ] = ACTIONS(4140),
    [anon_sym_PIPE_EQ] = ACTIONS(4140),
    [anon_sym_DASH_DASH] = ACTIONS(4140),
    [anon_sym_PLUS_PLUS] = ACTIONS(4140),
    [anon_sym_DOT] = ACTIONS(4138),
    [anon_sym_DASH_GT] = ACTIONS(4140),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4138),
    [anon_sym___typeof] = ACTIONS(4138),
    [anon_sym_typeof] = ACTIONS(4138),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4138),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4138),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4138),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4138),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4138),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4138),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4138),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_API_AVAILABLE] = ACTIONS(4138),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_API_DEPRECATED] = ACTIONS(4138),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4138),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym___deprecated_msg] = ACTIONS(4138),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4138),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4138),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4138),
    [anon_sym_BOOL] = ACTIONS(4138),
    [anon_sym_IMP] = ACTIONS(4138),
    [anon_sym_SEL] = ACTIONS(4138),
    [anon_sym_Class] = ACTIONS(4138),
    [anon_sym_id] = ACTIONS(4138),
  },
  [STATE(1475)] = {
    [sym_identifier] = ACTIONS(4142),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4144),
    [anon_sym_COMMA] = ACTIONS(4144),
    [anon_sym_RPAREN] = ACTIONS(4144),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4144),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4144),
    [anon_sym_LPAREN2] = ACTIONS(4144),
    [anon_sym_DASH] = ACTIONS(4142),
    [anon_sym_PLUS] = ACTIONS(4142),
    [anon_sym_STAR] = ACTIONS(4142),
    [anon_sym_SLASH] = ACTIONS(4142),
    [anon_sym_PERCENT] = ACTIONS(4142),
    [anon_sym_PIPE_PIPE] = ACTIONS(4144),
    [anon_sym_AMP_AMP] = ACTIONS(4144),
    [anon_sym_PIPE] = ACTIONS(4142),
    [anon_sym_CARET] = ACTIONS(4142),
    [anon_sym_AMP] = ACTIONS(4142),
    [anon_sym_EQ_EQ] = ACTIONS(4144),
    [anon_sym_BANG_EQ] = ACTIONS(4144),
    [anon_sym_GT] = ACTIONS(4142),
    [anon_sym_GT_EQ] = ACTIONS(4144),
    [anon_sym_LT_EQ] = ACTIONS(4144),
    [anon_sym_LT] = ACTIONS(4142),
    [anon_sym_LT_LT] = ACTIONS(4142),
    [anon_sym_GT_GT] = ACTIONS(4142),
    [anon_sym_SEMI] = ACTIONS(4144),
    [anon_sym___extension__] = ACTIONS(4142),
    [anon_sym_extern] = ACTIONS(4142),
    [anon_sym___attribute__] = ACTIONS(4142),
    [anon_sym___attribute] = ACTIONS(4142),
    [anon_sym_noreturn] = ACTIONS(4142),
    [anon_sym_LBRACK] = ACTIONS(4144),
    [anon_sym_RBRACK] = ACTIONS(4144),
    [anon_sym___declspec] = ACTIONS(4142),
    [anon_sym_LBRACE] = ACTIONS(4144),
    [anon_sym_RBRACE] = ACTIONS(4144),
    [anon_sym_signed] = ACTIONS(4142),
    [anon_sym_unsigned] = ACTIONS(4142),
    [anon_sym_long] = ACTIONS(4142),
    [anon_sym_short] = ACTIONS(4142),
    [anon_sym_static] = ACTIONS(4142),
    [anon_sym_EQ] = ACTIONS(4142),
    [anon_sym_ATautoreleasepool] = ACTIONS(4144),
    [anon_sym_auto] = ACTIONS(4142),
    [anon_sym_register] = ACTIONS(4142),
    [anon_sym_inline] = ACTIONS(4142),
    [anon_sym___inline] = ACTIONS(4142),
    [anon_sym___inline__] = ACTIONS(4142),
    [anon_sym___forceinline] = ACTIONS(4142),
    [anon_sym_thread_local] = ACTIONS(4142),
    [anon_sym___thread] = ACTIONS(4142),
    [anon_sym_CG_EXTERN] = ACTIONS(4142),
    [anon_sym_CG_INLINE] = ACTIONS(4142),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4142),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4142),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4142),
    [anon_sym_IBOutlet] = ACTIONS(4142),
    [anon_sym_IBInspectable] = ACTIONS(4142),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4142),
    [anon_sym_NS_INLINE] = ACTIONS(4142),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4142),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4142),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4142),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4142),
    [anon_sym_const] = ACTIONS(4142),
    [anon_sym_constexpr] = ACTIONS(4142),
    [anon_sym_volatile] = ACTIONS(4142),
    [anon_sym_restrict] = ACTIONS(4142),
    [anon_sym___restrict__] = ACTIONS(4142),
    [anon_sym__Atomic] = ACTIONS(4142),
    [anon_sym__Noreturn] = ACTIONS(4142),
    [anon_sym__Nonnull] = ACTIONS(4142),
    [anon_sym_nullable] = ACTIONS(4142),
    [anon_sym__Complex] = ACTIONS(4142),
    [anon_sym__Nullable] = ACTIONS(4142),
    [anon_sym__Nullable_result] = ACTIONS(4142),
    [anon_sym__Null_unspecified] = ACTIONS(4142),
    [anon_sym___autoreleasing] = ACTIONS(4142),
    [anon_sym___block] = ACTIONS(4142),
    [anon_sym___bridge] = ACTIONS(4142),
    [anon_sym___bridge_retained] = ACTIONS(4142),
    [anon_sym___bridge_transfer] = ACTIONS(4142),
    [anon_sym___complex] = ACTIONS(4142),
    [anon_sym___const] = ACTIONS(4142),
    [anon_sym___imag] = ACTIONS(4142),
    [anon_sym___kindof] = ACTIONS(4142),
    [anon_sym___nonnull] = ACTIONS(4142),
    [anon_sym___nullable] = ACTIONS(4142),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4142),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4142),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4142),
    [anon_sym___real] = ACTIONS(4142),
    [anon_sym___strong] = ACTIONS(4142),
    [anon_sym___unsafe_unretained] = ACTIONS(4142),
    [anon_sym___unused] = ACTIONS(4142),
    [anon_sym___weak] = ACTIONS(4142),
    [anon_sym_alignas] = ACTIONS(4142),
    [anon_sym__Alignas] = ACTIONS(4142),
    [sym_primitive_type] = ACTIONS(4142),
    [anon_sym_enum] = ACTIONS(4142),
    [anon_sym_COLON] = ACTIONS(4144),
    [anon_sym_struct] = ACTIONS(4142),
    [anon_sym_union] = ACTIONS(4142),
    [anon_sym_QMARK] = ACTIONS(4144),
    [anon_sym_STAR_EQ] = ACTIONS(4144),
    [anon_sym_SLASH_EQ] = ACTIONS(4144),
    [anon_sym_PERCENT_EQ] = ACTIONS(4144),
    [anon_sym_PLUS_EQ] = ACTIONS(4144),
    [anon_sym_DASH_EQ] = ACTIONS(4144),
    [anon_sym_LT_LT_EQ] = ACTIONS(4144),
    [anon_sym_GT_GT_EQ] = ACTIONS(4144),
    [anon_sym_AMP_EQ] = ACTIONS(4144),
    [anon_sym_CARET_EQ] = ACTIONS(4144),
    [anon_sym_PIPE_EQ] = ACTIONS(4144),
    [anon_sym_DASH_DASH] = ACTIONS(4144),
    [anon_sym_PLUS_PLUS] = ACTIONS(4144),
    [anon_sym_DOT] = ACTIONS(4142),
    [anon_sym_DASH_GT] = ACTIONS(4144),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4142),
    [anon_sym___typeof] = ACTIONS(4142),
    [anon_sym_typeof] = ACTIONS(4142),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4142),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4142),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4142),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4142),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4142),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4142),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4142),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4142),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4142),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4142),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4142),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4142),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4142),
    [anon_sym_API_AVAILABLE] = ACTIONS(4142),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4142),
    [anon_sym_API_DEPRECATED] = ACTIONS(4142),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4142),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4142),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4142),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4142),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4142),
    [anon_sym___deprecated_msg] = ACTIONS(4142),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4142),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4142),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4142),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4142),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4142),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4142),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4142),
    [anon_sym_BOOL] = ACTIONS(4142),
    [anon_sym_IMP] = ACTIONS(4142),
    [anon_sym_SEL] = ACTIONS(4142),
    [anon_sym_Class] = ACTIONS(4142),
    [anon_sym_id] = ACTIONS(4142),
  },
  [STATE(1476)] = {
    [sym_identifier] = ACTIONS(4146),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4148),
    [anon_sym_COMMA] = ACTIONS(4148),
    [anon_sym_RPAREN] = ACTIONS(4148),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4148),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4148),
    [anon_sym_LPAREN2] = ACTIONS(4148),
    [anon_sym_DASH] = ACTIONS(4146),
    [anon_sym_PLUS] = ACTIONS(4146),
    [anon_sym_STAR] = ACTIONS(4146),
    [anon_sym_SLASH] = ACTIONS(4146),
    [anon_sym_PERCENT] = ACTIONS(4146),
    [anon_sym_PIPE_PIPE] = ACTIONS(4148),
    [anon_sym_AMP_AMP] = ACTIONS(4148),
    [anon_sym_PIPE] = ACTIONS(4146),
    [anon_sym_CARET] = ACTIONS(4146),
    [anon_sym_AMP] = ACTIONS(4146),
    [anon_sym_EQ_EQ] = ACTIONS(4148),
    [anon_sym_BANG_EQ] = ACTIONS(4148),
    [anon_sym_GT] = ACTIONS(4146),
    [anon_sym_GT_EQ] = ACTIONS(4148),
    [anon_sym_LT_EQ] = ACTIONS(4148),
    [anon_sym_LT] = ACTIONS(4146),
    [anon_sym_LT_LT] = ACTIONS(4146),
    [anon_sym_GT_GT] = ACTIONS(4146),
    [anon_sym_SEMI] = ACTIONS(4148),
    [anon_sym___extension__] = ACTIONS(4146),
    [anon_sym_extern] = ACTIONS(4146),
    [anon_sym___attribute__] = ACTIONS(4146),
    [anon_sym___attribute] = ACTIONS(4146),
    [anon_sym_noreturn] = ACTIONS(4146),
    [anon_sym_LBRACK] = ACTIONS(4148),
    [anon_sym_RBRACK] = ACTIONS(4148),
    [anon_sym___declspec] = ACTIONS(4146),
    [anon_sym_LBRACE] = ACTIONS(4148),
    [anon_sym_RBRACE] = ACTIONS(4148),
    [anon_sym_signed] = ACTIONS(4146),
    [anon_sym_unsigned] = ACTIONS(4146),
    [anon_sym_long] = ACTIONS(4146),
    [anon_sym_short] = ACTIONS(4146),
    [anon_sym_static] = ACTIONS(4146),
    [anon_sym_EQ] = ACTIONS(4146),
    [anon_sym_ATautoreleasepool] = ACTIONS(4148),
    [anon_sym_auto] = ACTIONS(4146),
    [anon_sym_register] = ACTIONS(4146),
    [anon_sym_inline] = ACTIONS(4146),
    [anon_sym___inline] = ACTIONS(4146),
    [anon_sym___inline__] = ACTIONS(4146),
    [anon_sym___forceinline] = ACTIONS(4146),
    [anon_sym_thread_local] = ACTIONS(4146),
    [anon_sym___thread] = ACTIONS(4146),
    [anon_sym_CG_EXTERN] = ACTIONS(4146),
    [anon_sym_CG_INLINE] = ACTIONS(4146),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4146),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4146),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4146),
    [anon_sym_IBOutlet] = ACTIONS(4146),
    [anon_sym_IBInspectable] = ACTIONS(4146),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4146),
    [anon_sym_NS_INLINE] = ACTIONS(4146),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4146),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4146),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4146),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4146),
    [anon_sym_const] = ACTIONS(4146),
    [anon_sym_constexpr] = ACTIONS(4146),
    [anon_sym_volatile] = ACTIONS(4146),
    [anon_sym_restrict] = ACTIONS(4146),
    [anon_sym___restrict__] = ACTIONS(4146),
    [anon_sym__Atomic] = ACTIONS(4146),
    [anon_sym__Noreturn] = ACTIONS(4146),
    [anon_sym__Nonnull] = ACTIONS(4146),
    [anon_sym_nullable] = ACTIONS(4146),
    [anon_sym__Complex] = ACTIONS(4146),
    [anon_sym__Nullable] = ACTIONS(4146),
    [anon_sym__Nullable_result] = ACTIONS(4146),
    [anon_sym__Null_unspecified] = ACTIONS(4146),
    [anon_sym___autoreleasing] = ACTIONS(4146),
    [anon_sym___block] = ACTIONS(4146),
    [anon_sym___bridge] = ACTIONS(4146),
    [anon_sym___bridge_retained] = ACTIONS(4146),
    [anon_sym___bridge_transfer] = ACTIONS(4146),
    [anon_sym___complex] = ACTIONS(4146),
    [anon_sym___const] = ACTIONS(4146),
    [anon_sym___imag] = ACTIONS(4146),
    [anon_sym___kindof] = ACTIONS(4146),
    [anon_sym___nonnull] = ACTIONS(4146),
    [anon_sym___nullable] = ACTIONS(4146),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4146),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4146),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4146),
    [anon_sym___real] = ACTIONS(4146),
    [anon_sym___strong] = ACTIONS(4146),
    [anon_sym___unsafe_unretained] = ACTIONS(4146),
    [anon_sym___unused] = ACTIONS(4146),
    [anon_sym___weak] = ACTIONS(4146),
    [anon_sym_alignas] = ACTIONS(4146),
    [anon_sym__Alignas] = ACTIONS(4146),
    [sym_primitive_type] = ACTIONS(4146),
    [anon_sym_enum] = ACTIONS(4146),
    [anon_sym_COLON] = ACTIONS(4148),
    [anon_sym_struct] = ACTIONS(4146),
    [anon_sym_union] = ACTIONS(4146),
    [anon_sym_QMARK] = ACTIONS(4148),
    [anon_sym_STAR_EQ] = ACTIONS(4148),
    [anon_sym_SLASH_EQ] = ACTIONS(4148),
    [anon_sym_PERCENT_EQ] = ACTIONS(4148),
    [anon_sym_PLUS_EQ] = ACTIONS(4148),
    [anon_sym_DASH_EQ] = ACTIONS(4148),
    [anon_sym_LT_LT_EQ] = ACTIONS(4148),
    [anon_sym_GT_GT_EQ] = ACTIONS(4148),
    [anon_sym_AMP_EQ] = ACTIONS(4148),
    [anon_sym_CARET_EQ] = ACTIONS(4148),
    [anon_sym_PIPE_EQ] = ACTIONS(4148),
    [anon_sym_DASH_DASH] = ACTIONS(4148),
    [anon_sym_PLUS_PLUS] = ACTIONS(4148),
    [anon_sym_DOT] = ACTIONS(4146),
    [anon_sym_DASH_GT] = ACTIONS(4148),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4146),
    [anon_sym___typeof] = ACTIONS(4146),
    [anon_sym_typeof] = ACTIONS(4146),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4146),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4146),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4146),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4146),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4146),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4146),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4146),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4146),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4146),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4146),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4146),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4146),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4146),
    [anon_sym_API_AVAILABLE] = ACTIONS(4146),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4146),
    [anon_sym_API_DEPRECATED] = ACTIONS(4146),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4146),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4146),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4146),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4146),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4146),
    [anon_sym___deprecated_msg] = ACTIONS(4146),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4146),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4146),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4146),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4146),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4146),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4146),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4146),
    [anon_sym_BOOL] = ACTIONS(4146),
    [anon_sym_IMP] = ACTIONS(4146),
    [anon_sym_SEL] = ACTIONS(4146),
    [anon_sym_Class] = ACTIONS(4146),
    [anon_sym_id] = ACTIONS(4146),
  },
  [STATE(1477)] = {
    [sym_identifier] = ACTIONS(4150),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4152),
    [anon_sym_COMMA] = ACTIONS(4152),
    [anon_sym_RPAREN] = ACTIONS(4152),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4152),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4152),
    [anon_sym_LPAREN2] = ACTIONS(4152),
    [anon_sym_DASH] = ACTIONS(4150),
    [anon_sym_PLUS] = ACTIONS(4150),
    [anon_sym_STAR] = ACTIONS(4150),
    [anon_sym_SLASH] = ACTIONS(4150),
    [anon_sym_PERCENT] = ACTIONS(4150),
    [anon_sym_PIPE_PIPE] = ACTIONS(4152),
    [anon_sym_AMP_AMP] = ACTIONS(4152),
    [anon_sym_PIPE] = ACTIONS(4150),
    [anon_sym_CARET] = ACTIONS(4150),
    [anon_sym_AMP] = ACTIONS(4150),
    [anon_sym_EQ_EQ] = ACTIONS(4152),
    [anon_sym_BANG_EQ] = ACTIONS(4152),
    [anon_sym_GT] = ACTIONS(4150),
    [anon_sym_GT_EQ] = ACTIONS(4152),
    [anon_sym_LT_EQ] = ACTIONS(4152),
    [anon_sym_LT] = ACTIONS(4150),
    [anon_sym_LT_LT] = ACTIONS(4150),
    [anon_sym_GT_GT] = ACTIONS(4150),
    [anon_sym_SEMI] = ACTIONS(4152),
    [anon_sym___extension__] = ACTIONS(4150),
    [anon_sym_extern] = ACTIONS(4150),
    [anon_sym___attribute__] = ACTIONS(4150),
    [anon_sym___attribute] = ACTIONS(4150),
    [anon_sym_noreturn] = ACTIONS(4150),
    [anon_sym_LBRACK] = ACTIONS(4152),
    [anon_sym_RBRACK] = ACTIONS(4152),
    [anon_sym___declspec] = ACTIONS(4150),
    [anon_sym_LBRACE] = ACTIONS(4152),
    [anon_sym_RBRACE] = ACTIONS(4152),
    [anon_sym_signed] = ACTIONS(4150),
    [anon_sym_unsigned] = ACTIONS(4150),
    [anon_sym_long] = ACTIONS(4150),
    [anon_sym_short] = ACTIONS(4150),
    [anon_sym_static] = ACTIONS(4150),
    [anon_sym_EQ] = ACTIONS(4150),
    [anon_sym_ATautoreleasepool] = ACTIONS(4152),
    [anon_sym_auto] = ACTIONS(4150),
    [anon_sym_register] = ACTIONS(4150),
    [anon_sym_inline] = ACTIONS(4150),
    [anon_sym___inline] = ACTIONS(4150),
    [anon_sym___inline__] = ACTIONS(4150),
    [anon_sym___forceinline] = ACTIONS(4150),
    [anon_sym_thread_local] = ACTIONS(4150),
    [anon_sym___thread] = ACTIONS(4150),
    [anon_sym_CG_EXTERN] = ACTIONS(4150),
    [anon_sym_CG_INLINE] = ACTIONS(4150),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4150),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4150),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4150),
    [anon_sym_IBOutlet] = ACTIONS(4150),
    [anon_sym_IBInspectable] = ACTIONS(4150),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4150),
    [anon_sym_NS_INLINE] = ACTIONS(4150),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4150),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4150),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4150),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4150),
    [anon_sym_const] = ACTIONS(4150),
    [anon_sym_constexpr] = ACTIONS(4150),
    [anon_sym_volatile] = ACTIONS(4150),
    [anon_sym_restrict] = ACTIONS(4150),
    [anon_sym___restrict__] = ACTIONS(4150),
    [anon_sym__Atomic] = ACTIONS(4150),
    [anon_sym__Noreturn] = ACTIONS(4150),
    [anon_sym__Nonnull] = ACTIONS(4150),
    [anon_sym_nullable] = ACTIONS(4150),
    [anon_sym__Complex] = ACTIONS(4150),
    [anon_sym__Nullable] = ACTIONS(4150),
    [anon_sym__Nullable_result] = ACTIONS(4150),
    [anon_sym__Null_unspecified] = ACTIONS(4150),
    [anon_sym___autoreleasing] = ACTIONS(4150),
    [anon_sym___block] = ACTIONS(4150),
    [anon_sym___bridge] = ACTIONS(4150),
    [anon_sym___bridge_retained] = ACTIONS(4150),
    [anon_sym___bridge_transfer] = ACTIONS(4150),
    [anon_sym___complex] = ACTIONS(4150),
    [anon_sym___const] = ACTIONS(4150),
    [anon_sym___imag] = ACTIONS(4150),
    [anon_sym___kindof] = ACTIONS(4150),
    [anon_sym___nonnull] = ACTIONS(4150),
    [anon_sym___nullable] = ACTIONS(4150),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4150),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4150),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4150),
    [anon_sym___real] = ACTIONS(4150),
    [anon_sym___strong] = ACTIONS(4150),
    [anon_sym___unsafe_unretained] = ACTIONS(4150),
    [anon_sym___unused] = ACTIONS(4150),
    [anon_sym___weak] = ACTIONS(4150),
    [anon_sym_alignas] = ACTIONS(4150),
    [anon_sym__Alignas] = ACTIONS(4150),
    [sym_primitive_type] = ACTIONS(4150),
    [anon_sym_enum] = ACTIONS(4150),
    [anon_sym_COLON] = ACTIONS(4152),
    [anon_sym_struct] = ACTIONS(4150),
    [anon_sym_union] = ACTIONS(4150),
    [anon_sym_QMARK] = ACTIONS(4152),
    [anon_sym_STAR_EQ] = ACTIONS(4152),
    [anon_sym_SLASH_EQ] = ACTIONS(4152),
    [anon_sym_PERCENT_EQ] = ACTIONS(4152),
    [anon_sym_PLUS_EQ] = ACTIONS(4152),
    [anon_sym_DASH_EQ] = ACTIONS(4152),
    [anon_sym_LT_LT_EQ] = ACTIONS(4152),
    [anon_sym_GT_GT_EQ] = ACTIONS(4152),
    [anon_sym_AMP_EQ] = ACTIONS(4152),
    [anon_sym_CARET_EQ] = ACTIONS(4152),
    [anon_sym_PIPE_EQ] = ACTIONS(4152),
    [anon_sym_DASH_DASH] = ACTIONS(4152),
    [anon_sym_PLUS_PLUS] = ACTIONS(4152),
    [anon_sym_DOT] = ACTIONS(4150),
    [anon_sym_DASH_GT] = ACTIONS(4152),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4150),
    [anon_sym___typeof] = ACTIONS(4150),
    [anon_sym_typeof] = ACTIONS(4150),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4150),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4150),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4150),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4150),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4150),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4150),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4150),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4150),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4150),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4150),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4150),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4150),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4150),
    [anon_sym_API_AVAILABLE] = ACTIONS(4150),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4150),
    [anon_sym_API_DEPRECATED] = ACTIONS(4150),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4150),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4150),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4150),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4150),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4150),
    [anon_sym___deprecated_msg] = ACTIONS(4150),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4150),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4150),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4150),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4150),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4150),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4150),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4150),
    [anon_sym_BOOL] = ACTIONS(4150),
    [anon_sym_IMP] = ACTIONS(4150),
    [anon_sym_SEL] = ACTIONS(4150),
    [anon_sym_Class] = ACTIONS(4150),
    [anon_sym_id] = ACTIONS(4150),
  },
  [STATE(1478)] = {
    [sym_identifier] = ACTIONS(4154),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4157),
    [anon_sym_COMMA] = ACTIONS(4157),
    [anon_sym_LPAREN2] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4154),
    [anon_sym_PLUS] = ACTIONS(4154),
    [anon_sym_STAR] = ACTIONS(4154),
    [anon_sym_SLASH] = ACTIONS(4154),
    [anon_sym_PERCENT] = ACTIONS(4154),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4154),
    [anon_sym_CARET] = ACTIONS(4154),
    [anon_sym_AMP] = ACTIONS(4154),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4154),
    [anon_sym_GT_EQ] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_LT] = ACTIONS(4154),
    [anon_sym_LT_LT] = ACTIONS(4154),
    [anon_sym_GT_GT] = ACTIONS(4154),
    [anon_sym_SEMI] = ACTIONS(4157),
    [anon_sym___extension__] = ACTIONS(4154),
    [anon_sym_extern] = ACTIONS(4154),
    [anon_sym___attribute__] = ACTIONS(4154),
    [anon_sym___attribute] = ACTIONS(4154),
    [anon_sym_noreturn] = ACTIONS(4154),
    [anon_sym_LBRACK] = ACTIONS(4157),
    [anon_sym___declspec] = ACTIONS(4154),
    [anon_sym_LBRACE] = ACTIONS(4157),
    [anon_sym_signed] = ACTIONS(4154),
    [anon_sym_unsigned] = ACTIONS(4154),
    [anon_sym_long] = ACTIONS(4154),
    [anon_sym_short] = ACTIONS(4154),
    [anon_sym_static] = ACTIONS(4154),
    [anon_sym_EQ] = ACTIONS(4154),
    [anon_sym_ATautoreleasepool] = ACTIONS(4157),
    [anon_sym_auto] = ACTIONS(4154),
    [anon_sym_register] = ACTIONS(4154),
    [anon_sym_inline] = ACTIONS(4154),
    [anon_sym___inline] = ACTIONS(4154),
    [anon_sym___inline__] = ACTIONS(4154),
    [anon_sym___forceinline] = ACTIONS(4154),
    [anon_sym_thread_local] = ACTIONS(4154),
    [anon_sym___thread] = ACTIONS(4154),
    [anon_sym_CG_EXTERN] = ACTIONS(4154),
    [anon_sym_CG_INLINE] = ACTIONS(4154),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4154),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4154),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4154),
    [anon_sym_IBOutlet] = ACTIONS(4154),
    [anon_sym_IBInspectable] = ACTIONS(4154),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4154),
    [anon_sym_NS_INLINE] = ACTIONS(4154),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4154),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4154),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4154),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4154),
    [anon_sym_const] = ACTIONS(4154),
    [anon_sym_constexpr] = ACTIONS(4154),
    [anon_sym_volatile] = ACTIONS(4154),
    [anon_sym_restrict] = ACTIONS(4154),
    [anon_sym___restrict__] = ACTIONS(4154),
    [anon_sym__Atomic] = ACTIONS(4154),
    [anon_sym__Noreturn] = ACTIONS(4154),
    [anon_sym__Nonnull] = ACTIONS(4154),
    [anon_sym_nullable] = ACTIONS(4154),
    [anon_sym__Complex] = ACTIONS(4154),
    [anon_sym__Nullable] = ACTIONS(4154),
    [anon_sym__Nullable_result] = ACTIONS(4154),
    [anon_sym__Null_unspecified] = ACTIONS(4154),
    [anon_sym___autoreleasing] = ACTIONS(4154),
    [anon_sym___block] = ACTIONS(4154),
    [anon_sym___bridge] = ACTIONS(4154),
    [anon_sym___bridge_retained] = ACTIONS(4154),
    [anon_sym___bridge_transfer] = ACTIONS(4154),
    [anon_sym___complex] = ACTIONS(4154),
    [anon_sym___const] = ACTIONS(4154),
    [anon_sym___imag] = ACTIONS(4154),
    [anon_sym___kindof] = ACTIONS(4154),
    [anon_sym___nonnull] = ACTIONS(4154),
    [anon_sym___nullable] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4154),
    [anon_sym___real] = ACTIONS(4154),
    [anon_sym___strong] = ACTIONS(4154),
    [anon_sym___unsafe_unretained] = ACTIONS(4154),
    [anon_sym___unused] = ACTIONS(4154),
    [anon_sym___weak] = ACTIONS(4154),
    [anon_sym_alignas] = ACTIONS(4154),
    [anon_sym__Alignas] = ACTIONS(4154),
    [sym_primitive_type] = ACTIONS(4154),
    [anon_sym_enum] = ACTIONS(4154),
    [anon_sym_struct] = ACTIONS(4154),
    [anon_sym_union] = ACTIONS(4154),
    [anon_sym_QMARK] = ACTIONS(4157),
    [anon_sym_STAR_EQ] = ACTIONS(4157),
    [anon_sym_SLASH_EQ] = ACTIONS(4157),
    [anon_sym_PERCENT_EQ] = ACTIONS(4157),
    [anon_sym_PLUS_EQ] = ACTIONS(4157),
    [anon_sym_DASH_EQ] = ACTIONS(4157),
    [anon_sym_LT_LT_EQ] = ACTIONS(4157),
    [anon_sym_GT_GT_EQ] = ACTIONS(4157),
    [anon_sym_AMP_EQ] = ACTIONS(4157),
    [anon_sym_CARET_EQ] = ACTIONS(4157),
    [anon_sym_PIPE_EQ] = ACTIONS(4157),
    [anon_sym_DASH_DASH] = ACTIONS(4157),
    [anon_sym_PLUS_PLUS] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4154),
    [anon_sym_DASH_GT] = ACTIONS(4157),
    [anon_sym_AT] = ACTIONS(4154),
    [anon_sym_DQUOTE] = ACTIONS(4157),
    [anon_sym_L_DQUOTE] = ACTIONS(4157),
    [anon_sym_u_DQUOTE] = ACTIONS(4157),
    [anon_sym_U_DQUOTE] = ACTIONS(4157),
    [anon_sym_u8_DQUOTE] = ACTIONS(4157),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4154),
    [anon_sym___typeof] = ACTIONS(4154),
    [anon_sym_typeof] = ACTIONS(4154),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4154),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4154),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4154),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4154),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4154),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4154),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4154),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_API_AVAILABLE] = ACTIONS(4154),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_API_DEPRECATED] = ACTIONS(4154),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4154),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym___deprecated_msg] = ACTIONS(4154),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4154),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4154),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4154),
    [anon_sym_BOOL] = ACTIONS(4154),
    [anon_sym_IMP] = ACTIONS(4154),
    [anon_sym_SEL] = ACTIONS(4154),
    [anon_sym_Class] = ACTIONS(4154),
    [anon_sym_id] = ACTIONS(4154),
  },
  [STATE(1479)] = {
    [sym_identifier] = ACTIONS(4160),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4162),
    [anon_sym_COMMA] = ACTIONS(4162),
    [anon_sym_RPAREN] = ACTIONS(4162),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4162),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4162),
    [anon_sym_LPAREN2] = ACTIONS(4162),
    [anon_sym_DASH] = ACTIONS(4160),
    [anon_sym_PLUS] = ACTIONS(4160),
    [anon_sym_STAR] = ACTIONS(4160),
    [anon_sym_SLASH] = ACTIONS(4160),
    [anon_sym_PERCENT] = ACTIONS(4160),
    [anon_sym_PIPE_PIPE] = ACTIONS(4162),
    [anon_sym_AMP_AMP] = ACTIONS(4162),
    [anon_sym_PIPE] = ACTIONS(4160),
    [anon_sym_CARET] = ACTIONS(4160),
    [anon_sym_AMP] = ACTIONS(4160),
    [anon_sym_EQ_EQ] = ACTIONS(4162),
    [anon_sym_BANG_EQ] = ACTIONS(4162),
    [anon_sym_GT] = ACTIONS(4160),
    [anon_sym_GT_EQ] = ACTIONS(4162),
    [anon_sym_LT_EQ] = ACTIONS(4162),
    [anon_sym_LT] = ACTIONS(4160),
    [anon_sym_LT_LT] = ACTIONS(4160),
    [anon_sym_GT_GT] = ACTIONS(4160),
    [anon_sym_SEMI] = ACTIONS(4162),
    [anon_sym___extension__] = ACTIONS(4160),
    [anon_sym_extern] = ACTIONS(4160),
    [anon_sym___attribute__] = ACTIONS(4160),
    [anon_sym___attribute] = ACTIONS(4160),
    [anon_sym_noreturn] = ACTIONS(4160),
    [anon_sym_LBRACK] = ACTIONS(4162),
    [anon_sym_RBRACK] = ACTIONS(4162),
    [anon_sym___declspec] = ACTIONS(4160),
    [anon_sym_LBRACE] = ACTIONS(4162),
    [anon_sym_RBRACE] = ACTIONS(4162),
    [anon_sym_signed] = ACTIONS(4160),
    [anon_sym_unsigned] = ACTIONS(4160),
    [anon_sym_long] = ACTIONS(4160),
    [anon_sym_short] = ACTIONS(4160),
    [anon_sym_static] = ACTIONS(4160),
    [anon_sym_EQ] = ACTIONS(4160),
    [anon_sym_ATautoreleasepool] = ACTIONS(4162),
    [anon_sym_auto] = ACTIONS(4160),
    [anon_sym_register] = ACTIONS(4160),
    [anon_sym_inline] = ACTIONS(4160),
    [anon_sym___inline] = ACTIONS(4160),
    [anon_sym___inline__] = ACTIONS(4160),
    [anon_sym___forceinline] = ACTIONS(4160),
    [anon_sym_thread_local] = ACTIONS(4160),
    [anon_sym___thread] = ACTIONS(4160),
    [anon_sym_CG_EXTERN] = ACTIONS(4160),
    [anon_sym_CG_INLINE] = ACTIONS(4160),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4160),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4160),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4160),
    [anon_sym_IBOutlet] = ACTIONS(4160),
    [anon_sym_IBInspectable] = ACTIONS(4160),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4160),
    [anon_sym_NS_INLINE] = ACTIONS(4160),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4160),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4160),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4160),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4160),
    [anon_sym_const] = ACTIONS(4160),
    [anon_sym_constexpr] = ACTIONS(4160),
    [anon_sym_volatile] = ACTIONS(4160),
    [anon_sym_restrict] = ACTIONS(4160),
    [anon_sym___restrict__] = ACTIONS(4160),
    [anon_sym__Atomic] = ACTIONS(4160),
    [anon_sym__Noreturn] = ACTIONS(4160),
    [anon_sym__Nonnull] = ACTIONS(4160),
    [anon_sym_nullable] = ACTIONS(4160),
    [anon_sym__Complex] = ACTIONS(4160),
    [anon_sym__Nullable] = ACTIONS(4160),
    [anon_sym__Nullable_result] = ACTIONS(4160),
    [anon_sym__Null_unspecified] = ACTIONS(4160),
    [anon_sym___autoreleasing] = ACTIONS(4160),
    [anon_sym___block] = ACTIONS(4160),
    [anon_sym___bridge] = ACTIONS(4160),
    [anon_sym___bridge_retained] = ACTIONS(4160),
    [anon_sym___bridge_transfer] = ACTIONS(4160),
    [anon_sym___complex] = ACTIONS(4160),
    [anon_sym___const] = ACTIONS(4160),
    [anon_sym___imag] = ACTIONS(4160),
    [anon_sym___kindof] = ACTIONS(4160),
    [anon_sym___nonnull] = ACTIONS(4160),
    [anon_sym___nullable] = ACTIONS(4160),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4160),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4160),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4160),
    [anon_sym___real] = ACTIONS(4160),
    [anon_sym___strong] = ACTIONS(4160),
    [anon_sym___unsafe_unretained] = ACTIONS(4160),
    [anon_sym___unused] = ACTIONS(4160),
    [anon_sym___weak] = ACTIONS(4160),
    [anon_sym_alignas] = ACTIONS(4160),
    [anon_sym__Alignas] = ACTIONS(4160),
    [sym_primitive_type] = ACTIONS(4160),
    [anon_sym_enum] = ACTIONS(4160),
    [anon_sym_COLON] = ACTIONS(4162),
    [anon_sym_struct] = ACTIONS(4160),
    [anon_sym_union] = ACTIONS(4160),
    [anon_sym_QMARK] = ACTIONS(4162),
    [anon_sym_STAR_EQ] = ACTIONS(4162),
    [anon_sym_SLASH_EQ] = ACTIONS(4162),
    [anon_sym_PERCENT_EQ] = ACTIONS(4162),
    [anon_sym_PLUS_EQ] = ACTIONS(4162),
    [anon_sym_DASH_EQ] = ACTIONS(4162),
    [anon_sym_LT_LT_EQ] = ACTIONS(4162),
    [anon_sym_GT_GT_EQ] = ACTIONS(4162),
    [anon_sym_AMP_EQ] = ACTIONS(4162),
    [anon_sym_CARET_EQ] = ACTIONS(4162),
    [anon_sym_PIPE_EQ] = ACTIONS(4162),
    [anon_sym_DASH_DASH] = ACTIONS(4162),
    [anon_sym_PLUS_PLUS] = ACTIONS(4162),
    [anon_sym_DOT] = ACTIONS(4160),
    [anon_sym_DASH_GT] = ACTIONS(4162),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4160),
    [anon_sym___typeof] = ACTIONS(4160),
    [anon_sym_typeof] = ACTIONS(4160),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4160),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4160),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4160),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4160),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4160),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4160),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4160),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4160),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4160),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4160),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4160),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4160),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4160),
    [anon_sym_API_AVAILABLE] = ACTIONS(4160),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4160),
    [anon_sym_API_DEPRECATED] = ACTIONS(4160),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4160),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4160),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4160),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4160),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4160),
    [anon_sym___deprecated_msg] = ACTIONS(4160),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4160),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4160),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4160),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4160),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4160),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4160),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4160),
    [anon_sym_BOOL] = ACTIONS(4160),
    [anon_sym_IMP] = ACTIONS(4160),
    [anon_sym_SEL] = ACTIONS(4160),
    [anon_sym_Class] = ACTIONS(4160),
    [anon_sym_id] = ACTIONS(4160),
  },
  [STATE(1480)] = {
    [sym_identifier] = ACTIONS(4164),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4166),
    [anon_sym_COMMA] = ACTIONS(4166),
    [anon_sym_RPAREN] = ACTIONS(4166),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4166),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4166),
    [anon_sym_LPAREN2] = ACTIONS(4166),
    [anon_sym_DASH] = ACTIONS(4164),
    [anon_sym_PLUS] = ACTIONS(4164),
    [anon_sym_STAR] = ACTIONS(4164),
    [anon_sym_SLASH] = ACTIONS(4164),
    [anon_sym_PERCENT] = ACTIONS(4164),
    [anon_sym_PIPE_PIPE] = ACTIONS(4166),
    [anon_sym_AMP_AMP] = ACTIONS(4166),
    [anon_sym_PIPE] = ACTIONS(4164),
    [anon_sym_CARET] = ACTIONS(4164),
    [anon_sym_AMP] = ACTIONS(4164),
    [anon_sym_EQ_EQ] = ACTIONS(4166),
    [anon_sym_BANG_EQ] = ACTIONS(4166),
    [anon_sym_GT] = ACTIONS(4164),
    [anon_sym_GT_EQ] = ACTIONS(4166),
    [anon_sym_LT_EQ] = ACTIONS(4166),
    [anon_sym_LT] = ACTIONS(4164),
    [anon_sym_LT_LT] = ACTIONS(4164),
    [anon_sym_GT_GT] = ACTIONS(4164),
    [anon_sym_SEMI] = ACTIONS(4166),
    [anon_sym___extension__] = ACTIONS(4164),
    [anon_sym_extern] = ACTIONS(4164),
    [anon_sym___attribute__] = ACTIONS(4164),
    [anon_sym___attribute] = ACTIONS(4164),
    [anon_sym_noreturn] = ACTIONS(4164),
    [anon_sym_LBRACK] = ACTIONS(4166),
    [anon_sym_RBRACK] = ACTIONS(4166),
    [anon_sym___declspec] = ACTIONS(4164),
    [anon_sym_LBRACE] = ACTIONS(4166),
    [anon_sym_RBRACE] = ACTIONS(4166),
    [anon_sym_signed] = ACTIONS(4164),
    [anon_sym_unsigned] = ACTIONS(4164),
    [anon_sym_long] = ACTIONS(4164),
    [anon_sym_short] = ACTIONS(4164),
    [anon_sym_static] = ACTIONS(4164),
    [anon_sym_EQ] = ACTIONS(4164),
    [anon_sym_ATautoreleasepool] = ACTIONS(4166),
    [anon_sym_auto] = ACTIONS(4164),
    [anon_sym_register] = ACTIONS(4164),
    [anon_sym_inline] = ACTIONS(4164),
    [anon_sym___inline] = ACTIONS(4164),
    [anon_sym___inline__] = ACTIONS(4164),
    [anon_sym___forceinline] = ACTIONS(4164),
    [anon_sym_thread_local] = ACTIONS(4164),
    [anon_sym___thread] = ACTIONS(4164),
    [anon_sym_CG_EXTERN] = ACTIONS(4164),
    [anon_sym_CG_INLINE] = ACTIONS(4164),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4164),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4164),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4164),
    [anon_sym_IBOutlet] = ACTIONS(4164),
    [anon_sym_IBInspectable] = ACTIONS(4164),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4164),
    [anon_sym_NS_INLINE] = ACTIONS(4164),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4164),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4164),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4164),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4164),
    [anon_sym_const] = ACTIONS(4164),
    [anon_sym_constexpr] = ACTIONS(4164),
    [anon_sym_volatile] = ACTIONS(4164),
    [anon_sym_restrict] = ACTIONS(4164),
    [anon_sym___restrict__] = ACTIONS(4164),
    [anon_sym__Atomic] = ACTIONS(4164),
    [anon_sym__Noreturn] = ACTIONS(4164),
    [anon_sym__Nonnull] = ACTIONS(4164),
    [anon_sym_nullable] = ACTIONS(4164),
    [anon_sym__Complex] = ACTIONS(4164),
    [anon_sym__Nullable] = ACTIONS(4164),
    [anon_sym__Nullable_result] = ACTIONS(4164),
    [anon_sym__Null_unspecified] = ACTIONS(4164),
    [anon_sym___autoreleasing] = ACTIONS(4164),
    [anon_sym___block] = ACTIONS(4164),
    [anon_sym___bridge] = ACTIONS(4164),
    [anon_sym___bridge_retained] = ACTIONS(4164),
    [anon_sym___bridge_transfer] = ACTIONS(4164),
    [anon_sym___complex] = ACTIONS(4164),
    [anon_sym___const] = ACTIONS(4164),
    [anon_sym___imag] = ACTIONS(4164),
    [anon_sym___kindof] = ACTIONS(4164),
    [anon_sym___nonnull] = ACTIONS(4164),
    [anon_sym___nullable] = ACTIONS(4164),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4164),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4164),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4164),
    [anon_sym___real] = ACTIONS(4164),
    [anon_sym___strong] = ACTIONS(4164),
    [anon_sym___unsafe_unretained] = ACTIONS(4164),
    [anon_sym___unused] = ACTIONS(4164),
    [anon_sym___weak] = ACTIONS(4164),
    [anon_sym_alignas] = ACTIONS(4164),
    [anon_sym__Alignas] = ACTIONS(4164),
    [sym_primitive_type] = ACTIONS(4164),
    [anon_sym_enum] = ACTIONS(4164),
    [anon_sym_COLON] = ACTIONS(4166),
    [anon_sym_struct] = ACTIONS(4164),
    [anon_sym_union] = ACTIONS(4164),
    [anon_sym_QMARK] = ACTIONS(4166),
    [anon_sym_STAR_EQ] = ACTIONS(4166),
    [anon_sym_SLASH_EQ] = ACTIONS(4166),
    [anon_sym_PERCENT_EQ] = ACTIONS(4166),
    [anon_sym_PLUS_EQ] = ACTIONS(4166),
    [anon_sym_DASH_EQ] = ACTIONS(4166),
    [anon_sym_LT_LT_EQ] = ACTIONS(4166),
    [anon_sym_GT_GT_EQ] = ACTIONS(4166),
    [anon_sym_AMP_EQ] = ACTIONS(4166),
    [anon_sym_CARET_EQ] = ACTIONS(4166),
    [anon_sym_PIPE_EQ] = ACTIONS(4166),
    [anon_sym_DASH_DASH] = ACTIONS(4166),
    [anon_sym_PLUS_PLUS] = ACTIONS(4166),
    [anon_sym_DOT] = ACTIONS(4164),
    [anon_sym_DASH_GT] = ACTIONS(4166),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4164),
    [anon_sym___typeof] = ACTIONS(4164),
    [anon_sym_typeof] = ACTIONS(4164),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4164),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4164),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4164),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4164),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4164),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4164),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4164),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4164),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4164),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4164),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4164),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4164),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4164),
    [anon_sym_API_AVAILABLE] = ACTIONS(4164),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4164),
    [anon_sym_API_DEPRECATED] = ACTIONS(4164),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4164),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4164),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4164),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4164),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4164),
    [anon_sym___deprecated_msg] = ACTIONS(4164),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4164),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4164),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4164),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4164),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4164),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4164),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4164),
    [anon_sym_BOOL] = ACTIONS(4164),
    [anon_sym_IMP] = ACTIONS(4164),
    [anon_sym_SEL] = ACTIONS(4164),
    [anon_sym_Class] = ACTIONS(4164),
    [anon_sym_id] = ACTIONS(4164),
  },
  [STATE(1481)] = {
    [sym_identifier] = ACTIONS(4168),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4170),
    [anon_sym_COMMA] = ACTIONS(4170),
    [anon_sym_RPAREN] = ACTIONS(4170),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4170),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4170),
    [anon_sym_LPAREN2] = ACTIONS(4170),
    [anon_sym_DASH] = ACTIONS(4168),
    [anon_sym_PLUS] = ACTIONS(4168),
    [anon_sym_STAR] = ACTIONS(4168),
    [anon_sym_SLASH] = ACTIONS(4168),
    [anon_sym_PERCENT] = ACTIONS(4168),
    [anon_sym_PIPE_PIPE] = ACTIONS(4170),
    [anon_sym_AMP_AMP] = ACTIONS(4170),
    [anon_sym_PIPE] = ACTIONS(4168),
    [anon_sym_CARET] = ACTIONS(4168),
    [anon_sym_AMP] = ACTIONS(4168),
    [anon_sym_EQ_EQ] = ACTIONS(4170),
    [anon_sym_BANG_EQ] = ACTIONS(4170),
    [anon_sym_GT] = ACTIONS(4168),
    [anon_sym_GT_EQ] = ACTIONS(4170),
    [anon_sym_LT_EQ] = ACTIONS(4170),
    [anon_sym_LT] = ACTIONS(4168),
    [anon_sym_LT_LT] = ACTIONS(4168),
    [anon_sym_GT_GT] = ACTIONS(4168),
    [anon_sym_SEMI] = ACTIONS(4170),
    [anon_sym___extension__] = ACTIONS(4168),
    [anon_sym_extern] = ACTIONS(4168),
    [anon_sym___attribute__] = ACTIONS(4168),
    [anon_sym___attribute] = ACTIONS(4168),
    [anon_sym_noreturn] = ACTIONS(4168),
    [anon_sym_LBRACK] = ACTIONS(4170),
    [anon_sym_RBRACK] = ACTIONS(4170),
    [anon_sym___declspec] = ACTIONS(4168),
    [anon_sym_LBRACE] = ACTIONS(4170),
    [anon_sym_RBRACE] = ACTIONS(4170),
    [anon_sym_signed] = ACTIONS(4168),
    [anon_sym_unsigned] = ACTIONS(4168),
    [anon_sym_long] = ACTIONS(4168),
    [anon_sym_short] = ACTIONS(4168),
    [anon_sym_static] = ACTIONS(4168),
    [anon_sym_EQ] = ACTIONS(4168),
    [anon_sym_ATautoreleasepool] = ACTIONS(4170),
    [anon_sym_auto] = ACTIONS(4168),
    [anon_sym_register] = ACTIONS(4168),
    [anon_sym_inline] = ACTIONS(4168),
    [anon_sym___inline] = ACTIONS(4168),
    [anon_sym___inline__] = ACTIONS(4168),
    [anon_sym___forceinline] = ACTIONS(4168),
    [anon_sym_thread_local] = ACTIONS(4168),
    [anon_sym___thread] = ACTIONS(4168),
    [anon_sym_CG_EXTERN] = ACTIONS(4168),
    [anon_sym_CG_INLINE] = ACTIONS(4168),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4168),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4168),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4168),
    [anon_sym_IBOutlet] = ACTIONS(4168),
    [anon_sym_IBInspectable] = ACTIONS(4168),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4168),
    [anon_sym_NS_INLINE] = ACTIONS(4168),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4168),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4168),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4168),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4168),
    [anon_sym_const] = ACTIONS(4168),
    [anon_sym_constexpr] = ACTIONS(4168),
    [anon_sym_volatile] = ACTIONS(4168),
    [anon_sym_restrict] = ACTIONS(4168),
    [anon_sym___restrict__] = ACTIONS(4168),
    [anon_sym__Atomic] = ACTIONS(4168),
    [anon_sym__Noreturn] = ACTIONS(4168),
    [anon_sym__Nonnull] = ACTIONS(4168),
    [anon_sym_nullable] = ACTIONS(4168),
    [anon_sym__Complex] = ACTIONS(4168),
    [anon_sym__Nullable] = ACTIONS(4168),
    [anon_sym__Nullable_result] = ACTIONS(4168),
    [anon_sym__Null_unspecified] = ACTIONS(4168),
    [anon_sym___autoreleasing] = ACTIONS(4168),
    [anon_sym___block] = ACTIONS(4168),
    [anon_sym___bridge] = ACTIONS(4168),
    [anon_sym___bridge_retained] = ACTIONS(4168),
    [anon_sym___bridge_transfer] = ACTIONS(4168),
    [anon_sym___complex] = ACTIONS(4168),
    [anon_sym___const] = ACTIONS(4168),
    [anon_sym___imag] = ACTIONS(4168),
    [anon_sym___kindof] = ACTIONS(4168),
    [anon_sym___nonnull] = ACTIONS(4168),
    [anon_sym___nullable] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4168),
    [anon_sym___real] = ACTIONS(4168),
    [anon_sym___strong] = ACTIONS(4168),
    [anon_sym___unsafe_unretained] = ACTIONS(4168),
    [anon_sym___unused] = ACTIONS(4168),
    [anon_sym___weak] = ACTIONS(4168),
    [anon_sym_alignas] = ACTIONS(4168),
    [anon_sym__Alignas] = ACTIONS(4168),
    [sym_primitive_type] = ACTIONS(4168),
    [anon_sym_enum] = ACTIONS(4168),
    [anon_sym_COLON] = ACTIONS(4170),
    [anon_sym_struct] = ACTIONS(4168),
    [anon_sym_union] = ACTIONS(4168),
    [anon_sym_QMARK] = ACTIONS(4170),
    [anon_sym_STAR_EQ] = ACTIONS(4170),
    [anon_sym_SLASH_EQ] = ACTIONS(4170),
    [anon_sym_PERCENT_EQ] = ACTIONS(4170),
    [anon_sym_PLUS_EQ] = ACTIONS(4170),
    [anon_sym_DASH_EQ] = ACTIONS(4170),
    [anon_sym_LT_LT_EQ] = ACTIONS(4170),
    [anon_sym_GT_GT_EQ] = ACTIONS(4170),
    [anon_sym_AMP_EQ] = ACTIONS(4170),
    [anon_sym_CARET_EQ] = ACTIONS(4170),
    [anon_sym_PIPE_EQ] = ACTIONS(4170),
    [anon_sym_DASH_DASH] = ACTIONS(4170),
    [anon_sym_PLUS_PLUS] = ACTIONS(4170),
    [anon_sym_DOT] = ACTIONS(4168),
    [anon_sym_DASH_GT] = ACTIONS(4170),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4168),
    [anon_sym___typeof] = ACTIONS(4168),
    [anon_sym_typeof] = ACTIONS(4168),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4168),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4168),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4168),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4168),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4168),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4168),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4168),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4168),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4168),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_API_AVAILABLE] = ACTIONS(4168),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_API_DEPRECATED] = ACTIONS(4168),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4168),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4168),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4168),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4168),
    [anon_sym___deprecated_msg] = ACTIONS(4168),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4168),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4168),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4168),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4168),
    [anon_sym_BOOL] = ACTIONS(4168),
    [anon_sym_IMP] = ACTIONS(4168),
    [anon_sym_SEL] = ACTIONS(4168),
    [anon_sym_Class] = ACTIONS(4168),
    [anon_sym_id] = ACTIONS(4168),
  },
  [STATE(1482)] = {
    [sym_identifier] = ACTIONS(4172),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4174),
    [anon_sym_COMMA] = ACTIONS(4174),
    [anon_sym_RPAREN] = ACTIONS(4174),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4174),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4174),
    [anon_sym_LPAREN2] = ACTIONS(4174),
    [anon_sym_DASH] = ACTIONS(4172),
    [anon_sym_PLUS] = ACTIONS(4172),
    [anon_sym_STAR] = ACTIONS(4172),
    [anon_sym_SLASH] = ACTIONS(4172),
    [anon_sym_PERCENT] = ACTIONS(4172),
    [anon_sym_PIPE_PIPE] = ACTIONS(4174),
    [anon_sym_AMP_AMP] = ACTIONS(4174),
    [anon_sym_PIPE] = ACTIONS(4172),
    [anon_sym_CARET] = ACTIONS(4172),
    [anon_sym_AMP] = ACTIONS(4172),
    [anon_sym_EQ_EQ] = ACTIONS(4174),
    [anon_sym_BANG_EQ] = ACTIONS(4174),
    [anon_sym_GT] = ACTIONS(4172),
    [anon_sym_GT_EQ] = ACTIONS(4174),
    [anon_sym_LT_EQ] = ACTIONS(4174),
    [anon_sym_LT] = ACTIONS(4172),
    [anon_sym_LT_LT] = ACTIONS(4172),
    [anon_sym_GT_GT] = ACTIONS(4172),
    [anon_sym_SEMI] = ACTIONS(4174),
    [anon_sym___extension__] = ACTIONS(4172),
    [anon_sym_extern] = ACTIONS(4172),
    [anon_sym___attribute__] = ACTIONS(4172),
    [anon_sym___attribute] = ACTIONS(4172),
    [anon_sym_noreturn] = ACTIONS(4172),
    [anon_sym_LBRACK] = ACTIONS(4174),
    [anon_sym_RBRACK] = ACTIONS(4174),
    [anon_sym___declspec] = ACTIONS(4172),
    [anon_sym_LBRACE] = ACTIONS(4174),
    [anon_sym_RBRACE] = ACTIONS(4174),
    [anon_sym_signed] = ACTIONS(4172),
    [anon_sym_unsigned] = ACTIONS(4172),
    [anon_sym_long] = ACTIONS(4172),
    [anon_sym_short] = ACTIONS(4172),
    [anon_sym_static] = ACTIONS(4172),
    [anon_sym_EQ] = ACTIONS(4172),
    [anon_sym_ATautoreleasepool] = ACTIONS(4174),
    [anon_sym_auto] = ACTIONS(4172),
    [anon_sym_register] = ACTIONS(4172),
    [anon_sym_inline] = ACTIONS(4172),
    [anon_sym___inline] = ACTIONS(4172),
    [anon_sym___inline__] = ACTIONS(4172),
    [anon_sym___forceinline] = ACTIONS(4172),
    [anon_sym_thread_local] = ACTIONS(4172),
    [anon_sym___thread] = ACTIONS(4172),
    [anon_sym_CG_EXTERN] = ACTIONS(4172),
    [anon_sym_CG_INLINE] = ACTIONS(4172),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4172),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4172),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4172),
    [anon_sym_IBOutlet] = ACTIONS(4172),
    [anon_sym_IBInspectable] = ACTIONS(4172),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4172),
    [anon_sym_NS_INLINE] = ACTIONS(4172),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4172),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4172),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4172),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4172),
    [anon_sym_const] = ACTIONS(4172),
    [anon_sym_constexpr] = ACTIONS(4172),
    [anon_sym_volatile] = ACTIONS(4172),
    [anon_sym_restrict] = ACTIONS(4172),
    [anon_sym___restrict__] = ACTIONS(4172),
    [anon_sym__Atomic] = ACTIONS(4172),
    [anon_sym__Noreturn] = ACTIONS(4172),
    [anon_sym__Nonnull] = ACTIONS(4172),
    [anon_sym_nullable] = ACTIONS(4172),
    [anon_sym__Complex] = ACTIONS(4172),
    [anon_sym__Nullable] = ACTIONS(4172),
    [anon_sym__Nullable_result] = ACTIONS(4172),
    [anon_sym__Null_unspecified] = ACTIONS(4172),
    [anon_sym___autoreleasing] = ACTIONS(4172),
    [anon_sym___block] = ACTIONS(4172),
    [anon_sym___bridge] = ACTIONS(4172),
    [anon_sym___bridge_retained] = ACTIONS(4172),
    [anon_sym___bridge_transfer] = ACTIONS(4172),
    [anon_sym___complex] = ACTIONS(4172),
    [anon_sym___const] = ACTIONS(4172),
    [anon_sym___imag] = ACTIONS(4172),
    [anon_sym___kindof] = ACTIONS(4172),
    [anon_sym___nonnull] = ACTIONS(4172),
    [anon_sym___nullable] = ACTIONS(4172),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4172),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4172),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4172),
    [anon_sym___real] = ACTIONS(4172),
    [anon_sym___strong] = ACTIONS(4172),
    [anon_sym___unsafe_unretained] = ACTIONS(4172),
    [anon_sym___unused] = ACTIONS(4172),
    [anon_sym___weak] = ACTIONS(4172),
    [anon_sym_alignas] = ACTIONS(4172),
    [anon_sym__Alignas] = ACTIONS(4172),
    [sym_primitive_type] = ACTIONS(4172),
    [anon_sym_enum] = ACTIONS(4172),
    [anon_sym_COLON] = ACTIONS(4174),
    [anon_sym_struct] = ACTIONS(4172),
    [anon_sym_union] = ACTIONS(4172),
    [anon_sym_QMARK] = ACTIONS(4174),
    [anon_sym_STAR_EQ] = ACTIONS(4174),
    [anon_sym_SLASH_EQ] = ACTIONS(4174),
    [anon_sym_PERCENT_EQ] = ACTIONS(4174),
    [anon_sym_PLUS_EQ] = ACTIONS(4174),
    [anon_sym_DASH_EQ] = ACTIONS(4174),
    [anon_sym_LT_LT_EQ] = ACTIONS(4174),
    [anon_sym_GT_GT_EQ] = ACTIONS(4174),
    [anon_sym_AMP_EQ] = ACTIONS(4174),
    [anon_sym_CARET_EQ] = ACTIONS(4174),
    [anon_sym_PIPE_EQ] = ACTIONS(4174),
    [anon_sym_DASH_DASH] = ACTIONS(4174),
    [anon_sym_PLUS_PLUS] = ACTIONS(4174),
    [anon_sym_DOT] = ACTIONS(4172),
    [anon_sym_DASH_GT] = ACTIONS(4174),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4172),
    [anon_sym___typeof] = ACTIONS(4172),
    [anon_sym_typeof] = ACTIONS(4172),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4172),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4172),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4172),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4172),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4172),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4172),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4172),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4172),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4172),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4172),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4172),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4172),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4172),
    [anon_sym_API_AVAILABLE] = ACTIONS(4172),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4172),
    [anon_sym_API_DEPRECATED] = ACTIONS(4172),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4172),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4172),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4172),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4172),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4172),
    [anon_sym___deprecated_msg] = ACTIONS(4172),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4172),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4172),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4172),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4172),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4172),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4172),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4172),
    [anon_sym_BOOL] = ACTIONS(4172),
    [anon_sym_IMP] = ACTIONS(4172),
    [anon_sym_SEL] = ACTIONS(4172),
    [anon_sym_Class] = ACTIONS(4172),
    [anon_sym_id] = ACTIONS(4172),
  },
  [STATE(1483)] = {
    [sym_identifier] = ACTIONS(4176),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4178),
    [anon_sym_COMMA] = ACTIONS(4178),
    [anon_sym_RPAREN] = ACTIONS(4178),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4178),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4178),
    [anon_sym_LPAREN2] = ACTIONS(4178),
    [anon_sym_DASH] = ACTIONS(4176),
    [anon_sym_PLUS] = ACTIONS(4176),
    [anon_sym_STAR] = ACTIONS(4176),
    [anon_sym_SLASH] = ACTIONS(4176),
    [anon_sym_PERCENT] = ACTIONS(4176),
    [anon_sym_PIPE_PIPE] = ACTIONS(4178),
    [anon_sym_AMP_AMP] = ACTIONS(4178),
    [anon_sym_PIPE] = ACTIONS(4176),
    [anon_sym_CARET] = ACTIONS(4176),
    [anon_sym_AMP] = ACTIONS(4176),
    [anon_sym_EQ_EQ] = ACTIONS(4178),
    [anon_sym_BANG_EQ] = ACTIONS(4178),
    [anon_sym_GT] = ACTIONS(4176),
    [anon_sym_GT_EQ] = ACTIONS(4178),
    [anon_sym_LT_EQ] = ACTIONS(4178),
    [anon_sym_LT] = ACTIONS(4176),
    [anon_sym_LT_LT] = ACTIONS(4176),
    [anon_sym_GT_GT] = ACTIONS(4176),
    [anon_sym_SEMI] = ACTIONS(4178),
    [anon_sym___extension__] = ACTIONS(4176),
    [anon_sym_extern] = ACTIONS(4176),
    [anon_sym___attribute__] = ACTIONS(4176),
    [anon_sym___attribute] = ACTIONS(4176),
    [anon_sym_noreturn] = ACTIONS(4176),
    [anon_sym_LBRACK] = ACTIONS(4178),
    [anon_sym_RBRACK] = ACTIONS(4178),
    [anon_sym___declspec] = ACTIONS(4176),
    [anon_sym_LBRACE] = ACTIONS(4178),
    [anon_sym_RBRACE] = ACTIONS(4178),
    [anon_sym_signed] = ACTIONS(4176),
    [anon_sym_unsigned] = ACTIONS(4176),
    [anon_sym_long] = ACTIONS(4176),
    [anon_sym_short] = ACTIONS(4176),
    [anon_sym_static] = ACTIONS(4176),
    [anon_sym_EQ] = ACTIONS(4176),
    [anon_sym_ATautoreleasepool] = ACTIONS(4178),
    [anon_sym_auto] = ACTIONS(4176),
    [anon_sym_register] = ACTIONS(4176),
    [anon_sym_inline] = ACTIONS(4176),
    [anon_sym___inline] = ACTIONS(4176),
    [anon_sym___inline__] = ACTIONS(4176),
    [anon_sym___forceinline] = ACTIONS(4176),
    [anon_sym_thread_local] = ACTIONS(4176),
    [anon_sym___thread] = ACTIONS(4176),
    [anon_sym_CG_EXTERN] = ACTIONS(4176),
    [anon_sym_CG_INLINE] = ACTIONS(4176),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4176),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4176),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4176),
    [anon_sym_IBOutlet] = ACTIONS(4176),
    [anon_sym_IBInspectable] = ACTIONS(4176),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4176),
    [anon_sym_NS_INLINE] = ACTIONS(4176),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4176),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4176),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4176),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4176),
    [anon_sym_const] = ACTIONS(4176),
    [anon_sym_constexpr] = ACTIONS(4176),
    [anon_sym_volatile] = ACTIONS(4176),
    [anon_sym_restrict] = ACTIONS(4176),
    [anon_sym___restrict__] = ACTIONS(4176),
    [anon_sym__Atomic] = ACTIONS(4176),
    [anon_sym__Noreturn] = ACTIONS(4176),
    [anon_sym__Nonnull] = ACTIONS(4176),
    [anon_sym_nullable] = ACTIONS(4176),
    [anon_sym__Complex] = ACTIONS(4176),
    [anon_sym__Nullable] = ACTIONS(4176),
    [anon_sym__Nullable_result] = ACTIONS(4176),
    [anon_sym__Null_unspecified] = ACTIONS(4176),
    [anon_sym___autoreleasing] = ACTIONS(4176),
    [anon_sym___block] = ACTIONS(4176),
    [anon_sym___bridge] = ACTIONS(4176),
    [anon_sym___bridge_retained] = ACTIONS(4176),
    [anon_sym___bridge_transfer] = ACTIONS(4176),
    [anon_sym___complex] = ACTIONS(4176),
    [anon_sym___const] = ACTIONS(4176),
    [anon_sym___imag] = ACTIONS(4176),
    [anon_sym___kindof] = ACTIONS(4176),
    [anon_sym___nonnull] = ACTIONS(4176),
    [anon_sym___nullable] = ACTIONS(4176),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4176),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4176),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4176),
    [anon_sym___real] = ACTIONS(4176),
    [anon_sym___strong] = ACTIONS(4176),
    [anon_sym___unsafe_unretained] = ACTIONS(4176),
    [anon_sym___unused] = ACTIONS(4176),
    [anon_sym___weak] = ACTIONS(4176),
    [anon_sym_alignas] = ACTIONS(4176),
    [anon_sym__Alignas] = ACTIONS(4176),
    [sym_primitive_type] = ACTIONS(4176),
    [anon_sym_enum] = ACTIONS(4176),
    [anon_sym_COLON] = ACTIONS(4178),
    [anon_sym_struct] = ACTIONS(4176),
    [anon_sym_union] = ACTIONS(4176),
    [anon_sym_QMARK] = ACTIONS(4178),
    [anon_sym_STAR_EQ] = ACTIONS(4178),
    [anon_sym_SLASH_EQ] = ACTIONS(4178),
    [anon_sym_PERCENT_EQ] = ACTIONS(4178),
    [anon_sym_PLUS_EQ] = ACTIONS(4178),
    [anon_sym_DASH_EQ] = ACTIONS(4178),
    [anon_sym_LT_LT_EQ] = ACTIONS(4178),
    [anon_sym_GT_GT_EQ] = ACTIONS(4178),
    [anon_sym_AMP_EQ] = ACTIONS(4178),
    [anon_sym_CARET_EQ] = ACTIONS(4178),
    [anon_sym_PIPE_EQ] = ACTIONS(4178),
    [anon_sym_DASH_DASH] = ACTIONS(4178),
    [anon_sym_PLUS_PLUS] = ACTIONS(4178),
    [anon_sym_DOT] = ACTIONS(4176),
    [anon_sym_DASH_GT] = ACTIONS(4178),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4176),
    [anon_sym___typeof] = ACTIONS(4176),
    [anon_sym_typeof] = ACTIONS(4176),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4176),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4176),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4176),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4176),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4176),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4176),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4176),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4176),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4176),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4176),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4176),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4176),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4176),
    [anon_sym_API_AVAILABLE] = ACTIONS(4176),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4176),
    [anon_sym_API_DEPRECATED] = ACTIONS(4176),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4176),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4176),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4176),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4176),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4176),
    [anon_sym___deprecated_msg] = ACTIONS(4176),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4176),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4176),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4176),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4176),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4176),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4176),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4176),
    [anon_sym_BOOL] = ACTIONS(4176),
    [anon_sym_IMP] = ACTIONS(4176),
    [anon_sym_SEL] = ACTIONS(4176),
    [anon_sym_Class] = ACTIONS(4176),
    [anon_sym_id] = ACTIONS(4176),
  },
  [STATE(1484)] = {
    [sym_identifier] = ACTIONS(4180),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4182),
    [anon_sym_COMMA] = ACTIONS(4182),
    [anon_sym_RPAREN] = ACTIONS(4182),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4182),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4182),
    [anon_sym_LPAREN2] = ACTIONS(4182),
    [anon_sym_DASH] = ACTIONS(4180),
    [anon_sym_PLUS] = ACTIONS(4180),
    [anon_sym_STAR] = ACTIONS(4180),
    [anon_sym_SLASH] = ACTIONS(4180),
    [anon_sym_PERCENT] = ACTIONS(4180),
    [anon_sym_PIPE_PIPE] = ACTIONS(4182),
    [anon_sym_AMP_AMP] = ACTIONS(4182),
    [anon_sym_PIPE] = ACTIONS(4180),
    [anon_sym_CARET] = ACTIONS(4180),
    [anon_sym_AMP] = ACTIONS(4180),
    [anon_sym_EQ_EQ] = ACTIONS(4182),
    [anon_sym_BANG_EQ] = ACTIONS(4182),
    [anon_sym_GT] = ACTIONS(4180),
    [anon_sym_GT_EQ] = ACTIONS(4182),
    [anon_sym_LT_EQ] = ACTIONS(4182),
    [anon_sym_LT] = ACTIONS(4180),
    [anon_sym_LT_LT] = ACTIONS(4180),
    [anon_sym_GT_GT] = ACTIONS(4180),
    [anon_sym_SEMI] = ACTIONS(4182),
    [anon_sym___extension__] = ACTIONS(4180),
    [anon_sym_extern] = ACTIONS(4180),
    [anon_sym___attribute__] = ACTIONS(4180),
    [anon_sym___attribute] = ACTIONS(4180),
    [anon_sym_noreturn] = ACTIONS(4180),
    [anon_sym_LBRACK] = ACTIONS(4182),
    [anon_sym_RBRACK] = ACTIONS(4182),
    [anon_sym___declspec] = ACTIONS(4180),
    [anon_sym_LBRACE] = ACTIONS(4182),
    [anon_sym_RBRACE] = ACTIONS(4182),
    [anon_sym_signed] = ACTIONS(4180),
    [anon_sym_unsigned] = ACTIONS(4180),
    [anon_sym_long] = ACTIONS(4180),
    [anon_sym_short] = ACTIONS(4180),
    [anon_sym_static] = ACTIONS(4180),
    [anon_sym_EQ] = ACTIONS(4180),
    [anon_sym_ATautoreleasepool] = ACTIONS(4182),
    [anon_sym_auto] = ACTIONS(4180),
    [anon_sym_register] = ACTIONS(4180),
    [anon_sym_inline] = ACTIONS(4180),
    [anon_sym___inline] = ACTIONS(4180),
    [anon_sym___inline__] = ACTIONS(4180),
    [anon_sym___forceinline] = ACTIONS(4180),
    [anon_sym_thread_local] = ACTIONS(4180),
    [anon_sym___thread] = ACTIONS(4180),
    [anon_sym_CG_EXTERN] = ACTIONS(4180),
    [anon_sym_CG_INLINE] = ACTIONS(4180),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4180),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4180),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4180),
    [anon_sym_IBOutlet] = ACTIONS(4180),
    [anon_sym_IBInspectable] = ACTIONS(4180),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4180),
    [anon_sym_NS_INLINE] = ACTIONS(4180),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4180),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4180),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4180),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4180),
    [anon_sym_const] = ACTIONS(4180),
    [anon_sym_constexpr] = ACTIONS(4180),
    [anon_sym_volatile] = ACTIONS(4180),
    [anon_sym_restrict] = ACTIONS(4180),
    [anon_sym___restrict__] = ACTIONS(4180),
    [anon_sym__Atomic] = ACTIONS(4180),
    [anon_sym__Noreturn] = ACTIONS(4180),
    [anon_sym__Nonnull] = ACTIONS(4180),
    [anon_sym_nullable] = ACTIONS(4180),
    [anon_sym__Complex] = ACTIONS(4180),
    [anon_sym__Nullable] = ACTIONS(4180),
    [anon_sym__Nullable_result] = ACTIONS(4180),
    [anon_sym__Null_unspecified] = ACTIONS(4180),
    [anon_sym___autoreleasing] = ACTIONS(4180),
    [anon_sym___block] = ACTIONS(4180),
    [anon_sym___bridge] = ACTIONS(4180),
    [anon_sym___bridge_retained] = ACTIONS(4180),
    [anon_sym___bridge_transfer] = ACTIONS(4180),
    [anon_sym___complex] = ACTIONS(4180),
    [anon_sym___const] = ACTIONS(4180),
    [anon_sym___imag] = ACTIONS(4180),
    [anon_sym___kindof] = ACTIONS(4180),
    [anon_sym___nonnull] = ACTIONS(4180),
    [anon_sym___nullable] = ACTIONS(4180),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4180),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4180),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4180),
    [anon_sym___real] = ACTIONS(4180),
    [anon_sym___strong] = ACTIONS(4180),
    [anon_sym___unsafe_unretained] = ACTIONS(4180),
    [anon_sym___unused] = ACTIONS(4180),
    [anon_sym___weak] = ACTIONS(4180),
    [anon_sym_alignas] = ACTIONS(4180),
    [anon_sym__Alignas] = ACTIONS(4180),
    [sym_primitive_type] = ACTIONS(4180),
    [anon_sym_enum] = ACTIONS(4180),
    [anon_sym_COLON] = ACTIONS(4182),
    [anon_sym_struct] = ACTIONS(4180),
    [anon_sym_union] = ACTIONS(4180),
    [anon_sym_QMARK] = ACTIONS(4182),
    [anon_sym_STAR_EQ] = ACTIONS(4182),
    [anon_sym_SLASH_EQ] = ACTIONS(4182),
    [anon_sym_PERCENT_EQ] = ACTIONS(4182),
    [anon_sym_PLUS_EQ] = ACTIONS(4182),
    [anon_sym_DASH_EQ] = ACTIONS(4182),
    [anon_sym_LT_LT_EQ] = ACTIONS(4182),
    [anon_sym_GT_GT_EQ] = ACTIONS(4182),
    [anon_sym_AMP_EQ] = ACTIONS(4182),
    [anon_sym_CARET_EQ] = ACTIONS(4182),
    [anon_sym_PIPE_EQ] = ACTIONS(4182),
    [anon_sym_DASH_DASH] = ACTIONS(4182),
    [anon_sym_PLUS_PLUS] = ACTIONS(4182),
    [anon_sym_DOT] = ACTIONS(4180),
    [anon_sym_DASH_GT] = ACTIONS(4182),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4180),
    [anon_sym___typeof] = ACTIONS(4180),
    [anon_sym_typeof] = ACTIONS(4180),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4180),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4180),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4180),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4180),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4180),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4180),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4180),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4180),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4180),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4180),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4180),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4180),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4180),
    [anon_sym_API_AVAILABLE] = ACTIONS(4180),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4180),
    [anon_sym_API_DEPRECATED] = ACTIONS(4180),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4180),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4180),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4180),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4180),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4180),
    [anon_sym___deprecated_msg] = ACTIONS(4180),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4180),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4180),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4180),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4180),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4180),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4180),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4180),
    [anon_sym_BOOL] = ACTIONS(4180),
    [anon_sym_IMP] = ACTIONS(4180),
    [anon_sym_SEL] = ACTIONS(4180),
    [anon_sym_Class] = ACTIONS(4180),
    [anon_sym_id] = ACTIONS(4180),
  },
  [STATE(1485)] = {
    [sym_identifier] = ACTIONS(4184),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4186),
    [anon_sym_COMMA] = ACTIONS(4186),
    [anon_sym_RPAREN] = ACTIONS(4186),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4186),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4186),
    [anon_sym_LPAREN2] = ACTIONS(4186),
    [anon_sym_DASH] = ACTIONS(4184),
    [anon_sym_PLUS] = ACTIONS(4184),
    [anon_sym_STAR] = ACTIONS(4184),
    [anon_sym_SLASH] = ACTIONS(4184),
    [anon_sym_PERCENT] = ACTIONS(4184),
    [anon_sym_PIPE_PIPE] = ACTIONS(4186),
    [anon_sym_AMP_AMP] = ACTIONS(4186),
    [anon_sym_PIPE] = ACTIONS(4184),
    [anon_sym_CARET] = ACTIONS(4184),
    [anon_sym_AMP] = ACTIONS(4184),
    [anon_sym_EQ_EQ] = ACTIONS(4186),
    [anon_sym_BANG_EQ] = ACTIONS(4186),
    [anon_sym_GT] = ACTIONS(4184),
    [anon_sym_GT_EQ] = ACTIONS(4186),
    [anon_sym_LT_EQ] = ACTIONS(4186),
    [anon_sym_LT] = ACTIONS(4184),
    [anon_sym_LT_LT] = ACTIONS(4184),
    [anon_sym_GT_GT] = ACTIONS(4184),
    [anon_sym_SEMI] = ACTIONS(4186),
    [anon_sym___extension__] = ACTIONS(4184),
    [anon_sym_extern] = ACTIONS(4184),
    [anon_sym___attribute__] = ACTIONS(4184),
    [anon_sym___attribute] = ACTIONS(4184),
    [anon_sym_noreturn] = ACTIONS(4184),
    [anon_sym_LBRACK] = ACTIONS(4186),
    [anon_sym_RBRACK] = ACTIONS(4186),
    [anon_sym___declspec] = ACTIONS(4184),
    [anon_sym_LBRACE] = ACTIONS(4186),
    [anon_sym_RBRACE] = ACTIONS(4186),
    [anon_sym_signed] = ACTIONS(4184),
    [anon_sym_unsigned] = ACTIONS(4184),
    [anon_sym_long] = ACTIONS(4184),
    [anon_sym_short] = ACTIONS(4184),
    [anon_sym_static] = ACTIONS(4184),
    [anon_sym_EQ] = ACTIONS(4184),
    [anon_sym_ATautoreleasepool] = ACTIONS(4186),
    [anon_sym_auto] = ACTIONS(4184),
    [anon_sym_register] = ACTIONS(4184),
    [anon_sym_inline] = ACTIONS(4184),
    [anon_sym___inline] = ACTIONS(4184),
    [anon_sym___inline__] = ACTIONS(4184),
    [anon_sym___forceinline] = ACTIONS(4184),
    [anon_sym_thread_local] = ACTIONS(4184),
    [anon_sym___thread] = ACTIONS(4184),
    [anon_sym_CG_EXTERN] = ACTIONS(4184),
    [anon_sym_CG_INLINE] = ACTIONS(4184),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4184),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4184),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4184),
    [anon_sym_IBOutlet] = ACTIONS(4184),
    [anon_sym_IBInspectable] = ACTIONS(4184),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4184),
    [anon_sym_NS_INLINE] = ACTIONS(4184),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4184),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4184),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4184),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4184),
    [anon_sym_const] = ACTIONS(4184),
    [anon_sym_constexpr] = ACTIONS(4184),
    [anon_sym_volatile] = ACTIONS(4184),
    [anon_sym_restrict] = ACTIONS(4184),
    [anon_sym___restrict__] = ACTIONS(4184),
    [anon_sym__Atomic] = ACTIONS(4184),
    [anon_sym__Noreturn] = ACTIONS(4184),
    [anon_sym__Nonnull] = ACTIONS(4184),
    [anon_sym_nullable] = ACTIONS(4184),
    [anon_sym__Complex] = ACTIONS(4184),
    [anon_sym__Nullable] = ACTIONS(4184),
    [anon_sym__Nullable_result] = ACTIONS(4184),
    [anon_sym__Null_unspecified] = ACTIONS(4184),
    [anon_sym___autoreleasing] = ACTIONS(4184),
    [anon_sym___block] = ACTIONS(4184),
    [anon_sym___bridge] = ACTIONS(4184),
    [anon_sym___bridge_retained] = ACTIONS(4184),
    [anon_sym___bridge_transfer] = ACTIONS(4184),
    [anon_sym___complex] = ACTIONS(4184),
    [anon_sym___const] = ACTIONS(4184),
    [anon_sym___imag] = ACTIONS(4184),
    [anon_sym___kindof] = ACTIONS(4184),
    [anon_sym___nonnull] = ACTIONS(4184),
    [anon_sym___nullable] = ACTIONS(4184),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4184),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4184),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4184),
    [anon_sym___real] = ACTIONS(4184),
    [anon_sym___strong] = ACTIONS(4184),
    [anon_sym___unsafe_unretained] = ACTIONS(4184),
    [anon_sym___unused] = ACTIONS(4184),
    [anon_sym___weak] = ACTIONS(4184),
    [anon_sym_alignas] = ACTIONS(4184),
    [anon_sym__Alignas] = ACTIONS(4184),
    [sym_primitive_type] = ACTIONS(4184),
    [anon_sym_enum] = ACTIONS(4184),
    [anon_sym_COLON] = ACTIONS(4186),
    [anon_sym_struct] = ACTIONS(4184),
    [anon_sym_union] = ACTIONS(4184),
    [anon_sym_QMARK] = ACTIONS(4186),
    [anon_sym_STAR_EQ] = ACTIONS(4186),
    [anon_sym_SLASH_EQ] = ACTIONS(4186),
    [anon_sym_PERCENT_EQ] = ACTIONS(4186),
    [anon_sym_PLUS_EQ] = ACTIONS(4186),
    [anon_sym_DASH_EQ] = ACTIONS(4186),
    [anon_sym_LT_LT_EQ] = ACTIONS(4186),
    [anon_sym_GT_GT_EQ] = ACTIONS(4186),
    [anon_sym_AMP_EQ] = ACTIONS(4186),
    [anon_sym_CARET_EQ] = ACTIONS(4186),
    [anon_sym_PIPE_EQ] = ACTIONS(4186),
    [anon_sym_DASH_DASH] = ACTIONS(4186),
    [anon_sym_PLUS_PLUS] = ACTIONS(4186),
    [anon_sym_DOT] = ACTIONS(4184),
    [anon_sym_DASH_GT] = ACTIONS(4186),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4184),
    [anon_sym___typeof] = ACTIONS(4184),
    [anon_sym_typeof] = ACTIONS(4184),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4184),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4184),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4184),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4184),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4184),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4184),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4184),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4184),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4184),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4184),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4184),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4184),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4184),
    [anon_sym_API_AVAILABLE] = ACTIONS(4184),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4184),
    [anon_sym_API_DEPRECATED] = ACTIONS(4184),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4184),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4184),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4184),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4184),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4184),
    [anon_sym___deprecated_msg] = ACTIONS(4184),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4184),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4184),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4184),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4184),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4184),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4184),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4184),
    [anon_sym_BOOL] = ACTIONS(4184),
    [anon_sym_IMP] = ACTIONS(4184),
    [anon_sym_SEL] = ACTIONS(4184),
    [anon_sym_Class] = ACTIONS(4184),
    [anon_sym_id] = ACTIONS(4184),
  },
  [STATE(1486)] = {
    [sym_identifier] = ACTIONS(4188),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4190),
    [anon_sym_COMMA] = ACTIONS(4190),
    [anon_sym_RPAREN] = ACTIONS(4190),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4190),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4190),
    [anon_sym_LPAREN2] = ACTIONS(4190),
    [anon_sym_DASH] = ACTIONS(4188),
    [anon_sym_PLUS] = ACTIONS(4188),
    [anon_sym_STAR] = ACTIONS(4188),
    [anon_sym_SLASH] = ACTIONS(4188),
    [anon_sym_PERCENT] = ACTIONS(4188),
    [anon_sym_PIPE_PIPE] = ACTIONS(4190),
    [anon_sym_AMP_AMP] = ACTIONS(4190),
    [anon_sym_PIPE] = ACTIONS(4188),
    [anon_sym_CARET] = ACTIONS(4188),
    [anon_sym_AMP] = ACTIONS(4188),
    [anon_sym_EQ_EQ] = ACTIONS(4190),
    [anon_sym_BANG_EQ] = ACTIONS(4190),
    [anon_sym_GT] = ACTIONS(4188),
    [anon_sym_GT_EQ] = ACTIONS(4190),
    [anon_sym_LT_EQ] = ACTIONS(4190),
    [anon_sym_LT] = ACTIONS(4188),
    [anon_sym_LT_LT] = ACTIONS(4188),
    [anon_sym_GT_GT] = ACTIONS(4188),
    [anon_sym_SEMI] = ACTIONS(4190),
    [anon_sym___extension__] = ACTIONS(4188),
    [anon_sym_extern] = ACTIONS(4188),
    [anon_sym___attribute__] = ACTIONS(4188),
    [anon_sym___attribute] = ACTIONS(4188),
    [anon_sym_noreturn] = ACTIONS(4188),
    [anon_sym_LBRACK] = ACTIONS(4190),
    [anon_sym_RBRACK] = ACTIONS(4190),
    [anon_sym___declspec] = ACTIONS(4188),
    [anon_sym_LBRACE] = ACTIONS(4190),
    [anon_sym_RBRACE] = ACTIONS(4190),
    [anon_sym_signed] = ACTIONS(4188),
    [anon_sym_unsigned] = ACTIONS(4188),
    [anon_sym_long] = ACTIONS(4188),
    [anon_sym_short] = ACTIONS(4188),
    [anon_sym_static] = ACTIONS(4188),
    [anon_sym_EQ] = ACTIONS(4188),
    [anon_sym_ATautoreleasepool] = ACTIONS(4190),
    [anon_sym_auto] = ACTIONS(4188),
    [anon_sym_register] = ACTIONS(4188),
    [anon_sym_inline] = ACTIONS(4188),
    [anon_sym___inline] = ACTIONS(4188),
    [anon_sym___inline__] = ACTIONS(4188),
    [anon_sym___forceinline] = ACTIONS(4188),
    [anon_sym_thread_local] = ACTIONS(4188),
    [anon_sym___thread] = ACTIONS(4188),
    [anon_sym_CG_EXTERN] = ACTIONS(4188),
    [anon_sym_CG_INLINE] = ACTIONS(4188),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4188),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4188),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4188),
    [anon_sym_IBOutlet] = ACTIONS(4188),
    [anon_sym_IBInspectable] = ACTIONS(4188),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4188),
    [anon_sym_NS_INLINE] = ACTIONS(4188),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4188),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4188),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4188),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4188),
    [anon_sym_const] = ACTIONS(4188),
    [anon_sym_constexpr] = ACTIONS(4188),
    [anon_sym_volatile] = ACTIONS(4188),
    [anon_sym_restrict] = ACTIONS(4188),
    [anon_sym___restrict__] = ACTIONS(4188),
    [anon_sym__Atomic] = ACTIONS(4188),
    [anon_sym__Noreturn] = ACTIONS(4188),
    [anon_sym__Nonnull] = ACTIONS(4188),
    [anon_sym_nullable] = ACTIONS(4188),
    [anon_sym__Complex] = ACTIONS(4188),
    [anon_sym__Nullable] = ACTIONS(4188),
    [anon_sym__Nullable_result] = ACTIONS(4188),
    [anon_sym__Null_unspecified] = ACTIONS(4188),
    [anon_sym___autoreleasing] = ACTIONS(4188),
    [anon_sym___block] = ACTIONS(4188),
    [anon_sym___bridge] = ACTIONS(4188),
    [anon_sym___bridge_retained] = ACTIONS(4188),
    [anon_sym___bridge_transfer] = ACTIONS(4188),
    [anon_sym___complex] = ACTIONS(4188),
    [anon_sym___const] = ACTIONS(4188),
    [anon_sym___imag] = ACTIONS(4188),
    [anon_sym___kindof] = ACTIONS(4188),
    [anon_sym___nonnull] = ACTIONS(4188),
    [anon_sym___nullable] = ACTIONS(4188),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4188),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4188),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4188),
    [anon_sym___real] = ACTIONS(4188),
    [anon_sym___strong] = ACTIONS(4188),
    [anon_sym___unsafe_unretained] = ACTIONS(4188),
    [anon_sym___unused] = ACTIONS(4188),
    [anon_sym___weak] = ACTIONS(4188),
    [anon_sym_alignas] = ACTIONS(4188),
    [anon_sym__Alignas] = ACTIONS(4188),
    [sym_primitive_type] = ACTIONS(4188),
    [anon_sym_enum] = ACTIONS(4188),
    [anon_sym_COLON] = ACTIONS(4190),
    [anon_sym_struct] = ACTIONS(4188),
    [anon_sym_union] = ACTIONS(4188),
    [anon_sym_QMARK] = ACTIONS(4190),
    [anon_sym_STAR_EQ] = ACTIONS(4190),
    [anon_sym_SLASH_EQ] = ACTIONS(4190),
    [anon_sym_PERCENT_EQ] = ACTIONS(4190),
    [anon_sym_PLUS_EQ] = ACTIONS(4190),
    [anon_sym_DASH_EQ] = ACTIONS(4190),
    [anon_sym_LT_LT_EQ] = ACTIONS(4190),
    [anon_sym_GT_GT_EQ] = ACTIONS(4190),
    [anon_sym_AMP_EQ] = ACTIONS(4190),
    [anon_sym_CARET_EQ] = ACTIONS(4190),
    [anon_sym_PIPE_EQ] = ACTIONS(4190),
    [anon_sym_DASH_DASH] = ACTIONS(4190),
    [anon_sym_PLUS_PLUS] = ACTIONS(4190),
    [anon_sym_DOT] = ACTIONS(4188),
    [anon_sym_DASH_GT] = ACTIONS(4190),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4188),
    [anon_sym___typeof] = ACTIONS(4188),
    [anon_sym_typeof] = ACTIONS(4188),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4188),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4188),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4188),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4188),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4188),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4188),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4188),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4188),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4188),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4188),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4188),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4188),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4188),
    [anon_sym_API_AVAILABLE] = ACTIONS(4188),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4188),
    [anon_sym_API_DEPRECATED] = ACTIONS(4188),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4188),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4188),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4188),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4188),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4188),
    [anon_sym___deprecated_msg] = ACTIONS(4188),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4188),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4188),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4188),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4188),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4188),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4188),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4188),
    [anon_sym_BOOL] = ACTIONS(4188),
    [anon_sym_IMP] = ACTIONS(4188),
    [anon_sym_SEL] = ACTIONS(4188),
    [anon_sym_Class] = ACTIONS(4188),
    [anon_sym_id] = ACTIONS(4188),
  },
  [STATE(1487)] = {
    [sym_identifier] = ACTIONS(4192),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4194),
    [anon_sym_COMMA] = ACTIONS(4194),
    [anon_sym_RPAREN] = ACTIONS(4194),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4194),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4194),
    [anon_sym_LPAREN2] = ACTIONS(4194),
    [anon_sym_DASH] = ACTIONS(4192),
    [anon_sym_PLUS] = ACTIONS(4192),
    [anon_sym_STAR] = ACTIONS(4192),
    [anon_sym_SLASH] = ACTIONS(4192),
    [anon_sym_PERCENT] = ACTIONS(4192),
    [anon_sym_PIPE_PIPE] = ACTIONS(4194),
    [anon_sym_AMP_AMP] = ACTIONS(4194),
    [anon_sym_PIPE] = ACTIONS(4192),
    [anon_sym_CARET] = ACTIONS(4192),
    [anon_sym_AMP] = ACTIONS(4192),
    [anon_sym_EQ_EQ] = ACTIONS(4194),
    [anon_sym_BANG_EQ] = ACTIONS(4194),
    [anon_sym_GT] = ACTIONS(4192),
    [anon_sym_GT_EQ] = ACTIONS(4194),
    [anon_sym_LT_EQ] = ACTIONS(4194),
    [anon_sym_LT] = ACTIONS(4192),
    [anon_sym_LT_LT] = ACTIONS(4192),
    [anon_sym_GT_GT] = ACTIONS(4192),
    [anon_sym_SEMI] = ACTIONS(4194),
    [anon_sym___extension__] = ACTIONS(4192),
    [anon_sym_extern] = ACTIONS(4192),
    [anon_sym___attribute__] = ACTIONS(4192),
    [anon_sym___attribute] = ACTIONS(4192),
    [anon_sym_noreturn] = ACTIONS(4192),
    [anon_sym_LBRACK] = ACTIONS(4194),
    [anon_sym_RBRACK] = ACTIONS(4194),
    [anon_sym___declspec] = ACTIONS(4192),
    [anon_sym_LBRACE] = ACTIONS(4194),
    [anon_sym_RBRACE] = ACTIONS(4194),
    [anon_sym_signed] = ACTIONS(4192),
    [anon_sym_unsigned] = ACTIONS(4192),
    [anon_sym_long] = ACTIONS(4192),
    [anon_sym_short] = ACTIONS(4192),
    [anon_sym_static] = ACTIONS(4192),
    [anon_sym_EQ] = ACTIONS(4192),
    [anon_sym_ATautoreleasepool] = ACTIONS(4194),
    [anon_sym_auto] = ACTIONS(4192),
    [anon_sym_register] = ACTIONS(4192),
    [anon_sym_inline] = ACTIONS(4192),
    [anon_sym___inline] = ACTIONS(4192),
    [anon_sym___inline__] = ACTIONS(4192),
    [anon_sym___forceinline] = ACTIONS(4192),
    [anon_sym_thread_local] = ACTIONS(4192),
    [anon_sym___thread] = ACTIONS(4192),
    [anon_sym_CG_EXTERN] = ACTIONS(4192),
    [anon_sym_CG_INLINE] = ACTIONS(4192),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4192),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4192),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4192),
    [anon_sym_IBOutlet] = ACTIONS(4192),
    [anon_sym_IBInspectable] = ACTIONS(4192),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4192),
    [anon_sym_NS_INLINE] = ACTIONS(4192),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4192),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4192),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4192),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4192),
    [anon_sym_const] = ACTIONS(4192),
    [anon_sym_constexpr] = ACTIONS(4192),
    [anon_sym_volatile] = ACTIONS(4192),
    [anon_sym_restrict] = ACTIONS(4192),
    [anon_sym___restrict__] = ACTIONS(4192),
    [anon_sym__Atomic] = ACTIONS(4192),
    [anon_sym__Noreturn] = ACTIONS(4192),
    [anon_sym__Nonnull] = ACTIONS(4192),
    [anon_sym_nullable] = ACTIONS(4192),
    [anon_sym__Complex] = ACTIONS(4192),
    [anon_sym__Nullable] = ACTIONS(4192),
    [anon_sym__Nullable_result] = ACTIONS(4192),
    [anon_sym__Null_unspecified] = ACTIONS(4192),
    [anon_sym___autoreleasing] = ACTIONS(4192),
    [anon_sym___block] = ACTIONS(4192),
    [anon_sym___bridge] = ACTIONS(4192),
    [anon_sym___bridge_retained] = ACTIONS(4192),
    [anon_sym___bridge_transfer] = ACTIONS(4192),
    [anon_sym___complex] = ACTIONS(4192),
    [anon_sym___const] = ACTIONS(4192),
    [anon_sym___imag] = ACTIONS(4192),
    [anon_sym___kindof] = ACTIONS(4192),
    [anon_sym___nonnull] = ACTIONS(4192),
    [anon_sym___nullable] = ACTIONS(4192),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4192),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4192),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4192),
    [anon_sym___real] = ACTIONS(4192),
    [anon_sym___strong] = ACTIONS(4192),
    [anon_sym___unsafe_unretained] = ACTIONS(4192),
    [anon_sym___unused] = ACTIONS(4192),
    [anon_sym___weak] = ACTIONS(4192),
    [anon_sym_alignas] = ACTIONS(4192),
    [anon_sym__Alignas] = ACTIONS(4192),
    [sym_primitive_type] = ACTIONS(4192),
    [anon_sym_enum] = ACTIONS(4192),
    [anon_sym_COLON] = ACTIONS(4194),
    [anon_sym_struct] = ACTIONS(4192),
    [anon_sym_union] = ACTIONS(4192),
    [anon_sym_QMARK] = ACTIONS(4194),
    [anon_sym_STAR_EQ] = ACTIONS(4194),
    [anon_sym_SLASH_EQ] = ACTIONS(4194),
    [anon_sym_PERCENT_EQ] = ACTIONS(4194),
    [anon_sym_PLUS_EQ] = ACTIONS(4194),
    [anon_sym_DASH_EQ] = ACTIONS(4194),
    [anon_sym_LT_LT_EQ] = ACTIONS(4194),
    [anon_sym_GT_GT_EQ] = ACTIONS(4194),
    [anon_sym_AMP_EQ] = ACTIONS(4194),
    [anon_sym_CARET_EQ] = ACTIONS(4194),
    [anon_sym_PIPE_EQ] = ACTIONS(4194),
    [anon_sym_DASH_DASH] = ACTIONS(4194),
    [anon_sym_PLUS_PLUS] = ACTIONS(4194),
    [anon_sym_DOT] = ACTIONS(4192),
    [anon_sym_DASH_GT] = ACTIONS(4194),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4192),
    [anon_sym___typeof] = ACTIONS(4192),
    [anon_sym_typeof] = ACTIONS(4192),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4192),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4192),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4192),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4192),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4192),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4192),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4192),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4192),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4192),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4192),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4192),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4192),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4192),
    [anon_sym_API_AVAILABLE] = ACTIONS(4192),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4192),
    [anon_sym_API_DEPRECATED] = ACTIONS(4192),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4192),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4192),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4192),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4192),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4192),
    [anon_sym___deprecated_msg] = ACTIONS(4192),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4192),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4192),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4192),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4192),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4192),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4192),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4192),
    [anon_sym_BOOL] = ACTIONS(4192),
    [anon_sym_IMP] = ACTIONS(4192),
    [anon_sym_SEL] = ACTIONS(4192),
    [anon_sym_Class] = ACTIONS(4192),
    [anon_sym_id] = ACTIONS(4192),
  },
  [STATE(1488)] = {
    [sym_identifier] = ACTIONS(4196),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4198),
    [anon_sym_COMMA] = ACTIONS(4198),
    [anon_sym_RPAREN] = ACTIONS(4198),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4198),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4198),
    [anon_sym_LPAREN2] = ACTIONS(4198),
    [anon_sym_DASH] = ACTIONS(4196),
    [anon_sym_PLUS] = ACTIONS(4196),
    [anon_sym_STAR] = ACTIONS(4196),
    [anon_sym_SLASH] = ACTIONS(4196),
    [anon_sym_PERCENT] = ACTIONS(4196),
    [anon_sym_PIPE_PIPE] = ACTIONS(4198),
    [anon_sym_AMP_AMP] = ACTIONS(4198),
    [anon_sym_PIPE] = ACTIONS(4196),
    [anon_sym_CARET] = ACTIONS(4196),
    [anon_sym_AMP] = ACTIONS(4196),
    [anon_sym_EQ_EQ] = ACTIONS(4198),
    [anon_sym_BANG_EQ] = ACTIONS(4198),
    [anon_sym_GT] = ACTIONS(4196),
    [anon_sym_GT_EQ] = ACTIONS(4198),
    [anon_sym_LT_EQ] = ACTIONS(4198),
    [anon_sym_LT] = ACTIONS(4196),
    [anon_sym_LT_LT] = ACTIONS(4196),
    [anon_sym_GT_GT] = ACTIONS(4196),
    [anon_sym_SEMI] = ACTIONS(4198),
    [anon_sym___extension__] = ACTIONS(4196),
    [anon_sym_extern] = ACTIONS(4196),
    [anon_sym___attribute__] = ACTIONS(4196),
    [anon_sym___attribute] = ACTIONS(4196),
    [anon_sym_noreturn] = ACTIONS(4196),
    [anon_sym_LBRACK] = ACTIONS(4198),
    [anon_sym_RBRACK] = ACTIONS(4198),
    [anon_sym___declspec] = ACTIONS(4196),
    [anon_sym_LBRACE] = ACTIONS(4198),
    [anon_sym_RBRACE] = ACTIONS(4198),
    [anon_sym_signed] = ACTIONS(4196),
    [anon_sym_unsigned] = ACTIONS(4196),
    [anon_sym_long] = ACTIONS(4196),
    [anon_sym_short] = ACTIONS(4196),
    [anon_sym_static] = ACTIONS(4196),
    [anon_sym_EQ] = ACTIONS(4196),
    [anon_sym_ATautoreleasepool] = ACTIONS(4198),
    [anon_sym_auto] = ACTIONS(4196),
    [anon_sym_register] = ACTIONS(4196),
    [anon_sym_inline] = ACTIONS(4196),
    [anon_sym___inline] = ACTIONS(4196),
    [anon_sym___inline__] = ACTIONS(4196),
    [anon_sym___forceinline] = ACTIONS(4196),
    [anon_sym_thread_local] = ACTIONS(4196),
    [anon_sym___thread] = ACTIONS(4196),
    [anon_sym_CG_EXTERN] = ACTIONS(4196),
    [anon_sym_CG_INLINE] = ACTIONS(4196),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4196),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4196),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4196),
    [anon_sym_IBOutlet] = ACTIONS(4196),
    [anon_sym_IBInspectable] = ACTIONS(4196),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4196),
    [anon_sym_NS_INLINE] = ACTIONS(4196),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4196),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4196),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4196),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4196),
    [anon_sym_const] = ACTIONS(4196),
    [anon_sym_constexpr] = ACTIONS(4196),
    [anon_sym_volatile] = ACTIONS(4196),
    [anon_sym_restrict] = ACTIONS(4196),
    [anon_sym___restrict__] = ACTIONS(4196),
    [anon_sym__Atomic] = ACTIONS(4196),
    [anon_sym__Noreturn] = ACTIONS(4196),
    [anon_sym__Nonnull] = ACTIONS(4196),
    [anon_sym_nullable] = ACTIONS(4196),
    [anon_sym__Complex] = ACTIONS(4196),
    [anon_sym__Nullable] = ACTIONS(4196),
    [anon_sym__Nullable_result] = ACTIONS(4196),
    [anon_sym__Null_unspecified] = ACTIONS(4196),
    [anon_sym___autoreleasing] = ACTIONS(4196),
    [anon_sym___block] = ACTIONS(4196),
    [anon_sym___bridge] = ACTIONS(4196),
    [anon_sym___bridge_retained] = ACTIONS(4196),
    [anon_sym___bridge_transfer] = ACTIONS(4196),
    [anon_sym___complex] = ACTIONS(4196),
    [anon_sym___const] = ACTIONS(4196),
    [anon_sym___imag] = ACTIONS(4196),
    [anon_sym___kindof] = ACTIONS(4196),
    [anon_sym___nonnull] = ACTIONS(4196),
    [anon_sym___nullable] = ACTIONS(4196),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4196),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4196),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4196),
    [anon_sym___real] = ACTIONS(4196),
    [anon_sym___strong] = ACTIONS(4196),
    [anon_sym___unsafe_unretained] = ACTIONS(4196),
    [anon_sym___unused] = ACTIONS(4196),
    [anon_sym___weak] = ACTIONS(4196),
    [anon_sym_alignas] = ACTIONS(4196),
    [anon_sym__Alignas] = ACTIONS(4196),
    [sym_primitive_type] = ACTIONS(4196),
    [anon_sym_enum] = ACTIONS(4196),
    [anon_sym_COLON] = ACTIONS(4198),
    [anon_sym_struct] = ACTIONS(4196),
    [anon_sym_union] = ACTIONS(4196),
    [anon_sym_QMARK] = ACTIONS(4198),
    [anon_sym_STAR_EQ] = ACTIONS(4198),
    [anon_sym_SLASH_EQ] = ACTIONS(4198),
    [anon_sym_PERCENT_EQ] = ACTIONS(4198),
    [anon_sym_PLUS_EQ] = ACTIONS(4198),
    [anon_sym_DASH_EQ] = ACTIONS(4198),
    [anon_sym_LT_LT_EQ] = ACTIONS(4198),
    [anon_sym_GT_GT_EQ] = ACTIONS(4198),
    [anon_sym_AMP_EQ] = ACTIONS(4198),
    [anon_sym_CARET_EQ] = ACTIONS(4198),
    [anon_sym_PIPE_EQ] = ACTIONS(4198),
    [anon_sym_DASH_DASH] = ACTIONS(4198),
    [anon_sym_PLUS_PLUS] = ACTIONS(4198),
    [anon_sym_DOT] = ACTIONS(4196),
    [anon_sym_DASH_GT] = ACTIONS(4198),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4196),
    [anon_sym___typeof] = ACTIONS(4196),
    [anon_sym_typeof] = ACTIONS(4196),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4196),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4196),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4196),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4196),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4196),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4196),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4196),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4196),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4196),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4196),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4196),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4196),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4196),
    [anon_sym_API_AVAILABLE] = ACTIONS(4196),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4196),
    [anon_sym_API_DEPRECATED] = ACTIONS(4196),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4196),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4196),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4196),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4196),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4196),
    [anon_sym___deprecated_msg] = ACTIONS(4196),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4196),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4196),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4196),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4196),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4196),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4196),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4196),
    [anon_sym_BOOL] = ACTIONS(4196),
    [anon_sym_IMP] = ACTIONS(4196),
    [anon_sym_SEL] = ACTIONS(4196),
    [anon_sym_Class] = ACTIONS(4196),
    [anon_sym_id] = ACTIONS(4196),
  },
  [STATE(1489)] = {
    [sym_identifier] = ACTIONS(4200),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4202),
    [anon_sym_COMMA] = ACTIONS(4202),
    [anon_sym_RPAREN] = ACTIONS(4202),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4202),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4202),
    [anon_sym_LPAREN2] = ACTIONS(4202),
    [anon_sym_DASH] = ACTIONS(4200),
    [anon_sym_PLUS] = ACTIONS(4200),
    [anon_sym_STAR] = ACTIONS(4200),
    [anon_sym_SLASH] = ACTIONS(4200),
    [anon_sym_PERCENT] = ACTIONS(4200),
    [anon_sym_PIPE_PIPE] = ACTIONS(4202),
    [anon_sym_AMP_AMP] = ACTIONS(4202),
    [anon_sym_PIPE] = ACTIONS(4200),
    [anon_sym_CARET] = ACTIONS(4200),
    [anon_sym_AMP] = ACTIONS(4200),
    [anon_sym_EQ_EQ] = ACTIONS(4202),
    [anon_sym_BANG_EQ] = ACTIONS(4202),
    [anon_sym_GT] = ACTIONS(4200),
    [anon_sym_GT_EQ] = ACTIONS(4202),
    [anon_sym_LT_EQ] = ACTIONS(4202),
    [anon_sym_LT] = ACTIONS(4200),
    [anon_sym_LT_LT] = ACTIONS(4200),
    [anon_sym_GT_GT] = ACTIONS(4200),
    [anon_sym_SEMI] = ACTIONS(4202),
    [anon_sym___extension__] = ACTIONS(4200),
    [anon_sym_extern] = ACTIONS(4200),
    [anon_sym___attribute__] = ACTIONS(4200),
    [anon_sym___attribute] = ACTIONS(4200),
    [anon_sym_noreturn] = ACTIONS(4200),
    [anon_sym_LBRACK] = ACTIONS(4202),
    [anon_sym_RBRACK] = ACTIONS(4202),
    [anon_sym___declspec] = ACTIONS(4200),
    [anon_sym_LBRACE] = ACTIONS(4202),
    [anon_sym_RBRACE] = ACTIONS(4202),
    [anon_sym_signed] = ACTIONS(4200),
    [anon_sym_unsigned] = ACTIONS(4200),
    [anon_sym_long] = ACTIONS(4200),
    [anon_sym_short] = ACTIONS(4200),
    [anon_sym_static] = ACTIONS(4200),
    [anon_sym_EQ] = ACTIONS(4200),
    [anon_sym_ATautoreleasepool] = ACTIONS(4202),
    [anon_sym_auto] = ACTIONS(4200),
    [anon_sym_register] = ACTIONS(4200),
    [anon_sym_inline] = ACTIONS(4200),
    [anon_sym___inline] = ACTIONS(4200),
    [anon_sym___inline__] = ACTIONS(4200),
    [anon_sym___forceinline] = ACTIONS(4200),
    [anon_sym_thread_local] = ACTIONS(4200),
    [anon_sym___thread] = ACTIONS(4200),
    [anon_sym_CG_EXTERN] = ACTIONS(4200),
    [anon_sym_CG_INLINE] = ACTIONS(4200),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4200),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4200),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4200),
    [anon_sym_IBOutlet] = ACTIONS(4200),
    [anon_sym_IBInspectable] = ACTIONS(4200),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4200),
    [anon_sym_NS_INLINE] = ACTIONS(4200),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4200),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4200),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4200),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4200),
    [anon_sym_const] = ACTIONS(4200),
    [anon_sym_constexpr] = ACTIONS(4200),
    [anon_sym_volatile] = ACTIONS(4200),
    [anon_sym_restrict] = ACTIONS(4200),
    [anon_sym___restrict__] = ACTIONS(4200),
    [anon_sym__Atomic] = ACTIONS(4200),
    [anon_sym__Noreturn] = ACTIONS(4200),
    [anon_sym__Nonnull] = ACTIONS(4200),
    [anon_sym_nullable] = ACTIONS(4200),
    [anon_sym__Complex] = ACTIONS(4200),
    [anon_sym__Nullable] = ACTIONS(4200),
    [anon_sym__Nullable_result] = ACTIONS(4200),
    [anon_sym__Null_unspecified] = ACTIONS(4200),
    [anon_sym___autoreleasing] = ACTIONS(4200),
    [anon_sym___block] = ACTIONS(4200),
    [anon_sym___bridge] = ACTIONS(4200),
    [anon_sym___bridge_retained] = ACTIONS(4200),
    [anon_sym___bridge_transfer] = ACTIONS(4200),
    [anon_sym___complex] = ACTIONS(4200),
    [anon_sym___const] = ACTIONS(4200),
    [anon_sym___imag] = ACTIONS(4200),
    [anon_sym___kindof] = ACTIONS(4200),
    [anon_sym___nonnull] = ACTIONS(4200),
    [anon_sym___nullable] = ACTIONS(4200),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4200),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4200),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4200),
    [anon_sym___real] = ACTIONS(4200),
    [anon_sym___strong] = ACTIONS(4200),
    [anon_sym___unsafe_unretained] = ACTIONS(4200),
    [anon_sym___unused] = ACTIONS(4200),
    [anon_sym___weak] = ACTIONS(4200),
    [anon_sym_alignas] = ACTIONS(4200),
    [anon_sym__Alignas] = ACTIONS(4200),
    [sym_primitive_type] = ACTIONS(4200),
    [anon_sym_enum] = ACTIONS(4200),
    [anon_sym_COLON] = ACTIONS(4202),
    [anon_sym_struct] = ACTIONS(4200),
    [anon_sym_union] = ACTIONS(4200),
    [anon_sym_QMARK] = ACTIONS(4202),
    [anon_sym_STAR_EQ] = ACTIONS(4202),
    [anon_sym_SLASH_EQ] = ACTIONS(4202),
    [anon_sym_PERCENT_EQ] = ACTIONS(4202),
    [anon_sym_PLUS_EQ] = ACTIONS(4202),
    [anon_sym_DASH_EQ] = ACTIONS(4202),
    [anon_sym_LT_LT_EQ] = ACTIONS(4202),
    [anon_sym_GT_GT_EQ] = ACTIONS(4202),
    [anon_sym_AMP_EQ] = ACTIONS(4202),
    [anon_sym_CARET_EQ] = ACTIONS(4202),
    [anon_sym_PIPE_EQ] = ACTIONS(4202),
    [anon_sym_DASH_DASH] = ACTIONS(4202),
    [anon_sym_PLUS_PLUS] = ACTIONS(4202),
    [anon_sym_DOT] = ACTIONS(4200),
    [anon_sym_DASH_GT] = ACTIONS(4202),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4200),
    [anon_sym___typeof] = ACTIONS(4200),
    [anon_sym_typeof] = ACTIONS(4200),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4200),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4200),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4200),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4200),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4200),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4200),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4200),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4200),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4200),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4200),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4200),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4200),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4200),
    [anon_sym_API_AVAILABLE] = ACTIONS(4200),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4200),
    [anon_sym_API_DEPRECATED] = ACTIONS(4200),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4200),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4200),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4200),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4200),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4200),
    [anon_sym___deprecated_msg] = ACTIONS(4200),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4200),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4200),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4200),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4200),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4200),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4200),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4200),
    [anon_sym_BOOL] = ACTIONS(4200),
    [anon_sym_IMP] = ACTIONS(4200),
    [anon_sym_SEL] = ACTIONS(4200),
    [anon_sym_Class] = ACTIONS(4200),
    [anon_sym_id] = ACTIONS(4200),
  },
  [STATE(1490)] = {
    [sym_identifier] = ACTIONS(4204),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4206),
    [anon_sym_COMMA] = ACTIONS(4206),
    [anon_sym_RPAREN] = ACTIONS(4206),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4206),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4206),
    [anon_sym_LPAREN2] = ACTIONS(4206),
    [anon_sym_DASH] = ACTIONS(4204),
    [anon_sym_PLUS] = ACTIONS(4204),
    [anon_sym_STAR] = ACTIONS(4204),
    [anon_sym_SLASH] = ACTIONS(4204),
    [anon_sym_PERCENT] = ACTIONS(4204),
    [anon_sym_PIPE_PIPE] = ACTIONS(4206),
    [anon_sym_AMP_AMP] = ACTIONS(4206),
    [anon_sym_PIPE] = ACTIONS(4204),
    [anon_sym_CARET] = ACTIONS(4204),
    [anon_sym_AMP] = ACTIONS(4204),
    [anon_sym_EQ_EQ] = ACTIONS(4206),
    [anon_sym_BANG_EQ] = ACTIONS(4206),
    [anon_sym_GT] = ACTIONS(4204),
    [anon_sym_GT_EQ] = ACTIONS(4206),
    [anon_sym_LT_EQ] = ACTIONS(4206),
    [anon_sym_LT] = ACTIONS(4204),
    [anon_sym_LT_LT] = ACTIONS(4204),
    [anon_sym_GT_GT] = ACTIONS(4204),
    [anon_sym_SEMI] = ACTIONS(4206),
    [anon_sym___extension__] = ACTIONS(4204),
    [anon_sym_extern] = ACTIONS(4204),
    [anon_sym___attribute__] = ACTIONS(4204),
    [anon_sym___attribute] = ACTIONS(4204),
    [anon_sym_noreturn] = ACTIONS(4204),
    [anon_sym_LBRACK] = ACTIONS(4206),
    [anon_sym_RBRACK] = ACTIONS(4206),
    [anon_sym___declspec] = ACTIONS(4204),
    [anon_sym_LBRACE] = ACTIONS(4206),
    [anon_sym_RBRACE] = ACTIONS(4206),
    [anon_sym_signed] = ACTIONS(4204),
    [anon_sym_unsigned] = ACTIONS(4204),
    [anon_sym_long] = ACTIONS(4204),
    [anon_sym_short] = ACTIONS(4204),
    [anon_sym_static] = ACTIONS(4204),
    [anon_sym_EQ] = ACTIONS(4204),
    [anon_sym_ATautoreleasepool] = ACTIONS(4206),
    [anon_sym_auto] = ACTIONS(4204),
    [anon_sym_register] = ACTIONS(4204),
    [anon_sym_inline] = ACTIONS(4204),
    [anon_sym___inline] = ACTIONS(4204),
    [anon_sym___inline__] = ACTIONS(4204),
    [anon_sym___forceinline] = ACTIONS(4204),
    [anon_sym_thread_local] = ACTIONS(4204),
    [anon_sym___thread] = ACTIONS(4204),
    [anon_sym_CG_EXTERN] = ACTIONS(4204),
    [anon_sym_CG_INLINE] = ACTIONS(4204),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4204),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4204),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4204),
    [anon_sym_IBOutlet] = ACTIONS(4204),
    [anon_sym_IBInspectable] = ACTIONS(4204),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4204),
    [anon_sym_NS_INLINE] = ACTIONS(4204),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4204),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4204),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4204),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4204),
    [anon_sym_const] = ACTIONS(4204),
    [anon_sym_constexpr] = ACTIONS(4204),
    [anon_sym_volatile] = ACTIONS(4204),
    [anon_sym_restrict] = ACTIONS(4204),
    [anon_sym___restrict__] = ACTIONS(4204),
    [anon_sym__Atomic] = ACTIONS(4204),
    [anon_sym__Noreturn] = ACTIONS(4204),
    [anon_sym__Nonnull] = ACTIONS(4204),
    [anon_sym_nullable] = ACTIONS(4204),
    [anon_sym__Complex] = ACTIONS(4204),
    [anon_sym__Nullable] = ACTIONS(4204),
    [anon_sym__Nullable_result] = ACTIONS(4204),
    [anon_sym__Null_unspecified] = ACTIONS(4204),
    [anon_sym___autoreleasing] = ACTIONS(4204),
    [anon_sym___block] = ACTIONS(4204),
    [anon_sym___bridge] = ACTIONS(4204),
    [anon_sym___bridge_retained] = ACTIONS(4204),
    [anon_sym___bridge_transfer] = ACTIONS(4204),
    [anon_sym___complex] = ACTIONS(4204),
    [anon_sym___const] = ACTIONS(4204),
    [anon_sym___imag] = ACTIONS(4204),
    [anon_sym___kindof] = ACTIONS(4204),
    [anon_sym___nonnull] = ACTIONS(4204),
    [anon_sym___nullable] = ACTIONS(4204),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4204),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4204),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4204),
    [anon_sym___real] = ACTIONS(4204),
    [anon_sym___strong] = ACTIONS(4204),
    [anon_sym___unsafe_unretained] = ACTIONS(4204),
    [anon_sym___unused] = ACTIONS(4204),
    [anon_sym___weak] = ACTIONS(4204),
    [anon_sym_alignas] = ACTIONS(4204),
    [anon_sym__Alignas] = ACTIONS(4204),
    [sym_primitive_type] = ACTIONS(4204),
    [anon_sym_enum] = ACTIONS(4204),
    [anon_sym_COLON] = ACTIONS(4206),
    [anon_sym_struct] = ACTIONS(4204),
    [anon_sym_union] = ACTIONS(4204),
    [anon_sym_QMARK] = ACTIONS(4206),
    [anon_sym_STAR_EQ] = ACTIONS(4206),
    [anon_sym_SLASH_EQ] = ACTIONS(4206),
    [anon_sym_PERCENT_EQ] = ACTIONS(4206),
    [anon_sym_PLUS_EQ] = ACTIONS(4206),
    [anon_sym_DASH_EQ] = ACTIONS(4206),
    [anon_sym_LT_LT_EQ] = ACTIONS(4206),
    [anon_sym_GT_GT_EQ] = ACTIONS(4206),
    [anon_sym_AMP_EQ] = ACTIONS(4206),
    [anon_sym_CARET_EQ] = ACTIONS(4206),
    [anon_sym_PIPE_EQ] = ACTIONS(4206),
    [anon_sym_DASH_DASH] = ACTIONS(4206),
    [anon_sym_PLUS_PLUS] = ACTIONS(4206),
    [anon_sym_DOT] = ACTIONS(4204),
    [anon_sym_DASH_GT] = ACTIONS(4206),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4204),
    [anon_sym___typeof] = ACTIONS(4204),
    [anon_sym_typeof] = ACTIONS(4204),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4204),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4204),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4204),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4204),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4204),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4204),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4204),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4204),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4204),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4204),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4204),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4204),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4204),
    [anon_sym_API_AVAILABLE] = ACTIONS(4204),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4204),
    [anon_sym_API_DEPRECATED] = ACTIONS(4204),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4204),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4204),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4204),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4204),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4204),
    [anon_sym___deprecated_msg] = ACTIONS(4204),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4204),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4204),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4204),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4204),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4204),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4204),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4204),
    [anon_sym_BOOL] = ACTIONS(4204),
    [anon_sym_IMP] = ACTIONS(4204),
    [anon_sym_SEL] = ACTIONS(4204),
    [anon_sym_Class] = ACTIONS(4204),
    [anon_sym_id] = ACTIONS(4204),
  },
  [STATE(1491)] = {
    [sym_identifier] = ACTIONS(4208),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4210),
    [anon_sym_COMMA] = ACTIONS(4210),
    [anon_sym_RPAREN] = ACTIONS(4210),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4210),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4210),
    [anon_sym_LPAREN2] = ACTIONS(4210),
    [anon_sym_DASH] = ACTIONS(4208),
    [anon_sym_PLUS] = ACTIONS(4208),
    [anon_sym_STAR] = ACTIONS(4208),
    [anon_sym_SLASH] = ACTIONS(4208),
    [anon_sym_PERCENT] = ACTIONS(4208),
    [anon_sym_PIPE_PIPE] = ACTIONS(4210),
    [anon_sym_AMP_AMP] = ACTIONS(4210),
    [anon_sym_PIPE] = ACTIONS(4208),
    [anon_sym_CARET] = ACTIONS(4208),
    [anon_sym_AMP] = ACTIONS(4208),
    [anon_sym_EQ_EQ] = ACTIONS(4210),
    [anon_sym_BANG_EQ] = ACTIONS(4210),
    [anon_sym_GT] = ACTIONS(4208),
    [anon_sym_GT_EQ] = ACTIONS(4210),
    [anon_sym_LT_EQ] = ACTIONS(4210),
    [anon_sym_LT] = ACTIONS(4208),
    [anon_sym_LT_LT] = ACTIONS(4208),
    [anon_sym_GT_GT] = ACTIONS(4208),
    [anon_sym_SEMI] = ACTIONS(4210),
    [anon_sym___extension__] = ACTIONS(4208),
    [anon_sym_extern] = ACTIONS(4208),
    [anon_sym___attribute__] = ACTIONS(4208),
    [anon_sym___attribute] = ACTIONS(4208),
    [anon_sym_noreturn] = ACTIONS(4208),
    [anon_sym_LBRACK] = ACTIONS(4210),
    [anon_sym_RBRACK] = ACTIONS(4210),
    [anon_sym___declspec] = ACTIONS(4208),
    [anon_sym_LBRACE] = ACTIONS(4210),
    [anon_sym_RBRACE] = ACTIONS(4210),
    [anon_sym_signed] = ACTIONS(4208),
    [anon_sym_unsigned] = ACTIONS(4208),
    [anon_sym_long] = ACTIONS(4208),
    [anon_sym_short] = ACTIONS(4208),
    [anon_sym_static] = ACTIONS(4208),
    [anon_sym_EQ] = ACTIONS(4208),
    [anon_sym_ATautoreleasepool] = ACTIONS(4210),
    [anon_sym_auto] = ACTIONS(4208),
    [anon_sym_register] = ACTIONS(4208),
    [anon_sym_inline] = ACTIONS(4208),
    [anon_sym___inline] = ACTIONS(4208),
    [anon_sym___inline__] = ACTIONS(4208),
    [anon_sym___forceinline] = ACTIONS(4208),
    [anon_sym_thread_local] = ACTIONS(4208),
    [anon_sym___thread] = ACTIONS(4208),
    [anon_sym_CG_EXTERN] = ACTIONS(4208),
    [anon_sym_CG_INLINE] = ACTIONS(4208),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4208),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4208),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4208),
    [anon_sym_IBOutlet] = ACTIONS(4208),
    [anon_sym_IBInspectable] = ACTIONS(4208),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4208),
    [anon_sym_NS_INLINE] = ACTIONS(4208),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4208),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4208),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4208),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4208),
    [anon_sym_const] = ACTIONS(4208),
    [anon_sym_constexpr] = ACTIONS(4208),
    [anon_sym_volatile] = ACTIONS(4208),
    [anon_sym_restrict] = ACTIONS(4208),
    [anon_sym___restrict__] = ACTIONS(4208),
    [anon_sym__Atomic] = ACTIONS(4208),
    [anon_sym__Noreturn] = ACTIONS(4208),
    [anon_sym__Nonnull] = ACTIONS(4208),
    [anon_sym_nullable] = ACTIONS(4208),
    [anon_sym__Complex] = ACTIONS(4208),
    [anon_sym__Nullable] = ACTIONS(4208),
    [anon_sym__Nullable_result] = ACTIONS(4208),
    [anon_sym__Null_unspecified] = ACTIONS(4208),
    [anon_sym___autoreleasing] = ACTIONS(4208),
    [anon_sym___block] = ACTIONS(4208),
    [anon_sym___bridge] = ACTIONS(4208),
    [anon_sym___bridge_retained] = ACTIONS(4208),
    [anon_sym___bridge_transfer] = ACTIONS(4208),
    [anon_sym___complex] = ACTIONS(4208),
    [anon_sym___const] = ACTIONS(4208),
    [anon_sym___imag] = ACTIONS(4208),
    [anon_sym___kindof] = ACTIONS(4208),
    [anon_sym___nonnull] = ACTIONS(4208),
    [anon_sym___nullable] = ACTIONS(4208),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4208),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4208),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4208),
    [anon_sym___real] = ACTIONS(4208),
    [anon_sym___strong] = ACTIONS(4208),
    [anon_sym___unsafe_unretained] = ACTIONS(4208),
    [anon_sym___unused] = ACTIONS(4208),
    [anon_sym___weak] = ACTIONS(4208),
    [anon_sym_alignas] = ACTIONS(4208),
    [anon_sym__Alignas] = ACTIONS(4208),
    [sym_primitive_type] = ACTIONS(4208),
    [anon_sym_enum] = ACTIONS(4208),
    [anon_sym_COLON] = ACTIONS(4210),
    [anon_sym_struct] = ACTIONS(4208),
    [anon_sym_union] = ACTIONS(4208),
    [anon_sym_QMARK] = ACTIONS(4210),
    [anon_sym_STAR_EQ] = ACTIONS(4210),
    [anon_sym_SLASH_EQ] = ACTIONS(4210),
    [anon_sym_PERCENT_EQ] = ACTIONS(4210),
    [anon_sym_PLUS_EQ] = ACTIONS(4210),
    [anon_sym_DASH_EQ] = ACTIONS(4210),
    [anon_sym_LT_LT_EQ] = ACTIONS(4210),
    [anon_sym_GT_GT_EQ] = ACTIONS(4210),
    [anon_sym_AMP_EQ] = ACTIONS(4210),
    [anon_sym_CARET_EQ] = ACTIONS(4210),
    [anon_sym_PIPE_EQ] = ACTIONS(4210),
    [anon_sym_DASH_DASH] = ACTIONS(4210),
    [anon_sym_PLUS_PLUS] = ACTIONS(4210),
    [anon_sym_DOT] = ACTIONS(4208),
    [anon_sym_DASH_GT] = ACTIONS(4210),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4208),
    [anon_sym___typeof] = ACTIONS(4208),
    [anon_sym_typeof] = ACTIONS(4208),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4208),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4208),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4208),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4208),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4208),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4208),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4208),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4208),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4208),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4208),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4208),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4208),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4208),
    [anon_sym_API_AVAILABLE] = ACTIONS(4208),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4208),
    [anon_sym_API_DEPRECATED] = ACTIONS(4208),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4208),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4208),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4208),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4208),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4208),
    [anon_sym___deprecated_msg] = ACTIONS(4208),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4208),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4208),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4208),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4208),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4208),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4208),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4208),
    [anon_sym_BOOL] = ACTIONS(4208),
    [anon_sym_IMP] = ACTIONS(4208),
    [anon_sym_SEL] = ACTIONS(4208),
    [anon_sym_Class] = ACTIONS(4208),
    [anon_sym_id] = ACTIONS(4208),
  },
  [STATE(1492)] = {
    [sym_identifier] = ACTIONS(4212),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4214),
    [anon_sym_COMMA] = ACTIONS(4214),
    [anon_sym_RPAREN] = ACTIONS(4214),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4214),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4214),
    [anon_sym_LPAREN2] = ACTIONS(4214),
    [anon_sym_DASH] = ACTIONS(4212),
    [anon_sym_PLUS] = ACTIONS(4212),
    [anon_sym_STAR] = ACTIONS(4212),
    [anon_sym_SLASH] = ACTIONS(4212),
    [anon_sym_PERCENT] = ACTIONS(4212),
    [anon_sym_PIPE_PIPE] = ACTIONS(4214),
    [anon_sym_AMP_AMP] = ACTIONS(4214),
    [anon_sym_PIPE] = ACTIONS(4212),
    [anon_sym_CARET] = ACTIONS(4212),
    [anon_sym_AMP] = ACTIONS(4212),
    [anon_sym_EQ_EQ] = ACTIONS(4214),
    [anon_sym_BANG_EQ] = ACTIONS(4214),
    [anon_sym_GT] = ACTIONS(4212),
    [anon_sym_GT_EQ] = ACTIONS(4214),
    [anon_sym_LT_EQ] = ACTIONS(4214),
    [anon_sym_LT] = ACTIONS(4212),
    [anon_sym_LT_LT] = ACTIONS(4212),
    [anon_sym_GT_GT] = ACTIONS(4212),
    [anon_sym_SEMI] = ACTIONS(4214),
    [anon_sym___extension__] = ACTIONS(4212),
    [anon_sym_extern] = ACTIONS(4212),
    [anon_sym___attribute__] = ACTIONS(4212),
    [anon_sym___attribute] = ACTIONS(4212),
    [anon_sym_noreturn] = ACTIONS(4212),
    [anon_sym_LBRACK] = ACTIONS(4214),
    [anon_sym_RBRACK] = ACTIONS(4214),
    [anon_sym___declspec] = ACTIONS(4212),
    [anon_sym_LBRACE] = ACTIONS(4214),
    [anon_sym_RBRACE] = ACTIONS(4214),
    [anon_sym_signed] = ACTIONS(4212),
    [anon_sym_unsigned] = ACTIONS(4212),
    [anon_sym_long] = ACTIONS(4212),
    [anon_sym_short] = ACTIONS(4212),
    [anon_sym_static] = ACTIONS(4212),
    [anon_sym_EQ] = ACTIONS(4212),
    [anon_sym_ATautoreleasepool] = ACTIONS(4214),
    [anon_sym_auto] = ACTIONS(4212),
    [anon_sym_register] = ACTIONS(4212),
    [anon_sym_inline] = ACTIONS(4212),
    [anon_sym___inline] = ACTIONS(4212),
    [anon_sym___inline__] = ACTIONS(4212),
    [anon_sym___forceinline] = ACTIONS(4212),
    [anon_sym_thread_local] = ACTIONS(4212),
    [anon_sym___thread] = ACTIONS(4212),
    [anon_sym_CG_EXTERN] = ACTIONS(4212),
    [anon_sym_CG_INLINE] = ACTIONS(4212),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4212),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4212),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4212),
    [anon_sym_IBOutlet] = ACTIONS(4212),
    [anon_sym_IBInspectable] = ACTIONS(4212),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4212),
    [anon_sym_NS_INLINE] = ACTIONS(4212),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4212),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4212),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4212),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4212),
    [anon_sym_const] = ACTIONS(4212),
    [anon_sym_constexpr] = ACTIONS(4212),
    [anon_sym_volatile] = ACTIONS(4212),
    [anon_sym_restrict] = ACTIONS(4212),
    [anon_sym___restrict__] = ACTIONS(4212),
    [anon_sym__Atomic] = ACTIONS(4212),
    [anon_sym__Noreturn] = ACTIONS(4212),
    [anon_sym__Nonnull] = ACTIONS(4212),
    [anon_sym_nullable] = ACTIONS(4212),
    [anon_sym__Complex] = ACTIONS(4212),
    [anon_sym__Nullable] = ACTIONS(4212),
    [anon_sym__Nullable_result] = ACTIONS(4212),
    [anon_sym__Null_unspecified] = ACTIONS(4212),
    [anon_sym___autoreleasing] = ACTIONS(4212),
    [anon_sym___block] = ACTIONS(4212),
    [anon_sym___bridge] = ACTIONS(4212),
    [anon_sym___bridge_retained] = ACTIONS(4212),
    [anon_sym___bridge_transfer] = ACTIONS(4212),
    [anon_sym___complex] = ACTIONS(4212),
    [anon_sym___const] = ACTIONS(4212),
    [anon_sym___imag] = ACTIONS(4212),
    [anon_sym___kindof] = ACTIONS(4212),
    [anon_sym___nonnull] = ACTIONS(4212),
    [anon_sym___nullable] = ACTIONS(4212),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4212),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4212),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4212),
    [anon_sym___real] = ACTIONS(4212),
    [anon_sym___strong] = ACTIONS(4212),
    [anon_sym___unsafe_unretained] = ACTIONS(4212),
    [anon_sym___unused] = ACTIONS(4212),
    [anon_sym___weak] = ACTIONS(4212),
    [anon_sym_alignas] = ACTIONS(4212),
    [anon_sym__Alignas] = ACTIONS(4212),
    [sym_primitive_type] = ACTIONS(4212),
    [anon_sym_enum] = ACTIONS(4212),
    [anon_sym_COLON] = ACTIONS(4214),
    [anon_sym_struct] = ACTIONS(4212),
    [anon_sym_union] = ACTIONS(4212),
    [anon_sym_QMARK] = ACTIONS(4214),
    [anon_sym_STAR_EQ] = ACTIONS(4214),
    [anon_sym_SLASH_EQ] = ACTIONS(4214),
    [anon_sym_PERCENT_EQ] = ACTIONS(4214),
    [anon_sym_PLUS_EQ] = ACTIONS(4214),
    [anon_sym_DASH_EQ] = ACTIONS(4214),
    [anon_sym_LT_LT_EQ] = ACTIONS(4214),
    [anon_sym_GT_GT_EQ] = ACTIONS(4214),
    [anon_sym_AMP_EQ] = ACTIONS(4214),
    [anon_sym_CARET_EQ] = ACTIONS(4214),
    [anon_sym_PIPE_EQ] = ACTIONS(4214),
    [anon_sym_DASH_DASH] = ACTIONS(4214),
    [anon_sym_PLUS_PLUS] = ACTIONS(4214),
    [anon_sym_DOT] = ACTIONS(4212),
    [anon_sym_DASH_GT] = ACTIONS(4214),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4212),
    [anon_sym___typeof] = ACTIONS(4212),
    [anon_sym_typeof] = ACTIONS(4212),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4212),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4212),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4212),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4212),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4212),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4212),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4212),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4212),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4212),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4212),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4212),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4212),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4212),
    [anon_sym_API_AVAILABLE] = ACTIONS(4212),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4212),
    [anon_sym_API_DEPRECATED] = ACTIONS(4212),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4212),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4212),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4212),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4212),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4212),
    [anon_sym___deprecated_msg] = ACTIONS(4212),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4212),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4212),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4212),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4212),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4212),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4212),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4212),
    [anon_sym_BOOL] = ACTIONS(4212),
    [anon_sym_IMP] = ACTIONS(4212),
    [anon_sym_SEL] = ACTIONS(4212),
    [anon_sym_Class] = ACTIONS(4212),
    [anon_sym_id] = ACTIONS(4212),
  },
  [STATE(1493)] = {
    [sym_identifier] = ACTIONS(4216),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4218),
    [anon_sym_COMMA] = ACTIONS(4218),
    [anon_sym_RPAREN] = ACTIONS(4218),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4218),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4218),
    [anon_sym_LPAREN2] = ACTIONS(4218),
    [anon_sym_DASH] = ACTIONS(4216),
    [anon_sym_PLUS] = ACTIONS(4216),
    [anon_sym_STAR] = ACTIONS(4216),
    [anon_sym_SLASH] = ACTIONS(4216),
    [anon_sym_PERCENT] = ACTIONS(4216),
    [anon_sym_PIPE_PIPE] = ACTIONS(4218),
    [anon_sym_AMP_AMP] = ACTIONS(4218),
    [anon_sym_PIPE] = ACTIONS(4216),
    [anon_sym_CARET] = ACTIONS(4216),
    [anon_sym_AMP] = ACTIONS(4216),
    [anon_sym_EQ_EQ] = ACTIONS(4218),
    [anon_sym_BANG_EQ] = ACTIONS(4218),
    [anon_sym_GT] = ACTIONS(4216),
    [anon_sym_GT_EQ] = ACTIONS(4218),
    [anon_sym_LT_EQ] = ACTIONS(4218),
    [anon_sym_LT] = ACTIONS(4216),
    [anon_sym_LT_LT] = ACTIONS(4216),
    [anon_sym_GT_GT] = ACTIONS(4216),
    [anon_sym_SEMI] = ACTIONS(4218),
    [anon_sym___extension__] = ACTIONS(4216),
    [anon_sym_extern] = ACTIONS(4216),
    [anon_sym___attribute__] = ACTIONS(4216),
    [anon_sym___attribute] = ACTIONS(4216),
    [anon_sym_noreturn] = ACTIONS(4216),
    [anon_sym_LBRACK] = ACTIONS(4218),
    [anon_sym_RBRACK] = ACTIONS(4218),
    [anon_sym___declspec] = ACTIONS(4216),
    [anon_sym_LBRACE] = ACTIONS(4218),
    [anon_sym_RBRACE] = ACTIONS(4218),
    [anon_sym_signed] = ACTIONS(4216),
    [anon_sym_unsigned] = ACTIONS(4216),
    [anon_sym_long] = ACTIONS(4216),
    [anon_sym_short] = ACTIONS(4216),
    [anon_sym_static] = ACTIONS(4216),
    [anon_sym_EQ] = ACTIONS(4216),
    [anon_sym_ATautoreleasepool] = ACTIONS(4218),
    [anon_sym_auto] = ACTIONS(4216),
    [anon_sym_register] = ACTIONS(4216),
    [anon_sym_inline] = ACTIONS(4216),
    [anon_sym___inline] = ACTIONS(4216),
    [anon_sym___inline__] = ACTIONS(4216),
    [anon_sym___forceinline] = ACTIONS(4216),
    [anon_sym_thread_local] = ACTIONS(4216),
    [anon_sym___thread] = ACTIONS(4216),
    [anon_sym_CG_EXTERN] = ACTIONS(4216),
    [anon_sym_CG_INLINE] = ACTIONS(4216),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4216),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4216),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4216),
    [anon_sym_IBOutlet] = ACTIONS(4216),
    [anon_sym_IBInspectable] = ACTIONS(4216),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4216),
    [anon_sym_NS_INLINE] = ACTIONS(4216),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4216),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4216),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4216),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4216),
    [anon_sym_const] = ACTIONS(4216),
    [anon_sym_constexpr] = ACTIONS(4216),
    [anon_sym_volatile] = ACTIONS(4216),
    [anon_sym_restrict] = ACTIONS(4216),
    [anon_sym___restrict__] = ACTIONS(4216),
    [anon_sym__Atomic] = ACTIONS(4216),
    [anon_sym__Noreturn] = ACTIONS(4216),
    [anon_sym__Nonnull] = ACTIONS(4216),
    [anon_sym_nullable] = ACTIONS(4216),
    [anon_sym__Complex] = ACTIONS(4216),
    [anon_sym__Nullable] = ACTIONS(4216),
    [anon_sym__Nullable_result] = ACTIONS(4216),
    [anon_sym__Null_unspecified] = ACTIONS(4216),
    [anon_sym___autoreleasing] = ACTIONS(4216),
    [anon_sym___block] = ACTIONS(4216),
    [anon_sym___bridge] = ACTIONS(4216),
    [anon_sym___bridge_retained] = ACTIONS(4216),
    [anon_sym___bridge_transfer] = ACTIONS(4216),
    [anon_sym___complex] = ACTIONS(4216),
    [anon_sym___const] = ACTIONS(4216),
    [anon_sym___imag] = ACTIONS(4216),
    [anon_sym___kindof] = ACTIONS(4216),
    [anon_sym___nonnull] = ACTIONS(4216),
    [anon_sym___nullable] = ACTIONS(4216),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4216),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4216),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4216),
    [anon_sym___real] = ACTIONS(4216),
    [anon_sym___strong] = ACTIONS(4216),
    [anon_sym___unsafe_unretained] = ACTIONS(4216),
    [anon_sym___unused] = ACTIONS(4216),
    [anon_sym___weak] = ACTIONS(4216),
    [anon_sym_alignas] = ACTIONS(4216),
    [anon_sym__Alignas] = ACTIONS(4216),
    [sym_primitive_type] = ACTIONS(4216),
    [anon_sym_enum] = ACTIONS(4216),
    [anon_sym_COLON] = ACTIONS(4218),
    [anon_sym_struct] = ACTIONS(4216),
    [anon_sym_union] = ACTIONS(4216),
    [anon_sym_QMARK] = ACTIONS(4218),
    [anon_sym_STAR_EQ] = ACTIONS(4218),
    [anon_sym_SLASH_EQ] = ACTIONS(4218),
    [anon_sym_PERCENT_EQ] = ACTIONS(4218),
    [anon_sym_PLUS_EQ] = ACTIONS(4218),
    [anon_sym_DASH_EQ] = ACTIONS(4218),
    [anon_sym_LT_LT_EQ] = ACTIONS(4218),
    [anon_sym_GT_GT_EQ] = ACTIONS(4218),
    [anon_sym_AMP_EQ] = ACTIONS(4218),
    [anon_sym_CARET_EQ] = ACTIONS(4218),
    [anon_sym_PIPE_EQ] = ACTIONS(4218),
    [anon_sym_DASH_DASH] = ACTIONS(4218),
    [anon_sym_PLUS_PLUS] = ACTIONS(4218),
    [anon_sym_DOT] = ACTIONS(4216),
    [anon_sym_DASH_GT] = ACTIONS(4218),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4216),
    [anon_sym___typeof] = ACTIONS(4216),
    [anon_sym_typeof] = ACTIONS(4216),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4216),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4216),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4216),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4216),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4216),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4216),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4216),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4216),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4216),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4216),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4216),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4216),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4216),
    [anon_sym_API_AVAILABLE] = ACTIONS(4216),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4216),
    [anon_sym_API_DEPRECATED] = ACTIONS(4216),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4216),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4216),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4216),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4216),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4216),
    [anon_sym___deprecated_msg] = ACTIONS(4216),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4216),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4216),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4216),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4216),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4216),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4216),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4216),
    [anon_sym_BOOL] = ACTIONS(4216),
    [anon_sym_IMP] = ACTIONS(4216),
    [anon_sym_SEL] = ACTIONS(4216),
    [anon_sym_Class] = ACTIONS(4216),
    [anon_sym_id] = ACTIONS(4216),
  },
  [STATE(1494)] = {
    [sym_identifier] = ACTIONS(4220),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4222),
    [anon_sym_COMMA] = ACTIONS(4222),
    [anon_sym_RPAREN] = ACTIONS(4222),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4222),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4222),
    [anon_sym_LPAREN2] = ACTIONS(4222),
    [anon_sym_DASH] = ACTIONS(4220),
    [anon_sym_PLUS] = ACTIONS(4220),
    [anon_sym_STAR] = ACTIONS(4220),
    [anon_sym_SLASH] = ACTIONS(4220),
    [anon_sym_PERCENT] = ACTIONS(4220),
    [anon_sym_PIPE_PIPE] = ACTIONS(4222),
    [anon_sym_AMP_AMP] = ACTIONS(4222),
    [anon_sym_PIPE] = ACTIONS(4220),
    [anon_sym_CARET] = ACTIONS(4220),
    [anon_sym_AMP] = ACTIONS(4220),
    [anon_sym_EQ_EQ] = ACTIONS(4222),
    [anon_sym_BANG_EQ] = ACTIONS(4222),
    [anon_sym_GT] = ACTIONS(4220),
    [anon_sym_GT_EQ] = ACTIONS(4222),
    [anon_sym_LT_EQ] = ACTIONS(4222),
    [anon_sym_LT] = ACTIONS(4220),
    [anon_sym_LT_LT] = ACTIONS(4220),
    [anon_sym_GT_GT] = ACTIONS(4220),
    [anon_sym_SEMI] = ACTIONS(4222),
    [anon_sym___extension__] = ACTIONS(4220),
    [anon_sym_extern] = ACTIONS(4220),
    [anon_sym___attribute__] = ACTIONS(4220),
    [anon_sym___attribute] = ACTIONS(4220),
    [anon_sym_noreturn] = ACTIONS(4220),
    [anon_sym_LBRACK] = ACTIONS(4222),
    [anon_sym_RBRACK] = ACTIONS(4222),
    [anon_sym___declspec] = ACTIONS(4220),
    [anon_sym_LBRACE] = ACTIONS(4222),
    [anon_sym_RBRACE] = ACTIONS(4222),
    [anon_sym_signed] = ACTIONS(4220),
    [anon_sym_unsigned] = ACTIONS(4220),
    [anon_sym_long] = ACTIONS(4220),
    [anon_sym_short] = ACTIONS(4220),
    [anon_sym_static] = ACTIONS(4220),
    [anon_sym_EQ] = ACTIONS(4220),
    [anon_sym_ATautoreleasepool] = ACTIONS(4222),
    [anon_sym_auto] = ACTIONS(4220),
    [anon_sym_register] = ACTIONS(4220),
    [anon_sym_inline] = ACTIONS(4220),
    [anon_sym___inline] = ACTIONS(4220),
    [anon_sym___inline__] = ACTIONS(4220),
    [anon_sym___forceinline] = ACTIONS(4220),
    [anon_sym_thread_local] = ACTIONS(4220),
    [anon_sym___thread] = ACTIONS(4220),
    [anon_sym_CG_EXTERN] = ACTIONS(4220),
    [anon_sym_CG_INLINE] = ACTIONS(4220),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4220),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4220),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4220),
    [anon_sym_IBOutlet] = ACTIONS(4220),
    [anon_sym_IBInspectable] = ACTIONS(4220),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4220),
    [anon_sym_NS_INLINE] = ACTIONS(4220),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4220),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4220),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4220),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4220),
    [anon_sym_const] = ACTIONS(4220),
    [anon_sym_constexpr] = ACTIONS(4220),
    [anon_sym_volatile] = ACTIONS(4220),
    [anon_sym_restrict] = ACTIONS(4220),
    [anon_sym___restrict__] = ACTIONS(4220),
    [anon_sym__Atomic] = ACTIONS(4220),
    [anon_sym__Noreturn] = ACTIONS(4220),
    [anon_sym__Nonnull] = ACTIONS(4220),
    [anon_sym_nullable] = ACTIONS(4220),
    [anon_sym__Complex] = ACTIONS(4220),
    [anon_sym__Nullable] = ACTIONS(4220),
    [anon_sym__Nullable_result] = ACTIONS(4220),
    [anon_sym__Null_unspecified] = ACTIONS(4220),
    [anon_sym___autoreleasing] = ACTIONS(4220),
    [anon_sym___block] = ACTIONS(4220),
    [anon_sym___bridge] = ACTIONS(4220),
    [anon_sym___bridge_retained] = ACTIONS(4220),
    [anon_sym___bridge_transfer] = ACTIONS(4220),
    [anon_sym___complex] = ACTIONS(4220),
    [anon_sym___const] = ACTIONS(4220),
    [anon_sym___imag] = ACTIONS(4220),
    [anon_sym___kindof] = ACTIONS(4220),
    [anon_sym___nonnull] = ACTIONS(4220),
    [anon_sym___nullable] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4220),
    [anon_sym___real] = ACTIONS(4220),
    [anon_sym___strong] = ACTIONS(4220),
    [anon_sym___unsafe_unretained] = ACTIONS(4220),
    [anon_sym___unused] = ACTIONS(4220),
    [anon_sym___weak] = ACTIONS(4220),
    [anon_sym_alignas] = ACTIONS(4220),
    [anon_sym__Alignas] = ACTIONS(4220),
    [sym_primitive_type] = ACTIONS(4220),
    [anon_sym_enum] = ACTIONS(4220),
    [anon_sym_COLON] = ACTIONS(4222),
    [anon_sym_struct] = ACTIONS(4220),
    [anon_sym_union] = ACTIONS(4220),
    [anon_sym_QMARK] = ACTIONS(4222),
    [anon_sym_STAR_EQ] = ACTIONS(4222),
    [anon_sym_SLASH_EQ] = ACTIONS(4222),
    [anon_sym_PERCENT_EQ] = ACTIONS(4222),
    [anon_sym_PLUS_EQ] = ACTIONS(4222),
    [anon_sym_DASH_EQ] = ACTIONS(4222),
    [anon_sym_LT_LT_EQ] = ACTIONS(4222),
    [anon_sym_GT_GT_EQ] = ACTIONS(4222),
    [anon_sym_AMP_EQ] = ACTIONS(4222),
    [anon_sym_CARET_EQ] = ACTIONS(4222),
    [anon_sym_PIPE_EQ] = ACTIONS(4222),
    [anon_sym_DASH_DASH] = ACTIONS(4222),
    [anon_sym_PLUS_PLUS] = ACTIONS(4222),
    [anon_sym_DOT] = ACTIONS(4220),
    [anon_sym_DASH_GT] = ACTIONS(4222),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4220),
    [anon_sym___typeof] = ACTIONS(4220),
    [anon_sym_typeof] = ACTIONS(4220),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4220),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4220),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4220),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4220),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4220),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4220),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4220),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4220),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4220),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_API_AVAILABLE] = ACTIONS(4220),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_API_DEPRECATED] = ACTIONS(4220),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4220),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4220),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4220),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4220),
    [anon_sym___deprecated_msg] = ACTIONS(4220),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4220),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4220),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4220),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4220),
    [anon_sym_BOOL] = ACTIONS(4220),
    [anon_sym_IMP] = ACTIONS(4220),
    [anon_sym_SEL] = ACTIONS(4220),
    [anon_sym_Class] = ACTIONS(4220),
    [anon_sym_id] = ACTIONS(4220),
  },
  [STATE(1495)] = {
    [sym_identifier] = ACTIONS(4224),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4226),
    [anon_sym_COMMA] = ACTIONS(4226),
    [anon_sym_RPAREN] = ACTIONS(4226),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4226),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4226),
    [anon_sym_LPAREN2] = ACTIONS(4226),
    [anon_sym_DASH] = ACTIONS(4224),
    [anon_sym_PLUS] = ACTIONS(4224),
    [anon_sym_STAR] = ACTIONS(4224),
    [anon_sym_SLASH] = ACTIONS(4224),
    [anon_sym_PERCENT] = ACTIONS(4224),
    [anon_sym_PIPE_PIPE] = ACTIONS(4226),
    [anon_sym_AMP_AMP] = ACTIONS(4226),
    [anon_sym_PIPE] = ACTIONS(4224),
    [anon_sym_CARET] = ACTIONS(4224),
    [anon_sym_AMP] = ACTIONS(4224),
    [anon_sym_EQ_EQ] = ACTIONS(4226),
    [anon_sym_BANG_EQ] = ACTIONS(4226),
    [anon_sym_GT] = ACTIONS(4224),
    [anon_sym_GT_EQ] = ACTIONS(4226),
    [anon_sym_LT_EQ] = ACTIONS(4226),
    [anon_sym_LT] = ACTIONS(4224),
    [anon_sym_LT_LT] = ACTIONS(4224),
    [anon_sym_GT_GT] = ACTIONS(4224),
    [anon_sym_SEMI] = ACTIONS(4226),
    [anon_sym___extension__] = ACTIONS(4224),
    [anon_sym_extern] = ACTIONS(4224),
    [anon_sym___attribute__] = ACTIONS(4224),
    [anon_sym___attribute] = ACTIONS(4224),
    [anon_sym_noreturn] = ACTIONS(4224),
    [anon_sym_LBRACK] = ACTIONS(4226),
    [anon_sym_RBRACK] = ACTIONS(4226),
    [anon_sym___declspec] = ACTIONS(4224),
    [anon_sym_LBRACE] = ACTIONS(4226),
    [anon_sym_RBRACE] = ACTIONS(4226),
    [anon_sym_signed] = ACTIONS(4224),
    [anon_sym_unsigned] = ACTIONS(4224),
    [anon_sym_long] = ACTIONS(4224),
    [anon_sym_short] = ACTIONS(4224),
    [anon_sym_static] = ACTIONS(4224),
    [anon_sym_EQ] = ACTIONS(4224),
    [anon_sym_ATautoreleasepool] = ACTIONS(4226),
    [anon_sym_auto] = ACTIONS(4224),
    [anon_sym_register] = ACTIONS(4224),
    [anon_sym_inline] = ACTIONS(4224),
    [anon_sym___inline] = ACTIONS(4224),
    [anon_sym___inline__] = ACTIONS(4224),
    [anon_sym___forceinline] = ACTIONS(4224),
    [anon_sym_thread_local] = ACTIONS(4224),
    [anon_sym___thread] = ACTIONS(4224),
    [anon_sym_CG_EXTERN] = ACTIONS(4224),
    [anon_sym_CG_INLINE] = ACTIONS(4224),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4224),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4224),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4224),
    [anon_sym_IBOutlet] = ACTIONS(4224),
    [anon_sym_IBInspectable] = ACTIONS(4224),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4224),
    [anon_sym_NS_INLINE] = ACTIONS(4224),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4224),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4224),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4224),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4224),
    [anon_sym_const] = ACTIONS(4224),
    [anon_sym_constexpr] = ACTIONS(4224),
    [anon_sym_volatile] = ACTIONS(4224),
    [anon_sym_restrict] = ACTIONS(4224),
    [anon_sym___restrict__] = ACTIONS(4224),
    [anon_sym__Atomic] = ACTIONS(4224),
    [anon_sym__Noreturn] = ACTIONS(4224),
    [anon_sym__Nonnull] = ACTIONS(4224),
    [anon_sym_nullable] = ACTIONS(4224),
    [anon_sym__Complex] = ACTIONS(4224),
    [anon_sym__Nullable] = ACTIONS(4224),
    [anon_sym__Nullable_result] = ACTIONS(4224),
    [anon_sym__Null_unspecified] = ACTIONS(4224),
    [anon_sym___autoreleasing] = ACTIONS(4224),
    [anon_sym___block] = ACTIONS(4224),
    [anon_sym___bridge] = ACTIONS(4224),
    [anon_sym___bridge_retained] = ACTIONS(4224),
    [anon_sym___bridge_transfer] = ACTIONS(4224),
    [anon_sym___complex] = ACTIONS(4224),
    [anon_sym___const] = ACTIONS(4224),
    [anon_sym___imag] = ACTIONS(4224),
    [anon_sym___kindof] = ACTIONS(4224),
    [anon_sym___nonnull] = ACTIONS(4224),
    [anon_sym___nullable] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4224),
    [anon_sym___real] = ACTIONS(4224),
    [anon_sym___strong] = ACTIONS(4224),
    [anon_sym___unsafe_unretained] = ACTIONS(4224),
    [anon_sym___unused] = ACTIONS(4224),
    [anon_sym___weak] = ACTIONS(4224),
    [anon_sym_alignas] = ACTIONS(4224),
    [anon_sym__Alignas] = ACTIONS(4224),
    [sym_primitive_type] = ACTIONS(4224),
    [anon_sym_enum] = ACTIONS(4224),
    [anon_sym_COLON] = ACTIONS(4226),
    [anon_sym_struct] = ACTIONS(4224),
    [anon_sym_union] = ACTIONS(4224),
    [anon_sym_QMARK] = ACTIONS(4226),
    [anon_sym_STAR_EQ] = ACTIONS(4226),
    [anon_sym_SLASH_EQ] = ACTIONS(4226),
    [anon_sym_PERCENT_EQ] = ACTIONS(4226),
    [anon_sym_PLUS_EQ] = ACTIONS(4226),
    [anon_sym_DASH_EQ] = ACTIONS(4226),
    [anon_sym_LT_LT_EQ] = ACTIONS(4226),
    [anon_sym_GT_GT_EQ] = ACTIONS(4226),
    [anon_sym_AMP_EQ] = ACTIONS(4226),
    [anon_sym_CARET_EQ] = ACTIONS(4226),
    [anon_sym_PIPE_EQ] = ACTIONS(4226),
    [anon_sym_DASH_DASH] = ACTIONS(4226),
    [anon_sym_PLUS_PLUS] = ACTIONS(4226),
    [anon_sym_DOT] = ACTIONS(4224),
    [anon_sym_DASH_GT] = ACTIONS(4226),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4224),
    [anon_sym___typeof] = ACTIONS(4224),
    [anon_sym_typeof] = ACTIONS(4224),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4224),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4224),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4224),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4224),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4224),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4224),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4224),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4224),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4224),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_API_AVAILABLE] = ACTIONS(4224),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_API_DEPRECATED] = ACTIONS(4224),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4224),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4224),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4224),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4224),
    [anon_sym___deprecated_msg] = ACTIONS(4224),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4224),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4224),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4224),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4224),
    [anon_sym_BOOL] = ACTIONS(4224),
    [anon_sym_IMP] = ACTIONS(4224),
    [anon_sym_SEL] = ACTIONS(4224),
    [anon_sym_Class] = ACTIONS(4224),
    [anon_sym_id] = ACTIONS(4224),
  },
  [STATE(1496)] = {
    [sym_identifier] = ACTIONS(2350),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2352),
    [anon_sym_COMMA] = ACTIONS(2352),
    [anon_sym_RPAREN] = ACTIONS(2352),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2352),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2352),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2350),
    [anon_sym_SLASH] = ACTIONS(2350),
    [anon_sym_PERCENT] = ACTIONS(2350),
    [anon_sym_PIPE_PIPE] = ACTIONS(2352),
    [anon_sym_AMP_AMP] = ACTIONS(2352),
    [anon_sym_PIPE] = ACTIONS(2350),
    [anon_sym_CARET] = ACTIONS(2350),
    [anon_sym_AMP] = ACTIONS(2350),
    [anon_sym_EQ_EQ] = ACTIONS(2352),
    [anon_sym_BANG_EQ] = ACTIONS(2352),
    [anon_sym_GT] = ACTIONS(2350),
    [anon_sym_GT_EQ] = ACTIONS(2352),
    [anon_sym_LT_EQ] = ACTIONS(2352),
    [anon_sym_LT] = ACTIONS(2350),
    [anon_sym_LT_LT] = ACTIONS(2350),
    [anon_sym_GT_GT] = ACTIONS(2350),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym_RBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_RBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_EQ] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_COLON] = ACTIONS(2352),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_QMARK] = ACTIONS(2352),
    [anon_sym_STAR_EQ] = ACTIONS(2352),
    [anon_sym_SLASH_EQ] = ACTIONS(2352),
    [anon_sym_PERCENT_EQ] = ACTIONS(2352),
    [anon_sym_PLUS_EQ] = ACTIONS(2352),
    [anon_sym_DASH_EQ] = ACTIONS(2352),
    [anon_sym_LT_LT_EQ] = ACTIONS(2352),
    [anon_sym_GT_GT_EQ] = ACTIONS(2352),
    [anon_sym_AMP_EQ] = ACTIONS(2352),
    [anon_sym_CARET_EQ] = ACTIONS(2352),
    [anon_sym_PIPE_EQ] = ACTIONS(2352),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_DOT] = ACTIONS(2350),
    [anon_sym_DASH_GT] = ACTIONS(2352),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(1497)] = {
    [sym_identifier] = ACTIONS(4228),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4230),
    [anon_sym_COMMA] = ACTIONS(4230),
    [anon_sym_RPAREN] = ACTIONS(4230),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4230),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4230),
    [anon_sym_LPAREN2] = ACTIONS(4230),
    [anon_sym_DASH] = ACTIONS(4228),
    [anon_sym_PLUS] = ACTIONS(4228),
    [anon_sym_STAR] = ACTIONS(4228),
    [anon_sym_SLASH] = ACTIONS(4228),
    [anon_sym_PERCENT] = ACTIONS(4228),
    [anon_sym_PIPE_PIPE] = ACTIONS(4230),
    [anon_sym_AMP_AMP] = ACTIONS(4230),
    [anon_sym_PIPE] = ACTIONS(4228),
    [anon_sym_CARET] = ACTIONS(4228),
    [anon_sym_AMP] = ACTIONS(4228),
    [anon_sym_EQ_EQ] = ACTIONS(4230),
    [anon_sym_BANG_EQ] = ACTIONS(4230),
    [anon_sym_GT] = ACTIONS(4228),
    [anon_sym_GT_EQ] = ACTIONS(4230),
    [anon_sym_LT_EQ] = ACTIONS(4230),
    [anon_sym_LT] = ACTIONS(4228),
    [anon_sym_LT_LT] = ACTIONS(4228),
    [anon_sym_GT_GT] = ACTIONS(4228),
    [anon_sym_SEMI] = ACTIONS(4230),
    [anon_sym___extension__] = ACTIONS(4228),
    [anon_sym_extern] = ACTIONS(4228),
    [anon_sym___attribute__] = ACTIONS(4228),
    [anon_sym___attribute] = ACTIONS(4228),
    [anon_sym_noreturn] = ACTIONS(4228),
    [anon_sym_LBRACK] = ACTIONS(4230),
    [anon_sym_RBRACK] = ACTIONS(4230),
    [anon_sym___declspec] = ACTIONS(4228),
    [anon_sym_LBRACE] = ACTIONS(4230),
    [anon_sym_RBRACE] = ACTIONS(4230),
    [anon_sym_signed] = ACTIONS(4228),
    [anon_sym_unsigned] = ACTIONS(4228),
    [anon_sym_long] = ACTIONS(4228),
    [anon_sym_short] = ACTIONS(4228),
    [anon_sym_static] = ACTIONS(4228),
    [anon_sym_EQ] = ACTIONS(4228),
    [anon_sym_ATautoreleasepool] = ACTIONS(4230),
    [anon_sym_auto] = ACTIONS(4228),
    [anon_sym_register] = ACTIONS(4228),
    [anon_sym_inline] = ACTIONS(4228),
    [anon_sym___inline] = ACTIONS(4228),
    [anon_sym___inline__] = ACTIONS(4228),
    [anon_sym___forceinline] = ACTIONS(4228),
    [anon_sym_thread_local] = ACTIONS(4228),
    [anon_sym___thread] = ACTIONS(4228),
    [anon_sym_CG_EXTERN] = ACTIONS(4228),
    [anon_sym_CG_INLINE] = ACTIONS(4228),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4228),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4228),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4228),
    [anon_sym_IBOutlet] = ACTIONS(4228),
    [anon_sym_IBInspectable] = ACTIONS(4228),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4228),
    [anon_sym_NS_INLINE] = ACTIONS(4228),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4228),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4228),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4228),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4228),
    [anon_sym_const] = ACTIONS(4228),
    [anon_sym_constexpr] = ACTIONS(4228),
    [anon_sym_volatile] = ACTIONS(4228),
    [anon_sym_restrict] = ACTIONS(4228),
    [anon_sym___restrict__] = ACTIONS(4228),
    [anon_sym__Atomic] = ACTIONS(4228),
    [anon_sym__Noreturn] = ACTIONS(4228),
    [anon_sym__Nonnull] = ACTIONS(4228),
    [anon_sym_nullable] = ACTIONS(4228),
    [anon_sym__Complex] = ACTIONS(4228),
    [anon_sym__Nullable] = ACTIONS(4228),
    [anon_sym__Nullable_result] = ACTIONS(4228),
    [anon_sym__Null_unspecified] = ACTIONS(4228),
    [anon_sym___autoreleasing] = ACTIONS(4228),
    [anon_sym___block] = ACTIONS(4228),
    [anon_sym___bridge] = ACTIONS(4228),
    [anon_sym___bridge_retained] = ACTIONS(4228),
    [anon_sym___bridge_transfer] = ACTIONS(4228),
    [anon_sym___complex] = ACTIONS(4228),
    [anon_sym___const] = ACTIONS(4228),
    [anon_sym___imag] = ACTIONS(4228),
    [anon_sym___kindof] = ACTIONS(4228),
    [anon_sym___nonnull] = ACTIONS(4228),
    [anon_sym___nullable] = ACTIONS(4228),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4228),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4228),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4228),
    [anon_sym___real] = ACTIONS(4228),
    [anon_sym___strong] = ACTIONS(4228),
    [anon_sym___unsafe_unretained] = ACTIONS(4228),
    [anon_sym___unused] = ACTIONS(4228),
    [anon_sym___weak] = ACTIONS(4228),
    [anon_sym_alignas] = ACTIONS(4228),
    [anon_sym__Alignas] = ACTIONS(4228),
    [sym_primitive_type] = ACTIONS(4228),
    [anon_sym_enum] = ACTIONS(4228),
    [anon_sym_COLON] = ACTIONS(4230),
    [anon_sym_struct] = ACTIONS(4228),
    [anon_sym_union] = ACTIONS(4228),
    [anon_sym_QMARK] = ACTIONS(4230),
    [anon_sym_STAR_EQ] = ACTIONS(4230),
    [anon_sym_SLASH_EQ] = ACTIONS(4230),
    [anon_sym_PERCENT_EQ] = ACTIONS(4230),
    [anon_sym_PLUS_EQ] = ACTIONS(4230),
    [anon_sym_DASH_EQ] = ACTIONS(4230),
    [anon_sym_LT_LT_EQ] = ACTIONS(4230),
    [anon_sym_GT_GT_EQ] = ACTIONS(4230),
    [anon_sym_AMP_EQ] = ACTIONS(4230),
    [anon_sym_CARET_EQ] = ACTIONS(4230),
    [anon_sym_PIPE_EQ] = ACTIONS(4230),
    [anon_sym_DASH_DASH] = ACTIONS(4230),
    [anon_sym_PLUS_PLUS] = ACTIONS(4230),
    [anon_sym_DOT] = ACTIONS(4228),
    [anon_sym_DASH_GT] = ACTIONS(4230),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4228),
    [anon_sym___typeof] = ACTIONS(4228),
    [anon_sym_typeof] = ACTIONS(4228),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4228),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4228),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4228),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4228),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4228),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4228),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4228),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4228),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4228),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4228),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4228),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4228),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4228),
    [anon_sym_API_AVAILABLE] = ACTIONS(4228),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4228),
    [anon_sym_API_DEPRECATED] = ACTIONS(4228),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4228),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4228),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4228),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4228),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4228),
    [anon_sym___deprecated_msg] = ACTIONS(4228),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4228),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4228),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4228),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4228),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4228),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4228),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4228),
    [anon_sym_BOOL] = ACTIONS(4228),
    [anon_sym_IMP] = ACTIONS(4228),
    [anon_sym_SEL] = ACTIONS(4228),
    [anon_sym_Class] = ACTIONS(4228),
    [anon_sym_id] = ACTIONS(4228),
  },
  [STATE(1498)] = {
    [sym_identifier] = ACTIONS(4232),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4234),
    [anon_sym_COMMA] = ACTIONS(4234),
    [anon_sym_RPAREN] = ACTIONS(4234),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4234),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4234),
    [anon_sym_LPAREN2] = ACTIONS(4234),
    [anon_sym_DASH] = ACTIONS(4232),
    [anon_sym_PLUS] = ACTIONS(4232),
    [anon_sym_STAR] = ACTIONS(4232),
    [anon_sym_SLASH] = ACTIONS(4232),
    [anon_sym_PERCENT] = ACTIONS(4232),
    [anon_sym_PIPE_PIPE] = ACTIONS(4234),
    [anon_sym_AMP_AMP] = ACTIONS(4234),
    [anon_sym_PIPE] = ACTIONS(4232),
    [anon_sym_CARET] = ACTIONS(4232),
    [anon_sym_AMP] = ACTIONS(4232),
    [anon_sym_EQ_EQ] = ACTIONS(4234),
    [anon_sym_BANG_EQ] = ACTIONS(4234),
    [anon_sym_GT] = ACTIONS(4232),
    [anon_sym_GT_EQ] = ACTIONS(4234),
    [anon_sym_LT_EQ] = ACTIONS(4234),
    [anon_sym_LT] = ACTIONS(4232),
    [anon_sym_LT_LT] = ACTIONS(4232),
    [anon_sym_GT_GT] = ACTIONS(4232),
    [anon_sym_SEMI] = ACTIONS(4234),
    [anon_sym___extension__] = ACTIONS(4232),
    [anon_sym_extern] = ACTIONS(4232),
    [anon_sym___attribute__] = ACTIONS(4232),
    [anon_sym___attribute] = ACTIONS(4232),
    [anon_sym_noreturn] = ACTIONS(4232),
    [anon_sym_LBRACK] = ACTIONS(4234),
    [anon_sym_RBRACK] = ACTIONS(4234),
    [anon_sym___declspec] = ACTIONS(4232),
    [anon_sym_LBRACE] = ACTIONS(4234),
    [anon_sym_RBRACE] = ACTIONS(4234),
    [anon_sym_signed] = ACTIONS(4232),
    [anon_sym_unsigned] = ACTIONS(4232),
    [anon_sym_long] = ACTIONS(4232),
    [anon_sym_short] = ACTIONS(4232),
    [anon_sym_static] = ACTIONS(4232),
    [anon_sym_EQ] = ACTIONS(4232),
    [anon_sym_ATautoreleasepool] = ACTIONS(4234),
    [anon_sym_auto] = ACTIONS(4232),
    [anon_sym_register] = ACTIONS(4232),
    [anon_sym_inline] = ACTIONS(4232),
    [anon_sym___inline] = ACTIONS(4232),
    [anon_sym___inline__] = ACTIONS(4232),
    [anon_sym___forceinline] = ACTIONS(4232),
    [anon_sym_thread_local] = ACTIONS(4232),
    [anon_sym___thread] = ACTIONS(4232),
    [anon_sym_CG_EXTERN] = ACTIONS(4232),
    [anon_sym_CG_INLINE] = ACTIONS(4232),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4232),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4232),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4232),
    [anon_sym_IBOutlet] = ACTIONS(4232),
    [anon_sym_IBInspectable] = ACTIONS(4232),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4232),
    [anon_sym_NS_INLINE] = ACTIONS(4232),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4232),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4232),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4232),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4232),
    [anon_sym_const] = ACTIONS(4232),
    [anon_sym_constexpr] = ACTIONS(4232),
    [anon_sym_volatile] = ACTIONS(4232),
    [anon_sym_restrict] = ACTIONS(4232),
    [anon_sym___restrict__] = ACTIONS(4232),
    [anon_sym__Atomic] = ACTIONS(4232),
    [anon_sym__Noreturn] = ACTIONS(4232),
    [anon_sym__Nonnull] = ACTIONS(4232),
    [anon_sym_nullable] = ACTIONS(4232),
    [anon_sym__Complex] = ACTIONS(4232),
    [anon_sym__Nullable] = ACTIONS(4232),
    [anon_sym__Nullable_result] = ACTIONS(4232),
    [anon_sym__Null_unspecified] = ACTIONS(4232),
    [anon_sym___autoreleasing] = ACTIONS(4232),
    [anon_sym___block] = ACTIONS(4232),
    [anon_sym___bridge] = ACTIONS(4232),
    [anon_sym___bridge_retained] = ACTIONS(4232),
    [anon_sym___bridge_transfer] = ACTIONS(4232),
    [anon_sym___complex] = ACTIONS(4232),
    [anon_sym___const] = ACTIONS(4232),
    [anon_sym___imag] = ACTIONS(4232),
    [anon_sym___kindof] = ACTIONS(4232),
    [anon_sym___nonnull] = ACTIONS(4232),
    [anon_sym___nullable] = ACTIONS(4232),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4232),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4232),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4232),
    [anon_sym___real] = ACTIONS(4232),
    [anon_sym___strong] = ACTIONS(4232),
    [anon_sym___unsafe_unretained] = ACTIONS(4232),
    [anon_sym___unused] = ACTIONS(4232),
    [anon_sym___weak] = ACTIONS(4232),
    [anon_sym_alignas] = ACTIONS(4232),
    [anon_sym__Alignas] = ACTIONS(4232),
    [sym_primitive_type] = ACTIONS(4232),
    [anon_sym_enum] = ACTIONS(4232),
    [anon_sym_COLON] = ACTIONS(4234),
    [anon_sym_struct] = ACTIONS(4232),
    [anon_sym_union] = ACTIONS(4232),
    [anon_sym_QMARK] = ACTIONS(4234),
    [anon_sym_STAR_EQ] = ACTIONS(4234),
    [anon_sym_SLASH_EQ] = ACTIONS(4234),
    [anon_sym_PERCENT_EQ] = ACTIONS(4234),
    [anon_sym_PLUS_EQ] = ACTIONS(4234),
    [anon_sym_DASH_EQ] = ACTIONS(4234),
    [anon_sym_LT_LT_EQ] = ACTIONS(4234),
    [anon_sym_GT_GT_EQ] = ACTIONS(4234),
    [anon_sym_AMP_EQ] = ACTIONS(4234),
    [anon_sym_CARET_EQ] = ACTIONS(4234),
    [anon_sym_PIPE_EQ] = ACTIONS(4234),
    [anon_sym_DASH_DASH] = ACTIONS(4234),
    [anon_sym_PLUS_PLUS] = ACTIONS(4234),
    [anon_sym_DOT] = ACTIONS(4232),
    [anon_sym_DASH_GT] = ACTIONS(4234),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4232),
    [anon_sym___typeof] = ACTIONS(4232),
    [anon_sym_typeof] = ACTIONS(4232),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4232),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4232),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4232),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4232),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4232),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4232),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4232),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4232),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4232),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4232),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4232),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4232),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4232),
    [anon_sym_API_AVAILABLE] = ACTIONS(4232),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4232),
    [anon_sym_API_DEPRECATED] = ACTIONS(4232),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4232),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4232),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4232),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4232),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4232),
    [anon_sym___deprecated_msg] = ACTIONS(4232),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4232),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4232),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4232),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4232),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4232),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4232),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4232),
    [anon_sym_BOOL] = ACTIONS(4232),
    [anon_sym_IMP] = ACTIONS(4232),
    [anon_sym_SEL] = ACTIONS(4232),
    [anon_sym_Class] = ACTIONS(4232),
    [anon_sym_id] = ACTIONS(4232),
  },
  [STATE(1499)] = {
    [sym_identifier] = ACTIONS(4236),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4238),
    [anon_sym_COMMA] = ACTIONS(4238),
    [anon_sym_RPAREN] = ACTIONS(4238),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4238),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4238),
    [anon_sym_LPAREN2] = ACTIONS(4238),
    [anon_sym_DASH] = ACTIONS(4236),
    [anon_sym_PLUS] = ACTIONS(4236),
    [anon_sym_STAR] = ACTIONS(4236),
    [anon_sym_SLASH] = ACTIONS(4236),
    [anon_sym_PERCENT] = ACTIONS(4236),
    [anon_sym_PIPE_PIPE] = ACTIONS(4238),
    [anon_sym_AMP_AMP] = ACTIONS(4238),
    [anon_sym_PIPE] = ACTIONS(4236),
    [anon_sym_CARET] = ACTIONS(4236),
    [anon_sym_AMP] = ACTIONS(4236),
    [anon_sym_EQ_EQ] = ACTIONS(4238),
    [anon_sym_BANG_EQ] = ACTIONS(4238),
    [anon_sym_GT] = ACTIONS(4236),
    [anon_sym_GT_EQ] = ACTIONS(4238),
    [anon_sym_LT_EQ] = ACTIONS(4238),
    [anon_sym_LT] = ACTIONS(4236),
    [anon_sym_LT_LT] = ACTIONS(4236),
    [anon_sym_GT_GT] = ACTIONS(4236),
    [anon_sym_SEMI] = ACTIONS(4238),
    [anon_sym___extension__] = ACTIONS(4236),
    [anon_sym_extern] = ACTIONS(4236),
    [anon_sym___attribute__] = ACTIONS(4236),
    [anon_sym___attribute] = ACTIONS(4236),
    [anon_sym_noreturn] = ACTIONS(4236),
    [anon_sym_LBRACK] = ACTIONS(4238),
    [anon_sym_RBRACK] = ACTIONS(4238),
    [anon_sym___declspec] = ACTIONS(4236),
    [anon_sym_LBRACE] = ACTIONS(4238),
    [anon_sym_RBRACE] = ACTIONS(4238),
    [anon_sym_signed] = ACTIONS(4236),
    [anon_sym_unsigned] = ACTIONS(4236),
    [anon_sym_long] = ACTIONS(4236),
    [anon_sym_short] = ACTIONS(4236),
    [anon_sym_static] = ACTIONS(4236),
    [anon_sym_EQ] = ACTIONS(4236),
    [anon_sym_ATautoreleasepool] = ACTIONS(4238),
    [anon_sym_auto] = ACTIONS(4236),
    [anon_sym_register] = ACTIONS(4236),
    [anon_sym_inline] = ACTIONS(4236),
    [anon_sym___inline] = ACTIONS(4236),
    [anon_sym___inline__] = ACTIONS(4236),
    [anon_sym___forceinline] = ACTIONS(4236),
    [anon_sym_thread_local] = ACTIONS(4236),
    [anon_sym___thread] = ACTIONS(4236),
    [anon_sym_CG_EXTERN] = ACTIONS(4236),
    [anon_sym_CG_INLINE] = ACTIONS(4236),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4236),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4236),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4236),
    [anon_sym_IBOutlet] = ACTIONS(4236),
    [anon_sym_IBInspectable] = ACTIONS(4236),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4236),
    [anon_sym_NS_INLINE] = ACTIONS(4236),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4236),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4236),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4236),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4236),
    [anon_sym_const] = ACTIONS(4236),
    [anon_sym_constexpr] = ACTIONS(4236),
    [anon_sym_volatile] = ACTIONS(4236),
    [anon_sym_restrict] = ACTIONS(4236),
    [anon_sym___restrict__] = ACTIONS(4236),
    [anon_sym__Atomic] = ACTIONS(4236),
    [anon_sym__Noreturn] = ACTIONS(4236),
    [anon_sym__Nonnull] = ACTIONS(4236),
    [anon_sym_nullable] = ACTIONS(4236),
    [anon_sym__Complex] = ACTIONS(4236),
    [anon_sym__Nullable] = ACTIONS(4236),
    [anon_sym__Nullable_result] = ACTIONS(4236),
    [anon_sym__Null_unspecified] = ACTIONS(4236),
    [anon_sym___autoreleasing] = ACTIONS(4236),
    [anon_sym___block] = ACTIONS(4236),
    [anon_sym___bridge] = ACTIONS(4236),
    [anon_sym___bridge_retained] = ACTIONS(4236),
    [anon_sym___bridge_transfer] = ACTIONS(4236),
    [anon_sym___complex] = ACTIONS(4236),
    [anon_sym___const] = ACTIONS(4236),
    [anon_sym___imag] = ACTIONS(4236),
    [anon_sym___kindof] = ACTIONS(4236),
    [anon_sym___nonnull] = ACTIONS(4236),
    [anon_sym___nullable] = ACTIONS(4236),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4236),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4236),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4236),
    [anon_sym___real] = ACTIONS(4236),
    [anon_sym___strong] = ACTIONS(4236),
    [anon_sym___unsafe_unretained] = ACTIONS(4236),
    [anon_sym___unused] = ACTIONS(4236),
    [anon_sym___weak] = ACTIONS(4236),
    [anon_sym_alignas] = ACTIONS(4236),
    [anon_sym__Alignas] = ACTIONS(4236),
    [sym_primitive_type] = ACTIONS(4236),
    [anon_sym_enum] = ACTIONS(4236),
    [anon_sym_COLON] = ACTIONS(4238),
    [anon_sym_struct] = ACTIONS(4236),
    [anon_sym_union] = ACTIONS(4236),
    [anon_sym_QMARK] = ACTIONS(4238),
    [anon_sym_STAR_EQ] = ACTIONS(4238),
    [anon_sym_SLASH_EQ] = ACTIONS(4238),
    [anon_sym_PERCENT_EQ] = ACTIONS(4238),
    [anon_sym_PLUS_EQ] = ACTIONS(4238),
    [anon_sym_DASH_EQ] = ACTIONS(4238),
    [anon_sym_LT_LT_EQ] = ACTIONS(4238),
    [anon_sym_GT_GT_EQ] = ACTIONS(4238),
    [anon_sym_AMP_EQ] = ACTIONS(4238),
    [anon_sym_CARET_EQ] = ACTIONS(4238),
    [anon_sym_PIPE_EQ] = ACTIONS(4238),
    [anon_sym_DASH_DASH] = ACTIONS(4238),
    [anon_sym_PLUS_PLUS] = ACTIONS(4238),
    [anon_sym_DOT] = ACTIONS(4236),
    [anon_sym_DASH_GT] = ACTIONS(4238),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4236),
    [anon_sym___typeof] = ACTIONS(4236),
    [anon_sym_typeof] = ACTIONS(4236),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4236),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4236),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4236),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4236),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4236),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4236),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4236),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4236),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4236),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4236),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4236),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4236),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4236),
    [anon_sym_API_AVAILABLE] = ACTIONS(4236),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4236),
    [anon_sym_API_DEPRECATED] = ACTIONS(4236),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4236),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4236),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4236),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4236),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4236),
    [anon_sym___deprecated_msg] = ACTIONS(4236),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4236),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4236),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4236),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4236),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4236),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4236),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4236),
    [anon_sym_BOOL] = ACTIONS(4236),
    [anon_sym_IMP] = ACTIONS(4236),
    [anon_sym_SEL] = ACTIONS(4236),
    [anon_sym_Class] = ACTIONS(4236),
    [anon_sym_id] = ACTIONS(4236),
  },
  [STATE(1500)] = {
    [sym_identifier] = ACTIONS(4240),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4242),
    [anon_sym_COMMA] = ACTIONS(4242),
    [anon_sym_RPAREN] = ACTIONS(4242),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4242),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4242),
    [anon_sym_LPAREN2] = ACTIONS(4242),
    [anon_sym_DASH] = ACTIONS(4240),
    [anon_sym_PLUS] = ACTIONS(4240),
    [anon_sym_STAR] = ACTIONS(4240),
    [anon_sym_SLASH] = ACTIONS(4240),
    [anon_sym_PERCENT] = ACTIONS(4240),
    [anon_sym_PIPE_PIPE] = ACTIONS(4242),
    [anon_sym_AMP_AMP] = ACTIONS(4242),
    [anon_sym_PIPE] = ACTIONS(4240),
    [anon_sym_CARET] = ACTIONS(4240),
    [anon_sym_AMP] = ACTIONS(4240),
    [anon_sym_EQ_EQ] = ACTIONS(4242),
    [anon_sym_BANG_EQ] = ACTIONS(4242),
    [anon_sym_GT] = ACTIONS(4240),
    [anon_sym_GT_EQ] = ACTIONS(4242),
    [anon_sym_LT_EQ] = ACTIONS(4242),
    [anon_sym_LT] = ACTIONS(4240),
    [anon_sym_LT_LT] = ACTIONS(4240),
    [anon_sym_GT_GT] = ACTIONS(4240),
    [anon_sym_SEMI] = ACTIONS(4242),
    [anon_sym___extension__] = ACTIONS(4240),
    [anon_sym_extern] = ACTIONS(4240),
    [anon_sym___attribute__] = ACTIONS(4240),
    [anon_sym___attribute] = ACTIONS(4240),
    [anon_sym_noreturn] = ACTIONS(4240),
    [anon_sym_LBRACK] = ACTIONS(4242),
    [anon_sym_RBRACK] = ACTIONS(4242),
    [anon_sym___declspec] = ACTIONS(4240),
    [anon_sym_LBRACE] = ACTIONS(4242),
    [anon_sym_RBRACE] = ACTIONS(4242),
    [anon_sym_signed] = ACTIONS(4240),
    [anon_sym_unsigned] = ACTIONS(4240),
    [anon_sym_long] = ACTIONS(4240),
    [anon_sym_short] = ACTIONS(4240),
    [anon_sym_static] = ACTIONS(4240),
    [anon_sym_EQ] = ACTIONS(4240),
    [anon_sym_ATautoreleasepool] = ACTIONS(4242),
    [anon_sym_auto] = ACTIONS(4240),
    [anon_sym_register] = ACTIONS(4240),
    [anon_sym_inline] = ACTIONS(4240),
    [anon_sym___inline] = ACTIONS(4240),
    [anon_sym___inline__] = ACTIONS(4240),
    [anon_sym___forceinline] = ACTIONS(4240),
    [anon_sym_thread_local] = ACTIONS(4240),
    [anon_sym___thread] = ACTIONS(4240),
    [anon_sym_CG_EXTERN] = ACTIONS(4240),
    [anon_sym_CG_INLINE] = ACTIONS(4240),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4240),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4240),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4240),
    [anon_sym_IBOutlet] = ACTIONS(4240),
    [anon_sym_IBInspectable] = ACTIONS(4240),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4240),
    [anon_sym_NS_INLINE] = ACTIONS(4240),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4240),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4240),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4240),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4240),
    [anon_sym_const] = ACTIONS(4240),
    [anon_sym_constexpr] = ACTIONS(4240),
    [anon_sym_volatile] = ACTIONS(4240),
    [anon_sym_restrict] = ACTIONS(4240),
    [anon_sym___restrict__] = ACTIONS(4240),
    [anon_sym__Atomic] = ACTIONS(4240),
    [anon_sym__Noreturn] = ACTIONS(4240),
    [anon_sym__Nonnull] = ACTIONS(4240),
    [anon_sym_nullable] = ACTIONS(4240),
    [anon_sym__Complex] = ACTIONS(4240),
    [anon_sym__Nullable] = ACTIONS(4240),
    [anon_sym__Nullable_result] = ACTIONS(4240),
    [anon_sym__Null_unspecified] = ACTIONS(4240),
    [anon_sym___autoreleasing] = ACTIONS(4240),
    [anon_sym___block] = ACTIONS(4240),
    [anon_sym___bridge] = ACTIONS(4240),
    [anon_sym___bridge_retained] = ACTIONS(4240),
    [anon_sym___bridge_transfer] = ACTIONS(4240),
    [anon_sym___complex] = ACTIONS(4240),
    [anon_sym___const] = ACTIONS(4240),
    [anon_sym___imag] = ACTIONS(4240),
    [anon_sym___kindof] = ACTIONS(4240),
    [anon_sym___nonnull] = ACTIONS(4240),
    [anon_sym___nullable] = ACTIONS(4240),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4240),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4240),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4240),
    [anon_sym___real] = ACTIONS(4240),
    [anon_sym___strong] = ACTIONS(4240),
    [anon_sym___unsafe_unretained] = ACTIONS(4240),
    [anon_sym___unused] = ACTIONS(4240),
    [anon_sym___weak] = ACTIONS(4240),
    [anon_sym_alignas] = ACTIONS(4240),
    [anon_sym__Alignas] = ACTIONS(4240),
    [sym_primitive_type] = ACTIONS(4240),
    [anon_sym_enum] = ACTIONS(4240),
    [anon_sym_COLON] = ACTIONS(4242),
    [anon_sym_struct] = ACTIONS(4240),
    [anon_sym_union] = ACTIONS(4240),
    [anon_sym_QMARK] = ACTIONS(4242),
    [anon_sym_STAR_EQ] = ACTIONS(4242),
    [anon_sym_SLASH_EQ] = ACTIONS(4242),
    [anon_sym_PERCENT_EQ] = ACTIONS(4242),
    [anon_sym_PLUS_EQ] = ACTIONS(4242),
    [anon_sym_DASH_EQ] = ACTIONS(4242),
    [anon_sym_LT_LT_EQ] = ACTIONS(4242),
    [anon_sym_GT_GT_EQ] = ACTIONS(4242),
    [anon_sym_AMP_EQ] = ACTIONS(4242),
    [anon_sym_CARET_EQ] = ACTIONS(4242),
    [anon_sym_PIPE_EQ] = ACTIONS(4242),
    [anon_sym_DASH_DASH] = ACTIONS(4242),
    [anon_sym_PLUS_PLUS] = ACTIONS(4242),
    [anon_sym_DOT] = ACTIONS(4240),
    [anon_sym_DASH_GT] = ACTIONS(4242),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4240),
    [anon_sym___typeof] = ACTIONS(4240),
    [anon_sym_typeof] = ACTIONS(4240),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4240),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4240),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4240),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4240),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4240),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4240),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4240),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4240),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4240),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4240),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4240),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4240),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4240),
    [anon_sym_API_AVAILABLE] = ACTIONS(4240),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4240),
    [anon_sym_API_DEPRECATED] = ACTIONS(4240),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4240),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4240),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4240),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4240),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4240),
    [anon_sym___deprecated_msg] = ACTIONS(4240),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4240),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4240),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4240),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4240),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4240),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4240),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4240),
    [anon_sym_BOOL] = ACTIONS(4240),
    [anon_sym_IMP] = ACTIONS(4240),
    [anon_sym_SEL] = ACTIONS(4240),
    [anon_sym_Class] = ACTIONS(4240),
    [anon_sym_id] = ACTIONS(4240),
  },
  [STATE(1501)] = {
    [sym_identifier] = ACTIONS(4244),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4246),
    [anon_sym_COMMA] = ACTIONS(4246),
    [anon_sym_RPAREN] = ACTIONS(4246),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4246),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4246),
    [anon_sym_LPAREN2] = ACTIONS(4246),
    [anon_sym_DASH] = ACTIONS(4244),
    [anon_sym_PLUS] = ACTIONS(4244),
    [anon_sym_STAR] = ACTIONS(4244),
    [anon_sym_SLASH] = ACTIONS(4244),
    [anon_sym_PERCENT] = ACTIONS(4244),
    [anon_sym_PIPE_PIPE] = ACTIONS(4246),
    [anon_sym_AMP_AMP] = ACTIONS(4246),
    [anon_sym_PIPE] = ACTIONS(4244),
    [anon_sym_CARET] = ACTIONS(4244),
    [anon_sym_AMP] = ACTIONS(4244),
    [anon_sym_EQ_EQ] = ACTIONS(4246),
    [anon_sym_BANG_EQ] = ACTIONS(4246),
    [anon_sym_GT] = ACTIONS(4244),
    [anon_sym_GT_EQ] = ACTIONS(4246),
    [anon_sym_LT_EQ] = ACTIONS(4246),
    [anon_sym_LT] = ACTIONS(4244),
    [anon_sym_LT_LT] = ACTIONS(4244),
    [anon_sym_GT_GT] = ACTIONS(4244),
    [anon_sym_SEMI] = ACTIONS(4246),
    [anon_sym___extension__] = ACTIONS(4244),
    [anon_sym_extern] = ACTIONS(4244),
    [anon_sym___attribute__] = ACTIONS(4244),
    [anon_sym___attribute] = ACTIONS(4244),
    [anon_sym_noreturn] = ACTIONS(4244),
    [anon_sym_LBRACK] = ACTIONS(4246),
    [anon_sym_RBRACK] = ACTIONS(4246),
    [anon_sym___declspec] = ACTIONS(4244),
    [anon_sym_LBRACE] = ACTIONS(4246),
    [anon_sym_RBRACE] = ACTIONS(4246),
    [anon_sym_signed] = ACTIONS(4244),
    [anon_sym_unsigned] = ACTIONS(4244),
    [anon_sym_long] = ACTIONS(4244),
    [anon_sym_short] = ACTIONS(4244),
    [anon_sym_static] = ACTIONS(4244),
    [anon_sym_EQ] = ACTIONS(4244),
    [anon_sym_ATautoreleasepool] = ACTIONS(4246),
    [anon_sym_auto] = ACTIONS(4244),
    [anon_sym_register] = ACTIONS(4244),
    [anon_sym_inline] = ACTIONS(4244),
    [anon_sym___inline] = ACTIONS(4244),
    [anon_sym___inline__] = ACTIONS(4244),
    [anon_sym___forceinline] = ACTIONS(4244),
    [anon_sym_thread_local] = ACTIONS(4244),
    [anon_sym___thread] = ACTIONS(4244),
    [anon_sym_CG_EXTERN] = ACTIONS(4244),
    [anon_sym_CG_INLINE] = ACTIONS(4244),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4244),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4244),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4244),
    [anon_sym_IBOutlet] = ACTIONS(4244),
    [anon_sym_IBInspectable] = ACTIONS(4244),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4244),
    [anon_sym_NS_INLINE] = ACTIONS(4244),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4244),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4244),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4244),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4244),
    [anon_sym_const] = ACTIONS(4244),
    [anon_sym_constexpr] = ACTIONS(4244),
    [anon_sym_volatile] = ACTIONS(4244),
    [anon_sym_restrict] = ACTIONS(4244),
    [anon_sym___restrict__] = ACTIONS(4244),
    [anon_sym__Atomic] = ACTIONS(4244),
    [anon_sym__Noreturn] = ACTIONS(4244),
    [anon_sym__Nonnull] = ACTIONS(4244),
    [anon_sym_nullable] = ACTIONS(4244),
    [anon_sym__Complex] = ACTIONS(4244),
    [anon_sym__Nullable] = ACTIONS(4244),
    [anon_sym__Nullable_result] = ACTIONS(4244),
    [anon_sym__Null_unspecified] = ACTIONS(4244),
    [anon_sym___autoreleasing] = ACTIONS(4244),
    [anon_sym___block] = ACTIONS(4244),
    [anon_sym___bridge] = ACTIONS(4244),
    [anon_sym___bridge_retained] = ACTIONS(4244),
    [anon_sym___bridge_transfer] = ACTIONS(4244),
    [anon_sym___complex] = ACTIONS(4244),
    [anon_sym___const] = ACTIONS(4244),
    [anon_sym___imag] = ACTIONS(4244),
    [anon_sym___kindof] = ACTIONS(4244),
    [anon_sym___nonnull] = ACTIONS(4244),
    [anon_sym___nullable] = ACTIONS(4244),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4244),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4244),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4244),
    [anon_sym___real] = ACTIONS(4244),
    [anon_sym___strong] = ACTIONS(4244),
    [anon_sym___unsafe_unretained] = ACTIONS(4244),
    [anon_sym___unused] = ACTIONS(4244),
    [anon_sym___weak] = ACTIONS(4244),
    [anon_sym_alignas] = ACTIONS(4244),
    [anon_sym__Alignas] = ACTIONS(4244),
    [sym_primitive_type] = ACTIONS(4244),
    [anon_sym_enum] = ACTIONS(4244),
    [anon_sym_COLON] = ACTIONS(4246),
    [anon_sym_struct] = ACTIONS(4244),
    [anon_sym_union] = ACTIONS(4244),
    [anon_sym_QMARK] = ACTIONS(4246),
    [anon_sym_STAR_EQ] = ACTIONS(4246),
    [anon_sym_SLASH_EQ] = ACTIONS(4246),
    [anon_sym_PERCENT_EQ] = ACTIONS(4246),
    [anon_sym_PLUS_EQ] = ACTIONS(4246),
    [anon_sym_DASH_EQ] = ACTIONS(4246),
    [anon_sym_LT_LT_EQ] = ACTIONS(4246),
    [anon_sym_GT_GT_EQ] = ACTIONS(4246),
    [anon_sym_AMP_EQ] = ACTIONS(4246),
    [anon_sym_CARET_EQ] = ACTIONS(4246),
    [anon_sym_PIPE_EQ] = ACTIONS(4246),
    [anon_sym_DASH_DASH] = ACTIONS(4246),
    [anon_sym_PLUS_PLUS] = ACTIONS(4246),
    [anon_sym_DOT] = ACTIONS(4244),
    [anon_sym_DASH_GT] = ACTIONS(4246),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4244),
    [anon_sym___typeof] = ACTIONS(4244),
    [anon_sym_typeof] = ACTIONS(4244),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4244),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4244),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4244),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4244),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4244),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4244),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4244),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4244),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4244),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4244),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4244),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4244),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4244),
    [anon_sym_API_AVAILABLE] = ACTIONS(4244),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4244),
    [anon_sym_API_DEPRECATED] = ACTIONS(4244),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4244),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4244),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4244),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4244),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4244),
    [anon_sym___deprecated_msg] = ACTIONS(4244),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4244),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4244),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4244),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4244),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4244),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4244),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4244),
    [anon_sym_BOOL] = ACTIONS(4244),
    [anon_sym_IMP] = ACTIONS(4244),
    [anon_sym_SEL] = ACTIONS(4244),
    [anon_sym_Class] = ACTIONS(4244),
    [anon_sym_id] = ACTIONS(4244),
  },
  [STATE(1502)] = {
    [sym_identifier] = ACTIONS(4248),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4250),
    [anon_sym_COMMA] = ACTIONS(4250),
    [anon_sym_RPAREN] = ACTIONS(4250),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4250),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4250),
    [anon_sym_LPAREN2] = ACTIONS(4250),
    [anon_sym_DASH] = ACTIONS(4248),
    [anon_sym_PLUS] = ACTIONS(4248),
    [anon_sym_STAR] = ACTIONS(4248),
    [anon_sym_SLASH] = ACTIONS(4248),
    [anon_sym_PERCENT] = ACTIONS(4248),
    [anon_sym_PIPE_PIPE] = ACTIONS(4250),
    [anon_sym_AMP_AMP] = ACTIONS(4250),
    [anon_sym_PIPE] = ACTIONS(4248),
    [anon_sym_CARET] = ACTIONS(4248),
    [anon_sym_AMP] = ACTIONS(4248),
    [anon_sym_EQ_EQ] = ACTIONS(4250),
    [anon_sym_BANG_EQ] = ACTIONS(4250),
    [anon_sym_GT] = ACTIONS(4248),
    [anon_sym_GT_EQ] = ACTIONS(4250),
    [anon_sym_LT_EQ] = ACTIONS(4250),
    [anon_sym_LT] = ACTIONS(4248),
    [anon_sym_LT_LT] = ACTIONS(4248),
    [anon_sym_GT_GT] = ACTIONS(4248),
    [anon_sym_SEMI] = ACTIONS(4250),
    [anon_sym___extension__] = ACTIONS(4248),
    [anon_sym_extern] = ACTIONS(4248),
    [anon_sym___attribute__] = ACTIONS(4248),
    [anon_sym___attribute] = ACTIONS(4248),
    [anon_sym_noreturn] = ACTIONS(4248),
    [anon_sym_LBRACK] = ACTIONS(4250),
    [anon_sym_RBRACK] = ACTIONS(4250),
    [anon_sym___declspec] = ACTIONS(4248),
    [anon_sym_LBRACE] = ACTIONS(4250),
    [anon_sym_RBRACE] = ACTIONS(4250),
    [anon_sym_signed] = ACTIONS(4248),
    [anon_sym_unsigned] = ACTIONS(4248),
    [anon_sym_long] = ACTIONS(4248),
    [anon_sym_short] = ACTIONS(4248),
    [anon_sym_static] = ACTIONS(4248),
    [anon_sym_EQ] = ACTIONS(4248),
    [anon_sym_ATautoreleasepool] = ACTIONS(4250),
    [anon_sym_auto] = ACTIONS(4248),
    [anon_sym_register] = ACTIONS(4248),
    [anon_sym_inline] = ACTIONS(4248),
    [anon_sym___inline] = ACTIONS(4248),
    [anon_sym___inline__] = ACTIONS(4248),
    [anon_sym___forceinline] = ACTIONS(4248),
    [anon_sym_thread_local] = ACTIONS(4248),
    [anon_sym___thread] = ACTIONS(4248),
    [anon_sym_CG_EXTERN] = ACTIONS(4248),
    [anon_sym_CG_INLINE] = ACTIONS(4248),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4248),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4248),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4248),
    [anon_sym_IBOutlet] = ACTIONS(4248),
    [anon_sym_IBInspectable] = ACTIONS(4248),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4248),
    [anon_sym_NS_INLINE] = ACTIONS(4248),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4248),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4248),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4248),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4248),
    [anon_sym_const] = ACTIONS(4248),
    [anon_sym_constexpr] = ACTIONS(4248),
    [anon_sym_volatile] = ACTIONS(4248),
    [anon_sym_restrict] = ACTIONS(4248),
    [anon_sym___restrict__] = ACTIONS(4248),
    [anon_sym__Atomic] = ACTIONS(4248),
    [anon_sym__Noreturn] = ACTIONS(4248),
    [anon_sym__Nonnull] = ACTIONS(4248),
    [anon_sym_nullable] = ACTIONS(4248),
    [anon_sym__Complex] = ACTIONS(4248),
    [anon_sym__Nullable] = ACTIONS(4248),
    [anon_sym__Nullable_result] = ACTIONS(4248),
    [anon_sym__Null_unspecified] = ACTIONS(4248),
    [anon_sym___autoreleasing] = ACTIONS(4248),
    [anon_sym___block] = ACTIONS(4248),
    [anon_sym___bridge] = ACTIONS(4248),
    [anon_sym___bridge_retained] = ACTIONS(4248),
    [anon_sym___bridge_transfer] = ACTIONS(4248),
    [anon_sym___complex] = ACTIONS(4248),
    [anon_sym___const] = ACTIONS(4248),
    [anon_sym___imag] = ACTIONS(4248),
    [anon_sym___kindof] = ACTIONS(4248),
    [anon_sym___nonnull] = ACTIONS(4248),
    [anon_sym___nullable] = ACTIONS(4248),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4248),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4248),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4248),
    [anon_sym___real] = ACTIONS(4248),
    [anon_sym___strong] = ACTIONS(4248),
    [anon_sym___unsafe_unretained] = ACTIONS(4248),
    [anon_sym___unused] = ACTIONS(4248),
    [anon_sym___weak] = ACTIONS(4248),
    [anon_sym_alignas] = ACTIONS(4248),
    [anon_sym__Alignas] = ACTIONS(4248),
    [sym_primitive_type] = ACTIONS(4248),
    [anon_sym_enum] = ACTIONS(4248),
    [anon_sym_COLON] = ACTIONS(4250),
    [anon_sym_struct] = ACTIONS(4248),
    [anon_sym_union] = ACTIONS(4248),
    [anon_sym_QMARK] = ACTIONS(4250),
    [anon_sym_STAR_EQ] = ACTIONS(4250),
    [anon_sym_SLASH_EQ] = ACTIONS(4250),
    [anon_sym_PERCENT_EQ] = ACTIONS(4250),
    [anon_sym_PLUS_EQ] = ACTIONS(4250),
    [anon_sym_DASH_EQ] = ACTIONS(4250),
    [anon_sym_LT_LT_EQ] = ACTIONS(4250),
    [anon_sym_GT_GT_EQ] = ACTIONS(4250),
    [anon_sym_AMP_EQ] = ACTIONS(4250),
    [anon_sym_CARET_EQ] = ACTIONS(4250),
    [anon_sym_PIPE_EQ] = ACTIONS(4250),
    [anon_sym_DASH_DASH] = ACTIONS(4250),
    [anon_sym_PLUS_PLUS] = ACTIONS(4250),
    [anon_sym_DOT] = ACTIONS(4248),
    [anon_sym_DASH_GT] = ACTIONS(4250),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4248),
    [anon_sym___typeof] = ACTIONS(4248),
    [anon_sym_typeof] = ACTIONS(4248),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4248),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4248),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4248),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4248),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4248),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4248),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4248),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4248),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4248),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4248),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4248),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4248),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4248),
    [anon_sym_API_AVAILABLE] = ACTIONS(4248),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4248),
    [anon_sym_API_DEPRECATED] = ACTIONS(4248),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4248),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4248),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4248),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4248),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4248),
    [anon_sym___deprecated_msg] = ACTIONS(4248),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4248),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4248),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4248),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4248),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4248),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4248),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4248),
    [anon_sym_BOOL] = ACTIONS(4248),
    [anon_sym_IMP] = ACTIONS(4248),
    [anon_sym_SEL] = ACTIONS(4248),
    [anon_sym_Class] = ACTIONS(4248),
    [anon_sym_id] = ACTIONS(4248),
  },
  [STATE(1503)] = {
    [sym__declaration_modifiers] = STATE(1574),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(1574),
    [sym_attribute_declaration] = STATE(1574),
    [sym_ms_declspec_modifier] = STATE(1574),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5206),
    [sym__abstract_declarator] = STATE(5183),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_compound_statement] = STATE(6762),
    [sym_storage_class_specifier] = STATE(1574),
    [sym_type_qualifier] = STATE(1574),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5455),
    [sym_parameter_list] = STATE(4271),
    [sym_parameter_declaration] = STATE(5455),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1574),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1574),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4078),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym_RPAREN] = ACTIONS(4082),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(4086),
    [anon_sym_CARET] = ACTIONS(4088),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1504)] = {
    [sym_identifier] = ACTIONS(4252),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4255),
    [anon_sym_COMMA] = ACTIONS(4255),
    [anon_sym_LPAREN2] = ACTIONS(4255),
    [anon_sym_DASH] = ACTIONS(4252),
    [anon_sym_PLUS] = ACTIONS(4252),
    [anon_sym_STAR] = ACTIONS(4252),
    [anon_sym_SLASH] = ACTIONS(4252),
    [anon_sym_PERCENT] = ACTIONS(4252),
    [anon_sym_PIPE_PIPE] = ACTIONS(4255),
    [anon_sym_AMP_AMP] = ACTIONS(4255),
    [anon_sym_PIPE] = ACTIONS(4252),
    [anon_sym_CARET] = ACTIONS(4252),
    [anon_sym_AMP] = ACTIONS(4252),
    [anon_sym_EQ_EQ] = ACTIONS(4255),
    [anon_sym_BANG_EQ] = ACTIONS(4255),
    [anon_sym_GT] = ACTIONS(4252),
    [anon_sym_GT_EQ] = ACTIONS(4255),
    [anon_sym_LT_EQ] = ACTIONS(4255),
    [anon_sym_LT] = ACTIONS(4252),
    [anon_sym_LT_LT] = ACTIONS(4252),
    [anon_sym_GT_GT] = ACTIONS(4252),
    [anon_sym_SEMI] = ACTIONS(4255),
    [anon_sym___extension__] = ACTIONS(4252),
    [anon_sym_extern] = ACTIONS(4252),
    [anon_sym___attribute__] = ACTIONS(4252),
    [anon_sym___attribute] = ACTIONS(4252),
    [anon_sym_noreturn] = ACTIONS(4252),
    [anon_sym_LBRACK] = ACTIONS(4255),
    [anon_sym___declspec] = ACTIONS(4252),
    [anon_sym_LBRACE] = ACTIONS(4255),
    [anon_sym_signed] = ACTIONS(4252),
    [anon_sym_unsigned] = ACTIONS(4252),
    [anon_sym_long] = ACTIONS(4252),
    [anon_sym_short] = ACTIONS(4252),
    [anon_sym_static] = ACTIONS(4252),
    [anon_sym_EQ] = ACTIONS(4252),
    [anon_sym_ATautoreleasepool] = ACTIONS(4255),
    [anon_sym_auto] = ACTIONS(4252),
    [anon_sym_register] = ACTIONS(4252),
    [anon_sym_inline] = ACTIONS(4252),
    [anon_sym___inline] = ACTIONS(4252),
    [anon_sym___inline__] = ACTIONS(4252),
    [anon_sym___forceinline] = ACTIONS(4252),
    [anon_sym_thread_local] = ACTIONS(4252),
    [anon_sym___thread] = ACTIONS(4252),
    [anon_sym_CG_EXTERN] = ACTIONS(4252),
    [anon_sym_CG_INLINE] = ACTIONS(4252),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4252),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4252),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4252),
    [anon_sym_IBOutlet] = ACTIONS(4252),
    [anon_sym_IBInspectable] = ACTIONS(4252),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4252),
    [anon_sym_NS_INLINE] = ACTIONS(4252),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4252),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4252),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4252),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4252),
    [anon_sym_const] = ACTIONS(4252),
    [anon_sym_constexpr] = ACTIONS(4252),
    [anon_sym_volatile] = ACTIONS(4252),
    [anon_sym_restrict] = ACTIONS(4252),
    [anon_sym___restrict__] = ACTIONS(4252),
    [anon_sym__Atomic] = ACTIONS(4252),
    [anon_sym__Noreturn] = ACTIONS(4252),
    [anon_sym__Nonnull] = ACTIONS(4252),
    [anon_sym_nullable] = ACTIONS(4252),
    [anon_sym__Complex] = ACTIONS(4252),
    [anon_sym__Nullable] = ACTIONS(4252),
    [anon_sym__Nullable_result] = ACTIONS(4252),
    [anon_sym__Null_unspecified] = ACTIONS(4252),
    [anon_sym___autoreleasing] = ACTIONS(4252),
    [anon_sym___block] = ACTIONS(4252),
    [anon_sym___bridge] = ACTIONS(4252),
    [anon_sym___bridge_retained] = ACTIONS(4252),
    [anon_sym___bridge_transfer] = ACTIONS(4252),
    [anon_sym___complex] = ACTIONS(4252),
    [anon_sym___const] = ACTIONS(4252),
    [anon_sym___imag] = ACTIONS(4252),
    [anon_sym___kindof] = ACTIONS(4252),
    [anon_sym___nonnull] = ACTIONS(4252),
    [anon_sym___nullable] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4252),
    [anon_sym___real] = ACTIONS(4252),
    [anon_sym___strong] = ACTIONS(4252),
    [anon_sym___unsafe_unretained] = ACTIONS(4252),
    [anon_sym___unused] = ACTIONS(4252),
    [anon_sym___weak] = ACTIONS(4252),
    [anon_sym_alignas] = ACTIONS(4252),
    [anon_sym__Alignas] = ACTIONS(4252),
    [sym_primitive_type] = ACTIONS(4252),
    [anon_sym_enum] = ACTIONS(4252),
    [anon_sym_struct] = ACTIONS(4252),
    [anon_sym_union] = ACTIONS(4252),
    [anon_sym_QMARK] = ACTIONS(4255),
    [anon_sym_STAR_EQ] = ACTIONS(4255),
    [anon_sym_SLASH_EQ] = ACTIONS(4255),
    [anon_sym_PERCENT_EQ] = ACTIONS(4255),
    [anon_sym_PLUS_EQ] = ACTIONS(4255),
    [anon_sym_DASH_EQ] = ACTIONS(4255),
    [anon_sym_LT_LT_EQ] = ACTIONS(4255),
    [anon_sym_GT_GT_EQ] = ACTIONS(4255),
    [anon_sym_AMP_EQ] = ACTIONS(4255),
    [anon_sym_CARET_EQ] = ACTIONS(4255),
    [anon_sym_PIPE_EQ] = ACTIONS(4255),
    [anon_sym_DASH_DASH] = ACTIONS(4255),
    [anon_sym_PLUS_PLUS] = ACTIONS(4255),
    [anon_sym_DOT] = ACTIONS(4252),
    [anon_sym_DASH_GT] = ACTIONS(4255),
    [anon_sym_AT] = ACTIONS(4252),
    [anon_sym_DQUOTE] = ACTIONS(4255),
    [anon_sym_L_DQUOTE] = ACTIONS(4255),
    [anon_sym_u_DQUOTE] = ACTIONS(4255),
    [anon_sym_U_DQUOTE] = ACTIONS(4255),
    [anon_sym_u8_DQUOTE] = ACTIONS(4255),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4252),
    [anon_sym___typeof] = ACTIONS(4252),
    [anon_sym_typeof] = ACTIONS(4252),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4252),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4252),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4252),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4252),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4252),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4252),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4252),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_API_AVAILABLE] = ACTIONS(4252),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_API_DEPRECATED] = ACTIONS(4252),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4252),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym___deprecated_msg] = ACTIONS(4252),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4252),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4252),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4252),
    [anon_sym_BOOL] = ACTIONS(4252),
    [anon_sym_IMP] = ACTIONS(4252),
    [anon_sym_SEL] = ACTIONS(4252),
    [anon_sym_Class] = ACTIONS(4252),
    [anon_sym_id] = ACTIONS(4252),
  },
  [STATE(1505)] = {
    [sym_identifier] = ACTIONS(2360),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2362),
    [anon_sym_COMMA] = ACTIONS(2362),
    [anon_sym_RPAREN] = ACTIONS(2362),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2362),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2362),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2360),
    [anon_sym_SLASH] = ACTIONS(2360),
    [anon_sym_PERCENT] = ACTIONS(2360),
    [anon_sym_PIPE_PIPE] = ACTIONS(2362),
    [anon_sym_AMP_AMP] = ACTIONS(2362),
    [anon_sym_PIPE] = ACTIONS(2360),
    [anon_sym_CARET] = ACTIONS(2360),
    [anon_sym_AMP] = ACTIONS(2360),
    [anon_sym_EQ_EQ] = ACTIONS(2362),
    [anon_sym_BANG_EQ] = ACTIONS(2362),
    [anon_sym_GT] = ACTIONS(2360),
    [anon_sym_GT_EQ] = ACTIONS(2362),
    [anon_sym_LT_EQ] = ACTIONS(2362),
    [anon_sym_LT] = ACTIONS(2360),
    [anon_sym_LT_LT] = ACTIONS(2360),
    [anon_sym_GT_GT] = ACTIONS(2360),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym_RBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_RBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_EQ] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_COLON] = ACTIONS(2362),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_QMARK] = ACTIONS(2362),
    [anon_sym_STAR_EQ] = ACTIONS(2362),
    [anon_sym_SLASH_EQ] = ACTIONS(2362),
    [anon_sym_PERCENT_EQ] = ACTIONS(2362),
    [anon_sym_PLUS_EQ] = ACTIONS(2362),
    [anon_sym_DASH_EQ] = ACTIONS(2362),
    [anon_sym_LT_LT_EQ] = ACTIONS(2362),
    [anon_sym_GT_GT_EQ] = ACTIONS(2362),
    [anon_sym_AMP_EQ] = ACTIONS(2362),
    [anon_sym_CARET_EQ] = ACTIONS(2362),
    [anon_sym_PIPE_EQ] = ACTIONS(2362),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_DOT] = ACTIONS(2360),
    [anon_sym_DASH_GT] = ACTIONS(2362),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(1506)] = {
    [sym_identifier] = ACTIONS(4258),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4260),
    [anon_sym_COMMA] = ACTIONS(4260),
    [anon_sym_LPAREN2] = ACTIONS(4260),
    [anon_sym_DASH] = ACTIONS(4258),
    [anon_sym_PLUS] = ACTIONS(4258),
    [anon_sym_STAR] = ACTIONS(4258),
    [anon_sym_SLASH] = ACTIONS(4258),
    [anon_sym_PERCENT] = ACTIONS(4258),
    [anon_sym_PIPE_PIPE] = ACTIONS(4260),
    [anon_sym_AMP_AMP] = ACTIONS(4260),
    [anon_sym_PIPE] = ACTIONS(4258),
    [anon_sym_CARET] = ACTIONS(4258),
    [anon_sym_AMP] = ACTIONS(4258),
    [anon_sym_EQ_EQ] = ACTIONS(4260),
    [anon_sym_BANG_EQ] = ACTIONS(4260),
    [anon_sym_GT] = ACTIONS(4258),
    [anon_sym_GT_EQ] = ACTIONS(4260),
    [anon_sym_LT_EQ] = ACTIONS(4260),
    [anon_sym_LT] = ACTIONS(4258),
    [anon_sym_LT_LT] = ACTIONS(4258),
    [anon_sym_GT_GT] = ACTIONS(4258),
    [anon_sym_SEMI] = ACTIONS(4260),
    [anon_sym___extension__] = ACTIONS(4258),
    [anon_sym_extern] = ACTIONS(4258),
    [anon_sym___attribute__] = ACTIONS(4258),
    [anon_sym___attribute] = ACTIONS(4258),
    [anon_sym_noreturn] = ACTIONS(4258),
    [anon_sym_LBRACK] = ACTIONS(4260),
    [anon_sym___declspec] = ACTIONS(4258),
    [anon_sym_LBRACE] = ACTIONS(4260),
    [anon_sym_signed] = ACTIONS(4258),
    [anon_sym_unsigned] = ACTIONS(4258),
    [anon_sym_long] = ACTIONS(4258),
    [anon_sym_short] = ACTIONS(4258),
    [anon_sym_static] = ACTIONS(4258),
    [anon_sym_EQ] = ACTIONS(4258),
    [anon_sym_ATautoreleasepool] = ACTIONS(4260),
    [anon_sym_auto] = ACTIONS(4258),
    [anon_sym_register] = ACTIONS(4258),
    [anon_sym_inline] = ACTIONS(4258),
    [anon_sym___inline] = ACTIONS(4258),
    [anon_sym___inline__] = ACTIONS(4258),
    [anon_sym___forceinline] = ACTIONS(4258),
    [anon_sym_thread_local] = ACTIONS(4258),
    [anon_sym___thread] = ACTIONS(4258),
    [anon_sym_CG_EXTERN] = ACTIONS(4258),
    [anon_sym_CG_INLINE] = ACTIONS(4258),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4258),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4258),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4258),
    [anon_sym_IBOutlet] = ACTIONS(4258),
    [anon_sym_IBInspectable] = ACTIONS(4258),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4258),
    [anon_sym_NS_INLINE] = ACTIONS(4258),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4258),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4258),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4258),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4258),
    [anon_sym_const] = ACTIONS(4258),
    [anon_sym_constexpr] = ACTIONS(4258),
    [anon_sym_volatile] = ACTIONS(4258),
    [anon_sym_restrict] = ACTIONS(4258),
    [anon_sym___restrict__] = ACTIONS(4258),
    [anon_sym__Atomic] = ACTIONS(4258),
    [anon_sym__Noreturn] = ACTIONS(4258),
    [anon_sym__Nonnull] = ACTIONS(4258),
    [anon_sym_nullable] = ACTIONS(4258),
    [anon_sym__Complex] = ACTIONS(4258),
    [anon_sym__Nullable] = ACTIONS(4258),
    [anon_sym__Nullable_result] = ACTIONS(4258),
    [anon_sym__Null_unspecified] = ACTIONS(4258),
    [anon_sym___autoreleasing] = ACTIONS(4258),
    [anon_sym___block] = ACTIONS(4258),
    [anon_sym___bridge] = ACTIONS(4258),
    [anon_sym___bridge_retained] = ACTIONS(4258),
    [anon_sym___bridge_transfer] = ACTIONS(4258),
    [anon_sym___complex] = ACTIONS(4258),
    [anon_sym___const] = ACTIONS(4258),
    [anon_sym___imag] = ACTIONS(4258),
    [anon_sym___kindof] = ACTIONS(4258),
    [anon_sym___nonnull] = ACTIONS(4258),
    [anon_sym___nullable] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4258),
    [anon_sym___real] = ACTIONS(4258),
    [anon_sym___strong] = ACTIONS(4258),
    [anon_sym___unsafe_unretained] = ACTIONS(4258),
    [anon_sym___unused] = ACTIONS(4258),
    [anon_sym___weak] = ACTIONS(4258),
    [anon_sym_alignas] = ACTIONS(4258),
    [anon_sym__Alignas] = ACTIONS(4258),
    [sym_primitive_type] = ACTIONS(4258),
    [anon_sym_enum] = ACTIONS(4258),
    [anon_sym_struct] = ACTIONS(4258),
    [anon_sym_union] = ACTIONS(4258),
    [anon_sym_QMARK] = ACTIONS(4260),
    [anon_sym_STAR_EQ] = ACTIONS(4260),
    [anon_sym_SLASH_EQ] = ACTIONS(4260),
    [anon_sym_PERCENT_EQ] = ACTIONS(4260),
    [anon_sym_PLUS_EQ] = ACTIONS(4260),
    [anon_sym_DASH_EQ] = ACTIONS(4260),
    [anon_sym_LT_LT_EQ] = ACTIONS(4260),
    [anon_sym_GT_GT_EQ] = ACTIONS(4260),
    [anon_sym_AMP_EQ] = ACTIONS(4260),
    [anon_sym_CARET_EQ] = ACTIONS(4260),
    [anon_sym_PIPE_EQ] = ACTIONS(4260),
    [anon_sym_DASH_DASH] = ACTIONS(4260),
    [anon_sym_PLUS_PLUS] = ACTIONS(4260),
    [anon_sym_DOT] = ACTIONS(4258),
    [anon_sym_DASH_GT] = ACTIONS(4260),
    [anon_sym_AT] = ACTIONS(4258),
    [anon_sym_DQUOTE] = ACTIONS(4260),
    [anon_sym_L_DQUOTE] = ACTIONS(4260),
    [anon_sym_u_DQUOTE] = ACTIONS(4260),
    [anon_sym_U_DQUOTE] = ACTIONS(4260),
    [anon_sym_u8_DQUOTE] = ACTIONS(4260),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4258),
    [anon_sym___typeof] = ACTIONS(4258),
    [anon_sym_typeof] = ACTIONS(4258),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4258),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4258),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4258),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4258),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4258),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4258),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4258),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_API_AVAILABLE] = ACTIONS(4258),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_API_DEPRECATED] = ACTIONS(4258),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4258),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym___deprecated_msg] = ACTIONS(4258),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4258),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4258),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4258),
    [anon_sym_BOOL] = ACTIONS(4258),
    [anon_sym_IMP] = ACTIONS(4258),
    [anon_sym_SEL] = ACTIONS(4258),
    [anon_sym_Class] = ACTIONS(4258),
    [anon_sym_id] = ACTIONS(4258),
  },
  [STATE(1507)] = {
    [sym_identifier] = ACTIONS(4262),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4264),
    [anon_sym_COMMA] = ACTIONS(4264),
    [anon_sym_LPAREN2] = ACTIONS(4264),
    [anon_sym_DASH] = ACTIONS(4262),
    [anon_sym_PLUS] = ACTIONS(4262),
    [anon_sym_STAR] = ACTIONS(4262),
    [anon_sym_SLASH] = ACTIONS(4262),
    [anon_sym_PERCENT] = ACTIONS(4262),
    [anon_sym_PIPE_PIPE] = ACTIONS(4264),
    [anon_sym_AMP_AMP] = ACTIONS(4264),
    [anon_sym_PIPE] = ACTIONS(4262),
    [anon_sym_CARET] = ACTIONS(4262),
    [anon_sym_AMP] = ACTIONS(4262),
    [anon_sym_EQ_EQ] = ACTIONS(4264),
    [anon_sym_BANG_EQ] = ACTIONS(4264),
    [anon_sym_GT] = ACTIONS(4262),
    [anon_sym_GT_EQ] = ACTIONS(4264),
    [anon_sym_LT_EQ] = ACTIONS(4264),
    [anon_sym_LT] = ACTIONS(4262),
    [anon_sym_LT_LT] = ACTIONS(4262),
    [anon_sym_GT_GT] = ACTIONS(4262),
    [anon_sym_SEMI] = ACTIONS(4264),
    [anon_sym___extension__] = ACTIONS(4262),
    [anon_sym_extern] = ACTIONS(4262),
    [anon_sym___attribute__] = ACTIONS(4262),
    [anon_sym___attribute] = ACTIONS(4262),
    [anon_sym_noreturn] = ACTIONS(4262),
    [anon_sym_LBRACK] = ACTIONS(4264),
    [anon_sym___declspec] = ACTIONS(4262),
    [anon_sym_LBRACE] = ACTIONS(4264),
    [anon_sym_signed] = ACTIONS(4262),
    [anon_sym_unsigned] = ACTIONS(4262),
    [anon_sym_long] = ACTIONS(4262),
    [anon_sym_short] = ACTIONS(4262),
    [anon_sym_static] = ACTIONS(4262),
    [anon_sym_EQ] = ACTIONS(4262),
    [anon_sym_ATautoreleasepool] = ACTIONS(4264),
    [anon_sym_auto] = ACTIONS(4262),
    [anon_sym_register] = ACTIONS(4262),
    [anon_sym_inline] = ACTIONS(4262),
    [anon_sym___inline] = ACTIONS(4262),
    [anon_sym___inline__] = ACTIONS(4262),
    [anon_sym___forceinline] = ACTIONS(4262),
    [anon_sym_thread_local] = ACTIONS(4262),
    [anon_sym___thread] = ACTIONS(4262),
    [anon_sym_CG_EXTERN] = ACTIONS(4262),
    [anon_sym_CG_INLINE] = ACTIONS(4262),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4262),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4262),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4262),
    [anon_sym_IBOutlet] = ACTIONS(4262),
    [anon_sym_IBInspectable] = ACTIONS(4262),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4262),
    [anon_sym_NS_INLINE] = ACTIONS(4262),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4262),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4262),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4262),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4262),
    [anon_sym_const] = ACTIONS(4262),
    [anon_sym_constexpr] = ACTIONS(4262),
    [anon_sym_volatile] = ACTIONS(4262),
    [anon_sym_restrict] = ACTIONS(4262),
    [anon_sym___restrict__] = ACTIONS(4262),
    [anon_sym__Atomic] = ACTIONS(4262),
    [anon_sym__Noreturn] = ACTIONS(4262),
    [anon_sym__Nonnull] = ACTIONS(4262),
    [anon_sym_nullable] = ACTIONS(4262),
    [anon_sym__Complex] = ACTIONS(4262),
    [anon_sym__Nullable] = ACTIONS(4262),
    [anon_sym__Nullable_result] = ACTIONS(4262),
    [anon_sym__Null_unspecified] = ACTIONS(4262),
    [anon_sym___autoreleasing] = ACTIONS(4262),
    [anon_sym___block] = ACTIONS(4262),
    [anon_sym___bridge] = ACTIONS(4262),
    [anon_sym___bridge_retained] = ACTIONS(4262),
    [anon_sym___bridge_transfer] = ACTIONS(4262),
    [anon_sym___complex] = ACTIONS(4262),
    [anon_sym___const] = ACTIONS(4262),
    [anon_sym___imag] = ACTIONS(4262),
    [anon_sym___kindof] = ACTIONS(4262),
    [anon_sym___nonnull] = ACTIONS(4262),
    [anon_sym___nullable] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4262),
    [anon_sym___real] = ACTIONS(4262),
    [anon_sym___strong] = ACTIONS(4262),
    [anon_sym___unsafe_unretained] = ACTIONS(4262),
    [anon_sym___unused] = ACTIONS(4262),
    [anon_sym___weak] = ACTIONS(4262),
    [anon_sym_alignas] = ACTIONS(4262),
    [anon_sym__Alignas] = ACTIONS(4262),
    [sym_primitive_type] = ACTIONS(4262),
    [anon_sym_enum] = ACTIONS(4262),
    [anon_sym_struct] = ACTIONS(4262),
    [anon_sym_union] = ACTIONS(4262),
    [anon_sym_QMARK] = ACTIONS(4264),
    [anon_sym_STAR_EQ] = ACTIONS(4264),
    [anon_sym_SLASH_EQ] = ACTIONS(4264),
    [anon_sym_PERCENT_EQ] = ACTIONS(4264),
    [anon_sym_PLUS_EQ] = ACTIONS(4264),
    [anon_sym_DASH_EQ] = ACTIONS(4264),
    [anon_sym_LT_LT_EQ] = ACTIONS(4264),
    [anon_sym_GT_GT_EQ] = ACTIONS(4264),
    [anon_sym_AMP_EQ] = ACTIONS(4264),
    [anon_sym_CARET_EQ] = ACTIONS(4264),
    [anon_sym_PIPE_EQ] = ACTIONS(4264),
    [anon_sym_DASH_DASH] = ACTIONS(4264),
    [anon_sym_PLUS_PLUS] = ACTIONS(4264),
    [anon_sym_DOT] = ACTIONS(4262),
    [anon_sym_DASH_GT] = ACTIONS(4264),
    [anon_sym_AT] = ACTIONS(4262),
    [anon_sym_DQUOTE] = ACTIONS(4264),
    [anon_sym_L_DQUOTE] = ACTIONS(4264),
    [anon_sym_u_DQUOTE] = ACTIONS(4264),
    [anon_sym_U_DQUOTE] = ACTIONS(4264),
    [anon_sym_u8_DQUOTE] = ACTIONS(4264),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4262),
    [anon_sym___typeof] = ACTIONS(4262),
    [anon_sym_typeof] = ACTIONS(4262),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4262),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4262),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4262),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4262),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4262),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4262),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4262),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_API_AVAILABLE] = ACTIONS(4262),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_API_DEPRECATED] = ACTIONS(4262),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4262),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym___deprecated_msg] = ACTIONS(4262),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4262),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4262),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4262),
    [anon_sym_BOOL] = ACTIONS(4262),
    [anon_sym_IMP] = ACTIONS(4262),
    [anon_sym_SEL] = ACTIONS(4262),
    [anon_sym_Class] = ACTIONS(4262),
    [anon_sym_id] = ACTIONS(4262),
  },
  [STATE(1508)] = {
    [sym_identifier] = ACTIONS(4266),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4268),
    [anon_sym_COMMA] = ACTIONS(4268),
    [anon_sym_RPAREN] = ACTIONS(4268),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4268),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4268),
    [anon_sym_LPAREN2] = ACTIONS(4268),
    [anon_sym_DASH] = ACTIONS(4266),
    [anon_sym_PLUS] = ACTIONS(4266),
    [anon_sym_STAR] = ACTIONS(4266),
    [anon_sym_SLASH] = ACTIONS(4266),
    [anon_sym_PERCENT] = ACTIONS(4266),
    [anon_sym_PIPE_PIPE] = ACTIONS(4268),
    [anon_sym_AMP_AMP] = ACTIONS(4268),
    [anon_sym_PIPE] = ACTIONS(4266),
    [anon_sym_CARET] = ACTIONS(4266),
    [anon_sym_AMP] = ACTIONS(4266),
    [anon_sym_EQ_EQ] = ACTIONS(4268),
    [anon_sym_BANG_EQ] = ACTIONS(4268),
    [anon_sym_GT] = ACTIONS(4266),
    [anon_sym_GT_EQ] = ACTIONS(4268),
    [anon_sym_LT_EQ] = ACTIONS(4268),
    [anon_sym_LT] = ACTIONS(4266),
    [anon_sym_LT_LT] = ACTIONS(4266),
    [anon_sym_GT_GT] = ACTIONS(4266),
    [anon_sym_SEMI] = ACTIONS(4268),
    [anon_sym___extension__] = ACTIONS(4266),
    [anon_sym_extern] = ACTIONS(4266),
    [anon_sym___attribute__] = ACTIONS(4266),
    [anon_sym___attribute] = ACTIONS(4266),
    [anon_sym_noreturn] = ACTIONS(4266),
    [anon_sym_LBRACK] = ACTIONS(4268),
    [anon_sym_RBRACK] = ACTIONS(4268),
    [anon_sym___declspec] = ACTIONS(4266),
    [anon_sym_LBRACE] = ACTIONS(4268),
    [anon_sym_RBRACE] = ACTIONS(4268),
    [anon_sym_signed] = ACTIONS(4266),
    [anon_sym_unsigned] = ACTIONS(4266),
    [anon_sym_long] = ACTIONS(4266),
    [anon_sym_short] = ACTIONS(4266),
    [anon_sym_static] = ACTIONS(4266),
    [anon_sym_EQ] = ACTIONS(4266),
    [anon_sym_ATautoreleasepool] = ACTIONS(4268),
    [anon_sym_auto] = ACTIONS(4266),
    [anon_sym_register] = ACTIONS(4266),
    [anon_sym_inline] = ACTIONS(4266),
    [anon_sym___inline] = ACTIONS(4266),
    [anon_sym___inline__] = ACTIONS(4266),
    [anon_sym___forceinline] = ACTIONS(4266),
    [anon_sym_thread_local] = ACTIONS(4266),
    [anon_sym___thread] = ACTIONS(4266),
    [anon_sym_CG_EXTERN] = ACTIONS(4266),
    [anon_sym_CG_INLINE] = ACTIONS(4266),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4266),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4266),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4266),
    [anon_sym_IBOutlet] = ACTIONS(4266),
    [anon_sym_IBInspectable] = ACTIONS(4266),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4266),
    [anon_sym_NS_INLINE] = ACTIONS(4266),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4266),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4266),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4266),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4266),
    [anon_sym_const] = ACTIONS(4266),
    [anon_sym_constexpr] = ACTIONS(4266),
    [anon_sym_volatile] = ACTIONS(4266),
    [anon_sym_restrict] = ACTIONS(4266),
    [anon_sym___restrict__] = ACTIONS(4266),
    [anon_sym__Atomic] = ACTIONS(4266),
    [anon_sym__Noreturn] = ACTIONS(4266),
    [anon_sym__Nonnull] = ACTIONS(4266),
    [anon_sym_nullable] = ACTIONS(4266),
    [anon_sym__Complex] = ACTIONS(4266),
    [anon_sym__Nullable] = ACTIONS(4266),
    [anon_sym__Nullable_result] = ACTIONS(4266),
    [anon_sym__Null_unspecified] = ACTIONS(4266),
    [anon_sym___autoreleasing] = ACTIONS(4266),
    [anon_sym___block] = ACTIONS(4266),
    [anon_sym___bridge] = ACTIONS(4266),
    [anon_sym___bridge_retained] = ACTIONS(4266),
    [anon_sym___bridge_transfer] = ACTIONS(4266),
    [anon_sym___complex] = ACTIONS(4266),
    [anon_sym___const] = ACTIONS(4266),
    [anon_sym___imag] = ACTIONS(4266),
    [anon_sym___kindof] = ACTIONS(4266),
    [anon_sym___nonnull] = ACTIONS(4266),
    [anon_sym___nullable] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4266),
    [anon_sym___real] = ACTIONS(4266),
    [anon_sym___strong] = ACTIONS(4266),
    [anon_sym___unsafe_unretained] = ACTIONS(4266),
    [anon_sym___unused] = ACTIONS(4266),
    [anon_sym___weak] = ACTIONS(4266),
    [anon_sym_alignas] = ACTIONS(4266),
    [anon_sym__Alignas] = ACTIONS(4266),
    [sym_primitive_type] = ACTIONS(4266),
    [anon_sym_enum] = ACTIONS(4266),
    [anon_sym_COLON] = ACTIONS(4268),
    [anon_sym_struct] = ACTIONS(4266),
    [anon_sym_union] = ACTIONS(4266),
    [anon_sym_QMARK] = ACTIONS(4268),
    [anon_sym_STAR_EQ] = ACTIONS(4268),
    [anon_sym_SLASH_EQ] = ACTIONS(4268),
    [anon_sym_PERCENT_EQ] = ACTIONS(4268),
    [anon_sym_PLUS_EQ] = ACTIONS(4268),
    [anon_sym_DASH_EQ] = ACTIONS(4268),
    [anon_sym_LT_LT_EQ] = ACTIONS(4268),
    [anon_sym_GT_GT_EQ] = ACTIONS(4268),
    [anon_sym_AMP_EQ] = ACTIONS(4268),
    [anon_sym_CARET_EQ] = ACTIONS(4268),
    [anon_sym_PIPE_EQ] = ACTIONS(4268),
    [anon_sym_DASH_DASH] = ACTIONS(4268),
    [anon_sym_PLUS_PLUS] = ACTIONS(4268),
    [anon_sym_DOT] = ACTIONS(4266),
    [anon_sym_DASH_GT] = ACTIONS(4268),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4266),
    [anon_sym___typeof] = ACTIONS(4266),
    [anon_sym_typeof] = ACTIONS(4266),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4266),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4266),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4266),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4266),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4266),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4266),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4266),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4266),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4266),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_API_AVAILABLE] = ACTIONS(4266),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_API_DEPRECATED] = ACTIONS(4266),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4266),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4266),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4266),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4266),
    [anon_sym___deprecated_msg] = ACTIONS(4266),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4266),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4266),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4266),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4266),
    [anon_sym_BOOL] = ACTIONS(4266),
    [anon_sym_IMP] = ACTIONS(4266),
    [anon_sym_SEL] = ACTIONS(4266),
    [anon_sym_Class] = ACTIONS(4266),
    [anon_sym_id] = ACTIONS(4266),
  },
  [STATE(1509)] = {
    [sym_identifier] = ACTIONS(4270),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4272),
    [anon_sym_COMMA] = ACTIONS(4272),
    [anon_sym_RPAREN] = ACTIONS(4272),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4272),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4272),
    [anon_sym_LPAREN2] = ACTIONS(4272),
    [anon_sym_DASH] = ACTIONS(4270),
    [anon_sym_PLUS] = ACTIONS(4270),
    [anon_sym_STAR] = ACTIONS(4270),
    [anon_sym_SLASH] = ACTIONS(4270),
    [anon_sym_PERCENT] = ACTIONS(4270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4272),
    [anon_sym_AMP_AMP] = ACTIONS(4272),
    [anon_sym_PIPE] = ACTIONS(4270),
    [anon_sym_CARET] = ACTIONS(4270),
    [anon_sym_AMP] = ACTIONS(4270),
    [anon_sym_EQ_EQ] = ACTIONS(4272),
    [anon_sym_BANG_EQ] = ACTIONS(4272),
    [anon_sym_GT] = ACTIONS(4270),
    [anon_sym_GT_EQ] = ACTIONS(4272),
    [anon_sym_LT_EQ] = ACTIONS(4272),
    [anon_sym_LT] = ACTIONS(4270),
    [anon_sym_LT_LT] = ACTIONS(4270),
    [anon_sym_GT_GT] = ACTIONS(4270),
    [anon_sym_SEMI] = ACTIONS(4272),
    [anon_sym___extension__] = ACTIONS(4270),
    [anon_sym_extern] = ACTIONS(4270),
    [anon_sym___attribute__] = ACTIONS(4270),
    [anon_sym___attribute] = ACTIONS(4270),
    [anon_sym_noreturn] = ACTIONS(4270),
    [anon_sym_LBRACK] = ACTIONS(4272),
    [anon_sym_RBRACK] = ACTIONS(4272),
    [anon_sym___declspec] = ACTIONS(4270),
    [anon_sym_LBRACE] = ACTIONS(4272),
    [anon_sym_RBRACE] = ACTIONS(4272),
    [anon_sym_signed] = ACTIONS(4270),
    [anon_sym_unsigned] = ACTIONS(4270),
    [anon_sym_long] = ACTIONS(4270),
    [anon_sym_short] = ACTIONS(4270),
    [anon_sym_static] = ACTIONS(4270),
    [anon_sym_EQ] = ACTIONS(4270),
    [anon_sym_ATautoreleasepool] = ACTIONS(4272),
    [anon_sym_auto] = ACTIONS(4270),
    [anon_sym_register] = ACTIONS(4270),
    [anon_sym_inline] = ACTIONS(4270),
    [anon_sym___inline] = ACTIONS(4270),
    [anon_sym___inline__] = ACTIONS(4270),
    [anon_sym___forceinline] = ACTIONS(4270),
    [anon_sym_thread_local] = ACTIONS(4270),
    [anon_sym___thread] = ACTIONS(4270),
    [anon_sym_CG_EXTERN] = ACTIONS(4270),
    [anon_sym_CG_INLINE] = ACTIONS(4270),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4270),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4270),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4270),
    [anon_sym_IBOutlet] = ACTIONS(4270),
    [anon_sym_IBInspectable] = ACTIONS(4270),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4270),
    [anon_sym_NS_INLINE] = ACTIONS(4270),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4270),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4270),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4270),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4270),
    [anon_sym_const] = ACTIONS(4270),
    [anon_sym_constexpr] = ACTIONS(4270),
    [anon_sym_volatile] = ACTIONS(4270),
    [anon_sym_restrict] = ACTIONS(4270),
    [anon_sym___restrict__] = ACTIONS(4270),
    [anon_sym__Atomic] = ACTIONS(4270),
    [anon_sym__Noreturn] = ACTIONS(4270),
    [anon_sym__Nonnull] = ACTIONS(4270),
    [anon_sym_nullable] = ACTIONS(4270),
    [anon_sym__Complex] = ACTIONS(4270),
    [anon_sym__Nullable] = ACTIONS(4270),
    [anon_sym__Nullable_result] = ACTIONS(4270),
    [anon_sym__Null_unspecified] = ACTIONS(4270),
    [anon_sym___autoreleasing] = ACTIONS(4270),
    [anon_sym___block] = ACTIONS(4270),
    [anon_sym___bridge] = ACTIONS(4270),
    [anon_sym___bridge_retained] = ACTIONS(4270),
    [anon_sym___bridge_transfer] = ACTIONS(4270),
    [anon_sym___complex] = ACTIONS(4270),
    [anon_sym___const] = ACTIONS(4270),
    [anon_sym___imag] = ACTIONS(4270),
    [anon_sym___kindof] = ACTIONS(4270),
    [anon_sym___nonnull] = ACTIONS(4270),
    [anon_sym___nullable] = ACTIONS(4270),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4270),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4270),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4270),
    [anon_sym___real] = ACTIONS(4270),
    [anon_sym___strong] = ACTIONS(4270),
    [anon_sym___unsafe_unretained] = ACTIONS(4270),
    [anon_sym___unused] = ACTIONS(4270),
    [anon_sym___weak] = ACTIONS(4270),
    [anon_sym_alignas] = ACTIONS(4270),
    [anon_sym__Alignas] = ACTIONS(4270),
    [sym_primitive_type] = ACTIONS(4270),
    [anon_sym_enum] = ACTIONS(4270),
    [anon_sym_COLON] = ACTIONS(4272),
    [anon_sym_struct] = ACTIONS(4270),
    [anon_sym_union] = ACTIONS(4270),
    [anon_sym_QMARK] = ACTIONS(4272),
    [anon_sym_STAR_EQ] = ACTIONS(4272),
    [anon_sym_SLASH_EQ] = ACTIONS(4272),
    [anon_sym_PERCENT_EQ] = ACTIONS(4272),
    [anon_sym_PLUS_EQ] = ACTIONS(4272),
    [anon_sym_DASH_EQ] = ACTIONS(4272),
    [anon_sym_LT_LT_EQ] = ACTIONS(4272),
    [anon_sym_GT_GT_EQ] = ACTIONS(4272),
    [anon_sym_AMP_EQ] = ACTIONS(4272),
    [anon_sym_CARET_EQ] = ACTIONS(4272),
    [anon_sym_PIPE_EQ] = ACTIONS(4272),
    [anon_sym_DASH_DASH] = ACTIONS(4272),
    [anon_sym_PLUS_PLUS] = ACTIONS(4272),
    [anon_sym_DOT] = ACTIONS(4270),
    [anon_sym_DASH_GT] = ACTIONS(4272),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4270),
    [anon_sym___typeof] = ACTIONS(4270),
    [anon_sym_typeof] = ACTIONS(4270),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4270),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4270),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4270),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4270),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4270),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4270),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4270),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4270),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4270),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4270),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4270),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4270),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4270),
    [anon_sym_API_AVAILABLE] = ACTIONS(4270),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4270),
    [anon_sym_API_DEPRECATED] = ACTIONS(4270),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4270),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4270),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4270),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4270),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4270),
    [anon_sym___deprecated_msg] = ACTIONS(4270),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4270),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4270),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4270),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4270),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4270),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4270),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4270),
    [anon_sym_BOOL] = ACTIONS(4270),
    [anon_sym_IMP] = ACTIONS(4270),
    [anon_sym_SEL] = ACTIONS(4270),
    [anon_sym_Class] = ACTIONS(4270),
    [anon_sym_id] = ACTIONS(4270),
  },
  [STATE(1510)] = {
    [sym_identifier] = ACTIONS(4274),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4276),
    [anon_sym_COMMA] = ACTIONS(4276),
    [anon_sym_RPAREN] = ACTIONS(4276),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4276),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4276),
    [anon_sym_LPAREN2] = ACTIONS(4276),
    [anon_sym_DASH] = ACTIONS(4274),
    [anon_sym_PLUS] = ACTIONS(4274),
    [anon_sym_STAR] = ACTIONS(4274),
    [anon_sym_SLASH] = ACTIONS(4274),
    [anon_sym_PERCENT] = ACTIONS(4274),
    [anon_sym_PIPE_PIPE] = ACTIONS(4276),
    [anon_sym_AMP_AMP] = ACTIONS(4276),
    [anon_sym_PIPE] = ACTIONS(4274),
    [anon_sym_CARET] = ACTIONS(4274),
    [anon_sym_AMP] = ACTIONS(4274),
    [anon_sym_EQ_EQ] = ACTIONS(4276),
    [anon_sym_BANG_EQ] = ACTIONS(4276),
    [anon_sym_GT] = ACTIONS(4274),
    [anon_sym_GT_EQ] = ACTIONS(4276),
    [anon_sym_LT_EQ] = ACTIONS(4276),
    [anon_sym_LT] = ACTIONS(4274),
    [anon_sym_LT_LT] = ACTIONS(4274),
    [anon_sym_GT_GT] = ACTIONS(4274),
    [anon_sym_SEMI] = ACTIONS(4276),
    [anon_sym___extension__] = ACTIONS(4274),
    [anon_sym_extern] = ACTIONS(4274),
    [anon_sym___attribute__] = ACTIONS(4274),
    [anon_sym___attribute] = ACTIONS(4274),
    [anon_sym_noreturn] = ACTIONS(4274),
    [anon_sym_LBRACK] = ACTIONS(4276),
    [anon_sym_RBRACK] = ACTIONS(4276),
    [anon_sym___declspec] = ACTIONS(4274),
    [anon_sym_LBRACE] = ACTIONS(4276),
    [anon_sym_RBRACE] = ACTIONS(4276),
    [anon_sym_signed] = ACTIONS(4274),
    [anon_sym_unsigned] = ACTIONS(4274),
    [anon_sym_long] = ACTIONS(4274),
    [anon_sym_short] = ACTIONS(4274),
    [anon_sym_static] = ACTIONS(4274),
    [anon_sym_EQ] = ACTIONS(4274),
    [anon_sym_ATautoreleasepool] = ACTIONS(4276),
    [anon_sym_auto] = ACTIONS(4274),
    [anon_sym_register] = ACTIONS(4274),
    [anon_sym_inline] = ACTIONS(4274),
    [anon_sym___inline] = ACTIONS(4274),
    [anon_sym___inline__] = ACTIONS(4274),
    [anon_sym___forceinline] = ACTIONS(4274),
    [anon_sym_thread_local] = ACTIONS(4274),
    [anon_sym___thread] = ACTIONS(4274),
    [anon_sym_CG_EXTERN] = ACTIONS(4274),
    [anon_sym_CG_INLINE] = ACTIONS(4274),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4274),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4274),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4274),
    [anon_sym_IBOutlet] = ACTIONS(4274),
    [anon_sym_IBInspectable] = ACTIONS(4274),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4274),
    [anon_sym_NS_INLINE] = ACTIONS(4274),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4274),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4274),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4274),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4274),
    [anon_sym_const] = ACTIONS(4274),
    [anon_sym_constexpr] = ACTIONS(4274),
    [anon_sym_volatile] = ACTIONS(4274),
    [anon_sym_restrict] = ACTIONS(4274),
    [anon_sym___restrict__] = ACTIONS(4274),
    [anon_sym__Atomic] = ACTIONS(4274),
    [anon_sym__Noreturn] = ACTIONS(4274),
    [anon_sym__Nonnull] = ACTIONS(4274),
    [anon_sym_nullable] = ACTIONS(4274),
    [anon_sym__Complex] = ACTIONS(4274),
    [anon_sym__Nullable] = ACTIONS(4274),
    [anon_sym__Nullable_result] = ACTIONS(4274),
    [anon_sym__Null_unspecified] = ACTIONS(4274),
    [anon_sym___autoreleasing] = ACTIONS(4274),
    [anon_sym___block] = ACTIONS(4274),
    [anon_sym___bridge] = ACTIONS(4274),
    [anon_sym___bridge_retained] = ACTIONS(4274),
    [anon_sym___bridge_transfer] = ACTIONS(4274),
    [anon_sym___complex] = ACTIONS(4274),
    [anon_sym___const] = ACTIONS(4274),
    [anon_sym___imag] = ACTIONS(4274),
    [anon_sym___kindof] = ACTIONS(4274),
    [anon_sym___nonnull] = ACTIONS(4274),
    [anon_sym___nullable] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4274),
    [anon_sym___real] = ACTIONS(4274),
    [anon_sym___strong] = ACTIONS(4274),
    [anon_sym___unsafe_unretained] = ACTIONS(4274),
    [anon_sym___unused] = ACTIONS(4274),
    [anon_sym___weak] = ACTIONS(4274),
    [anon_sym_alignas] = ACTIONS(4274),
    [anon_sym__Alignas] = ACTIONS(4274),
    [sym_primitive_type] = ACTIONS(4274),
    [anon_sym_enum] = ACTIONS(4274),
    [anon_sym_COLON] = ACTIONS(4276),
    [anon_sym_struct] = ACTIONS(4274),
    [anon_sym_union] = ACTIONS(4274),
    [anon_sym_QMARK] = ACTIONS(4276),
    [anon_sym_STAR_EQ] = ACTIONS(4276),
    [anon_sym_SLASH_EQ] = ACTIONS(4276),
    [anon_sym_PERCENT_EQ] = ACTIONS(4276),
    [anon_sym_PLUS_EQ] = ACTIONS(4276),
    [anon_sym_DASH_EQ] = ACTIONS(4276),
    [anon_sym_LT_LT_EQ] = ACTIONS(4276),
    [anon_sym_GT_GT_EQ] = ACTIONS(4276),
    [anon_sym_AMP_EQ] = ACTIONS(4276),
    [anon_sym_CARET_EQ] = ACTIONS(4276),
    [anon_sym_PIPE_EQ] = ACTIONS(4276),
    [anon_sym_DASH_DASH] = ACTIONS(4276),
    [anon_sym_PLUS_PLUS] = ACTIONS(4276),
    [anon_sym_DOT] = ACTIONS(4274),
    [anon_sym_DASH_GT] = ACTIONS(4276),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4274),
    [anon_sym___typeof] = ACTIONS(4274),
    [anon_sym_typeof] = ACTIONS(4274),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4274),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4274),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4274),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4274),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4274),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4274),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4274),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4274),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4274),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_API_AVAILABLE] = ACTIONS(4274),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_API_DEPRECATED] = ACTIONS(4274),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4274),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4274),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4274),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4274),
    [anon_sym___deprecated_msg] = ACTIONS(4274),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4274),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4274),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4274),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4274),
    [anon_sym_BOOL] = ACTIONS(4274),
    [anon_sym_IMP] = ACTIONS(4274),
    [anon_sym_SEL] = ACTIONS(4274),
    [anon_sym_Class] = ACTIONS(4274),
    [anon_sym_id] = ACTIONS(4274),
  },
  [STATE(1511)] = {
    [sym_identifier] = ACTIONS(4278),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4280),
    [anon_sym_COMMA] = ACTIONS(4280),
    [anon_sym_RPAREN] = ACTIONS(4280),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4280),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4280),
    [anon_sym_LPAREN2] = ACTIONS(4280),
    [anon_sym_DASH] = ACTIONS(4278),
    [anon_sym_PLUS] = ACTIONS(4278),
    [anon_sym_STAR] = ACTIONS(4278),
    [anon_sym_SLASH] = ACTIONS(4278),
    [anon_sym_PERCENT] = ACTIONS(4278),
    [anon_sym_PIPE_PIPE] = ACTIONS(4280),
    [anon_sym_AMP_AMP] = ACTIONS(4280),
    [anon_sym_PIPE] = ACTIONS(4278),
    [anon_sym_CARET] = ACTIONS(4278),
    [anon_sym_AMP] = ACTIONS(4278),
    [anon_sym_EQ_EQ] = ACTIONS(4280),
    [anon_sym_BANG_EQ] = ACTIONS(4280),
    [anon_sym_GT] = ACTIONS(4278),
    [anon_sym_GT_EQ] = ACTIONS(4280),
    [anon_sym_LT_EQ] = ACTIONS(4280),
    [anon_sym_LT] = ACTIONS(4278),
    [anon_sym_LT_LT] = ACTIONS(4278),
    [anon_sym_GT_GT] = ACTIONS(4278),
    [anon_sym_SEMI] = ACTIONS(4280),
    [anon_sym___extension__] = ACTIONS(4278),
    [anon_sym_extern] = ACTIONS(4278),
    [anon_sym___attribute__] = ACTIONS(4278),
    [anon_sym___attribute] = ACTIONS(4278),
    [anon_sym_noreturn] = ACTIONS(4278),
    [anon_sym_LBRACK] = ACTIONS(4280),
    [anon_sym_RBRACK] = ACTIONS(4280),
    [anon_sym___declspec] = ACTIONS(4278),
    [anon_sym_LBRACE] = ACTIONS(4280),
    [anon_sym_RBRACE] = ACTIONS(4280),
    [anon_sym_signed] = ACTIONS(4278),
    [anon_sym_unsigned] = ACTIONS(4278),
    [anon_sym_long] = ACTIONS(4278),
    [anon_sym_short] = ACTIONS(4278),
    [anon_sym_static] = ACTIONS(4278),
    [anon_sym_EQ] = ACTIONS(4278),
    [anon_sym_ATautoreleasepool] = ACTIONS(4280),
    [anon_sym_auto] = ACTIONS(4278),
    [anon_sym_register] = ACTIONS(4278),
    [anon_sym_inline] = ACTIONS(4278),
    [anon_sym___inline] = ACTIONS(4278),
    [anon_sym___inline__] = ACTIONS(4278),
    [anon_sym___forceinline] = ACTIONS(4278),
    [anon_sym_thread_local] = ACTIONS(4278),
    [anon_sym___thread] = ACTIONS(4278),
    [anon_sym_CG_EXTERN] = ACTIONS(4278),
    [anon_sym_CG_INLINE] = ACTIONS(4278),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4278),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4278),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4278),
    [anon_sym_IBOutlet] = ACTIONS(4278),
    [anon_sym_IBInspectable] = ACTIONS(4278),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4278),
    [anon_sym_NS_INLINE] = ACTIONS(4278),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4278),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4278),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4278),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4278),
    [anon_sym_const] = ACTIONS(4278),
    [anon_sym_constexpr] = ACTIONS(4278),
    [anon_sym_volatile] = ACTIONS(4278),
    [anon_sym_restrict] = ACTIONS(4278),
    [anon_sym___restrict__] = ACTIONS(4278),
    [anon_sym__Atomic] = ACTIONS(4278),
    [anon_sym__Noreturn] = ACTIONS(4278),
    [anon_sym__Nonnull] = ACTIONS(4278),
    [anon_sym_nullable] = ACTIONS(4278),
    [anon_sym__Complex] = ACTIONS(4278),
    [anon_sym__Nullable] = ACTIONS(4278),
    [anon_sym__Nullable_result] = ACTIONS(4278),
    [anon_sym__Null_unspecified] = ACTIONS(4278),
    [anon_sym___autoreleasing] = ACTIONS(4278),
    [anon_sym___block] = ACTIONS(4278),
    [anon_sym___bridge] = ACTIONS(4278),
    [anon_sym___bridge_retained] = ACTIONS(4278),
    [anon_sym___bridge_transfer] = ACTIONS(4278),
    [anon_sym___complex] = ACTIONS(4278),
    [anon_sym___const] = ACTIONS(4278),
    [anon_sym___imag] = ACTIONS(4278),
    [anon_sym___kindof] = ACTIONS(4278),
    [anon_sym___nonnull] = ACTIONS(4278),
    [anon_sym___nullable] = ACTIONS(4278),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4278),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4278),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4278),
    [anon_sym___real] = ACTIONS(4278),
    [anon_sym___strong] = ACTIONS(4278),
    [anon_sym___unsafe_unretained] = ACTIONS(4278),
    [anon_sym___unused] = ACTIONS(4278),
    [anon_sym___weak] = ACTIONS(4278),
    [anon_sym_alignas] = ACTIONS(4278),
    [anon_sym__Alignas] = ACTIONS(4278),
    [sym_primitive_type] = ACTIONS(4278),
    [anon_sym_enum] = ACTIONS(4278),
    [anon_sym_COLON] = ACTIONS(4280),
    [anon_sym_struct] = ACTIONS(4278),
    [anon_sym_union] = ACTIONS(4278),
    [anon_sym_QMARK] = ACTIONS(4280),
    [anon_sym_STAR_EQ] = ACTIONS(4280),
    [anon_sym_SLASH_EQ] = ACTIONS(4280),
    [anon_sym_PERCENT_EQ] = ACTIONS(4280),
    [anon_sym_PLUS_EQ] = ACTIONS(4280),
    [anon_sym_DASH_EQ] = ACTIONS(4280),
    [anon_sym_LT_LT_EQ] = ACTIONS(4280),
    [anon_sym_GT_GT_EQ] = ACTIONS(4280),
    [anon_sym_AMP_EQ] = ACTIONS(4280),
    [anon_sym_CARET_EQ] = ACTIONS(4280),
    [anon_sym_PIPE_EQ] = ACTIONS(4280),
    [anon_sym_DASH_DASH] = ACTIONS(4280),
    [anon_sym_PLUS_PLUS] = ACTIONS(4280),
    [anon_sym_DOT] = ACTIONS(4278),
    [anon_sym_DASH_GT] = ACTIONS(4280),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4278),
    [anon_sym___typeof] = ACTIONS(4278),
    [anon_sym_typeof] = ACTIONS(4278),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4278),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4278),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4278),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4278),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4278),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4278),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4278),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4278),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4278),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4278),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4278),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4278),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4278),
    [anon_sym_API_AVAILABLE] = ACTIONS(4278),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4278),
    [anon_sym_API_DEPRECATED] = ACTIONS(4278),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4278),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4278),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4278),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4278),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4278),
    [anon_sym___deprecated_msg] = ACTIONS(4278),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4278),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4278),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4278),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4278),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4278),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4278),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4278),
    [anon_sym_BOOL] = ACTIONS(4278),
    [anon_sym_IMP] = ACTIONS(4278),
    [anon_sym_SEL] = ACTIONS(4278),
    [anon_sym_Class] = ACTIONS(4278),
    [anon_sym_id] = ACTIONS(4278),
  },
  [STATE(1512)] = {
    [sym_identifier] = ACTIONS(4282),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4284),
    [anon_sym_COMMA] = ACTIONS(4284),
    [anon_sym_RPAREN] = ACTIONS(4284),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4284),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4284),
    [anon_sym_LPAREN2] = ACTIONS(4284),
    [anon_sym_DASH] = ACTIONS(4282),
    [anon_sym_PLUS] = ACTIONS(4282),
    [anon_sym_STAR] = ACTIONS(4282),
    [anon_sym_SLASH] = ACTIONS(4282),
    [anon_sym_PERCENT] = ACTIONS(4282),
    [anon_sym_PIPE_PIPE] = ACTIONS(4284),
    [anon_sym_AMP_AMP] = ACTIONS(4284),
    [anon_sym_PIPE] = ACTIONS(4282),
    [anon_sym_CARET] = ACTIONS(4282),
    [anon_sym_AMP] = ACTIONS(4282),
    [anon_sym_EQ_EQ] = ACTIONS(4284),
    [anon_sym_BANG_EQ] = ACTIONS(4284),
    [anon_sym_GT] = ACTIONS(4282),
    [anon_sym_GT_EQ] = ACTIONS(4284),
    [anon_sym_LT_EQ] = ACTIONS(4284),
    [anon_sym_LT] = ACTIONS(4282),
    [anon_sym_LT_LT] = ACTIONS(4282),
    [anon_sym_GT_GT] = ACTIONS(4282),
    [anon_sym_SEMI] = ACTIONS(4284),
    [anon_sym___extension__] = ACTIONS(4282),
    [anon_sym_extern] = ACTIONS(4282),
    [anon_sym___attribute__] = ACTIONS(4282),
    [anon_sym___attribute] = ACTIONS(4282),
    [anon_sym_noreturn] = ACTIONS(4282),
    [anon_sym_LBRACK] = ACTIONS(4284),
    [anon_sym_RBRACK] = ACTIONS(4284),
    [anon_sym___declspec] = ACTIONS(4282),
    [anon_sym_LBRACE] = ACTIONS(4284),
    [anon_sym_RBRACE] = ACTIONS(4284),
    [anon_sym_signed] = ACTIONS(4282),
    [anon_sym_unsigned] = ACTIONS(4282),
    [anon_sym_long] = ACTIONS(4282),
    [anon_sym_short] = ACTIONS(4282),
    [anon_sym_static] = ACTIONS(4282),
    [anon_sym_EQ] = ACTIONS(4282),
    [anon_sym_ATautoreleasepool] = ACTIONS(4284),
    [anon_sym_auto] = ACTIONS(4282),
    [anon_sym_register] = ACTIONS(4282),
    [anon_sym_inline] = ACTIONS(4282),
    [anon_sym___inline] = ACTIONS(4282),
    [anon_sym___inline__] = ACTIONS(4282),
    [anon_sym___forceinline] = ACTIONS(4282),
    [anon_sym_thread_local] = ACTIONS(4282),
    [anon_sym___thread] = ACTIONS(4282),
    [anon_sym_CG_EXTERN] = ACTIONS(4282),
    [anon_sym_CG_INLINE] = ACTIONS(4282),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4282),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4282),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4282),
    [anon_sym_IBOutlet] = ACTIONS(4282),
    [anon_sym_IBInspectable] = ACTIONS(4282),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4282),
    [anon_sym_NS_INLINE] = ACTIONS(4282),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4282),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4282),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4282),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4282),
    [anon_sym_const] = ACTIONS(4282),
    [anon_sym_constexpr] = ACTIONS(4282),
    [anon_sym_volatile] = ACTIONS(4282),
    [anon_sym_restrict] = ACTIONS(4282),
    [anon_sym___restrict__] = ACTIONS(4282),
    [anon_sym__Atomic] = ACTIONS(4282),
    [anon_sym__Noreturn] = ACTIONS(4282),
    [anon_sym__Nonnull] = ACTIONS(4282),
    [anon_sym_nullable] = ACTIONS(4282),
    [anon_sym__Complex] = ACTIONS(4282),
    [anon_sym__Nullable] = ACTIONS(4282),
    [anon_sym__Nullable_result] = ACTIONS(4282),
    [anon_sym__Null_unspecified] = ACTIONS(4282),
    [anon_sym___autoreleasing] = ACTIONS(4282),
    [anon_sym___block] = ACTIONS(4282),
    [anon_sym___bridge] = ACTIONS(4282),
    [anon_sym___bridge_retained] = ACTIONS(4282),
    [anon_sym___bridge_transfer] = ACTIONS(4282),
    [anon_sym___complex] = ACTIONS(4282),
    [anon_sym___const] = ACTIONS(4282),
    [anon_sym___imag] = ACTIONS(4282),
    [anon_sym___kindof] = ACTIONS(4282),
    [anon_sym___nonnull] = ACTIONS(4282),
    [anon_sym___nullable] = ACTIONS(4282),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4282),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4282),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4282),
    [anon_sym___real] = ACTIONS(4282),
    [anon_sym___strong] = ACTIONS(4282),
    [anon_sym___unsafe_unretained] = ACTIONS(4282),
    [anon_sym___unused] = ACTIONS(4282),
    [anon_sym___weak] = ACTIONS(4282),
    [anon_sym_alignas] = ACTIONS(4282),
    [anon_sym__Alignas] = ACTIONS(4282),
    [sym_primitive_type] = ACTIONS(4282),
    [anon_sym_enum] = ACTIONS(4282),
    [anon_sym_COLON] = ACTIONS(4284),
    [anon_sym_struct] = ACTIONS(4282),
    [anon_sym_union] = ACTIONS(4282),
    [anon_sym_QMARK] = ACTIONS(4284),
    [anon_sym_STAR_EQ] = ACTIONS(4284),
    [anon_sym_SLASH_EQ] = ACTIONS(4284),
    [anon_sym_PERCENT_EQ] = ACTIONS(4284),
    [anon_sym_PLUS_EQ] = ACTIONS(4284),
    [anon_sym_DASH_EQ] = ACTIONS(4284),
    [anon_sym_LT_LT_EQ] = ACTIONS(4284),
    [anon_sym_GT_GT_EQ] = ACTIONS(4284),
    [anon_sym_AMP_EQ] = ACTIONS(4284),
    [anon_sym_CARET_EQ] = ACTIONS(4284),
    [anon_sym_PIPE_EQ] = ACTIONS(4284),
    [anon_sym_DASH_DASH] = ACTIONS(4284),
    [anon_sym_PLUS_PLUS] = ACTIONS(4284),
    [anon_sym_DOT] = ACTIONS(4282),
    [anon_sym_DASH_GT] = ACTIONS(4284),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4282),
    [anon_sym___typeof] = ACTIONS(4282),
    [anon_sym_typeof] = ACTIONS(4282),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4282),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4282),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4282),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4282),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4282),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4282),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4282),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4282),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4282),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4282),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4282),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4282),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4282),
    [anon_sym_API_AVAILABLE] = ACTIONS(4282),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4282),
    [anon_sym_API_DEPRECATED] = ACTIONS(4282),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4282),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4282),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4282),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4282),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4282),
    [anon_sym___deprecated_msg] = ACTIONS(4282),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4282),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4282),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4282),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4282),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4282),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4282),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4282),
    [anon_sym_BOOL] = ACTIONS(4282),
    [anon_sym_IMP] = ACTIONS(4282),
    [anon_sym_SEL] = ACTIONS(4282),
    [anon_sym_Class] = ACTIONS(4282),
    [anon_sym_id] = ACTIONS(4282),
  },
  [STATE(1513)] = {
    [sym_preproc_def] = STATE(1554),
    [sym_preproc_function_def] = STATE(1554),
    [sym_preproc_call] = STATE(1554),
    [sym_preproc_if_in_field_declaration_list] = STATE(1554),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1554),
    [sym_preproc_else_in_field_declaration_list] = STATE(5911),
    [sym_preproc_elif_in_field_declaration_list] = STATE(5911),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(5911),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1554),
    [sym_field_declaration] = STATE(1554),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1554),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1554),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4290),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1514)] = {
    [sym_preproc_def] = STATE(1524),
    [sym_preproc_function_def] = STATE(1524),
    [sym_preproc_call] = STATE(1524),
    [sym_preproc_if_in_field_declaration_list] = STATE(1524),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1524),
    [sym_preproc_else_in_field_declaration_list] = STATE(6504),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6504),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6504),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1524),
    [sym_field_declaration] = STATE(1524),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1524),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1524),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1515)] = {
    [sym_preproc_def] = STATE(1518),
    [sym_preproc_function_def] = STATE(1518),
    [sym_preproc_call] = STATE(1518),
    [sym_preproc_if_in_field_declaration_list] = STATE(1518),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1518),
    [sym_preproc_else_in_field_declaration_list] = STATE(6644),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6644),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6644),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1518),
    [sym_field_declaration] = STATE(1518),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1518),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1518),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4308),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1516)] = {
    [sym_preproc_def] = STATE(1523),
    [sym_preproc_function_def] = STATE(1523),
    [sym_preproc_call] = STATE(1523),
    [sym_preproc_if_in_field_declaration_list] = STATE(1523),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1523),
    [sym_preproc_else_in_field_declaration_list] = STATE(6241),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6241),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6241),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1523),
    [sym_field_declaration] = STATE(1523),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1523),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1523),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4310),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1517)] = {
    [sym_preproc_def] = STATE(1528),
    [sym_preproc_function_def] = STATE(1528),
    [sym_preproc_call] = STATE(1528),
    [sym_preproc_if_in_field_declaration_list] = STATE(1528),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1528),
    [sym_preproc_else_in_field_declaration_list] = STATE(6740),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6740),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6740),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1528),
    [sym_field_declaration] = STATE(1528),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1528),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1528),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4312),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1518)] = {
    [sym_preproc_def] = STATE(1554),
    [sym_preproc_function_def] = STATE(1554),
    [sym_preproc_call] = STATE(1554),
    [sym_preproc_if_in_field_declaration_list] = STATE(1554),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1554),
    [sym_preproc_else_in_field_declaration_list] = STATE(6742),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6742),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6742),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1554),
    [sym_field_declaration] = STATE(1554),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1554),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1554),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4314),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1519)] = {
    [sym_preproc_def] = STATE(1520),
    [sym_preproc_function_def] = STATE(1520),
    [sym_preproc_call] = STATE(1520),
    [sym_preproc_if_in_field_declaration_list] = STATE(1520),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1520),
    [sym_preproc_else_in_field_declaration_list] = STATE(6097),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6097),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6097),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1520),
    [sym_field_declaration] = STATE(1520),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1520),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1520),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4316),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1520)] = {
    [sym_preproc_def] = STATE(1554),
    [sym_preproc_function_def] = STATE(1554),
    [sym_preproc_call] = STATE(1554),
    [sym_preproc_if_in_field_declaration_list] = STATE(1554),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1554),
    [sym_preproc_else_in_field_declaration_list] = STATE(6341),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6341),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6341),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1554),
    [sym_field_declaration] = STATE(1554),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1554),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1554),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1521)] = {
    [sym_preproc_def] = STATE(1513),
    [sym_preproc_function_def] = STATE(1513),
    [sym_preproc_call] = STATE(1513),
    [sym_preproc_if_in_field_declaration_list] = STATE(1513),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1513),
    [sym_preproc_else_in_field_declaration_list] = STATE(6282),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6282),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6282),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1513),
    [sym_field_declaration] = STATE(1513),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1513),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1513),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4320),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1522)] = {
    [sym_preproc_def] = STATE(1525),
    [sym_preproc_function_def] = STATE(1525),
    [sym_preproc_call] = STATE(1525),
    [sym_preproc_if_in_field_declaration_list] = STATE(1525),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1525),
    [sym_preproc_else_in_field_declaration_list] = STATE(6428),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6428),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6428),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1525),
    [sym_field_declaration] = STATE(1525),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1525),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1525),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4322),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1523)] = {
    [sym_preproc_def] = STATE(1554),
    [sym_preproc_function_def] = STATE(1554),
    [sym_preproc_call] = STATE(1554),
    [sym_preproc_if_in_field_declaration_list] = STATE(1554),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1554),
    [sym_preproc_else_in_field_declaration_list] = STATE(6291),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6291),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6291),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1554),
    [sym_field_declaration] = STATE(1554),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1554),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1554),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4324),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1524)] = {
    [sym_preproc_def] = STATE(1554),
    [sym_preproc_function_def] = STATE(1554),
    [sym_preproc_call] = STATE(1554),
    [sym_preproc_if_in_field_declaration_list] = STATE(1554),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1554),
    [sym_preproc_else_in_field_declaration_list] = STATE(6482),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6482),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6482),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1554),
    [sym_field_declaration] = STATE(1554),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1554),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1554),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4326),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1525)] = {
    [sym_preproc_def] = STATE(1554),
    [sym_preproc_function_def] = STATE(1554),
    [sym_preproc_call] = STATE(1554),
    [sym_preproc_if_in_field_declaration_list] = STATE(1554),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1554),
    [sym_preproc_else_in_field_declaration_list] = STATE(6196),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6196),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6196),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1554),
    [sym_field_declaration] = STATE(1554),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1554),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1554),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4328),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1526)] = {
    [sym_preproc_def] = STATE(1554),
    [sym_preproc_function_def] = STATE(1554),
    [sym_preproc_call] = STATE(1554),
    [sym_preproc_if_in_field_declaration_list] = STATE(1554),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1554),
    [sym_preproc_else_in_field_declaration_list] = STATE(6105),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6105),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6105),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1554),
    [sym_field_declaration] = STATE(1554),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1554),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1554),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4330),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1527)] = {
    [sym_preproc_def] = STATE(1526),
    [sym_preproc_function_def] = STATE(1526),
    [sym_preproc_call] = STATE(1526),
    [sym_preproc_if_in_field_declaration_list] = STATE(1526),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1526),
    [sym_preproc_else_in_field_declaration_list] = STATE(5907),
    [sym_preproc_elif_in_field_declaration_list] = STATE(5907),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(5907),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1526),
    [sym_field_declaration] = STATE(1526),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1526),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1526),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4332),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1528)] = {
    [sym_preproc_def] = STATE(1554),
    [sym_preproc_function_def] = STATE(1554),
    [sym_preproc_call] = STATE(1554),
    [sym_preproc_if_in_field_declaration_list] = STATE(1554),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1554),
    [sym_preproc_else_in_field_declaration_list] = STATE(6229),
    [sym_preproc_elif_in_field_declaration_list] = STATE(6229),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(6229),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1554),
    [sym_field_declaration] = STATE(1554),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1554),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1554),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(4286),
    [aux_sym_preproc_if_token1] = ACTIONS(4288),
    [aux_sym_preproc_if_token2] = ACTIONS(4334),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4292),
    [aux_sym_preproc_else_token1] = ACTIONS(4294),
    [aux_sym_preproc_elif_token1] = ACTIONS(4296),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4298),
    [sym_preproc_directive] = ACTIONS(4300),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4304),
  },
  [STATE(1529)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6016),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4342),
    [sym_parameterized_arguments] = STATE(6402),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1530)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6672),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4331),
    [sym_parameterized_arguments] = STATE(6674),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1531)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6238),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4337),
    [sym_parameterized_arguments] = STATE(6333),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1532)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6072),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4338),
    [sym_parameterized_arguments] = STATE(6014),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1533)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6357),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4336),
    [sym_parameterized_arguments] = STATE(6191),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1534)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(5937),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4339),
    [sym_parameterized_arguments] = STATE(5903),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1535)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6183),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4336),
    [sym_parameterized_arguments] = STATE(6191),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1536)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(5964),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4335),
    [sym_parameterized_arguments] = STATE(6239),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1537)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6336),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4332),
    [sym_parameterized_arguments] = STATE(6725),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1538)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6593),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4340),
    [sym_parameterized_arguments] = STATE(6352),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1539)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(5921),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4338),
    [sym_parameterized_arguments] = STATE(6014),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1540)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6491),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4331),
    [sym_parameterized_arguments] = STATE(6674),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1541)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6781),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4339),
    [sym_parameterized_arguments] = STATE(5903),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1542)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6142),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4333),
    [sym_parameterized_arguments] = STATE(6361),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1543)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6400),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4342),
    [sym_parameterized_arguments] = STATE(6402),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1544)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6230),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4335),
    [sym_parameterized_arguments] = STATE(6239),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1545)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6350),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4340),
    [sym_parameterized_arguments] = STATE(6352),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1546)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6359),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4333),
    [sym_parameterized_arguments] = STATE(6361),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1547)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6597),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4330),
    [sym_parameterized_arguments] = STATE(6583),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1548)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6659),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4334),
    [sym_parameterized_arguments] = STATE(6653),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1549)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6651),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4334),
    [sym_parameterized_arguments] = STATE(6653),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1550)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6723),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4332),
    [sym_parameterized_arguments] = STATE(6725),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1551)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6581),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4330),
    [sym_parameterized_arguments] = STATE(6583),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1552)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6786),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(4337),
    [sym_parameterized_arguments] = STATE(6333),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1553)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4000),
    [anon_sym_PLUS] = ACTIONS(4000),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4000),
    [anon_sym_GT_GT] = ACTIONS(4000),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1554)] = {
    [sym_preproc_def] = STATE(1554),
    [sym_preproc_function_def] = STATE(1554),
    [sym_preproc_call] = STATE(1554),
    [sym_preproc_if_in_field_declaration_list] = STATE(1554),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1554),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4618),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1554),
    [sym_field_declaration] = STATE(1554),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1554),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1554),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4366),
    [aux_sym_preproc_def_token1] = ACTIONS(4369),
    [aux_sym_preproc_if_token1] = ACTIONS(4372),
    [aux_sym_preproc_if_token2] = ACTIONS(4375),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4377),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4377),
    [aux_sym_preproc_else_token1] = ACTIONS(4375),
    [aux_sym_preproc_elif_token1] = ACTIONS(4375),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(4375),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(4375),
    [sym_preproc_directive] = ACTIONS(4380),
    [anon_sym___extension__] = ACTIONS(4383),
    [anon_sym_extern] = ACTIONS(4386),
    [anon_sym___attribute__] = ACTIONS(4389),
    [anon_sym___attribute] = ACTIONS(4389),
    [anon_sym_noreturn] = ACTIONS(4383),
    [anon_sym_LBRACK] = ACTIONS(4392),
    [anon_sym___declspec] = ACTIONS(4395),
    [anon_sym_signed] = ACTIONS(4398),
    [anon_sym_unsigned] = ACTIONS(4398),
    [anon_sym_long] = ACTIONS(4398),
    [anon_sym_short] = ACTIONS(4398),
    [anon_sym_static] = ACTIONS(4386),
    [anon_sym_auto] = ACTIONS(4386),
    [anon_sym_register] = ACTIONS(4386),
    [anon_sym_inline] = ACTIONS(4386),
    [anon_sym___inline] = ACTIONS(4386),
    [anon_sym___inline__] = ACTIONS(4386),
    [anon_sym___forceinline] = ACTIONS(4386),
    [anon_sym_thread_local] = ACTIONS(4386),
    [anon_sym___thread] = ACTIONS(4386),
    [anon_sym_CG_EXTERN] = ACTIONS(4386),
    [anon_sym_CG_INLINE] = ACTIONS(4386),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4386),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4386),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4386),
    [anon_sym_IBOutlet] = ACTIONS(4386),
    [anon_sym_IBInspectable] = ACTIONS(4386),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4386),
    [anon_sym_NS_INLINE] = ACTIONS(4386),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4386),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4386),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4386),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4386),
    [anon_sym_const] = ACTIONS(4383),
    [anon_sym_constexpr] = ACTIONS(4383),
    [anon_sym_volatile] = ACTIONS(4383),
    [anon_sym_restrict] = ACTIONS(4383),
    [anon_sym___restrict__] = ACTIONS(4383),
    [anon_sym__Atomic] = ACTIONS(4383),
    [anon_sym__Noreturn] = ACTIONS(4383),
    [anon_sym__Nonnull] = ACTIONS(4383),
    [anon_sym_nullable] = ACTIONS(4383),
    [anon_sym__Complex] = ACTIONS(4383),
    [anon_sym__Nullable] = ACTIONS(4383),
    [anon_sym__Nullable_result] = ACTIONS(4383),
    [anon_sym__Null_unspecified] = ACTIONS(4383),
    [anon_sym___autoreleasing] = ACTIONS(4383),
    [anon_sym___block] = ACTIONS(4383),
    [anon_sym___bridge] = ACTIONS(4383),
    [anon_sym___bridge_retained] = ACTIONS(4383),
    [anon_sym___bridge_transfer] = ACTIONS(4383),
    [anon_sym___complex] = ACTIONS(4383),
    [anon_sym___const] = ACTIONS(4383),
    [anon_sym___imag] = ACTIONS(4383),
    [anon_sym___kindof] = ACTIONS(4383),
    [anon_sym___nonnull] = ACTIONS(4383),
    [anon_sym___nullable] = ACTIONS(4383),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4383),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4383),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4383),
    [anon_sym___real] = ACTIONS(4383),
    [anon_sym___strong] = ACTIONS(4383),
    [anon_sym___unsafe_unretained] = ACTIONS(4383),
    [anon_sym___unused] = ACTIONS(4383),
    [anon_sym___weak] = ACTIONS(4383),
    [anon_sym_alignas] = ACTIONS(4401),
    [anon_sym__Alignas] = ACTIONS(4401),
    [sym_primitive_type] = ACTIONS(4404),
    [anon_sym_enum] = ACTIONS(4407),
    [anon_sym_struct] = ACTIONS(4410),
    [anon_sym_union] = ACTIONS(4413),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4416),
    [anon_sym___typeof] = ACTIONS(4416),
    [anon_sym_typeof] = ACTIONS(4416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4419),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4419),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4419),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4419),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4419),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4419),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4419),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4419),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4419),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4422),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4422),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4422),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_API_AVAILABLE] = ACTIONS(4422),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4422),
    [anon_sym_API_DEPRECATED] = ACTIONS(4422),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4422),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4422),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4422),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4422),
    [anon_sym___deprecated_msg] = ACTIONS(4422),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4422),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4422),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4422),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4422),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4422),
    [anon_sym_BOOL] = ACTIONS(4425),
    [anon_sym_IMP] = ACTIONS(4425),
    [anon_sym_SEL] = ACTIONS(4425),
    [anon_sym_Class] = ACTIONS(4425),
    [anon_sym_id] = ACTIONS(4425),
    [anon_sym_ATdefs] = ACTIONS(4428),
  },
  [STATE(1555)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4433),
    [anon_sym_COMMA] = ACTIONS(4436),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4440),
    [anon_sym_AMP_AMP] = ACTIONS(4442),
    [anon_sym_PIPE] = ACTIONS(4444),
    [anon_sym_CARET] = ACTIONS(4446),
    [anon_sym_AMP] = ACTIONS(4448),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4436),
    [anon_sym___extension__] = ACTIONS(4431),
    [anon_sym_extern] = ACTIONS(4431),
    [anon_sym___attribute__] = ACTIONS(4431),
    [anon_sym___attribute] = ACTIONS(4431),
    [anon_sym_noreturn] = ACTIONS(4431),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4431),
    [anon_sym_LBRACE] = ACTIONS(4436),
    [anon_sym_signed] = ACTIONS(4431),
    [anon_sym_unsigned] = ACTIONS(4431),
    [anon_sym_long] = ACTIONS(4431),
    [anon_sym_short] = ACTIONS(4431),
    [anon_sym_static] = ACTIONS(4431),
    [anon_sym_EQ] = ACTIONS(4431),
    [anon_sym_ATautoreleasepool] = ACTIONS(4436),
    [anon_sym_auto] = ACTIONS(4431),
    [anon_sym_register] = ACTIONS(4431),
    [anon_sym_inline] = ACTIONS(4431),
    [anon_sym___inline] = ACTIONS(4431),
    [anon_sym___inline__] = ACTIONS(4431),
    [anon_sym___forceinline] = ACTIONS(4431),
    [anon_sym_thread_local] = ACTIONS(4431),
    [anon_sym___thread] = ACTIONS(4431),
    [anon_sym_CG_EXTERN] = ACTIONS(4431),
    [anon_sym_CG_INLINE] = ACTIONS(4431),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4431),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4431),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4431),
    [anon_sym_IBOutlet] = ACTIONS(4431),
    [anon_sym_IBInspectable] = ACTIONS(4431),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4431),
    [anon_sym_NS_INLINE] = ACTIONS(4431),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4431),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4431),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4431),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4431),
    [anon_sym_const] = ACTIONS(4431),
    [anon_sym_constexpr] = ACTIONS(4431),
    [anon_sym_volatile] = ACTIONS(4431),
    [anon_sym_restrict] = ACTIONS(4431),
    [anon_sym___restrict__] = ACTIONS(4431),
    [anon_sym__Atomic] = ACTIONS(4431),
    [anon_sym__Noreturn] = ACTIONS(4431),
    [anon_sym__Nonnull] = ACTIONS(4431),
    [anon_sym_nullable] = ACTIONS(4431),
    [anon_sym__Complex] = ACTIONS(4431),
    [anon_sym__Nullable] = ACTIONS(4431),
    [anon_sym__Nullable_result] = ACTIONS(4431),
    [anon_sym__Null_unspecified] = ACTIONS(4431),
    [anon_sym___autoreleasing] = ACTIONS(4431),
    [anon_sym___block] = ACTIONS(4431),
    [anon_sym___bridge] = ACTIONS(4431),
    [anon_sym___bridge_retained] = ACTIONS(4431),
    [anon_sym___bridge_transfer] = ACTIONS(4431),
    [anon_sym___complex] = ACTIONS(4431),
    [anon_sym___const] = ACTIONS(4431),
    [anon_sym___imag] = ACTIONS(4431),
    [anon_sym___kindof] = ACTIONS(4431),
    [anon_sym___nonnull] = ACTIONS(4431),
    [anon_sym___nullable] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4431),
    [anon_sym___real] = ACTIONS(4431),
    [anon_sym___strong] = ACTIONS(4431),
    [anon_sym___unsafe_unretained] = ACTIONS(4431),
    [anon_sym___unused] = ACTIONS(4431),
    [anon_sym___weak] = ACTIONS(4431),
    [anon_sym_alignas] = ACTIONS(4431),
    [anon_sym__Alignas] = ACTIONS(4431),
    [sym_primitive_type] = ACTIONS(4431),
    [anon_sym_enum] = ACTIONS(4431),
    [anon_sym_struct] = ACTIONS(4431),
    [anon_sym_union] = ACTIONS(4431),
    [anon_sym_QMARK] = ACTIONS(4436),
    [anon_sym_STAR_EQ] = ACTIONS(4436),
    [anon_sym_SLASH_EQ] = ACTIONS(4436),
    [anon_sym_PERCENT_EQ] = ACTIONS(4436),
    [anon_sym_PLUS_EQ] = ACTIONS(4436),
    [anon_sym_DASH_EQ] = ACTIONS(4436),
    [anon_sym_LT_LT_EQ] = ACTIONS(4436),
    [anon_sym_GT_GT_EQ] = ACTIONS(4436),
    [anon_sym_AMP_EQ] = ACTIONS(4436),
    [anon_sym_CARET_EQ] = ACTIONS(4436),
    [anon_sym_PIPE_EQ] = ACTIONS(4436),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4431),
    [anon_sym___typeof] = ACTIONS(4431),
    [anon_sym_typeof] = ACTIONS(4431),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4431),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4431),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4431),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4431),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4431),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4431),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4431),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_API_AVAILABLE] = ACTIONS(4431),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_API_DEPRECATED] = ACTIONS(4431),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4431),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym___deprecated_msg] = ACTIONS(4431),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4431),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4431),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4431),
    [anon_sym_BOOL] = ACTIONS(4431),
    [anon_sym_IMP] = ACTIONS(4431),
    [anon_sym_SEL] = ACTIONS(4431),
    [anon_sym_Class] = ACTIONS(4431),
    [anon_sym_id] = ACTIONS(4431),
  },
  [STATE(1556)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4458),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4460),
    [anon_sym_COMMA] = ACTIONS(4462),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4440),
    [anon_sym_AMP_AMP] = ACTIONS(4442),
    [anon_sym_PIPE] = ACTIONS(4444),
    [anon_sym_CARET] = ACTIONS(4446),
    [anon_sym_AMP] = ACTIONS(4448),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4462),
    [anon_sym___extension__] = ACTIONS(4458),
    [anon_sym_extern] = ACTIONS(4458),
    [anon_sym___attribute__] = ACTIONS(4458),
    [anon_sym___attribute] = ACTIONS(4458),
    [anon_sym_noreturn] = ACTIONS(4458),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4458),
    [anon_sym_LBRACE] = ACTIONS(4462),
    [anon_sym_signed] = ACTIONS(4458),
    [anon_sym_unsigned] = ACTIONS(4458),
    [anon_sym_long] = ACTIONS(4458),
    [anon_sym_short] = ACTIONS(4458),
    [anon_sym_static] = ACTIONS(4458),
    [anon_sym_EQ] = ACTIONS(4458),
    [anon_sym_ATautoreleasepool] = ACTIONS(4462),
    [anon_sym_auto] = ACTIONS(4458),
    [anon_sym_register] = ACTIONS(4458),
    [anon_sym_inline] = ACTIONS(4458),
    [anon_sym___inline] = ACTIONS(4458),
    [anon_sym___inline__] = ACTIONS(4458),
    [anon_sym___forceinline] = ACTIONS(4458),
    [anon_sym_thread_local] = ACTIONS(4458),
    [anon_sym___thread] = ACTIONS(4458),
    [anon_sym_CG_EXTERN] = ACTIONS(4458),
    [anon_sym_CG_INLINE] = ACTIONS(4458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4458),
    [anon_sym_IBOutlet] = ACTIONS(4458),
    [anon_sym_IBInspectable] = ACTIONS(4458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4458),
    [anon_sym_NS_INLINE] = ACTIONS(4458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4458),
    [anon_sym_const] = ACTIONS(4458),
    [anon_sym_constexpr] = ACTIONS(4458),
    [anon_sym_volatile] = ACTIONS(4458),
    [anon_sym_restrict] = ACTIONS(4458),
    [anon_sym___restrict__] = ACTIONS(4458),
    [anon_sym__Atomic] = ACTIONS(4458),
    [anon_sym__Noreturn] = ACTIONS(4458),
    [anon_sym__Nonnull] = ACTIONS(4458),
    [anon_sym_nullable] = ACTIONS(4458),
    [anon_sym__Complex] = ACTIONS(4458),
    [anon_sym__Nullable] = ACTIONS(4458),
    [anon_sym__Nullable_result] = ACTIONS(4458),
    [anon_sym__Null_unspecified] = ACTIONS(4458),
    [anon_sym___autoreleasing] = ACTIONS(4458),
    [anon_sym___block] = ACTIONS(4458),
    [anon_sym___bridge] = ACTIONS(4458),
    [anon_sym___bridge_retained] = ACTIONS(4458),
    [anon_sym___bridge_transfer] = ACTIONS(4458),
    [anon_sym___complex] = ACTIONS(4458),
    [anon_sym___const] = ACTIONS(4458),
    [anon_sym___imag] = ACTIONS(4458),
    [anon_sym___kindof] = ACTIONS(4458),
    [anon_sym___nonnull] = ACTIONS(4458),
    [anon_sym___nullable] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4458),
    [anon_sym___real] = ACTIONS(4458),
    [anon_sym___strong] = ACTIONS(4458),
    [anon_sym___unsafe_unretained] = ACTIONS(4458),
    [anon_sym___unused] = ACTIONS(4458),
    [anon_sym___weak] = ACTIONS(4458),
    [anon_sym_alignas] = ACTIONS(4458),
    [anon_sym__Alignas] = ACTIONS(4458),
    [sym_primitive_type] = ACTIONS(4458),
    [anon_sym_enum] = ACTIONS(4458),
    [anon_sym_struct] = ACTIONS(4458),
    [anon_sym_union] = ACTIONS(4458),
    [anon_sym_QMARK] = ACTIONS(4464),
    [anon_sym_STAR_EQ] = ACTIONS(4462),
    [anon_sym_SLASH_EQ] = ACTIONS(4462),
    [anon_sym_PERCENT_EQ] = ACTIONS(4462),
    [anon_sym_PLUS_EQ] = ACTIONS(4462),
    [anon_sym_DASH_EQ] = ACTIONS(4462),
    [anon_sym_LT_LT_EQ] = ACTIONS(4462),
    [anon_sym_GT_GT_EQ] = ACTIONS(4462),
    [anon_sym_AMP_EQ] = ACTIONS(4462),
    [anon_sym_CARET_EQ] = ACTIONS(4462),
    [anon_sym_PIPE_EQ] = ACTIONS(4462),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4458),
    [anon_sym___typeof] = ACTIONS(4458),
    [anon_sym_typeof] = ACTIONS(4458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_API_AVAILABLE] = ACTIONS(4458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_API_DEPRECATED] = ACTIONS(4458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym___deprecated_msg] = ACTIONS(4458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4458),
    [anon_sym_BOOL] = ACTIONS(4458),
    [anon_sym_IMP] = ACTIONS(4458),
    [anon_sym_SEL] = ACTIONS(4458),
    [anon_sym_Class] = ACTIONS(4458),
    [anon_sym_id] = ACTIONS(4458),
  },
  [STATE(1557)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4466),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4460),
    [anon_sym_COMMA] = ACTIONS(4468),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4440),
    [anon_sym_AMP_AMP] = ACTIONS(4442),
    [anon_sym_PIPE] = ACTIONS(4444),
    [anon_sym_CARET] = ACTIONS(4446),
    [anon_sym_AMP] = ACTIONS(4448),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4468),
    [anon_sym___extension__] = ACTIONS(4466),
    [anon_sym_extern] = ACTIONS(4466),
    [anon_sym___attribute__] = ACTIONS(4466),
    [anon_sym___attribute] = ACTIONS(4466),
    [anon_sym_noreturn] = ACTIONS(4466),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4466),
    [anon_sym_LBRACE] = ACTIONS(4468),
    [anon_sym_signed] = ACTIONS(4466),
    [anon_sym_unsigned] = ACTIONS(4466),
    [anon_sym_long] = ACTIONS(4466),
    [anon_sym_short] = ACTIONS(4466),
    [anon_sym_static] = ACTIONS(4466),
    [anon_sym_EQ] = ACTIONS(4466),
    [anon_sym_ATautoreleasepool] = ACTIONS(4468),
    [anon_sym_auto] = ACTIONS(4466),
    [anon_sym_register] = ACTIONS(4466),
    [anon_sym_inline] = ACTIONS(4466),
    [anon_sym___inline] = ACTIONS(4466),
    [anon_sym___inline__] = ACTIONS(4466),
    [anon_sym___forceinline] = ACTIONS(4466),
    [anon_sym_thread_local] = ACTIONS(4466),
    [anon_sym___thread] = ACTIONS(4466),
    [anon_sym_CG_EXTERN] = ACTIONS(4466),
    [anon_sym_CG_INLINE] = ACTIONS(4466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4466),
    [anon_sym_IBOutlet] = ACTIONS(4466),
    [anon_sym_IBInspectable] = ACTIONS(4466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4466),
    [anon_sym_NS_INLINE] = ACTIONS(4466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4466),
    [anon_sym_const] = ACTIONS(4466),
    [anon_sym_constexpr] = ACTIONS(4466),
    [anon_sym_volatile] = ACTIONS(4466),
    [anon_sym_restrict] = ACTIONS(4466),
    [anon_sym___restrict__] = ACTIONS(4466),
    [anon_sym__Atomic] = ACTIONS(4466),
    [anon_sym__Noreturn] = ACTIONS(4466),
    [anon_sym__Nonnull] = ACTIONS(4466),
    [anon_sym_nullable] = ACTIONS(4466),
    [anon_sym__Complex] = ACTIONS(4466),
    [anon_sym__Nullable] = ACTIONS(4466),
    [anon_sym__Nullable_result] = ACTIONS(4466),
    [anon_sym__Null_unspecified] = ACTIONS(4466),
    [anon_sym___autoreleasing] = ACTIONS(4466),
    [anon_sym___block] = ACTIONS(4466),
    [anon_sym___bridge] = ACTIONS(4466),
    [anon_sym___bridge_retained] = ACTIONS(4466),
    [anon_sym___bridge_transfer] = ACTIONS(4466),
    [anon_sym___complex] = ACTIONS(4466),
    [anon_sym___const] = ACTIONS(4466),
    [anon_sym___imag] = ACTIONS(4466),
    [anon_sym___kindof] = ACTIONS(4466),
    [anon_sym___nonnull] = ACTIONS(4466),
    [anon_sym___nullable] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4466),
    [anon_sym___real] = ACTIONS(4466),
    [anon_sym___strong] = ACTIONS(4466),
    [anon_sym___unsafe_unretained] = ACTIONS(4466),
    [anon_sym___unused] = ACTIONS(4466),
    [anon_sym___weak] = ACTIONS(4466),
    [anon_sym_alignas] = ACTIONS(4466),
    [anon_sym__Alignas] = ACTIONS(4466),
    [sym_primitive_type] = ACTIONS(4466),
    [anon_sym_enum] = ACTIONS(4466),
    [anon_sym_struct] = ACTIONS(4466),
    [anon_sym_union] = ACTIONS(4466),
    [anon_sym_QMARK] = ACTIONS(4464),
    [anon_sym_STAR_EQ] = ACTIONS(4468),
    [anon_sym_SLASH_EQ] = ACTIONS(4468),
    [anon_sym_PERCENT_EQ] = ACTIONS(4468),
    [anon_sym_PLUS_EQ] = ACTIONS(4468),
    [anon_sym_DASH_EQ] = ACTIONS(4468),
    [anon_sym_LT_LT_EQ] = ACTIONS(4468),
    [anon_sym_GT_GT_EQ] = ACTIONS(4468),
    [anon_sym_AMP_EQ] = ACTIONS(4468),
    [anon_sym_CARET_EQ] = ACTIONS(4468),
    [anon_sym_PIPE_EQ] = ACTIONS(4468),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4466),
    [anon_sym___typeof] = ACTIONS(4466),
    [anon_sym_typeof] = ACTIONS(4466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_API_AVAILABLE] = ACTIONS(4466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_API_DEPRECATED] = ACTIONS(4466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym___deprecated_msg] = ACTIONS(4466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4466),
    [anon_sym_BOOL] = ACTIONS(4466),
    [anon_sym_IMP] = ACTIONS(4466),
    [anon_sym_SEL] = ACTIONS(4466),
    [anon_sym_Class] = ACTIONS(4466),
    [anon_sym_id] = ACTIONS(4466),
  },
  [STATE(1558)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4470),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4460),
    [anon_sym_COMMA] = ACTIONS(4472),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4440),
    [anon_sym_AMP_AMP] = ACTIONS(4442),
    [anon_sym_PIPE] = ACTIONS(4444),
    [anon_sym_CARET] = ACTIONS(4446),
    [anon_sym_AMP] = ACTIONS(4448),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4472),
    [anon_sym___extension__] = ACTIONS(4470),
    [anon_sym_extern] = ACTIONS(4470),
    [anon_sym___attribute__] = ACTIONS(4470),
    [anon_sym___attribute] = ACTIONS(4470),
    [anon_sym_noreturn] = ACTIONS(4470),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4470),
    [anon_sym_LBRACE] = ACTIONS(4472),
    [anon_sym_signed] = ACTIONS(4470),
    [anon_sym_unsigned] = ACTIONS(4470),
    [anon_sym_long] = ACTIONS(4470),
    [anon_sym_short] = ACTIONS(4470),
    [anon_sym_static] = ACTIONS(4470),
    [anon_sym_EQ] = ACTIONS(4470),
    [anon_sym_ATautoreleasepool] = ACTIONS(4472),
    [anon_sym_auto] = ACTIONS(4470),
    [anon_sym_register] = ACTIONS(4470),
    [anon_sym_inline] = ACTIONS(4470),
    [anon_sym___inline] = ACTIONS(4470),
    [anon_sym___inline__] = ACTIONS(4470),
    [anon_sym___forceinline] = ACTIONS(4470),
    [anon_sym_thread_local] = ACTIONS(4470),
    [anon_sym___thread] = ACTIONS(4470),
    [anon_sym_CG_EXTERN] = ACTIONS(4470),
    [anon_sym_CG_INLINE] = ACTIONS(4470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4470),
    [anon_sym_IBOutlet] = ACTIONS(4470),
    [anon_sym_IBInspectable] = ACTIONS(4470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4470),
    [anon_sym_NS_INLINE] = ACTIONS(4470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4470),
    [anon_sym_const] = ACTIONS(4470),
    [anon_sym_constexpr] = ACTIONS(4470),
    [anon_sym_volatile] = ACTIONS(4470),
    [anon_sym_restrict] = ACTIONS(4470),
    [anon_sym___restrict__] = ACTIONS(4470),
    [anon_sym__Atomic] = ACTIONS(4470),
    [anon_sym__Noreturn] = ACTIONS(4470),
    [anon_sym__Nonnull] = ACTIONS(4470),
    [anon_sym_nullable] = ACTIONS(4470),
    [anon_sym__Complex] = ACTIONS(4470),
    [anon_sym__Nullable] = ACTIONS(4470),
    [anon_sym__Nullable_result] = ACTIONS(4470),
    [anon_sym__Null_unspecified] = ACTIONS(4470),
    [anon_sym___autoreleasing] = ACTIONS(4470),
    [anon_sym___block] = ACTIONS(4470),
    [anon_sym___bridge] = ACTIONS(4470),
    [anon_sym___bridge_retained] = ACTIONS(4470),
    [anon_sym___bridge_transfer] = ACTIONS(4470),
    [anon_sym___complex] = ACTIONS(4470),
    [anon_sym___const] = ACTIONS(4470),
    [anon_sym___imag] = ACTIONS(4470),
    [anon_sym___kindof] = ACTIONS(4470),
    [anon_sym___nonnull] = ACTIONS(4470),
    [anon_sym___nullable] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4470),
    [anon_sym___real] = ACTIONS(4470),
    [anon_sym___strong] = ACTIONS(4470),
    [anon_sym___unsafe_unretained] = ACTIONS(4470),
    [anon_sym___unused] = ACTIONS(4470),
    [anon_sym___weak] = ACTIONS(4470),
    [anon_sym_alignas] = ACTIONS(4470),
    [anon_sym__Alignas] = ACTIONS(4470),
    [sym_primitive_type] = ACTIONS(4470),
    [anon_sym_enum] = ACTIONS(4470),
    [anon_sym_struct] = ACTIONS(4470),
    [anon_sym_union] = ACTIONS(4470),
    [anon_sym_QMARK] = ACTIONS(4472),
    [anon_sym_STAR_EQ] = ACTIONS(4472),
    [anon_sym_SLASH_EQ] = ACTIONS(4472),
    [anon_sym_PERCENT_EQ] = ACTIONS(4472),
    [anon_sym_PLUS_EQ] = ACTIONS(4472),
    [anon_sym_DASH_EQ] = ACTIONS(4472),
    [anon_sym_LT_LT_EQ] = ACTIONS(4472),
    [anon_sym_GT_GT_EQ] = ACTIONS(4472),
    [anon_sym_AMP_EQ] = ACTIONS(4472),
    [anon_sym_CARET_EQ] = ACTIONS(4472),
    [anon_sym_PIPE_EQ] = ACTIONS(4472),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4470),
    [anon_sym___typeof] = ACTIONS(4470),
    [anon_sym_typeof] = ACTIONS(4470),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_API_AVAILABLE] = ACTIONS(4470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_API_DEPRECATED] = ACTIONS(4470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym___deprecated_msg] = ACTIONS(4470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4470),
    [anon_sym_BOOL] = ACTIONS(4470),
    [anon_sym_IMP] = ACTIONS(4470),
    [anon_sym_SEL] = ACTIONS(4470),
    [anon_sym_Class] = ACTIONS(4470),
    [anon_sym_id] = ACTIONS(4470),
  },
  [STATE(1559)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1560)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4442),
    [anon_sym_PIPE] = ACTIONS(4444),
    [anon_sym_CARET] = ACTIONS(4446),
    [anon_sym_AMP] = ACTIONS(4448),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1561)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4474),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4460),
    [anon_sym_COMMA] = ACTIONS(4476),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4440),
    [anon_sym_AMP_AMP] = ACTIONS(4442),
    [anon_sym_PIPE] = ACTIONS(4444),
    [anon_sym_CARET] = ACTIONS(4446),
    [anon_sym_AMP] = ACTIONS(4448),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4476),
    [anon_sym___extension__] = ACTIONS(4474),
    [anon_sym_extern] = ACTIONS(4474),
    [anon_sym___attribute__] = ACTIONS(4474),
    [anon_sym___attribute] = ACTIONS(4474),
    [anon_sym_noreturn] = ACTIONS(4474),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4474),
    [anon_sym_LBRACE] = ACTIONS(4476),
    [anon_sym_signed] = ACTIONS(4474),
    [anon_sym_unsigned] = ACTIONS(4474),
    [anon_sym_long] = ACTIONS(4474),
    [anon_sym_short] = ACTIONS(4474),
    [anon_sym_static] = ACTIONS(4474),
    [anon_sym_EQ] = ACTIONS(4474),
    [anon_sym_ATautoreleasepool] = ACTIONS(4476),
    [anon_sym_auto] = ACTIONS(4474),
    [anon_sym_register] = ACTIONS(4474),
    [anon_sym_inline] = ACTIONS(4474),
    [anon_sym___inline] = ACTIONS(4474),
    [anon_sym___inline__] = ACTIONS(4474),
    [anon_sym___forceinline] = ACTIONS(4474),
    [anon_sym_thread_local] = ACTIONS(4474),
    [anon_sym___thread] = ACTIONS(4474),
    [anon_sym_CG_EXTERN] = ACTIONS(4474),
    [anon_sym_CG_INLINE] = ACTIONS(4474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4474),
    [anon_sym_IBOutlet] = ACTIONS(4474),
    [anon_sym_IBInspectable] = ACTIONS(4474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4474),
    [anon_sym_NS_INLINE] = ACTIONS(4474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4474),
    [anon_sym_const] = ACTIONS(4474),
    [anon_sym_constexpr] = ACTIONS(4474),
    [anon_sym_volatile] = ACTIONS(4474),
    [anon_sym_restrict] = ACTIONS(4474),
    [anon_sym___restrict__] = ACTIONS(4474),
    [anon_sym__Atomic] = ACTIONS(4474),
    [anon_sym__Noreturn] = ACTIONS(4474),
    [anon_sym__Nonnull] = ACTIONS(4474),
    [anon_sym_nullable] = ACTIONS(4474),
    [anon_sym__Complex] = ACTIONS(4474),
    [anon_sym__Nullable] = ACTIONS(4474),
    [anon_sym__Nullable_result] = ACTIONS(4474),
    [anon_sym__Null_unspecified] = ACTIONS(4474),
    [anon_sym___autoreleasing] = ACTIONS(4474),
    [anon_sym___block] = ACTIONS(4474),
    [anon_sym___bridge] = ACTIONS(4474),
    [anon_sym___bridge_retained] = ACTIONS(4474),
    [anon_sym___bridge_transfer] = ACTIONS(4474),
    [anon_sym___complex] = ACTIONS(4474),
    [anon_sym___const] = ACTIONS(4474),
    [anon_sym___imag] = ACTIONS(4474),
    [anon_sym___kindof] = ACTIONS(4474),
    [anon_sym___nonnull] = ACTIONS(4474),
    [anon_sym___nullable] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4474),
    [anon_sym___real] = ACTIONS(4474),
    [anon_sym___strong] = ACTIONS(4474),
    [anon_sym___unsafe_unretained] = ACTIONS(4474),
    [anon_sym___unused] = ACTIONS(4474),
    [anon_sym___weak] = ACTIONS(4474),
    [anon_sym_alignas] = ACTIONS(4474),
    [anon_sym__Alignas] = ACTIONS(4474),
    [sym_primitive_type] = ACTIONS(4474),
    [anon_sym_enum] = ACTIONS(4474),
    [anon_sym_struct] = ACTIONS(4474),
    [anon_sym_union] = ACTIONS(4474),
    [anon_sym_QMARK] = ACTIONS(4476),
    [anon_sym_STAR_EQ] = ACTIONS(4476),
    [anon_sym_SLASH_EQ] = ACTIONS(4476),
    [anon_sym_PERCENT_EQ] = ACTIONS(4476),
    [anon_sym_PLUS_EQ] = ACTIONS(4476),
    [anon_sym_DASH_EQ] = ACTIONS(4476),
    [anon_sym_LT_LT_EQ] = ACTIONS(4476),
    [anon_sym_GT_GT_EQ] = ACTIONS(4476),
    [anon_sym_AMP_EQ] = ACTIONS(4476),
    [anon_sym_CARET_EQ] = ACTIONS(4476),
    [anon_sym_PIPE_EQ] = ACTIONS(4476),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4474),
    [anon_sym___typeof] = ACTIONS(4474),
    [anon_sym_typeof] = ACTIONS(4474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_API_AVAILABLE] = ACTIONS(4474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_API_DEPRECATED] = ACTIONS(4474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym___deprecated_msg] = ACTIONS(4474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4474),
    [anon_sym_BOOL] = ACTIONS(4474),
    [anon_sym_IMP] = ACTIONS(4474),
    [anon_sym_SEL] = ACTIONS(4474),
    [anon_sym_Class] = ACTIONS(4474),
    [anon_sym_id] = ACTIONS(4474),
  },
  [STATE(1562)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4000),
    [anon_sym_GT_GT] = ACTIONS(4000),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1563)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4446),
    [anon_sym_AMP] = ACTIONS(4448),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1564)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4448),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1565)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1566)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4478),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4460),
    [anon_sym_COMMA] = ACTIONS(4480),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4440),
    [anon_sym_AMP_AMP] = ACTIONS(4442),
    [anon_sym_PIPE] = ACTIONS(4444),
    [anon_sym_CARET] = ACTIONS(4446),
    [anon_sym_AMP] = ACTIONS(4448),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4480),
    [anon_sym___extension__] = ACTIONS(4478),
    [anon_sym_extern] = ACTIONS(4478),
    [anon_sym___attribute__] = ACTIONS(4478),
    [anon_sym___attribute] = ACTIONS(4478),
    [anon_sym_noreturn] = ACTIONS(4478),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4478),
    [anon_sym_LBRACE] = ACTIONS(4480),
    [anon_sym_signed] = ACTIONS(4478),
    [anon_sym_unsigned] = ACTIONS(4478),
    [anon_sym_long] = ACTIONS(4478),
    [anon_sym_short] = ACTIONS(4478),
    [anon_sym_static] = ACTIONS(4478),
    [anon_sym_EQ] = ACTIONS(4478),
    [anon_sym_ATautoreleasepool] = ACTIONS(4480),
    [anon_sym_auto] = ACTIONS(4478),
    [anon_sym_register] = ACTIONS(4478),
    [anon_sym_inline] = ACTIONS(4478),
    [anon_sym___inline] = ACTIONS(4478),
    [anon_sym___inline__] = ACTIONS(4478),
    [anon_sym___forceinline] = ACTIONS(4478),
    [anon_sym_thread_local] = ACTIONS(4478),
    [anon_sym___thread] = ACTIONS(4478),
    [anon_sym_CG_EXTERN] = ACTIONS(4478),
    [anon_sym_CG_INLINE] = ACTIONS(4478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4478),
    [anon_sym_IBOutlet] = ACTIONS(4478),
    [anon_sym_IBInspectable] = ACTIONS(4478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4478),
    [anon_sym_NS_INLINE] = ACTIONS(4478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4478),
    [anon_sym_const] = ACTIONS(4478),
    [anon_sym_constexpr] = ACTIONS(4478),
    [anon_sym_volatile] = ACTIONS(4478),
    [anon_sym_restrict] = ACTIONS(4478),
    [anon_sym___restrict__] = ACTIONS(4478),
    [anon_sym__Atomic] = ACTIONS(4478),
    [anon_sym__Noreturn] = ACTIONS(4478),
    [anon_sym__Nonnull] = ACTIONS(4478),
    [anon_sym_nullable] = ACTIONS(4478),
    [anon_sym__Complex] = ACTIONS(4478),
    [anon_sym__Nullable] = ACTIONS(4478),
    [anon_sym__Nullable_result] = ACTIONS(4478),
    [anon_sym__Null_unspecified] = ACTIONS(4478),
    [anon_sym___autoreleasing] = ACTIONS(4478),
    [anon_sym___block] = ACTIONS(4478),
    [anon_sym___bridge] = ACTIONS(4478),
    [anon_sym___bridge_retained] = ACTIONS(4478),
    [anon_sym___bridge_transfer] = ACTIONS(4478),
    [anon_sym___complex] = ACTIONS(4478),
    [anon_sym___const] = ACTIONS(4478),
    [anon_sym___imag] = ACTIONS(4478),
    [anon_sym___kindof] = ACTIONS(4478),
    [anon_sym___nonnull] = ACTIONS(4478),
    [anon_sym___nullable] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4478),
    [anon_sym___real] = ACTIONS(4478),
    [anon_sym___strong] = ACTIONS(4478),
    [anon_sym___unsafe_unretained] = ACTIONS(4478),
    [anon_sym___unused] = ACTIONS(4478),
    [anon_sym___weak] = ACTIONS(4478),
    [anon_sym_alignas] = ACTIONS(4478),
    [anon_sym__Alignas] = ACTIONS(4478),
    [sym_primitive_type] = ACTIONS(4478),
    [anon_sym_enum] = ACTIONS(4478),
    [anon_sym_struct] = ACTIONS(4478),
    [anon_sym_union] = ACTIONS(4478),
    [anon_sym_QMARK] = ACTIONS(4464),
    [anon_sym_STAR_EQ] = ACTIONS(4480),
    [anon_sym_SLASH_EQ] = ACTIONS(4480),
    [anon_sym_PERCENT_EQ] = ACTIONS(4480),
    [anon_sym_PLUS_EQ] = ACTIONS(4480),
    [anon_sym_DASH_EQ] = ACTIONS(4480),
    [anon_sym_LT_LT_EQ] = ACTIONS(4480),
    [anon_sym_GT_GT_EQ] = ACTIONS(4480),
    [anon_sym_AMP_EQ] = ACTIONS(4480),
    [anon_sym_CARET_EQ] = ACTIONS(4480),
    [anon_sym_PIPE_EQ] = ACTIONS(4480),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4478),
    [anon_sym___typeof] = ACTIONS(4478),
    [anon_sym_typeof] = ACTIONS(4478),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_API_AVAILABLE] = ACTIONS(4478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_API_DEPRECATED] = ACTIONS(4478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym___deprecated_msg] = ACTIONS(4478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4478),
    [anon_sym_BOOL] = ACTIONS(4478),
    [anon_sym_IMP] = ACTIONS(4478),
    [anon_sym_SEL] = ACTIONS(4478),
    [anon_sym_Class] = ACTIONS(4478),
    [anon_sym_id] = ACTIONS(4478),
  },
  [STATE(1567)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1568)] = {
    [sym_identifier] = ACTIONS(2346),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2348),
    [anon_sym_COMMA] = ACTIONS(2348),
    [anon_sym_RPAREN] = ACTIONS(2348),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2348),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2348),
    [anon_sym_LPAREN2] = ACTIONS(2348),
    [anon_sym_DASH] = ACTIONS(2346),
    [anon_sym_PLUS] = ACTIONS(2346),
    [anon_sym_STAR] = ACTIONS(2348),
    [anon_sym_SLASH] = ACTIONS(2346),
    [anon_sym_PERCENT] = ACTIONS(2348),
    [anon_sym_PIPE_PIPE] = ACTIONS(2348),
    [anon_sym_AMP_AMP] = ACTIONS(2348),
    [anon_sym_PIPE] = ACTIONS(2346),
    [anon_sym_CARET] = ACTIONS(2348),
    [anon_sym_AMP] = ACTIONS(2346),
    [anon_sym_EQ_EQ] = ACTIONS(2348),
    [anon_sym_BANG_EQ] = ACTIONS(2348),
    [anon_sym_GT] = ACTIONS(2346),
    [anon_sym_GT_EQ] = ACTIONS(2348),
    [anon_sym_LT_EQ] = ACTIONS(2348),
    [anon_sym_LT] = ACTIONS(2346),
    [anon_sym_LT_LT] = ACTIONS(2348),
    [anon_sym_GT_GT] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym_RBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym_LBRACE] = ACTIONS(2348),
    [anon_sym_RBRACE] = ACTIONS(2348),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_ATautoreleasepool] = ACTIONS(2348),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_COLON] = ACTIONS(2348),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [anon_sym_else] = ACTIONS(2346),
    [anon_sym_while] = ACTIONS(2346),
    [anon_sym_QMARK] = ACTIONS(2348),
    [anon_sym_DASH_DASH] = ACTIONS(2348),
    [anon_sym_PLUS_PLUS] = ACTIONS(2348),
    [anon_sym_DOT] = ACTIONS(2346),
    [anon_sym_DASH_GT] = ACTIONS(2348),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATcatch] = ACTIONS(2348),
    [anon_sym___catch] = ACTIONS(2346),
    [anon_sym_ATfinally] = ACTIONS(2348),
    [anon_sym___finally] = ACTIONS(2346),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(1569)] = {
    [sym_identifier] = ACTIONS(2350),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2352),
    [anon_sym_COMMA] = ACTIONS(2352),
    [anon_sym_RPAREN] = ACTIONS(2352),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2352),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2352),
    [anon_sym_LPAREN2] = ACTIONS(2352),
    [anon_sym_DASH] = ACTIONS(2350),
    [anon_sym_PLUS] = ACTIONS(2350),
    [anon_sym_STAR] = ACTIONS(2352),
    [anon_sym_SLASH] = ACTIONS(2350),
    [anon_sym_PERCENT] = ACTIONS(2352),
    [anon_sym_PIPE_PIPE] = ACTIONS(2352),
    [anon_sym_AMP_AMP] = ACTIONS(2352),
    [anon_sym_PIPE] = ACTIONS(2350),
    [anon_sym_CARET] = ACTIONS(2352),
    [anon_sym_AMP] = ACTIONS(2350),
    [anon_sym_EQ_EQ] = ACTIONS(2352),
    [anon_sym_BANG_EQ] = ACTIONS(2352),
    [anon_sym_GT] = ACTIONS(2350),
    [anon_sym_GT_EQ] = ACTIONS(2352),
    [anon_sym_LT_EQ] = ACTIONS(2352),
    [anon_sym_LT] = ACTIONS(2350),
    [anon_sym_LT_LT] = ACTIONS(2352),
    [anon_sym_GT_GT] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym_RBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym_LBRACE] = ACTIONS(2352),
    [anon_sym_RBRACE] = ACTIONS(2352),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_ATautoreleasepool] = ACTIONS(2352),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_COLON] = ACTIONS(2352),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [anon_sym_else] = ACTIONS(2350),
    [anon_sym_while] = ACTIONS(2350),
    [anon_sym_QMARK] = ACTIONS(2352),
    [anon_sym_DASH_DASH] = ACTIONS(2352),
    [anon_sym_PLUS_PLUS] = ACTIONS(2352),
    [anon_sym_DOT] = ACTIONS(2350),
    [anon_sym_DASH_GT] = ACTIONS(2352),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATcatch] = ACTIONS(2352),
    [anon_sym___catch] = ACTIONS(2350),
    [anon_sym_ATfinally] = ACTIONS(2352),
    [anon_sym___finally] = ACTIONS(2350),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(1570)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_STAR] = ACTIONS(4364),
    [anon_sym_SLASH] = ACTIONS(4364),
    [anon_sym_PERCENT] = ACTIONS(4364),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4444),
    [anon_sym_CARET] = ACTIONS(4446),
    [anon_sym_AMP] = ACTIONS(4448),
    [anon_sym_EQ_EQ] = ACTIONS(4450),
    [anon_sym_BANG_EQ] = ACTIONS(4450),
    [anon_sym_GT] = ACTIONS(4452),
    [anon_sym_GT_EQ] = ACTIONS(4454),
    [anon_sym_LT_EQ] = ACTIONS(4454),
    [anon_sym_LT] = ACTIONS(4452),
    [anon_sym_LT_LT] = ACTIONS(4456),
    [anon_sym_GT_GT] = ACTIONS(4456),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1571)] = {
    [sym_identifier] = ACTIONS(2360),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2362),
    [anon_sym_COMMA] = ACTIONS(2362),
    [anon_sym_RPAREN] = ACTIONS(2362),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2362),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2362),
    [anon_sym_LPAREN2] = ACTIONS(2362),
    [anon_sym_DASH] = ACTIONS(2360),
    [anon_sym_PLUS] = ACTIONS(2360),
    [anon_sym_STAR] = ACTIONS(2362),
    [anon_sym_SLASH] = ACTIONS(2360),
    [anon_sym_PERCENT] = ACTIONS(2362),
    [anon_sym_PIPE_PIPE] = ACTIONS(2362),
    [anon_sym_AMP_AMP] = ACTIONS(2362),
    [anon_sym_PIPE] = ACTIONS(2360),
    [anon_sym_CARET] = ACTIONS(2362),
    [anon_sym_AMP] = ACTIONS(2360),
    [anon_sym_EQ_EQ] = ACTIONS(2362),
    [anon_sym_BANG_EQ] = ACTIONS(2362),
    [anon_sym_GT] = ACTIONS(2360),
    [anon_sym_GT_EQ] = ACTIONS(2362),
    [anon_sym_LT_EQ] = ACTIONS(2362),
    [anon_sym_LT] = ACTIONS(2360),
    [anon_sym_LT_LT] = ACTIONS(2362),
    [anon_sym_GT_GT] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym_RBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym_LBRACE] = ACTIONS(2362),
    [anon_sym_RBRACE] = ACTIONS(2362),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_ATautoreleasepool] = ACTIONS(2362),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_COLON] = ACTIONS(2362),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [anon_sym_else] = ACTIONS(2360),
    [anon_sym_while] = ACTIONS(2360),
    [anon_sym_QMARK] = ACTIONS(2362),
    [anon_sym_DASH_DASH] = ACTIONS(2362),
    [anon_sym_PLUS_PLUS] = ACTIONS(2362),
    [anon_sym_DOT] = ACTIONS(2360),
    [anon_sym_DASH_GT] = ACTIONS(2362),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATcatch] = ACTIONS(2362),
    [anon_sym___catch] = ACTIONS(2360),
    [anon_sym_ATfinally] = ACTIONS(2362),
    [anon_sym___finally] = ACTIONS(2360),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(1572)] = {
    [sym_identifier] = ACTIONS(4024),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4024),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4024),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4024),
    [anon_sym_extern] = ACTIONS(4024),
    [anon_sym___attribute__] = ACTIONS(4024),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4024),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4024),
    [anon_sym_LBRACE] = ACTIONS(4026),
    [anon_sym_signed] = ACTIONS(4024),
    [anon_sym_unsigned] = ACTIONS(4024),
    [anon_sym_long] = ACTIONS(4024),
    [anon_sym_short] = ACTIONS(4024),
    [anon_sym_static] = ACTIONS(4024),
    [anon_sym_EQ] = ACTIONS(4028),
    [anon_sym_ATautoreleasepool] = ACTIONS(4026),
    [anon_sym_auto] = ACTIONS(4024),
    [anon_sym_register] = ACTIONS(4024),
    [anon_sym_inline] = ACTIONS(4024),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4024),
    [anon_sym___forceinline] = ACTIONS(4024),
    [anon_sym_thread_local] = ACTIONS(4024),
    [anon_sym___thread] = ACTIONS(4024),
    [anon_sym_CG_EXTERN] = ACTIONS(4024),
    [anon_sym_CG_INLINE] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4024),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4024),
    [anon_sym_IBOutlet] = ACTIONS(4024),
    [anon_sym_IBInspectable] = ACTIONS(4024),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4024),
    [anon_sym_NS_INLINE] = ACTIONS(4024),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4024),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4024),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4024),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4024),
    [anon_sym_volatile] = ACTIONS(4024),
    [anon_sym_restrict] = ACTIONS(4024),
    [anon_sym___restrict__] = ACTIONS(4024),
    [anon_sym__Atomic] = ACTIONS(4024),
    [anon_sym__Noreturn] = ACTIONS(4024),
    [anon_sym__Nonnull] = ACTIONS(4024),
    [anon_sym_nullable] = ACTIONS(4024),
    [anon_sym__Complex] = ACTIONS(4024),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4024),
    [anon_sym__Null_unspecified] = ACTIONS(4024),
    [anon_sym___autoreleasing] = ACTIONS(4024),
    [anon_sym___block] = ACTIONS(4024),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4024),
    [anon_sym___bridge_transfer] = ACTIONS(4024),
    [anon_sym___complex] = ACTIONS(4024),
    [anon_sym___const] = ACTIONS(4024),
    [anon_sym___imag] = ACTIONS(4024),
    [anon_sym___kindof] = ACTIONS(4024),
    [anon_sym___nonnull] = ACTIONS(4024),
    [anon_sym___nullable] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4024),
    [anon_sym___real] = ACTIONS(4024),
    [anon_sym___strong] = ACTIONS(4024),
    [anon_sym___unsafe_unretained] = ACTIONS(4024),
    [anon_sym___unused] = ACTIONS(4024),
    [anon_sym___weak] = ACTIONS(4024),
    [anon_sym_alignas] = ACTIONS(4024),
    [anon_sym__Alignas] = ACTIONS(4024),
    [sym_primitive_type] = ACTIONS(4024),
    [anon_sym_enum] = ACTIONS(4024),
    [anon_sym_struct] = ACTIONS(4024),
    [anon_sym_union] = ACTIONS(4024),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4030),
    [anon_sym_SLASH_EQ] = ACTIONS(4030),
    [anon_sym_PERCENT_EQ] = ACTIONS(4030),
    [anon_sym_PLUS_EQ] = ACTIONS(4030),
    [anon_sym_DASH_EQ] = ACTIONS(4030),
    [anon_sym_LT_LT_EQ] = ACTIONS(4030),
    [anon_sym_GT_GT_EQ] = ACTIONS(4030),
    [anon_sym_AMP_EQ] = ACTIONS(4030),
    [anon_sym_CARET_EQ] = ACTIONS(4030),
    [anon_sym_PIPE_EQ] = ACTIONS(4030),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4024),
    [anon_sym___typeof] = ACTIONS(4024),
    [anon_sym_typeof] = ACTIONS(4024),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4024),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4024),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_API_AVAILABLE] = ACTIONS(4024),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_API_DEPRECATED] = ACTIONS(4024),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym___deprecated_msg] = ACTIONS(4024),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4024),
    [anon_sym_BOOL] = ACTIONS(4024),
    [anon_sym_IMP] = ACTIONS(4024),
    [anon_sym_SEL] = ACTIONS(4024),
    [anon_sym_Class] = ACTIONS(4024),
    [anon_sym_id] = ACTIONS(4024),
  },
  [STATE(1573)] = {
    [sym__declaration_modifiers] = STATE(1726),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(1726),
    [sym_attribute_declaration] = STATE(1726),
    [sym_ms_declspec_modifier] = STATE(1726),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5206),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_compound_statement] = STATE(6762),
    [sym_storage_class_specifier] = STATE(1726),
    [sym_type_qualifier] = STATE(1726),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5455),
    [sym_parameter_declaration] = STATE(5455),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1726),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1726),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4078),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym_RPAREN] = ACTIONS(4082),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1574)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5287),
    [sym__abstract_declarator] = STATE(5277),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_parameter_list] = STATE(4271),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4078),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(4086),
    [anon_sym_CARET] = ACTIONS(4088),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1575)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5173),
    [sym__abstract_declarator] = STATE(5277),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_parameter_list] = STATE(4271),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4078),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(4086),
    [anon_sym_CARET] = ACTIONS(4088),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1576)] = {
    [sym__declaration_modifiers] = STATE(1735),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(1735),
    [sym_attribute_declaration] = STATE(1735),
    [sym_ms_declspec_modifier] = STATE(1735),
    [sym__abstract_declarator] = STATE(5183),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_compound_statement] = STATE(6762),
    [sym_storage_class_specifier] = STATE(1735),
    [sym_type_qualifier] = STATE(1735),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5455),
    [sym_parameter_list] = STATE(4271),
    [sym_parameter_declaration] = STATE(5455),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1735),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1735),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym_RPAREN] = ACTIONS(4082),
    [anon_sym_LPAREN2] = ACTIONS(4488),
    [anon_sym_STAR] = ACTIONS(4490),
    [anon_sym_CARET] = ACTIONS(4492),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1577)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4556),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6485),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1578)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4488),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(5901),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1579)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4513),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1580)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4519),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1581)] = {
    [sym_preproc_def] = STATE(1611),
    [sym_preproc_function_def] = STATE(1611),
    [sym_preproc_call] = STATE(1611),
    [sym_preproc_if_in_field_declaration_list] = STATE(1611),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1611),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4603),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1611),
    [sym_field_declaration] = STATE(1611),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1611),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1611),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(4521),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4523),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_RBRACE] = ACTIONS(4525),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4527),
  },
  [STATE(1582)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4529),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1583)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4531),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1584)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4535),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6373),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1585)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_in] = ACTIONS(4533),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1586)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4535),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1587)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4506),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4513),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1588)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4538),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6638),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1589)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_in] = ACTIONS(4537),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1590)] = {
    [sym_preproc_def] = STATE(1581),
    [sym_preproc_function_def] = STATE(1581),
    [sym_preproc_call] = STATE(1581),
    [sym_preproc_if_in_field_declaration_list] = STATE(1581),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1581),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4603),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1581),
    [sym_field_declaration] = STATE(1581),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1581),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1581),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(4521),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4523),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_RBRACE] = ACTIONS(4539),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4527),
  },
  [STATE(1591)] = {
    [sym_preproc_def] = STATE(1603),
    [sym_preproc_function_def] = STATE(1603),
    [sym_preproc_call] = STATE(1603),
    [sym_preproc_if_in_field_declaration_list] = STATE(1603),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1603),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4603),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1603),
    [sym_field_declaration] = STATE(1603),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1603),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1603),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(4521),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4523),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_RBRACE] = ACTIONS(4541),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4527),
  },
  [STATE(1592)] = {
    [sym_preproc_def] = STATE(1592),
    [sym_preproc_function_def] = STATE(1592),
    [sym_preproc_call] = STATE(1592),
    [sym_preproc_if_in_field_declaration_list] = STATE(1592),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1592),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4604),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1592),
    [sym_field_declaration] = STATE(1592),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1592),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1592),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4366),
    [aux_sym_preproc_def_token1] = ACTIONS(4543),
    [aux_sym_preproc_if_token1] = ACTIONS(4546),
    [aux_sym_preproc_if_token2] = ACTIONS(4375),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4549),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4549),
    [sym_preproc_directive] = ACTIONS(4552),
    [anon_sym___extension__] = ACTIONS(4383),
    [anon_sym_extern] = ACTIONS(4386),
    [anon_sym___attribute__] = ACTIONS(4389),
    [anon_sym___attribute] = ACTIONS(4389),
    [anon_sym_noreturn] = ACTIONS(4383),
    [anon_sym_LBRACK] = ACTIONS(4392),
    [anon_sym___declspec] = ACTIONS(4395),
    [anon_sym_signed] = ACTIONS(4398),
    [anon_sym_unsigned] = ACTIONS(4398),
    [anon_sym_long] = ACTIONS(4398),
    [anon_sym_short] = ACTIONS(4398),
    [anon_sym_static] = ACTIONS(4386),
    [anon_sym_auto] = ACTIONS(4386),
    [anon_sym_register] = ACTIONS(4386),
    [anon_sym_inline] = ACTIONS(4386),
    [anon_sym___inline] = ACTIONS(4386),
    [anon_sym___inline__] = ACTIONS(4386),
    [anon_sym___forceinline] = ACTIONS(4386),
    [anon_sym_thread_local] = ACTIONS(4386),
    [anon_sym___thread] = ACTIONS(4386),
    [anon_sym_CG_EXTERN] = ACTIONS(4386),
    [anon_sym_CG_INLINE] = ACTIONS(4386),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4386),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4386),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4386),
    [anon_sym_IBOutlet] = ACTIONS(4386),
    [anon_sym_IBInspectable] = ACTIONS(4386),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4386),
    [anon_sym_NS_INLINE] = ACTIONS(4386),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4386),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4386),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4386),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4386),
    [anon_sym_const] = ACTIONS(4383),
    [anon_sym_constexpr] = ACTIONS(4383),
    [anon_sym_volatile] = ACTIONS(4383),
    [anon_sym_restrict] = ACTIONS(4383),
    [anon_sym___restrict__] = ACTIONS(4383),
    [anon_sym__Atomic] = ACTIONS(4383),
    [anon_sym__Noreturn] = ACTIONS(4383),
    [anon_sym__Nonnull] = ACTIONS(4383),
    [anon_sym_nullable] = ACTIONS(4383),
    [anon_sym__Complex] = ACTIONS(4383),
    [anon_sym__Nullable] = ACTIONS(4383),
    [anon_sym__Nullable_result] = ACTIONS(4383),
    [anon_sym__Null_unspecified] = ACTIONS(4383),
    [anon_sym___autoreleasing] = ACTIONS(4383),
    [anon_sym___block] = ACTIONS(4383),
    [anon_sym___bridge] = ACTIONS(4383),
    [anon_sym___bridge_retained] = ACTIONS(4383),
    [anon_sym___bridge_transfer] = ACTIONS(4383),
    [anon_sym___complex] = ACTIONS(4383),
    [anon_sym___const] = ACTIONS(4383),
    [anon_sym___imag] = ACTIONS(4383),
    [anon_sym___kindof] = ACTIONS(4383),
    [anon_sym___nonnull] = ACTIONS(4383),
    [anon_sym___nullable] = ACTIONS(4383),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4383),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4383),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4383),
    [anon_sym___real] = ACTIONS(4383),
    [anon_sym___strong] = ACTIONS(4383),
    [anon_sym___unsafe_unretained] = ACTIONS(4383),
    [anon_sym___unused] = ACTIONS(4383),
    [anon_sym___weak] = ACTIONS(4383),
    [anon_sym_alignas] = ACTIONS(4401),
    [anon_sym__Alignas] = ACTIONS(4401),
    [sym_primitive_type] = ACTIONS(4404),
    [anon_sym_enum] = ACTIONS(4407),
    [anon_sym_struct] = ACTIONS(4410),
    [anon_sym_union] = ACTIONS(4413),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4416),
    [anon_sym___typeof] = ACTIONS(4416),
    [anon_sym_typeof] = ACTIONS(4416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4419),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4419),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4419),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4419),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4419),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4419),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4419),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4419),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4419),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4422),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4422),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4422),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_API_AVAILABLE] = ACTIONS(4422),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4422),
    [anon_sym_API_DEPRECATED] = ACTIONS(4422),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4422),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4422),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4422),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4422),
    [anon_sym___deprecated_msg] = ACTIONS(4422),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4422),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4422),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4422),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4422),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4422),
    [anon_sym_BOOL] = ACTIONS(4425),
    [anon_sym_IMP] = ACTIONS(4425),
    [anon_sym_SEL] = ACTIONS(4425),
    [anon_sym_Class] = ACTIONS(4425),
    [anon_sym_id] = ACTIONS(4425),
    [anon_sym_ATdefs] = ACTIONS(4555),
  },
  [STATE(1593)] = {
    [sym_string_literal] = STATE(1599),
    [aux_sym_concatenated_string_repeat1] = STATE(1599),
    [sym_identifier] = ACTIONS(4558),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3944),
    [anon_sym_COMMA] = ACTIONS(3944),
    [anon_sym_LPAREN2] = ACTIONS(3944),
    [anon_sym_DASH] = ACTIONS(3946),
    [anon_sym_PLUS] = ACTIONS(3946),
    [anon_sym_STAR] = ACTIONS(3944),
    [anon_sym_SLASH] = ACTIONS(3946),
    [anon_sym_PERCENT] = ACTIONS(3944),
    [anon_sym_PIPE_PIPE] = ACTIONS(3944),
    [anon_sym_AMP_AMP] = ACTIONS(3944),
    [anon_sym_PIPE] = ACTIONS(3946),
    [anon_sym_CARET] = ACTIONS(3944),
    [anon_sym_AMP] = ACTIONS(3946),
    [anon_sym_EQ_EQ] = ACTIONS(3944),
    [anon_sym_BANG_EQ] = ACTIONS(3944),
    [anon_sym_GT] = ACTIONS(3946),
    [anon_sym_GT_EQ] = ACTIONS(3944),
    [anon_sym_LT_EQ] = ACTIONS(3944),
    [anon_sym_LT] = ACTIONS(3946),
    [anon_sym_LT_LT] = ACTIONS(3944),
    [anon_sym_GT_GT] = ACTIONS(3944),
    [anon_sym_SEMI] = ACTIONS(3944),
    [anon_sym___extension__] = ACTIONS(3946),
    [anon_sym_extern] = ACTIONS(3946),
    [anon_sym___attribute__] = ACTIONS(3946),
    [anon_sym___attribute] = ACTIONS(3946),
    [anon_sym_noreturn] = ACTIONS(3946),
    [anon_sym_LBRACK] = ACTIONS(3944),
    [anon_sym___declspec] = ACTIONS(3946),
    [anon_sym_LBRACE] = ACTIONS(3944),
    [anon_sym_signed] = ACTIONS(3946),
    [anon_sym_unsigned] = ACTIONS(3946),
    [anon_sym_long] = ACTIONS(3946),
    [anon_sym_short] = ACTIONS(3946),
    [anon_sym_static] = ACTIONS(3946),
    [anon_sym_ATautoreleasepool] = ACTIONS(3944),
    [anon_sym_auto] = ACTIONS(3946),
    [anon_sym_register] = ACTIONS(3946),
    [anon_sym_inline] = ACTIONS(3946),
    [anon_sym___inline] = ACTIONS(3946),
    [anon_sym___inline__] = ACTIONS(3946),
    [anon_sym___forceinline] = ACTIONS(3946),
    [anon_sym_thread_local] = ACTIONS(3946),
    [anon_sym___thread] = ACTIONS(3946),
    [anon_sym_CG_EXTERN] = ACTIONS(3946),
    [anon_sym_CG_INLINE] = ACTIONS(3946),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3946),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3946),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3946),
    [anon_sym_IBOutlet] = ACTIONS(3946),
    [anon_sym_IBInspectable] = ACTIONS(3946),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3946),
    [anon_sym_NS_INLINE] = ACTIONS(3946),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3946),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3946),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3946),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3946),
    [anon_sym_const] = ACTIONS(3946),
    [anon_sym_constexpr] = ACTIONS(3946),
    [anon_sym_volatile] = ACTIONS(3946),
    [anon_sym_restrict] = ACTIONS(3946),
    [anon_sym___restrict__] = ACTIONS(3946),
    [anon_sym__Atomic] = ACTIONS(3946),
    [anon_sym__Noreturn] = ACTIONS(3946),
    [anon_sym__Nonnull] = ACTIONS(3946),
    [anon_sym_nullable] = ACTIONS(3946),
    [anon_sym__Complex] = ACTIONS(3946),
    [anon_sym__Nullable] = ACTIONS(3946),
    [anon_sym__Nullable_result] = ACTIONS(3946),
    [anon_sym__Null_unspecified] = ACTIONS(3946),
    [anon_sym___autoreleasing] = ACTIONS(3946),
    [anon_sym___block] = ACTIONS(3946),
    [anon_sym___bridge] = ACTIONS(3946),
    [anon_sym___bridge_retained] = ACTIONS(3946),
    [anon_sym___bridge_transfer] = ACTIONS(3946),
    [anon_sym___complex] = ACTIONS(3946),
    [anon_sym___const] = ACTIONS(3946),
    [anon_sym___imag] = ACTIONS(3946),
    [anon_sym___kindof] = ACTIONS(3946),
    [anon_sym___nonnull] = ACTIONS(3946),
    [anon_sym___nullable] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3946),
    [anon_sym___real] = ACTIONS(3946),
    [anon_sym___strong] = ACTIONS(3946),
    [anon_sym___unsafe_unretained] = ACTIONS(3946),
    [anon_sym___unused] = ACTIONS(3946),
    [anon_sym___weak] = ACTIONS(3946),
    [anon_sym_alignas] = ACTIONS(3946),
    [anon_sym__Alignas] = ACTIONS(3946),
    [sym_primitive_type] = ACTIONS(3946),
    [anon_sym_enum] = ACTIONS(3946),
    [anon_sym_struct] = ACTIONS(3946),
    [anon_sym_union] = ACTIONS(3946),
    [anon_sym_QMARK] = ACTIONS(3944),
    [anon_sym_DASH_DASH] = ACTIONS(3944),
    [anon_sym_PLUS_PLUS] = ACTIONS(3944),
    [anon_sym_DOT] = ACTIONS(3946),
    [anon_sym_DASH_GT] = ACTIONS(3944),
    [anon_sym_AT] = ACTIONS(4032),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3946),
    [anon_sym___typeof] = ACTIONS(3946),
    [anon_sym_typeof] = ACTIONS(3946),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3946),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3946),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3946),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3946),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3946),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3946),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3946),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_API_AVAILABLE] = ACTIONS(3946),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_API_DEPRECATED] = ACTIONS(3946),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3946),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym___deprecated_msg] = ACTIONS(3946),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3946),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3946),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3946),
    [anon_sym_BOOL] = ACTIONS(3946),
    [anon_sym_IMP] = ACTIONS(3946),
    [anon_sym_SEL] = ACTIONS(3946),
    [anon_sym_Class] = ACTIONS(3946),
    [anon_sym_id] = ACTIONS(3946),
  },
  [STATE(1594)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4560),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1595)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_in] = ACTIONS(4562),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1596)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4506),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4564),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1597)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4566),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1598)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_in] = ACTIONS(4568),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1599)] = {
    [sym_string_literal] = STATE(1599),
    [aux_sym_concatenated_string_repeat1] = STATE(1599),
    [sym_identifier] = ACTIONS(4570),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_LPAREN2] = ACTIONS(3953),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_STAR] = ACTIONS(3953),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_PERCENT] = ACTIONS(3953),
    [anon_sym_PIPE_PIPE] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3953),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3953),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_EQ_EQ] = ACTIONS(3953),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_LT_EQ] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3953),
    [anon_sym_GT_GT] = ACTIONS(3953),
    [anon_sym_SEMI] = ACTIONS(3953),
    [anon_sym___extension__] = ACTIONS(3955),
    [anon_sym_extern] = ACTIONS(3955),
    [anon_sym___attribute__] = ACTIONS(3955),
    [anon_sym___attribute] = ACTIONS(3955),
    [anon_sym_noreturn] = ACTIONS(3955),
    [anon_sym_LBRACK] = ACTIONS(3953),
    [anon_sym___declspec] = ACTIONS(3955),
    [anon_sym_LBRACE] = ACTIONS(3953),
    [anon_sym_signed] = ACTIONS(3955),
    [anon_sym_unsigned] = ACTIONS(3955),
    [anon_sym_long] = ACTIONS(3955),
    [anon_sym_short] = ACTIONS(3955),
    [anon_sym_static] = ACTIONS(3955),
    [anon_sym_ATautoreleasepool] = ACTIONS(3953),
    [anon_sym_auto] = ACTIONS(3955),
    [anon_sym_register] = ACTIONS(3955),
    [anon_sym_inline] = ACTIONS(3955),
    [anon_sym___inline] = ACTIONS(3955),
    [anon_sym___inline__] = ACTIONS(3955),
    [anon_sym___forceinline] = ACTIONS(3955),
    [anon_sym_thread_local] = ACTIONS(3955),
    [anon_sym___thread] = ACTIONS(3955),
    [anon_sym_CG_EXTERN] = ACTIONS(3955),
    [anon_sym_CG_INLINE] = ACTIONS(3955),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3955),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3955),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3955),
    [anon_sym_IBOutlet] = ACTIONS(3955),
    [anon_sym_IBInspectable] = ACTIONS(3955),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3955),
    [anon_sym_NS_INLINE] = ACTIONS(3955),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3955),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3955),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3955),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3955),
    [anon_sym_const] = ACTIONS(3955),
    [anon_sym_constexpr] = ACTIONS(3955),
    [anon_sym_volatile] = ACTIONS(3955),
    [anon_sym_restrict] = ACTIONS(3955),
    [anon_sym___restrict__] = ACTIONS(3955),
    [anon_sym__Atomic] = ACTIONS(3955),
    [anon_sym__Noreturn] = ACTIONS(3955),
    [anon_sym__Nonnull] = ACTIONS(3955),
    [anon_sym_nullable] = ACTIONS(3955),
    [anon_sym__Complex] = ACTIONS(3955),
    [anon_sym__Nullable] = ACTIONS(3955),
    [anon_sym__Nullable_result] = ACTIONS(3955),
    [anon_sym__Null_unspecified] = ACTIONS(3955),
    [anon_sym___autoreleasing] = ACTIONS(3955),
    [anon_sym___block] = ACTIONS(3955),
    [anon_sym___bridge] = ACTIONS(3955),
    [anon_sym___bridge_retained] = ACTIONS(3955),
    [anon_sym___bridge_transfer] = ACTIONS(3955),
    [anon_sym___complex] = ACTIONS(3955),
    [anon_sym___const] = ACTIONS(3955),
    [anon_sym___imag] = ACTIONS(3955),
    [anon_sym___kindof] = ACTIONS(3955),
    [anon_sym___nonnull] = ACTIONS(3955),
    [anon_sym___nullable] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3955),
    [anon_sym___real] = ACTIONS(3955),
    [anon_sym___strong] = ACTIONS(3955),
    [anon_sym___unsafe_unretained] = ACTIONS(3955),
    [anon_sym___unused] = ACTIONS(3955),
    [anon_sym___weak] = ACTIONS(3955),
    [anon_sym_alignas] = ACTIONS(3955),
    [anon_sym__Alignas] = ACTIONS(3955),
    [sym_primitive_type] = ACTIONS(3955),
    [anon_sym_enum] = ACTIONS(3955),
    [anon_sym_struct] = ACTIONS(3955),
    [anon_sym_union] = ACTIONS(3955),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_AT] = ACTIONS(4573),
    [anon_sym_DQUOTE] = ACTIONS(4576),
    [anon_sym_L_DQUOTE] = ACTIONS(4576),
    [anon_sym_u_DQUOTE] = ACTIONS(4576),
    [anon_sym_U_DQUOTE] = ACTIONS(4576),
    [anon_sym_u8_DQUOTE] = ACTIONS(4576),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3955),
    [anon_sym___typeof] = ACTIONS(3955),
    [anon_sym_typeof] = ACTIONS(3955),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3955),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3955),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3955),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3955),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3955),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3955),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3955),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_API_AVAILABLE] = ACTIONS(3955),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_API_DEPRECATED] = ACTIONS(3955),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3955),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym___deprecated_msg] = ACTIONS(3955),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3955),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3955),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3955),
    [anon_sym_BOOL] = ACTIONS(3955),
    [anon_sym_IMP] = ACTIONS(3955),
    [anon_sym_SEL] = ACTIONS(3955),
    [anon_sym_Class] = ACTIONS(3955),
    [anon_sym_id] = ACTIONS(3955),
  },
  [STATE(1600)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_in] = ACTIONS(4579),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1601)] = {
    [sym_string_literal] = STATE(1593),
    [aux_sym_concatenated_string_repeat1] = STATE(1593),
    [sym_identifier] = ACTIONS(4581),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_LPAREN2] = ACTIONS(3965),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_STAR] = ACTIONS(3965),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_PERCENT] = ACTIONS(3965),
    [anon_sym_PIPE_PIPE] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3965),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3965),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_EQ_EQ] = ACTIONS(3965),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_LT_EQ] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3965),
    [anon_sym_GT_GT] = ACTIONS(3965),
    [anon_sym_SEMI] = ACTIONS(3965),
    [anon_sym___extension__] = ACTIONS(3967),
    [anon_sym_extern] = ACTIONS(3967),
    [anon_sym___attribute__] = ACTIONS(3967),
    [anon_sym___attribute] = ACTIONS(3967),
    [anon_sym_noreturn] = ACTIONS(3967),
    [anon_sym_LBRACK] = ACTIONS(3965),
    [anon_sym___declspec] = ACTIONS(3967),
    [anon_sym_LBRACE] = ACTIONS(3965),
    [anon_sym_signed] = ACTIONS(3967),
    [anon_sym_unsigned] = ACTIONS(3967),
    [anon_sym_long] = ACTIONS(3967),
    [anon_sym_short] = ACTIONS(3967),
    [anon_sym_static] = ACTIONS(3967),
    [anon_sym_ATautoreleasepool] = ACTIONS(3965),
    [anon_sym_auto] = ACTIONS(3967),
    [anon_sym_register] = ACTIONS(3967),
    [anon_sym_inline] = ACTIONS(3967),
    [anon_sym___inline] = ACTIONS(3967),
    [anon_sym___inline__] = ACTIONS(3967),
    [anon_sym___forceinline] = ACTIONS(3967),
    [anon_sym_thread_local] = ACTIONS(3967),
    [anon_sym___thread] = ACTIONS(3967),
    [anon_sym_CG_EXTERN] = ACTIONS(3967),
    [anon_sym_CG_INLINE] = ACTIONS(3967),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3967),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3967),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3967),
    [anon_sym_IBOutlet] = ACTIONS(3967),
    [anon_sym_IBInspectable] = ACTIONS(3967),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3967),
    [anon_sym_NS_INLINE] = ACTIONS(3967),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3967),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3967),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3967),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3967),
    [anon_sym_const] = ACTIONS(3967),
    [anon_sym_constexpr] = ACTIONS(3967),
    [anon_sym_volatile] = ACTIONS(3967),
    [anon_sym_restrict] = ACTIONS(3967),
    [anon_sym___restrict__] = ACTIONS(3967),
    [anon_sym__Atomic] = ACTIONS(3967),
    [anon_sym__Noreturn] = ACTIONS(3967),
    [anon_sym__Nonnull] = ACTIONS(3967),
    [anon_sym_nullable] = ACTIONS(3967),
    [anon_sym__Complex] = ACTIONS(3967),
    [anon_sym__Nullable] = ACTIONS(3967),
    [anon_sym__Nullable_result] = ACTIONS(3967),
    [anon_sym__Null_unspecified] = ACTIONS(3967),
    [anon_sym___autoreleasing] = ACTIONS(3967),
    [anon_sym___block] = ACTIONS(3967),
    [anon_sym___bridge] = ACTIONS(3967),
    [anon_sym___bridge_retained] = ACTIONS(3967),
    [anon_sym___bridge_transfer] = ACTIONS(3967),
    [anon_sym___complex] = ACTIONS(3967),
    [anon_sym___const] = ACTIONS(3967),
    [anon_sym___imag] = ACTIONS(3967),
    [anon_sym___kindof] = ACTIONS(3967),
    [anon_sym___nonnull] = ACTIONS(3967),
    [anon_sym___nullable] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3967),
    [anon_sym___real] = ACTIONS(3967),
    [anon_sym___strong] = ACTIONS(3967),
    [anon_sym___unsafe_unretained] = ACTIONS(3967),
    [anon_sym___unused] = ACTIONS(3967),
    [anon_sym___weak] = ACTIONS(3967),
    [anon_sym_alignas] = ACTIONS(3967),
    [anon_sym__Alignas] = ACTIONS(3967),
    [sym_primitive_type] = ACTIONS(3967),
    [anon_sym_enum] = ACTIONS(3967),
    [anon_sym_struct] = ACTIONS(3967),
    [anon_sym_union] = ACTIONS(3967),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_AT] = ACTIONS(4032),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3967),
    [anon_sym___typeof] = ACTIONS(3967),
    [anon_sym_typeof] = ACTIONS(3967),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3967),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3967),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3967),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3967),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3967),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3967),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3967),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_API_AVAILABLE] = ACTIONS(3967),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_API_DEPRECATED] = ACTIONS(3967),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3967),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym___deprecated_msg] = ACTIONS(3967),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3967),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3967),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3967),
    [anon_sym_BOOL] = ACTIONS(3967),
    [anon_sym_IMP] = ACTIONS(3967),
    [anon_sym_SEL] = ACTIONS(3967),
    [anon_sym_Class] = ACTIONS(3967),
    [anon_sym_id] = ACTIONS(3967),
  },
  [STATE(1602)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4502),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(5943),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1603)] = {
    [sym_preproc_def] = STATE(1611),
    [sym_preproc_function_def] = STATE(1611),
    [sym_preproc_call] = STATE(1611),
    [sym_preproc_if_in_field_declaration_list] = STATE(1611),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1611),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4603),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1611),
    [sym_field_declaration] = STATE(1611),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1611),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1611),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(2996),
    [aux_sym_preproc_if_token1] = ACTIONS(4521),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4523),
    [sym_preproc_directive] = ACTIONS(3002),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_RBRACE] = ACTIONS(4583),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4527),
  },
  [STATE(1604)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_in] = ACTIONS(4585),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1605)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4534),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6339),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1606)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4506),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4531),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1607)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_in] = ACTIONS(4587),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1608)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4564),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1609)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_in] = ACTIONS(4589),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1610)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4506),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4529),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1611)] = {
    [sym_preproc_def] = STATE(1611),
    [sym_preproc_function_def] = STATE(1611),
    [sym_preproc_call] = STATE(1611),
    [sym_preproc_if_in_field_declaration_list] = STATE(1611),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1611),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4603),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1611),
    [sym_field_declaration] = STATE(1611),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1611),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1611),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4366),
    [aux_sym_preproc_def_token1] = ACTIONS(4591),
    [aux_sym_preproc_if_token1] = ACTIONS(4594),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4597),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4597),
    [sym_preproc_directive] = ACTIONS(4600),
    [anon_sym___extension__] = ACTIONS(4383),
    [anon_sym_extern] = ACTIONS(4386),
    [anon_sym___attribute__] = ACTIONS(4389),
    [anon_sym___attribute] = ACTIONS(4389),
    [anon_sym_noreturn] = ACTIONS(4383),
    [anon_sym_LBRACK] = ACTIONS(4392),
    [anon_sym___declspec] = ACTIONS(4395),
    [anon_sym_RBRACE] = ACTIONS(4603),
    [anon_sym_signed] = ACTIONS(4398),
    [anon_sym_unsigned] = ACTIONS(4398),
    [anon_sym_long] = ACTIONS(4398),
    [anon_sym_short] = ACTIONS(4398),
    [anon_sym_static] = ACTIONS(4386),
    [anon_sym_auto] = ACTIONS(4386),
    [anon_sym_register] = ACTIONS(4386),
    [anon_sym_inline] = ACTIONS(4386),
    [anon_sym___inline] = ACTIONS(4386),
    [anon_sym___inline__] = ACTIONS(4386),
    [anon_sym___forceinline] = ACTIONS(4386),
    [anon_sym_thread_local] = ACTIONS(4386),
    [anon_sym___thread] = ACTIONS(4386),
    [anon_sym_CG_EXTERN] = ACTIONS(4386),
    [anon_sym_CG_INLINE] = ACTIONS(4386),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4386),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4386),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4386),
    [anon_sym_IBOutlet] = ACTIONS(4386),
    [anon_sym_IBInspectable] = ACTIONS(4386),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4386),
    [anon_sym_NS_INLINE] = ACTIONS(4386),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4386),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4386),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4386),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4386),
    [anon_sym_const] = ACTIONS(4383),
    [anon_sym_constexpr] = ACTIONS(4383),
    [anon_sym_volatile] = ACTIONS(4383),
    [anon_sym_restrict] = ACTIONS(4383),
    [anon_sym___restrict__] = ACTIONS(4383),
    [anon_sym__Atomic] = ACTIONS(4383),
    [anon_sym__Noreturn] = ACTIONS(4383),
    [anon_sym__Nonnull] = ACTIONS(4383),
    [anon_sym_nullable] = ACTIONS(4383),
    [anon_sym__Complex] = ACTIONS(4383),
    [anon_sym__Nullable] = ACTIONS(4383),
    [anon_sym__Nullable_result] = ACTIONS(4383),
    [anon_sym__Null_unspecified] = ACTIONS(4383),
    [anon_sym___autoreleasing] = ACTIONS(4383),
    [anon_sym___block] = ACTIONS(4383),
    [anon_sym___bridge] = ACTIONS(4383),
    [anon_sym___bridge_retained] = ACTIONS(4383),
    [anon_sym___bridge_transfer] = ACTIONS(4383),
    [anon_sym___complex] = ACTIONS(4383),
    [anon_sym___const] = ACTIONS(4383),
    [anon_sym___imag] = ACTIONS(4383),
    [anon_sym___kindof] = ACTIONS(4383),
    [anon_sym___nonnull] = ACTIONS(4383),
    [anon_sym___nullable] = ACTIONS(4383),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4383),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4383),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4383),
    [anon_sym___real] = ACTIONS(4383),
    [anon_sym___strong] = ACTIONS(4383),
    [anon_sym___unsafe_unretained] = ACTIONS(4383),
    [anon_sym___unused] = ACTIONS(4383),
    [anon_sym___weak] = ACTIONS(4383),
    [anon_sym_alignas] = ACTIONS(4401),
    [anon_sym__Alignas] = ACTIONS(4401),
    [sym_primitive_type] = ACTIONS(4404),
    [anon_sym_enum] = ACTIONS(4407),
    [anon_sym_struct] = ACTIONS(4410),
    [anon_sym_union] = ACTIONS(4413),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4416),
    [anon_sym___typeof] = ACTIONS(4416),
    [anon_sym_typeof] = ACTIONS(4416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4419),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4419),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4419),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4419),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4419),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4419),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4419),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4419),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4419),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4422),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4422),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4422),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_API_AVAILABLE] = ACTIONS(4422),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4422),
    [anon_sym_API_DEPRECATED] = ACTIONS(4422),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4422),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4422),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4422),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4422),
    [anon_sym___deprecated_msg] = ACTIONS(4422),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4422),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4422),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4422),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4422),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4422),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4422),
    [anon_sym_BOOL] = ACTIONS(4425),
    [anon_sym_IMP] = ACTIONS(4425),
    [anon_sym_SEL] = ACTIONS(4425),
    [anon_sym_Class] = ACTIONS(4425),
    [anon_sym_id] = ACTIONS(4425),
    [anon_sym_ATdefs] = ACTIONS(4605),
  },
  [STATE(1612)] = {
    [sym_preproc_def] = STATE(1592),
    [sym_preproc_function_def] = STATE(1592),
    [sym_preproc_call] = STATE(1592),
    [sym_preproc_if_in_field_declaration_list] = STATE(1592),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1592),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4604),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1592),
    [sym_field_declaration] = STATE(1592),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1592),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1592),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3353),
    [aux_sym_preproc_if_token1] = ACTIONS(4608),
    [aux_sym_preproc_if_token2] = ACTIONS(4610),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4612),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4612),
    [sym_preproc_directive] = ACTIONS(3361),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4614),
  },
  [STATE(1613)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_expression] = STATE(4548),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_type_descriptor] = STATE(6062),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(4336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(4346),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(4350),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(4352),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(4352),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(187),
    [anon_sym_union] = ACTIONS(67),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(243),
  },
  [STATE(1614)] = {
    [sym_preproc_def] = STATE(1612),
    [sym_preproc_function_def] = STATE(1612),
    [sym_preproc_call] = STATE(1612),
    [sym_preproc_if_in_field_declaration_list] = STATE(1612),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(1612),
    [sym__declaration_modifiers] = STATE(2232),
    [sym__declaration_specifiers] = STATE(4604),
    [sym_attribute_specifier] = STATE(2232),
    [sym_attribute_declaration] = STATE(2232),
    [sym_ms_declspec_modifier] = STATE(2232),
    [sym_storage_class_specifier] = STATE(2232),
    [sym_type_qualifier] = STATE(2232),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym__field_declaration_list_item] = STATE(1612),
    [sym_field_declaration] = STATE(1612),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2232),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [sym_atdef_field] = STATE(1612),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(1612),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2232),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [aux_sym_preproc_def_token1] = ACTIONS(3353),
    [aux_sym_preproc_if_token1] = ACTIONS(4608),
    [aux_sym_preproc_if_token2] = ACTIONS(4616),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4612),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4612),
    [sym_preproc_directive] = ACTIONS(3361),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
    [anon_sym_ATdefs] = ACTIONS(4614),
  },
  [STATE(1615)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(3976),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3978),
    [anon_sym_COMMA] = ACTIONS(3978),
    [anon_sym_RPAREN] = ACTIONS(3978),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3978),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3978),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(3976),
    [anon_sym_PLUS] = ACTIONS(3976),
    [anon_sym_STAR] = ACTIONS(3978),
    [anon_sym_SLASH] = ACTIONS(3976),
    [anon_sym_PERCENT] = ACTIONS(3978),
    [anon_sym_PIPE_PIPE] = ACTIONS(3978),
    [anon_sym_AMP_AMP] = ACTIONS(3978),
    [anon_sym_PIPE] = ACTIONS(3976),
    [anon_sym_CARET] = ACTIONS(3978),
    [anon_sym_AMP] = ACTIONS(3976),
    [anon_sym_EQ_EQ] = ACTIONS(3978),
    [anon_sym_BANG_EQ] = ACTIONS(3978),
    [anon_sym_GT] = ACTIONS(3976),
    [anon_sym_GT_EQ] = ACTIONS(3978),
    [anon_sym_LT_EQ] = ACTIONS(3978),
    [anon_sym_LT] = ACTIONS(3976),
    [anon_sym_LT_LT] = ACTIONS(3978),
    [anon_sym_GT_GT] = ACTIONS(3978),
    [anon_sym_SEMI] = ACTIONS(3978),
    [anon_sym___extension__] = ACTIONS(3976),
    [anon_sym_extern] = ACTIONS(3976),
    [anon_sym___attribute__] = ACTIONS(3976),
    [anon_sym___attribute] = ACTIONS(3976),
    [anon_sym_noreturn] = ACTIONS(3976),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(3978),
    [anon_sym___declspec] = ACTIONS(3976),
    [anon_sym_LBRACE] = ACTIONS(3978),
    [anon_sym_RBRACE] = ACTIONS(3978),
    [anon_sym_signed] = ACTIONS(3976),
    [anon_sym_unsigned] = ACTIONS(3976),
    [anon_sym_long] = ACTIONS(3976),
    [anon_sym_short] = ACTIONS(3976),
    [anon_sym_static] = ACTIONS(3976),
    [anon_sym_ATautoreleasepool] = ACTIONS(3978),
    [anon_sym_auto] = ACTIONS(3976),
    [anon_sym_register] = ACTIONS(3976),
    [anon_sym_inline] = ACTIONS(3976),
    [anon_sym___inline] = ACTIONS(3976),
    [anon_sym___inline__] = ACTIONS(3976),
    [anon_sym___forceinline] = ACTIONS(3976),
    [anon_sym_thread_local] = ACTIONS(3976),
    [anon_sym___thread] = ACTIONS(3976),
    [anon_sym_CG_EXTERN] = ACTIONS(3976),
    [anon_sym_CG_INLINE] = ACTIONS(3976),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3976),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3976),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3976),
    [anon_sym_IBOutlet] = ACTIONS(3976),
    [anon_sym_IBInspectable] = ACTIONS(3976),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3976),
    [anon_sym_NS_INLINE] = ACTIONS(3976),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3976),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3976),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3976),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3976),
    [anon_sym_const] = ACTIONS(3976),
    [anon_sym_constexpr] = ACTIONS(3976),
    [anon_sym_volatile] = ACTIONS(3976),
    [anon_sym_restrict] = ACTIONS(3976),
    [anon_sym___restrict__] = ACTIONS(3976),
    [anon_sym__Atomic] = ACTIONS(3976),
    [anon_sym__Noreturn] = ACTIONS(3976),
    [anon_sym__Nonnull] = ACTIONS(3976),
    [anon_sym_nullable] = ACTIONS(3976),
    [anon_sym__Complex] = ACTIONS(3976),
    [anon_sym__Nullable] = ACTIONS(3976),
    [anon_sym__Nullable_result] = ACTIONS(3976),
    [anon_sym__Null_unspecified] = ACTIONS(3976),
    [anon_sym___autoreleasing] = ACTIONS(3976),
    [anon_sym___block] = ACTIONS(3976),
    [anon_sym___bridge] = ACTIONS(3976),
    [anon_sym___bridge_retained] = ACTIONS(3976),
    [anon_sym___bridge_transfer] = ACTIONS(3976),
    [anon_sym___complex] = ACTIONS(3976),
    [anon_sym___const] = ACTIONS(3976),
    [anon_sym___imag] = ACTIONS(3976),
    [anon_sym___kindof] = ACTIONS(3976),
    [anon_sym___nonnull] = ACTIONS(3976),
    [anon_sym___nullable] = ACTIONS(3976),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3976),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3976),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3976),
    [anon_sym___real] = ACTIONS(3976),
    [anon_sym___strong] = ACTIONS(3976),
    [anon_sym___unsafe_unretained] = ACTIONS(3976),
    [anon_sym___unused] = ACTIONS(3976),
    [anon_sym___weak] = ACTIONS(3976),
    [anon_sym_alignas] = ACTIONS(3976),
    [anon_sym__Alignas] = ACTIONS(3976),
    [sym_primitive_type] = ACTIONS(3976),
    [anon_sym_enum] = ACTIONS(3976),
    [anon_sym_COLON] = ACTIONS(3978),
    [anon_sym_struct] = ACTIONS(3976),
    [anon_sym_union] = ACTIONS(3976),
    [anon_sym_QMARK] = ACTIONS(3978),
    [anon_sym_DASH_DASH] = ACTIONS(3978),
    [anon_sym_PLUS_PLUS] = ACTIONS(3978),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3976),
    [anon_sym___typeof] = ACTIONS(3976),
    [anon_sym_typeof] = ACTIONS(3976),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3976),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3976),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3976),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3976),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3976),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3976),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3976),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3976),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3976),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3976),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3976),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3976),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3976),
    [anon_sym_API_AVAILABLE] = ACTIONS(3976),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3976),
    [anon_sym_API_DEPRECATED] = ACTIONS(3976),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3976),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3976),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3976),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3976),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3976),
    [anon_sym___deprecated_msg] = ACTIONS(3976),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3976),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3976),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3976),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3976),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3976),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3976),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3976),
    [anon_sym_BOOL] = ACTIONS(3976),
    [anon_sym_IMP] = ACTIONS(3976),
    [anon_sym_SEL] = ACTIONS(3976),
    [anon_sym_Class] = ACTIONS(3976),
    [anon_sym_id] = ACTIONS(3976),
  },
  [STATE(1616)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4618),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_EQ] = ACTIONS(4511),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4566),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4515),
    [anon_sym_SLASH_EQ] = ACTIONS(4515),
    [anon_sym_PERCENT_EQ] = ACTIONS(4515),
    [anon_sym_PLUS_EQ] = ACTIONS(4515),
    [anon_sym_DASH_EQ] = ACTIONS(4515),
    [anon_sym_LT_LT_EQ] = ACTIONS(4515),
    [anon_sym_GT_GT_EQ] = ACTIONS(4515),
    [anon_sym_AMP_EQ] = ACTIONS(4515),
    [anon_sym_CARET_EQ] = ACTIONS(4515),
    [anon_sym_PIPE_EQ] = ACTIONS(4515),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(1617)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(3988),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3990),
    [anon_sym_COMMA] = ACTIONS(3990),
    [anon_sym_RPAREN] = ACTIONS(3990),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3990),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3990),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(3988),
    [anon_sym_PLUS] = ACTIONS(3988),
    [anon_sym_STAR] = ACTIONS(3990),
    [anon_sym_SLASH] = ACTIONS(3988),
    [anon_sym_PERCENT] = ACTIONS(3990),
    [anon_sym_PIPE_PIPE] = ACTIONS(3990),
    [anon_sym_AMP_AMP] = ACTIONS(3990),
    [anon_sym_PIPE] = ACTIONS(3988),
    [anon_sym_CARET] = ACTIONS(3990),
    [anon_sym_AMP] = ACTIONS(3988),
    [anon_sym_EQ_EQ] = ACTIONS(3990),
    [anon_sym_BANG_EQ] = ACTIONS(3990),
    [anon_sym_GT] = ACTIONS(3988),
    [anon_sym_GT_EQ] = ACTIONS(3990),
    [anon_sym_LT_EQ] = ACTIONS(3990),
    [anon_sym_LT] = ACTIONS(3988),
    [anon_sym_LT_LT] = ACTIONS(3990),
    [anon_sym_GT_GT] = ACTIONS(3990),
    [anon_sym_SEMI] = ACTIONS(3990),
    [anon_sym___extension__] = ACTIONS(3988),
    [anon_sym_extern] = ACTIONS(3988),
    [anon_sym___attribute__] = ACTIONS(3988),
    [anon_sym___attribute] = ACTIONS(3988),
    [anon_sym_noreturn] = ACTIONS(3988),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(3990),
    [anon_sym___declspec] = ACTIONS(3988),
    [anon_sym_LBRACE] = ACTIONS(3990),
    [anon_sym_RBRACE] = ACTIONS(3990),
    [anon_sym_signed] = ACTIONS(3988),
    [anon_sym_unsigned] = ACTIONS(3988),
    [anon_sym_long] = ACTIONS(3988),
    [anon_sym_short] = ACTIONS(3988),
    [anon_sym_static] = ACTIONS(3988),
    [anon_sym_ATautoreleasepool] = ACTIONS(3990),
    [anon_sym_auto] = ACTIONS(3988),
    [anon_sym_register] = ACTIONS(3988),
    [anon_sym_inline] = ACTIONS(3988),
    [anon_sym___inline] = ACTIONS(3988),
    [anon_sym___inline__] = ACTIONS(3988),
    [anon_sym___forceinline] = ACTIONS(3988),
    [anon_sym_thread_local] = ACTIONS(3988),
    [anon_sym___thread] = ACTIONS(3988),
    [anon_sym_CG_EXTERN] = ACTIONS(3988),
    [anon_sym_CG_INLINE] = ACTIONS(3988),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3988),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3988),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3988),
    [anon_sym_IBOutlet] = ACTIONS(3988),
    [anon_sym_IBInspectable] = ACTIONS(3988),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3988),
    [anon_sym_NS_INLINE] = ACTIONS(3988),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3988),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3988),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3988),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3988),
    [anon_sym_const] = ACTIONS(3988),
    [anon_sym_constexpr] = ACTIONS(3988),
    [anon_sym_volatile] = ACTIONS(3988),
    [anon_sym_restrict] = ACTIONS(3988),
    [anon_sym___restrict__] = ACTIONS(3988),
    [anon_sym__Atomic] = ACTIONS(3988),
    [anon_sym__Noreturn] = ACTIONS(3988),
    [anon_sym__Nonnull] = ACTIONS(3988),
    [anon_sym_nullable] = ACTIONS(3988),
    [anon_sym__Complex] = ACTIONS(3988),
    [anon_sym__Nullable] = ACTIONS(3988),
    [anon_sym__Nullable_result] = ACTIONS(3988),
    [anon_sym__Null_unspecified] = ACTIONS(3988),
    [anon_sym___autoreleasing] = ACTIONS(3988),
    [anon_sym___block] = ACTIONS(3988),
    [anon_sym___bridge] = ACTIONS(3988),
    [anon_sym___bridge_retained] = ACTIONS(3988),
    [anon_sym___bridge_transfer] = ACTIONS(3988),
    [anon_sym___complex] = ACTIONS(3988),
    [anon_sym___const] = ACTIONS(3988),
    [anon_sym___imag] = ACTIONS(3988),
    [anon_sym___kindof] = ACTIONS(3988),
    [anon_sym___nonnull] = ACTIONS(3988),
    [anon_sym___nullable] = ACTIONS(3988),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3988),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3988),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3988),
    [anon_sym___real] = ACTIONS(3988),
    [anon_sym___strong] = ACTIONS(3988),
    [anon_sym___unsafe_unretained] = ACTIONS(3988),
    [anon_sym___unused] = ACTIONS(3988),
    [anon_sym___weak] = ACTIONS(3988),
    [anon_sym_alignas] = ACTIONS(3988),
    [anon_sym__Alignas] = ACTIONS(3988),
    [sym_primitive_type] = ACTIONS(3988),
    [anon_sym_enum] = ACTIONS(3988),
    [anon_sym_COLON] = ACTIONS(3990),
    [anon_sym_struct] = ACTIONS(3988),
    [anon_sym_union] = ACTIONS(3988),
    [anon_sym_QMARK] = ACTIONS(3990),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3988),
    [anon_sym___typeof] = ACTIONS(3988),
    [anon_sym_typeof] = ACTIONS(3988),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3988),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3988),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3988),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3988),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3988),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3988),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3988),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3988),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3988),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3988),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3988),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3988),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3988),
    [anon_sym_API_AVAILABLE] = ACTIONS(3988),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3988),
    [anon_sym_API_DEPRECATED] = ACTIONS(3988),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3988),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3988),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3988),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3988),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3988),
    [anon_sym___deprecated_msg] = ACTIONS(3988),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3988),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3988),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3988),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3988),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3988),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3988),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3988),
    [anon_sym_BOOL] = ACTIONS(3988),
    [anon_sym_IMP] = ACTIONS(3988),
    [anon_sym_SEL] = ACTIONS(3988),
    [anon_sym_Class] = ACTIONS(3988),
    [anon_sym_id] = ACTIONS(3988),
  },
  [STATE(1618)] = {
    [sym_string_literal] = STATE(1621),
    [aux_sym_concatenated_string_repeat1] = STATE(1621),
    [sym_identifier] = ACTIONS(4623),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3944),
    [anon_sym_COMMA] = ACTIONS(3944),
    [anon_sym_RPAREN] = ACTIONS(3944),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3944),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3944),
    [anon_sym_LPAREN2] = ACTIONS(3944),
    [anon_sym_DASH] = ACTIONS(3946),
    [anon_sym_PLUS] = ACTIONS(3946),
    [anon_sym_STAR] = ACTIONS(3946),
    [anon_sym_SLASH] = ACTIONS(3946),
    [anon_sym_PERCENT] = ACTIONS(3946),
    [anon_sym_PIPE_PIPE] = ACTIONS(3944),
    [anon_sym_AMP_AMP] = ACTIONS(3944),
    [anon_sym_PIPE] = ACTIONS(3946),
    [anon_sym_CARET] = ACTIONS(3946),
    [anon_sym_AMP] = ACTIONS(3946),
    [anon_sym_EQ_EQ] = ACTIONS(3944),
    [anon_sym_BANG_EQ] = ACTIONS(3944),
    [anon_sym_GT] = ACTIONS(3946),
    [anon_sym_GT_EQ] = ACTIONS(3944),
    [anon_sym_LT_EQ] = ACTIONS(3944),
    [anon_sym_LT] = ACTIONS(3946),
    [anon_sym_LT_LT] = ACTIONS(3946),
    [anon_sym_GT_GT] = ACTIONS(3946),
    [anon_sym_SEMI] = ACTIONS(3944),
    [anon_sym___extension__] = ACTIONS(3946),
    [anon_sym_extern] = ACTIONS(3946),
    [anon_sym___attribute__] = ACTIONS(3946),
    [anon_sym___attribute] = ACTIONS(3946),
    [anon_sym_noreturn] = ACTIONS(3946),
    [anon_sym_LBRACK] = ACTIONS(3944),
    [anon_sym_RBRACK] = ACTIONS(3944),
    [anon_sym___declspec] = ACTIONS(3946),
    [anon_sym_RBRACE] = ACTIONS(3944),
    [anon_sym_static] = ACTIONS(3946),
    [anon_sym_EQ] = ACTIONS(3946),
    [anon_sym_auto] = ACTIONS(3946),
    [anon_sym_register] = ACTIONS(3946),
    [anon_sym_inline] = ACTIONS(3946),
    [anon_sym___inline] = ACTIONS(3946),
    [anon_sym___inline__] = ACTIONS(3946),
    [anon_sym___forceinline] = ACTIONS(3946),
    [anon_sym_thread_local] = ACTIONS(3946),
    [anon_sym___thread] = ACTIONS(3946),
    [anon_sym_CG_EXTERN] = ACTIONS(3946),
    [anon_sym_CG_INLINE] = ACTIONS(3946),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3946),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3946),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3946),
    [anon_sym_IBOutlet] = ACTIONS(3946),
    [anon_sym_IBInspectable] = ACTIONS(3946),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3946),
    [anon_sym_NS_INLINE] = ACTIONS(3946),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3946),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3946),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3946),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3946),
    [anon_sym_const] = ACTIONS(3946),
    [anon_sym_constexpr] = ACTIONS(3946),
    [anon_sym_volatile] = ACTIONS(3946),
    [anon_sym_restrict] = ACTIONS(3946),
    [anon_sym___restrict__] = ACTIONS(3946),
    [anon_sym__Atomic] = ACTIONS(3946),
    [anon_sym__Noreturn] = ACTIONS(3946),
    [anon_sym__Nonnull] = ACTIONS(3946),
    [anon_sym_nullable] = ACTIONS(3946),
    [anon_sym__Complex] = ACTIONS(3946),
    [anon_sym__Nullable] = ACTIONS(3946),
    [anon_sym__Nullable_result] = ACTIONS(3946),
    [anon_sym__Null_unspecified] = ACTIONS(3946),
    [anon_sym___autoreleasing] = ACTIONS(3946),
    [anon_sym___block] = ACTIONS(3946),
    [anon_sym___bridge] = ACTIONS(3946),
    [anon_sym___bridge_retained] = ACTIONS(3946),
    [anon_sym___bridge_transfer] = ACTIONS(3946),
    [anon_sym___complex] = ACTIONS(3946),
    [anon_sym___const] = ACTIONS(3946),
    [anon_sym___imag] = ACTIONS(3946),
    [anon_sym___kindof] = ACTIONS(3946),
    [anon_sym___nonnull] = ACTIONS(3946),
    [anon_sym___nullable] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3946),
    [anon_sym___real] = ACTIONS(3946),
    [anon_sym___strong] = ACTIONS(3946),
    [anon_sym___unsafe_unretained] = ACTIONS(3946),
    [anon_sym___unused] = ACTIONS(3946),
    [anon_sym___weak] = ACTIONS(3946),
    [anon_sym_alignas] = ACTIONS(3946),
    [anon_sym__Alignas] = ACTIONS(3946),
    [anon_sym_COLON] = ACTIONS(3944),
    [anon_sym_QMARK] = ACTIONS(3944),
    [anon_sym_STAR_EQ] = ACTIONS(3944),
    [anon_sym_SLASH_EQ] = ACTIONS(3944),
    [anon_sym_PERCENT_EQ] = ACTIONS(3944),
    [anon_sym_PLUS_EQ] = ACTIONS(3944),
    [anon_sym_DASH_EQ] = ACTIONS(3944),
    [anon_sym_LT_LT_EQ] = ACTIONS(3944),
    [anon_sym_GT_GT_EQ] = ACTIONS(3944),
    [anon_sym_AMP_EQ] = ACTIONS(3944),
    [anon_sym_CARET_EQ] = ACTIONS(3944),
    [anon_sym_PIPE_EQ] = ACTIONS(3944),
    [anon_sym_DASH_DASH] = ACTIONS(3944),
    [anon_sym_PLUS_PLUS] = ACTIONS(3944),
    [anon_sym_DOT] = ACTIONS(3946),
    [anon_sym_DASH_GT] = ACTIONS(3944),
    [anon_sym_AT] = ACTIONS(4625),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3946),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3946),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3946),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3946),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3946),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3946),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3946),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_API_AVAILABLE] = ACTIONS(3946),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_API_DEPRECATED] = ACTIONS(3946),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3946),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym___deprecated_msg] = ACTIONS(3946),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3946),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3946),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3946),
  },
  [STATE(1619)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4020),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4022),
    [anon_sym_COMMA] = ACTIONS(4022),
    [anon_sym_RPAREN] = ACTIONS(4022),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4022),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4022),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4020),
    [anon_sym_PLUS] = ACTIONS(4020),
    [anon_sym_STAR] = ACTIONS(4022),
    [anon_sym_SLASH] = ACTIONS(4020),
    [anon_sym_PERCENT] = ACTIONS(4022),
    [anon_sym_PIPE_PIPE] = ACTIONS(4022),
    [anon_sym_AMP_AMP] = ACTIONS(4022),
    [anon_sym_PIPE] = ACTIONS(4020),
    [anon_sym_CARET] = ACTIONS(4022),
    [anon_sym_AMP] = ACTIONS(4020),
    [anon_sym_EQ_EQ] = ACTIONS(4022),
    [anon_sym_BANG_EQ] = ACTIONS(4022),
    [anon_sym_GT] = ACTIONS(4020),
    [anon_sym_GT_EQ] = ACTIONS(4022),
    [anon_sym_LT_EQ] = ACTIONS(4022),
    [anon_sym_LT] = ACTIONS(4020),
    [anon_sym_LT_LT] = ACTIONS(4022),
    [anon_sym_GT_GT] = ACTIONS(4022),
    [anon_sym_SEMI] = ACTIONS(4022),
    [anon_sym___extension__] = ACTIONS(4020),
    [anon_sym_extern] = ACTIONS(4020),
    [anon_sym___attribute__] = ACTIONS(4020),
    [anon_sym___attribute] = ACTIONS(4020),
    [anon_sym_noreturn] = ACTIONS(4020),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(4022),
    [anon_sym___declspec] = ACTIONS(4020),
    [anon_sym_LBRACE] = ACTIONS(4022),
    [anon_sym_RBRACE] = ACTIONS(4022),
    [anon_sym_signed] = ACTIONS(4020),
    [anon_sym_unsigned] = ACTIONS(4020),
    [anon_sym_long] = ACTIONS(4020),
    [anon_sym_short] = ACTIONS(4020),
    [anon_sym_static] = ACTIONS(4020),
    [anon_sym_ATautoreleasepool] = ACTIONS(4022),
    [anon_sym_auto] = ACTIONS(4020),
    [anon_sym_register] = ACTIONS(4020),
    [anon_sym_inline] = ACTIONS(4020),
    [anon_sym___inline] = ACTIONS(4020),
    [anon_sym___inline__] = ACTIONS(4020),
    [anon_sym___forceinline] = ACTIONS(4020),
    [anon_sym_thread_local] = ACTIONS(4020),
    [anon_sym___thread] = ACTIONS(4020),
    [anon_sym_CG_EXTERN] = ACTIONS(4020),
    [anon_sym_CG_INLINE] = ACTIONS(4020),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4020),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4020),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4020),
    [anon_sym_IBOutlet] = ACTIONS(4020),
    [anon_sym_IBInspectable] = ACTIONS(4020),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4020),
    [anon_sym_NS_INLINE] = ACTIONS(4020),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4020),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4020),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4020),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4020),
    [anon_sym_const] = ACTIONS(4020),
    [anon_sym_constexpr] = ACTIONS(4020),
    [anon_sym_volatile] = ACTIONS(4020),
    [anon_sym_restrict] = ACTIONS(4020),
    [anon_sym___restrict__] = ACTIONS(4020),
    [anon_sym__Atomic] = ACTIONS(4020),
    [anon_sym__Noreturn] = ACTIONS(4020),
    [anon_sym__Nonnull] = ACTIONS(4020),
    [anon_sym_nullable] = ACTIONS(4020),
    [anon_sym__Complex] = ACTIONS(4020),
    [anon_sym__Nullable] = ACTIONS(4020),
    [anon_sym__Nullable_result] = ACTIONS(4020),
    [anon_sym__Null_unspecified] = ACTIONS(4020),
    [anon_sym___autoreleasing] = ACTIONS(4020),
    [anon_sym___block] = ACTIONS(4020),
    [anon_sym___bridge] = ACTIONS(4020),
    [anon_sym___bridge_retained] = ACTIONS(4020),
    [anon_sym___bridge_transfer] = ACTIONS(4020),
    [anon_sym___complex] = ACTIONS(4020),
    [anon_sym___const] = ACTIONS(4020),
    [anon_sym___imag] = ACTIONS(4020),
    [anon_sym___kindof] = ACTIONS(4020),
    [anon_sym___nonnull] = ACTIONS(4020),
    [anon_sym___nullable] = ACTIONS(4020),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4020),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4020),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4020),
    [anon_sym___real] = ACTIONS(4020),
    [anon_sym___strong] = ACTIONS(4020),
    [anon_sym___unsafe_unretained] = ACTIONS(4020),
    [anon_sym___unused] = ACTIONS(4020),
    [anon_sym___weak] = ACTIONS(4020),
    [anon_sym_alignas] = ACTIONS(4020),
    [anon_sym__Alignas] = ACTIONS(4020),
    [sym_primitive_type] = ACTIONS(4020),
    [anon_sym_enum] = ACTIONS(4020),
    [anon_sym_COLON] = ACTIONS(4022),
    [anon_sym_struct] = ACTIONS(4020),
    [anon_sym_union] = ACTIONS(4020),
    [anon_sym_QMARK] = ACTIONS(4022),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4020),
    [anon_sym___typeof] = ACTIONS(4020),
    [anon_sym_typeof] = ACTIONS(4020),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4020),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4020),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4020),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4020),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4020),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4020),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4020),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4020),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4020),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4020),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4020),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4020),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4020),
    [anon_sym_API_AVAILABLE] = ACTIONS(4020),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4020),
    [anon_sym_API_DEPRECATED] = ACTIONS(4020),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4020),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4020),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4020),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4020),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4020),
    [anon_sym___deprecated_msg] = ACTIONS(4020),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4020),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4020),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4020),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4020),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4020),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4020),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4020),
    [anon_sym_BOOL] = ACTIONS(4020),
    [anon_sym_IMP] = ACTIONS(4020),
    [anon_sym_SEL] = ACTIONS(4020),
    [anon_sym_Class] = ACTIONS(4020),
    [anon_sym_id] = ACTIONS(4020),
  },
  [STATE(1620)] = {
    [aux_sym_selector_expression_repeat2] = STATE(1620),
    [sym_identifier] = ACTIONS(3969),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3971),
    [anon_sym_COMMA] = ACTIONS(3971),
    [anon_sym_RPAREN] = ACTIONS(4627),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3971),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3971),
    [anon_sym_LPAREN2] = ACTIONS(3971),
    [anon_sym_DASH] = ACTIONS(3969),
    [anon_sym_PLUS] = ACTIONS(3969),
    [anon_sym_STAR] = ACTIONS(3971),
    [anon_sym_SLASH] = ACTIONS(3969),
    [anon_sym_PERCENT] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE] = ACTIONS(3971),
    [anon_sym_AMP_AMP] = ACTIONS(3971),
    [anon_sym_PIPE] = ACTIONS(3969),
    [anon_sym_CARET] = ACTIONS(3971),
    [anon_sym_AMP] = ACTIONS(3969),
    [anon_sym_EQ_EQ] = ACTIONS(3971),
    [anon_sym_BANG_EQ] = ACTIONS(3971),
    [anon_sym_GT] = ACTIONS(3969),
    [anon_sym_GT_EQ] = ACTIONS(3971),
    [anon_sym_LT_EQ] = ACTIONS(3971),
    [anon_sym_LT] = ACTIONS(3969),
    [anon_sym_LT_LT] = ACTIONS(3971),
    [anon_sym_GT_GT] = ACTIONS(3971),
    [anon_sym_SEMI] = ACTIONS(3971),
    [anon_sym___extension__] = ACTIONS(3969),
    [anon_sym_extern] = ACTIONS(3969),
    [anon_sym___attribute__] = ACTIONS(3969),
    [anon_sym___attribute] = ACTIONS(3969),
    [anon_sym_noreturn] = ACTIONS(3969),
    [anon_sym_LBRACK] = ACTIONS(3971),
    [anon_sym_RBRACK] = ACTIONS(3971),
    [anon_sym___declspec] = ACTIONS(3969),
    [anon_sym_LBRACE] = ACTIONS(3971),
    [anon_sym_RBRACE] = ACTIONS(3971),
    [anon_sym_signed] = ACTIONS(3969),
    [anon_sym_unsigned] = ACTIONS(3969),
    [anon_sym_long] = ACTIONS(3969),
    [anon_sym_short] = ACTIONS(3969),
    [anon_sym_static] = ACTIONS(3969),
    [anon_sym_ATautoreleasepool] = ACTIONS(3971),
    [anon_sym_auto] = ACTIONS(3969),
    [anon_sym_register] = ACTIONS(3969),
    [anon_sym_inline] = ACTIONS(3969),
    [anon_sym___inline] = ACTIONS(3969),
    [anon_sym___inline__] = ACTIONS(3969),
    [anon_sym___forceinline] = ACTIONS(3969),
    [anon_sym_thread_local] = ACTIONS(3969),
    [anon_sym___thread] = ACTIONS(3969),
    [anon_sym_CG_EXTERN] = ACTIONS(3969),
    [anon_sym_CG_INLINE] = ACTIONS(3969),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3969),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3969),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3969),
    [anon_sym_IBOutlet] = ACTIONS(3969),
    [anon_sym_IBInspectable] = ACTIONS(3969),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3969),
    [anon_sym_NS_INLINE] = ACTIONS(3969),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3969),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3969),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3969),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3969),
    [anon_sym_const] = ACTIONS(3969),
    [anon_sym_constexpr] = ACTIONS(3969),
    [anon_sym_volatile] = ACTIONS(3969),
    [anon_sym_restrict] = ACTIONS(3969),
    [anon_sym___restrict__] = ACTIONS(3969),
    [anon_sym__Atomic] = ACTIONS(3969),
    [anon_sym__Noreturn] = ACTIONS(3969),
    [anon_sym__Nonnull] = ACTIONS(3969),
    [anon_sym_nullable] = ACTIONS(3969),
    [anon_sym__Complex] = ACTIONS(3969),
    [anon_sym__Nullable] = ACTIONS(3969),
    [anon_sym__Nullable_result] = ACTIONS(3969),
    [anon_sym__Null_unspecified] = ACTIONS(3969),
    [anon_sym___autoreleasing] = ACTIONS(3969),
    [anon_sym___block] = ACTIONS(3969),
    [anon_sym___bridge] = ACTIONS(3969),
    [anon_sym___bridge_retained] = ACTIONS(3969),
    [anon_sym___bridge_transfer] = ACTIONS(3969),
    [anon_sym___complex] = ACTIONS(3969),
    [anon_sym___const] = ACTIONS(3969),
    [anon_sym___imag] = ACTIONS(3969),
    [anon_sym___kindof] = ACTIONS(3969),
    [anon_sym___nonnull] = ACTIONS(3969),
    [anon_sym___nullable] = ACTIONS(3969),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3969),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3969),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3969),
    [anon_sym___real] = ACTIONS(3969),
    [anon_sym___strong] = ACTIONS(3969),
    [anon_sym___unsafe_unretained] = ACTIONS(3969),
    [anon_sym___unused] = ACTIONS(3969),
    [anon_sym___weak] = ACTIONS(3969),
    [anon_sym_alignas] = ACTIONS(3969),
    [anon_sym__Alignas] = ACTIONS(3969),
    [sym_primitive_type] = ACTIONS(3969),
    [anon_sym_enum] = ACTIONS(3969),
    [anon_sym_COLON] = ACTIONS(3971),
    [anon_sym_struct] = ACTIONS(3969),
    [anon_sym_union] = ACTIONS(3969),
    [anon_sym_QMARK] = ACTIONS(3971),
    [anon_sym_DASH_DASH] = ACTIONS(3971),
    [anon_sym_PLUS_PLUS] = ACTIONS(3971),
    [anon_sym_DOT] = ACTIONS(3969),
    [anon_sym_DASH_GT] = ACTIONS(3971),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3969),
    [anon_sym___typeof] = ACTIONS(3969),
    [anon_sym_typeof] = ACTIONS(3969),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3969),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3969),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3969),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3969),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3969),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3969),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3969),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3969),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3969),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3969),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3969),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3969),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3969),
    [anon_sym_API_AVAILABLE] = ACTIONS(3969),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3969),
    [anon_sym_API_DEPRECATED] = ACTIONS(3969),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3969),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3969),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3969),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3969),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3969),
    [anon_sym___deprecated_msg] = ACTIONS(3969),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3969),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3969),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3969),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3969),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3969),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3969),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3969),
    [anon_sym_BOOL] = ACTIONS(3969),
    [anon_sym_IMP] = ACTIONS(3969),
    [anon_sym_SEL] = ACTIONS(3969),
    [anon_sym_Class] = ACTIONS(3969),
    [anon_sym_id] = ACTIONS(3969),
  },
  [STATE(1621)] = {
    [sym_string_literal] = STATE(1621),
    [aux_sym_concatenated_string_repeat1] = STATE(1621),
    [sym_identifier] = ACTIONS(4630),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_RPAREN] = ACTIONS(3953),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3953),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3953),
    [anon_sym_LPAREN2] = ACTIONS(3953),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_STAR] = ACTIONS(3955),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_PERCENT] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3953),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3955),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_EQ_EQ] = ACTIONS(3953),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_LT_EQ] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3955),
    [anon_sym_GT_GT] = ACTIONS(3955),
    [anon_sym_SEMI] = ACTIONS(3953),
    [anon_sym___extension__] = ACTIONS(3955),
    [anon_sym_extern] = ACTIONS(3955),
    [anon_sym___attribute__] = ACTIONS(3955),
    [anon_sym___attribute] = ACTIONS(3955),
    [anon_sym_noreturn] = ACTIONS(3955),
    [anon_sym_LBRACK] = ACTIONS(3953),
    [anon_sym_RBRACK] = ACTIONS(3953),
    [anon_sym___declspec] = ACTIONS(3955),
    [anon_sym_RBRACE] = ACTIONS(3953),
    [anon_sym_static] = ACTIONS(3955),
    [anon_sym_EQ] = ACTIONS(3955),
    [anon_sym_auto] = ACTIONS(3955),
    [anon_sym_register] = ACTIONS(3955),
    [anon_sym_inline] = ACTIONS(3955),
    [anon_sym___inline] = ACTIONS(3955),
    [anon_sym___inline__] = ACTIONS(3955),
    [anon_sym___forceinline] = ACTIONS(3955),
    [anon_sym_thread_local] = ACTIONS(3955),
    [anon_sym___thread] = ACTIONS(3955),
    [anon_sym_CG_EXTERN] = ACTIONS(3955),
    [anon_sym_CG_INLINE] = ACTIONS(3955),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3955),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3955),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3955),
    [anon_sym_IBOutlet] = ACTIONS(3955),
    [anon_sym_IBInspectable] = ACTIONS(3955),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3955),
    [anon_sym_NS_INLINE] = ACTIONS(3955),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3955),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3955),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3955),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3955),
    [anon_sym_const] = ACTIONS(3955),
    [anon_sym_constexpr] = ACTIONS(3955),
    [anon_sym_volatile] = ACTIONS(3955),
    [anon_sym_restrict] = ACTIONS(3955),
    [anon_sym___restrict__] = ACTIONS(3955),
    [anon_sym__Atomic] = ACTIONS(3955),
    [anon_sym__Noreturn] = ACTIONS(3955),
    [anon_sym__Nonnull] = ACTIONS(3955),
    [anon_sym_nullable] = ACTIONS(3955),
    [anon_sym__Complex] = ACTIONS(3955),
    [anon_sym__Nullable] = ACTIONS(3955),
    [anon_sym__Nullable_result] = ACTIONS(3955),
    [anon_sym__Null_unspecified] = ACTIONS(3955),
    [anon_sym___autoreleasing] = ACTIONS(3955),
    [anon_sym___block] = ACTIONS(3955),
    [anon_sym___bridge] = ACTIONS(3955),
    [anon_sym___bridge_retained] = ACTIONS(3955),
    [anon_sym___bridge_transfer] = ACTIONS(3955),
    [anon_sym___complex] = ACTIONS(3955),
    [anon_sym___const] = ACTIONS(3955),
    [anon_sym___imag] = ACTIONS(3955),
    [anon_sym___kindof] = ACTIONS(3955),
    [anon_sym___nonnull] = ACTIONS(3955),
    [anon_sym___nullable] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3955),
    [anon_sym___real] = ACTIONS(3955),
    [anon_sym___strong] = ACTIONS(3955),
    [anon_sym___unsafe_unretained] = ACTIONS(3955),
    [anon_sym___unused] = ACTIONS(3955),
    [anon_sym___weak] = ACTIONS(3955),
    [anon_sym_alignas] = ACTIONS(3955),
    [anon_sym__Alignas] = ACTIONS(3955),
    [anon_sym_COLON] = ACTIONS(3953),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_STAR_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_PERCENT_EQ] = ACTIONS(3953),
    [anon_sym_PLUS_EQ] = ACTIONS(3953),
    [anon_sym_DASH_EQ] = ACTIONS(3953),
    [anon_sym_LT_LT_EQ] = ACTIONS(3953),
    [anon_sym_GT_GT_EQ] = ACTIONS(3953),
    [anon_sym_AMP_EQ] = ACTIONS(3953),
    [anon_sym_CARET_EQ] = ACTIONS(3953),
    [anon_sym_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_AT] = ACTIONS(4633),
    [anon_sym_DQUOTE] = ACTIONS(4636),
    [anon_sym_L_DQUOTE] = ACTIONS(4636),
    [anon_sym_u_DQUOTE] = ACTIONS(4636),
    [anon_sym_U_DQUOTE] = ACTIONS(4636),
    [anon_sym_u8_DQUOTE] = ACTIONS(4636),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3955),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3955),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3955),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3955),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3955),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3955),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3955),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_API_AVAILABLE] = ACTIONS(3955),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_API_DEPRECATED] = ACTIONS(3955),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3955),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym___deprecated_msg] = ACTIONS(3955),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3955),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3955),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3955),
  },
  [STATE(1622)] = {
    [aux_sym_selector_expression_repeat2] = STATE(1620),
    [sym_identifier] = ACTIONS(3994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3996),
    [anon_sym_COMMA] = ACTIONS(3996),
    [anon_sym_RPAREN] = ACTIONS(4639),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3996),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3996),
    [anon_sym_LPAREN2] = ACTIONS(3996),
    [anon_sym_DASH] = ACTIONS(3994),
    [anon_sym_PLUS] = ACTIONS(3994),
    [anon_sym_STAR] = ACTIONS(3996),
    [anon_sym_SLASH] = ACTIONS(3994),
    [anon_sym_PERCENT] = ACTIONS(3996),
    [anon_sym_PIPE_PIPE] = ACTIONS(3996),
    [anon_sym_AMP_AMP] = ACTIONS(3996),
    [anon_sym_PIPE] = ACTIONS(3994),
    [anon_sym_CARET] = ACTIONS(3996),
    [anon_sym_AMP] = ACTIONS(3994),
    [anon_sym_EQ_EQ] = ACTIONS(3996),
    [anon_sym_BANG_EQ] = ACTIONS(3996),
    [anon_sym_GT] = ACTIONS(3994),
    [anon_sym_GT_EQ] = ACTIONS(3996),
    [anon_sym_LT_EQ] = ACTIONS(3996),
    [anon_sym_LT] = ACTIONS(3994),
    [anon_sym_LT_LT] = ACTIONS(3996),
    [anon_sym_GT_GT] = ACTIONS(3996),
    [anon_sym_SEMI] = ACTIONS(3996),
    [anon_sym___extension__] = ACTIONS(3994),
    [anon_sym_extern] = ACTIONS(3994),
    [anon_sym___attribute__] = ACTIONS(3994),
    [anon_sym___attribute] = ACTIONS(3994),
    [anon_sym_noreturn] = ACTIONS(3994),
    [anon_sym_LBRACK] = ACTIONS(3996),
    [anon_sym_RBRACK] = ACTIONS(3996),
    [anon_sym___declspec] = ACTIONS(3994),
    [anon_sym_LBRACE] = ACTIONS(3996),
    [anon_sym_RBRACE] = ACTIONS(3996),
    [anon_sym_signed] = ACTIONS(3994),
    [anon_sym_unsigned] = ACTIONS(3994),
    [anon_sym_long] = ACTIONS(3994),
    [anon_sym_short] = ACTIONS(3994),
    [anon_sym_static] = ACTIONS(3994),
    [anon_sym_ATautoreleasepool] = ACTIONS(3996),
    [anon_sym_auto] = ACTIONS(3994),
    [anon_sym_register] = ACTIONS(3994),
    [anon_sym_inline] = ACTIONS(3994),
    [anon_sym___inline] = ACTIONS(3994),
    [anon_sym___inline__] = ACTIONS(3994),
    [anon_sym___forceinline] = ACTIONS(3994),
    [anon_sym_thread_local] = ACTIONS(3994),
    [anon_sym___thread] = ACTIONS(3994),
    [anon_sym_CG_EXTERN] = ACTIONS(3994),
    [anon_sym_CG_INLINE] = ACTIONS(3994),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3994),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3994),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3994),
    [anon_sym_IBOutlet] = ACTIONS(3994),
    [anon_sym_IBInspectable] = ACTIONS(3994),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3994),
    [anon_sym_NS_INLINE] = ACTIONS(3994),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3994),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3994),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3994),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3994),
    [anon_sym_const] = ACTIONS(3994),
    [anon_sym_constexpr] = ACTIONS(3994),
    [anon_sym_volatile] = ACTIONS(3994),
    [anon_sym_restrict] = ACTIONS(3994),
    [anon_sym___restrict__] = ACTIONS(3994),
    [anon_sym__Atomic] = ACTIONS(3994),
    [anon_sym__Noreturn] = ACTIONS(3994),
    [anon_sym__Nonnull] = ACTIONS(3994),
    [anon_sym_nullable] = ACTIONS(3994),
    [anon_sym__Complex] = ACTIONS(3994),
    [anon_sym__Nullable] = ACTIONS(3994),
    [anon_sym__Nullable_result] = ACTIONS(3994),
    [anon_sym__Null_unspecified] = ACTIONS(3994),
    [anon_sym___autoreleasing] = ACTIONS(3994),
    [anon_sym___block] = ACTIONS(3994),
    [anon_sym___bridge] = ACTIONS(3994),
    [anon_sym___bridge_retained] = ACTIONS(3994),
    [anon_sym___bridge_transfer] = ACTIONS(3994),
    [anon_sym___complex] = ACTIONS(3994),
    [anon_sym___const] = ACTIONS(3994),
    [anon_sym___imag] = ACTIONS(3994),
    [anon_sym___kindof] = ACTIONS(3994),
    [anon_sym___nonnull] = ACTIONS(3994),
    [anon_sym___nullable] = ACTIONS(3994),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3994),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3994),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3994),
    [anon_sym___real] = ACTIONS(3994),
    [anon_sym___strong] = ACTIONS(3994),
    [anon_sym___unsafe_unretained] = ACTIONS(3994),
    [anon_sym___unused] = ACTIONS(3994),
    [anon_sym___weak] = ACTIONS(3994),
    [anon_sym_alignas] = ACTIONS(3994),
    [anon_sym__Alignas] = ACTIONS(3994),
    [sym_primitive_type] = ACTIONS(3994),
    [anon_sym_enum] = ACTIONS(3994),
    [anon_sym_COLON] = ACTIONS(3996),
    [anon_sym_struct] = ACTIONS(3994),
    [anon_sym_union] = ACTIONS(3994),
    [anon_sym_QMARK] = ACTIONS(3996),
    [anon_sym_DASH_DASH] = ACTIONS(3996),
    [anon_sym_PLUS_PLUS] = ACTIONS(3996),
    [anon_sym_DOT] = ACTIONS(3994),
    [anon_sym_DASH_GT] = ACTIONS(3996),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3994),
    [anon_sym___typeof] = ACTIONS(3994),
    [anon_sym_typeof] = ACTIONS(3994),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3994),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3994),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3994),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3994),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3994),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3994),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3994),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3994),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3994),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3994),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3994),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3994),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3994),
    [anon_sym_API_AVAILABLE] = ACTIONS(3994),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3994),
    [anon_sym_API_DEPRECATED] = ACTIONS(3994),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3994),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3994),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3994),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3994),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3994),
    [anon_sym___deprecated_msg] = ACTIONS(3994),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3994),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3994),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3994),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3994),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3994),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3994),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3994),
    [anon_sym_BOOL] = ACTIONS(3994),
    [anon_sym_IMP] = ACTIONS(3994),
    [anon_sym_SEL] = ACTIONS(3994),
    [anon_sym_Class] = ACTIONS(3994),
    [anon_sym_id] = ACTIONS(3994),
  },
  [STATE(1623)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4008),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4010),
    [anon_sym_COMMA] = ACTIONS(4010),
    [anon_sym_RPAREN] = ACTIONS(4010),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4010),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4010),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4008),
    [anon_sym_PLUS] = ACTIONS(4008),
    [anon_sym_STAR] = ACTIONS(4010),
    [anon_sym_SLASH] = ACTIONS(4008),
    [anon_sym_PERCENT] = ACTIONS(4010),
    [anon_sym_PIPE_PIPE] = ACTIONS(4010),
    [anon_sym_AMP_AMP] = ACTIONS(4010),
    [anon_sym_PIPE] = ACTIONS(4008),
    [anon_sym_CARET] = ACTIONS(4010),
    [anon_sym_AMP] = ACTIONS(4008),
    [anon_sym_EQ_EQ] = ACTIONS(4010),
    [anon_sym_BANG_EQ] = ACTIONS(4010),
    [anon_sym_GT] = ACTIONS(4008),
    [anon_sym_GT_EQ] = ACTIONS(4010),
    [anon_sym_LT_EQ] = ACTIONS(4010),
    [anon_sym_LT] = ACTIONS(4008),
    [anon_sym_LT_LT] = ACTIONS(4010),
    [anon_sym_GT_GT] = ACTIONS(4010),
    [anon_sym_SEMI] = ACTIONS(4010),
    [anon_sym___extension__] = ACTIONS(4008),
    [anon_sym_extern] = ACTIONS(4008),
    [anon_sym___attribute__] = ACTIONS(4008),
    [anon_sym___attribute] = ACTIONS(4008),
    [anon_sym_noreturn] = ACTIONS(4008),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(4010),
    [anon_sym___declspec] = ACTIONS(4008),
    [anon_sym_LBRACE] = ACTIONS(4010),
    [anon_sym_RBRACE] = ACTIONS(4010),
    [anon_sym_signed] = ACTIONS(4008),
    [anon_sym_unsigned] = ACTIONS(4008),
    [anon_sym_long] = ACTIONS(4008),
    [anon_sym_short] = ACTIONS(4008),
    [anon_sym_static] = ACTIONS(4008),
    [anon_sym_ATautoreleasepool] = ACTIONS(4010),
    [anon_sym_auto] = ACTIONS(4008),
    [anon_sym_register] = ACTIONS(4008),
    [anon_sym_inline] = ACTIONS(4008),
    [anon_sym___inline] = ACTIONS(4008),
    [anon_sym___inline__] = ACTIONS(4008),
    [anon_sym___forceinline] = ACTIONS(4008),
    [anon_sym_thread_local] = ACTIONS(4008),
    [anon_sym___thread] = ACTIONS(4008),
    [anon_sym_CG_EXTERN] = ACTIONS(4008),
    [anon_sym_CG_INLINE] = ACTIONS(4008),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4008),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4008),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4008),
    [anon_sym_IBOutlet] = ACTIONS(4008),
    [anon_sym_IBInspectable] = ACTIONS(4008),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4008),
    [anon_sym_NS_INLINE] = ACTIONS(4008),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4008),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4008),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4008),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4008),
    [anon_sym_const] = ACTIONS(4008),
    [anon_sym_constexpr] = ACTIONS(4008),
    [anon_sym_volatile] = ACTIONS(4008),
    [anon_sym_restrict] = ACTIONS(4008),
    [anon_sym___restrict__] = ACTIONS(4008),
    [anon_sym__Atomic] = ACTIONS(4008),
    [anon_sym__Noreturn] = ACTIONS(4008),
    [anon_sym__Nonnull] = ACTIONS(4008),
    [anon_sym_nullable] = ACTIONS(4008),
    [anon_sym__Complex] = ACTIONS(4008),
    [anon_sym__Nullable] = ACTIONS(4008),
    [anon_sym__Nullable_result] = ACTIONS(4008),
    [anon_sym__Null_unspecified] = ACTIONS(4008),
    [anon_sym___autoreleasing] = ACTIONS(4008),
    [anon_sym___block] = ACTIONS(4008),
    [anon_sym___bridge] = ACTIONS(4008),
    [anon_sym___bridge_retained] = ACTIONS(4008),
    [anon_sym___bridge_transfer] = ACTIONS(4008),
    [anon_sym___complex] = ACTIONS(4008),
    [anon_sym___const] = ACTIONS(4008),
    [anon_sym___imag] = ACTIONS(4008),
    [anon_sym___kindof] = ACTIONS(4008),
    [anon_sym___nonnull] = ACTIONS(4008),
    [anon_sym___nullable] = ACTIONS(4008),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4008),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4008),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4008),
    [anon_sym___real] = ACTIONS(4008),
    [anon_sym___strong] = ACTIONS(4008),
    [anon_sym___unsafe_unretained] = ACTIONS(4008),
    [anon_sym___unused] = ACTIONS(4008),
    [anon_sym___weak] = ACTIONS(4008),
    [anon_sym_alignas] = ACTIONS(4008),
    [anon_sym__Alignas] = ACTIONS(4008),
    [sym_primitive_type] = ACTIONS(4008),
    [anon_sym_enum] = ACTIONS(4008),
    [anon_sym_COLON] = ACTIONS(4010),
    [anon_sym_struct] = ACTIONS(4008),
    [anon_sym_union] = ACTIONS(4008),
    [anon_sym_QMARK] = ACTIONS(4010),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4008),
    [anon_sym___typeof] = ACTIONS(4008),
    [anon_sym_typeof] = ACTIONS(4008),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4008),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4008),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4008),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4008),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4008),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4008),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4008),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4008),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4008),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4008),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4008),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4008),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4008),
    [anon_sym_API_AVAILABLE] = ACTIONS(4008),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4008),
    [anon_sym_API_DEPRECATED] = ACTIONS(4008),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4008),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4008),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4008),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4008),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4008),
    [anon_sym___deprecated_msg] = ACTIONS(4008),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4008),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4008),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4008),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4008),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4008),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4008),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4008),
    [anon_sym_BOOL] = ACTIONS(4008),
    [anon_sym_IMP] = ACTIONS(4008),
    [anon_sym_SEL] = ACTIONS(4008),
    [anon_sym_Class] = ACTIONS(4008),
    [anon_sym_id] = ACTIONS(4008),
  },
  [STATE(1624)] = {
    [sym_string_literal] = STATE(1601),
    [sym_identifier] = ACTIONS(4004),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4006),
    [anon_sym_COMMA] = ACTIONS(4006),
    [anon_sym_LPAREN2] = ACTIONS(4006),
    [anon_sym_DASH] = ACTIONS(4004),
    [anon_sym_PLUS] = ACTIONS(4004),
    [anon_sym_STAR] = ACTIONS(4006),
    [anon_sym_SLASH] = ACTIONS(4004),
    [anon_sym_PERCENT] = ACTIONS(4006),
    [anon_sym_PIPE_PIPE] = ACTIONS(4006),
    [anon_sym_AMP_AMP] = ACTIONS(4006),
    [anon_sym_PIPE] = ACTIONS(4004),
    [anon_sym_CARET] = ACTIONS(4006),
    [anon_sym_AMP] = ACTIONS(4004),
    [anon_sym_EQ_EQ] = ACTIONS(4006),
    [anon_sym_BANG_EQ] = ACTIONS(4006),
    [anon_sym_GT] = ACTIONS(4004),
    [anon_sym_GT_EQ] = ACTIONS(4006),
    [anon_sym_LT_EQ] = ACTIONS(4006),
    [anon_sym_LT] = ACTIONS(4004),
    [anon_sym_LT_LT] = ACTIONS(4006),
    [anon_sym_GT_GT] = ACTIONS(4006),
    [anon_sym_SEMI] = ACTIONS(4006),
    [anon_sym___extension__] = ACTIONS(4004),
    [anon_sym_extern] = ACTIONS(4004),
    [anon_sym___attribute__] = ACTIONS(4004),
    [anon_sym___attribute] = ACTIONS(4004),
    [anon_sym_noreturn] = ACTIONS(4004),
    [anon_sym_LBRACK] = ACTIONS(4006),
    [anon_sym___declspec] = ACTIONS(4004),
    [anon_sym_LBRACE] = ACTIONS(4006),
    [anon_sym_signed] = ACTIONS(4004),
    [anon_sym_unsigned] = ACTIONS(4004),
    [anon_sym_long] = ACTIONS(4004),
    [anon_sym_short] = ACTIONS(4004),
    [anon_sym_static] = ACTIONS(4004),
    [anon_sym_ATautoreleasepool] = ACTIONS(4006),
    [anon_sym_auto] = ACTIONS(4004),
    [anon_sym_register] = ACTIONS(4004),
    [anon_sym_inline] = ACTIONS(4004),
    [anon_sym___inline] = ACTIONS(4004),
    [anon_sym___inline__] = ACTIONS(4004),
    [anon_sym___forceinline] = ACTIONS(4004),
    [anon_sym_thread_local] = ACTIONS(4004),
    [anon_sym___thread] = ACTIONS(4004),
    [anon_sym_CG_EXTERN] = ACTIONS(4004),
    [anon_sym_CG_INLINE] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4004),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4004),
    [anon_sym_IBOutlet] = ACTIONS(4004),
    [anon_sym_IBInspectable] = ACTIONS(4004),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4004),
    [anon_sym_NS_INLINE] = ACTIONS(4004),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4004),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4004),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4004),
    [anon_sym_const] = ACTIONS(4004),
    [anon_sym_constexpr] = ACTIONS(4004),
    [anon_sym_volatile] = ACTIONS(4004),
    [anon_sym_restrict] = ACTIONS(4004),
    [anon_sym___restrict__] = ACTIONS(4004),
    [anon_sym__Atomic] = ACTIONS(4004),
    [anon_sym__Noreturn] = ACTIONS(4004),
    [anon_sym__Nonnull] = ACTIONS(4004),
    [anon_sym_nullable] = ACTIONS(4004),
    [anon_sym__Complex] = ACTIONS(4004),
    [anon_sym__Nullable] = ACTIONS(4004),
    [anon_sym__Nullable_result] = ACTIONS(4004),
    [anon_sym__Null_unspecified] = ACTIONS(4004),
    [anon_sym___autoreleasing] = ACTIONS(4004),
    [anon_sym___block] = ACTIONS(4004),
    [anon_sym___bridge] = ACTIONS(4004),
    [anon_sym___bridge_retained] = ACTIONS(4004),
    [anon_sym___bridge_transfer] = ACTIONS(4004),
    [anon_sym___complex] = ACTIONS(4004),
    [anon_sym___const] = ACTIONS(4004),
    [anon_sym___imag] = ACTIONS(4004),
    [anon_sym___kindof] = ACTIONS(4004),
    [anon_sym___nonnull] = ACTIONS(4004),
    [anon_sym___nullable] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4004),
    [anon_sym___real] = ACTIONS(4004),
    [anon_sym___strong] = ACTIONS(4004),
    [anon_sym___unsafe_unretained] = ACTIONS(4004),
    [anon_sym___unused] = ACTIONS(4004),
    [anon_sym___weak] = ACTIONS(4004),
    [anon_sym_alignas] = ACTIONS(4004),
    [anon_sym__Alignas] = ACTIONS(4004),
    [sym_primitive_type] = ACTIONS(4004),
    [anon_sym_enum] = ACTIONS(4004),
    [anon_sym_struct] = ACTIONS(4004),
    [anon_sym_union] = ACTIONS(4004),
    [anon_sym_QMARK] = ACTIONS(4006),
    [anon_sym_DASH_DASH] = ACTIONS(4006),
    [anon_sym_PLUS_PLUS] = ACTIONS(4006),
    [anon_sym_DOT] = ACTIONS(4004),
    [anon_sym_DASH_GT] = ACTIONS(4006),
    [anon_sym_AT] = ACTIONS(4032),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4004),
    [anon_sym___typeof] = ACTIONS(4004),
    [anon_sym_typeof] = ACTIONS(4004),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4004),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4004),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4004),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_API_AVAILABLE] = ACTIONS(4004),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_API_DEPRECATED] = ACTIONS(4004),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym___deprecated_msg] = ACTIONS(4004),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4004),
    [anon_sym_BOOL] = ACTIONS(4004),
    [anon_sym_IMP] = ACTIONS(4004),
    [anon_sym_SEL] = ACTIONS(4004),
    [anon_sym_Class] = ACTIONS(4004),
    [anon_sym_id] = ACTIONS(4004),
  },
  [STATE(1625)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4012),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4014),
    [anon_sym_COMMA] = ACTIONS(4014),
    [anon_sym_RPAREN] = ACTIONS(4014),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4014),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4014),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4012),
    [anon_sym_PLUS] = ACTIONS(4012),
    [anon_sym_STAR] = ACTIONS(4014),
    [anon_sym_SLASH] = ACTIONS(4012),
    [anon_sym_PERCENT] = ACTIONS(4014),
    [anon_sym_PIPE_PIPE] = ACTIONS(4014),
    [anon_sym_AMP_AMP] = ACTIONS(4014),
    [anon_sym_PIPE] = ACTIONS(4012),
    [anon_sym_CARET] = ACTIONS(4014),
    [anon_sym_AMP] = ACTIONS(4012),
    [anon_sym_EQ_EQ] = ACTIONS(4014),
    [anon_sym_BANG_EQ] = ACTIONS(4014),
    [anon_sym_GT] = ACTIONS(4012),
    [anon_sym_GT_EQ] = ACTIONS(4014),
    [anon_sym_LT_EQ] = ACTIONS(4014),
    [anon_sym_LT] = ACTIONS(4012),
    [anon_sym_LT_LT] = ACTIONS(4014),
    [anon_sym_GT_GT] = ACTIONS(4014),
    [anon_sym_SEMI] = ACTIONS(4014),
    [anon_sym___extension__] = ACTIONS(4012),
    [anon_sym_extern] = ACTIONS(4012),
    [anon_sym___attribute__] = ACTIONS(4012),
    [anon_sym___attribute] = ACTIONS(4012),
    [anon_sym_noreturn] = ACTIONS(4012),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(4014),
    [anon_sym___declspec] = ACTIONS(4012),
    [anon_sym_LBRACE] = ACTIONS(4014),
    [anon_sym_RBRACE] = ACTIONS(4014),
    [anon_sym_signed] = ACTIONS(4012),
    [anon_sym_unsigned] = ACTIONS(4012),
    [anon_sym_long] = ACTIONS(4012),
    [anon_sym_short] = ACTIONS(4012),
    [anon_sym_static] = ACTIONS(4012),
    [anon_sym_ATautoreleasepool] = ACTIONS(4014),
    [anon_sym_auto] = ACTIONS(4012),
    [anon_sym_register] = ACTIONS(4012),
    [anon_sym_inline] = ACTIONS(4012),
    [anon_sym___inline] = ACTIONS(4012),
    [anon_sym___inline__] = ACTIONS(4012),
    [anon_sym___forceinline] = ACTIONS(4012),
    [anon_sym_thread_local] = ACTIONS(4012),
    [anon_sym___thread] = ACTIONS(4012),
    [anon_sym_CG_EXTERN] = ACTIONS(4012),
    [anon_sym_CG_INLINE] = ACTIONS(4012),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4012),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4012),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4012),
    [anon_sym_IBOutlet] = ACTIONS(4012),
    [anon_sym_IBInspectable] = ACTIONS(4012),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4012),
    [anon_sym_NS_INLINE] = ACTIONS(4012),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4012),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4012),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4012),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4012),
    [anon_sym_const] = ACTIONS(4012),
    [anon_sym_constexpr] = ACTIONS(4012),
    [anon_sym_volatile] = ACTIONS(4012),
    [anon_sym_restrict] = ACTIONS(4012),
    [anon_sym___restrict__] = ACTIONS(4012),
    [anon_sym__Atomic] = ACTIONS(4012),
    [anon_sym__Noreturn] = ACTIONS(4012),
    [anon_sym__Nonnull] = ACTIONS(4012),
    [anon_sym_nullable] = ACTIONS(4012),
    [anon_sym__Complex] = ACTIONS(4012),
    [anon_sym__Nullable] = ACTIONS(4012),
    [anon_sym__Nullable_result] = ACTIONS(4012),
    [anon_sym__Null_unspecified] = ACTIONS(4012),
    [anon_sym___autoreleasing] = ACTIONS(4012),
    [anon_sym___block] = ACTIONS(4012),
    [anon_sym___bridge] = ACTIONS(4012),
    [anon_sym___bridge_retained] = ACTIONS(4012),
    [anon_sym___bridge_transfer] = ACTIONS(4012),
    [anon_sym___complex] = ACTIONS(4012),
    [anon_sym___const] = ACTIONS(4012),
    [anon_sym___imag] = ACTIONS(4012),
    [anon_sym___kindof] = ACTIONS(4012),
    [anon_sym___nonnull] = ACTIONS(4012),
    [anon_sym___nullable] = ACTIONS(4012),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4012),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4012),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4012),
    [anon_sym___real] = ACTIONS(4012),
    [anon_sym___strong] = ACTIONS(4012),
    [anon_sym___unsafe_unretained] = ACTIONS(4012),
    [anon_sym___unused] = ACTIONS(4012),
    [anon_sym___weak] = ACTIONS(4012),
    [anon_sym_alignas] = ACTIONS(4012),
    [anon_sym__Alignas] = ACTIONS(4012),
    [sym_primitive_type] = ACTIONS(4012),
    [anon_sym_enum] = ACTIONS(4012),
    [anon_sym_COLON] = ACTIONS(4014),
    [anon_sym_struct] = ACTIONS(4012),
    [anon_sym_union] = ACTIONS(4012),
    [anon_sym_QMARK] = ACTIONS(4014),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4012),
    [anon_sym___typeof] = ACTIONS(4012),
    [anon_sym_typeof] = ACTIONS(4012),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4012),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4012),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4012),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4012),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4012),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4012),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4012),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4012),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4012),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4012),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4012),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4012),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4012),
    [anon_sym_API_AVAILABLE] = ACTIONS(4012),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4012),
    [anon_sym_API_DEPRECATED] = ACTIONS(4012),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4012),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4012),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4012),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4012),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4012),
    [anon_sym___deprecated_msg] = ACTIONS(4012),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4012),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4012),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4012),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4012),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4012),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4012),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4012),
    [anon_sym_BOOL] = ACTIONS(4012),
    [anon_sym_IMP] = ACTIONS(4012),
    [anon_sym_SEL] = ACTIONS(4012),
    [anon_sym_Class] = ACTIONS(4012),
    [anon_sym_id] = ACTIONS(4012),
  },
  [STATE(1626)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_RPAREN] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4000),
    [anon_sym_PLUS] = ACTIONS(4000),
    [anon_sym_STAR] = ACTIONS(4002),
    [anon_sym_SLASH] = ACTIONS(4000),
    [anon_sym_PERCENT] = ACTIONS(4002),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4002),
    [anon_sym_GT_GT] = ACTIONS(4002),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym_RBRACK] = ACTIONS(4002),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_COLON] = ACTIONS(4002),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1627)] = {
    [sym_string_literal] = STATE(1618),
    [aux_sym_concatenated_string_repeat1] = STATE(1618),
    [sym_identifier] = ACTIONS(4641),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_RPAREN] = ACTIONS(3965),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3965),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3965),
    [anon_sym_LPAREN2] = ACTIONS(3965),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_STAR] = ACTIONS(3967),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_PERCENT] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3965),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3967),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_EQ_EQ] = ACTIONS(3965),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_LT_EQ] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3967),
    [anon_sym_GT_GT] = ACTIONS(3967),
    [anon_sym_SEMI] = ACTIONS(3965),
    [anon_sym___extension__] = ACTIONS(3967),
    [anon_sym_extern] = ACTIONS(3967),
    [anon_sym___attribute__] = ACTIONS(3967),
    [anon_sym___attribute] = ACTIONS(3967),
    [anon_sym_noreturn] = ACTIONS(3967),
    [anon_sym_LBRACK] = ACTIONS(3965),
    [anon_sym_RBRACK] = ACTIONS(3965),
    [anon_sym___declspec] = ACTIONS(3967),
    [anon_sym_RBRACE] = ACTIONS(3965),
    [anon_sym_static] = ACTIONS(3967),
    [anon_sym_EQ] = ACTIONS(3967),
    [anon_sym_auto] = ACTIONS(3967),
    [anon_sym_register] = ACTIONS(3967),
    [anon_sym_inline] = ACTIONS(3967),
    [anon_sym___inline] = ACTIONS(3967),
    [anon_sym___inline__] = ACTIONS(3967),
    [anon_sym___forceinline] = ACTIONS(3967),
    [anon_sym_thread_local] = ACTIONS(3967),
    [anon_sym___thread] = ACTIONS(3967),
    [anon_sym_CG_EXTERN] = ACTIONS(3967),
    [anon_sym_CG_INLINE] = ACTIONS(3967),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3967),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3967),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3967),
    [anon_sym_IBOutlet] = ACTIONS(3967),
    [anon_sym_IBInspectable] = ACTIONS(3967),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3967),
    [anon_sym_NS_INLINE] = ACTIONS(3967),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3967),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3967),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3967),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3967),
    [anon_sym_const] = ACTIONS(3967),
    [anon_sym_constexpr] = ACTIONS(3967),
    [anon_sym_volatile] = ACTIONS(3967),
    [anon_sym_restrict] = ACTIONS(3967),
    [anon_sym___restrict__] = ACTIONS(3967),
    [anon_sym__Atomic] = ACTIONS(3967),
    [anon_sym__Noreturn] = ACTIONS(3967),
    [anon_sym__Nonnull] = ACTIONS(3967),
    [anon_sym_nullable] = ACTIONS(3967),
    [anon_sym__Complex] = ACTIONS(3967),
    [anon_sym__Nullable] = ACTIONS(3967),
    [anon_sym__Nullable_result] = ACTIONS(3967),
    [anon_sym__Null_unspecified] = ACTIONS(3967),
    [anon_sym___autoreleasing] = ACTIONS(3967),
    [anon_sym___block] = ACTIONS(3967),
    [anon_sym___bridge] = ACTIONS(3967),
    [anon_sym___bridge_retained] = ACTIONS(3967),
    [anon_sym___bridge_transfer] = ACTIONS(3967),
    [anon_sym___complex] = ACTIONS(3967),
    [anon_sym___const] = ACTIONS(3967),
    [anon_sym___imag] = ACTIONS(3967),
    [anon_sym___kindof] = ACTIONS(3967),
    [anon_sym___nonnull] = ACTIONS(3967),
    [anon_sym___nullable] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3967),
    [anon_sym___real] = ACTIONS(3967),
    [anon_sym___strong] = ACTIONS(3967),
    [anon_sym___unsafe_unretained] = ACTIONS(3967),
    [anon_sym___unused] = ACTIONS(3967),
    [anon_sym___weak] = ACTIONS(3967),
    [anon_sym_alignas] = ACTIONS(3967),
    [anon_sym__Alignas] = ACTIONS(3967),
    [anon_sym_COLON] = ACTIONS(3965),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_STAR_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_PERCENT_EQ] = ACTIONS(3965),
    [anon_sym_PLUS_EQ] = ACTIONS(3965),
    [anon_sym_DASH_EQ] = ACTIONS(3965),
    [anon_sym_LT_LT_EQ] = ACTIONS(3965),
    [anon_sym_GT_GT_EQ] = ACTIONS(3965),
    [anon_sym_AMP_EQ] = ACTIONS(3965),
    [anon_sym_CARET_EQ] = ACTIONS(3965),
    [anon_sym_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_AT] = ACTIONS(4625),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3967),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3967),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3967),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3967),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3967),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3967),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3967),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_API_AVAILABLE] = ACTIONS(3967),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_API_DEPRECATED] = ACTIONS(3967),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3967),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym___deprecated_msg] = ACTIONS(3967),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3967),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3967),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3967),
  },
  [STATE(1628)] = {
    [sym_identifier] = ACTIONS(4160),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4162),
    [anon_sym_COMMA] = ACTIONS(4162),
    [anon_sym_RPAREN] = ACTIONS(4162),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4162),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4162),
    [anon_sym_LPAREN2] = ACTIONS(4162),
    [anon_sym_DASH] = ACTIONS(4160),
    [anon_sym_PLUS] = ACTIONS(4160),
    [anon_sym_STAR] = ACTIONS(4162),
    [anon_sym_SLASH] = ACTIONS(4160),
    [anon_sym_PERCENT] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE] = ACTIONS(4162),
    [anon_sym_AMP_AMP] = ACTIONS(4162),
    [anon_sym_PIPE] = ACTIONS(4160),
    [anon_sym_CARET] = ACTIONS(4162),
    [anon_sym_AMP] = ACTIONS(4160),
    [anon_sym_EQ_EQ] = ACTIONS(4162),
    [anon_sym_BANG_EQ] = ACTIONS(4162),
    [anon_sym_GT] = ACTIONS(4160),
    [anon_sym_GT_EQ] = ACTIONS(4162),
    [anon_sym_LT_EQ] = ACTIONS(4162),
    [anon_sym_LT] = ACTIONS(4160),
    [anon_sym_LT_LT] = ACTIONS(4162),
    [anon_sym_GT_GT] = ACTIONS(4162),
    [anon_sym_SEMI] = ACTIONS(4162),
    [anon_sym___extension__] = ACTIONS(4160),
    [anon_sym_extern] = ACTIONS(4160),
    [anon_sym___attribute__] = ACTIONS(4160),
    [anon_sym___attribute] = ACTIONS(4160),
    [anon_sym_noreturn] = ACTIONS(4160),
    [anon_sym_LBRACK] = ACTIONS(4162),
    [anon_sym_RBRACK] = ACTIONS(4162),
    [anon_sym___declspec] = ACTIONS(4160),
    [anon_sym_LBRACE] = ACTIONS(4162),
    [anon_sym_RBRACE] = ACTIONS(4162),
    [anon_sym_signed] = ACTIONS(4160),
    [anon_sym_unsigned] = ACTIONS(4160),
    [anon_sym_long] = ACTIONS(4160),
    [anon_sym_short] = ACTIONS(4160),
    [anon_sym_static] = ACTIONS(4160),
    [anon_sym_ATautoreleasepool] = ACTIONS(4162),
    [anon_sym_auto] = ACTIONS(4160),
    [anon_sym_register] = ACTIONS(4160),
    [anon_sym_inline] = ACTIONS(4160),
    [anon_sym___inline] = ACTIONS(4160),
    [anon_sym___inline__] = ACTIONS(4160),
    [anon_sym___forceinline] = ACTIONS(4160),
    [anon_sym_thread_local] = ACTIONS(4160),
    [anon_sym___thread] = ACTIONS(4160),
    [anon_sym_CG_EXTERN] = ACTIONS(4160),
    [anon_sym_CG_INLINE] = ACTIONS(4160),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4160),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4160),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4160),
    [anon_sym_IBOutlet] = ACTIONS(4160),
    [anon_sym_IBInspectable] = ACTIONS(4160),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4160),
    [anon_sym_NS_INLINE] = ACTIONS(4160),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4160),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4160),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4160),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4160),
    [anon_sym_const] = ACTIONS(4160),
    [anon_sym_constexpr] = ACTIONS(4160),
    [anon_sym_volatile] = ACTIONS(4160),
    [anon_sym_restrict] = ACTIONS(4160),
    [anon_sym___restrict__] = ACTIONS(4160),
    [anon_sym__Atomic] = ACTIONS(4160),
    [anon_sym__Noreturn] = ACTIONS(4160),
    [anon_sym__Nonnull] = ACTIONS(4160),
    [anon_sym_nullable] = ACTIONS(4160),
    [anon_sym__Complex] = ACTIONS(4160),
    [anon_sym__Nullable] = ACTIONS(4160),
    [anon_sym__Nullable_result] = ACTIONS(4160),
    [anon_sym__Null_unspecified] = ACTIONS(4160),
    [anon_sym___autoreleasing] = ACTIONS(4160),
    [anon_sym___block] = ACTIONS(4160),
    [anon_sym___bridge] = ACTIONS(4160),
    [anon_sym___bridge_retained] = ACTIONS(4160),
    [anon_sym___bridge_transfer] = ACTIONS(4160),
    [anon_sym___complex] = ACTIONS(4160),
    [anon_sym___const] = ACTIONS(4160),
    [anon_sym___imag] = ACTIONS(4160),
    [anon_sym___kindof] = ACTIONS(4160),
    [anon_sym___nonnull] = ACTIONS(4160),
    [anon_sym___nullable] = ACTIONS(4160),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4160),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4160),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4160),
    [anon_sym___real] = ACTIONS(4160),
    [anon_sym___strong] = ACTIONS(4160),
    [anon_sym___unsafe_unretained] = ACTIONS(4160),
    [anon_sym___unused] = ACTIONS(4160),
    [anon_sym___weak] = ACTIONS(4160),
    [anon_sym_alignas] = ACTIONS(4160),
    [anon_sym__Alignas] = ACTIONS(4160),
    [sym_primitive_type] = ACTIONS(4160),
    [anon_sym_enum] = ACTIONS(4160),
    [anon_sym_COLON] = ACTIONS(4162),
    [anon_sym_struct] = ACTIONS(4160),
    [anon_sym_union] = ACTIONS(4160),
    [anon_sym_QMARK] = ACTIONS(4162),
    [anon_sym_DASH_DASH] = ACTIONS(4162),
    [anon_sym_PLUS_PLUS] = ACTIONS(4162),
    [anon_sym_DOT] = ACTIONS(4160),
    [anon_sym_DASH_GT] = ACTIONS(4162),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4160),
    [anon_sym___typeof] = ACTIONS(4160),
    [anon_sym_typeof] = ACTIONS(4160),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4160),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4160),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4160),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4160),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4160),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4160),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4160),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4160),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4160),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4160),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4160),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4160),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4160),
    [anon_sym_API_AVAILABLE] = ACTIONS(4160),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4160),
    [anon_sym_API_DEPRECATED] = ACTIONS(4160),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4160),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4160),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4160),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4160),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4160),
    [anon_sym___deprecated_msg] = ACTIONS(4160),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4160),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4160),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4160),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4160),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4160),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4160),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4160),
    [anon_sym_BOOL] = ACTIONS(4160),
    [anon_sym_IMP] = ACTIONS(4160),
    [anon_sym_SEL] = ACTIONS(4160),
    [anon_sym_Class] = ACTIONS(4160),
    [anon_sym_id] = ACTIONS(4160),
  },
  [STATE(1629)] = {
    [sym_identifier] = ACTIONS(4042),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_COMMA] = ACTIONS(4044),
    [anon_sym_RPAREN] = ACTIONS(4044),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4044),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4044),
    [anon_sym_LPAREN2] = ACTIONS(4044),
    [anon_sym_DASH] = ACTIONS(4042),
    [anon_sym_PLUS] = ACTIONS(4042),
    [anon_sym_STAR] = ACTIONS(4044),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4044),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(4044),
    [anon_sym_AMP] = ACTIONS(4042),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(4042),
    [anon_sym_LT_LT] = ACTIONS(4044),
    [anon_sym_GT_GT] = ACTIONS(4044),
    [anon_sym_SEMI] = ACTIONS(4044),
    [anon_sym___extension__] = ACTIONS(4042),
    [anon_sym_extern] = ACTIONS(4042),
    [anon_sym___attribute__] = ACTIONS(4042),
    [anon_sym___attribute] = ACTIONS(4042),
    [anon_sym_noreturn] = ACTIONS(4042),
    [anon_sym_LBRACK] = ACTIONS(4044),
    [anon_sym_RBRACK] = ACTIONS(4044),
    [anon_sym___declspec] = ACTIONS(4042),
    [anon_sym_LBRACE] = ACTIONS(4044),
    [anon_sym_RBRACE] = ACTIONS(4044),
    [anon_sym_signed] = ACTIONS(4042),
    [anon_sym_unsigned] = ACTIONS(4042),
    [anon_sym_long] = ACTIONS(4042),
    [anon_sym_short] = ACTIONS(4042),
    [anon_sym_static] = ACTIONS(4042),
    [anon_sym_ATautoreleasepool] = ACTIONS(4044),
    [anon_sym_auto] = ACTIONS(4042),
    [anon_sym_register] = ACTIONS(4042),
    [anon_sym_inline] = ACTIONS(4042),
    [anon_sym___inline] = ACTIONS(4042),
    [anon_sym___inline__] = ACTIONS(4042),
    [anon_sym___forceinline] = ACTIONS(4042),
    [anon_sym_thread_local] = ACTIONS(4042),
    [anon_sym___thread] = ACTIONS(4042),
    [anon_sym_CG_EXTERN] = ACTIONS(4042),
    [anon_sym_CG_INLINE] = ACTIONS(4042),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4042),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4042),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4042),
    [anon_sym_IBOutlet] = ACTIONS(4042),
    [anon_sym_IBInspectable] = ACTIONS(4042),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4042),
    [anon_sym_NS_INLINE] = ACTIONS(4042),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4042),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4042),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4042),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4042),
    [anon_sym_const] = ACTIONS(4042),
    [anon_sym_constexpr] = ACTIONS(4042),
    [anon_sym_volatile] = ACTIONS(4042),
    [anon_sym_restrict] = ACTIONS(4042),
    [anon_sym___restrict__] = ACTIONS(4042),
    [anon_sym__Atomic] = ACTIONS(4042),
    [anon_sym__Noreturn] = ACTIONS(4042),
    [anon_sym__Nonnull] = ACTIONS(4042),
    [anon_sym_nullable] = ACTIONS(4042),
    [anon_sym__Complex] = ACTIONS(4042),
    [anon_sym__Nullable] = ACTIONS(4042),
    [anon_sym__Nullable_result] = ACTIONS(4042),
    [anon_sym__Null_unspecified] = ACTIONS(4042),
    [anon_sym___autoreleasing] = ACTIONS(4042),
    [anon_sym___block] = ACTIONS(4042),
    [anon_sym___bridge] = ACTIONS(4042),
    [anon_sym___bridge_retained] = ACTIONS(4042),
    [anon_sym___bridge_transfer] = ACTIONS(4042),
    [anon_sym___complex] = ACTIONS(4042),
    [anon_sym___const] = ACTIONS(4042),
    [anon_sym___imag] = ACTIONS(4042),
    [anon_sym___kindof] = ACTIONS(4042),
    [anon_sym___nonnull] = ACTIONS(4042),
    [anon_sym___nullable] = ACTIONS(4042),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4042),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4042),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4042),
    [anon_sym___real] = ACTIONS(4042),
    [anon_sym___strong] = ACTIONS(4042),
    [anon_sym___unsafe_unretained] = ACTIONS(4042),
    [anon_sym___unused] = ACTIONS(4042),
    [anon_sym___weak] = ACTIONS(4042),
    [anon_sym_alignas] = ACTIONS(4042),
    [anon_sym__Alignas] = ACTIONS(4042),
    [sym_primitive_type] = ACTIONS(4042),
    [anon_sym_enum] = ACTIONS(4042),
    [anon_sym_COLON] = ACTIONS(4044),
    [anon_sym_struct] = ACTIONS(4042),
    [anon_sym_union] = ACTIONS(4042),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(4044),
    [anon_sym_PLUS_PLUS] = ACTIONS(4044),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4042),
    [anon_sym___typeof] = ACTIONS(4042),
    [anon_sym_typeof] = ACTIONS(4042),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4042),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4042),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4042),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4042),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4042),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4042),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4042),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4042),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4042),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4042),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4042),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4042),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4042),
    [anon_sym_API_AVAILABLE] = ACTIONS(4042),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4042),
    [anon_sym_API_DEPRECATED] = ACTIONS(4042),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4042),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4042),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4042),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4042),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4042),
    [anon_sym___deprecated_msg] = ACTIONS(4042),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4042),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4042),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4042),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4042),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4042),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4042),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4042),
    [anon_sym_BOOL] = ACTIONS(4042),
    [anon_sym_IMP] = ACTIONS(4042),
    [anon_sym_SEL] = ACTIONS(4042),
    [anon_sym_Class] = ACTIONS(4042),
    [anon_sym_id] = ACTIONS(4042),
  },
  [STATE(1630)] = {
    [sym_identifier] = ACTIONS(4138),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4140),
    [anon_sym_COMMA] = ACTIONS(4140),
    [anon_sym_RPAREN] = ACTIONS(4140),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4140),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4140),
    [anon_sym_LPAREN2] = ACTIONS(4140),
    [anon_sym_DASH] = ACTIONS(4138),
    [anon_sym_PLUS] = ACTIONS(4138),
    [anon_sym_STAR] = ACTIONS(4140),
    [anon_sym_SLASH] = ACTIONS(4138),
    [anon_sym_PERCENT] = ACTIONS(4140),
    [anon_sym_PIPE_PIPE] = ACTIONS(4140),
    [anon_sym_AMP_AMP] = ACTIONS(4140),
    [anon_sym_PIPE] = ACTIONS(4138),
    [anon_sym_CARET] = ACTIONS(4140),
    [anon_sym_AMP] = ACTIONS(4138),
    [anon_sym_EQ_EQ] = ACTIONS(4140),
    [anon_sym_BANG_EQ] = ACTIONS(4140),
    [anon_sym_GT] = ACTIONS(4138),
    [anon_sym_GT_EQ] = ACTIONS(4140),
    [anon_sym_LT_EQ] = ACTIONS(4140),
    [anon_sym_LT] = ACTIONS(4138),
    [anon_sym_LT_LT] = ACTIONS(4140),
    [anon_sym_GT_GT] = ACTIONS(4140),
    [anon_sym_SEMI] = ACTIONS(4140),
    [anon_sym___extension__] = ACTIONS(4138),
    [anon_sym_extern] = ACTIONS(4138),
    [anon_sym___attribute__] = ACTIONS(4138),
    [anon_sym___attribute] = ACTIONS(4138),
    [anon_sym_noreturn] = ACTIONS(4138),
    [anon_sym_LBRACK] = ACTIONS(4140),
    [anon_sym_RBRACK] = ACTIONS(4140),
    [anon_sym___declspec] = ACTIONS(4138),
    [anon_sym_LBRACE] = ACTIONS(4140),
    [anon_sym_RBRACE] = ACTIONS(4140),
    [anon_sym_signed] = ACTIONS(4138),
    [anon_sym_unsigned] = ACTIONS(4138),
    [anon_sym_long] = ACTIONS(4138),
    [anon_sym_short] = ACTIONS(4138),
    [anon_sym_static] = ACTIONS(4138),
    [anon_sym_ATautoreleasepool] = ACTIONS(4140),
    [anon_sym_auto] = ACTIONS(4138),
    [anon_sym_register] = ACTIONS(4138),
    [anon_sym_inline] = ACTIONS(4138),
    [anon_sym___inline] = ACTIONS(4138),
    [anon_sym___inline__] = ACTIONS(4138),
    [anon_sym___forceinline] = ACTIONS(4138),
    [anon_sym_thread_local] = ACTIONS(4138),
    [anon_sym___thread] = ACTIONS(4138),
    [anon_sym_CG_EXTERN] = ACTIONS(4138),
    [anon_sym_CG_INLINE] = ACTIONS(4138),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4138),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4138),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4138),
    [anon_sym_IBOutlet] = ACTIONS(4138),
    [anon_sym_IBInspectable] = ACTIONS(4138),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4138),
    [anon_sym_NS_INLINE] = ACTIONS(4138),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4138),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4138),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4138),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4138),
    [anon_sym_const] = ACTIONS(4138),
    [anon_sym_constexpr] = ACTIONS(4138),
    [anon_sym_volatile] = ACTIONS(4138),
    [anon_sym_restrict] = ACTIONS(4138),
    [anon_sym___restrict__] = ACTIONS(4138),
    [anon_sym__Atomic] = ACTIONS(4138),
    [anon_sym__Noreturn] = ACTIONS(4138),
    [anon_sym__Nonnull] = ACTIONS(4138),
    [anon_sym_nullable] = ACTIONS(4138),
    [anon_sym__Complex] = ACTIONS(4138),
    [anon_sym__Nullable] = ACTIONS(4138),
    [anon_sym__Nullable_result] = ACTIONS(4138),
    [anon_sym__Null_unspecified] = ACTIONS(4138),
    [anon_sym___autoreleasing] = ACTIONS(4138),
    [anon_sym___block] = ACTIONS(4138),
    [anon_sym___bridge] = ACTIONS(4138),
    [anon_sym___bridge_retained] = ACTIONS(4138),
    [anon_sym___bridge_transfer] = ACTIONS(4138),
    [anon_sym___complex] = ACTIONS(4138),
    [anon_sym___const] = ACTIONS(4138),
    [anon_sym___imag] = ACTIONS(4138),
    [anon_sym___kindof] = ACTIONS(4138),
    [anon_sym___nonnull] = ACTIONS(4138),
    [anon_sym___nullable] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4138),
    [anon_sym___real] = ACTIONS(4138),
    [anon_sym___strong] = ACTIONS(4138),
    [anon_sym___unsafe_unretained] = ACTIONS(4138),
    [anon_sym___unused] = ACTIONS(4138),
    [anon_sym___weak] = ACTIONS(4138),
    [anon_sym_alignas] = ACTIONS(4138),
    [anon_sym__Alignas] = ACTIONS(4138),
    [sym_primitive_type] = ACTIONS(4138),
    [anon_sym_enum] = ACTIONS(4138),
    [anon_sym_COLON] = ACTIONS(4140),
    [anon_sym_struct] = ACTIONS(4138),
    [anon_sym_union] = ACTIONS(4138),
    [anon_sym_QMARK] = ACTIONS(4140),
    [anon_sym_DASH_DASH] = ACTIONS(4140),
    [anon_sym_PLUS_PLUS] = ACTIONS(4140),
    [anon_sym_DOT] = ACTIONS(4138),
    [anon_sym_DASH_GT] = ACTIONS(4140),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4138),
    [anon_sym___typeof] = ACTIONS(4138),
    [anon_sym_typeof] = ACTIONS(4138),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4138),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4138),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4138),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4138),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4138),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4138),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4138),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_API_AVAILABLE] = ACTIONS(4138),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_API_DEPRECATED] = ACTIONS(4138),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4138),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym___deprecated_msg] = ACTIONS(4138),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4138),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4138),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4138),
    [anon_sym_BOOL] = ACTIONS(4138),
    [anon_sym_IMP] = ACTIONS(4138),
    [anon_sym_SEL] = ACTIONS(4138),
    [anon_sym_Class] = ACTIONS(4138),
    [anon_sym_id] = ACTIONS(4138),
  },
  [STATE(1631)] = {
    [sym_identifier] = ACTIONS(4134),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4136),
    [anon_sym_COMMA] = ACTIONS(4136),
    [anon_sym_RPAREN] = ACTIONS(4136),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4136),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4136),
    [anon_sym_LPAREN2] = ACTIONS(4136),
    [anon_sym_DASH] = ACTIONS(4134),
    [anon_sym_PLUS] = ACTIONS(4134),
    [anon_sym_STAR] = ACTIONS(4136),
    [anon_sym_SLASH] = ACTIONS(4134),
    [anon_sym_PERCENT] = ACTIONS(4136),
    [anon_sym_PIPE_PIPE] = ACTIONS(4136),
    [anon_sym_AMP_AMP] = ACTIONS(4136),
    [anon_sym_PIPE] = ACTIONS(4134),
    [anon_sym_CARET] = ACTIONS(4136),
    [anon_sym_AMP] = ACTIONS(4134),
    [anon_sym_EQ_EQ] = ACTIONS(4136),
    [anon_sym_BANG_EQ] = ACTIONS(4136),
    [anon_sym_GT] = ACTIONS(4134),
    [anon_sym_GT_EQ] = ACTIONS(4136),
    [anon_sym_LT_EQ] = ACTIONS(4136),
    [anon_sym_LT] = ACTIONS(4134),
    [anon_sym_LT_LT] = ACTIONS(4136),
    [anon_sym_GT_GT] = ACTIONS(4136),
    [anon_sym_SEMI] = ACTIONS(4136),
    [anon_sym___extension__] = ACTIONS(4134),
    [anon_sym_extern] = ACTIONS(4134),
    [anon_sym___attribute__] = ACTIONS(4134),
    [anon_sym___attribute] = ACTIONS(4134),
    [anon_sym_noreturn] = ACTIONS(4134),
    [anon_sym_LBRACK] = ACTIONS(4136),
    [anon_sym_RBRACK] = ACTIONS(4136),
    [anon_sym___declspec] = ACTIONS(4134),
    [anon_sym_LBRACE] = ACTIONS(4136),
    [anon_sym_RBRACE] = ACTIONS(4136),
    [anon_sym_signed] = ACTIONS(4134),
    [anon_sym_unsigned] = ACTIONS(4134),
    [anon_sym_long] = ACTIONS(4134),
    [anon_sym_short] = ACTIONS(4134),
    [anon_sym_static] = ACTIONS(4134),
    [anon_sym_ATautoreleasepool] = ACTIONS(4136),
    [anon_sym_auto] = ACTIONS(4134),
    [anon_sym_register] = ACTIONS(4134),
    [anon_sym_inline] = ACTIONS(4134),
    [anon_sym___inline] = ACTIONS(4134),
    [anon_sym___inline__] = ACTIONS(4134),
    [anon_sym___forceinline] = ACTIONS(4134),
    [anon_sym_thread_local] = ACTIONS(4134),
    [anon_sym___thread] = ACTIONS(4134),
    [anon_sym_CG_EXTERN] = ACTIONS(4134),
    [anon_sym_CG_INLINE] = ACTIONS(4134),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4134),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4134),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4134),
    [anon_sym_IBOutlet] = ACTIONS(4134),
    [anon_sym_IBInspectable] = ACTIONS(4134),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4134),
    [anon_sym_NS_INLINE] = ACTIONS(4134),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4134),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4134),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4134),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4134),
    [anon_sym_const] = ACTIONS(4134),
    [anon_sym_constexpr] = ACTIONS(4134),
    [anon_sym_volatile] = ACTIONS(4134),
    [anon_sym_restrict] = ACTIONS(4134),
    [anon_sym___restrict__] = ACTIONS(4134),
    [anon_sym__Atomic] = ACTIONS(4134),
    [anon_sym__Noreturn] = ACTIONS(4134),
    [anon_sym__Nonnull] = ACTIONS(4134),
    [anon_sym_nullable] = ACTIONS(4134),
    [anon_sym__Complex] = ACTIONS(4134),
    [anon_sym__Nullable] = ACTIONS(4134),
    [anon_sym__Nullable_result] = ACTIONS(4134),
    [anon_sym__Null_unspecified] = ACTIONS(4134),
    [anon_sym___autoreleasing] = ACTIONS(4134),
    [anon_sym___block] = ACTIONS(4134),
    [anon_sym___bridge] = ACTIONS(4134),
    [anon_sym___bridge_retained] = ACTIONS(4134),
    [anon_sym___bridge_transfer] = ACTIONS(4134),
    [anon_sym___complex] = ACTIONS(4134),
    [anon_sym___const] = ACTIONS(4134),
    [anon_sym___imag] = ACTIONS(4134),
    [anon_sym___kindof] = ACTIONS(4134),
    [anon_sym___nonnull] = ACTIONS(4134),
    [anon_sym___nullable] = ACTIONS(4134),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4134),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4134),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4134),
    [anon_sym___real] = ACTIONS(4134),
    [anon_sym___strong] = ACTIONS(4134),
    [anon_sym___unsafe_unretained] = ACTIONS(4134),
    [anon_sym___unused] = ACTIONS(4134),
    [anon_sym___weak] = ACTIONS(4134),
    [anon_sym_alignas] = ACTIONS(4134),
    [anon_sym__Alignas] = ACTIONS(4134),
    [sym_primitive_type] = ACTIONS(4134),
    [anon_sym_enum] = ACTIONS(4134),
    [anon_sym_COLON] = ACTIONS(4136),
    [anon_sym_struct] = ACTIONS(4134),
    [anon_sym_union] = ACTIONS(4134),
    [anon_sym_QMARK] = ACTIONS(4136),
    [anon_sym_DASH_DASH] = ACTIONS(4136),
    [anon_sym_PLUS_PLUS] = ACTIONS(4136),
    [anon_sym_DOT] = ACTIONS(4134),
    [anon_sym_DASH_GT] = ACTIONS(4136),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4134),
    [anon_sym___typeof] = ACTIONS(4134),
    [anon_sym_typeof] = ACTIONS(4134),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4134),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4134),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4134),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4134),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4134),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4134),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4134),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4134),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4134),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4134),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4134),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4134),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4134),
    [anon_sym_API_AVAILABLE] = ACTIONS(4134),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4134),
    [anon_sym_API_DEPRECATED] = ACTIONS(4134),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4134),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4134),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4134),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4134),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4134),
    [anon_sym___deprecated_msg] = ACTIONS(4134),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4134),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4134),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4134),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4134),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4134),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4134),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4134),
    [anon_sym_BOOL] = ACTIONS(4134),
    [anon_sym_IMP] = ACTIONS(4134),
    [anon_sym_SEL] = ACTIONS(4134),
    [anon_sym_Class] = ACTIONS(4134),
    [anon_sym_id] = ACTIONS(4134),
  },
  [STATE(1632)] = {
    [sym_identifier] = ACTIONS(4138),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4140),
    [anon_sym_COMMA] = ACTIONS(4140),
    [anon_sym_RPAREN] = ACTIONS(4140),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4140),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4140),
    [anon_sym_LPAREN2] = ACTIONS(4140),
    [anon_sym_DASH] = ACTIONS(4138),
    [anon_sym_PLUS] = ACTIONS(4138),
    [anon_sym_STAR] = ACTIONS(4140),
    [anon_sym_SLASH] = ACTIONS(4138),
    [anon_sym_PERCENT] = ACTIONS(4140),
    [anon_sym_PIPE_PIPE] = ACTIONS(4140),
    [anon_sym_AMP_AMP] = ACTIONS(4140),
    [anon_sym_PIPE] = ACTIONS(4138),
    [anon_sym_CARET] = ACTIONS(4140),
    [anon_sym_AMP] = ACTIONS(4138),
    [anon_sym_EQ_EQ] = ACTIONS(4140),
    [anon_sym_BANG_EQ] = ACTIONS(4140),
    [anon_sym_GT] = ACTIONS(4138),
    [anon_sym_GT_EQ] = ACTIONS(4140),
    [anon_sym_LT_EQ] = ACTIONS(4140),
    [anon_sym_LT] = ACTIONS(4138),
    [anon_sym_LT_LT] = ACTIONS(4140),
    [anon_sym_GT_GT] = ACTIONS(4140),
    [anon_sym_SEMI] = ACTIONS(4140),
    [anon_sym___extension__] = ACTIONS(4138),
    [anon_sym_extern] = ACTIONS(4138),
    [anon_sym___attribute__] = ACTIONS(4138),
    [anon_sym___attribute] = ACTIONS(4138),
    [anon_sym_noreturn] = ACTIONS(4138),
    [anon_sym_LBRACK] = ACTIONS(4140),
    [anon_sym_RBRACK] = ACTIONS(4140),
    [anon_sym___declspec] = ACTIONS(4138),
    [anon_sym_LBRACE] = ACTIONS(4140),
    [anon_sym_RBRACE] = ACTIONS(4140),
    [anon_sym_signed] = ACTIONS(4138),
    [anon_sym_unsigned] = ACTIONS(4138),
    [anon_sym_long] = ACTIONS(4138),
    [anon_sym_short] = ACTIONS(4138),
    [anon_sym_static] = ACTIONS(4138),
    [anon_sym_ATautoreleasepool] = ACTIONS(4140),
    [anon_sym_auto] = ACTIONS(4138),
    [anon_sym_register] = ACTIONS(4138),
    [anon_sym_inline] = ACTIONS(4138),
    [anon_sym___inline] = ACTIONS(4138),
    [anon_sym___inline__] = ACTIONS(4138),
    [anon_sym___forceinline] = ACTIONS(4138),
    [anon_sym_thread_local] = ACTIONS(4138),
    [anon_sym___thread] = ACTIONS(4138),
    [anon_sym_CG_EXTERN] = ACTIONS(4138),
    [anon_sym_CG_INLINE] = ACTIONS(4138),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4138),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4138),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4138),
    [anon_sym_IBOutlet] = ACTIONS(4138),
    [anon_sym_IBInspectable] = ACTIONS(4138),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4138),
    [anon_sym_NS_INLINE] = ACTIONS(4138),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4138),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4138),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4138),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4138),
    [anon_sym_const] = ACTIONS(4138),
    [anon_sym_constexpr] = ACTIONS(4138),
    [anon_sym_volatile] = ACTIONS(4138),
    [anon_sym_restrict] = ACTIONS(4138),
    [anon_sym___restrict__] = ACTIONS(4138),
    [anon_sym__Atomic] = ACTIONS(4138),
    [anon_sym__Noreturn] = ACTIONS(4138),
    [anon_sym__Nonnull] = ACTIONS(4138),
    [anon_sym_nullable] = ACTIONS(4138),
    [anon_sym__Complex] = ACTIONS(4138),
    [anon_sym__Nullable] = ACTIONS(4138),
    [anon_sym__Nullable_result] = ACTIONS(4138),
    [anon_sym__Null_unspecified] = ACTIONS(4138),
    [anon_sym___autoreleasing] = ACTIONS(4138),
    [anon_sym___block] = ACTIONS(4138),
    [anon_sym___bridge] = ACTIONS(4138),
    [anon_sym___bridge_retained] = ACTIONS(4138),
    [anon_sym___bridge_transfer] = ACTIONS(4138),
    [anon_sym___complex] = ACTIONS(4138),
    [anon_sym___const] = ACTIONS(4138),
    [anon_sym___imag] = ACTIONS(4138),
    [anon_sym___kindof] = ACTIONS(4138),
    [anon_sym___nonnull] = ACTIONS(4138),
    [anon_sym___nullable] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4138),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4138),
    [anon_sym___real] = ACTIONS(4138),
    [anon_sym___strong] = ACTIONS(4138),
    [anon_sym___unsafe_unretained] = ACTIONS(4138),
    [anon_sym___unused] = ACTIONS(4138),
    [anon_sym___weak] = ACTIONS(4138),
    [anon_sym_alignas] = ACTIONS(4138),
    [anon_sym__Alignas] = ACTIONS(4138),
    [sym_primitive_type] = ACTIONS(4138),
    [anon_sym_enum] = ACTIONS(4138),
    [anon_sym_COLON] = ACTIONS(4140),
    [anon_sym_struct] = ACTIONS(4138),
    [anon_sym_union] = ACTIONS(4138),
    [anon_sym_QMARK] = ACTIONS(4140),
    [anon_sym_DASH_DASH] = ACTIONS(4140),
    [anon_sym_PLUS_PLUS] = ACTIONS(4140),
    [anon_sym_DOT] = ACTIONS(4138),
    [anon_sym_DASH_GT] = ACTIONS(4140),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4138),
    [anon_sym___typeof] = ACTIONS(4138),
    [anon_sym_typeof] = ACTIONS(4138),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4138),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4138),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4138),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4138),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4138),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4138),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4138),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_API_AVAILABLE] = ACTIONS(4138),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_API_DEPRECATED] = ACTIONS(4138),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4138),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4138),
    [anon_sym___deprecated_msg] = ACTIONS(4138),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4138),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4138),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4138),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4138),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4138),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4138),
    [anon_sym_BOOL] = ACTIONS(4138),
    [anon_sym_IMP] = ACTIONS(4138),
    [anon_sym_SEL] = ACTIONS(4138),
    [anon_sym_Class] = ACTIONS(4138),
    [anon_sym_id] = ACTIONS(4138),
  },
  [STATE(1633)] = {
    [sym_identifier] = ACTIONS(4122),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4124),
    [anon_sym_COMMA] = ACTIONS(4124),
    [anon_sym_RPAREN] = ACTIONS(4124),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4124),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4124),
    [anon_sym_LPAREN2] = ACTIONS(4124),
    [anon_sym_DASH] = ACTIONS(4122),
    [anon_sym_PLUS] = ACTIONS(4122),
    [anon_sym_STAR] = ACTIONS(4124),
    [anon_sym_SLASH] = ACTIONS(4122),
    [anon_sym_PERCENT] = ACTIONS(4124),
    [anon_sym_PIPE_PIPE] = ACTIONS(4124),
    [anon_sym_AMP_AMP] = ACTIONS(4124),
    [anon_sym_PIPE] = ACTIONS(4122),
    [anon_sym_CARET] = ACTIONS(4124),
    [anon_sym_AMP] = ACTIONS(4122),
    [anon_sym_EQ_EQ] = ACTIONS(4124),
    [anon_sym_BANG_EQ] = ACTIONS(4124),
    [anon_sym_GT] = ACTIONS(4122),
    [anon_sym_GT_EQ] = ACTIONS(4124),
    [anon_sym_LT_EQ] = ACTIONS(4124),
    [anon_sym_LT] = ACTIONS(4122),
    [anon_sym_LT_LT] = ACTIONS(4124),
    [anon_sym_GT_GT] = ACTIONS(4124),
    [anon_sym_SEMI] = ACTIONS(4124),
    [anon_sym___extension__] = ACTIONS(4122),
    [anon_sym_extern] = ACTIONS(4122),
    [anon_sym___attribute__] = ACTIONS(4122),
    [anon_sym___attribute] = ACTIONS(4122),
    [anon_sym_noreturn] = ACTIONS(4122),
    [anon_sym_LBRACK] = ACTIONS(4124),
    [anon_sym_RBRACK] = ACTIONS(4124),
    [anon_sym___declspec] = ACTIONS(4122),
    [anon_sym_LBRACE] = ACTIONS(4124),
    [anon_sym_RBRACE] = ACTIONS(4124),
    [anon_sym_signed] = ACTIONS(4122),
    [anon_sym_unsigned] = ACTIONS(4122),
    [anon_sym_long] = ACTIONS(4122),
    [anon_sym_short] = ACTIONS(4122),
    [anon_sym_static] = ACTIONS(4122),
    [anon_sym_ATautoreleasepool] = ACTIONS(4124),
    [anon_sym_auto] = ACTIONS(4122),
    [anon_sym_register] = ACTIONS(4122),
    [anon_sym_inline] = ACTIONS(4122),
    [anon_sym___inline] = ACTIONS(4122),
    [anon_sym___inline__] = ACTIONS(4122),
    [anon_sym___forceinline] = ACTIONS(4122),
    [anon_sym_thread_local] = ACTIONS(4122),
    [anon_sym___thread] = ACTIONS(4122),
    [anon_sym_CG_EXTERN] = ACTIONS(4122),
    [anon_sym_CG_INLINE] = ACTIONS(4122),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4122),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4122),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4122),
    [anon_sym_IBOutlet] = ACTIONS(4122),
    [anon_sym_IBInspectable] = ACTIONS(4122),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4122),
    [anon_sym_NS_INLINE] = ACTIONS(4122),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4122),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4122),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4122),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4122),
    [anon_sym_const] = ACTIONS(4122),
    [anon_sym_constexpr] = ACTIONS(4122),
    [anon_sym_volatile] = ACTIONS(4122),
    [anon_sym_restrict] = ACTIONS(4122),
    [anon_sym___restrict__] = ACTIONS(4122),
    [anon_sym__Atomic] = ACTIONS(4122),
    [anon_sym__Noreturn] = ACTIONS(4122),
    [anon_sym__Nonnull] = ACTIONS(4122),
    [anon_sym_nullable] = ACTIONS(4122),
    [anon_sym__Complex] = ACTIONS(4122),
    [anon_sym__Nullable] = ACTIONS(4122),
    [anon_sym__Nullable_result] = ACTIONS(4122),
    [anon_sym__Null_unspecified] = ACTIONS(4122),
    [anon_sym___autoreleasing] = ACTIONS(4122),
    [anon_sym___block] = ACTIONS(4122),
    [anon_sym___bridge] = ACTIONS(4122),
    [anon_sym___bridge_retained] = ACTIONS(4122),
    [anon_sym___bridge_transfer] = ACTIONS(4122),
    [anon_sym___complex] = ACTIONS(4122),
    [anon_sym___const] = ACTIONS(4122),
    [anon_sym___imag] = ACTIONS(4122),
    [anon_sym___kindof] = ACTIONS(4122),
    [anon_sym___nonnull] = ACTIONS(4122),
    [anon_sym___nullable] = ACTIONS(4122),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4122),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4122),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4122),
    [anon_sym___real] = ACTIONS(4122),
    [anon_sym___strong] = ACTIONS(4122),
    [anon_sym___unsafe_unretained] = ACTIONS(4122),
    [anon_sym___unused] = ACTIONS(4122),
    [anon_sym___weak] = ACTIONS(4122),
    [anon_sym_alignas] = ACTIONS(4122),
    [anon_sym__Alignas] = ACTIONS(4122),
    [sym_primitive_type] = ACTIONS(4122),
    [anon_sym_enum] = ACTIONS(4122),
    [anon_sym_COLON] = ACTIONS(4124),
    [anon_sym_struct] = ACTIONS(4122),
    [anon_sym_union] = ACTIONS(4122),
    [anon_sym_QMARK] = ACTIONS(4124),
    [anon_sym_DASH_DASH] = ACTIONS(4124),
    [anon_sym_PLUS_PLUS] = ACTIONS(4124),
    [anon_sym_DOT] = ACTIONS(4122),
    [anon_sym_DASH_GT] = ACTIONS(4124),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4122),
    [anon_sym___typeof] = ACTIONS(4122),
    [anon_sym_typeof] = ACTIONS(4122),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4122),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4122),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4122),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4122),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4122),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4122),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4122),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4122),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4122),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4122),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4122),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4122),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4122),
    [anon_sym_API_AVAILABLE] = ACTIONS(4122),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4122),
    [anon_sym_API_DEPRECATED] = ACTIONS(4122),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4122),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4122),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4122),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4122),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4122),
    [anon_sym___deprecated_msg] = ACTIONS(4122),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4122),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4122),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4122),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4122),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4122),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4122),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4122),
    [anon_sym_BOOL] = ACTIONS(4122),
    [anon_sym_IMP] = ACTIONS(4122),
    [anon_sym_SEL] = ACTIONS(4122),
    [anon_sym_Class] = ACTIONS(4122),
    [anon_sym_id] = ACTIONS(4122),
  },
  [STATE(1634)] = {
    [sym_identifier] = ACTIONS(4094),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4096),
    [anon_sym_COMMA] = ACTIONS(4096),
    [anon_sym_RPAREN] = ACTIONS(4096),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4096),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4096),
    [anon_sym_LPAREN2] = ACTIONS(4096),
    [anon_sym_DASH] = ACTIONS(4094),
    [anon_sym_PLUS] = ACTIONS(4094),
    [anon_sym_STAR] = ACTIONS(4096),
    [anon_sym_SLASH] = ACTIONS(4094),
    [anon_sym_PERCENT] = ACTIONS(4096),
    [anon_sym_PIPE_PIPE] = ACTIONS(4096),
    [anon_sym_AMP_AMP] = ACTIONS(4096),
    [anon_sym_PIPE] = ACTIONS(4094),
    [anon_sym_CARET] = ACTIONS(4096),
    [anon_sym_AMP] = ACTIONS(4094),
    [anon_sym_EQ_EQ] = ACTIONS(4096),
    [anon_sym_BANG_EQ] = ACTIONS(4096),
    [anon_sym_GT] = ACTIONS(4094),
    [anon_sym_GT_EQ] = ACTIONS(4096),
    [anon_sym_LT_EQ] = ACTIONS(4096),
    [anon_sym_LT] = ACTIONS(4094),
    [anon_sym_LT_LT] = ACTIONS(4096),
    [anon_sym_GT_GT] = ACTIONS(4096),
    [anon_sym_SEMI] = ACTIONS(4096),
    [anon_sym___extension__] = ACTIONS(4094),
    [anon_sym_extern] = ACTIONS(4094),
    [anon_sym___attribute__] = ACTIONS(4094),
    [anon_sym___attribute] = ACTIONS(4094),
    [anon_sym_noreturn] = ACTIONS(4094),
    [anon_sym_LBRACK] = ACTIONS(4096),
    [anon_sym_RBRACK] = ACTIONS(4096),
    [anon_sym___declspec] = ACTIONS(4094),
    [anon_sym_LBRACE] = ACTIONS(4096),
    [anon_sym_RBRACE] = ACTIONS(4096),
    [anon_sym_signed] = ACTIONS(4094),
    [anon_sym_unsigned] = ACTIONS(4094),
    [anon_sym_long] = ACTIONS(4094),
    [anon_sym_short] = ACTIONS(4094),
    [anon_sym_static] = ACTIONS(4094),
    [anon_sym_ATautoreleasepool] = ACTIONS(4096),
    [anon_sym_auto] = ACTIONS(4094),
    [anon_sym_register] = ACTIONS(4094),
    [anon_sym_inline] = ACTIONS(4094),
    [anon_sym___inline] = ACTIONS(4094),
    [anon_sym___inline__] = ACTIONS(4094),
    [anon_sym___forceinline] = ACTIONS(4094),
    [anon_sym_thread_local] = ACTIONS(4094),
    [anon_sym___thread] = ACTIONS(4094),
    [anon_sym_CG_EXTERN] = ACTIONS(4094),
    [anon_sym_CG_INLINE] = ACTIONS(4094),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4094),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4094),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4094),
    [anon_sym_IBOutlet] = ACTIONS(4094),
    [anon_sym_IBInspectable] = ACTIONS(4094),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4094),
    [anon_sym_NS_INLINE] = ACTIONS(4094),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4094),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4094),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4094),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4094),
    [anon_sym_const] = ACTIONS(4094),
    [anon_sym_constexpr] = ACTIONS(4094),
    [anon_sym_volatile] = ACTIONS(4094),
    [anon_sym_restrict] = ACTIONS(4094),
    [anon_sym___restrict__] = ACTIONS(4094),
    [anon_sym__Atomic] = ACTIONS(4094),
    [anon_sym__Noreturn] = ACTIONS(4094),
    [anon_sym__Nonnull] = ACTIONS(4094),
    [anon_sym_nullable] = ACTIONS(4094),
    [anon_sym__Complex] = ACTIONS(4094),
    [anon_sym__Nullable] = ACTIONS(4094),
    [anon_sym__Nullable_result] = ACTIONS(4094),
    [anon_sym__Null_unspecified] = ACTIONS(4094),
    [anon_sym___autoreleasing] = ACTIONS(4094),
    [anon_sym___block] = ACTIONS(4094),
    [anon_sym___bridge] = ACTIONS(4094),
    [anon_sym___bridge_retained] = ACTIONS(4094),
    [anon_sym___bridge_transfer] = ACTIONS(4094),
    [anon_sym___complex] = ACTIONS(4094),
    [anon_sym___const] = ACTIONS(4094),
    [anon_sym___imag] = ACTIONS(4094),
    [anon_sym___kindof] = ACTIONS(4094),
    [anon_sym___nonnull] = ACTIONS(4094),
    [anon_sym___nullable] = ACTIONS(4094),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4094),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4094),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4094),
    [anon_sym___real] = ACTIONS(4094),
    [anon_sym___strong] = ACTIONS(4094),
    [anon_sym___unsafe_unretained] = ACTIONS(4094),
    [anon_sym___unused] = ACTIONS(4094),
    [anon_sym___weak] = ACTIONS(4094),
    [anon_sym_alignas] = ACTIONS(4094),
    [anon_sym__Alignas] = ACTIONS(4094),
    [sym_primitive_type] = ACTIONS(4094),
    [anon_sym_enum] = ACTIONS(4094),
    [anon_sym_COLON] = ACTIONS(4096),
    [anon_sym_struct] = ACTIONS(4094),
    [anon_sym_union] = ACTIONS(4094),
    [anon_sym_QMARK] = ACTIONS(4096),
    [anon_sym_DASH_DASH] = ACTIONS(4096),
    [anon_sym_PLUS_PLUS] = ACTIONS(4096),
    [anon_sym_DOT] = ACTIONS(4094),
    [anon_sym_DASH_GT] = ACTIONS(4096),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4094),
    [anon_sym___typeof] = ACTIONS(4094),
    [anon_sym_typeof] = ACTIONS(4094),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4094),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4094),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4094),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4094),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4094),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4094),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4094),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4094),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4094),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4094),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4094),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4094),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4094),
    [anon_sym_API_AVAILABLE] = ACTIONS(4094),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4094),
    [anon_sym_API_DEPRECATED] = ACTIONS(4094),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4094),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4094),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4094),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4094),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4094),
    [anon_sym___deprecated_msg] = ACTIONS(4094),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4094),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4094),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4094),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4094),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4094),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4094),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4094),
    [anon_sym_BOOL] = ACTIONS(4094),
    [anon_sym_IMP] = ACTIONS(4094),
    [anon_sym_SEL] = ACTIONS(4094),
    [anon_sym_Class] = ACTIONS(4094),
    [anon_sym_id] = ACTIONS(4094),
  },
  [STATE(1635)] = {
    [sym_identifier] = ACTIONS(4098),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4100),
    [anon_sym_COMMA] = ACTIONS(4100),
    [anon_sym_RPAREN] = ACTIONS(4100),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4100),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4100),
    [anon_sym_LPAREN2] = ACTIONS(4100),
    [anon_sym_DASH] = ACTIONS(4098),
    [anon_sym_PLUS] = ACTIONS(4098),
    [anon_sym_STAR] = ACTIONS(4100),
    [anon_sym_SLASH] = ACTIONS(4098),
    [anon_sym_PERCENT] = ACTIONS(4100),
    [anon_sym_PIPE_PIPE] = ACTIONS(4100),
    [anon_sym_AMP_AMP] = ACTIONS(4100),
    [anon_sym_PIPE] = ACTIONS(4098),
    [anon_sym_CARET] = ACTIONS(4100),
    [anon_sym_AMP] = ACTIONS(4098),
    [anon_sym_EQ_EQ] = ACTIONS(4100),
    [anon_sym_BANG_EQ] = ACTIONS(4100),
    [anon_sym_GT] = ACTIONS(4098),
    [anon_sym_GT_EQ] = ACTIONS(4100),
    [anon_sym_LT_EQ] = ACTIONS(4100),
    [anon_sym_LT] = ACTIONS(4098),
    [anon_sym_LT_LT] = ACTIONS(4100),
    [anon_sym_GT_GT] = ACTIONS(4100),
    [anon_sym_SEMI] = ACTIONS(4100),
    [anon_sym___extension__] = ACTIONS(4098),
    [anon_sym_extern] = ACTIONS(4098),
    [anon_sym___attribute__] = ACTIONS(4098),
    [anon_sym___attribute] = ACTIONS(4098),
    [anon_sym_noreturn] = ACTIONS(4098),
    [anon_sym_LBRACK] = ACTIONS(4100),
    [anon_sym_RBRACK] = ACTIONS(4100),
    [anon_sym___declspec] = ACTIONS(4098),
    [anon_sym_LBRACE] = ACTIONS(4100),
    [anon_sym_RBRACE] = ACTIONS(4100),
    [anon_sym_signed] = ACTIONS(4098),
    [anon_sym_unsigned] = ACTIONS(4098),
    [anon_sym_long] = ACTIONS(4098),
    [anon_sym_short] = ACTIONS(4098),
    [anon_sym_static] = ACTIONS(4098),
    [anon_sym_ATautoreleasepool] = ACTIONS(4100),
    [anon_sym_auto] = ACTIONS(4098),
    [anon_sym_register] = ACTIONS(4098),
    [anon_sym_inline] = ACTIONS(4098),
    [anon_sym___inline] = ACTIONS(4098),
    [anon_sym___inline__] = ACTIONS(4098),
    [anon_sym___forceinline] = ACTIONS(4098),
    [anon_sym_thread_local] = ACTIONS(4098),
    [anon_sym___thread] = ACTIONS(4098),
    [anon_sym_CG_EXTERN] = ACTIONS(4098),
    [anon_sym_CG_INLINE] = ACTIONS(4098),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4098),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4098),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4098),
    [anon_sym_IBOutlet] = ACTIONS(4098),
    [anon_sym_IBInspectable] = ACTIONS(4098),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4098),
    [anon_sym_NS_INLINE] = ACTIONS(4098),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4098),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4098),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4098),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4098),
    [anon_sym_const] = ACTIONS(4098),
    [anon_sym_constexpr] = ACTIONS(4098),
    [anon_sym_volatile] = ACTIONS(4098),
    [anon_sym_restrict] = ACTIONS(4098),
    [anon_sym___restrict__] = ACTIONS(4098),
    [anon_sym__Atomic] = ACTIONS(4098),
    [anon_sym__Noreturn] = ACTIONS(4098),
    [anon_sym__Nonnull] = ACTIONS(4098),
    [anon_sym_nullable] = ACTIONS(4098),
    [anon_sym__Complex] = ACTIONS(4098),
    [anon_sym__Nullable] = ACTIONS(4098),
    [anon_sym__Nullable_result] = ACTIONS(4098),
    [anon_sym__Null_unspecified] = ACTIONS(4098),
    [anon_sym___autoreleasing] = ACTIONS(4098),
    [anon_sym___block] = ACTIONS(4098),
    [anon_sym___bridge] = ACTIONS(4098),
    [anon_sym___bridge_retained] = ACTIONS(4098),
    [anon_sym___bridge_transfer] = ACTIONS(4098),
    [anon_sym___complex] = ACTIONS(4098),
    [anon_sym___const] = ACTIONS(4098),
    [anon_sym___imag] = ACTIONS(4098),
    [anon_sym___kindof] = ACTIONS(4098),
    [anon_sym___nonnull] = ACTIONS(4098),
    [anon_sym___nullable] = ACTIONS(4098),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4098),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4098),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4098),
    [anon_sym___real] = ACTIONS(4098),
    [anon_sym___strong] = ACTIONS(4098),
    [anon_sym___unsafe_unretained] = ACTIONS(4098),
    [anon_sym___unused] = ACTIONS(4098),
    [anon_sym___weak] = ACTIONS(4098),
    [anon_sym_alignas] = ACTIONS(4098),
    [anon_sym__Alignas] = ACTIONS(4098),
    [sym_primitive_type] = ACTIONS(4098),
    [anon_sym_enum] = ACTIONS(4098),
    [anon_sym_COLON] = ACTIONS(4100),
    [anon_sym_struct] = ACTIONS(4098),
    [anon_sym_union] = ACTIONS(4098),
    [anon_sym_QMARK] = ACTIONS(4100),
    [anon_sym_DASH_DASH] = ACTIONS(4100),
    [anon_sym_PLUS_PLUS] = ACTIONS(4100),
    [anon_sym_DOT] = ACTIONS(4098),
    [anon_sym_DASH_GT] = ACTIONS(4100),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4098),
    [anon_sym___typeof] = ACTIONS(4098),
    [anon_sym_typeof] = ACTIONS(4098),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4098),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4098),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4098),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4098),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4098),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4098),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4098),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4098),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4098),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4098),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4098),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4098),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4098),
    [anon_sym_API_AVAILABLE] = ACTIONS(4098),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4098),
    [anon_sym_API_DEPRECATED] = ACTIONS(4098),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4098),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4098),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4098),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4098),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4098),
    [anon_sym___deprecated_msg] = ACTIONS(4098),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4098),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4098),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4098),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4098),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4098),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4098),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4098),
    [anon_sym_BOOL] = ACTIONS(4098),
    [anon_sym_IMP] = ACTIONS(4098),
    [anon_sym_SEL] = ACTIONS(4098),
    [anon_sym_Class] = ACTIONS(4098),
    [anon_sym_id] = ACTIONS(4098),
  },
  [STATE(1636)] = {
    [sym_identifier] = ACTIONS(4164),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4166),
    [anon_sym_COMMA] = ACTIONS(4166),
    [anon_sym_RPAREN] = ACTIONS(4166),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4166),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4166),
    [anon_sym_LPAREN2] = ACTIONS(4166),
    [anon_sym_DASH] = ACTIONS(4164),
    [anon_sym_PLUS] = ACTIONS(4164),
    [anon_sym_STAR] = ACTIONS(4166),
    [anon_sym_SLASH] = ACTIONS(4164),
    [anon_sym_PERCENT] = ACTIONS(4166),
    [anon_sym_PIPE_PIPE] = ACTIONS(4166),
    [anon_sym_AMP_AMP] = ACTIONS(4166),
    [anon_sym_PIPE] = ACTIONS(4164),
    [anon_sym_CARET] = ACTIONS(4166),
    [anon_sym_AMP] = ACTIONS(4164),
    [anon_sym_EQ_EQ] = ACTIONS(4166),
    [anon_sym_BANG_EQ] = ACTIONS(4166),
    [anon_sym_GT] = ACTIONS(4164),
    [anon_sym_GT_EQ] = ACTIONS(4166),
    [anon_sym_LT_EQ] = ACTIONS(4166),
    [anon_sym_LT] = ACTIONS(4164),
    [anon_sym_LT_LT] = ACTIONS(4166),
    [anon_sym_GT_GT] = ACTIONS(4166),
    [anon_sym_SEMI] = ACTIONS(4166),
    [anon_sym___extension__] = ACTIONS(4164),
    [anon_sym_extern] = ACTIONS(4164),
    [anon_sym___attribute__] = ACTIONS(4164),
    [anon_sym___attribute] = ACTIONS(4164),
    [anon_sym_noreturn] = ACTIONS(4164),
    [anon_sym_LBRACK] = ACTIONS(4166),
    [anon_sym_RBRACK] = ACTIONS(4166),
    [anon_sym___declspec] = ACTIONS(4164),
    [anon_sym_LBRACE] = ACTIONS(4166),
    [anon_sym_RBRACE] = ACTIONS(4166),
    [anon_sym_signed] = ACTIONS(4164),
    [anon_sym_unsigned] = ACTIONS(4164),
    [anon_sym_long] = ACTIONS(4164),
    [anon_sym_short] = ACTIONS(4164),
    [anon_sym_static] = ACTIONS(4164),
    [anon_sym_ATautoreleasepool] = ACTIONS(4166),
    [anon_sym_auto] = ACTIONS(4164),
    [anon_sym_register] = ACTIONS(4164),
    [anon_sym_inline] = ACTIONS(4164),
    [anon_sym___inline] = ACTIONS(4164),
    [anon_sym___inline__] = ACTIONS(4164),
    [anon_sym___forceinline] = ACTIONS(4164),
    [anon_sym_thread_local] = ACTIONS(4164),
    [anon_sym___thread] = ACTIONS(4164),
    [anon_sym_CG_EXTERN] = ACTIONS(4164),
    [anon_sym_CG_INLINE] = ACTIONS(4164),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4164),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4164),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4164),
    [anon_sym_IBOutlet] = ACTIONS(4164),
    [anon_sym_IBInspectable] = ACTIONS(4164),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4164),
    [anon_sym_NS_INLINE] = ACTIONS(4164),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4164),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4164),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4164),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4164),
    [anon_sym_const] = ACTIONS(4164),
    [anon_sym_constexpr] = ACTIONS(4164),
    [anon_sym_volatile] = ACTIONS(4164),
    [anon_sym_restrict] = ACTIONS(4164),
    [anon_sym___restrict__] = ACTIONS(4164),
    [anon_sym__Atomic] = ACTIONS(4164),
    [anon_sym__Noreturn] = ACTIONS(4164),
    [anon_sym__Nonnull] = ACTIONS(4164),
    [anon_sym_nullable] = ACTIONS(4164),
    [anon_sym__Complex] = ACTIONS(4164),
    [anon_sym__Nullable] = ACTIONS(4164),
    [anon_sym__Nullable_result] = ACTIONS(4164),
    [anon_sym__Null_unspecified] = ACTIONS(4164),
    [anon_sym___autoreleasing] = ACTIONS(4164),
    [anon_sym___block] = ACTIONS(4164),
    [anon_sym___bridge] = ACTIONS(4164),
    [anon_sym___bridge_retained] = ACTIONS(4164),
    [anon_sym___bridge_transfer] = ACTIONS(4164),
    [anon_sym___complex] = ACTIONS(4164),
    [anon_sym___const] = ACTIONS(4164),
    [anon_sym___imag] = ACTIONS(4164),
    [anon_sym___kindof] = ACTIONS(4164),
    [anon_sym___nonnull] = ACTIONS(4164),
    [anon_sym___nullable] = ACTIONS(4164),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4164),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4164),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4164),
    [anon_sym___real] = ACTIONS(4164),
    [anon_sym___strong] = ACTIONS(4164),
    [anon_sym___unsafe_unretained] = ACTIONS(4164),
    [anon_sym___unused] = ACTIONS(4164),
    [anon_sym___weak] = ACTIONS(4164),
    [anon_sym_alignas] = ACTIONS(4164),
    [anon_sym__Alignas] = ACTIONS(4164),
    [sym_primitive_type] = ACTIONS(4164),
    [anon_sym_enum] = ACTIONS(4164),
    [anon_sym_COLON] = ACTIONS(4166),
    [anon_sym_struct] = ACTIONS(4164),
    [anon_sym_union] = ACTIONS(4164),
    [anon_sym_QMARK] = ACTIONS(4166),
    [anon_sym_DASH_DASH] = ACTIONS(4166),
    [anon_sym_PLUS_PLUS] = ACTIONS(4166),
    [anon_sym_DOT] = ACTIONS(4164),
    [anon_sym_DASH_GT] = ACTIONS(4166),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4164),
    [anon_sym___typeof] = ACTIONS(4164),
    [anon_sym_typeof] = ACTIONS(4164),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4164),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4164),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4164),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4164),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4164),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4164),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4164),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4164),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4164),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4164),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4164),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4164),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4164),
    [anon_sym_API_AVAILABLE] = ACTIONS(4164),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4164),
    [anon_sym_API_DEPRECATED] = ACTIONS(4164),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4164),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4164),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4164),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4164),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4164),
    [anon_sym___deprecated_msg] = ACTIONS(4164),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4164),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4164),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4164),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4164),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4164),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4164),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4164),
    [anon_sym_BOOL] = ACTIONS(4164),
    [anon_sym_IMP] = ACTIONS(4164),
    [anon_sym_SEL] = ACTIONS(4164),
    [anon_sym_Class] = ACTIONS(4164),
    [anon_sym_id] = ACTIONS(4164),
  },
  [STATE(1637)] = {
    [sym_identifier] = ACTIONS(4172),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4174),
    [anon_sym_COMMA] = ACTIONS(4174),
    [anon_sym_RPAREN] = ACTIONS(4174),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4174),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4174),
    [anon_sym_LPAREN2] = ACTIONS(4174),
    [anon_sym_DASH] = ACTIONS(4172),
    [anon_sym_PLUS] = ACTIONS(4172),
    [anon_sym_STAR] = ACTIONS(4174),
    [anon_sym_SLASH] = ACTIONS(4172),
    [anon_sym_PERCENT] = ACTIONS(4174),
    [anon_sym_PIPE_PIPE] = ACTIONS(4174),
    [anon_sym_AMP_AMP] = ACTIONS(4174),
    [anon_sym_PIPE] = ACTIONS(4172),
    [anon_sym_CARET] = ACTIONS(4174),
    [anon_sym_AMP] = ACTIONS(4172),
    [anon_sym_EQ_EQ] = ACTIONS(4174),
    [anon_sym_BANG_EQ] = ACTIONS(4174),
    [anon_sym_GT] = ACTIONS(4172),
    [anon_sym_GT_EQ] = ACTIONS(4174),
    [anon_sym_LT_EQ] = ACTIONS(4174),
    [anon_sym_LT] = ACTIONS(4172),
    [anon_sym_LT_LT] = ACTIONS(4174),
    [anon_sym_GT_GT] = ACTIONS(4174),
    [anon_sym_SEMI] = ACTIONS(4174),
    [anon_sym___extension__] = ACTIONS(4172),
    [anon_sym_extern] = ACTIONS(4172),
    [anon_sym___attribute__] = ACTIONS(4172),
    [anon_sym___attribute] = ACTIONS(4172),
    [anon_sym_noreturn] = ACTIONS(4172),
    [anon_sym_LBRACK] = ACTIONS(4174),
    [anon_sym_RBRACK] = ACTIONS(4174),
    [anon_sym___declspec] = ACTIONS(4172),
    [anon_sym_LBRACE] = ACTIONS(4174),
    [anon_sym_RBRACE] = ACTIONS(4174),
    [anon_sym_signed] = ACTIONS(4172),
    [anon_sym_unsigned] = ACTIONS(4172),
    [anon_sym_long] = ACTIONS(4172),
    [anon_sym_short] = ACTIONS(4172),
    [anon_sym_static] = ACTIONS(4172),
    [anon_sym_ATautoreleasepool] = ACTIONS(4174),
    [anon_sym_auto] = ACTIONS(4172),
    [anon_sym_register] = ACTIONS(4172),
    [anon_sym_inline] = ACTIONS(4172),
    [anon_sym___inline] = ACTIONS(4172),
    [anon_sym___inline__] = ACTIONS(4172),
    [anon_sym___forceinline] = ACTIONS(4172),
    [anon_sym_thread_local] = ACTIONS(4172),
    [anon_sym___thread] = ACTIONS(4172),
    [anon_sym_CG_EXTERN] = ACTIONS(4172),
    [anon_sym_CG_INLINE] = ACTIONS(4172),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4172),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4172),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4172),
    [anon_sym_IBOutlet] = ACTIONS(4172),
    [anon_sym_IBInspectable] = ACTIONS(4172),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4172),
    [anon_sym_NS_INLINE] = ACTIONS(4172),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4172),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4172),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4172),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4172),
    [anon_sym_const] = ACTIONS(4172),
    [anon_sym_constexpr] = ACTIONS(4172),
    [anon_sym_volatile] = ACTIONS(4172),
    [anon_sym_restrict] = ACTIONS(4172),
    [anon_sym___restrict__] = ACTIONS(4172),
    [anon_sym__Atomic] = ACTIONS(4172),
    [anon_sym__Noreturn] = ACTIONS(4172),
    [anon_sym__Nonnull] = ACTIONS(4172),
    [anon_sym_nullable] = ACTIONS(4172),
    [anon_sym__Complex] = ACTIONS(4172),
    [anon_sym__Nullable] = ACTIONS(4172),
    [anon_sym__Nullable_result] = ACTIONS(4172),
    [anon_sym__Null_unspecified] = ACTIONS(4172),
    [anon_sym___autoreleasing] = ACTIONS(4172),
    [anon_sym___block] = ACTIONS(4172),
    [anon_sym___bridge] = ACTIONS(4172),
    [anon_sym___bridge_retained] = ACTIONS(4172),
    [anon_sym___bridge_transfer] = ACTIONS(4172),
    [anon_sym___complex] = ACTIONS(4172),
    [anon_sym___const] = ACTIONS(4172),
    [anon_sym___imag] = ACTIONS(4172),
    [anon_sym___kindof] = ACTIONS(4172),
    [anon_sym___nonnull] = ACTIONS(4172),
    [anon_sym___nullable] = ACTIONS(4172),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4172),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4172),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4172),
    [anon_sym___real] = ACTIONS(4172),
    [anon_sym___strong] = ACTIONS(4172),
    [anon_sym___unsafe_unretained] = ACTIONS(4172),
    [anon_sym___unused] = ACTIONS(4172),
    [anon_sym___weak] = ACTIONS(4172),
    [anon_sym_alignas] = ACTIONS(4172),
    [anon_sym__Alignas] = ACTIONS(4172),
    [sym_primitive_type] = ACTIONS(4172),
    [anon_sym_enum] = ACTIONS(4172),
    [anon_sym_COLON] = ACTIONS(4174),
    [anon_sym_struct] = ACTIONS(4172),
    [anon_sym_union] = ACTIONS(4172),
    [anon_sym_QMARK] = ACTIONS(4174),
    [anon_sym_DASH_DASH] = ACTIONS(4174),
    [anon_sym_PLUS_PLUS] = ACTIONS(4174),
    [anon_sym_DOT] = ACTIONS(4172),
    [anon_sym_DASH_GT] = ACTIONS(4174),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4172),
    [anon_sym___typeof] = ACTIONS(4172),
    [anon_sym_typeof] = ACTIONS(4172),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4172),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4172),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4172),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4172),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4172),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4172),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4172),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4172),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4172),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4172),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4172),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4172),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4172),
    [anon_sym_API_AVAILABLE] = ACTIONS(4172),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4172),
    [anon_sym_API_DEPRECATED] = ACTIONS(4172),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4172),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4172),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4172),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4172),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4172),
    [anon_sym___deprecated_msg] = ACTIONS(4172),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4172),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4172),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4172),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4172),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4172),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4172),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4172),
    [anon_sym_BOOL] = ACTIONS(4172),
    [anon_sym_IMP] = ACTIONS(4172),
    [anon_sym_SEL] = ACTIONS(4172),
    [anon_sym_Class] = ACTIONS(4172),
    [anon_sym_id] = ACTIONS(4172),
  },
  [STATE(1638)] = {
    [sym_identifier] = ACTIONS(4282),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4284),
    [anon_sym_COMMA] = ACTIONS(4284),
    [anon_sym_RPAREN] = ACTIONS(4284),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4284),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4284),
    [anon_sym_LPAREN2] = ACTIONS(4284),
    [anon_sym_DASH] = ACTIONS(4282),
    [anon_sym_PLUS] = ACTIONS(4282),
    [anon_sym_STAR] = ACTIONS(4284),
    [anon_sym_SLASH] = ACTIONS(4282),
    [anon_sym_PERCENT] = ACTIONS(4284),
    [anon_sym_PIPE_PIPE] = ACTIONS(4284),
    [anon_sym_AMP_AMP] = ACTIONS(4284),
    [anon_sym_PIPE] = ACTIONS(4282),
    [anon_sym_CARET] = ACTIONS(4284),
    [anon_sym_AMP] = ACTIONS(4282),
    [anon_sym_EQ_EQ] = ACTIONS(4284),
    [anon_sym_BANG_EQ] = ACTIONS(4284),
    [anon_sym_GT] = ACTIONS(4282),
    [anon_sym_GT_EQ] = ACTIONS(4284),
    [anon_sym_LT_EQ] = ACTIONS(4284),
    [anon_sym_LT] = ACTIONS(4282),
    [anon_sym_LT_LT] = ACTIONS(4284),
    [anon_sym_GT_GT] = ACTIONS(4284),
    [anon_sym_SEMI] = ACTIONS(4284),
    [anon_sym___extension__] = ACTIONS(4282),
    [anon_sym_extern] = ACTIONS(4282),
    [anon_sym___attribute__] = ACTIONS(4282),
    [anon_sym___attribute] = ACTIONS(4282),
    [anon_sym_noreturn] = ACTIONS(4282),
    [anon_sym_LBRACK] = ACTIONS(4284),
    [anon_sym_RBRACK] = ACTIONS(4284),
    [anon_sym___declspec] = ACTIONS(4282),
    [anon_sym_LBRACE] = ACTIONS(4284),
    [anon_sym_RBRACE] = ACTIONS(4284),
    [anon_sym_signed] = ACTIONS(4282),
    [anon_sym_unsigned] = ACTIONS(4282),
    [anon_sym_long] = ACTIONS(4282),
    [anon_sym_short] = ACTIONS(4282),
    [anon_sym_static] = ACTIONS(4282),
    [anon_sym_ATautoreleasepool] = ACTIONS(4284),
    [anon_sym_auto] = ACTIONS(4282),
    [anon_sym_register] = ACTIONS(4282),
    [anon_sym_inline] = ACTIONS(4282),
    [anon_sym___inline] = ACTIONS(4282),
    [anon_sym___inline__] = ACTIONS(4282),
    [anon_sym___forceinline] = ACTIONS(4282),
    [anon_sym_thread_local] = ACTIONS(4282),
    [anon_sym___thread] = ACTIONS(4282),
    [anon_sym_CG_EXTERN] = ACTIONS(4282),
    [anon_sym_CG_INLINE] = ACTIONS(4282),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4282),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4282),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4282),
    [anon_sym_IBOutlet] = ACTIONS(4282),
    [anon_sym_IBInspectable] = ACTIONS(4282),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4282),
    [anon_sym_NS_INLINE] = ACTIONS(4282),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4282),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4282),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4282),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4282),
    [anon_sym_const] = ACTIONS(4282),
    [anon_sym_constexpr] = ACTIONS(4282),
    [anon_sym_volatile] = ACTIONS(4282),
    [anon_sym_restrict] = ACTIONS(4282),
    [anon_sym___restrict__] = ACTIONS(4282),
    [anon_sym__Atomic] = ACTIONS(4282),
    [anon_sym__Noreturn] = ACTIONS(4282),
    [anon_sym__Nonnull] = ACTIONS(4282),
    [anon_sym_nullable] = ACTIONS(4282),
    [anon_sym__Complex] = ACTIONS(4282),
    [anon_sym__Nullable] = ACTIONS(4282),
    [anon_sym__Nullable_result] = ACTIONS(4282),
    [anon_sym__Null_unspecified] = ACTIONS(4282),
    [anon_sym___autoreleasing] = ACTIONS(4282),
    [anon_sym___block] = ACTIONS(4282),
    [anon_sym___bridge] = ACTIONS(4282),
    [anon_sym___bridge_retained] = ACTIONS(4282),
    [anon_sym___bridge_transfer] = ACTIONS(4282),
    [anon_sym___complex] = ACTIONS(4282),
    [anon_sym___const] = ACTIONS(4282),
    [anon_sym___imag] = ACTIONS(4282),
    [anon_sym___kindof] = ACTIONS(4282),
    [anon_sym___nonnull] = ACTIONS(4282),
    [anon_sym___nullable] = ACTIONS(4282),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4282),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4282),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4282),
    [anon_sym___real] = ACTIONS(4282),
    [anon_sym___strong] = ACTIONS(4282),
    [anon_sym___unsafe_unretained] = ACTIONS(4282),
    [anon_sym___unused] = ACTIONS(4282),
    [anon_sym___weak] = ACTIONS(4282),
    [anon_sym_alignas] = ACTIONS(4282),
    [anon_sym__Alignas] = ACTIONS(4282),
    [sym_primitive_type] = ACTIONS(4282),
    [anon_sym_enum] = ACTIONS(4282),
    [anon_sym_COLON] = ACTIONS(4284),
    [anon_sym_struct] = ACTIONS(4282),
    [anon_sym_union] = ACTIONS(4282),
    [anon_sym_QMARK] = ACTIONS(4284),
    [anon_sym_DASH_DASH] = ACTIONS(4284),
    [anon_sym_PLUS_PLUS] = ACTIONS(4284),
    [anon_sym_DOT] = ACTIONS(4282),
    [anon_sym_DASH_GT] = ACTIONS(4284),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4282),
    [anon_sym___typeof] = ACTIONS(4282),
    [anon_sym_typeof] = ACTIONS(4282),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4282),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4282),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4282),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4282),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4282),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4282),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4282),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4282),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4282),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4282),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4282),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4282),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4282),
    [anon_sym_API_AVAILABLE] = ACTIONS(4282),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4282),
    [anon_sym_API_DEPRECATED] = ACTIONS(4282),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4282),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4282),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4282),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4282),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4282),
    [anon_sym___deprecated_msg] = ACTIONS(4282),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4282),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4282),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4282),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4282),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4282),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4282),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4282),
    [anon_sym_BOOL] = ACTIONS(4282),
    [anon_sym_IMP] = ACTIONS(4282),
    [anon_sym_SEL] = ACTIONS(4282),
    [anon_sym_Class] = ACTIONS(4282),
    [anon_sym_id] = ACTIONS(4282),
  },
  [STATE(1639)] = {
    [sym_identifier] = ACTIONS(4142),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4144),
    [anon_sym_COMMA] = ACTIONS(4144),
    [anon_sym_RPAREN] = ACTIONS(4144),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4144),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4144),
    [anon_sym_LPAREN2] = ACTIONS(4144),
    [anon_sym_DASH] = ACTIONS(4142),
    [anon_sym_PLUS] = ACTIONS(4142),
    [anon_sym_STAR] = ACTIONS(4144),
    [anon_sym_SLASH] = ACTIONS(4142),
    [anon_sym_PERCENT] = ACTIONS(4144),
    [anon_sym_PIPE_PIPE] = ACTIONS(4144),
    [anon_sym_AMP_AMP] = ACTIONS(4144),
    [anon_sym_PIPE] = ACTIONS(4142),
    [anon_sym_CARET] = ACTIONS(4144),
    [anon_sym_AMP] = ACTIONS(4142),
    [anon_sym_EQ_EQ] = ACTIONS(4144),
    [anon_sym_BANG_EQ] = ACTIONS(4144),
    [anon_sym_GT] = ACTIONS(4142),
    [anon_sym_GT_EQ] = ACTIONS(4144),
    [anon_sym_LT_EQ] = ACTIONS(4144),
    [anon_sym_LT] = ACTIONS(4142),
    [anon_sym_LT_LT] = ACTIONS(4144),
    [anon_sym_GT_GT] = ACTIONS(4144),
    [anon_sym_SEMI] = ACTIONS(4144),
    [anon_sym___extension__] = ACTIONS(4142),
    [anon_sym_extern] = ACTIONS(4142),
    [anon_sym___attribute__] = ACTIONS(4142),
    [anon_sym___attribute] = ACTIONS(4142),
    [anon_sym_noreturn] = ACTIONS(4142),
    [anon_sym_LBRACK] = ACTIONS(4144),
    [anon_sym_RBRACK] = ACTIONS(4144),
    [anon_sym___declspec] = ACTIONS(4142),
    [anon_sym_LBRACE] = ACTIONS(4144),
    [anon_sym_RBRACE] = ACTIONS(4144),
    [anon_sym_signed] = ACTIONS(4142),
    [anon_sym_unsigned] = ACTIONS(4142),
    [anon_sym_long] = ACTIONS(4142),
    [anon_sym_short] = ACTIONS(4142),
    [anon_sym_static] = ACTIONS(4142),
    [anon_sym_ATautoreleasepool] = ACTIONS(4144),
    [anon_sym_auto] = ACTIONS(4142),
    [anon_sym_register] = ACTIONS(4142),
    [anon_sym_inline] = ACTIONS(4142),
    [anon_sym___inline] = ACTIONS(4142),
    [anon_sym___inline__] = ACTIONS(4142),
    [anon_sym___forceinline] = ACTIONS(4142),
    [anon_sym_thread_local] = ACTIONS(4142),
    [anon_sym___thread] = ACTIONS(4142),
    [anon_sym_CG_EXTERN] = ACTIONS(4142),
    [anon_sym_CG_INLINE] = ACTIONS(4142),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4142),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4142),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4142),
    [anon_sym_IBOutlet] = ACTIONS(4142),
    [anon_sym_IBInspectable] = ACTIONS(4142),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4142),
    [anon_sym_NS_INLINE] = ACTIONS(4142),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4142),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4142),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4142),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4142),
    [anon_sym_const] = ACTIONS(4142),
    [anon_sym_constexpr] = ACTIONS(4142),
    [anon_sym_volatile] = ACTIONS(4142),
    [anon_sym_restrict] = ACTIONS(4142),
    [anon_sym___restrict__] = ACTIONS(4142),
    [anon_sym__Atomic] = ACTIONS(4142),
    [anon_sym__Noreturn] = ACTIONS(4142),
    [anon_sym__Nonnull] = ACTIONS(4142),
    [anon_sym_nullable] = ACTIONS(4142),
    [anon_sym__Complex] = ACTIONS(4142),
    [anon_sym__Nullable] = ACTIONS(4142),
    [anon_sym__Nullable_result] = ACTIONS(4142),
    [anon_sym__Null_unspecified] = ACTIONS(4142),
    [anon_sym___autoreleasing] = ACTIONS(4142),
    [anon_sym___block] = ACTIONS(4142),
    [anon_sym___bridge] = ACTIONS(4142),
    [anon_sym___bridge_retained] = ACTIONS(4142),
    [anon_sym___bridge_transfer] = ACTIONS(4142),
    [anon_sym___complex] = ACTIONS(4142),
    [anon_sym___const] = ACTIONS(4142),
    [anon_sym___imag] = ACTIONS(4142),
    [anon_sym___kindof] = ACTIONS(4142),
    [anon_sym___nonnull] = ACTIONS(4142),
    [anon_sym___nullable] = ACTIONS(4142),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4142),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4142),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4142),
    [anon_sym___real] = ACTIONS(4142),
    [anon_sym___strong] = ACTIONS(4142),
    [anon_sym___unsafe_unretained] = ACTIONS(4142),
    [anon_sym___unused] = ACTIONS(4142),
    [anon_sym___weak] = ACTIONS(4142),
    [anon_sym_alignas] = ACTIONS(4142),
    [anon_sym__Alignas] = ACTIONS(4142),
    [sym_primitive_type] = ACTIONS(4142),
    [anon_sym_enum] = ACTIONS(4142),
    [anon_sym_COLON] = ACTIONS(4144),
    [anon_sym_struct] = ACTIONS(4142),
    [anon_sym_union] = ACTIONS(4142),
    [anon_sym_QMARK] = ACTIONS(4144),
    [anon_sym_DASH_DASH] = ACTIONS(4144),
    [anon_sym_PLUS_PLUS] = ACTIONS(4144),
    [anon_sym_DOT] = ACTIONS(4142),
    [anon_sym_DASH_GT] = ACTIONS(4144),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4142),
    [anon_sym___typeof] = ACTIONS(4142),
    [anon_sym_typeof] = ACTIONS(4142),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4142),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4142),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4142),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4142),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4142),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4142),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4142),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4142),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4142),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4142),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4142),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4142),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4142),
    [anon_sym_API_AVAILABLE] = ACTIONS(4142),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4142),
    [anon_sym_API_DEPRECATED] = ACTIONS(4142),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4142),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4142),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4142),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4142),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4142),
    [anon_sym___deprecated_msg] = ACTIONS(4142),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4142),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4142),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4142),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4142),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4142),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4142),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4142),
    [anon_sym_BOOL] = ACTIONS(4142),
    [anon_sym_IMP] = ACTIONS(4142),
    [anon_sym_SEL] = ACTIONS(4142),
    [anon_sym_Class] = ACTIONS(4142),
    [anon_sym_id] = ACTIONS(4142),
  },
  [STATE(1640)] = {
    [sym_identifier] = ACTIONS(4176),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4178),
    [anon_sym_COMMA] = ACTIONS(4178),
    [anon_sym_RPAREN] = ACTIONS(4178),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4178),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4178),
    [anon_sym_LPAREN2] = ACTIONS(4178),
    [anon_sym_DASH] = ACTIONS(4176),
    [anon_sym_PLUS] = ACTIONS(4176),
    [anon_sym_STAR] = ACTIONS(4178),
    [anon_sym_SLASH] = ACTIONS(4176),
    [anon_sym_PERCENT] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE] = ACTIONS(4178),
    [anon_sym_AMP_AMP] = ACTIONS(4178),
    [anon_sym_PIPE] = ACTIONS(4176),
    [anon_sym_CARET] = ACTIONS(4178),
    [anon_sym_AMP] = ACTIONS(4176),
    [anon_sym_EQ_EQ] = ACTIONS(4178),
    [anon_sym_BANG_EQ] = ACTIONS(4178),
    [anon_sym_GT] = ACTIONS(4176),
    [anon_sym_GT_EQ] = ACTIONS(4178),
    [anon_sym_LT_EQ] = ACTIONS(4178),
    [anon_sym_LT] = ACTIONS(4176),
    [anon_sym_LT_LT] = ACTIONS(4178),
    [anon_sym_GT_GT] = ACTIONS(4178),
    [anon_sym_SEMI] = ACTIONS(4178),
    [anon_sym___extension__] = ACTIONS(4176),
    [anon_sym_extern] = ACTIONS(4176),
    [anon_sym___attribute__] = ACTIONS(4176),
    [anon_sym___attribute] = ACTIONS(4176),
    [anon_sym_noreturn] = ACTIONS(4176),
    [anon_sym_LBRACK] = ACTIONS(4178),
    [anon_sym_RBRACK] = ACTIONS(4178),
    [anon_sym___declspec] = ACTIONS(4176),
    [anon_sym_LBRACE] = ACTIONS(4178),
    [anon_sym_RBRACE] = ACTIONS(4178),
    [anon_sym_signed] = ACTIONS(4176),
    [anon_sym_unsigned] = ACTIONS(4176),
    [anon_sym_long] = ACTIONS(4176),
    [anon_sym_short] = ACTIONS(4176),
    [anon_sym_static] = ACTIONS(4176),
    [anon_sym_ATautoreleasepool] = ACTIONS(4178),
    [anon_sym_auto] = ACTIONS(4176),
    [anon_sym_register] = ACTIONS(4176),
    [anon_sym_inline] = ACTIONS(4176),
    [anon_sym___inline] = ACTIONS(4176),
    [anon_sym___inline__] = ACTIONS(4176),
    [anon_sym___forceinline] = ACTIONS(4176),
    [anon_sym_thread_local] = ACTIONS(4176),
    [anon_sym___thread] = ACTIONS(4176),
    [anon_sym_CG_EXTERN] = ACTIONS(4176),
    [anon_sym_CG_INLINE] = ACTIONS(4176),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4176),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4176),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4176),
    [anon_sym_IBOutlet] = ACTIONS(4176),
    [anon_sym_IBInspectable] = ACTIONS(4176),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4176),
    [anon_sym_NS_INLINE] = ACTIONS(4176),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4176),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4176),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4176),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4176),
    [anon_sym_const] = ACTIONS(4176),
    [anon_sym_constexpr] = ACTIONS(4176),
    [anon_sym_volatile] = ACTIONS(4176),
    [anon_sym_restrict] = ACTIONS(4176),
    [anon_sym___restrict__] = ACTIONS(4176),
    [anon_sym__Atomic] = ACTIONS(4176),
    [anon_sym__Noreturn] = ACTIONS(4176),
    [anon_sym__Nonnull] = ACTIONS(4176),
    [anon_sym_nullable] = ACTIONS(4176),
    [anon_sym__Complex] = ACTIONS(4176),
    [anon_sym__Nullable] = ACTIONS(4176),
    [anon_sym__Nullable_result] = ACTIONS(4176),
    [anon_sym__Null_unspecified] = ACTIONS(4176),
    [anon_sym___autoreleasing] = ACTIONS(4176),
    [anon_sym___block] = ACTIONS(4176),
    [anon_sym___bridge] = ACTIONS(4176),
    [anon_sym___bridge_retained] = ACTIONS(4176),
    [anon_sym___bridge_transfer] = ACTIONS(4176),
    [anon_sym___complex] = ACTIONS(4176),
    [anon_sym___const] = ACTIONS(4176),
    [anon_sym___imag] = ACTIONS(4176),
    [anon_sym___kindof] = ACTIONS(4176),
    [anon_sym___nonnull] = ACTIONS(4176),
    [anon_sym___nullable] = ACTIONS(4176),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4176),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4176),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4176),
    [anon_sym___real] = ACTIONS(4176),
    [anon_sym___strong] = ACTIONS(4176),
    [anon_sym___unsafe_unretained] = ACTIONS(4176),
    [anon_sym___unused] = ACTIONS(4176),
    [anon_sym___weak] = ACTIONS(4176),
    [anon_sym_alignas] = ACTIONS(4176),
    [anon_sym__Alignas] = ACTIONS(4176),
    [sym_primitive_type] = ACTIONS(4176),
    [anon_sym_enum] = ACTIONS(4176),
    [anon_sym_COLON] = ACTIONS(4178),
    [anon_sym_struct] = ACTIONS(4176),
    [anon_sym_union] = ACTIONS(4176),
    [anon_sym_QMARK] = ACTIONS(4178),
    [anon_sym_DASH_DASH] = ACTIONS(4178),
    [anon_sym_PLUS_PLUS] = ACTIONS(4178),
    [anon_sym_DOT] = ACTIONS(4176),
    [anon_sym_DASH_GT] = ACTIONS(4178),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4176),
    [anon_sym___typeof] = ACTIONS(4176),
    [anon_sym_typeof] = ACTIONS(4176),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4176),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4176),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4176),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4176),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4176),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4176),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4176),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4176),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4176),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4176),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4176),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4176),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4176),
    [anon_sym_API_AVAILABLE] = ACTIONS(4176),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4176),
    [anon_sym_API_DEPRECATED] = ACTIONS(4176),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4176),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4176),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4176),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4176),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4176),
    [anon_sym___deprecated_msg] = ACTIONS(4176),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4176),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4176),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4176),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4176),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4176),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4176),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4176),
    [anon_sym_BOOL] = ACTIONS(4176),
    [anon_sym_IMP] = ACTIONS(4176),
    [anon_sym_SEL] = ACTIONS(4176),
    [anon_sym_Class] = ACTIONS(4176),
    [anon_sym_id] = ACTIONS(4176),
  },
  [STATE(1641)] = {
    [sym_identifier] = ACTIONS(4024),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_RPAREN] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4026),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4026),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4026),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4026),
    [anon_sym_GT_GT] = ACTIONS(4026),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4024),
    [anon_sym_extern] = ACTIONS(4024),
    [anon_sym___attribute__] = ACTIONS(4024),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4024),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym_RBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4024),
    [anon_sym_LBRACE] = ACTIONS(4026),
    [anon_sym_RBRACE] = ACTIONS(4026),
    [anon_sym_signed] = ACTIONS(4024),
    [anon_sym_unsigned] = ACTIONS(4024),
    [anon_sym_long] = ACTIONS(4024),
    [anon_sym_short] = ACTIONS(4024),
    [anon_sym_static] = ACTIONS(4024),
    [anon_sym_ATautoreleasepool] = ACTIONS(4026),
    [anon_sym_auto] = ACTIONS(4024),
    [anon_sym_register] = ACTIONS(4024),
    [anon_sym_inline] = ACTIONS(4024),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4024),
    [anon_sym___forceinline] = ACTIONS(4024),
    [anon_sym_thread_local] = ACTIONS(4024),
    [anon_sym___thread] = ACTIONS(4024),
    [anon_sym_CG_EXTERN] = ACTIONS(4024),
    [anon_sym_CG_INLINE] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4024),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4024),
    [anon_sym_IBOutlet] = ACTIONS(4024),
    [anon_sym_IBInspectable] = ACTIONS(4024),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4024),
    [anon_sym_NS_INLINE] = ACTIONS(4024),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4024),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4024),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4024),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4024),
    [anon_sym_volatile] = ACTIONS(4024),
    [anon_sym_restrict] = ACTIONS(4024),
    [anon_sym___restrict__] = ACTIONS(4024),
    [anon_sym__Atomic] = ACTIONS(4024),
    [anon_sym__Noreturn] = ACTIONS(4024),
    [anon_sym__Nonnull] = ACTIONS(4024),
    [anon_sym_nullable] = ACTIONS(4024),
    [anon_sym__Complex] = ACTIONS(4024),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4024),
    [anon_sym__Null_unspecified] = ACTIONS(4024),
    [anon_sym___autoreleasing] = ACTIONS(4024),
    [anon_sym___block] = ACTIONS(4024),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4024),
    [anon_sym___bridge_transfer] = ACTIONS(4024),
    [anon_sym___complex] = ACTIONS(4024),
    [anon_sym___const] = ACTIONS(4024),
    [anon_sym___imag] = ACTIONS(4024),
    [anon_sym___kindof] = ACTIONS(4024),
    [anon_sym___nonnull] = ACTIONS(4024),
    [anon_sym___nullable] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4024),
    [anon_sym___real] = ACTIONS(4024),
    [anon_sym___strong] = ACTIONS(4024),
    [anon_sym___unsafe_unretained] = ACTIONS(4024),
    [anon_sym___unused] = ACTIONS(4024),
    [anon_sym___weak] = ACTIONS(4024),
    [anon_sym_alignas] = ACTIONS(4024),
    [anon_sym__Alignas] = ACTIONS(4024),
    [sym_primitive_type] = ACTIONS(4024),
    [anon_sym_enum] = ACTIONS(4024),
    [anon_sym_COLON] = ACTIONS(4026),
    [anon_sym_struct] = ACTIONS(4024),
    [anon_sym_union] = ACTIONS(4024),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4024),
    [anon_sym___typeof] = ACTIONS(4024),
    [anon_sym_typeof] = ACTIONS(4024),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4024),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4024),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_API_AVAILABLE] = ACTIONS(4024),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_API_DEPRECATED] = ACTIONS(4024),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym___deprecated_msg] = ACTIONS(4024),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4024),
    [anon_sym_BOOL] = ACTIONS(4024),
    [anon_sym_IMP] = ACTIONS(4024),
    [anon_sym_SEL] = ACTIONS(4024),
    [anon_sym_Class] = ACTIONS(4024),
    [anon_sym_id] = ACTIONS(4024),
  },
  [STATE(1642)] = {
    [sym_identifier] = ACTIONS(4038),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4040),
    [anon_sym_COMMA] = ACTIONS(4040),
    [anon_sym_RPAREN] = ACTIONS(4040),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4040),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4040),
    [anon_sym_LPAREN2] = ACTIONS(4040),
    [anon_sym_DASH] = ACTIONS(4038),
    [anon_sym_PLUS] = ACTIONS(4038),
    [anon_sym_STAR] = ACTIONS(4040),
    [anon_sym_SLASH] = ACTIONS(4038),
    [anon_sym_PERCENT] = ACTIONS(4040),
    [anon_sym_PIPE_PIPE] = ACTIONS(4040),
    [anon_sym_AMP_AMP] = ACTIONS(4040),
    [anon_sym_PIPE] = ACTIONS(4038),
    [anon_sym_CARET] = ACTIONS(4040),
    [anon_sym_AMP] = ACTIONS(4038),
    [anon_sym_EQ_EQ] = ACTIONS(4040),
    [anon_sym_BANG_EQ] = ACTIONS(4040),
    [anon_sym_GT] = ACTIONS(4038),
    [anon_sym_GT_EQ] = ACTIONS(4040),
    [anon_sym_LT_EQ] = ACTIONS(4040),
    [anon_sym_LT] = ACTIONS(4038),
    [anon_sym_LT_LT] = ACTIONS(4040),
    [anon_sym_GT_GT] = ACTIONS(4040),
    [anon_sym_SEMI] = ACTIONS(4040),
    [anon_sym___extension__] = ACTIONS(4038),
    [anon_sym_extern] = ACTIONS(4038),
    [anon_sym___attribute__] = ACTIONS(4038),
    [anon_sym___attribute] = ACTIONS(4038),
    [anon_sym_noreturn] = ACTIONS(4038),
    [anon_sym_LBRACK] = ACTIONS(4040),
    [anon_sym_RBRACK] = ACTIONS(4040),
    [anon_sym___declspec] = ACTIONS(4038),
    [anon_sym_LBRACE] = ACTIONS(4040),
    [anon_sym_RBRACE] = ACTIONS(4040),
    [anon_sym_signed] = ACTIONS(4038),
    [anon_sym_unsigned] = ACTIONS(4038),
    [anon_sym_long] = ACTIONS(4038),
    [anon_sym_short] = ACTIONS(4038),
    [anon_sym_static] = ACTIONS(4038),
    [anon_sym_ATautoreleasepool] = ACTIONS(4040),
    [anon_sym_auto] = ACTIONS(4038),
    [anon_sym_register] = ACTIONS(4038),
    [anon_sym_inline] = ACTIONS(4038),
    [anon_sym___inline] = ACTIONS(4038),
    [anon_sym___inline__] = ACTIONS(4038),
    [anon_sym___forceinline] = ACTIONS(4038),
    [anon_sym_thread_local] = ACTIONS(4038),
    [anon_sym___thread] = ACTIONS(4038),
    [anon_sym_CG_EXTERN] = ACTIONS(4038),
    [anon_sym_CG_INLINE] = ACTIONS(4038),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4038),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4038),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4038),
    [anon_sym_IBOutlet] = ACTIONS(4038),
    [anon_sym_IBInspectable] = ACTIONS(4038),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4038),
    [anon_sym_NS_INLINE] = ACTIONS(4038),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4038),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4038),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4038),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4038),
    [anon_sym_const] = ACTIONS(4038),
    [anon_sym_constexpr] = ACTIONS(4038),
    [anon_sym_volatile] = ACTIONS(4038),
    [anon_sym_restrict] = ACTIONS(4038),
    [anon_sym___restrict__] = ACTIONS(4038),
    [anon_sym__Atomic] = ACTIONS(4038),
    [anon_sym__Noreturn] = ACTIONS(4038),
    [anon_sym__Nonnull] = ACTIONS(4038),
    [anon_sym_nullable] = ACTIONS(4038),
    [anon_sym__Complex] = ACTIONS(4038),
    [anon_sym__Nullable] = ACTIONS(4038),
    [anon_sym__Nullable_result] = ACTIONS(4038),
    [anon_sym__Null_unspecified] = ACTIONS(4038),
    [anon_sym___autoreleasing] = ACTIONS(4038),
    [anon_sym___block] = ACTIONS(4038),
    [anon_sym___bridge] = ACTIONS(4038),
    [anon_sym___bridge_retained] = ACTIONS(4038),
    [anon_sym___bridge_transfer] = ACTIONS(4038),
    [anon_sym___complex] = ACTIONS(4038),
    [anon_sym___const] = ACTIONS(4038),
    [anon_sym___imag] = ACTIONS(4038),
    [anon_sym___kindof] = ACTIONS(4038),
    [anon_sym___nonnull] = ACTIONS(4038),
    [anon_sym___nullable] = ACTIONS(4038),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4038),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4038),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4038),
    [anon_sym___real] = ACTIONS(4038),
    [anon_sym___strong] = ACTIONS(4038),
    [anon_sym___unsafe_unretained] = ACTIONS(4038),
    [anon_sym___unused] = ACTIONS(4038),
    [anon_sym___weak] = ACTIONS(4038),
    [anon_sym_alignas] = ACTIONS(4038),
    [anon_sym__Alignas] = ACTIONS(4038),
    [sym_primitive_type] = ACTIONS(4038),
    [anon_sym_enum] = ACTIONS(4038),
    [anon_sym_COLON] = ACTIONS(4040),
    [anon_sym_struct] = ACTIONS(4038),
    [anon_sym_union] = ACTIONS(4038),
    [anon_sym_QMARK] = ACTIONS(4040),
    [anon_sym_DASH_DASH] = ACTIONS(4040),
    [anon_sym_PLUS_PLUS] = ACTIONS(4040),
    [anon_sym_DOT] = ACTIONS(4038),
    [anon_sym_DASH_GT] = ACTIONS(4040),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4038),
    [anon_sym___typeof] = ACTIONS(4038),
    [anon_sym_typeof] = ACTIONS(4038),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4038),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4038),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4038),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4038),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4038),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4038),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4038),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4038),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4038),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4038),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4038),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4038),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4038),
    [anon_sym_API_AVAILABLE] = ACTIONS(4038),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4038),
    [anon_sym_API_DEPRECATED] = ACTIONS(4038),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4038),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4038),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4038),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4038),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4038),
    [anon_sym___deprecated_msg] = ACTIONS(4038),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4038),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4038),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4038),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4038),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4038),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4038),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4038),
    [anon_sym_BOOL] = ACTIONS(4038),
    [anon_sym_IMP] = ACTIONS(4038),
    [anon_sym_SEL] = ACTIONS(4038),
    [anon_sym_Class] = ACTIONS(4038),
    [anon_sym_id] = ACTIONS(4038),
  },
  [STATE(1643)] = {
    [sym_identifier] = ACTIONS(4066),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4068),
    [anon_sym_COMMA] = ACTIONS(4068),
    [anon_sym_RPAREN] = ACTIONS(4068),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4068),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4068),
    [anon_sym_LPAREN2] = ACTIONS(4068),
    [anon_sym_DASH] = ACTIONS(4066),
    [anon_sym_PLUS] = ACTIONS(4066),
    [anon_sym_STAR] = ACTIONS(4068),
    [anon_sym_SLASH] = ACTIONS(4066),
    [anon_sym_PERCENT] = ACTIONS(4068),
    [anon_sym_PIPE_PIPE] = ACTIONS(4068),
    [anon_sym_AMP_AMP] = ACTIONS(4068),
    [anon_sym_PIPE] = ACTIONS(4066),
    [anon_sym_CARET] = ACTIONS(4068),
    [anon_sym_AMP] = ACTIONS(4066),
    [anon_sym_EQ_EQ] = ACTIONS(4068),
    [anon_sym_BANG_EQ] = ACTIONS(4068),
    [anon_sym_GT] = ACTIONS(4066),
    [anon_sym_GT_EQ] = ACTIONS(4068),
    [anon_sym_LT_EQ] = ACTIONS(4068),
    [anon_sym_LT] = ACTIONS(4066),
    [anon_sym_LT_LT] = ACTIONS(4068),
    [anon_sym_GT_GT] = ACTIONS(4068),
    [anon_sym_SEMI] = ACTIONS(4068),
    [anon_sym___extension__] = ACTIONS(4066),
    [anon_sym_extern] = ACTIONS(4066),
    [anon_sym___attribute__] = ACTIONS(4066),
    [anon_sym___attribute] = ACTIONS(4066),
    [anon_sym_noreturn] = ACTIONS(4066),
    [anon_sym_LBRACK] = ACTIONS(4068),
    [anon_sym_RBRACK] = ACTIONS(4068),
    [anon_sym___declspec] = ACTIONS(4066),
    [anon_sym_LBRACE] = ACTIONS(4068),
    [anon_sym_RBRACE] = ACTIONS(4068),
    [anon_sym_signed] = ACTIONS(4066),
    [anon_sym_unsigned] = ACTIONS(4066),
    [anon_sym_long] = ACTIONS(4066),
    [anon_sym_short] = ACTIONS(4066),
    [anon_sym_static] = ACTIONS(4066),
    [anon_sym_ATautoreleasepool] = ACTIONS(4068),
    [anon_sym_auto] = ACTIONS(4066),
    [anon_sym_register] = ACTIONS(4066),
    [anon_sym_inline] = ACTIONS(4066),
    [anon_sym___inline] = ACTIONS(4066),
    [anon_sym___inline__] = ACTIONS(4066),
    [anon_sym___forceinline] = ACTIONS(4066),
    [anon_sym_thread_local] = ACTIONS(4066),
    [anon_sym___thread] = ACTIONS(4066),
    [anon_sym_CG_EXTERN] = ACTIONS(4066),
    [anon_sym_CG_INLINE] = ACTIONS(4066),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4066),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4066),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4066),
    [anon_sym_IBOutlet] = ACTIONS(4066),
    [anon_sym_IBInspectable] = ACTIONS(4066),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4066),
    [anon_sym_NS_INLINE] = ACTIONS(4066),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4066),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4066),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4066),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4066),
    [anon_sym_const] = ACTIONS(4066),
    [anon_sym_constexpr] = ACTIONS(4066),
    [anon_sym_volatile] = ACTIONS(4066),
    [anon_sym_restrict] = ACTIONS(4066),
    [anon_sym___restrict__] = ACTIONS(4066),
    [anon_sym__Atomic] = ACTIONS(4066),
    [anon_sym__Noreturn] = ACTIONS(4066),
    [anon_sym__Nonnull] = ACTIONS(4066),
    [anon_sym_nullable] = ACTIONS(4066),
    [anon_sym__Complex] = ACTIONS(4066),
    [anon_sym__Nullable] = ACTIONS(4066),
    [anon_sym__Nullable_result] = ACTIONS(4066),
    [anon_sym__Null_unspecified] = ACTIONS(4066),
    [anon_sym___autoreleasing] = ACTIONS(4066),
    [anon_sym___block] = ACTIONS(4066),
    [anon_sym___bridge] = ACTIONS(4066),
    [anon_sym___bridge_retained] = ACTIONS(4066),
    [anon_sym___bridge_transfer] = ACTIONS(4066),
    [anon_sym___complex] = ACTIONS(4066),
    [anon_sym___const] = ACTIONS(4066),
    [anon_sym___imag] = ACTIONS(4066),
    [anon_sym___kindof] = ACTIONS(4066),
    [anon_sym___nonnull] = ACTIONS(4066),
    [anon_sym___nullable] = ACTIONS(4066),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4066),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4066),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4066),
    [anon_sym___real] = ACTIONS(4066),
    [anon_sym___strong] = ACTIONS(4066),
    [anon_sym___unsafe_unretained] = ACTIONS(4066),
    [anon_sym___unused] = ACTIONS(4066),
    [anon_sym___weak] = ACTIONS(4066),
    [anon_sym_alignas] = ACTIONS(4066),
    [anon_sym__Alignas] = ACTIONS(4066),
    [sym_primitive_type] = ACTIONS(4066),
    [anon_sym_enum] = ACTIONS(4066),
    [anon_sym_COLON] = ACTIONS(4068),
    [anon_sym_struct] = ACTIONS(4066),
    [anon_sym_union] = ACTIONS(4066),
    [anon_sym_QMARK] = ACTIONS(4068),
    [anon_sym_DASH_DASH] = ACTIONS(4068),
    [anon_sym_PLUS_PLUS] = ACTIONS(4068),
    [anon_sym_DOT] = ACTIONS(4066),
    [anon_sym_DASH_GT] = ACTIONS(4068),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4066),
    [anon_sym___typeof] = ACTIONS(4066),
    [anon_sym_typeof] = ACTIONS(4066),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4066),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4066),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4066),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4066),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4066),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4066),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4066),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4066),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4066),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4066),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4066),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4066),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4066),
    [anon_sym_API_AVAILABLE] = ACTIONS(4066),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4066),
    [anon_sym_API_DEPRECATED] = ACTIONS(4066),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4066),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4066),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4066),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4066),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4066),
    [anon_sym___deprecated_msg] = ACTIONS(4066),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4066),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4066),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4066),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4066),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4066),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4066),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4066),
    [anon_sym_BOOL] = ACTIONS(4066),
    [anon_sym_IMP] = ACTIONS(4066),
    [anon_sym_SEL] = ACTIONS(4066),
    [anon_sym_Class] = ACTIONS(4066),
    [anon_sym_id] = ACTIONS(4066),
  },
  [STATE(1644)] = {
    [sym_identifier] = ACTIONS(4070),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4072),
    [anon_sym_COMMA] = ACTIONS(4072),
    [anon_sym_RPAREN] = ACTIONS(4072),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4072),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4072),
    [anon_sym_LPAREN2] = ACTIONS(4072),
    [anon_sym_DASH] = ACTIONS(4070),
    [anon_sym_PLUS] = ACTIONS(4070),
    [anon_sym_STAR] = ACTIONS(4072),
    [anon_sym_SLASH] = ACTIONS(4070),
    [anon_sym_PERCENT] = ACTIONS(4072),
    [anon_sym_PIPE_PIPE] = ACTIONS(4072),
    [anon_sym_AMP_AMP] = ACTIONS(4072),
    [anon_sym_PIPE] = ACTIONS(4070),
    [anon_sym_CARET] = ACTIONS(4072),
    [anon_sym_AMP] = ACTIONS(4070),
    [anon_sym_EQ_EQ] = ACTIONS(4072),
    [anon_sym_BANG_EQ] = ACTIONS(4072),
    [anon_sym_GT] = ACTIONS(4070),
    [anon_sym_GT_EQ] = ACTIONS(4072),
    [anon_sym_LT_EQ] = ACTIONS(4072),
    [anon_sym_LT] = ACTIONS(4070),
    [anon_sym_LT_LT] = ACTIONS(4072),
    [anon_sym_GT_GT] = ACTIONS(4072),
    [anon_sym_SEMI] = ACTIONS(4072),
    [anon_sym___extension__] = ACTIONS(4070),
    [anon_sym_extern] = ACTIONS(4070),
    [anon_sym___attribute__] = ACTIONS(4070),
    [anon_sym___attribute] = ACTIONS(4070),
    [anon_sym_noreturn] = ACTIONS(4070),
    [anon_sym_LBRACK] = ACTIONS(4072),
    [anon_sym_RBRACK] = ACTIONS(4072),
    [anon_sym___declspec] = ACTIONS(4070),
    [anon_sym_LBRACE] = ACTIONS(4072),
    [anon_sym_RBRACE] = ACTIONS(4072),
    [anon_sym_signed] = ACTIONS(4070),
    [anon_sym_unsigned] = ACTIONS(4070),
    [anon_sym_long] = ACTIONS(4070),
    [anon_sym_short] = ACTIONS(4070),
    [anon_sym_static] = ACTIONS(4070),
    [anon_sym_ATautoreleasepool] = ACTIONS(4072),
    [anon_sym_auto] = ACTIONS(4070),
    [anon_sym_register] = ACTIONS(4070),
    [anon_sym_inline] = ACTIONS(4070),
    [anon_sym___inline] = ACTIONS(4070),
    [anon_sym___inline__] = ACTIONS(4070),
    [anon_sym___forceinline] = ACTIONS(4070),
    [anon_sym_thread_local] = ACTIONS(4070),
    [anon_sym___thread] = ACTIONS(4070),
    [anon_sym_CG_EXTERN] = ACTIONS(4070),
    [anon_sym_CG_INLINE] = ACTIONS(4070),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4070),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4070),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4070),
    [anon_sym_IBOutlet] = ACTIONS(4070),
    [anon_sym_IBInspectable] = ACTIONS(4070),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4070),
    [anon_sym_NS_INLINE] = ACTIONS(4070),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4070),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4070),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4070),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4070),
    [anon_sym_const] = ACTIONS(4070),
    [anon_sym_constexpr] = ACTIONS(4070),
    [anon_sym_volatile] = ACTIONS(4070),
    [anon_sym_restrict] = ACTIONS(4070),
    [anon_sym___restrict__] = ACTIONS(4070),
    [anon_sym__Atomic] = ACTIONS(4070),
    [anon_sym__Noreturn] = ACTIONS(4070),
    [anon_sym__Nonnull] = ACTIONS(4070),
    [anon_sym_nullable] = ACTIONS(4070),
    [anon_sym__Complex] = ACTIONS(4070),
    [anon_sym__Nullable] = ACTIONS(4070),
    [anon_sym__Nullable_result] = ACTIONS(4070),
    [anon_sym__Null_unspecified] = ACTIONS(4070),
    [anon_sym___autoreleasing] = ACTIONS(4070),
    [anon_sym___block] = ACTIONS(4070),
    [anon_sym___bridge] = ACTIONS(4070),
    [anon_sym___bridge_retained] = ACTIONS(4070),
    [anon_sym___bridge_transfer] = ACTIONS(4070),
    [anon_sym___complex] = ACTIONS(4070),
    [anon_sym___const] = ACTIONS(4070),
    [anon_sym___imag] = ACTIONS(4070),
    [anon_sym___kindof] = ACTIONS(4070),
    [anon_sym___nonnull] = ACTIONS(4070),
    [anon_sym___nullable] = ACTIONS(4070),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4070),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4070),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4070),
    [anon_sym___real] = ACTIONS(4070),
    [anon_sym___strong] = ACTIONS(4070),
    [anon_sym___unsafe_unretained] = ACTIONS(4070),
    [anon_sym___unused] = ACTIONS(4070),
    [anon_sym___weak] = ACTIONS(4070),
    [anon_sym_alignas] = ACTIONS(4070),
    [anon_sym__Alignas] = ACTIONS(4070),
    [sym_primitive_type] = ACTIONS(4070),
    [anon_sym_enum] = ACTIONS(4070),
    [anon_sym_COLON] = ACTIONS(4072),
    [anon_sym_struct] = ACTIONS(4070),
    [anon_sym_union] = ACTIONS(4070),
    [anon_sym_QMARK] = ACTIONS(4072),
    [anon_sym_DASH_DASH] = ACTIONS(4072),
    [anon_sym_PLUS_PLUS] = ACTIONS(4072),
    [anon_sym_DOT] = ACTIONS(4070),
    [anon_sym_DASH_GT] = ACTIONS(4072),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4070),
    [anon_sym___typeof] = ACTIONS(4070),
    [anon_sym_typeof] = ACTIONS(4070),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4070),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4070),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4070),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4070),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4070),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4070),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4070),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4070),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4070),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4070),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4070),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4070),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4070),
    [anon_sym_API_AVAILABLE] = ACTIONS(4070),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4070),
    [anon_sym_API_DEPRECATED] = ACTIONS(4070),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4070),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4070),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4070),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4070),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4070),
    [anon_sym___deprecated_msg] = ACTIONS(4070),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4070),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4070),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4070),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4070),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4070),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4070),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4070),
    [anon_sym_BOOL] = ACTIONS(4070),
    [anon_sym_IMP] = ACTIONS(4070),
    [anon_sym_SEL] = ACTIONS(4070),
    [anon_sym_Class] = ACTIONS(4070),
    [anon_sym_id] = ACTIONS(4070),
  },
  [STATE(1645)] = {
    [sym_identifier] = ACTIONS(4278),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4280),
    [anon_sym_COMMA] = ACTIONS(4280),
    [anon_sym_RPAREN] = ACTIONS(4280),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4280),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4280),
    [anon_sym_LPAREN2] = ACTIONS(4280),
    [anon_sym_DASH] = ACTIONS(4278),
    [anon_sym_PLUS] = ACTIONS(4278),
    [anon_sym_STAR] = ACTIONS(4280),
    [anon_sym_SLASH] = ACTIONS(4278),
    [anon_sym_PERCENT] = ACTIONS(4280),
    [anon_sym_PIPE_PIPE] = ACTIONS(4280),
    [anon_sym_AMP_AMP] = ACTIONS(4280),
    [anon_sym_PIPE] = ACTIONS(4278),
    [anon_sym_CARET] = ACTIONS(4280),
    [anon_sym_AMP] = ACTIONS(4278),
    [anon_sym_EQ_EQ] = ACTIONS(4280),
    [anon_sym_BANG_EQ] = ACTIONS(4280),
    [anon_sym_GT] = ACTIONS(4278),
    [anon_sym_GT_EQ] = ACTIONS(4280),
    [anon_sym_LT_EQ] = ACTIONS(4280),
    [anon_sym_LT] = ACTIONS(4278),
    [anon_sym_LT_LT] = ACTIONS(4280),
    [anon_sym_GT_GT] = ACTIONS(4280),
    [anon_sym_SEMI] = ACTIONS(4280),
    [anon_sym___extension__] = ACTIONS(4278),
    [anon_sym_extern] = ACTIONS(4278),
    [anon_sym___attribute__] = ACTIONS(4278),
    [anon_sym___attribute] = ACTIONS(4278),
    [anon_sym_noreturn] = ACTIONS(4278),
    [anon_sym_LBRACK] = ACTIONS(4280),
    [anon_sym_RBRACK] = ACTIONS(4280),
    [anon_sym___declspec] = ACTIONS(4278),
    [anon_sym_LBRACE] = ACTIONS(4280),
    [anon_sym_RBRACE] = ACTIONS(4280),
    [anon_sym_signed] = ACTIONS(4278),
    [anon_sym_unsigned] = ACTIONS(4278),
    [anon_sym_long] = ACTIONS(4278),
    [anon_sym_short] = ACTIONS(4278),
    [anon_sym_static] = ACTIONS(4278),
    [anon_sym_ATautoreleasepool] = ACTIONS(4280),
    [anon_sym_auto] = ACTIONS(4278),
    [anon_sym_register] = ACTIONS(4278),
    [anon_sym_inline] = ACTIONS(4278),
    [anon_sym___inline] = ACTIONS(4278),
    [anon_sym___inline__] = ACTIONS(4278),
    [anon_sym___forceinline] = ACTIONS(4278),
    [anon_sym_thread_local] = ACTIONS(4278),
    [anon_sym___thread] = ACTIONS(4278),
    [anon_sym_CG_EXTERN] = ACTIONS(4278),
    [anon_sym_CG_INLINE] = ACTIONS(4278),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4278),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4278),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4278),
    [anon_sym_IBOutlet] = ACTIONS(4278),
    [anon_sym_IBInspectable] = ACTIONS(4278),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4278),
    [anon_sym_NS_INLINE] = ACTIONS(4278),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4278),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4278),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4278),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4278),
    [anon_sym_const] = ACTIONS(4278),
    [anon_sym_constexpr] = ACTIONS(4278),
    [anon_sym_volatile] = ACTIONS(4278),
    [anon_sym_restrict] = ACTIONS(4278),
    [anon_sym___restrict__] = ACTIONS(4278),
    [anon_sym__Atomic] = ACTIONS(4278),
    [anon_sym__Noreturn] = ACTIONS(4278),
    [anon_sym__Nonnull] = ACTIONS(4278),
    [anon_sym_nullable] = ACTIONS(4278),
    [anon_sym__Complex] = ACTIONS(4278),
    [anon_sym__Nullable] = ACTIONS(4278),
    [anon_sym__Nullable_result] = ACTIONS(4278),
    [anon_sym__Null_unspecified] = ACTIONS(4278),
    [anon_sym___autoreleasing] = ACTIONS(4278),
    [anon_sym___block] = ACTIONS(4278),
    [anon_sym___bridge] = ACTIONS(4278),
    [anon_sym___bridge_retained] = ACTIONS(4278),
    [anon_sym___bridge_transfer] = ACTIONS(4278),
    [anon_sym___complex] = ACTIONS(4278),
    [anon_sym___const] = ACTIONS(4278),
    [anon_sym___imag] = ACTIONS(4278),
    [anon_sym___kindof] = ACTIONS(4278),
    [anon_sym___nonnull] = ACTIONS(4278),
    [anon_sym___nullable] = ACTIONS(4278),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4278),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4278),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4278),
    [anon_sym___real] = ACTIONS(4278),
    [anon_sym___strong] = ACTIONS(4278),
    [anon_sym___unsafe_unretained] = ACTIONS(4278),
    [anon_sym___unused] = ACTIONS(4278),
    [anon_sym___weak] = ACTIONS(4278),
    [anon_sym_alignas] = ACTIONS(4278),
    [anon_sym__Alignas] = ACTIONS(4278),
    [sym_primitive_type] = ACTIONS(4278),
    [anon_sym_enum] = ACTIONS(4278),
    [anon_sym_COLON] = ACTIONS(4280),
    [anon_sym_struct] = ACTIONS(4278),
    [anon_sym_union] = ACTIONS(4278),
    [anon_sym_QMARK] = ACTIONS(4280),
    [anon_sym_DASH_DASH] = ACTIONS(4280),
    [anon_sym_PLUS_PLUS] = ACTIONS(4280),
    [anon_sym_DOT] = ACTIONS(4278),
    [anon_sym_DASH_GT] = ACTIONS(4280),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4278),
    [anon_sym___typeof] = ACTIONS(4278),
    [anon_sym_typeof] = ACTIONS(4278),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4278),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4278),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4278),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4278),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4278),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4278),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4278),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4278),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4278),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4278),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4278),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4278),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4278),
    [anon_sym_API_AVAILABLE] = ACTIONS(4278),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4278),
    [anon_sym_API_DEPRECATED] = ACTIONS(4278),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4278),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4278),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4278),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4278),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4278),
    [anon_sym___deprecated_msg] = ACTIONS(4278),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4278),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4278),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4278),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4278),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4278),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4278),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4278),
    [anon_sym_BOOL] = ACTIONS(4278),
    [anon_sym_IMP] = ACTIONS(4278),
    [anon_sym_SEL] = ACTIONS(4278),
    [anon_sym_Class] = ACTIONS(4278),
    [anon_sym_id] = ACTIONS(4278),
  },
  [STATE(1646)] = {
    [sym_identifier] = ACTIONS(4208),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4210),
    [anon_sym_COMMA] = ACTIONS(4210),
    [anon_sym_RPAREN] = ACTIONS(4210),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4210),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4210),
    [anon_sym_LPAREN2] = ACTIONS(4210),
    [anon_sym_DASH] = ACTIONS(4208),
    [anon_sym_PLUS] = ACTIONS(4208),
    [anon_sym_STAR] = ACTIONS(4210),
    [anon_sym_SLASH] = ACTIONS(4208),
    [anon_sym_PERCENT] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE] = ACTIONS(4210),
    [anon_sym_AMP_AMP] = ACTIONS(4210),
    [anon_sym_PIPE] = ACTIONS(4208),
    [anon_sym_CARET] = ACTIONS(4210),
    [anon_sym_AMP] = ACTIONS(4208),
    [anon_sym_EQ_EQ] = ACTIONS(4210),
    [anon_sym_BANG_EQ] = ACTIONS(4210),
    [anon_sym_GT] = ACTIONS(4208),
    [anon_sym_GT_EQ] = ACTIONS(4210),
    [anon_sym_LT_EQ] = ACTIONS(4210),
    [anon_sym_LT] = ACTIONS(4208),
    [anon_sym_LT_LT] = ACTIONS(4210),
    [anon_sym_GT_GT] = ACTIONS(4210),
    [anon_sym_SEMI] = ACTIONS(4210),
    [anon_sym___extension__] = ACTIONS(4208),
    [anon_sym_extern] = ACTIONS(4208),
    [anon_sym___attribute__] = ACTIONS(4208),
    [anon_sym___attribute] = ACTIONS(4208),
    [anon_sym_noreturn] = ACTIONS(4208),
    [anon_sym_LBRACK] = ACTIONS(4210),
    [anon_sym_RBRACK] = ACTIONS(4210),
    [anon_sym___declspec] = ACTIONS(4208),
    [anon_sym_LBRACE] = ACTIONS(4210),
    [anon_sym_RBRACE] = ACTIONS(4210),
    [anon_sym_signed] = ACTIONS(4208),
    [anon_sym_unsigned] = ACTIONS(4208),
    [anon_sym_long] = ACTIONS(4208),
    [anon_sym_short] = ACTIONS(4208),
    [anon_sym_static] = ACTIONS(4208),
    [anon_sym_ATautoreleasepool] = ACTIONS(4210),
    [anon_sym_auto] = ACTIONS(4208),
    [anon_sym_register] = ACTIONS(4208),
    [anon_sym_inline] = ACTIONS(4208),
    [anon_sym___inline] = ACTIONS(4208),
    [anon_sym___inline__] = ACTIONS(4208),
    [anon_sym___forceinline] = ACTIONS(4208),
    [anon_sym_thread_local] = ACTIONS(4208),
    [anon_sym___thread] = ACTIONS(4208),
    [anon_sym_CG_EXTERN] = ACTIONS(4208),
    [anon_sym_CG_INLINE] = ACTIONS(4208),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4208),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4208),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4208),
    [anon_sym_IBOutlet] = ACTIONS(4208),
    [anon_sym_IBInspectable] = ACTIONS(4208),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4208),
    [anon_sym_NS_INLINE] = ACTIONS(4208),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4208),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4208),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4208),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4208),
    [anon_sym_const] = ACTIONS(4208),
    [anon_sym_constexpr] = ACTIONS(4208),
    [anon_sym_volatile] = ACTIONS(4208),
    [anon_sym_restrict] = ACTIONS(4208),
    [anon_sym___restrict__] = ACTIONS(4208),
    [anon_sym__Atomic] = ACTIONS(4208),
    [anon_sym__Noreturn] = ACTIONS(4208),
    [anon_sym__Nonnull] = ACTIONS(4208),
    [anon_sym_nullable] = ACTIONS(4208),
    [anon_sym__Complex] = ACTIONS(4208),
    [anon_sym__Nullable] = ACTIONS(4208),
    [anon_sym__Nullable_result] = ACTIONS(4208),
    [anon_sym__Null_unspecified] = ACTIONS(4208),
    [anon_sym___autoreleasing] = ACTIONS(4208),
    [anon_sym___block] = ACTIONS(4208),
    [anon_sym___bridge] = ACTIONS(4208),
    [anon_sym___bridge_retained] = ACTIONS(4208),
    [anon_sym___bridge_transfer] = ACTIONS(4208),
    [anon_sym___complex] = ACTIONS(4208),
    [anon_sym___const] = ACTIONS(4208),
    [anon_sym___imag] = ACTIONS(4208),
    [anon_sym___kindof] = ACTIONS(4208),
    [anon_sym___nonnull] = ACTIONS(4208),
    [anon_sym___nullable] = ACTIONS(4208),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4208),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4208),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4208),
    [anon_sym___real] = ACTIONS(4208),
    [anon_sym___strong] = ACTIONS(4208),
    [anon_sym___unsafe_unretained] = ACTIONS(4208),
    [anon_sym___unused] = ACTIONS(4208),
    [anon_sym___weak] = ACTIONS(4208),
    [anon_sym_alignas] = ACTIONS(4208),
    [anon_sym__Alignas] = ACTIONS(4208),
    [sym_primitive_type] = ACTIONS(4208),
    [anon_sym_enum] = ACTIONS(4208),
    [anon_sym_COLON] = ACTIONS(4210),
    [anon_sym_struct] = ACTIONS(4208),
    [anon_sym_union] = ACTIONS(4208),
    [anon_sym_QMARK] = ACTIONS(4210),
    [anon_sym_DASH_DASH] = ACTIONS(4210),
    [anon_sym_PLUS_PLUS] = ACTIONS(4210),
    [anon_sym_DOT] = ACTIONS(4208),
    [anon_sym_DASH_GT] = ACTIONS(4210),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4208),
    [anon_sym___typeof] = ACTIONS(4208),
    [anon_sym_typeof] = ACTIONS(4208),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4208),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4208),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4208),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4208),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4208),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4208),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4208),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4208),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4208),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4208),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4208),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4208),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4208),
    [anon_sym_API_AVAILABLE] = ACTIONS(4208),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4208),
    [anon_sym_API_DEPRECATED] = ACTIONS(4208),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4208),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4208),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4208),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4208),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4208),
    [anon_sym___deprecated_msg] = ACTIONS(4208),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4208),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4208),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4208),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4208),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4208),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4208),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4208),
    [anon_sym_BOOL] = ACTIONS(4208),
    [anon_sym_IMP] = ACTIONS(4208),
    [anon_sym_SEL] = ACTIONS(4208),
    [anon_sym_Class] = ACTIONS(4208),
    [anon_sym_id] = ACTIONS(4208),
  },
  [STATE(1647)] = {
    [sym_identifier] = ACTIONS(4216),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4218),
    [anon_sym_COMMA] = ACTIONS(4218),
    [anon_sym_RPAREN] = ACTIONS(4218),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4218),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4218),
    [anon_sym_LPAREN2] = ACTIONS(4218),
    [anon_sym_DASH] = ACTIONS(4216),
    [anon_sym_PLUS] = ACTIONS(4216),
    [anon_sym_STAR] = ACTIONS(4218),
    [anon_sym_SLASH] = ACTIONS(4216),
    [anon_sym_PERCENT] = ACTIONS(4218),
    [anon_sym_PIPE_PIPE] = ACTIONS(4218),
    [anon_sym_AMP_AMP] = ACTIONS(4218),
    [anon_sym_PIPE] = ACTIONS(4216),
    [anon_sym_CARET] = ACTIONS(4218),
    [anon_sym_AMP] = ACTIONS(4216),
    [anon_sym_EQ_EQ] = ACTIONS(4218),
    [anon_sym_BANG_EQ] = ACTIONS(4218),
    [anon_sym_GT] = ACTIONS(4216),
    [anon_sym_GT_EQ] = ACTIONS(4218),
    [anon_sym_LT_EQ] = ACTIONS(4218),
    [anon_sym_LT] = ACTIONS(4216),
    [anon_sym_LT_LT] = ACTIONS(4218),
    [anon_sym_GT_GT] = ACTIONS(4218),
    [anon_sym_SEMI] = ACTIONS(4218),
    [anon_sym___extension__] = ACTIONS(4216),
    [anon_sym_extern] = ACTIONS(4216),
    [anon_sym___attribute__] = ACTIONS(4216),
    [anon_sym___attribute] = ACTIONS(4216),
    [anon_sym_noreturn] = ACTIONS(4216),
    [anon_sym_LBRACK] = ACTIONS(4218),
    [anon_sym_RBRACK] = ACTIONS(4218),
    [anon_sym___declspec] = ACTIONS(4216),
    [anon_sym_LBRACE] = ACTIONS(4218),
    [anon_sym_RBRACE] = ACTIONS(4218),
    [anon_sym_signed] = ACTIONS(4216),
    [anon_sym_unsigned] = ACTIONS(4216),
    [anon_sym_long] = ACTIONS(4216),
    [anon_sym_short] = ACTIONS(4216),
    [anon_sym_static] = ACTIONS(4216),
    [anon_sym_ATautoreleasepool] = ACTIONS(4218),
    [anon_sym_auto] = ACTIONS(4216),
    [anon_sym_register] = ACTIONS(4216),
    [anon_sym_inline] = ACTIONS(4216),
    [anon_sym___inline] = ACTIONS(4216),
    [anon_sym___inline__] = ACTIONS(4216),
    [anon_sym___forceinline] = ACTIONS(4216),
    [anon_sym_thread_local] = ACTIONS(4216),
    [anon_sym___thread] = ACTIONS(4216),
    [anon_sym_CG_EXTERN] = ACTIONS(4216),
    [anon_sym_CG_INLINE] = ACTIONS(4216),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4216),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4216),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4216),
    [anon_sym_IBOutlet] = ACTIONS(4216),
    [anon_sym_IBInspectable] = ACTIONS(4216),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4216),
    [anon_sym_NS_INLINE] = ACTIONS(4216),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4216),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4216),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4216),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4216),
    [anon_sym_const] = ACTIONS(4216),
    [anon_sym_constexpr] = ACTIONS(4216),
    [anon_sym_volatile] = ACTIONS(4216),
    [anon_sym_restrict] = ACTIONS(4216),
    [anon_sym___restrict__] = ACTIONS(4216),
    [anon_sym__Atomic] = ACTIONS(4216),
    [anon_sym__Noreturn] = ACTIONS(4216),
    [anon_sym__Nonnull] = ACTIONS(4216),
    [anon_sym_nullable] = ACTIONS(4216),
    [anon_sym__Complex] = ACTIONS(4216),
    [anon_sym__Nullable] = ACTIONS(4216),
    [anon_sym__Nullable_result] = ACTIONS(4216),
    [anon_sym__Null_unspecified] = ACTIONS(4216),
    [anon_sym___autoreleasing] = ACTIONS(4216),
    [anon_sym___block] = ACTIONS(4216),
    [anon_sym___bridge] = ACTIONS(4216),
    [anon_sym___bridge_retained] = ACTIONS(4216),
    [anon_sym___bridge_transfer] = ACTIONS(4216),
    [anon_sym___complex] = ACTIONS(4216),
    [anon_sym___const] = ACTIONS(4216),
    [anon_sym___imag] = ACTIONS(4216),
    [anon_sym___kindof] = ACTIONS(4216),
    [anon_sym___nonnull] = ACTIONS(4216),
    [anon_sym___nullable] = ACTIONS(4216),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4216),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4216),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4216),
    [anon_sym___real] = ACTIONS(4216),
    [anon_sym___strong] = ACTIONS(4216),
    [anon_sym___unsafe_unretained] = ACTIONS(4216),
    [anon_sym___unused] = ACTIONS(4216),
    [anon_sym___weak] = ACTIONS(4216),
    [anon_sym_alignas] = ACTIONS(4216),
    [anon_sym__Alignas] = ACTIONS(4216),
    [sym_primitive_type] = ACTIONS(4216),
    [anon_sym_enum] = ACTIONS(4216),
    [anon_sym_COLON] = ACTIONS(4218),
    [anon_sym_struct] = ACTIONS(4216),
    [anon_sym_union] = ACTIONS(4216),
    [anon_sym_QMARK] = ACTIONS(4218),
    [anon_sym_DASH_DASH] = ACTIONS(4218),
    [anon_sym_PLUS_PLUS] = ACTIONS(4218),
    [anon_sym_DOT] = ACTIONS(4216),
    [anon_sym_DASH_GT] = ACTIONS(4218),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4216),
    [anon_sym___typeof] = ACTIONS(4216),
    [anon_sym_typeof] = ACTIONS(4216),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4216),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4216),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4216),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4216),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4216),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4216),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4216),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4216),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4216),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4216),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4216),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4216),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4216),
    [anon_sym_API_AVAILABLE] = ACTIONS(4216),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4216),
    [anon_sym_API_DEPRECATED] = ACTIONS(4216),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4216),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4216),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4216),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4216),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4216),
    [anon_sym___deprecated_msg] = ACTIONS(4216),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4216),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4216),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4216),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4216),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4216),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4216),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4216),
    [anon_sym_BOOL] = ACTIONS(4216),
    [anon_sym_IMP] = ACTIONS(4216),
    [anon_sym_SEL] = ACTIONS(4216),
    [anon_sym_Class] = ACTIONS(4216),
    [anon_sym_id] = ACTIONS(4216),
  },
  [STATE(1648)] = {
    [sym_identifier] = ACTIONS(4192),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4194),
    [anon_sym_COMMA] = ACTIONS(4194),
    [anon_sym_RPAREN] = ACTIONS(4194),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4194),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4194),
    [anon_sym_LPAREN2] = ACTIONS(4194),
    [anon_sym_DASH] = ACTIONS(4192),
    [anon_sym_PLUS] = ACTIONS(4192),
    [anon_sym_STAR] = ACTIONS(4194),
    [anon_sym_SLASH] = ACTIONS(4192),
    [anon_sym_PERCENT] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE] = ACTIONS(4194),
    [anon_sym_AMP_AMP] = ACTIONS(4194),
    [anon_sym_PIPE] = ACTIONS(4192),
    [anon_sym_CARET] = ACTIONS(4194),
    [anon_sym_AMP] = ACTIONS(4192),
    [anon_sym_EQ_EQ] = ACTIONS(4194),
    [anon_sym_BANG_EQ] = ACTIONS(4194),
    [anon_sym_GT] = ACTIONS(4192),
    [anon_sym_GT_EQ] = ACTIONS(4194),
    [anon_sym_LT_EQ] = ACTIONS(4194),
    [anon_sym_LT] = ACTIONS(4192),
    [anon_sym_LT_LT] = ACTIONS(4194),
    [anon_sym_GT_GT] = ACTIONS(4194),
    [anon_sym_SEMI] = ACTIONS(4194),
    [anon_sym___extension__] = ACTIONS(4192),
    [anon_sym_extern] = ACTIONS(4192),
    [anon_sym___attribute__] = ACTIONS(4192),
    [anon_sym___attribute] = ACTIONS(4192),
    [anon_sym_noreturn] = ACTIONS(4192),
    [anon_sym_LBRACK] = ACTIONS(4194),
    [anon_sym_RBRACK] = ACTIONS(4194),
    [anon_sym___declspec] = ACTIONS(4192),
    [anon_sym_LBRACE] = ACTIONS(4194),
    [anon_sym_RBRACE] = ACTIONS(4194),
    [anon_sym_signed] = ACTIONS(4192),
    [anon_sym_unsigned] = ACTIONS(4192),
    [anon_sym_long] = ACTIONS(4192),
    [anon_sym_short] = ACTIONS(4192),
    [anon_sym_static] = ACTIONS(4192),
    [anon_sym_ATautoreleasepool] = ACTIONS(4194),
    [anon_sym_auto] = ACTIONS(4192),
    [anon_sym_register] = ACTIONS(4192),
    [anon_sym_inline] = ACTIONS(4192),
    [anon_sym___inline] = ACTIONS(4192),
    [anon_sym___inline__] = ACTIONS(4192),
    [anon_sym___forceinline] = ACTIONS(4192),
    [anon_sym_thread_local] = ACTIONS(4192),
    [anon_sym___thread] = ACTIONS(4192),
    [anon_sym_CG_EXTERN] = ACTIONS(4192),
    [anon_sym_CG_INLINE] = ACTIONS(4192),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4192),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4192),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4192),
    [anon_sym_IBOutlet] = ACTIONS(4192),
    [anon_sym_IBInspectable] = ACTIONS(4192),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4192),
    [anon_sym_NS_INLINE] = ACTIONS(4192),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4192),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4192),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4192),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4192),
    [anon_sym_const] = ACTIONS(4192),
    [anon_sym_constexpr] = ACTIONS(4192),
    [anon_sym_volatile] = ACTIONS(4192),
    [anon_sym_restrict] = ACTIONS(4192),
    [anon_sym___restrict__] = ACTIONS(4192),
    [anon_sym__Atomic] = ACTIONS(4192),
    [anon_sym__Noreturn] = ACTIONS(4192),
    [anon_sym__Nonnull] = ACTIONS(4192),
    [anon_sym_nullable] = ACTIONS(4192),
    [anon_sym__Complex] = ACTIONS(4192),
    [anon_sym__Nullable] = ACTIONS(4192),
    [anon_sym__Nullable_result] = ACTIONS(4192),
    [anon_sym__Null_unspecified] = ACTIONS(4192),
    [anon_sym___autoreleasing] = ACTIONS(4192),
    [anon_sym___block] = ACTIONS(4192),
    [anon_sym___bridge] = ACTIONS(4192),
    [anon_sym___bridge_retained] = ACTIONS(4192),
    [anon_sym___bridge_transfer] = ACTIONS(4192),
    [anon_sym___complex] = ACTIONS(4192),
    [anon_sym___const] = ACTIONS(4192),
    [anon_sym___imag] = ACTIONS(4192),
    [anon_sym___kindof] = ACTIONS(4192),
    [anon_sym___nonnull] = ACTIONS(4192),
    [anon_sym___nullable] = ACTIONS(4192),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4192),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4192),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4192),
    [anon_sym___real] = ACTIONS(4192),
    [anon_sym___strong] = ACTIONS(4192),
    [anon_sym___unsafe_unretained] = ACTIONS(4192),
    [anon_sym___unused] = ACTIONS(4192),
    [anon_sym___weak] = ACTIONS(4192),
    [anon_sym_alignas] = ACTIONS(4192),
    [anon_sym__Alignas] = ACTIONS(4192),
    [sym_primitive_type] = ACTIONS(4192),
    [anon_sym_enum] = ACTIONS(4192),
    [anon_sym_COLON] = ACTIONS(4194),
    [anon_sym_struct] = ACTIONS(4192),
    [anon_sym_union] = ACTIONS(4192),
    [anon_sym_QMARK] = ACTIONS(4194),
    [anon_sym_DASH_DASH] = ACTIONS(4194),
    [anon_sym_PLUS_PLUS] = ACTIONS(4194),
    [anon_sym_DOT] = ACTIONS(4192),
    [anon_sym_DASH_GT] = ACTIONS(4194),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4192),
    [anon_sym___typeof] = ACTIONS(4192),
    [anon_sym_typeof] = ACTIONS(4192),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4192),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4192),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4192),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4192),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4192),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4192),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4192),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4192),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4192),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4192),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4192),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4192),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4192),
    [anon_sym_API_AVAILABLE] = ACTIONS(4192),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4192),
    [anon_sym_API_DEPRECATED] = ACTIONS(4192),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4192),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4192),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4192),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4192),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4192),
    [anon_sym___deprecated_msg] = ACTIONS(4192),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4192),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4192),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4192),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4192),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4192),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4192),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4192),
    [anon_sym_BOOL] = ACTIONS(4192),
    [anon_sym_IMP] = ACTIONS(4192),
    [anon_sym_SEL] = ACTIONS(4192),
    [anon_sym_Class] = ACTIONS(4192),
    [anon_sym_id] = ACTIONS(4192),
  },
  [STATE(1649)] = {
    [sym_identifier] = ACTIONS(4196),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4198),
    [anon_sym_COMMA] = ACTIONS(4198),
    [anon_sym_RPAREN] = ACTIONS(4198),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4198),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4198),
    [anon_sym_LPAREN2] = ACTIONS(4198),
    [anon_sym_DASH] = ACTIONS(4196),
    [anon_sym_PLUS] = ACTIONS(4196),
    [anon_sym_STAR] = ACTIONS(4198),
    [anon_sym_SLASH] = ACTIONS(4196),
    [anon_sym_PERCENT] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE] = ACTIONS(4198),
    [anon_sym_AMP_AMP] = ACTIONS(4198),
    [anon_sym_PIPE] = ACTIONS(4196),
    [anon_sym_CARET] = ACTIONS(4198),
    [anon_sym_AMP] = ACTIONS(4196),
    [anon_sym_EQ_EQ] = ACTIONS(4198),
    [anon_sym_BANG_EQ] = ACTIONS(4198),
    [anon_sym_GT] = ACTIONS(4196),
    [anon_sym_GT_EQ] = ACTIONS(4198),
    [anon_sym_LT_EQ] = ACTIONS(4198),
    [anon_sym_LT] = ACTIONS(4196),
    [anon_sym_LT_LT] = ACTIONS(4198),
    [anon_sym_GT_GT] = ACTIONS(4198),
    [anon_sym_SEMI] = ACTIONS(4198),
    [anon_sym___extension__] = ACTIONS(4196),
    [anon_sym_extern] = ACTIONS(4196),
    [anon_sym___attribute__] = ACTIONS(4196),
    [anon_sym___attribute] = ACTIONS(4196),
    [anon_sym_noreturn] = ACTIONS(4196),
    [anon_sym_LBRACK] = ACTIONS(4198),
    [anon_sym_RBRACK] = ACTIONS(4198),
    [anon_sym___declspec] = ACTIONS(4196),
    [anon_sym_LBRACE] = ACTIONS(4198),
    [anon_sym_RBRACE] = ACTIONS(4198),
    [anon_sym_signed] = ACTIONS(4196),
    [anon_sym_unsigned] = ACTIONS(4196),
    [anon_sym_long] = ACTIONS(4196),
    [anon_sym_short] = ACTIONS(4196),
    [anon_sym_static] = ACTIONS(4196),
    [anon_sym_ATautoreleasepool] = ACTIONS(4198),
    [anon_sym_auto] = ACTIONS(4196),
    [anon_sym_register] = ACTIONS(4196),
    [anon_sym_inline] = ACTIONS(4196),
    [anon_sym___inline] = ACTIONS(4196),
    [anon_sym___inline__] = ACTIONS(4196),
    [anon_sym___forceinline] = ACTIONS(4196),
    [anon_sym_thread_local] = ACTIONS(4196),
    [anon_sym___thread] = ACTIONS(4196),
    [anon_sym_CG_EXTERN] = ACTIONS(4196),
    [anon_sym_CG_INLINE] = ACTIONS(4196),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4196),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4196),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4196),
    [anon_sym_IBOutlet] = ACTIONS(4196),
    [anon_sym_IBInspectable] = ACTIONS(4196),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4196),
    [anon_sym_NS_INLINE] = ACTIONS(4196),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4196),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4196),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4196),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4196),
    [anon_sym_const] = ACTIONS(4196),
    [anon_sym_constexpr] = ACTIONS(4196),
    [anon_sym_volatile] = ACTIONS(4196),
    [anon_sym_restrict] = ACTIONS(4196),
    [anon_sym___restrict__] = ACTIONS(4196),
    [anon_sym__Atomic] = ACTIONS(4196),
    [anon_sym__Noreturn] = ACTIONS(4196),
    [anon_sym__Nonnull] = ACTIONS(4196),
    [anon_sym_nullable] = ACTIONS(4196),
    [anon_sym__Complex] = ACTIONS(4196),
    [anon_sym__Nullable] = ACTIONS(4196),
    [anon_sym__Nullable_result] = ACTIONS(4196),
    [anon_sym__Null_unspecified] = ACTIONS(4196),
    [anon_sym___autoreleasing] = ACTIONS(4196),
    [anon_sym___block] = ACTIONS(4196),
    [anon_sym___bridge] = ACTIONS(4196),
    [anon_sym___bridge_retained] = ACTIONS(4196),
    [anon_sym___bridge_transfer] = ACTIONS(4196),
    [anon_sym___complex] = ACTIONS(4196),
    [anon_sym___const] = ACTIONS(4196),
    [anon_sym___imag] = ACTIONS(4196),
    [anon_sym___kindof] = ACTIONS(4196),
    [anon_sym___nonnull] = ACTIONS(4196),
    [anon_sym___nullable] = ACTIONS(4196),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4196),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4196),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4196),
    [anon_sym___real] = ACTIONS(4196),
    [anon_sym___strong] = ACTIONS(4196),
    [anon_sym___unsafe_unretained] = ACTIONS(4196),
    [anon_sym___unused] = ACTIONS(4196),
    [anon_sym___weak] = ACTIONS(4196),
    [anon_sym_alignas] = ACTIONS(4196),
    [anon_sym__Alignas] = ACTIONS(4196),
    [sym_primitive_type] = ACTIONS(4196),
    [anon_sym_enum] = ACTIONS(4196),
    [anon_sym_COLON] = ACTIONS(4198),
    [anon_sym_struct] = ACTIONS(4196),
    [anon_sym_union] = ACTIONS(4196),
    [anon_sym_QMARK] = ACTIONS(4198),
    [anon_sym_DASH_DASH] = ACTIONS(4198),
    [anon_sym_PLUS_PLUS] = ACTIONS(4198),
    [anon_sym_DOT] = ACTIONS(4196),
    [anon_sym_DASH_GT] = ACTIONS(4198),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4196),
    [anon_sym___typeof] = ACTIONS(4196),
    [anon_sym_typeof] = ACTIONS(4196),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4196),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4196),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4196),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4196),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4196),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4196),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4196),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4196),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4196),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4196),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4196),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4196),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4196),
    [anon_sym_API_AVAILABLE] = ACTIONS(4196),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4196),
    [anon_sym_API_DEPRECATED] = ACTIONS(4196),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4196),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4196),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4196),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4196),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4196),
    [anon_sym___deprecated_msg] = ACTIONS(4196),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4196),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4196),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4196),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4196),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4196),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4196),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4196),
    [anon_sym_BOOL] = ACTIONS(4196),
    [anon_sym_IMP] = ACTIONS(4196),
    [anon_sym_SEL] = ACTIONS(4196),
    [anon_sym_Class] = ACTIONS(4196),
    [anon_sym_id] = ACTIONS(4196),
  },
  [STATE(1650)] = {
    [sym_identifier] = ACTIONS(4150),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4152),
    [anon_sym_COMMA] = ACTIONS(4152),
    [anon_sym_RPAREN] = ACTIONS(4152),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4152),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4152),
    [anon_sym_LPAREN2] = ACTIONS(4152),
    [anon_sym_DASH] = ACTIONS(4150),
    [anon_sym_PLUS] = ACTIONS(4150),
    [anon_sym_STAR] = ACTIONS(4152),
    [anon_sym_SLASH] = ACTIONS(4150),
    [anon_sym_PERCENT] = ACTIONS(4152),
    [anon_sym_PIPE_PIPE] = ACTIONS(4152),
    [anon_sym_AMP_AMP] = ACTIONS(4152),
    [anon_sym_PIPE] = ACTIONS(4150),
    [anon_sym_CARET] = ACTIONS(4152),
    [anon_sym_AMP] = ACTIONS(4150),
    [anon_sym_EQ_EQ] = ACTIONS(4152),
    [anon_sym_BANG_EQ] = ACTIONS(4152),
    [anon_sym_GT] = ACTIONS(4150),
    [anon_sym_GT_EQ] = ACTIONS(4152),
    [anon_sym_LT_EQ] = ACTIONS(4152),
    [anon_sym_LT] = ACTIONS(4150),
    [anon_sym_LT_LT] = ACTIONS(4152),
    [anon_sym_GT_GT] = ACTIONS(4152),
    [anon_sym_SEMI] = ACTIONS(4152),
    [anon_sym___extension__] = ACTIONS(4150),
    [anon_sym_extern] = ACTIONS(4150),
    [anon_sym___attribute__] = ACTIONS(4150),
    [anon_sym___attribute] = ACTIONS(4150),
    [anon_sym_noreturn] = ACTIONS(4150),
    [anon_sym_LBRACK] = ACTIONS(4152),
    [anon_sym_RBRACK] = ACTIONS(4152),
    [anon_sym___declspec] = ACTIONS(4150),
    [anon_sym_LBRACE] = ACTIONS(4152),
    [anon_sym_RBRACE] = ACTIONS(4152),
    [anon_sym_signed] = ACTIONS(4150),
    [anon_sym_unsigned] = ACTIONS(4150),
    [anon_sym_long] = ACTIONS(4150),
    [anon_sym_short] = ACTIONS(4150),
    [anon_sym_static] = ACTIONS(4150),
    [anon_sym_ATautoreleasepool] = ACTIONS(4152),
    [anon_sym_auto] = ACTIONS(4150),
    [anon_sym_register] = ACTIONS(4150),
    [anon_sym_inline] = ACTIONS(4150),
    [anon_sym___inline] = ACTIONS(4150),
    [anon_sym___inline__] = ACTIONS(4150),
    [anon_sym___forceinline] = ACTIONS(4150),
    [anon_sym_thread_local] = ACTIONS(4150),
    [anon_sym___thread] = ACTIONS(4150),
    [anon_sym_CG_EXTERN] = ACTIONS(4150),
    [anon_sym_CG_INLINE] = ACTIONS(4150),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4150),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4150),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4150),
    [anon_sym_IBOutlet] = ACTIONS(4150),
    [anon_sym_IBInspectable] = ACTIONS(4150),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4150),
    [anon_sym_NS_INLINE] = ACTIONS(4150),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4150),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4150),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4150),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4150),
    [anon_sym_const] = ACTIONS(4150),
    [anon_sym_constexpr] = ACTIONS(4150),
    [anon_sym_volatile] = ACTIONS(4150),
    [anon_sym_restrict] = ACTIONS(4150),
    [anon_sym___restrict__] = ACTIONS(4150),
    [anon_sym__Atomic] = ACTIONS(4150),
    [anon_sym__Noreturn] = ACTIONS(4150),
    [anon_sym__Nonnull] = ACTIONS(4150),
    [anon_sym_nullable] = ACTIONS(4150),
    [anon_sym__Complex] = ACTIONS(4150),
    [anon_sym__Nullable] = ACTIONS(4150),
    [anon_sym__Nullable_result] = ACTIONS(4150),
    [anon_sym__Null_unspecified] = ACTIONS(4150),
    [anon_sym___autoreleasing] = ACTIONS(4150),
    [anon_sym___block] = ACTIONS(4150),
    [anon_sym___bridge] = ACTIONS(4150),
    [anon_sym___bridge_retained] = ACTIONS(4150),
    [anon_sym___bridge_transfer] = ACTIONS(4150),
    [anon_sym___complex] = ACTIONS(4150),
    [anon_sym___const] = ACTIONS(4150),
    [anon_sym___imag] = ACTIONS(4150),
    [anon_sym___kindof] = ACTIONS(4150),
    [anon_sym___nonnull] = ACTIONS(4150),
    [anon_sym___nullable] = ACTIONS(4150),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4150),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4150),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4150),
    [anon_sym___real] = ACTIONS(4150),
    [anon_sym___strong] = ACTIONS(4150),
    [anon_sym___unsafe_unretained] = ACTIONS(4150),
    [anon_sym___unused] = ACTIONS(4150),
    [anon_sym___weak] = ACTIONS(4150),
    [anon_sym_alignas] = ACTIONS(4150),
    [anon_sym__Alignas] = ACTIONS(4150),
    [sym_primitive_type] = ACTIONS(4150),
    [anon_sym_enum] = ACTIONS(4150),
    [anon_sym_COLON] = ACTIONS(4152),
    [anon_sym_struct] = ACTIONS(4150),
    [anon_sym_union] = ACTIONS(4150),
    [anon_sym_QMARK] = ACTIONS(4152),
    [anon_sym_DASH_DASH] = ACTIONS(4152),
    [anon_sym_PLUS_PLUS] = ACTIONS(4152),
    [anon_sym_DOT] = ACTIONS(4150),
    [anon_sym_DASH_GT] = ACTIONS(4152),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4150),
    [anon_sym___typeof] = ACTIONS(4150),
    [anon_sym_typeof] = ACTIONS(4150),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4150),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4150),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4150),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4150),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4150),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4150),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4150),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4150),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4150),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4150),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4150),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4150),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4150),
    [anon_sym_API_AVAILABLE] = ACTIONS(4150),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4150),
    [anon_sym_API_DEPRECATED] = ACTIONS(4150),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4150),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4150),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4150),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4150),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4150),
    [anon_sym___deprecated_msg] = ACTIONS(4150),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4150),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4150),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4150),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4150),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4150),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4150),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4150),
    [anon_sym_BOOL] = ACTIONS(4150),
    [anon_sym_IMP] = ACTIONS(4150),
    [anon_sym_SEL] = ACTIONS(4150),
    [anon_sym_Class] = ACTIONS(4150),
    [anon_sym_id] = ACTIONS(4150),
  },
  [STATE(1651)] = {
    [sym_identifier] = ACTIONS(4204),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4206),
    [anon_sym_COMMA] = ACTIONS(4206),
    [anon_sym_RPAREN] = ACTIONS(4206),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4206),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4206),
    [anon_sym_LPAREN2] = ACTIONS(4206),
    [anon_sym_DASH] = ACTIONS(4204),
    [anon_sym_PLUS] = ACTIONS(4204),
    [anon_sym_STAR] = ACTIONS(4206),
    [anon_sym_SLASH] = ACTIONS(4204),
    [anon_sym_PERCENT] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE] = ACTIONS(4206),
    [anon_sym_AMP_AMP] = ACTIONS(4206),
    [anon_sym_PIPE] = ACTIONS(4204),
    [anon_sym_CARET] = ACTIONS(4206),
    [anon_sym_AMP] = ACTIONS(4204),
    [anon_sym_EQ_EQ] = ACTIONS(4206),
    [anon_sym_BANG_EQ] = ACTIONS(4206),
    [anon_sym_GT] = ACTIONS(4204),
    [anon_sym_GT_EQ] = ACTIONS(4206),
    [anon_sym_LT_EQ] = ACTIONS(4206),
    [anon_sym_LT] = ACTIONS(4204),
    [anon_sym_LT_LT] = ACTIONS(4206),
    [anon_sym_GT_GT] = ACTIONS(4206),
    [anon_sym_SEMI] = ACTIONS(4206),
    [anon_sym___extension__] = ACTIONS(4204),
    [anon_sym_extern] = ACTIONS(4204),
    [anon_sym___attribute__] = ACTIONS(4204),
    [anon_sym___attribute] = ACTIONS(4204),
    [anon_sym_noreturn] = ACTIONS(4204),
    [anon_sym_LBRACK] = ACTIONS(4206),
    [anon_sym_RBRACK] = ACTIONS(4206),
    [anon_sym___declspec] = ACTIONS(4204),
    [anon_sym_LBRACE] = ACTIONS(4206),
    [anon_sym_RBRACE] = ACTIONS(4206),
    [anon_sym_signed] = ACTIONS(4204),
    [anon_sym_unsigned] = ACTIONS(4204),
    [anon_sym_long] = ACTIONS(4204),
    [anon_sym_short] = ACTIONS(4204),
    [anon_sym_static] = ACTIONS(4204),
    [anon_sym_ATautoreleasepool] = ACTIONS(4206),
    [anon_sym_auto] = ACTIONS(4204),
    [anon_sym_register] = ACTIONS(4204),
    [anon_sym_inline] = ACTIONS(4204),
    [anon_sym___inline] = ACTIONS(4204),
    [anon_sym___inline__] = ACTIONS(4204),
    [anon_sym___forceinline] = ACTIONS(4204),
    [anon_sym_thread_local] = ACTIONS(4204),
    [anon_sym___thread] = ACTIONS(4204),
    [anon_sym_CG_EXTERN] = ACTIONS(4204),
    [anon_sym_CG_INLINE] = ACTIONS(4204),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4204),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4204),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4204),
    [anon_sym_IBOutlet] = ACTIONS(4204),
    [anon_sym_IBInspectable] = ACTIONS(4204),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4204),
    [anon_sym_NS_INLINE] = ACTIONS(4204),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4204),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4204),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4204),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4204),
    [anon_sym_const] = ACTIONS(4204),
    [anon_sym_constexpr] = ACTIONS(4204),
    [anon_sym_volatile] = ACTIONS(4204),
    [anon_sym_restrict] = ACTIONS(4204),
    [anon_sym___restrict__] = ACTIONS(4204),
    [anon_sym__Atomic] = ACTIONS(4204),
    [anon_sym__Noreturn] = ACTIONS(4204),
    [anon_sym__Nonnull] = ACTIONS(4204),
    [anon_sym_nullable] = ACTIONS(4204),
    [anon_sym__Complex] = ACTIONS(4204),
    [anon_sym__Nullable] = ACTIONS(4204),
    [anon_sym__Nullable_result] = ACTIONS(4204),
    [anon_sym__Null_unspecified] = ACTIONS(4204),
    [anon_sym___autoreleasing] = ACTIONS(4204),
    [anon_sym___block] = ACTIONS(4204),
    [anon_sym___bridge] = ACTIONS(4204),
    [anon_sym___bridge_retained] = ACTIONS(4204),
    [anon_sym___bridge_transfer] = ACTIONS(4204),
    [anon_sym___complex] = ACTIONS(4204),
    [anon_sym___const] = ACTIONS(4204),
    [anon_sym___imag] = ACTIONS(4204),
    [anon_sym___kindof] = ACTIONS(4204),
    [anon_sym___nonnull] = ACTIONS(4204),
    [anon_sym___nullable] = ACTIONS(4204),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4204),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4204),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4204),
    [anon_sym___real] = ACTIONS(4204),
    [anon_sym___strong] = ACTIONS(4204),
    [anon_sym___unsafe_unretained] = ACTIONS(4204),
    [anon_sym___unused] = ACTIONS(4204),
    [anon_sym___weak] = ACTIONS(4204),
    [anon_sym_alignas] = ACTIONS(4204),
    [anon_sym__Alignas] = ACTIONS(4204),
    [sym_primitive_type] = ACTIONS(4204),
    [anon_sym_enum] = ACTIONS(4204),
    [anon_sym_COLON] = ACTIONS(4206),
    [anon_sym_struct] = ACTIONS(4204),
    [anon_sym_union] = ACTIONS(4204),
    [anon_sym_QMARK] = ACTIONS(4206),
    [anon_sym_DASH_DASH] = ACTIONS(4206),
    [anon_sym_PLUS_PLUS] = ACTIONS(4206),
    [anon_sym_DOT] = ACTIONS(4204),
    [anon_sym_DASH_GT] = ACTIONS(4206),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4204),
    [anon_sym___typeof] = ACTIONS(4204),
    [anon_sym_typeof] = ACTIONS(4204),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4204),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4204),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4204),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4204),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4204),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4204),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4204),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4204),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4204),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4204),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4204),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4204),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4204),
    [anon_sym_API_AVAILABLE] = ACTIONS(4204),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4204),
    [anon_sym_API_DEPRECATED] = ACTIONS(4204),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4204),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4204),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4204),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4204),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4204),
    [anon_sym___deprecated_msg] = ACTIONS(4204),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4204),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4204),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4204),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4204),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4204),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4204),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4204),
    [anon_sym_BOOL] = ACTIONS(4204),
    [anon_sym_IMP] = ACTIONS(4204),
    [anon_sym_SEL] = ACTIONS(4204),
    [anon_sym_Class] = ACTIONS(4204),
    [anon_sym_id] = ACTIONS(4204),
  },
  [STATE(1652)] = {
    [sym_identifier] = ACTIONS(4244),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4246),
    [anon_sym_COMMA] = ACTIONS(4246),
    [anon_sym_RPAREN] = ACTIONS(4246),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4246),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4246),
    [anon_sym_LPAREN2] = ACTIONS(4246),
    [anon_sym_DASH] = ACTIONS(4244),
    [anon_sym_PLUS] = ACTIONS(4244),
    [anon_sym_STAR] = ACTIONS(4246),
    [anon_sym_SLASH] = ACTIONS(4244),
    [anon_sym_PERCENT] = ACTIONS(4246),
    [anon_sym_PIPE_PIPE] = ACTIONS(4246),
    [anon_sym_AMP_AMP] = ACTIONS(4246),
    [anon_sym_PIPE] = ACTIONS(4244),
    [anon_sym_CARET] = ACTIONS(4246),
    [anon_sym_AMP] = ACTIONS(4244),
    [anon_sym_EQ_EQ] = ACTIONS(4246),
    [anon_sym_BANG_EQ] = ACTIONS(4246),
    [anon_sym_GT] = ACTIONS(4244),
    [anon_sym_GT_EQ] = ACTIONS(4246),
    [anon_sym_LT_EQ] = ACTIONS(4246),
    [anon_sym_LT] = ACTIONS(4244),
    [anon_sym_LT_LT] = ACTIONS(4246),
    [anon_sym_GT_GT] = ACTIONS(4246),
    [anon_sym_SEMI] = ACTIONS(4246),
    [anon_sym___extension__] = ACTIONS(4244),
    [anon_sym_extern] = ACTIONS(4244),
    [anon_sym___attribute__] = ACTIONS(4244),
    [anon_sym___attribute] = ACTIONS(4244),
    [anon_sym_noreturn] = ACTIONS(4244),
    [anon_sym_LBRACK] = ACTIONS(4246),
    [anon_sym_RBRACK] = ACTIONS(4246),
    [anon_sym___declspec] = ACTIONS(4244),
    [anon_sym_LBRACE] = ACTIONS(4246),
    [anon_sym_RBRACE] = ACTIONS(4246),
    [anon_sym_signed] = ACTIONS(4244),
    [anon_sym_unsigned] = ACTIONS(4244),
    [anon_sym_long] = ACTIONS(4244),
    [anon_sym_short] = ACTIONS(4244),
    [anon_sym_static] = ACTIONS(4244),
    [anon_sym_ATautoreleasepool] = ACTIONS(4246),
    [anon_sym_auto] = ACTIONS(4244),
    [anon_sym_register] = ACTIONS(4244),
    [anon_sym_inline] = ACTIONS(4244),
    [anon_sym___inline] = ACTIONS(4244),
    [anon_sym___inline__] = ACTIONS(4244),
    [anon_sym___forceinline] = ACTIONS(4244),
    [anon_sym_thread_local] = ACTIONS(4244),
    [anon_sym___thread] = ACTIONS(4244),
    [anon_sym_CG_EXTERN] = ACTIONS(4244),
    [anon_sym_CG_INLINE] = ACTIONS(4244),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4244),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4244),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4244),
    [anon_sym_IBOutlet] = ACTIONS(4244),
    [anon_sym_IBInspectable] = ACTIONS(4244),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4244),
    [anon_sym_NS_INLINE] = ACTIONS(4244),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4244),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4244),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4244),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4244),
    [anon_sym_const] = ACTIONS(4244),
    [anon_sym_constexpr] = ACTIONS(4244),
    [anon_sym_volatile] = ACTIONS(4244),
    [anon_sym_restrict] = ACTIONS(4244),
    [anon_sym___restrict__] = ACTIONS(4244),
    [anon_sym__Atomic] = ACTIONS(4244),
    [anon_sym__Noreturn] = ACTIONS(4244),
    [anon_sym__Nonnull] = ACTIONS(4244),
    [anon_sym_nullable] = ACTIONS(4244),
    [anon_sym__Complex] = ACTIONS(4244),
    [anon_sym__Nullable] = ACTIONS(4244),
    [anon_sym__Nullable_result] = ACTIONS(4244),
    [anon_sym__Null_unspecified] = ACTIONS(4244),
    [anon_sym___autoreleasing] = ACTIONS(4244),
    [anon_sym___block] = ACTIONS(4244),
    [anon_sym___bridge] = ACTIONS(4244),
    [anon_sym___bridge_retained] = ACTIONS(4244),
    [anon_sym___bridge_transfer] = ACTIONS(4244),
    [anon_sym___complex] = ACTIONS(4244),
    [anon_sym___const] = ACTIONS(4244),
    [anon_sym___imag] = ACTIONS(4244),
    [anon_sym___kindof] = ACTIONS(4244),
    [anon_sym___nonnull] = ACTIONS(4244),
    [anon_sym___nullable] = ACTIONS(4244),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4244),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4244),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4244),
    [anon_sym___real] = ACTIONS(4244),
    [anon_sym___strong] = ACTIONS(4244),
    [anon_sym___unsafe_unretained] = ACTIONS(4244),
    [anon_sym___unused] = ACTIONS(4244),
    [anon_sym___weak] = ACTIONS(4244),
    [anon_sym_alignas] = ACTIONS(4244),
    [anon_sym__Alignas] = ACTIONS(4244),
    [sym_primitive_type] = ACTIONS(4244),
    [anon_sym_enum] = ACTIONS(4244),
    [anon_sym_COLON] = ACTIONS(4246),
    [anon_sym_struct] = ACTIONS(4244),
    [anon_sym_union] = ACTIONS(4244),
    [anon_sym_QMARK] = ACTIONS(4246),
    [anon_sym_DASH_DASH] = ACTIONS(4246),
    [anon_sym_PLUS_PLUS] = ACTIONS(4246),
    [anon_sym_DOT] = ACTIONS(4244),
    [anon_sym_DASH_GT] = ACTIONS(4246),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4244),
    [anon_sym___typeof] = ACTIONS(4244),
    [anon_sym_typeof] = ACTIONS(4244),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4244),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4244),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4244),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4244),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4244),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4244),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4244),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4244),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4244),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4244),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4244),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4244),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4244),
    [anon_sym_API_AVAILABLE] = ACTIONS(4244),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4244),
    [anon_sym_API_DEPRECATED] = ACTIONS(4244),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4244),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4244),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4244),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4244),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4244),
    [anon_sym___deprecated_msg] = ACTIONS(4244),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4244),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4244),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4244),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4244),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4244),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4244),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4244),
    [anon_sym_BOOL] = ACTIONS(4244),
    [anon_sym_IMP] = ACTIONS(4244),
    [anon_sym_SEL] = ACTIONS(4244),
    [anon_sym_Class] = ACTIONS(4244),
    [anon_sym_id] = ACTIONS(4244),
  },
  [STATE(1653)] = {
    [sym_identifier] = ACTIONS(4262),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4264),
    [anon_sym_COMMA] = ACTIONS(4264),
    [anon_sym_LPAREN2] = ACTIONS(4264),
    [anon_sym_DASH] = ACTIONS(4262),
    [anon_sym_PLUS] = ACTIONS(4262),
    [anon_sym_STAR] = ACTIONS(4264),
    [anon_sym_SLASH] = ACTIONS(4262),
    [anon_sym_PERCENT] = ACTIONS(4264),
    [anon_sym_PIPE_PIPE] = ACTIONS(4264),
    [anon_sym_AMP_AMP] = ACTIONS(4264),
    [anon_sym_PIPE] = ACTIONS(4262),
    [anon_sym_CARET] = ACTIONS(4264),
    [anon_sym_AMP] = ACTIONS(4262),
    [anon_sym_EQ_EQ] = ACTIONS(4264),
    [anon_sym_BANG_EQ] = ACTIONS(4264),
    [anon_sym_GT] = ACTIONS(4262),
    [anon_sym_GT_EQ] = ACTIONS(4264),
    [anon_sym_LT_EQ] = ACTIONS(4264),
    [anon_sym_LT] = ACTIONS(4262),
    [anon_sym_LT_LT] = ACTIONS(4264),
    [anon_sym_GT_GT] = ACTIONS(4264),
    [anon_sym_SEMI] = ACTIONS(4264),
    [anon_sym___extension__] = ACTIONS(4262),
    [anon_sym_extern] = ACTIONS(4262),
    [anon_sym___attribute__] = ACTIONS(4262),
    [anon_sym___attribute] = ACTIONS(4262),
    [anon_sym_noreturn] = ACTIONS(4262),
    [anon_sym_LBRACK] = ACTIONS(4264),
    [anon_sym___declspec] = ACTIONS(4262),
    [anon_sym_LBRACE] = ACTIONS(4264),
    [anon_sym_signed] = ACTIONS(4262),
    [anon_sym_unsigned] = ACTIONS(4262),
    [anon_sym_long] = ACTIONS(4262),
    [anon_sym_short] = ACTIONS(4262),
    [anon_sym_static] = ACTIONS(4262),
    [anon_sym_ATautoreleasepool] = ACTIONS(4264),
    [anon_sym_auto] = ACTIONS(4262),
    [anon_sym_register] = ACTIONS(4262),
    [anon_sym_inline] = ACTIONS(4262),
    [anon_sym___inline] = ACTIONS(4262),
    [anon_sym___inline__] = ACTIONS(4262),
    [anon_sym___forceinline] = ACTIONS(4262),
    [anon_sym_thread_local] = ACTIONS(4262),
    [anon_sym___thread] = ACTIONS(4262),
    [anon_sym_CG_EXTERN] = ACTIONS(4262),
    [anon_sym_CG_INLINE] = ACTIONS(4262),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4262),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4262),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4262),
    [anon_sym_IBOutlet] = ACTIONS(4262),
    [anon_sym_IBInspectable] = ACTIONS(4262),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4262),
    [anon_sym_NS_INLINE] = ACTIONS(4262),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4262),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4262),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4262),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4262),
    [anon_sym_const] = ACTIONS(4262),
    [anon_sym_constexpr] = ACTIONS(4262),
    [anon_sym_volatile] = ACTIONS(4262),
    [anon_sym_restrict] = ACTIONS(4262),
    [anon_sym___restrict__] = ACTIONS(4262),
    [anon_sym__Atomic] = ACTIONS(4262),
    [anon_sym__Noreturn] = ACTIONS(4262),
    [anon_sym__Nonnull] = ACTIONS(4262),
    [anon_sym_nullable] = ACTIONS(4262),
    [anon_sym__Complex] = ACTIONS(4262),
    [anon_sym__Nullable] = ACTIONS(4262),
    [anon_sym__Nullable_result] = ACTIONS(4262),
    [anon_sym__Null_unspecified] = ACTIONS(4262),
    [anon_sym___autoreleasing] = ACTIONS(4262),
    [anon_sym___block] = ACTIONS(4262),
    [anon_sym___bridge] = ACTIONS(4262),
    [anon_sym___bridge_retained] = ACTIONS(4262),
    [anon_sym___bridge_transfer] = ACTIONS(4262),
    [anon_sym___complex] = ACTIONS(4262),
    [anon_sym___const] = ACTIONS(4262),
    [anon_sym___imag] = ACTIONS(4262),
    [anon_sym___kindof] = ACTIONS(4262),
    [anon_sym___nonnull] = ACTIONS(4262),
    [anon_sym___nullable] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4262),
    [anon_sym___real] = ACTIONS(4262),
    [anon_sym___strong] = ACTIONS(4262),
    [anon_sym___unsafe_unretained] = ACTIONS(4262),
    [anon_sym___unused] = ACTIONS(4262),
    [anon_sym___weak] = ACTIONS(4262),
    [anon_sym_alignas] = ACTIONS(4262),
    [anon_sym__Alignas] = ACTIONS(4262),
    [sym_primitive_type] = ACTIONS(4262),
    [anon_sym_enum] = ACTIONS(4262),
    [anon_sym_struct] = ACTIONS(4262),
    [anon_sym_union] = ACTIONS(4262),
    [anon_sym_QMARK] = ACTIONS(4264),
    [anon_sym_DASH_DASH] = ACTIONS(4264),
    [anon_sym_PLUS_PLUS] = ACTIONS(4264),
    [anon_sym_DOT] = ACTIONS(4262),
    [anon_sym_DASH_GT] = ACTIONS(4264),
    [anon_sym_AT] = ACTIONS(4262),
    [anon_sym_DQUOTE] = ACTIONS(4264),
    [anon_sym_L_DQUOTE] = ACTIONS(4264),
    [anon_sym_u_DQUOTE] = ACTIONS(4264),
    [anon_sym_U_DQUOTE] = ACTIONS(4264),
    [anon_sym_u8_DQUOTE] = ACTIONS(4264),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4262),
    [anon_sym___typeof] = ACTIONS(4262),
    [anon_sym_typeof] = ACTIONS(4262),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4262),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4262),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4262),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4262),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4262),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4262),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4262),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_API_AVAILABLE] = ACTIONS(4262),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_API_DEPRECATED] = ACTIONS(4262),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4262),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym___deprecated_msg] = ACTIONS(4262),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4262),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4262),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4262),
    [anon_sym_BOOL] = ACTIONS(4262),
    [anon_sym_IMP] = ACTIONS(4262),
    [anon_sym_SEL] = ACTIONS(4262),
    [anon_sym_Class] = ACTIONS(4262),
    [anon_sym_id] = ACTIONS(4262),
  },
  [STATE(1654)] = {
    [sym_identifier] = ACTIONS(4184),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4186),
    [anon_sym_COMMA] = ACTIONS(4186),
    [anon_sym_RPAREN] = ACTIONS(4186),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4186),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4186),
    [anon_sym_LPAREN2] = ACTIONS(4186),
    [anon_sym_DASH] = ACTIONS(4184),
    [anon_sym_PLUS] = ACTIONS(4184),
    [anon_sym_STAR] = ACTIONS(4186),
    [anon_sym_SLASH] = ACTIONS(4184),
    [anon_sym_PERCENT] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE] = ACTIONS(4186),
    [anon_sym_AMP_AMP] = ACTIONS(4186),
    [anon_sym_PIPE] = ACTIONS(4184),
    [anon_sym_CARET] = ACTIONS(4186),
    [anon_sym_AMP] = ACTIONS(4184),
    [anon_sym_EQ_EQ] = ACTIONS(4186),
    [anon_sym_BANG_EQ] = ACTIONS(4186),
    [anon_sym_GT] = ACTIONS(4184),
    [anon_sym_GT_EQ] = ACTIONS(4186),
    [anon_sym_LT_EQ] = ACTIONS(4186),
    [anon_sym_LT] = ACTIONS(4184),
    [anon_sym_LT_LT] = ACTIONS(4186),
    [anon_sym_GT_GT] = ACTIONS(4186),
    [anon_sym_SEMI] = ACTIONS(4186),
    [anon_sym___extension__] = ACTIONS(4184),
    [anon_sym_extern] = ACTIONS(4184),
    [anon_sym___attribute__] = ACTIONS(4184),
    [anon_sym___attribute] = ACTIONS(4184),
    [anon_sym_noreturn] = ACTIONS(4184),
    [anon_sym_LBRACK] = ACTIONS(4186),
    [anon_sym_RBRACK] = ACTIONS(4186),
    [anon_sym___declspec] = ACTIONS(4184),
    [anon_sym_LBRACE] = ACTIONS(4186),
    [anon_sym_RBRACE] = ACTIONS(4186),
    [anon_sym_signed] = ACTIONS(4184),
    [anon_sym_unsigned] = ACTIONS(4184),
    [anon_sym_long] = ACTIONS(4184),
    [anon_sym_short] = ACTIONS(4184),
    [anon_sym_static] = ACTIONS(4184),
    [anon_sym_ATautoreleasepool] = ACTIONS(4186),
    [anon_sym_auto] = ACTIONS(4184),
    [anon_sym_register] = ACTIONS(4184),
    [anon_sym_inline] = ACTIONS(4184),
    [anon_sym___inline] = ACTIONS(4184),
    [anon_sym___inline__] = ACTIONS(4184),
    [anon_sym___forceinline] = ACTIONS(4184),
    [anon_sym_thread_local] = ACTIONS(4184),
    [anon_sym___thread] = ACTIONS(4184),
    [anon_sym_CG_EXTERN] = ACTIONS(4184),
    [anon_sym_CG_INLINE] = ACTIONS(4184),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4184),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4184),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4184),
    [anon_sym_IBOutlet] = ACTIONS(4184),
    [anon_sym_IBInspectable] = ACTIONS(4184),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4184),
    [anon_sym_NS_INLINE] = ACTIONS(4184),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4184),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4184),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4184),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4184),
    [anon_sym_const] = ACTIONS(4184),
    [anon_sym_constexpr] = ACTIONS(4184),
    [anon_sym_volatile] = ACTIONS(4184),
    [anon_sym_restrict] = ACTIONS(4184),
    [anon_sym___restrict__] = ACTIONS(4184),
    [anon_sym__Atomic] = ACTIONS(4184),
    [anon_sym__Noreturn] = ACTIONS(4184),
    [anon_sym__Nonnull] = ACTIONS(4184),
    [anon_sym_nullable] = ACTIONS(4184),
    [anon_sym__Complex] = ACTIONS(4184),
    [anon_sym__Nullable] = ACTIONS(4184),
    [anon_sym__Nullable_result] = ACTIONS(4184),
    [anon_sym__Null_unspecified] = ACTIONS(4184),
    [anon_sym___autoreleasing] = ACTIONS(4184),
    [anon_sym___block] = ACTIONS(4184),
    [anon_sym___bridge] = ACTIONS(4184),
    [anon_sym___bridge_retained] = ACTIONS(4184),
    [anon_sym___bridge_transfer] = ACTIONS(4184),
    [anon_sym___complex] = ACTIONS(4184),
    [anon_sym___const] = ACTIONS(4184),
    [anon_sym___imag] = ACTIONS(4184),
    [anon_sym___kindof] = ACTIONS(4184),
    [anon_sym___nonnull] = ACTIONS(4184),
    [anon_sym___nullable] = ACTIONS(4184),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4184),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4184),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4184),
    [anon_sym___real] = ACTIONS(4184),
    [anon_sym___strong] = ACTIONS(4184),
    [anon_sym___unsafe_unretained] = ACTIONS(4184),
    [anon_sym___unused] = ACTIONS(4184),
    [anon_sym___weak] = ACTIONS(4184),
    [anon_sym_alignas] = ACTIONS(4184),
    [anon_sym__Alignas] = ACTIONS(4184),
    [sym_primitive_type] = ACTIONS(4184),
    [anon_sym_enum] = ACTIONS(4184),
    [anon_sym_COLON] = ACTIONS(4186),
    [anon_sym_struct] = ACTIONS(4184),
    [anon_sym_union] = ACTIONS(4184),
    [anon_sym_QMARK] = ACTIONS(4186),
    [anon_sym_DASH_DASH] = ACTIONS(4186),
    [anon_sym_PLUS_PLUS] = ACTIONS(4186),
    [anon_sym_DOT] = ACTIONS(4184),
    [anon_sym_DASH_GT] = ACTIONS(4186),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4184),
    [anon_sym___typeof] = ACTIONS(4184),
    [anon_sym_typeof] = ACTIONS(4184),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4184),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4184),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4184),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4184),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4184),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4184),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4184),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4184),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4184),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4184),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4184),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4184),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4184),
    [anon_sym_API_AVAILABLE] = ACTIONS(4184),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4184),
    [anon_sym_API_DEPRECATED] = ACTIONS(4184),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4184),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4184),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4184),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4184),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4184),
    [anon_sym___deprecated_msg] = ACTIONS(4184),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4184),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4184),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4184),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4184),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4184),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4184),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4184),
    [anon_sym_BOOL] = ACTIONS(4184),
    [anon_sym_IMP] = ACTIONS(4184),
    [anon_sym_SEL] = ACTIONS(4184),
    [anon_sym_Class] = ACTIONS(4184),
    [anon_sym_id] = ACTIONS(4184),
  },
  [STATE(1655)] = {
    [sym_identifier] = ACTIONS(4188),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4190),
    [anon_sym_COMMA] = ACTIONS(4190),
    [anon_sym_RPAREN] = ACTIONS(4190),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4190),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4190),
    [anon_sym_LPAREN2] = ACTIONS(4190),
    [anon_sym_DASH] = ACTIONS(4188),
    [anon_sym_PLUS] = ACTIONS(4188),
    [anon_sym_STAR] = ACTIONS(4190),
    [anon_sym_SLASH] = ACTIONS(4188),
    [anon_sym_PERCENT] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE] = ACTIONS(4190),
    [anon_sym_AMP_AMP] = ACTIONS(4190),
    [anon_sym_PIPE] = ACTIONS(4188),
    [anon_sym_CARET] = ACTIONS(4190),
    [anon_sym_AMP] = ACTIONS(4188),
    [anon_sym_EQ_EQ] = ACTIONS(4190),
    [anon_sym_BANG_EQ] = ACTIONS(4190),
    [anon_sym_GT] = ACTIONS(4188),
    [anon_sym_GT_EQ] = ACTIONS(4190),
    [anon_sym_LT_EQ] = ACTIONS(4190),
    [anon_sym_LT] = ACTIONS(4188),
    [anon_sym_LT_LT] = ACTIONS(4190),
    [anon_sym_GT_GT] = ACTIONS(4190),
    [anon_sym_SEMI] = ACTIONS(4190),
    [anon_sym___extension__] = ACTIONS(4188),
    [anon_sym_extern] = ACTIONS(4188),
    [anon_sym___attribute__] = ACTIONS(4188),
    [anon_sym___attribute] = ACTIONS(4188),
    [anon_sym_noreturn] = ACTIONS(4188),
    [anon_sym_LBRACK] = ACTIONS(4190),
    [anon_sym_RBRACK] = ACTIONS(4190),
    [anon_sym___declspec] = ACTIONS(4188),
    [anon_sym_LBRACE] = ACTIONS(4190),
    [anon_sym_RBRACE] = ACTIONS(4190),
    [anon_sym_signed] = ACTIONS(4188),
    [anon_sym_unsigned] = ACTIONS(4188),
    [anon_sym_long] = ACTIONS(4188),
    [anon_sym_short] = ACTIONS(4188),
    [anon_sym_static] = ACTIONS(4188),
    [anon_sym_ATautoreleasepool] = ACTIONS(4190),
    [anon_sym_auto] = ACTIONS(4188),
    [anon_sym_register] = ACTIONS(4188),
    [anon_sym_inline] = ACTIONS(4188),
    [anon_sym___inline] = ACTIONS(4188),
    [anon_sym___inline__] = ACTIONS(4188),
    [anon_sym___forceinline] = ACTIONS(4188),
    [anon_sym_thread_local] = ACTIONS(4188),
    [anon_sym___thread] = ACTIONS(4188),
    [anon_sym_CG_EXTERN] = ACTIONS(4188),
    [anon_sym_CG_INLINE] = ACTIONS(4188),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4188),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4188),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4188),
    [anon_sym_IBOutlet] = ACTIONS(4188),
    [anon_sym_IBInspectable] = ACTIONS(4188),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4188),
    [anon_sym_NS_INLINE] = ACTIONS(4188),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4188),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4188),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4188),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4188),
    [anon_sym_const] = ACTIONS(4188),
    [anon_sym_constexpr] = ACTIONS(4188),
    [anon_sym_volatile] = ACTIONS(4188),
    [anon_sym_restrict] = ACTIONS(4188),
    [anon_sym___restrict__] = ACTIONS(4188),
    [anon_sym__Atomic] = ACTIONS(4188),
    [anon_sym__Noreturn] = ACTIONS(4188),
    [anon_sym__Nonnull] = ACTIONS(4188),
    [anon_sym_nullable] = ACTIONS(4188),
    [anon_sym__Complex] = ACTIONS(4188),
    [anon_sym__Nullable] = ACTIONS(4188),
    [anon_sym__Nullable_result] = ACTIONS(4188),
    [anon_sym__Null_unspecified] = ACTIONS(4188),
    [anon_sym___autoreleasing] = ACTIONS(4188),
    [anon_sym___block] = ACTIONS(4188),
    [anon_sym___bridge] = ACTIONS(4188),
    [anon_sym___bridge_retained] = ACTIONS(4188),
    [anon_sym___bridge_transfer] = ACTIONS(4188),
    [anon_sym___complex] = ACTIONS(4188),
    [anon_sym___const] = ACTIONS(4188),
    [anon_sym___imag] = ACTIONS(4188),
    [anon_sym___kindof] = ACTIONS(4188),
    [anon_sym___nonnull] = ACTIONS(4188),
    [anon_sym___nullable] = ACTIONS(4188),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4188),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4188),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4188),
    [anon_sym___real] = ACTIONS(4188),
    [anon_sym___strong] = ACTIONS(4188),
    [anon_sym___unsafe_unretained] = ACTIONS(4188),
    [anon_sym___unused] = ACTIONS(4188),
    [anon_sym___weak] = ACTIONS(4188),
    [anon_sym_alignas] = ACTIONS(4188),
    [anon_sym__Alignas] = ACTIONS(4188),
    [sym_primitive_type] = ACTIONS(4188),
    [anon_sym_enum] = ACTIONS(4188),
    [anon_sym_COLON] = ACTIONS(4190),
    [anon_sym_struct] = ACTIONS(4188),
    [anon_sym_union] = ACTIONS(4188),
    [anon_sym_QMARK] = ACTIONS(4190),
    [anon_sym_DASH_DASH] = ACTIONS(4190),
    [anon_sym_PLUS_PLUS] = ACTIONS(4190),
    [anon_sym_DOT] = ACTIONS(4188),
    [anon_sym_DASH_GT] = ACTIONS(4190),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4188),
    [anon_sym___typeof] = ACTIONS(4188),
    [anon_sym_typeof] = ACTIONS(4188),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4188),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4188),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4188),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4188),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4188),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4188),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4188),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4188),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4188),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4188),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4188),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4188),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4188),
    [anon_sym_API_AVAILABLE] = ACTIONS(4188),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4188),
    [anon_sym_API_DEPRECATED] = ACTIONS(4188),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4188),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4188),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4188),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4188),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4188),
    [anon_sym___deprecated_msg] = ACTIONS(4188),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4188),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4188),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4188),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4188),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4188),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4188),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4188),
    [anon_sym_BOOL] = ACTIONS(4188),
    [anon_sym_IMP] = ACTIONS(4188),
    [anon_sym_SEL] = ACTIONS(4188),
    [anon_sym_Class] = ACTIONS(4188),
    [anon_sym_id] = ACTIONS(4188),
  },
  [STATE(1656)] = {
    [sym_identifier] = ACTIONS(4126),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4128),
    [anon_sym_COMMA] = ACTIONS(4128),
    [anon_sym_RPAREN] = ACTIONS(4128),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4128),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4128),
    [anon_sym_LPAREN2] = ACTIONS(4128),
    [anon_sym_DASH] = ACTIONS(4126),
    [anon_sym_PLUS] = ACTIONS(4126),
    [anon_sym_STAR] = ACTIONS(4128),
    [anon_sym_SLASH] = ACTIONS(4126),
    [anon_sym_PERCENT] = ACTIONS(4128),
    [anon_sym_PIPE_PIPE] = ACTIONS(4128),
    [anon_sym_AMP_AMP] = ACTIONS(4128),
    [anon_sym_PIPE] = ACTIONS(4126),
    [anon_sym_CARET] = ACTIONS(4128),
    [anon_sym_AMP] = ACTIONS(4126),
    [anon_sym_EQ_EQ] = ACTIONS(4128),
    [anon_sym_BANG_EQ] = ACTIONS(4128),
    [anon_sym_GT] = ACTIONS(4126),
    [anon_sym_GT_EQ] = ACTIONS(4128),
    [anon_sym_LT_EQ] = ACTIONS(4128),
    [anon_sym_LT] = ACTIONS(4126),
    [anon_sym_LT_LT] = ACTIONS(4128),
    [anon_sym_GT_GT] = ACTIONS(4128),
    [anon_sym_SEMI] = ACTIONS(4128),
    [anon_sym___extension__] = ACTIONS(4126),
    [anon_sym_extern] = ACTIONS(4126),
    [anon_sym___attribute__] = ACTIONS(4126),
    [anon_sym___attribute] = ACTIONS(4126),
    [anon_sym_noreturn] = ACTIONS(4126),
    [anon_sym_LBRACK] = ACTIONS(4128),
    [anon_sym_RBRACK] = ACTIONS(4128),
    [anon_sym___declspec] = ACTIONS(4126),
    [anon_sym_LBRACE] = ACTIONS(4128),
    [anon_sym_RBRACE] = ACTIONS(4128),
    [anon_sym_signed] = ACTIONS(4126),
    [anon_sym_unsigned] = ACTIONS(4126),
    [anon_sym_long] = ACTIONS(4126),
    [anon_sym_short] = ACTIONS(4126),
    [anon_sym_static] = ACTIONS(4126),
    [anon_sym_ATautoreleasepool] = ACTIONS(4128),
    [anon_sym_auto] = ACTIONS(4126),
    [anon_sym_register] = ACTIONS(4126),
    [anon_sym_inline] = ACTIONS(4126),
    [anon_sym___inline] = ACTIONS(4126),
    [anon_sym___inline__] = ACTIONS(4126),
    [anon_sym___forceinline] = ACTIONS(4126),
    [anon_sym_thread_local] = ACTIONS(4126),
    [anon_sym___thread] = ACTIONS(4126),
    [anon_sym_CG_EXTERN] = ACTIONS(4126),
    [anon_sym_CG_INLINE] = ACTIONS(4126),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4126),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4126),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4126),
    [anon_sym_IBOutlet] = ACTIONS(4126),
    [anon_sym_IBInspectable] = ACTIONS(4126),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4126),
    [anon_sym_NS_INLINE] = ACTIONS(4126),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4126),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4126),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4126),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4126),
    [anon_sym_const] = ACTIONS(4126),
    [anon_sym_constexpr] = ACTIONS(4126),
    [anon_sym_volatile] = ACTIONS(4126),
    [anon_sym_restrict] = ACTIONS(4126),
    [anon_sym___restrict__] = ACTIONS(4126),
    [anon_sym__Atomic] = ACTIONS(4126),
    [anon_sym__Noreturn] = ACTIONS(4126),
    [anon_sym__Nonnull] = ACTIONS(4126),
    [anon_sym_nullable] = ACTIONS(4126),
    [anon_sym__Complex] = ACTIONS(4126),
    [anon_sym__Nullable] = ACTIONS(4126),
    [anon_sym__Nullable_result] = ACTIONS(4126),
    [anon_sym__Null_unspecified] = ACTIONS(4126),
    [anon_sym___autoreleasing] = ACTIONS(4126),
    [anon_sym___block] = ACTIONS(4126),
    [anon_sym___bridge] = ACTIONS(4126),
    [anon_sym___bridge_retained] = ACTIONS(4126),
    [anon_sym___bridge_transfer] = ACTIONS(4126),
    [anon_sym___complex] = ACTIONS(4126),
    [anon_sym___const] = ACTIONS(4126),
    [anon_sym___imag] = ACTIONS(4126),
    [anon_sym___kindof] = ACTIONS(4126),
    [anon_sym___nonnull] = ACTIONS(4126),
    [anon_sym___nullable] = ACTIONS(4126),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4126),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4126),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4126),
    [anon_sym___real] = ACTIONS(4126),
    [anon_sym___strong] = ACTIONS(4126),
    [anon_sym___unsafe_unretained] = ACTIONS(4126),
    [anon_sym___unused] = ACTIONS(4126),
    [anon_sym___weak] = ACTIONS(4126),
    [anon_sym_alignas] = ACTIONS(4126),
    [anon_sym__Alignas] = ACTIONS(4126),
    [sym_primitive_type] = ACTIONS(4126),
    [anon_sym_enum] = ACTIONS(4126),
    [anon_sym_COLON] = ACTIONS(4128),
    [anon_sym_struct] = ACTIONS(4126),
    [anon_sym_union] = ACTIONS(4126),
    [anon_sym_QMARK] = ACTIONS(4128),
    [anon_sym_DASH_DASH] = ACTIONS(4128),
    [anon_sym_PLUS_PLUS] = ACTIONS(4128),
    [anon_sym_DOT] = ACTIONS(4126),
    [anon_sym_DASH_GT] = ACTIONS(4128),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4126),
    [anon_sym___typeof] = ACTIONS(4126),
    [anon_sym_typeof] = ACTIONS(4126),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4126),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4126),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4126),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4126),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4126),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4126),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4126),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4126),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4126),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4126),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4126),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4126),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4126),
    [anon_sym_API_AVAILABLE] = ACTIONS(4126),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4126),
    [anon_sym_API_DEPRECATED] = ACTIONS(4126),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4126),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4126),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4126),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4126),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4126),
    [anon_sym___deprecated_msg] = ACTIONS(4126),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4126),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4126),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4126),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4126),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4126),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4126),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4126),
    [anon_sym_BOOL] = ACTIONS(4126),
    [anon_sym_IMP] = ACTIONS(4126),
    [anon_sym_SEL] = ACTIONS(4126),
    [anon_sym_Class] = ACTIONS(4126),
    [anon_sym_id] = ACTIONS(4126),
  },
  [STATE(1657)] = {
    [sym_identifier] = ACTIONS(4118),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4120),
    [anon_sym_COMMA] = ACTIONS(4120),
    [anon_sym_LPAREN2] = ACTIONS(4120),
    [anon_sym_DASH] = ACTIONS(4118),
    [anon_sym_PLUS] = ACTIONS(4118),
    [anon_sym_STAR] = ACTIONS(4120),
    [anon_sym_SLASH] = ACTIONS(4118),
    [anon_sym_PERCENT] = ACTIONS(4120),
    [anon_sym_PIPE_PIPE] = ACTIONS(4120),
    [anon_sym_AMP_AMP] = ACTIONS(4120),
    [anon_sym_PIPE] = ACTIONS(4118),
    [anon_sym_CARET] = ACTIONS(4120),
    [anon_sym_AMP] = ACTIONS(4118),
    [anon_sym_EQ_EQ] = ACTIONS(4120),
    [anon_sym_BANG_EQ] = ACTIONS(4120),
    [anon_sym_GT] = ACTIONS(4118),
    [anon_sym_GT_EQ] = ACTIONS(4120),
    [anon_sym_LT_EQ] = ACTIONS(4120),
    [anon_sym_LT] = ACTIONS(4118),
    [anon_sym_LT_LT] = ACTIONS(4120),
    [anon_sym_GT_GT] = ACTIONS(4120),
    [anon_sym_SEMI] = ACTIONS(4120),
    [anon_sym___extension__] = ACTIONS(4118),
    [anon_sym_extern] = ACTIONS(4118),
    [anon_sym___attribute__] = ACTIONS(4118),
    [anon_sym___attribute] = ACTIONS(4118),
    [anon_sym_noreturn] = ACTIONS(4118),
    [anon_sym_LBRACK] = ACTIONS(4120),
    [anon_sym___declspec] = ACTIONS(4118),
    [anon_sym_LBRACE] = ACTIONS(4120),
    [anon_sym_signed] = ACTIONS(4118),
    [anon_sym_unsigned] = ACTIONS(4118),
    [anon_sym_long] = ACTIONS(4118),
    [anon_sym_short] = ACTIONS(4118),
    [anon_sym_static] = ACTIONS(4118),
    [anon_sym_ATautoreleasepool] = ACTIONS(4120),
    [anon_sym_auto] = ACTIONS(4118),
    [anon_sym_register] = ACTIONS(4118),
    [anon_sym_inline] = ACTIONS(4118),
    [anon_sym___inline] = ACTIONS(4118),
    [anon_sym___inline__] = ACTIONS(4118),
    [anon_sym___forceinline] = ACTIONS(4118),
    [anon_sym_thread_local] = ACTIONS(4118),
    [anon_sym___thread] = ACTIONS(4118),
    [anon_sym_CG_EXTERN] = ACTIONS(4118),
    [anon_sym_CG_INLINE] = ACTIONS(4118),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4118),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4118),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4118),
    [anon_sym_IBOutlet] = ACTIONS(4118),
    [anon_sym_IBInspectable] = ACTIONS(4118),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4118),
    [anon_sym_NS_INLINE] = ACTIONS(4118),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4118),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4118),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4118),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4118),
    [anon_sym_const] = ACTIONS(4118),
    [anon_sym_constexpr] = ACTIONS(4118),
    [anon_sym_volatile] = ACTIONS(4118),
    [anon_sym_restrict] = ACTIONS(4118),
    [anon_sym___restrict__] = ACTIONS(4118),
    [anon_sym__Atomic] = ACTIONS(4118),
    [anon_sym__Noreturn] = ACTIONS(4118),
    [anon_sym__Nonnull] = ACTIONS(4118),
    [anon_sym_nullable] = ACTIONS(4118),
    [anon_sym__Complex] = ACTIONS(4118),
    [anon_sym__Nullable] = ACTIONS(4118),
    [anon_sym__Nullable_result] = ACTIONS(4118),
    [anon_sym__Null_unspecified] = ACTIONS(4118),
    [anon_sym___autoreleasing] = ACTIONS(4118),
    [anon_sym___block] = ACTIONS(4118),
    [anon_sym___bridge] = ACTIONS(4118),
    [anon_sym___bridge_retained] = ACTIONS(4118),
    [anon_sym___bridge_transfer] = ACTIONS(4118),
    [anon_sym___complex] = ACTIONS(4118),
    [anon_sym___const] = ACTIONS(4118),
    [anon_sym___imag] = ACTIONS(4118),
    [anon_sym___kindof] = ACTIONS(4118),
    [anon_sym___nonnull] = ACTIONS(4118),
    [anon_sym___nullable] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4118),
    [anon_sym___real] = ACTIONS(4118),
    [anon_sym___strong] = ACTIONS(4118),
    [anon_sym___unsafe_unretained] = ACTIONS(4118),
    [anon_sym___unused] = ACTIONS(4118),
    [anon_sym___weak] = ACTIONS(4118),
    [anon_sym_alignas] = ACTIONS(4118),
    [anon_sym__Alignas] = ACTIONS(4118),
    [sym_primitive_type] = ACTIONS(4118),
    [anon_sym_enum] = ACTIONS(4118),
    [anon_sym_struct] = ACTIONS(4118),
    [anon_sym_union] = ACTIONS(4118),
    [anon_sym_QMARK] = ACTIONS(4120),
    [anon_sym_DASH_DASH] = ACTIONS(4120),
    [anon_sym_PLUS_PLUS] = ACTIONS(4120),
    [anon_sym_DOT] = ACTIONS(4118),
    [anon_sym_DASH_GT] = ACTIONS(4120),
    [anon_sym_AT] = ACTIONS(4118),
    [anon_sym_DQUOTE] = ACTIONS(4120),
    [anon_sym_L_DQUOTE] = ACTIONS(4120),
    [anon_sym_u_DQUOTE] = ACTIONS(4120),
    [anon_sym_U_DQUOTE] = ACTIONS(4120),
    [anon_sym_u8_DQUOTE] = ACTIONS(4120),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4118),
    [anon_sym___typeof] = ACTIONS(4118),
    [anon_sym_typeof] = ACTIONS(4118),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4118),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4118),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4118),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4118),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4118),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4118),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4118),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_API_AVAILABLE] = ACTIONS(4118),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_API_DEPRECATED] = ACTIONS(4118),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4118),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym___deprecated_msg] = ACTIONS(4118),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4118),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4118),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4118),
    [anon_sym_BOOL] = ACTIONS(4118),
    [anon_sym_IMP] = ACTIONS(4118),
    [anon_sym_SEL] = ACTIONS(4118),
    [anon_sym_Class] = ACTIONS(4118),
    [anon_sym_id] = ACTIONS(4118),
  },
  [STATE(1658)] = {
    [sym_identifier] = ACTIONS(4258),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4260),
    [anon_sym_COMMA] = ACTIONS(4260),
    [anon_sym_LPAREN2] = ACTIONS(4260),
    [anon_sym_DASH] = ACTIONS(4258),
    [anon_sym_PLUS] = ACTIONS(4258),
    [anon_sym_STAR] = ACTIONS(4260),
    [anon_sym_SLASH] = ACTIONS(4258),
    [anon_sym_PERCENT] = ACTIONS(4260),
    [anon_sym_PIPE_PIPE] = ACTIONS(4260),
    [anon_sym_AMP_AMP] = ACTIONS(4260),
    [anon_sym_PIPE] = ACTIONS(4258),
    [anon_sym_CARET] = ACTIONS(4260),
    [anon_sym_AMP] = ACTIONS(4258),
    [anon_sym_EQ_EQ] = ACTIONS(4260),
    [anon_sym_BANG_EQ] = ACTIONS(4260),
    [anon_sym_GT] = ACTIONS(4258),
    [anon_sym_GT_EQ] = ACTIONS(4260),
    [anon_sym_LT_EQ] = ACTIONS(4260),
    [anon_sym_LT] = ACTIONS(4258),
    [anon_sym_LT_LT] = ACTIONS(4260),
    [anon_sym_GT_GT] = ACTIONS(4260),
    [anon_sym_SEMI] = ACTIONS(4260),
    [anon_sym___extension__] = ACTIONS(4258),
    [anon_sym_extern] = ACTIONS(4258),
    [anon_sym___attribute__] = ACTIONS(4258),
    [anon_sym___attribute] = ACTIONS(4258),
    [anon_sym_noreturn] = ACTIONS(4258),
    [anon_sym_LBRACK] = ACTIONS(4260),
    [anon_sym___declspec] = ACTIONS(4258),
    [anon_sym_LBRACE] = ACTIONS(4260),
    [anon_sym_signed] = ACTIONS(4258),
    [anon_sym_unsigned] = ACTIONS(4258),
    [anon_sym_long] = ACTIONS(4258),
    [anon_sym_short] = ACTIONS(4258),
    [anon_sym_static] = ACTIONS(4258),
    [anon_sym_ATautoreleasepool] = ACTIONS(4260),
    [anon_sym_auto] = ACTIONS(4258),
    [anon_sym_register] = ACTIONS(4258),
    [anon_sym_inline] = ACTIONS(4258),
    [anon_sym___inline] = ACTIONS(4258),
    [anon_sym___inline__] = ACTIONS(4258),
    [anon_sym___forceinline] = ACTIONS(4258),
    [anon_sym_thread_local] = ACTIONS(4258),
    [anon_sym___thread] = ACTIONS(4258),
    [anon_sym_CG_EXTERN] = ACTIONS(4258),
    [anon_sym_CG_INLINE] = ACTIONS(4258),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4258),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4258),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4258),
    [anon_sym_IBOutlet] = ACTIONS(4258),
    [anon_sym_IBInspectable] = ACTIONS(4258),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4258),
    [anon_sym_NS_INLINE] = ACTIONS(4258),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4258),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4258),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4258),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4258),
    [anon_sym_const] = ACTIONS(4258),
    [anon_sym_constexpr] = ACTIONS(4258),
    [anon_sym_volatile] = ACTIONS(4258),
    [anon_sym_restrict] = ACTIONS(4258),
    [anon_sym___restrict__] = ACTIONS(4258),
    [anon_sym__Atomic] = ACTIONS(4258),
    [anon_sym__Noreturn] = ACTIONS(4258),
    [anon_sym__Nonnull] = ACTIONS(4258),
    [anon_sym_nullable] = ACTIONS(4258),
    [anon_sym__Complex] = ACTIONS(4258),
    [anon_sym__Nullable] = ACTIONS(4258),
    [anon_sym__Nullable_result] = ACTIONS(4258),
    [anon_sym__Null_unspecified] = ACTIONS(4258),
    [anon_sym___autoreleasing] = ACTIONS(4258),
    [anon_sym___block] = ACTIONS(4258),
    [anon_sym___bridge] = ACTIONS(4258),
    [anon_sym___bridge_retained] = ACTIONS(4258),
    [anon_sym___bridge_transfer] = ACTIONS(4258),
    [anon_sym___complex] = ACTIONS(4258),
    [anon_sym___const] = ACTIONS(4258),
    [anon_sym___imag] = ACTIONS(4258),
    [anon_sym___kindof] = ACTIONS(4258),
    [anon_sym___nonnull] = ACTIONS(4258),
    [anon_sym___nullable] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4258),
    [anon_sym___real] = ACTIONS(4258),
    [anon_sym___strong] = ACTIONS(4258),
    [anon_sym___unsafe_unretained] = ACTIONS(4258),
    [anon_sym___unused] = ACTIONS(4258),
    [anon_sym___weak] = ACTIONS(4258),
    [anon_sym_alignas] = ACTIONS(4258),
    [anon_sym__Alignas] = ACTIONS(4258),
    [sym_primitive_type] = ACTIONS(4258),
    [anon_sym_enum] = ACTIONS(4258),
    [anon_sym_struct] = ACTIONS(4258),
    [anon_sym_union] = ACTIONS(4258),
    [anon_sym_QMARK] = ACTIONS(4260),
    [anon_sym_DASH_DASH] = ACTIONS(4260),
    [anon_sym_PLUS_PLUS] = ACTIONS(4260),
    [anon_sym_DOT] = ACTIONS(4258),
    [anon_sym_DASH_GT] = ACTIONS(4260),
    [anon_sym_AT] = ACTIONS(4258),
    [anon_sym_DQUOTE] = ACTIONS(4260),
    [anon_sym_L_DQUOTE] = ACTIONS(4260),
    [anon_sym_u_DQUOTE] = ACTIONS(4260),
    [anon_sym_U_DQUOTE] = ACTIONS(4260),
    [anon_sym_u8_DQUOTE] = ACTIONS(4260),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4258),
    [anon_sym___typeof] = ACTIONS(4258),
    [anon_sym_typeof] = ACTIONS(4258),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4258),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4258),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4258),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4258),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4258),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4258),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4258),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_API_AVAILABLE] = ACTIONS(4258),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_API_DEPRECATED] = ACTIONS(4258),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4258),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym___deprecated_msg] = ACTIONS(4258),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4258),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4258),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4258),
    [anon_sym_BOOL] = ACTIONS(4258),
    [anon_sym_IMP] = ACTIONS(4258),
    [anon_sym_SEL] = ACTIONS(4258),
    [anon_sym_Class] = ACTIONS(4258),
    [anon_sym_id] = ACTIONS(4258),
  },
  [STATE(1659)] = {
    [sym_identifier] = ACTIONS(4248),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4250),
    [anon_sym_COMMA] = ACTIONS(4250),
    [anon_sym_RPAREN] = ACTIONS(4250),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4250),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4250),
    [anon_sym_LPAREN2] = ACTIONS(4250),
    [anon_sym_DASH] = ACTIONS(4248),
    [anon_sym_PLUS] = ACTIONS(4248),
    [anon_sym_STAR] = ACTIONS(4250),
    [anon_sym_SLASH] = ACTIONS(4248),
    [anon_sym_PERCENT] = ACTIONS(4250),
    [anon_sym_PIPE_PIPE] = ACTIONS(4250),
    [anon_sym_AMP_AMP] = ACTIONS(4250),
    [anon_sym_PIPE] = ACTIONS(4248),
    [anon_sym_CARET] = ACTIONS(4250),
    [anon_sym_AMP] = ACTIONS(4248),
    [anon_sym_EQ_EQ] = ACTIONS(4250),
    [anon_sym_BANG_EQ] = ACTIONS(4250),
    [anon_sym_GT] = ACTIONS(4248),
    [anon_sym_GT_EQ] = ACTIONS(4250),
    [anon_sym_LT_EQ] = ACTIONS(4250),
    [anon_sym_LT] = ACTIONS(4248),
    [anon_sym_LT_LT] = ACTIONS(4250),
    [anon_sym_GT_GT] = ACTIONS(4250),
    [anon_sym_SEMI] = ACTIONS(4250),
    [anon_sym___extension__] = ACTIONS(4248),
    [anon_sym_extern] = ACTIONS(4248),
    [anon_sym___attribute__] = ACTIONS(4248),
    [anon_sym___attribute] = ACTIONS(4248),
    [anon_sym_noreturn] = ACTIONS(4248),
    [anon_sym_LBRACK] = ACTIONS(4250),
    [anon_sym_RBRACK] = ACTIONS(4250),
    [anon_sym___declspec] = ACTIONS(4248),
    [anon_sym_LBRACE] = ACTIONS(4250),
    [anon_sym_RBRACE] = ACTIONS(4250),
    [anon_sym_signed] = ACTIONS(4248),
    [anon_sym_unsigned] = ACTIONS(4248),
    [anon_sym_long] = ACTIONS(4248),
    [anon_sym_short] = ACTIONS(4248),
    [anon_sym_static] = ACTIONS(4248),
    [anon_sym_ATautoreleasepool] = ACTIONS(4250),
    [anon_sym_auto] = ACTIONS(4248),
    [anon_sym_register] = ACTIONS(4248),
    [anon_sym_inline] = ACTIONS(4248),
    [anon_sym___inline] = ACTIONS(4248),
    [anon_sym___inline__] = ACTIONS(4248),
    [anon_sym___forceinline] = ACTIONS(4248),
    [anon_sym_thread_local] = ACTIONS(4248),
    [anon_sym___thread] = ACTIONS(4248),
    [anon_sym_CG_EXTERN] = ACTIONS(4248),
    [anon_sym_CG_INLINE] = ACTIONS(4248),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4248),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4248),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4248),
    [anon_sym_IBOutlet] = ACTIONS(4248),
    [anon_sym_IBInspectable] = ACTIONS(4248),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4248),
    [anon_sym_NS_INLINE] = ACTIONS(4248),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4248),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4248),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4248),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4248),
    [anon_sym_const] = ACTIONS(4248),
    [anon_sym_constexpr] = ACTIONS(4248),
    [anon_sym_volatile] = ACTIONS(4248),
    [anon_sym_restrict] = ACTIONS(4248),
    [anon_sym___restrict__] = ACTIONS(4248),
    [anon_sym__Atomic] = ACTIONS(4248),
    [anon_sym__Noreturn] = ACTIONS(4248),
    [anon_sym__Nonnull] = ACTIONS(4248),
    [anon_sym_nullable] = ACTIONS(4248),
    [anon_sym__Complex] = ACTIONS(4248),
    [anon_sym__Nullable] = ACTIONS(4248),
    [anon_sym__Nullable_result] = ACTIONS(4248),
    [anon_sym__Null_unspecified] = ACTIONS(4248),
    [anon_sym___autoreleasing] = ACTIONS(4248),
    [anon_sym___block] = ACTIONS(4248),
    [anon_sym___bridge] = ACTIONS(4248),
    [anon_sym___bridge_retained] = ACTIONS(4248),
    [anon_sym___bridge_transfer] = ACTIONS(4248),
    [anon_sym___complex] = ACTIONS(4248),
    [anon_sym___const] = ACTIONS(4248),
    [anon_sym___imag] = ACTIONS(4248),
    [anon_sym___kindof] = ACTIONS(4248),
    [anon_sym___nonnull] = ACTIONS(4248),
    [anon_sym___nullable] = ACTIONS(4248),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4248),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4248),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4248),
    [anon_sym___real] = ACTIONS(4248),
    [anon_sym___strong] = ACTIONS(4248),
    [anon_sym___unsafe_unretained] = ACTIONS(4248),
    [anon_sym___unused] = ACTIONS(4248),
    [anon_sym___weak] = ACTIONS(4248),
    [anon_sym_alignas] = ACTIONS(4248),
    [anon_sym__Alignas] = ACTIONS(4248),
    [sym_primitive_type] = ACTIONS(4248),
    [anon_sym_enum] = ACTIONS(4248),
    [anon_sym_COLON] = ACTIONS(4250),
    [anon_sym_struct] = ACTIONS(4248),
    [anon_sym_union] = ACTIONS(4248),
    [anon_sym_QMARK] = ACTIONS(4250),
    [anon_sym_DASH_DASH] = ACTIONS(4250),
    [anon_sym_PLUS_PLUS] = ACTIONS(4250),
    [anon_sym_DOT] = ACTIONS(4248),
    [anon_sym_DASH_GT] = ACTIONS(4250),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4248),
    [anon_sym___typeof] = ACTIONS(4248),
    [anon_sym_typeof] = ACTIONS(4248),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4248),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4248),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4248),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4248),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4248),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4248),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4248),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4248),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4248),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4248),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4248),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4248),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4248),
    [anon_sym_API_AVAILABLE] = ACTIONS(4248),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4248),
    [anon_sym_API_DEPRECATED] = ACTIONS(4248),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4248),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4248),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4248),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4248),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4248),
    [anon_sym___deprecated_msg] = ACTIONS(4248),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4248),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4248),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4248),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4248),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4248),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4248),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4248),
    [anon_sym_BOOL] = ACTIONS(4248),
    [anon_sym_IMP] = ACTIONS(4248),
    [anon_sym_SEL] = ACTIONS(4248),
    [anon_sym_Class] = ACTIONS(4248),
    [anon_sym_id] = ACTIONS(4248),
  },
  [STATE(1660)] = {
    [sym_identifier] = ACTIONS(4212),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4214),
    [anon_sym_COMMA] = ACTIONS(4214),
    [anon_sym_RPAREN] = ACTIONS(4214),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4214),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4214),
    [anon_sym_LPAREN2] = ACTIONS(4214),
    [anon_sym_DASH] = ACTIONS(4212),
    [anon_sym_PLUS] = ACTIONS(4212),
    [anon_sym_STAR] = ACTIONS(4214),
    [anon_sym_SLASH] = ACTIONS(4212),
    [anon_sym_PERCENT] = ACTIONS(4214),
    [anon_sym_PIPE_PIPE] = ACTIONS(4214),
    [anon_sym_AMP_AMP] = ACTIONS(4214),
    [anon_sym_PIPE] = ACTIONS(4212),
    [anon_sym_CARET] = ACTIONS(4214),
    [anon_sym_AMP] = ACTIONS(4212),
    [anon_sym_EQ_EQ] = ACTIONS(4214),
    [anon_sym_BANG_EQ] = ACTIONS(4214),
    [anon_sym_GT] = ACTIONS(4212),
    [anon_sym_GT_EQ] = ACTIONS(4214),
    [anon_sym_LT_EQ] = ACTIONS(4214),
    [anon_sym_LT] = ACTIONS(4212),
    [anon_sym_LT_LT] = ACTIONS(4214),
    [anon_sym_GT_GT] = ACTIONS(4214),
    [anon_sym_SEMI] = ACTIONS(4214),
    [anon_sym___extension__] = ACTIONS(4212),
    [anon_sym_extern] = ACTIONS(4212),
    [anon_sym___attribute__] = ACTIONS(4212),
    [anon_sym___attribute] = ACTIONS(4212),
    [anon_sym_noreturn] = ACTIONS(4212),
    [anon_sym_LBRACK] = ACTIONS(4214),
    [anon_sym_RBRACK] = ACTIONS(4214),
    [anon_sym___declspec] = ACTIONS(4212),
    [anon_sym_LBRACE] = ACTIONS(4214),
    [anon_sym_RBRACE] = ACTIONS(4214),
    [anon_sym_signed] = ACTIONS(4212),
    [anon_sym_unsigned] = ACTIONS(4212),
    [anon_sym_long] = ACTIONS(4212),
    [anon_sym_short] = ACTIONS(4212),
    [anon_sym_static] = ACTIONS(4212),
    [anon_sym_ATautoreleasepool] = ACTIONS(4214),
    [anon_sym_auto] = ACTIONS(4212),
    [anon_sym_register] = ACTIONS(4212),
    [anon_sym_inline] = ACTIONS(4212),
    [anon_sym___inline] = ACTIONS(4212),
    [anon_sym___inline__] = ACTIONS(4212),
    [anon_sym___forceinline] = ACTIONS(4212),
    [anon_sym_thread_local] = ACTIONS(4212),
    [anon_sym___thread] = ACTIONS(4212),
    [anon_sym_CG_EXTERN] = ACTIONS(4212),
    [anon_sym_CG_INLINE] = ACTIONS(4212),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4212),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4212),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4212),
    [anon_sym_IBOutlet] = ACTIONS(4212),
    [anon_sym_IBInspectable] = ACTIONS(4212),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4212),
    [anon_sym_NS_INLINE] = ACTIONS(4212),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4212),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4212),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4212),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4212),
    [anon_sym_const] = ACTIONS(4212),
    [anon_sym_constexpr] = ACTIONS(4212),
    [anon_sym_volatile] = ACTIONS(4212),
    [anon_sym_restrict] = ACTIONS(4212),
    [anon_sym___restrict__] = ACTIONS(4212),
    [anon_sym__Atomic] = ACTIONS(4212),
    [anon_sym__Noreturn] = ACTIONS(4212),
    [anon_sym__Nonnull] = ACTIONS(4212),
    [anon_sym_nullable] = ACTIONS(4212),
    [anon_sym__Complex] = ACTIONS(4212),
    [anon_sym__Nullable] = ACTIONS(4212),
    [anon_sym__Nullable_result] = ACTIONS(4212),
    [anon_sym__Null_unspecified] = ACTIONS(4212),
    [anon_sym___autoreleasing] = ACTIONS(4212),
    [anon_sym___block] = ACTIONS(4212),
    [anon_sym___bridge] = ACTIONS(4212),
    [anon_sym___bridge_retained] = ACTIONS(4212),
    [anon_sym___bridge_transfer] = ACTIONS(4212),
    [anon_sym___complex] = ACTIONS(4212),
    [anon_sym___const] = ACTIONS(4212),
    [anon_sym___imag] = ACTIONS(4212),
    [anon_sym___kindof] = ACTIONS(4212),
    [anon_sym___nonnull] = ACTIONS(4212),
    [anon_sym___nullable] = ACTIONS(4212),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4212),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4212),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4212),
    [anon_sym___real] = ACTIONS(4212),
    [anon_sym___strong] = ACTIONS(4212),
    [anon_sym___unsafe_unretained] = ACTIONS(4212),
    [anon_sym___unused] = ACTIONS(4212),
    [anon_sym___weak] = ACTIONS(4212),
    [anon_sym_alignas] = ACTIONS(4212),
    [anon_sym__Alignas] = ACTIONS(4212),
    [sym_primitive_type] = ACTIONS(4212),
    [anon_sym_enum] = ACTIONS(4212),
    [anon_sym_COLON] = ACTIONS(4214),
    [anon_sym_struct] = ACTIONS(4212),
    [anon_sym_union] = ACTIONS(4212),
    [anon_sym_QMARK] = ACTIONS(4214),
    [anon_sym_DASH_DASH] = ACTIONS(4214),
    [anon_sym_PLUS_PLUS] = ACTIONS(4214),
    [anon_sym_DOT] = ACTIONS(4212),
    [anon_sym_DASH_GT] = ACTIONS(4214),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4212),
    [anon_sym___typeof] = ACTIONS(4212),
    [anon_sym_typeof] = ACTIONS(4212),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4212),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4212),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4212),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4212),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4212),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4212),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4212),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4212),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4212),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4212),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4212),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4212),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4212),
    [anon_sym_API_AVAILABLE] = ACTIONS(4212),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4212),
    [anon_sym_API_DEPRECATED] = ACTIONS(4212),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4212),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4212),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4212),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4212),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4212),
    [anon_sym___deprecated_msg] = ACTIONS(4212),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4212),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4212),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4212),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4212),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4212),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4212),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4212),
    [anon_sym_BOOL] = ACTIONS(4212),
    [anon_sym_IMP] = ACTIONS(4212),
    [anon_sym_SEL] = ACTIONS(4212),
    [anon_sym_Class] = ACTIONS(4212),
    [anon_sym_id] = ACTIONS(4212),
  },
  [STATE(1661)] = {
    [sym_identifier] = ACTIONS(4130),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4132),
    [anon_sym_COMMA] = ACTIONS(4132),
    [anon_sym_RPAREN] = ACTIONS(4132),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4132),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4132),
    [anon_sym_LPAREN2] = ACTIONS(4132),
    [anon_sym_DASH] = ACTIONS(4130),
    [anon_sym_PLUS] = ACTIONS(4130),
    [anon_sym_STAR] = ACTIONS(4132),
    [anon_sym_SLASH] = ACTIONS(4130),
    [anon_sym_PERCENT] = ACTIONS(4132),
    [anon_sym_PIPE_PIPE] = ACTIONS(4132),
    [anon_sym_AMP_AMP] = ACTIONS(4132),
    [anon_sym_PIPE] = ACTIONS(4130),
    [anon_sym_CARET] = ACTIONS(4132),
    [anon_sym_AMP] = ACTIONS(4130),
    [anon_sym_EQ_EQ] = ACTIONS(4132),
    [anon_sym_BANG_EQ] = ACTIONS(4132),
    [anon_sym_GT] = ACTIONS(4130),
    [anon_sym_GT_EQ] = ACTIONS(4132),
    [anon_sym_LT_EQ] = ACTIONS(4132),
    [anon_sym_LT] = ACTIONS(4130),
    [anon_sym_LT_LT] = ACTIONS(4132),
    [anon_sym_GT_GT] = ACTIONS(4132),
    [anon_sym_SEMI] = ACTIONS(4132),
    [anon_sym___extension__] = ACTIONS(4130),
    [anon_sym_extern] = ACTIONS(4130),
    [anon_sym___attribute__] = ACTIONS(4130),
    [anon_sym___attribute] = ACTIONS(4130),
    [anon_sym_noreturn] = ACTIONS(4130),
    [anon_sym_LBRACK] = ACTIONS(4132),
    [anon_sym_RBRACK] = ACTIONS(4132),
    [anon_sym___declspec] = ACTIONS(4130),
    [anon_sym_LBRACE] = ACTIONS(4132),
    [anon_sym_RBRACE] = ACTIONS(4132),
    [anon_sym_signed] = ACTIONS(4130),
    [anon_sym_unsigned] = ACTIONS(4130),
    [anon_sym_long] = ACTIONS(4130),
    [anon_sym_short] = ACTIONS(4130),
    [anon_sym_static] = ACTIONS(4130),
    [anon_sym_ATautoreleasepool] = ACTIONS(4132),
    [anon_sym_auto] = ACTIONS(4130),
    [anon_sym_register] = ACTIONS(4130),
    [anon_sym_inline] = ACTIONS(4130),
    [anon_sym___inline] = ACTIONS(4130),
    [anon_sym___inline__] = ACTIONS(4130),
    [anon_sym___forceinline] = ACTIONS(4130),
    [anon_sym_thread_local] = ACTIONS(4130),
    [anon_sym___thread] = ACTIONS(4130),
    [anon_sym_CG_EXTERN] = ACTIONS(4130),
    [anon_sym_CG_INLINE] = ACTIONS(4130),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4130),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4130),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4130),
    [anon_sym_IBOutlet] = ACTIONS(4130),
    [anon_sym_IBInspectable] = ACTIONS(4130),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4130),
    [anon_sym_NS_INLINE] = ACTIONS(4130),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4130),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4130),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4130),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4130),
    [anon_sym_const] = ACTIONS(4130),
    [anon_sym_constexpr] = ACTIONS(4130),
    [anon_sym_volatile] = ACTIONS(4130),
    [anon_sym_restrict] = ACTIONS(4130),
    [anon_sym___restrict__] = ACTIONS(4130),
    [anon_sym__Atomic] = ACTIONS(4130),
    [anon_sym__Noreturn] = ACTIONS(4130),
    [anon_sym__Nonnull] = ACTIONS(4130),
    [anon_sym_nullable] = ACTIONS(4130),
    [anon_sym__Complex] = ACTIONS(4130),
    [anon_sym__Nullable] = ACTIONS(4130),
    [anon_sym__Nullable_result] = ACTIONS(4130),
    [anon_sym__Null_unspecified] = ACTIONS(4130),
    [anon_sym___autoreleasing] = ACTIONS(4130),
    [anon_sym___block] = ACTIONS(4130),
    [anon_sym___bridge] = ACTIONS(4130),
    [anon_sym___bridge_retained] = ACTIONS(4130),
    [anon_sym___bridge_transfer] = ACTIONS(4130),
    [anon_sym___complex] = ACTIONS(4130),
    [anon_sym___const] = ACTIONS(4130),
    [anon_sym___imag] = ACTIONS(4130),
    [anon_sym___kindof] = ACTIONS(4130),
    [anon_sym___nonnull] = ACTIONS(4130),
    [anon_sym___nullable] = ACTIONS(4130),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4130),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4130),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4130),
    [anon_sym___real] = ACTIONS(4130),
    [anon_sym___strong] = ACTIONS(4130),
    [anon_sym___unsafe_unretained] = ACTIONS(4130),
    [anon_sym___unused] = ACTIONS(4130),
    [anon_sym___weak] = ACTIONS(4130),
    [anon_sym_alignas] = ACTIONS(4130),
    [anon_sym__Alignas] = ACTIONS(4130),
    [sym_primitive_type] = ACTIONS(4130),
    [anon_sym_enum] = ACTIONS(4130),
    [anon_sym_COLON] = ACTIONS(4132),
    [anon_sym_struct] = ACTIONS(4130),
    [anon_sym_union] = ACTIONS(4130),
    [anon_sym_QMARK] = ACTIONS(4132),
    [anon_sym_DASH_DASH] = ACTIONS(4132),
    [anon_sym_PLUS_PLUS] = ACTIONS(4132),
    [anon_sym_DOT] = ACTIONS(4130),
    [anon_sym_DASH_GT] = ACTIONS(4132),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4130),
    [anon_sym___typeof] = ACTIONS(4130),
    [anon_sym_typeof] = ACTIONS(4130),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4130),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4130),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4130),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4130),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4130),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4130),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4130),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4130),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4130),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4130),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4130),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4130),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4130),
    [anon_sym_API_AVAILABLE] = ACTIONS(4130),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4130),
    [anon_sym_API_DEPRECATED] = ACTIONS(4130),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4130),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4130),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4130),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4130),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4130),
    [anon_sym___deprecated_msg] = ACTIONS(4130),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4130),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4130),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4130),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4130),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4130),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4130),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4130),
    [anon_sym_BOOL] = ACTIONS(4130),
    [anon_sym_IMP] = ACTIONS(4130),
    [anon_sym_SEL] = ACTIONS(4130),
    [anon_sym_Class] = ACTIONS(4130),
    [anon_sym_id] = ACTIONS(4130),
  },
  [STATE(1662)] = {
    [sym_identifier] = ACTIONS(4236),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4238),
    [anon_sym_COMMA] = ACTIONS(4238),
    [anon_sym_RPAREN] = ACTIONS(4238),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4238),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4238),
    [anon_sym_LPAREN2] = ACTIONS(4238),
    [anon_sym_DASH] = ACTIONS(4236),
    [anon_sym_PLUS] = ACTIONS(4236),
    [anon_sym_STAR] = ACTIONS(4238),
    [anon_sym_SLASH] = ACTIONS(4236),
    [anon_sym_PERCENT] = ACTIONS(4238),
    [anon_sym_PIPE_PIPE] = ACTIONS(4238),
    [anon_sym_AMP_AMP] = ACTIONS(4238),
    [anon_sym_PIPE] = ACTIONS(4236),
    [anon_sym_CARET] = ACTIONS(4238),
    [anon_sym_AMP] = ACTIONS(4236),
    [anon_sym_EQ_EQ] = ACTIONS(4238),
    [anon_sym_BANG_EQ] = ACTIONS(4238),
    [anon_sym_GT] = ACTIONS(4236),
    [anon_sym_GT_EQ] = ACTIONS(4238),
    [anon_sym_LT_EQ] = ACTIONS(4238),
    [anon_sym_LT] = ACTIONS(4236),
    [anon_sym_LT_LT] = ACTIONS(4238),
    [anon_sym_GT_GT] = ACTIONS(4238),
    [anon_sym_SEMI] = ACTIONS(4238),
    [anon_sym___extension__] = ACTIONS(4236),
    [anon_sym_extern] = ACTIONS(4236),
    [anon_sym___attribute__] = ACTIONS(4236),
    [anon_sym___attribute] = ACTIONS(4236),
    [anon_sym_noreturn] = ACTIONS(4236),
    [anon_sym_LBRACK] = ACTIONS(4238),
    [anon_sym_RBRACK] = ACTIONS(4238),
    [anon_sym___declspec] = ACTIONS(4236),
    [anon_sym_LBRACE] = ACTIONS(4238),
    [anon_sym_RBRACE] = ACTIONS(4238),
    [anon_sym_signed] = ACTIONS(4236),
    [anon_sym_unsigned] = ACTIONS(4236),
    [anon_sym_long] = ACTIONS(4236),
    [anon_sym_short] = ACTIONS(4236),
    [anon_sym_static] = ACTIONS(4236),
    [anon_sym_ATautoreleasepool] = ACTIONS(4238),
    [anon_sym_auto] = ACTIONS(4236),
    [anon_sym_register] = ACTIONS(4236),
    [anon_sym_inline] = ACTIONS(4236),
    [anon_sym___inline] = ACTIONS(4236),
    [anon_sym___inline__] = ACTIONS(4236),
    [anon_sym___forceinline] = ACTIONS(4236),
    [anon_sym_thread_local] = ACTIONS(4236),
    [anon_sym___thread] = ACTIONS(4236),
    [anon_sym_CG_EXTERN] = ACTIONS(4236),
    [anon_sym_CG_INLINE] = ACTIONS(4236),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4236),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4236),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4236),
    [anon_sym_IBOutlet] = ACTIONS(4236),
    [anon_sym_IBInspectable] = ACTIONS(4236),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4236),
    [anon_sym_NS_INLINE] = ACTIONS(4236),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4236),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4236),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4236),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4236),
    [anon_sym_const] = ACTIONS(4236),
    [anon_sym_constexpr] = ACTIONS(4236),
    [anon_sym_volatile] = ACTIONS(4236),
    [anon_sym_restrict] = ACTIONS(4236),
    [anon_sym___restrict__] = ACTIONS(4236),
    [anon_sym__Atomic] = ACTIONS(4236),
    [anon_sym__Noreturn] = ACTIONS(4236),
    [anon_sym__Nonnull] = ACTIONS(4236),
    [anon_sym_nullable] = ACTIONS(4236),
    [anon_sym__Complex] = ACTIONS(4236),
    [anon_sym__Nullable] = ACTIONS(4236),
    [anon_sym__Nullable_result] = ACTIONS(4236),
    [anon_sym__Null_unspecified] = ACTIONS(4236),
    [anon_sym___autoreleasing] = ACTIONS(4236),
    [anon_sym___block] = ACTIONS(4236),
    [anon_sym___bridge] = ACTIONS(4236),
    [anon_sym___bridge_retained] = ACTIONS(4236),
    [anon_sym___bridge_transfer] = ACTIONS(4236),
    [anon_sym___complex] = ACTIONS(4236),
    [anon_sym___const] = ACTIONS(4236),
    [anon_sym___imag] = ACTIONS(4236),
    [anon_sym___kindof] = ACTIONS(4236),
    [anon_sym___nonnull] = ACTIONS(4236),
    [anon_sym___nullable] = ACTIONS(4236),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4236),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4236),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4236),
    [anon_sym___real] = ACTIONS(4236),
    [anon_sym___strong] = ACTIONS(4236),
    [anon_sym___unsafe_unretained] = ACTIONS(4236),
    [anon_sym___unused] = ACTIONS(4236),
    [anon_sym___weak] = ACTIONS(4236),
    [anon_sym_alignas] = ACTIONS(4236),
    [anon_sym__Alignas] = ACTIONS(4236),
    [sym_primitive_type] = ACTIONS(4236),
    [anon_sym_enum] = ACTIONS(4236),
    [anon_sym_COLON] = ACTIONS(4238),
    [anon_sym_struct] = ACTIONS(4236),
    [anon_sym_union] = ACTIONS(4236),
    [anon_sym_QMARK] = ACTIONS(4238),
    [anon_sym_DASH_DASH] = ACTIONS(4238),
    [anon_sym_PLUS_PLUS] = ACTIONS(4238),
    [anon_sym_DOT] = ACTIONS(4236),
    [anon_sym_DASH_GT] = ACTIONS(4238),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4236),
    [anon_sym___typeof] = ACTIONS(4236),
    [anon_sym_typeof] = ACTIONS(4236),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4236),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4236),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4236),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4236),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4236),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4236),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4236),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4236),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4236),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4236),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4236),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4236),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4236),
    [anon_sym_API_AVAILABLE] = ACTIONS(4236),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4236),
    [anon_sym_API_DEPRECATED] = ACTIONS(4236),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4236),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4236),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4236),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4236),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4236),
    [anon_sym___deprecated_msg] = ACTIONS(4236),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4236),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4236),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4236),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4236),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4236),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4236),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4236),
    [anon_sym_BOOL] = ACTIONS(4236),
    [anon_sym_IMP] = ACTIONS(4236),
    [anon_sym_SEL] = ACTIONS(4236),
    [anon_sym_Class] = ACTIONS(4236),
    [anon_sym_id] = ACTIONS(4236),
  },
  [STATE(1663)] = {
    [sym_identifier] = ACTIONS(4046),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4048),
    [anon_sym_COMMA] = ACTIONS(4048),
    [anon_sym_RPAREN] = ACTIONS(4048),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4048),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4048),
    [anon_sym_LPAREN2] = ACTIONS(4048),
    [anon_sym_DASH] = ACTIONS(4046),
    [anon_sym_PLUS] = ACTIONS(4046),
    [anon_sym_STAR] = ACTIONS(4048),
    [anon_sym_SLASH] = ACTIONS(4046),
    [anon_sym_PERCENT] = ACTIONS(4048),
    [anon_sym_PIPE_PIPE] = ACTIONS(4048),
    [anon_sym_AMP_AMP] = ACTIONS(4048),
    [anon_sym_PIPE] = ACTIONS(4046),
    [anon_sym_CARET] = ACTIONS(4048),
    [anon_sym_AMP] = ACTIONS(4046),
    [anon_sym_EQ_EQ] = ACTIONS(4048),
    [anon_sym_BANG_EQ] = ACTIONS(4048),
    [anon_sym_GT] = ACTIONS(4046),
    [anon_sym_GT_EQ] = ACTIONS(4048),
    [anon_sym_LT_EQ] = ACTIONS(4048),
    [anon_sym_LT] = ACTIONS(4046),
    [anon_sym_LT_LT] = ACTIONS(4048),
    [anon_sym_GT_GT] = ACTIONS(4048),
    [anon_sym_SEMI] = ACTIONS(4048),
    [anon_sym___extension__] = ACTIONS(4046),
    [anon_sym_extern] = ACTIONS(4046),
    [anon_sym___attribute__] = ACTIONS(4046),
    [anon_sym___attribute] = ACTIONS(4046),
    [anon_sym_noreturn] = ACTIONS(4046),
    [anon_sym_LBRACK] = ACTIONS(4048),
    [anon_sym_RBRACK] = ACTIONS(4048),
    [anon_sym___declspec] = ACTIONS(4046),
    [anon_sym_LBRACE] = ACTIONS(4048),
    [anon_sym_RBRACE] = ACTIONS(4048),
    [anon_sym_signed] = ACTIONS(4046),
    [anon_sym_unsigned] = ACTIONS(4046),
    [anon_sym_long] = ACTIONS(4046),
    [anon_sym_short] = ACTIONS(4046),
    [anon_sym_static] = ACTIONS(4046),
    [anon_sym_ATautoreleasepool] = ACTIONS(4048),
    [anon_sym_auto] = ACTIONS(4046),
    [anon_sym_register] = ACTIONS(4046),
    [anon_sym_inline] = ACTIONS(4046),
    [anon_sym___inline] = ACTIONS(4046),
    [anon_sym___inline__] = ACTIONS(4046),
    [anon_sym___forceinline] = ACTIONS(4046),
    [anon_sym_thread_local] = ACTIONS(4046),
    [anon_sym___thread] = ACTIONS(4046),
    [anon_sym_CG_EXTERN] = ACTIONS(4046),
    [anon_sym_CG_INLINE] = ACTIONS(4046),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4046),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4046),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4046),
    [anon_sym_IBOutlet] = ACTIONS(4046),
    [anon_sym_IBInspectable] = ACTIONS(4046),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4046),
    [anon_sym_NS_INLINE] = ACTIONS(4046),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4046),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4046),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4046),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4046),
    [anon_sym_const] = ACTIONS(4046),
    [anon_sym_constexpr] = ACTIONS(4046),
    [anon_sym_volatile] = ACTIONS(4046),
    [anon_sym_restrict] = ACTIONS(4046),
    [anon_sym___restrict__] = ACTIONS(4046),
    [anon_sym__Atomic] = ACTIONS(4046),
    [anon_sym__Noreturn] = ACTIONS(4046),
    [anon_sym__Nonnull] = ACTIONS(4046),
    [anon_sym_nullable] = ACTIONS(4046),
    [anon_sym__Complex] = ACTIONS(4046),
    [anon_sym__Nullable] = ACTIONS(4046),
    [anon_sym__Nullable_result] = ACTIONS(4046),
    [anon_sym__Null_unspecified] = ACTIONS(4046),
    [anon_sym___autoreleasing] = ACTIONS(4046),
    [anon_sym___block] = ACTIONS(4046),
    [anon_sym___bridge] = ACTIONS(4046),
    [anon_sym___bridge_retained] = ACTIONS(4046),
    [anon_sym___bridge_transfer] = ACTIONS(4046),
    [anon_sym___complex] = ACTIONS(4046),
    [anon_sym___const] = ACTIONS(4046),
    [anon_sym___imag] = ACTIONS(4046),
    [anon_sym___kindof] = ACTIONS(4046),
    [anon_sym___nonnull] = ACTIONS(4046),
    [anon_sym___nullable] = ACTIONS(4046),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4046),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4046),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4046),
    [anon_sym___real] = ACTIONS(4046),
    [anon_sym___strong] = ACTIONS(4046),
    [anon_sym___unsafe_unretained] = ACTIONS(4046),
    [anon_sym___unused] = ACTIONS(4046),
    [anon_sym___weak] = ACTIONS(4046),
    [anon_sym_alignas] = ACTIONS(4046),
    [anon_sym__Alignas] = ACTIONS(4046),
    [sym_primitive_type] = ACTIONS(4046),
    [anon_sym_enum] = ACTIONS(4046),
    [anon_sym_COLON] = ACTIONS(4048),
    [anon_sym_struct] = ACTIONS(4046),
    [anon_sym_union] = ACTIONS(4046),
    [anon_sym_QMARK] = ACTIONS(4048),
    [anon_sym_DASH_DASH] = ACTIONS(4048),
    [anon_sym_PLUS_PLUS] = ACTIONS(4048),
    [anon_sym_DOT] = ACTIONS(4046),
    [anon_sym_DASH_GT] = ACTIONS(4048),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4046),
    [anon_sym___typeof] = ACTIONS(4046),
    [anon_sym_typeof] = ACTIONS(4046),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4046),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4046),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4046),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4046),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4046),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4046),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4046),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4046),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4046),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4046),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4046),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4046),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4046),
    [anon_sym_API_AVAILABLE] = ACTIONS(4046),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4046),
    [anon_sym_API_DEPRECATED] = ACTIONS(4046),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4046),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4046),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4046),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4046),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4046),
    [anon_sym___deprecated_msg] = ACTIONS(4046),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4046),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4046),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4046),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4046),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4046),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4046),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4046),
    [anon_sym_BOOL] = ACTIONS(4046),
    [anon_sym_IMP] = ACTIONS(4046),
    [anon_sym_SEL] = ACTIONS(4046),
    [anon_sym_Class] = ACTIONS(4046),
    [anon_sym_id] = ACTIONS(4046),
  },
  [STATE(1664)] = {
    [sym_identifier] = ACTIONS(4252),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4255),
    [anon_sym_COMMA] = ACTIONS(4255),
    [anon_sym_LPAREN2] = ACTIONS(4255),
    [anon_sym_DASH] = ACTIONS(4252),
    [anon_sym_PLUS] = ACTIONS(4252),
    [anon_sym_STAR] = ACTIONS(4255),
    [anon_sym_SLASH] = ACTIONS(4252),
    [anon_sym_PERCENT] = ACTIONS(4255),
    [anon_sym_PIPE_PIPE] = ACTIONS(4255),
    [anon_sym_AMP_AMP] = ACTIONS(4255),
    [anon_sym_PIPE] = ACTIONS(4252),
    [anon_sym_CARET] = ACTIONS(4255),
    [anon_sym_AMP] = ACTIONS(4252),
    [anon_sym_EQ_EQ] = ACTIONS(4255),
    [anon_sym_BANG_EQ] = ACTIONS(4255),
    [anon_sym_GT] = ACTIONS(4252),
    [anon_sym_GT_EQ] = ACTIONS(4255),
    [anon_sym_LT_EQ] = ACTIONS(4255),
    [anon_sym_LT] = ACTIONS(4252),
    [anon_sym_LT_LT] = ACTIONS(4255),
    [anon_sym_GT_GT] = ACTIONS(4255),
    [anon_sym_SEMI] = ACTIONS(4255),
    [anon_sym___extension__] = ACTIONS(4252),
    [anon_sym_extern] = ACTIONS(4252),
    [anon_sym___attribute__] = ACTIONS(4252),
    [anon_sym___attribute] = ACTIONS(4252),
    [anon_sym_noreturn] = ACTIONS(4252),
    [anon_sym_LBRACK] = ACTIONS(4255),
    [anon_sym___declspec] = ACTIONS(4252),
    [anon_sym_LBRACE] = ACTIONS(4255),
    [anon_sym_signed] = ACTIONS(4252),
    [anon_sym_unsigned] = ACTIONS(4252),
    [anon_sym_long] = ACTIONS(4252),
    [anon_sym_short] = ACTIONS(4252),
    [anon_sym_static] = ACTIONS(4252),
    [anon_sym_ATautoreleasepool] = ACTIONS(4255),
    [anon_sym_auto] = ACTIONS(4252),
    [anon_sym_register] = ACTIONS(4252),
    [anon_sym_inline] = ACTIONS(4252),
    [anon_sym___inline] = ACTIONS(4252),
    [anon_sym___inline__] = ACTIONS(4252),
    [anon_sym___forceinline] = ACTIONS(4252),
    [anon_sym_thread_local] = ACTIONS(4252),
    [anon_sym___thread] = ACTIONS(4252),
    [anon_sym_CG_EXTERN] = ACTIONS(4252),
    [anon_sym_CG_INLINE] = ACTIONS(4252),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4252),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4252),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4252),
    [anon_sym_IBOutlet] = ACTIONS(4252),
    [anon_sym_IBInspectable] = ACTIONS(4252),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4252),
    [anon_sym_NS_INLINE] = ACTIONS(4252),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4252),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4252),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4252),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4252),
    [anon_sym_const] = ACTIONS(4252),
    [anon_sym_constexpr] = ACTIONS(4252),
    [anon_sym_volatile] = ACTIONS(4252),
    [anon_sym_restrict] = ACTIONS(4252),
    [anon_sym___restrict__] = ACTIONS(4252),
    [anon_sym__Atomic] = ACTIONS(4252),
    [anon_sym__Noreturn] = ACTIONS(4252),
    [anon_sym__Nonnull] = ACTIONS(4252),
    [anon_sym_nullable] = ACTIONS(4252),
    [anon_sym__Complex] = ACTIONS(4252),
    [anon_sym__Nullable] = ACTIONS(4252),
    [anon_sym__Nullable_result] = ACTIONS(4252),
    [anon_sym__Null_unspecified] = ACTIONS(4252),
    [anon_sym___autoreleasing] = ACTIONS(4252),
    [anon_sym___block] = ACTIONS(4252),
    [anon_sym___bridge] = ACTIONS(4252),
    [anon_sym___bridge_retained] = ACTIONS(4252),
    [anon_sym___bridge_transfer] = ACTIONS(4252),
    [anon_sym___complex] = ACTIONS(4252),
    [anon_sym___const] = ACTIONS(4252),
    [anon_sym___imag] = ACTIONS(4252),
    [anon_sym___kindof] = ACTIONS(4252),
    [anon_sym___nonnull] = ACTIONS(4252),
    [anon_sym___nullable] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4252),
    [anon_sym___real] = ACTIONS(4252),
    [anon_sym___strong] = ACTIONS(4252),
    [anon_sym___unsafe_unretained] = ACTIONS(4252),
    [anon_sym___unused] = ACTIONS(4252),
    [anon_sym___weak] = ACTIONS(4252),
    [anon_sym_alignas] = ACTIONS(4252),
    [anon_sym__Alignas] = ACTIONS(4252),
    [sym_primitive_type] = ACTIONS(4252),
    [anon_sym_enum] = ACTIONS(4252),
    [anon_sym_struct] = ACTIONS(4252),
    [anon_sym_union] = ACTIONS(4252),
    [anon_sym_QMARK] = ACTIONS(4255),
    [anon_sym_DASH_DASH] = ACTIONS(4255),
    [anon_sym_PLUS_PLUS] = ACTIONS(4255),
    [anon_sym_DOT] = ACTIONS(4252),
    [anon_sym_DASH_GT] = ACTIONS(4255),
    [anon_sym_AT] = ACTIONS(4252),
    [anon_sym_DQUOTE] = ACTIONS(4255),
    [anon_sym_L_DQUOTE] = ACTIONS(4255),
    [anon_sym_u_DQUOTE] = ACTIONS(4255),
    [anon_sym_U_DQUOTE] = ACTIONS(4255),
    [anon_sym_u8_DQUOTE] = ACTIONS(4255),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4252),
    [anon_sym___typeof] = ACTIONS(4252),
    [anon_sym_typeof] = ACTIONS(4252),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4252),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4252),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4252),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4252),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4252),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4252),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4252),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_API_AVAILABLE] = ACTIONS(4252),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_API_DEPRECATED] = ACTIONS(4252),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4252),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym___deprecated_msg] = ACTIONS(4252),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4252),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4252),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4252),
    [anon_sym_BOOL] = ACTIONS(4252),
    [anon_sym_IMP] = ACTIONS(4252),
    [anon_sym_SEL] = ACTIONS(4252),
    [anon_sym_Class] = ACTIONS(4252),
    [anon_sym_id] = ACTIONS(4252),
  },
  [STATE(1665)] = {
    [sym_identifier] = ACTIONS(4106),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4108),
    [anon_sym_COMMA] = ACTIONS(4108),
    [anon_sym_RPAREN] = ACTIONS(4108),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4108),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4108),
    [anon_sym_LPAREN2] = ACTIONS(4108),
    [anon_sym_DASH] = ACTIONS(4106),
    [anon_sym_PLUS] = ACTIONS(4106),
    [anon_sym_STAR] = ACTIONS(4108),
    [anon_sym_SLASH] = ACTIONS(4106),
    [anon_sym_PERCENT] = ACTIONS(4108),
    [anon_sym_PIPE_PIPE] = ACTIONS(4108),
    [anon_sym_AMP_AMP] = ACTIONS(4108),
    [anon_sym_PIPE] = ACTIONS(4106),
    [anon_sym_CARET] = ACTIONS(4108),
    [anon_sym_AMP] = ACTIONS(4106),
    [anon_sym_EQ_EQ] = ACTIONS(4108),
    [anon_sym_BANG_EQ] = ACTIONS(4108),
    [anon_sym_GT] = ACTIONS(4106),
    [anon_sym_GT_EQ] = ACTIONS(4108),
    [anon_sym_LT_EQ] = ACTIONS(4108),
    [anon_sym_LT] = ACTIONS(4106),
    [anon_sym_LT_LT] = ACTIONS(4108),
    [anon_sym_GT_GT] = ACTIONS(4108),
    [anon_sym_SEMI] = ACTIONS(4108),
    [anon_sym___extension__] = ACTIONS(4106),
    [anon_sym_extern] = ACTIONS(4106),
    [anon_sym___attribute__] = ACTIONS(4106),
    [anon_sym___attribute] = ACTIONS(4106),
    [anon_sym_noreturn] = ACTIONS(4106),
    [anon_sym_LBRACK] = ACTIONS(4108),
    [anon_sym_RBRACK] = ACTIONS(4108),
    [anon_sym___declspec] = ACTIONS(4106),
    [anon_sym_LBRACE] = ACTIONS(4108),
    [anon_sym_RBRACE] = ACTIONS(4108),
    [anon_sym_signed] = ACTIONS(4106),
    [anon_sym_unsigned] = ACTIONS(4106),
    [anon_sym_long] = ACTIONS(4106),
    [anon_sym_short] = ACTIONS(4106),
    [anon_sym_static] = ACTIONS(4106),
    [anon_sym_ATautoreleasepool] = ACTIONS(4108),
    [anon_sym_auto] = ACTIONS(4106),
    [anon_sym_register] = ACTIONS(4106),
    [anon_sym_inline] = ACTIONS(4106),
    [anon_sym___inline] = ACTIONS(4106),
    [anon_sym___inline__] = ACTIONS(4106),
    [anon_sym___forceinline] = ACTIONS(4106),
    [anon_sym_thread_local] = ACTIONS(4106),
    [anon_sym___thread] = ACTIONS(4106),
    [anon_sym_CG_EXTERN] = ACTIONS(4106),
    [anon_sym_CG_INLINE] = ACTIONS(4106),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4106),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4106),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4106),
    [anon_sym_IBOutlet] = ACTIONS(4106),
    [anon_sym_IBInspectable] = ACTIONS(4106),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4106),
    [anon_sym_NS_INLINE] = ACTIONS(4106),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4106),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4106),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4106),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4106),
    [anon_sym_const] = ACTIONS(4106),
    [anon_sym_constexpr] = ACTIONS(4106),
    [anon_sym_volatile] = ACTIONS(4106),
    [anon_sym_restrict] = ACTIONS(4106),
    [anon_sym___restrict__] = ACTIONS(4106),
    [anon_sym__Atomic] = ACTIONS(4106),
    [anon_sym__Noreturn] = ACTIONS(4106),
    [anon_sym__Nonnull] = ACTIONS(4106),
    [anon_sym_nullable] = ACTIONS(4106),
    [anon_sym__Complex] = ACTIONS(4106),
    [anon_sym__Nullable] = ACTIONS(4106),
    [anon_sym__Nullable_result] = ACTIONS(4106),
    [anon_sym__Null_unspecified] = ACTIONS(4106),
    [anon_sym___autoreleasing] = ACTIONS(4106),
    [anon_sym___block] = ACTIONS(4106),
    [anon_sym___bridge] = ACTIONS(4106),
    [anon_sym___bridge_retained] = ACTIONS(4106),
    [anon_sym___bridge_transfer] = ACTIONS(4106),
    [anon_sym___complex] = ACTIONS(4106),
    [anon_sym___const] = ACTIONS(4106),
    [anon_sym___imag] = ACTIONS(4106),
    [anon_sym___kindof] = ACTIONS(4106),
    [anon_sym___nonnull] = ACTIONS(4106),
    [anon_sym___nullable] = ACTIONS(4106),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4106),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4106),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4106),
    [anon_sym___real] = ACTIONS(4106),
    [anon_sym___strong] = ACTIONS(4106),
    [anon_sym___unsafe_unretained] = ACTIONS(4106),
    [anon_sym___unused] = ACTIONS(4106),
    [anon_sym___weak] = ACTIONS(4106),
    [anon_sym_alignas] = ACTIONS(4106),
    [anon_sym__Alignas] = ACTIONS(4106),
    [sym_primitive_type] = ACTIONS(4106),
    [anon_sym_enum] = ACTIONS(4106),
    [anon_sym_COLON] = ACTIONS(4108),
    [anon_sym_struct] = ACTIONS(4106),
    [anon_sym_union] = ACTIONS(4106),
    [anon_sym_QMARK] = ACTIONS(4108),
    [anon_sym_DASH_DASH] = ACTIONS(4108),
    [anon_sym_PLUS_PLUS] = ACTIONS(4108),
    [anon_sym_DOT] = ACTIONS(4106),
    [anon_sym_DASH_GT] = ACTIONS(4108),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4106),
    [anon_sym___typeof] = ACTIONS(4106),
    [anon_sym_typeof] = ACTIONS(4106),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4106),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4106),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4106),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4106),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4106),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4106),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4106),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4106),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4106),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4106),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4106),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4106),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4106),
    [anon_sym_API_AVAILABLE] = ACTIONS(4106),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4106),
    [anon_sym_API_DEPRECATED] = ACTIONS(4106),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4106),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4106),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4106),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4106),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4106),
    [anon_sym___deprecated_msg] = ACTIONS(4106),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4106),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4106),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4106),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4106),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4106),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4106),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4106),
    [anon_sym_BOOL] = ACTIONS(4106),
    [anon_sym_IMP] = ACTIONS(4106),
    [anon_sym_SEL] = ACTIONS(4106),
    [anon_sym_Class] = ACTIONS(4106),
    [anon_sym_id] = ACTIONS(4106),
  },
  [STATE(1666)] = {
    [sym_identifier] = ACTIONS(4154),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4157),
    [anon_sym_COMMA] = ACTIONS(4157),
    [anon_sym_LPAREN2] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4154),
    [anon_sym_PLUS] = ACTIONS(4154),
    [anon_sym_STAR] = ACTIONS(4157),
    [anon_sym_SLASH] = ACTIONS(4154),
    [anon_sym_PERCENT] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4154),
    [anon_sym_CARET] = ACTIONS(4157),
    [anon_sym_AMP] = ACTIONS(4154),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4154),
    [anon_sym_GT_EQ] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_LT] = ACTIONS(4154),
    [anon_sym_LT_LT] = ACTIONS(4157),
    [anon_sym_GT_GT] = ACTIONS(4157),
    [anon_sym_SEMI] = ACTIONS(4157),
    [anon_sym___extension__] = ACTIONS(4154),
    [anon_sym_extern] = ACTIONS(4154),
    [anon_sym___attribute__] = ACTIONS(4154),
    [anon_sym___attribute] = ACTIONS(4154),
    [anon_sym_noreturn] = ACTIONS(4154),
    [anon_sym_LBRACK] = ACTIONS(4157),
    [anon_sym___declspec] = ACTIONS(4154),
    [anon_sym_LBRACE] = ACTIONS(4157),
    [anon_sym_signed] = ACTIONS(4154),
    [anon_sym_unsigned] = ACTIONS(4154),
    [anon_sym_long] = ACTIONS(4154),
    [anon_sym_short] = ACTIONS(4154),
    [anon_sym_static] = ACTIONS(4154),
    [anon_sym_ATautoreleasepool] = ACTIONS(4157),
    [anon_sym_auto] = ACTIONS(4154),
    [anon_sym_register] = ACTIONS(4154),
    [anon_sym_inline] = ACTIONS(4154),
    [anon_sym___inline] = ACTIONS(4154),
    [anon_sym___inline__] = ACTIONS(4154),
    [anon_sym___forceinline] = ACTIONS(4154),
    [anon_sym_thread_local] = ACTIONS(4154),
    [anon_sym___thread] = ACTIONS(4154),
    [anon_sym_CG_EXTERN] = ACTIONS(4154),
    [anon_sym_CG_INLINE] = ACTIONS(4154),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4154),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4154),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4154),
    [anon_sym_IBOutlet] = ACTIONS(4154),
    [anon_sym_IBInspectable] = ACTIONS(4154),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4154),
    [anon_sym_NS_INLINE] = ACTIONS(4154),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4154),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4154),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4154),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4154),
    [anon_sym_const] = ACTIONS(4154),
    [anon_sym_constexpr] = ACTIONS(4154),
    [anon_sym_volatile] = ACTIONS(4154),
    [anon_sym_restrict] = ACTIONS(4154),
    [anon_sym___restrict__] = ACTIONS(4154),
    [anon_sym__Atomic] = ACTIONS(4154),
    [anon_sym__Noreturn] = ACTIONS(4154),
    [anon_sym__Nonnull] = ACTIONS(4154),
    [anon_sym_nullable] = ACTIONS(4154),
    [anon_sym__Complex] = ACTIONS(4154),
    [anon_sym__Nullable] = ACTIONS(4154),
    [anon_sym__Nullable_result] = ACTIONS(4154),
    [anon_sym__Null_unspecified] = ACTIONS(4154),
    [anon_sym___autoreleasing] = ACTIONS(4154),
    [anon_sym___block] = ACTIONS(4154),
    [anon_sym___bridge] = ACTIONS(4154),
    [anon_sym___bridge_retained] = ACTIONS(4154),
    [anon_sym___bridge_transfer] = ACTIONS(4154),
    [anon_sym___complex] = ACTIONS(4154),
    [anon_sym___const] = ACTIONS(4154),
    [anon_sym___imag] = ACTIONS(4154),
    [anon_sym___kindof] = ACTIONS(4154),
    [anon_sym___nonnull] = ACTIONS(4154),
    [anon_sym___nullable] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4154),
    [anon_sym___real] = ACTIONS(4154),
    [anon_sym___strong] = ACTIONS(4154),
    [anon_sym___unsafe_unretained] = ACTIONS(4154),
    [anon_sym___unused] = ACTIONS(4154),
    [anon_sym___weak] = ACTIONS(4154),
    [anon_sym_alignas] = ACTIONS(4154),
    [anon_sym__Alignas] = ACTIONS(4154),
    [sym_primitive_type] = ACTIONS(4154),
    [anon_sym_enum] = ACTIONS(4154),
    [anon_sym_struct] = ACTIONS(4154),
    [anon_sym_union] = ACTIONS(4154),
    [anon_sym_QMARK] = ACTIONS(4157),
    [anon_sym_DASH_DASH] = ACTIONS(4157),
    [anon_sym_PLUS_PLUS] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4154),
    [anon_sym_DASH_GT] = ACTIONS(4157),
    [anon_sym_AT] = ACTIONS(4154),
    [anon_sym_DQUOTE] = ACTIONS(4157),
    [anon_sym_L_DQUOTE] = ACTIONS(4157),
    [anon_sym_u_DQUOTE] = ACTIONS(4157),
    [anon_sym_U_DQUOTE] = ACTIONS(4157),
    [anon_sym_u8_DQUOTE] = ACTIONS(4157),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4154),
    [anon_sym___typeof] = ACTIONS(4154),
    [anon_sym_typeof] = ACTIONS(4154),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4154),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4154),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4154),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4154),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4154),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4154),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4154),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_API_AVAILABLE] = ACTIONS(4154),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_API_DEPRECATED] = ACTIONS(4154),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4154),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym___deprecated_msg] = ACTIONS(4154),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4154),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4154),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4154),
    [anon_sym_BOOL] = ACTIONS(4154),
    [anon_sym_IMP] = ACTIONS(4154),
    [anon_sym_SEL] = ACTIONS(4154),
    [anon_sym_Class] = ACTIONS(4154),
    [anon_sym_id] = ACTIONS(4154),
  },
  [STATE(1667)] = {
    [sym_string_literal] = STATE(1627),
    [sym_identifier] = ACTIONS(4024),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_RPAREN] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4024),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4024),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4024),
    [anon_sym_extern] = ACTIONS(4024),
    [anon_sym___attribute__] = ACTIONS(4024),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4024),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym_RBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4024),
    [anon_sym_RBRACE] = ACTIONS(4026),
    [anon_sym_static] = ACTIONS(4024),
    [anon_sym_EQ] = ACTIONS(4024),
    [anon_sym_auto] = ACTIONS(4024),
    [anon_sym_register] = ACTIONS(4024),
    [anon_sym_inline] = ACTIONS(4024),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4024),
    [anon_sym___forceinline] = ACTIONS(4024),
    [anon_sym_thread_local] = ACTIONS(4024),
    [anon_sym___thread] = ACTIONS(4024),
    [anon_sym_CG_EXTERN] = ACTIONS(4024),
    [anon_sym_CG_INLINE] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4024),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4024),
    [anon_sym_IBOutlet] = ACTIONS(4024),
    [anon_sym_IBInspectable] = ACTIONS(4024),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4024),
    [anon_sym_NS_INLINE] = ACTIONS(4024),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4024),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4024),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4024),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4024),
    [anon_sym_volatile] = ACTIONS(4024),
    [anon_sym_restrict] = ACTIONS(4024),
    [anon_sym___restrict__] = ACTIONS(4024),
    [anon_sym__Atomic] = ACTIONS(4024),
    [anon_sym__Noreturn] = ACTIONS(4024),
    [anon_sym__Nonnull] = ACTIONS(4024),
    [anon_sym_nullable] = ACTIONS(4024),
    [anon_sym__Complex] = ACTIONS(4024),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4024),
    [anon_sym__Null_unspecified] = ACTIONS(4024),
    [anon_sym___autoreleasing] = ACTIONS(4024),
    [anon_sym___block] = ACTIONS(4024),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4024),
    [anon_sym___bridge_transfer] = ACTIONS(4024),
    [anon_sym___complex] = ACTIONS(4024),
    [anon_sym___const] = ACTIONS(4024),
    [anon_sym___imag] = ACTIONS(4024),
    [anon_sym___kindof] = ACTIONS(4024),
    [anon_sym___nonnull] = ACTIONS(4024),
    [anon_sym___nullable] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4024),
    [anon_sym___real] = ACTIONS(4024),
    [anon_sym___strong] = ACTIONS(4024),
    [anon_sym___unsafe_unretained] = ACTIONS(4024),
    [anon_sym___unused] = ACTIONS(4024),
    [anon_sym___weak] = ACTIONS(4024),
    [anon_sym_alignas] = ACTIONS(4024),
    [anon_sym__Alignas] = ACTIONS(4024),
    [anon_sym_COLON] = ACTIONS(4026),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4026),
    [anon_sym_SLASH_EQ] = ACTIONS(4026),
    [anon_sym_PERCENT_EQ] = ACTIONS(4026),
    [anon_sym_PLUS_EQ] = ACTIONS(4026),
    [anon_sym_DASH_EQ] = ACTIONS(4026),
    [anon_sym_LT_LT_EQ] = ACTIONS(4026),
    [anon_sym_GT_GT_EQ] = ACTIONS(4026),
    [anon_sym_AMP_EQ] = ACTIONS(4026),
    [anon_sym_CARET_EQ] = ACTIONS(4026),
    [anon_sym_PIPE_EQ] = ACTIONS(4026),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4625),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4024),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4024),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_API_AVAILABLE] = ACTIONS(4024),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_API_DEPRECATED] = ACTIONS(4024),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym___deprecated_msg] = ACTIONS(4024),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4024),
  },
  [STATE(1668)] = {
    [sym_identifier] = ACTIONS(4240),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4242),
    [anon_sym_COMMA] = ACTIONS(4242),
    [anon_sym_RPAREN] = ACTIONS(4242),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4242),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4242),
    [anon_sym_LPAREN2] = ACTIONS(4242),
    [anon_sym_DASH] = ACTIONS(4240),
    [anon_sym_PLUS] = ACTIONS(4240),
    [anon_sym_STAR] = ACTIONS(4242),
    [anon_sym_SLASH] = ACTIONS(4240),
    [anon_sym_PERCENT] = ACTIONS(4242),
    [anon_sym_PIPE_PIPE] = ACTIONS(4242),
    [anon_sym_AMP_AMP] = ACTIONS(4242),
    [anon_sym_PIPE] = ACTIONS(4240),
    [anon_sym_CARET] = ACTIONS(4242),
    [anon_sym_AMP] = ACTIONS(4240),
    [anon_sym_EQ_EQ] = ACTIONS(4242),
    [anon_sym_BANG_EQ] = ACTIONS(4242),
    [anon_sym_GT] = ACTIONS(4240),
    [anon_sym_GT_EQ] = ACTIONS(4242),
    [anon_sym_LT_EQ] = ACTIONS(4242),
    [anon_sym_LT] = ACTIONS(4240),
    [anon_sym_LT_LT] = ACTIONS(4242),
    [anon_sym_GT_GT] = ACTIONS(4242),
    [anon_sym_SEMI] = ACTIONS(4242),
    [anon_sym___extension__] = ACTIONS(4240),
    [anon_sym_extern] = ACTIONS(4240),
    [anon_sym___attribute__] = ACTIONS(4240),
    [anon_sym___attribute] = ACTIONS(4240),
    [anon_sym_noreturn] = ACTIONS(4240),
    [anon_sym_LBRACK] = ACTIONS(4242),
    [anon_sym_RBRACK] = ACTIONS(4242),
    [anon_sym___declspec] = ACTIONS(4240),
    [anon_sym_LBRACE] = ACTIONS(4242),
    [anon_sym_RBRACE] = ACTIONS(4242),
    [anon_sym_signed] = ACTIONS(4240),
    [anon_sym_unsigned] = ACTIONS(4240),
    [anon_sym_long] = ACTIONS(4240),
    [anon_sym_short] = ACTIONS(4240),
    [anon_sym_static] = ACTIONS(4240),
    [anon_sym_ATautoreleasepool] = ACTIONS(4242),
    [anon_sym_auto] = ACTIONS(4240),
    [anon_sym_register] = ACTIONS(4240),
    [anon_sym_inline] = ACTIONS(4240),
    [anon_sym___inline] = ACTIONS(4240),
    [anon_sym___inline__] = ACTIONS(4240),
    [anon_sym___forceinline] = ACTIONS(4240),
    [anon_sym_thread_local] = ACTIONS(4240),
    [anon_sym___thread] = ACTIONS(4240),
    [anon_sym_CG_EXTERN] = ACTIONS(4240),
    [anon_sym_CG_INLINE] = ACTIONS(4240),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4240),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4240),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4240),
    [anon_sym_IBOutlet] = ACTIONS(4240),
    [anon_sym_IBInspectable] = ACTIONS(4240),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4240),
    [anon_sym_NS_INLINE] = ACTIONS(4240),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4240),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4240),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4240),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4240),
    [anon_sym_const] = ACTIONS(4240),
    [anon_sym_constexpr] = ACTIONS(4240),
    [anon_sym_volatile] = ACTIONS(4240),
    [anon_sym_restrict] = ACTIONS(4240),
    [anon_sym___restrict__] = ACTIONS(4240),
    [anon_sym__Atomic] = ACTIONS(4240),
    [anon_sym__Noreturn] = ACTIONS(4240),
    [anon_sym__Nonnull] = ACTIONS(4240),
    [anon_sym_nullable] = ACTIONS(4240),
    [anon_sym__Complex] = ACTIONS(4240),
    [anon_sym__Nullable] = ACTIONS(4240),
    [anon_sym__Nullable_result] = ACTIONS(4240),
    [anon_sym__Null_unspecified] = ACTIONS(4240),
    [anon_sym___autoreleasing] = ACTIONS(4240),
    [anon_sym___block] = ACTIONS(4240),
    [anon_sym___bridge] = ACTIONS(4240),
    [anon_sym___bridge_retained] = ACTIONS(4240),
    [anon_sym___bridge_transfer] = ACTIONS(4240),
    [anon_sym___complex] = ACTIONS(4240),
    [anon_sym___const] = ACTIONS(4240),
    [anon_sym___imag] = ACTIONS(4240),
    [anon_sym___kindof] = ACTIONS(4240),
    [anon_sym___nonnull] = ACTIONS(4240),
    [anon_sym___nullable] = ACTIONS(4240),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4240),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4240),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4240),
    [anon_sym___real] = ACTIONS(4240),
    [anon_sym___strong] = ACTIONS(4240),
    [anon_sym___unsafe_unretained] = ACTIONS(4240),
    [anon_sym___unused] = ACTIONS(4240),
    [anon_sym___weak] = ACTIONS(4240),
    [anon_sym_alignas] = ACTIONS(4240),
    [anon_sym__Alignas] = ACTIONS(4240),
    [sym_primitive_type] = ACTIONS(4240),
    [anon_sym_enum] = ACTIONS(4240),
    [anon_sym_COLON] = ACTIONS(4242),
    [anon_sym_struct] = ACTIONS(4240),
    [anon_sym_union] = ACTIONS(4240),
    [anon_sym_QMARK] = ACTIONS(4242),
    [anon_sym_DASH_DASH] = ACTIONS(4242),
    [anon_sym_PLUS_PLUS] = ACTIONS(4242),
    [anon_sym_DOT] = ACTIONS(4240),
    [anon_sym_DASH_GT] = ACTIONS(4242),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4240),
    [anon_sym___typeof] = ACTIONS(4240),
    [anon_sym_typeof] = ACTIONS(4240),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4240),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4240),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4240),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4240),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4240),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4240),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4240),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4240),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4240),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4240),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4240),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4240),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4240),
    [anon_sym_API_AVAILABLE] = ACTIONS(4240),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4240),
    [anon_sym_API_DEPRECATED] = ACTIONS(4240),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4240),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4240),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4240),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4240),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4240),
    [anon_sym___deprecated_msg] = ACTIONS(4240),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4240),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4240),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4240),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4240),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4240),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4240),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4240),
    [anon_sym_BOOL] = ACTIONS(4240),
    [anon_sym_IMP] = ACTIONS(4240),
    [anon_sym_SEL] = ACTIONS(4240),
    [anon_sym_Class] = ACTIONS(4240),
    [anon_sym_id] = ACTIONS(4240),
  },
  [STATE(1669)] = {
    [sym_identifier] = ACTIONS(4258),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4260),
    [anon_sym_COMMA] = ACTIONS(4260),
    [anon_sym_RPAREN] = ACTIONS(4260),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4260),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4260),
    [anon_sym_LPAREN2] = ACTIONS(4260),
    [anon_sym_DASH] = ACTIONS(4258),
    [anon_sym_PLUS] = ACTIONS(4258),
    [anon_sym_STAR] = ACTIONS(4258),
    [anon_sym_SLASH] = ACTIONS(4258),
    [anon_sym_PERCENT] = ACTIONS(4258),
    [anon_sym_PIPE_PIPE] = ACTIONS(4260),
    [anon_sym_AMP_AMP] = ACTIONS(4260),
    [anon_sym_PIPE] = ACTIONS(4258),
    [anon_sym_CARET] = ACTIONS(4258),
    [anon_sym_AMP] = ACTIONS(4258),
    [anon_sym_EQ_EQ] = ACTIONS(4260),
    [anon_sym_BANG_EQ] = ACTIONS(4260),
    [anon_sym_GT] = ACTIONS(4258),
    [anon_sym_GT_EQ] = ACTIONS(4260),
    [anon_sym_LT_EQ] = ACTIONS(4260),
    [anon_sym_LT] = ACTIONS(4258),
    [anon_sym_LT_LT] = ACTIONS(4258),
    [anon_sym_GT_GT] = ACTIONS(4258),
    [anon_sym_SEMI] = ACTIONS(4260),
    [anon_sym___extension__] = ACTIONS(4258),
    [anon_sym_extern] = ACTIONS(4258),
    [anon_sym___attribute__] = ACTIONS(4258),
    [anon_sym___attribute] = ACTIONS(4258),
    [anon_sym_noreturn] = ACTIONS(4258),
    [anon_sym_LBRACK] = ACTIONS(4260),
    [anon_sym_RBRACK] = ACTIONS(4260),
    [anon_sym___declspec] = ACTIONS(4258),
    [anon_sym_RBRACE] = ACTIONS(4260),
    [anon_sym_static] = ACTIONS(4258),
    [anon_sym_EQ] = ACTIONS(4258),
    [anon_sym_auto] = ACTIONS(4258),
    [anon_sym_register] = ACTIONS(4258),
    [anon_sym_inline] = ACTIONS(4258),
    [anon_sym___inline] = ACTIONS(4258),
    [anon_sym___inline__] = ACTIONS(4258),
    [anon_sym___forceinline] = ACTIONS(4258),
    [anon_sym_thread_local] = ACTIONS(4258),
    [anon_sym___thread] = ACTIONS(4258),
    [anon_sym_CG_EXTERN] = ACTIONS(4258),
    [anon_sym_CG_INLINE] = ACTIONS(4258),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4258),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4258),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4258),
    [anon_sym_IBOutlet] = ACTIONS(4258),
    [anon_sym_IBInspectable] = ACTIONS(4258),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4258),
    [anon_sym_NS_INLINE] = ACTIONS(4258),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4258),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4258),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4258),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4258),
    [anon_sym_const] = ACTIONS(4258),
    [anon_sym_constexpr] = ACTIONS(4258),
    [anon_sym_volatile] = ACTIONS(4258),
    [anon_sym_restrict] = ACTIONS(4258),
    [anon_sym___restrict__] = ACTIONS(4258),
    [anon_sym__Atomic] = ACTIONS(4258),
    [anon_sym__Noreturn] = ACTIONS(4258),
    [anon_sym__Nonnull] = ACTIONS(4258),
    [anon_sym_nullable] = ACTIONS(4258),
    [anon_sym__Complex] = ACTIONS(4258),
    [anon_sym__Nullable] = ACTIONS(4258),
    [anon_sym__Nullable_result] = ACTIONS(4258),
    [anon_sym__Null_unspecified] = ACTIONS(4258),
    [anon_sym___autoreleasing] = ACTIONS(4258),
    [anon_sym___block] = ACTIONS(4258),
    [anon_sym___bridge] = ACTIONS(4258),
    [anon_sym___bridge_retained] = ACTIONS(4258),
    [anon_sym___bridge_transfer] = ACTIONS(4258),
    [anon_sym___complex] = ACTIONS(4258),
    [anon_sym___const] = ACTIONS(4258),
    [anon_sym___imag] = ACTIONS(4258),
    [anon_sym___kindof] = ACTIONS(4258),
    [anon_sym___nonnull] = ACTIONS(4258),
    [anon_sym___nullable] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4258),
    [anon_sym___real] = ACTIONS(4258),
    [anon_sym___strong] = ACTIONS(4258),
    [anon_sym___unsafe_unretained] = ACTIONS(4258),
    [anon_sym___unused] = ACTIONS(4258),
    [anon_sym___weak] = ACTIONS(4258),
    [anon_sym_alignas] = ACTIONS(4258),
    [anon_sym__Alignas] = ACTIONS(4258),
    [anon_sym_COLON] = ACTIONS(4260),
    [anon_sym_QMARK] = ACTIONS(4260),
    [anon_sym_STAR_EQ] = ACTIONS(4260),
    [anon_sym_SLASH_EQ] = ACTIONS(4260),
    [anon_sym_PERCENT_EQ] = ACTIONS(4260),
    [anon_sym_PLUS_EQ] = ACTIONS(4260),
    [anon_sym_DASH_EQ] = ACTIONS(4260),
    [anon_sym_LT_LT_EQ] = ACTIONS(4260),
    [anon_sym_GT_GT_EQ] = ACTIONS(4260),
    [anon_sym_AMP_EQ] = ACTIONS(4260),
    [anon_sym_CARET_EQ] = ACTIONS(4260),
    [anon_sym_PIPE_EQ] = ACTIONS(4260),
    [anon_sym_DASH_DASH] = ACTIONS(4260),
    [anon_sym_PLUS_PLUS] = ACTIONS(4260),
    [anon_sym_DOT] = ACTIONS(4258),
    [anon_sym_DASH_GT] = ACTIONS(4260),
    [anon_sym_AT] = ACTIONS(4260),
    [anon_sym_DQUOTE] = ACTIONS(4260),
    [anon_sym_L_DQUOTE] = ACTIONS(4260),
    [anon_sym_u_DQUOTE] = ACTIONS(4260),
    [anon_sym_U_DQUOTE] = ACTIONS(4260),
    [anon_sym_u8_DQUOTE] = ACTIONS(4260),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4258),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4258),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4258),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4258),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4258),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4258),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4258),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_API_AVAILABLE] = ACTIONS(4258),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_API_DEPRECATED] = ACTIONS(4258),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4258),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym___deprecated_msg] = ACTIONS(4258),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4258),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4258),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4258),
  },
  [STATE(1670)] = {
    [sym_identifier] = ACTIONS(4118),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4120),
    [anon_sym_COMMA] = ACTIONS(4120),
    [anon_sym_RPAREN] = ACTIONS(4120),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4120),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4120),
    [anon_sym_LPAREN2] = ACTIONS(4120),
    [anon_sym_DASH] = ACTIONS(4118),
    [anon_sym_PLUS] = ACTIONS(4118),
    [anon_sym_STAR] = ACTIONS(4118),
    [anon_sym_SLASH] = ACTIONS(4118),
    [anon_sym_PERCENT] = ACTIONS(4118),
    [anon_sym_PIPE_PIPE] = ACTIONS(4120),
    [anon_sym_AMP_AMP] = ACTIONS(4120),
    [anon_sym_PIPE] = ACTIONS(4118),
    [anon_sym_CARET] = ACTIONS(4118),
    [anon_sym_AMP] = ACTIONS(4118),
    [anon_sym_EQ_EQ] = ACTIONS(4120),
    [anon_sym_BANG_EQ] = ACTIONS(4120),
    [anon_sym_GT] = ACTIONS(4118),
    [anon_sym_GT_EQ] = ACTIONS(4120),
    [anon_sym_LT_EQ] = ACTIONS(4120),
    [anon_sym_LT] = ACTIONS(4118),
    [anon_sym_LT_LT] = ACTIONS(4118),
    [anon_sym_GT_GT] = ACTIONS(4118),
    [anon_sym_SEMI] = ACTIONS(4120),
    [anon_sym___extension__] = ACTIONS(4118),
    [anon_sym_extern] = ACTIONS(4118),
    [anon_sym___attribute__] = ACTIONS(4118),
    [anon_sym___attribute] = ACTIONS(4118),
    [anon_sym_noreturn] = ACTIONS(4118),
    [anon_sym_LBRACK] = ACTIONS(4120),
    [anon_sym_RBRACK] = ACTIONS(4120),
    [anon_sym___declspec] = ACTIONS(4118),
    [anon_sym_RBRACE] = ACTIONS(4120),
    [anon_sym_static] = ACTIONS(4118),
    [anon_sym_EQ] = ACTIONS(4118),
    [anon_sym_auto] = ACTIONS(4118),
    [anon_sym_register] = ACTIONS(4118),
    [anon_sym_inline] = ACTIONS(4118),
    [anon_sym___inline] = ACTIONS(4118),
    [anon_sym___inline__] = ACTIONS(4118),
    [anon_sym___forceinline] = ACTIONS(4118),
    [anon_sym_thread_local] = ACTIONS(4118),
    [anon_sym___thread] = ACTIONS(4118),
    [anon_sym_CG_EXTERN] = ACTIONS(4118),
    [anon_sym_CG_INLINE] = ACTIONS(4118),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4118),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4118),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4118),
    [anon_sym_IBOutlet] = ACTIONS(4118),
    [anon_sym_IBInspectable] = ACTIONS(4118),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4118),
    [anon_sym_NS_INLINE] = ACTIONS(4118),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4118),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4118),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4118),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4118),
    [anon_sym_const] = ACTIONS(4118),
    [anon_sym_constexpr] = ACTIONS(4118),
    [anon_sym_volatile] = ACTIONS(4118),
    [anon_sym_restrict] = ACTIONS(4118),
    [anon_sym___restrict__] = ACTIONS(4118),
    [anon_sym__Atomic] = ACTIONS(4118),
    [anon_sym__Noreturn] = ACTIONS(4118),
    [anon_sym__Nonnull] = ACTIONS(4118),
    [anon_sym_nullable] = ACTIONS(4118),
    [anon_sym__Complex] = ACTIONS(4118),
    [anon_sym__Nullable] = ACTIONS(4118),
    [anon_sym__Nullable_result] = ACTIONS(4118),
    [anon_sym__Null_unspecified] = ACTIONS(4118),
    [anon_sym___autoreleasing] = ACTIONS(4118),
    [anon_sym___block] = ACTIONS(4118),
    [anon_sym___bridge] = ACTIONS(4118),
    [anon_sym___bridge_retained] = ACTIONS(4118),
    [anon_sym___bridge_transfer] = ACTIONS(4118),
    [anon_sym___complex] = ACTIONS(4118),
    [anon_sym___const] = ACTIONS(4118),
    [anon_sym___imag] = ACTIONS(4118),
    [anon_sym___kindof] = ACTIONS(4118),
    [anon_sym___nonnull] = ACTIONS(4118),
    [anon_sym___nullable] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4118),
    [anon_sym___real] = ACTIONS(4118),
    [anon_sym___strong] = ACTIONS(4118),
    [anon_sym___unsafe_unretained] = ACTIONS(4118),
    [anon_sym___unused] = ACTIONS(4118),
    [anon_sym___weak] = ACTIONS(4118),
    [anon_sym_alignas] = ACTIONS(4118),
    [anon_sym__Alignas] = ACTIONS(4118),
    [anon_sym_COLON] = ACTIONS(4120),
    [anon_sym_QMARK] = ACTIONS(4120),
    [anon_sym_STAR_EQ] = ACTIONS(4120),
    [anon_sym_SLASH_EQ] = ACTIONS(4120),
    [anon_sym_PERCENT_EQ] = ACTIONS(4120),
    [anon_sym_PLUS_EQ] = ACTIONS(4120),
    [anon_sym_DASH_EQ] = ACTIONS(4120),
    [anon_sym_LT_LT_EQ] = ACTIONS(4120),
    [anon_sym_GT_GT_EQ] = ACTIONS(4120),
    [anon_sym_AMP_EQ] = ACTIONS(4120),
    [anon_sym_CARET_EQ] = ACTIONS(4120),
    [anon_sym_PIPE_EQ] = ACTIONS(4120),
    [anon_sym_DASH_DASH] = ACTIONS(4120),
    [anon_sym_PLUS_PLUS] = ACTIONS(4120),
    [anon_sym_DOT] = ACTIONS(4118),
    [anon_sym_DASH_GT] = ACTIONS(4120),
    [anon_sym_AT] = ACTIONS(4120),
    [anon_sym_DQUOTE] = ACTIONS(4120),
    [anon_sym_L_DQUOTE] = ACTIONS(4120),
    [anon_sym_u_DQUOTE] = ACTIONS(4120),
    [anon_sym_U_DQUOTE] = ACTIONS(4120),
    [anon_sym_u8_DQUOTE] = ACTIONS(4120),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4118),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4118),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4118),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4118),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4118),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4118),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4118),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_API_AVAILABLE] = ACTIONS(4118),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_API_DEPRECATED] = ACTIONS(4118),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4118),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym___deprecated_msg] = ACTIONS(4118),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4118),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4118),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4118),
  },
  [STATE(1671)] = {
    [sym_identifier] = ACTIONS(4154),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4157),
    [anon_sym_COMMA] = ACTIONS(4157),
    [anon_sym_RPAREN] = ACTIONS(4157),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4157),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4157),
    [anon_sym_LPAREN2] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4154),
    [anon_sym_PLUS] = ACTIONS(4154),
    [anon_sym_STAR] = ACTIONS(4154),
    [anon_sym_SLASH] = ACTIONS(4154),
    [anon_sym_PERCENT] = ACTIONS(4154),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4154),
    [anon_sym_CARET] = ACTIONS(4154),
    [anon_sym_AMP] = ACTIONS(4154),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4154),
    [anon_sym_GT_EQ] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_LT] = ACTIONS(4154),
    [anon_sym_LT_LT] = ACTIONS(4154),
    [anon_sym_GT_GT] = ACTIONS(4154),
    [anon_sym_SEMI] = ACTIONS(4157),
    [anon_sym___extension__] = ACTIONS(4154),
    [anon_sym_extern] = ACTIONS(4154),
    [anon_sym___attribute__] = ACTIONS(4154),
    [anon_sym___attribute] = ACTIONS(4154),
    [anon_sym_noreturn] = ACTIONS(4154),
    [anon_sym_LBRACK] = ACTIONS(4157),
    [anon_sym_RBRACK] = ACTIONS(4157),
    [anon_sym___declspec] = ACTIONS(4154),
    [anon_sym_RBRACE] = ACTIONS(4157),
    [anon_sym_static] = ACTIONS(4154),
    [anon_sym_EQ] = ACTIONS(4154),
    [anon_sym_auto] = ACTIONS(4154),
    [anon_sym_register] = ACTIONS(4154),
    [anon_sym_inline] = ACTIONS(4154),
    [anon_sym___inline] = ACTIONS(4154),
    [anon_sym___inline__] = ACTIONS(4154),
    [anon_sym___forceinline] = ACTIONS(4154),
    [anon_sym_thread_local] = ACTIONS(4154),
    [anon_sym___thread] = ACTIONS(4154),
    [anon_sym_CG_EXTERN] = ACTIONS(4154),
    [anon_sym_CG_INLINE] = ACTIONS(4154),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4154),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4154),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4154),
    [anon_sym_IBOutlet] = ACTIONS(4154),
    [anon_sym_IBInspectable] = ACTIONS(4154),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4154),
    [anon_sym_NS_INLINE] = ACTIONS(4154),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4154),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4154),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4154),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4154),
    [anon_sym_const] = ACTIONS(4154),
    [anon_sym_constexpr] = ACTIONS(4154),
    [anon_sym_volatile] = ACTIONS(4154),
    [anon_sym_restrict] = ACTIONS(4154),
    [anon_sym___restrict__] = ACTIONS(4154),
    [anon_sym__Atomic] = ACTIONS(4154),
    [anon_sym__Noreturn] = ACTIONS(4154),
    [anon_sym__Nonnull] = ACTIONS(4154),
    [anon_sym_nullable] = ACTIONS(4154),
    [anon_sym__Complex] = ACTIONS(4154),
    [anon_sym__Nullable] = ACTIONS(4154),
    [anon_sym__Nullable_result] = ACTIONS(4154),
    [anon_sym__Null_unspecified] = ACTIONS(4154),
    [anon_sym___autoreleasing] = ACTIONS(4154),
    [anon_sym___block] = ACTIONS(4154),
    [anon_sym___bridge] = ACTIONS(4154),
    [anon_sym___bridge_retained] = ACTIONS(4154),
    [anon_sym___bridge_transfer] = ACTIONS(4154),
    [anon_sym___complex] = ACTIONS(4154),
    [anon_sym___const] = ACTIONS(4154),
    [anon_sym___imag] = ACTIONS(4154),
    [anon_sym___kindof] = ACTIONS(4154),
    [anon_sym___nonnull] = ACTIONS(4154),
    [anon_sym___nullable] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4154),
    [anon_sym___real] = ACTIONS(4154),
    [anon_sym___strong] = ACTIONS(4154),
    [anon_sym___unsafe_unretained] = ACTIONS(4154),
    [anon_sym___unused] = ACTIONS(4154),
    [anon_sym___weak] = ACTIONS(4154),
    [anon_sym_alignas] = ACTIONS(4154),
    [anon_sym__Alignas] = ACTIONS(4154),
    [anon_sym_COLON] = ACTIONS(4157),
    [anon_sym_QMARK] = ACTIONS(4157),
    [anon_sym_STAR_EQ] = ACTIONS(4157),
    [anon_sym_SLASH_EQ] = ACTIONS(4157),
    [anon_sym_PERCENT_EQ] = ACTIONS(4157),
    [anon_sym_PLUS_EQ] = ACTIONS(4157),
    [anon_sym_DASH_EQ] = ACTIONS(4157),
    [anon_sym_LT_LT_EQ] = ACTIONS(4157),
    [anon_sym_GT_GT_EQ] = ACTIONS(4157),
    [anon_sym_AMP_EQ] = ACTIONS(4157),
    [anon_sym_CARET_EQ] = ACTIONS(4157),
    [anon_sym_PIPE_EQ] = ACTIONS(4157),
    [anon_sym_DASH_DASH] = ACTIONS(4157),
    [anon_sym_PLUS_PLUS] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4154),
    [anon_sym_DASH_GT] = ACTIONS(4157),
    [anon_sym_AT] = ACTIONS(4157),
    [anon_sym_DQUOTE] = ACTIONS(4157),
    [anon_sym_L_DQUOTE] = ACTIONS(4157),
    [anon_sym_u_DQUOTE] = ACTIONS(4157),
    [anon_sym_U_DQUOTE] = ACTIONS(4157),
    [anon_sym_u8_DQUOTE] = ACTIONS(4157),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4154),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4154),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4154),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4154),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4154),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4154),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4154),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_API_AVAILABLE] = ACTIONS(4154),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_API_DEPRECATED] = ACTIONS(4154),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4154),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym___deprecated_msg] = ACTIONS(4154),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4154),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4154),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4154),
  },
  [STATE(1672)] = {
    [sym_identifier] = ACTIONS(4252),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4255),
    [anon_sym_COMMA] = ACTIONS(4255),
    [anon_sym_RPAREN] = ACTIONS(4255),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4255),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4255),
    [anon_sym_LPAREN2] = ACTIONS(4255),
    [anon_sym_DASH] = ACTIONS(4252),
    [anon_sym_PLUS] = ACTIONS(4252),
    [anon_sym_STAR] = ACTIONS(4252),
    [anon_sym_SLASH] = ACTIONS(4252),
    [anon_sym_PERCENT] = ACTIONS(4252),
    [anon_sym_PIPE_PIPE] = ACTIONS(4255),
    [anon_sym_AMP_AMP] = ACTIONS(4255),
    [anon_sym_PIPE] = ACTIONS(4252),
    [anon_sym_CARET] = ACTIONS(4252),
    [anon_sym_AMP] = ACTIONS(4252),
    [anon_sym_EQ_EQ] = ACTIONS(4255),
    [anon_sym_BANG_EQ] = ACTIONS(4255),
    [anon_sym_GT] = ACTIONS(4252),
    [anon_sym_GT_EQ] = ACTIONS(4255),
    [anon_sym_LT_EQ] = ACTIONS(4255),
    [anon_sym_LT] = ACTIONS(4252),
    [anon_sym_LT_LT] = ACTIONS(4252),
    [anon_sym_GT_GT] = ACTIONS(4252),
    [anon_sym_SEMI] = ACTIONS(4255),
    [anon_sym___extension__] = ACTIONS(4252),
    [anon_sym_extern] = ACTIONS(4252),
    [anon_sym___attribute__] = ACTIONS(4252),
    [anon_sym___attribute] = ACTIONS(4252),
    [anon_sym_noreturn] = ACTIONS(4252),
    [anon_sym_LBRACK] = ACTIONS(4255),
    [anon_sym_RBRACK] = ACTIONS(4255),
    [anon_sym___declspec] = ACTIONS(4252),
    [anon_sym_RBRACE] = ACTIONS(4255),
    [anon_sym_static] = ACTIONS(4252),
    [anon_sym_EQ] = ACTIONS(4252),
    [anon_sym_auto] = ACTIONS(4252),
    [anon_sym_register] = ACTIONS(4252),
    [anon_sym_inline] = ACTIONS(4252),
    [anon_sym___inline] = ACTIONS(4252),
    [anon_sym___inline__] = ACTIONS(4252),
    [anon_sym___forceinline] = ACTIONS(4252),
    [anon_sym_thread_local] = ACTIONS(4252),
    [anon_sym___thread] = ACTIONS(4252),
    [anon_sym_CG_EXTERN] = ACTIONS(4252),
    [anon_sym_CG_INLINE] = ACTIONS(4252),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4252),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4252),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4252),
    [anon_sym_IBOutlet] = ACTIONS(4252),
    [anon_sym_IBInspectable] = ACTIONS(4252),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4252),
    [anon_sym_NS_INLINE] = ACTIONS(4252),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4252),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4252),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4252),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4252),
    [anon_sym_const] = ACTIONS(4252),
    [anon_sym_constexpr] = ACTIONS(4252),
    [anon_sym_volatile] = ACTIONS(4252),
    [anon_sym_restrict] = ACTIONS(4252),
    [anon_sym___restrict__] = ACTIONS(4252),
    [anon_sym__Atomic] = ACTIONS(4252),
    [anon_sym__Noreturn] = ACTIONS(4252),
    [anon_sym__Nonnull] = ACTIONS(4252),
    [anon_sym_nullable] = ACTIONS(4252),
    [anon_sym__Complex] = ACTIONS(4252),
    [anon_sym__Nullable] = ACTIONS(4252),
    [anon_sym__Nullable_result] = ACTIONS(4252),
    [anon_sym__Null_unspecified] = ACTIONS(4252),
    [anon_sym___autoreleasing] = ACTIONS(4252),
    [anon_sym___block] = ACTIONS(4252),
    [anon_sym___bridge] = ACTIONS(4252),
    [anon_sym___bridge_retained] = ACTIONS(4252),
    [anon_sym___bridge_transfer] = ACTIONS(4252),
    [anon_sym___complex] = ACTIONS(4252),
    [anon_sym___const] = ACTIONS(4252),
    [anon_sym___imag] = ACTIONS(4252),
    [anon_sym___kindof] = ACTIONS(4252),
    [anon_sym___nonnull] = ACTIONS(4252),
    [anon_sym___nullable] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4252),
    [anon_sym___real] = ACTIONS(4252),
    [anon_sym___strong] = ACTIONS(4252),
    [anon_sym___unsafe_unretained] = ACTIONS(4252),
    [anon_sym___unused] = ACTIONS(4252),
    [anon_sym___weak] = ACTIONS(4252),
    [anon_sym_alignas] = ACTIONS(4252),
    [anon_sym__Alignas] = ACTIONS(4252),
    [anon_sym_COLON] = ACTIONS(4255),
    [anon_sym_QMARK] = ACTIONS(4255),
    [anon_sym_STAR_EQ] = ACTIONS(4255),
    [anon_sym_SLASH_EQ] = ACTIONS(4255),
    [anon_sym_PERCENT_EQ] = ACTIONS(4255),
    [anon_sym_PLUS_EQ] = ACTIONS(4255),
    [anon_sym_DASH_EQ] = ACTIONS(4255),
    [anon_sym_LT_LT_EQ] = ACTIONS(4255),
    [anon_sym_GT_GT_EQ] = ACTIONS(4255),
    [anon_sym_AMP_EQ] = ACTIONS(4255),
    [anon_sym_CARET_EQ] = ACTIONS(4255),
    [anon_sym_PIPE_EQ] = ACTIONS(4255),
    [anon_sym_DASH_DASH] = ACTIONS(4255),
    [anon_sym_PLUS_PLUS] = ACTIONS(4255),
    [anon_sym_DOT] = ACTIONS(4252),
    [anon_sym_DASH_GT] = ACTIONS(4255),
    [anon_sym_AT] = ACTIONS(4255),
    [anon_sym_DQUOTE] = ACTIONS(4255),
    [anon_sym_L_DQUOTE] = ACTIONS(4255),
    [anon_sym_u_DQUOTE] = ACTIONS(4255),
    [anon_sym_U_DQUOTE] = ACTIONS(4255),
    [anon_sym_u8_DQUOTE] = ACTIONS(4255),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4252),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4252),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4252),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4252),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4252),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4252),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4252),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_API_AVAILABLE] = ACTIONS(4252),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_API_DEPRECATED] = ACTIONS(4252),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4252),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym___deprecated_msg] = ACTIONS(4252),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4252),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4252),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4252),
  },
  [STATE(1673)] = {
    [sym_identifier] = ACTIONS(4262),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4264),
    [anon_sym_COMMA] = ACTIONS(4264),
    [anon_sym_RPAREN] = ACTIONS(4264),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4264),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4264),
    [anon_sym_LPAREN2] = ACTIONS(4264),
    [anon_sym_DASH] = ACTIONS(4262),
    [anon_sym_PLUS] = ACTIONS(4262),
    [anon_sym_STAR] = ACTIONS(4262),
    [anon_sym_SLASH] = ACTIONS(4262),
    [anon_sym_PERCENT] = ACTIONS(4262),
    [anon_sym_PIPE_PIPE] = ACTIONS(4264),
    [anon_sym_AMP_AMP] = ACTIONS(4264),
    [anon_sym_PIPE] = ACTIONS(4262),
    [anon_sym_CARET] = ACTIONS(4262),
    [anon_sym_AMP] = ACTIONS(4262),
    [anon_sym_EQ_EQ] = ACTIONS(4264),
    [anon_sym_BANG_EQ] = ACTIONS(4264),
    [anon_sym_GT] = ACTIONS(4262),
    [anon_sym_GT_EQ] = ACTIONS(4264),
    [anon_sym_LT_EQ] = ACTIONS(4264),
    [anon_sym_LT] = ACTIONS(4262),
    [anon_sym_LT_LT] = ACTIONS(4262),
    [anon_sym_GT_GT] = ACTIONS(4262),
    [anon_sym_SEMI] = ACTIONS(4264),
    [anon_sym___extension__] = ACTIONS(4262),
    [anon_sym_extern] = ACTIONS(4262),
    [anon_sym___attribute__] = ACTIONS(4262),
    [anon_sym___attribute] = ACTIONS(4262),
    [anon_sym_noreturn] = ACTIONS(4262),
    [anon_sym_LBRACK] = ACTIONS(4264),
    [anon_sym_RBRACK] = ACTIONS(4264),
    [anon_sym___declspec] = ACTIONS(4262),
    [anon_sym_RBRACE] = ACTIONS(4264),
    [anon_sym_static] = ACTIONS(4262),
    [anon_sym_EQ] = ACTIONS(4262),
    [anon_sym_auto] = ACTIONS(4262),
    [anon_sym_register] = ACTIONS(4262),
    [anon_sym_inline] = ACTIONS(4262),
    [anon_sym___inline] = ACTIONS(4262),
    [anon_sym___inline__] = ACTIONS(4262),
    [anon_sym___forceinline] = ACTIONS(4262),
    [anon_sym_thread_local] = ACTIONS(4262),
    [anon_sym___thread] = ACTIONS(4262),
    [anon_sym_CG_EXTERN] = ACTIONS(4262),
    [anon_sym_CG_INLINE] = ACTIONS(4262),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4262),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4262),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4262),
    [anon_sym_IBOutlet] = ACTIONS(4262),
    [anon_sym_IBInspectable] = ACTIONS(4262),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4262),
    [anon_sym_NS_INLINE] = ACTIONS(4262),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4262),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4262),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4262),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4262),
    [anon_sym_const] = ACTIONS(4262),
    [anon_sym_constexpr] = ACTIONS(4262),
    [anon_sym_volatile] = ACTIONS(4262),
    [anon_sym_restrict] = ACTIONS(4262),
    [anon_sym___restrict__] = ACTIONS(4262),
    [anon_sym__Atomic] = ACTIONS(4262),
    [anon_sym__Noreturn] = ACTIONS(4262),
    [anon_sym__Nonnull] = ACTIONS(4262),
    [anon_sym_nullable] = ACTIONS(4262),
    [anon_sym__Complex] = ACTIONS(4262),
    [anon_sym__Nullable] = ACTIONS(4262),
    [anon_sym__Nullable_result] = ACTIONS(4262),
    [anon_sym__Null_unspecified] = ACTIONS(4262),
    [anon_sym___autoreleasing] = ACTIONS(4262),
    [anon_sym___block] = ACTIONS(4262),
    [anon_sym___bridge] = ACTIONS(4262),
    [anon_sym___bridge_retained] = ACTIONS(4262),
    [anon_sym___bridge_transfer] = ACTIONS(4262),
    [anon_sym___complex] = ACTIONS(4262),
    [anon_sym___const] = ACTIONS(4262),
    [anon_sym___imag] = ACTIONS(4262),
    [anon_sym___kindof] = ACTIONS(4262),
    [anon_sym___nonnull] = ACTIONS(4262),
    [anon_sym___nullable] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4262),
    [anon_sym___real] = ACTIONS(4262),
    [anon_sym___strong] = ACTIONS(4262),
    [anon_sym___unsafe_unretained] = ACTIONS(4262),
    [anon_sym___unused] = ACTIONS(4262),
    [anon_sym___weak] = ACTIONS(4262),
    [anon_sym_alignas] = ACTIONS(4262),
    [anon_sym__Alignas] = ACTIONS(4262),
    [anon_sym_COLON] = ACTIONS(4264),
    [anon_sym_QMARK] = ACTIONS(4264),
    [anon_sym_STAR_EQ] = ACTIONS(4264),
    [anon_sym_SLASH_EQ] = ACTIONS(4264),
    [anon_sym_PERCENT_EQ] = ACTIONS(4264),
    [anon_sym_PLUS_EQ] = ACTIONS(4264),
    [anon_sym_DASH_EQ] = ACTIONS(4264),
    [anon_sym_LT_LT_EQ] = ACTIONS(4264),
    [anon_sym_GT_GT_EQ] = ACTIONS(4264),
    [anon_sym_AMP_EQ] = ACTIONS(4264),
    [anon_sym_CARET_EQ] = ACTIONS(4264),
    [anon_sym_PIPE_EQ] = ACTIONS(4264),
    [anon_sym_DASH_DASH] = ACTIONS(4264),
    [anon_sym_PLUS_PLUS] = ACTIONS(4264),
    [anon_sym_DOT] = ACTIONS(4262),
    [anon_sym_DASH_GT] = ACTIONS(4264),
    [anon_sym_AT] = ACTIONS(4264),
    [anon_sym_DQUOTE] = ACTIONS(4264),
    [anon_sym_L_DQUOTE] = ACTIONS(4264),
    [anon_sym_u_DQUOTE] = ACTIONS(4264),
    [anon_sym_U_DQUOTE] = ACTIONS(4264),
    [anon_sym_u8_DQUOTE] = ACTIONS(4264),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4262),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4262),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4262),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4262),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4262),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4262),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4262),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_API_AVAILABLE] = ACTIONS(4262),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_API_DEPRECATED] = ACTIONS(4262),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4262),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym___deprecated_msg] = ACTIONS(4262),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4262),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4262),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4262),
  },
  [STATE(1674)] = {
    [sym_declaration] = STATE(1864),
    [sym__declaration_modifiers] = STATE(1939),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1939),
    [sym_attribute_declaration] = STATE(1939),
    [sym_ms_declspec_modifier] = STATE(1939),
    [sym_compound_statement] = STATE(2281),
    [sym_storage_class_specifier] = STATE(1939),
    [sym_type_qualifier] = STATE(1939),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1939),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1687),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1939),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1687),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4645),
    [anon_sym_SEMI] = ACTIONS(4647),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1675)] = {
    [sym_declaration] = STATE(1862),
    [sym__declaration_modifiers] = STATE(1933),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1933),
    [sym_attribute_declaration] = STATE(1933),
    [sym_ms_declspec_modifier] = STATE(1933),
    [sym_compound_statement] = STATE(2140),
    [sym_storage_class_specifier] = STATE(1933),
    [sym_type_qualifier] = STATE(1933),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1933),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1676),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1862),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1933),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1676),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4655),
    [anon_sym_SEMI] = ACTIONS(4657),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1676)] = {
    [sym_declaration] = STATE(1873),
    [sym__declaration_modifiers] = STATE(1955),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1955),
    [sym_attribute_declaration] = STATE(1955),
    [sym_ms_declspec_modifier] = STATE(1955),
    [sym_compound_statement] = STATE(2186),
    [sym_storage_class_specifier] = STATE(1955),
    [sym_type_qualifier] = STATE(1955),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1955),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1873),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1955),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4663),
    [anon_sym_SEMI] = ACTIONS(4665),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1677)] = {
    [sym_declaration] = STATE(1873),
    [sym__declaration_modifiers] = STATE(1955),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1955),
    [sym_attribute_declaration] = STATE(1955),
    [sym_ms_declspec_modifier] = STATE(1955),
    [sym_compound_statement] = STATE(2186),
    [sym_storage_class_specifier] = STATE(1955),
    [sym_type_qualifier] = STATE(1955),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1955),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1679),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1873),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1955),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1679),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4663),
    [anon_sym_SEMI] = ACTIONS(4665),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1678)] = {
    [sym_declaration] = STATE(1850),
    [sym__declaration_modifiers] = STATE(1971),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1971),
    [sym_attribute_declaration] = STATE(1971),
    [sym_ms_declspec_modifier] = STATE(1971),
    [sym_compound_statement] = STATE(2101),
    [sym_storage_class_specifier] = STATE(1971),
    [sym_type_qualifier] = STATE(1971),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1971),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1680),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1850),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1971),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1680),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4667),
    [anon_sym_SEMI] = ACTIONS(4669),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1679)] = {
    [sym_declaration] = STATE(1850),
    [sym__declaration_modifiers] = STATE(1971),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1971),
    [sym_attribute_declaration] = STATE(1971),
    [sym_ms_declspec_modifier] = STATE(1971),
    [sym_compound_statement] = STATE(2101),
    [sym_storage_class_specifier] = STATE(1971),
    [sym_type_qualifier] = STATE(1971),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1971),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1850),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1971),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4667),
    [anon_sym_SEMI] = ACTIONS(4669),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1680)] = {
    [sym_declaration] = STATE(1831),
    [sym__declaration_modifiers] = STATE(1972),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1972),
    [sym_attribute_declaration] = STATE(1972),
    [sym_ms_declspec_modifier] = STATE(1972),
    [sym_compound_statement] = STATE(2104),
    [sym_storage_class_specifier] = STATE(1972),
    [sym_type_qualifier] = STATE(1972),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1972),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1831),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1972),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4671),
    [anon_sym_SEMI] = ACTIONS(4673),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1681)] = {
    [sym_declaration] = STATE(1831),
    [sym__declaration_modifiers] = STATE(1972),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1972),
    [sym_attribute_declaration] = STATE(1972),
    [sym_ms_declspec_modifier] = STATE(1972),
    [sym_compound_statement] = STATE(2104),
    [sym_storage_class_specifier] = STATE(1972),
    [sym_type_qualifier] = STATE(1972),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1972),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1682),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1831),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1972),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1682),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4671),
    [anon_sym_SEMI] = ACTIONS(4673),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1682)] = {
    [sym_declaration] = STATE(1836),
    [sym__declaration_modifiers] = STATE(1951),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1951),
    [sym_attribute_declaration] = STATE(1951),
    [sym_ms_declspec_modifier] = STATE(1951),
    [sym_compound_statement] = STATE(2106),
    [sym_storage_class_specifier] = STATE(1951),
    [sym_type_qualifier] = STATE(1951),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1951),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1836),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1951),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4675),
    [anon_sym_SEMI] = ACTIONS(4677),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1683)] = {
    [sym_declaration] = STATE(1858),
    [sym__declaration_modifiers] = STATE(1926),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1926),
    [sym_attribute_declaration] = STATE(1926),
    [sym_ms_declspec_modifier] = STATE(1926),
    [sym_compound_statement] = STATE(2252),
    [sym_storage_class_specifier] = STATE(1926),
    [sym_type_qualifier] = STATE(1926),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1926),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1858),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1926),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4679),
    [anon_sym_SEMI] = ACTIONS(4681),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1684)] = {
    [sym_declaration] = STATE(1858),
    [sym__declaration_modifiers] = STATE(1926),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1926),
    [sym_attribute_declaration] = STATE(1926),
    [sym_ms_declspec_modifier] = STATE(1926),
    [sym_compound_statement] = STATE(2252),
    [sym_storage_class_specifier] = STATE(1926),
    [sym_type_qualifier] = STATE(1926),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1926),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1688),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1858),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1926),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1688),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4679),
    [anon_sym_SEMI] = ACTIONS(4681),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1685)] = {
    [sym_declaration] = STATE(1856),
    [sym__declaration_modifiers] = STATE(1921),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1921),
    [sym_attribute_declaration] = STATE(1921),
    [sym_ms_declspec_modifier] = STATE(1921),
    [sym_compound_statement] = STATE(2257),
    [sym_storage_class_specifier] = STATE(1921),
    [sym_type_qualifier] = STATE(1921),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1921),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1709),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1921),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1709),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4687),
    [anon_sym_SEMI] = ACTIONS(4689),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1686)] = {
    [sym_declaration] = STATE(1875),
    [sym__declaration_modifiers] = STATE(1979),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1979),
    [sym_attribute_declaration] = STATE(1979),
    [sym_ms_declspec_modifier] = STATE(1979),
    [sym_compound_statement] = STATE(2334),
    [sym_storage_class_specifier] = STATE(1979),
    [sym_type_qualifier] = STATE(1979),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1979),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1690),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1875),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1979),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1690),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4691),
    [anon_sym_SEMI] = ACTIONS(4693),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1687)] = {
    [sym_declaration] = STATE(1875),
    [sym__declaration_modifiers] = STATE(1979),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1979),
    [sym_attribute_declaration] = STATE(1979),
    [sym_ms_declspec_modifier] = STATE(1979),
    [sym_compound_statement] = STATE(2334),
    [sym_storage_class_specifier] = STATE(1979),
    [sym_type_qualifier] = STATE(1979),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1979),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1875),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1979),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4691),
    [anon_sym_SEMI] = ACTIONS(4693),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1688)] = {
    [sym_declaration] = STATE(1876),
    [sym__declaration_modifiers] = STATE(1932),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1932),
    [sym_attribute_declaration] = STATE(1932),
    [sym_ms_declspec_modifier] = STATE(1932),
    [sym_compound_statement] = STATE(2309),
    [sym_storage_class_specifier] = STATE(1932),
    [sym_type_qualifier] = STATE(1932),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1932),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1876),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1932),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4695),
    [anon_sym_SEMI] = ACTIONS(4697),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1689)] = {
    [sym_declaration] = STATE(1876),
    [sym__declaration_modifiers] = STATE(1932),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1932),
    [sym_attribute_declaration] = STATE(1932),
    [sym_ms_declspec_modifier] = STATE(1932),
    [sym_compound_statement] = STATE(2309),
    [sym_storage_class_specifier] = STATE(1932),
    [sym_type_qualifier] = STATE(1932),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1932),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1702),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1876),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1932),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1702),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4695),
    [anon_sym_SEMI] = ACTIONS(4697),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1690)] = {
    [sym_declaration] = STATE(1827),
    [sym__declaration_modifiers] = STATE(1966),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1966),
    [sym_attribute_declaration] = STATE(1966),
    [sym_ms_declspec_modifier] = STATE(1966),
    [sym_compound_statement] = STATE(2262),
    [sym_storage_class_specifier] = STATE(1966),
    [sym_type_qualifier] = STATE(1966),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1966),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1827),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1966),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4699),
    [anon_sym_SEMI] = ACTIONS(4701),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1691)] = {
    [sym_declaration] = STATE(1827),
    [sym__declaration_modifiers] = STATE(1966),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1966),
    [sym_attribute_declaration] = STATE(1966),
    [sym_ms_declspec_modifier] = STATE(1966),
    [sym_compound_statement] = STATE(2262),
    [sym_storage_class_specifier] = STATE(1966),
    [sym_type_qualifier] = STATE(1966),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1966),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1693),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1827),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1966),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1693),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4699),
    [anon_sym_SEMI] = ACTIONS(4701),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1692)] = {
    [sym_declaration] = STATE(1851),
    [sym__declaration_modifiers] = STATE(1938),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1938),
    [sym_attribute_declaration] = STATE(1938),
    [sym_ms_declspec_modifier] = STATE(1938),
    [sym_compound_statement] = STATE(2184),
    [sym_storage_class_specifier] = STATE(1938),
    [sym_type_qualifier] = STATE(1938),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1938),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1705),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1851),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1938),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1705),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4703),
    [anon_sym_SEMI] = ACTIONS(4705),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1693)] = {
    [sym_declaration] = STATE(1830),
    [sym__declaration_modifiers] = STATE(1980),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1980),
    [sym_attribute_declaration] = STATE(1980),
    [sym_ms_declspec_modifier] = STATE(1980),
    [sym_compound_statement] = STATE(2297),
    [sym_storage_class_specifier] = STATE(1980),
    [sym_type_qualifier] = STATE(1980),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1980),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1830),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1980),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4707),
    [anon_sym_SEMI] = ACTIONS(4709),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1694)] = {
    [sym_declaration] = STATE(1830),
    [sym__declaration_modifiers] = STATE(1980),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1980),
    [sym_attribute_declaration] = STATE(1980),
    [sym_ms_declspec_modifier] = STATE(1980),
    [sym_compound_statement] = STATE(2297),
    [sym_storage_class_specifier] = STATE(1980),
    [sym_type_qualifier] = STATE(1980),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1980),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1696),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1830),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1980),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1696),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4707),
    [anon_sym_SEMI] = ACTIONS(4709),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1695)] = {
    [sym_declaration] = STATE(1834),
    [sym__declaration_modifiers] = STATE(1931),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1931),
    [sym_attribute_declaration] = STATE(1931),
    [sym_ms_declspec_modifier] = STATE(1931),
    [sym_compound_statement] = STATE(2345),
    [sym_storage_class_specifier] = STATE(1931),
    [sym_type_qualifier] = STATE(1931),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1931),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1697),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1834),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1931),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1697),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4711),
    [anon_sym_SEMI] = ACTIONS(4713),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1696)] = {
    [sym_declaration] = STATE(1834),
    [sym__declaration_modifiers] = STATE(1931),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1931),
    [sym_attribute_declaration] = STATE(1931),
    [sym_ms_declspec_modifier] = STATE(1931),
    [sym_compound_statement] = STATE(2345),
    [sym_storage_class_specifier] = STATE(1931),
    [sym_type_qualifier] = STATE(1931),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1931),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1834),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1931),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4711),
    [anon_sym_SEMI] = ACTIONS(4713),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1697)] = {
    [sym_declaration] = STATE(1838),
    [sym__declaration_modifiers] = STATE(1975),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1975),
    [sym_attribute_declaration] = STATE(1975),
    [sym_ms_declspec_modifier] = STATE(1975),
    [sym_compound_statement] = STATE(2308),
    [sym_storage_class_specifier] = STATE(1975),
    [sym_type_qualifier] = STATE(1975),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1975),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1838),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1975),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4715),
    [anon_sym_SEMI] = ACTIONS(4717),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1698)] = {
    [sym_declaration] = STATE(1838),
    [sym__declaration_modifiers] = STATE(1975),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1975),
    [sym_attribute_declaration] = STATE(1975),
    [sym_ms_declspec_modifier] = STATE(1975),
    [sym_compound_statement] = STATE(2308),
    [sym_storage_class_specifier] = STATE(1975),
    [sym_type_qualifier] = STATE(1975),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1975),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1699),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1838),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1975),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1699),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4715),
    [anon_sym_SEMI] = ACTIONS(4717),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1699)] = {
    [sym_declaration] = STATE(1840),
    [sym__declaration_modifiers] = STATE(1935),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1935),
    [sym_attribute_declaration] = STATE(1935),
    [sym_ms_declspec_modifier] = STATE(1935),
    [sym_compound_statement] = STATE(2312),
    [sym_storage_class_specifier] = STATE(1935),
    [sym_type_qualifier] = STATE(1935),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1935),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1840),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1935),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4719),
    [anon_sym_SEMI] = ACTIONS(4721),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1700)] = {
    [sym_string_literal] = STATE(1627),
    [sym_identifier] = ACTIONS(4004),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4006),
    [anon_sym_COMMA] = ACTIONS(4006),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4006),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4006),
    [anon_sym_LPAREN2] = ACTIONS(4006),
    [anon_sym_DASH] = ACTIONS(4004),
    [anon_sym_PLUS] = ACTIONS(4004),
    [anon_sym_STAR] = ACTIONS(4004),
    [anon_sym_SLASH] = ACTIONS(4004),
    [anon_sym_PERCENT] = ACTIONS(4004),
    [anon_sym_PIPE_PIPE] = ACTIONS(4006),
    [anon_sym_AMP_AMP] = ACTIONS(4006),
    [anon_sym_PIPE] = ACTIONS(4004),
    [anon_sym_CARET] = ACTIONS(4004),
    [anon_sym_AMP] = ACTIONS(4004),
    [anon_sym_EQ_EQ] = ACTIONS(4006),
    [anon_sym_BANG_EQ] = ACTIONS(4006),
    [anon_sym_GT] = ACTIONS(4004),
    [anon_sym_GT_EQ] = ACTIONS(4006),
    [anon_sym_LT_EQ] = ACTIONS(4006),
    [anon_sym_LT] = ACTIONS(4004),
    [anon_sym_LT_LT] = ACTIONS(4004),
    [anon_sym_GT_GT] = ACTIONS(4004),
    [anon_sym___extension__] = ACTIONS(4004),
    [anon_sym_extern] = ACTIONS(4004),
    [anon_sym___attribute__] = ACTIONS(4004),
    [anon_sym___attribute] = ACTIONS(4004),
    [anon_sym_noreturn] = ACTIONS(4004),
    [anon_sym_LBRACK] = ACTIONS(4006),
    [anon_sym_RBRACK] = ACTIONS(4006),
    [anon_sym___declspec] = ACTIONS(4004),
    [anon_sym_RBRACE] = ACTIONS(4006),
    [anon_sym_static] = ACTIONS(4004),
    [anon_sym_EQ] = ACTIONS(4004),
    [anon_sym_auto] = ACTIONS(4004),
    [anon_sym_register] = ACTIONS(4004),
    [anon_sym_inline] = ACTIONS(4004),
    [anon_sym___inline] = ACTIONS(4004),
    [anon_sym___inline__] = ACTIONS(4004),
    [anon_sym___forceinline] = ACTIONS(4004),
    [anon_sym_thread_local] = ACTIONS(4004),
    [anon_sym___thread] = ACTIONS(4004),
    [anon_sym_CG_EXTERN] = ACTIONS(4004),
    [anon_sym_CG_INLINE] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4004),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4004),
    [anon_sym_IBOutlet] = ACTIONS(4004),
    [anon_sym_IBInspectable] = ACTIONS(4004),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4004),
    [anon_sym_NS_INLINE] = ACTIONS(4004),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4004),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4004),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4004),
    [anon_sym_const] = ACTIONS(4004),
    [anon_sym_constexpr] = ACTIONS(4004),
    [anon_sym_volatile] = ACTIONS(4004),
    [anon_sym_restrict] = ACTIONS(4004),
    [anon_sym___restrict__] = ACTIONS(4004),
    [anon_sym__Atomic] = ACTIONS(4004),
    [anon_sym__Noreturn] = ACTIONS(4004),
    [anon_sym__Nonnull] = ACTIONS(4004),
    [anon_sym_nullable] = ACTIONS(4004),
    [anon_sym__Complex] = ACTIONS(4004),
    [anon_sym__Nullable] = ACTIONS(4004),
    [anon_sym__Nullable_result] = ACTIONS(4004),
    [anon_sym__Null_unspecified] = ACTIONS(4004),
    [anon_sym___autoreleasing] = ACTIONS(4004),
    [anon_sym___block] = ACTIONS(4004),
    [anon_sym___bridge] = ACTIONS(4004),
    [anon_sym___bridge_retained] = ACTIONS(4004),
    [anon_sym___bridge_transfer] = ACTIONS(4004),
    [anon_sym___complex] = ACTIONS(4004),
    [anon_sym___const] = ACTIONS(4004),
    [anon_sym___imag] = ACTIONS(4004),
    [anon_sym___kindof] = ACTIONS(4004),
    [anon_sym___nonnull] = ACTIONS(4004),
    [anon_sym___nullable] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4004),
    [anon_sym___real] = ACTIONS(4004),
    [anon_sym___strong] = ACTIONS(4004),
    [anon_sym___unsafe_unretained] = ACTIONS(4004),
    [anon_sym___unused] = ACTIONS(4004),
    [anon_sym___weak] = ACTIONS(4004),
    [anon_sym_alignas] = ACTIONS(4004),
    [anon_sym__Alignas] = ACTIONS(4004),
    [anon_sym_COLON] = ACTIONS(4006),
    [anon_sym_QMARK] = ACTIONS(4006),
    [anon_sym_STAR_EQ] = ACTIONS(4006),
    [anon_sym_SLASH_EQ] = ACTIONS(4006),
    [anon_sym_PERCENT_EQ] = ACTIONS(4006),
    [anon_sym_PLUS_EQ] = ACTIONS(4006),
    [anon_sym_DASH_EQ] = ACTIONS(4006),
    [anon_sym_LT_LT_EQ] = ACTIONS(4006),
    [anon_sym_GT_GT_EQ] = ACTIONS(4006),
    [anon_sym_AMP_EQ] = ACTIONS(4006),
    [anon_sym_CARET_EQ] = ACTIONS(4006),
    [anon_sym_PIPE_EQ] = ACTIONS(4006),
    [anon_sym_DASH_DASH] = ACTIONS(4006),
    [anon_sym_PLUS_PLUS] = ACTIONS(4006),
    [anon_sym_DOT] = ACTIONS(4004),
    [anon_sym_DASH_GT] = ACTIONS(4006),
    [anon_sym_AT] = ACTIONS(4625),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4004),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4004),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4004),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_API_AVAILABLE] = ACTIONS(4004),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_API_DEPRECATED] = ACTIONS(4004),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym___deprecated_msg] = ACTIONS(4004),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4004),
  },
  [STATE(1701)] = {
    [sym_declaration] = STATE(1860),
    [sym__declaration_modifiers] = STATE(1930),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1930),
    [sym_attribute_declaration] = STATE(1930),
    [sym_ms_declspec_modifier] = STATE(1930),
    [sym_compound_statement] = STATE(2260),
    [sym_storage_class_specifier] = STATE(1930),
    [sym_type_qualifier] = STATE(1930),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1930),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1706),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1860),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1930),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1706),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4723),
    [anon_sym_SEMI] = ACTIONS(4725),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1702)] = {
    [sym_declaration] = STATE(1860),
    [sym__declaration_modifiers] = STATE(1930),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1930),
    [sym_attribute_declaration] = STATE(1930),
    [sym_ms_declspec_modifier] = STATE(1930),
    [sym_compound_statement] = STATE(2260),
    [sym_storage_class_specifier] = STATE(1930),
    [sym_type_qualifier] = STATE(1930),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1930),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1860),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1930),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4723),
    [anon_sym_SEMI] = ACTIONS(4725),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1703)] = {
    [sym_declaration] = STATE(1855),
    [sym__declaration_modifiers] = STATE(1982),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1982),
    [sym_attribute_declaration] = STATE(1982),
    [sym_ms_declspec_modifier] = STATE(1982),
    [sym_compound_statement] = STATE(2163),
    [sym_storage_class_specifier] = STATE(1982),
    [sym_type_qualifier] = STATE(1982),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1982),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1711),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1855),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1982),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1711),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4727),
    [anon_sym_SEMI] = ACTIONS(4729),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1704)] = {
    [sym_declaration] = STATE(1853),
    [sym__declaration_modifiers] = STATE(1954),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1954),
    [sym_attribute_declaration] = STATE(1954),
    [sym_ms_declspec_modifier] = STATE(1954),
    [sym_compound_statement] = STATE(2268),
    [sym_storage_class_specifier] = STATE(1954),
    [sym_type_qualifier] = STATE(1954),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1954),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1683),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1853),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1954),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1683),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4731),
    [anon_sym_SEMI] = ACTIONS(4733),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1705)] = {
    [sym_declaration] = STATE(1855),
    [sym__declaration_modifiers] = STATE(1982),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1982),
    [sym_attribute_declaration] = STATE(1982),
    [sym_ms_declspec_modifier] = STATE(1982),
    [sym_compound_statement] = STATE(2163),
    [sym_storage_class_specifier] = STATE(1982),
    [sym_type_qualifier] = STATE(1982),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1982),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1855),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1982),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4727),
    [anon_sym_SEMI] = ACTIONS(4729),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1706)] = {
    [sym_declaration] = STATE(1863),
    [sym__declaration_modifiers] = STATE(1937),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1937),
    [sym_attribute_declaration] = STATE(1937),
    [sym_ms_declspec_modifier] = STATE(1937),
    [sym_compound_statement] = STATE(2271),
    [sym_storage_class_specifier] = STATE(1937),
    [sym_type_qualifier] = STATE(1937),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1937),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1863),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1937),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4735),
    [anon_sym_SEMI] = ACTIONS(4737),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1707)] = {
    [sym_declaration] = STATE(1863),
    [sym__declaration_modifiers] = STATE(1937),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1937),
    [sym_attribute_declaration] = STATE(1937),
    [sym_ms_declspec_modifier] = STATE(1937),
    [sym_compound_statement] = STATE(2271),
    [sym_storage_class_specifier] = STATE(1937),
    [sym_type_qualifier] = STATE(1937),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1937),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(1708),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1863),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1937),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(1708),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4735),
    [anon_sym_SEMI] = ACTIONS(4737),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1708)] = {
    [sym_declaration] = STATE(1867),
    [sym__declaration_modifiers] = STATE(1950),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1950),
    [sym_attribute_declaration] = STATE(1950),
    [sym_ms_declspec_modifier] = STATE(1950),
    [sym_compound_statement] = STATE(2277),
    [sym_storage_class_specifier] = STATE(1950),
    [sym_type_qualifier] = STATE(1950),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1950),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1867),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1950),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4739),
    [anon_sym_SEMI] = ACTIONS(4741),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1709)] = {
    [sym_declaration] = STATE(1853),
    [sym__declaration_modifiers] = STATE(1954),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1954),
    [sym_attribute_declaration] = STATE(1954),
    [sym_ms_declspec_modifier] = STATE(1954),
    [sym_compound_statement] = STATE(2268),
    [sym_storage_class_specifier] = STATE(1954),
    [sym_type_qualifier] = STATE(1954),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1954),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1853),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1954),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4731),
    [anon_sym_SEMI] = ACTIONS(4733),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1710)] = {
    [sym_string_literal] = STATE(1627),
    [sym_identifier] = ACTIONS(4743),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4006),
    [anon_sym_COMMA] = ACTIONS(4006),
    [anon_sym_RPAREN] = ACTIONS(4006),
    [anon_sym_LPAREN2] = ACTIONS(4006),
    [anon_sym_DASH] = ACTIONS(4004),
    [anon_sym_PLUS] = ACTIONS(4004),
    [anon_sym_STAR] = ACTIONS(4004),
    [anon_sym_SLASH] = ACTIONS(4004),
    [anon_sym_PERCENT] = ACTIONS(4004),
    [anon_sym_PIPE_PIPE] = ACTIONS(4006),
    [anon_sym_AMP_AMP] = ACTIONS(4006),
    [anon_sym_PIPE] = ACTIONS(4004),
    [anon_sym_CARET] = ACTIONS(4004),
    [anon_sym_AMP] = ACTIONS(4004),
    [anon_sym_EQ_EQ] = ACTIONS(4006),
    [anon_sym_BANG_EQ] = ACTIONS(4006),
    [anon_sym_GT] = ACTIONS(4004),
    [anon_sym_GT_EQ] = ACTIONS(4006),
    [anon_sym_LT_EQ] = ACTIONS(4006),
    [anon_sym_LT] = ACTIONS(4004),
    [anon_sym_LT_LT] = ACTIONS(4004),
    [anon_sym_GT_GT] = ACTIONS(4004),
    [anon_sym_SEMI] = ACTIONS(4006),
    [anon_sym___extension__] = ACTIONS(4004),
    [anon_sym_extern] = ACTIONS(4004),
    [anon_sym___attribute__] = ACTIONS(4004),
    [anon_sym___attribute] = ACTIONS(4004),
    [anon_sym_noreturn] = ACTIONS(4004),
    [anon_sym_LBRACK] = ACTIONS(4006),
    [anon_sym_RBRACK] = ACTIONS(4006),
    [anon_sym___declspec] = ACTIONS(4004),
    [anon_sym_RBRACE] = ACTIONS(4006),
    [anon_sym_static] = ACTIONS(4004),
    [anon_sym_EQ] = ACTIONS(4004),
    [anon_sym_auto] = ACTIONS(4004),
    [anon_sym_register] = ACTIONS(4004),
    [anon_sym_inline] = ACTIONS(4004),
    [anon_sym___inline] = ACTIONS(4004),
    [anon_sym___inline__] = ACTIONS(4004),
    [anon_sym___forceinline] = ACTIONS(4004),
    [anon_sym_thread_local] = ACTIONS(4004),
    [anon_sym___thread] = ACTIONS(4004),
    [anon_sym_CG_EXTERN] = ACTIONS(4004),
    [anon_sym_CG_INLINE] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4004),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4004),
    [anon_sym_IBOutlet] = ACTIONS(4004),
    [anon_sym_IBInspectable] = ACTIONS(4004),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4004),
    [anon_sym_NS_INLINE] = ACTIONS(4004),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4004),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4004),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4004),
    [anon_sym_const] = ACTIONS(4004),
    [anon_sym_constexpr] = ACTIONS(4004),
    [anon_sym_volatile] = ACTIONS(4004),
    [anon_sym_restrict] = ACTIONS(4004),
    [anon_sym___restrict__] = ACTIONS(4004),
    [anon_sym__Atomic] = ACTIONS(4004),
    [anon_sym__Noreturn] = ACTIONS(4004),
    [anon_sym__Nonnull] = ACTIONS(4004),
    [anon_sym_nullable] = ACTIONS(4004),
    [anon_sym__Complex] = ACTIONS(4004),
    [anon_sym__Nullable] = ACTIONS(4004),
    [anon_sym__Nullable_result] = ACTIONS(4004),
    [anon_sym__Null_unspecified] = ACTIONS(4004),
    [anon_sym___autoreleasing] = ACTIONS(4004),
    [anon_sym___block] = ACTIONS(4004),
    [anon_sym___bridge] = ACTIONS(4004),
    [anon_sym___bridge_retained] = ACTIONS(4004),
    [anon_sym___bridge_transfer] = ACTIONS(4004),
    [anon_sym___complex] = ACTIONS(4004),
    [anon_sym___const] = ACTIONS(4004),
    [anon_sym___imag] = ACTIONS(4004),
    [anon_sym___kindof] = ACTIONS(4004),
    [anon_sym___nonnull] = ACTIONS(4004),
    [anon_sym___nullable] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4004),
    [anon_sym___real] = ACTIONS(4004),
    [anon_sym___strong] = ACTIONS(4004),
    [anon_sym___unsafe_unretained] = ACTIONS(4004),
    [anon_sym___unused] = ACTIONS(4004),
    [anon_sym___weak] = ACTIONS(4004),
    [anon_sym_alignas] = ACTIONS(4004),
    [anon_sym__Alignas] = ACTIONS(4004),
    [anon_sym_COLON] = ACTIONS(4006),
    [anon_sym_QMARK] = ACTIONS(4006),
    [anon_sym_STAR_EQ] = ACTIONS(4006),
    [anon_sym_SLASH_EQ] = ACTIONS(4006),
    [anon_sym_PERCENT_EQ] = ACTIONS(4006),
    [anon_sym_PLUS_EQ] = ACTIONS(4006),
    [anon_sym_DASH_EQ] = ACTIONS(4006),
    [anon_sym_LT_LT_EQ] = ACTIONS(4006),
    [anon_sym_GT_GT_EQ] = ACTIONS(4006),
    [anon_sym_AMP_EQ] = ACTIONS(4006),
    [anon_sym_CARET_EQ] = ACTIONS(4006),
    [anon_sym_PIPE_EQ] = ACTIONS(4006),
    [anon_sym_DASH_DASH] = ACTIONS(4006),
    [anon_sym_PLUS_PLUS] = ACTIONS(4006),
    [anon_sym_DOT] = ACTIONS(4004),
    [anon_sym_DASH_GT] = ACTIONS(4006),
    [anon_sym_AT] = ACTIONS(4625),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4004),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4004),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4004),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_API_AVAILABLE] = ACTIONS(4004),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_API_DEPRECATED] = ACTIONS(4004),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym___deprecated_msg] = ACTIONS(4004),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4004),
  },
  [STATE(1711)] = {
    [sym_declaration] = STATE(1862),
    [sym__declaration_modifiers] = STATE(1933),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1933),
    [sym_attribute_declaration] = STATE(1933),
    [sym_ms_declspec_modifier] = STATE(1933),
    [sym_compound_statement] = STATE(2140),
    [sym_storage_class_specifier] = STATE(1933),
    [sym_type_qualifier] = STATE(1933),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1933),
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1862),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1933),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(4643),
    [anon_sym_COMMA] = ACTIONS(4655),
    [anon_sym_SEMI] = ACTIONS(4657),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1712)] = {
    [sym_function_definition] = STATE(298),
    [sym_declaration] = STATE(298),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4210),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_ms_call_modifier] = STATE(2171),
    [sym_declaration_list] = STATE(298),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(4745),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1713)] = {
    [sym_function_definition] = STATE(1066),
    [sym_declaration] = STATE(1066),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4206),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_ms_call_modifier] = STATE(2161),
    [sym_declaration_list] = STATE(1066),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(4747),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1714)] = {
    [sym_function_definition] = STATE(668),
    [sym_declaration] = STATE(668),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4201),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_ms_call_modifier] = STATE(2177),
    [sym_declaration_list] = STATE(668),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(4749),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1715)] = {
    [sym_declaration] = STATE(1858),
    [sym__declaration_modifiers] = STATE(1926),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1926),
    [sym_attribute_declaration] = STATE(1926),
    [sym_ms_declspec_modifier] = STATE(1926),
    [sym_compound_statement] = STATE(2252),
    [sym_storage_class_specifier] = STATE(1926),
    [sym_type_qualifier] = STATE(1926),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1926),
    [sym_keyword_declarator] = STATE(2744),
    [sym_method_parameter] = STATE(1689),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1858),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1926),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(4751),
    [anon_sym_COMMA] = ACTIONS(4753),
    [anon_sym_SEMI] = ACTIONS(4681),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1716)] = {
    [sym_declaration] = STATE(1862),
    [sym__declaration_modifiers] = STATE(1933),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1933),
    [sym_attribute_declaration] = STATE(1933),
    [sym_ms_declspec_modifier] = STATE(1933),
    [sym_compound_statement] = STATE(2140),
    [sym_storage_class_specifier] = STATE(1933),
    [sym_type_qualifier] = STATE(1933),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1933),
    [sym_keyword_declarator] = STATE(2744),
    [sym_method_parameter] = STATE(1677),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1862),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1933),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(4751),
    [anon_sym_COMMA] = ACTIONS(4755),
    [anon_sym_SEMI] = ACTIONS(4657),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1717)] = {
    [sym_declaration] = STATE(1864),
    [sym__declaration_modifiers] = STATE(1939),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1939),
    [sym_attribute_declaration] = STATE(1939),
    [sym_ms_declspec_modifier] = STATE(1939),
    [sym_compound_statement] = STATE(2281),
    [sym_storage_class_specifier] = STATE(1939),
    [sym_type_qualifier] = STATE(1939),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1939),
    [sym_keyword_declarator] = STATE(2744),
    [sym_method_parameter] = STATE(1686),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1939),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(4751),
    [anon_sym_COMMA] = ACTIONS(4757),
    [anon_sym_SEMI] = ACTIONS(4647),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1718)] = {
    [sym_declaration] = STATE(1855),
    [sym__declaration_modifiers] = STATE(1982),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1982),
    [sym_attribute_declaration] = STATE(1982),
    [sym_ms_declspec_modifier] = STATE(1982),
    [sym_compound_statement] = STATE(2163),
    [sym_storage_class_specifier] = STATE(1982),
    [sym_type_qualifier] = STATE(1982),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1982),
    [sym_keyword_declarator] = STATE(2744),
    [sym_method_parameter] = STATE(1675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1855),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1982),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(4751),
    [anon_sym_COMMA] = ACTIONS(4759),
    [anon_sym_SEMI] = ACTIONS(4729),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1719)] = {
    [sym_declaration] = STATE(1856),
    [sym__declaration_modifiers] = STATE(1921),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1921),
    [sym_attribute_declaration] = STATE(1921),
    [sym_ms_declspec_modifier] = STATE(1921),
    [sym_compound_statement] = STATE(2257),
    [sym_storage_class_specifier] = STATE(1921),
    [sym_type_qualifier] = STATE(1921),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1921),
    [sym_keyword_declarator] = STATE(2744),
    [sym_method_parameter] = STATE(1704),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1921),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(4751),
    [anon_sym_COMMA] = ACTIONS(4761),
    [anon_sym_SEMI] = ACTIONS(4689),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1720)] = {
    [sym_declaration] = STATE(1827),
    [sym__declaration_modifiers] = STATE(1966),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1966),
    [sym_attribute_declaration] = STATE(1966),
    [sym_ms_declspec_modifier] = STATE(1966),
    [sym_compound_statement] = STATE(2262),
    [sym_storage_class_specifier] = STATE(1966),
    [sym_type_qualifier] = STATE(1966),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1966),
    [sym_keyword_declarator] = STATE(2744),
    [sym_method_parameter] = STATE(1694),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1827),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1966),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(4751),
    [anon_sym_COMMA] = ACTIONS(4763),
    [anon_sym_SEMI] = ACTIONS(4701),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1721)] = {
    [sym_function_definition] = STATE(896),
    [sym_declaration] = STATE(896),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4236),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_ms_call_modifier] = STATE(2168),
    [sym_declaration_list] = STATE(896),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(4765),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1722)] = {
    [sym_declaration] = STATE(1875),
    [sym__declaration_modifiers] = STATE(1979),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1979),
    [sym_attribute_declaration] = STATE(1979),
    [sym_ms_declspec_modifier] = STATE(1979),
    [sym_compound_statement] = STATE(2334),
    [sym_storage_class_specifier] = STATE(1979),
    [sym_type_qualifier] = STATE(1979),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1979),
    [sym_keyword_declarator] = STATE(2744),
    [sym_method_parameter] = STATE(1691),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1875),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1979),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(4751),
    [anon_sym_COMMA] = ACTIONS(4767),
    [anon_sym_SEMI] = ACTIONS(4693),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1723)] = {
    [sym_declaration] = STATE(1853),
    [sym__declaration_modifiers] = STATE(1954),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1954),
    [sym_attribute_declaration] = STATE(1954),
    [sym_ms_declspec_modifier] = STATE(1954),
    [sym_compound_statement] = STATE(2268),
    [sym_storage_class_specifier] = STATE(1954),
    [sym_type_qualifier] = STATE(1954),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1954),
    [sym_keyword_declarator] = STATE(2744),
    [sym_method_parameter] = STATE(1684),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1853),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1954),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(4751),
    [anon_sym_COMMA] = ACTIONS(4769),
    [anon_sym_SEMI] = ACTIONS(4733),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1724)] = {
    [sym_function_definition] = STATE(515),
    [sym_declaration] = STATE(515),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4223),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_ms_call_modifier] = STATE(2175),
    [sym_declaration_list] = STATE(515),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___cdecl] = ACTIONS(45),
    [anon_sym___clrcall] = ACTIONS(45),
    [anon_sym___stdcall] = ACTIONS(45),
    [anon_sym___fastcall] = ACTIONS(45),
    [anon_sym___thiscall] = ACTIONS(45),
    [anon_sym___vectorcall] = ACTIONS(45),
    [anon_sym_LBRACE] = ACTIONS(4771),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1725)] = {
    [sym_declaration] = STATE(1851),
    [sym__declaration_modifiers] = STATE(1938),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1938),
    [sym_attribute_declaration] = STATE(1938),
    [sym_ms_declspec_modifier] = STATE(1938),
    [sym_compound_statement] = STATE(2184),
    [sym_storage_class_specifier] = STATE(1938),
    [sym_type_qualifier] = STATE(1938),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1938),
    [sym_keyword_declarator] = STATE(2744),
    [sym_method_parameter] = STATE(1703),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1851),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1938),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(4751),
    [anon_sym_COMMA] = ACTIONS(4773),
    [anon_sym_SEMI] = ACTIONS(4705),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1726)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5287),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4078),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1727)] = {
    [sym_string_literal] = STATE(1916),
    [sym_identifier] = ACTIONS(4024),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4024),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4024),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym___extension__] = ACTIONS(4024),
    [anon_sym_extern] = ACTIONS(4024),
    [anon_sym___attribute__] = ACTIONS(4024),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4024),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4024),
    [anon_sym_RBRACE] = ACTIONS(4026),
    [anon_sym_static] = ACTIONS(4024),
    [anon_sym_EQ] = ACTIONS(4775),
    [anon_sym_auto] = ACTIONS(4024),
    [anon_sym_register] = ACTIONS(4024),
    [anon_sym_inline] = ACTIONS(4024),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4024),
    [anon_sym___forceinline] = ACTIONS(4024),
    [anon_sym_thread_local] = ACTIONS(4024),
    [anon_sym___thread] = ACTIONS(4024),
    [anon_sym_CG_EXTERN] = ACTIONS(4024),
    [anon_sym_CG_INLINE] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4024),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4024),
    [anon_sym_IBOutlet] = ACTIONS(4024),
    [anon_sym_IBInspectable] = ACTIONS(4024),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4024),
    [anon_sym_NS_INLINE] = ACTIONS(4024),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4024),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4024),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4024),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4024),
    [anon_sym_volatile] = ACTIONS(4024),
    [anon_sym_restrict] = ACTIONS(4024),
    [anon_sym___restrict__] = ACTIONS(4024),
    [anon_sym__Atomic] = ACTIONS(4024),
    [anon_sym__Noreturn] = ACTIONS(4024),
    [anon_sym__Nonnull] = ACTIONS(4024),
    [anon_sym_nullable] = ACTIONS(4024),
    [anon_sym__Complex] = ACTIONS(4024),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4024),
    [anon_sym__Null_unspecified] = ACTIONS(4024),
    [anon_sym___autoreleasing] = ACTIONS(4024),
    [anon_sym___block] = ACTIONS(4024),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4024),
    [anon_sym___bridge_transfer] = ACTIONS(4024),
    [anon_sym___complex] = ACTIONS(4024),
    [anon_sym___const] = ACTIONS(4024),
    [anon_sym___imag] = ACTIONS(4024),
    [anon_sym___kindof] = ACTIONS(4024),
    [anon_sym___nonnull] = ACTIONS(4024),
    [anon_sym___nullable] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4024),
    [anon_sym___real] = ACTIONS(4024),
    [anon_sym___strong] = ACTIONS(4024),
    [anon_sym___unsafe_unretained] = ACTIONS(4024),
    [anon_sym___unused] = ACTIONS(4024),
    [anon_sym___weak] = ACTIONS(4024),
    [anon_sym_alignas] = ACTIONS(4024),
    [anon_sym__Alignas] = ACTIONS(4024),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4777),
    [anon_sym_SLASH_EQ] = ACTIONS(4777),
    [anon_sym_PERCENT_EQ] = ACTIONS(4777),
    [anon_sym_PLUS_EQ] = ACTIONS(4777),
    [anon_sym_DASH_EQ] = ACTIONS(4777),
    [anon_sym_LT_LT_EQ] = ACTIONS(4777),
    [anon_sym_GT_GT_EQ] = ACTIONS(4777),
    [anon_sym_AMP_EQ] = ACTIONS(4777),
    [anon_sym_CARET_EQ] = ACTIONS(4777),
    [anon_sym_PIPE_EQ] = ACTIONS(4777),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4024),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4024),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_API_AVAILABLE] = ACTIONS(4024),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_API_DEPRECATED] = ACTIONS(4024),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym___deprecated_msg] = ACTIONS(4024),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4024),
  },
  [STATE(1728)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4779),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4781),
    [anon_sym_COMMA] = ACTIONS(4783),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4791),
    [anon_sym_AMP_AMP] = ACTIONS(4793),
    [anon_sym_PIPE] = ACTIONS(4795),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4783),
    [anon_sym___extension__] = ACTIONS(4779),
    [anon_sym_extern] = ACTIONS(4779),
    [anon_sym___attribute__] = ACTIONS(4779),
    [anon_sym___attribute] = ACTIONS(4779),
    [anon_sym_noreturn] = ACTIONS(4779),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4779),
    [anon_sym_LBRACE] = ACTIONS(4783),
    [anon_sym_signed] = ACTIONS(4779),
    [anon_sym_unsigned] = ACTIONS(4779),
    [anon_sym_long] = ACTIONS(4779),
    [anon_sym_short] = ACTIONS(4779),
    [anon_sym_static] = ACTIONS(4779),
    [anon_sym_ATautoreleasepool] = ACTIONS(4783),
    [anon_sym_auto] = ACTIONS(4779),
    [anon_sym_register] = ACTIONS(4779),
    [anon_sym_inline] = ACTIONS(4779),
    [anon_sym___inline] = ACTIONS(4779),
    [anon_sym___inline__] = ACTIONS(4779),
    [anon_sym___forceinline] = ACTIONS(4779),
    [anon_sym_thread_local] = ACTIONS(4779),
    [anon_sym___thread] = ACTIONS(4779),
    [anon_sym_CG_EXTERN] = ACTIONS(4779),
    [anon_sym_CG_INLINE] = ACTIONS(4779),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4779),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4779),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4779),
    [anon_sym_IBOutlet] = ACTIONS(4779),
    [anon_sym_IBInspectable] = ACTIONS(4779),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4779),
    [anon_sym_NS_INLINE] = ACTIONS(4779),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4779),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4779),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4779),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4779),
    [anon_sym_const] = ACTIONS(4779),
    [anon_sym_constexpr] = ACTIONS(4779),
    [anon_sym_volatile] = ACTIONS(4779),
    [anon_sym_restrict] = ACTIONS(4779),
    [anon_sym___restrict__] = ACTIONS(4779),
    [anon_sym__Atomic] = ACTIONS(4779),
    [anon_sym__Noreturn] = ACTIONS(4779),
    [anon_sym__Nonnull] = ACTIONS(4779),
    [anon_sym_nullable] = ACTIONS(4779),
    [anon_sym__Complex] = ACTIONS(4779),
    [anon_sym__Nullable] = ACTIONS(4779),
    [anon_sym__Nullable_result] = ACTIONS(4779),
    [anon_sym__Null_unspecified] = ACTIONS(4779),
    [anon_sym___autoreleasing] = ACTIONS(4779),
    [anon_sym___block] = ACTIONS(4779),
    [anon_sym___bridge] = ACTIONS(4779),
    [anon_sym___bridge_retained] = ACTIONS(4779),
    [anon_sym___bridge_transfer] = ACTIONS(4779),
    [anon_sym___complex] = ACTIONS(4779),
    [anon_sym___const] = ACTIONS(4779),
    [anon_sym___imag] = ACTIONS(4779),
    [anon_sym___kindof] = ACTIONS(4779),
    [anon_sym___nonnull] = ACTIONS(4779),
    [anon_sym___nullable] = ACTIONS(4779),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4779),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4779),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4779),
    [anon_sym___real] = ACTIONS(4779),
    [anon_sym___strong] = ACTIONS(4779),
    [anon_sym___unsafe_unretained] = ACTIONS(4779),
    [anon_sym___unused] = ACTIONS(4779),
    [anon_sym___weak] = ACTIONS(4779),
    [anon_sym_alignas] = ACTIONS(4779),
    [anon_sym__Alignas] = ACTIONS(4779),
    [sym_primitive_type] = ACTIONS(4779),
    [anon_sym_enum] = ACTIONS(4779),
    [anon_sym_struct] = ACTIONS(4779),
    [anon_sym_union] = ACTIONS(4779),
    [anon_sym_QMARK] = ACTIONS(4809),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4779),
    [anon_sym___typeof] = ACTIONS(4779),
    [anon_sym_typeof] = ACTIONS(4779),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4779),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4779),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4779),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4779),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4779),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4779),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4779),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4779),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4779),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4779),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4779),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4779),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4779),
    [anon_sym_API_AVAILABLE] = ACTIONS(4779),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4779),
    [anon_sym_API_DEPRECATED] = ACTIONS(4779),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4779),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4779),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4779),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4779),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4779),
    [anon_sym___deprecated_msg] = ACTIONS(4779),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4779),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4779),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4779),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4779),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4779),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4779),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4779),
    [anon_sym_BOOL] = ACTIONS(4779),
    [anon_sym_IMP] = ACTIONS(4779),
    [anon_sym_SEL] = ACTIONS(4779),
    [anon_sym_Class] = ACTIONS(4779),
    [anon_sym_id] = ACTIONS(4779),
  },
  [STATE(1729)] = {
    [sym_declaration] = STATE(1827),
    [sym__declaration_modifiers] = STATE(1966),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1966),
    [sym_attribute_declaration] = STATE(1966),
    [sym_ms_declspec_modifier] = STATE(1966),
    [sym_compound_statement] = STATE(2262),
    [sym_storage_class_specifier] = STATE(1966),
    [sym_type_qualifier] = STATE(1966),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1966),
    [sym_method_parameter] = STATE(1694),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1827),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1966),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4699),
    [anon_sym_SEMI] = ACTIONS(4811),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1730)] = {
    [sym_declaration] = STATE(1862),
    [sym__declaration_modifiers] = STATE(1933),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1933),
    [sym_attribute_declaration] = STATE(1933),
    [sym_ms_declspec_modifier] = STATE(1933),
    [sym_compound_statement] = STATE(2140),
    [sym_storage_class_specifier] = STATE(1933),
    [sym_type_qualifier] = STATE(1933),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1933),
    [sym_method_parameter] = STATE(1677),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1862),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1933),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4655),
    [anon_sym_SEMI] = ACTIONS(4813),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1731)] = {
    [sym_declaration] = STATE(1860),
    [sym__declaration_modifiers] = STATE(1930),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1930),
    [sym_attribute_declaration] = STATE(1930),
    [sym_ms_declspec_modifier] = STATE(1930),
    [sym_compound_statement] = STATE(2260),
    [sym_storage_class_specifier] = STATE(1930),
    [sym_type_qualifier] = STATE(1930),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1930),
    [sym_method_parameter] = STATE(1707),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1860),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1930),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4723),
    [anon_sym_SEMI] = ACTIONS(4815),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1732)] = {
    [sym_declaration] = STATE(1875),
    [sym__declaration_modifiers] = STATE(1979),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1979),
    [sym_attribute_declaration] = STATE(1979),
    [sym_ms_declspec_modifier] = STATE(1979),
    [sym_compound_statement] = STATE(2334),
    [sym_storage_class_specifier] = STATE(1979),
    [sym_type_qualifier] = STATE(1979),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1979),
    [sym_method_parameter] = STATE(1691),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1875),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1979),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4691),
    [anon_sym_SEMI] = ACTIONS(4817),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1733)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4781),
    [anon_sym_COMMA] = ACTIONS(4821),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4791),
    [anon_sym_AMP_AMP] = ACTIONS(4793),
    [anon_sym_PIPE] = ACTIONS(4795),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4821),
    [anon_sym___extension__] = ACTIONS(4819),
    [anon_sym_extern] = ACTIONS(4819),
    [anon_sym___attribute__] = ACTIONS(4819),
    [anon_sym___attribute] = ACTIONS(4819),
    [anon_sym_noreturn] = ACTIONS(4819),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4819),
    [anon_sym_LBRACE] = ACTIONS(4821),
    [anon_sym_signed] = ACTIONS(4819),
    [anon_sym_unsigned] = ACTIONS(4819),
    [anon_sym_long] = ACTIONS(4819),
    [anon_sym_short] = ACTIONS(4819),
    [anon_sym_static] = ACTIONS(4819),
    [anon_sym_ATautoreleasepool] = ACTIONS(4821),
    [anon_sym_auto] = ACTIONS(4819),
    [anon_sym_register] = ACTIONS(4819),
    [anon_sym_inline] = ACTIONS(4819),
    [anon_sym___inline] = ACTIONS(4819),
    [anon_sym___inline__] = ACTIONS(4819),
    [anon_sym___forceinline] = ACTIONS(4819),
    [anon_sym_thread_local] = ACTIONS(4819),
    [anon_sym___thread] = ACTIONS(4819),
    [anon_sym_CG_EXTERN] = ACTIONS(4819),
    [anon_sym_CG_INLINE] = ACTIONS(4819),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4819),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4819),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4819),
    [anon_sym_IBOutlet] = ACTIONS(4819),
    [anon_sym_IBInspectable] = ACTIONS(4819),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4819),
    [anon_sym_NS_INLINE] = ACTIONS(4819),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4819),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4819),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4819),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4819),
    [anon_sym_const] = ACTIONS(4819),
    [anon_sym_constexpr] = ACTIONS(4819),
    [anon_sym_volatile] = ACTIONS(4819),
    [anon_sym_restrict] = ACTIONS(4819),
    [anon_sym___restrict__] = ACTIONS(4819),
    [anon_sym__Atomic] = ACTIONS(4819),
    [anon_sym__Noreturn] = ACTIONS(4819),
    [anon_sym__Nonnull] = ACTIONS(4819),
    [anon_sym_nullable] = ACTIONS(4819),
    [anon_sym__Complex] = ACTIONS(4819),
    [anon_sym__Nullable] = ACTIONS(4819),
    [anon_sym__Nullable_result] = ACTIONS(4819),
    [anon_sym__Null_unspecified] = ACTIONS(4819),
    [anon_sym___autoreleasing] = ACTIONS(4819),
    [anon_sym___block] = ACTIONS(4819),
    [anon_sym___bridge] = ACTIONS(4819),
    [anon_sym___bridge_retained] = ACTIONS(4819),
    [anon_sym___bridge_transfer] = ACTIONS(4819),
    [anon_sym___complex] = ACTIONS(4819),
    [anon_sym___const] = ACTIONS(4819),
    [anon_sym___imag] = ACTIONS(4819),
    [anon_sym___kindof] = ACTIONS(4819),
    [anon_sym___nonnull] = ACTIONS(4819),
    [anon_sym___nullable] = ACTIONS(4819),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4819),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4819),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4819),
    [anon_sym___real] = ACTIONS(4819),
    [anon_sym___strong] = ACTIONS(4819),
    [anon_sym___unsafe_unretained] = ACTIONS(4819),
    [anon_sym___unused] = ACTIONS(4819),
    [anon_sym___weak] = ACTIONS(4819),
    [anon_sym_alignas] = ACTIONS(4819),
    [anon_sym__Alignas] = ACTIONS(4819),
    [sym_primitive_type] = ACTIONS(4819),
    [anon_sym_enum] = ACTIONS(4819),
    [anon_sym_struct] = ACTIONS(4819),
    [anon_sym_union] = ACTIONS(4819),
    [anon_sym_QMARK] = ACTIONS(4809),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4819),
    [anon_sym___typeof] = ACTIONS(4819),
    [anon_sym_typeof] = ACTIONS(4819),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4819),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4819),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4819),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4819),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4819),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4819),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4819),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4819),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4819),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4819),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4819),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4819),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4819),
    [anon_sym_API_AVAILABLE] = ACTIONS(4819),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4819),
    [anon_sym_API_DEPRECATED] = ACTIONS(4819),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4819),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4819),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4819),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4819),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4819),
    [anon_sym___deprecated_msg] = ACTIONS(4819),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4819),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4819),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4819),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4819),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4819),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4819),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4819),
    [anon_sym_BOOL] = ACTIONS(4819),
    [anon_sym_IMP] = ACTIONS(4819),
    [anon_sym_SEL] = ACTIONS(4819),
    [anon_sym_Class] = ACTIONS(4819),
    [anon_sym_id] = ACTIONS(4819),
  },
  [STATE(1734)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4823),
    [anon_sym_COMMA] = ACTIONS(4436),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4791),
    [anon_sym_AMP_AMP] = ACTIONS(4793),
    [anon_sym_PIPE] = ACTIONS(4795),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4436),
    [anon_sym___extension__] = ACTIONS(4431),
    [anon_sym_extern] = ACTIONS(4431),
    [anon_sym___attribute__] = ACTIONS(4431),
    [anon_sym___attribute] = ACTIONS(4431),
    [anon_sym_noreturn] = ACTIONS(4431),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4431),
    [anon_sym_LBRACE] = ACTIONS(4436),
    [anon_sym_signed] = ACTIONS(4431),
    [anon_sym_unsigned] = ACTIONS(4431),
    [anon_sym_long] = ACTIONS(4431),
    [anon_sym_short] = ACTIONS(4431),
    [anon_sym_static] = ACTIONS(4431),
    [anon_sym_ATautoreleasepool] = ACTIONS(4436),
    [anon_sym_auto] = ACTIONS(4431),
    [anon_sym_register] = ACTIONS(4431),
    [anon_sym_inline] = ACTIONS(4431),
    [anon_sym___inline] = ACTIONS(4431),
    [anon_sym___inline__] = ACTIONS(4431),
    [anon_sym___forceinline] = ACTIONS(4431),
    [anon_sym_thread_local] = ACTIONS(4431),
    [anon_sym___thread] = ACTIONS(4431),
    [anon_sym_CG_EXTERN] = ACTIONS(4431),
    [anon_sym_CG_INLINE] = ACTIONS(4431),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4431),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4431),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4431),
    [anon_sym_IBOutlet] = ACTIONS(4431),
    [anon_sym_IBInspectable] = ACTIONS(4431),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4431),
    [anon_sym_NS_INLINE] = ACTIONS(4431),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4431),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4431),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4431),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4431),
    [anon_sym_const] = ACTIONS(4431),
    [anon_sym_constexpr] = ACTIONS(4431),
    [anon_sym_volatile] = ACTIONS(4431),
    [anon_sym_restrict] = ACTIONS(4431),
    [anon_sym___restrict__] = ACTIONS(4431),
    [anon_sym__Atomic] = ACTIONS(4431),
    [anon_sym__Noreturn] = ACTIONS(4431),
    [anon_sym__Nonnull] = ACTIONS(4431),
    [anon_sym_nullable] = ACTIONS(4431),
    [anon_sym__Complex] = ACTIONS(4431),
    [anon_sym__Nullable] = ACTIONS(4431),
    [anon_sym__Nullable_result] = ACTIONS(4431),
    [anon_sym__Null_unspecified] = ACTIONS(4431),
    [anon_sym___autoreleasing] = ACTIONS(4431),
    [anon_sym___block] = ACTIONS(4431),
    [anon_sym___bridge] = ACTIONS(4431),
    [anon_sym___bridge_retained] = ACTIONS(4431),
    [anon_sym___bridge_transfer] = ACTIONS(4431),
    [anon_sym___complex] = ACTIONS(4431),
    [anon_sym___const] = ACTIONS(4431),
    [anon_sym___imag] = ACTIONS(4431),
    [anon_sym___kindof] = ACTIONS(4431),
    [anon_sym___nonnull] = ACTIONS(4431),
    [anon_sym___nullable] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4431),
    [anon_sym___real] = ACTIONS(4431),
    [anon_sym___strong] = ACTIONS(4431),
    [anon_sym___unsafe_unretained] = ACTIONS(4431),
    [anon_sym___unused] = ACTIONS(4431),
    [anon_sym___weak] = ACTIONS(4431),
    [anon_sym_alignas] = ACTIONS(4431),
    [anon_sym__Alignas] = ACTIONS(4431),
    [sym_primitive_type] = ACTIONS(4431),
    [anon_sym_enum] = ACTIONS(4431),
    [anon_sym_struct] = ACTIONS(4431),
    [anon_sym_union] = ACTIONS(4431),
    [anon_sym_QMARK] = ACTIONS(4436),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4431),
    [anon_sym___typeof] = ACTIONS(4431),
    [anon_sym_typeof] = ACTIONS(4431),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4431),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4431),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4431),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4431),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4431),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4431),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4431),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_API_AVAILABLE] = ACTIONS(4431),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_API_DEPRECATED] = ACTIONS(4431),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4431),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym___deprecated_msg] = ACTIONS(4431),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4431),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4431),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4431),
    [anon_sym_BOOL] = ACTIONS(4431),
    [anon_sym_IMP] = ACTIONS(4431),
    [anon_sym_SEL] = ACTIONS(4431),
    [anon_sym_Class] = ACTIONS(4431),
    [anon_sym_id] = ACTIONS(4431),
  },
  [STATE(1735)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym__abstract_declarator] = STATE(5277),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_parameter_list] = STATE(4271),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_LPAREN2] = ACTIONS(4488),
    [anon_sym_STAR] = ACTIONS(4490),
    [anon_sym_CARET] = ACTIONS(4492),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1736)] = {
    [sym_identifier] = ACTIONS(4826),
    [aux_sym_preproc_def_token1] = ACTIONS(4826),
    [anon_sym_COMMA] = ACTIONS(4828),
    [anon_sym_RPAREN] = ACTIONS(4828),
    [aux_sym_preproc_if_token1] = ACTIONS(4826),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4826),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4826),
    [sym_preproc_directive] = ACTIONS(4826),
    [anon_sym_LPAREN2] = ACTIONS(4828),
    [anon_sym_DASH] = ACTIONS(4828),
    [anon_sym_PLUS] = ACTIONS(4828),
    [anon_sym_STAR] = ACTIONS(4828),
    [anon_sym_CARET] = ACTIONS(4828),
    [anon_sym_GT] = ACTIONS(4828),
    [anon_sym_SEMI] = ACTIONS(4828),
    [anon_sym___extension__] = ACTIONS(4826),
    [anon_sym_typedef] = ACTIONS(4826),
    [anon_sym_extern] = ACTIONS(4826),
    [anon_sym___attribute__] = ACTIONS(4826),
    [anon_sym___attribute] = ACTIONS(4826),
    [anon_sym_noreturn] = ACTIONS(4826),
    [anon_sym_LBRACK] = ACTIONS(4828),
    [anon_sym___declspec] = ACTIONS(4826),
    [anon_sym___based] = ACTIONS(4826),
    [anon_sym___cdecl] = ACTIONS(4826),
    [anon_sym___clrcall] = ACTIONS(4826),
    [anon_sym___stdcall] = ACTIONS(4826),
    [anon_sym___fastcall] = ACTIONS(4826),
    [anon_sym___thiscall] = ACTIONS(4826),
    [anon_sym___vectorcall] = ACTIONS(4826),
    [anon_sym_LBRACE] = ACTIONS(4828),
    [anon_sym_signed] = ACTIONS(4826),
    [anon_sym_unsigned] = ACTIONS(4826),
    [anon_sym_long] = ACTIONS(4826),
    [anon_sym_short] = ACTIONS(4826),
    [anon_sym_static] = ACTIONS(4826),
    [anon_sym_ATautoreleasepool] = ACTIONS(4828),
    [anon_sym_auto] = ACTIONS(4826),
    [anon_sym_register] = ACTIONS(4826),
    [anon_sym_inline] = ACTIONS(4826),
    [anon_sym___inline] = ACTIONS(4826),
    [anon_sym___inline__] = ACTIONS(4826),
    [anon_sym___forceinline] = ACTIONS(4826),
    [anon_sym_thread_local] = ACTIONS(4826),
    [anon_sym___thread] = ACTIONS(4826),
    [anon_sym_CG_EXTERN] = ACTIONS(4826),
    [anon_sym_CG_INLINE] = ACTIONS(4826),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4826),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4826),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4826),
    [anon_sym_IBOutlet] = ACTIONS(4826),
    [anon_sym_IBInspectable] = ACTIONS(4826),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4826),
    [anon_sym_NS_INLINE] = ACTIONS(4826),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4826),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4826),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4826),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4826),
    [anon_sym_const] = ACTIONS(4826),
    [anon_sym_constexpr] = ACTIONS(4826),
    [anon_sym_volatile] = ACTIONS(4826),
    [anon_sym_restrict] = ACTIONS(4826),
    [anon_sym___restrict__] = ACTIONS(4826),
    [anon_sym__Atomic] = ACTIONS(4826),
    [anon_sym__Noreturn] = ACTIONS(4826),
    [anon_sym__Nonnull] = ACTIONS(4826),
    [anon_sym_nullable] = ACTIONS(4826),
    [anon_sym__Complex] = ACTIONS(4826),
    [anon_sym__Nullable] = ACTIONS(4826),
    [anon_sym__Nullable_result] = ACTIONS(4826),
    [anon_sym__Null_unspecified] = ACTIONS(4826),
    [anon_sym___autoreleasing] = ACTIONS(4826),
    [anon_sym___block] = ACTIONS(4826),
    [anon_sym___bridge] = ACTIONS(4826),
    [anon_sym___bridge_retained] = ACTIONS(4826),
    [anon_sym___bridge_transfer] = ACTIONS(4826),
    [anon_sym___complex] = ACTIONS(4826),
    [anon_sym___const] = ACTIONS(4826),
    [anon_sym___imag] = ACTIONS(4826),
    [anon_sym___kindof] = ACTIONS(4826),
    [anon_sym___nonnull] = ACTIONS(4826),
    [anon_sym___nullable] = ACTIONS(4826),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4826),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4826),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4826),
    [anon_sym___real] = ACTIONS(4826),
    [anon_sym___strong] = ACTIONS(4826),
    [anon_sym___unsafe_unretained] = ACTIONS(4826),
    [anon_sym___unused] = ACTIONS(4826),
    [anon_sym___weak] = ACTIONS(4826),
    [anon_sym_alignas] = ACTIONS(4826),
    [anon_sym__Alignas] = ACTIONS(4826),
    [sym_primitive_type] = ACTIONS(4826),
    [anon_sym_enum] = ACTIONS(4826),
    [anon_sym_COLON] = ACTIONS(4828),
    [anon_sym_struct] = ACTIONS(4826),
    [anon_sym_union] = ACTIONS(4826),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4826),
    [anon_sym___typeof] = ACTIONS(4826),
    [anon_sym_typeof] = ACTIONS(4826),
    [aux_sym_preproc_undef_token1] = ACTIONS(4826),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4826),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4826),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4826),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4826),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4826),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4826),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4826),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4826),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4826),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4826),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4826),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4826),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4826),
    [anon_sym_API_AVAILABLE] = ACTIONS(4826),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4826),
    [anon_sym_API_DEPRECATED] = ACTIONS(4826),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4826),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4826),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4826),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4826),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4826),
    [anon_sym___deprecated_msg] = ACTIONS(4826),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4826),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4826),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4826),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4826),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4826),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4826),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4826),
    [anon_sym_ATend] = ACTIONS(4828),
    [anon_sym_AToptional] = ACTIONS(4828),
    [anon_sym_ATrequired] = ACTIONS(4828),
    [anon_sym_ATproperty] = ACTIONS(4828),
    [anon_sym_BOOL] = ACTIONS(4826),
    [anon_sym_IMP] = ACTIONS(4826),
    [anon_sym_SEL] = ACTIONS(4826),
    [anon_sym_Class] = ACTIONS(4826),
    [anon_sym_id] = ACTIONS(4826),
  },
  [STATE(1737)] = {
    [sym_declaration] = STATE(1830),
    [sym__declaration_modifiers] = STATE(1980),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1980),
    [sym_attribute_declaration] = STATE(1980),
    [sym_ms_declspec_modifier] = STATE(1980),
    [sym_compound_statement] = STATE(2297),
    [sym_storage_class_specifier] = STATE(1980),
    [sym_type_qualifier] = STATE(1980),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1980),
    [sym_method_parameter] = STATE(1695),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1830),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1980),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4707),
    [anon_sym_SEMI] = ACTIONS(4830),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1738)] = {
    [sym_declaration] = STATE(1855),
    [sym__declaration_modifiers] = STATE(1982),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1982),
    [sym_attribute_declaration] = STATE(1982),
    [sym_ms_declspec_modifier] = STATE(1982),
    [sym_compound_statement] = STATE(2163),
    [sym_storage_class_specifier] = STATE(1982),
    [sym_type_qualifier] = STATE(1982),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1982),
    [sym_method_parameter] = STATE(1675),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1855),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1982),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4727),
    [anon_sym_SEMI] = ACTIONS(4832),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1739)] = {
    [sym_declaration] = STATE(1873),
    [sym__declaration_modifiers] = STATE(1955),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1955),
    [sym_attribute_declaration] = STATE(1955),
    [sym_ms_declspec_modifier] = STATE(1955),
    [sym_compound_statement] = STATE(2186),
    [sym_storage_class_specifier] = STATE(1955),
    [sym_type_qualifier] = STATE(1955),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1955),
    [sym_method_parameter] = STATE(1678),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1873),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1955),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4663),
    [anon_sym_SEMI] = ACTIONS(4834),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1740)] = {
    [sym_declaration] = STATE(1858),
    [sym__declaration_modifiers] = STATE(1926),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1926),
    [sym_attribute_declaration] = STATE(1926),
    [sym_ms_declspec_modifier] = STATE(1926),
    [sym_compound_statement] = STATE(2252),
    [sym_storage_class_specifier] = STATE(1926),
    [sym_type_qualifier] = STATE(1926),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1926),
    [sym_method_parameter] = STATE(1689),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1858),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1926),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4679),
    [anon_sym_SEMI] = ACTIONS(4836),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1741)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4466),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4781),
    [anon_sym_COMMA] = ACTIONS(4468),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4791),
    [anon_sym_AMP_AMP] = ACTIONS(4793),
    [anon_sym_PIPE] = ACTIONS(4795),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4468),
    [anon_sym___extension__] = ACTIONS(4466),
    [anon_sym_extern] = ACTIONS(4466),
    [anon_sym___attribute__] = ACTIONS(4466),
    [anon_sym___attribute] = ACTIONS(4466),
    [anon_sym_noreturn] = ACTIONS(4466),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4466),
    [anon_sym_LBRACE] = ACTIONS(4468),
    [anon_sym_signed] = ACTIONS(4466),
    [anon_sym_unsigned] = ACTIONS(4466),
    [anon_sym_long] = ACTIONS(4466),
    [anon_sym_short] = ACTIONS(4466),
    [anon_sym_static] = ACTIONS(4466),
    [anon_sym_ATautoreleasepool] = ACTIONS(4468),
    [anon_sym_auto] = ACTIONS(4466),
    [anon_sym_register] = ACTIONS(4466),
    [anon_sym_inline] = ACTIONS(4466),
    [anon_sym___inline] = ACTIONS(4466),
    [anon_sym___inline__] = ACTIONS(4466),
    [anon_sym___forceinline] = ACTIONS(4466),
    [anon_sym_thread_local] = ACTIONS(4466),
    [anon_sym___thread] = ACTIONS(4466),
    [anon_sym_CG_EXTERN] = ACTIONS(4466),
    [anon_sym_CG_INLINE] = ACTIONS(4466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4466),
    [anon_sym_IBOutlet] = ACTIONS(4466),
    [anon_sym_IBInspectable] = ACTIONS(4466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4466),
    [anon_sym_NS_INLINE] = ACTIONS(4466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4466),
    [anon_sym_const] = ACTIONS(4466),
    [anon_sym_constexpr] = ACTIONS(4466),
    [anon_sym_volatile] = ACTIONS(4466),
    [anon_sym_restrict] = ACTIONS(4466),
    [anon_sym___restrict__] = ACTIONS(4466),
    [anon_sym__Atomic] = ACTIONS(4466),
    [anon_sym__Noreturn] = ACTIONS(4466),
    [anon_sym__Nonnull] = ACTIONS(4466),
    [anon_sym_nullable] = ACTIONS(4466),
    [anon_sym__Complex] = ACTIONS(4466),
    [anon_sym__Nullable] = ACTIONS(4466),
    [anon_sym__Nullable_result] = ACTIONS(4466),
    [anon_sym__Null_unspecified] = ACTIONS(4466),
    [anon_sym___autoreleasing] = ACTIONS(4466),
    [anon_sym___block] = ACTIONS(4466),
    [anon_sym___bridge] = ACTIONS(4466),
    [anon_sym___bridge_retained] = ACTIONS(4466),
    [anon_sym___bridge_transfer] = ACTIONS(4466),
    [anon_sym___complex] = ACTIONS(4466),
    [anon_sym___const] = ACTIONS(4466),
    [anon_sym___imag] = ACTIONS(4466),
    [anon_sym___kindof] = ACTIONS(4466),
    [anon_sym___nonnull] = ACTIONS(4466),
    [anon_sym___nullable] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4466),
    [anon_sym___real] = ACTIONS(4466),
    [anon_sym___strong] = ACTIONS(4466),
    [anon_sym___unsafe_unretained] = ACTIONS(4466),
    [anon_sym___unused] = ACTIONS(4466),
    [anon_sym___weak] = ACTIONS(4466),
    [anon_sym_alignas] = ACTIONS(4466),
    [anon_sym__Alignas] = ACTIONS(4466),
    [sym_primitive_type] = ACTIONS(4466),
    [anon_sym_enum] = ACTIONS(4466),
    [anon_sym_struct] = ACTIONS(4466),
    [anon_sym_union] = ACTIONS(4466),
    [anon_sym_QMARK] = ACTIONS(4809),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4466),
    [anon_sym___typeof] = ACTIONS(4466),
    [anon_sym_typeof] = ACTIONS(4466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_API_AVAILABLE] = ACTIONS(4466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_API_DEPRECATED] = ACTIONS(4466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym___deprecated_msg] = ACTIONS(4466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4466),
    [anon_sym_BOOL] = ACTIONS(4466),
    [anon_sym_IMP] = ACTIONS(4466),
    [anon_sym_SEL] = ACTIONS(4466),
    [anon_sym_Class] = ACTIONS(4466),
    [anon_sym_id] = ACTIONS(4466),
  },
  [STATE(1742)] = {
    [sym_declaration] = STATE(1853),
    [sym__declaration_modifiers] = STATE(1954),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1954),
    [sym_attribute_declaration] = STATE(1954),
    [sym_ms_declspec_modifier] = STATE(1954),
    [sym_compound_statement] = STATE(2268),
    [sym_storage_class_specifier] = STATE(1954),
    [sym_type_qualifier] = STATE(1954),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1954),
    [sym_method_parameter] = STATE(1684),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1853),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1954),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4731),
    [anon_sym_SEMI] = ACTIONS(4838),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1743)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4470),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4781),
    [anon_sym_COMMA] = ACTIONS(4472),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4791),
    [anon_sym_AMP_AMP] = ACTIONS(4793),
    [anon_sym_PIPE] = ACTIONS(4795),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4472),
    [anon_sym___extension__] = ACTIONS(4470),
    [anon_sym_extern] = ACTIONS(4470),
    [anon_sym___attribute__] = ACTIONS(4470),
    [anon_sym___attribute] = ACTIONS(4470),
    [anon_sym_noreturn] = ACTIONS(4470),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4470),
    [anon_sym_LBRACE] = ACTIONS(4472),
    [anon_sym_signed] = ACTIONS(4470),
    [anon_sym_unsigned] = ACTIONS(4470),
    [anon_sym_long] = ACTIONS(4470),
    [anon_sym_short] = ACTIONS(4470),
    [anon_sym_static] = ACTIONS(4470),
    [anon_sym_ATautoreleasepool] = ACTIONS(4472),
    [anon_sym_auto] = ACTIONS(4470),
    [anon_sym_register] = ACTIONS(4470),
    [anon_sym_inline] = ACTIONS(4470),
    [anon_sym___inline] = ACTIONS(4470),
    [anon_sym___inline__] = ACTIONS(4470),
    [anon_sym___forceinline] = ACTIONS(4470),
    [anon_sym_thread_local] = ACTIONS(4470),
    [anon_sym___thread] = ACTIONS(4470),
    [anon_sym_CG_EXTERN] = ACTIONS(4470),
    [anon_sym_CG_INLINE] = ACTIONS(4470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4470),
    [anon_sym_IBOutlet] = ACTIONS(4470),
    [anon_sym_IBInspectable] = ACTIONS(4470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4470),
    [anon_sym_NS_INLINE] = ACTIONS(4470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4470),
    [anon_sym_const] = ACTIONS(4470),
    [anon_sym_constexpr] = ACTIONS(4470),
    [anon_sym_volatile] = ACTIONS(4470),
    [anon_sym_restrict] = ACTIONS(4470),
    [anon_sym___restrict__] = ACTIONS(4470),
    [anon_sym__Atomic] = ACTIONS(4470),
    [anon_sym__Noreturn] = ACTIONS(4470),
    [anon_sym__Nonnull] = ACTIONS(4470),
    [anon_sym_nullable] = ACTIONS(4470),
    [anon_sym__Complex] = ACTIONS(4470),
    [anon_sym__Nullable] = ACTIONS(4470),
    [anon_sym__Nullable_result] = ACTIONS(4470),
    [anon_sym__Null_unspecified] = ACTIONS(4470),
    [anon_sym___autoreleasing] = ACTIONS(4470),
    [anon_sym___block] = ACTIONS(4470),
    [anon_sym___bridge] = ACTIONS(4470),
    [anon_sym___bridge_retained] = ACTIONS(4470),
    [anon_sym___bridge_transfer] = ACTIONS(4470),
    [anon_sym___complex] = ACTIONS(4470),
    [anon_sym___const] = ACTIONS(4470),
    [anon_sym___imag] = ACTIONS(4470),
    [anon_sym___kindof] = ACTIONS(4470),
    [anon_sym___nonnull] = ACTIONS(4470),
    [anon_sym___nullable] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4470),
    [anon_sym___real] = ACTIONS(4470),
    [anon_sym___strong] = ACTIONS(4470),
    [anon_sym___unsafe_unretained] = ACTIONS(4470),
    [anon_sym___unused] = ACTIONS(4470),
    [anon_sym___weak] = ACTIONS(4470),
    [anon_sym_alignas] = ACTIONS(4470),
    [anon_sym__Alignas] = ACTIONS(4470),
    [sym_primitive_type] = ACTIONS(4470),
    [anon_sym_enum] = ACTIONS(4470),
    [anon_sym_struct] = ACTIONS(4470),
    [anon_sym_union] = ACTIONS(4470),
    [anon_sym_QMARK] = ACTIONS(4472),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4470),
    [anon_sym___typeof] = ACTIONS(4470),
    [anon_sym_typeof] = ACTIONS(4470),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_API_AVAILABLE] = ACTIONS(4470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_API_DEPRECATED] = ACTIONS(4470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym___deprecated_msg] = ACTIONS(4470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4470),
    [anon_sym_BOOL] = ACTIONS(4470),
    [anon_sym_IMP] = ACTIONS(4470),
    [anon_sym_SEL] = ACTIONS(4470),
    [anon_sym_Class] = ACTIONS(4470),
    [anon_sym_id] = ACTIONS(4470),
  },
  [STATE(1744)] = {
    [sym_declaration] = STATE(1858),
    [sym__declaration_modifiers] = STATE(1926),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1926),
    [sym_attribute_declaration] = STATE(1926),
    [sym_ms_declspec_modifier] = STATE(1926),
    [sym_compound_statement] = STATE(2252),
    [sym_storage_class_specifier] = STATE(1926),
    [sym_type_qualifier] = STATE(1926),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1926),
    [sym_method_parameter] = STATE(1689),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1858),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1926),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4679),
    [anon_sym_SEMI] = ACTIONS(4840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1745)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4793),
    [anon_sym_PIPE] = ACTIONS(4795),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1746)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4795),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1747)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1748)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1749)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1750)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1751)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1752)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4002),
    [anon_sym_GT_GT] = ACTIONS(4002),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1753)] = {
    [sym_declaration] = STATE(1876),
    [sym__declaration_modifiers] = STATE(1932),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1932),
    [sym_attribute_declaration] = STATE(1932),
    [sym_ms_declspec_modifier] = STATE(1932),
    [sym_compound_statement] = STATE(2309),
    [sym_storage_class_specifier] = STATE(1932),
    [sym_type_qualifier] = STATE(1932),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1932),
    [sym_method_parameter] = STATE(1701),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1876),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1932),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4695),
    [anon_sym_SEMI] = ACTIONS(4842),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1754)] = {
    [sym_declaration] = STATE(1834),
    [sym__declaration_modifiers] = STATE(1931),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1931),
    [sym_attribute_declaration] = STATE(1931),
    [sym_ms_declspec_modifier] = STATE(1931),
    [sym_compound_statement] = STATE(2345),
    [sym_storage_class_specifier] = STATE(1931),
    [sym_type_qualifier] = STATE(1931),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1931),
    [sym_method_parameter] = STATE(1698),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1834),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1931),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4711),
    [anon_sym_SEMI] = ACTIONS(4844),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1755)] = {
    [sym_declaration] = STATE(1856),
    [sym__declaration_modifiers] = STATE(1921),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1921),
    [sym_attribute_declaration] = STATE(1921),
    [sym_ms_declspec_modifier] = STATE(1921),
    [sym_compound_statement] = STATE(2257),
    [sym_storage_class_specifier] = STATE(1921),
    [sym_type_qualifier] = STATE(1921),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1921),
    [sym_method_parameter] = STATE(1704),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1921),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4687),
    [anon_sym_SEMI] = ACTIONS(4846),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1756)] = {
    [sym_declaration] = STATE(1851),
    [sym__declaration_modifiers] = STATE(1938),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1938),
    [sym_attribute_declaration] = STATE(1938),
    [sym_ms_declspec_modifier] = STATE(1938),
    [sym_compound_statement] = STATE(2184),
    [sym_storage_class_specifier] = STATE(1938),
    [sym_type_qualifier] = STATE(1938),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1938),
    [sym_method_parameter] = STATE(1703),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1851),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1938),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4703),
    [anon_sym_SEMI] = ACTIONS(4848),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1757)] = {
    [sym_declaration] = STATE(1850),
    [sym__declaration_modifiers] = STATE(1971),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1971),
    [sym_attribute_declaration] = STATE(1971),
    [sym_ms_declspec_modifier] = STATE(1971),
    [sym_compound_statement] = STATE(2101),
    [sym_storage_class_specifier] = STATE(1971),
    [sym_type_qualifier] = STATE(1971),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1971),
    [sym_method_parameter] = STATE(1681),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1850),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1971),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4667),
    [anon_sym_SEMI] = ACTIONS(4850),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1758)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4478),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4781),
    [anon_sym_COMMA] = ACTIONS(4480),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4791),
    [anon_sym_AMP_AMP] = ACTIONS(4793),
    [anon_sym_PIPE] = ACTIONS(4795),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4480),
    [anon_sym___extension__] = ACTIONS(4478),
    [anon_sym_extern] = ACTIONS(4478),
    [anon_sym___attribute__] = ACTIONS(4478),
    [anon_sym___attribute] = ACTIONS(4478),
    [anon_sym_noreturn] = ACTIONS(4478),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4478),
    [anon_sym_LBRACE] = ACTIONS(4480),
    [anon_sym_signed] = ACTIONS(4478),
    [anon_sym_unsigned] = ACTIONS(4478),
    [anon_sym_long] = ACTIONS(4478),
    [anon_sym_short] = ACTIONS(4478),
    [anon_sym_static] = ACTIONS(4478),
    [anon_sym_ATautoreleasepool] = ACTIONS(4480),
    [anon_sym_auto] = ACTIONS(4478),
    [anon_sym_register] = ACTIONS(4478),
    [anon_sym_inline] = ACTIONS(4478),
    [anon_sym___inline] = ACTIONS(4478),
    [anon_sym___inline__] = ACTIONS(4478),
    [anon_sym___forceinline] = ACTIONS(4478),
    [anon_sym_thread_local] = ACTIONS(4478),
    [anon_sym___thread] = ACTIONS(4478),
    [anon_sym_CG_EXTERN] = ACTIONS(4478),
    [anon_sym_CG_INLINE] = ACTIONS(4478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4478),
    [anon_sym_IBOutlet] = ACTIONS(4478),
    [anon_sym_IBInspectable] = ACTIONS(4478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4478),
    [anon_sym_NS_INLINE] = ACTIONS(4478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4478),
    [anon_sym_const] = ACTIONS(4478),
    [anon_sym_constexpr] = ACTIONS(4478),
    [anon_sym_volatile] = ACTIONS(4478),
    [anon_sym_restrict] = ACTIONS(4478),
    [anon_sym___restrict__] = ACTIONS(4478),
    [anon_sym__Atomic] = ACTIONS(4478),
    [anon_sym__Noreturn] = ACTIONS(4478),
    [anon_sym__Nonnull] = ACTIONS(4478),
    [anon_sym_nullable] = ACTIONS(4478),
    [anon_sym__Complex] = ACTIONS(4478),
    [anon_sym__Nullable] = ACTIONS(4478),
    [anon_sym__Nullable_result] = ACTIONS(4478),
    [anon_sym__Null_unspecified] = ACTIONS(4478),
    [anon_sym___autoreleasing] = ACTIONS(4478),
    [anon_sym___block] = ACTIONS(4478),
    [anon_sym___bridge] = ACTIONS(4478),
    [anon_sym___bridge_retained] = ACTIONS(4478),
    [anon_sym___bridge_transfer] = ACTIONS(4478),
    [anon_sym___complex] = ACTIONS(4478),
    [anon_sym___const] = ACTIONS(4478),
    [anon_sym___imag] = ACTIONS(4478),
    [anon_sym___kindof] = ACTIONS(4478),
    [anon_sym___nonnull] = ACTIONS(4478),
    [anon_sym___nullable] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4478),
    [anon_sym___real] = ACTIONS(4478),
    [anon_sym___strong] = ACTIONS(4478),
    [anon_sym___unsafe_unretained] = ACTIONS(4478),
    [anon_sym___unused] = ACTIONS(4478),
    [anon_sym___weak] = ACTIONS(4478),
    [anon_sym_alignas] = ACTIONS(4478),
    [anon_sym__Alignas] = ACTIONS(4478),
    [sym_primitive_type] = ACTIONS(4478),
    [anon_sym_enum] = ACTIONS(4478),
    [anon_sym_struct] = ACTIONS(4478),
    [anon_sym_union] = ACTIONS(4478),
    [anon_sym_QMARK] = ACTIONS(4809),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4478),
    [anon_sym___typeof] = ACTIONS(4478),
    [anon_sym_typeof] = ACTIONS(4478),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_API_AVAILABLE] = ACTIONS(4478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_API_DEPRECATED] = ACTIONS(4478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym___deprecated_msg] = ACTIONS(4478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4478),
    [anon_sym_BOOL] = ACTIONS(4478),
    [anon_sym_IMP] = ACTIONS(4478),
    [anon_sym_SEL] = ACTIONS(4478),
    [anon_sym_Class] = ACTIONS(4478),
    [anon_sym_id] = ACTIONS(4478),
  },
  [STATE(1759)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4458),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4781),
    [anon_sym_COMMA] = ACTIONS(4462),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4791),
    [anon_sym_AMP_AMP] = ACTIONS(4793),
    [anon_sym_PIPE] = ACTIONS(4795),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4462),
    [anon_sym___extension__] = ACTIONS(4458),
    [anon_sym_extern] = ACTIONS(4458),
    [anon_sym___attribute__] = ACTIONS(4458),
    [anon_sym___attribute] = ACTIONS(4458),
    [anon_sym_noreturn] = ACTIONS(4458),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4458),
    [anon_sym_LBRACE] = ACTIONS(4462),
    [anon_sym_signed] = ACTIONS(4458),
    [anon_sym_unsigned] = ACTIONS(4458),
    [anon_sym_long] = ACTIONS(4458),
    [anon_sym_short] = ACTIONS(4458),
    [anon_sym_static] = ACTIONS(4458),
    [anon_sym_ATautoreleasepool] = ACTIONS(4462),
    [anon_sym_auto] = ACTIONS(4458),
    [anon_sym_register] = ACTIONS(4458),
    [anon_sym_inline] = ACTIONS(4458),
    [anon_sym___inline] = ACTIONS(4458),
    [anon_sym___inline__] = ACTIONS(4458),
    [anon_sym___forceinline] = ACTIONS(4458),
    [anon_sym_thread_local] = ACTIONS(4458),
    [anon_sym___thread] = ACTIONS(4458),
    [anon_sym_CG_EXTERN] = ACTIONS(4458),
    [anon_sym_CG_INLINE] = ACTIONS(4458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4458),
    [anon_sym_IBOutlet] = ACTIONS(4458),
    [anon_sym_IBInspectable] = ACTIONS(4458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4458),
    [anon_sym_NS_INLINE] = ACTIONS(4458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4458),
    [anon_sym_const] = ACTIONS(4458),
    [anon_sym_constexpr] = ACTIONS(4458),
    [anon_sym_volatile] = ACTIONS(4458),
    [anon_sym_restrict] = ACTIONS(4458),
    [anon_sym___restrict__] = ACTIONS(4458),
    [anon_sym__Atomic] = ACTIONS(4458),
    [anon_sym__Noreturn] = ACTIONS(4458),
    [anon_sym__Nonnull] = ACTIONS(4458),
    [anon_sym_nullable] = ACTIONS(4458),
    [anon_sym__Complex] = ACTIONS(4458),
    [anon_sym__Nullable] = ACTIONS(4458),
    [anon_sym__Nullable_result] = ACTIONS(4458),
    [anon_sym__Null_unspecified] = ACTIONS(4458),
    [anon_sym___autoreleasing] = ACTIONS(4458),
    [anon_sym___block] = ACTIONS(4458),
    [anon_sym___bridge] = ACTIONS(4458),
    [anon_sym___bridge_retained] = ACTIONS(4458),
    [anon_sym___bridge_transfer] = ACTIONS(4458),
    [anon_sym___complex] = ACTIONS(4458),
    [anon_sym___const] = ACTIONS(4458),
    [anon_sym___imag] = ACTIONS(4458),
    [anon_sym___kindof] = ACTIONS(4458),
    [anon_sym___nonnull] = ACTIONS(4458),
    [anon_sym___nullable] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4458),
    [anon_sym___real] = ACTIONS(4458),
    [anon_sym___strong] = ACTIONS(4458),
    [anon_sym___unsafe_unretained] = ACTIONS(4458),
    [anon_sym___unused] = ACTIONS(4458),
    [anon_sym___weak] = ACTIONS(4458),
    [anon_sym_alignas] = ACTIONS(4458),
    [anon_sym__Alignas] = ACTIONS(4458),
    [sym_primitive_type] = ACTIONS(4458),
    [anon_sym_enum] = ACTIONS(4458),
    [anon_sym_struct] = ACTIONS(4458),
    [anon_sym_union] = ACTIONS(4458),
    [anon_sym_QMARK] = ACTIONS(4809),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4458),
    [anon_sym___typeof] = ACTIONS(4458),
    [anon_sym_typeof] = ACTIONS(4458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_API_AVAILABLE] = ACTIONS(4458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_API_DEPRECATED] = ACTIONS(4458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym___deprecated_msg] = ACTIONS(4458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4458),
    [anon_sym_BOOL] = ACTIONS(4458),
    [anon_sym_IMP] = ACTIONS(4458),
    [anon_sym_SEL] = ACTIONS(4458),
    [anon_sym_Class] = ACTIONS(4458),
    [anon_sym_id] = ACTIONS(4458),
  },
  [STATE(1760)] = {
    [sym_declaration] = STATE(1864),
    [sym__declaration_modifiers] = STATE(1939),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1939),
    [sym_attribute_declaration] = STATE(1939),
    [sym_ms_declspec_modifier] = STATE(1939),
    [sym_compound_statement] = STATE(2281),
    [sym_storage_class_specifier] = STATE(1939),
    [sym_type_qualifier] = STATE(1939),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1939),
    [sym_method_parameter] = STATE(1686),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1939),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4645),
    [anon_sym_SEMI] = ACTIONS(4852),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1761)] = {
    [sym_declaration] = STATE(1862),
    [sym__declaration_modifiers] = STATE(1933),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1933),
    [sym_attribute_declaration] = STATE(1933),
    [sym_ms_declspec_modifier] = STATE(1933),
    [sym_compound_statement] = STATE(2140),
    [sym_storage_class_specifier] = STATE(1933),
    [sym_type_qualifier] = STATE(1933),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1933),
    [sym_method_parameter] = STATE(1677),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1862),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1933),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4655),
    [anon_sym_SEMI] = ACTIONS(4854),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1762)] = {
    [sym_identifier] = ACTIONS(4856),
    [aux_sym_preproc_def_token1] = ACTIONS(4856),
    [anon_sym_COMMA] = ACTIONS(4858),
    [anon_sym_RPAREN] = ACTIONS(4858),
    [aux_sym_preproc_if_token1] = ACTIONS(4856),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4856),
    [sym_preproc_directive] = ACTIONS(4856),
    [anon_sym_LPAREN2] = ACTIONS(4858),
    [anon_sym_DASH] = ACTIONS(4858),
    [anon_sym_PLUS] = ACTIONS(4858),
    [anon_sym_STAR] = ACTIONS(4858),
    [anon_sym_CARET] = ACTIONS(4858),
    [anon_sym_GT] = ACTIONS(4858),
    [anon_sym_SEMI] = ACTIONS(4858),
    [anon_sym___extension__] = ACTIONS(4856),
    [anon_sym_typedef] = ACTIONS(4856),
    [anon_sym_extern] = ACTIONS(4856),
    [anon_sym___attribute__] = ACTIONS(4856),
    [anon_sym___attribute] = ACTIONS(4856),
    [anon_sym_noreturn] = ACTIONS(4856),
    [anon_sym_LBRACK] = ACTIONS(4858),
    [anon_sym___declspec] = ACTIONS(4856),
    [anon_sym___based] = ACTIONS(4856),
    [anon_sym___cdecl] = ACTIONS(4856),
    [anon_sym___clrcall] = ACTIONS(4856),
    [anon_sym___stdcall] = ACTIONS(4856),
    [anon_sym___fastcall] = ACTIONS(4856),
    [anon_sym___thiscall] = ACTIONS(4856),
    [anon_sym___vectorcall] = ACTIONS(4856),
    [anon_sym_LBRACE] = ACTIONS(4858),
    [anon_sym_signed] = ACTIONS(4856),
    [anon_sym_unsigned] = ACTIONS(4856),
    [anon_sym_long] = ACTIONS(4856),
    [anon_sym_short] = ACTIONS(4856),
    [anon_sym_static] = ACTIONS(4856),
    [anon_sym_ATautoreleasepool] = ACTIONS(4858),
    [anon_sym_auto] = ACTIONS(4856),
    [anon_sym_register] = ACTIONS(4856),
    [anon_sym_inline] = ACTIONS(4856),
    [anon_sym___inline] = ACTIONS(4856),
    [anon_sym___inline__] = ACTIONS(4856),
    [anon_sym___forceinline] = ACTIONS(4856),
    [anon_sym_thread_local] = ACTIONS(4856),
    [anon_sym___thread] = ACTIONS(4856),
    [anon_sym_CG_EXTERN] = ACTIONS(4856),
    [anon_sym_CG_INLINE] = ACTIONS(4856),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4856),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4856),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4856),
    [anon_sym_IBOutlet] = ACTIONS(4856),
    [anon_sym_IBInspectable] = ACTIONS(4856),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4856),
    [anon_sym_NS_INLINE] = ACTIONS(4856),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4856),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4856),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4856),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4856),
    [anon_sym_const] = ACTIONS(4856),
    [anon_sym_constexpr] = ACTIONS(4856),
    [anon_sym_volatile] = ACTIONS(4856),
    [anon_sym_restrict] = ACTIONS(4856),
    [anon_sym___restrict__] = ACTIONS(4856),
    [anon_sym__Atomic] = ACTIONS(4856),
    [anon_sym__Noreturn] = ACTIONS(4856),
    [anon_sym__Nonnull] = ACTIONS(4856),
    [anon_sym_nullable] = ACTIONS(4856),
    [anon_sym__Complex] = ACTIONS(4856),
    [anon_sym__Nullable] = ACTIONS(4856),
    [anon_sym__Nullable_result] = ACTIONS(4856),
    [anon_sym__Null_unspecified] = ACTIONS(4856),
    [anon_sym___autoreleasing] = ACTIONS(4856),
    [anon_sym___block] = ACTIONS(4856),
    [anon_sym___bridge] = ACTIONS(4856),
    [anon_sym___bridge_retained] = ACTIONS(4856),
    [anon_sym___bridge_transfer] = ACTIONS(4856),
    [anon_sym___complex] = ACTIONS(4856),
    [anon_sym___const] = ACTIONS(4856),
    [anon_sym___imag] = ACTIONS(4856),
    [anon_sym___kindof] = ACTIONS(4856),
    [anon_sym___nonnull] = ACTIONS(4856),
    [anon_sym___nullable] = ACTIONS(4856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4856),
    [anon_sym___real] = ACTIONS(4856),
    [anon_sym___strong] = ACTIONS(4856),
    [anon_sym___unsafe_unretained] = ACTIONS(4856),
    [anon_sym___unused] = ACTIONS(4856),
    [anon_sym___weak] = ACTIONS(4856),
    [anon_sym_alignas] = ACTIONS(4856),
    [anon_sym__Alignas] = ACTIONS(4856),
    [sym_primitive_type] = ACTIONS(4856),
    [anon_sym_enum] = ACTIONS(4856),
    [anon_sym_COLON] = ACTIONS(4858),
    [anon_sym_struct] = ACTIONS(4856),
    [anon_sym_union] = ACTIONS(4856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4856),
    [anon_sym___typeof] = ACTIONS(4856),
    [anon_sym_typeof] = ACTIONS(4856),
    [aux_sym_preproc_undef_token1] = ACTIONS(4856),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4856),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4856),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4856),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4856),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4856),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4856),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4856),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4856),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4856),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4856),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4856),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4856),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4856),
    [anon_sym_API_AVAILABLE] = ACTIONS(4856),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4856),
    [anon_sym_API_DEPRECATED] = ACTIONS(4856),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4856),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4856),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4856),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4856),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4856),
    [anon_sym___deprecated_msg] = ACTIONS(4856),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4856),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4856),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4856),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4856),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4856),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4856),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4856),
    [anon_sym_ATend] = ACTIONS(4858),
    [anon_sym_AToptional] = ACTIONS(4858),
    [anon_sym_ATrequired] = ACTIONS(4858),
    [anon_sym_ATproperty] = ACTIONS(4858),
    [anon_sym_BOOL] = ACTIONS(4856),
    [anon_sym_IMP] = ACTIONS(4856),
    [anon_sym_SEL] = ACTIONS(4856),
    [anon_sym_Class] = ACTIONS(4856),
    [anon_sym_id] = ACTIONS(4856),
  },
  [STATE(1763)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4474),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4781),
    [anon_sym_COMMA] = ACTIONS(4476),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4785),
    [anon_sym_PLUS] = ACTIONS(4785),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4791),
    [anon_sym_AMP_AMP] = ACTIONS(4793),
    [anon_sym_PIPE] = ACTIONS(4795),
    [anon_sym_CARET] = ACTIONS(4797),
    [anon_sym_AMP] = ACTIONS(4799),
    [anon_sym_EQ_EQ] = ACTIONS(4801),
    [anon_sym_BANG_EQ] = ACTIONS(4801),
    [anon_sym_GT] = ACTIONS(4803),
    [anon_sym_GT_EQ] = ACTIONS(4805),
    [anon_sym_LT_EQ] = ACTIONS(4805),
    [anon_sym_LT] = ACTIONS(4803),
    [anon_sym_LT_LT] = ACTIONS(4807),
    [anon_sym_GT_GT] = ACTIONS(4807),
    [anon_sym_SEMI] = ACTIONS(4476),
    [anon_sym___extension__] = ACTIONS(4474),
    [anon_sym_extern] = ACTIONS(4474),
    [anon_sym___attribute__] = ACTIONS(4474),
    [anon_sym___attribute] = ACTIONS(4474),
    [anon_sym_noreturn] = ACTIONS(4474),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4474),
    [anon_sym_LBRACE] = ACTIONS(4476),
    [anon_sym_signed] = ACTIONS(4474),
    [anon_sym_unsigned] = ACTIONS(4474),
    [anon_sym_long] = ACTIONS(4474),
    [anon_sym_short] = ACTIONS(4474),
    [anon_sym_static] = ACTIONS(4474),
    [anon_sym_ATautoreleasepool] = ACTIONS(4476),
    [anon_sym_auto] = ACTIONS(4474),
    [anon_sym_register] = ACTIONS(4474),
    [anon_sym_inline] = ACTIONS(4474),
    [anon_sym___inline] = ACTIONS(4474),
    [anon_sym___inline__] = ACTIONS(4474),
    [anon_sym___forceinline] = ACTIONS(4474),
    [anon_sym_thread_local] = ACTIONS(4474),
    [anon_sym___thread] = ACTIONS(4474),
    [anon_sym_CG_EXTERN] = ACTIONS(4474),
    [anon_sym_CG_INLINE] = ACTIONS(4474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4474),
    [anon_sym_IBOutlet] = ACTIONS(4474),
    [anon_sym_IBInspectable] = ACTIONS(4474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4474),
    [anon_sym_NS_INLINE] = ACTIONS(4474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4474),
    [anon_sym_const] = ACTIONS(4474),
    [anon_sym_constexpr] = ACTIONS(4474),
    [anon_sym_volatile] = ACTIONS(4474),
    [anon_sym_restrict] = ACTIONS(4474),
    [anon_sym___restrict__] = ACTIONS(4474),
    [anon_sym__Atomic] = ACTIONS(4474),
    [anon_sym__Noreturn] = ACTIONS(4474),
    [anon_sym__Nonnull] = ACTIONS(4474),
    [anon_sym_nullable] = ACTIONS(4474),
    [anon_sym__Complex] = ACTIONS(4474),
    [anon_sym__Nullable] = ACTIONS(4474),
    [anon_sym__Nullable_result] = ACTIONS(4474),
    [anon_sym__Null_unspecified] = ACTIONS(4474),
    [anon_sym___autoreleasing] = ACTIONS(4474),
    [anon_sym___block] = ACTIONS(4474),
    [anon_sym___bridge] = ACTIONS(4474),
    [anon_sym___bridge_retained] = ACTIONS(4474),
    [anon_sym___bridge_transfer] = ACTIONS(4474),
    [anon_sym___complex] = ACTIONS(4474),
    [anon_sym___const] = ACTIONS(4474),
    [anon_sym___imag] = ACTIONS(4474),
    [anon_sym___kindof] = ACTIONS(4474),
    [anon_sym___nonnull] = ACTIONS(4474),
    [anon_sym___nullable] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4474),
    [anon_sym___real] = ACTIONS(4474),
    [anon_sym___strong] = ACTIONS(4474),
    [anon_sym___unsafe_unretained] = ACTIONS(4474),
    [anon_sym___unused] = ACTIONS(4474),
    [anon_sym___weak] = ACTIONS(4474),
    [anon_sym_alignas] = ACTIONS(4474),
    [anon_sym__Alignas] = ACTIONS(4474),
    [sym_primitive_type] = ACTIONS(4474),
    [anon_sym_enum] = ACTIONS(4474),
    [anon_sym_struct] = ACTIONS(4474),
    [anon_sym_union] = ACTIONS(4474),
    [anon_sym_QMARK] = ACTIONS(4476),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4474),
    [anon_sym___typeof] = ACTIONS(4474),
    [anon_sym_typeof] = ACTIONS(4474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_API_AVAILABLE] = ACTIONS(4474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_API_DEPRECATED] = ACTIONS(4474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym___deprecated_msg] = ACTIONS(4474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4474),
    [anon_sym_BOOL] = ACTIONS(4474),
    [anon_sym_IMP] = ACTIONS(4474),
    [anon_sym_SEL] = ACTIONS(4474),
    [anon_sym_Class] = ACTIONS(4474),
    [anon_sym_id] = ACTIONS(4474),
  },
  [STATE(1764)] = {
    [sym_declaration] = STATE(1827),
    [sym__declaration_modifiers] = STATE(1966),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1966),
    [sym_attribute_declaration] = STATE(1966),
    [sym_ms_declspec_modifier] = STATE(1966),
    [sym_compound_statement] = STATE(2262),
    [sym_storage_class_specifier] = STATE(1966),
    [sym_type_qualifier] = STATE(1966),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1966),
    [sym_method_parameter] = STATE(1694),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1827),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1966),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4699),
    [anon_sym_SEMI] = ACTIONS(4860),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_COLON] = ACTIONS(4653),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1765)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_COMMA] = ACTIONS(4044),
    [anon_sym_LPAREN2] = ACTIONS(4044),
    [anon_sym_DASH] = ACTIONS(4042),
    [anon_sym_PLUS] = ACTIONS(4042),
    [anon_sym_STAR] = ACTIONS(4044),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4044),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(4044),
    [anon_sym_AMP] = ACTIONS(4042),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(4042),
    [anon_sym_LT_LT] = ACTIONS(4044),
    [anon_sym_GT_GT] = ACTIONS(4044),
    [anon_sym_SEMI] = ACTIONS(4044),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym___declspec] = ACTIONS(2914),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(4044),
    [anon_sym_PLUS_PLUS] = ACTIONS(4044),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATprotocol] = ACTIONS(2920),
    [anon_sym_ATinterface] = ACTIONS(2920),
    [anon_sym_ATimplementation] = ACTIONS(2920),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(1766)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4000),
    [anon_sym_PLUS] = ACTIONS(4000),
    [anon_sym_STAR] = ACTIONS(4787),
    [anon_sym_SLASH] = ACTIONS(4789),
    [anon_sym_PERCENT] = ACTIONS(4787),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4002),
    [anon_sym_GT_GT] = ACTIONS(4002),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_LBRACE] = ACTIONS(4002),
    [anon_sym_signed] = ACTIONS(4000),
    [anon_sym_unsigned] = ACTIONS(4000),
    [anon_sym_long] = ACTIONS(4000),
    [anon_sym_short] = ACTIONS(4000),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_ATautoreleasepool] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [sym_primitive_type] = ACTIONS(4000),
    [anon_sym_enum] = ACTIONS(4000),
    [anon_sym_struct] = ACTIONS(4000),
    [anon_sym_union] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4000),
    [anon_sym___typeof] = ACTIONS(4000),
    [anon_sym_typeof] = ACTIONS(4000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
    [anon_sym_BOOL] = ACTIONS(4000),
    [anon_sym_IMP] = ACTIONS(4000),
    [anon_sym_SEL] = ACTIONS(4000),
    [anon_sym_Class] = ACTIONS(4000),
    [anon_sym_id] = ACTIONS(4000),
  },
  [STATE(1767)] = {
    [sym_declaration] = STATE(1840),
    [sym__declaration_modifiers] = STATE(1935),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1935),
    [sym_attribute_declaration] = STATE(1935),
    [sym_ms_declspec_modifier] = STATE(1935),
    [sym_compound_statement] = STATE(2312),
    [sym_storage_class_specifier] = STATE(1935),
    [sym_type_qualifier] = STATE(1935),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1935),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1840),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1935),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1813),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4864),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1768)] = {
    [sym_declaration] = STATE(1849),
    [sym__declaration_modifiers] = STATE(1924),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1924),
    [sym_attribute_declaration] = STATE(1924),
    [sym_ms_declspec_modifier] = STATE(1924),
    [sym_compound_statement] = STATE(2109),
    [sym_storage_class_specifier] = STATE(1924),
    [sym_type_qualifier] = STATE(1924),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1924),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1849),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1924),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4866),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1769)] = {
    [sym_declaration] = STATE(1850),
    [sym__declaration_modifiers] = STATE(1971),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1971),
    [sym_attribute_declaration] = STATE(1971),
    [sym_ms_declspec_modifier] = STATE(1971),
    [sym_compound_statement] = STATE(2101),
    [sym_storage_class_specifier] = STATE(1971),
    [sym_type_qualifier] = STATE(1971),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1850),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1971),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4850),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1770)] = {
    [sym_declaration] = STATE(1860),
    [sym__declaration_modifiers] = STATE(1930),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1930),
    [sym_attribute_declaration] = STATE(1930),
    [sym_ms_declspec_modifier] = STATE(1930),
    [sym_compound_statement] = STATE(2260),
    [sym_storage_class_specifier] = STATE(1930),
    [sym_type_qualifier] = STATE(1930),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1930),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1860),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1930),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4815),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1771)] = {
    [sym_declaration] = STATE(1860),
    [sym__declaration_modifiers] = STATE(1930),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1930),
    [sym_attribute_declaration] = STATE(1930),
    [sym_ms_declspec_modifier] = STATE(1930),
    [sym_compound_statement] = STATE(2260),
    [sym_storage_class_specifier] = STATE(1930),
    [sym_type_qualifier] = STATE(1930),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1930),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1860),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1930),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1796),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4815),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1772)] = {
    [sym_declaration] = STATE(1831),
    [sym__declaration_modifiers] = STATE(1972),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1972),
    [sym_attribute_declaration] = STATE(1972),
    [sym_ms_declspec_modifier] = STATE(1972),
    [sym_compound_statement] = STATE(2104),
    [sym_storage_class_specifier] = STATE(1972),
    [sym_type_qualifier] = STATE(1972),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1972),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1831),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1972),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4868),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1773)] = {
    [sym_declaration] = STATE(1831),
    [sym__declaration_modifiers] = STATE(1972),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1972),
    [sym_attribute_declaration] = STATE(1972),
    [sym_ms_declspec_modifier] = STATE(1972),
    [sym_compound_statement] = STATE(2104),
    [sym_storage_class_specifier] = STATE(1972),
    [sym_type_qualifier] = STATE(1972),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1972),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1831),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1972),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1782),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4868),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1774)] = {
    [sym_declaration] = STATE(1850),
    [sym__declaration_modifiers] = STATE(1971),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1971),
    [sym_attribute_declaration] = STATE(1971),
    [sym_ms_declspec_modifier] = STATE(1971),
    [sym_compound_statement] = STATE(2101),
    [sym_storage_class_specifier] = STATE(1971),
    [sym_type_qualifier] = STATE(1971),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1850),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1971),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1772),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4850),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1775)] = {
    [sym_identifier] = ACTIONS(4168),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4170),
    [anon_sym_COMMA] = ACTIONS(4170),
    [anon_sym_LPAREN2] = ACTIONS(4170),
    [anon_sym_DASH] = ACTIONS(4168),
    [anon_sym_PLUS] = ACTIONS(4168),
    [anon_sym_STAR] = ACTIONS(4170),
    [anon_sym_SLASH] = ACTIONS(4168),
    [anon_sym_PERCENT] = ACTIONS(4170),
    [anon_sym_PIPE_PIPE] = ACTIONS(4170),
    [anon_sym_AMP_AMP] = ACTIONS(4170),
    [anon_sym_PIPE] = ACTIONS(4168),
    [anon_sym_CARET] = ACTIONS(4170),
    [anon_sym_AMP] = ACTIONS(4168),
    [anon_sym_EQ_EQ] = ACTIONS(4170),
    [anon_sym_BANG_EQ] = ACTIONS(4170),
    [anon_sym_GT] = ACTIONS(4168),
    [anon_sym_GT_EQ] = ACTIONS(4170),
    [anon_sym_LT_EQ] = ACTIONS(4170),
    [anon_sym_LT] = ACTIONS(4168),
    [anon_sym_LT_LT] = ACTIONS(4170),
    [anon_sym_GT_GT] = ACTIONS(4170),
    [anon_sym_SEMI] = ACTIONS(4170),
    [anon_sym___extension__] = ACTIONS(4168),
    [anon_sym_extern] = ACTIONS(4168),
    [anon_sym___attribute__] = ACTIONS(4168),
    [anon_sym___attribute] = ACTIONS(4168),
    [anon_sym_noreturn] = ACTIONS(4168),
    [anon_sym_LBRACK] = ACTIONS(4170),
    [anon_sym___declspec] = ACTIONS(4168),
    [anon_sym_LBRACE] = ACTIONS(4170),
    [anon_sym_signed] = ACTIONS(4168),
    [anon_sym_unsigned] = ACTIONS(4168),
    [anon_sym_long] = ACTIONS(4168),
    [anon_sym_short] = ACTIONS(4168),
    [anon_sym_static] = ACTIONS(4168),
    [anon_sym_ATautoreleasepool] = ACTIONS(4170),
    [anon_sym_auto] = ACTIONS(4168),
    [anon_sym_register] = ACTIONS(4168),
    [anon_sym_inline] = ACTIONS(4168),
    [anon_sym___inline] = ACTIONS(4168),
    [anon_sym___inline__] = ACTIONS(4168),
    [anon_sym___forceinline] = ACTIONS(4168),
    [anon_sym_thread_local] = ACTIONS(4168),
    [anon_sym___thread] = ACTIONS(4168),
    [anon_sym_CG_EXTERN] = ACTIONS(4168),
    [anon_sym_CG_INLINE] = ACTIONS(4168),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4168),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4168),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4168),
    [anon_sym_IBOutlet] = ACTIONS(4168),
    [anon_sym_IBInspectable] = ACTIONS(4168),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4168),
    [anon_sym_NS_INLINE] = ACTIONS(4168),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4168),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4168),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4168),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4168),
    [anon_sym_const] = ACTIONS(4168),
    [anon_sym_constexpr] = ACTIONS(4168),
    [anon_sym_volatile] = ACTIONS(4168),
    [anon_sym_restrict] = ACTIONS(4168),
    [anon_sym___restrict__] = ACTIONS(4168),
    [anon_sym__Atomic] = ACTIONS(4168),
    [anon_sym__Noreturn] = ACTIONS(4168),
    [anon_sym__Nonnull] = ACTIONS(4168),
    [anon_sym_nullable] = ACTIONS(4168),
    [anon_sym__Complex] = ACTIONS(4168),
    [anon_sym__Nullable] = ACTIONS(4168),
    [anon_sym__Nullable_result] = ACTIONS(4168),
    [anon_sym__Null_unspecified] = ACTIONS(4168),
    [anon_sym___autoreleasing] = ACTIONS(4168),
    [anon_sym___block] = ACTIONS(4168),
    [anon_sym___bridge] = ACTIONS(4168),
    [anon_sym___bridge_retained] = ACTIONS(4168),
    [anon_sym___bridge_transfer] = ACTIONS(4168),
    [anon_sym___complex] = ACTIONS(4168),
    [anon_sym___const] = ACTIONS(4168),
    [anon_sym___imag] = ACTIONS(4168),
    [anon_sym___kindof] = ACTIONS(4168),
    [anon_sym___nonnull] = ACTIONS(4168),
    [anon_sym___nullable] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4168),
    [anon_sym___real] = ACTIONS(4168),
    [anon_sym___strong] = ACTIONS(4168),
    [anon_sym___unsafe_unretained] = ACTIONS(4168),
    [anon_sym___unused] = ACTIONS(4168),
    [anon_sym___weak] = ACTIONS(4168),
    [anon_sym_alignas] = ACTIONS(4168),
    [anon_sym__Alignas] = ACTIONS(4168),
    [sym_primitive_type] = ACTIONS(4168),
    [anon_sym_enum] = ACTIONS(4168),
    [anon_sym_struct] = ACTIONS(4168),
    [anon_sym_union] = ACTIONS(4168),
    [anon_sym_QMARK] = ACTIONS(4170),
    [anon_sym_DASH_DASH] = ACTIONS(4170),
    [anon_sym_PLUS_PLUS] = ACTIONS(4170),
    [anon_sym_DOT] = ACTIONS(4168),
    [anon_sym_DASH_GT] = ACTIONS(4170),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4168),
    [anon_sym___typeof] = ACTIONS(4168),
    [anon_sym_typeof] = ACTIONS(4168),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4168),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4168),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4168),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4168),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4168),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4168),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4168),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4168),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4168),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_API_AVAILABLE] = ACTIONS(4168),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_API_DEPRECATED] = ACTIONS(4168),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4168),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4168),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4168),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4168),
    [anon_sym___deprecated_msg] = ACTIONS(4168),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4168),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4168),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4168),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4168),
    [anon_sym_BOOL] = ACTIONS(4168),
    [anon_sym_IMP] = ACTIONS(4168),
    [anon_sym_SEL] = ACTIONS(4168),
    [anon_sym_Class] = ACTIONS(4168),
    [anon_sym_id] = ACTIONS(4168),
  },
  [STATE(1776)] = {
    [sym_identifier] = ACTIONS(4054),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4056),
    [anon_sym_COMMA] = ACTIONS(4056),
    [anon_sym_LPAREN2] = ACTIONS(4056),
    [anon_sym_DASH] = ACTIONS(4054),
    [anon_sym_PLUS] = ACTIONS(4054),
    [anon_sym_STAR] = ACTIONS(4056),
    [anon_sym_SLASH] = ACTIONS(4054),
    [anon_sym_PERCENT] = ACTIONS(4056),
    [anon_sym_PIPE_PIPE] = ACTIONS(4056),
    [anon_sym_AMP_AMP] = ACTIONS(4056),
    [anon_sym_PIPE] = ACTIONS(4054),
    [anon_sym_CARET] = ACTIONS(4056),
    [anon_sym_AMP] = ACTIONS(4054),
    [anon_sym_EQ_EQ] = ACTIONS(4056),
    [anon_sym_BANG_EQ] = ACTIONS(4056),
    [anon_sym_GT] = ACTIONS(4054),
    [anon_sym_GT_EQ] = ACTIONS(4056),
    [anon_sym_LT_EQ] = ACTIONS(4056),
    [anon_sym_LT] = ACTIONS(4054),
    [anon_sym_LT_LT] = ACTIONS(4056),
    [anon_sym_GT_GT] = ACTIONS(4056),
    [anon_sym_SEMI] = ACTIONS(4056),
    [anon_sym___extension__] = ACTIONS(4054),
    [anon_sym_extern] = ACTIONS(4054),
    [anon_sym___attribute__] = ACTIONS(4054),
    [anon_sym___attribute] = ACTIONS(4054),
    [anon_sym_noreturn] = ACTIONS(4054),
    [anon_sym_LBRACK] = ACTIONS(4056),
    [anon_sym___declspec] = ACTIONS(4054),
    [anon_sym_LBRACE] = ACTIONS(4056),
    [anon_sym_signed] = ACTIONS(4054),
    [anon_sym_unsigned] = ACTIONS(4054),
    [anon_sym_long] = ACTIONS(4054),
    [anon_sym_short] = ACTIONS(4054),
    [anon_sym_static] = ACTIONS(4054),
    [anon_sym_ATautoreleasepool] = ACTIONS(4056),
    [anon_sym_auto] = ACTIONS(4054),
    [anon_sym_register] = ACTIONS(4054),
    [anon_sym_inline] = ACTIONS(4054),
    [anon_sym___inline] = ACTIONS(4054),
    [anon_sym___inline__] = ACTIONS(4054),
    [anon_sym___forceinline] = ACTIONS(4054),
    [anon_sym_thread_local] = ACTIONS(4054),
    [anon_sym___thread] = ACTIONS(4054),
    [anon_sym_CG_EXTERN] = ACTIONS(4054),
    [anon_sym_CG_INLINE] = ACTIONS(4054),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4054),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4054),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4054),
    [anon_sym_IBOutlet] = ACTIONS(4054),
    [anon_sym_IBInspectable] = ACTIONS(4054),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4054),
    [anon_sym_NS_INLINE] = ACTIONS(4054),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4054),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4054),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4054),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4054),
    [anon_sym_const] = ACTIONS(4054),
    [anon_sym_constexpr] = ACTIONS(4054),
    [anon_sym_volatile] = ACTIONS(4054),
    [anon_sym_restrict] = ACTIONS(4054),
    [anon_sym___restrict__] = ACTIONS(4054),
    [anon_sym__Atomic] = ACTIONS(4054),
    [anon_sym__Noreturn] = ACTIONS(4054),
    [anon_sym__Nonnull] = ACTIONS(4054),
    [anon_sym_nullable] = ACTIONS(4054),
    [anon_sym__Complex] = ACTIONS(4054),
    [anon_sym__Nullable] = ACTIONS(4054),
    [anon_sym__Nullable_result] = ACTIONS(4054),
    [anon_sym__Null_unspecified] = ACTIONS(4054),
    [anon_sym___autoreleasing] = ACTIONS(4054),
    [anon_sym___block] = ACTIONS(4054),
    [anon_sym___bridge] = ACTIONS(4054),
    [anon_sym___bridge_retained] = ACTIONS(4054),
    [anon_sym___bridge_transfer] = ACTIONS(4054),
    [anon_sym___complex] = ACTIONS(4054),
    [anon_sym___const] = ACTIONS(4054),
    [anon_sym___imag] = ACTIONS(4054),
    [anon_sym___kindof] = ACTIONS(4054),
    [anon_sym___nonnull] = ACTIONS(4054),
    [anon_sym___nullable] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4054),
    [anon_sym___real] = ACTIONS(4054),
    [anon_sym___strong] = ACTIONS(4054),
    [anon_sym___unsafe_unretained] = ACTIONS(4054),
    [anon_sym___unused] = ACTIONS(4054),
    [anon_sym___weak] = ACTIONS(4054),
    [anon_sym_alignas] = ACTIONS(4054),
    [anon_sym__Alignas] = ACTIONS(4054),
    [sym_primitive_type] = ACTIONS(4054),
    [anon_sym_enum] = ACTIONS(4054),
    [anon_sym_struct] = ACTIONS(4054),
    [anon_sym_union] = ACTIONS(4054),
    [anon_sym_QMARK] = ACTIONS(4056),
    [anon_sym_DASH_DASH] = ACTIONS(4056),
    [anon_sym_PLUS_PLUS] = ACTIONS(4056),
    [anon_sym_DOT] = ACTIONS(4054),
    [anon_sym_DASH_GT] = ACTIONS(4056),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4054),
    [anon_sym___typeof] = ACTIONS(4054),
    [anon_sym_typeof] = ACTIONS(4054),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4054),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4054),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4054),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4054),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4054),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4054),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4054),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4054),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4054),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_API_AVAILABLE] = ACTIONS(4054),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_API_DEPRECATED] = ACTIONS(4054),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4054),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4054),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4054),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4054),
    [anon_sym___deprecated_msg] = ACTIONS(4054),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4054),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4054),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4054),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4054),
    [anon_sym_BOOL] = ACTIONS(4054),
    [anon_sym_IMP] = ACTIONS(4054),
    [anon_sym_SEL] = ACTIONS(4054),
    [anon_sym_Class] = ACTIONS(4054),
    [anon_sym_id] = ACTIONS(4054),
  },
  [STATE(1777)] = {
    [sym_identifier] = ACTIONS(4266),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4268),
    [anon_sym_COMMA] = ACTIONS(4268),
    [anon_sym_LPAREN2] = ACTIONS(4268),
    [anon_sym_DASH] = ACTIONS(4266),
    [anon_sym_PLUS] = ACTIONS(4266),
    [anon_sym_STAR] = ACTIONS(4268),
    [anon_sym_SLASH] = ACTIONS(4266),
    [anon_sym_PERCENT] = ACTIONS(4268),
    [anon_sym_PIPE_PIPE] = ACTIONS(4268),
    [anon_sym_AMP_AMP] = ACTIONS(4268),
    [anon_sym_PIPE] = ACTIONS(4266),
    [anon_sym_CARET] = ACTIONS(4268),
    [anon_sym_AMP] = ACTIONS(4266),
    [anon_sym_EQ_EQ] = ACTIONS(4268),
    [anon_sym_BANG_EQ] = ACTIONS(4268),
    [anon_sym_GT] = ACTIONS(4266),
    [anon_sym_GT_EQ] = ACTIONS(4268),
    [anon_sym_LT_EQ] = ACTIONS(4268),
    [anon_sym_LT] = ACTIONS(4266),
    [anon_sym_LT_LT] = ACTIONS(4268),
    [anon_sym_GT_GT] = ACTIONS(4268),
    [anon_sym_SEMI] = ACTIONS(4268),
    [anon_sym___extension__] = ACTIONS(4266),
    [anon_sym_extern] = ACTIONS(4266),
    [anon_sym___attribute__] = ACTIONS(4266),
    [anon_sym___attribute] = ACTIONS(4266),
    [anon_sym_noreturn] = ACTIONS(4266),
    [anon_sym_LBRACK] = ACTIONS(4268),
    [anon_sym___declspec] = ACTIONS(4266),
    [anon_sym_LBRACE] = ACTIONS(4268),
    [anon_sym_signed] = ACTIONS(4266),
    [anon_sym_unsigned] = ACTIONS(4266),
    [anon_sym_long] = ACTIONS(4266),
    [anon_sym_short] = ACTIONS(4266),
    [anon_sym_static] = ACTIONS(4266),
    [anon_sym_ATautoreleasepool] = ACTIONS(4268),
    [anon_sym_auto] = ACTIONS(4266),
    [anon_sym_register] = ACTIONS(4266),
    [anon_sym_inline] = ACTIONS(4266),
    [anon_sym___inline] = ACTIONS(4266),
    [anon_sym___inline__] = ACTIONS(4266),
    [anon_sym___forceinline] = ACTIONS(4266),
    [anon_sym_thread_local] = ACTIONS(4266),
    [anon_sym___thread] = ACTIONS(4266),
    [anon_sym_CG_EXTERN] = ACTIONS(4266),
    [anon_sym_CG_INLINE] = ACTIONS(4266),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4266),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4266),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4266),
    [anon_sym_IBOutlet] = ACTIONS(4266),
    [anon_sym_IBInspectable] = ACTIONS(4266),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4266),
    [anon_sym_NS_INLINE] = ACTIONS(4266),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4266),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4266),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4266),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4266),
    [anon_sym_const] = ACTIONS(4266),
    [anon_sym_constexpr] = ACTIONS(4266),
    [anon_sym_volatile] = ACTIONS(4266),
    [anon_sym_restrict] = ACTIONS(4266),
    [anon_sym___restrict__] = ACTIONS(4266),
    [anon_sym__Atomic] = ACTIONS(4266),
    [anon_sym__Noreturn] = ACTIONS(4266),
    [anon_sym__Nonnull] = ACTIONS(4266),
    [anon_sym_nullable] = ACTIONS(4266),
    [anon_sym__Complex] = ACTIONS(4266),
    [anon_sym__Nullable] = ACTIONS(4266),
    [anon_sym__Nullable_result] = ACTIONS(4266),
    [anon_sym__Null_unspecified] = ACTIONS(4266),
    [anon_sym___autoreleasing] = ACTIONS(4266),
    [anon_sym___block] = ACTIONS(4266),
    [anon_sym___bridge] = ACTIONS(4266),
    [anon_sym___bridge_retained] = ACTIONS(4266),
    [anon_sym___bridge_transfer] = ACTIONS(4266),
    [anon_sym___complex] = ACTIONS(4266),
    [anon_sym___const] = ACTIONS(4266),
    [anon_sym___imag] = ACTIONS(4266),
    [anon_sym___kindof] = ACTIONS(4266),
    [anon_sym___nonnull] = ACTIONS(4266),
    [anon_sym___nullable] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4266),
    [anon_sym___real] = ACTIONS(4266),
    [anon_sym___strong] = ACTIONS(4266),
    [anon_sym___unsafe_unretained] = ACTIONS(4266),
    [anon_sym___unused] = ACTIONS(4266),
    [anon_sym___weak] = ACTIONS(4266),
    [anon_sym_alignas] = ACTIONS(4266),
    [anon_sym__Alignas] = ACTIONS(4266),
    [sym_primitive_type] = ACTIONS(4266),
    [anon_sym_enum] = ACTIONS(4266),
    [anon_sym_struct] = ACTIONS(4266),
    [anon_sym_union] = ACTIONS(4266),
    [anon_sym_QMARK] = ACTIONS(4268),
    [anon_sym_DASH_DASH] = ACTIONS(4268),
    [anon_sym_PLUS_PLUS] = ACTIONS(4268),
    [anon_sym_DOT] = ACTIONS(4266),
    [anon_sym_DASH_GT] = ACTIONS(4268),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4266),
    [anon_sym___typeof] = ACTIONS(4266),
    [anon_sym_typeof] = ACTIONS(4266),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4266),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4266),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4266),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4266),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4266),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4266),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4266),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4266),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4266),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_API_AVAILABLE] = ACTIONS(4266),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_API_DEPRECATED] = ACTIONS(4266),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4266),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4266),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4266),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4266),
    [anon_sym___deprecated_msg] = ACTIONS(4266),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4266),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4266),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4266),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4266),
    [anon_sym_BOOL] = ACTIONS(4266),
    [anon_sym_IMP] = ACTIONS(4266),
    [anon_sym_SEL] = ACTIONS(4266),
    [anon_sym_Class] = ACTIONS(4266),
    [anon_sym_id] = ACTIONS(4266),
  },
  [STATE(1778)] = {
    [sym_identifier] = ACTIONS(4274),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4276),
    [anon_sym_COMMA] = ACTIONS(4276),
    [anon_sym_LPAREN2] = ACTIONS(4276),
    [anon_sym_DASH] = ACTIONS(4274),
    [anon_sym_PLUS] = ACTIONS(4274),
    [anon_sym_STAR] = ACTIONS(4276),
    [anon_sym_SLASH] = ACTIONS(4274),
    [anon_sym_PERCENT] = ACTIONS(4276),
    [anon_sym_PIPE_PIPE] = ACTIONS(4276),
    [anon_sym_AMP_AMP] = ACTIONS(4276),
    [anon_sym_PIPE] = ACTIONS(4274),
    [anon_sym_CARET] = ACTIONS(4276),
    [anon_sym_AMP] = ACTIONS(4274),
    [anon_sym_EQ_EQ] = ACTIONS(4276),
    [anon_sym_BANG_EQ] = ACTIONS(4276),
    [anon_sym_GT] = ACTIONS(4274),
    [anon_sym_GT_EQ] = ACTIONS(4276),
    [anon_sym_LT_EQ] = ACTIONS(4276),
    [anon_sym_LT] = ACTIONS(4274),
    [anon_sym_LT_LT] = ACTIONS(4276),
    [anon_sym_GT_GT] = ACTIONS(4276),
    [anon_sym_SEMI] = ACTIONS(4276),
    [anon_sym___extension__] = ACTIONS(4274),
    [anon_sym_extern] = ACTIONS(4274),
    [anon_sym___attribute__] = ACTIONS(4274),
    [anon_sym___attribute] = ACTIONS(4274),
    [anon_sym_noreturn] = ACTIONS(4274),
    [anon_sym_LBRACK] = ACTIONS(4276),
    [anon_sym___declspec] = ACTIONS(4274),
    [anon_sym_LBRACE] = ACTIONS(4276),
    [anon_sym_signed] = ACTIONS(4274),
    [anon_sym_unsigned] = ACTIONS(4274),
    [anon_sym_long] = ACTIONS(4274),
    [anon_sym_short] = ACTIONS(4274),
    [anon_sym_static] = ACTIONS(4274),
    [anon_sym_ATautoreleasepool] = ACTIONS(4276),
    [anon_sym_auto] = ACTIONS(4274),
    [anon_sym_register] = ACTIONS(4274),
    [anon_sym_inline] = ACTIONS(4274),
    [anon_sym___inline] = ACTIONS(4274),
    [anon_sym___inline__] = ACTIONS(4274),
    [anon_sym___forceinline] = ACTIONS(4274),
    [anon_sym_thread_local] = ACTIONS(4274),
    [anon_sym___thread] = ACTIONS(4274),
    [anon_sym_CG_EXTERN] = ACTIONS(4274),
    [anon_sym_CG_INLINE] = ACTIONS(4274),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4274),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4274),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4274),
    [anon_sym_IBOutlet] = ACTIONS(4274),
    [anon_sym_IBInspectable] = ACTIONS(4274),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4274),
    [anon_sym_NS_INLINE] = ACTIONS(4274),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4274),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4274),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4274),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4274),
    [anon_sym_const] = ACTIONS(4274),
    [anon_sym_constexpr] = ACTIONS(4274),
    [anon_sym_volatile] = ACTIONS(4274),
    [anon_sym_restrict] = ACTIONS(4274),
    [anon_sym___restrict__] = ACTIONS(4274),
    [anon_sym__Atomic] = ACTIONS(4274),
    [anon_sym__Noreturn] = ACTIONS(4274),
    [anon_sym__Nonnull] = ACTIONS(4274),
    [anon_sym_nullable] = ACTIONS(4274),
    [anon_sym__Complex] = ACTIONS(4274),
    [anon_sym__Nullable] = ACTIONS(4274),
    [anon_sym__Nullable_result] = ACTIONS(4274),
    [anon_sym__Null_unspecified] = ACTIONS(4274),
    [anon_sym___autoreleasing] = ACTIONS(4274),
    [anon_sym___block] = ACTIONS(4274),
    [anon_sym___bridge] = ACTIONS(4274),
    [anon_sym___bridge_retained] = ACTIONS(4274),
    [anon_sym___bridge_transfer] = ACTIONS(4274),
    [anon_sym___complex] = ACTIONS(4274),
    [anon_sym___const] = ACTIONS(4274),
    [anon_sym___imag] = ACTIONS(4274),
    [anon_sym___kindof] = ACTIONS(4274),
    [anon_sym___nonnull] = ACTIONS(4274),
    [anon_sym___nullable] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4274),
    [anon_sym___real] = ACTIONS(4274),
    [anon_sym___strong] = ACTIONS(4274),
    [anon_sym___unsafe_unretained] = ACTIONS(4274),
    [anon_sym___unused] = ACTIONS(4274),
    [anon_sym___weak] = ACTIONS(4274),
    [anon_sym_alignas] = ACTIONS(4274),
    [anon_sym__Alignas] = ACTIONS(4274),
    [sym_primitive_type] = ACTIONS(4274),
    [anon_sym_enum] = ACTIONS(4274),
    [anon_sym_struct] = ACTIONS(4274),
    [anon_sym_union] = ACTIONS(4274),
    [anon_sym_QMARK] = ACTIONS(4276),
    [anon_sym_DASH_DASH] = ACTIONS(4276),
    [anon_sym_PLUS_PLUS] = ACTIONS(4276),
    [anon_sym_DOT] = ACTIONS(4274),
    [anon_sym_DASH_GT] = ACTIONS(4276),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4274),
    [anon_sym___typeof] = ACTIONS(4274),
    [anon_sym_typeof] = ACTIONS(4274),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4274),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4274),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4274),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4274),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4274),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4274),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4274),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4274),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4274),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_API_AVAILABLE] = ACTIONS(4274),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_API_DEPRECATED] = ACTIONS(4274),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4274),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4274),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4274),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4274),
    [anon_sym___deprecated_msg] = ACTIONS(4274),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4274),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4274),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4274),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4274),
    [anon_sym_BOOL] = ACTIONS(4274),
    [anon_sym_IMP] = ACTIONS(4274),
    [anon_sym_SEL] = ACTIONS(4274),
    [anon_sym_Class] = ACTIONS(4274),
    [anon_sym_id] = ACTIONS(4274),
  },
  [STATE(1779)] = {
    [sym_declaration] = STATE(1827),
    [sym__declaration_modifiers] = STATE(1966),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1966),
    [sym_attribute_declaration] = STATE(1966),
    [sym_ms_declspec_modifier] = STATE(1966),
    [sym_compound_statement] = STATE(2262),
    [sym_storage_class_specifier] = STATE(1966),
    [sym_type_qualifier] = STATE(1966),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1966),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1827),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1966),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1786),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4811),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1780)] = {
    [sym_identifier] = ACTIONS(4058),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4060),
    [anon_sym_COMMA] = ACTIONS(4060),
    [anon_sym_LPAREN2] = ACTIONS(4060),
    [anon_sym_DASH] = ACTIONS(4058),
    [anon_sym_PLUS] = ACTIONS(4058),
    [anon_sym_STAR] = ACTIONS(4060),
    [anon_sym_SLASH] = ACTIONS(4058),
    [anon_sym_PERCENT] = ACTIONS(4060),
    [anon_sym_PIPE_PIPE] = ACTIONS(4060),
    [anon_sym_AMP_AMP] = ACTIONS(4060),
    [anon_sym_PIPE] = ACTIONS(4058),
    [anon_sym_CARET] = ACTIONS(4060),
    [anon_sym_AMP] = ACTIONS(4058),
    [anon_sym_EQ_EQ] = ACTIONS(4060),
    [anon_sym_BANG_EQ] = ACTIONS(4060),
    [anon_sym_GT] = ACTIONS(4058),
    [anon_sym_GT_EQ] = ACTIONS(4060),
    [anon_sym_LT_EQ] = ACTIONS(4060),
    [anon_sym_LT] = ACTIONS(4058),
    [anon_sym_LT_LT] = ACTIONS(4060),
    [anon_sym_GT_GT] = ACTIONS(4060),
    [anon_sym_SEMI] = ACTIONS(4060),
    [anon_sym___extension__] = ACTIONS(4058),
    [anon_sym_extern] = ACTIONS(4058),
    [anon_sym___attribute__] = ACTIONS(4058),
    [anon_sym___attribute] = ACTIONS(4058),
    [anon_sym_noreturn] = ACTIONS(4058),
    [anon_sym_LBRACK] = ACTIONS(4060),
    [anon_sym___declspec] = ACTIONS(4058),
    [anon_sym_LBRACE] = ACTIONS(4060),
    [anon_sym_signed] = ACTIONS(4058),
    [anon_sym_unsigned] = ACTIONS(4058),
    [anon_sym_long] = ACTIONS(4058),
    [anon_sym_short] = ACTIONS(4058),
    [anon_sym_static] = ACTIONS(4058),
    [anon_sym_ATautoreleasepool] = ACTIONS(4060),
    [anon_sym_auto] = ACTIONS(4058),
    [anon_sym_register] = ACTIONS(4058),
    [anon_sym_inline] = ACTIONS(4058),
    [anon_sym___inline] = ACTIONS(4058),
    [anon_sym___inline__] = ACTIONS(4058),
    [anon_sym___forceinline] = ACTIONS(4058),
    [anon_sym_thread_local] = ACTIONS(4058),
    [anon_sym___thread] = ACTIONS(4058),
    [anon_sym_CG_EXTERN] = ACTIONS(4058),
    [anon_sym_CG_INLINE] = ACTIONS(4058),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4058),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4058),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4058),
    [anon_sym_IBOutlet] = ACTIONS(4058),
    [anon_sym_IBInspectable] = ACTIONS(4058),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4058),
    [anon_sym_NS_INLINE] = ACTIONS(4058),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4058),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4058),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4058),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4058),
    [anon_sym_const] = ACTIONS(4058),
    [anon_sym_constexpr] = ACTIONS(4058),
    [anon_sym_volatile] = ACTIONS(4058),
    [anon_sym_restrict] = ACTIONS(4058),
    [anon_sym___restrict__] = ACTIONS(4058),
    [anon_sym__Atomic] = ACTIONS(4058),
    [anon_sym__Noreturn] = ACTIONS(4058),
    [anon_sym__Nonnull] = ACTIONS(4058),
    [anon_sym_nullable] = ACTIONS(4058),
    [anon_sym__Complex] = ACTIONS(4058),
    [anon_sym__Nullable] = ACTIONS(4058),
    [anon_sym__Nullable_result] = ACTIONS(4058),
    [anon_sym__Null_unspecified] = ACTIONS(4058),
    [anon_sym___autoreleasing] = ACTIONS(4058),
    [anon_sym___block] = ACTIONS(4058),
    [anon_sym___bridge] = ACTIONS(4058),
    [anon_sym___bridge_retained] = ACTIONS(4058),
    [anon_sym___bridge_transfer] = ACTIONS(4058),
    [anon_sym___complex] = ACTIONS(4058),
    [anon_sym___const] = ACTIONS(4058),
    [anon_sym___imag] = ACTIONS(4058),
    [anon_sym___kindof] = ACTIONS(4058),
    [anon_sym___nonnull] = ACTIONS(4058),
    [anon_sym___nullable] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4058),
    [anon_sym___real] = ACTIONS(4058),
    [anon_sym___strong] = ACTIONS(4058),
    [anon_sym___unsafe_unretained] = ACTIONS(4058),
    [anon_sym___unused] = ACTIONS(4058),
    [anon_sym___weak] = ACTIONS(4058),
    [anon_sym_alignas] = ACTIONS(4058),
    [anon_sym__Alignas] = ACTIONS(4058),
    [sym_primitive_type] = ACTIONS(4058),
    [anon_sym_enum] = ACTIONS(4058),
    [anon_sym_struct] = ACTIONS(4058),
    [anon_sym_union] = ACTIONS(4058),
    [anon_sym_QMARK] = ACTIONS(4060),
    [anon_sym_DASH_DASH] = ACTIONS(4060),
    [anon_sym_PLUS_PLUS] = ACTIONS(4060),
    [anon_sym_DOT] = ACTIONS(4058),
    [anon_sym_DASH_GT] = ACTIONS(4060),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4058),
    [anon_sym___typeof] = ACTIONS(4058),
    [anon_sym_typeof] = ACTIONS(4058),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4058),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4058),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4058),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4058),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4058),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4058),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4058),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4058),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4058),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_API_AVAILABLE] = ACTIONS(4058),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_API_DEPRECATED] = ACTIONS(4058),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4058),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4058),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4058),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4058),
    [anon_sym___deprecated_msg] = ACTIONS(4058),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4058),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4058),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4058),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4058),
    [anon_sym_BOOL] = ACTIONS(4058),
    [anon_sym_IMP] = ACTIONS(4058),
    [anon_sym_SEL] = ACTIONS(4058),
    [anon_sym_Class] = ACTIONS(4058),
    [anon_sym_id] = ACTIONS(4058),
  },
  [STATE(1781)] = {
    [sym_declaration] = STATE(1873),
    [sym__declaration_modifiers] = STATE(1955),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1955),
    [sym_attribute_declaration] = STATE(1955),
    [sym_ms_declspec_modifier] = STATE(1955),
    [sym_compound_statement] = STATE(2186),
    [sym_storage_class_specifier] = STATE(1955),
    [sym_type_qualifier] = STATE(1955),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1955),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1873),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1955),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1769),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4834),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1782)] = {
    [sym_declaration] = STATE(1836),
    [sym__declaration_modifiers] = STATE(1951),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1951),
    [sym_attribute_declaration] = STATE(1951),
    [sym_ms_declspec_modifier] = STATE(1951),
    [sym_compound_statement] = STATE(2106),
    [sym_storage_class_specifier] = STATE(1951),
    [sym_type_qualifier] = STATE(1951),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1951),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1836),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1951),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4870),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1783)] = {
    [sym_declaration] = STATE(1836),
    [sym__declaration_modifiers] = STATE(1951),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1951),
    [sym_attribute_declaration] = STATE(1951),
    [sym_ms_declspec_modifier] = STATE(1951),
    [sym_compound_statement] = STATE(2106),
    [sym_storage_class_specifier] = STATE(1951),
    [sym_type_qualifier] = STATE(1951),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1951),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1836),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1951),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1789),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4870),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1784)] = {
    [sym_identifier] = ACTIONS(4220),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4222),
    [anon_sym_COMMA] = ACTIONS(4222),
    [anon_sym_LPAREN2] = ACTIONS(4222),
    [anon_sym_DASH] = ACTIONS(4220),
    [anon_sym_PLUS] = ACTIONS(4220),
    [anon_sym_STAR] = ACTIONS(4222),
    [anon_sym_SLASH] = ACTIONS(4220),
    [anon_sym_PERCENT] = ACTIONS(4222),
    [anon_sym_PIPE_PIPE] = ACTIONS(4222),
    [anon_sym_AMP_AMP] = ACTIONS(4222),
    [anon_sym_PIPE] = ACTIONS(4220),
    [anon_sym_CARET] = ACTIONS(4222),
    [anon_sym_AMP] = ACTIONS(4220),
    [anon_sym_EQ_EQ] = ACTIONS(4222),
    [anon_sym_BANG_EQ] = ACTIONS(4222),
    [anon_sym_GT] = ACTIONS(4220),
    [anon_sym_GT_EQ] = ACTIONS(4222),
    [anon_sym_LT_EQ] = ACTIONS(4222),
    [anon_sym_LT] = ACTIONS(4220),
    [anon_sym_LT_LT] = ACTIONS(4222),
    [anon_sym_GT_GT] = ACTIONS(4222),
    [anon_sym_SEMI] = ACTIONS(4222),
    [anon_sym___extension__] = ACTIONS(4220),
    [anon_sym_extern] = ACTIONS(4220),
    [anon_sym___attribute__] = ACTIONS(4220),
    [anon_sym___attribute] = ACTIONS(4220),
    [anon_sym_noreturn] = ACTIONS(4220),
    [anon_sym_LBRACK] = ACTIONS(4222),
    [anon_sym___declspec] = ACTIONS(4220),
    [anon_sym_LBRACE] = ACTIONS(4222),
    [anon_sym_signed] = ACTIONS(4220),
    [anon_sym_unsigned] = ACTIONS(4220),
    [anon_sym_long] = ACTIONS(4220),
    [anon_sym_short] = ACTIONS(4220),
    [anon_sym_static] = ACTIONS(4220),
    [anon_sym_ATautoreleasepool] = ACTIONS(4222),
    [anon_sym_auto] = ACTIONS(4220),
    [anon_sym_register] = ACTIONS(4220),
    [anon_sym_inline] = ACTIONS(4220),
    [anon_sym___inline] = ACTIONS(4220),
    [anon_sym___inline__] = ACTIONS(4220),
    [anon_sym___forceinline] = ACTIONS(4220),
    [anon_sym_thread_local] = ACTIONS(4220),
    [anon_sym___thread] = ACTIONS(4220),
    [anon_sym_CG_EXTERN] = ACTIONS(4220),
    [anon_sym_CG_INLINE] = ACTIONS(4220),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4220),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4220),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4220),
    [anon_sym_IBOutlet] = ACTIONS(4220),
    [anon_sym_IBInspectable] = ACTIONS(4220),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4220),
    [anon_sym_NS_INLINE] = ACTIONS(4220),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4220),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4220),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4220),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4220),
    [anon_sym_const] = ACTIONS(4220),
    [anon_sym_constexpr] = ACTIONS(4220),
    [anon_sym_volatile] = ACTIONS(4220),
    [anon_sym_restrict] = ACTIONS(4220),
    [anon_sym___restrict__] = ACTIONS(4220),
    [anon_sym__Atomic] = ACTIONS(4220),
    [anon_sym__Noreturn] = ACTIONS(4220),
    [anon_sym__Nonnull] = ACTIONS(4220),
    [anon_sym_nullable] = ACTIONS(4220),
    [anon_sym__Complex] = ACTIONS(4220),
    [anon_sym__Nullable] = ACTIONS(4220),
    [anon_sym__Nullable_result] = ACTIONS(4220),
    [anon_sym__Null_unspecified] = ACTIONS(4220),
    [anon_sym___autoreleasing] = ACTIONS(4220),
    [anon_sym___block] = ACTIONS(4220),
    [anon_sym___bridge] = ACTIONS(4220),
    [anon_sym___bridge_retained] = ACTIONS(4220),
    [anon_sym___bridge_transfer] = ACTIONS(4220),
    [anon_sym___complex] = ACTIONS(4220),
    [anon_sym___const] = ACTIONS(4220),
    [anon_sym___imag] = ACTIONS(4220),
    [anon_sym___kindof] = ACTIONS(4220),
    [anon_sym___nonnull] = ACTIONS(4220),
    [anon_sym___nullable] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4220),
    [anon_sym___real] = ACTIONS(4220),
    [anon_sym___strong] = ACTIONS(4220),
    [anon_sym___unsafe_unretained] = ACTIONS(4220),
    [anon_sym___unused] = ACTIONS(4220),
    [anon_sym___weak] = ACTIONS(4220),
    [anon_sym_alignas] = ACTIONS(4220),
    [anon_sym__Alignas] = ACTIONS(4220),
    [sym_primitive_type] = ACTIONS(4220),
    [anon_sym_enum] = ACTIONS(4220),
    [anon_sym_struct] = ACTIONS(4220),
    [anon_sym_union] = ACTIONS(4220),
    [anon_sym_QMARK] = ACTIONS(4222),
    [anon_sym_DASH_DASH] = ACTIONS(4222),
    [anon_sym_PLUS_PLUS] = ACTIONS(4222),
    [anon_sym_DOT] = ACTIONS(4220),
    [anon_sym_DASH_GT] = ACTIONS(4222),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4220),
    [anon_sym___typeof] = ACTIONS(4220),
    [anon_sym_typeof] = ACTIONS(4220),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4220),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4220),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4220),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4220),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4220),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4220),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4220),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4220),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4220),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_API_AVAILABLE] = ACTIONS(4220),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_API_DEPRECATED] = ACTIONS(4220),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4220),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4220),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4220),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4220),
    [anon_sym___deprecated_msg] = ACTIONS(4220),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4220),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4220),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4220),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4220),
    [anon_sym_BOOL] = ACTIONS(4220),
    [anon_sym_IMP] = ACTIONS(4220),
    [anon_sym_SEL] = ACTIONS(4220),
    [anon_sym_Class] = ACTIONS(4220),
    [anon_sym_id] = ACTIONS(4220),
  },
  [STATE(1785)] = {
    [sym_identifier] = ACTIONS(4224),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4226),
    [anon_sym_COMMA] = ACTIONS(4226),
    [anon_sym_LPAREN2] = ACTIONS(4226),
    [anon_sym_DASH] = ACTIONS(4224),
    [anon_sym_PLUS] = ACTIONS(4224),
    [anon_sym_STAR] = ACTIONS(4226),
    [anon_sym_SLASH] = ACTIONS(4224),
    [anon_sym_PERCENT] = ACTIONS(4226),
    [anon_sym_PIPE_PIPE] = ACTIONS(4226),
    [anon_sym_AMP_AMP] = ACTIONS(4226),
    [anon_sym_PIPE] = ACTIONS(4224),
    [anon_sym_CARET] = ACTIONS(4226),
    [anon_sym_AMP] = ACTIONS(4224),
    [anon_sym_EQ_EQ] = ACTIONS(4226),
    [anon_sym_BANG_EQ] = ACTIONS(4226),
    [anon_sym_GT] = ACTIONS(4224),
    [anon_sym_GT_EQ] = ACTIONS(4226),
    [anon_sym_LT_EQ] = ACTIONS(4226),
    [anon_sym_LT] = ACTIONS(4224),
    [anon_sym_LT_LT] = ACTIONS(4226),
    [anon_sym_GT_GT] = ACTIONS(4226),
    [anon_sym_SEMI] = ACTIONS(4226),
    [anon_sym___extension__] = ACTIONS(4224),
    [anon_sym_extern] = ACTIONS(4224),
    [anon_sym___attribute__] = ACTIONS(4224),
    [anon_sym___attribute] = ACTIONS(4224),
    [anon_sym_noreturn] = ACTIONS(4224),
    [anon_sym_LBRACK] = ACTIONS(4226),
    [anon_sym___declspec] = ACTIONS(4224),
    [anon_sym_LBRACE] = ACTIONS(4226),
    [anon_sym_signed] = ACTIONS(4224),
    [anon_sym_unsigned] = ACTIONS(4224),
    [anon_sym_long] = ACTIONS(4224),
    [anon_sym_short] = ACTIONS(4224),
    [anon_sym_static] = ACTIONS(4224),
    [anon_sym_ATautoreleasepool] = ACTIONS(4226),
    [anon_sym_auto] = ACTIONS(4224),
    [anon_sym_register] = ACTIONS(4224),
    [anon_sym_inline] = ACTIONS(4224),
    [anon_sym___inline] = ACTIONS(4224),
    [anon_sym___inline__] = ACTIONS(4224),
    [anon_sym___forceinline] = ACTIONS(4224),
    [anon_sym_thread_local] = ACTIONS(4224),
    [anon_sym___thread] = ACTIONS(4224),
    [anon_sym_CG_EXTERN] = ACTIONS(4224),
    [anon_sym_CG_INLINE] = ACTIONS(4224),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4224),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4224),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4224),
    [anon_sym_IBOutlet] = ACTIONS(4224),
    [anon_sym_IBInspectable] = ACTIONS(4224),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4224),
    [anon_sym_NS_INLINE] = ACTIONS(4224),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4224),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4224),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4224),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4224),
    [anon_sym_const] = ACTIONS(4224),
    [anon_sym_constexpr] = ACTIONS(4224),
    [anon_sym_volatile] = ACTIONS(4224),
    [anon_sym_restrict] = ACTIONS(4224),
    [anon_sym___restrict__] = ACTIONS(4224),
    [anon_sym__Atomic] = ACTIONS(4224),
    [anon_sym__Noreturn] = ACTIONS(4224),
    [anon_sym__Nonnull] = ACTIONS(4224),
    [anon_sym_nullable] = ACTIONS(4224),
    [anon_sym__Complex] = ACTIONS(4224),
    [anon_sym__Nullable] = ACTIONS(4224),
    [anon_sym__Nullable_result] = ACTIONS(4224),
    [anon_sym__Null_unspecified] = ACTIONS(4224),
    [anon_sym___autoreleasing] = ACTIONS(4224),
    [anon_sym___block] = ACTIONS(4224),
    [anon_sym___bridge] = ACTIONS(4224),
    [anon_sym___bridge_retained] = ACTIONS(4224),
    [anon_sym___bridge_transfer] = ACTIONS(4224),
    [anon_sym___complex] = ACTIONS(4224),
    [anon_sym___const] = ACTIONS(4224),
    [anon_sym___imag] = ACTIONS(4224),
    [anon_sym___kindof] = ACTIONS(4224),
    [anon_sym___nonnull] = ACTIONS(4224),
    [anon_sym___nullable] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4224),
    [anon_sym___real] = ACTIONS(4224),
    [anon_sym___strong] = ACTIONS(4224),
    [anon_sym___unsafe_unretained] = ACTIONS(4224),
    [anon_sym___unused] = ACTIONS(4224),
    [anon_sym___weak] = ACTIONS(4224),
    [anon_sym_alignas] = ACTIONS(4224),
    [anon_sym__Alignas] = ACTIONS(4224),
    [sym_primitive_type] = ACTIONS(4224),
    [anon_sym_enum] = ACTIONS(4224),
    [anon_sym_struct] = ACTIONS(4224),
    [anon_sym_union] = ACTIONS(4224),
    [anon_sym_QMARK] = ACTIONS(4226),
    [anon_sym_DASH_DASH] = ACTIONS(4226),
    [anon_sym_PLUS_PLUS] = ACTIONS(4226),
    [anon_sym_DOT] = ACTIONS(4224),
    [anon_sym_DASH_GT] = ACTIONS(4226),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4224),
    [anon_sym___typeof] = ACTIONS(4224),
    [anon_sym_typeof] = ACTIONS(4224),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4224),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4224),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4224),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4224),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4224),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4224),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4224),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4224),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4224),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_API_AVAILABLE] = ACTIONS(4224),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_API_DEPRECATED] = ACTIONS(4224),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4224),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4224),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4224),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4224),
    [anon_sym___deprecated_msg] = ACTIONS(4224),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4224),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4224),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4224),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4224),
    [anon_sym_BOOL] = ACTIONS(4224),
    [anon_sym_IMP] = ACTIONS(4224),
    [anon_sym_SEL] = ACTIONS(4224),
    [anon_sym_Class] = ACTIONS(4224),
    [anon_sym_id] = ACTIONS(4224),
  },
  [STATE(1786)] = {
    [sym_declaration] = STATE(1830),
    [sym__declaration_modifiers] = STATE(1980),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1980),
    [sym_attribute_declaration] = STATE(1980),
    [sym_ms_declspec_modifier] = STATE(1980),
    [sym_compound_statement] = STATE(2297),
    [sym_storage_class_specifier] = STATE(1980),
    [sym_type_qualifier] = STATE(1980),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1980),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1830),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1980),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4830),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1787)] = {
    [sym_declaration] = STATE(1830),
    [sym__declaration_modifiers] = STATE(1980),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1980),
    [sym_attribute_declaration] = STATE(1980),
    [sym_ms_declspec_modifier] = STATE(1980),
    [sym_compound_statement] = STATE(2297),
    [sym_storage_class_specifier] = STATE(1980),
    [sym_type_qualifier] = STATE(1980),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1980),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1830),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1980),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1819),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4830),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1788)] = {
    [sym_declaration] = STATE(1873),
    [sym__declaration_modifiers] = STATE(1955),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1955),
    [sym_attribute_declaration] = STATE(1955),
    [sym_ms_declspec_modifier] = STATE(1955),
    [sym_compound_statement] = STATE(2186),
    [sym_storage_class_specifier] = STATE(1955),
    [sym_type_qualifier] = STATE(1955),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1955),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1873),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1955),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4834),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1789)] = {
    [sym_declaration] = STATE(1839),
    [sym__declaration_modifiers] = STATE(1925),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1925),
    [sym_attribute_declaration] = STATE(1925),
    [sym_ms_declspec_modifier] = STATE(1925),
    [sym_compound_statement] = STATE(2107),
    [sym_storage_class_specifier] = STATE(1925),
    [sym_type_qualifier] = STATE(1925),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1925),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1839),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1925),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4872),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1790)] = {
    [sym_declaration] = STATE(1839),
    [sym__declaration_modifiers] = STATE(1925),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1925),
    [sym_attribute_declaration] = STATE(1925),
    [sym_ms_declspec_modifier] = STATE(1925),
    [sym_compound_statement] = STATE(2107),
    [sym_storage_class_specifier] = STATE(1925),
    [sym_type_qualifier] = STATE(1925),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1925),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1839),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1925),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1793),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4872),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1791)] = {
    [sym_declaration] = STATE(1862),
    [sym__declaration_modifiers] = STATE(1933),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1933),
    [sym_attribute_declaration] = STATE(1933),
    [sym_ms_declspec_modifier] = STATE(1933),
    [sym_compound_statement] = STATE(2140),
    [sym_storage_class_specifier] = STATE(1933),
    [sym_type_qualifier] = STATE(1933),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1933),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1862),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1933),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1788),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4854),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1792)] = {
    [sym_declaration] = STATE(1834),
    [sym__declaration_modifiers] = STATE(1931),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1931),
    [sym_attribute_declaration] = STATE(1931),
    [sym_ms_declspec_modifier] = STATE(1931),
    [sym_compound_statement] = STATE(2345),
    [sym_storage_class_specifier] = STATE(1931),
    [sym_type_qualifier] = STATE(1931),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1931),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1834),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1931),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1798),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4844),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1793)] = {
    [sym_declaration] = STATE(1845),
    [sym__declaration_modifiers] = STATE(1969),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1969),
    [sym_attribute_declaration] = STATE(1969),
    [sym_ms_declspec_modifier] = STATE(1969),
    [sym_compound_statement] = STATE(2108),
    [sym_storage_class_specifier] = STATE(1969),
    [sym_type_qualifier] = STATE(1969),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1969),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1845),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1969),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4874),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1794)] = {
    [sym_declaration] = STATE(1876),
    [sym__declaration_modifiers] = STATE(1932),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1932),
    [sym_attribute_declaration] = STATE(1932),
    [sym_ms_declspec_modifier] = STATE(1932),
    [sym_compound_statement] = STATE(2309),
    [sym_storage_class_specifier] = STATE(1932),
    [sym_type_qualifier] = STATE(1932),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1932),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1876),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1932),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4842),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1795)] = {
    [sym_identifier] = ACTIONS(4050),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4052),
    [anon_sym_COMMA] = ACTIONS(4052),
    [anon_sym_LPAREN2] = ACTIONS(4052),
    [anon_sym_DASH] = ACTIONS(4050),
    [anon_sym_PLUS] = ACTIONS(4050),
    [anon_sym_STAR] = ACTIONS(4052),
    [anon_sym_SLASH] = ACTIONS(4050),
    [anon_sym_PERCENT] = ACTIONS(4052),
    [anon_sym_PIPE_PIPE] = ACTIONS(4052),
    [anon_sym_AMP_AMP] = ACTIONS(4052),
    [anon_sym_PIPE] = ACTIONS(4050),
    [anon_sym_CARET] = ACTIONS(4052),
    [anon_sym_AMP] = ACTIONS(4050),
    [anon_sym_EQ_EQ] = ACTIONS(4052),
    [anon_sym_BANG_EQ] = ACTIONS(4052),
    [anon_sym_GT] = ACTIONS(4050),
    [anon_sym_GT_EQ] = ACTIONS(4052),
    [anon_sym_LT_EQ] = ACTIONS(4052),
    [anon_sym_LT] = ACTIONS(4050),
    [anon_sym_LT_LT] = ACTIONS(4052),
    [anon_sym_GT_GT] = ACTIONS(4052),
    [anon_sym_SEMI] = ACTIONS(4052),
    [anon_sym___extension__] = ACTIONS(4050),
    [anon_sym_extern] = ACTIONS(4050),
    [anon_sym___attribute__] = ACTIONS(4050),
    [anon_sym___attribute] = ACTIONS(4050),
    [anon_sym_noreturn] = ACTIONS(4050),
    [anon_sym_LBRACK] = ACTIONS(4052),
    [anon_sym___declspec] = ACTIONS(4050),
    [anon_sym_LBRACE] = ACTIONS(4052),
    [anon_sym_signed] = ACTIONS(4050),
    [anon_sym_unsigned] = ACTIONS(4050),
    [anon_sym_long] = ACTIONS(4050),
    [anon_sym_short] = ACTIONS(4050),
    [anon_sym_static] = ACTIONS(4050),
    [anon_sym_ATautoreleasepool] = ACTIONS(4052),
    [anon_sym_auto] = ACTIONS(4050),
    [anon_sym_register] = ACTIONS(4050),
    [anon_sym_inline] = ACTIONS(4050),
    [anon_sym___inline] = ACTIONS(4050),
    [anon_sym___inline__] = ACTIONS(4050),
    [anon_sym___forceinline] = ACTIONS(4050),
    [anon_sym_thread_local] = ACTIONS(4050),
    [anon_sym___thread] = ACTIONS(4050),
    [anon_sym_CG_EXTERN] = ACTIONS(4050),
    [anon_sym_CG_INLINE] = ACTIONS(4050),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4050),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4050),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4050),
    [anon_sym_IBOutlet] = ACTIONS(4050),
    [anon_sym_IBInspectable] = ACTIONS(4050),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4050),
    [anon_sym_NS_INLINE] = ACTIONS(4050),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4050),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4050),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4050),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4050),
    [anon_sym_const] = ACTIONS(4050),
    [anon_sym_constexpr] = ACTIONS(4050),
    [anon_sym_volatile] = ACTIONS(4050),
    [anon_sym_restrict] = ACTIONS(4050),
    [anon_sym___restrict__] = ACTIONS(4050),
    [anon_sym__Atomic] = ACTIONS(4050),
    [anon_sym__Noreturn] = ACTIONS(4050),
    [anon_sym__Nonnull] = ACTIONS(4050),
    [anon_sym_nullable] = ACTIONS(4050),
    [anon_sym__Complex] = ACTIONS(4050),
    [anon_sym__Nullable] = ACTIONS(4050),
    [anon_sym__Nullable_result] = ACTIONS(4050),
    [anon_sym__Null_unspecified] = ACTIONS(4050),
    [anon_sym___autoreleasing] = ACTIONS(4050),
    [anon_sym___block] = ACTIONS(4050),
    [anon_sym___bridge] = ACTIONS(4050),
    [anon_sym___bridge_retained] = ACTIONS(4050),
    [anon_sym___bridge_transfer] = ACTIONS(4050),
    [anon_sym___complex] = ACTIONS(4050),
    [anon_sym___const] = ACTIONS(4050),
    [anon_sym___imag] = ACTIONS(4050),
    [anon_sym___kindof] = ACTIONS(4050),
    [anon_sym___nonnull] = ACTIONS(4050),
    [anon_sym___nullable] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4050),
    [anon_sym___real] = ACTIONS(4050),
    [anon_sym___strong] = ACTIONS(4050),
    [anon_sym___unsafe_unretained] = ACTIONS(4050),
    [anon_sym___unused] = ACTIONS(4050),
    [anon_sym___weak] = ACTIONS(4050),
    [anon_sym_alignas] = ACTIONS(4050),
    [anon_sym__Alignas] = ACTIONS(4050),
    [sym_primitive_type] = ACTIONS(4050),
    [anon_sym_enum] = ACTIONS(4050),
    [anon_sym_struct] = ACTIONS(4050),
    [anon_sym_union] = ACTIONS(4050),
    [anon_sym_QMARK] = ACTIONS(4052),
    [anon_sym_DASH_DASH] = ACTIONS(4052),
    [anon_sym_PLUS_PLUS] = ACTIONS(4052),
    [anon_sym_DOT] = ACTIONS(4050),
    [anon_sym_DASH_GT] = ACTIONS(4052),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4050),
    [anon_sym___typeof] = ACTIONS(4050),
    [anon_sym_typeof] = ACTIONS(4050),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4050),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4050),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4050),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4050),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4050),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4050),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4050),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4050),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4050),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_API_AVAILABLE] = ACTIONS(4050),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_API_DEPRECATED] = ACTIONS(4050),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4050),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4050),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4050),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4050),
    [anon_sym___deprecated_msg] = ACTIONS(4050),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4050),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4050),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4050),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4050),
    [anon_sym_BOOL] = ACTIONS(4050),
    [anon_sym_IMP] = ACTIONS(4050),
    [anon_sym_SEL] = ACTIONS(4050),
    [anon_sym_Class] = ACTIONS(4050),
    [anon_sym_id] = ACTIONS(4050),
  },
  [STATE(1796)] = {
    [sym_declaration] = STATE(1863),
    [sym__declaration_modifiers] = STATE(1937),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1937),
    [sym_attribute_declaration] = STATE(1937),
    [sym_ms_declspec_modifier] = STATE(1937),
    [sym_compound_statement] = STATE(2271),
    [sym_storage_class_specifier] = STATE(1937),
    [sym_type_qualifier] = STATE(1937),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1937),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1863),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1937),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4876),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1797)] = {
    [sym_declaration] = STATE(1863),
    [sym__declaration_modifiers] = STATE(1937),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1937),
    [sym_attribute_declaration] = STATE(1937),
    [sym_ms_declspec_modifier] = STATE(1937),
    [sym_compound_statement] = STATE(2271),
    [sym_storage_class_specifier] = STATE(1937),
    [sym_type_qualifier] = STATE(1937),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1937),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1863),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1937),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1802),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4876),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1798)] = {
    [sym_declaration] = STATE(1838),
    [sym__declaration_modifiers] = STATE(1975),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1975),
    [sym_attribute_declaration] = STATE(1975),
    [sym_ms_declspec_modifier] = STATE(1975),
    [sym_compound_statement] = STATE(2308),
    [sym_storage_class_specifier] = STATE(1975),
    [sym_type_qualifier] = STATE(1975),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1975),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1838),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1975),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4878),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1799)] = {
    [sym_declaration] = STATE(1838),
    [sym__declaration_modifiers] = STATE(1975),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1975),
    [sym_attribute_declaration] = STATE(1975),
    [sym_ms_declspec_modifier] = STATE(1975),
    [sym_compound_statement] = STATE(2308),
    [sym_storage_class_specifier] = STATE(1975),
    [sym_type_qualifier] = STATE(1975),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1975),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1838),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1975),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1804),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4878),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1800)] = {
    [sym_identifier] = ACTIONS(4062),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4064),
    [anon_sym_COMMA] = ACTIONS(4064),
    [anon_sym_LPAREN2] = ACTIONS(4064),
    [anon_sym_DASH] = ACTIONS(4062),
    [anon_sym_PLUS] = ACTIONS(4062),
    [anon_sym_STAR] = ACTIONS(4064),
    [anon_sym_SLASH] = ACTIONS(4062),
    [anon_sym_PERCENT] = ACTIONS(4064),
    [anon_sym_PIPE_PIPE] = ACTIONS(4064),
    [anon_sym_AMP_AMP] = ACTIONS(4064),
    [anon_sym_PIPE] = ACTIONS(4062),
    [anon_sym_CARET] = ACTIONS(4064),
    [anon_sym_AMP] = ACTIONS(4062),
    [anon_sym_EQ_EQ] = ACTIONS(4064),
    [anon_sym_BANG_EQ] = ACTIONS(4064),
    [anon_sym_GT] = ACTIONS(4062),
    [anon_sym_GT_EQ] = ACTIONS(4064),
    [anon_sym_LT_EQ] = ACTIONS(4064),
    [anon_sym_LT] = ACTIONS(4062),
    [anon_sym_LT_LT] = ACTIONS(4064),
    [anon_sym_GT_GT] = ACTIONS(4064),
    [anon_sym_SEMI] = ACTIONS(4064),
    [anon_sym___extension__] = ACTIONS(4062),
    [anon_sym_extern] = ACTIONS(4062),
    [anon_sym___attribute__] = ACTIONS(4062),
    [anon_sym___attribute] = ACTIONS(4062),
    [anon_sym_noreturn] = ACTIONS(4062),
    [anon_sym_LBRACK] = ACTIONS(4064),
    [anon_sym___declspec] = ACTIONS(4062),
    [anon_sym_LBRACE] = ACTIONS(4064),
    [anon_sym_signed] = ACTIONS(4062),
    [anon_sym_unsigned] = ACTIONS(4062),
    [anon_sym_long] = ACTIONS(4062),
    [anon_sym_short] = ACTIONS(4062),
    [anon_sym_static] = ACTIONS(4062),
    [anon_sym_ATautoreleasepool] = ACTIONS(4064),
    [anon_sym_auto] = ACTIONS(4062),
    [anon_sym_register] = ACTIONS(4062),
    [anon_sym_inline] = ACTIONS(4062),
    [anon_sym___inline] = ACTIONS(4062),
    [anon_sym___inline__] = ACTIONS(4062),
    [anon_sym___forceinline] = ACTIONS(4062),
    [anon_sym_thread_local] = ACTIONS(4062),
    [anon_sym___thread] = ACTIONS(4062),
    [anon_sym_CG_EXTERN] = ACTIONS(4062),
    [anon_sym_CG_INLINE] = ACTIONS(4062),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4062),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4062),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4062),
    [anon_sym_IBOutlet] = ACTIONS(4062),
    [anon_sym_IBInspectable] = ACTIONS(4062),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4062),
    [anon_sym_NS_INLINE] = ACTIONS(4062),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4062),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4062),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4062),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4062),
    [anon_sym_const] = ACTIONS(4062),
    [anon_sym_constexpr] = ACTIONS(4062),
    [anon_sym_volatile] = ACTIONS(4062),
    [anon_sym_restrict] = ACTIONS(4062),
    [anon_sym___restrict__] = ACTIONS(4062),
    [anon_sym__Atomic] = ACTIONS(4062),
    [anon_sym__Noreturn] = ACTIONS(4062),
    [anon_sym__Nonnull] = ACTIONS(4062),
    [anon_sym_nullable] = ACTIONS(4062),
    [anon_sym__Complex] = ACTIONS(4062),
    [anon_sym__Nullable] = ACTIONS(4062),
    [anon_sym__Nullable_result] = ACTIONS(4062),
    [anon_sym__Null_unspecified] = ACTIONS(4062),
    [anon_sym___autoreleasing] = ACTIONS(4062),
    [anon_sym___block] = ACTIONS(4062),
    [anon_sym___bridge] = ACTIONS(4062),
    [anon_sym___bridge_retained] = ACTIONS(4062),
    [anon_sym___bridge_transfer] = ACTIONS(4062),
    [anon_sym___complex] = ACTIONS(4062),
    [anon_sym___const] = ACTIONS(4062),
    [anon_sym___imag] = ACTIONS(4062),
    [anon_sym___kindof] = ACTIONS(4062),
    [anon_sym___nonnull] = ACTIONS(4062),
    [anon_sym___nullable] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4062),
    [anon_sym___real] = ACTIONS(4062),
    [anon_sym___strong] = ACTIONS(4062),
    [anon_sym___unsafe_unretained] = ACTIONS(4062),
    [anon_sym___unused] = ACTIONS(4062),
    [anon_sym___weak] = ACTIONS(4062),
    [anon_sym_alignas] = ACTIONS(4062),
    [anon_sym__Alignas] = ACTIONS(4062),
    [sym_primitive_type] = ACTIONS(4062),
    [anon_sym_enum] = ACTIONS(4062),
    [anon_sym_struct] = ACTIONS(4062),
    [anon_sym_union] = ACTIONS(4062),
    [anon_sym_QMARK] = ACTIONS(4064),
    [anon_sym_DASH_DASH] = ACTIONS(4064),
    [anon_sym_PLUS_PLUS] = ACTIONS(4064),
    [anon_sym_DOT] = ACTIONS(4062),
    [anon_sym_DASH_GT] = ACTIONS(4064),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4062),
    [anon_sym___typeof] = ACTIONS(4062),
    [anon_sym_typeof] = ACTIONS(4062),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4062),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4062),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4062),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4062),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4062),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4062),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4062),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4062),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4062),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_API_AVAILABLE] = ACTIONS(4062),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_API_DEPRECATED] = ACTIONS(4062),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4062),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4062),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4062),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4062),
    [anon_sym___deprecated_msg] = ACTIONS(4062),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4062),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4062),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4062),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4062),
    [anon_sym_BOOL] = ACTIONS(4062),
    [anon_sym_IMP] = ACTIONS(4062),
    [anon_sym_SEL] = ACTIONS(4062),
    [anon_sym_Class] = ACTIONS(4062),
    [anon_sym_id] = ACTIONS(4062),
  },
  [STATE(1801)] = {
    [sym_identifier] = ACTIONS(4110),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4112),
    [anon_sym_COMMA] = ACTIONS(4112),
    [anon_sym_LPAREN2] = ACTIONS(4112),
    [anon_sym_DASH] = ACTIONS(4110),
    [anon_sym_PLUS] = ACTIONS(4110),
    [anon_sym_STAR] = ACTIONS(4112),
    [anon_sym_SLASH] = ACTIONS(4110),
    [anon_sym_PERCENT] = ACTIONS(4112),
    [anon_sym_PIPE_PIPE] = ACTIONS(4112),
    [anon_sym_AMP_AMP] = ACTIONS(4112),
    [anon_sym_PIPE] = ACTIONS(4110),
    [anon_sym_CARET] = ACTIONS(4112),
    [anon_sym_AMP] = ACTIONS(4110),
    [anon_sym_EQ_EQ] = ACTIONS(4112),
    [anon_sym_BANG_EQ] = ACTIONS(4112),
    [anon_sym_GT] = ACTIONS(4110),
    [anon_sym_GT_EQ] = ACTIONS(4112),
    [anon_sym_LT_EQ] = ACTIONS(4112),
    [anon_sym_LT] = ACTIONS(4110),
    [anon_sym_LT_LT] = ACTIONS(4112),
    [anon_sym_GT_GT] = ACTIONS(4112),
    [anon_sym_SEMI] = ACTIONS(4112),
    [anon_sym___extension__] = ACTIONS(4110),
    [anon_sym_extern] = ACTIONS(4110),
    [anon_sym___attribute__] = ACTIONS(4110),
    [anon_sym___attribute] = ACTIONS(4110),
    [anon_sym_noreturn] = ACTIONS(4110),
    [anon_sym_LBRACK] = ACTIONS(4112),
    [anon_sym___declspec] = ACTIONS(4110),
    [anon_sym_LBRACE] = ACTIONS(4112),
    [anon_sym_signed] = ACTIONS(4110),
    [anon_sym_unsigned] = ACTIONS(4110),
    [anon_sym_long] = ACTIONS(4110),
    [anon_sym_short] = ACTIONS(4110),
    [anon_sym_static] = ACTIONS(4110),
    [anon_sym_ATautoreleasepool] = ACTIONS(4112),
    [anon_sym_auto] = ACTIONS(4110),
    [anon_sym_register] = ACTIONS(4110),
    [anon_sym_inline] = ACTIONS(4110),
    [anon_sym___inline] = ACTIONS(4110),
    [anon_sym___inline__] = ACTIONS(4110),
    [anon_sym___forceinline] = ACTIONS(4110),
    [anon_sym_thread_local] = ACTIONS(4110),
    [anon_sym___thread] = ACTIONS(4110),
    [anon_sym_CG_EXTERN] = ACTIONS(4110),
    [anon_sym_CG_INLINE] = ACTIONS(4110),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4110),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4110),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4110),
    [anon_sym_IBOutlet] = ACTIONS(4110),
    [anon_sym_IBInspectable] = ACTIONS(4110),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4110),
    [anon_sym_NS_INLINE] = ACTIONS(4110),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4110),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4110),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4110),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4110),
    [anon_sym_const] = ACTIONS(4110),
    [anon_sym_constexpr] = ACTIONS(4110),
    [anon_sym_volatile] = ACTIONS(4110),
    [anon_sym_restrict] = ACTIONS(4110),
    [anon_sym___restrict__] = ACTIONS(4110),
    [anon_sym__Atomic] = ACTIONS(4110),
    [anon_sym__Noreturn] = ACTIONS(4110),
    [anon_sym__Nonnull] = ACTIONS(4110),
    [anon_sym_nullable] = ACTIONS(4110),
    [anon_sym__Complex] = ACTIONS(4110),
    [anon_sym__Nullable] = ACTIONS(4110),
    [anon_sym__Nullable_result] = ACTIONS(4110),
    [anon_sym__Null_unspecified] = ACTIONS(4110),
    [anon_sym___autoreleasing] = ACTIONS(4110),
    [anon_sym___block] = ACTIONS(4110),
    [anon_sym___bridge] = ACTIONS(4110),
    [anon_sym___bridge_retained] = ACTIONS(4110),
    [anon_sym___bridge_transfer] = ACTIONS(4110),
    [anon_sym___complex] = ACTIONS(4110),
    [anon_sym___const] = ACTIONS(4110),
    [anon_sym___imag] = ACTIONS(4110),
    [anon_sym___kindof] = ACTIONS(4110),
    [anon_sym___nonnull] = ACTIONS(4110),
    [anon_sym___nullable] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4110),
    [anon_sym___real] = ACTIONS(4110),
    [anon_sym___strong] = ACTIONS(4110),
    [anon_sym___unsafe_unretained] = ACTIONS(4110),
    [anon_sym___unused] = ACTIONS(4110),
    [anon_sym___weak] = ACTIONS(4110),
    [anon_sym_alignas] = ACTIONS(4110),
    [anon_sym__Alignas] = ACTIONS(4110),
    [sym_primitive_type] = ACTIONS(4110),
    [anon_sym_enum] = ACTIONS(4110),
    [anon_sym_struct] = ACTIONS(4110),
    [anon_sym_union] = ACTIONS(4110),
    [anon_sym_QMARK] = ACTIONS(4112),
    [anon_sym_DASH_DASH] = ACTIONS(4112),
    [anon_sym_PLUS_PLUS] = ACTIONS(4112),
    [anon_sym_DOT] = ACTIONS(4110),
    [anon_sym_DASH_GT] = ACTIONS(4112),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4110),
    [anon_sym___typeof] = ACTIONS(4110),
    [anon_sym_typeof] = ACTIONS(4110),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4110),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4110),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4110),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4110),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4110),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4110),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4110),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4110),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4110),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_API_AVAILABLE] = ACTIONS(4110),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_API_DEPRECATED] = ACTIONS(4110),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4110),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4110),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4110),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4110),
    [anon_sym___deprecated_msg] = ACTIONS(4110),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4110),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4110),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4110),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4110),
    [anon_sym_BOOL] = ACTIONS(4110),
    [anon_sym_IMP] = ACTIONS(4110),
    [anon_sym_SEL] = ACTIONS(4110),
    [anon_sym_Class] = ACTIONS(4110),
    [anon_sym_id] = ACTIONS(4110),
  },
  [STATE(1802)] = {
    [sym_declaration] = STATE(1867),
    [sym__declaration_modifiers] = STATE(1950),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1950),
    [sym_attribute_declaration] = STATE(1950),
    [sym_ms_declspec_modifier] = STATE(1950),
    [sym_compound_statement] = STATE(2277),
    [sym_storage_class_specifier] = STATE(1950),
    [sym_type_qualifier] = STATE(1950),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1950),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1867),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1950),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4880),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1803)] = {
    [sym_declaration] = STATE(1867),
    [sym__declaration_modifiers] = STATE(1950),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1950),
    [sym_attribute_declaration] = STATE(1950),
    [sym_ms_declspec_modifier] = STATE(1950),
    [sym_compound_statement] = STATE(2277),
    [sym_storage_class_specifier] = STATE(1950),
    [sym_type_qualifier] = STATE(1950),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1950),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1867),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1950),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1806),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4880),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1804)] = {
    [sym_declaration] = STATE(1840),
    [sym__declaration_modifiers] = STATE(1935),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1935),
    [sym_attribute_declaration] = STATE(1935),
    [sym_ms_declspec_modifier] = STATE(1935),
    [sym_compound_statement] = STATE(2312),
    [sym_storage_class_specifier] = STATE(1935),
    [sym_type_qualifier] = STATE(1935),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1935),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1840),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1935),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4864),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1805)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_LPAREN2] = ACTIONS(4044),
    [anon_sym_DASH] = ACTIONS(4042),
    [anon_sym_PLUS] = ACTIONS(4042),
    [anon_sym_STAR] = ACTIONS(4044),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4044),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(4044),
    [anon_sym_AMP] = ACTIONS(4042),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(4042),
    [anon_sym_LT_LT] = ACTIONS(4044),
    [anon_sym_GT_GT] = ACTIONS(4044),
    [anon_sym_SEMI] = ACTIONS(4882),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym___declspec] = ACTIONS(2914),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(4044),
    [anon_sym_PLUS_PLUS] = ACTIONS(4044),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATprotocol] = ACTIONS(2920),
    [anon_sym_ATinterface] = ACTIONS(2920),
    [anon_sym_ATimplementation] = ACTIONS(2920),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(1806)] = {
    [sym_declaration] = STATE(1869),
    [sym__declaration_modifiers] = STATE(1953),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1953),
    [sym_attribute_declaration] = STATE(1953),
    [sym_ms_declspec_modifier] = STATE(1953),
    [sym_compound_statement] = STATE(2283),
    [sym_storage_class_specifier] = STATE(1953),
    [sym_type_qualifier] = STATE(1953),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1953),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1869),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1953),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4884),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1807)] = {
    [sym_declaration] = STATE(1869),
    [sym__declaration_modifiers] = STATE(1953),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1953),
    [sym_attribute_declaration] = STATE(1953),
    [sym_ms_declspec_modifier] = STATE(1953),
    [sym_compound_statement] = STATE(2283),
    [sym_storage_class_specifier] = STATE(1953),
    [sym_type_qualifier] = STATE(1953),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1953),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1869),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1953),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1808),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4884),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1808)] = {
    [sym_declaration] = STATE(1871),
    [sym__declaration_modifiers] = STATE(1956),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1956),
    [sym_attribute_declaration] = STATE(1956),
    [sym_ms_declspec_modifier] = STATE(1956),
    [sym_compound_statement] = STATE(2289),
    [sym_storage_class_specifier] = STATE(1956),
    [sym_type_qualifier] = STATE(1956),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1956),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1871),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1956),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4886),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1809)] = {
    [sym_declaration] = STATE(1871),
    [sym__declaration_modifiers] = STATE(1956),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1956),
    [sym_attribute_declaration] = STATE(1956),
    [sym_ms_declspec_modifier] = STATE(1956),
    [sym_compound_statement] = STATE(2289),
    [sym_storage_class_specifier] = STATE(1956),
    [sym_type_qualifier] = STATE(1956),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1956),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1871),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1956),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1810),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4886),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1810)] = {
    [sym_declaration] = STATE(1872),
    [sym__declaration_modifiers] = STATE(1957),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1957),
    [sym_attribute_declaration] = STATE(1957),
    [sym_ms_declspec_modifier] = STATE(1957),
    [sym_compound_statement] = STATE(2299),
    [sym_storage_class_specifier] = STATE(1957),
    [sym_type_qualifier] = STATE(1957),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1957),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1872),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1957),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4888),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1811)] = {
    [sym_declaration] = STATE(1858),
    [sym__declaration_modifiers] = STATE(1926),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1926),
    [sym_attribute_declaration] = STATE(1926),
    [sym_ms_declspec_modifier] = STATE(1926),
    [sym_compound_statement] = STATE(2252),
    [sym_storage_class_specifier] = STATE(1926),
    [sym_type_qualifier] = STATE(1926),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1926),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1858),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1926),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1794),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4836),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1812)] = {
    [sym_declaration] = STATE(1876),
    [sym__declaration_modifiers] = STATE(1932),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1932),
    [sym_attribute_declaration] = STATE(1932),
    [sym_ms_declspec_modifier] = STATE(1932),
    [sym_compound_statement] = STATE(2309),
    [sym_storage_class_specifier] = STATE(1932),
    [sym_type_qualifier] = STATE(1932),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1932),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1876),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1932),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1770),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4842),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1813)] = {
    [sym_declaration] = STATE(1843),
    [sym__declaration_modifiers] = STATE(1965),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1965),
    [sym_attribute_declaration] = STATE(1965),
    [sym_ms_declspec_modifier] = STATE(1965),
    [sym_compound_statement] = STATE(2315),
    [sym_storage_class_specifier] = STATE(1965),
    [sym_type_qualifier] = STATE(1965),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1965),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1843),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1965),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4890),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1814)] = {
    [sym_declaration] = STATE(1843),
    [sym__declaration_modifiers] = STATE(1965),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1965),
    [sym_attribute_declaration] = STATE(1965),
    [sym_ms_declspec_modifier] = STATE(1965),
    [sym_compound_statement] = STATE(2315),
    [sym_storage_class_specifier] = STATE(1965),
    [sym_type_qualifier] = STATE(1965),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1965),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1843),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1965),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1815),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4890),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1815)] = {
    [sym_declaration] = STATE(1846),
    [sym__declaration_modifiers] = STATE(1974),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1974),
    [sym_attribute_declaration] = STATE(1974),
    [sym_ms_declspec_modifier] = STATE(1974),
    [sym_compound_statement] = STATE(2339),
    [sym_storage_class_specifier] = STATE(1974),
    [sym_type_qualifier] = STATE(1974),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1974),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1846),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1974),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4892),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1816)] = {
    [sym_declaration] = STATE(1846),
    [sym__declaration_modifiers] = STATE(1974),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1974),
    [sym_attribute_declaration] = STATE(1974),
    [sym_ms_declspec_modifier] = STATE(1974),
    [sym_compound_statement] = STATE(2339),
    [sym_storage_class_specifier] = STATE(1974),
    [sym_type_qualifier] = STATE(1974),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1974),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1846),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1974),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1817),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4892),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1817)] = {
    [sym_declaration] = STATE(1848),
    [sym__declaration_modifiers] = STATE(1919),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1919),
    [sym_attribute_declaration] = STATE(1919),
    [sym_ms_declspec_modifier] = STATE(1919),
    [sym_compound_statement] = STATE(2342),
    [sym_storage_class_specifier] = STATE(1919),
    [sym_type_qualifier] = STATE(1919),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1919),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1848),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1919),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4894),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1818)] = {
    [sym_declaration] = STATE(1845),
    [sym__declaration_modifiers] = STATE(1969),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1969),
    [sym_attribute_declaration] = STATE(1969),
    [sym_ms_declspec_modifier] = STATE(1969),
    [sym_compound_statement] = STATE(2108),
    [sym_storage_class_specifier] = STATE(1969),
    [sym_type_qualifier] = STATE(1969),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1969),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1845),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1969),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(1768),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4874),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1819)] = {
    [sym_declaration] = STATE(1834),
    [sym__declaration_modifiers] = STATE(1931),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1931),
    [sym_attribute_declaration] = STATE(1931),
    [sym_ms_declspec_modifier] = STATE(1931),
    [sym_compound_statement] = STATE(2345),
    [sym_storage_class_specifier] = STATE(1931),
    [sym_type_qualifier] = STATE(1931),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1931),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1834),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1931),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(4862),
    [anon_sym_SEMI] = ACTIONS(4844),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1820)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(6313),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5171),
    [sym_parameter_declaration] = STATE(5567),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4896),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym_RPAREN] = ACTIONS(4898),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1821)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(6762),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5204),
    [sym_parameter_declaration] = STATE(5455),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4896),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym_RPAREN] = ACTIONS(4900),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1822)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(6694),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5447),
    [sym_parameter_declaration] = STATE(5447),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym_RPAREN] = ACTIONS(4902),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1823)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(6762),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5455),
    [sym_parameter_declaration] = STATE(5455),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym_RPAREN] = ACTIONS(4082),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1824)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(6313),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5567),
    [sym_parameter_declaration] = STATE(5567),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym_RPAREN] = ACTIONS(4904),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1825)] = {
    [sym_string_literal] = STATE(1916),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4024),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4024),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4026),
    [anon_sym_extern] = ACTIONS(4026),
    [anon_sym___attribute__] = ACTIONS(4026),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4026),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4026),
    [anon_sym_static] = ACTIONS(4026),
    [anon_sym_EQ] = ACTIONS(4906),
    [anon_sym_auto] = ACTIONS(4026),
    [anon_sym_register] = ACTIONS(4026),
    [anon_sym_inline] = ACTIONS(4026),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4026),
    [anon_sym___forceinline] = ACTIONS(4026),
    [anon_sym_thread_local] = ACTIONS(4026),
    [anon_sym___thread] = ACTIONS(4026),
    [anon_sym_CG_EXTERN] = ACTIONS(4026),
    [anon_sym_CG_INLINE] = ACTIONS(4026),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4026),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4026),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4026),
    [anon_sym_IBOutlet] = ACTIONS(4026),
    [anon_sym_IBInspectable] = ACTIONS(4026),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4026),
    [anon_sym_NS_INLINE] = ACTIONS(4026),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4026),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4026),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4026),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4026),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4026),
    [anon_sym_volatile] = ACTIONS(4026),
    [anon_sym_restrict] = ACTIONS(4026),
    [anon_sym___restrict__] = ACTIONS(4026),
    [anon_sym__Atomic] = ACTIONS(4026),
    [anon_sym__Noreturn] = ACTIONS(4026),
    [anon_sym__Nonnull] = ACTIONS(4026),
    [anon_sym_nullable] = ACTIONS(4026),
    [anon_sym__Complex] = ACTIONS(4026),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4026),
    [anon_sym__Null_unspecified] = ACTIONS(4026),
    [anon_sym___autoreleasing] = ACTIONS(4026),
    [anon_sym___block] = ACTIONS(4026),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4026),
    [anon_sym___bridge_transfer] = ACTIONS(4026),
    [anon_sym___complex] = ACTIONS(4026),
    [anon_sym___const] = ACTIONS(4026),
    [anon_sym___imag] = ACTIONS(4026),
    [anon_sym___kindof] = ACTIONS(4026),
    [anon_sym___nonnull] = ACTIONS(4026),
    [anon_sym___nullable] = ACTIONS(4026),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4026),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4026),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4026),
    [anon_sym___real] = ACTIONS(4026),
    [anon_sym___strong] = ACTIONS(4026),
    [anon_sym___unsafe_unretained] = ACTIONS(4026),
    [anon_sym___unused] = ACTIONS(4026),
    [anon_sym___weak] = ACTIONS(4026),
    [anon_sym_alignas] = ACTIONS(4026),
    [anon_sym__Alignas] = ACTIONS(4026),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4908),
    [anon_sym_SLASH_EQ] = ACTIONS(4908),
    [anon_sym_PERCENT_EQ] = ACTIONS(4908),
    [anon_sym_PLUS_EQ] = ACTIONS(4908),
    [anon_sym_DASH_EQ] = ACTIONS(4908),
    [anon_sym_LT_LT_EQ] = ACTIONS(4908),
    [anon_sym_GT_GT_EQ] = ACTIONS(4908),
    [anon_sym_AMP_EQ] = ACTIONS(4908),
    [anon_sym_CARET_EQ] = ACTIONS(4908),
    [anon_sym_PIPE_EQ] = ACTIONS(4908),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4026),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4026),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4026),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4026),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4026),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4026),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4026),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4026),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4026),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4026),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4026),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4026),
    [anon_sym_API_AVAILABLE] = ACTIONS(4026),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4026),
    [anon_sym_API_DEPRECATED] = ACTIONS(4026),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4026),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4026),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4026),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4026),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4026),
    [anon_sym___deprecated_msg] = ACTIONS(4026),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4026),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4026),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4026),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4026),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4026),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4026),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4026),
  },
  [STATE(1826)] = {
    [sym_declaration] = STATE(1840),
    [sym__declaration_modifiers] = STATE(1935),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1935),
    [sym_attribute_declaration] = STATE(1935),
    [sym_ms_declspec_modifier] = STATE(1935),
    [sym_compound_statement] = STATE(2312),
    [sym_storage_class_specifier] = STATE(1935),
    [sym_type_qualifier] = STATE(1935),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1935),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1840),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1935),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4864),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1827)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1980),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1980),
    [sym_attribute_declaration] = STATE(1980),
    [sym_ms_declspec_modifier] = STATE(1980),
    [sym_compound_statement] = STATE(2297),
    [sym_storage_class_specifier] = STATE(1980),
    [sym_type_qualifier] = STATE(1980),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1980),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1980),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4830),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1828)] = {
    [sym_declaration] = STATE(1831),
    [sym__declaration_modifiers] = STATE(1972),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1972),
    [sym_attribute_declaration] = STATE(1972),
    [sym_ms_declspec_modifier] = STATE(1972),
    [sym_compound_statement] = STATE(2104),
    [sym_storage_class_specifier] = STATE(1972),
    [sym_type_qualifier] = STATE(1972),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1972),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1831),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1972),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4868),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1829)] = {
    [sym_declaration] = STATE(1830),
    [sym__declaration_modifiers] = STATE(1980),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1980),
    [sym_attribute_declaration] = STATE(1980),
    [sym_ms_declspec_modifier] = STATE(1980),
    [sym_compound_statement] = STATE(2297),
    [sym_storage_class_specifier] = STATE(1980),
    [sym_type_qualifier] = STATE(1980),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1980),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1830),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1980),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4830),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1830)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1931),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1931),
    [sym_attribute_declaration] = STATE(1931),
    [sym_ms_declspec_modifier] = STATE(1931),
    [sym_compound_statement] = STATE(2345),
    [sym_storage_class_specifier] = STATE(1931),
    [sym_type_qualifier] = STATE(1931),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1931),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1931),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4844),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1831)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1951),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1951),
    [sym_attribute_declaration] = STATE(1951),
    [sym_ms_declspec_modifier] = STATE(1951),
    [sym_compound_statement] = STATE(2106),
    [sym_storage_class_specifier] = STATE(1951),
    [sym_type_qualifier] = STATE(1951),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1951),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1951),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4870),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1832)] = {
    [sym_declaration] = STATE(1834),
    [sym__declaration_modifiers] = STATE(1931),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1931),
    [sym_attribute_declaration] = STATE(1931),
    [sym_ms_declspec_modifier] = STATE(1931),
    [sym_compound_statement] = STATE(2345),
    [sym_storage_class_specifier] = STATE(1931),
    [sym_type_qualifier] = STATE(1931),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1931),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1834),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1931),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4844),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1833)] = {
    [sym_declaration] = STATE(1836),
    [sym__declaration_modifiers] = STATE(1951),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1951),
    [sym_attribute_declaration] = STATE(1951),
    [sym_ms_declspec_modifier] = STATE(1951),
    [sym_compound_statement] = STATE(2106),
    [sym_storage_class_specifier] = STATE(1951),
    [sym_type_qualifier] = STATE(1951),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1951),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1836),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1951),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4870),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1834)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1975),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1975),
    [sym_attribute_declaration] = STATE(1975),
    [sym_ms_declspec_modifier] = STATE(1975),
    [sym_compound_statement] = STATE(2308),
    [sym_storage_class_specifier] = STATE(1975),
    [sym_type_qualifier] = STATE(1975),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1975),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1975),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4878),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1835)] = {
    [sym_declaration] = STATE(1838),
    [sym__declaration_modifiers] = STATE(1975),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1975),
    [sym_attribute_declaration] = STATE(1975),
    [sym_ms_declspec_modifier] = STATE(1975),
    [sym_compound_statement] = STATE(2308),
    [sym_storage_class_specifier] = STATE(1975),
    [sym_type_qualifier] = STATE(1975),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1975),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1838),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1975),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4878),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1836)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1925),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1925),
    [sym_attribute_declaration] = STATE(1925),
    [sym_ms_declspec_modifier] = STATE(1925),
    [sym_compound_statement] = STATE(2107),
    [sym_storage_class_specifier] = STATE(1925),
    [sym_type_qualifier] = STATE(1925),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1925),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1925),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4872),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1837)] = {
    [sym_declaration] = STATE(1839),
    [sym__declaration_modifiers] = STATE(1925),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1925),
    [sym_attribute_declaration] = STATE(1925),
    [sym_ms_declspec_modifier] = STATE(1925),
    [sym_compound_statement] = STATE(2107),
    [sym_storage_class_specifier] = STATE(1925),
    [sym_type_qualifier] = STATE(1925),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1925),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1839),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1925),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4872),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1838)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1935),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1935),
    [sym_attribute_declaration] = STATE(1935),
    [sym_ms_declspec_modifier] = STATE(1935),
    [sym_compound_statement] = STATE(2312),
    [sym_storage_class_specifier] = STATE(1935),
    [sym_type_qualifier] = STATE(1935),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1935),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1935),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4864),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1839)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1969),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1969),
    [sym_attribute_declaration] = STATE(1969),
    [sym_ms_declspec_modifier] = STATE(1969),
    [sym_compound_statement] = STATE(2108),
    [sym_storage_class_specifier] = STATE(1969),
    [sym_type_qualifier] = STATE(1969),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1969),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1969),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4874),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1840)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1965),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1965),
    [sym_attribute_declaration] = STATE(1965),
    [sym_ms_declspec_modifier] = STATE(1965),
    [sym_compound_statement] = STATE(2315),
    [sym_storage_class_specifier] = STATE(1965),
    [sym_type_qualifier] = STATE(1965),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1965),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1965),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4890),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1841)] = {
    [sym_declaration] = STATE(1845),
    [sym__declaration_modifiers] = STATE(1969),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1969),
    [sym_attribute_declaration] = STATE(1969),
    [sym_ms_declspec_modifier] = STATE(1969),
    [sym_compound_statement] = STATE(2108),
    [sym_storage_class_specifier] = STATE(1969),
    [sym_type_qualifier] = STATE(1969),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1969),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1845),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1969),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4874),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1842)] = {
    [sym_declaration] = STATE(1843),
    [sym__declaration_modifiers] = STATE(1965),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1965),
    [sym_attribute_declaration] = STATE(1965),
    [sym_ms_declspec_modifier] = STATE(1965),
    [sym_compound_statement] = STATE(2315),
    [sym_storage_class_specifier] = STATE(1965),
    [sym_type_qualifier] = STATE(1965),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1965),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1843),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1965),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4890),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1843)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1974),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1974),
    [sym_attribute_declaration] = STATE(1974),
    [sym_ms_declspec_modifier] = STATE(1974),
    [sym_compound_statement] = STATE(2339),
    [sym_storage_class_specifier] = STATE(1974),
    [sym_type_qualifier] = STATE(1974),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1974),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1974),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4892),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1844)] = {
    [sym_declaration] = STATE(1846),
    [sym__declaration_modifiers] = STATE(1974),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1974),
    [sym_attribute_declaration] = STATE(1974),
    [sym_ms_declspec_modifier] = STATE(1974),
    [sym_compound_statement] = STATE(2339),
    [sym_storage_class_specifier] = STATE(1974),
    [sym_type_qualifier] = STATE(1974),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1974),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1846),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1974),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4892),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1845)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1924),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1924),
    [sym_attribute_declaration] = STATE(1924),
    [sym_ms_declspec_modifier] = STATE(1924),
    [sym_compound_statement] = STATE(2109),
    [sym_storage_class_specifier] = STATE(1924),
    [sym_type_qualifier] = STATE(1924),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1924),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1924),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4866),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1846)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1919),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1919),
    [sym_attribute_declaration] = STATE(1919),
    [sym_ms_declspec_modifier] = STATE(1919),
    [sym_compound_statement] = STATE(2342),
    [sym_storage_class_specifier] = STATE(1919),
    [sym_type_qualifier] = STATE(1919),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1919),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1919),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4894),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1847)] = {
    [sym_declaration] = STATE(1858),
    [sym__declaration_modifiers] = STATE(1926),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1926),
    [sym_attribute_declaration] = STATE(1926),
    [sym_ms_declspec_modifier] = STATE(1926),
    [sym_compound_statement] = STATE(2252),
    [sym_storage_class_specifier] = STATE(1926),
    [sym_type_qualifier] = STATE(1926),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1926),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1858),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1926),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4836),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1848)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1920),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1920),
    [sym_attribute_declaration] = STATE(1920),
    [sym_ms_declspec_modifier] = STATE(1920),
    [sym_compound_statement] = STATE(2295),
    [sym_storage_class_specifier] = STATE(1920),
    [sym_type_qualifier] = STATE(1920),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1920),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1920),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4910),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1849)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1927),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1927),
    [sym_attribute_declaration] = STATE(1927),
    [sym_ms_declspec_modifier] = STATE(1927),
    [sym_compound_statement] = STATE(2110),
    [sym_storage_class_specifier] = STATE(1927),
    [sym_type_qualifier] = STATE(1927),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1927),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1927),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4912),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1850)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1972),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1972),
    [sym_attribute_declaration] = STATE(1972),
    [sym_ms_declspec_modifier] = STATE(1972),
    [sym_compound_statement] = STATE(2104),
    [sym_storage_class_specifier] = STATE(1972),
    [sym_type_qualifier] = STATE(1972),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1972),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1972),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4868),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1851)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1982),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1982),
    [sym_attribute_declaration] = STATE(1982),
    [sym_ms_declspec_modifier] = STATE(1982),
    [sym_compound_statement] = STATE(2163),
    [sym_storage_class_specifier] = STATE(1982),
    [sym_type_qualifier] = STATE(1982),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1982),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1982),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4914),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1852)] = {
    [sym_declaration] = STATE(1860),
    [sym__declaration_modifiers] = STATE(1930),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1930),
    [sym_attribute_declaration] = STATE(1930),
    [sym_ms_declspec_modifier] = STATE(1930),
    [sym_compound_statement] = STATE(2260),
    [sym_storage_class_specifier] = STATE(1930),
    [sym_type_qualifier] = STATE(1930),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1930),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1860),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1930),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4815),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1853)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1926),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1926),
    [sym_attribute_declaration] = STATE(1926),
    [sym_ms_declspec_modifier] = STATE(1926),
    [sym_compound_statement] = STATE(2252),
    [sym_storage_class_specifier] = STATE(1926),
    [sym_type_qualifier] = STATE(1926),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1926),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1926),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4836),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1854)] = {
    [sym_declaration] = STATE(1862),
    [sym__declaration_modifiers] = STATE(1933),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1933),
    [sym_attribute_declaration] = STATE(1933),
    [sym_ms_declspec_modifier] = STATE(1933),
    [sym_compound_statement] = STATE(2140),
    [sym_storage_class_specifier] = STATE(1933),
    [sym_type_qualifier] = STATE(1933),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1933),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1862),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1933),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4854),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1855)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1933),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1933),
    [sym_attribute_declaration] = STATE(1933),
    [sym_ms_declspec_modifier] = STATE(1933),
    [sym_compound_statement] = STATE(2140),
    [sym_storage_class_specifier] = STATE(1933),
    [sym_type_qualifier] = STATE(1933),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1933),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1933),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4854),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1856)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1954),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1954),
    [sym_attribute_declaration] = STATE(1954),
    [sym_ms_declspec_modifier] = STATE(1954),
    [sym_compound_statement] = STATE(2268),
    [sym_storage_class_specifier] = STATE(1954),
    [sym_type_qualifier] = STATE(1954),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1954),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1954),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4916),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1857)] = {
    [sym_declaration] = STATE(1850),
    [sym__declaration_modifiers] = STATE(1971),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1971),
    [sym_attribute_declaration] = STATE(1971),
    [sym_ms_declspec_modifier] = STATE(1971),
    [sym_compound_statement] = STATE(2101),
    [sym_storage_class_specifier] = STATE(1971),
    [sym_type_qualifier] = STATE(1971),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1850),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1971),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4850),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1858)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1932),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1932),
    [sym_attribute_declaration] = STATE(1932),
    [sym_ms_declspec_modifier] = STATE(1932),
    [sym_compound_statement] = STATE(2309),
    [sym_storage_class_specifier] = STATE(1932),
    [sym_type_qualifier] = STATE(1932),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1932),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1932),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4842),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1859)] = {
    [sym_declaration] = STATE(1876),
    [sym__declaration_modifiers] = STATE(1932),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1932),
    [sym_attribute_declaration] = STATE(1932),
    [sym_ms_declspec_modifier] = STATE(1932),
    [sym_compound_statement] = STATE(2309),
    [sym_storage_class_specifier] = STATE(1932),
    [sym_type_qualifier] = STATE(1932),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1932),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1876),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1932),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4842),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1860)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1937),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1937),
    [sym_attribute_declaration] = STATE(1937),
    [sym_ms_declspec_modifier] = STATE(1937),
    [sym_compound_statement] = STATE(2271),
    [sym_storage_class_specifier] = STATE(1937),
    [sym_type_qualifier] = STATE(1937),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1937),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1937),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4876),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1861)] = {
    [sym_declaration] = STATE(1863),
    [sym__declaration_modifiers] = STATE(1937),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1937),
    [sym_attribute_declaration] = STATE(1937),
    [sym_ms_declspec_modifier] = STATE(1937),
    [sym_compound_statement] = STATE(2271),
    [sym_storage_class_specifier] = STATE(1937),
    [sym_type_qualifier] = STATE(1937),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1937),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1863),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1937),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4876),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1862)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1955),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1955),
    [sym_attribute_declaration] = STATE(1955),
    [sym_ms_declspec_modifier] = STATE(1955),
    [sym_compound_statement] = STATE(2186),
    [sym_storage_class_specifier] = STATE(1955),
    [sym_type_qualifier] = STATE(1955),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1955),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1955),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4834),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1863)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1950),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1950),
    [sym_attribute_declaration] = STATE(1950),
    [sym_ms_declspec_modifier] = STATE(1950),
    [sym_compound_statement] = STATE(2277),
    [sym_storage_class_specifier] = STATE(1950),
    [sym_type_qualifier] = STATE(1950),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1950),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1950),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4880),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1864)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1979),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1979),
    [sym_attribute_declaration] = STATE(1979),
    [sym_ms_declspec_modifier] = STATE(1979),
    [sym_compound_statement] = STATE(2334),
    [sym_storage_class_specifier] = STATE(1979),
    [sym_type_qualifier] = STATE(1979),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1979),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1979),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4918),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1865)] = {
    [sym_declaration] = STATE(1867),
    [sym__declaration_modifiers] = STATE(1950),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1950),
    [sym_attribute_declaration] = STATE(1950),
    [sym_ms_declspec_modifier] = STATE(1950),
    [sym_compound_statement] = STATE(2277),
    [sym_storage_class_specifier] = STATE(1950),
    [sym_type_qualifier] = STATE(1950),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1950),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1867),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1950),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4880),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1866)] = {
    [sym_declaration] = STATE(1873),
    [sym__declaration_modifiers] = STATE(1955),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1955),
    [sym_attribute_declaration] = STATE(1955),
    [sym_ms_declspec_modifier] = STATE(1955),
    [sym_compound_statement] = STATE(2186),
    [sym_storage_class_specifier] = STATE(1955),
    [sym_type_qualifier] = STATE(1955),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1955),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1873),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1955),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4834),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1867)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1953),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1953),
    [sym_attribute_declaration] = STATE(1953),
    [sym_ms_declspec_modifier] = STATE(1953),
    [sym_compound_statement] = STATE(2283),
    [sym_storage_class_specifier] = STATE(1953),
    [sym_type_qualifier] = STATE(1953),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1953),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1953),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4884),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1868)] = {
    [sym_declaration] = STATE(1869),
    [sym__declaration_modifiers] = STATE(1953),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1953),
    [sym_attribute_declaration] = STATE(1953),
    [sym_ms_declspec_modifier] = STATE(1953),
    [sym_compound_statement] = STATE(2283),
    [sym_storage_class_specifier] = STATE(1953),
    [sym_type_qualifier] = STATE(1953),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1953),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1869),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1953),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4884),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1869)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1956),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1956),
    [sym_attribute_declaration] = STATE(1956),
    [sym_ms_declspec_modifier] = STATE(1956),
    [sym_compound_statement] = STATE(2289),
    [sym_storage_class_specifier] = STATE(1956),
    [sym_type_qualifier] = STATE(1956),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1956),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1956),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4886),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1870)] = {
    [sym_declaration] = STATE(1871),
    [sym__declaration_modifiers] = STATE(1956),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1956),
    [sym_attribute_declaration] = STATE(1956),
    [sym_ms_declspec_modifier] = STATE(1956),
    [sym_compound_statement] = STATE(2289),
    [sym_storage_class_specifier] = STATE(1956),
    [sym_type_qualifier] = STATE(1956),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1956),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1871),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1956),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4886),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1871)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1957),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1957),
    [sym_attribute_declaration] = STATE(1957),
    [sym_ms_declspec_modifier] = STATE(1957),
    [sym_compound_statement] = STATE(2299),
    [sym_storage_class_specifier] = STATE(1957),
    [sym_type_qualifier] = STATE(1957),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1957),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1957),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4888),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1872)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1964),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1964),
    [sym_attribute_declaration] = STATE(1964),
    [sym_ms_declspec_modifier] = STATE(1964),
    [sym_compound_statement] = STATE(2301),
    [sym_storage_class_specifier] = STATE(1964),
    [sym_type_qualifier] = STATE(1964),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1964),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1964),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4920),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1873)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1971),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1971),
    [sym_attribute_declaration] = STATE(1971),
    [sym_ms_declspec_modifier] = STATE(1971),
    [sym_compound_statement] = STATE(2101),
    [sym_storage_class_specifier] = STATE(1971),
    [sym_type_qualifier] = STATE(1971),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1971),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4850),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1874)] = {
    [sym_declaration] = STATE(1827),
    [sym__declaration_modifiers] = STATE(1966),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1966),
    [sym_attribute_declaration] = STATE(1966),
    [sym_ms_declspec_modifier] = STATE(1966),
    [sym_compound_statement] = STATE(2262),
    [sym_storage_class_specifier] = STATE(1966),
    [sym_type_qualifier] = STATE(1966),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1966),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1827),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1966),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4811),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1875)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1966),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1966),
    [sym_attribute_declaration] = STATE(1966),
    [sym_ms_declspec_modifier] = STATE(1966),
    [sym_compound_statement] = STATE(2262),
    [sym_storage_class_specifier] = STATE(1966),
    [sym_type_qualifier] = STATE(1966),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1966),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1966),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4811),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1876)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(1930),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(1930),
    [sym_attribute_declaration] = STATE(1930),
    [sym_ms_declspec_modifier] = STATE(1930),
    [sym_compound_statement] = STATE(2260),
    [sym_storage_class_specifier] = STATE(1930),
    [sym_type_qualifier] = STATE(1930),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1930),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1930),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4815),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1877)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(984),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1878)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(476),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1879)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(335),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1880)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(865),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1881)] = {
    [sym_identifier] = ACTIONS(4922),
    [aux_sym_preproc_def_token1] = ACTIONS(4922),
    [anon_sym_COMMA] = ACTIONS(4924),
    [anon_sym_RPAREN] = ACTIONS(4924),
    [aux_sym_preproc_if_token1] = ACTIONS(4922),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4922),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4922),
    [sym_preproc_directive] = ACTIONS(4922),
    [anon_sym_LPAREN2] = ACTIONS(4924),
    [anon_sym_DASH] = ACTIONS(4924),
    [anon_sym_PLUS] = ACTIONS(4924),
    [anon_sym_LT] = ACTIONS(4924),
    [anon_sym_SEMI] = ACTIONS(4924),
    [anon_sym___extension__] = ACTIONS(4922),
    [anon_sym_typedef] = ACTIONS(4922),
    [anon_sym_extern] = ACTIONS(4922),
    [anon_sym___attribute__] = ACTIONS(4922),
    [anon_sym___attribute] = ACTIONS(4922),
    [anon_sym_noreturn] = ACTIONS(4922),
    [anon_sym_LBRACK] = ACTIONS(4924),
    [anon_sym___declspec] = ACTIONS(4922),
    [anon_sym___cdecl] = ACTIONS(4922),
    [anon_sym___clrcall] = ACTIONS(4922),
    [anon_sym___stdcall] = ACTIONS(4922),
    [anon_sym___fastcall] = ACTIONS(4922),
    [anon_sym___thiscall] = ACTIONS(4922),
    [anon_sym___vectorcall] = ACTIONS(4922),
    [anon_sym_LBRACE] = ACTIONS(4924),
    [anon_sym_signed] = ACTIONS(4922),
    [anon_sym_unsigned] = ACTIONS(4922),
    [anon_sym_long] = ACTIONS(4922),
    [anon_sym_short] = ACTIONS(4922),
    [anon_sym_static] = ACTIONS(4922),
    [anon_sym_auto] = ACTIONS(4922),
    [anon_sym_register] = ACTIONS(4922),
    [anon_sym_inline] = ACTIONS(4922),
    [anon_sym___inline] = ACTIONS(4922),
    [anon_sym___inline__] = ACTIONS(4922),
    [anon_sym___forceinline] = ACTIONS(4922),
    [anon_sym_thread_local] = ACTIONS(4922),
    [anon_sym___thread] = ACTIONS(4922),
    [anon_sym_CG_EXTERN] = ACTIONS(4922),
    [anon_sym_CG_INLINE] = ACTIONS(4922),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4922),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4922),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4922),
    [anon_sym_IBOutlet] = ACTIONS(4922),
    [anon_sym_IBInspectable] = ACTIONS(4922),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4922),
    [anon_sym_NS_INLINE] = ACTIONS(4922),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4922),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4922),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4922),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4922),
    [anon_sym_const] = ACTIONS(4922),
    [anon_sym_constexpr] = ACTIONS(4922),
    [anon_sym_volatile] = ACTIONS(4922),
    [anon_sym_restrict] = ACTIONS(4922),
    [anon_sym___restrict__] = ACTIONS(4922),
    [anon_sym__Atomic] = ACTIONS(4922),
    [anon_sym__Noreturn] = ACTIONS(4922),
    [anon_sym__Nonnull] = ACTIONS(4922),
    [anon_sym_nullable] = ACTIONS(4922),
    [anon_sym__Complex] = ACTIONS(4922),
    [anon_sym__Nullable] = ACTIONS(4922),
    [anon_sym__Nullable_result] = ACTIONS(4922),
    [anon_sym__Null_unspecified] = ACTIONS(4922),
    [anon_sym___autoreleasing] = ACTIONS(4922),
    [anon_sym___block] = ACTIONS(4922),
    [anon_sym___bridge] = ACTIONS(4922),
    [anon_sym___bridge_retained] = ACTIONS(4922),
    [anon_sym___bridge_transfer] = ACTIONS(4922),
    [anon_sym___complex] = ACTIONS(4922),
    [anon_sym___const] = ACTIONS(4922),
    [anon_sym___imag] = ACTIONS(4922),
    [anon_sym___kindof] = ACTIONS(4922),
    [anon_sym___nonnull] = ACTIONS(4922),
    [anon_sym___nullable] = ACTIONS(4922),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4922),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4922),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4922),
    [anon_sym___real] = ACTIONS(4922),
    [anon_sym___strong] = ACTIONS(4922),
    [anon_sym___unsafe_unretained] = ACTIONS(4922),
    [anon_sym___unused] = ACTIONS(4922),
    [anon_sym___weak] = ACTIONS(4922),
    [anon_sym_alignas] = ACTIONS(4922),
    [anon_sym__Alignas] = ACTIONS(4922),
    [sym_primitive_type] = ACTIONS(4922),
    [anon_sym_enum] = ACTIONS(4922),
    [anon_sym_COLON] = ACTIONS(4924),
    [anon_sym_struct] = ACTIONS(4922),
    [anon_sym_union] = ACTIONS(4922),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4922),
    [anon_sym___typeof] = ACTIONS(4922),
    [anon_sym_typeof] = ACTIONS(4922),
    [aux_sym_preproc_undef_token1] = ACTIONS(4922),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4922),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4922),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4922),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4922),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4922),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4922),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4922),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4922),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4922),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4922),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4922),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4922),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4922),
    [anon_sym_API_AVAILABLE] = ACTIONS(4922),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4922),
    [anon_sym_API_DEPRECATED] = ACTIONS(4922),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4922),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4922),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4922),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4922),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4922),
    [anon_sym___deprecated_msg] = ACTIONS(4922),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4922),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4922),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4922),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4922),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4922),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4922),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4922),
    [anon_sym_ATend] = ACTIONS(4924),
    [anon_sym_ATsynthesize] = ACTIONS(4924),
    [anon_sym_ATdynamic] = ACTIONS(4924),
    [anon_sym_ATproperty] = ACTIONS(4924),
    [anon_sym_BOOL] = ACTIONS(4922),
    [anon_sym_IMP] = ACTIONS(4922),
    [anon_sym_SEL] = ACTIONS(4922),
    [anon_sym_Class] = ACTIONS(4922),
    [anon_sym_id] = ACTIONS(4922),
  },
  [STATE(1882)] = {
    [sym_identifier] = ACTIONS(4926),
    [aux_sym_preproc_def_token1] = ACTIONS(4926),
    [anon_sym_COMMA] = ACTIONS(4928),
    [anon_sym_RPAREN] = ACTIONS(4928),
    [aux_sym_preproc_if_token1] = ACTIONS(4926),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4926),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4926),
    [sym_preproc_directive] = ACTIONS(4926),
    [anon_sym_LPAREN2] = ACTIONS(4928),
    [anon_sym_DASH] = ACTIONS(4928),
    [anon_sym_PLUS] = ACTIONS(4928),
    [anon_sym_STAR] = ACTIONS(4928),
    [anon_sym_CARET] = ACTIONS(4928),
    [anon_sym_SEMI] = ACTIONS(4928),
    [anon_sym___extension__] = ACTIONS(4926),
    [anon_sym_typedef] = ACTIONS(4926),
    [anon_sym_extern] = ACTIONS(4926),
    [anon_sym___attribute__] = ACTIONS(4926),
    [anon_sym___attribute] = ACTIONS(4926),
    [anon_sym_noreturn] = ACTIONS(4926),
    [anon_sym_LBRACK] = ACTIONS(4928),
    [anon_sym___declspec] = ACTIONS(4926),
    [anon_sym___based] = ACTIONS(4926),
    [anon_sym___cdecl] = ACTIONS(4926),
    [anon_sym___clrcall] = ACTIONS(4926),
    [anon_sym___stdcall] = ACTIONS(4926),
    [anon_sym___fastcall] = ACTIONS(4926),
    [anon_sym___thiscall] = ACTIONS(4926),
    [anon_sym___vectorcall] = ACTIONS(4926),
    [anon_sym_signed] = ACTIONS(4926),
    [anon_sym_unsigned] = ACTIONS(4926),
    [anon_sym_long] = ACTIONS(4926),
    [anon_sym_short] = ACTIONS(4926),
    [anon_sym_static] = ACTIONS(4926),
    [anon_sym_auto] = ACTIONS(4926),
    [anon_sym_register] = ACTIONS(4926),
    [anon_sym_inline] = ACTIONS(4926),
    [anon_sym___inline] = ACTIONS(4926),
    [anon_sym___inline__] = ACTIONS(4926),
    [anon_sym___forceinline] = ACTIONS(4926),
    [anon_sym_thread_local] = ACTIONS(4926),
    [anon_sym___thread] = ACTIONS(4926),
    [anon_sym_CG_EXTERN] = ACTIONS(4926),
    [anon_sym_CG_INLINE] = ACTIONS(4926),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4926),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4926),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4926),
    [anon_sym_IBOutlet] = ACTIONS(4926),
    [anon_sym_IBInspectable] = ACTIONS(4926),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4926),
    [anon_sym_NS_INLINE] = ACTIONS(4926),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4926),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4926),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4926),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4926),
    [anon_sym_const] = ACTIONS(4926),
    [anon_sym_constexpr] = ACTIONS(4926),
    [anon_sym_volatile] = ACTIONS(4926),
    [anon_sym_restrict] = ACTIONS(4926),
    [anon_sym___restrict__] = ACTIONS(4926),
    [anon_sym__Atomic] = ACTIONS(4926),
    [anon_sym__Noreturn] = ACTIONS(4926),
    [anon_sym__Nonnull] = ACTIONS(4926),
    [anon_sym_nullable] = ACTIONS(4926),
    [anon_sym__Complex] = ACTIONS(4926),
    [anon_sym__Nullable] = ACTIONS(4926),
    [anon_sym__Nullable_result] = ACTIONS(4926),
    [anon_sym__Null_unspecified] = ACTIONS(4926),
    [anon_sym___autoreleasing] = ACTIONS(4926),
    [anon_sym___block] = ACTIONS(4926),
    [anon_sym___bridge] = ACTIONS(4926),
    [anon_sym___bridge_retained] = ACTIONS(4926),
    [anon_sym___bridge_transfer] = ACTIONS(4926),
    [anon_sym___complex] = ACTIONS(4926),
    [anon_sym___const] = ACTIONS(4926),
    [anon_sym___imag] = ACTIONS(4926),
    [anon_sym___kindof] = ACTIONS(4926),
    [anon_sym___nonnull] = ACTIONS(4926),
    [anon_sym___nullable] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4926),
    [anon_sym___real] = ACTIONS(4926),
    [anon_sym___strong] = ACTIONS(4926),
    [anon_sym___unsafe_unretained] = ACTIONS(4926),
    [anon_sym___unused] = ACTIONS(4926),
    [anon_sym___weak] = ACTIONS(4926),
    [anon_sym_alignas] = ACTIONS(4926),
    [anon_sym__Alignas] = ACTIONS(4926),
    [sym_primitive_type] = ACTIONS(4926),
    [anon_sym_enum] = ACTIONS(4926),
    [anon_sym_COLON] = ACTIONS(4928),
    [anon_sym_struct] = ACTIONS(4926),
    [anon_sym_union] = ACTIONS(4926),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4926),
    [anon_sym___typeof] = ACTIONS(4926),
    [anon_sym_typeof] = ACTIONS(4926),
    [aux_sym_preproc_undef_token1] = ACTIONS(4926),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4926),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4926),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4926),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4926),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4926),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4926),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4926),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4926),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4926),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_API_AVAILABLE] = ACTIONS(4926),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_API_DEPRECATED] = ACTIONS(4926),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4926),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4926),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4926),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4926),
    [anon_sym___deprecated_msg] = ACTIONS(4926),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4926),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4926),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4926),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4926),
    [anon_sym_ATend] = ACTIONS(4928),
    [anon_sym_ATsynthesize] = ACTIONS(4928),
    [anon_sym_ATdynamic] = ACTIONS(4928),
    [anon_sym_BOOL] = ACTIONS(4926),
    [anon_sym_IMP] = ACTIONS(4926),
    [anon_sym_SEL] = ACTIONS(4926),
    [anon_sym_Class] = ACTIONS(4926),
    [anon_sym_id] = ACTIONS(4926),
  },
  [STATE(1883)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(872),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1884)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(452),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1885)] = {
    [sym_identifier] = ACTIONS(4930),
    [aux_sym_preproc_def_token1] = ACTIONS(4930),
    [anon_sym_COMMA] = ACTIONS(4932),
    [anon_sym_RPAREN] = ACTIONS(4932),
    [aux_sym_preproc_if_token1] = ACTIONS(4930),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4930),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4930),
    [sym_preproc_directive] = ACTIONS(4930),
    [anon_sym_LPAREN2] = ACTIONS(4932),
    [anon_sym_DASH] = ACTIONS(4932),
    [anon_sym_PLUS] = ACTIONS(4932),
    [anon_sym_LT] = ACTIONS(4932),
    [anon_sym_SEMI] = ACTIONS(4932),
    [anon_sym___extension__] = ACTIONS(4930),
    [anon_sym_typedef] = ACTIONS(4930),
    [anon_sym_extern] = ACTIONS(4930),
    [anon_sym___attribute__] = ACTIONS(4930),
    [anon_sym___attribute] = ACTIONS(4930),
    [anon_sym_noreturn] = ACTIONS(4930),
    [anon_sym_LBRACK] = ACTIONS(4932),
    [anon_sym___declspec] = ACTIONS(4930),
    [anon_sym___cdecl] = ACTIONS(4930),
    [anon_sym___clrcall] = ACTIONS(4930),
    [anon_sym___stdcall] = ACTIONS(4930),
    [anon_sym___fastcall] = ACTIONS(4930),
    [anon_sym___thiscall] = ACTIONS(4930),
    [anon_sym___vectorcall] = ACTIONS(4930),
    [anon_sym_LBRACE] = ACTIONS(4932),
    [anon_sym_signed] = ACTIONS(4930),
    [anon_sym_unsigned] = ACTIONS(4930),
    [anon_sym_long] = ACTIONS(4930),
    [anon_sym_short] = ACTIONS(4930),
    [anon_sym_static] = ACTIONS(4930),
    [anon_sym_auto] = ACTIONS(4930),
    [anon_sym_register] = ACTIONS(4930),
    [anon_sym_inline] = ACTIONS(4930),
    [anon_sym___inline] = ACTIONS(4930),
    [anon_sym___inline__] = ACTIONS(4930),
    [anon_sym___forceinline] = ACTIONS(4930),
    [anon_sym_thread_local] = ACTIONS(4930),
    [anon_sym___thread] = ACTIONS(4930),
    [anon_sym_CG_EXTERN] = ACTIONS(4930),
    [anon_sym_CG_INLINE] = ACTIONS(4930),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4930),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4930),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4930),
    [anon_sym_IBOutlet] = ACTIONS(4930),
    [anon_sym_IBInspectable] = ACTIONS(4930),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4930),
    [anon_sym_NS_INLINE] = ACTIONS(4930),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4930),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4930),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4930),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4930),
    [anon_sym_const] = ACTIONS(4930),
    [anon_sym_constexpr] = ACTIONS(4930),
    [anon_sym_volatile] = ACTIONS(4930),
    [anon_sym_restrict] = ACTIONS(4930),
    [anon_sym___restrict__] = ACTIONS(4930),
    [anon_sym__Atomic] = ACTIONS(4930),
    [anon_sym__Noreturn] = ACTIONS(4930),
    [anon_sym__Nonnull] = ACTIONS(4930),
    [anon_sym_nullable] = ACTIONS(4930),
    [anon_sym__Complex] = ACTIONS(4930),
    [anon_sym__Nullable] = ACTIONS(4930),
    [anon_sym__Nullable_result] = ACTIONS(4930),
    [anon_sym__Null_unspecified] = ACTIONS(4930),
    [anon_sym___autoreleasing] = ACTIONS(4930),
    [anon_sym___block] = ACTIONS(4930),
    [anon_sym___bridge] = ACTIONS(4930),
    [anon_sym___bridge_retained] = ACTIONS(4930),
    [anon_sym___bridge_transfer] = ACTIONS(4930),
    [anon_sym___complex] = ACTIONS(4930),
    [anon_sym___const] = ACTIONS(4930),
    [anon_sym___imag] = ACTIONS(4930),
    [anon_sym___kindof] = ACTIONS(4930),
    [anon_sym___nonnull] = ACTIONS(4930),
    [anon_sym___nullable] = ACTIONS(4930),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4930),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4930),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4930),
    [anon_sym___real] = ACTIONS(4930),
    [anon_sym___strong] = ACTIONS(4930),
    [anon_sym___unsafe_unretained] = ACTIONS(4930),
    [anon_sym___unused] = ACTIONS(4930),
    [anon_sym___weak] = ACTIONS(4930),
    [anon_sym_alignas] = ACTIONS(4930),
    [anon_sym__Alignas] = ACTIONS(4930),
    [sym_primitive_type] = ACTIONS(4930),
    [anon_sym_enum] = ACTIONS(4930),
    [anon_sym_COLON] = ACTIONS(4932),
    [anon_sym_struct] = ACTIONS(4930),
    [anon_sym_union] = ACTIONS(4930),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4930),
    [anon_sym___typeof] = ACTIONS(4930),
    [anon_sym_typeof] = ACTIONS(4930),
    [aux_sym_preproc_undef_token1] = ACTIONS(4930),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4930),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4930),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4930),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4930),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4930),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4930),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4930),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4930),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4930),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4930),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4930),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4930),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4930),
    [anon_sym_API_AVAILABLE] = ACTIONS(4930),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4930),
    [anon_sym_API_DEPRECATED] = ACTIONS(4930),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4930),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4930),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4930),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4930),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4930),
    [anon_sym___deprecated_msg] = ACTIONS(4930),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4930),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4930),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4930),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4930),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4930),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4930),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4930),
    [anon_sym_ATend] = ACTIONS(4932),
    [anon_sym_ATsynthesize] = ACTIONS(4932),
    [anon_sym_ATdynamic] = ACTIONS(4932),
    [anon_sym_ATproperty] = ACTIONS(4932),
    [anon_sym_BOOL] = ACTIONS(4930),
    [anon_sym_IMP] = ACTIONS(4930),
    [anon_sym_SEL] = ACTIONS(4930),
    [anon_sym_Class] = ACTIONS(4930),
    [anon_sym_id] = ACTIONS(4930),
  },
  [STATE(1886)] = {
    [sym_identifier] = ACTIONS(4934),
    [aux_sym_preproc_def_token1] = ACTIONS(4934),
    [anon_sym_COMMA] = ACTIONS(4936),
    [anon_sym_RPAREN] = ACTIONS(4936),
    [aux_sym_preproc_if_token1] = ACTIONS(4934),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4934),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4934),
    [sym_preproc_directive] = ACTIONS(4934),
    [anon_sym_LPAREN2] = ACTIONS(4936),
    [anon_sym_DASH] = ACTIONS(4936),
    [anon_sym_PLUS] = ACTIONS(4936),
    [anon_sym_LT] = ACTIONS(4936),
    [anon_sym_SEMI] = ACTIONS(4936),
    [anon_sym___extension__] = ACTIONS(4934),
    [anon_sym_typedef] = ACTIONS(4934),
    [anon_sym_extern] = ACTIONS(4934),
    [anon_sym___attribute__] = ACTIONS(4934),
    [anon_sym___attribute] = ACTIONS(4934),
    [anon_sym_noreturn] = ACTIONS(4934),
    [anon_sym_LBRACK] = ACTIONS(4936),
    [anon_sym___declspec] = ACTIONS(4934),
    [anon_sym___cdecl] = ACTIONS(4934),
    [anon_sym___clrcall] = ACTIONS(4934),
    [anon_sym___stdcall] = ACTIONS(4934),
    [anon_sym___fastcall] = ACTIONS(4934),
    [anon_sym___thiscall] = ACTIONS(4934),
    [anon_sym___vectorcall] = ACTIONS(4934),
    [anon_sym_LBRACE] = ACTIONS(4936),
    [anon_sym_signed] = ACTIONS(4934),
    [anon_sym_unsigned] = ACTIONS(4934),
    [anon_sym_long] = ACTIONS(4934),
    [anon_sym_short] = ACTIONS(4934),
    [anon_sym_static] = ACTIONS(4934),
    [anon_sym_auto] = ACTIONS(4934),
    [anon_sym_register] = ACTIONS(4934),
    [anon_sym_inline] = ACTIONS(4934),
    [anon_sym___inline] = ACTIONS(4934),
    [anon_sym___inline__] = ACTIONS(4934),
    [anon_sym___forceinline] = ACTIONS(4934),
    [anon_sym_thread_local] = ACTIONS(4934),
    [anon_sym___thread] = ACTIONS(4934),
    [anon_sym_CG_EXTERN] = ACTIONS(4934),
    [anon_sym_CG_INLINE] = ACTIONS(4934),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4934),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4934),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4934),
    [anon_sym_IBOutlet] = ACTIONS(4934),
    [anon_sym_IBInspectable] = ACTIONS(4934),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4934),
    [anon_sym_NS_INLINE] = ACTIONS(4934),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4934),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4934),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4934),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4934),
    [anon_sym_const] = ACTIONS(4934),
    [anon_sym_constexpr] = ACTIONS(4934),
    [anon_sym_volatile] = ACTIONS(4934),
    [anon_sym_restrict] = ACTIONS(4934),
    [anon_sym___restrict__] = ACTIONS(4934),
    [anon_sym__Atomic] = ACTIONS(4934),
    [anon_sym__Noreturn] = ACTIONS(4934),
    [anon_sym__Nonnull] = ACTIONS(4934),
    [anon_sym_nullable] = ACTIONS(4934),
    [anon_sym__Complex] = ACTIONS(4934),
    [anon_sym__Nullable] = ACTIONS(4934),
    [anon_sym__Nullable_result] = ACTIONS(4934),
    [anon_sym__Null_unspecified] = ACTIONS(4934),
    [anon_sym___autoreleasing] = ACTIONS(4934),
    [anon_sym___block] = ACTIONS(4934),
    [anon_sym___bridge] = ACTIONS(4934),
    [anon_sym___bridge_retained] = ACTIONS(4934),
    [anon_sym___bridge_transfer] = ACTIONS(4934),
    [anon_sym___complex] = ACTIONS(4934),
    [anon_sym___const] = ACTIONS(4934),
    [anon_sym___imag] = ACTIONS(4934),
    [anon_sym___kindof] = ACTIONS(4934),
    [anon_sym___nonnull] = ACTIONS(4934),
    [anon_sym___nullable] = ACTIONS(4934),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4934),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4934),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4934),
    [anon_sym___real] = ACTIONS(4934),
    [anon_sym___strong] = ACTIONS(4934),
    [anon_sym___unsafe_unretained] = ACTIONS(4934),
    [anon_sym___unused] = ACTIONS(4934),
    [anon_sym___weak] = ACTIONS(4934),
    [anon_sym_alignas] = ACTIONS(4934),
    [anon_sym__Alignas] = ACTIONS(4934),
    [sym_primitive_type] = ACTIONS(4934),
    [anon_sym_enum] = ACTIONS(4934),
    [anon_sym_COLON] = ACTIONS(4936),
    [anon_sym_struct] = ACTIONS(4934),
    [anon_sym_union] = ACTIONS(4934),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4934),
    [anon_sym___typeof] = ACTIONS(4934),
    [anon_sym_typeof] = ACTIONS(4934),
    [aux_sym_preproc_undef_token1] = ACTIONS(4934),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4934),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4934),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4934),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4934),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4934),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4934),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4934),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4934),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4934),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4934),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4934),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4934),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4934),
    [anon_sym_API_AVAILABLE] = ACTIONS(4934),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4934),
    [anon_sym_API_DEPRECATED] = ACTIONS(4934),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4934),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4934),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4934),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4934),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4934),
    [anon_sym___deprecated_msg] = ACTIONS(4934),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4934),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4934),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4934),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4934),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4934),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4934),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4934),
    [anon_sym_ATend] = ACTIONS(4936),
    [anon_sym_ATsynthesize] = ACTIONS(4936),
    [anon_sym_ATdynamic] = ACTIONS(4936),
    [anon_sym_ATproperty] = ACTIONS(4936),
    [anon_sym_BOOL] = ACTIONS(4934),
    [anon_sym_IMP] = ACTIONS(4934),
    [anon_sym_SEL] = ACTIONS(4934),
    [anon_sym_Class] = ACTIONS(4934),
    [anon_sym_id] = ACTIONS(4934),
  },
  [STATE(1887)] = {
    [sym_identifier] = ACTIONS(4938),
    [aux_sym_preproc_def_token1] = ACTIONS(4938),
    [anon_sym_COMMA] = ACTIONS(4940),
    [anon_sym_RPAREN] = ACTIONS(4940),
    [aux_sym_preproc_if_token1] = ACTIONS(4938),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4938),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4938),
    [sym_preproc_directive] = ACTIONS(4938),
    [anon_sym_LPAREN2] = ACTIONS(4940),
    [anon_sym_DASH] = ACTIONS(4940),
    [anon_sym_PLUS] = ACTIONS(4940),
    [anon_sym_LT] = ACTIONS(4940),
    [anon_sym_SEMI] = ACTIONS(4940),
    [anon_sym___extension__] = ACTIONS(4938),
    [anon_sym_typedef] = ACTIONS(4938),
    [anon_sym_extern] = ACTIONS(4938),
    [anon_sym___attribute__] = ACTIONS(4938),
    [anon_sym___attribute] = ACTIONS(4938),
    [anon_sym_noreturn] = ACTIONS(4938),
    [anon_sym_LBRACK] = ACTIONS(4940),
    [anon_sym___declspec] = ACTIONS(4938),
    [anon_sym___cdecl] = ACTIONS(4938),
    [anon_sym___clrcall] = ACTIONS(4938),
    [anon_sym___stdcall] = ACTIONS(4938),
    [anon_sym___fastcall] = ACTIONS(4938),
    [anon_sym___thiscall] = ACTIONS(4938),
    [anon_sym___vectorcall] = ACTIONS(4938),
    [anon_sym_LBRACE] = ACTIONS(4940),
    [anon_sym_signed] = ACTIONS(4938),
    [anon_sym_unsigned] = ACTIONS(4938),
    [anon_sym_long] = ACTIONS(4938),
    [anon_sym_short] = ACTIONS(4938),
    [anon_sym_static] = ACTIONS(4938),
    [anon_sym_auto] = ACTIONS(4938),
    [anon_sym_register] = ACTIONS(4938),
    [anon_sym_inline] = ACTIONS(4938),
    [anon_sym___inline] = ACTIONS(4938),
    [anon_sym___inline__] = ACTIONS(4938),
    [anon_sym___forceinline] = ACTIONS(4938),
    [anon_sym_thread_local] = ACTIONS(4938),
    [anon_sym___thread] = ACTIONS(4938),
    [anon_sym_CG_EXTERN] = ACTIONS(4938),
    [anon_sym_CG_INLINE] = ACTIONS(4938),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4938),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4938),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4938),
    [anon_sym_IBOutlet] = ACTIONS(4938),
    [anon_sym_IBInspectable] = ACTIONS(4938),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4938),
    [anon_sym_NS_INLINE] = ACTIONS(4938),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4938),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4938),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4938),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4938),
    [anon_sym_const] = ACTIONS(4938),
    [anon_sym_constexpr] = ACTIONS(4938),
    [anon_sym_volatile] = ACTIONS(4938),
    [anon_sym_restrict] = ACTIONS(4938),
    [anon_sym___restrict__] = ACTIONS(4938),
    [anon_sym__Atomic] = ACTIONS(4938),
    [anon_sym__Noreturn] = ACTIONS(4938),
    [anon_sym__Nonnull] = ACTIONS(4938),
    [anon_sym_nullable] = ACTIONS(4938),
    [anon_sym__Complex] = ACTIONS(4938),
    [anon_sym__Nullable] = ACTIONS(4938),
    [anon_sym__Nullable_result] = ACTIONS(4938),
    [anon_sym__Null_unspecified] = ACTIONS(4938),
    [anon_sym___autoreleasing] = ACTIONS(4938),
    [anon_sym___block] = ACTIONS(4938),
    [anon_sym___bridge] = ACTIONS(4938),
    [anon_sym___bridge_retained] = ACTIONS(4938),
    [anon_sym___bridge_transfer] = ACTIONS(4938),
    [anon_sym___complex] = ACTIONS(4938),
    [anon_sym___const] = ACTIONS(4938),
    [anon_sym___imag] = ACTIONS(4938),
    [anon_sym___kindof] = ACTIONS(4938),
    [anon_sym___nonnull] = ACTIONS(4938),
    [anon_sym___nullable] = ACTIONS(4938),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4938),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4938),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4938),
    [anon_sym___real] = ACTIONS(4938),
    [anon_sym___strong] = ACTIONS(4938),
    [anon_sym___unsafe_unretained] = ACTIONS(4938),
    [anon_sym___unused] = ACTIONS(4938),
    [anon_sym___weak] = ACTIONS(4938),
    [anon_sym_alignas] = ACTIONS(4938),
    [anon_sym__Alignas] = ACTIONS(4938),
    [sym_primitive_type] = ACTIONS(4938),
    [anon_sym_enum] = ACTIONS(4938),
    [anon_sym_COLON] = ACTIONS(4940),
    [anon_sym_struct] = ACTIONS(4938),
    [anon_sym_union] = ACTIONS(4938),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4938),
    [anon_sym___typeof] = ACTIONS(4938),
    [anon_sym_typeof] = ACTIONS(4938),
    [aux_sym_preproc_undef_token1] = ACTIONS(4938),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4938),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4938),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4938),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4938),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4938),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4938),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4938),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4938),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4938),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4938),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4938),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4938),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4938),
    [anon_sym_API_AVAILABLE] = ACTIONS(4938),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4938),
    [anon_sym_API_DEPRECATED] = ACTIONS(4938),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4938),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4938),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4938),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4938),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4938),
    [anon_sym___deprecated_msg] = ACTIONS(4938),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4938),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4938),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4938),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4938),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4938),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4938),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4938),
    [anon_sym_ATend] = ACTIONS(4940),
    [anon_sym_ATsynthesize] = ACTIONS(4940),
    [anon_sym_ATdynamic] = ACTIONS(4940),
    [anon_sym_ATproperty] = ACTIONS(4940),
    [anon_sym_BOOL] = ACTIONS(4938),
    [anon_sym_IMP] = ACTIONS(4938),
    [anon_sym_SEL] = ACTIONS(4938),
    [anon_sym_Class] = ACTIONS(4938),
    [anon_sym_id] = ACTIONS(4938),
  },
  [STATE(1888)] = {
    [sym_identifier] = ACTIONS(4942),
    [aux_sym_preproc_def_token1] = ACTIONS(4942),
    [anon_sym_COMMA] = ACTIONS(4944),
    [anon_sym_RPAREN] = ACTIONS(4944),
    [aux_sym_preproc_if_token1] = ACTIONS(4942),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4942),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4942),
    [sym_preproc_directive] = ACTIONS(4942),
    [anon_sym_LPAREN2] = ACTIONS(4944),
    [anon_sym_DASH] = ACTIONS(4944),
    [anon_sym_PLUS] = ACTIONS(4944),
    [anon_sym_LT] = ACTIONS(4944),
    [anon_sym_SEMI] = ACTIONS(4944),
    [anon_sym___extension__] = ACTIONS(4942),
    [anon_sym_typedef] = ACTIONS(4942),
    [anon_sym_extern] = ACTIONS(4942),
    [anon_sym___attribute__] = ACTIONS(4942),
    [anon_sym___attribute] = ACTIONS(4942),
    [anon_sym_noreturn] = ACTIONS(4942),
    [anon_sym_LBRACK] = ACTIONS(4944),
    [anon_sym___declspec] = ACTIONS(4942),
    [anon_sym___cdecl] = ACTIONS(4942),
    [anon_sym___clrcall] = ACTIONS(4942),
    [anon_sym___stdcall] = ACTIONS(4942),
    [anon_sym___fastcall] = ACTIONS(4942),
    [anon_sym___thiscall] = ACTIONS(4942),
    [anon_sym___vectorcall] = ACTIONS(4942),
    [anon_sym_LBRACE] = ACTIONS(4944),
    [anon_sym_signed] = ACTIONS(4942),
    [anon_sym_unsigned] = ACTIONS(4942),
    [anon_sym_long] = ACTIONS(4942),
    [anon_sym_short] = ACTIONS(4942),
    [anon_sym_static] = ACTIONS(4942),
    [anon_sym_auto] = ACTIONS(4942),
    [anon_sym_register] = ACTIONS(4942),
    [anon_sym_inline] = ACTIONS(4942),
    [anon_sym___inline] = ACTIONS(4942),
    [anon_sym___inline__] = ACTIONS(4942),
    [anon_sym___forceinline] = ACTIONS(4942),
    [anon_sym_thread_local] = ACTIONS(4942),
    [anon_sym___thread] = ACTIONS(4942),
    [anon_sym_CG_EXTERN] = ACTIONS(4942),
    [anon_sym_CG_INLINE] = ACTIONS(4942),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4942),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4942),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4942),
    [anon_sym_IBOutlet] = ACTIONS(4942),
    [anon_sym_IBInspectable] = ACTIONS(4942),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4942),
    [anon_sym_NS_INLINE] = ACTIONS(4942),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4942),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4942),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4942),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4942),
    [anon_sym_const] = ACTIONS(4942),
    [anon_sym_constexpr] = ACTIONS(4942),
    [anon_sym_volatile] = ACTIONS(4942),
    [anon_sym_restrict] = ACTIONS(4942),
    [anon_sym___restrict__] = ACTIONS(4942),
    [anon_sym__Atomic] = ACTIONS(4942),
    [anon_sym__Noreturn] = ACTIONS(4942),
    [anon_sym__Nonnull] = ACTIONS(4942),
    [anon_sym_nullable] = ACTIONS(4942),
    [anon_sym__Complex] = ACTIONS(4942),
    [anon_sym__Nullable] = ACTIONS(4942),
    [anon_sym__Nullable_result] = ACTIONS(4942),
    [anon_sym__Null_unspecified] = ACTIONS(4942),
    [anon_sym___autoreleasing] = ACTIONS(4942),
    [anon_sym___block] = ACTIONS(4942),
    [anon_sym___bridge] = ACTIONS(4942),
    [anon_sym___bridge_retained] = ACTIONS(4942),
    [anon_sym___bridge_transfer] = ACTIONS(4942),
    [anon_sym___complex] = ACTIONS(4942),
    [anon_sym___const] = ACTIONS(4942),
    [anon_sym___imag] = ACTIONS(4942),
    [anon_sym___kindof] = ACTIONS(4942),
    [anon_sym___nonnull] = ACTIONS(4942),
    [anon_sym___nullable] = ACTIONS(4942),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4942),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4942),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4942),
    [anon_sym___real] = ACTIONS(4942),
    [anon_sym___strong] = ACTIONS(4942),
    [anon_sym___unsafe_unretained] = ACTIONS(4942),
    [anon_sym___unused] = ACTIONS(4942),
    [anon_sym___weak] = ACTIONS(4942),
    [anon_sym_alignas] = ACTIONS(4942),
    [anon_sym__Alignas] = ACTIONS(4942),
    [sym_primitive_type] = ACTIONS(4942),
    [anon_sym_enum] = ACTIONS(4942),
    [anon_sym_COLON] = ACTIONS(4944),
    [anon_sym_struct] = ACTIONS(4942),
    [anon_sym_union] = ACTIONS(4942),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4942),
    [anon_sym___typeof] = ACTIONS(4942),
    [anon_sym_typeof] = ACTIONS(4942),
    [aux_sym_preproc_undef_token1] = ACTIONS(4942),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4942),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4942),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4942),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4942),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4942),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4942),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4942),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4942),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4942),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4942),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4942),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4942),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4942),
    [anon_sym_API_AVAILABLE] = ACTIONS(4942),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4942),
    [anon_sym_API_DEPRECATED] = ACTIONS(4942),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4942),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4942),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4942),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4942),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4942),
    [anon_sym___deprecated_msg] = ACTIONS(4942),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4942),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4942),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4942),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4942),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4942),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4942),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4942),
    [anon_sym_ATend] = ACTIONS(4944),
    [anon_sym_ATsynthesize] = ACTIONS(4944),
    [anon_sym_ATdynamic] = ACTIONS(4944),
    [anon_sym_ATproperty] = ACTIONS(4944),
    [anon_sym_BOOL] = ACTIONS(4942),
    [anon_sym_IMP] = ACTIONS(4942),
    [anon_sym_SEL] = ACTIONS(4942),
    [anon_sym_Class] = ACTIONS(4942),
    [anon_sym_id] = ACTIONS(4942),
  },
  [STATE(1889)] = {
    [sym_identifier] = ACTIONS(4946),
    [aux_sym_preproc_def_token1] = ACTIONS(4946),
    [anon_sym_COMMA] = ACTIONS(4948),
    [anon_sym_RPAREN] = ACTIONS(4948),
    [aux_sym_preproc_if_token1] = ACTIONS(4946),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4946),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4946),
    [sym_preproc_directive] = ACTIONS(4946),
    [anon_sym_LPAREN2] = ACTIONS(4948),
    [anon_sym_DASH] = ACTIONS(4948),
    [anon_sym_PLUS] = ACTIONS(4948),
    [anon_sym_LT] = ACTIONS(4948),
    [anon_sym_SEMI] = ACTIONS(4948),
    [anon_sym___extension__] = ACTIONS(4946),
    [anon_sym_typedef] = ACTIONS(4946),
    [anon_sym_extern] = ACTIONS(4946),
    [anon_sym___attribute__] = ACTIONS(4946),
    [anon_sym___attribute] = ACTIONS(4946),
    [anon_sym_noreturn] = ACTIONS(4946),
    [anon_sym_LBRACK] = ACTIONS(4948),
    [anon_sym___declspec] = ACTIONS(4946),
    [anon_sym___cdecl] = ACTIONS(4946),
    [anon_sym___clrcall] = ACTIONS(4946),
    [anon_sym___stdcall] = ACTIONS(4946),
    [anon_sym___fastcall] = ACTIONS(4946),
    [anon_sym___thiscall] = ACTIONS(4946),
    [anon_sym___vectorcall] = ACTIONS(4946),
    [anon_sym_LBRACE] = ACTIONS(4948),
    [anon_sym_signed] = ACTIONS(4946),
    [anon_sym_unsigned] = ACTIONS(4946),
    [anon_sym_long] = ACTIONS(4946),
    [anon_sym_short] = ACTIONS(4946),
    [anon_sym_static] = ACTIONS(4946),
    [anon_sym_auto] = ACTIONS(4946),
    [anon_sym_register] = ACTIONS(4946),
    [anon_sym_inline] = ACTIONS(4946),
    [anon_sym___inline] = ACTIONS(4946),
    [anon_sym___inline__] = ACTIONS(4946),
    [anon_sym___forceinline] = ACTIONS(4946),
    [anon_sym_thread_local] = ACTIONS(4946),
    [anon_sym___thread] = ACTIONS(4946),
    [anon_sym_CG_EXTERN] = ACTIONS(4946),
    [anon_sym_CG_INLINE] = ACTIONS(4946),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4946),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4946),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4946),
    [anon_sym_IBOutlet] = ACTIONS(4946),
    [anon_sym_IBInspectable] = ACTIONS(4946),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4946),
    [anon_sym_NS_INLINE] = ACTIONS(4946),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4946),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4946),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4946),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4946),
    [anon_sym_const] = ACTIONS(4946),
    [anon_sym_constexpr] = ACTIONS(4946),
    [anon_sym_volatile] = ACTIONS(4946),
    [anon_sym_restrict] = ACTIONS(4946),
    [anon_sym___restrict__] = ACTIONS(4946),
    [anon_sym__Atomic] = ACTIONS(4946),
    [anon_sym__Noreturn] = ACTIONS(4946),
    [anon_sym__Nonnull] = ACTIONS(4946),
    [anon_sym_nullable] = ACTIONS(4946),
    [anon_sym__Complex] = ACTIONS(4946),
    [anon_sym__Nullable] = ACTIONS(4946),
    [anon_sym__Nullable_result] = ACTIONS(4946),
    [anon_sym__Null_unspecified] = ACTIONS(4946),
    [anon_sym___autoreleasing] = ACTIONS(4946),
    [anon_sym___block] = ACTIONS(4946),
    [anon_sym___bridge] = ACTIONS(4946),
    [anon_sym___bridge_retained] = ACTIONS(4946),
    [anon_sym___bridge_transfer] = ACTIONS(4946),
    [anon_sym___complex] = ACTIONS(4946),
    [anon_sym___const] = ACTIONS(4946),
    [anon_sym___imag] = ACTIONS(4946),
    [anon_sym___kindof] = ACTIONS(4946),
    [anon_sym___nonnull] = ACTIONS(4946),
    [anon_sym___nullable] = ACTIONS(4946),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4946),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4946),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4946),
    [anon_sym___real] = ACTIONS(4946),
    [anon_sym___strong] = ACTIONS(4946),
    [anon_sym___unsafe_unretained] = ACTIONS(4946),
    [anon_sym___unused] = ACTIONS(4946),
    [anon_sym___weak] = ACTIONS(4946),
    [anon_sym_alignas] = ACTIONS(4946),
    [anon_sym__Alignas] = ACTIONS(4946),
    [sym_primitive_type] = ACTIONS(4946),
    [anon_sym_enum] = ACTIONS(4946),
    [anon_sym_COLON] = ACTIONS(4948),
    [anon_sym_struct] = ACTIONS(4946),
    [anon_sym_union] = ACTIONS(4946),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4946),
    [anon_sym___typeof] = ACTIONS(4946),
    [anon_sym_typeof] = ACTIONS(4946),
    [aux_sym_preproc_undef_token1] = ACTIONS(4946),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4946),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4946),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4946),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4946),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4946),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4946),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4946),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4946),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4946),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4946),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4946),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4946),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4946),
    [anon_sym_API_AVAILABLE] = ACTIONS(4946),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4946),
    [anon_sym_API_DEPRECATED] = ACTIONS(4946),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4946),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4946),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4946),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4946),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4946),
    [anon_sym___deprecated_msg] = ACTIONS(4946),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4946),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4946),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4946),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4946),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4946),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4946),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4946),
    [anon_sym_ATend] = ACTIONS(4948),
    [anon_sym_ATsynthesize] = ACTIONS(4948),
    [anon_sym_ATdynamic] = ACTIONS(4948),
    [anon_sym_ATproperty] = ACTIONS(4948),
    [anon_sym_BOOL] = ACTIONS(4946),
    [anon_sym_IMP] = ACTIONS(4946),
    [anon_sym_SEL] = ACTIONS(4946),
    [anon_sym_Class] = ACTIONS(4946),
    [anon_sym_id] = ACTIONS(4946),
  },
  [STATE(1890)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(644),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1891)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(319),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1892)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(4950),
    [anon_sym_SEMI] = ACTIONS(4953),
    [anon_sym___extension__] = ACTIONS(4955),
    [anon_sym_extern] = ACTIONS(4958),
    [anon_sym___attribute__] = ACTIONS(4961),
    [anon_sym___attribute] = ACTIONS(4961),
    [anon_sym_noreturn] = ACTIONS(4955),
    [anon_sym_LBRACK] = ACTIONS(4964),
    [anon_sym___declspec] = ACTIONS(4967),
    [anon_sym_LBRACE] = ACTIONS(4953),
    [anon_sym_signed] = ACTIONS(4970),
    [anon_sym_unsigned] = ACTIONS(4970),
    [anon_sym_long] = ACTIONS(4970),
    [anon_sym_short] = ACTIONS(4970),
    [anon_sym_static] = ACTIONS(4958),
    [anon_sym_ATautoreleasepool] = ACTIONS(4953),
    [anon_sym_auto] = ACTIONS(4958),
    [anon_sym_register] = ACTIONS(4958),
    [anon_sym_inline] = ACTIONS(4958),
    [anon_sym___inline] = ACTIONS(4958),
    [anon_sym___inline__] = ACTIONS(4958),
    [anon_sym___forceinline] = ACTIONS(4958),
    [anon_sym_thread_local] = ACTIONS(4958),
    [anon_sym___thread] = ACTIONS(4958),
    [anon_sym_CG_EXTERN] = ACTIONS(4958),
    [anon_sym_CG_INLINE] = ACTIONS(4958),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4958),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4958),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4958),
    [anon_sym_IBOutlet] = ACTIONS(4958),
    [anon_sym_IBInspectable] = ACTIONS(4958),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4958),
    [anon_sym_NS_INLINE] = ACTIONS(4958),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4958),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4958),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4958),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4958),
    [anon_sym_const] = ACTIONS(4955),
    [anon_sym_constexpr] = ACTIONS(4955),
    [anon_sym_volatile] = ACTIONS(4955),
    [anon_sym_restrict] = ACTIONS(4955),
    [anon_sym___restrict__] = ACTIONS(4955),
    [anon_sym__Atomic] = ACTIONS(4955),
    [anon_sym__Noreturn] = ACTIONS(4955),
    [anon_sym__Nonnull] = ACTIONS(4955),
    [anon_sym_nullable] = ACTIONS(4955),
    [anon_sym__Complex] = ACTIONS(4955),
    [anon_sym__Nullable] = ACTIONS(4955),
    [anon_sym__Nullable_result] = ACTIONS(4955),
    [anon_sym__Null_unspecified] = ACTIONS(4955),
    [anon_sym___autoreleasing] = ACTIONS(4955),
    [anon_sym___block] = ACTIONS(4955),
    [anon_sym___bridge] = ACTIONS(4955),
    [anon_sym___bridge_retained] = ACTIONS(4955),
    [anon_sym___bridge_transfer] = ACTIONS(4955),
    [anon_sym___complex] = ACTIONS(4955),
    [anon_sym___const] = ACTIONS(4955),
    [anon_sym___imag] = ACTIONS(4955),
    [anon_sym___kindof] = ACTIONS(4955),
    [anon_sym___nonnull] = ACTIONS(4955),
    [anon_sym___nullable] = ACTIONS(4955),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4955),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4955),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4955),
    [anon_sym___real] = ACTIONS(4955),
    [anon_sym___strong] = ACTIONS(4955),
    [anon_sym___unsafe_unretained] = ACTIONS(4955),
    [anon_sym___unused] = ACTIONS(4955),
    [anon_sym___weak] = ACTIONS(4955),
    [anon_sym_alignas] = ACTIONS(4973),
    [anon_sym__Alignas] = ACTIONS(4973),
    [sym_primitive_type] = ACTIONS(4976),
    [anon_sym_enum] = ACTIONS(4979),
    [anon_sym_struct] = ACTIONS(4982),
    [anon_sym_union] = ACTIONS(4985),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4988),
    [anon_sym___typeof] = ACTIONS(4988),
    [anon_sym_typeof] = ACTIONS(4988),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4991),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4991),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4991),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4991),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4991),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4991),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4991),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4991),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4991),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4994),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4994),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4994),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4994),
    [anon_sym_API_AVAILABLE] = ACTIONS(4994),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4994),
    [anon_sym_API_DEPRECATED] = ACTIONS(4994),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4994),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4994),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4994),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4994),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4994),
    [anon_sym___deprecated_msg] = ACTIONS(4994),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4994),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4994),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4994),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4994),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4994),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4994),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4994),
    [anon_sym_BOOL] = ACTIONS(4997),
    [anon_sym_IMP] = ACTIONS(4997),
    [anon_sym_SEL] = ACTIONS(4997),
    [anon_sym_Class] = ACTIONS(4997),
    [anon_sym_id] = ACTIONS(4997),
  },
  [STATE(1893)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(614),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1894)] = {
    [sym_declaration] = STATE(1892),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_compound_statement] = STATE(1045),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1892),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1895)] = {
    [sym_identifier] = ACTIONS(5000),
    [aux_sym_preproc_def_token1] = ACTIONS(5000),
    [anon_sym_COMMA] = ACTIONS(5002),
    [anon_sym_RPAREN] = ACTIONS(5002),
    [aux_sym_preproc_if_token1] = ACTIONS(5000),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5000),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5000),
    [sym_preproc_directive] = ACTIONS(5000),
    [anon_sym_LPAREN2] = ACTIONS(5002),
    [anon_sym_DASH] = ACTIONS(5002),
    [anon_sym_PLUS] = ACTIONS(5002),
    [anon_sym_LT] = ACTIONS(5002),
    [anon_sym_SEMI] = ACTIONS(5002),
    [anon_sym___extension__] = ACTIONS(5000),
    [anon_sym_typedef] = ACTIONS(5000),
    [anon_sym_extern] = ACTIONS(5000),
    [anon_sym___attribute__] = ACTIONS(5000),
    [anon_sym___attribute] = ACTIONS(5000),
    [anon_sym_noreturn] = ACTIONS(5000),
    [anon_sym_LBRACK] = ACTIONS(5002),
    [anon_sym___declspec] = ACTIONS(5000),
    [anon_sym___cdecl] = ACTIONS(5000),
    [anon_sym___clrcall] = ACTIONS(5000),
    [anon_sym___stdcall] = ACTIONS(5000),
    [anon_sym___fastcall] = ACTIONS(5000),
    [anon_sym___thiscall] = ACTIONS(5000),
    [anon_sym___vectorcall] = ACTIONS(5000),
    [anon_sym_LBRACE] = ACTIONS(5002),
    [anon_sym_signed] = ACTIONS(5000),
    [anon_sym_unsigned] = ACTIONS(5000),
    [anon_sym_long] = ACTIONS(5000),
    [anon_sym_short] = ACTIONS(5000),
    [anon_sym_static] = ACTIONS(5000),
    [anon_sym_auto] = ACTIONS(5000),
    [anon_sym_register] = ACTIONS(5000),
    [anon_sym_inline] = ACTIONS(5000),
    [anon_sym___inline] = ACTIONS(5000),
    [anon_sym___inline__] = ACTIONS(5000),
    [anon_sym___forceinline] = ACTIONS(5000),
    [anon_sym_thread_local] = ACTIONS(5000),
    [anon_sym___thread] = ACTIONS(5000),
    [anon_sym_CG_EXTERN] = ACTIONS(5000),
    [anon_sym_CG_INLINE] = ACTIONS(5000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5000),
    [anon_sym_IBOutlet] = ACTIONS(5000),
    [anon_sym_IBInspectable] = ACTIONS(5000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5000),
    [anon_sym_NS_INLINE] = ACTIONS(5000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5000),
    [anon_sym_const] = ACTIONS(5000),
    [anon_sym_constexpr] = ACTIONS(5000),
    [anon_sym_volatile] = ACTIONS(5000),
    [anon_sym_restrict] = ACTIONS(5000),
    [anon_sym___restrict__] = ACTIONS(5000),
    [anon_sym__Atomic] = ACTIONS(5000),
    [anon_sym__Noreturn] = ACTIONS(5000),
    [anon_sym__Nonnull] = ACTIONS(5000),
    [anon_sym_nullable] = ACTIONS(5000),
    [anon_sym__Complex] = ACTIONS(5000),
    [anon_sym__Nullable] = ACTIONS(5000),
    [anon_sym__Nullable_result] = ACTIONS(5000),
    [anon_sym__Null_unspecified] = ACTIONS(5000),
    [anon_sym___autoreleasing] = ACTIONS(5000),
    [anon_sym___block] = ACTIONS(5000),
    [anon_sym___bridge] = ACTIONS(5000),
    [anon_sym___bridge_retained] = ACTIONS(5000),
    [anon_sym___bridge_transfer] = ACTIONS(5000),
    [anon_sym___complex] = ACTIONS(5000),
    [anon_sym___const] = ACTIONS(5000),
    [anon_sym___imag] = ACTIONS(5000),
    [anon_sym___kindof] = ACTIONS(5000),
    [anon_sym___nonnull] = ACTIONS(5000),
    [anon_sym___nullable] = ACTIONS(5000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5000),
    [anon_sym___real] = ACTIONS(5000),
    [anon_sym___strong] = ACTIONS(5000),
    [anon_sym___unsafe_unretained] = ACTIONS(5000),
    [anon_sym___unused] = ACTIONS(5000),
    [anon_sym___weak] = ACTIONS(5000),
    [anon_sym_alignas] = ACTIONS(5000),
    [anon_sym__Alignas] = ACTIONS(5000),
    [sym_primitive_type] = ACTIONS(5000),
    [anon_sym_enum] = ACTIONS(5000),
    [anon_sym_COLON] = ACTIONS(5002),
    [anon_sym_struct] = ACTIONS(5000),
    [anon_sym_union] = ACTIONS(5000),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5000),
    [anon_sym___typeof] = ACTIONS(5000),
    [anon_sym_typeof] = ACTIONS(5000),
    [aux_sym_preproc_undef_token1] = ACTIONS(5000),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5000),
    [anon_sym_API_AVAILABLE] = ACTIONS(5000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5000),
    [anon_sym_API_DEPRECATED] = ACTIONS(5000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5000),
    [anon_sym___deprecated_msg] = ACTIONS(5000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5000),
    [anon_sym_ATend] = ACTIONS(5002),
    [anon_sym_ATsynthesize] = ACTIONS(5002),
    [anon_sym_ATdynamic] = ACTIONS(5002),
    [anon_sym_ATproperty] = ACTIONS(5002),
    [anon_sym_BOOL] = ACTIONS(5000),
    [anon_sym_IMP] = ACTIONS(5000),
    [anon_sym_SEL] = ACTIONS(5000),
    [anon_sym_Class] = ACTIONS(5000),
    [anon_sym_id] = ACTIONS(5000),
  },
  [STATE(1896)] = {
    [sym_identifier] = ACTIONS(5004),
    [aux_sym_preproc_def_token1] = ACTIONS(5004),
    [anon_sym_COMMA] = ACTIONS(5006),
    [anon_sym_RPAREN] = ACTIONS(5006),
    [aux_sym_preproc_if_token1] = ACTIONS(5004),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5004),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5004),
    [sym_preproc_directive] = ACTIONS(5004),
    [anon_sym_LPAREN2] = ACTIONS(5006),
    [anon_sym_DASH] = ACTIONS(5006),
    [anon_sym_PLUS] = ACTIONS(5006),
    [anon_sym_LT] = ACTIONS(5006),
    [anon_sym_SEMI] = ACTIONS(5006),
    [anon_sym___extension__] = ACTIONS(5004),
    [anon_sym_typedef] = ACTIONS(5004),
    [anon_sym_extern] = ACTIONS(5004),
    [anon_sym___attribute__] = ACTIONS(5004),
    [anon_sym___attribute] = ACTIONS(5004),
    [anon_sym_noreturn] = ACTIONS(5004),
    [anon_sym_LBRACK] = ACTIONS(5006),
    [anon_sym___declspec] = ACTIONS(5004),
    [anon_sym___cdecl] = ACTIONS(5004),
    [anon_sym___clrcall] = ACTIONS(5004),
    [anon_sym___stdcall] = ACTIONS(5004),
    [anon_sym___fastcall] = ACTIONS(5004),
    [anon_sym___thiscall] = ACTIONS(5004),
    [anon_sym___vectorcall] = ACTIONS(5004),
    [anon_sym_LBRACE] = ACTIONS(5006),
    [anon_sym_signed] = ACTIONS(5004),
    [anon_sym_unsigned] = ACTIONS(5004),
    [anon_sym_long] = ACTIONS(5004),
    [anon_sym_short] = ACTIONS(5004),
    [anon_sym_static] = ACTIONS(5004),
    [anon_sym_auto] = ACTIONS(5004),
    [anon_sym_register] = ACTIONS(5004),
    [anon_sym_inline] = ACTIONS(5004),
    [anon_sym___inline] = ACTIONS(5004),
    [anon_sym___inline__] = ACTIONS(5004),
    [anon_sym___forceinline] = ACTIONS(5004),
    [anon_sym_thread_local] = ACTIONS(5004),
    [anon_sym___thread] = ACTIONS(5004),
    [anon_sym_CG_EXTERN] = ACTIONS(5004),
    [anon_sym_CG_INLINE] = ACTIONS(5004),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5004),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5004),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5004),
    [anon_sym_IBOutlet] = ACTIONS(5004),
    [anon_sym_IBInspectable] = ACTIONS(5004),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5004),
    [anon_sym_NS_INLINE] = ACTIONS(5004),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5004),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5004),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5004),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5004),
    [anon_sym_const] = ACTIONS(5004),
    [anon_sym_constexpr] = ACTIONS(5004),
    [anon_sym_volatile] = ACTIONS(5004),
    [anon_sym_restrict] = ACTIONS(5004),
    [anon_sym___restrict__] = ACTIONS(5004),
    [anon_sym__Atomic] = ACTIONS(5004),
    [anon_sym__Noreturn] = ACTIONS(5004),
    [anon_sym__Nonnull] = ACTIONS(5004),
    [anon_sym_nullable] = ACTIONS(5004),
    [anon_sym__Complex] = ACTIONS(5004),
    [anon_sym__Nullable] = ACTIONS(5004),
    [anon_sym__Nullable_result] = ACTIONS(5004),
    [anon_sym__Null_unspecified] = ACTIONS(5004),
    [anon_sym___autoreleasing] = ACTIONS(5004),
    [anon_sym___block] = ACTIONS(5004),
    [anon_sym___bridge] = ACTIONS(5004),
    [anon_sym___bridge_retained] = ACTIONS(5004),
    [anon_sym___bridge_transfer] = ACTIONS(5004),
    [anon_sym___complex] = ACTIONS(5004),
    [anon_sym___const] = ACTIONS(5004),
    [anon_sym___imag] = ACTIONS(5004),
    [anon_sym___kindof] = ACTIONS(5004),
    [anon_sym___nonnull] = ACTIONS(5004),
    [anon_sym___nullable] = ACTIONS(5004),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5004),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5004),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5004),
    [anon_sym___real] = ACTIONS(5004),
    [anon_sym___strong] = ACTIONS(5004),
    [anon_sym___unsafe_unretained] = ACTIONS(5004),
    [anon_sym___unused] = ACTIONS(5004),
    [anon_sym___weak] = ACTIONS(5004),
    [anon_sym_alignas] = ACTIONS(5004),
    [anon_sym__Alignas] = ACTIONS(5004),
    [sym_primitive_type] = ACTIONS(5004),
    [anon_sym_enum] = ACTIONS(5004),
    [anon_sym_COLON] = ACTIONS(5006),
    [anon_sym_struct] = ACTIONS(5004),
    [anon_sym_union] = ACTIONS(5004),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5004),
    [anon_sym___typeof] = ACTIONS(5004),
    [anon_sym_typeof] = ACTIONS(5004),
    [aux_sym_preproc_undef_token1] = ACTIONS(5004),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5004),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5004),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5004),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5004),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5004),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5004),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5004),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5004),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5004),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5004),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5004),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5004),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5004),
    [anon_sym_API_AVAILABLE] = ACTIONS(5004),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5004),
    [anon_sym_API_DEPRECATED] = ACTIONS(5004),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5004),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5004),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5004),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5004),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5004),
    [anon_sym___deprecated_msg] = ACTIONS(5004),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5004),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5004),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5004),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5004),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5004),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5004),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5004),
    [anon_sym_ATend] = ACTIONS(5006),
    [anon_sym_ATsynthesize] = ACTIONS(5006),
    [anon_sym_ATdynamic] = ACTIONS(5006),
    [anon_sym_ATproperty] = ACTIONS(5006),
    [anon_sym_BOOL] = ACTIONS(5004),
    [anon_sym_IMP] = ACTIONS(5004),
    [anon_sym_SEL] = ACTIONS(5004),
    [anon_sym_Class] = ACTIONS(5004),
    [anon_sym_id] = ACTIONS(5004),
  },
  [STATE(1897)] = {
    [sym_identifier] = ACTIONS(5008),
    [aux_sym_preproc_def_token1] = ACTIONS(5008),
    [anon_sym_COMMA] = ACTIONS(5010),
    [anon_sym_RPAREN] = ACTIONS(5010),
    [aux_sym_preproc_if_token1] = ACTIONS(5008),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5008),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5008),
    [sym_preproc_directive] = ACTIONS(5008),
    [anon_sym_LPAREN2] = ACTIONS(5010),
    [anon_sym_DASH] = ACTIONS(5010),
    [anon_sym_PLUS] = ACTIONS(5010),
    [anon_sym_LT] = ACTIONS(5010),
    [anon_sym_SEMI] = ACTIONS(5010),
    [anon_sym___extension__] = ACTIONS(5008),
    [anon_sym_typedef] = ACTIONS(5008),
    [anon_sym_extern] = ACTIONS(5008),
    [anon_sym___attribute__] = ACTIONS(5008),
    [anon_sym___attribute] = ACTIONS(5008),
    [anon_sym_noreturn] = ACTIONS(5008),
    [anon_sym_LBRACK] = ACTIONS(5010),
    [anon_sym___declspec] = ACTIONS(5008),
    [anon_sym___cdecl] = ACTIONS(5008),
    [anon_sym___clrcall] = ACTIONS(5008),
    [anon_sym___stdcall] = ACTIONS(5008),
    [anon_sym___fastcall] = ACTIONS(5008),
    [anon_sym___thiscall] = ACTIONS(5008),
    [anon_sym___vectorcall] = ACTIONS(5008),
    [anon_sym_LBRACE] = ACTIONS(5010),
    [anon_sym_signed] = ACTIONS(5008),
    [anon_sym_unsigned] = ACTIONS(5008),
    [anon_sym_long] = ACTIONS(5008),
    [anon_sym_short] = ACTIONS(5008),
    [anon_sym_static] = ACTIONS(5008),
    [anon_sym_auto] = ACTIONS(5008),
    [anon_sym_register] = ACTIONS(5008),
    [anon_sym_inline] = ACTIONS(5008),
    [anon_sym___inline] = ACTIONS(5008),
    [anon_sym___inline__] = ACTIONS(5008),
    [anon_sym___forceinline] = ACTIONS(5008),
    [anon_sym_thread_local] = ACTIONS(5008),
    [anon_sym___thread] = ACTIONS(5008),
    [anon_sym_CG_EXTERN] = ACTIONS(5008),
    [anon_sym_CG_INLINE] = ACTIONS(5008),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5008),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5008),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5008),
    [anon_sym_IBOutlet] = ACTIONS(5008),
    [anon_sym_IBInspectable] = ACTIONS(5008),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5008),
    [anon_sym_NS_INLINE] = ACTIONS(5008),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5008),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5008),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5008),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5008),
    [anon_sym_const] = ACTIONS(5008),
    [anon_sym_constexpr] = ACTIONS(5008),
    [anon_sym_volatile] = ACTIONS(5008),
    [anon_sym_restrict] = ACTIONS(5008),
    [anon_sym___restrict__] = ACTIONS(5008),
    [anon_sym__Atomic] = ACTIONS(5008),
    [anon_sym__Noreturn] = ACTIONS(5008),
    [anon_sym__Nonnull] = ACTIONS(5008),
    [anon_sym_nullable] = ACTIONS(5008),
    [anon_sym__Complex] = ACTIONS(5008),
    [anon_sym__Nullable] = ACTIONS(5008),
    [anon_sym__Nullable_result] = ACTIONS(5008),
    [anon_sym__Null_unspecified] = ACTIONS(5008),
    [anon_sym___autoreleasing] = ACTIONS(5008),
    [anon_sym___block] = ACTIONS(5008),
    [anon_sym___bridge] = ACTIONS(5008),
    [anon_sym___bridge_retained] = ACTIONS(5008),
    [anon_sym___bridge_transfer] = ACTIONS(5008),
    [anon_sym___complex] = ACTIONS(5008),
    [anon_sym___const] = ACTIONS(5008),
    [anon_sym___imag] = ACTIONS(5008),
    [anon_sym___kindof] = ACTIONS(5008),
    [anon_sym___nonnull] = ACTIONS(5008),
    [anon_sym___nullable] = ACTIONS(5008),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5008),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5008),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5008),
    [anon_sym___real] = ACTIONS(5008),
    [anon_sym___strong] = ACTIONS(5008),
    [anon_sym___unsafe_unretained] = ACTIONS(5008),
    [anon_sym___unused] = ACTIONS(5008),
    [anon_sym___weak] = ACTIONS(5008),
    [anon_sym_alignas] = ACTIONS(5008),
    [anon_sym__Alignas] = ACTIONS(5008),
    [sym_primitive_type] = ACTIONS(5008),
    [anon_sym_enum] = ACTIONS(5008),
    [anon_sym_COLON] = ACTIONS(5010),
    [anon_sym_struct] = ACTIONS(5008),
    [anon_sym_union] = ACTIONS(5008),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5008),
    [anon_sym___typeof] = ACTIONS(5008),
    [anon_sym_typeof] = ACTIONS(5008),
    [aux_sym_preproc_undef_token1] = ACTIONS(5008),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5008),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5008),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5008),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5008),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5008),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5008),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5008),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5008),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5008),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5008),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5008),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5008),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5008),
    [anon_sym_API_AVAILABLE] = ACTIONS(5008),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5008),
    [anon_sym_API_DEPRECATED] = ACTIONS(5008),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5008),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5008),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5008),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5008),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5008),
    [anon_sym___deprecated_msg] = ACTIONS(5008),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5008),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5008),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5008),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5008),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5008),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5008),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5008),
    [anon_sym_ATend] = ACTIONS(5010),
    [anon_sym_ATsynthesize] = ACTIONS(5010),
    [anon_sym_ATdynamic] = ACTIONS(5010),
    [anon_sym_ATproperty] = ACTIONS(5010),
    [anon_sym_BOOL] = ACTIONS(5008),
    [anon_sym_IMP] = ACTIONS(5008),
    [anon_sym_SEL] = ACTIONS(5008),
    [anon_sym_Class] = ACTIONS(5008),
    [anon_sym_id] = ACTIONS(5008),
  },
  [STATE(1898)] = {
    [sym_identifier] = ACTIONS(5012),
    [aux_sym_preproc_def_token1] = ACTIONS(5012),
    [anon_sym_COMMA] = ACTIONS(5014),
    [anon_sym_RPAREN] = ACTIONS(5014),
    [aux_sym_preproc_if_token1] = ACTIONS(5012),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5012),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5012),
    [sym_preproc_directive] = ACTIONS(5012),
    [anon_sym_LPAREN2] = ACTIONS(5014),
    [anon_sym_DASH] = ACTIONS(5014),
    [anon_sym_PLUS] = ACTIONS(5014),
    [anon_sym_LT] = ACTIONS(5014),
    [anon_sym_SEMI] = ACTIONS(5014),
    [anon_sym___extension__] = ACTIONS(5012),
    [anon_sym_typedef] = ACTIONS(5012),
    [anon_sym_extern] = ACTIONS(5012),
    [anon_sym___attribute__] = ACTIONS(5012),
    [anon_sym___attribute] = ACTIONS(5012),
    [anon_sym_noreturn] = ACTIONS(5012),
    [anon_sym_LBRACK] = ACTIONS(5014),
    [anon_sym___declspec] = ACTIONS(5012),
    [anon_sym___cdecl] = ACTIONS(5012),
    [anon_sym___clrcall] = ACTIONS(5012),
    [anon_sym___stdcall] = ACTIONS(5012),
    [anon_sym___fastcall] = ACTIONS(5012),
    [anon_sym___thiscall] = ACTIONS(5012),
    [anon_sym___vectorcall] = ACTIONS(5012),
    [anon_sym_LBRACE] = ACTIONS(5014),
    [anon_sym_signed] = ACTIONS(5012),
    [anon_sym_unsigned] = ACTIONS(5012),
    [anon_sym_long] = ACTIONS(5012),
    [anon_sym_short] = ACTIONS(5012),
    [anon_sym_static] = ACTIONS(5012),
    [anon_sym_auto] = ACTIONS(5012),
    [anon_sym_register] = ACTIONS(5012),
    [anon_sym_inline] = ACTIONS(5012),
    [anon_sym___inline] = ACTIONS(5012),
    [anon_sym___inline__] = ACTIONS(5012),
    [anon_sym___forceinline] = ACTIONS(5012),
    [anon_sym_thread_local] = ACTIONS(5012),
    [anon_sym___thread] = ACTIONS(5012),
    [anon_sym_CG_EXTERN] = ACTIONS(5012),
    [anon_sym_CG_INLINE] = ACTIONS(5012),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5012),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5012),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5012),
    [anon_sym_IBOutlet] = ACTIONS(5012),
    [anon_sym_IBInspectable] = ACTIONS(5012),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5012),
    [anon_sym_NS_INLINE] = ACTIONS(5012),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5012),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5012),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5012),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5012),
    [anon_sym_const] = ACTIONS(5012),
    [anon_sym_constexpr] = ACTIONS(5012),
    [anon_sym_volatile] = ACTIONS(5012),
    [anon_sym_restrict] = ACTIONS(5012),
    [anon_sym___restrict__] = ACTIONS(5012),
    [anon_sym__Atomic] = ACTIONS(5012),
    [anon_sym__Noreturn] = ACTIONS(5012),
    [anon_sym__Nonnull] = ACTIONS(5012),
    [anon_sym_nullable] = ACTIONS(5012),
    [anon_sym__Complex] = ACTIONS(5012),
    [anon_sym__Nullable] = ACTIONS(5012),
    [anon_sym__Nullable_result] = ACTIONS(5012),
    [anon_sym__Null_unspecified] = ACTIONS(5012),
    [anon_sym___autoreleasing] = ACTIONS(5012),
    [anon_sym___block] = ACTIONS(5012),
    [anon_sym___bridge] = ACTIONS(5012),
    [anon_sym___bridge_retained] = ACTIONS(5012),
    [anon_sym___bridge_transfer] = ACTIONS(5012),
    [anon_sym___complex] = ACTIONS(5012),
    [anon_sym___const] = ACTIONS(5012),
    [anon_sym___imag] = ACTIONS(5012),
    [anon_sym___kindof] = ACTIONS(5012),
    [anon_sym___nonnull] = ACTIONS(5012),
    [anon_sym___nullable] = ACTIONS(5012),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5012),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5012),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5012),
    [anon_sym___real] = ACTIONS(5012),
    [anon_sym___strong] = ACTIONS(5012),
    [anon_sym___unsafe_unretained] = ACTIONS(5012),
    [anon_sym___unused] = ACTIONS(5012),
    [anon_sym___weak] = ACTIONS(5012),
    [anon_sym_alignas] = ACTIONS(5012),
    [anon_sym__Alignas] = ACTIONS(5012),
    [sym_primitive_type] = ACTIONS(5012),
    [anon_sym_enum] = ACTIONS(5012),
    [anon_sym_COLON] = ACTIONS(5014),
    [anon_sym_struct] = ACTIONS(5012),
    [anon_sym_union] = ACTIONS(5012),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5012),
    [anon_sym___typeof] = ACTIONS(5012),
    [anon_sym_typeof] = ACTIONS(5012),
    [aux_sym_preproc_undef_token1] = ACTIONS(5012),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5012),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5012),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5012),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5012),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5012),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5012),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5012),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5012),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5012),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5012),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5012),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5012),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5012),
    [anon_sym_API_AVAILABLE] = ACTIONS(5012),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5012),
    [anon_sym_API_DEPRECATED] = ACTIONS(5012),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5012),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5012),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5012),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5012),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5012),
    [anon_sym___deprecated_msg] = ACTIONS(5012),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5012),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5012),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5012),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5012),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5012),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5012),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5012),
    [anon_sym_ATend] = ACTIONS(5014),
    [anon_sym_ATsynthesize] = ACTIONS(5014),
    [anon_sym_ATdynamic] = ACTIONS(5014),
    [anon_sym_ATproperty] = ACTIONS(5014),
    [anon_sym_BOOL] = ACTIONS(5012),
    [anon_sym_IMP] = ACTIONS(5012),
    [anon_sym_SEL] = ACTIONS(5012),
    [anon_sym_Class] = ACTIONS(5012),
    [anon_sym_id] = ACTIONS(5012),
  },
  [STATE(1899)] = {
    [sym_identifier] = ACTIONS(5016),
    [aux_sym_preproc_def_token1] = ACTIONS(5016),
    [anon_sym_COMMA] = ACTIONS(5018),
    [anon_sym_RPAREN] = ACTIONS(5018),
    [aux_sym_preproc_if_token1] = ACTIONS(5016),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5016),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5016),
    [sym_preproc_directive] = ACTIONS(5016),
    [anon_sym_LPAREN2] = ACTIONS(5018),
    [anon_sym_DASH] = ACTIONS(5018),
    [anon_sym_PLUS] = ACTIONS(5018),
    [anon_sym_LT] = ACTIONS(5018),
    [anon_sym_SEMI] = ACTIONS(5018),
    [anon_sym___extension__] = ACTIONS(5016),
    [anon_sym_typedef] = ACTIONS(5016),
    [anon_sym_extern] = ACTIONS(5016),
    [anon_sym___attribute__] = ACTIONS(5016),
    [anon_sym___attribute] = ACTIONS(5016),
    [anon_sym_noreturn] = ACTIONS(5016),
    [anon_sym_LBRACK] = ACTIONS(5018),
    [anon_sym___declspec] = ACTIONS(5016),
    [anon_sym___cdecl] = ACTIONS(5016),
    [anon_sym___clrcall] = ACTIONS(5016),
    [anon_sym___stdcall] = ACTIONS(5016),
    [anon_sym___fastcall] = ACTIONS(5016),
    [anon_sym___thiscall] = ACTIONS(5016),
    [anon_sym___vectorcall] = ACTIONS(5016),
    [anon_sym_LBRACE] = ACTIONS(5018),
    [anon_sym_signed] = ACTIONS(5016),
    [anon_sym_unsigned] = ACTIONS(5016),
    [anon_sym_long] = ACTIONS(5016),
    [anon_sym_short] = ACTIONS(5016),
    [anon_sym_static] = ACTIONS(5016),
    [anon_sym_auto] = ACTIONS(5016),
    [anon_sym_register] = ACTIONS(5016),
    [anon_sym_inline] = ACTIONS(5016),
    [anon_sym___inline] = ACTIONS(5016),
    [anon_sym___inline__] = ACTIONS(5016),
    [anon_sym___forceinline] = ACTIONS(5016),
    [anon_sym_thread_local] = ACTIONS(5016),
    [anon_sym___thread] = ACTIONS(5016),
    [anon_sym_CG_EXTERN] = ACTIONS(5016),
    [anon_sym_CG_INLINE] = ACTIONS(5016),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5016),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5016),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5016),
    [anon_sym_IBOutlet] = ACTIONS(5016),
    [anon_sym_IBInspectable] = ACTIONS(5016),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5016),
    [anon_sym_NS_INLINE] = ACTIONS(5016),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5016),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5016),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5016),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5016),
    [anon_sym_const] = ACTIONS(5016),
    [anon_sym_constexpr] = ACTIONS(5016),
    [anon_sym_volatile] = ACTIONS(5016),
    [anon_sym_restrict] = ACTIONS(5016),
    [anon_sym___restrict__] = ACTIONS(5016),
    [anon_sym__Atomic] = ACTIONS(5016),
    [anon_sym__Noreturn] = ACTIONS(5016),
    [anon_sym__Nonnull] = ACTIONS(5016),
    [anon_sym_nullable] = ACTIONS(5016),
    [anon_sym__Complex] = ACTIONS(5016),
    [anon_sym__Nullable] = ACTIONS(5016),
    [anon_sym__Nullable_result] = ACTIONS(5016),
    [anon_sym__Null_unspecified] = ACTIONS(5016),
    [anon_sym___autoreleasing] = ACTIONS(5016),
    [anon_sym___block] = ACTIONS(5016),
    [anon_sym___bridge] = ACTIONS(5016),
    [anon_sym___bridge_retained] = ACTIONS(5016),
    [anon_sym___bridge_transfer] = ACTIONS(5016),
    [anon_sym___complex] = ACTIONS(5016),
    [anon_sym___const] = ACTIONS(5016),
    [anon_sym___imag] = ACTIONS(5016),
    [anon_sym___kindof] = ACTIONS(5016),
    [anon_sym___nonnull] = ACTIONS(5016),
    [anon_sym___nullable] = ACTIONS(5016),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5016),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5016),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5016),
    [anon_sym___real] = ACTIONS(5016),
    [anon_sym___strong] = ACTIONS(5016),
    [anon_sym___unsafe_unretained] = ACTIONS(5016),
    [anon_sym___unused] = ACTIONS(5016),
    [anon_sym___weak] = ACTIONS(5016),
    [anon_sym_alignas] = ACTIONS(5016),
    [anon_sym__Alignas] = ACTIONS(5016),
    [sym_primitive_type] = ACTIONS(5016),
    [anon_sym_enum] = ACTIONS(5016),
    [anon_sym_COLON] = ACTIONS(5018),
    [anon_sym_struct] = ACTIONS(5016),
    [anon_sym_union] = ACTIONS(5016),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5016),
    [anon_sym___typeof] = ACTIONS(5016),
    [anon_sym_typeof] = ACTIONS(5016),
    [aux_sym_preproc_undef_token1] = ACTIONS(5016),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5016),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5016),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5016),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5016),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5016),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5016),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5016),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5016),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5016),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5016),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5016),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5016),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5016),
    [anon_sym_API_AVAILABLE] = ACTIONS(5016),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5016),
    [anon_sym_API_DEPRECATED] = ACTIONS(5016),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5016),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5016),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5016),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5016),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5016),
    [anon_sym___deprecated_msg] = ACTIONS(5016),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5016),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5016),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5016),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5016),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5016),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5016),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5016),
    [anon_sym_ATend] = ACTIONS(5018),
    [anon_sym_ATsynthesize] = ACTIONS(5018),
    [anon_sym_ATdynamic] = ACTIONS(5018),
    [anon_sym_ATproperty] = ACTIONS(5018),
    [anon_sym_BOOL] = ACTIONS(5016),
    [anon_sym_IMP] = ACTIONS(5016),
    [anon_sym_SEL] = ACTIONS(5016),
    [anon_sym_Class] = ACTIONS(5016),
    [anon_sym_id] = ACTIONS(5016),
  },
  [STATE(1900)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4470),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5020),
    [anon_sym_COMMA] = ACTIONS(4472),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4472),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4472),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(5026),
    [anon_sym_AMP_AMP] = ACTIONS(5028),
    [anon_sym_PIPE] = ACTIONS(5030),
    [anon_sym_CARET] = ACTIONS(5032),
    [anon_sym_AMP] = ACTIONS(5034),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4470),
    [anon_sym_extern] = ACTIONS(4470),
    [anon_sym___attribute__] = ACTIONS(4470),
    [anon_sym___attribute] = ACTIONS(4470),
    [anon_sym_noreturn] = ACTIONS(4470),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4470),
    [anon_sym_RBRACE] = ACTIONS(4472),
    [anon_sym_static] = ACTIONS(4470),
    [anon_sym_EQ] = ACTIONS(4470),
    [anon_sym_auto] = ACTIONS(4470),
    [anon_sym_register] = ACTIONS(4470),
    [anon_sym_inline] = ACTIONS(4470),
    [anon_sym___inline] = ACTIONS(4470),
    [anon_sym___inline__] = ACTIONS(4470),
    [anon_sym___forceinline] = ACTIONS(4470),
    [anon_sym_thread_local] = ACTIONS(4470),
    [anon_sym___thread] = ACTIONS(4470),
    [anon_sym_CG_EXTERN] = ACTIONS(4470),
    [anon_sym_CG_INLINE] = ACTIONS(4470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4470),
    [anon_sym_IBOutlet] = ACTIONS(4470),
    [anon_sym_IBInspectable] = ACTIONS(4470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4470),
    [anon_sym_NS_INLINE] = ACTIONS(4470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4470),
    [anon_sym_const] = ACTIONS(4470),
    [anon_sym_constexpr] = ACTIONS(4470),
    [anon_sym_volatile] = ACTIONS(4470),
    [anon_sym_restrict] = ACTIONS(4470),
    [anon_sym___restrict__] = ACTIONS(4470),
    [anon_sym__Atomic] = ACTIONS(4470),
    [anon_sym__Noreturn] = ACTIONS(4470),
    [anon_sym__Nonnull] = ACTIONS(4470),
    [anon_sym_nullable] = ACTIONS(4470),
    [anon_sym__Complex] = ACTIONS(4470),
    [anon_sym__Nullable] = ACTIONS(4470),
    [anon_sym__Nullable_result] = ACTIONS(4470),
    [anon_sym__Null_unspecified] = ACTIONS(4470),
    [anon_sym___autoreleasing] = ACTIONS(4470),
    [anon_sym___block] = ACTIONS(4470),
    [anon_sym___bridge] = ACTIONS(4470),
    [anon_sym___bridge_retained] = ACTIONS(4470),
    [anon_sym___bridge_transfer] = ACTIONS(4470),
    [anon_sym___complex] = ACTIONS(4470),
    [anon_sym___const] = ACTIONS(4470),
    [anon_sym___imag] = ACTIONS(4470),
    [anon_sym___kindof] = ACTIONS(4470),
    [anon_sym___nonnull] = ACTIONS(4470),
    [anon_sym___nullable] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4470),
    [anon_sym___real] = ACTIONS(4470),
    [anon_sym___strong] = ACTIONS(4470),
    [anon_sym___unsafe_unretained] = ACTIONS(4470),
    [anon_sym___unused] = ACTIONS(4470),
    [anon_sym___weak] = ACTIONS(4470),
    [anon_sym_alignas] = ACTIONS(4470),
    [anon_sym__Alignas] = ACTIONS(4470),
    [anon_sym_QMARK] = ACTIONS(4472),
    [anon_sym_STAR_EQ] = ACTIONS(4472),
    [anon_sym_SLASH_EQ] = ACTIONS(4472),
    [anon_sym_PERCENT_EQ] = ACTIONS(4472),
    [anon_sym_PLUS_EQ] = ACTIONS(4472),
    [anon_sym_DASH_EQ] = ACTIONS(4472),
    [anon_sym_LT_LT_EQ] = ACTIONS(4472),
    [anon_sym_GT_GT_EQ] = ACTIONS(4472),
    [anon_sym_AMP_EQ] = ACTIONS(4472),
    [anon_sym_CARET_EQ] = ACTIONS(4472),
    [anon_sym_PIPE_EQ] = ACTIONS(4472),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_API_AVAILABLE] = ACTIONS(4470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_API_DEPRECATED] = ACTIONS(4470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym___deprecated_msg] = ACTIONS(4470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4470),
  },
  [STATE(1901)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4000),
    [anon_sym_PLUS] = ACTIONS(4000),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4000),
    [anon_sym_GT_GT] = ACTIONS(4000),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(1902)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(5028),
    [anon_sym_PIPE] = ACTIONS(5030),
    [anon_sym_CARET] = ACTIONS(5032),
    [anon_sym_AMP] = ACTIONS(5034),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(1903)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(5030),
    [anon_sym_CARET] = ACTIONS(5032),
    [anon_sym_AMP] = ACTIONS(5034),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(1904)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(5032),
    [anon_sym_AMP] = ACTIONS(5034),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(1905)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(5034),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(1906)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(1907)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(1908)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(1909)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4000),
    [anon_sym_GT_GT] = ACTIONS(4000),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(1910)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4478),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5020),
    [anon_sym_COMMA] = ACTIONS(4480),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4480),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4480),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(5026),
    [anon_sym_AMP_AMP] = ACTIONS(5028),
    [anon_sym_PIPE] = ACTIONS(5030),
    [anon_sym_CARET] = ACTIONS(5032),
    [anon_sym_AMP] = ACTIONS(5034),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4478),
    [anon_sym_extern] = ACTIONS(4478),
    [anon_sym___attribute__] = ACTIONS(4478),
    [anon_sym___attribute] = ACTIONS(4478),
    [anon_sym_noreturn] = ACTIONS(4478),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4478),
    [anon_sym_RBRACE] = ACTIONS(4480),
    [anon_sym_static] = ACTIONS(4478),
    [anon_sym_EQ] = ACTIONS(4478),
    [anon_sym_auto] = ACTIONS(4478),
    [anon_sym_register] = ACTIONS(4478),
    [anon_sym_inline] = ACTIONS(4478),
    [anon_sym___inline] = ACTIONS(4478),
    [anon_sym___inline__] = ACTIONS(4478),
    [anon_sym___forceinline] = ACTIONS(4478),
    [anon_sym_thread_local] = ACTIONS(4478),
    [anon_sym___thread] = ACTIONS(4478),
    [anon_sym_CG_EXTERN] = ACTIONS(4478),
    [anon_sym_CG_INLINE] = ACTIONS(4478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4478),
    [anon_sym_IBOutlet] = ACTIONS(4478),
    [anon_sym_IBInspectable] = ACTIONS(4478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4478),
    [anon_sym_NS_INLINE] = ACTIONS(4478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4478),
    [anon_sym_const] = ACTIONS(4478),
    [anon_sym_constexpr] = ACTIONS(4478),
    [anon_sym_volatile] = ACTIONS(4478),
    [anon_sym_restrict] = ACTIONS(4478),
    [anon_sym___restrict__] = ACTIONS(4478),
    [anon_sym__Atomic] = ACTIONS(4478),
    [anon_sym__Noreturn] = ACTIONS(4478),
    [anon_sym__Nonnull] = ACTIONS(4478),
    [anon_sym_nullable] = ACTIONS(4478),
    [anon_sym__Complex] = ACTIONS(4478),
    [anon_sym__Nullable] = ACTIONS(4478),
    [anon_sym__Nullable_result] = ACTIONS(4478),
    [anon_sym__Null_unspecified] = ACTIONS(4478),
    [anon_sym___autoreleasing] = ACTIONS(4478),
    [anon_sym___block] = ACTIONS(4478),
    [anon_sym___bridge] = ACTIONS(4478),
    [anon_sym___bridge_retained] = ACTIONS(4478),
    [anon_sym___bridge_transfer] = ACTIONS(4478),
    [anon_sym___complex] = ACTIONS(4478),
    [anon_sym___const] = ACTIONS(4478),
    [anon_sym___imag] = ACTIONS(4478),
    [anon_sym___kindof] = ACTIONS(4478),
    [anon_sym___nonnull] = ACTIONS(4478),
    [anon_sym___nullable] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4478),
    [anon_sym___real] = ACTIONS(4478),
    [anon_sym___strong] = ACTIONS(4478),
    [anon_sym___unsafe_unretained] = ACTIONS(4478),
    [anon_sym___unused] = ACTIONS(4478),
    [anon_sym___weak] = ACTIONS(4478),
    [anon_sym_alignas] = ACTIONS(4478),
    [anon_sym__Alignas] = ACTIONS(4478),
    [anon_sym_QMARK] = ACTIONS(5044),
    [anon_sym_STAR_EQ] = ACTIONS(4480),
    [anon_sym_SLASH_EQ] = ACTIONS(4480),
    [anon_sym_PERCENT_EQ] = ACTIONS(4480),
    [anon_sym_PLUS_EQ] = ACTIONS(4480),
    [anon_sym_DASH_EQ] = ACTIONS(4480),
    [anon_sym_LT_LT_EQ] = ACTIONS(4480),
    [anon_sym_GT_GT_EQ] = ACTIONS(4480),
    [anon_sym_AMP_EQ] = ACTIONS(4480),
    [anon_sym_CARET_EQ] = ACTIONS(4480),
    [anon_sym_PIPE_EQ] = ACTIONS(4480),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_API_AVAILABLE] = ACTIONS(4478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_API_DEPRECATED] = ACTIONS(4478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym___deprecated_msg] = ACTIONS(4478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4478),
  },
  [STATE(1911)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4458),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5020),
    [anon_sym_COMMA] = ACTIONS(4462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4462),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(5026),
    [anon_sym_AMP_AMP] = ACTIONS(5028),
    [anon_sym_PIPE] = ACTIONS(5030),
    [anon_sym_CARET] = ACTIONS(5032),
    [anon_sym_AMP] = ACTIONS(5034),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4458),
    [anon_sym_extern] = ACTIONS(4458),
    [anon_sym___attribute__] = ACTIONS(4458),
    [anon_sym___attribute] = ACTIONS(4458),
    [anon_sym_noreturn] = ACTIONS(4458),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4458),
    [anon_sym_RBRACE] = ACTIONS(4462),
    [anon_sym_static] = ACTIONS(4458),
    [anon_sym_EQ] = ACTIONS(4458),
    [anon_sym_auto] = ACTIONS(4458),
    [anon_sym_register] = ACTIONS(4458),
    [anon_sym_inline] = ACTIONS(4458),
    [anon_sym___inline] = ACTIONS(4458),
    [anon_sym___inline__] = ACTIONS(4458),
    [anon_sym___forceinline] = ACTIONS(4458),
    [anon_sym_thread_local] = ACTIONS(4458),
    [anon_sym___thread] = ACTIONS(4458),
    [anon_sym_CG_EXTERN] = ACTIONS(4458),
    [anon_sym_CG_INLINE] = ACTIONS(4458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4458),
    [anon_sym_IBOutlet] = ACTIONS(4458),
    [anon_sym_IBInspectable] = ACTIONS(4458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4458),
    [anon_sym_NS_INLINE] = ACTIONS(4458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4458),
    [anon_sym_const] = ACTIONS(4458),
    [anon_sym_constexpr] = ACTIONS(4458),
    [anon_sym_volatile] = ACTIONS(4458),
    [anon_sym_restrict] = ACTIONS(4458),
    [anon_sym___restrict__] = ACTIONS(4458),
    [anon_sym__Atomic] = ACTIONS(4458),
    [anon_sym__Noreturn] = ACTIONS(4458),
    [anon_sym__Nonnull] = ACTIONS(4458),
    [anon_sym_nullable] = ACTIONS(4458),
    [anon_sym__Complex] = ACTIONS(4458),
    [anon_sym__Nullable] = ACTIONS(4458),
    [anon_sym__Nullable_result] = ACTIONS(4458),
    [anon_sym__Null_unspecified] = ACTIONS(4458),
    [anon_sym___autoreleasing] = ACTIONS(4458),
    [anon_sym___block] = ACTIONS(4458),
    [anon_sym___bridge] = ACTIONS(4458),
    [anon_sym___bridge_retained] = ACTIONS(4458),
    [anon_sym___bridge_transfer] = ACTIONS(4458),
    [anon_sym___complex] = ACTIONS(4458),
    [anon_sym___const] = ACTIONS(4458),
    [anon_sym___imag] = ACTIONS(4458),
    [anon_sym___kindof] = ACTIONS(4458),
    [anon_sym___nonnull] = ACTIONS(4458),
    [anon_sym___nullable] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4458),
    [anon_sym___real] = ACTIONS(4458),
    [anon_sym___strong] = ACTIONS(4458),
    [anon_sym___unsafe_unretained] = ACTIONS(4458),
    [anon_sym___unused] = ACTIONS(4458),
    [anon_sym___weak] = ACTIONS(4458),
    [anon_sym_alignas] = ACTIONS(4458),
    [anon_sym__Alignas] = ACTIONS(4458),
    [anon_sym_QMARK] = ACTIONS(5044),
    [anon_sym_STAR_EQ] = ACTIONS(4462),
    [anon_sym_SLASH_EQ] = ACTIONS(4462),
    [anon_sym_PERCENT_EQ] = ACTIONS(4462),
    [anon_sym_PLUS_EQ] = ACTIONS(4462),
    [anon_sym_DASH_EQ] = ACTIONS(4462),
    [anon_sym_LT_LT_EQ] = ACTIONS(4462),
    [anon_sym_GT_GT_EQ] = ACTIONS(4462),
    [anon_sym_AMP_EQ] = ACTIONS(4462),
    [anon_sym_CARET_EQ] = ACTIONS(4462),
    [anon_sym_PIPE_EQ] = ACTIONS(4462),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_API_AVAILABLE] = ACTIONS(4458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_API_DEPRECATED] = ACTIONS(4458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym___deprecated_msg] = ACTIONS(4458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4458),
  },
  [STATE(1912)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5046),
    [anon_sym_COMMA] = ACTIONS(4436),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4436),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4436),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(5026),
    [anon_sym_AMP_AMP] = ACTIONS(5028),
    [anon_sym_PIPE] = ACTIONS(5030),
    [anon_sym_CARET] = ACTIONS(5032),
    [anon_sym_AMP] = ACTIONS(5034),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4431),
    [anon_sym_extern] = ACTIONS(4431),
    [anon_sym___attribute__] = ACTIONS(4431),
    [anon_sym___attribute] = ACTIONS(4431),
    [anon_sym_noreturn] = ACTIONS(4431),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4431),
    [anon_sym_RBRACE] = ACTIONS(4436),
    [anon_sym_static] = ACTIONS(4431),
    [anon_sym_EQ] = ACTIONS(4431),
    [anon_sym_auto] = ACTIONS(4431),
    [anon_sym_register] = ACTIONS(4431),
    [anon_sym_inline] = ACTIONS(4431),
    [anon_sym___inline] = ACTIONS(4431),
    [anon_sym___inline__] = ACTIONS(4431),
    [anon_sym___forceinline] = ACTIONS(4431),
    [anon_sym_thread_local] = ACTIONS(4431),
    [anon_sym___thread] = ACTIONS(4431),
    [anon_sym_CG_EXTERN] = ACTIONS(4431),
    [anon_sym_CG_INLINE] = ACTIONS(4431),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4431),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4431),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4431),
    [anon_sym_IBOutlet] = ACTIONS(4431),
    [anon_sym_IBInspectable] = ACTIONS(4431),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4431),
    [anon_sym_NS_INLINE] = ACTIONS(4431),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4431),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4431),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4431),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4431),
    [anon_sym_const] = ACTIONS(4431),
    [anon_sym_constexpr] = ACTIONS(4431),
    [anon_sym_volatile] = ACTIONS(4431),
    [anon_sym_restrict] = ACTIONS(4431),
    [anon_sym___restrict__] = ACTIONS(4431),
    [anon_sym__Atomic] = ACTIONS(4431),
    [anon_sym__Noreturn] = ACTIONS(4431),
    [anon_sym__Nonnull] = ACTIONS(4431),
    [anon_sym_nullable] = ACTIONS(4431),
    [anon_sym__Complex] = ACTIONS(4431),
    [anon_sym__Nullable] = ACTIONS(4431),
    [anon_sym__Nullable_result] = ACTIONS(4431),
    [anon_sym__Null_unspecified] = ACTIONS(4431),
    [anon_sym___autoreleasing] = ACTIONS(4431),
    [anon_sym___block] = ACTIONS(4431),
    [anon_sym___bridge] = ACTIONS(4431),
    [anon_sym___bridge_retained] = ACTIONS(4431),
    [anon_sym___bridge_transfer] = ACTIONS(4431),
    [anon_sym___complex] = ACTIONS(4431),
    [anon_sym___const] = ACTIONS(4431),
    [anon_sym___imag] = ACTIONS(4431),
    [anon_sym___kindof] = ACTIONS(4431),
    [anon_sym___nonnull] = ACTIONS(4431),
    [anon_sym___nullable] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4431),
    [anon_sym___real] = ACTIONS(4431),
    [anon_sym___strong] = ACTIONS(4431),
    [anon_sym___unsafe_unretained] = ACTIONS(4431),
    [anon_sym___unused] = ACTIONS(4431),
    [anon_sym___weak] = ACTIONS(4431),
    [anon_sym_alignas] = ACTIONS(4431),
    [anon_sym__Alignas] = ACTIONS(4431),
    [anon_sym_QMARK] = ACTIONS(4436),
    [anon_sym_STAR_EQ] = ACTIONS(4436),
    [anon_sym_SLASH_EQ] = ACTIONS(4436),
    [anon_sym_PERCENT_EQ] = ACTIONS(4436),
    [anon_sym_PLUS_EQ] = ACTIONS(4436),
    [anon_sym_DASH_EQ] = ACTIONS(4436),
    [anon_sym_LT_LT_EQ] = ACTIONS(4436),
    [anon_sym_GT_GT_EQ] = ACTIONS(4436),
    [anon_sym_AMP_EQ] = ACTIONS(4436),
    [anon_sym_CARET_EQ] = ACTIONS(4436),
    [anon_sym_PIPE_EQ] = ACTIONS(4436),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4431),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4431),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4431),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4431),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4431),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4431),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4431),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_API_AVAILABLE] = ACTIONS(4431),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_API_DEPRECATED] = ACTIONS(4431),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4431),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym___deprecated_msg] = ACTIONS(4431),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4431),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4431),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4431),
  },
  [STATE(1913)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4474),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5020),
    [anon_sym_COMMA] = ACTIONS(4476),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4476),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4476),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(5026),
    [anon_sym_AMP_AMP] = ACTIONS(5028),
    [anon_sym_PIPE] = ACTIONS(5030),
    [anon_sym_CARET] = ACTIONS(5032),
    [anon_sym_AMP] = ACTIONS(5034),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4474),
    [anon_sym_extern] = ACTIONS(4474),
    [anon_sym___attribute__] = ACTIONS(4474),
    [anon_sym___attribute] = ACTIONS(4474),
    [anon_sym_noreturn] = ACTIONS(4474),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4474),
    [anon_sym_RBRACE] = ACTIONS(4476),
    [anon_sym_static] = ACTIONS(4474),
    [anon_sym_EQ] = ACTIONS(4474),
    [anon_sym_auto] = ACTIONS(4474),
    [anon_sym_register] = ACTIONS(4474),
    [anon_sym_inline] = ACTIONS(4474),
    [anon_sym___inline] = ACTIONS(4474),
    [anon_sym___inline__] = ACTIONS(4474),
    [anon_sym___forceinline] = ACTIONS(4474),
    [anon_sym_thread_local] = ACTIONS(4474),
    [anon_sym___thread] = ACTIONS(4474),
    [anon_sym_CG_EXTERN] = ACTIONS(4474),
    [anon_sym_CG_INLINE] = ACTIONS(4474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4474),
    [anon_sym_IBOutlet] = ACTIONS(4474),
    [anon_sym_IBInspectable] = ACTIONS(4474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4474),
    [anon_sym_NS_INLINE] = ACTIONS(4474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4474),
    [anon_sym_const] = ACTIONS(4474),
    [anon_sym_constexpr] = ACTIONS(4474),
    [anon_sym_volatile] = ACTIONS(4474),
    [anon_sym_restrict] = ACTIONS(4474),
    [anon_sym___restrict__] = ACTIONS(4474),
    [anon_sym__Atomic] = ACTIONS(4474),
    [anon_sym__Noreturn] = ACTIONS(4474),
    [anon_sym__Nonnull] = ACTIONS(4474),
    [anon_sym_nullable] = ACTIONS(4474),
    [anon_sym__Complex] = ACTIONS(4474),
    [anon_sym__Nullable] = ACTIONS(4474),
    [anon_sym__Nullable_result] = ACTIONS(4474),
    [anon_sym__Null_unspecified] = ACTIONS(4474),
    [anon_sym___autoreleasing] = ACTIONS(4474),
    [anon_sym___block] = ACTIONS(4474),
    [anon_sym___bridge] = ACTIONS(4474),
    [anon_sym___bridge_retained] = ACTIONS(4474),
    [anon_sym___bridge_transfer] = ACTIONS(4474),
    [anon_sym___complex] = ACTIONS(4474),
    [anon_sym___const] = ACTIONS(4474),
    [anon_sym___imag] = ACTIONS(4474),
    [anon_sym___kindof] = ACTIONS(4474),
    [anon_sym___nonnull] = ACTIONS(4474),
    [anon_sym___nullable] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4474),
    [anon_sym___real] = ACTIONS(4474),
    [anon_sym___strong] = ACTIONS(4474),
    [anon_sym___unsafe_unretained] = ACTIONS(4474),
    [anon_sym___unused] = ACTIONS(4474),
    [anon_sym___weak] = ACTIONS(4474),
    [anon_sym_alignas] = ACTIONS(4474),
    [anon_sym__Alignas] = ACTIONS(4474),
    [anon_sym_QMARK] = ACTIONS(4476),
    [anon_sym_STAR_EQ] = ACTIONS(4476),
    [anon_sym_SLASH_EQ] = ACTIONS(4476),
    [anon_sym_PERCENT_EQ] = ACTIONS(4476),
    [anon_sym_PLUS_EQ] = ACTIONS(4476),
    [anon_sym_DASH_EQ] = ACTIONS(4476),
    [anon_sym_LT_LT_EQ] = ACTIONS(4476),
    [anon_sym_GT_GT_EQ] = ACTIONS(4476),
    [anon_sym_AMP_EQ] = ACTIONS(4476),
    [anon_sym_CARET_EQ] = ACTIONS(4476),
    [anon_sym_PIPE_EQ] = ACTIONS(4476),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_API_AVAILABLE] = ACTIONS(4474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_API_DEPRECATED] = ACTIONS(4474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym___deprecated_msg] = ACTIONS(4474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4474),
  },
  [STATE(1914)] = {
    [sym_string_literal] = STATE(1915),
    [aux_sym_concatenated_string_repeat1] = STATE(1915),
    [sym_identifier] = ACTIONS(5049),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3944),
    [anon_sym_COMMA] = ACTIONS(3944),
    [anon_sym_RPAREN] = ACTIONS(3944),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3944),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3944),
    [anon_sym_LPAREN2] = ACTIONS(3944),
    [anon_sym_DASH] = ACTIONS(3946),
    [anon_sym_PLUS] = ACTIONS(3946),
    [anon_sym_STAR] = ACTIONS(3944),
    [anon_sym_SLASH] = ACTIONS(3946),
    [anon_sym_PERCENT] = ACTIONS(3944),
    [anon_sym_PIPE_PIPE] = ACTIONS(3944),
    [anon_sym_AMP_AMP] = ACTIONS(3944),
    [anon_sym_PIPE] = ACTIONS(3946),
    [anon_sym_CARET] = ACTIONS(3944),
    [anon_sym_AMP] = ACTIONS(3946),
    [anon_sym_EQ_EQ] = ACTIONS(3944),
    [anon_sym_BANG_EQ] = ACTIONS(3944),
    [anon_sym_GT] = ACTIONS(3946),
    [anon_sym_GT_EQ] = ACTIONS(3944),
    [anon_sym_LT_EQ] = ACTIONS(3944),
    [anon_sym_LT] = ACTIONS(3946),
    [anon_sym_LT_LT] = ACTIONS(3944),
    [anon_sym_GT_GT] = ACTIONS(3944),
    [anon_sym_SEMI] = ACTIONS(3944),
    [anon_sym___extension__] = ACTIONS(3946),
    [anon_sym_extern] = ACTIONS(3946),
    [anon_sym___attribute__] = ACTIONS(3946),
    [anon_sym___attribute] = ACTIONS(3946),
    [anon_sym_noreturn] = ACTIONS(3946),
    [anon_sym_LBRACK] = ACTIONS(3944),
    [anon_sym_RBRACK] = ACTIONS(3944),
    [anon_sym___declspec] = ACTIONS(3946),
    [anon_sym_RBRACE] = ACTIONS(3944),
    [anon_sym_static] = ACTIONS(3946),
    [anon_sym_auto] = ACTIONS(3946),
    [anon_sym_register] = ACTIONS(3946),
    [anon_sym_inline] = ACTIONS(3946),
    [anon_sym___inline] = ACTIONS(3946),
    [anon_sym___inline__] = ACTIONS(3946),
    [anon_sym___forceinline] = ACTIONS(3946),
    [anon_sym_thread_local] = ACTIONS(3946),
    [anon_sym___thread] = ACTIONS(3946),
    [anon_sym_CG_EXTERN] = ACTIONS(3946),
    [anon_sym_CG_INLINE] = ACTIONS(3946),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3946),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3946),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3946),
    [anon_sym_IBOutlet] = ACTIONS(3946),
    [anon_sym_IBInspectable] = ACTIONS(3946),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3946),
    [anon_sym_NS_INLINE] = ACTIONS(3946),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3946),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3946),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3946),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3946),
    [anon_sym_const] = ACTIONS(3946),
    [anon_sym_constexpr] = ACTIONS(3946),
    [anon_sym_volatile] = ACTIONS(3946),
    [anon_sym_restrict] = ACTIONS(3946),
    [anon_sym___restrict__] = ACTIONS(3946),
    [anon_sym__Atomic] = ACTIONS(3946),
    [anon_sym__Noreturn] = ACTIONS(3946),
    [anon_sym__Nonnull] = ACTIONS(3946),
    [anon_sym_nullable] = ACTIONS(3946),
    [anon_sym__Complex] = ACTIONS(3946),
    [anon_sym__Nullable] = ACTIONS(3946),
    [anon_sym__Nullable_result] = ACTIONS(3946),
    [anon_sym__Null_unspecified] = ACTIONS(3946),
    [anon_sym___autoreleasing] = ACTIONS(3946),
    [anon_sym___block] = ACTIONS(3946),
    [anon_sym___bridge] = ACTIONS(3946),
    [anon_sym___bridge_retained] = ACTIONS(3946),
    [anon_sym___bridge_transfer] = ACTIONS(3946),
    [anon_sym___complex] = ACTIONS(3946),
    [anon_sym___const] = ACTIONS(3946),
    [anon_sym___imag] = ACTIONS(3946),
    [anon_sym___kindof] = ACTIONS(3946),
    [anon_sym___nonnull] = ACTIONS(3946),
    [anon_sym___nullable] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3946),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3946),
    [anon_sym___real] = ACTIONS(3946),
    [anon_sym___strong] = ACTIONS(3946),
    [anon_sym___unsafe_unretained] = ACTIONS(3946),
    [anon_sym___unused] = ACTIONS(3946),
    [anon_sym___weak] = ACTIONS(3946),
    [anon_sym_alignas] = ACTIONS(3946),
    [anon_sym__Alignas] = ACTIONS(3946),
    [anon_sym_COLON] = ACTIONS(3944),
    [anon_sym_QMARK] = ACTIONS(3944),
    [anon_sym_DASH_DASH] = ACTIONS(3944),
    [anon_sym_PLUS_PLUS] = ACTIONS(3944),
    [anon_sym_DOT] = ACTIONS(3946),
    [anon_sym_DASH_GT] = ACTIONS(3944),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3946),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3946),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3946),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3946),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3946),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3946),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3946),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_API_AVAILABLE] = ACTIONS(3946),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_API_DEPRECATED] = ACTIONS(3946),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3946),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3946),
    [anon_sym___deprecated_msg] = ACTIONS(3946),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3946),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3946),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3946),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3946),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3946),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3946),
  },
  [STATE(1915)] = {
    [sym_string_literal] = STATE(1915),
    [aux_sym_concatenated_string_repeat1] = STATE(1915),
    [sym_identifier] = ACTIONS(5051),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_RPAREN] = ACTIONS(3953),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3953),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3953),
    [anon_sym_LPAREN2] = ACTIONS(3953),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_STAR] = ACTIONS(3953),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_PERCENT] = ACTIONS(3953),
    [anon_sym_PIPE_PIPE] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3953),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3953),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_EQ_EQ] = ACTIONS(3953),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_LT_EQ] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3953),
    [anon_sym_GT_GT] = ACTIONS(3953),
    [anon_sym_SEMI] = ACTIONS(3953),
    [anon_sym___extension__] = ACTIONS(3955),
    [anon_sym_extern] = ACTIONS(3955),
    [anon_sym___attribute__] = ACTIONS(3955),
    [anon_sym___attribute] = ACTIONS(3955),
    [anon_sym_noreturn] = ACTIONS(3955),
    [anon_sym_LBRACK] = ACTIONS(3953),
    [anon_sym_RBRACK] = ACTIONS(3953),
    [anon_sym___declspec] = ACTIONS(3955),
    [anon_sym_RBRACE] = ACTIONS(3953),
    [anon_sym_static] = ACTIONS(3955),
    [anon_sym_auto] = ACTIONS(3955),
    [anon_sym_register] = ACTIONS(3955),
    [anon_sym_inline] = ACTIONS(3955),
    [anon_sym___inline] = ACTIONS(3955),
    [anon_sym___inline__] = ACTIONS(3955),
    [anon_sym___forceinline] = ACTIONS(3955),
    [anon_sym_thread_local] = ACTIONS(3955),
    [anon_sym___thread] = ACTIONS(3955),
    [anon_sym_CG_EXTERN] = ACTIONS(3955),
    [anon_sym_CG_INLINE] = ACTIONS(3955),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3955),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3955),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3955),
    [anon_sym_IBOutlet] = ACTIONS(3955),
    [anon_sym_IBInspectable] = ACTIONS(3955),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3955),
    [anon_sym_NS_INLINE] = ACTIONS(3955),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3955),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3955),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3955),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3955),
    [anon_sym_const] = ACTIONS(3955),
    [anon_sym_constexpr] = ACTIONS(3955),
    [anon_sym_volatile] = ACTIONS(3955),
    [anon_sym_restrict] = ACTIONS(3955),
    [anon_sym___restrict__] = ACTIONS(3955),
    [anon_sym__Atomic] = ACTIONS(3955),
    [anon_sym__Noreturn] = ACTIONS(3955),
    [anon_sym__Nonnull] = ACTIONS(3955),
    [anon_sym_nullable] = ACTIONS(3955),
    [anon_sym__Complex] = ACTIONS(3955),
    [anon_sym__Nullable] = ACTIONS(3955),
    [anon_sym__Nullable_result] = ACTIONS(3955),
    [anon_sym__Null_unspecified] = ACTIONS(3955),
    [anon_sym___autoreleasing] = ACTIONS(3955),
    [anon_sym___block] = ACTIONS(3955),
    [anon_sym___bridge] = ACTIONS(3955),
    [anon_sym___bridge_retained] = ACTIONS(3955),
    [anon_sym___bridge_transfer] = ACTIONS(3955),
    [anon_sym___complex] = ACTIONS(3955),
    [anon_sym___const] = ACTIONS(3955),
    [anon_sym___imag] = ACTIONS(3955),
    [anon_sym___kindof] = ACTIONS(3955),
    [anon_sym___nonnull] = ACTIONS(3955),
    [anon_sym___nullable] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3955),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3955),
    [anon_sym___real] = ACTIONS(3955),
    [anon_sym___strong] = ACTIONS(3955),
    [anon_sym___unsafe_unretained] = ACTIONS(3955),
    [anon_sym___unused] = ACTIONS(3955),
    [anon_sym___weak] = ACTIONS(3955),
    [anon_sym_alignas] = ACTIONS(3955),
    [anon_sym__Alignas] = ACTIONS(3955),
    [anon_sym_COLON] = ACTIONS(3953),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_AT] = ACTIONS(5054),
    [anon_sym_DQUOTE] = ACTIONS(5057),
    [anon_sym_L_DQUOTE] = ACTIONS(5057),
    [anon_sym_u_DQUOTE] = ACTIONS(5057),
    [anon_sym_U_DQUOTE] = ACTIONS(5057),
    [anon_sym_u8_DQUOTE] = ACTIONS(5057),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3955),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3955),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3955),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3955),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3955),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3955),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3955),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_API_AVAILABLE] = ACTIONS(3955),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_API_DEPRECATED] = ACTIONS(3955),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3955),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3955),
    [anon_sym___deprecated_msg] = ACTIONS(3955),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3955),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3955),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3955),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3955),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3955),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3955),
  },
  [STATE(1916)] = {
    [sym_string_literal] = STATE(1914),
    [aux_sym_concatenated_string_repeat1] = STATE(1914),
    [sym_identifier] = ACTIONS(5060),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_RPAREN] = ACTIONS(3965),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3965),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3965),
    [anon_sym_LPAREN2] = ACTIONS(3965),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_STAR] = ACTIONS(3965),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_PERCENT] = ACTIONS(3965),
    [anon_sym_PIPE_PIPE] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3965),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3965),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_EQ_EQ] = ACTIONS(3965),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_LT_EQ] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3965),
    [anon_sym_GT_GT] = ACTIONS(3965),
    [anon_sym_SEMI] = ACTIONS(3965),
    [anon_sym___extension__] = ACTIONS(3967),
    [anon_sym_extern] = ACTIONS(3967),
    [anon_sym___attribute__] = ACTIONS(3967),
    [anon_sym___attribute] = ACTIONS(3967),
    [anon_sym_noreturn] = ACTIONS(3967),
    [anon_sym_LBRACK] = ACTIONS(3965),
    [anon_sym_RBRACK] = ACTIONS(3965),
    [anon_sym___declspec] = ACTIONS(3967),
    [anon_sym_RBRACE] = ACTIONS(3965),
    [anon_sym_static] = ACTIONS(3967),
    [anon_sym_auto] = ACTIONS(3967),
    [anon_sym_register] = ACTIONS(3967),
    [anon_sym_inline] = ACTIONS(3967),
    [anon_sym___inline] = ACTIONS(3967),
    [anon_sym___inline__] = ACTIONS(3967),
    [anon_sym___forceinline] = ACTIONS(3967),
    [anon_sym_thread_local] = ACTIONS(3967),
    [anon_sym___thread] = ACTIONS(3967),
    [anon_sym_CG_EXTERN] = ACTIONS(3967),
    [anon_sym_CG_INLINE] = ACTIONS(3967),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3967),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3967),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3967),
    [anon_sym_IBOutlet] = ACTIONS(3967),
    [anon_sym_IBInspectable] = ACTIONS(3967),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3967),
    [anon_sym_NS_INLINE] = ACTIONS(3967),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3967),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3967),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3967),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3967),
    [anon_sym_const] = ACTIONS(3967),
    [anon_sym_constexpr] = ACTIONS(3967),
    [anon_sym_volatile] = ACTIONS(3967),
    [anon_sym_restrict] = ACTIONS(3967),
    [anon_sym___restrict__] = ACTIONS(3967),
    [anon_sym__Atomic] = ACTIONS(3967),
    [anon_sym__Noreturn] = ACTIONS(3967),
    [anon_sym__Nonnull] = ACTIONS(3967),
    [anon_sym_nullable] = ACTIONS(3967),
    [anon_sym__Complex] = ACTIONS(3967),
    [anon_sym__Nullable] = ACTIONS(3967),
    [anon_sym__Nullable_result] = ACTIONS(3967),
    [anon_sym__Null_unspecified] = ACTIONS(3967),
    [anon_sym___autoreleasing] = ACTIONS(3967),
    [anon_sym___block] = ACTIONS(3967),
    [anon_sym___bridge] = ACTIONS(3967),
    [anon_sym___bridge_retained] = ACTIONS(3967),
    [anon_sym___bridge_transfer] = ACTIONS(3967),
    [anon_sym___complex] = ACTIONS(3967),
    [anon_sym___const] = ACTIONS(3967),
    [anon_sym___imag] = ACTIONS(3967),
    [anon_sym___kindof] = ACTIONS(3967),
    [anon_sym___nonnull] = ACTIONS(3967),
    [anon_sym___nullable] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3967),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3967),
    [anon_sym___real] = ACTIONS(3967),
    [anon_sym___strong] = ACTIONS(3967),
    [anon_sym___unsafe_unretained] = ACTIONS(3967),
    [anon_sym___unused] = ACTIONS(3967),
    [anon_sym___weak] = ACTIONS(3967),
    [anon_sym_alignas] = ACTIONS(3967),
    [anon_sym__Alignas] = ACTIONS(3967),
    [anon_sym_COLON] = ACTIONS(3965),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3967),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3967),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3967),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3967),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3967),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3967),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3967),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_API_AVAILABLE] = ACTIONS(3967),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_API_DEPRECATED] = ACTIONS(3967),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3967),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3967),
    [anon_sym___deprecated_msg] = ACTIONS(3967),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3967),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3967),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3967),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3967),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3967),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3967),
  },
  [STATE(1917)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4466),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5020),
    [anon_sym_COMMA] = ACTIONS(4468),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4468),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4468),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5022),
    [anon_sym_PLUS] = ACTIONS(5022),
    [anon_sym_STAR] = ACTIONS(5024),
    [anon_sym_SLASH] = ACTIONS(5024),
    [anon_sym_PERCENT] = ACTIONS(5024),
    [anon_sym_PIPE_PIPE] = ACTIONS(5026),
    [anon_sym_AMP_AMP] = ACTIONS(5028),
    [anon_sym_PIPE] = ACTIONS(5030),
    [anon_sym_CARET] = ACTIONS(5032),
    [anon_sym_AMP] = ACTIONS(5034),
    [anon_sym_EQ_EQ] = ACTIONS(5036),
    [anon_sym_BANG_EQ] = ACTIONS(5036),
    [anon_sym_GT] = ACTIONS(5038),
    [anon_sym_GT_EQ] = ACTIONS(5040),
    [anon_sym_LT_EQ] = ACTIONS(5040),
    [anon_sym_LT] = ACTIONS(5038),
    [anon_sym_LT_LT] = ACTIONS(5042),
    [anon_sym_GT_GT] = ACTIONS(5042),
    [anon_sym___extension__] = ACTIONS(4466),
    [anon_sym_extern] = ACTIONS(4466),
    [anon_sym___attribute__] = ACTIONS(4466),
    [anon_sym___attribute] = ACTIONS(4466),
    [anon_sym_noreturn] = ACTIONS(4466),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4466),
    [anon_sym_RBRACE] = ACTIONS(4468),
    [anon_sym_static] = ACTIONS(4466),
    [anon_sym_EQ] = ACTIONS(4466),
    [anon_sym_auto] = ACTIONS(4466),
    [anon_sym_register] = ACTIONS(4466),
    [anon_sym_inline] = ACTIONS(4466),
    [anon_sym___inline] = ACTIONS(4466),
    [anon_sym___inline__] = ACTIONS(4466),
    [anon_sym___forceinline] = ACTIONS(4466),
    [anon_sym_thread_local] = ACTIONS(4466),
    [anon_sym___thread] = ACTIONS(4466),
    [anon_sym_CG_EXTERN] = ACTIONS(4466),
    [anon_sym_CG_INLINE] = ACTIONS(4466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4466),
    [anon_sym_IBOutlet] = ACTIONS(4466),
    [anon_sym_IBInspectable] = ACTIONS(4466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4466),
    [anon_sym_NS_INLINE] = ACTIONS(4466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4466),
    [anon_sym_const] = ACTIONS(4466),
    [anon_sym_constexpr] = ACTIONS(4466),
    [anon_sym_volatile] = ACTIONS(4466),
    [anon_sym_restrict] = ACTIONS(4466),
    [anon_sym___restrict__] = ACTIONS(4466),
    [anon_sym__Atomic] = ACTIONS(4466),
    [anon_sym__Noreturn] = ACTIONS(4466),
    [anon_sym__Nonnull] = ACTIONS(4466),
    [anon_sym_nullable] = ACTIONS(4466),
    [anon_sym__Complex] = ACTIONS(4466),
    [anon_sym__Nullable] = ACTIONS(4466),
    [anon_sym__Nullable_result] = ACTIONS(4466),
    [anon_sym__Null_unspecified] = ACTIONS(4466),
    [anon_sym___autoreleasing] = ACTIONS(4466),
    [anon_sym___block] = ACTIONS(4466),
    [anon_sym___bridge] = ACTIONS(4466),
    [anon_sym___bridge_retained] = ACTIONS(4466),
    [anon_sym___bridge_transfer] = ACTIONS(4466),
    [anon_sym___complex] = ACTIONS(4466),
    [anon_sym___const] = ACTIONS(4466),
    [anon_sym___imag] = ACTIONS(4466),
    [anon_sym___kindof] = ACTIONS(4466),
    [anon_sym___nonnull] = ACTIONS(4466),
    [anon_sym___nullable] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4466),
    [anon_sym___real] = ACTIONS(4466),
    [anon_sym___strong] = ACTIONS(4466),
    [anon_sym___unsafe_unretained] = ACTIONS(4466),
    [anon_sym___unused] = ACTIONS(4466),
    [anon_sym___weak] = ACTIONS(4466),
    [anon_sym_alignas] = ACTIONS(4466),
    [anon_sym__Alignas] = ACTIONS(4466),
    [anon_sym_QMARK] = ACTIONS(5044),
    [anon_sym_STAR_EQ] = ACTIONS(4468),
    [anon_sym_SLASH_EQ] = ACTIONS(4468),
    [anon_sym_PERCENT_EQ] = ACTIONS(4468),
    [anon_sym_PLUS_EQ] = ACTIONS(4468),
    [anon_sym_DASH_EQ] = ACTIONS(4468),
    [anon_sym_LT_LT_EQ] = ACTIONS(4468),
    [anon_sym_GT_GT_EQ] = ACTIONS(4468),
    [anon_sym_AMP_EQ] = ACTIONS(4468),
    [anon_sym_CARET_EQ] = ACTIONS(4468),
    [anon_sym_PIPE_EQ] = ACTIONS(4468),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_API_AVAILABLE] = ACTIONS(4466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_API_DEPRECATED] = ACTIONS(4466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym___deprecated_msg] = ACTIONS(4466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4466),
  },
  [STATE(1918)] = {
    [sym_identifier] = ACTIONS(4024),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4026),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4024),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4024),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym___extension__] = ACTIONS(4024),
    [anon_sym_extern] = ACTIONS(4024),
    [anon_sym___attribute__] = ACTIONS(4024),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4024),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4024),
    [anon_sym_RBRACE] = ACTIONS(4026),
    [anon_sym_static] = ACTIONS(4024),
    [anon_sym_EQ] = ACTIONS(4775),
    [anon_sym_auto] = ACTIONS(4024),
    [anon_sym_register] = ACTIONS(4024),
    [anon_sym_inline] = ACTIONS(4024),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4024),
    [anon_sym___forceinline] = ACTIONS(4024),
    [anon_sym_thread_local] = ACTIONS(4024),
    [anon_sym___thread] = ACTIONS(4024),
    [anon_sym_CG_EXTERN] = ACTIONS(4024),
    [anon_sym_CG_INLINE] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4024),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4024),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4024),
    [anon_sym_IBOutlet] = ACTIONS(4024),
    [anon_sym_IBInspectable] = ACTIONS(4024),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4024),
    [anon_sym_NS_INLINE] = ACTIONS(4024),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4024),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4024),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4024),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4024),
    [anon_sym_volatile] = ACTIONS(4024),
    [anon_sym_restrict] = ACTIONS(4024),
    [anon_sym___restrict__] = ACTIONS(4024),
    [anon_sym__Atomic] = ACTIONS(4024),
    [anon_sym__Noreturn] = ACTIONS(4024),
    [anon_sym__Nonnull] = ACTIONS(4024),
    [anon_sym_nullable] = ACTIONS(4024),
    [anon_sym__Complex] = ACTIONS(4024),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4024),
    [anon_sym__Null_unspecified] = ACTIONS(4024),
    [anon_sym___autoreleasing] = ACTIONS(4024),
    [anon_sym___block] = ACTIONS(4024),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4024),
    [anon_sym___bridge_transfer] = ACTIONS(4024),
    [anon_sym___complex] = ACTIONS(4024),
    [anon_sym___const] = ACTIONS(4024),
    [anon_sym___imag] = ACTIONS(4024),
    [anon_sym___kindof] = ACTIONS(4024),
    [anon_sym___nonnull] = ACTIONS(4024),
    [anon_sym___nullable] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4024),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4024),
    [anon_sym___real] = ACTIONS(4024),
    [anon_sym___strong] = ACTIONS(4024),
    [anon_sym___unsafe_unretained] = ACTIONS(4024),
    [anon_sym___unused] = ACTIONS(4024),
    [anon_sym___weak] = ACTIONS(4024),
    [anon_sym_alignas] = ACTIONS(4024),
    [anon_sym__Alignas] = ACTIONS(4024),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4777),
    [anon_sym_SLASH_EQ] = ACTIONS(4777),
    [anon_sym_PERCENT_EQ] = ACTIONS(4777),
    [anon_sym_PLUS_EQ] = ACTIONS(4777),
    [anon_sym_DASH_EQ] = ACTIONS(4777),
    [anon_sym_LT_LT_EQ] = ACTIONS(4777),
    [anon_sym_GT_GT_EQ] = ACTIONS(4777),
    [anon_sym_AMP_EQ] = ACTIONS(4777),
    [anon_sym_CARET_EQ] = ACTIONS(4777),
    [anon_sym_PIPE_EQ] = ACTIONS(4777),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4024),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4024),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4024),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4024),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_API_AVAILABLE] = ACTIONS(4024),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_API_DEPRECATED] = ACTIONS(4024),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4024),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4024),
    [anon_sym___deprecated_msg] = ACTIONS(4024),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4024),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4024),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4024),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4024),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4024),
  },
  [STATE(1919)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2295),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4910),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1920)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2250),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(5062),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1921)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2268),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4916),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1922)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_DASH] = ACTIONS(2476),
    [anon_sym_PLUS] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATend] = ACTIONS(2476),
    [anon_sym_AToptional] = ACTIONS(2476),
    [anon_sym_ATrequired] = ACTIONS(2476),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATproperty] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(1923)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_DASH] = ACTIONS(2476),
    [anon_sym_PLUS] = ACTIONS(2476),
    [anon_sym_SEMI] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_LBRACE] = ACTIONS(2476),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_ATautoreleasepool] = ACTIONS(2476),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATend] = ACTIONS(2476),
    [anon_sym_AToptional] = ACTIONS(2476),
    [anon_sym_ATrequired] = ACTIONS(2476),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATproperty] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(1924)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2110),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4912),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1925)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2108),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4874),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1926)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2309),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4842),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1927)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2111),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(5064),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1928)] = {
    [sym_identifier] = ACTIONS(4220),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4222),
    [anon_sym_COMMA] = ACTIONS(4222),
    [anon_sym_RPAREN] = ACTIONS(4222),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4222),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4222),
    [anon_sym_LPAREN2] = ACTIONS(4222),
    [anon_sym_DASH] = ACTIONS(4220),
    [anon_sym_PLUS] = ACTIONS(4220),
    [anon_sym_STAR] = ACTIONS(4222),
    [anon_sym_SLASH] = ACTIONS(4220),
    [anon_sym_PERCENT] = ACTIONS(4222),
    [anon_sym_PIPE_PIPE] = ACTIONS(4222),
    [anon_sym_AMP_AMP] = ACTIONS(4222),
    [anon_sym_PIPE] = ACTIONS(4220),
    [anon_sym_CARET] = ACTIONS(4222),
    [anon_sym_AMP] = ACTIONS(4220),
    [anon_sym_EQ_EQ] = ACTIONS(4222),
    [anon_sym_BANG_EQ] = ACTIONS(4222),
    [anon_sym_GT] = ACTIONS(4220),
    [anon_sym_GT_EQ] = ACTIONS(4222),
    [anon_sym_LT_EQ] = ACTIONS(4222),
    [anon_sym_LT] = ACTIONS(4220),
    [anon_sym_LT_LT] = ACTIONS(4222),
    [anon_sym_GT_GT] = ACTIONS(4222),
    [anon_sym_SEMI] = ACTIONS(4222),
    [anon_sym___extension__] = ACTIONS(4220),
    [anon_sym_extern] = ACTIONS(4220),
    [anon_sym___attribute__] = ACTIONS(4220),
    [anon_sym___attribute] = ACTIONS(4220),
    [anon_sym_noreturn] = ACTIONS(4220),
    [anon_sym_LBRACK] = ACTIONS(4222),
    [anon_sym_RBRACK] = ACTIONS(4222),
    [anon_sym___declspec] = ACTIONS(4220),
    [anon_sym_LBRACE] = ACTIONS(4222),
    [anon_sym_RBRACE] = ACTIONS(4222),
    [anon_sym_static] = ACTIONS(4220),
    [anon_sym_EQ] = ACTIONS(4220),
    [anon_sym_ATautoreleasepool] = ACTIONS(4222),
    [anon_sym_auto] = ACTIONS(4220),
    [anon_sym_register] = ACTIONS(4220),
    [anon_sym_inline] = ACTIONS(4220),
    [anon_sym___inline] = ACTIONS(4220),
    [anon_sym___inline__] = ACTIONS(4220),
    [anon_sym___forceinline] = ACTIONS(4220),
    [anon_sym_thread_local] = ACTIONS(4220),
    [anon_sym___thread] = ACTIONS(4220),
    [anon_sym_CG_EXTERN] = ACTIONS(4220),
    [anon_sym_CG_INLINE] = ACTIONS(4220),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4220),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4220),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4220),
    [anon_sym_IBOutlet] = ACTIONS(4220),
    [anon_sym_IBInspectable] = ACTIONS(4220),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4220),
    [anon_sym_NS_INLINE] = ACTIONS(4220),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4220),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4220),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4220),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4220),
    [anon_sym_const] = ACTIONS(4220),
    [anon_sym_constexpr] = ACTIONS(4220),
    [anon_sym_volatile] = ACTIONS(4220),
    [anon_sym_restrict] = ACTIONS(4220),
    [anon_sym___restrict__] = ACTIONS(4220),
    [anon_sym__Atomic] = ACTIONS(4220),
    [anon_sym__Noreturn] = ACTIONS(4220),
    [anon_sym__Nonnull] = ACTIONS(4220),
    [anon_sym_nullable] = ACTIONS(4220),
    [anon_sym__Complex] = ACTIONS(4220),
    [anon_sym__Nullable] = ACTIONS(4220),
    [anon_sym__Nullable_result] = ACTIONS(4220),
    [anon_sym__Null_unspecified] = ACTIONS(4220),
    [anon_sym___autoreleasing] = ACTIONS(4220),
    [anon_sym___block] = ACTIONS(4220),
    [anon_sym___bridge] = ACTIONS(4220),
    [anon_sym___bridge_retained] = ACTIONS(4220),
    [anon_sym___bridge_transfer] = ACTIONS(4220),
    [anon_sym___complex] = ACTIONS(4220),
    [anon_sym___const] = ACTIONS(4220),
    [anon_sym___imag] = ACTIONS(4220),
    [anon_sym___kindof] = ACTIONS(4220),
    [anon_sym___nonnull] = ACTIONS(4220),
    [anon_sym___nullable] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4220),
    [anon_sym___real] = ACTIONS(4220),
    [anon_sym___strong] = ACTIONS(4220),
    [anon_sym___unsafe_unretained] = ACTIONS(4220),
    [anon_sym___unused] = ACTIONS(4220),
    [anon_sym___weak] = ACTIONS(4220),
    [anon_sym_alignas] = ACTIONS(4220),
    [anon_sym__Alignas] = ACTIONS(4220),
    [anon_sym_COLON] = ACTIONS(4222),
    [anon_sym_in] = ACTIONS(4220),
    [anon_sym_QMARK] = ACTIONS(4222),
    [anon_sym_DASH_DASH] = ACTIONS(4222),
    [anon_sym_PLUS_PLUS] = ACTIONS(4222),
    [anon_sym_asm] = ACTIONS(4220),
    [anon_sym___asm__] = ACTIONS(4220),
    [anon_sym___asm] = ACTIONS(4220),
    [anon_sym_DOT] = ACTIONS(4220),
    [anon_sym_DASH_GT] = ACTIONS(4222),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4220),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4220),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4220),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4220),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4220),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4220),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4220),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4220),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4220),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_API_AVAILABLE] = ACTIONS(4220),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_API_DEPRECATED] = ACTIONS(4220),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4220),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4220),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4220),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4220),
    [anon_sym___deprecated_msg] = ACTIONS(4220),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4220),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4220),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4220),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4220),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4220),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4220),
  },
  [STATE(1929)] = {
    [sym_identifier] = ACTIONS(4224),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4226),
    [anon_sym_COMMA] = ACTIONS(4226),
    [anon_sym_RPAREN] = ACTIONS(4226),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4226),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4226),
    [anon_sym_LPAREN2] = ACTIONS(4226),
    [anon_sym_DASH] = ACTIONS(4224),
    [anon_sym_PLUS] = ACTIONS(4224),
    [anon_sym_STAR] = ACTIONS(4226),
    [anon_sym_SLASH] = ACTIONS(4224),
    [anon_sym_PERCENT] = ACTIONS(4226),
    [anon_sym_PIPE_PIPE] = ACTIONS(4226),
    [anon_sym_AMP_AMP] = ACTIONS(4226),
    [anon_sym_PIPE] = ACTIONS(4224),
    [anon_sym_CARET] = ACTIONS(4226),
    [anon_sym_AMP] = ACTIONS(4224),
    [anon_sym_EQ_EQ] = ACTIONS(4226),
    [anon_sym_BANG_EQ] = ACTIONS(4226),
    [anon_sym_GT] = ACTIONS(4224),
    [anon_sym_GT_EQ] = ACTIONS(4226),
    [anon_sym_LT_EQ] = ACTIONS(4226),
    [anon_sym_LT] = ACTIONS(4224),
    [anon_sym_LT_LT] = ACTIONS(4226),
    [anon_sym_GT_GT] = ACTIONS(4226),
    [anon_sym_SEMI] = ACTIONS(4226),
    [anon_sym___extension__] = ACTIONS(4224),
    [anon_sym_extern] = ACTIONS(4224),
    [anon_sym___attribute__] = ACTIONS(4224),
    [anon_sym___attribute] = ACTIONS(4224),
    [anon_sym_noreturn] = ACTIONS(4224),
    [anon_sym_LBRACK] = ACTIONS(4226),
    [anon_sym_RBRACK] = ACTIONS(4226),
    [anon_sym___declspec] = ACTIONS(4224),
    [anon_sym_LBRACE] = ACTIONS(4226),
    [anon_sym_RBRACE] = ACTIONS(4226),
    [anon_sym_static] = ACTIONS(4224),
    [anon_sym_EQ] = ACTIONS(4224),
    [anon_sym_ATautoreleasepool] = ACTIONS(4226),
    [anon_sym_auto] = ACTIONS(4224),
    [anon_sym_register] = ACTIONS(4224),
    [anon_sym_inline] = ACTIONS(4224),
    [anon_sym___inline] = ACTIONS(4224),
    [anon_sym___inline__] = ACTIONS(4224),
    [anon_sym___forceinline] = ACTIONS(4224),
    [anon_sym_thread_local] = ACTIONS(4224),
    [anon_sym___thread] = ACTIONS(4224),
    [anon_sym_CG_EXTERN] = ACTIONS(4224),
    [anon_sym_CG_INLINE] = ACTIONS(4224),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4224),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4224),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4224),
    [anon_sym_IBOutlet] = ACTIONS(4224),
    [anon_sym_IBInspectable] = ACTIONS(4224),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4224),
    [anon_sym_NS_INLINE] = ACTIONS(4224),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4224),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4224),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4224),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4224),
    [anon_sym_const] = ACTIONS(4224),
    [anon_sym_constexpr] = ACTIONS(4224),
    [anon_sym_volatile] = ACTIONS(4224),
    [anon_sym_restrict] = ACTIONS(4224),
    [anon_sym___restrict__] = ACTIONS(4224),
    [anon_sym__Atomic] = ACTIONS(4224),
    [anon_sym__Noreturn] = ACTIONS(4224),
    [anon_sym__Nonnull] = ACTIONS(4224),
    [anon_sym_nullable] = ACTIONS(4224),
    [anon_sym__Complex] = ACTIONS(4224),
    [anon_sym__Nullable] = ACTIONS(4224),
    [anon_sym__Nullable_result] = ACTIONS(4224),
    [anon_sym__Null_unspecified] = ACTIONS(4224),
    [anon_sym___autoreleasing] = ACTIONS(4224),
    [anon_sym___block] = ACTIONS(4224),
    [anon_sym___bridge] = ACTIONS(4224),
    [anon_sym___bridge_retained] = ACTIONS(4224),
    [anon_sym___bridge_transfer] = ACTIONS(4224),
    [anon_sym___complex] = ACTIONS(4224),
    [anon_sym___const] = ACTIONS(4224),
    [anon_sym___imag] = ACTIONS(4224),
    [anon_sym___kindof] = ACTIONS(4224),
    [anon_sym___nonnull] = ACTIONS(4224),
    [anon_sym___nullable] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4224),
    [anon_sym___real] = ACTIONS(4224),
    [anon_sym___strong] = ACTIONS(4224),
    [anon_sym___unsafe_unretained] = ACTIONS(4224),
    [anon_sym___unused] = ACTIONS(4224),
    [anon_sym___weak] = ACTIONS(4224),
    [anon_sym_alignas] = ACTIONS(4224),
    [anon_sym__Alignas] = ACTIONS(4224),
    [anon_sym_COLON] = ACTIONS(4226),
    [anon_sym_in] = ACTIONS(4224),
    [anon_sym_QMARK] = ACTIONS(4226),
    [anon_sym_DASH_DASH] = ACTIONS(4226),
    [anon_sym_PLUS_PLUS] = ACTIONS(4226),
    [anon_sym_asm] = ACTIONS(4224),
    [anon_sym___asm__] = ACTIONS(4224),
    [anon_sym___asm] = ACTIONS(4224),
    [anon_sym_DOT] = ACTIONS(4224),
    [anon_sym_DASH_GT] = ACTIONS(4226),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4224),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4224),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4224),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4224),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4224),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4224),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4224),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4224),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4224),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_API_AVAILABLE] = ACTIONS(4224),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_API_DEPRECATED] = ACTIONS(4224),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4224),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4224),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4224),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4224),
    [anon_sym___deprecated_msg] = ACTIONS(4224),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4224),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4224),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4224),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4224),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4224),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4224),
  },
  [STATE(1930)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2271),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4876),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1931)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2308),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4878),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1932)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2260),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4815),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1933)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2186),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4834),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1934)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5793),
    [sym_parameter_declaration] = STATE(5807),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(5066),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1935)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2315),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4890),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1936)] = {
    [sym_identifier] = ACTIONS(4110),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4112),
    [anon_sym_COMMA] = ACTIONS(4112),
    [anon_sym_RPAREN] = ACTIONS(4112),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4112),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4112),
    [anon_sym_LPAREN2] = ACTIONS(4112),
    [anon_sym_DASH] = ACTIONS(4110),
    [anon_sym_PLUS] = ACTIONS(4110),
    [anon_sym_STAR] = ACTIONS(4112),
    [anon_sym_SLASH] = ACTIONS(4110),
    [anon_sym_PERCENT] = ACTIONS(4112),
    [anon_sym_PIPE_PIPE] = ACTIONS(4112),
    [anon_sym_AMP_AMP] = ACTIONS(4112),
    [anon_sym_PIPE] = ACTIONS(4110),
    [anon_sym_CARET] = ACTIONS(4112),
    [anon_sym_AMP] = ACTIONS(4110),
    [anon_sym_EQ_EQ] = ACTIONS(4112),
    [anon_sym_BANG_EQ] = ACTIONS(4112),
    [anon_sym_GT] = ACTIONS(4110),
    [anon_sym_GT_EQ] = ACTIONS(4112),
    [anon_sym_LT_EQ] = ACTIONS(4112),
    [anon_sym_LT] = ACTIONS(4110),
    [anon_sym_LT_LT] = ACTIONS(4112),
    [anon_sym_GT_GT] = ACTIONS(4112),
    [anon_sym_SEMI] = ACTIONS(4112),
    [anon_sym___extension__] = ACTIONS(4110),
    [anon_sym_extern] = ACTIONS(4110),
    [anon_sym___attribute__] = ACTIONS(4110),
    [anon_sym___attribute] = ACTIONS(4110),
    [anon_sym_noreturn] = ACTIONS(4110),
    [anon_sym_LBRACK] = ACTIONS(4112),
    [anon_sym_RBRACK] = ACTIONS(4112),
    [anon_sym___declspec] = ACTIONS(4110),
    [anon_sym_LBRACE] = ACTIONS(4112),
    [anon_sym_RBRACE] = ACTIONS(4112),
    [anon_sym_static] = ACTIONS(4110),
    [anon_sym_EQ] = ACTIONS(4110),
    [anon_sym_ATautoreleasepool] = ACTIONS(4112),
    [anon_sym_auto] = ACTIONS(4110),
    [anon_sym_register] = ACTIONS(4110),
    [anon_sym_inline] = ACTIONS(4110),
    [anon_sym___inline] = ACTIONS(4110),
    [anon_sym___inline__] = ACTIONS(4110),
    [anon_sym___forceinline] = ACTIONS(4110),
    [anon_sym_thread_local] = ACTIONS(4110),
    [anon_sym___thread] = ACTIONS(4110),
    [anon_sym_CG_EXTERN] = ACTIONS(4110),
    [anon_sym_CG_INLINE] = ACTIONS(4110),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4110),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4110),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4110),
    [anon_sym_IBOutlet] = ACTIONS(4110),
    [anon_sym_IBInspectable] = ACTIONS(4110),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4110),
    [anon_sym_NS_INLINE] = ACTIONS(4110),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4110),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4110),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4110),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4110),
    [anon_sym_const] = ACTIONS(4110),
    [anon_sym_constexpr] = ACTIONS(4110),
    [anon_sym_volatile] = ACTIONS(4110),
    [anon_sym_restrict] = ACTIONS(4110),
    [anon_sym___restrict__] = ACTIONS(4110),
    [anon_sym__Atomic] = ACTIONS(4110),
    [anon_sym__Noreturn] = ACTIONS(4110),
    [anon_sym__Nonnull] = ACTIONS(4110),
    [anon_sym_nullable] = ACTIONS(4110),
    [anon_sym__Complex] = ACTIONS(4110),
    [anon_sym__Nullable] = ACTIONS(4110),
    [anon_sym__Nullable_result] = ACTIONS(4110),
    [anon_sym__Null_unspecified] = ACTIONS(4110),
    [anon_sym___autoreleasing] = ACTIONS(4110),
    [anon_sym___block] = ACTIONS(4110),
    [anon_sym___bridge] = ACTIONS(4110),
    [anon_sym___bridge_retained] = ACTIONS(4110),
    [anon_sym___bridge_transfer] = ACTIONS(4110),
    [anon_sym___complex] = ACTIONS(4110),
    [anon_sym___const] = ACTIONS(4110),
    [anon_sym___imag] = ACTIONS(4110),
    [anon_sym___kindof] = ACTIONS(4110),
    [anon_sym___nonnull] = ACTIONS(4110),
    [anon_sym___nullable] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4110),
    [anon_sym___real] = ACTIONS(4110),
    [anon_sym___strong] = ACTIONS(4110),
    [anon_sym___unsafe_unretained] = ACTIONS(4110),
    [anon_sym___unused] = ACTIONS(4110),
    [anon_sym___weak] = ACTIONS(4110),
    [anon_sym_alignas] = ACTIONS(4110),
    [anon_sym__Alignas] = ACTIONS(4110),
    [anon_sym_COLON] = ACTIONS(4112),
    [anon_sym_in] = ACTIONS(4110),
    [anon_sym_QMARK] = ACTIONS(4112),
    [anon_sym_DASH_DASH] = ACTIONS(4112),
    [anon_sym_PLUS_PLUS] = ACTIONS(4112),
    [anon_sym_asm] = ACTIONS(4110),
    [anon_sym___asm__] = ACTIONS(4110),
    [anon_sym___asm] = ACTIONS(4110),
    [anon_sym_DOT] = ACTIONS(4110),
    [anon_sym_DASH_GT] = ACTIONS(4112),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4110),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4110),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4110),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4110),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4110),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4110),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4110),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4110),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4110),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_API_AVAILABLE] = ACTIONS(4110),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_API_DEPRECATED] = ACTIONS(4110),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4110),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4110),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4110),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4110),
    [anon_sym___deprecated_msg] = ACTIONS(4110),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4110),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4110),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4110),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4110),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4110),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4110),
  },
  [STATE(1937)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2277),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4880),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1938)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2163),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4914),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1939)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2334),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4918),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1940)] = {
    [sym_identifier] = ACTIONS(4926),
    [aux_sym_preproc_def_token1] = ACTIONS(4926),
    [aux_sym_preproc_if_token1] = ACTIONS(4926),
    [aux_sym_preproc_if_token2] = ACTIONS(4926),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4926),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4926),
    [aux_sym_preproc_else_token1] = ACTIONS(4926),
    [aux_sym_preproc_elif_token1] = ACTIONS(4926),
    [sym_preproc_directive] = ACTIONS(4926),
    [anon_sym_LPAREN2] = ACTIONS(4928),
    [anon_sym_DASH] = ACTIONS(4928),
    [anon_sym_PLUS] = ACTIONS(4928),
    [anon_sym_STAR] = ACTIONS(4928),
    [anon_sym_CARET] = ACTIONS(4928),
    [anon_sym___extension__] = ACTIONS(4926),
    [anon_sym_typedef] = ACTIONS(4926),
    [anon_sym_extern] = ACTIONS(4926),
    [anon_sym___attribute__] = ACTIONS(4926),
    [anon_sym___attribute] = ACTIONS(4926),
    [anon_sym_noreturn] = ACTIONS(4926),
    [anon_sym_LBRACK] = ACTIONS(4928),
    [anon_sym___declspec] = ACTIONS(4926),
    [anon_sym___based] = ACTIONS(4926),
    [anon_sym___cdecl] = ACTIONS(4926),
    [anon_sym___clrcall] = ACTIONS(4926),
    [anon_sym___stdcall] = ACTIONS(4926),
    [anon_sym___fastcall] = ACTIONS(4926),
    [anon_sym___thiscall] = ACTIONS(4926),
    [anon_sym___vectorcall] = ACTIONS(4926),
    [anon_sym_signed] = ACTIONS(4926),
    [anon_sym_unsigned] = ACTIONS(4926),
    [anon_sym_long] = ACTIONS(4926),
    [anon_sym_short] = ACTIONS(4926),
    [anon_sym_static] = ACTIONS(4926),
    [anon_sym_auto] = ACTIONS(4926),
    [anon_sym_register] = ACTIONS(4926),
    [anon_sym_inline] = ACTIONS(4926),
    [anon_sym___inline] = ACTIONS(4926),
    [anon_sym___inline__] = ACTIONS(4926),
    [anon_sym___forceinline] = ACTIONS(4926),
    [anon_sym_thread_local] = ACTIONS(4926),
    [anon_sym___thread] = ACTIONS(4926),
    [anon_sym_CG_EXTERN] = ACTIONS(4926),
    [anon_sym_CG_INLINE] = ACTIONS(4926),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4926),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4926),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4926),
    [anon_sym_IBOutlet] = ACTIONS(4926),
    [anon_sym_IBInspectable] = ACTIONS(4926),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4926),
    [anon_sym_NS_INLINE] = ACTIONS(4926),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4926),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4926),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4926),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4926),
    [anon_sym_const] = ACTIONS(4926),
    [anon_sym_constexpr] = ACTIONS(4926),
    [anon_sym_volatile] = ACTIONS(4926),
    [anon_sym_restrict] = ACTIONS(4926),
    [anon_sym___restrict__] = ACTIONS(4926),
    [anon_sym__Atomic] = ACTIONS(4926),
    [anon_sym__Noreturn] = ACTIONS(4926),
    [anon_sym__Nonnull] = ACTIONS(4926),
    [anon_sym_nullable] = ACTIONS(4926),
    [anon_sym__Complex] = ACTIONS(4926),
    [anon_sym__Nullable] = ACTIONS(4926),
    [anon_sym__Nullable_result] = ACTIONS(4926),
    [anon_sym__Null_unspecified] = ACTIONS(4926),
    [anon_sym___autoreleasing] = ACTIONS(4926),
    [anon_sym___block] = ACTIONS(4926),
    [anon_sym___bridge] = ACTIONS(4926),
    [anon_sym___bridge_retained] = ACTIONS(4926),
    [anon_sym___bridge_transfer] = ACTIONS(4926),
    [anon_sym___complex] = ACTIONS(4926),
    [anon_sym___const] = ACTIONS(4926),
    [anon_sym___imag] = ACTIONS(4926),
    [anon_sym___kindof] = ACTIONS(4926),
    [anon_sym___nonnull] = ACTIONS(4926),
    [anon_sym___nullable] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4926),
    [anon_sym___real] = ACTIONS(4926),
    [anon_sym___strong] = ACTIONS(4926),
    [anon_sym___unsafe_unretained] = ACTIONS(4926),
    [anon_sym___unused] = ACTIONS(4926),
    [anon_sym___weak] = ACTIONS(4926),
    [anon_sym_alignas] = ACTIONS(4926),
    [anon_sym__Alignas] = ACTIONS(4926),
    [sym_primitive_type] = ACTIONS(4926),
    [anon_sym_enum] = ACTIONS(4926),
    [anon_sym_struct] = ACTIONS(4926),
    [anon_sym_union] = ACTIONS(4926),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4926),
    [anon_sym___typeof] = ACTIONS(4926),
    [anon_sym_typeof] = ACTIONS(4926),
    [aux_sym_preproc_undef_token1] = ACTIONS(4926),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4926),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4926),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4926),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4926),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4926),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4926),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4926),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4926),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4926),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_API_AVAILABLE] = ACTIONS(4926),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_API_DEPRECATED] = ACTIONS(4926),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4926),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4926),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4926),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4926),
    [anon_sym___deprecated_msg] = ACTIONS(4926),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4926),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4926),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4926),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4926),
    [anon_sym_ATsynthesize] = ACTIONS(4928),
    [anon_sym_ATdynamic] = ACTIONS(4928),
    [anon_sym_BOOL] = ACTIONS(4926),
    [anon_sym_IMP] = ACTIONS(4926),
    [anon_sym_SEL] = ACTIONS(4926),
    [anon_sym_Class] = ACTIONS(4926),
    [anon_sym_id] = ACTIONS(4926),
  },
  [STATE(1941)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4566),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_variadic_parameter] = STATE(5807),
    [sym_parameter_declaration] = STATE(5807),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4080),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1942)] = {
    [sym_identifier] = ACTIONS(4266),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4268),
    [anon_sym_COMMA] = ACTIONS(4268),
    [anon_sym_RPAREN] = ACTIONS(4268),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4268),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4268),
    [anon_sym_LPAREN2] = ACTIONS(4268),
    [anon_sym_DASH] = ACTIONS(4266),
    [anon_sym_PLUS] = ACTIONS(4266),
    [anon_sym_STAR] = ACTIONS(4268),
    [anon_sym_SLASH] = ACTIONS(4266),
    [anon_sym_PERCENT] = ACTIONS(4268),
    [anon_sym_PIPE_PIPE] = ACTIONS(4268),
    [anon_sym_AMP_AMP] = ACTIONS(4268),
    [anon_sym_PIPE] = ACTIONS(4266),
    [anon_sym_CARET] = ACTIONS(4268),
    [anon_sym_AMP] = ACTIONS(4266),
    [anon_sym_EQ_EQ] = ACTIONS(4268),
    [anon_sym_BANG_EQ] = ACTIONS(4268),
    [anon_sym_GT] = ACTIONS(4266),
    [anon_sym_GT_EQ] = ACTIONS(4268),
    [anon_sym_LT_EQ] = ACTIONS(4268),
    [anon_sym_LT] = ACTIONS(4266),
    [anon_sym_LT_LT] = ACTIONS(4268),
    [anon_sym_GT_GT] = ACTIONS(4268),
    [anon_sym_SEMI] = ACTIONS(4268),
    [anon_sym___extension__] = ACTIONS(4266),
    [anon_sym_extern] = ACTIONS(4266),
    [anon_sym___attribute__] = ACTIONS(4266),
    [anon_sym___attribute] = ACTIONS(4266),
    [anon_sym_noreturn] = ACTIONS(4266),
    [anon_sym_LBRACK] = ACTIONS(4268),
    [anon_sym_RBRACK] = ACTIONS(4268),
    [anon_sym___declspec] = ACTIONS(4266),
    [anon_sym_LBRACE] = ACTIONS(4268),
    [anon_sym_RBRACE] = ACTIONS(4268),
    [anon_sym_static] = ACTIONS(4266),
    [anon_sym_EQ] = ACTIONS(4266),
    [anon_sym_ATautoreleasepool] = ACTIONS(4268),
    [anon_sym_auto] = ACTIONS(4266),
    [anon_sym_register] = ACTIONS(4266),
    [anon_sym_inline] = ACTIONS(4266),
    [anon_sym___inline] = ACTIONS(4266),
    [anon_sym___inline__] = ACTIONS(4266),
    [anon_sym___forceinline] = ACTIONS(4266),
    [anon_sym_thread_local] = ACTIONS(4266),
    [anon_sym___thread] = ACTIONS(4266),
    [anon_sym_CG_EXTERN] = ACTIONS(4266),
    [anon_sym_CG_INLINE] = ACTIONS(4266),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4266),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4266),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4266),
    [anon_sym_IBOutlet] = ACTIONS(4266),
    [anon_sym_IBInspectable] = ACTIONS(4266),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4266),
    [anon_sym_NS_INLINE] = ACTIONS(4266),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4266),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4266),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4266),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4266),
    [anon_sym_const] = ACTIONS(4266),
    [anon_sym_constexpr] = ACTIONS(4266),
    [anon_sym_volatile] = ACTIONS(4266),
    [anon_sym_restrict] = ACTIONS(4266),
    [anon_sym___restrict__] = ACTIONS(4266),
    [anon_sym__Atomic] = ACTIONS(4266),
    [anon_sym__Noreturn] = ACTIONS(4266),
    [anon_sym__Nonnull] = ACTIONS(4266),
    [anon_sym_nullable] = ACTIONS(4266),
    [anon_sym__Complex] = ACTIONS(4266),
    [anon_sym__Nullable] = ACTIONS(4266),
    [anon_sym__Nullable_result] = ACTIONS(4266),
    [anon_sym__Null_unspecified] = ACTIONS(4266),
    [anon_sym___autoreleasing] = ACTIONS(4266),
    [anon_sym___block] = ACTIONS(4266),
    [anon_sym___bridge] = ACTIONS(4266),
    [anon_sym___bridge_retained] = ACTIONS(4266),
    [anon_sym___bridge_transfer] = ACTIONS(4266),
    [anon_sym___complex] = ACTIONS(4266),
    [anon_sym___const] = ACTIONS(4266),
    [anon_sym___imag] = ACTIONS(4266),
    [anon_sym___kindof] = ACTIONS(4266),
    [anon_sym___nonnull] = ACTIONS(4266),
    [anon_sym___nullable] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4266),
    [anon_sym___real] = ACTIONS(4266),
    [anon_sym___strong] = ACTIONS(4266),
    [anon_sym___unsafe_unretained] = ACTIONS(4266),
    [anon_sym___unused] = ACTIONS(4266),
    [anon_sym___weak] = ACTIONS(4266),
    [anon_sym_alignas] = ACTIONS(4266),
    [anon_sym__Alignas] = ACTIONS(4266),
    [anon_sym_COLON] = ACTIONS(4268),
    [anon_sym_in] = ACTIONS(4266),
    [anon_sym_QMARK] = ACTIONS(4268),
    [anon_sym_DASH_DASH] = ACTIONS(4268),
    [anon_sym_PLUS_PLUS] = ACTIONS(4268),
    [anon_sym_asm] = ACTIONS(4266),
    [anon_sym___asm__] = ACTIONS(4266),
    [anon_sym___asm] = ACTIONS(4266),
    [anon_sym_DOT] = ACTIONS(4266),
    [anon_sym_DASH_GT] = ACTIONS(4268),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4266),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4266),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4266),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4266),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4266),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4266),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4266),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4266),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4266),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_API_AVAILABLE] = ACTIONS(4266),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_API_DEPRECATED] = ACTIONS(4266),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4266),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4266),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4266),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4266),
    [anon_sym___deprecated_msg] = ACTIONS(4266),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4266),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4266),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4266),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4266),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4266),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4266),
  },
  [STATE(1943)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_DASH] = ACTIONS(2418),
    [anon_sym_PLUS] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATend] = ACTIONS(2418),
    [anon_sym_AToptional] = ACTIONS(2418),
    [anon_sym_ATrequired] = ACTIONS(2418),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATproperty] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(1944)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_DASH] = ACTIONS(2422),
    [anon_sym_PLUS] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATend] = ACTIONS(2422),
    [anon_sym_AToptional] = ACTIONS(2422),
    [anon_sym_ATrequired] = ACTIONS(2422),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATproperty] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(1945)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_DASH] = ACTIONS(2426),
    [anon_sym_PLUS] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATend] = ACTIONS(2426),
    [anon_sym_AToptional] = ACTIONS(2426),
    [anon_sym_ATrequired] = ACTIONS(2426),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATproperty] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(1946)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_DASH] = ACTIONS(2418),
    [anon_sym_PLUS] = ACTIONS(2418),
    [anon_sym_SEMI] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_LBRACE] = ACTIONS(2418),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_ATautoreleasepool] = ACTIONS(2418),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATend] = ACTIONS(2418),
    [anon_sym_AToptional] = ACTIONS(2418),
    [anon_sym_ATrequired] = ACTIONS(2418),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATproperty] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(1947)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_DASH] = ACTIONS(2426),
    [anon_sym_PLUS] = ACTIONS(2426),
    [anon_sym_SEMI] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_LBRACE] = ACTIONS(2426),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_ATautoreleasepool] = ACTIONS(2426),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATend] = ACTIONS(2426),
    [anon_sym_AToptional] = ACTIONS(2426),
    [anon_sym_ATrequired] = ACTIONS(2426),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATproperty] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(1948)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_DASH] = ACTIONS(2422),
    [anon_sym_PLUS] = ACTIONS(2422),
    [anon_sym_SEMI] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_LBRACE] = ACTIONS(2422),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_ATautoreleasepool] = ACTIONS(2422),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATend] = ACTIONS(2422),
    [anon_sym_AToptional] = ACTIONS(2422),
    [anon_sym_ATrequired] = ACTIONS(2422),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATproperty] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(1949)] = {
    [sym_identifier] = ACTIONS(4274),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4276),
    [anon_sym_COMMA] = ACTIONS(4276),
    [anon_sym_RPAREN] = ACTIONS(4276),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4276),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4276),
    [anon_sym_LPAREN2] = ACTIONS(4276),
    [anon_sym_DASH] = ACTIONS(4274),
    [anon_sym_PLUS] = ACTIONS(4274),
    [anon_sym_STAR] = ACTIONS(4276),
    [anon_sym_SLASH] = ACTIONS(4274),
    [anon_sym_PERCENT] = ACTIONS(4276),
    [anon_sym_PIPE_PIPE] = ACTIONS(4276),
    [anon_sym_AMP_AMP] = ACTIONS(4276),
    [anon_sym_PIPE] = ACTIONS(4274),
    [anon_sym_CARET] = ACTIONS(4276),
    [anon_sym_AMP] = ACTIONS(4274),
    [anon_sym_EQ_EQ] = ACTIONS(4276),
    [anon_sym_BANG_EQ] = ACTIONS(4276),
    [anon_sym_GT] = ACTIONS(4274),
    [anon_sym_GT_EQ] = ACTIONS(4276),
    [anon_sym_LT_EQ] = ACTIONS(4276),
    [anon_sym_LT] = ACTIONS(4274),
    [anon_sym_LT_LT] = ACTIONS(4276),
    [anon_sym_GT_GT] = ACTIONS(4276),
    [anon_sym_SEMI] = ACTIONS(4276),
    [anon_sym___extension__] = ACTIONS(4274),
    [anon_sym_extern] = ACTIONS(4274),
    [anon_sym___attribute__] = ACTIONS(4274),
    [anon_sym___attribute] = ACTIONS(4274),
    [anon_sym_noreturn] = ACTIONS(4274),
    [anon_sym_LBRACK] = ACTIONS(4276),
    [anon_sym_RBRACK] = ACTIONS(4276),
    [anon_sym___declspec] = ACTIONS(4274),
    [anon_sym_LBRACE] = ACTIONS(4276),
    [anon_sym_RBRACE] = ACTIONS(4276),
    [anon_sym_static] = ACTIONS(4274),
    [anon_sym_EQ] = ACTIONS(4274),
    [anon_sym_ATautoreleasepool] = ACTIONS(4276),
    [anon_sym_auto] = ACTIONS(4274),
    [anon_sym_register] = ACTIONS(4274),
    [anon_sym_inline] = ACTIONS(4274),
    [anon_sym___inline] = ACTIONS(4274),
    [anon_sym___inline__] = ACTIONS(4274),
    [anon_sym___forceinline] = ACTIONS(4274),
    [anon_sym_thread_local] = ACTIONS(4274),
    [anon_sym___thread] = ACTIONS(4274),
    [anon_sym_CG_EXTERN] = ACTIONS(4274),
    [anon_sym_CG_INLINE] = ACTIONS(4274),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4274),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4274),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4274),
    [anon_sym_IBOutlet] = ACTIONS(4274),
    [anon_sym_IBInspectable] = ACTIONS(4274),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4274),
    [anon_sym_NS_INLINE] = ACTIONS(4274),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4274),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4274),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4274),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4274),
    [anon_sym_const] = ACTIONS(4274),
    [anon_sym_constexpr] = ACTIONS(4274),
    [anon_sym_volatile] = ACTIONS(4274),
    [anon_sym_restrict] = ACTIONS(4274),
    [anon_sym___restrict__] = ACTIONS(4274),
    [anon_sym__Atomic] = ACTIONS(4274),
    [anon_sym__Noreturn] = ACTIONS(4274),
    [anon_sym__Nonnull] = ACTIONS(4274),
    [anon_sym_nullable] = ACTIONS(4274),
    [anon_sym__Complex] = ACTIONS(4274),
    [anon_sym__Nullable] = ACTIONS(4274),
    [anon_sym__Nullable_result] = ACTIONS(4274),
    [anon_sym__Null_unspecified] = ACTIONS(4274),
    [anon_sym___autoreleasing] = ACTIONS(4274),
    [anon_sym___block] = ACTIONS(4274),
    [anon_sym___bridge] = ACTIONS(4274),
    [anon_sym___bridge_retained] = ACTIONS(4274),
    [anon_sym___bridge_transfer] = ACTIONS(4274),
    [anon_sym___complex] = ACTIONS(4274),
    [anon_sym___const] = ACTIONS(4274),
    [anon_sym___imag] = ACTIONS(4274),
    [anon_sym___kindof] = ACTIONS(4274),
    [anon_sym___nonnull] = ACTIONS(4274),
    [anon_sym___nullable] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4274),
    [anon_sym___real] = ACTIONS(4274),
    [anon_sym___strong] = ACTIONS(4274),
    [anon_sym___unsafe_unretained] = ACTIONS(4274),
    [anon_sym___unused] = ACTIONS(4274),
    [anon_sym___weak] = ACTIONS(4274),
    [anon_sym_alignas] = ACTIONS(4274),
    [anon_sym__Alignas] = ACTIONS(4274),
    [anon_sym_COLON] = ACTIONS(4276),
    [anon_sym_in] = ACTIONS(4274),
    [anon_sym_QMARK] = ACTIONS(4276),
    [anon_sym_DASH_DASH] = ACTIONS(4276),
    [anon_sym_PLUS_PLUS] = ACTIONS(4276),
    [anon_sym_asm] = ACTIONS(4274),
    [anon_sym___asm__] = ACTIONS(4274),
    [anon_sym___asm] = ACTIONS(4274),
    [anon_sym_DOT] = ACTIONS(4274),
    [anon_sym_DASH_GT] = ACTIONS(4276),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4274),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4274),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4274),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4274),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4274),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4274),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4274),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4274),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4274),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_API_AVAILABLE] = ACTIONS(4274),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_API_DEPRECATED] = ACTIONS(4274),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4274),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4274),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4274),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4274),
    [anon_sym___deprecated_msg] = ACTIONS(4274),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4274),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4274),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4274),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4274),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4274),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4274),
  },
  [STATE(1950)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2283),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4884),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1951)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2107),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4872),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1952)] = {
    [sym_identifier] = ACTIONS(5068),
    [aux_sym_preproc_def_token1] = ACTIONS(5068),
    [aux_sym_preproc_if_token1] = ACTIONS(5068),
    [aux_sym_preproc_if_token2] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5068),
    [aux_sym_preproc_else_token1] = ACTIONS(5068),
    [aux_sym_preproc_elif_token1] = ACTIONS(5068),
    [sym_preproc_directive] = ACTIONS(5068),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_DASH] = ACTIONS(5072),
    [anon_sym_PLUS] = ACTIONS(5072),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym___extension__] = ACTIONS(5068),
    [anon_sym_typedef] = ACTIONS(5068),
    [anon_sym_extern] = ACTIONS(5068),
    [anon_sym___attribute__] = ACTIONS(5068),
    [anon_sym___attribute] = ACTIONS(5068),
    [anon_sym_noreturn] = ACTIONS(5068),
    [anon_sym_LBRACK] = ACTIONS(5072),
    [anon_sym___declspec] = ACTIONS(5068),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym___cdecl] = ACTIONS(5068),
    [anon_sym___clrcall] = ACTIONS(5068),
    [anon_sym___stdcall] = ACTIONS(5068),
    [anon_sym___fastcall] = ACTIONS(5068),
    [anon_sym___thiscall] = ACTIONS(5068),
    [anon_sym___vectorcall] = ACTIONS(5068),
    [anon_sym_signed] = ACTIONS(5068),
    [anon_sym_unsigned] = ACTIONS(5068),
    [anon_sym_long] = ACTIONS(5068),
    [anon_sym_short] = ACTIONS(5068),
    [anon_sym_static] = ACTIONS(5068),
    [anon_sym_auto] = ACTIONS(5068),
    [anon_sym_register] = ACTIONS(5068),
    [anon_sym_inline] = ACTIONS(5068),
    [anon_sym___inline] = ACTIONS(5068),
    [anon_sym___inline__] = ACTIONS(5068),
    [anon_sym___forceinline] = ACTIONS(5068),
    [anon_sym_thread_local] = ACTIONS(5068),
    [anon_sym___thread] = ACTIONS(5068),
    [anon_sym_CG_EXTERN] = ACTIONS(5068),
    [anon_sym_CG_INLINE] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5068),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5068),
    [anon_sym_IBOutlet] = ACTIONS(5068),
    [anon_sym_IBInspectable] = ACTIONS(5068),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5068),
    [anon_sym_NS_INLINE] = ACTIONS(5068),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5068),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5068),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5068),
    [anon_sym_const] = ACTIONS(5068),
    [anon_sym_constexpr] = ACTIONS(5068),
    [anon_sym_volatile] = ACTIONS(5068),
    [anon_sym_restrict] = ACTIONS(5068),
    [anon_sym___restrict__] = ACTIONS(5068),
    [anon_sym__Atomic] = ACTIONS(5068),
    [anon_sym__Noreturn] = ACTIONS(5068),
    [anon_sym__Nonnull] = ACTIONS(5068),
    [anon_sym_nullable] = ACTIONS(5068),
    [anon_sym__Complex] = ACTIONS(5068),
    [anon_sym__Nullable] = ACTIONS(5068),
    [anon_sym__Nullable_result] = ACTIONS(5068),
    [anon_sym__Null_unspecified] = ACTIONS(5068),
    [anon_sym___autoreleasing] = ACTIONS(5068),
    [anon_sym___block] = ACTIONS(5068),
    [anon_sym___bridge] = ACTIONS(5068),
    [anon_sym___bridge_retained] = ACTIONS(5068),
    [anon_sym___bridge_transfer] = ACTIONS(5068),
    [anon_sym___complex] = ACTIONS(5068),
    [anon_sym___const] = ACTIONS(5068),
    [anon_sym___imag] = ACTIONS(5068),
    [anon_sym___kindof] = ACTIONS(5068),
    [anon_sym___nonnull] = ACTIONS(5068),
    [anon_sym___nullable] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5068),
    [anon_sym___real] = ACTIONS(5068),
    [anon_sym___strong] = ACTIONS(5068),
    [anon_sym___unsafe_unretained] = ACTIONS(5068),
    [anon_sym___unused] = ACTIONS(5068),
    [anon_sym___weak] = ACTIONS(5068),
    [anon_sym_alignas] = ACTIONS(5068),
    [anon_sym__Alignas] = ACTIONS(5068),
    [sym_primitive_type] = ACTIONS(5068),
    [anon_sym_enum] = ACTIONS(5068),
    [anon_sym_struct] = ACTIONS(5068),
    [anon_sym_union] = ACTIONS(5068),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5068),
    [anon_sym___typeof] = ACTIONS(5068),
    [anon_sym_typeof] = ACTIONS(5068),
    [aux_sym_preproc_undef_token1] = ACTIONS(5068),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5068),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5068),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5068),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_API_AVAILABLE] = ACTIONS(5068),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_API_DEPRECATED] = ACTIONS(5068),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym___deprecated_msg] = ACTIONS(5068),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5068),
    [anon_sym_ATsynthesize] = ACTIONS(5072),
    [anon_sym_ATdynamic] = ACTIONS(5072),
    [anon_sym_BOOL] = ACTIONS(5068),
    [anon_sym_IMP] = ACTIONS(5068),
    [anon_sym_SEL] = ACTIONS(5068),
    [anon_sym_Class] = ACTIONS(5068),
    [anon_sym_id] = ACTIONS(5068),
  },
  [STATE(1953)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2289),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4886),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1954)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2252),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4836),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1955)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2101),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4850),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1956)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2299),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4888),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1957)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2301),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4920),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1958)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_DASH] = ACTIONS(2440),
    [anon_sym_PLUS] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATend] = ACTIONS(2440),
    [anon_sym_AToptional] = ACTIONS(2440),
    [anon_sym_ATrequired] = ACTIONS(2440),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATproperty] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(1959)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_DASH] = ACTIONS(2444),
    [anon_sym_PLUS] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATend] = ACTIONS(2444),
    [anon_sym_AToptional] = ACTIONS(2444),
    [anon_sym_ATrequired] = ACTIONS(2444),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATproperty] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(1960)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_DASH] = ACTIONS(2448),
    [anon_sym_PLUS] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATend] = ACTIONS(2448),
    [anon_sym_AToptional] = ACTIONS(2448),
    [anon_sym_ATrequired] = ACTIONS(2448),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATproperty] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(1961)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_DASH] = ACTIONS(2448),
    [anon_sym_PLUS] = ACTIONS(2448),
    [anon_sym_SEMI] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_LBRACE] = ACTIONS(2448),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_ATautoreleasepool] = ACTIONS(2448),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATend] = ACTIONS(2448),
    [anon_sym_AToptional] = ACTIONS(2448),
    [anon_sym_ATrequired] = ACTIONS(2448),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATproperty] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(1962)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_DASH] = ACTIONS(2440),
    [anon_sym_PLUS] = ACTIONS(2440),
    [anon_sym_SEMI] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_LBRACE] = ACTIONS(2440),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_ATautoreleasepool] = ACTIONS(2440),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATend] = ACTIONS(2440),
    [anon_sym_AToptional] = ACTIONS(2440),
    [anon_sym_ATrequired] = ACTIONS(2440),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATproperty] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(1963)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_DASH] = ACTIONS(2444),
    [anon_sym_PLUS] = ACTIONS(2444),
    [anon_sym_SEMI] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_LBRACE] = ACTIONS(2444),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_ATautoreleasepool] = ACTIONS(2444),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATend] = ACTIONS(2444),
    [anon_sym_AToptional] = ACTIONS(2444),
    [anon_sym_ATrequired] = ACTIONS(2444),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATproperty] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(1964)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2328),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(5076),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1965)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2339),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4892),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1966)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2297),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4830),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1967)] = {
    [sym_identifier] = ACTIONS(4054),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4056),
    [anon_sym_COMMA] = ACTIONS(4056),
    [anon_sym_RPAREN] = ACTIONS(4056),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4056),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4056),
    [anon_sym_LPAREN2] = ACTIONS(4056),
    [anon_sym_DASH] = ACTIONS(4054),
    [anon_sym_PLUS] = ACTIONS(4054),
    [anon_sym_STAR] = ACTIONS(4056),
    [anon_sym_SLASH] = ACTIONS(4054),
    [anon_sym_PERCENT] = ACTIONS(4056),
    [anon_sym_PIPE_PIPE] = ACTIONS(4056),
    [anon_sym_AMP_AMP] = ACTIONS(4056),
    [anon_sym_PIPE] = ACTIONS(4054),
    [anon_sym_CARET] = ACTIONS(4056),
    [anon_sym_AMP] = ACTIONS(4054),
    [anon_sym_EQ_EQ] = ACTIONS(4056),
    [anon_sym_BANG_EQ] = ACTIONS(4056),
    [anon_sym_GT] = ACTIONS(4054),
    [anon_sym_GT_EQ] = ACTIONS(4056),
    [anon_sym_LT_EQ] = ACTIONS(4056),
    [anon_sym_LT] = ACTIONS(4054),
    [anon_sym_LT_LT] = ACTIONS(4056),
    [anon_sym_GT_GT] = ACTIONS(4056),
    [anon_sym_SEMI] = ACTIONS(4056),
    [anon_sym___extension__] = ACTIONS(4054),
    [anon_sym_extern] = ACTIONS(4054),
    [anon_sym___attribute__] = ACTIONS(4054),
    [anon_sym___attribute] = ACTIONS(4054),
    [anon_sym_noreturn] = ACTIONS(4054),
    [anon_sym_LBRACK] = ACTIONS(4056),
    [anon_sym_RBRACK] = ACTIONS(4056),
    [anon_sym___declspec] = ACTIONS(4054),
    [anon_sym_LBRACE] = ACTIONS(4056),
    [anon_sym_RBRACE] = ACTIONS(4056),
    [anon_sym_static] = ACTIONS(4054),
    [anon_sym_EQ] = ACTIONS(4054),
    [anon_sym_ATautoreleasepool] = ACTIONS(4056),
    [anon_sym_auto] = ACTIONS(4054),
    [anon_sym_register] = ACTIONS(4054),
    [anon_sym_inline] = ACTIONS(4054),
    [anon_sym___inline] = ACTIONS(4054),
    [anon_sym___inline__] = ACTIONS(4054),
    [anon_sym___forceinline] = ACTIONS(4054),
    [anon_sym_thread_local] = ACTIONS(4054),
    [anon_sym___thread] = ACTIONS(4054),
    [anon_sym_CG_EXTERN] = ACTIONS(4054),
    [anon_sym_CG_INLINE] = ACTIONS(4054),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4054),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4054),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4054),
    [anon_sym_IBOutlet] = ACTIONS(4054),
    [anon_sym_IBInspectable] = ACTIONS(4054),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4054),
    [anon_sym_NS_INLINE] = ACTIONS(4054),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4054),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4054),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4054),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4054),
    [anon_sym_const] = ACTIONS(4054),
    [anon_sym_constexpr] = ACTIONS(4054),
    [anon_sym_volatile] = ACTIONS(4054),
    [anon_sym_restrict] = ACTIONS(4054),
    [anon_sym___restrict__] = ACTIONS(4054),
    [anon_sym__Atomic] = ACTIONS(4054),
    [anon_sym__Noreturn] = ACTIONS(4054),
    [anon_sym__Nonnull] = ACTIONS(4054),
    [anon_sym_nullable] = ACTIONS(4054),
    [anon_sym__Complex] = ACTIONS(4054),
    [anon_sym__Nullable] = ACTIONS(4054),
    [anon_sym__Nullable_result] = ACTIONS(4054),
    [anon_sym__Null_unspecified] = ACTIONS(4054),
    [anon_sym___autoreleasing] = ACTIONS(4054),
    [anon_sym___block] = ACTIONS(4054),
    [anon_sym___bridge] = ACTIONS(4054),
    [anon_sym___bridge_retained] = ACTIONS(4054),
    [anon_sym___bridge_transfer] = ACTIONS(4054),
    [anon_sym___complex] = ACTIONS(4054),
    [anon_sym___const] = ACTIONS(4054),
    [anon_sym___imag] = ACTIONS(4054),
    [anon_sym___kindof] = ACTIONS(4054),
    [anon_sym___nonnull] = ACTIONS(4054),
    [anon_sym___nullable] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4054),
    [anon_sym___real] = ACTIONS(4054),
    [anon_sym___strong] = ACTIONS(4054),
    [anon_sym___unsafe_unretained] = ACTIONS(4054),
    [anon_sym___unused] = ACTIONS(4054),
    [anon_sym___weak] = ACTIONS(4054),
    [anon_sym_alignas] = ACTIONS(4054),
    [anon_sym__Alignas] = ACTIONS(4054),
    [anon_sym_COLON] = ACTIONS(4056),
    [anon_sym_in] = ACTIONS(4054),
    [anon_sym_QMARK] = ACTIONS(4056),
    [anon_sym_DASH_DASH] = ACTIONS(4056),
    [anon_sym_PLUS_PLUS] = ACTIONS(4056),
    [anon_sym_asm] = ACTIONS(4054),
    [anon_sym___asm__] = ACTIONS(4054),
    [anon_sym___asm] = ACTIONS(4054),
    [anon_sym_DOT] = ACTIONS(4054),
    [anon_sym_DASH_GT] = ACTIONS(4056),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4054),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4054),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4054),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4054),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4054),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4054),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4054),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4054),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4054),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_API_AVAILABLE] = ACTIONS(4054),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_API_DEPRECATED] = ACTIONS(4054),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4054),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4054),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4054),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4054),
    [anon_sym___deprecated_msg] = ACTIONS(4054),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4054),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4054),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4054),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4054),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4054),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4054),
  },
  [STATE(1968)] = {
    [sym_identifier] = ACTIONS(4058),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4060),
    [anon_sym_COMMA] = ACTIONS(4060),
    [anon_sym_RPAREN] = ACTIONS(4060),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4060),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4060),
    [anon_sym_LPAREN2] = ACTIONS(4060),
    [anon_sym_DASH] = ACTIONS(4058),
    [anon_sym_PLUS] = ACTIONS(4058),
    [anon_sym_STAR] = ACTIONS(4060),
    [anon_sym_SLASH] = ACTIONS(4058),
    [anon_sym_PERCENT] = ACTIONS(4060),
    [anon_sym_PIPE_PIPE] = ACTIONS(4060),
    [anon_sym_AMP_AMP] = ACTIONS(4060),
    [anon_sym_PIPE] = ACTIONS(4058),
    [anon_sym_CARET] = ACTIONS(4060),
    [anon_sym_AMP] = ACTIONS(4058),
    [anon_sym_EQ_EQ] = ACTIONS(4060),
    [anon_sym_BANG_EQ] = ACTIONS(4060),
    [anon_sym_GT] = ACTIONS(4058),
    [anon_sym_GT_EQ] = ACTIONS(4060),
    [anon_sym_LT_EQ] = ACTIONS(4060),
    [anon_sym_LT] = ACTIONS(4058),
    [anon_sym_LT_LT] = ACTIONS(4060),
    [anon_sym_GT_GT] = ACTIONS(4060),
    [anon_sym_SEMI] = ACTIONS(4060),
    [anon_sym___extension__] = ACTIONS(4058),
    [anon_sym_extern] = ACTIONS(4058),
    [anon_sym___attribute__] = ACTIONS(4058),
    [anon_sym___attribute] = ACTIONS(4058),
    [anon_sym_noreturn] = ACTIONS(4058),
    [anon_sym_LBRACK] = ACTIONS(4060),
    [anon_sym_RBRACK] = ACTIONS(4060),
    [anon_sym___declspec] = ACTIONS(4058),
    [anon_sym_LBRACE] = ACTIONS(4060),
    [anon_sym_RBRACE] = ACTIONS(4060),
    [anon_sym_static] = ACTIONS(4058),
    [anon_sym_EQ] = ACTIONS(4058),
    [anon_sym_ATautoreleasepool] = ACTIONS(4060),
    [anon_sym_auto] = ACTIONS(4058),
    [anon_sym_register] = ACTIONS(4058),
    [anon_sym_inline] = ACTIONS(4058),
    [anon_sym___inline] = ACTIONS(4058),
    [anon_sym___inline__] = ACTIONS(4058),
    [anon_sym___forceinline] = ACTIONS(4058),
    [anon_sym_thread_local] = ACTIONS(4058),
    [anon_sym___thread] = ACTIONS(4058),
    [anon_sym_CG_EXTERN] = ACTIONS(4058),
    [anon_sym_CG_INLINE] = ACTIONS(4058),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4058),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4058),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4058),
    [anon_sym_IBOutlet] = ACTIONS(4058),
    [anon_sym_IBInspectable] = ACTIONS(4058),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4058),
    [anon_sym_NS_INLINE] = ACTIONS(4058),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4058),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4058),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4058),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4058),
    [anon_sym_const] = ACTIONS(4058),
    [anon_sym_constexpr] = ACTIONS(4058),
    [anon_sym_volatile] = ACTIONS(4058),
    [anon_sym_restrict] = ACTIONS(4058),
    [anon_sym___restrict__] = ACTIONS(4058),
    [anon_sym__Atomic] = ACTIONS(4058),
    [anon_sym__Noreturn] = ACTIONS(4058),
    [anon_sym__Nonnull] = ACTIONS(4058),
    [anon_sym_nullable] = ACTIONS(4058),
    [anon_sym__Complex] = ACTIONS(4058),
    [anon_sym__Nullable] = ACTIONS(4058),
    [anon_sym__Nullable_result] = ACTIONS(4058),
    [anon_sym__Null_unspecified] = ACTIONS(4058),
    [anon_sym___autoreleasing] = ACTIONS(4058),
    [anon_sym___block] = ACTIONS(4058),
    [anon_sym___bridge] = ACTIONS(4058),
    [anon_sym___bridge_retained] = ACTIONS(4058),
    [anon_sym___bridge_transfer] = ACTIONS(4058),
    [anon_sym___complex] = ACTIONS(4058),
    [anon_sym___const] = ACTIONS(4058),
    [anon_sym___imag] = ACTIONS(4058),
    [anon_sym___kindof] = ACTIONS(4058),
    [anon_sym___nonnull] = ACTIONS(4058),
    [anon_sym___nullable] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4058),
    [anon_sym___real] = ACTIONS(4058),
    [anon_sym___strong] = ACTIONS(4058),
    [anon_sym___unsafe_unretained] = ACTIONS(4058),
    [anon_sym___unused] = ACTIONS(4058),
    [anon_sym___weak] = ACTIONS(4058),
    [anon_sym_alignas] = ACTIONS(4058),
    [anon_sym__Alignas] = ACTIONS(4058),
    [anon_sym_COLON] = ACTIONS(4060),
    [anon_sym_in] = ACTIONS(4058),
    [anon_sym_QMARK] = ACTIONS(4060),
    [anon_sym_DASH_DASH] = ACTIONS(4060),
    [anon_sym_PLUS_PLUS] = ACTIONS(4060),
    [anon_sym_asm] = ACTIONS(4058),
    [anon_sym___asm__] = ACTIONS(4058),
    [anon_sym___asm] = ACTIONS(4058),
    [anon_sym_DOT] = ACTIONS(4058),
    [anon_sym_DASH_GT] = ACTIONS(4060),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4058),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4058),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4058),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4058),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4058),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4058),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4058),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4058),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4058),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_API_AVAILABLE] = ACTIONS(4058),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_API_DEPRECATED] = ACTIONS(4058),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4058),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4058),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4058),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4058),
    [anon_sym___deprecated_msg] = ACTIONS(4058),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4058),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4058),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4058),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4058),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4058),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4058),
  },
  [STATE(1969)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2109),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4866),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1970)] = {
    [sym_parameterized_arguments] = STATE(2052),
    [sym_identifier] = ACTIONS(5078),
    [aux_sym_preproc_def_token1] = ACTIONS(5078),
    [aux_sym_preproc_if_token1] = ACTIONS(5078),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5078),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5078),
    [sym_preproc_directive] = ACTIONS(5078),
    [anon_sym_LPAREN2] = ACTIONS(3259),
    [anon_sym_DASH] = ACTIONS(3259),
    [anon_sym_PLUS] = ACTIONS(3259),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(5078),
    [anon_sym_typedef] = ACTIONS(5078),
    [anon_sym_extern] = ACTIONS(5078),
    [anon_sym___attribute__] = ACTIONS(5078),
    [anon_sym___attribute] = ACTIONS(5078),
    [anon_sym_noreturn] = ACTIONS(5078),
    [anon_sym_LBRACK] = ACTIONS(3259),
    [anon_sym___declspec] = ACTIONS(5078),
    [anon_sym___cdecl] = ACTIONS(5078),
    [anon_sym___clrcall] = ACTIONS(5078),
    [anon_sym___stdcall] = ACTIONS(5078),
    [anon_sym___fastcall] = ACTIONS(5078),
    [anon_sym___thiscall] = ACTIONS(5078),
    [anon_sym___vectorcall] = ACTIONS(5078),
    [anon_sym_LBRACE] = ACTIONS(3259),
    [anon_sym_signed] = ACTIONS(5078),
    [anon_sym_unsigned] = ACTIONS(5078),
    [anon_sym_long] = ACTIONS(5078),
    [anon_sym_short] = ACTIONS(5078),
    [anon_sym_static] = ACTIONS(5078),
    [anon_sym_auto] = ACTIONS(5078),
    [anon_sym_register] = ACTIONS(5078),
    [anon_sym_inline] = ACTIONS(5078),
    [anon_sym___inline] = ACTIONS(5078),
    [anon_sym___inline__] = ACTIONS(5078),
    [anon_sym___forceinline] = ACTIONS(5078),
    [anon_sym_thread_local] = ACTIONS(5078),
    [anon_sym___thread] = ACTIONS(5078),
    [anon_sym_CG_EXTERN] = ACTIONS(5078),
    [anon_sym_CG_INLINE] = ACTIONS(5078),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5078),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5078),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5078),
    [anon_sym_IBOutlet] = ACTIONS(5078),
    [anon_sym_IBInspectable] = ACTIONS(5078),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5078),
    [anon_sym_NS_INLINE] = ACTIONS(5078),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5078),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5078),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5078),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5078),
    [anon_sym_const] = ACTIONS(5078),
    [anon_sym_constexpr] = ACTIONS(5078),
    [anon_sym_volatile] = ACTIONS(5078),
    [anon_sym_restrict] = ACTIONS(5078),
    [anon_sym___restrict__] = ACTIONS(5078),
    [anon_sym__Atomic] = ACTIONS(5078),
    [anon_sym__Noreturn] = ACTIONS(5078),
    [anon_sym__Nonnull] = ACTIONS(5078),
    [anon_sym_nullable] = ACTIONS(5078),
    [anon_sym__Complex] = ACTIONS(5078),
    [anon_sym__Nullable] = ACTIONS(5078),
    [anon_sym__Nullable_result] = ACTIONS(5078),
    [anon_sym__Null_unspecified] = ACTIONS(5078),
    [anon_sym___autoreleasing] = ACTIONS(5078),
    [anon_sym___block] = ACTIONS(5078),
    [anon_sym___bridge] = ACTIONS(5078),
    [anon_sym___bridge_retained] = ACTIONS(5078),
    [anon_sym___bridge_transfer] = ACTIONS(5078),
    [anon_sym___complex] = ACTIONS(5078),
    [anon_sym___const] = ACTIONS(5078),
    [anon_sym___imag] = ACTIONS(5078),
    [anon_sym___kindof] = ACTIONS(5078),
    [anon_sym___nonnull] = ACTIONS(5078),
    [anon_sym___nullable] = ACTIONS(5078),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5078),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5078),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5078),
    [anon_sym___real] = ACTIONS(5078),
    [anon_sym___strong] = ACTIONS(5078),
    [anon_sym___unsafe_unretained] = ACTIONS(5078),
    [anon_sym___unused] = ACTIONS(5078),
    [anon_sym___weak] = ACTIONS(5078),
    [anon_sym_alignas] = ACTIONS(5078),
    [anon_sym__Alignas] = ACTIONS(5078),
    [sym_primitive_type] = ACTIONS(5078),
    [anon_sym_enum] = ACTIONS(5078),
    [anon_sym_COLON] = ACTIONS(3259),
    [anon_sym_struct] = ACTIONS(5078),
    [anon_sym_union] = ACTIONS(5078),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5078),
    [anon_sym___typeof] = ACTIONS(5078),
    [anon_sym_typeof] = ACTIONS(5078),
    [aux_sym_preproc_undef_token1] = ACTIONS(5078),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5078),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5078),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5078),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5078),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5078),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5078),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5078),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5078),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5078),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5078),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5078),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5078),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5078),
    [anon_sym_API_AVAILABLE] = ACTIONS(5078),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5078),
    [anon_sym_API_DEPRECATED] = ACTIONS(5078),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5078),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5078),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5078),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5078),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5078),
    [anon_sym___deprecated_msg] = ACTIONS(5078),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5078),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5078),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5078),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5078),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5078),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5078),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5078),
    [anon_sym_ATend] = ACTIONS(3259),
    [anon_sym_ATsynthesize] = ACTIONS(3259),
    [anon_sym_ATdynamic] = ACTIONS(3259),
    [anon_sym_ATproperty] = ACTIONS(3259),
    [anon_sym_BOOL] = ACTIONS(5078),
    [anon_sym_IMP] = ACTIONS(5078),
    [anon_sym_SEL] = ACTIONS(5078),
    [anon_sym_Class] = ACTIONS(5078),
    [anon_sym_id] = ACTIONS(5078),
  },
  [STATE(1971)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2104),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4868),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1972)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2106),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4870),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1973)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_PLUS] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATend] = ACTIONS(2548),
    [anon_sym_AToptional] = ACTIONS(2548),
    [anon_sym_ATrequired] = ACTIONS(2548),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATproperty] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(1974)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2342),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4894),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1975)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2312),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4864),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1976)] = {
    [sym_identifier] = ACTIONS(4050),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4052),
    [anon_sym_COMMA] = ACTIONS(4052),
    [anon_sym_RPAREN] = ACTIONS(4052),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4052),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4052),
    [anon_sym_LPAREN2] = ACTIONS(4052),
    [anon_sym_DASH] = ACTIONS(4050),
    [anon_sym_PLUS] = ACTIONS(4050),
    [anon_sym_STAR] = ACTIONS(4052),
    [anon_sym_SLASH] = ACTIONS(4050),
    [anon_sym_PERCENT] = ACTIONS(4052),
    [anon_sym_PIPE_PIPE] = ACTIONS(4052),
    [anon_sym_AMP_AMP] = ACTIONS(4052),
    [anon_sym_PIPE] = ACTIONS(4050),
    [anon_sym_CARET] = ACTIONS(4052),
    [anon_sym_AMP] = ACTIONS(4050),
    [anon_sym_EQ_EQ] = ACTIONS(4052),
    [anon_sym_BANG_EQ] = ACTIONS(4052),
    [anon_sym_GT] = ACTIONS(4050),
    [anon_sym_GT_EQ] = ACTIONS(4052),
    [anon_sym_LT_EQ] = ACTIONS(4052),
    [anon_sym_LT] = ACTIONS(4050),
    [anon_sym_LT_LT] = ACTIONS(4052),
    [anon_sym_GT_GT] = ACTIONS(4052),
    [anon_sym_SEMI] = ACTIONS(4052),
    [anon_sym___extension__] = ACTIONS(4050),
    [anon_sym_extern] = ACTIONS(4050),
    [anon_sym___attribute__] = ACTIONS(4050),
    [anon_sym___attribute] = ACTIONS(4050),
    [anon_sym_noreturn] = ACTIONS(4050),
    [anon_sym_LBRACK] = ACTIONS(4052),
    [anon_sym_RBRACK] = ACTIONS(4052),
    [anon_sym___declspec] = ACTIONS(4050),
    [anon_sym_LBRACE] = ACTIONS(4052),
    [anon_sym_RBRACE] = ACTIONS(4052),
    [anon_sym_static] = ACTIONS(4050),
    [anon_sym_EQ] = ACTIONS(4050),
    [anon_sym_ATautoreleasepool] = ACTIONS(4052),
    [anon_sym_auto] = ACTIONS(4050),
    [anon_sym_register] = ACTIONS(4050),
    [anon_sym_inline] = ACTIONS(4050),
    [anon_sym___inline] = ACTIONS(4050),
    [anon_sym___inline__] = ACTIONS(4050),
    [anon_sym___forceinline] = ACTIONS(4050),
    [anon_sym_thread_local] = ACTIONS(4050),
    [anon_sym___thread] = ACTIONS(4050),
    [anon_sym_CG_EXTERN] = ACTIONS(4050),
    [anon_sym_CG_INLINE] = ACTIONS(4050),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4050),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4050),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4050),
    [anon_sym_IBOutlet] = ACTIONS(4050),
    [anon_sym_IBInspectable] = ACTIONS(4050),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4050),
    [anon_sym_NS_INLINE] = ACTIONS(4050),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4050),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4050),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4050),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4050),
    [anon_sym_const] = ACTIONS(4050),
    [anon_sym_constexpr] = ACTIONS(4050),
    [anon_sym_volatile] = ACTIONS(4050),
    [anon_sym_restrict] = ACTIONS(4050),
    [anon_sym___restrict__] = ACTIONS(4050),
    [anon_sym__Atomic] = ACTIONS(4050),
    [anon_sym__Noreturn] = ACTIONS(4050),
    [anon_sym__Nonnull] = ACTIONS(4050),
    [anon_sym_nullable] = ACTIONS(4050),
    [anon_sym__Complex] = ACTIONS(4050),
    [anon_sym__Nullable] = ACTIONS(4050),
    [anon_sym__Nullable_result] = ACTIONS(4050),
    [anon_sym__Null_unspecified] = ACTIONS(4050),
    [anon_sym___autoreleasing] = ACTIONS(4050),
    [anon_sym___block] = ACTIONS(4050),
    [anon_sym___bridge] = ACTIONS(4050),
    [anon_sym___bridge_retained] = ACTIONS(4050),
    [anon_sym___bridge_transfer] = ACTIONS(4050),
    [anon_sym___complex] = ACTIONS(4050),
    [anon_sym___const] = ACTIONS(4050),
    [anon_sym___imag] = ACTIONS(4050),
    [anon_sym___kindof] = ACTIONS(4050),
    [anon_sym___nonnull] = ACTIONS(4050),
    [anon_sym___nullable] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4050),
    [anon_sym___real] = ACTIONS(4050),
    [anon_sym___strong] = ACTIONS(4050),
    [anon_sym___unsafe_unretained] = ACTIONS(4050),
    [anon_sym___unused] = ACTIONS(4050),
    [anon_sym___weak] = ACTIONS(4050),
    [anon_sym_alignas] = ACTIONS(4050),
    [anon_sym__Alignas] = ACTIONS(4050),
    [anon_sym_COLON] = ACTIONS(4052),
    [anon_sym_in] = ACTIONS(4050),
    [anon_sym_QMARK] = ACTIONS(4052),
    [anon_sym_DASH_DASH] = ACTIONS(4052),
    [anon_sym_PLUS_PLUS] = ACTIONS(4052),
    [anon_sym_asm] = ACTIONS(4050),
    [anon_sym___asm__] = ACTIONS(4050),
    [anon_sym___asm] = ACTIONS(4050),
    [anon_sym_DOT] = ACTIONS(4050),
    [anon_sym_DASH_GT] = ACTIONS(4052),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4050),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4050),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4050),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4050),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4050),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4050),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4050),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4050),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4050),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_API_AVAILABLE] = ACTIONS(4050),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_API_DEPRECATED] = ACTIONS(4050),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4050),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4050),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4050),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4050),
    [anon_sym___deprecated_msg] = ACTIONS(4050),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4050),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4050),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4050),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4050),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4050),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4050),
  },
  [STATE(1977)] = {
    [sym_identifier] = ACTIONS(4062),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4064),
    [anon_sym_COMMA] = ACTIONS(4064),
    [anon_sym_RPAREN] = ACTIONS(4064),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4064),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4064),
    [anon_sym_LPAREN2] = ACTIONS(4064),
    [anon_sym_DASH] = ACTIONS(4062),
    [anon_sym_PLUS] = ACTIONS(4062),
    [anon_sym_STAR] = ACTIONS(4064),
    [anon_sym_SLASH] = ACTIONS(4062),
    [anon_sym_PERCENT] = ACTIONS(4064),
    [anon_sym_PIPE_PIPE] = ACTIONS(4064),
    [anon_sym_AMP_AMP] = ACTIONS(4064),
    [anon_sym_PIPE] = ACTIONS(4062),
    [anon_sym_CARET] = ACTIONS(4064),
    [anon_sym_AMP] = ACTIONS(4062),
    [anon_sym_EQ_EQ] = ACTIONS(4064),
    [anon_sym_BANG_EQ] = ACTIONS(4064),
    [anon_sym_GT] = ACTIONS(4062),
    [anon_sym_GT_EQ] = ACTIONS(4064),
    [anon_sym_LT_EQ] = ACTIONS(4064),
    [anon_sym_LT] = ACTIONS(4062),
    [anon_sym_LT_LT] = ACTIONS(4064),
    [anon_sym_GT_GT] = ACTIONS(4064),
    [anon_sym_SEMI] = ACTIONS(4064),
    [anon_sym___extension__] = ACTIONS(4062),
    [anon_sym_extern] = ACTIONS(4062),
    [anon_sym___attribute__] = ACTIONS(4062),
    [anon_sym___attribute] = ACTIONS(4062),
    [anon_sym_noreturn] = ACTIONS(4062),
    [anon_sym_LBRACK] = ACTIONS(4064),
    [anon_sym_RBRACK] = ACTIONS(4064),
    [anon_sym___declspec] = ACTIONS(4062),
    [anon_sym_LBRACE] = ACTIONS(4064),
    [anon_sym_RBRACE] = ACTIONS(4064),
    [anon_sym_static] = ACTIONS(4062),
    [anon_sym_EQ] = ACTIONS(4062),
    [anon_sym_ATautoreleasepool] = ACTIONS(4064),
    [anon_sym_auto] = ACTIONS(4062),
    [anon_sym_register] = ACTIONS(4062),
    [anon_sym_inline] = ACTIONS(4062),
    [anon_sym___inline] = ACTIONS(4062),
    [anon_sym___inline__] = ACTIONS(4062),
    [anon_sym___forceinline] = ACTIONS(4062),
    [anon_sym_thread_local] = ACTIONS(4062),
    [anon_sym___thread] = ACTIONS(4062),
    [anon_sym_CG_EXTERN] = ACTIONS(4062),
    [anon_sym_CG_INLINE] = ACTIONS(4062),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4062),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4062),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4062),
    [anon_sym_IBOutlet] = ACTIONS(4062),
    [anon_sym_IBInspectable] = ACTIONS(4062),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4062),
    [anon_sym_NS_INLINE] = ACTIONS(4062),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4062),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4062),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4062),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4062),
    [anon_sym_const] = ACTIONS(4062),
    [anon_sym_constexpr] = ACTIONS(4062),
    [anon_sym_volatile] = ACTIONS(4062),
    [anon_sym_restrict] = ACTIONS(4062),
    [anon_sym___restrict__] = ACTIONS(4062),
    [anon_sym__Atomic] = ACTIONS(4062),
    [anon_sym__Noreturn] = ACTIONS(4062),
    [anon_sym__Nonnull] = ACTIONS(4062),
    [anon_sym_nullable] = ACTIONS(4062),
    [anon_sym__Complex] = ACTIONS(4062),
    [anon_sym__Nullable] = ACTIONS(4062),
    [anon_sym__Nullable_result] = ACTIONS(4062),
    [anon_sym__Null_unspecified] = ACTIONS(4062),
    [anon_sym___autoreleasing] = ACTIONS(4062),
    [anon_sym___block] = ACTIONS(4062),
    [anon_sym___bridge] = ACTIONS(4062),
    [anon_sym___bridge_retained] = ACTIONS(4062),
    [anon_sym___bridge_transfer] = ACTIONS(4062),
    [anon_sym___complex] = ACTIONS(4062),
    [anon_sym___const] = ACTIONS(4062),
    [anon_sym___imag] = ACTIONS(4062),
    [anon_sym___kindof] = ACTIONS(4062),
    [anon_sym___nonnull] = ACTIONS(4062),
    [anon_sym___nullable] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4062),
    [anon_sym___real] = ACTIONS(4062),
    [anon_sym___strong] = ACTIONS(4062),
    [anon_sym___unsafe_unretained] = ACTIONS(4062),
    [anon_sym___unused] = ACTIONS(4062),
    [anon_sym___weak] = ACTIONS(4062),
    [anon_sym_alignas] = ACTIONS(4062),
    [anon_sym__Alignas] = ACTIONS(4062),
    [anon_sym_COLON] = ACTIONS(4064),
    [anon_sym_in] = ACTIONS(4062),
    [anon_sym_QMARK] = ACTIONS(4064),
    [anon_sym_DASH_DASH] = ACTIONS(4064),
    [anon_sym_PLUS_PLUS] = ACTIONS(4064),
    [anon_sym_asm] = ACTIONS(4062),
    [anon_sym___asm__] = ACTIONS(4062),
    [anon_sym___asm] = ACTIONS(4062),
    [anon_sym_DOT] = ACTIONS(4062),
    [anon_sym_DASH_GT] = ACTIONS(4064),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4062),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4062),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4062),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4062),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4062),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4062),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4062),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4062),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4062),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_API_AVAILABLE] = ACTIONS(4062),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_API_DEPRECATED] = ACTIONS(4062),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4062),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4062),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4062),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4062),
    [anon_sym___deprecated_msg] = ACTIONS(4062),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4062),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4062),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4062),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4062),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4062),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4062),
  },
  [STATE(1978)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_PLUS] = ACTIONS(2548),
    [anon_sym_SEMI] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_LBRACE] = ACTIONS(2548),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_ATautoreleasepool] = ACTIONS(2548),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATend] = ACTIONS(2548),
    [anon_sym_AToptional] = ACTIONS(2548),
    [anon_sym_ATrequired] = ACTIONS(2548),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATproperty] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(1979)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2262),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4811),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1980)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2345),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4844),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1981)] = {
    [sym_identifier] = ACTIONS(4168),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4170),
    [anon_sym_COMMA] = ACTIONS(4170),
    [anon_sym_RPAREN] = ACTIONS(4170),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4170),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4170),
    [anon_sym_LPAREN2] = ACTIONS(4170),
    [anon_sym_DASH] = ACTIONS(4168),
    [anon_sym_PLUS] = ACTIONS(4168),
    [anon_sym_STAR] = ACTIONS(4170),
    [anon_sym_SLASH] = ACTIONS(4168),
    [anon_sym_PERCENT] = ACTIONS(4170),
    [anon_sym_PIPE_PIPE] = ACTIONS(4170),
    [anon_sym_AMP_AMP] = ACTIONS(4170),
    [anon_sym_PIPE] = ACTIONS(4168),
    [anon_sym_CARET] = ACTIONS(4170),
    [anon_sym_AMP] = ACTIONS(4168),
    [anon_sym_EQ_EQ] = ACTIONS(4170),
    [anon_sym_BANG_EQ] = ACTIONS(4170),
    [anon_sym_GT] = ACTIONS(4168),
    [anon_sym_GT_EQ] = ACTIONS(4170),
    [anon_sym_LT_EQ] = ACTIONS(4170),
    [anon_sym_LT] = ACTIONS(4168),
    [anon_sym_LT_LT] = ACTIONS(4170),
    [anon_sym_GT_GT] = ACTIONS(4170),
    [anon_sym_SEMI] = ACTIONS(4170),
    [anon_sym___extension__] = ACTIONS(4168),
    [anon_sym_extern] = ACTIONS(4168),
    [anon_sym___attribute__] = ACTIONS(4168),
    [anon_sym___attribute] = ACTIONS(4168),
    [anon_sym_noreturn] = ACTIONS(4168),
    [anon_sym_LBRACK] = ACTIONS(4170),
    [anon_sym_RBRACK] = ACTIONS(4170),
    [anon_sym___declspec] = ACTIONS(4168),
    [anon_sym_LBRACE] = ACTIONS(4170),
    [anon_sym_RBRACE] = ACTIONS(4170),
    [anon_sym_static] = ACTIONS(4168),
    [anon_sym_EQ] = ACTIONS(4168),
    [anon_sym_ATautoreleasepool] = ACTIONS(4170),
    [anon_sym_auto] = ACTIONS(4168),
    [anon_sym_register] = ACTIONS(4168),
    [anon_sym_inline] = ACTIONS(4168),
    [anon_sym___inline] = ACTIONS(4168),
    [anon_sym___inline__] = ACTIONS(4168),
    [anon_sym___forceinline] = ACTIONS(4168),
    [anon_sym_thread_local] = ACTIONS(4168),
    [anon_sym___thread] = ACTIONS(4168),
    [anon_sym_CG_EXTERN] = ACTIONS(4168),
    [anon_sym_CG_INLINE] = ACTIONS(4168),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4168),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4168),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4168),
    [anon_sym_IBOutlet] = ACTIONS(4168),
    [anon_sym_IBInspectable] = ACTIONS(4168),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4168),
    [anon_sym_NS_INLINE] = ACTIONS(4168),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4168),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4168),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4168),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4168),
    [anon_sym_const] = ACTIONS(4168),
    [anon_sym_constexpr] = ACTIONS(4168),
    [anon_sym_volatile] = ACTIONS(4168),
    [anon_sym_restrict] = ACTIONS(4168),
    [anon_sym___restrict__] = ACTIONS(4168),
    [anon_sym__Atomic] = ACTIONS(4168),
    [anon_sym__Noreturn] = ACTIONS(4168),
    [anon_sym__Nonnull] = ACTIONS(4168),
    [anon_sym_nullable] = ACTIONS(4168),
    [anon_sym__Complex] = ACTIONS(4168),
    [anon_sym__Nullable] = ACTIONS(4168),
    [anon_sym__Nullable_result] = ACTIONS(4168),
    [anon_sym__Null_unspecified] = ACTIONS(4168),
    [anon_sym___autoreleasing] = ACTIONS(4168),
    [anon_sym___block] = ACTIONS(4168),
    [anon_sym___bridge] = ACTIONS(4168),
    [anon_sym___bridge_retained] = ACTIONS(4168),
    [anon_sym___bridge_transfer] = ACTIONS(4168),
    [anon_sym___complex] = ACTIONS(4168),
    [anon_sym___const] = ACTIONS(4168),
    [anon_sym___imag] = ACTIONS(4168),
    [anon_sym___kindof] = ACTIONS(4168),
    [anon_sym___nonnull] = ACTIONS(4168),
    [anon_sym___nullable] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4168),
    [anon_sym___real] = ACTIONS(4168),
    [anon_sym___strong] = ACTIONS(4168),
    [anon_sym___unsafe_unretained] = ACTIONS(4168),
    [anon_sym___unused] = ACTIONS(4168),
    [anon_sym___weak] = ACTIONS(4168),
    [anon_sym_alignas] = ACTIONS(4168),
    [anon_sym__Alignas] = ACTIONS(4168),
    [anon_sym_COLON] = ACTIONS(4170),
    [anon_sym_in] = ACTIONS(4168),
    [anon_sym_QMARK] = ACTIONS(4170),
    [anon_sym_DASH_DASH] = ACTIONS(4170),
    [anon_sym_PLUS_PLUS] = ACTIONS(4170),
    [anon_sym_asm] = ACTIONS(4168),
    [anon_sym___asm__] = ACTIONS(4168),
    [anon_sym___asm] = ACTIONS(4168),
    [anon_sym_DOT] = ACTIONS(4168),
    [anon_sym_DASH_GT] = ACTIONS(4170),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4168),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4168),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4168),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4168),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4168),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4168),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4168),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4168),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4168),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_API_AVAILABLE] = ACTIONS(4168),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_API_DEPRECATED] = ACTIONS(4168),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4168),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4168),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4168),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4168),
    [anon_sym___deprecated_msg] = ACTIONS(4168),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4168),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4168),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4168),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4168),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4168),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4168),
  },
  [STATE(1982)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_compound_statement] = STATE(2140),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(4854),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1983)] = {
    [sym_compound_statement] = STATE(5317),
    [sym_type_qualifier] = STATE(3320),
    [sym_alignas_qualifier] = STATE(4149),
    [sym_expression] = STATE(4407),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_objc_bridge] = STATE(5947),
    [sym_typeof_specifier] = STATE(5317),
    [sym_availability] = STATE(5947),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5080),
    [anon_sym_RPAREN] = ACTIONS(5082),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5084),
    [anon_sym_noreturn] = ACTIONS(5086),
    [anon_sym_nothrow] = ACTIONS(5088),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(5090),
    [anon_sym_constexpr] = ACTIONS(5092),
    [anon_sym_volatile] = ACTIONS(5092),
    [anon_sym_restrict] = ACTIONS(5092),
    [anon_sym___restrict__] = ACTIONS(5092),
    [anon_sym__Atomic] = ACTIONS(5092),
    [anon_sym__Noreturn] = ACTIONS(5092),
    [anon_sym__Nonnull] = ACTIONS(5092),
    [anon_sym_nullable] = ACTIONS(5092),
    [anon_sym__Complex] = ACTIONS(5092),
    [anon_sym__Nullable] = ACTIONS(5092),
    [anon_sym__Nullable_result] = ACTIONS(5092),
    [anon_sym__Null_unspecified] = ACTIONS(5092),
    [anon_sym___autoreleasing] = ACTIONS(5092),
    [anon_sym___block] = ACTIONS(5092),
    [anon_sym___bridge] = ACTIONS(5092),
    [anon_sym___bridge_retained] = ACTIONS(5092),
    [anon_sym___bridge_transfer] = ACTIONS(5092),
    [anon_sym___complex] = ACTIONS(5092),
    [anon_sym___const] = ACTIONS(5092),
    [anon_sym___imag] = ACTIONS(5094),
    [anon_sym___kindof] = ACTIONS(5092),
    [anon_sym___nonnull] = ACTIONS(5092),
    [anon_sym___nullable] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5092),
    [anon_sym___real] = ACTIONS(5094),
    [anon_sym___strong] = ACTIONS(5092),
    [anon_sym___unsafe_unretained] = ACTIONS(5092),
    [anon_sym___unused] = ACTIONS(5092),
    [anon_sym___weak] = ACTIONS(5092),
    [anon_sym_alignas] = ACTIONS(5096),
    [anon_sym__Alignas] = ACTIONS(5096),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_objc_bridge_related] = ACTIONS(5098),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_availability] = ACTIONS(5100),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(1984)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1774),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5102),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1985)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1774),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5104),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1986)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1773),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5106),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1987)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1783),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5108),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1988)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1790),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5110),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1989)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1818),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5112),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1990)] = {
    [sym_identifier] = ACTIONS(2856),
    [aux_sym_preproc_def_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2856),
    [sym_preproc_directive] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2858),
    [anon_sym_PLUS] = ACTIONS(2858),
    [anon_sym___extension__] = ACTIONS(2856),
    [anon_sym_typedef] = ACTIONS(2856),
    [anon_sym_extern] = ACTIONS(2856),
    [anon_sym___attribute__] = ACTIONS(2856),
    [anon_sym___attribute] = ACTIONS(2856),
    [anon_sym_noreturn] = ACTIONS(2856),
    [anon_sym_LBRACK] = ACTIONS(2858),
    [anon_sym___declspec] = ACTIONS(2856),
    [anon_sym___cdecl] = ACTIONS(2856),
    [anon_sym___clrcall] = ACTIONS(2856),
    [anon_sym___stdcall] = ACTIONS(2856),
    [anon_sym___fastcall] = ACTIONS(2856),
    [anon_sym___thiscall] = ACTIONS(2856),
    [anon_sym___vectorcall] = ACTIONS(2856),
    [anon_sym_RBRACE] = ACTIONS(2858),
    [anon_sym_signed] = ACTIONS(2856),
    [anon_sym_unsigned] = ACTIONS(2856),
    [anon_sym_long] = ACTIONS(2856),
    [anon_sym_short] = ACTIONS(2856),
    [anon_sym_static] = ACTIONS(2856),
    [anon_sym_auto] = ACTIONS(2856),
    [anon_sym_register] = ACTIONS(2856),
    [anon_sym_inline] = ACTIONS(2856),
    [anon_sym___inline] = ACTIONS(2856),
    [anon_sym___inline__] = ACTIONS(2856),
    [anon_sym___forceinline] = ACTIONS(2856),
    [anon_sym_thread_local] = ACTIONS(2856),
    [anon_sym___thread] = ACTIONS(2856),
    [anon_sym_CG_EXTERN] = ACTIONS(2856),
    [anon_sym_CG_INLINE] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2856),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2856),
    [anon_sym_IBOutlet] = ACTIONS(2856),
    [anon_sym_IBInspectable] = ACTIONS(2856),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2856),
    [anon_sym_NS_INLINE] = ACTIONS(2856),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2856),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2856),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2856),
    [anon_sym_const] = ACTIONS(2856),
    [anon_sym_constexpr] = ACTIONS(2856),
    [anon_sym_volatile] = ACTIONS(2856),
    [anon_sym_restrict] = ACTIONS(2856),
    [anon_sym___restrict__] = ACTIONS(2856),
    [anon_sym__Atomic] = ACTIONS(2856),
    [anon_sym__Noreturn] = ACTIONS(2856),
    [anon_sym__Nonnull] = ACTIONS(2856),
    [anon_sym_nullable] = ACTIONS(2856),
    [anon_sym__Complex] = ACTIONS(2856),
    [anon_sym__Nullable] = ACTIONS(2856),
    [anon_sym__Nullable_result] = ACTIONS(2856),
    [anon_sym__Null_unspecified] = ACTIONS(2856),
    [anon_sym___autoreleasing] = ACTIONS(2856),
    [anon_sym___block] = ACTIONS(2856),
    [anon_sym___bridge] = ACTIONS(2856),
    [anon_sym___bridge_retained] = ACTIONS(2856),
    [anon_sym___bridge_transfer] = ACTIONS(2856),
    [anon_sym___complex] = ACTIONS(2856),
    [anon_sym___const] = ACTIONS(2856),
    [anon_sym___imag] = ACTIONS(2856),
    [anon_sym___kindof] = ACTIONS(2856),
    [anon_sym___nonnull] = ACTIONS(2856),
    [anon_sym___nullable] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2856),
    [anon_sym___real] = ACTIONS(2856),
    [anon_sym___strong] = ACTIONS(2856),
    [anon_sym___unsafe_unretained] = ACTIONS(2856),
    [anon_sym___unused] = ACTIONS(2856),
    [anon_sym___weak] = ACTIONS(2856),
    [anon_sym_alignas] = ACTIONS(2856),
    [anon_sym__Alignas] = ACTIONS(2856),
    [sym_primitive_type] = ACTIONS(2856),
    [anon_sym_enum] = ACTIONS(2856),
    [anon_sym_struct] = ACTIONS(2856),
    [anon_sym_union] = ACTIONS(2856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2856),
    [anon_sym___typeof] = ACTIONS(2856),
    [anon_sym_typeof] = ACTIONS(2856),
    [aux_sym_preproc_undef_token1] = ACTIONS(2856),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2856),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2856),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2856),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_API_AVAILABLE] = ACTIONS(2856),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_API_DEPRECATED] = ACTIONS(2856),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym___deprecated_msg] = ACTIONS(2856),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2856),
    [anon_sym_ATend] = ACTIONS(2858),
    [anon_sym_AToptional] = ACTIONS(2858),
    [anon_sym_ATrequired] = ACTIONS(2858),
    [anon_sym_ATsynthesize] = ACTIONS(2858),
    [anon_sym_ATdynamic] = ACTIONS(2858),
    [anon_sym_ATproperty] = ACTIONS(2858),
    [anon_sym_BOOL] = ACTIONS(2856),
    [anon_sym_IMP] = ACTIONS(2856),
    [anon_sym_SEL] = ACTIONS(2856),
    [anon_sym_Class] = ACTIONS(2856),
    [anon_sym_id] = ACTIONS(2856),
    [anon_sym_ATdefs] = ACTIONS(2858),
  },
  [STATE(1991)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4801),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(3017),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5114),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1992)] = {
    [sym_identifier] = ACTIONS(5116),
    [aux_sym_preproc_def_token1] = ACTIONS(5116),
    [aux_sym_preproc_if_token1] = ACTIONS(5116),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5116),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5116),
    [sym_preproc_directive] = ACTIONS(5116),
    [anon_sym_LPAREN2] = ACTIONS(5118),
    [anon_sym_DASH] = ACTIONS(5118),
    [anon_sym_PLUS] = ACTIONS(5118),
    [anon_sym_LT] = ACTIONS(5118),
    [anon_sym___extension__] = ACTIONS(5116),
    [anon_sym_typedef] = ACTIONS(5116),
    [anon_sym_extern] = ACTIONS(5116),
    [anon_sym___attribute__] = ACTIONS(5116),
    [anon_sym___attribute] = ACTIONS(5116),
    [anon_sym_noreturn] = ACTIONS(5116),
    [anon_sym_LBRACK] = ACTIONS(5118),
    [anon_sym___declspec] = ACTIONS(5116),
    [anon_sym___cdecl] = ACTIONS(5116),
    [anon_sym___clrcall] = ACTIONS(5116),
    [anon_sym___stdcall] = ACTIONS(5116),
    [anon_sym___fastcall] = ACTIONS(5116),
    [anon_sym___thiscall] = ACTIONS(5116),
    [anon_sym___vectorcall] = ACTIONS(5116),
    [anon_sym_LBRACE] = ACTIONS(5118),
    [anon_sym_signed] = ACTIONS(5116),
    [anon_sym_unsigned] = ACTIONS(5116),
    [anon_sym_long] = ACTIONS(5116),
    [anon_sym_short] = ACTIONS(5116),
    [anon_sym_static] = ACTIONS(5116),
    [anon_sym_auto] = ACTIONS(5116),
    [anon_sym_register] = ACTIONS(5116),
    [anon_sym_inline] = ACTIONS(5116),
    [anon_sym___inline] = ACTIONS(5116),
    [anon_sym___inline__] = ACTIONS(5116),
    [anon_sym___forceinline] = ACTIONS(5116),
    [anon_sym_thread_local] = ACTIONS(5116),
    [anon_sym___thread] = ACTIONS(5116),
    [anon_sym_CG_EXTERN] = ACTIONS(5116),
    [anon_sym_CG_INLINE] = ACTIONS(5116),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5116),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5116),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5116),
    [anon_sym_IBOutlet] = ACTIONS(5116),
    [anon_sym_IBInspectable] = ACTIONS(5116),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5116),
    [anon_sym_NS_INLINE] = ACTIONS(5116),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5116),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5116),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5116),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5116),
    [anon_sym_const] = ACTIONS(5116),
    [anon_sym_constexpr] = ACTIONS(5116),
    [anon_sym_volatile] = ACTIONS(5116),
    [anon_sym_restrict] = ACTIONS(5116),
    [anon_sym___restrict__] = ACTIONS(5116),
    [anon_sym__Atomic] = ACTIONS(5116),
    [anon_sym__Noreturn] = ACTIONS(5116),
    [anon_sym__Nonnull] = ACTIONS(5116),
    [anon_sym_nullable] = ACTIONS(5116),
    [anon_sym__Complex] = ACTIONS(5116),
    [anon_sym__Nullable] = ACTIONS(5116),
    [anon_sym__Nullable_result] = ACTIONS(5116),
    [anon_sym__Null_unspecified] = ACTIONS(5116),
    [anon_sym___autoreleasing] = ACTIONS(5116),
    [anon_sym___block] = ACTIONS(5116),
    [anon_sym___bridge] = ACTIONS(5116),
    [anon_sym___bridge_retained] = ACTIONS(5116),
    [anon_sym___bridge_transfer] = ACTIONS(5116),
    [anon_sym___complex] = ACTIONS(5116),
    [anon_sym___const] = ACTIONS(5116),
    [anon_sym___imag] = ACTIONS(5116),
    [anon_sym___kindof] = ACTIONS(5116),
    [anon_sym___nonnull] = ACTIONS(5116),
    [anon_sym___nullable] = ACTIONS(5116),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5116),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5116),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5116),
    [anon_sym___real] = ACTIONS(5116),
    [anon_sym___strong] = ACTIONS(5116),
    [anon_sym___unsafe_unretained] = ACTIONS(5116),
    [anon_sym___unused] = ACTIONS(5116),
    [anon_sym___weak] = ACTIONS(5116),
    [anon_sym_alignas] = ACTIONS(5116),
    [anon_sym__Alignas] = ACTIONS(5116),
    [sym_primitive_type] = ACTIONS(5116),
    [anon_sym_enum] = ACTIONS(5116),
    [anon_sym_COLON] = ACTIONS(5118),
    [anon_sym_struct] = ACTIONS(5116),
    [anon_sym_union] = ACTIONS(5116),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5116),
    [anon_sym___typeof] = ACTIONS(5116),
    [anon_sym_typeof] = ACTIONS(5116),
    [aux_sym_preproc_undef_token1] = ACTIONS(5116),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5116),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5116),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5116),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5116),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5116),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5116),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5116),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5116),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5116),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5116),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5116),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5116),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5116),
    [anon_sym_API_AVAILABLE] = ACTIONS(5116),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5116),
    [anon_sym_API_DEPRECATED] = ACTIONS(5116),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5116),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5116),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5116),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5116),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5116),
    [anon_sym___deprecated_msg] = ACTIONS(5116),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5116),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5116),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5116),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5116),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5116),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5116),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5116),
    [anon_sym_ATend] = ACTIONS(5118),
    [anon_sym_ATsynthesize] = ACTIONS(5118),
    [anon_sym_ATdynamic] = ACTIONS(5118),
    [anon_sym_ATproperty] = ACTIONS(5118),
    [anon_sym_BOOL] = ACTIONS(5116),
    [anon_sym_IMP] = ACTIONS(5116),
    [anon_sym_SEL] = ACTIONS(5116),
    [anon_sym_Class] = ACTIONS(5116),
    [anon_sym_id] = ACTIONS(5116),
  },
  [STATE(1993)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5120),
    [anon_sym_COMMA] = ACTIONS(5122),
    [anon_sym_RPAREN] = ACTIONS(5122),
    [anon_sym_LPAREN2] = ACTIONS(5122),
    [anon_sym_STAR] = ACTIONS(5122),
    [anon_sym_CARET] = ACTIONS(5122),
    [anon_sym_GT] = ACTIONS(5122),
    [anon_sym_SEMI] = ACTIONS(5122),
    [anon_sym___extension__] = ACTIONS(5124),
    [anon_sym_extern] = ACTIONS(5127),
    [anon_sym___attribute__] = ACTIONS(5130),
    [anon_sym___attribute] = ACTIONS(5130),
    [anon_sym_noreturn] = ACTIONS(5124),
    [anon_sym_LBRACK] = ACTIONS(5133),
    [anon_sym___declspec] = ACTIONS(5136),
    [anon_sym___based] = ACTIONS(5120),
    [anon_sym_LBRACE] = ACTIONS(5122),
    [anon_sym_signed] = ACTIONS(5120),
    [anon_sym_unsigned] = ACTIONS(5120),
    [anon_sym_long] = ACTIONS(5120),
    [anon_sym_short] = ACTIONS(5120),
    [anon_sym_static] = ACTIONS(5127),
    [anon_sym_ATautoreleasepool] = ACTIONS(5122),
    [anon_sym_auto] = ACTIONS(5127),
    [anon_sym_register] = ACTIONS(5127),
    [anon_sym_inline] = ACTIONS(5127),
    [anon_sym___inline] = ACTIONS(5127),
    [anon_sym___inline__] = ACTIONS(5127),
    [anon_sym___forceinline] = ACTIONS(5127),
    [anon_sym_thread_local] = ACTIONS(5127),
    [anon_sym___thread] = ACTIONS(5127),
    [anon_sym_CG_EXTERN] = ACTIONS(5127),
    [anon_sym_CG_INLINE] = ACTIONS(5127),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5127),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5127),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5127),
    [anon_sym_IBOutlet] = ACTIONS(5127),
    [anon_sym_IBInspectable] = ACTIONS(5127),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5127),
    [anon_sym_NS_INLINE] = ACTIONS(5127),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5127),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5127),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5127),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5127),
    [anon_sym_const] = ACTIONS(5124),
    [anon_sym_constexpr] = ACTIONS(5124),
    [anon_sym_volatile] = ACTIONS(5124),
    [anon_sym_restrict] = ACTIONS(5124),
    [anon_sym___restrict__] = ACTIONS(5124),
    [anon_sym__Atomic] = ACTIONS(5124),
    [anon_sym__Noreturn] = ACTIONS(5124),
    [anon_sym__Nonnull] = ACTIONS(5124),
    [anon_sym_nullable] = ACTIONS(5124),
    [anon_sym__Complex] = ACTIONS(5124),
    [anon_sym__Nullable] = ACTIONS(5124),
    [anon_sym__Nullable_result] = ACTIONS(5124),
    [anon_sym__Null_unspecified] = ACTIONS(5124),
    [anon_sym___autoreleasing] = ACTIONS(5124),
    [anon_sym___block] = ACTIONS(5124),
    [anon_sym___bridge] = ACTIONS(5124),
    [anon_sym___bridge_retained] = ACTIONS(5124),
    [anon_sym___bridge_transfer] = ACTIONS(5124),
    [anon_sym___complex] = ACTIONS(5124),
    [anon_sym___const] = ACTIONS(5124),
    [anon_sym___imag] = ACTIONS(5124),
    [anon_sym___kindof] = ACTIONS(5124),
    [anon_sym___nonnull] = ACTIONS(5124),
    [anon_sym___nullable] = ACTIONS(5124),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5124),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5124),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5124),
    [anon_sym___real] = ACTIONS(5124),
    [anon_sym___strong] = ACTIONS(5124),
    [anon_sym___unsafe_unretained] = ACTIONS(5124),
    [anon_sym___unused] = ACTIONS(5124),
    [anon_sym___weak] = ACTIONS(5124),
    [anon_sym_alignas] = ACTIONS(5139),
    [anon_sym__Alignas] = ACTIONS(5139),
    [sym_primitive_type] = ACTIONS(5120),
    [anon_sym_enum] = ACTIONS(5120),
    [anon_sym_COLON] = ACTIONS(5122),
    [anon_sym_struct] = ACTIONS(5120),
    [anon_sym_union] = ACTIONS(5120),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5120),
    [anon_sym___typeof] = ACTIONS(5120),
    [anon_sym_typeof] = ACTIONS(5120),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5142),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5142),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5142),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5142),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5142),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5142),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5142),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5142),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5142),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5145),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5145),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5145),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5145),
    [anon_sym_API_AVAILABLE] = ACTIONS(5145),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5145),
    [anon_sym_API_DEPRECATED] = ACTIONS(5145),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5145),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5145),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5145),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5145),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5145),
    [anon_sym___deprecated_msg] = ACTIONS(5145),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5145),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5145),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5145),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5145),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5145),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5145),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5145),
    [anon_sym_ATprotocol] = ACTIONS(5122),
    [anon_sym_ATinterface] = ACTIONS(5122),
    [anon_sym_ATimplementation] = ACTIONS(5122),
    [anon_sym_BOOL] = ACTIONS(5120),
    [anon_sym_IMP] = ACTIONS(5120),
    [anon_sym_SEL] = ACTIONS(5120),
    [anon_sym_Class] = ACTIONS(5120),
    [anon_sym_id] = ACTIONS(5120),
  },
  [STATE(1994)] = {
    [sym_declaration] = STATE(1891),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1891),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1995)] = {
    [sym_declaration] = STATE(1879),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1879),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1996)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4801),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(3042),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5148),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1997)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1779),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5150),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1998)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1779),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5152),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(1999)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1787),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5154),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2000)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1787),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5156),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2001)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1792),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5158),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2002)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1792),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5160),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2003)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1799),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5162),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2004)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1767),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5164),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2005)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1814),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5166),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2006)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1816),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5168),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2007)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4801),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(3030),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5170),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2008)] = {
    [sym_identifier] = ACTIONS(4118),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4120),
    [anon_sym_COMMA] = ACTIONS(4120),
    [anon_sym_RPAREN] = ACTIONS(4120),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4120),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4120),
    [anon_sym_LPAREN2] = ACTIONS(4120),
    [anon_sym_DASH] = ACTIONS(4118),
    [anon_sym_PLUS] = ACTIONS(4118),
    [anon_sym_STAR] = ACTIONS(4120),
    [anon_sym_SLASH] = ACTIONS(4118),
    [anon_sym_PERCENT] = ACTIONS(4120),
    [anon_sym_PIPE_PIPE] = ACTIONS(4120),
    [anon_sym_AMP_AMP] = ACTIONS(4120),
    [anon_sym_PIPE] = ACTIONS(4118),
    [anon_sym_CARET] = ACTIONS(4120),
    [anon_sym_AMP] = ACTIONS(4118),
    [anon_sym_EQ_EQ] = ACTIONS(4120),
    [anon_sym_BANG_EQ] = ACTIONS(4120),
    [anon_sym_GT] = ACTIONS(4118),
    [anon_sym_GT_EQ] = ACTIONS(4120),
    [anon_sym_LT_EQ] = ACTIONS(4120),
    [anon_sym_LT] = ACTIONS(4118),
    [anon_sym_LT_LT] = ACTIONS(4120),
    [anon_sym_GT_GT] = ACTIONS(4120),
    [anon_sym_SEMI] = ACTIONS(4120),
    [anon_sym___extension__] = ACTIONS(4118),
    [anon_sym_extern] = ACTIONS(4118),
    [anon_sym___attribute__] = ACTIONS(4118),
    [anon_sym___attribute] = ACTIONS(4118),
    [anon_sym_noreturn] = ACTIONS(4118),
    [anon_sym_LBRACK] = ACTIONS(4120),
    [anon_sym_RBRACK] = ACTIONS(4120),
    [anon_sym___declspec] = ACTIONS(4118),
    [anon_sym_RBRACE] = ACTIONS(4120),
    [anon_sym_static] = ACTIONS(4118),
    [anon_sym_auto] = ACTIONS(4118),
    [anon_sym_register] = ACTIONS(4118),
    [anon_sym_inline] = ACTIONS(4118),
    [anon_sym___inline] = ACTIONS(4118),
    [anon_sym___inline__] = ACTIONS(4118),
    [anon_sym___forceinline] = ACTIONS(4118),
    [anon_sym_thread_local] = ACTIONS(4118),
    [anon_sym___thread] = ACTIONS(4118),
    [anon_sym_CG_EXTERN] = ACTIONS(4118),
    [anon_sym_CG_INLINE] = ACTIONS(4118),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4118),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4118),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4118),
    [anon_sym_IBOutlet] = ACTIONS(4118),
    [anon_sym_IBInspectable] = ACTIONS(4118),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4118),
    [anon_sym_NS_INLINE] = ACTIONS(4118),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4118),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4118),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4118),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4118),
    [anon_sym_const] = ACTIONS(4118),
    [anon_sym_constexpr] = ACTIONS(4118),
    [anon_sym_volatile] = ACTIONS(4118),
    [anon_sym_restrict] = ACTIONS(4118),
    [anon_sym___restrict__] = ACTIONS(4118),
    [anon_sym__Atomic] = ACTIONS(4118),
    [anon_sym__Noreturn] = ACTIONS(4118),
    [anon_sym__Nonnull] = ACTIONS(4118),
    [anon_sym_nullable] = ACTIONS(4118),
    [anon_sym__Complex] = ACTIONS(4118),
    [anon_sym__Nullable] = ACTIONS(4118),
    [anon_sym__Nullable_result] = ACTIONS(4118),
    [anon_sym__Null_unspecified] = ACTIONS(4118),
    [anon_sym___autoreleasing] = ACTIONS(4118),
    [anon_sym___block] = ACTIONS(4118),
    [anon_sym___bridge] = ACTIONS(4118),
    [anon_sym___bridge_retained] = ACTIONS(4118),
    [anon_sym___bridge_transfer] = ACTIONS(4118),
    [anon_sym___complex] = ACTIONS(4118),
    [anon_sym___const] = ACTIONS(4118),
    [anon_sym___imag] = ACTIONS(4118),
    [anon_sym___kindof] = ACTIONS(4118),
    [anon_sym___nonnull] = ACTIONS(4118),
    [anon_sym___nullable] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4118),
    [anon_sym___real] = ACTIONS(4118),
    [anon_sym___strong] = ACTIONS(4118),
    [anon_sym___unsafe_unretained] = ACTIONS(4118),
    [anon_sym___unused] = ACTIONS(4118),
    [anon_sym___weak] = ACTIONS(4118),
    [anon_sym_alignas] = ACTIONS(4118),
    [anon_sym__Alignas] = ACTIONS(4118),
    [anon_sym_COLON] = ACTIONS(4120),
    [anon_sym_QMARK] = ACTIONS(4120),
    [anon_sym_DASH_DASH] = ACTIONS(4120),
    [anon_sym_PLUS_PLUS] = ACTIONS(4120),
    [anon_sym_DOT] = ACTIONS(4118),
    [anon_sym_DASH_GT] = ACTIONS(4120),
    [anon_sym_AT] = ACTIONS(4120),
    [anon_sym_DQUOTE] = ACTIONS(4120),
    [anon_sym_L_DQUOTE] = ACTIONS(4120),
    [anon_sym_u_DQUOTE] = ACTIONS(4120),
    [anon_sym_U_DQUOTE] = ACTIONS(4120),
    [anon_sym_u8_DQUOTE] = ACTIONS(4120),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4118),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4118),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4118),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4118),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4118),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4118),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4118),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_API_AVAILABLE] = ACTIONS(4118),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_API_DEPRECATED] = ACTIONS(4118),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4118),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym___deprecated_msg] = ACTIONS(4118),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4118),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4118),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4118),
  },
  [STATE(2009)] = {
    [sym_identifier] = ACTIONS(5172),
    [aux_sym_preproc_def_token1] = ACTIONS(5172),
    [aux_sym_preproc_if_token1] = ACTIONS(5172),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5172),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5172),
    [sym_preproc_directive] = ACTIONS(5172),
    [anon_sym_LPAREN2] = ACTIONS(5174),
    [anon_sym_DASH] = ACTIONS(5174),
    [anon_sym_PLUS] = ACTIONS(5174),
    [anon_sym_LT] = ACTIONS(5174),
    [anon_sym_SEMI] = ACTIONS(5176),
    [anon_sym___extension__] = ACTIONS(5172),
    [anon_sym_typedef] = ACTIONS(5172),
    [anon_sym_extern] = ACTIONS(5172),
    [anon_sym___attribute__] = ACTIONS(5172),
    [anon_sym___attribute] = ACTIONS(5172),
    [anon_sym_noreturn] = ACTIONS(5172),
    [anon_sym_LBRACK] = ACTIONS(5174),
    [anon_sym___declspec] = ACTIONS(5172),
    [anon_sym___cdecl] = ACTIONS(5172),
    [anon_sym___clrcall] = ACTIONS(5172),
    [anon_sym___stdcall] = ACTIONS(5172),
    [anon_sym___fastcall] = ACTIONS(5172),
    [anon_sym___thiscall] = ACTIONS(5172),
    [anon_sym___vectorcall] = ACTIONS(5172),
    [anon_sym_LBRACE] = ACTIONS(5174),
    [anon_sym_signed] = ACTIONS(5172),
    [anon_sym_unsigned] = ACTIONS(5172),
    [anon_sym_long] = ACTIONS(5172),
    [anon_sym_short] = ACTIONS(5172),
    [anon_sym_static] = ACTIONS(5172),
    [anon_sym_auto] = ACTIONS(5172),
    [anon_sym_register] = ACTIONS(5172),
    [anon_sym_inline] = ACTIONS(5172),
    [anon_sym___inline] = ACTIONS(5172),
    [anon_sym___inline__] = ACTIONS(5172),
    [anon_sym___forceinline] = ACTIONS(5172),
    [anon_sym_thread_local] = ACTIONS(5172),
    [anon_sym___thread] = ACTIONS(5172),
    [anon_sym_CG_EXTERN] = ACTIONS(5172),
    [anon_sym_CG_INLINE] = ACTIONS(5172),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5172),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5172),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5172),
    [anon_sym_IBOutlet] = ACTIONS(5172),
    [anon_sym_IBInspectable] = ACTIONS(5172),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5172),
    [anon_sym_NS_INLINE] = ACTIONS(5172),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5172),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5172),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5172),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5172),
    [anon_sym_const] = ACTIONS(5172),
    [anon_sym_constexpr] = ACTIONS(5172),
    [anon_sym_volatile] = ACTIONS(5172),
    [anon_sym_restrict] = ACTIONS(5172),
    [anon_sym___restrict__] = ACTIONS(5172),
    [anon_sym__Atomic] = ACTIONS(5172),
    [anon_sym__Noreturn] = ACTIONS(5172),
    [anon_sym__Nonnull] = ACTIONS(5172),
    [anon_sym_nullable] = ACTIONS(5172),
    [anon_sym__Complex] = ACTIONS(5172),
    [anon_sym__Nullable] = ACTIONS(5172),
    [anon_sym__Nullable_result] = ACTIONS(5172),
    [anon_sym__Null_unspecified] = ACTIONS(5172),
    [anon_sym___autoreleasing] = ACTIONS(5172),
    [anon_sym___block] = ACTIONS(5172),
    [anon_sym___bridge] = ACTIONS(5172),
    [anon_sym___bridge_retained] = ACTIONS(5172),
    [anon_sym___bridge_transfer] = ACTIONS(5172),
    [anon_sym___complex] = ACTIONS(5172),
    [anon_sym___const] = ACTIONS(5172),
    [anon_sym___imag] = ACTIONS(5172),
    [anon_sym___kindof] = ACTIONS(5172),
    [anon_sym___nonnull] = ACTIONS(5172),
    [anon_sym___nullable] = ACTIONS(5172),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5172),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5172),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5172),
    [anon_sym___real] = ACTIONS(5172),
    [anon_sym___strong] = ACTIONS(5172),
    [anon_sym___unsafe_unretained] = ACTIONS(5172),
    [anon_sym___unused] = ACTIONS(5172),
    [anon_sym___weak] = ACTIONS(5172),
    [anon_sym_alignas] = ACTIONS(5172),
    [anon_sym__Alignas] = ACTIONS(5172),
    [sym_primitive_type] = ACTIONS(5172),
    [anon_sym_enum] = ACTIONS(5172),
    [anon_sym_COLON] = ACTIONS(5174),
    [anon_sym_struct] = ACTIONS(5172),
    [anon_sym_union] = ACTIONS(5172),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5172),
    [anon_sym___typeof] = ACTIONS(5172),
    [anon_sym_typeof] = ACTIONS(5172),
    [aux_sym_preproc_undef_token1] = ACTIONS(5172),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5172),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5172),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5172),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5172),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5172),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5172),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5172),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5172),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5172),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5172),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5172),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5172),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5172),
    [anon_sym_API_AVAILABLE] = ACTIONS(5172),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5172),
    [anon_sym_API_DEPRECATED] = ACTIONS(5172),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5172),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5172),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5172),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5172),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5172),
    [anon_sym___deprecated_msg] = ACTIONS(5172),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5172),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5172),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5172),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5172),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5172),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5172),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5172),
    [anon_sym_ATend] = ACTIONS(5174),
    [anon_sym_ATsynthesize] = ACTIONS(5174),
    [anon_sym_ATdynamic] = ACTIONS(5174),
    [anon_sym_BOOL] = ACTIONS(5172),
    [anon_sym_IMP] = ACTIONS(5172),
    [anon_sym_SEL] = ACTIONS(5172),
    [anon_sym_Class] = ACTIONS(5172),
    [anon_sym_id] = ACTIONS(5172),
  },
  [STATE(2010)] = {
    [sym_declaration] = STATE(1877),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1877),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2011)] = {
    [sym_declaration] = STATE(1878),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1878),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2012)] = {
    [sym_declaration] = STATE(1884),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1884),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2013)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4801),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(3015),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5178),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2014)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4801),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(3027),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5180),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2015)] = {
    [sym_declaration] = STATE(1890),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1890),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2016)] = {
    [sym_declaration] = STATE(1893),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1893),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2017)] = {
    [sym_declaration] = STATE(1894),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1894),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2018)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1812),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5182),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2019)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(5184),
    [anon_sym_ATinterface] = ACTIONS(5186),
    [anon_sym_ATimplementation] = ACTIONS(5188),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2020)] = {
    [sym_compound_statement] = STATE(5317),
    [sym_type_qualifier] = STATE(3320),
    [sym_alignas_qualifier] = STATE(4149),
    [sym_expression] = STATE(4407),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_objc_bridge] = STATE(5947),
    [sym_typeof_specifier] = STATE(5317),
    [sym_availability] = STATE(5947),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5080),
    [anon_sym_RPAREN] = ACTIONS(5082),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5084),
    [anon_sym_noreturn] = ACTIONS(5190),
    [anon_sym_nothrow] = ACTIONS(5192),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(5090),
    [anon_sym_constexpr] = ACTIONS(5092),
    [anon_sym_volatile] = ACTIONS(5092),
    [anon_sym_restrict] = ACTIONS(5092),
    [anon_sym___restrict__] = ACTIONS(5092),
    [anon_sym__Atomic] = ACTIONS(5092),
    [anon_sym__Noreturn] = ACTIONS(5092),
    [anon_sym__Nonnull] = ACTIONS(5092),
    [anon_sym_nullable] = ACTIONS(5092),
    [anon_sym__Complex] = ACTIONS(5092),
    [anon_sym__Nullable] = ACTIONS(5092),
    [anon_sym__Nullable_result] = ACTIONS(5092),
    [anon_sym__Null_unspecified] = ACTIONS(5092),
    [anon_sym___autoreleasing] = ACTIONS(5092),
    [anon_sym___block] = ACTIONS(5092),
    [anon_sym___bridge] = ACTIONS(5092),
    [anon_sym___bridge_retained] = ACTIONS(5092),
    [anon_sym___bridge_transfer] = ACTIONS(5092),
    [anon_sym___complex] = ACTIONS(5092),
    [anon_sym___const] = ACTIONS(5092),
    [anon_sym___imag] = ACTIONS(5094),
    [anon_sym___kindof] = ACTIONS(5092),
    [anon_sym___nonnull] = ACTIONS(5092),
    [anon_sym___nullable] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5092),
    [anon_sym___real] = ACTIONS(5094),
    [anon_sym___strong] = ACTIONS(5092),
    [anon_sym___unsafe_unretained] = ACTIONS(5092),
    [anon_sym___unused] = ACTIONS(5092),
    [anon_sym___weak] = ACTIONS(5092),
    [anon_sym_alignas] = ACTIONS(5096),
    [anon_sym__Alignas] = ACTIONS(5096),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_objc_bridge_related] = ACTIONS(5098),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_availability] = ACTIONS(5100),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2021)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1771),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5194),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2022)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1781),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5196),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2023)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1781),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5198),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2024)] = {
    [sym_identifier] = ACTIONS(4262),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4264),
    [anon_sym_COMMA] = ACTIONS(4264),
    [anon_sym_RPAREN] = ACTIONS(4264),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4264),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4264),
    [anon_sym_LPAREN2] = ACTIONS(4264),
    [anon_sym_DASH] = ACTIONS(4262),
    [anon_sym_PLUS] = ACTIONS(4262),
    [anon_sym_STAR] = ACTIONS(4264),
    [anon_sym_SLASH] = ACTIONS(4262),
    [anon_sym_PERCENT] = ACTIONS(4264),
    [anon_sym_PIPE_PIPE] = ACTIONS(4264),
    [anon_sym_AMP_AMP] = ACTIONS(4264),
    [anon_sym_PIPE] = ACTIONS(4262),
    [anon_sym_CARET] = ACTIONS(4264),
    [anon_sym_AMP] = ACTIONS(4262),
    [anon_sym_EQ_EQ] = ACTIONS(4264),
    [anon_sym_BANG_EQ] = ACTIONS(4264),
    [anon_sym_GT] = ACTIONS(4262),
    [anon_sym_GT_EQ] = ACTIONS(4264),
    [anon_sym_LT_EQ] = ACTIONS(4264),
    [anon_sym_LT] = ACTIONS(4262),
    [anon_sym_LT_LT] = ACTIONS(4264),
    [anon_sym_GT_GT] = ACTIONS(4264),
    [anon_sym_SEMI] = ACTIONS(4264),
    [anon_sym___extension__] = ACTIONS(4262),
    [anon_sym_extern] = ACTIONS(4262),
    [anon_sym___attribute__] = ACTIONS(4262),
    [anon_sym___attribute] = ACTIONS(4262),
    [anon_sym_noreturn] = ACTIONS(4262),
    [anon_sym_LBRACK] = ACTIONS(4264),
    [anon_sym_RBRACK] = ACTIONS(4264),
    [anon_sym___declspec] = ACTIONS(4262),
    [anon_sym_RBRACE] = ACTIONS(4264),
    [anon_sym_static] = ACTIONS(4262),
    [anon_sym_auto] = ACTIONS(4262),
    [anon_sym_register] = ACTIONS(4262),
    [anon_sym_inline] = ACTIONS(4262),
    [anon_sym___inline] = ACTIONS(4262),
    [anon_sym___inline__] = ACTIONS(4262),
    [anon_sym___forceinline] = ACTIONS(4262),
    [anon_sym_thread_local] = ACTIONS(4262),
    [anon_sym___thread] = ACTIONS(4262),
    [anon_sym_CG_EXTERN] = ACTIONS(4262),
    [anon_sym_CG_INLINE] = ACTIONS(4262),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4262),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4262),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4262),
    [anon_sym_IBOutlet] = ACTIONS(4262),
    [anon_sym_IBInspectable] = ACTIONS(4262),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4262),
    [anon_sym_NS_INLINE] = ACTIONS(4262),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4262),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4262),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4262),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4262),
    [anon_sym_const] = ACTIONS(4262),
    [anon_sym_constexpr] = ACTIONS(4262),
    [anon_sym_volatile] = ACTIONS(4262),
    [anon_sym_restrict] = ACTIONS(4262),
    [anon_sym___restrict__] = ACTIONS(4262),
    [anon_sym__Atomic] = ACTIONS(4262),
    [anon_sym__Noreturn] = ACTIONS(4262),
    [anon_sym__Nonnull] = ACTIONS(4262),
    [anon_sym_nullable] = ACTIONS(4262),
    [anon_sym__Complex] = ACTIONS(4262),
    [anon_sym__Nullable] = ACTIONS(4262),
    [anon_sym__Nullable_result] = ACTIONS(4262),
    [anon_sym__Null_unspecified] = ACTIONS(4262),
    [anon_sym___autoreleasing] = ACTIONS(4262),
    [anon_sym___block] = ACTIONS(4262),
    [anon_sym___bridge] = ACTIONS(4262),
    [anon_sym___bridge_retained] = ACTIONS(4262),
    [anon_sym___bridge_transfer] = ACTIONS(4262),
    [anon_sym___complex] = ACTIONS(4262),
    [anon_sym___const] = ACTIONS(4262),
    [anon_sym___imag] = ACTIONS(4262),
    [anon_sym___kindof] = ACTIONS(4262),
    [anon_sym___nonnull] = ACTIONS(4262),
    [anon_sym___nullable] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4262),
    [anon_sym___real] = ACTIONS(4262),
    [anon_sym___strong] = ACTIONS(4262),
    [anon_sym___unsafe_unretained] = ACTIONS(4262),
    [anon_sym___unused] = ACTIONS(4262),
    [anon_sym___weak] = ACTIONS(4262),
    [anon_sym_alignas] = ACTIONS(4262),
    [anon_sym__Alignas] = ACTIONS(4262),
    [anon_sym_COLON] = ACTIONS(4264),
    [anon_sym_QMARK] = ACTIONS(4264),
    [anon_sym_DASH_DASH] = ACTIONS(4264),
    [anon_sym_PLUS_PLUS] = ACTIONS(4264),
    [anon_sym_DOT] = ACTIONS(4262),
    [anon_sym_DASH_GT] = ACTIONS(4264),
    [anon_sym_AT] = ACTIONS(4264),
    [anon_sym_DQUOTE] = ACTIONS(4264),
    [anon_sym_L_DQUOTE] = ACTIONS(4264),
    [anon_sym_u_DQUOTE] = ACTIONS(4264),
    [anon_sym_U_DQUOTE] = ACTIONS(4264),
    [anon_sym_u8_DQUOTE] = ACTIONS(4264),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4262),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4262),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4262),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4262),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4262),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4262),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4262),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_API_AVAILABLE] = ACTIONS(4262),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_API_DEPRECATED] = ACTIONS(4262),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4262),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym___deprecated_msg] = ACTIONS(4262),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4262),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4262),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4262),
  },
  [STATE(2025)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1809),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2026)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1797),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5202),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2027)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4801),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(3019),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5204),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2028)] = {
    [sym_identifier] = ACTIONS(2588),
    [aux_sym_preproc_def_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2588),
    [sym_preproc_directive] = ACTIONS(2588),
    [anon_sym_DASH] = ACTIONS(2590),
    [anon_sym_PLUS] = ACTIONS(2590),
    [anon_sym___extension__] = ACTIONS(2588),
    [anon_sym_typedef] = ACTIONS(2588),
    [anon_sym_extern] = ACTIONS(2588),
    [anon_sym___attribute__] = ACTIONS(2588),
    [anon_sym___attribute] = ACTIONS(2588),
    [anon_sym_noreturn] = ACTIONS(2588),
    [anon_sym_LBRACK] = ACTIONS(2590),
    [anon_sym___declspec] = ACTIONS(2588),
    [anon_sym___cdecl] = ACTIONS(2588),
    [anon_sym___clrcall] = ACTIONS(2588),
    [anon_sym___stdcall] = ACTIONS(2588),
    [anon_sym___fastcall] = ACTIONS(2588),
    [anon_sym___thiscall] = ACTIONS(2588),
    [anon_sym___vectorcall] = ACTIONS(2588),
    [anon_sym_RBRACE] = ACTIONS(2590),
    [anon_sym_signed] = ACTIONS(2588),
    [anon_sym_unsigned] = ACTIONS(2588),
    [anon_sym_long] = ACTIONS(2588),
    [anon_sym_short] = ACTIONS(2588),
    [anon_sym_static] = ACTIONS(2588),
    [anon_sym_auto] = ACTIONS(2588),
    [anon_sym_register] = ACTIONS(2588),
    [anon_sym_inline] = ACTIONS(2588),
    [anon_sym___inline] = ACTIONS(2588),
    [anon_sym___inline__] = ACTIONS(2588),
    [anon_sym___forceinline] = ACTIONS(2588),
    [anon_sym_thread_local] = ACTIONS(2588),
    [anon_sym___thread] = ACTIONS(2588),
    [anon_sym_CG_EXTERN] = ACTIONS(2588),
    [anon_sym_CG_INLINE] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2588),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2588),
    [anon_sym_IBOutlet] = ACTIONS(2588),
    [anon_sym_IBInspectable] = ACTIONS(2588),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2588),
    [anon_sym_NS_INLINE] = ACTIONS(2588),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2588),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2588),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2588),
    [anon_sym_const] = ACTIONS(2588),
    [anon_sym_constexpr] = ACTIONS(2588),
    [anon_sym_volatile] = ACTIONS(2588),
    [anon_sym_restrict] = ACTIONS(2588),
    [anon_sym___restrict__] = ACTIONS(2588),
    [anon_sym__Atomic] = ACTIONS(2588),
    [anon_sym__Noreturn] = ACTIONS(2588),
    [anon_sym__Nonnull] = ACTIONS(2588),
    [anon_sym_nullable] = ACTIONS(2588),
    [anon_sym__Complex] = ACTIONS(2588),
    [anon_sym__Nullable] = ACTIONS(2588),
    [anon_sym__Nullable_result] = ACTIONS(2588),
    [anon_sym__Null_unspecified] = ACTIONS(2588),
    [anon_sym___autoreleasing] = ACTIONS(2588),
    [anon_sym___block] = ACTIONS(2588),
    [anon_sym___bridge] = ACTIONS(2588),
    [anon_sym___bridge_retained] = ACTIONS(2588),
    [anon_sym___bridge_transfer] = ACTIONS(2588),
    [anon_sym___complex] = ACTIONS(2588),
    [anon_sym___const] = ACTIONS(2588),
    [anon_sym___imag] = ACTIONS(2588),
    [anon_sym___kindof] = ACTIONS(2588),
    [anon_sym___nonnull] = ACTIONS(2588),
    [anon_sym___nullable] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2588),
    [anon_sym___real] = ACTIONS(2588),
    [anon_sym___strong] = ACTIONS(2588),
    [anon_sym___unsafe_unretained] = ACTIONS(2588),
    [anon_sym___unused] = ACTIONS(2588),
    [anon_sym___weak] = ACTIONS(2588),
    [anon_sym_alignas] = ACTIONS(2588),
    [anon_sym__Alignas] = ACTIONS(2588),
    [sym_primitive_type] = ACTIONS(2588),
    [anon_sym_enum] = ACTIONS(2588),
    [anon_sym_struct] = ACTIONS(2588),
    [anon_sym_union] = ACTIONS(2588),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2588),
    [anon_sym___typeof] = ACTIONS(2588),
    [anon_sym_typeof] = ACTIONS(2588),
    [aux_sym_preproc_undef_token1] = ACTIONS(2588),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2588),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2588),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2588),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_API_AVAILABLE] = ACTIONS(2588),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_API_DEPRECATED] = ACTIONS(2588),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym___deprecated_msg] = ACTIONS(2588),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2588),
    [anon_sym_ATend] = ACTIONS(2590),
    [anon_sym_AToptional] = ACTIONS(2590),
    [anon_sym_ATrequired] = ACTIONS(2590),
    [anon_sym_ATsynthesize] = ACTIONS(2590),
    [anon_sym_ATdynamic] = ACTIONS(2590),
    [anon_sym_ATproperty] = ACTIONS(2590),
    [anon_sym_BOOL] = ACTIONS(2588),
    [anon_sym_IMP] = ACTIONS(2588),
    [anon_sym_SEL] = ACTIONS(2588),
    [anon_sym_Class] = ACTIONS(2588),
    [anon_sym_id] = ACTIONS(2588),
    [anon_sym_ATdefs] = ACTIONS(2590),
  },
  [STATE(2029)] = {
    [sym_identifier] = ACTIONS(4154),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4157),
    [anon_sym_COMMA] = ACTIONS(4157),
    [anon_sym_RPAREN] = ACTIONS(4157),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4157),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4157),
    [anon_sym_LPAREN2] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4154),
    [anon_sym_PLUS] = ACTIONS(4154),
    [anon_sym_STAR] = ACTIONS(4157),
    [anon_sym_SLASH] = ACTIONS(4154),
    [anon_sym_PERCENT] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4154),
    [anon_sym_CARET] = ACTIONS(4157),
    [anon_sym_AMP] = ACTIONS(4154),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4154),
    [anon_sym_GT_EQ] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_LT] = ACTIONS(4154),
    [anon_sym_LT_LT] = ACTIONS(4157),
    [anon_sym_GT_GT] = ACTIONS(4157),
    [anon_sym_SEMI] = ACTIONS(4157),
    [anon_sym___extension__] = ACTIONS(4154),
    [anon_sym_extern] = ACTIONS(4154),
    [anon_sym___attribute__] = ACTIONS(4154),
    [anon_sym___attribute] = ACTIONS(4154),
    [anon_sym_noreturn] = ACTIONS(4154),
    [anon_sym_LBRACK] = ACTIONS(4157),
    [anon_sym_RBRACK] = ACTIONS(4157),
    [anon_sym___declspec] = ACTIONS(4154),
    [anon_sym_RBRACE] = ACTIONS(4157),
    [anon_sym_static] = ACTIONS(4154),
    [anon_sym_auto] = ACTIONS(4154),
    [anon_sym_register] = ACTIONS(4154),
    [anon_sym_inline] = ACTIONS(4154),
    [anon_sym___inline] = ACTIONS(4154),
    [anon_sym___inline__] = ACTIONS(4154),
    [anon_sym___forceinline] = ACTIONS(4154),
    [anon_sym_thread_local] = ACTIONS(4154),
    [anon_sym___thread] = ACTIONS(4154),
    [anon_sym_CG_EXTERN] = ACTIONS(4154),
    [anon_sym_CG_INLINE] = ACTIONS(4154),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4154),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4154),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4154),
    [anon_sym_IBOutlet] = ACTIONS(4154),
    [anon_sym_IBInspectable] = ACTIONS(4154),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4154),
    [anon_sym_NS_INLINE] = ACTIONS(4154),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4154),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4154),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4154),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4154),
    [anon_sym_const] = ACTIONS(4154),
    [anon_sym_constexpr] = ACTIONS(4154),
    [anon_sym_volatile] = ACTIONS(4154),
    [anon_sym_restrict] = ACTIONS(4154),
    [anon_sym___restrict__] = ACTIONS(4154),
    [anon_sym__Atomic] = ACTIONS(4154),
    [anon_sym__Noreturn] = ACTIONS(4154),
    [anon_sym__Nonnull] = ACTIONS(4154),
    [anon_sym_nullable] = ACTIONS(4154),
    [anon_sym__Complex] = ACTIONS(4154),
    [anon_sym__Nullable] = ACTIONS(4154),
    [anon_sym__Nullable_result] = ACTIONS(4154),
    [anon_sym__Null_unspecified] = ACTIONS(4154),
    [anon_sym___autoreleasing] = ACTIONS(4154),
    [anon_sym___block] = ACTIONS(4154),
    [anon_sym___bridge] = ACTIONS(4154),
    [anon_sym___bridge_retained] = ACTIONS(4154),
    [anon_sym___bridge_transfer] = ACTIONS(4154),
    [anon_sym___complex] = ACTIONS(4154),
    [anon_sym___const] = ACTIONS(4154),
    [anon_sym___imag] = ACTIONS(4154),
    [anon_sym___kindof] = ACTIONS(4154),
    [anon_sym___nonnull] = ACTIONS(4154),
    [anon_sym___nullable] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4154),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4154),
    [anon_sym___real] = ACTIONS(4154),
    [anon_sym___strong] = ACTIONS(4154),
    [anon_sym___unsafe_unretained] = ACTIONS(4154),
    [anon_sym___unused] = ACTIONS(4154),
    [anon_sym___weak] = ACTIONS(4154),
    [anon_sym_alignas] = ACTIONS(4154),
    [anon_sym__Alignas] = ACTIONS(4154),
    [anon_sym_COLON] = ACTIONS(4157),
    [anon_sym_QMARK] = ACTIONS(4157),
    [anon_sym_DASH_DASH] = ACTIONS(4157),
    [anon_sym_PLUS_PLUS] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4154),
    [anon_sym_DASH_GT] = ACTIONS(4157),
    [anon_sym_AT] = ACTIONS(4157),
    [anon_sym_DQUOTE] = ACTIONS(4157),
    [anon_sym_L_DQUOTE] = ACTIONS(4157),
    [anon_sym_u_DQUOTE] = ACTIONS(4157),
    [anon_sym_U_DQUOTE] = ACTIONS(4157),
    [anon_sym_u8_DQUOTE] = ACTIONS(4157),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4154),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4154),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4154),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4154),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4154),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4154),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4154),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_API_AVAILABLE] = ACTIONS(4154),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_API_DEPRECATED] = ACTIONS(4154),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4154),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4154),
    [anon_sym___deprecated_msg] = ACTIONS(4154),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4154),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4154),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4154),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4154),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4154),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4154),
  },
  [STATE(2030)] = {
    [sym_identifier] = ACTIONS(2852),
    [aux_sym_preproc_def_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2852),
    [sym_preproc_directive] = ACTIONS(2852),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym___extension__] = ACTIONS(2852),
    [anon_sym_typedef] = ACTIONS(2852),
    [anon_sym_extern] = ACTIONS(2852),
    [anon_sym___attribute__] = ACTIONS(2852),
    [anon_sym___attribute] = ACTIONS(2852),
    [anon_sym_noreturn] = ACTIONS(2852),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym___declspec] = ACTIONS(2852),
    [anon_sym___cdecl] = ACTIONS(2852),
    [anon_sym___clrcall] = ACTIONS(2852),
    [anon_sym___stdcall] = ACTIONS(2852),
    [anon_sym___fastcall] = ACTIONS(2852),
    [anon_sym___thiscall] = ACTIONS(2852),
    [anon_sym___vectorcall] = ACTIONS(2852),
    [anon_sym_RBRACE] = ACTIONS(2854),
    [anon_sym_signed] = ACTIONS(2852),
    [anon_sym_unsigned] = ACTIONS(2852),
    [anon_sym_long] = ACTIONS(2852),
    [anon_sym_short] = ACTIONS(2852),
    [anon_sym_static] = ACTIONS(2852),
    [anon_sym_auto] = ACTIONS(2852),
    [anon_sym_register] = ACTIONS(2852),
    [anon_sym_inline] = ACTIONS(2852),
    [anon_sym___inline] = ACTIONS(2852),
    [anon_sym___inline__] = ACTIONS(2852),
    [anon_sym___forceinline] = ACTIONS(2852),
    [anon_sym_thread_local] = ACTIONS(2852),
    [anon_sym___thread] = ACTIONS(2852),
    [anon_sym_CG_EXTERN] = ACTIONS(2852),
    [anon_sym_CG_INLINE] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2852),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2852),
    [anon_sym_IBOutlet] = ACTIONS(2852),
    [anon_sym_IBInspectable] = ACTIONS(2852),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2852),
    [anon_sym_NS_INLINE] = ACTIONS(2852),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2852),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2852),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2852),
    [anon_sym_const] = ACTIONS(2852),
    [anon_sym_constexpr] = ACTIONS(2852),
    [anon_sym_volatile] = ACTIONS(2852),
    [anon_sym_restrict] = ACTIONS(2852),
    [anon_sym___restrict__] = ACTIONS(2852),
    [anon_sym__Atomic] = ACTIONS(2852),
    [anon_sym__Noreturn] = ACTIONS(2852),
    [anon_sym__Nonnull] = ACTIONS(2852),
    [anon_sym_nullable] = ACTIONS(2852),
    [anon_sym__Complex] = ACTIONS(2852),
    [anon_sym__Nullable] = ACTIONS(2852),
    [anon_sym__Nullable_result] = ACTIONS(2852),
    [anon_sym__Null_unspecified] = ACTIONS(2852),
    [anon_sym___autoreleasing] = ACTIONS(2852),
    [anon_sym___block] = ACTIONS(2852),
    [anon_sym___bridge] = ACTIONS(2852),
    [anon_sym___bridge_retained] = ACTIONS(2852),
    [anon_sym___bridge_transfer] = ACTIONS(2852),
    [anon_sym___complex] = ACTIONS(2852),
    [anon_sym___const] = ACTIONS(2852),
    [anon_sym___imag] = ACTIONS(2852),
    [anon_sym___kindof] = ACTIONS(2852),
    [anon_sym___nonnull] = ACTIONS(2852),
    [anon_sym___nullable] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2852),
    [anon_sym___real] = ACTIONS(2852),
    [anon_sym___strong] = ACTIONS(2852),
    [anon_sym___unsafe_unretained] = ACTIONS(2852),
    [anon_sym___unused] = ACTIONS(2852),
    [anon_sym___weak] = ACTIONS(2852),
    [anon_sym_alignas] = ACTIONS(2852),
    [anon_sym__Alignas] = ACTIONS(2852),
    [sym_primitive_type] = ACTIONS(2852),
    [anon_sym_enum] = ACTIONS(2852),
    [anon_sym_struct] = ACTIONS(2852),
    [anon_sym_union] = ACTIONS(2852),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2852),
    [anon_sym___typeof] = ACTIONS(2852),
    [anon_sym_typeof] = ACTIONS(2852),
    [aux_sym_preproc_undef_token1] = ACTIONS(2852),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2852),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2852),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2852),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_API_AVAILABLE] = ACTIONS(2852),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_API_DEPRECATED] = ACTIONS(2852),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym___deprecated_msg] = ACTIONS(2852),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2852),
    [anon_sym_ATend] = ACTIONS(2854),
    [anon_sym_AToptional] = ACTIONS(2854),
    [anon_sym_ATrequired] = ACTIONS(2854),
    [anon_sym_ATsynthesize] = ACTIONS(2854),
    [anon_sym_ATdynamic] = ACTIONS(2854),
    [anon_sym_ATproperty] = ACTIONS(2854),
    [anon_sym_BOOL] = ACTIONS(2852),
    [anon_sym_IMP] = ACTIONS(2852),
    [anon_sym_SEL] = ACTIONS(2852),
    [anon_sym_Class] = ACTIONS(2852),
    [anon_sym_id] = ACTIONS(2852),
    [anon_sym_ATdefs] = ACTIONS(2854),
  },
  [STATE(2031)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4801),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(3041),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5206),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2032)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1812),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5208),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2033)] = {
    [sym_identifier] = ACTIONS(4258),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4260),
    [anon_sym_COMMA] = ACTIONS(4260),
    [anon_sym_RPAREN] = ACTIONS(4260),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4260),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4260),
    [anon_sym_LPAREN2] = ACTIONS(4260),
    [anon_sym_DASH] = ACTIONS(4258),
    [anon_sym_PLUS] = ACTIONS(4258),
    [anon_sym_STAR] = ACTIONS(4260),
    [anon_sym_SLASH] = ACTIONS(4258),
    [anon_sym_PERCENT] = ACTIONS(4260),
    [anon_sym_PIPE_PIPE] = ACTIONS(4260),
    [anon_sym_AMP_AMP] = ACTIONS(4260),
    [anon_sym_PIPE] = ACTIONS(4258),
    [anon_sym_CARET] = ACTIONS(4260),
    [anon_sym_AMP] = ACTIONS(4258),
    [anon_sym_EQ_EQ] = ACTIONS(4260),
    [anon_sym_BANG_EQ] = ACTIONS(4260),
    [anon_sym_GT] = ACTIONS(4258),
    [anon_sym_GT_EQ] = ACTIONS(4260),
    [anon_sym_LT_EQ] = ACTIONS(4260),
    [anon_sym_LT] = ACTIONS(4258),
    [anon_sym_LT_LT] = ACTIONS(4260),
    [anon_sym_GT_GT] = ACTIONS(4260),
    [anon_sym_SEMI] = ACTIONS(4260),
    [anon_sym___extension__] = ACTIONS(4258),
    [anon_sym_extern] = ACTIONS(4258),
    [anon_sym___attribute__] = ACTIONS(4258),
    [anon_sym___attribute] = ACTIONS(4258),
    [anon_sym_noreturn] = ACTIONS(4258),
    [anon_sym_LBRACK] = ACTIONS(4260),
    [anon_sym_RBRACK] = ACTIONS(4260),
    [anon_sym___declspec] = ACTIONS(4258),
    [anon_sym_RBRACE] = ACTIONS(4260),
    [anon_sym_static] = ACTIONS(4258),
    [anon_sym_auto] = ACTIONS(4258),
    [anon_sym_register] = ACTIONS(4258),
    [anon_sym_inline] = ACTIONS(4258),
    [anon_sym___inline] = ACTIONS(4258),
    [anon_sym___inline__] = ACTIONS(4258),
    [anon_sym___forceinline] = ACTIONS(4258),
    [anon_sym_thread_local] = ACTIONS(4258),
    [anon_sym___thread] = ACTIONS(4258),
    [anon_sym_CG_EXTERN] = ACTIONS(4258),
    [anon_sym_CG_INLINE] = ACTIONS(4258),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4258),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4258),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4258),
    [anon_sym_IBOutlet] = ACTIONS(4258),
    [anon_sym_IBInspectable] = ACTIONS(4258),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4258),
    [anon_sym_NS_INLINE] = ACTIONS(4258),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4258),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4258),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4258),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4258),
    [anon_sym_const] = ACTIONS(4258),
    [anon_sym_constexpr] = ACTIONS(4258),
    [anon_sym_volatile] = ACTIONS(4258),
    [anon_sym_restrict] = ACTIONS(4258),
    [anon_sym___restrict__] = ACTIONS(4258),
    [anon_sym__Atomic] = ACTIONS(4258),
    [anon_sym__Noreturn] = ACTIONS(4258),
    [anon_sym__Nonnull] = ACTIONS(4258),
    [anon_sym_nullable] = ACTIONS(4258),
    [anon_sym__Complex] = ACTIONS(4258),
    [anon_sym__Nullable] = ACTIONS(4258),
    [anon_sym__Nullable_result] = ACTIONS(4258),
    [anon_sym__Null_unspecified] = ACTIONS(4258),
    [anon_sym___autoreleasing] = ACTIONS(4258),
    [anon_sym___block] = ACTIONS(4258),
    [anon_sym___bridge] = ACTIONS(4258),
    [anon_sym___bridge_retained] = ACTIONS(4258),
    [anon_sym___bridge_transfer] = ACTIONS(4258),
    [anon_sym___complex] = ACTIONS(4258),
    [anon_sym___const] = ACTIONS(4258),
    [anon_sym___imag] = ACTIONS(4258),
    [anon_sym___kindof] = ACTIONS(4258),
    [anon_sym___nonnull] = ACTIONS(4258),
    [anon_sym___nullable] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4258),
    [anon_sym___real] = ACTIONS(4258),
    [anon_sym___strong] = ACTIONS(4258),
    [anon_sym___unsafe_unretained] = ACTIONS(4258),
    [anon_sym___unused] = ACTIONS(4258),
    [anon_sym___weak] = ACTIONS(4258),
    [anon_sym_alignas] = ACTIONS(4258),
    [anon_sym__Alignas] = ACTIONS(4258),
    [anon_sym_COLON] = ACTIONS(4260),
    [anon_sym_QMARK] = ACTIONS(4260),
    [anon_sym_DASH_DASH] = ACTIONS(4260),
    [anon_sym_PLUS_PLUS] = ACTIONS(4260),
    [anon_sym_DOT] = ACTIONS(4258),
    [anon_sym_DASH_GT] = ACTIONS(4260),
    [anon_sym_AT] = ACTIONS(4260),
    [anon_sym_DQUOTE] = ACTIONS(4260),
    [anon_sym_L_DQUOTE] = ACTIONS(4260),
    [anon_sym_u_DQUOTE] = ACTIONS(4260),
    [anon_sym_U_DQUOTE] = ACTIONS(4260),
    [anon_sym_u8_DQUOTE] = ACTIONS(4260),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4258),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4258),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4258),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4258),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4258),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4258),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4258),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_API_AVAILABLE] = ACTIONS(4258),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_API_DEPRECATED] = ACTIONS(4258),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4258),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym___deprecated_msg] = ACTIONS(4258),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4258),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4258),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4258),
  },
  [STATE(2034)] = {
    [sym_identifier] = ACTIONS(5210),
    [aux_sym_preproc_def_token1] = ACTIONS(5210),
    [aux_sym_preproc_if_token1] = ACTIONS(5210),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5210),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5210),
    [sym_preproc_directive] = ACTIONS(5210),
    [anon_sym_LPAREN2] = ACTIONS(5212),
    [anon_sym_DASH] = ACTIONS(5212),
    [anon_sym_PLUS] = ACTIONS(5212),
    [anon_sym_LT] = ACTIONS(5212),
    [anon_sym_SEMI] = ACTIONS(5214),
    [anon_sym___extension__] = ACTIONS(5210),
    [anon_sym_typedef] = ACTIONS(5210),
    [anon_sym_extern] = ACTIONS(5210),
    [anon_sym___attribute__] = ACTIONS(5210),
    [anon_sym___attribute] = ACTIONS(5210),
    [anon_sym_noreturn] = ACTIONS(5210),
    [anon_sym_LBRACK] = ACTIONS(5212),
    [anon_sym___declspec] = ACTIONS(5210),
    [anon_sym___cdecl] = ACTIONS(5210),
    [anon_sym___clrcall] = ACTIONS(5210),
    [anon_sym___stdcall] = ACTIONS(5210),
    [anon_sym___fastcall] = ACTIONS(5210),
    [anon_sym___thiscall] = ACTIONS(5210),
    [anon_sym___vectorcall] = ACTIONS(5210),
    [anon_sym_LBRACE] = ACTIONS(5212),
    [anon_sym_signed] = ACTIONS(5210),
    [anon_sym_unsigned] = ACTIONS(5210),
    [anon_sym_long] = ACTIONS(5210),
    [anon_sym_short] = ACTIONS(5210),
    [anon_sym_static] = ACTIONS(5210),
    [anon_sym_auto] = ACTIONS(5210),
    [anon_sym_register] = ACTIONS(5210),
    [anon_sym_inline] = ACTIONS(5210),
    [anon_sym___inline] = ACTIONS(5210),
    [anon_sym___inline__] = ACTIONS(5210),
    [anon_sym___forceinline] = ACTIONS(5210),
    [anon_sym_thread_local] = ACTIONS(5210),
    [anon_sym___thread] = ACTIONS(5210),
    [anon_sym_CG_EXTERN] = ACTIONS(5210),
    [anon_sym_CG_INLINE] = ACTIONS(5210),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5210),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5210),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5210),
    [anon_sym_IBOutlet] = ACTIONS(5210),
    [anon_sym_IBInspectable] = ACTIONS(5210),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5210),
    [anon_sym_NS_INLINE] = ACTIONS(5210),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5210),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5210),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5210),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5210),
    [anon_sym_const] = ACTIONS(5210),
    [anon_sym_constexpr] = ACTIONS(5210),
    [anon_sym_volatile] = ACTIONS(5210),
    [anon_sym_restrict] = ACTIONS(5210),
    [anon_sym___restrict__] = ACTIONS(5210),
    [anon_sym__Atomic] = ACTIONS(5210),
    [anon_sym__Noreturn] = ACTIONS(5210),
    [anon_sym__Nonnull] = ACTIONS(5210),
    [anon_sym_nullable] = ACTIONS(5210),
    [anon_sym__Complex] = ACTIONS(5210),
    [anon_sym__Nullable] = ACTIONS(5210),
    [anon_sym__Nullable_result] = ACTIONS(5210),
    [anon_sym__Null_unspecified] = ACTIONS(5210),
    [anon_sym___autoreleasing] = ACTIONS(5210),
    [anon_sym___block] = ACTIONS(5210),
    [anon_sym___bridge] = ACTIONS(5210),
    [anon_sym___bridge_retained] = ACTIONS(5210),
    [anon_sym___bridge_transfer] = ACTIONS(5210),
    [anon_sym___complex] = ACTIONS(5210),
    [anon_sym___const] = ACTIONS(5210),
    [anon_sym___imag] = ACTIONS(5210),
    [anon_sym___kindof] = ACTIONS(5210),
    [anon_sym___nonnull] = ACTIONS(5210),
    [anon_sym___nullable] = ACTIONS(5210),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5210),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5210),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5210),
    [anon_sym___real] = ACTIONS(5210),
    [anon_sym___strong] = ACTIONS(5210),
    [anon_sym___unsafe_unretained] = ACTIONS(5210),
    [anon_sym___unused] = ACTIONS(5210),
    [anon_sym___weak] = ACTIONS(5210),
    [anon_sym_alignas] = ACTIONS(5210),
    [anon_sym__Alignas] = ACTIONS(5210),
    [sym_primitive_type] = ACTIONS(5210),
    [anon_sym_enum] = ACTIONS(5210),
    [anon_sym_COLON] = ACTIONS(5212),
    [anon_sym_struct] = ACTIONS(5210),
    [anon_sym_union] = ACTIONS(5210),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5210),
    [anon_sym___typeof] = ACTIONS(5210),
    [anon_sym_typeof] = ACTIONS(5210),
    [aux_sym_preproc_undef_token1] = ACTIONS(5210),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5210),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5210),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5210),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5210),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5210),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5210),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5210),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5210),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5210),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5210),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5210),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5210),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5210),
    [anon_sym_API_AVAILABLE] = ACTIONS(5210),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5210),
    [anon_sym_API_DEPRECATED] = ACTIONS(5210),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5210),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5210),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5210),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5210),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5210),
    [anon_sym___deprecated_msg] = ACTIONS(5210),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5210),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5210),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5210),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5210),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5210),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5210),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5210),
    [anon_sym_ATend] = ACTIONS(5212),
    [anon_sym_ATsynthesize] = ACTIONS(5212),
    [anon_sym_ATdynamic] = ACTIONS(5212),
    [anon_sym_BOOL] = ACTIONS(5210),
    [anon_sym_IMP] = ACTIONS(5210),
    [anon_sym_SEL] = ACTIONS(5210),
    [anon_sym_Class] = ACTIONS(5210),
    [anon_sym_id] = ACTIONS(5210),
  },
  [STATE(2035)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1771),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5216),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2036)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1811),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5218),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2037)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4801),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(3036),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5220),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2038)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(5222),
    [anon_sym_ATinterface] = ACTIONS(5186),
    [anon_sym_ATimplementation] = ACTIONS(5188),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2039)] = {
    [sym_compound_statement] = STATE(5317),
    [sym_type_qualifier] = STATE(3320),
    [sym_alignas_qualifier] = STATE(4149),
    [sym_expression] = STATE(4407),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_objc_bridge] = STATE(5947),
    [sym_typeof_specifier] = STATE(5317),
    [sym_availability] = STATE(5947),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5080),
    [anon_sym_RPAREN] = ACTIONS(5082),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5084),
    [anon_sym_noreturn] = ACTIONS(5224),
    [anon_sym_nothrow] = ACTIONS(5226),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(5090),
    [anon_sym_constexpr] = ACTIONS(5092),
    [anon_sym_volatile] = ACTIONS(5092),
    [anon_sym_restrict] = ACTIONS(5092),
    [anon_sym___restrict__] = ACTIONS(5092),
    [anon_sym__Atomic] = ACTIONS(5092),
    [anon_sym__Noreturn] = ACTIONS(5092),
    [anon_sym__Nonnull] = ACTIONS(5092),
    [anon_sym_nullable] = ACTIONS(5092),
    [anon_sym__Complex] = ACTIONS(5092),
    [anon_sym__Nullable] = ACTIONS(5092),
    [anon_sym__Nullable_result] = ACTIONS(5092),
    [anon_sym__Null_unspecified] = ACTIONS(5092),
    [anon_sym___autoreleasing] = ACTIONS(5092),
    [anon_sym___block] = ACTIONS(5092),
    [anon_sym___bridge] = ACTIONS(5092),
    [anon_sym___bridge_retained] = ACTIONS(5092),
    [anon_sym___bridge_transfer] = ACTIONS(5092),
    [anon_sym___complex] = ACTIONS(5092),
    [anon_sym___const] = ACTIONS(5092),
    [anon_sym___imag] = ACTIONS(5094),
    [anon_sym___kindof] = ACTIONS(5092),
    [anon_sym___nonnull] = ACTIONS(5092),
    [anon_sym___nullable] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5092),
    [anon_sym___real] = ACTIONS(5094),
    [anon_sym___strong] = ACTIONS(5092),
    [anon_sym___unsafe_unretained] = ACTIONS(5092),
    [anon_sym___unused] = ACTIONS(5092),
    [anon_sym___weak] = ACTIONS(5092),
    [anon_sym_alignas] = ACTIONS(5096),
    [anon_sym__Alignas] = ACTIONS(5096),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_objc_bridge_related] = ACTIONS(5098),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_availability] = ACTIONS(5100),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2040)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(5228),
    [anon_sym_ATinterface] = ACTIONS(5186),
    [anon_sym_ATimplementation] = ACTIONS(5188),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2041)] = {
    [sym_compound_statement] = STATE(5317),
    [sym_type_qualifier] = STATE(3320),
    [sym_alignas_qualifier] = STATE(4149),
    [sym_expression] = STATE(4407),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_objc_bridge] = STATE(5947),
    [sym_typeof_specifier] = STATE(5317),
    [sym_availability] = STATE(5947),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5080),
    [anon_sym_RPAREN] = ACTIONS(5082),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5084),
    [anon_sym_noreturn] = ACTIONS(5230),
    [anon_sym_nothrow] = ACTIONS(5232),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(5090),
    [anon_sym_constexpr] = ACTIONS(5092),
    [anon_sym_volatile] = ACTIONS(5092),
    [anon_sym_restrict] = ACTIONS(5092),
    [anon_sym___restrict__] = ACTIONS(5092),
    [anon_sym__Atomic] = ACTIONS(5092),
    [anon_sym__Noreturn] = ACTIONS(5092),
    [anon_sym__Nonnull] = ACTIONS(5092),
    [anon_sym_nullable] = ACTIONS(5092),
    [anon_sym__Complex] = ACTIONS(5092),
    [anon_sym__Nullable] = ACTIONS(5092),
    [anon_sym__Nullable_result] = ACTIONS(5092),
    [anon_sym__Null_unspecified] = ACTIONS(5092),
    [anon_sym___autoreleasing] = ACTIONS(5092),
    [anon_sym___block] = ACTIONS(5092),
    [anon_sym___bridge] = ACTIONS(5092),
    [anon_sym___bridge_retained] = ACTIONS(5092),
    [anon_sym___bridge_transfer] = ACTIONS(5092),
    [anon_sym___complex] = ACTIONS(5092),
    [anon_sym___const] = ACTIONS(5092),
    [anon_sym___imag] = ACTIONS(5094),
    [anon_sym___kindof] = ACTIONS(5092),
    [anon_sym___nonnull] = ACTIONS(5092),
    [anon_sym___nullable] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5092),
    [anon_sym___real] = ACTIONS(5094),
    [anon_sym___strong] = ACTIONS(5092),
    [anon_sym___unsafe_unretained] = ACTIONS(5092),
    [anon_sym___unused] = ACTIONS(5092),
    [anon_sym___weak] = ACTIONS(5092),
    [anon_sym_alignas] = ACTIONS(5096),
    [anon_sym__Alignas] = ACTIONS(5096),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_objc_bridge_related] = ACTIONS(5098),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_availability] = ACTIONS(5100),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2042)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(5234),
    [anon_sym_ATinterface] = ACTIONS(5186),
    [anon_sym_ATimplementation] = ACTIONS(5188),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2043)] = {
    [sym_compound_statement] = STATE(5317),
    [sym_type_qualifier] = STATE(3320),
    [sym_alignas_qualifier] = STATE(4149),
    [sym_expression] = STATE(4407),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_objc_bridge] = STATE(5947),
    [sym_typeof_specifier] = STATE(5317),
    [sym_availability] = STATE(5947),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5080),
    [anon_sym_RPAREN] = ACTIONS(5082),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5084),
    [anon_sym_noreturn] = ACTIONS(5236),
    [anon_sym_nothrow] = ACTIONS(5238),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(5090),
    [anon_sym_constexpr] = ACTIONS(5092),
    [anon_sym_volatile] = ACTIONS(5092),
    [anon_sym_restrict] = ACTIONS(5092),
    [anon_sym___restrict__] = ACTIONS(5092),
    [anon_sym__Atomic] = ACTIONS(5092),
    [anon_sym__Noreturn] = ACTIONS(5092),
    [anon_sym__Nonnull] = ACTIONS(5092),
    [anon_sym_nullable] = ACTIONS(5092),
    [anon_sym__Complex] = ACTIONS(5092),
    [anon_sym__Nullable] = ACTIONS(5092),
    [anon_sym__Nullable_result] = ACTIONS(5092),
    [anon_sym__Null_unspecified] = ACTIONS(5092),
    [anon_sym___autoreleasing] = ACTIONS(5092),
    [anon_sym___block] = ACTIONS(5092),
    [anon_sym___bridge] = ACTIONS(5092),
    [anon_sym___bridge_retained] = ACTIONS(5092),
    [anon_sym___bridge_transfer] = ACTIONS(5092),
    [anon_sym___complex] = ACTIONS(5092),
    [anon_sym___const] = ACTIONS(5092),
    [anon_sym___imag] = ACTIONS(5094),
    [anon_sym___kindof] = ACTIONS(5092),
    [anon_sym___nonnull] = ACTIONS(5092),
    [anon_sym___nullable] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5092),
    [anon_sym___real] = ACTIONS(5094),
    [anon_sym___strong] = ACTIONS(5092),
    [anon_sym___unsafe_unretained] = ACTIONS(5092),
    [anon_sym___unused] = ACTIONS(5092),
    [anon_sym___weak] = ACTIONS(5092),
    [anon_sym_alignas] = ACTIONS(5096),
    [anon_sym__Alignas] = ACTIONS(5096),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_objc_bridge_related] = ACTIONS(5098),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_availability] = ACTIONS(5100),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2044)] = {
    [sym_compound_statement] = STATE(5317),
    [sym_type_qualifier] = STATE(3320),
    [sym_alignas_qualifier] = STATE(4149),
    [sym_expression] = STATE(4407),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_objc_bridge] = STATE(5947),
    [sym_typeof_specifier] = STATE(5317),
    [sym_availability] = STATE(5947),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5080),
    [anon_sym_RPAREN] = ACTIONS(5082),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5084),
    [anon_sym_noreturn] = ACTIONS(5240),
    [anon_sym_nothrow] = ACTIONS(5242),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(5090),
    [anon_sym_constexpr] = ACTIONS(5092),
    [anon_sym_volatile] = ACTIONS(5092),
    [anon_sym_restrict] = ACTIONS(5092),
    [anon_sym___restrict__] = ACTIONS(5092),
    [anon_sym__Atomic] = ACTIONS(5092),
    [anon_sym__Noreturn] = ACTIONS(5092),
    [anon_sym__Nonnull] = ACTIONS(5092),
    [anon_sym_nullable] = ACTIONS(5092),
    [anon_sym__Complex] = ACTIONS(5092),
    [anon_sym__Nullable] = ACTIONS(5092),
    [anon_sym__Nullable_result] = ACTIONS(5092),
    [anon_sym__Null_unspecified] = ACTIONS(5092),
    [anon_sym___autoreleasing] = ACTIONS(5092),
    [anon_sym___block] = ACTIONS(5092),
    [anon_sym___bridge] = ACTIONS(5092),
    [anon_sym___bridge_retained] = ACTIONS(5092),
    [anon_sym___bridge_transfer] = ACTIONS(5092),
    [anon_sym___complex] = ACTIONS(5092),
    [anon_sym___const] = ACTIONS(5092),
    [anon_sym___imag] = ACTIONS(5094),
    [anon_sym___kindof] = ACTIONS(5092),
    [anon_sym___nonnull] = ACTIONS(5092),
    [anon_sym___nullable] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5092),
    [anon_sym___real] = ACTIONS(5094),
    [anon_sym___strong] = ACTIONS(5092),
    [anon_sym___unsafe_unretained] = ACTIONS(5092),
    [anon_sym___unused] = ACTIONS(5092),
    [anon_sym___weak] = ACTIONS(5092),
    [anon_sym_alignas] = ACTIONS(5096),
    [anon_sym__Alignas] = ACTIONS(5096),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_objc_bridge_related] = ACTIONS(5098),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_availability] = ACTIONS(5100),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2045)] = {
    [sym_compound_statement] = STATE(5317),
    [sym_type_qualifier] = STATE(3320),
    [sym_alignas_qualifier] = STATE(4149),
    [sym_expression] = STATE(4407),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_objc_bridge] = STATE(5947),
    [sym_typeof_specifier] = STATE(5317),
    [sym_availability] = STATE(5947),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5080),
    [anon_sym_RPAREN] = ACTIONS(5082),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5084),
    [anon_sym_noreturn] = ACTIONS(5244),
    [anon_sym_nothrow] = ACTIONS(5246),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(5090),
    [anon_sym_constexpr] = ACTIONS(5092),
    [anon_sym_volatile] = ACTIONS(5092),
    [anon_sym_restrict] = ACTIONS(5092),
    [anon_sym___restrict__] = ACTIONS(5092),
    [anon_sym__Atomic] = ACTIONS(5092),
    [anon_sym__Noreturn] = ACTIONS(5092),
    [anon_sym__Nonnull] = ACTIONS(5092),
    [anon_sym_nullable] = ACTIONS(5092),
    [anon_sym__Complex] = ACTIONS(5092),
    [anon_sym__Nullable] = ACTIONS(5092),
    [anon_sym__Nullable_result] = ACTIONS(5092),
    [anon_sym__Null_unspecified] = ACTIONS(5092),
    [anon_sym___autoreleasing] = ACTIONS(5092),
    [anon_sym___block] = ACTIONS(5092),
    [anon_sym___bridge] = ACTIONS(5092),
    [anon_sym___bridge_retained] = ACTIONS(5092),
    [anon_sym___bridge_transfer] = ACTIONS(5092),
    [anon_sym___complex] = ACTIONS(5092),
    [anon_sym___const] = ACTIONS(5092),
    [anon_sym___imag] = ACTIONS(5094),
    [anon_sym___kindof] = ACTIONS(5092),
    [anon_sym___nonnull] = ACTIONS(5092),
    [anon_sym___nullable] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5092),
    [anon_sym___real] = ACTIONS(5094),
    [anon_sym___strong] = ACTIONS(5092),
    [anon_sym___unsafe_unretained] = ACTIONS(5092),
    [anon_sym___unused] = ACTIONS(5092),
    [anon_sym___weak] = ACTIONS(5092),
    [anon_sym_alignas] = ACTIONS(5096),
    [anon_sym__Alignas] = ACTIONS(5096),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_objc_bridge_related] = ACTIONS(5098),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_availability] = ACTIONS(5100),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2046)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1807),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5248),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2047)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1803),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5250),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2048)] = {
    [sym_declaration] = STATE(1880),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1880),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2049)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1811),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5252),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2050)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_ATprotocol] = ACTIONS(5254),
    [anon_sym_ATinterface] = ACTIONS(5186),
    [anon_sym_ATimplementation] = ACTIONS(5188),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2051)] = {
    [sym_declaration] = STATE(1883),
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4205),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__old_style_function_definition_repeat1] = STATE(1883),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2052)] = {
    [sym_identifier] = ACTIONS(5256),
    [aux_sym_preproc_def_token1] = ACTIONS(5256),
    [aux_sym_preproc_if_token1] = ACTIONS(5256),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5256),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5256),
    [sym_preproc_directive] = ACTIONS(5256),
    [anon_sym_LPAREN2] = ACTIONS(5258),
    [anon_sym_DASH] = ACTIONS(5258),
    [anon_sym_PLUS] = ACTIONS(5258),
    [anon_sym_LT] = ACTIONS(5258),
    [anon_sym___extension__] = ACTIONS(5256),
    [anon_sym_typedef] = ACTIONS(5256),
    [anon_sym_extern] = ACTIONS(5256),
    [anon_sym___attribute__] = ACTIONS(5256),
    [anon_sym___attribute] = ACTIONS(5256),
    [anon_sym_noreturn] = ACTIONS(5256),
    [anon_sym_LBRACK] = ACTIONS(5258),
    [anon_sym___declspec] = ACTIONS(5256),
    [anon_sym___cdecl] = ACTIONS(5256),
    [anon_sym___clrcall] = ACTIONS(5256),
    [anon_sym___stdcall] = ACTIONS(5256),
    [anon_sym___fastcall] = ACTIONS(5256),
    [anon_sym___thiscall] = ACTIONS(5256),
    [anon_sym___vectorcall] = ACTIONS(5256),
    [anon_sym_LBRACE] = ACTIONS(5258),
    [anon_sym_signed] = ACTIONS(5256),
    [anon_sym_unsigned] = ACTIONS(5256),
    [anon_sym_long] = ACTIONS(5256),
    [anon_sym_short] = ACTIONS(5256),
    [anon_sym_static] = ACTIONS(5256),
    [anon_sym_auto] = ACTIONS(5256),
    [anon_sym_register] = ACTIONS(5256),
    [anon_sym_inline] = ACTIONS(5256),
    [anon_sym___inline] = ACTIONS(5256),
    [anon_sym___inline__] = ACTIONS(5256),
    [anon_sym___forceinline] = ACTIONS(5256),
    [anon_sym_thread_local] = ACTIONS(5256),
    [anon_sym___thread] = ACTIONS(5256),
    [anon_sym_CG_EXTERN] = ACTIONS(5256),
    [anon_sym_CG_INLINE] = ACTIONS(5256),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5256),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5256),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5256),
    [anon_sym_IBOutlet] = ACTIONS(5256),
    [anon_sym_IBInspectable] = ACTIONS(5256),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5256),
    [anon_sym_NS_INLINE] = ACTIONS(5256),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5256),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5256),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5256),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5256),
    [anon_sym_const] = ACTIONS(5256),
    [anon_sym_constexpr] = ACTIONS(5256),
    [anon_sym_volatile] = ACTIONS(5256),
    [anon_sym_restrict] = ACTIONS(5256),
    [anon_sym___restrict__] = ACTIONS(5256),
    [anon_sym__Atomic] = ACTIONS(5256),
    [anon_sym__Noreturn] = ACTIONS(5256),
    [anon_sym__Nonnull] = ACTIONS(5256),
    [anon_sym_nullable] = ACTIONS(5256),
    [anon_sym__Complex] = ACTIONS(5256),
    [anon_sym__Nullable] = ACTIONS(5256),
    [anon_sym__Nullable_result] = ACTIONS(5256),
    [anon_sym__Null_unspecified] = ACTIONS(5256),
    [anon_sym___autoreleasing] = ACTIONS(5256),
    [anon_sym___block] = ACTIONS(5256),
    [anon_sym___bridge] = ACTIONS(5256),
    [anon_sym___bridge_retained] = ACTIONS(5256),
    [anon_sym___bridge_transfer] = ACTIONS(5256),
    [anon_sym___complex] = ACTIONS(5256),
    [anon_sym___const] = ACTIONS(5256),
    [anon_sym___imag] = ACTIONS(5256),
    [anon_sym___kindof] = ACTIONS(5256),
    [anon_sym___nonnull] = ACTIONS(5256),
    [anon_sym___nullable] = ACTIONS(5256),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5256),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5256),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5256),
    [anon_sym___real] = ACTIONS(5256),
    [anon_sym___strong] = ACTIONS(5256),
    [anon_sym___unsafe_unretained] = ACTIONS(5256),
    [anon_sym___unused] = ACTIONS(5256),
    [anon_sym___weak] = ACTIONS(5256),
    [anon_sym_alignas] = ACTIONS(5256),
    [anon_sym__Alignas] = ACTIONS(5256),
    [sym_primitive_type] = ACTIONS(5256),
    [anon_sym_enum] = ACTIONS(5256),
    [anon_sym_COLON] = ACTIONS(5258),
    [anon_sym_struct] = ACTIONS(5256),
    [anon_sym_union] = ACTIONS(5256),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5256),
    [anon_sym___typeof] = ACTIONS(5256),
    [anon_sym_typeof] = ACTIONS(5256),
    [aux_sym_preproc_undef_token1] = ACTIONS(5256),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5256),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5256),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5256),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5256),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5256),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5256),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5256),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5256),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5256),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5256),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5256),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5256),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5256),
    [anon_sym_API_AVAILABLE] = ACTIONS(5256),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5256),
    [anon_sym_API_DEPRECATED] = ACTIONS(5256),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5256),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5256),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5256),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5256),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5256),
    [anon_sym___deprecated_msg] = ACTIONS(5256),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5256),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5256),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5256),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5256),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5256),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5256),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5256),
    [anon_sym_ATend] = ACTIONS(5258),
    [anon_sym_ATsynthesize] = ACTIONS(5258),
    [anon_sym_ATdynamic] = ACTIONS(5258),
    [anon_sym_ATproperty] = ACTIONS(5258),
    [anon_sym_BOOL] = ACTIONS(5256),
    [anon_sym_IMP] = ACTIONS(5256),
    [anon_sym_SEL] = ACTIONS(5256),
    [anon_sym_Class] = ACTIONS(5256),
    [anon_sym_id] = ACTIONS(5256),
  },
  [STATE(2053)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4801),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(3018),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5260),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2054)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1791),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5262),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2055)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(1791),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5264),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2056)] = {
    [sym_identifier] = ACTIONS(4252),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4255),
    [anon_sym_COMMA] = ACTIONS(4255),
    [anon_sym_RPAREN] = ACTIONS(4255),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4255),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4255),
    [anon_sym_LPAREN2] = ACTIONS(4255),
    [anon_sym_DASH] = ACTIONS(4252),
    [anon_sym_PLUS] = ACTIONS(4252),
    [anon_sym_STAR] = ACTIONS(4255),
    [anon_sym_SLASH] = ACTIONS(4252),
    [anon_sym_PERCENT] = ACTIONS(4255),
    [anon_sym_PIPE_PIPE] = ACTIONS(4255),
    [anon_sym_AMP_AMP] = ACTIONS(4255),
    [anon_sym_PIPE] = ACTIONS(4252),
    [anon_sym_CARET] = ACTIONS(4255),
    [anon_sym_AMP] = ACTIONS(4252),
    [anon_sym_EQ_EQ] = ACTIONS(4255),
    [anon_sym_BANG_EQ] = ACTIONS(4255),
    [anon_sym_GT] = ACTIONS(4252),
    [anon_sym_GT_EQ] = ACTIONS(4255),
    [anon_sym_LT_EQ] = ACTIONS(4255),
    [anon_sym_LT] = ACTIONS(4252),
    [anon_sym_LT_LT] = ACTIONS(4255),
    [anon_sym_GT_GT] = ACTIONS(4255),
    [anon_sym_SEMI] = ACTIONS(4255),
    [anon_sym___extension__] = ACTIONS(4252),
    [anon_sym_extern] = ACTIONS(4252),
    [anon_sym___attribute__] = ACTIONS(4252),
    [anon_sym___attribute] = ACTIONS(4252),
    [anon_sym_noreturn] = ACTIONS(4252),
    [anon_sym_LBRACK] = ACTIONS(4255),
    [anon_sym_RBRACK] = ACTIONS(4255),
    [anon_sym___declspec] = ACTIONS(4252),
    [anon_sym_RBRACE] = ACTIONS(4255),
    [anon_sym_static] = ACTIONS(4252),
    [anon_sym_auto] = ACTIONS(4252),
    [anon_sym_register] = ACTIONS(4252),
    [anon_sym_inline] = ACTIONS(4252),
    [anon_sym___inline] = ACTIONS(4252),
    [anon_sym___inline__] = ACTIONS(4252),
    [anon_sym___forceinline] = ACTIONS(4252),
    [anon_sym_thread_local] = ACTIONS(4252),
    [anon_sym___thread] = ACTIONS(4252),
    [anon_sym_CG_EXTERN] = ACTIONS(4252),
    [anon_sym_CG_INLINE] = ACTIONS(4252),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4252),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4252),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4252),
    [anon_sym_IBOutlet] = ACTIONS(4252),
    [anon_sym_IBInspectable] = ACTIONS(4252),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4252),
    [anon_sym_NS_INLINE] = ACTIONS(4252),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4252),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4252),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4252),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4252),
    [anon_sym_const] = ACTIONS(4252),
    [anon_sym_constexpr] = ACTIONS(4252),
    [anon_sym_volatile] = ACTIONS(4252),
    [anon_sym_restrict] = ACTIONS(4252),
    [anon_sym___restrict__] = ACTIONS(4252),
    [anon_sym__Atomic] = ACTIONS(4252),
    [anon_sym__Noreturn] = ACTIONS(4252),
    [anon_sym__Nonnull] = ACTIONS(4252),
    [anon_sym_nullable] = ACTIONS(4252),
    [anon_sym__Complex] = ACTIONS(4252),
    [anon_sym__Nullable] = ACTIONS(4252),
    [anon_sym__Nullable_result] = ACTIONS(4252),
    [anon_sym__Null_unspecified] = ACTIONS(4252),
    [anon_sym___autoreleasing] = ACTIONS(4252),
    [anon_sym___block] = ACTIONS(4252),
    [anon_sym___bridge] = ACTIONS(4252),
    [anon_sym___bridge_retained] = ACTIONS(4252),
    [anon_sym___bridge_transfer] = ACTIONS(4252),
    [anon_sym___complex] = ACTIONS(4252),
    [anon_sym___const] = ACTIONS(4252),
    [anon_sym___imag] = ACTIONS(4252),
    [anon_sym___kindof] = ACTIONS(4252),
    [anon_sym___nonnull] = ACTIONS(4252),
    [anon_sym___nullable] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4252),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4252),
    [anon_sym___real] = ACTIONS(4252),
    [anon_sym___strong] = ACTIONS(4252),
    [anon_sym___unsafe_unretained] = ACTIONS(4252),
    [anon_sym___unused] = ACTIONS(4252),
    [anon_sym___weak] = ACTIONS(4252),
    [anon_sym_alignas] = ACTIONS(4252),
    [anon_sym__Alignas] = ACTIONS(4252),
    [anon_sym_COLON] = ACTIONS(4255),
    [anon_sym_QMARK] = ACTIONS(4255),
    [anon_sym_DASH_DASH] = ACTIONS(4255),
    [anon_sym_PLUS_PLUS] = ACTIONS(4255),
    [anon_sym_DOT] = ACTIONS(4252),
    [anon_sym_DASH_GT] = ACTIONS(4255),
    [anon_sym_AT] = ACTIONS(4255),
    [anon_sym_DQUOTE] = ACTIONS(4255),
    [anon_sym_L_DQUOTE] = ACTIONS(4255),
    [anon_sym_u_DQUOTE] = ACTIONS(4255),
    [anon_sym_U_DQUOTE] = ACTIONS(4255),
    [anon_sym_u8_DQUOTE] = ACTIONS(4255),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4252),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4252),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4252),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4252),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4252),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4252),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4252),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_API_AVAILABLE] = ACTIONS(4252),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_API_DEPRECATED] = ACTIONS(4252),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4252),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4252),
    [anon_sym___deprecated_msg] = ACTIONS(4252),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4252),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4252),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4252),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4252),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4252),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4252),
  },
  [STATE(2057)] = {
    [sym_identifier] = ACTIONS(2608),
    [aux_sym_preproc_def_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2608),
    [sym_preproc_directive] = ACTIONS(2608),
    [anon_sym_DASH] = ACTIONS(2610),
    [anon_sym_PLUS] = ACTIONS(2610),
    [anon_sym___extension__] = ACTIONS(2608),
    [anon_sym_typedef] = ACTIONS(2608),
    [anon_sym_extern] = ACTIONS(2608),
    [anon_sym___attribute__] = ACTIONS(2608),
    [anon_sym___attribute] = ACTIONS(2608),
    [anon_sym_noreturn] = ACTIONS(2608),
    [anon_sym_LBRACK] = ACTIONS(2610),
    [anon_sym___declspec] = ACTIONS(2608),
    [anon_sym___cdecl] = ACTIONS(2608),
    [anon_sym___clrcall] = ACTIONS(2608),
    [anon_sym___stdcall] = ACTIONS(2608),
    [anon_sym___fastcall] = ACTIONS(2608),
    [anon_sym___thiscall] = ACTIONS(2608),
    [anon_sym___vectorcall] = ACTIONS(2608),
    [anon_sym_RBRACE] = ACTIONS(2610),
    [anon_sym_signed] = ACTIONS(2608),
    [anon_sym_unsigned] = ACTIONS(2608),
    [anon_sym_long] = ACTIONS(2608),
    [anon_sym_short] = ACTIONS(2608),
    [anon_sym_static] = ACTIONS(2608),
    [anon_sym_auto] = ACTIONS(2608),
    [anon_sym_register] = ACTIONS(2608),
    [anon_sym_inline] = ACTIONS(2608),
    [anon_sym___inline] = ACTIONS(2608),
    [anon_sym___inline__] = ACTIONS(2608),
    [anon_sym___forceinline] = ACTIONS(2608),
    [anon_sym_thread_local] = ACTIONS(2608),
    [anon_sym___thread] = ACTIONS(2608),
    [anon_sym_CG_EXTERN] = ACTIONS(2608),
    [anon_sym_CG_INLINE] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2608),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2608),
    [anon_sym_IBOutlet] = ACTIONS(2608),
    [anon_sym_IBInspectable] = ACTIONS(2608),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2608),
    [anon_sym_NS_INLINE] = ACTIONS(2608),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2608),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2608),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2608),
    [anon_sym_const] = ACTIONS(2608),
    [anon_sym_constexpr] = ACTIONS(2608),
    [anon_sym_volatile] = ACTIONS(2608),
    [anon_sym_restrict] = ACTIONS(2608),
    [anon_sym___restrict__] = ACTIONS(2608),
    [anon_sym__Atomic] = ACTIONS(2608),
    [anon_sym__Noreturn] = ACTIONS(2608),
    [anon_sym__Nonnull] = ACTIONS(2608),
    [anon_sym_nullable] = ACTIONS(2608),
    [anon_sym__Complex] = ACTIONS(2608),
    [anon_sym__Nullable] = ACTIONS(2608),
    [anon_sym__Nullable_result] = ACTIONS(2608),
    [anon_sym__Null_unspecified] = ACTIONS(2608),
    [anon_sym___autoreleasing] = ACTIONS(2608),
    [anon_sym___block] = ACTIONS(2608),
    [anon_sym___bridge] = ACTIONS(2608),
    [anon_sym___bridge_retained] = ACTIONS(2608),
    [anon_sym___bridge_transfer] = ACTIONS(2608),
    [anon_sym___complex] = ACTIONS(2608),
    [anon_sym___const] = ACTIONS(2608),
    [anon_sym___imag] = ACTIONS(2608),
    [anon_sym___kindof] = ACTIONS(2608),
    [anon_sym___nonnull] = ACTIONS(2608),
    [anon_sym___nullable] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2608),
    [anon_sym___real] = ACTIONS(2608),
    [anon_sym___strong] = ACTIONS(2608),
    [anon_sym___unsafe_unretained] = ACTIONS(2608),
    [anon_sym___unused] = ACTIONS(2608),
    [anon_sym___weak] = ACTIONS(2608),
    [anon_sym_alignas] = ACTIONS(2608),
    [anon_sym__Alignas] = ACTIONS(2608),
    [sym_primitive_type] = ACTIONS(2608),
    [anon_sym_enum] = ACTIONS(2608),
    [anon_sym_struct] = ACTIONS(2608),
    [anon_sym_union] = ACTIONS(2608),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2608),
    [anon_sym___typeof] = ACTIONS(2608),
    [anon_sym_typeof] = ACTIONS(2608),
    [aux_sym_preproc_undef_token1] = ACTIONS(2608),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2608),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2608),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2608),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_API_AVAILABLE] = ACTIONS(2608),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_API_DEPRECATED] = ACTIONS(2608),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym___deprecated_msg] = ACTIONS(2608),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2608),
    [anon_sym_ATend] = ACTIONS(2610),
    [anon_sym_AToptional] = ACTIONS(2610),
    [anon_sym_ATrequired] = ACTIONS(2610),
    [anon_sym_ATsynthesize] = ACTIONS(2610),
    [anon_sym_ATdynamic] = ACTIONS(2610),
    [anon_sym_ATproperty] = ACTIONS(2610),
    [anon_sym_BOOL] = ACTIONS(2608),
    [anon_sym_IMP] = ACTIONS(2608),
    [anon_sym_SEL] = ACTIONS(2608),
    [anon_sym_Class] = ACTIONS(2608),
    [anon_sym_id] = ACTIONS(2608),
    [anon_sym_ATdefs] = ACTIONS(2610),
  },
  [STATE(2058)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5266),
    [anon_sym_COMMA] = ACTIONS(4462),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(5272),
    [anon_sym_AMP_AMP] = ACTIONS(5274),
    [anon_sym_PIPE] = ACTIONS(5276),
    [anon_sym_CARET] = ACTIONS(5278),
    [anon_sym_AMP] = ACTIONS(5280),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4462),
    [anon_sym___extension__] = ACTIONS(4462),
    [anon_sym_extern] = ACTIONS(4462),
    [anon_sym___attribute__] = ACTIONS(4462),
    [anon_sym___attribute] = ACTIONS(4458),
    [anon_sym_noreturn] = ACTIONS(4462),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4462),
    [anon_sym_static] = ACTIONS(4462),
    [anon_sym_EQ] = ACTIONS(4458),
    [anon_sym_auto] = ACTIONS(4462),
    [anon_sym_register] = ACTIONS(4462),
    [anon_sym_inline] = ACTIONS(4462),
    [anon_sym___inline] = ACTIONS(4458),
    [anon_sym___inline__] = ACTIONS(4462),
    [anon_sym___forceinline] = ACTIONS(4462),
    [anon_sym_thread_local] = ACTIONS(4462),
    [anon_sym___thread] = ACTIONS(4462),
    [anon_sym_CG_EXTERN] = ACTIONS(4462),
    [anon_sym_CG_INLINE] = ACTIONS(4462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4462),
    [anon_sym_IBOutlet] = ACTIONS(4462),
    [anon_sym_IBInspectable] = ACTIONS(4462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4462),
    [anon_sym_NS_INLINE] = ACTIONS(4462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4462),
    [anon_sym_const] = ACTIONS(4458),
    [anon_sym_constexpr] = ACTIONS(4462),
    [anon_sym_volatile] = ACTIONS(4462),
    [anon_sym_restrict] = ACTIONS(4462),
    [anon_sym___restrict__] = ACTIONS(4462),
    [anon_sym__Atomic] = ACTIONS(4462),
    [anon_sym__Noreturn] = ACTIONS(4462),
    [anon_sym__Nonnull] = ACTIONS(4462),
    [anon_sym_nullable] = ACTIONS(4462),
    [anon_sym__Complex] = ACTIONS(4462),
    [anon_sym__Nullable] = ACTIONS(4458),
    [anon_sym__Nullable_result] = ACTIONS(4462),
    [anon_sym__Null_unspecified] = ACTIONS(4462),
    [anon_sym___autoreleasing] = ACTIONS(4462),
    [anon_sym___block] = ACTIONS(4462),
    [anon_sym___bridge] = ACTIONS(4458),
    [anon_sym___bridge_retained] = ACTIONS(4462),
    [anon_sym___bridge_transfer] = ACTIONS(4462),
    [anon_sym___complex] = ACTIONS(4462),
    [anon_sym___const] = ACTIONS(4462),
    [anon_sym___imag] = ACTIONS(4462),
    [anon_sym___kindof] = ACTIONS(4462),
    [anon_sym___nonnull] = ACTIONS(4462),
    [anon_sym___nullable] = ACTIONS(4462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4462),
    [anon_sym___real] = ACTIONS(4462),
    [anon_sym___strong] = ACTIONS(4462),
    [anon_sym___unsafe_unretained] = ACTIONS(4462),
    [anon_sym___unused] = ACTIONS(4462),
    [anon_sym___weak] = ACTIONS(4462),
    [anon_sym_alignas] = ACTIONS(4462),
    [anon_sym__Alignas] = ACTIONS(4462),
    [anon_sym_QMARK] = ACTIONS(5290),
    [anon_sym_STAR_EQ] = ACTIONS(4462),
    [anon_sym_SLASH_EQ] = ACTIONS(4462),
    [anon_sym_PERCENT_EQ] = ACTIONS(4462),
    [anon_sym_PLUS_EQ] = ACTIONS(4462),
    [anon_sym_DASH_EQ] = ACTIONS(4462),
    [anon_sym_LT_LT_EQ] = ACTIONS(4462),
    [anon_sym_GT_GT_EQ] = ACTIONS(4462),
    [anon_sym_AMP_EQ] = ACTIONS(4462),
    [anon_sym_CARET_EQ] = ACTIONS(4462),
    [anon_sym_PIPE_EQ] = ACTIONS(4462),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4462),
    [anon_sym_API_AVAILABLE] = ACTIONS(4462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4462),
    [anon_sym_API_DEPRECATED] = ACTIONS(4462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4462),
    [anon_sym___deprecated_msg] = ACTIONS(4462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4462),
  },
  [STATE(2059)] = {
    [sym_identifier] = ACTIONS(2712),
    [anon_sym_COMMA] = ACTIONS(2714),
    [anon_sym_RPAREN] = ACTIONS(2714),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2714),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2714),
    [anon_sym_LPAREN2] = ACTIONS(2714),
    [anon_sym_STAR] = ACTIONS(2714),
    [anon_sym_CARET] = ACTIONS(2714),
    [anon_sym_GT] = ACTIONS(2714),
    [anon_sym_SEMI] = ACTIONS(2714),
    [anon_sym___extension__] = ACTIONS(2712),
    [anon_sym_extern] = ACTIONS(2712),
    [anon_sym___attribute__] = ACTIONS(2712),
    [anon_sym___attribute] = ACTIONS(2712),
    [anon_sym_noreturn] = ACTIONS(2712),
    [anon_sym_LBRACK] = ACTIONS(2714),
    [anon_sym___declspec] = ACTIONS(2712),
    [anon_sym___based] = ACTIONS(2712),
    [anon_sym_LBRACE] = ACTIONS(2714),
    [anon_sym_RBRACE] = ACTIONS(2714),
    [anon_sym_signed] = ACTIONS(2712),
    [anon_sym_unsigned] = ACTIONS(2712),
    [anon_sym_long] = ACTIONS(2712),
    [anon_sym_short] = ACTIONS(2712),
    [anon_sym_static] = ACTIONS(2712),
    [anon_sym_EQ] = ACTIONS(2714),
    [anon_sym_ATautoreleasepool] = ACTIONS(2714),
    [anon_sym_auto] = ACTIONS(2712),
    [anon_sym_register] = ACTIONS(2712),
    [anon_sym_inline] = ACTIONS(2712),
    [anon_sym___inline] = ACTIONS(2712),
    [anon_sym___inline__] = ACTIONS(2712),
    [anon_sym___forceinline] = ACTIONS(2712),
    [anon_sym_thread_local] = ACTIONS(2712),
    [anon_sym___thread] = ACTIONS(2712),
    [anon_sym_CG_EXTERN] = ACTIONS(2712),
    [anon_sym_CG_INLINE] = ACTIONS(2712),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2712),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2712),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2712),
    [anon_sym_IBOutlet] = ACTIONS(2712),
    [anon_sym_IBInspectable] = ACTIONS(2712),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2712),
    [anon_sym_NS_INLINE] = ACTIONS(2712),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2712),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2712),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2712),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2712),
    [anon_sym_const] = ACTIONS(2712),
    [anon_sym_constexpr] = ACTIONS(2712),
    [anon_sym_volatile] = ACTIONS(2712),
    [anon_sym_restrict] = ACTIONS(2712),
    [anon_sym___restrict__] = ACTIONS(2712),
    [anon_sym__Atomic] = ACTIONS(2712),
    [anon_sym__Noreturn] = ACTIONS(2712),
    [anon_sym__Nonnull] = ACTIONS(2712),
    [anon_sym_nullable] = ACTIONS(2712),
    [anon_sym__Complex] = ACTIONS(2712),
    [anon_sym__Nullable] = ACTIONS(2712),
    [anon_sym__Nullable_result] = ACTIONS(2712),
    [anon_sym__Null_unspecified] = ACTIONS(2712),
    [anon_sym___autoreleasing] = ACTIONS(2712),
    [anon_sym___block] = ACTIONS(2712),
    [anon_sym___bridge] = ACTIONS(2712),
    [anon_sym___bridge_retained] = ACTIONS(2712),
    [anon_sym___bridge_transfer] = ACTIONS(2712),
    [anon_sym___complex] = ACTIONS(2712),
    [anon_sym___const] = ACTIONS(2712),
    [anon_sym___imag] = ACTIONS(2712),
    [anon_sym___kindof] = ACTIONS(2712),
    [anon_sym___nonnull] = ACTIONS(2712),
    [anon_sym___nullable] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2712),
    [anon_sym___real] = ACTIONS(2712),
    [anon_sym___strong] = ACTIONS(2712),
    [anon_sym___unsafe_unretained] = ACTIONS(2712),
    [anon_sym___unused] = ACTIONS(2712),
    [anon_sym___weak] = ACTIONS(2712),
    [anon_sym_alignas] = ACTIONS(2712),
    [anon_sym__Alignas] = ACTIONS(2712),
    [sym_primitive_type] = ACTIONS(2712),
    [anon_sym_enum] = ACTIONS(2712),
    [anon_sym_COLON] = ACTIONS(2714),
    [anon_sym_struct] = ACTIONS(2712),
    [anon_sym_union] = ACTIONS(2712),
    [anon_sym_asm] = ACTIONS(2712),
    [anon_sym___asm__] = ACTIONS(2712),
    [anon_sym___asm] = ACTIONS(2712),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2712),
    [anon_sym___typeof] = ACTIONS(2712),
    [anon_sym_typeof] = ACTIONS(2712),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2712),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2712),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2712),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2712),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2712),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2712),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2712),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_API_AVAILABLE] = ACTIONS(2712),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_API_DEPRECATED] = ACTIONS(2712),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2712),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2712),
    [anon_sym___deprecated_msg] = ACTIONS(2712),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2712),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2712),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2712),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2712),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2712),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2712),
    [anon_sym_ATprotocol] = ACTIONS(2714),
    [anon_sym_ATinterface] = ACTIONS(2714),
    [anon_sym_ATimplementation] = ACTIONS(2714),
    [anon_sym_ATproperty] = ACTIONS(2714),
    [anon_sym_BOOL] = ACTIONS(2712),
    [anon_sym_IMP] = ACTIONS(2712),
    [anon_sym_SEL] = ACTIONS(2712),
    [anon_sym_Class] = ACTIONS(2712),
    [anon_sym_id] = ACTIONS(2712),
  },
  [STATE(2060)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token2] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [aux_sym_preproc_else_token1] = ACTIONS(2346),
    [aux_sym_preproc_elif_token1] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_DASH] = ACTIONS(2348),
    [anon_sym_PLUS] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATsynthesize] = ACTIONS(2348),
    [anon_sym_ATdynamic] = ACTIONS(2348),
    [anon_sym_ATproperty] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(2061)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token2] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [aux_sym_preproc_else_token1] = ACTIONS(2350),
    [aux_sym_preproc_elif_token1] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_DASH] = ACTIONS(2352),
    [anon_sym_PLUS] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATsynthesize] = ACTIONS(2352),
    [anon_sym_ATdynamic] = ACTIONS(2352),
    [anon_sym_ATproperty] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(2062)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token2] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [aux_sym_preproc_else_token1] = ACTIONS(2360),
    [aux_sym_preproc_elif_token1] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_DASH] = ACTIONS(2362),
    [anon_sym_PLUS] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATsynthesize] = ACTIONS(2362),
    [anon_sym_ATdynamic] = ACTIONS(2362),
    [anon_sym_ATproperty] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(2063)] = {
    [sym_identifier] = ACTIONS(4926),
    [aux_sym_preproc_def_token1] = ACTIONS(4926),
    [aux_sym_preproc_if_token1] = ACTIONS(4926),
    [aux_sym_preproc_if_token2] = ACTIONS(4926),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4926),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4926),
    [sym_preproc_directive] = ACTIONS(4926),
    [anon_sym_LPAREN2] = ACTIONS(4928),
    [anon_sym_DASH] = ACTIONS(4928),
    [anon_sym_PLUS] = ACTIONS(4928),
    [anon_sym_STAR] = ACTIONS(4928),
    [anon_sym_CARET] = ACTIONS(4928),
    [anon_sym___extension__] = ACTIONS(4926),
    [anon_sym_typedef] = ACTIONS(4926),
    [anon_sym_extern] = ACTIONS(4926),
    [anon_sym___attribute__] = ACTIONS(4926),
    [anon_sym___attribute] = ACTIONS(4926),
    [anon_sym_noreturn] = ACTIONS(4926),
    [anon_sym_LBRACK] = ACTIONS(4928),
    [anon_sym___declspec] = ACTIONS(4926),
    [anon_sym___based] = ACTIONS(4926),
    [anon_sym___cdecl] = ACTIONS(4926),
    [anon_sym___clrcall] = ACTIONS(4926),
    [anon_sym___stdcall] = ACTIONS(4926),
    [anon_sym___fastcall] = ACTIONS(4926),
    [anon_sym___thiscall] = ACTIONS(4926),
    [anon_sym___vectorcall] = ACTIONS(4926),
    [anon_sym_signed] = ACTIONS(4926),
    [anon_sym_unsigned] = ACTIONS(4926),
    [anon_sym_long] = ACTIONS(4926),
    [anon_sym_short] = ACTIONS(4926),
    [anon_sym_static] = ACTIONS(4926),
    [anon_sym_auto] = ACTIONS(4926),
    [anon_sym_register] = ACTIONS(4926),
    [anon_sym_inline] = ACTIONS(4926),
    [anon_sym___inline] = ACTIONS(4926),
    [anon_sym___inline__] = ACTIONS(4926),
    [anon_sym___forceinline] = ACTIONS(4926),
    [anon_sym_thread_local] = ACTIONS(4926),
    [anon_sym___thread] = ACTIONS(4926),
    [anon_sym_CG_EXTERN] = ACTIONS(4926),
    [anon_sym_CG_INLINE] = ACTIONS(4926),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4926),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4926),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4926),
    [anon_sym_IBOutlet] = ACTIONS(4926),
    [anon_sym_IBInspectable] = ACTIONS(4926),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4926),
    [anon_sym_NS_INLINE] = ACTIONS(4926),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4926),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4926),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4926),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4926),
    [anon_sym_const] = ACTIONS(4926),
    [anon_sym_constexpr] = ACTIONS(4926),
    [anon_sym_volatile] = ACTIONS(4926),
    [anon_sym_restrict] = ACTIONS(4926),
    [anon_sym___restrict__] = ACTIONS(4926),
    [anon_sym__Atomic] = ACTIONS(4926),
    [anon_sym__Noreturn] = ACTIONS(4926),
    [anon_sym__Nonnull] = ACTIONS(4926),
    [anon_sym_nullable] = ACTIONS(4926),
    [anon_sym__Complex] = ACTIONS(4926),
    [anon_sym__Nullable] = ACTIONS(4926),
    [anon_sym__Nullable_result] = ACTIONS(4926),
    [anon_sym__Null_unspecified] = ACTIONS(4926),
    [anon_sym___autoreleasing] = ACTIONS(4926),
    [anon_sym___block] = ACTIONS(4926),
    [anon_sym___bridge] = ACTIONS(4926),
    [anon_sym___bridge_retained] = ACTIONS(4926),
    [anon_sym___bridge_transfer] = ACTIONS(4926),
    [anon_sym___complex] = ACTIONS(4926),
    [anon_sym___const] = ACTIONS(4926),
    [anon_sym___imag] = ACTIONS(4926),
    [anon_sym___kindof] = ACTIONS(4926),
    [anon_sym___nonnull] = ACTIONS(4926),
    [anon_sym___nullable] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4926),
    [anon_sym___real] = ACTIONS(4926),
    [anon_sym___strong] = ACTIONS(4926),
    [anon_sym___unsafe_unretained] = ACTIONS(4926),
    [anon_sym___unused] = ACTIONS(4926),
    [anon_sym___weak] = ACTIONS(4926),
    [anon_sym_alignas] = ACTIONS(4926),
    [anon_sym__Alignas] = ACTIONS(4926),
    [sym_primitive_type] = ACTIONS(4926),
    [anon_sym_enum] = ACTIONS(4926),
    [anon_sym_struct] = ACTIONS(4926),
    [anon_sym_union] = ACTIONS(4926),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4926),
    [anon_sym___typeof] = ACTIONS(4926),
    [anon_sym_typeof] = ACTIONS(4926),
    [aux_sym_preproc_undef_token1] = ACTIONS(4926),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4926),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4926),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4926),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4926),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4926),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4926),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4926),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4926),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4926),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_API_AVAILABLE] = ACTIONS(4926),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_API_DEPRECATED] = ACTIONS(4926),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4926),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4926),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4926),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4926),
    [anon_sym___deprecated_msg] = ACTIONS(4926),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4926),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4926),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4926),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4926),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4926),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4926),
    [anon_sym_ATsynthesize] = ACTIONS(4928),
    [anon_sym_ATdynamic] = ACTIONS(4928),
    [anon_sym_BOOL] = ACTIONS(4926),
    [anon_sym_IMP] = ACTIONS(4926),
    [anon_sym_SEL] = ACTIONS(4926),
    [anon_sym_Class] = ACTIONS(4926),
    [anon_sym_id] = ACTIONS(4926),
  },
  [STATE(2064)] = {
    [sym_identifier] = ACTIONS(5292),
    [aux_sym_preproc_def_token1] = ACTIONS(5292),
    [aux_sym_preproc_if_token1] = ACTIONS(5292),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5292),
    [sym_preproc_directive] = ACTIONS(5292),
    [anon_sym_LPAREN2] = ACTIONS(5294),
    [anon_sym_DASH] = ACTIONS(5294),
    [anon_sym_PLUS] = ACTIONS(5294),
    [anon_sym_LT] = ACTIONS(5294),
    [anon_sym_SEMI] = ACTIONS(5296),
    [anon_sym___extension__] = ACTIONS(5292),
    [anon_sym_typedef] = ACTIONS(5292),
    [anon_sym_extern] = ACTIONS(5292),
    [anon_sym___attribute__] = ACTIONS(5292),
    [anon_sym___attribute] = ACTIONS(5292),
    [anon_sym_noreturn] = ACTIONS(5292),
    [anon_sym_LBRACK] = ACTIONS(5294),
    [anon_sym___declspec] = ACTIONS(5292),
    [anon_sym___cdecl] = ACTIONS(5292),
    [anon_sym___clrcall] = ACTIONS(5292),
    [anon_sym___stdcall] = ACTIONS(5292),
    [anon_sym___fastcall] = ACTIONS(5292),
    [anon_sym___thiscall] = ACTIONS(5292),
    [anon_sym___vectorcall] = ACTIONS(5292),
    [anon_sym_LBRACE] = ACTIONS(5294),
    [anon_sym_signed] = ACTIONS(5292),
    [anon_sym_unsigned] = ACTIONS(5292),
    [anon_sym_long] = ACTIONS(5292),
    [anon_sym_short] = ACTIONS(5292),
    [anon_sym_static] = ACTIONS(5292),
    [anon_sym_auto] = ACTIONS(5292),
    [anon_sym_register] = ACTIONS(5292),
    [anon_sym_inline] = ACTIONS(5292),
    [anon_sym___inline] = ACTIONS(5292),
    [anon_sym___inline__] = ACTIONS(5292),
    [anon_sym___forceinline] = ACTIONS(5292),
    [anon_sym_thread_local] = ACTIONS(5292),
    [anon_sym___thread] = ACTIONS(5292),
    [anon_sym_CG_EXTERN] = ACTIONS(5292),
    [anon_sym_CG_INLINE] = ACTIONS(5292),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5292),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5292),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5292),
    [anon_sym_IBOutlet] = ACTIONS(5292),
    [anon_sym_IBInspectable] = ACTIONS(5292),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5292),
    [anon_sym_NS_INLINE] = ACTIONS(5292),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5292),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5292),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5292),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5292),
    [anon_sym_const] = ACTIONS(5292),
    [anon_sym_constexpr] = ACTIONS(5292),
    [anon_sym_volatile] = ACTIONS(5292),
    [anon_sym_restrict] = ACTIONS(5292),
    [anon_sym___restrict__] = ACTIONS(5292),
    [anon_sym__Atomic] = ACTIONS(5292),
    [anon_sym__Noreturn] = ACTIONS(5292),
    [anon_sym__Nonnull] = ACTIONS(5292),
    [anon_sym_nullable] = ACTIONS(5292),
    [anon_sym__Complex] = ACTIONS(5292),
    [anon_sym__Nullable] = ACTIONS(5292),
    [anon_sym__Nullable_result] = ACTIONS(5292),
    [anon_sym__Null_unspecified] = ACTIONS(5292),
    [anon_sym___autoreleasing] = ACTIONS(5292),
    [anon_sym___block] = ACTIONS(5292),
    [anon_sym___bridge] = ACTIONS(5292),
    [anon_sym___bridge_retained] = ACTIONS(5292),
    [anon_sym___bridge_transfer] = ACTIONS(5292),
    [anon_sym___complex] = ACTIONS(5292),
    [anon_sym___const] = ACTIONS(5292),
    [anon_sym___imag] = ACTIONS(5292),
    [anon_sym___kindof] = ACTIONS(5292),
    [anon_sym___nonnull] = ACTIONS(5292),
    [anon_sym___nullable] = ACTIONS(5292),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5292),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5292),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5292),
    [anon_sym___real] = ACTIONS(5292),
    [anon_sym___strong] = ACTIONS(5292),
    [anon_sym___unsafe_unretained] = ACTIONS(5292),
    [anon_sym___unused] = ACTIONS(5292),
    [anon_sym___weak] = ACTIONS(5292),
    [anon_sym_alignas] = ACTIONS(5292),
    [anon_sym__Alignas] = ACTIONS(5292),
    [sym_primitive_type] = ACTIONS(5292),
    [anon_sym_enum] = ACTIONS(5292),
    [anon_sym_COLON] = ACTIONS(5294),
    [anon_sym_struct] = ACTIONS(5292),
    [anon_sym_union] = ACTIONS(5292),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5292),
    [anon_sym___typeof] = ACTIONS(5292),
    [anon_sym_typeof] = ACTIONS(5292),
    [aux_sym_preproc_undef_token1] = ACTIONS(5292),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5292),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5292),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5292),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5292),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5292),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5292),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5292),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5292),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5292),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5292),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5292),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5292),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5292),
    [anon_sym_API_AVAILABLE] = ACTIONS(5292),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5292),
    [anon_sym_API_DEPRECATED] = ACTIONS(5292),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5292),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5292),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5292),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5292),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5292),
    [anon_sym___deprecated_msg] = ACTIONS(5292),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5292),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5292),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5292),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5292),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5292),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5292),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5292),
    [anon_sym_ATend] = ACTIONS(5294),
    [anon_sym_ATproperty] = ACTIONS(5294),
    [anon_sym_BOOL] = ACTIONS(5292),
    [anon_sym_IMP] = ACTIONS(5292),
    [anon_sym_SEL] = ACTIONS(5292),
    [anon_sym_Class] = ACTIONS(5292),
    [anon_sym_id] = ACTIONS(5292),
  },
  [STATE(2065)] = {
    [sym_identifier] = ACTIONS(5298),
    [aux_sym_preproc_def_token1] = ACTIONS(5298),
    [aux_sym_preproc_if_token1] = ACTIONS(5298),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5298),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5298),
    [sym_preproc_directive] = ACTIONS(5298),
    [anon_sym_LPAREN2] = ACTIONS(5300),
    [anon_sym_DASH] = ACTIONS(5300),
    [anon_sym_PLUS] = ACTIONS(5300),
    [anon_sym_LT] = ACTIONS(5300),
    [anon_sym_SEMI] = ACTIONS(5302),
    [anon_sym___extension__] = ACTIONS(5298),
    [anon_sym_typedef] = ACTIONS(5298),
    [anon_sym_extern] = ACTIONS(5298),
    [anon_sym___attribute__] = ACTIONS(5298),
    [anon_sym___attribute] = ACTIONS(5298),
    [anon_sym_noreturn] = ACTIONS(5298),
    [anon_sym_LBRACK] = ACTIONS(5300),
    [anon_sym___declspec] = ACTIONS(5298),
    [anon_sym___cdecl] = ACTIONS(5298),
    [anon_sym___clrcall] = ACTIONS(5298),
    [anon_sym___stdcall] = ACTIONS(5298),
    [anon_sym___fastcall] = ACTIONS(5298),
    [anon_sym___thiscall] = ACTIONS(5298),
    [anon_sym___vectorcall] = ACTIONS(5298),
    [anon_sym_LBRACE] = ACTIONS(5300),
    [anon_sym_signed] = ACTIONS(5298),
    [anon_sym_unsigned] = ACTIONS(5298),
    [anon_sym_long] = ACTIONS(5298),
    [anon_sym_short] = ACTIONS(5298),
    [anon_sym_static] = ACTIONS(5298),
    [anon_sym_auto] = ACTIONS(5298),
    [anon_sym_register] = ACTIONS(5298),
    [anon_sym_inline] = ACTIONS(5298),
    [anon_sym___inline] = ACTIONS(5298),
    [anon_sym___inline__] = ACTIONS(5298),
    [anon_sym___forceinline] = ACTIONS(5298),
    [anon_sym_thread_local] = ACTIONS(5298),
    [anon_sym___thread] = ACTIONS(5298),
    [anon_sym_CG_EXTERN] = ACTIONS(5298),
    [anon_sym_CG_INLINE] = ACTIONS(5298),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5298),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5298),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5298),
    [anon_sym_IBOutlet] = ACTIONS(5298),
    [anon_sym_IBInspectable] = ACTIONS(5298),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5298),
    [anon_sym_NS_INLINE] = ACTIONS(5298),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5298),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5298),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5298),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5298),
    [anon_sym_const] = ACTIONS(5298),
    [anon_sym_constexpr] = ACTIONS(5298),
    [anon_sym_volatile] = ACTIONS(5298),
    [anon_sym_restrict] = ACTIONS(5298),
    [anon_sym___restrict__] = ACTIONS(5298),
    [anon_sym__Atomic] = ACTIONS(5298),
    [anon_sym__Noreturn] = ACTIONS(5298),
    [anon_sym__Nonnull] = ACTIONS(5298),
    [anon_sym_nullable] = ACTIONS(5298),
    [anon_sym__Complex] = ACTIONS(5298),
    [anon_sym__Nullable] = ACTIONS(5298),
    [anon_sym__Nullable_result] = ACTIONS(5298),
    [anon_sym__Null_unspecified] = ACTIONS(5298),
    [anon_sym___autoreleasing] = ACTIONS(5298),
    [anon_sym___block] = ACTIONS(5298),
    [anon_sym___bridge] = ACTIONS(5298),
    [anon_sym___bridge_retained] = ACTIONS(5298),
    [anon_sym___bridge_transfer] = ACTIONS(5298),
    [anon_sym___complex] = ACTIONS(5298),
    [anon_sym___const] = ACTIONS(5298),
    [anon_sym___imag] = ACTIONS(5298),
    [anon_sym___kindof] = ACTIONS(5298),
    [anon_sym___nonnull] = ACTIONS(5298),
    [anon_sym___nullable] = ACTIONS(5298),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5298),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5298),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5298),
    [anon_sym___real] = ACTIONS(5298),
    [anon_sym___strong] = ACTIONS(5298),
    [anon_sym___unsafe_unretained] = ACTIONS(5298),
    [anon_sym___unused] = ACTIONS(5298),
    [anon_sym___weak] = ACTIONS(5298),
    [anon_sym_alignas] = ACTIONS(5298),
    [anon_sym__Alignas] = ACTIONS(5298),
    [sym_primitive_type] = ACTIONS(5298),
    [anon_sym_enum] = ACTIONS(5298),
    [anon_sym_COLON] = ACTIONS(5300),
    [anon_sym_struct] = ACTIONS(5298),
    [anon_sym_union] = ACTIONS(5298),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5298),
    [anon_sym___typeof] = ACTIONS(5298),
    [anon_sym_typeof] = ACTIONS(5298),
    [aux_sym_preproc_undef_token1] = ACTIONS(5298),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5298),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5298),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5298),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5298),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5298),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5298),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5298),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5298),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5298),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5298),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5298),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5298),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5298),
    [anon_sym_API_AVAILABLE] = ACTIONS(5298),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5298),
    [anon_sym_API_DEPRECATED] = ACTIONS(5298),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5298),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5298),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5298),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5298),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5298),
    [anon_sym___deprecated_msg] = ACTIONS(5298),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5298),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5298),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5298),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5298),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5298),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5298),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5298),
    [anon_sym_ATend] = ACTIONS(5300),
    [anon_sym_ATproperty] = ACTIONS(5300),
    [anon_sym_BOOL] = ACTIONS(5298),
    [anon_sym_IMP] = ACTIONS(5298),
    [anon_sym_SEL] = ACTIONS(5298),
    [anon_sym_Class] = ACTIONS(5298),
    [anon_sym_id] = ACTIONS(5298),
  },
  [STATE(2066)] = {
    [sym_method_type] = STATE(6560),
    [sym_identifier] = ACTIONS(5304),
    [aux_sym_preproc_def_token1] = ACTIONS(5306),
    [aux_sym_preproc_if_token1] = ACTIONS(5306),
    [aux_sym_preproc_if_token2] = ACTIONS(5306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5306),
    [aux_sym_preproc_else_token1] = ACTIONS(5306),
    [aux_sym_preproc_elif_token1] = ACTIONS(5306),
    [sym_preproc_directive] = ACTIONS(5306),
    [anon_sym_LPAREN2] = ACTIONS(5308),
    [anon_sym_DASH] = ACTIONS(5310),
    [anon_sym_PLUS] = ACTIONS(5310),
    [anon_sym_SEMI] = ACTIONS(5310),
    [anon_sym___extension__] = ACTIONS(5306),
    [anon_sym_typedef] = ACTIONS(5306),
    [anon_sym_extern] = ACTIONS(5306),
    [anon_sym___attribute__] = ACTIONS(5306),
    [anon_sym___attribute] = ACTIONS(5306),
    [anon_sym_noreturn] = ACTIONS(5306),
    [anon_sym_LBRACK] = ACTIONS(5310),
    [anon_sym___declspec] = ACTIONS(5306),
    [anon_sym___cdecl] = ACTIONS(5306),
    [anon_sym___clrcall] = ACTIONS(5306),
    [anon_sym___stdcall] = ACTIONS(5306),
    [anon_sym___fastcall] = ACTIONS(5306),
    [anon_sym___thiscall] = ACTIONS(5306),
    [anon_sym___vectorcall] = ACTIONS(5306),
    [anon_sym_signed] = ACTIONS(5306),
    [anon_sym_unsigned] = ACTIONS(5306),
    [anon_sym_long] = ACTIONS(5306),
    [anon_sym_short] = ACTIONS(5306),
    [anon_sym_static] = ACTIONS(5306),
    [anon_sym_auto] = ACTIONS(5306),
    [anon_sym_register] = ACTIONS(5306),
    [anon_sym_inline] = ACTIONS(5306),
    [anon_sym___inline] = ACTIONS(5306),
    [anon_sym___inline__] = ACTIONS(5306),
    [anon_sym___forceinline] = ACTIONS(5306),
    [anon_sym_thread_local] = ACTIONS(5306),
    [anon_sym___thread] = ACTIONS(5306),
    [anon_sym_CG_EXTERN] = ACTIONS(5306),
    [anon_sym_CG_INLINE] = ACTIONS(5306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5306),
    [anon_sym_IBOutlet] = ACTIONS(5306),
    [anon_sym_IBInspectable] = ACTIONS(5306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5306),
    [anon_sym_NS_INLINE] = ACTIONS(5306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5306),
    [anon_sym_const] = ACTIONS(5306),
    [anon_sym_constexpr] = ACTIONS(5306),
    [anon_sym_volatile] = ACTIONS(5306),
    [anon_sym_restrict] = ACTIONS(5306),
    [anon_sym___restrict__] = ACTIONS(5306),
    [anon_sym__Atomic] = ACTIONS(5306),
    [anon_sym__Noreturn] = ACTIONS(5306),
    [anon_sym__Nonnull] = ACTIONS(5306),
    [anon_sym_nullable] = ACTIONS(5306),
    [anon_sym__Complex] = ACTIONS(5306),
    [anon_sym__Nullable] = ACTIONS(5306),
    [anon_sym__Nullable_result] = ACTIONS(5306),
    [anon_sym__Null_unspecified] = ACTIONS(5306),
    [anon_sym___autoreleasing] = ACTIONS(5306),
    [anon_sym___block] = ACTIONS(5306),
    [anon_sym___bridge] = ACTIONS(5306),
    [anon_sym___bridge_retained] = ACTIONS(5306),
    [anon_sym___bridge_transfer] = ACTIONS(5306),
    [anon_sym___complex] = ACTIONS(5306),
    [anon_sym___const] = ACTIONS(5306),
    [anon_sym___imag] = ACTIONS(5306),
    [anon_sym___kindof] = ACTIONS(5306),
    [anon_sym___nonnull] = ACTIONS(5306),
    [anon_sym___nullable] = ACTIONS(5306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5306),
    [anon_sym___real] = ACTIONS(5306),
    [anon_sym___strong] = ACTIONS(5306),
    [anon_sym___unsafe_unretained] = ACTIONS(5306),
    [anon_sym___unused] = ACTIONS(5306),
    [anon_sym___weak] = ACTIONS(5306),
    [anon_sym_alignas] = ACTIONS(5306),
    [anon_sym__Alignas] = ACTIONS(5306),
    [sym_primitive_type] = ACTIONS(5306),
    [anon_sym_enum] = ACTIONS(5306),
    [anon_sym_struct] = ACTIONS(5306),
    [anon_sym_union] = ACTIONS(5306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5306),
    [anon_sym___typeof] = ACTIONS(5306),
    [anon_sym_typeof] = ACTIONS(5306),
    [aux_sym_preproc_undef_token1] = ACTIONS(5306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_API_AVAILABLE] = ACTIONS(5306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_API_DEPRECATED] = ACTIONS(5306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5306),
    [anon_sym___deprecated_msg] = ACTIONS(5306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5306),
    [anon_sym_ATproperty] = ACTIONS(5310),
    [anon_sym_BOOL] = ACTIONS(5306),
    [anon_sym_IMP] = ACTIONS(5306),
    [anon_sym_SEL] = ACTIONS(5306),
    [anon_sym_Class] = ACTIONS(5306),
    [anon_sym_id] = ACTIONS(5306),
  },
  [STATE(2067)] = {
    [sym__declaration_modifiers] = STATE(2514),
    [sym_attribute_specifier] = STATE(2514),
    [sym_attribute_declaration] = STATE(2514),
    [sym_ms_declspec_modifier] = STATE(2514),
    [sym_ms_based_modifier] = STATE(6497),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2553),
    [sym__declarator] = STATE(3831),
    [sym__abstract_declarator] = STATE(4978),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(2514),
    [sym_type_qualifier] = STATE(2514),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4271),
    [sym_availability_attribute_specifier] = STATE(2514),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2514),
    [aux_sym_pointer_declarator_repeat1] = STATE(2553),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_COMMA] = ACTIONS(5314),
    [anon_sym_RPAREN] = ACTIONS(5314),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(5316),
    [anon_sym_CARET] = ACTIONS(5318),
    [anon_sym_GT] = ACTIONS(5314),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2068)] = {
    [sym_identifier] = ACTIONS(2592),
    [anon_sym_COMMA] = ACTIONS(2594),
    [anon_sym_RPAREN] = ACTIONS(2594),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2594),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2594),
    [anon_sym_LPAREN2] = ACTIONS(2594),
    [anon_sym_STAR] = ACTIONS(2594),
    [anon_sym_CARET] = ACTIONS(2594),
    [anon_sym_GT] = ACTIONS(2594),
    [anon_sym_SEMI] = ACTIONS(2594),
    [anon_sym___extension__] = ACTIONS(2592),
    [anon_sym_extern] = ACTIONS(2592),
    [anon_sym___attribute__] = ACTIONS(2592),
    [anon_sym___attribute] = ACTIONS(2592),
    [anon_sym_noreturn] = ACTIONS(2592),
    [anon_sym_LBRACK] = ACTIONS(2594),
    [anon_sym___declspec] = ACTIONS(2592),
    [anon_sym___based] = ACTIONS(2592),
    [anon_sym_LBRACE] = ACTIONS(2594),
    [anon_sym_RBRACE] = ACTIONS(2594),
    [anon_sym_signed] = ACTIONS(2592),
    [anon_sym_unsigned] = ACTIONS(2592),
    [anon_sym_long] = ACTIONS(2592),
    [anon_sym_short] = ACTIONS(2592),
    [anon_sym_static] = ACTIONS(2592),
    [anon_sym_EQ] = ACTIONS(2594),
    [anon_sym_ATautoreleasepool] = ACTIONS(2594),
    [anon_sym_auto] = ACTIONS(2592),
    [anon_sym_register] = ACTIONS(2592),
    [anon_sym_inline] = ACTIONS(2592),
    [anon_sym___inline] = ACTIONS(2592),
    [anon_sym___inline__] = ACTIONS(2592),
    [anon_sym___forceinline] = ACTIONS(2592),
    [anon_sym_thread_local] = ACTIONS(2592),
    [anon_sym___thread] = ACTIONS(2592),
    [anon_sym_CG_EXTERN] = ACTIONS(2592),
    [anon_sym_CG_INLINE] = ACTIONS(2592),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2592),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2592),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2592),
    [anon_sym_IBOutlet] = ACTIONS(2592),
    [anon_sym_IBInspectable] = ACTIONS(2592),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2592),
    [anon_sym_NS_INLINE] = ACTIONS(2592),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2592),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2592),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2592),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2592),
    [anon_sym_const] = ACTIONS(2592),
    [anon_sym_constexpr] = ACTIONS(2592),
    [anon_sym_volatile] = ACTIONS(2592),
    [anon_sym_restrict] = ACTIONS(2592),
    [anon_sym___restrict__] = ACTIONS(2592),
    [anon_sym__Atomic] = ACTIONS(2592),
    [anon_sym__Noreturn] = ACTIONS(2592),
    [anon_sym__Nonnull] = ACTIONS(2592),
    [anon_sym_nullable] = ACTIONS(2592),
    [anon_sym__Complex] = ACTIONS(2592),
    [anon_sym__Nullable] = ACTIONS(2592),
    [anon_sym__Nullable_result] = ACTIONS(2592),
    [anon_sym__Null_unspecified] = ACTIONS(2592),
    [anon_sym___autoreleasing] = ACTIONS(2592),
    [anon_sym___block] = ACTIONS(2592),
    [anon_sym___bridge] = ACTIONS(2592),
    [anon_sym___bridge_retained] = ACTIONS(2592),
    [anon_sym___bridge_transfer] = ACTIONS(2592),
    [anon_sym___complex] = ACTIONS(2592),
    [anon_sym___const] = ACTIONS(2592),
    [anon_sym___imag] = ACTIONS(2592),
    [anon_sym___kindof] = ACTIONS(2592),
    [anon_sym___nonnull] = ACTIONS(2592),
    [anon_sym___nullable] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2592),
    [anon_sym___real] = ACTIONS(2592),
    [anon_sym___strong] = ACTIONS(2592),
    [anon_sym___unsafe_unretained] = ACTIONS(2592),
    [anon_sym___unused] = ACTIONS(2592),
    [anon_sym___weak] = ACTIONS(2592),
    [anon_sym_alignas] = ACTIONS(2592),
    [anon_sym__Alignas] = ACTIONS(2592),
    [sym_primitive_type] = ACTIONS(2592),
    [anon_sym_enum] = ACTIONS(2592),
    [anon_sym_COLON] = ACTIONS(2594),
    [anon_sym_struct] = ACTIONS(2592),
    [anon_sym_union] = ACTIONS(2592),
    [anon_sym_asm] = ACTIONS(2592),
    [anon_sym___asm__] = ACTIONS(2592),
    [anon_sym___asm] = ACTIONS(2592),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2592),
    [anon_sym___typeof] = ACTIONS(2592),
    [anon_sym_typeof] = ACTIONS(2592),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2592),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2592),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2592),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2592),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2592),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2592),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2592),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_API_AVAILABLE] = ACTIONS(2592),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_API_DEPRECATED] = ACTIONS(2592),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2592),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2592),
    [anon_sym___deprecated_msg] = ACTIONS(2592),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2592),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2592),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2592),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2592),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2592),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2592),
    [anon_sym_ATprotocol] = ACTIONS(2594),
    [anon_sym_ATinterface] = ACTIONS(2594),
    [anon_sym_ATimplementation] = ACTIONS(2594),
    [anon_sym_ATproperty] = ACTIONS(2594),
    [anon_sym_BOOL] = ACTIONS(2592),
    [anon_sym_IMP] = ACTIONS(2592),
    [anon_sym_SEL] = ACTIONS(2592),
    [anon_sym_Class] = ACTIONS(2592),
    [anon_sym_id] = ACTIONS(2592),
  },
  [STATE(2069)] = {
    [sym_string_literal] = STATE(1916),
    [sym_identifier] = ACTIONS(5324),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4006),
    [anon_sym_COMMA] = ACTIONS(4006),
    [anon_sym_RPAREN] = ACTIONS(4006),
    [anon_sym_LPAREN2] = ACTIONS(4006),
    [anon_sym_DASH] = ACTIONS(4004),
    [anon_sym_PLUS] = ACTIONS(4004),
    [anon_sym_STAR] = ACTIONS(4006),
    [anon_sym_SLASH] = ACTIONS(4004),
    [anon_sym_PERCENT] = ACTIONS(4006),
    [anon_sym_PIPE_PIPE] = ACTIONS(4006),
    [anon_sym_AMP_AMP] = ACTIONS(4006),
    [anon_sym_PIPE] = ACTIONS(4004),
    [anon_sym_CARET] = ACTIONS(4006),
    [anon_sym_AMP] = ACTIONS(4004),
    [anon_sym_EQ_EQ] = ACTIONS(4006),
    [anon_sym_BANG_EQ] = ACTIONS(4006),
    [anon_sym_GT] = ACTIONS(4004),
    [anon_sym_GT_EQ] = ACTIONS(4006),
    [anon_sym_LT_EQ] = ACTIONS(4006),
    [anon_sym_LT] = ACTIONS(4004),
    [anon_sym_LT_LT] = ACTIONS(4006),
    [anon_sym_GT_GT] = ACTIONS(4006),
    [anon_sym_SEMI] = ACTIONS(4006),
    [anon_sym___extension__] = ACTIONS(4004),
    [anon_sym_extern] = ACTIONS(4004),
    [anon_sym___attribute__] = ACTIONS(4004),
    [anon_sym___attribute] = ACTIONS(4004),
    [anon_sym_noreturn] = ACTIONS(4004),
    [anon_sym_LBRACK] = ACTIONS(4006),
    [anon_sym_RBRACK] = ACTIONS(4006),
    [anon_sym___declspec] = ACTIONS(4004),
    [anon_sym_RBRACE] = ACTIONS(4006),
    [anon_sym_static] = ACTIONS(4004),
    [anon_sym_auto] = ACTIONS(4004),
    [anon_sym_register] = ACTIONS(4004),
    [anon_sym_inline] = ACTIONS(4004),
    [anon_sym___inline] = ACTIONS(4004),
    [anon_sym___inline__] = ACTIONS(4004),
    [anon_sym___forceinline] = ACTIONS(4004),
    [anon_sym_thread_local] = ACTIONS(4004),
    [anon_sym___thread] = ACTIONS(4004),
    [anon_sym_CG_EXTERN] = ACTIONS(4004),
    [anon_sym_CG_INLINE] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4004),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4004),
    [anon_sym_IBOutlet] = ACTIONS(4004),
    [anon_sym_IBInspectable] = ACTIONS(4004),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4004),
    [anon_sym_NS_INLINE] = ACTIONS(4004),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4004),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4004),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4004),
    [anon_sym_const] = ACTIONS(4004),
    [anon_sym_constexpr] = ACTIONS(4004),
    [anon_sym_volatile] = ACTIONS(4004),
    [anon_sym_restrict] = ACTIONS(4004),
    [anon_sym___restrict__] = ACTIONS(4004),
    [anon_sym__Atomic] = ACTIONS(4004),
    [anon_sym__Noreturn] = ACTIONS(4004),
    [anon_sym__Nonnull] = ACTIONS(4004),
    [anon_sym_nullable] = ACTIONS(4004),
    [anon_sym__Complex] = ACTIONS(4004),
    [anon_sym__Nullable] = ACTIONS(4004),
    [anon_sym__Nullable_result] = ACTIONS(4004),
    [anon_sym__Null_unspecified] = ACTIONS(4004),
    [anon_sym___autoreleasing] = ACTIONS(4004),
    [anon_sym___block] = ACTIONS(4004),
    [anon_sym___bridge] = ACTIONS(4004),
    [anon_sym___bridge_retained] = ACTIONS(4004),
    [anon_sym___bridge_transfer] = ACTIONS(4004),
    [anon_sym___complex] = ACTIONS(4004),
    [anon_sym___const] = ACTIONS(4004),
    [anon_sym___imag] = ACTIONS(4004),
    [anon_sym___kindof] = ACTIONS(4004),
    [anon_sym___nonnull] = ACTIONS(4004),
    [anon_sym___nullable] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4004),
    [anon_sym___real] = ACTIONS(4004),
    [anon_sym___strong] = ACTIONS(4004),
    [anon_sym___unsafe_unretained] = ACTIONS(4004),
    [anon_sym___unused] = ACTIONS(4004),
    [anon_sym___weak] = ACTIONS(4004),
    [anon_sym_alignas] = ACTIONS(4004),
    [anon_sym__Alignas] = ACTIONS(4004),
    [anon_sym_COLON] = ACTIONS(4006),
    [anon_sym_QMARK] = ACTIONS(4006),
    [anon_sym_DASH_DASH] = ACTIONS(4006),
    [anon_sym_PLUS_PLUS] = ACTIONS(4006),
    [anon_sym_DOT] = ACTIONS(4004),
    [anon_sym_DASH_GT] = ACTIONS(4006),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4004),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4004),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4004),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_API_AVAILABLE] = ACTIONS(4004),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_API_DEPRECATED] = ACTIONS(4004),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym___deprecated_msg] = ACTIONS(4004),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4004),
  },
  [STATE(2070)] = {
    [sym_method_type] = STATE(6560),
    [sym_identifier] = ACTIONS(5304),
    [aux_sym_preproc_def_token1] = ACTIONS(5306),
    [aux_sym_preproc_if_token1] = ACTIONS(5306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5306),
    [sym_preproc_directive] = ACTIONS(5306),
    [anon_sym_LPAREN2] = ACTIONS(5308),
    [anon_sym_DASH] = ACTIONS(5310),
    [anon_sym_PLUS] = ACTIONS(5310),
    [anon_sym_SEMI] = ACTIONS(5310),
    [anon_sym___extension__] = ACTIONS(5306),
    [anon_sym_typedef] = ACTIONS(5306),
    [anon_sym_extern] = ACTIONS(5306),
    [anon_sym___attribute__] = ACTIONS(5306),
    [anon_sym___attribute] = ACTIONS(5306),
    [anon_sym_noreturn] = ACTIONS(5306),
    [anon_sym_LBRACK] = ACTIONS(5310),
    [anon_sym___declspec] = ACTIONS(5306),
    [anon_sym___cdecl] = ACTIONS(5306),
    [anon_sym___clrcall] = ACTIONS(5306),
    [anon_sym___stdcall] = ACTIONS(5306),
    [anon_sym___fastcall] = ACTIONS(5306),
    [anon_sym___thiscall] = ACTIONS(5306),
    [anon_sym___vectorcall] = ACTIONS(5306),
    [anon_sym_signed] = ACTIONS(5306),
    [anon_sym_unsigned] = ACTIONS(5306),
    [anon_sym_long] = ACTIONS(5306),
    [anon_sym_short] = ACTIONS(5306),
    [anon_sym_static] = ACTIONS(5306),
    [anon_sym_auto] = ACTIONS(5306),
    [anon_sym_register] = ACTIONS(5306),
    [anon_sym_inline] = ACTIONS(5306),
    [anon_sym___inline] = ACTIONS(5306),
    [anon_sym___inline__] = ACTIONS(5306),
    [anon_sym___forceinline] = ACTIONS(5306),
    [anon_sym_thread_local] = ACTIONS(5306),
    [anon_sym___thread] = ACTIONS(5306),
    [anon_sym_CG_EXTERN] = ACTIONS(5306),
    [anon_sym_CG_INLINE] = ACTIONS(5306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5306),
    [anon_sym_IBOutlet] = ACTIONS(5306),
    [anon_sym_IBInspectable] = ACTIONS(5306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5306),
    [anon_sym_NS_INLINE] = ACTIONS(5306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5306),
    [anon_sym_const] = ACTIONS(5306),
    [anon_sym_constexpr] = ACTIONS(5306),
    [anon_sym_volatile] = ACTIONS(5306),
    [anon_sym_restrict] = ACTIONS(5306),
    [anon_sym___restrict__] = ACTIONS(5306),
    [anon_sym__Atomic] = ACTIONS(5306),
    [anon_sym__Noreturn] = ACTIONS(5306),
    [anon_sym__Nonnull] = ACTIONS(5306),
    [anon_sym_nullable] = ACTIONS(5306),
    [anon_sym__Complex] = ACTIONS(5306),
    [anon_sym__Nullable] = ACTIONS(5306),
    [anon_sym__Nullable_result] = ACTIONS(5306),
    [anon_sym__Null_unspecified] = ACTIONS(5306),
    [anon_sym___autoreleasing] = ACTIONS(5306),
    [anon_sym___block] = ACTIONS(5306),
    [anon_sym___bridge] = ACTIONS(5306),
    [anon_sym___bridge_retained] = ACTIONS(5306),
    [anon_sym___bridge_transfer] = ACTIONS(5306),
    [anon_sym___complex] = ACTIONS(5306),
    [anon_sym___const] = ACTIONS(5306),
    [anon_sym___imag] = ACTIONS(5306),
    [anon_sym___kindof] = ACTIONS(5306),
    [anon_sym___nonnull] = ACTIONS(5306),
    [anon_sym___nullable] = ACTIONS(5306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5306),
    [anon_sym___real] = ACTIONS(5306),
    [anon_sym___strong] = ACTIONS(5306),
    [anon_sym___unsafe_unretained] = ACTIONS(5306),
    [anon_sym___unused] = ACTIONS(5306),
    [anon_sym___weak] = ACTIONS(5306),
    [anon_sym_alignas] = ACTIONS(5306),
    [anon_sym__Alignas] = ACTIONS(5306),
    [sym_primitive_type] = ACTIONS(5306),
    [anon_sym_enum] = ACTIONS(5306),
    [anon_sym_struct] = ACTIONS(5306),
    [anon_sym_union] = ACTIONS(5306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5306),
    [anon_sym___typeof] = ACTIONS(5306),
    [anon_sym_typeof] = ACTIONS(5306),
    [aux_sym_preproc_undef_token1] = ACTIONS(5306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_API_AVAILABLE] = ACTIONS(5306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_API_DEPRECATED] = ACTIONS(5306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5306),
    [anon_sym___deprecated_msg] = ACTIONS(5306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5306),
    [anon_sym_ATend] = ACTIONS(5310),
    [anon_sym_AToptional] = ACTIONS(5310),
    [anon_sym_ATrequired] = ACTIONS(5310),
    [anon_sym_ATproperty] = ACTIONS(5310),
    [anon_sym_BOOL] = ACTIONS(5306),
    [anon_sym_IMP] = ACTIONS(5306),
    [anon_sym_SEL] = ACTIONS(5306),
    [anon_sym_Class] = ACTIONS(5306),
    [anon_sym_id] = ACTIONS(5306),
  },
  [STATE(2071)] = {
    [sym_identifier] = ACTIONS(5068),
    [aux_sym_preproc_def_token1] = ACTIONS(5068),
    [aux_sym_preproc_if_token1] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5068),
    [sym_preproc_directive] = ACTIONS(5068),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_DASH] = ACTIONS(5072),
    [anon_sym_PLUS] = ACTIONS(5072),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym___extension__] = ACTIONS(5068),
    [anon_sym_typedef] = ACTIONS(5068),
    [anon_sym_extern] = ACTIONS(5068),
    [anon_sym___attribute__] = ACTIONS(5068),
    [anon_sym___attribute] = ACTIONS(5068),
    [anon_sym_noreturn] = ACTIONS(5068),
    [anon_sym_LBRACK] = ACTIONS(5072),
    [anon_sym___declspec] = ACTIONS(5068),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym___cdecl] = ACTIONS(5068),
    [anon_sym___clrcall] = ACTIONS(5068),
    [anon_sym___stdcall] = ACTIONS(5068),
    [anon_sym___fastcall] = ACTIONS(5068),
    [anon_sym___thiscall] = ACTIONS(5068),
    [anon_sym___vectorcall] = ACTIONS(5068),
    [anon_sym_signed] = ACTIONS(5068),
    [anon_sym_unsigned] = ACTIONS(5068),
    [anon_sym_long] = ACTIONS(5068),
    [anon_sym_short] = ACTIONS(5068),
    [anon_sym_static] = ACTIONS(5068),
    [anon_sym_auto] = ACTIONS(5068),
    [anon_sym_register] = ACTIONS(5068),
    [anon_sym_inline] = ACTIONS(5068),
    [anon_sym___inline] = ACTIONS(5068),
    [anon_sym___inline__] = ACTIONS(5068),
    [anon_sym___forceinline] = ACTIONS(5068),
    [anon_sym_thread_local] = ACTIONS(5068),
    [anon_sym___thread] = ACTIONS(5068),
    [anon_sym_CG_EXTERN] = ACTIONS(5068),
    [anon_sym_CG_INLINE] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5068),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5068),
    [anon_sym_IBOutlet] = ACTIONS(5068),
    [anon_sym_IBInspectable] = ACTIONS(5068),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5068),
    [anon_sym_NS_INLINE] = ACTIONS(5068),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5068),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5068),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5068),
    [anon_sym_const] = ACTIONS(5068),
    [anon_sym_constexpr] = ACTIONS(5068),
    [anon_sym_volatile] = ACTIONS(5068),
    [anon_sym_restrict] = ACTIONS(5068),
    [anon_sym___restrict__] = ACTIONS(5068),
    [anon_sym__Atomic] = ACTIONS(5068),
    [anon_sym__Noreturn] = ACTIONS(5068),
    [anon_sym__Nonnull] = ACTIONS(5068),
    [anon_sym_nullable] = ACTIONS(5068),
    [anon_sym__Complex] = ACTIONS(5068),
    [anon_sym__Nullable] = ACTIONS(5068),
    [anon_sym__Nullable_result] = ACTIONS(5068),
    [anon_sym__Null_unspecified] = ACTIONS(5068),
    [anon_sym___autoreleasing] = ACTIONS(5068),
    [anon_sym___block] = ACTIONS(5068),
    [anon_sym___bridge] = ACTIONS(5068),
    [anon_sym___bridge_retained] = ACTIONS(5068),
    [anon_sym___bridge_transfer] = ACTIONS(5068),
    [anon_sym___complex] = ACTIONS(5068),
    [anon_sym___const] = ACTIONS(5068),
    [anon_sym___imag] = ACTIONS(5068),
    [anon_sym___kindof] = ACTIONS(5068),
    [anon_sym___nonnull] = ACTIONS(5068),
    [anon_sym___nullable] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5068),
    [anon_sym___real] = ACTIONS(5068),
    [anon_sym___strong] = ACTIONS(5068),
    [anon_sym___unsafe_unretained] = ACTIONS(5068),
    [anon_sym___unused] = ACTIONS(5068),
    [anon_sym___weak] = ACTIONS(5068),
    [anon_sym_alignas] = ACTIONS(5068),
    [anon_sym__Alignas] = ACTIONS(5068),
    [sym_primitive_type] = ACTIONS(5068),
    [anon_sym_enum] = ACTIONS(5068),
    [anon_sym_struct] = ACTIONS(5068),
    [anon_sym_union] = ACTIONS(5068),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5068),
    [anon_sym___typeof] = ACTIONS(5068),
    [anon_sym_typeof] = ACTIONS(5068),
    [aux_sym_preproc_undef_token1] = ACTIONS(5068),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5068),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5068),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5068),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_API_AVAILABLE] = ACTIONS(5068),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_API_DEPRECATED] = ACTIONS(5068),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym___deprecated_msg] = ACTIONS(5068),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5068),
    [anon_sym_ATend] = ACTIONS(5072),
    [anon_sym_ATsynthesize] = ACTIONS(5072),
    [anon_sym_ATdynamic] = ACTIONS(5072),
    [anon_sym_BOOL] = ACTIONS(5068),
    [anon_sym_IMP] = ACTIONS(5068),
    [anon_sym_SEL] = ACTIONS(5068),
    [anon_sym_Class] = ACTIONS(5068),
    [anon_sym_id] = ACTIONS(5068),
  },
  [STATE(2072)] = {
    [sym_string_literal] = STATE(1916),
    [sym_identifier] = ACTIONS(4004),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4006),
    [anon_sym_COMMA] = ACTIONS(4006),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4006),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4006),
    [anon_sym_LPAREN2] = ACTIONS(4006),
    [anon_sym_DASH] = ACTIONS(4004),
    [anon_sym_PLUS] = ACTIONS(4004),
    [anon_sym_STAR] = ACTIONS(4006),
    [anon_sym_SLASH] = ACTIONS(4004),
    [anon_sym_PERCENT] = ACTIONS(4006),
    [anon_sym_PIPE_PIPE] = ACTIONS(4006),
    [anon_sym_AMP_AMP] = ACTIONS(4006),
    [anon_sym_PIPE] = ACTIONS(4004),
    [anon_sym_CARET] = ACTIONS(4006),
    [anon_sym_AMP] = ACTIONS(4004),
    [anon_sym_EQ_EQ] = ACTIONS(4006),
    [anon_sym_BANG_EQ] = ACTIONS(4006),
    [anon_sym_GT] = ACTIONS(4004),
    [anon_sym_GT_EQ] = ACTIONS(4006),
    [anon_sym_LT_EQ] = ACTIONS(4006),
    [anon_sym_LT] = ACTIONS(4004),
    [anon_sym_LT_LT] = ACTIONS(4006),
    [anon_sym_GT_GT] = ACTIONS(4006),
    [anon_sym___extension__] = ACTIONS(4004),
    [anon_sym_extern] = ACTIONS(4004),
    [anon_sym___attribute__] = ACTIONS(4004),
    [anon_sym___attribute] = ACTIONS(4004),
    [anon_sym_noreturn] = ACTIONS(4004),
    [anon_sym_LBRACK] = ACTIONS(4006),
    [anon_sym_RBRACK] = ACTIONS(4006),
    [anon_sym___declspec] = ACTIONS(4004),
    [anon_sym_RBRACE] = ACTIONS(4006),
    [anon_sym_static] = ACTIONS(4004),
    [anon_sym_auto] = ACTIONS(4004),
    [anon_sym_register] = ACTIONS(4004),
    [anon_sym_inline] = ACTIONS(4004),
    [anon_sym___inline] = ACTIONS(4004),
    [anon_sym___inline__] = ACTIONS(4004),
    [anon_sym___forceinline] = ACTIONS(4004),
    [anon_sym_thread_local] = ACTIONS(4004),
    [anon_sym___thread] = ACTIONS(4004),
    [anon_sym_CG_EXTERN] = ACTIONS(4004),
    [anon_sym_CG_INLINE] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4004),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4004),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4004),
    [anon_sym_IBOutlet] = ACTIONS(4004),
    [anon_sym_IBInspectable] = ACTIONS(4004),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4004),
    [anon_sym_NS_INLINE] = ACTIONS(4004),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4004),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4004),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4004),
    [anon_sym_const] = ACTIONS(4004),
    [anon_sym_constexpr] = ACTIONS(4004),
    [anon_sym_volatile] = ACTIONS(4004),
    [anon_sym_restrict] = ACTIONS(4004),
    [anon_sym___restrict__] = ACTIONS(4004),
    [anon_sym__Atomic] = ACTIONS(4004),
    [anon_sym__Noreturn] = ACTIONS(4004),
    [anon_sym__Nonnull] = ACTIONS(4004),
    [anon_sym_nullable] = ACTIONS(4004),
    [anon_sym__Complex] = ACTIONS(4004),
    [anon_sym__Nullable] = ACTIONS(4004),
    [anon_sym__Nullable_result] = ACTIONS(4004),
    [anon_sym__Null_unspecified] = ACTIONS(4004),
    [anon_sym___autoreleasing] = ACTIONS(4004),
    [anon_sym___block] = ACTIONS(4004),
    [anon_sym___bridge] = ACTIONS(4004),
    [anon_sym___bridge_retained] = ACTIONS(4004),
    [anon_sym___bridge_transfer] = ACTIONS(4004),
    [anon_sym___complex] = ACTIONS(4004),
    [anon_sym___const] = ACTIONS(4004),
    [anon_sym___imag] = ACTIONS(4004),
    [anon_sym___kindof] = ACTIONS(4004),
    [anon_sym___nonnull] = ACTIONS(4004),
    [anon_sym___nullable] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4004),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4004),
    [anon_sym___real] = ACTIONS(4004),
    [anon_sym___strong] = ACTIONS(4004),
    [anon_sym___unsafe_unretained] = ACTIONS(4004),
    [anon_sym___unused] = ACTIONS(4004),
    [anon_sym___weak] = ACTIONS(4004),
    [anon_sym_alignas] = ACTIONS(4004),
    [anon_sym__Alignas] = ACTIONS(4004),
    [anon_sym_COLON] = ACTIONS(4006),
    [anon_sym_QMARK] = ACTIONS(4006),
    [anon_sym_DASH_DASH] = ACTIONS(4006),
    [anon_sym_PLUS_PLUS] = ACTIONS(4006),
    [anon_sym_DOT] = ACTIONS(4004),
    [anon_sym_DASH_GT] = ACTIONS(4006),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4004),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4004),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4004),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4004),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4004),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_API_AVAILABLE] = ACTIONS(4004),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_API_DEPRECATED] = ACTIONS(4004),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4004),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4004),
    [anon_sym___deprecated_msg] = ACTIONS(4004),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4004),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4004),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4004),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4004),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4004),
  },
  [STATE(2073)] = {
    [sym_identifier] = ACTIONS(5068),
    [aux_sym_preproc_def_token1] = ACTIONS(5068),
    [aux_sym_preproc_if_token1] = ACTIONS(5068),
    [aux_sym_preproc_if_token2] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5068),
    [sym_preproc_directive] = ACTIONS(5068),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_DASH] = ACTIONS(5072),
    [anon_sym_PLUS] = ACTIONS(5072),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym___extension__] = ACTIONS(5068),
    [anon_sym_typedef] = ACTIONS(5068),
    [anon_sym_extern] = ACTIONS(5068),
    [anon_sym___attribute__] = ACTIONS(5068),
    [anon_sym___attribute] = ACTIONS(5068),
    [anon_sym_noreturn] = ACTIONS(5068),
    [anon_sym_LBRACK] = ACTIONS(5072),
    [anon_sym___declspec] = ACTIONS(5068),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym___cdecl] = ACTIONS(5068),
    [anon_sym___clrcall] = ACTIONS(5068),
    [anon_sym___stdcall] = ACTIONS(5068),
    [anon_sym___fastcall] = ACTIONS(5068),
    [anon_sym___thiscall] = ACTIONS(5068),
    [anon_sym___vectorcall] = ACTIONS(5068),
    [anon_sym_signed] = ACTIONS(5068),
    [anon_sym_unsigned] = ACTIONS(5068),
    [anon_sym_long] = ACTIONS(5068),
    [anon_sym_short] = ACTIONS(5068),
    [anon_sym_static] = ACTIONS(5068),
    [anon_sym_auto] = ACTIONS(5068),
    [anon_sym_register] = ACTIONS(5068),
    [anon_sym_inline] = ACTIONS(5068),
    [anon_sym___inline] = ACTIONS(5068),
    [anon_sym___inline__] = ACTIONS(5068),
    [anon_sym___forceinline] = ACTIONS(5068),
    [anon_sym_thread_local] = ACTIONS(5068),
    [anon_sym___thread] = ACTIONS(5068),
    [anon_sym_CG_EXTERN] = ACTIONS(5068),
    [anon_sym_CG_INLINE] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5068),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5068),
    [anon_sym_IBOutlet] = ACTIONS(5068),
    [anon_sym_IBInspectable] = ACTIONS(5068),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5068),
    [anon_sym_NS_INLINE] = ACTIONS(5068),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5068),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5068),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5068),
    [anon_sym_const] = ACTIONS(5068),
    [anon_sym_constexpr] = ACTIONS(5068),
    [anon_sym_volatile] = ACTIONS(5068),
    [anon_sym_restrict] = ACTIONS(5068),
    [anon_sym___restrict__] = ACTIONS(5068),
    [anon_sym__Atomic] = ACTIONS(5068),
    [anon_sym__Noreturn] = ACTIONS(5068),
    [anon_sym__Nonnull] = ACTIONS(5068),
    [anon_sym_nullable] = ACTIONS(5068),
    [anon_sym__Complex] = ACTIONS(5068),
    [anon_sym__Nullable] = ACTIONS(5068),
    [anon_sym__Nullable_result] = ACTIONS(5068),
    [anon_sym__Null_unspecified] = ACTIONS(5068),
    [anon_sym___autoreleasing] = ACTIONS(5068),
    [anon_sym___block] = ACTIONS(5068),
    [anon_sym___bridge] = ACTIONS(5068),
    [anon_sym___bridge_retained] = ACTIONS(5068),
    [anon_sym___bridge_transfer] = ACTIONS(5068),
    [anon_sym___complex] = ACTIONS(5068),
    [anon_sym___const] = ACTIONS(5068),
    [anon_sym___imag] = ACTIONS(5068),
    [anon_sym___kindof] = ACTIONS(5068),
    [anon_sym___nonnull] = ACTIONS(5068),
    [anon_sym___nullable] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5068),
    [anon_sym___real] = ACTIONS(5068),
    [anon_sym___strong] = ACTIONS(5068),
    [anon_sym___unsafe_unretained] = ACTIONS(5068),
    [anon_sym___unused] = ACTIONS(5068),
    [anon_sym___weak] = ACTIONS(5068),
    [anon_sym_alignas] = ACTIONS(5068),
    [anon_sym__Alignas] = ACTIONS(5068),
    [sym_primitive_type] = ACTIONS(5068),
    [anon_sym_enum] = ACTIONS(5068),
    [anon_sym_struct] = ACTIONS(5068),
    [anon_sym_union] = ACTIONS(5068),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5068),
    [anon_sym___typeof] = ACTIONS(5068),
    [anon_sym_typeof] = ACTIONS(5068),
    [aux_sym_preproc_undef_token1] = ACTIONS(5068),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5068),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5068),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5068),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_API_AVAILABLE] = ACTIONS(5068),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_API_DEPRECATED] = ACTIONS(5068),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym___deprecated_msg] = ACTIONS(5068),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5068),
    [anon_sym_ATsynthesize] = ACTIONS(5072),
    [anon_sym_ATdynamic] = ACTIONS(5072),
    [anon_sym_BOOL] = ACTIONS(5068),
    [anon_sym_IMP] = ACTIONS(5068),
    [anon_sym_SEL] = ACTIONS(5068),
    [anon_sym_Class] = ACTIONS(5068),
    [anon_sym_id] = ACTIONS(5068),
  },
  [STATE(2074)] = {
    [sym_identifier] = ACTIONS(5326),
    [aux_sym_preproc_def_token1] = ACTIONS(5326),
    [aux_sym_preproc_if_token1] = ACTIONS(5326),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5326),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5326),
    [sym_preproc_directive] = ACTIONS(5326),
    [anon_sym_LPAREN2] = ACTIONS(5328),
    [anon_sym_DASH] = ACTIONS(5330),
    [anon_sym_PLUS] = ACTIONS(5330),
    [anon_sym_LT] = ACTIONS(5328),
    [anon_sym___extension__] = ACTIONS(5326),
    [anon_sym_typedef] = ACTIONS(5326),
    [anon_sym_extern] = ACTIONS(5326),
    [anon_sym___attribute__] = ACTIONS(5326),
    [anon_sym___attribute] = ACTIONS(5326),
    [anon_sym_noreturn] = ACTIONS(5326),
    [anon_sym_LBRACK] = ACTIONS(5330),
    [anon_sym___declspec] = ACTIONS(5326),
    [anon_sym___cdecl] = ACTIONS(5326),
    [anon_sym___clrcall] = ACTIONS(5326),
    [anon_sym___stdcall] = ACTIONS(5326),
    [anon_sym___fastcall] = ACTIONS(5326),
    [anon_sym___thiscall] = ACTIONS(5326),
    [anon_sym___vectorcall] = ACTIONS(5326),
    [anon_sym_LBRACE] = ACTIONS(5330),
    [anon_sym_signed] = ACTIONS(5326),
    [anon_sym_unsigned] = ACTIONS(5326),
    [anon_sym_long] = ACTIONS(5326),
    [anon_sym_short] = ACTIONS(5326),
    [anon_sym_static] = ACTIONS(5326),
    [anon_sym_auto] = ACTIONS(5326),
    [anon_sym_register] = ACTIONS(5326),
    [anon_sym_inline] = ACTIONS(5326),
    [anon_sym___inline] = ACTIONS(5326),
    [anon_sym___inline__] = ACTIONS(5326),
    [anon_sym___forceinline] = ACTIONS(5326),
    [anon_sym_thread_local] = ACTIONS(5326),
    [anon_sym___thread] = ACTIONS(5326),
    [anon_sym_CG_EXTERN] = ACTIONS(5326),
    [anon_sym_CG_INLINE] = ACTIONS(5326),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5326),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5326),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5326),
    [anon_sym_IBOutlet] = ACTIONS(5326),
    [anon_sym_IBInspectable] = ACTIONS(5326),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5326),
    [anon_sym_NS_INLINE] = ACTIONS(5326),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5326),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5326),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5326),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5326),
    [anon_sym_const] = ACTIONS(5326),
    [anon_sym_constexpr] = ACTIONS(5326),
    [anon_sym_volatile] = ACTIONS(5326),
    [anon_sym_restrict] = ACTIONS(5326),
    [anon_sym___restrict__] = ACTIONS(5326),
    [anon_sym__Atomic] = ACTIONS(5326),
    [anon_sym__Noreturn] = ACTIONS(5326),
    [anon_sym__Nonnull] = ACTIONS(5326),
    [anon_sym_nullable] = ACTIONS(5326),
    [anon_sym__Complex] = ACTIONS(5326),
    [anon_sym__Nullable] = ACTIONS(5326),
    [anon_sym__Nullable_result] = ACTIONS(5326),
    [anon_sym__Null_unspecified] = ACTIONS(5326),
    [anon_sym___autoreleasing] = ACTIONS(5326),
    [anon_sym___block] = ACTIONS(5326),
    [anon_sym___bridge] = ACTIONS(5326),
    [anon_sym___bridge_retained] = ACTIONS(5326),
    [anon_sym___bridge_transfer] = ACTIONS(5326),
    [anon_sym___complex] = ACTIONS(5326),
    [anon_sym___const] = ACTIONS(5326),
    [anon_sym___imag] = ACTIONS(5326),
    [anon_sym___kindof] = ACTIONS(5326),
    [anon_sym___nonnull] = ACTIONS(5326),
    [anon_sym___nullable] = ACTIONS(5326),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5326),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5326),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5326),
    [anon_sym___real] = ACTIONS(5326),
    [anon_sym___strong] = ACTIONS(5326),
    [anon_sym___unsafe_unretained] = ACTIONS(5326),
    [anon_sym___unused] = ACTIONS(5326),
    [anon_sym___weak] = ACTIONS(5326),
    [anon_sym_alignas] = ACTIONS(5326),
    [anon_sym__Alignas] = ACTIONS(5326),
    [sym_primitive_type] = ACTIONS(5326),
    [anon_sym_enum] = ACTIONS(5326),
    [anon_sym_COLON] = ACTIONS(5328),
    [anon_sym_struct] = ACTIONS(5326),
    [anon_sym_union] = ACTIONS(5326),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5326),
    [anon_sym___typeof] = ACTIONS(5326),
    [anon_sym_typeof] = ACTIONS(5326),
    [aux_sym_preproc_undef_token1] = ACTIONS(5326),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5326),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5326),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5326),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5326),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5326),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5326),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5326),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5326),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5326),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5326),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5326),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5326),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5326),
    [anon_sym_API_AVAILABLE] = ACTIONS(5326),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5326),
    [anon_sym_API_DEPRECATED] = ACTIONS(5326),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5326),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5326),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5326),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5326),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5326),
    [anon_sym___deprecated_msg] = ACTIONS(5326),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5326),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5326),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5326),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5326),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5326),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5326),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5326),
    [anon_sym_ATend] = ACTIONS(5330),
    [anon_sym_ATsynthesize] = ACTIONS(5330),
    [anon_sym_ATdynamic] = ACTIONS(5330),
    [anon_sym_BOOL] = ACTIONS(5326),
    [anon_sym_IMP] = ACTIONS(5326),
    [anon_sym_SEL] = ACTIONS(5326),
    [anon_sym_Class] = ACTIONS(5326),
    [anon_sym_id] = ACTIONS(5326),
  },
  [STATE(2075)] = {
    [sym_identifier] = ACTIONS(2828),
    [anon_sym_COMMA] = ACTIONS(2830),
    [anon_sym_RPAREN] = ACTIONS(2830),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2830),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2830),
    [anon_sym_LPAREN2] = ACTIONS(2830),
    [anon_sym_STAR] = ACTIONS(2830),
    [anon_sym_CARET] = ACTIONS(2830),
    [anon_sym_GT] = ACTIONS(2830),
    [anon_sym_SEMI] = ACTIONS(2830),
    [anon_sym___extension__] = ACTIONS(2828),
    [anon_sym_extern] = ACTIONS(2828),
    [anon_sym___attribute__] = ACTIONS(2828),
    [anon_sym___attribute] = ACTIONS(2828),
    [anon_sym_noreturn] = ACTIONS(2828),
    [anon_sym_LBRACK] = ACTIONS(2830),
    [anon_sym___declspec] = ACTIONS(2828),
    [anon_sym___based] = ACTIONS(2828),
    [anon_sym_LBRACE] = ACTIONS(2830),
    [anon_sym_RBRACE] = ACTIONS(2830),
    [anon_sym_signed] = ACTIONS(2828),
    [anon_sym_unsigned] = ACTIONS(2828),
    [anon_sym_long] = ACTIONS(2828),
    [anon_sym_short] = ACTIONS(2828),
    [anon_sym_static] = ACTIONS(2828),
    [anon_sym_EQ] = ACTIONS(2830),
    [anon_sym_ATautoreleasepool] = ACTIONS(2830),
    [anon_sym_auto] = ACTIONS(2828),
    [anon_sym_register] = ACTIONS(2828),
    [anon_sym_inline] = ACTIONS(2828),
    [anon_sym___inline] = ACTIONS(2828),
    [anon_sym___inline__] = ACTIONS(2828),
    [anon_sym___forceinline] = ACTIONS(2828),
    [anon_sym_thread_local] = ACTIONS(2828),
    [anon_sym___thread] = ACTIONS(2828),
    [anon_sym_CG_EXTERN] = ACTIONS(2828),
    [anon_sym_CG_INLINE] = ACTIONS(2828),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2828),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2828),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2828),
    [anon_sym_IBOutlet] = ACTIONS(2828),
    [anon_sym_IBInspectable] = ACTIONS(2828),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2828),
    [anon_sym_NS_INLINE] = ACTIONS(2828),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2828),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2828),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2828),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2828),
    [anon_sym_const] = ACTIONS(2828),
    [anon_sym_constexpr] = ACTIONS(2828),
    [anon_sym_volatile] = ACTIONS(2828),
    [anon_sym_restrict] = ACTIONS(2828),
    [anon_sym___restrict__] = ACTIONS(2828),
    [anon_sym__Atomic] = ACTIONS(2828),
    [anon_sym__Noreturn] = ACTIONS(2828),
    [anon_sym__Nonnull] = ACTIONS(2828),
    [anon_sym_nullable] = ACTIONS(2828),
    [anon_sym__Complex] = ACTIONS(2828),
    [anon_sym__Nullable] = ACTIONS(2828),
    [anon_sym__Nullable_result] = ACTIONS(2828),
    [anon_sym__Null_unspecified] = ACTIONS(2828),
    [anon_sym___autoreleasing] = ACTIONS(2828),
    [anon_sym___block] = ACTIONS(2828),
    [anon_sym___bridge] = ACTIONS(2828),
    [anon_sym___bridge_retained] = ACTIONS(2828),
    [anon_sym___bridge_transfer] = ACTIONS(2828),
    [anon_sym___complex] = ACTIONS(2828),
    [anon_sym___const] = ACTIONS(2828),
    [anon_sym___imag] = ACTIONS(2828),
    [anon_sym___kindof] = ACTIONS(2828),
    [anon_sym___nonnull] = ACTIONS(2828),
    [anon_sym___nullable] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2828),
    [anon_sym___real] = ACTIONS(2828),
    [anon_sym___strong] = ACTIONS(2828),
    [anon_sym___unsafe_unretained] = ACTIONS(2828),
    [anon_sym___unused] = ACTIONS(2828),
    [anon_sym___weak] = ACTIONS(2828),
    [anon_sym_alignas] = ACTIONS(2828),
    [anon_sym__Alignas] = ACTIONS(2828),
    [sym_primitive_type] = ACTIONS(2828),
    [anon_sym_enum] = ACTIONS(2828),
    [anon_sym_COLON] = ACTIONS(2830),
    [anon_sym_struct] = ACTIONS(2828),
    [anon_sym_union] = ACTIONS(2828),
    [anon_sym_asm] = ACTIONS(2828),
    [anon_sym___asm__] = ACTIONS(2828),
    [anon_sym___asm] = ACTIONS(2828),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2828),
    [anon_sym___typeof] = ACTIONS(2828),
    [anon_sym_typeof] = ACTIONS(2828),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2828),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2828),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2828),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2828),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2828),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2828),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2828),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_API_AVAILABLE] = ACTIONS(2828),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_API_DEPRECATED] = ACTIONS(2828),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2828),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2828),
    [anon_sym___deprecated_msg] = ACTIONS(2828),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2828),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2828),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2828),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2828),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2828),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2828),
    [anon_sym_ATprotocol] = ACTIONS(2830),
    [anon_sym_ATinterface] = ACTIONS(2830),
    [anon_sym_ATimplementation] = ACTIONS(2830),
    [anon_sym_ATproperty] = ACTIONS(2830),
    [anon_sym_BOOL] = ACTIONS(2828),
    [anon_sym_IMP] = ACTIONS(2828),
    [anon_sym_SEL] = ACTIONS(2828),
    [anon_sym_Class] = ACTIONS(2828),
    [anon_sym_id] = ACTIONS(2828),
  },
  [STATE(2076)] = {
    [sym_compound_statement] = STATE(5317),
    [sym_type_qualifier] = STATE(3320),
    [sym_alignas_qualifier] = STATE(4149),
    [sym_expression] = STATE(4407),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_objc_bridge] = STATE(5947),
    [sym_typeof_specifier] = STATE(5317),
    [sym_availability] = STATE(5947),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5080),
    [anon_sym_RPAREN] = ACTIONS(5082),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5084),
    [anon_sym_noreturn] = ACTIONS(5092),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(5090),
    [anon_sym_constexpr] = ACTIONS(5092),
    [anon_sym_volatile] = ACTIONS(5092),
    [anon_sym_restrict] = ACTIONS(5092),
    [anon_sym___restrict__] = ACTIONS(5092),
    [anon_sym__Atomic] = ACTIONS(5092),
    [anon_sym__Noreturn] = ACTIONS(5092),
    [anon_sym__Nonnull] = ACTIONS(5092),
    [anon_sym_nullable] = ACTIONS(5092),
    [anon_sym__Complex] = ACTIONS(5092),
    [anon_sym__Nullable] = ACTIONS(5092),
    [anon_sym__Nullable_result] = ACTIONS(5092),
    [anon_sym__Null_unspecified] = ACTIONS(5092),
    [anon_sym___autoreleasing] = ACTIONS(5092),
    [anon_sym___block] = ACTIONS(5092),
    [anon_sym___bridge] = ACTIONS(5092),
    [anon_sym___bridge_retained] = ACTIONS(5092),
    [anon_sym___bridge_transfer] = ACTIONS(5092),
    [anon_sym___complex] = ACTIONS(5092),
    [anon_sym___const] = ACTIONS(5092),
    [anon_sym___imag] = ACTIONS(5094),
    [anon_sym___kindof] = ACTIONS(5092),
    [anon_sym___nonnull] = ACTIONS(5092),
    [anon_sym___nullable] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5092),
    [anon_sym___real] = ACTIONS(5094),
    [anon_sym___strong] = ACTIONS(5092),
    [anon_sym___unsafe_unretained] = ACTIONS(5092),
    [anon_sym___unused] = ACTIONS(5092),
    [anon_sym___weak] = ACTIONS(5092),
    [anon_sym_alignas] = ACTIONS(5096),
    [anon_sym__Alignas] = ACTIONS(5096),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_objc_bridge_related] = ACTIONS(5098),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_availability] = ACTIONS(5100),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2077)] = {
    [sym_identifier] = ACTIONS(5210),
    [aux_sym_preproc_def_token1] = ACTIONS(5210),
    [aux_sym_preproc_if_token1] = ACTIONS(5210),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5210),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5210),
    [sym_preproc_directive] = ACTIONS(5210),
    [anon_sym_LPAREN2] = ACTIONS(5212),
    [anon_sym_DASH] = ACTIONS(5212),
    [anon_sym_PLUS] = ACTIONS(5212),
    [anon_sym_LT] = ACTIONS(5212),
    [anon_sym___extension__] = ACTIONS(5210),
    [anon_sym_typedef] = ACTIONS(5210),
    [anon_sym_extern] = ACTIONS(5210),
    [anon_sym___attribute__] = ACTIONS(5210),
    [anon_sym___attribute] = ACTIONS(5210),
    [anon_sym_noreturn] = ACTIONS(5210),
    [anon_sym_LBRACK] = ACTIONS(5212),
    [anon_sym___declspec] = ACTIONS(5210),
    [anon_sym___cdecl] = ACTIONS(5210),
    [anon_sym___clrcall] = ACTIONS(5210),
    [anon_sym___stdcall] = ACTIONS(5210),
    [anon_sym___fastcall] = ACTIONS(5210),
    [anon_sym___thiscall] = ACTIONS(5210),
    [anon_sym___vectorcall] = ACTIONS(5210),
    [anon_sym_LBRACE] = ACTIONS(5212),
    [anon_sym_signed] = ACTIONS(5210),
    [anon_sym_unsigned] = ACTIONS(5210),
    [anon_sym_long] = ACTIONS(5210),
    [anon_sym_short] = ACTIONS(5210),
    [anon_sym_static] = ACTIONS(5210),
    [anon_sym_auto] = ACTIONS(5210),
    [anon_sym_register] = ACTIONS(5210),
    [anon_sym_inline] = ACTIONS(5210),
    [anon_sym___inline] = ACTIONS(5210),
    [anon_sym___inline__] = ACTIONS(5210),
    [anon_sym___forceinline] = ACTIONS(5210),
    [anon_sym_thread_local] = ACTIONS(5210),
    [anon_sym___thread] = ACTIONS(5210),
    [anon_sym_CG_EXTERN] = ACTIONS(5210),
    [anon_sym_CG_INLINE] = ACTIONS(5210),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5210),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5210),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5210),
    [anon_sym_IBOutlet] = ACTIONS(5210),
    [anon_sym_IBInspectable] = ACTIONS(5210),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5210),
    [anon_sym_NS_INLINE] = ACTIONS(5210),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5210),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5210),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5210),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5210),
    [anon_sym_const] = ACTIONS(5210),
    [anon_sym_constexpr] = ACTIONS(5210),
    [anon_sym_volatile] = ACTIONS(5210),
    [anon_sym_restrict] = ACTIONS(5210),
    [anon_sym___restrict__] = ACTIONS(5210),
    [anon_sym__Atomic] = ACTIONS(5210),
    [anon_sym__Noreturn] = ACTIONS(5210),
    [anon_sym__Nonnull] = ACTIONS(5210),
    [anon_sym_nullable] = ACTIONS(5210),
    [anon_sym__Complex] = ACTIONS(5210),
    [anon_sym__Nullable] = ACTIONS(5210),
    [anon_sym__Nullable_result] = ACTIONS(5210),
    [anon_sym__Null_unspecified] = ACTIONS(5210),
    [anon_sym___autoreleasing] = ACTIONS(5210),
    [anon_sym___block] = ACTIONS(5210),
    [anon_sym___bridge] = ACTIONS(5210),
    [anon_sym___bridge_retained] = ACTIONS(5210),
    [anon_sym___bridge_transfer] = ACTIONS(5210),
    [anon_sym___complex] = ACTIONS(5210),
    [anon_sym___const] = ACTIONS(5210),
    [anon_sym___imag] = ACTIONS(5210),
    [anon_sym___kindof] = ACTIONS(5210),
    [anon_sym___nonnull] = ACTIONS(5210),
    [anon_sym___nullable] = ACTIONS(5210),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5210),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5210),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5210),
    [anon_sym___real] = ACTIONS(5210),
    [anon_sym___strong] = ACTIONS(5210),
    [anon_sym___unsafe_unretained] = ACTIONS(5210),
    [anon_sym___unused] = ACTIONS(5210),
    [anon_sym___weak] = ACTIONS(5210),
    [anon_sym_alignas] = ACTIONS(5210),
    [anon_sym__Alignas] = ACTIONS(5210),
    [sym_primitive_type] = ACTIONS(5210),
    [anon_sym_enum] = ACTIONS(5210),
    [anon_sym_COLON] = ACTIONS(5212),
    [anon_sym_struct] = ACTIONS(5210),
    [anon_sym_union] = ACTIONS(5210),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5210),
    [anon_sym___typeof] = ACTIONS(5210),
    [anon_sym_typeof] = ACTIONS(5210),
    [aux_sym_preproc_undef_token1] = ACTIONS(5210),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5210),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5210),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5210),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5210),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5210),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5210),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5210),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5210),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5210),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5210),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5210),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5210),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5210),
    [anon_sym_API_AVAILABLE] = ACTIONS(5210),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5210),
    [anon_sym_API_DEPRECATED] = ACTIONS(5210),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5210),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5210),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5210),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5210),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5210),
    [anon_sym___deprecated_msg] = ACTIONS(5210),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5210),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5210),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5210),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5210),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5210),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5210),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5210),
    [anon_sym_ATend] = ACTIONS(5212),
    [anon_sym_ATsynthesize] = ACTIONS(5212),
    [anon_sym_ATdynamic] = ACTIONS(5212),
    [anon_sym_BOOL] = ACTIONS(5210),
    [anon_sym_IMP] = ACTIONS(5210),
    [anon_sym_SEL] = ACTIONS(5210),
    [anon_sym_Class] = ACTIONS(5210),
    [anon_sym_id] = ACTIONS(5210),
  },
  [STATE(2078)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5266),
    [anon_sym_COMMA] = ACTIONS(4476),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(5272),
    [anon_sym_AMP_AMP] = ACTIONS(5274),
    [anon_sym_PIPE] = ACTIONS(5276),
    [anon_sym_CARET] = ACTIONS(5278),
    [anon_sym_AMP] = ACTIONS(5280),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4476),
    [anon_sym___extension__] = ACTIONS(4476),
    [anon_sym_extern] = ACTIONS(4476),
    [anon_sym___attribute__] = ACTIONS(4476),
    [anon_sym___attribute] = ACTIONS(4474),
    [anon_sym_noreturn] = ACTIONS(4476),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4476),
    [anon_sym_static] = ACTIONS(4476),
    [anon_sym_EQ] = ACTIONS(4474),
    [anon_sym_auto] = ACTIONS(4476),
    [anon_sym_register] = ACTIONS(4476),
    [anon_sym_inline] = ACTIONS(4476),
    [anon_sym___inline] = ACTIONS(4474),
    [anon_sym___inline__] = ACTIONS(4476),
    [anon_sym___forceinline] = ACTIONS(4476),
    [anon_sym_thread_local] = ACTIONS(4476),
    [anon_sym___thread] = ACTIONS(4476),
    [anon_sym_CG_EXTERN] = ACTIONS(4476),
    [anon_sym_CG_INLINE] = ACTIONS(4476),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4476),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4476),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4476),
    [anon_sym_IBOutlet] = ACTIONS(4476),
    [anon_sym_IBInspectable] = ACTIONS(4476),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4476),
    [anon_sym_NS_INLINE] = ACTIONS(4476),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4476),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4476),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4476),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4476),
    [anon_sym_const] = ACTIONS(4474),
    [anon_sym_constexpr] = ACTIONS(4476),
    [anon_sym_volatile] = ACTIONS(4476),
    [anon_sym_restrict] = ACTIONS(4476),
    [anon_sym___restrict__] = ACTIONS(4476),
    [anon_sym__Atomic] = ACTIONS(4476),
    [anon_sym__Noreturn] = ACTIONS(4476),
    [anon_sym__Nonnull] = ACTIONS(4476),
    [anon_sym_nullable] = ACTIONS(4476),
    [anon_sym__Complex] = ACTIONS(4476),
    [anon_sym__Nullable] = ACTIONS(4474),
    [anon_sym__Nullable_result] = ACTIONS(4476),
    [anon_sym__Null_unspecified] = ACTIONS(4476),
    [anon_sym___autoreleasing] = ACTIONS(4476),
    [anon_sym___block] = ACTIONS(4476),
    [anon_sym___bridge] = ACTIONS(4474),
    [anon_sym___bridge_retained] = ACTIONS(4476),
    [anon_sym___bridge_transfer] = ACTIONS(4476),
    [anon_sym___complex] = ACTIONS(4476),
    [anon_sym___const] = ACTIONS(4476),
    [anon_sym___imag] = ACTIONS(4476),
    [anon_sym___kindof] = ACTIONS(4476),
    [anon_sym___nonnull] = ACTIONS(4476),
    [anon_sym___nullable] = ACTIONS(4476),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4476),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4476),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4476),
    [anon_sym___real] = ACTIONS(4476),
    [anon_sym___strong] = ACTIONS(4476),
    [anon_sym___unsafe_unretained] = ACTIONS(4476),
    [anon_sym___unused] = ACTIONS(4476),
    [anon_sym___weak] = ACTIONS(4476),
    [anon_sym_alignas] = ACTIONS(4476),
    [anon_sym__Alignas] = ACTIONS(4476),
    [anon_sym_QMARK] = ACTIONS(4476),
    [anon_sym_STAR_EQ] = ACTIONS(4476),
    [anon_sym_SLASH_EQ] = ACTIONS(4476),
    [anon_sym_PERCENT_EQ] = ACTIONS(4476),
    [anon_sym_PLUS_EQ] = ACTIONS(4476),
    [anon_sym_DASH_EQ] = ACTIONS(4476),
    [anon_sym_LT_LT_EQ] = ACTIONS(4476),
    [anon_sym_GT_GT_EQ] = ACTIONS(4476),
    [anon_sym_AMP_EQ] = ACTIONS(4476),
    [anon_sym_CARET_EQ] = ACTIONS(4476),
    [anon_sym_PIPE_EQ] = ACTIONS(4476),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4476),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4476),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4476),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4476),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4476),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4476),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4476),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4476),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4476),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4476),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4476),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4476),
    [anon_sym_API_AVAILABLE] = ACTIONS(4476),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4476),
    [anon_sym_API_DEPRECATED] = ACTIONS(4476),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4476),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4476),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4476),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4476),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4476),
    [anon_sym___deprecated_msg] = ACTIONS(4476),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4476),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4476),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4476),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4476),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4476),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4476),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4476),
  },
  [STATE(2079)] = {
    [sym_identifier] = ACTIONS(5332),
    [aux_sym_preproc_def_token1] = ACTIONS(5332),
    [aux_sym_preproc_if_token1] = ACTIONS(5332),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5332),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5332),
    [sym_preproc_directive] = ACTIONS(5332),
    [anon_sym_LPAREN2] = ACTIONS(5334),
    [anon_sym_DASH] = ACTIONS(5334),
    [anon_sym_PLUS] = ACTIONS(5334),
    [anon_sym_LT] = ACTIONS(5334),
    [anon_sym___extension__] = ACTIONS(5332),
    [anon_sym_typedef] = ACTIONS(5332),
    [anon_sym_extern] = ACTIONS(5332),
    [anon_sym___attribute__] = ACTIONS(5332),
    [anon_sym___attribute] = ACTIONS(5332),
    [anon_sym_noreturn] = ACTIONS(5332),
    [anon_sym_LBRACK] = ACTIONS(5334),
    [anon_sym___declspec] = ACTIONS(5332),
    [anon_sym___cdecl] = ACTIONS(5332),
    [anon_sym___clrcall] = ACTIONS(5332),
    [anon_sym___stdcall] = ACTIONS(5332),
    [anon_sym___fastcall] = ACTIONS(5332),
    [anon_sym___thiscall] = ACTIONS(5332),
    [anon_sym___vectorcall] = ACTIONS(5332),
    [anon_sym_LBRACE] = ACTIONS(5334),
    [anon_sym_signed] = ACTIONS(5332),
    [anon_sym_unsigned] = ACTIONS(5332),
    [anon_sym_long] = ACTIONS(5332),
    [anon_sym_short] = ACTIONS(5332),
    [anon_sym_static] = ACTIONS(5332),
    [anon_sym_auto] = ACTIONS(5332),
    [anon_sym_register] = ACTIONS(5332),
    [anon_sym_inline] = ACTIONS(5332),
    [anon_sym___inline] = ACTIONS(5332),
    [anon_sym___inline__] = ACTIONS(5332),
    [anon_sym___forceinline] = ACTIONS(5332),
    [anon_sym_thread_local] = ACTIONS(5332),
    [anon_sym___thread] = ACTIONS(5332),
    [anon_sym_CG_EXTERN] = ACTIONS(5332),
    [anon_sym_CG_INLINE] = ACTIONS(5332),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5332),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5332),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5332),
    [anon_sym_IBOutlet] = ACTIONS(5332),
    [anon_sym_IBInspectable] = ACTIONS(5332),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5332),
    [anon_sym_NS_INLINE] = ACTIONS(5332),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5332),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5332),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5332),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5332),
    [anon_sym_const] = ACTIONS(5332),
    [anon_sym_constexpr] = ACTIONS(5332),
    [anon_sym_volatile] = ACTIONS(5332),
    [anon_sym_restrict] = ACTIONS(5332),
    [anon_sym___restrict__] = ACTIONS(5332),
    [anon_sym__Atomic] = ACTIONS(5332),
    [anon_sym__Noreturn] = ACTIONS(5332),
    [anon_sym__Nonnull] = ACTIONS(5332),
    [anon_sym_nullable] = ACTIONS(5332),
    [anon_sym__Complex] = ACTIONS(5332),
    [anon_sym__Nullable] = ACTIONS(5332),
    [anon_sym__Nullable_result] = ACTIONS(5332),
    [anon_sym__Null_unspecified] = ACTIONS(5332),
    [anon_sym___autoreleasing] = ACTIONS(5332),
    [anon_sym___block] = ACTIONS(5332),
    [anon_sym___bridge] = ACTIONS(5332),
    [anon_sym___bridge_retained] = ACTIONS(5332),
    [anon_sym___bridge_transfer] = ACTIONS(5332),
    [anon_sym___complex] = ACTIONS(5332),
    [anon_sym___const] = ACTIONS(5332),
    [anon_sym___imag] = ACTIONS(5332),
    [anon_sym___kindof] = ACTIONS(5332),
    [anon_sym___nonnull] = ACTIONS(5332),
    [anon_sym___nullable] = ACTIONS(5332),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5332),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5332),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5332),
    [anon_sym___real] = ACTIONS(5332),
    [anon_sym___strong] = ACTIONS(5332),
    [anon_sym___unsafe_unretained] = ACTIONS(5332),
    [anon_sym___unused] = ACTIONS(5332),
    [anon_sym___weak] = ACTIONS(5332),
    [anon_sym_alignas] = ACTIONS(5332),
    [anon_sym__Alignas] = ACTIONS(5332),
    [sym_primitive_type] = ACTIONS(5332),
    [anon_sym_enum] = ACTIONS(5332),
    [anon_sym_COLON] = ACTIONS(5334),
    [anon_sym_struct] = ACTIONS(5332),
    [anon_sym_union] = ACTIONS(5332),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5332),
    [anon_sym___typeof] = ACTIONS(5332),
    [anon_sym_typeof] = ACTIONS(5332),
    [aux_sym_preproc_undef_token1] = ACTIONS(5332),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5332),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5332),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5332),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5332),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5332),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5332),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5332),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5332),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5332),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5332),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5332),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5332),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5332),
    [anon_sym_API_AVAILABLE] = ACTIONS(5332),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5332),
    [anon_sym_API_DEPRECATED] = ACTIONS(5332),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5332),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5332),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5332),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5332),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5332),
    [anon_sym___deprecated_msg] = ACTIONS(5332),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5332),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5332),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5332),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5332),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5332),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5332),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5332),
    [anon_sym_ATend] = ACTIONS(5334),
    [anon_sym_ATsynthesize] = ACTIONS(5334),
    [anon_sym_ATdynamic] = ACTIONS(5334),
    [anon_sym_BOOL] = ACTIONS(5332),
    [anon_sym_IMP] = ACTIONS(5332),
    [anon_sym_SEL] = ACTIONS(5332),
    [anon_sym_Class] = ACTIONS(5332),
    [anon_sym_id] = ACTIONS(5332),
  },
  [STATE(2080)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4808),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(2833),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2081)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_DASH] = ACTIONS(2348),
    [anon_sym_PLUS] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATend] = ACTIONS(2348),
    [anon_sym_AToptional] = ACTIONS(2348),
    [anon_sym_ATrequired] = ACTIONS(2348),
    [anon_sym_ATsynthesize] = ACTIONS(2348),
    [anon_sym_ATdynamic] = ACTIONS(2348),
    [anon_sym_ATproperty] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(2082)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_DASH] = ACTIONS(2352),
    [anon_sym_PLUS] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATend] = ACTIONS(2352),
    [anon_sym_AToptional] = ACTIONS(2352),
    [anon_sym_ATrequired] = ACTIONS(2352),
    [anon_sym_ATsynthesize] = ACTIONS(2352),
    [anon_sym_ATdynamic] = ACTIONS(2352),
    [anon_sym_ATproperty] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(2083)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_DASH] = ACTIONS(2362),
    [anon_sym_PLUS] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATend] = ACTIONS(2362),
    [anon_sym_AToptional] = ACTIONS(2362),
    [anon_sym_ATrequired] = ACTIONS(2362),
    [anon_sym_ATsynthesize] = ACTIONS(2362),
    [anon_sym_ATdynamic] = ACTIONS(2362),
    [anon_sym_ATproperty] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(2084)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4801),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_c_method_parameter] = STATE(2833),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2085)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2086)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5266),
    [anon_sym_COMMA] = ACTIONS(4468),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(5272),
    [anon_sym_AMP_AMP] = ACTIONS(5274),
    [anon_sym_PIPE] = ACTIONS(5276),
    [anon_sym_CARET] = ACTIONS(5278),
    [anon_sym_AMP] = ACTIONS(5280),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4468),
    [anon_sym___extension__] = ACTIONS(4468),
    [anon_sym_extern] = ACTIONS(4468),
    [anon_sym___attribute__] = ACTIONS(4468),
    [anon_sym___attribute] = ACTIONS(4466),
    [anon_sym_noreturn] = ACTIONS(4468),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4468),
    [anon_sym_static] = ACTIONS(4468),
    [anon_sym_EQ] = ACTIONS(4466),
    [anon_sym_auto] = ACTIONS(4468),
    [anon_sym_register] = ACTIONS(4468),
    [anon_sym_inline] = ACTIONS(4468),
    [anon_sym___inline] = ACTIONS(4466),
    [anon_sym___inline__] = ACTIONS(4468),
    [anon_sym___forceinline] = ACTIONS(4468),
    [anon_sym_thread_local] = ACTIONS(4468),
    [anon_sym___thread] = ACTIONS(4468),
    [anon_sym_CG_EXTERN] = ACTIONS(4468),
    [anon_sym_CG_INLINE] = ACTIONS(4468),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4468),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4468),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4468),
    [anon_sym_IBOutlet] = ACTIONS(4468),
    [anon_sym_IBInspectable] = ACTIONS(4468),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4468),
    [anon_sym_NS_INLINE] = ACTIONS(4468),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4468),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4468),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4468),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4468),
    [anon_sym_const] = ACTIONS(4466),
    [anon_sym_constexpr] = ACTIONS(4468),
    [anon_sym_volatile] = ACTIONS(4468),
    [anon_sym_restrict] = ACTIONS(4468),
    [anon_sym___restrict__] = ACTIONS(4468),
    [anon_sym__Atomic] = ACTIONS(4468),
    [anon_sym__Noreturn] = ACTIONS(4468),
    [anon_sym__Nonnull] = ACTIONS(4468),
    [anon_sym_nullable] = ACTIONS(4468),
    [anon_sym__Complex] = ACTIONS(4468),
    [anon_sym__Nullable] = ACTIONS(4466),
    [anon_sym__Nullable_result] = ACTIONS(4468),
    [anon_sym__Null_unspecified] = ACTIONS(4468),
    [anon_sym___autoreleasing] = ACTIONS(4468),
    [anon_sym___block] = ACTIONS(4468),
    [anon_sym___bridge] = ACTIONS(4466),
    [anon_sym___bridge_retained] = ACTIONS(4468),
    [anon_sym___bridge_transfer] = ACTIONS(4468),
    [anon_sym___complex] = ACTIONS(4468),
    [anon_sym___const] = ACTIONS(4468),
    [anon_sym___imag] = ACTIONS(4468),
    [anon_sym___kindof] = ACTIONS(4468),
    [anon_sym___nonnull] = ACTIONS(4468),
    [anon_sym___nullable] = ACTIONS(4468),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4468),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4468),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4468),
    [anon_sym___real] = ACTIONS(4468),
    [anon_sym___strong] = ACTIONS(4468),
    [anon_sym___unsafe_unretained] = ACTIONS(4468),
    [anon_sym___unused] = ACTIONS(4468),
    [anon_sym___weak] = ACTIONS(4468),
    [anon_sym_alignas] = ACTIONS(4468),
    [anon_sym__Alignas] = ACTIONS(4468),
    [anon_sym_QMARK] = ACTIONS(5290),
    [anon_sym_STAR_EQ] = ACTIONS(4468),
    [anon_sym_SLASH_EQ] = ACTIONS(4468),
    [anon_sym_PERCENT_EQ] = ACTIONS(4468),
    [anon_sym_PLUS_EQ] = ACTIONS(4468),
    [anon_sym_DASH_EQ] = ACTIONS(4468),
    [anon_sym_LT_LT_EQ] = ACTIONS(4468),
    [anon_sym_GT_GT_EQ] = ACTIONS(4468),
    [anon_sym_AMP_EQ] = ACTIONS(4468),
    [anon_sym_CARET_EQ] = ACTIONS(4468),
    [anon_sym_PIPE_EQ] = ACTIONS(4468),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4468),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4468),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4468),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4468),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4468),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4468),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4468),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4468),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4468),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4468),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4468),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4468),
    [anon_sym_API_AVAILABLE] = ACTIONS(4468),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4468),
    [anon_sym_API_DEPRECATED] = ACTIONS(4468),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4468),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4468),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4468),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4468),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4468),
    [anon_sym___deprecated_msg] = ACTIONS(4468),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4468),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4468),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4468),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4468),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4468),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4468),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4468),
  },
  [STATE(2087)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5266),
    [anon_sym_COMMA] = ACTIONS(4472),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(5272),
    [anon_sym_AMP_AMP] = ACTIONS(5274),
    [anon_sym_PIPE] = ACTIONS(5276),
    [anon_sym_CARET] = ACTIONS(5278),
    [anon_sym_AMP] = ACTIONS(5280),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4472),
    [anon_sym___extension__] = ACTIONS(4472),
    [anon_sym_extern] = ACTIONS(4472),
    [anon_sym___attribute__] = ACTIONS(4472),
    [anon_sym___attribute] = ACTIONS(4470),
    [anon_sym_noreturn] = ACTIONS(4472),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4472),
    [anon_sym_static] = ACTIONS(4472),
    [anon_sym_EQ] = ACTIONS(4470),
    [anon_sym_auto] = ACTIONS(4472),
    [anon_sym_register] = ACTIONS(4472),
    [anon_sym_inline] = ACTIONS(4472),
    [anon_sym___inline] = ACTIONS(4470),
    [anon_sym___inline__] = ACTIONS(4472),
    [anon_sym___forceinline] = ACTIONS(4472),
    [anon_sym_thread_local] = ACTIONS(4472),
    [anon_sym___thread] = ACTIONS(4472),
    [anon_sym_CG_EXTERN] = ACTIONS(4472),
    [anon_sym_CG_INLINE] = ACTIONS(4472),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4472),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4472),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4472),
    [anon_sym_IBOutlet] = ACTIONS(4472),
    [anon_sym_IBInspectable] = ACTIONS(4472),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4472),
    [anon_sym_NS_INLINE] = ACTIONS(4472),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4472),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4472),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4472),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4472),
    [anon_sym_const] = ACTIONS(4470),
    [anon_sym_constexpr] = ACTIONS(4472),
    [anon_sym_volatile] = ACTIONS(4472),
    [anon_sym_restrict] = ACTIONS(4472),
    [anon_sym___restrict__] = ACTIONS(4472),
    [anon_sym__Atomic] = ACTIONS(4472),
    [anon_sym__Noreturn] = ACTIONS(4472),
    [anon_sym__Nonnull] = ACTIONS(4472),
    [anon_sym_nullable] = ACTIONS(4472),
    [anon_sym__Complex] = ACTIONS(4472),
    [anon_sym__Nullable] = ACTIONS(4470),
    [anon_sym__Nullable_result] = ACTIONS(4472),
    [anon_sym__Null_unspecified] = ACTIONS(4472),
    [anon_sym___autoreleasing] = ACTIONS(4472),
    [anon_sym___block] = ACTIONS(4472),
    [anon_sym___bridge] = ACTIONS(4470),
    [anon_sym___bridge_retained] = ACTIONS(4472),
    [anon_sym___bridge_transfer] = ACTIONS(4472),
    [anon_sym___complex] = ACTIONS(4472),
    [anon_sym___const] = ACTIONS(4472),
    [anon_sym___imag] = ACTIONS(4472),
    [anon_sym___kindof] = ACTIONS(4472),
    [anon_sym___nonnull] = ACTIONS(4472),
    [anon_sym___nullable] = ACTIONS(4472),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4472),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4472),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4472),
    [anon_sym___real] = ACTIONS(4472),
    [anon_sym___strong] = ACTIONS(4472),
    [anon_sym___unsafe_unretained] = ACTIONS(4472),
    [anon_sym___unused] = ACTIONS(4472),
    [anon_sym___weak] = ACTIONS(4472),
    [anon_sym_alignas] = ACTIONS(4472),
    [anon_sym__Alignas] = ACTIONS(4472),
    [anon_sym_QMARK] = ACTIONS(4472),
    [anon_sym_STAR_EQ] = ACTIONS(4472),
    [anon_sym_SLASH_EQ] = ACTIONS(4472),
    [anon_sym_PERCENT_EQ] = ACTIONS(4472),
    [anon_sym_PLUS_EQ] = ACTIONS(4472),
    [anon_sym_DASH_EQ] = ACTIONS(4472),
    [anon_sym_LT_LT_EQ] = ACTIONS(4472),
    [anon_sym_GT_GT_EQ] = ACTIONS(4472),
    [anon_sym_AMP_EQ] = ACTIONS(4472),
    [anon_sym_CARET_EQ] = ACTIONS(4472),
    [anon_sym_PIPE_EQ] = ACTIONS(4472),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4472),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4472),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4472),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4472),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4472),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4472),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4472),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4472),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4472),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4472),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4472),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4472),
    [anon_sym_API_AVAILABLE] = ACTIONS(4472),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4472),
    [anon_sym_API_DEPRECATED] = ACTIONS(4472),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4472),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4472),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4472),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4472),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4472),
    [anon_sym___deprecated_msg] = ACTIONS(4472),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4472),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4472),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4472),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4472),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4472),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4472),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4472),
  },
  [STATE(2088)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4000),
    [anon_sym_PLUS] = ACTIONS(4000),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4000),
    [anon_sym_GT_GT] = ACTIONS(4000),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2089)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(5274),
    [anon_sym_PIPE] = ACTIONS(5276),
    [anon_sym_CARET] = ACTIONS(5278),
    [anon_sym_AMP] = ACTIONS(5280),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2090)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(5276),
    [anon_sym_CARET] = ACTIONS(5278),
    [anon_sym_AMP] = ACTIONS(5280),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2091)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(5278),
    [anon_sym_AMP] = ACTIONS(5280),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2092)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(5280),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2093)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2094)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2095)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2096)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4000),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4000),
    [anon_sym_GT_GT] = ACTIONS(4000),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_EQ] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_STAR_EQ] = ACTIONS(4002),
    [anon_sym_SLASH_EQ] = ACTIONS(4002),
    [anon_sym_PERCENT_EQ] = ACTIONS(4002),
    [anon_sym_PLUS_EQ] = ACTIONS(4002),
    [anon_sym_DASH_EQ] = ACTIONS(4002),
    [anon_sym_LT_LT_EQ] = ACTIONS(4002),
    [anon_sym_GT_GT_EQ] = ACTIONS(4002),
    [anon_sym_AMP_EQ] = ACTIONS(4002),
    [anon_sym_CARET_EQ] = ACTIONS(4002),
    [anon_sym_PIPE_EQ] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2097)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5266),
    [anon_sym_COMMA] = ACTIONS(4480),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(5272),
    [anon_sym_AMP_AMP] = ACTIONS(5274),
    [anon_sym_PIPE] = ACTIONS(5276),
    [anon_sym_CARET] = ACTIONS(5278),
    [anon_sym_AMP] = ACTIONS(5280),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4480),
    [anon_sym___extension__] = ACTIONS(4480),
    [anon_sym_extern] = ACTIONS(4480),
    [anon_sym___attribute__] = ACTIONS(4480),
    [anon_sym___attribute] = ACTIONS(4478),
    [anon_sym_noreturn] = ACTIONS(4480),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4480),
    [anon_sym_static] = ACTIONS(4480),
    [anon_sym_EQ] = ACTIONS(4478),
    [anon_sym_auto] = ACTIONS(4480),
    [anon_sym_register] = ACTIONS(4480),
    [anon_sym_inline] = ACTIONS(4480),
    [anon_sym___inline] = ACTIONS(4478),
    [anon_sym___inline__] = ACTIONS(4480),
    [anon_sym___forceinline] = ACTIONS(4480),
    [anon_sym_thread_local] = ACTIONS(4480),
    [anon_sym___thread] = ACTIONS(4480),
    [anon_sym_CG_EXTERN] = ACTIONS(4480),
    [anon_sym_CG_INLINE] = ACTIONS(4480),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4480),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4480),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4480),
    [anon_sym_IBOutlet] = ACTIONS(4480),
    [anon_sym_IBInspectable] = ACTIONS(4480),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4480),
    [anon_sym_NS_INLINE] = ACTIONS(4480),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4480),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4480),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4480),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4480),
    [anon_sym_const] = ACTIONS(4478),
    [anon_sym_constexpr] = ACTIONS(4480),
    [anon_sym_volatile] = ACTIONS(4480),
    [anon_sym_restrict] = ACTIONS(4480),
    [anon_sym___restrict__] = ACTIONS(4480),
    [anon_sym__Atomic] = ACTIONS(4480),
    [anon_sym__Noreturn] = ACTIONS(4480),
    [anon_sym__Nonnull] = ACTIONS(4480),
    [anon_sym_nullable] = ACTIONS(4480),
    [anon_sym__Complex] = ACTIONS(4480),
    [anon_sym__Nullable] = ACTIONS(4478),
    [anon_sym__Nullable_result] = ACTIONS(4480),
    [anon_sym__Null_unspecified] = ACTIONS(4480),
    [anon_sym___autoreleasing] = ACTIONS(4480),
    [anon_sym___block] = ACTIONS(4480),
    [anon_sym___bridge] = ACTIONS(4478),
    [anon_sym___bridge_retained] = ACTIONS(4480),
    [anon_sym___bridge_transfer] = ACTIONS(4480),
    [anon_sym___complex] = ACTIONS(4480),
    [anon_sym___const] = ACTIONS(4480),
    [anon_sym___imag] = ACTIONS(4480),
    [anon_sym___kindof] = ACTIONS(4480),
    [anon_sym___nonnull] = ACTIONS(4480),
    [anon_sym___nullable] = ACTIONS(4480),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4480),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4480),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4480),
    [anon_sym___real] = ACTIONS(4480),
    [anon_sym___strong] = ACTIONS(4480),
    [anon_sym___unsafe_unretained] = ACTIONS(4480),
    [anon_sym___unused] = ACTIONS(4480),
    [anon_sym___weak] = ACTIONS(4480),
    [anon_sym_alignas] = ACTIONS(4480),
    [anon_sym__Alignas] = ACTIONS(4480),
    [anon_sym_QMARK] = ACTIONS(5290),
    [anon_sym_STAR_EQ] = ACTIONS(4480),
    [anon_sym_SLASH_EQ] = ACTIONS(4480),
    [anon_sym_PERCENT_EQ] = ACTIONS(4480),
    [anon_sym_PLUS_EQ] = ACTIONS(4480),
    [anon_sym_DASH_EQ] = ACTIONS(4480),
    [anon_sym_LT_LT_EQ] = ACTIONS(4480),
    [anon_sym_GT_GT_EQ] = ACTIONS(4480),
    [anon_sym_AMP_EQ] = ACTIONS(4480),
    [anon_sym_CARET_EQ] = ACTIONS(4480),
    [anon_sym_PIPE_EQ] = ACTIONS(4480),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4480),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4480),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4480),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4480),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4480),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4480),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4480),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4480),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4480),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4480),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4480),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4480),
    [anon_sym_API_AVAILABLE] = ACTIONS(4480),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4480),
    [anon_sym_API_DEPRECATED] = ACTIONS(4480),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4480),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4480),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4480),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4480),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4480),
    [anon_sym___deprecated_msg] = ACTIONS(4480),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4480),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4480),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4480),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4480),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4480),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4480),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4480),
  },
  [STATE(2098)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5342),
    [anon_sym_COMMA] = ACTIONS(4436),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(5268),
    [anon_sym_PLUS] = ACTIONS(5268),
    [anon_sym_STAR] = ACTIONS(5270),
    [anon_sym_SLASH] = ACTIONS(5270),
    [anon_sym_PERCENT] = ACTIONS(5270),
    [anon_sym_PIPE_PIPE] = ACTIONS(5272),
    [anon_sym_AMP_AMP] = ACTIONS(5274),
    [anon_sym_PIPE] = ACTIONS(5276),
    [anon_sym_CARET] = ACTIONS(5278),
    [anon_sym_AMP] = ACTIONS(5280),
    [anon_sym_EQ_EQ] = ACTIONS(5282),
    [anon_sym_BANG_EQ] = ACTIONS(5282),
    [anon_sym_GT] = ACTIONS(5284),
    [anon_sym_GT_EQ] = ACTIONS(5286),
    [anon_sym_LT_EQ] = ACTIONS(5286),
    [anon_sym_LT] = ACTIONS(5284),
    [anon_sym_LT_LT] = ACTIONS(5288),
    [anon_sym_GT_GT] = ACTIONS(5288),
    [anon_sym_SEMI] = ACTIONS(4436),
    [anon_sym___extension__] = ACTIONS(4436),
    [anon_sym_extern] = ACTIONS(4436),
    [anon_sym___attribute__] = ACTIONS(4436),
    [anon_sym___attribute] = ACTIONS(4431),
    [anon_sym_noreturn] = ACTIONS(4436),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4436),
    [anon_sym_static] = ACTIONS(4436),
    [anon_sym_EQ] = ACTIONS(4431),
    [anon_sym_auto] = ACTIONS(4436),
    [anon_sym_register] = ACTIONS(4436),
    [anon_sym_inline] = ACTIONS(4436),
    [anon_sym___inline] = ACTIONS(4431),
    [anon_sym___inline__] = ACTIONS(4436),
    [anon_sym___forceinline] = ACTIONS(4436),
    [anon_sym_thread_local] = ACTIONS(4436),
    [anon_sym___thread] = ACTIONS(4436),
    [anon_sym_CG_EXTERN] = ACTIONS(4436),
    [anon_sym_CG_INLINE] = ACTIONS(4436),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4436),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4436),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4436),
    [anon_sym_IBOutlet] = ACTIONS(4436),
    [anon_sym_IBInspectable] = ACTIONS(4436),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4436),
    [anon_sym_NS_INLINE] = ACTIONS(4436),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4436),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4436),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4436),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4436),
    [anon_sym_const] = ACTIONS(4431),
    [anon_sym_constexpr] = ACTIONS(4436),
    [anon_sym_volatile] = ACTIONS(4436),
    [anon_sym_restrict] = ACTIONS(4436),
    [anon_sym___restrict__] = ACTIONS(4436),
    [anon_sym__Atomic] = ACTIONS(4436),
    [anon_sym__Noreturn] = ACTIONS(4436),
    [anon_sym__Nonnull] = ACTIONS(4436),
    [anon_sym_nullable] = ACTIONS(4436),
    [anon_sym__Complex] = ACTIONS(4436),
    [anon_sym__Nullable] = ACTIONS(4431),
    [anon_sym__Nullable_result] = ACTIONS(4436),
    [anon_sym__Null_unspecified] = ACTIONS(4436),
    [anon_sym___autoreleasing] = ACTIONS(4436),
    [anon_sym___block] = ACTIONS(4436),
    [anon_sym___bridge] = ACTIONS(4431),
    [anon_sym___bridge_retained] = ACTIONS(4436),
    [anon_sym___bridge_transfer] = ACTIONS(4436),
    [anon_sym___complex] = ACTIONS(4436),
    [anon_sym___const] = ACTIONS(4436),
    [anon_sym___imag] = ACTIONS(4436),
    [anon_sym___kindof] = ACTIONS(4436),
    [anon_sym___nonnull] = ACTIONS(4436),
    [anon_sym___nullable] = ACTIONS(4436),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4436),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4436),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4436),
    [anon_sym___real] = ACTIONS(4436),
    [anon_sym___strong] = ACTIONS(4436),
    [anon_sym___unsafe_unretained] = ACTIONS(4436),
    [anon_sym___unused] = ACTIONS(4436),
    [anon_sym___weak] = ACTIONS(4436),
    [anon_sym_alignas] = ACTIONS(4436),
    [anon_sym__Alignas] = ACTIONS(4436),
    [anon_sym_QMARK] = ACTIONS(4436),
    [anon_sym_STAR_EQ] = ACTIONS(4436),
    [anon_sym_SLASH_EQ] = ACTIONS(4436),
    [anon_sym_PERCENT_EQ] = ACTIONS(4436),
    [anon_sym_PLUS_EQ] = ACTIONS(4436),
    [anon_sym_DASH_EQ] = ACTIONS(4436),
    [anon_sym_LT_LT_EQ] = ACTIONS(4436),
    [anon_sym_GT_GT_EQ] = ACTIONS(4436),
    [anon_sym_AMP_EQ] = ACTIONS(4436),
    [anon_sym_CARET_EQ] = ACTIONS(4436),
    [anon_sym_PIPE_EQ] = ACTIONS(4436),
    [anon_sym_DASH_DASH] = ACTIONS(3992),
    [anon_sym_PLUS_PLUS] = ACTIONS(3992),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4436),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4436),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4436),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4436),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4436),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4436),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4436),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4436),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4436),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4436),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4431),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4436),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4436),
    [anon_sym_API_AVAILABLE] = ACTIONS(4436),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4436),
    [anon_sym_API_DEPRECATED] = ACTIONS(4436),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4436),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4436),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4436),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4436),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4436),
    [anon_sym___deprecated_msg] = ACTIONS(4436),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4436),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4436),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4436),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4436),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4436),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4436),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4436),
  },
  [STATE(2099)] = {
    [sym_expression] = STATE(4458),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2100)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2185),
    [sym_identifier] = ACTIONS(5345),
    [aux_sym_preproc_def_token1] = ACTIONS(5345),
    [aux_sym_preproc_if_token1] = ACTIONS(5345),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5345),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5345),
    [sym_preproc_directive] = ACTIONS(5345),
    [anon_sym_DASH] = ACTIONS(5347),
    [anon_sym_PLUS] = ACTIONS(5347),
    [anon_sym_SEMI] = ACTIONS(5349),
    [anon_sym___extension__] = ACTIONS(5345),
    [anon_sym_typedef] = ACTIONS(5345),
    [anon_sym_extern] = ACTIONS(5345),
    [anon_sym___attribute__] = ACTIONS(5345),
    [anon_sym___attribute] = ACTIONS(5345),
    [anon_sym_noreturn] = ACTIONS(5345),
    [anon_sym_LBRACK] = ACTIONS(5347),
    [anon_sym___declspec] = ACTIONS(5345),
    [anon_sym___cdecl] = ACTIONS(5345),
    [anon_sym___clrcall] = ACTIONS(5345),
    [anon_sym___stdcall] = ACTIONS(5345),
    [anon_sym___fastcall] = ACTIONS(5345),
    [anon_sym___thiscall] = ACTIONS(5345),
    [anon_sym___vectorcall] = ACTIONS(5345),
    [anon_sym_signed] = ACTIONS(5345),
    [anon_sym_unsigned] = ACTIONS(5345),
    [anon_sym_long] = ACTIONS(5345),
    [anon_sym_short] = ACTIONS(5345),
    [anon_sym_static] = ACTIONS(5345),
    [anon_sym_auto] = ACTIONS(5345),
    [anon_sym_register] = ACTIONS(5345),
    [anon_sym_inline] = ACTIONS(5345),
    [anon_sym___inline] = ACTIONS(5345),
    [anon_sym___inline__] = ACTIONS(5345),
    [anon_sym___forceinline] = ACTIONS(5345),
    [anon_sym_thread_local] = ACTIONS(5345),
    [anon_sym___thread] = ACTIONS(5345),
    [anon_sym_CG_EXTERN] = ACTIONS(5345),
    [anon_sym_CG_INLINE] = ACTIONS(5345),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5345),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5345),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5345),
    [anon_sym_IBOutlet] = ACTIONS(5345),
    [anon_sym_IBInspectable] = ACTIONS(5345),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5345),
    [anon_sym_NS_INLINE] = ACTIONS(5345),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5345),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5345),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5345),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5345),
    [anon_sym_const] = ACTIONS(5345),
    [anon_sym_constexpr] = ACTIONS(5345),
    [anon_sym_volatile] = ACTIONS(5345),
    [anon_sym_restrict] = ACTIONS(5345),
    [anon_sym___restrict__] = ACTIONS(5345),
    [anon_sym__Atomic] = ACTIONS(5345),
    [anon_sym__Noreturn] = ACTIONS(5345),
    [anon_sym__Nonnull] = ACTIONS(5345),
    [anon_sym_nullable] = ACTIONS(5345),
    [anon_sym__Complex] = ACTIONS(5345),
    [anon_sym__Nullable] = ACTIONS(5345),
    [anon_sym__Nullable_result] = ACTIONS(5345),
    [anon_sym__Null_unspecified] = ACTIONS(5345),
    [anon_sym___autoreleasing] = ACTIONS(5345),
    [anon_sym___block] = ACTIONS(5345),
    [anon_sym___bridge] = ACTIONS(5345),
    [anon_sym___bridge_retained] = ACTIONS(5345),
    [anon_sym___bridge_transfer] = ACTIONS(5345),
    [anon_sym___complex] = ACTIONS(5345),
    [anon_sym___const] = ACTIONS(5345),
    [anon_sym___imag] = ACTIONS(5345),
    [anon_sym___kindof] = ACTIONS(5345),
    [anon_sym___nonnull] = ACTIONS(5345),
    [anon_sym___nullable] = ACTIONS(5345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5345),
    [anon_sym___real] = ACTIONS(5345),
    [anon_sym___strong] = ACTIONS(5345),
    [anon_sym___unsafe_unretained] = ACTIONS(5345),
    [anon_sym___unused] = ACTIONS(5345),
    [anon_sym___weak] = ACTIONS(5345),
    [anon_sym_alignas] = ACTIONS(5345),
    [anon_sym__Alignas] = ACTIONS(5345),
    [sym_primitive_type] = ACTIONS(5345),
    [anon_sym_enum] = ACTIONS(5345),
    [anon_sym_struct] = ACTIONS(5345),
    [anon_sym_union] = ACTIONS(5345),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5345),
    [anon_sym___typeof] = ACTIONS(5345),
    [anon_sym_typeof] = ACTIONS(5345),
    [aux_sym_preproc_undef_token1] = ACTIONS(5345),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5345),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5345),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5345),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5345),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5345),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5345),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5345),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5345),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5345),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_API_AVAILABLE] = ACTIONS(5345),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_API_DEPRECATED] = ACTIONS(5345),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5345),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5345),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5345),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5345),
    [anon_sym___deprecated_msg] = ACTIONS(5345),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5345),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5345),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5345),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5345),
    [anon_sym_ATend] = ACTIONS(5347),
    [anon_sym_AToptional] = ACTIONS(5347),
    [anon_sym_ATrequired] = ACTIONS(5347),
    [anon_sym_ATproperty] = ACTIONS(5347),
    [anon_sym_BOOL] = ACTIONS(5345),
    [anon_sym_IMP] = ACTIONS(5345),
    [anon_sym_SEL] = ACTIONS(5345),
    [anon_sym_Class] = ACTIONS(5345),
    [anon_sym_id] = ACTIONS(5345),
  },
  [STATE(2101)] = {
    [sym_identifier] = ACTIONS(5351),
    [aux_sym_preproc_def_token1] = ACTIONS(5351),
    [aux_sym_preproc_if_token1] = ACTIONS(5351),
    [aux_sym_preproc_if_token2] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5351),
    [aux_sym_preproc_else_token1] = ACTIONS(5351),
    [aux_sym_preproc_elif_token1] = ACTIONS(5351),
    [sym_preproc_directive] = ACTIONS(5351),
    [anon_sym_DASH] = ACTIONS(5353),
    [anon_sym_PLUS] = ACTIONS(5353),
    [anon_sym_SEMI] = ACTIONS(5355),
    [anon_sym___extension__] = ACTIONS(5351),
    [anon_sym_typedef] = ACTIONS(5351),
    [anon_sym_extern] = ACTIONS(5351),
    [anon_sym___attribute__] = ACTIONS(5351),
    [anon_sym___attribute] = ACTIONS(5351),
    [anon_sym_noreturn] = ACTIONS(5351),
    [anon_sym_LBRACK] = ACTIONS(5353),
    [anon_sym___declspec] = ACTIONS(5351),
    [anon_sym___cdecl] = ACTIONS(5351),
    [anon_sym___clrcall] = ACTIONS(5351),
    [anon_sym___stdcall] = ACTIONS(5351),
    [anon_sym___fastcall] = ACTIONS(5351),
    [anon_sym___thiscall] = ACTIONS(5351),
    [anon_sym___vectorcall] = ACTIONS(5351),
    [anon_sym_signed] = ACTIONS(5351),
    [anon_sym_unsigned] = ACTIONS(5351),
    [anon_sym_long] = ACTIONS(5351),
    [anon_sym_short] = ACTIONS(5351),
    [anon_sym_static] = ACTIONS(5351),
    [anon_sym_auto] = ACTIONS(5351),
    [anon_sym_register] = ACTIONS(5351),
    [anon_sym_inline] = ACTIONS(5351),
    [anon_sym___inline] = ACTIONS(5351),
    [anon_sym___inline__] = ACTIONS(5351),
    [anon_sym___forceinline] = ACTIONS(5351),
    [anon_sym_thread_local] = ACTIONS(5351),
    [anon_sym___thread] = ACTIONS(5351),
    [anon_sym_CG_EXTERN] = ACTIONS(5351),
    [anon_sym_CG_INLINE] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5351),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5351),
    [anon_sym_IBOutlet] = ACTIONS(5351),
    [anon_sym_IBInspectable] = ACTIONS(5351),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5351),
    [anon_sym_NS_INLINE] = ACTIONS(5351),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5351),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5351),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5351),
    [anon_sym_const] = ACTIONS(5351),
    [anon_sym_constexpr] = ACTIONS(5351),
    [anon_sym_volatile] = ACTIONS(5351),
    [anon_sym_restrict] = ACTIONS(5351),
    [anon_sym___restrict__] = ACTIONS(5351),
    [anon_sym__Atomic] = ACTIONS(5351),
    [anon_sym__Noreturn] = ACTIONS(5351),
    [anon_sym__Nonnull] = ACTIONS(5351),
    [anon_sym_nullable] = ACTIONS(5351),
    [anon_sym__Complex] = ACTIONS(5351),
    [anon_sym__Nullable] = ACTIONS(5351),
    [anon_sym__Nullable_result] = ACTIONS(5351),
    [anon_sym__Null_unspecified] = ACTIONS(5351),
    [anon_sym___autoreleasing] = ACTIONS(5351),
    [anon_sym___block] = ACTIONS(5351),
    [anon_sym___bridge] = ACTIONS(5351),
    [anon_sym___bridge_retained] = ACTIONS(5351),
    [anon_sym___bridge_transfer] = ACTIONS(5351),
    [anon_sym___complex] = ACTIONS(5351),
    [anon_sym___const] = ACTIONS(5351),
    [anon_sym___imag] = ACTIONS(5351),
    [anon_sym___kindof] = ACTIONS(5351),
    [anon_sym___nonnull] = ACTIONS(5351),
    [anon_sym___nullable] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5351),
    [anon_sym___real] = ACTIONS(5351),
    [anon_sym___strong] = ACTIONS(5351),
    [anon_sym___unsafe_unretained] = ACTIONS(5351),
    [anon_sym___unused] = ACTIONS(5351),
    [anon_sym___weak] = ACTIONS(5351),
    [anon_sym_alignas] = ACTIONS(5351),
    [anon_sym__Alignas] = ACTIONS(5351),
    [sym_primitive_type] = ACTIONS(5351),
    [anon_sym_enum] = ACTIONS(5351),
    [anon_sym_struct] = ACTIONS(5351),
    [anon_sym_union] = ACTIONS(5351),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5351),
    [anon_sym___typeof] = ACTIONS(5351),
    [anon_sym_typeof] = ACTIONS(5351),
    [aux_sym_preproc_undef_token1] = ACTIONS(5351),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5351),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5351),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5351),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_API_AVAILABLE] = ACTIONS(5351),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_API_DEPRECATED] = ACTIONS(5351),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym___deprecated_msg] = ACTIONS(5351),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5351),
    [anon_sym_ATsynthesize] = ACTIONS(5353),
    [anon_sym_ATdynamic] = ACTIONS(5353),
    [anon_sym_BOOL] = ACTIONS(5351),
    [anon_sym_IMP] = ACTIONS(5351),
    [anon_sym_SEL] = ACTIONS(5351),
    [anon_sym_Class] = ACTIONS(5351),
    [anon_sym_id] = ACTIONS(5351),
  },
  [STATE(2102)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2138),
    [sym_identifier] = ACTIONS(5345),
    [aux_sym_preproc_def_token1] = ACTIONS(5345),
    [aux_sym_preproc_if_token1] = ACTIONS(5345),
    [aux_sym_preproc_if_token2] = ACTIONS(5345),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5345),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5345),
    [aux_sym_preproc_else_token1] = ACTIONS(5345),
    [aux_sym_preproc_elif_token1] = ACTIONS(5345),
    [sym_preproc_directive] = ACTIONS(5345),
    [anon_sym_DASH] = ACTIONS(5347),
    [anon_sym_PLUS] = ACTIONS(5347),
    [anon_sym_SEMI] = ACTIONS(5357),
    [anon_sym___extension__] = ACTIONS(5345),
    [anon_sym_typedef] = ACTIONS(5345),
    [anon_sym_extern] = ACTIONS(5345),
    [anon_sym___attribute__] = ACTIONS(5345),
    [anon_sym___attribute] = ACTIONS(5345),
    [anon_sym_noreturn] = ACTIONS(5345),
    [anon_sym_LBRACK] = ACTIONS(5347),
    [anon_sym___declspec] = ACTIONS(5345),
    [anon_sym___cdecl] = ACTIONS(5345),
    [anon_sym___clrcall] = ACTIONS(5345),
    [anon_sym___stdcall] = ACTIONS(5345),
    [anon_sym___fastcall] = ACTIONS(5345),
    [anon_sym___thiscall] = ACTIONS(5345),
    [anon_sym___vectorcall] = ACTIONS(5345),
    [anon_sym_signed] = ACTIONS(5345),
    [anon_sym_unsigned] = ACTIONS(5345),
    [anon_sym_long] = ACTIONS(5345),
    [anon_sym_short] = ACTIONS(5345),
    [anon_sym_static] = ACTIONS(5345),
    [anon_sym_auto] = ACTIONS(5345),
    [anon_sym_register] = ACTIONS(5345),
    [anon_sym_inline] = ACTIONS(5345),
    [anon_sym___inline] = ACTIONS(5345),
    [anon_sym___inline__] = ACTIONS(5345),
    [anon_sym___forceinline] = ACTIONS(5345),
    [anon_sym_thread_local] = ACTIONS(5345),
    [anon_sym___thread] = ACTIONS(5345),
    [anon_sym_CG_EXTERN] = ACTIONS(5345),
    [anon_sym_CG_INLINE] = ACTIONS(5345),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5345),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5345),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5345),
    [anon_sym_IBOutlet] = ACTIONS(5345),
    [anon_sym_IBInspectable] = ACTIONS(5345),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5345),
    [anon_sym_NS_INLINE] = ACTIONS(5345),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5345),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5345),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5345),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5345),
    [anon_sym_const] = ACTIONS(5345),
    [anon_sym_constexpr] = ACTIONS(5345),
    [anon_sym_volatile] = ACTIONS(5345),
    [anon_sym_restrict] = ACTIONS(5345),
    [anon_sym___restrict__] = ACTIONS(5345),
    [anon_sym__Atomic] = ACTIONS(5345),
    [anon_sym__Noreturn] = ACTIONS(5345),
    [anon_sym__Nonnull] = ACTIONS(5345),
    [anon_sym_nullable] = ACTIONS(5345),
    [anon_sym__Complex] = ACTIONS(5345),
    [anon_sym__Nullable] = ACTIONS(5345),
    [anon_sym__Nullable_result] = ACTIONS(5345),
    [anon_sym__Null_unspecified] = ACTIONS(5345),
    [anon_sym___autoreleasing] = ACTIONS(5345),
    [anon_sym___block] = ACTIONS(5345),
    [anon_sym___bridge] = ACTIONS(5345),
    [anon_sym___bridge_retained] = ACTIONS(5345),
    [anon_sym___bridge_transfer] = ACTIONS(5345),
    [anon_sym___complex] = ACTIONS(5345),
    [anon_sym___const] = ACTIONS(5345),
    [anon_sym___imag] = ACTIONS(5345),
    [anon_sym___kindof] = ACTIONS(5345),
    [anon_sym___nonnull] = ACTIONS(5345),
    [anon_sym___nullable] = ACTIONS(5345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5345),
    [anon_sym___real] = ACTIONS(5345),
    [anon_sym___strong] = ACTIONS(5345),
    [anon_sym___unsafe_unretained] = ACTIONS(5345),
    [anon_sym___unused] = ACTIONS(5345),
    [anon_sym___weak] = ACTIONS(5345),
    [anon_sym_alignas] = ACTIONS(5345),
    [anon_sym__Alignas] = ACTIONS(5345),
    [sym_primitive_type] = ACTIONS(5345),
    [anon_sym_enum] = ACTIONS(5345),
    [anon_sym_struct] = ACTIONS(5345),
    [anon_sym_union] = ACTIONS(5345),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5345),
    [anon_sym___typeof] = ACTIONS(5345),
    [anon_sym_typeof] = ACTIONS(5345),
    [aux_sym_preproc_undef_token1] = ACTIONS(5345),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5345),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5345),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5345),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5345),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5345),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5345),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5345),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5345),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5345),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_API_AVAILABLE] = ACTIONS(5345),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_API_DEPRECATED] = ACTIONS(5345),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5345),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5345),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5345),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5345),
    [anon_sym___deprecated_msg] = ACTIONS(5345),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5345),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5345),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5345),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5345),
    [anon_sym_ATproperty] = ACTIONS(5347),
    [anon_sym_BOOL] = ACTIONS(5345),
    [anon_sym_IMP] = ACTIONS(5345),
    [anon_sym_SEL] = ACTIONS(5345),
    [anon_sym_Class] = ACTIONS(5345),
    [anon_sym_id] = ACTIONS(5345),
  },
  [STATE(2103)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token2] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [aux_sym_preproc_else_token1] = ACTIONS(2546),
    [aux_sym_preproc_elif_token1] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_PLUS] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATproperty] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(2104)] = {
    [sym_identifier] = ACTIONS(5359),
    [aux_sym_preproc_def_token1] = ACTIONS(5359),
    [aux_sym_preproc_if_token1] = ACTIONS(5359),
    [aux_sym_preproc_if_token2] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5359),
    [aux_sym_preproc_else_token1] = ACTIONS(5359),
    [aux_sym_preproc_elif_token1] = ACTIONS(5359),
    [sym_preproc_directive] = ACTIONS(5359),
    [anon_sym_DASH] = ACTIONS(5361),
    [anon_sym_PLUS] = ACTIONS(5361),
    [anon_sym_SEMI] = ACTIONS(5363),
    [anon_sym___extension__] = ACTIONS(5359),
    [anon_sym_typedef] = ACTIONS(5359),
    [anon_sym_extern] = ACTIONS(5359),
    [anon_sym___attribute__] = ACTIONS(5359),
    [anon_sym___attribute] = ACTIONS(5359),
    [anon_sym_noreturn] = ACTIONS(5359),
    [anon_sym_LBRACK] = ACTIONS(5361),
    [anon_sym___declspec] = ACTIONS(5359),
    [anon_sym___cdecl] = ACTIONS(5359),
    [anon_sym___clrcall] = ACTIONS(5359),
    [anon_sym___stdcall] = ACTIONS(5359),
    [anon_sym___fastcall] = ACTIONS(5359),
    [anon_sym___thiscall] = ACTIONS(5359),
    [anon_sym___vectorcall] = ACTIONS(5359),
    [anon_sym_signed] = ACTIONS(5359),
    [anon_sym_unsigned] = ACTIONS(5359),
    [anon_sym_long] = ACTIONS(5359),
    [anon_sym_short] = ACTIONS(5359),
    [anon_sym_static] = ACTIONS(5359),
    [anon_sym_auto] = ACTIONS(5359),
    [anon_sym_register] = ACTIONS(5359),
    [anon_sym_inline] = ACTIONS(5359),
    [anon_sym___inline] = ACTIONS(5359),
    [anon_sym___inline__] = ACTIONS(5359),
    [anon_sym___forceinline] = ACTIONS(5359),
    [anon_sym_thread_local] = ACTIONS(5359),
    [anon_sym___thread] = ACTIONS(5359),
    [anon_sym_CG_EXTERN] = ACTIONS(5359),
    [anon_sym_CG_INLINE] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5359),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5359),
    [anon_sym_IBOutlet] = ACTIONS(5359),
    [anon_sym_IBInspectable] = ACTIONS(5359),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5359),
    [anon_sym_NS_INLINE] = ACTIONS(5359),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5359),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5359),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5359),
    [anon_sym_const] = ACTIONS(5359),
    [anon_sym_constexpr] = ACTIONS(5359),
    [anon_sym_volatile] = ACTIONS(5359),
    [anon_sym_restrict] = ACTIONS(5359),
    [anon_sym___restrict__] = ACTIONS(5359),
    [anon_sym__Atomic] = ACTIONS(5359),
    [anon_sym__Noreturn] = ACTIONS(5359),
    [anon_sym__Nonnull] = ACTIONS(5359),
    [anon_sym_nullable] = ACTIONS(5359),
    [anon_sym__Complex] = ACTIONS(5359),
    [anon_sym__Nullable] = ACTIONS(5359),
    [anon_sym__Nullable_result] = ACTIONS(5359),
    [anon_sym__Null_unspecified] = ACTIONS(5359),
    [anon_sym___autoreleasing] = ACTIONS(5359),
    [anon_sym___block] = ACTIONS(5359),
    [anon_sym___bridge] = ACTIONS(5359),
    [anon_sym___bridge_retained] = ACTIONS(5359),
    [anon_sym___bridge_transfer] = ACTIONS(5359),
    [anon_sym___complex] = ACTIONS(5359),
    [anon_sym___const] = ACTIONS(5359),
    [anon_sym___imag] = ACTIONS(5359),
    [anon_sym___kindof] = ACTIONS(5359),
    [anon_sym___nonnull] = ACTIONS(5359),
    [anon_sym___nullable] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5359),
    [anon_sym___real] = ACTIONS(5359),
    [anon_sym___strong] = ACTIONS(5359),
    [anon_sym___unsafe_unretained] = ACTIONS(5359),
    [anon_sym___unused] = ACTIONS(5359),
    [anon_sym___weak] = ACTIONS(5359),
    [anon_sym_alignas] = ACTIONS(5359),
    [anon_sym__Alignas] = ACTIONS(5359),
    [sym_primitive_type] = ACTIONS(5359),
    [anon_sym_enum] = ACTIONS(5359),
    [anon_sym_struct] = ACTIONS(5359),
    [anon_sym_union] = ACTIONS(5359),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5359),
    [anon_sym___typeof] = ACTIONS(5359),
    [anon_sym_typeof] = ACTIONS(5359),
    [aux_sym_preproc_undef_token1] = ACTIONS(5359),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5359),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5359),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5359),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_API_AVAILABLE] = ACTIONS(5359),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_API_DEPRECATED] = ACTIONS(5359),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym___deprecated_msg] = ACTIONS(5359),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5359),
    [anon_sym_ATsynthesize] = ACTIONS(5361),
    [anon_sym_ATdynamic] = ACTIONS(5361),
    [anon_sym_BOOL] = ACTIONS(5359),
    [anon_sym_IMP] = ACTIONS(5359),
    [anon_sym_SEL] = ACTIONS(5359),
    [anon_sym_Class] = ACTIONS(5359),
    [anon_sym_id] = ACTIONS(5359),
  },
  [STATE(2105)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2138),
    [sym_identifier] = ACTIONS(5365),
    [aux_sym_preproc_def_token1] = ACTIONS(5365),
    [aux_sym_preproc_if_token1] = ACTIONS(5365),
    [aux_sym_preproc_if_token2] = ACTIONS(5365),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5365),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5365),
    [aux_sym_preproc_else_token1] = ACTIONS(5365),
    [aux_sym_preproc_elif_token1] = ACTIONS(5365),
    [sym_preproc_directive] = ACTIONS(5365),
    [anon_sym_DASH] = ACTIONS(5367),
    [anon_sym_PLUS] = ACTIONS(5367),
    [anon_sym_SEMI] = ACTIONS(5357),
    [anon_sym___extension__] = ACTIONS(5365),
    [anon_sym_typedef] = ACTIONS(5365),
    [anon_sym_extern] = ACTIONS(5365),
    [anon_sym___attribute__] = ACTIONS(5365),
    [anon_sym___attribute] = ACTIONS(5365),
    [anon_sym_noreturn] = ACTIONS(5365),
    [anon_sym_LBRACK] = ACTIONS(5367),
    [anon_sym___declspec] = ACTIONS(5365),
    [anon_sym___cdecl] = ACTIONS(5365),
    [anon_sym___clrcall] = ACTIONS(5365),
    [anon_sym___stdcall] = ACTIONS(5365),
    [anon_sym___fastcall] = ACTIONS(5365),
    [anon_sym___thiscall] = ACTIONS(5365),
    [anon_sym___vectorcall] = ACTIONS(5365),
    [anon_sym_signed] = ACTIONS(5365),
    [anon_sym_unsigned] = ACTIONS(5365),
    [anon_sym_long] = ACTIONS(5365),
    [anon_sym_short] = ACTIONS(5365),
    [anon_sym_static] = ACTIONS(5365),
    [anon_sym_auto] = ACTIONS(5365),
    [anon_sym_register] = ACTIONS(5365),
    [anon_sym_inline] = ACTIONS(5365),
    [anon_sym___inline] = ACTIONS(5365),
    [anon_sym___inline__] = ACTIONS(5365),
    [anon_sym___forceinline] = ACTIONS(5365),
    [anon_sym_thread_local] = ACTIONS(5365),
    [anon_sym___thread] = ACTIONS(5365),
    [anon_sym_CG_EXTERN] = ACTIONS(5365),
    [anon_sym_CG_INLINE] = ACTIONS(5365),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5365),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5365),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5365),
    [anon_sym_IBOutlet] = ACTIONS(5365),
    [anon_sym_IBInspectable] = ACTIONS(5365),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5365),
    [anon_sym_NS_INLINE] = ACTIONS(5365),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5365),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5365),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5365),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5365),
    [anon_sym_const] = ACTIONS(5365),
    [anon_sym_constexpr] = ACTIONS(5365),
    [anon_sym_volatile] = ACTIONS(5365),
    [anon_sym_restrict] = ACTIONS(5365),
    [anon_sym___restrict__] = ACTIONS(5365),
    [anon_sym__Atomic] = ACTIONS(5365),
    [anon_sym__Noreturn] = ACTIONS(5365),
    [anon_sym__Nonnull] = ACTIONS(5365),
    [anon_sym_nullable] = ACTIONS(5365),
    [anon_sym__Complex] = ACTIONS(5365),
    [anon_sym__Nullable] = ACTIONS(5365),
    [anon_sym__Nullable_result] = ACTIONS(5365),
    [anon_sym__Null_unspecified] = ACTIONS(5365),
    [anon_sym___autoreleasing] = ACTIONS(5365),
    [anon_sym___block] = ACTIONS(5365),
    [anon_sym___bridge] = ACTIONS(5365),
    [anon_sym___bridge_retained] = ACTIONS(5365),
    [anon_sym___bridge_transfer] = ACTIONS(5365),
    [anon_sym___complex] = ACTIONS(5365),
    [anon_sym___const] = ACTIONS(5365),
    [anon_sym___imag] = ACTIONS(5365),
    [anon_sym___kindof] = ACTIONS(5365),
    [anon_sym___nonnull] = ACTIONS(5365),
    [anon_sym___nullable] = ACTIONS(5365),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5365),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5365),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5365),
    [anon_sym___real] = ACTIONS(5365),
    [anon_sym___strong] = ACTIONS(5365),
    [anon_sym___unsafe_unretained] = ACTIONS(5365),
    [anon_sym___unused] = ACTIONS(5365),
    [anon_sym___weak] = ACTIONS(5365),
    [anon_sym_alignas] = ACTIONS(5365),
    [anon_sym__Alignas] = ACTIONS(5365),
    [sym_primitive_type] = ACTIONS(5365),
    [anon_sym_enum] = ACTIONS(5365),
    [anon_sym_struct] = ACTIONS(5365),
    [anon_sym_union] = ACTIONS(5365),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5365),
    [anon_sym___typeof] = ACTIONS(5365),
    [anon_sym_typeof] = ACTIONS(5365),
    [aux_sym_preproc_undef_token1] = ACTIONS(5365),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5365),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5365),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5365),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5365),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5365),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5365),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5365),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5365),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5365),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_API_AVAILABLE] = ACTIONS(5365),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_API_DEPRECATED] = ACTIONS(5365),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5365),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5365),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5365),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5365),
    [anon_sym___deprecated_msg] = ACTIONS(5365),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5365),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5365),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5365),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5365),
    [anon_sym_ATproperty] = ACTIONS(5367),
    [anon_sym_BOOL] = ACTIONS(5365),
    [anon_sym_IMP] = ACTIONS(5365),
    [anon_sym_SEL] = ACTIONS(5365),
    [anon_sym_Class] = ACTIONS(5365),
    [anon_sym_id] = ACTIONS(5365),
  },
  [STATE(2106)] = {
    [sym_identifier] = ACTIONS(5369),
    [aux_sym_preproc_def_token1] = ACTIONS(5369),
    [aux_sym_preproc_if_token1] = ACTIONS(5369),
    [aux_sym_preproc_if_token2] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5369),
    [aux_sym_preproc_else_token1] = ACTIONS(5369),
    [aux_sym_preproc_elif_token1] = ACTIONS(5369),
    [sym_preproc_directive] = ACTIONS(5369),
    [anon_sym_DASH] = ACTIONS(5371),
    [anon_sym_PLUS] = ACTIONS(5371),
    [anon_sym_SEMI] = ACTIONS(5373),
    [anon_sym___extension__] = ACTIONS(5369),
    [anon_sym_typedef] = ACTIONS(5369),
    [anon_sym_extern] = ACTIONS(5369),
    [anon_sym___attribute__] = ACTIONS(5369),
    [anon_sym___attribute] = ACTIONS(5369),
    [anon_sym_noreturn] = ACTIONS(5369),
    [anon_sym_LBRACK] = ACTIONS(5371),
    [anon_sym___declspec] = ACTIONS(5369),
    [anon_sym___cdecl] = ACTIONS(5369),
    [anon_sym___clrcall] = ACTIONS(5369),
    [anon_sym___stdcall] = ACTIONS(5369),
    [anon_sym___fastcall] = ACTIONS(5369),
    [anon_sym___thiscall] = ACTIONS(5369),
    [anon_sym___vectorcall] = ACTIONS(5369),
    [anon_sym_signed] = ACTIONS(5369),
    [anon_sym_unsigned] = ACTIONS(5369),
    [anon_sym_long] = ACTIONS(5369),
    [anon_sym_short] = ACTIONS(5369),
    [anon_sym_static] = ACTIONS(5369),
    [anon_sym_auto] = ACTIONS(5369),
    [anon_sym_register] = ACTIONS(5369),
    [anon_sym_inline] = ACTIONS(5369),
    [anon_sym___inline] = ACTIONS(5369),
    [anon_sym___inline__] = ACTIONS(5369),
    [anon_sym___forceinline] = ACTIONS(5369),
    [anon_sym_thread_local] = ACTIONS(5369),
    [anon_sym___thread] = ACTIONS(5369),
    [anon_sym_CG_EXTERN] = ACTIONS(5369),
    [anon_sym_CG_INLINE] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5369),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5369),
    [anon_sym_IBOutlet] = ACTIONS(5369),
    [anon_sym_IBInspectable] = ACTIONS(5369),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5369),
    [anon_sym_NS_INLINE] = ACTIONS(5369),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5369),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5369),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5369),
    [anon_sym_const] = ACTIONS(5369),
    [anon_sym_constexpr] = ACTIONS(5369),
    [anon_sym_volatile] = ACTIONS(5369),
    [anon_sym_restrict] = ACTIONS(5369),
    [anon_sym___restrict__] = ACTIONS(5369),
    [anon_sym__Atomic] = ACTIONS(5369),
    [anon_sym__Noreturn] = ACTIONS(5369),
    [anon_sym__Nonnull] = ACTIONS(5369),
    [anon_sym_nullable] = ACTIONS(5369),
    [anon_sym__Complex] = ACTIONS(5369),
    [anon_sym__Nullable] = ACTIONS(5369),
    [anon_sym__Nullable_result] = ACTIONS(5369),
    [anon_sym__Null_unspecified] = ACTIONS(5369),
    [anon_sym___autoreleasing] = ACTIONS(5369),
    [anon_sym___block] = ACTIONS(5369),
    [anon_sym___bridge] = ACTIONS(5369),
    [anon_sym___bridge_retained] = ACTIONS(5369),
    [anon_sym___bridge_transfer] = ACTIONS(5369),
    [anon_sym___complex] = ACTIONS(5369),
    [anon_sym___const] = ACTIONS(5369),
    [anon_sym___imag] = ACTIONS(5369),
    [anon_sym___kindof] = ACTIONS(5369),
    [anon_sym___nonnull] = ACTIONS(5369),
    [anon_sym___nullable] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5369),
    [anon_sym___real] = ACTIONS(5369),
    [anon_sym___strong] = ACTIONS(5369),
    [anon_sym___unsafe_unretained] = ACTIONS(5369),
    [anon_sym___unused] = ACTIONS(5369),
    [anon_sym___weak] = ACTIONS(5369),
    [anon_sym_alignas] = ACTIONS(5369),
    [anon_sym__Alignas] = ACTIONS(5369),
    [sym_primitive_type] = ACTIONS(5369),
    [anon_sym_enum] = ACTIONS(5369),
    [anon_sym_struct] = ACTIONS(5369),
    [anon_sym_union] = ACTIONS(5369),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5369),
    [anon_sym___typeof] = ACTIONS(5369),
    [anon_sym_typeof] = ACTIONS(5369),
    [aux_sym_preproc_undef_token1] = ACTIONS(5369),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5369),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5369),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5369),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_API_AVAILABLE] = ACTIONS(5369),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_API_DEPRECATED] = ACTIONS(5369),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym___deprecated_msg] = ACTIONS(5369),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5369),
    [anon_sym_ATsynthesize] = ACTIONS(5371),
    [anon_sym_ATdynamic] = ACTIONS(5371),
    [anon_sym_BOOL] = ACTIONS(5369),
    [anon_sym_IMP] = ACTIONS(5369),
    [anon_sym_SEL] = ACTIONS(5369),
    [anon_sym_Class] = ACTIONS(5369),
    [anon_sym_id] = ACTIONS(5369),
  },
  [STATE(2107)] = {
    [sym_identifier] = ACTIONS(5375),
    [aux_sym_preproc_def_token1] = ACTIONS(5375),
    [aux_sym_preproc_if_token1] = ACTIONS(5375),
    [aux_sym_preproc_if_token2] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5375),
    [aux_sym_preproc_else_token1] = ACTIONS(5375),
    [aux_sym_preproc_elif_token1] = ACTIONS(5375),
    [sym_preproc_directive] = ACTIONS(5375),
    [anon_sym_DASH] = ACTIONS(5377),
    [anon_sym_PLUS] = ACTIONS(5377),
    [anon_sym_SEMI] = ACTIONS(5379),
    [anon_sym___extension__] = ACTIONS(5375),
    [anon_sym_typedef] = ACTIONS(5375),
    [anon_sym_extern] = ACTIONS(5375),
    [anon_sym___attribute__] = ACTIONS(5375),
    [anon_sym___attribute] = ACTIONS(5375),
    [anon_sym_noreturn] = ACTIONS(5375),
    [anon_sym_LBRACK] = ACTIONS(5377),
    [anon_sym___declspec] = ACTIONS(5375),
    [anon_sym___cdecl] = ACTIONS(5375),
    [anon_sym___clrcall] = ACTIONS(5375),
    [anon_sym___stdcall] = ACTIONS(5375),
    [anon_sym___fastcall] = ACTIONS(5375),
    [anon_sym___thiscall] = ACTIONS(5375),
    [anon_sym___vectorcall] = ACTIONS(5375),
    [anon_sym_signed] = ACTIONS(5375),
    [anon_sym_unsigned] = ACTIONS(5375),
    [anon_sym_long] = ACTIONS(5375),
    [anon_sym_short] = ACTIONS(5375),
    [anon_sym_static] = ACTIONS(5375),
    [anon_sym_auto] = ACTIONS(5375),
    [anon_sym_register] = ACTIONS(5375),
    [anon_sym_inline] = ACTIONS(5375),
    [anon_sym___inline] = ACTIONS(5375),
    [anon_sym___inline__] = ACTIONS(5375),
    [anon_sym___forceinline] = ACTIONS(5375),
    [anon_sym_thread_local] = ACTIONS(5375),
    [anon_sym___thread] = ACTIONS(5375),
    [anon_sym_CG_EXTERN] = ACTIONS(5375),
    [anon_sym_CG_INLINE] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5375),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5375),
    [anon_sym_IBOutlet] = ACTIONS(5375),
    [anon_sym_IBInspectable] = ACTIONS(5375),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5375),
    [anon_sym_NS_INLINE] = ACTIONS(5375),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5375),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5375),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5375),
    [anon_sym_const] = ACTIONS(5375),
    [anon_sym_constexpr] = ACTIONS(5375),
    [anon_sym_volatile] = ACTIONS(5375),
    [anon_sym_restrict] = ACTIONS(5375),
    [anon_sym___restrict__] = ACTIONS(5375),
    [anon_sym__Atomic] = ACTIONS(5375),
    [anon_sym__Noreturn] = ACTIONS(5375),
    [anon_sym__Nonnull] = ACTIONS(5375),
    [anon_sym_nullable] = ACTIONS(5375),
    [anon_sym__Complex] = ACTIONS(5375),
    [anon_sym__Nullable] = ACTIONS(5375),
    [anon_sym__Nullable_result] = ACTIONS(5375),
    [anon_sym__Null_unspecified] = ACTIONS(5375),
    [anon_sym___autoreleasing] = ACTIONS(5375),
    [anon_sym___block] = ACTIONS(5375),
    [anon_sym___bridge] = ACTIONS(5375),
    [anon_sym___bridge_retained] = ACTIONS(5375),
    [anon_sym___bridge_transfer] = ACTIONS(5375),
    [anon_sym___complex] = ACTIONS(5375),
    [anon_sym___const] = ACTIONS(5375),
    [anon_sym___imag] = ACTIONS(5375),
    [anon_sym___kindof] = ACTIONS(5375),
    [anon_sym___nonnull] = ACTIONS(5375),
    [anon_sym___nullable] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5375),
    [anon_sym___real] = ACTIONS(5375),
    [anon_sym___strong] = ACTIONS(5375),
    [anon_sym___unsafe_unretained] = ACTIONS(5375),
    [anon_sym___unused] = ACTIONS(5375),
    [anon_sym___weak] = ACTIONS(5375),
    [anon_sym_alignas] = ACTIONS(5375),
    [anon_sym__Alignas] = ACTIONS(5375),
    [sym_primitive_type] = ACTIONS(5375),
    [anon_sym_enum] = ACTIONS(5375),
    [anon_sym_struct] = ACTIONS(5375),
    [anon_sym_union] = ACTIONS(5375),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5375),
    [anon_sym___typeof] = ACTIONS(5375),
    [anon_sym_typeof] = ACTIONS(5375),
    [aux_sym_preproc_undef_token1] = ACTIONS(5375),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5375),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5375),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5375),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_API_AVAILABLE] = ACTIONS(5375),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_API_DEPRECATED] = ACTIONS(5375),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym___deprecated_msg] = ACTIONS(5375),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5375),
    [anon_sym_ATsynthesize] = ACTIONS(5377),
    [anon_sym_ATdynamic] = ACTIONS(5377),
    [anon_sym_BOOL] = ACTIONS(5375),
    [anon_sym_IMP] = ACTIONS(5375),
    [anon_sym_SEL] = ACTIONS(5375),
    [anon_sym_Class] = ACTIONS(5375),
    [anon_sym_id] = ACTIONS(5375),
  },
  [STATE(2108)] = {
    [sym_identifier] = ACTIONS(5381),
    [aux_sym_preproc_def_token1] = ACTIONS(5381),
    [aux_sym_preproc_if_token1] = ACTIONS(5381),
    [aux_sym_preproc_if_token2] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5381),
    [aux_sym_preproc_else_token1] = ACTIONS(5381),
    [aux_sym_preproc_elif_token1] = ACTIONS(5381),
    [sym_preproc_directive] = ACTIONS(5381),
    [anon_sym_DASH] = ACTIONS(5383),
    [anon_sym_PLUS] = ACTIONS(5383),
    [anon_sym_SEMI] = ACTIONS(5385),
    [anon_sym___extension__] = ACTIONS(5381),
    [anon_sym_typedef] = ACTIONS(5381),
    [anon_sym_extern] = ACTIONS(5381),
    [anon_sym___attribute__] = ACTIONS(5381),
    [anon_sym___attribute] = ACTIONS(5381),
    [anon_sym_noreturn] = ACTIONS(5381),
    [anon_sym_LBRACK] = ACTIONS(5383),
    [anon_sym___declspec] = ACTIONS(5381),
    [anon_sym___cdecl] = ACTIONS(5381),
    [anon_sym___clrcall] = ACTIONS(5381),
    [anon_sym___stdcall] = ACTIONS(5381),
    [anon_sym___fastcall] = ACTIONS(5381),
    [anon_sym___thiscall] = ACTIONS(5381),
    [anon_sym___vectorcall] = ACTIONS(5381),
    [anon_sym_signed] = ACTIONS(5381),
    [anon_sym_unsigned] = ACTIONS(5381),
    [anon_sym_long] = ACTIONS(5381),
    [anon_sym_short] = ACTIONS(5381),
    [anon_sym_static] = ACTIONS(5381),
    [anon_sym_auto] = ACTIONS(5381),
    [anon_sym_register] = ACTIONS(5381),
    [anon_sym_inline] = ACTIONS(5381),
    [anon_sym___inline] = ACTIONS(5381),
    [anon_sym___inline__] = ACTIONS(5381),
    [anon_sym___forceinline] = ACTIONS(5381),
    [anon_sym_thread_local] = ACTIONS(5381),
    [anon_sym___thread] = ACTIONS(5381),
    [anon_sym_CG_EXTERN] = ACTIONS(5381),
    [anon_sym_CG_INLINE] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5381),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5381),
    [anon_sym_IBOutlet] = ACTIONS(5381),
    [anon_sym_IBInspectable] = ACTIONS(5381),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5381),
    [anon_sym_NS_INLINE] = ACTIONS(5381),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5381),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5381),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5381),
    [anon_sym_const] = ACTIONS(5381),
    [anon_sym_constexpr] = ACTIONS(5381),
    [anon_sym_volatile] = ACTIONS(5381),
    [anon_sym_restrict] = ACTIONS(5381),
    [anon_sym___restrict__] = ACTIONS(5381),
    [anon_sym__Atomic] = ACTIONS(5381),
    [anon_sym__Noreturn] = ACTIONS(5381),
    [anon_sym__Nonnull] = ACTIONS(5381),
    [anon_sym_nullable] = ACTIONS(5381),
    [anon_sym__Complex] = ACTIONS(5381),
    [anon_sym__Nullable] = ACTIONS(5381),
    [anon_sym__Nullable_result] = ACTIONS(5381),
    [anon_sym__Null_unspecified] = ACTIONS(5381),
    [anon_sym___autoreleasing] = ACTIONS(5381),
    [anon_sym___block] = ACTIONS(5381),
    [anon_sym___bridge] = ACTIONS(5381),
    [anon_sym___bridge_retained] = ACTIONS(5381),
    [anon_sym___bridge_transfer] = ACTIONS(5381),
    [anon_sym___complex] = ACTIONS(5381),
    [anon_sym___const] = ACTIONS(5381),
    [anon_sym___imag] = ACTIONS(5381),
    [anon_sym___kindof] = ACTIONS(5381),
    [anon_sym___nonnull] = ACTIONS(5381),
    [anon_sym___nullable] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5381),
    [anon_sym___real] = ACTIONS(5381),
    [anon_sym___strong] = ACTIONS(5381),
    [anon_sym___unsafe_unretained] = ACTIONS(5381),
    [anon_sym___unused] = ACTIONS(5381),
    [anon_sym___weak] = ACTIONS(5381),
    [anon_sym_alignas] = ACTIONS(5381),
    [anon_sym__Alignas] = ACTIONS(5381),
    [sym_primitive_type] = ACTIONS(5381),
    [anon_sym_enum] = ACTIONS(5381),
    [anon_sym_struct] = ACTIONS(5381),
    [anon_sym_union] = ACTIONS(5381),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5381),
    [anon_sym___typeof] = ACTIONS(5381),
    [anon_sym_typeof] = ACTIONS(5381),
    [aux_sym_preproc_undef_token1] = ACTIONS(5381),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5381),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5381),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5381),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_API_AVAILABLE] = ACTIONS(5381),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_API_DEPRECATED] = ACTIONS(5381),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym___deprecated_msg] = ACTIONS(5381),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5381),
    [anon_sym_ATsynthesize] = ACTIONS(5383),
    [anon_sym_ATdynamic] = ACTIONS(5383),
    [anon_sym_BOOL] = ACTIONS(5381),
    [anon_sym_IMP] = ACTIONS(5381),
    [anon_sym_SEL] = ACTIONS(5381),
    [anon_sym_Class] = ACTIONS(5381),
    [anon_sym_id] = ACTIONS(5381),
  },
  [STATE(2109)] = {
    [sym_identifier] = ACTIONS(5387),
    [aux_sym_preproc_def_token1] = ACTIONS(5387),
    [aux_sym_preproc_if_token1] = ACTIONS(5387),
    [aux_sym_preproc_if_token2] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5387),
    [aux_sym_preproc_else_token1] = ACTIONS(5387),
    [aux_sym_preproc_elif_token1] = ACTIONS(5387),
    [sym_preproc_directive] = ACTIONS(5387),
    [anon_sym_DASH] = ACTIONS(5389),
    [anon_sym_PLUS] = ACTIONS(5389),
    [anon_sym_SEMI] = ACTIONS(5391),
    [anon_sym___extension__] = ACTIONS(5387),
    [anon_sym_typedef] = ACTIONS(5387),
    [anon_sym_extern] = ACTIONS(5387),
    [anon_sym___attribute__] = ACTIONS(5387),
    [anon_sym___attribute] = ACTIONS(5387),
    [anon_sym_noreturn] = ACTIONS(5387),
    [anon_sym_LBRACK] = ACTIONS(5389),
    [anon_sym___declspec] = ACTIONS(5387),
    [anon_sym___cdecl] = ACTIONS(5387),
    [anon_sym___clrcall] = ACTIONS(5387),
    [anon_sym___stdcall] = ACTIONS(5387),
    [anon_sym___fastcall] = ACTIONS(5387),
    [anon_sym___thiscall] = ACTIONS(5387),
    [anon_sym___vectorcall] = ACTIONS(5387),
    [anon_sym_signed] = ACTIONS(5387),
    [anon_sym_unsigned] = ACTIONS(5387),
    [anon_sym_long] = ACTIONS(5387),
    [anon_sym_short] = ACTIONS(5387),
    [anon_sym_static] = ACTIONS(5387),
    [anon_sym_auto] = ACTIONS(5387),
    [anon_sym_register] = ACTIONS(5387),
    [anon_sym_inline] = ACTIONS(5387),
    [anon_sym___inline] = ACTIONS(5387),
    [anon_sym___inline__] = ACTIONS(5387),
    [anon_sym___forceinline] = ACTIONS(5387),
    [anon_sym_thread_local] = ACTIONS(5387),
    [anon_sym___thread] = ACTIONS(5387),
    [anon_sym_CG_EXTERN] = ACTIONS(5387),
    [anon_sym_CG_INLINE] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5387),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5387),
    [anon_sym_IBOutlet] = ACTIONS(5387),
    [anon_sym_IBInspectable] = ACTIONS(5387),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5387),
    [anon_sym_NS_INLINE] = ACTIONS(5387),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5387),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5387),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5387),
    [anon_sym_const] = ACTIONS(5387),
    [anon_sym_constexpr] = ACTIONS(5387),
    [anon_sym_volatile] = ACTIONS(5387),
    [anon_sym_restrict] = ACTIONS(5387),
    [anon_sym___restrict__] = ACTIONS(5387),
    [anon_sym__Atomic] = ACTIONS(5387),
    [anon_sym__Noreturn] = ACTIONS(5387),
    [anon_sym__Nonnull] = ACTIONS(5387),
    [anon_sym_nullable] = ACTIONS(5387),
    [anon_sym__Complex] = ACTIONS(5387),
    [anon_sym__Nullable] = ACTIONS(5387),
    [anon_sym__Nullable_result] = ACTIONS(5387),
    [anon_sym__Null_unspecified] = ACTIONS(5387),
    [anon_sym___autoreleasing] = ACTIONS(5387),
    [anon_sym___block] = ACTIONS(5387),
    [anon_sym___bridge] = ACTIONS(5387),
    [anon_sym___bridge_retained] = ACTIONS(5387),
    [anon_sym___bridge_transfer] = ACTIONS(5387),
    [anon_sym___complex] = ACTIONS(5387),
    [anon_sym___const] = ACTIONS(5387),
    [anon_sym___imag] = ACTIONS(5387),
    [anon_sym___kindof] = ACTIONS(5387),
    [anon_sym___nonnull] = ACTIONS(5387),
    [anon_sym___nullable] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5387),
    [anon_sym___real] = ACTIONS(5387),
    [anon_sym___strong] = ACTIONS(5387),
    [anon_sym___unsafe_unretained] = ACTIONS(5387),
    [anon_sym___unused] = ACTIONS(5387),
    [anon_sym___weak] = ACTIONS(5387),
    [anon_sym_alignas] = ACTIONS(5387),
    [anon_sym__Alignas] = ACTIONS(5387),
    [sym_primitive_type] = ACTIONS(5387),
    [anon_sym_enum] = ACTIONS(5387),
    [anon_sym_struct] = ACTIONS(5387),
    [anon_sym_union] = ACTIONS(5387),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5387),
    [anon_sym___typeof] = ACTIONS(5387),
    [anon_sym_typeof] = ACTIONS(5387),
    [aux_sym_preproc_undef_token1] = ACTIONS(5387),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5387),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5387),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5387),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_API_AVAILABLE] = ACTIONS(5387),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_API_DEPRECATED] = ACTIONS(5387),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym___deprecated_msg] = ACTIONS(5387),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5387),
    [anon_sym_ATsynthesize] = ACTIONS(5389),
    [anon_sym_ATdynamic] = ACTIONS(5389),
    [anon_sym_BOOL] = ACTIONS(5387),
    [anon_sym_IMP] = ACTIONS(5387),
    [anon_sym_SEL] = ACTIONS(5387),
    [anon_sym_Class] = ACTIONS(5387),
    [anon_sym_id] = ACTIONS(5387),
  },
  [STATE(2110)] = {
    [sym_identifier] = ACTIONS(5393),
    [aux_sym_preproc_def_token1] = ACTIONS(5393),
    [aux_sym_preproc_if_token1] = ACTIONS(5393),
    [aux_sym_preproc_if_token2] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5393),
    [aux_sym_preproc_else_token1] = ACTIONS(5393),
    [aux_sym_preproc_elif_token1] = ACTIONS(5393),
    [sym_preproc_directive] = ACTIONS(5393),
    [anon_sym_DASH] = ACTIONS(5395),
    [anon_sym_PLUS] = ACTIONS(5395),
    [anon_sym_SEMI] = ACTIONS(5397),
    [anon_sym___extension__] = ACTIONS(5393),
    [anon_sym_typedef] = ACTIONS(5393),
    [anon_sym_extern] = ACTIONS(5393),
    [anon_sym___attribute__] = ACTIONS(5393),
    [anon_sym___attribute] = ACTIONS(5393),
    [anon_sym_noreturn] = ACTIONS(5393),
    [anon_sym_LBRACK] = ACTIONS(5395),
    [anon_sym___declspec] = ACTIONS(5393),
    [anon_sym___cdecl] = ACTIONS(5393),
    [anon_sym___clrcall] = ACTIONS(5393),
    [anon_sym___stdcall] = ACTIONS(5393),
    [anon_sym___fastcall] = ACTIONS(5393),
    [anon_sym___thiscall] = ACTIONS(5393),
    [anon_sym___vectorcall] = ACTIONS(5393),
    [anon_sym_signed] = ACTIONS(5393),
    [anon_sym_unsigned] = ACTIONS(5393),
    [anon_sym_long] = ACTIONS(5393),
    [anon_sym_short] = ACTIONS(5393),
    [anon_sym_static] = ACTIONS(5393),
    [anon_sym_auto] = ACTIONS(5393),
    [anon_sym_register] = ACTIONS(5393),
    [anon_sym_inline] = ACTIONS(5393),
    [anon_sym___inline] = ACTIONS(5393),
    [anon_sym___inline__] = ACTIONS(5393),
    [anon_sym___forceinline] = ACTIONS(5393),
    [anon_sym_thread_local] = ACTIONS(5393),
    [anon_sym___thread] = ACTIONS(5393),
    [anon_sym_CG_EXTERN] = ACTIONS(5393),
    [anon_sym_CG_INLINE] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5393),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5393),
    [anon_sym_IBOutlet] = ACTIONS(5393),
    [anon_sym_IBInspectable] = ACTIONS(5393),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5393),
    [anon_sym_NS_INLINE] = ACTIONS(5393),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5393),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5393),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5393),
    [anon_sym_const] = ACTIONS(5393),
    [anon_sym_constexpr] = ACTIONS(5393),
    [anon_sym_volatile] = ACTIONS(5393),
    [anon_sym_restrict] = ACTIONS(5393),
    [anon_sym___restrict__] = ACTIONS(5393),
    [anon_sym__Atomic] = ACTIONS(5393),
    [anon_sym__Noreturn] = ACTIONS(5393),
    [anon_sym__Nonnull] = ACTIONS(5393),
    [anon_sym_nullable] = ACTIONS(5393),
    [anon_sym__Complex] = ACTIONS(5393),
    [anon_sym__Nullable] = ACTIONS(5393),
    [anon_sym__Nullable_result] = ACTIONS(5393),
    [anon_sym__Null_unspecified] = ACTIONS(5393),
    [anon_sym___autoreleasing] = ACTIONS(5393),
    [anon_sym___block] = ACTIONS(5393),
    [anon_sym___bridge] = ACTIONS(5393),
    [anon_sym___bridge_retained] = ACTIONS(5393),
    [anon_sym___bridge_transfer] = ACTIONS(5393),
    [anon_sym___complex] = ACTIONS(5393),
    [anon_sym___const] = ACTIONS(5393),
    [anon_sym___imag] = ACTIONS(5393),
    [anon_sym___kindof] = ACTIONS(5393),
    [anon_sym___nonnull] = ACTIONS(5393),
    [anon_sym___nullable] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5393),
    [anon_sym___real] = ACTIONS(5393),
    [anon_sym___strong] = ACTIONS(5393),
    [anon_sym___unsafe_unretained] = ACTIONS(5393),
    [anon_sym___unused] = ACTIONS(5393),
    [anon_sym___weak] = ACTIONS(5393),
    [anon_sym_alignas] = ACTIONS(5393),
    [anon_sym__Alignas] = ACTIONS(5393),
    [sym_primitive_type] = ACTIONS(5393),
    [anon_sym_enum] = ACTIONS(5393),
    [anon_sym_struct] = ACTIONS(5393),
    [anon_sym_union] = ACTIONS(5393),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5393),
    [anon_sym___typeof] = ACTIONS(5393),
    [anon_sym_typeof] = ACTIONS(5393),
    [aux_sym_preproc_undef_token1] = ACTIONS(5393),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5393),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5393),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5393),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_API_AVAILABLE] = ACTIONS(5393),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_API_DEPRECATED] = ACTIONS(5393),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym___deprecated_msg] = ACTIONS(5393),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5393),
    [anon_sym_ATsynthesize] = ACTIONS(5395),
    [anon_sym_ATdynamic] = ACTIONS(5395),
    [anon_sym_BOOL] = ACTIONS(5393),
    [anon_sym_IMP] = ACTIONS(5393),
    [anon_sym_SEL] = ACTIONS(5393),
    [anon_sym_Class] = ACTIONS(5393),
    [anon_sym_id] = ACTIONS(5393),
  },
  [STATE(2111)] = {
    [sym_identifier] = ACTIONS(5399),
    [aux_sym_preproc_def_token1] = ACTIONS(5399),
    [aux_sym_preproc_if_token1] = ACTIONS(5399),
    [aux_sym_preproc_if_token2] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5399),
    [aux_sym_preproc_else_token1] = ACTIONS(5399),
    [aux_sym_preproc_elif_token1] = ACTIONS(5399),
    [sym_preproc_directive] = ACTIONS(5399),
    [anon_sym_DASH] = ACTIONS(5401),
    [anon_sym_PLUS] = ACTIONS(5401),
    [anon_sym_SEMI] = ACTIONS(5403),
    [anon_sym___extension__] = ACTIONS(5399),
    [anon_sym_typedef] = ACTIONS(5399),
    [anon_sym_extern] = ACTIONS(5399),
    [anon_sym___attribute__] = ACTIONS(5399),
    [anon_sym___attribute] = ACTIONS(5399),
    [anon_sym_noreturn] = ACTIONS(5399),
    [anon_sym_LBRACK] = ACTIONS(5401),
    [anon_sym___declspec] = ACTIONS(5399),
    [anon_sym___cdecl] = ACTIONS(5399),
    [anon_sym___clrcall] = ACTIONS(5399),
    [anon_sym___stdcall] = ACTIONS(5399),
    [anon_sym___fastcall] = ACTIONS(5399),
    [anon_sym___thiscall] = ACTIONS(5399),
    [anon_sym___vectorcall] = ACTIONS(5399),
    [anon_sym_signed] = ACTIONS(5399),
    [anon_sym_unsigned] = ACTIONS(5399),
    [anon_sym_long] = ACTIONS(5399),
    [anon_sym_short] = ACTIONS(5399),
    [anon_sym_static] = ACTIONS(5399),
    [anon_sym_auto] = ACTIONS(5399),
    [anon_sym_register] = ACTIONS(5399),
    [anon_sym_inline] = ACTIONS(5399),
    [anon_sym___inline] = ACTIONS(5399),
    [anon_sym___inline__] = ACTIONS(5399),
    [anon_sym___forceinline] = ACTIONS(5399),
    [anon_sym_thread_local] = ACTIONS(5399),
    [anon_sym___thread] = ACTIONS(5399),
    [anon_sym_CG_EXTERN] = ACTIONS(5399),
    [anon_sym_CG_INLINE] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5399),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5399),
    [anon_sym_IBOutlet] = ACTIONS(5399),
    [anon_sym_IBInspectable] = ACTIONS(5399),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5399),
    [anon_sym_NS_INLINE] = ACTIONS(5399),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5399),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5399),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5399),
    [anon_sym_const] = ACTIONS(5399),
    [anon_sym_constexpr] = ACTIONS(5399),
    [anon_sym_volatile] = ACTIONS(5399),
    [anon_sym_restrict] = ACTIONS(5399),
    [anon_sym___restrict__] = ACTIONS(5399),
    [anon_sym__Atomic] = ACTIONS(5399),
    [anon_sym__Noreturn] = ACTIONS(5399),
    [anon_sym__Nonnull] = ACTIONS(5399),
    [anon_sym_nullable] = ACTIONS(5399),
    [anon_sym__Complex] = ACTIONS(5399),
    [anon_sym__Nullable] = ACTIONS(5399),
    [anon_sym__Nullable_result] = ACTIONS(5399),
    [anon_sym__Null_unspecified] = ACTIONS(5399),
    [anon_sym___autoreleasing] = ACTIONS(5399),
    [anon_sym___block] = ACTIONS(5399),
    [anon_sym___bridge] = ACTIONS(5399),
    [anon_sym___bridge_retained] = ACTIONS(5399),
    [anon_sym___bridge_transfer] = ACTIONS(5399),
    [anon_sym___complex] = ACTIONS(5399),
    [anon_sym___const] = ACTIONS(5399),
    [anon_sym___imag] = ACTIONS(5399),
    [anon_sym___kindof] = ACTIONS(5399),
    [anon_sym___nonnull] = ACTIONS(5399),
    [anon_sym___nullable] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5399),
    [anon_sym___real] = ACTIONS(5399),
    [anon_sym___strong] = ACTIONS(5399),
    [anon_sym___unsafe_unretained] = ACTIONS(5399),
    [anon_sym___unused] = ACTIONS(5399),
    [anon_sym___weak] = ACTIONS(5399),
    [anon_sym_alignas] = ACTIONS(5399),
    [anon_sym__Alignas] = ACTIONS(5399),
    [sym_primitive_type] = ACTIONS(5399),
    [anon_sym_enum] = ACTIONS(5399),
    [anon_sym_struct] = ACTIONS(5399),
    [anon_sym_union] = ACTIONS(5399),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5399),
    [anon_sym___typeof] = ACTIONS(5399),
    [anon_sym_typeof] = ACTIONS(5399),
    [aux_sym_preproc_undef_token1] = ACTIONS(5399),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5399),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5399),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5399),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_API_AVAILABLE] = ACTIONS(5399),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_API_DEPRECATED] = ACTIONS(5399),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym___deprecated_msg] = ACTIONS(5399),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5399),
    [anon_sym_ATsynthesize] = ACTIONS(5401),
    [anon_sym_ATdynamic] = ACTIONS(5401),
    [anon_sym_BOOL] = ACTIONS(5399),
    [anon_sym_IMP] = ACTIONS(5399),
    [anon_sym_SEL] = ACTIONS(5399),
    [anon_sym_Class] = ACTIONS(5399),
    [anon_sym_id] = ACTIONS(5399),
  },
  [STATE(2112)] = {
    [sym_identifier] = ACTIONS(5405),
    [aux_sym_preproc_def_token1] = ACTIONS(5405),
    [aux_sym_preproc_if_token1] = ACTIONS(5405),
    [aux_sym_preproc_if_token2] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5405),
    [aux_sym_preproc_else_token1] = ACTIONS(5405),
    [aux_sym_preproc_elif_token1] = ACTIONS(5405),
    [sym_preproc_directive] = ACTIONS(5405),
    [anon_sym_DASH] = ACTIONS(5407),
    [anon_sym_PLUS] = ACTIONS(5407),
    [anon_sym_SEMI] = ACTIONS(5409),
    [anon_sym___extension__] = ACTIONS(5405),
    [anon_sym_typedef] = ACTIONS(5405),
    [anon_sym_extern] = ACTIONS(5405),
    [anon_sym___attribute__] = ACTIONS(5405),
    [anon_sym___attribute] = ACTIONS(5405),
    [anon_sym_noreturn] = ACTIONS(5405),
    [anon_sym_LBRACK] = ACTIONS(5407),
    [anon_sym___declspec] = ACTIONS(5405),
    [anon_sym___cdecl] = ACTIONS(5405),
    [anon_sym___clrcall] = ACTIONS(5405),
    [anon_sym___stdcall] = ACTIONS(5405),
    [anon_sym___fastcall] = ACTIONS(5405),
    [anon_sym___thiscall] = ACTIONS(5405),
    [anon_sym___vectorcall] = ACTIONS(5405),
    [anon_sym_signed] = ACTIONS(5405),
    [anon_sym_unsigned] = ACTIONS(5405),
    [anon_sym_long] = ACTIONS(5405),
    [anon_sym_short] = ACTIONS(5405),
    [anon_sym_static] = ACTIONS(5405),
    [anon_sym_auto] = ACTIONS(5405),
    [anon_sym_register] = ACTIONS(5405),
    [anon_sym_inline] = ACTIONS(5405),
    [anon_sym___inline] = ACTIONS(5405),
    [anon_sym___inline__] = ACTIONS(5405),
    [anon_sym___forceinline] = ACTIONS(5405),
    [anon_sym_thread_local] = ACTIONS(5405),
    [anon_sym___thread] = ACTIONS(5405),
    [anon_sym_CG_EXTERN] = ACTIONS(5405),
    [anon_sym_CG_INLINE] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5405),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5405),
    [anon_sym_IBOutlet] = ACTIONS(5405),
    [anon_sym_IBInspectable] = ACTIONS(5405),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5405),
    [anon_sym_NS_INLINE] = ACTIONS(5405),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5405),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5405),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5405),
    [anon_sym_const] = ACTIONS(5405),
    [anon_sym_constexpr] = ACTIONS(5405),
    [anon_sym_volatile] = ACTIONS(5405),
    [anon_sym_restrict] = ACTIONS(5405),
    [anon_sym___restrict__] = ACTIONS(5405),
    [anon_sym__Atomic] = ACTIONS(5405),
    [anon_sym__Noreturn] = ACTIONS(5405),
    [anon_sym__Nonnull] = ACTIONS(5405),
    [anon_sym_nullable] = ACTIONS(5405),
    [anon_sym__Complex] = ACTIONS(5405),
    [anon_sym__Nullable] = ACTIONS(5405),
    [anon_sym__Nullable_result] = ACTIONS(5405),
    [anon_sym__Null_unspecified] = ACTIONS(5405),
    [anon_sym___autoreleasing] = ACTIONS(5405),
    [anon_sym___block] = ACTIONS(5405),
    [anon_sym___bridge] = ACTIONS(5405),
    [anon_sym___bridge_retained] = ACTIONS(5405),
    [anon_sym___bridge_transfer] = ACTIONS(5405),
    [anon_sym___complex] = ACTIONS(5405),
    [anon_sym___const] = ACTIONS(5405),
    [anon_sym___imag] = ACTIONS(5405),
    [anon_sym___kindof] = ACTIONS(5405),
    [anon_sym___nonnull] = ACTIONS(5405),
    [anon_sym___nullable] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5405),
    [anon_sym___real] = ACTIONS(5405),
    [anon_sym___strong] = ACTIONS(5405),
    [anon_sym___unsafe_unretained] = ACTIONS(5405),
    [anon_sym___unused] = ACTIONS(5405),
    [anon_sym___weak] = ACTIONS(5405),
    [anon_sym_alignas] = ACTIONS(5405),
    [anon_sym__Alignas] = ACTIONS(5405),
    [sym_primitive_type] = ACTIONS(5405),
    [anon_sym_enum] = ACTIONS(5405),
    [anon_sym_struct] = ACTIONS(5405),
    [anon_sym_union] = ACTIONS(5405),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5405),
    [anon_sym___typeof] = ACTIONS(5405),
    [anon_sym_typeof] = ACTIONS(5405),
    [aux_sym_preproc_undef_token1] = ACTIONS(5405),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5405),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5405),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5405),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_API_AVAILABLE] = ACTIONS(5405),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_API_DEPRECATED] = ACTIONS(5405),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym___deprecated_msg] = ACTIONS(5405),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5405),
    [anon_sym_ATsynthesize] = ACTIONS(5407),
    [anon_sym_ATdynamic] = ACTIONS(5407),
    [anon_sym_BOOL] = ACTIONS(5405),
    [anon_sym_IMP] = ACTIONS(5405),
    [anon_sym_SEL] = ACTIONS(5405),
    [anon_sym_Class] = ACTIONS(5405),
    [anon_sym_id] = ACTIONS(5405),
  },
  [STATE(2113)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token2] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [aux_sym_preproc_else_token1] = ACTIONS(2546),
    [aux_sym_preproc_elif_token1] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_PLUS] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATproperty] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(2114)] = {
    [sym_identifier] = ACTIONS(2856),
    [aux_sym_preproc_def_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token2] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2856),
    [aux_sym_preproc_else_token1] = ACTIONS(2856),
    [aux_sym_preproc_elif_token1] = ACTIONS(2856),
    [sym_preproc_directive] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2858),
    [anon_sym_PLUS] = ACTIONS(2858),
    [anon_sym___extension__] = ACTIONS(2856),
    [anon_sym_typedef] = ACTIONS(2856),
    [anon_sym_extern] = ACTIONS(2856),
    [anon_sym___attribute__] = ACTIONS(2856),
    [anon_sym___attribute] = ACTIONS(2856),
    [anon_sym_noreturn] = ACTIONS(2856),
    [anon_sym_LBRACK] = ACTIONS(2858),
    [anon_sym___declspec] = ACTIONS(2856),
    [anon_sym___cdecl] = ACTIONS(2856),
    [anon_sym___clrcall] = ACTIONS(2856),
    [anon_sym___stdcall] = ACTIONS(2856),
    [anon_sym___fastcall] = ACTIONS(2856),
    [anon_sym___thiscall] = ACTIONS(2856),
    [anon_sym___vectorcall] = ACTIONS(2856),
    [anon_sym_signed] = ACTIONS(2856),
    [anon_sym_unsigned] = ACTIONS(2856),
    [anon_sym_long] = ACTIONS(2856),
    [anon_sym_short] = ACTIONS(2856),
    [anon_sym_static] = ACTIONS(2856),
    [anon_sym_auto] = ACTIONS(2856),
    [anon_sym_register] = ACTIONS(2856),
    [anon_sym_inline] = ACTIONS(2856),
    [anon_sym___inline] = ACTIONS(2856),
    [anon_sym___inline__] = ACTIONS(2856),
    [anon_sym___forceinline] = ACTIONS(2856),
    [anon_sym_thread_local] = ACTIONS(2856),
    [anon_sym___thread] = ACTIONS(2856),
    [anon_sym_CG_EXTERN] = ACTIONS(2856),
    [anon_sym_CG_INLINE] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2856),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2856),
    [anon_sym_IBOutlet] = ACTIONS(2856),
    [anon_sym_IBInspectable] = ACTIONS(2856),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2856),
    [anon_sym_NS_INLINE] = ACTIONS(2856),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2856),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2856),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2856),
    [anon_sym_const] = ACTIONS(2856),
    [anon_sym_constexpr] = ACTIONS(2856),
    [anon_sym_volatile] = ACTIONS(2856),
    [anon_sym_restrict] = ACTIONS(2856),
    [anon_sym___restrict__] = ACTIONS(2856),
    [anon_sym__Atomic] = ACTIONS(2856),
    [anon_sym__Noreturn] = ACTIONS(2856),
    [anon_sym__Nonnull] = ACTIONS(2856),
    [anon_sym_nullable] = ACTIONS(2856),
    [anon_sym__Complex] = ACTIONS(2856),
    [anon_sym__Nullable] = ACTIONS(2856),
    [anon_sym__Nullable_result] = ACTIONS(2856),
    [anon_sym__Null_unspecified] = ACTIONS(2856),
    [anon_sym___autoreleasing] = ACTIONS(2856),
    [anon_sym___block] = ACTIONS(2856),
    [anon_sym___bridge] = ACTIONS(2856),
    [anon_sym___bridge_retained] = ACTIONS(2856),
    [anon_sym___bridge_transfer] = ACTIONS(2856),
    [anon_sym___complex] = ACTIONS(2856),
    [anon_sym___const] = ACTIONS(2856),
    [anon_sym___imag] = ACTIONS(2856),
    [anon_sym___kindof] = ACTIONS(2856),
    [anon_sym___nonnull] = ACTIONS(2856),
    [anon_sym___nullable] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2856),
    [anon_sym___real] = ACTIONS(2856),
    [anon_sym___strong] = ACTIONS(2856),
    [anon_sym___unsafe_unretained] = ACTIONS(2856),
    [anon_sym___unused] = ACTIONS(2856),
    [anon_sym___weak] = ACTIONS(2856),
    [anon_sym_alignas] = ACTIONS(2856),
    [anon_sym__Alignas] = ACTIONS(2856),
    [sym_primitive_type] = ACTIONS(2856),
    [anon_sym_enum] = ACTIONS(2856),
    [anon_sym_struct] = ACTIONS(2856),
    [anon_sym_union] = ACTIONS(2856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2856),
    [anon_sym___typeof] = ACTIONS(2856),
    [anon_sym_typeof] = ACTIONS(2856),
    [aux_sym_preproc_undef_token1] = ACTIONS(2856),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2856),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2856),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2856),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_API_AVAILABLE] = ACTIONS(2856),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_API_DEPRECATED] = ACTIONS(2856),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym___deprecated_msg] = ACTIONS(2856),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2856),
    [anon_sym_ATsynthesize] = ACTIONS(2858),
    [anon_sym_ATdynamic] = ACTIONS(2858),
    [anon_sym_ATproperty] = ACTIONS(2858),
    [anon_sym_BOOL] = ACTIONS(2856),
    [anon_sym_IMP] = ACTIONS(2856),
    [anon_sym_SEL] = ACTIONS(2856),
    [anon_sym_Class] = ACTIONS(2856),
    [anon_sym_id] = ACTIONS(2856),
  },
  [STATE(2115)] = {
    [sym_identifier] = ACTIONS(2588),
    [aux_sym_preproc_def_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token2] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2588),
    [aux_sym_preproc_else_token1] = ACTIONS(2588),
    [aux_sym_preproc_elif_token1] = ACTIONS(2588),
    [sym_preproc_directive] = ACTIONS(2588),
    [anon_sym_DASH] = ACTIONS(2590),
    [anon_sym_PLUS] = ACTIONS(2590),
    [anon_sym___extension__] = ACTIONS(2588),
    [anon_sym_typedef] = ACTIONS(2588),
    [anon_sym_extern] = ACTIONS(2588),
    [anon_sym___attribute__] = ACTIONS(2588),
    [anon_sym___attribute] = ACTIONS(2588),
    [anon_sym_noreturn] = ACTIONS(2588),
    [anon_sym_LBRACK] = ACTIONS(2590),
    [anon_sym___declspec] = ACTIONS(2588),
    [anon_sym___cdecl] = ACTIONS(2588),
    [anon_sym___clrcall] = ACTIONS(2588),
    [anon_sym___stdcall] = ACTIONS(2588),
    [anon_sym___fastcall] = ACTIONS(2588),
    [anon_sym___thiscall] = ACTIONS(2588),
    [anon_sym___vectorcall] = ACTIONS(2588),
    [anon_sym_signed] = ACTIONS(2588),
    [anon_sym_unsigned] = ACTIONS(2588),
    [anon_sym_long] = ACTIONS(2588),
    [anon_sym_short] = ACTIONS(2588),
    [anon_sym_static] = ACTIONS(2588),
    [anon_sym_auto] = ACTIONS(2588),
    [anon_sym_register] = ACTIONS(2588),
    [anon_sym_inline] = ACTIONS(2588),
    [anon_sym___inline] = ACTIONS(2588),
    [anon_sym___inline__] = ACTIONS(2588),
    [anon_sym___forceinline] = ACTIONS(2588),
    [anon_sym_thread_local] = ACTIONS(2588),
    [anon_sym___thread] = ACTIONS(2588),
    [anon_sym_CG_EXTERN] = ACTIONS(2588),
    [anon_sym_CG_INLINE] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2588),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2588),
    [anon_sym_IBOutlet] = ACTIONS(2588),
    [anon_sym_IBInspectable] = ACTIONS(2588),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2588),
    [anon_sym_NS_INLINE] = ACTIONS(2588),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2588),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2588),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2588),
    [anon_sym_const] = ACTIONS(2588),
    [anon_sym_constexpr] = ACTIONS(2588),
    [anon_sym_volatile] = ACTIONS(2588),
    [anon_sym_restrict] = ACTIONS(2588),
    [anon_sym___restrict__] = ACTIONS(2588),
    [anon_sym__Atomic] = ACTIONS(2588),
    [anon_sym__Noreturn] = ACTIONS(2588),
    [anon_sym__Nonnull] = ACTIONS(2588),
    [anon_sym_nullable] = ACTIONS(2588),
    [anon_sym__Complex] = ACTIONS(2588),
    [anon_sym__Nullable] = ACTIONS(2588),
    [anon_sym__Nullable_result] = ACTIONS(2588),
    [anon_sym__Null_unspecified] = ACTIONS(2588),
    [anon_sym___autoreleasing] = ACTIONS(2588),
    [anon_sym___block] = ACTIONS(2588),
    [anon_sym___bridge] = ACTIONS(2588),
    [anon_sym___bridge_retained] = ACTIONS(2588),
    [anon_sym___bridge_transfer] = ACTIONS(2588),
    [anon_sym___complex] = ACTIONS(2588),
    [anon_sym___const] = ACTIONS(2588),
    [anon_sym___imag] = ACTIONS(2588),
    [anon_sym___kindof] = ACTIONS(2588),
    [anon_sym___nonnull] = ACTIONS(2588),
    [anon_sym___nullable] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2588),
    [anon_sym___real] = ACTIONS(2588),
    [anon_sym___strong] = ACTIONS(2588),
    [anon_sym___unsafe_unretained] = ACTIONS(2588),
    [anon_sym___unused] = ACTIONS(2588),
    [anon_sym___weak] = ACTIONS(2588),
    [anon_sym_alignas] = ACTIONS(2588),
    [anon_sym__Alignas] = ACTIONS(2588),
    [sym_primitive_type] = ACTIONS(2588),
    [anon_sym_enum] = ACTIONS(2588),
    [anon_sym_struct] = ACTIONS(2588),
    [anon_sym_union] = ACTIONS(2588),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2588),
    [anon_sym___typeof] = ACTIONS(2588),
    [anon_sym_typeof] = ACTIONS(2588),
    [aux_sym_preproc_undef_token1] = ACTIONS(2588),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2588),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2588),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2588),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_API_AVAILABLE] = ACTIONS(2588),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_API_DEPRECATED] = ACTIONS(2588),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym___deprecated_msg] = ACTIONS(2588),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2588),
    [anon_sym_ATsynthesize] = ACTIONS(2590),
    [anon_sym_ATdynamic] = ACTIONS(2590),
    [anon_sym_ATproperty] = ACTIONS(2590),
    [anon_sym_BOOL] = ACTIONS(2588),
    [anon_sym_IMP] = ACTIONS(2588),
    [anon_sym_SEL] = ACTIONS(2588),
    [anon_sym_Class] = ACTIONS(2588),
    [anon_sym_id] = ACTIONS(2588),
  },
  [STATE(2116)] = {
    [sym_identifier] = ACTIONS(2716),
    [aux_sym_preproc_def_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token2] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2716),
    [aux_sym_preproc_else_token1] = ACTIONS(2716),
    [aux_sym_preproc_elif_token1] = ACTIONS(2716),
    [sym_preproc_directive] = ACTIONS(2716),
    [anon_sym_DASH] = ACTIONS(2718),
    [anon_sym_PLUS] = ACTIONS(2718),
    [anon_sym___extension__] = ACTIONS(2716),
    [anon_sym_typedef] = ACTIONS(2716),
    [anon_sym_extern] = ACTIONS(2716),
    [anon_sym___attribute__] = ACTIONS(2716),
    [anon_sym___attribute] = ACTIONS(2716),
    [anon_sym_noreturn] = ACTIONS(2716),
    [anon_sym_LBRACK] = ACTIONS(2718),
    [anon_sym___declspec] = ACTIONS(2716),
    [anon_sym___cdecl] = ACTIONS(2716),
    [anon_sym___clrcall] = ACTIONS(2716),
    [anon_sym___stdcall] = ACTIONS(2716),
    [anon_sym___fastcall] = ACTIONS(2716),
    [anon_sym___thiscall] = ACTIONS(2716),
    [anon_sym___vectorcall] = ACTIONS(2716),
    [anon_sym_signed] = ACTIONS(2716),
    [anon_sym_unsigned] = ACTIONS(2716),
    [anon_sym_long] = ACTIONS(2716),
    [anon_sym_short] = ACTIONS(2716),
    [anon_sym_static] = ACTIONS(2716),
    [anon_sym_auto] = ACTIONS(2716),
    [anon_sym_register] = ACTIONS(2716),
    [anon_sym_inline] = ACTIONS(2716),
    [anon_sym___inline] = ACTIONS(2716),
    [anon_sym___inline__] = ACTIONS(2716),
    [anon_sym___forceinline] = ACTIONS(2716),
    [anon_sym_thread_local] = ACTIONS(2716),
    [anon_sym___thread] = ACTIONS(2716),
    [anon_sym_CG_EXTERN] = ACTIONS(2716),
    [anon_sym_CG_INLINE] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2716),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2716),
    [anon_sym_IBOutlet] = ACTIONS(2716),
    [anon_sym_IBInspectable] = ACTIONS(2716),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2716),
    [anon_sym_NS_INLINE] = ACTIONS(2716),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2716),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2716),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2716),
    [anon_sym_const] = ACTIONS(2716),
    [anon_sym_constexpr] = ACTIONS(2716),
    [anon_sym_volatile] = ACTIONS(2716),
    [anon_sym_restrict] = ACTIONS(2716),
    [anon_sym___restrict__] = ACTIONS(2716),
    [anon_sym__Atomic] = ACTIONS(2716),
    [anon_sym__Noreturn] = ACTIONS(2716),
    [anon_sym__Nonnull] = ACTIONS(2716),
    [anon_sym_nullable] = ACTIONS(2716),
    [anon_sym__Complex] = ACTIONS(2716),
    [anon_sym__Nullable] = ACTIONS(2716),
    [anon_sym__Nullable_result] = ACTIONS(2716),
    [anon_sym__Null_unspecified] = ACTIONS(2716),
    [anon_sym___autoreleasing] = ACTIONS(2716),
    [anon_sym___block] = ACTIONS(2716),
    [anon_sym___bridge] = ACTIONS(2716),
    [anon_sym___bridge_retained] = ACTIONS(2716),
    [anon_sym___bridge_transfer] = ACTIONS(2716),
    [anon_sym___complex] = ACTIONS(2716),
    [anon_sym___const] = ACTIONS(2716),
    [anon_sym___imag] = ACTIONS(2716),
    [anon_sym___kindof] = ACTIONS(2716),
    [anon_sym___nonnull] = ACTIONS(2716),
    [anon_sym___nullable] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2716),
    [anon_sym___real] = ACTIONS(2716),
    [anon_sym___strong] = ACTIONS(2716),
    [anon_sym___unsafe_unretained] = ACTIONS(2716),
    [anon_sym___unused] = ACTIONS(2716),
    [anon_sym___weak] = ACTIONS(2716),
    [anon_sym_alignas] = ACTIONS(2716),
    [anon_sym__Alignas] = ACTIONS(2716),
    [sym_primitive_type] = ACTIONS(2716),
    [anon_sym_enum] = ACTIONS(2716),
    [anon_sym_struct] = ACTIONS(2716),
    [anon_sym_union] = ACTIONS(2716),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2716),
    [anon_sym___typeof] = ACTIONS(2716),
    [anon_sym_typeof] = ACTIONS(2716),
    [aux_sym_preproc_undef_token1] = ACTIONS(2716),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2716),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2716),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2716),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_API_AVAILABLE] = ACTIONS(2716),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_API_DEPRECATED] = ACTIONS(2716),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym___deprecated_msg] = ACTIONS(2716),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2716),
    [anon_sym_ATsynthesize] = ACTIONS(2718),
    [anon_sym_ATdynamic] = ACTIONS(2718),
    [anon_sym_ATproperty] = ACTIONS(2718),
    [anon_sym_BOOL] = ACTIONS(2716),
    [anon_sym_IMP] = ACTIONS(2716),
    [anon_sym_SEL] = ACTIONS(2716),
    [anon_sym_Class] = ACTIONS(2716),
    [anon_sym_id] = ACTIONS(2716),
  },
  [STATE(2117)] = {
    [sym_identifier] = ACTIONS(2716),
    [aux_sym_preproc_def_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2716),
    [sym_preproc_directive] = ACTIONS(2716),
    [anon_sym_DASH] = ACTIONS(2718),
    [anon_sym_PLUS] = ACTIONS(2718),
    [anon_sym___extension__] = ACTIONS(2716),
    [anon_sym_typedef] = ACTIONS(2716),
    [anon_sym_extern] = ACTIONS(2716),
    [anon_sym___attribute__] = ACTIONS(2716),
    [anon_sym___attribute] = ACTIONS(2716),
    [anon_sym_noreturn] = ACTIONS(2716),
    [anon_sym_LBRACK] = ACTIONS(2718),
    [anon_sym___declspec] = ACTIONS(2716),
    [anon_sym___cdecl] = ACTIONS(2716),
    [anon_sym___clrcall] = ACTIONS(2716),
    [anon_sym___stdcall] = ACTIONS(2716),
    [anon_sym___fastcall] = ACTIONS(2716),
    [anon_sym___thiscall] = ACTIONS(2716),
    [anon_sym___vectorcall] = ACTIONS(2716),
    [anon_sym_signed] = ACTIONS(2716),
    [anon_sym_unsigned] = ACTIONS(2716),
    [anon_sym_long] = ACTIONS(2716),
    [anon_sym_short] = ACTIONS(2716),
    [anon_sym_static] = ACTIONS(2716),
    [anon_sym_auto] = ACTIONS(2716),
    [anon_sym_register] = ACTIONS(2716),
    [anon_sym_inline] = ACTIONS(2716),
    [anon_sym___inline] = ACTIONS(2716),
    [anon_sym___inline__] = ACTIONS(2716),
    [anon_sym___forceinline] = ACTIONS(2716),
    [anon_sym_thread_local] = ACTIONS(2716),
    [anon_sym___thread] = ACTIONS(2716),
    [anon_sym_CG_EXTERN] = ACTIONS(2716),
    [anon_sym_CG_INLINE] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2716),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2716),
    [anon_sym_IBOutlet] = ACTIONS(2716),
    [anon_sym_IBInspectable] = ACTIONS(2716),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2716),
    [anon_sym_NS_INLINE] = ACTIONS(2716),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2716),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2716),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2716),
    [anon_sym_const] = ACTIONS(2716),
    [anon_sym_constexpr] = ACTIONS(2716),
    [anon_sym_volatile] = ACTIONS(2716),
    [anon_sym_restrict] = ACTIONS(2716),
    [anon_sym___restrict__] = ACTIONS(2716),
    [anon_sym__Atomic] = ACTIONS(2716),
    [anon_sym__Noreturn] = ACTIONS(2716),
    [anon_sym__Nonnull] = ACTIONS(2716),
    [anon_sym_nullable] = ACTIONS(2716),
    [anon_sym__Complex] = ACTIONS(2716),
    [anon_sym__Nullable] = ACTIONS(2716),
    [anon_sym__Nullable_result] = ACTIONS(2716),
    [anon_sym__Null_unspecified] = ACTIONS(2716),
    [anon_sym___autoreleasing] = ACTIONS(2716),
    [anon_sym___block] = ACTIONS(2716),
    [anon_sym___bridge] = ACTIONS(2716),
    [anon_sym___bridge_retained] = ACTIONS(2716),
    [anon_sym___bridge_transfer] = ACTIONS(2716),
    [anon_sym___complex] = ACTIONS(2716),
    [anon_sym___const] = ACTIONS(2716),
    [anon_sym___imag] = ACTIONS(2716),
    [anon_sym___kindof] = ACTIONS(2716),
    [anon_sym___nonnull] = ACTIONS(2716),
    [anon_sym___nullable] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2716),
    [anon_sym___real] = ACTIONS(2716),
    [anon_sym___strong] = ACTIONS(2716),
    [anon_sym___unsafe_unretained] = ACTIONS(2716),
    [anon_sym___unused] = ACTIONS(2716),
    [anon_sym___weak] = ACTIONS(2716),
    [anon_sym_alignas] = ACTIONS(2716),
    [anon_sym__Alignas] = ACTIONS(2716),
    [sym_primitive_type] = ACTIONS(2716),
    [anon_sym_enum] = ACTIONS(2716),
    [anon_sym_struct] = ACTIONS(2716),
    [anon_sym_union] = ACTIONS(2716),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2716),
    [anon_sym___typeof] = ACTIONS(2716),
    [anon_sym_typeof] = ACTIONS(2716),
    [aux_sym_preproc_undef_token1] = ACTIONS(2716),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2716),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2716),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2716),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_API_AVAILABLE] = ACTIONS(2716),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_API_DEPRECATED] = ACTIONS(2716),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym___deprecated_msg] = ACTIONS(2716),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2716),
    [anon_sym_ATend] = ACTIONS(2718),
    [anon_sym_AToptional] = ACTIONS(2718),
    [anon_sym_ATrequired] = ACTIONS(2718),
    [anon_sym_ATsynthesize] = ACTIONS(2718),
    [anon_sym_ATdynamic] = ACTIONS(2718),
    [anon_sym_ATproperty] = ACTIONS(2718),
    [anon_sym_BOOL] = ACTIONS(2716),
    [anon_sym_IMP] = ACTIONS(2716),
    [anon_sym_SEL] = ACTIONS(2716),
    [anon_sym_Class] = ACTIONS(2716),
    [anon_sym_id] = ACTIONS(2716),
  },
  [STATE(2118)] = {
    [sym_identifier] = ACTIONS(2852),
    [aux_sym_preproc_def_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token2] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2852),
    [aux_sym_preproc_else_token1] = ACTIONS(2852),
    [aux_sym_preproc_elif_token1] = ACTIONS(2852),
    [sym_preproc_directive] = ACTIONS(2852),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym___extension__] = ACTIONS(2852),
    [anon_sym_typedef] = ACTIONS(2852),
    [anon_sym_extern] = ACTIONS(2852),
    [anon_sym___attribute__] = ACTIONS(2852),
    [anon_sym___attribute] = ACTIONS(2852),
    [anon_sym_noreturn] = ACTIONS(2852),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym___declspec] = ACTIONS(2852),
    [anon_sym___cdecl] = ACTIONS(2852),
    [anon_sym___clrcall] = ACTIONS(2852),
    [anon_sym___stdcall] = ACTIONS(2852),
    [anon_sym___fastcall] = ACTIONS(2852),
    [anon_sym___thiscall] = ACTIONS(2852),
    [anon_sym___vectorcall] = ACTIONS(2852),
    [anon_sym_signed] = ACTIONS(2852),
    [anon_sym_unsigned] = ACTIONS(2852),
    [anon_sym_long] = ACTIONS(2852),
    [anon_sym_short] = ACTIONS(2852),
    [anon_sym_static] = ACTIONS(2852),
    [anon_sym_auto] = ACTIONS(2852),
    [anon_sym_register] = ACTIONS(2852),
    [anon_sym_inline] = ACTIONS(2852),
    [anon_sym___inline] = ACTIONS(2852),
    [anon_sym___inline__] = ACTIONS(2852),
    [anon_sym___forceinline] = ACTIONS(2852),
    [anon_sym_thread_local] = ACTIONS(2852),
    [anon_sym___thread] = ACTIONS(2852),
    [anon_sym_CG_EXTERN] = ACTIONS(2852),
    [anon_sym_CG_INLINE] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2852),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2852),
    [anon_sym_IBOutlet] = ACTIONS(2852),
    [anon_sym_IBInspectable] = ACTIONS(2852),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2852),
    [anon_sym_NS_INLINE] = ACTIONS(2852),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2852),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2852),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2852),
    [anon_sym_const] = ACTIONS(2852),
    [anon_sym_constexpr] = ACTIONS(2852),
    [anon_sym_volatile] = ACTIONS(2852),
    [anon_sym_restrict] = ACTIONS(2852),
    [anon_sym___restrict__] = ACTIONS(2852),
    [anon_sym__Atomic] = ACTIONS(2852),
    [anon_sym__Noreturn] = ACTIONS(2852),
    [anon_sym__Nonnull] = ACTIONS(2852),
    [anon_sym_nullable] = ACTIONS(2852),
    [anon_sym__Complex] = ACTIONS(2852),
    [anon_sym__Nullable] = ACTIONS(2852),
    [anon_sym__Nullable_result] = ACTIONS(2852),
    [anon_sym__Null_unspecified] = ACTIONS(2852),
    [anon_sym___autoreleasing] = ACTIONS(2852),
    [anon_sym___block] = ACTIONS(2852),
    [anon_sym___bridge] = ACTIONS(2852),
    [anon_sym___bridge_retained] = ACTIONS(2852),
    [anon_sym___bridge_transfer] = ACTIONS(2852),
    [anon_sym___complex] = ACTIONS(2852),
    [anon_sym___const] = ACTIONS(2852),
    [anon_sym___imag] = ACTIONS(2852),
    [anon_sym___kindof] = ACTIONS(2852),
    [anon_sym___nonnull] = ACTIONS(2852),
    [anon_sym___nullable] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2852),
    [anon_sym___real] = ACTIONS(2852),
    [anon_sym___strong] = ACTIONS(2852),
    [anon_sym___unsafe_unretained] = ACTIONS(2852),
    [anon_sym___unused] = ACTIONS(2852),
    [anon_sym___weak] = ACTIONS(2852),
    [anon_sym_alignas] = ACTIONS(2852),
    [anon_sym__Alignas] = ACTIONS(2852),
    [sym_primitive_type] = ACTIONS(2852),
    [anon_sym_enum] = ACTIONS(2852),
    [anon_sym_struct] = ACTIONS(2852),
    [anon_sym_union] = ACTIONS(2852),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2852),
    [anon_sym___typeof] = ACTIONS(2852),
    [anon_sym_typeof] = ACTIONS(2852),
    [aux_sym_preproc_undef_token1] = ACTIONS(2852),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2852),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2852),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2852),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_API_AVAILABLE] = ACTIONS(2852),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_API_DEPRECATED] = ACTIONS(2852),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym___deprecated_msg] = ACTIONS(2852),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2852),
    [anon_sym_ATsynthesize] = ACTIONS(2854),
    [anon_sym_ATdynamic] = ACTIONS(2854),
    [anon_sym_ATproperty] = ACTIONS(2854),
    [anon_sym_BOOL] = ACTIONS(2852),
    [anon_sym_IMP] = ACTIONS(2852),
    [anon_sym_SEL] = ACTIONS(2852),
    [anon_sym_Class] = ACTIONS(2852),
    [anon_sym_id] = ACTIONS(2852),
  },
  [STATE(2119)] = {
    [sym_identifier] = ACTIONS(2608),
    [aux_sym_preproc_def_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token2] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2608),
    [aux_sym_preproc_else_token1] = ACTIONS(2608),
    [aux_sym_preproc_elif_token1] = ACTIONS(2608),
    [sym_preproc_directive] = ACTIONS(2608),
    [anon_sym_DASH] = ACTIONS(2610),
    [anon_sym_PLUS] = ACTIONS(2610),
    [anon_sym___extension__] = ACTIONS(2608),
    [anon_sym_typedef] = ACTIONS(2608),
    [anon_sym_extern] = ACTIONS(2608),
    [anon_sym___attribute__] = ACTIONS(2608),
    [anon_sym___attribute] = ACTIONS(2608),
    [anon_sym_noreturn] = ACTIONS(2608),
    [anon_sym_LBRACK] = ACTIONS(2610),
    [anon_sym___declspec] = ACTIONS(2608),
    [anon_sym___cdecl] = ACTIONS(2608),
    [anon_sym___clrcall] = ACTIONS(2608),
    [anon_sym___stdcall] = ACTIONS(2608),
    [anon_sym___fastcall] = ACTIONS(2608),
    [anon_sym___thiscall] = ACTIONS(2608),
    [anon_sym___vectorcall] = ACTIONS(2608),
    [anon_sym_signed] = ACTIONS(2608),
    [anon_sym_unsigned] = ACTIONS(2608),
    [anon_sym_long] = ACTIONS(2608),
    [anon_sym_short] = ACTIONS(2608),
    [anon_sym_static] = ACTIONS(2608),
    [anon_sym_auto] = ACTIONS(2608),
    [anon_sym_register] = ACTIONS(2608),
    [anon_sym_inline] = ACTIONS(2608),
    [anon_sym___inline] = ACTIONS(2608),
    [anon_sym___inline__] = ACTIONS(2608),
    [anon_sym___forceinline] = ACTIONS(2608),
    [anon_sym_thread_local] = ACTIONS(2608),
    [anon_sym___thread] = ACTIONS(2608),
    [anon_sym_CG_EXTERN] = ACTIONS(2608),
    [anon_sym_CG_INLINE] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2608),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2608),
    [anon_sym_IBOutlet] = ACTIONS(2608),
    [anon_sym_IBInspectable] = ACTIONS(2608),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2608),
    [anon_sym_NS_INLINE] = ACTIONS(2608),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2608),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2608),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2608),
    [anon_sym_const] = ACTIONS(2608),
    [anon_sym_constexpr] = ACTIONS(2608),
    [anon_sym_volatile] = ACTIONS(2608),
    [anon_sym_restrict] = ACTIONS(2608),
    [anon_sym___restrict__] = ACTIONS(2608),
    [anon_sym__Atomic] = ACTIONS(2608),
    [anon_sym__Noreturn] = ACTIONS(2608),
    [anon_sym__Nonnull] = ACTIONS(2608),
    [anon_sym_nullable] = ACTIONS(2608),
    [anon_sym__Complex] = ACTIONS(2608),
    [anon_sym__Nullable] = ACTIONS(2608),
    [anon_sym__Nullable_result] = ACTIONS(2608),
    [anon_sym__Null_unspecified] = ACTIONS(2608),
    [anon_sym___autoreleasing] = ACTIONS(2608),
    [anon_sym___block] = ACTIONS(2608),
    [anon_sym___bridge] = ACTIONS(2608),
    [anon_sym___bridge_retained] = ACTIONS(2608),
    [anon_sym___bridge_transfer] = ACTIONS(2608),
    [anon_sym___complex] = ACTIONS(2608),
    [anon_sym___const] = ACTIONS(2608),
    [anon_sym___imag] = ACTIONS(2608),
    [anon_sym___kindof] = ACTIONS(2608),
    [anon_sym___nonnull] = ACTIONS(2608),
    [anon_sym___nullable] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2608),
    [anon_sym___real] = ACTIONS(2608),
    [anon_sym___strong] = ACTIONS(2608),
    [anon_sym___unsafe_unretained] = ACTIONS(2608),
    [anon_sym___unused] = ACTIONS(2608),
    [anon_sym___weak] = ACTIONS(2608),
    [anon_sym_alignas] = ACTIONS(2608),
    [anon_sym__Alignas] = ACTIONS(2608),
    [sym_primitive_type] = ACTIONS(2608),
    [anon_sym_enum] = ACTIONS(2608),
    [anon_sym_struct] = ACTIONS(2608),
    [anon_sym_union] = ACTIONS(2608),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2608),
    [anon_sym___typeof] = ACTIONS(2608),
    [anon_sym_typeof] = ACTIONS(2608),
    [aux_sym_preproc_undef_token1] = ACTIONS(2608),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2608),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2608),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2608),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_API_AVAILABLE] = ACTIONS(2608),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_API_DEPRECATED] = ACTIONS(2608),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym___deprecated_msg] = ACTIONS(2608),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2608),
    [anon_sym_ATsynthesize] = ACTIONS(2610),
    [anon_sym_ATdynamic] = ACTIONS(2610),
    [anon_sym_ATproperty] = ACTIONS(2610),
    [anon_sym_BOOL] = ACTIONS(2608),
    [anon_sym_IMP] = ACTIONS(2608),
    [anon_sym_SEL] = ACTIONS(2608),
    [anon_sym_Class] = ACTIONS(2608),
    [anon_sym_id] = ACTIONS(2608),
  },
  [STATE(2120)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4798),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2121)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2185),
    [sym_identifier] = ACTIONS(5411),
    [aux_sym_preproc_def_token1] = ACTIONS(5411),
    [aux_sym_preproc_if_token1] = ACTIONS(5411),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5411),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5411),
    [sym_preproc_directive] = ACTIONS(5411),
    [anon_sym_DASH] = ACTIONS(5413),
    [anon_sym_PLUS] = ACTIONS(5413),
    [anon_sym_SEMI] = ACTIONS(5349),
    [anon_sym___extension__] = ACTIONS(5411),
    [anon_sym_typedef] = ACTIONS(5411),
    [anon_sym_extern] = ACTIONS(5411),
    [anon_sym___attribute__] = ACTIONS(5411),
    [anon_sym___attribute] = ACTIONS(5411),
    [anon_sym_noreturn] = ACTIONS(5411),
    [anon_sym_LBRACK] = ACTIONS(5413),
    [anon_sym___declspec] = ACTIONS(5411),
    [anon_sym___cdecl] = ACTIONS(5411),
    [anon_sym___clrcall] = ACTIONS(5411),
    [anon_sym___stdcall] = ACTIONS(5411),
    [anon_sym___fastcall] = ACTIONS(5411),
    [anon_sym___thiscall] = ACTIONS(5411),
    [anon_sym___vectorcall] = ACTIONS(5411),
    [anon_sym_signed] = ACTIONS(5411),
    [anon_sym_unsigned] = ACTIONS(5411),
    [anon_sym_long] = ACTIONS(5411),
    [anon_sym_short] = ACTIONS(5411),
    [anon_sym_static] = ACTIONS(5411),
    [anon_sym_auto] = ACTIONS(5411),
    [anon_sym_register] = ACTIONS(5411),
    [anon_sym_inline] = ACTIONS(5411),
    [anon_sym___inline] = ACTIONS(5411),
    [anon_sym___inline__] = ACTIONS(5411),
    [anon_sym___forceinline] = ACTIONS(5411),
    [anon_sym_thread_local] = ACTIONS(5411),
    [anon_sym___thread] = ACTIONS(5411),
    [anon_sym_CG_EXTERN] = ACTIONS(5411),
    [anon_sym_CG_INLINE] = ACTIONS(5411),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5411),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5411),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5411),
    [anon_sym_IBOutlet] = ACTIONS(5411),
    [anon_sym_IBInspectable] = ACTIONS(5411),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5411),
    [anon_sym_NS_INLINE] = ACTIONS(5411),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5411),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5411),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5411),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5411),
    [anon_sym_const] = ACTIONS(5411),
    [anon_sym_constexpr] = ACTIONS(5411),
    [anon_sym_volatile] = ACTIONS(5411),
    [anon_sym_restrict] = ACTIONS(5411),
    [anon_sym___restrict__] = ACTIONS(5411),
    [anon_sym__Atomic] = ACTIONS(5411),
    [anon_sym__Noreturn] = ACTIONS(5411),
    [anon_sym__Nonnull] = ACTIONS(5411),
    [anon_sym_nullable] = ACTIONS(5411),
    [anon_sym__Complex] = ACTIONS(5411),
    [anon_sym__Nullable] = ACTIONS(5411),
    [anon_sym__Nullable_result] = ACTIONS(5411),
    [anon_sym__Null_unspecified] = ACTIONS(5411),
    [anon_sym___autoreleasing] = ACTIONS(5411),
    [anon_sym___block] = ACTIONS(5411),
    [anon_sym___bridge] = ACTIONS(5411),
    [anon_sym___bridge_retained] = ACTIONS(5411),
    [anon_sym___bridge_transfer] = ACTIONS(5411),
    [anon_sym___complex] = ACTIONS(5411),
    [anon_sym___const] = ACTIONS(5411),
    [anon_sym___imag] = ACTIONS(5411),
    [anon_sym___kindof] = ACTIONS(5411),
    [anon_sym___nonnull] = ACTIONS(5411),
    [anon_sym___nullable] = ACTIONS(5411),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5411),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5411),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5411),
    [anon_sym___real] = ACTIONS(5411),
    [anon_sym___strong] = ACTIONS(5411),
    [anon_sym___unsafe_unretained] = ACTIONS(5411),
    [anon_sym___unused] = ACTIONS(5411),
    [anon_sym___weak] = ACTIONS(5411),
    [anon_sym_alignas] = ACTIONS(5411),
    [anon_sym__Alignas] = ACTIONS(5411),
    [sym_primitive_type] = ACTIONS(5411),
    [anon_sym_enum] = ACTIONS(5411),
    [anon_sym_struct] = ACTIONS(5411),
    [anon_sym_union] = ACTIONS(5411),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5411),
    [anon_sym___typeof] = ACTIONS(5411),
    [anon_sym_typeof] = ACTIONS(5411),
    [aux_sym_preproc_undef_token1] = ACTIONS(5411),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5411),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5411),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5411),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5411),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5411),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5411),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5411),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5411),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5411),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_API_AVAILABLE] = ACTIONS(5411),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_API_DEPRECATED] = ACTIONS(5411),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5411),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5411),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5411),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5411),
    [anon_sym___deprecated_msg] = ACTIONS(5411),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5411),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5411),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5411),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5411),
    [anon_sym_ATend] = ACTIONS(5413),
    [anon_sym_AToptional] = ACTIONS(5413),
    [anon_sym_ATrequired] = ACTIONS(5413),
    [anon_sym_ATproperty] = ACTIONS(5413),
    [anon_sym_BOOL] = ACTIONS(5411),
    [anon_sym_IMP] = ACTIONS(5411),
    [anon_sym_SEL] = ACTIONS(5411),
    [anon_sym_Class] = ACTIONS(5411),
    [anon_sym_id] = ACTIONS(5411),
  },
  [STATE(2122)] = {
    [sym_identifier] = ACTIONS(2688),
    [aux_sym_preproc_def_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2688),
    [sym_preproc_directive] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2690),
    [anon_sym_PLUS] = ACTIONS(2690),
    [anon_sym___extension__] = ACTIONS(2688),
    [anon_sym_typedef] = ACTIONS(2688),
    [anon_sym_extern] = ACTIONS(2688),
    [anon_sym___attribute__] = ACTIONS(2688),
    [anon_sym___attribute] = ACTIONS(2688),
    [anon_sym_noreturn] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym___declspec] = ACTIONS(2688),
    [anon_sym___cdecl] = ACTIONS(2688),
    [anon_sym___clrcall] = ACTIONS(2688),
    [anon_sym___stdcall] = ACTIONS(2688),
    [anon_sym___fastcall] = ACTIONS(2688),
    [anon_sym___thiscall] = ACTIONS(2688),
    [anon_sym___vectorcall] = ACTIONS(2688),
    [anon_sym_signed] = ACTIONS(2688),
    [anon_sym_unsigned] = ACTIONS(2688),
    [anon_sym_long] = ACTIONS(2688),
    [anon_sym_short] = ACTIONS(2688),
    [anon_sym_static] = ACTIONS(2688),
    [anon_sym_auto] = ACTIONS(2688),
    [anon_sym_register] = ACTIONS(2688),
    [anon_sym_inline] = ACTIONS(2688),
    [anon_sym___inline] = ACTIONS(2688),
    [anon_sym___inline__] = ACTIONS(2688),
    [anon_sym___forceinline] = ACTIONS(2688),
    [anon_sym_thread_local] = ACTIONS(2688),
    [anon_sym___thread] = ACTIONS(2688),
    [anon_sym_CG_EXTERN] = ACTIONS(2688),
    [anon_sym_CG_INLINE] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2688),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2688),
    [anon_sym_IBOutlet] = ACTIONS(2688),
    [anon_sym_IBInspectable] = ACTIONS(2688),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2688),
    [anon_sym_NS_INLINE] = ACTIONS(2688),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2688),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2688),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2688),
    [anon_sym_const] = ACTIONS(2688),
    [anon_sym_constexpr] = ACTIONS(2688),
    [anon_sym_volatile] = ACTIONS(2688),
    [anon_sym_restrict] = ACTIONS(2688),
    [anon_sym___restrict__] = ACTIONS(2688),
    [anon_sym__Atomic] = ACTIONS(2688),
    [anon_sym__Noreturn] = ACTIONS(2688),
    [anon_sym__Nonnull] = ACTIONS(2688),
    [anon_sym_nullable] = ACTIONS(2688),
    [anon_sym__Complex] = ACTIONS(2688),
    [anon_sym__Nullable] = ACTIONS(2688),
    [anon_sym__Nullable_result] = ACTIONS(2688),
    [anon_sym__Null_unspecified] = ACTIONS(2688),
    [anon_sym___autoreleasing] = ACTIONS(2688),
    [anon_sym___block] = ACTIONS(2688),
    [anon_sym___bridge] = ACTIONS(2688),
    [anon_sym___bridge_retained] = ACTIONS(2688),
    [anon_sym___bridge_transfer] = ACTIONS(2688),
    [anon_sym___complex] = ACTIONS(2688),
    [anon_sym___const] = ACTIONS(2688),
    [anon_sym___imag] = ACTIONS(2688),
    [anon_sym___kindof] = ACTIONS(2688),
    [anon_sym___nonnull] = ACTIONS(2688),
    [anon_sym___nullable] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2688),
    [anon_sym___real] = ACTIONS(2688),
    [anon_sym___strong] = ACTIONS(2688),
    [anon_sym___unsafe_unretained] = ACTIONS(2688),
    [anon_sym___unused] = ACTIONS(2688),
    [anon_sym___weak] = ACTIONS(2688),
    [anon_sym_alignas] = ACTIONS(2688),
    [anon_sym__Alignas] = ACTIONS(2688),
    [sym_primitive_type] = ACTIONS(2688),
    [anon_sym_enum] = ACTIONS(2688),
    [anon_sym_struct] = ACTIONS(2688),
    [anon_sym_union] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2688),
    [anon_sym___typeof] = ACTIONS(2688),
    [anon_sym_typeof] = ACTIONS(2688),
    [aux_sym_preproc_undef_token1] = ACTIONS(2688),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2688),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2688),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2688),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_API_AVAILABLE] = ACTIONS(2688),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_API_DEPRECATED] = ACTIONS(2688),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym___deprecated_msg] = ACTIONS(2688),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2688),
    [anon_sym_ATend] = ACTIONS(2690),
    [anon_sym_AToptional] = ACTIONS(2690),
    [anon_sym_ATrequired] = ACTIONS(2690),
    [anon_sym_ATsynthesize] = ACTIONS(2690),
    [anon_sym_ATdynamic] = ACTIONS(2690),
    [anon_sym_ATproperty] = ACTIONS(2690),
    [anon_sym_BOOL] = ACTIONS(2688),
    [anon_sym_IMP] = ACTIONS(2688),
    [anon_sym_SEL] = ACTIONS(2688),
    [anon_sym_Class] = ACTIONS(2688),
    [anon_sym_id] = ACTIONS(2688),
  },
  [STATE(2123)] = {
    [sym_identifier] = ACTIONS(2584),
    [aux_sym_preproc_def_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2584),
    [sym_preproc_directive] = ACTIONS(2584),
    [anon_sym_DASH] = ACTIONS(2586),
    [anon_sym_PLUS] = ACTIONS(2586),
    [anon_sym___extension__] = ACTIONS(2584),
    [anon_sym_typedef] = ACTIONS(2584),
    [anon_sym_extern] = ACTIONS(2584),
    [anon_sym___attribute__] = ACTIONS(2584),
    [anon_sym___attribute] = ACTIONS(2584),
    [anon_sym_noreturn] = ACTIONS(2584),
    [anon_sym_LBRACK] = ACTIONS(2586),
    [anon_sym___declspec] = ACTIONS(2584),
    [anon_sym___cdecl] = ACTIONS(2584),
    [anon_sym___clrcall] = ACTIONS(2584),
    [anon_sym___stdcall] = ACTIONS(2584),
    [anon_sym___fastcall] = ACTIONS(2584),
    [anon_sym___thiscall] = ACTIONS(2584),
    [anon_sym___vectorcall] = ACTIONS(2584),
    [anon_sym_signed] = ACTIONS(2584),
    [anon_sym_unsigned] = ACTIONS(2584),
    [anon_sym_long] = ACTIONS(2584),
    [anon_sym_short] = ACTIONS(2584),
    [anon_sym_static] = ACTIONS(2584),
    [anon_sym_auto] = ACTIONS(2584),
    [anon_sym_register] = ACTIONS(2584),
    [anon_sym_inline] = ACTIONS(2584),
    [anon_sym___inline] = ACTIONS(2584),
    [anon_sym___inline__] = ACTIONS(2584),
    [anon_sym___forceinline] = ACTIONS(2584),
    [anon_sym_thread_local] = ACTIONS(2584),
    [anon_sym___thread] = ACTIONS(2584),
    [anon_sym_CG_EXTERN] = ACTIONS(2584),
    [anon_sym_CG_INLINE] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2584),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2584),
    [anon_sym_IBOutlet] = ACTIONS(2584),
    [anon_sym_IBInspectable] = ACTIONS(2584),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2584),
    [anon_sym_NS_INLINE] = ACTIONS(2584),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2584),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2584),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2584),
    [anon_sym_const] = ACTIONS(2584),
    [anon_sym_constexpr] = ACTIONS(2584),
    [anon_sym_volatile] = ACTIONS(2584),
    [anon_sym_restrict] = ACTIONS(2584),
    [anon_sym___restrict__] = ACTIONS(2584),
    [anon_sym__Atomic] = ACTIONS(2584),
    [anon_sym__Noreturn] = ACTIONS(2584),
    [anon_sym__Nonnull] = ACTIONS(2584),
    [anon_sym_nullable] = ACTIONS(2584),
    [anon_sym__Complex] = ACTIONS(2584),
    [anon_sym__Nullable] = ACTIONS(2584),
    [anon_sym__Nullable_result] = ACTIONS(2584),
    [anon_sym__Null_unspecified] = ACTIONS(2584),
    [anon_sym___autoreleasing] = ACTIONS(2584),
    [anon_sym___block] = ACTIONS(2584),
    [anon_sym___bridge] = ACTIONS(2584),
    [anon_sym___bridge_retained] = ACTIONS(2584),
    [anon_sym___bridge_transfer] = ACTIONS(2584),
    [anon_sym___complex] = ACTIONS(2584),
    [anon_sym___const] = ACTIONS(2584),
    [anon_sym___imag] = ACTIONS(2584),
    [anon_sym___kindof] = ACTIONS(2584),
    [anon_sym___nonnull] = ACTIONS(2584),
    [anon_sym___nullable] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2584),
    [anon_sym___real] = ACTIONS(2584),
    [anon_sym___strong] = ACTIONS(2584),
    [anon_sym___unsafe_unretained] = ACTIONS(2584),
    [anon_sym___unused] = ACTIONS(2584),
    [anon_sym___weak] = ACTIONS(2584),
    [anon_sym_alignas] = ACTIONS(2584),
    [anon_sym__Alignas] = ACTIONS(2584),
    [sym_primitive_type] = ACTIONS(2584),
    [anon_sym_enum] = ACTIONS(2584),
    [anon_sym_struct] = ACTIONS(2584),
    [anon_sym_union] = ACTIONS(2584),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2584),
    [anon_sym___typeof] = ACTIONS(2584),
    [anon_sym_typeof] = ACTIONS(2584),
    [aux_sym_preproc_undef_token1] = ACTIONS(2584),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2584),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2584),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2584),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_API_AVAILABLE] = ACTIONS(2584),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_API_DEPRECATED] = ACTIONS(2584),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym___deprecated_msg] = ACTIONS(2584),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2584),
    [anon_sym_ATend] = ACTIONS(2586),
    [anon_sym_AToptional] = ACTIONS(2586),
    [anon_sym_ATrequired] = ACTIONS(2586),
    [anon_sym_ATsynthesize] = ACTIONS(2586),
    [anon_sym_ATdynamic] = ACTIONS(2586),
    [anon_sym_ATproperty] = ACTIONS(2586),
    [anon_sym_BOOL] = ACTIONS(2584),
    [anon_sym_IMP] = ACTIONS(2584),
    [anon_sym_SEL] = ACTIONS(2584),
    [anon_sym_Class] = ACTIONS(2584),
    [anon_sym_id] = ACTIONS(2584),
  },
  [STATE(2124)] = {
    [sym_identifier] = ACTIONS(2634),
    [aux_sym_preproc_def_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token2] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2634),
    [aux_sym_preproc_else_token1] = ACTIONS(2634),
    [aux_sym_preproc_elif_token1] = ACTIONS(2634),
    [sym_preproc_directive] = ACTIONS(2634),
    [anon_sym_DASH] = ACTIONS(2636),
    [anon_sym_PLUS] = ACTIONS(2636),
    [anon_sym___extension__] = ACTIONS(2634),
    [anon_sym_typedef] = ACTIONS(2634),
    [anon_sym_extern] = ACTIONS(2634),
    [anon_sym___attribute__] = ACTIONS(2634),
    [anon_sym___attribute] = ACTIONS(2634),
    [anon_sym_noreturn] = ACTIONS(2634),
    [anon_sym_LBRACK] = ACTIONS(2636),
    [anon_sym___declspec] = ACTIONS(2634),
    [anon_sym___cdecl] = ACTIONS(2634),
    [anon_sym___clrcall] = ACTIONS(2634),
    [anon_sym___stdcall] = ACTIONS(2634),
    [anon_sym___fastcall] = ACTIONS(2634),
    [anon_sym___thiscall] = ACTIONS(2634),
    [anon_sym___vectorcall] = ACTIONS(2634),
    [anon_sym_signed] = ACTIONS(2634),
    [anon_sym_unsigned] = ACTIONS(2634),
    [anon_sym_long] = ACTIONS(2634),
    [anon_sym_short] = ACTIONS(2634),
    [anon_sym_static] = ACTIONS(2634),
    [anon_sym_auto] = ACTIONS(2634),
    [anon_sym_register] = ACTIONS(2634),
    [anon_sym_inline] = ACTIONS(2634),
    [anon_sym___inline] = ACTIONS(2634),
    [anon_sym___inline__] = ACTIONS(2634),
    [anon_sym___forceinline] = ACTIONS(2634),
    [anon_sym_thread_local] = ACTIONS(2634),
    [anon_sym___thread] = ACTIONS(2634),
    [anon_sym_CG_EXTERN] = ACTIONS(2634),
    [anon_sym_CG_INLINE] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2634),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2634),
    [anon_sym_IBOutlet] = ACTIONS(2634),
    [anon_sym_IBInspectable] = ACTIONS(2634),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2634),
    [anon_sym_NS_INLINE] = ACTIONS(2634),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2634),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2634),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2634),
    [anon_sym_const] = ACTIONS(2634),
    [anon_sym_constexpr] = ACTIONS(2634),
    [anon_sym_volatile] = ACTIONS(2634),
    [anon_sym_restrict] = ACTIONS(2634),
    [anon_sym___restrict__] = ACTIONS(2634),
    [anon_sym__Atomic] = ACTIONS(2634),
    [anon_sym__Noreturn] = ACTIONS(2634),
    [anon_sym__Nonnull] = ACTIONS(2634),
    [anon_sym_nullable] = ACTIONS(2634),
    [anon_sym__Complex] = ACTIONS(2634),
    [anon_sym__Nullable] = ACTIONS(2634),
    [anon_sym__Nullable_result] = ACTIONS(2634),
    [anon_sym__Null_unspecified] = ACTIONS(2634),
    [anon_sym___autoreleasing] = ACTIONS(2634),
    [anon_sym___block] = ACTIONS(2634),
    [anon_sym___bridge] = ACTIONS(2634),
    [anon_sym___bridge_retained] = ACTIONS(2634),
    [anon_sym___bridge_transfer] = ACTIONS(2634),
    [anon_sym___complex] = ACTIONS(2634),
    [anon_sym___const] = ACTIONS(2634),
    [anon_sym___imag] = ACTIONS(2634),
    [anon_sym___kindof] = ACTIONS(2634),
    [anon_sym___nonnull] = ACTIONS(2634),
    [anon_sym___nullable] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2634),
    [anon_sym___real] = ACTIONS(2634),
    [anon_sym___strong] = ACTIONS(2634),
    [anon_sym___unsafe_unretained] = ACTIONS(2634),
    [anon_sym___unused] = ACTIONS(2634),
    [anon_sym___weak] = ACTIONS(2634),
    [anon_sym_alignas] = ACTIONS(2634),
    [anon_sym__Alignas] = ACTIONS(2634),
    [sym_primitive_type] = ACTIONS(2634),
    [anon_sym_enum] = ACTIONS(2634),
    [anon_sym_struct] = ACTIONS(2634),
    [anon_sym_union] = ACTIONS(2634),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2634),
    [anon_sym___typeof] = ACTIONS(2634),
    [anon_sym_typeof] = ACTIONS(2634),
    [aux_sym_preproc_undef_token1] = ACTIONS(2634),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2634),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2634),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2634),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_API_AVAILABLE] = ACTIONS(2634),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_API_DEPRECATED] = ACTIONS(2634),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym___deprecated_msg] = ACTIONS(2634),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2634),
    [anon_sym_ATsynthesize] = ACTIONS(2636),
    [anon_sym_ATdynamic] = ACTIONS(2636),
    [anon_sym_ATproperty] = ACTIONS(2636),
    [anon_sym_BOOL] = ACTIONS(2634),
    [anon_sym_IMP] = ACTIONS(2634),
    [anon_sym_SEL] = ACTIONS(2634),
    [anon_sym_Class] = ACTIONS(2634),
    [anon_sym_id] = ACTIONS(2634),
  },
  [STATE(2125)] = {
    [sym_identifier] = ACTIONS(2668),
    [aux_sym_preproc_def_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token2] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2668),
    [aux_sym_preproc_else_token1] = ACTIONS(2668),
    [aux_sym_preproc_elif_token1] = ACTIONS(2668),
    [sym_preproc_directive] = ACTIONS(2668),
    [anon_sym_DASH] = ACTIONS(2670),
    [anon_sym_PLUS] = ACTIONS(2670),
    [anon_sym___extension__] = ACTIONS(2668),
    [anon_sym_typedef] = ACTIONS(2668),
    [anon_sym_extern] = ACTIONS(2668),
    [anon_sym___attribute__] = ACTIONS(2668),
    [anon_sym___attribute] = ACTIONS(2668),
    [anon_sym_noreturn] = ACTIONS(2668),
    [anon_sym_LBRACK] = ACTIONS(2670),
    [anon_sym___declspec] = ACTIONS(2668),
    [anon_sym___cdecl] = ACTIONS(2668),
    [anon_sym___clrcall] = ACTIONS(2668),
    [anon_sym___stdcall] = ACTIONS(2668),
    [anon_sym___fastcall] = ACTIONS(2668),
    [anon_sym___thiscall] = ACTIONS(2668),
    [anon_sym___vectorcall] = ACTIONS(2668),
    [anon_sym_signed] = ACTIONS(2668),
    [anon_sym_unsigned] = ACTIONS(2668),
    [anon_sym_long] = ACTIONS(2668),
    [anon_sym_short] = ACTIONS(2668),
    [anon_sym_static] = ACTIONS(2668),
    [anon_sym_auto] = ACTIONS(2668),
    [anon_sym_register] = ACTIONS(2668),
    [anon_sym_inline] = ACTIONS(2668),
    [anon_sym___inline] = ACTIONS(2668),
    [anon_sym___inline__] = ACTIONS(2668),
    [anon_sym___forceinline] = ACTIONS(2668),
    [anon_sym_thread_local] = ACTIONS(2668),
    [anon_sym___thread] = ACTIONS(2668),
    [anon_sym_CG_EXTERN] = ACTIONS(2668),
    [anon_sym_CG_INLINE] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2668),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2668),
    [anon_sym_IBOutlet] = ACTIONS(2668),
    [anon_sym_IBInspectable] = ACTIONS(2668),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2668),
    [anon_sym_NS_INLINE] = ACTIONS(2668),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2668),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2668),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2668),
    [anon_sym_const] = ACTIONS(2668),
    [anon_sym_constexpr] = ACTIONS(2668),
    [anon_sym_volatile] = ACTIONS(2668),
    [anon_sym_restrict] = ACTIONS(2668),
    [anon_sym___restrict__] = ACTIONS(2668),
    [anon_sym__Atomic] = ACTIONS(2668),
    [anon_sym__Noreturn] = ACTIONS(2668),
    [anon_sym__Nonnull] = ACTIONS(2668),
    [anon_sym_nullable] = ACTIONS(2668),
    [anon_sym__Complex] = ACTIONS(2668),
    [anon_sym__Nullable] = ACTIONS(2668),
    [anon_sym__Nullable_result] = ACTIONS(2668),
    [anon_sym__Null_unspecified] = ACTIONS(2668),
    [anon_sym___autoreleasing] = ACTIONS(2668),
    [anon_sym___block] = ACTIONS(2668),
    [anon_sym___bridge] = ACTIONS(2668),
    [anon_sym___bridge_retained] = ACTIONS(2668),
    [anon_sym___bridge_transfer] = ACTIONS(2668),
    [anon_sym___complex] = ACTIONS(2668),
    [anon_sym___const] = ACTIONS(2668),
    [anon_sym___imag] = ACTIONS(2668),
    [anon_sym___kindof] = ACTIONS(2668),
    [anon_sym___nonnull] = ACTIONS(2668),
    [anon_sym___nullable] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2668),
    [anon_sym___real] = ACTIONS(2668),
    [anon_sym___strong] = ACTIONS(2668),
    [anon_sym___unsafe_unretained] = ACTIONS(2668),
    [anon_sym___unused] = ACTIONS(2668),
    [anon_sym___weak] = ACTIONS(2668),
    [anon_sym_alignas] = ACTIONS(2668),
    [anon_sym__Alignas] = ACTIONS(2668),
    [sym_primitive_type] = ACTIONS(2668),
    [anon_sym_enum] = ACTIONS(2668),
    [anon_sym_struct] = ACTIONS(2668),
    [anon_sym_union] = ACTIONS(2668),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2668),
    [anon_sym___typeof] = ACTIONS(2668),
    [anon_sym_typeof] = ACTIONS(2668),
    [aux_sym_preproc_undef_token1] = ACTIONS(2668),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2668),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2668),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2668),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_API_AVAILABLE] = ACTIONS(2668),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_API_DEPRECATED] = ACTIONS(2668),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym___deprecated_msg] = ACTIONS(2668),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2668),
    [anon_sym_ATsynthesize] = ACTIONS(2670),
    [anon_sym_ATdynamic] = ACTIONS(2670),
    [anon_sym_ATproperty] = ACTIONS(2670),
    [anon_sym_BOOL] = ACTIONS(2668),
    [anon_sym_IMP] = ACTIONS(2668),
    [anon_sym_SEL] = ACTIONS(2668),
    [anon_sym_Class] = ACTIONS(2668),
    [anon_sym_id] = ACTIONS(2668),
  },
  [STATE(2126)] = {
    [sym__declaration_modifiers] = STATE(2534),
    [sym_attribute_specifier] = STATE(2534),
    [sym_attribute_declaration] = STATE(2534),
    [sym_ms_declspec_modifier] = STATE(2534),
    [sym_ms_based_modifier] = STATE(6199),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2572),
    [sym__declarator] = STATE(3049),
    [sym__abstract_declarator] = STATE(4978),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3065),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3065),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3065),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(2534),
    [sym_type_qualifier] = STATE(2534),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4271),
    [sym_availability_attribute_specifier] = STATE(2534),
    [sym_block_pointer_declarator] = STATE(3065),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2534),
    [aux_sym_pointer_declarator_repeat1] = STATE(2572),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_COMMA] = ACTIONS(5314),
    [anon_sym_RPAREN] = ACTIONS(5314),
    [anon_sym_LPAREN2] = ACTIONS(5417),
    [anon_sym_STAR] = ACTIONS(5419),
    [anon_sym_CARET] = ACTIONS(5421),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2127)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4811),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2128)] = {
    [sym_identifier] = ACTIONS(2634),
    [aux_sym_preproc_def_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2634),
    [sym_preproc_directive] = ACTIONS(2634),
    [anon_sym_DASH] = ACTIONS(2636),
    [anon_sym_PLUS] = ACTIONS(2636),
    [anon_sym___extension__] = ACTIONS(2634),
    [anon_sym_typedef] = ACTIONS(2634),
    [anon_sym_extern] = ACTIONS(2634),
    [anon_sym___attribute__] = ACTIONS(2634),
    [anon_sym___attribute] = ACTIONS(2634),
    [anon_sym_noreturn] = ACTIONS(2634),
    [anon_sym_LBRACK] = ACTIONS(2636),
    [anon_sym___declspec] = ACTIONS(2634),
    [anon_sym___cdecl] = ACTIONS(2634),
    [anon_sym___clrcall] = ACTIONS(2634),
    [anon_sym___stdcall] = ACTIONS(2634),
    [anon_sym___fastcall] = ACTIONS(2634),
    [anon_sym___thiscall] = ACTIONS(2634),
    [anon_sym___vectorcall] = ACTIONS(2634),
    [anon_sym_signed] = ACTIONS(2634),
    [anon_sym_unsigned] = ACTIONS(2634),
    [anon_sym_long] = ACTIONS(2634),
    [anon_sym_short] = ACTIONS(2634),
    [anon_sym_static] = ACTIONS(2634),
    [anon_sym_auto] = ACTIONS(2634),
    [anon_sym_register] = ACTIONS(2634),
    [anon_sym_inline] = ACTIONS(2634),
    [anon_sym___inline] = ACTIONS(2634),
    [anon_sym___inline__] = ACTIONS(2634),
    [anon_sym___forceinline] = ACTIONS(2634),
    [anon_sym_thread_local] = ACTIONS(2634),
    [anon_sym___thread] = ACTIONS(2634),
    [anon_sym_CG_EXTERN] = ACTIONS(2634),
    [anon_sym_CG_INLINE] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2634),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2634),
    [anon_sym_IBOutlet] = ACTIONS(2634),
    [anon_sym_IBInspectable] = ACTIONS(2634),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2634),
    [anon_sym_NS_INLINE] = ACTIONS(2634),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2634),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2634),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2634),
    [anon_sym_const] = ACTIONS(2634),
    [anon_sym_constexpr] = ACTIONS(2634),
    [anon_sym_volatile] = ACTIONS(2634),
    [anon_sym_restrict] = ACTIONS(2634),
    [anon_sym___restrict__] = ACTIONS(2634),
    [anon_sym__Atomic] = ACTIONS(2634),
    [anon_sym__Noreturn] = ACTIONS(2634),
    [anon_sym__Nonnull] = ACTIONS(2634),
    [anon_sym_nullable] = ACTIONS(2634),
    [anon_sym__Complex] = ACTIONS(2634),
    [anon_sym__Nullable] = ACTIONS(2634),
    [anon_sym__Nullable_result] = ACTIONS(2634),
    [anon_sym__Null_unspecified] = ACTIONS(2634),
    [anon_sym___autoreleasing] = ACTIONS(2634),
    [anon_sym___block] = ACTIONS(2634),
    [anon_sym___bridge] = ACTIONS(2634),
    [anon_sym___bridge_retained] = ACTIONS(2634),
    [anon_sym___bridge_transfer] = ACTIONS(2634),
    [anon_sym___complex] = ACTIONS(2634),
    [anon_sym___const] = ACTIONS(2634),
    [anon_sym___imag] = ACTIONS(2634),
    [anon_sym___kindof] = ACTIONS(2634),
    [anon_sym___nonnull] = ACTIONS(2634),
    [anon_sym___nullable] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2634),
    [anon_sym___real] = ACTIONS(2634),
    [anon_sym___strong] = ACTIONS(2634),
    [anon_sym___unsafe_unretained] = ACTIONS(2634),
    [anon_sym___unused] = ACTIONS(2634),
    [anon_sym___weak] = ACTIONS(2634),
    [anon_sym_alignas] = ACTIONS(2634),
    [anon_sym__Alignas] = ACTIONS(2634),
    [sym_primitive_type] = ACTIONS(2634),
    [anon_sym_enum] = ACTIONS(2634),
    [anon_sym_struct] = ACTIONS(2634),
    [anon_sym_union] = ACTIONS(2634),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2634),
    [anon_sym___typeof] = ACTIONS(2634),
    [anon_sym_typeof] = ACTIONS(2634),
    [aux_sym_preproc_undef_token1] = ACTIONS(2634),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2634),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2634),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2634),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_API_AVAILABLE] = ACTIONS(2634),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_API_DEPRECATED] = ACTIONS(2634),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym___deprecated_msg] = ACTIONS(2634),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2634),
    [anon_sym_ATend] = ACTIONS(2636),
    [anon_sym_AToptional] = ACTIONS(2636),
    [anon_sym_ATrequired] = ACTIONS(2636),
    [anon_sym_ATsynthesize] = ACTIONS(2636),
    [anon_sym_ATdynamic] = ACTIONS(2636),
    [anon_sym_ATproperty] = ACTIONS(2636),
    [anon_sym_BOOL] = ACTIONS(2634),
    [anon_sym_IMP] = ACTIONS(2634),
    [anon_sym_SEL] = ACTIONS(2634),
    [anon_sym_Class] = ACTIONS(2634),
    [anon_sym_id] = ACTIONS(2634),
  },
  [STATE(2129)] = {
    [sym_identifier] = ACTIONS(2668),
    [aux_sym_preproc_def_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2668),
    [sym_preproc_directive] = ACTIONS(2668),
    [anon_sym_DASH] = ACTIONS(2670),
    [anon_sym_PLUS] = ACTIONS(2670),
    [anon_sym___extension__] = ACTIONS(2668),
    [anon_sym_typedef] = ACTIONS(2668),
    [anon_sym_extern] = ACTIONS(2668),
    [anon_sym___attribute__] = ACTIONS(2668),
    [anon_sym___attribute] = ACTIONS(2668),
    [anon_sym_noreturn] = ACTIONS(2668),
    [anon_sym_LBRACK] = ACTIONS(2670),
    [anon_sym___declspec] = ACTIONS(2668),
    [anon_sym___cdecl] = ACTIONS(2668),
    [anon_sym___clrcall] = ACTIONS(2668),
    [anon_sym___stdcall] = ACTIONS(2668),
    [anon_sym___fastcall] = ACTIONS(2668),
    [anon_sym___thiscall] = ACTIONS(2668),
    [anon_sym___vectorcall] = ACTIONS(2668),
    [anon_sym_signed] = ACTIONS(2668),
    [anon_sym_unsigned] = ACTIONS(2668),
    [anon_sym_long] = ACTIONS(2668),
    [anon_sym_short] = ACTIONS(2668),
    [anon_sym_static] = ACTIONS(2668),
    [anon_sym_auto] = ACTIONS(2668),
    [anon_sym_register] = ACTIONS(2668),
    [anon_sym_inline] = ACTIONS(2668),
    [anon_sym___inline] = ACTIONS(2668),
    [anon_sym___inline__] = ACTIONS(2668),
    [anon_sym___forceinline] = ACTIONS(2668),
    [anon_sym_thread_local] = ACTIONS(2668),
    [anon_sym___thread] = ACTIONS(2668),
    [anon_sym_CG_EXTERN] = ACTIONS(2668),
    [anon_sym_CG_INLINE] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2668),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2668),
    [anon_sym_IBOutlet] = ACTIONS(2668),
    [anon_sym_IBInspectable] = ACTIONS(2668),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2668),
    [anon_sym_NS_INLINE] = ACTIONS(2668),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2668),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2668),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2668),
    [anon_sym_const] = ACTIONS(2668),
    [anon_sym_constexpr] = ACTIONS(2668),
    [anon_sym_volatile] = ACTIONS(2668),
    [anon_sym_restrict] = ACTIONS(2668),
    [anon_sym___restrict__] = ACTIONS(2668),
    [anon_sym__Atomic] = ACTIONS(2668),
    [anon_sym__Noreturn] = ACTIONS(2668),
    [anon_sym__Nonnull] = ACTIONS(2668),
    [anon_sym_nullable] = ACTIONS(2668),
    [anon_sym__Complex] = ACTIONS(2668),
    [anon_sym__Nullable] = ACTIONS(2668),
    [anon_sym__Nullable_result] = ACTIONS(2668),
    [anon_sym__Null_unspecified] = ACTIONS(2668),
    [anon_sym___autoreleasing] = ACTIONS(2668),
    [anon_sym___block] = ACTIONS(2668),
    [anon_sym___bridge] = ACTIONS(2668),
    [anon_sym___bridge_retained] = ACTIONS(2668),
    [anon_sym___bridge_transfer] = ACTIONS(2668),
    [anon_sym___complex] = ACTIONS(2668),
    [anon_sym___const] = ACTIONS(2668),
    [anon_sym___imag] = ACTIONS(2668),
    [anon_sym___kindof] = ACTIONS(2668),
    [anon_sym___nonnull] = ACTIONS(2668),
    [anon_sym___nullable] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2668),
    [anon_sym___real] = ACTIONS(2668),
    [anon_sym___strong] = ACTIONS(2668),
    [anon_sym___unsafe_unretained] = ACTIONS(2668),
    [anon_sym___unused] = ACTIONS(2668),
    [anon_sym___weak] = ACTIONS(2668),
    [anon_sym_alignas] = ACTIONS(2668),
    [anon_sym__Alignas] = ACTIONS(2668),
    [sym_primitive_type] = ACTIONS(2668),
    [anon_sym_enum] = ACTIONS(2668),
    [anon_sym_struct] = ACTIONS(2668),
    [anon_sym_union] = ACTIONS(2668),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2668),
    [anon_sym___typeof] = ACTIONS(2668),
    [anon_sym_typeof] = ACTIONS(2668),
    [aux_sym_preproc_undef_token1] = ACTIONS(2668),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2668),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2668),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2668),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_API_AVAILABLE] = ACTIONS(2668),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_API_DEPRECATED] = ACTIONS(2668),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym___deprecated_msg] = ACTIONS(2668),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2668),
    [anon_sym_ATend] = ACTIONS(2670),
    [anon_sym_AToptional] = ACTIONS(2670),
    [anon_sym_ATrequired] = ACTIONS(2670),
    [anon_sym_ATsynthesize] = ACTIONS(2670),
    [anon_sym_ATdynamic] = ACTIONS(2670),
    [anon_sym_ATproperty] = ACTIONS(2670),
    [anon_sym_BOOL] = ACTIONS(2668),
    [anon_sym_IMP] = ACTIONS(2668),
    [anon_sym_SEL] = ACTIONS(2668),
    [anon_sym_Class] = ACTIONS(2668),
    [anon_sym_id] = ACTIONS(2668),
  },
  [STATE(2130)] = {
    [sym_identifier] = ACTIONS(2604),
    [aux_sym_preproc_def_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token2] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2604),
    [aux_sym_preproc_else_token1] = ACTIONS(2604),
    [aux_sym_preproc_elif_token1] = ACTIONS(2604),
    [sym_preproc_directive] = ACTIONS(2604),
    [anon_sym_DASH] = ACTIONS(2606),
    [anon_sym_PLUS] = ACTIONS(2606),
    [anon_sym___extension__] = ACTIONS(2604),
    [anon_sym_typedef] = ACTIONS(2604),
    [anon_sym_extern] = ACTIONS(2604),
    [anon_sym___attribute__] = ACTIONS(2604),
    [anon_sym___attribute] = ACTIONS(2604),
    [anon_sym_noreturn] = ACTIONS(2604),
    [anon_sym_LBRACK] = ACTIONS(2606),
    [anon_sym___declspec] = ACTIONS(2604),
    [anon_sym___cdecl] = ACTIONS(2604),
    [anon_sym___clrcall] = ACTIONS(2604),
    [anon_sym___stdcall] = ACTIONS(2604),
    [anon_sym___fastcall] = ACTIONS(2604),
    [anon_sym___thiscall] = ACTIONS(2604),
    [anon_sym___vectorcall] = ACTIONS(2604),
    [anon_sym_signed] = ACTIONS(2604),
    [anon_sym_unsigned] = ACTIONS(2604),
    [anon_sym_long] = ACTIONS(2604),
    [anon_sym_short] = ACTIONS(2604),
    [anon_sym_static] = ACTIONS(2604),
    [anon_sym_auto] = ACTIONS(2604),
    [anon_sym_register] = ACTIONS(2604),
    [anon_sym_inline] = ACTIONS(2604),
    [anon_sym___inline] = ACTIONS(2604),
    [anon_sym___inline__] = ACTIONS(2604),
    [anon_sym___forceinline] = ACTIONS(2604),
    [anon_sym_thread_local] = ACTIONS(2604),
    [anon_sym___thread] = ACTIONS(2604),
    [anon_sym_CG_EXTERN] = ACTIONS(2604),
    [anon_sym_CG_INLINE] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2604),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2604),
    [anon_sym_IBOutlet] = ACTIONS(2604),
    [anon_sym_IBInspectable] = ACTIONS(2604),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2604),
    [anon_sym_NS_INLINE] = ACTIONS(2604),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2604),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2604),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2604),
    [anon_sym_const] = ACTIONS(2604),
    [anon_sym_constexpr] = ACTIONS(2604),
    [anon_sym_volatile] = ACTIONS(2604),
    [anon_sym_restrict] = ACTIONS(2604),
    [anon_sym___restrict__] = ACTIONS(2604),
    [anon_sym__Atomic] = ACTIONS(2604),
    [anon_sym__Noreturn] = ACTIONS(2604),
    [anon_sym__Nonnull] = ACTIONS(2604),
    [anon_sym_nullable] = ACTIONS(2604),
    [anon_sym__Complex] = ACTIONS(2604),
    [anon_sym__Nullable] = ACTIONS(2604),
    [anon_sym__Nullable_result] = ACTIONS(2604),
    [anon_sym__Null_unspecified] = ACTIONS(2604),
    [anon_sym___autoreleasing] = ACTIONS(2604),
    [anon_sym___block] = ACTIONS(2604),
    [anon_sym___bridge] = ACTIONS(2604),
    [anon_sym___bridge_retained] = ACTIONS(2604),
    [anon_sym___bridge_transfer] = ACTIONS(2604),
    [anon_sym___complex] = ACTIONS(2604),
    [anon_sym___const] = ACTIONS(2604),
    [anon_sym___imag] = ACTIONS(2604),
    [anon_sym___kindof] = ACTIONS(2604),
    [anon_sym___nonnull] = ACTIONS(2604),
    [anon_sym___nullable] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2604),
    [anon_sym___real] = ACTIONS(2604),
    [anon_sym___strong] = ACTIONS(2604),
    [anon_sym___unsafe_unretained] = ACTIONS(2604),
    [anon_sym___unused] = ACTIONS(2604),
    [anon_sym___weak] = ACTIONS(2604),
    [anon_sym_alignas] = ACTIONS(2604),
    [anon_sym__Alignas] = ACTIONS(2604),
    [sym_primitive_type] = ACTIONS(2604),
    [anon_sym_enum] = ACTIONS(2604),
    [anon_sym_struct] = ACTIONS(2604),
    [anon_sym_union] = ACTIONS(2604),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2604),
    [anon_sym___typeof] = ACTIONS(2604),
    [anon_sym_typeof] = ACTIONS(2604),
    [aux_sym_preproc_undef_token1] = ACTIONS(2604),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2604),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2604),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2604),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_API_AVAILABLE] = ACTIONS(2604),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_API_DEPRECATED] = ACTIONS(2604),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym___deprecated_msg] = ACTIONS(2604),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2604),
    [anon_sym_ATsynthesize] = ACTIONS(2606),
    [anon_sym_ATdynamic] = ACTIONS(2606),
    [anon_sym_ATproperty] = ACTIONS(2606),
    [anon_sym_BOOL] = ACTIONS(2604),
    [anon_sym_IMP] = ACTIONS(2604),
    [anon_sym_SEL] = ACTIONS(2604),
    [anon_sym_Class] = ACTIONS(2604),
    [anon_sym_id] = ACTIONS(2604),
  },
  [STATE(2131)] = {
    [sym_identifier] = ACTIONS(2412),
    [aux_sym_preproc_def_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2412),
    [sym_preproc_directive] = ACTIONS(2412),
    [anon_sym_DASH] = ACTIONS(2414),
    [anon_sym_PLUS] = ACTIONS(2414),
    [anon_sym___extension__] = ACTIONS(2412),
    [anon_sym_typedef] = ACTIONS(2412),
    [anon_sym_extern] = ACTIONS(2412),
    [anon_sym___attribute__] = ACTIONS(2412),
    [anon_sym___attribute] = ACTIONS(2412),
    [anon_sym_noreturn] = ACTIONS(2412),
    [anon_sym_LBRACK] = ACTIONS(2414),
    [anon_sym___declspec] = ACTIONS(2412),
    [anon_sym___cdecl] = ACTIONS(2412),
    [anon_sym___clrcall] = ACTIONS(2412),
    [anon_sym___stdcall] = ACTIONS(2412),
    [anon_sym___fastcall] = ACTIONS(2412),
    [anon_sym___thiscall] = ACTIONS(2412),
    [anon_sym___vectorcall] = ACTIONS(2412),
    [anon_sym_signed] = ACTIONS(2412),
    [anon_sym_unsigned] = ACTIONS(2412),
    [anon_sym_long] = ACTIONS(2412),
    [anon_sym_short] = ACTIONS(2412),
    [anon_sym_static] = ACTIONS(2412),
    [anon_sym_auto] = ACTIONS(2412),
    [anon_sym_register] = ACTIONS(2412),
    [anon_sym_inline] = ACTIONS(2412),
    [anon_sym___inline] = ACTIONS(2412),
    [anon_sym___inline__] = ACTIONS(2412),
    [anon_sym___forceinline] = ACTIONS(2412),
    [anon_sym_thread_local] = ACTIONS(2412),
    [anon_sym___thread] = ACTIONS(2412),
    [anon_sym_CG_EXTERN] = ACTIONS(2412),
    [anon_sym_CG_INLINE] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2412),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2412),
    [anon_sym_IBOutlet] = ACTIONS(2412),
    [anon_sym_IBInspectable] = ACTIONS(2412),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2412),
    [anon_sym_NS_INLINE] = ACTIONS(2412),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2412),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2412),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2412),
    [anon_sym_const] = ACTIONS(2412),
    [anon_sym_constexpr] = ACTIONS(2412),
    [anon_sym_volatile] = ACTIONS(2412),
    [anon_sym_restrict] = ACTIONS(2412),
    [anon_sym___restrict__] = ACTIONS(2412),
    [anon_sym__Atomic] = ACTIONS(2412),
    [anon_sym__Noreturn] = ACTIONS(2412),
    [anon_sym__Nonnull] = ACTIONS(2412),
    [anon_sym_nullable] = ACTIONS(2412),
    [anon_sym__Complex] = ACTIONS(2412),
    [anon_sym__Nullable] = ACTIONS(2412),
    [anon_sym__Nullable_result] = ACTIONS(2412),
    [anon_sym__Null_unspecified] = ACTIONS(2412),
    [anon_sym___autoreleasing] = ACTIONS(2412),
    [anon_sym___block] = ACTIONS(2412),
    [anon_sym___bridge] = ACTIONS(2412),
    [anon_sym___bridge_retained] = ACTIONS(2412),
    [anon_sym___bridge_transfer] = ACTIONS(2412),
    [anon_sym___complex] = ACTIONS(2412),
    [anon_sym___const] = ACTIONS(2412),
    [anon_sym___imag] = ACTIONS(2412),
    [anon_sym___kindof] = ACTIONS(2412),
    [anon_sym___nonnull] = ACTIONS(2412),
    [anon_sym___nullable] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2412),
    [anon_sym___real] = ACTIONS(2412),
    [anon_sym___strong] = ACTIONS(2412),
    [anon_sym___unsafe_unretained] = ACTIONS(2412),
    [anon_sym___unused] = ACTIONS(2412),
    [anon_sym___weak] = ACTIONS(2412),
    [anon_sym_alignas] = ACTIONS(2412),
    [anon_sym__Alignas] = ACTIONS(2412),
    [sym_primitive_type] = ACTIONS(2412),
    [anon_sym_enum] = ACTIONS(2412),
    [anon_sym_struct] = ACTIONS(2412),
    [anon_sym_union] = ACTIONS(2412),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2412),
    [anon_sym___typeof] = ACTIONS(2412),
    [anon_sym_typeof] = ACTIONS(2412),
    [aux_sym_preproc_undef_token1] = ACTIONS(2412),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2412),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2412),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2412),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_API_AVAILABLE] = ACTIONS(2412),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_API_DEPRECATED] = ACTIONS(2412),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym___deprecated_msg] = ACTIONS(2412),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2412),
    [anon_sym_ATend] = ACTIONS(2414),
    [anon_sym_AToptional] = ACTIONS(2414),
    [anon_sym_ATrequired] = ACTIONS(2414),
    [anon_sym_ATsynthesize] = ACTIONS(2414),
    [anon_sym_ATdynamic] = ACTIONS(2414),
    [anon_sym_ATproperty] = ACTIONS(2414),
    [anon_sym_BOOL] = ACTIONS(2412),
    [anon_sym_IMP] = ACTIONS(2412),
    [anon_sym_SEL] = ACTIONS(2412),
    [anon_sym_Class] = ACTIONS(2412),
    [anon_sym_id] = ACTIONS(2412),
  },
  [STATE(2132)] = {
    [sym_identifier] = ACTIONS(5292),
    [aux_sym_preproc_def_token1] = ACTIONS(5292),
    [aux_sym_preproc_if_token1] = ACTIONS(5292),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5292),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5292),
    [sym_preproc_directive] = ACTIONS(5292),
    [anon_sym_LPAREN2] = ACTIONS(5294),
    [anon_sym_DASH] = ACTIONS(5294),
    [anon_sym_PLUS] = ACTIONS(5294),
    [anon_sym_LT] = ACTIONS(5294),
    [anon_sym___extension__] = ACTIONS(5292),
    [anon_sym_typedef] = ACTIONS(5292),
    [anon_sym_extern] = ACTIONS(5292),
    [anon_sym___attribute__] = ACTIONS(5292),
    [anon_sym___attribute] = ACTIONS(5292),
    [anon_sym_noreturn] = ACTIONS(5292),
    [anon_sym_LBRACK] = ACTIONS(5294),
    [anon_sym___declspec] = ACTIONS(5292),
    [anon_sym___cdecl] = ACTIONS(5292),
    [anon_sym___clrcall] = ACTIONS(5292),
    [anon_sym___stdcall] = ACTIONS(5292),
    [anon_sym___fastcall] = ACTIONS(5292),
    [anon_sym___thiscall] = ACTIONS(5292),
    [anon_sym___vectorcall] = ACTIONS(5292),
    [anon_sym_LBRACE] = ACTIONS(5294),
    [anon_sym_signed] = ACTIONS(5292),
    [anon_sym_unsigned] = ACTIONS(5292),
    [anon_sym_long] = ACTIONS(5292),
    [anon_sym_short] = ACTIONS(5292),
    [anon_sym_static] = ACTIONS(5292),
    [anon_sym_auto] = ACTIONS(5292),
    [anon_sym_register] = ACTIONS(5292),
    [anon_sym_inline] = ACTIONS(5292),
    [anon_sym___inline] = ACTIONS(5292),
    [anon_sym___inline__] = ACTIONS(5292),
    [anon_sym___forceinline] = ACTIONS(5292),
    [anon_sym_thread_local] = ACTIONS(5292),
    [anon_sym___thread] = ACTIONS(5292),
    [anon_sym_CG_EXTERN] = ACTIONS(5292),
    [anon_sym_CG_INLINE] = ACTIONS(5292),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5292),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5292),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5292),
    [anon_sym_IBOutlet] = ACTIONS(5292),
    [anon_sym_IBInspectable] = ACTIONS(5292),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5292),
    [anon_sym_NS_INLINE] = ACTIONS(5292),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5292),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5292),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5292),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5292),
    [anon_sym_const] = ACTIONS(5292),
    [anon_sym_constexpr] = ACTIONS(5292),
    [anon_sym_volatile] = ACTIONS(5292),
    [anon_sym_restrict] = ACTIONS(5292),
    [anon_sym___restrict__] = ACTIONS(5292),
    [anon_sym__Atomic] = ACTIONS(5292),
    [anon_sym__Noreturn] = ACTIONS(5292),
    [anon_sym__Nonnull] = ACTIONS(5292),
    [anon_sym_nullable] = ACTIONS(5292),
    [anon_sym__Complex] = ACTIONS(5292),
    [anon_sym__Nullable] = ACTIONS(5292),
    [anon_sym__Nullable_result] = ACTIONS(5292),
    [anon_sym__Null_unspecified] = ACTIONS(5292),
    [anon_sym___autoreleasing] = ACTIONS(5292),
    [anon_sym___block] = ACTIONS(5292),
    [anon_sym___bridge] = ACTIONS(5292),
    [anon_sym___bridge_retained] = ACTIONS(5292),
    [anon_sym___bridge_transfer] = ACTIONS(5292),
    [anon_sym___complex] = ACTIONS(5292),
    [anon_sym___const] = ACTIONS(5292),
    [anon_sym___imag] = ACTIONS(5292),
    [anon_sym___kindof] = ACTIONS(5292),
    [anon_sym___nonnull] = ACTIONS(5292),
    [anon_sym___nullable] = ACTIONS(5292),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5292),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5292),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5292),
    [anon_sym___real] = ACTIONS(5292),
    [anon_sym___strong] = ACTIONS(5292),
    [anon_sym___unsafe_unretained] = ACTIONS(5292),
    [anon_sym___unused] = ACTIONS(5292),
    [anon_sym___weak] = ACTIONS(5292),
    [anon_sym_alignas] = ACTIONS(5292),
    [anon_sym__Alignas] = ACTIONS(5292),
    [sym_primitive_type] = ACTIONS(5292),
    [anon_sym_enum] = ACTIONS(5292),
    [anon_sym_COLON] = ACTIONS(5294),
    [anon_sym_struct] = ACTIONS(5292),
    [anon_sym_union] = ACTIONS(5292),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5292),
    [anon_sym___typeof] = ACTIONS(5292),
    [anon_sym_typeof] = ACTIONS(5292),
    [aux_sym_preproc_undef_token1] = ACTIONS(5292),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5292),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5292),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5292),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5292),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5292),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5292),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5292),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5292),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5292),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5292),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5292),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5292),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5292),
    [anon_sym_API_AVAILABLE] = ACTIONS(5292),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5292),
    [anon_sym_API_DEPRECATED] = ACTIONS(5292),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5292),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5292),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5292),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5292),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5292),
    [anon_sym___deprecated_msg] = ACTIONS(5292),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5292),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5292),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5292),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5292),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5292),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5292),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5292),
    [anon_sym_ATend] = ACTIONS(5294),
    [anon_sym_ATproperty] = ACTIONS(5294),
    [anon_sym_BOOL] = ACTIONS(5292),
    [anon_sym_IMP] = ACTIONS(5292),
    [anon_sym_SEL] = ACTIONS(5292),
    [anon_sym_Class] = ACTIONS(5292),
    [anon_sym_id] = ACTIONS(5292),
  },
  [STATE(2133)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2138),
    [sym_identifier] = ACTIONS(5423),
    [aux_sym_preproc_def_token1] = ACTIONS(5423),
    [aux_sym_preproc_if_token1] = ACTIONS(5423),
    [aux_sym_preproc_if_token2] = ACTIONS(5423),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5423),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5423),
    [aux_sym_preproc_else_token1] = ACTIONS(5423),
    [aux_sym_preproc_elif_token1] = ACTIONS(5423),
    [sym_preproc_directive] = ACTIONS(5423),
    [anon_sym_DASH] = ACTIONS(5425),
    [anon_sym_PLUS] = ACTIONS(5425),
    [anon_sym_SEMI] = ACTIONS(5357),
    [anon_sym___extension__] = ACTIONS(5423),
    [anon_sym_typedef] = ACTIONS(5423),
    [anon_sym_extern] = ACTIONS(5423),
    [anon_sym___attribute__] = ACTIONS(5423),
    [anon_sym___attribute] = ACTIONS(5423),
    [anon_sym_noreturn] = ACTIONS(5423),
    [anon_sym_LBRACK] = ACTIONS(5425),
    [anon_sym___declspec] = ACTIONS(5423),
    [anon_sym___cdecl] = ACTIONS(5423),
    [anon_sym___clrcall] = ACTIONS(5423),
    [anon_sym___stdcall] = ACTIONS(5423),
    [anon_sym___fastcall] = ACTIONS(5423),
    [anon_sym___thiscall] = ACTIONS(5423),
    [anon_sym___vectorcall] = ACTIONS(5423),
    [anon_sym_signed] = ACTIONS(5423),
    [anon_sym_unsigned] = ACTIONS(5423),
    [anon_sym_long] = ACTIONS(5423),
    [anon_sym_short] = ACTIONS(5423),
    [anon_sym_static] = ACTIONS(5423),
    [anon_sym_auto] = ACTIONS(5423),
    [anon_sym_register] = ACTIONS(5423),
    [anon_sym_inline] = ACTIONS(5423),
    [anon_sym___inline] = ACTIONS(5423),
    [anon_sym___inline__] = ACTIONS(5423),
    [anon_sym___forceinline] = ACTIONS(5423),
    [anon_sym_thread_local] = ACTIONS(5423),
    [anon_sym___thread] = ACTIONS(5423),
    [anon_sym_CG_EXTERN] = ACTIONS(5423),
    [anon_sym_CG_INLINE] = ACTIONS(5423),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5423),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5423),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5423),
    [anon_sym_IBOutlet] = ACTIONS(5423),
    [anon_sym_IBInspectable] = ACTIONS(5423),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5423),
    [anon_sym_NS_INLINE] = ACTIONS(5423),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5423),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5423),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5423),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5423),
    [anon_sym_const] = ACTIONS(5423),
    [anon_sym_constexpr] = ACTIONS(5423),
    [anon_sym_volatile] = ACTIONS(5423),
    [anon_sym_restrict] = ACTIONS(5423),
    [anon_sym___restrict__] = ACTIONS(5423),
    [anon_sym__Atomic] = ACTIONS(5423),
    [anon_sym__Noreturn] = ACTIONS(5423),
    [anon_sym__Nonnull] = ACTIONS(5423),
    [anon_sym_nullable] = ACTIONS(5423),
    [anon_sym__Complex] = ACTIONS(5423),
    [anon_sym__Nullable] = ACTIONS(5423),
    [anon_sym__Nullable_result] = ACTIONS(5423),
    [anon_sym__Null_unspecified] = ACTIONS(5423),
    [anon_sym___autoreleasing] = ACTIONS(5423),
    [anon_sym___block] = ACTIONS(5423),
    [anon_sym___bridge] = ACTIONS(5423),
    [anon_sym___bridge_retained] = ACTIONS(5423),
    [anon_sym___bridge_transfer] = ACTIONS(5423),
    [anon_sym___complex] = ACTIONS(5423),
    [anon_sym___const] = ACTIONS(5423),
    [anon_sym___imag] = ACTIONS(5423),
    [anon_sym___kindof] = ACTIONS(5423),
    [anon_sym___nonnull] = ACTIONS(5423),
    [anon_sym___nullable] = ACTIONS(5423),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5423),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5423),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5423),
    [anon_sym___real] = ACTIONS(5423),
    [anon_sym___strong] = ACTIONS(5423),
    [anon_sym___unsafe_unretained] = ACTIONS(5423),
    [anon_sym___unused] = ACTIONS(5423),
    [anon_sym___weak] = ACTIONS(5423),
    [anon_sym_alignas] = ACTIONS(5423),
    [anon_sym__Alignas] = ACTIONS(5423),
    [sym_primitive_type] = ACTIONS(5423),
    [anon_sym_enum] = ACTIONS(5423),
    [anon_sym_struct] = ACTIONS(5423),
    [anon_sym_union] = ACTIONS(5423),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5423),
    [anon_sym___typeof] = ACTIONS(5423),
    [anon_sym_typeof] = ACTIONS(5423),
    [aux_sym_preproc_undef_token1] = ACTIONS(5423),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5423),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5423),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5423),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5423),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5423),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5423),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5423),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5423),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5423),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_API_AVAILABLE] = ACTIONS(5423),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_API_DEPRECATED] = ACTIONS(5423),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5423),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5423),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5423),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5423),
    [anon_sym___deprecated_msg] = ACTIONS(5423),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5423),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5423),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5423),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5423),
    [anon_sym_ATproperty] = ACTIONS(5425),
    [anon_sym_BOOL] = ACTIONS(5423),
    [anon_sym_IMP] = ACTIONS(5423),
    [anon_sym_SEL] = ACTIONS(5423),
    [anon_sym_Class] = ACTIONS(5423),
    [anon_sym_id] = ACTIONS(5423),
  },
  [STATE(2134)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token2] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [aux_sym_preproc_else_token1] = ACTIONS(2438),
    [aux_sym_preproc_elif_token1] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_DASH] = ACTIONS(2440),
    [anon_sym_PLUS] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATproperty] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(2135)] = {
    [sym__declaration_modifiers] = STATE(2539),
    [sym_attribute_specifier] = STATE(2539),
    [sym_attribute_declaration] = STATE(2539),
    [sym_ms_declspec_modifier] = STATE(2539),
    [sym_ms_based_modifier] = STATE(6497),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2553),
    [sym__declarator] = STATE(3831),
    [sym__abstract_declarator] = STATE(4988),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(2539),
    [sym_type_qualifier] = STATE(2539),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4290),
    [sym_availability_attribute_specifier] = STATE(2539),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2539),
    [aux_sym_pointer_declarator_repeat1] = STATE(2553),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(5427),
    [anon_sym_CARET] = ACTIONS(5429),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_LBRACE] = ACTIONS(5314),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(5314),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2136)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_DASH] = ACTIONS(2556),
    [anon_sym_PLUS] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATend] = ACTIONS(2556),
    [anon_sym_AToptional] = ACTIONS(2556),
    [anon_sym_ATrequired] = ACTIONS(2556),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATproperty] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(2137)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2185),
    [sym_identifier] = ACTIONS(5365),
    [aux_sym_preproc_def_token1] = ACTIONS(5365),
    [aux_sym_preproc_if_token1] = ACTIONS(5365),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5365),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5365),
    [sym_preproc_directive] = ACTIONS(5365),
    [anon_sym_DASH] = ACTIONS(5367),
    [anon_sym_PLUS] = ACTIONS(5367),
    [anon_sym_SEMI] = ACTIONS(5349),
    [anon_sym___extension__] = ACTIONS(5365),
    [anon_sym_typedef] = ACTIONS(5365),
    [anon_sym_extern] = ACTIONS(5365),
    [anon_sym___attribute__] = ACTIONS(5365),
    [anon_sym___attribute] = ACTIONS(5365),
    [anon_sym_noreturn] = ACTIONS(5365),
    [anon_sym_LBRACK] = ACTIONS(5367),
    [anon_sym___declspec] = ACTIONS(5365),
    [anon_sym___cdecl] = ACTIONS(5365),
    [anon_sym___clrcall] = ACTIONS(5365),
    [anon_sym___stdcall] = ACTIONS(5365),
    [anon_sym___fastcall] = ACTIONS(5365),
    [anon_sym___thiscall] = ACTIONS(5365),
    [anon_sym___vectorcall] = ACTIONS(5365),
    [anon_sym_signed] = ACTIONS(5365),
    [anon_sym_unsigned] = ACTIONS(5365),
    [anon_sym_long] = ACTIONS(5365),
    [anon_sym_short] = ACTIONS(5365),
    [anon_sym_static] = ACTIONS(5365),
    [anon_sym_auto] = ACTIONS(5365),
    [anon_sym_register] = ACTIONS(5365),
    [anon_sym_inline] = ACTIONS(5365),
    [anon_sym___inline] = ACTIONS(5365),
    [anon_sym___inline__] = ACTIONS(5365),
    [anon_sym___forceinline] = ACTIONS(5365),
    [anon_sym_thread_local] = ACTIONS(5365),
    [anon_sym___thread] = ACTIONS(5365),
    [anon_sym_CG_EXTERN] = ACTIONS(5365),
    [anon_sym_CG_INLINE] = ACTIONS(5365),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5365),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5365),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5365),
    [anon_sym_IBOutlet] = ACTIONS(5365),
    [anon_sym_IBInspectable] = ACTIONS(5365),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5365),
    [anon_sym_NS_INLINE] = ACTIONS(5365),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5365),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5365),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5365),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5365),
    [anon_sym_const] = ACTIONS(5365),
    [anon_sym_constexpr] = ACTIONS(5365),
    [anon_sym_volatile] = ACTIONS(5365),
    [anon_sym_restrict] = ACTIONS(5365),
    [anon_sym___restrict__] = ACTIONS(5365),
    [anon_sym__Atomic] = ACTIONS(5365),
    [anon_sym__Noreturn] = ACTIONS(5365),
    [anon_sym__Nonnull] = ACTIONS(5365),
    [anon_sym_nullable] = ACTIONS(5365),
    [anon_sym__Complex] = ACTIONS(5365),
    [anon_sym__Nullable] = ACTIONS(5365),
    [anon_sym__Nullable_result] = ACTIONS(5365),
    [anon_sym__Null_unspecified] = ACTIONS(5365),
    [anon_sym___autoreleasing] = ACTIONS(5365),
    [anon_sym___block] = ACTIONS(5365),
    [anon_sym___bridge] = ACTIONS(5365),
    [anon_sym___bridge_retained] = ACTIONS(5365),
    [anon_sym___bridge_transfer] = ACTIONS(5365),
    [anon_sym___complex] = ACTIONS(5365),
    [anon_sym___const] = ACTIONS(5365),
    [anon_sym___imag] = ACTIONS(5365),
    [anon_sym___kindof] = ACTIONS(5365),
    [anon_sym___nonnull] = ACTIONS(5365),
    [anon_sym___nullable] = ACTIONS(5365),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5365),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5365),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5365),
    [anon_sym___real] = ACTIONS(5365),
    [anon_sym___strong] = ACTIONS(5365),
    [anon_sym___unsafe_unretained] = ACTIONS(5365),
    [anon_sym___unused] = ACTIONS(5365),
    [anon_sym___weak] = ACTIONS(5365),
    [anon_sym_alignas] = ACTIONS(5365),
    [anon_sym__Alignas] = ACTIONS(5365),
    [sym_primitive_type] = ACTIONS(5365),
    [anon_sym_enum] = ACTIONS(5365),
    [anon_sym_struct] = ACTIONS(5365),
    [anon_sym_union] = ACTIONS(5365),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5365),
    [anon_sym___typeof] = ACTIONS(5365),
    [anon_sym_typeof] = ACTIONS(5365),
    [aux_sym_preproc_undef_token1] = ACTIONS(5365),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5365),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5365),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5365),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5365),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5365),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5365),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5365),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5365),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5365),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_API_AVAILABLE] = ACTIONS(5365),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_API_DEPRECATED] = ACTIONS(5365),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5365),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5365),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5365),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5365),
    [anon_sym___deprecated_msg] = ACTIONS(5365),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5365),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5365),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5365),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5365),
    [anon_sym_ATend] = ACTIONS(5367),
    [anon_sym_AToptional] = ACTIONS(5367),
    [anon_sym_ATrequired] = ACTIONS(5367),
    [anon_sym_ATproperty] = ACTIONS(5367),
    [anon_sym_BOOL] = ACTIONS(5365),
    [anon_sym_IMP] = ACTIONS(5365),
    [anon_sym_SEL] = ACTIONS(5365),
    [anon_sym_Class] = ACTIONS(5365),
    [anon_sym_id] = ACTIONS(5365),
  },
  [STATE(2138)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2138),
    [sym_identifier] = ACTIONS(5431),
    [aux_sym_preproc_def_token1] = ACTIONS(5431),
    [aux_sym_preproc_if_token1] = ACTIONS(5431),
    [aux_sym_preproc_if_token2] = ACTIONS(5431),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5431),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5431),
    [aux_sym_preproc_else_token1] = ACTIONS(5431),
    [aux_sym_preproc_elif_token1] = ACTIONS(5431),
    [sym_preproc_directive] = ACTIONS(5431),
    [anon_sym_DASH] = ACTIONS(5433),
    [anon_sym_PLUS] = ACTIONS(5433),
    [anon_sym_SEMI] = ACTIONS(5435),
    [anon_sym___extension__] = ACTIONS(5431),
    [anon_sym_typedef] = ACTIONS(5431),
    [anon_sym_extern] = ACTIONS(5431),
    [anon_sym___attribute__] = ACTIONS(5431),
    [anon_sym___attribute] = ACTIONS(5431),
    [anon_sym_noreturn] = ACTIONS(5431),
    [anon_sym_LBRACK] = ACTIONS(5433),
    [anon_sym___declspec] = ACTIONS(5431),
    [anon_sym___cdecl] = ACTIONS(5431),
    [anon_sym___clrcall] = ACTIONS(5431),
    [anon_sym___stdcall] = ACTIONS(5431),
    [anon_sym___fastcall] = ACTIONS(5431),
    [anon_sym___thiscall] = ACTIONS(5431),
    [anon_sym___vectorcall] = ACTIONS(5431),
    [anon_sym_signed] = ACTIONS(5431),
    [anon_sym_unsigned] = ACTIONS(5431),
    [anon_sym_long] = ACTIONS(5431),
    [anon_sym_short] = ACTIONS(5431),
    [anon_sym_static] = ACTIONS(5431),
    [anon_sym_auto] = ACTIONS(5431),
    [anon_sym_register] = ACTIONS(5431),
    [anon_sym_inline] = ACTIONS(5431),
    [anon_sym___inline] = ACTIONS(5431),
    [anon_sym___inline__] = ACTIONS(5431),
    [anon_sym___forceinline] = ACTIONS(5431),
    [anon_sym_thread_local] = ACTIONS(5431),
    [anon_sym___thread] = ACTIONS(5431),
    [anon_sym_CG_EXTERN] = ACTIONS(5431),
    [anon_sym_CG_INLINE] = ACTIONS(5431),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5431),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5431),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5431),
    [anon_sym_IBOutlet] = ACTIONS(5431),
    [anon_sym_IBInspectable] = ACTIONS(5431),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5431),
    [anon_sym_NS_INLINE] = ACTIONS(5431),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5431),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5431),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5431),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5431),
    [anon_sym_const] = ACTIONS(5431),
    [anon_sym_constexpr] = ACTIONS(5431),
    [anon_sym_volatile] = ACTIONS(5431),
    [anon_sym_restrict] = ACTIONS(5431),
    [anon_sym___restrict__] = ACTIONS(5431),
    [anon_sym__Atomic] = ACTIONS(5431),
    [anon_sym__Noreturn] = ACTIONS(5431),
    [anon_sym__Nonnull] = ACTIONS(5431),
    [anon_sym_nullable] = ACTIONS(5431),
    [anon_sym__Complex] = ACTIONS(5431),
    [anon_sym__Nullable] = ACTIONS(5431),
    [anon_sym__Nullable_result] = ACTIONS(5431),
    [anon_sym__Null_unspecified] = ACTIONS(5431),
    [anon_sym___autoreleasing] = ACTIONS(5431),
    [anon_sym___block] = ACTIONS(5431),
    [anon_sym___bridge] = ACTIONS(5431),
    [anon_sym___bridge_retained] = ACTIONS(5431),
    [anon_sym___bridge_transfer] = ACTIONS(5431),
    [anon_sym___complex] = ACTIONS(5431),
    [anon_sym___const] = ACTIONS(5431),
    [anon_sym___imag] = ACTIONS(5431),
    [anon_sym___kindof] = ACTIONS(5431),
    [anon_sym___nonnull] = ACTIONS(5431),
    [anon_sym___nullable] = ACTIONS(5431),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5431),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5431),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5431),
    [anon_sym___real] = ACTIONS(5431),
    [anon_sym___strong] = ACTIONS(5431),
    [anon_sym___unsafe_unretained] = ACTIONS(5431),
    [anon_sym___unused] = ACTIONS(5431),
    [anon_sym___weak] = ACTIONS(5431),
    [anon_sym_alignas] = ACTIONS(5431),
    [anon_sym__Alignas] = ACTIONS(5431),
    [sym_primitive_type] = ACTIONS(5431),
    [anon_sym_enum] = ACTIONS(5431),
    [anon_sym_struct] = ACTIONS(5431),
    [anon_sym_union] = ACTIONS(5431),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5431),
    [anon_sym___typeof] = ACTIONS(5431),
    [anon_sym_typeof] = ACTIONS(5431),
    [aux_sym_preproc_undef_token1] = ACTIONS(5431),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5431),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5431),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5431),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5431),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5431),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5431),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5431),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5431),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5431),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_API_AVAILABLE] = ACTIONS(5431),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_API_DEPRECATED] = ACTIONS(5431),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5431),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5431),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5431),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5431),
    [anon_sym___deprecated_msg] = ACTIONS(5431),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5431),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5431),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5431),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5431),
    [anon_sym_ATproperty] = ACTIONS(5433),
    [anon_sym_BOOL] = ACTIONS(5431),
    [anon_sym_IMP] = ACTIONS(5431),
    [anon_sym_SEL] = ACTIONS(5431),
    [anon_sym_Class] = ACTIONS(5431),
    [anon_sym_id] = ACTIONS(5431),
  },
  [STATE(2139)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_DASH] = ACTIONS(2556),
    [anon_sym_PLUS] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATend] = ACTIONS(2556),
    [anon_sym_AToptional] = ACTIONS(2556),
    [anon_sym_ATrequired] = ACTIONS(2556),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATproperty] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(2140)] = {
    [sym_identifier] = ACTIONS(5438),
    [aux_sym_preproc_def_token1] = ACTIONS(5438),
    [aux_sym_preproc_if_token1] = ACTIONS(5438),
    [aux_sym_preproc_if_token2] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5438),
    [aux_sym_preproc_else_token1] = ACTIONS(5438),
    [aux_sym_preproc_elif_token1] = ACTIONS(5438),
    [sym_preproc_directive] = ACTIONS(5438),
    [anon_sym_DASH] = ACTIONS(5440),
    [anon_sym_PLUS] = ACTIONS(5440),
    [anon_sym_SEMI] = ACTIONS(5442),
    [anon_sym___extension__] = ACTIONS(5438),
    [anon_sym_typedef] = ACTIONS(5438),
    [anon_sym_extern] = ACTIONS(5438),
    [anon_sym___attribute__] = ACTIONS(5438),
    [anon_sym___attribute] = ACTIONS(5438),
    [anon_sym_noreturn] = ACTIONS(5438),
    [anon_sym_LBRACK] = ACTIONS(5440),
    [anon_sym___declspec] = ACTIONS(5438),
    [anon_sym___cdecl] = ACTIONS(5438),
    [anon_sym___clrcall] = ACTIONS(5438),
    [anon_sym___stdcall] = ACTIONS(5438),
    [anon_sym___fastcall] = ACTIONS(5438),
    [anon_sym___thiscall] = ACTIONS(5438),
    [anon_sym___vectorcall] = ACTIONS(5438),
    [anon_sym_signed] = ACTIONS(5438),
    [anon_sym_unsigned] = ACTIONS(5438),
    [anon_sym_long] = ACTIONS(5438),
    [anon_sym_short] = ACTIONS(5438),
    [anon_sym_static] = ACTIONS(5438),
    [anon_sym_auto] = ACTIONS(5438),
    [anon_sym_register] = ACTIONS(5438),
    [anon_sym_inline] = ACTIONS(5438),
    [anon_sym___inline] = ACTIONS(5438),
    [anon_sym___inline__] = ACTIONS(5438),
    [anon_sym___forceinline] = ACTIONS(5438),
    [anon_sym_thread_local] = ACTIONS(5438),
    [anon_sym___thread] = ACTIONS(5438),
    [anon_sym_CG_EXTERN] = ACTIONS(5438),
    [anon_sym_CG_INLINE] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5438),
    [anon_sym_IBOutlet] = ACTIONS(5438),
    [anon_sym_IBInspectable] = ACTIONS(5438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5438),
    [anon_sym_NS_INLINE] = ACTIONS(5438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5438),
    [anon_sym_const] = ACTIONS(5438),
    [anon_sym_constexpr] = ACTIONS(5438),
    [anon_sym_volatile] = ACTIONS(5438),
    [anon_sym_restrict] = ACTIONS(5438),
    [anon_sym___restrict__] = ACTIONS(5438),
    [anon_sym__Atomic] = ACTIONS(5438),
    [anon_sym__Noreturn] = ACTIONS(5438),
    [anon_sym__Nonnull] = ACTIONS(5438),
    [anon_sym_nullable] = ACTIONS(5438),
    [anon_sym__Complex] = ACTIONS(5438),
    [anon_sym__Nullable] = ACTIONS(5438),
    [anon_sym__Nullable_result] = ACTIONS(5438),
    [anon_sym__Null_unspecified] = ACTIONS(5438),
    [anon_sym___autoreleasing] = ACTIONS(5438),
    [anon_sym___block] = ACTIONS(5438),
    [anon_sym___bridge] = ACTIONS(5438),
    [anon_sym___bridge_retained] = ACTIONS(5438),
    [anon_sym___bridge_transfer] = ACTIONS(5438),
    [anon_sym___complex] = ACTIONS(5438),
    [anon_sym___const] = ACTIONS(5438),
    [anon_sym___imag] = ACTIONS(5438),
    [anon_sym___kindof] = ACTIONS(5438),
    [anon_sym___nonnull] = ACTIONS(5438),
    [anon_sym___nullable] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5438),
    [anon_sym___real] = ACTIONS(5438),
    [anon_sym___strong] = ACTIONS(5438),
    [anon_sym___unsafe_unretained] = ACTIONS(5438),
    [anon_sym___unused] = ACTIONS(5438),
    [anon_sym___weak] = ACTIONS(5438),
    [anon_sym_alignas] = ACTIONS(5438),
    [anon_sym__Alignas] = ACTIONS(5438),
    [sym_primitive_type] = ACTIONS(5438),
    [anon_sym_enum] = ACTIONS(5438),
    [anon_sym_struct] = ACTIONS(5438),
    [anon_sym_union] = ACTIONS(5438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5438),
    [anon_sym___typeof] = ACTIONS(5438),
    [anon_sym_typeof] = ACTIONS(5438),
    [aux_sym_preproc_undef_token1] = ACTIONS(5438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_API_AVAILABLE] = ACTIONS(5438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_API_DEPRECATED] = ACTIONS(5438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym___deprecated_msg] = ACTIONS(5438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5438),
    [anon_sym_ATsynthesize] = ACTIONS(5440),
    [anon_sym_ATdynamic] = ACTIONS(5440),
    [anon_sym_BOOL] = ACTIONS(5438),
    [anon_sym_IMP] = ACTIONS(5438),
    [anon_sym_SEL] = ACTIONS(5438),
    [anon_sym_Class] = ACTIONS(5438),
    [anon_sym_id] = ACTIONS(5438),
  },
  [STATE(2141)] = {
    [sym_identifier] = ACTIONS(2596),
    [aux_sym_preproc_def_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2596),
    [sym_preproc_directive] = ACTIONS(2596),
    [anon_sym_DASH] = ACTIONS(2598),
    [anon_sym_PLUS] = ACTIONS(2598),
    [anon_sym___extension__] = ACTIONS(2596),
    [anon_sym_typedef] = ACTIONS(2596),
    [anon_sym_extern] = ACTIONS(2596),
    [anon_sym___attribute__] = ACTIONS(2596),
    [anon_sym___attribute] = ACTIONS(2596),
    [anon_sym_noreturn] = ACTIONS(2596),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym___declspec] = ACTIONS(2596),
    [anon_sym___cdecl] = ACTIONS(2596),
    [anon_sym___clrcall] = ACTIONS(2596),
    [anon_sym___stdcall] = ACTIONS(2596),
    [anon_sym___fastcall] = ACTIONS(2596),
    [anon_sym___thiscall] = ACTIONS(2596),
    [anon_sym___vectorcall] = ACTIONS(2596),
    [anon_sym_signed] = ACTIONS(2596),
    [anon_sym_unsigned] = ACTIONS(2596),
    [anon_sym_long] = ACTIONS(2596),
    [anon_sym_short] = ACTIONS(2596),
    [anon_sym_static] = ACTIONS(2596),
    [anon_sym_auto] = ACTIONS(2596),
    [anon_sym_register] = ACTIONS(2596),
    [anon_sym_inline] = ACTIONS(2596),
    [anon_sym___inline] = ACTIONS(2596),
    [anon_sym___inline__] = ACTIONS(2596),
    [anon_sym___forceinline] = ACTIONS(2596),
    [anon_sym_thread_local] = ACTIONS(2596),
    [anon_sym___thread] = ACTIONS(2596),
    [anon_sym_CG_EXTERN] = ACTIONS(2596),
    [anon_sym_CG_INLINE] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2596),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2596),
    [anon_sym_IBOutlet] = ACTIONS(2596),
    [anon_sym_IBInspectable] = ACTIONS(2596),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2596),
    [anon_sym_NS_INLINE] = ACTIONS(2596),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2596),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2596),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2596),
    [anon_sym_const] = ACTIONS(2596),
    [anon_sym_constexpr] = ACTIONS(2596),
    [anon_sym_volatile] = ACTIONS(2596),
    [anon_sym_restrict] = ACTIONS(2596),
    [anon_sym___restrict__] = ACTIONS(2596),
    [anon_sym__Atomic] = ACTIONS(2596),
    [anon_sym__Noreturn] = ACTIONS(2596),
    [anon_sym__Nonnull] = ACTIONS(2596),
    [anon_sym_nullable] = ACTIONS(2596),
    [anon_sym__Complex] = ACTIONS(2596),
    [anon_sym__Nullable] = ACTIONS(2596),
    [anon_sym__Nullable_result] = ACTIONS(2596),
    [anon_sym__Null_unspecified] = ACTIONS(2596),
    [anon_sym___autoreleasing] = ACTIONS(2596),
    [anon_sym___block] = ACTIONS(2596),
    [anon_sym___bridge] = ACTIONS(2596),
    [anon_sym___bridge_retained] = ACTIONS(2596),
    [anon_sym___bridge_transfer] = ACTIONS(2596),
    [anon_sym___complex] = ACTIONS(2596),
    [anon_sym___const] = ACTIONS(2596),
    [anon_sym___imag] = ACTIONS(2596),
    [anon_sym___kindof] = ACTIONS(2596),
    [anon_sym___nonnull] = ACTIONS(2596),
    [anon_sym___nullable] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2596),
    [anon_sym___real] = ACTIONS(2596),
    [anon_sym___strong] = ACTIONS(2596),
    [anon_sym___unsafe_unretained] = ACTIONS(2596),
    [anon_sym___unused] = ACTIONS(2596),
    [anon_sym___weak] = ACTIONS(2596),
    [anon_sym_alignas] = ACTIONS(2596),
    [anon_sym__Alignas] = ACTIONS(2596),
    [sym_primitive_type] = ACTIONS(2596),
    [anon_sym_enum] = ACTIONS(2596),
    [anon_sym_struct] = ACTIONS(2596),
    [anon_sym_union] = ACTIONS(2596),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2596),
    [anon_sym___typeof] = ACTIONS(2596),
    [anon_sym_typeof] = ACTIONS(2596),
    [aux_sym_preproc_undef_token1] = ACTIONS(2596),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2596),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2596),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2596),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_API_AVAILABLE] = ACTIONS(2596),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_API_DEPRECATED] = ACTIONS(2596),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym___deprecated_msg] = ACTIONS(2596),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2596),
    [anon_sym_ATend] = ACTIONS(2598),
    [anon_sym_AToptional] = ACTIONS(2598),
    [anon_sym_ATrequired] = ACTIONS(2598),
    [anon_sym_ATsynthesize] = ACTIONS(2598),
    [anon_sym_ATdynamic] = ACTIONS(2598),
    [anon_sym_ATproperty] = ACTIONS(2598),
    [anon_sym_BOOL] = ACTIONS(2596),
    [anon_sym_IMP] = ACTIONS(2596),
    [anon_sym_SEL] = ACTIONS(2596),
    [anon_sym_Class] = ACTIONS(2596),
    [anon_sym_id] = ACTIONS(2596),
  },
  [STATE(2142)] = {
    [sym_identifier] = ACTIONS(2596),
    [aux_sym_preproc_def_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token2] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2596),
    [aux_sym_preproc_else_token1] = ACTIONS(2596),
    [aux_sym_preproc_elif_token1] = ACTIONS(2596),
    [sym_preproc_directive] = ACTIONS(2596),
    [anon_sym_DASH] = ACTIONS(2598),
    [anon_sym_PLUS] = ACTIONS(2598),
    [anon_sym___extension__] = ACTIONS(2596),
    [anon_sym_typedef] = ACTIONS(2596),
    [anon_sym_extern] = ACTIONS(2596),
    [anon_sym___attribute__] = ACTIONS(2596),
    [anon_sym___attribute] = ACTIONS(2596),
    [anon_sym_noreturn] = ACTIONS(2596),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym___declspec] = ACTIONS(2596),
    [anon_sym___cdecl] = ACTIONS(2596),
    [anon_sym___clrcall] = ACTIONS(2596),
    [anon_sym___stdcall] = ACTIONS(2596),
    [anon_sym___fastcall] = ACTIONS(2596),
    [anon_sym___thiscall] = ACTIONS(2596),
    [anon_sym___vectorcall] = ACTIONS(2596),
    [anon_sym_signed] = ACTIONS(2596),
    [anon_sym_unsigned] = ACTIONS(2596),
    [anon_sym_long] = ACTIONS(2596),
    [anon_sym_short] = ACTIONS(2596),
    [anon_sym_static] = ACTIONS(2596),
    [anon_sym_auto] = ACTIONS(2596),
    [anon_sym_register] = ACTIONS(2596),
    [anon_sym_inline] = ACTIONS(2596),
    [anon_sym___inline] = ACTIONS(2596),
    [anon_sym___inline__] = ACTIONS(2596),
    [anon_sym___forceinline] = ACTIONS(2596),
    [anon_sym_thread_local] = ACTIONS(2596),
    [anon_sym___thread] = ACTIONS(2596),
    [anon_sym_CG_EXTERN] = ACTIONS(2596),
    [anon_sym_CG_INLINE] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2596),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2596),
    [anon_sym_IBOutlet] = ACTIONS(2596),
    [anon_sym_IBInspectable] = ACTIONS(2596),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2596),
    [anon_sym_NS_INLINE] = ACTIONS(2596),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2596),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2596),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2596),
    [anon_sym_const] = ACTIONS(2596),
    [anon_sym_constexpr] = ACTIONS(2596),
    [anon_sym_volatile] = ACTIONS(2596),
    [anon_sym_restrict] = ACTIONS(2596),
    [anon_sym___restrict__] = ACTIONS(2596),
    [anon_sym__Atomic] = ACTIONS(2596),
    [anon_sym__Noreturn] = ACTIONS(2596),
    [anon_sym__Nonnull] = ACTIONS(2596),
    [anon_sym_nullable] = ACTIONS(2596),
    [anon_sym__Complex] = ACTIONS(2596),
    [anon_sym__Nullable] = ACTIONS(2596),
    [anon_sym__Nullable_result] = ACTIONS(2596),
    [anon_sym__Null_unspecified] = ACTIONS(2596),
    [anon_sym___autoreleasing] = ACTIONS(2596),
    [anon_sym___block] = ACTIONS(2596),
    [anon_sym___bridge] = ACTIONS(2596),
    [anon_sym___bridge_retained] = ACTIONS(2596),
    [anon_sym___bridge_transfer] = ACTIONS(2596),
    [anon_sym___complex] = ACTIONS(2596),
    [anon_sym___const] = ACTIONS(2596),
    [anon_sym___imag] = ACTIONS(2596),
    [anon_sym___kindof] = ACTIONS(2596),
    [anon_sym___nonnull] = ACTIONS(2596),
    [anon_sym___nullable] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2596),
    [anon_sym___real] = ACTIONS(2596),
    [anon_sym___strong] = ACTIONS(2596),
    [anon_sym___unsafe_unretained] = ACTIONS(2596),
    [anon_sym___unused] = ACTIONS(2596),
    [anon_sym___weak] = ACTIONS(2596),
    [anon_sym_alignas] = ACTIONS(2596),
    [anon_sym__Alignas] = ACTIONS(2596),
    [sym_primitive_type] = ACTIONS(2596),
    [anon_sym_enum] = ACTIONS(2596),
    [anon_sym_struct] = ACTIONS(2596),
    [anon_sym_union] = ACTIONS(2596),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2596),
    [anon_sym___typeof] = ACTIONS(2596),
    [anon_sym_typeof] = ACTIONS(2596),
    [aux_sym_preproc_undef_token1] = ACTIONS(2596),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2596),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2596),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2596),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_API_AVAILABLE] = ACTIONS(2596),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_API_DEPRECATED] = ACTIONS(2596),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym___deprecated_msg] = ACTIONS(2596),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2596),
    [anon_sym_ATsynthesize] = ACTIONS(2598),
    [anon_sym_ATdynamic] = ACTIONS(2598),
    [anon_sym_ATproperty] = ACTIONS(2598),
    [anon_sym_BOOL] = ACTIONS(2596),
    [anon_sym_IMP] = ACTIONS(2596),
    [anon_sym_SEL] = ACTIONS(2596),
    [anon_sym_Class] = ACTIONS(2596),
    [anon_sym_id] = ACTIONS(2596),
  },
  [STATE(2143)] = {
    [sym_identifier] = ACTIONS(2604),
    [aux_sym_preproc_def_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2604),
    [sym_preproc_directive] = ACTIONS(2604),
    [anon_sym_DASH] = ACTIONS(2606),
    [anon_sym_PLUS] = ACTIONS(2606),
    [anon_sym___extension__] = ACTIONS(2604),
    [anon_sym_typedef] = ACTIONS(2604),
    [anon_sym_extern] = ACTIONS(2604),
    [anon_sym___attribute__] = ACTIONS(2604),
    [anon_sym___attribute] = ACTIONS(2604),
    [anon_sym_noreturn] = ACTIONS(2604),
    [anon_sym_LBRACK] = ACTIONS(2606),
    [anon_sym___declspec] = ACTIONS(2604),
    [anon_sym___cdecl] = ACTIONS(2604),
    [anon_sym___clrcall] = ACTIONS(2604),
    [anon_sym___stdcall] = ACTIONS(2604),
    [anon_sym___fastcall] = ACTIONS(2604),
    [anon_sym___thiscall] = ACTIONS(2604),
    [anon_sym___vectorcall] = ACTIONS(2604),
    [anon_sym_signed] = ACTIONS(2604),
    [anon_sym_unsigned] = ACTIONS(2604),
    [anon_sym_long] = ACTIONS(2604),
    [anon_sym_short] = ACTIONS(2604),
    [anon_sym_static] = ACTIONS(2604),
    [anon_sym_auto] = ACTIONS(2604),
    [anon_sym_register] = ACTIONS(2604),
    [anon_sym_inline] = ACTIONS(2604),
    [anon_sym___inline] = ACTIONS(2604),
    [anon_sym___inline__] = ACTIONS(2604),
    [anon_sym___forceinline] = ACTIONS(2604),
    [anon_sym_thread_local] = ACTIONS(2604),
    [anon_sym___thread] = ACTIONS(2604),
    [anon_sym_CG_EXTERN] = ACTIONS(2604),
    [anon_sym_CG_INLINE] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2604),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2604),
    [anon_sym_IBOutlet] = ACTIONS(2604),
    [anon_sym_IBInspectable] = ACTIONS(2604),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2604),
    [anon_sym_NS_INLINE] = ACTIONS(2604),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2604),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2604),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2604),
    [anon_sym_const] = ACTIONS(2604),
    [anon_sym_constexpr] = ACTIONS(2604),
    [anon_sym_volatile] = ACTIONS(2604),
    [anon_sym_restrict] = ACTIONS(2604),
    [anon_sym___restrict__] = ACTIONS(2604),
    [anon_sym__Atomic] = ACTIONS(2604),
    [anon_sym__Noreturn] = ACTIONS(2604),
    [anon_sym__Nonnull] = ACTIONS(2604),
    [anon_sym_nullable] = ACTIONS(2604),
    [anon_sym__Complex] = ACTIONS(2604),
    [anon_sym__Nullable] = ACTIONS(2604),
    [anon_sym__Nullable_result] = ACTIONS(2604),
    [anon_sym__Null_unspecified] = ACTIONS(2604),
    [anon_sym___autoreleasing] = ACTIONS(2604),
    [anon_sym___block] = ACTIONS(2604),
    [anon_sym___bridge] = ACTIONS(2604),
    [anon_sym___bridge_retained] = ACTIONS(2604),
    [anon_sym___bridge_transfer] = ACTIONS(2604),
    [anon_sym___complex] = ACTIONS(2604),
    [anon_sym___const] = ACTIONS(2604),
    [anon_sym___imag] = ACTIONS(2604),
    [anon_sym___kindof] = ACTIONS(2604),
    [anon_sym___nonnull] = ACTIONS(2604),
    [anon_sym___nullable] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2604),
    [anon_sym___real] = ACTIONS(2604),
    [anon_sym___strong] = ACTIONS(2604),
    [anon_sym___unsafe_unretained] = ACTIONS(2604),
    [anon_sym___unused] = ACTIONS(2604),
    [anon_sym___weak] = ACTIONS(2604),
    [anon_sym_alignas] = ACTIONS(2604),
    [anon_sym__Alignas] = ACTIONS(2604),
    [sym_primitive_type] = ACTIONS(2604),
    [anon_sym_enum] = ACTIONS(2604),
    [anon_sym_struct] = ACTIONS(2604),
    [anon_sym_union] = ACTIONS(2604),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2604),
    [anon_sym___typeof] = ACTIONS(2604),
    [anon_sym_typeof] = ACTIONS(2604),
    [aux_sym_preproc_undef_token1] = ACTIONS(2604),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2604),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2604),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2604),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_API_AVAILABLE] = ACTIONS(2604),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_API_DEPRECATED] = ACTIONS(2604),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym___deprecated_msg] = ACTIONS(2604),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2604),
    [anon_sym_ATend] = ACTIONS(2606),
    [anon_sym_AToptional] = ACTIONS(2606),
    [anon_sym_ATrequired] = ACTIONS(2606),
    [anon_sym_ATsynthesize] = ACTIONS(2606),
    [anon_sym_ATdynamic] = ACTIONS(2606),
    [anon_sym_ATproperty] = ACTIONS(2606),
    [anon_sym_BOOL] = ACTIONS(2604),
    [anon_sym_IMP] = ACTIONS(2604),
    [anon_sym_SEL] = ACTIONS(2604),
    [anon_sym_Class] = ACTIONS(2604),
    [anon_sym_id] = ACTIONS(2604),
  },
  [STATE(2144)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_DASH] = ACTIONS(2430),
    [anon_sym_PLUS] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATend] = ACTIONS(2430),
    [anon_sym_AToptional] = ACTIONS(2430),
    [anon_sym_ATrequired] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATproperty] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(2145)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token2] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [aux_sym_preproc_else_token1] = ACTIONS(2442),
    [aux_sym_preproc_elif_token1] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_DASH] = ACTIONS(2444),
    [anon_sym_PLUS] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATproperty] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(2146)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_DASH] = ACTIONS(2430),
    [anon_sym_PLUS] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATend] = ACTIONS(2430),
    [anon_sym_AToptional] = ACTIONS(2430),
    [anon_sym_ATrequired] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATproperty] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(2147)] = {
    [sym_identifier] = ACTIONS(2412),
    [aux_sym_preproc_def_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token2] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2412),
    [aux_sym_preproc_else_token1] = ACTIONS(2412),
    [aux_sym_preproc_elif_token1] = ACTIONS(2412),
    [sym_preproc_directive] = ACTIONS(2412),
    [anon_sym_DASH] = ACTIONS(2414),
    [anon_sym_PLUS] = ACTIONS(2414),
    [anon_sym___extension__] = ACTIONS(2412),
    [anon_sym_typedef] = ACTIONS(2412),
    [anon_sym_extern] = ACTIONS(2412),
    [anon_sym___attribute__] = ACTIONS(2412),
    [anon_sym___attribute] = ACTIONS(2412),
    [anon_sym_noreturn] = ACTIONS(2412),
    [anon_sym_LBRACK] = ACTIONS(2414),
    [anon_sym___declspec] = ACTIONS(2412),
    [anon_sym___cdecl] = ACTIONS(2412),
    [anon_sym___clrcall] = ACTIONS(2412),
    [anon_sym___stdcall] = ACTIONS(2412),
    [anon_sym___fastcall] = ACTIONS(2412),
    [anon_sym___thiscall] = ACTIONS(2412),
    [anon_sym___vectorcall] = ACTIONS(2412),
    [anon_sym_signed] = ACTIONS(2412),
    [anon_sym_unsigned] = ACTIONS(2412),
    [anon_sym_long] = ACTIONS(2412),
    [anon_sym_short] = ACTIONS(2412),
    [anon_sym_static] = ACTIONS(2412),
    [anon_sym_auto] = ACTIONS(2412),
    [anon_sym_register] = ACTIONS(2412),
    [anon_sym_inline] = ACTIONS(2412),
    [anon_sym___inline] = ACTIONS(2412),
    [anon_sym___inline__] = ACTIONS(2412),
    [anon_sym___forceinline] = ACTIONS(2412),
    [anon_sym_thread_local] = ACTIONS(2412),
    [anon_sym___thread] = ACTIONS(2412),
    [anon_sym_CG_EXTERN] = ACTIONS(2412),
    [anon_sym_CG_INLINE] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2412),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2412),
    [anon_sym_IBOutlet] = ACTIONS(2412),
    [anon_sym_IBInspectable] = ACTIONS(2412),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2412),
    [anon_sym_NS_INLINE] = ACTIONS(2412),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2412),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2412),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2412),
    [anon_sym_const] = ACTIONS(2412),
    [anon_sym_constexpr] = ACTIONS(2412),
    [anon_sym_volatile] = ACTIONS(2412),
    [anon_sym_restrict] = ACTIONS(2412),
    [anon_sym___restrict__] = ACTIONS(2412),
    [anon_sym__Atomic] = ACTIONS(2412),
    [anon_sym__Noreturn] = ACTIONS(2412),
    [anon_sym__Nonnull] = ACTIONS(2412),
    [anon_sym_nullable] = ACTIONS(2412),
    [anon_sym__Complex] = ACTIONS(2412),
    [anon_sym__Nullable] = ACTIONS(2412),
    [anon_sym__Nullable_result] = ACTIONS(2412),
    [anon_sym__Null_unspecified] = ACTIONS(2412),
    [anon_sym___autoreleasing] = ACTIONS(2412),
    [anon_sym___block] = ACTIONS(2412),
    [anon_sym___bridge] = ACTIONS(2412),
    [anon_sym___bridge_retained] = ACTIONS(2412),
    [anon_sym___bridge_transfer] = ACTIONS(2412),
    [anon_sym___complex] = ACTIONS(2412),
    [anon_sym___const] = ACTIONS(2412),
    [anon_sym___imag] = ACTIONS(2412),
    [anon_sym___kindof] = ACTIONS(2412),
    [anon_sym___nonnull] = ACTIONS(2412),
    [anon_sym___nullable] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2412),
    [anon_sym___real] = ACTIONS(2412),
    [anon_sym___strong] = ACTIONS(2412),
    [anon_sym___unsafe_unretained] = ACTIONS(2412),
    [anon_sym___unused] = ACTIONS(2412),
    [anon_sym___weak] = ACTIONS(2412),
    [anon_sym_alignas] = ACTIONS(2412),
    [anon_sym__Alignas] = ACTIONS(2412),
    [sym_primitive_type] = ACTIONS(2412),
    [anon_sym_enum] = ACTIONS(2412),
    [anon_sym_struct] = ACTIONS(2412),
    [anon_sym_union] = ACTIONS(2412),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2412),
    [anon_sym___typeof] = ACTIONS(2412),
    [anon_sym_typeof] = ACTIONS(2412),
    [aux_sym_preproc_undef_token1] = ACTIONS(2412),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2412),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2412),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2412),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_API_AVAILABLE] = ACTIONS(2412),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_API_DEPRECATED] = ACTIONS(2412),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym___deprecated_msg] = ACTIONS(2412),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2412),
    [anon_sym_ATsynthesize] = ACTIONS(2414),
    [anon_sym_ATdynamic] = ACTIONS(2414),
    [anon_sym_ATproperty] = ACTIONS(2414),
    [anon_sym_BOOL] = ACTIONS(2412),
    [anon_sym_IMP] = ACTIONS(2412),
    [anon_sym_SEL] = ACTIONS(2412),
    [anon_sym_Class] = ACTIONS(2412),
    [anon_sym_id] = ACTIONS(2412),
  },
  [STATE(2148)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [aux_sym_preproc_else_token1] = ACTIONS(2428),
    [aux_sym_preproc_elif_token1] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_DASH] = ACTIONS(2430),
    [anon_sym_PLUS] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATproperty] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(2149)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [aux_sym_preproc_else_token1] = ACTIONS(2428),
    [aux_sym_preproc_elif_token1] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_DASH] = ACTIONS(2430),
    [anon_sym_PLUS] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATproperty] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(2150)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [aux_sym_preproc_else_token1] = ACTIONS(2428),
    [aux_sym_preproc_elif_token1] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_DASH] = ACTIONS(2430),
    [anon_sym_PLUS] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATproperty] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(2151)] = {
    [sym_identifier] = ACTIONS(2558),
    [aux_sym_preproc_def_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2558),
    [sym_preproc_directive] = ACTIONS(2558),
    [anon_sym_DASH] = ACTIONS(2560),
    [anon_sym_PLUS] = ACTIONS(2560),
    [anon_sym___extension__] = ACTIONS(2558),
    [anon_sym_typedef] = ACTIONS(2558),
    [anon_sym_extern] = ACTIONS(2558),
    [anon_sym___attribute__] = ACTIONS(2558),
    [anon_sym___attribute] = ACTIONS(2558),
    [anon_sym_noreturn] = ACTIONS(2558),
    [anon_sym_LBRACK] = ACTIONS(2560),
    [anon_sym___declspec] = ACTIONS(2558),
    [anon_sym___cdecl] = ACTIONS(2558),
    [anon_sym___clrcall] = ACTIONS(2558),
    [anon_sym___stdcall] = ACTIONS(2558),
    [anon_sym___fastcall] = ACTIONS(2558),
    [anon_sym___thiscall] = ACTIONS(2558),
    [anon_sym___vectorcall] = ACTIONS(2558),
    [anon_sym_signed] = ACTIONS(2558),
    [anon_sym_unsigned] = ACTIONS(2558),
    [anon_sym_long] = ACTIONS(2558),
    [anon_sym_short] = ACTIONS(2558),
    [anon_sym_static] = ACTIONS(2558),
    [anon_sym_auto] = ACTIONS(2558),
    [anon_sym_register] = ACTIONS(2558),
    [anon_sym_inline] = ACTIONS(2558),
    [anon_sym___inline] = ACTIONS(2558),
    [anon_sym___inline__] = ACTIONS(2558),
    [anon_sym___forceinline] = ACTIONS(2558),
    [anon_sym_thread_local] = ACTIONS(2558),
    [anon_sym___thread] = ACTIONS(2558),
    [anon_sym_CG_EXTERN] = ACTIONS(2558),
    [anon_sym_CG_INLINE] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2558),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2558),
    [anon_sym_IBOutlet] = ACTIONS(2558),
    [anon_sym_IBInspectable] = ACTIONS(2558),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2558),
    [anon_sym_NS_INLINE] = ACTIONS(2558),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2558),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2558),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2558),
    [anon_sym_const] = ACTIONS(2558),
    [anon_sym_constexpr] = ACTIONS(2558),
    [anon_sym_volatile] = ACTIONS(2558),
    [anon_sym_restrict] = ACTIONS(2558),
    [anon_sym___restrict__] = ACTIONS(2558),
    [anon_sym__Atomic] = ACTIONS(2558),
    [anon_sym__Noreturn] = ACTIONS(2558),
    [anon_sym__Nonnull] = ACTIONS(2558),
    [anon_sym_nullable] = ACTIONS(2558),
    [anon_sym__Complex] = ACTIONS(2558),
    [anon_sym__Nullable] = ACTIONS(2558),
    [anon_sym__Nullable_result] = ACTIONS(2558),
    [anon_sym__Null_unspecified] = ACTIONS(2558),
    [anon_sym___autoreleasing] = ACTIONS(2558),
    [anon_sym___block] = ACTIONS(2558),
    [anon_sym___bridge] = ACTIONS(2558),
    [anon_sym___bridge_retained] = ACTIONS(2558),
    [anon_sym___bridge_transfer] = ACTIONS(2558),
    [anon_sym___complex] = ACTIONS(2558),
    [anon_sym___const] = ACTIONS(2558),
    [anon_sym___imag] = ACTIONS(2558),
    [anon_sym___kindof] = ACTIONS(2558),
    [anon_sym___nonnull] = ACTIONS(2558),
    [anon_sym___nullable] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2558),
    [anon_sym___real] = ACTIONS(2558),
    [anon_sym___strong] = ACTIONS(2558),
    [anon_sym___unsafe_unretained] = ACTIONS(2558),
    [anon_sym___unused] = ACTIONS(2558),
    [anon_sym___weak] = ACTIONS(2558),
    [anon_sym_alignas] = ACTIONS(2558),
    [anon_sym__Alignas] = ACTIONS(2558),
    [sym_primitive_type] = ACTIONS(2558),
    [anon_sym_enum] = ACTIONS(2558),
    [anon_sym_struct] = ACTIONS(2558),
    [anon_sym_union] = ACTIONS(2558),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2558),
    [anon_sym___typeof] = ACTIONS(2558),
    [anon_sym_typeof] = ACTIONS(2558),
    [aux_sym_preproc_undef_token1] = ACTIONS(2558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2558),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2558),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2558),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_API_AVAILABLE] = ACTIONS(2558),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_API_DEPRECATED] = ACTIONS(2558),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym___deprecated_msg] = ACTIONS(2558),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2558),
    [anon_sym_ATend] = ACTIONS(2560),
    [anon_sym_AToptional] = ACTIONS(2560),
    [anon_sym_ATrequired] = ACTIONS(2560),
    [anon_sym_ATsynthesize] = ACTIONS(2560),
    [anon_sym_ATdynamic] = ACTIONS(2560),
    [anon_sym_ATproperty] = ACTIONS(2560),
    [anon_sym_BOOL] = ACTIONS(2558),
    [anon_sym_IMP] = ACTIONS(2558),
    [anon_sym_SEL] = ACTIONS(2558),
    [anon_sym_Class] = ACTIONS(2558),
    [anon_sym_id] = ACTIONS(2558),
  },
  [STATE(2152)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [aux_sym_preproc_else_token1] = ACTIONS(2510),
    [aux_sym_preproc_elif_token1] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_DASH] = ACTIONS(2512),
    [anon_sym_PLUS] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATproperty] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(2153)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [aux_sym_preproc_else_token1] = ACTIONS(2510),
    [aux_sym_preproc_elif_token1] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_DASH] = ACTIONS(2512),
    [anon_sym_PLUS] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATproperty] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(2154)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [aux_sym_preproc_else_token1] = ACTIONS(2510),
    [aux_sym_preproc_elif_token1] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_DASH] = ACTIONS(2512),
    [anon_sym_PLUS] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATproperty] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(2155)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token2] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [aux_sym_preproc_else_token1] = ACTIONS(2554),
    [aux_sym_preproc_elif_token1] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_DASH] = ACTIONS(2556),
    [anon_sym_PLUS] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATproperty] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(2156)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token2] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [aux_sym_preproc_else_token1] = ACTIONS(2554),
    [aux_sym_preproc_elif_token1] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_DASH] = ACTIONS(2556),
    [anon_sym_PLUS] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATproperty] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(2157)] = {
    [sym_identifier] = ACTIONS(2558),
    [aux_sym_preproc_def_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token2] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2558),
    [aux_sym_preproc_else_token1] = ACTIONS(2558),
    [aux_sym_preproc_elif_token1] = ACTIONS(2558),
    [sym_preproc_directive] = ACTIONS(2558),
    [anon_sym_DASH] = ACTIONS(2560),
    [anon_sym_PLUS] = ACTIONS(2560),
    [anon_sym___extension__] = ACTIONS(2558),
    [anon_sym_typedef] = ACTIONS(2558),
    [anon_sym_extern] = ACTIONS(2558),
    [anon_sym___attribute__] = ACTIONS(2558),
    [anon_sym___attribute] = ACTIONS(2558),
    [anon_sym_noreturn] = ACTIONS(2558),
    [anon_sym_LBRACK] = ACTIONS(2560),
    [anon_sym___declspec] = ACTIONS(2558),
    [anon_sym___cdecl] = ACTIONS(2558),
    [anon_sym___clrcall] = ACTIONS(2558),
    [anon_sym___stdcall] = ACTIONS(2558),
    [anon_sym___fastcall] = ACTIONS(2558),
    [anon_sym___thiscall] = ACTIONS(2558),
    [anon_sym___vectorcall] = ACTIONS(2558),
    [anon_sym_signed] = ACTIONS(2558),
    [anon_sym_unsigned] = ACTIONS(2558),
    [anon_sym_long] = ACTIONS(2558),
    [anon_sym_short] = ACTIONS(2558),
    [anon_sym_static] = ACTIONS(2558),
    [anon_sym_auto] = ACTIONS(2558),
    [anon_sym_register] = ACTIONS(2558),
    [anon_sym_inline] = ACTIONS(2558),
    [anon_sym___inline] = ACTIONS(2558),
    [anon_sym___inline__] = ACTIONS(2558),
    [anon_sym___forceinline] = ACTIONS(2558),
    [anon_sym_thread_local] = ACTIONS(2558),
    [anon_sym___thread] = ACTIONS(2558),
    [anon_sym_CG_EXTERN] = ACTIONS(2558),
    [anon_sym_CG_INLINE] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2558),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2558),
    [anon_sym_IBOutlet] = ACTIONS(2558),
    [anon_sym_IBInspectable] = ACTIONS(2558),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2558),
    [anon_sym_NS_INLINE] = ACTIONS(2558),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2558),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2558),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2558),
    [anon_sym_const] = ACTIONS(2558),
    [anon_sym_constexpr] = ACTIONS(2558),
    [anon_sym_volatile] = ACTIONS(2558),
    [anon_sym_restrict] = ACTIONS(2558),
    [anon_sym___restrict__] = ACTIONS(2558),
    [anon_sym__Atomic] = ACTIONS(2558),
    [anon_sym__Noreturn] = ACTIONS(2558),
    [anon_sym__Nonnull] = ACTIONS(2558),
    [anon_sym_nullable] = ACTIONS(2558),
    [anon_sym__Complex] = ACTIONS(2558),
    [anon_sym__Nullable] = ACTIONS(2558),
    [anon_sym__Nullable_result] = ACTIONS(2558),
    [anon_sym__Null_unspecified] = ACTIONS(2558),
    [anon_sym___autoreleasing] = ACTIONS(2558),
    [anon_sym___block] = ACTIONS(2558),
    [anon_sym___bridge] = ACTIONS(2558),
    [anon_sym___bridge_retained] = ACTIONS(2558),
    [anon_sym___bridge_transfer] = ACTIONS(2558),
    [anon_sym___complex] = ACTIONS(2558),
    [anon_sym___const] = ACTIONS(2558),
    [anon_sym___imag] = ACTIONS(2558),
    [anon_sym___kindof] = ACTIONS(2558),
    [anon_sym___nonnull] = ACTIONS(2558),
    [anon_sym___nullable] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2558),
    [anon_sym___real] = ACTIONS(2558),
    [anon_sym___strong] = ACTIONS(2558),
    [anon_sym___unsafe_unretained] = ACTIONS(2558),
    [anon_sym___unused] = ACTIONS(2558),
    [anon_sym___weak] = ACTIONS(2558),
    [anon_sym_alignas] = ACTIONS(2558),
    [anon_sym__Alignas] = ACTIONS(2558),
    [sym_primitive_type] = ACTIONS(2558),
    [anon_sym_enum] = ACTIONS(2558),
    [anon_sym_struct] = ACTIONS(2558),
    [anon_sym_union] = ACTIONS(2558),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2558),
    [anon_sym___typeof] = ACTIONS(2558),
    [anon_sym_typeof] = ACTIONS(2558),
    [aux_sym_preproc_undef_token1] = ACTIONS(2558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2558),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2558),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2558),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_API_AVAILABLE] = ACTIONS(2558),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_API_DEPRECATED] = ACTIONS(2558),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym___deprecated_msg] = ACTIONS(2558),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2558),
    [anon_sym_ATsynthesize] = ACTIONS(2560),
    [anon_sym_ATdynamic] = ACTIONS(2560),
    [anon_sym_ATproperty] = ACTIONS(2560),
    [anon_sym_BOOL] = ACTIONS(2558),
    [anon_sym_IMP] = ACTIONS(2558),
    [anon_sym_SEL] = ACTIONS(2558),
    [anon_sym_Class] = ACTIONS(2558),
    [anon_sym_id] = ACTIONS(2558),
  },
  [STATE(2158)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token2] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [aux_sym_preproc_else_token1] = ACTIONS(2474),
    [aux_sym_preproc_elif_token1] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_DASH] = ACTIONS(2476),
    [anon_sym_PLUS] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATproperty] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(2159)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token2] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [aux_sym_preproc_else_token1] = ACTIONS(2474),
    [aux_sym_preproc_elif_token1] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_DASH] = ACTIONS(2476),
    [anon_sym_PLUS] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATproperty] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(2160)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4796),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2161)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4820),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2162)] = {
    [sym_identifier] = ACTIONS(5444),
    [aux_sym_preproc_def_token1] = ACTIONS(5444),
    [aux_sym_preproc_if_token1] = ACTIONS(5444),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5444),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5444),
    [sym_preproc_directive] = ACTIONS(5444),
    [anon_sym_LPAREN2] = ACTIONS(5446),
    [anon_sym_DASH] = ACTIONS(5446),
    [anon_sym_PLUS] = ACTIONS(5446),
    [anon_sym_LT] = ACTIONS(5446),
    [anon_sym___extension__] = ACTIONS(5444),
    [anon_sym_typedef] = ACTIONS(5444),
    [anon_sym_extern] = ACTIONS(5444),
    [anon_sym___attribute__] = ACTIONS(5444),
    [anon_sym___attribute] = ACTIONS(5444),
    [anon_sym_noreturn] = ACTIONS(5444),
    [anon_sym_LBRACK] = ACTIONS(5446),
    [anon_sym___declspec] = ACTIONS(5444),
    [anon_sym___cdecl] = ACTIONS(5444),
    [anon_sym___clrcall] = ACTIONS(5444),
    [anon_sym___stdcall] = ACTIONS(5444),
    [anon_sym___fastcall] = ACTIONS(5444),
    [anon_sym___thiscall] = ACTIONS(5444),
    [anon_sym___vectorcall] = ACTIONS(5444),
    [anon_sym_LBRACE] = ACTIONS(5446),
    [anon_sym_signed] = ACTIONS(5444),
    [anon_sym_unsigned] = ACTIONS(5444),
    [anon_sym_long] = ACTIONS(5444),
    [anon_sym_short] = ACTIONS(5444),
    [anon_sym_static] = ACTIONS(5444),
    [anon_sym_auto] = ACTIONS(5444),
    [anon_sym_register] = ACTIONS(5444),
    [anon_sym_inline] = ACTIONS(5444),
    [anon_sym___inline] = ACTIONS(5444),
    [anon_sym___inline__] = ACTIONS(5444),
    [anon_sym___forceinline] = ACTIONS(5444),
    [anon_sym_thread_local] = ACTIONS(5444),
    [anon_sym___thread] = ACTIONS(5444),
    [anon_sym_CG_EXTERN] = ACTIONS(5444),
    [anon_sym_CG_INLINE] = ACTIONS(5444),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5444),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5444),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5444),
    [anon_sym_IBOutlet] = ACTIONS(5444),
    [anon_sym_IBInspectable] = ACTIONS(5444),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5444),
    [anon_sym_NS_INLINE] = ACTIONS(5444),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5444),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5444),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5444),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5444),
    [anon_sym_const] = ACTIONS(5444),
    [anon_sym_constexpr] = ACTIONS(5444),
    [anon_sym_volatile] = ACTIONS(5444),
    [anon_sym_restrict] = ACTIONS(5444),
    [anon_sym___restrict__] = ACTIONS(5444),
    [anon_sym__Atomic] = ACTIONS(5444),
    [anon_sym__Noreturn] = ACTIONS(5444),
    [anon_sym__Nonnull] = ACTIONS(5444),
    [anon_sym_nullable] = ACTIONS(5444),
    [anon_sym__Complex] = ACTIONS(5444),
    [anon_sym__Nullable] = ACTIONS(5444),
    [anon_sym__Nullable_result] = ACTIONS(5444),
    [anon_sym__Null_unspecified] = ACTIONS(5444),
    [anon_sym___autoreleasing] = ACTIONS(5444),
    [anon_sym___block] = ACTIONS(5444),
    [anon_sym___bridge] = ACTIONS(5444),
    [anon_sym___bridge_retained] = ACTIONS(5444),
    [anon_sym___bridge_transfer] = ACTIONS(5444),
    [anon_sym___complex] = ACTIONS(5444),
    [anon_sym___const] = ACTIONS(5444),
    [anon_sym___imag] = ACTIONS(5444),
    [anon_sym___kindof] = ACTIONS(5444),
    [anon_sym___nonnull] = ACTIONS(5444),
    [anon_sym___nullable] = ACTIONS(5444),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5444),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5444),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5444),
    [anon_sym___real] = ACTIONS(5444),
    [anon_sym___strong] = ACTIONS(5444),
    [anon_sym___unsafe_unretained] = ACTIONS(5444),
    [anon_sym___unused] = ACTIONS(5444),
    [anon_sym___weak] = ACTIONS(5444),
    [anon_sym_alignas] = ACTIONS(5444),
    [anon_sym__Alignas] = ACTIONS(5444),
    [sym_primitive_type] = ACTIONS(5444),
    [anon_sym_enum] = ACTIONS(5444),
    [anon_sym_COLON] = ACTIONS(5446),
    [anon_sym_struct] = ACTIONS(5444),
    [anon_sym_union] = ACTIONS(5444),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5444),
    [anon_sym___typeof] = ACTIONS(5444),
    [anon_sym_typeof] = ACTIONS(5444),
    [aux_sym_preproc_undef_token1] = ACTIONS(5444),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5444),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5444),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5444),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5444),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5444),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5444),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5444),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5444),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5444),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5444),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5444),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5444),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5444),
    [anon_sym_API_AVAILABLE] = ACTIONS(5444),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5444),
    [anon_sym_API_DEPRECATED] = ACTIONS(5444),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5444),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5444),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5444),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5444),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5444),
    [anon_sym___deprecated_msg] = ACTIONS(5444),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5444),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5444),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5444),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5444),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5444),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5444),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5444),
    [anon_sym_ATend] = ACTIONS(5446),
    [anon_sym_ATproperty] = ACTIONS(5446),
    [anon_sym_BOOL] = ACTIONS(5444),
    [anon_sym_IMP] = ACTIONS(5444),
    [anon_sym_SEL] = ACTIONS(5444),
    [anon_sym_Class] = ACTIONS(5444),
    [anon_sym_id] = ACTIONS(5444),
  },
  [STATE(2163)] = {
    [sym_identifier] = ACTIONS(5448),
    [aux_sym_preproc_def_token1] = ACTIONS(5448),
    [aux_sym_preproc_if_token1] = ACTIONS(5448),
    [aux_sym_preproc_if_token2] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5448),
    [aux_sym_preproc_else_token1] = ACTIONS(5448),
    [aux_sym_preproc_elif_token1] = ACTIONS(5448),
    [sym_preproc_directive] = ACTIONS(5448),
    [anon_sym_DASH] = ACTIONS(5450),
    [anon_sym_PLUS] = ACTIONS(5450),
    [anon_sym_SEMI] = ACTIONS(5452),
    [anon_sym___extension__] = ACTIONS(5448),
    [anon_sym_typedef] = ACTIONS(5448),
    [anon_sym_extern] = ACTIONS(5448),
    [anon_sym___attribute__] = ACTIONS(5448),
    [anon_sym___attribute] = ACTIONS(5448),
    [anon_sym_noreturn] = ACTIONS(5448),
    [anon_sym_LBRACK] = ACTIONS(5450),
    [anon_sym___declspec] = ACTIONS(5448),
    [anon_sym___cdecl] = ACTIONS(5448),
    [anon_sym___clrcall] = ACTIONS(5448),
    [anon_sym___stdcall] = ACTIONS(5448),
    [anon_sym___fastcall] = ACTIONS(5448),
    [anon_sym___thiscall] = ACTIONS(5448),
    [anon_sym___vectorcall] = ACTIONS(5448),
    [anon_sym_signed] = ACTIONS(5448),
    [anon_sym_unsigned] = ACTIONS(5448),
    [anon_sym_long] = ACTIONS(5448),
    [anon_sym_short] = ACTIONS(5448),
    [anon_sym_static] = ACTIONS(5448),
    [anon_sym_auto] = ACTIONS(5448),
    [anon_sym_register] = ACTIONS(5448),
    [anon_sym_inline] = ACTIONS(5448),
    [anon_sym___inline] = ACTIONS(5448),
    [anon_sym___inline__] = ACTIONS(5448),
    [anon_sym___forceinline] = ACTIONS(5448),
    [anon_sym_thread_local] = ACTIONS(5448),
    [anon_sym___thread] = ACTIONS(5448),
    [anon_sym_CG_EXTERN] = ACTIONS(5448),
    [anon_sym_CG_INLINE] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5448),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5448),
    [anon_sym_IBOutlet] = ACTIONS(5448),
    [anon_sym_IBInspectable] = ACTIONS(5448),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5448),
    [anon_sym_NS_INLINE] = ACTIONS(5448),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5448),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5448),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5448),
    [anon_sym_const] = ACTIONS(5448),
    [anon_sym_constexpr] = ACTIONS(5448),
    [anon_sym_volatile] = ACTIONS(5448),
    [anon_sym_restrict] = ACTIONS(5448),
    [anon_sym___restrict__] = ACTIONS(5448),
    [anon_sym__Atomic] = ACTIONS(5448),
    [anon_sym__Noreturn] = ACTIONS(5448),
    [anon_sym__Nonnull] = ACTIONS(5448),
    [anon_sym_nullable] = ACTIONS(5448),
    [anon_sym__Complex] = ACTIONS(5448),
    [anon_sym__Nullable] = ACTIONS(5448),
    [anon_sym__Nullable_result] = ACTIONS(5448),
    [anon_sym__Null_unspecified] = ACTIONS(5448),
    [anon_sym___autoreleasing] = ACTIONS(5448),
    [anon_sym___block] = ACTIONS(5448),
    [anon_sym___bridge] = ACTIONS(5448),
    [anon_sym___bridge_retained] = ACTIONS(5448),
    [anon_sym___bridge_transfer] = ACTIONS(5448),
    [anon_sym___complex] = ACTIONS(5448),
    [anon_sym___const] = ACTIONS(5448),
    [anon_sym___imag] = ACTIONS(5448),
    [anon_sym___kindof] = ACTIONS(5448),
    [anon_sym___nonnull] = ACTIONS(5448),
    [anon_sym___nullable] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5448),
    [anon_sym___real] = ACTIONS(5448),
    [anon_sym___strong] = ACTIONS(5448),
    [anon_sym___unsafe_unretained] = ACTIONS(5448),
    [anon_sym___unused] = ACTIONS(5448),
    [anon_sym___weak] = ACTIONS(5448),
    [anon_sym_alignas] = ACTIONS(5448),
    [anon_sym__Alignas] = ACTIONS(5448),
    [sym_primitive_type] = ACTIONS(5448),
    [anon_sym_enum] = ACTIONS(5448),
    [anon_sym_struct] = ACTIONS(5448),
    [anon_sym_union] = ACTIONS(5448),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5448),
    [anon_sym___typeof] = ACTIONS(5448),
    [anon_sym_typeof] = ACTIONS(5448),
    [aux_sym_preproc_undef_token1] = ACTIONS(5448),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5448),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5448),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5448),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_API_AVAILABLE] = ACTIONS(5448),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_API_DEPRECATED] = ACTIONS(5448),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym___deprecated_msg] = ACTIONS(5448),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5448),
    [anon_sym_ATsynthesize] = ACTIONS(5450),
    [anon_sym_ATdynamic] = ACTIONS(5450),
    [anon_sym_BOOL] = ACTIONS(5448),
    [anon_sym_IMP] = ACTIONS(5448),
    [anon_sym_SEL] = ACTIONS(5448),
    [anon_sym_Class] = ACTIONS(5448),
    [anon_sym_id] = ACTIONS(5448),
  },
  [STATE(2164)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token2] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [aux_sym_preproc_else_token1] = ACTIONS(2416),
    [aux_sym_preproc_elif_token1] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_DASH] = ACTIONS(2418),
    [anon_sym_PLUS] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATproperty] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(2165)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token2] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [aux_sym_preproc_else_token1] = ACTIONS(2420),
    [aux_sym_preproc_elif_token1] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_DASH] = ACTIONS(2422),
    [anon_sym_PLUS] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATproperty] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(2166)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4803),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2167)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2185),
    [sym_identifier] = ACTIONS(5454),
    [aux_sym_preproc_def_token1] = ACTIONS(5454),
    [aux_sym_preproc_if_token1] = ACTIONS(5454),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5454),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5454),
    [sym_preproc_directive] = ACTIONS(5454),
    [anon_sym_DASH] = ACTIONS(5456),
    [anon_sym_PLUS] = ACTIONS(5456),
    [anon_sym_SEMI] = ACTIONS(5349),
    [anon_sym___extension__] = ACTIONS(5454),
    [anon_sym_typedef] = ACTIONS(5454),
    [anon_sym_extern] = ACTIONS(5454),
    [anon_sym___attribute__] = ACTIONS(5454),
    [anon_sym___attribute] = ACTIONS(5454),
    [anon_sym_noreturn] = ACTIONS(5454),
    [anon_sym_LBRACK] = ACTIONS(5456),
    [anon_sym___declspec] = ACTIONS(5454),
    [anon_sym___cdecl] = ACTIONS(5454),
    [anon_sym___clrcall] = ACTIONS(5454),
    [anon_sym___stdcall] = ACTIONS(5454),
    [anon_sym___fastcall] = ACTIONS(5454),
    [anon_sym___thiscall] = ACTIONS(5454),
    [anon_sym___vectorcall] = ACTIONS(5454),
    [anon_sym_signed] = ACTIONS(5454),
    [anon_sym_unsigned] = ACTIONS(5454),
    [anon_sym_long] = ACTIONS(5454),
    [anon_sym_short] = ACTIONS(5454),
    [anon_sym_static] = ACTIONS(5454),
    [anon_sym_auto] = ACTIONS(5454),
    [anon_sym_register] = ACTIONS(5454),
    [anon_sym_inline] = ACTIONS(5454),
    [anon_sym___inline] = ACTIONS(5454),
    [anon_sym___inline__] = ACTIONS(5454),
    [anon_sym___forceinline] = ACTIONS(5454),
    [anon_sym_thread_local] = ACTIONS(5454),
    [anon_sym___thread] = ACTIONS(5454),
    [anon_sym_CG_EXTERN] = ACTIONS(5454),
    [anon_sym_CG_INLINE] = ACTIONS(5454),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5454),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5454),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5454),
    [anon_sym_IBOutlet] = ACTIONS(5454),
    [anon_sym_IBInspectable] = ACTIONS(5454),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5454),
    [anon_sym_NS_INLINE] = ACTIONS(5454),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5454),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5454),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5454),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5454),
    [anon_sym_const] = ACTIONS(5454),
    [anon_sym_constexpr] = ACTIONS(5454),
    [anon_sym_volatile] = ACTIONS(5454),
    [anon_sym_restrict] = ACTIONS(5454),
    [anon_sym___restrict__] = ACTIONS(5454),
    [anon_sym__Atomic] = ACTIONS(5454),
    [anon_sym__Noreturn] = ACTIONS(5454),
    [anon_sym__Nonnull] = ACTIONS(5454),
    [anon_sym_nullable] = ACTIONS(5454),
    [anon_sym__Complex] = ACTIONS(5454),
    [anon_sym__Nullable] = ACTIONS(5454),
    [anon_sym__Nullable_result] = ACTIONS(5454),
    [anon_sym__Null_unspecified] = ACTIONS(5454),
    [anon_sym___autoreleasing] = ACTIONS(5454),
    [anon_sym___block] = ACTIONS(5454),
    [anon_sym___bridge] = ACTIONS(5454),
    [anon_sym___bridge_retained] = ACTIONS(5454),
    [anon_sym___bridge_transfer] = ACTIONS(5454),
    [anon_sym___complex] = ACTIONS(5454),
    [anon_sym___const] = ACTIONS(5454),
    [anon_sym___imag] = ACTIONS(5454),
    [anon_sym___kindof] = ACTIONS(5454),
    [anon_sym___nonnull] = ACTIONS(5454),
    [anon_sym___nullable] = ACTIONS(5454),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5454),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5454),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5454),
    [anon_sym___real] = ACTIONS(5454),
    [anon_sym___strong] = ACTIONS(5454),
    [anon_sym___unsafe_unretained] = ACTIONS(5454),
    [anon_sym___unused] = ACTIONS(5454),
    [anon_sym___weak] = ACTIONS(5454),
    [anon_sym_alignas] = ACTIONS(5454),
    [anon_sym__Alignas] = ACTIONS(5454),
    [sym_primitive_type] = ACTIONS(5454),
    [anon_sym_enum] = ACTIONS(5454),
    [anon_sym_struct] = ACTIONS(5454),
    [anon_sym_union] = ACTIONS(5454),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5454),
    [anon_sym___typeof] = ACTIONS(5454),
    [anon_sym_typeof] = ACTIONS(5454),
    [aux_sym_preproc_undef_token1] = ACTIONS(5454),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5454),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5454),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5454),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5454),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5454),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5454),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5454),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5454),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5454),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_API_AVAILABLE] = ACTIONS(5454),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_API_DEPRECATED] = ACTIONS(5454),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5454),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5454),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5454),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5454),
    [anon_sym___deprecated_msg] = ACTIONS(5454),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5454),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5454),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5454),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5454),
    [anon_sym_ATend] = ACTIONS(5456),
    [anon_sym_AToptional] = ACTIONS(5456),
    [anon_sym_ATrequired] = ACTIONS(5456),
    [anon_sym_ATproperty] = ACTIONS(5456),
    [anon_sym_BOOL] = ACTIONS(5454),
    [anon_sym_IMP] = ACTIONS(5454),
    [anon_sym_SEL] = ACTIONS(5454),
    [anon_sym_Class] = ACTIONS(5454),
    [anon_sym_id] = ACTIONS(5454),
  },
  [STATE(2168)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4818),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2169)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token2] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [aux_sym_preproc_else_token1] = ACTIONS(2424),
    [aux_sym_preproc_elif_token1] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_DASH] = ACTIONS(2426),
    [anon_sym_PLUS] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATproperty] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(2170)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4806),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2171)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4814),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2172)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2185),
    [sym_identifier] = ACTIONS(5423),
    [aux_sym_preproc_def_token1] = ACTIONS(5423),
    [aux_sym_preproc_if_token1] = ACTIONS(5423),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5423),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5423),
    [sym_preproc_directive] = ACTIONS(5423),
    [anon_sym_DASH] = ACTIONS(5425),
    [anon_sym_PLUS] = ACTIONS(5425),
    [anon_sym_SEMI] = ACTIONS(5349),
    [anon_sym___extension__] = ACTIONS(5423),
    [anon_sym_typedef] = ACTIONS(5423),
    [anon_sym_extern] = ACTIONS(5423),
    [anon_sym___attribute__] = ACTIONS(5423),
    [anon_sym___attribute] = ACTIONS(5423),
    [anon_sym_noreturn] = ACTIONS(5423),
    [anon_sym_LBRACK] = ACTIONS(5425),
    [anon_sym___declspec] = ACTIONS(5423),
    [anon_sym___cdecl] = ACTIONS(5423),
    [anon_sym___clrcall] = ACTIONS(5423),
    [anon_sym___stdcall] = ACTIONS(5423),
    [anon_sym___fastcall] = ACTIONS(5423),
    [anon_sym___thiscall] = ACTIONS(5423),
    [anon_sym___vectorcall] = ACTIONS(5423),
    [anon_sym_signed] = ACTIONS(5423),
    [anon_sym_unsigned] = ACTIONS(5423),
    [anon_sym_long] = ACTIONS(5423),
    [anon_sym_short] = ACTIONS(5423),
    [anon_sym_static] = ACTIONS(5423),
    [anon_sym_auto] = ACTIONS(5423),
    [anon_sym_register] = ACTIONS(5423),
    [anon_sym_inline] = ACTIONS(5423),
    [anon_sym___inline] = ACTIONS(5423),
    [anon_sym___inline__] = ACTIONS(5423),
    [anon_sym___forceinline] = ACTIONS(5423),
    [anon_sym_thread_local] = ACTIONS(5423),
    [anon_sym___thread] = ACTIONS(5423),
    [anon_sym_CG_EXTERN] = ACTIONS(5423),
    [anon_sym_CG_INLINE] = ACTIONS(5423),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5423),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5423),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5423),
    [anon_sym_IBOutlet] = ACTIONS(5423),
    [anon_sym_IBInspectable] = ACTIONS(5423),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5423),
    [anon_sym_NS_INLINE] = ACTIONS(5423),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5423),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5423),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5423),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5423),
    [anon_sym_const] = ACTIONS(5423),
    [anon_sym_constexpr] = ACTIONS(5423),
    [anon_sym_volatile] = ACTIONS(5423),
    [anon_sym_restrict] = ACTIONS(5423),
    [anon_sym___restrict__] = ACTIONS(5423),
    [anon_sym__Atomic] = ACTIONS(5423),
    [anon_sym__Noreturn] = ACTIONS(5423),
    [anon_sym__Nonnull] = ACTIONS(5423),
    [anon_sym_nullable] = ACTIONS(5423),
    [anon_sym__Complex] = ACTIONS(5423),
    [anon_sym__Nullable] = ACTIONS(5423),
    [anon_sym__Nullable_result] = ACTIONS(5423),
    [anon_sym__Null_unspecified] = ACTIONS(5423),
    [anon_sym___autoreleasing] = ACTIONS(5423),
    [anon_sym___block] = ACTIONS(5423),
    [anon_sym___bridge] = ACTIONS(5423),
    [anon_sym___bridge_retained] = ACTIONS(5423),
    [anon_sym___bridge_transfer] = ACTIONS(5423),
    [anon_sym___complex] = ACTIONS(5423),
    [anon_sym___const] = ACTIONS(5423),
    [anon_sym___imag] = ACTIONS(5423),
    [anon_sym___kindof] = ACTIONS(5423),
    [anon_sym___nonnull] = ACTIONS(5423),
    [anon_sym___nullable] = ACTIONS(5423),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5423),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5423),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5423),
    [anon_sym___real] = ACTIONS(5423),
    [anon_sym___strong] = ACTIONS(5423),
    [anon_sym___unsafe_unretained] = ACTIONS(5423),
    [anon_sym___unused] = ACTIONS(5423),
    [anon_sym___weak] = ACTIONS(5423),
    [anon_sym_alignas] = ACTIONS(5423),
    [anon_sym__Alignas] = ACTIONS(5423),
    [sym_primitive_type] = ACTIONS(5423),
    [anon_sym_enum] = ACTIONS(5423),
    [anon_sym_struct] = ACTIONS(5423),
    [anon_sym_union] = ACTIONS(5423),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5423),
    [anon_sym___typeof] = ACTIONS(5423),
    [anon_sym_typeof] = ACTIONS(5423),
    [aux_sym_preproc_undef_token1] = ACTIONS(5423),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5423),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5423),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5423),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5423),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5423),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5423),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5423),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5423),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5423),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_API_AVAILABLE] = ACTIONS(5423),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_API_DEPRECATED] = ACTIONS(5423),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5423),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5423),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5423),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5423),
    [anon_sym___deprecated_msg] = ACTIONS(5423),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5423),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5423),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5423),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5423),
    [anon_sym_ATend] = ACTIONS(5425),
    [anon_sym_AToptional] = ACTIONS(5425),
    [anon_sym_ATrequired] = ACTIONS(5425),
    [anon_sym_ATproperty] = ACTIONS(5425),
    [anon_sym_BOOL] = ACTIONS(5423),
    [anon_sym_IMP] = ACTIONS(5423),
    [anon_sym_SEL] = ACTIONS(5423),
    [anon_sym_Class] = ACTIONS(5423),
    [anon_sym_id] = ACTIONS(5423),
  },
  [STATE(2173)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4807),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2174)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4822),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2175)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4812),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2176)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token2] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [aux_sym_preproc_else_token1] = ACTIONS(2416),
    [aux_sym_preproc_elif_token1] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_DASH] = ACTIONS(2418),
    [anon_sym_PLUS] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATproperty] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(2177)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4810),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2178)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token2] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [aux_sym_preproc_else_token1] = ACTIONS(2424),
    [aux_sym_preproc_elif_token1] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_DASH] = ACTIONS(2426),
    [anon_sym_PLUS] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATproperty] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(2179)] = {
    [sym__declaration_modifiers] = STATE(2230),
    [sym__declaration_specifiers] = STATE(4813),
    [sym_attribute_specifier] = STATE(2230),
    [sym_attribute_declaration] = STATE(2230),
    [sym_ms_declspec_modifier] = STATE(2230),
    [sym_storage_class_specifier] = STATE(2230),
    [sym_type_qualifier] = STATE(2230),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2859),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(2230),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2230),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2180)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2138),
    [sym_identifier] = ACTIONS(5458),
    [aux_sym_preproc_def_token1] = ACTIONS(5458),
    [aux_sym_preproc_if_token1] = ACTIONS(5458),
    [aux_sym_preproc_if_token2] = ACTIONS(5458),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5458),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5458),
    [aux_sym_preproc_else_token1] = ACTIONS(5458),
    [aux_sym_preproc_elif_token1] = ACTIONS(5458),
    [sym_preproc_directive] = ACTIONS(5458),
    [anon_sym_DASH] = ACTIONS(5460),
    [anon_sym_PLUS] = ACTIONS(5460),
    [anon_sym_SEMI] = ACTIONS(5357),
    [anon_sym___extension__] = ACTIONS(5458),
    [anon_sym_typedef] = ACTIONS(5458),
    [anon_sym_extern] = ACTIONS(5458),
    [anon_sym___attribute__] = ACTIONS(5458),
    [anon_sym___attribute] = ACTIONS(5458),
    [anon_sym_noreturn] = ACTIONS(5458),
    [anon_sym_LBRACK] = ACTIONS(5460),
    [anon_sym___declspec] = ACTIONS(5458),
    [anon_sym___cdecl] = ACTIONS(5458),
    [anon_sym___clrcall] = ACTIONS(5458),
    [anon_sym___stdcall] = ACTIONS(5458),
    [anon_sym___fastcall] = ACTIONS(5458),
    [anon_sym___thiscall] = ACTIONS(5458),
    [anon_sym___vectorcall] = ACTIONS(5458),
    [anon_sym_signed] = ACTIONS(5458),
    [anon_sym_unsigned] = ACTIONS(5458),
    [anon_sym_long] = ACTIONS(5458),
    [anon_sym_short] = ACTIONS(5458),
    [anon_sym_static] = ACTIONS(5458),
    [anon_sym_auto] = ACTIONS(5458),
    [anon_sym_register] = ACTIONS(5458),
    [anon_sym_inline] = ACTIONS(5458),
    [anon_sym___inline] = ACTIONS(5458),
    [anon_sym___inline__] = ACTIONS(5458),
    [anon_sym___forceinline] = ACTIONS(5458),
    [anon_sym_thread_local] = ACTIONS(5458),
    [anon_sym___thread] = ACTIONS(5458),
    [anon_sym_CG_EXTERN] = ACTIONS(5458),
    [anon_sym_CG_INLINE] = ACTIONS(5458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5458),
    [anon_sym_IBOutlet] = ACTIONS(5458),
    [anon_sym_IBInspectable] = ACTIONS(5458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5458),
    [anon_sym_NS_INLINE] = ACTIONS(5458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5458),
    [anon_sym_const] = ACTIONS(5458),
    [anon_sym_constexpr] = ACTIONS(5458),
    [anon_sym_volatile] = ACTIONS(5458),
    [anon_sym_restrict] = ACTIONS(5458),
    [anon_sym___restrict__] = ACTIONS(5458),
    [anon_sym__Atomic] = ACTIONS(5458),
    [anon_sym__Noreturn] = ACTIONS(5458),
    [anon_sym__Nonnull] = ACTIONS(5458),
    [anon_sym_nullable] = ACTIONS(5458),
    [anon_sym__Complex] = ACTIONS(5458),
    [anon_sym__Nullable] = ACTIONS(5458),
    [anon_sym__Nullable_result] = ACTIONS(5458),
    [anon_sym__Null_unspecified] = ACTIONS(5458),
    [anon_sym___autoreleasing] = ACTIONS(5458),
    [anon_sym___block] = ACTIONS(5458),
    [anon_sym___bridge] = ACTIONS(5458),
    [anon_sym___bridge_retained] = ACTIONS(5458),
    [anon_sym___bridge_transfer] = ACTIONS(5458),
    [anon_sym___complex] = ACTIONS(5458),
    [anon_sym___const] = ACTIONS(5458),
    [anon_sym___imag] = ACTIONS(5458),
    [anon_sym___kindof] = ACTIONS(5458),
    [anon_sym___nonnull] = ACTIONS(5458),
    [anon_sym___nullable] = ACTIONS(5458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5458),
    [anon_sym___real] = ACTIONS(5458),
    [anon_sym___strong] = ACTIONS(5458),
    [anon_sym___unsafe_unretained] = ACTIONS(5458),
    [anon_sym___unused] = ACTIONS(5458),
    [anon_sym___weak] = ACTIONS(5458),
    [anon_sym_alignas] = ACTIONS(5458),
    [anon_sym__Alignas] = ACTIONS(5458),
    [sym_primitive_type] = ACTIONS(5458),
    [anon_sym_enum] = ACTIONS(5458),
    [anon_sym_struct] = ACTIONS(5458),
    [anon_sym_union] = ACTIONS(5458),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5458),
    [anon_sym___typeof] = ACTIONS(5458),
    [anon_sym_typeof] = ACTIONS(5458),
    [aux_sym_preproc_undef_token1] = ACTIONS(5458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_API_AVAILABLE] = ACTIONS(5458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_API_DEPRECATED] = ACTIONS(5458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5458),
    [anon_sym___deprecated_msg] = ACTIONS(5458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5458),
    [anon_sym_ATproperty] = ACTIONS(5460),
    [anon_sym_BOOL] = ACTIONS(5458),
    [anon_sym_IMP] = ACTIONS(5458),
    [anon_sym_SEL] = ACTIONS(5458),
    [anon_sym_Class] = ACTIONS(5458),
    [anon_sym_id] = ACTIONS(5458),
  },
  [STATE(2181)] = {
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_LPAREN2] = ACTIONS(4026),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4024),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4024),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4024),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym_SEMI] = ACTIONS(4026),
    [anon_sym___extension__] = ACTIONS(4026),
    [anon_sym_extern] = ACTIONS(4026),
    [anon_sym___attribute__] = ACTIONS(4026),
    [anon_sym___attribute] = ACTIONS(4024),
    [anon_sym_noreturn] = ACTIONS(4026),
    [anon_sym_LBRACK] = ACTIONS(4026),
    [anon_sym___declspec] = ACTIONS(4026),
    [anon_sym_static] = ACTIONS(4026),
    [anon_sym_EQ] = ACTIONS(4906),
    [anon_sym_auto] = ACTIONS(4026),
    [anon_sym_register] = ACTIONS(4026),
    [anon_sym_inline] = ACTIONS(4026),
    [anon_sym___inline] = ACTIONS(4024),
    [anon_sym___inline__] = ACTIONS(4026),
    [anon_sym___forceinline] = ACTIONS(4026),
    [anon_sym_thread_local] = ACTIONS(4026),
    [anon_sym___thread] = ACTIONS(4026),
    [anon_sym_CG_EXTERN] = ACTIONS(4026),
    [anon_sym_CG_INLINE] = ACTIONS(4026),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4026),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4026),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4026),
    [anon_sym_IBOutlet] = ACTIONS(4026),
    [anon_sym_IBInspectable] = ACTIONS(4026),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4026),
    [anon_sym_NS_INLINE] = ACTIONS(4026),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4026),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4026),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4026),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4026),
    [anon_sym_const] = ACTIONS(4024),
    [anon_sym_constexpr] = ACTIONS(4026),
    [anon_sym_volatile] = ACTIONS(4026),
    [anon_sym_restrict] = ACTIONS(4026),
    [anon_sym___restrict__] = ACTIONS(4026),
    [anon_sym__Atomic] = ACTIONS(4026),
    [anon_sym__Noreturn] = ACTIONS(4026),
    [anon_sym__Nonnull] = ACTIONS(4026),
    [anon_sym_nullable] = ACTIONS(4026),
    [anon_sym__Complex] = ACTIONS(4026),
    [anon_sym__Nullable] = ACTIONS(4024),
    [anon_sym__Nullable_result] = ACTIONS(4026),
    [anon_sym__Null_unspecified] = ACTIONS(4026),
    [anon_sym___autoreleasing] = ACTIONS(4026),
    [anon_sym___block] = ACTIONS(4026),
    [anon_sym___bridge] = ACTIONS(4024),
    [anon_sym___bridge_retained] = ACTIONS(4026),
    [anon_sym___bridge_transfer] = ACTIONS(4026),
    [anon_sym___complex] = ACTIONS(4026),
    [anon_sym___const] = ACTIONS(4026),
    [anon_sym___imag] = ACTIONS(4026),
    [anon_sym___kindof] = ACTIONS(4026),
    [anon_sym___nonnull] = ACTIONS(4026),
    [anon_sym___nullable] = ACTIONS(4026),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4026),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4026),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4026),
    [anon_sym___real] = ACTIONS(4026),
    [anon_sym___strong] = ACTIONS(4026),
    [anon_sym___unsafe_unretained] = ACTIONS(4026),
    [anon_sym___unused] = ACTIONS(4026),
    [anon_sym___weak] = ACTIONS(4026),
    [anon_sym_alignas] = ACTIONS(4026),
    [anon_sym__Alignas] = ACTIONS(4026),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(4908),
    [anon_sym_SLASH_EQ] = ACTIONS(4908),
    [anon_sym_PERCENT_EQ] = ACTIONS(4908),
    [anon_sym_PLUS_EQ] = ACTIONS(4908),
    [anon_sym_DASH_EQ] = ACTIONS(4908),
    [anon_sym_LT_LT_EQ] = ACTIONS(4908),
    [anon_sym_GT_GT_EQ] = ACTIONS(4908),
    [anon_sym_AMP_EQ] = ACTIONS(4908),
    [anon_sym_CARET_EQ] = ACTIONS(4908),
    [anon_sym_PIPE_EQ] = ACTIONS(4908),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4026),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4026),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4026),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4026),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4026),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4026),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4026),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4026),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4026),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4026),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4024),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4026),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4026),
    [anon_sym_API_AVAILABLE] = ACTIONS(4026),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4026),
    [anon_sym_API_DEPRECATED] = ACTIONS(4026),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4026),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4026),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4026),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4026),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4026),
    [anon_sym___deprecated_msg] = ACTIONS(4026),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4026),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4026),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4026),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4026),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4026),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4026),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4026),
  },
  [STATE(2182)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token2] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [aux_sym_preproc_else_token1] = ACTIONS(2420),
    [aux_sym_preproc_elif_token1] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_DASH] = ACTIONS(2422),
    [anon_sym_PLUS] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATproperty] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(2183)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2138),
    [sym_identifier] = ACTIONS(5462),
    [aux_sym_preproc_def_token1] = ACTIONS(5462),
    [aux_sym_preproc_if_token1] = ACTIONS(5462),
    [aux_sym_preproc_if_token2] = ACTIONS(5462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5462),
    [aux_sym_preproc_else_token1] = ACTIONS(5462),
    [aux_sym_preproc_elif_token1] = ACTIONS(5462),
    [sym_preproc_directive] = ACTIONS(5462),
    [anon_sym_DASH] = ACTIONS(5464),
    [anon_sym_PLUS] = ACTIONS(5464),
    [anon_sym_SEMI] = ACTIONS(5357),
    [anon_sym___extension__] = ACTIONS(5462),
    [anon_sym_typedef] = ACTIONS(5462),
    [anon_sym_extern] = ACTIONS(5462),
    [anon_sym___attribute__] = ACTIONS(5462),
    [anon_sym___attribute] = ACTIONS(5462),
    [anon_sym_noreturn] = ACTIONS(5462),
    [anon_sym_LBRACK] = ACTIONS(5464),
    [anon_sym___declspec] = ACTIONS(5462),
    [anon_sym___cdecl] = ACTIONS(5462),
    [anon_sym___clrcall] = ACTIONS(5462),
    [anon_sym___stdcall] = ACTIONS(5462),
    [anon_sym___fastcall] = ACTIONS(5462),
    [anon_sym___thiscall] = ACTIONS(5462),
    [anon_sym___vectorcall] = ACTIONS(5462),
    [anon_sym_signed] = ACTIONS(5462),
    [anon_sym_unsigned] = ACTIONS(5462),
    [anon_sym_long] = ACTIONS(5462),
    [anon_sym_short] = ACTIONS(5462),
    [anon_sym_static] = ACTIONS(5462),
    [anon_sym_auto] = ACTIONS(5462),
    [anon_sym_register] = ACTIONS(5462),
    [anon_sym_inline] = ACTIONS(5462),
    [anon_sym___inline] = ACTIONS(5462),
    [anon_sym___inline__] = ACTIONS(5462),
    [anon_sym___forceinline] = ACTIONS(5462),
    [anon_sym_thread_local] = ACTIONS(5462),
    [anon_sym___thread] = ACTIONS(5462),
    [anon_sym_CG_EXTERN] = ACTIONS(5462),
    [anon_sym_CG_INLINE] = ACTIONS(5462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5462),
    [anon_sym_IBOutlet] = ACTIONS(5462),
    [anon_sym_IBInspectable] = ACTIONS(5462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5462),
    [anon_sym_NS_INLINE] = ACTIONS(5462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5462),
    [anon_sym_const] = ACTIONS(5462),
    [anon_sym_constexpr] = ACTIONS(5462),
    [anon_sym_volatile] = ACTIONS(5462),
    [anon_sym_restrict] = ACTIONS(5462),
    [anon_sym___restrict__] = ACTIONS(5462),
    [anon_sym__Atomic] = ACTIONS(5462),
    [anon_sym__Noreturn] = ACTIONS(5462),
    [anon_sym__Nonnull] = ACTIONS(5462),
    [anon_sym_nullable] = ACTIONS(5462),
    [anon_sym__Complex] = ACTIONS(5462),
    [anon_sym__Nullable] = ACTIONS(5462),
    [anon_sym__Nullable_result] = ACTIONS(5462),
    [anon_sym__Null_unspecified] = ACTIONS(5462),
    [anon_sym___autoreleasing] = ACTIONS(5462),
    [anon_sym___block] = ACTIONS(5462),
    [anon_sym___bridge] = ACTIONS(5462),
    [anon_sym___bridge_retained] = ACTIONS(5462),
    [anon_sym___bridge_transfer] = ACTIONS(5462),
    [anon_sym___complex] = ACTIONS(5462),
    [anon_sym___const] = ACTIONS(5462),
    [anon_sym___imag] = ACTIONS(5462),
    [anon_sym___kindof] = ACTIONS(5462),
    [anon_sym___nonnull] = ACTIONS(5462),
    [anon_sym___nullable] = ACTIONS(5462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5462),
    [anon_sym___real] = ACTIONS(5462),
    [anon_sym___strong] = ACTIONS(5462),
    [anon_sym___unsafe_unretained] = ACTIONS(5462),
    [anon_sym___unused] = ACTIONS(5462),
    [anon_sym___weak] = ACTIONS(5462),
    [anon_sym_alignas] = ACTIONS(5462),
    [anon_sym__Alignas] = ACTIONS(5462),
    [sym_primitive_type] = ACTIONS(5462),
    [anon_sym_enum] = ACTIONS(5462),
    [anon_sym_struct] = ACTIONS(5462),
    [anon_sym_union] = ACTIONS(5462),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5462),
    [anon_sym___typeof] = ACTIONS(5462),
    [anon_sym_typeof] = ACTIONS(5462),
    [aux_sym_preproc_undef_token1] = ACTIONS(5462),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5462),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_API_AVAILABLE] = ACTIONS(5462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_API_DEPRECATED] = ACTIONS(5462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5462),
    [anon_sym___deprecated_msg] = ACTIONS(5462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5462),
    [anon_sym_ATproperty] = ACTIONS(5464),
    [anon_sym_BOOL] = ACTIONS(5462),
    [anon_sym_IMP] = ACTIONS(5462),
    [anon_sym_SEL] = ACTIONS(5462),
    [anon_sym_Class] = ACTIONS(5462),
    [anon_sym_id] = ACTIONS(5462),
  },
  [STATE(2184)] = {
    [sym_identifier] = ACTIONS(5466),
    [aux_sym_preproc_def_token1] = ACTIONS(5466),
    [aux_sym_preproc_if_token1] = ACTIONS(5466),
    [aux_sym_preproc_if_token2] = ACTIONS(5466),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5466),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5466),
    [aux_sym_preproc_else_token1] = ACTIONS(5466),
    [aux_sym_preproc_elif_token1] = ACTIONS(5466),
    [sym_preproc_directive] = ACTIONS(5466),
    [anon_sym_DASH] = ACTIONS(5468),
    [anon_sym_PLUS] = ACTIONS(5468),
    [anon_sym_SEMI] = ACTIONS(5470),
    [anon_sym___extension__] = ACTIONS(5466),
    [anon_sym_typedef] = ACTIONS(5466),
    [anon_sym_extern] = ACTIONS(5466),
    [anon_sym___attribute__] = ACTIONS(5466),
    [anon_sym___attribute] = ACTIONS(5466),
    [anon_sym_noreturn] = ACTIONS(5466),
    [anon_sym_LBRACK] = ACTIONS(5468),
    [anon_sym___declspec] = ACTIONS(5466),
    [anon_sym___cdecl] = ACTIONS(5466),
    [anon_sym___clrcall] = ACTIONS(5466),
    [anon_sym___stdcall] = ACTIONS(5466),
    [anon_sym___fastcall] = ACTIONS(5466),
    [anon_sym___thiscall] = ACTIONS(5466),
    [anon_sym___vectorcall] = ACTIONS(5466),
    [anon_sym_signed] = ACTIONS(5466),
    [anon_sym_unsigned] = ACTIONS(5466),
    [anon_sym_long] = ACTIONS(5466),
    [anon_sym_short] = ACTIONS(5466),
    [anon_sym_static] = ACTIONS(5466),
    [anon_sym_auto] = ACTIONS(5466),
    [anon_sym_register] = ACTIONS(5466),
    [anon_sym_inline] = ACTIONS(5466),
    [anon_sym___inline] = ACTIONS(5466),
    [anon_sym___inline__] = ACTIONS(5466),
    [anon_sym___forceinline] = ACTIONS(5466),
    [anon_sym_thread_local] = ACTIONS(5466),
    [anon_sym___thread] = ACTIONS(5466),
    [anon_sym_CG_EXTERN] = ACTIONS(5466),
    [anon_sym_CG_INLINE] = ACTIONS(5466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5466),
    [anon_sym_IBOutlet] = ACTIONS(5466),
    [anon_sym_IBInspectable] = ACTIONS(5466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5466),
    [anon_sym_NS_INLINE] = ACTIONS(5466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5466),
    [anon_sym_const] = ACTIONS(5466),
    [anon_sym_constexpr] = ACTIONS(5466),
    [anon_sym_volatile] = ACTIONS(5466),
    [anon_sym_restrict] = ACTIONS(5466),
    [anon_sym___restrict__] = ACTIONS(5466),
    [anon_sym__Atomic] = ACTIONS(5466),
    [anon_sym__Noreturn] = ACTIONS(5466),
    [anon_sym__Nonnull] = ACTIONS(5466),
    [anon_sym_nullable] = ACTIONS(5466),
    [anon_sym__Complex] = ACTIONS(5466),
    [anon_sym__Nullable] = ACTIONS(5466),
    [anon_sym__Nullable_result] = ACTIONS(5466),
    [anon_sym__Null_unspecified] = ACTIONS(5466),
    [anon_sym___autoreleasing] = ACTIONS(5466),
    [anon_sym___block] = ACTIONS(5466),
    [anon_sym___bridge] = ACTIONS(5466),
    [anon_sym___bridge_retained] = ACTIONS(5466),
    [anon_sym___bridge_transfer] = ACTIONS(5466),
    [anon_sym___complex] = ACTIONS(5466),
    [anon_sym___const] = ACTIONS(5466),
    [anon_sym___imag] = ACTIONS(5466),
    [anon_sym___kindof] = ACTIONS(5466),
    [anon_sym___nonnull] = ACTIONS(5466),
    [anon_sym___nullable] = ACTIONS(5466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5466),
    [anon_sym___real] = ACTIONS(5466),
    [anon_sym___strong] = ACTIONS(5466),
    [anon_sym___unsafe_unretained] = ACTIONS(5466),
    [anon_sym___unused] = ACTIONS(5466),
    [anon_sym___weak] = ACTIONS(5466),
    [anon_sym_alignas] = ACTIONS(5466),
    [anon_sym__Alignas] = ACTIONS(5466),
    [sym_primitive_type] = ACTIONS(5466),
    [anon_sym_enum] = ACTIONS(5466),
    [anon_sym_struct] = ACTIONS(5466),
    [anon_sym_union] = ACTIONS(5466),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5466),
    [anon_sym___typeof] = ACTIONS(5466),
    [anon_sym_typeof] = ACTIONS(5466),
    [aux_sym_preproc_undef_token1] = ACTIONS(5466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_API_AVAILABLE] = ACTIONS(5466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_API_DEPRECATED] = ACTIONS(5466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5466),
    [anon_sym___deprecated_msg] = ACTIONS(5466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5466),
    [anon_sym_ATsynthesize] = ACTIONS(5468),
    [anon_sym_ATdynamic] = ACTIONS(5468),
    [anon_sym_BOOL] = ACTIONS(5466),
    [anon_sym_IMP] = ACTIONS(5466),
    [anon_sym_SEL] = ACTIONS(5466),
    [anon_sym_Class] = ACTIONS(5466),
    [anon_sym_id] = ACTIONS(5466),
  },
  [STATE(2185)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2185),
    [sym_identifier] = ACTIONS(5431),
    [aux_sym_preproc_def_token1] = ACTIONS(5431),
    [aux_sym_preproc_if_token1] = ACTIONS(5431),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5431),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5431),
    [sym_preproc_directive] = ACTIONS(5431),
    [anon_sym_DASH] = ACTIONS(5433),
    [anon_sym_PLUS] = ACTIONS(5433),
    [anon_sym_SEMI] = ACTIONS(5472),
    [anon_sym___extension__] = ACTIONS(5431),
    [anon_sym_typedef] = ACTIONS(5431),
    [anon_sym_extern] = ACTIONS(5431),
    [anon_sym___attribute__] = ACTIONS(5431),
    [anon_sym___attribute] = ACTIONS(5431),
    [anon_sym_noreturn] = ACTIONS(5431),
    [anon_sym_LBRACK] = ACTIONS(5433),
    [anon_sym___declspec] = ACTIONS(5431),
    [anon_sym___cdecl] = ACTIONS(5431),
    [anon_sym___clrcall] = ACTIONS(5431),
    [anon_sym___stdcall] = ACTIONS(5431),
    [anon_sym___fastcall] = ACTIONS(5431),
    [anon_sym___thiscall] = ACTIONS(5431),
    [anon_sym___vectorcall] = ACTIONS(5431),
    [anon_sym_signed] = ACTIONS(5431),
    [anon_sym_unsigned] = ACTIONS(5431),
    [anon_sym_long] = ACTIONS(5431),
    [anon_sym_short] = ACTIONS(5431),
    [anon_sym_static] = ACTIONS(5431),
    [anon_sym_auto] = ACTIONS(5431),
    [anon_sym_register] = ACTIONS(5431),
    [anon_sym_inline] = ACTIONS(5431),
    [anon_sym___inline] = ACTIONS(5431),
    [anon_sym___inline__] = ACTIONS(5431),
    [anon_sym___forceinline] = ACTIONS(5431),
    [anon_sym_thread_local] = ACTIONS(5431),
    [anon_sym___thread] = ACTIONS(5431),
    [anon_sym_CG_EXTERN] = ACTIONS(5431),
    [anon_sym_CG_INLINE] = ACTIONS(5431),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5431),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5431),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5431),
    [anon_sym_IBOutlet] = ACTIONS(5431),
    [anon_sym_IBInspectable] = ACTIONS(5431),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5431),
    [anon_sym_NS_INLINE] = ACTIONS(5431),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5431),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5431),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5431),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5431),
    [anon_sym_const] = ACTIONS(5431),
    [anon_sym_constexpr] = ACTIONS(5431),
    [anon_sym_volatile] = ACTIONS(5431),
    [anon_sym_restrict] = ACTIONS(5431),
    [anon_sym___restrict__] = ACTIONS(5431),
    [anon_sym__Atomic] = ACTIONS(5431),
    [anon_sym__Noreturn] = ACTIONS(5431),
    [anon_sym__Nonnull] = ACTIONS(5431),
    [anon_sym_nullable] = ACTIONS(5431),
    [anon_sym__Complex] = ACTIONS(5431),
    [anon_sym__Nullable] = ACTIONS(5431),
    [anon_sym__Nullable_result] = ACTIONS(5431),
    [anon_sym__Null_unspecified] = ACTIONS(5431),
    [anon_sym___autoreleasing] = ACTIONS(5431),
    [anon_sym___block] = ACTIONS(5431),
    [anon_sym___bridge] = ACTIONS(5431),
    [anon_sym___bridge_retained] = ACTIONS(5431),
    [anon_sym___bridge_transfer] = ACTIONS(5431),
    [anon_sym___complex] = ACTIONS(5431),
    [anon_sym___const] = ACTIONS(5431),
    [anon_sym___imag] = ACTIONS(5431),
    [anon_sym___kindof] = ACTIONS(5431),
    [anon_sym___nonnull] = ACTIONS(5431),
    [anon_sym___nullable] = ACTIONS(5431),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5431),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5431),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5431),
    [anon_sym___real] = ACTIONS(5431),
    [anon_sym___strong] = ACTIONS(5431),
    [anon_sym___unsafe_unretained] = ACTIONS(5431),
    [anon_sym___unused] = ACTIONS(5431),
    [anon_sym___weak] = ACTIONS(5431),
    [anon_sym_alignas] = ACTIONS(5431),
    [anon_sym__Alignas] = ACTIONS(5431),
    [sym_primitive_type] = ACTIONS(5431),
    [anon_sym_enum] = ACTIONS(5431),
    [anon_sym_struct] = ACTIONS(5431),
    [anon_sym_union] = ACTIONS(5431),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5431),
    [anon_sym___typeof] = ACTIONS(5431),
    [anon_sym_typeof] = ACTIONS(5431),
    [aux_sym_preproc_undef_token1] = ACTIONS(5431),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5431),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5431),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5431),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5431),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5431),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5431),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5431),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5431),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5431),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_API_AVAILABLE] = ACTIONS(5431),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_API_DEPRECATED] = ACTIONS(5431),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5431),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5431),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5431),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5431),
    [anon_sym___deprecated_msg] = ACTIONS(5431),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5431),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5431),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5431),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5431),
    [anon_sym_ATend] = ACTIONS(5433),
    [anon_sym_AToptional] = ACTIONS(5433),
    [anon_sym_ATrequired] = ACTIONS(5433),
    [anon_sym_ATproperty] = ACTIONS(5433),
    [anon_sym_BOOL] = ACTIONS(5431),
    [anon_sym_IMP] = ACTIONS(5431),
    [anon_sym_SEL] = ACTIONS(5431),
    [anon_sym_Class] = ACTIONS(5431),
    [anon_sym_id] = ACTIONS(5431),
  },
  [STATE(2186)] = {
    [sym_identifier] = ACTIONS(5475),
    [aux_sym_preproc_def_token1] = ACTIONS(5475),
    [aux_sym_preproc_if_token1] = ACTIONS(5475),
    [aux_sym_preproc_if_token2] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5475),
    [aux_sym_preproc_else_token1] = ACTIONS(5475),
    [aux_sym_preproc_elif_token1] = ACTIONS(5475),
    [sym_preproc_directive] = ACTIONS(5475),
    [anon_sym_DASH] = ACTIONS(5477),
    [anon_sym_PLUS] = ACTIONS(5477),
    [anon_sym_SEMI] = ACTIONS(5479),
    [anon_sym___extension__] = ACTIONS(5475),
    [anon_sym_typedef] = ACTIONS(5475),
    [anon_sym_extern] = ACTIONS(5475),
    [anon_sym___attribute__] = ACTIONS(5475),
    [anon_sym___attribute] = ACTIONS(5475),
    [anon_sym_noreturn] = ACTIONS(5475),
    [anon_sym_LBRACK] = ACTIONS(5477),
    [anon_sym___declspec] = ACTIONS(5475),
    [anon_sym___cdecl] = ACTIONS(5475),
    [anon_sym___clrcall] = ACTIONS(5475),
    [anon_sym___stdcall] = ACTIONS(5475),
    [anon_sym___fastcall] = ACTIONS(5475),
    [anon_sym___thiscall] = ACTIONS(5475),
    [anon_sym___vectorcall] = ACTIONS(5475),
    [anon_sym_signed] = ACTIONS(5475),
    [anon_sym_unsigned] = ACTIONS(5475),
    [anon_sym_long] = ACTIONS(5475),
    [anon_sym_short] = ACTIONS(5475),
    [anon_sym_static] = ACTIONS(5475),
    [anon_sym_auto] = ACTIONS(5475),
    [anon_sym_register] = ACTIONS(5475),
    [anon_sym_inline] = ACTIONS(5475),
    [anon_sym___inline] = ACTIONS(5475),
    [anon_sym___inline__] = ACTIONS(5475),
    [anon_sym___forceinline] = ACTIONS(5475),
    [anon_sym_thread_local] = ACTIONS(5475),
    [anon_sym___thread] = ACTIONS(5475),
    [anon_sym_CG_EXTERN] = ACTIONS(5475),
    [anon_sym_CG_INLINE] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5475),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5475),
    [anon_sym_IBOutlet] = ACTIONS(5475),
    [anon_sym_IBInspectable] = ACTIONS(5475),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5475),
    [anon_sym_NS_INLINE] = ACTIONS(5475),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5475),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5475),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5475),
    [anon_sym_const] = ACTIONS(5475),
    [anon_sym_constexpr] = ACTIONS(5475),
    [anon_sym_volatile] = ACTIONS(5475),
    [anon_sym_restrict] = ACTIONS(5475),
    [anon_sym___restrict__] = ACTIONS(5475),
    [anon_sym__Atomic] = ACTIONS(5475),
    [anon_sym__Noreturn] = ACTIONS(5475),
    [anon_sym__Nonnull] = ACTIONS(5475),
    [anon_sym_nullable] = ACTIONS(5475),
    [anon_sym__Complex] = ACTIONS(5475),
    [anon_sym__Nullable] = ACTIONS(5475),
    [anon_sym__Nullable_result] = ACTIONS(5475),
    [anon_sym__Null_unspecified] = ACTIONS(5475),
    [anon_sym___autoreleasing] = ACTIONS(5475),
    [anon_sym___block] = ACTIONS(5475),
    [anon_sym___bridge] = ACTIONS(5475),
    [anon_sym___bridge_retained] = ACTIONS(5475),
    [anon_sym___bridge_transfer] = ACTIONS(5475),
    [anon_sym___complex] = ACTIONS(5475),
    [anon_sym___const] = ACTIONS(5475),
    [anon_sym___imag] = ACTIONS(5475),
    [anon_sym___kindof] = ACTIONS(5475),
    [anon_sym___nonnull] = ACTIONS(5475),
    [anon_sym___nullable] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5475),
    [anon_sym___real] = ACTIONS(5475),
    [anon_sym___strong] = ACTIONS(5475),
    [anon_sym___unsafe_unretained] = ACTIONS(5475),
    [anon_sym___unused] = ACTIONS(5475),
    [anon_sym___weak] = ACTIONS(5475),
    [anon_sym_alignas] = ACTIONS(5475),
    [anon_sym__Alignas] = ACTIONS(5475),
    [sym_primitive_type] = ACTIONS(5475),
    [anon_sym_enum] = ACTIONS(5475),
    [anon_sym_struct] = ACTIONS(5475),
    [anon_sym_union] = ACTIONS(5475),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5475),
    [anon_sym___typeof] = ACTIONS(5475),
    [anon_sym_typeof] = ACTIONS(5475),
    [aux_sym_preproc_undef_token1] = ACTIONS(5475),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5475),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5475),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5475),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_API_AVAILABLE] = ACTIONS(5475),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_API_DEPRECATED] = ACTIONS(5475),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym___deprecated_msg] = ACTIONS(5475),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5475),
    [anon_sym_ATsynthesize] = ACTIONS(5477),
    [anon_sym_ATdynamic] = ACTIONS(5477),
    [anon_sym_BOOL] = ACTIONS(5475),
    [anon_sym_IMP] = ACTIONS(5475),
    [anon_sym_SEL] = ACTIONS(5475),
    [anon_sym_Class] = ACTIONS(5475),
    [anon_sym_id] = ACTIONS(5475),
  },
  [STATE(2187)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2138),
    [sym_identifier] = ACTIONS(5454),
    [aux_sym_preproc_def_token1] = ACTIONS(5454),
    [aux_sym_preproc_if_token1] = ACTIONS(5454),
    [aux_sym_preproc_if_token2] = ACTIONS(5454),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5454),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5454),
    [aux_sym_preproc_else_token1] = ACTIONS(5454),
    [aux_sym_preproc_elif_token1] = ACTIONS(5454),
    [sym_preproc_directive] = ACTIONS(5454),
    [anon_sym_DASH] = ACTIONS(5456),
    [anon_sym_PLUS] = ACTIONS(5456),
    [anon_sym_SEMI] = ACTIONS(5357),
    [anon_sym___extension__] = ACTIONS(5454),
    [anon_sym_typedef] = ACTIONS(5454),
    [anon_sym_extern] = ACTIONS(5454),
    [anon_sym___attribute__] = ACTIONS(5454),
    [anon_sym___attribute] = ACTIONS(5454),
    [anon_sym_noreturn] = ACTIONS(5454),
    [anon_sym_LBRACK] = ACTIONS(5456),
    [anon_sym___declspec] = ACTIONS(5454),
    [anon_sym___cdecl] = ACTIONS(5454),
    [anon_sym___clrcall] = ACTIONS(5454),
    [anon_sym___stdcall] = ACTIONS(5454),
    [anon_sym___fastcall] = ACTIONS(5454),
    [anon_sym___thiscall] = ACTIONS(5454),
    [anon_sym___vectorcall] = ACTIONS(5454),
    [anon_sym_signed] = ACTIONS(5454),
    [anon_sym_unsigned] = ACTIONS(5454),
    [anon_sym_long] = ACTIONS(5454),
    [anon_sym_short] = ACTIONS(5454),
    [anon_sym_static] = ACTIONS(5454),
    [anon_sym_auto] = ACTIONS(5454),
    [anon_sym_register] = ACTIONS(5454),
    [anon_sym_inline] = ACTIONS(5454),
    [anon_sym___inline] = ACTIONS(5454),
    [anon_sym___inline__] = ACTIONS(5454),
    [anon_sym___forceinline] = ACTIONS(5454),
    [anon_sym_thread_local] = ACTIONS(5454),
    [anon_sym___thread] = ACTIONS(5454),
    [anon_sym_CG_EXTERN] = ACTIONS(5454),
    [anon_sym_CG_INLINE] = ACTIONS(5454),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5454),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5454),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5454),
    [anon_sym_IBOutlet] = ACTIONS(5454),
    [anon_sym_IBInspectable] = ACTIONS(5454),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5454),
    [anon_sym_NS_INLINE] = ACTIONS(5454),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5454),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5454),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5454),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5454),
    [anon_sym_const] = ACTIONS(5454),
    [anon_sym_constexpr] = ACTIONS(5454),
    [anon_sym_volatile] = ACTIONS(5454),
    [anon_sym_restrict] = ACTIONS(5454),
    [anon_sym___restrict__] = ACTIONS(5454),
    [anon_sym__Atomic] = ACTIONS(5454),
    [anon_sym__Noreturn] = ACTIONS(5454),
    [anon_sym__Nonnull] = ACTIONS(5454),
    [anon_sym_nullable] = ACTIONS(5454),
    [anon_sym__Complex] = ACTIONS(5454),
    [anon_sym__Nullable] = ACTIONS(5454),
    [anon_sym__Nullable_result] = ACTIONS(5454),
    [anon_sym__Null_unspecified] = ACTIONS(5454),
    [anon_sym___autoreleasing] = ACTIONS(5454),
    [anon_sym___block] = ACTIONS(5454),
    [anon_sym___bridge] = ACTIONS(5454),
    [anon_sym___bridge_retained] = ACTIONS(5454),
    [anon_sym___bridge_transfer] = ACTIONS(5454),
    [anon_sym___complex] = ACTIONS(5454),
    [anon_sym___const] = ACTIONS(5454),
    [anon_sym___imag] = ACTIONS(5454),
    [anon_sym___kindof] = ACTIONS(5454),
    [anon_sym___nonnull] = ACTIONS(5454),
    [anon_sym___nullable] = ACTIONS(5454),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5454),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5454),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5454),
    [anon_sym___real] = ACTIONS(5454),
    [anon_sym___strong] = ACTIONS(5454),
    [anon_sym___unsafe_unretained] = ACTIONS(5454),
    [anon_sym___unused] = ACTIONS(5454),
    [anon_sym___weak] = ACTIONS(5454),
    [anon_sym_alignas] = ACTIONS(5454),
    [anon_sym__Alignas] = ACTIONS(5454),
    [sym_primitive_type] = ACTIONS(5454),
    [anon_sym_enum] = ACTIONS(5454),
    [anon_sym_struct] = ACTIONS(5454),
    [anon_sym_union] = ACTIONS(5454),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5454),
    [anon_sym___typeof] = ACTIONS(5454),
    [anon_sym_typeof] = ACTIONS(5454),
    [aux_sym_preproc_undef_token1] = ACTIONS(5454),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5454),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5454),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5454),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5454),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5454),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5454),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5454),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5454),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5454),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_API_AVAILABLE] = ACTIONS(5454),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_API_DEPRECATED] = ACTIONS(5454),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5454),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5454),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5454),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5454),
    [anon_sym___deprecated_msg] = ACTIONS(5454),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5454),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5454),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5454),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5454),
    [anon_sym_ATproperty] = ACTIONS(5456),
    [anon_sym_BOOL] = ACTIONS(5454),
    [anon_sym_IMP] = ACTIONS(5454),
    [anon_sym_SEL] = ACTIONS(5454),
    [anon_sym_Class] = ACTIONS(5454),
    [anon_sym_id] = ACTIONS(5454),
  },
  [STATE(2188)] = {
    [sym_identifier] = ACTIONS(2688),
    [aux_sym_preproc_def_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token2] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2688),
    [aux_sym_preproc_else_token1] = ACTIONS(2688),
    [aux_sym_preproc_elif_token1] = ACTIONS(2688),
    [sym_preproc_directive] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2690),
    [anon_sym_PLUS] = ACTIONS(2690),
    [anon_sym___extension__] = ACTIONS(2688),
    [anon_sym_typedef] = ACTIONS(2688),
    [anon_sym_extern] = ACTIONS(2688),
    [anon_sym___attribute__] = ACTIONS(2688),
    [anon_sym___attribute] = ACTIONS(2688),
    [anon_sym_noreturn] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym___declspec] = ACTIONS(2688),
    [anon_sym___cdecl] = ACTIONS(2688),
    [anon_sym___clrcall] = ACTIONS(2688),
    [anon_sym___stdcall] = ACTIONS(2688),
    [anon_sym___fastcall] = ACTIONS(2688),
    [anon_sym___thiscall] = ACTIONS(2688),
    [anon_sym___vectorcall] = ACTIONS(2688),
    [anon_sym_signed] = ACTIONS(2688),
    [anon_sym_unsigned] = ACTIONS(2688),
    [anon_sym_long] = ACTIONS(2688),
    [anon_sym_short] = ACTIONS(2688),
    [anon_sym_static] = ACTIONS(2688),
    [anon_sym_auto] = ACTIONS(2688),
    [anon_sym_register] = ACTIONS(2688),
    [anon_sym_inline] = ACTIONS(2688),
    [anon_sym___inline] = ACTIONS(2688),
    [anon_sym___inline__] = ACTIONS(2688),
    [anon_sym___forceinline] = ACTIONS(2688),
    [anon_sym_thread_local] = ACTIONS(2688),
    [anon_sym___thread] = ACTIONS(2688),
    [anon_sym_CG_EXTERN] = ACTIONS(2688),
    [anon_sym_CG_INLINE] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2688),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2688),
    [anon_sym_IBOutlet] = ACTIONS(2688),
    [anon_sym_IBInspectable] = ACTIONS(2688),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2688),
    [anon_sym_NS_INLINE] = ACTIONS(2688),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2688),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2688),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2688),
    [anon_sym_const] = ACTIONS(2688),
    [anon_sym_constexpr] = ACTIONS(2688),
    [anon_sym_volatile] = ACTIONS(2688),
    [anon_sym_restrict] = ACTIONS(2688),
    [anon_sym___restrict__] = ACTIONS(2688),
    [anon_sym__Atomic] = ACTIONS(2688),
    [anon_sym__Noreturn] = ACTIONS(2688),
    [anon_sym__Nonnull] = ACTIONS(2688),
    [anon_sym_nullable] = ACTIONS(2688),
    [anon_sym__Complex] = ACTIONS(2688),
    [anon_sym__Nullable] = ACTIONS(2688),
    [anon_sym__Nullable_result] = ACTIONS(2688),
    [anon_sym__Null_unspecified] = ACTIONS(2688),
    [anon_sym___autoreleasing] = ACTIONS(2688),
    [anon_sym___block] = ACTIONS(2688),
    [anon_sym___bridge] = ACTIONS(2688),
    [anon_sym___bridge_retained] = ACTIONS(2688),
    [anon_sym___bridge_transfer] = ACTIONS(2688),
    [anon_sym___complex] = ACTIONS(2688),
    [anon_sym___const] = ACTIONS(2688),
    [anon_sym___imag] = ACTIONS(2688),
    [anon_sym___kindof] = ACTIONS(2688),
    [anon_sym___nonnull] = ACTIONS(2688),
    [anon_sym___nullable] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2688),
    [anon_sym___real] = ACTIONS(2688),
    [anon_sym___strong] = ACTIONS(2688),
    [anon_sym___unsafe_unretained] = ACTIONS(2688),
    [anon_sym___unused] = ACTIONS(2688),
    [anon_sym___weak] = ACTIONS(2688),
    [anon_sym_alignas] = ACTIONS(2688),
    [anon_sym__Alignas] = ACTIONS(2688),
    [sym_primitive_type] = ACTIONS(2688),
    [anon_sym_enum] = ACTIONS(2688),
    [anon_sym_struct] = ACTIONS(2688),
    [anon_sym_union] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2688),
    [anon_sym___typeof] = ACTIONS(2688),
    [anon_sym_typeof] = ACTIONS(2688),
    [aux_sym_preproc_undef_token1] = ACTIONS(2688),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2688),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2688),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2688),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_API_AVAILABLE] = ACTIONS(2688),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_API_DEPRECATED] = ACTIONS(2688),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym___deprecated_msg] = ACTIONS(2688),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2688),
    [anon_sym_ATsynthesize] = ACTIONS(2690),
    [anon_sym_ATdynamic] = ACTIONS(2690),
    [anon_sym_ATproperty] = ACTIONS(2690),
    [anon_sym_BOOL] = ACTIONS(2688),
    [anon_sym_IMP] = ACTIONS(2688),
    [anon_sym_SEL] = ACTIONS(2688),
    [anon_sym_Class] = ACTIONS(2688),
    [anon_sym_id] = ACTIONS(2688),
  },
  [STATE(2189)] = {
    [sym_identifier] = ACTIONS(5481),
    [aux_sym_preproc_def_token1] = ACTIONS(5481),
    [aux_sym_preproc_if_token1] = ACTIONS(5481),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5481),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5481),
    [sym_preproc_directive] = ACTIONS(5481),
    [anon_sym_LPAREN2] = ACTIONS(5328),
    [anon_sym_DASH] = ACTIONS(5483),
    [anon_sym_PLUS] = ACTIONS(5483),
    [anon_sym_LT] = ACTIONS(5483),
    [anon_sym___extension__] = ACTIONS(5481),
    [anon_sym_typedef] = ACTIONS(5481),
    [anon_sym_extern] = ACTIONS(5481),
    [anon_sym___attribute__] = ACTIONS(5481),
    [anon_sym___attribute] = ACTIONS(5481),
    [anon_sym_noreturn] = ACTIONS(5481),
    [anon_sym_LBRACK] = ACTIONS(5483),
    [anon_sym___declspec] = ACTIONS(5481),
    [anon_sym___cdecl] = ACTIONS(5481),
    [anon_sym___clrcall] = ACTIONS(5481),
    [anon_sym___stdcall] = ACTIONS(5481),
    [anon_sym___fastcall] = ACTIONS(5481),
    [anon_sym___thiscall] = ACTIONS(5481),
    [anon_sym___vectorcall] = ACTIONS(5481),
    [anon_sym_LBRACE] = ACTIONS(5483),
    [anon_sym_signed] = ACTIONS(5481),
    [anon_sym_unsigned] = ACTIONS(5481),
    [anon_sym_long] = ACTIONS(5481),
    [anon_sym_short] = ACTIONS(5481),
    [anon_sym_static] = ACTIONS(5481),
    [anon_sym_auto] = ACTIONS(5481),
    [anon_sym_register] = ACTIONS(5481),
    [anon_sym_inline] = ACTIONS(5481),
    [anon_sym___inline] = ACTIONS(5481),
    [anon_sym___inline__] = ACTIONS(5481),
    [anon_sym___forceinline] = ACTIONS(5481),
    [anon_sym_thread_local] = ACTIONS(5481),
    [anon_sym___thread] = ACTIONS(5481),
    [anon_sym_CG_EXTERN] = ACTIONS(5481),
    [anon_sym_CG_INLINE] = ACTIONS(5481),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5481),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5481),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5481),
    [anon_sym_IBOutlet] = ACTIONS(5481),
    [anon_sym_IBInspectable] = ACTIONS(5481),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5481),
    [anon_sym_NS_INLINE] = ACTIONS(5481),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5481),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5481),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5481),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5481),
    [anon_sym_const] = ACTIONS(5481),
    [anon_sym_constexpr] = ACTIONS(5481),
    [anon_sym_volatile] = ACTIONS(5481),
    [anon_sym_restrict] = ACTIONS(5481),
    [anon_sym___restrict__] = ACTIONS(5481),
    [anon_sym__Atomic] = ACTIONS(5481),
    [anon_sym__Noreturn] = ACTIONS(5481),
    [anon_sym__Nonnull] = ACTIONS(5481),
    [anon_sym_nullable] = ACTIONS(5481),
    [anon_sym__Complex] = ACTIONS(5481),
    [anon_sym__Nullable] = ACTIONS(5481),
    [anon_sym__Nullable_result] = ACTIONS(5481),
    [anon_sym__Null_unspecified] = ACTIONS(5481),
    [anon_sym___autoreleasing] = ACTIONS(5481),
    [anon_sym___block] = ACTIONS(5481),
    [anon_sym___bridge] = ACTIONS(5481),
    [anon_sym___bridge_retained] = ACTIONS(5481),
    [anon_sym___bridge_transfer] = ACTIONS(5481),
    [anon_sym___complex] = ACTIONS(5481),
    [anon_sym___const] = ACTIONS(5481),
    [anon_sym___imag] = ACTIONS(5481),
    [anon_sym___kindof] = ACTIONS(5481),
    [anon_sym___nonnull] = ACTIONS(5481),
    [anon_sym___nullable] = ACTIONS(5481),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5481),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5481),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5481),
    [anon_sym___real] = ACTIONS(5481),
    [anon_sym___strong] = ACTIONS(5481),
    [anon_sym___unsafe_unretained] = ACTIONS(5481),
    [anon_sym___unused] = ACTIONS(5481),
    [anon_sym___weak] = ACTIONS(5481),
    [anon_sym_alignas] = ACTIONS(5481),
    [anon_sym__Alignas] = ACTIONS(5481),
    [sym_primitive_type] = ACTIONS(5481),
    [anon_sym_enum] = ACTIONS(5481),
    [anon_sym_COLON] = ACTIONS(5328),
    [anon_sym_struct] = ACTIONS(5481),
    [anon_sym_union] = ACTIONS(5481),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5481),
    [anon_sym___typeof] = ACTIONS(5481),
    [anon_sym_typeof] = ACTIONS(5481),
    [aux_sym_preproc_undef_token1] = ACTIONS(5481),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5481),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5481),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5481),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5481),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5481),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5481),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5481),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5481),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5481),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5481),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5481),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5481),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5481),
    [anon_sym_API_AVAILABLE] = ACTIONS(5481),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5481),
    [anon_sym_API_DEPRECATED] = ACTIONS(5481),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5481),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5481),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5481),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5481),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5481),
    [anon_sym___deprecated_msg] = ACTIONS(5481),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5481),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5481),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5481),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5481),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5481),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5481),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5481),
    [anon_sym_ATend] = ACTIONS(5483),
    [anon_sym_ATproperty] = ACTIONS(5483),
    [anon_sym_BOOL] = ACTIONS(5481),
    [anon_sym_IMP] = ACTIONS(5481),
    [anon_sym_SEL] = ACTIONS(5481),
    [anon_sym_Class] = ACTIONS(5481),
    [anon_sym_id] = ACTIONS(5481),
  },
  [STATE(2190)] = {
    [sym_identifier] = ACTIONS(2584),
    [aux_sym_preproc_def_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token2] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2584),
    [aux_sym_preproc_else_token1] = ACTIONS(2584),
    [aux_sym_preproc_elif_token1] = ACTIONS(2584),
    [sym_preproc_directive] = ACTIONS(2584),
    [anon_sym_DASH] = ACTIONS(2586),
    [anon_sym_PLUS] = ACTIONS(2586),
    [anon_sym___extension__] = ACTIONS(2584),
    [anon_sym_typedef] = ACTIONS(2584),
    [anon_sym_extern] = ACTIONS(2584),
    [anon_sym___attribute__] = ACTIONS(2584),
    [anon_sym___attribute] = ACTIONS(2584),
    [anon_sym_noreturn] = ACTIONS(2584),
    [anon_sym_LBRACK] = ACTIONS(2586),
    [anon_sym___declspec] = ACTIONS(2584),
    [anon_sym___cdecl] = ACTIONS(2584),
    [anon_sym___clrcall] = ACTIONS(2584),
    [anon_sym___stdcall] = ACTIONS(2584),
    [anon_sym___fastcall] = ACTIONS(2584),
    [anon_sym___thiscall] = ACTIONS(2584),
    [anon_sym___vectorcall] = ACTIONS(2584),
    [anon_sym_signed] = ACTIONS(2584),
    [anon_sym_unsigned] = ACTIONS(2584),
    [anon_sym_long] = ACTIONS(2584),
    [anon_sym_short] = ACTIONS(2584),
    [anon_sym_static] = ACTIONS(2584),
    [anon_sym_auto] = ACTIONS(2584),
    [anon_sym_register] = ACTIONS(2584),
    [anon_sym_inline] = ACTIONS(2584),
    [anon_sym___inline] = ACTIONS(2584),
    [anon_sym___inline__] = ACTIONS(2584),
    [anon_sym___forceinline] = ACTIONS(2584),
    [anon_sym_thread_local] = ACTIONS(2584),
    [anon_sym___thread] = ACTIONS(2584),
    [anon_sym_CG_EXTERN] = ACTIONS(2584),
    [anon_sym_CG_INLINE] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2584),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2584),
    [anon_sym_IBOutlet] = ACTIONS(2584),
    [anon_sym_IBInspectable] = ACTIONS(2584),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2584),
    [anon_sym_NS_INLINE] = ACTIONS(2584),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2584),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2584),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2584),
    [anon_sym_const] = ACTIONS(2584),
    [anon_sym_constexpr] = ACTIONS(2584),
    [anon_sym_volatile] = ACTIONS(2584),
    [anon_sym_restrict] = ACTIONS(2584),
    [anon_sym___restrict__] = ACTIONS(2584),
    [anon_sym__Atomic] = ACTIONS(2584),
    [anon_sym__Noreturn] = ACTIONS(2584),
    [anon_sym__Nonnull] = ACTIONS(2584),
    [anon_sym_nullable] = ACTIONS(2584),
    [anon_sym__Complex] = ACTIONS(2584),
    [anon_sym__Nullable] = ACTIONS(2584),
    [anon_sym__Nullable_result] = ACTIONS(2584),
    [anon_sym__Null_unspecified] = ACTIONS(2584),
    [anon_sym___autoreleasing] = ACTIONS(2584),
    [anon_sym___block] = ACTIONS(2584),
    [anon_sym___bridge] = ACTIONS(2584),
    [anon_sym___bridge_retained] = ACTIONS(2584),
    [anon_sym___bridge_transfer] = ACTIONS(2584),
    [anon_sym___complex] = ACTIONS(2584),
    [anon_sym___const] = ACTIONS(2584),
    [anon_sym___imag] = ACTIONS(2584),
    [anon_sym___kindof] = ACTIONS(2584),
    [anon_sym___nonnull] = ACTIONS(2584),
    [anon_sym___nullable] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2584),
    [anon_sym___real] = ACTIONS(2584),
    [anon_sym___strong] = ACTIONS(2584),
    [anon_sym___unsafe_unretained] = ACTIONS(2584),
    [anon_sym___unused] = ACTIONS(2584),
    [anon_sym___weak] = ACTIONS(2584),
    [anon_sym_alignas] = ACTIONS(2584),
    [anon_sym__Alignas] = ACTIONS(2584),
    [sym_primitive_type] = ACTIONS(2584),
    [anon_sym_enum] = ACTIONS(2584),
    [anon_sym_struct] = ACTIONS(2584),
    [anon_sym_union] = ACTIONS(2584),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2584),
    [anon_sym___typeof] = ACTIONS(2584),
    [anon_sym_typeof] = ACTIONS(2584),
    [aux_sym_preproc_undef_token1] = ACTIONS(2584),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2584),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2584),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2584),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_API_AVAILABLE] = ACTIONS(2584),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_API_DEPRECATED] = ACTIONS(2584),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym___deprecated_msg] = ACTIONS(2584),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2584),
    [anon_sym_ATsynthesize] = ACTIONS(2586),
    [anon_sym_ATdynamic] = ACTIONS(2586),
    [anon_sym_ATproperty] = ACTIONS(2586),
    [anon_sym_BOOL] = ACTIONS(2584),
    [anon_sym_IMP] = ACTIONS(2584),
    [anon_sym_SEL] = ACTIONS(2584),
    [anon_sym_Class] = ACTIONS(2584),
    [anon_sym_id] = ACTIONS(2584),
  },
  [STATE(2191)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2185),
    [sym_identifier] = ACTIONS(5462),
    [aux_sym_preproc_def_token1] = ACTIONS(5462),
    [aux_sym_preproc_if_token1] = ACTIONS(5462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5462),
    [sym_preproc_directive] = ACTIONS(5462),
    [anon_sym_DASH] = ACTIONS(5464),
    [anon_sym_PLUS] = ACTIONS(5464),
    [anon_sym_SEMI] = ACTIONS(5349),
    [anon_sym___extension__] = ACTIONS(5462),
    [anon_sym_typedef] = ACTIONS(5462),
    [anon_sym_extern] = ACTIONS(5462),
    [anon_sym___attribute__] = ACTIONS(5462),
    [anon_sym___attribute] = ACTIONS(5462),
    [anon_sym_noreturn] = ACTIONS(5462),
    [anon_sym_LBRACK] = ACTIONS(5464),
    [anon_sym___declspec] = ACTIONS(5462),
    [anon_sym___cdecl] = ACTIONS(5462),
    [anon_sym___clrcall] = ACTIONS(5462),
    [anon_sym___stdcall] = ACTIONS(5462),
    [anon_sym___fastcall] = ACTIONS(5462),
    [anon_sym___thiscall] = ACTIONS(5462),
    [anon_sym___vectorcall] = ACTIONS(5462),
    [anon_sym_signed] = ACTIONS(5462),
    [anon_sym_unsigned] = ACTIONS(5462),
    [anon_sym_long] = ACTIONS(5462),
    [anon_sym_short] = ACTIONS(5462),
    [anon_sym_static] = ACTIONS(5462),
    [anon_sym_auto] = ACTIONS(5462),
    [anon_sym_register] = ACTIONS(5462),
    [anon_sym_inline] = ACTIONS(5462),
    [anon_sym___inline] = ACTIONS(5462),
    [anon_sym___inline__] = ACTIONS(5462),
    [anon_sym___forceinline] = ACTIONS(5462),
    [anon_sym_thread_local] = ACTIONS(5462),
    [anon_sym___thread] = ACTIONS(5462),
    [anon_sym_CG_EXTERN] = ACTIONS(5462),
    [anon_sym_CG_INLINE] = ACTIONS(5462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5462),
    [anon_sym_IBOutlet] = ACTIONS(5462),
    [anon_sym_IBInspectable] = ACTIONS(5462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5462),
    [anon_sym_NS_INLINE] = ACTIONS(5462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5462),
    [anon_sym_const] = ACTIONS(5462),
    [anon_sym_constexpr] = ACTIONS(5462),
    [anon_sym_volatile] = ACTIONS(5462),
    [anon_sym_restrict] = ACTIONS(5462),
    [anon_sym___restrict__] = ACTIONS(5462),
    [anon_sym__Atomic] = ACTIONS(5462),
    [anon_sym__Noreturn] = ACTIONS(5462),
    [anon_sym__Nonnull] = ACTIONS(5462),
    [anon_sym_nullable] = ACTIONS(5462),
    [anon_sym__Complex] = ACTIONS(5462),
    [anon_sym__Nullable] = ACTIONS(5462),
    [anon_sym__Nullable_result] = ACTIONS(5462),
    [anon_sym__Null_unspecified] = ACTIONS(5462),
    [anon_sym___autoreleasing] = ACTIONS(5462),
    [anon_sym___block] = ACTIONS(5462),
    [anon_sym___bridge] = ACTIONS(5462),
    [anon_sym___bridge_retained] = ACTIONS(5462),
    [anon_sym___bridge_transfer] = ACTIONS(5462),
    [anon_sym___complex] = ACTIONS(5462),
    [anon_sym___const] = ACTIONS(5462),
    [anon_sym___imag] = ACTIONS(5462),
    [anon_sym___kindof] = ACTIONS(5462),
    [anon_sym___nonnull] = ACTIONS(5462),
    [anon_sym___nullable] = ACTIONS(5462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5462),
    [anon_sym___real] = ACTIONS(5462),
    [anon_sym___strong] = ACTIONS(5462),
    [anon_sym___unsafe_unretained] = ACTIONS(5462),
    [anon_sym___unused] = ACTIONS(5462),
    [anon_sym___weak] = ACTIONS(5462),
    [anon_sym_alignas] = ACTIONS(5462),
    [anon_sym__Alignas] = ACTIONS(5462),
    [sym_primitive_type] = ACTIONS(5462),
    [anon_sym_enum] = ACTIONS(5462),
    [anon_sym_struct] = ACTIONS(5462),
    [anon_sym_union] = ACTIONS(5462),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5462),
    [anon_sym___typeof] = ACTIONS(5462),
    [anon_sym_typeof] = ACTIONS(5462),
    [aux_sym_preproc_undef_token1] = ACTIONS(5462),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5462),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_API_AVAILABLE] = ACTIONS(5462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_API_DEPRECATED] = ACTIONS(5462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5462),
    [anon_sym___deprecated_msg] = ACTIONS(5462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5462),
    [anon_sym_ATend] = ACTIONS(5464),
    [anon_sym_AToptional] = ACTIONS(5464),
    [anon_sym_ATrequired] = ACTIONS(5464),
    [anon_sym_ATproperty] = ACTIONS(5464),
    [anon_sym_BOOL] = ACTIONS(5462),
    [anon_sym_IMP] = ACTIONS(5462),
    [anon_sym_SEL] = ACTIONS(5462),
    [anon_sym_Class] = ACTIONS(5462),
    [anon_sym_id] = ACTIONS(5462),
  },
  [STATE(2192)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2185),
    [sym_identifier] = ACTIONS(5458),
    [aux_sym_preproc_def_token1] = ACTIONS(5458),
    [aux_sym_preproc_if_token1] = ACTIONS(5458),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5458),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5458),
    [sym_preproc_directive] = ACTIONS(5458),
    [anon_sym_DASH] = ACTIONS(5460),
    [anon_sym_PLUS] = ACTIONS(5460),
    [anon_sym_SEMI] = ACTIONS(5349),
    [anon_sym___extension__] = ACTIONS(5458),
    [anon_sym_typedef] = ACTIONS(5458),
    [anon_sym_extern] = ACTIONS(5458),
    [anon_sym___attribute__] = ACTIONS(5458),
    [anon_sym___attribute] = ACTIONS(5458),
    [anon_sym_noreturn] = ACTIONS(5458),
    [anon_sym_LBRACK] = ACTIONS(5460),
    [anon_sym___declspec] = ACTIONS(5458),
    [anon_sym___cdecl] = ACTIONS(5458),
    [anon_sym___clrcall] = ACTIONS(5458),
    [anon_sym___stdcall] = ACTIONS(5458),
    [anon_sym___fastcall] = ACTIONS(5458),
    [anon_sym___thiscall] = ACTIONS(5458),
    [anon_sym___vectorcall] = ACTIONS(5458),
    [anon_sym_signed] = ACTIONS(5458),
    [anon_sym_unsigned] = ACTIONS(5458),
    [anon_sym_long] = ACTIONS(5458),
    [anon_sym_short] = ACTIONS(5458),
    [anon_sym_static] = ACTIONS(5458),
    [anon_sym_auto] = ACTIONS(5458),
    [anon_sym_register] = ACTIONS(5458),
    [anon_sym_inline] = ACTIONS(5458),
    [anon_sym___inline] = ACTIONS(5458),
    [anon_sym___inline__] = ACTIONS(5458),
    [anon_sym___forceinline] = ACTIONS(5458),
    [anon_sym_thread_local] = ACTIONS(5458),
    [anon_sym___thread] = ACTIONS(5458),
    [anon_sym_CG_EXTERN] = ACTIONS(5458),
    [anon_sym_CG_INLINE] = ACTIONS(5458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5458),
    [anon_sym_IBOutlet] = ACTIONS(5458),
    [anon_sym_IBInspectable] = ACTIONS(5458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5458),
    [anon_sym_NS_INLINE] = ACTIONS(5458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5458),
    [anon_sym_const] = ACTIONS(5458),
    [anon_sym_constexpr] = ACTIONS(5458),
    [anon_sym_volatile] = ACTIONS(5458),
    [anon_sym_restrict] = ACTIONS(5458),
    [anon_sym___restrict__] = ACTIONS(5458),
    [anon_sym__Atomic] = ACTIONS(5458),
    [anon_sym__Noreturn] = ACTIONS(5458),
    [anon_sym__Nonnull] = ACTIONS(5458),
    [anon_sym_nullable] = ACTIONS(5458),
    [anon_sym__Complex] = ACTIONS(5458),
    [anon_sym__Nullable] = ACTIONS(5458),
    [anon_sym__Nullable_result] = ACTIONS(5458),
    [anon_sym__Null_unspecified] = ACTIONS(5458),
    [anon_sym___autoreleasing] = ACTIONS(5458),
    [anon_sym___block] = ACTIONS(5458),
    [anon_sym___bridge] = ACTIONS(5458),
    [anon_sym___bridge_retained] = ACTIONS(5458),
    [anon_sym___bridge_transfer] = ACTIONS(5458),
    [anon_sym___complex] = ACTIONS(5458),
    [anon_sym___const] = ACTIONS(5458),
    [anon_sym___imag] = ACTIONS(5458),
    [anon_sym___kindof] = ACTIONS(5458),
    [anon_sym___nonnull] = ACTIONS(5458),
    [anon_sym___nullable] = ACTIONS(5458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5458),
    [anon_sym___real] = ACTIONS(5458),
    [anon_sym___strong] = ACTIONS(5458),
    [anon_sym___unsafe_unretained] = ACTIONS(5458),
    [anon_sym___unused] = ACTIONS(5458),
    [anon_sym___weak] = ACTIONS(5458),
    [anon_sym_alignas] = ACTIONS(5458),
    [anon_sym__Alignas] = ACTIONS(5458),
    [sym_primitive_type] = ACTIONS(5458),
    [anon_sym_enum] = ACTIONS(5458),
    [anon_sym_struct] = ACTIONS(5458),
    [anon_sym_union] = ACTIONS(5458),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5458),
    [anon_sym___typeof] = ACTIONS(5458),
    [anon_sym_typeof] = ACTIONS(5458),
    [aux_sym_preproc_undef_token1] = ACTIONS(5458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_API_AVAILABLE] = ACTIONS(5458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_API_DEPRECATED] = ACTIONS(5458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5458),
    [anon_sym___deprecated_msg] = ACTIONS(5458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5458),
    [anon_sym_ATend] = ACTIONS(5460),
    [anon_sym_AToptional] = ACTIONS(5460),
    [anon_sym_ATrequired] = ACTIONS(5460),
    [anon_sym_ATproperty] = ACTIONS(5460),
    [anon_sym_BOOL] = ACTIONS(5458),
    [anon_sym_IMP] = ACTIONS(5458),
    [anon_sym_SEL] = ACTIONS(5458),
    [anon_sym_Class] = ACTIONS(5458),
    [anon_sym_id] = ACTIONS(5458),
  },
  [STATE(2193)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token2] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [aux_sym_preproc_else_token1] = ACTIONS(2446),
    [aux_sym_preproc_elif_token1] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_DASH] = ACTIONS(2448),
    [anon_sym_PLUS] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATproperty] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(2194)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token2] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [aux_sym_preproc_else_token1] = ACTIONS(2438),
    [aux_sym_preproc_elif_token1] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_DASH] = ACTIONS(2440),
    [anon_sym_PLUS] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATproperty] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(2195)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token2] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [aux_sym_preproc_else_token1] = ACTIONS(2442),
    [aux_sym_preproc_elif_token1] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_DASH] = ACTIONS(2444),
    [anon_sym_PLUS] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATproperty] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(2196)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_DASH] = ACTIONS(2512),
    [anon_sym_PLUS] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATend] = ACTIONS(2512),
    [anon_sym_AToptional] = ACTIONS(2512),
    [anon_sym_ATrequired] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATproperty] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(2197)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_DASH] = ACTIONS(2512),
    [anon_sym_PLUS] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATend] = ACTIONS(2512),
    [anon_sym_AToptional] = ACTIONS(2512),
    [anon_sym_ATrequired] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATproperty] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(2198)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_DASH] = ACTIONS(2512),
    [anon_sym_PLUS] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATend] = ACTIONS(2512),
    [anon_sym_AToptional] = ACTIONS(2512),
    [anon_sym_ATrequired] = ACTIONS(2512),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATproperty] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(2199)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token2] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [aux_sym_preproc_else_token1] = ACTIONS(2446),
    [aux_sym_preproc_elif_token1] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_DASH] = ACTIONS(2448),
    [anon_sym_PLUS] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATproperty] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(2200)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2138),
    [sym_identifier] = ACTIONS(5411),
    [aux_sym_preproc_def_token1] = ACTIONS(5411),
    [aux_sym_preproc_if_token1] = ACTIONS(5411),
    [aux_sym_preproc_if_token2] = ACTIONS(5411),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5411),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5411),
    [aux_sym_preproc_else_token1] = ACTIONS(5411),
    [aux_sym_preproc_elif_token1] = ACTIONS(5411),
    [sym_preproc_directive] = ACTIONS(5411),
    [anon_sym_DASH] = ACTIONS(5413),
    [anon_sym_PLUS] = ACTIONS(5413),
    [anon_sym_SEMI] = ACTIONS(5357),
    [anon_sym___extension__] = ACTIONS(5411),
    [anon_sym_typedef] = ACTIONS(5411),
    [anon_sym_extern] = ACTIONS(5411),
    [anon_sym___attribute__] = ACTIONS(5411),
    [anon_sym___attribute] = ACTIONS(5411),
    [anon_sym_noreturn] = ACTIONS(5411),
    [anon_sym_LBRACK] = ACTIONS(5413),
    [anon_sym___declspec] = ACTIONS(5411),
    [anon_sym___cdecl] = ACTIONS(5411),
    [anon_sym___clrcall] = ACTIONS(5411),
    [anon_sym___stdcall] = ACTIONS(5411),
    [anon_sym___fastcall] = ACTIONS(5411),
    [anon_sym___thiscall] = ACTIONS(5411),
    [anon_sym___vectorcall] = ACTIONS(5411),
    [anon_sym_signed] = ACTIONS(5411),
    [anon_sym_unsigned] = ACTIONS(5411),
    [anon_sym_long] = ACTIONS(5411),
    [anon_sym_short] = ACTIONS(5411),
    [anon_sym_static] = ACTIONS(5411),
    [anon_sym_auto] = ACTIONS(5411),
    [anon_sym_register] = ACTIONS(5411),
    [anon_sym_inline] = ACTIONS(5411),
    [anon_sym___inline] = ACTIONS(5411),
    [anon_sym___inline__] = ACTIONS(5411),
    [anon_sym___forceinline] = ACTIONS(5411),
    [anon_sym_thread_local] = ACTIONS(5411),
    [anon_sym___thread] = ACTIONS(5411),
    [anon_sym_CG_EXTERN] = ACTIONS(5411),
    [anon_sym_CG_INLINE] = ACTIONS(5411),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5411),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5411),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5411),
    [anon_sym_IBOutlet] = ACTIONS(5411),
    [anon_sym_IBInspectable] = ACTIONS(5411),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5411),
    [anon_sym_NS_INLINE] = ACTIONS(5411),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5411),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5411),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5411),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5411),
    [anon_sym_const] = ACTIONS(5411),
    [anon_sym_constexpr] = ACTIONS(5411),
    [anon_sym_volatile] = ACTIONS(5411),
    [anon_sym_restrict] = ACTIONS(5411),
    [anon_sym___restrict__] = ACTIONS(5411),
    [anon_sym__Atomic] = ACTIONS(5411),
    [anon_sym__Noreturn] = ACTIONS(5411),
    [anon_sym__Nonnull] = ACTIONS(5411),
    [anon_sym_nullable] = ACTIONS(5411),
    [anon_sym__Complex] = ACTIONS(5411),
    [anon_sym__Nullable] = ACTIONS(5411),
    [anon_sym__Nullable_result] = ACTIONS(5411),
    [anon_sym__Null_unspecified] = ACTIONS(5411),
    [anon_sym___autoreleasing] = ACTIONS(5411),
    [anon_sym___block] = ACTIONS(5411),
    [anon_sym___bridge] = ACTIONS(5411),
    [anon_sym___bridge_retained] = ACTIONS(5411),
    [anon_sym___bridge_transfer] = ACTIONS(5411),
    [anon_sym___complex] = ACTIONS(5411),
    [anon_sym___const] = ACTIONS(5411),
    [anon_sym___imag] = ACTIONS(5411),
    [anon_sym___kindof] = ACTIONS(5411),
    [anon_sym___nonnull] = ACTIONS(5411),
    [anon_sym___nullable] = ACTIONS(5411),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5411),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5411),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5411),
    [anon_sym___real] = ACTIONS(5411),
    [anon_sym___strong] = ACTIONS(5411),
    [anon_sym___unsafe_unretained] = ACTIONS(5411),
    [anon_sym___unused] = ACTIONS(5411),
    [anon_sym___weak] = ACTIONS(5411),
    [anon_sym_alignas] = ACTIONS(5411),
    [anon_sym__Alignas] = ACTIONS(5411),
    [sym_primitive_type] = ACTIONS(5411),
    [anon_sym_enum] = ACTIONS(5411),
    [anon_sym_struct] = ACTIONS(5411),
    [anon_sym_union] = ACTIONS(5411),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5411),
    [anon_sym___typeof] = ACTIONS(5411),
    [anon_sym_typeof] = ACTIONS(5411),
    [aux_sym_preproc_undef_token1] = ACTIONS(5411),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5411),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5411),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5411),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5411),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5411),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5411),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5411),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5411),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5411),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_API_AVAILABLE] = ACTIONS(5411),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_API_DEPRECATED] = ACTIONS(5411),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5411),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5411),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5411),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5411),
    [anon_sym___deprecated_msg] = ACTIONS(5411),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5411),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5411),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5411),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5411),
    [anon_sym_ATproperty] = ACTIONS(5413),
    [anon_sym_BOOL] = ACTIONS(5411),
    [anon_sym_IMP] = ACTIONS(5411),
    [anon_sym_SEL] = ACTIONS(5411),
    [anon_sym_Class] = ACTIONS(5411),
    [anon_sym_id] = ACTIONS(5411),
  },
  [STATE(2201)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_DASH] = ACTIONS(2430),
    [anon_sym_PLUS] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATend] = ACTIONS(2430),
    [anon_sym_AToptional] = ACTIONS(2430),
    [anon_sym_ATrequired] = ACTIONS(2430),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATproperty] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(2202)] = {
    [sym_identifier] = ACTIONS(2618),
    [aux_sym_preproc_def_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2618),
    [sym_preproc_directive] = ACTIONS(2618),
    [anon_sym_DASH] = ACTIONS(2620),
    [anon_sym_PLUS] = ACTIONS(2620),
    [anon_sym___extension__] = ACTIONS(2618),
    [anon_sym_typedef] = ACTIONS(2618),
    [anon_sym_extern] = ACTIONS(2618),
    [anon_sym___attribute__] = ACTIONS(2618),
    [anon_sym___attribute] = ACTIONS(2618),
    [anon_sym_noreturn] = ACTIONS(2618),
    [anon_sym_LBRACK] = ACTIONS(2620),
    [anon_sym___declspec] = ACTIONS(2618),
    [anon_sym___cdecl] = ACTIONS(2618),
    [anon_sym___clrcall] = ACTIONS(2618),
    [anon_sym___stdcall] = ACTIONS(2618),
    [anon_sym___fastcall] = ACTIONS(2618),
    [anon_sym___thiscall] = ACTIONS(2618),
    [anon_sym___vectorcall] = ACTIONS(2618),
    [anon_sym_RBRACE] = ACTIONS(2620),
    [anon_sym_signed] = ACTIONS(2618),
    [anon_sym_unsigned] = ACTIONS(2618),
    [anon_sym_long] = ACTIONS(2618),
    [anon_sym_short] = ACTIONS(2618),
    [anon_sym_static] = ACTIONS(2618),
    [anon_sym_auto] = ACTIONS(2618),
    [anon_sym_register] = ACTIONS(2618),
    [anon_sym_inline] = ACTIONS(2618),
    [anon_sym___inline] = ACTIONS(2618),
    [anon_sym___inline__] = ACTIONS(2618),
    [anon_sym___forceinline] = ACTIONS(2618),
    [anon_sym_thread_local] = ACTIONS(2618),
    [anon_sym___thread] = ACTIONS(2618),
    [anon_sym_CG_EXTERN] = ACTIONS(2618),
    [anon_sym_CG_INLINE] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2618),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2618),
    [anon_sym_IBOutlet] = ACTIONS(2618),
    [anon_sym_IBInspectable] = ACTIONS(2618),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2618),
    [anon_sym_NS_INLINE] = ACTIONS(2618),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2618),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2618),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2618),
    [anon_sym_const] = ACTIONS(2618),
    [anon_sym_constexpr] = ACTIONS(2618),
    [anon_sym_volatile] = ACTIONS(2618),
    [anon_sym_restrict] = ACTIONS(2618),
    [anon_sym___restrict__] = ACTIONS(2618),
    [anon_sym__Atomic] = ACTIONS(2618),
    [anon_sym__Noreturn] = ACTIONS(2618),
    [anon_sym__Nonnull] = ACTIONS(2618),
    [anon_sym_nullable] = ACTIONS(2618),
    [anon_sym__Complex] = ACTIONS(2618),
    [anon_sym__Nullable] = ACTIONS(2618),
    [anon_sym__Nullable_result] = ACTIONS(2618),
    [anon_sym__Null_unspecified] = ACTIONS(2618),
    [anon_sym___autoreleasing] = ACTIONS(2618),
    [anon_sym___block] = ACTIONS(2618),
    [anon_sym___bridge] = ACTIONS(2618),
    [anon_sym___bridge_retained] = ACTIONS(2618),
    [anon_sym___bridge_transfer] = ACTIONS(2618),
    [anon_sym___complex] = ACTIONS(2618),
    [anon_sym___const] = ACTIONS(2618),
    [anon_sym___imag] = ACTIONS(2618),
    [anon_sym___kindof] = ACTIONS(2618),
    [anon_sym___nonnull] = ACTIONS(2618),
    [anon_sym___nullable] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2618),
    [anon_sym___real] = ACTIONS(2618),
    [anon_sym___strong] = ACTIONS(2618),
    [anon_sym___unsafe_unretained] = ACTIONS(2618),
    [anon_sym___unused] = ACTIONS(2618),
    [anon_sym___weak] = ACTIONS(2618),
    [anon_sym_alignas] = ACTIONS(2618),
    [anon_sym__Alignas] = ACTIONS(2618),
    [sym_primitive_type] = ACTIONS(2618),
    [anon_sym_enum] = ACTIONS(2618),
    [anon_sym_struct] = ACTIONS(2618),
    [anon_sym_union] = ACTIONS(2618),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2618),
    [anon_sym___typeof] = ACTIONS(2618),
    [anon_sym_typeof] = ACTIONS(2618),
    [aux_sym_preproc_undef_token1] = ACTIONS(2618),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2618),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2618),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2618),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_API_AVAILABLE] = ACTIONS(2618),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_API_DEPRECATED] = ACTIONS(2618),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym___deprecated_msg] = ACTIONS(2618),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2618),
    [anon_sym_ATend] = ACTIONS(2620),
    [anon_sym_ATsynthesize] = ACTIONS(2620),
    [anon_sym_ATdynamic] = ACTIONS(2620),
    [anon_sym_BOOL] = ACTIONS(2618),
    [anon_sym_IMP] = ACTIONS(2618),
    [anon_sym_SEL] = ACTIONS(2618),
    [anon_sym_Class] = ACTIONS(2618),
    [anon_sym_id] = ACTIONS(2618),
    [anon_sym_ATdefs] = ACTIONS(2620),
  },
  [STATE(2203)] = {
    [sym_identifier] = ACTIONS(5485),
    [aux_sym_preproc_def_token1] = ACTIONS(5485),
    [aux_sym_preproc_if_token1] = ACTIONS(5485),
    [aux_sym_preproc_if_token2] = ACTIONS(5485),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5485),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5485),
    [aux_sym_preproc_else_token1] = ACTIONS(5485),
    [aux_sym_preproc_elif_token1] = ACTIONS(5485),
    [sym_preproc_directive] = ACTIONS(5485),
    [anon_sym_DASH] = ACTIONS(5487),
    [anon_sym_PLUS] = ACTIONS(5487),
    [anon_sym___extension__] = ACTIONS(5485),
    [anon_sym_typedef] = ACTIONS(5485),
    [anon_sym_extern] = ACTIONS(5485),
    [anon_sym___attribute__] = ACTIONS(5485),
    [anon_sym___attribute] = ACTIONS(5485),
    [anon_sym_noreturn] = ACTIONS(5485),
    [anon_sym_LBRACK] = ACTIONS(5487),
    [anon_sym___declspec] = ACTIONS(5485),
    [anon_sym___cdecl] = ACTIONS(5485),
    [anon_sym___clrcall] = ACTIONS(5485),
    [anon_sym___stdcall] = ACTIONS(5485),
    [anon_sym___fastcall] = ACTIONS(5485),
    [anon_sym___thiscall] = ACTIONS(5485),
    [anon_sym___vectorcall] = ACTIONS(5485),
    [anon_sym_signed] = ACTIONS(5485),
    [anon_sym_unsigned] = ACTIONS(5485),
    [anon_sym_long] = ACTIONS(5485),
    [anon_sym_short] = ACTIONS(5485),
    [anon_sym_static] = ACTIONS(5485),
    [anon_sym_auto] = ACTIONS(5485),
    [anon_sym_register] = ACTIONS(5485),
    [anon_sym_inline] = ACTIONS(5485),
    [anon_sym___inline] = ACTIONS(5485),
    [anon_sym___inline__] = ACTIONS(5485),
    [anon_sym___forceinline] = ACTIONS(5485),
    [anon_sym_thread_local] = ACTIONS(5485),
    [anon_sym___thread] = ACTIONS(5485),
    [anon_sym_CG_EXTERN] = ACTIONS(5485),
    [anon_sym_CG_INLINE] = ACTIONS(5485),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5485),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5485),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5485),
    [anon_sym_IBOutlet] = ACTIONS(5485),
    [anon_sym_IBInspectable] = ACTIONS(5485),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5485),
    [anon_sym_NS_INLINE] = ACTIONS(5485),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5485),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5485),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5485),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5485),
    [anon_sym_const] = ACTIONS(5485),
    [anon_sym_constexpr] = ACTIONS(5485),
    [anon_sym_volatile] = ACTIONS(5485),
    [anon_sym_restrict] = ACTIONS(5485),
    [anon_sym___restrict__] = ACTIONS(5485),
    [anon_sym__Atomic] = ACTIONS(5485),
    [anon_sym__Noreturn] = ACTIONS(5485),
    [anon_sym__Nonnull] = ACTIONS(5485),
    [anon_sym_nullable] = ACTIONS(5485),
    [anon_sym__Complex] = ACTIONS(5485),
    [anon_sym__Nullable] = ACTIONS(5485),
    [anon_sym__Nullable_result] = ACTIONS(5485),
    [anon_sym__Null_unspecified] = ACTIONS(5485),
    [anon_sym___autoreleasing] = ACTIONS(5485),
    [anon_sym___block] = ACTIONS(5485),
    [anon_sym___bridge] = ACTIONS(5485),
    [anon_sym___bridge_retained] = ACTIONS(5485),
    [anon_sym___bridge_transfer] = ACTIONS(5485),
    [anon_sym___complex] = ACTIONS(5485),
    [anon_sym___const] = ACTIONS(5485),
    [anon_sym___imag] = ACTIONS(5485),
    [anon_sym___kindof] = ACTIONS(5485),
    [anon_sym___nonnull] = ACTIONS(5485),
    [anon_sym___nullable] = ACTIONS(5485),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5485),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5485),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5485),
    [anon_sym___real] = ACTIONS(5485),
    [anon_sym___strong] = ACTIONS(5485),
    [anon_sym___unsafe_unretained] = ACTIONS(5485),
    [anon_sym___unused] = ACTIONS(5485),
    [anon_sym___weak] = ACTIONS(5485),
    [anon_sym_alignas] = ACTIONS(5485),
    [anon_sym__Alignas] = ACTIONS(5485),
    [sym_primitive_type] = ACTIONS(5485),
    [anon_sym_enum] = ACTIONS(5485),
    [anon_sym_struct] = ACTIONS(5485),
    [anon_sym_union] = ACTIONS(5485),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5485),
    [anon_sym___typeof] = ACTIONS(5485),
    [anon_sym_typeof] = ACTIONS(5485),
    [aux_sym_preproc_undef_token1] = ACTIONS(5485),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5485),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5485),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5485),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5485),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5485),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5485),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5485),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5485),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5485),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_API_AVAILABLE] = ACTIONS(5485),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_API_DEPRECATED] = ACTIONS(5485),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5485),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5485),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5485),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5485),
    [anon_sym___deprecated_msg] = ACTIONS(5485),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5485),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5485),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5485),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5485),
    [anon_sym_ATsynthesize] = ACTIONS(5487),
    [anon_sym_ATdynamic] = ACTIONS(5487),
    [anon_sym_BOOL] = ACTIONS(5485),
    [anon_sym_IMP] = ACTIONS(5485),
    [anon_sym_SEL] = ACTIONS(5485),
    [anon_sym_Class] = ACTIONS(5485),
    [anon_sym_id] = ACTIONS(5485),
  },
  [STATE(2204)] = {
    [sym_identifier] = ACTIONS(2618),
    [aux_sym_preproc_def_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token2] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2618),
    [aux_sym_preproc_else_token1] = ACTIONS(2618),
    [aux_sym_preproc_elif_token1] = ACTIONS(2618),
    [sym_preproc_directive] = ACTIONS(2618),
    [anon_sym_DASH] = ACTIONS(2620),
    [anon_sym_PLUS] = ACTIONS(2620),
    [anon_sym___extension__] = ACTIONS(2618),
    [anon_sym_typedef] = ACTIONS(2618),
    [anon_sym_extern] = ACTIONS(2618),
    [anon_sym___attribute__] = ACTIONS(2618),
    [anon_sym___attribute] = ACTIONS(2618),
    [anon_sym_noreturn] = ACTIONS(2618),
    [anon_sym_LBRACK] = ACTIONS(2620),
    [anon_sym___declspec] = ACTIONS(2618),
    [anon_sym___cdecl] = ACTIONS(2618),
    [anon_sym___clrcall] = ACTIONS(2618),
    [anon_sym___stdcall] = ACTIONS(2618),
    [anon_sym___fastcall] = ACTIONS(2618),
    [anon_sym___thiscall] = ACTIONS(2618),
    [anon_sym___vectorcall] = ACTIONS(2618),
    [anon_sym_signed] = ACTIONS(2618),
    [anon_sym_unsigned] = ACTIONS(2618),
    [anon_sym_long] = ACTIONS(2618),
    [anon_sym_short] = ACTIONS(2618),
    [anon_sym_static] = ACTIONS(2618),
    [anon_sym_auto] = ACTIONS(2618),
    [anon_sym_register] = ACTIONS(2618),
    [anon_sym_inline] = ACTIONS(2618),
    [anon_sym___inline] = ACTIONS(2618),
    [anon_sym___inline__] = ACTIONS(2618),
    [anon_sym___forceinline] = ACTIONS(2618),
    [anon_sym_thread_local] = ACTIONS(2618),
    [anon_sym___thread] = ACTIONS(2618),
    [anon_sym_CG_EXTERN] = ACTIONS(2618),
    [anon_sym_CG_INLINE] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2618),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2618),
    [anon_sym_IBOutlet] = ACTIONS(2618),
    [anon_sym_IBInspectable] = ACTIONS(2618),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2618),
    [anon_sym_NS_INLINE] = ACTIONS(2618),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2618),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2618),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2618),
    [anon_sym_const] = ACTIONS(2618),
    [anon_sym_constexpr] = ACTIONS(2618),
    [anon_sym_volatile] = ACTIONS(2618),
    [anon_sym_restrict] = ACTIONS(2618),
    [anon_sym___restrict__] = ACTIONS(2618),
    [anon_sym__Atomic] = ACTIONS(2618),
    [anon_sym__Noreturn] = ACTIONS(2618),
    [anon_sym__Nonnull] = ACTIONS(2618),
    [anon_sym_nullable] = ACTIONS(2618),
    [anon_sym__Complex] = ACTIONS(2618),
    [anon_sym__Nullable] = ACTIONS(2618),
    [anon_sym__Nullable_result] = ACTIONS(2618),
    [anon_sym__Null_unspecified] = ACTIONS(2618),
    [anon_sym___autoreleasing] = ACTIONS(2618),
    [anon_sym___block] = ACTIONS(2618),
    [anon_sym___bridge] = ACTIONS(2618),
    [anon_sym___bridge_retained] = ACTIONS(2618),
    [anon_sym___bridge_transfer] = ACTIONS(2618),
    [anon_sym___complex] = ACTIONS(2618),
    [anon_sym___const] = ACTIONS(2618),
    [anon_sym___imag] = ACTIONS(2618),
    [anon_sym___kindof] = ACTIONS(2618),
    [anon_sym___nonnull] = ACTIONS(2618),
    [anon_sym___nullable] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2618),
    [anon_sym___real] = ACTIONS(2618),
    [anon_sym___strong] = ACTIONS(2618),
    [anon_sym___unsafe_unretained] = ACTIONS(2618),
    [anon_sym___unused] = ACTIONS(2618),
    [anon_sym___weak] = ACTIONS(2618),
    [anon_sym_alignas] = ACTIONS(2618),
    [anon_sym__Alignas] = ACTIONS(2618),
    [sym_primitive_type] = ACTIONS(2618),
    [anon_sym_enum] = ACTIONS(2618),
    [anon_sym_struct] = ACTIONS(2618),
    [anon_sym_union] = ACTIONS(2618),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2618),
    [anon_sym___typeof] = ACTIONS(2618),
    [anon_sym_typeof] = ACTIONS(2618),
    [aux_sym_preproc_undef_token1] = ACTIONS(2618),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2618),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2618),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2618),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_API_AVAILABLE] = ACTIONS(2618),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_API_DEPRECATED] = ACTIONS(2618),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym___deprecated_msg] = ACTIONS(2618),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2618),
    [anon_sym_ATsynthesize] = ACTIONS(2620),
    [anon_sym_ATdynamic] = ACTIONS(2620),
    [anon_sym_BOOL] = ACTIONS(2618),
    [anon_sym_IMP] = ACTIONS(2618),
    [anon_sym_SEL] = ACTIONS(2618),
    [anon_sym_Class] = ACTIONS(2618),
    [anon_sym_id] = ACTIONS(2618),
  },
  [STATE(2205)] = {
    [sym_identifier] = ACTIONS(5387),
    [aux_sym_preproc_def_token1] = ACTIONS(5387),
    [aux_sym_preproc_if_token1] = ACTIONS(5387),
    [aux_sym_preproc_if_token2] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5387),
    [aux_sym_preproc_else_token1] = ACTIONS(5387),
    [aux_sym_preproc_elif_token1] = ACTIONS(5387),
    [sym_preproc_directive] = ACTIONS(5387),
    [anon_sym_DASH] = ACTIONS(5389),
    [anon_sym_PLUS] = ACTIONS(5389),
    [anon_sym___extension__] = ACTIONS(5387),
    [anon_sym_typedef] = ACTIONS(5387),
    [anon_sym_extern] = ACTIONS(5387),
    [anon_sym___attribute__] = ACTIONS(5387),
    [anon_sym___attribute] = ACTIONS(5387),
    [anon_sym_noreturn] = ACTIONS(5387),
    [anon_sym_LBRACK] = ACTIONS(5389),
    [anon_sym___declspec] = ACTIONS(5387),
    [anon_sym___cdecl] = ACTIONS(5387),
    [anon_sym___clrcall] = ACTIONS(5387),
    [anon_sym___stdcall] = ACTIONS(5387),
    [anon_sym___fastcall] = ACTIONS(5387),
    [anon_sym___thiscall] = ACTIONS(5387),
    [anon_sym___vectorcall] = ACTIONS(5387),
    [anon_sym_signed] = ACTIONS(5387),
    [anon_sym_unsigned] = ACTIONS(5387),
    [anon_sym_long] = ACTIONS(5387),
    [anon_sym_short] = ACTIONS(5387),
    [anon_sym_static] = ACTIONS(5387),
    [anon_sym_auto] = ACTIONS(5387),
    [anon_sym_register] = ACTIONS(5387),
    [anon_sym_inline] = ACTIONS(5387),
    [anon_sym___inline] = ACTIONS(5387),
    [anon_sym___inline__] = ACTIONS(5387),
    [anon_sym___forceinline] = ACTIONS(5387),
    [anon_sym_thread_local] = ACTIONS(5387),
    [anon_sym___thread] = ACTIONS(5387),
    [anon_sym_CG_EXTERN] = ACTIONS(5387),
    [anon_sym_CG_INLINE] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5387),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5387),
    [anon_sym_IBOutlet] = ACTIONS(5387),
    [anon_sym_IBInspectable] = ACTIONS(5387),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5387),
    [anon_sym_NS_INLINE] = ACTIONS(5387),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5387),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5387),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5387),
    [anon_sym_const] = ACTIONS(5387),
    [anon_sym_constexpr] = ACTIONS(5387),
    [anon_sym_volatile] = ACTIONS(5387),
    [anon_sym_restrict] = ACTIONS(5387),
    [anon_sym___restrict__] = ACTIONS(5387),
    [anon_sym__Atomic] = ACTIONS(5387),
    [anon_sym__Noreturn] = ACTIONS(5387),
    [anon_sym__Nonnull] = ACTIONS(5387),
    [anon_sym_nullable] = ACTIONS(5387),
    [anon_sym__Complex] = ACTIONS(5387),
    [anon_sym__Nullable] = ACTIONS(5387),
    [anon_sym__Nullable_result] = ACTIONS(5387),
    [anon_sym__Null_unspecified] = ACTIONS(5387),
    [anon_sym___autoreleasing] = ACTIONS(5387),
    [anon_sym___block] = ACTIONS(5387),
    [anon_sym___bridge] = ACTIONS(5387),
    [anon_sym___bridge_retained] = ACTIONS(5387),
    [anon_sym___bridge_transfer] = ACTIONS(5387),
    [anon_sym___complex] = ACTIONS(5387),
    [anon_sym___const] = ACTIONS(5387),
    [anon_sym___imag] = ACTIONS(5387),
    [anon_sym___kindof] = ACTIONS(5387),
    [anon_sym___nonnull] = ACTIONS(5387),
    [anon_sym___nullable] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5387),
    [anon_sym___real] = ACTIONS(5387),
    [anon_sym___strong] = ACTIONS(5387),
    [anon_sym___unsafe_unretained] = ACTIONS(5387),
    [anon_sym___unused] = ACTIONS(5387),
    [anon_sym___weak] = ACTIONS(5387),
    [anon_sym_alignas] = ACTIONS(5387),
    [anon_sym__Alignas] = ACTIONS(5387),
    [sym_primitive_type] = ACTIONS(5387),
    [anon_sym_enum] = ACTIONS(5387),
    [anon_sym_struct] = ACTIONS(5387),
    [anon_sym_union] = ACTIONS(5387),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5387),
    [anon_sym___typeof] = ACTIONS(5387),
    [anon_sym_typeof] = ACTIONS(5387),
    [aux_sym_preproc_undef_token1] = ACTIONS(5387),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5387),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5387),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5387),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_API_AVAILABLE] = ACTIONS(5387),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_API_DEPRECATED] = ACTIONS(5387),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym___deprecated_msg] = ACTIONS(5387),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5387),
    [anon_sym_ATsynthesize] = ACTIONS(5389),
    [anon_sym_ATdynamic] = ACTIONS(5389),
    [anon_sym_BOOL] = ACTIONS(5387),
    [anon_sym_IMP] = ACTIONS(5387),
    [anon_sym_SEL] = ACTIONS(5387),
    [anon_sym_Class] = ACTIONS(5387),
    [anon_sym_id] = ACTIONS(5387),
  },
  [STATE(2206)] = {
    [sym_method_type] = STATE(6560),
    [sym_identifier] = ACTIONS(5304),
    [aux_sym_preproc_def_token1] = ACTIONS(5306),
    [aux_sym_preproc_if_token1] = ACTIONS(5306),
    [aux_sym_preproc_if_token2] = ACTIONS(5306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5306),
    [sym_preproc_directive] = ACTIONS(5306),
    [anon_sym_LPAREN2] = ACTIONS(5308),
    [anon_sym_DASH] = ACTIONS(5310),
    [anon_sym_PLUS] = ACTIONS(5310),
    [anon_sym_SEMI] = ACTIONS(5310),
    [anon_sym___extension__] = ACTIONS(5306),
    [anon_sym_typedef] = ACTIONS(5306),
    [anon_sym_extern] = ACTIONS(5306),
    [anon_sym___attribute__] = ACTIONS(5306),
    [anon_sym___attribute] = ACTIONS(5306),
    [anon_sym_noreturn] = ACTIONS(5306),
    [anon_sym_LBRACK] = ACTIONS(5310),
    [anon_sym___declspec] = ACTIONS(5306),
    [anon_sym___cdecl] = ACTIONS(5306),
    [anon_sym___clrcall] = ACTIONS(5306),
    [anon_sym___stdcall] = ACTIONS(5306),
    [anon_sym___fastcall] = ACTIONS(5306),
    [anon_sym___thiscall] = ACTIONS(5306),
    [anon_sym___vectorcall] = ACTIONS(5306),
    [anon_sym_signed] = ACTIONS(5306),
    [anon_sym_unsigned] = ACTIONS(5306),
    [anon_sym_long] = ACTIONS(5306),
    [anon_sym_short] = ACTIONS(5306),
    [anon_sym_static] = ACTIONS(5306),
    [anon_sym_auto] = ACTIONS(5306),
    [anon_sym_register] = ACTIONS(5306),
    [anon_sym_inline] = ACTIONS(5306),
    [anon_sym___inline] = ACTIONS(5306),
    [anon_sym___inline__] = ACTIONS(5306),
    [anon_sym___forceinline] = ACTIONS(5306),
    [anon_sym_thread_local] = ACTIONS(5306),
    [anon_sym___thread] = ACTIONS(5306),
    [anon_sym_CG_EXTERN] = ACTIONS(5306),
    [anon_sym_CG_INLINE] = ACTIONS(5306),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5306),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5306),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5306),
    [anon_sym_IBOutlet] = ACTIONS(5306),
    [anon_sym_IBInspectable] = ACTIONS(5306),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5306),
    [anon_sym_NS_INLINE] = ACTIONS(5306),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5306),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5306),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5306),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5306),
    [anon_sym_const] = ACTIONS(5306),
    [anon_sym_constexpr] = ACTIONS(5306),
    [anon_sym_volatile] = ACTIONS(5306),
    [anon_sym_restrict] = ACTIONS(5306),
    [anon_sym___restrict__] = ACTIONS(5306),
    [anon_sym__Atomic] = ACTIONS(5306),
    [anon_sym__Noreturn] = ACTIONS(5306),
    [anon_sym__Nonnull] = ACTIONS(5306),
    [anon_sym_nullable] = ACTIONS(5306),
    [anon_sym__Complex] = ACTIONS(5306),
    [anon_sym__Nullable] = ACTIONS(5306),
    [anon_sym__Nullable_result] = ACTIONS(5306),
    [anon_sym__Null_unspecified] = ACTIONS(5306),
    [anon_sym___autoreleasing] = ACTIONS(5306),
    [anon_sym___block] = ACTIONS(5306),
    [anon_sym___bridge] = ACTIONS(5306),
    [anon_sym___bridge_retained] = ACTIONS(5306),
    [anon_sym___bridge_transfer] = ACTIONS(5306),
    [anon_sym___complex] = ACTIONS(5306),
    [anon_sym___const] = ACTIONS(5306),
    [anon_sym___imag] = ACTIONS(5306),
    [anon_sym___kindof] = ACTIONS(5306),
    [anon_sym___nonnull] = ACTIONS(5306),
    [anon_sym___nullable] = ACTIONS(5306),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5306),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5306),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5306),
    [anon_sym___real] = ACTIONS(5306),
    [anon_sym___strong] = ACTIONS(5306),
    [anon_sym___unsafe_unretained] = ACTIONS(5306),
    [anon_sym___unused] = ACTIONS(5306),
    [anon_sym___weak] = ACTIONS(5306),
    [anon_sym_alignas] = ACTIONS(5306),
    [anon_sym__Alignas] = ACTIONS(5306),
    [sym_primitive_type] = ACTIONS(5306),
    [anon_sym_enum] = ACTIONS(5306),
    [anon_sym_struct] = ACTIONS(5306),
    [anon_sym_union] = ACTIONS(5306),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5306),
    [anon_sym___typeof] = ACTIONS(5306),
    [anon_sym_typeof] = ACTIONS(5306),
    [aux_sym_preproc_undef_token1] = ACTIONS(5306),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5306),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5306),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5306),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5306),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5306),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5306),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5306),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5306),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5306),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_API_AVAILABLE] = ACTIONS(5306),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_API_DEPRECATED] = ACTIONS(5306),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5306),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5306),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5306),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5306),
    [anon_sym___deprecated_msg] = ACTIONS(5306),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5306),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5306),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5306),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5306),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5306),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5306),
    [anon_sym_ATproperty] = ACTIONS(5310),
    [anon_sym_BOOL] = ACTIONS(5306),
    [anon_sym_IMP] = ACTIONS(5306),
    [anon_sym_SEL] = ACTIONS(5306),
    [anon_sym_Class] = ACTIONS(5306),
    [anon_sym_id] = ACTIONS(5306),
  },
  [STATE(2207)] = {
    [sym_identifier] = ACTIONS(5489),
    [aux_sym_preproc_def_token1] = ACTIONS(5489),
    [aux_sym_preproc_if_token1] = ACTIONS(5489),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5489),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5489),
    [sym_preproc_directive] = ACTIONS(5489),
    [anon_sym_DASH] = ACTIONS(5491),
    [anon_sym_PLUS] = ACTIONS(5491),
    [anon_sym_SEMI] = ACTIONS(5493),
    [anon_sym___extension__] = ACTIONS(5489),
    [anon_sym_typedef] = ACTIONS(5489),
    [anon_sym_extern] = ACTIONS(5489),
    [anon_sym___attribute__] = ACTIONS(5489),
    [anon_sym___attribute] = ACTIONS(5489),
    [anon_sym_noreturn] = ACTIONS(5489),
    [anon_sym_LBRACK] = ACTIONS(5491),
    [anon_sym___declspec] = ACTIONS(5489),
    [anon_sym___cdecl] = ACTIONS(5489),
    [anon_sym___clrcall] = ACTIONS(5489),
    [anon_sym___stdcall] = ACTIONS(5489),
    [anon_sym___fastcall] = ACTIONS(5489),
    [anon_sym___thiscall] = ACTIONS(5489),
    [anon_sym___vectorcall] = ACTIONS(5489),
    [anon_sym_signed] = ACTIONS(5489),
    [anon_sym_unsigned] = ACTIONS(5489),
    [anon_sym_long] = ACTIONS(5489),
    [anon_sym_short] = ACTIONS(5489),
    [anon_sym_static] = ACTIONS(5489),
    [anon_sym_auto] = ACTIONS(5489),
    [anon_sym_register] = ACTIONS(5489),
    [anon_sym_inline] = ACTIONS(5489),
    [anon_sym___inline] = ACTIONS(5489),
    [anon_sym___inline__] = ACTIONS(5489),
    [anon_sym___forceinline] = ACTIONS(5489),
    [anon_sym_thread_local] = ACTIONS(5489),
    [anon_sym___thread] = ACTIONS(5489),
    [anon_sym_CG_EXTERN] = ACTIONS(5489),
    [anon_sym_CG_INLINE] = ACTIONS(5489),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5489),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5489),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5489),
    [anon_sym_IBOutlet] = ACTIONS(5489),
    [anon_sym_IBInspectable] = ACTIONS(5489),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5489),
    [anon_sym_NS_INLINE] = ACTIONS(5489),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5489),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5489),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5489),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5489),
    [anon_sym_const] = ACTIONS(5489),
    [anon_sym_constexpr] = ACTIONS(5489),
    [anon_sym_volatile] = ACTIONS(5489),
    [anon_sym_restrict] = ACTIONS(5489),
    [anon_sym___restrict__] = ACTIONS(5489),
    [anon_sym__Atomic] = ACTIONS(5489),
    [anon_sym__Noreturn] = ACTIONS(5489),
    [anon_sym__Nonnull] = ACTIONS(5489),
    [anon_sym_nullable] = ACTIONS(5489),
    [anon_sym__Complex] = ACTIONS(5489),
    [anon_sym__Nullable] = ACTIONS(5489),
    [anon_sym__Nullable_result] = ACTIONS(5489),
    [anon_sym__Null_unspecified] = ACTIONS(5489),
    [anon_sym___autoreleasing] = ACTIONS(5489),
    [anon_sym___block] = ACTIONS(5489),
    [anon_sym___bridge] = ACTIONS(5489),
    [anon_sym___bridge_retained] = ACTIONS(5489),
    [anon_sym___bridge_transfer] = ACTIONS(5489),
    [anon_sym___complex] = ACTIONS(5489),
    [anon_sym___const] = ACTIONS(5489),
    [anon_sym___imag] = ACTIONS(5489),
    [anon_sym___kindof] = ACTIONS(5489),
    [anon_sym___nonnull] = ACTIONS(5489),
    [anon_sym___nullable] = ACTIONS(5489),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5489),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5489),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5489),
    [anon_sym___real] = ACTIONS(5489),
    [anon_sym___strong] = ACTIONS(5489),
    [anon_sym___unsafe_unretained] = ACTIONS(5489),
    [anon_sym___unused] = ACTIONS(5489),
    [anon_sym___weak] = ACTIONS(5489),
    [anon_sym_alignas] = ACTIONS(5489),
    [anon_sym__Alignas] = ACTIONS(5489),
    [sym_primitive_type] = ACTIONS(5489),
    [anon_sym_enum] = ACTIONS(5489),
    [anon_sym_struct] = ACTIONS(5489),
    [anon_sym_union] = ACTIONS(5489),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5489),
    [anon_sym___typeof] = ACTIONS(5489),
    [anon_sym_typeof] = ACTIONS(5489),
    [aux_sym_preproc_undef_token1] = ACTIONS(5489),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5489),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5489),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5489),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5489),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5489),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5489),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5489),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5489),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5489),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5489),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5489),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5489),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5489),
    [anon_sym_API_AVAILABLE] = ACTIONS(5489),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5489),
    [anon_sym_API_DEPRECATED] = ACTIONS(5489),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5489),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5489),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5489),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5489),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5489),
    [anon_sym___deprecated_msg] = ACTIONS(5489),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5489),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5489),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5489),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5489),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5489),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5489),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5489),
    [anon_sym_ATend] = ACTIONS(5491),
    [anon_sym_ATsynthesize] = ACTIONS(5491),
    [anon_sym_ATdynamic] = ACTIONS(5491),
    [anon_sym_ATproperty] = ACTIONS(5491),
    [anon_sym_BOOL] = ACTIONS(5489),
    [anon_sym_IMP] = ACTIONS(5489),
    [anon_sym_SEL] = ACTIONS(5489),
    [anon_sym_Class] = ACTIONS(5489),
    [anon_sym_id] = ACTIONS(5489),
  },
  [STATE(2208)] = {
    [sym_identifier] = ACTIONS(2856),
    [aux_sym_preproc_def_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token2] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2856),
    [sym_preproc_directive] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2858),
    [anon_sym_PLUS] = ACTIONS(2858),
    [anon_sym___extension__] = ACTIONS(2856),
    [anon_sym_typedef] = ACTIONS(2856),
    [anon_sym_extern] = ACTIONS(2856),
    [anon_sym___attribute__] = ACTIONS(2856),
    [anon_sym___attribute] = ACTIONS(2856),
    [anon_sym_noreturn] = ACTIONS(2856),
    [anon_sym_LBRACK] = ACTIONS(2858),
    [anon_sym___declspec] = ACTIONS(2856),
    [anon_sym___cdecl] = ACTIONS(2856),
    [anon_sym___clrcall] = ACTIONS(2856),
    [anon_sym___stdcall] = ACTIONS(2856),
    [anon_sym___fastcall] = ACTIONS(2856),
    [anon_sym___thiscall] = ACTIONS(2856),
    [anon_sym___vectorcall] = ACTIONS(2856),
    [anon_sym_signed] = ACTIONS(2856),
    [anon_sym_unsigned] = ACTIONS(2856),
    [anon_sym_long] = ACTIONS(2856),
    [anon_sym_short] = ACTIONS(2856),
    [anon_sym_static] = ACTIONS(2856),
    [anon_sym_auto] = ACTIONS(2856),
    [anon_sym_register] = ACTIONS(2856),
    [anon_sym_inline] = ACTIONS(2856),
    [anon_sym___inline] = ACTIONS(2856),
    [anon_sym___inline__] = ACTIONS(2856),
    [anon_sym___forceinline] = ACTIONS(2856),
    [anon_sym_thread_local] = ACTIONS(2856),
    [anon_sym___thread] = ACTIONS(2856),
    [anon_sym_CG_EXTERN] = ACTIONS(2856),
    [anon_sym_CG_INLINE] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2856),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2856),
    [anon_sym_IBOutlet] = ACTIONS(2856),
    [anon_sym_IBInspectable] = ACTIONS(2856),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2856),
    [anon_sym_NS_INLINE] = ACTIONS(2856),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2856),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2856),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2856),
    [anon_sym_const] = ACTIONS(2856),
    [anon_sym_constexpr] = ACTIONS(2856),
    [anon_sym_volatile] = ACTIONS(2856),
    [anon_sym_restrict] = ACTIONS(2856),
    [anon_sym___restrict__] = ACTIONS(2856),
    [anon_sym__Atomic] = ACTIONS(2856),
    [anon_sym__Noreturn] = ACTIONS(2856),
    [anon_sym__Nonnull] = ACTIONS(2856),
    [anon_sym_nullable] = ACTIONS(2856),
    [anon_sym__Complex] = ACTIONS(2856),
    [anon_sym__Nullable] = ACTIONS(2856),
    [anon_sym__Nullable_result] = ACTIONS(2856),
    [anon_sym__Null_unspecified] = ACTIONS(2856),
    [anon_sym___autoreleasing] = ACTIONS(2856),
    [anon_sym___block] = ACTIONS(2856),
    [anon_sym___bridge] = ACTIONS(2856),
    [anon_sym___bridge_retained] = ACTIONS(2856),
    [anon_sym___bridge_transfer] = ACTIONS(2856),
    [anon_sym___complex] = ACTIONS(2856),
    [anon_sym___const] = ACTIONS(2856),
    [anon_sym___imag] = ACTIONS(2856),
    [anon_sym___kindof] = ACTIONS(2856),
    [anon_sym___nonnull] = ACTIONS(2856),
    [anon_sym___nullable] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2856),
    [anon_sym___real] = ACTIONS(2856),
    [anon_sym___strong] = ACTIONS(2856),
    [anon_sym___unsafe_unretained] = ACTIONS(2856),
    [anon_sym___unused] = ACTIONS(2856),
    [anon_sym___weak] = ACTIONS(2856),
    [anon_sym_alignas] = ACTIONS(2856),
    [anon_sym__Alignas] = ACTIONS(2856),
    [sym_primitive_type] = ACTIONS(2856),
    [anon_sym_enum] = ACTIONS(2856),
    [anon_sym_struct] = ACTIONS(2856),
    [anon_sym_union] = ACTIONS(2856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2856),
    [anon_sym___typeof] = ACTIONS(2856),
    [anon_sym_typeof] = ACTIONS(2856),
    [aux_sym_preproc_undef_token1] = ACTIONS(2856),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2856),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2856),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2856),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_API_AVAILABLE] = ACTIONS(2856),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_API_DEPRECATED] = ACTIONS(2856),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym___deprecated_msg] = ACTIONS(2856),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2856),
    [anon_sym_ATsynthesize] = ACTIONS(2858),
    [anon_sym_ATdynamic] = ACTIONS(2858),
    [anon_sym_ATproperty] = ACTIONS(2858),
    [anon_sym_BOOL] = ACTIONS(2856),
    [anon_sym_IMP] = ACTIONS(2856),
    [anon_sym_SEL] = ACTIONS(2856),
    [anon_sym_Class] = ACTIONS(2856),
    [anon_sym_id] = ACTIONS(2856),
    [anon_sym_ATdefs] = ACTIONS(2858),
  },
  [STATE(2209)] = {
    [sym_identifier] = ACTIONS(2588),
    [aux_sym_preproc_def_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token2] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2588),
    [sym_preproc_directive] = ACTIONS(2588),
    [anon_sym_DASH] = ACTIONS(2590),
    [anon_sym_PLUS] = ACTIONS(2590),
    [anon_sym___extension__] = ACTIONS(2588),
    [anon_sym_typedef] = ACTIONS(2588),
    [anon_sym_extern] = ACTIONS(2588),
    [anon_sym___attribute__] = ACTIONS(2588),
    [anon_sym___attribute] = ACTIONS(2588),
    [anon_sym_noreturn] = ACTIONS(2588),
    [anon_sym_LBRACK] = ACTIONS(2590),
    [anon_sym___declspec] = ACTIONS(2588),
    [anon_sym___cdecl] = ACTIONS(2588),
    [anon_sym___clrcall] = ACTIONS(2588),
    [anon_sym___stdcall] = ACTIONS(2588),
    [anon_sym___fastcall] = ACTIONS(2588),
    [anon_sym___thiscall] = ACTIONS(2588),
    [anon_sym___vectorcall] = ACTIONS(2588),
    [anon_sym_signed] = ACTIONS(2588),
    [anon_sym_unsigned] = ACTIONS(2588),
    [anon_sym_long] = ACTIONS(2588),
    [anon_sym_short] = ACTIONS(2588),
    [anon_sym_static] = ACTIONS(2588),
    [anon_sym_auto] = ACTIONS(2588),
    [anon_sym_register] = ACTIONS(2588),
    [anon_sym_inline] = ACTIONS(2588),
    [anon_sym___inline] = ACTIONS(2588),
    [anon_sym___inline__] = ACTIONS(2588),
    [anon_sym___forceinline] = ACTIONS(2588),
    [anon_sym_thread_local] = ACTIONS(2588),
    [anon_sym___thread] = ACTIONS(2588),
    [anon_sym_CG_EXTERN] = ACTIONS(2588),
    [anon_sym_CG_INLINE] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2588),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2588),
    [anon_sym_IBOutlet] = ACTIONS(2588),
    [anon_sym_IBInspectable] = ACTIONS(2588),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2588),
    [anon_sym_NS_INLINE] = ACTIONS(2588),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2588),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2588),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2588),
    [anon_sym_const] = ACTIONS(2588),
    [anon_sym_constexpr] = ACTIONS(2588),
    [anon_sym_volatile] = ACTIONS(2588),
    [anon_sym_restrict] = ACTIONS(2588),
    [anon_sym___restrict__] = ACTIONS(2588),
    [anon_sym__Atomic] = ACTIONS(2588),
    [anon_sym__Noreturn] = ACTIONS(2588),
    [anon_sym__Nonnull] = ACTIONS(2588),
    [anon_sym_nullable] = ACTIONS(2588),
    [anon_sym__Complex] = ACTIONS(2588),
    [anon_sym__Nullable] = ACTIONS(2588),
    [anon_sym__Nullable_result] = ACTIONS(2588),
    [anon_sym__Null_unspecified] = ACTIONS(2588),
    [anon_sym___autoreleasing] = ACTIONS(2588),
    [anon_sym___block] = ACTIONS(2588),
    [anon_sym___bridge] = ACTIONS(2588),
    [anon_sym___bridge_retained] = ACTIONS(2588),
    [anon_sym___bridge_transfer] = ACTIONS(2588),
    [anon_sym___complex] = ACTIONS(2588),
    [anon_sym___const] = ACTIONS(2588),
    [anon_sym___imag] = ACTIONS(2588),
    [anon_sym___kindof] = ACTIONS(2588),
    [anon_sym___nonnull] = ACTIONS(2588),
    [anon_sym___nullable] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2588),
    [anon_sym___real] = ACTIONS(2588),
    [anon_sym___strong] = ACTIONS(2588),
    [anon_sym___unsafe_unretained] = ACTIONS(2588),
    [anon_sym___unused] = ACTIONS(2588),
    [anon_sym___weak] = ACTIONS(2588),
    [anon_sym_alignas] = ACTIONS(2588),
    [anon_sym__Alignas] = ACTIONS(2588),
    [sym_primitive_type] = ACTIONS(2588),
    [anon_sym_enum] = ACTIONS(2588),
    [anon_sym_struct] = ACTIONS(2588),
    [anon_sym_union] = ACTIONS(2588),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2588),
    [anon_sym___typeof] = ACTIONS(2588),
    [anon_sym_typeof] = ACTIONS(2588),
    [aux_sym_preproc_undef_token1] = ACTIONS(2588),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2588),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2588),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2588),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_API_AVAILABLE] = ACTIONS(2588),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_API_DEPRECATED] = ACTIONS(2588),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym___deprecated_msg] = ACTIONS(2588),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2588),
    [anon_sym_ATsynthesize] = ACTIONS(2590),
    [anon_sym_ATdynamic] = ACTIONS(2590),
    [anon_sym_ATproperty] = ACTIONS(2590),
    [anon_sym_BOOL] = ACTIONS(2588),
    [anon_sym_IMP] = ACTIONS(2588),
    [anon_sym_SEL] = ACTIONS(2588),
    [anon_sym_Class] = ACTIONS(2588),
    [anon_sym_id] = ACTIONS(2588),
    [anon_sym_ATdefs] = ACTIONS(2590),
  },
  [STATE(2210)] = {
    [sym_identifier] = ACTIONS(2600),
    [aux_sym_preproc_def_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2600),
    [sym_preproc_directive] = ACTIONS(2600),
    [anon_sym_DASH] = ACTIONS(2602),
    [anon_sym_PLUS] = ACTIONS(2602),
    [anon_sym___extension__] = ACTIONS(2600),
    [anon_sym_typedef] = ACTIONS(2600),
    [anon_sym_extern] = ACTIONS(2600),
    [anon_sym___attribute__] = ACTIONS(2600),
    [anon_sym___attribute] = ACTIONS(2600),
    [anon_sym_noreturn] = ACTIONS(2600),
    [anon_sym_LBRACK] = ACTIONS(2602),
    [anon_sym___declspec] = ACTIONS(2600),
    [anon_sym___cdecl] = ACTIONS(2600),
    [anon_sym___clrcall] = ACTIONS(2600),
    [anon_sym___stdcall] = ACTIONS(2600),
    [anon_sym___fastcall] = ACTIONS(2600),
    [anon_sym___thiscall] = ACTIONS(2600),
    [anon_sym___vectorcall] = ACTIONS(2600),
    [anon_sym_RBRACE] = ACTIONS(2602),
    [anon_sym_signed] = ACTIONS(2600),
    [anon_sym_unsigned] = ACTIONS(2600),
    [anon_sym_long] = ACTIONS(2600),
    [anon_sym_short] = ACTIONS(2600),
    [anon_sym_static] = ACTIONS(2600),
    [anon_sym_auto] = ACTIONS(2600),
    [anon_sym_register] = ACTIONS(2600),
    [anon_sym_inline] = ACTIONS(2600),
    [anon_sym___inline] = ACTIONS(2600),
    [anon_sym___inline__] = ACTIONS(2600),
    [anon_sym___forceinline] = ACTIONS(2600),
    [anon_sym_thread_local] = ACTIONS(2600),
    [anon_sym___thread] = ACTIONS(2600),
    [anon_sym_CG_EXTERN] = ACTIONS(2600),
    [anon_sym_CG_INLINE] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2600),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2600),
    [anon_sym_IBOutlet] = ACTIONS(2600),
    [anon_sym_IBInspectable] = ACTIONS(2600),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2600),
    [anon_sym_NS_INLINE] = ACTIONS(2600),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2600),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2600),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2600),
    [anon_sym_const] = ACTIONS(2600),
    [anon_sym_constexpr] = ACTIONS(2600),
    [anon_sym_volatile] = ACTIONS(2600),
    [anon_sym_restrict] = ACTIONS(2600),
    [anon_sym___restrict__] = ACTIONS(2600),
    [anon_sym__Atomic] = ACTIONS(2600),
    [anon_sym__Noreturn] = ACTIONS(2600),
    [anon_sym__Nonnull] = ACTIONS(2600),
    [anon_sym_nullable] = ACTIONS(2600),
    [anon_sym__Complex] = ACTIONS(2600),
    [anon_sym__Nullable] = ACTIONS(2600),
    [anon_sym__Nullable_result] = ACTIONS(2600),
    [anon_sym__Null_unspecified] = ACTIONS(2600),
    [anon_sym___autoreleasing] = ACTIONS(2600),
    [anon_sym___block] = ACTIONS(2600),
    [anon_sym___bridge] = ACTIONS(2600),
    [anon_sym___bridge_retained] = ACTIONS(2600),
    [anon_sym___bridge_transfer] = ACTIONS(2600),
    [anon_sym___complex] = ACTIONS(2600),
    [anon_sym___const] = ACTIONS(2600),
    [anon_sym___imag] = ACTIONS(2600),
    [anon_sym___kindof] = ACTIONS(2600),
    [anon_sym___nonnull] = ACTIONS(2600),
    [anon_sym___nullable] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2600),
    [anon_sym___real] = ACTIONS(2600),
    [anon_sym___strong] = ACTIONS(2600),
    [anon_sym___unsafe_unretained] = ACTIONS(2600),
    [anon_sym___unused] = ACTIONS(2600),
    [anon_sym___weak] = ACTIONS(2600),
    [anon_sym_alignas] = ACTIONS(2600),
    [anon_sym__Alignas] = ACTIONS(2600),
    [sym_primitive_type] = ACTIONS(2600),
    [anon_sym_enum] = ACTIONS(2600),
    [anon_sym_struct] = ACTIONS(2600),
    [anon_sym_union] = ACTIONS(2600),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2600),
    [anon_sym___typeof] = ACTIONS(2600),
    [anon_sym_typeof] = ACTIONS(2600),
    [aux_sym_preproc_undef_token1] = ACTIONS(2600),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2600),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2600),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2600),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_API_AVAILABLE] = ACTIONS(2600),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_API_DEPRECATED] = ACTIONS(2600),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym___deprecated_msg] = ACTIONS(2600),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2600),
    [anon_sym_ATend] = ACTIONS(2602),
    [anon_sym_ATsynthesize] = ACTIONS(2602),
    [anon_sym_ATdynamic] = ACTIONS(2602),
    [anon_sym_BOOL] = ACTIONS(2600),
    [anon_sym_IMP] = ACTIONS(2600),
    [anon_sym_SEL] = ACTIONS(2600),
    [anon_sym_Class] = ACTIONS(2600),
    [anon_sym_id] = ACTIONS(2600),
    [anon_sym_ATdefs] = ACTIONS(2602),
  },
  [STATE(2211)] = {
    [sym_identifier] = ACTIONS(2852),
    [aux_sym_preproc_def_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token2] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2852),
    [sym_preproc_directive] = ACTIONS(2852),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym___extension__] = ACTIONS(2852),
    [anon_sym_typedef] = ACTIONS(2852),
    [anon_sym_extern] = ACTIONS(2852),
    [anon_sym___attribute__] = ACTIONS(2852),
    [anon_sym___attribute] = ACTIONS(2852),
    [anon_sym_noreturn] = ACTIONS(2852),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym___declspec] = ACTIONS(2852),
    [anon_sym___cdecl] = ACTIONS(2852),
    [anon_sym___clrcall] = ACTIONS(2852),
    [anon_sym___stdcall] = ACTIONS(2852),
    [anon_sym___fastcall] = ACTIONS(2852),
    [anon_sym___thiscall] = ACTIONS(2852),
    [anon_sym___vectorcall] = ACTIONS(2852),
    [anon_sym_signed] = ACTIONS(2852),
    [anon_sym_unsigned] = ACTIONS(2852),
    [anon_sym_long] = ACTIONS(2852),
    [anon_sym_short] = ACTIONS(2852),
    [anon_sym_static] = ACTIONS(2852),
    [anon_sym_auto] = ACTIONS(2852),
    [anon_sym_register] = ACTIONS(2852),
    [anon_sym_inline] = ACTIONS(2852),
    [anon_sym___inline] = ACTIONS(2852),
    [anon_sym___inline__] = ACTIONS(2852),
    [anon_sym___forceinline] = ACTIONS(2852),
    [anon_sym_thread_local] = ACTIONS(2852),
    [anon_sym___thread] = ACTIONS(2852),
    [anon_sym_CG_EXTERN] = ACTIONS(2852),
    [anon_sym_CG_INLINE] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2852),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2852),
    [anon_sym_IBOutlet] = ACTIONS(2852),
    [anon_sym_IBInspectable] = ACTIONS(2852),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2852),
    [anon_sym_NS_INLINE] = ACTIONS(2852),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2852),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2852),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2852),
    [anon_sym_const] = ACTIONS(2852),
    [anon_sym_constexpr] = ACTIONS(2852),
    [anon_sym_volatile] = ACTIONS(2852),
    [anon_sym_restrict] = ACTIONS(2852),
    [anon_sym___restrict__] = ACTIONS(2852),
    [anon_sym__Atomic] = ACTIONS(2852),
    [anon_sym__Noreturn] = ACTIONS(2852),
    [anon_sym__Nonnull] = ACTIONS(2852),
    [anon_sym_nullable] = ACTIONS(2852),
    [anon_sym__Complex] = ACTIONS(2852),
    [anon_sym__Nullable] = ACTIONS(2852),
    [anon_sym__Nullable_result] = ACTIONS(2852),
    [anon_sym__Null_unspecified] = ACTIONS(2852),
    [anon_sym___autoreleasing] = ACTIONS(2852),
    [anon_sym___block] = ACTIONS(2852),
    [anon_sym___bridge] = ACTIONS(2852),
    [anon_sym___bridge_retained] = ACTIONS(2852),
    [anon_sym___bridge_transfer] = ACTIONS(2852),
    [anon_sym___complex] = ACTIONS(2852),
    [anon_sym___const] = ACTIONS(2852),
    [anon_sym___imag] = ACTIONS(2852),
    [anon_sym___kindof] = ACTIONS(2852),
    [anon_sym___nonnull] = ACTIONS(2852),
    [anon_sym___nullable] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2852),
    [anon_sym___real] = ACTIONS(2852),
    [anon_sym___strong] = ACTIONS(2852),
    [anon_sym___unsafe_unretained] = ACTIONS(2852),
    [anon_sym___unused] = ACTIONS(2852),
    [anon_sym___weak] = ACTIONS(2852),
    [anon_sym_alignas] = ACTIONS(2852),
    [anon_sym__Alignas] = ACTIONS(2852),
    [sym_primitive_type] = ACTIONS(2852),
    [anon_sym_enum] = ACTIONS(2852),
    [anon_sym_struct] = ACTIONS(2852),
    [anon_sym_union] = ACTIONS(2852),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2852),
    [anon_sym___typeof] = ACTIONS(2852),
    [anon_sym_typeof] = ACTIONS(2852),
    [aux_sym_preproc_undef_token1] = ACTIONS(2852),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2852),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2852),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2852),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_API_AVAILABLE] = ACTIONS(2852),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_API_DEPRECATED] = ACTIONS(2852),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym___deprecated_msg] = ACTIONS(2852),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2852),
    [anon_sym_ATsynthesize] = ACTIONS(2854),
    [anon_sym_ATdynamic] = ACTIONS(2854),
    [anon_sym_ATproperty] = ACTIONS(2854),
    [anon_sym_BOOL] = ACTIONS(2852),
    [anon_sym_IMP] = ACTIONS(2852),
    [anon_sym_SEL] = ACTIONS(2852),
    [anon_sym_Class] = ACTIONS(2852),
    [anon_sym_id] = ACTIONS(2852),
    [anon_sym_ATdefs] = ACTIONS(2854),
  },
  [STATE(2212)] = {
    [sym_identifier] = ACTIONS(5495),
    [aux_sym_preproc_def_token1] = ACTIONS(5495),
    [aux_sym_preproc_if_token1] = ACTIONS(5495),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5495),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5495),
    [sym_preproc_directive] = ACTIONS(5495),
    [anon_sym_DASH] = ACTIONS(5497),
    [anon_sym_PLUS] = ACTIONS(5497),
    [anon_sym_SEMI] = ACTIONS(5499),
    [anon_sym___extension__] = ACTIONS(5495),
    [anon_sym_typedef] = ACTIONS(5495),
    [anon_sym_extern] = ACTIONS(5495),
    [anon_sym___attribute__] = ACTIONS(5495),
    [anon_sym___attribute] = ACTIONS(5495),
    [anon_sym_noreturn] = ACTIONS(5495),
    [anon_sym_LBRACK] = ACTIONS(5497),
    [anon_sym___declspec] = ACTIONS(5495),
    [anon_sym___cdecl] = ACTIONS(5495),
    [anon_sym___clrcall] = ACTIONS(5495),
    [anon_sym___stdcall] = ACTIONS(5495),
    [anon_sym___fastcall] = ACTIONS(5495),
    [anon_sym___thiscall] = ACTIONS(5495),
    [anon_sym___vectorcall] = ACTIONS(5495),
    [anon_sym_signed] = ACTIONS(5495),
    [anon_sym_unsigned] = ACTIONS(5495),
    [anon_sym_long] = ACTIONS(5495),
    [anon_sym_short] = ACTIONS(5495),
    [anon_sym_static] = ACTIONS(5495),
    [anon_sym_auto] = ACTIONS(5495),
    [anon_sym_register] = ACTIONS(5495),
    [anon_sym_inline] = ACTIONS(5495),
    [anon_sym___inline] = ACTIONS(5495),
    [anon_sym___inline__] = ACTIONS(5495),
    [anon_sym___forceinline] = ACTIONS(5495),
    [anon_sym_thread_local] = ACTIONS(5495),
    [anon_sym___thread] = ACTIONS(5495),
    [anon_sym_CG_EXTERN] = ACTIONS(5495),
    [anon_sym_CG_INLINE] = ACTIONS(5495),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5495),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5495),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5495),
    [anon_sym_IBOutlet] = ACTIONS(5495),
    [anon_sym_IBInspectable] = ACTIONS(5495),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5495),
    [anon_sym_NS_INLINE] = ACTIONS(5495),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5495),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5495),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5495),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5495),
    [anon_sym_const] = ACTIONS(5495),
    [anon_sym_constexpr] = ACTIONS(5495),
    [anon_sym_volatile] = ACTIONS(5495),
    [anon_sym_restrict] = ACTIONS(5495),
    [anon_sym___restrict__] = ACTIONS(5495),
    [anon_sym__Atomic] = ACTIONS(5495),
    [anon_sym__Noreturn] = ACTIONS(5495),
    [anon_sym__Nonnull] = ACTIONS(5495),
    [anon_sym_nullable] = ACTIONS(5495),
    [anon_sym__Complex] = ACTIONS(5495),
    [anon_sym__Nullable] = ACTIONS(5495),
    [anon_sym__Nullable_result] = ACTIONS(5495),
    [anon_sym__Null_unspecified] = ACTIONS(5495),
    [anon_sym___autoreleasing] = ACTIONS(5495),
    [anon_sym___block] = ACTIONS(5495),
    [anon_sym___bridge] = ACTIONS(5495),
    [anon_sym___bridge_retained] = ACTIONS(5495),
    [anon_sym___bridge_transfer] = ACTIONS(5495),
    [anon_sym___complex] = ACTIONS(5495),
    [anon_sym___const] = ACTIONS(5495),
    [anon_sym___imag] = ACTIONS(5495),
    [anon_sym___kindof] = ACTIONS(5495),
    [anon_sym___nonnull] = ACTIONS(5495),
    [anon_sym___nullable] = ACTIONS(5495),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5495),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5495),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5495),
    [anon_sym___real] = ACTIONS(5495),
    [anon_sym___strong] = ACTIONS(5495),
    [anon_sym___unsafe_unretained] = ACTIONS(5495),
    [anon_sym___unused] = ACTIONS(5495),
    [anon_sym___weak] = ACTIONS(5495),
    [anon_sym_alignas] = ACTIONS(5495),
    [anon_sym__Alignas] = ACTIONS(5495),
    [sym_primitive_type] = ACTIONS(5495),
    [anon_sym_enum] = ACTIONS(5495),
    [anon_sym_struct] = ACTIONS(5495),
    [anon_sym_union] = ACTIONS(5495),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5495),
    [anon_sym___typeof] = ACTIONS(5495),
    [anon_sym_typeof] = ACTIONS(5495),
    [aux_sym_preproc_undef_token1] = ACTIONS(5495),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5495),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5495),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5495),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5495),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5495),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5495),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5495),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5495),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5495),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5495),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5495),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5495),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5495),
    [anon_sym_API_AVAILABLE] = ACTIONS(5495),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5495),
    [anon_sym_API_DEPRECATED] = ACTIONS(5495),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5495),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5495),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5495),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5495),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5495),
    [anon_sym___deprecated_msg] = ACTIONS(5495),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5495),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5495),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5495),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5495),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5495),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5495),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5495),
    [anon_sym_ATend] = ACTIONS(5497),
    [anon_sym_ATsynthesize] = ACTIONS(5497),
    [anon_sym_ATdynamic] = ACTIONS(5497),
    [anon_sym_ATproperty] = ACTIONS(5497),
    [anon_sym_BOOL] = ACTIONS(5495),
    [anon_sym_IMP] = ACTIONS(5495),
    [anon_sym_SEL] = ACTIONS(5495),
    [anon_sym_Class] = ACTIONS(5495),
    [anon_sym_id] = ACTIONS(5495),
  },
  [STATE(2213)] = {
    [sym_identifier] = ACTIONS(2626),
    [aux_sym_preproc_def_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token2] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2626),
    [aux_sym_preproc_else_token1] = ACTIONS(2626),
    [aux_sym_preproc_elif_token1] = ACTIONS(2626),
    [sym_preproc_directive] = ACTIONS(2626),
    [anon_sym_DASH] = ACTIONS(2628),
    [anon_sym_PLUS] = ACTIONS(2628),
    [anon_sym___extension__] = ACTIONS(2626),
    [anon_sym_typedef] = ACTIONS(2626),
    [anon_sym_extern] = ACTIONS(2626),
    [anon_sym___attribute__] = ACTIONS(2626),
    [anon_sym___attribute] = ACTIONS(2626),
    [anon_sym_noreturn] = ACTIONS(2626),
    [anon_sym_LBRACK] = ACTIONS(2628),
    [anon_sym___declspec] = ACTIONS(2626),
    [anon_sym___cdecl] = ACTIONS(2626),
    [anon_sym___clrcall] = ACTIONS(2626),
    [anon_sym___stdcall] = ACTIONS(2626),
    [anon_sym___fastcall] = ACTIONS(2626),
    [anon_sym___thiscall] = ACTIONS(2626),
    [anon_sym___vectorcall] = ACTIONS(2626),
    [anon_sym_signed] = ACTIONS(2626),
    [anon_sym_unsigned] = ACTIONS(2626),
    [anon_sym_long] = ACTIONS(2626),
    [anon_sym_short] = ACTIONS(2626),
    [anon_sym_static] = ACTIONS(2626),
    [anon_sym_auto] = ACTIONS(2626),
    [anon_sym_register] = ACTIONS(2626),
    [anon_sym_inline] = ACTIONS(2626),
    [anon_sym___inline] = ACTIONS(2626),
    [anon_sym___inline__] = ACTIONS(2626),
    [anon_sym___forceinline] = ACTIONS(2626),
    [anon_sym_thread_local] = ACTIONS(2626),
    [anon_sym___thread] = ACTIONS(2626),
    [anon_sym_CG_EXTERN] = ACTIONS(2626),
    [anon_sym_CG_INLINE] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2626),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2626),
    [anon_sym_IBOutlet] = ACTIONS(2626),
    [anon_sym_IBInspectable] = ACTIONS(2626),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2626),
    [anon_sym_NS_INLINE] = ACTIONS(2626),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2626),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2626),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2626),
    [anon_sym_const] = ACTIONS(2626),
    [anon_sym_constexpr] = ACTIONS(2626),
    [anon_sym_volatile] = ACTIONS(2626),
    [anon_sym_restrict] = ACTIONS(2626),
    [anon_sym___restrict__] = ACTIONS(2626),
    [anon_sym__Atomic] = ACTIONS(2626),
    [anon_sym__Noreturn] = ACTIONS(2626),
    [anon_sym__Nonnull] = ACTIONS(2626),
    [anon_sym_nullable] = ACTIONS(2626),
    [anon_sym__Complex] = ACTIONS(2626),
    [anon_sym__Nullable] = ACTIONS(2626),
    [anon_sym__Nullable_result] = ACTIONS(2626),
    [anon_sym__Null_unspecified] = ACTIONS(2626),
    [anon_sym___autoreleasing] = ACTIONS(2626),
    [anon_sym___block] = ACTIONS(2626),
    [anon_sym___bridge] = ACTIONS(2626),
    [anon_sym___bridge_retained] = ACTIONS(2626),
    [anon_sym___bridge_transfer] = ACTIONS(2626),
    [anon_sym___complex] = ACTIONS(2626),
    [anon_sym___const] = ACTIONS(2626),
    [anon_sym___imag] = ACTIONS(2626),
    [anon_sym___kindof] = ACTIONS(2626),
    [anon_sym___nonnull] = ACTIONS(2626),
    [anon_sym___nullable] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2626),
    [anon_sym___real] = ACTIONS(2626),
    [anon_sym___strong] = ACTIONS(2626),
    [anon_sym___unsafe_unretained] = ACTIONS(2626),
    [anon_sym___unused] = ACTIONS(2626),
    [anon_sym___weak] = ACTIONS(2626),
    [anon_sym_alignas] = ACTIONS(2626),
    [anon_sym__Alignas] = ACTIONS(2626),
    [sym_primitive_type] = ACTIONS(2626),
    [anon_sym_enum] = ACTIONS(2626),
    [anon_sym_struct] = ACTIONS(2626),
    [anon_sym_union] = ACTIONS(2626),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2626),
    [anon_sym___typeof] = ACTIONS(2626),
    [anon_sym_typeof] = ACTIONS(2626),
    [aux_sym_preproc_undef_token1] = ACTIONS(2626),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2626),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2626),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2626),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_API_AVAILABLE] = ACTIONS(2626),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_API_DEPRECATED] = ACTIONS(2626),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym___deprecated_msg] = ACTIONS(2626),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2626),
    [anon_sym_ATsynthesize] = ACTIONS(2628),
    [anon_sym_ATdynamic] = ACTIONS(2628),
    [anon_sym_BOOL] = ACTIONS(2626),
    [anon_sym_IMP] = ACTIONS(2626),
    [anon_sym_SEL] = ACTIONS(2626),
    [anon_sym_Class] = ACTIONS(2626),
    [anon_sym_id] = ACTIONS(2626),
  },
  [STATE(2214)] = {
    [sym_parameterized_arguments] = STATE(2244),
    [sym_identifier] = ACTIONS(5501),
    [aux_sym_preproc_def_token1] = ACTIONS(5501),
    [aux_sym_preproc_if_token1] = ACTIONS(5501),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5501),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5501),
    [sym_preproc_directive] = ACTIONS(5501),
    [anon_sym_DASH] = ACTIONS(5503),
    [anon_sym_PLUS] = ACTIONS(5503),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(5501),
    [anon_sym_typedef] = ACTIONS(5501),
    [anon_sym_extern] = ACTIONS(5501),
    [anon_sym___attribute__] = ACTIONS(5501),
    [anon_sym___attribute] = ACTIONS(5501),
    [anon_sym_noreturn] = ACTIONS(5501),
    [anon_sym_LBRACK] = ACTIONS(5503),
    [anon_sym___declspec] = ACTIONS(5501),
    [anon_sym___cdecl] = ACTIONS(5501),
    [anon_sym___clrcall] = ACTIONS(5501),
    [anon_sym___stdcall] = ACTIONS(5501),
    [anon_sym___fastcall] = ACTIONS(5501),
    [anon_sym___thiscall] = ACTIONS(5501),
    [anon_sym___vectorcall] = ACTIONS(5501),
    [anon_sym_LBRACE] = ACTIONS(5503),
    [anon_sym_signed] = ACTIONS(5501),
    [anon_sym_unsigned] = ACTIONS(5501),
    [anon_sym_long] = ACTIONS(5501),
    [anon_sym_short] = ACTIONS(5501),
    [anon_sym_static] = ACTIONS(5501),
    [anon_sym_auto] = ACTIONS(5501),
    [anon_sym_register] = ACTIONS(5501),
    [anon_sym_inline] = ACTIONS(5501),
    [anon_sym___inline] = ACTIONS(5501),
    [anon_sym___inline__] = ACTIONS(5501),
    [anon_sym___forceinline] = ACTIONS(5501),
    [anon_sym_thread_local] = ACTIONS(5501),
    [anon_sym___thread] = ACTIONS(5501),
    [anon_sym_CG_EXTERN] = ACTIONS(5501),
    [anon_sym_CG_INLINE] = ACTIONS(5501),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5501),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5501),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5501),
    [anon_sym_IBOutlet] = ACTIONS(5501),
    [anon_sym_IBInspectable] = ACTIONS(5501),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5501),
    [anon_sym_NS_INLINE] = ACTIONS(5501),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5501),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5501),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5501),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5501),
    [anon_sym_const] = ACTIONS(5501),
    [anon_sym_constexpr] = ACTIONS(5501),
    [anon_sym_volatile] = ACTIONS(5501),
    [anon_sym_restrict] = ACTIONS(5501),
    [anon_sym___restrict__] = ACTIONS(5501),
    [anon_sym__Atomic] = ACTIONS(5501),
    [anon_sym__Noreturn] = ACTIONS(5501),
    [anon_sym__Nonnull] = ACTIONS(5501),
    [anon_sym_nullable] = ACTIONS(5501),
    [anon_sym__Complex] = ACTIONS(5501),
    [anon_sym__Nullable] = ACTIONS(5501),
    [anon_sym__Nullable_result] = ACTIONS(5501),
    [anon_sym__Null_unspecified] = ACTIONS(5501),
    [anon_sym___autoreleasing] = ACTIONS(5501),
    [anon_sym___block] = ACTIONS(5501),
    [anon_sym___bridge] = ACTIONS(5501),
    [anon_sym___bridge_retained] = ACTIONS(5501),
    [anon_sym___bridge_transfer] = ACTIONS(5501),
    [anon_sym___complex] = ACTIONS(5501),
    [anon_sym___const] = ACTIONS(5501),
    [anon_sym___imag] = ACTIONS(5501),
    [anon_sym___kindof] = ACTIONS(5501),
    [anon_sym___nonnull] = ACTIONS(5501),
    [anon_sym___nullable] = ACTIONS(5501),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5501),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5501),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5501),
    [anon_sym___real] = ACTIONS(5501),
    [anon_sym___strong] = ACTIONS(5501),
    [anon_sym___unsafe_unretained] = ACTIONS(5501),
    [anon_sym___unused] = ACTIONS(5501),
    [anon_sym___weak] = ACTIONS(5501),
    [anon_sym_alignas] = ACTIONS(5501),
    [anon_sym__Alignas] = ACTIONS(5501),
    [sym_primitive_type] = ACTIONS(5501),
    [anon_sym_enum] = ACTIONS(5501),
    [anon_sym_struct] = ACTIONS(5501),
    [anon_sym_union] = ACTIONS(5501),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5501),
    [anon_sym___typeof] = ACTIONS(5501),
    [anon_sym_typeof] = ACTIONS(5501),
    [aux_sym_preproc_undef_token1] = ACTIONS(5501),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5501),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5501),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5501),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5501),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5501),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5501),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5501),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5501),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5501),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5501),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5501),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5501),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5501),
    [anon_sym_API_AVAILABLE] = ACTIONS(5501),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5501),
    [anon_sym_API_DEPRECATED] = ACTIONS(5501),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5501),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5501),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5501),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5501),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5501),
    [anon_sym___deprecated_msg] = ACTIONS(5501),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5501),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5501),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5501),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5501),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5501),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5501),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5501),
    [anon_sym_ATend] = ACTIONS(5503),
    [anon_sym_ATproperty] = ACTIONS(5503),
    [anon_sym_BOOL] = ACTIONS(5501),
    [anon_sym_IMP] = ACTIONS(5501),
    [anon_sym_SEL] = ACTIONS(5501),
    [anon_sym_Class] = ACTIONS(5501),
    [anon_sym_id] = ACTIONS(5501),
  },
  [STATE(2215)] = {
    [sym_identifier] = ACTIONS(2360),
    [aux_sym_preproc_def_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token1] = ACTIONS(2360),
    [aux_sym_preproc_if_token2] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2360),
    [sym_preproc_directive] = ACTIONS(2360),
    [anon_sym_DASH] = ACTIONS(2362),
    [anon_sym_PLUS] = ACTIONS(2362),
    [anon_sym_SEMI] = ACTIONS(2362),
    [anon_sym___extension__] = ACTIONS(2360),
    [anon_sym_typedef] = ACTIONS(2360),
    [anon_sym_extern] = ACTIONS(2360),
    [anon_sym___attribute__] = ACTIONS(2360),
    [anon_sym___attribute] = ACTIONS(2360),
    [anon_sym_noreturn] = ACTIONS(2360),
    [anon_sym_LBRACK] = ACTIONS(2362),
    [anon_sym___declspec] = ACTIONS(2360),
    [anon_sym___cdecl] = ACTIONS(2360),
    [anon_sym___clrcall] = ACTIONS(2360),
    [anon_sym___stdcall] = ACTIONS(2360),
    [anon_sym___fastcall] = ACTIONS(2360),
    [anon_sym___thiscall] = ACTIONS(2360),
    [anon_sym___vectorcall] = ACTIONS(2360),
    [anon_sym_signed] = ACTIONS(2360),
    [anon_sym_unsigned] = ACTIONS(2360),
    [anon_sym_long] = ACTIONS(2360),
    [anon_sym_short] = ACTIONS(2360),
    [anon_sym_static] = ACTIONS(2360),
    [anon_sym_auto] = ACTIONS(2360),
    [anon_sym_register] = ACTIONS(2360),
    [anon_sym_inline] = ACTIONS(2360),
    [anon_sym___inline] = ACTIONS(2360),
    [anon_sym___inline__] = ACTIONS(2360),
    [anon_sym___forceinline] = ACTIONS(2360),
    [anon_sym_thread_local] = ACTIONS(2360),
    [anon_sym___thread] = ACTIONS(2360),
    [anon_sym_CG_EXTERN] = ACTIONS(2360),
    [anon_sym_CG_INLINE] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2360),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2360),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2360),
    [anon_sym_IBOutlet] = ACTIONS(2360),
    [anon_sym_IBInspectable] = ACTIONS(2360),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2360),
    [anon_sym_NS_INLINE] = ACTIONS(2360),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2360),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2360),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2360),
    [anon_sym_const] = ACTIONS(2360),
    [anon_sym_constexpr] = ACTIONS(2360),
    [anon_sym_volatile] = ACTIONS(2360),
    [anon_sym_restrict] = ACTIONS(2360),
    [anon_sym___restrict__] = ACTIONS(2360),
    [anon_sym__Atomic] = ACTIONS(2360),
    [anon_sym__Noreturn] = ACTIONS(2360),
    [anon_sym__Nonnull] = ACTIONS(2360),
    [anon_sym_nullable] = ACTIONS(2360),
    [anon_sym__Complex] = ACTIONS(2360),
    [anon_sym__Nullable] = ACTIONS(2360),
    [anon_sym__Nullable_result] = ACTIONS(2360),
    [anon_sym__Null_unspecified] = ACTIONS(2360),
    [anon_sym___autoreleasing] = ACTIONS(2360),
    [anon_sym___block] = ACTIONS(2360),
    [anon_sym___bridge] = ACTIONS(2360),
    [anon_sym___bridge_retained] = ACTIONS(2360),
    [anon_sym___bridge_transfer] = ACTIONS(2360),
    [anon_sym___complex] = ACTIONS(2360),
    [anon_sym___const] = ACTIONS(2360),
    [anon_sym___imag] = ACTIONS(2360),
    [anon_sym___kindof] = ACTIONS(2360),
    [anon_sym___nonnull] = ACTIONS(2360),
    [anon_sym___nullable] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2360),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2360),
    [anon_sym___real] = ACTIONS(2360),
    [anon_sym___strong] = ACTIONS(2360),
    [anon_sym___unsafe_unretained] = ACTIONS(2360),
    [anon_sym___unused] = ACTIONS(2360),
    [anon_sym___weak] = ACTIONS(2360),
    [anon_sym_alignas] = ACTIONS(2360),
    [anon_sym__Alignas] = ACTIONS(2360),
    [sym_primitive_type] = ACTIONS(2360),
    [anon_sym_enum] = ACTIONS(2360),
    [anon_sym_struct] = ACTIONS(2360),
    [anon_sym_union] = ACTIONS(2360),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2360),
    [anon_sym___typeof] = ACTIONS(2360),
    [anon_sym_typeof] = ACTIONS(2360),
    [aux_sym_preproc_undef_token1] = ACTIONS(2360),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2360),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2360),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2360),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2360),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2360),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_API_AVAILABLE] = ACTIONS(2360),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_API_DEPRECATED] = ACTIONS(2360),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2360),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2360),
    [anon_sym___deprecated_msg] = ACTIONS(2360),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2360),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2360),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2360),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2360),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2360),
    [anon_sym_ATsynthesize] = ACTIONS(2362),
    [anon_sym_ATdynamic] = ACTIONS(2362),
    [anon_sym_ATproperty] = ACTIONS(2362),
    [anon_sym_BOOL] = ACTIONS(2360),
    [anon_sym_IMP] = ACTIONS(2360),
    [anon_sym_SEL] = ACTIONS(2360),
    [anon_sym_Class] = ACTIONS(2360),
    [anon_sym_id] = ACTIONS(2360),
  },
  [STATE(2216)] = {
    [sym_identifier] = ACTIONS(2608),
    [aux_sym_preproc_def_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token2] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2608),
    [sym_preproc_directive] = ACTIONS(2608),
    [anon_sym_DASH] = ACTIONS(2610),
    [anon_sym_PLUS] = ACTIONS(2610),
    [anon_sym___extension__] = ACTIONS(2608),
    [anon_sym_typedef] = ACTIONS(2608),
    [anon_sym_extern] = ACTIONS(2608),
    [anon_sym___attribute__] = ACTIONS(2608),
    [anon_sym___attribute] = ACTIONS(2608),
    [anon_sym_noreturn] = ACTIONS(2608),
    [anon_sym_LBRACK] = ACTIONS(2610),
    [anon_sym___declspec] = ACTIONS(2608),
    [anon_sym___cdecl] = ACTIONS(2608),
    [anon_sym___clrcall] = ACTIONS(2608),
    [anon_sym___stdcall] = ACTIONS(2608),
    [anon_sym___fastcall] = ACTIONS(2608),
    [anon_sym___thiscall] = ACTIONS(2608),
    [anon_sym___vectorcall] = ACTIONS(2608),
    [anon_sym_signed] = ACTIONS(2608),
    [anon_sym_unsigned] = ACTIONS(2608),
    [anon_sym_long] = ACTIONS(2608),
    [anon_sym_short] = ACTIONS(2608),
    [anon_sym_static] = ACTIONS(2608),
    [anon_sym_auto] = ACTIONS(2608),
    [anon_sym_register] = ACTIONS(2608),
    [anon_sym_inline] = ACTIONS(2608),
    [anon_sym___inline] = ACTIONS(2608),
    [anon_sym___inline__] = ACTIONS(2608),
    [anon_sym___forceinline] = ACTIONS(2608),
    [anon_sym_thread_local] = ACTIONS(2608),
    [anon_sym___thread] = ACTIONS(2608),
    [anon_sym_CG_EXTERN] = ACTIONS(2608),
    [anon_sym_CG_INLINE] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2608),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2608),
    [anon_sym_IBOutlet] = ACTIONS(2608),
    [anon_sym_IBInspectable] = ACTIONS(2608),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2608),
    [anon_sym_NS_INLINE] = ACTIONS(2608),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2608),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2608),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2608),
    [anon_sym_const] = ACTIONS(2608),
    [anon_sym_constexpr] = ACTIONS(2608),
    [anon_sym_volatile] = ACTIONS(2608),
    [anon_sym_restrict] = ACTIONS(2608),
    [anon_sym___restrict__] = ACTIONS(2608),
    [anon_sym__Atomic] = ACTIONS(2608),
    [anon_sym__Noreturn] = ACTIONS(2608),
    [anon_sym__Nonnull] = ACTIONS(2608),
    [anon_sym_nullable] = ACTIONS(2608),
    [anon_sym__Complex] = ACTIONS(2608),
    [anon_sym__Nullable] = ACTIONS(2608),
    [anon_sym__Nullable_result] = ACTIONS(2608),
    [anon_sym__Null_unspecified] = ACTIONS(2608),
    [anon_sym___autoreleasing] = ACTIONS(2608),
    [anon_sym___block] = ACTIONS(2608),
    [anon_sym___bridge] = ACTIONS(2608),
    [anon_sym___bridge_retained] = ACTIONS(2608),
    [anon_sym___bridge_transfer] = ACTIONS(2608),
    [anon_sym___complex] = ACTIONS(2608),
    [anon_sym___const] = ACTIONS(2608),
    [anon_sym___imag] = ACTIONS(2608),
    [anon_sym___kindof] = ACTIONS(2608),
    [anon_sym___nonnull] = ACTIONS(2608),
    [anon_sym___nullable] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2608),
    [anon_sym___real] = ACTIONS(2608),
    [anon_sym___strong] = ACTIONS(2608),
    [anon_sym___unsafe_unretained] = ACTIONS(2608),
    [anon_sym___unused] = ACTIONS(2608),
    [anon_sym___weak] = ACTIONS(2608),
    [anon_sym_alignas] = ACTIONS(2608),
    [anon_sym__Alignas] = ACTIONS(2608),
    [sym_primitive_type] = ACTIONS(2608),
    [anon_sym_enum] = ACTIONS(2608),
    [anon_sym_struct] = ACTIONS(2608),
    [anon_sym_union] = ACTIONS(2608),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2608),
    [anon_sym___typeof] = ACTIONS(2608),
    [anon_sym_typeof] = ACTIONS(2608),
    [aux_sym_preproc_undef_token1] = ACTIONS(2608),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2608),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2608),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2608),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_API_AVAILABLE] = ACTIONS(2608),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_API_DEPRECATED] = ACTIONS(2608),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym___deprecated_msg] = ACTIONS(2608),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2608),
    [anon_sym_ATsynthesize] = ACTIONS(2610),
    [anon_sym_ATdynamic] = ACTIONS(2610),
    [anon_sym_ATproperty] = ACTIONS(2610),
    [anon_sym_BOOL] = ACTIONS(2608),
    [anon_sym_IMP] = ACTIONS(2608),
    [anon_sym_SEL] = ACTIONS(2608),
    [anon_sym_Class] = ACTIONS(2608),
    [anon_sym_id] = ACTIONS(2608),
    [anon_sym_ATdefs] = ACTIONS(2610),
  },
  [STATE(2217)] = {
    [sym_identifier] = ACTIONS(2630),
    [aux_sym_preproc_def_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token2] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2630),
    [aux_sym_preproc_else_token1] = ACTIONS(2630),
    [aux_sym_preproc_elif_token1] = ACTIONS(2630),
    [sym_preproc_directive] = ACTIONS(2630),
    [anon_sym_DASH] = ACTIONS(2632),
    [anon_sym_PLUS] = ACTIONS(2632),
    [anon_sym___extension__] = ACTIONS(2630),
    [anon_sym_typedef] = ACTIONS(2630),
    [anon_sym_extern] = ACTIONS(2630),
    [anon_sym___attribute__] = ACTIONS(2630),
    [anon_sym___attribute] = ACTIONS(2630),
    [anon_sym_noreturn] = ACTIONS(2630),
    [anon_sym_LBRACK] = ACTIONS(2632),
    [anon_sym___declspec] = ACTIONS(2630),
    [anon_sym___cdecl] = ACTIONS(2630),
    [anon_sym___clrcall] = ACTIONS(2630),
    [anon_sym___stdcall] = ACTIONS(2630),
    [anon_sym___fastcall] = ACTIONS(2630),
    [anon_sym___thiscall] = ACTIONS(2630),
    [anon_sym___vectorcall] = ACTIONS(2630),
    [anon_sym_signed] = ACTIONS(2630),
    [anon_sym_unsigned] = ACTIONS(2630),
    [anon_sym_long] = ACTIONS(2630),
    [anon_sym_short] = ACTIONS(2630),
    [anon_sym_static] = ACTIONS(2630),
    [anon_sym_auto] = ACTIONS(2630),
    [anon_sym_register] = ACTIONS(2630),
    [anon_sym_inline] = ACTIONS(2630),
    [anon_sym___inline] = ACTIONS(2630),
    [anon_sym___inline__] = ACTIONS(2630),
    [anon_sym___forceinline] = ACTIONS(2630),
    [anon_sym_thread_local] = ACTIONS(2630),
    [anon_sym___thread] = ACTIONS(2630),
    [anon_sym_CG_EXTERN] = ACTIONS(2630),
    [anon_sym_CG_INLINE] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2630),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2630),
    [anon_sym_IBOutlet] = ACTIONS(2630),
    [anon_sym_IBInspectable] = ACTIONS(2630),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2630),
    [anon_sym_NS_INLINE] = ACTIONS(2630),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2630),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2630),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2630),
    [anon_sym_const] = ACTIONS(2630),
    [anon_sym_constexpr] = ACTIONS(2630),
    [anon_sym_volatile] = ACTIONS(2630),
    [anon_sym_restrict] = ACTIONS(2630),
    [anon_sym___restrict__] = ACTIONS(2630),
    [anon_sym__Atomic] = ACTIONS(2630),
    [anon_sym__Noreturn] = ACTIONS(2630),
    [anon_sym__Nonnull] = ACTIONS(2630),
    [anon_sym_nullable] = ACTIONS(2630),
    [anon_sym__Complex] = ACTIONS(2630),
    [anon_sym__Nullable] = ACTIONS(2630),
    [anon_sym__Nullable_result] = ACTIONS(2630),
    [anon_sym__Null_unspecified] = ACTIONS(2630),
    [anon_sym___autoreleasing] = ACTIONS(2630),
    [anon_sym___block] = ACTIONS(2630),
    [anon_sym___bridge] = ACTIONS(2630),
    [anon_sym___bridge_retained] = ACTIONS(2630),
    [anon_sym___bridge_transfer] = ACTIONS(2630),
    [anon_sym___complex] = ACTIONS(2630),
    [anon_sym___const] = ACTIONS(2630),
    [anon_sym___imag] = ACTIONS(2630),
    [anon_sym___kindof] = ACTIONS(2630),
    [anon_sym___nonnull] = ACTIONS(2630),
    [anon_sym___nullable] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2630),
    [anon_sym___real] = ACTIONS(2630),
    [anon_sym___strong] = ACTIONS(2630),
    [anon_sym___unsafe_unretained] = ACTIONS(2630),
    [anon_sym___unused] = ACTIONS(2630),
    [anon_sym___weak] = ACTIONS(2630),
    [anon_sym_alignas] = ACTIONS(2630),
    [anon_sym__Alignas] = ACTIONS(2630),
    [sym_primitive_type] = ACTIONS(2630),
    [anon_sym_enum] = ACTIONS(2630),
    [anon_sym_struct] = ACTIONS(2630),
    [anon_sym_union] = ACTIONS(2630),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2630),
    [anon_sym___typeof] = ACTIONS(2630),
    [anon_sym_typeof] = ACTIONS(2630),
    [aux_sym_preproc_undef_token1] = ACTIONS(2630),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2630),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2630),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2630),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_API_AVAILABLE] = ACTIONS(2630),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_API_DEPRECATED] = ACTIONS(2630),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym___deprecated_msg] = ACTIONS(2630),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2630),
    [anon_sym_ATsynthesize] = ACTIONS(2632),
    [anon_sym_ATdynamic] = ACTIONS(2632),
    [anon_sym_BOOL] = ACTIONS(2630),
    [anon_sym_IMP] = ACTIONS(2630),
    [anon_sym_SEL] = ACTIONS(2630),
    [anon_sym_Class] = ACTIONS(2630),
    [anon_sym_id] = ACTIONS(2630),
  },
  [STATE(2218)] = {
    [sym_identifier] = ACTIONS(5505),
    [aux_sym_preproc_def_token1] = ACTIONS(5505),
    [aux_sym_preproc_if_token1] = ACTIONS(5505),
    [aux_sym_preproc_if_token2] = ACTIONS(5505),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5505),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5505),
    [aux_sym_preproc_else_token1] = ACTIONS(5505),
    [aux_sym_preproc_elif_token1] = ACTIONS(5505),
    [sym_preproc_directive] = ACTIONS(5505),
    [anon_sym_DASH] = ACTIONS(5507),
    [anon_sym_PLUS] = ACTIONS(5507),
    [anon_sym___extension__] = ACTIONS(5505),
    [anon_sym_typedef] = ACTIONS(5505),
    [anon_sym_extern] = ACTIONS(5505),
    [anon_sym___attribute__] = ACTIONS(5505),
    [anon_sym___attribute] = ACTIONS(5505),
    [anon_sym_noreturn] = ACTIONS(5505),
    [anon_sym_LBRACK] = ACTIONS(5507),
    [anon_sym___declspec] = ACTIONS(5505),
    [anon_sym___cdecl] = ACTIONS(5505),
    [anon_sym___clrcall] = ACTIONS(5505),
    [anon_sym___stdcall] = ACTIONS(5505),
    [anon_sym___fastcall] = ACTIONS(5505),
    [anon_sym___thiscall] = ACTIONS(5505),
    [anon_sym___vectorcall] = ACTIONS(5505),
    [anon_sym_signed] = ACTIONS(5505),
    [anon_sym_unsigned] = ACTIONS(5505),
    [anon_sym_long] = ACTIONS(5505),
    [anon_sym_short] = ACTIONS(5505),
    [anon_sym_static] = ACTIONS(5505),
    [anon_sym_auto] = ACTIONS(5505),
    [anon_sym_register] = ACTIONS(5505),
    [anon_sym_inline] = ACTIONS(5505),
    [anon_sym___inline] = ACTIONS(5505),
    [anon_sym___inline__] = ACTIONS(5505),
    [anon_sym___forceinline] = ACTIONS(5505),
    [anon_sym_thread_local] = ACTIONS(5505),
    [anon_sym___thread] = ACTIONS(5505),
    [anon_sym_CG_EXTERN] = ACTIONS(5505),
    [anon_sym_CG_INLINE] = ACTIONS(5505),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5505),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5505),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5505),
    [anon_sym_IBOutlet] = ACTIONS(5505),
    [anon_sym_IBInspectable] = ACTIONS(5505),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5505),
    [anon_sym_NS_INLINE] = ACTIONS(5505),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5505),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5505),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5505),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5505),
    [anon_sym_const] = ACTIONS(5505),
    [anon_sym_constexpr] = ACTIONS(5505),
    [anon_sym_volatile] = ACTIONS(5505),
    [anon_sym_restrict] = ACTIONS(5505),
    [anon_sym___restrict__] = ACTIONS(5505),
    [anon_sym__Atomic] = ACTIONS(5505),
    [anon_sym__Noreturn] = ACTIONS(5505),
    [anon_sym__Nonnull] = ACTIONS(5505),
    [anon_sym_nullable] = ACTIONS(5505),
    [anon_sym__Complex] = ACTIONS(5505),
    [anon_sym__Nullable] = ACTIONS(5505),
    [anon_sym__Nullable_result] = ACTIONS(5505),
    [anon_sym__Null_unspecified] = ACTIONS(5505),
    [anon_sym___autoreleasing] = ACTIONS(5505),
    [anon_sym___block] = ACTIONS(5505),
    [anon_sym___bridge] = ACTIONS(5505),
    [anon_sym___bridge_retained] = ACTIONS(5505),
    [anon_sym___bridge_transfer] = ACTIONS(5505),
    [anon_sym___complex] = ACTIONS(5505),
    [anon_sym___const] = ACTIONS(5505),
    [anon_sym___imag] = ACTIONS(5505),
    [anon_sym___kindof] = ACTIONS(5505),
    [anon_sym___nonnull] = ACTIONS(5505),
    [anon_sym___nullable] = ACTIONS(5505),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5505),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5505),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5505),
    [anon_sym___real] = ACTIONS(5505),
    [anon_sym___strong] = ACTIONS(5505),
    [anon_sym___unsafe_unretained] = ACTIONS(5505),
    [anon_sym___unused] = ACTIONS(5505),
    [anon_sym___weak] = ACTIONS(5505),
    [anon_sym_alignas] = ACTIONS(5505),
    [anon_sym__Alignas] = ACTIONS(5505),
    [sym_primitive_type] = ACTIONS(5505),
    [anon_sym_enum] = ACTIONS(5505),
    [anon_sym_struct] = ACTIONS(5505),
    [anon_sym_union] = ACTIONS(5505),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5505),
    [anon_sym___typeof] = ACTIONS(5505),
    [anon_sym_typeof] = ACTIONS(5505),
    [aux_sym_preproc_undef_token1] = ACTIONS(5505),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5505),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5505),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5505),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5505),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5505),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5505),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5505),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5505),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5505),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_API_AVAILABLE] = ACTIONS(5505),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_API_DEPRECATED] = ACTIONS(5505),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5505),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5505),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5505),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5505),
    [anon_sym___deprecated_msg] = ACTIONS(5505),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5505),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5505),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5505),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5505),
    [anon_sym_ATsynthesize] = ACTIONS(5507),
    [anon_sym_ATdynamic] = ACTIONS(5507),
    [anon_sym_BOOL] = ACTIONS(5505),
    [anon_sym_IMP] = ACTIONS(5505),
    [anon_sym_SEL] = ACTIONS(5505),
    [anon_sym_Class] = ACTIONS(5505),
    [anon_sym_id] = ACTIONS(5505),
  },
  [STATE(2219)] = {
    [sym__declaration_modifiers] = STATE(2566),
    [sym_attribute_specifier] = STATE(2566),
    [sym_attribute_declaration] = STATE(2566),
    [sym_ms_declspec_modifier] = STATE(2566),
    [sym_ms_based_modifier] = STATE(6198),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2579),
    [sym__declarator] = STATE(5024),
    [sym__abstract_declarator] = STATE(4978),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(2566),
    [sym_type_qualifier] = STATE(2566),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4271),
    [sym_availability_attribute_specifier] = STATE(2566),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2566),
    [aux_sym_pointer_declarator_repeat1] = STATE(2579),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_RPAREN] = ACTIONS(5314),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(4086),
    [anon_sym_CARET] = ACTIONS(4088),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2220)] = {
    [sym_identifier] = ACTIONS(5375),
    [aux_sym_preproc_def_token1] = ACTIONS(5375),
    [aux_sym_preproc_if_token1] = ACTIONS(5375),
    [aux_sym_preproc_if_token2] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5375),
    [aux_sym_preproc_else_token1] = ACTIONS(5375),
    [aux_sym_preproc_elif_token1] = ACTIONS(5375),
    [sym_preproc_directive] = ACTIONS(5375),
    [anon_sym_DASH] = ACTIONS(5377),
    [anon_sym_PLUS] = ACTIONS(5377),
    [anon_sym___extension__] = ACTIONS(5375),
    [anon_sym_typedef] = ACTIONS(5375),
    [anon_sym_extern] = ACTIONS(5375),
    [anon_sym___attribute__] = ACTIONS(5375),
    [anon_sym___attribute] = ACTIONS(5375),
    [anon_sym_noreturn] = ACTIONS(5375),
    [anon_sym_LBRACK] = ACTIONS(5377),
    [anon_sym___declspec] = ACTIONS(5375),
    [anon_sym___cdecl] = ACTIONS(5375),
    [anon_sym___clrcall] = ACTIONS(5375),
    [anon_sym___stdcall] = ACTIONS(5375),
    [anon_sym___fastcall] = ACTIONS(5375),
    [anon_sym___thiscall] = ACTIONS(5375),
    [anon_sym___vectorcall] = ACTIONS(5375),
    [anon_sym_signed] = ACTIONS(5375),
    [anon_sym_unsigned] = ACTIONS(5375),
    [anon_sym_long] = ACTIONS(5375),
    [anon_sym_short] = ACTIONS(5375),
    [anon_sym_static] = ACTIONS(5375),
    [anon_sym_auto] = ACTIONS(5375),
    [anon_sym_register] = ACTIONS(5375),
    [anon_sym_inline] = ACTIONS(5375),
    [anon_sym___inline] = ACTIONS(5375),
    [anon_sym___inline__] = ACTIONS(5375),
    [anon_sym___forceinline] = ACTIONS(5375),
    [anon_sym_thread_local] = ACTIONS(5375),
    [anon_sym___thread] = ACTIONS(5375),
    [anon_sym_CG_EXTERN] = ACTIONS(5375),
    [anon_sym_CG_INLINE] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5375),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5375),
    [anon_sym_IBOutlet] = ACTIONS(5375),
    [anon_sym_IBInspectable] = ACTIONS(5375),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5375),
    [anon_sym_NS_INLINE] = ACTIONS(5375),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5375),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5375),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5375),
    [anon_sym_const] = ACTIONS(5375),
    [anon_sym_constexpr] = ACTIONS(5375),
    [anon_sym_volatile] = ACTIONS(5375),
    [anon_sym_restrict] = ACTIONS(5375),
    [anon_sym___restrict__] = ACTIONS(5375),
    [anon_sym__Atomic] = ACTIONS(5375),
    [anon_sym__Noreturn] = ACTIONS(5375),
    [anon_sym__Nonnull] = ACTIONS(5375),
    [anon_sym_nullable] = ACTIONS(5375),
    [anon_sym__Complex] = ACTIONS(5375),
    [anon_sym__Nullable] = ACTIONS(5375),
    [anon_sym__Nullable_result] = ACTIONS(5375),
    [anon_sym__Null_unspecified] = ACTIONS(5375),
    [anon_sym___autoreleasing] = ACTIONS(5375),
    [anon_sym___block] = ACTIONS(5375),
    [anon_sym___bridge] = ACTIONS(5375),
    [anon_sym___bridge_retained] = ACTIONS(5375),
    [anon_sym___bridge_transfer] = ACTIONS(5375),
    [anon_sym___complex] = ACTIONS(5375),
    [anon_sym___const] = ACTIONS(5375),
    [anon_sym___imag] = ACTIONS(5375),
    [anon_sym___kindof] = ACTIONS(5375),
    [anon_sym___nonnull] = ACTIONS(5375),
    [anon_sym___nullable] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5375),
    [anon_sym___real] = ACTIONS(5375),
    [anon_sym___strong] = ACTIONS(5375),
    [anon_sym___unsafe_unretained] = ACTIONS(5375),
    [anon_sym___unused] = ACTIONS(5375),
    [anon_sym___weak] = ACTIONS(5375),
    [anon_sym_alignas] = ACTIONS(5375),
    [anon_sym__Alignas] = ACTIONS(5375),
    [sym_primitive_type] = ACTIONS(5375),
    [anon_sym_enum] = ACTIONS(5375),
    [anon_sym_struct] = ACTIONS(5375),
    [anon_sym_union] = ACTIONS(5375),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5375),
    [anon_sym___typeof] = ACTIONS(5375),
    [anon_sym_typeof] = ACTIONS(5375),
    [aux_sym_preproc_undef_token1] = ACTIONS(5375),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5375),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5375),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5375),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_API_AVAILABLE] = ACTIONS(5375),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_API_DEPRECATED] = ACTIONS(5375),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym___deprecated_msg] = ACTIONS(5375),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5375),
    [anon_sym_ATsynthesize] = ACTIONS(5377),
    [anon_sym_ATdynamic] = ACTIONS(5377),
    [anon_sym_BOOL] = ACTIONS(5375),
    [anon_sym_IMP] = ACTIONS(5375),
    [anon_sym_SEL] = ACTIONS(5375),
    [anon_sym_Class] = ACTIONS(5375),
    [anon_sym_id] = ACTIONS(5375),
  },
  [STATE(2221)] = {
    [sym_identifier] = ACTIONS(5448),
    [aux_sym_preproc_def_token1] = ACTIONS(5448),
    [aux_sym_preproc_if_token1] = ACTIONS(5448),
    [aux_sym_preproc_if_token2] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5448),
    [aux_sym_preproc_else_token1] = ACTIONS(5448),
    [aux_sym_preproc_elif_token1] = ACTIONS(5448),
    [sym_preproc_directive] = ACTIONS(5448),
    [anon_sym_DASH] = ACTIONS(5450),
    [anon_sym_PLUS] = ACTIONS(5450),
    [anon_sym___extension__] = ACTIONS(5448),
    [anon_sym_typedef] = ACTIONS(5448),
    [anon_sym_extern] = ACTIONS(5448),
    [anon_sym___attribute__] = ACTIONS(5448),
    [anon_sym___attribute] = ACTIONS(5448),
    [anon_sym_noreturn] = ACTIONS(5448),
    [anon_sym_LBRACK] = ACTIONS(5450),
    [anon_sym___declspec] = ACTIONS(5448),
    [anon_sym___cdecl] = ACTIONS(5448),
    [anon_sym___clrcall] = ACTIONS(5448),
    [anon_sym___stdcall] = ACTIONS(5448),
    [anon_sym___fastcall] = ACTIONS(5448),
    [anon_sym___thiscall] = ACTIONS(5448),
    [anon_sym___vectorcall] = ACTIONS(5448),
    [anon_sym_signed] = ACTIONS(5448),
    [anon_sym_unsigned] = ACTIONS(5448),
    [anon_sym_long] = ACTIONS(5448),
    [anon_sym_short] = ACTIONS(5448),
    [anon_sym_static] = ACTIONS(5448),
    [anon_sym_auto] = ACTIONS(5448),
    [anon_sym_register] = ACTIONS(5448),
    [anon_sym_inline] = ACTIONS(5448),
    [anon_sym___inline] = ACTIONS(5448),
    [anon_sym___inline__] = ACTIONS(5448),
    [anon_sym___forceinline] = ACTIONS(5448),
    [anon_sym_thread_local] = ACTIONS(5448),
    [anon_sym___thread] = ACTIONS(5448),
    [anon_sym_CG_EXTERN] = ACTIONS(5448),
    [anon_sym_CG_INLINE] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5448),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5448),
    [anon_sym_IBOutlet] = ACTIONS(5448),
    [anon_sym_IBInspectable] = ACTIONS(5448),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5448),
    [anon_sym_NS_INLINE] = ACTIONS(5448),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5448),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5448),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5448),
    [anon_sym_const] = ACTIONS(5448),
    [anon_sym_constexpr] = ACTIONS(5448),
    [anon_sym_volatile] = ACTIONS(5448),
    [anon_sym_restrict] = ACTIONS(5448),
    [anon_sym___restrict__] = ACTIONS(5448),
    [anon_sym__Atomic] = ACTIONS(5448),
    [anon_sym__Noreturn] = ACTIONS(5448),
    [anon_sym__Nonnull] = ACTIONS(5448),
    [anon_sym_nullable] = ACTIONS(5448),
    [anon_sym__Complex] = ACTIONS(5448),
    [anon_sym__Nullable] = ACTIONS(5448),
    [anon_sym__Nullable_result] = ACTIONS(5448),
    [anon_sym__Null_unspecified] = ACTIONS(5448),
    [anon_sym___autoreleasing] = ACTIONS(5448),
    [anon_sym___block] = ACTIONS(5448),
    [anon_sym___bridge] = ACTIONS(5448),
    [anon_sym___bridge_retained] = ACTIONS(5448),
    [anon_sym___bridge_transfer] = ACTIONS(5448),
    [anon_sym___complex] = ACTIONS(5448),
    [anon_sym___const] = ACTIONS(5448),
    [anon_sym___imag] = ACTIONS(5448),
    [anon_sym___kindof] = ACTIONS(5448),
    [anon_sym___nonnull] = ACTIONS(5448),
    [anon_sym___nullable] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5448),
    [anon_sym___real] = ACTIONS(5448),
    [anon_sym___strong] = ACTIONS(5448),
    [anon_sym___unsafe_unretained] = ACTIONS(5448),
    [anon_sym___unused] = ACTIONS(5448),
    [anon_sym___weak] = ACTIONS(5448),
    [anon_sym_alignas] = ACTIONS(5448),
    [anon_sym__Alignas] = ACTIONS(5448),
    [sym_primitive_type] = ACTIONS(5448),
    [anon_sym_enum] = ACTIONS(5448),
    [anon_sym_struct] = ACTIONS(5448),
    [anon_sym_union] = ACTIONS(5448),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5448),
    [anon_sym___typeof] = ACTIONS(5448),
    [anon_sym_typeof] = ACTIONS(5448),
    [aux_sym_preproc_undef_token1] = ACTIONS(5448),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5448),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5448),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5448),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_API_AVAILABLE] = ACTIONS(5448),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_API_DEPRECATED] = ACTIONS(5448),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym___deprecated_msg] = ACTIONS(5448),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5448),
    [anon_sym_ATsynthesize] = ACTIONS(5450),
    [anon_sym_ATdynamic] = ACTIONS(5450),
    [anon_sym_BOOL] = ACTIONS(5448),
    [anon_sym_IMP] = ACTIONS(5448),
    [anon_sym_SEL] = ACTIONS(5448),
    [anon_sym_Class] = ACTIONS(5448),
    [anon_sym_id] = ACTIONS(5448),
  },
  [STATE(2222)] = {
    [sym_identifier] = ACTIONS(5359),
    [aux_sym_preproc_def_token1] = ACTIONS(5359),
    [aux_sym_preproc_if_token1] = ACTIONS(5359),
    [aux_sym_preproc_if_token2] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5359),
    [aux_sym_preproc_else_token1] = ACTIONS(5359),
    [aux_sym_preproc_elif_token1] = ACTIONS(5359),
    [sym_preproc_directive] = ACTIONS(5359),
    [anon_sym_DASH] = ACTIONS(5361),
    [anon_sym_PLUS] = ACTIONS(5361),
    [anon_sym___extension__] = ACTIONS(5359),
    [anon_sym_typedef] = ACTIONS(5359),
    [anon_sym_extern] = ACTIONS(5359),
    [anon_sym___attribute__] = ACTIONS(5359),
    [anon_sym___attribute] = ACTIONS(5359),
    [anon_sym_noreturn] = ACTIONS(5359),
    [anon_sym_LBRACK] = ACTIONS(5361),
    [anon_sym___declspec] = ACTIONS(5359),
    [anon_sym___cdecl] = ACTIONS(5359),
    [anon_sym___clrcall] = ACTIONS(5359),
    [anon_sym___stdcall] = ACTIONS(5359),
    [anon_sym___fastcall] = ACTIONS(5359),
    [anon_sym___thiscall] = ACTIONS(5359),
    [anon_sym___vectorcall] = ACTIONS(5359),
    [anon_sym_signed] = ACTIONS(5359),
    [anon_sym_unsigned] = ACTIONS(5359),
    [anon_sym_long] = ACTIONS(5359),
    [anon_sym_short] = ACTIONS(5359),
    [anon_sym_static] = ACTIONS(5359),
    [anon_sym_auto] = ACTIONS(5359),
    [anon_sym_register] = ACTIONS(5359),
    [anon_sym_inline] = ACTIONS(5359),
    [anon_sym___inline] = ACTIONS(5359),
    [anon_sym___inline__] = ACTIONS(5359),
    [anon_sym___forceinline] = ACTIONS(5359),
    [anon_sym_thread_local] = ACTIONS(5359),
    [anon_sym___thread] = ACTIONS(5359),
    [anon_sym_CG_EXTERN] = ACTIONS(5359),
    [anon_sym_CG_INLINE] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5359),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5359),
    [anon_sym_IBOutlet] = ACTIONS(5359),
    [anon_sym_IBInspectable] = ACTIONS(5359),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5359),
    [anon_sym_NS_INLINE] = ACTIONS(5359),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5359),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5359),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5359),
    [anon_sym_const] = ACTIONS(5359),
    [anon_sym_constexpr] = ACTIONS(5359),
    [anon_sym_volatile] = ACTIONS(5359),
    [anon_sym_restrict] = ACTIONS(5359),
    [anon_sym___restrict__] = ACTIONS(5359),
    [anon_sym__Atomic] = ACTIONS(5359),
    [anon_sym__Noreturn] = ACTIONS(5359),
    [anon_sym__Nonnull] = ACTIONS(5359),
    [anon_sym_nullable] = ACTIONS(5359),
    [anon_sym__Complex] = ACTIONS(5359),
    [anon_sym__Nullable] = ACTIONS(5359),
    [anon_sym__Nullable_result] = ACTIONS(5359),
    [anon_sym__Null_unspecified] = ACTIONS(5359),
    [anon_sym___autoreleasing] = ACTIONS(5359),
    [anon_sym___block] = ACTIONS(5359),
    [anon_sym___bridge] = ACTIONS(5359),
    [anon_sym___bridge_retained] = ACTIONS(5359),
    [anon_sym___bridge_transfer] = ACTIONS(5359),
    [anon_sym___complex] = ACTIONS(5359),
    [anon_sym___const] = ACTIONS(5359),
    [anon_sym___imag] = ACTIONS(5359),
    [anon_sym___kindof] = ACTIONS(5359),
    [anon_sym___nonnull] = ACTIONS(5359),
    [anon_sym___nullable] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5359),
    [anon_sym___real] = ACTIONS(5359),
    [anon_sym___strong] = ACTIONS(5359),
    [anon_sym___unsafe_unretained] = ACTIONS(5359),
    [anon_sym___unused] = ACTIONS(5359),
    [anon_sym___weak] = ACTIONS(5359),
    [anon_sym_alignas] = ACTIONS(5359),
    [anon_sym__Alignas] = ACTIONS(5359),
    [sym_primitive_type] = ACTIONS(5359),
    [anon_sym_enum] = ACTIONS(5359),
    [anon_sym_struct] = ACTIONS(5359),
    [anon_sym_union] = ACTIONS(5359),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5359),
    [anon_sym___typeof] = ACTIONS(5359),
    [anon_sym_typeof] = ACTIONS(5359),
    [aux_sym_preproc_undef_token1] = ACTIONS(5359),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5359),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5359),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5359),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_API_AVAILABLE] = ACTIONS(5359),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_API_DEPRECATED] = ACTIONS(5359),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym___deprecated_msg] = ACTIONS(5359),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5359),
    [anon_sym_ATsynthesize] = ACTIONS(5361),
    [anon_sym_ATdynamic] = ACTIONS(5361),
    [anon_sym_BOOL] = ACTIONS(5359),
    [anon_sym_IMP] = ACTIONS(5359),
    [anon_sym_SEL] = ACTIONS(5359),
    [anon_sym_Class] = ACTIONS(5359),
    [anon_sym_id] = ACTIONS(5359),
  },
  [STATE(2223)] = {
    [sym_identifier] = ACTIONS(5509),
    [aux_sym_preproc_def_token1] = ACTIONS(5509),
    [aux_sym_preproc_if_token1] = ACTIONS(5509),
    [aux_sym_preproc_if_token2] = ACTIONS(5509),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5509),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5509),
    [aux_sym_preproc_else_token1] = ACTIONS(5509),
    [aux_sym_preproc_elif_token1] = ACTIONS(5509),
    [sym_preproc_directive] = ACTIONS(5509),
    [anon_sym_DASH] = ACTIONS(5511),
    [anon_sym_PLUS] = ACTIONS(5511),
    [anon_sym___extension__] = ACTIONS(5509),
    [anon_sym_typedef] = ACTIONS(5509),
    [anon_sym_extern] = ACTIONS(5509),
    [anon_sym___attribute__] = ACTIONS(5509),
    [anon_sym___attribute] = ACTIONS(5509),
    [anon_sym_noreturn] = ACTIONS(5509),
    [anon_sym_LBRACK] = ACTIONS(5511),
    [anon_sym___declspec] = ACTIONS(5509),
    [anon_sym___cdecl] = ACTIONS(5509),
    [anon_sym___clrcall] = ACTIONS(5509),
    [anon_sym___stdcall] = ACTIONS(5509),
    [anon_sym___fastcall] = ACTIONS(5509),
    [anon_sym___thiscall] = ACTIONS(5509),
    [anon_sym___vectorcall] = ACTIONS(5509),
    [anon_sym_signed] = ACTIONS(5509),
    [anon_sym_unsigned] = ACTIONS(5509),
    [anon_sym_long] = ACTIONS(5509),
    [anon_sym_short] = ACTIONS(5509),
    [anon_sym_static] = ACTIONS(5509),
    [anon_sym_auto] = ACTIONS(5509),
    [anon_sym_register] = ACTIONS(5509),
    [anon_sym_inline] = ACTIONS(5509),
    [anon_sym___inline] = ACTIONS(5509),
    [anon_sym___inline__] = ACTIONS(5509),
    [anon_sym___forceinline] = ACTIONS(5509),
    [anon_sym_thread_local] = ACTIONS(5509),
    [anon_sym___thread] = ACTIONS(5509),
    [anon_sym_CG_EXTERN] = ACTIONS(5509),
    [anon_sym_CG_INLINE] = ACTIONS(5509),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5509),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5509),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5509),
    [anon_sym_IBOutlet] = ACTIONS(5509),
    [anon_sym_IBInspectable] = ACTIONS(5509),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5509),
    [anon_sym_NS_INLINE] = ACTIONS(5509),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5509),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5509),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5509),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5509),
    [anon_sym_const] = ACTIONS(5509),
    [anon_sym_constexpr] = ACTIONS(5509),
    [anon_sym_volatile] = ACTIONS(5509),
    [anon_sym_restrict] = ACTIONS(5509),
    [anon_sym___restrict__] = ACTIONS(5509),
    [anon_sym__Atomic] = ACTIONS(5509),
    [anon_sym__Noreturn] = ACTIONS(5509),
    [anon_sym__Nonnull] = ACTIONS(5509),
    [anon_sym_nullable] = ACTIONS(5509),
    [anon_sym__Complex] = ACTIONS(5509),
    [anon_sym__Nullable] = ACTIONS(5509),
    [anon_sym__Nullable_result] = ACTIONS(5509),
    [anon_sym__Null_unspecified] = ACTIONS(5509),
    [anon_sym___autoreleasing] = ACTIONS(5509),
    [anon_sym___block] = ACTIONS(5509),
    [anon_sym___bridge] = ACTIONS(5509),
    [anon_sym___bridge_retained] = ACTIONS(5509),
    [anon_sym___bridge_transfer] = ACTIONS(5509),
    [anon_sym___complex] = ACTIONS(5509),
    [anon_sym___const] = ACTIONS(5509),
    [anon_sym___imag] = ACTIONS(5509),
    [anon_sym___kindof] = ACTIONS(5509),
    [anon_sym___nonnull] = ACTIONS(5509),
    [anon_sym___nullable] = ACTIONS(5509),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5509),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5509),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5509),
    [anon_sym___real] = ACTIONS(5509),
    [anon_sym___strong] = ACTIONS(5509),
    [anon_sym___unsafe_unretained] = ACTIONS(5509),
    [anon_sym___unused] = ACTIONS(5509),
    [anon_sym___weak] = ACTIONS(5509),
    [anon_sym_alignas] = ACTIONS(5509),
    [anon_sym__Alignas] = ACTIONS(5509),
    [sym_primitive_type] = ACTIONS(5509),
    [anon_sym_enum] = ACTIONS(5509),
    [anon_sym_struct] = ACTIONS(5509),
    [anon_sym_union] = ACTIONS(5509),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5509),
    [anon_sym___typeof] = ACTIONS(5509),
    [anon_sym_typeof] = ACTIONS(5509),
    [aux_sym_preproc_undef_token1] = ACTIONS(5509),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5509),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5509),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5509),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5509),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5509),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5509),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5509),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5509),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5509),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_API_AVAILABLE] = ACTIONS(5509),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_API_DEPRECATED] = ACTIONS(5509),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5509),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5509),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5509),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5509),
    [anon_sym___deprecated_msg] = ACTIONS(5509),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5509),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5509),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5509),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5509),
    [anon_sym_ATsynthesize] = ACTIONS(5511),
    [anon_sym_ATdynamic] = ACTIONS(5511),
    [anon_sym_BOOL] = ACTIONS(5509),
    [anon_sym_IMP] = ACTIONS(5509),
    [anon_sym_SEL] = ACTIONS(5509),
    [anon_sym_Class] = ACTIONS(5509),
    [anon_sym_id] = ACTIONS(5509),
  },
  [STATE(2224)] = {
    [sym_identifier] = ACTIONS(5513),
    [aux_sym_preproc_def_token1] = ACTIONS(5513),
    [aux_sym_preproc_if_token1] = ACTIONS(5513),
    [aux_sym_preproc_if_token2] = ACTIONS(5513),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5513),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5513),
    [aux_sym_preproc_else_token1] = ACTIONS(5513),
    [aux_sym_preproc_elif_token1] = ACTIONS(5513),
    [sym_preproc_directive] = ACTIONS(5513),
    [anon_sym_DASH] = ACTIONS(5515),
    [anon_sym_PLUS] = ACTIONS(5515),
    [anon_sym___extension__] = ACTIONS(5513),
    [anon_sym_typedef] = ACTIONS(5513),
    [anon_sym_extern] = ACTIONS(5513),
    [anon_sym___attribute__] = ACTIONS(5513),
    [anon_sym___attribute] = ACTIONS(5513),
    [anon_sym_noreturn] = ACTIONS(5513),
    [anon_sym_LBRACK] = ACTIONS(5515),
    [anon_sym___declspec] = ACTIONS(5513),
    [anon_sym___cdecl] = ACTIONS(5513),
    [anon_sym___clrcall] = ACTIONS(5513),
    [anon_sym___stdcall] = ACTIONS(5513),
    [anon_sym___fastcall] = ACTIONS(5513),
    [anon_sym___thiscall] = ACTIONS(5513),
    [anon_sym___vectorcall] = ACTIONS(5513),
    [anon_sym_signed] = ACTIONS(5513),
    [anon_sym_unsigned] = ACTIONS(5513),
    [anon_sym_long] = ACTIONS(5513),
    [anon_sym_short] = ACTIONS(5513),
    [anon_sym_static] = ACTIONS(5513),
    [anon_sym_auto] = ACTIONS(5513),
    [anon_sym_register] = ACTIONS(5513),
    [anon_sym_inline] = ACTIONS(5513),
    [anon_sym___inline] = ACTIONS(5513),
    [anon_sym___inline__] = ACTIONS(5513),
    [anon_sym___forceinline] = ACTIONS(5513),
    [anon_sym_thread_local] = ACTIONS(5513),
    [anon_sym___thread] = ACTIONS(5513),
    [anon_sym_CG_EXTERN] = ACTIONS(5513),
    [anon_sym_CG_INLINE] = ACTIONS(5513),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5513),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5513),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5513),
    [anon_sym_IBOutlet] = ACTIONS(5513),
    [anon_sym_IBInspectable] = ACTIONS(5513),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5513),
    [anon_sym_NS_INLINE] = ACTIONS(5513),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5513),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5513),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5513),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5513),
    [anon_sym_const] = ACTIONS(5513),
    [anon_sym_constexpr] = ACTIONS(5513),
    [anon_sym_volatile] = ACTIONS(5513),
    [anon_sym_restrict] = ACTIONS(5513),
    [anon_sym___restrict__] = ACTIONS(5513),
    [anon_sym__Atomic] = ACTIONS(5513),
    [anon_sym__Noreturn] = ACTIONS(5513),
    [anon_sym__Nonnull] = ACTIONS(5513),
    [anon_sym_nullable] = ACTIONS(5513),
    [anon_sym__Complex] = ACTIONS(5513),
    [anon_sym__Nullable] = ACTIONS(5513),
    [anon_sym__Nullable_result] = ACTIONS(5513),
    [anon_sym__Null_unspecified] = ACTIONS(5513),
    [anon_sym___autoreleasing] = ACTIONS(5513),
    [anon_sym___block] = ACTIONS(5513),
    [anon_sym___bridge] = ACTIONS(5513),
    [anon_sym___bridge_retained] = ACTIONS(5513),
    [anon_sym___bridge_transfer] = ACTIONS(5513),
    [anon_sym___complex] = ACTIONS(5513),
    [anon_sym___const] = ACTIONS(5513),
    [anon_sym___imag] = ACTIONS(5513),
    [anon_sym___kindof] = ACTIONS(5513),
    [anon_sym___nonnull] = ACTIONS(5513),
    [anon_sym___nullable] = ACTIONS(5513),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5513),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5513),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5513),
    [anon_sym___real] = ACTIONS(5513),
    [anon_sym___strong] = ACTIONS(5513),
    [anon_sym___unsafe_unretained] = ACTIONS(5513),
    [anon_sym___unused] = ACTIONS(5513),
    [anon_sym___weak] = ACTIONS(5513),
    [anon_sym_alignas] = ACTIONS(5513),
    [anon_sym__Alignas] = ACTIONS(5513),
    [sym_primitive_type] = ACTIONS(5513),
    [anon_sym_enum] = ACTIONS(5513),
    [anon_sym_struct] = ACTIONS(5513),
    [anon_sym_union] = ACTIONS(5513),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5513),
    [anon_sym___typeof] = ACTIONS(5513),
    [anon_sym_typeof] = ACTIONS(5513),
    [aux_sym_preproc_undef_token1] = ACTIONS(5513),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5513),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5513),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5513),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5513),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5513),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5513),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5513),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5513),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5513),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_API_AVAILABLE] = ACTIONS(5513),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_API_DEPRECATED] = ACTIONS(5513),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5513),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5513),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5513),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5513),
    [anon_sym___deprecated_msg] = ACTIONS(5513),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5513),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5513),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5513),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5513),
    [anon_sym_ATsynthesize] = ACTIONS(5515),
    [anon_sym_ATdynamic] = ACTIONS(5515),
    [anon_sym_BOOL] = ACTIONS(5513),
    [anon_sym_IMP] = ACTIONS(5513),
    [anon_sym_SEL] = ACTIONS(5513),
    [anon_sym_Class] = ACTIONS(5513),
    [anon_sym_id] = ACTIONS(5513),
  },
  [STATE(2225)] = {
    [sym_identifier] = ACTIONS(5393),
    [aux_sym_preproc_def_token1] = ACTIONS(5393),
    [aux_sym_preproc_if_token1] = ACTIONS(5393),
    [aux_sym_preproc_if_token2] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5393),
    [aux_sym_preproc_else_token1] = ACTIONS(5393),
    [aux_sym_preproc_elif_token1] = ACTIONS(5393),
    [sym_preproc_directive] = ACTIONS(5393),
    [anon_sym_DASH] = ACTIONS(5395),
    [anon_sym_PLUS] = ACTIONS(5395),
    [anon_sym___extension__] = ACTIONS(5393),
    [anon_sym_typedef] = ACTIONS(5393),
    [anon_sym_extern] = ACTIONS(5393),
    [anon_sym___attribute__] = ACTIONS(5393),
    [anon_sym___attribute] = ACTIONS(5393),
    [anon_sym_noreturn] = ACTIONS(5393),
    [anon_sym_LBRACK] = ACTIONS(5395),
    [anon_sym___declspec] = ACTIONS(5393),
    [anon_sym___cdecl] = ACTIONS(5393),
    [anon_sym___clrcall] = ACTIONS(5393),
    [anon_sym___stdcall] = ACTIONS(5393),
    [anon_sym___fastcall] = ACTIONS(5393),
    [anon_sym___thiscall] = ACTIONS(5393),
    [anon_sym___vectorcall] = ACTIONS(5393),
    [anon_sym_signed] = ACTIONS(5393),
    [anon_sym_unsigned] = ACTIONS(5393),
    [anon_sym_long] = ACTIONS(5393),
    [anon_sym_short] = ACTIONS(5393),
    [anon_sym_static] = ACTIONS(5393),
    [anon_sym_auto] = ACTIONS(5393),
    [anon_sym_register] = ACTIONS(5393),
    [anon_sym_inline] = ACTIONS(5393),
    [anon_sym___inline] = ACTIONS(5393),
    [anon_sym___inline__] = ACTIONS(5393),
    [anon_sym___forceinline] = ACTIONS(5393),
    [anon_sym_thread_local] = ACTIONS(5393),
    [anon_sym___thread] = ACTIONS(5393),
    [anon_sym_CG_EXTERN] = ACTIONS(5393),
    [anon_sym_CG_INLINE] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5393),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5393),
    [anon_sym_IBOutlet] = ACTIONS(5393),
    [anon_sym_IBInspectable] = ACTIONS(5393),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5393),
    [anon_sym_NS_INLINE] = ACTIONS(5393),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5393),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5393),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5393),
    [anon_sym_const] = ACTIONS(5393),
    [anon_sym_constexpr] = ACTIONS(5393),
    [anon_sym_volatile] = ACTIONS(5393),
    [anon_sym_restrict] = ACTIONS(5393),
    [anon_sym___restrict__] = ACTIONS(5393),
    [anon_sym__Atomic] = ACTIONS(5393),
    [anon_sym__Noreturn] = ACTIONS(5393),
    [anon_sym__Nonnull] = ACTIONS(5393),
    [anon_sym_nullable] = ACTIONS(5393),
    [anon_sym__Complex] = ACTIONS(5393),
    [anon_sym__Nullable] = ACTIONS(5393),
    [anon_sym__Nullable_result] = ACTIONS(5393),
    [anon_sym__Null_unspecified] = ACTIONS(5393),
    [anon_sym___autoreleasing] = ACTIONS(5393),
    [anon_sym___block] = ACTIONS(5393),
    [anon_sym___bridge] = ACTIONS(5393),
    [anon_sym___bridge_retained] = ACTIONS(5393),
    [anon_sym___bridge_transfer] = ACTIONS(5393),
    [anon_sym___complex] = ACTIONS(5393),
    [anon_sym___const] = ACTIONS(5393),
    [anon_sym___imag] = ACTIONS(5393),
    [anon_sym___kindof] = ACTIONS(5393),
    [anon_sym___nonnull] = ACTIONS(5393),
    [anon_sym___nullable] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5393),
    [anon_sym___real] = ACTIONS(5393),
    [anon_sym___strong] = ACTIONS(5393),
    [anon_sym___unsafe_unretained] = ACTIONS(5393),
    [anon_sym___unused] = ACTIONS(5393),
    [anon_sym___weak] = ACTIONS(5393),
    [anon_sym_alignas] = ACTIONS(5393),
    [anon_sym__Alignas] = ACTIONS(5393),
    [sym_primitive_type] = ACTIONS(5393),
    [anon_sym_enum] = ACTIONS(5393),
    [anon_sym_struct] = ACTIONS(5393),
    [anon_sym_union] = ACTIONS(5393),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5393),
    [anon_sym___typeof] = ACTIONS(5393),
    [anon_sym_typeof] = ACTIONS(5393),
    [aux_sym_preproc_undef_token1] = ACTIONS(5393),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5393),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5393),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5393),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_API_AVAILABLE] = ACTIONS(5393),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_API_DEPRECATED] = ACTIONS(5393),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym___deprecated_msg] = ACTIONS(5393),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5393),
    [anon_sym_ATsynthesize] = ACTIONS(5395),
    [anon_sym_ATdynamic] = ACTIONS(5395),
    [anon_sym_BOOL] = ACTIONS(5393),
    [anon_sym_IMP] = ACTIONS(5393),
    [anon_sym_SEL] = ACTIONS(5393),
    [anon_sym_Class] = ACTIONS(5393),
    [anon_sym_id] = ACTIONS(5393),
  },
  [STATE(2226)] = {
    [sym_identifier] = ACTIONS(5475),
    [aux_sym_preproc_def_token1] = ACTIONS(5475),
    [aux_sym_preproc_if_token1] = ACTIONS(5475),
    [aux_sym_preproc_if_token2] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5475),
    [aux_sym_preproc_else_token1] = ACTIONS(5475),
    [aux_sym_preproc_elif_token1] = ACTIONS(5475),
    [sym_preproc_directive] = ACTIONS(5475),
    [anon_sym_DASH] = ACTIONS(5477),
    [anon_sym_PLUS] = ACTIONS(5477),
    [anon_sym___extension__] = ACTIONS(5475),
    [anon_sym_typedef] = ACTIONS(5475),
    [anon_sym_extern] = ACTIONS(5475),
    [anon_sym___attribute__] = ACTIONS(5475),
    [anon_sym___attribute] = ACTIONS(5475),
    [anon_sym_noreturn] = ACTIONS(5475),
    [anon_sym_LBRACK] = ACTIONS(5477),
    [anon_sym___declspec] = ACTIONS(5475),
    [anon_sym___cdecl] = ACTIONS(5475),
    [anon_sym___clrcall] = ACTIONS(5475),
    [anon_sym___stdcall] = ACTIONS(5475),
    [anon_sym___fastcall] = ACTIONS(5475),
    [anon_sym___thiscall] = ACTIONS(5475),
    [anon_sym___vectorcall] = ACTIONS(5475),
    [anon_sym_signed] = ACTIONS(5475),
    [anon_sym_unsigned] = ACTIONS(5475),
    [anon_sym_long] = ACTIONS(5475),
    [anon_sym_short] = ACTIONS(5475),
    [anon_sym_static] = ACTIONS(5475),
    [anon_sym_auto] = ACTIONS(5475),
    [anon_sym_register] = ACTIONS(5475),
    [anon_sym_inline] = ACTIONS(5475),
    [anon_sym___inline] = ACTIONS(5475),
    [anon_sym___inline__] = ACTIONS(5475),
    [anon_sym___forceinline] = ACTIONS(5475),
    [anon_sym_thread_local] = ACTIONS(5475),
    [anon_sym___thread] = ACTIONS(5475),
    [anon_sym_CG_EXTERN] = ACTIONS(5475),
    [anon_sym_CG_INLINE] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5475),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5475),
    [anon_sym_IBOutlet] = ACTIONS(5475),
    [anon_sym_IBInspectable] = ACTIONS(5475),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5475),
    [anon_sym_NS_INLINE] = ACTIONS(5475),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5475),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5475),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5475),
    [anon_sym_const] = ACTIONS(5475),
    [anon_sym_constexpr] = ACTIONS(5475),
    [anon_sym_volatile] = ACTIONS(5475),
    [anon_sym_restrict] = ACTIONS(5475),
    [anon_sym___restrict__] = ACTIONS(5475),
    [anon_sym__Atomic] = ACTIONS(5475),
    [anon_sym__Noreturn] = ACTIONS(5475),
    [anon_sym__Nonnull] = ACTIONS(5475),
    [anon_sym_nullable] = ACTIONS(5475),
    [anon_sym__Complex] = ACTIONS(5475),
    [anon_sym__Nullable] = ACTIONS(5475),
    [anon_sym__Nullable_result] = ACTIONS(5475),
    [anon_sym__Null_unspecified] = ACTIONS(5475),
    [anon_sym___autoreleasing] = ACTIONS(5475),
    [anon_sym___block] = ACTIONS(5475),
    [anon_sym___bridge] = ACTIONS(5475),
    [anon_sym___bridge_retained] = ACTIONS(5475),
    [anon_sym___bridge_transfer] = ACTIONS(5475),
    [anon_sym___complex] = ACTIONS(5475),
    [anon_sym___const] = ACTIONS(5475),
    [anon_sym___imag] = ACTIONS(5475),
    [anon_sym___kindof] = ACTIONS(5475),
    [anon_sym___nonnull] = ACTIONS(5475),
    [anon_sym___nullable] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5475),
    [anon_sym___real] = ACTIONS(5475),
    [anon_sym___strong] = ACTIONS(5475),
    [anon_sym___unsafe_unretained] = ACTIONS(5475),
    [anon_sym___unused] = ACTIONS(5475),
    [anon_sym___weak] = ACTIONS(5475),
    [anon_sym_alignas] = ACTIONS(5475),
    [anon_sym__Alignas] = ACTIONS(5475),
    [sym_primitive_type] = ACTIONS(5475),
    [anon_sym_enum] = ACTIONS(5475),
    [anon_sym_struct] = ACTIONS(5475),
    [anon_sym_union] = ACTIONS(5475),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5475),
    [anon_sym___typeof] = ACTIONS(5475),
    [anon_sym_typeof] = ACTIONS(5475),
    [aux_sym_preproc_undef_token1] = ACTIONS(5475),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5475),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5475),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5475),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_API_AVAILABLE] = ACTIONS(5475),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_API_DEPRECATED] = ACTIONS(5475),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym___deprecated_msg] = ACTIONS(5475),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5475),
    [anon_sym_ATsynthesize] = ACTIONS(5477),
    [anon_sym_ATdynamic] = ACTIONS(5477),
    [anon_sym_BOOL] = ACTIONS(5475),
    [anon_sym_IMP] = ACTIONS(5475),
    [anon_sym_SEL] = ACTIONS(5475),
    [anon_sym_Class] = ACTIONS(5475),
    [anon_sym_id] = ACTIONS(5475),
  },
  [STATE(2227)] = {
    [sym_identifier] = ACTIONS(5369),
    [aux_sym_preproc_def_token1] = ACTIONS(5369),
    [aux_sym_preproc_if_token1] = ACTIONS(5369),
    [aux_sym_preproc_if_token2] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5369),
    [aux_sym_preproc_else_token1] = ACTIONS(5369),
    [aux_sym_preproc_elif_token1] = ACTIONS(5369),
    [sym_preproc_directive] = ACTIONS(5369),
    [anon_sym_DASH] = ACTIONS(5371),
    [anon_sym_PLUS] = ACTIONS(5371),
    [anon_sym___extension__] = ACTIONS(5369),
    [anon_sym_typedef] = ACTIONS(5369),
    [anon_sym_extern] = ACTIONS(5369),
    [anon_sym___attribute__] = ACTIONS(5369),
    [anon_sym___attribute] = ACTIONS(5369),
    [anon_sym_noreturn] = ACTIONS(5369),
    [anon_sym_LBRACK] = ACTIONS(5371),
    [anon_sym___declspec] = ACTIONS(5369),
    [anon_sym___cdecl] = ACTIONS(5369),
    [anon_sym___clrcall] = ACTIONS(5369),
    [anon_sym___stdcall] = ACTIONS(5369),
    [anon_sym___fastcall] = ACTIONS(5369),
    [anon_sym___thiscall] = ACTIONS(5369),
    [anon_sym___vectorcall] = ACTIONS(5369),
    [anon_sym_signed] = ACTIONS(5369),
    [anon_sym_unsigned] = ACTIONS(5369),
    [anon_sym_long] = ACTIONS(5369),
    [anon_sym_short] = ACTIONS(5369),
    [anon_sym_static] = ACTIONS(5369),
    [anon_sym_auto] = ACTIONS(5369),
    [anon_sym_register] = ACTIONS(5369),
    [anon_sym_inline] = ACTIONS(5369),
    [anon_sym___inline] = ACTIONS(5369),
    [anon_sym___inline__] = ACTIONS(5369),
    [anon_sym___forceinline] = ACTIONS(5369),
    [anon_sym_thread_local] = ACTIONS(5369),
    [anon_sym___thread] = ACTIONS(5369),
    [anon_sym_CG_EXTERN] = ACTIONS(5369),
    [anon_sym_CG_INLINE] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5369),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5369),
    [anon_sym_IBOutlet] = ACTIONS(5369),
    [anon_sym_IBInspectable] = ACTIONS(5369),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5369),
    [anon_sym_NS_INLINE] = ACTIONS(5369),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5369),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5369),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5369),
    [anon_sym_const] = ACTIONS(5369),
    [anon_sym_constexpr] = ACTIONS(5369),
    [anon_sym_volatile] = ACTIONS(5369),
    [anon_sym_restrict] = ACTIONS(5369),
    [anon_sym___restrict__] = ACTIONS(5369),
    [anon_sym__Atomic] = ACTIONS(5369),
    [anon_sym__Noreturn] = ACTIONS(5369),
    [anon_sym__Nonnull] = ACTIONS(5369),
    [anon_sym_nullable] = ACTIONS(5369),
    [anon_sym__Complex] = ACTIONS(5369),
    [anon_sym__Nullable] = ACTIONS(5369),
    [anon_sym__Nullable_result] = ACTIONS(5369),
    [anon_sym__Null_unspecified] = ACTIONS(5369),
    [anon_sym___autoreleasing] = ACTIONS(5369),
    [anon_sym___block] = ACTIONS(5369),
    [anon_sym___bridge] = ACTIONS(5369),
    [anon_sym___bridge_retained] = ACTIONS(5369),
    [anon_sym___bridge_transfer] = ACTIONS(5369),
    [anon_sym___complex] = ACTIONS(5369),
    [anon_sym___const] = ACTIONS(5369),
    [anon_sym___imag] = ACTIONS(5369),
    [anon_sym___kindof] = ACTIONS(5369),
    [anon_sym___nonnull] = ACTIONS(5369),
    [anon_sym___nullable] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5369),
    [anon_sym___real] = ACTIONS(5369),
    [anon_sym___strong] = ACTIONS(5369),
    [anon_sym___unsafe_unretained] = ACTIONS(5369),
    [anon_sym___unused] = ACTIONS(5369),
    [anon_sym___weak] = ACTIONS(5369),
    [anon_sym_alignas] = ACTIONS(5369),
    [anon_sym__Alignas] = ACTIONS(5369),
    [sym_primitive_type] = ACTIONS(5369),
    [anon_sym_enum] = ACTIONS(5369),
    [anon_sym_struct] = ACTIONS(5369),
    [anon_sym_union] = ACTIONS(5369),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5369),
    [anon_sym___typeof] = ACTIONS(5369),
    [anon_sym_typeof] = ACTIONS(5369),
    [aux_sym_preproc_undef_token1] = ACTIONS(5369),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5369),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5369),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5369),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_API_AVAILABLE] = ACTIONS(5369),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_API_DEPRECATED] = ACTIONS(5369),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym___deprecated_msg] = ACTIONS(5369),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5369),
    [anon_sym_ATsynthesize] = ACTIONS(5371),
    [anon_sym_ATdynamic] = ACTIONS(5371),
    [anon_sym_BOOL] = ACTIONS(5369),
    [anon_sym_IMP] = ACTIONS(5369),
    [anon_sym_SEL] = ACTIONS(5369),
    [anon_sym_Class] = ACTIONS(5369),
    [anon_sym_id] = ACTIONS(5369),
  },
  [STATE(2228)] = {
    [sym_identifier] = ACTIONS(5399),
    [aux_sym_preproc_def_token1] = ACTIONS(5399),
    [aux_sym_preproc_if_token1] = ACTIONS(5399),
    [aux_sym_preproc_if_token2] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5399),
    [aux_sym_preproc_else_token1] = ACTIONS(5399),
    [aux_sym_preproc_elif_token1] = ACTIONS(5399),
    [sym_preproc_directive] = ACTIONS(5399),
    [anon_sym_DASH] = ACTIONS(5401),
    [anon_sym_PLUS] = ACTIONS(5401),
    [anon_sym___extension__] = ACTIONS(5399),
    [anon_sym_typedef] = ACTIONS(5399),
    [anon_sym_extern] = ACTIONS(5399),
    [anon_sym___attribute__] = ACTIONS(5399),
    [anon_sym___attribute] = ACTIONS(5399),
    [anon_sym_noreturn] = ACTIONS(5399),
    [anon_sym_LBRACK] = ACTIONS(5401),
    [anon_sym___declspec] = ACTIONS(5399),
    [anon_sym___cdecl] = ACTIONS(5399),
    [anon_sym___clrcall] = ACTIONS(5399),
    [anon_sym___stdcall] = ACTIONS(5399),
    [anon_sym___fastcall] = ACTIONS(5399),
    [anon_sym___thiscall] = ACTIONS(5399),
    [anon_sym___vectorcall] = ACTIONS(5399),
    [anon_sym_signed] = ACTIONS(5399),
    [anon_sym_unsigned] = ACTIONS(5399),
    [anon_sym_long] = ACTIONS(5399),
    [anon_sym_short] = ACTIONS(5399),
    [anon_sym_static] = ACTIONS(5399),
    [anon_sym_auto] = ACTIONS(5399),
    [anon_sym_register] = ACTIONS(5399),
    [anon_sym_inline] = ACTIONS(5399),
    [anon_sym___inline] = ACTIONS(5399),
    [anon_sym___inline__] = ACTIONS(5399),
    [anon_sym___forceinline] = ACTIONS(5399),
    [anon_sym_thread_local] = ACTIONS(5399),
    [anon_sym___thread] = ACTIONS(5399),
    [anon_sym_CG_EXTERN] = ACTIONS(5399),
    [anon_sym_CG_INLINE] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5399),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5399),
    [anon_sym_IBOutlet] = ACTIONS(5399),
    [anon_sym_IBInspectable] = ACTIONS(5399),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5399),
    [anon_sym_NS_INLINE] = ACTIONS(5399),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5399),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5399),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5399),
    [anon_sym_const] = ACTIONS(5399),
    [anon_sym_constexpr] = ACTIONS(5399),
    [anon_sym_volatile] = ACTIONS(5399),
    [anon_sym_restrict] = ACTIONS(5399),
    [anon_sym___restrict__] = ACTIONS(5399),
    [anon_sym__Atomic] = ACTIONS(5399),
    [anon_sym__Noreturn] = ACTIONS(5399),
    [anon_sym__Nonnull] = ACTIONS(5399),
    [anon_sym_nullable] = ACTIONS(5399),
    [anon_sym__Complex] = ACTIONS(5399),
    [anon_sym__Nullable] = ACTIONS(5399),
    [anon_sym__Nullable_result] = ACTIONS(5399),
    [anon_sym__Null_unspecified] = ACTIONS(5399),
    [anon_sym___autoreleasing] = ACTIONS(5399),
    [anon_sym___block] = ACTIONS(5399),
    [anon_sym___bridge] = ACTIONS(5399),
    [anon_sym___bridge_retained] = ACTIONS(5399),
    [anon_sym___bridge_transfer] = ACTIONS(5399),
    [anon_sym___complex] = ACTIONS(5399),
    [anon_sym___const] = ACTIONS(5399),
    [anon_sym___imag] = ACTIONS(5399),
    [anon_sym___kindof] = ACTIONS(5399),
    [anon_sym___nonnull] = ACTIONS(5399),
    [anon_sym___nullable] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5399),
    [anon_sym___real] = ACTIONS(5399),
    [anon_sym___strong] = ACTIONS(5399),
    [anon_sym___unsafe_unretained] = ACTIONS(5399),
    [anon_sym___unused] = ACTIONS(5399),
    [anon_sym___weak] = ACTIONS(5399),
    [anon_sym_alignas] = ACTIONS(5399),
    [anon_sym__Alignas] = ACTIONS(5399),
    [sym_primitive_type] = ACTIONS(5399),
    [anon_sym_enum] = ACTIONS(5399),
    [anon_sym_struct] = ACTIONS(5399),
    [anon_sym_union] = ACTIONS(5399),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5399),
    [anon_sym___typeof] = ACTIONS(5399),
    [anon_sym_typeof] = ACTIONS(5399),
    [aux_sym_preproc_undef_token1] = ACTIONS(5399),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5399),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5399),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5399),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_API_AVAILABLE] = ACTIONS(5399),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_API_DEPRECATED] = ACTIONS(5399),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym___deprecated_msg] = ACTIONS(5399),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5399),
    [anon_sym_ATsynthesize] = ACTIONS(5401),
    [anon_sym_ATdynamic] = ACTIONS(5401),
    [anon_sym_BOOL] = ACTIONS(5399),
    [anon_sym_IMP] = ACTIONS(5399),
    [anon_sym_SEL] = ACTIONS(5399),
    [anon_sym_Class] = ACTIONS(5399),
    [anon_sym_id] = ACTIONS(5399),
  },
  [STATE(2229)] = {
    [sym_identifier] = ACTIONS(5381),
    [aux_sym_preproc_def_token1] = ACTIONS(5381),
    [aux_sym_preproc_if_token1] = ACTIONS(5381),
    [aux_sym_preproc_if_token2] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5381),
    [aux_sym_preproc_else_token1] = ACTIONS(5381),
    [aux_sym_preproc_elif_token1] = ACTIONS(5381),
    [sym_preproc_directive] = ACTIONS(5381),
    [anon_sym_DASH] = ACTIONS(5383),
    [anon_sym_PLUS] = ACTIONS(5383),
    [anon_sym___extension__] = ACTIONS(5381),
    [anon_sym_typedef] = ACTIONS(5381),
    [anon_sym_extern] = ACTIONS(5381),
    [anon_sym___attribute__] = ACTIONS(5381),
    [anon_sym___attribute] = ACTIONS(5381),
    [anon_sym_noreturn] = ACTIONS(5381),
    [anon_sym_LBRACK] = ACTIONS(5383),
    [anon_sym___declspec] = ACTIONS(5381),
    [anon_sym___cdecl] = ACTIONS(5381),
    [anon_sym___clrcall] = ACTIONS(5381),
    [anon_sym___stdcall] = ACTIONS(5381),
    [anon_sym___fastcall] = ACTIONS(5381),
    [anon_sym___thiscall] = ACTIONS(5381),
    [anon_sym___vectorcall] = ACTIONS(5381),
    [anon_sym_signed] = ACTIONS(5381),
    [anon_sym_unsigned] = ACTIONS(5381),
    [anon_sym_long] = ACTIONS(5381),
    [anon_sym_short] = ACTIONS(5381),
    [anon_sym_static] = ACTIONS(5381),
    [anon_sym_auto] = ACTIONS(5381),
    [anon_sym_register] = ACTIONS(5381),
    [anon_sym_inline] = ACTIONS(5381),
    [anon_sym___inline] = ACTIONS(5381),
    [anon_sym___inline__] = ACTIONS(5381),
    [anon_sym___forceinline] = ACTIONS(5381),
    [anon_sym_thread_local] = ACTIONS(5381),
    [anon_sym___thread] = ACTIONS(5381),
    [anon_sym_CG_EXTERN] = ACTIONS(5381),
    [anon_sym_CG_INLINE] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5381),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5381),
    [anon_sym_IBOutlet] = ACTIONS(5381),
    [anon_sym_IBInspectable] = ACTIONS(5381),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5381),
    [anon_sym_NS_INLINE] = ACTIONS(5381),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5381),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5381),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5381),
    [anon_sym_const] = ACTIONS(5381),
    [anon_sym_constexpr] = ACTIONS(5381),
    [anon_sym_volatile] = ACTIONS(5381),
    [anon_sym_restrict] = ACTIONS(5381),
    [anon_sym___restrict__] = ACTIONS(5381),
    [anon_sym__Atomic] = ACTIONS(5381),
    [anon_sym__Noreturn] = ACTIONS(5381),
    [anon_sym__Nonnull] = ACTIONS(5381),
    [anon_sym_nullable] = ACTIONS(5381),
    [anon_sym__Complex] = ACTIONS(5381),
    [anon_sym__Nullable] = ACTIONS(5381),
    [anon_sym__Nullable_result] = ACTIONS(5381),
    [anon_sym__Null_unspecified] = ACTIONS(5381),
    [anon_sym___autoreleasing] = ACTIONS(5381),
    [anon_sym___block] = ACTIONS(5381),
    [anon_sym___bridge] = ACTIONS(5381),
    [anon_sym___bridge_retained] = ACTIONS(5381),
    [anon_sym___bridge_transfer] = ACTIONS(5381),
    [anon_sym___complex] = ACTIONS(5381),
    [anon_sym___const] = ACTIONS(5381),
    [anon_sym___imag] = ACTIONS(5381),
    [anon_sym___kindof] = ACTIONS(5381),
    [anon_sym___nonnull] = ACTIONS(5381),
    [anon_sym___nullable] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5381),
    [anon_sym___real] = ACTIONS(5381),
    [anon_sym___strong] = ACTIONS(5381),
    [anon_sym___unsafe_unretained] = ACTIONS(5381),
    [anon_sym___unused] = ACTIONS(5381),
    [anon_sym___weak] = ACTIONS(5381),
    [anon_sym_alignas] = ACTIONS(5381),
    [anon_sym__Alignas] = ACTIONS(5381),
    [sym_primitive_type] = ACTIONS(5381),
    [anon_sym_enum] = ACTIONS(5381),
    [anon_sym_struct] = ACTIONS(5381),
    [anon_sym_union] = ACTIONS(5381),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5381),
    [anon_sym___typeof] = ACTIONS(5381),
    [anon_sym_typeof] = ACTIONS(5381),
    [aux_sym_preproc_undef_token1] = ACTIONS(5381),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5381),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5381),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5381),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_API_AVAILABLE] = ACTIONS(5381),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_API_DEPRECATED] = ACTIONS(5381),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym___deprecated_msg] = ACTIONS(5381),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5381),
    [anon_sym_ATsynthesize] = ACTIONS(5383),
    [anon_sym_ATdynamic] = ACTIONS(5383),
    [anon_sym_BOOL] = ACTIONS(5381),
    [anon_sym_IMP] = ACTIONS(5381),
    [anon_sym_SEL] = ACTIONS(5381),
    [anon_sym_Class] = ACTIONS(5381),
    [anon_sym_id] = ACTIONS(5381),
  },
  [STATE(2230)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(63),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2231)] = {
    [sym_identifier] = ACTIONS(2346),
    [aux_sym_preproc_def_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token1] = ACTIONS(2346),
    [aux_sym_preproc_if_token2] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2346),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2346),
    [sym_preproc_directive] = ACTIONS(2346),
    [anon_sym_DASH] = ACTIONS(2348),
    [anon_sym_PLUS] = ACTIONS(2348),
    [anon_sym_SEMI] = ACTIONS(2348),
    [anon_sym___extension__] = ACTIONS(2346),
    [anon_sym_typedef] = ACTIONS(2346),
    [anon_sym_extern] = ACTIONS(2346),
    [anon_sym___attribute__] = ACTIONS(2346),
    [anon_sym___attribute] = ACTIONS(2346),
    [anon_sym_noreturn] = ACTIONS(2346),
    [anon_sym_LBRACK] = ACTIONS(2348),
    [anon_sym___declspec] = ACTIONS(2346),
    [anon_sym___cdecl] = ACTIONS(2346),
    [anon_sym___clrcall] = ACTIONS(2346),
    [anon_sym___stdcall] = ACTIONS(2346),
    [anon_sym___fastcall] = ACTIONS(2346),
    [anon_sym___thiscall] = ACTIONS(2346),
    [anon_sym___vectorcall] = ACTIONS(2346),
    [anon_sym_signed] = ACTIONS(2346),
    [anon_sym_unsigned] = ACTIONS(2346),
    [anon_sym_long] = ACTIONS(2346),
    [anon_sym_short] = ACTIONS(2346),
    [anon_sym_static] = ACTIONS(2346),
    [anon_sym_auto] = ACTIONS(2346),
    [anon_sym_register] = ACTIONS(2346),
    [anon_sym_inline] = ACTIONS(2346),
    [anon_sym___inline] = ACTIONS(2346),
    [anon_sym___inline__] = ACTIONS(2346),
    [anon_sym___forceinline] = ACTIONS(2346),
    [anon_sym_thread_local] = ACTIONS(2346),
    [anon_sym___thread] = ACTIONS(2346),
    [anon_sym_CG_EXTERN] = ACTIONS(2346),
    [anon_sym_CG_INLINE] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2346),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2346),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2346),
    [anon_sym_IBOutlet] = ACTIONS(2346),
    [anon_sym_IBInspectable] = ACTIONS(2346),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2346),
    [anon_sym_NS_INLINE] = ACTIONS(2346),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2346),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2346),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2346),
    [anon_sym_const] = ACTIONS(2346),
    [anon_sym_constexpr] = ACTIONS(2346),
    [anon_sym_volatile] = ACTIONS(2346),
    [anon_sym_restrict] = ACTIONS(2346),
    [anon_sym___restrict__] = ACTIONS(2346),
    [anon_sym__Atomic] = ACTIONS(2346),
    [anon_sym__Noreturn] = ACTIONS(2346),
    [anon_sym__Nonnull] = ACTIONS(2346),
    [anon_sym_nullable] = ACTIONS(2346),
    [anon_sym__Complex] = ACTIONS(2346),
    [anon_sym__Nullable] = ACTIONS(2346),
    [anon_sym__Nullable_result] = ACTIONS(2346),
    [anon_sym__Null_unspecified] = ACTIONS(2346),
    [anon_sym___autoreleasing] = ACTIONS(2346),
    [anon_sym___block] = ACTIONS(2346),
    [anon_sym___bridge] = ACTIONS(2346),
    [anon_sym___bridge_retained] = ACTIONS(2346),
    [anon_sym___bridge_transfer] = ACTIONS(2346),
    [anon_sym___complex] = ACTIONS(2346),
    [anon_sym___const] = ACTIONS(2346),
    [anon_sym___imag] = ACTIONS(2346),
    [anon_sym___kindof] = ACTIONS(2346),
    [anon_sym___nonnull] = ACTIONS(2346),
    [anon_sym___nullable] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2346),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2346),
    [anon_sym___real] = ACTIONS(2346),
    [anon_sym___strong] = ACTIONS(2346),
    [anon_sym___unsafe_unretained] = ACTIONS(2346),
    [anon_sym___unused] = ACTIONS(2346),
    [anon_sym___weak] = ACTIONS(2346),
    [anon_sym_alignas] = ACTIONS(2346),
    [anon_sym__Alignas] = ACTIONS(2346),
    [sym_primitive_type] = ACTIONS(2346),
    [anon_sym_enum] = ACTIONS(2346),
    [anon_sym_struct] = ACTIONS(2346),
    [anon_sym_union] = ACTIONS(2346),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2346),
    [anon_sym___typeof] = ACTIONS(2346),
    [anon_sym_typeof] = ACTIONS(2346),
    [aux_sym_preproc_undef_token1] = ACTIONS(2346),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2346),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2346),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2346),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2346),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2346),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_API_AVAILABLE] = ACTIONS(2346),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_API_DEPRECATED] = ACTIONS(2346),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2346),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2346),
    [anon_sym___deprecated_msg] = ACTIONS(2346),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2346),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2346),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2346),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2346),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2346),
    [anon_sym_ATsynthesize] = ACTIONS(2348),
    [anon_sym_ATdynamic] = ACTIONS(2348),
    [anon_sym_ATproperty] = ACTIONS(2348),
    [anon_sym_BOOL] = ACTIONS(2346),
    [anon_sym_IMP] = ACTIONS(2346),
    [anon_sym_SEL] = ACTIONS(2346),
    [anon_sym_Class] = ACTIONS(2346),
    [anon_sym_id] = ACTIONS(2346),
  },
  [STATE(2232)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(2857),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2862),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(49),
    [anon_sym_unsigned] = ACTIONS(49),
    [anon_sym_long] = ACTIONS(49),
    [anon_sym_short] = ACTIONS(49),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4302),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(2233)] = {
    [sym_identifier] = ACTIONS(5438),
    [aux_sym_preproc_def_token1] = ACTIONS(5438),
    [aux_sym_preproc_if_token1] = ACTIONS(5438),
    [aux_sym_preproc_if_token2] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5438),
    [aux_sym_preproc_else_token1] = ACTIONS(5438),
    [aux_sym_preproc_elif_token1] = ACTIONS(5438),
    [sym_preproc_directive] = ACTIONS(5438),
    [anon_sym_DASH] = ACTIONS(5440),
    [anon_sym_PLUS] = ACTIONS(5440),
    [anon_sym___extension__] = ACTIONS(5438),
    [anon_sym_typedef] = ACTIONS(5438),
    [anon_sym_extern] = ACTIONS(5438),
    [anon_sym___attribute__] = ACTIONS(5438),
    [anon_sym___attribute] = ACTIONS(5438),
    [anon_sym_noreturn] = ACTIONS(5438),
    [anon_sym_LBRACK] = ACTIONS(5440),
    [anon_sym___declspec] = ACTIONS(5438),
    [anon_sym___cdecl] = ACTIONS(5438),
    [anon_sym___clrcall] = ACTIONS(5438),
    [anon_sym___stdcall] = ACTIONS(5438),
    [anon_sym___fastcall] = ACTIONS(5438),
    [anon_sym___thiscall] = ACTIONS(5438),
    [anon_sym___vectorcall] = ACTIONS(5438),
    [anon_sym_signed] = ACTIONS(5438),
    [anon_sym_unsigned] = ACTIONS(5438),
    [anon_sym_long] = ACTIONS(5438),
    [anon_sym_short] = ACTIONS(5438),
    [anon_sym_static] = ACTIONS(5438),
    [anon_sym_auto] = ACTIONS(5438),
    [anon_sym_register] = ACTIONS(5438),
    [anon_sym_inline] = ACTIONS(5438),
    [anon_sym___inline] = ACTIONS(5438),
    [anon_sym___inline__] = ACTIONS(5438),
    [anon_sym___forceinline] = ACTIONS(5438),
    [anon_sym_thread_local] = ACTIONS(5438),
    [anon_sym___thread] = ACTIONS(5438),
    [anon_sym_CG_EXTERN] = ACTIONS(5438),
    [anon_sym_CG_INLINE] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5438),
    [anon_sym_IBOutlet] = ACTIONS(5438),
    [anon_sym_IBInspectable] = ACTIONS(5438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5438),
    [anon_sym_NS_INLINE] = ACTIONS(5438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5438),
    [anon_sym_const] = ACTIONS(5438),
    [anon_sym_constexpr] = ACTIONS(5438),
    [anon_sym_volatile] = ACTIONS(5438),
    [anon_sym_restrict] = ACTIONS(5438),
    [anon_sym___restrict__] = ACTIONS(5438),
    [anon_sym__Atomic] = ACTIONS(5438),
    [anon_sym__Noreturn] = ACTIONS(5438),
    [anon_sym__Nonnull] = ACTIONS(5438),
    [anon_sym_nullable] = ACTIONS(5438),
    [anon_sym__Complex] = ACTIONS(5438),
    [anon_sym__Nullable] = ACTIONS(5438),
    [anon_sym__Nullable_result] = ACTIONS(5438),
    [anon_sym__Null_unspecified] = ACTIONS(5438),
    [anon_sym___autoreleasing] = ACTIONS(5438),
    [anon_sym___block] = ACTIONS(5438),
    [anon_sym___bridge] = ACTIONS(5438),
    [anon_sym___bridge_retained] = ACTIONS(5438),
    [anon_sym___bridge_transfer] = ACTIONS(5438),
    [anon_sym___complex] = ACTIONS(5438),
    [anon_sym___const] = ACTIONS(5438),
    [anon_sym___imag] = ACTIONS(5438),
    [anon_sym___kindof] = ACTIONS(5438),
    [anon_sym___nonnull] = ACTIONS(5438),
    [anon_sym___nullable] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5438),
    [anon_sym___real] = ACTIONS(5438),
    [anon_sym___strong] = ACTIONS(5438),
    [anon_sym___unsafe_unretained] = ACTIONS(5438),
    [anon_sym___unused] = ACTIONS(5438),
    [anon_sym___weak] = ACTIONS(5438),
    [anon_sym_alignas] = ACTIONS(5438),
    [anon_sym__Alignas] = ACTIONS(5438),
    [sym_primitive_type] = ACTIONS(5438),
    [anon_sym_enum] = ACTIONS(5438),
    [anon_sym_struct] = ACTIONS(5438),
    [anon_sym_union] = ACTIONS(5438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5438),
    [anon_sym___typeof] = ACTIONS(5438),
    [anon_sym_typeof] = ACTIONS(5438),
    [aux_sym_preproc_undef_token1] = ACTIONS(5438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_API_AVAILABLE] = ACTIONS(5438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_API_DEPRECATED] = ACTIONS(5438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym___deprecated_msg] = ACTIONS(5438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5438),
    [anon_sym_ATsynthesize] = ACTIONS(5440),
    [anon_sym_ATdynamic] = ACTIONS(5440),
    [anon_sym_BOOL] = ACTIONS(5438),
    [anon_sym_IMP] = ACTIONS(5438),
    [anon_sym_SEL] = ACTIONS(5438),
    [anon_sym_Class] = ACTIONS(5438),
    [anon_sym_id] = ACTIONS(5438),
  },
  [STATE(2234)] = {
    [sym_identifier] = ACTIONS(2622),
    [aux_sym_preproc_def_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token2] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2622),
    [aux_sym_preproc_else_token1] = ACTIONS(2622),
    [aux_sym_preproc_elif_token1] = ACTIONS(2622),
    [sym_preproc_directive] = ACTIONS(2622),
    [anon_sym_DASH] = ACTIONS(2624),
    [anon_sym_PLUS] = ACTIONS(2624),
    [anon_sym___extension__] = ACTIONS(2622),
    [anon_sym_typedef] = ACTIONS(2622),
    [anon_sym_extern] = ACTIONS(2622),
    [anon_sym___attribute__] = ACTIONS(2622),
    [anon_sym___attribute] = ACTIONS(2622),
    [anon_sym_noreturn] = ACTIONS(2622),
    [anon_sym_LBRACK] = ACTIONS(2624),
    [anon_sym___declspec] = ACTIONS(2622),
    [anon_sym___cdecl] = ACTIONS(2622),
    [anon_sym___clrcall] = ACTIONS(2622),
    [anon_sym___stdcall] = ACTIONS(2622),
    [anon_sym___fastcall] = ACTIONS(2622),
    [anon_sym___thiscall] = ACTIONS(2622),
    [anon_sym___vectorcall] = ACTIONS(2622),
    [anon_sym_signed] = ACTIONS(2622),
    [anon_sym_unsigned] = ACTIONS(2622),
    [anon_sym_long] = ACTIONS(2622),
    [anon_sym_short] = ACTIONS(2622),
    [anon_sym_static] = ACTIONS(2622),
    [anon_sym_auto] = ACTIONS(2622),
    [anon_sym_register] = ACTIONS(2622),
    [anon_sym_inline] = ACTIONS(2622),
    [anon_sym___inline] = ACTIONS(2622),
    [anon_sym___inline__] = ACTIONS(2622),
    [anon_sym___forceinline] = ACTIONS(2622),
    [anon_sym_thread_local] = ACTIONS(2622),
    [anon_sym___thread] = ACTIONS(2622),
    [anon_sym_CG_EXTERN] = ACTIONS(2622),
    [anon_sym_CG_INLINE] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2622),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2622),
    [anon_sym_IBOutlet] = ACTIONS(2622),
    [anon_sym_IBInspectable] = ACTIONS(2622),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2622),
    [anon_sym_NS_INLINE] = ACTIONS(2622),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2622),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2622),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2622),
    [anon_sym_const] = ACTIONS(2622),
    [anon_sym_constexpr] = ACTIONS(2622),
    [anon_sym_volatile] = ACTIONS(2622),
    [anon_sym_restrict] = ACTIONS(2622),
    [anon_sym___restrict__] = ACTIONS(2622),
    [anon_sym__Atomic] = ACTIONS(2622),
    [anon_sym__Noreturn] = ACTIONS(2622),
    [anon_sym__Nonnull] = ACTIONS(2622),
    [anon_sym_nullable] = ACTIONS(2622),
    [anon_sym__Complex] = ACTIONS(2622),
    [anon_sym__Nullable] = ACTIONS(2622),
    [anon_sym__Nullable_result] = ACTIONS(2622),
    [anon_sym__Null_unspecified] = ACTIONS(2622),
    [anon_sym___autoreleasing] = ACTIONS(2622),
    [anon_sym___block] = ACTIONS(2622),
    [anon_sym___bridge] = ACTIONS(2622),
    [anon_sym___bridge_retained] = ACTIONS(2622),
    [anon_sym___bridge_transfer] = ACTIONS(2622),
    [anon_sym___complex] = ACTIONS(2622),
    [anon_sym___const] = ACTIONS(2622),
    [anon_sym___imag] = ACTIONS(2622),
    [anon_sym___kindof] = ACTIONS(2622),
    [anon_sym___nonnull] = ACTIONS(2622),
    [anon_sym___nullable] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2622),
    [anon_sym___real] = ACTIONS(2622),
    [anon_sym___strong] = ACTIONS(2622),
    [anon_sym___unsafe_unretained] = ACTIONS(2622),
    [anon_sym___unused] = ACTIONS(2622),
    [anon_sym___weak] = ACTIONS(2622),
    [anon_sym_alignas] = ACTIONS(2622),
    [anon_sym__Alignas] = ACTIONS(2622),
    [sym_primitive_type] = ACTIONS(2622),
    [anon_sym_enum] = ACTIONS(2622),
    [anon_sym_struct] = ACTIONS(2622),
    [anon_sym_union] = ACTIONS(2622),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2622),
    [anon_sym___typeof] = ACTIONS(2622),
    [anon_sym_typeof] = ACTIONS(2622),
    [aux_sym_preproc_undef_token1] = ACTIONS(2622),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2622),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2622),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2622),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_API_AVAILABLE] = ACTIONS(2622),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_API_DEPRECATED] = ACTIONS(2622),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym___deprecated_msg] = ACTIONS(2622),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2622),
    [anon_sym_ATsynthesize] = ACTIONS(2624),
    [anon_sym_ATdynamic] = ACTIONS(2624),
    [anon_sym_BOOL] = ACTIONS(2622),
    [anon_sym_IMP] = ACTIONS(2622),
    [anon_sym_SEL] = ACTIONS(2622),
    [anon_sym_Class] = ACTIONS(2622),
    [anon_sym_id] = ACTIONS(2622),
  },
  [STATE(2235)] = {
    [sym_identifier] = ACTIONS(5068),
    [aux_sym_preproc_def_token1] = ACTIONS(5068),
    [aux_sym_preproc_if_token1] = ACTIONS(5068),
    [aux_sym_preproc_if_token2] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5068),
    [aux_sym_preproc_else_token1] = ACTIONS(5068),
    [aux_sym_preproc_elif_token1] = ACTIONS(5068),
    [sym_preproc_directive] = ACTIONS(5068),
    [anon_sym_DASH] = ACTIONS(5072),
    [anon_sym_PLUS] = ACTIONS(5072),
    [anon_sym___extension__] = ACTIONS(5068),
    [anon_sym_typedef] = ACTIONS(5068),
    [anon_sym_extern] = ACTIONS(5068),
    [anon_sym___attribute__] = ACTIONS(5068),
    [anon_sym___attribute] = ACTIONS(5068),
    [anon_sym_noreturn] = ACTIONS(5068),
    [anon_sym_LBRACK] = ACTIONS(5072),
    [anon_sym___declspec] = ACTIONS(5068),
    [anon_sym___cdecl] = ACTIONS(5068),
    [anon_sym___clrcall] = ACTIONS(5068),
    [anon_sym___stdcall] = ACTIONS(5068),
    [anon_sym___fastcall] = ACTIONS(5068),
    [anon_sym___thiscall] = ACTIONS(5068),
    [anon_sym___vectorcall] = ACTIONS(5068),
    [anon_sym_signed] = ACTIONS(5068),
    [anon_sym_unsigned] = ACTIONS(5068),
    [anon_sym_long] = ACTIONS(5068),
    [anon_sym_short] = ACTIONS(5068),
    [anon_sym_static] = ACTIONS(5068),
    [anon_sym_auto] = ACTIONS(5068),
    [anon_sym_register] = ACTIONS(5068),
    [anon_sym_inline] = ACTIONS(5068),
    [anon_sym___inline] = ACTIONS(5068),
    [anon_sym___inline__] = ACTIONS(5068),
    [anon_sym___forceinline] = ACTIONS(5068),
    [anon_sym_thread_local] = ACTIONS(5068),
    [anon_sym___thread] = ACTIONS(5068),
    [anon_sym_CG_EXTERN] = ACTIONS(5068),
    [anon_sym_CG_INLINE] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5068),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5068),
    [anon_sym_IBOutlet] = ACTIONS(5068),
    [anon_sym_IBInspectable] = ACTIONS(5068),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5068),
    [anon_sym_NS_INLINE] = ACTIONS(5068),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5068),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5068),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5068),
    [anon_sym_const] = ACTIONS(5068),
    [anon_sym_constexpr] = ACTIONS(5068),
    [anon_sym_volatile] = ACTIONS(5068),
    [anon_sym_restrict] = ACTIONS(5068),
    [anon_sym___restrict__] = ACTIONS(5068),
    [anon_sym__Atomic] = ACTIONS(5068),
    [anon_sym__Noreturn] = ACTIONS(5068),
    [anon_sym__Nonnull] = ACTIONS(5068),
    [anon_sym_nullable] = ACTIONS(5068),
    [anon_sym__Complex] = ACTIONS(5068),
    [anon_sym__Nullable] = ACTIONS(5068),
    [anon_sym__Nullable_result] = ACTIONS(5068),
    [anon_sym__Null_unspecified] = ACTIONS(5068),
    [anon_sym___autoreleasing] = ACTIONS(5068),
    [anon_sym___block] = ACTIONS(5068),
    [anon_sym___bridge] = ACTIONS(5068),
    [anon_sym___bridge_retained] = ACTIONS(5068),
    [anon_sym___bridge_transfer] = ACTIONS(5068),
    [anon_sym___complex] = ACTIONS(5068),
    [anon_sym___const] = ACTIONS(5068),
    [anon_sym___imag] = ACTIONS(5068),
    [anon_sym___kindof] = ACTIONS(5068),
    [anon_sym___nonnull] = ACTIONS(5068),
    [anon_sym___nullable] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5068),
    [anon_sym___real] = ACTIONS(5068),
    [anon_sym___strong] = ACTIONS(5068),
    [anon_sym___unsafe_unretained] = ACTIONS(5068),
    [anon_sym___unused] = ACTIONS(5068),
    [anon_sym___weak] = ACTIONS(5068),
    [anon_sym_alignas] = ACTIONS(5068),
    [anon_sym__Alignas] = ACTIONS(5068),
    [sym_primitive_type] = ACTIONS(5068),
    [anon_sym_enum] = ACTIONS(5068),
    [anon_sym_struct] = ACTIONS(5068),
    [anon_sym_union] = ACTIONS(5068),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5068),
    [anon_sym___typeof] = ACTIONS(5068),
    [anon_sym_typeof] = ACTIONS(5068),
    [aux_sym_preproc_undef_token1] = ACTIONS(5068),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5068),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5068),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5068),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_API_AVAILABLE] = ACTIONS(5068),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_API_DEPRECATED] = ACTIONS(5068),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym___deprecated_msg] = ACTIONS(5068),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5068),
    [anon_sym_ATsynthesize] = ACTIONS(5072),
    [anon_sym_ATdynamic] = ACTIONS(5072),
    [anon_sym_BOOL] = ACTIONS(5068),
    [anon_sym_IMP] = ACTIONS(5068),
    [anon_sym_SEL] = ACTIONS(5068),
    [anon_sym_Class] = ACTIONS(5068),
    [anon_sym_id] = ACTIONS(5068),
  },
  [STATE(2236)] = {
    [sym_identifier] = ACTIONS(5517),
    [aux_sym_preproc_def_token1] = ACTIONS(5517),
    [aux_sym_preproc_if_token1] = ACTIONS(5517),
    [aux_sym_preproc_if_token2] = ACTIONS(5517),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5517),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5517),
    [aux_sym_preproc_else_token1] = ACTIONS(5517),
    [aux_sym_preproc_elif_token1] = ACTIONS(5517),
    [sym_preproc_directive] = ACTIONS(5517),
    [anon_sym_DASH] = ACTIONS(5519),
    [anon_sym_PLUS] = ACTIONS(5519),
    [anon_sym___extension__] = ACTIONS(5517),
    [anon_sym_typedef] = ACTIONS(5517),
    [anon_sym_extern] = ACTIONS(5517),
    [anon_sym___attribute__] = ACTIONS(5517),
    [anon_sym___attribute] = ACTIONS(5517),
    [anon_sym_noreturn] = ACTIONS(5517),
    [anon_sym_LBRACK] = ACTIONS(5519),
    [anon_sym___declspec] = ACTIONS(5517),
    [anon_sym___cdecl] = ACTIONS(5517),
    [anon_sym___clrcall] = ACTIONS(5517),
    [anon_sym___stdcall] = ACTIONS(5517),
    [anon_sym___fastcall] = ACTIONS(5517),
    [anon_sym___thiscall] = ACTIONS(5517),
    [anon_sym___vectorcall] = ACTIONS(5517),
    [anon_sym_signed] = ACTIONS(5517),
    [anon_sym_unsigned] = ACTIONS(5517),
    [anon_sym_long] = ACTIONS(5517),
    [anon_sym_short] = ACTIONS(5517),
    [anon_sym_static] = ACTIONS(5517),
    [anon_sym_auto] = ACTIONS(5517),
    [anon_sym_register] = ACTIONS(5517),
    [anon_sym_inline] = ACTIONS(5517),
    [anon_sym___inline] = ACTIONS(5517),
    [anon_sym___inline__] = ACTIONS(5517),
    [anon_sym___forceinline] = ACTIONS(5517),
    [anon_sym_thread_local] = ACTIONS(5517),
    [anon_sym___thread] = ACTIONS(5517),
    [anon_sym_CG_EXTERN] = ACTIONS(5517),
    [anon_sym_CG_INLINE] = ACTIONS(5517),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5517),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5517),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5517),
    [anon_sym_IBOutlet] = ACTIONS(5517),
    [anon_sym_IBInspectable] = ACTIONS(5517),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5517),
    [anon_sym_NS_INLINE] = ACTIONS(5517),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5517),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5517),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5517),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5517),
    [anon_sym_const] = ACTIONS(5517),
    [anon_sym_constexpr] = ACTIONS(5517),
    [anon_sym_volatile] = ACTIONS(5517),
    [anon_sym_restrict] = ACTIONS(5517),
    [anon_sym___restrict__] = ACTIONS(5517),
    [anon_sym__Atomic] = ACTIONS(5517),
    [anon_sym__Noreturn] = ACTIONS(5517),
    [anon_sym__Nonnull] = ACTIONS(5517),
    [anon_sym_nullable] = ACTIONS(5517),
    [anon_sym__Complex] = ACTIONS(5517),
    [anon_sym__Nullable] = ACTIONS(5517),
    [anon_sym__Nullable_result] = ACTIONS(5517),
    [anon_sym__Null_unspecified] = ACTIONS(5517),
    [anon_sym___autoreleasing] = ACTIONS(5517),
    [anon_sym___block] = ACTIONS(5517),
    [anon_sym___bridge] = ACTIONS(5517),
    [anon_sym___bridge_retained] = ACTIONS(5517),
    [anon_sym___bridge_transfer] = ACTIONS(5517),
    [anon_sym___complex] = ACTIONS(5517),
    [anon_sym___const] = ACTIONS(5517),
    [anon_sym___imag] = ACTIONS(5517),
    [anon_sym___kindof] = ACTIONS(5517),
    [anon_sym___nonnull] = ACTIONS(5517),
    [anon_sym___nullable] = ACTIONS(5517),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5517),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5517),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5517),
    [anon_sym___real] = ACTIONS(5517),
    [anon_sym___strong] = ACTIONS(5517),
    [anon_sym___unsafe_unretained] = ACTIONS(5517),
    [anon_sym___unused] = ACTIONS(5517),
    [anon_sym___weak] = ACTIONS(5517),
    [anon_sym_alignas] = ACTIONS(5517),
    [anon_sym__Alignas] = ACTIONS(5517),
    [sym_primitive_type] = ACTIONS(5517),
    [anon_sym_enum] = ACTIONS(5517),
    [anon_sym_struct] = ACTIONS(5517),
    [anon_sym_union] = ACTIONS(5517),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5517),
    [anon_sym___typeof] = ACTIONS(5517),
    [anon_sym_typeof] = ACTIONS(5517),
    [aux_sym_preproc_undef_token1] = ACTIONS(5517),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5517),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5517),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5517),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5517),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5517),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5517),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5517),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5517),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5517),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_API_AVAILABLE] = ACTIONS(5517),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_API_DEPRECATED] = ACTIONS(5517),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5517),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5517),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5517),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5517),
    [anon_sym___deprecated_msg] = ACTIONS(5517),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5517),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5517),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5517),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5517),
    [anon_sym_ATsynthesize] = ACTIONS(5519),
    [anon_sym_ATdynamic] = ACTIONS(5519),
    [anon_sym_BOOL] = ACTIONS(5517),
    [anon_sym_IMP] = ACTIONS(5517),
    [anon_sym_SEL] = ACTIONS(5517),
    [anon_sym_Class] = ACTIONS(5517),
    [anon_sym_id] = ACTIONS(5517),
  },
  [STATE(2237)] = {
    [sym_identifier] = ACTIONS(2614),
    [aux_sym_preproc_def_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token2] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2614),
    [aux_sym_preproc_else_token1] = ACTIONS(2614),
    [aux_sym_preproc_elif_token1] = ACTIONS(2614),
    [sym_preproc_directive] = ACTIONS(2614),
    [anon_sym_DASH] = ACTIONS(2616),
    [anon_sym_PLUS] = ACTIONS(2616),
    [anon_sym___extension__] = ACTIONS(2614),
    [anon_sym_typedef] = ACTIONS(2614),
    [anon_sym_extern] = ACTIONS(2614),
    [anon_sym___attribute__] = ACTIONS(2614),
    [anon_sym___attribute] = ACTIONS(2614),
    [anon_sym_noreturn] = ACTIONS(2614),
    [anon_sym_LBRACK] = ACTIONS(2616),
    [anon_sym___declspec] = ACTIONS(2614),
    [anon_sym___cdecl] = ACTIONS(2614),
    [anon_sym___clrcall] = ACTIONS(2614),
    [anon_sym___stdcall] = ACTIONS(2614),
    [anon_sym___fastcall] = ACTIONS(2614),
    [anon_sym___thiscall] = ACTIONS(2614),
    [anon_sym___vectorcall] = ACTIONS(2614),
    [anon_sym_signed] = ACTIONS(2614),
    [anon_sym_unsigned] = ACTIONS(2614),
    [anon_sym_long] = ACTIONS(2614),
    [anon_sym_short] = ACTIONS(2614),
    [anon_sym_static] = ACTIONS(2614),
    [anon_sym_auto] = ACTIONS(2614),
    [anon_sym_register] = ACTIONS(2614),
    [anon_sym_inline] = ACTIONS(2614),
    [anon_sym___inline] = ACTIONS(2614),
    [anon_sym___inline__] = ACTIONS(2614),
    [anon_sym___forceinline] = ACTIONS(2614),
    [anon_sym_thread_local] = ACTIONS(2614),
    [anon_sym___thread] = ACTIONS(2614),
    [anon_sym_CG_EXTERN] = ACTIONS(2614),
    [anon_sym_CG_INLINE] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2614),
    [anon_sym_IBOutlet] = ACTIONS(2614),
    [anon_sym_IBInspectable] = ACTIONS(2614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2614),
    [anon_sym_NS_INLINE] = ACTIONS(2614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2614),
    [anon_sym_const] = ACTIONS(2614),
    [anon_sym_constexpr] = ACTIONS(2614),
    [anon_sym_volatile] = ACTIONS(2614),
    [anon_sym_restrict] = ACTIONS(2614),
    [anon_sym___restrict__] = ACTIONS(2614),
    [anon_sym__Atomic] = ACTIONS(2614),
    [anon_sym__Noreturn] = ACTIONS(2614),
    [anon_sym__Nonnull] = ACTIONS(2614),
    [anon_sym_nullable] = ACTIONS(2614),
    [anon_sym__Complex] = ACTIONS(2614),
    [anon_sym__Nullable] = ACTIONS(2614),
    [anon_sym__Nullable_result] = ACTIONS(2614),
    [anon_sym__Null_unspecified] = ACTIONS(2614),
    [anon_sym___autoreleasing] = ACTIONS(2614),
    [anon_sym___block] = ACTIONS(2614),
    [anon_sym___bridge] = ACTIONS(2614),
    [anon_sym___bridge_retained] = ACTIONS(2614),
    [anon_sym___bridge_transfer] = ACTIONS(2614),
    [anon_sym___complex] = ACTIONS(2614),
    [anon_sym___const] = ACTIONS(2614),
    [anon_sym___imag] = ACTIONS(2614),
    [anon_sym___kindof] = ACTIONS(2614),
    [anon_sym___nonnull] = ACTIONS(2614),
    [anon_sym___nullable] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2614),
    [anon_sym___real] = ACTIONS(2614),
    [anon_sym___strong] = ACTIONS(2614),
    [anon_sym___unsafe_unretained] = ACTIONS(2614),
    [anon_sym___unused] = ACTIONS(2614),
    [anon_sym___weak] = ACTIONS(2614),
    [anon_sym_alignas] = ACTIONS(2614),
    [anon_sym__Alignas] = ACTIONS(2614),
    [sym_primitive_type] = ACTIONS(2614),
    [anon_sym_enum] = ACTIONS(2614),
    [anon_sym_struct] = ACTIONS(2614),
    [anon_sym_union] = ACTIONS(2614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2614),
    [anon_sym___typeof] = ACTIONS(2614),
    [anon_sym_typeof] = ACTIONS(2614),
    [aux_sym_preproc_undef_token1] = ACTIONS(2614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_API_AVAILABLE] = ACTIONS(2614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_API_DEPRECATED] = ACTIONS(2614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym___deprecated_msg] = ACTIONS(2614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2614),
    [anon_sym_ATsynthesize] = ACTIONS(2616),
    [anon_sym_ATdynamic] = ACTIONS(2616),
    [anon_sym_BOOL] = ACTIONS(2614),
    [anon_sym_IMP] = ACTIONS(2614),
    [anon_sym_SEL] = ACTIONS(2614),
    [anon_sym_Class] = ACTIONS(2614),
    [anon_sym_id] = ACTIONS(2614),
  },
  [STATE(2238)] = {
    [sym_identifier] = ACTIONS(2350),
    [aux_sym_preproc_def_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token1] = ACTIONS(2350),
    [aux_sym_preproc_if_token2] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2350),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2350),
    [sym_preproc_directive] = ACTIONS(2350),
    [anon_sym_DASH] = ACTIONS(2352),
    [anon_sym_PLUS] = ACTIONS(2352),
    [anon_sym_SEMI] = ACTIONS(2352),
    [anon_sym___extension__] = ACTIONS(2350),
    [anon_sym_typedef] = ACTIONS(2350),
    [anon_sym_extern] = ACTIONS(2350),
    [anon_sym___attribute__] = ACTIONS(2350),
    [anon_sym___attribute] = ACTIONS(2350),
    [anon_sym_noreturn] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(2352),
    [anon_sym___declspec] = ACTIONS(2350),
    [anon_sym___cdecl] = ACTIONS(2350),
    [anon_sym___clrcall] = ACTIONS(2350),
    [anon_sym___stdcall] = ACTIONS(2350),
    [anon_sym___fastcall] = ACTIONS(2350),
    [anon_sym___thiscall] = ACTIONS(2350),
    [anon_sym___vectorcall] = ACTIONS(2350),
    [anon_sym_signed] = ACTIONS(2350),
    [anon_sym_unsigned] = ACTIONS(2350),
    [anon_sym_long] = ACTIONS(2350),
    [anon_sym_short] = ACTIONS(2350),
    [anon_sym_static] = ACTIONS(2350),
    [anon_sym_auto] = ACTIONS(2350),
    [anon_sym_register] = ACTIONS(2350),
    [anon_sym_inline] = ACTIONS(2350),
    [anon_sym___inline] = ACTIONS(2350),
    [anon_sym___inline__] = ACTIONS(2350),
    [anon_sym___forceinline] = ACTIONS(2350),
    [anon_sym_thread_local] = ACTIONS(2350),
    [anon_sym___thread] = ACTIONS(2350),
    [anon_sym_CG_EXTERN] = ACTIONS(2350),
    [anon_sym_CG_INLINE] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2350),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2350),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2350),
    [anon_sym_IBOutlet] = ACTIONS(2350),
    [anon_sym_IBInspectable] = ACTIONS(2350),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2350),
    [anon_sym_NS_INLINE] = ACTIONS(2350),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2350),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2350),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2350),
    [anon_sym_const] = ACTIONS(2350),
    [anon_sym_constexpr] = ACTIONS(2350),
    [anon_sym_volatile] = ACTIONS(2350),
    [anon_sym_restrict] = ACTIONS(2350),
    [anon_sym___restrict__] = ACTIONS(2350),
    [anon_sym__Atomic] = ACTIONS(2350),
    [anon_sym__Noreturn] = ACTIONS(2350),
    [anon_sym__Nonnull] = ACTIONS(2350),
    [anon_sym_nullable] = ACTIONS(2350),
    [anon_sym__Complex] = ACTIONS(2350),
    [anon_sym__Nullable] = ACTIONS(2350),
    [anon_sym__Nullable_result] = ACTIONS(2350),
    [anon_sym__Null_unspecified] = ACTIONS(2350),
    [anon_sym___autoreleasing] = ACTIONS(2350),
    [anon_sym___block] = ACTIONS(2350),
    [anon_sym___bridge] = ACTIONS(2350),
    [anon_sym___bridge_retained] = ACTIONS(2350),
    [anon_sym___bridge_transfer] = ACTIONS(2350),
    [anon_sym___complex] = ACTIONS(2350),
    [anon_sym___const] = ACTIONS(2350),
    [anon_sym___imag] = ACTIONS(2350),
    [anon_sym___kindof] = ACTIONS(2350),
    [anon_sym___nonnull] = ACTIONS(2350),
    [anon_sym___nullable] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2350),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2350),
    [anon_sym___real] = ACTIONS(2350),
    [anon_sym___strong] = ACTIONS(2350),
    [anon_sym___unsafe_unretained] = ACTIONS(2350),
    [anon_sym___unused] = ACTIONS(2350),
    [anon_sym___weak] = ACTIONS(2350),
    [anon_sym_alignas] = ACTIONS(2350),
    [anon_sym__Alignas] = ACTIONS(2350),
    [sym_primitive_type] = ACTIONS(2350),
    [anon_sym_enum] = ACTIONS(2350),
    [anon_sym_struct] = ACTIONS(2350),
    [anon_sym_union] = ACTIONS(2350),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2350),
    [anon_sym___typeof] = ACTIONS(2350),
    [anon_sym_typeof] = ACTIONS(2350),
    [aux_sym_preproc_undef_token1] = ACTIONS(2350),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2350),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2350),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2350),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2350),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2350),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_API_AVAILABLE] = ACTIONS(2350),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_API_DEPRECATED] = ACTIONS(2350),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2350),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2350),
    [anon_sym___deprecated_msg] = ACTIONS(2350),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2350),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2350),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2350),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2350),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2350),
    [anon_sym_ATsynthesize] = ACTIONS(2352),
    [anon_sym_ATdynamic] = ACTIONS(2352),
    [anon_sym_ATproperty] = ACTIONS(2352),
    [anon_sym_BOOL] = ACTIONS(2350),
    [anon_sym_IMP] = ACTIONS(2350),
    [anon_sym_SEL] = ACTIONS(2350),
    [anon_sym_Class] = ACTIONS(2350),
    [anon_sym_id] = ACTIONS(2350),
  },
  [STATE(2239)] = {
    [sym_identifier] = ACTIONS(5521),
    [aux_sym_preproc_def_token1] = ACTIONS(5521),
    [aux_sym_preproc_if_token1] = ACTIONS(5521),
    [aux_sym_preproc_if_token2] = ACTIONS(5521),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5521),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5521),
    [aux_sym_preproc_else_token1] = ACTIONS(5521),
    [aux_sym_preproc_elif_token1] = ACTIONS(5521),
    [sym_preproc_directive] = ACTIONS(5521),
    [anon_sym_DASH] = ACTIONS(5523),
    [anon_sym_PLUS] = ACTIONS(5523),
    [anon_sym___extension__] = ACTIONS(5521),
    [anon_sym_typedef] = ACTIONS(5521),
    [anon_sym_extern] = ACTIONS(5521),
    [anon_sym___attribute__] = ACTIONS(5521),
    [anon_sym___attribute] = ACTIONS(5521),
    [anon_sym_noreturn] = ACTIONS(5521),
    [anon_sym_LBRACK] = ACTIONS(5523),
    [anon_sym___declspec] = ACTIONS(5521),
    [anon_sym___cdecl] = ACTIONS(5521),
    [anon_sym___clrcall] = ACTIONS(5521),
    [anon_sym___stdcall] = ACTIONS(5521),
    [anon_sym___fastcall] = ACTIONS(5521),
    [anon_sym___thiscall] = ACTIONS(5521),
    [anon_sym___vectorcall] = ACTIONS(5521),
    [anon_sym_signed] = ACTIONS(5521),
    [anon_sym_unsigned] = ACTIONS(5521),
    [anon_sym_long] = ACTIONS(5521),
    [anon_sym_short] = ACTIONS(5521),
    [anon_sym_static] = ACTIONS(5521),
    [anon_sym_auto] = ACTIONS(5521),
    [anon_sym_register] = ACTIONS(5521),
    [anon_sym_inline] = ACTIONS(5521),
    [anon_sym___inline] = ACTIONS(5521),
    [anon_sym___inline__] = ACTIONS(5521),
    [anon_sym___forceinline] = ACTIONS(5521),
    [anon_sym_thread_local] = ACTIONS(5521),
    [anon_sym___thread] = ACTIONS(5521),
    [anon_sym_CG_EXTERN] = ACTIONS(5521),
    [anon_sym_CG_INLINE] = ACTIONS(5521),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5521),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5521),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5521),
    [anon_sym_IBOutlet] = ACTIONS(5521),
    [anon_sym_IBInspectable] = ACTIONS(5521),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5521),
    [anon_sym_NS_INLINE] = ACTIONS(5521),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5521),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5521),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5521),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5521),
    [anon_sym_const] = ACTIONS(5521),
    [anon_sym_constexpr] = ACTIONS(5521),
    [anon_sym_volatile] = ACTIONS(5521),
    [anon_sym_restrict] = ACTIONS(5521),
    [anon_sym___restrict__] = ACTIONS(5521),
    [anon_sym__Atomic] = ACTIONS(5521),
    [anon_sym__Noreturn] = ACTIONS(5521),
    [anon_sym__Nonnull] = ACTIONS(5521),
    [anon_sym_nullable] = ACTIONS(5521),
    [anon_sym__Complex] = ACTIONS(5521),
    [anon_sym__Nullable] = ACTIONS(5521),
    [anon_sym__Nullable_result] = ACTIONS(5521),
    [anon_sym__Null_unspecified] = ACTIONS(5521),
    [anon_sym___autoreleasing] = ACTIONS(5521),
    [anon_sym___block] = ACTIONS(5521),
    [anon_sym___bridge] = ACTIONS(5521),
    [anon_sym___bridge_retained] = ACTIONS(5521),
    [anon_sym___bridge_transfer] = ACTIONS(5521),
    [anon_sym___complex] = ACTIONS(5521),
    [anon_sym___const] = ACTIONS(5521),
    [anon_sym___imag] = ACTIONS(5521),
    [anon_sym___kindof] = ACTIONS(5521),
    [anon_sym___nonnull] = ACTIONS(5521),
    [anon_sym___nullable] = ACTIONS(5521),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5521),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5521),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5521),
    [anon_sym___real] = ACTIONS(5521),
    [anon_sym___strong] = ACTIONS(5521),
    [anon_sym___unsafe_unretained] = ACTIONS(5521),
    [anon_sym___unused] = ACTIONS(5521),
    [anon_sym___weak] = ACTIONS(5521),
    [anon_sym_alignas] = ACTIONS(5521),
    [anon_sym__Alignas] = ACTIONS(5521),
    [sym_primitive_type] = ACTIONS(5521),
    [anon_sym_enum] = ACTIONS(5521),
    [anon_sym_struct] = ACTIONS(5521),
    [anon_sym_union] = ACTIONS(5521),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5521),
    [anon_sym___typeof] = ACTIONS(5521),
    [anon_sym_typeof] = ACTIONS(5521),
    [aux_sym_preproc_undef_token1] = ACTIONS(5521),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5521),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5521),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5521),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5521),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5521),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5521),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5521),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5521),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5521),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_API_AVAILABLE] = ACTIONS(5521),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_API_DEPRECATED] = ACTIONS(5521),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5521),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5521),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5521),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5521),
    [anon_sym___deprecated_msg] = ACTIONS(5521),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5521),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5521),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5521),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5521),
    [anon_sym_ATsynthesize] = ACTIONS(5523),
    [anon_sym_ATdynamic] = ACTIONS(5523),
    [anon_sym_BOOL] = ACTIONS(5521),
    [anon_sym_IMP] = ACTIONS(5521),
    [anon_sym_SEL] = ACTIONS(5521),
    [anon_sym_Class] = ACTIONS(5521),
    [anon_sym_id] = ACTIONS(5521),
  },
  [STATE(2240)] = {
    [sym_identifier] = ACTIONS(5405),
    [aux_sym_preproc_def_token1] = ACTIONS(5405),
    [aux_sym_preproc_if_token1] = ACTIONS(5405),
    [aux_sym_preproc_if_token2] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5405),
    [aux_sym_preproc_else_token1] = ACTIONS(5405),
    [aux_sym_preproc_elif_token1] = ACTIONS(5405),
    [sym_preproc_directive] = ACTIONS(5405),
    [anon_sym_DASH] = ACTIONS(5407),
    [anon_sym_PLUS] = ACTIONS(5407),
    [anon_sym___extension__] = ACTIONS(5405),
    [anon_sym_typedef] = ACTIONS(5405),
    [anon_sym_extern] = ACTIONS(5405),
    [anon_sym___attribute__] = ACTIONS(5405),
    [anon_sym___attribute] = ACTIONS(5405),
    [anon_sym_noreturn] = ACTIONS(5405),
    [anon_sym_LBRACK] = ACTIONS(5407),
    [anon_sym___declspec] = ACTIONS(5405),
    [anon_sym___cdecl] = ACTIONS(5405),
    [anon_sym___clrcall] = ACTIONS(5405),
    [anon_sym___stdcall] = ACTIONS(5405),
    [anon_sym___fastcall] = ACTIONS(5405),
    [anon_sym___thiscall] = ACTIONS(5405),
    [anon_sym___vectorcall] = ACTIONS(5405),
    [anon_sym_signed] = ACTIONS(5405),
    [anon_sym_unsigned] = ACTIONS(5405),
    [anon_sym_long] = ACTIONS(5405),
    [anon_sym_short] = ACTIONS(5405),
    [anon_sym_static] = ACTIONS(5405),
    [anon_sym_auto] = ACTIONS(5405),
    [anon_sym_register] = ACTIONS(5405),
    [anon_sym_inline] = ACTIONS(5405),
    [anon_sym___inline] = ACTIONS(5405),
    [anon_sym___inline__] = ACTIONS(5405),
    [anon_sym___forceinline] = ACTIONS(5405),
    [anon_sym_thread_local] = ACTIONS(5405),
    [anon_sym___thread] = ACTIONS(5405),
    [anon_sym_CG_EXTERN] = ACTIONS(5405),
    [anon_sym_CG_INLINE] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5405),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5405),
    [anon_sym_IBOutlet] = ACTIONS(5405),
    [anon_sym_IBInspectable] = ACTIONS(5405),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5405),
    [anon_sym_NS_INLINE] = ACTIONS(5405),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5405),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5405),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5405),
    [anon_sym_const] = ACTIONS(5405),
    [anon_sym_constexpr] = ACTIONS(5405),
    [anon_sym_volatile] = ACTIONS(5405),
    [anon_sym_restrict] = ACTIONS(5405),
    [anon_sym___restrict__] = ACTIONS(5405),
    [anon_sym__Atomic] = ACTIONS(5405),
    [anon_sym__Noreturn] = ACTIONS(5405),
    [anon_sym__Nonnull] = ACTIONS(5405),
    [anon_sym_nullable] = ACTIONS(5405),
    [anon_sym__Complex] = ACTIONS(5405),
    [anon_sym__Nullable] = ACTIONS(5405),
    [anon_sym__Nullable_result] = ACTIONS(5405),
    [anon_sym__Null_unspecified] = ACTIONS(5405),
    [anon_sym___autoreleasing] = ACTIONS(5405),
    [anon_sym___block] = ACTIONS(5405),
    [anon_sym___bridge] = ACTIONS(5405),
    [anon_sym___bridge_retained] = ACTIONS(5405),
    [anon_sym___bridge_transfer] = ACTIONS(5405),
    [anon_sym___complex] = ACTIONS(5405),
    [anon_sym___const] = ACTIONS(5405),
    [anon_sym___imag] = ACTIONS(5405),
    [anon_sym___kindof] = ACTIONS(5405),
    [anon_sym___nonnull] = ACTIONS(5405),
    [anon_sym___nullable] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5405),
    [anon_sym___real] = ACTIONS(5405),
    [anon_sym___strong] = ACTIONS(5405),
    [anon_sym___unsafe_unretained] = ACTIONS(5405),
    [anon_sym___unused] = ACTIONS(5405),
    [anon_sym___weak] = ACTIONS(5405),
    [anon_sym_alignas] = ACTIONS(5405),
    [anon_sym__Alignas] = ACTIONS(5405),
    [sym_primitive_type] = ACTIONS(5405),
    [anon_sym_enum] = ACTIONS(5405),
    [anon_sym_struct] = ACTIONS(5405),
    [anon_sym_union] = ACTIONS(5405),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5405),
    [anon_sym___typeof] = ACTIONS(5405),
    [anon_sym_typeof] = ACTIONS(5405),
    [aux_sym_preproc_undef_token1] = ACTIONS(5405),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5405),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5405),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5405),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_API_AVAILABLE] = ACTIONS(5405),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_API_DEPRECATED] = ACTIONS(5405),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym___deprecated_msg] = ACTIONS(5405),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5405),
    [anon_sym_ATsynthesize] = ACTIONS(5407),
    [anon_sym_ATdynamic] = ACTIONS(5407),
    [anon_sym_BOOL] = ACTIONS(5405),
    [anon_sym_IMP] = ACTIONS(5405),
    [anon_sym_SEL] = ACTIONS(5405),
    [anon_sym_Class] = ACTIONS(5405),
    [anon_sym_id] = ACTIONS(5405),
  },
  [STATE(2241)] = {
    [sym_identifier] = ACTIONS(5351),
    [aux_sym_preproc_def_token1] = ACTIONS(5351),
    [aux_sym_preproc_if_token1] = ACTIONS(5351),
    [aux_sym_preproc_if_token2] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5351),
    [aux_sym_preproc_else_token1] = ACTIONS(5351),
    [aux_sym_preproc_elif_token1] = ACTIONS(5351),
    [sym_preproc_directive] = ACTIONS(5351),
    [anon_sym_DASH] = ACTIONS(5353),
    [anon_sym_PLUS] = ACTIONS(5353),
    [anon_sym___extension__] = ACTIONS(5351),
    [anon_sym_typedef] = ACTIONS(5351),
    [anon_sym_extern] = ACTIONS(5351),
    [anon_sym___attribute__] = ACTIONS(5351),
    [anon_sym___attribute] = ACTIONS(5351),
    [anon_sym_noreturn] = ACTIONS(5351),
    [anon_sym_LBRACK] = ACTIONS(5353),
    [anon_sym___declspec] = ACTIONS(5351),
    [anon_sym___cdecl] = ACTIONS(5351),
    [anon_sym___clrcall] = ACTIONS(5351),
    [anon_sym___stdcall] = ACTIONS(5351),
    [anon_sym___fastcall] = ACTIONS(5351),
    [anon_sym___thiscall] = ACTIONS(5351),
    [anon_sym___vectorcall] = ACTIONS(5351),
    [anon_sym_signed] = ACTIONS(5351),
    [anon_sym_unsigned] = ACTIONS(5351),
    [anon_sym_long] = ACTIONS(5351),
    [anon_sym_short] = ACTIONS(5351),
    [anon_sym_static] = ACTIONS(5351),
    [anon_sym_auto] = ACTIONS(5351),
    [anon_sym_register] = ACTIONS(5351),
    [anon_sym_inline] = ACTIONS(5351),
    [anon_sym___inline] = ACTIONS(5351),
    [anon_sym___inline__] = ACTIONS(5351),
    [anon_sym___forceinline] = ACTIONS(5351),
    [anon_sym_thread_local] = ACTIONS(5351),
    [anon_sym___thread] = ACTIONS(5351),
    [anon_sym_CG_EXTERN] = ACTIONS(5351),
    [anon_sym_CG_INLINE] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5351),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5351),
    [anon_sym_IBOutlet] = ACTIONS(5351),
    [anon_sym_IBInspectable] = ACTIONS(5351),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5351),
    [anon_sym_NS_INLINE] = ACTIONS(5351),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5351),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5351),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5351),
    [anon_sym_const] = ACTIONS(5351),
    [anon_sym_constexpr] = ACTIONS(5351),
    [anon_sym_volatile] = ACTIONS(5351),
    [anon_sym_restrict] = ACTIONS(5351),
    [anon_sym___restrict__] = ACTIONS(5351),
    [anon_sym__Atomic] = ACTIONS(5351),
    [anon_sym__Noreturn] = ACTIONS(5351),
    [anon_sym__Nonnull] = ACTIONS(5351),
    [anon_sym_nullable] = ACTIONS(5351),
    [anon_sym__Complex] = ACTIONS(5351),
    [anon_sym__Nullable] = ACTIONS(5351),
    [anon_sym__Nullable_result] = ACTIONS(5351),
    [anon_sym__Null_unspecified] = ACTIONS(5351),
    [anon_sym___autoreleasing] = ACTIONS(5351),
    [anon_sym___block] = ACTIONS(5351),
    [anon_sym___bridge] = ACTIONS(5351),
    [anon_sym___bridge_retained] = ACTIONS(5351),
    [anon_sym___bridge_transfer] = ACTIONS(5351),
    [anon_sym___complex] = ACTIONS(5351),
    [anon_sym___const] = ACTIONS(5351),
    [anon_sym___imag] = ACTIONS(5351),
    [anon_sym___kindof] = ACTIONS(5351),
    [anon_sym___nonnull] = ACTIONS(5351),
    [anon_sym___nullable] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5351),
    [anon_sym___real] = ACTIONS(5351),
    [anon_sym___strong] = ACTIONS(5351),
    [anon_sym___unsafe_unretained] = ACTIONS(5351),
    [anon_sym___unused] = ACTIONS(5351),
    [anon_sym___weak] = ACTIONS(5351),
    [anon_sym_alignas] = ACTIONS(5351),
    [anon_sym__Alignas] = ACTIONS(5351),
    [sym_primitive_type] = ACTIONS(5351),
    [anon_sym_enum] = ACTIONS(5351),
    [anon_sym_struct] = ACTIONS(5351),
    [anon_sym_union] = ACTIONS(5351),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5351),
    [anon_sym___typeof] = ACTIONS(5351),
    [anon_sym_typeof] = ACTIONS(5351),
    [aux_sym_preproc_undef_token1] = ACTIONS(5351),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5351),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5351),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5351),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_API_AVAILABLE] = ACTIONS(5351),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_API_DEPRECATED] = ACTIONS(5351),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym___deprecated_msg] = ACTIONS(5351),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5351),
    [anon_sym_ATsynthesize] = ACTIONS(5353),
    [anon_sym_ATdynamic] = ACTIONS(5353),
    [anon_sym_BOOL] = ACTIONS(5351),
    [anon_sym_IMP] = ACTIONS(5351),
    [anon_sym_SEL] = ACTIONS(5351),
    [anon_sym_Class] = ACTIONS(5351),
    [anon_sym_id] = ACTIONS(5351),
  },
  [STATE(2242)] = {
    [sym_identifier] = ACTIONS(2600),
    [aux_sym_preproc_def_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token2] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2600),
    [aux_sym_preproc_else_token1] = ACTIONS(2600),
    [aux_sym_preproc_elif_token1] = ACTIONS(2600),
    [sym_preproc_directive] = ACTIONS(2600),
    [anon_sym_DASH] = ACTIONS(2602),
    [anon_sym_PLUS] = ACTIONS(2602),
    [anon_sym___extension__] = ACTIONS(2600),
    [anon_sym_typedef] = ACTIONS(2600),
    [anon_sym_extern] = ACTIONS(2600),
    [anon_sym___attribute__] = ACTIONS(2600),
    [anon_sym___attribute] = ACTIONS(2600),
    [anon_sym_noreturn] = ACTIONS(2600),
    [anon_sym_LBRACK] = ACTIONS(2602),
    [anon_sym___declspec] = ACTIONS(2600),
    [anon_sym___cdecl] = ACTIONS(2600),
    [anon_sym___clrcall] = ACTIONS(2600),
    [anon_sym___stdcall] = ACTIONS(2600),
    [anon_sym___fastcall] = ACTIONS(2600),
    [anon_sym___thiscall] = ACTIONS(2600),
    [anon_sym___vectorcall] = ACTIONS(2600),
    [anon_sym_signed] = ACTIONS(2600),
    [anon_sym_unsigned] = ACTIONS(2600),
    [anon_sym_long] = ACTIONS(2600),
    [anon_sym_short] = ACTIONS(2600),
    [anon_sym_static] = ACTIONS(2600),
    [anon_sym_auto] = ACTIONS(2600),
    [anon_sym_register] = ACTIONS(2600),
    [anon_sym_inline] = ACTIONS(2600),
    [anon_sym___inline] = ACTIONS(2600),
    [anon_sym___inline__] = ACTIONS(2600),
    [anon_sym___forceinline] = ACTIONS(2600),
    [anon_sym_thread_local] = ACTIONS(2600),
    [anon_sym___thread] = ACTIONS(2600),
    [anon_sym_CG_EXTERN] = ACTIONS(2600),
    [anon_sym_CG_INLINE] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2600),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2600),
    [anon_sym_IBOutlet] = ACTIONS(2600),
    [anon_sym_IBInspectable] = ACTIONS(2600),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2600),
    [anon_sym_NS_INLINE] = ACTIONS(2600),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2600),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2600),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2600),
    [anon_sym_const] = ACTIONS(2600),
    [anon_sym_constexpr] = ACTIONS(2600),
    [anon_sym_volatile] = ACTIONS(2600),
    [anon_sym_restrict] = ACTIONS(2600),
    [anon_sym___restrict__] = ACTIONS(2600),
    [anon_sym__Atomic] = ACTIONS(2600),
    [anon_sym__Noreturn] = ACTIONS(2600),
    [anon_sym__Nonnull] = ACTIONS(2600),
    [anon_sym_nullable] = ACTIONS(2600),
    [anon_sym__Complex] = ACTIONS(2600),
    [anon_sym__Nullable] = ACTIONS(2600),
    [anon_sym__Nullable_result] = ACTIONS(2600),
    [anon_sym__Null_unspecified] = ACTIONS(2600),
    [anon_sym___autoreleasing] = ACTIONS(2600),
    [anon_sym___block] = ACTIONS(2600),
    [anon_sym___bridge] = ACTIONS(2600),
    [anon_sym___bridge_retained] = ACTIONS(2600),
    [anon_sym___bridge_transfer] = ACTIONS(2600),
    [anon_sym___complex] = ACTIONS(2600),
    [anon_sym___const] = ACTIONS(2600),
    [anon_sym___imag] = ACTIONS(2600),
    [anon_sym___kindof] = ACTIONS(2600),
    [anon_sym___nonnull] = ACTIONS(2600),
    [anon_sym___nullable] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2600),
    [anon_sym___real] = ACTIONS(2600),
    [anon_sym___strong] = ACTIONS(2600),
    [anon_sym___unsafe_unretained] = ACTIONS(2600),
    [anon_sym___unused] = ACTIONS(2600),
    [anon_sym___weak] = ACTIONS(2600),
    [anon_sym_alignas] = ACTIONS(2600),
    [anon_sym__Alignas] = ACTIONS(2600),
    [sym_primitive_type] = ACTIONS(2600),
    [anon_sym_enum] = ACTIONS(2600),
    [anon_sym_struct] = ACTIONS(2600),
    [anon_sym_union] = ACTIONS(2600),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2600),
    [anon_sym___typeof] = ACTIONS(2600),
    [anon_sym_typeof] = ACTIONS(2600),
    [aux_sym_preproc_undef_token1] = ACTIONS(2600),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2600),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2600),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2600),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_API_AVAILABLE] = ACTIONS(2600),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_API_DEPRECATED] = ACTIONS(2600),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym___deprecated_msg] = ACTIONS(2600),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2600),
    [anon_sym_ATsynthesize] = ACTIONS(2602),
    [anon_sym_ATdynamic] = ACTIONS(2602),
    [anon_sym_BOOL] = ACTIONS(2600),
    [anon_sym_IMP] = ACTIONS(2600),
    [anon_sym_SEL] = ACTIONS(2600),
    [anon_sym_Class] = ACTIONS(2600),
    [anon_sym_id] = ACTIONS(2600),
  },
  [STATE(2243)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_DASH] = ACTIONS(2430),
    [anon_sym_PLUS] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATproperty] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(2244)] = {
    [sym_identifier] = ACTIONS(5525),
    [aux_sym_preproc_def_token1] = ACTIONS(5525),
    [aux_sym_preproc_if_token1] = ACTIONS(5525),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5525),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5525),
    [sym_preproc_directive] = ACTIONS(5525),
    [anon_sym_DASH] = ACTIONS(5527),
    [anon_sym_PLUS] = ACTIONS(5527),
    [anon_sym_LT] = ACTIONS(5527),
    [anon_sym___extension__] = ACTIONS(5525),
    [anon_sym_typedef] = ACTIONS(5525),
    [anon_sym_extern] = ACTIONS(5525),
    [anon_sym___attribute__] = ACTIONS(5525),
    [anon_sym___attribute] = ACTIONS(5525),
    [anon_sym_noreturn] = ACTIONS(5525),
    [anon_sym_LBRACK] = ACTIONS(5527),
    [anon_sym___declspec] = ACTIONS(5525),
    [anon_sym___cdecl] = ACTIONS(5525),
    [anon_sym___clrcall] = ACTIONS(5525),
    [anon_sym___stdcall] = ACTIONS(5525),
    [anon_sym___fastcall] = ACTIONS(5525),
    [anon_sym___thiscall] = ACTIONS(5525),
    [anon_sym___vectorcall] = ACTIONS(5525),
    [anon_sym_LBRACE] = ACTIONS(5527),
    [anon_sym_signed] = ACTIONS(5525),
    [anon_sym_unsigned] = ACTIONS(5525),
    [anon_sym_long] = ACTIONS(5525),
    [anon_sym_short] = ACTIONS(5525),
    [anon_sym_static] = ACTIONS(5525),
    [anon_sym_auto] = ACTIONS(5525),
    [anon_sym_register] = ACTIONS(5525),
    [anon_sym_inline] = ACTIONS(5525),
    [anon_sym___inline] = ACTIONS(5525),
    [anon_sym___inline__] = ACTIONS(5525),
    [anon_sym___forceinline] = ACTIONS(5525),
    [anon_sym_thread_local] = ACTIONS(5525),
    [anon_sym___thread] = ACTIONS(5525),
    [anon_sym_CG_EXTERN] = ACTIONS(5525),
    [anon_sym_CG_INLINE] = ACTIONS(5525),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5525),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5525),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5525),
    [anon_sym_IBOutlet] = ACTIONS(5525),
    [anon_sym_IBInspectable] = ACTIONS(5525),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5525),
    [anon_sym_NS_INLINE] = ACTIONS(5525),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5525),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5525),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5525),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5525),
    [anon_sym_const] = ACTIONS(5525),
    [anon_sym_constexpr] = ACTIONS(5525),
    [anon_sym_volatile] = ACTIONS(5525),
    [anon_sym_restrict] = ACTIONS(5525),
    [anon_sym___restrict__] = ACTIONS(5525),
    [anon_sym__Atomic] = ACTIONS(5525),
    [anon_sym__Noreturn] = ACTIONS(5525),
    [anon_sym__Nonnull] = ACTIONS(5525),
    [anon_sym_nullable] = ACTIONS(5525),
    [anon_sym__Complex] = ACTIONS(5525),
    [anon_sym__Nullable] = ACTIONS(5525),
    [anon_sym__Nullable_result] = ACTIONS(5525),
    [anon_sym__Null_unspecified] = ACTIONS(5525),
    [anon_sym___autoreleasing] = ACTIONS(5525),
    [anon_sym___block] = ACTIONS(5525),
    [anon_sym___bridge] = ACTIONS(5525),
    [anon_sym___bridge_retained] = ACTIONS(5525),
    [anon_sym___bridge_transfer] = ACTIONS(5525),
    [anon_sym___complex] = ACTIONS(5525),
    [anon_sym___const] = ACTIONS(5525),
    [anon_sym___imag] = ACTIONS(5525),
    [anon_sym___kindof] = ACTIONS(5525),
    [anon_sym___nonnull] = ACTIONS(5525),
    [anon_sym___nullable] = ACTIONS(5525),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5525),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5525),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5525),
    [anon_sym___real] = ACTIONS(5525),
    [anon_sym___strong] = ACTIONS(5525),
    [anon_sym___unsafe_unretained] = ACTIONS(5525),
    [anon_sym___unused] = ACTIONS(5525),
    [anon_sym___weak] = ACTIONS(5525),
    [anon_sym_alignas] = ACTIONS(5525),
    [anon_sym__Alignas] = ACTIONS(5525),
    [sym_primitive_type] = ACTIONS(5525),
    [anon_sym_enum] = ACTIONS(5525),
    [anon_sym_struct] = ACTIONS(5525),
    [anon_sym_union] = ACTIONS(5525),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5525),
    [anon_sym___typeof] = ACTIONS(5525),
    [anon_sym_typeof] = ACTIONS(5525),
    [aux_sym_preproc_undef_token1] = ACTIONS(5525),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5525),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5525),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5525),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5525),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5525),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5525),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5525),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5525),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5525),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5525),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5525),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5525),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5525),
    [anon_sym_API_AVAILABLE] = ACTIONS(5525),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5525),
    [anon_sym_API_DEPRECATED] = ACTIONS(5525),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5525),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5525),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5525),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5525),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5525),
    [anon_sym___deprecated_msg] = ACTIONS(5525),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5525),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5525),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5525),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5525),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5525),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5525),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5525),
    [anon_sym_ATend] = ACTIONS(5527),
    [anon_sym_ATproperty] = ACTIONS(5527),
    [anon_sym_BOOL] = ACTIONS(5525),
    [anon_sym_IMP] = ACTIONS(5525),
    [anon_sym_SEL] = ACTIONS(5525),
    [anon_sym_Class] = ACTIONS(5525),
    [anon_sym_id] = ACTIONS(5525),
  },
  [STATE(2245)] = {
    [sym_identifier] = ACTIONS(5529),
    [aux_sym_preproc_def_token1] = ACTIONS(5529),
    [aux_sym_preproc_if_token1] = ACTIONS(5529),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5529),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5529),
    [sym_preproc_directive] = ACTIONS(5529),
    [anon_sym_DASH] = ACTIONS(5531),
    [anon_sym_PLUS] = ACTIONS(5531),
    [anon_sym___extension__] = ACTIONS(5529),
    [anon_sym_typedef] = ACTIONS(5529),
    [anon_sym_extern] = ACTIONS(5529),
    [anon_sym___attribute__] = ACTIONS(5529),
    [anon_sym___attribute] = ACTIONS(5529),
    [anon_sym_noreturn] = ACTIONS(5529),
    [anon_sym_LBRACK] = ACTIONS(5531),
    [anon_sym___declspec] = ACTIONS(5529),
    [anon_sym___cdecl] = ACTIONS(5529),
    [anon_sym___clrcall] = ACTIONS(5529),
    [anon_sym___stdcall] = ACTIONS(5529),
    [anon_sym___fastcall] = ACTIONS(5529),
    [anon_sym___thiscall] = ACTIONS(5529),
    [anon_sym___vectorcall] = ACTIONS(5529),
    [anon_sym_signed] = ACTIONS(5529),
    [anon_sym_unsigned] = ACTIONS(5529),
    [anon_sym_long] = ACTIONS(5529),
    [anon_sym_short] = ACTIONS(5529),
    [anon_sym_static] = ACTIONS(5529),
    [anon_sym_auto] = ACTIONS(5529),
    [anon_sym_register] = ACTIONS(5529),
    [anon_sym_inline] = ACTIONS(5529),
    [anon_sym___inline] = ACTIONS(5529),
    [anon_sym___inline__] = ACTIONS(5529),
    [anon_sym___forceinline] = ACTIONS(5529),
    [anon_sym_thread_local] = ACTIONS(5529),
    [anon_sym___thread] = ACTIONS(5529),
    [anon_sym_CG_EXTERN] = ACTIONS(5529),
    [anon_sym_CG_INLINE] = ACTIONS(5529),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5529),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5529),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5529),
    [anon_sym_IBOutlet] = ACTIONS(5529),
    [anon_sym_IBInspectable] = ACTIONS(5529),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5529),
    [anon_sym_NS_INLINE] = ACTIONS(5529),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5529),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5529),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5529),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5529),
    [anon_sym_const] = ACTIONS(5529),
    [anon_sym_constexpr] = ACTIONS(5529),
    [anon_sym_volatile] = ACTIONS(5529),
    [anon_sym_restrict] = ACTIONS(5529),
    [anon_sym___restrict__] = ACTIONS(5529),
    [anon_sym__Atomic] = ACTIONS(5529),
    [anon_sym__Noreturn] = ACTIONS(5529),
    [anon_sym__Nonnull] = ACTIONS(5529),
    [anon_sym_nullable] = ACTIONS(5529),
    [anon_sym__Complex] = ACTIONS(5529),
    [anon_sym__Nullable] = ACTIONS(5529),
    [anon_sym__Nullable_result] = ACTIONS(5529),
    [anon_sym__Null_unspecified] = ACTIONS(5529),
    [anon_sym___autoreleasing] = ACTIONS(5529),
    [anon_sym___block] = ACTIONS(5529),
    [anon_sym___bridge] = ACTIONS(5529),
    [anon_sym___bridge_retained] = ACTIONS(5529),
    [anon_sym___bridge_transfer] = ACTIONS(5529),
    [anon_sym___complex] = ACTIONS(5529),
    [anon_sym___const] = ACTIONS(5529),
    [anon_sym___imag] = ACTIONS(5529),
    [anon_sym___kindof] = ACTIONS(5529),
    [anon_sym___nonnull] = ACTIONS(5529),
    [anon_sym___nullable] = ACTIONS(5529),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5529),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5529),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5529),
    [anon_sym___real] = ACTIONS(5529),
    [anon_sym___strong] = ACTIONS(5529),
    [anon_sym___unsafe_unretained] = ACTIONS(5529),
    [anon_sym___unused] = ACTIONS(5529),
    [anon_sym___weak] = ACTIONS(5529),
    [anon_sym_alignas] = ACTIONS(5529),
    [anon_sym__Alignas] = ACTIONS(5529),
    [sym_primitive_type] = ACTIONS(5529),
    [anon_sym_enum] = ACTIONS(5529),
    [anon_sym_struct] = ACTIONS(5529),
    [anon_sym_union] = ACTIONS(5529),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5529),
    [anon_sym___typeof] = ACTIONS(5529),
    [anon_sym_typeof] = ACTIONS(5529),
    [aux_sym_preproc_undef_token1] = ACTIONS(5529),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5529),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5529),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5529),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5529),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5529),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5529),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5529),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5529),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5529),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_API_AVAILABLE] = ACTIONS(5529),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_API_DEPRECATED] = ACTIONS(5529),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5529),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5529),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5529),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5529),
    [anon_sym___deprecated_msg] = ACTIONS(5529),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5529),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5529),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5529),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5529),
    [anon_sym_ATend] = ACTIONS(5531),
    [anon_sym_AToptional] = ACTIONS(5531),
    [anon_sym_ATrequired] = ACTIONS(5531),
    [anon_sym_ATproperty] = ACTIONS(5531),
    [anon_sym_BOOL] = ACTIONS(5529),
    [anon_sym_IMP] = ACTIONS(5529),
    [anon_sym_SEL] = ACTIONS(5529),
    [anon_sym_Class] = ACTIONS(5529),
    [anon_sym_id] = ACTIONS(5529),
  },
  [STATE(2246)] = {
    [sym_identifier] = ACTIONS(5533),
    [aux_sym_preproc_def_token1] = ACTIONS(5533),
    [aux_sym_preproc_if_token1] = ACTIONS(5533),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5533),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5533),
    [sym_preproc_directive] = ACTIONS(5533),
    [anon_sym_DASH] = ACTIONS(5535),
    [anon_sym_PLUS] = ACTIONS(5535),
    [anon_sym_LT] = ACTIONS(5535),
    [anon_sym___extension__] = ACTIONS(5533),
    [anon_sym_typedef] = ACTIONS(5533),
    [anon_sym_extern] = ACTIONS(5533),
    [anon_sym___attribute__] = ACTIONS(5533),
    [anon_sym___attribute] = ACTIONS(5533),
    [anon_sym_noreturn] = ACTIONS(5533),
    [anon_sym_LBRACK] = ACTIONS(5535),
    [anon_sym___declspec] = ACTIONS(5533),
    [anon_sym___cdecl] = ACTIONS(5533),
    [anon_sym___clrcall] = ACTIONS(5533),
    [anon_sym___stdcall] = ACTIONS(5533),
    [anon_sym___fastcall] = ACTIONS(5533),
    [anon_sym___thiscall] = ACTIONS(5533),
    [anon_sym___vectorcall] = ACTIONS(5533),
    [anon_sym_LBRACE] = ACTIONS(5535),
    [anon_sym_signed] = ACTIONS(5533),
    [anon_sym_unsigned] = ACTIONS(5533),
    [anon_sym_long] = ACTIONS(5533),
    [anon_sym_short] = ACTIONS(5533),
    [anon_sym_static] = ACTIONS(5533),
    [anon_sym_auto] = ACTIONS(5533),
    [anon_sym_register] = ACTIONS(5533),
    [anon_sym_inline] = ACTIONS(5533),
    [anon_sym___inline] = ACTIONS(5533),
    [anon_sym___inline__] = ACTIONS(5533),
    [anon_sym___forceinline] = ACTIONS(5533),
    [anon_sym_thread_local] = ACTIONS(5533),
    [anon_sym___thread] = ACTIONS(5533),
    [anon_sym_CG_EXTERN] = ACTIONS(5533),
    [anon_sym_CG_INLINE] = ACTIONS(5533),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5533),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5533),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5533),
    [anon_sym_IBOutlet] = ACTIONS(5533),
    [anon_sym_IBInspectable] = ACTIONS(5533),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5533),
    [anon_sym_NS_INLINE] = ACTIONS(5533),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5533),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5533),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5533),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5533),
    [anon_sym_const] = ACTIONS(5533),
    [anon_sym_constexpr] = ACTIONS(5533),
    [anon_sym_volatile] = ACTIONS(5533),
    [anon_sym_restrict] = ACTIONS(5533),
    [anon_sym___restrict__] = ACTIONS(5533),
    [anon_sym__Atomic] = ACTIONS(5533),
    [anon_sym__Noreturn] = ACTIONS(5533),
    [anon_sym__Nonnull] = ACTIONS(5533),
    [anon_sym_nullable] = ACTIONS(5533),
    [anon_sym__Complex] = ACTIONS(5533),
    [anon_sym__Nullable] = ACTIONS(5533),
    [anon_sym__Nullable_result] = ACTIONS(5533),
    [anon_sym__Null_unspecified] = ACTIONS(5533),
    [anon_sym___autoreleasing] = ACTIONS(5533),
    [anon_sym___block] = ACTIONS(5533),
    [anon_sym___bridge] = ACTIONS(5533),
    [anon_sym___bridge_retained] = ACTIONS(5533),
    [anon_sym___bridge_transfer] = ACTIONS(5533),
    [anon_sym___complex] = ACTIONS(5533),
    [anon_sym___const] = ACTIONS(5533),
    [anon_sym___imag] = ACTIONS(5533),
    [anon_sym___kindof] = ACTIONS(5533),
    [anon_sym___nonnull] = ACTIONS(5533),
    [anon_sym___nullable] = ACTIONS(5533),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5533),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5533),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5533),
    [anon_sym___real] = ACTIONS(5533),
    [anon_sym___strong] = ACTIONS(5533),
    [anon_sym___unsafe_unretained] = ACTIONS(5533),
    [anon_sym___unused] = ACTIONS(5533),
    [anon_sym___weak] = ACTIONS(5533),
    [anon_sym_alignas] = ACTIONS(5533),
    [anon_sym__Alignas] = ACTIONS(5533),
    [sym_primitive_type] = ACTIONS(5533),
    [anon_sym_enum] = ACTIONS(5533),
    [anon_sym_struct] = ACTIONS(5533),
    [anon_sym_union] = ACTIONS(5533),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5533),
    [anon_sym___typeof] = ACTIONS(5533),
    [anon_sym_typeof] = ACTIONS(5533),
    [aux_sym_preproc_undef_token1] = ACTIONS(5533),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5533),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5533),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5533),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5533),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5533),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5533),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5533),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5533),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5533),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5533),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5533),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5533),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5533),
    [anon_sym_API_AVAILABLE] = ACTIONS(5533),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5533),
    [anon_sym_API_DEPRECATED] = ACTIONS(5533),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5533),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5533),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5533),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5533),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5533),
    [anon_sym___deprecated_msg] = ACTIONS(5533),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5533),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5533),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5533),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5533),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5533),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5533),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5533),
    [anon_sym_ATend] = ACTIONS(5535),
    [anon_sym_ATproperty] = ACTIONS(5535),
    [anon_sym_BOOL] = ACTIONS(5533),
    [anon_sym_IMP] = ACTIONS(5533),
    [anon_sym_SEL] = ACTIONS(5533),
    [anon_sym_Class] = ACTIONS(5533),
    [anon_sym_id] = ACTIONS(5533),
  },
  [STATE(2247)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_DASH] = ACTIONS(2512),
    [anon_sym_PLUS] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATproperty] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(2248)] = {
    [sym_identifier] = ACTIONS(5537),
    [aux_sym_preproc_def_token1] = ACTIONS(5537),
    [aux_sym_preproc_if_token1] = ACTIONS(5537),
    [aux_sym_preproc_if_token2] = ACTIONS(5537),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5537),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5537),
    [aux_sym_preproc_else_token1] = ACTIONS(5537),
    [aux_sym_preproc_elif_token1] = ACTIONS(5537),
    [sym_preproc_directive] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5539),
    [anon_sym_PLUS] = ACTIONS(5539),
    [anon_sym___extension__] = ACTIONS(5537),
    [anon_sym_typedef] = ACTIONS(5537),
    [anon_sym_extern] = ACTIONS(5537),
    [anon_sym___attribute__] = ACTIONS(5537),
    [anon_sym___attribute] = ACTIONS(5537),
    [anon_sym_noreturn] = ACTIONS(5537),
    [anon_sym_LBRACK] = ACTIONS(5539),
    [anon_sym___declspec] = ACTIONS(5537),
    [anon_sym___cdecl] = ACTIONS(5537),
    [anon_sym___clrcall] = ACTIONS(5537),
    [anon_sym___stdcall] = ACTIONS(5537),
    [anon_sym___fastcall] = ACTIONS(5537),
    [anon_sym___thiscall] = ACTIONS(5537),
    [anon_sym___vectorcall] = ACTIONS(5537),
    [anon_sym_signed] = ACTIONS(5537),
    [anon_sym_unsigned] = ACTIONS(5537),
    [anon_sym_long] = ACTIONS(5537),
    [anon_sym_short] = ACTIONS(5537),
    [anon_sym_static] = ACTIONS(5537),
    [anon_sym_auto] = ACTIONS(5537),
    [anon_sym_register] = ACTIONS(5537),
    [anon_sym_inline] = ACTIONS(5537),
    [anon_sym___inline] = ACTIONS(5537),
    [anon_sym___inline__] = ACTIONS(5537),
    [anon_sym___forceinline] = ACTIONS(5537),
    [anon_sym_thread_local] = ACTIONS(5537),
    [anon_sym___thread] = ACTIONS(5537),
    [anon_sym_CG_EXTERN] = ACTIONS(5537),
    [anon_sym_CG_INLINE] = ACTIONS(5537),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5537),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5537),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5537),
    [anon_sym_IBOutlet] = ACTIONS(5537),
    [anon_sym_IBInspectable] = ACTIONS(5537),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5537),
    [anon_sym_NS_INLINE] = ACTIONS(5537),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5537),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5537),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5537),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5537),
    [anon_sym_const] = ACTIONS(5537),
    [anon_sym_constexpr] = ACTIONS(5537),
    [anon_sym_volatile] = ACTIONS(5537),
    [anon_sym_restrict] = ACTIONS(5537),
    [anon_sym___restrict__] = ACTIONS(5537),
    [anon_sym__Atomic] = ACTIONS(5537),
    [anon_sym__Noreturn] = ACTIONS(5537),
    [anon_sym__Nonnull] = ACTIONS(5537),
    [anon_sym_nullable] = ACTIONS(5537),
    [anon_sym__Complex] = ACTIONS(5537),
    [anon_sym__Nullable] = ACTIONS(5537),
    [anon_sym__Nullable_result] = ACTIONS(5537),
    [anon_sym__Null_unspecified] = ACTIONS(5537),
    [anon_sym___autoreleasing] = ACTIONS(5537),
    [anon_sym___block] = ACTIONS(5537),
    [anon_sym___bridge] = ACTIONS(5537),
    [anon_sym___bridge_retained] = ACTIONS(5537),
    [anon_sym___bridge_transfer] = ACTIONS(5537),
    [anon_sym___complex] = ACTIONS(5537),
    [anon_sym___const] = ACTIONS(5537),
    [anon_sym___imag] = ACTIONS(5537),
    [anon_sym___kindof] = ACTIONS(5537),
    [anon_sym___nonnull] = ACTIONS(5537),
    [anon_sym___nullable] = ACTIONS(5537),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5537),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5537),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5537),
    [anon_sym___real] = ACTIONS(5537),
    [anon_sym___strong] = ACTIONS(5537),
    [anon_sym___unsafe_unretained] = ACTIONS(5537),
    [anon_sym___unused] = ACTIONS(5537),
    [anon_sym___weak] = ACTIONS(5537),
    [anon_sym_alignas] = ACTIONS(5537),
    [anon_sym__Alignas] = ACTIONS(5537),
    [sym_primitive_type] = ACTIONS(5537),
    [anon_sym_enum] = ACTIONS(5537),
    [anon_sym_struct] = ACTIONS(5537),
    [anon_sym_union] = ACTIONS(5537),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5537),
    [anon_sym___typeof] = ACTIONS(5537),
    [anon_sym_typeof] = ACTIONS(5537),
    [aux_sym_preproc_undef_token1] = ACTIONS(5537),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5537),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5537),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5537),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5537),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5537),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5537),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5537),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5537),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5537),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_API_AVAILABLE] = ACTIONS(5537),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_API_DEPRECATED] = ACTIONS(5537),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5537),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5537),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5537),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5537),
    [anon_sym___deprecated_msg] = ACTIONS(5537),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5537),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5537),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5537),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5537),
    [anon_sym_ATproperty] = ACTIONS(5539),
    [anon_sym_BOOL] = ACTIONS(5537),
    [anon_sym_IMP] = ACTIONS(5537),
    [anon_sym_SEL] = ACTIONS(5537),
    [anon_sym_Class] = ACTIONS(5537),
    [anon_sym_id] = ACTIONS(5537),
  },
  [STATE(2249)] = {
    [sym_identifier] = ACTIONS(5541),
    [aux_sym_preproc_def_token1] = ACTIONS(5541),
    [aux_sym_preproc_if_token1] = ACTIONS(5541),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5541),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5541),
    [sym_preproc_directive] = ACTIONS(5541),
    [anon_sym_DASH] = ACTIONS(5543),
    [anon_sym_PLUS] = ACTIONS(5543),
    [anon_sym___extension__] = ACTIONS(5541),
    [anon_sym_typedef] = ACTIONS(5541),
    [anon_sym_extern] = ACTIONS(5541),
    [anon_sym___attribute__] = ACTIONS(5541),
    [anon_sym___attribute] = ACTIONS(5541),
    [anon_sym_noreturn] = ACTIONS(5541),
    [anon_sym_LBRACK] = ACTIONS(5543),
    [anon_sym___declspec] = ACTIONS(5541),
    [anon_sym___cdecl] = ACTIONS(5541),
    [anon_sym___clrcall] = ACTIONS(5541),
    [anon_sym___stdcall] = ACTIONS(5541),
    [anon_sym___fastcall] = ACTIONS(5541),
    [anon_sym___thiscall] = ACTIONS(5541),
    [anon_sym___vectorcall] = ACTIONS(5541),
    [anon_sym_signed] = ACTIONS(5541),
    [anon_sym_unsigned] = ACTIONS(5541),
    [anon_sym_long] = ACTIONS(5541),
    [anon_sym_short] = ACTIONS(5541),
    [anon_sym_static] = ACTIONS(5541),
    [anon_sym_auto] = ACTIONS(5541),
    [anon_sym_register] = ACTIONS(5541),
    [anon_sym_inline] = ACTIONS(5541),
    [anon_sym___inline] = ACTIONS(5541),
    [anon_sym___inline__] = ACTIONS(5541),
    [anon_sym___forceinline] = ACTIONS(5541),
    [anon_sym_thread_local] = ACTIONS(5541),
    [anon_sym___thread] = ACTIONS(5541),
    [anon_sym_CG_EXTERN] = ACTIONS(5541),
    [anon_sym_CG_INLINE] = ACTIONS(5541),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5541),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5541),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5541),
    [anon_sym_IBOutlet] = ACTIONS(5541),
    [anon_sym_IBInspectable] = ACTIONS(5541),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5541),
    [anon_sym_NS_INLINE] = ACTIONS(5541),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5541),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5541),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5541),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5541),
    [anon_sym_const] = ACTIONS(5541),
    [anon_sym_constexpr] = ACTIONS(5541),
    [anon_sym_volatile] = ACTIONS(5541),
    [anon_sym_restrict] = ACTIONS(5541),
    [anon_sym___restrict__] = ACTIONS(5541),
    [anon_sym__Atomic] = ACTIONS(5541),
    [anon_sym__Noreturn] = ACTIONS(5541),
    [anon_sym__Nonnull] = ACTIONS(5541),
    [anon_sym_nullable] = ACTIONS(5541),
    [anon_sym__Complex] = ACTIONS(5541),
    [anon_sym__Nullable] = ACTIONS(5541),
    [anon_sym__Nullable_result] = ACTIONS(5541),
    [anon_sym__Null_unspecified] = ACTIONS(5541),
    [anon_sym___autoreleasing] = ACTIONS(5541),
    [anon_sym___block] = ACTIONS(5541),
    [anon_sym___bridge] = ACTIONS(5541),
    [anon_sym___bridge_retained] = ACTIONS(5541),
    [anon_sym___bridge_transfer] = ACTIONS(5541),
    [anon_sym___complex] = ACTIONS(5541),
    [anon_sym___const] = ACTIONS(5541),
    [anon_sym___imag] = ACTIONS(5541),
    [anon_sym___kindof] = ACTIONS(5541),
    [anon_sym___nonnull] = ACTIONS(5541),
    [anon_sym___nullable] = ACTIONS(5541),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5541),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5541),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5541),
    [anon_sym___real] = ACTIONS(5541),
    [anon_sym___strong] = ACTIONS(5541),
    [anon_sym___unsafe_unretained] = ACTIONS(5541),
    [anon_sym___unused] = ACTIONS(5541),
    [anon_sym___weak] = ACTIONS(5541),
    [anon_sym_alignas] = ACTIONS(5541),
    [anon_sym__Alignas] = ACTIONS(5541),
    [sym_primitive_type] = ACTIONS(5541),
    [anon_sym_enum] = ACTIONS(5541),
    [anon_sym_struct] = ACTIONS(5541),
    [anon_sym_union] = ACTIONS(5541),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5541),
    [anon_sym___typeof] = ACTIONS(5541),
    [anon_sym_typeof] = ACTIONS(5541),
    [aux_sym_preproc_undef_token1] = ACTIONS(5541),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5541),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5541),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5541),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5541),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5541),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5541),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5541),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5541),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5541),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_API_AVAILABLE] = ACTIONS(5541),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_API_DEPRECATED] = ACTIONS(5541),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5541),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5541),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5541),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5541),
    [anon_sym___deprecated_msg] = ACTIONS(5541),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5541),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5541),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5541),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5541),
    [anon_sym_ATend] = ACTIONS(5543),
    [anon_sym_AToptional] = ACTIONS(5543),
    [anon_sym_ATrequired] = ACTIONS(5543),
    [anon_sym_ATproperty] = ACTIONS(5543),
    [anon_sym_BOOL] = ACTIONS(5541),
    [anon_sym_IMP] = ACTIONS(5541),
    [anon_sym_SEL] = ACTIONS(5541),
    [anon_sym_Class] = ACTIONS(5541),
    [anon_sym_id] = ACTIONS(5541),
  },
  [STATE(2250)] = {
    [sym_identifier] = ACTIONS(5399),
    [aux_sym_preproc_def_token1] = ACTIONS(5399),
    [aux_sym_preproc_if_token1] = ACTIONS(5399),
    [aux_sym_preproc_if_token2] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5399),
    [sym_preproc_directive] = ACTIONS(5399),
    [anon_sym_DASH] = ACTIONS(5401),
    [anon_sym_PLUS] = ACTIONS(5401),
    [anon_sym_SEMI] = ACTIONS(5545),
    [anon_sym___extension__] = ACTIONS(5399),
    [anon_sym_typedef] = ACTIONS(5399),
    [anon_sym_extern] = ACTIONS(5399),
    [anon_sym___attribute__] = ACTIONS(5399),
    [anon_sym___attribute] = ACTIONS(5399),
    [anon_sym_noreturn] = ACTIONS(5399),
    [anon_sym_LBRACK] = ACTIONS(5401),
    [anon_sym___declspec] = ACTIONS(5399),
    [anon_sym___cdecl] = ACTIONS(5399),
    [anon_sym___clrcall] = ACTIONS(5399),
    [anon_sym___stdcall] = ACTIONS(5399),
    [anon_sym___fastcall] = ACTIONS(5399),
    [anon_sym___thiscall] = ACTIONS(5399),
    [anon_sym___vectorcall] = ACTIONS(5399),
    [anon_sym_signed] = ACTIONS(5399),
    [anon_sym_unsigned] = ACTIONS(5399),
    [anon_sym_long] = ACTIONS(5399),
    [anon_sym_short] = ACTIONS(5399),
    [anon_sym_static] = ACTIONS(5399),
    [anon_sym_auto] = ACTIONS(5399),
    [anon_sym_register] = ACTIONS(5399),
    [anon_sym_inline] = ACTIONS(5399),
    [anon_sym___inline] = ACTIONS(5399),
    [anon_sym___inline__] = ACTIONS(5399),
    [anon_sym___forceinline] = ACTIONS(5399),
    [anon_sym_thread_local] = ACTIONS(5399),
    [anon_sym___thread] = ACTIONS(5399),
    [anon_sym_CG_EXTERN] = ACTIONS(5399),
    [anon_sym_CG_INLINE] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5399),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5399),
    [anon_sym_IBOutlet] = ACTIONS(5399),
    [anon_sym_IBInspectable] = ACTIONS(5399),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5399),
    [anon_sym_NS_INLINE] = ACTIONS(5399),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5399),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5399),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5399),
    [anon_sym_const] = ACTIONS(5399),
    [anon_sym_constexpr] = ACTIONS(5399),
    [anon_sym_volatile] = ACTIONS(5399),
    [anon_sym_restrict] = ACTIONS(5399),
    [anon_sym___restrict__] = ACTIONS(5399),
    [anon_sym__Atomic] = ACTIONS(5399),
    [anon_sym__Noreturn] = ACTIONS(5399),
    [anon_sym__Nonnull] = ACTIONS(5399),
    [anon_sym_nullable] = ACTIONS(5399),
    [anon_sym__Complex] = ACTIONS(5399),
    [anon_sym__Nullable] = ACTIONS(5399),
    [anon_sym__Nullable_result] = ACTIONS(5399),
    [anon_sym__Null_unspecified] = ACTIONS(5399),
    [anon_sym___autoreleasing] = ACTIONS(5399),
    [anon_sym___block] = ACTIONS(5399),
    [anon_sym___bridge] = ACTIONS(5399),
    [anon_sym___bridge_retained] = ACTIONS(5399),
    [anon_sym___bridge_transfer] = ACTIONS(5399),
    [anon_sym___complex] = ACTIONS(5399),
    [anon_sym___const] = ACTIONS(5399),
    [anon_sym___imag] = ACTIONS(5399),
    [anon_sym___kindof] = ACTIONS(5399),
    [anon_sym___nonnull] = ACTIONS(5399),
    [anon_sym___nullable] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5399),
    [anon_sym___real] = ACTIONS(5399),
    [anon_sym___strong] = ACTIONS(5399),
    [anon_sym___unsafe_unretained] = ACTIONS(5399),
    [anon_sym___unused] = ACTIONS(5399),
    [anon_sym___weak] = ACTIONS(5399),
    [anon_sym_alignas] = ACTIONS(5399),
    [anon_sym__Alignas] = ACTIONS(5399),
    [sym_primitive_type] = ACTIONS(5399),
    [anon_sym_enum] = ACTIONS(5399),
    [anon_sym_struct] = ACTIONS(5399),
    [anon_sym_union] = ACTIONS(5399),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5399),
    [anon_sym___typeof] = ACTIONS(5399),
    [anon_sym_typeof] = ACTIONS(5399),
    [aux_sym_preproc_undef_token1] = ACTIONS(5399),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5399),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5399),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5399),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_API_AVAILABLE] = ACTIONS(5399),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_API_DEPRECATED] = ACTIONS(5399),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym___deprecated_msg] = ACTIONS(5399),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5399),
    [anon_sym_ATsynthesize] = ACTIONS(5401),
    [anon_sym_ATdynamic] = ACTIONS(5401),
    [anon_sym_BOOL] = ACTIONS(5399),
    [anon_sym_IMP] = ACTIONS(5399),
    [anon_sym_SEL] = ACTIONS(5399),
    [anon_sym_Class] = ACTIONS(5399),
    [anon_sym_id] = ACTIONS(5399),
  },
  [STATE(2251)] = {
    [sym_identifier] = ACTIONS(5547),
    [aux_sym_preproc_def_token1] = ACTIONS(5547),
    [aux_sym_preproc_if_token1] = ACTIONS(5547),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5547),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5547),
    [sym_preproc_directive] = ACTIONS(5547),
    [anon_sym_DASH] = ACTIONS(3755),
    [anon_sym_PLUS] = ACTIONS(3755),
    [anon_sym___extension__] = ACTIONS(5547),
    [anon_sym_typedef] = ACTIONS(5547),
    [anon_sym_extern] = ACTIONS(5547),
    [anon_sym___attribute__] = ACTIONS(5547),
    [anon_sym___attribute] = ACTIONS(5547),
    [anon_sym_noreturn] = ACTIONS(5547),
    [anon_sym_LBRACK] = ACTIONS(3755),
    [anon_sym___declspec] = ACTIONS(5547),
    [anon_sym___cdecl] = ACTIONS(5547),
    [anon_sym___clrcall] = ACTIONS(5547),
    [anon_sym___stdcall] = ACTIONS(5547),
    [anon_sym___fastcall] = ACTIONS(5547),
    [anon_sym___thiscall] = ACTIONS(5547),
    [anon_sym___vectorcall] = ACTIONS(5547),
    [anon_sym_signed] = ACTIONS(5547),
    [anon_sym_unsigned] = ACTIONS(5547),
    [anon_sym_long] = ACTIONS(5547),
    [anon_sym_short] = ACTIONS(5547),
    [anon_sym_static] = ACTIONS(5547),
    [anon_sym_auto] = ACTIONS(5547),
    [anon_sym_register] = ACTIONS(5547),
    [anon_sym_inline] = ACTIONS(5547),
    [anon_sym___inline] = ACTIONS(5547),
    [anon_sym___inline__] = ACTIONS(5547),
    [anon_sym___forceinline] = ACTIONS(5547),
    [anon_sym_thread_local] = ACTIONS(5547),
    [anon_sym___thread] = ACTIONS(5547),
    [anon_sym_CG_EXTERN] = ACTIONS(5547),
    [anon_sym_CG_INLINE] = ACTIONS(5547),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5547),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5547),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5547),
    [anon_sym_IBOutlet] = ACTIONS(5547),
    [anon_sym_IBInspectable] = ACTIONS(5547),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5547),
    [anon_sym_NS_INLINE] = ACTIONS(5547),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5547),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5547),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5547),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5547),
    [anon_sym_const] = ACTIONS(5547),
    [anon_sym_constexpr] = ACTIONS(5547),
    [anon_sym_volatile] = ACTIONS(5547),
    [anon_sym_restrict] = ACTIONS(5547),
    [anon_sym___restrict__] = ACTIONS(5547),
    [anon_sym__Atomic] = ACTIONS(5547),
    [anon_sym__Noreturn] = ACTIONS(5547),
    [anon_sym__Nonnull] = ACTIONS(5547),
    [anon_sym_nullable] = ACTIONS(5547),
    [anon_sym__Complex] = ACTIONS(5547),
    [anon_sym__Nullable] = ACTIONS(5547),
    [anon_sym__Nullable_result] = ACTIONS(5547),
    [anon_sym__Null_unspecified] = ACTIONS(5547),
    [anon_sym___autoreleasing] = ACTIONS(5547),
    [anon_sym___block] = ACTIONS(5547),
    [anon_sym___bridge] = ACTIONS(5547),
    [anon_sym___bridge_retained] = ACTIONS(5547),
    [anon_sym___bridge_transfer] = ACTIONS(5547),
    [anon_sym___complex] = ACTIONS(5547),
    [anon_sym___const] = ACTIONS(5547),
    [anon_sym___imag] = ACTIONS(5547),
    [anon_sym___kindof] = ACTIONS(5547),
    [anon_sym___nonnull] = ACTIONS(5547),
    [anon_sym___nullable] = ACTIONS(5547),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5547),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5547),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5547),
    [anon_sym___real] = ACTIONS(5547),
    [anon_sym___strong] = ACTIONS(5547),
    [anon_sym___unsafe_unretained] = ACTIONS(5547),
    [anon_sym___unused] = ACTIONS(5547),
    [anon_sym___weak] = ACTIONS(5547),
    [anon_sym_alignas] = ACTIONS(5547),
    [anon_sym__Alignas] = ACTIONS(5547),
    [sym_primitive_type] = ACTIONS(5547),
    [anon_sym_enum] = ACTIONS(5547),
    [anon_sym_struct] = ACTIONS(5547),
    [anon_sym_union] = ACTIONS(5547),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5547),
    [anon_sym___typeof] = ACTIONS(5547),
    [anon_sym_typeof] = ACTIONS(5547),
    [aux_sym_preproc_undef_token1] = ACTIONS(5547),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5547),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5547),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5547),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5547),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5547),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5547),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5547),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5547),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5547),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5547),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5547),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5547),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5547),
    [anon_sym_API_AVAILABLE] = ACTIONS(5547),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5547),
    [anon_sym_API_DEPRECATED] = ACTIONS(5547),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5547),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5547),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5547),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5547),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5547),
    [anon_sym___deprecated_msg] = ACTIONS(5547),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5547),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5547),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5547),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5547),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5547),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5547),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5547),
    [anon_sym_ATend] = ACTIONS(3755),
    [anon_sym_AToptional] = ACTIONS(3755),
    [anon_sym_ATrequired] = ACTIONS(3755),
    [anon_sym_ATproperty] = ACTIONS(3755),
    [anon_sym_BOOL] = ACTIONS(5547),
    [anon_sym_IMP] = ACTIONS(5547),
    [anon_sym_SEL] = ACTIONS(5547),
    [anon_sym_Class] = ACTIONS(5547),
    [anon_sym_id] = ACTIONS(5547),
  },
  [STATE(2252)] = {
    [sym_identifier] = ACTIONS(5438),
    [aux_sym_preproc_def_token1] = ACTIONS(5438),
    [aux_sym_preproc_if_token1] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5438),
    [sym_preproc_directive] = ACTIONS(5438),
    [anon_sym_DASH] = ACTIONS(5440),
    [anon_sym_PLUS] = ACTIONS(5440),
    [anon_sym_SEMI] = ACTIONS(5549),
    [anon_sym___extension__] = ACTIONS(5438),
    [anon_sym_typedef] = ACTIONS(5438),
    [anon_sym_extern] = ACTIONS(5438),
    [anon_sym___attribute__] = ACTIONS(5438),
    [anon_sym___attribute] = ACTIONS(5438),
    [anon_sym_noreturn] = ACTIONS(5438),
    [anon_sym_LBRACK] = ACTIONS(5440),
    [anon_sym___declspec] = ACTIONS(5438),
    [anon_sym___cdecl] = ACTIONS(5438),
    [anon_sym___clrcall] = ACTIONS(5438),
    [anon_sym___stdcall] = ACTIONS(5438),
    [anon_sym___fastcall] = ACTIONS(5438),
    [anon_sym___thiscall] = ACTIONS(5438),
    [anon_sym___vectorcall] = ACTIONS(5438),
    [anon_sym_signed] = ACTIONS(5438),
    [anon_sym_unsigned] = ACTIONS(5438),
    [anon_sym_long] = ACTIONS(5438),
    [anon_sym_short] = ACTIONS(5438),
    [anon_sym_static] = ACTIONS(5438),
    [anon_sym_auto] = ACTIONS(5438),
    [anon_sym_register] = ACTIONS(5438),
    [anon_sym_inline] = ACTIONS(5438),
    [anon_sym___inline] = ACTIONS(5438),
    [anon_sym___inline__] = ACTIONS(5438),
    [anon_sym___forceinline] = ACTIONS(5438),
    [anon_sym_thread_local] = ACTIONS(5438),
    [anon_sym___thread] = ACTIONS(5438),
    [anon_sym_CG_EXTERN] = ACTIONS(5438),
    [anon_sym_CG_INLINE] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5438),
    [anon_sym_IBOutlet] = ACTIONS(5438),
    [anon_sym_IBInspectable] = ACTIONS(5438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5438),
    [anon_sym_NS_INLINE] = ACTIONS(5438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5438),
    [anon_sym_const] = ACTIONS(5438),
    [anon_sym_constexpr] = ACTIONS(5438),
    [anon_sym_volatile] = ACTIONS(5438),
    [anon_sym_restrict] = ACTIONS(5438),
    [anon_sym___restrict__] = ACTIONS(5438),
    [anon_sym__Atomic] = ACTIONS(5438),
    [anon_sym__Noreturn] = ACTIONS(5438),
    [anon_sym__Nonnull] = ACTIONS(5438),
    [anon_sym_nullable] = ACTIONS(5438),
    [anon_sym__Complex] = ACTIONS(5438),
    [anon_sym__Nullable] = ACTIONS(5438),
    [anon_sym__Nullable_result] = ACTIONS(5438),
    [anon_sym__Null_unspecified] = ACTIONS(5438),
    [anon_sym___autoreleasing] = ACTIONS(5438),
    [anon_sym___block] = ACTIONS(5438),
    [anon_sym___bridge] = ACTIONS(5438),
    [anon_sym___bridge_retained] = ACTIONS(5438),
    [anon_sym___bridge_transfer] = ACTIONS(5438),
    [anon_sym___complex] = ACTIONS(5438),
    [anon_sym___const] = ACTIONS(5438),
    [anon_sym___imag] = ACTIONS(5438),
    [anon_sym___kindof] = ACTIONS(5438),
    [anon_sym___nonnull] = ACTIONS(5438),
    [anon_sym___nullable] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5438),
    [anon_sym___real] = ACTIONS(5438),
    [anon_sym___strong] = ACTIONS(5438),
    [anon_sym___unsafe_unretained] = ACTIONS(5438),
    [anon_sym___unused] = ACTIONS(5438),
    [anon_sym___weak] = ACTIONS(5438),
    [anon_sym_alignas] = ACTIONS(5438),
    [anon_sym__Alignas] = ACTIONS(5438),
    [sym_primitive_type] = ACTIONS(5438),
    [anon_sym_enum] = ACTIONS(5438),
    [anon_sym_struct] = ACTIONS(5438),
    [anon_sym_union] = ACTIONS(5438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5438),
    [anon_sym___typeof] = ACTIONS(5438),
    [anon_sym_typeof] = ACTIONS(5438),
    [aux_sym_preproc_undef_token1] = ACTIONS(5438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_API_AVAILABLE] = ACTIONS(5438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_API_DEPRECATED] = ACTIONS(5438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym___deprecated_msg] = ACTIONS(5438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5438),
    [anon_sym_ATend] = ACTIONS(5440),
    [anon_sym_ATsynthesize] = ACTIONS(5440),
    [anon_sym_ATdynamic] = ACTIONS(5440),
    [anon_sym_BOOL] = ACTIONS(5438),
    [anon_sym_IMP] = ACTIONS(5438),
    [anon_sym_SEL] = ACTIONS(5438),
    [anon_sym_Class] = ACTIONS(5438),
    [anon_sym_id] = ACTIONS(5438),
  },
  [STATE(2253)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_DASH] = ACTIONS(2512),
    [anon_sym_PLUS] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATproperty] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(2254)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token2] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_DASH] = ACTIONS(2556),
    [anon_sym_PLUS] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATproperty] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(2255)] = {
    [sym_identifier] = ACTIONS(2554),
    [aux_sym_preproc_def_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token1] = ACTIONS(2554),
    [aux_sym_preproc_if_token2] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2554),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2554),
    [sym_preproc_directive] = ACTIONS(2554),
    [anon_sym_DASH] = ACTIONS(2556),
    [anon_sym_PLUS] = ACTIONS(2556),
    [anon_sym___extension__] = ACTIONS(2554),
    [anon_sym_typedef] = ACTIONS(2554),
    [anon_sym_extern] = ACTIONS(2554),
    [anon_sym___attribute__] = ACTIONS(2554),
    [anon_sym___attribute] = ACTIONS(2554),
    [anon_sym_noreturn] = ACTIONS(2554),
    [anon_sym_LBRACK] = ACTIONS(2556),
    [anon_sym___declspec] = ACTIONS(2554),
    [anon_sym___cdecl] = ACTIONS(2554),
    [anon_sym___clrcall] = ACTIONS(2554),
    [anon_sym___stdcall] = ACTIONS(2554),
    [anon_sym___fastcall] = ACTIONS(2554),
    [anon_sym___thiscall] = ACTIONS(2554),
    [anon_sym___vectorcall] = ACTIONS(2554),
    [anon_sym_signed] = ACTIONS(2554),
    [anon_sym_unsigned] = ACTIONS(2554),
    [anon_sym_long] = ACTIONS(2554),
    [anon_sym_short] = ACTIONS(2554),
    [anon_sym_static] = ACTIONS(2554),
    [anon_sym_auto] = ACTIONS(2554),
    [anon_sym_register] = ACTIONS(2554),
    [anon_sym_inline] = ACTIONS(2554),
    [anon_sym___inline] = ACTIONS(2554),
    [anon_sym___inline__] = ACTIONS(2554),
    [anon_sym___forceinline] = ACTIONS(2554),
    [anon_sym_thread_local] = ACTIONS(2554),
    [anon_sym___thread] = ACTIONS(2554),
    [anon_sym_CG_EXTERN] = ACTIONS(2554),
    [anon_sym_CG_INLINE] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2554),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2554),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2554),
    [anon_sym_IBOutlet] = ACTIONS(2554),
    [anon_sym_IBInspectable] = ACTIONS(2554),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2554),
    [anon_sym_NS_INLINE] = ACTIONS(2554),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2554),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2554),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2554),
    [anon_sym_const] = ACTIONS(2554),
    [anon_sym_constexpr] = ACTIONS(2554),
    [anon_sym_volatile] = ACTIONS(2554),
    [anon_sym_restrict] = ACTIONS(2554),
    [anon_sym___restrict__] = ACTIONS(2554),
    [anon_sym__Atomic] = ACTIONS(2554),
    [anon_sym__Noreturn] = ACTIONS(2554),
    [anon_sym__Nonnull] = ACTIONS(2554),
    [anon_sym_nullable] = ACTIONS(2554),
    [anon_sym__Complex] = ACTIONS(2554),
    [anon_sym__Nullable] = ACTIONS(2554),
    [anon_sym__Nullable_result] = ACTIONS(2554),
    [anon_sym__Null_unspecified] = ACTIONS(2554),
    [anon_sym___autoreleasing] = ACTIONS(2554),
    [anon_sym___block] = ACTIONS(2554),
    [anon_sym___bridge] = ACTIONS(2554),
    [anon_sym___bridge_retained] = ACTIONS(2554),
    [anon_sym___bridge_transfer] = ACTIONS(2554),
    [anon_sym___complex] = ACTIONS(2554),
    [anon_sym___const] = ACTIONS(2554),
    [anon_sym___imag] = ACTIONS(2554),
    [anon_sym___kindof] = ACTIONS(2554),
    [anon_sym___nonnull] = ACTIONS(2554),
    [anon_sym___nullable] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2554),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2554),
    [anon_sym___real] = ACTIONS(2554),
    [anon_sym___strong] = ACTIONS(2554),
    [anon_sym___unsafe_unretained] = ACTIONS(2554),
    [anon_sym___unused] = ACTIONS(2554),
    [anon_sym___weak] = ACTIONS(2554),
    [anon_sym_alignas] = ACTIONS(2554),
    [anon_sym__Alignas] = ACTIONS(2554),
    [sym_primitive_type] = ACTIONS(2554),
    [anon_sym_enum] = ACTIONS(2554),
    [anon_sym_struct] = ACTIONS(2554),
    [anon_sym_union] = ACTIONS(2554),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2554),
    [anon_sym___typeof] = ACTIONS(2554),
    [anon_sym_typeof] = ACTIONS(2554),
    [aux_sym_preproc_undef_token1] = ACTIONS(2554),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2554),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2554),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2554),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2554),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2554),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_API_AVAILABLE] = ACTIONS(2554),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_API_DEPRECATED] = ACTIONS(2554),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2554),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2554),
    [anon_sym___deprecated_msg] = ACTIONS(2554),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2554),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2554),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2554),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2554),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2554),
    [anon_sym_ATsynthesize] = ACTIONS(2556),
    [anon_sym_ATdynamic] = ACTIONS(2556),
    [anon_sym_ATproperty] = ACTIONS(2556),
    [anon_sym_BOOL] = ACTIONS(2554),
    [anon_sym_IMP] = ACTIONS(2554),
    [anon_sym_SEL] = ACTIONS(2554),
    [anon_sym_Class] = ACTIONS(2554),
    [anon_sym_id] = ACTIONS(2554),
  },
  [STATE(2256)] = {
    [sym_identifier] = ACTIONS(5405),
    [aux_sym_preproc_def_token1] = ACTIONS(5405),
    [aux_sym_preproc_if_token1] = ACTIONS(5405),
    [aux_sym_preproc_if_token2] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5405),
    [sym_preproc_directive] = ACTIONS(5405),
    [anon_sym_DASH] = ACTIONS(5407),
    [anon_sym_PLUS] = ACTIONS(5407),
    [anon_sym_SEMI] = ACTIONS(5551),
    [anon_sym___extension__] = ACTIONS(5405),
    [anon_sym_typedef] = ACTIONS(5405),
    [anon_sym_extern] = ACTIONS(5405),
    [anon_sym___attribute__] = ACTIONS(5405),
    [anon_sym___attribute] = ACTIONS(5405),
    [anon_sym_noreturn] = ACTIONS(5405),
    [anon_sym_LBRACK] = ACTIONS(5407),
    [anon_sym___declspec] = ACTIONS(5405),
    [anon_sym___cdecl] = ACTIONS(5405),
    [anon_sym___clrcall] = ACTIONS(5405),
    [anon_sym___stdcall] = ACTIONS(5405),
    [anon_sym___fastcall] = ACTIONS(5405),
    [anon_sym___thiscall] = ACTIONS(5405),
    [anon_sym___vectorcall] = ACTIONS(5405),
    [anon_sym_signed] = ACTIONS(5405),
    [anon_sym_unsigned] = ACTIONS(5405),
    [anon_sym_long] = ACTIONS(5405),
    [anon_sym_short] = ACTIONS(5405),
    [anon_sym_static] = ACTIONS(5405),
    [anon_sym_auto] = ACTIONS(5405),
    [anon_sym_register] = ACTIONS(5405),
    [anon_sym_inline] = ACTIONS(5405),
    [anon_sym___inline] = ACTIONS(5405),
    [anon_sym___inline__] = ACTIONS(5405),
    [anon_sym___forceinline] = ACTIONS(5405),
    [anon_sym_thread_local] = ACTIONS(5405),
    [anon_sym___thread] = ACTIONS(5405),
    [anon_sym_CG_EXTERN] = ACTIONS(5405),
    [anon_sym_CG_INLINE] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5405),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5405),
    [anon_sym_IBOutlet] = ACTIONS(5405),
    [anon_sym_IBInspectable] = ACTIONS(5405),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5405),
    [anon_sym_NS_INLINE] = ACTIONS(5405),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5405),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5405),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5405),
    [anon_sym_const] = ACTIONS(5405),
    [anon_sym_constexpr] = ACTIONS(5405),
    [anon_sym_volatile] = ACTIONS(5405),
    [anon_sym_restrict] = ACTIONS(5405),
    [anon_sym___restrict__] = ACTIONS(5405),
    [anon_sym__Atomic] = ACTIONS(5405),
    [anon_sym__Noreturn] = ACTIONS(5405),
    [anon_sym__Nonnull] = ACTIONS(5405),
    [anon_sym_nullable] = ACTIONS(5405),
    [anon_sym__Complex] = ACTIONS(5405),
    [anon_sym__Nullable] = ACTIONS(5405),
    [anon_sym__Nullable_result] = ACTIONS(5405),
    [anon_sym__Null_unspecified] = ACTIONS(5405),
    [anon_sym___autoreleasing] = ACTIONS(5405),
    [anon_sym___block] = ACTIONS(5405),
    [anon_sym___bridge] = ACTIONS(5405),
    [anon_sym___bridge_retained] = ACTIONS(5405),
    [anon_sym___bridge_transfer] = ACTIONS(5405),
    [anon_sym___complex] = ACTIONS(5405),
    [anon_sym___const] = ACTIONS(5405),
    [anon_sym___imag] = ACTIONS(5405),
    [anon_sym___kindof] = ACTIONS(5405),
    [anon_sym___nonnull] = ACTIONS(5405),
    [anon_sym___nullable] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5405),
    [anon_sym___real] = ACTIONS(5405),
    [anon_sym___strong] = ACTIONS(5405),
    [anon_sym___unsafe_unretained] = ACTIONS(5405),
    [anon_sym___unused] = ACTIONS(5405),
    [anon_sym___weak] = ACTIONS(5405),
    [anon_sym_alignas] = ACTIONS(5405),
    [anon_sym__Alignas] = ACTIONS(5405),
    [sym_primitive_type] = ACTIONS(5405),
    [anon_sym_enum] = ACTIONS(5405),
    [anon_sym_struct] = ACTIONS(5405),
    [anon_sym_union] = ACTIONS(5405),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5405),
    [anon_sym___typeof] = ACTIONS(5405),
    [anon_sym_typeof] = ACTIONS(5405),
    [aux_sym_preproc_undef_token1] = ACTIONS(5405),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5405),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5405),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5405),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_API_AVAILABLE] = ACTIONS(5405),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_API_DEPRECATED] = ACTIONS(5405),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym___deprecated_msg] = ACTIONS(5405),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5405),
    [anon_sym_ATsynthesize] = ACTIONS(5407),
    [anon_sym_ATdynamic] = ACTIONS(5407),
    [anon_sym_BOOL] = ACTIONS(5405),
    [anon_sym_IMP] = ACTIONS(5405),
    [anon_sym_SEL] = ACTIONS(5405),
    [anon_sym_Class] = ACTIONS(5405),
    [anon_sym_id] = ACTIONS(5405),
  },
  [STATE(2257)] = {
    [sym_identifier] = ACTIONS(5466),
    [aux_sym_preproc_def_token1] = ACTIONS(5466),
    [aux_sym_preproc_if_token1] = ACTIONS(5466),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5466),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5466),
    [sym_preproc_directive] = ACTIONS(5466),
    [anon_sym_DASH] = ACTIONS(5468),
    [anon_sym_PLUS] = ACTIONS(5468),
    [anon_sym_SEMI] = ACTIONS(5553),
    [anon_sym___extension__] = ACTIONS(5466),
    [anon_sym_typedef] = ACTIONS(5466),
    [anon_sym_extern] = ACTIONS(5466),
    [anon_sym___attribute__] = ACTIONS(5466),
    [anon_sym___attribute] = ACTIONS(5466),
    [anon_sym_noreturn] = ACTIONS(5466),
    [anon_sym_LBRACK] = ACTIONS(5468),
    [anon_sym___declspec] = ACTIONS(5466),
    [anon_sym___cdecl] = ACTIONS(5466),
    [anon_sym___clrcall] = ACTIONS(5466),
    [anon_sym___stdcall] = ACTIONS(5466),
    [anon_sym___fastcall] = ACTIONS(5466),
    [anon_sym___thiscall] = ACTIONS(5466),
    [anon_sym___vectorcall] = ACTIONS(5466),
    [anon_sym_signed] = ACTIONS(5466),
    [anon_sym_unsigned] = ACTIONS(5466),
    [anon_sym_long] = ACTIONS(5466),
    [anon_sym_short] = ACTIONS(5466),
    [anon_sym_static] = ACTIONS(5466),
    [anon_sym_auto] = ACTIONS(5466),
    [anon_sym_register] = ACTIONS(5466),
    [anon_sym_inline] = ACTIONS(5466),
    [anon_sym___inline] = ACTIONS(5466),
    [anon_sym___inline__] = ACTIONS(5466),
    [anon_sym___forceinline] = ACTIONS(5466),
    [anon_sym_thread_local] = ACTIONS(5466),
    [anon_sym___thread] = ACTIONS(5466),
    [anon_sym_CG_EXTERN] = ACTIONS(5466),
    [anon_sym_CG_INLINE] = ACTIONS(5466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5466),
    [anon_sym_IBOutlet] = ACTIONS(5466),
    [anon_sym_IBInspectable] = ACTIONS(5466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5466),
    [anon_sym_NS_INLINE] = ACTIONS(5466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5466),
    [anon_sym_const] = ACTIONS(5466),
    [anon_sym_constexpr] = ACTIONS(5466),
    [anon_sym_volatile] = ACTIONS(5466),
    [anon_sym_restrict] = ACTIONS(5466),
    [anon_sym___restrict__] = ACTIONS(5466),
    [anon_sym__Atomic] = ACTIONS(5466),
    [anon_sym__Noreturn] = ACTIONS(5466),
    [anon_sym__Nonnull] = ACTIONS(5466),
    [anon_sym_nullable] = ACTIONS(5466),
    [anon_sym__Complex] = ACTIONS(5466),
    [anon_sym__Nullable] = ACTIONS(5466),
    [anon_sym__Nullable_result] = ACTIONS(5466),
    [anon_sym__Null_unspecified] = ACTIONS(5466),
    [anon_sym___autoreleasing] = ACTIONS(5466),
    [anon_sym___block] = ACTIONS(5466),
    [anon_sym___bridge] = ACTIONS(5466),
    [anon_sym___bridge_retained] = ACTIONS(5466),
    [anon_sym___bridge_transfer] = ACTIONS(5466),
    [anon_sym___complex] = ACTIONS(5466),
    [anon_sym___const] = ACTIONS(5466),
    [anon_sym___imag] = ACTIONS(5466),
    [anon_sym___kindof] = ACTIONS(5466),
    [anon_sym___nonnull] = ACTIONS(5466),
    [anon_sym___nullable] = ACTIONS(5466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5466),
    [anon_sym___real] = ACTIONS(5466),
    [anon_sym___strong] = ACTIONS(5466),
    [anon_sym___unsafe_unretained] = ACTIONS(5466),
    [anon_sym___unused] = ACTIONS(5466),
    [anon_sym___weak] = ACTIONS(5466),
    [anon_sym_alignas] = ACTIONS(5466),
    [anon_sym__Alignas] = ACTIONS(5466),
    [sym_primitive_type] = ACTIONS(5466),
    [anon_sym_enum] = ACTIONS(5466),
    [anon_sym_struct] = ACTIONS(5466),
    [anon_sym_union] = ACTIONS(5466),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5466),
    [anon_sym___typeof] = ACTIONS(5466),
    [anon_sym_typeof] = ACTIONS(5466),
    [aux_sym_preproc_undef_token1] = ACTIONS(5466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_API_AVAILABLE] = ACTIONS(5466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_API_DEPRECATED] = ACTIONS(5466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5466),
    [anon_sym___deprecated_msg] = ACTIONS(5466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5466),
    [anon_sym_ATend] = ACTIONS(5468),
    [anon_sym_ATsynthesize] = ACTIONS(5468),
    [anon_sym_ATdynamic] = ACTIONS(5468),
    [anon_sym_BOOL] = ACTIONS(5466),
    [anon_sym_IMP] = ACTIONS(5466),
    [anon_sym_SEL] = ACTIONS(5466),
    [anon_sym_Class] = ACTIONS(5466),
    [anon_sym_id] = ACTIONS(5466),
  },
  [STATE(2258)] = {
    [sym_identifier] = ACTIONS(2688),
    [aux_sym_preproc_def_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token1] = ACTIONS(2688),
    [aux_sym_preproc_if_token2] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2688),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2688),
    [sym_preproc_directive] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2690),
    [anon_sym_PLUS] = ACTIONS(2690),
    [anon_sym___extension__] = ACTIONS(2688),
    [anon_sym_typedef] = ACTIONS(2688),
    [anon_sym_extern] = ACTIONS(2688),
    [anon_sym___attribute__] = ACTIONS(2688),
    [anon_sym___attribute] = ACTIONS(2688),
    [anon_sym_noreturn] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym___declspec] = ACTIONS(2688),
    [anon_sym___cdecl] = ACTIONS(2688),
    [anon_sym___clrcall] = ACTIONS(2688),
    [anon_sym___stdcall] = ACTIONS(2688),
    [anon_sym___fastcall] = ACTIONS(2688),
    [anon_sym___thiscall] = ACTIONS(2688),
    [anon_sym___vectorcall] = ACTIONS(2688),
    [anon_sym_signed] = ACTIONS(2688),
    [anon_sym_unsigned] = ACTIONS(2688),
    [anon_sym_long] = ACTIONS(2688),
    [anon_sym_short] = ACTIONS(2688),
    [anon_sym_static] = ACTIONS(2688),
    [anon_sym_auto] = ACTIONS(2688),
    [anon_sym_register] = ACTIONS(2688),
    [anon_sym_inline] = ACTIONS(2688),
    [anon_sym___inline] = ACTIONS(2688),
    [anon_sym___inline__] = ACTIONS(2688),
    [anon_sym___forceinline] = ACTIONS(2688),
    [anon_sym_thread_local] = ACTIONS(2688),
    [anon_sym___thread] = ACTIONS(2688),
    [anon_sym_CG_EXTERN] = ACTIONS(2688),
    [anon_sym_CG_INLINE] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2688),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2688),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2688),
    [anon_sym_IBOutlet] = ACTIONS(2688),
    [anon_sym_IBInspectable] = ACTIONS(2688),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2688),
    [anon_sym_NS_INLINE] = ACTIONS(2688),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2688),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2688),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2688),
    [anon_sym_const] = ACTIONS(2688),
    [anon_sym_constexpr] = ACTIONS(2688),
    [anon_sym_volatile] = ACTIONS(2688),
    [anon_sym_restrict] = ACTIONS(2688),
    [anon_sym___restrict__] = ACTIONS(2688),
    [anon_sym__Atomic] = ACTIONS(2688),
    [anon_sym__Noreturn] = ACTIONS(2688),
    [anon_sym__Nonnull] = ACTIONS(2688),
    [anon_sym_nullable] = ACTIONS(2688),
    [anon_sym__Complex] = ACTIONS(2688),
    [anon_sym__Nullable] = ACTIONS(2688),
    [anon_sym__Nullable_result] = ACTIONS(2688),
    [anon_sym__Null_unspecified] = ACTIONS(2688),
    [anon_sym___autoreleasing] = ACTIONS(2688),
    [anon_sym___block] = ACTIONS(2688),
    [anon_sym___bridge] = ACTIONS(2688),
    [anon_sym___bridge_retained] = ACTIONS(2688),
    [anon_sym___bridge_transfer] = ACTIONS(2688),
    [anon_sym___complex] = ACTIONS(2688),
    [anon_sym___const] = ACTIONS(2688),
    [anon_sym___imag] = ACTIONS(2688),
    [anon_sym___kindof] = ACTIONS(2688),
    [anon_sym___nonnull] = ACTIONS(2688),
    [anon_sym___nullable] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2688),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2688),
    [anon_sym___real] = ACTIONS(2688),
    [anon_sym___strong] = ACTIONS(2688),
    [anon_sym___unsafe_unretained] = ACTIONS(2688),
    [anon_sym___unused] = ACTIONS(2688),
    [anon_sym___weak] = ACTIONS(2688),
    [anon_sym_alignas] = ACTIONS(2688),
    [anon_sym__Alignas] = ACTIONS(2688),
    [sym_primitive_type] = ACTIONS(2688),
    [anon_sym_enum] = ACTIONS(2688),
    [anon_sym_struct] = ACTIONS(2688),
    [anon_sym_union] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2688),
    [anon_sym___typeof] = ACTIONS(2688),
    [anon_sym_typeof] = ACTIONS(2688),
    [aux_sym_preproc_undef_token1] = ACTIONS(2688),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2688),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2688),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2688),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2688),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2688),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_API_AVAILABLE] = ACTIONS(2688),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_API_DEPRECATED] = ACTIONS(2688),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2688),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2688),
    [anon_sym___deprecated_msg] = ACTIONS(2688),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2688),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2688),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2688),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2688),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2688),
    [anon_sym_ATsynthesize] = ACTIONS(2690),
    [anon_sym_ATdynamic] = ACTIONS(2690),
    [anon_sym_ATproperty] = ACTIONS(2690),
    [anon_sym_BOOL] = ACTIONS(2688),
    [anon_sym_IMP] = ACTIONS(2688),
    [anon_sym_SEL] = ACTIONS(2688),
    [anon_sym_Class] = ACTIONS(2688),
    [anon_sym_id] = ACTIONS(2688),
  },
  [STATE(2259)] = {
    [sym_identifier] = ACTIONS(2584),
    [aux_sym_preproc_def_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token1] = ACTIONS(2584),
    [aux_sym_preproc_if_token2] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2584),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2584),
    [sym_preproc_directive] = ACTIONS(2584),
    [anon_sym_DASH] = ACTIONS(2586),
    [anon_sym_PLUS] = ACTIONS(2586),
    [anon_sym___extension__] = ACTIONS(2584),
    [anon_sym_typedef] = ACTIONS(2584),
    [anon_sym_extern] = ACTIONS(2584),
    [anon_sym___attribute__] = ACTIONS(2584),
    [anon_sym___attribute] = ACTIONS(2584),
    [anon_sym_noreturn] = ACTIONS(2584),
    [anon_sym_LBRACK] = ACTIONS(2586),
    [anon_sym___declspec] = ACTIONS(2584),
    [anon_sym___cdecl] = ACTIONS(2584),
    [anon_sym___clrcall] = ACTIONS(2584),
    [anon_sym___stdcall] = ACTIONS(2584),
    [anon_sym___fastcall] = ACTIONS(2584),
    [anon_sym___thiscall] = ACTIONS(2584),
    [anon_sym___vectorcall] = ACTIONS(2584),
    [anon_sym_signed] = ACTIONS(2584),
    [anon_sym_unsigned] = ACTIONS(2584),
    [anon_sym_long] = ACTIONS(2584),
    [anon_sym_short] = ACTIONS(2584),
    [anon_sym_static] = ACTIONS(2584),
    [anon_sym_auto] = ACTIONS(2584),
    [anon_sym_register] = ACTIONS(2584),
    [anon_sym_inline] = ACTIONS(2584),
    [anon_sym___inline] = ACTIONS(2584),
    [anon_sym___inline__] = ACTIONS(2584),
    [anon_sym___forceinline] = ACTIONS(2584),
    [anon_sym_thread_local] = ACTIONS(2584),
    [anon_sym___thread] = ACTIONS(2584),
    [anon_sym_CG_EXTERN] = ACTIONS(2584),
    [anon_sym_CG_INLINE] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2584),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2584),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2584),
    [anon_sym_IBOutlet] = ACTIONS(2584),
    [anon_sym_IBInspectable] = ACTIONS(2584),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2584),
    [anon_sym_NS_INLINE] = ACTIONS(2584),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2584),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2584),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2584),
    [anon_sym_const] = ACTIONS(2584),
    [anon_sym_constexpr] = ACTIONS(2584),
    [anon_sym_volatile] = ACTIONS(2584),
    [anon_sym_restrict] = ACTIONS(2584),
    [anon_sym___restrict__] = ACTIONS(2584),
    [anon_sym__Atomic] = ACTIONS(2584),
    [anon_sym__Noreturn] = ACTIONS(2584),
    [anon_sym__Nonnull] = ACTIONS(2584),
    [anon_sym_nullable] = ACTIONS(2584),
    [anon_sym__Complex] = ACTIONS(2584),
    [anon_sym__Nullable] = ACTIONS(2584),
    [anon_sym__Nullable_result] = ACTIONS(2584),
    [anon_sym__Null_unspecified] = ACTIONS(2584),
    [anon_sym___autoreleasing] = ACTIONS(2584),
    [anon_sym___block] = ACTIONS(2584),
    [anon_sym___bridge] = ACTIONS(2584),
    [anon_sym___bridge_retained] = ACTIONS(2584),
    [anon_sym___bridge_transfer] = ACTIONS(2584),
    [anon_sym___complex] = ACTIONS(2584),
    [anon_sym___const] = ACTIONS(2584),
    [anon_sym___imag] = ACTIONS(2584),
    [anon_sym___kindof] = ACTIONS(2584),
    [anon_sym___nonnull] = ACTIONS(2584),
    [anon_sym___nullable] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2584),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2584),
    [anon_sym___real] = ACTIONS(2584),
    [anon_sym___strong] = ACTIONS(2584),
    [anon_sym___unsafe_unretained] = ACTIONS(2584),
    [anon_sym___unused] = ACTIONS(2584),
    [anon_sym___weak] = ACTIONS(2584),
    [anon_sym_alignas] = ACTIONS(2584),
    [anon_sym__Alignas] = ACTIONS(2584),
    [sym_primitive_type] = ACTIONS(2584),
    [anon_sym_enum] = ACTIONS(2584),
    [anon_sym_struct] = ACTIONS(2584),
    [anon_sym_union] = ACTIONS(2584),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2584),
    [anon_sym___typeof] = ACTIONS(2584),
    [anon_sym_typeof] = ACTIONS(2584),
    [aux_sym_preproc_undef_token1] = ACTIONS(2584),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2584),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2584),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2584),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2584),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2584),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_API_AVAILABLE] = ACTIONS(2584),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_API_DEPRECATED] = ACTIONS(2584),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2584),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2584),
    [anon_sym___deprecated_msg] = ACTIONS(2584),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2584),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2584),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2584),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2584),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2584),
    [anon_sym_ATsynthesize] = ACTIONS(2586),
    [anon_sym_ATdynamic] = ACTIONS(2586),
    [anon_sym_ATproperty] = ACTIONS(2586),
    [anon_sym_BOOL] = ACTIONS(2584),
    [anon_sym_IMP] = ACTIONS(2584),
    [anon_sym_SEL] = ACTIONS(2584),
    [anon_sym_Class] = ACTIONS(2584),
    [anon_sym_id] = ACTIONS(2584),
  },
  [STATE(2260)] = {
    [sym_identifier] = ACTIONS(5351),
    [aux_sym_preproc_def_token1] = ACTIONS(5351),
    [aux_sym_preproc_if_token1] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5351),
    [sym_preproc_directive] = ACTIONS(5351),
    [anon_sym_DASH] = ACTIONS(5353),
    [anon_sym_PLUS] = ACTIONS(5353),
    [anon_sym_SEMI] = ACTIONS(5555),
    [anon_sym___extension__] = ACTIONS(5351),
    [anon_sym_typedef] = ACTIONS(5351),
    [anon_sym_extern] = ACTIONS(5351),
    [anon_sym___attribute__] = ACTIONS(5351),
    [anon_sym___attribute] = ACTIONS(5351),
    [anon_sym_noreturn] = ACTIONS(5351),
    [anon_sym_LBRACK] = ACTIONS(5353),
    [anon_sym___declspec] = ACTIONS(5351),
    [anon_sym___cdecl] = ACTIONS(5351),
    [anon_sym___clrcall] = ACTIONS(5351),
    [anon_sym___stdcall] = ACTIONS(5351),
    [anon_sym___fastcall] = ACTIONS(5351),
    [anon_sym___thiscall] = ACTIONS(5351),
    [anon_sym___vectorcall] = ACTIONS(5351),
    [anon_sym_signed] = ACTIONS(5351),
    [anon_sym_unsigned] = ACTIONS(5351),
    [anon_sym_long] = ACTIONS(5351),
    [anon_sym_short] = ACTIONS(5351),
    [anon_sym_static] = ACTIONS(5351),
    [anon_sym_auto] = ACTIONS(5351),
    [anon_sym_register] = ACTIONS(5351),
    [anon_sym_inline] = ACTIONS(5351),
    [anon_sym___inline] = ACTIONS(5351),
    [anon_sym___inline__] = ACTIONS(5351),
    [anon_sym___forceinline] = ACTIONS(5351),
    [anon_sym_thread_local] = ACTIONS(5351),
    [anon_sym___thread] = ACTIONS(5351),
    [anon_sym_CG_EXTERN] = ACTIONS(5351),
    [anon_sym_CG_INLINE] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5351),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5351),
    [anon_sym_IBOutlet] = ACTIONS(5351),
    [anon_sym_IBInspectable] = ACTIONS(5351),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5351),
    [anon_sym_NS_INLINE] = ACTIONS(5351),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5351),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5351),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5351),
    [anon_sym_const] = ACTIONS(5351),
    [anon_sym_constexpr] = ACTIONS(5351),
    [anon_sym_volatile] = ACTIONS(5351),
    [anon_sym_restrict] = ACTIONS(5351),
    [anon_sym___restrict__] = ACTIONS(5351),
    [anon_sym__Atomic] = ACTIONS(5351),
    [anon_sym__Noreturn] = ACTIONS(5351),
    [anon_sym__Nonnull] = ACTIONS(5351),
    [anon_sym_nullable] = ACTIONS(5351),
    [anon_sym__Complex] = ACTIONS(5351),
    [anon_sym__Nullable] = ACTIONS(5351),
    [anon_sym__Nullable_result] = ACTIONS(5351),
    [anon_sym__Null_unspecified] = ACTIONS(5351),
    [anon_sym___autoreleasing] = ACTIONS(5351),
    [anon_sym___block] = ACTIONS(5351),
    [anon_sym___bridge] = ACTIONS(5351),
    [anon_sym___bridge_retained] = ACTIONS(5351),
    [anon_sym___bridge_transfer] = ACTIONS(5351),
    [anon_sym___complex] = ACTIONS(5351),
    [anon_sym___const] = ACTIONS(5351),
    [anon_sym___imag] = ACTIONS(5351),
    [anon_sym___kindof] = ACTIONS(5351),
    [anon_sym___nonnull] = ACTIONS(5351),
    [anon_sym___nullable] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5351),
    [anon_sym___real] = ACTIONS(5351),
    [anon_sym___strong] = ACTIONS(5351),
    [anon_sym___unsafe_unretained] = ACTIONS(5351),
    [anon_sym___unused] = ACTIONS(5351),
    [anon_sym___weak] = ACTIONS(5351),
    [anon_sym_alignas] = ACTIONS(5351),
    [anon_sym__Alignas] = ACTIONS(5351),
    [sym_primitive_type] = ACTIONS(5351),
    [anon_sym_enum] = ACTIONS(5351),
    [anon_sym_struct] = ACTIONS(5351),
    [anon_sym_union] = ACTIONS(5351),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5351),
    [anon_sym___typeof] = ACTIONS(5351),
    [anon_sym_typeof] = ACTIONS(5351),
    [aux_sym_preproc_undef_token1] = ACTIONS(5351),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5351),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5351),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5351),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_API_AVAILABLE] = ACTIONS(5351),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_API_DEPRECATED] = ACTIONS(5351),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym___deprecated_msg] = ACTIONS(5351),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5351),
    [anon_sym_ATend] = ACTIONS(5353),
    [anon_sym_ATsynthesize] = ACTIONS(5353),
    [anon_sym_ATdynamic] = ACTIONS(5353),
    [anon_sym_BOOL] = ACTIONS(5351),
    [anon_sym_IMP] = ACTIONS(5351),
    [anon_sym_SEL] = ACTIONS(5351),
    [anon_sym_Class] = ACTIONS(5351),
    [anon_sym_id] = ACTIONS(5351),
  },
  [STATE(2261)] = {
    [sym_identifier] = ACTIONS(2558),
    [aux_sym_preproc_def_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token1] = ACTIONS(2558),
    [aux_sym_preproc_if_token2] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2558),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2558),
    [sym_preproc_directive] = ACTIONS(2558),
    [anon_sym_DASH] = ACTIONS(2560),
    [anon_sym_PLUS] = ACTIONS(2560),
    [anon_sym___extension__] = ACTIONS(2558),
    [anon_sym_typedef] = ACTIONS(2558),
    [anon_sym_extern] = ACTIONS(2558),
    [anon_sym___attribute__] = ACTIONS(2558),
    [anon_sym___attribute] = ACTIONS(2558),
    [anon_sym_noreturn] = ACTIONS(2558),
    [anon_sym_LBRACK] = ACTIONS(2560),
    [anon_sym___declspec] = ACTIONS(2558),
    [anon_sym___cdecl] = ACTIONS(2558),
    [anon_sym___clrcall] = ACTIONS(2558),
    [anon_sym___stdcall] = ACTIONS(2558),
    [anon_sym___fastcall] = ACTIONS(2558),
    [anon_sym___thiscall] = ACTIONS(2558),
    [anon_sym___vectorcall] = ACTIONS(2558),
    [anon_sym_signed] = ACTIONS(2558),
    [anon_sym_unsigned] = ACTIONS(2558),
    [anon_sym_long] = ACTIONS(2558),
    [anon_sym_short] = ACTIONS(2558),
    [anon_sym_static] = ACTIONS(2558),
    [anon_sym_auto] = ACTIONS(2558),
    [anon_sym_register] = ACTIONS(2558),
    [anon_sym_inline] = ACTIONS(2558),
    [anon_sym___inline] = ACTIONS(2558),
    [anon_sym___inline__] = ACTIONS(2558),
    [anon_sym___forceinline] = ACTIONS(2558),
    [anon_sym_thread_local] = ACTIONS(2558),
    [anon_sym___thread] = ACTIONS(2558),
    [anon_sym_CG_EXTERN] = ACTIONS(2558),
    [anon_sym_CG_INLINE] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2558),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2558),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2558),
    [anon_sym_IBOutlet] = ACTIONS(2558),
    [anon_sym_IBInspectable] = ACTIONS(2558),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2558),
    [anon_sym_NS_INLINE] = ACTIONS(2558),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2558),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2558),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2558),
    [anon_sym_const] = ACTIONS(2558),
    [anon_sym_constexpr] = ACTIONS(2558),
    [anon_sym_volatile] = ACTIONS(2558),
    [anon_sym_restrict] = ACTIONS(2558),
    [anon_sym___restrict__] = ACTIONS(2558),
    [anon_sym__Atomic] = ACTIONS(2558),
    [anon_sym__Noreturn] = ACTIONS(2558),
    [anon_sym__Nonnull] = ACTIONS(2558),
    [anon_sym_nullable] = ACTIONS(2558),
    [anon_sym__Complex] = ACTIONS(2558),
    [anon_sym__Nullable] = ACTIONS(2558),
    [anon_sym__Nullable_result] = ACTIONS(2558),
    [anon_sym__Null_unspecified] = ACTIONS(2558),
    [anon_sym___autoreleasing] = ACTIONS(2558),
    [anon_sym___block] = ACTIONS(2558),
    [anon_sym___bridge] = ACTIONS(2558),
    [anon_sym___bridge_retained] = ACTIONS(2558),
    [anon_sym___bridge_transfer] = ACTIONS(2558),
    [anon_sym___complex] = ACTIONS(2558),
    [anon_sym___const] = ACTIONS(2558),
    [anon_sym___imag] = ACTIONS(2558),
    [anon_sym___kindof] = ACTIONS(2558),
    [anon_sym___nonnull] = ACTIONS(2558),
    [anon_sym___nullable] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2558),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2558),
    [anon_sym___real] = ACTIONS(2558),
    [anon_sym___strong] = ACTIONS(2558),
    [anon_sym___unsafe_unretained] = ACTIONS(2558),
    [anon_sym___unused] = ACTIONS(2558),
    [anon_sym___weak] = ACTIONS(2558),
    [anon_sym_alignas] = ACTIONS(2558),
    [anon_sym__Alignas] = ACTIONS(2558),
    [sym_primitive_type] = ACTIONS(2558),
    [anon_sym_enum] = ACTIONS(2558),
    [anon_sym_struct] = ACTIONS(2558),
    [anon_sym_union] = ACTIONS(2558),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2558),
    [anon_sym___typeof] = ACTIONS(2558),
    [anon_sym_typeof] = ACTIONS(2558),
    [aux_sym_preproc_undef_token1] = ACTIONS(2558),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2558),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2558),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2558),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2558),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2558),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_API_AVAILABLE] = ACTIONS(2558),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_API_DEPRECATED] = ACTIONS(2558),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2558),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2558),
    [anon_sym___deprecated_msg] = ACTIONS(2558),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2558),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2558),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2558),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2558),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2558),
    [anon_sym_ATsynthesize] = ACTIONS(2560),
    [anon_sym_ATdynamic] = ACTIONS(2560),
    [anon_sym_ATproperty] = ACTIONS(2560),
    [anon_sym_BOOL] = ACTIONS(2558),
    [anon_sym_IMP] = ACTIONS(2558),
    [anon_sym_SEL] = ACTIONS(2558),
    [anon_sym_Class] = ACTIONS(2558),
    [anon_sym_id] = ACTIONS(2558),
  },
  [STATE(2262)] = {
    [sym_identifier] = ACTIONS(5438),
    [aux_sym_preproc_def_token1] = ACTIONS(5438),
    [aux_sym_preproc_if_token1] = ACTIONS(5438),
    [aux_sym_preproc_if_token2] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5438),
    [sym_preproc_directive] = ACTIONS(5438),
    [anon_sym_DASH] = ACTIONS(5440),
    [anon_sym_PLUS] = ACTIONS(5440),
    [anon_sym_SEMI] = ACTIONS(5557),
    [anon_sym___extension__] = ACTIONS(5438),
    [anon_sym_typedef] = ACTIONS(5438),
    [anon_sym_extern] = ACTIONS(5438),
    [anon_sym___attribute__] = ACTIONS(5438),
    [anon_sym___attribute] = ACTIONS(5438),
    [anon_sym_noreturn] = ACTIONS(5438),
    [anon_sym_LBRACK] = ACTIONS(5440),
    [anon_sym___declspec] = ACTIONS(5438),
    [anon_sym___cdecl] = ACTIONS(5438),
    [anon_sym___clrcall] = ACTIONS(5438),
    [anon_sym___stdcall] = ACTIONS(5438),
    [anon_sym___fastcall] = ACTIONS(5438),
    [anon_sym___thiscall] = ACTIONS(5438),
    [anon_sym___vectorcall] = ACTIONS(5438),
    [anon_sym_signed] = ACTIONS(5438),
    [anon_sym_unsigned] = ACTIONS(5438),
    [anon_sym_long] = ACTIONS(5438),
    [anon_sym_short] = ACTIONS(5438),
    [anon_sym_static] = ACTIONS(5438),
    [anon_sym_auto] = ACTIONS(5438),
    [anon_sym_register] = ACTIONS(5438),
    [anon_sym_inline] = ACTIONS(5438),
    [anon_sym___inline] = ACTIONS(5438),
    [anon_sym___inline__] = ACTIONS(5438),
    [anon_sym___forceinline] = ACTIONS(5438),
    [anon_sym_thread_local] = ACTIONS(5438),
    [anon_sym___thread] = ACTIONS(5438),
    [anon_sym_CG_EXTERN] = ACTIONS(5438),
    [anon_sym_CG_INLINE] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5438),
    [anon_sym_IBOutlet] = ACTIONS(5438),
    [anon_sym_IBInspectable] = ACTIONS(5438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5438),
    [anon_sym_NS_INLINE] = ACTIONS(5438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5438),
    [anon_sym_const] = ACTIONS(5438),
    [anon_sym_constexpr] = ACTIONS(5438),
    [anon_sym_volatile] = ACTIONS(5438),
    [anon_sym_restrict] = ACTIONS(5438),
    [anon_sym___restrict__] = ACTIONS(5438),
    [anon_sym__Atomic] = ACTIONS(5438),
    [anon_sym__Noreturn] = ACTIONS(5438),
    [anon_sym__Nonnull] = ACTIONS(5438),
    [anon_sym_nullable] = ACTIONS(5438),
    [anon_sym__Complex] = ACTIONS(5438),
    [anon_sym__Nullable] = ACTIONS(5438),
    [anon_sym__Nullable_result] = ACTIONS(5438),
    [anon_sym__Null_unspecified] = ACTIONS(5438),
    [anon_sym___autoreleasing] = ACTIONS(5438),
    [anon_sym___block] = ACTIONS(5438),
    [anon_sym___bridge] = ACTIONS(5438),
    [anon_sym___bridge_retained] = ACTIONS(5438),
    [anon_sym___bridge_transfer] = ACTIONS(5438),
    [anon_sym___complex] = ACTIONS(5438),
    [anon_sym___const] = ACTIONS(5438),
    [anon_sym___imag] = ACTIONS(5438),
    [anon_sym___kindof] = ACTIONS(5438),
    [anon_sym___nonnull] = ACTIONS(5438),
    [anon_sym___nullable] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5438),
    [anon_sym___real] = ACTIONS(5438),
    [anon_sym___strong] = ACTIONS(5438),
    [anon_sym___unsafe_unretained] = ACTIONS(5438),
    [anon_sym___unused] = ACTIONS(5438),
    [anon_sym___weak] = ACTIONS(5438),
    [anon_sym_alignas] = ACTIONS(5438),
    [anon_sym__Alignas] = ACTIONS(5438),
    [sym_primitive_type] = ACTIONS(5438),
    [anon_sym_enum] = ACTIONS(5438),
    [anon_sym_struct] = ACTIONS(5438),
    [anon_sym_union] = ACTIONS(5438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5438),
    [anon_sym___typeof] = ACTIONS(5438),
    [anon_sym_typeof] = ACTIONS(5438),
    [aux_sym_preproc_undef_token1] = ACTIONS(5438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_API_AVAILABLE] = ACTIONS(5438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_API_DEPRECATED] = ACTIONS(5438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym___deprecated_msg] = ACTIONS(5438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5438),
    [anon_sym_ATsynthesize] = ACTIONS(5440),
    [anon_sym_ATdynamic] = ACTIONS(5440),
    [anon_sym_BOOL] = ACTIONS(5438),
    [anon_sym_IMP] = ACTIONS(5438),
    [anon_sym_SEL] = ACTIONS(5438),
    [anon_sym_Class] = ACTIONS(5438),
    [anon_sym_id] = ACTIONS(5438),
  },
  [STATE(2263)] = {
    [sym_identifier] = ACTIONS(2596),
    [aux_sym_preproc_def_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token1] = ACTIONS(2596),
    [aux_sym_preproc_if_token2] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2596),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2596),
    [sym_preproc_directive] = ACTIONS(2596),
    [anon_sym_DASH] = ACTIONS(2598),
    [anon_sym_PLUS] = ACTIONS(2598),
    [anon_sym___extension__] = ACTIONS(2596),
    [anon_sym_typedef] = ACTIONS(2596),
    [anon_sym_extern] = ACTIONS(2596),
    [anon_sym___attribute__] = ACTIONS(2596),
    [anon_sym___attribute] = ACTIONS(2596),
    [anon_sym_noreturn] = ACTIONS(2596),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym___declspec] = ACTIONS(2596),
    [anon_sym___cdecl] = ACTIONS(2596),
    [anon_sym___clrcall] = ACTIONS(2596),
    [anon_sym___stdcall] = ACTIONS(2596),
    [anon_sym___fastcall] = ACTIONS(2596),
    [anon_sym___thiscall] = ACTIONS(2596),
    [anon_sym___vectorcall] = ACTIONS(2596),
    [anon_sym_signed] = ACTIONS(2596),
    [anon_sym_unsigned] = ACTIONS(2596),
    [anon_sym_long] = ACTIONS(2596),
    [anon_sym_short] = ACTIONS(2596),
    [anon_sym_static] = ACTIONS(2596),
    [anon_sym_auto] = ACTIONS(2596),
    [anon_sym_register] = ACTIONS(2596),
    [anon_sym_inline] = ACTIONS(2596),
    [anon_sym___inline] = ACTIONS(2596),
    [anon_sym___inline__] = ACTIONS(2596),
    [anon_sym___forceinline] = ACTIONS(2596),
    [anon_sym_thread_local] = ACTIONS(2596),
    [anon_sym___thread] = ACTIONS(2596),
    [anon_sym_CG_EXTERN] = ACTIONS(2596),
    [anon_sym_CG_INLINE] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2596),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2596),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2596),
    [anon_sym_IBOutlet] = ACTIONS(2596),
    [anon_sym_IBInspectable] = ACTIONS(2596),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2596),
    [anon_sym_NS_INLINE] = ACTIONS(2596),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2596),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2596),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2596),
    [anon_sym_const] = ACTIONS(2596),
    [anon_sym_constexpr] = ACTIONS(2596),
    [anon_sym_volatile] = ACTIONS(2596),
    [anon_sym_restrict] = ACTIONS(2596),
    [anon_sym___restrict__] = ACTIONS(2596),
    [anon_sym__Atomic] = ACTIONS(2596),
    [anon_sym__Noreturn] = ACTIONS(2596),
    [anon_sym__Nonnull] = ACTIONS(2596),
    [anon_sym_nullable] = ACTIONS(2596),
    [anon_sym__Complex] = ACTIONS(2596),
    [anon_sym__Nullable] = ACTIONS(2596),
    [anon_sym__Nullable_result] = ACTIONS(2596),
    [anon_sym__Null_unspecified] = ACTIONS(2596),
    [anon_sym___autoreleasing] = ACTIONS(2596),
    [anon_sym___block] = ACTIONS(2596),
    [anon_sym___bridge] = ACTIONS(2596),
    [anon_sym___bridge_retained] = ACTIONS(2596),
    [anon_sym___bridge_transfer] = ACTIONS(2596),
    [anon_sym___complex] = ACTIONS(2596),
    [anon_sym___const] = ACTIONS(2596),
    [anon_sym___imag] = ACTIONS(2596),
    [anon_sym___kindof] = ACTIONS(2596),
    [anon_sym___nonnull] = ACTIONS(2596),
    [anon_sym___nullable] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2596),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2596),
    [anon_sym___real] = ACTIONS(2596),
    [anon_sym___strong] = ACTIONS(2596),
    [anon_sym___unsafe_unretained] = ACTIONS(2596),
    [anon_sym___unused] = ACTIONS(2596),
    [anon_sym___weak] = ACTIONS(2596),
    [anon_sym_alignas] = ACTIONS(2596),
    [anon_sym__Alignas] = ACTIONS(2596),
    [sym_primitive_type] = ACTIONS(2596),
    [anon_sym_enum] = ACTIONS(2596),
    [anon_sym_struct] = ACTIONS(2596),
    [anon_sym_union] = ACTIONS(2596),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2596),
    [anon_sym___typeof] = ACTIONS(2596),
    [anon_sym_typeof] = ACTIONS(2596),
    [aux_sym_preproc_undef_token1] = ACTIONS(2596),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2596),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2596),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2596),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2596),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2596),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_API_AVAILABLE] = ACTIONS(2596),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_API_DEPRECATED] = ACTIONS(2596),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2596),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2596),
    [anon_sym___deprecated_msg] = ACTIONS(2596),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2596),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2596),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2596),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2596),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2596),
    [anon_sym_ATsynthesize] = ACTIONS(2598),
    [anon_sym_ATdynamic] = ACTIONS(2598),
    [anon_sym_ATproperty] = ACTIONS(2598),
    [anon_sym_BOOL] = ACTIONS(2596),
    [anon_sym_IMP] = ACTIONS(2596),
    [anon_sym_SEL] = ACTIONS(2596),
    [anon_sym_Class] = ACTIONS(2596),
    [anon_sym_id] = ACTIONS(2596),
  },
  [STATE(2264)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token2] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_DASH] = ACTIONS(2476),
    [anon_sym_PLUS] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATproperty] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(2265)] = {
    [sym_identifier] = ACTIONS(2474),
    [aux_sym_preproc_def_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token1] = ACTIONS(2474),
    [aux_sym_preproc_if_token2] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2474),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2474),
    [sym_preproc_directive] = ACTIONS(2474),
    [anon_sym_DASH] = ACTIONS(2476),
    [anon_sym_PLUS] = ACTIONS(2476),
    [anon_sym___extension__] = ACTIONS(2474),
    [anon_sym_typedef] = ACTIONS(2474),
    [anon_sym_extern] = ACTIONS(2474),
    [anon_sym___attribute__] = ACTIONS(2474),
    [anon_sym___attribute] = ACTIONS(2474),
    [anon_sym_noreturn] = ACTIONS(2474),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym___declspec] = ACTIONS(2474),
    [anon_sym___cdecl] = ACTIONS(2474),
    [anon_sym___clrcall] = ACTIONS(2474),
    [anon_sym___stdcall] = ACTIONS(2474),
    [anon_sym___fastcall] = ACTIONS(2474),
    [anon_sym___thiscall] = ACTIONS(2474),
    [anon_sym___vectorcall] = ACTIONS(2474),
    [anon_sym_signed] = ACTIONS(2474),
    [anon_sym_unsigned] = ACTIONS(2474),
    [anon_sym_long] = ACTIONS(2474),
    [anon_sym_short] = ACTIONS(2474),
    [anon_sym_static] = ACTIONS(2474),
    [anon_sym_auto] = ACTIONS(2474),
    [anon_sym_register] = ACTIONS(2474),
    [anon_sym_inline] = ACTIONS(2474),
    [anon_sym___inline] = ACTIONS(2474),
    [anon_sym___inline__] = ACTIONS(2474),
    [anon_sym___forceinline] = ACTIONS(2474),
    [anon_sym_thread_local] = ACTIONS(2474),
    [anon_sym___thread] = ACTIONS(2474),
    [anon_sym_CG_EXTERN] = ACTIONS(2474),
    [anon_sym_CG_INLINE] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2474),
    [anon_sym_IBOutlet] = ACTIONS(2474),
    [anon_sym_IBInspectable] = ACTIONS(2474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2474),
    [anon_sym_NS_INLINE] = ACTIONS(2474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2474),
    [anon_sym_const] = ACTIONS(2474),
    [anon_sym_constexpr] = ACTIONS(2474),
    [anon_sym_volatile] = ACTIONS(2474),
    [anon_sym_restrict] = ACTIONS(2474),
    [anon_sym___restrict__] = ACTIONS(2474),
    [anon_sym__Atomic] = ACTIONS(2474),
    [anon_sym__Noreturn] = ACTIONS(2474),
    [anon_sym__Nonnull] = ACTIONS(2474),
    [anon_sym_nullable] = ACTIONS(2474),
    [anon_sym__Complex] = ACTIONS(2474),
    [anon_sym__Nullable] = ACTIONS(2474),
    [anon_sym__Nullable_result] = ACTIONS(2474),
    [anon_sym__Null_unspecified] = ACTIONS(2474),
    [anon_sym___autoreleasing] = ACTIONS(2474),
    [anon_sym___block] = ACTIONS(2474),
    [anon_sym___bridge] = ACTIONS(2474),
    [anon_sym___bridge_retained] = ACTIONS(2474),
    [anon_sym___bridge_transfer] = ACTIONS(2474),
    [anon_sym___complex] = ACTIONS(2474),
    [anon_sym___const] = ACTIONS(2474),
    [anon_sym___imag] = ACTIONS(2474),
    [anon_sym___kindof] = ACTIONS(2474),
    [anon_sym___nonnull] = ACTIONS(2474),
    [anon_sym___nullable] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2474),
    [anon_sym___real] = ACTIONS(2474),
    [anon_sym___strong] = ACTIONS(2474),
    [anon_sym___unsafe_unretained] = ACTIONS(2474),
    [anon_sym___unused] = ACTIONS(2474),
    [anon_sym___weak] = ACTIONS(2474),
    [anon_sym_alignas] = ACTIONS(2474),
    [anon_sym__Alignas] = ACTIONS(2474),
    [sym_primitive_type] = ACTIONS(2474),
    [anon_sym_enum] = ACTIONS(2474),
    [anon_sym_struct] = ACTIONS(2474),
    [anon_sym_union] = ACTIONS(2474),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2474),
    [anon_sym___typeof] = ACTIONS(2474),
    [anon_sym_typeof] = ACTIONS(2474),
    [aux_sym_preproc_undef_token1] = ACTIONS(2474),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_API_AVAILABLE] = ACTIONS(2474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_API_DEPRECATED] = ACTIONS(2474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2474),
    [anon_sym___deprecated_msg] = ACTIONS(2474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2474),
    [anon_sym_ATsynthesize] = ACTIONS(2476),
    [anon_sym_ATdynamic] = ACTIONS(2476),
    [anon_sym_ATproperty] = ACTIONS(2476),
    [anon_sym_BOOL] = ACTIONS(2474),
    [anon_sym_IMP] = ACTIONS(2474),
    [anon_sym_SEL] = ACTIONS(2474),
    [anon_sym_Class] = ACTIONS(2474),
    [anon_sym_id] = ACTIONS(2474),
  },
  [STATE(2266)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token2] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_DASH] = ACTIONS(2418),
    [anon_sym_PLUS] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATproperty] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(2267)] = {
    [sym_identifier] = ACTIONS(5529),
    [aux_sym_preproc_def_token1] = ACTIONS(5529),
    [aux_sym_preproc_if_token1] = ACTIONS(5529),
    [aux_sym_preproc_if_token2] = ACTIONS(5529),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5529),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5529),
    [aux_sym_preproc_else_token1] = ACTIONS(5529),
    [aux_sym_preproc_elif_token1] = ACTIONS(5529),
    [sym_preproc_directive] = ACTIONS(5529),
    [anon_sym_DASH] = ACTIONS(5531),
    [anon_sym_PLUS] = ACTIONS(5531),
    [anon_sym___extension__] = ACTIONS(5529),
    [anon_sym_typedef] = ACTIONS(5529),
    [anon_sym_extern] = ACTIONS(5529),
    [anon_sym___attribute__] = ACTIONS(5529),
    [anon_sym___attribute] = ACTIONS(5529),
    [anon_sym_noreturn] = ACTIONS(5529),
    [anon_sym_LBRACK] = ACTIONS(5531),
    [anon_sym___declspec] = ACTIONS(5529),
    [anon_sym___cdecl] = ACTIONS(5529),
    [anon_sym___clrcall] = ACTIONS(5529),
    [anon_sym___stdcall] = ACTIONS(5529),
    [anon_sym___fastcall] = ACTIONS(5529),
    [anon_sym___thiscall] = ACTIONS(5529),
    [anon_sym___vectorcall] = ACTIONS(5529),
    [anon_sym_signed] = ACTIONS(5529),
    [anon_sym_unsigned] = ACTIONS(5529),
    [anon_sym_long] = ACTIONS(5529),
    [anon_sym_short] = ACTIONS(5529),
    [anon_sym_static] = ACTIONS(5529),
    [anon_sym_auto] = ACTIONS(5529),
    [anon_sym_register] = ACTIONS(5529),
    [anon_sym_inline] = ACTIONS(5529),
    [anon_sym___inline] = ACTIONS(5529),
    [anon_sym___inline__] = ACTIONS(5529),
    [anon_sym___forceinline] = ACTIONS(5529),
    [anon_sym_thread_local] = ACTIONS(5529),
    [anon_sym___thread] = ACTIONS(5529),
    [anon_sym_CG_EXTERN] = ACTIONS(5529),
    [anon_sym_CG_INLINE] = ACTIONS(5529),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5529),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5529),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5529),
    [anon_sym_IBOutlet] = ACTIONS(5529),
    [anon_sym_IBInspectable] = ACTIONS(5529),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5529),
    [anon_sym_NS_INLINE] = ACTIONS(5529),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5529),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5529),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5529),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5529),
    [anon_sym_const] = ACTIONS(5529),
    [anon_sym_constexpr] = ACTIONS(5529),
    [anon_sym_volatile] = ACTIONS(5529),
    [anon_sym_restrict] = ACTIONS(5529),
    [anon_sym___restrict__] = ACTIONS(5529),
    [anon_sym__Atomic] = ACTIONS(5529),
    [anon_sym__Noreturn] = ACTIONS(5529),
    [anon_sym__Nonnull] = ACTIONS(5529),
    [anon_sym_nullable] = ACTIONS(5529),
    [anon_sym__Complex] = ACTIONS(5529),
    [anon_sym__Nullable] = ACTIONS(5529),
    [anon_sym__Nullable_result] = ACTIONS(5529),
    [anon_sym__Null_unspecified] = ACTIONS(5529),
    [anon_sym___autoreleasing] = ACTIONS(5529),
    [anon_sym___block] = ACTIONS(5529),
    [anon_sym___bridge] = ACTIONS(5529),
    [anon_sym___bridge_retained] = ACTIONS(5529),
    [anon_sym___bridge_transfer] = ACTIONS(5529),
    [anon_sym___complex] = ACTIONS(5529),
    [anon_sym___const] = ACTIONS(5529),
    [anon_sym___imag] = ACTIONS(5529),
    [anon_sym___kindof] = ACTIONS(5529),
    [anon_sym___nonnull] = ACTIONS(5529),
    [anon_sym___nullable] = ACTIONS(5529),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5529),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5529),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5529),
    [anon_sym___real] = ACTIONS(5529),
    [anon_sym___strong] = ACTIONS(5529),
    [anon_sym___unsafe_unretained] = ACTIONS(5529),
    [anon_sym___unused] = ACTIONS(5529),
    [anon_sym___weak] = ACTIONS(5529),
    [anon_sym_alignas] = ACTIONS(5529),
    [anon_sym__Alignas] = ACTIONS(5529),
    [sym_primitive_type] = ACTIONS(5529),
    [anon_sym_enum] = ACTIONS(5529),
    [anon_sym_struct] = ACTIONS(5529),
    [anon_sym_union] = ACTIONS(5529),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5529),
    [anon_sym___typeof] = ACTIONS(5529),
    [anon_sym_typeof] = ACTIONS(5529),
    [aux_sym_preproc_undef_token1] = ACTIONS(5529),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5529),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5529),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5529),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5529),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5529),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5529),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5529),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5529),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5529),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_API_AVAILABLE] = ACTIONS(5529),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_API_DEPRECATED] = ACTIONS(5529),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5529),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5529),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5529),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5529),
    [anon_sym___deprecated_msg] = ACTIONS(5529),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5529),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5529),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5529),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5529),
    [anon_sym_ATproperty] = ACTIONS(5531),
    [anon_sym_BOOL] = ACTIONS(5529),
    [anon_sym_IMP] = ACTIONS(5529),
    [anon_sym_SEL] = ACTIONS(5529),
    [anon_sym_Class] = ACTIONS(5529),
    [anon_sym_id] = ACTIONS(5529),
  },
  [STATE(2268)] = {
    [sym_identifier] = ACTIONS(5448),
    [aux_sym_preproc_def_token1] = ACTIONS(5448),
    [aux_sym_preproc_if_token1] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5448),
    [sym_preproc_directive] = ACTIONS(5448),
    [anon_sym_DASH] = ACTIONS(5450),
    [anon_sym_PLUS] = ACTIONS(5450),
    [anon_sym_SEMI] = ACTIONS(5559),
    [anon_sym___extension__] = ACTIONS(5448),
    [anon_sym_typedef] = ACTIONS(5448),
    [anon_sym_extern] = ACTIONS(5448),
    [anon_sym___attribute__] = ACTIONS(5448),
    [anon_sym___attribute] = ACTIONS(5448),
    [anon_sym_noreturn] = ACTIONS(5448),
    [anon_sym_LBRACK] = ACTIONS(5450),
    [anon_sym___declspec] = ACTIONS(5448),
    [anon_sym___cdecl] = ACTIONS(5448),
    [anon_sym___clrcall] = ACTIONS(5448),
    [anon_sym___stdcall] = ACTIONS(5448),
    [anon_sym___fastcall] = ACTIONS(5448),
    [anon_sym___thiscall] = ACTIONS(5448),
    [anon_sym___vectorcall] = ACTIONS(5448),
    [anon_sym_signed] = ACTIONS(5448),
    [anon_sym_unsigned] = ACTIONS(5448),
    [anon_sym_long] = ACTIONS(5448),
    [anon_sym_short] = ACTIONS(5448),
    [anon_sym_static] = ACTIONS(5448),
    [anon_sym_auto] = ACTIONS(5448),
    [anon_sym_register] = ACTIONS(5448),
    [anon_sym_inline] = ACTIONS(5448),
    [anon_sym___inline] = ACTIONS(5448),
    [anon_sym___inline__] = ACTIONS(5448),
    [anon_sym___forceinline] = ACTIONS(5448),
    [anon_sym_thread_local] = ACTIONS(5448),
    [anon_sym___thread] = ACTIONS(5448),
    [anon_sym_CG_EXTERN] = ACTIONS(5448),
    [anon_sym_CG_INLINE] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5448),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5448),
    [anon_sym_IBOutlet] = ACTIONS(5448),
    [anon_sym_IBInspectable] = ACTIONS(5448),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5448),
    [anon_sym_NS_INLINE] = ACTIONS(5448),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5448),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5448),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5448),
    [anon_sym_const] = ACTIONS(5448),
    [anon_sym_constexpr] = ACTIONS(5448),
    [anon_sym_volatile] = ACTIONS(5448),
    [anon_sym_restrict] = ACTIONS(5448),
    [anon_sym___restrict__] = ACTIONS(5448),
    [anon_sym__Atomic] = ACTIONS(5448),
    [anon_sym__Noreturn] = ACTIONS(5448),
    [anon_sym__Nonnull] = ACTIONS(5448),
    [anon_sym_nullable] = ACTIONS(5448),
    [anon_sym__Complex] = ACTIONS(5448),
    [anon_sym__Nullable] = ACTIONS(5448),
    [anon_sym__Nullable_result] = ACTIONS(5448),
    [anon_sym__Null_unspecified] = ACTIONS(5448),
    [anon_sym___autoreleasing] = ACTIONS(5448),
    [anon_sym___block] = ACTIONS(5448),
    [anon_sym___bridge] = ACTIONS(5448),
    [anon_sym___bridge_retained] = ACTIONS(5448),
    [anon_sym___bridge_transfer] = ACTIONS(5448),
    [anon_sym___complex] = ACTIONS(5448),
    [anon_sym___const] = ACTIONS(5448),
    [anon_sym___imag] = ACTIONS(5448),
    [anon_sym___kindof] = ACTIONS(5448),
    [anon_sym___nonnull] = ACTIONS(5448),
    [anon_sym___nullable] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5448),
    [anon_sym___real] = ACTIONS(5448),
    [anon_sym___strong] = ACTIONS(5448),
    [anon_sym___unsafe_unretained] = ACTIONS(5448),
    [anon_sym___unused] = ACTIONS(5448),
    [anon_sym___weak] = ACTIONS(5448),
    [anon_sym_alignas] = ACTIONS(5448),
    [anon_sym__Alignas] = ACTIONS(5448),
    [sym_primitive_type] = ACTIONS(5448),
    [anon_sym_enum] = ACTIONS(5448),
    [anon_sym_struct] = ACTIONS(5448),
    [anon_sym_union] = ACTIONS(5448),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5448),
    [anon_sym___typeof] = ACTIONS(5448),
    [anon_sym_typeof] = ACTIONS(5448),
    [aux_sym_preproc_undef_token1] = ACTIONS(5448),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5448),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5448),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5448),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_API_AVAILABLE] = ACTIONS(5448),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_API_DEPRECATED] = ACTIONS(5448),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym___deprecated_msg] = ACTIONS(5448),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5448),
    [anon_sym_ATend] = ACTIONS(5450),
    [anon_sym_ATsynthesize] = ACTIONS(5450),
    [anon_sym_ATdynamic] = ACTIONS(5450),
    [anon_sym_BOOL] = ACTIONS(5448),
    [anon_sym_IMP] = ACTIONS(5448),
    [anon_sym_SEL] = ACTIONS(5448),
    [anon_sym_Class] = ACTIONS(5448),
    [anon_sym_id] = ACTIONS(5448),
  },
  [STATE(2269)] = {
    [sym_identifier] = ACTIONS(5561),
    [aux_sym_preproc_def_token1] = ACTIONS(5561),
    [aux_sym_preproc_if_token1] = ACTIONS(5561),
    [aux_sym_preproc_if_token2] = ACTIONS(5561),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5561),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5561),
    [aux_sym_preproc_else_token1] = ACTIONS(5561),
    [aux_sym_preproc_elif_token1] = ACTIONS(5561),
    [sym_preproc_directive] = ACTIONS(5561),
    [anon_sym_DASH] = ACTIONS(5563),
    [anon_sym_PLUS] = ACTIONS(5563),
    [anon_sym___extension__] = ACTIONS(5561),
    [anon_sym_typedef] = ACTIONS(5561),
    [anon_sym_extern] = ACTIONS(5561),
    [anon_sym___attribute__] = ACTIONS(5561),
    [anon_sym___attribute] = ACTIONS(5561),
    [anon_sym_noreturn] = ACTIONS(5561),
    [anon_sym_LBRACK] = ACTIONS(5563),
    [anon_sym___declspec] = ACTIONS(5561),
    [anon_sym___cdecl] = ACTIONS(5561),
    [anon_sym___clrcall] = ACTIONS(5561),
    [anon_sym___stdcall] = ACTIONS(5561),
    [anon_sym___fastcall] = ACTIONS(5561),
    [anon_sym___thiscall] = ACTIONS(5561),
    [anon_sym___vectorcall] = ACTIONS(5561),
    [anon_sym_signed] = ACTIONS(5561),
    [anon_sym_unsigned] = ACTIONS(5561),
    [anon_sym_long] = ACTIONS(5561),
    [anon_sym_short] = ACTIONS(5561),
    [anon_sym_static] = ACTIONS(5561),
    [anon_sym_auto] = ACTIONS(5561),
    [anon_sym_register] = ACTIONS(5561),
    [anon_sym_inline] = ACTIONS(5561),
    [anon_sym___inline] = ACTIONS(5561),
    [anon_sym___inline__] = ACTIONS(5561),
    [anon_sym___forceinline] = ACTIONS(5561),
    [anon_sym_thread_local] = ACTIONS(5561),
    [anon_sym___thread] = ACTIONS(5561),
    [anon_sym_CG_EXTERN] = ACTIONS(5561),
    [anon_sym_CG_INLINE] = ACTIONS(5561),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5561),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5561),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5561),
    [anon_sym_IBOutlet] = ACTIONS(5561),
    [anon_sym_IBInspectable] = ACTIONS(5561),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5561),
    [anon_sym_NS_INLINE] = ACTIONS(5561),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5561),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5561),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5561),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5561),
    [anon_sym_const] = ACTIONS(5561),
    [anon_sym_constexpr] = ACTIONS(5561),
    [anon_sym_volatile] = ACTIONS(5561),
    [anon_sym_restrict] = ACTIONS(5561),
    [anon_sym___restrict__] = ACTIONS(5561),
    [anon_sym__Atomic] = ACTIONS(5561),
    [anon_sym__Noreturn] = ACTIONS(5561),
    [anon_sym__Nonnull] = ACTIONS(5561),
    [anon_sym_nullable] = ACTIONS(5561),
    [anon_sym__Complex] = ACTIONS(5561),
    [anon_sym__Nullable] = ACTIONS(5561),
    [anon_sym__Nullable_result] = ACTIONS(5561),
    [anon_sym__Null_unspecified] = ACTIONS(5561),
    [anon_sym___autoreleasing] = ACTIONS(5561),
    [anon_sym___block] = ACTIONS(5561),
    [anon_sym___bridge] = ACTIONS(5561),
    [anon_sym___bridge_retained] = ACTIONS(5561),
    [anon_sym___bridge_transfer] = ACTIONS(5561),
    [anon_sym___complex] = ACTIONS(5561),
    [anon_sym___const] = ACTIONS(5561),
    [anon_sym___imag] = ACTIONS(5561),
    [anon_sym___kindof] = ACTIONS(5561),
    [anon_sym___nonnull] = ACTIONS(5561),
    [anon_sym___nullable] = ACTIONS(5561),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5561),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5561),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5561),
    [anon_sym___real] = ACTIONS(5561),
    [anon_sym___strong] = ACTIONS(5561),
    [anon_sym___unsafe_unretained] = ACTIONS(5561),
    [anon_sym___unused] = ACTIONS(5561),
    [anon_sym___weak] = ACTIONS(5561),
    [anon_sym_alignas] = ACTIONS(5561),
    [anon_sym__Alignas] = ACTIONS(5561),
    [sym_primitive_type] = ACTIONS(5561),
    [anon_sym_enum] = ACTIONS(5561),
    [anon_sym_struct] = ACTIONS(5561),
    [anon_sym_union] = ACTIONS(5561),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5561),
    [anon_sym___typeof] = ACTIONS(5561),
    [anon_sym_typeof] = ACTIONS(5561),
    [aux_sym_preproc_undef_token1] = ACTIONS(5561),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5561),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5561),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5561),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5561),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5561),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5561),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5561),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5561),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5561),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_API_AVAILABLE] = ACTIONS(5561),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_API_DEPRECATED] = ACTIONS(5561),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5561),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5561),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5561),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5561),
    [anon_sym___deprecated_msg] = ACTIONS(5561),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5561),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5561),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5561),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5561),
    [anon_sym_ATproperty] = ACTIONS(5563),
    [anon_sym_BOOL] = ACTIONS(5561),
    [anon_sym_IMP] = ACTIONS(5561),
    [anon_sym_SEL] = ACTIONS(5561),
    [anon_sym_Class] = ACTIONS(5561),
    [anon_sym_id] = ACTIONS(5561),
  },
  [STATE(2270)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2321),
    [sym_identifier] = ACTIONS(5411),
    [aux_sym_preproc_def_token1] = ACTIONS(5411),
    [aux_sym_preproc_if_token1] = ACTIONS(5411),
    [aux_sym_preproc_if_token2] = ACTIONS(5411),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5411),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5411),
    [sym_preproc_directive] = ACTIONS(5411),
    [anon_sym_DASH] = ACTIONS(5413),
    [anon_sym_PLUS] = ACTIONS(5413),
    [anon_sym_SEMI] = ACTIONS(5565),
    [anon_sym___extension__] = ACTIONS(5411),
    [anon_sym_typedef] = ACTIONS(5411),
    [anon_sym_extern] = ACTIONS(5411),
    [anon_sym___attribute__] = ACTIONS(5411),
    [anon_sym___attribute] = ACTIONS(5411),
    [anon_sym_noreturn] = ACTIONS(5411),
    [anon_sym_LBRACK] = ACTIONS(5413),
    [anon_sym___declspec] = ACTIONS(5411),
    [anon_sym___cdecl] = ACTIONS(5411),
    [anon_sym___clrcall] = ACTIONS(5411),
    [anon_sym___stdcall] = ACTIONS(5411),
    [anon_sym___fastcall] = ACTIONS(5411),
    [anon_sym___thiscall] = ACTIONS(5411),
    [anon_sym___vectorcall] = ACTIONS(5411),
    [anon_sym_signed] = ACTIONS(5411),
    [anon_sym_unsigned] = ACTIONS(5411),
    [anon_sym_long] = ACTIONS(5411),
    [anon_sym_short] = ACTIONS(5411),
    [anon_sym_static] = ACTIONS(5411),
    [anon_sym_auto] = ACTIONS(5411),
    [anon_sym_register] = ACTIONS(5411),
    [anon_sym_inline] = ACTIONS(5411),
    [anon_sym___inline] = ACTIONS(5411),
    [anon_sym___inline__] = ACTIONS(5411),
    [anon_sym___forceinline] = ACTIONS(5411),
    [anon_sym_thread_local] = ACTIONS(5411),
    [anon_sym___thread] = ACTIONS(5411),
    [anon_sym_CG_EXTERN] = ACTIONS(5411),
    [anon_sym_CG_INLINE] = ACTIONS(5411),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5411),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5411),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5411),
    [anon_sym_IBOutlet] = ACTIONS(5411),
    [anon_sym_IBInspectable] = ACTIONS(5411),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5411),
    [anon_sym_NS_INLINE] = ACTIONS(5411),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5411),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5411),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5411),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5411),
    [anon_sym_const] = ACTIONS(5411),
    [anon_sym_constexpr] = ACTIONS(5411),
    [anon_sym_volatile] = ACTIONS(5411),
    [anon_sym_restrict] = ACTIONS(5411),
    [anon_sym___restrict__] = ACTIONS(5411),
    [anon_sym__Atomic] = ACTIONS(5411),
    [anon_sym__Noreturn] = ACTIONS(5411),
    [anon_sym__Nonnull] = ACTIONS(5411),
    [anon_sym_nullable] = ACTIONS(5411),
    [anon_sym__Complex] = ACTIONS(5411),
    [anon_sym__Nullable] = ACTIONS(5411),
    [anon_sym__Nullable_result] = ACTIONS(5411),
    [anon_sym__Null_unspecified] = ACTIONS(5411),
    [anon_sym___autoreleasing] = ACTIONS(5411),
    [anon_sym___block] = ACTIONS(5411),
    [anon_sym___bridge] = ACTIONS(5411),
    [anon_sym___bridge_retained] = ACTIONS(5411),
    [anon_sym___bridge_transfer] = ACTIONS(5411),
    [anon_sym___complex] = ACTIONS(5411),
    [anon_sym___const] = ACTIONS(5411),
    [anon_sym___imag] = ACTIONS(5411),
    [anon_sym___kindof] = ACTIONS(5411),
    [anon_sym___nonnull] = ACTIONS(5411),
    [anon_sym___nullable] = ACTIONS(5411),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5411),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5411),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5411),
    [anon_sym___real] = ACTIONS(5411),
    [anon_sym___strong] = ACTIONS(5411),
    [anon_sym___unsafe_unretained] = ACTIONS(5411),
    [anon_sym___unused] = ACTIONS(5411),
    [anon_sym___weak] = ACTIONS(5411),
    [anon_sym_alignas] = ACTIONS(5411),
    [anon_sym__Alignas] = ACTIONS(5411),
    [sym_primitive_type] = ACTIONS(5411),
    [anon_sym_enum] = ACTIONS(5411),
    [anon_sym_struct] = ACTIONS(5411),
    [anon_sym_union] = ACTIONS(5411),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5411),
    [anon_sym___typeof] = ACTIONS(5411),
    [anon_sym_typeof] = ACTIONS(5411),
    [aux_sym_preproc_undef_token1] = ACTIONS(5411),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5411),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5411),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5411),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5411),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5411),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5411),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5411),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5411),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5411),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_API_AVAILABLE] = ACTIONS(5411),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_API_DEPRECATED] = ACTIONS(5411),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5411),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5411),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5411),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5411),
    [anon_sym___deprecated_msg] = ACTIONS(5411),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5411),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5411),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5411),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5411),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5411),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5411),
    [anon_sym_ATproperty] = ACTIONS(5413),
    [anon_sym_BOOL] = ACTIONS(5411),
    [anon_sym_IMP] = ACTIONS(5411),
    [anon_sym_SEL] = ACTIONS(5411),
    [anon_sym_Class] = ACTIONS(5411),
    [anon_sym_id] = ACTIONS(5411),
  },
  [STATE(2271)] = {
    [sym_identifier] = ACTIONS(5359),
    [aux_sym_preproc_def_token1] = ACTIONS(5359),
    [aux_sym_preproc_if_token1] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5359),
    [sym_preproc_directive] = ACTIONS(5359),
    [anon_sym_DASH] = ACTIONS(5361),
    [anon_sym_PLUS] = ACTIONS(5361),
    [anon_sym_SEMI] = ACTIONS(5567),
    [anon_sym___extension__] = ACTIONS(5359),
    [anon_sym_typedef] = ACTIONS(5359),
    [anon_sym_extern] = ACTIONS(5359),
    [anon_sym___attribute__] = ACTIONS(5359),
    [anon_sym___attribute] = ACTIONS(5359),
    [anon_sym_noreturn] = ACTIONS(5359),
    [anon_sym_LBRACK] = ACTIONS(5361),
    [anon_sym___declspec] = ACTIONS(5359),
    [anon_sym___cdecl] = ACTIONS(5359),
    [anon_sym___clrcall] = ACTIONS(5359),
    [anon_sym___stdcall] = ACTIONS(5359),
    [anon_sym___fastcall] = ACTIONS(5359),
    [anon_sym___thiscall] = ACTIONS(5359),
    [anon_sym___vectorcall] = ACTIONS(5359),
    [anon_sym_signed] = ACTIONS(5359),
    [anon_sym_unsigned] = ACTIONS(5359),
    [anon_sym_long] = ACTIONS(5359),
    [anon_sym_short] = ACTIONS(5359),
    [anon_sym_static] = ACTIONS(5359),
    [anon_sym_auto] = ACTIONS(5359),
    [anon_sym_register] = ACTIONS(5359),
    [anon_sym_inline] = ACTIONS(5359),
    [anon_sym___inline] = ACTIONS(5359),
    [anon_sym___inline__] = ACTIONS(5359),
    [anon_sym___forceinline] = ACTIONS(5359),
    [anon_sym_thread_local] = ACTIONS(5359),
    [anon_sym___thread] = ACTIONS(5359),
    [anon_sym_CG_EXTERN] = ACTIONS(5359),
    [anon_sym_CG_INLINE] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5359),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5359),
    [anon_sym_IBOutlet] = ACTIONS(5359),
    [anon_sym_IBInspectable] = ACTIONS(5359),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5359),
    [anon_sym_NS_INLINE] = ACTIONS(5359),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5359),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5359),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5359),
    [anon_sym_const] = ACTIONS(5359),
    [anon_sym_constexpr] = ACTIONS(5359),
    [anon_sym_volatile] = ACTIONS(5359),
    [anon_sym_restrict] = ACTIONS(5359),
    [anon_sym___restrict__] = ACTIONS(5359),
    [anon_sym__Atomic] = ACTIONS(5359),
    [anon_sym__Noreturn] = ACTIONS(5359),
    [anon_sym__Nonnull] = ACTIONS(5359),
    [anon_sym_nullable] = ACTIONS(5359),
    [anon_sym__Complex] = ACTIONS(5359),
    [anon_sym__Nullable] = ACTIONS(5359),
    [anon_sym__Nullable_result] = ACTIONS(5359),
    [anon_sym__Null_unspecified] = ACTIONS(5359),
    [anon_sym___autoreleasing] = ACTIONS(5359),
    [anon_sym___block] = ACTIONS(5359),
    [anon_sym___bridge] = ACTIONS(5359),
    [anon_sym___bridge_retained] = ACTIONS(5359),
    [anon_sym___bridge_transfer] = ACTIONS(5359),
    [anon_sym___complex] = ACTIONS(5359),
    [anon_sym___const] = ACTIONS(5359),
    [anon_sym___imag] = ACTIONS(5359),
    [anon_sym___kindof] = ACTIONS(5359),
    [anon_sym___nonnull] = ACTIONS(5359),
    [anon_sym___nullable] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5359),
    [anon_sym___real] = ACTIONS(5359),
    [anon_sym___strong] = ACTIONS(5359),
    [anon_sym___unsafe_unretained] = ACTIONS(5359),
    [anon_sym___unused] = ACTIONS(5359),
    [anon_sym___weak] = ACTIONS(5359),
    [anon_sym_alignas] = ACTIONS(5359),
    [anon_sym__Alignas] = ACTIONS(5359),
    [sym_primitive_type] = ACTIONS(5359),
    [anon_sym_enum] = ACTIONS(5359),
    [anon_sym_struct] = ACTIONS(5359),
    [anon_sym_union] = ACTIONS(5359),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5359),
    [anon_sym___typeof] = ACTIONS(5359),
    [anon_sym_typeof] = ACTIONS(5359),
    [aux_sym_preproc_undef_token1] = ACTIONS(5359),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5359),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5359),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5359),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_API_AVAILABLE] = ACTIONS(5359),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_API_DEPRECATED] = ACTIONS(5359),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym___deprecated_msg] = ACTIONS(5359),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5359),
    [anon_sym_ATend] = ACTIONS(5361),
    [anon_sym_ATsynthesize] = ACTIONS(5361),
    [anon_sym_ATdynamic] = ACTIONS(5361),
    [anon_sym_BOOL] = ACTIONS(5359),
    [anon_sym_IMP] = ACTIONS(5359),
    [anon_sym_SEL] = ACTIONS(5359),
    [anon_sym_Class] = ACTIONS(5359),
    [anon_sym_id] = ACTIONS(5359),
  },
  [STATE(2272)] = {
    [sym_identifier] = ACTIONS(5569),
    [aux_sym_preproc_def_token1] = ACTIONS(5569),
    [aux_sym_preproc_if_token1] = ACTIONS(5569),
    [aux_sym_preproc_if_token2] = ACTIONS(5569),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5569),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5569),
    [aux_sym_preproc_else_token1] = ACTIONS(5569),
    [aux_sym_preproc_elif_token1] = ACTIONS(5569),
    [sym_preproc_directive] = ACTIONS(5569),
    [anon_sym_DASH] = ACTIONS(5571),
    [anon_sym_PLUS] = ACTIONS(5571),
    [anon_sym___extension__] = ACTIONS(5569),
    [anon_sym_typedef] = ACTIONS(5569),
    [anon_sym_extern] = ACTIONS(5569),
    [anon_sym___attribute__] = ACTIONS(5569),
    [anon_sym___attribute] = ACTIONS(5569),
    [anon_sym_noreturn] = ACTIONS(5569),
    [anon_sym_LBRACK] = ACTIONS(5571),
    [anon_sym___declspec] = ACTIONS(5569),
    [anon_sym___cdecl] = ACTIONS(5569),
    [anon_sym___clrcall] = ACTIONS(5569),
    [anon_sym___stdcall] = ACTIONS(5569),
    [anon_sym___fastcall] = ACTIONS(5569),
    [anon_sym___thiscall] = ACTIONS(5569),
    [anon_sym___vectorcall] = ACTIONS(5569),
    [anon_sym_signed] = ACTIONS(5569),
    [anon_sym_unsigned] = ACTIONS(5569),
    [anon_sym_long] = ACTIONS(5569),
    [anon_sym_short] = ACTIONS(5569),
    [anon_sym_static] = ACTIONS(5569),
    [anon_sym_auto] = ACTIONS(5569),
    [anon_sym_register] = ACTIONS(5569),
    [anon_sym_inline] = ACTIONS(5569),
    [anon_sym___inline] = ACTIONS(5569),
    [anon_sym___inline__] = ACTIONS(5569),
    [anon_sym___forceinline] = ACTIONS(5569),
    [anon_sym_thread_local] = ACTIONS(5569),
    [anon_sym___thread] = ACTIONS(5569),
    [anon_sym_CG_EXTERN] = ACTIONS(5569),
    [anon_sym_CG_INLINE] = ACTIONS(5569),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5569),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5569),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5569),
    [anon_sym_IBOutlet] = ACTIONS(5569),
    [anon_sym_IBInspectable] = ACTIONS(5569),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5569),
    [anon_sym_NS_INLINE] = ACTIONS(5569),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5569),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5569),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5569),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5569),
    [anon_sym_const] = ACTIONS(5569),
    [anon_sym_constexpr] = ACTIONS(5569),
    [anon_sym_volatile] = ACTIONS(5569),
    [anon_sym_restrict] = ACTIONS(5569),
    [anon_sym___restrict__] = ACTIONS(5569),
    [anon_sym__Atomic] = ACTIONS(5569),
    [anon_sym__Noreturn] = ACTIONS(5569),
    [anon_sym__Nonnull] = ACTIONS(5569),
    [anon_sym_nullable] = ACTIONS(5569),
    [anon_sym__Complex] = ACTIONS(5569),
    [anon_sym__Nullable] = ACTIONS(5569),
    [anon_sym__Nullable_result] = ACTIONS(5569),
    [anon_sym__Null_unspecified] = ACTIONS(5569),
    [anon_sym___autoreleasing] = ACTIONS(5569),
    [anon_sym___block] = ACTIONS(5569),
    [anon_sym___bridge] = ACTIONS(5569),
    [anon_sym___bridge_retained] = ACTIONS(5569),
    [anon_sym___bridge_transfer] = ACTIONS(5569),
    [anon_sym___complex] = ACTIONS(5569),
    [anon_sym___const] = ACTIONS(5569),
    [anon_sym___imag] = ACTIONS(5569),
    [anon_sym___kindof] = ACTIONS(5569),
    [anon_sym___nonnull] = ACTIONS(5569),
    [anon_sym___nullable] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5569),
    [anon_sym___real] = ACTIONS(5569),
    [anon_sym___strong] = ACTIONS(5569),
    [anon_sym___unsafe_unretained] = ACTIONS(5569),
    [anon_sym___unused] = ACTIONS(5569),
    [anon_sym___weak] = ACTIONS(5569),
    [anon_sym_alignas] = ACTIONS(5569),
    [anon_sym__Alignas] = ACTIONS(5569),
    [sym_primitive_type] = ACTIONS(5569),
    [anon_sym_enum] = ACTIONS(5569),
    [anon_sym_struct] = ACTIONS(5569),
    [anon_sym_union] = ACTIONS(5569),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5569),
    [anon_sym___typeof] = ACTIONS(5569),
    [anon_sym_typeof] = ACTIONS(5569),
    [aux_sym_preproc_undef_token1] = ACTIONS(5569),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5569),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5569),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5569),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5569),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5569),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5569),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5569),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5569),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5569),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_API_AVAILABLE] = ACTIONS(5569),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_API_DEPRECATED] = ACTIONS(5569),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5569),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5569),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5569),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5569),
    [anon_sym___deprecated_msg] = ACTIONS(5569),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5569),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5569),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5569),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5569),
    [anon_sym_ATproperty] = ACTIONS(5571),
    [anon_sym_BOOL] = ACTIONS(5569),
    [anon_sym_IMP] = ACTIONS(5569),
    [anon_sym_SEL] = ACTIONS(5569),
    [anon_sym_Class] = ACTIONS(5569),
    [anon_sym_id] = ACTIONS(5569),
  },
  [STATE(2273)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token2] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_DASH] = ACTIONS(2422),
    [anon_sym_PLUS] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATproperty] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(2274)] = {
    [sym_identifier] = ACTIONS(5573),
    [aux_sym_preproc_def_token1] = ACTIONS(5573),
    [aux_sym_preproc_if_token1] = ACTIONS(5573),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5573),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5573),
    [sym_preproc_directive] = ACTIONS(5573),
    [anon_sym_DASH] = ACTIONS(5575),
    [anon_sym_PLUS] = ACTIONS(5575),
    [anon_sym___extension__] = ACTIONS(5573),
    [anon_sym_typedef] = ACTIONS(5573),
    [anon_sym_extern] = ACTIONS(5573),
    [anon_sym___attribute__] = ACTIONS(5573),
    [anon_sym___attribute] = ACTIONS(5573),
    [anon_sym_noreturn] = ACTIONS(5573),
    [anon_sym_LBRACK] = ACTIONS(5575),
    [anon_sym___declspec] = ACTIONS(5573),
    [anon_sym___cdecl] = ACTIONS(5573),
    [anon_sym___clrcall] = ACTIONS(5573),
    [anon_sym___stdcall] = ACTIONS(5573),
    [anon_sym___fastcall] = ACTIONS(5573),
    [anon_sym___thiscall] = ACTIONS(5573),
    [anon_sym___vectorcall] = ACTIONS(5573),
    [anon_sym_signed] = ACTIONS(5573),
    [anon_sym_unsigned] = ACTIONS(5573),
    [anon_sym_long] = ACTIONS(5573),
    [anon_sym_short] = ACTIONS(5573),
    [anon_sym_static] = ACTIONS(5573),
    [anon_sym_auto] = ACTIONS(5573),
    [anon_sym_register] = ACTIONS(5573),
    [anon_sym_inline] = ACTIONS(5573),
    [anon_sym___inline] = ACTIONS(5573),
    [anon_sym___inline__] = ACTIONS(5573),
    [anon_sym___forceinline] = ACTIONS(5573),
    [anon_sym_thread_local] = ACTIONS(5573),
    [anon_sym___thread] = ACTIONS(5573),
    [anon_sym_CG_EXTERN] = ACTIONS(5573),
    [anon_sym_CG_INLINE] = ACTIONS(5573),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5573),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5573),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5573),
    [anon_sym_IBOutlet] = ACTIONS(5573),
    [anon_sym_IBInspectable] = ACTIONS(5573),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5573),
    [anon_sym_NS_INLINE] = ACTIONS(5573),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5573),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5573),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5573),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5573),
    [anon_sym_const] = ACTIONS(5573),
    [anon_sym_constexpr] = ACTIONS(5573),
    [anon_sym_volatile] = ACTIONS(5573),
    [anon_sym_restrict] = ACTIONS(5573),
    [anon_sym___restrict__] = ACTIONS(5573),
    [anon_sym__Atomic] = ACTIONS(5573),
    [anon_sym__Noreturn] = ACTIONS(5573),
    [anon_sym__Nonnull] = ACTIONS(5573),
    [anon_sym_nullable] = ACTIONS(5573),
    [anon_sym__Complex] = ACTIONS(5573),
    [anon_sym__Nullable] = ACTIONS(5573),
    [anon_sym__Nullable_result] = ACTIONS(5573),
    [anon_sym__Null_unspecified] = ACTIONS(5573),
    [anon_sym___autoreleasing] = ACTIONS(5573),
    [anon_sym___block] = ACTIONS(5573),
    [anon_sym___bridge] = ACTIONS(5573),
    [anon_sym___bridge_retained] = ACTIONS(5573),
    [anon_sym___bridge_transfer] = ACTIONS(5573),
    [anon_sym___complex] = ACTIONS(5573),
    [anon_sym___const] = ACTIONS(5573),
    [anon_sym___imag] = ACTIONS(5573),
    [anon_sym___kindof] = ACTIONS(5573),
    [anon_sym___nonnull] = ACTIONS(5573),
    [anon_sym___nullable] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5573),
    [anon_sym___real] = ACTIONS(5573),
    [anon_sym___strong] = ACTIONS(5573),
    [anon_sym___unsafe_unretained] = ACTIONS(5573),
    [anon_sym___unused] = ACTIONS(5573),
    [anon_sym___weak] = ACTIONS(5573),
    [anon_sym_alignas] = ACTIONS(5573),
    [anon_sym__Alignas] = ACTIONS(5573),
    [sym_primitive_type] = ACTIONS(5573),
    [anon_sym_enum] = ACTIONS(5573),
    [anon_sym_struct] = ACTIONS(5573),
    [anon_sym_union] = ACTIONS(5573),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5573),
    [anon_sym___typeof] = ACTIONS(5573),
    [anon_sym_typeof] = ACTIONS(5573),
    [aux_sym_preproc_undef_token1] = ACTIONS(5573),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5573),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5573),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5573),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5573),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5573),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5573),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5573),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5573),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5573),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_API_AVAILABLE] = ACTIONS(5573),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_API_DEPRECATED] = ACTIONS(5573),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5573),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5573),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5573),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5573),
    [anon_sym___deprecated_msg] = ACTIONS(5573),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5573),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5573),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5573),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5573),
    [anon_sym_ATend] = ACTIONS(5575),
    [anon_sym_AToptional] = ACTIONS(5575),
    [anon_sym_ATrequired] = ACTIONS(5575),
    [anon_sym_ATproperty] = ACTIONS(5575),
    [anon_sym_BOOL] = ACTIONS(5573),
    [anon_sym_IMP] = ACTIONS(5573),
    [anon_sym_SEL] = ACTIONS(5573),
    [anon_sym_Class] = ACTIONS(5573),
    [anon_sym_id] = ACTIONS(5573),
  },
  [STATE(2275)] = {
    [sym_identifier] = ACTIONS(5577),
    [aux_sym_preproc_def_token1] = ACTIONS(5577),
    [aux_sym_preproc_if_token1] = ACTIONS(5577),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5577),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5577),
    [sym_preproc_directive] = ACTIONS(5577),
    [anon_sym_DASH] = ACTIONS(5579),
    [anon_sym_PLUS] = ACTIONS(5579),
    [anon_sym___extension__] = ACTIONS(5577),
    [anon_sym_typedef] = ACTIONS(5577),
    [anon_sym_extern] = ACTIONS(5577),
    [anon_sym___attribute__] = ACTIONS(5577),
    [anon_sym___attribute] = ACTIONS(5577),
    [anon_sym_noreturn] = ACTIONS(5577),
    [anon_sym_LBRACK] = ACTIONS(5579),
    [anon_sym___declspec] = ACTIONS(5577),
    [anon_sym___cdecl] = ACTIONS(5577),
    [anon_sym___clrcall] = ACTIONS(5577),
    [anon_sym___stdcall] = ACTIONS(5577),
    [anon_sym___fastcall] = ACTIONS(5577),
    [anon_sym___thiscall] = ACTIONS(5577),
    [anon_sym___vectorcall] = ACTIONS(5577),
    [anon_sym_signed] = ACTIONS(5577),
    [anon_sym_unsigned] = ACTIONS(5577),
    [anon_sym_long] = ACTIONS(5577),
    [anon_sym_short] = ACTIONS(5577),
    [anon_sym_static] = ACTIONS(5577),
    [anon_sym_auto] = ACTIONS(5577),
    [anon_sym_register] = ACTIONS(5577),
    [anon_sym_inline] = ACTIONS(5577),
    [anon_sym___inline] = ACTIONS(5577),
    [anon_sym___inline__] = ACTIONS(5577),
    [anon_sym___forceinline] = ACTIONS(5577),
    [anon_sym_thread_local] = ACTIONS(5577),
    [anon_sym___thread] = ACTIONS(5577),
    [anon_sym_CG_EXTERN] = ACTIONS(5577),
    [anon_sym_CG_INLINE] = ACTIONS(5577),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5577),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5577),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5577),
    [anon_sym_IBOutlet] = ACTIONS(5577),
    [anon_sym_IBInspectable] = ACTIONS(5577),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5577),
    [anon_sym_NS_INLINE] = ACTIONS(5577),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5577),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5577),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5577),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5577),
    [anon_sym_const] = ACTIONS(5577),
    [anon_sym_constexpr] = ACTIONS(5577),
    [anon_sym_volatile] = ACTIONS(5577),
    [anon_sym_restrict] = ACTIONS(5577),
    [anon_sym___restrict__] = ACTIONS(5577),
    [anon_sym__Atomic] = ACTIONS(5577),
    [anon_sym__Noreturn] = ACTIONS(5577),
    [anon_sym__Nonnull] = ACTIONS(5577),
    [anon_sym_nullable] = ACTIONS(5577),
    [anon_sym__Complex] = ACTIONS(5577),
    [anon_sym__Nullable] = ACTIONS(5577),
    [anon_sym__Nullable_result] = ACTIONS(5577),
    [anon_sym__Null_unspecified] = ACTIONS(5577),
    [anon_sym___autoreleasing] = ACTIONS(5577),
    [anon_sym___block] = ACTIONS(5577),
    [anon_sym___bridge] = ACTIONS(5577),
    [anon_sym___bridge_retained] = ACTIONS(5577),
    [anon_sym___bridge_transfer] = ACTIONS(5577),
    [anon_sym___complex] = ACTIONS(5577),
    [anon_sym___const] = ACTIONS(5577),
    [anon_sym___imag] = ACTIONS(5577),
    [anon_sym___kindof] = ACTIONS(5577),
    [anon_sym___nonnull] = ACTIONS(5577),
    [anon_sym___nullable] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5577),
    [anon_sym___real] = ACTIONS(5577),
    [anon_sym___strong] = ACTIONS(5577),
    [anon_sym___unsafe_unretained] = ACTIONS(5577),
    [anon_sym___unused] = ACTIONS(5577),
    [anon_sym___weak] = ACTIONS(5577),
    [anon_sym_alignas] = ACTIONS(5577),
    [anon_sym__Alignas] = ACTIONS(5577),
    [sym_primitive_type] = ACTIONS(5577),
    [anon_sym_enum] = ACTIONS(5577),
    [anon_sym_struct] = ACTIONS(5577),
    [anon_sym_union] = ACTIONS(5577),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5577),
    [anon_sym___typeof] = ACTIONS(5577),
    [anon_sym_typeof] = ACTIONS(5577),
    [aux_sym_preproc_undef_token1] = ACTIONS(5577),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5577),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5577),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5577),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5577),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5577),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5577),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5577),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5577),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5577),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_API_AVAILABLE] = ACTIONS(5577),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_API_DEPRECATED] = ACTIONS(5577),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5577),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5577),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5577),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5577),
    [anon_sym___deprecated_msg] = ACTIONS(5577),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5577),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5577),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5577),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5577),
    [anon_sym_ATend] = ACTIONS(5579),
    [anon_sym_AToptional] = ACTIONS(5579),
    [anon_sym_ATrequired] = ACTIONS(5579),
    [anon_sym_ATproperty] = ACTIONS(5579),
    [anon_sym_BOOL] = ACTIONS(5577),
    [anon_sym_IMP] = ACTIONS(5577),
    [anon_sym_SEL] = ACTIONS(5577),
    [anon_sym_Class] = ACTIONS(5577),
    [anon_sym_id] = ACTIONS(5577),
  },
  [STATE(2276)] = {
    [sym_identifier] = ACTIONS(5581),
    [aux_sym_preproc_def_token1] = ACTIONS(5581),
    [aux_sym_preproc_if_token1] = ACTIONS(5581),
    [aux_sym_preproc_if_token2] = ACTIONS(5581),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5581),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5581),
    [aux_sym_preproc_else_token1] = ACTIONS(5581),
    [aux_sym_preproc_elif_token1] = ACTIONS(5581),
    [sym_preproc_directive] = ACTIONS(5581),
    [anon_sym_DASH] = ACTIONS(5583),
    [anon_sym_PLUS] = ACTIONS(5583),
    [anon_sym___extension__] = ACTIONS(5581),
    [anon_sym_typedef] = ACTIONS(5581),
    [anon_sym_extern] = ACTIONS(5581),
    [anon_sym___attribute__] = ACTIONS(5581),
    [anon_sym___attribute] = ACTIONS(5581),
    [anon_sym_noreturn] = ACTIONS(5581),
    [anon_sym_LBRACK] = ACTIONS(5583),
    [anon_sym___declspec] = ACTIONS(5581),
    [anon_sym___cdecl] = ACTIONS(5581),
    [anon_sym___clrcall] = ACTIONS(5581),
    [anon_sym___stdcall] = ACTIONS(5581),
    [anon_sym___fastcall] = ACTIONS(5581),
    [anon_sym___thiscall] = ACTIONS(5581),
    [anon_sym___vectorcall] = ACTIONS(5581),
    [anon_sym_signed] = ACTIONS(5581),
    [anon_sym_unsigned] = ACTIONS(5581),
    [anon_sym_long] = ACTIONS(5581),
    [anon_sym_short] = ACTIONS(5581),
    [anon_sym_static] = ACTIONS(5581),
    [anon_sym_auto] = ACTIONS(5581),
    [anon_sym_register] = ACTIONS(5581),
    [anon_sym_inline] = ACTIONS(5581),
    [anon_sym___inline] = ACTIONS(5581),
    [anon_sym___inline__] = ACTIONS(5581),
    [anon_sym___forceinline] = ACTIONS(5581),
    [anon_sym_thread_local] = ACTIONS(5581),
    [anon_sym___thread] = ACTIONS(5581),
    [anon_sym_CG_EXTERN] = ACTIONS(5581),
    [anon_sym_CG_INLINE] = ACTIONS(5581),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5581),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5581),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5581),
    [anon_sym_IBOutlet] = ACTIONS(5581),
    [anon_sym_IBInspectable] = ACTIONS(5581),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5581),
    [anon_sym_NS_INLINE] = ACTIONS(5581),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5581),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5581),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5581),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5581),
    [anon_sym_const] = ACTIONS(5581),
    [anon_sym_constexpr] = ACTIONS(5581),
    [anon_sym_volatile] = ACTIONS(5581),
    [anon_sym_restrict] = ACTIONS(5581),
    [anon_sym___restrict__] = ACTIONS(5581),
    [anon_sym__Atomic] = ACTIONS(5581),
    [anon_sym__Noreturn] = ACTIONS(5581),
    [anon_sym__Nonnull] = ACTIONS(5581),
    [anon_sym_nullable] = ACTIONS(5581),
    [anon_sym__Complex] = ACTIONS(5581),
    [anon_sym__Nullable] = ACTIONS(5581),
    [anon_sym__Nullable_result] = ACTIONS(5581),
    [anon_sym__Null_unspecified] = ACTIONS(5581),
    [anon_sym___autoreleasing] = ACTIONS(5581),
    [anon_sym___block] = ACTIONS(5581),
    [anon_sym___bridge] = ACTIONS(5581),
    [anon_sym___bridge_retained] = ACTIONS(5581),
    [anon_sym___bridge_transfer] = ACTIONS(5581),
    [anon_sym___complex] = ACTIONS(5581),
    [anon_sym___const] = ACTIONS(5581),
    [anon_sym___imag] = ACTIONS(5581),
    [anon_sym___kindof] = ACTIONS(5581),
    [anon_sym___nonnull] = ACTIONS(5581),
    [anon_sym___nullable] = ACTIONS(5581),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5581),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5581),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5581),
    [anon_sym___real] = ACTIONS(5581),
    [anon_sym___strong] = ACTIONS(5581),
    [anon_sym___unsafe_unretained] = ACTIONS(5581),
    [anon_sym___unused] = ACTIONS(5581),
    [anon_sym___weak] = ACTIONS(5581),
    [anon_sym_alignas] = ACTIONS(5581),
    [anon_sym__Alignas] = ACTIONS(5581),
    [sym_primitive_type] = ACTIONS(5581),
    [anon_sym_enum] = ACTIONS(5581),
    [anon_sym_struct] = ACTIONS(5581),
    [anon_sym_union] = ACTIONS(5581),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5581),
    [anon_sym___typeof] = ACTIONS(5581),
    [anon_sym_typeof] = ACTIONS(5581),
    [aux_sym_preproc_undef_token1] = ACTIONS(5581),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5581),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5581),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5581),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5581),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5581),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5581),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5581),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5581),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5581),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_API_AVAILABLE] = ACTIONS(5581),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_API_DEPRECATED] = ACTIONS(5581),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5581),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5581),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5581),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5581),
    [anon_sym___deprecated_msg] = ACTIONS(5581),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5581),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5581),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5581),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5581),
    [anon_sym_ATproperty] = ACTIONS(5583),
    [anon_sym_BOOL] = ACTIONS(5581),
    [anon_sym_IMP] = ACTIONS(5581),
    [anon_sym_SEL] = ACTIONS(5581),
    [anon_sym_Class] = ACTIONS(5581),
    [anon_sym_id] = ACTIONS(5581),
  },
  [STATE(2277)] = {
    [sym_identifier] = ACTIONS(5369),
    [aux_sym_preproc_def_token1] = ACTIONS(5369),
    [aux_sym_preproc_if_token1] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5369),
    [sym_preproc_directive] = ACTIONS(5369),
    [anon_sym_DASH] = ACTIONS(5371),
    [anon_sym_PLUS] = ACTIONS(5371),
    [anon_sym_SEMI] = ACTIONS(5585),
    [anon_sym___extension__] = ACTIONS(5369),
    [anon_sym_typedef] = ACTIONS(5369),
    [anon_sym_extern] = ACTIONS(5369),
    [anon_sym___attribute__] = ACTIONS(5369),
    [anon_sym___attribute] = ACTIONS(5369),
    [anon_sym_noreturn] = ACTIONS(5369),
    [anon_sym_LBRACK] = ACTIONS(5371),
    [anon_sym___declspec] = ACTIONS(5369),
    [anon_sym___cdecl] = ACTIONS(5369),
    [anon_sym___clrcall] = ACTIONS(5369),
    [anon_sym___stdcall] = ACTIONS(5369),
    [anon_sym___fastcall] = ACTIONS(5369),
    [anon_sym___thiscall] = ACTIONS(5369),
    [anon_sym___vectorcall] = ACTIONS(5369),
    [anon_sym_signed] = ACTIONS(5369),
    [anon_sym_unsigned] = ACTIONS(5369),
    [anon_sym_long] = ACTIONS(5369),
    [anon_sym_short] = ACTIONS(5369),
    [anon_sym_static] = ACTIONS(5369),
    [anon_sym_auto] = ACTIONS(5369),
    [anon_sym_register] = ACTIONS(5369),
    [anon_sym_inline] = ACTIONS(5369),
    [anon_sym___inline] = ACTIONS(5369),
    [anon_sym___inline__] = ACTIONS(5369),
    [anon_sym___forceinline] = ACTIONS(5369),
    [anon_sym_thread_local] = ACTIONS(5369),
    [anon_sym___thread] = ACTIONS(5369),
    [anon_sym_CG_EXTERN] = ACTIONS(5369),
    [anon_sym_CG_INLINE] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5369),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5369),
    [anon_sym_IBOutlet] = ACTIONS(5369),
    [anon_sym_IBInspectable] = ACTIONS(5369),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5369),
    [anon_sym_NS_INLINE] = ACTIONS(5369),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5369),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5369),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5369),
    [anon_sym_const] = ACTIONS(5369),
    [anon_sym_constexpr] = ACTIONS(5369),
    [anon_sym_volatile] = ACTIONS(5369),
    [anon_sym_restrict] = ACTIONS(5369),
    [anon_sym___restrict__] = ACTIONS(5369),
    [anon_sym__Atomic] = ACTIONS(5369),
    [anon_sym__Noreturn] = ACTIONS(5369),
    [anon_sym__Nonnull] = ACTIONS(5369),
    [anon_sym_nullable] = ACTIONS(5369),
    [anon_sym__Complex] = ACTIONS(5369),
    [anon_sym__Nullable] = ACTIONS(5369),
    [anon_sym__Nullable_result] = ACTIONS(5369),
    [anon_sym__Null_unspecified] = ACTIONS(5369),
    [anon_sym___autoreleasing] = ACTIONS(5369),
    [anon_sym___block] = ACTIONS(5369),
    [anon_sym___bridge] = ACTIONS(5369),
    [anon_sym___bridge_retained] = ACTIONS(5369),
    [anon_sym___bridge_transfer] = ACTIONS(5369),
    [anon_sym___complex] = ACTIONS(5369),
    [anon_sym___const] = ACTIONS(5369),
    [anon_sym___imag] = ACTIONS(5369),
    [anon_sym___kindof] = ACTIONS(5369),
    [anon_sym___nonnull] = ACTIONS(5369),
    [anon_sym___nullable] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5369),
    [anon_sym___real] = ACTIONS(5369),
    [anon_sym___strong] = ACTIONS(5369),
    [anon_sym___unsafe_unretained] = ACTIONS(5369),
    [anon_sym___unused] = ACTIONS(5369),
    [anon_sym___weak] = ACTIONS(5369),
    [anon_sym_alignas] = ACTIONS(5369),
    [anon_sym__Alignas] = ACTIONS(5369),
    [sym_primitive_type] = ACTIONS(5369),
    [anon_sym_enum] = ACTIONS(5369),
    [anon_sym_struct] = ACTIONS(5369),
    [anon_sym_union] = ACTIONS(5369),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5369),
    [anon_sym___typeof] = ACTIONS(5369),
    [anon_sym_typeof] = ACTIONS(5369),
    [aux_sym_preproc_undef_token1] = ACTIONS(5369),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5369),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5369),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5369),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_API_AVAILABLE] = ACTIONS(5369),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_API_DEPRECATED] = ACTIONS(5369),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym___deprecated_msg] = ACTIONS(5369),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5369),
    [anon_sym_ATend] = ACTIONS(5371),
    [anon_sym_ATsynthesize] = ACTIONS(5371),
    [anon_sym_ATdynamic] = ACTIONS(5371),
    [anon_sym_BOOL] = ACTIONS(5369),
    [anon_sym_IMP] = ACTIONS(5369),
    [anon_sym_SEL] = ACTIONS(5369),
    [anon_sym_Class] = ACTIONS(5369),
    [anon_sym_id] = ACTIONS(5369),
  },
  [STATE(2278)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token2] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_DASH] = ACTIONS(2426),
    [anon_sym_PLUS] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATproperty] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(2279)] = {
    [sym_identifier] = ACTIONS(2416),
    [aux_sym_preproc_def_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token1] = ACTIONS(2416),
    [aux_sym_preproc_if_token2] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2416),
    [sym_preproc_directive] = ACTIONS(2416),
    [anon_sym_DASH] = ACTIONS(2418),
    [anon_sym_PLUS] = ACTIONS(2418),
    [anon_sym___extension__] = ACTIONS(2416),
    [anon_sym_typedef] = ACTIONS(2416),
    [anon_sym_extern] = ACTIONS(2416),
    [anon_sym___attribute__] = ACTIONS(2416),
    [anon_sym___attribute] = ACTIONS(2416),
    [anon_sym_noreturn] = ACTIONS(2416),
    [anon_sym_LBRACK] = ACTIONS(2418),
    [anon_sym___declspec] = ACTIONS(2416),
    [anon_sym___cdecl] = ACTIONS(2416),
    [anon_sym___clrcall] = ACTIONS(2416),
    [anon_sym___stdcall] = ACTIONS(2416),
    [anon_sym___fastcall] = ACTIONS(2416),
    [anon_sym___thiscall] = ACTIONS(2416),
    [anon_sym___vectorcall] = ACTIONS(2416),
    [anon_sym_signed] = ACTIONS(2416),
    [anon_sym_unsigned] = ACTIONS(2416),
    [anon_sym_long] = ACTIONS(2416),
    [anon_sym_short] = ACTIONS(2416),
    [anon_sym_static] = ACTIONS(2416),
    [anon_sym_auto] = ACTIONS(2416),
    [anon_sym_register] = ACTIONS(2416),
    [anon_sym_inline] = ACTIONS(2416),
    [anon_sym___inline] = ACTIONS(2416),
    [anon_sym___inline__] = ACTIONS(2416),
    [anon_sym___forceinline] = ACTIONS(2416),
    [anon_sym_thread_local] = ACTIONS(2416),
    [anon_sym___thread] = ACTIONS(2416),
    [anon_sym_CG_EXTERN] = ACTIONS(2416),
    [anon_sym_CG_INLINE] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2416),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2416),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2416),
    [anon_sym_IBOutlet] = ACTIONS(2416),
    [anon_sym_IBInspectable] = ACTIONS(2416),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2416),
    [anon_sym_NS_INLINE] = ACTIONS(2416),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2416),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2416),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2416),
    [anon_sym_const] = ACTIONS(2416),
    [anon_sym_constexpr] = ACTIONS(2416),
    [anon_sym_volatile] = ACTIONS(2416),
    [anon_sym_restrict] = ACTIONS(2416),
    [anon_sym___restrict__] = ACTIONS(2416),
    [anon_sym__Atomic] = ACTIONS(2416),
    [anon_sym__Noreturn] = ACTIONS(2416),
    [anon_sym__Nonnull] = ACTIONS(2416),
    [anon_sym_nullable] = ACTIONS(2416),
    [anon_sym__Complex] = ACTIONS(2416),
    [anon_sym__Nullable] = ACTIONS(2416),
    [anon_sym__Nullable_result] = ACTIONS(2416),
    [anon_sym__Null_unspecified] = ACTIONS(2416),
    [anon_sym___autoreleasing] = ACTIONS(2416),
    [anon_sym___block] = ACTIONS(2416),
    [anon_sym___bridge] = ACTIONS(2416),
    [anon_sym___bridge_retained] = ACTIONS(2416),
    [anon_sym___bridge_transfer] = ACTIONS(2416),
    [anon_sym___complex] = ACTIONS(2416),
    [anon_sym___const] = ACTIONS(2416),
    [anon_sym___imag] = ACTIONS(2416),
    [anon_sym___kindof] = ACTIONS(2416),
    [anon_sym___nonnull] = ACTIONS(2416),
    [anon_sym___nullable] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2416),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2416),
    [anon_sym___real] = ACTIONS(2416),
    [anon_sym___strong] = ACTIONS(2416),
    [anon_sym___unsafe_unretained] = ACTIONS(2416),
    [anon_sym___unused] = ACTIONS(2416),
    [anon_sym___weak] = ACTIONS(2416),
    [anon_sym_alignas] = ACTIONS(2416),
    [anon_sym__Alignas] = ACTIONS(2416),
    [sym_primitive_type] = ACTIONS(2416),
    [anon_sym_enum] = ACTIONS(2416),
    [anon_sym_struct] = ACTIONS(2416),
    [anon_sym_union] = ACTIONS(2416),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2416),
    [anon_sym___typeof] = ACTIONS(2416),
    [anon_sym_typeof] = ACTIONS(2416),
    [aux_sym_preproc_undef_token1] = ACTIONS(2416),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2416),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2416),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2416),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2416),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2416),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_API_AVAILABLE] = ACTIONS(2416),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_API_DEPRECATED] = ACTIONS(2416),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2416),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2416),
    [anon_sym___deprecated_msg] = ACTIONS(2416),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2416),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2416),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2416),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2416),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2416),
    [anon_sym_ATsynthesize] = ACTIONS(2418),
    [anon_sym_ATdynamic] = ACTIONS(2418),
    [anon_sym_ATproperty] = ACTIONS(2418),
    [anon_sym_BOOL] = ACTIONS(2416),
    [anon_sym_IMP] = ACTIONS(2416),
    [anon_sym_SEL] = ACTIONS(2416),
    [anon_sym_Class] = ACTIONS(2416),
    [anon_sym_id] = ACTIONS(2416),
  },
  [STATE(2280)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2321),
    [sym_identifier] = ACTIONS(5458),
    [aux_sym_preproc_def_token1] = ACTIONS(5458),
    [aux_sym_preproc_if_token1] = ACTIONS(5458),
    [aux_sym_preproc_if_token2] = ACTIONS(5458),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5458),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5458),
    [sym_preproc_directive] = ACTIONS(5458),
    [anon_sym_DASH] = ACTIONS(5460),
    [anon_sym_PLUS] = ACTIONS(5460),
    [anon_sym_SEMI] = ACTIONS(5565),
    [anon_sym___extension__] = ACTIONS(5458),
    [anon_sym_typedef] = ACTIONS(5458),
    [anon_sym_extern] = ACTIONS(5458),
    [anon_sym___attribute__] = ACTIONS(5458),
    [anon_sym___attribute] = ACTIONS(5458),
    [anon_sym_noreturn] = ACTIONS(5458),
    [anon_sym_LBRACK] = ACTIONS(5460),
    [anon_sym___declspec] = ACTIONS(5458),
    [anon_sym___cdecl] = ACTIONS(5458),
    [anon_sym___clrcall] = ACTIONS(5458),
    [anon_sym___stdcall] = ACTIONS(5458),
    [anon_sym___fastcall] = ACTIONS(5458),
    [anon_sym___thiscall] = ACTIONS(5458),
    [anon_sym___vectorcall] = ACTIONS(5458),
    [anon_sym_signed] = ACTIONS(5458),
    [anon_sym_unsigned] = ACTIONS(5458),
    [anon_sym_long] = ACTIONS(5458),
    [anon_sym_short] = ACTIONS(5458),
    [anon_sym_static] = ACTIONS(5458),
    [anon_sym_auto] = ACTIONS(5458),
    [anon_sym_register] = ACTIONS(5458),
    [anon_sym_inline] = ACTIONS(5458),
    [anon_sym___inline] = ACTIONS(5458),
    [anon_sym___inline__] = ACTIONS(5458),
    [anon_sym___forceinline] = ACTIONS(5458),
    [anon_sym_thread_local] = ACTIONS(5458),
    [anon_sym___thread] = ACTIONS(5458),
    [anon_sym_CG_EXTERN] = ACTIONS(5458),
    [anon_sym_CG_INLINE] = ACTIONS(5458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5458),
    [anon_sym_IBOutlet] = ACTIONS(5458),
    [anon_sym_IBInspectable] = ACTIONS(5458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5458),
    [anon_sym_NS_INLINE] = ACTIONS(5458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5458),
    [anon_sym_const] = ACTIONS(5458),
    [anon_sym_constexpr] = ACTIONS(5458),
    [anon_sym_volatile] = ACTIONS(5458),
    [anon_sym_restrict] = ACTIONS(5458),
    [anon_sym___restrict__] = ACTIONS(5458),
    [anon_sym__Atomic] = ACTIONS(5458),
    [anon_sym__Noreturn] = ACTIONS(5458),
    [anon_sym__Nonnull] = ACTIONS(5458),
    [anon_sym_nullable] = ACTIONS(5458),
    [anon_sym__Complex] = ACTIONS(5458),
    [anon_sym__Nullable] = ACTIONS(5458),
    [anon_sym__Nullable_result] = ACTIONS(5458),
    [anon_sym__Null_unspecified] = ACTIONS(5458),
    [anon_sym___autoreleasing] = ACTIONS(5458),
    [anon_sym___block] = ACTIONS(5458),
    [anon_sym___bridge] = ACTIONS(5458),
    [anon_sym___bridge_retained] = ACTIONS(5458),
    [anon_sym___bridge_transfer] = ACTIONS(5458),
    [anon_sym___complex] = ACTIONS(5458),
    [anon_sym___const] = ACTIONS(5458),
    [anon_sym___imag] = ACTIONS(5458),
    [anon_sym___kindof] = ACTIONS(5458),
    [anon_sym___nonnull] = ACTIONS(5458),
    [anon_sym___nullable] = ACTIONS(5458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5458),
    [anon_sym___real] = ACTIONS(5458),
    [anon_sym___strong] = ACTIONS(5458),
    [anon_sym___unsafe_unretained] = ACTIONS(5458),
    [anon_sym___unused] = ACTIONS(5458),
    [anon_sym___weak] = ACTIONS(5458),
    [anon_sym_alignas] = ACTIONS(5458),
    [anon_sym__Alignas] = ACTIONS(5458),
    [sym_primitive_type] = ACTIONS(5458),
    [anon_sym_enum] = ACTIONS(5458),
    [anon_sym_struct] = ACTIONS(5458),
    [anon_sym_union] = ACTIONS(5458),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5458),
    [anon_sym___typeof] = ACTIONS(5458),
    [anon_sym_typeof] = ACTIONS(5458),
    [aux_sym_preproc_undef_token1] = ACTIONS(5458),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_API_AVAILABLE] = ACTIONS(5458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_API_DEPRECATED] = ACTIONS(5458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5458),
    [anon_sym___deprecated_msg] = ACTIONS(5458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5458),
    [anon_sym_ATproperty] = ACTIONS(5460),
    [anon_sym_BOOL] = ACTIONS(5458),
    [anon_sym_IMP] = ACTIONS(5458),
    [anon_sym_SEL] = ACTIONS(5458),
    [anon_sym_Class] = ACTIONS(5458),
    [anon_sym_id] = ACTIONS(5458),
  },
  [STATE(2281)] = {
    [sym_identifier] = ACTIONS(5466),
    [aux_sym_preproc_def_token1] = ACTIONS(5466),
    [aux_sym_preproc_if_token1] = ACTIONS(5466),
    [aux_sym_preproc_if_token2] = ACTIONS(5466),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5466),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5466),
    [sym_preproc_directive] = ACTIONS(5466),
    [anon_sym_DASH] = ACTIONS(5468),
    [anon_sym_PLUS] = ACTIONS(5468),
    [anon_sym_SEMI] = ACTIONS(5587),
    [anon_sym___extension__] = ACTIONS(5466),
    [anon_sym_typedef] = ACTIONS(5466),
    [anon_sym_extern] = ACTIONS(5466),
    [anon_sym___attribute__] = ACTIONS(5466),
    [anon_sym___attribute] = ACTIONS(5466),
    [anon_sym_noreturn] = ACTIONS(5466),
    [anon_sym_LBRACK] = ACTIONS(5468),
    [anon_sym___declspec] = ACTIONS(5466),
    [anon_sym___cdecl] = ACTIONS(5466),
    [anon_sym___clrcall] = ACTIONS(5466),
    [anon_sym___stdcall] = ACTIONS(5466),
    [anon_sym___fastcall] = ACTIONS(5466),
    [anon_sym___thiscall] = ACTIONS(5466),
    [anon_sym___vectorcall] = ACTIONS(5466),
    [anon_sym_signed] = ACTIONS(5466),
    [anon_sym_unsigned] = ACTIONS(5466),
    [anon_sym_long] = ACTIONS(5466),
    [anon_sym_short] = ACTIONS(5466),
    [anon_sym_static] = ACTIONS(5466),
    [anon_sym_auto] = ACTIONS(5466),
    [anon_sym_register] = ACTIONS(5466),
    [anon_sym_inline] = ACTIONS(5466),
    [anon_sym___inline] = ACTIONS(5466),
    [anon_sym___inline__] = ACTIONS(5466),
    [anon_sym___forceinline] = ACTIONS(5466),
    [anon_sym_thread_local] = ACTIONS(5466),
    [anon_sym___thread] = ACTIONS(5466),
    [anon_sym_CG_EXTERN] = ACTIONS(5466),
    [anon_sym_CG_INLINE] = ACTIONS(5466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5466),
    [anon_sym_IBOutlet] = ACTIONS(5466),
    [anon_sym_IBInspectable] = ACTIONS(5466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5466),
    [anon_sym_NS_INLINE] = ACTIONS(5466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5466),
    [anon_sym_const] = ACTIONS(5466),
    [anon_sym_constexpr] = ACTIONS(5466),
    [anon_sym_volatile] = ACTIONS(5466),
    [anon_sym_restrict] = ACTIONS(5466),
    [anon_sym___restrict__] = ACTIONS(5466),
    [anon_sym__Atomic] = ACTIONS(5466),
    [anon_sym__Noreturn] = ACTIONS(5466),
    [anon_sym__Nonnull] = ACTIONS(5466),
    [anon_sym_nullable] = ACTIONS(5466),
    [anon_sym__Complex] = ACTIONS(5466),
    [anon_sym__Nullable] = ACTIONS(5466),
    [anon_sym__Nullable_result] = ACTIONS(5466),
    [anon_sym__Null_unspecified] = ACTIONS(5466),
    [anon_sym___autoreleasing] = ACTIONS(5466),
    [anon_sym___block] = ACTIONS(5466),
    [anon_sym___bridge] = ACTIONS(5466),
    [anon_sym___bridge_retained] = ACTIONS(5466),
    [anon_sym___bridge_transfer] = ACTIONS(5466),
    [anon_sym___complex] = ACTIONS(5466),
    [anon_sym___const] = ACTIONS(5466),
    [anon_sym___imag] = ACTIONS(5466),
    [anon_sym___kindof] = ACTIONS(5466),
    [anon_sym___nonnull] = ACTIONS(5466),
    [anon_sym___nullable] = ACTIONS(5466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5466),
    [anon_sym___real] = ACTIONS(5466),
    [anon_sym___strong] = ACTIONS(5466),
    [anon_sym___unsafe_unretained] = ACTIONS(5466),
    [anon_sym___unused] = ACTIONS(5466),
    [anon_sym___weak] = ACTIONS(5466),
    [anon_sym_alignas] = ACTIONS(5466),
    [anon_sym__Alignas] = ACTIONS(5466),
    [sym_primitive_type] = ACTIONS(5466),
    [anon_sym_enum] = ACTIONS(5466),
    [anon_sym_struct] = ACTIONS(5466),
    [anon_sym_union] = ACTIONS(5466),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5466),
    [anon_sym___typeof] = ACTIONS(5466),
    [anon_sym_typeof] = ACTIONS(5466),
    [aux_sym_preproc_undef_token1] = ACTIONS(5466),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_API_AVAILABLE] = ACTIONS(5466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_API_DEPRECATED] = ACTIONS(5466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5466),
    [anon_sym___deprecated_msg] = ACTIONS(5466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5466),
    [anon_sym_ATsynthesize] = ACTIONS(5468),
    [anon_sym_ATdynamic] = ACTIONS(5468),
    [anon_sym_BOOL] = ACTIONS(5466),
    [anon_sym_IMP] = ACTIONS(5466),
    [anon_sym_SEL] = ACTIONS(5466),
    [anon_sym_Class] = ACTIONS(5466),
    [anon_sym_id] = ACTIONS(5466),
  },
  [STATE(2282)] = {
    [sym_identifier] = ACTIONS(5589),
    [aux_sym_preproc_def_token1] = ACTIONS(5589),
    [aux_sym_preproc_if_token1] = ACTIONS(5589),
    [aux_sym_preproc_if_token2] = ACTIONS(5589),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5589),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5589),
    [aux_sym_preproc_else_token1] = ACTIONS(5589),
    [aux_sym_preproc_elif_token1] = ACTIONS(5589),
    [sym_preproc_directive] = ACTIONS(5589),
    [anon_sym_DASH] = ACTIONS(5591),
    [anon_sym_PLUS] = ACTIONS(5591),
    [anon_sym___extension__] = ACTIONS(5589),
    [anon_sym_typedef] = ACTIONS(5589),
    [anon_sym_extern] = ACTIONS(5589),
    [anon_sym___attribute__] = ACTIONS(5589),
    [anon_sym___attribute] = ACTIONS(5589),
    [anon_sym_noreturn] = ACTIONS(5589),
    [anon_sym_LBRACK] = ACTIONS(5591),
    [anon_sym___declspec] = ACTIONS(5589),
    [anon_sym___cdecl] = ACTIONS(5589),
    [anon_sym___clrcall] = ACTIONS(5589),
    [anon_sym___stdcall] = ACTIONS(5589),
    [anon_sym___fastcall] = ACTIONS(5589),
    [anon_sym___thiscall] = ACTIONS(5589),
    [anon_sym___vectorcall] = ACTIONS(5589),
    [anon_sym_signed] = ACTIONS(5589),
    [anon_sym_unsigned] = ACTIONS(5589),
    [anon_sym_long] = ACTIONS(5589),
    [anon_sym_short] = ACTIONS(5589),
    [anon_sym_static] = ACTIONS(5589),
    [anon_sym_auto] = ACTIONS(5589),
    [anon_sym_register] = ACTIONS(5589),
    [anon_sym_inline] = ACTIONS(5589),
    [anon_sym___inline] = ACTIONS(5589),
    [anon_sym___inline__] = ACTIONS(5589),
    [anon_sym___forceinline] = ACTIONS(5589),
    [anon_sym_thread_local] = ACTIONS(5589),
    [anon_sym___thread] = ACTIONS(5589),
    [anon_sym_CG_EXTERN] = ACTIONS(5589),
    [anon_sym_CG_INLINE] = ACTIONS(5589),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5589),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5589),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5589),
    [anon_sym_IBOutlet] = ACTIONS(5589),
    [anon_sym_IBInspectable] = ACTIONS(5589),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5589),
    [anon_sym_NS_INLINE] = ACTIONS(5589),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5589),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5589),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5589),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5589),
    [anon_sym_const] = ACTIONS(5589),
    [anon_sym_constexpr] = ACTIONS(5589),
    [anon_sym_volatile] = ACTIONS(5589),
    [anon_sym_restrict] = ACTIONS(5589),
    [anon_sym___restrict__] = ACTIONS(5589),
    [anon_sym__Atomic] = ACTIONS(5589),
    [anon_sym__Noreturn] = ACTIONS(5589),
    [anon_sym__Nonnull] = ACTIONS(5589),
    [anon_sym_nullable] = ACTIONS(5589),
    [anon_sym__Complex] = ACTIONS(5589),
    [anon_sym__Nullable] = ACTIONS(5589),
    [anon_sym__Nullable_result] = ACTIONS(5589),
    [anon_sym__Null_unspecified] = ACTIONS(5589),
    [anon_sym___autoreleasing] = ACTIONS(5589),
    [anon_sym___block] = ACTIONS(5589),
    [anon_sym___bridge] = ACTIONS(5589),
    [anon_sym___bridge_retained] = ACTIONS(5589),
    [anon_sym___bridge_transfer] = ACTIONS(5589),
    [anon_sym___complex] = ACTIONS(5589),
    [anon_sym___const] = ACTIONS(5589),
    [anon_sym___imag] = ACTIONS(5589),
    [anon_sym___kindof] = ACTIONS(5589),
    [anon_sym___nonnull] = ACTIONS(5589),
    [anon_sym___nullable] = ACTIONS(5589),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5589),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5589),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5589),
    [anon_sym___real] = ACTIONS(5589),
    [anon_sym___strong] = ACTIONS(5589),
    [anon_sym___unsafe_unretained] = ACTIONS(5589),
    [anon_sym___unused] = ACTIONS(5589),
    [anon_sym___weak] = ACTIONS(5589),
    [anon_sym_alignas] = ACTIONS(5589),
    [anon_sym__Alignas] = ACTIONS(5589),
    [sym_primitive_type] = ACTIONS(5589),
    [anon_sym_enum] = ACTIONS(5589),
    [anon_sym_struct] = ACTIONS(5589),
    [anon_sym_union] = ACTIONS(5589),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5589),
    [anon_sym___typeof] = ACTIONS(5589),
    [anon_sym_typeof] = ACTIONS(5589),
    [aux_sym_preproc_undef_token1] = ACTIONS(5589),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5589),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5589),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5589),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5589),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5589),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5589),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5589),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5589),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5589),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_API_AVAILABLE] = ACTIONS(5589),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_API_DEPRECATED] = ACTIONS(5589),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5589),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5589),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5589),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5589),
    [anon_sym___deprecated_msg] = ACTIONS(5589),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5589),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5589),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5589),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5589),
    [anon_sym_ATproperty] = ACTIONS(5591),
    [anon_sym_BOOL] = ACTIONS(5589),
    [anon_sym_IMP] = ACTIONS(5589),
    [anon_sym_SEL] = ACTIONS(5589),
    [anon_sym_Class] = ACTIONS(5589),
    [anon_sym_id] = ACTIONS(5589),
  },
  [STATE(2283)] = {
    [sym_identifier] = ACTIONS(5375),
    [aux_sym_preproc_def_token1] = ACTIONS(5375),
    [aux_sym_preproc_if_token1] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5375),
    [sym_preproc_directive] = ACTIONS(5375),
    [anon_sym_DASH] = ACTIONS(5377),
    [anon_sym_PLUS] = ACTIONS(5377),
    [anon_sym_SEMI] = ACTIONS(5593),
    [anon_sym___extension__] = ACTIONS(5375),
    [anon_sym_typedef] = ACTIONS(5375),
    [anon_sym_extern] = ACTIONS(5375),
    [anon_sym___attribute__] = ACTIONS(5375),
    [anon_sym___attribute] = ACTIONS(5375),
    [anon_sym_noreturn] = ACTIONS(5375),
    [anon_sym_LBRACK] = ACTIONS(5377),
    [anon_sym___declspec] = ACTIONS(5375),
    [anon_sym___cdecl] = ACTIONS(5375),
    [anon_sym___clrcall] = ACTIONS(5375),
    [anon_sym___stdcall] = ACTIONS(5375),
    [anon_sym___fastcall] = ACTIONS(5375),
    [anon_sym___thiscall] = ACTIONS(5375),
    [anon_sym___vectorcall] = ACTIONS(5375),
    [anon_sym_signed] = ACTIONS(5375),
    [anon_sym_unsigned] = ACTIONS(5375),
    [anon_sym_long] = ACTIONS(5375),
    [anon_sym_short] = ACTIONS(5375),
    [anon_sym_static] = ACTIONS(5375),
    [anon_sym_auto] = ACTIONS(5375),
    [anon_sym_register] = ACTIONS(5375),
    [anon_sym_inline] = ACTIONS(5375),
    [anon_sym___inline] = ACTIONS(5375),
    [anon_sym___inline__] = ACTIONS(5375),
    [anon_sym___forceinline] = ACTIONS(5375),
    [anon_sym_thread_local] = ACTIONS(5375),
    [anon_sym___thread] = ACTIONS(5375),
    [anon_sym_CG_EXTERN] = ACTIONS(5375),
    [anon_sym_CG_INLINE] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5375),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5375),
    [anon_sym_IBOutlet] = ACTIONS(5375),
    [anon_sym_IBInspectable] = ACTIONS(5375),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5375),
    [anon_sym_NS_INLINE] = ACTIONS(5375),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5375),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5375),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5375),
    [anon_sym_const] = ACTIONS(5375),
    [anon_sym_constexpr] = ACTIONS(5375),
    [anon_sym_volatile] = ACTIONS(5375),
    [anon_sym_restrict] = ACTIONS(5375),
    [anon_sym___restrict__] = ACTIONS(5375),
    [anon_sym__Atomic] = ACTIONS(5375),
    [anon_sym__Noreturn] = ACTIONS(5375),
    [anon_sym__Nonnull] = ACTIONS(5375),
    [anon_sym_nullable] = ACTIONS(5375),
    [anon_sym__Complex] = ACTIONS(5375),
    [anon_sym__Nullable] = ACTIONS(5375),
    [anon_sym__Nullable_result] = ACTIONS(5375),
    [anon_sym__Null_unspecified] = ACTIONS(5375),
    [anon_sym___autoreleasing] = ACTIONS(5375),
    [anon_sym___block] = ACTIONS(5375),
    [anon_sym___bridge] = ACTIONS(5375),
    [anon_sym___bridge_retained] = ACTIONS(5375),
    [anon_sym___bridge_transfer] = ACTIONS(5375),
    [anon_sym___complex] = ACTIONS(5375),
    [anon_sym___const] = ACTIONS(5375),
    [anon_sym___imag] = ACTIONS(5375),
    [anon_sym___kindof] = ACTIONS(5375),
    [anon_sym___nonnull] = ACTIONS(5375),
    [anon_sym___nullable] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5375),
    [anon_sym___real] = ACTIONS(5375),
    [anon_sym___strong] = ACTIONS(5375),
    [anon_sym___unsafe_unretained] = ACTIONS(5375),
    [anon_sym___unused] = ACTIONS(5375),
    [anon_sym___weak] = ACTIONS(5375),
    [anon_sym_alignas] = ACTIONS(5375),
    [anon_sym__Alignas] = ACTIONS(5375),
    [sym_primitive_type] = ACTIONS(5375),
    [anon_sym_enum] = ACTIONS(5375),
    [anon_sym_struct] = ACTIONS(5375),
    [anon_sym_union] = ACTIONS(5375),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5375),
    [anon_sym___typeof] = ACTIONS(5375),
    [anon_sym_typeof] = ACTIONS(5375),
    [aux_sym_preproc_undef_token1] = ACTIONS(5375),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5375),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5375),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5375),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_API_AVAILABLE] = ACTIONS(5375),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_API_DEPRECATED] = ACTIONS(5375),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym___deprecated_msg] = ACTIONS(5375),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5375),
    [anon_sym_ATend] = ACTIONS(5377),
    [anon_sym_ATsynthesize] = ACTIONS(5377),
    [anon_sym_ATdynamic] = ACTIONS(5377),
    [anon_sym_BOOL] = ACTIONS(5375),
    [anon_sym_IMP] = ACTIONS(5375),
    [anon_sym_SEL] = ACTIONS(5375),
    [anon_sym_Class] = ACTIONS(5375),
    [anon_sym_id] = ACTIONS(5375),
  },
  [STATE(2284)] = {
    [sym_identifier] = ACTIONS(2424),
    [aux_sym_preproc_def_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token1] = ACTIONS(2424),
    [aux_sym_preproc_if_token2] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2424),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2424),
    [sym_preproc_directive] = ACTIONS(2424),
    [anon_sym_DASH] = ACTIONS(2426),
    [anon_sym_PLUS] = ACTIONS(2426),
    [anon_sym___extension__] = ACTIONS(2424),
    [anon_sym_typedef] = ACTIONS(2424),
    [anon_sym_extern] = ACTIONS(2424),
    [anon_sym___attribute__] = ACTIONS(2424),
    [anon_sym___attribute] = ACTIONS(2424),
    [anon_sym_noreturn] = ACTIONS(2424),
    [anon_sym_LBRACK] = ACTIONS(2426),
    [anon_sym___declspec] = ACTIONS(2424),
    [anon_sym___cdecl] = ACTIONS(2424),
    [anon_sym___clrcall] = ACTIONS(2424),
    [anon_sym___stdcall] = ACTIONS(2424),
    [anon_sym___fastcall] = ACTIONS(2424),
    [anon_sym___thiscall] = ACTIONS(2424),
    [anon_sym___vectorcall] = ACTIONS(2424),
    [anon_sym_signed] = ACTIONS(2424),
    [anon_sym_unsigned] = ACTIONS(2424),
    [anon_sym_long] = ACTIONS(2424),
    [anon_sym_short] = ACTIONS(2424),
    [anon_sym_static] = ACTIONS(2424),
    [anon_sym_auto] = ACTIONS(2424),
    [anon_sym_register] = ACTIONS(2424),
    [anon_sym_inline] = ACTIONS(2424),
    [anon_sym___inline] = ACTIONS(2424),
    [anon_sym___inline__] = ACTIONS(2424),
    [anon_sym___forceinline] = ACTIONS(2424),
    [anon_sym_thread_local] = ACTIONS(2424),
    [anon_sym___thread] = ACTIONS(2424),
    [anon_sym_CG_EXTERN] = ACTIONS(2424),
    [anon_sym_CG_INLINE] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2424),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2424),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2424),
    [anon_sym_IBOutlet] = ACTIONS(2424),
    [anon_sym_IBInspectable] = ACTIONS(2424),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2424),
    [anon_sym_NS_INLINE] = ACTIONS(2424),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2424),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2424),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2424),
    [anon_sym_const] = ACTIONS(2424),
    [anon_sym_constexpr] = ACTIONS(2424),
    [anon_sym_volatile] = ACTIONS(2424),
    [anon_sym_restrict] = ACTIONS(2424),
    [anon_sym___restrict__] = ACTIONS(2424),
    [anon_sym__Atomic] = ACTIONS(2424),
    [anon_sym__Noreturn] = ACTIONS(2424),
    [anon_sym__Nonnull] = ACTIONS(2424),
    [anon_sym_nullable] = ACTIONS(2424),
    [anon_sym__Complex] = ACTIONS(2424),
    [anon_sym__Nullable] = ACTIONS(2424),
    [anon_sym__Nullable_result] = ACTIONS(2424),
    [anon_sym__Null_unspecified] = ACTIONS(2424),
    [anon_sym___autoreleasing] = ACTIONS(2424),
    [anon_sym___block] = ACTIONS(2424),
    [anon_sym___bridge] = ACTIONS(2424),
    [anon_sym___bridge_retained] = ACTIONS(2424),
    [anon_sym___bridge_transfer] = ACTIONS(2424),
    [anon_sym___complex] = ACTIONS(2424),
    [anon_sym___const] = ACTIONS(2424),
    [anon_sym___imag] = ACTIONS(2424),
    [anon_sym___kindof] = ACTIONS(2424),
    [anon_sym___nonnull] = ACTIONS(2424),
    [anon_sym___nullable] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2424),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2424),
    [anon_sym___real] = ACTIONS(2424),
    [anon_sym___strong] = ACTIONS(2424),
    [anon_sym___unsafe_unretained] = ACTIONS(2424),
    [anon_sym___unused] = ACTIONS(2424),
    [anon_sym___weak] = ACTIONS(2424),
    [anon_sym_alignas] = ACTIONS(2424),
    [anon_sym__Alignas] = ACTIONS(2424),
    [sym_primitive_type] = ACTIONS(2424),
    [anon_sym_enum] = ACTIONS(2424),
    [anon_sym_struct] = ACTIONS(2424),
    [anon_sym_union] = ACTIONS(2424),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2424),
    [anon_sym___typeof] = ACTIONS(2424),
    [anon_sym_typeof] = ACTIONS(2424),
    [aux_sym_preproc_undef_token1] = ACTIONS(2424),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2424),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2424),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2424),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2424),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2424),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_API_AVAILABLE] = ACTIONS(2424),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_API_DEPRECATED] = ACTIONS(2424),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2424),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2424),
    [anon_sym___deprecated_msg] = ACTIONS(2424),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2424),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2424),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2424),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2424),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2424),
    [anon_sym_ATsynthesize] = ACTIONS(2426),
    [anon_sym_ATdynamic] = ACTIONS(2426),
    [anon_sym_ATproperty] = ACTIONS(2426),
    [anon_sym_BOOL] = ACTIONS(2424),
    [anon_sym_IMP] = ACTIONS(2424),
    [anon_sym_SEL] = ACTIONS(2424),
    [anon_sym_Class] = ACTIONS(2424),
    [anon_sym_id] = ACTIONS(2424),
  },
  [STATE(2285)] = {
    [sym_identifier] = ACTIONS(2716),
    [aux_sym_preproc_def_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token1] = ACTIONS(2716),
    [aux_sym_preproc_if_token2] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2716),
    [sym_preproc_directive] = ACTIONS(2716),
    [anon_sym_DASH] = ACTIONS(2718),
    [anon_sym_PLUS] = ACTIONS(2718),
    [anon_sym___extension__] = ACTIONS(2716),
    [anon_sym_typedef] = ACTIONS(2716),
    [anon_sym_extern] = ACTIONS(2716),
    [anon_sym___attribute__] = ACTIONS(2716),
    [anon_sym___attribute] = ACTIONS(2716),
    [anon_sym_noreturn] = ACTIONS(2716),
    [anon_sym_LBRACK] = ACTIONS(2718),
    [anon_sym___declspec] = ACTIONS(2716),
    [anon_sym___cdecl] = ACTIONS(2716),
    [anon_sym___clrcall] = ACTIONS(2716),
    [anon_sym___stdcall] = ACTIONS(2716),
    [anon_sym___fastcall] = ACTIONS(2716),
    [anon_sym___thiscall] = ACTIONS(2716),
    [anon_sym___vectorcall] = ACTIONS(2716),
    [anon_sym_signed] = ACTIONS(2716),
    [anon_sym_unsigned] = ACTIONS(2716),
    [anon_sym_long] = ACTIONS(2716),
    [anon_sym_short] = ACTIONS(2716),
    [anon_sym_static] = ACTIONS(2716),
    [anon_sym_auto] = ACTIONS(2716),
    [anon_sym_register] = ACTIONS(2716),
    [anon_sym_inline] = ACTIONS(2716),
    [anon_sym___inline] = ACTIONS(2716),
    [anon_sym___inline__] = ACTIONS(2716),
    [anon_sym___forceinline] = ACTIONS(2716),
    [anon_sym_thread_local] = ACTIONS(2716),
    [anon_sym___thread] = ACTIONS(2716),
    [anon_sym_CG_EXTERN] = ACTIONS(2716),
    [anon_sym_CG_INLINE] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2716),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2716),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2716),
    [anon_sym_IBOutlet] = ACTIONS(2716),
    [anon_sym_IBInspectable] = ACTIONS(2716),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2716),
    [anon_sym_NS_INLINE] = ACTIONS(2716),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2716),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2716),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2716),
    [anon_sym_const] = ACTIONS(2716),
    [anon_sym_constexpr] = ACTIONS(2716),
    [anon_sym_volatile] = ACTIONS(2716),
    [anon_sym_restrict] = ACTIONS(2716),
    [anon_sym___restrict__] = ACTIONS(2716),
    [anon_sym__Atomic] = ACTIONS(2716),
    [anon_sym__Noreturn] = ACTIONS(2716),
    [anon_sym__Nonnull] = ACTIONS(2716),
    [anon_sym_nullable] = ACTIONS(2716),
    [anon_sym__Complex] = ACTIONS(2716),
    [anon_sym__Nullable] = ACTIONS(2716),
    [anon_sym__Nullable_result] = ACTIONS(2716),
    [anon_sym__Null_unspecified] = ACTIONS(2716),
    [anon_sym___autoreleasing] = ACTIONS(2716),
    [anon_sym___block] = ACTIONS(2716),
    [anon_sym___bridge] = ACTIONS(2716),
    [anon_sym___bridge_retained] = ACTIONS(2716),
    [anon_sym___bridge_transfer] = ACTIONS(2716),
    [anon_sym___complex] = ACTIONS(2716),
    [anon_sym___const] = ACTIONS(2716),
    [anon_sym___imag] = ACTIONS(2716),
    [anon_sym___kindof] = ACTIONS(2716),
    [anon_sym___nonnull] = ACTIONS(2716),
    [anon_sym___nullable] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2716),
    [anon_sym___real] = ACTIONS(2716),
    [anon_sym___strong] = ACTIONS(2716),
    [anon_sym___unsafe_unretained] = ACTIONS(2716),
    [anon_sym___unused] = ACTIONS(2716),
    [anon_sym___weak] = ACTIONS(2716),
    [anon_sym_alignas] = ACTIONS(2716),
    [anon_sym__Alignas] = ACTIONS(2716),
    [sym_primitive_type] = ACTIONS(2716),
    [anon_sym_enum] = ACTIONS(2716),
    [anon_sym_struct] = ACTIONS(2716),
    [anon_sym_union] = ACTIONS(2716),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2716),
    [anon_sym___typeof] = ACTIONS(2716),
    [anon_sym_typeof] = ACTIONS(2716),
    [aux_sym_preproc_undef_token1] = ACTIONS(2716),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2716),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2716),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2716),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2716),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2716),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_API_AVAILABLE] = ACTIONS(2716),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_API_DEPRECATED] = ACTIONS(2716),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2716),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2716),
    [anon_sym___deprecated_msg] = ACTIONS(2716),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2716),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2716),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2716),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2716),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2716),
    [anon_sym_ATsynthesize] = ACTIONS(2718),
    [anon_sym_ATdynamic] = ACTIONS(2718),
    [anon_sym_ATproperty] = ACTIONS(2718),
    [anon_sym_BOOL] = ACTIONS(2716),
    [anon_sym_IMP] = ACTIONS(2716),
    [anon_sym_SEL] = ACTIONS(2716),
    [anon_sym_Class] = ACTIONS(2716),
    [anon_sym_id] = ACTIONS(2716),
  },
  [STATE(2286)] = {
    [sym_identifier] = ACTIONS(5595),
    [aux_sym_preproc_def_token1] = ACTIONS(5595),
    [aux_sym_preproc_if_token1] = ACTIONS(5595),
    [aux_sym_preproc_if_token2] = ACTIONS(5595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5595),
    [aux_sym_preproc_else_token1] = ACTIONS(5595),
    [aux_sym_preproc_elif_token1] = ACTIONS(5595),
    [sym_preproc_directive] = ACTIONS(5595),
    [anon_sym_DASH] = ACTIONS(5597),
    [anon_sym_PLUS] = ACTIONS(5597),
    [anon_sym___extension__] = ACTIONS(5595),
    [anon_sym_typedef] = ACTIONS(5595),
    [anon_sym_extern] = ACTIONS(5595),
    [anon_sym___attribute__] = ACTIONS(5595),
    [anon_sym___attribute] = ACTIONS(5595),
    [anon_sym_noreturn] = ACTIONS(5595),
    [anon_sym_LBRACK] = ACTIONS(5597),
    [anon_sym___declspec] = ACTIONS(5595),
    [anon_sym___cdecl] = ACTIONS(5595),
    [anon_sym___clrcall] = ACTIONS(5595),
    [anon_sym___stdcall] = ACTIONS(5595),
    [anon_sym___fastcall] = ACTIONS(5595),
    [anon_sym___thiscall] = ACTIONS(5595),
    [anon_sym___vectorcall] = ACTIONS(5595),
    [anon_sym_signed] = ACTIONS(5595),
    [anon_sym_unsigned] = ACTIONS(5595),
    [anon_sym_long] = ACTIONS(5595),
    [anon_sym_short] = ACTIONS(5595),
    [anon_sym_static] = ACTIONS(5595),
    [anon_sym_auto] = ACTIONS(5595),
    [anon_sym_register] = ACTIONS(5595),
    [anon_sym_inline] = ACTIONS(5595),
    [anon_sym___inline] = ACTIONS(5595),
    [anon_sym___inline__] = ACTIONS(5595),
    [anon_sym___forceinline] = ACTIONS(5595),
    [anon_sym_thread_local] = ACTIONS(5595),
    [anon_sym___thread] = ACTIONS(5595),
    [anon_sym_CG_EXTERN] = ACTIONS(5595),
    [anon_sym_CG_INLINE] = ACTIONS(5595),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5595),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5595),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5595),
    [anon_sym_IBOutlet] = ACTIONS(5595),
    [anon_sym_IBInspectable] = ACTIONS(5595),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5595),
    [anon_sym_NS_INLINE] = ACTIONS(5595),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5595),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5595),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5595),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5595),
    [anon_sym_const] = ACTIONS(5595),
    [anon_sym_constexpr] = ACTIONS(5595),
    [anon_sym_volatile] = ACTIONS(5595),
    [anon_sym_restrict] = ACTIONS(5595),
    [anon_sym___restrict__] = ACTIONS(5595),
    [anon_sym__Atomic] = ACTIONS(5595),
    [anon_sym__Noreturn] = ACTIONS(5595),
    [anon_sym__Nonnull] = ACTIONS(5595),
    [anon_sym_nullable] = ACTIONS(5595),
    [anon_sym__Complex] = ACTIONS(5595),
    [anon_sym__Nullable] = ACTIONS(5595),
    [anon_sym__Nullable_result] = ACTIONS(5595),
    [anon_sym__Null_unspecified] = ACTIONS(5595),
    [anon_sym___autoreleasing] = ACTIONS(5595),
    [anon_sym___block] = ACTIONS(5595),
    [anon_sym___bridge] = ACTIONS(5595),
    [anon_sym___bridge_retained] = ACTIONS(5595),
    [anon_sym___bridge_transfer] = ACTIONS(5595),
    [anon_sym___complex] = ACTIONS(5595),
    [anon_sym___const] = ACTIONS(5595),
    [anon_sym___imag] = ACTIONS(5595),
    [anon_sym___kindof] = ACTIONS(5595),
    [anon_sym___nonnull] = ACTIONS(5595),
    [anon_sym___nullable] = ACTIONS(5595),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5595),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5595),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5595),
    [anon_sym___real] = ACTIONS(5595),
    [anon_sym___strong] = ACTIONS(5595),
    [anon_sym___unsafe_unretained] = ACTIONS(5595),
    [anon_sym___unused] = ACTIONS(5595),
    [anon_sym___weak] = ACTIONS(5595),
    [anon_sym_alignas] = ACTIONS(5595),
    [anon_sym__Alignas] = ACTIONS(5595),
    [sym_primitive_type] = ACTIONS(5595),
    [anon_sym_enum] = ACTIONS(5595),
    [anon_sym_struct] = ACTIONS(5595),
    [anon_sym_union] = ACTIONS(5595),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5595),
    [anon_sym___typeof] = ACTIONS(5595),
    [anon_sym_typeof] = ACTIONS(5595),
    [aux_sym_preproc_undef_token1] = ACTIONS(5595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5595),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5595),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5595),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5595),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5595),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5595),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5595),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5595),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5595),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_API_AVAILABLE] = ACTIONS(5595),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_API_DEPRECATED] = ACTIONS(5595),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5595),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5595),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5595),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5595),
    [anon_sym___deprecated_msg] = ACTIONS(5595),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5595),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5595),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5595),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5595),
    [anon_sym_ATproperty] = ACTIONS(5597),
    [anon_sym_BOOL] = ACTIONS(5595),
    [anon_sym_IMP] = ACTIONS(5595),
    [anon_sym_SEL] = ACTIONS(5595),
    [anon_sym_Class] = ACTIONS(5595),
    [anon_sym_id] = ACTIONS(5595),
  },
  [STATE(2287)] = {
    [sym_identifier] = ACTIONS(2420),
    [aux_sym_preproc_def_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token1] = ACTIONS(2420),
    [aux_sym_preproc_if_token2] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2420),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2420),
    [sym_preproc_directive] = ACTIONS(2420),
    [anon_sym_DASH] = ACTIONS(2422),
    [anon_sym_PLUS] = ACTIONS(2422),
    [anon_sym___extension__] = ACTIONS(2420),
    [anon_sym_typedef] = ACTIONS(2420),
    [anon_sym_extern] = ACTIONS(2420),
    [anon_sym___attribute__] = ACTIONS(2420),
    [anon_sym___attribute] = ACTIONS(2420),
    [anon_sym_noreturn] = ACTIONS(2420),
    [anon_sym_LBRACK] = ACTIONS(2422),
    [anon_sym___declspec] = ACTIONS(2420),
    [anon_sym___cdecl] = ACTIONS(2420),
    [anon_sym___clrcall] = ACTIONS(2420),
    [anon_sym___stdcall] = ACTIONS(2420),
    [anon_sym___fastcall] = ACTIONS(2420),
    [anon_sym___thiscall] = ACTIONS(2420),
    [anon_sym___vectorcall] = ACTIONS(2420),
    [anon_sym_signed] = ACTIONS(2420),
    [anon_sym_unsigned] = ACTIONS(2420),
    [anon_sym_long] = ACTIONS(2420),
    [anon_sym_short] = ACTIONS(2420),
    [anon_sym_static] = ACTIONS(2420),
    [anon_sym_auto] = ACTIONS(2420),
    [anon_sym_register] = ACTIONS(2420),
    [anon_sym_inline] = ACTIONS(2420),
    [anon_sym___inline] = ACTIONS(2420),
    [anon_sym___inline__] = ACTIONS(2420),
    [anon_sym___forceinline] = ACTIONS(2420),
    [anon_sym_thread_local] = ACTIONS(2420),
    [anon_sym___thread] = ACTIONS(2420),
    [anon_sym_CG_EXTERN] = ACTIONS(2420),
    [anon_sym_CG_INLINE] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2420),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2420),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2420),
    [anon_sym_IBOutlet] = ACTIONS(2420),
    [anon_sym_IBInspectable] = ACTIONS(2420),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2420),
    [anon_sym_NS_INLINE] = ACTIONS(2420),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2420),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2420),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2420),
    [anon_sym_const] = ACTIONS(2420),
    [anon_sym_constexpr] = ACTIONS(2420),
    [anon_sym_volatile] = ACTIONS(2420),
    [anon_sym_restrict] = ACTIONS(2420),
    [anon_sym___restrict__] = ACTIONS(2420),
    [anon_sym__Atomic] = ACTIONS(2420),
    [anon_sym__Noreturn] = ACTIONS(2420),
    [anon_sym__Nonnull] = ACTIONS(2420),
    [anon_sym_nullable] = ACTIONS(2420),
    [anon_sym__Complex] = ACTIONS(2420),
    [anon_sym__Nullable] = ACTIONS(2420),
    [anon_sym__Nullable_result] = ACTIONS(2420),
    [anon_sym__Null_unspecified] = ACTIONS(2420),
    [anon_sym___autoreleasing] = ACTIONS(2420),
    [anon_sym___block] = ACTIONS(2420),
    [anon_sym___bridge] = ACTIONS(2420),
    [anon_sym___bridge_retained] = ACTIONS(2420),
    [anon_sym___bridge_transfer] = ACTIONS(2420),
    [anon_sym___complex] = ACTIONS(2420),
    [anon_sym___const] = ACTIONS(2420),
    [anon_sym___imag] = ACTIONS(2420),
    [anon_sym___kindof] = ACTIONS(2420),
    [anon_sym___nonnull] = ACTIONS(2420),
    [anon_sym___nullable] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2420),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2420),
    [anon_sym___real] = ACTIONS(2420),
    [anon_sym___strong] = ACTIONS(2420),
    [anon_sym___unsafe_unretained] = ACTIONS(2420),
    [anon_sym___unused] = ACTIONS(2420),
    [anon_sym___weak] = ACTIONS(2420),
    [anon_sym_alignas] = ACTIONS(2420),
    [anon_sym__Alignas] = ACTIONS(2420),
    [sym_primitive_type] = ACTIONS(2420),
    [anon_sym_enum] = ACTIONS(2420),
    [anon_sym_struct] = ACTIONS(2420),
    [anon_sym_union] = ACTIONS(2420),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2420),
    [anon_sym___typeof] = ACTIONS(2420),
    [anon_sym_typeof] = ACTIONS(2420),
    [aux_sym_preproc_undef_token1] = ACTIONS(2420),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2420),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2420),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2420),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2420),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2420),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_API_AVAILABLE] = ACTIONS(2420),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_API_DEPRECATED] = ACTIONS(2420),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2420),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2420),
    [anon_sym___deprecated_msg] = ACTIONS(2420),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2420),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2420),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2420),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2420),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2420),
    [anon_sym_ATsynthesize] = ACTIONS(2422),
    [anon_sym_ATdynamic] = ACTIONS(2422),
    [anon_sym_ATproperty] = ACTIONS(2422),
    [anon_sym_BOOL] = ACTIONS(2420),
    [anon_sym_IMP] = ACTIONS(2420),
    [anon_sym_SEL] = ACTIONS(2420),
    [anon_sym_Class] = ACTIONS(2420),
    [anon_sym_id] = ACTIONS(2420),
  },
  [STATE(2288)] = {
    [sym_identifier] = ACTIONS(5541),
    [aux_sym_preproc_def_token1] = ACTIONS(5541),
    [aux_sym_preproc_if_token1] = ACTIONS(5541),
    [aux_sym_preproc_if_token2] = ACTIONS(5541),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5541),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5541),
    [aux_sym_preproc_else_token1] = ACTIONS(5541),
    [aux_sym_preproc_elif_token1] = ACTIONS(5541),
    [sym_preproc_directive] = ACTIONS(5541),
    [anon_sym_DASH] = ACTIONS(5543),
    [anon_sym_PLUS] = ACTIONS(5543),
    [anon_sym___extension__] = ACTIONS(5541),
    [anon_sym_typedef] = ACTIONS(5541),
    [anon_sym_extern] = ACTIONS(5541),
    [anon_sym___attribute__] = ACTIONS(5541),
    [anon_sym___attribute] = ACTIONS(5541),
    [anon_sym_noreturn] = ACTIONS(5541),
    [anon_sym_LBRACK] = ACTIONS(5543),
    [anon_sym___declspec] = ACTIONS(5541),
    [anon_sym___cdecl] = ACTIONS(5541),
    [anon_sym___clrcall] = ACTIONS(5541),
    [anon_sym___stdcall] = ACTIONS(5541),
    [anon_sym___fastcall] = ACTIONS(5541),
    [anon_sym___thiscall] = ACTIONS(5541),
    [anon_sym___vectorcall] = ACTIONS(5541),
    [anon_sym_signed] = ACTIONS(5541),
    [anon_sym_unsigned] = ACTIONS(5541),
    [anon_sym_long] = ACTIONS(5541),
    [anon_sym_short] = ACTIONS(5541),
    [anon_sym_static] = ACTIONS(5541),
    [anon_sym_auto] = ACTIONS(5541),
    [anon_sym_register] = ACTIONS(5541),
    [anon_sym_inline] = ACTIONS(5541),
    [anon_sym___inline] = ACTIONS(5541),
    [anon_sym___inline__] = ACTIONS(5541),
    [anon_sym___forceinline] = ACTIONS(5541),
    [anon_sym_thread_local] = ACTIONS(5541),
    [anon_sym___thread] = ACTIONS(5541),
    [anon_sym_CG_EXTERN] = ACTIONS(5541),
    [anon_sym_CG_INLINE] = ACTIONS(5541),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5541),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5541),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5541),
    [anon_sym_IBOutlet] = ACTIONS(5541),
    [anon_sym_IBInspectable] = ACTIONS(5541),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5541),
    [anon_sym_NS_INLINE] = ACTIONS(5541),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5541),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5541),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5541),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5541),
    [anon_sym_const] = ACTIONS(5541),
    [anon_sym_constexpr] = ACTIONS(5541),
    [anon_sym_volatile] = ACTIONS(5541),
    [anon_sym_restrict] = ACTIONS(5541),
    [anon_sym___restrict__] = ACTIONS(5541),
    [anon_sym__Atomic] = ACTIONS(5541),
    [anon_sym__Noreturn] = ACTIONS(5541),
    [anon_sym__Nonnull] = ACTIONS(5541),
    [anon_sym_nullable] = ACTIONS(5541),
    [anon_sym__Complex] = ACTIONS(5541),
    [anon_sym__Nullable] = ACTIONS(5541),
    [anon_sym__Nullable_result] = ACTIONS(5541),
    [anon_sym__Null_unspecified] = ACTIONS(5541),
    [anon_sym___autoreleasing] = ACTIONS(5541),
    [anon_sym___block] = ACTIONS(5541),
    [anon_sym___bridge] = ACTIONS(5541),
    [anon_sym___bridge_retained] = ACTIONS(5541),
    [anon_sym___bridge_transfer] = ACTIONS(5541),
    [anon_sym___complex] = ACTIONS(5541),
    [anon_sym___const] = ACTIONS(5541),
    [anon_sym___imag] = ACTIONS(5541),
    [anon_sym___kindof] = ACTIONS(5541),
    [anon_sym___nonnull] = ACTIONS(5541),
    [anon_sym___nullable] = ACTIONS(5541),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5541),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5541),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5541),
    [anon_sym___real] = ACTIONS(5541),
    [anon_sym___strong] = ACTIONS(5541),
    [anon_sym___unsafe_unretained] = ACTIONS(5541),
    [anon_sym___unused] = ACTIONS(5541),
    [anon_sym___weak] = ACTIONS(5541),
    [anon_sym_alignas] = ACTIONS(5541),
    [anon_sym__Alignas] = ACTIONS(5541),
    [sym_primitive_type] = ACTIONS(5541),
    [anon_sym_enum] = ACTIONS(5541),
    [anon_sym_struct] = ACTIONS(5541),
    [anon_sym_union] = ACTIONS(5541),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5541),
    [anon_sym___typeof] = ACTIONS(5541),
    [anon_sym_typeof] = ACTIONS(5541),
    [aux_sym_preproc_undef_token1] = ACTIONS(5541),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5541),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5541),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5541),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5541),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5541),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5541),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5541),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5541),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5541),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_API_AVAILABLE] = ACTIONS(5541),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_API_DEPRECATED] = ACTIONS(5541),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5541),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5541),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5541),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5541),
    [anon_sym___deprecated_msg] = ACTIONS(5541),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5541),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5541),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5541),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5541),
    [anon_sym_ATproperty] = ACTIONS(5543),
    [anon_sym_BOOL] = ACTIONS(5541),
    [anon_sym_IMP] = ACTIONS(5541),
    [anon_sym_SEL] = ACTIONS(5541),
    [anon_sym_Class] = ACTIONS(5541),
    [anon_sym_id] = ACTIONS(5541),
  },
  [STATE(2289)] = {
    [sym_identifier] = ACTIONS(5381),
    [aux_sym_preproc_def_token1] = ACTIONS(5381),
    [aux_sym_preproc_if_token1] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5381),
    [sym_preproc_directive] = ACTIONS(5381),
    [anon_sym_DASH] = ACTIONS(5383),
    [anon_sym_PLUS] = ACTIONS(5383),
    [anon_sym_SEMI] = ACTIONS(5599),
    [anon_sym___extension__] = ACTIONS(5381),
    [anon_sym_typedef] = ACTIONS(5381),
    [anon_sym_extern] = ACTIONS(5381),
    [anon_sym___attribute__] = ACTIONS(5381),
    [anon_sym___attribute] = ACTIONS(5381),
    [anon_sym_noreturn] = ACTIONS(5381),
    [anon_sym_LBRACK] = ACTIONS(5383),
    [anon_sym___declspec] = ACTIONS(5381),
    [anon_sym___cdecl] = ACTIONS(5381),
    [anon_sym___clrcall] = ACTIONS(5381),
    [anon_sym___stdcall] = ACTIONS(5381),
    [anon_sym___fastcall] = ACTIONS(5381),
    [anon_sym___thiscall] = ACTIONS(5381),
    [anon_sym___vectorcall] = ACTIONS(5381),
    [anon_sym_signed] = ACTIONS(5381),
    [anon_sym_unsigned] = ACTIONS(5381),
    [anon_sym_long] = ACTIONS(5381),
    [anon_sym_short] = ACTIONS(5381),
    [anon_sym_static] = ACTIONS(5381),
    [anon_sym_auto] = ACTIONS(5381),
    [anon_sym_register] = ACTIONS(5381),
    [anon_sym_inline] = ACTIONS(5381),
    [anon_sym___inline] = ACTIONS(5381),
    [anon_sym___inline__] = ACTIONS(5381),
    [anon_sym___forceinline] = ACTIONS(5381),
    [anon_sym_thread_local] = ACTIONS(5381),
    [anon_sym___thread] = ACTIONS(5381),
    [anon_sym_CG_EXTERN] = ACTIONS(5381),
    [anon_sym_CG_INLINE] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5381),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5381),
    [anon_sym_IBOutlet] = ACTIONS(5381),
    [anon_sym_IBInspectable] = ACTIONS(5381),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5381),
    [anon_sym_NS_INLINE] = ACTIONS(5381),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5381),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5381),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5381),
    [anon_sym_const] = ACTIONS(5381),
    [anon_sym_constexpr] = ACTIONS(5381),
    [anon_sym_volatile] = ACTIONS(5381),
    [anon_sym_restrict] = ACTIONS(5381),
    [anon_sym___restrict__] = ACTIONS(5381),
    [anon_sym__Atomic] = ACTIONS(5381),
    [anon_sym__Noreturn] = ACTIONS(5381),
    [anon_sym__Nonnull] = ACTIONS(5381),
    [anon_sym_nullable] = ACTIONS(5381),
    [anon_sym__Complex] = ACTIONS(5381),
    [anon_sym__Nullable] = ACTIONS(5381),
    [anon_sym__Nullable_result] = ACTIONS(5381),
    [anon_sym__Null_unspecified] = ACTIONS(5381),
    [anon_sym___autoreleasing] = ACTIONS(5381),
    [anon_sym___block] = ACTIONS(5381),
    [anon_sym___bridge] = ACTIONS(5381),
    [anon_sym___bridge_retained] = ACTIONS(5381),
    [anon_sym___bridge_transfer] = ACTIONS(5381),
    [anon_sym___complex] = ACTIONS(5381),
    [anon_sym___const] = ACTIONS(5381),
    [anon_sym___imag] = ACTIONS(5381),
    [anon_sym___kindof] = ACTIONS(5381),
    [anon_sym___nonnull] = ACTIONS(5381),
    [anon_sym___nullable] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5381),
    [anon_sym___real] = ACTIONS(5381),
    [anon_sym___strong] = ACTIONS(5381),
    [anon_sym___unsafe_unretained] = ACTIONS(5381),
    [anon_sym___unused] = ACTIONS(5381),
    [anon_sym___weak] = ACTIONS(5381),
    [anon_sym_alignas] = ACTIONS(5381),
    [anon_sym__Alignas] = ACTIONS(5381),
    [sym_primitive_type] = ACTIONS(5381),
    [anon_sym_enum] = ACTIONS(5381),
    [anon_sym_struct] = ACTIONS(5381),
    [anon_sym_union] = ACTIONS(5381),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5381),
    [anon_sym___typeof] = ACTIONS(5381),
    [anon_sym_typeof] = ACTIONS(5381),
    [aux_sym_preproc_undef_token1] = ACTIONS(5381),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5381),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5381),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5381),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_API_AVAILABLE] = ACTIONS(5381),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_API_DEPRECATED] = ACTIONS(5381),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym___deprecated_msg] = ACTIONS(5381),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5381),
    [anon_sym_ATend] = ACTIONS(5383),
    [anon_sym_ATsynthesize] = ACTIONS(5383),
    [anon_sym_ATdynamic] = ACTIONS(5383),
    [anon_sym_BOOL] = ACTIONS(5381),
    [anon_sym_IMP] = ACTIONS(5381),
    [anon_sym_SEL] = ACTIONS(5381),
    [anon_sym_Class] = ACTIONS(5381),
    [anon_sym_id] = ACTIONS(5381),
  },
  [STATE(2290)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2321),
    [sym_identifier] = ACTIONS(5462),
    [aux_sym_preproc_def_token1] = ACTIONS(5462),
    [aux_sym_preproc_if_token1] = ACTIONS(5462),
    [aux_sym_preproc_if_token2] = ACTIONS(5462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5462),
    [sym_preproc_directive] = ACTIONS(5462),
    [anon_sym_DASH] = ACTIONS(5464),
    [anon_sym_PLUS] = ACTIONS(5464),
    [anon_sym_SEMI] = ACTIONS(5565),
    [anon_sym___extension__] = ACTIONS(5462),
    [anon_sym_typedef] = ACTIONS(5462),
    [anon_sym_extern] = ACTIONS(5462),
    [anon_sym___attribute__] = ACTIONS(5462),
    [anon_sym___attribute] = ACTIONS(5462),
    [anon_sym_noreturn] = ACTIONS(5462),
    [anon_sym_LBRACK] = ACTIONS(5464),
    [anon_sym___declspec] = ACTIONS(5462),
    [anon_sym___cdecl] = ACTIONS(5462),
    [anon_sym___clrcall] = ACTIONS(5462),
    [anon_sym___stdcall] = ACTIONS(5462),
    [anon_sym___fastcall] = ACTIONS(5462),
    [anon_sym___thiscall] = ACTIONS(5462),
    [anon_sym___vectorcall] = ACTIONS(5462),
    [anon_sym_signed] = ACTIONS(5462),
    [anon_sym_unsigned] = ACTIONS(5462),
    [anon_sym_long] = ACTIONS(5462),
    [anon_sym_short] = ACTIONS(5462),
    [anon_sym_static] = ACTIONS(5462),
    [anon_sym_auto] = ACTIONS(5462),
    [anon_sym_register] = ACTIONS(5462),
    [anon_sym_inline] = ACTIONS(5462),
    [anon_sym___inline] = ACTIONS(5462),
    [anon_sym___inline__] = ACTIONS(5462),
    [anon_sym___forceinline] = ACTIONS(5462),
    [anon_sym_thread_local] = ACTIONS(5462),
    [anon_sym___thread] = ACTIONS(5462),
    [anon_sym_CG_EXTERN] = ACTIONS(5462),
    [anon_sym_CG_INLINE] = ACTIONS(5462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5462),
    [anon_sym_IBOutlet] = ACTIONS(5462),
    [anon_sym_IBInspectable] = ACTIONS(5462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5462),
    [anon_sym_NS_INLINE] = ACTIONS(5462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5462),
    [anon_sym_const] = ACTIONS(5462),
    [anon_sym_constexpr] = ACTIONS(5462),
    [anon_sym_volatile] = ACTIONS(5462),
    [anon_sym_restrict] = ACTIONS(5462),
    [anon_sym___restrict__] = ACTIONS(5462),
    [anon_sym__Atomic] = ACTIONS(5462),
    [anon_sym__Noreturn] = ACTIONS(5462),
    [anon_sym__Nonnull] = ACTIONS(5462),
    [anon_sym_nullable] = ACTIONS(5462),
    [anon_sym__Complex] = ACTIONS(5462),
    [anon_sym__Nullable] = ACTIONS(5462),
    [anon_sym__Nullable_result] = ACTIONS(5462),
    [anon_sym__Null_unspecified] = ACTIONS(5462),
    [anon_sym___autoreleasing] = ACTIONS(5462),
    [anon_sym___block] = ACTIONS(5462),
    [anon_sym___bridge] = ACTIONS(5462),
    [anon_sym___bridge_retained] = ACTIONS(5462),
    [anon_sym___bridge_transfer] = ACTIONS(5462),
    [anon_sym___complex] = ACTIONS(5462),
    [anon_sym___const] = ACTIONS(5462),
    [anon_sym___imag] = ACTIONS(5462),
    [anon_sym___kindof] = ACTIONS(5462),
    [anon_sym___nonnull] = ACTIONS(5462),
    [anon_sym___nullable] = ACTIONS(5462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5462),
    [anon_sym___real] = ACTIONS(5462),
    [anon_sym___strong] = ACTIONS(5462),
    [anon_sym___unsafe_unretained] = ACTIONS(5462),
    [anon_sym___unused] = ACTIONS(5462),
    [anon_sym___weak] = ACTIONS(5462),
    [anon_sym_alignas] = ACTIONS(5462),
    [anon_sym__Alignas] = ACTIONS(5462),
    [sym_primitive_type] = ACTIONS(5462),
    [anon_sym_enum] = ACTIONS(5462),
    [anon_sym_struct] = ACTIONS(5462),
    [anon_sym_union] = ACTIONS(5462),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5462),
    [anon_sym___typeof] = ACTIONS(5462),
    [anon_sym_typeof] = ACTIONS(5462),
    [aux_sym_preproc_undef_token1] = ACTIONS(5462),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5462),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_API_AVAILABLE] = ACTIONS(5462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_API_DEPRECATED] = ACTIONS(5462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5462),
    [anon_sym___deprecated_msg] = ACTIONS(5462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5462),
    [anon_sym_ATproperty] = ACTIONS(5464),
    [anon_sym_BOOL] = ACTIONS(5462),
    [anon_sym_IMP] = ACTIONS(5462),
    [anon_sym_SEL] = ACTIONS(5462),
    [anon_sym_Class] = ACTIONS(5462),
    [anon_sym_id] = ACTIONS(5462),
  },
  [STATE(2291)] = {
    [sym_identifier] = ACTIONS(5573),
    [aux_sym_preproc_def_token1] = ACTIONS(5573),
    [aux_sym_preproc_if_token1] = ACTIONS(5573),
    [aux_sym_preproc_if_token2] = ACTIONS(5573),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5573),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5573),
    [aux_sym_preproc_else_token1] = ACTIONS(5573),
    [aux_sym_preproc_elif_token1] = ACTIONS(5573),
    [sym_preproc_directive] = ACTIONS(5573),
    [anon_sym_DASH] = ACTIONS(5575),
    [anon_sym_PLUS] = ACTIONS(5575),
    [anon_sym___extension__] = ACTIONS(5573),
    [anon_sym_typedef] = ACTIONS(5573),
    [anon_sym_extern] = ACTIONS(5573),
    [anon_sym___attribute__] = ACTIONS(5573),
    [anon_sym___attribute] = ACTIONS(5573),
    [anon_sym_noreturn] = ACTIONS(5573),
    [anon_sym_LBRACK] = ACTIONS(5575),
    [anon_sym___declspec] = ACTIONS(5573),
    [anon_sym___cdecl] = ACTIONS(5573),
    [anon_sym___clrcall] = ACTIONS(5573),
    [anon_sym___stdcall] = ACTIONS(5573),
    [anon_sym___fastcall] = ACTIONS(5573),
    [anon_sym___thiscall] = ACTIONS(5573),
    [anon_sym___vectorcall] = ACTIONS(5573),
    [anon_sym_signed] = ACTIONS(5573),
    [anon_sym_unsigned] = ACTIONS(5573),
    [anon_sym_long] = ACTIONS(5573),
    [anon_sym_short] = ACTIONS(5573),
    [anon_sym_static] = ACTIONS(5573),
    [anon_sym_auto] = ACTIONS(5573),
    [anon_sym_register] = ACTIONS(5573),
    [anon_sym_inline] = ACTIONS(5573),
    [anon_sym___inline] = ACTIONS(5573),
    [anon_sym___inline__] = ACTIONS(5573),
    [anon_sym___forceinline] = ACTIONS(5573),
    [anon_sym_thread_local] = ACTIONS(5573),
    [anon_sym___thread] = ACTIONS(5573),
    [anon_sym_CG_EXTERN] = ACTIONS(5573),
    [anon_sym_CG_INLINE] = ACTIONS(5573),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5573),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5573),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5573),
    [anon_sym_IBOutlet] = ACTIONS(5573),
    [anon_sym_IBInspectable] = ACTIONS(5573),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5573),
    [anon_sym_NS_INLINE] = ACTIONS(5573),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5573),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5573),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5573),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5573),
    [anon_sym_const] = ACTIONS(5573),
    [anon_sym_constexpr] = ACTIONS(5573),
    [anon_sym_volatile] = ACTIONS(5573),
    [anon_sym_restrict] = ACTIONS(5573),
    [anon_sym___restrict__] = ACTIONS(5573),
    [anon_sym__Atomic] = ACTIONS(5573),
    [anon_sym__Noreturn] = ACTIONS(5573),
    [anon_sym__Nonnull] = ACTIONS(5573),
    [anon_sym_nullable] = ACTIONS(5573),
    [anon_sym__Complex] = ACTIONS(5573),
    [anon_sym__Nullable] = ACTIONS(5573),
    [anon_sym__Nullable_result] = ACTIONS(5573),
    [anon_sym__Null_unspecified] = ACTIONS(5573),
    [anon_sym___autoreleasing] = ACTIONS(5573),
    [anon_sym___block] = ACTIONS(5573),
    [anon_sym___bridge] = ACTIONS(5573),
    [anon_sym___bridge_retained] = ACTIONS(5573),
    [anon_sym___bridge_transfer] = ACTIONS(5573),
    [anon_sym___complex] = ACTIONS(5573),
    [anon_sym___const] = ACTIONS(5573),
    [anon_sym___imag] = ACTIONS(5573),
    [anon_sym___kindof] = ACTIONS(5573),
    [anon_sym___nonnull] = ACTIONS(5573),
    [anon_sym___nullable] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5573),
    [anon_sym___real] = ACTIONS(5573),
    [anon_sym___strong] = ACTIONS(5573),
    [anon_sym___unsafe_unretained] = ACTIONS(5573),
    [anon_sym___unused] = ACTIONS(5573),
    [anon_sym___weak] = ACTIONS(5573),
    [anon_sym_alignas] = ACTIONS(5573),
    [anon_sym__Alignas] = ACTIONS(5573),
    [sym_primitive_type] = ACTIONS(5573),
    [anon_sym_enum] = ACTIONS(5573),
    [anon_sym_struct] = ACTIONS(5573),
    [anon_sym_union] = ACTIONS(5573),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5573),
    [anon_sym___typeof] = ACTIONS(5573),
    [anon_sym_typeof] = ACTIONS(5573),
    [aux_sym_preproc_undef_token1] = ACTIONS(5573),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5573),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5573),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5573),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5573),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5573),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5573),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5573),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5573),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5573),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_API_AVAILABLE] = ACTIONS(5573),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_API_DEPRECATED] = ACTIONS(5573),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5573),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5573),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5573),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5573),
    [anon_sym___deprecated_msg] = ACTIONS(5573),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5573),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5573),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5573),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5573),
    [anon_sym_ATproperty] = ACTIONS(5575),
    [anon_sym_BOOL] = ACTIONS(5573),
    [anon_sym_IMP] = ACTIONS(5573),
    [anon_sym_SEL] = ACTIONS(5573),
    [anon_sym_Class] = ACTIONS(5573),
    [anon_sym_id] = ACTIONS(5573),
  },
  [STATE(2292)] = {
    [sym_identifier] = ACTIONS(2618),
    [aux_sym_preproc_def_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token2] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2618),
    [sym_preproc_directive] = ACTIONS(2618),
    [anon_sym_DASH] = ACTIONS(2620),
    [anon_sym_PLUS] = ACTIONS(2620),
    [anon_sym___extension__] = ACTIONS(2618),
    [anon_sym_typedef] = ACTIONS(2618),
    [anon_sym_extern] = ACTIONS(2618),
    [anon_sym___attribute__] = ACTIONS(2618),
    [anon_sym___attribute] = ACTIONS(2618),
    [anon_sym_noreturn] = ACTIONS(2618),
    [anon_sym_LBRACK] = ACTIONS(2620),
    [anon_sym___declspec] = ACTIONS(2618),
    [anon_sym___cdecl] = ACTIONS(2618),
    [anon_sym___clrcall] = ACTIONS(2618),
    [anon_sym___stdcall] = ACTIONS(2618),
    [anon_sym___fastcall] = ACTIONS(2618),
    [anon_sym___thiscall] = ACTIONS(2618),
    [anon_sym___vectorcall] = ACTIONS(2618),
    [anon_sym_signed] = ACTIONS(2618),
    [anon_sym_unsigned] = ACTIONS(2618),
    [anon_sym_long] = ACTIONS(2618),
    [anon_sym_short] = ACTIONS(2618),
    [anon_sym_static] = ACTIONS(2618),
    [anon_sym_auto] = ACTIONS(2618),
    [anon_sym_register] = ACTIONS(2618),
    [anon_sym_inline] = ACTIONS(2618),
    [anon_sym___inline] = ACTIONS(2618),
    [anon_sym___inline__] = ACTIONS(2618),
    [anon_sym___forceinline] = ACTIONS(2618),
    [anon_sym_thread_local] = ACTIONS(2618),
    [anon_sym___thread] = ACTIONS(2618),
    [anon_sym_CG_EXTERN] = ACTIONS(2618),
    [anon_sym_CG_INLINE] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2618),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2618),
    [anon_sym_IBOutlet] = ACTIONS(2618),
    [anon_sym_IBInspectable] = ACTIONS(2618),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2618),
    [anon_sym_NS_INLINE] = ACTIONS(2618),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2618),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2618),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2618),
    [anon_sym_const] = ACTIONS(2618),
    [anon_sym_constexpr] = ACTIONS(2618),
    [anon_sym_volatile] = ACTIONS(2618),
    [anon_sym_restrict] = ACTIONS(2618),
    [anon_sym___restrict__] = ACTIONS(2618),
    [anon_sym__Atomic] = ACTIONS(2618),
    [anon_sym__Noreturn] = ACTIONS(2618),
    [anon_sym__Nonnull] = ACTIONS(2618),
    [anon_sym_nullable] = ACTIONS(2618),
    [anon_sym__Complex] = ACTIONS(2618),
    [anon_sym__Nullable] = ACTIONS(2618),
    [anon_sym__Nullable_result] = ACTIONS(2618),
    [anon_sym__Null_unspecified] = ACTIONS(2618),
    [anon_sym___autoreleasing] = ACTIONS(2618),
    [anon_sym___block] = ACTIONS(2618),
    [anon_sym___bridge] = ACTIONS(2618),
    [anon_sym___bridge_retained] = ACTIONS(2618),
    [anon_sym___bridge_transfer] = ACTIONS(2618),
    [anon_sym___complex] = ACTIONS(2618),
    [anon_sym___const] = ACTIONS(2618),
    [anon_sym___imag] = ACTIONS(2618),
    [anon_sym___kindof] = ACTIONS(2618),
    [anon_sym___nonnull] = ACTIONS(2618),
    [anon_sym___nullable] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2618),
    [anon_sym___real] = ACTIONS(2618),
    [anon_sym___strong] = ACTIONS(2618),
    [anon_sym___unsafe_unretained] = ACTIONS(2618),
    [anon_sym___unused] = ACTIONS(2618),
    [anon_sym___weak] = ACTIONS(2618),
    [anon_sym_alignas] = ACTIONS(2618),
    [anon_sym__Alignas] = ACTIONS(2618),
    [sym_primitive_type] = ACTIONS(2618),
    [anon_sym_enum] = ACTIONS(2618),
    [anon_sym_struct] = ACTIONS(2618),
    [anon_sym_union] = ACTIONS(2618),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2618),
    [anon_sym___typeof] = ACTIONS(2618),
    [anon_sym_typeof] = ACTIONS(2618),
    [aux_sym_preproc_undef_token1] = ACTIONS(2618),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2618),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2618),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2618),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_API_AVAILABLE] = ACTIONS(2618),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_API_DEPRECATED] = ACTIONS(2618),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym___deprecated_msg] = ACTIONS(2618),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2618),
    [anon_sym_ATsynthesize] = ACTIONS(2620),
    [anon_sym_ATdynamic] = ACTIONS(2620),
    [anon_sym_BOOL] = ACTIONS(2618),
    [anon_sym_IMP] = ACTIONS(2618),
    [anon_sym_SEL] = ACTIONS(2618),
    [anon_sym_Class] = ACTIONS(2618),
    [anon_sym_id] = ACTIONS(2618),
    [anon_sym_ATdefs] = ACTIONS(2620),
  },
  [STATE(2293)] = {
    [sym_identifier] = ACTIONS(2634),
    [aux_sym_preproc_def_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token1] = ACTIONS(2634),
    [aux_sym_preproc_if_token2] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2634),
    [sym_preproc_directive] = ACTIONS(2634),
    [anon_sym_DASH] = ACTIONS(2636),
    [anon_sym_PLUS] = ACTIONS(2636),
    [anon_sym___extension__] = ACTIONS(2634),
    [anon_sym_typedef] = ACTIONS(2634),
    [anon_sym_extern] = ACTIONS(2634),
    [anon_sym___attribute__] = ACTIONS(2634),
    [anon_sym___attribute] = ACTIONS(2634),
    [anon_sym_noreturn] = ACTIONS(2634),
    [anon_sym_LBRACK] = ACTIONS(2636),
    [anon_sym___declspec] = ACTIONS(2634),
    [anon_sym___cdecl] = ACTIONS(2634),
    [anon_sym___clrcall] = ACTIONS(2634),
    [anon_sym___stdcall] = ACTIONS(2634),
    [anon_sym___fastcall] = ACTIONS(2634),
    [anon_sym___thiscall] = ACTIONS(2634),
    [anon_sym___vectorcall] = ACTIONS(2634),
    [anon_sym_signed] = ACTIONS(2634),
    [anon_sym_unsigned] = ACTIONS(2634),
    [anon_sym_long] = ACTIONS(2634),
    [anon_sym_short] = ACTIONS(2634),
    [anon_sym_static] = ACTIONS(2634),
    [anon_sym_auto] = ACTIONS(2634),
    [anon_sym_register] = ACTIONS(2634),
    [anon_sym_inline] = ACTIONS(2634),
    [anon_sym___inline] = ACTIONS(2634),
    [anon_sym___inline__] = ACTIONS(2634),
    [anon_sym___forceinline] = ACTIONS(2634),
    [anon_sym_thread_local] = ACTIONS(2634),
    [anon_sym___thread] = ACTIONS(2634),
    [anon_sym_CG_EXTERN] = ACTIONS(2634),
    [anon_sym_CG_INLINE] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2634),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2634),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2634),
    [anon_sym_IBOutlet] = ACTIONS(2634),
    [anon_sym_IBInspectable] = ACTIONS(2634),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2634),
    [anon_sym_NS_INLINE] = ACTIONS(2634),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2634),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2634),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2634),
    [anon_sym_const] = ACTIONS(2634),
    [anon_sym_constexpr] = ACTIONS(2634),
    [anon_sym_volatile] = ACTIONS(2634),
    [anon_sym_restrict] = ACTIONS(2634),
    [anon_sym___restrict__] = ACTIONS(2634),
    [anon_sym__Atomic] = ACTIONS(2634),
    [anon_sym__Noreturn] = ACTIONS(2634),
    [anon_sym__Nonnull] = ACTIONS(2634),
    [anon_sym_nullable] = ACTIONS(2634),
    [anon_sym__Complex] = ACTIONS(2634),
    [anon_sym__Nullable] = ACTIONS(2634),
    [anon_sym__Nullable_result] = ACTIONS(2634),
    [anon_sym__Null_unspecified] = ACTIONS(2634),
    [anon_sym___autoreleasing] = ACTIONS(2634),
    [anon_sym___block] = ACTIONS(2634),
    [anon_sym___bridge] = ACTIONS(2634),
    [anon_sym___bridge_retained] = ACTIONS(2634),
    [anon_sym___bridge_transfer] = ACTIONS(2634),
    [anon_sym___complex] = ACTIONS(2634),
    [anon_sym___const] = ACTIONS(2634),
    [anon_sym___imag] = ACTIONS(2634),
    [anon_sym___kindof] = ACTIONS(2634),
    [anon_sym___nonnull] = ACTIONS(2634),
    [anon_sym___nullable] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2634),
    [anon_sym___real] = ACTIONS(2634),
    [anon_sym___strong] = ACTIONS(2634),
    [anon_sym___unsafe_unretained] = ACTIONS(2634),
    [anon_sym___unused] = ACTIONS(2634),
    [anon_sym___weak] = ACTIONS(2634),
    [anon_sym_alignas] = ACTIONS(2634),
    [anon_sym__Alignas] = ACTIONS(2634),
    [sym_primitive_type] = ACTIONS(2634),
    [anon_sym_enum] = ACTIONS(2634),
    [anon_sym_struct] = ACTIONS(2634),
    [anon_sym_union] = ACTIONS(2634),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2634),
    [anon_sym___typeof] = ACTIONS(2634),
    [anon_sym_typeof] = ACTIONS(2634),
    [aux_sym_preproc_undef_token1] = ACTIONS(2634),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2634),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2634),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2634),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2634),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2634),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_API_AVAILABLE] = ACTIONS(2634),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_API_DEPRECATED] = ACTIONS(2634),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2634),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2634),
    [anon_sym___deprecated_msg] = ACTIONS(2634),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2634),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2634),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2634),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2634),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2634),
    [anon_sym_ATsynthesize] = ACTIONS(2636),
    [anon_sym_ATdynamic] = ACTIONS(2636),
    [anon_sym_ATproperty] = ACTIONS(2636),
    [anon_sym_BOOL] = ACTIONS(2634),
    [anon_sym_IMP] = ACTIONS(2634),
    [anon_sym_SEL] = ACTIONS(2634),
    [anon_sym_Class] = ACTIONS(2634),
    [anon_sym_id] = ACTIONS(2634),
  },
  [STATE(2294)] = {
    [sym_identifier] = ACTIONS(2668),
    [aux_sym_preproc_def_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token1] = ACTIONS(2668),
    [aux_sym_preproc_if_token2] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2668),
    [sym_preproc_directive] = ACTIONS(2668),
    [anon_sym_DASH] = ACTIONS(2670),
    [anon_sym_PLUS] = ACTIONS(2670),
    [anon_sym___extension__] = ACTIONS(2668),
    [anon_sym_typedef] = ACTIONS(2668),
    [anon_sym_extern] = ACTIONS(2668),
    [anon_sym___attribute__] = ACTIONS(2668),
    [anon_sym___attribute] = ACTIONS(2668),
    [anon_sym_noreturn] = ACTIONS(2668),
    [anon_sym_LBRACK] = ACTIONS(2670),
    [anon_sym___declspec] = ACTIONS(2668),
    [anon_sym___cdecl] = ACTIONS(2668),
    [anon_sym___clrcall] = ACTIONS(2668),
    [anon_sym___stdcall] = ACTIONS(2668),
    [anon_sym___fastcall] = ACTIONS(2668),
    [anon_sym___thiscall] = ACTIONS(2668),
    [anon_sym___vectorcall] = ACTIONS(2668),
    [anon_sym_signed] = ACTIONS(2668),
    [anon_sym_unsigned] = ACTIONS(2668),
    [anon_sym_long] = ACTIONS(2668),
    [anon_sym_short] = ACTIONS(2668),
    [anon_sym_static] = ACTIONS(2668),
    [anon_sym_auto] = ACTIONS(2668),
    [anon_sym_register] = ACTIONS(2668),
    [anon_sym_inline] = ACTIONS(2668),
    [anon_sym___inline] = ACTIONS(2668),
    [anon_sym___inline__] = ACTIONS(2668),
    [anon_sym___forceinline] = ACTIONS(2668),
    [anon_sym_thread_local] = ACTIONS(2668),
    [anon_sym___thread] = ACTIONS(2668),
    [anon_sym_CG_EXTERN] = ACTIONS(2668),
    [anon_sym_CG_INLINE] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2668),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2668),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2668),
    [anon_sym_IBOutlet] = ACTIONS(2668),
    [anon_sym_IBInspectable] = ACTIONS(2668),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2668),
    [anon_sym_NS_INLINE] = ACTIONS(2668),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2668),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2668),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2668),
    [anon_sym_const] = ACTIONS(2668),
    [anon_sym_constexpr] = ACTIONS(2668),
    [anon_sym_volatile] = ACTIONS(2668),
    [anon_sym_restrict] = ACTIONS(2668),
    [anon_sym___restrict__] = ACTIONS(2668),
    [anon_sym__Atomic] = ACTIONS(2668),
    [anon_sym__Noreturn] = ACTIONS(2668),
    [anon_sym__Nonnull] = ACTIONS(2668),
    [anon_sym_nullable] = ACTIONS(2668),
    [anon_sym__Complex] = ACTIONS(2668),
    [anon_sym__Nullable] = ACTIONS(2668),
    [anon_sym__Nullable_result] = ACTIONS(2668),
    [anon_sym__Null_unspecified] = ACTIONS(2668),
    [anon_sym___autoreleasing] = ACTIONS(2668),
    [anon_sym___block] = ACTIONS(2668),
    [anon_sym___bridge] = ACTIONS(2668),
    [anon_sym___bridge_retained] = ACTIONS(2668),
    [anon_sym___bridge_transfer] = ACTIONS(2668),
    [anon_sym___complex] = ACTIONS(2668),
    [anon_sym___const] = ACTIONS(2668),
    [anon_sym___imag] = ACTIONS(2668),
    [anon_sym___kindof] = ACTIONS(2668),
    [anon_sym___nonnull] = ACTIONS(2668),
    [anon_sym___nullable] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2668),
    [anon_sym___real] = ACTIONS(2668),
    [anon_sym___strong] = ACTIONS(2668),
    [anon_sym___unsafe_unretained] = ACTIONS(2668),
    [anon_sym___unused] = ACTIONS(2668),
    [anon_sym___weak] = ACTIONS(2668),
    [anon_sym_alignas] = ACTIONS(2668),
    [anon_sym__Alignas] = ACTIONS(2668),
    [sym_primitive_type] = ACTIONS(2668),
    [anon_sym_enum] = ACTIONS(2668),
    [anon_sym_struct] = ACTIONS(2668),
    [anon_sym_union] = ACTIONS(2668),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2668),
    [anon_sym___typeof] = ACTIONS(2668),
    [anon_sym_typeof] = ACTIONS(2668),
    [aux_sym_preproc_undef_token1] = ACTIONS(2668),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2668),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2668),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2668),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2668),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2668),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_API_AVAILABLE] = ACTIONS(2668),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_API_DEPRECATED] = ACTIONS(2668),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2668),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2668),
    [anon_sym___deprecated_msg] = ACTIONS(2668),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2668),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2668),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2668),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2668),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2668),
    [anon_sym_ATsynthesize] = ACTIONS(2670),
    [anon_sym_ATdynamic] = ACTIONS(2670),
    [anon_sym_ATproperty] = ACTIONS(2670),
    [anon_sym_BOOL] = ACTIONS(2668),
    [anon_sym_IMP] = ACTIONS(2668),
    [anon_sym_SEL] = ACTIONS(2668),
    [anon_sym_Class] = ACTIONS(2668),
    [anon_sym_id] = ACTIONS(2668),
  },
  [STATE(2295)] = {
    [sym_identifier] = ACTIONS(5393),
    [aux_sym_preproc_def_token1] = ACTIONS(5393),
    [aux_sym_preproc_if_token1] = ACTIONS(5393),
    [aux_sym_preproc_if_token2] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5393),
    [sym_preproc_directive] = ACTIONS(5393),
    [anon_sym_DASH] = ACTIONS(5395),
    [anon_sym_PLUS] = ACTIONS(5395),
    [anon_sym_SEMI] = ACTIONS(5601),
    [anon_sym___extension__] = ACTIONS(5393),
    [anon_sym_typedef] = ACTIONS(5393),
    [anon_sym_extern] = ACTIONS(5393),
    [anon_sym___attribute__] = ACTIONS(5393),
    [anon_sym___attribute] = ACTIONS(5393),
    [anon_sym_noreturn] = ACTIONS(5393),
    [anon_sym_LBRACK] = ACTIONS(5395),
    [anon_sym___declspec] = ACTIONS(5393),
    [anon_sym___cdecl] = ACTIONS(5393),
    [anon_sym___clrcall] = ACTIONS(5393),
    [anon_sym___stdcall] = ACTIONS(5393),
    [anon_sym___fastcall] = ACTIONS(5393),
    [anon_sym___thiscall] = ACTIONS(5393),
    [anon_sym___vectorcall] = ACTIONS(5393),
    [anon_sym_signed] = ACTIONS(5393),
    [anon_sym_unsigned] = ACTIONS(5393),
    [anon_sym_long] = ACTIONS(5393),
    [anon_sym_short] = ACTIONS(5393),
    [anon_sym_static] = ACTIONS(5393),
    [anon_sym_auto] = ACTIONS(5393),
    [anon_sym_register] = ACTIONS(5393),
    [anon_sym_inline] = ACTIONS(5393),
    [anon_sym___inline] = ACTIONS(5393),
    [anon_sym___inline__] = ACTIONS(5393),
    [anon_sym___forceinline] = ACTIONS(5393),
    [anon_sym_thread_local] = ACTIONS(5393),
    [anon_sym___thread] = ACTIONS(5393),
    [anon_sym_CG_EXTERN] = ACTIONS(5393),
    [anon_sym_CG_INLINE] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5393),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5393),
    [anon_sym_IBOutlet] = ACTIONS(5393),
    [anon_sym_IBInspectable] = ACTIONS(5393),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5393),
    [anon_sym_NS_INLINE] = ACTIONS(5393),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5393),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5393),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5393),
    [anon_sym_const] = ACTIONS(5393),
    [anon_sym_constexpr] = ACTIONS(5393),
    [anon_sym_volatile] = ACTIONS(5393),
    [anon_sym_restrict] = ACTIONS(5393),
    [anon_sym___restrict__] = ACTIONS(5393),
    [anon_sym__Atomic] = ACTIONS(5393),
    [anon_sym__Noreturn] = ACTIONS(5393),
    [anon_sym__Nonnull] = ACTIONS(5393),
    [anon_sym_nullable] = ACTIONS(5393),
    [anon_sym__Complex] = ACTIONS(5393),
    [anon_sym__Nullable] = ACTIONS(5393),
    [anon_sym__Nullable_result] = ACTIONS(5393),
    [anon_sym__Null_unspecified] = ACTIONS(5393),
    [anon_sym___autoreleasing] = ACTIONS(5393),
    [anon_sym___block] = ACTIONS(5393),
    [anon_sym___bridge] = ACTIONS(5393),
    [anon_sym___bridge_retained] = ACTIONS(5393),
    [anon_sym___bridge_transfer] = ACTIONS(5393),
    [anon_sym___complex] = ACTIONS(5393),
    [anon_sym___const] = ACTIONS(5393),
    [anon_sym___imag] = ACTIONS(5393),
    [anon_sym___kindof] = ACTIONS(5393),
    [anon_sym___nonnull] = ACTIONS(5393),
    [anon_sym___nullable] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5393),
    [anon_sym___real] = ACTIONS(5393),
    [anon_sym___strong] = ACTIONS(5393),
    [anon_sym___unsafe_unretained] = ACTIONS(5393),
    [anon_sym___unused] = ACTIONS(5393),
    [anon_sym___weak] = ACTIONS(5393),
    [anon_sym_alignas] = ACTIONS(5393),
    [anon_sym__Alignas] = ACTIONS(5393),
    [sym_primitive_type] = ACTIONS(5393),
    [anon_sym_enum] = ACTIONS(5393),
    [anon_sym_struct] = ACTIONS(5393),
    [anon_sym_union] = ACTIONS(5393),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5393),
    [anon_sym___typeof] = ACTIONS(5393),
    [anon_sym_typeof] = ACTIONS(5393),
    [aux_sym_preproc_undef_token1] = ACTIONS(5393),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5393),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5393),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5393),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_API_AVAILABLE] = ACTIONS(5393),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_API_DEPRECATED] = ACTIONS(5393),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym___deprecated_msg] = ACTIONS(5393),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5393),
    [anon_sym_ATsynthesize] = ACTIONS(5395),
    [anon_sym_ATdynamic] = ACTIONS(5395),
    [anon_sym_BOOL] = ACTIONS(5393),
    [anon_sym_IMP] = ACTIONS(5393),
    [anon_sym_SEL] = ACTIONS(5393),
    [anon_sym_Class] = ACTIONS(5393),
    [anon_sym_id] = ACTIONS(5393),
  },
  [STATE(2296)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2321),
    [sym_identifier] = ACTIONS(5345),
    [aux_sym_preproc_def_token1] = ACTIONS(5345),
    [aux_sym_preproc_if_token1] = ACTIONS(5345),
    [aux_sym_preproc_if_token2] = ACTIONS(5345),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5345),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5345),
    [sym_preproc_directive] = ACTIONS(5345),
    [anon_sym_DASH] = ACTIONS(5347),
    [anon_sym_PLUS] = ACTIONS(5347),
    [anon_sym_SEMI] = ACTIONS(5565),
    [anon_sym___extension__] = ACTIONS(5345),
    [anon_sym_typedef] = ACTIONS(5345),
    [anon_sym_extern] = ACTIONS(5345),
    [anon_sym___attribute__] = ACTIONS(5345),
    [anon_sym___attribute] = ACTIONS(5345),
    [anon_sym_noreturn] = ACTIONS(5345),
    [anon_sym_LBRACK] = ACTIONS(5347),
    [anon_sym___declspec] = ACTIONS(5345),
    [anon_sym___cdecl] = ACTIONS(5345),
    [anon_sym___clrcall] = ACTIONS(5345),
    [anon_sym___stdcall] = ACTIONS(5345),
    [anon_sym___fastcall] = ACTIONS(5345),
    [anon_sym___thiscall] = ACTIONS(5345),
    [anon_sym___vectorcall] = ACTIONS(5345),
    [anon_sym_signed] = ACTIONS(5345),
    [anon_sym_unsigned] = ACTIONS(5345),
    [anon_sym_long] = ACTIONS(5345),
    [anon_sym_short] = ACTIONS(5345),
    [anon_sym_static] = ACTIONS(5345),
    [anon_sym_auto] = ACTIONS(5345),
    [anon_sym_register] = ACTIONS(5345),
    [anon_sym_inline] = ACTIONS(5345),
    [anon_sym___inline] = ACTIONS(5345),
    [anon_sym___inline__] = ACTIONS(5345),
    [anon_sym___forceinline] = ACTIONS(5345),
    [anon_sym_thread_local] = ACTIONS(5345),
    [anon_sym___thread] = ACTIONS(5345),
    [anon_sym_CG_EXTERN] = ACTIONS(5345),
    [anon_sym_CG_INLINE] = ACTIONS(5345),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5345),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5345),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5345),
    [anon_sym_IBOutlet] = ACTIONS(5345),
    [anon_sym_IBInspectable] = ACTIONS(5345),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5345),
    [anon_sym_NS_INLINE] = ACTIONS(5345),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5345),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5345),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5345),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5345),
    [anon_sym_const] = ACTIONS(5345),
    [anon_sym_constexpr] = ACTIONS(5345),
    [anon_sym_volatile] = ACTIONS(5345),
    [anon_sym_restrict] = ACTIONS(5345),
    [anon_sym___restrict__] = ACTIONS(5345),
    [anon_sym__Atomic] = ACTIONS(5345),
    [anon_sym__Noreturn] = ACTIONS(5345),
    [anon_sym__Nonnull] = ACTIONS(5345),
    [anon_sym_nullable] = ACTIONS(5345),
    [anon_sym__Complex] = ACTIONS(5345),
    [anon_sym__Nullable] = ACTIONS(5345),
    [anon_sym__Nullable_result] = ACTIONS(5345),
    [anon_sym__Null_unspecified] = ACTIONS(5345),
    [anon_sym___autoreleasing] = ACTIONS(5345),
    [anon_sym___block] = ACTIONS(5345),
    [anon_sym___bridge] = ACTIONS(5345),
    [anon_sym___bridge_retained] = ACTIONS(5345),
    [anon_sym___bridge_transfer] = ACTIONS(5345),
    [anon_sym___complex] = ACTIONS(5345),
    [anon_sym___const] = ACTIONS(5345),
    [anon_sym___imag] = ACTIONS(5345),
    [anon_sym___kindof] = ACTIONS(5345),
    [anon_sym___nonnull] = ACTIONS(5345),
    [anon_sym___nullable] = ACTIONS(5345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5345),
    [anon_sym___real] = ACTIONS(5345),
    [anon_sym___strong] = ACTIONS(5345),
    [anon_sym___unsafe_unretained] = ACTIONS(5345),
    [anon_sym___unused] = ACTIONS(5345),
    [anon_sym___weak] = ACTIONS(5345),
    [anon_sym_alignas] = ACTIONS(5345),
    [anon_sym__Alignas] = ACTIONS(5345),
    [sym_primitive_type] = ACTIONS(5345),
    [anon_sym_enum] = ACTIONS(5345),
    [anon_sym_struct] = ACTIONS(5345),
    [anon_sym_union] = ACTIONS(5345),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5345),
    [anon_sym___typeof] = ACTIONS(5345),
    [anon_sym_typeof] = ACTIONS(5345),
    [aux_sym_preproc_undef_token1] = ACTIONS(5345),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5345),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5345),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5345),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5345),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5345),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5345),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5345),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5345),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5345),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_API_AVAILABLE] = ACTIONS(5345),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_API_DEPRECATED] = ACTIONS(5345),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5345),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5345),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5345),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5345),
    [anon_sym___deprecated_msg] = ACTIONS(5345),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5345),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5345),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5345),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5345),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5345),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5345),
    [anon_sym_ATproperty] = ACTIONS(5347),
    [anon_sym_BOOL] = ACTIONS(5345),
    [anon_sym_IMP] = ACTIONS(5345),
    [anon_sym_SEL] = ACTIONS(5345),
    [anon_sym_Class] = ACTIONS(5345),
    [anon_sym_id] = ACTIONS(5345),
  },
  [STATE(2297)] = {
    [sym_identifier] = ACTIONS(5475),
    [aux_sym_preproc_def_token1] = ACTIONS(5475),
    [aux_sym_preproc_if_token1] = ACTIONS(5475),
    [aux_sym_preproc_if_token2] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5475),
    [sym_preproc_directive] = ACTIONS(5475),
    [anon_sym_DASH] = ACTIONS(5477),
    [anon_sym_PLUS] = ACTIONS(5477),
    [anon_sym_SEMI] = ACTIONS(5603),
    [anon_sym___extension__] = ACTIONS(5475),
    [anon_sym_typedef] = ACTIONS(5475),
    [anon_sym_extern] = ACTIONS(5475),
    [anon_sym___attribute__] = ACTIONS(5475),
    [anon_sym___attribute] = ACTIONS(5475),
    [anon_sym_noreturn] = ACTIONS(5475),
    [anon_sym_LBRACK] = ACTIONS(5477),
    [anon_sym___declspec] = ACTIONS(5475),
    [anon_sym___cdecl] = ACTIONS(5475),
    [anon_sym___clrcall] = ACTIONS(5475),
    [anon_sym___stdcall] = ACTIONS(5475),
    [anon_sym___fastcall] = ACTIONS(5475),
    [anon_sym___thiscall] = ACTIONS(5475),
    [anon_sym___vectorcall] = ACTIONS(5475),
    [anon_sym_signed] = ACTIONS(5475),
    [anon_sym_unsigned] = ACTIONS(5475),
    [anon_sym_long] = ACTIONS(5475),
    [anon_sym_short] = ACTIONS(5475),
    [anon_sym_static] = ACTIONS(5475),
    [anon_sym_auto] = ACTIONS(5475),
    [anon_sym_register] = ACTIONS(5475),
    [anon_sym_inline] = ACTIONS(5475),
    [anon_sym___inline] = ACTIONS(5475),
    [anon_sym___inline__] = ACTIONS(5475),
    [anon_sym___forceinline] = ACTIONS(5475),
    [anon_sym_thread_local] = ACTIONS(5475),
    [anon_sym___thread] = ACTIONS(5475),
    [anon_sym_CG_EXTERN] = ACTIONS(5475),
    [anon_sym_CG_INLINE] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5475),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5475),
    [anon_sym_IBOutlet] = ACTIONS(5475),
    [anon_sym_IBInspectable] = ACTIONS(5475),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5475),
    [anon_sym_NS_INLINE] = ACTIONS(5475),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5475),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5475),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5475),
    [anon_sym_const] = ACTIONS(5475),
    [anon_sym_constexpr] = ACTIONS(5475),
    [anon_sym_volatile] = ACTIONS(5475),
    [anon_sym_restrict] = ACTIONS(5475),
    [anon_sym___restrict__] = ACTIONS(5475),
    [anon_sym__Atomic] = ACTIONS(5475),
    [anon_sym__Noreturn] = ACTIONS(5475),
    [anon_sym__Nonnull] = ACTIONS(5475),
    [anon_sym_nullable] = ACTIONS(5475),
    [anon_sym__Complex] = ACTIONS(5475),
    [anon_sym__Nullable] = ACTIONS(5475),
    [anon_sym__Nullable_result] = ACTIONS(5475),
    [anon_sym__Null_unspecified] = ACTIONS(5475),
    [anon_sym___autoreleasing] = ACTIONS(5475),
    [anon_sym___block] = ACTIONS(5475),
    [anon_sym___bridge] = ACTIONS(5475),
    [anon_sym___bridge_retained] = ACTIONS(5475),
    [anon_sym___bridge_transfer] = ACTIONS(5475),
    [anon_sym___complex] = ACTIONS(5475),
    [anon_sym___const] = ACTIONS(5475),
    [anon_sym___imag] = ACTIONS(5475),
    [anon_sym___kindof] = ACTIONS(5475),
    [anon_sym___nonnull] = ACTIONS(5475),
    [anon_sym___nullable] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5475),
    [anon_sym___real] = ACTIONS(5475),
    [anon_sym___strong] = ACTIONS(5475),
    [anon_sym___unsafe_unretained] = ACTIONS(5475),
    [anon_sym___unused] = ACTIONS(5475),
    [anon_sym___weak] = ACTIONS(5475),
    [anon_sym_alignas] = ACTIONS(5475),
    [anon_sym__Alignas] = ACTIONS(5475),
    [sym_primitive_type] = ACTIONS(5475),
    [anon_sym_enum] = ACTIONS(5475),
    [anon_sym_struct] = ACTIONS(5475),
    [anon_sym_union] = ACTIONS(5475),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5475),
    [anon_sym___typeof] = ACTIONS(5475),
    [anon_sym_typeof] = ACTIONS(5475),
    [aux_sym_preproc_undef_token1] = ACTIONS(5475),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5475),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5475),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5475),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_API_AVAILABLE] = ACTIONS(5475),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_API_DEPRECATED] = ACTIONS(5475),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym___deprecated_msg] = ACTIONS(5475),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5475),
    [anon_sym_ATsynthesize] = ACTIONS(5477),
    [anon_sym_ATdynamic] = ACTIONS(5477),
    [anon_sym_BOOL] = ACTIONS(5475),
    [anon_sym_IMP] = ACTIONS(5475),
    [anon_sym_SEL] = ACTIONS(5475),
    [anon_sym_Class] = ACTIONS(5475),
    [anon_sym_id] = ACTIONS(5475),
  },
  [STATE(2298)] = {
    [sym_identifier] = ACTIONS(2600),
    [aux_sym_preproc_def_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token2] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2600),
    [sym_preproc_directive] = ACTIONS(2600),
    [anon_sym_DASH] = ACTIONS(2602),
    [anon_sym_PLUS] = ACTIONS(2602),
    [anon_sym___extension__] = ACTIONS(2600),
    [anon_sym_typedef] = ACTIONS(2600),
    [anon_sym_extern] = ACTIONS(2600),
    [anon_sym___attribute__] = ACTIONS(2600),
    [anon_sym___attribute] = ACTIONS(2600),
    [anon_sym_noreturn] = ACTIONS(2600),
    [anon_sym_LBRACK] = ACTIONS(2602),
    [anon_sym___declspec] = ACTIONS(2600),
    [anon_sym___cdecl] = ACTIONS(2600),
    [anon_sym___clrcall] = ACTIONS(2600),
    [anon_sym___stdcall] = ACTIONS(2600),
    [anon_sym___fastcall] = ACTIONS(2600),
    [anon_sym___thiscall] = ACTIONS(2600),
    [anon_sym___vectorcall] = ACTIONS(2600),
    [anon_sym_signed] = ACTIONS(2600),
    [anon_sym_unsigned] = ACTIONS(2600),
    [anon_sym_long] = ACTIONS(2600),
    [anon_sym_short] = ACTIONS(2600),
    [anon_sym_static] = ACTIONS(2600),
    [anon_sym_auto] = ACTIONS(2600),
    [anon_sym_register] = ACTIONS(2600),
    [anon_sym_inline] = ACTIONS(2600),
    [anon_sym___inline] = ACTIONS(2600),
    [anon_sym___inline__] = ACTIONS(2600),
    [anon_sym___forceinline] = ACTIONS(2600),
    [anon_sym_thread_local] = ACTIONS(2600),
    [anon_sym___thread] = ACTIONS(2600),
    [anon_sym_CG_EXTERN] = ACTIONS(2600),
    [anon_sym_CG_INLINE] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2600),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2600),
    [anon_sym_IBOutlet] = ACTIONS(2600),
    [anon_sym_IBInspectable] = ACTIONS(2600),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2600),
    [anon_sym_NS_INLINE] = ACTIONS(2600),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2600),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2600),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2600),
    [anon_sym_const] = ACTIONS(2600),
    [anon_sym_constexpr] = ACTIONS(2600),
    [anon_sym_volatile] = ACTIONS(2600),
    [anon_sym_restrict] = ACTIONS(2600),
    [anon_sym___restrict__] = ACTIONS(2600),
    [anon_sym__Atomic] = ACTIONS(2600),
    [anon_sym__Noreturn] = ACTIONS(2600),
    [anon_sym__Nonnull] = ACTIONS(2600),
    [anon_sym_nullable] = ACTIONS(2600),
    [anon_sym__Complex] = ACTIONS(2600),
    [anon_sym__Nullable] = ACTIONS(2600),
    [anon_sym__Nullable_result] = ACTIONS(2600),
    [anon_sym__Null_unspecified] = ACTIONS(2600),
    [anon_sym___autoreleasing] = ACTIONS(2600),
    [anon_sym___block] = ACTIONS(2600),
    [anon_sym___bridge] = ACTIONS(2600),
    [anon_sym___bridge_retained] = ACTIONS(2600),
    [anon_sym___bridge_transfer] = ACTIONS(2600),
    [anon_sym___complex] = ACTIONS(2600),
    [anon_sym___const] = ACTIONS(2600),
    [anon_sym___imag] = ACTIONS(2600),
    [anon_sym___kindof] = ACTIONS(2600),
    [anon_sym___nonnull] = ACTIONS(2600),
    [anon_sym___nullable] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2600),
    [anon_sym___real] = ACTIONS(2600),
    [anon_sym___strong] = ACTIONS(2600),
    [anon_sym___unsafe_unretained] = ACTIONS(2600),
    [anon_sym___unused] = ACTIONS(2600),
    [anon_sym___weak] = ACTIONS(2600),
    [anon_sym_alignas] = ACTIONS(2600),
    [anon_sym__Alignas] = ACTIONS(2600),
    [sym_primitive_type] = ACTIONS(2600),
    [anon_sym_enum] = ACTIONS(2600),
    [anon_sym_struct] = ACTIONS(2600),
    [anon_sym_union] = ACTIONS(2600),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2600),
    [anon_sym___typeof] = ACTIONS(2600),
    [anon_sym_typeof] = ACTIONS(2600),
    [aux_sym_preproc_undef_token1] = ACTIONS(2600),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2600),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2600),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2600),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_API_AVAILABLE] = ACTIONS(2600),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_API_DEPRECATED] = ACTIONS(2600),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym___deprecated_msg] = ACTIONS(2600),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2600),
    [anon_sym_ATsynthesize] = ACTIONS(2602),
    [anon_sym_ATdynamic] = ACTIONS(2602),
    [anon_sym_BOOL] = ACTIONS(2600),
    [anon_sym_IMP] = ACTIONS(2600),
    [anon_sym_SEL] = ACTIONS(2600),
    [anon_sym_Class] = ACTIONS(2600),
    [anon_sym_id] = ACTIONS(2600),
    [anon_sym_ATdefs] = ACTIONS(2602),
  },
  [STATE(2299)] = {
    [sym_identifier] = ACTIONS(5387),
    [aux_sym_preproc_def_token1] = ACTIONS(5387),
    [aux_sym_preproc_if_token1] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5387),
    [sym_preproc_directive] = ACTIONS(5387),
    [anon_sym_DASH] = ACTIONS(5389),
    [anon_sym_PLUS] = ACTIONS(5389),
    [anon_sym_SEMI] = ACTIONS(5605),
    [anon_sym___extension__] = ACTIONS(5387),
    [anon_sym_typedef] = ACTIONS(5387),
    [anon_sym_extern] = ACTIONS(5387),
    [anon_sym___attribute__] = ACTIONS(5387),
    [anon_sym___attribute] = ACTIONS(5387),
    [anon_sym_noreturn] = ACTIONS(5387),
    [anon_sym_LBRACK] = ACTIONS(5389),
    [anon_sym___declspec] = ACTIONS(5387),
    [anon_sym___cdecl] = ACTIONS(5387),
    [anon_sym___clrcall] = ACTIONS(5387),
    [anon_sym___stdcall] = ACTIONS(5387),
    [anon_sym___fastcall] = ACTIONS(5387),
    [anon_sym___thiscall] = ACTIONS(5387),
    [anon_sym___vectorcall] = ACTIONS(5387),
    [anon_sym_signed] = ACTIONS(5387),
    [anon_sym_unsigned] = ACTIONS(5387),
    [anon_sym_long] = ACTIONS(5387),
    [anon_sym_short] = ACTIONS(5387),
    [anon_sym_static] = ACTIONS(5387),
    [anon_sym_auto] = ACTIONS(5387),
    [anon_sym_register] = ACTIONS(5387),
    [anon_sym_inline] = ACTIONS(5387),
    [anon_sym___inline] = ACTIONS(5387),
    [anon_sym___inline__] = ACTIONS(5387),
    [anon_sym___forceinline] = ACTIONS(5387),
    [anon_sym_thread_local] = ACTIONS(5387),
    [anon_sym___thread] = ACTIONS(5387),
    [anon_sym_CG_EXTERN] = ACTIONS(5387),
    [anon_sym_CG_INLINE] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5387),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5387),
    [anon_sym_IBOutlet] = ACTIONS(5387),
    [anon_sym_IBInspectable] = ACTIONS(5387),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5387),
    [anon_sym_NS_INLINE] = ACTIONS(5387),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5387),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5387),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5387),
    [anon_sym_const] = ACTIONS(5387),
    [anon_sym_constexpr] = ACTIONS(5387),
    [anon_sym_volatile] = ACTIONS(5387),
    [anon_sym_restrict] = ACTIONS(5387),
    [anon_sym___restrict__] = ACTIONS(5387),
    [anon_sym__Atomic] = ACTIONS(5387),
    [anon_sym__Noreturn] = ACTIONS(5387),
    [anon_sym__Nonnull] = ACTIONS(5387),
    [anon_sym_nullable] = ACTIONS(5387),
    [anon_sym__Complex] = ACTIONS(5387),
    [anon_sym__Nullable] = ACTIONS(5387),
    [anon_sym__Nullable_result] = ACTIONS(5387),
    [anon_sym__Null_unspecified] = ACTIONS(5387),
    [anon_sym___autoreleasing] = ACTIONS(5387),
    [anon_sym___block] = ACTIONS(5387),
    [anon_sym___bridge] = ACTIONS(5387),
    [anon_sym___bridge_retained] = ACTIONS(5387),
    [anon_sym___bridge_transfer] = ACTIONS(5387),
    [anon_sym___complex] = ACTIONS(5387),
    [anon_sym___const] = ACTIONS(5387),
    [anon_sym___imag] = ACTIONS(5387),
    [anon_sym___kindof] = ACTIONS(5387),
    [anon_sym___nonnull] = ACTIONS(5387),
    [anon_sym___nullable] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5387),
    [anon_sym___real] = ACTIONS(5387),
    [anon_sym___strong] = ACTIONS(5387),
    [anon_sym___unsafe_unretained] = ACTIONS(5387),
    [anon_sym___unused] = ACTIONS(5387),
    [anon_sym___weak] = ACTIONS(5387),
    [anon_sym_alignas] = ACTIONS(5387),
    [anon_sym__Alignas] = ACTIONS(5387),
    [sym_primitive_type] = ACTIONS(5387),
    [anon_sym_enum] = ACTIONS(5387),
    [anon_sym_struct] = ACTIONS(5387),
    [anon_sym_union] = ACTIONS(5387),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5387),
    [anon_sym___typeof] = ACTIONS(5387),
    [anon_sym_typeof] = ACTIONS(5387),
    [aux_sym_preproc_undef_token1] = ACTIONS(5387),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5387),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5387),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5387),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_API_AVAILABLE] = ACTIONS(5387),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_API_DEPRECATED] = ACTIONS(5387),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym___deprecated_msg] = ACTIONS(5387),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5387),
    [anon_sym_ATend] = ACTIONS(5389),
    [anon_sym_ATsynthesize] = ACTIONS(5389),
    [anon_sym_ATdynamic] = ACTIONS(5389),
    [anon_sym_BOOL] = ACTIONS(5387),
    [anon_sym_IMP] = ACTIONS(5387),
    [anon_sym_SEL] = ACTIONS(5387),
    [anon_sym_Class] = ACTIONS(5387),
    [anon_sym_id] = ACTIONS(5387),
  },
  [STATE(2300)] = {
    [sym_identifier] = ACTIONS(2604),
    [aux_sym_preproc_def_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token1] = ACTIONS(2604),
    [aux_sym_preproc_if_token2] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2604),
    [sym_preproc_directive] = ACTIONS(2604),
    [anon_sym_DASH] = ACTIONS(2606),
    [anon_sym_PLUS] = ACTIONS(2606),
    [anon_sym___extension__] = ACTIONS(2604),
    [anon_sym_typedef] = ACTIONS(2604),
    [anon_sym_extern] = ACTIONS(2604),
    [anon_sym___attribute__] = ACTIONS(2604),
    [anon_sym___attribute] = ACTIONS(2604),
    [anon_sym_noreturn] = ACTIONS(2604),
    [anon_sym_LBRACK] = ACTIONS(2606),
    [anon_sym___declspec] = ACTIONS(2604),
    [anon_sym___cdecl] = ACTIONS(2604),
    [anon_sym___clrcall] = ACTIONS(2604),
    [anon_sym___stdcall] = ACTIONS(2604),
    [anon_sym___fastcall] = ACTIONS(2604),
    [anon_sym___thiscall] = ACTIONS(2604),
    [anon_sym___vectorcall] = ACTIONS(2604),
    [anon_sym_signed] = ACTIONS(2604),
    [anon_sym_unsigned] = ACTIONS(2604),
    [anon_sym_long] = ACTIONS(2604),
    [anon_sym_short] = ACTIONS(2604),
    [anon_sym_static] = ACTIONS(2604),
    [anon_sym_auto] = ACTIONS(2604),
    [anon_sym_register] = ACTIONS(2604),
    [anon_sym_inline] = ACTIONS(2604),
    [anon_sym___inline] = ACTIONS(2604),
    [anon_sym___inline__] = ACTIONS(2604),
    [anon_sym___forceinline] = ACTIONS(2604),
    [anon_sym_thread_local] = ACTIONS(2604),
    [anon_sym___thread] = ACTIONS(2604),
    [anon_sym_CG_EXTERN] = ACTIONS(2604),
    [anon_sym_CG_INLINE] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2604),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2604),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2604),
    [anon_sym_IBOutlet] = ACTIONS(2604),
    [anon_sym_IBInspectable] = ACTIONS(2604),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2604),
    [anon_sym_NS_INLINE] = ACTIONS(2604),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2604),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2604),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2604),
    [anon_sym_const] = ACTIONS(2604),
    [anon_sym_constexpr] = ACTIONS(2604),
    [anon_sym_volatile] = ACTIONS(2604),
    [anon_sym_restrict] = ACTIONS(2604),
    [anon_sym___restrict__] = ACTIONS(2604),
    [anon_sym__Atomic] = ACTIONS(2604),
    [anon_sym__Noreturn] = ACTIONS(2604),
    [anon_sym__Nonnull] = ACTIONS(2604),
    [anon_sym_nullable] = ACTIONS(2604),
    [anon_sym__Complex] = ACTIONS(2604),
    [anon_sym__Nullable] = ACTIONS(2604),
    [anon_sym__Nullable_result] = ACTIONS(2604),
    [anon_sym__Null_unspecified] = ACTIONS(2604),
    [anon_sym___autoreleasing] = ACTIONS(2604),
    [anon_sym___block] = ACTIONS(2604),
    [anon_sym___bridge] = ACTIONS(2604),
    [anon_sym___bridge_retained] = ACTIONS(2604),
    [anon_sym___bridge_transfer] = ACTIONS(2604),
    [anon_sym___complex] = ACTIONS(2604),
    [anon_sym___const] = ACTIONS(2604),
    [anon_sym___imag] = ACTIONS(2604),
    [anon_sym___kindof] = ACTIONS(2604),
    [anon_sym___nonnull] = ACTIONS(2604),
    [anon_sym___nullable] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2604),
    [anon_sym___real] = ACTIONS(2604),
    [anon_sym___strong] = ACTIONS(2604),
    [anon_sym___unsafe_unretained] = ACTIONS(2604),
    [anon_sym___unused] = ACTIONS(2604),
    [anon_sym___weak] = ACTIONS(2604),
    [anon_sym_alignas] = ACTIONS(2604),
    [anon_sym__Alignas] = ACTIONS(2604),
    [sym_primitive_type] = ACTIONS(2604),
    [anon_sym_enum] = ACTIONS(2604),
    [anon_sym_struct] = ACTIONS(2604),
    [anon_sym_union] = ACTIONS(2604),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2604),
    [anon_sym___typeof] = ACTIONS(2604),
    [anon_sym_typeof] = ACTIONS(2604),
    [aux_sym_preproc_undef_token1] = ACTIONS(2604),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2604),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2604),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2604),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2604),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2604),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_API_AVAILABLE] = ACTIONS(2604),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_API_DEPRECATED] = ACTIONS(2604),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2604),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2604),
    [anon_sym___deprecated_msg] = ACTIONS(2604),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2604),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2604),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2604),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2604),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2604),
    [anon_sym_ATsynthesize] = ACTIONS(2606),
    [anon_sym_ATdynamic] = ACTIONS(2606),
    [anon_sym_ATproperty] = ACTIONS(2606),
    [anon_sym_BOOL] = ACTIONS(2604),
    [anon_sym_IMP] = ACTIONS(2604),
    [anon_sym_SEL] = ACTIONS(2604),
    [anon_sym_Class] = ACTIONS(2604),
    [anon_sym_id] = ACTIONS(2604),
  },
  [STATE(2301)] = {
    [sym_identifier] = ACTIONS(5393),
    [aux_sym_preproc_def_token1] = ACTIONS(5393),
    [aux_sym_preproc_if_token1] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5393),
    [sym_preproc_directive] = ACTIONS(5393),
    [anon_sym_DASH] = ACTIONS(5395),
    [anon_sym_PLUS] = ACTIONS(5395),
    [anon_sym_SEMI] = ACTIONS(5607),
    [anon_sym___extension__] = ACTIONS(5393),
    [anon_sym_typedef] = ACTIONS(5393),
    [anon_sym_extern] = ACTIONS(5393),
    [anon_sym___attribute__] = ACTIONS(5393),
    [anon_sym___attribute] = ACTIONS(5393),
    [anon_sym_noreturn] = ACTIONS(5393),
    [anon_sym_LBRACK] = ACTIONS(5395),
    [anon_sym___declspec] = ACTIONS(5393),
    [anon_sym___cdecl] = ACTIONS(5393),
    [anon_sym___clrcall] = ACTIONS(5393),
    [anon_sym___stdcall] = ACTIONS(5393),
    [anon_sym___fastcall] = ACTIONS(5393),
    [anon_sym___thiscall] = ACTIONS(5393),
    [anon_sym___vectorcall] = ACTIONS(5393),
    [anon_sym_signed] = ACTIONS(5393),
    [anon_sym_unsigned] = ACTIONS(5393),
    [anon_sym_long] = ACTIONS(5393),
    [anon_sym_short] = ACTIONS(5393),
    [anon_sym_static] = ACTIONS(5393),
    [anon_sym_auto] = ACTIONS(5393),
    [anon_sym_register] = ACTIONS(5393),
    [anon_sym_inline] = ACTIONS(5393),
    [anon_sym___inline] = ACTIONS(5393),
    [anon_sym___inline__] = ACTIONS(5393),
    [anon_sym___forceinline] = ACTIONS(5393),
    [anon_sym_thread_local] = ACTIONS(5393),
    [anon_sym___thread] = ACTIONS(5393),
    [anon_sym_CG_EXTERN] = ACTIONS(5393),
    [anon_sym_CG_INLINE] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5393),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5393),
    [anon_sym_IBOutlet] = ACTIONS(5393),
    [anon_sym_IBInspectable] = ACTIONS(5393),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5393),
    [anon_sym_NS_INLINE] = ACTIONS(5393),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5393),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5393),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5393),
    [anon_sym_const] = ACTIONS(5393),
    [anon_sym_constexpr] = ACTIONS(5393),
    [anon_sym_volatile] = ACTIONS(5393),
    [anon_sym_restrict] = ACTIONS(5393),
    [anon_sym___restrict__] = ACTIONS(5393),
    [anon_sym__Atomic] = ACTIONS(5393),
    [anon_sym__Noreturn] = ACTIONS(5393),
    [anon_sym__Nonnull] = ACTIONS(5393),
    [anon_sym_nullable] = ACTIONS(5393),
    [anon_sym__Complex] = ACTIONS(5393),
    [anon_sym__Nullable] = ACTIONS(5393),
    [anon_sym__Nullable_result] = ACTIONS(5393),
    [anon_sym__Null_unspecified] = ACTIONS(5393),
    [anon_sym___autoreleasing] = ACTIONS(5393),
    [anon_sym___block] = ACTIONS(5393),
    [anon_sym___bridge] = ACTIONS(5393),
    [anon_sym___bridge_retained] = ACTIONS(5393),
    [anon_sym___bridge_transfer] = ACTIONS(5393),
    [anon_sym___complex] = ACTIONS(5393),
    [anon_sym___const] = ACTIONS(5393),
    [anon_sym___imag] = ACTIONS(5393),
    [anon_sym___kindof] = ACTIONS(5393),
    [anon_sym___nonnull] = ACTIONS(5393),
    [anon_sym___nullable] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5393),
    [anon_sym___real] = ACTIONS(5393),
    [anon_sym___strong] = ACTIONS(5393),
    [anon_sym___unsafe_unretained] = ACTIONS(5393),
    [anon_sym___unused] = ACTIONS(5393),
    [anon_sym___weak] = ACTIONS(5393),
    [anon_sym_alignas] = ACTIONS(5393),
    [anon_sym__Alignas] = ACTIONS(5393),
    [sym_primitive_type] = ACTIONS(5393),
    [anon_sym_enum] = ACTIONS(5393),
    [anon_sym_struct] = ACTIONS(5393),
    [anon_sym_union] = ACTIONS(5393),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5393),
    [anon_sym___typeof] = ACTIONS(5393),
    [anon_sym_typeof] = ACTIONS(5393),
    [aux_sym_preproc_undef_token1] = ACTIONS(5393),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5393),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5393),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5393),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_API_AVAILABLE] = ACTIONS(5393),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_API_DEPRECATED] = ACTIONS(5393),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym___deprecated_msg] = ACTIONS(5393),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5393),
    [anon_sym_ATend] = ACTIONS(5395),
    [anon_sym_ATsynthesize] = ACTIONS(5395),
    [anon_sym_ATdynamic] = ACTIONS(5395),
    [anon_sym_BOOL] = ACTIONS(5393),
    [anon_sym_IMP] = ACTIONS(5393),
    [anon_sym_SEL] = ACTIONS(5393),
    [anon_sym_Class] = ACTIONS(5393),
    [anon_sym_id] = ACTIONS(5393),
  },
  [STATE(2302)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token2] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_DASH] = ACTIONS(2440),
    [anon_sym_PLUS] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATproperty] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(2303)] = {
    [sym_identifier] = ACTIONS(5609),
    [aux_sym_preproc_def_token1] = ACTIONS(5609),
    [aux_sym_preproc_if_token1] = ACTIONS(5609),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5609),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5609),
    [sym_preproc_directive] = ACTIONS(5609),
    [anon_sym_DASH] = ACTIONS(5611),
    [anon_sym_PLUS] = ACTIONS(5611),
    [anon_sym___extension__] = ACTIONS(5609),
    [anon_sym_typedef] = ACTIONS(5609),
    [anon_sym_extern] = ACTIONS(5609),
    [anon_sym___attribute__] = ACTIONS(5609),
    [anon_sym___attribute] = ACTIONS(5609),
    [anon_sym_noreturn] = ACTIONS(5609),
    [anon_sym_LBRACK] = ACTIONS(5611),
    [anon_sym___declspec] = ACTIONS(5609),
    [anon_sym___cdecl] = ACTIONS(5609),
    [anon_sym___clrcall] = ACTIONS(5609),
    [anon_sym___stdcall] = ACTIONS(5609),
    [anon_sym___fastcall] = ACTIONS(5609),
    [anon_sym___thiscall] = ACTIONS(5609),
    [anon_sym___vectorcall] = ACTIONS(5609),
    [anon_sym_signed] = ACTIONS(5609),
    [anon_sym_unsigned] = ACTIONS(5609),
    [anon_sym_long] = ACTIONS(5609),
    [anon_sym_short] = ACTIONS(5609),
    [anon_sym_static] = ACTIONS(5609),
    [anon_sym_auto] = ACTIONS(5609),
    [anon_sym_register] = ACTIONS(5609),
    [anon_sym_inline] = ACTIONS(5609),
    [anon_sym___inline] = ACTIONS(5609),
    [anon_sym___inline__] = ACTIONS(5609),
    [anon_sym___forceinline] = ACTIONS(5609),
    [anon_sym_thread_local] = ACTIONS(5609),
    [anon_sym___thread] = ACTIONS(5609),
    [anon_sym_CG_EXTERN] = ACTIONS(5609),
    [anon_sym_CG_INLINE] = ACTIONS(5609),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5609),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5609),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5609),
    [anon_sym_IBOutlet] = ACTIONS(5609),
    [anon_sym_IBInspectable] = ACTIONS(5609),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5609),
    [anon_sym_NS_INLINE] = ACTIONS(5609),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5609),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5609),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5609),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5609),
    [anon_sym_const] = ACTIONS(5609),
    [anon_sym_constexpr] = ACTIONS(5609),
    [anon_sym_volatile] = ACTIONS(5609),
    [anon_sym_restrict] = ACTIONS(5609),
    [anon_sym___restrict__] = ACTIONS(5609),
    [anon_sym__Atomic] = ACTIONS(5609),
    [anon_sym__Noreturn] = ACTIONS(5609),
    [anon_sym__Nonnull] = ACTIONS(5609),
    [anon_sym_nullable] = ACTIONS(5609),
    [anon_sym__Complex] = ACTIONS(5609),
    [anon_sym__Nullable] = ACTIONS(5609),
    [anon_sym__Nullable_result] = ACTIONS(5609),
    [anon_sym__Null_unspecified] = ACTIONS(5609),
    [anon_sym___autoreleasing] = ACTIONS(5609),
    [anon_sym___block] = ACTIONS(5609),
    [anon_sym___bridge] = ACTIONS(5609),
    [anon_sym___bridge_retained] = ACTIONS(5609),
    [anon_sym___bridge_transfer] = ACTIONS(5609),
    [anon_sym___complex] = ACTIONS(5609),
    [anon_sym___const] = ACTIONS(5609),
    [anon_sym___imag] = ACTIONS(5609),
    [anon_sym___kindof] = ACTIONS(5609),
    [anon_sym___nonnull] = ACTIONS(5609),
    [anon_sym___nullable] = ACTIONS(5609),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5609),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5609),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5609),
    [anon_sym___real] = ACTIONS(5609),
    [anon_sym___strong] = ACTIONS(5609),
    [anon_sym___unsafe_unretained] = ACTIONS(5609),
    [anon_sym___unused] = ACTIONS(5609),
    [anon_sym___weak] = ACTIONS(5609),
    [anon_sym_alignas] = ACTIONS(5609),
    [anon_sym__Alignas] = ACTIONS(5609),
    [sym_primitive_type] = ACTIONS(5609),
    [anon_sym_enum] = ACTIONS(5609),
    [anon_sym_struct] = ACTIONS(5609),
    [anon_sym_union] = ACTIONS(5609),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5609),
    [anon_sym___typeof] = ACTIONS(5609),
    [anon_sym_typeof] = ACTIONS(5609),
    [aux_sym_preproc_undef_token1] = ACTIONS(5609),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5609),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5609),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5609),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5609),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5609),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5609),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5609),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5609),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5609),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_API_AVAILABLE] = ACTIONS(5609),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_API_DEPRECATED] = ACTIONS(5609),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5609),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5609),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5609),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5609),
    [anon_sym___deprecated_msg] = ACTIONS(5609),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5609),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5609),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5609),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5609),
    [anon_sym_ATend] = ACTIONS(5611),
    [anon_sym_AToptional] = ACTIONS(5611),
    [anon_sym_ATrequired] = ACTIONS(5611),
    [anon_sym_ATproperty] = ACTIONS(5611),
    [anon_sym_BOOL] = ACTIONS(5609),
    [anon_sym_IMP] = ACTIONS(5609),
    [anon_sym_SEL] = ACTIONS(5609),
    [anon_sym_Class] = ACTIONS(5609),
    [anon_sym_id] = ACTIONS(5609),
  },
  [STATE(2304)] = {
    [sym_identifier] = ACTIONS(5561),
    [aux_sym_preproc_def_token1] = ACTIONS(5561),
    [aux_sym_preproc_if_token1] = ACTIONS(5561),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5561),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5561),
    [sym_preproc_directive] = ACTIONS(5561),
    [anon_sym_DASH] = ACTIONS(5563),
    [anon_sym_PLUS] = ACTIONS(5563),
    [anon_sym___extension__] = ACTIONS(5561),
    [anon_sym_typedef] = ACTIONS(5561),
    [anon_sym_extern] = ACTIONS(5561),
    [anon_sym___attribute__] = ACTIONS(5561),
    [anon_sym___attribute] = ACTIONS(5561),
    [anon_sym_noreturn] = ACTIONS(5561),
    [anon_sym_LBRACK] = ACTIONS(5563),
    [anon_sym___declspec] = ACTIONS(5561),
    [anon_sym___cdecl] = ACTIONS(5561),
    [anon_sym___clrcall] = ACTIONS(5561),
    [anon_sym___stdcall] = ACTIONS(5561),
    [anon_sym___fastcall] = ACTIONS(5561),
    [anon_sym___thiscall] = ACTIONS(5561),
    [anon_sym___vectorcall] = ACTIONS(5561),
    [anon_sym_signed] = ACTIONS(5561),
    [anon_sym_unsigned] = ACTIONS(5561),
    [anon_sym_long] = ACTIONS(5561),
    [anon_sym_short] = ACTIONS(5561),
    [anon_sym_static] = ACTIONS(5561),
    [anon_sym_auto] = ACTIONS(5561),
    [anon_sym_register] = ACTIONS(5561),
    [anon_sym_inline] = ACTIONS(5561),
    [anon_sym___inline] = ACTIONS(5561),
    [anon_sym___inline__] = ACTIONS(5561),
    [anon_sym___forceinline] = ACTIONS(5561),
    [anon_sym_thread_local] = ACTIONS(5561),
    [anon_sym___thread] = ACTIONS(5561),
    [anon_sym_CG_EXTERN] = ACTIONS(5561),
    [anon_sym_CG_INLINE] = ACTIONS(5561),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5561),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5561),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5561),
    [anon_sym_IBOutlet] = ACTIONS(5561),
    [anon_sym_IBInspectable] = ACTIONS(5561),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5561),
    [anon_sym_NS_INLINE] = ACTIONS(5561),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5561),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5561),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5561),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5561),
    [anon_sym_const] = ACTIONS(5561),
    [anon_sym_constexpr] = ACTIONS(5561),
    [anon_sym_volatile] = ACTIONS(5561),
    [anon_sym_restrict] = ACTIONS(5561),
    [anon_sym___restrict__] = ACTIONS(5561),
    [anon_sym__Atomic] = ACTIONS(5561),
    [anon_sym__Noreturn] = ACTIONS(5561),
    [anon_sym__Nonnull] = ACTIONS(5561),
    [anon_sym_nullable] = ACTIONS(5561),
    [anon_sym__Complex] = ACTIONS(5561),
    [anon_sym__Nullable] = ACTIONS(5561),
    [anon_sym__Nullable_result] = ACTIONS(5561),
    [anon_sym__Null_unspecified] = ACTIONS(5561),
    [anon_sym___autoreleasing] = ACTIONS(5561),
    [anon_sym___block] = ACTIONS(5561),
    [anon_sym___bridge] = ACTIONS(5561),
    [anon_sym___bridge_retained] = ACTIONS(5561),
    [anon_sym___bridge_transfer] = ACTIONS(5561),
    [anon_sym___complex] = ACTIONS(5561),
    [anon_sym___const] = ACTIONS(5561),
    [anon_sym___imag] = ACTIONS(5561),
    [anon_sym___kindof] = ACTIONS(5561),
    [anon_sym___nonnull] = ACTIONS(5561),
    [anon_sym___nullable] = ACTIONS(5561),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5561),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5561),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5561),
    [anon_sym___real] = ACTIONS(5561),
    [anon_sym___strong] = ACTIONS(5561),
    [anon_sym___unsafe_unretained] = ACTIONS(5561),
    [anon_sym___unused] = ACTIONS(5561),
    [anon_sym___weak] = ACTIONS(5561),
    [anon_sym_alignas] = ACTIONS(5561),
    [anon_sym__Alignas] = ACTIONS(5561),
    [sym_primitive_type] = ACTIONS(5561),
    [anon_sym_enum] = ACTIONS(5561),
    [anon_sym_struct] = ACTIONS(5561),
    [anon_sym_union] = ACTIONS(5561),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5561),
    [anon_sym___typeof] = ACTIONS(5561),
    [anon_sym_typeof] = ACTIONS(5561),
    [aux_sym_preproc_undef_token1] = ACTIONS(5561),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5561),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5561),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5561),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5561),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5561),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5561),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5561),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5561),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5561),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_API_AVAILABLE] = ACTIONS(5561),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_API_DEPRECATED] = ACTIONS(5561),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5561),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5561),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5561),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5561),
    [anon_sym___deprecated_msg] = ACTIONS(5561),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5561),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5561),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5561),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5561),
    [anon_sym_ATend] = ACTIONS(5563),
    [anon_sym_AToptional] = ACTIONS(5563),
    [anon_sym_ATrequired] = ACTIONS(5563),
    [anon_sym_ATproperty] = ACTIONS(5563),
    [anon_sym_BOOL] = ACTIONS(5561),
    [anon_sym_IMP] = ACTIONS(5561),
    [anon_sym_SEL] = ACTIONS(5561),
    [anon_sym_Class] = ACTIONS(5561),
    [anon_sym_id] = ACTIONS(5561),
  },
  [STATE(2305)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token2] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_DASH] = ACTIONS(2444),
    [anon_sym_PLUS] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATproperty] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(2306)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token2] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_DASH] = ACTIONS(2448),
    [anon_sym_PLUS] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATproperty] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(2307)] = {
    [sym_identifier] = ACTIONS(5489),
    [aux_sym_preproc_def_token1] = ACTIONS(5489),
    [aux_sym_preproc_if_token1] = ACTIONS(5489),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5489),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5489),
    [sym_preproc_directive] = ACTIONS(5489),
    [anon_sym_DASH] = ACTIONS(5491),
    [anon_sym_PLUS] = ACTIONS(5491),
    [anon_sym___extension__] = ACTIONS(5489),
    [anon_sym_typedef] = ACTIONS(5489),
    [anon_sym_extern] = ACTIONS(5489),
    [anon_sym___attribute__] = ACTIONS(5489),
    [anon_sym___attribute] = ACTIONS(5489),
    [anon_sym_noreturn] = ACTIONS(5489),
    [anon_sym_LBRACK] = ACTIONS(5491),
    [anon_sym___declspec] = ACTIONS(5489),
    [anon_sym___cdecl] = ACTIONS(5489),
    [anon_sym___clrcall] = ACTIONS(5489),
    [anon_sym___stdcall] = ACTIONS(5489),
    [anon_sym___fastcall] = ACTIONS(5489),
    [anon_sym___thiscall] = ACTIONS(5489),
    [anon_sym___vectorcall] = ACTIONS(5489),
    [anon_sym_signed] = ACTIONS(5489),
    [anon_sym_unsigned] = ACTIONS(5489),
    [anon_sym_long] = ACTIONS(5489),
    [anon_sym_short] = ACTIONS(5489),
    [anon_sym_static] = ACTIONS(5489),
    [anon_sym_auto] = ACTIONS(5489),
    [anon_sym_register] = ACTIONS(5489),
    [anon_sym_inline] = ACTIONS(5489),
    [anon_sym___inline] = ACTIONS(5489),
    [anon_sym___inline__] = ACTIONS(5489),
    [anon_sym___forceinline] = ACTIONS(5489),
    [anon_sym_thread_local] = ACTIONS(5489),
    [anon_sym___thread] = ACTIONS(5489),
    [anon_sym_CG_EXTERN] = ACTIONS(5489),
    [anon_sym_CG_INLINE] = ACTIONS(5489),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5489),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5489),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5489),
    [anon_sym_IBOutlet] = ACTIONS(5489),
    [anon_sym_IBInspectable] = ACTIONS(5489),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5489),
    [anon_sym_NS_INLINE] = ACTIONS(5489),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5489),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5489),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5489),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5489),
    [anon_sym_const] = ACTIONS(5489),
    [anon_sym_constexpr] = ACTIONS(5489),
    [anon_sym_volatile] = ACTIONS(5489),
    [anon_sym_restrict] = ACTIONS(5489),
    [anon_sym___restrict__] = ACTIONS(5489),
    [anon_sym__Atomic] = ACTIONS(5489),
    [anon_sym__Noreturn] = ACTIONS(5489),
    [anon_sym__Nonnull] = ACTIONS(5489),
    [anon_sym_nullable] = ACTIONS(5489),
    [anon_sym__Complex] = ACTIONS(5489),
    [anon_sym__Nullable] = ACTIONS(5489),
    [anon_sym__Nullable_result] = ACTIONS(5489),
    [anon_sym__Null_unspecified] = ACTIONS(5489),
    [anon_sym___autoreleasing] = ACTIONS(5489),
    [anon_sym___block] = ACTIONS(5489),
    [anon_sym___bridge] = ACTIONS(5489),
    [anon_sym___bridge_retained] = ACTIONS(5489),
    [anon_sym___bridge_transfer] = ACTIONS(5489),
    [anon_sym___complex] = ACTIONS(5489),
    [anon_sym___const] = ACTIONS(5489),
    [anon_sym___imag] = ACTIONS(5489),
    [anon_sym___kindof] = ACTIONS(5489),
    [anon_sym___nonnull] = ACTIONS(5489),
    [anon_sym___nullable] = ACTIONS(5489),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5489),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5489),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5489),
    [anon_sym___real] = ACTIONS(5489),
    [anon_sym___strong] = ACTIONS(5489),
    [anon_sym___unsafe_unretained] = ACTIONS(5489),
    [anon_sym___unused] = ACTIONS(5489),
    [anon_sym___weak] = ACTIONS(5489),
    [anon_sym_alignas] = ACTIONS(5489),
    [anon_sym__Alignas] = ACTIONS(5489),
    [sym_primitive_type] = ACTIONS(5489),
    [anon_sym_enum] = ACTIONS(5489),
    [anon_sym_struct] = ACTIONS(5489),
    [anon_sym_union] = ACTIONS(5489),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5489),
    [anon_sym___typeof] = ACTIONS(5489),
    [anon_sym_typeof] = ACTIONS(5489),
    [aux_sym_preproc_undef_token1] = ACTIONS(5489),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5489),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5489),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5489),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5489),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5489),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5489),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5489),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5489),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5489),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5489),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5489),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5489),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5489),
    [anon_sym_API_AVAILABLE] = ACTIONS(5489),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5489),
    [anon_sym_API_DEPRECATED] = ACTIONS(5489),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5489),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5489),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5489),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5489),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5489),
    [anon_sym___deprecated_msg] = ACTIONS(5489),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5489),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5489),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5489),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5489),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5489),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5489),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5489),
    [anon_sym_ATend] = ACTIONS(5491),
    [anon_sym_ATsynthesize] = ACTIONS(5491),
    [anon_sym_ATdynamic] = ACTIONS(5491),
    [anon_sym_ATproperty] = ACTIONS(5491),
    [anon_sym_BOOL] = ACTIONS(5489),
    [anon_sym_IMP] = ACTIONS(5489),
    [anon_sym_SEL] = ACTIONS(5489),
    [anon_sym_Class] = ACTIONS(5489),
    [anon_sym_id] = ACTIONS(5489),
  },
  [STATE(2308)] = {
    [sym_identifier] = ACTIONS(5359),
    [aux_sym_preproc_def_token1] = ACTIONS(5359),
    [aux_sym_preproc_if_token1] = ACTIONS(5359),
    [aux_sym_preproc_if_token2] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5359),
    [sym_preproc_directive] = ACTIONS(5359),
    [anon_sym_DASH] = ACTIONS(5361),
    [anon_sym_PLUS] = ACTIONS(5361),
    [anon_sym_SEMI] = ACTIONS(5613),
    [anon_sym___extension__] = ACTIONS(5359),
    [anon_sym_typedef] = ACTIONS(5359),
    [anon_sym_extern] = ACTIONS(5359),
    [anon_sym___attribute__] = ACTIONS(5359),
    [anon_sym___attribute] = ACTIONS(5359),
    [anon_sym_noreturn] = ACTIONS(5359),
    [anon_sym_LBRACK] = ACTIONS(5361),
    [anon_sym___declspec] = ACTIONS(5359),
    [anon_sym___cdecl] = ACTIONS(5359),
    [anon_sym___clrcall] = ACTIONS(5359),
    [anon_sym___stdcall] = ACTIONS(5359),
    [anon_sym___fastcall] = ACTIONS(5359),
    [anon_sym___thiscall] = ACTIONS(5359),
    [anon_sym___vectorcall] = ACTIONS(5359),
    [anon_sym_signed] = ACTIONS(5359),
    [anon_sym_unsigned] = ACTIONS(5359),
    [anon_sym_long] = ACTIONS(5359),
    [anon_sym_short] = ACTIONS(5359),
    [anon_sym_static] = ACTIONS(5359),
    [anon_sym_auto] = ACTIONS(5359),
    [anon_sym_register] = ACTIONS(5359),
    [anon_sym_inline] = ACTIONS(5359),
    [anon_sym___inline] = ACTIONS(5359),
    [anon_sym___inline__] = ACTIONS(5359),
    [anon_sym___forceinline] = ACTIONS(5359),
    [anon_sym_thread_local] = ACTIONS(5359),
    [anon_sym___thread] = ACTIONS(5359),
    [anon_sym_CG_EXTERN] = ACTIONS(5359),
    [anon_sym_CG_INLINE] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5359),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5359),
    [anon_sym_IBOutlet] = ACTIONS(5359),
    [anon_sym_IBInspectable] = ACTIONS(5359),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5359),
    [anon_sym_NS_INLINE] = ACTIONS(5359),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5359),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5359),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5359),
    [anon_sym_const] = ACTIONS(5359),
    [anon_sym_constexpr] = ACTIONS(5359),
    [anon_sym_volatile] = ACTIONS(5359),
    [anon_sym_restrict] = ACTIONS(5359),
    [anon_sym___restrict__] = ACTIONS(5359),
    [anon_sym__Atomic] = ACTIONS(5359),
    [anon_sym__Noreturn] = ACTIONS(5359),
    [anon_sym__Nonnull] = ACTIONS(5359),
    [anon_sym_nullable] = ACTIONS(5359),
    [anon_sym__Complex] = ACTIONS(5359),
    [anon_sym__Nullable] = ACTIONS(5359),
    [anon_sym__Nullable_result] = ACTIONS(5359),
    [anon_sym__Null_unspecified] = ACTIONS(5359),
    [anon_sym___autoreleasing] = ACTIONS(5359),
    [anon_sym___block] = ACTIONS(5359),
    [anon_sym___bridge] = ACTIONS(5359),
    [anon_sym___bridge_retained] = ACTIONS(5359),
    [anon_sym___bridge_transfer] = ACTIONS(5359),
    [anon_sym___complex] = ACTIONS(5359),
    [anon_sym___const] = ACTIONS(5359),
    [anon_sym___imag] = ACTIONS(5359),
    [anon_sym___kindof] = ACTIONS(5359),
    [anon_sym___nonnull] = ACTIONS(5359),
    [anon_sym___nullable] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5359),
    [anon_sym___real] = ACTIONS(5359),
    [anon_sym___strong] = ACTIONS(5359),
    [anon_sym___unsafe_unretained] = ACTIONS(5359),
    [anon_sym___unused] = ACTIONS(5359),
    [anon_sym___weak] = ACTIONS(5359),
    [anon_sym_alignas] = ACTIONS(5359),
    [anon_sym__Alignas] = ACTIONS(5359),
    [sym_primitive_type] = ACTIONS(5359),
    [anon_sym_enum] = ACTIONS(5359),
    [anon_sym_struct] = ACTIONS(5359),
    [anon_sym_union] = ACTIONS(5359),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5359),
    [anon_sym___typeof] = ACTIONS(5359),
    [anon_sym_typeof] = ACTIONS(5359),
    [aux_sym_preproc_undef_token1] = ACTIONS(5359),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5359),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5359),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5359),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_API_AVAILABLE] = ACTIONS(5359),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_API_DEPRECATED] = ACTIONS(5359),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym___deprecated_msg] = ACTIONS(5359),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5359),
    [anon_sym_ATsynthesize] = ACTIONS(5361),
    [anon_sym_ATdynamic] = ACTIONS(5361),
    [anon_sym_BOOL] = ACTIONS(5359),
    [anon_sym_IMP] = ACTIONS(5359),
    [anon_sym_SEL] = ACTIONS(5359),
    [anon_sym_Class] = ACTIONS(5359),
    [anon_sym_id] = ACTIONS(5359),
  },
  [STATE(2309)] = {
    [sym_identifier] = ACTIONS(5475),
    [aux_sym_preproc_def_token1] = ACTIONS(5475),
    [aux_sym_preproc_if_token1] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5475),
    [sym_preproc_directive] = ACTIONS(5475),
    [anon_sym_DASH] = ACTIONS(5477),
    [anon_sym_PLUS] = ACTIONS(5477),
    [anon_sym_SEMI] = ACTIONS(5615),
    [anon_sym___extension__] = ACTIONS(5475),
    [anon_sym_typedef] = ACTIONS(5475),
    [anon_sym_extern] = ACTIONS(5475),
    [anon_sym___attribute__] = ACTIONS(5475),
    [anon_sym___attribute] = ACTIONS(5475),
    [anon_sym_noreturn] = ACTIONS(5475),
    [anon_sym_LBRACK] = ACTIONS(5477),
    [anon_sym___declspec] = ACTIONS(5475),
    [anon_sym___cdecl] = ACTIONS(5475),
    [anon_sym___clrcall] = ACTIONS(5475),
    [anon_sym___stdcall] = ACTIONS(5475),
    [anon_sym___fastcall] = ACTIONS(5475),
    [anon_sym___thiscall] = ACTIONS(5475),
    [anon_sym___vectorcall] = ACTIONS(5475),
    [anon_sym_signed] = ACTIONS(5475),
    [anon_sym_unsigned] = ACTIONS(5475),
    [anon_sym_long] = ACTIONS(5475),
    [anon_sym_short] = ACTIONS(5475),
    [anon_sym_static] = ACTIONS(5475),
    [anon_sym_auto] = ACTIONS(5475),
    [anon_sym_register] = ACTIONS(5475),
    [anon_sym_inline] = ACTIONS(5475),
    [anon_sym___inline] = ACTIONS(5475),
    [anon_sym___inline__] = ACTIONS(5475),
    [anon_sym___forceinline] = ACTIONS(5475),
    [anon_sym_thread_local] = ACTIONS(5475),
    [anon_sym___thread] = ACTIONS(5475),
    [anon_sym_CG_EXTERN] = ACTIONS(5475),
    [anon_sym_CG_INLINE] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5475),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5475),
    [anon_sym_IBOutlet] = ACTIONS(5475),
    [anon_sym_IBInspectable] = ACTIONS(5475),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5475),
    [anon_sym_NS_INLINE] = ACTIONS(5475),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5475),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5475),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5475),
    [anon_sym_const] = ACTIONS(5475),
    [anon_sym_constexpr] = ACTIONS(5475),
    [anon_sym_volatile] = ACTIONS(5475),
    [anon_sym_restrict] = ACTIONS(5475),
    [anon_sym___restrict__] = ACTIONS(5475),
    [anon_sym__Atomic] = ACTIONS(5475),
    [anon_sym__Noreturn] = ACTIONS(5475),
    [anon_sym__Nonnull] = ACTIONS(5475),
    [anon_sym_nullable] = ACTIONS(5475),
    [anon_sym__Complex] = ACTIONS(5475),
    [anon_sym__Nullable] = ACTIONS(5475),
    [anon_sym__Nullable_result] = ACTIONS(5475),
    [anon_sym__Null_unspecified] = ACTIONS(5475),
    [anon_sym___autoreleasing] = ACTIONS(5475),
    [anon_sym___block] = ACTIONS(5475),
    [anon_sym___bridge] = ACTIONS(5475),
    [anon_sym___bridge_retained] = ACTIONS(5475),
    [anon_sym___bridge_transfer] = ACTIONS(5475),
    [anon_sym___complex] = ACTIONS(5475),
    [anon_sym___const] = ACTIONS(5475),
    [anon_sym___imag] = ACTIONS(5475),
    [anon_sym___kindof] = ACTIONS(5475),
    [anon_sym___nonnull] = ACTIONS(5475),
    [anon_sym___nullable] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5475),
    [anon_sym___real] = ACTIONS(5475),
    [anon_sym___strong] = ACTIONS(5475),
    [anon_sym___unsafe_unretained] = ACTIONS(5475),
    [anon_sym___unused] = ACTIONS(5475),
    [anon_sym___weak] = ACTIONS(5475),
    [anon_sym_alignas] = ACTIONS(5475),
    [anon_sym__Alignas] = ACTIONS(5475),
    [sym_primitive_type] = ACTIONS(5475),
    [anon_sym_enum] = ACTIONS(5475),
    [anon_sym_struct] = ACTIONS(5475),
    [anon_sym_union] = ACTIONS(5475),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5475),
    [anon_sym___typeof] = ACTIONS(5475),
    [anon_sym_typeof] = ACTIONS(5475),
    [aux_sym_preproc_undef_token1] = ACTIONS(5475),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5475),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5475),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5475),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_API_AVAILABLE] = ACTIONS(5475),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_API_DEPRECATED] = ACTIONS(5475),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym___deprecated_msg] = ACTIONS(5475),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5475),
    [anon_sym_ATend] = ACTIONS(5477),
    [anon_sym_ATsynthesize] = ACTIONS(5477),
    [anon_sym_ATdynamic] = ACTIONS(5477),
    [anon_sym_BOOL] = ACTIONS(5475),
    [anon_sym_IMP] = ACTIONS(5475),
    [anon_sym_SEL] = ACTIONS(5475),
    [anon_sym_Class] = ACTIONS(5475),
    [anon_sym_id] = ACTIONS(5475),
  },
  [STATE(2310)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2321),
    [sym_identifier] = ACTIONS(5365),
    [aux_sym_preproc_def_token1] = ACTIONS(5365),
    [aux_sym_preproc_if_token1] = ACTIONS(5365),
    [aux_sym_preproc_if_token2] = ACTIONS(5365),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5365),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5365),
    [sym_preproc_directive] = ACTIONS(5365),
    [anon_sym_DASH] = ACTIONS(5367),
    [anon_sym_PLUS] = ACTIONS(5367),
    [anon_sym_SEMI] = ACTIONS(5565),
    [anon_sym___extension__] = ACTIONS(5365),
    [anon_sym_typedef] = ACTIONS(5365),
    [anon_sym_extern] = ACTIONS(5365),
    [anon_sym___attribute__] = ACTIONS(5365),
    [anon_sym___attribute] = ACTIONS(5365),
    [anon_sym_noreturn] = ACTIONS(5365),
    [anon_sym_LBRACK] = ACTIONS(5367),
    [anon_sym___declspec] = ACTIONS(5365),
    [anon_sym___cdecl] = ACTIONS(5365),
    [anon_sym___clrcall] = ACTIONS(5365),
    [anon_sym___stdcall] = ACTIONS(5365),
    [anon_sym___fastcall] = ACTIONS(5365),
    [anon_sym___thiscall] = ACTIONS(5365),
    [anon_sym___vectorcall] = ACTIONS(5365),
    [anon_sym_signed] = ACTIONS(5365),
    [anon_sym_unsigned] = ACTIONS(5365),
    [anon_sym_long] = ACTIONS(5365),
    [anon_sym_short] = ACTIONS(5365),
    [anon_sym_static] = ACTIONS(5365),
    [anon_sym_auto] = ACTIONS(5365),
    [anon_sym_register] = ACTIONS(5365),
    [anon_sym_inline] = ACTIONS(5365),
    [anon_sym___inline] = ACTIONS(5365),
    [anon_sym___inline__] = ACTIONS(5365),
    [anon_sym___forceinline] = ACTIONS(5365),
    [anon_sym_thread_local] = ACTIONS(5365),
    [anon_sym___thread] = ACTIONS(5365),
    [anon_sym_CG_EXTERN] = ACTIONS(5365),
    [anon_sym_CG_INLINE] = ACTIONS(5365),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5365),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5365),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5365),
    [anon_sym_IBOutlet] = ACTIONS(5365),
    [anon_sym_IBInspectable] = ACTIONS(5365),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5365),
    [anon_sym_NS_INLINE] = ACTIONS(5365),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5365),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5365),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5365),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5365),
    [anon_sym_const] = ACTIONS(5365),
    [anon_sym_constexpr] = ACTIONS(5365),
    [anon_sym_volatile] = ACTIONS(5365),
    [anon_sym_restrict] = ACTIONS(5365),
    [anon_sym___restrict__] = ACTIONS(5365),
    [anon_sym__Atomic] = ACTIONS(5365),
    [anon_sym__Noreturn] = ACTIONS(5365),
    [anon_sym__Nonnull] = ACTIONS(5365),
    [anon_sym_nullable] = ACTIONS(5365),
    [anon_sym__Complex] = ACTIONS(5365),
    [anon_sym__Nullable] = ACTIONS(5365),
    [anon_sym__Nullable_result] = ACTIONS(5365),
    [anon_sym__Null_unspecified] = ACTIONS(5365),
    [anon_sym___autoreleasing] = ACTIONS(5365),
    [anon_sym___block] = ACTIONS(5365),
    [anon_sym___bridge] = ACTIONS(5365),
    [anon_sym___bridge_retained] = ACTIONS(5365),
    [anon_sym___bridge_transfer] = ACTIONS(5365),
    [anon_sym___complex] = ACTIONS(5365),
    [anon_sym___const] = ACTIONS(5365),
    [anon_sym___imag] = ACTIONS(5365),
    [anon_sym___kindof] = ACTIONS(5365),
    [anon_sym___nonnull] = ACTIONS(5365),
    [anon_sym___nullable] = ACTIONS(5365),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5365),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5365),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5365),
    [anon_sym___real] = ACTIONS(5365),
    [anon_sym___strong] = ACTIONS(5365),
    [anon_sym___unsafe_unretained] = ACTIONS(5365),
    [anon_sym___unused] = ACTIONS(5365),
    [anon_sym___weak] = ACTIONS(5365),
    [anon_sym_alignas] = ACTIONS(5365),
    [anon_sym__Alignas] = ACTIONS(5365),
    [sym_primitive_type] = ACTIONS(5365),
    [anon_sym_enum] = ACTIONS(5365),
    [anon_sym_struct] = ACTIONS(5365),
    [anon_sym_union] = ACTIONS(5365),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5365),
    [anon_sym___typeof] = ACTIONS(5365),
    [anon_sym_typeof] = ACTIONS(5365),
    [aux_sym_preproc_undef_token1] = ACTIONS(5365),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5365),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5365),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5365),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5365),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5365),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5365),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5365),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5365),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5365),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_API_AVAILABLE] = ACTIONS(5365),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_API_DEPRECATED] = ACTIONS(5365),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5365),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5365),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5365),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5365),
    [anon_sym___deprecated_msg] = ACTIONS(5365),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5365),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5365),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5365),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5365),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5365),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5365),
    [anon_sym_ATproperty] = ACTIONS(5367),
    [anon_sym_BOOL] = ACTIONS(5365),
    [anon_sym_IMP] = ACTIONS(5365),
    [anon_sym_SEL] = ACTIONS(5365),
    [anon_sym_Class] = ACTIONS(5365),
    [anon_sym_id] = ACTIONS(5365),
  },
  [STATE(2311)] = {
    [sym_identifier] = ACTIONS(2412),
    [aux_sym_preproc_def_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token1] = ACTIONS(2412),
    [aux_sym_preproc_if_token2] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2412),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2412),
    [sym_preproc_directive] = ACTIONS(2412),
    [anon_sym_DASH] = ACTIONS(2414),
    [anon_sym_PLUS] = ACTIONS(2414),
    [anon_sym___extension__] = ACTIONS(2412),
    [anon_sym_typedef] = ACTIONS(2412),
    [anon_sym_extern] = ACTIONS(2412),
    [anon_sym___attribute__] = ACTIONS(2412),
    [anon_sym___attribute] = ACTIONS(2412),
    [anon_sym_noreturn] = ACTIONS(2412),
    [anon_sym_LBRACK] = ACTIONS(2414),
    [anon_sym___declspec] = ACTIONS(2412),
    [anon_sym___cdecl] = ACTIONS(2412),
    [anon_sym___clrcall] = ACTIONS(2412),
    [anon_sym___stdcall] = ACTIONS(2412),
    [anon_sym___fastcall] = ACTIONS(2412),
    [anon_sym___thiscall] = ACTIONS(2412),
    [anon_sym___vectorcall] = ACTIONS(2412),
    [anon_sym_signed] = ACTIONS(2412),
    [anon_sym_unsigned] = ACTIONS(2412),
    [anon_sym_long] = ACTIONS(2412),
    [anon_sym_short] = ACTIONS(2412),
    [anon_sym_static] = ACTIONS(2412),
    [anon_sym_auto] = ACTIONS(2412),
    [anon_sym_register] = ACTIONS(2412),
    [anon_sym_inline] = ACTIONS(2412),
    [anon_sym___inline] = ACTIONS(2412),
    [anon_sym___inline__] = ACTIONS(2412),
    [anon_sym___forceinline] = ACTIONS(2412),
    [anon_sym_thread_local] = ACTIONS(2412),
    [anon_sym___thread] = ACTIONS(2412),
    [anon_sym_CG_EXTERN] = ACTIONS(2412),
    [anon_sym_CG_INLINE] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2412),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2412),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2412),
    [anon_sym_IBOutlet] = ACTIONS(2412),
    [anon_sym_IBInspectable] = ACTIONS(2412),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2412),
    [anon_sym_NS_INLINE] = ACTIONS(2412),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2412),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2412),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2412),
    [anon_sym_const] = ACTIONS(2412),
    [anon_sym_constexpr] = ACTIONS(2412),
    [anon_sym_volatile] = ACTIONS(2412),
    [anon_sym_restrict] = ACTIONS(2412),
    [anon_sym___restrict__] = ACTIONS(2412),
    [anon_sym__Atomic] = ACTIONS(2412),
    [anon_sym__Noreturn] = ACTIONS(2412),
    [anon_sym__Nonnull] = ACTIONS(2412),
    [anon_sym_nullable] = ACTIONS(2412),
    [anon_sym__Complex] = ACTIONS(2412),
    [anon_sym__Nullable] = ACTIONS(2412),
    [anon_sym__Nullable_result] = ACTIONS(2412),
    [anon_sym__Null_unspecified] = ACTIONS(2412),
    [anon_sym___autoreleasing] = ACTIONS(2412),
    [anon_sym___block] = ACTIONS(2412),
    [anon_sym___bridge] = ACTIONS(2412),
    [anon_sym___bridge_retained] = ACTIONS(2412),
    [anon_sym___bridge_transfer] = ACTIONS(2412),
    [anon_sym___complex] = ACTIONS(2412),
    [anon_sym___const] = ACTIONS(2412),
    [anon_sym___imag] = ACTIONS(2412),
    [anon_sym___kindof] = ACTIONS(2412),
    [anon_sym___nonnull] = ACTIONS(2412),
    [anon_sym___nullable] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2412),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2412),
    [anon_sym___real] = ACTIONS(2412),
    [anon_sym___strong] = ACTIONS(2412),
    [anon_sym___unsafe_unretained] = ACTIONS(2412),
    [anon_sym___unused] = ACTIONS(2412),
    [anon_sym___weak] = ACTIONS(2412),
    [anon_sym_alignas] = ACTIONS(2412),
    [anon_sym__Alignas] = ACTIONS(2412),
    [sym_primitive_type] = ACTIONS(2412),
    [anon_sym_enum] = ACTIONS(2412),
    [anon_sym_struct] = ACTIONS(2412),
    [anon_sym_union] = ACTIONS(2412),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2412),
    [anon_sym___typeof] = ACTIONS(2412),
    [anon_sym_typeof] = ACTIONS(2412),
    [aux_sym_preproc_undef_token1] = ACTIONS(2412),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2412),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2412),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2412),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2412),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2412),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_API_AVAILABLE] = ACTIONS(2412),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_API_DEPRECATED] = ACTIONS(2412),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2412),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2412),
    [anon_sym___deprecated_msg] = ACTIONS(2412),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2412),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2412),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2412),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2412),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2412),
    [anon_sym_ATsynthesize] = ACTIONS(2414),
    [anon_sym_ATdynamic] = ACTIONS(2414),
    [anon_sym_ATproperty] = ACTIONS(2414),
    [anon_sym_BOOL] = ACTIONS(2412),
    [anon_sym_IMP] = ACTIONS(2412),
    [anon_sym_SEL] = ACTIONS(2412),
    [anon_sym_Class] = ACTIONS(2412),
    [anon_sym_id] = ACTIONS(2412),
  },
  [STATE(2312)] = {
    [sym_identifier] = ACTIONS(5369),
    [aux_sym_preproc_def_token1] = ACTIONS(5369),
    [aux_sym_preproc_if_token1] = ACTIONS(5369),
    [aux_sym_preproc_if_token2] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5369),
    [sym_preproc_directive] = ACTIONS(5369),
    [anon_sym_DASH] = ACTIONS(5371),
    [anon_sym_PLUS] = ACTIONS(5371),
    [anon_sym_SEMI] = ACTIONS(5617),
    [anon_sym___extension__] = ACTIONS(5369),
    [anon_sym_typedef] = ACTIONS(5369),
    [anon_sym_extern] = ACTIONS(5369),
    [anon_sym___attribute__] = ACTIONS(5369),
    [anon_sym___attribute] = ACTIONS(5369),
    [anon_sym_noreturn] = ACTIONS(5369),
    [anon_sym_LBRACK] = ACTIONS(5371),
    [anon_sym___declspec] = ACTIONS(5369),
    [anon_sym___cdecl] = ACTIONS(5369),
    [anon_sym___clrcall] = ACTIONS(5369),
    [anon_sym___stdcall] = ACTIONS(5369),
    [anon_sym___fastcall] = ACTIONS(5369),
    [anon_sym___thiscall] = ACTIONS(5369),
    [anon_sym___vectorcall] = ACTIONS(5369),
    [anon_sym_signed] = ACTIONS(5369),
    [anon_sym_unsigned] = ACTIONS(5369),
    [anon_sym_long] = ACTIONS(5369),
    [anon_sym_short] = ACTIONS(5369),
    [anon_sym_static] = ACTIONS(5369),
    [anon_sym_auto] = ACTIONS(5369),
    [anon_sym_register] = ACTIONS(5369),
    [anon_sym_inline] = ACTIONS(5369),
    [anon_sym___inline] = ACTIONS(5369),
    [anon_sym___inline__] = ACTIONS(5369),
    [anon_sym___forceinline] = ACTIONS(5369),
    [anon_sym_thread_local] = ACTIONS(5369),
    [anon_sym___thread] = ACTIONS(5369),
    [anon_sym_CG_EXTERN] = ACTIONS(5369),
    [anon_sym_CG_INLINE] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5369),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5369),
    [anon_sym_IBOutlet] = ACTIONS(5369),
    [anon_sym_IBInspectable] = ACTIONS(5369),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5369),
    [anon_sym_NS_INLINE] = ACTIONS(5369),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5369),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5369),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5369),
    [anon_sym_const] = ACTIONS(5369),
    [anon_sym_constexpr] = ACTIONS(5369),
    [anon_sym_volatile] = ACTIONS(5369),
    [anon_sym_restrict] = ACTIONS(5369),
    [anon_sym___restrict__] = ACTIONS(5369),
    [anon_sym__Atomic] = ACTIONS(5369),
    [anon_sym__Noreturn] = ACTIONS(5369),
    [anon_sym__Nonnull] = ACTIONS(5369),
    [anon_sym_nullable] = ACTIONS(5369),
    [anon_sym__Complex] = ACTIONS(5369),
    [anon_sym__Nullable] = ACTIONS(5369),
    [anon_sym__Nullable_result] = ACTIONS(5369),
    [anon_sym__Null_unspecified] = ACTIONS(5369),
    [anon_sym___autoreleasing] = ACTIONS(5369),
    [anon_sym___block] = ACTIONS(5369),
    [anon_sym___bridge] = ACTIONS(5369),
    [anon_sym___bridge_retained] = ACTIONS(5369),
    [anon_sym___bridge_transfer] = ACTIONS(5369),
    [anon_sym___complex] = ACTIONS(5369),
    [anon_sym___const] = ACTIONS(5369),
    [anon_sym___imag] = ACTIONS(5369),
    [anon_sym___kindof] = ACTIONS(5369),
    [anon_sym___nonnull] = ACTIONS(5369),
    [anon_sym___nullable] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5369),
    [anon_sym___real] = ACTIONS(5369),
    [anon_sym___strong] = ACTIONS(5369),
    [anon_sym___unsafe_unretained] = ACTIONS(5369),
    [anon_sym___unused] = ACTIONS(5369),
    [anon_sym___weak] = ACTIONS(5369),
    [anon_sym_alignas] = ACTIONS(5369),
    [anon_sym__Alignas] = ACTIONS(5369),
    [sym_primitive_type] = ACTIONS(5369),
    [anon_sym_enum] = ACTIONS(5369),
    [anon_sym_struct] = ACTIONS(5369),
    [anon_sym_union] = ACTIONS(5369),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5369),
    [anon_sym___typeof] = ACTIONS(5369),
    [anon_sym_typeof] = ACTIONS(5369),
    [aux_sym_preproc_undef_token1] = ACTIONS(5369),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5369),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5369),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5369),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_API_AVAILABLE] = ACTIONS(5369),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_API_DEPRECATED] = ACTIONS(5369),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym___deprecated_msg] = ACTIONS(5369),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5369),
    [anon_sym_ATsynthesize] = ACTIONS(5371),
    [anon_sym_ATdynamic] = ACTIONS(5371),
    [anon_sym_BOOL] = ACTIONS(5369),
    [anon_sym_IMP] = ACTIONS(5369),
    [anon_sym_SEL] = ACTIONS(5369),
    [anon_sym_Class] = ACTIONS(5369),
    [anon_sym_id] = ACTIONS(5369),
  },
  [STATE(2313)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2321),
    [sym_identifier] = ACTIONS(5423),
    [aux_sym_preproc_def_token1] = ACTIONS(5423),
    [aux_sym_preproc_if_token1] = ACTIONS(5423),
    [aux_sym_preproc_if_token2] = ACTIONS(5423),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5423),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5423),
    [sym_preproc_directive] = ACTIONS(5423),
    [anon_sym_DASH] = ACTIONS(5425),
    [anon_sym_PLUS] = ACTIONS(5425),
    [anon_sym_SEMI] = ACTIONS(5565),
    [anon_sym___extension__] = ACTIONS(5423),
    [anon_sym_typedef] = ACTIONS(5423),
    [anon_sym_extern] = ACTIONS(5423),
    [anon_sym___attribute__] = ACTIONS(5423),
    [anon_sym___attribute] = ACTIONS(5423),
    [anon_sym_noreturn] = ACTIONS(5423),
    [anon_sym_LBRACK] = ACTIONS(5425),
    [anon_sym___declspec] = ACTIONS(5423),
    [anon_sym___cdecl] = ACTIONS(5423),
    [anon_sym___clrcall] = ACTIONS(5423),
    [anon_sym___stdcall] = ACTIONS(5423),
    [anon_sym___fastcall] = ACTIONS(5423),
    [anon_sym___thiscall] = ACTIONS(5423),
    [anon_sym___vectorcall] = ACTIONS(5423),
    [anon_sym_signed] = ACTIONS(5423),
    [anon_sym_unsigned] = ACTIONS(5423),
    [anon_sym_long] = ACTIONS(5423),
    [anon_sym_short] = ACTIONS(5423),
    [anon_sym_static] = ACTIONS(5423),
    [anon_sym_auto] = ACTIONS(5423),
    [anon_sym_register] = ACTIONS(5423),
    [anon_sym_inline] = ACTIONS(5423),
    [anon_sym___inline] = ACTIONS(5423),
    [anon_sym___inline__] = ACTIONS(5423),
    [anon_sym___forceinline] = ACTIONS(5423),
    [anon_sym_thread_local] = ACTIONS(5423),
    [anon_sym___thread] = ACTIONS(5423),
    [anon_sym_CG_EXTERN] = ACTIONS(5423),
    [anon_sym_CG_INLINE] = ACTIONS(5423),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5423),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5423),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5423),
    [anon_sym_IBOutlet] = ACTIONS(5423),
    [anon_sym_IBInspectable] = ACTIONS(5423),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5423),
    [anon_sym_NS_INLINE] = ACTIONS(5423),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5423),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5423),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5423),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5423),
    [anon_sym_const] = ACTIONS(5423),
    [anon_sym_constexpr] = ACTIONS(5423),
    [anon_sym_volatile] = ACTIONS(5423),
    [anon_sym_restrict] = ACTIONS(5423),
    [anon_sym___restrict__] = ACTIONS(5423),
    [anon_sym__Atomic] = ACTIONS(5423),
    [anon_sym__Noreturn] = ACTIONS(5423),
    [anon_sym__Nonnull] = ACTIONS(5423),
    [anon_sym_nullable] = ACTIONS(5423),
    [anon_sym__Complex] = ACTIONS(5423),
    [anon_sym__Nullable] = ACTIONS(5423),
    [anon_sym__Nullable_result] = ACTIONS(5423),
    [anon_sym__Null_unspecified] = ACTIONS(5423),
    [anon_sym___autoreleasing] = ACTIONS(5423),
    [anon_sym___block] = ACTIONS(5423),
    [anon_sym___bridge] = ACTIONS(5423),
    [anon_sym___bridge_retained] = ACTIONS(5423),
    [anon_sym___bridge_transfer] = ACTIONS(5423),
    [anon_sym___complex] = ACTIONS(5423),
    [anon_sym___const] = ACTIONS(5423),
    [anon_sym___imag] = ACTIONS(5423),
    [anon_sym___kindof] = ACTIONS(5423),
    [anon_sym___nonnull] = ACTIONS(5423),
    [anon_sym___nullable] = ACTIONS(5423),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5423),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5423),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5423),
    [anon_sym___real] = ACTIONS(5423),
    [anon_sym___strong] = ACTIONS(5423),
    [anon_sym___unsafe_unretained] = ACTIONS(5423),
    [anon_sym___unused] = ACTIONS(5423),
    [anon_sym___weak] = ACTIONS(5423),
    [anon_sym_alignas] = ACTIONS(5423),
    [anon_sym__Alignas] = ACTIONS(5423),
    [sym_primitive_type] = ACTIONS(5423),
    [anon_sym_enum] = ACTIONS(5423),
    [anon_sym_struct] = ACTIONS(5423),
    [anon_sym_union] = ACTIONS(5423),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5423),
    [anon_sym___typeof] = ACTIONS(5423),
    [anon_sym_typeof] = ACTIONS(5423),
    [aux_sym_preproc_undef_token1] = ACTIONS(5423),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5423),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5423),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5423),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5423),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5423),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5423),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5423),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5423),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5423),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_API_AVAILABLE] = ACTIONS(5423),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_API_DEPRECATED] = ACTIONS(5423),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5423),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5423),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5423),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5423),
    [anon_sym___deprecated_msg] = ACTIONS(5423),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5423),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5423),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5423),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5423),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5423),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5423),
    [anon_sym_ATproperty] = ACTIONS(5425),
    [anon_sym_BOOL] = ACTIONS(5423),
    [anon_sym_IMP] = ACTIONS(5423),
    [anon_sym_SEL] = ACTIONS(5423),
    [anon_sym_Class] = ACTIONS(5423),
    [anon_sym_id] = ACTIONS(5423),
  },
  [STATE(2314)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2321),
    [sym_identifier] = ACTIONS(5454),
    [aux_sym_preproc_def_token1] = ACTIONS(5454),
    [aux_sym_preproc_if_token1] = ACTIONS(5454),
    [aux_sym_preproc_if_token2] = ACTIONS(5454),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5454),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5454),
    [sym_preproc_directive] = ACTIONS(5454),
    [anon_sym_DASH] = ACTIONS(5456),
    [anon_sym_PLUS] = ACTIONS(5456),
    [anon_sym_SEMI] = ACTIONS(5565),
    [anon_sym___extension__] = ACTIONS(5454),
    [anon_sym_typedef] = ACTIONS(5454),
    [anon_sym_extern] = ACTIONS(5454),
    [anon_sym___attribute__] = ACTIONS(5454),
    [anon_sym___attribute] = ACTIONS(5454),
    [anon_sym_noreturn] = ACTIONS(5454),
    [anon_sym_LBRACK] = ACTIONS(5456),
    [anon_sym___declspec] = ACTIONS(5454),
    [anon_sym___cdecl] = ACTIONS(5454),
    [anon_sym___clrcall] = ACTIONS(5454),
    [anon_sym___stdcall] = ACTIONS(5454),
    [anon_sym___fastcall] = ACTIONS(5454),
    [anon_sym___thiscall] = ACTIONS(5454),
    [anon_sym___vectorcall] = ACTIONS(5454),
    [anon_sym_signed] = ACTIONS(5454),
    [anon_sym_unsigned] = ACTIONS(5454),
    [anon_sym_long] = ACTIONS(5454),
    [anon_sym_short] = ACTIONS(5454),
    [anon_sym_static] = ACTIONS(5454),
    [anon_sym_auto] = ACTIONS(5454),
    [anon_sym_register] = ACTIONS(5454),
    [anon_sym_inline] = ACTIONS(5454),
    [anon_sym___inline] = ACTIONS(5454),
    [anon_sym___inline__] = ACTIONS(5454),
    [anon_sym___forceinline] = ACTIONS(5454),
    [anon_sym_thread_local] = ACTIONS(5454),
    [anon_sym___thread] = ACTIONS(5454),
    [anon_sym_CG_EXTERN] = ACTIONS(5454),
    [anon_sym_CG_INLINE] = ACTIONS(5454),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5454),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5454),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5454),
    [anon_sym_IBOutlet] = ACTIONS(5454),
    [anon_sym_IBInspectable] = ACTIONS(5454),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5454),
    [anon_sym_NS_INLINE] = ACTIONS(5454),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5454),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5454),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5454),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5454),
    [anon_sym_const] = ACTIONS(5454),
    [anon_sym_constexpr] = ACTIONS(5454),
    [anon_sym_volatile] = ACTIONS(5454),
    [anon_sym_restrict] = ACTIONS(5454),
    [anon_sym___restrict__] = ACTIONS(5454),
    [anon_sym__Atomic] = ACTIONS(5454),
    [anon_sym__Noreturn] = ACTIONS(5454),
    [anon_sym__Nonnull] = ACTIONS(5454),
    [anon_sym_nullable] = ACTIONS(5454),
    [anon_sym__Complex] = ACTIONS(5454),
    [anon_sym__Nullable] = ACTIONS(5454),
    [anon_sym__Nullable_result] = ACTIONS(5454),
    [anon_sym__Null_unspecified] = ACTIONS(5454),
    [anon_sym___autoreleasing] = ACTIONS(5454),
    [anon_sym___block] = ACTIONS(5454),
    [anon_sym___bridge] = ACTIONS(5454),
    [anon_sym___bridge_retained] = ACTIONS(5454),
    [anon_sym___bridge_transfer] = ACTIONS(5454),
    [anon_sym___complex] = ACTIONS(5454),
    [anon_sym___const] = ACTIONS(5454),
    [anon_sym___imag] = ACTIONS(5454),
    [anon_sym___kindof] = ACTIONS(5454),
    [anon_sym___nonnull] = ACTIONS(5454),
    [anon_sym___nullable] = ACTIONS(5454),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5454),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5454),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5454),
    [anon_sym___real] = ACTIONS(5454),
    [anon_sym___strong] = ACTIONS(5454),
    [anon_sym___unsafe_unretained] = ACTIONS(5454),
    [anon_sym___unused] = ACTIONS(5454),
    [anon_sym___weak] = ACTIONS(5454),
    [anon_sym_alignas] = ACTIONS(5454),
    [anon_sym__Alignas] = ACTIONS(5454),
    [sym_primitive_type] = ACTIONS(5454),
    [anon_sym_enum] = ACTIONS(5454),
    [anon_sym_struct] = ACTIONS(5454),
    [anon_sym_union] = ACTIONS(5454),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5454),
    [anon_sym___typeof] = ACTIONS(5454),
    [anon_sym_typeof] = ACTIONS(5454),
    [aux_sym_preproc_undef_token1] = ACTIONS(5454),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5454),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5454),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5454),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5454),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5454),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5454),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5454),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5454),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5454),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_API_AVAILABLE] = ACTIONS(5454),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_API_DEPRECATED] = ACTIONS(5454),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5454),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5454),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5454),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5454),
    [anon_sym___deprecated_msg] = ACTIONS(5454),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5454),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5454),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5454),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5454),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5454),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5454),
    [anon_sym_ATproperty] = ACTIONS(5456),
    [anon_sym_BOOL] = ACTIONS(5454),
    [anon_sym_IMP] = ACTIONS(5454),
    [anon_sym_SEL] = ACTIONS(5454),
    [anon_sym_Class] = ACTIONS(5454),
    [anon_sym_id] = ACTIONS(5454),
  },
  [STATE(2315)] = {
    [sym_identifier] = ACTIONS(5375),
    [aux_sym_preproc_def_token1] = ACTIONS(5375),
    [aux_sym_preproc_if_token1] = ACTIONS(5375),
    [aux_sym_preproc_if_token2] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5375),
    [sym_preproc_directive] = ACTIONS(5375),
    [anon_sym_DASH] = ACTIONS(5377),
    [anon_sym_PLUS] = ACTIONS(5377),
    [anon_sym_SEMI] = ACTIONS(5619),
    [anon_sym___extension__] = ACTIONS(5375),
    [anon_sym_typedef] = ACTIONS(5375),
    [anon_sym_extern] = ACTIONS(5375),
    [anon_sym___attribute__] = ACTIONS(5375),
    [anon_sym___attribute] = ACTIONS(5375),
    [anon_sym_noreturn] = ACTIONS(5375),
    [anon_sym_LBRACK] = ACTIONS(5377),
    [anon_sym___declspec] = ACTIONS(5375),
    [anon_sym___cdecl] = ACTIONS(5375),
    [anon_sym___clrcall] = ACTIONS(5375),
    [anon_sym___stdcall] = ACTIONS(5375),
    [anon_sym___fastcall] = ACTIONS(5375),
    [anon_sym___thiscall] = ACTIONS(5375),
    [anon_sym___vectorcall] = ACTIONS(5375),
    [anon_sym_signed] = ACTIONS(5375),
    [anon_sym_unsigned] = ACTIONS(5375),
    [anon_sym_long] = ACTIONS(5375),
    [anon_sym_short] = ACTIONS(5375),
    [anon_sym_static] = ACTIONS(5375),
    [anon_sym_auto] = ACTIONS(5375),
    [anon_sym_register] = ACTIONS(5375),
    [anon_sym_inline] = ACTIONS(5375),
    [anon_sym___inline] = ACTIONS(5375),
    [anon_sym___inline__] = ACTIONS(5375),
    [anon_sym___forceinline] = ACTIONS(5375),
    [anon_sym_thread_local] = ACTIONS(5375),
    [anon_sym___thread] = ACTIONS(5375),
    [anon_sym_CG_EXTERN] = ACTIONS(5375),
    [anon_sym_CG_INLINE] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5375),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5375),
    [anon_sym_IBOutlet] = ACTIONS(5375),
    [anon_sym_IBInspectable] = ACTIONS(5375),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5375),
    [anon_sym_NS_INLINE] = ACTIONS(5375),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5375),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5375),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5375),
    [anon_sym_const] = ACTIONS(5375),
    [anon_sym_constexpr] = ACTIONS(5375),
    [anon_sym_volatile] = ACTIONS(5375),
    [anon_sym_restrict] = ACTIONS(5375),
    [anon_sym___restrict__] = ACTIONS(5375),
    [anon_sym__Atomic] = ACTIONS(5375),
    [anon_sym__Noreturn] = ACTIONS(5375),
    [anon_sym__Nonnull] = ACTIONS(5375),
    [anon_sym_nullable] = ACTIONS(5375),
    [anon_sym__Complex] = ACTIONS(5375),
    [anon_sym__Nullable] = ACTIONS(5375),
    [anon_sym__Nullable_result] = ACTIONS(5375),
    [anon_sym__Null_unspecified] = ACTIONS(5375),
    [anon_sym___autoreleasing] = ACTIONS(5375),
    [anon_sym___block] = ACTIONS(5375),
    [anon_sym___bridge] = ACTIONS(5375),
    [anon_sym___bridge_retained] = ACTIONS(5375),
    [anon_sym___bridge_transfer] = ACTIONS(5375),
    [anon_sym___complex] = ACTIONS(5375),
    [anon_sym___const] = ACTIONS(5375),
    [anon_sym___imag] = ACTIONS(5375),
    [anon_sym___kindof] = ACTIONS(5375),
    [anon_sym___nonnull] = ACTIONS(5375),
    [anon_sym___nullable] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5375),
    [anon_sym___real] = ACTIONS(5375),
    [anon_sym___strong] = ACTIONS(5375),
    [anon_sym___unsafe_unretained] = ACTIONS(5375),
    [anon_sym___unused] = ACTIONS(5375),
    [anon_sym___weak] = ACTIONS(5375),
    [anon_sym_alignas] = ACTIONS(5375),
    [anon_sym__Alignas] = ACTIONS(5375),
    [sym_primitive_type] = ACTIONS(5375),
    [anon_sym_enum] = ACTIONS(5375),
    [anon_sym_struct] = ACTIONS(5375),
    [anon_sym_union] = ACTIONS(5375),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5375),
    [anon_sym___typeof] = ACTIONS(5375),
    [anon_sym_typeof] = ACTIONS(5375),
    [aux_sym_preproc_undef_token1] = ACTIONS(5375),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5375),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5375),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5375),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_API_AVAILABLE] = ACTIONS(5375),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_API_DEPRECATED] = ACTIONS(5375),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym___deprecated_msg] = ACTIONS(5375),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5375),
    [anon_sym_ATsynthesize] = ACTIONS(5377),
    [anon_sym_ATdynamic] = ACTIONS(5377),
    [anon_sym_BOOL] = ACTIONS(5375),
    [anon_sym_IMP] = ACTIONS(5375),
    [anon_sym_SEL] = ACTIONS(5375),
    [anon_sym_Class] = ACTIONS(5375),
    [anon_sym_id] = ACTIONS(5375),
  },
  [STATE(2316)] = {
    [sym_identifier] = ACTIONS(2446),
    [aux_sym_preproc_def_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token1] = ACTIONS(2446),
    [aux_sym_preproc_if_token2] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2446),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2446),
    [sym_preproc_directive] = ACTIONS(2446),
    [anon_sym_DASH] = ACTIONS(2448),
    [anon_sym_PLUS] = ACTIONS(2448),
    [anon_sym___extension__] = ACTIONS(2446),
    [anon_sym_typedef] = ACTIONS(2446),
    [anon_sym_extern] = ACTIONS(2446),
    [anon_sym___attribute__] = ACTIONS(2446),
    [anon_sym___attribute] = ACTIONS(2446),
    [anon_sym_noreturn] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym___declspec] = ACTIONS(2446),
    [anon_sym___cdecl] = ACTIONS(2446),
    [anon_sym___clrcall] = ACTIONS(2446),
    [anon_sym___stdcall] = ACTIONS(2446),
    [anon_sym___fastcall] = ACTIONS(2446),
    [anon_sym___thiscall] = ACTIONS(2446),
    [anon_sym___vectorcall] = ACTIONS(2446),
    [anon_sym_signed] = ACTIONS(2446),
    [anon_sym_unsigned] = ACTIONS(2446),
    [anon_sym_long] = ACTIONS(2446),
    [anon_sym_short] = ACTIONS(2446),
    [anon_sym_static] = ACTIONS(2446),
    [anon_sym_auto] = ACTIONS(2446),
    [anon_sym_register] = ACTIONS(2446),
    [anon_sym_inline] = ACTIONS(2446),
    [anon_sym___inline] = ACTIONS(2446),
    [anon_sym___inline__] = ACTIONS(2446),
    [anon_sym___forceinline] = ACTIONS(2446),
    [anon_sym_thread_local] = ACTIONS(2446),
    [anon_sym___thread] = ACTIONS(2446),
    [anon_sym_CG_EXTERN] = ACTIONS(2446),
    [anon_sym_CG_INLINE] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2446),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2446),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2446),
    [anon_sym_IBOutlet] = ACTIONS(2446),
    [anon_sym_IBInspectable] = ACTIONS(2446),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2446),
    [anon_sym_NS_INLINE] = ACTIONS(2446),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2446),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2446),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2446),
    [anon_sym_const] = ACTIONS(2446),
    [anon_sym_constexpr] = ACTIONS(2446),
    [anon_sym_volatile] = ACTIONS(2446),
    [anon_sym_restrict] = ACTIONS(2446),
    [anon_sym___restrict__] = ACTIONS(2446),
    [anon_sym__Atomic] = ACTIONS(2446),
    [anon_sym__Noreturn] = ACTIONS(2446),
    [anon_sym__Nonnull] = ACTIONS(2446),
    [anon_sym_nullable] = ACTIONS(2446),
    [anon_sym__Complex] = ACTIONS(2446),
    [anon_sym__Nullable] = ACTIONS(2446),
    [anon_sym__Nullable_result] = ACTIONS(2446),
    [anon_sym__Null_unspecified] = ACTIONS(2446),
    [anon_sym___autoreleasing] = ACTIONS(2446),
    [anon_sym___block] = ACTIONS(2446),
    [anon_sym___bridge] = ACTIONS(2446),
    [anon_sym___bridge_retained] = ACTIONS(2446),
    [anon_sym___bridge_transfer] = ACTIONS(2446),
    [anon_sym___complex] = ACTIONS(2446),
    [anon_sym___const] = ACTIONS(2446),
    [anon_sym___imag] = ACTIONS(2446),
    [anon_sym___kindof] = ACTIONS(2446),
    [anon_sym___nonnull] = ACTIONS(2446),
    [anon_sym___nullable] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2446),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2446),
    [anon_sym___real] = ACTIONS(2446),
    [anon_sym___strong] = ACTIONS(2446),
    [anon_sym___unsafe_unretained] = ACTIONS(2446),
    [anon_sym___unused] = ACTIONS(2446),
    [anon_sym___weak] = ACTIONS(2446),
    [anon_sym_alignas] = ACTIONS(2446),
    [anon_sym__Alignas] = ACTIONS(2446),
    [sym_primitive_type] = ACTIONS(2446),
    [anon_sym_enum] = ACTIONS(2446),
    [anon_sym_struct] = ACTIONS(2446),
    [anon_sym_union] = ACTIONS(2446),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2446),
    [anon_sym___typeof] = ACTIONS(2446),
    [anon_sym_typeof] = ACTIONS(2446),
    [aux_sym_preproc_undef_token1] = ACTIONS(2446),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2446),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2446),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2446),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2446),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2446),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_API_AVAILABLE] = ACTIONS(2446),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_API_DEPRECATED] = ACTIONS(2446),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2446),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2446),
    [anon_sym___deprecated_msg] = ACTIONS(2446),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2446),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2446),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2446),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2446),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2446),
    [anon_sym_ATsynthesize] = ACTIONS(2448),
    [anon_sym_ATdynamic] = ACTIONS(2448),
    [anon_sym_ATproperty] = ACTIONS(2448),
    [anon_sym_BOOL] = ACTIONS(2446),
    [anon_sym_IMP] = ACTIONS(2446),
    [anon_sym_SEL] = ACTIONS(2446),
    [anon_sym_Class] = ACTIONS(2446),
    [anon_sym_id] = ACTIONS(2446),
  },
  [STATE(2317)] = {
    [sym_identifier] = ACTIONS(5569),
    [aux_sym_preproc_def_token1] = ACTIONS(5569),
    [aux_sym_preproc_if_token1] = ACTIONS(5569),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5569),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5569),
    [sym_preproc_directive] = ACTIONS(5569),
    [anon_sym_DASH] = ACTIONS(5571),
    [anon_sym_PLUS] = ACTIONS(5571),
    [anon_sym___extension__] = ACTIONS(5569),
    [anon_sym_typedef] = ACTIONS(5569),
    [anon_sym_extern] = ACTIONS(5569),
    [anon_sym___attribute__] = ACTIONS(5569),
    [anon_sym___attribute] = ACTIONS(5569),
    [anon_sym_noreturn] = ACTIONS(5569),
    [anon_sym_LBRACK] = ACTIONS(5571),
    [anon_sym___declspec] = ACTIONS(5569),
    [anon_sym___cdecl] = ACTIONS(5569),
    [anon_sym___clrcall] = ACTIONS(5569),
    [anon_sym___stdcall] = ACTIONS(5569),
    [anon_sym___fastcall] = ACTIONS(5569),
    [anon_sym___thiscall] = ACTIONS(5569),
    [anon_sym___vectorcall] = ACTIONS(5569),
    [anon_sym_signed] = ACTIONS(5569),
    [anon_sym_unsigned] = ACTIONS(5569),
    [anon_sym_long] = ACTIONS(5569),
    [anon_sym_short] = ACTIONS(5569),
    [anon_sym_static] = ACTIONS(5569),
    [anon_sym_auto] = ACTIONS(5569),
    [anon_sym_register] = ACTIONS(5569),
    [anon_sym_inline] = ACTIONS(5569),
    [anon_sym___inline] = ACTIONS(5569),
    [anon_sym___inline__] = ACTIONS(5569),
    [anon_sym___forceinline] = ACTIONS(5569),
    [anon_sym_thread_local] = ACTIONS(5569),
    [anon_sym___thread] = ACTIONS(5569),
    [anon_sym_CG_EXTERN] = ACTIONS(5569),
    [anon_sym_CG_INLINE] = ACTIONS(5569),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5569),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5569),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5569),
    [anon_sym_IBOutlet] = ACTIONS(5569),
    [anon_sym_IBInspectable] = ACTIONS(5569),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5569),
    [anon_sym_NS_INLINE] = ACTIONS(5569),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5569),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5569),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5569),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5569),
    [anon_sym_const] = ACTIONS(5569),
    [anon_sym_constexpr] = ACTIONS(5569),
    [anon_sym_volatile] = ACTIONS(5569),
    [anon_sym_restrict] = ACTIONS(5569),
    [anon_sym___restrict__] = ACTIONS(5569),
    [anon_sym__Atomic] = ACTIONS(5569),
    [anon_sym__Noreturn] = ACTIONS(5569),
    [anon_sym__Nonnull] = ACTIONS(5569),
    [anon_sym_nullable] = ACTIONS(5569),
    [anon_sym__Complex] = ACTIONS(5569),
    [anon_sym__Nullable] = ACTIONS(5569),
    [anon_sym__Nullable_result] = ACTIONS(5569),
    [anon_sym__Null_unspecified] = ACTIONS(5569),
    [anon_sym___autoreleasing] = ACTIONS(5569),
    [anon_sym___block] = ACTIONS(5569),
    [anon_sym___bridge] = ACTIONS(5569),
    [anon_sym___bridge_retained] = ACTIONS(5569),
    [anon_sym___bridge_transfer] = ACTIONS(5569),
    [anon_sym___complex] = ACTIONS(5569),
    [anon_sym___const] = ACTIONS(5569),
    [anon_sym___imag] = ACTIONS(5569),
    [anon_sym___kindof] = ACTIONS(5569),
    [anon_sym___nonnull] = ACTIONS(5569),
    [anon_sym___nullable] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5569),
    [anon_sym___real] = ACTIONS(5569),
    [anon_sym___strong] = ACTIONS(5569),
    [anon_sym___unsafe_unretained] = ACTIONS(5569),
    [anon_sym___unused] = ACTIONS(5569),
    [anon_sym___weak] = ACTIONS(5569),
    [anon_sym_alignas] = ACTIONS(5569),
    [anon_sym__Alignas] = ACTIONS(5569),
    [sym_primitive_type] = ACTIONS(5569),
    [anon_sym_enum] = ACTIONS(5569),
    [anon_sym_struct] = ACTIONS(5569),
    [anon_sym_union] = ACTIONS(5569),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5569),
    [anon_sym___typeof] = ACTIONS(5569),
    [anon_sym_typeof] = ACTIONS(5569),
    [aux_sym_preproc_undef_token1] = ACTIONS(5569),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5569),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5569),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5569),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5569),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5569),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5569),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5569),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5569),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5569),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_API_AVAILABLE] = ACTIONS(5569),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_API_DEPRECATED] = ACTIONS(5569),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5569),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5569),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5569),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5569),
    [anon_sym___deprecated_msg] = ACTIONS(5569),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5569),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5569),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5569),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5569),
    [anon_sym_ATend] = ACTIONS(5571),
    [anon_sym_AToptional] = ACTIONS(5571),
    [anon_sym_ATrequired] = ACTIONS(5571),
    [anon_sym_ATproperty] = ACTIONS(5571),
    [anon_sym_BOOL] = ACTIONS(5569),
    [anon_sym_IMP] = ACTIONS(5569),
    [anon_sym_SEL] = ACTIONS(5569),
    [anon_sym_Class] = ACTIONS(5569),
    [anon_sym_id] = ACTIONS(5569),
  },
  [STATE(2318)] = {
    [sym_identifier] = ACTIONS(3592),
    [aux_sym_preproc_def_token1] = ACTIONS(3592),
    [aux_sym_preproc_if_token1] = ACTIONS(3592),
    [aux_sym_preproc_if_token2] = ACTIONS(3592),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3592),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3592),
    [aux_sym_preproc_else_token1] = ACTIONS(3592),
    [aux_sym_preproc_elif_token1] = ACTIONS(3592),
    [sym_preproc_directive] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(5621),
    [anon_sym_PLUS] = ACTIONS(5621),
    [anon_sym___extension__] = ACTIONS(3592),
    [anon_sym_typedef] = ACTIONS(3592),
    [anon_sym_extern] = ACTIONS(3592),
    [anon_sym___attribute__] = ACTIONS(3592),
    [anon_sym___attribute] = ACTIONS(3592),
    [anon_sym_noreturn] = ACTIONS(3592),
    [anon_sym_LBRACK] = ACTIONS(5621),
    [anon_sym___declspec] = ACTIONS(3592),
    [anon_sym___cdecl] = ACTIONS(3592),
    [anon_sym___clrcall] = ACTIONS(3592),
    [anon_sym___stdcall] = ACTIONS(3592),
    [anon_sym___fastcall] = ACTIONS(3592),
    [anon_sym___thiscall] = ACTIONS(3592),
    [anon_sym___vectorcall] = ACTIONS(3592),
    [anon_sym_signed] = ACTIONS(3592),
    [anon_sym_unsigned] = ACTIONS(3592),
    [anon_sym_long] = ACTIONS(3592),
    [anon_sym_short] = ACTIONS(3592),
    [anon_sym_static] = ACTIONS(3592),
    [anon_sym_auto] = ACTIONS(3592),
    [anon_sym_register] = ACTIONS(3592),
    [anon_sym_inline] = ACTIONS(3592),
    [anon_sym___inline] = ACTIONS(3592),
    [anon_sym___inline__] = ACTIONS(3592),
    [anon_sym___forceinline] = ACTIONS(3592),
    [anon_sym_thread_local] = ACTIONS(3592),
    [anon_sym___thread] = ACTIONS(3592),
    [anon_sym_CG_EXTERN] = ACTIONS(3592),
    [anon_sym_CG_INLINE] = ACTIONS(3592),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3592),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3592),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3592),
    [anon_sym_IBOutlet] = ACTIONS(3592),
    [anon_sym_IBInspectable] = ACTIONS(3592),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3592),
    [anon_sym_NS_INLINE] = ACTIONS(3592),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3592),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3592),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3592),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3592),
    [anon_sym_const] = ACTIONS(3592),
    [anon_sym_constexpr] = ACTIONS(3592),
    [anon_sym_volatile] = ACTIONS(3592),
    [anon_sym_restrict] = ACTIONS(3592),
    [anon_sym___restrict__] = ACTIONS(3592),
    [anon_sym__Atomic] = ACTIONS(3592),
    [anon_sym__Noreturn] = ACTIONS(3592),
    [anon_sym__Nonnull] = ACTIONS(3592),
    [anon_sym_nullable] = ACTIONS(3592),
    [anon_sym__Complex] = ACTIONS(3592),
    [anon_sym__Nullable] = ACTIONS(3592),
    [anon_sym__Nullable_result] = ACTIONS(3592),
    [anon_sym__Null_unspecified] = ACTIONS(3592),
    [anon_sym___autoreleasing] = ACTIONS(3592),
    [anon_sym___block] = ACTIONS(3592),
    [anon_sym___bridge] = ACTIONS(3592),
    [anon_sym___bridge_retained] = ACTIONS(3592),
    [anon_sym___bridge_transfer] = ACTIONS(3592),
    [anon_sym___complex] = ACTIONS(3592),
    [anon_sym___const] = ACTIONS(3592),
    [anon_sym___imag] = ACTIONS(3592),
    [anon_sym___kindof] = ACTIONS(3592),
    [anon_sym___nonnull] = ACTIONS(3592),
    [anon_sym___nullable] = ACTIONS(3592),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3592),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3592),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3592),
    [anon_sym___real] = ACTIONS(3592),
    [anon_sym___strong] = ACTIONS(3592),
    [anon_sym___unsafe_unretained] = ACTIONS(3592),
    [anon_sym___unused] = ACTIONS(3592),
    [anon_sym___weak] = ACTIONS(3592),
    [anon_sym_alignas] = ACTIONS(3592),
    [anon_sym__Alignas] = ACTIONS(3592),
    [sym_primitive_type] = ACTIONS(3592),
    [anon_sym_enum] = ACTIONS(3592),
    [anon_sym_struct] = ACTIONS(3592),
    [anon_sym_union] = ACTIONS(3592),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3592),
    [anon_sym___typeof] = ACTIONS(3592),
    [anon_sym_typeof] = ACTIONS(3592),
    [aux_sym_preproc_undef_token1] = ACTIONS(3592),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3592),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3592),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3592),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3592),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3592),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3592),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3592),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3592),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3592),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3592),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3592),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3592),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3592),
    [anon_sym_API_AVAILABLE] = ACTIONS(3592),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3592),
    [anon_sym_API_DEPRECATED] = ACTIONS(3592),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3592),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3592),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3592),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3592),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3592),
    [anon_sym___deprecated_msg] = ACTIONS(3592),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3592),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3592),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3592),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3592),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3592),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3592),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3592),
    [anon_sym_ATproperty] = ACTIONS(5621),
    [anon_sym_BOOL] = ACTIONS(3592),
    [anon_sym_IMP] = ACTIONS(3592),
    [anon_sym_SEL] = ACTIONS(3592),
    [anon_sym_Class] = ACTIONS(3592),
    [anon_sym_id] = ACTIONS(3592),
  },
  [STATE(2319)] = {
    [sym_identifier] = ACTIONS(5589),
    [aux_sym_preproc_def_token1] = ACTIONS(5589),
    [aux_sym_preproc_if_token1] = ACTIONS(5589),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5589),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5589),
    [sym_preproc_directive] = ACTIONS(5589),
    [anon_sym_DASH] = ACTIONS(5591),
    [anon_sym_PLUS] = ACTIONS(5591),
    [anon_sym___extension__] = ACTIONS(5589),
    [anon_sym_typedef] = ACTIONS(5589),
    [anon_sym_extern] = ACTIONS(5589),
    [anon_sym___attribute__] = ACTIONS(5589),
    [anon_sym___attribute] = ACTIONS(5589),
    [anon_sym_noreturn] = ACTIONS(5589),
    [anon_sym_LBRACK] = ACTIONS(5591),
    [anon_sym___declspec] = ACTIONS(5589),
    [anon_sym___cdecl] = ACTIONS(5589),
    [anon_sym___clrcall] = ACTIONS(5589),
    [anon_sym___stdcall] = ACTIONS(5589),
    [anon_sym___fastcall] = ACTIONS(5589),
    [anon_sym___thiscall] = ACTIONS(5589),
    [anon_sym___vectorcall] = ACTIONS(5589),
    [anon_sym_signed] = ACTIONS(5589),
    [anon_sym_unsigned] = ACTIONS(5589),
    [anon_sym_long] = ACTIONS(5589),
    [anon_sym_short] = ACTIONS(5589),
    [anon_sym_static] = ACTIONS(5589),
    [anon_sym_auto] = ACTIONS(5589),
    [anon_sym_register] = ACTIONS(5589),
    [anon_sym_inline] = ACTIONS(5589),
    [anon_sym___inline] = ACTIONS(5589),
    [anon_sym___inline__] = ACTIONS(5589),
    [anon_sym___forceinline] = ACTIONS(5589),
    [anon_sym_thread_local] = ACTIONS(5589),
    [anon_sym___thread] = ACTIONS(5589),
    [anon_sym_CG_EXTERN] = ACTIONS(5589),
    [anon_sym_CG_INLINE] = ACTIONS(5589),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5589),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5589),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5589),
    [anon_sym_IBOutlet] = ACTIONS(5589),
    [anon_sym_IBInspectable] = ACTIONS(5589),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5589),
    [anon_sym_NS_INLINE] = ACTIONS(5589),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5589),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5589),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5589),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5589),
    [anon_sym_const] = ACTIONS(5589),
    [anon_sym_constexpr] = ACTIONS(5589),
    [anon_sym_volatile] = ACTIONS(5589),
    [anon_sym_restrict] = ACTIONS(5589),
    [anon_sym___restrict__] = ACTIONS(5589),
    [anon_sym__Atomic] = ACTIONS(5589),
    [anon_sym__Noreturn] = ACTIONS(5589),
    [anon_sym__Nonnull] = ACTIONS(5589),
    [anon_sym_nullable] = ACTIONS(5589),
    [anon_sym__Complex] = ACTIONS(5589),
    [anon_sym__Nullable] = ACTIONS(5589),
    [anon_sym__Nullable_result] = ACTIONS(5589),
    [anon_sym__Null_unspecified] = ACTIONS(5589),
    [anon_sym___autoreleasing] = ACTIONS(5589),
    [anon_sym___block] = ACTIONS(5589),
    [anon_sym___bridge] = ACTIONS(5589),
    [anon_sym___bridge_retained] = ACTIONS(5589),
    [anon_sym___bridge_transfer] = ACTIONS(5589),
    [anon_sym___complex] = ACTIONS(5589),
    [anon_sym___const] = ACTIONS(5589),
    [anon_sym___imag] = ACTIONS(5589),
    [anon_sym___kindof] = ACTIONS(5589),
    [anon_sym___nonnull] = ACTIONS(5589),
    [anon_sym___nullable] = ACTIONS(5589),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5589),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5589),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5589),
    [anon_sym___real] = ACTIONS(5589),
    [anon_sym___strong] = ACTIONS(5589),
    [anon_sym___unsafe_unretained] = ACTIONS(5589),
    [anon_sym___unused] = ACTIONS(5589),
    [anon_sym___weak] = ACTIONS(5589),
    [anon_sym_alignas] = ACTIONS(5589),
    [anon_sym__Alignas] = ACTIONS(5589),
    [sym_primitive_type] = ACTIONS(5589),
    [anon_sym_enum] = ACTIONS(5589),
    [anon_sym_struct] = ACTIONS(5589),
    [anon_sym_union] = ACTIONS(5589),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5589),
    [anon_sym___typeof] = ACTIONS(5589),
    [anon_sym_typeof] = ACTIONS(5589),
    [aux_sym_preproc_undef_token1] = ACTIONS(5589),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5589),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5589),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5589),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5589),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5589),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5589),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5589),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5589),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5589),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_API_AVAILABLE] = ACTIONS(5589),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_API_DEPRECATED] = ACTIONS(5589),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5589),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5589),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5589),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5589),
    [anon_sym___deprecated_msg] = ACTIONS(5589),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5589),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5589),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5589),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5589),
    [anon_sym_ATend] = ACTIONS(5591),
    [anon_sym_AToptional] = ACTIONS(5591),
    [anon_sym_ATrequired] = ACTIONS(5591),
    [anon_sym_ATproperty] = ACTIONS(5591),
    [anon_sym_BOOL] = ACTIONS(5589),
    [anon_sym_IMP] = ACTIONS(5589),
    [anon_sym_SEL] = ACTIONS(5589),
    [anon_sym_Class] = ACTIONS(5589),
    [anon_sym_id] = ACTIONS(5589),
  },
  [STATE(2320)] = {
    [sym_identifier] = ACTIONS(5595),
    [aux_sym_preproc_def_token1] = ACTIONS(5595),
    [aux_sym_preproc_if_token1] = ACTIONS(5595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5595),
    [sym_preproc_directive] = ACTIONS(5595),
    [anon_sym_DASH] = ACTIONS(5597),
    [anon_sym_PLUS] = ACTIONS(5597),
    [anon_sym___extension__] = ACTIONS(5595),
    [anon_sym_typedef] = ACTIONS(5595),
    [anon_sym_extern] = ACTIONS(5595),
    [anon_sym___attribute__] = ACTIONS(5595),
    [anon_sym___attribute] = ACTIONS(5595),
    [anon_sym_noreturn] = ACTIONS(5595),
    [anon_sym_LBRACK] = ACTIONS(5597),
    [anon_sym___declspec] = ACTIONS(5595),
    [anon_sym___cdecl] = ACTIONS(5595),
    [anon_sym___clrcall] = ACTIONS(5595),
    [anon_sym___stdcall] = ACTIONS(5595),
    [anon_sym___fastcall] = ACTIONS(5595),
    [anon_sym___thiscall] = ACTIONS(5595),
    [anon_sym___vectorcall] = ACTIONS(5595),
    [anon_sym_signed] = ACTIONS(5595),
    [anon_sym_unsigned] = ACTIONS(5595),
    [anon_sym_long] = ACTIONS(5595),
    [anon_sym_short] = ACTIONS(5595),
    [anon_sym_static] = ACTIONS(5595),
    [anon_sym_auto] = ACTIONS(5595),
    [anon_sym_register] = ACTIONS(5595),
    [anon_sym_inline] = ACTIONS(5595),
    [anon_sym___inline] = ACTIONS(5595),
    [anon_sym___inline__] = ACTIONS(5595),
    [anon_sym___forceinline] = ACTIONS(5595),
    [anon_sym_thread_local] = ACTIONS(5595),
    [anon_sym___thread] = ACTIONS(5595),
    [anon_sym_CG_EXTERN] = ACTIONS(5595),
    [anon_sym_CG_INLINE] = ACTIONS(5595),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5595),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5595),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5595),
    [anon_sym_IBOutlet] = ACTIONS(5595),
    [anon_sym_IBInspectable] = ACTIONS(5595),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5595),
    [anon_sym_NS_INLINE] = ACTIONS(5595),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5595),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5595),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5595),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5595),
    [anon_sym_const] = ACTIONS(5595),
    [anon_sym_constexpr] = ACTIONS(5595),
    [anon_sym_volatile] = ACTIONS(5595),
    [anon_sym_restrict] = ACTIONS(5595),
    [anon_sym___restrict__] = ACTIONS(5595),
    [anon_sym__Atomic] = ACTIONS(5595),
    [anon_sym__Noreturn] = ACTIONS(5595),
    [anon_sym__Nonnull] = ACTIONS(5595),
    [anon_sym_nullable] = ACTIONS(5595),
    [anon_sym__Complex] = ACTIONS(5595),
    [anon_sym__Nullable] = ACTIONS(5595),
    [anon_sym__Nullable_result] = ACTIONS(5595),
    [anon_sym__Null_unspecified] = ACTIONS(5595),
    [anon_sym___autoreleasing] = ACTIONS(5595),
    [anon_sym___block] = ACTIONS(5595),
    [anon_sym___bridge] = ACTIONS(5595),
    [anon_sym___bridge_retained] = ACTIONS(5595),
    [anon_sym___bridge_transfer] = ACTIONS(5595),
    [anon_sym___complex] = ACTIONS(5595),
    [anon_sym___const] = ACTIONS(5595),
    [anon_sym___imag] = ACTIONS(5595),
    [anon_sym___kindof] = ACTIONS(5595),
    [anon_sym___nonnull] = ACTIONS(5595),
    [anon_sym___nullable] = ACTIONS(5595),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5595),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5595),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5595),
    [anon_sym___real] = ACTIONS(5595),
    [anon_sym___strong] = ACTIONS(5595),
    [anon_sym___unsafe_unretained] = ACTIONS(5595),
    [anon_sym___unused] = ACTIONS(5595),
    [anon_sym___weak] = ACTIONS(5595),
    [anon_sym_alignas] = ACTIONS(5595),
    [anon_sym__Alignas] = ACTIONS(5595),
    [sym_primitive_type] = ACTIONS(5595),
    [anon_sym_enum] = ACTIONS(5595),
    [anon_sym_struct] = ACTIONS(5595),
    [anon_sym_union] = ACTIONS(5595),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5595),
    [anon_sym___typeof] = ACTIONS(5595),
    [anon_sym_typeof] = ACTIONS(5595),
    [aux_sym_preproc_undef_token1] = ACTIONS(5595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5595),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5595),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5595),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5595),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5595),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5595),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5595),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5595),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5595),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_API_AVAILABLE] = ACTIONS(5595),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_API_DEPRECATED] = ACTIONS(5595),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5595),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5595),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5595),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5595),
    [anon_sym___deprecated_msg] = ACTIONS(5595),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5595),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5595),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5595),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5595),
    [anon_sym_ATend] = ACTIONS(5597),
    [anon_sym_AToptional] = ACTIONS(5597),
    [anon_sym_ATrequired] = ACTIONS(5597),
    [anon_sym_ATproperty] = ACTIONS(5597),
    [anon_sym_BOOL] = ACTIONS(5595),
    [anon_sym_IMP] = ACTIONS(5595),
    [anon_sym_SEL] = ACTIONS(5595),
    [anon_sym_Class] = ACTIONS(5595),
    [anon_sym_id] = ACTIONS(5595),
  },
  [STATE(2321)] = {
    [aux_sym_method_declaration_repeat2] = STATE(2321),
    [sym_identifier] = ACTIONS(5431),
    [aux_sym_preproc_def_token1] = ACTIONS(5431),
    [aux_sym_preproc_if_token1] = ACTIONS(5431),
    [aux_sym_preproc_if_token2] = ACTIONS(5431),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5431),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5431),
    [sym_preproc_directive] = ACTIONS(5431),
    [anon_sym_DASH] = ACTIONS(5433),
    [anon_sym_PLUS] = ACTIONS(5433),
    [anon_sym_SEMI] = ACTIONS(5623),
    [anon_sym___extension__] = ACTIONS(5431),
    [anon_sym_typedef] = ACTIONS(5431),
    [anon_sym_extern] = ACTIONS(5431),
    [anon_sym___attribute__] = ACTIONS(5431),
    [anon_sym___attribute] = ACTIONS(5431),
    [anon_sym_noreturn] = ACTIONS(5431),
    [anon_sym_LBRACK] = ACTIONS(5433),
    [anon_sym___declspec] = ACTIONS(5431),
    [anon_sym___cdecl] = ACTIONS(5431),
    [anon_sym___clrcall] = ACTIONS(5431),
    [anon_sym___stdcall] = ACTIONS(5431),
    [anon_sym___fastcall] = ACTIONS(5431),
    [anon_sym___thiscall] = ACTIONS(5431),
    [anon_sym___vectorcall] = ACTIONS(5431),
    [anon_sym_signed] = ACTIONS(5431),
    [anon_sym_unsigned] = ACTIONS(5431),
    [anon_sym_long] = ACTIONS(5431),
    [anon_sym_short] = ACTIONS(5431),
    [anon_sym_static] = ACTIONS(5431),
    [anon_sym_auto] = ACTIONS(5431),
    [anon_sym_register] = ACTIONS(5431),
    [anon_sym_inline] = ACTIONS(5431),
    [anon_sym___inline] = ACTIONS(5431),
    [anon_sym___inline__] = ACTIONS(5431),
    [anon_sym___forceinline] = ACTIONS(5431),
    [anon_sym_thread_local] = ACTIONS(5431),
    [anon_sym___thread] = ACTIONS(5431),
    [anon_sym_CG_EXTERN] = ACTIONS(5431),
    [anon_sym_CG_INLINE] = ACTIONS(5431),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5431),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5431),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5431),
    [anon_sym_IBOutlet] = ACTIONS(5431),
    [anon_sym_IBInspectable] = ACTIONS(5431),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5431),
    [anon_sym_NS_INLINE] = ACTIONS(5431),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5431),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5431),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5431),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5431),
    [anon_sym_const] = ACTIONS(5431),
    [anon_sym_constexpr] = ACTIONS(5431),
    [anon_sym_volatile] = ACTIONS(5431),
    [anon_sym_restrict] = ACTIONS(5431),
    [anon_sym___restrict__] = ACTIONS(5431),
    [anon_sym__Atomic] = ACTIONS(5431),
    [anon_sym__Noreturn] = ACTIONS(5431),
    [anon_sym__Nonnull] = ACTIONS(5431),
    [anon_sym_nullable] = ACTIONS(5431),
    [anon_sym__Complex] = ACTIONS(5431),
    [anon_sym__Nullable] = ACTIONS(5431),
    [anon_sym__Nullable_result] = ACTIONS(5431),
    [anon_sym__Null_unspecified] = ACTIONS(5431),
    [anon_sym___autoreleasing] = ACTIONS(5431),
    [anon_sym___block] = ACTIONS(5431),
    [anon_sym___bridge] = ACTIONS(5431),
    [anon_sym___bridge_retained] = ACTIONS(5431),
    [anon_sym___bridge_transfer] = ACTIONS(5431),
    [anon_sym___complex] = ACTIONS(5431),
    [anon_sym___const] = ACTIONS(5431),
    [anon_sym___imag] = ACTIONS(5431),
    [anon_sym___kindof] = ACTIONS(5431),
    [anon_sym___nonnull] = ACTIONS(5431),
    [anon_sym___nullable] = ACTIONS(5431),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5431),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5431),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5431),
    [anon_sym___real] = ACTIONS(5431),
    [anon_sym___strong] = ACTIONS(5431),
    [anon_sym___unsafe_unretained] = ACTIONS(5431),
    [anon_sym___unused] = ACTIONS(5431),
    [anon_sym___weak] = ACTIONS(5431),
    [anon_sym_alignas] = ACTIONS(5431),
    [anon_sym__Alignas] = ACTIONS(5431),
    [sym_primitive_type] = ACTIONS(5431),
    [anon_sym_enum] = ACTIONS(5431),
    [anon_sym_struct] = ACTIONS(5431),
    [anon_sym_union] = ACTIONS(5431),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5431),
    [anon_sym___typeof] = ACTIONS(5431),
    [anon_sym_typeof] = ACTIONS(5431),
    [aux_sym_preproc_undef_token1] = ACTIONS(5431),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5431),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5431),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5431),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5431),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5431),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5431),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5431),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5431),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5431),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_API_AVAILABLE] = ACTIONS(5431),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_API_DEPRECATED] = ACTIONS(5431),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5431),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5431),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5431),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5431),
    [anon_sym___deprecated_msg] = ACTIONS(5431),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5431),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5431),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5431),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5431),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5431),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5431),
    [anon_sym_ATproperty] = ACTIONS(5433),
    [anon_sym_BOOL] = ACTIONS(5431),
    [anon_sym_IMP] = ACTIONS(5431),
    [anon_sym_SEL] = ACTIONS(5431),
    [anon_sym_Class] = ACTIONS(5431),
    [anon_sym_id] = ACTIONS(5431),
  },
  [STATE(2322)] = {
    [sym_identifier] = ACTIONS(5326),
    [aux_sym_preproc_def_token1] = ACTIONS(5326),
    [aux_sym_preproc_if_token1] = ACTIONS(5326),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5326),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5326),
    [sym_preproc_directive] = ACTIONS(5326),
    [anon_sym_DASH] = ACTIONS(5330),
    [anon_sym_PLUS] = ACTIONS(5330),
    [anon_sym___extension__] = ACTIONS(5326),
    [anon_sym_typedef] = ACTIONS(5326),
    [anon_sym_extern] = ACTIONS(5326),
    [anon_sym___attribute__] = ACTIONS(5326),
    [anon_sym___attribute] = ACTIONS(5326),
    [anon_sym_noreturn] = ACTIONS(5326),
    [anon_sym_LBRACK] = ACTIONS(5330),
    [anon_sym___declspec] = ACTIONS(5326),
    [anon_sym___cdecl] = ACTIONS(5326),
    [anon_sym___clrcall] = ACTIONS(5326),
    [anon_sym___stdcall] = ACTIONS(5326),
    [anon_sym___fastcall] = ACTIONS(5326),
    [anon_sym___thiscall] = ACTIONS(5326),
    [anon_sym___vectorcall] = ACTIONS(5326),
    [anon_sym_LBRACE] = ACTIONS(5330),
    [anon_sym_signed] = ACTIONS(5326),
    [anon_sym_unsigned] = ACTIONS(5326),
    [anon_sym_long] = ACTIONS(5326),
    [anon_sym_short] = ACTIONS(5326),
    [anon_sym_static] = ACTIONS(5326),
    [anon_sym_auto] = ACTIONS(5326),
    [anon_sym_register] = ACTIONS(5326),
    [anon_sym_inline] = ACTIONS(5326),
    [anon_sym___inline] = ACTIONS(5326),
    [anon_sym___inline__] = ACTIONS(5326),
    [anon_sym___forceinline] = ACTIONS(5326),
    [anon_sym_thread_local] = ACTIONS(5326),
    [anon_sym___thread] = ACTIONS(5326),
    [anon_sym_CG_EXTERN] = ACTIONS(5326),
    [anon_sym_CG_INLINE] = ACTIONS(5326),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5326),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5326),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5326),
    [anon_sym_IBOutlet] = ACTIONS(5326),
    [anon_sym_IBInspectable] = ACTIONS(5326),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5326),
    [anon_sym_NS_INLINE] = ACTIONS(5326),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5326),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5326),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5326),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5326),
    [anon_sym_const] = ACTIONS(5326),
    [anon_sym_constexpr] = ACTIONS(5326),
    [anon_sym_volatile] = ACTIONS(5326),
    [anon_sym_restrict] = ACTIONS(5326),
    [anon_sym___restrict__] = ACTIONS(5326),
    [anon_sym__Atomic] = ACTIONS(5326),
    [anon_sym__Noreturn] = ACTIONS(5326),
    [anon_sym__Nonnull] = ACTIONS(5326),
    [anon_sym_nullable] = ACTIONS(5326),
    [anon_sym__Complex] = ACTIONS(5326),
    [anon_sym__Nullable] = ACTIONS(5326),
    [anon_sym__Nullable_result] = ACTIONS(5326),
    [anon_sym__Null_unspecified] = ACTIONS(5326),
    [anon_sym___autoreleasing] = ACTIONS(5326),
    [anon_sym___block] = ACTIONS(5326),
    [anon_sym___bridge] = ACTIONS(5326),
    [anon_sym___bridge_retained] = ACTIONS(5326),
    [anon_sym___bridge_transfer] = ACTIONS(5326),
    [anon_sym___complex] = ACTIONS(5326),
    [anon_sym___const] = ACTIONS(5326),
    [anon_sym___imag] = ACTIONS(5326),
    [anon_sym___kindof] = ACTIONS(5326),
    [anon_sym___nonnull] = ACTIONS(5326),
    [anon_sym___nullable] = ACTIONS(5326),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5326),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5326),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5326),
    [anon_sym___real] = ACTIONS(5326),
    [anon_sym___strong] = ACTIONS(5326),
    [anon_sym___unsafe_unretained] = ACTIONS(5326),
    [anon_sym___unused] = ACTIONS(5326),
    [anon_sym___weak] = ACTIONS(5326),
    [anon_sym_alignas] = ACTIONS(5326),
    [anon_sym__Alignas] = ACTIONS(5326),
    [sym_primitive_type] = ACTIONS(5326),
    [anon_sym_enum] = ACTIONS(5326),
    [anon_sym_struct] = ACTIONS(5326),
    [anon_sym_union] = ACTIONS(5326),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5326),
    [anon_sym___typeof] = ACTIONS(5326),
    [anon_sym_typeof] = ACTIONS(5326),
    [aux_sym_preproc_undef_token1] = ACTIONS(5326),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5326),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5326),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5326),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5326),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5326),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5326),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5326),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5326),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5326),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5326),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5326),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5326),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5326),
    [anon_sym_API_AVAILABLE] = ACTIONS(5326),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5326),
    [anon_sym_API_DEPRECATED] = ACTIONS(5326),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5326),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5326),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5326),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5326),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5326),
    [anon_sym___deprecated_msg] = ACTIONS(5326),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5326),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5326),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5326),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5326),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5326),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5326),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5326),
    [anon_sym_ATend] = ACTIONS(5330),
    [anon_sym_ATsynthesize] = ACTIONS(5330),
    [anon_sym_ATdynamic] = ACTIONS(5330),
    [anon_sym_BOOL] = ACTIONS(5326),
    [anon_sym_IMP] = ACTIONS(5326),
    [anon_sym_SEL] = ACTIONS(5326),
    [anon_sym_Class] = ACTIONS(5326),
    [anon_sym_id] = ACTIONS(5326),
  },
  [STATE(2323)] = {
    [sym_identifier] = ACTIONS(5577),
    [aux_sym_preproc_def_token1] = ACTIONS(5577),
    [aux_sym_preproc_if_token1] = ACTIONS(5577),
    [aux_sym_preproc_if_token2] = ACTIONS(5577),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5577),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5577),
    [aux_sym_preproc_else_token1] = ACTIONS(5577),
    [aux_sym_preproc_elif_token1] = ACTIONS(5577),
    [sym_preproc_directive] = ACTIONS(5577),
    [anon_sym_DASH] = ACTIONS(5579),
    [anon_sym_PLUS] = ACTIONS(5579),
    [anon_sym___extension__] = ACTIONS(5577),
    [anon_sym_typedef] = ACTIONS(5577),
    [anon_sym_extern] = ACTIONS(5577),
    [anon_sym___attribute__] = ACTIONS(5577),
    [anon_sym___attribute] = ACTIONS(5577),
    [anon_sym_noreturn] = ACTIONS(5577),
    [anon_sym_LBRACK] = ACTIONS(5579),
    [anon_sym___declspec] = ACTIONS(5577),
    [anon_sym___cdecl] = ACTIONS(5577),
    [anon_sym___clrcall] = ACTIONS(5577),
    [anon_sym___stdcall] = ACTIONS(5577),
    [anon_sym___fastcall] = ACTIONS(5577),
    [anon_sym___thiscall] = ACTIONS(5577),
    [anon_sym___vectorcall] = ACTIONS(5577),
    [anon_sym_signed] = ACTIONS(5577),
    [anon_sym_unsigned] = ACTIONS(5577),
    [anon_sym_long] = ACTIONS(5577),
    [anon_sym_short] = ACTIONS(5577),
    [anon_sym_static] = ACTIONS(5577),
    [anon_sym_auto] = ACTIONS(5577),
    [anon_sym_register] = ACTIONS(5577),
    [anon_sym_inline] = ACTIONS(5577),
    [anon_sym___inline] = ACTIONS(5577),
    [anon_sym___inline__] = ACTIONS(5577),
    [anon_sym___forceinline] = ACTIONS(5577),
    [anon_sym_thread_local] = ACTIONS(5577),
    [anon_sym___thread] = ACTIONS(5577),
    [anon_sym_CG_EXTERN] = ACTIONS(5577),
    [anon_sym_CG_INLINE] = ACTIONS(5577),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5577),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5577),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5577),
    [anon_sym_IBOutlet] = ACTIONS(5577),
    [anon_sym_IBInspectable] = ACTIONS(5577),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5577),
    [anon_sym_NS_INLINE] = ACTIONS(5577),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5577),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5577),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5577),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5577),
    [anon_sym_const] = ACTIONS(5577),
    [anon_sym_constexpr] = ACTIONS(5577),
    [anon_sym_volatile] = ACTIONS(5577),
    [anon_sym_restrict] = ACTIONS(5577),
    [anon_sym___restrict__] = ACTIONS(5577),
    [anon_sym__Atomic] = ACTIONS(5577),
    [anon_sym__Noreturn] = ACTIONS(5577),
    [anon_sym__Nonnull] = ACTIONS(5577),
    [anon_sym_nullable] = ACTIONS(5577),
    [anon_sym__Complex] = ACTIONS(5577),
    [anon_sym__Nullable] = ACTIONS(5577),
    [anon_sym__Nullable_result] = ACTIONS(5577),
    [anon_sym__Null_unspecified] = ACTIONS(5577),
    [anon_sym___autoreleasing] = ACTIONS(5577),
    [anon_sym___block] = ACTIONS(5577),
    [anon_sym___bridge] = ACTIONS(5577),
    [anon_sym___bridge_retained] = ACTIONS(5577),
    [anon_sym___bridge_transfer] = ACTIONS(5577),
    [anon_sym___complex] = ACTIONS(5577),
    [anon_sym___const] = ACTIONS(5577),
    [anon_sym___imag] = ACTIONS(5577),
    [anon_sym___kindof] = ACTIONS(5577),
    [anon_sym___nonnull] = ACTIONS(5577),
    [anon_sym___nullable] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5577),
    [anon_sym___real] = ACTIONS(5577),
    [anon_sym___strong] = ACTIONS(5577),
    [anon_sym___unsafe_unretained] = ACTIONS(5577),
    [anon_sym___unused] = ACTIONS(5577),
    [anon_sym___weak] = ACTIONS(5577),
    [anon_sym_alignas] = ACTIONS(5577),
    [anon_sym__Alignas] = ACTIONS(5577),
    [sym_primitive_type] = ACTIONS(5577),
    [anon_sym_enum] = ACTIONS(5577),
    [anon_sym_struct] = ACTIONS(5577),
    [anon_sym_union] = ACTIONS(5577),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5577),
    [anon_sym___typeof] = ACTIONS(5577),
    [anon_sym_typeof] = ACTIONS(5577),
    [aux_sym_preproc_undef_token1] = ACTIONS(5577),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5577),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5577),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5577),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5577),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5577),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5577),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5577),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5577),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5577),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_API_AVAILABLE] = ACTIONS(5577),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_API_DEPRECATED] = ACTIONS(5577),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5577),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5577),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5577),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5577),
    [anon_sym___deprecated_msg] = ACTIONS(5577),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5577),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5577),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5577),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5577),
    [anon_sym_ATproperty] = ACTIONS(5579),
    [anon_sym_BOOL] = ACTIONS(5577),
    [anon_sym_IMP] = ACTIONS(5577),
    [anon_sym_SEL] = ACTIONS(5577),
    [anon_sym_Class] = ACTIONS(5577),
    [anon_sym_id] = ACTIONS(5577),
  },
  [STATE(2324)] = {
    [sym_identifier] = ACTIONS(2438),
    [aux_sym_preproc_def_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token1] = ACTIONS(2438),
    [aux_sym_preproc_if_token2] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2438),
    [sym_preproc_directive] = ACTIONS(2438),
    [anon_sym_DASH] = ACTIONS(2440),
    [anon_sym_PLUS] = ACTIONS(2440),
    [anon_sym___extension__] = ACTIONS(2438),
    [anon_sym_typedef] = ACTIONS(2438),
    [anon_sym_extern] = ACTIONS(2438),
    [anon_sym___attribute__] = ACTIONS(2438),
    [anon_sym___attribute] = ACTIONS(2438),
    [anon_sym_noreturn] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(2440),
    [anon_sym___declspec] = ACTIONS(2438),
    [anon_sym___cdecl] = ACTIONS(2438),
    [anon_sym___clrcall] = ACTIONS(2438),
    [anon_sym___stdcall] = ACTIONS(2438),
    [anon_sym___fastcall] = ACTIONS(2438),
    [anon_sym___thiscall] = ACTIONS(2438),
    [anon_sym___vectorcall] = ACTIONS(2438),
    [anon_sym_signed] = ACTIONS(2438),
    [anon_sym_unsigned] = ACTIONS(2438),
    [anon_sym_long] = ACTIONS(2438),
    [anon_sym_short] = ACTIONS(2438),
    [anon_sym_static] = ACTIONS(2438),
    [anon_sym_auto] = ACTIONS(2438),
    [anon_sym_register] = ACTIONS(2438),
    [anon_sym_inline] = ACTIONS(2438),
    [anon_sym___inline] = ACTIONS(2438),
    [anon_sym___inline__] = ACTIONS(2438),
    [anon_sym___forceinline] = ACTIONS(2438),
    [anon_sym_thread_local] = ACTIONS(2438),
    [anon_sym___thread] = ACTIONS(2438),
    [anon_sym_CG_EXTERN] = ACTIONS(2438),
    [anon_sym_CG_INLINE] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2438),
    [anon_sym_IBOutlet] = ACTIONS(2438),
    [anon_sym_IBInspectable] = ACTIONS(2438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2438),
    [anon_sym_NS_INLINE] = ACTIONS(2438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2438),
    [anon_sym_const] = ACTIONS(2438),
    [anon_sym_constexpr] = ACTIONS(2438),
    [anon_sym_volatile] = ACTIONS(2438),
    [anon_sym_restrict] = ACTIONS(2438),
    [anon_sym___restrict__] = ACTIONS(2438),
    [anon_sym__Atomic] = ACTIONS(2438),
    [anon_sym__Noreturn] = ACTIONS(2438),
    [anon_sym__Nonnull] = ACTIONS(2438),
    [anon_sym_nullable] = ACTIONS(2438),
    [anon_sym__Complex] = ACTIONS(2438),
    [anon_sym__Nullable] = ACTIONS(2438),
    [anon_sym__Nullable_result] = ACTIONS(2438),
    [anon_sym__Null_unspecified] = ACTIONS(2438),
    [anon_sym___autoreleasing] = ACTIONS(2438),
    [anon_sym___block] = ACTIONS(2438),
    [anon_sym___bridge] = ACTIONS(2438),
    [anon_sym___bridge_retained] = ACTIONS(2438),
    [anon_sym___bridge_transfer] = ACTIONS(2438),
    [anon_sym___complex] = ACTIONS(2438),
    [anon_sym___const] = ACTIONS(2438),
    [anon_sym___imag] = ACTIONS(2438),
    [anon_sym___kindof] = ACTIONS(2438),
    [anon_sym___nonnull] = ACTIONS(2438),
    [anon_sym___nullable] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2438),
    [anon_sym___real] = ACTIONS(2438),
    [anon_sym___strong] = ACTIONS(2438),
    [anon_sym___unsafe_unretained] = ACTIONS(2438),
    [anon_sym___unused] = ACTIONS(2438),
    [anon_sym___weak] = ACTIONS(2438),
    [anon_sym_alignas] = ACTIONS(2438),
    [anon_sym__Alignas] = ACTIONS(2438),
    [sym_primitive_type] = ACTIONS(2438),
    [anon_sym_enum] = ACTIONS(2438),
    [anon_sym_struct] = ACTIONS(2438),
    [anon_sym_union] = ACTIONS(2438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2438),
    [anon_sym___typeof] = ACTIONS(2438),
    [anon_sym_typeof] = ACTIONS(2438),
    [aux_sym_preproc_undef_token1] = ACTIONS(2438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_API_AVAILABLE] = ACTIONS(2438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_API_DEPRECATED] = ACTIONS(2438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2438),
    [anon_sym___deprecated_msg] = ACTIONS(2438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2438),
    [anon_sym_ATsynthesize] = ACTIONS(2440),
    [anon_sym_ATdynamic] = ACTIONS(2440),
    [anon_sym_ATproperty] = ACTIONS(2440),
    [anon_sym_BOOL] = ACTIONS(2438),
    [anon_sym_IMP] = ACTIONS(2438),
    [anon_sym_SEL] = ACTIONS(2438),
    [anon_sym_Class] = ACTIONS(2438),
    [anon_sym_id] = ACTIONS(2438),
  },
  [STATE(2325)] = {
    [sym_identifier] = ACTIONS(2442),
    [aux_sym_preproc_def_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token1] = ACTIONS(2442),
    [aux_sym_preproc_if_token2] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2442),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2442),
    [sym_preproc_directive] = ACTIONS(2442),
    [anon_sym_DASH] = ACTIONS(2444),
    [anon_sym_PLUS] = ACTIONS(2444),
    [anon_sym___extension__] = ACTIONS(2442),
    [anon_sym_typedef] = ACTIONS(2442),
    [anon_sym_extern] = ACTIONS(2442),
    [anon_sym___attribute__] = ACTIONS(2442),
    [anon_sym___attribute] = ACTIONS(2442),
    [anon_sym_noreturn] = ACTIONS(2442),
    [anon_sym_LBRACK] = ACTIONS(2444),
    [anon_sym___declspec] = ACTIONS(2442),
    [anon_sym___cdecl] = ACTIONS(2442),
    [anon_sym___clrcall] = ACTIONS(2442),
    [anon_sym___stdcall] = ACTIONS(2442),
    [anon_sym___fastcall] = ACTIONS(2442),
    [anon_sym___thiscall] = ACTIONS(2442),
    [anon_sym___vectorcall] = ACTIONS(2442),
    [anon_sym_signed] = ACTIONS(2442),
    [anon_sym_unsigned] = ACTIONS(2442),
    [anon_sym_long] = ACTIONS(2442),
    [anon_sym_short] = ACTIONS(2442),
    [anon_sym_static] = ACTIONS(2442),
    [anon_sym_auto] = ACTIONS(2442),
    [anon_sym_register] = ACTIONS(2442),
    [anon_sym_inline] = ACTIONS(2442),
    [anon_sym___inline] = ACTIONS(2442),
    [anon_sym___inline__] = ACTIONS(2442),
    [anon_sym___forceinline] = ACTIONS(2442),
    [anon_sym_thread_local] = ACTIONS(2442),
    [anon_sym___thread] = ACTIONS(2442),
    [anon_sym_CG_EXTERN] = ACTIONS(2442),
    [anon_sym_CG_INLINE] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2442),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2442),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2442),
    [anon_sym_IBOutlet] = ACTIONS(2442),
    [anon_sym_IBInspectable] = ACTIONS(2442),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2442),
    [anon_sym_NS_INLINE] = ACTIONS(2442),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2442),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2442),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2442),
    [anon_sym_const] = ACTIONS(2442),
    [anon_sym_constexpr] = ACTIONS(2442),
    [anon_sym_volatile] = ACTIONS(2442),
    [anon_sym_restrict] = ACTIONS(2442),
    [anon_sym___restrict__] = ACTIONS(2442),
    [anon_sym__Atomic] = ACTIONS(2442),
    [anon_sym__Noreturn] = ACTIONS(2442),
    [anon_sym__Nonnull] = ACTIONS(2442),
    [anon_sym_nullable] = ACTIONS(2442),
    [anon_sym__Complex] = ACTIONS(2442),
    [anon_sym__Nullable] = ACTIONS(2442),
    [anon_sym__Nullable_result] = ACTIONS(2442),
    [anon_sym__Null_unspecified] = ACTIONS(2442),
    [anon_sym___autoreleasing] = ACTIONS(2442),
    [anon_sym___block] = ACTIONS(2442),
    [anon_sym___bridge] = ACTIONS(2442),
    [anon_sym___bridge_retained] = ACTIONS(2442),
    [anon_sym___bridge_transfer] = ACTIONS(2442),
    [anon_sym___complex] = ACTIONS(2442),
    [anon_sym___const] = ACTIONS(2442),
    [anon_sym___imag] = ACTIONS(2442),
    [anon_sym___kindof] = ACTIONS(2442),
    [anon_sym___nonnull] = ACTIONS(2442),
    [anon_sym___nullable] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2442),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2442),
    [anon_sym___real] = ACTIONS(2442),
    [anon_sym___strong] = ACTIONS(2442),
    [anon_sym___unsafe_unretained] = ACTIONS(2442),
    [anon_sym___unused] = ACTIONS(2442),
    [anon_sym___weak] = ACTIONS(2442),
    [anon_sym_alignas] = ACTIONS(2442),
    [anon_sym__Alignas] = ACTIONS(2442),
    [sym_primitive_type] = ACTIONS(2442),
    [anon_sym_enum] = ACTIONS(2442),
    [anon_sym_struct] = ACTIONS(2442),
    [anon_sym_union] = ACTIONS(2442),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2442),
    [anon_sym___typeof] = ACTIONS(2442),
    [anon_sym_typeof] = ACTIONS(2442),
    [aux_sym_preproc_undef_token1] = ACTIONS(2442),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2442),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2442),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2442),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2442),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2442),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_API_AVAILABLE] = ACTIONS(2442),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_API_DEPRECATED] = ACTIONS(2442),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2442),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2442),
    [anon_sym___deprecated_msg] = ACTIONS(2442),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2442),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2442),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2442),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2442),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2442),
    [anon_sym_ATsynthesize] = ACTIONS(2444),
    [anon_sym_ATdynamic] = ACTIONS(2444),
    [anon_sym_ATproperty] = ACTIONS(2444),
    [anon_sym_BOOL] = ACTIONS(2442),
    [anon_sym_IMP] = ACTIONS(2442),
    [anon_sym_SEL] = ACTIONS(2442),
    [anon_sym_Class] = ACTIONS(2442),
    [anon_sym_id] = ACTIONS(2442),
  },
  [STATE(2326)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token2] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_PLUS] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATproperty] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(2327)] = {
    [sym_identifier] = ACTIONS(5537),
    [aux_sym_preproc_def_token1] = ACTIONS(5537),
    [aux_sym_preproc_if_token1] = ACTIONS(5537),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5537),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5537),
    [sym_preproc_directive] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5539),
    [anon_sym_PLUS] = ACTIONS(5539),
    [anon_sym___extension__] = ACTIONS(5537),
    [anon_sym_typedef] = ACTIONS(5537),
    [anon_sym_extern] = ACTIONS(5537),
    [anon_sym___attribute__] = ACTIONS(5537),
    [anon_sym___attribute] = ACTIONS(5537),
    [anon_sym_noreturn] = ACTIONS(5537),
    [anon_sym_LBRACK] = ACTIONS(5539),
    [anon_sym___declspec] = ACTIONS(5537),
    [anon_sym___cdecl] = ACTIONS(5537),
    [anon_sym___clrcall] = ACTIONS(5537),
    [anon_sym___stdcall] = ACTIONS(5537),
    [anon_sym___fastcall] = ACTIONS(5537),
    [anon_sym___thiscall] = ACTIONS(5537),
    [anon_sym___vectorcall] = ACTIONS(5537),
    [anon_sym_signed] = ACTIONS(5537),
    [anon_sym_unsigned] = ACTIONS(5537),
    [anon_sym_long] = ACTIONS(5537),
    [anon_sym_short] = ACTIONS(5537),
    [anon_sym_static] = ACTIONS(5537),
    [anon_sym_auto] = ACTIONS(5537),
    [anon_sym_register] = ACTIONS(5537),
    [anon_sym_inline] = ACTIONS(5537),
    [anon_sym___inline] = ACTIONS(5537),
    [anon_sym___inline__] = ACTIONS(5537),
    [anon_sym___forceinline] = ACTIONS(5537),
    [anon_sym_thread_local] = ACTIONS(5537),
    [anon_sym___thread] = ACTIONS(5537),
    [anon_sym_CG_EXTERN] = ACTIONS(5537),
    [anon_sym_CG_INLINE] = ACTIONS(5537),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5537),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5537),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5537),
    [anon_sym_IBOutlet] = ACTIONS(5537),
    [anon_sym_IBInspectable] = ACTIONS(5537),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5537),
    [anon_sym_NS_INLINE] = ACTIONS(5537),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5537),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5537),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5537),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5537),
    [anon_sym_const] = ACTIONS(5537),
    [anon_sym_constexpr] = ACTIONS(5537),
    [anon_sym_volatile] = ACTIONS(5537),
    [anon_sym_restrict] = ACTIONS(5537),
    [anon_sym___restrict__] = ACTIONS(5537),
    [anon_sym__Atomic] = ACTIONS(5537),
    [anon_sym__Noreturn] = ACTIONS(5537),
    [anon_sym__Nonnull] = ACTIONS(5537),
    [anon_sym_nullable] = ACTIONS(5537),
    [anon_sym__Complex] = ACTIONS(5537),
    [anon_sym__Nullable] = ACTIONS(5537),
    [anon_sym__Nullable_result] = ACTIONS(5537),
    [anon_sym__Null_unspecified] = ACTIONS(5537),
    [anon_sym___autoreleasing] = ACTIONS(5537),
    [anon_sym___block] = ACTIONS(5537),
    [anon_sym___bridge] = ACTIONS(5537),
    [anon_sym___bridge_retained] = ACTIONS(5537),
    [anon_sym___bridge_transfer] = ACTIONS(5537),
    [anon_sym___complex] = ACTIONS(5537),
    [anon_sym___const] = ACTIONS(5537),
    [anon_sym___imag] = ACTIONS(5537),
    [anon_sym___kindof] = ACTIONS(5537),
    [anon_sym___nonnull] = ACTIONS(5537),
    [anon_sym___nullable] = ACTIONS(5537),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5537),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5537),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5537),
    [anon_sym___real] = ACTIONS(5537),
    [anon_sym___strong] = ACTIONS(5537),
    [anon_sym___unsafe_unretained] = ACTIONS(5537),
    [anon_sym___unused] = ACTIONS(5537),
    [anon_sym___weak] = ACTIONS(5537),
    [anon_sym_alignas] = ACTIONS(5537),
    [anon_sym__Alignas] = ACTIONS(5537),
    [sym_primitive_type] = ACTIONS(5537),
    [anon_sym_enum] = ACTIONS(5537),
    [anon_sym_struct] = ACTIONS(5537),
    [anon_sym_union] = ACTIONS(5537),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5537),
    [anon_sym___typeof] = ACTIONS(5537),
    [anon_sym_typeof] = ACTIONS(5537),
    [aux_sym_preproc_undef_token1] = ACTIONS(5537),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5537),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5537),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5537),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5537),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5537),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5537),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5537),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5537),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5537),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_API_AVAILABLE] = ACTIONS(5537),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_API_DEPRECATED] = ACTIONS(5537),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5537),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5537),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5537),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5537),
    [anon_sym___deprecated_msg] = ACTIONS(5537),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5537),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5537),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5537),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5537),
    [anon_sym_ATend] = ACTIONS(5539),
    [anon_sym_AToptional] = ACTIONS(5539),
    [anon_sym_ATrequired] = ACTIONS(5539),
    [anon_sym_ATproperty] = ACTIONS(5539),
    [anon_sym_BOOL] = ACTIONS(5537),
    [anon_sym_IMP] = ACTIONS(5537),
    [anon_sym_SEL] = ACTIONS(5537),
    [anon_sym_Class] = ACTIONS(5537),
    [anon_sym_id] = ACTIONS(5537),
  },
  [STATE(2328)] = {
    [sym_identifier] = ACTIONS(5399),
    [aux_sym_preproc_def_token1] = ACTIONS(5399),
    [aux_sym_preproc_if_token1] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5399),
    [sym_preproc_directive] = ACTIONS(5399),
    [anon_sym_DASH] = ACTIONS(5401),
    [anon_sym_PLUS] = ACTIONS(5401),
    [anon_sym_SEMI] = ACTIONS(5626),
    [anon_sym___extension__] = ACTIONS(5399),
    [anon_sym_typedef] = ACTIONS(5399),
    [anon_sym_extern] = ACTIONS(5399),
    [anon_sym___attribute__] = ACTIONS(5399),
    [anon_sym___attribute] = ACTIONS(5399),
    [anon_sym_noreturn] = ACTIONS(5399),
    [anon_sym_LBRACK] = ACTIONS(5401),
    [anon_sym___declspec] = ACTIONS(5399),
    [anon_sym___cdecl] = ACTIONS(5399),
    [anon_sym___clrcall] = ACTIONS(5399),
    [anon_sym___stdcall] = ACTIONS(5399),
    [anon_sym___fastcall] = ACTIONS(5399),
    [anon_sym___thiscall] = ACTIONS(5399),
    [anon_sym___vectorcall] = ACTIONS(5399),
    [anon_sym_signed] = ACTIONS(5399),
    [anon_sym_unsigned] = ACTIONS(5399),
    [anon_sym_long] = ACTIONS(5399),
    [anon_sym_short] = ACTIONS(5399),
    [anon_sym_static] = ACTIONS(5399),
    [anon_sym_auto] = ACTIONS(5399),
    [anon_sym_register] = ACTIONS(5399),
    [anon_sym_inline] = ACTIONS(5399),
    [anon_sym___inline] = ACTIONS(5399),
    [anon_sym___inline__] = ACTIONS(5399),
    [anon_sym___forceinline] = ACTIONS(5399),
    [anon_sym_thread_local] = ACTIONS(5399),
    [anon_sym___thread] = ACTIONS(5399),
    [anon_sym_CG_EXTERN] = ACTIONS(5399),
    [anon_sym_CG_INLINE] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5399),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5399),
    [anon_sym_IBOutlet] = ACTIONS(5399),
    [anon_sym_IBInspectable] = ACTIONS(5399),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5399),
    [anon_sym_NS_INLINE] = ACTIONS(5399),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5399),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5399),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5399),
    [anon_sym_const] = ACTIONS(5399),
    [anon_sym_constexpr] = ACTIONS(5399),
    [anon_sym_volatile] = ACTIONS(5399),
    [anon_sym_restrict] = ACTIONS(5399),
    [anon_sym___restrict__] = ACTIONS(5399),
    [anon_sym__Atomic] = ACTIONS(5399),
    [anon_sym__Noreturn] = ACTIONS(5399),
    [anon_sym__Nonnull] = ACTIONS(5399),
    [anon_sym_nullable] = ACTIONS(5399),
    [anon_sym__Complex] = ACTIONS(5399),
    [anon_sym__Nullable] = ACTIONS(5399),
    [anon_sym__Nullable_result] = ACTIONS(5399),
    [anon_sym__Null_unspecified] = ACTIONS(5399),
    [anon_sym___autoreleasing] = ACTIONS(5399),
    [anon_sym___block] = ACTIONS(5399),
    [anon_sym___bridge] = ACTIONS(5399),
    [anon_sym___bridge_retained] = ACTIONS(5399),
    [anon_sym___bridge_transfer] = ACTIONS(5399),
    [anon_sym___complex] = ACTIONS(5399),
    [anon_sym___const] = ACTIONS(5399),
    [anon_sym___imag] = ACTIONS(5399),
    [anon_sym___kindof] = ACTIONS(5399),
    [anon_sym___nonnull] = ACTIONS(5399),
    [anon_sym___nullable] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5399),
    [anon_sym___real] = ACTIONS(5399),
    [anon_sym___strong] = ACTIONS(5399),
    [anon_sym___unsafe_unretained] = ACTIONS(5399),
    [anon_sym___unused] = ACTIONS(5399),
    [anon_sym___weak] = ACTIONS(5399),
    [anon_sym_alignas] = ACTIONS(5399),
    [anon_sym__Alignas] = ACTIONS(5399),
    [sym_primitive_type] = ACTIONS(5399),
    [anon_sym_enum] = ACTIONS(5399),
    [anon_sym_struct] = ACTIONS(5399),
    [anon_sym_union] = ACTIONS(5399),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5399),
    [anon_sym___typeof] = ACTIONS(5399),
    [anon_sym_typeof] = ACTIONS(5399),
    [aux_sym_preproc_undef_token1] = ACTIONS(5399),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5399),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5399),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5399),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_API_AVAILABLE] = ACTIONS(5399),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_API_DEPRECATED] = ACTIONS(5399),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym___deprecated_msg] = ACTIONS(5399),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5399),
    [anon_sym_ATend] = ACTIONS(5401),
    [anon_sym_ATsynthesize] = ACTIONS(5401),
    [anon_sym_ATdynamic] = ACTIONS(5401),
    [anon_sym_BOOL] = ACTIONS(5399),
    [anon_sym_IMP] = ACTIONS(5399),
    [anon_sym_SEL] = ACTIONS(5399),
    [anon_sym_Class] = ACTIONS(5399),
    [anon_sym_id] = ACTIONS(5399),
  },
  [STATE(2329)] = {
    [sym_identifier] = ACTIONS(2546),
    [aux_sym_preproc_def_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token1] = ACTIONS(2546),
    [aux_sym_preproc_if_token2] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2546),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2546),
    [sym_preproc_directive] = ACTIONS(2546),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_PLUS] = ACTIONS(2548),
    [anon_sym___extension__] = ACTIONS(2546),
    [anon_sym_typedef] = ACTIONS(2546),
    [anon_sym_extern] = ACTIONS(2546),
    [anon_sym___attribute__] = ACTIONS(2546),
    [anon_sym___attribute] = ACTIONS(2546),
    [anon_sym_noreturn] = ACTIONS(2546),
    [anon_sym_LBRACK] = ACTIONS(2548),
    [anon_sym___declspec] = ACTIONS(2546),
    [anon_sym___cdecl] = ACTIONS(2546),
    [anon_sym___clrcall] = ACTIONS(2546),
    [anon_sym___stdcall] = ACTIONS(2546),
    [anon_sym___fastcall] = ACTIONS(2546),
    [anon_sym___thiscall] = ACTIONS(2546),
    [anon_sym___vectorcall] = ACTIONS(2546),
    [anon_sym_signed] = ACTIONS(2546),
    [anon_sym_unsigned] = ACTIONS(2546),
    [anon_sym_long] = ACTIONS(2546),
    [anon_sym_short] = ACTIONS(2546),
    [anon_sym_static] = ACTIONS(2546),
    [anon_sym_auto] = ACTIONS(2546),
    [anon_sym_register] = ACTIONS(2546),
    [anon_sym_inline] = ACTIONS(2546),
    [anon_sym___inline] = ACTIONS(2546),
    [anon_sym___inline__] = ACTIONS(2546),
    [anon_sym___forceinline] = ACTIONS(2546),
    [anon_sym_thread_local] = ACTIONS(2546),
    [anon_sym___thread] = ACTIONS(2546),
    [anon_sym_CG_EXTERN] = ACTIONS(2546),
    [anon_sym_CG_INLINE] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2546),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2546),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2546),
    [anon_sym_IBOutlet] = ACTIONS(2546),
    [anon_sym_IBInspectable] = ACTIONS(2546),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2546),
    [anon_sym_NS_INLINE] = ACTIONS(2546),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2546),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2546),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2546),
    [anon_sym_const] = ACTIONS(2546),
    [anon_sym_constexpr] = ACTIONS(2546),
    [anon_sym_volatile] = ACTIONS(2546),
    [anon_sym_restrict] = ACTIONS(2546),
    [anon_sym___restrict__] = ACTIONS(2546),
    [anon_sym__Atomic] = ACTIONS(2546),
    [anon_sym__Noreturn] = ACTIONS(2546),
    [anon_sym__Nonnull] = ACTIONS(2546),
    [anon_sym_nullable] = ACTIONS(2546),
    [anon_sym__Complex] = ACTIONS(2546),
    [anon_sym__Nullable] = ACTIONS(2546),
    [anon_sym__Nullable_result] = ACTIONS(2546),
    [anon_sym__Null_unspecified] = ACTIONS(2546),
    [anon_sym___autoreleasing] = ACTIONS(2546),
    [anon_sym___block] = ACTIONS(2546),
    [anon_sym___bridge] = ACTIONS(2546),
    [anon_sym___bridge_retained] = ACTIONS(2546),
    [anon_sym___bridge_transfer] = ACTIONS(2546),
    [anon_sym___complex] = ACTIONS(2546),
    [anon_sym___const] = ACTIONS(2546),
    [anon_sym___imag] = ACTIONS(2546),
    [anon_sym___kindof] = ACTIONS(2546),
    [anon_sym___nonnull] = ACTIONS(2546),
    [anon_sym___nullable] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2546),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2546),
    [anon_sym___real] = ACTIONS(2546),
    [anon_sym___strong] = ACTIONS(2546),
    [anon_sym___unsafe_unretained] = ACTIONS(2546),
    [anon_sym___unused] = ACTIONS(2546),
    [anon_sym___weak] = ACTIONS(2546),
    [anon_sym_alignas] = ACTIONS(2546),
    [anon_sym__Alignas] = ACTIONS(2546),
    [sym_primitive_type] = ACTIONS(2546),
    [anon_sym_enum] = ACTIONS(2546),
    [anon_sym_struct] = ACTIONS(2546),
    [anon_sym_union] = ACTIONS(2546),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2546),
    [anon_sym___typeof] = ACTIONS(2546),
    [anon_sym_typeof] = ACTIONS(2546),
    [aux_sym_preproc_undef_token1] = ACTIONS(2546),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2546),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2546),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2546),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2546),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2546),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_API_AVAILABLE] = ACTIONS(2546),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_API_DEPRECATED] = ACTIONS(2546),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2546),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2546),
    [anon_sym___deprecated_msg] = ACTIONS(2546),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2546),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2546),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2546),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2546),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2546),
    [anon_sym_ATsynthesize] = ACTIONS(2548),
    [anon_sym_ATdynamic] = ACTIONS(2548),
    [anon_sym_ATproperty] = ACTIONS(2548),
    [anon_sym_BOOL] = ACTIONS(2546),
    [anon_sym_IMP] = ACTIONS(2546),
    [anon_sym_SEL] = ACTIONS(2546),
    [anon_sym_Class] = ACTIONS(2546),
    [anon_sym_id] = ACTIONS(2546),
  },
  [STATE(2330)] = {
    [sym_identifier] = ACTIONS(5628),
    [aux_sym_preproc_def_token1] = ACTIONS(5628),
    [aux_sym_preproc_if_token1] = ACTIONS(5628),
    [aux_sym_preproc_if_token2] = ACTIONS(5628),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5628),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5628),
    [aux_sym_preproc_else_token1] = ACTIONS(5628),
    [aux_sym_preproc_elif_token1] = ACTIONS(5628),
    [sym_preproc_directive] = ACTIONS(5628),
    [anon_sym_DASH] = ACTIONS(5630),
    [anon_sym_PLUS] = ACTIONS(5630),
    [anon_sym___extension__] = ACTIONS(5628),
    [anon_sym_typedef] = ACTIONS(5628),
    [anon_sym_extern] = ACTIONS(5628),
    [anon_sym___attribute__] = ACTIONS(5628),
    [anon_sym___attribute] = ACTIONS(5628),
    [anon_sym_noreturn] = ACTIONS(5628),
    [anon_sym_LBRACK] = ACTIONS(5630),
    [anon_sym___declspec] = ACTIONS(5628),
    [anon_sym___cdecl] = ACTIONS(5628),
    [anon_sym___clrcall] = ACTIONS(5628),
    [anon_sym___stdcall] = ACTIONS(5628),
    [anon_sym___fastcall] = ACTIONS(5628),
    [anon_sym___thiscall] = ACTIONS(5628),
    [anon_sym___vectorcall] = ACTIONS(5628),
    [anon_sym_signed] = ACTIONS(5628),
    [anon_sym_unsigned] = ACTIONS(5628),
    [anon_sym_long] = ACTIONS(5628),
    [anon_sym_short] = ACTIONS(5628),
    [anon_sym_static] = ACTIONS(5628),
    [anon_sym_auto] = ACTIONS(5628),
    [anon_sym_register] = ACTIONS(5628),
    [anon_sym_inline] = ACTIONS(5628),
    [anon_sym___inline] = ACTIONS(5628),
    [anon_sym___inline__] = ACTIONS(5628),
    [anon_sym___forceinline] = ACTIONS(5628),
    [anon_sym_thread_local] = ACTIONS(5628),
    [anon_sym___thread] = ACTIONS(5628),
    [anon_sym_CG_EXTERN] = ACTIONS(5628),
    [anon_sym_CG_INLINE] = ACTIONS(5628),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5628),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5628),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5628),
    [anon_sym_IBOutlet] = ACTIONS(5628),
    [anon_sym_IBInspectable] = ACTIONS(5628),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5628),
    [anon_sym_NS_INLINE] = ACTIONS(5628),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5628),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5628),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5628),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5628),
    [anon_sym_const] = ACTIONS(5628),
    [anon_sym_constexpr] = ACTIONS(5628),
    [anon_sym_volatile] = ACTIONS(5628),
    [anon_sym_restrict] = ACTIONS(5628),
    [anon_sym___restrict__] = ACTIONS(5628),
    [anon_sym__Atomic] = ACTIONS(5628),
    [anon_sym__Noreturn] = ACTIONS(5628),
    [anon_sym__Nonnull] = ACTIONS(5628),
    [anon_sym_nullable] = ACTIONS(5628),
    [anon_sym__Complex] = ACTIONS(5628),
    [anon_sym__Nullable] = ACTIONS(5628),
    [anon_sym__Nullable_result] = ACTIONS(5628),
    [anon_sym__Null_unspecified] = ACTIONS(5628),
    [anon_sym___autoreleasing] = ACTIONS(5628),
    [anon_sym___block] = ACTIONS(5628),
    [anon_sym___bridge] = ACTIONS(5628),
    [anon_sym___bridge_retained] = ACTIONS(5628),
    [anon_sym___bridge_transfer] = ACTIONS(5628),
    [anon_sym___complex] = ACTIONS(5628),
    [anon_sym___const] = ACTIONS(5628),
    [anon_sym___imag] = ACTIONS(5628),
    [anon_sym___kindof] = ACTIONS(5628),
    [anon_sym___nonnull] = ACTIONS(5628),
    [anon_sym___nullable] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5628),
    [anon_sym___real] = ACTIONS(5628),
    [anon_sym___strong] = ACTIONS(5628),
    [anon_sym___unsafe_unretained] = ACTIONS(5628),
    [anon_sym___unused] = ACTIONS(5628),
    [anon_sym___weak] = ACTIONS(5628),
    [anon_sym_alignas] = ACTIONS(5628),
    [anon_sym__Alignas] = ACTIONS(5628),
    [sym_primitive_type] = ACTIONS(5628),
    [anon_sym_enum] = ACTIONS(5628),
    [anon_sym_struct] = ACTIONS(5628),
    [anon_sym_union] = ACTIONS(5628),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5628),
    [anon_sym___typeof] = ACTIONS(5628),
    [anon_sym_typeof] = ACTIONS(5628),
    [aux_sym_preproc_undef_token1] = ACTIONS(5628),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5628),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5628),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5628),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5628),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5628),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5628),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5628),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5628),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5628),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_API_AVAILABLE] = ACTIONS(5628),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_API_DEPRECATED] = ACTIONS(5628),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5628),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5628),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5628),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5628),
    [anon_sym___deprecated_msg] = ACTIONS(5628),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5628),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5628),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5628),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5628),
    [anon_sym_ATproperty] = ACTIONS(5630),
    [anon_sym_BOOL] = ACTIONS(5628),
    [anon_sym_IMP] = ACTIONS(5628),
    [anon_sym_SEL] = ACTIONS(5628),
    [anon_sym_Class] = ACTIONS(5628),
    [anon_sym_id] = ACTIONS(5628),
  },
  [STATE(2331)] = {
    [sym_identifier] = ACTIONS(2510),
    [aux_sym_preproc_def_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token1] = ACTIONS(2510),
    [aux_sym_preproc_if_token2] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2510),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2510),
    [sym_preproc_directive] = ACTIONS(2510),
    [anon_sym_DASH] = ACTIONS(2512),
    [anon_sym_PLUS] = ACTIONS(2512),
    [anon_sym___extension__] = ACTIONS(2510),
    [anon_sym_typedef] = ACTIONS(2510),
    [anon_sym_extern] = ACTIONS(2510),
    [anon_sym___attribute__] = ACTIONS(2510),
    [anon_sym___attribute] = ACTIONS(2510),
    [anon_sym_noreturn] = ACTIONS(2510),
    [anon_sym_LBRACK] = ACTIONS(2512),
    [anon_sym___declspec] = ACTIONS(2510),
    [anon_sym___cdecl] = ACTIONS(2510),
    [anon_sym___clrcall] = ACTIONS(2510),
    [anon_sym___stdcall] = ACTIONS(2510),
    [anon_sym___fastcall] = ACTIONS(2510),
    [anon_sym___thiscall] = ACTIONS(2510),
    [anon_sym___vectorcall] = ACTIONS(2510),
    [anon_sym_signed] = ACTIONS(2510),
    [anon_sym_unsigned] = ACTIONS(2510),
    [anon_sym_long] = ACTIONS(2510),
    [anon_sym_short] = ACTIONS(2510),
    [anon_sym_static] = ACTIONS(2510),
    [anon_sym_auto] = ACTIONS(2510),
    [anon_sym_register] = ACTIONS(2510),
    [anon_sym_inline] = ACTIONS(2510),
    [anon_sym___inline] = ACTIONS(2510),
    [anon_sym___inline__] = ACTIONS(2510),
    [anon_sym___forceinline] = ACTIONS(2510),
    [anon_sym_thread_local] = ACTIONS(2510),
    [anon_sym___thread] = ACTIONS(2510),
    [anon_sym_CG_EXTERN] = ACTIONS(2510),
    [anon_sym_CG_INLINE] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2510),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2510),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2510),
    [anon_sym_IBOutlet] = ACTIONS(2510),
    [anon_sym_IBInspectable] = ACTIONS(2510),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2510),
    [anon_sym_NS_INLINE] = ACTIONS(2510),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2510),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2510),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2510),
    [anon_sym_const] = ACTIONS(2510),
    [anon_sym_constexpr] = ACTIONS(2510),
    [anon_sym_volatile] = ACTIONS(2510),
    [anon_sym_restrict] = ACTIONS(2510),
    [anon_sym___restrict__] = ACTIONS(2510),
    [anon_sym__Atomic] = ACTIONS(2510),
    [anon_sym__Noreturn] = ACTIONS(2510),
    [anon_sym__Nonnull] = ACTIONS(2510),
    [anon_sym_nullable] = ACTIONS(2510),
    [anon_sym__Complex] = ACTIONS(2510),
    [anon_sym__Nullable] = ACTIONS(2510),
    [anon_sym__Nullable_result] = ACTIONS(2510),
    [anon_sym__Null_unspecified] = ACTIONS(2510),
    [anon_sym___autoreleasing] = ACTIONS(2510),
    [anon_sym___block] = ACTIONS(2510),
    [anon_sym___bridge] = ACTIONS(2510),
    [anon_sym___bridge_retained] = ACTIONS(2510),
    [anon_sym___bridge_transfer] = ACTIONS(2510),
    [anon_sym___complex] = ACTIONS(2510),
    [anon_sym___const] = ACTIONS(2510),
    [anon_sym___imag] = ACTIONS(2510),
    [anon_sym___kindof] = ACTIONS(2510),
    [anon_sym___nonnull] = ACTIONS(2510),
    [anon_sym___nullable] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2510),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2510),
    [anon_sym___real] = ACTIONS(2510),
    [anon_sym___strong] = ACTIONS(2510),
    [anon_sym___unsafe_unretained] = ACTIONS(2510),
    [anon_sym___unused] = ACTIONS(2510),
    [anon_sym___weak] = ACTIONS(2510),
    [anon_sym_alignas] = ACTIONS(2510),
    [anon_sym__Alignas] = ACTIONS(2510),
    [sym_primitive_type] = ACTIONS(2510),
    [anon_sym_enum] = ACTIONS(2510),
    [anon_sym_struct] = ACTIONS(2510),
    [anon_sym_union] = ACTIONS(2510),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2510),
    [anon_sym___typeof] = ACTIONS(2510),
    [anon_sym_typeof] = ACTIONS(2510),
    [aux_sym_preproc_undef_token1] = ACTIONS(2510),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2510),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2510),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2510),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2510),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2510),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_API_AVAILABLE] = ACTIONS(2510),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_API_DEPRECATED] = ACTIONS(2510),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2510),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2510),
    [anon_sym___deprecated_msg] = ACTIONS(2510),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2510),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2510),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2510),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2510),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2510),
    [anon_sym_ATsynthesize] = ACTIONS(2512),
    [anon_sym_ATdynamic] = ACTIONS(2512),
    [anon_sym_ATproperty] = ACTIONS(2512),
    [anon_sym_BOOL] = ACTIONS(2510),
    [anon_sym_IMP] = ACTIONS(2510),
    [anon_sym_SEL] = ACTIONS(2510),
    [anon_sym_Class] = ACTIONS(2510),
    [anon_sym_id] = ACTIONS(2510),
  },
  [STATE(2332)] = {
    [sym_identifier] = ACTIONS(5405),
    [aux_sym_preproc_def_token1] = ACTIONS(5405),
    [aux_sym_preproc_if_token1] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5405),
    [sym_preproc_directive] = ACTIONS(5405),
    [anon_sym_DASH] = ACTIONS(5407),
    [anon_sym_PLUS] = ACTIONS(5407),
    [anon_sym_SEMI] = ACTIONS(5632),
    [anon_sym___extension__] = ACTIONS(5405),
    [anon_sym_typedef] = ACTIONS(5405),
    [anon_sym_extern] = ACTIONS(5405),
    [anon_sym___attribute__] = ACTIONS(5405),
    [anon_sym___attribute] = ACTIONS(5405),
    [anon_sym_noreturn] = ACTIONS(5405),
    [anon_sym_LBRACK] = ACTIONS(5407),
    [anon_sym___declspec] = ACTIONS(5405),
    [anon_sym___cdecl] = ACTIONS(5405),
    [anon_sym___clrcall] = ACTIONS(5405),
    [anon_sym___stdcall] = ACTIONS(5405),
    [anon_sym___fastcall] = ACTIONS(5405),
    [anon_sym___thiscall] = ACTIONS(5405),
    [anon_sym___vectorcall] = ACTIONS(5405),
    [anon_sym_signed] = ACTIONS(5405),
    [anon_sym_unsigned] = ACTIONS(5405),
    [anon_sym_long] = ACTIONS(5405),
    [anon_sym_short] = ACTIONS(5405),
    [anon_sym_static] = ACTIONS(5405),
    [anon_sym_auto] = ACTIONS(5405),
    [anon_sym_register] = ACTIONS(5405),
    [anon_sym_inline] = ACTIONS(5405),
    [anon_sym___inline] = ACTIONS(5405),
    [anon_sym___inline__] = ACTIONS(5405),
    [anon_sym___forceinline] = ACTIONS(5405),
    [anon_sym_thread_local] = ACTIONS(5405),
    [anon_sym___thread] = ACTIONS(5405),
    [anon_sym_CG_EXTERN] = ACTIONS(5405),
    [anon_sym_CG_INLINE] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5405),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5405),
    [anon_sym_IBOutlet] = ACTIONS(5405),
    [anon_sym_IBInspectable] = ACTIONS(5405),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5405),
    [anon_sym_NS_INLINE] = ACTIONS(5405),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5405),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5405),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5405),
    [anon_sym_const] = ACTIONS(5405),
    [anon_sym_constexpr] = ACTIONS(5405),
    [anon_sym_volatile] = ACTIONS(5405),
    [anon_sym_restrict] = ACTIONS(5405),
    [anon_sym___restrict__] = ACTIONS(5405),
    [anon_sym__Atomic] = ACTIONS(5405),
    [anon_sym__Noreturn] = ACTIONS(5405),
    [anon_sym__Nonnull] = ACTIONS(5405),
    [anon_sym_nullable] = ACTIONS(5405),
    [anon_sym__Complex] = ACTIONS(5405),
    [anon_sym__Nullable] = ACTIONS(5405),
    [anon_sym__Nullable_result] = ACTIONS(5405),
    [anon_sym__Null_unspecified] = ACTIONS(5405),
    [anon_sym___autoreleasing] = ACTIONS(5405),
    [anon_sym___block] = ACTIONS(5405),
    [anon_sym___bridge] = ACTIONS(5405),
    [anon_sym___bridge_retained] = ACTIONS(5405),
    [anon_sym___bridge_transfer] = ACTIONS(5405),
    [anon_sym___complex] = ACTIONS(5405),
    [anon_sym___const] = ACTIONS(5405),
    [anon_sym___imag] = ACTIONS(5405),
    [anon_sym___kindof] = ACTIONS(5405),
    [anon_sym___nonnull] = ACTIONS(5405),
    [anon_sym___nullable] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5405),
    [anon_sym___real] = ACTIONS(5405),
    [anon_sym___strong] = ACTIONS(5405),
    [anon_sym___unsafe_unretained] = ACTIONS(5405),
    [anon_sym___unused] = ACTIONS(5405),
    [anon_sym___weak] = ACTIONS(5405),
    [anon_sym_alignas] = ACTIONS(5405),
    [anon_sym__Alignas] = ACTIONS(5405),
    [sym_primitive_type] = ACTIONS(5405),
    [anon_sym_enum] = ACTIONS(5405),
    [anon_sym_struct] = ACTIONS(5405),
    [anon_sym_union] = ACTIONS(5405),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5405),
    [anon_sym___typeof] = ACTIONS(5405),
    [anon_sym_typeof] = ACTIONS(5405),
    [aux_sym_preproc_undef_token1] = ACTIONS(5405),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5405),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5405),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5405),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_API_AVAILABLE] = ACTIONS(5405),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_API_DEPRECATED] = ACTIONS(5405),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym___deprecated_msg] = ACTIONS(5405),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5405),
    [anon_sym_ATend] = ACTIONS(5407),
    [anon_sym_ATsynthesize] = ACTIONS(5407),
    [anon_sym_ATdynamic] = ACTIONS(5407),
    [anon_sym_BOOL] = ACTIONS(5405),
    [anon_sym_IMP] = ACTIONS(5405),
    [anon_sym_SEL] = ACTIONS(5405),
    [anon_sym_Class] = ACTIONS(5405),
    [anon_sym_id] = ACTIONS(5405),
  },
  [STATE(2333)] = {
    [sym_identifier] = ACTIONS(5634),
    [aux_sym_preproc_def_token1] = ACTIONS(5634),
    [aux_sym_preproc_if_token1] = ACTIONS(5634),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5634),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5634),
    [sym_preproc_directive] = ACTIONS(5634),
    [anon_sym_DASH] = ACTIONS(5636),
    [anon_sym_PLUS] = ACTIONS(5636),
    [anon_sym___extension__] = ACTIONS(5634),
    [anon_sym_typedef] = ACTIONS(5634),
    [anon_sym_extern] = ACTIONS(5634),
    [anon_sym___attribute__] = ACTIONS(5634),
    [anon_sym___attribute] = ACTIONS(5634),
    [anon_sym_noreturn] = ACTIONS(5634),
    [anon_sym_LBRACK] = ACTIONS(5636),
    [anon_sym___declspec] = ACTIONS(5634),
    [anon_sym___cdecl] = ACTIONS(5634),
    [anon_sym___clrcall] = ACTIONS(5634),
    [anon_sym___stdcall] = ACTIONS(5634),
    [anon_sym___fastcall] = ACTIONS(5634),
    [anon_sym___thiscall] = ACTIONS(5634),
    [anon_sym___vectorcall] = ACTIONS(5634),
    [anon_sym_signed] = ACTIONS(5634),
    [anon_sym_unsigned] = ACTIONS(5634),
    [anon_sym_long] = ACTIONS(5634),
    [anon_sym_short] = ACTIONS(5634),
    [anon_sym_static] = ACTIONS(5634),
    [anon_sym_auto] = ACTIONS(5634),
    [anon_sym_register] = ACTIONS(5634),
    [anon_sym_inline] = ACTIONS(5634),
    [anon_sym___inline] = ACTIONS(5634),
    [anon_sym___inline__] = ACTIONS(5634),
    [anon_sym___forceinline] = ACTIONS(5634),
    [anon_sym_thread_local] = ACTIONS(5634),
    [anon_sym___thread] = ACTIONS(5634),
    [anon_sym_CG_EXTERN] = ACTIONS(5634),
    [anon_sym_CG_INLINE] = ACTIONS(5634),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5634),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5634),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5634),
    [anon_sym_IBOutlet] = ACTIONS(5634),
    [anon_sym_IBInspectable] = ACTIONS(5634),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5634),
    [anon_sym_NS_INLINE] = ACTIONS(5634),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5634),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5634),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5634),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5634),
    [anon_sym_const] = ACTIONS(5634),
    [anon_sym_constexpr] = ACTIONS(5634),
    [anon_sym_volatile] = ACTIONS(5634),
    [anon_sym_restrict] = ACTIONS(5634),
    [anon_sym___restrict__] = ACTIONS(5634),
    [anon_sym__Atomic] = ACTIONS(5634),
    [anon_sym__Noreturn] = ACTIONS(5634),
    [anon_sym__Nonnull] = ACTIONS(5634),
    [anon_sym_nullable] = ACTIONS(5634),
    [anon_sym__Complex] = ACTIONS(5634),
    [anon_sym__Nullable] = ACTIONS(5634),
    [anon_sym__Nullable_result] = ACTIONS(5634),
    [anon_sym__Null_unspecified] = ACTIONS(5634),
    [anon_sym___autoreleasing] = ACTIONS(5634),
    [anon_sym___block] = ACTIONS(5634),
    [anon_sym___bridge] = ACTIONS(5634),
    [anon_sym___bridge_retained] = ACTIONS(5634),
    [anon_sym___bridge_transfer] = ACTIONS(5634),
    [anon_sym___complex] = ACTIONS(5634),
    [anon_sym___const] = ACTIONS(5634),
    [anon_sym___imag] = ACTIONS(5634),
    [anon_sym___kindof] = ACTIONS(5634),
    [anon_sym___nonnull] = ACTIONS(5634),
    [anon_sym___nullable] = ACTIONS(5634),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5634),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5634),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5634),
    [anon_sym___real] = ACTIONS(5634),
    [anon_sym___strong] = ACTIONS(5634),
    [anon_sym___unsafe_unretained] = ACTIONS(5634),
    [anon_sym___unused] = ACTIONS(5634),
    [anon_sym___weak] = ACTIONS(5634),
    [anon_sym_alignas] = ACTIONS(5634),
    [anon_sym__Alignas] = ACTIONS(5634),
    [sym_primitive_type] = ACTIONS(5634),
    [anon_sym_enum] = ACTIONS(5634),
    [anon_sym_struct] = ACTIONS(5634),
    [anon_sym_union] = ACTIONS(5634),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5634),
    [anon_sym___typeof] = ACTIONS(5634),
    [anon_sym_typeof] = ACTIONS(5634),
    [aux_sym_preproc_undef_token1] = ACTIONS(5634),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5634),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5634),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5634),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5634),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5634),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5634),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5634),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5634),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5634),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5634),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5634),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5634),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5634),
    [anon_sym_API_AVAILABLE] = ACTIONS(5634),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5634),
    [anon_sym_API_DEPRECATED] = ACTIONS(5634),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5634),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5634),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5634),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5634),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5634),
    [anon_sym___deprecated_msg] = ACTIONS(5634),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5634),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5634),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5634),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5634),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5634),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5634),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5634),
    [anon_sym_ATend] = ACTIONS(5636),
    [anon_sym_ATsynthesize] = ACTIONS(5636),
    [anon_sym_ATdynamic] = ACTIONS(5636),
    [anon_sym_ATproperty] = ACTIONS(5636),
    [anon_sym_BOOL] = ACTIONS(5634),
    [anon_sym_IMP] = ACTIONS(5634),
    [anon_sym_SEL] = ACTIONS(5634),
    [anon_sym_Class] = ACTIONS(5634),
    [anon_sym_id] = ACTIONS(5634),
  },
  [STATE(2334)] = {
    [sym_identifier] = ACTIONS(5448),
    [aux_sym_preproc_def_token1] = ACTIONS(5448),
    [aux_sym_preproc_if_token1] = ACTIONS(5448),
    [aux_sym_preproc_if_token2] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5448),
    [sym_preproc_directive] = ACTIONS(5448),
    [anon_sym_DASH] = ACTIONS(5450),
    [anon_sym_PLUS] = ACTIONS(5450),
    [anon_sym_SEMI] = ACTIONS(5638),
    [anon_sym___extension__] = ACTIONS(5448),
    [anon_sym_typedef] = ACTIONS(5448),
    [anon_sym_extern] = ACTIONS(5448),
    [anon_sym___attribute__] = ACTIONS(5448),
    [anon_sym___attribute] = ACTIONS(5448),
    [anon_sym_noreturn] = ACTIONS(5448),
    [anon_sym_LBRACK] = ACTIONS(5450),
    [anon_sym___declspec] = ACTIONS(5448),
    [anon_sym___cdecl] = ACTIONS(5448),
    [anon_sym___clrcall] = ACTIONS(5448),
    [anon_sym___stdcall] = ACTIONS(5448),
    [anon_sym___fastcall] = ACTIONS(5448),
    [anon_sym___thiscall] = ACTIONS(5448),
    [anon_sym___vectorcall] = ACTIONS(5448),
    [anon_sym_signed] = ACTIONS(5448),
    [anon_sym_unsigned] = ACTIONS(5448),
    [anon_sym_long] = ACTIONS(5448),
    [anon_sym_short] = ACTIONS(5448),
    [anon_sym_static] = ACTIONS(5448),
    [anon_sym_auto] = ACTIONS(5448),
    [anon_sym_register] = ACTIONS(5448),
    [anon_sym_inline] = ACTIONS(5448),
    [anon_sym___inline] = ACTIONS(5448),
    [anon_sym___inline__] = ACTIONS(5448),
    [anon_sym___forceinline] = ACTIONS(5448),
    [anon_sym_thread_local] = ACTIONS(5448),
    [anon_sym___thread] = ACTIONS(5448),
    [anon_sym_CG_EXTERN] = ACTIONS(5448),
    [anon_sym_CG_INLINE] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5448),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5448),
    [anon_sym_IBOutlet] = ACTIONS(5448),
    [anon_sym_IBInspectable] = ACTIONS(5448),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5448),
    [anon_sym_NS_INLINE] = ACTIONS(5448),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5448),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5448),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5448),
    [anon_sym_const] = ACTIONS(5448),
    [anon_sym_constexpr] = ACTIONS(5448),
    [anon_sym_volatile] = ACTIONS(5448),
    [anon_sym_restrict] = ACTIONS(5448),
    [anon_sym___restrict__] = ACTIONS(5448),
    [anon_sym__Atomic] = ACTIONS(5448),
    [anon_sym__Noreturn] = ACTIONS(5448),
    [anon_sym__Nonnull] = ACTIONS(5448),
    [anon_sym_nullable] = ACTIONS(5448),
    [anon_sym__Complex] = ACTIONS(5448),
    [anon_sym__Nullable] = ACTIONS(5448),
    [anon_sym__Nullable_result] = ACTIONS(5448),
    [anon_sym__Null_unspecified] = ACTIONS(5448),
    [anon_sym___autoreleasing] = ACTIONS(5448),
    [anon_sym___block] = ACTIONS(5448),
    [anon_sym___bridge] = ACTIONS(5448),
    [anon_sym___bridge_retained] = ACTIONS(5448),
    [anon_sym___bridge_transfer] = ACTIONS(5448),
    [anon_sym___complex] = ACTIONS(5448),
    [anon_sym___const] = ACTIONS(5448),
    [anon_sym___imag] = ACTIONS(5448),
    [anon_sym___kindof] = ACTIONS(5448),
    [anon_sym___nonnull] = ACTIONS(5448),
    [anon_sym___nullable] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5448),
    [anon_sym___real] = ACTIONS(5448),
    [anon_sym___strong] = ACTIONS(5448),
    [anon_sym___unsafe_unretained] = ACTIONS(5448),
    [anon_sym___unused] = ACTIONS(5448),
    [anon_sym___weak] = ACTIONS(5448),
    [anon_sym_alignas] = ACTIONS(5448),
    [anon_sym__Alignas] = ACTIONS(5448),
    [sym_primitive_type] = ACTIONS(5448),
    [anon_sym_enum] = ACTIONS(5448),
    [anon_sym_struct] = ACTIONS(5448),
    [anon_sym_union] = ACTIONS(5448),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5448),
    [anon_sym___typeof] = ACTIONS(5448),
    [anon_sym_typeof] = ACTIONS(5448),
    [aux_sym_preproc_undef_token1] = ACTIONS(5448),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5448),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5448),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5448),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_API_AVAILABLE] = ACTIONS(5448),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_API_DEPRECATED] = ACTIONS(5448),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym___deprecated_msg] = ACTIONS(5448),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5448),
    [anon_sym_ATsynthesize] = ACTIONS(5450),
    [anon_sym_ATdynamic] = ACTIONS(5450),
    [anon_sym_BOOL] = ACTIONS(5448),
    [anon_sym_IMP] = ACTIONS(5448),
    [anon_sym_SEL] = ACTIONS(5448),
    [anon_sym_Class] = ACTIONS(5448),
    [anon_sym_id] = ACTIONS(5448),
  },
  [STATE(2335)] = {
    [sym_identifier] = ACTIONS(5581),
    [aux_sym_preproc_def_token1] = ACTIONS(5581),
    [aux_sym_preproc_if_token1] = ACTIONS(5581),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5581),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5581),
    [sym_preproc_directive] = ACTIONS(5581),
    [anon_sym_DASH] = ACTIONS(5583),
    [anon_sym_PLUS] = ACTIONS(5583),
    [anon_sym___extension__] = ACTIONS(5581),
    [anon_sym_typedef] = ACTIONS(5581),
    [anon_sym_extern] = ACTIONS(5581),
    [anon_sym___attribute__] = ACTIONS(5581),
    [anon_sym___attribute] = ACTIONS(5581),
    [anon_sym_noreturn] = ACTIONS(5581),
    [anon_sym_LBRACK] = ACTIONS(5583),
    [anon_sym___declspec] = ACTIONS(5581),
    [anon_sym___cdecl] = ACTIONS(5581),
    [anon_sym___clrcall] = ACTIONS(5581),
    [anon_sym___stdcall] = ACTIONS(5581),
    [anon_sym___fastcall] = ACTIONS(5581),
    [anon_sym___thiscall] = ACTIONS(5581),
    [anon_sym___vectorcall] = ACTIONS(5581),
    [anon_sym_signed] = ACTIONS(5581),
    [anon_sym_unsigned] = ACTIONS(5581),
    [anon_sym_long] = ACTIONS(5581),
    [anon_sym_short] = ACTIONS(5581),
    [anon_sym_static] = ACTIONS(5581),
    [anon_sym_auto] = ACTIONS(5581),
    [anon_sym_register] = ACTIONS(5581),
    [anon_sym_inline] = ACTIONS(5581),
    [anon_sym___inline] = ACTIONS(5581),
    [anon_sym___inline__] = ACTIONS(5581),
    [anon_sym___forceinline] = ACTIONS(5581),
    [anon_sym_thread_local] = ACTIONS(5581),
    [anon_sym___thread] = ACTIONS(5581),
    [anon_sym_CG_EXTERN] = ACTIONS(5581),
    [anon_sym_CG_INLINE] = ACTIONS(5581),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5581),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5581),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5581),
    [anon_sym_IBOutlet] = ACTIONS(5581),
    [anon_sym_IBInspectable] = ACTIONS(5581),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5581),
    [anon_sym_NS_INLINE] = ACTIONS(5581),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5581),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5581),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5581),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5581),
    [anon_sym_const] = ACTIONS(5581),
    [anon_sym_constexpr] = ACTIONS(5581),
    [anon_sym_volatile] = ACTIONS(5581),
    [anon_sym_restrict] = ACTIONS(5581),
    [anon_sym___restrict__] = ACTIONS(5581),
    [anon_sym__Atomic] = ACTIONS(5581),
    [anon_sym__Noreturn] = ACTIONS(5581),
    [anon_sym__Nonnull] = ACTIONS(5581),
    [anon_sym_nullable] = ACTIONS(5581),
    [anon_sym__Complex] = ACTIONS(5581),
    [anon_sym__Nullable] = ACTIONS(5581),
    [anon_sym__Nullable_result] = ACTIONS(5581),
    [anon_sym__Null_unspecified] = ACTIONS(5581),
    [anon_sym___autoreleasing] = ACTIONS(5581),
    [anon_sym___block] = ACTIONS(5581),
    [anon_sym___bridge] = ACTIONS(5581),
    [anon_sym___bridge_retained] = ACTIONS(5581),
    [anon_sym___bridge_transfer] = ACTIONS(5581),
    [anon_sym___complex] = ACTIONS(5581),
    [anon_sym___const] = ACTIONS(5581),
    [anon_sym___imag] = ACTIONS(5581),
    [anon_sym___kindof] = ACTIONS(5581),
    [anon_sym___nonnull] = ACTIONS(5581),
    [anon_sym___nullable] = ACTIONS(5581),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5581),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5581),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5581),
    [anon_sym___real] = ACTIONS(5581),
    [anon_sym___strong] = ACTIONS(5581),
    [anon_sym___unsafe_unretained] = ACTIONS(5581),
    [anon_sym___unused] = ACTIONS(5581),
    [anon_sym___weak] = ACTIONS(5581),
    [anon_sym_alignas] = ACTIONS(5581),
    [anon_sym__Alignas] = ACTIONS(5581),
    [sym_primitive_type] = ACTIONS(5581),
    [anon_sym_enum] = ACTIONS(5581),
    [anon_sym_struct] = ACTIONS(5581),
    [anon_sym_union] = ACTIONS(5581),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5581),
    [anon_sym___typeof] = ACTIONS(5581),
    [anon_sym_typeof] = ACTIONS(5581),
    [aux_sym_preproc_undef_token1] = ACTIONS(5581),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5581),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5581),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5581),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5581),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5581),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5581),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5581),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5581),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5581),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_API_AVAILABLE] = ACTIONS(5581),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_API_DEPRECATED] = ACTIONS(5581),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5581),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5581),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5581),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5581),
    [anon_sym___deprecated_msg] = ACTIONS(5581),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5581),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5581),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5581),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5581),
    [anon_sym_ATend] = ACTIONS(5583),
    [anon_sym_AToptional] = ACTIONS(5583),
    [anon_sym_ATrequired] = ACTIONS(5583),
    [anon_sym_ATproperty] = ACTIONS(5583),
    [anon_sym_BOOL] = ACTIONS(5581),
    [anon_sym_IMP] = ACTIONS(5581),
    [anon_sym_SEL] = ACTIONS(5581),
    [anon_sym_Class] = ACTIONS(5581),
    [anon_sym_id] = ACTIONS(5581),
  },
  [STATE(2336)] = {
    [sym_identifier] = ACTIONS(5481),
    [aux_sym_preproc_def_token1] = ACTIONS(5481),
    [aux_sym_preproc_if_token1] = ACTIONS(5481),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5481),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5481),
    [sym_preproc_directive] = ACTIONS(5481),
    [anon_sym_DASH] = ACTIONS(5483),
    [anon_sym_PLUS] = ACTIONS(5483),
    [anon_sym_LT] = ACTIONS(5483),
    [anon_sym___extension__] = ACTIONS(5481),
    [anon_sym_typedef] = ACTIONS(5481),
    [anon_sym_extern] = ACTIONS(5481),
    [anon_sym___attribute__] = ACTIONS(5481),
    [anon_sym___attribute] = ACTIONS(5481),
    [anon_sym_noreturn] = ACTIONS(5481),
    [anon_sym_LBRACK] = ACTIONS(5483),
    [anon_sym___declspec] = ACTIONS(5481),
    [anon_sym___cdecl] = ACTIONS(5481),
    [anon_sym___clrcall] = ACTIONS(5481),
    [anon_sym___stdcall] = ACTIONS(5481),
    [anon_sym___fastcall] = ACTIONS(5481),
    [anon_sym___thiscall] = ACTIONS(5481),
    [anon_sym___vectorcall] = ACTIONS(5481),
    [anon_sym_LBRACE] = ACTIONS(5483),
    [anon_sym_signed] = ACTIONS(5481),
    [anon_sym_unsigned] = ACTIONS(5481),
    [anon_sym_long] = ACTIONS(5481),
    [anon_sym_short] = ACTIONS(5481),
    [anon_sym_static] = ACTIONS(5481),
    [anon_sym_auto] = ACTIONS(5481),
    [anon_sym_register] = ACTIONS(5481),
    [anon_sym_inline] = ACTIONS(5481),
    [anon_sym___inline] = ACTIONS(5481),
    [anon_sym___inline__] = ACTIONS(5481),
    [anon_sym___forceinline] = ACTIONS(5481),
    [anon_sym_thread_local] = ACTIONS(5481),
    [anon_sym___thread] = ACTIONS(5481),
    [anon_sym_CG_EXTERN] = ACTIONS(5481),
    [anon_sym_CG_INLINE] = ACTIONS(5481),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5481),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5481),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5481),
    [anon_sym_IBOutlet] = ACTIONS(5481),
    [anon_sym_IBInspectable] = ACTIONS(5481),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5481),
    [anon_sym_NS_INLINE] = ACTIONS(5481),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5481),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5481),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5481),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5481),
    [anon_sym_const] = ACTIONS(5481),
    [anon_sym_constexpr] = ACTIONS(5481),
    [anon_sym_volatile] = ACTIONS(5481),
    [anon_sym_restrict] = ACTIONS(5481),
    [anon_sym___restrict__] = ACTIONS(5481),
    [anon_sym__Atomic] = ACTIONS(5481),
    [anon_sym__Noreturn] = ACTIONS(5481),
    [anon_sym__Nonnull] = ACTIONS(5481),
    [anon_sym_nullable] = ACTIONS(5481),
    [anon_sym__Complex] = ACTIONS(5481),
    [anon_sym__Nullable] = ACTIONS(5481),
    [anon_sym__Nullable_result] = ACTIONS(5481),
    [anon_sym__Null_unspecified] = ACTIONS(5481),
    [anon_sym___autoreleasing] = ACTIONS(5481),
    [anon_sym___block] = ACTIONS(5481),
    [anon_sym___bridge] = ACTIONS(5481),
    [anon_sym___bridge_retained] = ACTIONS(5481),
    [anon_sym___bridge_transfer] = ACTIONS(5481),
    [anon_sym___complex] = ACTIONS(5481),
    [anon_sym___const] = ACTIONS(5481),
    [anon_sym___imag] = ACTIONS(5481),
    [anon_sym___kindof] = ACTIONS(5481),
    [anon_sym___nonnull] = ACTIONS(5481),
    [anon_sym___nullable] = ACTIONS(5481),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5481),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5481),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5481),
    [anon_sym___real] = ACTIONS(5481),
    [anon_sym___strong] = ACTIONS(5481),
    [anon_sym___unsafe_unretained] = ACTIONS(5481),
    [anon_sym___unused] = ACTIONS(5481),
    [anon_sym___weak] = ACTIONS(5481),
    [anon_sym_alignas] = ACTIONS(5481),
    [anon_sym__Alignas] = ACTIONS(5481),
    [sym_primitive_type] = ACTIONS(5481),
    [anon_sym_enum] = ACTIONS(5481),
    [anon_sym_struct] = ACTIONS(5481),
    [anon_sym_union] = ACTIONS(5481),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5481),
    [anon_sym___typeof] = ACTIONS(5481),
    [anon_sym_typeof] = ACTIONS(5481),
    [aux_sym_preproc_undef_token1] = ACTIONS(5481),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5481),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5481),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5481),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5481),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5481),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5481),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5481),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5481),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5481),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5481),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5481),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5481),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5481),
    [anon_sym_API_AVAILABLE] = ACTIONS(5481),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5481),
    [anon_sym_API_DEPRECATED] = ACTIONS(5481),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5481),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5481),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5481),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5481),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5481),
    [anon_sym___deprecated_msg] = ACTIONS(5481),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5481),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5481),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5481),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5481),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5481),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5481),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5481),
    [anon_sym_ATend] = ACTIONS(5483),
    [anon_sym_ATproperty] = ACTIONS(5483),
    [anon_sym_BOOL] = ACTIONS(5481),
    [anon_sym_IMP] = ACTIONS(5481),
    [anon_sym_SEL] = ACTIONS(5481),
    [anon_sym_Class] = ACTIONS(5481),
    [anon_sym_id] = ACTIONS(5481),
  },
  [STATE(2337)] = {
    [sym_identifier] = ACTIONS(5609),
    [aux_sym_preproc_def_token1] = ACTIONS(5609),
    [aux_sym_preproc_if_token1] = ACTIONS(5609),
    [aux_sym_preproc_if_token2] = ACTIONS(5609),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5609),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5609),
    [aux_sym_preproc_else_token1] = ACTIONS(5609),
    [aux_sym_preproc_elif_token1] = ACTIONS(5609),
    [sym_preproc_directive] = ACTIONS(5609),
    [anon_sym_DASH] = ACTIONS(5611),
    [anon_sym_PLUS] = ACTIONS(5611),
    [anon_sym___extension__] = ACTIONS(5609),
    [anon_sym_typedef] = ACTIONS(5609),
    [anon_sym_extern] = ACTIONS(5609),
    [anon_sym___attribute__] = ACTIONS(5609),
    [anon_sym___attribute] = ACTIONS(5609),
    [anon_sym_noreturn] = ACTIONS(5609),
    [anon_sym_LBRACK] = ACTIONS(5611),
    [anon_sym___declspec] = ACTIONS(5609),
    [anon_sym___cdecl] = ACTIONS(5609),
    [anon_sym___clrcall] = ACTIONS(5609),
    [anon_sym___stdcall] = ACTIONS(5609),
    [anon_sym___fastcall] = ACTIONS(5609),
    [anon_sym___thiscall] = ACTIONS(5609),
    [anon_sym___vectorcall] = ACTIONS(5609),
    [anon_sym_signed] = ACTIONS(5609),
    [anon_sym_unsigned] = ACTIONS(5609),
    [anon_sym_long] = ACTIONS(5609),
    [anon_sym_short] = ACTIONS(5609),
    [anon_sym_static] = ACTIONS(5609),
    [anon_sym_auto] = ACTIONS(5609),
    [anon_sym_register] = ACTIONS(5609),
    [anon_sym_inline] = ACTIONS(5609),
    [anon_sym___inline] = ACTIONS(5609),
    [anon_sym___inline__] = ACTIONS(5609),
    [anon_sym___forceinline] = ACTIONS(5609),
    [anon_sym_thread_local] = ACTIONS(5609),
    [anon_sym___thread] = ACTIONS(5609),
    [anon_sym_CG_EXTERN] = ACTIONS(5609),
    [anon_sym_CG_INLINE] = ACTIONS(5609),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5609),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5609),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5609),
    [anon_sym_IBOutlet] = ACTIONS(5609),
    [anon_sym_IBInspectable] = ACTIONS(5609),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5609),
    [anon_sym_NS_INLINE] = ACTIONS(5609),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5609),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5609),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5609),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5609),
    [anon_sym_const] = ACTIONS(5609),
    [anon_sym_constexpr] = ACTIONS(5609),
    [anon_sym_volatile] = ACTIONS(5609),
    [anon_sym_restrict] = ACTIONS(5609),
    [anon_sym___restrict__] = ACTIONS(5609),
    [anon_sym__Atomic] = ACTIONS(5609),
    [anon_sym__Noreturn] = ACTIONS(5609),
    [anon_sym__Nonnull] = ACTIONS(5609),
    [anon_sym_nullable] = ACTIONS(5609),
    [anon_sym__Complex] = ACTIONS(5609),
    [anon_sym__Nullable] = ACTIONS(5609),
    [anon_sym__Nullable_result] = ACTIONS(5609),
    [anon_sym__Null_unspecified] = ACTIONS(5609),
    [anon_sym___autoreleasing] = ACTIONS(5609),
    [anon_sym___block] = ACTIONS(5609),
    [anon_sym___bridge] = ACTIONS(5609),
    [anon_sym___bridge_retained] = ACTIONS(5609),
    [anon_sym___bridge_transfer] = ACTIONS(5609),
    [anon_sym___complex] = ACTIONS(5609),
    [anon_sym___const] = ACTIONS(5609),
    [anon_sym___imag] = ACTIONS(5609),
    [anon_sym___kindof] = ACTIONS(5609),
    [anon_sym___nonnull] = ACTIONS(5609),
    [anon_sym___nullable] = ACTIONS(5609),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5609),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5609),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5609),
    [anon_sym___real] = ACTIONS(5609),
    [anon_sym___strong] = ACTIONS(5609),
    [anon_sym___unsafe_unretained] = ACTIONS(5609),
    [anon_sym___unused] = ACTIONS(5609),
    [anon_sym___weak] = ACTIONS(5609),
    [anon_sym_alignas] = ACTIONS(5609),
    [anon_sym__Alignas] = ACTIONS(5609),
    [sym_primitive_type] = ACTIONS(5609),
    [anon_sym_enum] = ACTIONS(5609),
    [anon_sym_struct] = ACTIONS(5609),
    [anon_sym_union] = ACTIONS(5609),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5609),
    [anon_sym___typeof] = ACTIONS(5609),
    [anon_sym_typeof] = ACTIONS(5609),
    [aux_sym_preproc_undef_token1] = ACTIONS(5609),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5609),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5609),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5609),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5609),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5609),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5609),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5609),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5609),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5609),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_API_AVAILABLE] = ACTIONS(5609),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_API_DEPRECATED] = ACTIONS(5609),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5609),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5609),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5609),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5609),
    [anon_sym___deprecated_msg] = ACTIONS(5609),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5609),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5609),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5609),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5609),
    [anon_sym_ATproperty] = ACTIONS(5611),
    [anon_sym_BOOL] = ACTIONS(5609),
    [anon_sym_IMP] = ACTIONS(5609),
    [anon_sym_SEL] = ACTIONS(5609),
    [anon_sym_Class] = ACTIONS(5609),
    [anon_sym_id] = ACTIONS(5609),
  },
  [STATE(2338)] = {
    [sym_identifier] = ACTIONS(5640),
    [aux_sym_preproc_def_token1] = ACTIONS(5640),
    [aux_sym_preproc_if_token1] = ACTIONS(5640),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5640),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5640),
    [sym_preproc_directive] = ACTIONS(5640),
    [anon_sym_DASH] = ACTIONS(5642),
    [anon_sym_PLUS] = ACTIONS(5642),
    [anon_sym___extension__] = ACTIONS(5640),
    [anon_sym_typedef] = ACTIONS(5640),
    [anon_sym_extern] = ACTIONS(5640),
    [anon_sym___attribute__] = ACTIONS(5640),
    [anon_sym___attribute] = ACTIONS(5640),
    [anon_sym_noreturn] = ACTIONS(5640),
    [anon_sym_LBRACK] = ACTIONS(5642),
    [anon_sym___declspec] = ACTIONS(5640),
    [anon_sym___cdecl] = ACTIONS(5640),
    [anon_sym___clrcall] = ACTIONS(5640),
    [anon_sym___stdcall] = ACTIONS(5640),
    [anon_sym___fastcall] = ACTIONS(5640),
    [anon_sym___thiscall] = ACTIONS(5640),
    [anon_sym___vectorcall] = ACTIONS(5640),
    [anon_sym_LBRACE] = ACTIONS(5642),
    [anon_sym_signed] = ACTIONS(5640),
    [anon_sym_unsigned] = ACTIONS(5640),
    [anon_sym_long] = ACTIONS(5640),
    [anon_sym_short] = ACTIONS(5640),
    [anon_sym_static] = ACTIONS(5640),
    [anon_sym_auto] = ACTIONS(5640),
    [anon_sym_register] = ACTIONS(5640),
    [anon_sym_inline] = ACTIONS(5640),
    [anon_sym___inline] = ACTIONS(5640),
    [anon_sym___inline__] = ACTIONS(5640),
    [anon_sym___forceinline] = ACTIONS(5640),
    [anon_sym_thread_local] = ACTIONS(5640),
    [anon_sym___thread] = ACTIONS(5640),
    [anon_sym_CG_EXTERN] = ACTIONS(5640),
    [anon_sym_CG_INLINE] = ACTIONS(5640),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5640),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5640),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5640),
    [anon_sym_IBOutlet] = ACTIONS(5640),
    [anon_sym_IBInspectable] = ACTIONS(5640),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5640),
    [anon_sym_NS_INLINE] = ACTIONS(5640),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5640),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5640),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5640),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5640),
    [anon_sym_const] = ACTIONS(5640),
    [anon_sym_constexpr] = ACTIONS(5640),
    [anon_sym_volatile] = ACTIONS(5640),
    [anon_sym_restrict] = ACTIONS(5640),
    [anon_sym___restrict__] = ACTIONS(5640),
    [anon_sym__Atomic] = ACTIONS(5640),
    [anon_sym__Noreturn] = ACTIONS(5640),
    [anon_sym__Nonnull] = ACTIONS(5640),
    [anon_sym_nullable] = ACTIONS(5640),
    [anon_sym__Complex] = ACTIONS(5640),
    [anon_sym__Nullable] = ACTIONS(5640),
    [anon_sym__Nullable_result] = ACTIONS(5640),
    [anon_sym__Null_unspecified] = ACTIONS(5640),
    [anon_sym___autoreleasing] = ACTIONS(5640),
    [anon_sym___block] = ACTIONS(5640),
    [anon_sym___bridge] = ACTIONS(5640),
    [anon_sym___bridge_retained] = ACTIONS(5640),
    [anon_sym___bridge_transfer] = ACTIONS(5640),
    [anon_sym___complex] = ACTIONS(5640),
    [anon_sym___const] = ACTIONS(5640),
    [anon_sym___imag] = ACTIONS(5640),
    [anon_sym___kindof] = ACTIONS(5640),
    [anon_sym___nonnull] = ACTIONS(5640),
    [anon_sym___nullable] = ACTIONS(5640),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5640),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5640),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5640),
    [anon_sym___real] = ACTIONS(5640),
    [anon_sym___strong] = ACTIONS(5640),
    [anon_sym___unsafe_unretained] = ACTIONS(5640),
    [anon_sym___unused] = ACTIONS(5640),
    [anon_sym___weak] = ACTIONS(5640),
    [anon_sym_alignas] = ACTIONS(5640),
    [anon_sym__Alignas] = ACTIONS(5640),
    [sym_primitive_type] = ACTIONS(5640),
    [anon_sym_enum] = ACTIONS(5640),
    [anon_sym_struct] = ACTIONS(5640),
    [anon_sym_union] = ACTIONS(5640),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5640),
    [anon_sym___typeof] = ACTIONS(5640),
    [anon_sym_typeof] = ACTIONS(5640),
    [aux_sym_preproc_undef_token1] = ACTIONS(5640),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5640),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5640),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5640),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5640),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5640),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5640),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5640),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5640),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5640),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5640),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5640),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5640),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5640),
    [anon_sym_API_AVAILABLE] = ACTIONS(5640),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5640),
    [anon_sym_API_DEPRECATED] = ACTIONS(5640),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5640),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5640),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5640),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5640),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5640),
    [anon_sym___deprecated_msg] = ACTIONS(5640),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5640),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5640),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5640),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5640),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5640),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5640),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5640),
    [anon_sym_ATend] = ACTIONS(5642),
    [anon_sym_ATsynthesize] = ACTIONS(5642),
    [anon_sym_ATdynamic] = ACTIONS(5642),
    [anon_sym_BOOL] = ACTIONS(5640),
    [anon_sym_IMP] = ACTIONS(5640),
    [anon_sym_SEL] = ACTIONS(5640),
    [anon_sym_Class] = ACTIONS(5640),
    [anon_sym_id] = ACTIONS(5640),
  },
  [STATE(2339)] = {
    [sym_identifier] = ACTIONS(5381),
    [aux_sym_preproc_def_token1] = ACTIONS(5381),
    [aux_sym_preproc_if_token1] = ACTIONS(5381),
    [aux_sym_preproc_if_token2] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5381),
    [sym_preproc_directive] = ACTIONS(5381),
    [anon_sym_DASH] = ACTIONS(5383),
    [anon_sym_PLUS] = ACTIONS(5383),
    [anon_sym_SEMI] = ACTIONS(5644),
    [anon_sym___extension__] = ACTIONS(5381),
    [anon_sym_typedef] = ACTIONS(5381),
    [anon_sym_extern] = ACTIONS(5381),
    [anon_sym___attribute__] = ACTIONS(5381),
    [anon_sym___attribute] = ACTIONS(5381),
    [anon_sym_noreturn] = ACTIONS(5381),
    [anon_sym_LBRACK] = ACTIONS(5383),
    [anon_sym___declspec] = ACTIONS(5381),
    [anon_sym___cdecl] = ACTIONS(5381),
    [anon_sym___clrcall] = ACTIONS(5381),
    [anon_sym___stdcall] = ACTIONS(5381),
    [anon_sym___fastcall] = ACTIONS(5381),
    [anon_sym___thiscall] = ACTIONS(5381),
    [anon_sym___vectorcall] = ACTIONS(5381),
    [anon_sym_signed] = ACTIONS(5381),
    [anon_sym_unsigned] = ACTIONS(5381),
    [anon_sym_long] = ACTIONS(5381),
    [anon_sym_short] = ACTIONS(5381),
    [anon_sym_static] = ACTIONS(5381),
    [anon_sym_auto] = ACTIONS(5381),
    [anon_sym_register] = ACTIONS(5381),
    [anon_sym_inline] = ACTIONS(5381),
    [anon_sym___inline] = ACTIONS(5381),
    [anon_sym___inline__] = ACTIONS(5381),
    [anon_sym___forceinline] = ACTIONS(5381),
    [anon_sym_thread_local] = ACTIONS(5381),
    [anon_sym___thread] = ACTIONS(5381),
    [anon_sym_CG_EXTERN] = ACTIONS(5381),
    [anon_sym_CG_INLINE] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5381),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5381),
    [anon_sym_IBOutlet] = ACTIONS(5381),
    [anon_sym_IBInspectable] = ACTIONS(5381),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5381),
    [anon_sym_NS_INLINE] = ACTIONS(5381),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5381),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5381),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5381),
    [anon_sym_const] = ACTIONS(5381),
    [anon_sym_constexpr] = ACTIONS(5381),
    [anon_sym_volatile] = ACTIONS(5381),
    [anon_sym_restrict] = ACTIONS(5381),
    [anon_sym___restrict__] = ACTIONS(5381),
    [anon_sym__Atomic] = ACTIONS(5381),
    [anon_sym__Noreturn] = ACTIONS(5381),
    [anon_sym__Nonnull] = ACTIONS(5381),
    [anon_sym_nullable] = ACTIONS(5381),
    [anon_sym__Complex] = ACTIONS(5381),
    [anon_sym__Nullable] = ACTIONS(5381),
    [anon_sym__Nullable_result] = ACTIONS(5381),
    [anon_sym__Null_unspecified] = ACTIONS(5381),
    [anon_sym___autoreleasing] = ACTIONS(5381),
    [anon_sym___block] = ACTIONS(5381),
    [anon_sym___bridge] = ACTIONS(5381),
    [anon_sym___bridge_retained] = ACTIONS(5381),
    [anon_sym___bridge_transfer] = ACTIONS(5381),
    [anon_sym___complex] = ACTIONS(5381),
    [anon_sym___const] = ACTIONS(5381),
    [anon_sym___imag] = ACTIONS(5381),
    [anon_sym___kindof] = ACTIONS(5381),
    [anon_sym___nonnull] = ACTIONS(5381),
    [anon_sym___nullable] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5381),
    [anon_sym___real] = ACTIONS(5381),
    [anon_sym___strong] = ACTIONS(5381),
    [anon_sym___unsafe_unretained] = ACTIONS(5381),
    [anon_sym___unused] = ACTIONS(5381),
    [anon_sym___weak] = ACTIONS(5381),
    [anon_sym_alignas] = ACTIONS(5381),
    [anon_sym__Alignas] = ACTIONS(5381),
    [sym_primitive_type] = ACTIONS(5381),
    [anon_sym_enum] = ACTIONS(5381),
    [anon_sym_struct] = ACTIONS(5381),
    [anon_sym_union] = ACTIONS(5381),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5381),
    [anon_sym___typeof] = ACTIONS(5381),
    [anon_sym_typeof] = ACTIONS(5381),
    [aux_sym_preproc_undef_token1] = ACTIONS(5381),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5381),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5381),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5381),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_API_AVAILABLE] = ACTIONS(5381),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_API_DEPRECATED] = ACTIONS(5381),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym___deprecated_msg] = ACTIONS(5381),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5381),
    [anon_sym_ATsynthesize] = ACTIONS(5383),
    [anon_sym_ATdynamic] = ACTIONS(5383),
    [anon_sym_BOOL] = ACTIONS(5381),
    [anon_sym_IMP] = ACTIONS(5381),
    [anon_sym_SEL] = ACTIONS(5381),
    [anon_sym_Class] = ACTIONS(5381),
    [anon_sym_id] = ACTIONS(5381),
  },
  [STATE(2340)] = {
    [sym_identifier] = ACTIONS(5646),
    [anon_sym_COMMA] = ACTIONS(5648),
    [anon_sym_RPAREN] = ACTIONS(5648),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5648),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5648),
    [anon_sym_LPAREN2] = ACTIONS(5648),
    [anon_sym_STAR] = ACTIONS(5648),
    [anon_sym_CARET] = ACTIONS(5648),
    [anon_sym_GT] = ACTIONS(5648),
    [anon_sym_SEMI] = ACTIONS(5648),
    [anon_sym___extension__] = ACTIONS(5646),
    [anon_sym_typedef] = ACTIONS(5646),
    [anon_sym_extern] = ACTIONS(5646),
    [anon_sym___attribute__] = ACTIONS(5646),
    [anon_sym___attribute] = ACTIONS(5646),
    [anon_sym_noreturn] = ACTIONS(5646),
    [anon_sym_LBRACK] = ACTIONS(5648),
    [anon_sym___declspec] = ACTIONS(5646),
    [anon_sym___based] = ACTIONS(5646),
    [anon_sym_LBRACE] = ACTIONS(5648),
    [anon_sym_RBRACE] = ACTIONS(5648),
    [anon_sym_signed] = ACTIONS(5646),
    [anon_sym_unsigned] = ACTIONS(5646),
    [anon_sym_long] = ACTIONS(5646),
    [anon_sym_short] = ACTIONS(5646),
    [anon_sym_static] = ACTIONS(5646),
    [anon_sym_ATautoreleasepool] = ACTIONS(5648),
    [anon_sym_auto] = ACTIONS(5646),
    [anon_sym_register] = ACTIONS(5646),
    [anon_sym_inline] = ACTIONS(5646),
    [anon_sym___inline] = ACTIONS(5646),
    [anon_sym___inline__] = ACTIONS(5646),
    [anon_sym___forceinline] = ACTIONS(5646),
    [anon_sym_thread_local] = ACTIONS(5646),
    [anon_sym___thread] = ACTIONS(5646),
    [anon_sym_CG_EXTERN] = ACTIONS(5646),
    [anon_sym_CG_INLINE] = ACTIONS(5646),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5646),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5646),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5646),
    [anon_sym_IBOutlet] = ACTIONS(5646),
    [anon_sym_IBInspectable] = ACTIONS(5646),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5646),
    [anon_sym_NS_INLINE] = ACTIONS(5646),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5646),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5646),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5646),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5646),
    [anon_sym_const] = ACTIONS(5646),
    [anon_sym_constexpr] = ACTIONS(5646),
    [anon_sym_volatile] = ACTIONS(5646),
    [anon_sym_restrict] = ACTIONS(5646),
    [anon_sym___restrict__] = ACTIONS(5646),
    [anon_sym__Atomic] = ACTIONS(5646),
    [anon_sym__Noreturn] = ACTIONS(5646),
    [anon_sym__Nonnull] = ACTIONS(5646),
    [anon_sym_nullable] = ACTIONS(5646),
    [anon_sym__Complex] = ACTIONS(5646),
    [anon_sym__Nullable] = ACTIONS(5646),
    [anon_sym__Nullable_result] = ACTIONS(5646),
    [anon_sym__Null_unspecified] = ACTIONS(5646),
    [anon_sym___autoreleasing] = ACTIONS(5646),
    [anon_sym___block] = ACTIONS(5646),
    [anon_sym___bridge] = ACTIONS(5646),
    [anon_sym___bridge_retained] = ACTIONS(5646),
    [anon_sym___bridge_transfer] = ACTIONS(5646),
    [anon_sym___complex] = ACTIONS(5646),
    [anon_sym___const] = ACTIONS(5646),
    [anon_sym___imag] = ACTIONS(5646),
    [anon_sym___kindof] = ACTIONS(5646),
    [anon_sym___nonnull] = ACTIONS(5646),
    [anon_sym___nullable] = ACTIONS(5646),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5646),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5646),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5646),
    [anon_sym___real] = ACTIONS(5646),
    [anon_sym___strong] = ACTIONS(5646),
    [anon_sym___unsafe_unretained] = ACTIONS(5646),
    [anon_sym___unused] = ACTIONS(5646),
    [anon_sym___weak] = ACTIONS(5646),
    [anon_sym_alignas] = ACTIONS(5646),
    [anon_sym__Alignas] = ACTIONS(5646),
    [sym_primitive_type] = ACTIONS(5646),
    [anon_sym_enum] = ACTIONS(5646),
    [anon_sym_COLON] = ACTIONS(5648),
    [anon_sym_struct] = ACTIONS(5646),
    [anon_sym_union] = ACTIONS(5646),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5646),
    [anon_sym___typeof] = ACTIONS(5646),
    [anon_sym_typeof] = ACTIONS(5646),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5646),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5646),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5646),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5646),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5646),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5646),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5646),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5646),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5646),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5646),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5646),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5646),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5646),
    [anon_sym_API_AVAILABLE] = ACTIONS(5646),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5646),
    [anon_sym_API_DEPRECATED] = ACTIONS(5646),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5646),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5646),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5646),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5646),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5646),
    [anon_sym___deprecated_msg] = ACTIONS(5646),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5646),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5646),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5646),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5646),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5646),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5646),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5646),
    [anon_sym_ATprotocol] = ACTIONS(5648),
    [anon_sym_ATinterface] = ACTIONS(5648),
    [anon_sym_ATimplementation] = ACTIONS(5648),
    [anon_sym_ATproperty] = ACTIONS(5648),
    [anon_sym_BOOL] = ACTIONS(5646),
    [anon_sym_IMP] = ACTIONS(5646),
    [anon_sym_SEL] = ACTIONS(5646),
    [anon_sym_Class] = ACTIONS(5646),
    [anon_sym_id] = ACTIONS(5646),
  },
  [STATE(2341)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_DASH] = ACTIONS(2430),
    [anon_sym_PLUS] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATproperty] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(2342)] = {
    [sym_identifier] = ACTIONS(5387),
    [aux_sym_preproc_def_token1] = ACTIONS(5387),
    [aux_sym_preproc_if_token1] = ACTIONS(5387),
    [aux_sym_preproc_if_token2] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5387),
    [sym_preproc_directive] = ACTIONS(5387),
    [anon_sym_DASH] = ACTIONS(5389),
    [anon_sym_PLUS] = ACTIONS(5389),
    [anon_sym_SEMI] = ACTIONS(5650),
    [anon_sym___extension__] = ACTIONS(5387),
    [anon_sym_typedef] = ACTIONS(5387),
    [anon_sym_extern] = ACTIONS(5387),
    [anon_sym___attribute__] = ACTIONS(5387),
    [anon_sym___attribute] = ACTIONS(5387),
    [anon_sym_noreturn] = ACTIONS(5387),
    [anon_sym_LBRACK] = ACTIONS(5389),
    [anon_sym___declspec] = ACTIONS(5387),
    [anon_sym___cdecl] = ACTIONS(5387),
    [anon_sym___clrcall] = ACTIONS(5387),
    [anon_sym___stdcall] = ACTIONS(5387),
    [anon_sym___fastcall] = ACTIONS(5387),
    [anon_sym___thiscall] = ACTIONS(5387),
    [anon_sym___vectorcall] = ACTIONS(5387),
    [anon_sym_signed] = ACTIONS(5387),
    [anon_sym_unsigned] = ACTIONS(5387),
    [anon_sym_long] = ACTIONS(5387),
    [anon_sym_short] = ACTIONS(5387),
    [anon_sym_static] = ACTIONS(5387),
    [anon_sym_auto] = ACTIONS(5387),
    [anon_sym_register] = ACTIONS(5387),
    [anon_sym_inline] = ACTIONS(5387),
    [anon_sym___inline] = ACTIONS(5387),
    [anon_sym___inline__] = ACTIONS(5387),
    [anon_sym___forceinline] = ACTIONS(5387),
    [anon_sym_thread_local] = ACTIONS(5387),
    [anon_sym___thread] = ACTIONS(5387),
    [anon_sym_CG_EXTERN] = ACTIONS(5387),
    [anon_sym_CG_INLINE] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5387),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5387),
    [anon_sym_IBOutlet] = ACTIONS(5387),
    [anon_sym_IBInspectable] = ACTIONS(5387),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5387),
    [anon_sym_NS_INLINE] = ACTIONS(5387),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5387),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5387),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5387),
    [anon_sym_const] = ACTIONS(5387),
    [anon_sym_constexpr] = ACTIONS(5387),
    [anon_sym_volatile] = ACTIONS(5387),
    [anon_sym_restrict] = ACTIONS(5387),
    [anon_sym___restrict__] = ACTIONS(5387),
    [anon_sym__Atomic] = ACTIONS(5387),
    [anon_sym__Noreturn] = ACTIONS(5387),
    [anon_sym__Nonnull] = ACTIONS(5387),
    [anon_sym_nullable] = ACTIONS(5387),
    [anon_sym__Complex] = ACTIONS(5387),
    [anon_sym__Nullable] = ACTIONS(5387),
    [anon_sym__Nullable_result] = ACTIONS(5387),
    [anon_sym__Null_unspecified] = ACTIONS(5387),
    [anon_sym___autoreleasing] = ACTIONS(5387),
    [anon_sym___block] = ACTIONS(5387),
    [anon_sym___bridge] = ACTIONS(5387),
    [anon_sym___bridge_retained] = ACTIONS(5387),
    [anon_sym___bridge_transfer] = ACTIONS(5387),
    [anon_sym___complex] = ACTIONS(5387),
    [anon_sym___const] = ACTIONS(5387),
    [anon_sym___imag] = ACTIONS(5387),
    [anon_sym___kindof] = ACTIONS(5387),
    [anon_sym___nonnull] = ACTIONS(5387),
    [anon_sym___nullable] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5387),
    [anon_sym___real] = ACTIONS(5387),
    [anon_sym___strong] = ACTIONS(5387),
    [anon_sym___unsafe_unretained] = ACTIONS(5387),
    [anon_sym___unused] = ACTIONS(5387),
    [anon_sym___weak] = ACTIONS(5387),
    [anon_sym_alignas] = ACTIONS(5387),
    [anon_sym__Alignas] = ACTIONS(5387),
    [sym_primitive_type] = ACTIONS(5387),
    [anon_sym_enum] = ACTIONS(5387),
    [anon_sym_struct] = ACTIONS(5387),
    [anon_sym_union] = ACTIONS(5387),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5387),
    [anon_sym___typeof] = ACTIONS(5387),
    [anon_sym_typeof] = ACTIONS(5387),
    [aux_sym_preproc_undef_token1] = ACTIONS(5387),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5387),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5387),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5387),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_API_AVAILABLE] = ACTIONS(5387),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_API_DEPRECATED] = ACTIONS(5387),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym___deprecated_msg] = ACTIONS(5387),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5387),
    [anon_sym_ATsynthesize] = ACTIONS(5389),
    [anon_sym_ATdynamic] = ACTIONS(5389),
    [anon_sym_BOOL] = ACTIONS(5387),
    [anon_sym_IMP] = ACTIONS(5387),
    [anon_sym_SEL] = ACTIONS(5387),
    [anon_sym_Class] = ACTIONS(5387),
    [anon_sym_id] = ACTIONS(5387),
  },
  [STATE(2343)] = {
    [sym_identifier] = ACTIONS(2428),
    [aux_sym_preproc_def_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token1] = ACTIONS(2428),
    [aux_sym_preproc_if_token2] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2428),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2428),
    [sym_preproc_directive] = ACTIONS(2428),
    [anon_sym_DASH] = ACTIONS(2430),
    [anon_sym_PLUS] = ACTIONS(2430),
    [anon_sym___extension__] = ACTIONS(2428),
    [anon_sym_typedef] = ACTIONS(2428),
    [anon_sym_extern] = ACTIONS(2428),
    [anon_sym___attribute__] = ACTIONS(2428),
    [anon_sym___attribute] = ACTIONS(2428),
    [anon_sym_noreturn] = ACTIONS(2428),
    [anon_sym_LBRACK] = ACTIONS(2430),
    [anon_sym___declspec] = ACTIONS(2428),
    [anon_sym___cdecl] = ACTIONS(2428),
    [anon_sym___clrcall] = ACTIONS(2428),
    [anon_sym___stdcall] = ACTIONS(2428),
    [anon_sym___fastcall] = ACTIONS(2428),
    [anon_sym___thiscall] = ACTIONS(2428),
    [anon_sym___vectorcall] = ACTIONS(2428),
    [anon_sym_signed] = ACTIONS(2428),
    [anon_sym_unsigned] = ACTIONS(2428),
    [anon_sym_long] = ACTIONS(2428),
    [anon_sym_short] = ACTIONS(2428),
    [anon_sym_static] = ACTIONS(2428),
    [anon_sym_auto] = ACTIONS(2428),
    [anon_sym_register] = ACTIONS(2428),
    [anon_sym_inline] = ACTIONS(2428),
    [anon_sym___inline] = ACTIONS(2428),
    [anon_sym___inline__] = ACTIONS(2428),
    [anon_sym___forceinline] = ACTIONS(2428),
    [anon_sym_thread_local] = ACTIONS(2428),
    [anon_sym___thread] = ACTIONS(2428),
    [anon_sym_CG_EXTERN] = ACTIONS(2428),
    [anon_sym_CG_INLINE] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2428),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2428),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2428),
    [anon_sym_IBOutlet] = ACTIONS(2428),
    [anon_sym_IBInspectable] = ACTIONS(2428),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2428),
    [anon_sym_NS_INLINE] = ACTIONS(2428),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2428),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2428),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2428),
    [anon_sym_const] = ACTIONS(2428),
    [anon_sym_constexpr] = ACTIONS(2428),
    [anon_sym_volatile] = ACTIONS(2428),
    [anon_sym_restrict] = ACTIONS(2428),
    [anon_sym___restrict__] = ACTIONS(2428),
    [anon_sym__Atomic] = ACTIONS(2428),
    [anon_sym__Noreturn] = ACTIONS(2428),
    [anon_sym__Nonnull] = ACTIONS(2428),
    [anon_sym_nullable] = ACTIONS(2428),
    [anon_sym__Complex] = ACTIONS(2428),
    [anon_sym__Nullable] = ACTIONS(2428),
    [anon_sym__Nullable_result] = ACTIONS(2428),
    [anon_sym__Null_unspecified] = ACTIONS(2428),
    [anon_sym___autoreleasing] = ACTIONS(2428),
    [anon_sym___block] = ACTIONS(2428),
    [anon_sym___bridge] = ACTIONS(2428),
    [anon_sym___bridge_retained] = ACTIONS(2428),
    [anon_sym___bridge_transfer] = ACTIONS(2428),
    [anon_sym___complex] = ACTIONS(2428),
    [anon_sym___const] = ACTIONS(2428),
    [anon_sym___imag] = ACTIONS(2428),
    [anon_sym___kindof] = ACTIONS(2428),
    [anon_sym___nonnull] = ACTIONS(2428),
    [anon_sym___nullable] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2428),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2428),
    [anon_sym___real] = ACTIONS(2428),
    [anon_sym___strong] = ACTIONS(2428),
    [anon_sym___unsafe_unretained] = ACTIONS(2428),
    [anon_sym___unused] = ACTIONS(2428),
    [anon_sym___weak] = ACTIONS(2428),
    [anon_sym_alignas] = ACTIONS(2428),
    [anon_sym__Alignas] = ACTIONS(2428),
    [sym_primitive_type] = ACTIONS(2428),
    [anon_sym_enum] = ACTIONS(2428),
    [anon_sym_struct] = ACTIONS(2428),
    [anon_sym_union] = ACTIONS(2428),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2428),
    [anon_sym___typeof] = ACTIONS(2428),
    [anon_sym_typeof] = ACTIONS(2428),
    [aux_sym_preproc_undef_token1] = ACTIONS(2428),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2428),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2428),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2428),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2428),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2428),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_API_AVAILABLE] = ACTIONS(2428),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_API_DEPRECATED] = ACTIONS(2428),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2428),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2428),
    [anon_sym___deprecated_msg] = ACTIONS(2428),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2428),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2428),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2428),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2428),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2428),
    [anon_sym_ATsynthesize] = ACTIONS(2430),
    [anon_sym_ATdynamic] = ACTIONS(2430),
    [anon_sym_ATproperty] = ACTIONS(2430),
    [anon_sym_BOOL] = ACTIONS(2428),
    [anon_sym_IMP] = ACTIONS(2428),
    [anon_sym_SEL] = ACTIONS(2428),
    [anon_sym_Class] = ACTIONS(2428),
    [anon_sym_id] = ACTIONS(2428),
  },
  [STATE(2344)] = {
    [sym_identifier] = ACTIONS(5628),
    [aux_sym_preproc_def_token1] = ACTIONS(5628),
    [aux_sym_preproc_if_token1] = ACTIONS(5628),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5628),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5628),
    [sym_preproc_directive] = ACTIONS(5628),
    [anon_sym_DASH] = ACTIONS(5630),
    [anon_sym_PLUS] = ACTIONS(5630),
    [anon_sym___extension__] = ACTIONS(5628),
    [anon_sym_typedef] = ACTIONS(5628),
    [anon_sym_extern] = ACTIONS(5628),
    [anon_sym___attribute__] = ACTIONS(5628),
    [anon_sym___attribute] = ACTIONS(5628),
    [anon_sym_noreturn] = ACTIONS(5628),
    [anon_sym_LBRACK] = ACTIONS(5630),
    [anon_sym___declspec] = ACTIONS(5628),
    [anon_sym___cdecl] = ACTIONS(5628),
    [anon_sym___clrcall] = ACTIONS(5628),
    [anon_sym___stdcall] = ACTIONS(5628),
    [anon_sym___fastcall] = ACTIONS(5628),
    [anon_sym___thiscall] = ACTIONS(5628),
    [anon_sym___vectorcall] = ACTIONS(5628),
    [anon_sym_signed] = ACTIONS(5628),
    [anon_sym_unsigned] = ACTIONS(5628),
    [anon_sym_long] = ACTIONS(5628),
    [anon_sym_short] = ACTIONS(5628),
    [anon_sym_static] = ACTIONS(5628),
    [anon_sym_auto] = ACTIONS(5628),
    [anon_sym_register] = ACTIONS(5628),
    [anon_sym_inline] = ACTIONS(5628),
    [anon_sym___inline] = ACTIONS(5628),
    [anon_sym___inline__] = ACTIONS(5628),
    [anon_sym___forceinline] = ACTIONS(5628),
    [anon_sym_thread_local] = ACTIONS(5628),
    [anon_sym___thread] = ACTIONS(5628),
    [anon_sym_CG_EXTERN] = ACTIONS(5628),
    [anon_sym_CG_INLINE] = ACTIONS(5628),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5628),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5628),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5628),
    [anon_sym_IBOutlet] = ACTIONS(5628),
    [anon_sym_IBInspectable] = ACTIONS(5628),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5628),
    [anon_sym_NS_INLINE] = ACTIONS(5628),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5628),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5628),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5628),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5628),
    [anon_sym_const] = ACTIONS(5628),
    [anon_sym_constexpr] = ACTIONS(5628),
    [anon_sym_volatile] = ACTIONS(5628),
    [anon_sym_restrict] = ACTIONS(5628),
    [anon_sym___restrict__] = ACTIONS(5628),
    [anon_sym__Atomic] = ACTIONS(5628),
    [anon_sym__Noreturn] = ACTIONS(5628),
    [anon_sym__Nonnull] = ACTIONS(5628),
    [anon_sym_nullable] = ACTIONS(5628),
    [anon_sym__Complex] = ACTIONS(5628),
    [anon_sym__Nullable] = ACTIONS(5628),
    [anon_sym__Nullable_result] = ACTIONS(5628),
    [anon_sym__Null_unspecified] = ACTIONS(5628),
    [anon_sym___autoreleasing] = ACTIONS(5628),
    [anon_sym___block] = ACTIONS(5628),
    [anon_sym___bridge] = ACTIONS(5628),
    [anon_sym___bridge_retained] = ACTIONS(5628),
    [anon_sym___bridge_transfer] = ACTIONS(5628),
    [anon_sym___complex] = ACTIONS(5628),
    [anon_sym___const] = ACTIONS(5628),
    [anon_sym___imag] = ACTIONS(5628),
    [anon_sym___kindof] = ACTIONS(5628),
    [anon_sym___nonnull] = ACTIONS(5628),
    [anon_sym___nullable] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5628),
    [anon_sym___real] = ACTIONS(5628),
    [anon_sym___strong] = ACTIONS(5628),
    [anon_sym___unsafe_unretained] = ACTIONS(5628),
    [anon_sym___unused] = ACTIONS(5628),
    [anon_sym___weak] = ACTIONS(5628),
    [anon_sym_alignas] = ACTIONS(5628),
    [anon_sym__Alignas] = ACTIONS(5628),
    [sym_primitive_type] = ACTIONS(5628),
    [anon_sym_enum] = ACTIONS(5628),
    [anon_sym_struct] = ACTIONS(5628),
    [anon_sym_union] = ACTIONS(5628),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5628),
    [anon_sym___typeof] = ACTIONS(5628),
    [anon_sym_typeof] = ACTIONS(5628),
    [aux_sym_preproc_undef_token1] = ACTIONS(5628),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5628),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5628),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5628),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5628),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5628),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5628),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5628),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5628),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5628),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_API_AVAILABLE] = ACTIONS(5628),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_API_DEPRECATED] = ACTIONS(5628),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5628),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5628),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5628),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5628),
    [anon_sym___deprecated_msg] = ACTIONS(5628),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5628),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5628),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5628),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5628),
    [anon_sym_ATend] = ACTIONS(5630),
    [anon_sym_AToptional] = ACTIONS(5630),
    [anon_sym_ATrequired] = ACTIONS(5630),
    [anon_sym_ATproperty] = ACTIONS(5630),
    [anon_sym_BOOL] = ACTIONS(5628),
    [anon_sym_IMP] = ACTIONS(5628),
    [anon_sym_SEL] = ACTIONS(5628),
    [anon_sym_Class] = ACTIONS(5628),
    [anon_sym_id] = ACTIONS(5628),
  },
  [STATE(2345)] = {
    [sym_identifier] = ACTIONS(5351),
    [aux_sym_preproc_def_token1] = ACTIONS(5351),
    [aux_sym_preproc_if_token1] = ACTIONS(5351),
    [aux_sym_preproc_if_token2] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5351),
    [sym_preproc_directive] = ACTIONS(5351),
    [anon_sym_DASH] = ACTIONS(5353),
    [anon_sym_PLUS] = ACTIONS(5353),
    [anon_sym_SEMI] = ACTIONS(5652),
    [anon_sym___extension__] = ACTIONS(5351),
    [anon_sym_typedef] = ACTIONS(5351),
    [anon_sym_extern] = ACTIONS(5351),
    [anon_sym___attribute__] = ACTIONS(5351),
    [anon_sym___attribute] = ACTIONS(5351),
    [anon_sym_noreturn] = ACTIONS(5351),
    [anon_sym_LBRACK] = ACTIONS(5353),
    [anon_sym___declspec] = ACTIONS(5351),
    [anon_sym___cdecl] = ACTIONS(5351),
    [anon_sym___clrcall] = ACTIONS(5351),
    [anon_sym___stdcall] = ACTIONS(5351),
    [anon_sym___fastcall] = ACTIONS(5351),
    [anon_sym___thiscall] = ACTIONS(5351),
    [anon_sym___vectorcall] = ACTIONS(5351),
    [anon_sym_signed] = ACTIONS(5351),
    [anon_sym_unsigned] = ACTIONS(5351),
    [anon_sym_long] = ACTIONS(5351),
    [anon_sym_short] = ACTIONS(5351),
    [anon_sym_static] = ACTIONS(5351),
    [anon_sym_auto] = ACTIONS(5351),
    [anon_sym_register] = ACTIONS(5351),
    [anon_sym_inline] = ACTIONS(5351),
    [anon_sym___inline] = ACTIONS(5351),
    [anon_sym___inline__] = ACTIONS(5351),
    [anon_sym___forceinline] = ACTIONS(5351),
    [anon_sym_thread_local] = ACTIONS(5351),
    [anon_sym___thread] = ACTIONS(5351),
    [anon_sym_CG_EXTERN] = ACTIONS(5351),
    [anon_sym_CG_INLINE] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5351),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5351),
    [anon_sym_IBOutlet] = ACTIONS(5351),
    [anon_sym_IBInspectable] = ACTIONS(5351),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5351),
    [anon_sym_NS_INLINE] = ACTIONS(5351),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5351),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5351),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5351),
    [anon_sym_const] = ACTIONS(5351),
    [anon_sym_constexpr] = ACTIONS(5351),
    [anon_sym_volatile] = ACTIONS(5351),
    [anon_sym_restrict] = ACTIONS(5351),
    [anon_sym___restrict__] = ACTIONS(5351),
    [anon_sym__Atomic] = ACTIONS(5351),
    [anon_sym__Noreturn] = ACTIONS(5351),
    [anon_sym__Nonnull] = ACTIONS(5351),
    [anon_sym_nullable] = ACTIONS(5351),
    [anon_sym__Complex] = ACTIONS(5351),
    [anon_sym__Nullable] = ACTIONS(5351),
    [anon_sym__Nullable_result] = ACTIONS(5351),
    [anon_sym__Null_unspecified] = ACTIONS(5351),
    [anon_sym___autoreleasing] = ACTIONS(5351),
    [anon_sym___block] = ACTIONS(5351),
    [anon_sym___bridge] = ACTIONS(5351),
    [anon_sym___bridge_retained] = ACTIONS(5351),
    [anon_sym___bridge_transfer] = ACTIONS(5351),
    [anon_sym___complex] = ACTIONS(5351),
    [anon_sym___const] = ACTIONS(5351),
    [anon_sym___imag] = ACTIONS(5351),
    [anon_sym___kindof] = ACTIONS(5351),
    [anon_sym___nonnull] = ACTIONS(5351),
    [anon_sym___nullable] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5351),
    [anon_sym___real] = ACTIONS(5351),
    [anon_sym___strong] = ACTIONS(5351),
    [anon_sym___unsafe_unretained] = ACTIONS(5351),
    [anon_sym___unused] = ACTIONS(5351),
    [anon_sym___weak] = ACTIONS(5351),
    [anon_sym_alignas] = ACTIONS(5351),
    [anon_sym__Alignas] = ACTIONS(5351),
    [sym_primitive_type] = ACTIONS(5351),
    [anon_sym_enum] = ACTIONS(5351),
    [anon_sym_struct] = ACTIONS(5351),
    [anon_sym_union] = ACTIONS(5351),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5351),
    [anon_sym___typeof] = ACTIONS(5351),
    [anon_sym_typeof] = ACTIONS(5351),
    [aux_sym_preproc_undef_token1] = ACTIONS(5351),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5351),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5351),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5351),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_API_AVAILABLE] = ACTIONS(5351),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_API_DEPRECATED] = ACTIONS(5351),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym___deprecated_msg] = ACTIONS(5351),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5351),
    [anon_sym_ATsynthesize] = ACTIONS(5353),
    [anon_sym_ATdynamic] = ACTIONS(5353),
    [anon_sym_BOOL] = ACTIONS(5351),
    [anon_sym_IMP] = ACTIONS(5351),
    [anon_sym_SEL] = ACTIONS(5351),
    [anon_sym_Class] = ACTIONS(5351),
    [anon_sym_id] = ACTIONS(5351),
  },
  [STATE(2346)] = {
    [sym_identifier] = ACTIONS(5485),
    [aux_sym_preproc_def_token1] = ACTIONS(5485),
    [aux_sym_preproc_if_token1] = ACTIONS(5485),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5485),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5485),
    [sym_preproc_directive] = ACTIONS(5485),
    [anon_sym_DASH] = ACTIONS(5487),
    [anon_sym_PLUS] = ACTIONS(5487),
    [anon_sym___extension__] = ACTIONS(5485),
    [anon_sym_typedef] = ACTIONS(5485),
    [anon_sym_extern] = ACTIONS(5485),
    [anon_sym___attribute__] = ACTIONS(5485),
    [anon_sym___attribute] = ACTIONS(5485),
    [anon_sym_noreturn] = ACTIONS(5485),
    [anon_sym_LBRACK] = ACTIONS(5487),
    [anon_sym___declspec] = ACTIONS(5485),
    [anon_sym___cdecl] = ACTIONS(5485),
    [anon_sym___clrcall] = ACTIONS(5485),
    [anon_sym___stdcall] = ACTIONS(5485),
    [anon_sym___fastcall] = ACTIONS(5485),
    [anon_sym___thiscall] = ACTIONS(5485),
    [anon_sym___vectorcall] = ACTIONS(5485),
    [anon_sym_signed] = ACTIONS(5485),
    [anon_sym_unsigned] = ACTIONS(5485),
    [anon_sym_long] = ACTIONS(5485),
    [anon_sym_short] = ACTIONS(5485),
    [anon_sym_static] = ACTIONS(5485),
    [anon_sym_auto] = ACTIONS(5485),
    [anon_sym_register] = ACTIONS(5485),
    [anon_sym_inline] = ACTIONS(5485),
    [anon_sym___inline] = ACTIONS(5485),
    [anon_sym___inline__] = ACTIONS(5485),
    [anon_sym___forceinline] = ACTIONS(5485),
    [anon_sym_thread_local] = ACTIONS(5485),
    [anon_sym___thread] = ACTIONS(5485),
    [anon_sym_CG_EXTERN] = ACTIONS(5485),
    [anon_sym_CG_INLINE] = ACTIONS(5485),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5485),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5485),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5485),
    [anon_sym_IBOutlet] = ACTIONS(5485),
    [anon_sym_IBInspectable] = ACTIONS(5485),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5485),
    [anon_sym_NS_INLINE] = ACTIONS(5485),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5485),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5485),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5485),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5485),
    [anon_sym_const] = ACTIONS(5485),
    [anon_sym_constexpr] = ACTIONS(5485),
    [anon_sym_volatile] = ACTIONS(5485),
    [anon_sym_restrict] = ACTIONS(5485),
    [anon_sym___restrict__] = ACTIONS(5485),
    [anon_sym__Atomic] = ACTIONS(5485),
    [anon_sym__Noreturn] = ACTIONS(5485),
    [anon_sym__Nonnull] = ACTIONS(5485),
    [anon_sym_nullable] = ACTIONS(5485),
    [anon_sym__Complex] = ACTIONS(5485),
    [anon_sym__Nullable] = ACTIONS(5485),
    [anon_sym__Nullable_result] = ACTIONS(5485),
    [anon_sym__Null_unspecified] = ACTIONS(5485),
    [anon_sym___autoreleasing] = ACTIONS(5485),
    [anon_sym___block] = ACTIONS(5485),
    [anon_sym___bridge] = ACTIONS(5485),
    [anon_sym___bridge_retained] = ACTIONS(5485),
    [anon_sym___bridge_transfer] = ACTIONS(5485),
    [anon_sym___complex] = ACTIONS(5485),
    [anon_sym___const] = ACTIONS(5485),
    [anon_sym___imag] = ACTIONS(5485),
    [anon_sym___kindof] = ACTIONS(5485),
    [anon_sym___nonnull] = ACTIONS(5485),
    [anon_sym___nullable] = ACTIONS(5485),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5485),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5485),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5485),
    [anon_sym___real] = ACTIONS(5485),
    [anon_sym___strong] = ACTIONS(5485),
    [anon_sym___unsafe_unretained] = ACTIONS(5485),
    [anon_sym___unused] = ACTIONS(5485),
    [anon_sym___weak] = ACTIONS(5485),
    [anon_sym_alignas] = ACTIONS(5485),
    [anon_sym__Alignas] = ACTIONS(5485),
    [sym_primitive_type] = ACTIONS(5485),
    [anon_sym_enum] = ACTIONS(5485),
    [anon_sym_struct] = ACTIONS(5485),
    [anon_sym_union] = ACTIONS(5485),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5485),
    [anon_sym___typeof] = ACTIONS(5485),
    [anon_sym_typeof] = ACTIONS(5485),
    [aux_sym_preproc_undef_token1] = ACTIONS(5485),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5485),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5485),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5485),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5485),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5485),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5485),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5485),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5485),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5485),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_API_AVAILABLE] = ACTIONS(5485),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_API_DEPRECATED] = ACTIONS(5485),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5485),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5485),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5485),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5485),
    [anon_sym___deprecated_msg] = ACTIONS(5485),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5485),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5485),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5485),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5485),
    [anon_sym_ATend] = ACTIONS(5487),
    [anon_sym_ATsynthesize] = ACTIONS(5487),
    [anon_sym_ATdynamic] = ACTIONS(5487),
    [anon_sym_BOOL] = ACTIONS(5485),
    [anon_sym_IMP] = ACTIONS(5485),
    [anon_sym_SEL] = ACTIONS(5485),
    [anon_sym_Class] = ACTIONS(5485),
    [anon_sym_id] = ACTIONS(5485),
  },
  [STATE(2347)] = {
    [sym_identifier] = ACTIONS(5485),
    [aux_sym_preproc_def_token1] = ACTIONS(5485),
    [aux_sym_preproc_if_token1] = ACTIONS(5485),
    [aux_sym_preproc_if_token2] = ACTIONS(5485),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5485),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5485),
    [sym_preproc_directive] = ACTIONS(5485),
    [anon_sym_DASH] = ACTIONS(5487),
    [anon_sym_PLUS] = ACTIONS(5487),
    [anon_sym___extension__] = ACTIONS(5485),
    [anon_sym_typedef] = ACTIONS(5485),
    [anon_sym_extern] = ACTIONS(5485),
    [anon_sym___attribute__] = ACTIONS(5485),
    [anon_sym___attribute] = ACTIONS(5485),
    [anon_sym_noreturn] = ACTIONS(5485),
    [anon_sym_LBRACK] = ACTIONS(5487),
    [anon_sym___declspec] = ACTIONS(5485),
    [anon_sym___cdecl] = ACTIONS(5485),
    [anon_sym___clrcall] = ACTIONS(5485),
    [anon_sym___stdcall] = ACTIONS(5485),
    [anon_sym___fastcall] = ACTIONS(5485),
    [anon_sym___thiscall] = ACTIONS(5485),
    [anon_sym___vectorcall] = ACTIONS(5485),
    [anon_sym_signed] = ACTIONS(5485),
    [anon_sym_unsigned] = ACTIONS(5485),
    [anon_sym_long] = ACTIONS(5485),
    [anon_sym_short] = ACTIONS(5485),
    [anon_sym_static] = ACTIONS(5485),
    [anon_sym_auto] = ACTIONS(5485),
    [anon_sym_register] = ACTIONS(5485),
    [anon_sym_inline] = ACTIONS(5485),
    [anon_sym___inline] = ACTIONS(5485),
    [anon_sym___inline__] = ACTIONS(5485),
    [anon_sym___forceinline] = ACTIONS(5485),
    [anon_sym_thread_local] = ACTIONS(5485),
    [anon_sym___thread] = ACTIONS(5485),
    [anon_sym_CG_EXTERN] = ACTIONS(5485),
    [anon_sym_CG_INLINE] = ACTIONS(5485),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5485),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5485),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5485),
    [anon_sym_IBOutlet] = ACTIONS(5485),
    [anon_sym_IBInspectable] = ACTIONS(5485),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5485),
    [anon_sym_NS_INLINE] = ACTIONS(5485),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5485),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5485),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5485),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5485),
    [anon_sym_const] = ACTIONS(5485),
    [anon_sym_constexpr] = ACTIONS(5485),
    [anon_sym_volatile] = ACTIONS(5485),
    [anon_sym_restrict] = ACTIONS(5485),
    [anon_sym___restrict__] = ACTIONS(5485),
    [anon_sym__Atomic] = ACTIONS(5485),
    [anon_sym__Noreturn] = ACTIONS(5485),
    [anon_sym__Nonnull] = ACTIONS(5485),
    [anon_sym_nullable] = ACTIONS(5485),
    [anon_sym__Complex] = ACTIONS(5485),
    [anon_sym__Nullable] = ACTIONS(5485),
    [anon_sym__Nullable_result] = ACTIONS(5485),
    [anon_sym__Null_unspecified] = ACTIONS(5485),
    [anon_sym___autoreleasing] = ACTIONS(5485),
    [anon_sym___block] = ACTIONS(5485),
    [anon_sym___bridge] = ACTIONS(5485),
    [anon_sym___bridge_retained] = ACTIONS(5485),
    [anon_sym___bridge_transfer] = ACTIONS(5485),
    [anon_sym___complex] = ACTIONS(5485),
    [anon_sym___const] = ACTIONS(5485),
    [anon_sym___imag] = ACTIONS(5485),
    [anon_sym___kindof] = ACTIONS(5485),
    [anon_sym___nonnull] = ACTIONS(5485),
    [anon_sym___nullable] = ACTIONS(5485),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5485),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5485),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5485),
    [anon_sym___real] = ACTIONS(5485),
    [anon_sym___strong] = ACTIONS(5485),
    [anon_sym___unsafe_unretained] = ACTIONS(5485),
    [anon_sym___unused] = ACTIONS(5485),
    [anon_sym___weak] = ACTIONS(5485),
    [anon_sym_alignas] = ACTIONS(5485),
    [anon_sym__Alignas] = ACTIONS(5485),
    [sym_primitive_type] = ACTIONS(5485),
    [anon_sym_enum] = ACTIONS(5485),
    [anon_sym_struct] = ACTIONS(5485),
    [anon_sym_union] = ACTIONS(5485),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5485),
    [anon_sym___typeof] = ACTIONS(5485),
    [anon_sym_typeof] = ACTIONS(5485),
    [aux_sym_preproc_undef_token1] = ACTIONS(5485),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5485),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5485),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5485),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5485),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5485),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5485),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5485),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5485),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5485),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_API_AVAILABLE] = ACTIONS(5485),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_API_DEPRECATED] = ACTIONS(5485),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5485),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5485),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5485),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5485),
    [anon_sym___deprecated_msg] = ACTIONS(5485),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5485),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5485),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5485),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5485),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5485),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5485),
    [anon_sym_ATsynthesize] = ACTIONS(5487),
    [anon_sym_ATdynamic] = ACTIONS(5487),
    [anon_sym_BOOL] = ACTIONS(5485),
    [anon_sym_IMP] = ACTIONS(5485),
    [anon_sym_SEL] = ACTIONS(5485),
    [anon_sym_Class] = ACTIONS(5485),
    [anon_sym_id] = ACTIONS(5485),
  },
  [STATE(2348)] = {
    [sym_identifier] = ACTIONS(5654),
    [anon_sym_COMMA] = ACTIONS(5656),
    [anon_sym_RPAREN] = ACTIONS(5656),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5656),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5656),
    [anon_sym_LPAREN2] = ACTIONS(5656),
    [anon_sym_STAR] = ACTIONS(5656),
    [anon_sym_CARET] = ACTIONS(5656),
    [anon_sym_GT] = ACTIONS(5656),
    [anon_sym_SEMI] = ACTIONS(5656),
    [anon_sym___extension__] = ACTIONS(5654),
    [anon_sym_extern] = ACTIONS(5654),
    [anon_sym___attribute__] = ACTIONS(5654),
    [anon_sym___attribute] = ACTIONS(5654),
    [anon_sym_noreturn] = ACTIONS(5654),
    [anon_sym_LBRACK] = ACTIONS(5656),
    [anon_sym___declspec] = ACTIONS(5654),
    [anon_sym___based] = ACTIONS(5654),
    [anon_sym_LBRACE] = ACTIONS(5656),
    [anon_sym_RBRACE] = ACTIONS(5656),
    [anon_sym_signed] = ACTIONS(5654),
    [anon_sym_unsigned] = ACTIONS(5654),
    [anon_sym_long] = ACTIONS(5654),
    [anon_sym_short] = ACTIONS(5654),
    [anon_sym_static] = ACTIONS(5654),
    [anon_sym_ATautoreleasepool] = ACTIONS(5656),
    [anon_sym_auto] = ACTIONS(5654),
    [anon_sym_register] = ACTIONS(5654),
    [anon_sym_inline] = ACTIONS(5654),
    [anon_sym___inline] = ACTIONS(5654),
    [anon_sym___inline__] = ACTIONS(5654),
    [anon_sym___forceinline] = ACTIONS(5654),
    [anon_sym_thread_local] = ACTIONS(5654),
    [anon_sym___thread] = ACTIONS(5654),
    [anon_sym_CG_EXTERN] = ACTIONS(5654),
    [anon_sym_CG_INLINE] = ACTIONS(5654),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5654),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5654),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5654),
    [anon_sym_IBOutlet] = ACTIONS(5654),
    [anon_sym_IBInspectable] = ACTIONS(5654),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5654),
    [anon_sym_NS_INLINE] = ACTIONS(5654),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5654),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5654),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5654),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5654),
    [anon_sym_const] = ACTIONS(5654),
    [anon_sym_constexpr] = ACTIONS(5654),
    [anon_sym_volatile] = ACTIONS(5654),
    [anon_sym_restrict] = ACTIONS(5654),
    [anon_sym___restrict__] = ACTIONS(5654),
    [anon_sym__Atomic] = ACTIONS(5654),
    [anon_sym__Noreturn] = ACTIONS(5654),
    [anon_sym__Nonnull] = ACTIONS(5654),
    [anon_sym_nullable] = ACTIONS(5654),
    [anon_sym__Complex] = ACTIONS(5654),
    [anon_sym__Nullable] = ACTIONS(5654),
    [anon_sym__Nullable_result] = ACTIONS(5654),
    [anon_sym__Null_unspecified] = ACTIONS(5654),
    [anon_sym___autoreleasing] = ACTIONS(5654),
    [anon_sym___block] = ACTIONS(5654),
    [anon_sym___bridge] = ACTIONS(5654),
    [anon_sym___bridge_retained] = ACTIONS(5654),
    [anon_sym___bridge_transfer] = ACTIONS(5654),
    [anon_sym___complex] = ACTIONS(5654),
    [anon_sym___const] = ACTIONS(5654),
    [anon_sym___imag] = ACTIONS(5654),
    [anon_sym___kindof] = ACTIONS(5654),
    [anon_sym___nonnull] = ACTIONS(5654),
    [anon_sym___nullable] = ACTIONS(5654),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5654),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5654),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5654),
    [anon_sym___real] = ACTIONS(5654),
    [anon_sym___strong] = ACTIONS(5654),
    [anon_sym___unsafe_unretained] = ACTIONS(5654),
    [anon_sym___unused] = ACTIONS(5654),
    [anon_sym___weak] = ACTIONS(5654),
    [anon_sym_alignas] = ACTIONS(5654),
    [anon_sym__Alignas] = ACTIONS(5654),
    [sym_primitive_type] = ACTIONS(5654),
    [anon_sym_enum] = ACTIONS(5654),
    [anon_sym_COLON] = ACTIONS(5656),
    [anon_sym_struct] = ACTIONS(5654),
    [anon_sym_union] = ACTIONS(5654),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5654),
    [anon_sym___typeof] = ACTIONS(5654),
    [anon_sym_typeof] = ACTIONS(5654),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5654),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5654),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5654),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5654),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5654),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5654),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5654),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5654),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5654),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5654),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5654),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5654),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5654),
    [anon_sym_API_AVAILABLE] = ACTIONS(5654),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5654),
    [anon_sym_API_DEPRECATED] = ACTIONS(5654),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5654),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5654),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5654),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5654),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5654),
    [anon_sym___deprecated_msg] = ACTIONS(5654),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5654),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5654),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5654),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5654),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5654),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5654),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5654),
    [anon_sym_ATprotocol] = ACTIONS(5656),
    [anon_sym_ATinterface] = ACTIONS(5656),
    [anon_sym_ATimplementation] = ACTIONS(5656),
    [anon_sym_ATproperty] = ACTIONS(5656),
    [anon_sym_BOOL] = ACTIONS(5654),
    [anon_sym_IMP] = ACTIONS(5654),
    [anon_sym_SEL] = ACTIONS(5654),
    [anon_sym_Class] = ACTIONS(5654),
    [anon_sym_id] = ACTIONS(5654),
  },
  [STATE(2349)] = {
    [sym_identifier] = ACTIONS(5658),
    [anon_sym_COMMA] = ACTIONS(5660),
    [anon_sym_RPAREN] = ACTIONS(5660),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5660),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5660),
    [anon_sym_LPAREN2] = ACTIONS(5660),
    [anon_sym_STAR] = ACTIONS(5660),
    [anon_sym_CARET] = ACTIONS(5660),
    [anon_sym_GT] = ACTIONS(5660),
    [anon_sym_SEMI] = ACTIONS(5660),
    [anon_sym___extension__] = ACTIONS(5658),
    [anon_sym_extern] = ACTIONS(5658),
    [anon_sym___attribute__] = ACTIONS(5658),
    [anon_sym___attribute] = ACTIONS(5658),
    [anon_sym_noreturn] = ACTIONS(5658),
    [anon_sym_LBRACK] = ACTIONS(5660),
    [anon_sym___declspec] = ACTIONS(5658),
    [anon_sym___based] = ACTIONS(5658),
    [anon_sym_LBRACE] = ACTIONS(5660),
    [anon_sym_RBRACE] = ACTIONS(5660),
    [anon_sym_signed] = ACTIONS(5658),
    [anon_sym_unsigned] = ACTIONS(5658),
    [anon_sym_long] = ACTIONS(5658),
    [anon_sym_short] = ACTIONS(5658),
    [anon_sym_static] = ACTIONS(5658),
    [anon_sym_ATautoreleasepool] = ACTIONS(5660),
    [anon_sym_auto] = ACTIONS(5658),
    [anon_sym_register] = ACTIONS(5658),
    [anon_sym_inline] = ACTIONS(5658),
    [anon_sym___inline] = ACTIONS(5658),
    [anon_sym___inline__] = ACTIONS(5658),
    [anon_sym___forceinline] = ACTIONS(5658),
    [anon_sym_thread_local] = ACTIONS(5658),
    [anon_sym___thread] = ACTIONS(5658),
    [anon_sym_CG_EXTERN] = ACTIONS(5658),
    [anon_sym_CG_INLINE] = ACTIONS(5658),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5658),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5658),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5658),
    [anon_sym_IBOutlet] = ACTIONS(5658),
    [anon_sym_IBInspectable] = ACTIONS(5658),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5658),
    [anon_sym_NS_INLINE] = ACTIONS(5658),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5658),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5658),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5658),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5658),
    [anon_sym_const] = ACTIONS(5658),
    [anon_sym_constexpr] = ACTIONS(5658),
    [anon_sym_volatile] = ACTIONS(5658),
    [anon_sym_restrict] = ACTIONS(5658),
    [anon_sym___restrict__] = ACTIONS(5658),
    [anon_sym__Atomic] = ACTIONS(5658),
    [anon_sym__Noreturn] = ACTIONS(5658),
    [anon_sym__Nonnull] = ACTIONS(5658),
    [anon_sym_nullable] = ACTIONS(5658),
    [anon_sym__Complex] = ACTIONS(5658),
    [anon_sym__Nullable] = ACTIONS(5658),
    [anon_sym__Nullable_result] = ACTIONS(5658),
    [anon_sym__Null_unspecified] = ACTIONS(5658),
    [anon_sym___autoreleasing] = ACTIONS(5658),
    [anon_sym___block] = ACTIONS(5658),
    [anon_sym___bridge] = ACTIONS(5658),
    [anon_sym___bridge_retained] = ACTIONS(5658),
    [anon_sym___bridge_transfer] = ACTIONS(5658),
    [anon_sym___complex] = ACTIONS(5658),
    [anon_sym___const] = ACTIONS(5658),
    [anon_sym___imag] = ACTIONS(5658),
    [anon_sym___kindof] = ACTIONS(5658),
    [anon_sym___nonnull] = ACTIONS(5658),
    [anon_sym___nullable] = ACTIONS(5658),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5658),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5658),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5658),
    [anon_sym___real] = ACTIONS(5658),
    [anon_sym___strong] = ACTIONS(5658),
    [anon_sym___unsafe_unretained] = ACTIONS(5658),
    [anon_sym___unused] = ACTIONS(5658),
    [anon_sym___weak] = ACTIONS(5658),
    [anon_sym_alignas] = ACTIONS(5658),
    [anon_sym__Alignas] = ACTIONS(5658),
    [sym_primitive_type] = ACTIONS(5658),
    [anon_sym_enum] = ACTIONS(5658),
    [anon_sym_COLON] = ACTIONS(5660),
    [anon_sym_struct] = ACTIONS(5658),
    [anon_sym_union] = ACTIONS(5658),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5658),
    [anon_sym___typeof] = ACTIONS(5658),
    [anon_sym_typeof] = ACTIONS(5658),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5658),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5658),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5658),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5658),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5658),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5658),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5658),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5658),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5658),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5658),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5658),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5658),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5658),
    [anon_sym_API_AVAILABLE] = ACTIONS(5658),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5658),
    [anon_sym_API_DEPRECATED] = ACTIONS(5658),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5658),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5658),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5658),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5658),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5658),
    [anon_sym___deprecated_msg] = ACTIONS(5658),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5658),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5658),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5658),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5658),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5658),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5658),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5658),
    [anon_sym_ATprotocol] = ACTIONS(5660),
    [anon_sym_ATinterface] = ACTIONS(5660),
    [anon_sym_ATimplementation] = ACTIONS(5660),
    [anon_sym_ATproperty] = ACTIONS(5660),
    [anon_sym_BOOL] = ACTIONS(5658),
    [anon_sym_IMP] = ACTIONS(5658),
    [anon_sym_SEL] = ACTIONS(5658),
    [anon_sym_Class] = ACTIONS(5658),
    [anon_sym_id] = ACTIONS(5658),
  },
  [STATE(2350)] = {
    [sym_identifier] = ACTIONS(5399),
    [aux_sym_preproc_def_token1] = ACTIONS(5399),
    [aux_sym_preproc_if_token1] = ACTIONS(5399),
    [aux_sym_preproc_if_token2] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5399),
    [sym_preproc_directive] = ACTIONS(5399),
    [anon_sym_DASH] = ACTIONS(5401),
    [anon_sym_PLUS] = ACTIONS(5401),
    [anon_sym___extension__] = ACTIONS(5399),
    [anon_sym_typedef] = ACTIONS(5399),
    [anon_sym_extern] = ACTIONS(5399),
    [anon_sym___attribute__] = ACTIONS(5399),
    [anon_sym___attribute] = ACTIONS(5399),
    [anon_sym_noreturn] = ACTIONS(5399),
    [anon_sym_LBRACK] = ACTIONS(5401),
    [anon_sym___declspec] = ACTIONS(5399),
    [anon_sym___cdecl] = ACTIONS(5399),
    [anon_sym___clrcall] = ACTIONS(5399),
    [anon_sym___stdcall] = ACTIONS(5399),
    [anon_sym___fastcall] = ACTIONS(5399),
    [anon_sym___thiscall] = ACTIONS(5399),
    [anon_sym___vectorcall] = ACTIONS(5399),
    [anon_sym_signed] = ACTIONS(5399),
    [anon_sym_unsigned] = ACTIONS(5399),
    [anon_sym_long] = ACTIONS(5399),
    [anon_sym_short] = ACTIONS(5399),
    [anon_sym_static] = ACTIONS(5399),
    [anon_sym_auto] = ACTIONS(5399),
    [anon_sym_register] = ACTIONS(5399),
    [anon_sym_inline] = ACTIONS(5399),
    [anon_sym___inline] = ACTIONS(5399),
    [anon_sym___inline__] = ACTIONS(5399),
    [anon_sym___forceinline] = ACTIONS(5399),
    [anon_sym_thread_local] = ACTIONS(5399),
    [anon_sym___thread] = ACTIONS(5399),
    [anon_sym_CG_EXTERN] = ACTIONS(5399),
    [anon_sym_CG_INLINE] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5399),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5399),
    [anon_sym_IBOutlet] = ACTIONS(5399),
    [anon_sym_IBInspectable] = ACTIONS(5399),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5399),
    [anon_sym_NS_INLINE] = ACTIONS(5399),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5399),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5399),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5399),
    [anon_sym_const] = ACTIONS(5399),
    [anon_sym_constexpr] = ACTIONS(5399),
    [anon_sym_volatile] = ACTIONS(5399),
    [anon_sym_restrict] = ACTIONS(5399),
    [anon_sym___restrict__] = ACTIONS(5399),
    [anon_sym__Atomic] = ACTIONS(5399),
    [anon_sym__Noreturn] = ACTIONS(5399),
    [anon_sym__Nonnull] = ACTIONS(5399),
    [anon_sym_nullable] = ACTIONS(5399),
    [anon_sym__Complex] = ACTIONS(5399),
    [anon_sym__Nullable] = ACTIONS(5399),
    [anon_sym__Nullable_result] = ACTIONS(5399),
    [anon_sym__Null_unspecified] = ACTIONS(5399),
    [anon_sym___autoreleasing] = ACTIONS(5399),
    [anon_sym___block] = ACTIONS(5399),
    [anon_sym___bridge] = ACTIONS(5399),
    [anon_sym___bridge_retained] = ACTIONS(5399),
    [anon_sym___bridge_transfer] = ACTIONS(5399),
    [anon_sym___complex] = ACTIONS(5399),
    [anon_sym___const] = ACTIONS(5399),
    [anon_sym___imag] = ACTIONS(5399),
    [anon_sym___kindof] = ACTIONS(5399),
    [anon_sym___nonnull] = ACTIONS(5399),
    [anon_sym___nullable] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5399),
    [anon_sym___real] = ACTIONS(5399),
    [anon_sym___strong] = ACTIONS(5399),
    [anon_sym___unsafe_unretained] = ACTIONS(5399),
    [anon_sym___unused] = ACTIONS(5399),
    [anon_sym___weak] = ACTIONS(5399),
    [anon_sym_alignas] = ACTIONS(5399),
    [anon_sym__Alignas] = ACTIONS(5399),
    [sym_primitive_type] = ACTIONS(5399),
    [anon_sym_enum] = ACTIONS(5399),
    [anon_sym_struct] = ACTIONS(5399),
    [anon_sym_union] = ACTIONS(5399),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5399),
    [anon_sym___typeof] = ACTIONS(5399),
    [anon_sym_typeof] = ACTIONS(5399),
    [aux_sym_preproc_undef_token1] = ACTIONS(5399),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5399),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5399),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5399),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_API_AVAILABLE] = ACTIONS(5399),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_API_DEPRECATED] = ACTIONS(5399),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym___deprecated_msg] = ACTIONS(5399),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5399),
    [anon_sym_ATsynthesize] = ACTIONS(5401),
    [anon_sym_ATdynamic] = ACTIONS(5401),
    [anon_sym_BOOL] = ACTIONS(5399),
    [anon_sym_IMP] = ACTIONS(5399),
    [anon_sym_SEL] = ACTIONS(5399),
    [anon_sym_Class] = ACTIONS(5399),
    [anon_sym_id] = ACTIONS(5399),
  },
  [STATE(2351)] = {
    [sym_identifier] = ACTIONS(5513),
    [aux_sym_preproc_def_token1] = ACTIONS(5513),
    [aux_sym_preproc_if_token1] = ACTIONS(5513),
    [aux_sym_preproc_if_token2] = ACTIONS(5513),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5513),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5513),
    [sym_preproc_directive] = ACTIONS(5513),
    [anon_sym_DASH] = ACTIONS(5515),
    [anon_sym_PLUS] = ACTIONS(5515),
    [anon_sym___extension__] = ACTIONS(5513),
    [anon_sym_typedef] = ACTIONS(5513),
    [anon_sym_extern] = ACTIONS(5513),
    [anon_sym___attribute__] = ACTIONS(5513),
    [anon_sym___attribute] = ACTIONS(5513),
    [anon_sym_noreturn] = ACTIONS(5513),
    [anon_sym_LBRACK] = ACTIONS(5515),
    [anon_sym___declspec] = ACTIONS(5513),
    [anon_sym___cdecl] = ACTIONS(5513),
    [anon_sym___clrcall] = ACTIONS(5513),
    [anon_sym___stdcall] = ACTIONS(5513),
    [anon_sym___fastcall] = ACTIONS(5513),
    [anon_sym___thiscall] = ACTIONS(5513),
    [anon_sym___vectorcall] = ACTIONS(5513),
    [anon_sym_signed] = ACTIONS(5513),
    [anon_sym_unsigned] = ACTIONS(5513),
    [anon_sym_long] = ACTIONS(5513),
    [anon_sym_short] = ACTIONS(5513),
    [anon_sym_static] = ACTIONS(5513),
    [anon_sym_auto] = ACTIONS(5513),
    [anon_sym_register] = ACTIONS(5513),
    [anon_sym_inline] = ACTIONS(5513),
    [anon_sym___inline] = ACTIONS(5513),
    [anon_sym___inline__] = ACTIONS(5513),
    [anon_sym___forceinline] = ACTIONS(5513),
    [anon_sym_thread_local] = ACTIONS(5513),
    [anon_sym___thread] = ACTIONS(5513),
    [anon_sym_CG_EXTERN] = ACTIONS(5513),
    [anon_sym_CG_INLINE] = ACTIONS(5513),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5513),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5513),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5513),
    [anon_sym_IBOutlet] = ACTIONS(5513),
    [anon_sym_IBInspectable] = ACTIONS(5513),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5513),
    [anon_sym_NS_INLINE] = ACTIONS(5513),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5513),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5513),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5513),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5513),
    [anon_sym_const] = ACTIONS(5513),
    [anon_sym_constexpr] = ACTIONS(5513),
    [anon_sym_volatile] = ACTIONS(5513),
    [anon_sym_restrict] = ACTIONS(5513),
    [anon_sym___restrict__] = ACTIONS(5513),
    [anon_sym__Atomic] = ACTIONS(5513),
    [anon_sym__Noreturn] = ACTIONS(5513),
    [anon_sym__Nonnull] = ACTIONS(5513),
    [anon_sym_nullable] = ACTIONS(5513),
    [anon_sym__Complex] = ACTIONS(5513),
    [anon_sym__Nullable] = ACTIONS(5513),
    [anon_sym__Nullable_result] = ACTIONS(5513),
    [anon_sym__Null_unspecified] = ACTIONS(5513),
    [anon_sym___autoreleasing] = ACTIONS(5513),
    [anon_sym___block] = ACTIONS(5513),
    [anon_sym___bridge] = ACTIONS(5513),
    [anon_sym___bridge_retained] = ACTIONS(5513),
    [anon_sym___bridge_transfer] = ACTIONS(5513),
    [anon_sym___complex] = ACTIONS(5513),
    [anon_sym___const] = ACTIONS(5513),
    [anon_sym___imag] = ACTIONS(5513),
    [anon_sym___kindof] = ACTIONS(5513),
    [anon_sym___nonnull] = ACTIONS(5513),
    [anon_sym___nullable] = ACTIONS(5513),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5513),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5513),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5513),
    [anon_sym___real] = ACTIONS(5513),
    [anon_sym___strong] = ACTIONS(5513),
    [anon_sym___unsafe_unretained] = ACTIONS(5513),
    [anon_sym___unused] = ACTIONS(5513),
    [anon_sym___weak] = ACTIONS(5513),
    [anon_sym_alignas] = ACTIONS(5513),
    [anon_sym__Alignas] = ACTIONS(5513),
    [sym_primitive_type] = ACTIONS(5513),
    [anon_sym_enum] = ACTIONS(5513),
    [anon_sym_struct] = ACTIONS(5513),
    [anon_sym_union] = ACTIONS(5513),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5513),
    [anon_sym___typeof] = ACTIONS(5513),
    [anon_sym_typeof] = ACTIONS(5513),
    [aux_sym_preproc_undef_token1] = ACTIONS(5513),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5513),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5513),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5513),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5513),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5513),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5513),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5513),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5513),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5513),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_API_AVAILABLE] = ACTIONS(5513),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_API_DEPRECATED] = ACTIONS(5513),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5513),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5513),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5513),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5513),
    [anon_sym___deprecated_msg] = ACTIONS(5513),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5513),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5513),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5513),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5513),
    [anon_sym_ATsynthesize] = ACTIONS(5515),
    [anon_sym_ATdynamic] = ACTIONS(5515),
    [anon_sym_BOOL] = ACTIONS(5513),
    [anon_sym_IMP] = ACTIONS(5513),
    [anon_sym_SEL] = ACTIONS(5513),
    [anon_sym_Class] = ACTIONS(5513),
    [anon_sym_id] = ACTIONS(5513),
  },
  [STATE(2352)] = {
    [sym_identifier] = ACTIONS(5375),
    [aux_sym_preproc_def_token1] = ACTIONS(5375),
    [aux_sym_preproc_if_token1] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5375),
    [sym_preproc_directive] = ACTIONS(5375),
    [anon_sym_DASH] = ACTIONS(5377),
    [anon_sym_PLUS] = ACTIONS(5377),
    [anon_sym___extension__] = ACTIONS(5375),
    [anon_sym_typedef] = ACTIONS(5375),
    [anon_sym_extern] = ACTIONS(5375),
    [anon_sym___attribute__] = ACTIONS(5375),
    [anon_sym___attribute] = ACTIONS(5375),
    [anon_sym_noreturn] = ACTIONS(5375),
    [anon_sym_LBRACK] = ACTIONS(5377),
    [anon_sym___declspec] = ACTIONS(5375),
    [anon_sym___cdecl] = ACTIONS(5375),
    [anon_sym___clrcall] = ACTIONS(5375),
    [anon_sym___stdcall] = ACTIONS(5375),
    [anon_sym___fastcall] = ACTIONS(5375),
    [anon_sym___thiscall] = ACTIONS(5375),
    [anon_sym___vectorcall] = ACTIONS(5375),
    [anon_sym_signed] = ACTIONS(5375),
    [anon_sym_unsigned] = ACTIONS(5375),
    [anon_sym_long] = ACTIONS(5375),
    [anon_sym_short] = ACTIONS(5375),
    [anon_sym_static] = ACTIONS(5375),
    [anon_sym_auto] = ACTIONS(5375),
    [anon_sym_register] = ACTIONS(5375),
    [anon_sym_inline] = ACTIONS(5375),
    [anon_sym___inline] = ACTIONS(5375),
    [anon_sym___inline__] = ACTIONS(5375),
    [anon_sym___forceinline] = ACTIONS(5375),
    [anon_sym_thread_local] = ACTIONS(5375),
    [anon_sym___thread] = ACTIONS(5375),
    [anon_sym_CG_EXTERN] = ACTIONS(5375),
    [anon_sym_CG_INLINE] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5375),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5375),
    [anon_sym_IBOutlet] = ACTIONS(5375),
    [anon_sym_IBInspectable] = ACTIONS(5375),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5375),
    [anon_sym_NS_INLINE] = ACTIONS(5375),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5375),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5375),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5375),
    [anon_sym_const] = ACTIONS(5375),
    [anon_sym_constexpr] = ACTIONS(5375),
    [anon_sym_volatile] = ACTIONS(5375),
    [anon_sym_restrict] = ACTIONS(5375),
    [anon_sym___restrict__] = ACTIONS(5375),
    [anon_sym__Atomic] = ACTIONS(5375),
    [anon_sym__Noreturn] = ACTIONS(5375),
    [anon_sym__Nonnull] = ACTIONS(5375),
    [anon_sym_nullable] = ACTIONS(5375),
    [anon_sym__Complex] = ACTIONS(5375),
    [anon_sym__Nullable] = ACTIONS(5375),
    [anon_sym__Nullable_result] = ACTIONS(5375),
    [anon_sym__Null_unspecified] = ACTIONS(5375),
    [anon_sym___autoreleasing] = ACTIONS(5375),
    [anon_sym___block] = ACTIONS(5375),
    [anon_sym___bridge] = ACTIONS(5375),
    [anon_sym___bridge_retained] = ACTIONS(5375),
    [anon_sym___bridge_transfer] = ACTIONS(5375),
    [anon_sym___complex] = ACTIONS(5375),
    [anon_sym___const] = ACTIONS(5375),
    [anon_sym___imag] = ACTIONS(5375),
    [anon_sym___kindof] = ACTIONS(5375),
    [anon_sym___nonnull] = ACTIONS(5375),
    [anon_sym___nullable] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5375),
    [anon_sym___real] = ACTIONS(5375),
    [anon_sym___strong] = ACTIONS(5375),
    [anon_sym___unsafe_unretained] = ACTIONS(5375),
    [anon_sym___unused] = ACTIONS(5375),
    [anon_sym___weak] = ACTIONS(5375),
    [anon_sym_alignas] = ACTIONS(5375),
    [anon_sym__Alignas] = ACTIONS(5375),
    [sym_primitive_type] = ACTIONS(5375),
    [anon_sym_enum] = ACTIONS(5375),
    [anon_sym_struct] = ACTIONS(5375),
    [anon_sym_union] = ACTIONS(5375),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5375),
    [anon_sym___typeof] = ACTIONS(5375),
    [anon_sym_typeof] = ACTIONS(5375),
    [aux_sym_preproc_undef_token1] = ACTIONS(5375),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5375),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5375),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5375),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_API_AVAILABLE] = ACTIONS(5375),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_API_DEPRECATED] = ACTIONS(5375),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym___deprecated_msg] = ACTIONS(5375),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5375),
    [anon_sym_ATend] = ACTIONS(5377),
    [anon_sym_ATsynthesize] = ACTIONS(5377),
    [anon_sym_ATdynamic] = ACTIONS(5377),
    [anon_sym_BOOL] = ACTIONS(5375),
    [anon_sym_IMP] = ACTIONS(5375),
    [anon_sym_SEL] = ACTIONS(5375),
    [anon_sym_Class] = ACTIONS(5375),
    [anon_sym_id] = ACTIONS(5375),
  },
  [STATE(2353)] = {
    [sym_identifier] = ACTIONS(5662),
    [anon_sym_COMMA] = ACTIONS(5664),
    [anon_sym_RPAREN] = ACTIONS(5664),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5664),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5664),
    [anon_sym_LPAREN2] = ACTIONS(5664),
    [anon_sym_STAR] = ACTIONS(5664),
    [anon_sym_CARET] = ACTIONS(5664),
    [anon_sym_GT] = ACTIONS(5664),
    [anon_sym_SEMI] = ACTIONS(5664),
    [anon_sym___extension__] = ACTIONS(5662),
    [anon_sym_extern] = ACTIONS(5662),
    [anon_sym___attribute__] = ACTIONS(5662),
    [anon_sym___attribute] = ACTIONS(5662),
    [anon_sym_noreturn] = ACTIONS(5662),
    [anon_sym_LBRACK] = ACTIONS(5664),
    [anon_sym___declspec] = ACTIONS(5662),
    [anon_sym___based] = ACTIONS(5662),
    [anon_sym_LBRACE] = ACTIONS(5664),
    [anon_sym_RBRACE] = ACTIONS(5664),
    [anon_sym_signed] = ACTIONS(5662),
    [anon_sym_unsigned] = ACTIONS(5662),
    [anon_sym_long] = ACTIONS(5662),
    [anon_sym_short] = ACTIONS(5662),
    [anon_sym_static] = ACTIONS(5662),
    [anon_sym_ATautoreleasepool] = ACTIONS(5664),
    [anon_sym_auto] = ACTIONS(5662),
    [anon_sym_register] = ACTIONS(5662),
    [anon_sym_inline] = ACTIONS(5662),
    [anon_sym___inline] = ACTIONS(5662),
    [anon_sym___inline__] = ACTIONS(5662),
    [anon_sym___forceinline] = ACTIONS(5662),
    [anon_sym_thread_local] = ACTIONS(5662),
    [anon_sym___thread] = ACTIONS(5662),
    [anon_sym_CG_EXTERN] = ACTIONS(5662),
    [anon_sym_CG_INLINE] = ACTIONS(5662),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5662),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5662),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5662),
    [anon_sym_IBOutlet] = ACTIONS(5662),
    [anon_sym_IBInspectable] = ACTIONS(5662),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5662),
    [anon_sym_NS_INLINE] = ACTIONS(5662),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5662),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5662),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5662),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5662),
    [anon_sym_const] = ACTIONS(5662),
    [anon_sym_constexpr] = ACTIONS(5662),
    [anon_sym_volatile] = ACTIONS(5662),
    [anon_sym_restrict] = ACTIONS(5662),
    [anon_sym___restrict__] = ACTIONS(5662),
    [anon_sym__Atomic] = ACTIONS(5662),
    [anon_sym__Noreturn] = ACTIONS(5662),
    [anon_sym__Nonnull] = ACTIONS(5662),
    [anon_sym_nullable] = ACTIONS(5662),
    [anon_sym__Complex] = ACTIONS(5662),
    [anon_sym__Nullable] = ACTIONS(5662),
    [anon_sym__Nullable_result] = ACTIONS(5662),
    [anon_sym__Null_unspecified] = ACTIONS(5662),
    [anon_sym___autoreleasing] = ACTIONS(5662),
    [anon_sym___block] = ACTIONS(5662),
    [anon_sym___bridge] = ACTIONS(5662),
    [anon_sym___bridge_retained] = ACTIONS(5662),
    [anon_sym___bridge_transfer] = ACTIONS(5662),
    [anon_sym___complex] = ACTIONS(5662),
    [anon_sym___const] = ACTIONS(5662),
    [anon_sym___imag] = ACTIONS(5662),
    [anon_sym___kindof] = ACTIONS(5662),
    [anon_sym___nonnull] = ACTIONS(5662),
    [anon_sym___nullable] = ACTIONS(5662),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5662),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5662),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5662),
    [anon_sym___real] = ACTIONS(5662),
    [anon_sym___strong] = ACTIONS(5662),
    [anon_sym___unsafe_unretained] = ACTIONS(5662),
    [anon_sym___unused] = ACTIONS(5662),
    [anon_sym___weak] = ACTIONS(5662),
    [anon_sym_alignas] = ACTIONS(5662),
    [anon_sym__Alignas] = ACTIONS(5662),
    [sym_primitive_type] = ACTIONS(5662),
    [anon_sym_enum] = ACTIONS(5662),
    [anon_sym_COLON] = ACTIONS(5664),
    [anon_sym_struct] = ACTIONS(5662),
    [anon_sym_union] = ACTIONS(5662),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5662),
    [anon_sym___typeof] = ACTIONS(5662),
    [anon_sym_typeof] = ACTIONS(5662),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5662),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5662),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5662),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5662),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5662),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5662),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5662),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5662),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5662),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5662),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5662),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5662),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5662),
    [anon_sym_API_AVAILABLE] = ACTIONS(5662),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5662),
    [anon_sym_API_DEPRECATED] = ACTIONS(5662),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5662),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5662),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5662),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5662),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5662),
    [anon_sym___deprecated_msg] = ACTIONS(5662),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5662),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5662),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5662),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5662),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5662),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5662),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5662),
    [anon_sym_ATprotocol] = ACTIONS(5664),
    [anon_sym_ATinterface] = ACTIONS(5664),
    [anon_sym_ATimplementation] = ACTIONS(5664),
    [anon_sym_ATproperty] = ACTIONS(5664),
    [anon_sym_BOOL] = ACTIONS(5662),
    [anon_sym_IMP] = ACTIONS(5662),
    [anon_sym_SEL] = ACTIONS(5662),
    [anon_sym_Class] = ACTIONS(5662),
    [anon_sym_id] = ACTIONS(5662),
  },
  [STATE(2354)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_COMMA] = ACTIONS(2920),
    [anon_sym_RPAREN] = ACTIONS(2920),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2920),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2920),
    [anon_sym_LPAREN2] = ACTIONS(2920),
    [anon_sym_STAR] = ACTIONS(2920),
    [anon_sym_CARET] = ACTIONS(2920),
    [anon_sym_GT] = ACTIONS(2920),
    [anon_sym_SEMI] = ACTIONS(2920),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym___declspec] = ACTIONS(2914),
    [anon_sym___based] = ACTIONS(2914),
    [anon_sym_LBRACE] = ACTIONS(2920),
    [anon_sym_RBRACE] = ACTIONS(2920),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_ATautoreleasepool] = ACTIONS(2920),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_COLON] = ACTIONS(2920),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATprotocol] = ACTIONS(2920),
    [anon_sym_ATinterface] = ACTIONS(2920),
    [anon_sym_ATimplementation] = ACTIONS(2920),
    [anon_sym_ATproperty] = ACTIONS(2920),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(2355)] = {
    [sym_identifier] = ACTIONS(2614),
    [aux_sym_preproc_def_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token2] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2614),
    [sym_preproc_directive] = ACTIONS(2614),
    [anon_sym_DASH] = ACTIONS(2616),
    [anon_sym_PLUS] = ACTIONS(2616),
    [anon_sym___extension__] = ACTIONS(2614),
    [anon_sym_typedef] = ACTIONS(2614),
    [anon_sym_extern] = ACTIONS(2614),
    [anon_sym___attribute__] = ACTIONS(2614),
    [anon_sym___attribute] = ACTIONS(2614),
    [anon_sym_noreturn] = ACTIONS(2614),
    [anon_sym_LBRACK] = ACTIONS(2616),
    [anon_sym___declspec] = ACTIONS(2614),
    [anon_sym___cdecl] = ACTIONS(2614),
    [anon_sym___clrcall] = ACTIONS(2614),
    [anon_sym___stdcall] = ACTIONS(2614),
    [anon_sym___fastcall] = ACTIONS(2614),
    [anon_sym___thiscall] = ACTIONS(2614),
    [anon_sym___vectorcall] = ACTIONS(2614),
    [anon_sym_signed] = ACTIONS(2614),
    [anon_sym_unsigned] = ACTIONS(2614),
    [anon_sym_long] = ACTIONS(2614),
    [anon_sym_short] = ACTIONS(2614),
    [anon_sym_static] = ACTIONS(2614),
    [anon_sym_auto] = ACTIONS(2614),
    [anon_sym_register] = ACTIONS(2614),
    [anon_sym_inline] = ACTIONS(2614),
    [anon_sym___inline] = ACTIONS(2614),
    [anon_sym___inline__] = ACTIONS(2614),
    [anon_sym___forceinline] = ACTIONS(2614),
    [anon_sym_thread_local] = ACTIONS(2614),
    [anon_sym___thread] = ACTIONS(2614),
    [anon_sym_CG_EXTERN] = ACTIONS(2614),
    [anon_sym_CG_INLINE] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2614),
    [anon_sym_IBOutlet] = ACTIONS(2614),
    [anon_sym_IBInspectable] = ACTIONS(2614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2614),
    [anon_sym_NS_INLINE] = ACTIONS(2614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2614),
    [anon_sym_const] = ACTIONS(2614),
    [anon_sym_constexpr] = ACTIONS(2614),
    [anon_sym_volatile] = ACTIONS(2614),
    [anon_sym_restrict] = ACTIONS(2614),
    [anon_sym___restrict__] = ACTIONS(2614),
    [anon_sym__Atomic] = ACTIONS(2614),
    [anon_sym__Noreturn] = ACTIONS(2614),
    [anon_sym__Nonnull] = ACTIONS(2614),
    [anon_sym_nullable] = ACTIONS(2614),
    [anon_sym__Complex] = ACTIONS(2614),
    [anon_sym__Nullable] = ACTIONS(2614),
    [anon_sym__Nullable_result] = ACTIONS(2614),
    [anon_sym__Null_unspecified] = ACTIONS(2614),
    [anon_sym___autoreleasing] = ACTIONS(2614),
    [anon_sym___block] = ACTIONS(2614),
    [anon_sym___bridge] = ACTIONS(2614),
    [anon_sym___bridge_retained] = ACTIONS(2614),
    [anon_sym___bridge_transfer] = ACTIONS(2614),
    [anon_sym___complex] = ACTIONS(2614),
    [anon_sym___const] = ACTIONS(2614),
    [anon_sym___imag] = ACTIONS(2614),
    [anon_sym___kindof] = ACTIONS(2614),
    [anon_sym___nonnull] = ACTIONS(2614),
    [anon_sym___nullable] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2614),
    [anon_sym___real] = ACTIONS(2614),
    [anon_sym___strong] = ACTIONS(2614),
    [anon_sym___unsafe_unretained] = ACTIONS(2614),
    [anon_sym___unused] = ACTIONS(2614),
    [anon_sym___weak] = ACTIONS(2614),
    [anon_sym_alignas] = ACTIONS(2614),
    [anon_sym__Alignas] = ACTIONS(2614),
    [sym_primitive_type] = ACTIONS(2614),
    [anon_sym_enum] = ACTIONS(2614),
    [anon_sym_struct] = ACTIONS(2614),
    [anon_sym_union] = ACTIONS(2614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2614),
    [anon_sym___typeof] = ACTIONS(2614),
    [anon_sym_typeof] = ACTIONS(2614),
    [aux_sym_preproc_undef_token1] = ACTIONS(2614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_API_AVAILABLE] = ACTIONS(2614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_API_DEPRECATED] = ACTIONS(2614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym___deprecated_msg] = ACTIONS(2614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2614),
    [anon_sym_ATsynthesize] = ACTIONS(2616),
    [anon_sym_ATdynamic] = ACTIONS(2616),
    [anon_sym_BOOL] = ACTIONS(2614),
    [anon_sym_IMP] = ACTIONS(2614),
    [anon_sym_SEL] = ACTIONS(2614),
    [anon_sym_Class] = ACTIONS(2614),
    [anon_sym_id] = ACTIONS(2614),
  },
  [STATE(2356)] = {
    [sym_identifier] = ACTIONS(5666),
    [anon_sym_COMMA] = ACTIONS(5668),
    [anon_sym_RPAREN] = ACTIONS(5668),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5668),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5668),
    [anon_sym_LPAREN2] = ACTIONS(5668),
    [anon_sym_STAR] = ACTIONS(5668),
    [anon_sym_CARET] = ACTIONS(5668),
    [anon_sym_GT] = ACTIONS(5668),
    [anon_sym_SEMI] = ACTIONS(5668),
    [anon_sym___extension__] = ACTIONS(5666),
    [anon_sym_extern] = ACTIONS(5666),
    [anon_sym___attribute__] = ACTIONS(5666),
    [anon_sym___attribute] = ACTIONS(5666),
    [anon_sym_noreturn] = ACTIONS(5666),
    [anon_sym_LBRACK] = ACTIONS(5668),
    [anon_sym___declspec] = ACTIONS(5666),
    [anon_sym___based] = ACTIONS(5666),
    [anon_sym_LBRACE] = ACTIONS(5668),
    [anon_sym_RBRACE] = ACTIONS(5668),
    [anon_sym_signed] = ACTIONS(5666),
    [anon_sym_unsigned] = ACTIONS(5666),
    [anon_sym_long] = ACTIONS(5666),
    [anon_sym_short] = ACTIONS(5666),
    [anon_sym_static] = ACTIONS(5666),
    [anon_sym_ATautoreleasepool] = ACTIONS(5668),
    [anon_sym_auto] = ACTIONS(5666),
    [anon_sym_register] = ACTIONS(5666),
    [anon_sym_inline] = ACTIONS(5666),
    [anon_sym___inline] = ACTIONS(5666),
    [anon_sym___inline__] = ACTIONS(5666),
    [anon_sym___forceinline] = ACTIONS(5666),
    [anon_sym_thread_local] = ACTIONS(5666),
    [anon_sym___thread] = ACTIONS(5666),
    [anon_sym_CG_EXTERN] = ACTIONS(5666),
    [anon_sym_CG_INLINE] = ACTIONS(5666),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5666),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5666),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5666),
    [anon_sym_IBOutlet] = ACTIONS(5666),
    [anon_sym_IBInspectable] = ACTIONS(5666),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5666),
    [anon_sym_NS_INLINE] = ACTIONS(5666),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5666),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5666),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5666),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5666),
    [anon_sym_const] = ACTIONS(5666),
    [anon_sym_constexpr] = ACTIONS(5666),
    [anon_sym_volatile] = ACTIONS(5666),
    [anon_sym_restrict] = ACTIONS(5666),
    [anon_sym___restrict__] = ACTIONS(5666),
    [anon_sym__Atomic] = ACTIONS(5666),
    [anon_sym__Noreturn] = ACTIONS(5666),
    [anon_sym__Nonnull] = ACTIONS(5666),
    [anon_sym_nullable] = ACTIONS(5666),
    [anon_sym__Complex] = ACTIONS(5666),
    [anon_sym__Nullable] = ACTIONS(5666),
    [anon_sym__Nullable_result] = ACTIONS(5666),
    [anon_sym__Null_unspecified] = ACTIONS(5666),
    [anon_sym___autoreleasing] = ACTIONS(5666),
    [anon_sym___block] = ACTIONS(5666),
    [anon_sym___bridge] = ACTIONS(5666),
    [anon_sym___bridge_retained] = ACTIONS(5666),
    [anon_sym___bridge_transfer] = ACTIONS(5666),
    [anon_sym___complex] = ACTIONS(5666),
    [anon_sym___const] = ACTIONS(5666),
    [anon_sym___imag] = ACTIONS(5666),
    [anon_sym___kindof] = ACTIONS(5666),
    [anon_sym___nonnull] = ACTIONS(5666),
    [anon_sym___nullable] = ACTIONS(5666),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5666),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5666),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5666),
    [anon_sym___real] = ACTIONS(5666),
    [anon_sym___strong] = ACTIONS(5666),
    [anon_sym___unsafe_unretained] = ACTIONS(5666),
    [anon_sym___unused] = ACTIONS(5666),
    [anon_sym___weak] = ACTIONS(5666),
    [anon_sym_alignas] = ACTIONS(5666),
    [anon_sym__Alignas] = ACTIONS(5666),
    [sym_primitive_type] = ACTIONS(5666),
    [anon_sym_enum] = ACTIONS(5666),
    [anon_sym_COLON] = ACTIONS(5668),
    [anon_sym_struct] = ACTIONS(5666),
    [anon_sym_union] = ACTIONS(5666),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5666),
    [anon_sym___typeof] = ACTIONS(5666),
    [anon_sym_typeof] = ACTIONS(5666),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5666),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5666),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5666),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5666),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5666),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5666),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5666),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5666),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5666),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5666),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5666),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5666),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5666),
    [anon_sym_API_AVAILABLE] = ACTIONS(5666),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5666),
    [anon_sym_API_DEPRECATED] = ACTIONS(5666),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5666),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5666),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5666),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5666),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5666),
    [anon_sym___deprecated_msg] = ACTIONS(5666),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5666),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5666),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5666),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5666),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5666),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5666),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5666),
    [anon_sym_ATprotocol] = ACTIONS(5668),
    [anon_sym_ATinterface] = ACTIONS(5668),
    [anon_sym_ATimplementation] = ACTIONS(5668),
    [anon_sym_ATproperty] = ACTIONS(5668),
    [anon_sym_BOOL] = ACTIONS(5666),
    [anon_sym_IMP] = ACTIONS(5666),
    [anon_sym_SEL] = ACTIONS(5666),
    [anon_sym_Class] = ACTIONS(5666),
    [anon_sym_id] = ACTIONS(5666),
  },
  [STATE(2357)] = {
    [sym_identifier] = ACTIONS(5670),
    [anon_sym_COMMA] = ACTIONS(5672),
    [anon_sym_RPAREN] = ACTIONS(5672),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5672),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5672),
    [anon_sym_LPAREN2] = ACTIONS(5672),
    [anon_sym_STAR] = ACTIONS(5672),
    [anon_sym_CARET] = ACTIONS(5672),
    [anon_sym_GT] = ACTIONS(5672),
    [anon_sym_SEMI] = ACTIONS(5672),
    [anon_sym___extension__] = ACTIONS(5670),
    [anon_sym_extern] = ACTIONS(5670),
    [anon_sym___attribute__] = ACTIONS(5670),
    [anon_sym___attribute] = ACTIONS(5670),
    [anon_sym_noreturn] = ACTIONS(5670),
    [anon_sym_LBRACK] = ACTIONS(5672),
    [anon_sym___declspec] = ACTIONS(5670),
    [anon_sym___based] = ACTIONS(5670),
    [anon_sym_LBRACE] = ACTIONS(5672),
    [anon_sym_RBRACE] = ACTIONS(5672),
    [anon_sym_signed] = ACTIONS(5670),
    [anon_sym_unsigned] = ACTIONS(5670),
    [anon_sym_long] = ACTIONS(5670),
    [anon_sym_short] = ACTIONS(5670),
    [anon_sym_static] = ACTIONS(5670),
    [anon_sym_ATautoreleasepool] = ACTIONS(5672),
    [anon_sym_auto] = ACTIONS(5670),
    [anon_sym_register] = ACTIONS(5670),
    [anon_sym_inline] = ACTIONS(5670),
    [anon_sym___inline] = ACTIONS(5670),
    [anon_sym___inline__] = ACTIONS(5670),
    [anon_sym___forceinline] = ACTIONS(5670),
    [anon_sym_thread_local] = ACTIONS(5670),
    [anon_sym___thread] = ACTIONS(5670),
    [anon_sym_CG_EXTERN] = ACTIONS(5670),
    [anon_sym_CG_INLINE] = ACTIONS(5670),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5670),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5670),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5670),
    [anon_sym_IBOutlet] = ACTIONS(5670),
    [anon_sym_IBInspectable] = ACTIONS(5670),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5670),
    [anon_sym_NS_INLINE] = ACTIONS(5670),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5670),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5670),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5670),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5670),
    [anon_sym_const] = ACTIONS(5670),
    [anon_sym_constexpr] = ACTIONS(5670),
    [anon_sym_volatile] = ACTIONS(5670),
    [anon_sym_restrict] = ACTIONS(5670),
    [anon_sym___restrict__] = ACTIONS(5670),
    [anon_sym__Atomic] = ACTIONS(5670),
    [anon_sym__Noreturn] = ACTIONS(5670),
    [anon_sym__Nonnull] = ACTIONS(5670),
    [anon_sym_nullable] = ACTIONS(5670),
    [anon_sym__Complex] = ACTIONS(5670),
    [anon_sym__Nullable] = ACTIONS(5670),
    [anon_sym__Nullable_result] = ACTIONS(5670),
    [anon_sym__Null_unspecified] = ACTIONS(5670),
    [anon_sym___autoreleasing] = ACTIONS(5670),
    [anon_sym___block] = ACTIONS(5670),
    [anon_sym___bridge] = ACTIONS(5670),
    [anon_sym___bridge_retained] = ACTIONS(5670),
    [anon_sym___bridge_transfer] = ACTIONS(5670),
    [anon_sym___complex] = ACTIONS(5670),
    [anon_sym___const] = ACTIONS(5670),
    [anon_sym___imag] = ACTIONS(5670),
    [anon_sym___kindof] = ACTIONS(5670),
    [anon_sym___nonnull] = ACTIONS(5670),
    [anon_sym___nullable] = ACTIONS(5670),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5670),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5670),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5670),
    [anon_sym___real] = ACTIONS(5670),
    [anon_sym___strong] = ACTIONS(5670),
    [anon_sym___unsafe_unretained] = ACTIONS(5670),
    [anon_sym___unused] = ACTIONS(5670),
    [anon_sym___weak] = ACTIONS(5670),
    [anon_sym_alignas] = ACTIONS(5670),
    [anon_sym__Alignas] = ACTIONS(5670),
    [sym_primitive_type] = ACTIONS(5670),
    [anon_sym_enum] = ACTIONS(5670),
    [anon_sym_COLON] = ACTIONS(5672),
    [anon_sym_struct] = ACTIONS(5670),
    [anon_sym_union] = ACTIONS(5670),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5670),
    [anon_sym___typeof] = ACTIONS(5670),
    [anon_sym_typeof] = ACTIONS(5670),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5670),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5670),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5670),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5670),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5670),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5670),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5670),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5670),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5670),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5670),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5670),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5670),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5670),
    [anon_sym_API_AVAILABLE] = ACTIONS(5670),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5670),
    [anon_sym_API_DEPRECATED] = ACTIONS(5670),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5670),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5670),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5670),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5670),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5670),
    [anon_sym___deprecated_msg] = ACTIONS(5670),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5670),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5670),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5670),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5670),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5670),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5670),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5670),
    [anon_sym_ATprotocol] = ACTIONS(5672),
    [anon_sym_ATinterface] = ACTIONS(5672),
    [anon_sym_ATimplementation] = ACTIONS(5672),
    [anon_sym_ATproperty] = ACTIONS(5672),
    [anon_sym_BOOL] = ACTIONS(5670),
    [anon_sym_IMP] = ACTIONS(5670),
    [anon_sym_SEL] = ACTIONS(5670),
    [anon_sym_Class] = ACTIONS(5670),
    [anon_sym_id] = ACTIONS(5670),
  },
  [STATE(2358)] = {
    [sym_identifier] = ACTIONS(2958),
    [anon_sym_COMMA] = ACTIONS(2960),
    [anon_sym_RPAREN] = ACTIONS(2960),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2960),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2960),
    [anon_sym_LPAREN2] = ACTIONS(2960),
    [anon_sym_STAR] = ACTIONS(2960),
    [anon_sym_CARET] = ACTIONS(2960),
    [anon_sym_GT] = ACTIONS(2960),
    [anon_sym_SEMI] = ACTIONS(2960),
    [anon_sym___extension__] = ACTIONS(2958),
    [anon_sym_extern] = ACTIONS(2958),
    [anon_sym___attribute__] = ACTIONS(2958),
    [anon_sym___attribute] = ACTIONS(2958),
    [anon_sym_noreturn] = ACTIONS(2958),
    [anon_sym_LBRACK] = ACTIONS(2960),
    [anon_sym___declspec] = ACTIONS(2958),
    [anon_sym___based] = ACTIONS(2958),
    [anon_sym_LBRACE] = ACTIONS(2960),
    [anon_sym_RBRACE] = ACTIONS(2960),
    [anon_sym_signed] = ACTIONS(2958),
    [anon_sym_unsigned] = ACTIONS(2958),
    [anon_sym_long] = ACTIONS(2958),
    [anon_sym_short] = ACTIONS(2958),
    [anon_sym_static] = ACTIONS(2958),
    [anon_sym_ATautoreleasepool] = ACTIONS(2960),
    [anon_sym_auto] = ACTIONS(2958),
    [anon_sym_register] = ACTIONS(2958),
    [anon_sym_inline] = ACTIONS(2958),
    [anon_sym___inline] = ACTIONS(2958),
    [anon_sym___inline__] = ACTIONS(2958),
    [anon_sym___forceinline] = ACTIONS(2958),
    [anon_sym_thread_local] = ACTIONS(2958),
    [anon_sym___thread] = ACTIONS(2958),
    [anon_sym_CG_EXTERN] = ACTIONS(2958),
    [anon_sym_CG_INLINE] = ACTIONS(2958),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2958),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2958),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2958),
    [anon_sym_IBOutlet] = ACTIONS(2958),
    [anon_sym_IBInspectable] = ACTIONS(2958),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2958),
    [anon_sym_NS_INLINE] = ACTIONS(2958),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2958),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2958),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2958),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2958),
    [anon_sym_const] = ACTIONS(2958),
    [anon_sym_constexpr] = ACTIONS(2958),
    [anon_sym_volatile] = ACTIONS(2958),
    [anon_sym_restrict] = ACTIONS(2958),
    [anon_sym___restrict__] = ACTIONS(2958),
    [anon_sym__Atomic] = ACTIONS(2958),
    [anon_sym__Noreturn] = ACTIONS(2958),
    [anon_sym__Nonnull] = ACTIONS(2958),
    [anon_sym_nullable] = ACTIONS(2958),
    [anon_sym__Complex] = ACTIONS(2958),
    [anon_sym__Nullable] = ACTIONS(2958),
    [anon_sym__Nullable_result] = ACTIONS(2958),
    [anon_sym__Null_unspecified] = ACTIONS(2958),
    [anon_sym___autoreleasing] = ACTIONS(2958),
    [anon_sym___block] = ACTIONS(2958),
    [anon_sym___bridge] = ACTIONS(2958),
    [anon_sym___bridge_retained] = ACTIONS(2958),
    [anon_sym___bridge_transfer] = ACTIONS(2958),
    [anon_sym___complex] = ACTIONS(2958),
    [anon_sym___const] = ACTIONS(2958),
    [anon_sym___imag] = ACTIONS(2958),
    [anon_sym___kindof] = ACTIONS(2958),
    [anon_sym___nonnull] = ACTIONS(2958),
    [anon_sym___nullable] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2958),
    [anon_sym___real] = ACTIONS(2958),
    [anon_sym___strong] = ACTIONS(2958),
    [anon_sym___unsafe_unretained] = ACTIONS(2958),
    [anon_sym___unused] = ACTIONS(2958),
    [anon_sym___weak] = ACTIONS(2958),
    [anon_sym_alignas] = ACTIONS(2958),
    [anon_sym__Alignas] = ACTIONS(2958),
    [sym_primitive_type] = ACTIONS(2958),
    [anon_sym_enum] = ACTIONS(2958),
    [anon_sym_COLON] = ACTIONS(2960),
    [anon_sym_struct] = ACTIONS(2958),
    [anon_sym_union] = ACTIONS(2958),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2958),
    [anon_sym___typeof] = ACTIONS(2958),
    [anon_sym_typeof] = ACTIONS(2958),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2958),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2958),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2958),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2958),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2958),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2958),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2958),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2958),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2958),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_API_AVAILABLE] = ACTIONS(2958),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_API_DEPRECATED] = ACTIONS(2958),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2958),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2958),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2958),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2958),
    [anon_sym___deprecated_msg] = ACTIONS(2958),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2958),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2958),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2958),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2958),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2958),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2958),
    [anon_sym_ATprotocol] = ACTIONS(2960),
    [anon_sym_ATinterface] = ACTIONS(2960),
    [anon_sym_ATimplementation] = ACTIONS(2960),
    [anon_sym_ATproperty] = ACTIONS(2960),
    [anon_sym_BOOL] = ACTIONS(2958),
    [anon_sym_IMP] = ACTIONS(2958),
    [anon_sym_SEL] = ACTIONS(2958),
    [anon_sym_Class] = ACTIONS(2958),
    [anon_sym_id] = ACTIONS(2958),
  },
  [STATE(2359)] = {
    [sym_identifier] = ACTIONS(5381),
    [aux_sym_preproc_def_token1] = ACTIONS(5381),
    [aux_sym_preproc_if_token1] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5381),
    [sym_preproc_directive] = ACTIONS(5381),
    [anon_sym_DASH] = ACTIONS(5383),
    [anon_sym_PLUS] = ACTIONS(5383),
    [anon_sym___extension__] = ACTIONS(5381),
    [anon_sym_typedef] = ACTIONS(5381),
    [anon_sym_extern] = ACTIONS(5381),
    [anon_sym___attribute__] = ACTIONS(5381),
    [anon_sym___attribute] = ACTIONS(5381),
    [anon_sym_noreturn] = ACTIONS(5381),
    [anon_sym_LBRACK] = ACTIONS(5383),
    [anon_sym___declspec] = ACTIONS(5381),
    [anon_sym___cdecl] = ACTIONS(5381),
    [anon_sym___clrcall] = ACTIONS(5381),
    [anon_sym___stdcall] = ACTIONS(5381),
    [anon_sym___fastcall] = ACTIONS(5381),
    [anon_sym___thiscall] = ACTIONS(5381),
    [anon_sym___vectorcall] = ACTIONS(5381),
    [anon_sym_signed] = ACTIONS(5381),
    [anon_sym_unsigned] = ACTIONS(5381),
    [anon_sym_long] = ACTIONS(5381),
    [anon_sym_short] = ACTIONS(5381),
    [anon_sym_static] = ACTIONS(5381),
    [anon_sym_auto] = ACTIONS(5381),
    [anon_sym_register] = ACTIONS(5381),
    [anon_sym_inline] = ACTIONS(5381),
    [anon_sym___inline] = ACTIONS(5381),
    [anon_sym___inline__] = ACTIONS(5381),
    [anon_sym___forceinline] = ACTIONS(5381),
    [anon_sym_thread_local] = ACTIONS(5381),
    [anon_sym___thread] = ACTIONS(5381),
    [anon_sym_CG_EXTERN] = ACTIONS(5381),
    [anon_sym_CG_INLINE] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5381),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5381),
    [anon_sym_IBOutlet] = ACTIONS(5381),
    [anon_sym_IBInspectable] = ACTIONS(5381),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5381),
    [anon_sym_NS_INLINE] = ACTIONS(5381),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5381),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5381),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5381),
    [anon_sym_const] = ACTIONS(5381),
    [anon_sym_constexpr] = ACTIONS(5381),
    [anon_sym_volatile] = ACTIONS(5381),
    [anon_sym_restrict] = ACTIONS(5381),
    [anon_sym___restrict__] = ACTIONS(5381),
    [anon_sym__Atomic] = ACTIONS(5381),
    [anon_sym__Noreturn] = ACTIONS(5381),
    [anon_sym__Nonnull] = ACTIONS(5381),
    [anon_sym_nullable] = ACTIONS(5381),
    [anon_sym__Complex] = ACTIONS(5381),
    [anon_sym__Nullable] = ACTIONS(5381),
    [anon_sym__Nullable_result] = ACTIONS(5381),
    [anon_sym__Null_unspecified] = ACTIONS(5381),
    [anon_sym___autoreleasing] = ACTIONS(5381),
    [anon_sym___block] = ACTIONS(5381),
    [anon_sym___bridge] = ACTIONS(5381),
    [anon_sym___bridge_retained] = ACTIONS(5381),
    [anon_sym___bridge_transfer] = ACTIONS(5381),
    [anon_sym___complex] = ACTIONS(5381),
    [anon_sym___const] = ACTIONS(5381),
    [anon_sym___imag] = ACTIONS(5381),
    [anon_sym___kindof] = ACTIONS(5381),
    [anon_sym___nonnull] = ACTIONS(5381),
    [anon_sym___nullable] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5381),
    [anon_sym___real] = ACTIONS(5381),
    [anon_sym___strong] = ACTIONS(5381),
    [anon_sym___unsafe_unretained] = ACTIONS(5381),
    [anon_sym___unused] = ACTIONS(5381),
    [anon_sym___weak] = ACTIONS(5381),
    [anon_sym_alignas] = ACTIONS(5381),
    [anon_sym__Alignas] = ACTIONS(5381),
    [sym_primitive_type] = ACTIONS(5381),
    [anon_sym_enum] = ACTIONS(5381),
    [anon_sym_struct] = ACTIONS(5381),
    [anon_sym_union] = ACTIONS(5381),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5381),
    [anon_sym___typeof] = ACTIONS(5381),
    [anon_sym_typeof] = ACTIONS(5381),
    [aux_sym_preproc_undef_token1] = ACTIONS(5381),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5381),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5381),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5381),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_API_AVAILABLE] = ACTIONS(5381),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_API_DEPRECATED] = ACTIONS(5381),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym___deprecated_msg] = ACTIONS(5381),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5381),
    [anon_sym_ATend] = ACTIONS(5383),
    [anon_sym_ATsynthesize] = ACTIONS(5383),
    [anon_sym_ATdynamic] = ACTIONS(5383),
    [anon_sym_BOOL] = ACTIONS(5381),
    [anon_sym_IMP] = ACTIONS(5381),
    [anon_sym_SEL] = ACTIONS(5381),
    [anon_sym_Class] = ACTIONS(5381),
    [anon_sym_id] = ACTIONS(5381),
  },
  [STATE(2360)] = {
    [sym_identifier] = ACTIONS(5351),
    [aux_sym_preproc_def_token1] = ACTIONS(5351),
    [aux_sym_preproc_if_token1] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5351),
    [sym_preproc_directive] = ACTIONS(5351),
    [anon_sym_DASH] = ACTIONS(5353),
    [anon_sym_PLUS] = ACTIONS(5353),
    [anon_sym___extension__] = ACTIONS(5351),
    [anon_sym_typedef] = ACTIONS(5351),
    [anon_sym_extern] = ACTIONS(5351),
    [anon_sym___attribute__] = ACTIONS(5351),
    [anon_sym___attribute] = ACTIONS(5351),
    [anon_sym_noreturn] = ACTIONS(5351),
    [anon_sym_LBRACK] = ACTIONS(5353),
    [anon_sym___declspec] = ACTIONS(5351),
    [anon_sym___cdecl] = ACTIONS(5351),
    [anon_sym___clrcall] = ACTIONS(5351),
    [anon_sym___stdcall] = ACTIONS(5351),
    [anon_sym___fastcall] = ACTIONS(5351),
    [anon_sym___thiscall] = ACTIONS(5351),
    [anon_sym___vectorcall] = ACTIONS(5351),
    [anon_sym_signed] = ACTIONS(5351),
    [anon_sym_unsigned] = ACTIONS(5351),
    [anon_sym_long] = ACTIONS(5351),
    [anon_sym_short] = ACTIONS(5351),
    [anon_sym_static] = ACTIONS(5351),
    [anon_sym_auto] = ACTIONS(5351),
    [anon_sym_register] = ACTIONS(5351),
    [anon_sym_inline] = ACTIONS(5351),
    [anon_sym___inline] = ACTIONS(5351),
    [anon_sym___inline__] = ACTIONS(5351),
    [anon_sym___forceinline] = ACTIONS(5351),
    [anon_sym_thread_local] = ACTIONS(5351),
    [anon_sym___thread] = ACTIONS(5351),
    [anon_sym_CG_EXTERN] = ACTIONS(5351),
    [anon_sym_CG_INLINE] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5351),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5351),
    [anon_sym_IBOutlet] = ACTIONS(5351),
    [anon_sym_IBInspectable] = ACTIONS(5351),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5351),
    [anon_sym_NS_INLINE] = ACTIONS(5351),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5351),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5351),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5351),
    [anon_sym_const] = ACTIONS(5351),
    [anon_sym_constexpr] = ACTIONS(5351),
    [anon_sym_volatile] = ACTIONS(5351),
    [anon_sym_restrict] = ACTIONS(5351),
    [anon_sym___restrict__] = ACTIONS(5351),
    [anon_sym__Atomic] = ACTIONS(5351),
    [anon_sym__Noreturn] = ACTIONS(5351),
    [anon_sym__Nonnull] = ACTIONS(5351),
    [anon_sym_nullable] = ACTIONS(5351),
    [anon_sym__Complex] = ACTIONS(5351),
    [anon_sym__Nullable] = ACTIONS(5351),
    [anon_sym__Nullable_result] = ACTIONS(5351),
    [anon_sym__Null_unspecified] = ACTIONS(5351),
    [anon_sym___autoreleasing] = ACTIONS(5351),
    [anon_sym___block] = ACTIONS(5351),
    [anon_sym___bridge] = ACTIONS(5351),
    [anon_sym___bridge_retained] = ACTIONS(5351),
    [anon_sym___bridge_transfer] = ACTIONS(5351),
    [anon_sym___complex] = ACTIONS(5351),
    [anon_sym___const] = ACTIONS(5351),
    [anon_sym___imag] = ACTIONS(5351),
    [anon_sym___kindof] = ACTIONS(5351),
    [anon_sym___nonnull] = ACTIONS(5351),
    [anon_sym___nullable] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5351),
    [anon_sym___real] = ACTIONS(5351),
    [anon_sym___strong] = ACTIONS(5351),
    [anon_sym___unsafe_unretained] = ACTIONS(5351),
    [anon_sym___unused] = ACTIONS(5351),
    [anon_sym___weak] = ACTIONS(5351),
    [anon_sym_alignas] = ACTIONS(5351),
    [anon_sym__Alignas] = ACTIONS(5351),
    [sym_primitive_type] = ACTIONS(5351),
    [anon_sym_enum] = ACTIONS(5351),
    [anon_sym_struct] = ACTIONS(5351),
    [anon_sym_union] = ACTIONS(5351),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5351),
    [anon_sym___typeof] = ACTIONS(5351),
    [anon_sym_typeof] = ACTIONS(5351),
    [aux_sym_preproc_undef_token1] = ACTIONS(5351),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5351),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5351),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5351),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_API_AVAILABLE] = ACTIONS(5351),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_API_DEPRECATED] = ACTIONS(5351),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym___deprecated_msg] = ACTIONS(5351),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5351),
    [anon_sym_ATend] = ACTIONS(5353),
    [anon_sym_ATsynthesize] = ACTIONS(5353),
    [anon_sym_ATdynamic] = ACTIONS(5353),
    [anon_sym_BOOL] = ACTIONS(5351),
    [anon_sym_IMP] = ACTIONS(5351),
    [anon_sym_SEL] = ACTIONS(5351),
    [anon_sym_Class] = ACTIONS(5351),
    [anon_sym_id] = ACTIONS(5351),
  },
  [STATE(2361)] = {
    [sym_identifier] = ACTIONS(5405),
    [aux_sym_preproc_def_token1] = ACTIONS(5405),
    [aux_sym_preproc_if_token1] = ACTIONS(5405),
    [aux_sym_preproc_if_token2] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5405),
    [sym_preproc_directive] = ACTIONS(5405),
    [anon_sym_DASH] = ACTIONS(5407),
    [anon_sym_PLUS] = ACTIONS(5407),
    [anon_sym___extension__] = ACTIONS(5405),
    [anon_sym_typedef] = ACTIONS(5405),
    [anon_sym_extern] = ACTIONS(5405),
    [anon_sym___attribute__] = ACTIONS(5405),
    [anon_sym___attribute] = ACTIONS(5405),
    [anon_sym_noreturn] = ACTIONS(5405),
    [anon_sym_LBRACK] = ACTIONS(5407),
    [anon_sym___declspec] = ACTIONS(5405),
    [anon_sym___cdecl] = ACTIONS(5405),
    [anon_sym___clrcall] = ACTIONS(5405),
    [anon_sym___stdcall] = ACTIONS(5405),
    [anon_sym___fastcall] = ACTIONS(5405),
    [anon_sym___thiscall] = ACTIONS(5405),
    [anon_sym___vectorcall] = ACTIONS(5405),
    [anon_sym_signed] = ACTIONS(5405),
    [anon_sym_unsigned] = ACTIONS(5405),
    [anon_sym_long] = ACTIONS(5405),
    [anon_sym_short] = ACTIONS(5405),
    [anon_sym_static] = ACTIONS(5405),
    [anon_sym_auto] = ACTIONS(5405),
    [anon_sym_register] = ACTIONS(5405),
    [anon_sym_inline] = ACTIONS(5405),
    [anon_sym___inline] = ACTIONS(5405),
    [anon_sym___inline__] = ACTIONS(5405),
    [anon_sym___forceinline] = ACTIONS(5405),
    [anon_sym_thread_local] = ACTIONS(5405),
    [anon_sym___thread] = ACTIONS(5405),
    [anon_sym_CG_EXTERN] = ACTIONS(5405),
    [anon_sym_CG_INLINE] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5405),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5405),
    [anon_sym_IBOutlet] = ACTIONS(5405),
    [anon_sym_IBInspectable] = ACTIONS(5405),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5405),
    [anon_sym_NS_INLINE] = ACTIONS(5405),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5405),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5405),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5405),
    [anon_sym_const] = ACTIONS(5405),
    [anon_sym_constexpr] = ACTIONS(5405),
    [anon_sym_volatile] = ACTIONS(5405),
    [anon_sym_restrict] = ACTIONS(5405),
    [anon_sym___restrict__] = ACTIONS(5405),
    [anon_sym__Atomic] = ACTIONS(5405),
    [anon_sym__Noreturn] = ACTIONS(5405),
    [anon_sym__Nonnull] = ACTIONS(5405),
    [anon_sym_nullable] = ACTIONS(5405),
    [anon_sym__Complex] = ACTIONS(5405),
    [anon_sym__Nullable] = ACTIONS(5405),
    [anon_sym__Nullable_result] = ACTIONS(5405),
    [anon_sym__Null_unspecified] = ACTIONS(5405),
    [anon_sym___autoreleasing] = ACTIONS(5405),
    [anon_sym___block] = ACTIONS(5405),
    [anon_sym___bridge] = ACTIONS(5405),
    [anon_sym___bridge_retained] = ACTIONS(5405),
    [anon_sym___bridge_transfer] = ACTIONS(5405),
    [anon_sym___complex] = ACTIONS(5405),
    [anon_sym___const] = ACTIONS(5405),
    [anon_sym___imag] = ACTIONS(5405),
    [anon_sym___kindof] = ACTIONS(5405),
    [anon_sym___nonnull] = ACTIONS(5405),
    [anon_sym___nullable] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5405),
    [anon_sym___real] = ACTIONS(5405),
    [anon_sym___strong] = ACTIONS(5405),
    [anon_sym___unsafe_unretained] = ACTIONS(5405),
    [anon_sym___unused] = ACTIONS(5405),
    [anon_sym___weak] = ACTIONS(5405),
    [anon_sym_alignas] = ACTIONS(5405),
    [anon_sym__Alignas] = ACTIONS(5405),
    [sym_primitive_type] = ACTIONS(5405),
    [anon_sym_enum] = ACTIONS(5405),
    [anon_sym_struct] = ACTIONS(5405),
    [anon_sym_union] = ACTIONS(5405),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5405),
    [anon_sym___typeof] = ACTIONS(5405),
    [anon_sym_typeof] = ACTIONS(5405),
    [aux_sym_preproc_undef_token1] = ACTIONS(5405),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5405),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5405),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5405),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_API_AVAILABLE] = ACTIONS(5405),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_API_DEPRECATED] = ACTIONS(5405),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym___deprecated_msg] = ACTIONS(5405),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5405),
    [anon_sym_ATsynthesize] = ACTIONS(5407),
    [anon_sym_ATdynamic] = ACTIONS(5407),
    [anon_sym_BOOL] = ACTIONS(5405),
    [anon_sym_IMP] = ACTIONS(5405),
    [anon_sym_SEL] = ACTIONS(5405),
    [anon_sym_Class] = ACTIONS(5405),
    [anon_sym_id] = ACTIONS(5405),
  },
  [STATE(2362)] = {
    [sym_identifier] = ACTIONS(5505),
    [aux_sym_preproc_def_token1] = ACTIONS(5505),
    [aux_sym_preproc_if_token1] = ACTIONS(5505),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5505),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5505),
    [sym_preproc_directive] = ACTIONS(5505),
    [anon_sym_DASH] = ACTIONS(5507),
    [anon_sym_PLUS] = ACTIONS(5507),
    [anon_sym___extension__] = ACTIONS(5505),
    [anon_sym_typedef] = ACTIONS(5505),
    [anon_sym_extern] = ACTIONS(5505),
    [anon_sym___attribute__] = ACTIONS(5505),
    [anon_sym___attribute] = ACTIONS(5505),
    [anon_sym_noreturn] = ACTIONS(5505),
    [anon_sym_LBRACK] = ACTIONS(5507),
    [anon_sym___declspec] = ACTIONS(5505),
    [anon_sym___cdecl] = ACTIONS(5505),
    [anon_sym___clrcall] = ACTIONS(5505),
    [anon_sym___stdcall] = ACTIONS(5505),
    [anon_sym___fastcall] = ACTIONS(5505),
    [anon_sym___thiscall] = ACTIONS(5505),
    [anon_sym___vectorcall] = ACTIONS(5505),
    [anon_sym_signed] = ACTIONS(5505),
    [anon_sym_unsigned] = ACTIONS(5505),
    [anon_sym_long] = ACTIONS(5505),
    [anon_sym_short] = ACTIONS(5505),
    [anon_sym_static] = ACTIONS(5505),
    [anon_sym_auto] = ACTIONS(5505),
    [anon_sym_register] = ACTIONS(5505),
    [anon_sym_inline] = ACTIONS(5505),
    [anon_sym___inline] = ACTIONS(5505),
    [anon_sym___inline__] = ACTIONS(5505),
    [anon_sym___forceinline] = ACTIONS(5505),
    [anon_sym_thread_local] = ACTIONS(5505),
    [anon_sym___thread] = ACTIONS(5505),
    [anon_sym_CG_EXTERN] = ACTIONS(5505),
    [anon_sym_CG_INLINE] = ACTIONS(5505),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5505),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5505),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5505),
    [anon_sym_IBOutlet] = ACTIONS(5505),
    [anon_sym_IBInspectable] = ACTIONS(5505),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5505),
    [anon_sym_NS_INLINE] = ACTIONS(5505),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5505),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5505),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5505),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5505),
    [anon_sym_const] = ACTIONS(5505),
    [anon_sym_constexpr] = ACTIONS(5505),
    [anon_sym_volatile] = ACTIONS(5505),
    [anon_sym_restrict] = ACTIONS(5505),
    [anon_sym___restrict__] = ACTIONS(5505),
    [anon_sym__Atomic] = ACTIONS(5505),
    [anon_sym__Noreturn] = ACTIONS(5505),
    [anon_sym__Nonnull] = ACTIONS(5505),
    [anon_sym_nullable] = ACTIONS(5505),
    [anon_sym__Complex] = ACTIONS(5505),
    [anon_sym__Nullable] = ACTIONS(5505),
    [anon_sym__Nullable_result] = ACTIONS(5505),
    [anon_sym__Null_unspecified] = ACTIONS(5505),
    [anon_sym___autoreleasing] = ACTIONS(5505),
    [anon_sym___block] = ACTIONS(5505),
    [anon_sym___bridge] = ACTIONS(5505),
    [anon_sym___bridge_retained] = ACTIONS(5505),
    [anon_sym___bridge_transfer] = ACTIONS(5505),
    [anon_sym___complex] = ACTIONS(5505),
    [anon_sym___const] = ACTIONS(5505),
    [anon_sym___imag] = ACTIONS(5505),
    [anon_sym___kindof] = ACTIONS(5505),
    [anon_sym___nonnull] = ACTIONS(5505),
    [anon_sym___nullable] = ACTIONS(5505),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5505),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5505),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5505),
    [anon_sym___real] = ACTIONS(5505),
    [anon_sym___strong] = ACTIONS(5505),
    [anon_sym___unsafe_unretained] = ACTIONS(5505),
    [anon_sym___unused] = ACTIONS(5505),
    [anon_sym___weak] = ACTIONS(5505),
    [anon_sym_alignas] = ACTIONS(5505),
    [anon_sym__Alignas] = ACTIONS(5505),
    [sym_primitive_type] = ACTIONS(5505),
    [anon_sym_enum] = ACTIONS(5505),
    [anon_sym_struct] = ACTIONS(5505),
    [anon_sym_union] = ACTIONS(5505),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5505),
    [anon_sym___typeof] = ACTIONS(5505),
    [anon_sym_typeof] = ACTIONS(5505),
    [aux_sym_preproc_undef_token1] = ACTIONS(5505),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5505),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5505),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5505),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5505),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5505),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5505),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5505),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5505),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5505),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_API_AVAILABLE] = ACTIONS(5505),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_API_DEPRECATED] = ACTIONS(5505),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5505),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5505),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5505),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5505),
    [anon_sym___deprecated_msg] = ACTIONS(5505),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5505),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5505),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5505),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5505),
    [anon_sym_ATend] = ACTIONS(5507),
    [anon_sym_ATsynthesize] = ACTIONS(5507),
    [anon_sym_ATdynamic] = ACTIONS(5507),
    [anon_sym_BOOL] = ACTIONS(5505),
    [anon_sym_IMP] = ACTIONS(5505),
    [anon_sym_SEL] = ACTIONS(5505),
    [anon_sym_Class] = ACTIONS(5505),
    [anon_sym_id] = ACTIONS(5505),
  },
  [STATE(2363)] = {
    [sym_identifier] = ACTIONS(5517),
    [aux_sym_preproc_def_token1] = ACTIONS(5517),
    [aux_sym_preproc_if_token1] = ACTIONS(5517),
    [aux_sym_preproc_if_token2] = ACTIONS(5517),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5517),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5517),
    [sym_preproc_directive] = ACTIONS(5517),
    [anon_sym_DASH] = ACTIONS(5519),
    [anon_sym_PLUS] = ACTIONS(5519),
    [anon_sym___extension__] = ACTIONS(5517),
    [anon_sym_typedef] = ACTIONS(5517),
    [anon_sym_extern] = ACTIONS(5517),
    [anon_sym___attribute__] = ACTIONS(5517),
    [anon_sym___attribute] = ACTIONS(5517),
    [anon_sym_noreturn] = ACTIONS(5517),
    [anon_sym_LBRACK] = ACTIONS(5519),
    [anon_sym___declspec] = ACTIONS(5517),
    [anon_sym___cdecl] = ACTIONS(5517),
    [anon_sym___clrcall] = ACTIONS(5517),
    [anon_sym___stdcall] = ACTIONS(5517),
    [anon_sym___fastcall] = ACTIONS(5517),
    [anon_sym___thiscall] = ACTIONS(5517),
    [anon_sym___vectorcall] = ACTIONS(5517),
    [anon_sym_signed] = ACTIONS(5517),
    [anon_sym_unsigned] = ACTIONS(5517),
    [anon_sym_long] = ACTIONS(5517),
    [anon_sym_short] = ACTIONS(5517),
    [anon_sym_static] = ACTIONS(5517),
    [anon_sym_auto] = ACTIONS(5517),
    [anon_sym_register] = ACTIONS(5517),
    [anon_sym_inline] = ACTIONS(5517),
    [anon_sym___inline] = ACTIONS(5517),
    [anon_sym___inline__] = ACTIONS(5517),
    [anon_sym___forceinline] = ACTIONS(5517),
    [anon_sym_thread_local] = ACTIONS(5517),
    [anon_sym___thread] = ACTIONS(5517),
    [anon_sym_CG_EXTERN] = ACTIONS(5517),
    [anon_sym_CG_INLINE] = ACTIONS(5517),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5517),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5517),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5517),
    [anon_sym_IBOutlet] = ACTIONS(5517),
    [anon_sym_IBInspectable] = ACTIONS(5517),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5517),
    [anon_sym_NS_INLINE] = ACTIONS(5517),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5517),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5517),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5517),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5517),
    [anon_sym_const] = ACTIONS(5517),
    [anon_sym_constexpr] = ACTIONS(5517),
    [anon_sym_volatile] = ACTIONS(5517),
    [anon_sym_restrict] = ACTIONS(5517),
    [anon_sym___restrict__] = ACTIONS(5517),
    [anon_sym__Atomic] = ACTIONS(5517),
    [anon_sym__Noreturn] = ACTIONS(5517),
    [anon_sym__Nonnull] = ACTIONS(5517),
    [anon_sym_nullable] = ACTIONS(5517),
    [anon_sym__Complex] = ACTIONS(5517),
    [anon_sym__Nullable] = ACTIONS(5517),
    [anon_sym__Nullable_result] = ACTIONS(5517),
    [anon_sym__Null_unspecified] = ACTIONS(5517),
    [anon_sym___autoreleasing] = ACTIONS(5517),
    [anon_sym___block] = ACTIONS(5517),
    [anon_sym___bridge] = ACTIONS(5517),
    [anon_sym___bridge_retained] = ACTIONS(5517),
    [anon_sym___bridge_transfer] = ACTIONS(5517),
    [anon_sym___complex] = ACTIONS(5517),
    [anon_sym___const] = ACTIONS(5517),
    [anon_sym___imag] = ACTIONS(5517),
    [anon_sym___kindof] = ACTIONS(5517),
    [anon_sym___nonnull] = ACTIONS(5517),
    [anon_sym___nullable] = ACTIONS(5517),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5517),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5517),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5517),
    [anon_sym___real] = ACTIONS(5517),
    [anon_sym___strong] = ACTIONS(5517),
    [anon_sym___unsafe_unretained] = ACTIONS(5517),
    [anon_sym___unused] = ACTIONS(5517),
    [anon_sym___weak] = ACTIONS(5517),
    [anon_sym_alignas] = ACTIONS(5517),
    [anon_sym__Alignas] = ACTIONS(5517),
    [sym_primitive_type] = ACTIONS(5517),
    [anon_sym_enum] = ACTIONS(5517),
    [anon_sym_struct] = ACTIONS(5517),
    [anon_sym_union] = ACTIONS(5517),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5517),
    [anon_sym___typeof] = ACTIONS(5517),
    [anon_sym_typeof] = ACTIONS(5517),
    [aux_sym_preproc_undef_token1] = ACTIONS(5517),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5517),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5517),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5517),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5517),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5517),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5517),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5517),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5517),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5517),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_API_AVAILABLE] = ACTIONS(5517),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_API_DEPRECATED] = ACTIONS(5517),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5517),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5517),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5517),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5517),
    [anon_sym___deprecated_msg] = ACTIONS(5517),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5517),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5517),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5517),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5517),
    [anon_sym_ATsynthesize] = ACTIONS(5519),
    [anon_sym_ATdynamic] = ACTIONS(5519),
    [anon_sym_BOOL] = ACTIONS(5517),
    [anon_sym_IMP] = ACTIONS(5517),
    [anon_sym_SEL] = ACTIONS(5517),
    [anon_sym_Class] = ACTIONS(5517),
    [anon_sym_id] = ACTIONS(5517),
  },
  [STATE(2364)] = {
    [sym_identifier] = ACTIONS(5509),
    [aux_sym_preproc_def_token1] = ACTIONS(5509),
    [aux_sym_preproc_if_token1] = ACTIONS(5509),
    [aux_sym_preproc_if_token2] = ACTIONS(5509),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5509),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5509),
    [sym_preproc_directive] = ACTIONS(5509),
    [anon_sym_DASH] = ACTIONS(5511),
    [anon_sym_PLUS] = ACTIONS(5511),
    [anon_sym___extension__] = ACTIONS(5509),
    [anon_sym_typedef] = ACTIONS(5509),
    [anon_sym_extern] = ACTIONS(5509),
    [anon_sym___attribute__] = ACTIONS(5509),
    [anon_sym___attribute] = ACTIONS(5509),
    [anon_sym_noreturn] = ACTIONS(5509),
    [anon_sym_LBRACK] = ACTIONS(5511),
    [anon_sym___declspec] = ACTIONS(5509),
    [anon_sym___cdecl] = ACTIONS(5509),
    [anon_sym___clrcall] = ACTIONS(5509),
    [anon_sym___stdcall] = ACTIONS(5509),
    [anon_sym___fastcall] = ACTIONS(5509),
    [anon_sym___thiscall] = ACTIONS(5509),
    [anon_sym___vectorcall] = ACTIONS(5509),
    [anon_sym_signed] = ACTIONS(5509),
    [anon_sym_unsigned] = ACTIONS(5509),
    [anon_sym_long] = ACTIONS(5509),
    [anon_sym_short] = ACTIONS(5509),
    [anon_sym_static] = ACTIONS(5509),
    [anon_sym_auto] = ACTIONS(5509),
    [anon_sym_register] = ACTIONS(5509),
    [anon_sym_inline] = ACTIONS(5509),
    [anon_sym___inline] = ACTIONS(5509),
    [anon_sym___inline__] = ACTIONS(5509),
    [anon_sym___forceinline] = ACTIONS(5509),
    [anon_sym_thread_local] = ACTIONS(5509),
    [anon_sym___thread] = ACTIONS(5509),
    [anon_sym_CG_EXTERN] = ACTIONS(5509),
    [anon_sym_CG_INLINE] = ACTIONS(5509),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5509),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5509),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5509),
    [anon_sym_IBOutlet] = ACTIONS(5509),
    [anon_sym_IBInspectable] = ACTIONS(5509),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5509),
    [anon_sym_NS_INLINE] = ACTIONS(5509),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5509),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5509),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5509),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5509),
    [anon_sym_const] = ACTIONS(5509),
    [anon_sym_constexpr] = ACTIONS(5509),
    [anon_sym_volatile] = ACTIONS(5509),
    [anon_sym_restrict] = ACTIONS(5509),
    [anon_sym___restrict__] = ACTIONS(5509),
    [anon_sym__Atomic] = ACTIONS(5509),
    [anon_sym__Noreturn] = ACTIONS(5509),
    [anon_sym__Nonnull] = ACTIONS(5509),
    [anon_sym_nullable] = ACTIONS(5509),
    [anon_sym__Complex] = ACTIONS(5509),
    [anon_sym__Nullable] = ACTIONS(5509),
    [anon_sym__Nullable_result] = ACTIONS(5509),
    [anon_sym__Null_unspecified] = ACTIONS(5509),
    [anon_sym___autoreleasing] = ACTIONS(5509),
    [anon_sym___block] = ACTIONS(5509),
    [anon_sym___bridge] = ACTIONS(5509),
    [anon_sym___bridge_retained] = ACTIONS(5509),
    [anon_sym___bridge_transfer] = ACTIONS(5509),
    [anon_sym___complex] = ACTIONS(5509),
    [anon_sym___const] = ACTIONS(5509),
    [anon_sym___imag] = ACTIONS(5509),
    [anon_sym___kindof] = ACTIONS(5509),
    [anon_sym___nonnull] = ACTIONS(5509),
    [anon_sym___nullable] = ACTIONS(5509),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5509),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5509),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5509),
    [anon_sym___real] = ACTIONS(5509),
    [anon_sym___strong] = ACTIONS(5509),
    [anon_sym___unsafe_unretained] = ACTIONS(5509),
    [anon_sym___unused] = ACTIONS(5509),
    [anon_sym___weak] = ACTIONS(5509),
    [anon_sym_alignas] = ACTIONS(5509),
    [anon_sym__Alignas] = ACTIONS(5509),
    [sym_primitive_type] = ACTIONS(5509),
    [anon_sym_enum] = ACTIONS(5509),
    [anon_sym_struct] = ACTIONS(5509),
    [anon_sym_union] = ACTIONS(5509),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5509),
    [anon_sym___typeof] = ACTIONS(5509),
    [anon_sym_typeof] = ACTIONS(5509),
    [aux_sym_preproc_undef_token1] = ACTIONS(5509),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5509),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5509),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5509),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5509),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5509),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5509),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5509),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5509),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5509),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_API_AVAILABLE] = ACTIONS(5509),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_API_DEPRECATED] = ACTIONS(5509),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5509),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5509),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5509),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5509),
    [anon_sym___deprecated_msg] = ACTIONS(5509),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5509),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5509),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5509),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5509),
    [anon_sym_ATsynthesize] = ACTIONS(5511),
    [anon_sym_ATdynamic] = ACTIONS(5511),
    [anon_sym_BOOL] = ACTIONS(5509),
    [anon_sym_IMP] = ACTIONS(5509),
    [anon_sym_SEL] = ACTIONS(5509),
    [anon_sym_Class] = ACTIONS(5509),
    [anon_sym_id] = ACTIONS(5509),
  },
  [STATE(2365)] = {
    [sym_identifier] = ACTIONS(5521),
    [aux_sym_preproc_def_token1] = ACTIONS(5521),
    [aux_sym_preproc_if_token1] = ACTIONS(5521),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5521),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5521),
    [sym_preproc_directive] = ACTIONS(5521),
    [anon_sym_DASH] = ACTIONS(5523),
    [anon_sym_PLUS] = ACTIONS(5523),
    [anon_sym___extension__] = ACTIONS(5521),
    [anon_sym_typedef] = ACTIONS(5521),
    [anon_sym_extern] = ACTIONS(5521),
    [anon_sym___attribute__] = ACTIONS(5521),
    [anon_sym___attribute] = ACTIONS(5521),
    [anon_sym_noreturn] = ACTIONS(5521),
    [anon_sym_LBRACK] = ACTIONS(5523),
    [anon_sym___declspec] = ACTIONS(5521),
    [anon_sym___cdecl] = ACTIONS(5521),
    [anon_sym___clrcall] = ACTIONS(5521),
    [anon_sym___stdcall] = ACTIONS(5521),
    [anon_sym___fastcall] = ACTIONS(5521),
    [anon_sym___thiscall] = ACTIONS(5521),
    [anon_sym___vectorcall] = ACTIONS(5521),
    [anon_sym_signed] = ACTIONS(5521),
    [anon_sym_unsigned] = ACTIONS(5521),
    [anon_sym_long] = ACTIONS(5521),
    [anon_sym_short] = ACTIONS(5521),
    [anon_sym_static] = ACTIONS(5521),
    [anon_sym_auto] = ACTIONS(5521),
    [anon_sym_register] = ACTIONS(5521),
    [anon_sym_inline] = ACTIONS(5521),
    [anon_sym___inline] = ACTIONS(5521),
    [anon_sym___inline__] = ACTIONS(5521),
    [anon_sym___forceinline] = ACTIONS(5521),
    [anon_sym_thread_local] = ACTIONS(5521),
    [anon_sym___thread] = ACTIONS(5521),
    [anon_sym_CG_EXTERN] = ACTIONS(5521),
    [anon_sym_CG_INLINE] = ACTIONS(5521),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5521),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5521),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5521),
    [anon_sym_IBOutlet] = ACTIONS(5521),
    [anon_sym_IBInspectable] = ACTIONS(5521),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5521),
    [anon_sym_NS_INLINE] = ACTIONS(5521),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5521),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5521),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5521),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5521),
    [anon_sym_const] = ACTIONS(5521),
    [anon_sym_constexpr] = ACTIONS(5521),
    [anon_sym_volatile] = ACTIONS(5521),
    [anon_sym_restrict] = ACTIONS(5521),
    [anon_sym___restrict__] = ACTIONS(5521),
    [anon_sym__Atomic] = ACTIONS(5521),
    [anon_sym__Noreturn] = ACTIONS(5521),
    [anon_sym__Nonnull] = ACTIONS(5521),
    [anon_sym_nullable] = ACTIONS(5521),
    [anon_sym__Complex] = ACTIONS(5521),
    [anon_sym__Nullable] = ACTIONS(5521),
    [anon_sym__Nullable_result] = ACTIONS(5521),
    [anon_sym__Null_unspecified] = ACTIONS(5521),
    [anon_sym___autoreleasing] = ACTIONS(5521),
    [anon_sym___block] = ACTIONS(5521),
    [anon_sym___bridge] = ACTIONS(5521),
    [anon_sym___bridge_retained] = ACTIONS(5521),
    [anon_sym___bridge_transfer] = ACTIONS(5521),
    [anon_sym___complex] = ACTIONS(5521),
    [anon_sym___const] = ACTIONS(5521),
    [anon_sym___imag] = ACTIONS(5521),
    [anon_sym___kindof] = ACTIONS(5521),
    [anon_sym___nonnull] = ACTIONS(5521),
    [anon_sym___nullable] = ACTIONS(5521),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5521),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5521),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5521),
    [anon_sym___real] = ACTIONS(5521),
    [anon_sym___strong] = ACTIONS(5521),
    [anon_sym___unsafe_unretained] = ACTIONS(5521),
    [anon_sym___unused] = ACTIONS(5521),
    [anon_sym___weak] = ACTIONS(5521),
    [anon_sym_alignas] = ACTIONS(5521),
    [anon_sym__Alignas] = ACTIONS(5521),
    [sym_primitive_type] = ACTIONS(5521),
    [anon_sym_enum] = ACTIONS(5521),
    [anon_sym_struct] = ACTIONS(5521),
    [anon_sym_union] = ACTIONS(5521),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5521),
    [anon_sym___typeof] = ACTIONS(5521),
    [anon_sym_typeof] = ACTIONS(5521),
    [aux_sym_preproc_undef_token1] = ACTIONS(5521),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5521),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5521),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5521),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5521),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5521),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5521),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5521),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5521),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5521),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_API_AVAILABLE] = ACTIONS(5521),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_API_DEPRECATED] = ACTIONS(5521),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5521),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5521),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5521),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5521),
    [anon_sym___deprecated_msg] = ACTIONS(5521),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5521),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5521),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5521),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5521),
    [anon_sym_ATend] = ACTIONS(5523),
    [anon_sym_ATsynthesize] = ACTIONS(5523),
    [anon_sym_ATdynamic] = ACTIONS(5523),
    [anon_sym_BOOL] = ACTIONS(5521),
    [anon_sym_IMP] = ACTIONS(5521),
    [anon_sym_SEL] = ACTIONS(5521),
    [anon_sym_Class] = ACTIONS(5521),
    [anon_sym_id] = ACTIONS(5521),
  },
  [STATE(2366)] = {
    [sym_identifier] = ACTIONS(5475),
    [aux_sym_preproc_def_token1] = ACTIONS(5475),
    [aux_sym_preproc_if_token1] = ACTIONS(5475),
    [aux_sym_preproc_if_token2] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5475),
    [sym_preproc_directive] = ACTIONS(5475),
    [anon_sym_DASH] = ACTIONS(5477),
    [anon_sym_PLUS] = ACTIONS(5477),
    [anon_sym___extension__] = ACTIONS(5475),
    [anon_sym_typedef] = ACTIONS(5475),
    [anon_sym_extern] = ACTIONS(5475),
    [anon_sym___attribute__] = ACTIONS(5475),
    [anon_sym___attribute] = ACTIONS(5475),
    [anon_sym_noreturn] = ACTIONS(5475),
    [anon_sym_LBRACK] = ACTIONS(5477),
    [anon_sym___declspec] = ACTIONS(5475),
    [anon_sym___cdecl] = ACTIONS(5475),
    [anon_sym___clrcall] = ACTIONS(5475),
    [anon_sym___stdcall] = ACTIONS(5475),
    [anon_sym___fastcall] = ACTIONS(5475),
    [anon_sym___thiscall] = ACTIONS(5475),
    [anon_sym___vectorcall] = ACTIONS(5475),
    [anon_sym_signed] = ACTIONS(5475),
    [anon_sym_unsigned] = ACTIONS(5475),
    [anon_sym_long] = ACTIONS(5475),
    [anon_sym_short] = ACTIONS(5475),
    [anon_sym_static] = ACTIONS(5475),
    [anon_sym_auto] = ACTIONS(5475),
    [anon_sym_register] = ACTIONS(5475),
    [anon_sym_inline] = ACTIONS(5475),
    [anon_sym___inline] = ACTIONS(5475),
    [anon_sym___inline__] = ACTIONS(5475),
    [anon_sym___forceinline] = ACTIONS(5475),
    [anon_sym_thread_local] = ACTIONS(5475),
    [anon_sym___thread] = ACTIONS(5475),
    [anon_sym_CG_EXTERN] = ACTIONS(5475),
    [anon_sym_CG_INLINE] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5475),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5475),
    [anon_sym_IBOutlet] = ACTIONS(5475),
    [anon_sym_IBInspectable] = ACTIONS(5475),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5475),
    [anon_sym_NS_INLINE] = ACTIONS(5475),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5475),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5475),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5475),
    [anon_sym_const] = ACTIONS(5475),
    [anon_sym_constexpr] = ACTIONS(5475),
    [anon_sym_volatile] = ACTIONS(5475),
    [anon_sym_restrict] = ACTIONS(5475),
    [anon_sym___restrict__] = ACTIONS(5475),
    [anon_sym__Atomic] = ACTIONS(5475),
    [anon_sym__Noreturn] = ACTIONS(5475),
    [anon_sym__Nonnull] = ACTIONS(5475),
    [anon_sym_nullable] = ACTIONS(5475),
    [anon_sym__Complex] = ACTIONS(5475),
    [anon_sym__Nullable] = ACTIONS(5475),
    [anon_sym__Nullable_result] = ACTIONS(5475),
    [anon_sym__Null_unspecified] = ACTIONS(5475),
    [anon_sym___autoreleasing] = ACTIONS(5475),
    [anon_sym___block] = ACTIONS(5475),
    [anon_sym___bridge] = ACTIONS(5475),
    [anon_sym___bridge_retained] = ACTIONS(5475),
    [anon_sym___bridge_transfer] = ACTIONS(5475),
    [anon_sym___complex] = ACTIONS(5475),
    [anon_sym___const] = ACTIONS(5475),
    [anon_sym___imag] = ACTIONS(5475),
    [anon_sym___kindof] = ACTIONS(5475),
    [anon_sym___nonnull] = ACTIONS(5475),
    [anon_sym___nullable] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5475),
    [anon_sym___real] = ACTIONS(5475),
    [anon_sym___strong] = ACTIONS(5475),
    [anon_sym___unsafe_unretained] = ACTIONS(5475),
    [anon_sym___unused] = ACTIONS(5475),
    [anon_sym___weak] = ACTIONS(5475),
    [anon_sym_alignas] = ACTIONS(5475),
    [anon_sym__Alignas] = ACTIONS(5475),
    [sym_primitive_type] = ACTIONS(5475),
    [anon_sym_enum] = ACTIONS(5475),
    [anon_sym_struct] = ACTIONS(5475),
    [anon_sym_union] = ACTIONS(5475),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5475),
    [anon_sym___typeof] = ACTIONS(5475),
    [anon_sym_typeof] = ACTIONS(5475),
    [aux_sym_preproc_undef_token1] = ACTIONS(5475),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5475),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5475),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5475),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_API_AVAILABLE] = ACTIONS(5475),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_API_DEPRECATED] = ACTIONS(5475),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym___deprecated_msg] = ACTIONS(5475),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5475),
    [anon_sym_ATsynthesize] = ACTIONS(5477),
    [anon_sym_ATdynamic] = ACTIONS(5477),
    [anon_sym_BOOL] = ACTIONS(5475),
    [anon_sym_IMP] = ACTIONS(5475),
    [anon_sym_SEL] = ACTIONS(5475),
    [anon_sym_Class] = ACTIONS(5475),
    [anon_sym_id] = ACTIONS(5475),
  },
  [STATE(2367)] = {
    [sym_identifier] = ACTIONS(5387),
    [aux_sym_preproc_def_token1] = ACTIONS(5387),
    [aux_sym_preproc_if_token1] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5387),
    [sym_preproc_directive] = ACTIONS(5387),
    [anon_sym_DASH] = ACTIONS(5389),
    [anon_sym_PLUS] = ACTIONS(5389),
    [anon_sym___extension__] = ACTIONS(5387),
    [anon_sym_typedef] = ACTIONS(5387),
    [anon_sym_extern] = ACTIONS(5387),
    [anon_sym___attribute__] = ACTIONS(5387),
    [anon_sym___attribute] = ACTIONS(5387),
    [anon_sym_noreturn] = ACTIONS(5387),
    [anon_sym_LBRACK] = ACTIONS(5389),
    [anon_sym___declspec] = ACTIONS(5387),
    [anon_sym___cdecl] = ACTIONS(5387),
    [anon_sym___clrcall] = ACTIONS(5387),
    [anon_sym___stdcall] = ACTIONS(5387),
    [anon_sym___fastcall] = ACTIONS(5387),
    [anon_sym___thiscall] = ACTIONS(5387),
    [anon_sym___vectorcall] = ACTIONS(5387),
    [anon_sym_signed] = ACTIONS(5387),
    [anon_sym_unsigned] = ACTIONS(5387),
    [anon_sym_long] = ACTIONS(5387),
    [anon_sym_short] = ACTIONS(5387),
    [anon_sym_static] = ACTIONS(5387),
    [anon_sym_auto] = ACTIONS(5387),
    [anon_sym_register] = ACTIONS(5387),
    [anon_sym_inline] = ACTIONS(5387),
    [anon_sym___inline] = ACTIONS(5387),
    [anon_sym___inline__] = ACTIONS(5387),
    [anon_sym___forceinline] = ACTIONS(5387),
    [anon_sym_thread_local] = ACTIONS(5387),
    [anon_sym___thread] = ACTIONS(5387),
    [anon_sym_CG_EXTERN] = ACTIONS(5387),
    [anon_sym_CG_INLINE] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5387),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5387),
    [anon_sym_IBOutlet] = ACTIONS(5387),
    [anon_sym_IBInspectable] = ACTIONS(5387),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5387),
    [anon_sym_NS_INLINE] = ACTIONS(5387),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5387),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5387),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5387),
    [anon_sym_const] = ACTIONS(5387),
    [anon_sym_constexpr] = ACTIONS(5387),
    [anon_sym_volatile] = ACTIONS(5387),
    [anon_sym_restrict] = ACTIONS(5387),
    [anon_sym___restrict__] = ACTIONS(5387),
    [anon_sym__Atomic] = ACTIONS(5387),
    [anon_sym__Noreturn] = ACTIONS(5387),
    [anon_sym__Nonnull] = ACTIONS(5387),
    [anon_sym_nullable] = ACTIONS(5387),
    [anon_sym__Complex] = ACTIONS(5387),
    [anon_sym__Nullable] = ACTIONS(5387),
    [anon_sym__Nullable_result] = ACTIONS(5387),
    [anon_sym__Null_unspecified] = ACTIONS(5387),
    [anon_sym___autoreleasing] = ACTIONS(5387),
    [anon_sym___block] = ACTIONS(5387),
    [anon_sym___bridge] = ACTIONS(5387),
    [anon_sym___bridge_retained] = ACTIONS(5387),
    [anon_sym___bridge_transfer] = ACTIONS(5387),
    [anon_sym___complex] = ACTIONS(5387),
    [anon_sym___const] = ACTIONS(5387),
    [anon_sym___imag] = ACTIONS(5387),
    [anon_sym___kindof] = ACTIONS(5387),
    [anon_sym___nonnull] = ACTIONS(5387),
    [anon_sym___nullable] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5387),
    [anon_sym___real] = ACTIONS(5387),
    [anon_sym___strong] = ACTIONS(5387),
    [anon_sym___unsafe_unretained] = ACTIONS(5387),
    [anon_sym___unused] = ACTIONS(5387),
    [anon_sym___weak] = ACTIONS(5387),
    [anon_sym_alignas] = ACTIONS(5387),
    [anon_sym__Alignas] = ACTIONS(5387),
    [sym_primitive_type] = ACTIONS(5387),
    [anon_sym_enum] = ACTIONS(5387),
    [anon_sym_struct] = ACTIONS(5387),
    [anon_sym_union] = ACTIONS(5387),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5387),
    [anon_sym___typeof] = ACTIONS(5387),
    [anon_sym_typeof] = ACTIONS(5387),
    [aux_sym_preproc_undef_token1] = ACTIONS(5387),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5387),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5387),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5387),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_API_AVAILABLE] = ACTIONS(5387),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_API_DEPRECATED] = ACTIONS(5387),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym___deprecated_msg] = ACTIONS(5387),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5387),
    [anon_sym_ATend] = ACTIONS(5389),
    [anon_sym_ATsynthesize] = ACTIONS(5389),
    [anon_sym_ATdynamic] = ACTIONS(5389),
    [anon_sym_BOOL] = ACTIONS(5387),
    [anon_sym_IMP] = ACTIONS(5387),
    [anon_sym_SEL] = ACTIONS(5387),
    [anon_sym_Class] = ACTIONS(5387),
    [anon_sym_id] = ACTIONS(5387),
  },
  [STATE(2368)] = {
    [sym_identifier] = ACTIONS(5674),
    [anon_sym_COMMA] = ACTIONS(5676),
    [anon_sym_RPAREN] = ACTIONS(5676),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5676),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5676),
    [anon_sym_LPAREN2] = ACTIONS(5676),
    [anon_sym_STAR] = ACTIONS(5676),
    [anon_sym_CARET] = ACTIONS(5676),
    [anon_sym_GT] = ACTIONS(5676),
    [anon_sym_SEMI] = ACTIONS(5676),
    [anon_sym___extension__] = ACTIONS(5674),
    [anon_sym_extern] = ACTIONS(5674),
    [anon_sym___attribute__] = ACTIONS(5674),
    [anon_sym___attribute] = ACTIONS(5674),
    [anon_sym_noreturn] = ACTIONS(5674),
    [anon_sym_LBRACK] = ACTIONS(5676),
    [anon_sym___declspec] = ACTIONS(5674),
    [anon_sym___based] = ACTIONS(5674),
    [anon_sym_LBRACE] = ACTIONS(5676),
    [anon_sym_RBRACE] = ACTIONS(5676),
    [anon_sym_signed] = ACTIONS(5674),
    [anon_sym_unsigned] = ACTIONS(5674),
    [anon_sym_long] = ACTIONS(5674),
    [anon_sym_short] = ACTIONS(5674),
    [anon_sym_static] = ACTIONS(5674),
    [anon_sym_ATautoreleasepool] = ACTIONS(5676),
    [anon_sym_auto] = ACTIONS(5674),
    [anon_sym_register] = ACTIONS(5674),
    [anon_sym_inline] = ACTIONS(5674),
    [anon_sym___inline] = ACTIONS(5674),
    [anon_sym___inline__] = ACTIONS(5674),
    [anon_sym___forceinline] = ACTIONS(5674),
    [anon_sym_thread_local] = ACTIONS(5674),
    [anon_sym___thread] = ACTIONS(5674),
    [anon_sym_CG_EXTERN] = ACTIONS(5674),
    [anon_sym_CG_INLINE] = ACTIONS(5674),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5674),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5674),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5674),
    [anon_sym_IBOutlet] = ACTIONS(5674),
    [anon_sym_IBInspectable] = ACTIONS(5674),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5674),
    [anon_sym_NS_INLINE] = ACTIONS(5674),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5674),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5674),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5674),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5674),
    [anon_sym_const] = ACTIONS(5674),
    [anon_sym_constexpr] = ACTIONS(5674),
    [anon_sym_volatile] = ACTIONS(5674),
    [anon_sym_restrict] = ACTIONS(5674),
    [anon_sym___restrict__] = ACTIONS(5674),
    [anon_sym__Atomic] = ACTIONS(5674),
    [anon_sym__Noreturn] = ACTIONS(5674),
    [anon_sym__Nonnull] = ACTIONS(5674),
    [anon_sym_nullable] = ACTIONS(5674),
    [anon_sym__Complex] = ACTIONS(5674),
    [anon_sym__Nullable] = ACTIONS(5674),
    [anon_sym__Nullable_result] = ACTIONS(5674),
    [anon_sym__Null_unspecified] = ACTIONS(5674),
    [anon_sym___autoreleasing] = ACTIONS(5674),
    [anon_sym___block] = ACTIONS(5674),
    [anon_sym___bridge] = ACTIONS(5674),
    [anon_sym___bridge_retained] = ACTIONS(5674),
    [anon_sym___bridge_transfer] = ACTIONS(5674),
    [anon_sym___complex] = ACTIONS(5674),
    [anon_sym___const] = ACTIONS(5674),
    [anon_sym___imag] = ACTIONS(5674),
    [anon_sym___kindof] = ACTIONS(5674),
    [anon_sym___nonnull] = ACTIONS(5674),
    [anon_sym___nullable] = ACTIONS(5674),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5674),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5674),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5674),
    [anon_sym___real] = ACTIONS(5674),
    [anon_sym___strong] = ACTIONS(5674),
    [anon_sym___unsafe_unretained] = ACTIONS(5674),
    [anon_sym___unused] = ACTIONS(5674),
    [anon_sym___weak] = ACTIONS(5674),
    [anon_sym_alignas] = ACTIONS(5674),
    [anon_sym__Alignas] = ACTIONS(5674),
    [sym_primitive_type] = ACTIONS(5674),
    [anon_sym_enum] = ACTIONS(5674),
    [anon_sym_COLON] = ACTIONS(5676),
    [anon_sym_struct] = ACTIONS(5674),
    [anon_sym_union] = ACTIONS(5674),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5674),
    [anon_sym___typeof] = ACTIONS(5674),
    [anon_sym_typeof] = ACTIONS(5674),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5674),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5674),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5674),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5674),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5674),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5674),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5674),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5674),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5674),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5674),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5674),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5674),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5674),
    [anon_sym_API_AVAILABLE] = ACTIONS(5674),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5674),
    [anon_sym_API_DEPRECATED] = ACTIONS(5674),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5674),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5674),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5674),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5674),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5674),
    [anon_sym___deprecated_msg] = ACTIONS(5674),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5674),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5674),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5674),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5674),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5674),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5674),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5674),
    [anon_sym_ATprotocol] = ACTIONS(5676),
    [anon_sym_ATinterface] = ACTIONS(5676),
    [anon_sym_ATimplementation] = ACTIONS(5676),
    [anon_sym_ATproperty] = ACTIONS(5676),
    [anon_sym_BOOL] = ACTIONS(5674),
    [anon_sym_IMP] = ACTIONS(5674),
    [anon_sym_SEL] = ACTIONS(5674),
    [anon_sym_Class] = ACTIONS(5674),
    [anon_sym_id] = ACTIONS(5674),
  },
  [STATE(2369)] = {
    [sym_identifier] = ACTIONS(5359),
    [aux_sym_preproc_def_token1] = ACTIONS(5359),
    [aux_sym_preproc_if_token1] = ACTIONS(5359),
    [aux_sym_preproc_if_token2] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5359),
    [sym_preproc_directive] = ACTIONS(5359),
    [anon_sym_DASH] = ACTIONS(5361),
    [anon_sym_PLUS] = ACTIONS(5361),
    [anon_sym___extension__] = ACTIONS(5359),
    [anon_sym_typedef] = ACTIONS(5359),
    [anon_sym_extern] = ACTIONS(5359),
    [anon_sym___attribute__] = ACTIONS(5359),
    [anon_sym___attribute] = ACTIONS(5359),
    [anon_sym_noreturn] = ACTIONS(5359),
    [anon_sym_LBRACK] = ACTIONS(5361),
    [anon_sym___declspec] = ACTIONS(5359),
    [anon_sym___cdecl] = ACTIONS(5359),
    [anon_sym___clrcall] = ACTIONS(5359),
    [anon_sym___stdcall] = ACTIONS(5359),
    [anon_sym___fastcall] = ACTIONS(5359),
    [anon_sym___thiscall] = ACTIONS(5359),
    [anon_sym___vectorcall] = ACTIONS(5359),
    [anon_sym_signed] = ACTIONS(5359),
    [anon_sym_unsigned] = ACTIONS(5359),
    [anon_sym_long] = ACTIONS(5359),
    [anon_sym_short] = ACTIONS(5359),
    [anon_sym_static] = ACTIONS(5359),
    [anon_sym_auto] = ACTIONS(5359),
    [anon_sym_register] = ACTIONS(5359),
    [anon_sym_inline] = ACTIONS(5359),
    [anon_sym___inline] = ACTIONS(5359),
    [anon_sym___inline__] = ACTIONS(5359),
    [anon_sym___forceinline] = ACTIONS(5359),
    [anon_sym_thread_local] = ACTIONS(5359),
    [anon_sym___thread] = ACTIONS(5359),
    [anon_sym_CG_EXTERN] = ACTIONS(5359),
    [anon_sym_CG_INLINE] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5359),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5359),
    [anon_sym_IBOutlet] = ACTIONS(5359),
    [anon_sym_IBInspectable] = ACTIONS(5359),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5359),
    [anon_sym_NS_INLINE] = ACTIONS(5359),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5359),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5359),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5359),
    [anon_sym_const] = ACTIONS(5359),
    [anon_sym_constexpr] = ACTIONS(5359),
    [anon_sym_volatile] = ACTIONS(5359),
    [anon_sym_restrict] = ACTIONS(5359),
    [anon_sym___restrict__] = ACTIONS(5359),
    [anon_sym__Atomic] = ACTIONS(5359),
    [anon_sym__Noreturn] = ACTIONS(5359),
    [anon_sym__Nonnull] = ACTIONS(5359),
    [anon_sym_nullable] = ACTIONS(5359),
    [anon_sym__Complex] = ACTIONS(5359),
    [anon_sym__Nullable] = ACTIONS(5359),
    [anon_sym__Nullable_result] = ACTIONS(5359),
    [anon_sym__Null_unspecified] = ACTIONS(5359),
    [anon_sym___autoreleasing] = ACTIONS(5359),
    [anon_sym___block] = ACTIONS(5359),
    [anon_sym___bridge] = ACTIONS(5359),
    [anon_sym___bridge_retained] = ACTIONS(5359),
    [anon_sym___bridge_transfer] = ACTIONS(5359),
    [anon_sym___complex] = ACTIONS(5359),
    [anon_sym___const] = ACTIONS(5359),
    [anon_sym___imag] = ACTIONS(5359),
    [anon_sym___kindof] = ACTIONS(5359),
    [anon_sym___nonnull] = ACTIONS(5359),
    [anon_sym___nullable] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5359),
    [anon_sym___real] = ACTIONS(5359),
    [anon_sym___strong] = ACTIONS(5359),
    [anon_sym___unsafe_unretained] = ACTIONS(5359),
    [anon_sym___unused] = ACTIONS(5359),
    [anon_sym___weak] = ACTIONS(5359),
    [anon_sym_alignas] = ACTIONS(5359),
    [anon_sym__Alignas] = ACTIONS(5359),
    [sym_primitive_type] = ACTIONS(5359),
    [anon_sym_enum] = ACTIONS(5359),
    [anon_sym_struct] = ACTIONS(5359),
    [anon_sym_union] = ACTIONS(5359),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5359),
    [anon_sym___typeof] = ACTIONS(5359),
    [anon_sym_typeof] = ACTIONS(5359),
    [aux_sym_preproc_undef_token1] = ACTIONS(5359),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5359),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5359),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5359),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_API_AVAILABLE] = ACTIONS(5359),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_API_DEPRECATED] = ACTIONS(5359),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym___deprecated_msg] = ACTIONS(5359),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5359),
    [anon_sym_ATsynthesize] = ACTIONS(5361),
    [anon_sym_ATdynamic] = ACTIONS(5361),
    [anon_sym_BOOL] = ACTIONS(5359),
    [anon_sym_IMP] = ACTIONS(5359),
    [anon_sym_SEL] = ACTIONS(5359),
    [anon_sym_Class] = ACTIONS(5359),
    [anon_sym_id] = ACTIONS(5359),
  },
  [STATE(2370)] = {
    [sym_identifier] = ACTIONS(5393),
    [aux_sym_preproc_def_token1] = ACTIONS(5393),
    [aux_sym_preproc_if_token1] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5393),
    [sym_preproc_directive] = ACTIONS(5393),
    [anon_sym_DASH] = ACTIONS(5395),
    [anon_sym_PLUS] = ACTIONS(5395),
    [anon_sym___extension__] = ACTIONS(5393),
    [anon_sym_typedef] = ACTIONS(5393),
    [anon_sym_extern] = ACTIONS(5393),
    [anon_sym___attribute__] = ACTIONS(5393),
    [anon_sym___attribute] = ACTIONS(5393),
    [anon_sym_noreturn] = ACTIONS(5393),
    [anon_sym_LBRACK] = ACTIONS(5395),
    [anon_sym___declspec] = ACTIONS(5393),
    [anon_sym___cdecl] = ACTIONS(5393),
    [anon_sym___clrcall] = ACTIONS(5393),
    [anon_sym___stdcall] = ACTIONS(5393),
    [anon_sym___fastcall] = ACTIONS(5393),
    [anon_sym___thiscall] = ACTIONS(5393),
    [anon_sym___vectorcall] = ACTIONS(5393),
    [anon_sym_signed] = ACTIONS(5393),
    [anon_sym_unsigned] = ACTIONS(5393),
    [anon_sym_long] = ACTIONS(5393),
    [anon_sym_short] = ACTIONS(5393),
    [anon_sym_static] = ACTIONS(5393),
    [anon_sym_auto] = ACTIONS(5393),
    [anon_sym_register] = ACTIONS(5393),
    [anon_sym_inline] = ACTIONS(5393),
    [anon_sym___inline] = ACTIONS(5393),
    [anon_sym___inline__] = ACTIONS(5393),
    [anon_sym___forceinline] = ACTIONS(5393),
    [anon_sym_thread_local] = ACTIONS(5393),
    [anon_sym___thread] = ACTIONS(5393),
    [anon_sym_CG_EXTERN] = ACTIONS(5393),
    [anon_sym_CG_INLINE] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5393),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5393),
    [anon_sym_IBOutlet] = ACTIONS(5393),
    [anon_sym_IBInspectable] = ACTIONS(5393),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5393),
    [anon_sym_NS_INLINE] = ACTIONS(5393),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5393),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5393),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5393),
    [anon_sym_const] = ACTIONS(5393),
    [anon_sym_constexpr] = ACTIONS(5393),
    [anon_sym_volatile] = ACTIONS(5393),
    [anon_sym_restrict] = ACTIONS(5393),
    [anon_sym___restrict__] = ACTIONS(5393),
    [anon_sym__Atomic] = ACTIONS(5393),
    [anon_sym__Noreturn] = ACTIONS(5393),
    [anon_sym__Nonnull] = ACTIONS(5393),
    [anon_sym_nullable] = ACTIONS(5393),
    [anon_sym__Complex] = ACTIONS(5393),
    [anon_sym__Nullable] = ACTIONS(5393),
    [anon_sym__Nullable_result] = ACTIONS(5393),
    [anon_sym__Null_unspecified] = ACTIONS(5393),
    [anon_sym___autoreleasing] = ACTIONS(5393),
    [anon_sym___block] = ACTIONS(5393),
    [anon_sym___bridge] = ACTIONS(5393),
    [anon_sym___bridge_retained] = ACTIONS(5393),
    [anon_sym___bridge_transfer] = ACTIONS(5393),
    [anon_sym___complex] = ACTIONS(5393),
    [anon_sym___const] = ACTIONS(5393),
    [anon_sym___imag] = ACTIONS(5393),
    [anon_sym___kindof] = ACTIONS(5393),
    [anon_sym___nonnull] = ACTIONS(5393),
    [anon_sym___nullable] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5393),
    [anon_sym___real] = ACTIONS(5393),
    [anon_sym___strong] = ACTIONS(5393),
    [anon_sym___unsafe_unretained] = ACTIONS(5393),
    [anon_sym___unused] = ACTIONS(5393),
    [anon_sym___weak] = ACTIONS(5393),
    [anon_sym_alignas] = ACTIONS(5393),
    [anon_sym__Alignas] = ACTIONS(5393),
    [sym_primitive_type] = ACTIONS(5393),
    [anon_sym_enum] = ACTIONS(5393),
    [anon_sym_struct] = ACTIONS(5393),
    [anon_sym_union] = ACTIONS(5393),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5393),
    [anon_sym___typeof] = ACTIONS(5393),
    [anon_sym_typeof] = ACTIONS(5393),
    [aux_sym_preproc_undef_token1] = ACTIONS(5393),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5393),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5393),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5393),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_API_AVAILABLE] = ACTIONS(5393),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_API_DEPRECATED] = ACTIONS(5393),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym___deprecated_msg] = ACTIONS(5393),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5393),
    [anon_sym_ATend] = ACTIONS(5395),
    [anon_sym_ATsynthesize] = ACTIONS(5395),
    [anon_sym_ATdynamic] = ACTIONS(5395),
    [anon_sym_BOOL] = ACTIONS(5393),
    [anon_sym_IMP] = ACTIONS(5393),
    [anon_sym_SEL] = ACTIONS(5393),
    [anon_sym_Class] = ACTIONS(5393),
    [anon_sym_id] = ACTIONS(5393),
  },
  [STATE(2371)] = {
    [sym_identifier] = ACTIONS(5068),
    [aux_sym_preproc_def_token1] = ACTIONS(5068),
    [aux_sym_preproc_if_token1] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5068),
    [sym_preproc_directive] = ACTIONS(5068),
    [anon_sym_DASH] = ACTIONS(5072),
    [anon_sym_PLUS] = ACTIONS(5072),
    [anon_sym___extension__] = ACTIONS(5068),
    [anon_sym_typedef] = ACTIONS(5068),
    [anon_sym_extern] = ACTIONS(5068),
    [anon_sym___attribute__] = ACTIONS(5068),
    [anon_sym___attribute] = ACTIONS(5068),
    [anon_sym_noreturn] = ACTIONS(5068),
    [anon_sym_LBRACK] = ACTIONS(5072),
    [anon_sym___declspec] = ACTIONS(5068),
    [anon_sym___cdecl] = ACTIONS(5068),
    [anon_sym___clrcall] = ACTIONS(5068),
    [anon_sym___stdcall] = ACTIONS(5068),
    [anon_sym___fastcall] = ACTIONS(5068),
    [anon_sym___thiscall] = ACTIONS(5068),
    [anon_sym___vectorcall] = ACTIONS(5068),
    [anon_sym_signed] = ACTIONS(5068),
    [anon_sym_unsigned] = ACTIONS(5068),
    [anon_sym_long] = ACTIONS(5068),
    [anon_sym_short] = ACTIONS(5068),
    [anon_sym_static] = ACTIONS(5068),
    [anon_sym_auto] = ACTIONS(5068),
    [anon_sym_register] = ACTIONS(5068),
    [anon_sym_inline] = ACTIONS(5068),
    [anon_sym___inline] = ACTIONS(5068),
    [anon_sym___inline__] = ACTIONS(5068),
    [anon_sym___forceinline] = ACTIONS(5068),
    [anon_sym_thread_local] = ACTIONS(5068),
    [anon_sym___thread] = ACTIONS(5068),
    [anon_sym_CG_EXTERN] = ACTIONS(5068),
    [anon_sym_CG_INLINE] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5068),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5068),
    [anon_sym_IBOutlet] = ACTIONS(5068),
    [anon_sym_IBInspectable] = ACTIONS(5068),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5068),
    [anon_sym_NS_INLINE] = ACTIONS(5068),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5068),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5068),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5068),
    [anon_sym_const] = ACTIONS(5068),
    [anon_sym_constexpr] = ACTIONS(5068),
    [anon_sym_volatile] = ACTIONS(5068),
    [anon_sym_restrict] = ACTIONS(5068),
    [anon_sym___restrict__] = ACTIONS(5068),
    [anon_sym__Atomic] = ACTIONS(5068),
    [anon_sym__Noreturn] = ACTIONS(5068),
    [anon_sym__Nonnull] = ACTIONS(5068),
    [anon_sym_nullable] = ACTIONS(5068),
    [anon_sym__Complex] = ACTIONS(5068),
    [anon_sym__Nullable] = ACTIONS(5068),
    [anon_sym__Nullable_result] = ACTIONS(5068),
    [anon_sym__Null_unspecified] = ACTIONS(5068),
    [anon_sym___autoreleasing] = ACTIONS(5068),
    [anon_sym___block] = ACTIONS(5068),
    [anon_sym___bridge] = ACTIONS(5068),
    [anon_sym___bridge_retained] = ACTIONS(5068),
    [anon_sym___bridge_transfer] = ACTIONS(5068),
    [anon_sym___complex] = ACTIONS(5068),
    [anon_sym___const] = ACTIONS(5068),
    [anon_sym___imag] = ACTIONS(5068),
    [anon_sym___kindof] = ACTIONS(5068),
    [anon_sym___nonnull] = ACTIONS(5068),
    [anon_sym___nullable] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5068),
    [anon_sym___real] = ACTIONS(5068),
    [anon_sym___strong] = ACTIONS(5068),
    [anon_sym___unsafe_unretained] = ACTIONS(5068),
    [anon_sym___unused] = ACTIONS(5068),
    [anon_sym___weak] = ACTIONS(5068),
    [anon_sym_alignas] = ACTIONS(5068),
    [anon_sym__Alignas] = ACTIONS(5068),
    [sym_primitive_type] = ACTIONS(5068),
    [anon_sym_enum] = ACTIONS(5068),
    [anon_sym_struct] = ACTIONS(5068),
    [anon_sym_union] = ACTIONS(5068),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5068),
    [anon_sym___typeof] = ACTIONS(5068),
    [anon_sym_typeof] = ACTIONS(5068),
    [aux_sym_preproc_undef_token1] = ACTIONS(5068),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5068),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5068),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5068),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_API_AVAILABLE] = ACTIONS(5068),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_API_DEPRECATED] = ACTIONS(5068),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym___deprecated_msg] = ACTIONS(5068),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5068),
    [anon_sym_ATend] = ACTIONS(5072),
    [anon_sym_ATsynthesize] = ACTIONS(5072),
    [anon_sym_ATdynamic] = ACTIONS(5072),
    [anon_sym_BOOL] = ACTIONS(5068),
    [anon_sym_IMP] = ACTIONS(5068),
    [anon_sym_SEL] = ACTIONS(5068),
    [anon_sym_Class] = ACTIONS(5068),
    [anon_sym_id] = ACTIONS(5068),
  },
  [STATE(2372)] = {
    [sym_identifier] = ACTIONS(5678),
    [anon_sym_COMMA] = ACTIONS(5680),
    [anon_sym_RPAREN] = ACTIONS(5680),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5680),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5680),
    [anon_sym_LPAREN2] = ACTIONS(5680),
    [anon_sym_STAR] = ACTIONS(5680),
    [anon_sym_CARET] = ACTIONS(5680),
    [anon_sym_GT] = ACTIONS(5680),
    [anon_sym_SEMI] = ACTIONS(5680),
    [anon_sym___extension__] = ACTIONS(5678),
    [anon_sym_extern] = ACTIONS(5678),
    [anon_sym___attribute__] = ACTIONS(5678),
    [anon_sym___attribute] = ACTIONS(5678),
    [anon_sym_noreturn] = ACTIONS(5678),
    [anon_sym_LBRACK] = ACTIONS(5680),
    [anon_sym___declspec] = ACTIONS(5678),
    [anon_sym___based] = ACTIONS(5678),
    [anon_sym_LBRACE] = ACTIONS(5680),
    [anon_sym_RBRACE] = ACTIONS(5680),
    [anon_sym_signed] = ACTIONS(5678),
    [anon_sym_unsigned] = ACTIONS(5678),
    [anon_sym_long] = ACTIONS(5678),
    [anon_sym_short] = ACTIONS(5678),
    [anon_sym_static] = ACTIONS(5678),
    [anon_sym_ATautoreleasepool] = ACTIONS(5680),
    [anon_sym_auto] = ACTIONS(5678),
    [anon_sym_register] = ACTIONS(5678),
    [anon_sym_inline] = ACTIONS(5678),
    [anon_sym___inline] = ACTIONS(5678),
    [anon_sym___inline__] = ACTIONS(5678),
    [anon_sym___forceinline] = ACTIONS(5678),
    [anon_sym_thread_local] = ACTIONS(5678),
    [anon_sym___thread] = ACTIONS(5678),
    [anon_sym_CG_EXTERN] = ACTIONS(5678),
    [anon_sym_CG_INLINE] = ACTIONS(5678),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5678),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5678),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5678),
    [anon_sym_IBOutlet] = ACTIONS(5678),
    [anon_sym_IBInspectable] = ACTIONS(5678),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5678),
    [anon_sym_NS_INLINE] = ACTIONS(5678),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5678),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5678),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5678),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5678),
    [anon_sym_const] = ACTIONS(5678),
    [anon_sym_constexpr] = ACTIONS(5678),
    [anon_sym_volatile] = ACTIONS(5678),
    [anon_sym_restrict] = ACTIONS(5678),
    [anon_sym___restrict__] = ACTIONS(5678),
    [anon_sym__Atomic] = ACTIONS(5678),
    [anon_sym__Noreturn] = ACTIONS(5678),
    [anon_sym__Nonnull] = ACTIONS(5678),
    [anon_sym_nullable] = ACTIONS(5678),
    [anon_sym__Complex] = ACTIONS(5678),
    [anon_sym__Nullable] = ACTIONS(5678),
    [anon_sym__Nullable_result] = ACTIONS(5678),
    [anon_sym__Null_unspecified] = ACTIONS(5678),
    [anon_sym___autoreleasing] = ACTIONS(5678),
    [anon_sym___block] = ACTIONS(5678),
    [anon_sym___bridge] = ACTIONS(5678),
    [anon_sym___bridge_retained] = ACTIONS(5678),
    [anon_sym___bridge_transfer] = ACTIONS(5678),
    [anon_sym___complex] = ACTIONS(5678),
    [anon_sym___const] = ACTIONS(5678),
    [anon_sym___imag] = ACTIONS(5678),
    [anon_sym___kindof] = ACTIONS(5678),
    [anon_sym___nonnull] = ACTIONS(5678),
    [anon_sym___nullable] = ACTIONS(5678),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5678),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5678),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5678),
    [anon_sym___real] = ACTIONS(5678),
    [anon_sym___strong] = ACTIONS(5678),
    [anon_sym___unsafe_unretained] = ACTIONS(5678),
    [anon_sym___unused] = ACTIONS(5678),
    [anon_sym___weak] = ACTIONS(5678),
    [anon_sym_alignas] = ACTIONS(5678),
    [anon_sym__Alignas] = ACTIONS(5678),
    [sym_primitive_type] = ACTIONS(5678),
    [anon_sym_enum] = ACTIONS(5678),
    [anon_sym_COLON] = ACTIONS(5680),
    [anon_sym_struct] = ACTIONS(5678),
    [anon_sym_union] = ACTIONS(5678),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5678),
    [anon_sym___typeof] = ACTIONS(5678),
    [anon_sym_typeof] = ACTIONS(5678),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5678),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5678),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5678),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5678),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5678),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5678),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5678),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5678),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5678),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5678),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5678),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5678),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5678),
    [anon_sym_API_AVAILABLE] = ACTIONS(5678),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5678),
    [anon_sym_API_DEPRECATED] = ACTIONS(5678),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5678),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5678),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5678),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5678),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5678),
    [anon_sym___deprecated_msg] = ACTIONS(5678),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5678),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5678),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5678),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5678),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5678),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5678),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5678),
    [anon_sym_ATprotocol] = ACTIONS(5680),
    [anon_sym_ATinterface] = ACTIONS(5680),
    [anon_sym_ATimplementation] = ACTIONS(5680),
    [anon_sym_ATproperty] = ACTIONS(5680),
    [anon_sym_BOOL] = ACTIONS(5678),
    [anon_sym_IMP] = ACTIONS(5678),
    [anon_sym_SEL] = ACTIONS(5678),
    [anon_sym_Class] = ACTIONS(5678),
    [anon_sym_id] = ACTIONS(5678),
  },
  [STATE(2373)] = {
    [sym_identifier] = ACTIONS(2622),
    [aux_sym_preproc_def_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token2] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2622),
    [sym_preproc_directive] = ACTIONS(2622),
    [anon_sym_DASH] = ACTIONS(2624),
    [anon_sym_PLUS] = ACTIONS(2624),
    [anon_sym___extension__] = ACTIONS(2622),
    [anon_sym_typedef] = ACTIONS(2622),
    [anon_sym_extern] = ACTIONS(2622),
    [anon_sym___attribute__] = ACTIONS(2622),
    [anon_sym___attribute] = ACTIONS(2622),
    [anon_sym_noreturn] = ACTIONS(2622),
    [anon_sym_LBRACK] = ACTIONS(2624),
    [anon_sym___declspec] = ACTIONS(2622),
    [anon_sym___cdecl] = ACTIONS(2622),
    [anon_sym___clrcall] = ACTIONS(2622),
    [anon_sym___stdcall] = ACTIONS(2622),
    [anon_sym___fastcall] = ACTIONS(2622),
    [anon_sym___thiscall] = ACTIONS(2622),
    [anon_sym___vectorcall] = ACTIONS(2622),
    [anon_sym_signed] = ACTIONS(2622),
    [anon_sym_unsigned] = ACTIONS(2622),
    [anon_sym_long] = ACTIONS(2622),
    [anon_sym_short] = ACTIONS(2622),
    [anon_sym_static] = ACTIONS(2622),
    [anon_sym_auto] = ACTIONS(2622),
    [anon_sym_register] = ACTIONS(2622),
    [anon_sym_inline] = ACTIONS(2622),
    [anon_sym___inline] = ACTIONS(2622),
    [anon_sym___inline__] = ACTIONS(2622),
    [anon_sym___forceinline] = ACTIONS(2622),
    [anon_sym_thread_local] = ACTIONS(2622),
    [anon_sym___thread] = ACTIONS(2622),
    [anon_sym_CG_EXTERN] = ACTIONS(2622),
    [anon_sym_CG_INLINE] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2622),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2622),
    [anon_sym_IBOutlet] = ACTIONS(2622),
    [anon_sym_IBInspectable] = ACTIONS(2622),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2622),
    [anon_sym_NS_INLINE] = ACTIONS(2622),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2622),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2622),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2622),
    [anon_sym_const] = ACTIONS(2622),
    [anon_sym_constexpr] = ACTIONS(2622),
    [anon_sym_volatile] = ACTIONS(2622),
    [anon_sym_restrict] = ACTIONS(2622),
    [anon_sym___restrict__] = ACTIONS(2622),
    [anon_sym__Atomic] = ACTIONS(2622),
    [anon_sym__Noreturn] = ACTIONS(2622),
    [anon_sym__Nonnull] = ACTIONS(2622),
    [anon_sym_nullable] = ACTIONS(2622),
    [anon_sym__Complex] = ACTIONS(2622),
    [anon_sym__Nullable] = ACTIONS(2622),
    [anon_sym__Nullable_result] = ACTIONS(2622),
    [anon_sym__Null_unspecified] = ACTIONS(2622),
    [anon_sym___autoreleasing] = ACTIONS(2622),
    [anon_sym___block] = ACTIONS(2622),
    [anon_sym___bridge] = ACTIONS(2622),
    [anon_sym___bridge_retained] = ACTIONS(2622),
    [anon_sym___bridge_transfer] = ACTIONS(2622),
    [anon_sym___complex] = ACTIONS(2622),
    [anon_sym___const] = ACTIONS(2622),
    [anon_sym___imag] = ACTIONS(2622),
    [anon_sym___kindof] = ACTIONS(2622),
    [anon_sym___nonnull] = ACTIONS(2622),
    [anon_sym___nullable] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2622),
    [anon_sym___real] = ACTIONS(2622),
    [anon_sym___strong] = ACTIONS(2622),
    [anon_sym___unsafe_unretained] = ACTIONS(2622),
    [anon_sym___unused] = ACTIONS(2622),
    [anon_sym___weak] = ACTIONS(2622),
    [anon_sym_alignas] = ACTIONS(2622),
    [anon_sym__Alignas] = ACTIONS(2622),
    [sym_primitive_type] = ACTIONS(2622),
    [anon_sym_enum] = ACTIONS(2622),
    [anon_sym_struct] = ACTIONS(2622),
    [anon_sym_union] = ACTIONS(2622),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2622),
    [anon_sym___typeof] = ACTIONS(2622),
    [anon_sym_typeof] = ACTIONS(2622),
    [aux_sym_preproc_undef_token1] = ACTIONS(2622),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2622),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2622),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2622),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_API_AVAILABLE] = ACTIONS(2622),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_API_DEPRECATED] = ACTIONS(2622),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym___deprecated_msg] = ACTIONS(2622),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2622),
    [anon_sym_ATsynthesize] = ACTIONS(2624),
    [anon_sym_ATdynamic] = ACTIONS(2624),
    [anon_sym_BOOL] = ACTIONS(2622),
    [anon_sym_IMP] = ACTIONS(2622),
    [anon_sym_SEL] = ACTIONS(2622),
    [anon_sym_Class] = ACTIONS(2622),
    [anon_sym_id] = ACTIONS(2622),
  },
  [STATE(2374)] = {
    [sym_identifier] = ACTIONS(5369),
    [aux_sym_preproc_def_token1] = ACTIONS(5369),
    [aux_sym_preproc_if_token1] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5369),
    [sym_preproc_directive] = ACTIONS(5369),
    [anon_sym_DASH] = ACTIONS(5371),
    [anon_sym_PLUS] = ACTIONS(5371),
    [anon_sym___extension__] = ACTIONS(5369),
    [anon_sym_typedef] = ACTIONS(5369),
    [anon_sym_extern] = ACTIONS(5369),
    [anon_sym___attribute__] = ACTIONS(5369),
    [anon_sym___attribute] = ACTIONS(5369),
    [anon_sym_noreturn] = ACTIONS(5369),
    [anon_sym_LBRACK] = ACTIONS(5371),
    [anon_sym___declspec] = ACTIONS(5369),
    [anon_sym___cdecl] = ACTIONS(5369),
    [anon_sym___clrcall] = ACTIONS(5369),
    [anon_sym___stdcall] = ACTIONS(5369),
    [anon_sym___fastcall] = ACTIONS(5369),
    [anon_sym___thiscall] = ACTIONS(5369),
    [anon_sym___vectorcall] = ACTIONS(5369),
    [anon_sym_signed] = ACTIONS(5369),
    [anon_sym_unsigned] = ACTIONS(5369),
    [anon_sym_long] = ACTIONS(5369),
    [anon_sym_short] = ACTIONS(5369),
    [anon_sym_static] = ACTIONS(5369),
    [anon_sym_auto] = ACTIONS(5369),
    [anon_sym_register] = ACTIONS(5369),
    [anon_sym_inline] = ACTIONS(5369),
    [anon_sym___inline] = ACTIONS(5369),
    [anon_sym___inline__] = ACTIONS(5369),
    [anon_sym___forceinline] = ACTIONS(5369),
    [anon_sym_thread_local] = ACTIONS(5369),
    [anon_sym___thread] = ACTIONS(5369),
    [anon_sym_CG_EXTERN] = ACTIONS(5369),
    [anon_sym_CG_INLINE] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5369),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5369),
    [anon_sym_IBOutlet] = ACTIONS(5369),
    [anon_sym_IBInspectable] = ACTIONS(5369),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5369),
    [anon_sym_NS_INLINE] = ACTIONS(5369),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5369),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5369),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5369),
    [anon_sym_const] = ACTIONS(5369),
    [anon_sym_constexpr] = ACTIONS(5369),
    [anon_sym_volatile] = ACTIONS(5369),
    [anon_sym_restrict] = ACTIONS(5369),
    [anon_sym___restrict__] = ACTIONS(5369),
    [anon_sym__Atomic] = ACTIONS(5369),
    [anon_sym__Noreturn] = ACTIONS(5369),
    [anon_sym__Nonnull] = ACTIONS(5369),
    [anon_sym_nullable] = ACTIONS(5369),
    [anon_sym__Complex] = ACTIONS(5369),
    [anon_sym__Nullable] = ACTIONS(5369),
    [anon_sym__Nullable_result] = ACTIONS(5369),
    [anon_sym__Null_unspecified] = ACTIONS(5369),
    [anon_sym___autoreleasing] = ACTIONS(5369),
    [anon_sym___block] = ACTIONS(5369),
    [anon_sym___bridge] = ACTIONS(5369),
    [anon_sym___bridge_retained] = ACTIONS(5369),
    [anon_sym___bridge_transfer] = ACTIONS(5369),
    [anon_sym___complex] = ACTIONS(5369),
    [anon_sym___const] = ACTIONS(5369),
    [anon_sym___imag] = ACTIONS(5369),
    [anon_sym___kindof] = ACTIONS(5369),
    [anon_sym___nonnull] = ACTIONS(5369),
    [anon_sym___nullable] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5369),
    [anon_sym___real] = ACTIONS(5369),
    [anon_sym___strong] = ACTIONS(5369),
    [anon_sym___unsafe_unretained] = ACTIONS(5369),
    [anon_sym___unused] = ACTIONS(5369),
    [anon_sym___weak] = ACTIONS(5369),
    [anon_sym_alignas] = ACTIONS(5369),
    [anon_sym__Alignas] = ACTIONS(5369),
    [sym_primitive_type] = ACTIONS(5369),
    [anon_sym_enum] = ACTIONS(5369),
    [anon_sym_struct] = ACTIONS(5369),
    [anon_sym_union] = ACTIONS(5369),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5369),
    [anon_sym___typeof] = ACTIONS(5369),
    [anon_sym_typeof] = ACTIONS(5369),
    [aux_sym_preproc_undef_token1] = ACTIONS(5369),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5369),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5369),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5369),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_API_AVAILABLE] = ACTIONS(5369),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_API_DEPRECATED] = ACTIONS(5369),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym___deprecated_msg] = ACTIONS(5369),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5369),
    [anon_sym_ATend] = ACTIONS(5371),
    [anon_sym_ATsynthesize] = ACTIONS(5371),
    [anon_sym_ATdynamic] = ACTIONS(5371),
    [anon_sym_BOOL] = ACTIONS(5369),
    [anon_sym_IMP] = ACTIONS(5369),
    [anon_sym_SEL] = ACTIONS(5369),
    [anon_sym_Class] = ACTIONS(5369),
    [anon_sym_id] = ACTIONS(5369),
  },
  [STATE(2375)] = {
    [sym_identifier] = ACTIONS(2626),
    [aux_sym_preproc_def_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token2] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2626),
    [sym_preproc_directive] = ACTIONS(2626),
    [anon_sym_DASH] = ACTIONS(2628),
    [anon_sym_PLUS] = ACTIONS(2628),
    [anon_sym___extension__] = ACTIONS(2626),
    [anon_sym_typedef] = ACTIONS(2626),
    [anon_sym_extern] = ACTIONS(2626),
    [anon_sym___attribute__] = ACTIONS(2626),
    [anon_sym___attribute] = ACTIONS(2626),
    [anon_sym_noreturn] = ACTIONS(2626),
    [anon_sym_LBRACK] = ACTIONS(2628),
    [anon_sym___declspec] = ACTIONS(2626),
    [anon_sym___cdecl] = ACTIONS(2626),
    [anon_sym___clrcall] = ACTIONS(2626),
    [anon_sym___stdcall] = ACTIONS(2626),
    [anon_sym___fastcall] = ACTIONS(2626),
    [anon_sym___thiscall] = ACTIONS(2626),
    [anon_sym___vectorcall] = ACTIONS(2626),
    [anon_sym_signed] = ACTIONS(2626),
    [anon_sym_unsigned] = ACTIONS(2626),
    [anon_sym_long] = ACTIONS(2626),
    [anon_sym_short] = ACTIONS(2626),
    [anon_sym_static] = ACTIONS(2626),
    [anon_sym_auto] = ACTIONS(2626),
    [anon_sym_register] = ACTIONS(2626),
    [anon_sym_inline] = ACTIONS(2626),
    [anon_sym___inline] = ACTIONS(2626),
    [anon_sym___inline__] = ACTIONS(2626),
    [anon_sym___forceinline] = ACTIONS(2626),
    [anon_sym_thread_local] = ACTIONS(2626),
    [anon_sym___thread] = ACTIONS(2626),
    [anon_sym_CG_EXTERN] = ACTIONS(2626),
    [anon_sym_CG_INLINE] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2626),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2626),
    [anon_sym_IBOutlet] = ACTIONS(2626),
    [anon_sym_IBInspectable] = ACTIONS(2626),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2626),
    [anon_sym_NS_INLINE] = ACTIONS(2626),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2626),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2626),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2626),
    [anon_sym_const] = ACTIONS(2626),
    [anon_sym_constexpr] = ACTIONS(2626),
    [anon_sym_volatile] = ACTIONS(2626),
    [anon_sym_restrict] = ACTIONS(2626),
    [anon_sym___restrict__] = ACTIONS(2626),
    [anon_sym__Atomic] = ACTIONS(2626),
    [anon_sym__Noreturn] = ACTIONS(2626),
    [anon_sym__Nonnull] = ACTIONS(2626),
    [anon_sym_nullable] = ACTIONS(2626),
    [anon_sym__Complex] = ACTIONS(2626),
    [anon_sym__Nullable] = ACTIONS(2626),
    [anon_sym__Nullable_result] = ACTIONS(2626),
    [anon_sym__Null_unspecified] = ACTIONS(2626),
    [anon_sym___autoreleasing] = ACTIONS(2626),
    [anon_sym___block] = ACTIONS(2626),
    [anon_sym___bridge] = ACTIONS(2626),
    [anon_sym___bridge_retained] = ACTIONS(2626),
    [anon_sym___bridge_transfer] = ACTIONS(2626),
    [anon_sym___complex] = ACTIONS(2626),
    [anon_sym___const] = ACTIONS(2626),
    [anon_sym___imag] = ACTIONS(2626),
    [anon_sym___kindof] = ACTIONS(2626),
    [anon_sym___nonnull] = ACTIONS(2626),
    [anon_sym___nullable] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2626),
    [anon_sym___real] = ACTIONS(2626),
    [anon_sym___strong] = ACTIONS(2626),
    [anon_sym___unsafe_unretained] = ACTIONS(2626),
    [anon_sym___unused] = ACTIONS(2626),
    [anon_sym___weak] = ACTIONS(2626),
    [anon_sym_alignas] = ACTIONS(2626),
    [anon_sym__Alignas] = ACTIONS(2626),
    [sym_primitive_type] = ACTIONS(2626),
    [anon_sym_enum] = ACTIONS(2626),
    [anon_sym_struct] = ACTIONS(2626),
    [anon_sym_union] = ACTIONS(2626),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2626),
    [anon_sym___typeof] = ACTIONS(2626),
    [anon_sym_typeof] = ACTIONS(2626),
    [aux_sym_preproc_undef_token1] = ACTIONS(2626),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2626),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2626),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2626),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_API_AVAILABLE] = ACTIONS(2626),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_API_DEPRECATED] = ACTIONS(2626),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym___deprecated_msg] = ACTIONS(2626),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2626),
    [anon_sym_ATsynthesize] = ACTIONS(2628),
    [anon_sym_ATdynamic] = ACTIONS(2628),
    [anon_sym_BOOL] = ACTIONS(2626),
    [anon_sym_IMP] = ACTIONS(2626),
    [anon_sym_SEL] = ACTIONS(2626),
    [anon_sym_Class] = ACTIONS(2626),
    [anon_sym_id] = ACTIONS(2626),
  },
  [STATE(2376)] = {
    [sym_identifier] = ACTIONS(5475),
    [aux_sym_preproc_def_token1] = ACTIONS(5475),
    [aux_sym_preproc_if_token1] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5475),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5475),
    [sym_preproc_directive] = ACTIONS(5475),
    [anon_sym_DASH] = ACTIONS(5477),
    [anon_sym_PLUS] = ACTIONS(5477),
    [anon_sym___extension__] = ACTIONS(5475),
    [anon_sym_typedef] = ACTIONS(5475),
    [anon_sym_extern] = ACTIONS(5475),
    [anon_sym___attribute__] = ACTIONS(5475),
    [anon_sym___attribute] = ACTIONS(5475),
    [anon_sym_noreturn] = ACTIONS(5475),
    [anon_sym_LBRACK] = ACTIONS(5477),
    [anon_sym___declspec] = ACTIONS(5475),
    [anon_sym___cdecl] = ACTIONS(5475),
    [anon_sym___clrcall] = ACTIONS(5475),
    [anon_sym___stdcall] = ACTIONS(5475),
    [anon_sym___fastcall] = ACTIONS(5475),
    [anon_sym___thiscall] = ACTIONS(5475),
    [anon_sym___vectorcall] = ACTIONS(5475),
    [anon_sym_signed] = ACTIONS(5475),
    [anon_sym_unsigned] = ACTIONS(5475),
    [anon_sym_long] = ACTIONS(5475),
    [anon_sym_short] = ACTIONS(5475),
    [anon_sym_static] = ACTIONS(5475),
    [anon_sym_auto] = ACTIONS(5475),
    [anon_sym_register] = ACTIONS(5475),
    [anon_sym_inline] = ACTIONS(5475),
    [anon_sym___inline] = ACTIONS(5475),
    [anon_sym___inline__] = ACTIONS(5475),
    [anon_sym___forceinline] = ACTIONS(5475),
    [anon_sym_thread_local] = ACTIONS(5475),
    [anon_sym___thread] = ACTIONS(5475),
    [anon_sym_CG_EXTERN] = ACTIONS(5475),
    [anon_sym_CG_INLINE] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5475),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5475),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5475),
    [anon_sym_IBOutlet] = ACTIONS(5475),
    [anon_sym_IBInspectable] = ACTIONS(5475),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5475),
    [anon_sym_NS_INLINE] = ACTIONS(5475),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5475),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5475),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5475),
    [anon_sym_const] = ACTIONS(5475),
    [anon_sym_constexpr] = ACTIONS(5475),
    [anon_sym_volatile] = ACTIONS(5475),
    [anon_sym_restrict] = ACTIONS(5475),
    [anon_sym___restrict__] = ACTIONS(5475),
    [anon_sym__Atomic] = ACTIONS(5475),
    [anon_sym__Noreturn] = ACTIONS(5475),
    [anon_sym__Nonnull] = ACTIONS(5475),
    [anon_sym_nullable] = ACTIONS(5475),
    [anon_sym__Complex] = ACTIONS(5475),
    [anon_sym__Nullable] = ACTIONS(5475),
    [anon_sym__Nullable_result] = ACTIONS(5475),
    [anon_sym__Null_unspecified] = ACTIONS(5475),
    [anon_sym___autoreleasing] = ACTIONS(5475),
    [anon_sym___block] = ACTIONS(5475),
    [anon_sym___bridge] = ACTIONS(5475),
    [anon_sym___bridge_retained] = ACTIONS(5475),
    [anon_sym___bridge_transfer] = ACTIONS(5475),
    [anon_sym___complex] = ACTIONS(5475),
    [anon_sym___const] = ACTIONS(5475),
    [anon_sym___imag] = ACTIONS(5475),
    [anon_sym___kindof] = ACTIONS(5475),
    [anon_sym___nonnull] = ACTIONS(5475),
    [anon_sym___nullable] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5475),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5475),
    [anon_sym___real] = ACTIONS(5475),
    [anon_sym___strong] = ACTIONS(5475),
    [anon_sym___unsafe_unretained] = ACTIONS(5475),
    [anon_sym___unused] = ACTIONS(5475),
    [anon_sym___weak] = ACTIONS(5475),
    [anon_sym_alignas] = ACTIONS(5475),
    [anon_sym__Alignas] = ACTIONS(5475),
    [sym_primitive_type] = ACTIONS(5475),
    [anon_sym_enum] = ACTIONS(5475),
    [anon_sym_struct] = ACTIONS(5475),
    [anon_sym_union] = ACTIONS(5475),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5475),
    [anon_sym___typeof] = ACTIONS(5475),
    [anon_sym_typeof] = ACTIONS(5475),
    [aux_sym_preproc_undef_token1] = ACTIONS(5475),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5475),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5475),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5475),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5475),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5475),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_API_AVAILABLE] = ACTIONS(5475),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_API_DEPRECATED] = ACTIONS(5475),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5475),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5475),
    [anon_sym___deprecated_msg] = ACTIONS(5475),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5475),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5475),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5475),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5475),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5475),
    [anon_sym_ATend] = ACTIONS(5477),
    [anon_sym_ATsynthesize] = ACTIONS(5477),
    [anon_sym_ATdynamic] = ACTIONS(5477),
    [anon_sym_BOOL] = ACTIONS(5475),
    [anon_sym_IMP] = ACTIONS(5475),
    [anon_sym_SEL] = ACTIONS(5475),
    [anon_sym_Class] = ACTIONS(5475),
    [anon_sym_id] = ACTIONS(5475),
  },
  [STATE(2377)] = {
    [sym_identifier] = ACTIONS(5369),
    [aux_sym_preproc_def_token1] = ACTIONS(5369),
    [aux_sym_preproc_if_token1] = ACTIONS(5369),
    [aux_sym_preproc_if_token2] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5369),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5369),
    [sym_preproc_directive] = ACTIONS(5369),
    [anon_sym_DASH] = ACTIONS(5371),
    [anon_sym_PLUS] = ACTIONS(5371),
    [anon_sym___extension__] = ACTIONS(5369),
    [anon_sym_typedef] = ACTIONS(5369),
    [anon_sym_extern] = ACTIONS(5369),
    [anon_sym___attribute__] = ACTIONS(5369),
    [anon_sym___attribute] = ACTIONS(5369),
    [anon_sym_noreturn] = ACTIONS(5369),
    [anon_sym_LBRACK] = ACTIONS(5371),
    [anon_sym___declspec] = ACTIONS(5369),
    [anon_sym___cdecl] = ACTIONS(5369),
    [anon_sym___clrcall] = ACTIONS(5369),
    [anon_sym___stdcall] = ACTIONS(5369),
    [anon_sym___fastcall] = ACTIONS(5369),
    [anon_sym___thiscall] = ACTIONS(5369),
    [anon_sym___vectorcall] = ACTIONS(5369),
    [anon_sym_signed] = ACTIONS(5369),
    [anon_sym_unsigned] = ACTIONS(5369),
    [anon_sym_long] = ACTIONS(5369),
    [anon_sym_short] = ACTIONS(5369),
    [anon_sym_static] = ACTIONS(5369),
    [anon_sym_auto] = ACTIONS(5369),
    [anon_sym_register] = ACTIONS(5369),
    [anon_sym_inline] = ACTIONS(5369),
    [anon_sym___inline] = ACTIONS(5369),
    [anon_sym___inline__] = ACTIONS(5369),
    [anon_sym___forceinline] = ACTIONS(5369),
    [anon_sym_thread_local] = ACTIONS(5369),
    [anon_sym___thread] = ACTIONS(5369),
    [anon_sym_CG_EXTERN] = ACTIONS(5369),
    [anon_sym_CG_INLINE] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5369),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5369),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5369),
    [anon_sym_IBOutlet] = ACTIONS(5369),
    [anon_sym_IBInspectable] = ACTIONS(5369),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5369),
    [anon_sym_NS_INLINE] = ACTIONS(5369),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5369),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5369),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5369),
    [anon_sym_const] = ACTIONS(5369),
    [anon_sym_constexpr] = ACTIONS(5369),
    [anon_sym_volatile] = ACTIONS(5369),
    [anon_sym_restrict] = ACTIONS(5369),
    [anon_sym___restrict__] = ACTIONS(5369),
    [anon_sym__Atomic] = ACTIONS(5369),
    [anon_sym__Noreturn] = ACTIONS(5369),
    [anon_sym__Nonnull] = ACTIONS(5369),
    [anon_sym_nullable] = ACTIONS(5369),
    [anon_sym__Complex] = ACTIONS(5369),
    [anon_sym__Nullable] = ACTIONS(5369),
    [anon_sym__Nullable_result] = ACTIONS(5369),
    [anon_sym__Null_unspecified] = ACTIONS(5369),
    [anon_sym___autoreleasing] = ACTIONS(5369),
    [anon_sym___block] = ACTIONS(5369),
    [anon_sym___bridge] = ACTIONS(5369),
    [anon_sym___bridge_retained] = ACTIONS(5369),
    [anon_sym___bridge_transfer] = ACTIONS(5369),
    [anon_sym___complex] = ACTIONS(5369),
    [anon_sym___const] = ACTIONS(5369),
    [anon_sym___imag] = ACTIONS(5369),
    [anon_sym___kindof] = ACTIONS(5369),
    [anon_sym___nonnull] = ACTIONS(5369),
    [anon_sym___nullable] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5369),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5369),
    [anon_sym___real] = ACTIONS(5369),
    [anon_sym___strong] = ACTIONS(5369),
    [anon_sym___unsafe_unretained] = ACTIONS(5369),
    [anon_sym___unused] = ACTIONS(5369),
    [anon_sym___weak] = ACTIONS(5369),
    [anon_sym_alignas] = ACTIONS(5369),
    [anon_sym__Alignas] = ACTIONS(5369),
    [sym_primitive_type] = ACTIONS(5369),
    [anon_sym_enum] = ACTIONS(5369),
    [anon_sym_struct] = ACTIONS(5369),
    [anon_sym_union] = ACTIONS(5369),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5369),
    [anon_sym___typeof] = ACTIONS(5369),
    [anon_sym_typeof] = ACTIONS(5369),
    [aux_sym_preproc_undef_token1] = ACTIONS(5369),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5369),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5369),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5369),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5369),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5369),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_API_AVAILABLE] = ACTIONS(5369),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_API_DEPRECATED] = ACTIONS(5369),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5369),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5369),
    [anon_sym___deprecated_msg] = ACTIONS(5369),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5369),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5369),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5369),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5369),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5369),
    [anon_sym_ATsynthesize] = ACTIONS(5371),
    [anon_sym_ATdynamic] = ACTIONS(5371),
    [anon_sym_BOOL] = ACTIONS(5369),
    [anon_sym_IMP] = ACTIONS(5369),
    [anon_sym_SEL] = ACTIONS(5369),
    [anon_sym_Class] = ACTIONS(5369),
    [anon_sym_id] = ACTIONS(5369),
  },
  [STATE(2378)] = {
    [sym_identifier] = ACTIONS(5375),
    [aux_sym_preproc_def_token1] = ACTIONS(5375),
    [aux_sym_preproc_if_token1] = ACTIONS(5375),
    [aux_sym_preproc_if_token2] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5375),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5375),
    [sym_preproc_directive] = ACTIONS(5375),
    [anon_sym_DASH] = ACTIONS(5377),
    [anon_sym_PLUS] = ACTIONS(5377),
    [anon_sym___extension__] = ACTIONS(5375),
    [anon_sym_typedef] = ACTIONS(5375),
    [anon_sym_extern] = ACTIONS(5375),
    [anon_sym___attribute__] = ACTIONS(5375),
    [anon_sym___attribute] = ACTIONS(5375),
    [anon_sym_noreturn] = ACTIONS(5375),
    [anon_sym_LBRACK] = ACTIONS(5377),
    [anon_sym___declspec] = ACTIONS(5375),
    [anon_sym___cdecl] = ACTIONS(5375),
    [anon_sym___clrcall] = ACTIONS(5375),
    [anon_sym___stdcall] = ACTIONS(5375),
    [anon_sym___fastcall] = ACTIONS(5375),
    [anon_sym___thiscall] = ACTIONS(5375),
    [anon_sym___vectorcall] = ACTIONS(5375),
    [anon_sym_signed] = ACTIONS(5375),
    [anon_sym_unsigned] = ACTIONS(5375),
    [anon_sym_long] = ACTIONS(5375),
    [anon_sym_short] = ACTIONS(5375),
    [anon_sym_static] = ACTIONS(5375),
    [anon_sym_auto] = ACTIONS(5375),
    [anon_sym_register] = ACTIONS(5375),
    [anon_sym_inline] = ACTIONS(5375),
    [anon_sym___inline] = ACTIONS(5375),
    [anon_sym___inline__] = ACTIONS(5375),
    [anon_sym___forceinline] = ACTIONS(5375),
    [anon_sym_thread_local] = ACTIONS(5375),
    [anon_sym___thread] = ACTIONS(5375),
    [anon_sym_CG_EXTERN] = ACTIONS(5375),
    [anon_sym_CG_INLINE] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5375),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5375),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5375),
    [anon_sym_IBOutlet] = ACTIONS(5375),
    [anon_sym_IBInspectable] = ACTIONS(5375),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5375),
    [anon_sym_NS_INLINE] = ACTIONS(5375),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5375),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5375),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5375),
    [anon_sym_const] = ACTIONS(5375),
    [anon_sym_constexpr] = ACTIONS(5375),
    [anon_sym_volatile] = ACTIONS(5375),
    [anon_sym_restrict] = ACTIONS(5375),
    [anon_sym___restrict__] = ACTIONS(5375),
    [anon_sym__Atomic] = ACTIONS(5375),
    [anon_sym__Noreturn] = ACTIONS(5375),
    [anon_sym__Nonnull] = ACTIONS(5375),
    [anon_sym_nullable] = ACTIONS(5375),
    [anon_sym__Complex] = ACTIONS(5375),
    [anon_sym__Nullable] = ACTIONS(5375),
    [anon_sym__Nullable_result] = ACTIONS(5375),
    [anon_sym__Null_unspecified] = ACTIONS(5375),
    [anon_sym___autoreleasing] = ACTIONS(5375),
    [anon_sym___block] = ACTIONS(5375),
    [anon_sym___bridge] = ACTIONS(5375),
    [anon_sym___bridge_retained] = ACTIONS(5375),
    [anon_sym___bridge_transfer] = ACTIONS(5375),
    [anon_sym___complex] = ACTIONS(5375),
    [anon_sym___const] = ACTIONS(5375),
    [anon_sym___imag] = ACTIONS(5375),
    [anon_sym___kindof] = ACTIONS(5375),
    [anon_sym___nonnull] = ACTIONS(5375),
    [anon_sym___nullable] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5375),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5375),
    [anon_sym___real] = ACTIONS(5375),
    [anon_sym___strong] = ACTIONS(5375),
    [anon_sym___unsafe_unretained] = ACTIONS(5375),
    [anon_sym___unused] = ACTIONS(5375),
    [anon_sym___weak] = ACTIONS(5375),
    [anon_sym_alignas] = ACTIONS(5375),
    [anon_sym__Alignas] = ACTIONS(5375),
    [sym_primitive_type] = ACTIONS(5375),
    [anon_sym_enum] = ACTIONS(5375),
    [anon_sym_struct] = ACTIONS(5375),
    [anon_sym_union] = ACTIONS(5375),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5375),
    [anon_sym___typeof] = ACTIONS(5375),
    [anon_sym_typeof] = ACTIONS(5375),
    [aux_sym_preproc_undef_token1] = ACTIONS(5375),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5375),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5375),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5375),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5375),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5375),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_API_AVAILABLE] = ACTIONS(5375),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_API_DEPRECATED] = ACTIONS(5375),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5375),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5375),
    [anon_sym___deprecated_msg] = ACTIONS(5375),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5375),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5375),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5375),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5375),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5375),
    [anon_sym_ATsynthesize] = ACTIONS(5377),
    [anon_sym_ATdynamic] = ACTIONS(5377),
    [anon_sym_BOOL] = ACTIONS(5375),
    [anon_sym_IMP] = ACTIONS(5375),
    [anon_sym_SEL] = ACTIONS(5375),
    [anon_sym_Class] = ACTIONS(5375),
    [anon_sym_id] = ACTIONS(5375),
  },
  [STATE(2379)] = {
    [sym_identifier] = ACTIONS(5505),
    [aux_sym_preproc_def_token1] = ACTIONS(5505),
    [aux_sym_preproc_if_token1] = ACTIONS(5505),
    [aux_sym_preproc_if_token2] = ACTIONS(5505),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5505),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5505),
    [sym_preproc_directive] = ACTIONS(5505),
    [anon_sym_DASH] = ACTIONS(5507),
    [anon_sym_PLUS] = ACTIONS(5507),
    [anon_sym___extension__] = ACTIONS(5505),
    [anon_sym_typedef] = ACTIONS(5505),
    [anon_sym_extern] = ACTIONS(5505),
    [anon_sym___attribute__] = ACTIONS(5505),
    [anon_sym___attribute] = ACTIONS(5505),
    [anon_sym_noreturn] = ACTIONS(5505),
    [anon_sym_LBRACK] = ACTIONS(5507),
    [anon_sym___declspec] = ACTIONS(5505),
    [anon_sym___cdecl] = ACTIONS(5505),
    [anon_sym___clrcall] = ACTIONS(5505),
    [anon_sym___stdcall] = ACTIONS(5505),
    [anon_sym___fastcall] = ACTIONS(5505),
    [anon_sym___thiscall] = ACTIONS(5505),
    [anon_sym___vectorcall] = ACTIONS(5505),
    [anon_sym_signed] = ACTIONS(5505),
    [anon_sym_unsigned] = ACTIONS(5505),
    [anon_sym_long] = ACTIONS(5505),
    [anon_sym_short] = ACTIONS(5505),
    [anon_sym_static] = ACTIONS(5505),
    [anon_sym_auto] = ACTIONS(5505),
    [anon_sym_register] = ACTIONS(5505),
    [anon_sym_inline] = ACTIONS(5505),
    [anon_sym___inline] = ACTIONS(5505),
    [anon_sym___inline__] = ACTIONS(5505),
    [anon_sym___forceinline] = ACTIONS(5505),
    [anon_sym_thread_local] = ACTIONS(5505),
    [anon_sym___thread] = ACTIONS(5505),
    [anon_sym_CG_EXTERN] = ACTIONS(5505),
    [anon_sym_CG_INLINE] = ACTIONS(5505),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5505),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5505),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5505),
    [anon_sym_IBOutlet] = ACTIONS(5505),
    [anon_sym_IBInspectable] = ACTIONS(5505),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5505),
    [anon_sym_NS_INLINE] = ACTIONS(5505),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5505),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5505),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5505),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5505),
    [anon_sym_const] = ACTIONS(5505),
    [anon_sym_constexpr] = ACTIONS(5505),
    [anon_sym_volatile] = ACTIONS(5505),
    [anon_sym_restrict] = ACTIONS(5505),
    [anon_sym___restrict__] = ACTIONS(5505),
    [anon_sym__Atomic] = ACTIONS(5505),
    [anon_sym__Noreturn] = ACTIONS(5505),
    [anon_sym__Nonnull] = ACTIONS(5505),
    [anon_sym_nullable] = ACTIONS(5505),
    [anon_sym__Complex] = ACTIONS(5505),
    [anon_sym__Nullable] = ACTIONS(5505),
    [anon_sym__Nullable_result] = ACTIONS(5505),
    [anon_sym__Null_unspecified] = ACTIONS(5505),
    [anon_sym___autoreleasing] = ACTIONS(5505),
    [anon_sym___block] = ACTIONS(5505),
    [anon_sym___bridge] = ACTIONS(5505),
    [anon_sym___bridge_retained] = ACTIONS(5505),
    [anon_sym___bridge_transfer] = ACTIONS(5505),
    [anon_sym___complex] = ACTIONS(5505),
    [anon_sym___const] = ACTIONS(5505),
    [anon_sym___imag] = ACTIONS(5505),
    [anon_sym___kindof] = ACTIONS(5505),
    [anon_sym___nonnull] = ACTIONS(5505),
    [anon_sym___nullable] = ACTIONS(5505),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5505),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5505),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5505),
    [anon_sym___real] = ACTIONS(5505),
    [anon_sym___strong] = ACTIONS(5505),
    [anon_sym___unsafe_unretained] = ACTIONS(5505),
    [anon_sym___unused] = ACTIONS(5505),
    [anon_sym___weak] = ACTIONS(5505),
    [anon_sym_alignas] = ACTIONS(5505),
    [anon_sym__Alignas] = ACTIONS(5505),
    [sym_primitive_type] = ACTIONS(5505),
    [anon_sym_enum] = ACTIONS(5505),
    [anon_sym_struct] = ACTIONS(5505),
    [anon_sym_union] = ACTIONS(5505),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5505),
    [anon_sym___typeof] = ACTIONS(5505),
    [anon_sym_typeof] = ACTIONS(5505),
    [aux_sym_preproc_undef_token1] = ACTIONS(5505),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5505),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5505),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5505),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5505),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5505),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5505),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5505),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5505),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5505),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_API_AVAILABLE] = ACTIONS(5505),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_API_DEPRECATED] = ACTIONS(5505),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5505),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5505),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5505),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5505),
    [anon_sym___deprecated_msg] = ACTIONS(5505),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5505),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5505),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5505),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5505),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5505),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5505),
    [anon_sym_ATsynthesize] = ACTIONS(5507),
    [anon_sym_ATdynamic] = ACTIONS(5507),
    [anon_sym_BOOL] = ACTIONS(5505),
    [anon_sym_IMP] = ACTIONS(5505),
    [anon_sym_SEL] = ACTIONS(5505),
    [anon_sym_Class] = ACTIONS(5505),
    [anon_sym_id] = ACTIONS(5505),
  },
  [STATE(2380)] = {
    [sym_identifier] = ACTIONS(2622),
    [aux_sym_preproc_def_token1] = ACTIONS(2622),
    [aux_sym_preproc_if_token1] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2622),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2622),
    [sym_preproc_directive] = ACTIONS(2622),
    [anon_sym_DASH] = ACTIONS(2624),
    [anon_sym_PLUS] = ACTIONS(2624),
    [anon_sym___extension__] = ACTIONS(2622),
    [anon_sym_typedef] = ACTIONS(2622),
    [anon_sym_extern] = ACTIONS(2622),
    [anon_sym___attribute__] = ACTIONS(2622),
    [anon_sym___attribute] = ACTIONS(2622),
    [anon_sym_noreturn] = ACTIONS(2622),
    [anon_sym_LBRACK] = ACTIONS(2624),
    [anon_sym___declspec] = ACTIONS(2622),
    [anon_sym___cdecl] = ACTIONS(2622),
    [anon_sym___clrcall] = ACTIONS(2622),
    [anon_sym___stdcall] = ACTIONS(2622),
    [anon_sym___fastcall] = ACTIONS(2622),
    [anon_sym___thiscall] = ACTIONS(2622),
    [anon_sym___vectorcall] = ACTIONS(2622),
    [anon_sym_signed] = ACTIONS(2622),
    [anon_sym_unsigned] = ACTIONS(2622),
    [anon_sym_long] = ACTIONS(2622),
    [anon_sym_short] = ACTIONS(2622),
    [anon_sym_static] = ACTIONS(2622),
    [anon_sym_auto] = ACTIONS(2622),
    [anon_sym_register] = ACTIONS(2622),
    [anon_sym_inline] = ACTIONS(2622),
    [anon_sym___inline] = ACTIONS(2622),
    [anon_sym___inline__] = ACTIONS(2622),
    [anon_sym___forceinline] = ACTIONS(2622),
    [anon_sym_thread_local] = ACTIONS(2622),
    [anon_sym___thread] = ACTIONS(2622),
    [anon_sym_CG_EXTERN] = ACTIONS(2622),
    [anon_sym_CG_INLINE] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2622),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2622),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2622),
    [anon_sym_IBOutlet] = ACTIONS(2622),
    [anon_sym_IBInspectable] = ACTIONS(2622),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2622),
    [anon_sym_NS_INLINE] = ACTIONS(2622),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2622),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2622),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2622),
    [anon_sym_const] = ACTIONS(2622),
    [anon_sym_constexpr] = ACTIONS(2622),
    [anon_sym_volatile] = ACTIONS(2622),
    [anon_sym_restrict] = ACTIONS(2622),
    [anon_sym___restrict__] = ACTIONS(2622),
    [anon_sym__Atomic] = ACTIONS(2622),
    [anon_sym__Noreturn] = ACTIONS(2622),
    [anon_sym__Nonnull] = ACTIONS(2622),
    [anon_sym_nullable] = ACTIONS(2622),
    [anon_sym__Complex] = ACTIONS(2622),
    [anon_sym__Nullable] = ACTIONS(2622),
    [anon_sym__Nullable_result] = ACTIONS(2622),
    [anon_sym__Null_unspecified] = ACTIONS(2622),
    [anon_sym___autoreleasing] = ACTIONS(2622),
    [anon_sym___block] = ACTIONS(2622),
    [anon_sym___bridge] = ACTIONS(2622),
    [anon_sym___bridge_retained] = ACTIONS(2622),
    [anon_sym___bridge_transfer] = ACTIONS(2622),
    [anon_sym___complex] = ACTIONS(2622),
    [anon_sym___const] = ACTIONS(2622),
    [anon_sym___imag] = ACTIONS(2622),
    [anon_sym___kindof] = ACTIONS(2622),
    [anon_sym___nonnull] = ACTIONS(2622),
    [anon_sym___nullable] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2622),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2622),
    [anon_sym___real] = ACTIONS(2622),
    [anon_sym___strong] = ACTIONS(2622),
    [anon_sym___unsafe_unretained] = ACTIONS(2622),
    [anon_sym___unused] = ACTIONS(2622),
    [anon_sym___weak] = ACTIONS(2622),
    [anon_sym_alignas] = ACTIONS(2622),
    [anon_sym__Alignas] = ACTIONS(2622),
    [sym_primitive_type] = ACTIONS(2622),
    [anon_sym_enum] = ACTIONS(2622),
    [anon_sym_struct] = ACTIONS(2622),
    [anon_sym_union] = ACTIONS(2622),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2622),
    [anon_sym___typeof] = ACTIONS(2622),
    [anon_sym_typeof] = ACTIONS(2622),
    [aux_sym_preproc_undef_token1] = ACTIONS(2622),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2622),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2622),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2622),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2622),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2622),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_API_AVAILABLE] = ACTIONS(2622),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_API_DEPRECATED] = ACTIONS(2622),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2622),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2622),
    [anon_sym___deprecated_msg] = ACTIONS(2622),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2622),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2622),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2622),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2622),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2622),
    [anon_sym_ATend] = ACTIONS(2624),
    [anon_sym_ATsynthesize] = ACTIONS(2624),
    [anon_sym_ATdynamic] = ACTIONS(2624),
    [anon_sym_BOOL] = ACTIONS(2622),
    [anon_sym_IMP] = ACTIONS(2622),
    [anon_sym_SEL] = ACTIONS(2622),
    [anon_sym_Class] = ACTIONS(2622),
    [anon_sym_id] = ACTIONS(2622),
  },
  [STATE(2381)] = {
    [sym_identifier] = ACTIONS(5682),
    [anon_sym_COMMA] = ACTIONS(5684),
    [anon_sym_RPAREN] = ACTIONS(5684),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5684),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5684),
    [anon_sym_LPAREN2] = ACTIONS(5684),
    [anon_sym_STAR] = ACTIONS(5684),
    [anon_sym_CARET] = ACTIONS(5684),
    [anon_sym_GT] = ACTIONS(5684),
    [anon_sym_SEMI] = ACTIONS(5684),
    [anon_sym___extension__] = ACTIONS(5682),
    [anon_sym_extern] = ACTIONS(5682),
    [anon_sym___attribute__] = ACTIONS(5682),
    [anon_sym___attribute] = ACTIONS(5682),
    [anon_sym_noreturn] = ACTIONS(5682),
    [anon_sym_LBRACK] = ACTIONS(5684),
    [anon_sym___declspec] = ACTIONS(5682),
    [anon_sym___based] = ACTIONS(5682),
    [anon_sym_LBRACE] = ACTIONS(5684),
    [anon_sym_RBRACE] = ACTIONS(5684),
    [anon_sym_signed] = ACTIONS(5682),
    [anon_sym_unsigned] = ACTIONS(5682),
    [anon_sym_long] = ACTIONS(5682),
    [anon_sym_short] = ACTIONS(5682),
    [anon_sym_static] = ACTIONS(5682),
    [anon_sym_ATautoreleasepool] = ACTIONS(5684),
    [anon_sym_auto] = ACTIONS(5682),
    [anon_sym_register] = ACTIONS(5682),
    [anon_sym_inline] = ACTIONS(5682),
    [anon_sym___inline] = ACTIONS(5682),
    [anon_sym___inline__] = ACTIONS(5682),
    [anon_sym___forceinline] = ACTIONS(5682),
    [anon_sym_thread_local] = ACTIONS(5682),
    [anon_sym___thread] = ACTIONS(5682),
    [anon_sym_CG_EXTERN] = ACTIONS(5682),
    [anon_sym_CG_INLINE] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5682),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5682),
    [anon_sym_IBOutlet] = ACTIONS(5682),
    [anon_sym_IBInspectable] = ACTIONS(5682),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5682),
    [anon_sym_NS_INLINE] = ACTIONS(5682),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5682),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5682),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5682),
    [anon_sym_const] = ACTIONS(5682),
    [anon_sym_constexpr] = ACTIONS(5682),
    [anon_sym_volatile] = ACTIONS(5682),
    [anon_sym_restrict] = ACTIONS(5682),
    [anon_sym___restrict__] = ACTIONS(5682),
    [anon_sym__Atomic] = ACTIONS(5682),
    [anon_sym__Noreturn] = ACTIONS(5682),
    [anon_sym__Nonnull] = ACTIONS(5682),
    [anon_sym_nullable] = ACTIONS(5682),
    [anon_sym__Complex] = ACTIONS(5682),
    [anon_sym__Nullable] = ACTIONS(5682),
    [anon_sym__Nullable_result] = ACTIONS(5682),
    [anon_sym__Null_unspecified] = ACTIONS(5682),
    [anon_sym___autoreleasing] = ACTIONS(5682),
    [anon_sym___block] = ACTIONS(5682),
    [anon_sym___bridge] = ACTIONS(5682),
    [anon_sym___bridge_retained] = ACTIONS(5682),
    [anon_sym___bridge_transfer] = ACTIONS(5682),
    [anon_sym___complex] = ACTIONS(5682),
    [anon_sym___const] = ACTIONS(5682),
    [anon_sym___imag] = ACTIONS(5682),
    [anon_sym___kindof] = ACTIONS(5682),
    [anon_sym___nonnull] = ACTIONS(5682),
    [anon_sym___nullable] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5682),
    [anon_sym___real] = ACTIONS(5682),
    [anon_sym___strong] = ACTIONS(5682),
    [anon_sym___unsafe_unretained] = ACTIONS(5682),
    [anon_sym___unused] = ACTIONS(5682),
    [anon_sym___weak] = ACTIONS(5682),
    [anon_sym_alignas] = ACTIONS(5682),
    [anon_sym__Alignas] = ACTIONS(5682),
    [sym_primitive_type] = ACTIONS(5682),
    [anon_sym_enum] = ACTIONS(5682),
    [anon_sym_COLON] = ACTIONS(5684),
    [anon_sym_struct] = ACTIONS(5682),
    [anon_sym_union] = ACTIONS(5682),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5682),
    [anon_sym___typeof] = ACTIONS(5682),
    [anon_sym_typeof] = ACTIONS(5682),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5682),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5682),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5682),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_API_AVAILABLE] = ACTIONS(5682),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_API_DEPRECATED] = ACTIONS(5682),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym___deprecated_msg] = ACTIONS(5682),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5682),
    [anon_sym_ATprotocol] = ACTIONS(5684),
    [anon_sym_ATinterface] = ACTIONS(5684),
    [anon_sym_ATimplementation] = ACTIONS(5684),
    [anon_sym_ATproperty] = ACTIONS(5684),
    [anon_sym_BOOL] = ACTIONS(5682),
    [anon_sym_IMP] = ACTIONS(5682),
    [anon_sym_SEL] = ACTIONS(5682),
    [anon_sym_Class] = ACTIONS(5682),
    [anon_sym_id] = ACTIONS(5682),
  },
  [STATE(2382)] = {
    [sym_identifier] = ACTIONS(2626),
    [aux_sym_preproc_def_token1] = ACTIONS(2626),
    [aux_sym_preproc_if_token1] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2626),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2626),
    [sym_preproc_directive] = ACTIONS(2626),
    [anon_sym_DASH] = ACTIONS(2628),
    [anon_sym_PLUS] = ACTIONS(2628),
    [anon_sym___extension__] = ACTIONS(2626),
    [anon_sym_typedef] = ACTIONS(2626),
    [anon_sym_extern] = ACTIONS(2626),
    [anon_sym___attribute__] = ACTIONS(2626),
    [anon_sym___attribute] = ACTIONS(2626),
    [anon_sym_noreturn] = ACTIONS(2626),
    [anon_sym_LBRACK] = ACTIONS(2628),
    [anon_sym___declspec] = ACTIONS(2626),
    [anon_sym___cdecl] = ACTIONS(2626),
    [anon_sym___clrcall] = ACTIONS(2626),
    [anon_sym___stdcall] = ACTIONS(2626),
    [anon_sym___fastcall] = ACTIONS(2626),
    [anon_sym___thiscall] = ACTIONS(2626),
    [anon_sym___vectorcall] = ACTIONS(2626),
    [anon_sym_signed] = ACTIONS(2626),
    [anon_sym_unsigned] = ACTIONS(2626),
    [anon_sym_long] = ACTIONS(2626),
    [anon_sym_short] = ACTIONS(2626),
    [anon_sym_static] = ACTIONS(2626),
    [anon_sym_auto] = ACTIONS(2626),
    [anon_sym_register] = ACTIONS(2626),
    [anon_sym_inline] = ACTIONS(2626),
    [anon_sym___inline] = ACTIONS(2626),
    [anon_sym___inline__] = ACTIONS(2626),
    [anon_sym___forceinline] = ACTIONS(2626),
    [anon_sym_thread_local] = ACTIONS(2626),
    [anon_sym___thread] = ACTIONS(2626),
    [anon_sym_CG_EXTERN] = ACTIONS(2626),
    [anon_sym_CG_INLINE] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2626),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2626),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2626),
    [anon_sym_IBOutlet] = ACTIONS(2626),
    [anon_sym_IBInspectable] = ACTIONS(2626),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2626),
    [anon_sym_NS_INLINE] = ACTIONS(2626),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2626),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2626),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2626),
    [anon_sym_const] = ACTIONS(2626),
    [anon_sym_constexpr] = ACTIONS(2626),
    [anon_sym_volatile] = ACTIONS(2626),
    [anon_sym_restrict] = ACTIONS(2626),
    [anon_sym___restrict__] = ACTIONS(2626),
    [anon_sym__Atomic] = ACTIONS(2626),
    [anon_sym__Noreturn] = ACTIONS(2626),
    [anon_sym__Nonnull] = ACTIONS(2626),
    [anon_sym_nullable] = ACTIONS(2626),
    [anon_sym__Complex] = ACTIONS(2626),
    [anon_sym__Nullable] = ACTIONS(2626),
    [anon_sym__Nullable_result] = ACTIONS(2626),
    [anon_sym__Null_unspecified] = ACTIONS(2626),
    [anon_sym___autoreleasing] = ACTIONS(2626),
    [anon_sym___block] = ACTIONS(2626),
    [anon_sym___bridge] = ACTIONS(2626),
    [anon_sym___bridge_retained] = ACTIONS(2626),
    [anon_sym___bridge_transfer] = ACTIONS(2626),
    [anon_sym___complex] = ACTIONS(2626),
    [anon_sym___const] = ACTIONS(2626),
    [anon_sym___imag] = ACTIONS(2626),
    [anon_sym___kindof] = ACTIONS(2626),
    [anon_sym___nonnull] = ACTIONS(2626),
    [anon_sym___nullable] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2626),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2626),
    [anon_sym___real] = ACTIONS(2626),
    [anon_sym___strong] = ACTIONS(2626),
    [anon_sym___unsafe_unretained] = ACTIONS(2626),
    [anon_sym___unused] = ACTIONS(2626),
    [anon_sym___weak] = ACTIONS(2626),
    [anon_sym_alignas] = ACTIONS(2626),
    [anon_sym__Alignas] = ACTIONS(2626),
    [sym_primitive_type] = ACTIONS(2626),
    [anon_sym_enum] = ACTIONS(2626),
    [anon_sym_struct] = ACTIONS(2626),
    [anon_sym_union] = ACTIONS(2626),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2626),
    [anon_sym___typeof] = ACTIONS(2626),
    [anon_sym_typeof] = ACTIONS(2626),
    [aux_sym_preproc_undef_token1] = ACTIONS(2626),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2626),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2626),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2626),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2626),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2626),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_API_AVAILABLE] = ACTIONS(2626),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_API_DEPRECATED] = ACTIONS(2626),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2626),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2626),
    [anon_sym___deprecated_msg] = ACTIONS(2626),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2626),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2626),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2626),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2626),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2626),
    [anon_sym_ATend] = ACTIONS(2628),
    [anon_sym_ATsynthesize] = ACTIONS(2628),
    [anon_sym_ATdynamic] = ACTIONS(2628),
    [anon_sym_BOOL] = ACTIONS(2626),
    [anon_sym_IMP] = ACTIONS(2626),
    [anon_sym_SEL] = ACTIONS(2626),
    [anon_sym_Class] = ACTIONS(2626),
    [anon_sym_id] = ACTIONS(2626),
  },
  [STATE(2383)] = {
    [sym_identifier] = ACTIONS(2630),
    [aux_sym_preproc_def_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token2] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2630),
    [sym_preproc_directive] = ACTIONS(2630),
    [anon_sym_DASH] = ACTIONS(2632),
    [anon_sym_PLUS] = ACTIONS(2632),
    [anon_sym___extension__] = ACTIONS(2630),
    [anon_sym_typedef] = ACTIONS(2630),
    [anon_sym_extern] = ACTIONS(2630),
    [anon_sym___attribute__] = ACTIONS(2630),
    [anon_sym___attribute] = ACTIONS(2630),
    [anon_sym_noreturn] = ACTIONS(2630),
    [anon_sym_LBRACK] = ACTIONS(2632),
    [anon_sym___declspec] = ACTIONS(2630),
    [anon_sym___cdecl] = ACTIONS(2630),
    [anon_sym___clrcall] = ACTIONS(2630),
    [anon_sym___stdcall] = ACTIONS(2630),
    [anon_sym___fastcall] = ACTIONS(2630),
    [anon_sym___thiscall] = ACTIONS(2630),
    [anon_sym___vectorcall] = ACTIONS(2630),
    [anon_sym_signed] = ACTIONS(2630),
    [anon_sym_unsigned] = ACTIONS(2630),
    [anon_sym_long] = ACTIONS(2630),
    [anon_sym_short] = ACTIONS(2630),
    [anon_sym_static] = ACTIONS(2630),
    [anon_sym_auto] = ACTIONS(2630),
    [anon_sym_register] = ACTIONS(2630),
    [anon_sym_inline] = ACTIONS(2630),
    [anon_sym___inline] = ACTIONS(2630),
    [anon_sym___inline__] = ACTIONS(2630),
    [anon_sym___forceinline] = ACTIONS(2630),
    [anon_sym_thread_local] = ACTIONS(2630),
    [anon_sym___thread] = ACTIONS(2630),
    [anon_sym_CG_EXTERN] = ACTIONS(2630),
    [anon_sym_CG_INLINE] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2630),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2630),
    [anon_sym_IBOutlet] = ACTIONS(2630),
    [anon_sym_IBInspectable] = ACTIONS(2630),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2630),
    [anon_sym_NS_INLINE] = ACTIONS(2630),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2630),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2630),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2630),
    [anon_sym_const] = ACTIONS(2630),
    [anon_sym_constexpr] = ACTIONS(2630),
    [anon_sym_volatile] = ACTIONS(2630),
    [anon_sym_restrict] = ACTIONS(2630),
    [anon_sym___restrict__] = ACTIONS(2630),
    [anon_sym__Atomic] = ACTIONS(2630),
    [anon_sym__Noreturn] = ACTIONS(2630),
    [anon_sym__Nonnull] = ACTIONS(2630),
    [anon_sym_nullable] = ACTIONS(2630),
    [anon_sym__Complex] = ACTIONS(2630),
    [anon_sym__Nullable] = ACTIONS(2630),
    [anon_sym__Nullable_result] = ACTIONS(2630),
    [anon_sym__Null_unspecified] = ACTIONS(2630),
    [anon_sym___autoreleasing] = ACTIONS(2630),
    [anon_sym___block] = ACTIONS(2630),
    [anon_sym___bridge] = ACTIONS(2630),
    [anon_sym___bridge_retained] = ACTIONS(2630),
    [anon_sym___bridge_transfer] = ACTIONS(2630),
    [anon_sym___complex] = ACTIONS(2630),
    [anon_sym___const] = ACTIONS(2630),
    [anon_sym___imag] = ACTIONS(2630),
    [anon_sym___kindof] = ACTIONS(2630),
    [anon_sym___nonnull] = ACTIONS(2630),
    [anon_sym___nullable] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2630),
    [anon_sym___real] = ACTIONS(2630),
    [anon_sym___strong] = ACTIONS(2630),
    [anon_sym___unsafe_unretained] = ACTIONS(2630),
    [anon_sym___unused] = ACTIONS(2630),
    [anon_sym___weak] = ACTIONS(2630),
    [anon_sym_alignas] = ACTIONS(2630),
    [anon_sym__Alignas] = ACTIONS(2630),
    [sym_primitive_type] = ACTIONS(2630),
    [anon_sym_enum] = ACTIONS(2630),
    [anon_sym_struct] = ACTIONS(2630),
    [anon_sym_union] = ACTIONS(2630),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2630),
    [anon_sym___typeof] = ACTIONS(2630),
    [anon_sym_typeof] = ACTIONS(2630),
    [aux_sym_preproc_undef_token1] = ACTIONS(2630),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2630),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2630),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2630),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_API_AVAILABLE] = ACTIONS(2630),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_API_DEPRECATED] = ACTIONS(2630),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym___deprecated_msg] = ACTIONS(2630),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2630),
    [anon_sym_ATsynthesize] = ACTIONS(2632),
    [anon_sym_ATdynamic] = ACTIONS(2632),
    [anon_sym_BOOL] = ACTIONS(2630),
    [anon_sym_IMP] = ACTIONS(2630),
    [anon_sym_SEL] = ACTIONS(2630),
    [anon_sym_Class] = ACTIONS(2630),
    [anon_sym_id] = ACTIONS(2630),
  },
  [STATE(2384)] = {
    [sym_identifier] = ACTIONS(5381),
    [aux_sym_preproc_def_token1] = ACTIONS(5381),
    [aux_sym_preproc_if_token1] = ACTIONS(5381),
    [aux_sym_preproc_if_token2] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5381),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5381),
    [sym_preproc_directive] = ACTIONS(5381),
    [anon_sym_DASH] = ACTIONS(5383),
    [anon_sym_PLUS] = ACTIONS(5383),
    [anon_sym___extension__] = ACTIONS(5381),
    [anon_sym_typedef] = ACTIONS(5381),
    [anon_sym_extern] = ACTIONS(5381),
    [anon_sym___attribute__] = ACTIONS(5381),
    [anon_sym___attribute] = ACTIONS(5381),
    [anon_sym_noreturn] = ACTIONS(5381),
    [anon_sym_LBRACK] = ACTIONS(5383),
    [anon_sym___declspec] = ACTIONS(5381),
    [anon_sym___cdecl] = ACTIONS(5381),
    [anon_sym___clrcall] = ACTIONS(5381),
    [anon_sym___stdcall] = ACTIONS(5381),
    [anon_sym___fastcall] = ACTIONS(5381),
    [anon_sym___thiscall] = ACTIONS(5381),
    [anon_sym___vectorcall] = ACTIONS(5381),
    [anon_sym_signed] = ACTIONS(5381),
    [anon_sym_unsigned] = ACTIONS(5381),
    [anon_sym_long] = ACTIONS(5381),
    [anon_sym_short] = ACTIONS(5381),
    [anon_sym_static] = ACTIONS(5381),
    [anon_sym_auto] = ACTIONS(5381),
    [anon_sym_register] = ACTIONS(5381),
    [anon_sym_inline] = ACTIONS(5381),
    [anon_sym___inline] = ACTIONS(5381),
    [anon_sym___inline__] = ACTIONS(5381),
    [anon_sym___forceinline] = ACTIONS(5381),
    [anon_sym_thread_local] = ACTIONS(5381),
    [anon_sym___thread] = ACTIONS(5381),
    [anon_sym_CG_EXTERN] = ACTIONS(5381),
    [anon_sym_CG_INLINE] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5381),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5381),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5381),
    [anon_sym_IBOutlet] = ACTIONS(5381),
    [anon_sym_IBInspectable] = ACTIONS(5381),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5381),
    [anon_sym_NS_INLINE] = ACTIONS(5381),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5381),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5381),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5381),
    [anon_sym_const] = ACTIONS(5381),
    [anon_sym_constexpr] = ACTIONS(5381),
    [anon_sym_volatile] = ACTIONS(5381),
    [anon_sym_restrict] = ACTIONS(5381),
    [anon_sym___restrict__] = ACTIONS(5381),
    [anon_sym__Atomic] = ACTIONS(5381),
    [anon_sym__Noreturn] = ACTIONS(5381),
    [anon_sym__Nonnull] = ACTIONS(5381),
    [anon_sym_nullable] = ACTIONS(5381),
    [anon_sym__Complex] = ACTIONS(5381),
    [anon_sym__Nullable] = ACTIONS(5381),
    [anon_sym__Nullable_result] = ACTIONS(5381),
    [anon_sym__Null_unspecified] = ACTIONS(5381),
    [anon_sym___autoreleasing] = ACTIONS(5381),
    [anon_sym___block] = ACTIONS(5381),
    [anon_sym___bridge] = ACTIONS(5381),
    [anon_sym___bridge_retained] = ACTIONS(5381),
    [anon_sym___bridge_transfer] = ACTIONS(5381),
    [anon_sym___complex] = ACTIONS(5381),
    [anon_sym___const] = ACTIONS(5381),
    [anon_sym___imag] = ACTIONS(5381),
    [anon_sym___kindof] = ACTIONS(5381),
    [anon_sym___nonnull] = ACTIONS(5381),
    [anon_sym___nullable] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5381),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5381),
    [anon_sym___real] = ACTIONS(5381),
    [anon_sym___strong] = ACTIONS(5381),
    [anon_sym___unsafe_unretained] = ACTIONS(5381),
    [anon_sym___unused] = ACTIONS(5381),
    [anon_sym___weak] = ACTIONS(5381),
    [anon_sym_alignas] = ACTIONS(5381),
    [anon_sym__Alignas] = ACTIONS(5381),
    [sym_primitive_type] = ACTIONS(5381),
    [anon_sym_enum] = ACTIONS(5381),
    [anon_sym_struct] = ACTIONS(5381),
    [anon_sym_union] = ACTIONS(5381),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5381),
    [anon_sym___typeof] = ACTIONS(5381),
    [anon_sym_typeof] = ACTIONS(5381),
    [aux_sym_preproc_undef_token1] = ACTIONS(5381),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5381),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5381),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5381),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5381),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5381),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_API_AVAILABLE] = ACTIONS(5381),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_API_DEPRECATED] = ACTIONS(5381),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5381),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5381),
    [anon_sym___deprecated_msg] = ACTIONS(5381),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5381),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5381),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5381),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5381),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5381),
    [anon_sym_ATsynthesize] = ACTIONS(5383),
    [anon_sym_ATdynamic] = ACTIONS(5383),
    [anon_sym_BOOL] = ACTIONS(5381),
    [anon_sym_IMP] = ACTIONS(5381),
    [anon_sym_SEL] = ACTIONS(5381),
    [anon_sym_Class] = ACTIONS(5381),
    [anon_sym_id] = ACTIONS(5381),
  },
  [STATE(2385)] = {
    [sym_identifier] = ACTIONS(5399),
    [aux_sym_preproc_def_token1] = ACTIONS(5399),
    [aux_sym_preproc_if_token1] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5399),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5399),
    [sym_preproc_directive] = ACTIONS(5399),
    [anon_sym_DASH] = ACTIONS(5401),
    [anon_sym_PLUS] = ACTIONS(5401),
    [anon_sym___extension__] = ACTIONS(5399),
    [anon_sym_typedef] = ACTIONS(5399),
    [anon_sym_extern] = ACTIONS(5399),
    [anon_sym___attribute__] = ACTIONS(5399),
    [anon_sym___attribute] = ACTIONS(5399),
    [anon_sym_noreturn] = ACTIONS(5399),
    [anon_sym_LBRACK] = ACTIONS(5401),
    [anon_sym___declspec] = ACTIONS(5399),
    [anon_sym___cdecl] = ACTIONS(5399),
    [anon_sym___clrcall] = ACTIONS(5399),
    [anon_sym___stdcall] = ACTIONS(5399),
    [anon_sym___fastcall] = ACTIONS(5399),
    [anon_sym___thiscall] = ACTIONS(5399),
    [anon_sym___vectorcall] = ACTIONS(5399),
    [anon_sym_signed] = ACTIONS(5399),
    [anon_sym_unsigned] = ACTIONS(5399),
    [anon_sym_long] = ACTIONS(5399),
    [anon_sym_short] = ACTIONS(5399),
    [anon_sym_static] = ACTIONS(5399),
    [anon_sym_auto] = ACTIONS(5399),
    [anon_sym_register] = ACTIONS(5399),
    [anon_sym_inline] = ACTIONS(5399),
    [anon_sym___inline] = ACTIONS(5399),
    [anon_sym___inline__] = ACTIONS(5399),
    [anon_sym___forceinline] = ACTIONS(5399),
    [anon_sym_thread_local] = ACTIONS(5399),
    [anon_sym___thread] = ACTIONS(5399),
    [anon_sym_CG_EXTERN] = ACTIONS(5399),
    [anon_sym_CG_INLINE] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5399),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5399),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5399),
    [anon_sym_IBOutlet] = ACTIONS(5399),
    [anon_sym_IBInspectable] = ACTIONS(5399),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5399),
    [anon_sym_NS_INLINE] = ACTIONS(5399),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5399),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5399),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5399),
    [anon_sym_const] = ACTIONS(5399),
    [anon_sym_constexpr] = ACTIONS(5399),
    [anon_sym_volatile] = ACTIONS(5399),
    [anon_sym_restrict] = ACTIONS(5399),
    [anon_sym___restrict__] = ACTIONS(5399),
    [anon_sym__Atomic] = ACTIONS(5399),
    [anon_sym__Noreturn] = ACTIONS(5399),
    [anon_sym__Nonnull] = ACTIONS(5399),
    [anon_sym_nullable] = ACTIONS(5399),
    [anon_sym__Complex] = ACTIONS(5399),
    [anon_sym__Nullable] = ACTIONS(5399),
    [anon_sym__Nullable_result] = ACTIONS(5399),
    [anon_sym__Null_unspecified] = ACTIONS(5399),
    [anon_sym___autoreleasing] = ACTIONS(5399),
    [anon_sym___block] = ACTIONS(5399),
    [anon_sym___bridge] = ACTIONS(5399),
    [anon_sym___bridge_retained] = ACTIONS(5399),
    [anon_sym___bridge_transfer] = ACTIONS(5399),
    [anon_sym___complex] = ACTIONS(5399),
    [anon_sym___const] = ACTIONS(5399),
    [anon_sym___imag] = ACTIONS(5399),
    [anon_sym___kindof] = ACTIONS(5399),
    [anon_sym___nonnull] = ACTIONS(5399),
    [anon_sym___nullable] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5399),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5399),
    [anon_sym___real] = ACTIONS(5399),
    [anon_sym___strong] = ACTIONS(5399),
    [anon_sym___unsafe_unretained] = ACTIONS(5399),
    [anon_sym___unused] = ACTIONS(5399),
    [anon_sym___weak] = ACTIONS(5399),
    [anon_sym_alignas] = ACTIONS(5399),
    [anon_sym__Alignas] = ACTIONS(5399),
    [sym_primitive_type] = ACTIONS(5399),
    [anon_sym_enum] = ACTIONS(5399),
    [anon_sym_struct] = ACTIONS(5399),
    [anon_sym_union] = ACTIONS(5399),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5399),
    [anon_sym___typeof] = ACTIONS(5399),
    [anon_sym_typeof] = ACTIONS(5399),
    [aux_sym_preproc_undef_token1] = ACTIONS(5399),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5399),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5399),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5399),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5399),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5399),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_API_AVAILABLE] = ACTIONS(5399),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_API_DEPRECATED] = ACTIONS(5399),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5399),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5399),
    [anon_sym___deprecated_msg] = ACTIONS(5399),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5399),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5399),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5399),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5399),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5399),
    [anon_sym_ATend] = ACTIONS(5401),
    [anon_sym_ATsynthesize] = ACTIONS(5401),
    [anon_sym_ATdynamic] = ACTIONS(5401),
    [anon_sym_BOOL] = ACTIONS(5399),
    [anon_sym_IMP] = ACTIONS(5399),
    [anon_sym_SEL] = ACTIONS(5399),
    [anon_sym_Class] = ACTIONS(5399),
    [anon_sym_id] = ACTIONS(5399),
  },
  [STATE(2386)] = {
    [sym_identifier] = ACTIONS(5068),
    [aux_sym_preproc_def_token1] = ACTIONS(5068),
    [aux_sym_preproc_if_token1] = ACTIONS(5068),
    [aux_sym_preproc_if_token2] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5068),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5068),
    [sym_preproc_directive] = ACTIONS(5068),
    [anon_sym_DASH] = ACTIONS(5072),
    [anon_sym_PLUS] = ACTIONS(5072),
    [anon_sym___extension__] = ACTIONS(5068),
    [anon_sym_typedef] = ACTIONS(5068),
    [anon_sym_extern] = ACTIONS(5068),
    [anon_sym___attribute__] = ACTIONS(5068),
    [anon_sym___attribute] = ACTIONS(5068),
    [anon_sym_noreturn] = ACTIONS(5068),
    [anon_sym_LBRACK] = ACTIONS(5072),
    [anon_sym___declspec] = ACTIONS(5068),
    [anon_sym___cdecl] = ACTIONS(5068),
    [anon_sym___clrcall] = ACTIONS(5068),
    [anon_sym___stdcall] = ACTIONS(5068),
    [anon_sym___fastcall] = ACTIONS(5068),
    [anon_sym___thiscall] = ACTIONS(5068),
    [anon_sym___vectorcall] = ACTIONS(5068),
    [anon_sym_signed] = ACTIONS(5068),
    [anon_sym_unsigned] = ACTIONS(5068),
    [anon_sym_long] = ACTIONS(5068),
    [anon_sym_short] = ACTIONS(5068),
    [anon_sym_static] = ACTIONS(5068),
    [anon_sym_auto] = ACTIONS(5068),
    [anon_sym_register] = ACTIONS(5068),
    [anon_sym_inline] = ACTIONS(5068),
    [anon_sym___inline] = ACTIONS(5068),
    [anon_sym___inline__] = ACTIONS(5068),
    [anon_sym___forceinline] = ACTIONS(5068),
    [anon_sym_thread_local] = ACTIONS(5068),
    [anon_sym___thread] = ACTIONS(5068),
    [anon_sym_CG_EXTERN] = ACTIONS(5068),
    [anon_sym_CG_INLINE] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5068),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5068),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5068),
    [anon_sym_IBOutlet] = ACTIONS(5068),
    [anon_sym_IBInspectable] = ACTIONS(5068),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5068),
    [anon_sym_NS_INLINE] = ACTIONS(5068),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5068),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5068),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5068),
    [anon_sym_const] = ACTIONS(5068),
    [anon_sym_constexpr] = ACTIONS(5068),
    [anon_sym_volatile] = ACTIONS(5068),
    [anon_sym_restrict] = ACTIONS(5068),
    [anon_sym___restrict__] = ACTIONS(5068),
    [anon_sym__Atomic] = ACTIONS(5068),
    [anon_sym__Noreturn] = ACTIONS(5068),
    [anon_sym__Nonnull] = ACTIONS(5068),
    [anon_sym_nullable] = ACTIONS(5068),
    [anon_sym__Complex] = ACTIONS(5068),
    [anon_sym__Nullable] = ACTIONS(5068),
    [anon_sym__Nullable_result] = ACTIONS(5068),
    [anon_sym__Null_unspecified] = ACTIONS(5068),
    [anon_sym___autoreleasing] = ACTIONS(5068),
    [anon_sym___block] = ACTIONS(5068),
    [anon_sym___bridge] = ACTIONS(5068),
    [anon_sym___bridge_retained] = ACTIONS(5068),
    [anon_sym___bridge_transfer] = ACTIONS(5068),
    [anon_sym___complex] = ACTIONS(5068),
    [anon_sym___const] = ACTIONS(5068),
    [anon_sym___imag] = ACTIONS(5068),
    [anon_sym___kindof] = ACTIONS(5068),
    [anon_sym___nonnull] = ACTIONS(5068),
    [anon_sym___nullable] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5068),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5068),
    [anon_sym___real] = ACTIONS(5068),
    [anon_sym___strong] = ACTIONS(5068),
    [anon_sym___unsafe_unretained] = ACTIONS(5068),
    [anon_sym___unused] = ACTIONS(5068),
    [anon_sym___weak] = ACTIONS(5068),
    [anon_sym_alignas] = ACTIONS(5068),
    [anon_sym__Alignas] = ACTIONS(5068),
    [sym_primitive_type] = ACTIONS(5068),
    [anon_sym_enum] = ACTIONS(5068),
    [anon_sym_struct] = ACTIONS(5068),
    [anon_sym_union] = ACTIONS(5068),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5068),
    [anon_sym___typeof] = ACTIONS(5068),
    [anon_sym_typeof] = ACTIONS(5068),
    [aux_sym_preproc_undef_token1] = ACTIONS(5068),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5068),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5068),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5068),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5068),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5068),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_API_AVAILABLE] = ACTIONS(5068),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_API_DEPRECATED] = ACTIONS(5068),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5068),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5068),
    [anon_sym___deprecated_msg] = ACTIONS(5068),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5068),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5068),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5068),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5068),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5068),
    [anon_sym_ATsynthesize] = ACTIONS(5072),
    [anon_sym_ATdynamic] = ACTIONS(5072),
    [anon_sym_BOOL] = ACTIONS(5068),
    [anon_sym_IMP] = ACTIONS(5068),
    [anon_sym_SEL] = ACTIONS(5068),
    [anon_sym_Class] = ACTIONS(5068),
    [anon_sym_id] = ACTIONS(5068),
  },
  [STATE(2387)] = {
    [sym_identifier] = ACTIONS(5405),
    [aux_sym_preproc_def_token1] = ACTIONS(5405),
    [aux_sym_preproc_if_token1] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5405),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5405),
    [sym_preproc_directive] = ACTIONS(5405),
    [anon_sym_DASH] = ACTIONS(5407),
    [anon_sym_PLUS] = ACTIONS(5407),
    [anon_sym___extension__] = ACTIONS(5405),
    [anon_sym_typedef] = ACTIONS(5405),
    [anon_sym_extern] = ACTIONS(5405),
    [anon_sym___attribute__] = ACTIONS(5405),
    [anon_sym___attribute] = ACTIONS(5405),
    [anon_sym_noreturn] = ACTIONS(5405),
    [anon_sym_LBRACK] = ACTIONS(5407),
    [anon_sym___declspec] = ACTIONS(5405),
    [anon_sym___cdecl] = ACTIONS(5405),
    [anon_sym___clrcall] = ACTIONS(5405),
    [anon_sym___stdcall] = ACTIONS(5405),
    [anon_sym___fastcall] = ACTIONS(5405),
    [anon_sym___thiscall] = ACTIONS(5405),
    [anon_sym___vectorcall] = ACTIONS(5405),
    [anon_sym_signed] = ACTIONS(5405),
    [anon_sym_unsigned] = ACTIONS(5405),
    [anon_sym_long] = ACTIONS(5405),
    [anon_sym_short] = ACTIONS(5405),
    [anon_sym_static] = ACTIONS(5405),
    [anon_sym_auto] = ACTIONS(5405),
    [anon_sym_register] = ACTIONS(5405),
    [anon_sym_inline] = ACTIONS(5405),
    [anon_sym___inline] = ACTIONS(5405),
    [anon_sym___inline__] = ACTIONS(5405),
    [anon_sym___forceinline] = ACTIONS(5405),
    [anon_sym_thread_local] = ACTIONS(5405),
    [anon_sym___thread] = ACTIONS(5405),
    [anon_sym_CG_EXTERN] = ACTIONS(5405),
    [anon_sym_CG_INLINE] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5405),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5405),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5405),
    [anon_sym_IBOutlet] = ACTIONS(5405),
    [anon_sym_IBInspectable] = ACTIONS(5405),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5405),
    [anon_sym_NS_INLINE] = ACTIONS(5405),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5405),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5405),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5405),
    [anon_sym_const] = ACTIONS(5405),
    [anon_sym_constexpr] = ACTIONS(5405),
    [anon_sym_volatile] = ACTIONS(5405),
    [anon_sym_restrict] = ACTIONS(5405),
    [anon_sym___restrict__] = ACTIONS(5405),
    [anon_sym__Atomic] = ACTIONS(5405),
    [anon_sym__Noreturn] = ACTIONS(5405),
    [anon_sym__Nonnull] = ACTIONS(5405),
    [anon_sym_nullable] = ACTIONS(5405),
    [anon_sym__Complex] = ACTIONS(5405),
    [anon_sym__Nullable] = ACTIONS(5405),
    [anon_sym__Nullable_result] = ACTIONS(5405),
    [anon_sym__Null_unspecified] = ACTIONS(5405),
    [anon_sym___autoreleasing] = ACTIONS(5405),
    [anon_sym___block] = ACTIONS(5405),
    [anon_sym___bridge] = ACTIONS(5405),
    [anon_sym___bridge_retained] = ACTIONS(5405),
    [anon_sym___bridge_transfer] = ACTIONS(5405),
    [anon_sym___complex] = ACTIONS(5405),
    [anon_sym___const] = ACTIONS(5405),
    [anon_sym___imag] = ACTIONS(5405),
    [anon_sym___kindof] = ACTIONS(5405),
    [anon_sym___nonnull] = ACTIONS(5405),
    [anon_sym___nullable] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5405),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5405),
    [anon_sym___real] = ACTIONS(5405),
    [anon_sym___strong] = ACTIONS(5405),
    [anon_sym___unsafe_unretained] = ACTIONS(5405),
    [anon_sym___unused] = ACTIONS(5405),
    [anon_sym___weak] = ACTIONS(5405),
    [anon_sym_alignas] = ACTIONS(5405),
    [anon_sym__Alignas] = ACTIONS(5405),
    [sym_primitive_type] = ACTIONS(5405),
    [anon_sym_enum] = ACTIONS(5405),
    [anon_sym_struct] = ACTIONS(5405),
    [anon_sym_union] = ACTIONS(5405),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5405),
    [anon_sym___typeof] = ACTIONS(5405),
    [anon_sym_typeof] = ACTIONS(5405),
    [aux_sym_preproc_undef_token1] = ACTIONS(5405),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5405),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5405),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5405),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5405),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5405),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_API_AVAILABLE] = ACTIONS(5405),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_API_DEPRECATED] = ACTIONS(5405),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5405),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5405),
    [anon_sym___deprecated_msg] = ACTIONS(5405),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5405),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5405),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5405),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5405),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5405),
    [anon_sym_ATend] = ACTIONS(5407),
    [anon_sym_ATsynthesize] = ACTIONS(5407),
    [anon_sym_ATdynamic] = ACTIONS(5407),
    [anon_sym_BOOL] = ACTIONS(5405),
    [anon_sym_IMP] = ACTIONS(5405),
    [anon_sym_SEL] = ACTIONS(5405),
    [anon_sym_Class] = ACTIONS(5405),
    [anon_sym_id] = ACTIONS(5405),
  },
  [STATE(2388)] = {
    [sym_identifier] = ACTIONS(5448),
    [aux_sym_preproc_def_token1] = ACTIONS(5448),
    [aux_sym_preproc_if_token1] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5448),
    [sym_preproc_directive] = ACTIONS(5448),
    [anon_sym_DASH] = ACTIONS(5450),
    [anon_sym_PLUS] = ACTIONS(5450),
    [anon_sym___extension__] = ACTIONS(5448),
    [anon_sym_typedef] = ACTIONS(5448),
    [anon_sym_extern] = ACTIONS(5448),
    [anon_sym___attribute__] = ACTIONS(5448),
    [anon_sym___attribute] = ACTIONS(5448),
    [anon_sym_noreturn] = ACTIONS(5448),
    [anon_sym_LBRACK] = ACTIONS(5450),
    [anon_sym___declspec] = ACTIONS(5448),
    [anon_sym___cdecl] = ACTIONS(5448),
    [anon_sym___clrcall] = ACTIONS(5448),
    [anon_sym___stdcall] = ACTIONS(5448),
    [anon_sym___fastcall] = ACTIONS(5448),
    [anon_sym___thiscall] = ACTIONS(5448),
    [anon_sym___vectorcall] = ACTIONS(5448),
    [anon_sym_signed] = ACTIONS(5448),
    [anon_sym_unsigned] = ACTIONS(5448),
    [anon_sym_long] = ACTIONS(5448),
    [anon_sym_short] = ACTIONS(5448),
    [anon_sym_static] = ACTIONS(5448),
    [anon_sym_auto] = ACTIONS(5448),
    [anon_sym_register] = ACTIONS(5448),
    [anon_sym_inline] = ACTIONS(5448),
    [anon_sym___inline] = ACTIONS(5448),
    [anon_sym___inline__] = ACTIONS(5448),
    [anon_sym___forceinline] = ACTIONS(5448),
    [anon_sym_thread_local] = ACTIONS(5448),
    [anon_sym___thread] = ACTIONS(5448),
    [anon_sym_CG_EXTERN] = ACTIONS(5448),
    [anon_sym_CG_INLINE] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5448),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5448),
    [anon_sym_IBOutlet] = ACTIONS(5448),
    [anon_sym_IBInspectable] = ACTIONS(5448),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5448),
    [anon_sym_NS_INLINE] = ACTIONS(5448),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5448),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5448),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5448),
    [anon_sym_const] = ACTIONS(5448),
    [anon_sym_constexpr] = ACTIONS(5448),
    [anon_sym_volatile] = ACTIONS(5448),
    [anon_sym_restrict] = ACTIONS(5448),
    [anon_sym___restrict__] = ACTIONS(5448),
    [anon_sym__Atomic] = ACTIONS(5448),
    [anon_sym__Noreturn] = ACTIONS(5448),
    [anon_sym__Nonnull] = ACTIONS(5448),
    [anon_sym_nullable] = ACTIONS(5448),
    [anon_sym__Complex] = ACTIONS(5448),
    [anon_sym__Nullable] = ACTIONS(5448),
    [anon_sym__Nullable_result] = ACTIONS(5448),
    [anon_sym__Null_unspecified] = ACTIONS(5448),
    [anon_sym___autoreleasing] = ACTIONS(5448),
    [anon_sym___block] = ACTIONS(5448),
    [anon_sym___bridge] = ACTIONS(5448),
    [anon_sym___bridge_retained] = ACTIONS(5448),
    [anon_sym___bridge_transfer] = ACTIONS(5448),
    [anon_sym___complex] = ACTIONS(5448),
    [anon_sym___const] = ACTIONS(5448),
    [anon_sym___imag] = ACTIONS(5448),
    [anon_sym___kindof] = ACTIONS(5448),
    [anon_sym___nonnull] = ACTIONS(5448),
    [anon_sym___nullable] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5448),
    [anon_sym___real] = ACTIONS(5448),
    [anon_sym___strong] = ACTIONS(5448),
    [anon_sym___unsafe_unretained] = ACTIONS(5448),
    [anon_sym___unused] = ACTIONS(5448),
    [anon_sym___weak] = ACTIONS(5448),
    [anon_sym_alignas] = ACTIONS(5448),
    [anon_sym__Alignas] = ACTIONS(5448),
    [sym_primitive_type] = ACTIONS(5448),
    [anon_sym_enum] = ACTIONS(5448),
    [anon_sym_struct] = ACTIONS(5448),
    [anon_sym_union] = ACTIONS(5448),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5448),
    [anon_sym___typeof] = ACTIONS(5448),
    [anon_sym_typeof] = ACTIONS(5448),
    [aux_sym_preproc_undef_token1] = ACTIONS(5448),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5448),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5448),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5448),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_API_AVAILABLE] = ACTIONS(5448),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_API_DEPRECATED] = ACTIONS(5448),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym___deprecated_msg] = ACTIONS(5448),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5448),
    [anon_sym_ATend] = ACTIONS(5450),
    [anon_sym_ATsynthesize] = ACTIONS(5450),
    [anon_sym_ATdynamic] = ACTIONS(5450),
    [anon_sym_BOOL] = ACTIONS(5448),
    [anon_sym_IMP] = ACTIONS(5448),
    [anon_sym_SEL] = ACTIONS(5448),
    [anon_sym_Class] = ACTIONS(5448),
    [anon_sym_id] = ACTIONS(5448),
  },
  [STATE(2389)] = {
    [sym_identifier] = ACTIONS(5517),
    [aux_sym_preproc_def_token1] = ACTIONS(5517),
    [aux_sym_preproc_if_token1] = ACTIONS(5517),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5517),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5517),
    [sym_preproc_directive] = ACTIONS(5517),
    [anon_sym_DASH] = ACTIONS(5519),
    [anon_sym_PLUS] = ACTIONS(5519),
    [anon_sym___extension__] = ACTIONS(5517),
    [anon_sym_typedef] = ACTIONS(5517),
    [anon_sym_extern] = ACTIONS(5517),
    [anon_sym___attribute__] = ACTIONS(5517),
    [anon_sym___attribute] = ACTIONS(5517),
    [anon_sym_noreturn] = ACTIONS(5517),
    [anon_sym_LBRACK] = ACTIONS(5519),
    [anon_sym___declspec] = ACTIONS(5517),
    [anon_sym___cdecl] = ACTIONS(5517),
    [anon_sym___clrcall] = ACTIONS(5517),
    [anon_sym___stdcall] = ACTIONS(5517),
    [anon_sym___fastcall] = ACTIONS(5517),
    [anon_sym___thiscall] = ACTIONS(5517),
    [anon_sym___vectorcall] = ACTIONS(5517),
    [anon_sym_signed] = ACTIONS(5517),
    [anon_sym_unsigned] = ACTIONS(5517),
    [anon_sym_long] = ACTIONS(5517),
    [anon_sym_short] = ACTIONS(5517),
    [anon_sym_static] = ACTIONS(5517),
    [anon_sym_auto] = ACTIONS(5517),
    [anon_sym_register] = ACTIONS(5517),
    [anon_sym_inline] = ACTIONS(5517),
    [anon_sym___inline] = ACTIONS(5517),
    [anon_sym___inline__] = ACTIONS(5517),
    [anon_sym___forceinline] = ACTIONS(5517),
    [anon_sym_thread_local] = ACTIONS(5517),
    [anon_sym___thread] = ACTIONS(5517),
    [anon_sym_CG_EXTERN] = ACTIONS(5517),
    [anon_sym_CG_INLINE] = ACTIONS(5517),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5517),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5517),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5517),
    [anon_sym_IBOutlet] = ACTIONS(5517),
    [anon_sym_IBInspectable] = ACTIONS(5517),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5517),
    [anon_sym_NS_INLINE] = ACTIONS(5517),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5517),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5517),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5517),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5517),
    [anon_sym_const] = ACTIONS(5517),
    [anon_sym_constexpr] = ACTIONS(5517),
    [anon_sym_volatile] = ACTIONS(5517),
    [anon_sym_restrict] = ACTIONS(5517),
    [anon_sym___restrict__] = ACTIONS(5517),
    [anon_sym__Atomic] = ACTIONS(5517),
    [anon_sym__Noreturn] = ACTIONS(5517),
    [anon_sym__Nonnull] = ACTIONS(5517),
    [anon_sym_nullable] = ACTIONS(5517),
    [anon_sym__Complex] = ACTIONS(5517),
    [anon_sym__Nullable] = ACTIONS(5517),
    [anon_sym__Nullable_result] = ACTIONS(5517),
    [anon_sym__Null_unspecified] = ACTIONS(5517),
    [anon_sym___autoreleasing] = ACTIONS(5517),
    [anon_sym___block] = ACTIONS(5517),
    [anon_sym___bridge] = ACTIONS(5517),
    [anon_sym___bridge_retained] = ACTIONS(5517),
    [anon_sym___bridge_transfer] = ACTIONS(5517),
    [anon_sym___complex] = ACTIONS(5517),
    [anon_sym___const] = ACTIONS(5517),
    [anon_sym___imag] = ACTIONS(5517),
    [anon_sym___kindof] = ACTIONS(5517),
    [anon_sym___nonnull] = ACTIONS(5517),
    [anon_sym___nullable] = ACTIONS(5517),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5517),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5517),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5517),
    [anon_sym___real] = ACTIONS(5517),
    [anon_sym___strong] = ACTIONS(5517),
    [anon_sym___unsafe_unretained] = ACTIONS(5517),
    [anon_sym___unused] = ACTIONS(5517),
    [anon_sym___weak] = ACTIONS(5517),
    [anon_sym_alignas] = ACTIONS(5517),
    [anon_sym__Alignas] = ACTIONS(5517),
    [sym_primitive_type] = ACTIONS(5517),
    [anon_sym_enum] = ACTIONS(5517),
    [anon_sym_struct] = ACTIONS(5517),
    [anon_sym_union] = ACTIONS(5517),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5517),
    [anon_sym___typeof] = ACTIONS(5517),
    [anon_sym_typeof] = ACTIONS(5517),
    [aux_sym_preproc_undef_token1] = ACTIONS(5517),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5517),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5517),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5517),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5517),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5517),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5517),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5517),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5517),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5517),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_API_AVAILABLE] = ACTIONS(5517),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_API_DEPRECATED] = ACTIONS(5517),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5517),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5517),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5517),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5517),
    [anon_sym___deprecated_msg] = ACTIONS(5517),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5517),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5517),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5517),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5517),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5517),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5517),
    [anon_sym_ATend] = ACTIONS(5519),
    [anon_sym_ATsynthesize] = ACTIONS(5519),
    [anon_sym_ATdynamic] = ACTIONS(5519),
    [anon_sym_BOOL] = ACTIONS(5517),
    [anon_sym_IMP] = ACTIONS(5517),
    [anon_sym_SEL] = ACTIONS(5517),
    [anon_sym_Class] = ACTIONS(5517),
    [anon_sym_id] = ACTIONS(5517),
  },
  [STATE(2390)] = {
    [sym_identifier] = ACTIONS(5359),
    [aux_sym_preproc_def_token1] = ACTIONS(5359),
    [aux_sym_preproc_if_token1] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5359),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5359),
    [sym_preproc_directive] = ACTIONS(5359),
    [anon_sym_DASH] = ACTIONS(5361),
    [anon_sym_PLUS] = ACTIONS(5361),
    [anon_sym___extension__] = ACTIONS(5359),
    [anon_sym_typedef] = ACTIONS(5359),
    [anon_sym_extern] = ACTIONS(5359),
    [anon_sym___attribute__] = ACTIONS(5359),
    [anon_sym___attribute] = ACTIONS(5359),
    [anon_sym_noreturn] = ACTIONS(5359),
    [anon_sym_LBRACK] = ACTIONS(5361),
    [anon_sym___declspec] = ACTIONS(5359),
    [anon_sym___cdecl] = ACTIONS(5359),
    [anon_sym___clrcall] = ACTIONS(5359),
    [anon_sym___stdcall] = ACTIONS(5359),
    [anon_sym___fastcall] = ACTIONS(5359),
    [anon_sym___thiscall] = ACTIONS(5359),
    [anon_sym___vectorcall] = ACTIONS(5359),
    [anon_sym_signed] = ACTIONS(5359),
    [anon_sym_unsigned] = ACTIONS(5359),
    [anon_sym_long] = ACTIONS(5359),
    [anon_sym_short] = ACTIONS(5359),
    [anon_sym_static] = ACTIONS(5359),
    [anon_sym_auto] = ACTIONS(5359),
    [anon_sym_register] = ACTIONS(5359),
    [anon_sym_inline] = ACTIONS(5359),
    [anon_sym___inline] = ACTIONS(5359),
    [anon_sym___inline__] = ACTIONS(5359),
    [anon_sym___forceinline] = ACTIONS(5359),
    [anon_sym_thread_local] = ACTIONS(5359),
    [anon_sym___thread] = ACTIONS(5359),
    [anon_sym_CG_EXTERN] = ACTIONS(5359),
    [anon_sym_CG_INLINE] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5359),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5359),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5359),
    [anon_sym_IBOutlet] = ACTIONS(5359),
    [anon_sym_IBInspectable] = ACTIONS(5359),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5359),
    [anon_sym_NS_INLINE] = ACTIONS(5359),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5359),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5359),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5359),
    [anon_sym_const] = ACTIONS(5359),
    [anon_sym_constexpr] = ACTIONS(5359),
    [anon_sym_volatile] = ACTIONS(5359),
    [anon_sym_restrict] = ACTIONS(5359),
    [anon_sym___restrict__] = ACTIONS(5359),
    [anon_sym__Atomic] = ACTIONS(5359),
    [anon_sym__Noreturn] = ACTIONS(5359),
    [anon_sym__Nonnull] = ACTIONS(5359),
    [anon_sym_nullable] = ACTIONS(5359),
    [anon_sym__Complex] = ACTIONS(5359),
    [anon_sym__Nullable] = ACTIONS(5359),
    [anon_sym__Nullable_result] = ACTIONS(5359),
    [anon_sym__Null_unspecified] = ACTIONS(5359),
    [anon_sym___autoreleasing] = ACTIONS(5359),
    [anon_sym___block] = ACTIONS(5359),
    [anon_sym___bridge] = ACTIONS(5359),
    [anon_sym___bridge_retained] = ACTIONS(5359),
    [anon_sym___bridge_transfer] = ACTIONS(5359),
    [anon_sym___complex] = ACTIONS(5359),
    [anon_sym___const] = ACTIONS(5359),
    [anon_sym___imag] = ACTIONS(5359),
    [anon_sym___kindof] = ACTIONS(5359),
    [anon_sym___nonnull] = ACTIONS(5359),
    [anon_sym___nullable] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5359),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5359),
    [anon_sym___real] = ACTIONS(5359),
    [anon_sym___strong] = ACTIONS(5359),
    [anon_sym___unsafe_unretained] = ACTIONS(5359),
    [anon_sym___unused] = ACTIONS(5359),
    [anon_sym___weak] = ACTIONS(5359),
    [anon_sym_alignas] = ACTIONS(5359),
    [anon_sym__Alignas] = ACTIONS(5359),
    [sym_primitive_type] = ACTIONS(5359),
    [anon_sym_enum] = ACTIONS(5359),
    [anon_sym_struct] = ACTIONS(5359),
    [anon_sym_union] = ACTIONS(5359),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5359),
    [anon_sym___typeof] = ACTIONS(5359),
    [anon_sym_typeof] = ACTIONS(5359),
    [aux_sym_preproc_undef_token1] = ACTIONS(5359),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5359),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5359),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5359),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5359),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5359),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_API_AVAILABLE] = ACTIONS(5359),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_API_DEPRECATED] = ACTIONS(5359),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5359),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5359),
    [anon_sym___deprecated_msg] = ACTIONS(5359),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5359),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5359),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5359),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5359),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5359),
    [anon_sym_ATend] = ACTIONS(5361),
    [anon_sym_ATsynthesize] = ACTIONS(5361),
    [anon_sym_ATdynamic] = ACTIONS(5361),
    [anon_sym_BOOL] = ACTIONS(5359),
    [anon_sym_IMP] = ACTIONS(5359),
    [anon_sym_SEL] = ACTIONS(5359),
    [anon_sym_Class] = ACTIONS(5359),
    [anon_sym_id] = ACTIONS(5359),
  },
  [STATE(2391)] = {
    [sym_identifier] = ACTIONS(5521),
    [aux_sym_preproc_def_token1] = ACTIONS(5521),
    [aux_sym_preproc_if_token1] = ACTIONS(5521),
    [aux_sym_preproc_if_token2] = ACTIONS(5521),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5521),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5521),
    [sym_preproc_directive] = ACTIONS(5521),
    [anon_sym_DASH] = ACTIONS(5523),
    [anon_sym_PLUS] = ACTIONS(5523),
    [anon_sym___extension__] = ACTIONS(5521),
    [anon_sym_typedef] = ACTIONS(5521),
    [anon_sym_extern] = ACTIONS(5521),
    [anon_sym___attribute__] = ACTIONS(5521),
    [anon_sym___attribute] = ACTIONS(5521),
    [anon_sym_noreturn] = ACTIONS(5521),
    [anon_sym_LBRACK] = ACTIONS(5523),
    [anon_sym___declspec] = ACTIONS(5521),
    [anon_sym___cdecl] = ACTIONS(5521),
    [anon_sym___clrcall] = ACTIONS(5521),
    [anon_sym___stdcall] = ACTIONS(5521),
    [anon_sym___fastcall] = ACTIONS(5521),
    [anon_sym___thiscall] = ACTIONS(5521),
    [anon_sym___vectorcall] = ACTIONS(5521),
    [anon_sym_signed] = ACTIONS(5521),
    [anon_sym_unsigned] = ACTIONS(5521),
    [anon_sym_long] = ACTIONS(5521),
    [anon_sym_short] = ACTIONS(5521),
    [anon_sym_static] = ACTIONS(5521),
    [anon_sym_auto] = ACTIONS(5521),
    [anon_sym_register] = ACTIONS(5521),
    [anon_sym_inline] = ACTIONS(5521),
    [anon_sym___inline] = ACTIONS(5521),
    [anon_sym___inline__] = ACTIONS(5521),
    [anon_sym___forceinline] = ACTIONS(5521),
    [anon_sym_thread_local] = ACTIONS(5521),
    [anon_sym___thread] = ACTIONS(5521),
    [anon_sym_CG_EXTERN] = ACTIONS(5521),
    [anon_sym_CG_INLINE] = ACTIONS(5521),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5521),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5521),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5521),
    [anon_sym_IBOutlet] = ACTIONS(5521),
    [anon_sym_IBInspectable] = ACTIONS(5521),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5521),
    [anon_sym_NS_INLINE] = ACTIONS(5521),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5521),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5521),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5521),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5521),
    [anon_sym_const] = ACTIONS(5521),
    [anon_sym_constexpr] = ACTIONS(5521),
    [anon_sym_volatile] = ACTIONS(5521),
    [anon_sym_restrict] = ACTIONS(5521),
    [anon_sym___restrict__] = ACTIONS(5521),
    [anon_sym__Atomic] = ACTIONS(5521),
    [anon_sym__Noreturn] = ACTIONS(5521),
    [anon_sym__Nonnull] = ACTIONS(5521),
    [anon_sym_nullable] = ACTIONS(5521),
    [anon_sym__Complex] = ACTIONS(5521),
    [anon_sym__Nullable] = ACTIONS(5521),
    [anon_sym__Nullable_result] = ACTIONS(5521),
    [anon_sym__Null_unspecified] = ACTIONS(5521),
    [anon_sym___autoreleasing] = ACTIONS(5521),
    [anon_sym___block] = ACTIONS(5521),
    [anon_sym___bridge] = ACTIONS(5521),
    [anon_sym___bridge_retained] = ACTIONS(5521),
    [anon_sym___bridge_transfer] = ACTIONS(5521),
    [anon_sym___complex] = ACTIONS(5521),
    [anon_sym___const] = ACTIONS(5521),
    [anon_sym___imag] = ACTIONS(5521),
    [anon_sym___kindof] = ACTIONS(5521),
    [anon_sym___nonnull] = ACTIONS(5521),
    [anon_sym___nullable] = ACTIONS(5521),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5521),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5521),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5521),
    [anon_sym___real] = ACTIONS(5521),
    [anon_sym___strong] = ACTIONS(5521),
    [anon_sym___unsafe_unretained] = ACTIONS(5521),
    [anon_sym___unused] = ACTIONS(5521),
    [anon_sym___weak] = ACTIONS(5521),
    [anon_sym_alignas] = ACTIONS(5521),
    [anon_sym__Alignas] = ACTIONS(5521),
    [sym_primitive_type] = ACTIONS(5521),
    [anon_sym_enum] = ACTIONS(5521),
    [anon_sym_struct] = ACTIONS(5521),
    [anon_sym_union] = ACTIONS(5521),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5521),
    [anon_sym___typeof] = ACTIONS(5521),
    [anon_sym_typeof] = ACTIONS(5521),
    [aux_sym_preproc_undef_token1] = ACTIONS(5521),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5521),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5521),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5521),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5521),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5521),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5521),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5521),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5521),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5521),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_API_AVAILABLE] = ACTIONS(5521),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_API_DEPRECATED] = ACTIONS(5521),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5521),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5521),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5521),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5521),
    [anon_sym___deprecated_msg] = ACTIONS(5521),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5521),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5521),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5521),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5521),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5521),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5521),
    [anon_sym_ATsynthesize] = ACTIONS(5523),
    [anon_sym_ATdynamic] = ACTIONS(5523),
    [anon_sym_BOOL] = ACTIONS(5521),
    [anon_sym_IMP] = ACTIONS(5521),
    [anon_sym_SEL] = ACTIONS(5521),
    [anon_sym_Class] = ACTIONS(5521),
    [anon_sym_id] = ACTIONS(5521),
  },
  [STATE(2392)] = {
    [sym_identifier] = ACTIONS(5448),
    [aux_sym_preproc_def_token1] = ACTIONS(5448),
    [aux_sym_preproc_if_token1] = ACTIONS(5448),
    [aux_sym_preproc_if_token2] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5448),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5448),
    [sym_preproc_directive] = ACTIONS(5448),
    [anon_sym_DASH] = ACTIONS(5450),
    [anon_sym_PLUS] = ACTIONS(5450),
    [anon_sym___extension__] = ACTIONS(5448),
    [anon_sym_typedef] = ACTIONS(5448),
    [anon_sym_extern] = ACTIONS(5448),
    [anon_sym___attribute__] = ACTIONS(5448),
    [anon_sym___attribute] = ACTIONS(5448),
    [anon_sym_noreturn] = ACTIONS(5448),
    [anon_sym_LBRACK] = ACTIONS(5450),
    [anon_sym___declspec] = ACTIONS(5448),
    [anon_sym___cdecl] = ACTIONS(5448),
    [anon_sym___clrcall] = ACTIONS(5448),
    [anon_sym___stdcall] = ACTIONS(5448),
    [anon_sym___fastcall] = ACTIONS(5448),
    [anon_sym___thiscall] = ACTIONS(5448),
    [anon_sym___vectorcall] = ACTIONS(5448),
    [anon_sym_signed] = ACTIONS(5448),
    [anon_sym_unsigned] = ACTIONS(5448),
    [anon_sym_long] = ACTIONS(5448),
    [anon_sym_short] = ACTIONS(5448),
    [anon_sym_static] = ACTIONS(5448),
    [anon_sym_auto] = ACTIONS(5448),
    [anon_sym_register] = ACTIONS(5448),
    [anon_sym_inline] = ACTIONS(5448),
    [anon_sym___inline] = ACTIONS(5448),
    [anon_sym___inline__] = ACTIONS(5448),
    [anon_sym___forceinline] = ACTIONS(5448),
    [anon_sym_thread_local] = ACTIONS(5448),
    [anon_sym___thread] = ACTIONS(5448),
    [anon_sym_CG_EXTERN] = ACTIONS(5448),
    [anon_sym_CG_INLINE] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5448),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5448),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5448),
    [anon_sym_IBOutlet] = ACTIONS(5448),
    [anon_sym_IBInspectable] = ACTIONS(5448),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5448),
    [anon_sym_NS_INLINE] = ACTIONS(5448),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5448),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5448),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5448),
    [anon_sym_const] = ACTIONS(5448),
    [anon_sym_constexpr] = ACTIONS(5448),
    [anon_sym_volatile] = ACTIONS(5448),
    [anon_sym_restrict] = ACTIONS(5448),
    [anon_sym___restrict__] = ACTIONS(5448),
    [anon_sym__Atomic] = ACTIONS(5448),
    [anon_sym__Noreturn] = ACTIONS(5448),
    [anon_sym__Nonnull] = ACTIONS(5448),
    [anon_sym_nullable] = ACTIONS(5448),
    [anon_sym__Complex] = ACTIONS(5448),
    [anon_sym__Nullable] = ACTIONS(5448),
    [anon_sym__Nullable_result] = ACTIONS(5448),
    [anon_sym__Null_unspecified] = ACTIONS(5448),
    [anon_sym___autoreleasing] = ACTIONS(5448),
    [anon_sym___block] = ACTIONS(5448),
    [anon_sym___bridge] = ACTIONS(5448),
    [anon_sym___bridge_retained] = ACTIONS(5448),
    [anon_sym___bridge_transfer] = ACTIONS(5448),
    [anon_sym___complex] = ACTIONS(5448),
    [anon_sym___const] = ACTIONS(5448),
    [anon_sym___imag] = ACTIONS(5448),
    [anon_sym___kindof] = ACTIONS(5448),
    [anon_sym___nonnull] = ACTIONS(5448),
    [anon_sym___nullable] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5448),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5448),
    [anon_sym___real] = ACTIONS(5448),
    [anon_sym___strong] = ACTIONS(5448),
    [anon_sym___unsafe_unretained] = ACTIONS(5448),
    [anon_sym___unused] = ACTIONS(5448),
    [anon_sym___weak] = ACTIONS(5448),
    [anon_sym_alignas] = ACTIONS(5448),
    [anon_sym__Alignas] = ACTIONS(5448),
    [sym_primitive_type] = ACTIONS(5448),
    [anon_sym_enum] = ACTIONS(5448),
    [anon_sym_struct] = ACTIONS(5448),
    [anon_sym_union] = ACTIONS(5448),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5448),
    [anon_sym___typeof] = ACTIONS(5448),
    [anon_sym_typeof] = ACTIONS(5448),
    [aux_sym_preproc_undef_token1] = ACTIONS(5448),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5448),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5448),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5448),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5448),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5448),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_API_AVAILABLE] = ACTIONS(5448),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_API_DEPRECATED] = ACTIONS(5448),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5448),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5448),
    [anon_sym___deprecated_msg] = ACTIONS(5448),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5448),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5448),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5448),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5448),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5448),
    [anon_sym_ATsynthesize] = ACTIONS(5450),
    [anon_sym_ATdynamic] = ACTIONS(5450),
    [anon_sym_BOOL] = ACTIONS(5448),
    [anon_sym_IMP] = ACTIONS(5448),
    [anon_sym_SEL] = ACTIONS(5448),
    [anon_sym_Class] = ACTIONS(5448),
    [anon_sym_id] = ACTIONS(5448),
  },
  [STATE(2393)] = {
    [sym_identifier] = ACTIONS(5438),
    [aux_sym_preproc_def_token1] = ACTIONS(5438),
    [aux_sym_preproc_if_token1] = ACTIONS(5438),
    [aux_sym_preproc_if_token2] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5438),
    [sym_preproc_directive] = ACTIONS(5438),
    [anon_sym_DASH] = ACTIONS(5440),
    [anon_sym_PLUS] = ACTIONS(5440),
    [anon_sym___extension__] = ACTIONS(5438),
    [anon_sym_typedef] = ACTIONS(5438),
    [anon_sym_extern] = ACTIONS(5438),
    [anon_sym___attribute__] = ACTIONS(5438),
    [anon_sym___attribute] = ACTIONS(5438),
    [anon_sym_noreturn] = ACTIONS(5438),
    [anon_sym_LBRACK] = ACTIONS(5440),
    [anon_sym___declspec] = ACTIONS(5438),
    [anon_sym___cdecl] = ACTIONS(5438),
    [anon_sym___clrcall] = ACTIONS(5438),
    [anon_sym___stdcall] = ACTIONS(5438),
    [anon_sym___fastcall] = ACTIONS(5438),
    [anon_sym___thiscall] = ACTIONS(5438),
    [anon_sym___vectorcall] = ACTIONS(5438),
    [anon_sym_signed] = ACTIONS(5438),
    [anon_sym_unsigned] = ACTIONS(5438),
    [anon_sym_long] = ACTIONS(5438),
    [anon_sym_short] = ACTIONS(5438),
    [anon_sym_static] = ACTIONS(5438),
    [anon_sym_auto] = ACTIONS(5438),
    [anon_sym_register] = ACTIONS(5438),
    [anon_sym_inline] = ACTIONS(5438),
    [anon_sym___inline] = ACTIONS(5438),
    [anon_sym___inline__] = ACTIONS(5438),
    [anon_sym___forceinline] = ACTIONS(5438),
    [anon_sym_thread_local] = ACTIONS(5438),
    [anon_sym___thread] = ACTIONS(5438),
    [anon_sym_CG_EXTERN] = ACTIONS(5438),
    [anon_sym_CG_INLINE] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5438),
    [anon_sym_IBOutlet] = ACTIONS(5438),
    [anon_sym_IBInspectable] = ACTIONS(5438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5438),
    [anon_sym_NS_INLINE] = ACTIONS(5438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5438),
    [anon_sym_const] = ACTIONS(5438),
    [anon_sym_constexpr] = ACTIONS(5438),
    [anon_sym_volatile] = ACTIONS(5438),
    [anon_sym_restrict] = ACTIONS(5438),
    [anon_sym___restrict__] = ACTIONS(5438),
    [anon_sym__Atomic] = ACTIONS(5438),
    [anon_sym__Noreturn] = ACTIONS(5438),
    [anon_sym__Nonnull] = ACTIONS(5438),
    [anon_sym_nullable] = ACTIONS(5438),
    [anon_sym__Complex] = ACTIONS(5438),
    [anon_sym__Nullable] = ACTIONS(5438),
    [anon_sym__Nullable_result] = ACTIONS(5438),
    [anon_sym__Null_unspecified] = ACTIONS(5438),
    [anon_sym___autoreleasing] = ACTIONS(5438),
    [anon_sym___block] = ACTIONS(5438),
    [anon_sym___bridge] = ACTIONS(5438),
    [anon_sym___bridge_retained] = ACTIONS(5438),
    [anon_sym___bridge_transfer] = ACTIONS(5438),
    [anon_sym___complex] = ACTIONS(5438),
    [anon_sym___const] = ACTIONS(5438),
    [anon_sym___imag] = ACTIONS(5438),
    [anon_sym___kindof] = ACTIONS(5438),
    [anon_sym___nonnull] = ACTIONS(5438),
    [anon_sym___nullable] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5438),
    [anon_sym___real] = ACTIONS(5438),
    [anon_sym___strong] = ACTIONS(5438),
    [anon_sym___unsafe_unretained] = ACTIONS(5438),
    [anon_sym___unused] = ACTIONS(5438),
    [anon_sym___weak] = ACTIONS(5438),
    [anon_sym_alignas] = ACTIONS(5438),
    [anon_sym__Alignas] = ACTIONS(5438),
    [sym_primitive_type] = ACTIONS(5438),
    [anon_sym_enum] = ACTIONS(5438),
    [anon_sym_struct] = ACTIONS(5438),
    [anon_sym_union] = ACTIONS(5438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5438),
    [anon_sym___typeof] = ACTIONS(5438),
    [anon_sym_typeof] = ACTIONS(5438),
    [aux_sym_preproc_undef_token1] = ACTIONS(5438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_API_AVAILABLE] = ACTIONS(5438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_API_DEPRECATED] = ACTIONS(5438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym___deprecated_msg] = ACTIONS(5438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5438),
    [anon_sym_ATsynthesize] = ACTIONS(5440),
    [anon_sym_ATdynamic] = ACTIONS(5440),
    [anon_sym_BOOL] = ACTIONS(5438),
    [anon_sym_IMP] = ACTIONS(5438),
    [anon_sym_SEL] = ACTIONS(5438),
    [anon_sym_Class] = ACTIONS(5438),
    [anon_sym_id] = ACTIONS(5438),
  },
  [STATE(2394)] = {
    [sym_identifier] = ACTIONS(5438),
    [aux_sym_preproc_def_token1] = ACTIONS(5438),
    [aux_sym_preproc_if_token1] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5438),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5438),
    [sym_preproc_directive] = ACTIONS(5438),
    [anon_sym_DASH] = ACTIONS(5440),
    [anon_sym_PLUS] = ACTIONS(5440),
    [anon_sym___extension__] = ACTIONS(5438),
    [anon_sym_typedef] = ACTIONS(5438),
    [anon_sym_extern] = ACTIONS(5438),
    [anon_sym___attribute__] = ACTIONS(5438),
    [anon_sym___attribute] = ACTIONS(5438),
    [anon_sym_noreturn] = ACTIONS(5438),
    [anon_sym_LBRACK] = ACTIONS(5440),
    [anon_sym___declspec] = ACTIONS(5438),
    [anon_sym___cdecl] = ACTIONS(5438),
    [anon_sym___clrcall] = ACTIONS(5438),
    [anon_sym___stdcall] = ACTIONS(5438),
    [anon_sym___fastcall] = ACTIONS(5438),
    [anon_sym___thiscall] = ACTIONS(5438),
    [anon_sym___vectorcall] = ACTIONS(5438),
    [anon_sym_signed] = ACTIONS(5438),
    [anon_sym_unsigned] = ACTIONS(5438),
    [anon_sym_long] = ACTIONS(5438),
    [anon_sym_short] = ACTIONS(5438),
    [anon_sym_static] = ACTIONS(5438),
    [anon_sym_auto] = ACTIONS(5438),
    [anon_sym_register] = ACTIONS(5438),
    [anon_sym_inline] = ACTIONS(5438),
    [anon_sym___inline] = ACTIONS(5438),
    [anon_sym___inline__] = ACTIONS(5438),
    [anon_sym___forceinline] = ACTIONS(5438),
    [anon_sym_thread_local] = ACTIONS(5438),
    [anon_sym___thread] = ACTIONS(5438),
    [anon_sym_CG_EXTERN] = ACTIONS(5438),
    [anon_sym_CG_INLINE] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5438),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5438),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5438),
    [anon_sym_IBOutlet] = ACTIONS(5438),
    [anon_sym_IBInspectable] = ACTIONS(5438),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5438),
    [anon_sym_NS_INLINE] = ACTIONS(5438),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5438),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5438),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5438),
    [anon_sym_const] = ACTIONS(5438),
    [anon_sym_constexpr] = ACTIONS(5438),
    [anon_sym_volatile] = ACTIONS(5438),
    [anon_sym_restrict] = ACTIONS(5438),
    [anon_sym___restrict__] = ACTIONS(5438),
    [anon_sym__Atomic] = ACTIONS(5438),
    [anon_sym__Noreturn] = ACTIONS(5438),
    [anon_sym__Nonnull] = ACTIONS(5438),
    [anon_sym_nullable] = ACTIONS(5438),
    [anon_sym__Complex] = ACTIONS(5438),
    [anon_sym__Nullable] = ACTIONS(5438),
    [anon_sym__Nullable_result] = ACTIONS(5438),
    [anon_sym__Null_unspecified] = ACTIONS(5438),
    [anon_sym___autoreleasing] = ACTIONS(5438),
    [anon_sym___block] = ACTIONS(5438),
    [anon_sym___bridge] = ACTIONS(5438),
    [anon_sym___bridge_retained] = ACTIONS(5438),
    [anon_sym___bridge_transfer] = ACTIONS(5438),
    [anon_sym___complex] = ACTIONS(5438),
    [anon_sym___const] = ACTIONS(5438),
    [anon_sym___imag] = ACTIONS(5438),
    [anon_sym___kindof] = ACTIONS(5438),
    [anon_sym___nonnull] = ACTIONS(5438),
    [anon_sym___nullable] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5438),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5438),
    [anon_sym___real] = ACTIONS(5438),
    [anon_sym___strong] = ACTIONS(5438),
    [anon_sym___unsafe_unretained] = ACTIONS(5438),
    [anon_sym___unused] = ACTIONS(5438),
    [anon_sym___weak] = ACTIONS(5438),
    [anon_sym_alignas] = ACTIONS(5438),
    [anon_sym__Alignas] = ACTIONS(5438),
    [sym_primitive_type] = ACTIONS(5438),
    [anon_sym_enum] = ACTIONS(5438),
    [anon_sym_struct] = ACTIONS(5438),
    [anon_sym_union] = ACTIONS(5438),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5438),
    [anon_sym___typeof] = ACTIONS(5438),
    [anon_sym_typeof] = ACTIONS(5438),
    [aux_sym_preproc_undef_token1] = ACTIONS(5438),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5438),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5438),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5438),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5438),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5438),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_API_AVAILABLE] = ACTIONS(5438),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_API_DEPRECATED] = ACTIONS(5438),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5438),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5438),
    [anon_sym___deprecated_msg] = ACTIONS(5438),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5438),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5438),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5438),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5438),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5438),
    [anon_sym_ATend] = ACTIONS(5440),
    [anon_sym_ATsynthesize] = ACTIONS(5440),
    [anon_sym_ATdynamic] = ACTIONS(5440),
    [anon_sym_BOOL] = ACTIONS(5438),
    [anon_sym_IMP] = ACTIONS(5438),
    [anon_sym_SEL] = ACTIONS(5438),
    [anon_sym_Class] = ACTIONS(5438),
    [anon_sym_id] = ACTIONS(5438),
  },
  [STATE(2395)] = {
    [sym_identifier] = ACTIONS(5387),
    [aux_sym_preproc_def_token1] = ACTIONS(5387),
    [aux_sym_preproc_if_token1] = ACTIONS(5387),
    [aux_sym_preproc_if_token2] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5387),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5387),
    [sym_preproc_directive] = ACTIONS(5387),
    [anon_sym_DASH] = ACTIONS(5389),
    [anon_sym_PLUS] = ACTIONS(5389),
    [anon_sym___extension__] = ACTIONS(5387),
    [anon_sym_typedef] = ACTIONS(5387),
    [anon_sym_extern] = ACTIONS(5387),
    [anon_sym___attribute__] = ACTIONS(5387),
    [anon_sym___attribute] = ACTIONS(5387),
    [anon_sym_noreturn] = ACTIONS(5387),
    [anon_sym_LBRACK] = ACTIONS(5389),
    [anon_sym___declspec] = ACTIONS(5387),
    [anon_sym___cdecl] = ACTIONS(5387),
    [anon_sym___clrcall] = ACTIONS(5387),
    [anon_sym___stdcall] = ACTIONS(5387),
    [anon_sym___fastcall] = ACTIONS(5387),
    [anon_sym___thiscall] = ACTIONS(5387),
    [anon_sym___vectorcall] = ACTIONS(5387),
    [anon_sym_signed] = ACTIONS(5387),
    [anon_sym_unsigned] = ACTIONS(5387),
    [anon_sym_long] = ACTIONS(5387),
    [anon_sym_short] = ACTIONS(5387),
    [anon_sym_static] = ACTIONS(5387),
    [anon_sym_auto] = ACTIONS(5387),
    [anon_sym_register] = ACTIONS(5387),
    [anon_sym_inline] = ACTIONS(5387),
    [anon_sym___inline] = ACTIONS(5387),
    [anon_sym___inline__] = ACTIONS(5387),
    [anon_sym___forceinline] = ACTIONS(5387),
    [anon_sym_thread_local] = ACTIONS(5387),
    [anon_sym___thread] = ACTIONS(5387),
    [anon_sym_CG_EXTERN] = ACTIONS(5387),
    [anon_sym_CG_INLINE] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5387),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5387),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5387),
    [anon_sym_IBOutlet] = ACTIONS(5387),
    [anon_sym_IBInspectable] = ACTIONS(5387),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5387),
    [anon_sym_NS_INLINE] = ACTIONS(5387),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5387),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5387),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5387),
    [anon_sym_const] = ACTIONS(5387),
    [anon_sym_constexpr] = ACTIONS(5387),
    [anon_sym_volatile] = ACTIONS(5387),
    [anon_sym_restrict] = ACTIONS(5387),
    [anon_sym___restrict__] = ACTIONS(5387),
    [anon_sym__Atomic] = ACTIONS(5387),
    [anon_sym__Noreturn] = ACTIONS(5387),
    [anon_sym__Nonnull] = ACTIONS(5387),
    [anon_sym_nullable] = ACTIONS(5387),
    [anon_sym__Complex] = ACTIONS(5387),
    [anon_sym__Nullable] = ACTIONS(5387),
    [anon_sym__Nullable_result] = ACTIONS(5387),
    [anon_sym__Null_unspecified] = ACTIONS(5387),
    [anon_sym___autoreleasing] = ACTIONS(5387),
    [anon_sym___block] = ACTIONS(5387),
    [anon_sym___bridge] = ACTIONS(5387),
    [anon_sym___bridge_retained] = ACTIONS(5387),
    [anon_sym___bridge_transfer] = ACTIONS(5387),
    [anon_sym___complex] = ACTIONS(5387),
    [anon_sym___const] = ACTIONS(5387),
    [anon_sym___imag] = ACTIONS(5387),
    [anon_sym___kindof] = ACTIONS(5387),
    [anon_sym___nonnull] = ACTIONS(5387),
    [anon_sym___nullable] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5387),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5387),
    [anon_sym___real] = ACTIONS(5387),
    [anon_sym___strong] = ACTIONS(5387),
    [anon_sym___unsafe_unretained] = ACTIONS(5387),
    [anon_sym___unused] = ACTIONS(5387),
    [anon_sym___weak] = ACTIONS(5387),
    [anon_sym_alignas] = ACTIONS(5387),
    [anon_sym__Alignas] = ACTIONS(5387),
    [sym_primitive_type] = ACTIONS(5387),
    [anon_sym_enum] = ACTIONS(5387),
    [anon_sym_struct] = ACTIONS(5387),
    [anon_sym_union] = ACTIONS(5387),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5387),
    [anon_sym___typeof] = ACTIONS(5387),
    [anon_sym_typeof] = ACTIONS(5387),
    [aux_sym_preproc_undef_token1] = ACTIONS(5387),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5387),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5387),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5387),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5387),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5387),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_API_AVAILABLE] = ACTIONS(5387),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_API_DEPRECATED] = ACTIONS(5387),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5387),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5387),
    [anon_sym___deprecated_msg] = ACTIONS(5387),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5387),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5387),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5387),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5387),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5387),
    [anon_sym_ATsynthesize] = ACTIONS(5389),
    [anon_sym_ATdynamic] = ACTIONS(5389),
    [anon_sym_BOOL] = ACTIONS(5387),
    [anon_sym_IMP] = ACTIONS(5387),
    [anon_sym_SEL] = ACTIONS(5387),
    [anon_sym_Class] = ACTIONS(5387),
    [anon_sym_id] = ACTIONS(5387),
  },
  [STATE(2396)] = {
    [sym_identifier] = ACTIONS(5351),
    [aux_sym_preproc_def_token1] = ACTIONS(5351),
    [aux_sym_preproc_if_token1] = ACTIONS(5351),
    [aux_sym_preproc_if_token2] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5351),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5351),
    [sym_preproc_directive] = ACTIONS(5351),
    [anon_sym_DASH] = ACTIONS(5353),
    [anon_sym_PLUS] = ACTIONS(5353),
    [anon_sym___extension__] = ACTIONS(5351),
    [anon_sym_typedef] = ACTIONS(5351),
    [anon_sym_extern] = ACTIONS(5351),
    [anon_sym___attribute__] = ACTIONS(5351),
    [anon_sym___attribute] = ACTIONS(5351),
    [anon_sym_noreturn] = ACTIONS(5351),
    [anon_sym_LBRACK] = ACTIONS(5353),
    [anon_sym___declspec] = ACTIONS(5351),
    [anon_sym___cdecl] = ACTIONS(5351),
    [anon_sym___clrcall] = ACTIONS(5351),
    [anon_sym___stdcall] = ACTIONS(5351),
    [anon_sym___fastcall] = ACTIONS(5351),
    [anon_sym___thiscall] = ACTIONS(5351),
    [anon_sym___vectorcall] = ACTIONS(5351),
    [anon_sym_signed] = ACTIONS(5351),
    [anon_sym_unsigned] = ACTIONS(5351),
    [anon_sym_long] = ACTIONS(5351),
    [anon_sym_short] = ACTIONS(5351),
    [anon_sym_static] = ACTIONS(5351),
    [anon_sym_auto] = ACTIONS(5351),
    [anon_sym_register] = ACTIONS(5351),
    [anon_sym_inline] = ACTIONS(5351),
    [anon_sym___inline] = ACTIONS(5351),
    [anon_sym___inline__] = ACTIONS(5351),
    [anon_sym___forceinline] = ACTIONS(5351),
    [anon_sym_thread_local] = ACTIONS(5351),
    [anon_sym___thread] = ACTIONS(5351),
    [anon_sym_CG_EXTERN] = ACTIONS(5351),
    [anon_sym_CG_INLINE] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5351),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5351),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5351),
    [anon_sym_IBOutlet] = ACTIONS(5351),
    [anon_sym_IBInspectable] = ACTIONS(5351),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5351),
    [anon_sym_NS_INLINE] = ACTIONS(5351),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5351),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5351),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5351),
    [anon_sym_const] = ACTIONS(5351),
    [anon_sym_constexpr] = ACTIONS(5351),
    [anon_sym_volatile] = ACTIONS(5351),
    [anon_sym_restrict] = ACTIONS(5351),
    [anon_sym___restrict__] = ACTIONS(5351),
    [anon_sym__Atomic] = ACTIONS(5351),
    [anon_sym__Noreturn] = ACTIONS(5351),
    [anon_sym__Nonnull] = ACTIONS(5351),
    [anon_sym_nullable] = ACTIONS(5351),
    [anon_sym__Complex] = ACTIONS(5351),
    [anon_sym__Nullable] = ACTIONS(5351),
    [anon_sym__Nullable_result] = ACTIONS(5351),
    [anon_sym__Null_unspecified] = ACTIONS(5351),
    [anon_sym___autoreleasing] = ACTIONS(5351),
    [anon_sym___block] = ACTIONS(5351),
    [anon_sym___bridge] = ACTIONS(5351),
    [anon_sym___bridge_retained] = ACTIONS(5351),
    [anon_sym___bridge_transfer] = ACTIONS(5351),
    [anon_sym___complex] = ACTIONS(5351),
    [anon_sym___const] = ACTIONS(5351),
    [anon_sym___imag] = ACTIONS(5351),
    [anon_sym___kindof] = ACTIONS(5351),
    [anon_sym___nonnull] = ACTIONS(5351),
    [anon_sym___nullable] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5351),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5351),
    [anon_sym___real] = ACTIONS(5351),
    [anon_sym___strong] = ACTIONS(5351),
    [anon_sym___unsafe_unretained] = ACTIONS(5351),
    [anon_sym___unused] = ACTIONS(5351),
    [anon_sym___weak] = ACTIONS(5351),
    [anon_sym_alignas] = ACTIONS(5351),
    [anon_sym__Alignas] = ACTIONS(5351),
    [sym_primitive_type] = ACTIONS(5351),
    [anon_sym_enum] = ACTIONS(5351),
    [anon_sym_struct] = ACTIONS(5351),
    [anon_sym_union] = ACTIONS(5351),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5351),
    [anon_sym___typeof] = ACTIONS(5351),
    [anon_sym_typeof] = ACTIONS(5351),
    [aux_sym_preproc_undef_token1] = ACTIONS(5351),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5351),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5351),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5351),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5351),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5351),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_API_AVAILABLE] = ACTIONS(5351),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_API_DEPRECATED] = ACTIONS(5351),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5351),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5351),
    [anon_sym___deprecated_msg] = ACTIONS(5351),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5351),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5351),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5351),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5351),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5351),
    [anon_sym_ATsynthesize] = ACTIONS(5353),
    [anon_sym_ATdynamic] = ACTIONS(5353),
    [anon_sym_BOOL] = ACTIONS(5351),
    [anon_sym_IMP] = ACTIONS(5351),
    [anon_sym_SEL] = ACTIONS(5351),
    [anon_sym_Class] = ACTIONS(5351),
    [anon_sym_id] = ACTIONS(5351),
  },
  [STATE(2397)] = {
    [sym_identifier] = ACTIONS(2614),
    [aux_sym_preproc_def_token1] = ACTIONS(2614),
    [aux_sym_preproc_if_token1] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2614),
    [sym_preproc_directive] = ACTIONS(2614),
    [anon_sym_DASH] = ACTIONS(2616),
    [anon_sym_PLUS] = ACTIONS(2616),
    [anon_sym___extension__] = ACTIONS(2614),
    [anon_sym_typedef] = ACTIONS(2614),
    [anon_sym_extern] = ACTIONS(2614),
    [anon_sym___attribute__] = ACTIONS(2614),
    [anon_sym___attribute] = ACTIONS(2614),
    [anon_sym_noreturn] = ACTIONS(2614),
    [anon_sym_LBRACK] = ACTIONS(2616),
    [anon_sym___declspec] = ACTIONS(2614),
    [anon_sym___cdecl] = ACTIONS(2614),
    [anon_sym___clrcall] = ACTIONS(2614),
    [anon_sym___stdcall] = ACTIONS(2614),
    [anon_sym___fastcall] = ACTIONS(2614),
    [anon_sym___thiscall] = ACTIONS(2614),
    [anon_sym___vectorcall] = ACTIONS(2614),
    [anon_sym_signed] = ACTIONS(2614),
    [anon_sym_unsigned] = ACTIONS(2614),
    [anon_sym_long] = ACTIONS(2614),
    [anon_sym_short] = ACTIONS(2614),
    [anon_sym_static] = ACTIONS(2614),
    [anon_sym_auto] = ACTIONS(2614),
    [anon_sym_register] = ACTIONS(2614),
    [anon_sym_inline] = ACTIONS(2614),
    [anon_sym___inline] = ACTIONS(2614),
    [anon_sym___inline__] = ACTIONS(2614),
    [anon_sym___forceinline] = ACTIONS(2614),
    [anon_sym_thread_local] = ACTIONS(2614),
    [anon_sym___thread] = ACTIONS(2614),
    [anon_sym_CG_EXTERN] = ACTIONS(2614),
    [anon_sym_CG_INLINE] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2614),
    [anon_sym_IBOutlet] = ACTIONS(2614),
    [anon_sym_IBInspectable] = ACTIONS(2614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2614),
    [anon_sym_NS_INLINE] = ACTIONS(2614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2614),
    [anon_sym_const] = ACTIONS(2614),
    [anon_sym_constexpr] = ACTIONS(2614),
    [anon_sym_volatile] = ACTIONS(2614),
    [anon_sym_restrict] = ACTIONS(2614),
    [anon_sym___restrict__] = ACTIONS(2614),
    [anon_sym__Atomic] = ACTIONS(2614),
    [anon_sym__Noreturn] = ACTIONS(2614),
    [anon_sym__Nonnull] = ACTIONS(2614),
    [anon_sym_nullable] = ACTIONS(2614),
    [anon_sym__Complex] = ACTIONS(2614),
    [anon_sym__Nullable] = ACTIONS(2614),
    [anon_sym__Nullable_result] = ACTIONS(2614),
    [anon_sym__Null_unspecified] = ACTIONS(2614),
    [anon_sym___autoreleasing] = ACTIONS(2614),
    [anon_sym___block] = ACTIONS(2614),
    [anon_sym___bridge] = ACTIONS(2614),
    [anon_sym___bridge_retained] = ACTIONS(2614),
    [anon_sym___bridge_transfer] = ACTIONS(2614),
    [anon_sym___complex] = ACTIONS(2614),
    [anon_sym___const] = ACTIONS(2614),
    [anon_sym___imag] = ACTIONS(2614),
    [anon_sym___kindof] = ACTIONS(2614),
    [anon_sym___nonnull] = ACTIONS(2614),
    [anon_sym___nullable] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2614),
    [anon_sym___real] = ACTIONS(2614),
    [anon_sym___strong] = ACTIONS(2614),
    [anon_sym___unsafe_unretained] = ACTIONS(2614),
    [anon_sym___unused] = ACTIONS(2614),
    [anon_sym___weak] = ACTIONS(2614),
    [anon_sym_alignas] = ACTIONS(2614),
    [anon_sym__Alignas] = ACTIONS(2614),
    [sym_primitive_type] = ACTIONS(2614),
    [anon_sym_enum] = ACTIONS(2614),
    [anon_sym_struct] = ACTIONS(2614),
    [anon_sym_union] = ACTIONS(2614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2614),
    [anon_sym___typeof] = ACTIONS(2614),
    [anon_sym_typeof] = ACTIONS(2614),
    [aux_sym_preproc_undef_token1] = ACTIONS(2614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_API_AVAILABLE] = ACTIONS(2614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_API_DEPRECATED] = ACTIONS(2614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2614),
    [anon_sym___deprecated_msg] = ACTIONS(2614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2614),
    [anon_sym_ATend] = ACTIONS(2616),
    [anon_sym_ATsynthesize] = ACTIONS(2616),
    [anon_sym_ATdynamic] = ACTIONS(2616),
    [anon_sym_BOOL] = ACTIONS(2614),
    [anon_sym_IMP] = ACTIONS(2614),
    [anon_sym_SEL] = ACTIONS(2614),
    [anon_sym_Class] = ACTIONS(2614),
    [anon_sym_id] = ACTIONS(2614),
  },
  [STATE(2398)] = {
    [sym_identifier] = ACTIONS(2630),
    [aux_sym_preproc_def_token1] = ACTIONS(2630),
    [aux_sym_preproc_if_token1] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2630),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2630),
    [sym_preproc_directive] = ACTIONS(2630),
    [anon_sym_DASH] = ACTIONS(2632),
    [anon_sym_PLUS] = ACTIONS(2632),
    [anon_sym___extension__] = ACTIONS(2630),
    [anon_sym_typedef] = ACTIONS(2630),
    [anon_sym_extern] = ACTIONS(2630),
    [anon_sym___attribute__] = ACTIONS(2630),
    [anon_sym___attribute] = ACTIONS(2630),
    [anon_sym_noreturn] = ACTIONS(2630),
    [anon_sym_LBRACK] = ACTIONS(2632),
    [anon_sym___declspec] = ACTIONS(2630),
    [anon_sym___cdecl] = ACTIONS(2630),
    [anon_sym___clrcall] = ACTIONS(2630),
    [anon_sym___stdcall] = ACTIONS(2630),
    [anon_sym___fastcall] = ACTIONS(2630),
    [anon_sym___thiscall] = ACTIONS(2630),
    [anon_sym___vectorcall] = ACTIONS(2630),
    [anon_sym_signed] = ACTIONS(2630),
    [anon_sym_unsigned] = ACTIONS(2630),
    [anon_sym_long] = ACTIONS(2630),
    [anon_sym_short] = ACTIONS(2630),
    [anon_sym_static] = ACTIONS(2630),
    [anon_sym_auto] = ACTIONS(2630),
    [anon_sym_register] = ACTIONS(2630),
    [anon_sym_inline] = ACTIONS(2630),
    [anon_sym___inline] = ACTIONS(2630),
    [anon_sym___inline__] = ACTIONS(2630),
    [anon_sym___forceinline] = ACTIONS(2630),
    [anon_sym_thread_local] = ACTIONS(2630),
    [anon_sym___thread] = ACTIONS(2630),
    [anon_sym_CG_EXTERN] = ACTIONS(2630),
    [anon_sym_CG_INLINE] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2630),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2630),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2630),
    [anon_sym_IBOutlet] = ACTIONS(2630),
    [anon_sym_IBInspectable] = ACTIONS(2630),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2630),
    [anon_sym_NS_INLINE] = ACTIONS(2630),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2630),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2630),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2630),
    [anon_sym_const] = ACTIONS(2630),
    [anon_sym_constexpr] = ACTIONS(2630),
    [anon_sym_volatile] = ACTIONS(2630),
    [anon_sym_restrict] = ACTIONS(2630),
    [anon_sym___restrict__] = ACTIONS(2630),
    [anon_sym__Atomic] = ACTIONS(2630),
    [anon_sym__Noreturn] = ACTIONS(2630),
    [anon_sym__Nonnull] = ACTIONS(2630),
    [anon_sym_nullable] = ACTIONS(2630),
    [anon_sym__Complex] = ACTIONS(2630),
    [anon_sym__Nullable] = ACTIONS(2630),
    [anon_sym__Nullable_result] = ACTIONS(2630),
    [anon_sym__Null_unspecified] = ACTIONS(2630),
    [anon_sym___autoreleasing] = ACTIONS(2630),
    [anon_sym___block] = ACTIONS(2630),
    [anon_sym___bridge] = ACTIONS(2630),
    [anon_sym___bridge_retained] = ACTIONS(2630),
    [anon_sym___bridge_transfer] = ACTIONS(2630),
    [anon_sym___complex] = ACTIONS(2630),
    [anon_sym___const] = ACTIONS(2630),
    [anon_sym___imag] = ACTIONS(2630),
    [anon_sym___kindof] = ACTIONS(2630),
    [anon_sym___nonnull] = ACTIONS(2630),
    [anon_sym___nullable] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2630),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2630),
    [anon_sym___real] = ACTIONS(2630),
    [anon_sym___strong] = ACTIONS(2630),
    [anon_sym___unsafe_unretained] = ACTIONS(2630),
    [anon_sym___unused] = ACTIONS(2630),
    [anon_sym___weak] = ACTIONS(2630),
    [anon_sym_alignas] = ACTIONS(2630),
    [anon_sym__Alignas] = ACTIONS(2630),
    [sym_primitive_type] = ACTIONS(2630),
    [anon_sym_enum] = ACTIONS(2630),
    [anon_sym_struct] = ACTIONS(2630),
    [anon_sym_union] = ACTIONS(2630),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2630),
    [anon_sym___typeof] = ACTIONS(2630),
    [anon_sym_typeof] = ACTIONS(2630),
    [aux_sym_preproc_undef_token1] = ACTIONS(2630),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2630),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2630),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2630),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2630),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2630),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_API_AVAILABLE] = ACTIONS(2630),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_API_DEPRECATED] = ACTIONS(2630),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2630),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2630),
    [anon_sym___deprecated_msg] = ACTIONS(2630),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2630),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2630),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2630),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2630),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2630),
    [anon_sym_ATend] = ACTIONS(2632),
    [anon_sym_ATsynthesize] = ACTIONS(2632),
    [anon_sym_ATdynamic] = ACTIONS(2632),
    [anon_sym_BOOL] = ACTIONS(2630),
    [anon_sym_IMP] = ACTIONS(2630),
    [anon_sym_SEL] = ACTIONS(2630),
    [anon_sym_Class] = ACTIONS(2630),
    [anon_sym_id] = ACTIONS(2630),
  },
  [STATE(2399)] = {
    [sym_identifier] = ACTIONS(5393),
    [aux_sym_preproc_def_token1] = ACTIONS(5393),
    [aux_sym_preproc_if_token1] = ACTIONS(5393),
    [aux_sym_preproc_if_token2] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5393),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5393),
    [sym_preproc_directive] = ACTIONS(5393),
    [anon_sym_DASH] = ACTIONS(5395),
    [anon_sym_PLUS] = ACTIONS(5395),
    [anon_sym___extension__] = ACTIONS(5393),
    [anon_sym_typedef] = ACTIONS(5393),
    [anon_sym_extern] = ACTIONS(5393),
    [anon_sym___attribute__] = ACTIONS(5393),
    [anon_sym___attribute] = ACTIONS(5393),
    [anon_sym_noreturn] = ACTIONS(5393),
    [anon_sym_LBRACK] = ACTIONS(5395),
    [anon_sym___declspec] = ACTIONS(5393),
    [anon_sym___cdecl] = ACTIONS(5393),
    [anon_sym___clrcall] = ACTIONS(5393),
    [anon_sym___stdcall] = ACTIONS(5393),
    [anon_sym___fastcall] = ACTIONS(5393),
    [anon_sym___thiscall] = ACTIONS(5393),
    [anon_sym___vectorcall] = ACTIONS(5393),
    [anon_sym_signed] = ACTIONS(5393),
    [anon_sym_unsigned] = ACTIONS(5393),
    [anon_sym_long] = ACTIONS(5393),
    [anon_sym_short] = ACTIONS(5393),
    [anon_sym_static] = ACTIONS(5393),
    [anon_sym_auto] = ACTIONS(5393),
    [anon_sym_register] = ACTIONS(5393),
    [anon_sym_inline] = ACTIONS(5393),
    [anon_sym___inline] = ACTIONS(5393),
    [anon_sym___inline__] = ACTIONS(5393),
    [anon_sym___forceinline] = ACTIONS(5393),
    [anon_sym_thread_local] = ACTIONS(5393),
    [anon_sym___thread] = ACTIONS(5393),
    [anon_sym_CG_EXTERN] = ACTIONS(5393),
    [anon_sym_CG_INLINE] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5393),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5393),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5393),
    [anon_sym_IBOutlet] = ACTIONS(5393),
    [anon_sym_IBInspectable] = ACTIONS(5393),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5393),
    [anon_sym_NS_INLINE] = ACTIONS(5393),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5393),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5393),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5393),
    [anon_sym_const] = ACTIONS(5393),
    [anon_sym_constexpr] = ACTIONS(5393),
    [anon_sym_volatile] = ACTIONS(5393),
    [anon_sym_restrict] = ACTIONS(5393),
    [anon_sym___restrict__] = ACTIONS(5393),
    [anon_sym__Atomic] = ACTIONS(5393),
    [anon_sym__Noreturn] = ACTIONS(5393),
    [anon_sym__Nonnull] = ACTIONS(5393),
    [anon_sym_nullable] = ACTIONS(5393),
    [anon_sym__Complex] = ACTIONS(5393),
    [anon_sym__Nullable] = ACTIONS(5393),
    [anon_sym__Nullable_result] = ACTIONS(5393),
    [anon_sym__Null_unspecified] = ACTIONS(5393),
    [anon_sym___autoreleasing] = ACTIONS(5393),
    [anon_sym___block] = ACTIONS(5393),
    [anon_sym___bridge] = ACTIONS(5393),
    [anon_sym___bridge_retained] = ACTIONS(5393),
    [anon_sym___bridge_transfer] = ACTIONS(5393),
    [anon_sym___complex] = ACTIONS(5393),
    [anon_sym___const] = ACTIONS(5393),
    [anon_sym___imag] = ACTIONS(5393),
    [anon_sym___kindof] = ACTIONS(5393),
    [anon_sym___nonnull] = ACTIONS(5393),
    [anon_sym___nullable] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5393),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5393),
    [anon_sym___real] = ACTIONS(5393),
    [anon_sym___strong] = ACTIONS(5393),
    [anon_sym___unsafe_unretained] = ACTIONS(5393),
    [anon_sym___unused] = ACTIONS(5393),
    [anon_sym___weak] = ACTIONS(5393),
    [anon_sym_alignas] = ACTIONS(5393),
    [anon_sym__Alignas] = ACTIONS(5393),
    [sym_primitive_type] = ACTIONS(5393),
    [anon_sym_enum] = ACTIONS(5393),
    [anon_sym_struct] = ACTIONS(5393),
    [anon_sym_union] = ACTIONS(5393),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5393),
    [anon_sym___typeof] = ACTIONS(5393),
    [anon_sym_typeof] = ACTIONS(5393),
    [aux_sym_preproc_undef_token1] = ACTIONS(5393),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5393),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5393),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5393),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5393),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5393),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_API_AVAILABLE] = ACTIONS(5393),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_API_DEPRECATED] = ACTIONS(5393),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5393),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5393),
    [anon_sym___deprecated_msg] = ACTIONS(5393),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5393),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5393),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5393),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5393),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5393),
    [anon_sym_ATsynthesize] = ACTIONS(5395),
    [anon_sym_ATdynamic] = ACTIONS(5395),
    [anon_sym_BOOL] = ACTIONS(5393),
    [anon_sym_IMP] = ACTIONS(5393),
    [anon_sym_SEL] = ACTIONS(5393),
    [anon_sym_Class] = ACTIONS(5393),
    [anon_sym_id] = ACTIONS(5393),
  },
  [STATE(2400)] = {
    [sym_identifier] = ACTIONS(5513),
    [aux_sym_preproc_def_token1] = ACTIONS(5513),
    [aux_sym_preproc_if_token1] = ACTIONS(5513),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5513),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5513),
    [sym_preproc_directive] = ACTIONS(5513),
    [anon_sym_DASH] = ACTIONS(5515),
    [anon_sym_PLUS] = ACTIONS(5515),
    [anon_sym___extension__] = ACTIONS(5513),
    [anon_sym_typedef] = ACTIONS(5513),
    [anon_sym_extern] = ACTIONS(5513),
    [anon_sym___attribute__] = ACTIONS(5513),
    [anon_sym___attribute] = ACTIONS(5513),
    [anon_sym_noreturn] = ACTIONS(5513),
    [anon_sym_LBRACK] = ACTIONS(5515),
    [anon_sym___declspec] = ACTIONS(5513),
    [anon_sym___cdecl] = ACTIONS(5513),
    [anon_sym___clrcall] = ACTIONS(5513),
    [anon_sym___stdcall] = ACTIONS(5513),
    [anon_sym___fastcall] = ACTIONS(5513),
    [anon_sym___thiscall] = ACTIONS(5513),
    [anon_sym___vectorcall] = ACTIONS(5513),
    [anon_sym_signed] = ACTIONS(5513),
    [anon_sym_unsigned] = ACTIONS(5513),
    [anon_sym_long] = ACTIONS(5513),
    [anon_sym_short] = ACTIONS(5513),
    [anon_sym_static] = ACTIONS(5513),
    [anon_sym_auto] = ACTIONS(5513),
    [anon_sym_register] = ACTIONS(5513),
    [anon_sym_inline] = ACTIONS(5513),
    [anon_sym___inline] = ACTIONS(5513),
    [anon_sym___inline__] = ACTIONS(5513),
    [anon_sym___forceinline] = ACTIONS(5513),
    [anon_sym_thread_local] = ACTIONS(5513),
    [anon_sym___thread] = ACTIONS(5513),
    [anon_sym_CG_EXTERN] = ACTIONS(5513),
    [anon_sym_CG_INLINE] = ACTIONS(5513),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5513),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5513),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5513),
    [anon_sym_IBOutlet] = ACTIONS(5513),
    [anon_sym_IBInspectable] = ACTIONS(5513),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5513),
    [anon_sym_NS_INLINE] = ACTIONS(5513),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5513),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5513),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5513),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5513),
    [anon_sym_const] = ACTIONS(5513),
    [anon_sym_constexpr] = ACTIONS(5513),
    [anon_sym_volatile] = ACTIONS(5513),
    [anon_sym_restrict] = ACTIONS(5513),
    [anon_sym___restrict__] = ACTIONS(5513),
    [anon_sym__Atomic] = ACTIONS(5513),
    [anon_sym__Noreturn] = ACTIONS(5513),
    [anon_sym__Nonnull] = ACTIONS(5513),
    [anon_sym_nullable] = ACTIONS(5513),
    [anon_sym__Complex] = ACTIONS(5513),
    [anon_sym__Nullable] = ACTIONS(5513),
    [anon_sym__Nullable_result] = ACTIONS(5513),
    [anon_sym__Null_unspecified] = ACTIONS(5513),
    [anon_sym___autoreleasing] = ACTIONS(5513),
    [anon_sym___block] = ACTIONS(5513),
    [anon_sym___bridge] = ACTIONS(5513),
    [anon_sym___bridge_retained] = ACTIONS(5513),
    [anon_sym___bridge_transfer] = ACTIONS(5513),
    [anon_sym___complex] = ACTIONS(5513),
    [anon_sym___const] = ACTIONS(5513),
    [anon_sym___imag] = ACTIONS(5513),
    [anon_sym___kindof] = ACTIONS(5513),
    [anon_sym___nonnull] = ACTIONS(5513),
    [anon_sym___nullable] = ACTIONS(5513),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5513),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5513),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5513),
    [anon_sym___real] = ACTIONS(5513),
    [anon_sym___strong] = ACTIONS(5513),
    [anon_sym___unsafe_unretained] = ACTIONS(5513),
    [anon_sym___unused] = ACTIONS(5513),
    [anon_sym___weak] = ACTIONS(5513),
    [anon_sym_alignas] = ACTIONS(5513),
    [anon_sym__Alignas] = ACTIONS(5513),
    [sym_primitive_type] = ACTIONS(5513),
    [anon_sym_enum] = ACTIONS(5513),
    [anon_sym_struct] = ACTIONS(5513),
    [anon_sym_union] = ACTIONS(5513),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5513),
    [anon_sym___typeof] = ACTIONS(5513),
    [anon_sym_typeof] = ACTIONS(5513),
    [aux_sym_preproc_undef_token1] = ACTIONS(5513),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5513),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5513),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5513),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5513),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5513),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5513),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5513),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5513),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5513),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_API_AVAILABLE] = ACTIONS(5513),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_API_DEPRECATED] = ACTIONS(5513),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5513),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5513),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5513),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5513),
    [anon_sym___deprecated_msg] = ACTIONS(5513),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5513),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5513),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5513),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5513),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5513),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5513),
    [anon_sym_ATend] = ACTIONS(5515),
    [anon_sym_ATsynthesize] = ACTIONS(5515),
    [anon_sym_ATdynamic] = ACTIONS(5515),
    [anon_sym_BOOL] = ACTIONS(5513),
    [anon_sym_IMP] = ACTIONS(5513),
    [anon_sym_SEL] = ACTIONS(5513),
    [anon_sym_Class] = ACTIONS(5513),
    [anon_sym_id] = ACTIONS(5513),
  },
  [STATE(2401)] = {
    [sym_identifier] = ACTIONS(2962),
    [anon_sym_COMMA] = ACTIONS(2964),
    [anon_sym_RPAREN] = ACTIONS(2964),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2964),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2964),
    [anon_sym_LPAREN2] = ACTIONS(2964),
    [anon_sym_STAR] = ACTIONS(2964),
    [anon_sym_CARET] = ACTIONS(2964),
    [anon_sym_GT] = ACTIONS(2964),
    [anon_sym_SEMI] = ACTIONS(2964),
    [anon_sym___extension__] = ACTIONS(2962),
    [anon_sym_extern] = ACTIONS(2962),
    [anon_sym___attribute__] = ACTIONS(2962),
    [anon_sym___attribute] = ACTIONS(2962),
    [anon_sym_noreturn] = ACTIONS(2962),
    [anon_sym_LBRACK] = ACTIONS(2964),
    [anon_sym___declspec] = ACTIONS(2962),
    [anon_sym___based] = ACTIONS(2962),
    [anon_sym_LBRACE] = ACTIONS(2964),
    [anon_sym_RBRACE] = ACTIONS(2964),
    [anon_sym_signed] = ACTIONS(2962),
    [anon_sym_unsigned] = ACTIONS(2962),
    [anon_sym_long] = ACTIONS(2962),
    [anon_sym_short] = ACTIONS(2962),
    [anon_sym_static] = ACTIONS(2962),
    [anon_sym_ATautoreleasepool] = ACTIONS(2964),
    [anon_sym_auto] = ACTIONS(2962),
    [anon_sym_register] = ACTIONS(2962),
    [anon_sym_inline] = ACTIONS(2962),
    [anon_sym___inline] = ACTIONS(2962),
    [anon_sym___inline__] = ACTIONS(2962),
    [anon_sym___forceinline] = ACTIONS(2962),
    [anon_sym_thread_local] = ACTIONS(2962),
    [anon_sym___thread] = ACTIONS(2962),
    [anon_sym_CG_EXTERN] = ACTIONS(2962),
    [anon_sym_CG_INLINE] = ACTIONS(2962),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2962),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2962),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2962),
    [anon_sym_IBOutlet] = ACTIONS(2962),
    [anon_sym_IBInspectable] = ACTIONS(2962),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2962),
    [anon_sym_NS_INLINE] = ACTIONS(2962),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2962),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2962),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2962),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2962),
    [anon_sym_const] = ACTIONS(2962),
    [anon_sym_constexpr] = ACTIONS(2962),
    [anon_sym_volatile] = ACTIONS(2962),
    [anon_sym_restrict] = ACTIONS(2962),
    [anon_sym___restrict__] = ACTIONS(2962),
    [anon_sym__Atomic] = ACTIONS(2962),
    [anon_sym__Noreturn] = ACTIONS(2962),
    [anon_sym__Nonnull] = ACTIONS(2962),
    [anon_sym_nullable] = ACTIONS(2962),
    [anon_sym__Complex] = ACTIONS(2962),
    [anon_sym__Nullable] = ACTIONS(2962),
    [anon_sym__Nullable_result] = ACTIONS(2962),
    [anon_sym__Null_unspecified] = ACTIONS(2962),
    [anon_sym___autoreleasing] = ACTIONS(2962),
    [anon_sym___block] = ACTIONS(2962),
    [anon_sym___bridge] = ACTIONS(2962),
    [anon_sym___bridge_retained] = ACTIONS(2962),
    [anon_sym___bridge_transfer] = ACTIONS(2962),
    [anon_sym___complex] = ACTIONS(2962),
    [anon_sym___const] = ACTIONS(2962),
    [anon_sym___imag] = ACTIONS(2962),
    [anon_sym___kindof] = ACTIONS(2962),
    [anon_sym___nonnull] = ACTIONS(2962),
    [anon_sym___nullable] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2962),
    [anon_sym___real] = ACTIONS(2962),
    [anon_sym___strong] = ACTIONS(2962),
    [anon_sym___unsafe_unretained] = ACTIONS(2962),
    [anon_sym___unused] = ACTIONS(2962),
    [anon_sym___weak] = ACTIONS(2962),
    [anon_sym_alignas] = ACTIONS(2962),
    [anon_sym__Alignas] = ACTIONS(2962),
    [sym_primitive_type] = ACTIONS(2962),
    [anon_sym_enum] = ACTIONS(2962),
    [anon_sym_COLON] = ACTIONS(2964),
    [anon_sym_struct] = ACTIONS(2962),
    [anon_sym_union] = ACTIONS(2962),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2962),
    [anon_sym___typeof] = ACTIONS(2962),
    [anon_sym_typeof] = ACTIONS(2962),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2962),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2962),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2962),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2962),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2962),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2962),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2962),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2962),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2962),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_API_AVAILABLE] = ACTIONS(2962),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_API_DEPRECATED] = ACTIONS(2962),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2962),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2962),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2962),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2962),
    [anon_sym___deprecated_msg] = ACTIONS(2962),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2962),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2962),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2962),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2962),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2962),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2962),
    [anon_sym_ATprotocol] = ACTIONS(2964),
    [anon_sym_ATinterface] = ACTIONS(2964),
    [anon_sym_ATimplementation] = ACTIONS(2964),
    [anon_sym_ATproperty] = ACTIONS(2964),
    [anon_sym_BOOL] = ACTIONS(2962),
    [anon_sym_IMP] = ACTIONS(2962),
    [anon_sym_SEL] = ACTIONS(2962),
    [anon_sym_Class] = ACTIONS(2962),
    [anon_sym_id] = ACTIONS(2962),
  },
  [STATE(2402)] = {
    [sym_identifier] = ACTIONS(5509),
    [aux_sym_preproc_def_token1] = ACTIONS(5509),
    [aux_sym_preproc_if_token1] = ACTIONS(5509),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5509),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5509),
    [sym_preproc_directive] = ACTIONS(5509),
    [anon_sym_DASH] = ACTIONS(5511),
    [anon_sym_PLUS] = ACTIONS(5511),
    [anon_sym___extension__] = ACTIONS(5509),
    [anon_sym_typedef] = ACTIONS(5509),
    [anon_sym_extern] = ACTIONS(5509),
    [anon_sym___attribute__] = ACTIONS(5509),
    [anon_sym___attribute] = ACTIONS(5509),
    [anon_sym_noreturn] = ACTIONS(5509),
    [anon_sym_LBRACK] = ACTIONS(5511),
    [anon_sym___declspec] = ACTIONS(5509),
    [anon_sym___cdecl] = ACTIONS(5509),
    [anon_sym___clrcall] = ACTIONS(5509),
    [anon_sym___stdcall] = ACTIONS(5509),
    [anon_sym___fastcall] = ACTIONS(5509),
    [anon_sym___thiscall] = ACTIONS(5509),
    [anon_sym___vectorcall] = ACTIONS(5509),
    [anon_sym_signed] = ACTIONS(5509),
    [anon_sym_unsigned] = ACTIONS(5509),
    [anon_sym_long] = ACTIONS(5509),
    [anon_sym_short] = ACTIONS(5509),
    [anon_sym_static] = ACTIONS(5509),
    [anon_sym_auto] = ACTIONS(5509),
    [anon_sym_register] = ACTIONS(5509),
    [anon_sym_inline] = ACTIONS(5509),
    [anon_sym___inline] = ACTIONS(5509),
    [anon_sym___inline__] = ACTIONS(5509),
    [anon_sym___forceinline] = ACTIONS(5509),
    [anon_sym_thread_local] = ACTIONS(5509),
    [anon_sym___thread] = ACTIONS(5509),
    [anon_sym_CG_EXTERN] = ACTIONS(5509),
    [anon_sym_CG_INLINE] = ACTIONS(5509),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5509),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5509),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5509),
    [anon_sym_IBOutlet] = ACTIONS(5509),
    [anon_sym_IBInspectable] = ACTIONS(5509),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5509),
    [anon_sym_NS_INLINE] = ACTIONS(5509),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5509),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5509),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5509),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5509),
    [anon_sym_const] = ACTIONS(5509),
    [anon_sym_constexpr] = ACTIONS(5509),
    [anon_sym_volatile] = ACTIONS(5509),
    [anon_sym_restrict] = ACTIONS(5509),
    [anon_sym___restrict__] = ACTIONS(5509),
    [anon_sym__Atomic] = ACTIONS(5509),
    [anon_sym__Noreturn] = ACTIONS(5509),
    [anon_sym__Nonnull] = ACTIONS(5509),
    [anon_sym_nullable] = ACTIONS(5509),
    [anon_sym__Complex] = ACTIONS(5509),
    [anon_sym__Nullable] = ACTIONS(5509),
    [anon_sym__Nullable_result] = ACTIONS(5509),
    [anon_sym__Null_unspecified] = ACTIONS(5509),
    [anon_sym___autoreleasing] = ACTIONS(5509),
    [anon_sym___block] = ACTIONS(5509),
    [anon_sym___bridge] = ACTIONS(5509),
    [anon_sym___bridge_retained] = ACTIONS(5509),
    [anon_sym___bridge_transfer] = ACTIONS(5509),
    [anon_sym___complex] = ACTIONS(5509),
    [anon_sym___const] = ACTIONS(5509),
    [anon_sym___imag] = ACTIONS(5509),
    [anon_sym___kindof] = ACTIONS(5509),
    [anon_sym___nonnull] = ACTIONS(5509),
    [anon_sym___nullable] = ACTIONS(5509),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5509),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5509),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5509),
    [anon_sym___real] = ACTIONS(5509),
    [anon_sym___strong] = ACTIONS(5509),
    [anon_sym___unsafe_unretained] = ACTIONS(5509),
    [anon_sym___unused] = ACTIONS(5509),
    [anon_sym___weak] = ACTIONS(5509),
    [anon_sym_alignas] = ACTIONS(5509),
    [anon_sym__Alignas] = ACTIONS(5509),
    [sym_primitive_type] = ACTIONS(5509),
    [anon_sym_enum] = ACTIONS(5509),
    [anon_sym_struct] = ACTIONS(5509),
    [anon_sym_union] = ACTIONS(5509),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5509),
    [anon_sym___typeof] = ACTIONS(5509),
    [anon_sym_typeof] = ACTIONS(5509),
    [aux_sym_preproc_undef_token1] = ACTIONS(5509),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5509),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5509),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5509),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5509),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5509),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5509),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5509),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5509),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5509),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_API_AVAILABLE] = ACTIONS(5509),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_API_DEPRECATED] = ACTIONS(5509),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5509),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5509),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5509),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5509),
    [anon_sym___deprecated_msg] = ACTIONS(5509),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5509),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5509),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5509),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5509),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5509),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5509),
    [anon_sym_ATend] = ACTIONS(5511),
    [anon_sym_ATsynthesize] = ACTIONS(5511),
    [anon_sym_ATdynamic] = ACTIONS(5511),
    [anon_sym_BOOL] = ACTIONS(5509),
    [anon_sym_IMP] = ACTIONS(5509),
    [anon_sym_SEL] = ACTIONS(5509),
    [anon_sym_Class] = ACTIONS(5509),
    [anon_sym_id] = ACTIONS(5509),
  },
  [STATE(2403)] = {
    [sym_identifier] = ACTIONS(5529),
    [aux_sym_preproc_def_token1] = ACTIONS(5529),
    [aux_sym_preproc_if_token1] = ACTIONS(5529),
    [aux_sym_preproc_if_token2] = ACTIONS(5529),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5529),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5529),
    [sym_preproc_directive] = ACTIONS(5529),
    [anon_sym_DASH] = ACTIONS(5531),
    [anon_sym_PLUS] = ACTIONS(5531),
    [anon_sym___extension__] = ACTIONS(5529),
    [anon_sym_typedef] = ACTIONS(5529),
    [anon_sym_extern] = ACTIONS(5529),
    [anon_sym___attribute__] = ACTIONS(5529),
    [anon_sym___attribute] = ACTIONS(5529),
    [anon_sym_noreturn] = ACTIONS(5529),
    [anon_sym_LBRACK] = ACTIONS(5531),
    [anon_sym___declspec] = ACTIONS(5529),
    [anon_sym___cdecl] = ACTIONS(5529),
    [anon_sym___clrcall] = ACTIONS(5529),
    [anon_sym___stdcall] = ACTIONS(5529),
    [anon_sym___fastcall] = ACTIONS(5529),
    [anon_sym___thiscall] = ACTIONS(5529),
    [anon_sym___vectorcall] = ACTIONS(5529),
    [anon_sym_signed] = ACTIONS(5529),
    [anon_sym_unsigned] = ACTIONS(5529),
    [anon_sym_long] = ACTIONS(5529),
    [anon_sym_short] = ACTIONS(5529),
    [anon_sym_static] = ACTIONS(5529),
    [anon_sym_auto] = ACTIONS(5529),
    [anon_sym_register] = ACTIONS(5529),
    [anon_sym_inline] = ACTIONS(5529),
    [anon_sym___inline] = ACTIONS(5529),
    [anon_sym___inline__] = ACTIONS(5529),
    [anon_sym___forceinline] = ACTIONS(5529),
    [anon_sym_thread_local] = ACTIONS(5529),
    [anon_sym___thread] = ACTIONS(5529),
    [anon_sym_CG_EXTERN] = ACTIONS(5529),
    [anon_sym_CG_INLINE] = ACTIONS(5529),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5529),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5529),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5529),
    [anon_sym_IBOutlet] = ACTIONS(5529),
    [anon_sym_IBInspectable] = ACTIONS(5529),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5529),
    [anon_sym_NS_INLINE] = ACTIONS(5529),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5529),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5529),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5529),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5529),
    [anon_sym_const] = ACTIONS(5529),
    [anon_sym_constexpr] = ACTIONS(5529),
    [anon_sym_volatile] = ACTIONS(5529),
    [anon_sym_restrict] = ACTIONS(5529),
    [anon_sym___restrict__] = ACTIONS(5529),
    [anon_sym__Atomic] = ACTIONS(5529),
    [anon_sym__Noreturn] = ACTIONS(5529),
    [anon_sym__Nonnull] = ACTIONS(5529),
    [anon_sym_nullable] = ACTIONS(5529),
    [anon_sym__Complex] = ACTIONS(5529),
    [anon_sym__Nullable] = ACTIONS(5529),
    [anon_sym__Nullable_result] = ACTIONS(5529),
    [anon_sym__Null_unspecified] = ACTIONS(5529),
    [anon_sym___autoreleasing] = ACTIONS(5529),
    [anon_sym___block] = ACTIONS(5529),
    [anon_sym___bridge] = ACTIONS(5529),
    [anon_sym___bridge_retained] = ACTIONS(5529),
    [anon_sym___bridge_transfer] = ACTIONS(5529),
    [anon_sym___complex] = ACTIONS(5529),
    [anon_sym___const] = ACTIONS(5529),
    [anon_sym___imag] = ACTIONS(5529),
    [anon_sym___kindof] = ACTIONS(5529),
    [anon_sym___nonnull] = ACTIONS(5529),
    [anon_sym___nullable] = ACTIONS(5529),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5529),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5529),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5529),
    [anon_sym___real] = ACTIONS(5529),
    [anon_sym___strong] = ACTIONS(5529),
    [anon_sym___unsafe_unretained] = ACTIONS(5529),
    [anon_sym___unused] = ACTIONS(5529),
    [anon_sym___weak] = ACTIONS(5529),
    [anon_sym_alignas] = ACTIONS(5529),
    [anon_sym__Alignas] = ACTIONS(5529),
    [sym_primitive_type] = ACTIONS(5529),
    [anon_sym_enum] = ACTIONS(5529),
    [anon_sym_struct] = ACTIONS(5529),
    [anon_sym_union] = ACTIONS(5529),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5529),
    [anon_sym___typeof] = ACTIONS(5529),
    [anon_sym_typeof] = ACTIONS(5529),
    [aux_sym_preproc_undef_token1] = ACTIONS(5529),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5529),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5529),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5529),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5529),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5529),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5529),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5529),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5529),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5529),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_API_AVAILABLE] = ACTIONS(5529),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_API_DEPRECATED] = ACTIONS(5529),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5529),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5529),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5529),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5529),
    [anon_sym___deprecated_msg] = ACTIONS(5529),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5529),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5529),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5529),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5529),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5529),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5529),
    [anon_sym_ATproperty] = ACTIONS(5531),
    [anon_sym_BOOL] = ACTIONS(5529),
    [anon_sym_IMP] = ACTIONS(5529),
    [anon_sym_SEL] = ACTIONS(5529),
    [anon_sym_Class] = ACTIONS(5529),
    [anon_sym_id] = ACTIONS(5529),
  },
  [STATE(2404)] = {
    [sym_identifier] = ACTIONS(5595),
    [aux_sym_preproc_def_token1] = ACTIONS(5595),
    [aux_sym_preproc_if_token1] = ACTIONS(5595),
    [aux_sym_preproc_if_token2] = ACTIONS(5595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5595),
    [sym_preproc_directive] = ACTIONS(5595),
    [anon_sym_DASH] = ACTIONS(5597),
    [anon_sym_PLUS] = ACTIONS(5597),
    [anon_sym___extension__] = ACTIONS(5595),
    [anon_sym_typedef] = ACTIONS(5595),
    [anon_sym_extern] = ACTIONS(5595),
    [anon_sym___attribute__] = ACTIONS(5595),
    [anon_sym___attribute] = ACTIONS(5595),
    [anon_sym_noreturn] = ACTIONS(5595),
    [anon_sym_LBRACK] = ACTIONS(5597),
    [anon_sym___declspec] = ACTIONS(5595),
    [anon_sym___cdecl] = ACTIONS(5595),
    [anon_sym___clrcall] = ACTIONS(5595),
    [anon_sym___stdcall] = ACTIONS(5595),
    [anon_sym___fastcall] = ACTIONS(5595),
    [anon_sym___thiscall] = ACTIONS(5595),
    [anon_sym___vectorcall] = ACTIONS(5595),
    [anon_sym_signed] = ACTIONS(5595),
    [anon_sym_unsigned] = ACTIONS(5595),
    [anon_sym_long] = ACTIONS(5595),
    [anon_sym_short] = ACTIONS(5595),
    [anon_sym_static] = ACTIONS(5595),
    [anon_sym_auto] = ACTIONS(5595),
    [anon_sym_register] = ACTIONS(5595),
    [anon_sym_inline] = ACTIONS(5595),
    [anon_sym___inline] = ACTIONS(5595),
    [anon_sym___inline__] = ACTIONS(5595),
    [anon_sym___forceinline] = ACTIONS(5595),
    [anon_sym_thread_local] = ACTIONS(5595),
    [anon_sym___thread] = ACTIONS(5595),
    [anon_sym_CG_EXTERN] = ACTIONS(5595),
    [anon_sym_CG_INLINE] = ACTIONS(5595),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5595),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5595),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5595),
    [anon_sym_IBOutlet] = ACTIONS(5595),
    [anon_sym_IBInspectable] = ACTIONS(5595),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5595),
    [anon_sym_NS_INLINE] = ACTIONS(5595),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5595),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5595),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5595),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5595),
    [anon_sym_const] = ACTIONS(5595),
    [anon_sym_constexpr] = ACTIONS(5595),
    [anon_sym_volatile] = ACTIONS(5595),
    [anon_sym_restrict] = ACTIONS(5595),
    [anon_sym___restrict__] = ACTIONS(5595),
    [anon_sym__Atomic] = ACTIONS(5595),
    [anon_sym__Noreturn] = ACTIONS(5595),
    [anon_sym__Nonnull] = ACTIONS(5595),
    [anon_sym_nullable] = ACTIONS(5595),
    [anon_sym__Complex] = ACTIONS(5595),
    [anon_sym__Nullable] = ACTIONS(5595),
    [anon_sym__Nullable_result] = ACTIONS(5595),
    [anon_sym__Null_unspecified] = ACTIONS(5595),
    [anon_sym___autoreleasing] = ACTIONS(5595),
    [anon_sym___block] = ACTIONS(5595),
    [anon_sym___bridge] = ACTIONS(5595),
    [anon_sym___bridge_retained] = ACTIONS(5595),
    [anon_sym___bridge_transfer] = ACTIONS(5595),
    [anon_sym___complex] = ACTIONS(5595),
    [anon_sym___const] = ACTIONS(5595),
    [anon_sym___imag] = ACTIONS(5595),
    [anon_sym___kindof] = ACTIONS(5595),
    [anon_sym___nonnull] = ACTIONS(5595),
    [anon_sym___nullable] = ACTIONS(5595),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5595),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5595),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5595),
    [anon_sym___real] = ACTIONS(5595),
    [anon_sym___strong] = ACTIONS(5595),
    [anon_sym___unsafe_unretained] = ACTIONS(5595),
    [anon_sym___unused] = ACTIONS(5595),
    [anon_sym___weak] = ACTIONS(5595),
    [anon_sym_alignas] = ACTIONS(5595),
    [anon_sym__Alignas] = ACTIONS(5595),
    [sym_primitive_type] = ACTIONS(5595),
    [anon_sym_enum] = ACTIONS(5595),
    [anon_sym_struct] = ACTIONS(5595),
    [anon_sym_union] = ACTIONS(5595),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5595),
    [anon_sym___typeof] = ACTIONS(5595),
    [anon_sym_typeof] = ACTIONS(5595),
    [aux_sym_preproc_undef_token1] = ACTIONS(5595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5595),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5595),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5595),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5595),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5595),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5595),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5595),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5595),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5595),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_API_AVAILABLE] = ACTIONS(5595),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_API_DEPRECATED] = ACTIONS(5595),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5595),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5595),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5595),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5595),
    [anon_sym___deprecated_msg] = ACTIONS(5595),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5595),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5595),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5595),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5595),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5595),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5595),
    [anon_sym_ATproperty] = ACTIONS(5597),
    [anon_sym_BOOL] = ACTIONS(5595),
    [anon_sym_IMP] = ACTIONS(5595),
    [anon_sym_SEL] = ACTIONS(5595),
    [anon_sym_Class] = ACTIONS(5595),
    [anon_sym_id] = ACTIONS(5595),
  },
  [STATE(2405)] = {
    [sym__declaration_modifiers] = STATE(2676),
    [sym_attribute_specifier] = STATE(2676),
    [sym_attribute_declaration] = STATE(2676),
    [sym_ms_declspec_modifier] = STATE(2676),
    [sym_ms_based_modifier] = STATE(6578),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__type_declarator] = STATE(3260),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(2676),
    [sym_type_qualifier] = STATE(2676),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2676),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2676),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5690),
    [anon_sym_CARET] = ACTIONS(5692),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2406)] = {
    [sym__declaration_modifiers] = STATE(2676),
    [sym_attribute_specifier] = STATE(2676),
    [sym_attribute_declaration] = STATE(2676),
    [sym_ms_declspec_modifier] = STATE(2676),
    [sym_ms_based_modifier] = STATE(6578),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2407),
    [sym__type_declarator] = STATE(3260),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(2676),
    [sym_type_qualifier] = STATE(2676),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2676),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2676),
    [aux_sym_pointer_declarator_repeat1] = STATE(2407),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5690),
    [anon_sym_CARET] = ACTIONS(5692),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2407)] = {
    [sym__declaration_modifiers] = STATE(2670),
    [sym_attribute_specifier] = STATE(2670),
    [sym_attribute_declaration] = STATE(2670),
    [sym_ms_declspec_modifier] = STATE(2670),
    [sym_ms_based_modifier] = STATE(6578),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__type_declarator] = STATE(3245),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(2670),
    [sym_type_qualifier] = STATE(2670),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2670),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2670),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5690),
    [anon_sym_CARET] = ACTIONS(5692),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2408)] = {
    [sym_identifier] = ACTIONS(5577),
    [aux_sym_preproc_def_token1] = ACTIONS(5577),
    [aux_sym_preproc_if_token1] = ACTIONS(5577),
    [aux_sym_preproc_if_token2] = ACTIONS(5577),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5577),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5577),
    [sym_preproc_directive] = ACTIONS(5577),
    [anon_sym_DASH] = ACTIONS(5579),
    [anon_sym_PLUS] = ACTIONS(5579),
    [anon_sym___extension__] = ACTIONS(5577),
    [anon_sym_typedef] = ACTIONS(5577),
    [anon_sym_extern] = ACTIONS(5577),
    [anon_sym___attribute__] = ACTIONS(5577),
    [anon_sym___attribute] = ACTIONS(5577),
    [anon_sym_noreturn] = ACTIONS(5577),
    [anon_sym_LBRACK] = ACTIONS(5579),
    [anon_sym___declspec] = ACTIONS(5577),
    [anon_sym___cdecl] = ACTIONS(5577),
    [anon_sym___clrcall] = ACTIONS(5577),
    [anon_sym___stdcall] = ACTIONS(5577),
    [anon_sym___fastcall] = ACTIONS(5577),
    [anon_sym___thiscall] = ACTIONS(5577),
    [anon_sym___vectorcall] = ACTIONS(5577),
    [anon_sym_signed] = ACTIONS(5577),
    [anon_sym_unsigned] = ACTIONS(5577),
    [anon_sym_long] = ACTIONS(5577),
    [anon_sym_short] = ACTIONS(5577),
    [anon_sym_static] = ACTIONS(5577),
    [anon_sym_auto] = ACTIONS(5577),
    [anon_sym_register] = ACTIONS(5577),
    [anon_sym_inline] = ACTIONS(5577),
    [anon_sym___inline] = ACTIONS(5577),
    [anon_sym___inline__] = ACTIONS(5577),
    [anon_sym___forceinline] = ACTIONS(5577),
    [anon_sym_thread_local] = ACTIONS(5577),
    [anon_sym___thread] = ACTIONS(5577),
    [anon_sym_CG_EXTERN] = ACTIONS(5577),
    [anon_sym_CG_INLINE] = ACTIONS(5577),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5577),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5577),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5577),
    [anon_sym_IBOutlet] = ACTIONS(5577),
    [anon_sym_IBInspectable] = ACTIONS(5577),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5577),
    [anon_sym_NS_INLINE] = ACTIONS(5577),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5577),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5577),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5577),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5577),
    [anon_sym_const] = ACTIONS(5577),
    [anon_sym_constexpr] = ACTIONS(5577),
    [anon_sym_volatile] = ACTIONS(5577),
    [anon_sym_restrict] = ACTIONS(5577),
    [anon_sym___restrict__] = ACTIONS(5577),
    [anon_sym__Atomic] = ACTIONS(5577),
    [anon_sym__Noreturn] = ACTIONS(5577),
    [anon_sym__Nonnull] = ACTIONS(5577),
    [anon_sym_nullable] = ACTIONS(5577),
    [anon_sym__Complex] = ACTIONS(5577),
    [anon_sym__Nullable] = ACTIONS(5577),
    [anon_sym__Nullable_result] = ACTIONS(5577),
    [anon_sym__Null_unspecified] = ACTIONS(5577),
    [anon_sym___autoreleasing] = ACTIONS(5577),
    [anon_sym___block] = ACTIONS(5577),
    [anon_sym___bridge] = ACTIONS(5577),
    [anon_sym___bridge_retained] = ACTIONS(5577),
    [anon_sym___bridge_transfer] = ACTIONS(5577),
    [anon_sym___complex] = ACTIONS(5577),
    [anon_sym___const] = ACTIONS(5577),
    [anon_sym___imag] = ACTIONS(5577),
    [anon_sym___kindof] = ACTIONS(5577),
    [anon_sym___nonnull] = ACTIONS(5577),
    [anon_sym___nullable] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5577),
    [anon_sym___real] = ACTIONS(5577),
    [anon_sym___strong] = ACTIONS(5577),
    [anon_sym___unsafe_unretained] = ACTIONS(5577),
    [anon_sym___unused] = ACTIONS(5577),
    [anon_sym___weak] = ACTIONS(5577),
    [anon_sym_alignas] = ACTIONS(5577),
    [anon_sym__Alignas] = ACTIONS(5577),
    [sym_primitive_type] = ACTIONS(5577),
    [anon_sym_enum] = ACTIONS(5577),
    [anon_sym_struct] = ACTIONS(5577),
    [anon_sym_union] = ACTIONS(5577),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5577),
    [anon_sym___typeof] = ACTIONS(5577),
    [anon_sym_typeof] = ACTIONS(5577),
    [aux_sym_preproc_undef_token1] = ACTIONS(5577),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5577),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5577),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5577),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5577),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5577),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5577),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5577),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5577),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5577),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_API_AVAILABLE] = ACTIONS(5577),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_API_DEPRECATED] = ACTIONS(5577),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5577),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5577),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5577),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5577),
    [anon_sym___deprecated_msg] = ACTIONS(5577),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5577),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5577),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5577),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5577),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5577),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5577),
    [anon_sym_ATproperty] = ACTIONS(5579),
    [anon_sym_BOOL] = ACTIONS(5577),
    [anon_sym_IMP] = ACTIONS(5577),
    [anon_sym_SEL] = ACTIONS(5577),
    [anon_sym_Class] = ACTIONS(5577),
    [anon_sym_id] = ACTIONS(5577),
  },
  [STATE(2409)] = {
    [sym_identifier] = ACTIONS(5541),
    [aux_sym_preproc_def_token1] = ACTIONS(5541),
    [aux_sym_preproc_if_token1] = ACTIONS(5541),
    [aux_sym_preproc_if_token2] = ACTIONS(5541),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5541),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5541),
    [sym_preproc_directive] = ACTIONS(5541),
    [anon_sym_DASH] = ACTIONS(5543),
    [anon_sym_PLUS] = ACTIONS(5543),
    [anon_sym___extension__] = ACTIONS(5541),
    [anon_sym_typedef] = ACTIONS(5541),
    [anon_sym_extern] = ACTIONS(5541),
    [anon_sym___attribute__] = ACTIONS(5541),
    [anon_sym___attribute] = ACTIONS(5541),
    [anon_sym_noreturn] = ACTIONS(5541),
    [anon_sym_LBRACK] = ACTIONS(5543),
    [anon_sym___declspec] = ACTIONS(5541),
    [anon_sym___cdecl] = ACTIONS(5541),
    [anon_sym___clrcall] = ACTIONS(5541),
    [anon_sym___stdcall] = ACTIONS(5541),
    [anon_sym___fastcall] = ACTIONS(5541),
    [anon_sym___thiscall] = ACTIONS(5541),
    [anon_sym___vectorcall] = ACTIONS(5541),
    [anon_sym_signed] = ACTIONS(5541),
    [anon_sym_unsigned] = ACTIONS(5541),
    [anon_sym_long] = ACTIONS(5541),
    [anon_sym_short] = ACTIONS(5541),
    [anon_sym_static] = ACTIONS(5541),
    [anon_sym_auto] = ACTIONS(5541),
    [anon_sym_register] = ACTIONS(5541),
    [anon_sym_inline] = ACTIONS(5541),
    [anon_sym___inline] = ACTIONS(5541),
    [anon_sym___inline__] = ACTIONS(5541),
    [anon_sym___forceinline] = ACTIONS(5541),
    [anon_sym_thread_local] = ACTIONS(5541),
    [anon_sym___thread] = ACTIONS(5541),
    [anon_sym_CG_EXTERN] = ACTIONS(5541),
    [anon_sym_CG_INLINE] = ACTIONS(5541),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5541),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5541),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5541),
    [anon_sym_IBOutlet] = ACTIONS(5541),
    [anon_sym_IBInspectable] = ACTIONS(5541),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5541),
    [anon_sym_NS_INLINE] = ACTIONS(5541),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5541),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5541),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5541),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5541),
    [anon_sym_const] = ACTIONS(5541),
    [anon_sym_constexpr] = ACTIONS(5541),
    [anon_sym_volatile] = ACTIONS(5541),
    [anon_sym_restrict] = ACTIONS(5541),
    [anon_sym___restrict__] = ACTIONS(5541),
    [anon_sym__Atomic] = ACTIONS(5541),
    [anon_sym__Noreturn] = ACTIONS(5541),
    [anon_sym__Nonnull] = ACTIONS(5541),
    [anon_sym_nullable] = ACTIONS(5541),
    [anon_sym__Complex] = ACTIONS(5541),
    [anon_sym__Nullable] = ACTIONS(5541),
    [anon_sym__Nullable_result] = ACTIONS(5541),
    [anon_sym__Null_unspecified] = ACTIONS(5541),
    [anon_sym___autoreleasing] = ACTIONS(5541),
    [anon_sym___block] = ACTIONS(5541),
    [anon_sym___bridge] = ACTIONS(5541),
    [anon_sym___bridge_retained] = ACTIONS(5541),
    [anon_sym___bridge_transfer] = ACTIONS(5541),
    [anon_sym___complex] = ACTIONS(5541),
    [anon_sym___const] = ACTIONS(5541),
    [anon_sym___imag] = ACTIONS(5541),
    [anon_sym___kindof] = ACTIONS(5541),
    [anon_sym___nonnull] = ACTIONS(5541),
    [anon_sym___nullable] = ACTIONS(5541),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5541),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5541),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5541),
    [anon_sym___real] = ACTIONS(5541),
    [anon_sym___strong] = ACTIONS(5541),
    [anon_sym___unsafe_unretained] = ACTIONS(5541),
    [anon_sym___unused] = ACTIONS(5541),
    [anon_sym___weak] = ACTIONS(5541),
    [anon_sym_alignas] = ACTIONS(5541),
    [anon_sym__Alignas] = ACTIONS(5541),
    [sym_primitive_type] = ACTIONS(5541),
    [anon_sym_enum] = ACTIONS(5541),
    [anon_sym_struct] = ACTIONS(5541),
    [anon_sym_union] = ACTIONS(5541),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5541),
    [anon_sym___typeof] = ACTIONS(5541),
    [anon_sym_typeof] = ACTIONS(5541),
    [aux_sym_preproc_undef_token1] = ACTIONS(5541),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5541),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5541),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5541),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5541),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5541),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5541),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5541),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5541),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5541),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_API_AVAILABLE] = ACTIONS(5541),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_API_DEPRECATED] = ACTIONS(5541),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5541),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5541),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5541),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5541),
    [anon_sym___deprecated_msg] = ACTIONS(5541),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5541),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5541),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5541),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5541),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5541),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5541),
    [anon_sym_ATproperty] = ACTIONS(5543),
    [anon_sym_BOOL] = ACTIONS(5541),
    [anon_sym_IMP] = ACTIONS(5541),
    [anon_sym_SEL] = ACTIONS(5541),
    [anon_sym_Class] = ACTIONS(5541),
    [anon_sym_id] = ACTIONS(5541),
  },
  [STATE(2410)] = {
    [sym__declaration_modifiers] = STATE(2677),
    [sym_attribute_specifier] = STATE(2677),
    [sym_attribute_declaration] = STATE(2677),
    [sym_ms_declspec_modifier] = STATE(2677),
    [sym_ms_based_modifier] = STATE(6343),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2416),
    [sym__type_declarator] = STATE(5187),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(2677),
    [sym_type_qualifier] = STATE(2677),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2677),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2677),
    [aux_sym_pointer_declarator_repeat1] = STATE(2416),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5698),
    [anon_sym_CARET] = ACTIONS(5700),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2411)] = {
    [sym_identifier] = ACTIONS(3592),
    [aux_sym_preproc_def_token1] = ACTIONS(3592),
    [aux_sym_preproc_if_token1] = ACTIONS(3592),
    [aux_sym_preproc_if_token2] = ACTIONS(3592),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(3592),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(3592),
    [sym_preproc_directive] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(5621),
    [anon_sym_PLUS] = ACTIONS(5621),
    [anon_sym___extension__] = ACTIONS(3592),
    [anon_sym_typedef] = ACTIONS(3592),
    [anon_sym_extern] = ACTIONS(3592),
    [anon_sym___attribute__] = ACTIONS(3592),
    [anon_sym___attribute] = ACTIONS(3592),
    [anon_sym_noreturn] = ACTIONS(3592),
    [anon_sym_LBRACK] = ACTIONS(5621),
    [anon_sym___declspec] = ACTIONS(3592),
    [anon_sym___cdecl] = ACTIONS(3592),
    [anon_sym___clrcall] = ACTIONS(3592),
    [anon_sym___stdcall] = ACTIONS(3592),
    [anon_sym___fastcall] = ACTIONS(3592),
    [anon_sym___thiscall] = ACTIONS(3592),
    [anon_sym___vectorcall] = ACTIONS(3592),
    [anon_sym_signed] = ACTIONS(3592),
    [anon_sym_unsigned] = ACTIONS(3592),
    [anon_sym_long] = ACTIONS(3592),
    [anon_sym_short] = ACTIONS(3592),
    [anon_sym_static] = ACTIONS(3592),
    [anon_sym_auto] = ACTIONS(3592),
    [anon_sym_register] = ACTIONS(3592),
    [anon_sym_inline] = ACTIONS(3592),
    [anon_sym___inline] = ACTIONS(3592),
    [anon_sym___inline__] = ACTIONS(3592),
    [anon_sym___forceinline] = ACTIONS(3592),
    [anon_sym_thread_local] = ACTIONS(3592),
    [anon_sym___thread] = ACTIONS(3592),
    [anon_sym_CG_EXTERN] = ACTIONS(3592),
    [anon_sym_CG_INLINE] = ACTIONS(3592),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(3592),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(3592),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(3592),
    [anon_sym_IBOutlet] = ACTIONS(3592),
    [anon_sym_IBInspectable] = ACTIONS(3592),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(3592),
    [anon_sym_NS_INLINE] = ACTIONS(3592),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(3592),
    [anon_sym_OBJC_EXPORT] = ACTIONS(3592),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(3592),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(3592),
    [anon_sym_const] = ACTIONS(3592),
    [anon_sym_constexpr] = ACTIONS(3592),
    [anon_sym_volatile] = ACTIONS(3592),
    [anon_sym_restrict] = ACTIONS(3592),
    [anon_sym___restrict__] = ACTIONS(3592),
    [anon_sym__Atomic] = ACTIONS(3592),
    [anon_sym__Noreturn] = ACTIONS(3592),
    [anon_sym__Nonnull] = ACTIONS(3592),
    [anon_sym_nullable] = ACTIONS(3592),
    [anon_sym__Complex] = ACTIONS(3592),
    [anon_sym__Nullable] = ACTIONS(3592),
    [anon_sym__Nullable_result] = ACTIONS(3592),
    [anon_sym__Null_unspecified] = ACTIONS(3592),
    [anon_sym___autoreleasing] = ACTIONS(3592),
    [anon_sym___block] = ACTIONS(3592),
    [anon_sym___bridge] = ACTIONS(3592),
    [anon_sym___bridge_retained] = ACTIONS(3592),
    [anon_sym___bridge_transfer] = ACTIONS(3592),
    [anon_sym___complex] = ACTIONS(3592),
    [anon_sym___const] = ACTIONS(3592),
    [anon_sym___imag] = ACTIONS(3592),
    [anon_sym___kindof] = ACTIONS(3592),
    [anon_sym___nonnull] = ACTIONS(3592),
    [anon_sym___nullable] = ACTIONS(3592),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(3592),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(3592),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(3592),
    [anon_sym___real] = ACTIONS(3592),
    [anon_sym___strong] = ACTIONS(3592),
    [anon_sym___unsafe_unretained] = ACTIONS(3592),
    [anon_sym___unused] = ACTIONS(3592),
    [anon_sym___weak] = ACTIONS(3592),
    [anon_sym_alignas] = ACTIONS(3592),
    [anon_sym__Alignas] = ACTIONS(3592),
    [sym_primitive_type] = ACTIONS(3592),
    [anon_sym_enum] = ACTIONS(3592),
    [anon_sym_struct] = ACTIONS(3592),
    [anon_sym_union] = ACTIONS(3592),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(3592),
    [anon_sym___typeof] = ACTIONS(3592),
    [anon_sym_typeof] = ACTIONS(3592),
    [aux_sym_preproc_undef_token1] = ACTIONS(3592),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(3592),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(3592),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(3592),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(3592),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(3592),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(3592),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(3592),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(3592),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(3592),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(3592),
    [anon_sym_NS_AVAILABLE] = ACTIONS(3592),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(3592),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(3592),
    [anon_sym_API_AVAILABLE] = ACTIONS(3592),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(3592),
    [anon_sym_API_DEPRECATED] = ACTIONS(3592),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(3592),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(3592),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(3592),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(3592),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(3592),
    [anon_sym___deprecated_msg] = ACTIONS(3592),
    [anon_sym___deprecated_enum_msg] = ACTIONS(3592),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(3592),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(3592),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(3592),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(3592),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(3592),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(3592),
    [anon_sym_ATproperty] = ACTIONS(5621),
    [anon_sym_BOOL] = ACTIONS(3592),
    [anon_sym_IMP] = ACTIONS(3592),
    [anon_sym_SEL] = ACTIONS(3592),
    [anon_sym_Class] = ACTIONS(3592),
    [anon_sym_id] = ACTIONS(3592),
  },
  [STATE(2412)] = {
    [sym_identifier] = ACTIONS(5561),
    [aux_sym_preproc_def_token1] = ACTIONS(5561),
    [aux_sym_preproc_if_token1] = ACTIONS(5561),
    [aux_sym_preproc_if_token2] = ACTIONS(5561),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5561),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5561),
    [sym_preproc_directive] = ACTIONS(5561),
    [anon_sym_DASH] = ACTIONS(5563),
    [anon_sym_PLUS] = ACTIONS(5563),
    [anon_sym___extension__] = ACTIONS(5561),
    [anon_sym_typedef] = ACTIONS(5561),
    [anon_sym_extern] = ACTIONS(5561),
    [anon_sym___attribute__] = ACTIONS(5561),
    [anon_sym___attribute] = ACTIONS(5561),
    [anon_sym_noreturn] = ACTIONS(5561),
    [anon_sym_LBRACK] = ACTIONS(5563),
    [anon_sym___declspec] = ACTIONS(5561),
    [anon_sym___cdecl] = ACTIONS(5561),
    [anon_sym___clrcall] = ACTIONS(5561),
    [anon_sym___stdcall] = ACTIONS(5561),
    [anon_sym___fastcall] = ACTIONS(5561),
    [anon_sym___thiscall] = ACTIONS(5561),
    [anon_sym___vectorcall] = ACTIONS(5561),
    [anon_sym_signed] = ACTIONS(5561),
    [anon_sym_unsigned] = ACTIONS(5561),
    [anon_sym_long] = ACTIONS(5561),
    [anon_sym_short] = ACTIONS(5561),
    [anon_sym_static] = ACTIONS(5561),
    [anon_sym_auto] = ACTIONS(5561),
    [anon_sym_register] = ACTIONS(5561),
    [anon_sym_inline] = ACTIONS(5561),
    [anon_sym___inline] = ACTIONS(5561),
    [anon_sym___inline__] = ACTIONS(5561),
    [anon_sym___forceinline] = ACTIONS(5561),
    [anon_sym_thread_local] = ACTIONS(5561),
    [anon_sym___thread] = ACTIONS(5561),
    [anon_sym_CG_EXTERN] = ACTIONS(5561),
    [anon_sym_CG_INLINE] = ACTIONS(5561),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5561),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5561),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5561),
    [anon_sym_IBOutlet] = ACTIONS(5561),
    [anon_sym_IBInspectable] = ACTIONS(5561),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5561),
    [anon_sym_NS_INLINE] = ACTIONS(5561),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5561),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5561),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5561),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5561),
    [anon_sym_const] = ACTIONS(5561),
    [anon_sym_constexpr] = ACTIONS(5561),
    [anon_sym_volatile] = ACTIONS(5561),
    [anon_sym_restrict] = ACTIONS(5561),
    [anon_sym___restrict__] = ACTIONS(5561),
    [anon_sym__Atomic] = ACTIONS(5561),
    [anon_sym__Noreturn] = ACTIONS(5561),
    [anon_sym__Nonnull] = ACTIONS(5561),
    [anon_sym_nullable] = ACTIONS(5561),
    [anon_sym__Complex] = ACTIONS(5561),
    [anon_sym__Nullable] = ACTIONS(5561),
    [anon_sym__Nullable_result] = ACTIONS(5561),
    [anon_sym__Null_unspecified] = ACTIONS(5561),
    [anon_sym___autoreleasing] = ACTIONS(5561),
    [anon_sym___block] = ACTIONS(5561),
    [anon_sym___bridge] = ACTIONS(5561),
    [anon_sym___bridge_retained] = ACTIONS(5561),
    [anon_sym___bridge_transfer] = ACTIONS(5561),
    [anon_sym___complex] = ACTIONS(5561),
    [anon_sym___const] = ACTIONS(5561),
    [anon_sym___imag] = ACTIONS(5561),
    [anon_sym___kindof] = ACTIONS(5561),
    [anon_sym___nonnull] = ACTIONS(5561),
    [anon_sym___nullable] = ACTIONS(5561),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5561),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5561),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5561),
    [anon_sym___real] = ACTIONS(5561),
    [anon_sym___strong] = ACTIONS(5561),
    [anon_sym___unsafe_unretained] = ACTIONS(5561),
    [anon_sym___unused] = ACTIONS(5561),
    [anon_sym___weak] = ACTIONS(5561),
    [anon_sym_alignas] = ACTIONS(5561),
    [anon_sym__Alignas] = ACTIONS(5561),
    [sym_primitive_type] = ACTIONS(5561),
    [anon_sym_enum] = ACTIONS(5561),
    [anon_sym_struct] = ACTIONS(5561),
    [anon_sym_union] = ACTIONS(5561),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5561),
    [anon_sym___typeof] = ACTIONS(5561),
    [anon_sym_typeof] = ACTIONS(5561),
    [aux_sym_preproc_undef_token1] = ACTIONS(5561),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5561),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5561),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5561),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5561),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5561),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5561),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5561),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5561),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5561),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_API_AVAILABLE] = ACTIONS(5561),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_API_DEPRECATED] = ACTIONS(5561),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5561),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5561),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5561),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5561),
    [anon_sym___deprecated_msg] = ACTIONS(5561),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5561),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5561),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5561),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5561),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5561),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5561),
    [anon_sym_ATproperty] = ACTIONS(5563),
    [anon_sym_BOOL] = ACTIONS(5561),
    [anon_sym_IMP] = ACTIONS(5561),
    [anon_sym_SEL] = ACTIONS(5561),
    [anon_sym_Class] = ACTIONS(5561),
    [anon_sym_id] = ACTIONS(5561),
  },
  [STATE(2413)] = {
    [sym__declaration_modifiers] = STATE(2664),
    [sym_attribute_specifier] = STATE(2664),
    [sym_attribute_declaration] = STATE(2664),
    [sym_ms_declspec_modifier] = STATE(2664),
    [sym_ms_based_modifier] = STATE(6343),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__type_declarator] = STATE(5216),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(2664),
    [sym_type_qualifier] = STATE(2664),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2664),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2664),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5698),
    [anon_sym_CARET] = ACTIONS(5700),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2414)] = {
    [sym__declaration_modifiers] = STATE(2672),
    [sym_attribute_specifier] = STATE(2672),
    [sym_attribute_declaration] = STATE(2672),
    [sym_ms_declspec_modifier] = STATE(2672),
    [sym_ms_based_modifier] = STATE(6578),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2405),
    [sym__type_declarator] = STATE(3246),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(2672),
    [sym_type_qualifier] = STATE(2672),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2672),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2672),
    [aux_sym_pointer_declarator_repeat1] = STATE(2405),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5690),
    [anon_sym_CARET] = ACTIONS(5692),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2415)] = {
    [sym_identifier] = ACTIONS(5628),
    [aux_sym_preproc_def_token1] = ACTIONS(5628),
    [aux_sym_preproc_if_token1] = ACTIONS(5628),
    [aux_sym_preproc_if_token2] = ACTIONS(5628),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5628),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5628),
    [sym_preproc_directive] = ACTIONS(5628),
    [anon_sym_DASH] = ACTIONS(5630),
    [anon_sym_PLUS] = ACTIONS(5630),
    [anon_sym___extension__] = ACTIONS(5628),
    [anon_sym_typedef] = ACTIONS(5628),
    [anon_sym_extern] = ACTIONS(5628),
    [anon_sym___attribute__] = ACTIONS(5628),
    [anon_sym___attribute] = ACTIONS(5628),
    [anon_sym_noreturn] = ACTIONS(5628),
    [anon_sym_LBRACK] = ACTIONS(5630),
    [anon_sym___declspec] = ACTIONS(5628),
    [anon_sym___cdecl] = ACTIONS(5628),
    [anon_sym___clrcall] = ACTIONS(5628),
    [anon_sym___stdcall] = ACTIONS(5628),
    [anon_sym___fastcall] = ACTIONS(5628),
    [anon_sym___thiscall] = ACTIONS(5628),
    [anon_sym___vectorcall] = ACTIONS(5628),
    [anon_sym_signed] = ACTIONS(5628),
    [anon_sym_unsigned] = ACTIONS(5628),
    [anon_sym_long] = ACTIONS(5628),
    [anon_sym_short] = ACTIONS(5628),
    [anon_sym_static] = ACTIONS(5628),
    [anon_sym_auto] = ACTIONS(5628),
    [anon_sym_register] = ACTIONS(5628),
    [anon_sym_inline] = ACTIONS(5628),
    [anon_sym___inline] = ACTIONS(5628),
    [anon_sym___inline__] = ACTIONS(5628),
    [anon_sym___forceinline] = ACTIONS(5628),
    [anon_sym_thread_local] = ACTIONS(5628),
    [anon_sym___thread] = ACTIONS(5628),
    [anon_sym_CG_EXTERN] = ACTIONS(5628),
    [anon_sym_CG_INLINE] = ACTIONS(5628),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5628),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5628),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5628),
    [anon_sym_IBOutlet] = ACTIONS(5628),
    [anon_sym_IBInspectable] = ACTIONS(5628),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5628),
    [anon_sym_NS_INLINE] = ACTIONS(5628),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5628),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5628),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5628),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5628),
    [anon_sym_const] = ACTIONS(5628),
    [anon_sym_constexpr] = ACTIONS(5628),
    [anon_sym_volatile] = ACTIONS(5628),
    [anon_sym_restrict] = ACTIONS(5628),
    [anon_sym___restrict__] = ACTIONS(5628),
    [anon_sym__Atomic] = ACTIONS(5628),
    [anon_sym__Noreturn] = ACTIONS(5628),
    [anon_sym__Nonnull] = ACTIONS(5628),
    [anon_sym_nullable] = ACTIONS(5628),
    [anon_sym__Complex] = ACTIONS(5628),
    [anon_sym__Nullable] = ACTIONS(5628),
    [anon_sym__Nullable_result] = ACTIONS(5628),
    [anon_sym__Null_unspecified] = ACTIONS(5628),
    [anon_sym___autoreleasing] = ACTIONS(5628),
    [anon_sym___block] = ACTIONS(5628),
    [anon_sym___bridge] = ACTIONS(5628),
    [anon_sym___bridge_retained] = ACTIONS(5628),
    [anon_sym___bridge_transfer] = ACTIONS(5628),
    [anon_sym___complex] = ACTIONS(5628),
    [anon_sym___const] = ACTIONS(5628),
    [anon_sym___imag] = ACTIONS(5628),
    [anon_sym___kindof] = ACTIONS(5628),
    [anon_sym___nonnull] = ACTIONS(5628),
    [anon_sym___nullable] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5628),
    [anon_sym___real] = ACTIONS(5628),
    [anon_sym___strong] = ACTIONS(5628),
    [anon_sym___unsafe_unretained] = ACTIONS(5628),
    [anon_sym___unused] = ACTIONS(5628),
    [anon_sym___weak] = ACTIONS(5628),
    [anon_sym_alignas] = ACTIONS(5628),
    [anon_sym__Alignas] = ACTIONS(5628),
    [sym_primitive_type] = ACTIONS(5628),
    [anon_sym_enum] = ACTIONS(5628),
    [anon_sym_struct] = ACTIONS(5628),
    [anon_sym_union] = ACTIONS(5628),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5628),
    [anon_sym___typeof] = ACTIONS(5628),
    [anon_sym_typeof] = ACTIONS(5628),
    [aux_sym_preproc_undef_token1] = ACTIONS(5628),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5628),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5628),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5628),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5628),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5628),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5628),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5628),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5628),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5628),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_API_AVAILABLE] = ACTIONS(5628),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_API_DEPRECATED] = ACTIONS(5628),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5628),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5628),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5628),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5628),
    [anon_sym___deprecated_msg] = ACTIONS(5628),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5628),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5628),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5628),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5628),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5628),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5628),
    [anon_sym_ATproperty] = ACTIONS(5630),
    [anon_sym_BOOL] = ACTIONS(5628),
    [anon_sym_IMP] = ACTIONS(5628),
    [anon_sym_SEL] = ACTIONS(5628),
    [anon_sym_Class] = ACTIONS(5628),
    [anon_sym_id] = ACTIONS(5628),
  },
  [STATE(2416)] = {
    [sym__declaration_modifiers] = STATE(2665),
    [sym_attribute_specifier] = STATE(2665),
    [sym_attribute_declaration] = STATE(2665),
    [sym_ms_declspec_modifier] = STATE(2665),
    [sym_ms_based_modifier] = STATE(6343),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__type_declarator] = STATE(5164),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(2665),
    [sym_type_qualifier] = STATE(2665),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2665),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2665),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5698),
    [anon_sym_CARET] = ACTIONS(5700),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2417)] = {
    [sym_identifier] = ACTIONS(5609),
    [aux_sym_preproc_def_token1] = ACTIONS(5609),
    [aux_sym_preproc_if_token1] = ACTIONS(5609),
    [aux_sym_preproc_if_token2] = ACTIONS(5609),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5609),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5609),
    [sym_preproc_directive] = ACTIONS(5609),
    [anon_sym_DASH] = ACTIONS(5611),
    [anon_sym_PLUS] = ACTIONS(5611),
    [anon_sym___extension__] = ACTIONS(5609),
    [anon_sym_typedef] = ACTIONS(5609),
    [anon_sym_extern] = ACTIONS(5609),
    [anon_sym___attribute__] = ACTIONS(5609),
    [anon_sym___attribute] = ACTIONS(5609),
    [anon_sym_noreturn] = ACTIONS(5609),
    [anon_sym_LBRACK] = ACTIONS(5611),
    [anon_sym___declspec] = ACTIONS(5609),
    [anon_sym___cdecl] = ACTIONS(5609),
    [anon_sym___clrcall] = ACTIONS(5609),
    [anon_sym___stdcall] = ACTIONS(5609),
    [anon_sym___fastcall] = ACTIONS(5609),
    [anon_sym___thiscall] = ACTIONS(5609),
    [anon_sym___vectorcall] = ACTIONS(5609),
    [anon_sym_signed] = ACTIONS(5609),
    [anon_sym_unsigned] = ACTIONS(5609),
    [anon_sym_long] = ACTIONS(5609),
    [anon_sym_short] = ACTIONS(5609),
    [anon_sym_static] = ACTIONS(5609),
    [anon_sym_auto] = ACTIONS(5609),
    [anon_sym_register] = ACTIONS(5609),
    [anon_sym_inline] = ACTIONS(5609),
    [anon_sym___inline] = ACTIONS(5609),
    [anon_sym___inline__] = ACTIONS(5609),
    [anon_sym___forceinline] = ACTIONS(5609),
    [anon_sym_thread_local] = ACTIONS(5609),
    [anon_sym___thread] = ACTIONS(5609),
    [anon_sym_CG_EXTERN] = ACTIONS(5609),
    [anon_sym_CG_INLINE] = ACTIONS(5609),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5609),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5609),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5609),
    [anon_sym_IBOutlet] = ACTIONS(5609),
    [anon_sym_IBInspectable] = ACTIONS(5609),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5609),
    [anon_sym_NS_INLINE] = ACTIONS(5609),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5609),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5609),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5609),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5609),
    [anon_sym_const] = ACTIONS(5609),
    [anon_sym_constexpr] = ACTIONS(5609),
    [anon_sym_volatile] = ACTIONS(5609),
    [anon_sym_restrict] = ACTIONS(5609),
    [anon_sym___restrict__] = ACTIONS(5609),
    [anon_sym__Atomic] = ACTIONS(5609),
    [anon_sym__Noreturn] = ACTIONS(5609),
    [anon_sym__Nonnull] = ACTIONS(5609),
    [anon_sym_nullable] = ACTIONS(5609),
    [anon_sym__Complex] = ACTIONS(5609),
    [anon_sym__Nullable] = ACTIONS(5609),
    [anon_sym__Nullable_result] = ACTIONS(5609),
    [anon_sym__Null_unspecified] = ACTIONS(5609),
    [anon_sym___autoreleasing] = ACTIONS(5609),
    [anon_sym___block] = ACTIONS(5609),
    [anon_sym___bridge] = ACTIONS(5609),
    [anon_sym___bridge_retained] = ACTIONS(5609),
    [anon_sym___bridge_transfer] = ACTIONS(5609),
    [anon_sym___complex] = ACTIONS(5609),
    [anon_sym___const] = ACTIONS(5609),
    [anon_sym___imag] = ACTIONS(5609),
    [anon_sym___kindof] = ACTIONS(5609),
    [anon_sym___nonnull] = ACTIONS(5609),
    [anon_sym___nullable] = ACTIONS(5609),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5609),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5609),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5609),
    [anon_sym___real] = ACTIONS(5609),
    [anon_sym___strong] = ACTIONS(5609),
    [anon_sym___unsafe_unretained] = ACTIONS(5609),
    [anon_sym___unused] = ACTIONS(5609),
    [anon_sym___weak] = ACTIONS(5609),
    [anon_sym_alignas] = ACTIONS(5609),
    [anon_sym__Alignas] = ACTIONS(5609),
    [sym_primitive_type] = ACTIONS(5609),
    [anon_sym_enum] = ACTIONS(5609),
    [anon_sym_struct] = ACTIONS(5609),
    [anon_sym_union] = ACTIONS(5609),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5609),
    [anon_sym___typeof] = ACTIONS(5609),
    [anon_sym_typeof] = ACTIONS(5609),
    [aux_sym_preproc_undef_token1] = ACTIONS(5609),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5609),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5609),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5609),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5609),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5609),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5609),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5609),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5609),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5609),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_API_AVAILABLE] = ACTIONS(5609),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_API_DEPRECATED] = ACTIONS(5609),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5609),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5609),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5609),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5609),
    [anon_sym___deprecated_msg] = ACTIONS(5609),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5609),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5609),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5609),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5609),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5609),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5609),
    [anon_sym_ATproperty] = ACTIONS(5611),
    [anon_sym_BOOL] = ACTIONS(5609),
    [anon_sym_IMP] = ACTIONS(5609),
    [anon_sym_SEL] = ACTIONS(5609),
    [anon_sym_Class] = ACTIONS(5609),
    [anon_sym_id] = ACTIONS(5609),
  },
  [STATE(2418)] = {
    [sym__declaration_modifiers] = STATE(2665),
    [sym_attribute_specifier] = STATE(2665),
    [sym_attribute_declaration] = STATE(2665),
    [sym_ms_declspec_modifier] = STATE(2665),
    [sym_ms_based_modifier] = STATE(6343),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2413),
    [sym__type_declarator] = STATE(5164),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(2665),
    [sym_type_qualifier] = STATE(2665),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2665),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2665),
    [aux_sym_pointer_declarator_repeat1] = STATE(2413),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5698),
    [anon_sym_CARET] = ACTIONS(5700),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2419)] = {
    [sym_compound_statement] = STATE(5809),
    [sym_type_qualifier] = STATE(3322),
    [sym_alignas_qualifier] = STATE(4149),
    [sym_expression] = STATE(4444),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_typeof_specifier] = STATE(5809),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5084),
    [anon_sym_noreturn] = ACTIONS(5092),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(5090),
    [anon_sym_constexpr] = ACTIONS(5092),
    [anon_sym_volatile] = ACTIONS(5092),
    [anon_sym_restrict] = ACTIONS(5092),
    [anon_sym___restrict__] = ACTIONS(5092),
    [anon_sym__Atomic] = ACTIONS(5092),
    [anon_sym__Noreturn] = ACTIONS(5092),
    [anon_sym__Nonnull] = ACTIONS(5092),
    [anon_sym_nullable] = ACTIONS(5092),
    [anon_sym__Complex] = ACTIONS(5092),
    [anon_sym__Nullable] = ACTIONS(5092),
    [anon_sym__Nullable_result] = ACTIONS(5092),
    [anon_sym__Null_unspecified] = ACTIONS(5092),
    [anon_sym___autoreleasing] = ACTIONS(5092),
    [anon_sym___block] = ACTIONS(5092),
    [anon_sym___bridge] = ACTIONS(5092),
    [anon_sym___bridge_retained] = ACTIONS(5092),
    [anon_sym___bridge_transfer] = ACTIONS(5092),
    [anon_sym___complex] = ACTIONS(5092),
    [anon_sym___const] = ACTIONS(5092),
    [anon_sym___imag] = ACTIONS(5094),
    [anon_sym___kindof] = ACTIONS(5092),
    [anon_sym___nonnull] = ACTIONS(5092),
    [anon_sym___nullable] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5092),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5092),
    [anon_sym___real] = ACTIONS(5094),
    [anon_sym___strong] = ACTIONS(5092),
    [anon_sym___unsafe_unretained] = ACTIONS(5092),
    [anon_sym___unused] = ACTIONS(5092),
    [anon_sym___weak] = ACTIONS(5092),
    [anon_sym_alignas] = ACTIONS(5096),
    [anon_sym__Alignas] = ACTIONS(5096),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2420)] = {
    [sym_identifier] = ACTIONS(5581),
    [aux_sym_preproc_def_token1] = ACTIONS(5581),
    [aux_sym_preproc_if_token1] = ACTIONS(5581),
    [aux_sym_preproc_if_token2] = ACTIONS(5581),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5581),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5581),
    [sym_preproc_directive] = ACTIONS(5581),
    [anon_sym_DASH] = ACTIONS(5583),
    [anon_sym_PLUS] = ACTIONS(5583),
    [anon_sym___extension__] = ACTIONS(5581),
    [anon_sym_typedef] = ACTIONS(5581),
    [anon_sym_extern] = ACTIONS(5581),
    [anon_sym___attribute__] = ACTIONS(5581),
    [anon_sym___attribute] = ACTIONS(5581),
    [anon_sym_noreturn] = ACTIONS(5581),
    [anon_sym_LBRACK] = ACTIONS(5583),
    [anon_sym___declspec] = ACTIONS(5581),
    [anon_sym___cdecl] = ACTIONS(5581),
    [anon_sym___clrcall] = ACTIONS(5581),
    [anon_sym___stdcall] = ACTIONS(5581),
    [anon_sym___fastcall] = ACTIONS(5581),
    [anon_sym___thiscall] = ACTIONS(5581),
    [anon_sym___vectorcall] = ACTIONS(5581),
    [anon_sym_signed] = ACTIONS(5581),
    [anon_sym_unsigned] = ACTIONS(5581),
    [anon_sym_long] = ACTIONS(5581),
    [anon_sym_short] = ACTIONS(5581),
    [anon_sym_static] = ACTIONS(5581),
    [anon_sym_auto] = ACTIONS(5581),
    [anon_sym_register] = ACTIONS(5581),
    [anon_sym_inline] = ACTIONS(5581),
    [anon_sym___inline] = ACTIONS(5581),
    [anon_sym___inline__] = ACTIONS(5581),
    [anon_sym___forceinline] = ACTIONS(5581),
    [anon_sym_thread_local] = ACTIONS(5581),
    [anon_sym___thread] = ACTIONS(5581),
    [anon_sym_CG_EXTERN] = ACTIONS(5581),
    [anon_sym_CG_INLINE] = ACTIONS(5581),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5581),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5581),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5581),
    [anon_sym_IBOutlet] = ACTIONS(5581),
    [anon_sym_IBInspectable] = ACTIONS(5581),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5581),
    [anon_sym_NS_INLINE] = ACTIONS(5581),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5581),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5581),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5581),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5581),
    [anon_sym_const] = ACTIONS(5581),
    [anon_sym_constexpr] = ACTIONS(5581),
    [anon_sym_volatile] = ACTIONS(5581),
    [anon_sym_restrict] = ACTIONS(5581),
    [anon_sym___restrict__] = ACTIONS(5581),
    [anon_sym__Atomic] = ACTIONS(5581),
    [anon_sym__Noreturn] = ACTIONS(5581),
    [anon_sym__Nonnull] = ACTIONS(5581),
    [anon_sym_nullable] = ACTIONS(5581),
    [anon_sym__Complex] = ACTIONS(5581),
    [anon_sym__Nullable] = ACTIONS(5581),
    [anon_sym__Nullable_result] = ACTIONS(5581),
    [anon_sym__Null_unspecified] = ACTIONS(5581),
    [anon_sym___autoreleasing] = ACTIONS(5581),
    [anon_sym___block] = ACTIONS(5581),
    [anon_sym___bridge] = ACTIONS(5581),
    [anon_sym___bridge_retained] = ACTIONS(5581),
    [anon_sym___bridge_transfer] = ACTIONS(5581),
    [anon_sym___complex] = ACTIONS(5581),
    [anon_sym___const] = ACTIONS(5581),
    [anon_sym___imag] = ACTIONS(5581),
    [anon_sym___kindof] = ACTIONS(5581),
    [anon_sym___nonnull] = ACTIONS(5581),
    [anon_sym___nullable] = ACTIONS(5581),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5581),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5581),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5581),
    [anon_sym___real] = ACTIONS(5581),
    [anon_sym___strong] = ACTIONS(5581),
    [anon_sym___unsafe_unretained] = ACTIONS(5581),
    [anon_sym___unused] = ACTIONS(5581),
    [anon_sym___weak] = ACTIONS(5581),
    [anon_sym_alignas] = ACTIONS(5581),
    [anon_sym__Alignas] = ACTIONS(5581),
    [sym_primitive_type] = ACTIONS(5581),
    [anon_sym_enum] = ACTIONS(5581),
    [anon_sym_struct] = ACTIONS(5581),
    [anon_sym_union] = ACTIONS(5581),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5581),
    [anon_sym___typeof] = ACTIONS(5581),
    [anon_sym_typeof] = ACTIONS(5581),
    [aux_sym_preproc_undef_token1] = ACTIONS(5581),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5581),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5581),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5581),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5581),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5581),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5581),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5581),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5581),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5581),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_API_AVAILABLE] = ACTIONS(5581),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_API_DEPRECATED] = ACTIONS(5581),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5581),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5581),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5581),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5581),
    [anon_sym___deprecated_msg] = ACTIONS(5581),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5581),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5581),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5581),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5581),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5581),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5581),
    [anon_sym_ATproperty] = ACTIONS(5583),
    [anon_sym_BOOL] = ACTIONS(5581),
    [anon_sym_IMP] = ACTIONS(5581),
    [anon_sym_SEL] = ACTIONS(5581),
    [anon_sym_Class] = ACTIONS(5581),
    [anon_sym_id] = ACTIONS(5581),
  },
  [STATE(2421)] = {
    [sym_identifier] = ACTIONS(5569),
    [aux_sym_preproc_def_token1] = ACTIONS(5569),
    [aux_sym_preproc_if_token1] = ACTIONS(5569),
    [aux_sym_preproc_if_token2] = ACTIONS(5569),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5569),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5569),
    [sym_preproc_directive] = ACTIONS(5569),
    [anon_sym_DASH] = ACTIONS(5571),
    [anon_sym_PLUS] = ACTIONS(5571),
    [anon_sym___extension__] = ACTIONS(5569),
    [anon_sym_typedef] = ACTIONS(5569),
    [anon_sym_extern] = ACTIONS(5569),
    [anon_sym___attribute__] = ACTIONS(5569),
    [anon_sym___attribute] = ACTIONS(5569),
    [anon_sym_noreturn] = ACTIONS(5569),
    [anon_sym_LBRACK] = ACTIONS(5571),
    [anon_sym___declspec] = ACTIONS(5569),
    [anon_sym___cdecl] = ACTIONS(5569),
    [anon_sym___clrcall] = ACTIONS(5569),
    [anon_sym___stdcall] = ACTIONS(5569),
    [anon_sym___fastcall] = ACTIONS(5569),
    [anon_sym___thiscall] = ACTIONS(5569),
    [anon_sym___vectorcall] = ACTIONS(5569),
    [anon_sym_signed] = ACTIONS(5569),
    [anon_sym_unsigned] = ACTIONS(5569),
    [anon_sym_long] = ACTIONS(5569),
    [anon_sym_short] = ACTIONS(5569),
    [anon_sym_static] = ACTIONS(5569),
    [anon_sym_auto] = ACTIONS(5569),
    [anon_sym_register] = ACTIONS(5569),
    [anon_sym_inline] = ACTIONS(5569),
    [anon_sym___inline] = ACTIONS(5569),
    [anon_sym___inline__] = ACTIONS(5569),
    [anon_sym___forceinline] = ACTIONS(5569),
    [anon_sym_thread_local] = ACTIONS(5569),
    [anon_sym___thread] = ACTIONS(5569),
    [anon_sym_CG_EXTERN] = ACTIONS(5569),
    [anon_sym_CG_INLINE] = ACTIONS(5569),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5569),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5569),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5569),
    [anon_sym_IBOutlet] = ACTIONS(5569),
    [anon_sym_IBInspectable] = ACTIONS(5569),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5569),
    [anon_sym_NS_INLINE] = ACTIONS(5569),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5569),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5569),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5569),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5569),
    [anon_sym_const] = ACTIONS(5569),
    [anon_sym_constexpr] = ACTIONS(5569),
    [anon_sym_volatile] = ACTIONS(5569),
    [anon_sym_restrict] = ACTIONS(5569),
    [anon_sym___restrict__] = ACTIONS(5569),
    [anon_sym__Atomic] = ACTIONS(5569),
    [anon_sym__Noreturn] = ACTIONS(5569),
    [anon_sym__Nonnull] = ACTIONS(5569),
    [anon_sym_nullable] = ACTIONS(5569),
    [anon_sym__Complex] = ACTIONS(5569),
    [anon_sym__Nullable] = ACTIONS(5569),
    [anon_sym__Nullable_result] = ACTIONS(5569),
    [anon_sym__Null_unspecified] = ACTIONS(5569),
    [anon_sym___autoreleasing] = ACTIONS(5569),
    [anon_sym___block] = ACTIONS(5569),
    [anon_sym___bridge] = ACTIONS(5569),
    [anon_sym___bridge_retained] = ACTIONS(5569),
    [anon_sym___bridge_transfer] = ACTIONS(5569),
    [anon_sym___complex] = ACTIONS(5569),
    [anon_sym___const] = ACTIONS(5569),
    [anon_sym___imag] = ACTIONS(5569),
    [anon_sym___kindof] = ACTIONS(5569),
    [anon_sym___nonnull] = ACTIONS(5569),
    [anon_sym___nullable] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5569),
    [anon_sym___real] = ACTIONS(5569),
    [anon_sym___strong] = ACTIONS(5569),
    [anon_sym___unsafe_unretained] = ACTIONS(5569),
    [anon_sym___unused] = ACTIONS(5569),
    [anon_sym___weak] = ACTIONS(5569),
    [anon_sym_alignas] = ACTIONS(5569),
    [anon_sym__Alignas] = ACTIONS(5569),
    [sym_primitive_type] = ACTIONS(5569),
    [anon_sym_enum] = ACTIONS(5569),
    [anon_sym_struct] = ACTIONS(5569),
    [anon_sym_union] = ACTIONS(5569),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5569),
    [anon_sym___typeof] = ACTIONS(5569),
    [anon_sym_typeof] = ACTIONS(5569),
    [aux_sym_preproc_undef_token1] = ACTIONS(5569),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5569),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5569),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5569),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5569),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5569),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5569),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5569),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5569),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5569),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_API_AVAILABLE] = ACTIONS(5569),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_API_DEPRECATED] = ACTIONS(5569),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5569),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5569),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5569),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5569),
    [anon_sym___deprecated_msg] = ACTIONS(5569),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5569),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5569),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5569),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5569),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5569),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5569),
    [anon_sym_ATproperty] = ACTIONS(5571),
    [anon_sym_BOOL] = ACTIONS(5569),
    [anon_sym_IMP] = ACTIONS(5569),
    [anon_sym_SEL] = ACTIONS(5569),
    [anon_sym_Class] = ACTIONS(5569),
    [anon_sym_id] = ACTIONS(5569),
  },
  [STATE(2422)] = {
    [sym_identifier] = ACTIONS(5573),
    [aux_sym_preproc_def_token1] = ACTIONS(5573),
    [aux_sym_preproc_if_token1] = ACTIONS(5573),
    [aux_sym_preproc_if_token2] = ACTIONS(5573),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5573),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5573),
    [sym_preproc_directive] = ACTIONS(5573),
    [anon_sym_DASH] = ACTIONS(5575),
    [anon_sym_PLUS] = ACTIONS(5575),
    [anon_sym___extension__] = ACTIONS(5573),
    [anon_sym_typedef] = ACTIONS(5573),
    [anon_sym_extern] = ACTIONS(5573),
    [anon_sym___attribute__] = ACTIONS(5573),
    [anon_sym___attribute] = ACTIONS(5573),
    [anon_sym_noreturn] = ACTIONS(5573),
    [anon_sym_LBRACK] = ACTIONS(5575),
    [anon_sym___declspec] = ACTIONS(5573),
    [anon_sym___cdecl] = ACTIONS(5573),
    [anon_sym___clrcall] = ACTIONS(5573),
    [anon_sym___stdcall] = ACTIONS(5573),
    [anon_sym___fastcall] = ACTIONS(5573),
    [anon_sym___thiscall] = ACTIONS(5573),
    [anon_sym___vectorcall] = ACTIONS(5573),
    [anon_sym_signed] = ACTIONS(5573),
    [anon_sym_unsigned] = ACTIONS(5573),
    [anon_sym_long] = ACTIONS(5573),
    [anon_sym_short] = ACTIONS(5573),
    [anon_sym_static] = ACTIONS(5573),
    [anon_sym_auto] = ACTIONS(5573),
    [anon_sym_register] = ACTIONS(5573),
    [anon_sym_inline] = ACTIONS(5573),
    [anon_sym___inline] = ACTIONS(5573),
    [anon_sym___inline__] = ACTIONS(5573),
    [anon_sym___forceinline] = ACTIONS(5573),
    [anon_sym_thread_local] = ACTIONS(5573),
    [anon_sym___thread] = ACTIONS(5573),
    [anon_sym_CG_EXTERN] = ACTIONS(5573),
    [anon_sym_CG_INLINE] = ACTIONS(5573),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5573),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5573),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5573),
    [anon_sym_IBOutlet] = ACTIONS(5573),
    [anon_sym_IBInspectable] = ACTIONS(5573),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5573),
    [anon_sym_NS_INLINE] = ACTIONS(5573),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5573),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5573),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5573),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5573),
    [anon_sym_const] = ACTIONS(5573),
    [anon_sym_constexpr] = ACTIONS(5573),
    [anon_sym_volatile] = ACTIONS(5573),
    [anon_sym_restrict] = ACTIONS(5573),
    [anon_sym___restrict__] = ACTIONS(5573),
    [anon_sym__Atomic] = ACTIONS(5573),
    [anon_sym__Noreturn] = ACTIONS(5573),
    [anon_sym__Nonnull] = ACTIONS(5573),
    [anon_sym_nullable] = ACTIONS(5573),
    [anon_sym__Complex] = ACTIONS(5573),
    [anon_sym__Nullable] = ACTIONS(5573),
    [anon_sym__Nullable_result] = ACTIONS(5573),
    [anon_sym__Null_unspecified] = ACTIONS(5573),
    [anon_sym___autoreleasing] = ACTIONS(5573),
    [anon_sym___block] = ACTIONS(5573),
    [anon_sym___bridge] = ACTIONS(5573),
    [anon_sym___bridge_retained] = ACTIONS(5573),
    [anon_sym___bridge_transfer] = ACTIONS(5573),
    [anon_sym___complex] = ACTIONS(5573),
    [anon_sym___const] = ACTIONS(5573),
    [anon_sym___imag] = ACTIONS(5573),
    [anon_sym___kindof] = ACTIONS(5573),
    [anon_sym___nonnull] = ACTIONS(5573),
    [anon_sym___nullable] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5573),
    [anon_sym___real] = ACTIONS(5573),
    [anon_sym___strong] = ACTIONS(5573),
    [anon_sym___unsafe_unretained] = ACTIONS(5573),
    [anon_sym___unused] = ACTIONS(5573),
    [anon_sym___weak] = ACTIONS(5573),
    [anon_sym_alignas] = ACTIONS(5573),
    [anon_sym__Alignas] = ACTIONS(5573),
    [sym_primitive_type] = ACTIONS(5573),
    [anon_sym_enum] = ACTIONS(5573),
    [anon_sym_struct] = ACTIONS(5573),
    [anon_sym_union] = ACTIONS(5573),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5573),
    [anon_sym___typeof] = ACTIONS(5573),
    [anon_sym_typeof] = ACTIONS(5573),
    [aux_sym_preproc_undef_token1] = ACTIONS(5573),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5573),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5573),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5573),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5573),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5573),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5573),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5573),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5573),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5573),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_API_AVAILABLE] = ACTIONS(5573),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_API_DEPRECATED] = ACTIONS(5573),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5573),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5573),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5573),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5573),
    [anon_sym___deprecated_msg] = ACTIONS(5573),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5573),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5573),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5573),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5573),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5573),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5573),
    [anon_sym_ATproperty] = ACTIONS(5575),
    [anon_sym_BOOL] = ACTIONS(5573),
    [anon_sym_IMP] = ACTIONS(5573),
    [anon_sym_SEL] = ACTIONS(5573),
    [anon_sym_Class] = ACTIONS(5573),
    [anon_sym_id] = ACTIONS(5573),
  },
  [STATE(2423)] = {
    [sym_identifier] = ACTIONS(5537),
    [aux_sym_preproc_def_token1] = ACTIONS(5537),
    [aux_sym_preproc_if_token1] = ACTIONS(5537),
    [aux_sym_preproc_if_token2] = ACTIONS(5537),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5537),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5537),
    [sym_preproc_directive] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5539),
    [anon_sym_PLUS] = ACTIONS(5539),
    [anon_sym___extension__] = ACTIONS(5537),
    [anon_sym_typedef] = ACTIONS(5537),
    [anon_sym_extern] = ACTIONS(5537),
    [anon_sym___attribute__] = ACTIONS(5537),
    [anon_sym___attribute] = ACTIONS(5537),
    [anon_sym_noreturn] = ACTIONS(5537),
    [anon_sym_LBRACK] = ACTIONS(5539),
    [anon_sym___declspec] = ACTIONS(5537),
    [anon_sym___cdecl] = ACTIONS(5537),
    [anon_sym___clrcall] = ACTIONS(5537),
    [anon_sym___stdcall] = ACTIONS(5537),
    [anon_sym___fastcall] = ACTIONS(5537),
    [anon_sym___thiscall] = ACTIONS(5537),
    [anon_sym___vectorcall] = ACTIONS(5537),
    [anon_sym_signed] = ACTIONS(5537),
    [anon_sym_unsigned] = ACTIONS(5537),
    [anon_sym_long] = ACTIONS(5537),
    [anon_sym_short] = ACTIONS(5537),
    [anon_sym_static] = ACTIONS(5537),
    [anon_sym_auto] = ACTIONS(5537),
    [anon_sym_register] = ACTIONS(5537),
    [anon_sym_inline] = ACTIONS(5537),
    [anon_sym___inline] = ACTIONS(5537),
    [anon_sym___inline__] = ACTIONS(5537),
    [anon_sym___forceinline] = ACTIONS(5537),
    [anon_sym_thread_local] = ACTIONS(5537),
    [anon_sym___thread] = ACTIONS(5537),
    [anon_sym_CG_EXTERN] = ACTIONS(5537),
    [anon_sym_CG_INLINE] = ACTIONS(5537),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5537),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5537),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5537),
    [anon_sym_IBOutlet] = ACTIONS(5537),
    [anon_sym_IBInspectable] = ACTIONS(5537),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5537),
    [anon_sym_NS_INLINE] = ACTIONS(5537),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5537),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5537),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5537),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5537),
    [anon_sym_const] = ACTIONS(5537),
    [anon_sym_constexpr] = ACTIONS(5537),
    [anon_sym_volatile] = ACTIONS(5537),
    [anon_sym_restrict] = ACTIONS(5537),
    [anon_sym___restrict__] = ACTIONS(5537),
    [anon_sym__Atomic] = ACTIONS(5537),
    [anon_sym__Noreturn] = ACTIONS(5537),
    [anon_sym__Nonnull] = ACTIONS(5537),
    [anon_sym_nullable] = ACTIONS(5537),
    [anon_sym__Complex] = ACTIONS(5537),
    [anon_sym__Nullable] = ACTIONS(5537),
    [anon_sym__Nullable_result] = ACTIONS(5537),
    [anon_sym__Null_unspecified] = ACTIONS(5537),
    [anon_sym___autoreleasing] = ACTIONS(5537),
    [anon_sym___block] = ACTIONS(5537),
    [anon_sym___bridge] = ACTIONS(5537),
    [anon_sym___bridge_retained] = ACTIONS(5537),
    [anon_sym___bridge_transfer] = ACTIONS(5537),
    [anon_sym___complex] = ACTIONS(5537),
    [anon_sym___const] = ACTIONS(5537),
    [anon_sym___imag] = ACTIONS(5537),
    [anon_sym___kindof] = ACTIONS(5537),
    [anon_sym___nonnull] = ACTIONS(5537),
    [anon_sym___nullable] = ACTIONS(5537),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5537),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5537),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5537),
    [anon_sym___real] = ACTIONS(5537),
    [anon_sym___strong] = ACTIONS(5537),
    [anon_sym___unsafe_unretained] = ACTIONS(5537),
    [anon_sym___unused] = ACTIONS(5537),
    [anon_sym___weak] = ACTIONS(5537),
    [anon_sym_alignas] = ACTIONS(5537),
    [anon_sym__Alignas] = ACTIONS(5537),
    [sym_primitive_type] = ACTIONS(5537),
    [anon_sym_enum] = ACTIONS(5537),
    [anon_sym_struct] = ACTIONS(5537),
    [anon_sym_union] = ACTIONS(5537),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5537),
    [anon_sym___typeof] = ACTIONS(5537),
    [anon_sym_typeof] = ACTIONS(5537),
    [aux_sym_preproc_undef_token1] = ACTIONS(5537),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5537),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5537),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5537),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5537),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5537),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5537),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5537),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5537),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5537),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_API_AVAILABLE] = ACTIONS(5537),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_API_DEPRECATED] = ACTIONS(5537),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5537),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5537),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5537),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5537),
    [anon_sym___deprecated_msg] = ACTIONS(5537),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5537),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5537),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5537),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5537),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5537),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5537),
    [anon_sym_ATproperty] = ACTIONS(5539),
    [anon_sym_BOOL] = ACTIONS(5537),
    [anon_sym_IMP] = ACTIONS(5537),
    [anon_sym_SEL] = ACTIONS(5537),
    [anon_sym_Class] = ACTIONS(5537),
    [anon_sym_id] = ACTIONS(5537),
  },
  [STATE(2424)] = {
    [sym_identifier] = ACTIONS(5589),
    [aux_sym_preproc_def_token1] = ACTIONS(5589),
    [aux_sym_preproc_if_token1] = ACTIONS(5589),
    [aux_sym_preproc_if_token2] = ACTIONS(5589),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5589),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5589),
    [sym_preproc_directive] = ACTIONS(5589),
    [anon_sym_DASH] = ACTIONS(5591),
    [anon_sym_PLUS] = ACTIONS(5591),
    [anon_sym___extension__] = ACTIONS(5589),
    [anon_sym_typedef] = ACTIONS(5589),
    [anon_sym_extern] = ACTIONS(5589),
    [anon_sym___attribute__] = ACTIONS(5589),
    [anon_sym___attribute] = ACTIONS(5589),
    [anon_sym_noreturn] = ACTIONS(5589),
    [anon_sym_LBRACK] = ACTIONS(5591),
    [anon_sym___declspec] = ACTIONS(5589),
    [anon_sym___cdecl] = ACTIONS(5589),
    [anon_sym___clrcall] = ACTIONS(5589),
    [anon_sym___stdcall] = ACTIONS(5589),
    [anon_sym___fastcall] = ACTIONS(5589),
    [anon_sym___thiscall] = ACTIONS(5589),
    [anon_sym___vectorcall] = ACTIONS(5589),
    [anon_sym_signed] = ACTIONS(5589),
    [anon_sym_unsigned] = ACTIONS(5589),
    [anon_sym_long] = ACTIONS(5589),
    [anon_sym_short] = ACTIONS(5589),
    [anon_sym_static] = ACTIONS(5589),
    [anon_sym_auto] = ACTIONS(5589),
    [anon_sym_register] = ACTIONS(5589),
    [anon_sym_inline] = ACTIONS(5589),
    [anon_sym___inline] = ACTIONS(5589),
    [anon_sym___inline__] = ACTIONS(5589),
    [anon_sym___forceinline] = ACTIONS(5589),
    [anon_sym_thread_local] = ACTIONS(5589),
    [anon_sym___thread] = ACTIONS(5589),
    [anon_sym_CG_EXTERN] = ACTIONS(5589),
    [anon_sym_CG_INLINE] = ACTIONS(5589),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5589),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5589),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5589),
    [anon_sym_IBOutlet] = ACTIONS(5589),
    [anon_sym_IBInspectable] = ACTIONS(5589),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5589),
    [anon_sym_NS_INLINE] = ACTIONS(5589),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5589),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5589),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5589),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5589),
    [anon_sym_const] = ACTIONS(5589),
    [anon_sym_constexpr] = ACTIONS(5589),
    [anon_sym_volatile] = ACTIONS(5589),
    [anon_sym_restrict] = ACTIONS(5589),
    [anon_sym___restrict__] = ACTIONS(5589),
    [anon_sym__Atomic] = ACTIONS(5589),
    [anon_sym__Noreturn] = ACTIONS(5589),
    [anon_sym__Nonnull] = ACTIONS(5589),
    [anon_sym_nullable] = ACTIONS(5589),
    [anon_sym__Complex] = ACTIONS(5589),
    [anon_sym__Nullable] = ACTIONS(5589),
    [anon_sym__Nullable_result] = ACTIONS(5589),
    [anon_sym__Null_unspecified] = ACTIONS(5589),
    [anon_sym___autoreleasing] = ACTIONS(5589),
    [anon_sym___block] = ACTIONS(5589),
    [anon_sym___bridge] = ACTIONS(5589),
    [anon_sym___bridge_retained] = ACTIONS(5589),
    [anon_sym___bridge_transfer] = ACTIONS(5589),
    [anon_sym___complex] = ACTIONS(5589),
    [anon_sym___const] = ACTIONS(5589),
    [anon_sym___imag] = ACTIONS(5589),
    [anon_sym___kindof] = ACTIONS(5589),
    [anon_sym___nonnull] = ACTIONS(5589),
    [anon_sym___nullable] = ACTIONS(5589),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5589),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5589),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5589),
    [anon_sym___real] = ACTIONS(5589),
    [anon_sym___strong] = ACTIONS(5589),
    [anon_sym___unsafe_unretained] = ACTIONS(5589),
    [anon_sym___unused] = ACTIONS(5589),
    [anon_sym___weak] = ACTIONS(5589),
    [anon_sym_alignas] = ACTIONS(5589),
    [anon_sym__Alignas] = ACTIONS(5589),
    [sym_primitive_type] = ACTIONS(5589),
    [anon_sym_enum] = ACTIONS(5589),
    [anon_sym_struct] = ACTIONS(5589),
    [anon_sym_union] = ACTIONS(5589),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5589),
    [anon_sym___typeof] = ACTIONS(5589),
    [anon_sym_typeof] = ACTIONS(5589),
    [aux_sym_preproc_undef_token1] = ACTIONS(5589),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5589),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5589),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5589),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5589),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5589),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5589),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5589),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5589),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5589),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_API_AVAILABLE] = ACTIONS(5589),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_API_DEPRECATED] = ACTIONS(5589),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5589),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5589),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5589),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5589),
    [anon_sym___deprecated_msg] = ACTIONS(5589),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5589),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5589),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5589),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5589),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5589),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5589),
    [anon_sym_ATproperty] = ACTIONS(5591),
    [anon_sym_BOOL] = ACTIONS(5589),
    [anon_sym_IMP] = ACTIONS(5589),
    [anon_sym_SEL] = ACTIONS(5589),
    [anon_sym_Class] = ACTIONS(5589),
    [anon_sym_id] = ACTIONS(5589),
  },
  [STATE(2425)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(6487),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2426)] = {
    [sym_attribute_declaration] = STATE(2436),
    [sym_compound_statement] = STATE(439),
    [sym_attributed_statement] = STATE(439),
    [sym_statement] = STATE(287),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [sym_identifier] = ACTIONS(5706),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2427)] = {
    [sym_attribute_declaration] = STATE(2427),
    [sym_compound_statement] = STATE(439),
    [sym_attributed_statement] = STATE(439),
    [sym_statement] = STATE(391),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2427),
    [sym_identifier] = ACTIONS(5708),
    [anon_sym_LPAREN2] = ACTIONS(5711),
    [anon_sym_BANG] = ACTIONS(5714),
    [anon_sym_TILDE] = ACTIONS(5714),
    [anon_sym_DASH] = ACTIONS(5717),
    [anon_sym_PLUS] = ACTIONS(5717),
    [anon_sym_STAR] = ACTIONS(5720),
    [anon_sym_CARET] = ACTIONS(5723),
    [anon_sym_AMP] = ACTIONS(5720),
    [anon_sym_SEMI] = ACTIONS(5726),
    [anon_sym___extension__] = ACTIONS(5729),
    [anon_sym_LBRACK] = ACTIONS(5732),
    [anon_sym_LBRACE] = ACTIONS(5735),
    [anon_sym_ATautoreleasepool] = ACTIONS(5738),
    [anon_sym_const] = ACTIONS(5741),
    [anon_sym___imag] = ACTIONS(5744),
    [anon_sym___real] = ACTIONS(5744),
    [anon_sym_struct] = ACTIONS(5741),
    [anon_sym_if] = ACTIONS(5747),
    [anon_sym_switch] = ACTIONS(5750),
    [anon_sym_case] = ACTIONS(5753),
    [anon_sym_default] = ACTIONS(5756),
    [anon_sym_while] = ACTIONS(5759),
    [anon_sym_do] = ACTIONS(5762),
    [anon_sym_for] = ACTIONS(5765),
    [anon_sym_in] = ACTIONS(5741),
    [anon_sym_return] = ACTIONS(5768),
    [anon_sym_break] = ACTIONS(5771),
    [anon_sym_continue] = ACTIONS(5774),
    [anon_sym_goto] = ACTIONS(5777),
    [anon_sym_DASH_DASH] = ACTIONS(5780),
    [anon_sym_PLUS_PLUS] = ACTIONS(5780),
    [anon_sym_sizeof] = ACTIONS(5783),
    [anon_sym___alignof__] = ACTIONS(5786),
    [anon_sym___alignof] = ACTIONS(5786),
    [anon_sym__alignof] = ACTIONS(5786),
    [anon_sym_alignof] = ACTIONS(5786),
    [anon_sym__Alignof] = ACTIONS(5786),
    [anon_sym_offsetof] = ACTIONS(5789),
    [anon_sym__Generic] = ACTIONS(5792),
    [anon_sym_asm] = ACTIONS(5795),
    [anon_sym___asm__] = ACTIONS(5795),
    [anon_sym___asm] = ACTIONS(5798),
    [sym_number_literal] = ACTIONS(5801),
    [anon_sym_L_SQUOTE] = ACTIONS(5804),
    [anon_sym_u_SQUOTE] = ACTIONS(5804),
    [anon_sym_U_SQUOTE] = ACTIONS(5804),
    [anon_sym_u8_SQUOTE] = ACTIONS(5804),
    [anon_sym_SQUOTE] = ACTIONS(5804),
    [anon_sym_AT] = ACTIONS(5807),
    [anon_sym_DQUOTE] = ACTIONS(5810),
    [anon_sym_L_DQUOTE] = ACTIONS(5810),
    [anon_sym_u_DQUOTE] = ACTIONS(5810),
    [anon_sym_U_DQUOTE] = ACTIONS(5810),
    [anon_sym_u8_DQUOTE] = ACTIONS(5810),
    [sym_true] = ACTIONS(5813),
    [sym_false] = ACTIONS(5813),
    [anon_sym_NULL] = ACTIONS(5816),
    [anon_sym_nullptr] = ACTIONS(5816),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(5819),
    [anon_sym___try] = ACTIONS(5822),
    [anon_sym_ATthrow] = ACTIONS(5825),
    [anon_sym_ATselector] = ACTIONS(5828),
    [anon_sym_ATavailable] = ACTIONS(5831),
    [anon_sym___builtin_available] = ACTIONS(5834),
    [anon_sym_va_arg] = ACTIONS(5837),
    [anon_sym_ATencode] = ACTIONS(5840),
    [anon_sym_ATsynchronized] = ACTIONS(5843),
    [anon_sym_id] = ACTIONS(5741),
  },
  [STATE(2428)] = {
    [sym_attribute_declaration] = STATE(2438),
    [sym_compound_statement] = STATE(535),
    [sym_attributed_statement] = STATE(535),
    [sym_statement] = STATE(536),
    [sym_labeled_statement] = STATE(535),
    [sym_expression_statement] = STATE(535),
    [sym_if_statement] = STATE(535),
    [sym_switch_statement] = STATE(535),
    [sym_case_statement] = STATE(535),
    [sym_while_statement] = STATE(535),
    [sym_do_statement] = STATE(535),
    [sym_for_statement] = STATE(535),
    [sym_return_statement] = STATE(535),
    [sym_break_statement] = STATE(535),
    [sym_continue_statement] = STATE(535),
    [sym_goto_statement] = STATE(535),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(535),
    [sym_throw_statement] = STATE(535),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(535),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(535),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [sym_identifier] = ACTIONS(5846),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(878),
    [anon_sym_default] = ACTIONS(880),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2429)] = {
    [sym_attribute_declaration] = STATE(2442),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(942),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2442),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2430)] = {
    [sym_attribute_declaration] = STATE(2480),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(640),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [sym_identifier] = ACTIONS(5848),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2431)] = {
    [sym_attribute_declaration] = STATE(2436),
    [sym_compound_statement] = STATE(439),
    [sym_attributed_statement] = STATE(439),
    [sym_statement] = STATE(410),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [sym_identifier] = ACTIONS(5706),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2432)] = {
    [sym_attribute_declaration] = STATE(2446),
    [sym_compound_statement] = STATE(1159),
    [sym_attributed_statement] = STATE(1159),
    [sym_statement] = STATE(1150),
    [sym_labeled_statement] = STATE(1159),
    [sym_expression_statement] = STATE(1159),
    [sym_if_statement] = STATE(1159),
    [sym_switch_statement] = STATE(1159),
    [sym_case_statement] = STATE(1159),
    [sym_while_statement] = STATE(1159),
    [sym_do_statement] = STATE(1159),
    [sym_for_statement] = STATE(1159),
    [sym_return_statement] = STATE(1159),
    [sym_break_statement] = STATE(1159),
    [sym_continue_statement] = STATE(1159),
    [sym_goto_statement] = STATE(1159),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(1159),
    [sym_throw_statement] = STATE(1159),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(1159),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(1159),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [sym_identifier] = ACTIONS(5850),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2433)] = {
    [sym_attribute_declaration] = STATE(2472),
    [sym_compound_statement] = STATE(690),
    [sym_attributed_statement] = STATE(690),
    [sym_statement] = STATE(656),
    [sym_labeled_statement] = STATE(690),
    [sym_expression_statement] = STATE(690),
    [sym_if_statement] = STATE(690),
    [sym_switch_statement] = STATE(690),
    [sym_case_statement] = STATE(690),
    [sym_while_statement] = STATE(690),
    [sym_do_statement] = STATE(690),
    [sym_for_statement] = STATE(690),
    [sym_return_statement] = STATE(690),
    [sym_break_statement] = STATE(690),
    [sym_continue_statement] = STATE(690),
    [sym_goto_statement] = STATE(690),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(690),
    [sym_throw_statement] = STATE(690),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(690),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(690),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [sym_identifier] = ACTIONS(5852),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2434)] = {
    [sym_attribute_declaration] = STATE(2472),
    [sym_compound_statement] = STATE(690),
    [sym_attributed_statement] = STATE(690),
    [sym_statement] = STATE(724),
    [sym_labeled_statement] = STATE(690),
    [sym_expression_statement] = STATE(690),
    [sym_if_statement] = STATE(690),
    [sym_switch_statement] = STATE(690),
    [sym_case_statement] = STATE(690),
    [sym_while_statement] = STATE(690),
    [sym_do_statement] = STATE(690),
    [sym_for_statement] = STATE(690),
    [sym_return_statement] = STATE(690),
    [sym_break_statement] = STATE(690),
    [sym_continue_statement] = STATE(690),
    [sym_goto_statement] = STATE(690),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(690),
    [sym_throw_statement] = STATE(690),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(690),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(690),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [sym_identifier] = ACTIONS(5852),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2435)] = {
    [sym_attribute_declaration] = STATE(2446),
    [sym_compound_statement] = STATE(1159),
    [sym_attributed_statement] = STATE(1159),
    [sym_statement] = STATE(1152),
    [sym_labeled_statement] = STATE(1159),
    [sym_expression_statement] = STATE(1159),
    [sym_if_statement] = STATE(1159),
    [sym_switch_statement] = STATE(1159),
    [sym_case_statement] = STATE(1159),
    [sym_while_statement] = STATE(1159),
    [sym_do_statement] = STATE(1159),
    [sym_for_statement] = STATE(1159),
    [sym_return_statement] = STATE(1159),
    [sym_break_statement] = STATE(1159),
    [sym_continue_statement] = STATE(1159),
    [sym_goto_statement] = STATE(1159),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(1159),
    [sym_throw_statement] = STATE(1159),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(1159),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(1159),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [sym_identifier] = ACTIONS(5850),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2436)] = {
    [sym_attribute_declaration] = STATE(2427),
    [sym_compound_statement] = STATE(439),
    [sym_attributed_statement] = STATE(439),
    [sym_statement] = STATE(391),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2427),
    [sym_identifier] = ACTIONS(5706),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2437)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(5584),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2438)] = {
    [sym_attribute_declaration] = STATE(2447),
    [sym_compound_statement] = STATE(535),
    [sym_attributed_statement] = STATE(535),
    [sym_statement] = STATE(599),
    [sym_labeled_statement] = STATE(535),
    [sym_expression_statement] = STATE(535),
    [sym_if_statement] = STATE(535),
    [sym_switch_statement] = STATE(535),
    [sym_case_statement] = STATE(535),
    [sym_while_statement] = STATE(535),
    [sym_do_statement] = STATE(535),
    [sym_for_statement] = STATE(535),
    [sym_return_statement] = STATE(535),
    [sym_break_statement] = STATE(535),
    [sym_continue_statement] = STATE(535),
    [sym_goto_statement] = STATE(535),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(535),
    [sym_throw_statement] = STATE(535),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(535),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(535),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2447),
    [sym_identifier] = ACTIONS(5846),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(878),
    [anon_sym_default] = ACTIONS(880),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2439)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(948),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2440)] = {
    [sym_attribute_declaration] = STATE(2446),
    [sym_compound_statement] = STATE(1159),
    [sym_attributed_statement] = STATE(1159),
    [sym_statement] = STATE(1085),
    [sym_labeled_statement] = STATE(1159),
    [sym_expression_statement] = STATE(1159),
    [sym_if_statement] = STATE(1159),
    [sym_switch_statement] = STATE(1159),
    [sym_case_statement] = STATE(1159),
    [sym_while_statement] = STATE(1159),
    [sym_do_statement] = STATE(1159),
    [sym_for_statement] = STATE(1159),
    [sym_return_statement] = STATE(1159),
    [sym_break_statement] = STATE(1159),
    [sym_continue_statement] = STATE(1159),
    [sym_goto_statement] = STATE(1159),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(1159),
    [sym_throw_statement] = STATE(1159),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(1159),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(1159),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [sym_identifier] = ACTIONS(5850),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2441)] = {
    [sym_attribute_declaration] = STATE(2446),
    [sym_compound_statement] = STATE(1159),
    [sym_attributed_statement] = STATE(1159),
    [sym_statement] = STATE(1132),
    [sym_labeled_statement] = STATE(1159),
    [sym_expression_statement] = STATE(1159),
    [sym_if_statement] = STATE(1159),
    [sym_switch_statement] = STATE(1159),
    [sym_case_statement] = STATE(1159),
    [sym_while_statement] = STATE(1159),
    [sym_do_statement] = STATE(1159),
    [sym_for_statement] = STATE(1159),
    [sym_return_statement] = STATE(1159),
    [sym_break_statement] = STATE(1159),
    [sym_continue_statement] = STATE(1159),
    [sym_goto_statement] = STATE(1159),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(1159),
    [sym_throw_statement] = STATE(1159),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(1159),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(1159),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [sym_identifier] = ACTIONS(5850),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2442)] = {
    [sym_attribute_declaration] = STATE(2442),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(942),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2442),
    [sym_identifier] = ACTIONS(5854),
    [anon_sym_LPAREN2] = ACTIONS(5711),
    [anon_sym_BANG] = ACTIONS(5714),
    [anon_sym_TILDE] = ACTIONS(5714),
    [anon_sym_DASH] = ACTIONS(5717),
    [anon_sym_PLUS] = ACTIONS(5717),
    [anon_sym_STAR] = ACTIONS(5720),
    [anon_sym_CARET] = ACTIONS(5723),
    [anon_sym_AMP] = ACTIONS(5720),
    [anon_sym_SEMI] = ACTIONS(5857),
    [anon_sym___extension__] = ACTIONS(5729),
    [anon_sym_LBRACK] = ACTIONS(5732),
    [anon_sym_LBRACE] = ACTIONS(5860),
    [anon_sym_ATautoreleasepool] = ACTIONS(5863),
    [anon_sym_const] = ACTIONS(5741),
    [anon_sym___imag] = ACTIONS(5744),
    [anon_sym___real] = ACTIONS(5744),
    [anon_sym_struct] = ACTIONS(5741),
    [anon_sym_if] = ACTIONS(5866),
    [anon_sym_switch] = ACTIONS(5869),
    [anon_sym_case] = ACTIONS(5872),
    [anon_sym_default] = ACTIONS(5875),
    [anon_sym_while] = ACTIONS(5878),
    [anon_sym_do] = ACTIONS(5881),
    [anon_sym_for] = ACTIONS(5884),
    [anon_sym_in] = ACTIONS(5741),
    [anon_sym_return] = ACTIONS(5887),
    [anon_sym_break] = ACTIONS(5890),
    [anon_sym_continue] = ACTIONS(5893),
    [anon_sym_goto] = ACTIONS(5896),
    [anon_sym_DASH_DASH] = ACTIONS(5780),
    [anon_sym_PLUS_PLUS] = ACTIONS(5780),
    [anon_sym_sizeof] = ACTIONS(5783),
    [anon_sym___alignof__] = ACTIONS(5786),
    [anon_sym___alignof] = ACTIONS(5786),
    [anon_sym__alignof] = ACTIONS(5786),
    [anon_sym_alignof] = ACTIONS(5786),
    [anon_sym__Alignof] = ACTIONS(5786),
    [anon_sym_offsetof] = ACTIONS(5789),
    [anon_sym__Generic] = ACTIONS(5792),
    [anon_sym_asm] = ACTIONS(5795),
    [anon_sym___asm__] = ACTIONS(5795),
    [anon_sym___asm] = ACTIONS(5899),
    [sym_number_literal] = ACTIONS(5801),
    [anon_sym_L_SQUOTE] = ACTIONS(5804),
    [anon_sym_u_SQUOTE] = ACTIONS(5804),
    [anon_sym_U_SQUOTE] = ACTIONS(5804),
    [anon_sym_u8_SQUOTE] = ACTIONS(5804),
    [anon_sym_SQUOTE] = ACTIONS(5804),
    [anon_sym_AT] = ACTIONS(5807),
    [anon_sym_DQUOTE] = ACTIONS(5810),
    [anon_sym_L_DQUOTE] = ACTIONS(5810),
    [anon_sym_u_DQUOTE] = ACTIONS(5810),
    [anon_sym_U_DQUOTE] = ACTIONS(5810),
    [anon_sym_u8_DQUOTE] = ACTIONS(5810),
    [sym_true] = ACTIONS(5813),
    [sym_false] = ACTIONS(5813),
    [anon_sym_NULL] = ACTIONS(5816),
    [anon_sym_nullptr] = ACTIONS(5816),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(5902),
    [anon_sym___try] = ACTIONS(5905),
    [anon_sym_ATthrow] = ACTIONS(5908),
    [anon_sym_ATselector] = ACTIONS(5828),
    [anon_sym_ATavailable] = ACTIONS(5831),
    [anon_sym___builtin_available] = ACTIONS(5834),
    [anon_sym_va_arg] = ACTIONS(5837),
    [anon_sym_ATencode] = ACTIONS(5840),
    [anon_sym_ATsynchronized] = ACTIONS(5911),
    [anon_sym_id] = ACTIONS(5741),
  },
  [STATE(2443)] = {
    [sym_attribute_declaration] = STATE(2436),
    [sym_compound_statement] = STATE(439),
    [sym_attributed_statement] = STATE(439),
    [sym_statement] = STATE(431),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [sym_identifier] = ACTIONS(5706),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2444)] = {
    [sym_attribute_declaration] = STATE(2444),
    [sym_compound_statement] = STATE(690),
    [sym_attributed_statement] = STATE(690),
    [sym_statement] = STATE(716),
    [sym_labeled_statement] = STATE(690),
    [sym_expression_statement] = STATE(690),
    [sym_if_statement] = STATE(690),
    [sym_switch_statement] = STATE(690),
    [sym_case_statement] = STATE(690),
    [sym_while_statement] = STATE(690),
    [sym_do_statement] = STATE(690),
    [sym_for_statement] = STATE(690),
    [sym_return_statement] = STATE(690),
    [sym_break_statement] = STATE(690),
    [sym_continue_statement] = STATE(690),
    [sym_goto_statement] = STATE(690),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(690),
    [sym_throw_statement] = STATE(690),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(690),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(690),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2444),
    [sym_identifier] = ACTIONS(5914),
    [anon_sym_LPAREN2] = ACTIONS(5711),
    [anon_sym_BANG] = ACTIONS(5714),
    [anon_sym_TILDE] = ACTIONS(5714),
    [anon_sym_DASH] = ACTIONS(5717),
    [anon_sym_PLUS] = ACTIONS(5717),
    [anon_sym_STAR] = ACTIONS(5720),
    [anon_sym_CARET] = ACTIONS(5723),
    [anon_sym_AMP] = ACTIONS(5720),
    [anon_sym_SEMI] = ACTIONS(5917),
    [anon_sym___extension__] = ACTIONS(5729),
    [anon_sym_LBRACK] = ACTIONS(5732),
    [anon_sym_LBRACE] = ACTIONS(5920),
    [anon_sym_ATautoreleasepool] = ACTIONS(5923),
    [anon_sym_const] = ACTIONS(5741),
    [anon_sym___imag] = ACTIONS(5744),
    [anon_sym___real] = ACTIONS(5744),
    [anon_sym_struct] = ACTIONS(5741),
    [anon_sym_if] = ACTIONS(5926),
    [anon_sym_switch] = ACTIONS(5929),
    [anon_sym_case] = ACTIONS(5932),
    [anon_sym_default] = ACTIONS(5935),
    [anon_sym_while] = ACTIONS(5938),
    [anon_sym_do] = ACTIONS(5941),
    [anon_sym_for] = ACTIONS(5944),
    [anon_sym_in] = ACTIONS(5741),
    [anon_sym_return] = ACTIONS(5947),
    [anon_sym_break] = ACTIONS(5950),
    [anon_sym_continue] = ACTIONS(5953),
    [anon_sym_goto] = ACTIONS(5956),
    [anon_sym_DASH_DASH] = ACTIONS(5780),
    [anon_sym_PLUS_PLUS] = ACTIONS(5780),
    [anon_sym_sizeof] = ACTIONS(5783),
    [anon_sym___alignof__] = ACTIONS(5786),
    [anon_sym___alignof] = ACTIONS(5786),
    [anon_sym__alignof] = ACTIONS(5786),
    [anon_sym_alignof] = ACTIONS(5786),
    [anon_sym__Alignof] = ACTIONS(5786),
    [anon_sym_offsetof] = ACTIONS(5789),
    [anon_sym__Generic] = ACTIONS(5792),
    [anon_sym_asm] = ACTIONS(5795),
    [anon_sym___asm__] = ACTIONS(5795),
    [anon_sym___asm] = ACTIONS(5959),
    [sym_number_literal] = ACTIONS(5801),
    [anon_sym_L_SQUOTE] = ACTIONS(5804),
    [anon_sym_u_SQUOTE] = ACTIONS(5804),
    [anon_sym_U_SQUOTE] = ACTIONS(5804),
    [anon_sym_u8_SQUOTE] = ACTIONS(5804),
    [anon_sym_SQUOTE] = ACTIONS(5804),
    [anon_sym_AT] = ACTIONS(5807),
    [anon_sym_DQUOTE] = ACTIONS(5810),
    [anon_sym_L_DQUOTE] = ACTIONS(5810),
    [anon_sym_u_DQUOTE] = ACTIONS(5810),
    [anon_sym_U_DQUOTE] = ACTIONS(5810),
    [anon_sym_u8_DQUOTE] = ACTIONS(5810),
    [sym_true] = ACTIONS(5813),
    [sym_false] = ACTIONS(5813),
    [anon_sym_NULL] = ACTIONS(5816),
    [anon_sym_nullptr] = ACTIONS(5816),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(5962),
    [anon_sym___try] = ACTIONS(5965),
    [anon_sym_ATthrow] = ACTIONS(5968),
    [anon_sym_ATselector] = ACTIONS(5828),
    [anon_sym_ATavailable] = ACTIONS(5831),
    [anon_sym___builtin_available] = ACTIONS(5834),
    [anon_sym_va_arg] = ACTIONS(5837),
    [anon_sym_ATencode] = ACTIONS(5840),
    [anon_sym_ATsynchronized] = ACTIONS(5971),
    [anon_sym_id] = ACTIONS(5741),
  },
  [STATE(2445)] = {
    [sym_attribute_declaration] = STATE(2436),
    [sym_compound_statement] = STATE(439),
    [sym_attributed_statement] = STATE(439),
    [sym_statement] = STATE(435),
    [sym_labeled_statement] = STATE(439),
    [sym_expression_statement] = STATE(439),
    [sym_if_statement] = STATE(439),
    [sym_switch_statement] = STATE(439),
    [sym_case_statement] = STATE(439),
    [sym_while_statement] = STATE(439),
    [sym_do_statement] = STATE(439),
    [sym_for_statement] = STATE(439),
    [sym_return_statement] = STATE(439),
    [sym_break_statement] = STATE(439),
    [sym_continue_statement] = STATE(439),
    [sym_goto_statement] = STATE(439),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(439),
    [sym_throw_statement] = STATE(439),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(439),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(439),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [sym_identifier] = ACTIONS(5706),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_case] = ACTIONS(311),
    [anon_sym_default] = ACTIONS(313),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2446)] = {
    [sym_attribute_declaration] = STATE(2475),
    [sym_compound_statement] = STATE(1159),
    [sym_attributed_statement] = STATE(1159),
    [sym_statement] = STATE(1092),
    [sym_labeled_statement] = STATE(1159),
    [sym_expression_statement] = STATE(1159),
    [sym_if_statement] = STATE(1159),
    [sym_switch_statement] = STATE(1159),
    [sym_case_statement] = STATE(1159),
    [sym_while_statement] = STATE(1159),
    [sym_do_statement] = STATE(1159),
    [sym_for_statement] = STATE(1159),
    [sym_return_statement] = STATE(1159),
    [sym_break_statement] = STATE(1159),
    [sym_continue_statement] = STATE(1159),
    [sym_goto_statement] = STATE(1159),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(1159),
    [sym_throw_statement] = STATE(1159),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(1159),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(1159),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2475),
    [sym_identifier] = ACTIONS(5850),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2447)] = {
    [sym_attribute_declaration] = STATE(2447),
    [sym_compound_statement] = STATE(535),
    [sym_attributed_statement] = STATE(535),
    [sym_statement] = STATE(599),
    [sym_labeled_statement] = STATE(535),
    [sym_expression_statement] = STATE(535),
    [sym_if_statement] = STATE(535),
    [sym_switch_statement] = STATE(535),
    [sym_case_statement] = STATE(535),
    [sym_while_statement] = STATE(535),
    [sym_do_statement] = STATE(535),
    [sym_for_statement] = STATE(535),
    [sym_return_statement] = STATE(535),
    [sym_break_statement] = STATE(535),
    [sym_continue_statement] = STATE(535),
    [sym_goto_statement] = STATE(535),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(535),
    [sym_throw_statement] = STATE(535),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(535),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(535),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2447),
    [sym_identifier] = ACTIONS(5974),
    [anon_sym_LPAREN2] = ACTIONS(5711),
    [anon_sym_BANG] = ACTIONS(5714),
    [anon_sym_TILDE] = ACTIONS(5714),
    [anon_sym_DASH] = ACTIONS(5717),
    [anon_sym_PLUS] = ACTIONS(5717),
    [anon_sym_STAR] = ACTIONS(5720),
    [anon_sym_CARET] = ACTIONS(5723),
    [anon_sym_AMP] = ACTIONS(5720),
    [anon_sym_SEMI] = ACTIONS(5977),
    [anon_sym___extension__] = ACTIONS(5729),
    [anon_sym_LBRACK] = ACTIONS(5732),
    [anon_sym_LBRACE] = ACTIONS(5980),
    [anon_sym_ATautoreleasepool] = ACTIONS(5983),
    [anon_sym_const] = ACTIONS(5741),
    [anon_sym___imag] = ACTIONS(5744),
    [anon_sym___real] = ACTIONS(5744),
    [anon_sym_struct] = ACTIONS(5741),
    [anon_sym_if] = ACTIONS(5986),
    [anon_sym_switch] = ACTIONS(5989),
    [anon_sym_case] = ACTIONS(5992),
    [anon_sym_default] = ACTIONS(5995),
    [anon_sym_while] = ACTIONS(5998),
    [anon_sym_do] = ACTIONS(6001),
    [anon_sym_for] = ACTIONS(6004),
    [anon_sym_in] = ACTIONS(5741),
    [anon_sym_return] = ACTIONS(6007),
    [anon_sym_break] = ACTIONS(6010),
    [anon_sym_continue] = ACTIONS(6013),
    [anon_sym_goto] = ACTIONS(6016),
    [anon_sym_DASH_DASH] = ACTIONS(5780),
    [anon_sym_PLUS_PLUS] = ACTIONS(5780),
    [anon_sym_sizeof] = ACTIONS(5783),
    [anon_sym___alignof__] = ACTIONS(5786),
    [anon_sym___alignof] = ACTIONS(5786),
    [anon_sym__alignof] = ACTIONS(5786),
    [anon_sym_alignof] = ACTIONS(5786),
    [anon_sym__Alignof] = ACTIONS(5786),
    [anon_sym_offsetof] = ACTIONS(5789),
    [anon_sym__Generic] = ACTIONS(5792),
    [anon_sym_asm] = ACTIONS(5795),
    [anon_sym___asm__] = ACTIONS(5795),
    [anon_sym___asm] = ACTIONS(6019),
    [sym_number_literal] = ACTIONS(5801),
    [anon_sym_L_SQUOTE] = ACTIONS(5804),
    [anon_sym_u_SQUOTE] = ACTIONS(5804),
    [anon_sym_U_SQUOTE] = ACTIONS(5804),
    [anon_sym_u8_SQUOTE] = ACTIONS(5804),
    [anon_sym_SQUOTE] = ACTIONS(5804),
    [anon_sym_AT] = ACTIONS(5807),
    [anon_sym_DQUOTE] = ACTIONS(5810),
    [anon_sym_L_DQUOTE] = ACTIONS(5810),
    [anon_sym_u_DQUOTE] = ACTIONS(5810),
    [anon_sym_U_DQUOTE] = ACTIONS(5810),
    [anon_sym_u8_DQUOTE] = ACTIONS(5810),
    [sym_true] = ACTIONS(5813),
    [sym_false] = ACTIONS(5813),
    [anon_sym_NULL] = ACTIONS(5816),
    [anon_sym_nullptr] = ACTIONS(5816),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(6022),
    [anon_sym___try] = ACTIONS(6025),
    [anon_sym_ATthrow] = ACTIONS(6028),
    [anon_sym_ATselector] = ACTIONS(5828),
    [anon_sym_ATavailable] = ACTIONS(5831),
    [anon_sym___builtin_available] = ACTIONS(5834),
    [anon_sym_va_arg] = ACTIONS(5837),
    [anon_sym_ATencode] = ACTIONS(5840),
    [anon_sym_ATsynchronized] = ACTIONS(6031),
    [anon_sym_id] = ACTIONS(5741),
  },
  [STATE(2448)] = {
    [sym_attribute_declaration] = STATE(2451),
    [sym_compound_statement] = STATE(955),
    [sym_attributed_statement] = STATE(955),
    [sym_statement] = STATE(948),
    [sym_labeled_statement] = STATE(955),
    [sym_expression_statement] = STATE(955),
    [sym_if_statement] = STATE(955),
    [sym_switch_statement] = STATE(955),
    [sym_case_statement] = STATE(955),
    [sym_while_statement] = STATE(955),
    [sym_do_statement] = STATE(955),
    [sym_for_statement] = STATE(955),
    [sym_return_statement] = STATE(955),
    [sym_break_statement] = STATE(955),
    [sym_continue_statement] = STATE(955),
    [sym_goto_statement] = STATE(955),
    [sym_expression] = STATE(4419),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6212),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(955),
    [sym_throw_statement] = STATE(955),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(955),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(955),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2451),
    [sym_identifier] = ACTIONS(6034),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2150),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(69),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(77),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2152),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2154),
    [anon_sym___try] = ACTIONS(2156),
    [anon_sym_ATthrow] = ACTIONS(2158),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2160),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2449)] = {
    [sym_attribute_declaration] = STATE(2456),
    [sym_compound_statement] = STATE(232),
    [sym_attributed_statement] = STATE(232),
    [sym_statement] = STATE(197),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [sym_identifier] = ACTIONS(6036),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2450)] = {
    [sym_attribute_declaration] = STATE(2480),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(728),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [sym_identifier] = ACTIONS(5848),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2451)] = {
    [sym_attribute_declaration] = STATE(2460),
    [sym_compound_statement] = STATE(955),
    [sym_attributed_statement] = STATE(955),
    [sym_statement] = STATE(942),
    [sym_labeled_statement] = STATE(955),
    [sym_expression_statement] = STATE(955),
    [sym_if_statement] = STATE(955),
    [sym_switch_statement] = STATE(955),
    [sym_case_statement] = STATE(955),
    [sym_while_statement] = STATE(955),
    [sym_do_statement] = STATE(955),
    [sym_for_statement] = STATE(955),
    [sym_return_statement] = STATE(955),
    [sym_break_statement] = STATE(955),
    [sym_continue_statement] = STATE(955),
    [sym_goto_statement] = STATE(955),
    [sym_expression] = STATE(4419),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6212),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(955),
    [sym_throw_statement] = STATE(955),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(955),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(955),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2460),
    [sym_identifier] = ACTIONS(6034),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2150),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(69),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(77),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2152),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2154),
    [anon_sym___try] = ACTIONS(2156),
    [anon_sym_ATthrow] = ACTIONS(2158),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2160),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2452)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(6205),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2453)] = {
    [sym_attribute_declaration] = STATE(2438),
    [sym_compound_statement] = STATE(535),
    [sym_attributed_statement] = STATE(535),
    [sym_statement] = STATE(576),
    [sym_labeled_statement] = STATE(535),
    [sym_expression_statement] = STATE(535),
    [sym_if_statement] = STATE(535),
    [sym_switch_statement] = STATE(535),
    [sym_case_statement] = STATE(535),
    [sym_while_statement] = STATE(535),
    [sym_do_statement] = STATE(535),
    [sym_for_statement] = STATE(535),
    [sym_return_statement] = STATE(535),
    [sym_break_statement] = STATE(535),
    [sym_continue_statement] = STATE(535),
    [sym_goto_statement] = STATE(535),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(535),
    [sym_throw_statement] = STATE(535),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(535),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(535),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [sym_identifier] = ACTIONS(5846),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(878),
    [anon_sym_default] = ACTIONS(880),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2454)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(923),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2455)] = {
    [sym_attribute_declaration] = STATE(2456),
    [sym_compound_statement] = STATE(232),
    [sym_attributed_statement] = STATE(232),
    [sym_statement] = STATE(246),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [sym_identifier] = ACTIONS(6036),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2456)] = {
    [sym_attribute_declaration] = STATE(2461),
    [sym_compound_statement] = STATE(232),
    [sym_attributed_statement] = STATE(232),
    [sym_statement] = STATE(238),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2461),
    [sym_identifier] = ACTIONS(6036),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2457)] = {
    [sym_attribute_declaration] = STATE(2456),
    [sym_compound_statement] = STATE(232),
    [sym_attributed_statement] = STATE(232),
    [sym_statement] = STATE(264),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [sym_identifier] = ACTIONS(6036),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2458)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(6392),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2459)] = {
    [sym_attribute_declaration] = STATE(2480),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(731),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [sym_identifier] = ACTIONS(5848),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2460)] = {
    [sym_attribute_declaration] = STATE(2460),
    [sym_compound_statement] = STATE(955),
    [sym_attributed_statement] = STATE(955),
    [sym_statement] = STATE(942),
    [sym_labeled_statement] = STATE(955),
    [sym_expression_statement] = STATE(955),
    [sym_if_statement] = STATE(955),
    [sym_switch_statement] = STATE(955),
    [sym_case_statement] = STATE(955),
    [sym_while_statement] = STATE(955),
    [sym_do_statement] = STATE(955),
    [sym_for_statement] = STATE(955),
    [sym_return_statement] = STATE(955),
    [sym_break_statement] = STATE(955),
    [sym_continue_statement] = STATE(955),
    [sym_goto_statement] = STATE(955),
    [sym_expression] = STATE(4419),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6212),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(955),
    [sym_throw_statement] = STATE(955),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(955),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(955),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2460),
    [sym_identifier] = ACTIONS(6038),
    [anon_sym_LPAREN2] = ACTIONS(5711),
    [anon_sym_BANG] = ACTIONS(5714),
    [anon_sym_TILDE] = ACTIONS(5714),
    [anon_sym_DASH] = ACTIONS(5717),
    [anon_sym_PLUS] = ACTIONS(5717),
    [anon_sym_STAR] = ACTIONS(5720),
    [anon_sym_CARET] = ACTIONS(5723),
    [anon_sym_AMP] = ACTIONS(5720),
    [anon_sym_SEMI] = ACTIONS(6041),
    [anon_sym___extension__] = ACTIONS(5729),
    [anon_sym_LBRACK] = ACTIONS(5732),
    [anon_sym_LBRACE] = ACTIONS(5860),
    [anon_sym_ATautoreleasepool] = ACTIONS(5863),
    [anon_sym_const] = ACTIONS(5741),
    [anon_sym___imag] = ACTIONS(5744),
    [anon_sym___real] = ACTIONS(5744),
    [anon_sym_struct] = ACTIONS(5741),
    [anon_sym_if] = ACTIONS(6044),
    [anon_sym_switch] = ACTIONS(5869),
    [anon_sym_case] = ACTIONS(5932),
    [anon_sym_default] = ACTIONS(5935),
    [anon_sym_while] = ACTIONS(6047),
    [anon_sym_do] = ACTIONS(5881),
    [anon_sym_for] = ACTIONS(6050),
    [anon_sym_in] = ACTIONS(5741),
    [anon_sym_return] = ACTIONS(5887),
    [anon_sym_break] = ACTIONS(5890),
    [anon_sym_continue] = ACTIONS(5893),
    [anon_sym_goto] = ACTIONS(5896),
    [anon_sym_DASH_DASH] = ACTIONS(5780),
    [anon_sym_PLUS_PLUS] = ACTIONS(5780),
    [anon_sym_sizeof] = ACTIONS(5783),
    [anon_sym___alignof__] = ACTIONS(5786),
    [anon_sym___alignof] = ACTIONS(5786),
    [anon_sym__alignof] = ACTIONS(5786),
    [anon_sym_alignof] = ACTIONS(5786),
    [anon_sym__Alignof] = ACTIONS(5786),
    [anon_sym_offsetof] = ACTIONS(5789),
    [anon_sym__Generic] = ACTIONS(5792),
    [anon_sym_asm] = ACTIONS(5795),
    [anon_sym___asm__] = ACTIONS(5795),
    [anon_sym___asm] = ACTIONS(6053),
    [sym_number_literal] = ACTIONS(5801),
    [anon_sym_L_SQUOTE] = ACTIONS(5804),
    [anon_sym_u_SQUOTE] = ACTIONS(5804),
    [anon_sym_U_SQUOTE] = ACTIONS(5804),
    [anon_sym_u8_SQUOTE] = ACTIONS(5804),
    [anon_sym_SQUOTE] = ACTIONS(5804),
    [anon_sym_AT] = ACTIONS(5807),
    [anon_sym_DQUOTE] = ACTIONS(5810),
    [anon_sym_L_DQUOTE] = ACTIONS(5810),
    [anon_sym_u_DQUOTE] = ACTIONS(5810),
    [anon_sym_U_DQUOTE] = ACTIONS(5810),
    [anon_sym_u8_DQUOTE] = ACTIONS(5810),
    [sym_true] = ACTIONS(5813),
    [sym_false] = ACTIONS(5813),
    [anon_sym_NULL] = ACTIONS(5816),
    [anon_sym_nullptr] = ACTIONS(5816),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(6056),
    [anon_sym___try] = ACTIONS(6059),
    [anon_sym_ATthrow] = ACTIONS(6062),
    [anon_sym_ATselector] = ACTIONS(5828),
    [anon_sym_ATavailable] = ACTIONS(5831),
    [anon_sym___builtin_available] = ACTIONS(5834),
    [anon_sym_va_arg] = ACTIONS(5837),
    [anon_sym_ATencode] = ACTIONS(5840),
    [anon_sym_ATsynchronized] = ACTIONS(6065),
    [anon_sym_id] = ACTIONS(5741),
  },
  [STATE(2461)] = {
    [sym_attribute_declaration] = STATE(2461),
    [sym_compound_statement] = STATE(232),
    [sym_attributed_statement] = STATE(232),
    [sym_statement] = STATE(238),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2461),
    [sym_identifier] = ACTIONS(6068),
    [anon_sym_LPAREN2] = ACTIONS(5711),
    [anon_sym_BANG] = ACTIONS(5714),
    [anon_sym_TILDE] = ACTIONS(5714),
    [anon_sym_DASH] = ACTIONS(5717),
    [anon_sym_PLUS] = ACTIONS(5717),
    [anon_sym_STAR] = ACTIONS(5720),
    [anon_sym_CARET] = ACTIONS(5723),
    [anon_sym_AMP] = ACTIONS(5720),
    [anon_sym_SEMI] = ACTIONS(6071),
    [anon_sym___extension__] = ACTIONS(5729),
    [anon_sym_LBRACK] = ACTIONS(5732),
    [anon_sym_LBRACE] = ACTIONS(6074),
    [anon_sym_ATautoreleasepool] = ACTIONS(6077),
    [anon_sym_const] = ACTIONS(5741),
    [anon_sym___imag] = ACTIONS(5744),
    [anon_sym___real] = ACTIONS(5744),
    [anon_sym_struct] = ACTIONS(5741),
    [anon_sym_if] = ACTIONS(6080),
    [anon_sym_switch] = ACTIONS(6083),
    [anon_sym_case] = ACTIONS(6086),
    [anon_sym_default] = ACTIONS(6089),
    [anon_sym_while] = ACTIONS(6092),
    [anon_sym_do] = ACTIONS(6095),
    [anon_sym_for] = ACTIONS(6098),
    [anon_sym_in] = ACTIONS(5741),
    [anon_sym_return] = ACTIONS(6101),
    [anon_sym_break] = ACTIONS(6104),
    [anon_sym_continue] = ACTIONS(6107),
    [anon_sym_goto] = ACTIONS(6110),
    [anon_sym_DASH_DASH] = ACTIONS(5780),
    [anon_sym_PLUS_PLUS] = ACTIONS(5780),
    [anon_sym_sizeof] = ACTIONS(5783),
    [anon_sym___alignof__] = ACTIONS(5786),
    [anon_sym___alignof] = ACTIONS(5786),
    [anon_sym__alignof] = ACTIONS(5786),
    [anon_sym_alignof] = ACTIONS(5786),
    [anon_sym__Alignof] = ACTIONS(5786),
    [anon_sym_offsetof] = ACTIONS(5789),
    [anon_sym__Generic] = ACTIONS(5792),
    [anon_sym_asm] = ACTIONS(5795),
    [anon_sym___asm__] = ACTIONS(5795),
    [anon_sym___asm] = ACTIONS(6113),
    [sym_number_literal] = ACTIONS(5801),
    [anon_sym_L_SQUOTE] = ACTIONS(5804),
    [anon_sym_u_SQUOTE] = ACTIONS(5804),
    [anon_sym_U_SQUOTE] = ACTIONS(5804),
    [anon_sym_u8_SQUOTE] = ACTIONS(5804),
    [anon_sym_SQUOTE] = ACTIONS(5804),
    [anon_sym_AT] = ACTIONS(5807),
    [anon_sym_DQUOTE] = ACTIONS(5810),
    [anon_sym_L_DQUOTE] = ACTIONS(5810),
    [anon_sym_u_DQUOTE] = ACTIONS(5810),
    [anon_sym_U_DQUOTE] = ACTIONS(5810),
    [anon_sym_u8_DQUOTE] = ACTIONS(5810),
    [sym_true] = ACTIONS(5813),
    [sym_false] = ACTIONS(5813),
    [anon_sym_NULL] = ACTIONS(5816),
    [anon_sym_nullptr] = ACTIONS(5816),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(6116),
    [anon_sym___try] = ACTIONS(6119),
    [anon_sym_ATthrow] = ACTIONS(6122),
    [anon_sym_ATselector] = ACTIONS(5828),
    [anon_sym_ATavailable] = ACTIONS(5831),
    [anon_sym___builtin_available] = ACTIONS(5834),
    [anon_sym_va_arg] = ACTIONS(5837),
    [anon_sym_ATencode] = ACTIONS(5840),
    [anon_sym_ATsynchronized] = ACTIONS(6125),
    [anon_sym_id] = ACTIONS(5741),
  },
  [STATE(2462)] = {
    [sym_attribute_declaration] = STATE(2480),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(828),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [sym_identifier] = ACTIONS(5848),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2463)] = {
    [sym_attribute_declaration] = STATE(2451),
    [sym_compound_statement] = STATE(955),
    [sym_attributed_statement] = STATE(955),
    [sym_statement] = STATE(917),
    [sym_labeled_statement] = STATE(955),
    [sym_expression_statement] = STATE(955),
    [sym_if_statement] = STATE(955),
    [sym_switch_statement] = STATE(955),
    [sym_case_statement] = STATE(955),
    [sym_while_statement] = STATE(955),
    [sym_do_statement] = STATE(955),
    [sym_for_statement] = STATE(955),
    [sym_return_statement] = STATE(955),
    [sym_break_statement] = STATE(955),
    [sym_continue_statement] = STATE(955),
    [sym_goto_statement] = STATE(955),
    [sym_expression] = STATE(4419),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6212),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(955),
    [sym_throw_statement] = STATE(955),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(955),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(955),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2451),
    [sym_identifier] = ACTIONS(6034),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2150),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(69),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(77),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2152),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2154),
    [anon_sym___try] = ACTIONS(2156),
    [anon_sym_ATthrow] = ACTIONS(2158),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2160),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2464)] = {
    [sym_attribute_declaration] = STATE(2472),
    [sym_compound_statement] = STATE(690),
    [sym_attributed_statement] = STATE(690),
    [sym_statement] = STATE(759),
    [sym_labeled_statement] = STATE(690),
    [sym_expression_statement] = STATE(690),
    [sym_if_statement] = STATE(690),
    [sym_switch_statement] = STATE(690),
    [sym_case_statement] = STATE(690),
    [sym_while_statement] = STATE(690),
    [sym_do_statement] = STATE(690),
    [sym_for_statement] = STATE(690),
    [sym_return_statement] = STATE(690),
    [sym_break_statement] = STATE(690),
    [sym_continue_statement] = STATE(690),
    [sym_goto_statement] = STATE(690),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(690),
    [sym_throw_statement] = STATE(690),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(690),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(690),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [sym_identifier] = ACTIONS(5852),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2465)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(940),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2466)] = {
    [sym_attribute_declaration] = STATE(2472),
    [sym_compound_statement] = STATE(690),
    [sym_attributed_statement] = STATE(690),
    [sym_statement] = STATE(761),
    [sym_labeled_statement] = STATE(690),
    [sym_expression_statement] = STATE(690),
    [sym_if_statement] = STATE(690),
    [sym_switch_statement] = STATE(690),
    [sym_case_statement] = STATE(690),
    [sym_while_statement] = STATE(690),
    [sym_do_statement] = STATE(690),
    [sym_for_statement] = STATE(690),
    [sym_return_statement] = STATE(690),
    [sym_break_statement] = STATE(690),
    [sym_continue_statement] = STATE(690),
    [sym_goto_statement] = STATE(690),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(690),
    [sym_throw_statement] = STATE(690),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(690),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(690),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [sym_identifier] = ACTIONS(5852),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2467)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(6080),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2468)] = {
    [sym_attribute_declaration] = STATE(2451),
    [sym_compound_statement] = STATE(955),
    [sym_attributed_statement] = STATE(955),
    [sym_statement] = STATE(923),
    [sym_labeled_statement] = STATE(955),
    [sym_expression_statement] = STATE(955),
    [sym_if_statement] = STATE(955),
    [sym_switch_statement] = STATE(955),
    [sym_case_statement] = STATE(955),
    [sym_while_statement] = STATE(955),
    [sym_do_statement] = STATE(955),
    [sym_for_statement] = STATE(955),
    [sym_return_statement] = STATE(955),
    [sym_break_statement] = STATE(955),
    [sym_continue_statement] = STATE(955),
    [sym_goto_statement] = STATE(955),
    [sym_expression] = STATE(4419),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6212),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(955),
    [sym_throw_statement] = STATE(955),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(955),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(955),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2451),
    [sym_identifier] = ACTIONS(6034),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2150),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(69),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(77),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2152),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2154),
    [anon_sym___try] = ACTIONS(2156),
    [anon_sym_ATthrow] = ACTIONS(2158),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2160),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2469)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(6459),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2470)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3812),
    [aux_sym_generic_specifier_repeat1] = STATE(3813),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_COMMA] = ACTIONS(6128),
    [anon_sym_LPAREN2] = ACTIONS(6130),
    [anon_sym_STAR] = ACTIONS(6134),
    [anon_sym_CARET] = ACTIONS(6134),
    [anon_sym_LT] = ACTIONS(6136),
    [anon_sym_SEMI] = ACTIONS(6128),
    [anon_sym___extension__] = ACTIONS(6138),
    [anon_sym_extern] = ACTIONS(6141),
    [anon_sym___attribute__] = ACTIONS(6141),
    [anon_sym___attribute] = ACTIONS(6141),
    [anon_sym_noreturn] = ACTIONS(6138),
    [anon_sym_LBRACK] = ACTIONS(6143),
    [anon_sym___declspec] = ACTIONS(6141),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(6146),
    [anon_sym_unsigned] = ACTIONS(6146),
    [anon_sym_long] = ACTIONS(6146),
    [anon_sym_short] = ACTIONS(6146),
    [anon_sym_static] = ACTIONS(6141),
    [anon_sym_auto] = ACTIONS(6141),
    [anon_sym_register] = ACTIONS(6141),
    [anon_sym_inline] = ACTIONS(6141),
    [anon_sym___inline] = ACTIONS(6141),
    [anon_sym___inline__] = ACTIONS(6141),
    [anon_sym___forceinline] = ACTIONS(6141),
    [anon_sym_thread_local] = ACTIONS(6141),
    [anon_sym___thread] = ACTIONS(6141),
    [anon_sym_CG_EXTERN] = ACTIONS(6141),
    [anon_sym_CG_INLINE] = ACTIONS(6141),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6141),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6141),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6141),
    [anon_sym_IBOutlet] = ACTIONS(6141),
    [anon_sym_IBInspectable] = ACTIONS(6141),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6141),
    [anon_sym_NS_INLINE] = ACTIONS(6141),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6141),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6141),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6141),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6141),
    [anon_sym_const] = ACTIONS(6138),
    [anon_sym_constexpr] = ACTIONS(6138),
    [anon_sym_volatile] = ACTIONS(6138),
    [anon_sym_restrict] = ACTIONS(6138),
    [anon_sym___restrict__] = ACTIONS(6138),
    [anon_sym__Atomic] = ACTIONS(6138),
    [anon_sym__Noreturn] = ACTIONS(6138),
    [anon_sym__Nonnull] = ACTIONS(6138),
    [anon_sym_nullable] = ACTIONS(6138),
    [anon_sym__Complex] = ACTIONS(6138),
    [anon_sym__Nullable] = ACTIONS(6138),
    [anon_sym__Nullable_result] = ACTIONS(6138),
    [anon_sym__Null_unspecified] = ACTIONS(6138),
    [anon_sym___autoreleasing] = ACTIONS(6138),
    [anon_sym___block] = ACTIONS(6138),
    [anon_sym___bridge] = ACTIONS(6138),
    [anon_sym___bridge_retained] = ACTIONS(6138),
    [anon_sym___bridge_transfer] = ACTIONS(6138),
    [anon_sym___complex] = ACTIONS(6138),
    [anon_sym___const] = ACTIONS(6138),
    [anon_sym___imag] = ACTIONS(6138),
    [anon_sym___kindof] = ACTIONS(6138),
    [anon_sym___nonnull] = ACTIONS(6138),
    [anon_sym___nullable] = ACTIONS(6138),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6138),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6138),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6138),
    [anon_sym___real] = ACTIONS(6138),
    [anon_sym___strong] = ACTIONS(6138),
    [anon_sym___unsafe_unretained] = ACTIONS(6138),
    [anon_sym___unused] = ACTIONS(6138),
    [anon_sym___weak] = ACTIONS(6138),
    [anon_sym_alignas] = ACTIONS(6138),
    [anon_sym__Alignas] = ACTIONS(6138),
    [sym_primitive_type] = ACTIONS(4494),
    [anon_sym_enum] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(6143),
    [anon_sym_struct] = ACTIONS(4494),
    [anon_sym_union] = ACTIONS(4494),
    [anon_sym_in] = ACTIONS(4494),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4494),
    [anon_sym___typeof] = ACTIONS(4494),
    [anon_sym_typeof] = ACTIONS(4494),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6141),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6141),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6141),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6141),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6141),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6141),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6141),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6141),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6141),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6141),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6141),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6141),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6141),
    [anon_sym_API_AVAILABLE] = ACTIONS(6141),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6141),
    [anon_sym_API_DEPRECATED] = ACTIONS(6141),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6141),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6141),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6141),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6141),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6141),
    [anon_sym___deprecated_msg] = ACTIONS(6141),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6141),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6141),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6141),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6141),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6141),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6141),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6141),
    [anon_sym_BOOL] = ACTIONS(4494),
    [anon_sym_IMP] = ACTIONS(4494),
    [anon_sym_SEL] = ACTIONS(4494),
    [anon_sym_Class] = ACTIONS(4494),
    [anon_sym_id] = ACTIONS(4494),
    [anon_sym_out] = ACTIONS(4494),
    [anon_sym_inout] = ACTIONS(4494),
    [anon_sym_bycopy] = ACTIONS(4494),
    [anon_sym_byref] = ACTIONS(4494),
    [anon_sym_oneway] = ACTIONS(4494),
  },
  [STATE(2471)] = {
    [sym_attribute_declaration] = STATE(2438),
    [sym_compound_statement] = STATE(535),
    [sym_attributed_statement] = STATE(535),
    [sym_statement] = STATE(484),
    [sym_labeled_statement] = STATE(535),
    [sym_expression_statement] = STATE(535),
    [sym_if_statement] = STATE(535),
    [sym_switch_statement] = STATE(535),
    [sym_case_statement] = STATE(535),
    [sym_while_statement] = STATE(535),
    [sym_do_statement] = STATE(535),
    [sym_for_statement] = STATE(535),
    [sym_return_statement] = STATE(535),
    [sym_break_statement] = STATE(535),
    [sym_continue_statement] = STATE(535),
    [sym_goto_statement] = STATE(535),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(535),
    [sym_throw_statement] = STATE(535),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(535),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(535),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [sym_identifier] = ACTIONS(5846),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(878),
    [anon_sym_default] = ACTIONS(880),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2472)] = {
    [sym_attribute_declaration] = STATE(2444),
    [sym_compound_statement] = STATE(690),
    [sym_attributed_statement] = STATE(690),
    [sym_statement] = STATE(716),
    [sym_labeled_statement] = STATE(690),
    [sym_expression_statement] = STATE(690),
    [sym_if_statement] = STATE(690),
    [sym_switch_statement] = STATE(690),
    [sym_case_statement] = STATE(690),
    [sym_while_statement] = STATE(690),
    [sym_do_statement] = STATE(690),
    [sym_for_statement] = STATE(690),
    [sym_return_statement] = STATE(690),
    [sym_break_statement] = STATE(690),
    [sym_continue_statement] = STATE(690),
    [sym_goto_statement] = STATE(690),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(690),
    [sym_throw_statement] = STATE(690),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(690),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(690),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2444),
    [sym_identifier] = ACTIONS(5852),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2473)] = {
    [sym_attribute_declaration] = STATE(2456),
    [sym_compound_statement] = STATE(232),
    [sym_attributed_statement] = STATE(232),
    [sym_statement] = STATE(266),
    [sym_labeled_statement] = STATE(232),
    [sym_expression_statement] = STATE(232),
    [sym_if_statement] = STATE(232),
    [sym_switch_statement] = STATE(232),
    [sym_case_statement] = STATE(232),
    [sym_while_statement] = STATE(232),
    [sym_do_statement] = STATE(232),
    [sym_for_statement] = STATE(232),
    [sym_return_statement] = STATE(232),
    [sym_break_statement] = STATE(232),
    [sym_continue_statement] = STATE(232),
    [sym_goto_statement] = STATE(232),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(232),
    [sym_throw_statement] = STATE(232),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(232),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(232),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [sym_identifier] = ACTIONS(6036),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_case] = ACTIONS(193),
    [anon_sym_default] = ACTIONS(195),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2474)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(6508),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2475)] = {
    [sym_attribute_declaration] = STATE(2475),
    [sym_compound_statement] = STATE(1159),
    [sym_attributed_statement] = STATE(1159),
    [sym_statement] = STATE(1092),
    [sym_labeled_statement] = STATE(1159),
    [sym_expression_statement] = STATE(1159),
    [sym_if_statement] = STATE(1159),
    [sym_switch_statement] = STATE(1159),
    [sym_case_statement] = STATE(1159),
    [sym_while_statement] = STATE(1159),
    [sym_do_statement] = STATE(1159),
    [sym_for_statement] = STATE(1159),
    [sym_return_statement] = STATE(1159),
    [sym_break_statement] = STATE(1159),
    [sym_continue_statement] = STATE(1159),
    [sym_goto_statement] = STATE(1159),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(1159),
    [sym_throw_statement] = STATE(1159),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(1159),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(1159),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2475),
    [sym_identifier] = ACTIONS(6149),
    [anon_sym_LPAREN2] = ACTIONS(5711),
    [anon_sym_BANG] = ACTIONS(5714),
    [anon_sym_TILDE] = ACTIONS(5714),
    [anon_sym_DASH] = ACTIONS(5717),
    [anon_sym_PLUS] = ACTIONS(5717),
    [anon_sym_STAR] = ACTIONS(5720),
    [anon_sym_CARET] = ACTIONS(5723),
    [anon_sym_AMP] = ACTIONS(5720),
    [anon_sym_SEMI] = ACTIONS(6152),
    [anon_sym___extension__] = ACTIONS(5729),
    [anon_sym_LBRACK] = ACTIONS(5732),
    [anon_sym_LBRACE] = ACTIONS(6155),
    [anon_sym_ATautoreleasepool] = ACTIONS(6158),
    [anon_sym_const] = ACTIONS(5741),
    [anon_sym___imag] = ACTIONS(5744),
    [anon_sym___real] = ACTIONS(5744),
    [anon_sym_struct] = ACTIONS(5741),
    [anon_sym_if] = ACTIONS(6161),
    [anon_sym_switch] = ACTIONS(6164),
    [anon_sym_case] = ACTIONS(5872),
    [anon_sym_default] = ACTIONS(5875),
    [anon_sym_while] = ACTIONS(6167),
    [anon_sym_do] = ACTIONS(6170),
    [anon_sym_for] = ACTIONS(6173),
    [anon_sym_in] = ACTIONS(5741),
    [anon_sym_return] = ACTIONS(6176),
    [anon_sym_break] = ACTIONS(6179),
    [anon_sym_continue] = ACTIONS(6182),
    [anon_sym_goto] = ACTIONS(6185),
    [anon_sym_DASH_DASH] = ACTIONS(5780),
    [anon_sym_PLUS_PLUS] = ACTIONS(5780),
    [anon_sym_sizeof] = ACTIONS(5783),
    [anon_sym___alignof__] = ACTIONS(5786),
    [anon_sym___alignof] = ACTIONS(5786),
    [anon_sym__alignof] = ACTIONS(5786),
    [anon_sym_alignof] = ACTIONS(5786),
    [anon_sym__Alignof] = ACTIONS(5786),
    [anon_sym_offsetof] = ACTIONS(5789),
    [anon_sym__Generic] = ACTIONS(5792),
    [anon_sym_asm] = ACTIONS(5795),
    [anon_sym___asm__] = ACTIONS(5795),
    [anon_sym___asm] = ACTIONS(6188),
    [sym_number_literal] = ACTIONS(5801),
    [anon_sym_L_SQUOTE] = ACTIONS(5804),
    [anon_sym_u_SQUOTE] = ACTIONS(5804),
    [anon_sym_U_SQUOTE] = ACTIONS(5804),
    [anon_sym_u8_SQUOTE] = ACTIONS(5804),
    [anon_sym_SQUOTE] = ACTIONS(5804),
    [anon_sym_AT] = ACTIONS(5807),
    [anon_sym_DQUOTE] = ACTIONS(5810),
    [anon_sym_L_DQUOTE] = ACTIONS(5810),
    [anon_sym_u_DQUOTE] = ACTIONS(5810),
    [anon_sym_U_DQUOTE] = ACTIONS(5810),
    [anon_sym_u8_DQUOTE] = ACTIONS(5810),
    [sym_true] = ACTIONS(5813),
    [sym_false] = ACTIONS(5813),
    [anon_sym_NULL] = ACTIONS(5816),
    [anon_sym_nullptr] = ACTIONS(5816),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(6191),
    [anon_sym___try] = ACTIONS(6194),
    [anon_sym_ATthrow] = ACTIONS(6197),
    [anon_sym_ATselector] = ACTIONS(5828),
    [anon_sym_ATavailable] = ACTIONS(5831),
    [anon_sym___builtin_available] = ACTIONS(5834),
    [anon_sym_va_arg] = ACTIONS(5837),
    [anon_sym_ATencode] = ACTIONS(5840),
    [anon_sym_ATsynchronized] = ACTIONS(6200),
    [anon_sym_id] = ACTIONS(5741),
  },
  [STATE(2476)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(6528),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(2166),
    [anon_sym_default] = ACTIONS(2168),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2477)] = {
    [sym_attribute_declaration] = STATE(2477),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(712),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2477),
    [sym_identifier] = ACTIONS(6203),
    [anon_sym_LPAREN2] = ACTIONS(5711),
    [anon_sym_BANG] = ACTIONS(5714),
    [anon_sym_TILDE] = ACTIONS(5714),
    [anon_sym_DASH] = ACTIONS(5717),
    [anon_sym_PLUS] = ACTIONS(5717),
    [anon_sym_STAR] = ACTIONS(5720),
    [anon_sym_CARET] = ACTIONS(5723),
    [anon_sym_AMP] = ACTIONS(5720),
    [anon_sym_SEMI] = ACTIONS(5857),
    [anon_sym___extension__] = ACTIONS(5729),
    [anon_sym_LBRACK] = ACTIONS(5732),
    [anon_sym_LBRACE] = ACTIONS(6206),
    [anon_sym_ATautoreleasepool] = ACTIONS(6209),
    [anon_sym_const] = ACTIONS(5741),
    [anon_sym___imag] = ACTIONS(5744),
    [anon_sym___real] = ACTIONS(5744),
    [anon_sym_struct] = ACTIONS(5741),
    [anon_sym_if] = ACTIONS(6212),
    [anon_sym_switch] = ACTIONS(6215),
    [anon_sym_case] = ACTIONS(6218),
    [anon_sym_default] = ACTIONS(6221),
    [anon_sym_while] = ACTIONS(6224),
    [anon_sym_do] = ACTIONS(6227),
    [anon_sym_for] = ACTIONS(6230),
    [anon_sym_in] = ACTIONS(5741),
    [anon_sym_return] = ACTIONS(6233),
    [anon_sym_break] = ACTIONS(6236),
    [anon_sym_continue] = ACTIONS(6239),
    [anon_sym_goto] = ACTIONS(6242),
    [anon_sym_DASH_DASH] = ACTIONS(5780),
    [anon_sym_PLUS_PLUS] = ACTIONS(5780),
    [anon_sym_sizeof] = ACTIONS(5783),
    [anon_sym___alignof__] = ACTIONS(5786),
    [anon_sym___alignof] = ACTIONS(5786),
    [anon_sym__alignof] = ACTIONS(5786),
    [anon_sym_alignof] = ACTIONS(5786),
    [anon_sym__Alignof] = ACTIONS(5786),
    [anon_sym_offsetof] = ACTIONS(5789),
    [anon_sym__Generic] = ACTIONS(5792),
    [anon_sym_asm] = ACTIONS(5795),
    [anon_sym___asm__] = ACTIONS(5795),
    [anon_sym___asm] = ACTIONS(5899),
    [sym_number_literal] = ACTIONS(5801),
    [anon_sym_L_SQUOTE] = ACTIONS(5804),
    [anon_sym_u_SQUOTE] = ACTIONS(5804),
    [anon_sym_U_SQUOTE] = ACTIONS(5804),
    [anon_sym_u8_SQUOTE] = ACTIONS(5804),
    [anon_sym_SQUOTE] = ACTIONS(5804),
    [anon_sym_AT] = ACTIONS(5807),
    [anon_sym_DQUOTE] = ACTIONS(5810),
    [anon_sym_L_DQUOTE] = ACTIONS(5810),
    [anon_sym_u_DQUOTE] = ACTIONS(5810),
    [anon_sym_U_DQUOTE] = ACTIONS(5810),
    [anon_sym_u8_DQUOTE] = ACTIONS(5810),
    [sym_true] = ACTIONS(5813),
    [sym_false] = ACTIONS(5813),
    [anon_sym_NULL] = ACTIONS(5816),
    [anon_sym_nullptr] = ACTIONS(5816),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(6245),
    [anon_sym___try] = ACTIONS(6248),
    [anon_sym_ATthrow] = ACTIONS(5908),
    [anon_sym_ATselector] = ACTIONS(5828),
    [anon_sym_ATavailable] = ACTIONS(5831),
    [anon_sym___builtin_available] = ACTIONS(5834),
    [anon_sym_va_arg] = ACTIONS(5837),
    [anon_sym_ATencode] = ACTIONS(5840),
    [anon_sym_ATsynchronized] = ACTIONS(6251),
    [anon_sym_id] = ACTIONS(5741),
  },
  [STATE(2478)] = {
    [sym_attribute_declaration] = STATE(2451),
    [sym_compound_statement] = STATE(955),
    [sym_attributed_statement] = STATE(955),
    [sym_statement] = STATE(940),
    [sym_labeled_statement] = STATE(955),
    [sym_expression_statement] = STATE(955),
    [sym_if_statement] = STATE(955),
    [sym_switch_statement] = STATE(955),
    [sym_case_statement] = STATE(955),
    [sym_while_statement] = STATE(955),
    [sym_do_statement] = STATE(955),
    [sym_for_statement] = STATE(955),
    [sym_return_statement] = STATE(955),
    [sym_break_statement] = STATE(955),
    [sym_continue_statement] = STATE(955),
    [sym_goto_statement] = STATE(955),
    [sym_expression] = STATE(4419),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6212),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(955),
    [sym_throw_statement] = STATE(955),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(955),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(955),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2451),
    [sym_identifier] = ACTIONS(6034),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2150),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(69),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_case] = ACTIONS(73),
    [anon_sym_default] = ACTIONS(75),
    [anon_sym_while] = ACTIONS(77),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2152),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2154),
    [anon_sym___try] = ACTIONS(2156),
    [anon_sym_ATthrow] = ACTIONS(2158),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2160),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2479)] = {
    [sym_attribute_declaration] = STATE(2438),
    [sym_compound_statement] = STATE(535),
    [sym_attributed_statement] = STATE(535),
    [sym_statement] = STATE(538),
    [sym_labeled_statement] = STATE(535),
    [sym_expression_statement] = STATE(535),
    [sym_if_statement] = STATE(535),
    [sym_switch_statement] = STATE(535),
    [sym_case_statement] = STATE(535),
    [sym_while_statement] = STATE(535),
    [sym_do_statement] = STATE(535),
    [sym_for_statement] = STATE(535),
    [sym_return_statement] = STATE(535),
    [sym_break_statement] = STATE(535),
    [sym_continue_statement] = STATE(535),
    [sym_goto_statement] = STATE(535),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(535),
    [sym_throw_statement] = STATE(535),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(535),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(535),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [sym_identifier] = ACTIONS(5846),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_case] = ACTIONS(878),
    [anon_sym_default] = ACTIONS(880),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2480)] = {
    [sym_attribute_declaration] = STATE(2477),
    [sym_compound_statement] = STATE(825),
    [sym_attributed_statement] = STATE(825),
    [sym_statement] = STATE(712),
    [sym_labeled_statement] = STATE(825),
    [sym_expression_statement] = STATE(825),
    [sym_if_statement] = STATE(825),
    [sym_switch_statement] = STATE(825),
    [sym_case_statement] = STATE(825),
    [sym_while_statement] = STATE(825),
    [sym_do_statement] = STATE(825),
    [sym_for_statement] = STATE(825),
    [sym_return_statement] = STATE(825),
    [sym_break_statement] = STATE(825),
    [sym_continue_statement] = STATE(825),
    [sym_goto_statement] = STATE(825),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(825),
    [sym_throw_statement] = STATE(825),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(825),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(825),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2477),
    [sym_identifier] = ACTIONS(5848),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_case] = ACTIONS(954),
    [anon_sym_default] = ACTIONS(956),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2481)] = {
    [sym__declaration_modifiers] = STATE(2509),
    [sym_attribute_specifier] = STATE(2509),
    [sym_attribute_declaration] = STATE(2509),
    [sym_ms_declspec_modifier] = STATE(2509),
    [sym_storage_class_specifier] = STATE(2509),
    [sym_type_qualifier] = STATE(2509),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2509),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2509),
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_COMMA] = ACTIONS(6254),
    [anon_sym_SEMI] = ACTIONS(6254),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6254),
    [anon_sym_signed] = ACTIONS(6256),
    [anon_sym_unsigned] = ACTIONS(6256),
    [anon_sym_long] = ACTIONS(6256),
    [anon_sym_short] = ACTIONS(6256),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6254),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6256),
    [anon_sym_enum] = ACTIONS(6256),
    [anon_sym_COLON] = ACTIONS(6254),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(6256),
    [anon_sym_in] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6256),
    [anon_sym___typeof] = ACTIONS(6256),
    [anon_sym_typeof] = ACTIONS(6256),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6256),
    [anon_sym_IMP] = ACTIONS(6256),
    [anon_sym_SEL] = ACTIONS(6256),
    [anon_sym_Class] = ACTIONS(6256),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(2482)] = {
    [sym__declaration_modifiers] = STATE(2503),
    [sym_attribute_specifier] = STATE(2503),
    [sym_attribute_declaration] = STATE(2503),
    [sym_ms_declspec_modifier] = STATE(2503),
    [sym_storage_class_specifier] = STATE(2503),
    [sym_type_qualifier] = STATE(2503),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2503),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2503),
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_COMMA] = ACTIONS(6258),
    [anon_sym_SEMI] = ACTIONS(6258),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6258),
    [anon_sym_signed] = ACTIONS(6260),
    [anon_sym_unsigned] = ACTIONS(6260),
    [anon_sym_long] = ACTIONS(6260),
    [anon_sym_short] = ACTIONS(6260),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6258),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6260),
    [anon_sym_enum] = ACTIONS(6260),
    [anon_sym_COLON] = ACTIONS(6258),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(6260),
    [anon_sym_in] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6260),
    [anon_sym___typeof] = ACTIONS(6260),
    [anon_sym_typeof] = ACTIONS(6260),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6260),
    [anon_sym_IMP] = ACTIONS(6260),
    [anon_sym_SEL] = ACTIONS(6260),
    [anon_sym_Class] = ACTIONS(6260),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(2483)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6262),
    [anon_sym_COMMA] = ACTIONS(6264),
    [anon_sym_SEMI] = ACTIONS(6264),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6264),
    [anon_sym_signed] = ACTIONS(6262),
    [anon_sym_unsigned] = ACTIONS(6262),
    [anon_sym_long] = ACTIONS(6262),
    [anon_sym_short] = ACTIONS(6262),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6264),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6262),
    [anon_sym_enum] = ACTIONS(6262),
    [anon_sym_COLON] = ACTIONS(6264),
    [anon_sym_struct] = ACTIONS(6262),
    [anon_sym_union] = ACTIONS(6262),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6262),
    [anon_sym___typeof] = ACTIONS(6262),
    [anon_sym_typeof] = ACTIONS(6262),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6262),
    [anon_sym_IMP] = ACTIONS(6262),
    [anon_sym_SEL] = ACTIONS(6262),
    [anon_sym_Class] = ACTIONS(6262),
    [anon_sym_id] = ACTIONS(6262),
  },
  [STATE(2484)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4470),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6266),
    [anon_sym_COMMA] = ACTIONS(4472),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4472),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4472),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(6274),
    [anon_sym_AMP_AMP] = ACTIONS(6276),
    [anon_sym_PIPE] = ACTIONS(6278),
    [anon_sym_CARET] = ACTIONS(6280),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4470),
    [anon_sym_extern] = ACTIONS(4470),
    [anon_sym___attribute__] = ACTIONS(4470),
    [anon_sym___attribute] = ACTIONS(4470),
    [anon_sym_noreturn] = ACTIONS(4470),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4470),
    [anon_sym_RBRACE] = ACTIONS(4472),
    [anon_sym_static] = ACTIONS(4470),
    [anon_sym_auto] = ACTIONS(4470),
    [anon_sym_register] = ACTIONS(4470),
    [anon_sym_inline] = ACTIONS(4470),
    [anon_sym___inline] = ACTIONS(4470),
    [anon_sym___inline__] = ACTIONS(4470),
    [anon_sym___forceinline] = ACTIONS(4470),
    [anon_sym_thread_local] = ACTIONS(4470),
    [anon_sym___thread] = ACTIONS(4470),
    [anon_sym_CG_EXTERN] = ACTIONS(4470),
    [anon_sym_CG_INLINE] = ACTIONS(4470),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4470),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4470),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4470),
    [anon_sym_IBOutlet] = ACTIONS(4470),
    [anon_sym_IBInspectable] = ACTIONS(4470),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4470),
    [anon_sym_NS_INLINE] = ACTIONS(4470),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4470),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4470),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4470),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4470),
    [anon_sym_const] = ACTIONS(4470),
    [anon_sym_constexpr] = ACTIONS(4470),
    [anon_sym_volatile] = ACTIONS(4470),
    [anon_sym_restrict] = ACTIONS(4470),
    [anon_sym___restrict__] = ACTIONS(4470),
    [anon_sym__Atomic] = ACTIONS(4470),
    [anon_sym__Noreturn] = ACTIONS(4470),
    [anon_sym__Nonnull] = ACTIONS(4470),
    [anon_sym_nullable] = ACTIONS(4470),
    [anon_sym__Complex] = ACTIONS(4470),
    [anon_sym__Nullable] = ACTIONS(4470),
    [anon_sym__Nullable_result] = ACTIONS(4470),
    [anon_sym__Null_unspecified] = ACTIONS(4470),
    [anon_sym___autoreleasing] = ACTIONS(4470),
    [anon_sym___block] = ACTIONS(4470),
    [anon_sym___bridge] = ACTIONS(4470),
    [anon_sym___bridge_retained] = ACTIONS(4470),
    [anon_sym___bridge_transfer] = ACTIONS(4470),
    [anon_sym___complex] = ACTIONS(4470),
    [anon_sym___const] = ACTIONS(4470),
    [anon_sym___imag] = ACTIONS(4470),
    [anon_sym___kindof] = ACTIONS(4470),
    [anon_sym___nonnull] = ACTIONS(4470),
    [anon_sym___nullable] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4470),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4470),
    [anon_sym___real] = ACTIONS(4470),
    [anon_sym___strong] = ACTIONS(4470),
    [anon_sym___unsafe_unretained] = ACTIONS(4470),
    [anon_sym___unused] = ACTIONS(4470),
    [anon_sym___weak] = ACTIONS(4470),
    [anon_sym_alignas] = ACTIONS(4470),
    [anon_sym__Alignas] = ACTIONS(4470),
    [anon_sym_QMARK] = ACTIONS(4472),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4470),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4470),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4470),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4470),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4470),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4470),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_API_AVAILABLE] = ACTIONS(4470),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_API_DEPRECATED] = ACTIONS(4470),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4470),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4470),
    [anon_sym___deprecated_msg] = ACTIONS(4470),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4470),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4470),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4470),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4470),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4470),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4470),
  },
  [STATE(2485)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4000),
    [anon_sym_PLUS] = ACTIONS(4000),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4002),
    [anon_sym_GT_GT] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(2486)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(6276),
    [anon_sym_PIPE] = ACTIONS(6278),
    [anon_sym_CARET] = ACTIONS(6280),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(2487)] = {
    [sym_type_qualifier] = STATE(2487),
    [sym_alignas_qualifier] = STATE(2354),
    [aux_sym__type_definition_type_repeat1] = STATE(2487),
    [sym_identifier] = ACTIONS(6292),
    [anon_sym_COMMA] = ACTIONS(6294),
    [anon_sym_RPAREN] = ACTIONS(6294),
    [anon_sym_LPAREN2] = ACTIONS(6294),
    [anon_sym_STAR] = ACTIONS(6294),
    [anon_sym_CARET] = ACTIONS(6294),
    [anon_sym_GT] = ACTIONS(6294),
    [anon_sym_SEMI] = ACTIONS(6294),
    [anon_sym___extension__] = ACTIONS(6296),
    [anon_sym_extern] = ACTIONS(6292),
    [anon_sym___attribute__] = ACTIONS(6292),
    [anon_sym___attribute] = ACTIONS(6292),
    [anon_sym_noreturn] = ACTIONS(6296),
    [anon_sym_LBRACK] = ACTIONS(6294),
    [anon_sym___declspec] = ACTIONS(6292),
    [anon_sym___based] = ACTIONS(6292),
    [anon_sym_LBRACE] = ACTIONS(6294),
    [anon_sym_signed] = ACTIONS(6292),
    [anon_sym_unsigned] = ACTIONS(6292),
    [anon_sym_long] = ACTIONS(6292),
    [anon_sym_short] = ACTIONS(6292),
    [anon_sym_static] = ACTIONS(6292),
    [anon_sym_ATautoreleasepool] = ACTIONS(6294),
    [anon_sym_auto] = ACTIONS(6292),
    [anon_sym_register] = ACTIONS(6292),
    [anon_sym_inline] = ACTIONS(6292),
    [anon_sym___inline] = ACTIONS(6292),
    [anon_sym___inline__] = ACTIONS(6292),
    [anon_sym___forceinline] = ACTIONS(6292),
    [anon_sym_thread_local] = ACTIONS(6292),
    [anon_sym___thread] = ACTIONS(6292),
    [anon_sym_CG_EXTERN] = ACTIONS(6292),
    [anon_sym_CG_INLINE] = ACTIONS(6292),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6292),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6292),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6292),
    [anon_sym_IBOutlet] = ACTIONS(6292),
    [anon_sym_IBInspectable] = ACTIONS(6292),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6292),
    [anon_sym_NS_INLINE] = ACTIONS(6292),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6292),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6292),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6292),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6292),
    [anon_sym_const] = ACTIONS(6296),
    [anon_sym_constexpr] = ACTIONS(6296),
    [anon_sym_volatile] = ACTIONS(6296),
    [anon_sym_restrict] = ACTIONS(6296),
    [anon_sym___restrict__] = ACTIONS(6296),
    [anon_sym__Atomic] = ACTIONS(6296),
    [anon_sym__Noreturn] = ACTIONS(6296),
    [anon_sym__Nonnull] = ACTIONS(6296),
    [anon_sym_nullable] = ACTIONS(6296),
    [anon_sym__Complex] = ACTIONS(6296),
    [anon_sym__Nullable] = ACTIONS(6296),
    [anon_sym__Nullable_result] = ACTIONS(6296),
    [anon_sym__Null_unspecified] = ACTIONS(6296),
    [anon_sym___autoreleasing] = ACTIONS(6296),
    [anon_sym___block] = ACTIONS(6296),
    [anon_sym___bridge] = ACTIONS(6296),
    [anon_sym___bridge_retained] = ACTIONS(6296),
    [anon_sym___bridge_transfer] = ACTIONS(6296),
    [anon_sym___complex] = ACTIONS(6296),
    [anon_sym___const] = ACTIONS(6296),
    [anon_sym___imag] = ACTIONS(6296),
    [anon_sym___kindof] = ACTIONS(6296),
    [anon_sym___nonnull] = ACTIONS(6296),
    [anon_sym___nullable] = ACTIONS(6296),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6296),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6296),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6296),
    [anon_sym___real] = ACTIONS(6296),
    [anon_sym___strong] = ACTIONS(6296),
    [anon_sym___unsafe_unretained] = ACTIONS(6296),
    [anon_sym___unused] = ACTIONS(6296),
    [anon_sym___weak] = ACTIONS(6296),
    [anon_sym_alignas] = ACTIONS(6299),
    [anon_sym__Alignas] = ACTIONS(6299),
    [sym_primitive_type] = ACTIONS(6292),
    [anon_sym_enum] = ACTIONS(6292),
    [anon_sym_COLON] = ACTIONS(6294),
    [anon_sym_struct] = ACTIONS(6292),
    [anon_sym_union] = ACTIONS(6292),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6292),
    [anon_sym___typeof] = ACTIONS(6292),
    [anon_sym_typeof] = ACTIONS(6292),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6292),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6292),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6292),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6292),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6292),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6292),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6292),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6292),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6292),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6292),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6292),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6292),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6292),
    [anon_sym_API_AVAILABLE] = ACTIONS(6292),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6292),
    [anon_sym_API_DEPRECATED] = ACTIONS(6292),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6292),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6292),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6292),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6292),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6292),
    [anon_sym___deprecated_msg] = ACTIONS(6292),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6292),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6292),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6292),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6292),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6292),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6292),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6292),
    [anon_sym_BOOL] = ACTIONS(6292),
    [anon_sym_IMP] = ACTIONS(6292),
    [anon_sym_SEL] = ACTIONS(6292),
    [anon_sym_Class] = ACTIONS(6292),
    [anon_sym_id] = ACTIONS(6292),
  },
  [STATE(2488)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(6278),
    [anon_sym_CARET] = ACTIONS(6280),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(2489)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(6280),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(2490)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(2491)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(2492)] = {
    [sym__declaration_modifiers] = STATE(2504),
    [sym_attribute_specifier] = STATE(2504),
    [sym_attribute_declaration] = STATE(2504),
    [sym_ms_declspec_modifier] = STATE(2504),
    [sym_storage_class_specifier] = STATE(2504),
    [sym_type_qualifier] = STATE(2504),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2504),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2504),
    [sym_identifier] = ACTIONS(6302),
    [anon_sym_COMMA] = ACTIONS(6304),
    [anon_sym_SEMI] = ACTIONS(6304),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6304),
    [anon_sym_signed] = ACTIONS(6302),
    [anon_sym_unsigned] = ACTIONS(6302),
    [anon_sym_long] = ACTIONS(6302),
    [anon_sym_short] = ACTIONS(6302),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6304),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6302),
    [anon_sym_enum] = ACTIONS(6302),
    [anon_sym_COLON] = ACTIONS(6304),
    [anon_sym_struct] = ACTIONS(6302),
    [anon_sym_union] = ACTIONS(6302),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6302),
    [anon_sym___typeof] = ACTIONS(6302),
    [anon_sym_typeof] = ACTIONS(6302),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6302),
    [anon_sym_IMP] = ACTIONS(6302),
    [anon_sym_SEL] = ACTIONS(6302),
    [anon_sym_Class] = ACTIONS(6302),
    [anon_sym_id] = ACTIONS(6302),
  },
  [STATE(2493)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(2494)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(2495)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4000),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4002),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4002),
    [anon_sym_GT_GT] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4000),
    [anon_sym_extern] = ACTIONS(4000),
    [anon_sym___attribute__] = ACTIONS(4000),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4000),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4000),
    [anon_sym_RBRACE] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4000),
    [anon_sym_auto] = ACTIONS(4000),
    [anon_sym_register] = ACTIONS(4000),
    [anon_sym_inline] = ACTIONS(4000),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4000),
    [anon_sym___forceinline] = ACTIONS(4000),
    [anon_sym_thread_local] = ACTIONS(4000),
    [anon_sym___thread] = ACTIONS(4000),
    [anon_sym_CG_EXTERN] = ACTIONS(4000),
    [anon_sym_CG_INLINE] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4000),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4000),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4000),
    [anon_sym_IBOutlet] = ACTIONS(4000),
    [anon_sym_IBInspectable] = ACTIONS(4000),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4000),
    [anon_sym_NS_INLINE] = ACTIONS(4000),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4000),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4000),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4000),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4000),
    [anon_sym_volatile] = ACTIONS(4000),
    [anon_sym_restrict] = ACTIONS(4000),
    [anon_sym___restrict__] = ACTIONS(4000),
    [anon_sym__Atomic] = ACTIONS(4000),
    [anon_sym__Noreturn] = ACTIONS(4000),
    [anon_sym__Nonnull] = ACTIONS(4000),
    [anon_sym_nullable] = ACTIONS(4000),
    [anon_sym__Complex] = ACTIONS(4000),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4000),
    [anon_sym__Null_unspecified] = ACTIONS(4000),
    [anon_sym___autoreleasing] = ACTIONS(4000),
    [anon_sym___block] = ACTIONS(4000),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4000),
    [anon_sym___bridge_transfer] = ACTIONS(4000),
    [anon_sym___complex] = ACTIONS(4000),
    [anon_sym___const] = ACTIONS(4000),
    [anon_sym___imag] = ACTIONS(4000),
    [anon_sym___kindof] = ACTIONS(4000),
    [anon_sym___nonnull] = ACTIONS(4000),
    [anon_sym___nullable] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4000),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4000),
    [anon_sym___real] = ACTIONS(4000),
    [anon_sym___strong] = ACTIONS(4000),
    [anon_sym___unsafe_unretained] = ACTIONS(4000),
    [anon_sym___unused] = ACTIONS(4000),
    [anon_sym___weak] = ACTIONS(4000),
    [anon_sym_alignas] = ACTIONS(4000),
    [anon_sym__Alignas] = ACTIONS(4000),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4000),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4000),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4000),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4000),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_API_AVAILABLE] = ACTIONS(4000),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_API_DEPRECATED] = ACTIONS(4000),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4000),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4000),
    [anon_sym___deprecated_msg] = ACTIONS(4000),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4000),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4000),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4000),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4000),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4000),
  },
  [STATE(2496)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4478),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6266),
    [anon_sym_COMMA] = ACTIONS(4480),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4480),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4480),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(6274),
    [anon_sym_AMP_AMP] = ACTIONS(6276),
    [anon_sym_PIPE] = ACTIONS(6278),
    [anon_sym_CARET] = ACTIONS(6280),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4478),
    [anon_sym_extern] = ACTIONS(4478),
    [anon_sym___attribute__] = ACTIONS(4478),
    [anon_sym___attribute] = ACTIONS(4478),
    [anon_sym_noreturn] = ACTIONS(4478),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4478),
    [anon_sym_RBRACE] = ACTIONS(4480),
    [anon_sym_static] = ACTIONS(4478),
    [anon_sym_auto] = ACTIONS(4478),
    [anon_sym_register] = ACTIONS(4478),
    [anon_sym_inline] = ACTIONS(4478),
    [anon_sym___inline] = ACTIONS(4478),
    [anon_sym___inline__] = ACTIONS(4478),
    [anon_sym___forceinline] = ACTIONS(4478),
    [anon_sym_thread_local] = ACTIONS(4478),
    [anon_sym___thread] = ACTIONS(4478),
    [anon_sym_CG_EXTERN] = ACTIONS(4478),
    [anon_sym_CG_INLINE] = ACTIONS(4478),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4478),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4478),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4478),
    [anon_sym_IBOutlet] = ACTIONS(4478),
    [anon_sym_IBInspectable] = ACTIONS(4478),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4478),
    [anon_sym_NS_INLINE] = ACTIONS(4478),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4478),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4478),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4478),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4478),
    [anon_sym_const] = ACTIONS(4478),
    [anon_sym_constexpr] = ACTIONS(4478),
    [anon_sym_volatile] = ACTIONS(4478),
    [anon_sym_restrict] = ACTIONS(4478),
    [anon_sym___restrict__] = ACTIONS(4478),
    [anon_sym__Atomic] = ACTIONS(4478),
    [anon_sym__Noreturn] = ACTIONS(4478),
    [anon_sym__Nonnull] = ACTIONS(4478),
    [anon_sym_nullable] = ACTIONS(4478),
    [anon_sym__Complex] = ACTIONS(4478),
    [anon_sym__Nullable] = ACTIONS(4478),
    [anon_sym__Nullable_result] = ACTIONS(4478),
    [anon_sym__Null_unspecified] = ACTIONS(4478),
    [anon_sym___autoreleasing] = ACTIONS(4478),
    [anon_sym___block] = ACTIONS(4478),
    [anon_sym___bridge] = ACTIONS(4478),
    [anon_sym___bridge_retained] = ACTIONS(4478),
    [anon_sym___bridge_transfer] = ACTIONS(4478),
    [anon_sym___complex] = ACTIONS(4478),
    [anon_sym___const] = ACTIONS(4478),
    [anon_sym___imag] = ACTIONS(4478),
    [anon_sym___kindof] = ACTIONS(4478),
    [anon_sym___nonnull] = ACTIONS(4478),
    [anon_sym___nullable] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4478),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4478),
    [anon_sym___real] = ACTIONS(4478),
    [anon_sym___strong] = ACTIONS(4478),
    [anon_sym___unsafe_unretained] = ACTIONS(4478),
    [anon_sym___unused] = ACTIONS(4478),
    [anon_sym___weak] = ACTIONS(4478),
    [anon_sym_alignas] = ACTIONS(4478),
    [anon_sym__Alignas] = ACTIONS(4478),
    [anon_sym_QMARK] = ACTIONS(6306),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4478),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4478),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4478),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4478),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4478),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4478),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_API_AVAILABLE] = ACTIONS(4478),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_API_DEPRECATED] = ACTIONS(4478),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4478),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4478),
    [anon_sym___deprecated_msg] = ACTIONS(4478),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4478),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4478),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4478),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4478),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4478),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4478),
  },
  [STATE(2497)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4458),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6266),
    [anon_sym_COMMA] = ACTIONS(4462),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4462),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4462),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(6274),
    [anon_sym_AMP_AMP] = ACTIONS(6276),
    [anon_sym_PIPE] = ACTIONS(6278),
    [anon_sym_CARET] = ACTIONS(6280),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4458),
    [anon_sym_extern] = ACTIONS(4458),
    [anon_sym___attribute__] = ACTIONS(4458),
    [anon_sym___attribute] = ACTIONS(4458),
    [anon_sym_noreturn] = ACTIONS(4458),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4458),
    [anon_sym_RBRACE] = ACTIONS(4462),
    [anon_sym_static] = ACTIONS(4458),
    [anon_sym_auto] = ACTIONS(4458),
    [anon_sym_register] = ACTIONS(4458),
    [anon_sym_inline] = ACTIONS(4458),
    [anon_sym___inline] = ACTIONS(4458),
    [anon_sym___inline__] = ACTIONS(4458),
    [anon_sym___forceinline] = ACTIONS(4458),
    [anon_sym_thread_local] = ACTIONS(4458),
    [anon_sym___thread] = ACTIONS(4458),
    [anon_sym_CG_EXTERN] = ACTIONS(4458),
    [anon_sym_CG_INLINE] = ACTIONS(4458),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4458),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4458),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4458),
    [anon_sym_IBOutlet] = ACTIONS(4458),
    [anon_sym_IBInspectable] = ACTIONS(4458),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4458),
    [anon_sym_NS_INLINE] = ACTIONS(4458),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4458),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4458),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4458),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4458),
    [anon_sym_const] = ACTIONS(4458),
    [anon_sym_constexpr] = ACTIONS(4458),
    [anon_sym_volatile] = ACTIONS(4458),
    [anon_sym_restrict] = ACTIONS(4458),
    [anon_sym___restrict__] = ACTIONS(4458),
    [anon_sym__Atomic] = ACTIONS(4458),
    [anon_sym__Noreturn] = ACTIONS(4458),
    [anon_sym__Nonnull] = ACTIONS(4458),
    [anon_sym_nullable] = ACTIONS(4458),
    [anon_sym__Complex] = ACTIONS(4458),
    [anon_sym__Nullable] = ACTIONS(4458),
    [anon_sym__Nullable_result] = ACTIONS(4458),
    [anon_sym__Null_unspecified] = ACTIONS(4458),
    [anon_sym___autoreleasing] = ACTIONS(4458),
    [anon_sym___block] = ACTIONS(4458),
    [anon_sym___bridge] = ACTIONS(4458),
    [anon_sym___bridge_retained] = ACTIONS(4458),
    [anon_sym___bridge_transfer] = ACTIONS(4458),
    [anon_sym___complex] = ACTIONS(4458),
    [anon_sym___const] = ACTIONS(4458),
    [anon_sym___imag] = ACTIONS(4458),
    [anon_sym___kindof] = ACTIONS(4458),
    [anon_sym___nonnull] = ACTIONS(4458),
    [anon_sym___nullable] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4458),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4458),
    [anon_sym___real] = ACTIONS(4458),
    [anon_sym___strong] = ACTIONS(4458),
    [anon_sym___unsafe_unretained] = ACTIONS(4458),
    [anon_sym___unused] = ACTIONS(4458),
    [anon_sym___weak] = ACTIONS(4458),
    [anon_sym_alignas] = ACTIONS(4458),
    [anon_sym__Alignas] = ACTIONS(4458),
    [anon_sym_QMARK] = ACTIONS(6306),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4458),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4458),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4458),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4458),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4458),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4458),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_API_AVAILABLE] = ACTIONS(4458),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_API_DEPRECATED] = ACTIONS(4458),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4458),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4458),
    [anon_sym___deprecated_msg] = ACTIONS(4458),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4458),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4458),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4458),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4458),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4458),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4458),
  },
  [STATE(2498)] = {
    [sym_protocol_reference_list] = STATE(2964),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_COMMA] = ACTIONS(4044),
    [anon_sym_RPAREN] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6310),
    [anon_sym_DASH] = ACTIONS(4042),
    [anon_sym_PLUS] = ACTIONS(4042),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4044),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(4042),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LT_LT] = ACTIONS(4044),
    [anon_sym_GT_GT] = ACTIONS(4044),
    [anon_sym_SEMI] = ACTIONS(6310),
    [anon_sym___extension__] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [anon_sym___attribute__] = ACTIONS(6308),
    [anon_sym___attribute] = ACTIONS(6308),
    [anon_sym_noreturn] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6310),
    [anon_sym___declspec] = ACTIONS(6308),
    [anon_sym___based] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_auto] = ACTIONS(6308),
    [anon_sym_register] = ACTIONS(6308),
    [anon_sym_inline] = ACTIONS(6308),
    [anon_sym___inline] = ACTIONS(6308),
    [anon_sym___inline__] = ACTIONS(6308),
    [anon_sym___forceinline] = ACTIONS(6308),
    [anon_sym_thread_local] = ACTIONS(6308),
    [anon_sym___thread] = ACTIONS(6308),
    [anon_sym_CG_EXTERN] = ACTIONS(6308),
    [anon_sym_CG_INLINE] = ACTIONS(6308),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6308),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6308),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6308),
    [anon_sym_IBOutlet] = ACTIONS(6308),
    [anon_sym_IBInspectable] = ACTIONS(6308),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6308),
    [anon_sym_NS_INLINE] = ACTIONS(6308),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6308),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6308),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6308),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6308),
    [anon_sym_const] = ACTIONS(6308),
    [anon_sym_constexpr] = ACTIONS(6308),
    [anon_sym_volatile] = ACTIONS(6308),
    [anon_sym_restrict] = ACTIONS(6308),
    [anon_sym___restrict__] = ACTIONS(6308),
    [anon_sym__Atomic] = ACTIONS(6308),
    [anon_sym__Noreturn] = ACTIONS(6308),
    [anon_sym__Nonnull] = ACTIONS(6308),
    [anon_sym_nullable] = ACTIONS(6308),
    [anon_sym__Complex] = ACTIONS(6308),
    [anon_sym__Nullable] = ACTIONS(6308),
    [anon_sym__Nullable_result] = ACTIONS(6308),
    [anon_sym__Null_unspecified] = ACTIONS(6308),
    [anon_sym___autoreleasing] = ACTIONS(6308),
    [anon_sym___block] = ACTIONS(6308),
    [anon_sym___bridge] = ACTIONS(6308),
    [anon_sym___bridge_retained] = ACTIONS(6308),
    [anon_sym___bridge_transfer] = ACTIONS(6308),
    [anon_sym___complex] = ACTIONS(6308),
    [anon_sym___const] = ACTIONS(6308),
    [anon_sym___imag] = ACTIONS(6308),
    [anon_sym___kindof] = ACTIONS(6308),
    [anon_sym___nonnull] = ACTIONS(6308),
    [anon_sym___nullable] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6308),
    [anon_sym___real] = ACTIONS(6308),
    [anon_sym___strong] = ACTIONS(6308),
    [anon_sym___unsafe_unretained] = ACTIONS(6308),
    [anon_sym___unused] = ACTIONS(6308),
    [anon_sym___weak] = ACTIONS(6308),
    [anon_sym_alignas] = ACTIONS(6308),
    [anon_sym__Alignas] = ACTIONS(6308),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(4044),
    [anon_sym_PLUS_PLUS] = ACTIONS(4044),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6308),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6308),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6308),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6308),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6308),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6308),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6308),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_API_AVAILABLE] = ACTIONS(6308),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_API_DEPRECATED] = ACTIONS(6308),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6308),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym___deprecated_msg] = ACTIONS(6308),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6308),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6308),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6308),
  },
  [STATE(2499)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(6314),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6266),
    [anon_sym_COMMA] = ACTIONS(6316),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6316),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6316),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(6274),
    [anon_sym_AMP_AMP] = ACTIONS(6276),
    [anon_sym_PIPE] = ACTIONS(6278),
    [anon_sym_CARET] = ACTIONS(6280),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(6314),
    [anon_sym_extern] = ACTIONS(6314),
    [anon_sym___attribute__] = ACTIONS(6314),
    [anon_sym___attribute] = ACTIONS(6314),
    [anon_sym_noreturn] = ACTIONS(6314),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(6314),
    [anon_sym_RBRACE] = ACTIONS(6316),
    [anon_sym_static] = ACTIONS(6314),
    [anon_sym_auto] = ACTIONS(6314),
    [anon_sym_register] = ACTIONS(6314),
    [anon_sym_inline] = ACTIONS(6314),
    [anon_sym___inline] = ACTIONS(6314),
    [anon_sym___inline__] = ACTIONS(6314),
    [anon_sym___forceinline] = ACTIONS(6314),
    [anon_sym_thread_local] = ACTIONS(6314),
    [anon_sym___thread] = ACTIONS(6314),
    [anon_sym_CG_EXTERN] = ACTIONS(6314),
    [anon_sym_CG_INLINE] = ACTIONS(6314),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6314),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6314),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6314),
    [anon_sym_IBOutlet] = ACTIONS(6314),
    [anon_sym_IBInspectable] = ACTIONS(6314),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6314),
    [anon_sym_NS_INLINE] = ACTIONS(6314),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6314),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6314),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6314),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6314),
    [anon_sym_const] = ACTIONS(6314),
    [anon_sym_constexpr] = ACTIONS(6314),
    [anon_sym_volatile] = ACTIONS(6314),
    [anon_sym_restrict] = ACTIONS(6314),
    [anon_sym___restrict__] = ACTIONS(6314),
    [anon_sym__Atomic] = ACTIONS(6314),
    [anon_sym__Noreturn] = ACTIONS(6314),
    [anon_sym__Nonnull] = ACTIONS(6314),
    [anon_sym_nullable] = ACTIONS(6314),
    [anon_sym__Complex] = ACTIONS(6314),
    [anon_sym__Nullable] = ACTIONS(6314),
    [anon_sym__Nullable_result] = ACTIONS(6314),
    [anon_sym__Null_unspecified] = ACTIONS(6314),
    [anon_sym___autoreleasing] = ACTIONS(6314),
    [anon_sym___block] = ACTIONS(6314),
    [anon_sym___bridge] = ACTIONS(6314),
    [anon_sym___bridge_retained] = ACTIONS(6314),
    [anon_sym___bridge_transfer] = ACTIONS(6314),
    [anon_sym___complex] = ACTIONS(6314),
    [anon_sym___const] = ACTIONS(6314),
    [anon_sym___imag] = ACTIONS(6314),
    [anon_sym___kindof] = ACTIONS(6314),
    [anon_sym___nonnull] = ACTIONS(6314),
    [anon_sym___nullable] = ACTIONS(6314),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6314),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6314),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6314),
    [anon_sym___real] = ACTIONS(6314),
    [anon_sym___strong] = ACTIONS(6314),
    [anon_sym___unsafe_unretained] = ACTIONS(6314),
    [anon_sym___unused] = ACTIONS(6314),
    [anon_sym___weak] = ACTIONS(6314),
    [anon_sym_alignas] = ACTIONS(6314),
    [anon_sym__Alignas] = ACTIONS(6314),
    [anon_sym_QMARK] = ACTIONS(6306),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6314),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6314),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6314),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6314),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6314),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6314),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6314),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6314),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6314),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6314),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6314),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6314),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6314),
    [anon_sym_API_AVAILABLE] = ACTIONS(6314),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6314),
    [anon_sym_API_DEPRECATED] = ACTIONS(6314),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6314),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6314),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6314),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6314),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6314),
    [anon_sym___deprecated_msg] = ACTIONS(6314),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6314),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6314),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6314),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6314),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6314),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6314),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6314),
  },
  [STATE(2500)] = {
    [sym__declaration_modifiers] = STATE(2509),
    [sym_attribute_specifier] = STATE(2509),
    [sym_attribute_declaration] = STATE(2509),
    [sym_ms_declspec_modifier] = STATE(2509),
    [sym_storage_class_specifier] = STATE(2509),
    [sym_type_qualifier] = STATE(2509),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2509),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2509),
    [sym_identifier] = ACTIONS(6256),
    [anon_sym_COMMA] = ACTIONS(6254),
    [anon_sym_SEMI] = ACTIONS(6254),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6254),
    [anon_sym_signed] = ACTIONS(6256),
    [anon_sym_unsigned] = ACTIONS(6256),
    [anon_sym_long] = ACTIONS(6256),
    [anon_sym_short] = ACTIONS(6256),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6254),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6256),
    [anon_sym_enum] = ACTIONS(6256),
    [anon_sym_COLON] = ACTIONS(6254),
    [anon_sym_struct] = ACTIONS(6256),
    [anon_sym_union] = ACTIONS(6256),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6256),
    [anon_sym___typeof] = ACTIONS(6256),
    [anon_sym_typeof] = ACTIONS(6256),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6256),
    [anon_sym_IMP] = ACTIONS(6256),
    [anon_sym_SEL] = ACTIONS(6256),
    [anon_sym_Class] = ACTIONS(6256),
    [anon_sym_id] = ACTIONS(6256),
  },
  [STATE(2501)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6318),
    [anon_sym_COMMA] = ACTIONS(6320),
    [anon_sym_SEMI] = ACTIONS(6320),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6320),
    [anon_sym_signed] = ACTIONS(6318),
    [anon_sym_unsigned] = ACTIONS(6318),
    [anon_sym_long] = ACTIONS(6318),
    [anon_sym_short] = ACTIONS(6318),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6320),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6318),
    [anon_sym_enum] = ACTIONS(6318),
    [anon_sym_COLON] = ACTIONS(6320),
    [anon_sym_struct] = ACTIONS(6318),
    [anon_sym_union] = ACTIONS(6318),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6318),
    [anon_sym___typeof] = ACTIONS(6318),
    [anon_sym_typeof] = ACTIONS(6318),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6318),
    [anon_sym_IMP] = ACTIONS(6318),
    [anon_sym_SEL] = ACTIONS(6318),
    [anon_sym_Class] = ACTIONS(6318),
    [anon_sym_id] = ACTIONS(6318),
  },
  [STATE(2502)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6302),
    [anon_sym_COMMA] = ACTIONS(6304),
    [anon_sym_SEMI] = ACTIONS(6304),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6304),
    [anon_sym_signed] = ACTIONS(6302),
    [anon_sym_unsigned] = ACTIONS(6302),
    [anon_sym_long] = ACTIONS(6302),
    [anon_sym_short] = ACTIONS(6302),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6304),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6302),
    [anon_sym_enum] = ACTIONS(6302),
    [anon_sym_COLON] = ACTIONS(6304),
    [anon_sym_struct] = ACTIONS(6302),
    [anon_sym_union] = ACTIONS(6302),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6302),
    [anon_sym___typeof] = ACTIONS(6302),
    [anon_sym_typeof] = ACTIONS(6302),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6302),
    [anon_sym_IMP] = ACTIONS(6302),
    [anon_sym_SEL] = ACTIONS(6302),
    [anon_sym_Class] = ACTIONS(6302),
    [anon_sym_id] = ACTIONS(6302),
  },
  [STATE(2503)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6322),
    [anon_sym_COMMA] = ACTIONS(6324),
    [anon_sym_SEMI] = ACTIONS(6324),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6324),
    [anon_sym_signed] = ACTIONS(6322),
    [anon_sym_unsigned] = ACTIONS(6322),
    [anon_sym_long] = ACTIONS(6322),
    [anon_sym_short] = ACTIONS(6322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6324),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6322),
    [anon_sym_enum] = ACTIONS(6322),
    [anon_sym_COLON] = ACTIONS(6324),
    [anon_sym_struct] = ACTIONS(6322),
    [anon_sym_union] = ACTIONS(6322),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6322),
    [anon_sym___typeof] = ACTIONS(6322),
    [anon_sym_typeof] = ACTIONS(6322),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6322),
    [anon_sym_IMP] = ACTIONS(6322),
    [anon_sym_SEL] = ACTIONS(6322),
    [anon_sym_Class] = ACTIONS(6322),
    [anon_sym_id] = ACTIONS(6322),
  },
  [STATE(2504)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6326),
    [anon_sym_COMMA] = ACTIONS(6328),
    [anon_sym_SEMI] = ACTIONS(6328),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6328),
    [anon_sym_signed] = ACTIONS(6326),
    [anon_sym_unsigned] = ACTIONS(6326),
    [anon_sym_long] = ACTIONS(6326),
    [anon_sym_short] = ACTIONS(6326),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6328),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6326),
    [anon_sym_enum] = ACTIONS(6326),
    [anon_sym_COLON] = ACTIONS(6328),
    [anon_sym_struct] = ACTIONS(6326),
    [anon_sym_union] = ACTIONS(6326),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6326),
    [anon_sym___typeof] = ACTIONS(6326),
    [anon_sym_typeof] = ACTIONS(6326),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6326),
    [anon_sym_IMP] = ACTIONS(6326),
    [anon_sym_SEL] = ACTIONS(6326),
    [anon_sym_Class] = ACTIONS(6326),
    [anon_sym_id] = ACTIONS(6326),
  },
  [STATE(2505)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6330),
    [anon_sym_COMMA] = ACTIONS(4436),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4436),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4436),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(6274),
    [anon_sym_AMP_AMP] = ACTIONS(6276),
    [anon_sym_PIPE] = ACTIONS(6278),
    [anon_sym_CARET] = ACTIONS(6280),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4431),
    [anon_sym_extern] = ACTIONS(4431),
    [anon_sym___attribute__] = ACTIONS(4431),
    [anon_sym___attribute] = ACTIONS(4431),
    [anon_sym_noreturn] = ACTIONS(4431),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4431),
    [anon_sym_RBRACE] = ACTIONS(4436),
    [anon_sym_static] = ACTIONS(4431),
    [anon_sym_auto] = ACTIONS(4431),
    [anon_sym_register] = ACTIONS(4431),
    [anon_sym_inline] = ACTIONS(4431),
    [anon_sym___inline] = ACTIONS(4431),
    [anon_sym___inline__] = ACTIONS(4431),
    [anon_sym___forceinline] = ACTIONS(4431),
    [anon_sym_thread_local] = ACTIONS(4431),
    [anon_sym___thread] = ACTIONS(4431),
    [anon_sym_CG_EXTERN] = ACTIONS(4431),
    [anon_sym_CG_INLINE] = ACTIONS(4431),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4431),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4431),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4431),
    [anon_sym_IBOutlet] = ACTIONS(4431),
    [anon_sym_IBInspectable] = ACTIONS(4431),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4431),
    [anon_sym_NS_INLINE] = ACTIONS(4431),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4431),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4431),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4431),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4431),
    [anon_sym_const] = ACTIONS(4431),
    [anon_sym_constexpr] = ACTIONS(4431),
    [anon_sym_volatile] = ACTIONS(4431),
    [anon_sym_restrict] = ACTIONS(4431),
    [anon_sym___restrict__] = ACTIONS(4431),
    [anon_sym__Atomic] = ACTIONS(4431),
    [anon_sym__Noreturn] = ACTIONS(4431),
    [anon_sym__Nonnull] = ACTIONS(4431),
    [anon_sym_nullable] = ACTIONS(4431),
    [anon_sym__Complex] = ACTIONS(4431),
    [anon_sym__Nullable] = ACTIONS(4431),
    [anon_sym__Nullable_result] = ACTIONS(4431),
    [anon_sym__Null_unspecified] = ACTIONS(4431),
    [anon_sym___autoreleasing] = ACTIONS(4431),
    [anon_sym___block] = ACTIONS(4431),
    [anon_sym___bridge] = ACTIONS(4431),
    [anon_sym___bridge_retained] = ACTIONS(4431),
    [anon_sym___bridge_transfer] = ACTIONS(4431),
    [anon_sym___complex] = ACTIONS(4431),
    [anon_sym___const] = ACTIONS(4431),
    [anon_sym___imag] = ACTIONS(4431),
    [anon_sym___kindof] = ACTIONS(4431),
    [anon_sym___nonnull] = ACTIONS(4431),
    [anon_sym___nullable] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4431),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4431),
    [anon_sym___real] = ACTIONS(4431),
    [anon_sym___strong] = ACTIONS(4431),
    [anon_sym___unsafe_unretained] = ACTIONS(4431),
    [anon_sym___unused] = ACTIONS(4431),
    [anon_sym___weak] = ACTIONS(4431),
    [anon_sym_alignas] = ACTIONS(4431),
    [anon_sym__Alignas] = ACTIONS(4431),
    [anon_sym_QMARK] = ACTIONS(4436),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4431),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4431),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4431),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4431),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4431),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4431),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4431),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_API_AVAILABLE] = ACTIONS(4431),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_API_DEPRECATED] = ACTIONS(4431),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4431),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4431),
    [anon_sym___deprecated_msg] = ACTIONS(4431),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4431),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4431),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4431),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4431),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4431),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4431),
  },
  [STATE(2506)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4474),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6266),
    [anon_sym_COMMA] = ACTIONS(4476),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4476),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4476),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(6274),
    [anon_sym_AMP_AMP] = ACTIONS(6276),
    [anon_sym_PIPE] = ACTIONS(6278),
    [anon_sym_CARET] = ACTIONS(6280),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4474),
    [anon_sym_extern] = ACTIONS(4474),
    [anon_sym___attribute__] = ACTIONS(4474),
    [anon_sym___attribute] = ACTIONS(4474),
    [anon_sym_noreturn] = ACTIONS(4474),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4474),
    [anon_sym_RBRACE] = ACTIONS(4476),
    [anon_sym_static] = ACTIONS(4474),
    [anon_sym_auto] = ACTIONS(4474),
    [anon_sym_register] = ACTIONS(4474),
    [anon_sym_inline] = ACTIONS(4474),
    [anon_sym___inline] = ACTIONS(4474),
    [anon_sym___inline__] = ACTIONS(4474),
    [anon_sym___forceinline] = ACTIONS(4474),
    [anon_sym_thread_local] = ACTIONS(4474),
    [anon_sym___thread] = ACTIONS(4474),
    [anon_sym_CG_EXTERN] = ACTIONS(4474),
    [anon_sym_CG_INLINE] = ACTIONS(4474),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4474),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4474),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4474),
    [anon_sym_IBOutlet] = ACTIONS(4474),
    [anon_sym_IBInspectable] = ACTIONS(4474),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4474),
    [anon_sym_NS_INLINE] = ACTIONS(4474),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4474),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4474),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4474),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4474),
    [anon_sym_const] = ACTIONS(4474),
    [anon_sym_constexpr] = ACTIONS(4474),
    [anon_sym_volatile] = ACTIONS(4474),
    [anon_sym_restrict] = ACTIONS(4474),
    [anon_sym___restrict__] = ACTIONS(4474),
    [anon_sym__Atomic] = ACTIONS(4474),
    [anon_sym__Noreturn] = ACTIONS(4474),
    [anon_sym__Nonnull] = ACTIONS(4474),
    [anon_sym_nullable] = ACTIONS(4474),
    [anon_sym__Complex] = ACTIONS(4474),
    [anon_sym__Nullable] = ACTIONS(4474),
    [anon_sym__Nullable_result] = ACTIONS(4474),
    [anon_sym__Null_unspecified] = ACTIONS(4474),
    [anon_sym___autoreleasing] = ACTIONS(4474),
    [anon_sym___block] = ACTIONS(4474),
    [anon_sym___bridge] = ACTIONS(4474),
    [anon_sym___bridge_retained] = ACTIONS(4474),
    [anon_sym___bridge_transfer] = ACTIONS(4474),
    [anon_sym___complex] = ACTIONS(4474),
    [anon_sym___const] = ACTIONS(4474),
    [anon_sym___imag] = ACTIONS(4474),
    [anon_sym___kindof] = ACTIONS(4474),
    [anon_sym___nonnull] = ACTIONS(4474),
    [anon_sym___nullable] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4474),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4474),
    [anon_sym___real] = ACTIONS(4474),
    [anon_sym___strong] = ACTIONS(4474),
    [anon_sym___unsafe_unretained] = ACTIONS(4474),
    [anon_sym___unused] = ACTIONS(4474),
    [anon_sym___weak] = ACTIONS(4474),
    [anon_sym_alignas] = ACTIONS(4474),
    [anon_sym__Alignas] = ACTIONS(4474),
    [anon_sym_QMARK] = ACTIONS(4476),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4474),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4474),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4474),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4474),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4474),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4474),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_API_AVAILABLE] = ACTIONS(4474),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_API_DEPRECATED] = ACTIONS(4474),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4474),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4474),
    [anon_sym___deprecated_msg] = ACTIONS(4474),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4474),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4474),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4474),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4474),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4474),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4474),
  },
  [STATE(2507)] = {
    [sym__declaration_modifiers] = STATE(2483),
    [sym_attribute_specifier] = STATE(2483),
    [sym_attribute_declaration] = STATE(2483),
    [sym_ms_declspec_modifier] = STATE(2483),
    [sym_storage_class_specifier] = STATE(2483),
    [sym_type_qualifier] = STATE(2483),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2483),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2483),
    [sym_identifier] = ACTIONS(6333),
    [anon_sym_COMMA] = ACTIONS(6335),
    [anon_sym_SEMI] = ACTIONS(6335),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6335),
    [anon_sym_signed] = ACTIONS(6333),
    [anon_sym_unsigned] = ACTIONS(6333),
    [anon_sym_long] = ACTIONS(6333),
    [anon_sym_short] = ACTIONS(6333),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6335),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6333),
    [anon_sym_enum] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6335),
    [anon_sym_struct] = ACTIONS(6333),
    [anon_sym_union] = ACTIONS(6333),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6333),
    [anon_sym___typeof] = ACTIONS(6333),
    [anon_sym_typeof] = ACTIONS(6333),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6333),
    [anon_sym_IMP] = ACTIONS(6333),
    [anon_sym_SEL] = ACTIONS(6333),
    [anon_sym_Class] = ACTIONS(6333),
    [anon_sym_id] = ACTIONS(6333),
  },
  [STATE(2508)] = {
    [sym__declaration_modifiers] = STATE(2501),
    [sym_attribute_specifier] = STATE(2501),
    [sym_attribute_declaration] = STATE(2501),
    [sym_ms_declspec_modifier] = STATE(2501),
    [sym_storage_class_specifier] = STATE(2501),
    [sym_type_qualifier] = STATE(2501),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2501),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2501),
    [sym_identifier] = ACTIONS(6326),
    [anon_sym_COMMA] = ACTIONS(6328),
    [anon_sym_SEMI] = ACTIONS(6328),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6328),
    [anon_sym_signed] = ACTIONS(6326),
    [anon_sym_unsigned] = ACTIONS(6326),
    [anon_sym_long] = ACTIONS(6326),
    [anon_sym_short] = ACTIONS(6326),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6328),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6326),
    [anon_sym_enum] = ACTIONS(6326),
    [anon_sym_COLON] = ACTIONS(6328),
    [anon_sym_struct] = ACTIONS(6326),
    [anon_sym_union] = ACTIONS(6326),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6326),
    [anon_sym___typeof] = ACTIONS(6326),
    [anon_sym_typeof] = ACTIONS(6326),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6326),
    [anon_sym_IMP] = ACTIONS(6326),
    [anon_sym_SEL] = ACTIONS(6326),
    [anon_sym_Class] = ACTIONS(6326),
    [anon_sym_id] = ACTIONS(6326),
  },
  [STATE(2509)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6337),
    [anon_sym_COMMA] = ACTIONS(6339),
    [anon_sym_SEMI] = ACTIONS(6339),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6339),
    [anon_sym_signed] = ACTIONS(6337),
    [anon_sym_unsigned] = ACTIONS(6337),
    [anon_sym_long] = ACTIONS(6337),
    [anon_sym_short] = ACTIONS(6337),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6339),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6337),
    [anon_sym_enum] = ACTIONS(6337),
    [anon_sym_COLON] = ACTIONS(6339),
    [anon_sym_struct] = ACTIONS(6337),
    [anon_sym_union] = ACTIONS(6337),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6337),
    [anon_sym___typeof] = ACTIONS(6337),
    [anon_sym_typeof] = ACTIONS(6337),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6337),
    [anon_sym_IMP] = ACTIONS(6337),
    [anon_sym_SEL] = ACTIONS(6337),
    [anon_sym_Class] = ACTIONS(6337),
    [anon_sym_id] = ACTIONS(6337),
  },
  [STATE(2510)] = {
    [sym_type_qualifier] = STATE(3484),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4543),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(2623),
    [aux_sym_array_declarator_repeat1] = STATE(2527),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6341),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6347),
    [anon_sym_static] = ACTIONS(6349),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2511)] = {
    [sym__declaration_modifiers] = STATE(2502),
    [sym_attribute_specifier] = STATE(2502),
    [sym_attribute_declaration] = STATE(2502),
    [sym_ms_declspec_modifier] = STATE(2502),
    [sym_storage_class_specifier] = STATE(2502),
    [sym_type_qualifier] = STATE(2502),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2502),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2502),
    [sym_identifier] = ACTIONS(6357),
    [anon_sym_COMMA] = ACTIONS(6359),
    [anon_sym_SEMI] = ACTIONS(6359),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6359),
    [anon_sym_signed] = ACTIONS(6357),
    [anon_sym_unsigned] = ACTIONS(6357),
    [anon_sym_long] = ACTIONS(6357),
    [anon_sym_short] = ACTIONS(6357),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6359),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6357),
    [anon_sym_enum] = ACTIONS(6357),
    [anon_sym_COLON] = ACTIONS(6359),
    [anon_sym_struct] = ACTIONS(6357),
    [anon_sym_union] = ACTIONS(6357),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6357),
    [anon_sym___typeof] = ACTIONS(6357),
    [anon_sym_typeof] = ACTIONS(6357),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6357),
    [anon_sym_IMP] = ACTIONS(6357),
    [anon_sym_SEL] = ACTIONS(6357),
    [anon_sym_Class] = ACTIONS(6357),
    [anon_sym_id] = ACTIONS(6357),
  },
  [STATE(2512)] = {
    [sym__declaration_modifiers] = STATE(2503),
    [sym_attribute_specifier] = STATE(2503),
    [sym_attribute_declaration] = STATE(2503),
    [sym_ms_declspec_modifier] = STATE(2503),
    [sym_storage_class_specifier] = STATE(2503),
    [sym_type_qualifier] = STATE(2503),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2503),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2503),
    [sym_identifier] = ACTIONS(6260),
    [anon_sym_COMMA] = ACTIONS(6258),
    [anon_sym_SEMI] = ACTIONS(6258),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_LBRACE] = ACTIONS(6258),
    [anon_sym_signed] = ACTIONS(6260),
    [anon_sym_unsigned] = ACTIONS(6260),
    [anon_sym_long] = ACTIONS(6260),
    [anon_sym_short] = ACTIONS(6260),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6258),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(6260),
    [anon_sym_enum] = ACTIONS(6260),
    [anon_sym_COLON] = ACTIONS(6258),
    [anon_sym_struct] = ACTIONS(6260),
    [anon_sym_union] = ACTIONS(6260),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6260),
    [anon_sym___typeof] = ACTIONS(6260),
    [anon_sym_typeof] = ACTIONS(6260),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_BOOL] = ACTIONS(6260),
    [anon_sym_IMP] = ACTIONS(6260),
    [anon_sym_SEL] = ACTIONS(6260),
    [anon_sym_Class] = ACTIONS(6260),
    [anon_sym_id] = ACTIONS(6260),
  },
  [STATE(2513)] = {
    [sym_argument_list] = STATE(1509),
    [sym_identifier] = ACTIONS(4466),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6266),
    [anon_sym_COMMA] = ACTIONS(4468),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(4468),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(4468),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6268),
    [anon_sym_PLUS] = ACTIONS(6268),
    [anon_sym_STAR] = ACTIONS(6270),
    [anon_sym_SLASH] = ACTIONS(6272),
    [anon_sym_PERCENT] = ACTIONS(6270),
    [anon_sym_PIPE_PIPE] = ACTIONS(6274),
    [anon_sym_AMP_AMP] = ACTIONS(6276),
    [anon_sym_PIPE] = ACTIONS(6278),
    [anon_sym_CARET] = ACTIONS(6280),
    [anon_sym_AMP] = ACTIONS(6282),
    [anon_sym_EQ_EQ] = ACTIONS(6284),
    [anon_sym_BANG_EQ] = ACTIONS(6284),
    [anon_sym_GT] = ACTIONS(6286),
    [anon_sym_GT_EQ] = ACTIONS(6288),
    [anon_sym_LT_EQ] = ACTIONS(6288),
    [anon_sym_LT] = ACTIONS(6286),
    [anon_sym_LT_LT] = ACTIONS(6290),
    [anon_sym_GT_GT] = ACTIONS(6290),
    [anon_sym___extension__] = ACTIONS(4466),
    [anon_sym_extern] = ACTIONS(4466),
    [anon_sym___attribute__] = ACTIONS(4466),
    [anon_sym___attribute] = ACTIONS(4466),
    [anon_sym_noreturn] = ACTIONS(4466),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4466),
    [anon_sym_RBRACE] = ACTIONS(4468),
    [anon_sym_static] = ACTIONS(4466),
    [anon_sym_auto] = ACTIONS(4466),
    [anon_sym_register] = ACTIONS(4466),
    [anon_sym_inline] = ACTIONS(4466),
    [anon_sym___inline] = ACTIONS(4466),
    [anon_sym___inline__] = ACTIONS(4466),
    [anon_sym___forceinline] = ACTIONS(4466),
    [anon_sym_thread_local] = ACTIONS(4466),
    [anon_sym___thread] = ACTIONS(4466),
    [anon_sym_CG_EXTERN] = ACTIONS(4466),
    [anon_sym_CG_INLINE] = ACTIONS(4466),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4466),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4466),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4466),
    [anon_sym_IBOutlet] = ACTIONS(4466),
    [anon_sym_IBInspectable] = ACTIONS(4466),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4466),
    [anon_sym_NS_INLINE] = ACTIONS(4466),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4466),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4466),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4466),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4466),
    [anon_sym_const] = ACTIONS(4466),
    [anon_sym_constexpr] = ACTIONS(4466),
    [anon_sym_volatile] = ACTIONS(4466),
    [anon_sym_restrict] = ACTIONS(4466),
    [anon_sym___restrict__] = ACTIONS(4466),
    [anon_sym__Atomic] = ACTIONS(4466),
    [anon_sym__Noreturn] = ACTIONS(4466),
    [anon_sym__Nonnull] = ACTIONS(4466),
    [anon_sym_nullable] = ACTIONS(4466),
    [anon_sym__Complex] = ACTIONS(4466),
    [anon_sym__Nullable] = ACTIONS(4466),
    [anon_sym__Nullable_result] = ACTIONS(4466),
    [anon_sym__Null_unspecified] = ACTIONS(4466),
    [anon_sym___autoreleasing] = ACTIONS(4466),
    [anon_sym___block] = ACTIONS(4466),
    [anon_sym___bridge] = ACTIONS(4466),
    [anon_sym___bridge_retained] = ACTIONS(4466),
    [anon_sym___bridge_transfer] = ACTIONS(4466),
    [anon_sym___complex] = ACTIONS(4466),
    [anon_sym___const] = ACTIONS(4466),
    [anon_sym___imag] = ACTIONS(4466),
    [anon_sym___kindof] = ACTIONS(4466),
    [anon_sym___nonnull] = ACTIONS(4466),
    [anon_sym___nullable] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4466),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4466),
    [anon_sym___real] = ACTIONS(4466),
    [anon_sym___strong] = ACTIONS(4466),
    [anon_sym___unsafe_unretained] = ACTIONS(4466),
    [anon_sym___unused] = ACTIONS(4466),
    [anon_sym___weak] = ACTIONS(4466),
    [anon_sym_alignas] = ACTIONS(4466),
    [anon_sym__Alignas] = ACTIONS(4466),
    [anon_sym_QMARK] = ACTIONS(6306),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4466),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4466),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4466),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4466),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4466),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4466),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_API_AVAILABLE] = ACTIONS(4466),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_API_DEPRECATED] = ACTIONS(4466),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4466),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4466),
    [anon_sym___deprecated_msg] = ACTIONS(4466),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4466),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4466),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4466),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4466),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4466),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4466),
  },
  [STATE(2514)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3866),
    [sym__abstract_declarator] = STATE(4999),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4271),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_COMMA] = ACTIONS(6361),
    [anon_sym_RPAREN] = ACTIONS(6361),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(5316),
    [anon_sym_CARET] = ACTIONS(5318),
    [anon_sym_GT] = ACTIONS(6361),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2515)] = {
    [sym__declaration_modifiers] = STATE(2798),
    [sym_attribute_specifier] = STATE(2798),
    [sym_attribute_declaration] = STATE(2798),
    [sym_ms_declspec_modifier] = STATE(2798),
    [sym_ms_based_modifier] = STATE(6222),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__field_declarator] = STATE(4865),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(2798),
    [sym_type_qualifier] = STATE(2798),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2798),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2798),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2516)] = {
    [sym_type_qualifier] = STATE(2529),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4513),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2529),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6371),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6373),
    [anon_sym_static] = ACTIONS(6375),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2517)] = {
    [sym_type_qualifier] = STATE(2531),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4485),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2531),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6377),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6379),
    [anon_sym_static] = ACTIONS(6381),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2518)] = {
    [sym_type_qualifier] = STATE(3325),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4494),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(3325),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6385),
    [anon_sym_static] = ACTIONS(6387),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2519)] = {
    [sym__declaration_modifiers] = STATE(2804),
    [sym_attribute_specifier] = STATE(2804),
    [sym_attribute_declaration] = STATE(2804),
    [sym_ms_declspec_modifier] = STATE(2804),
    [sym_ms_based_modifier] = STATE(6222),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2521),
    [sym__field_declarator] = STATE(4851),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(2804),
    [sym_type_qualifier] = STATE(2804),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2804),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2804),
    [aux_sym_pointer_declarator_repeat1] = STATE(2521),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2520)] = {
    [sym_type_qualifier] = STATE(2544),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4532),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2544),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_RBRACK] = ACTIONS(6393),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2521)] = {
    [sym__declaration_modifiers] = STATE(2787),
    [sym_attribute_specifier] = STATE(2787),
    [sym_attribute_declaration] = STATE(2787),
    [sym_ms_declspec_modifier] = STATE(2787),
    [sym_ms_based_modifier] = STATE(6222),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__field_declarator] = STATE(4866),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(2787),
    [sym_type_qualifier] = STATE(2787),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2787),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2787),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2522)] = {
    [sym_type_qualifier] = STATE(3325),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4518),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(3325),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6397),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6399),
    [anon_sym_static] = ACTIONS(6387),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2523)] = {
    [sym__declaration_modifiers] = STATE(2787),
    [sym_attribute_specifier] = STATE(2787),
    [sym_attribute_declaration] = STATE(2787),
    [sym_ms_declspec_modifier] = STATE(2787),
    [sym_ms_based_modifier] = STATE(6222),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2515),
    [sym__field_declarator] = STATE(4866),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(2787),
    [sym_type_qualifier] = STATE(2787),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2787),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2787),
    [aux_sym_pointer_declarator_repeat1] = STATE(2515),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2524)] = {
    [sym__declaration_modifiers] = STATE(2793),
    [sym_attribute_specifier] = STATE(2793),
    [sym_attribute_declaration] = STATE(2793),
    [sym_ms_declspec_modifier] = STATE(2793),
    [sym_ms_based_modifier] = STATE(6257),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__field_declarator] = STATE(5088),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(2793),
    [sym_type_qualifier] = STATE(2793),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2793),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2793),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6401),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2525)] = {
    [sym_type_qualifier] = STATE(2518),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4537),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2518),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_RBRACK] = ACTIONS(6407),
    [anon_sym_static] = ACTIONS(6409),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2526)] = {
    [sym_type_qualifier] = STATE(2544),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4532),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2544),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6393),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2527)] = {
    [sym_type_qualifier] = STATE(3325),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4507),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(3325),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6413),
    [anon_sym_static] = ACTIONS(6387),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2528)] = {
    [sym_type_qualifier] = STATE(2527),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4495),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2527),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6341),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(6415),
    [anon_sym_RBRACK] = ACTIONS(6417),
    [anon_sym_static] = ACTIONS(6349),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2529)] = {
    [sym_type_qualifier] = STATE(3325),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4525),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(3325),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6419),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6421),
    [anon_sym_static] = ACTIONS(6387),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2530)] = {
    [sym_type_qualifier] = STATE(2538),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4540),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2538),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6423),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6425),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2531)] = {
    [sym_type_qualifier] = STATE(3325),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4519),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(3325),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6429),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6431),
    [anon_sym_static] = ACTIONS(6387),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2532)] = {
    [sym__declaration_modifiers] = STATE(2793),
    [sym_attribute_specifier] = STATE(2793),
    [sym_attribute_declaration] = STATE(2793),
    [sym_ms_declspec_modifier] = STATE(2793),
    [sym_ms_based_modifier] = STATE(6257),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2533),
    [sym__field_declarator] = STATE(5088),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(2793),
    [sym_type_qualifier] = STATE(2793),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2793),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2793),
    [aux_sym_pointer_declarator_repeat1] = STATE(2533),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6401),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2533)] = {
    [sym__declaration_modifiers] = STATE(2788),
    [sym_attribute_specifier] = STATE(2788),
    [sym_attribute_declaration] = STATE(2788),
    [sym_ms_declspec_modifier] = STATE(2788),
    [sym_ms_based_modifier] = STATE(6257),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__field_declarator] = STATE(5082),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(2788),
    [sym_type_qualifier] = STATE(2788),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2788),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2788),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6401),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2534)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6199),
    [sym__declarator] = STATE(3053),
    [sym__abstract_declarator] = STATE(4999),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3065),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3065),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3065),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4271),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3065),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_COMMA] = ACTIONS(6361),
    [anon_sym_RPAREN] = ACTIONS(6361),
    [anon_sym_LPAREN2] = ACTIONS(5417),
    [anon_sym_STAR] = ACTIONS(5419),
    [anon_sym_CARET] = ACTIONS(5421),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2535)] = {
    [sym_type_qualifier] = STATE(2522),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4508),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2522),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6433),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6435),
    [anon_sym_static] = ACTIONS(6437),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2536)] = {
    [sym_type_qualifier] = STATE(2527),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4495),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2527),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6341),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6417),
    [anon_sym_static] = ACTIONS(6349),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2537)] = {
    [sym_type_qualifier] = STATE(2527),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4495),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2527),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6341),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_RBRACK] = ACTIONS(6417),
    [anon_sym_static] = ACTIONS(6349),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2538)] = {
    [sym_type_qualifier] = STATE(3325),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4546),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(3325),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6439),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6441),
    [anon_sym_static] = ACTIONS(6387),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2539)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3866),
    [sym__abstract_declarator] = STATE(4990),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4290),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(5427),
    [anon_sym_CARET] = ACTIONS(5429),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_LBRACE] = ACTIONS(6361),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_ATautoreleasepool] = ACTIONS(6361),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2540)] = {
    [sym__declaration_modifiers] = STATE(2810),
    [sym_attribute_specifier] = STATE(2810),
    [sym_attribute_declaration] = STATE(2810),
    [sym_ms_declspec_modifier] = STATE(2810),
    [sym_ms_based_modifier] = STATE(6257),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2524),
    [sym__field_declarator] = STATE(5063),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(2810),
    [sym_type_qualifier] = STATE(2810),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2810),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2810),
    [aux_sym_pointer_declarator_repeat1] = STATE(2524),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6401),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2541)] = {
    [sym_protocol_reference_list] = STATE(2964),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_COMMA] = ACTIONS(4044),
    [anon_sym_LPAREN2] = ACTIONS(6310),
    [anon_sym_DASH] = ACTIONS(4042),
    [anon_sym_PLUS] = ACTIONS(4042),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4044),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(4042),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LT_LT] = ACTIONS(4044),
    [anon_sym_GT_GT] = ACTIONS(4044),
    [anon_sym_SEMI] = ACTIONS(4044),
    [anon_sym___extension__] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [anon_sym___attribute__] = ACTIONS(6308),
    [anon_sym___attribute] = ACTIONS(6308),
    [anon_sym_noreturn] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6310),
    [anon_sym___declspec] = ACTIONS(6308),
    [anon_sym___based] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_auto] = ACTIONS(6308),
    [anon_sym_register] = ACTIONS(6308),
    [anon_sym_inline] = ACTIONS(6308),
    [anon_sym___inline] = ACTIONS(6308),
    [anon_sym___inline__] = ACTIONS(6308),
    [anon_sym___forceinline] = ACTIONS(6308),
    [anon_sym_thread_local] = ACTIONS(6308),
    [anon_sym___thread] = ACTIONS(6308),
    [anon_sym_CG_EXTERN] = ACTIONS(6308),
    [anon_sym_CG_INLINE] = ACTIONS(6308),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6308),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6308),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6308),
    [anon_sym_IBOutlet] = ACTIONS(6308),
    [anon_sym_IBInspectable] = ACTIONS(6308),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6308),
    [anon_sym_NS_INLINE] = ACTIONS(6308),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6308),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6308),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6308),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6308),
    [anon_sym_const] = ACTIONS(6308),
    [anon_sym_constexpr] = ACTIONS(6308),
    [anon_sym_volatile] = ACTIONS(6308),
    [anon_sym_restrict] = ACTIONS(6308),
    [anon_sym___restrict__] = ACTIONS(6308),
    [anon_sym__Atomic] = ACTIONS(6308),
    [anon_sym__Noreturn] = ACTIONS(6308),
    [anon_sym__Nonnull] = ACTIONS(6308),
    [anon_sym_nullable] = ACTIONS(6308),
    [anon_sym__Complex] = ACTIONS(6308),
    [anon_sym__Nullable] = ACTIONS(6308),
    [anon_sym__Nullable_result] = ACTIONS(6308),
    [anon_sym__Null_unspecified] = ACTIONS(6308),
    [anon_sym___autoreleasing] = ACTIONS(6308),
    [anon_sym___block] = ACTIONS(6308),
    [anon_sym___bridge] = ACTIONS(6308),
    [anon_sym___bridge_retained] = ACTIONS(6308),
    [anon_sym___bridge_transfer] = ACTIONS(6308),
    [anon_sym___complex] = ACTIONS(6308),
    [anon_sym___const] = ACTIONS(6308),
    [anon_sym___imag] = ACTIONS(6308),
    [anon_sym___kindof] = ACTIONS(6308),
    [anon_sym___nonnull] = ACTIONS(6308),
    [anon_sym___nullable] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6308),
    [anon_sym___real] = ACTIONS(6308),
    [anon_sym___strong] = ACTIONS(6308),
    [anon_sym___unsafe_unretained] = ACTIONS(6308),
    [anon_sym___unused] = ACTIONS(6308),
    [anon_sym___weak] = ACTIONS(6308),
    [anon_sym_alignas] = ACTIONS(6308),
    [anon_sym__Alignas] = ACTIONS(6308),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(4044),
    [anon_sym_PLUS_PLUS] = ACTIONS(4044),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6308),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6308),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6308),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6308),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6308),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6308),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6308),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_API_AVAILABLE] = ACTIONS(6308),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_API_DEPRECATED] = ACTIONS(6308),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6308),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym___deprecated_msg] = ACTIONS(6308),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6308),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6308),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6308),
  },
  [STATE(2542)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(6443),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2220),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2220),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2220),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2220),
    [anon_sym_GT_GT] = ACTIONS(2220),
    [anon_sym_SEMI] = ACTIONS(2222),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym___attribute__] = ACTIONS(2220),
    [anon_sym___attribute] = ACTIONS(2220),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym_RBRACK] = ACTIONS(2222),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_RBRACE] = ACTIONS(2222),
    [anon_sym_EQ] = ACTIONS(2220),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_COLON] = ACTIONS(2222),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_STAR_EQ] = ACTIONS(2222),
    [anon_sym_SLASH_EQ] = ACTIONS(2222),
    [anon_sym_PERCENT_EQ] = ACTIONS(2222),
    [anon_sym_PLUS_EQ] = ACTIONS(2222),
    [anon_sym_DASH_EQ] = ACTIONS(2222),
    [anon_sym_LT_LT_EQ] = ACTIONS(2222),
    [anon_sym_GT_GT_EQ] = ACTIONS(2222),
    [anon_sym_AMP_EQ] = ACTIONS(2222),
    [anon_sym_CARET_EQ] = ACTIONS(2222),
    [anon_sym_PIPE_EQ] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(2543)] = {
    [sym_type_qualifier] = STATE(2538),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4540),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(2538),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6423),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_RBRACK] = ACTIONS(6425),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2544)] = {
    [sym_type_qualifier] = STATE(3325),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4527),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_array_declarator_repeat1] = STATE(3325),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(6455),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6457),
    [anon_sym_static] = ACTIONS(6387),
    [anon_sym_const] = ACTIONS(6351),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2545)] = {
    [sym__declaration_modifiers] = STATE(2815),
    [sym_attribute_specifier] = STATE(2815),
    [sym_attribute_declaration] = STATE(2815),
    [sym_ms_declspec_modifier] = STATE(2815),
    [sym_ms_based_modifier] = STATE(6497),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2553),
    [sym__declarator] = STATE(3831),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2815),
    [sym_type_qualifier] = STATE(2815),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2815),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2815),
    [aux_sym_pointer_declarator_repeat1] = STATE(2553),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2546)] = {
    [sym_attribute_declaration] = STATE(2480),
    [sym_compound_statement] = STATE(751),
    [sym_attributed_statement] = STATE(751),
    [sym_labeled_statement] = STATE(751),
    [sym_expression_statement] = STATE(751),
    [sym_if_statement] = STATE(751),
    [sym_switch_statement] = STATE(751),
    [sym_while_statement] = STATE(751),
    [sym_do_statement] = STATE(751),
    [sym_for_statement] = STATE(751),
    [sym_return_statement] = STATE(751),
    [sym_break_statement] = STATE(751),
    [sym_continue_statement] = STATE(751),
    [sym_goto_statement] = STATE(751),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(751),
    [sym_throw_statement] = STATE(751),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(751),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(751),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [sym_identifier] = ACTIONS(5848),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2547)] = {
    [sym_attribute_declaration] = STATE(2480),
    [sym_compound_statement] = STATE(764),
    [sym_attributed_statement] = STATE(832),
    [sym_labeled_statement] = STATE(766),
    [sym_expression_statement] = STATE(767),
    [sym_if_statement] = STATE(768),
    [sym_switch_statement] = STATE(769),
    [sym_while_statement] = STATE(770),
    [sym_do_statement] = STATE(771),
    [sym_for_statement] = STATE(772),
    [sym_return_statement] = STATE(773),
    [sym_break_statement] = STATE(774),
    [sym_continue_statement] = STATE(775),
    [sym_goto_statement] = STATE(776),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(777),
    [sym_throw_statement] = STATE(778),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(780),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(781),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2480),
    [sym_identifier] = ACTIONS(5848),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(950),
    [anon_sym_switch] = ACTIONS(952),
    [anon_sym_while] = ACTIONS(958),
    [anon_sym_do] = ACTIONS(960),
    [anon_sym_for] = ACTIONS(962),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(964),
    [anon_sym_break] = ACTIONS(966),
    [anon_sym_continue] = ACTIONS(968),
    [anon_sym_goto] = ACTIONS(970),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(986),
    [anon_sym___try] = ACTIONS(988),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(992),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2548)] = {
    [sym__declaration_modifiers] = STATE(2814),
    [sym_attribute_specifier] = STATE(2814),
    [sym_attribute_declaration] = STATE(2814),
    [sym_ms_declspec_modifier] = STATE(2814),
    [sym_ms_based_modifier] = STATE(6497),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__declarator] = STATE(3832),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2814),
    [sym_type_qualifier] = STATE(2814),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2814),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2814),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2549)] = {
    [sym_attribute_declaration] = STATE(2451),
    [sym_compound_statement] = STATE(927),
    [sym_attributed_statement] = STATE(927),
    [sym_labeled_statement] = STATE(927),
    [sym_expression_statement] = STATE(927),
    [sym_if_statement] = STATE(927),
    [sym_switch_statement] = STATE(927),
    [sym_while_statement] = STATE(927),
    [sym_do_statement] = STATE(927),
    [sym_for_statement] = STATE(927),
    [sym_return_statement] = STATE(927),
    [sym_break_statement] = STATE(927),
    [sym_continue_statement] = STATE(927),
    [sym_goto_statement] = STATE(927),
    [sym_expression] = STATE(4419),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6212),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(927),
    [sym_throw_statement] = STATE(927),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(927),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(927),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2451),
    [sym_identifier] = ACTIONS(6034),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2150),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(69),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_while] = ACTIONS(77),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2152),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2154),
    [anon_sym___try] = ACTIONS(2156),
    [anon_sym_ATthrow] = ACTIONS(2158),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2160),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2550)] = {
    [sym__declaration_modifiers] = STATE(2828),
    [sym_attribute_specifier] = STATE(2828),
    [sym_attribute_declaration] = STATE(2828),
    [sym_ms_declspec_modifier] = STATE(2828),
    [sym_ms_based_modifier] = STATE(6101),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__declarator] = STATE(2753),
    [sym_parenthesized_declarator] = STATE(2778),
    [sym_pointer_declarator] = STATE(2778),
    [sym_function_declarator] = STATE(2778),
    [sym_array_declarator] = STATE(2778),
    [sym_storage_class_specifier] = STATE(2828),
    [sym_type_qualifier] = STATE(2828),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2828),
    [sym_block_pointer_declarator] = STATE(2778),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2828),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(6463),
    [anon_sym_LPAREN2] = ACTIONS(6465),
    [anon_sym_STAR] = ACTIONS(6467),
    [anon_sym_CARET] = ACTIONS(6469),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2551)] = {
    [sym__declaration_modifiers] = STATE(2828),
    [sym_attribute_specifier] = STATE(2828),
    [sym_attribute_declaration] = STATE(2828),
    [sym_ms_declspec_modifier] = STATE(2828),
    [sym_ms_based_modifier] = STATE(6101),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2556),
    [sym__declarator] = STATE(2753),
    [sym_parenthesized_declarator] = STATE(2778),
    [sym_pointer_declarator] = STATE(2778),
    [sym_function_declarator] = STATE(2778),
    [sym_array_declarator] = STATE(2778),
    [sym_storage_class_specifier] = STATE(2828),
    [sym_type_qualifier] = STATE(2828),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2828),
    [sym_block_pointer_declarator] = STATE(2778),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2828),
    [aux_sym_pointer_declarator_repeat1] = STATE(2556),
    [sym_identifier] = ACTIONS(6463),
    [anon_sym_LPAREN2] = ACTIONS(6465),
    [anon_sym_STAR] = ACTIONS(6467),
    [anon_sym_CARET] = ACTIONS(6469),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2552)] = {
    [sym_attribute_specifier] = STATE(2678),
    [sym_gnu_asm_expression] = STATE(2682),
    [aux_sym_function_declarator_repeat1] = STATE(2678),
    [sym_identifier] = ACTIONS(6471),
    [anon_sym_COMMA] = ACTIONS(6473),
    [anon_sym_LPAREN2] = ACTIONS(6473),
    [anon_sym_SEMI] = ACTIONS(6473),
    [anon_sym___extension__] = ACTIONS(6471),
    [anon_sym_extern] = ACTIONS(6471),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6471),
    [anon_sym_LBRACK] = ACTIONS(6473),
    [anon_sym___declspec] = ACTIONS(6471),
    [anon_sym_LBRACE] = ACTIONS(6473),
    [anon_sym_signed] = ACTIONS(6471),
    [anon_sym_unsigned] = ACTIONS(6471),
    [anon_sym_long] = ACTIONS(6471),
    [anon_sym_short] = ACTIONS(6471),
    [anon_sym_static] = ACTIONS(6471),
    [anon_sym_EQ] = ACTIONS(6473),
    [anon_sym_ATautoreleasepool] = ACTIONS(6473),
    [anon_sym_auto] = ACTIONS(6471),
    [anon_sym_register] = ACTIONS(6471),
    [anon_sym_inline] = ACTIONS(6471),
    [anon_sym___inline] = ACTIONS(6471),
    [anon_sym___inline__] = ACTIONS(6471),
    [anon_sym___forceinline] = ACTIONS(6471),
    [anon_sym_thread_local] = ACTIONS(6471),
    [anon_sym___thread] = ACTIONS(6471),
    [anon_sym_CG_EXTERN] = ACTIONS(6471),
    [anon_sym_CG_INLINE] = ACTIONS(6471),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6471),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6471),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6471),
    [anon_sym_IBOutlet] = ACTIONS(6471),
    [anon_sym_IBInspectable] = ACTIONS(6471),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6471),
    [anon_sym_NS_INLINE] = ACTIONS(6471),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6471),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6471),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6471),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6471),
    [anon_sym_const] = ACTIONS(6471),
    [anon_sym_constexpr] = ACTIONS(6471),
    [anon_sym_volatile] = ACTIONS(6471),
    [anon_sym_restrict] = ACTIONS(6471),
    [anon_sym___restrict__] = ACTIONS(6471),
    [anon_sym__Atomic] = ACTIONS(6471),
    [anon_sym__Noreturn] = ACTIONS(6471),
    [anon_sym__Nonnull] = ACTIONS(6471),
    [anon_sym_nullable] = ACTIONS(6471),
    [anon_sym__Complex] = ACTIONS(6471),
    [anon_sym__Nullable] = ACTIONS(6471),
    [anon_sym__Nullable_result] = ACTIONS(6471),
    [anon_sym__Null_unspecified] = ACTIONS(6471),
    [anon_sym___autoreleasing] = ACTIONS(6471),
    [anon_sym___block] = ACTIONS(6471),
    [anon_sym___bridge] = ACTIONS(6471),
    [anon_sym___bridge_retained] = ACTIONS(6471),
    [anon_sym___bridge_transfer] = ACTIONS(6471),
    [anon_sym___complex] = ACTIONS(6471),
    [anon_sym___const] = ACTIONS(6471),
    [anon_sym___imag] = ACTIONS(6471),
    [anon_sym___kindof] = ACTIONS(6471),
    [anon_sym___nonnull] = ACTIONS(6471),
    [anon_sym___nullable] = ACTIONS(6471),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6471),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6471),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6471),
    [anon_sym___real] = ACTIONS(6471),
    [anon_sym___strong] = ACTIONS(6471),
    [anon_sym___unsafe_unretained] = ACTIONS(6471),
    [anon_sym___unused] = ACTIONS(6471),
    [anon_sym___weak] = ACTIONS(6471),
    [anon_sym_alignas] = ACTIONS(6471),
    [anon_sym__Alignas] = ACTIONS(6471),
    [sym_primitive_type] = ACTIONS(6471),
    [anon_sym_enum] = ACTIONS(6471),
    [anon_sym_struct] = ACTIONS(6471),
    [anon_sym_union] = ACTIONS(6471),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6471),
    [anon_sym___typeof] = ACTIONS(6471),
    [anon_sym_typeof] = ACTIONS(6471),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6471),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6471),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6471),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6471),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6471),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6471),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6471),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6471),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6471),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6471),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6471),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6471),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6471),
    [anon_sym_API_AVAILABLE] = ACTIONS(6471),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6471),
    [anon_sym_API_DEPRECATED] = ACTIONS(6471),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6471),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6471),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6471),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6471),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6471),
    [anon_sym___deprecated_msg] = ACTIONS(6471),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6471),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6471),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6471),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6471),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6471),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6471),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6471),
    [anon_sym_BOOL] = ACTIONS(6471),
    [anon_sym_IMP] = ACTIONS(6471),
    [anon_sym_SEL] = ACTIONS(6471),
    [anon_sym_Class] = ACTIONS(6471),
    [anon_sym_id] = ACTIONS(6471),
  },
  [STATE(2553)] = {
    [sym__declaration_modifiers] = STATE(2821),
    [sym_attribute_specifier] = STATE(2821),
    [sym_attribute_declaration] = STATE(2821),
    [sym_ms_declspec_modifier] = STATE(2821),
    [sym_ms_based_modifier] = STATE(6497),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__declarator] = STATE(3866),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2821),
    [sym_type_qualifier] = STATE(2821),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2821),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2821),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2554)] = {
    [sym__declaration_modifiers] = STATE(2821),
    [sym_attribute_specifier] = STATE(2821),
    [sym_attribute_declaration] = STATE(2821),
    [sym_ms_declspec_modifier] = STATE(2821),
    [sym_ms_based_modifier] = STATE(6497),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2548),
    [sym__declarator] = STATE(3866),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2821),
    [sym_type_qualifier] = STATE(2821),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2821),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2821),
    [aux_sym_pointer_declarator_repeat1] = STATE(2548),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2555)] = {
    [sym__declaration_modifiers] = STATE(2813),
    [sym_attribute_specifier] = STATE(2813),
    [sym_attribute_declaration] = STATE(2813),
    [sym_ms_declspec_modifier] = STATE(2813),
    [sym_ms_based_modifier] = STATE(6199),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2572),
    [sym__declarator] = STATE(3049),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(2813),
    [sym_type_qualifier] = STATE(2813),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2813),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2813),
    [aux_sym_pointer_declarator_repeat1] = STATE(2572),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2556)] = {
    [sym__declaration_modifiers] = STATE(2818),
    [sym_attribute_specifier] = STATE(2818),
    [sym_attribute_declaration] = STATE(2818),
    [sym_ms_declspec_modifier] = STATE(2818),
    [sym_ms_based_modifier] = STATE(6101),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__declarator] = STATE(2687),
    [sym_parenthesized_declarator] = STATE(2778),
    [sym_pointer_declarator] = STATE(2778),
    [sym_function_declarator] = STATE(2778),
    [sym_array_declarator] = STATE(2778),
    [sym_storage_class_specifier] = STATE(2818),
    [sym_type_qualifier] = STATE(2818),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2818),
    [sym_block_pointer_declarator] = STATE(2778),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2818),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(6463),
    [anon_sym_LPAREN2] = ACTIONS(6465),
    [anon_sym_STAR] = ACTIONS(6467),
    [anon_sym_CARET] = ACTIONS(6469),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2557)] = {
    [sym_type_qualifier] = STATE(2623),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4489),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(2623),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6481),
    [anon_sym_const] = ACTIONS(6483),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2558)] = {
    [sym_attribute_declaration] = STATE(2451),
    [sym_compound_statement] = STATE(930),
    [sym_attributed_statement] = STATE(931),
    [sym_labeled_statement] = STATE(932),
    [sym_expression_statement] = STATE(933),
    [sym_if_statement] = STATE(934),
    [sym_switch_statement] = STATE(935),
    [sym_while_statement] = STATE(937),
    [sym_do_statement] = STATE(938),
    [sym_for_statement] = STATE(939),
    [sym_return_statement] = STATE(941),
    [sym_break_statement] = STATE(976),
    [sym_continue_statement] = STATE(952),
    [sym_goto_statement] = STATE(953),
    [sym_expression] = STATE(4419),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6212),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(954),
    [sym_throw_statement] = STATE(958),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(959),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(960),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2451),
    [sym_identifier] = ACTIONS(6034),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2150),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(69),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_while] = ACTIONS(77),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2152),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2154),
    [anon_sym___try] = ACTIONS(2156),
    [anon_sym_ATthrow] = ACTIONS(2158),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2160),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2559)] = {
    [sym_type_qualifier] = STATE(2627),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4536),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(2627),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(6485),
    [anon_sym_const] = ACTIONS(6483),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2560)] = {
    [sym_attribute_declaration] = STATE(2456),
    [sym_compound_statement] = STATE(271),
    [sym_attributed_statement] = STATE(271),
    [sym_labeled_statement] = STATE(271),
    [sym_expression_statement] = STATE(271),
    [sym_if_statement] = STATE(271),
    [sym_switch_statement] = STATE(271),
    [sym_while_statement] = STATE(271),
    [sym_do_statement] = STATE(271),
    [sym_for_statement] = STATE(271),
    [sym_return_statement] = STATE(271),
    [sym_break_statement] = STATE(271),
    [sym_continue_statement] = STATE(271),
    [sym_goto_statement] = STATE(271),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(271),
    [sym_throw_statement] = STATE(271),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(271),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(271),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [sym_identifier] = ACTIONS(6036),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2561)] = {
    [sym_attribute_declaration] = STATE(2456),
    [sym_compound_statement] = STATE(245),
    [sym_attributed_statement] = STATE(252),
    [sym_labeled_statement] = STATE(203),
    [sym_expression_statement] = STATE(200),
    [sym_if_statement] = STATE(205),
    [sym_switch_statement] = STATE(206),
    [sym_while_statement] = STATE(207),
    [sym_do_statement] = STATE(209),
    [sym_for_statement] = STATE(210),
    [sym_return_statement] = STATE(217),
    [sym_break_statement] = STATE(227),
    [sym_continue_statement] = STATE(199),
    [sym_goto_statement] = STATE(229),
    [sym_expression] = STATE(4468),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6175),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(231),
    [sym_throw_statement] = STATE(233),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(235),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(241),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2456),
    [sym_identifier] = ACTIONS(6036),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(171),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(189),
    [anon_sym_switch] = ACTIONS(191),
    [anon_sym_while] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(199),
    [anon_sym_for] = ACTIONS(201),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(205),
    [anon_sym_break] = ACTIONS(207),
    [anon_sym_continue] = ACTIONS(209),
    [anon_sym_goto] = ACTIONS(211),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(213),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(235),
    [anon_sym___try] = ACTIONS(237),
    [anon_sym_ATthrow] = ACTIONS(239),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(241),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2562)] = {
    [sym_protocol_reference_list] = STATE(2964),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_LPAREN2] = ACTIONS(6310),
    [anon_sym_DASH] = ACTIONS(4042),
    [anon_sym_PLUS] = ACTIONS(4042),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4044),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(4042),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LT_LT] = ACTIONS(4044),
    [anon_sym_GT_GT] = ACTIONS(4044),
    [anon_sym_SEMI] = ACTIONS(6310),
    [anon_sym___extension__] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [anon_sym___attribute__] = ACTIONS(6308),
    [anon_sym___attribute] = ACTIONS(6308),
    [anon_sym_noreturn] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6310),
    [anon_sym___declspec] = ACTIONS(6308),
    [anon_sym___based] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_auto] = ACTIONS(6308),
    [anon_sym_register] = ACTIONS(6308),
    [anon_sym_inline] = ACTIONS(6308),
    [anon_sym___inline] = ACTIONS(6308),
    [anon_sym___inline__] = ACTIONS(6308),
    [anon_sym___forceinline] = ACTIONS(6308),
    [anon_sym_thread_local] = ACTIONS(6308),
    [anon_sym___thread] = ACTIONS(6308),
    [anon_sym_CG_EXTERN] = ACTIONS(6308),
    [anon_sym_CG_INLINE] = ACTIONS(6308),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6308),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6308),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6308),
    [anon_sym_IBOutlet] = ACTIONS(6308),
    [anon_sym_IBInspectable] = ACTIONS(6308),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6308),
    [anon_sym_NS_INLINE] = ACTIONS(6308),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6308),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6308),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6308),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6308),
    [anon_sym_const] = ACTIONS(6308),
    [anon_sym_constexpr] = ACTIONS(6308),
    [anon_sym_volatile] = ACTIONS(6308),
    [anon_sym_restrict] = ACTIONS(6308),
    [anon_sym___restrict__] = ACTIONS(6308),
    [anon_sym__Atomic] = ACTIONS(6308),
    [anon_sym__Noreturn] = ACTIONS(6308),
    [anon_sym__Nonnull] = ACTIONS(6308),
    [anon_sym_nullable] = ACTIONS(6308),
    [anon_sym__Complex] = ACTIONS(6308),
    [anon_sym__Nullable] = ACTIONS(6308),
    [anon_sym__Nullable_result] = ACTIONS(6308),
    [anon_sym__Null_unspecified] = ACTIONS(6308),
    [anon_sym___autoreleasing] = ACTIONS(6308),
    [anon_sym___block] = ACTIONS(6308),
    [anon_sym___bridge] = ACTIONS(6308),
    [anon_sym___bridge_retained] = ACTIONS(6308),
    [anon_sym___bridge_transfer] = ACTIONS(6308),
    [anon_sym___complex] = ACTIONS(6308),
    [anon_sym___const] = ACTIONS(6308),
    [anon_sym___imag] = ACTIONS(6308),
    [anon_sym___kindof] = ACTIONS(6308),
    [anon_sym___nonnull] = ACTIONS(6308),
    [anon_sym___nullable] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6308),
    [anon_sym___real] = ACTIONS(6308),
    [anon_sym___strong] = ACTIONS(6308),
    [anon_sym___unsafe_unretained] = ACTIONS(6308),
    [anon_sym___unused] = ACTIONS(6308),
    [anon_sym___weak] = ACTIONS(6308),
    [anon_sym_alignas] = ACTIONS(6308),
    [anon_sym__Alignas] = ACTIONS(6308),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(4044),
    [anon_sym_PLUS_PLUS] = ACTIONS(4044),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6308),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6308),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6308),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6308),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6308),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6308),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6308),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_API_AVAILABLE] = ACTIONS(6308),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_API_DEPRECATED] = ACTIONS(6308),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6308),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym___deprecated_msg] = ACTIONS(6308),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6308),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6308),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6308),
  },
  [STATE(2563)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(927),
    [sym_attributed_statement] = STATE(927),
    [sym_labeled_statement] = STATE(927),
    [sym_expression_statement] = STATE(927),
    [sym_if_statement] = STATE(927),
    [sym_switch_statement] = STATE(927),
    [sym_while_statement] = STATE(927),
    [sym_do_statement] = STATE(927),
    [sym_for_statement] = STATE(927),
    [sym_return_statement] = STATE(927),
    [sym_break_statement] = STATE(927),
    [sym_continue_statement] = STATE(927),
    [sym_goto_statement] = STATE(927),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(927),
    [sym_throw_statement] = STATE(927),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(927),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(927),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2564)] = {
    [sym_attribute_declaration] = STATE(2429),
    [sym_compound_statement] = STATE(930),
    [sym_attributed_statement] = STATE(931),
    [sym_labeled_statement] = STATE(932),
    [sym_expression_statement] = STATE(933),
    [sym_if_statement] = STATE(934),
    [sym_switch_statement] = STATE(935),
    [sym_while_statement] = STATE(937),
    [sym_do_statement] = STATE(938),
    [sym_for_statement] = STATE(939),
    [sym_return_statement] = STATE(941),
    [sym_break_statement] = STATE(976),
    [sym_continue_statement] = STATE(952),
    [sym_goto_statement] = STATE(953),
    [sym_expression] = STATE(4418),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6117),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(954),
    [sym_throw_statement] = STATE(958),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(959),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(960),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2429),
    [sym_identifier] = ACTIONS(5702),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2164),
    [anon_sym_switch] = ACTIONS(71),
    [anon_sym_while] = ACTIONS(2170),
    [anon_sym_do] = ACTIONS(79),
    [anon_sym_for] = ACTIONS(2172),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(85),
    [anon_sym_break] = ACTIONS(87),
    [anon_sym_continue] = ACTIONS(89),
    [anon_sym_goto] = ACTIONS(91),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(972),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2174),
    [anon_sym___try] = ACTIONS(2176),
    [anon_sym_ATthrow] = ACTIONS(990),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2178),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2565)] = {
    [sym__declaration_modifiers] = STATE(2841),
    [sym_attribute_specifier] = STATE(2841),
    [sym_attribute_declaration] = STATE(2841),
    [sym_ms_declspec_modifier] = STATE(2841),
    [sym_ms_based_modifier] = STATE(6656),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2583),
    [sym__declarator] = STATE(3209),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(2841),
    [sym_type_qualifier] = STATE(2841),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2841),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2841),
    [aux_sym_pointer_declarator_repeat1] = STATE(2583),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6487),
    [anon_sym_CARET] = ACTIONS(6489),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2566)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5012),
    [sym__abstract_declarator] = STATE(4999),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4271),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_RPAREN] = ACTIONS(6361),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(4086),
    [anon_sym_CARET] = ACTIONS(4088),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(4090),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2567)] = {
    [sym_attribute_declaration] = STATE(2472),
    [sym_compound_statement] = STATE(792),
    [sym_attributed_statement] = STATE(792),
    [sym_labeled_statement] = STATE(792),
    [sym_expression_statement] = STATE(792),
    [sym_if_statement] = STATE(792),
    [sym_switch_statement] = STATE(792),
    [sym_while_statement] = STATE(792),
    [sym_do_statement] = STATE(792),
    [sym_for_statement] = STATE(792),
    [sym_return_statement] = STATE(792),
    [sym_break_statement] = STATE(792),
    [sym_continue_statement] = STATE(792),
    [sym_goto_statement] = STATE(792),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(792),
    [sym_throw_statement] = STATE(792),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(792),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(792),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [sym_identifier] = ACTIONS(5852),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2568)] = {
    [sym_attribute_declaration] = STATE(2472),
    [sym_compound_statement] = STATE(794),
    [sym_attributed_statement] = STATE(795),
    [sym_labeled_statement] = STATE(796),
    [sym_expression_statement] = STATE(797),
    [sym_if_statement] = STATE(798),
    [sym_switch_statement] = STATE(799),
    [sym_while_statement] = STATE(800),
    [sym_do_statement] = STATE(801),
    [sym_for_statement] = STATE(802),
    [sym_return_statement] = STATE(803),
    [sym_break_statement] = STATE(804),
    [sym_continue_statement] = STATE(805),
    [sym_goto_statement] = STATE(806),
    [sym_expression] = STATE(4446),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5915),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(807),
    [sym_throw_statement] = STATE(808),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(809),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(810),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2472),
    [sym_identifier] = ACTIONS(5852),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_ATautoreleasepool] = ACTIONS(1936),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_switch] = ACTIONS(1940),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1944),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(1948),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1952),
    [anon_sym_goto] = ACTIONS(1954),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(1956),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(1958),
    [anon_sym___try] = ACTIONS(1960),
    [anon_sym_ATthrow] = ACTIONS(1962),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(1964),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2569)] = {
    [sym_type_qualifier] = STATE(2623),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4489),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(2623),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(6415),
    [anon_sym_RBRACK] = ACTIONS(6481),
    [anon_sym_const] = ACTIONS(6483),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2570)] = {
    [sym_attribute_declaration] = STATE(2436),
    [sym_compound_statement] = STATE(370),
    [sym_attributed_statement] = STATE(370),
    [sym_labeled_statement] = STATE(370),
    [sym_expression_statement] = STATE(370),
    [sym_if_statement] = STATE(370),
    [sym_switch_statement] = STATE(370),
    [sym_while_statement] = STATE(370),
    [sym_do_statement] = STATE(370),
    [sym_for_statement] = STATE(370),
    [sym_return_statement] = STATE(370),
    [sym_break_statement] = STATE(370),
    [sym_continue_statement] = STATE(370),
    [sym_goto_statement] = STATE(370),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(370),
    [sym_throw_statement] = STATE(370),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(370),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(370),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [sym_identifier] = ACTIONS(5706),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2571)] = {
    [sym_attribute_declaration] = STATE(2436),
    [sym_compound_statement] = STATE(374),
    [sym_attributed_statement] = STATE(375),
    [sym_labeled_statement] = STATE(376),
    [sym_expression_statement] = STATE(377),
    [sym_if_statement] = STATE(378),
    [sym_switch_statement] = STATE(379),
    [sym_while_statement] = STATE(398),
    [sym_do_statement] = STATE(399),
    [sym_for_statement] = STATE(400),
    [sym_return_statement] = STATE(401),
    [sym_break_statement] = STATE(402),
    [sym_continue_statement] = STATE(403),
    [sym_goto_statement] = STATE(361),
    [sym_expression] = STATE(4464),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6407),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(441),
    [sym_throw_statement] = STATE(424),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(432),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(433),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2436),
    [sym_identifier] = ACTIONS(5706),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(293),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(307),
    [anon_sym_switch] = ACTIONS(309),
    [anon_sym_while] = ACTIONS(315),
    [anon_sym_do] = ACTIONS(317),
    [anon_sym_for] = ACTIONS(319),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(321),
    [anon_sym_break] = ACTIONS(323),
    [anon_sym_continue] = ACTIONS(325),
    [anon_sym_goto] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(329),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(347),
    [anon_sym___try] = ACTIONS(349),
    [anon_sym_ATthrow] = ACTIONS(351),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(353),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2572)] = {
    [sym__declaration_modifiers] = STATE(2820),
    [sym_attribute_specifier] = STATE(2820),
    [sym_attribute_declaration] = STATE(2820),
    [sym_ms_declspec_modifier] = STATE(2820),
    [sym_ms_based_modifier] = STATE(6199),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__declarator] = STATE(3053),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(2820),
    [sym_type_qualifier] = STATE(2820),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2820),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2820),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2573)] = {
    [sym__declaration_modifiers] = STATE(2820),
    [sym_attribute_specifier] = STATE(2820),
    [sym_attribute_declaration] = STATE(2820),
    [sym_ms_declspec_modifier] = STATE(2820),
    [sym_ms_based_modifier] = STATE(6199),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2578),
    [sym__declarator] = STATE(3053),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(2820),
    [sym_type_qualifier] = STATE(2820),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2820),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2820),
    [aux_sym_pointer_declarator_repeat1] = STATE(2578),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2574)] = {
    [sym_attribute_declaration] = STATE(2438),
    [sym_compound_statement] = STATE(547),
    [sym_attributed_statement] = STATE(547),
    [sym_labeled_statement] = STATE(547),
    [sym_expression_statement] = STATE(547),
    [sym_if_statement] = STATE(547),
    [sym_switch_statement] = STATE(547),
    [sym_while_statement] = STATE(547),
    [sym_do_statement] = STATE(547),
    [sym_for_statement] = STATE(547),
    [sym_return_statement] = STATE(547),
    [sym_break_statement] = STATE(547),
    [sym_continue_statement] = STATE(547),
    [sym_goto_statement] = STATE(547),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(547),
    [sym_throw_statement] = STATE(547),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(547),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(547),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [sym_identifier] = ACTIONS(5846),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2575)] = {
    [sym_attribute_declaration] = STATE(2438),
    [sym_compound_statement] = STATE(550),
    [sym_attributed_statement] = STATE(551),
    [sym_labeled_statement] = STATE(552),
    [sym_expression_statement] = STATE(553),
    [sym_if_statement] = STATE(554),
    [sym_switch_statement] = STATE(555),
    [sym_while_statement] = STATE(556),
    [sym_do_statement] = STATE(557),
    [sym_for_statement] = STATE(558),
    [sym_return_statement] = STATE(559),
    [sym_break_statement] = STATE(560),
    [sym_continue_statement] = STATE(561),
    [sym_goto_statement] = STATE(562),
    [sym_expression] = STATE(4425),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6397),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(563),
    [sym_throw_statement] = STATE(564),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(565),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(527),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2438),
    [sym_identifier] = ACTIONS(5846),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(860),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(874),
    [anon_sym_switch] = ACTIONS(876),
    [anon_sym_while] = ACTIONS(882),
    [anon_sym_do] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(888),
    [anon_sym_break] = ACTIONS(890),
    [anon_sym_continue] = ACTIONS(892),
    [anon_sym_goto] = ACTIONS(894),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(896),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(914),
    [anon_sym___try] = ACTIONS(916),
    [anon_sym_ATthrow] = ACTIONS(918),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(920),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2576)] = {
    [sym_attribute_declaration] = STATE(2446),
    [sym_compound_statement] = STATE(1157),
    [sym_attributed_statement] = STATE(1157),
    [sym_labeled_statement] = STATE(1157),
    [sym_expression_statement] = STATE(1157),
    [sym_if_statement] = STATE(1157),
    [sym_switch_statement] = STATE(1157),
    [sym_while_statement] = STATE(1157),
    [sym_do_statement] = STATE(1157),
    [sym_for_statement] = STATE(1157),
    [sym_return_statement] = STATE(1157),
    [sym_break_statement] = STATE(1157),
    [sym_continue_statement] = STATE(1157),
    [sym_goto_statement] = STATE(1157),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(1157),
    [sym_throw_statement] = STATE(1157),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(1157),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(1157),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [sym_identifier] = ACTIONS(5850),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2577)] = {
    [sym_attribute_declaration] = STATE(2446),
    [sym_compound_statement] = STATE(1161),
    [sym_attributed_statement] = STATE(1162),
    [sym_labeled_statement] = STATE(1091),
    [sym_expression_statement] = STATE(1090),
    [sym_if_statement] = STATE(1137),
    [sym_switch_statement] = STATE(1124),
    [sym_while_statement] = STATE(1125),
    [sym_do_statement] = STATE(1126),
    [sym_for_statement] = STATE(1127),
    [sym_return_statement] = STATE(1128),
    [sym_break_statement] = STATE(1129),
    [sym_continue_statement] = STATE(1130),
    [sym_goto_statement] = STATE(1136),
    [sym_expression] = STATE(4447),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6206),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_try_statement] = STATE(1138),
    [sym_throw_statement] = STATE(1139),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_ms_asm_block] = STATE(1141),
    [sym_encode_expression] = STATE(1641),
    [sym_synchronized_statement] = STATE(1145),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_attributed_field_declarator_repeat1] = STATE(2446),
    [sym_identifier] = ACTIONS(5850),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(2182),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(5704),
    [anon_sym_LBRACE] = ACTIONS(2188),
    [anon_sym_ATautoreleasepool] = ACTIONS(2190),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_if] = ACTIONS(2192),
    [anon_sym_switch] = ACTIONS(2194),
    [anon_sym_while] = ACTIONS(2196),
    [anon_sym_do] = ACTIONS(2198),
    [anon_sym_for] = ACTIONS(2200),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_return] = ACTIONS(2202),
    [anon_sym_break] = ACTIONS(2204),
    [anon_sym_continue] = ACTIONS(2206),
    [anon_sym_goto] = ACTIONS(2208),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(2210),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2212),
    [anon_sym___try] = ACTIONS(2214),
    [anon_sym_ATthrow] = ACTIONS(2216),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_ATsynchronized] = ACTIONS(2218),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2578)] = {
    [sym__declaration_modifiers] = STATE(2837),
    [sym_attribute_specifier] = STATE(2837),
    [sym_attribute_declaration] = STATE(2837),
    [sym_ms_declspec_modifier] = STATE(2837),
    [sym_ms_based_modifier] = STATE(6199),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__declarator] = STATE(3068),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(2837),
    [sym_type_qualifier] = STATE(2837),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2837),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2837),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2579)] = {
    [sym__declaration_modifiers] = STATE(2827),
    [sym_attribute_specifier] = STATE(2827),
    [sym_attribute_declaration] = STATE(2827),
    [sym_ms_declspec_modifier] = STATE(2827),
    [sym_ms_based_modifier] = STATE(6198),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__declarator] = STATE(5012),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2827),
    [sym_type_qualifier] = STATE(2827),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2827),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2827),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2580)] = {
    [sym__declaration_modifiers] = STATE(2827),
    [sym_attribute_specifier] = STATE(2827),
    [sym_attribute_declaration] = STATE(2827),
    [sym_ms_declspec_modifier] = STATE(2827),
    [sym_ms_based_modifier] = STATE(6198),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2581),
    [sym__declarator] = STATE(5012),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2827),
    [sym_type_qualifier] = STATE(2827),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2827),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2827),
    [aux_sym_pointer_declarator_repeat1] = STATE(2581),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2581)] = {
    [sym__declaration_modifiers] = STATE(2829),
    [sym_attribute_specifier] = STATE(2829),
    [sym_attribute_declaration] = STATE(2829),
    [sym_ms_declspec_modifier] = STATE(2829),
    [sym_ms_based_modifier] = STATE(6198),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__declarator] = STATE(5058),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2829),
    [sym_type_qualifier] = STATE(2829),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2829),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2829),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2582)] = {
    [sym__declaration_modifiers] = STATE(2824),
    [sym_attribute_specifier] = STATE(2824),
    [sym_attribute_declaration] = STATE(2824),
    [sym_ms_declspec_modifier] = STATE(2824),
    [sym_ms_based_modifier] = STATE(6198),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2579),
    [sym__declarator] = STATE(5024),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2824),
    [sym_type_qualifier] = STATE(2824),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2824),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2824),
    [aux_sym_pointer_declarator_repeat1] = STATE(2579),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2583)] = {
    [sym__declaration_modifiers] = STATE(2843),
    [sym_attribute_specifier] = STATE(2843),
    [sym_attribute_declaration] = STATE(2843),
    [sym_ms_declspec_modifier] = STATE(2843),
    [sym_ms_based_modifier] = STATE(6656),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__declarator] = STATE(3215),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(2843),
    [sym_type_qualifier] = STATE(2843),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2843),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2843),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6487),
    [anon_sym_CARET] = ACTIONS(6489),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2584)] = {
    [sym__declaration_modifiers] = STATE(2843),
    [sym_attribute_specifier] = STATE(2843),
    [sym_attribute_declaration] = STATE(2843),
    [sym_ms_declspec_modifier] = STATE(2843),
    [sym_ms_based_modifier] = STATE(6656),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2585),
    [sym__declarator] = STATE(3215),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(2843),
    [sym_type_qualifier] = STATE(2843),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2843),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2843),
    [aux_sym_pointer_declarator_repeat1] = STATE(2585),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6487),
    [anon_sym_CARET] = ACTIONS(6489),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2585)] = {
    [sym__declaration_modifiers] = STATE(2844),
    [sym_attribute_specifier] = STATE(2844),
    [sym_attribute_declaration] = STATE(2844),
    [sym_ms_declspec_modifier] = STATE(2844),
    [sym_ms_based_modifier] = STATE(6656),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [sym__declarator] = STATE(3211),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(2844),
    [sym_type_qualifier] = STATE(2844),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2844),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2844),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6487),
    [anon_sym_CARET] = ACTIONS(6489),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2586)] = {
    [sym__declaration_modifiers] = STATE(2846),
    [sym_attribute_specifier] = STATE(2846),
    [sym_attribute_declaration] = STATE(2846),
    [sym_ms_declspec_modifier] = STATE(2846),
    [sym_ms_based_modifier] = STATE(6101),
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2550),
    [sym__declarator] = STATE(2748),
    [sym_parenthesized_declarator] = STATE(2778),
    [sym_pointer_declarator] = STATE(2778),
    [sym_function_declarator] = STATE(2778),
    [sym_array_declarator] = STATE(2778),
    [sym_storage_class_specifier] = STATE(2846),
    [sym_type_qualifier] = STATE(2846),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2846),
    [sym_block_pointer_declarator] = STATE(2778),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2846),
    [aux_sym_pointer_declarator_repeat1] = STATE(2550),
    [sym_identifier] = ACTIONS(6463),
    [anon_sym_LPAREN2] = ACTIONS(6465),
    [anon_sym_STAR] = ACTIONS(6467),
    [anon_sym_CARET] = ACTIONS(6469),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [sym_ms_restrict_modifier] = ACTIONS(5320),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(5320),
    [sym_ms_signed_ptr_modifier] = ACTIONS(5320),
    [anon_sym__unaligned] = ACTIONS(5322),
    [anon_sym___unaligned] = ACTIONS(5322),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2587)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6491),
    [anon_sym_COMMA] = ACTIONS(4783),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(6499),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4783),
    [anon_sym___extension__] = ACTIONS(4783),
    [anon_sym_extern] = ACTIONS(4783),
    [anon_sym___attribute__] = ACTIONS(4783),
    [anon_sym___attribute] = ACTIONS(4779),
    [anon_sym_noreturn] = ACTIONS(4783),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4783),
    [anon_sym_static] = ACTIONS(4783),
    [anon_sym_auto] = ACTIONS(4783),
    [anon_sym_register] = ACTIONS(4783),
    [anon_sym_inline] = ACTIONS(4783),
    [anon_sym___inline] = ACTIONS(4779),
    [anon_sym___inline__] = ACTIONS(4783),
    [anon_sym___forceinline] = ACTIONS(4783),
    [anon_sym_thread_local] = ACTIONS(4783),
    [anon_sym___thread] = ACTIONS(4783),
    [anon_sym_CG_EXTERN] = ACTIONS(4783),
    [anon_sym_CG_INLINE] = ACTIONS(4783),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4783),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4783),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4783),
    [anon_sym_IBOutlet] = ACTIONS(4783),
    [anon_sym_IBInspectable] = ACTIONS(4783),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4783),
    [anon_sym_NS_INLINE] = ACTIONS(4783),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4783),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4783),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4783),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4783),
    [anon_sym_const] = ACTIONS(4779),
    [anon_sym_constexpr] = ACTIONS(4783),
    [anon_sym_volatile] = ACTIONS(4783),
    [anon_sym_restrict] = ACTIONS(4783),
    [anon_sym___restrict__] = ACTIONS(4783),
    [anon_sym__Atomic] = ACTIONS(4783),
    [anon_sym__Noreturn] = ACTIONS(4783),
    [anon_sym__Nonnull] = ACTIONS(4783),
    [anon_sym_nullable] = ACTIONS(4783),
    [anon_sym__Complex] = ACTIONS(4783),
    [anon_sym__Nullable] = ACTIONS(4779),
    [anon_sym__Nullable_result] = ACTIONS(4783),
    [anon_sym__Null_unspecified] = ACTIONS(4783),
    [anon_sym___autoreleasing] = ACTIONS(4783),
    [anon_sym___block] = ACTIONS(4783),
    [anon_sym___bridge] = ACTIONS(4779),
    [anon_sym___bridge_retained] = ACTIONS(4783),
    [anon_sym___bridge_transfer] = ACTIONS(4783),
    [anon_sym___complex] = ACTIONS(4783),
    [anon_sym___const] = ACTIONS(4783),
    [anon_sym___imag] = ACTIONS(4783),
    [anon_sym___kindof] = ACTIONS(4783),
    [anon_sym___nonnull] = ACTIONS(4783),
    [anon_sym___nullable] = ACTIONS(4783),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4783),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4783),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4783),
    [anon_sym___real] = ACTIONS(4783),
    [anon_sym___strong] = ACTIONS(4783),
    [anon_sym___unsafe_unretained] = ACTIONS(4783),
    [anon_sym___unused] = ACTIONS(4783),
    [anon_sym___weak] = ACTIONS(4783),
    [anon_sym_alignas] = ACTIONS(4783),
    [anon_sym__Alignas] = ACTIONS(4783),
    [anon_sym_QMARK] = ACTIONS(6517),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4783),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4783),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4783),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4783),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4783),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4783),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4783),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4783),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4783),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4783),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4779),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4783),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4783),
    [anon_sym_API_AVAILABLE] = ACTIONS(4783),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4783),
    [anon_sym_API_DEPRECATED] = ACTIONS(4783),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4783),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4783),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4783),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4783),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4783),
    [anon_sym___deprecated_msg] = ACTIONS(4783),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4783),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4783),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4783),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4783),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4783),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4783),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4783),
  },
  [STATE(2588)] = {
    [sym_identifier] = ACTIONS(6519),
    [aux_sym_preproc_def_token1] = ACTIONS(6519),
    [aux_sym_preproc_if_token1] = ACTIONS(6519),
    [aux_sym_preproc_if_token2] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6519),
    [aux_sym_preproc_else_token1] = ACTIONS(6519),
    [aux_sym_preproc_elif_token1] = ACTIONS(6519),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6519),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6519),
    [sym_preproc_directive] = ACTIONS(6519),
    [anon_sym___extension__] = ACTIONS(6519),
    [anon_sym_extern] = ACTIONS(6519),
    [anon_sym___attribute__] = ACTIONS(6519),
    [anon_sym___attribute] = ACTIONS(6519),
    [anon_sym_noreturn] = ACTIONS(6519),
    [anon_sym_LBRACK] = ACTIONS(6521),
    [anon_sym___declspec] = ACTIONS(6519),
    [anon_sym_signed] = ACTIONS(6519),
    [anon_sym_unsigned] = ACTIONS(6519),
    [anon_sym_long] = ACTIONS(6519),
    [anon_sym_short] = ACTIONS(6519),
    [anon_sym_static] = ACTIONS(6519),
    [anon_sym_auto] = ACTIONS(6519),
    [anon_sym_register] = ACTIONS(6519),
    [anon_sym_inline] = ACTIONS(6519),
    [anon_sym___inline] = ACTIONS(6519),
    [anon_sym___inline__] = ACTIONS(6519),
    [anon_sym___forceinline] = ACTIONS(6519),
    [anon_sym_thread_local] = ACTIONS(6519),
    [anon_sym___thread] = ACTIONS(6519),
    [anon_sym_CG_EXTERN] = ACTIONS(6519),
    [anon_sym_CG_INLINE] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6519),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6519),
    [anon_sym_IBOutlet] = ACTIONS(6519),
    [anon_sym_IBInspectable] = ACTIONS(6519),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6519),
    [anon_sym_NS_INLINE] = ACTIONS(6519),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6519),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6519),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6519),
    [anon_sym_const] = ACTIONS(6519),
    [anon_sym_constexpr] = ACTIONS(6519),
    [anon_sym_volatile] = ACTIONS(6519),
    [anon_sym_restrict] = ACTIONS(6519),
    [anon_sym___restrict__] = ACTIONS(6519),
    [anon_sym__Atomic] = ACTIONS(6519),
    [anon_sym__Noreturn] = ACTIONS(6519),
    [anon_sym__Nonnull] = ACTIONS(6519),
    [anon_sym_nullable] = ACTIONS(6519),
    [anon_sym__Complex] = ACTIONS(6519),
    [anon_sym__Nullable] = ACTIONS(6519),
    [anon_sym__Nullable_result] = ACTIONS(6519),
    [anon_sym__Null_unspecified] = ACTIONS(6519),
    [anon_sym___autoreleasing] = ACTIONS(6519),
    [anon_sym___block] = ACTIONS(6519),
    [anon_sym___bridge] = ACTIONS(6519),
    [anon_sym___bridge_retained] = ACTIONS(6519),
    [anon_sym___bridge_transfer] = ACTIONS(6519),
    [anon_sym___complex] = ACTIONS(6519),
    [anon_sym___const] = ACTIONS(6519),
    [anon_sym___imag] = ACTIONS(6519),
    [anon_sym___kindof] = ACTIONS(6519),
    [anon_sym___nonnull] = ACTIONS(6519),
    [anon_sym___nullable] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6519),
    [anon_sym___real] = ACTIONS(6519),
    [anon_sym___strong] = ACTIONS(6519),
    [anon_sym___unsafe_unretained] = ACTIONS(6519),
    [anon_sym___unused] = ACTIONS(6519),
    [anon_sym___weak] = ACTIONS(6519),
    [anon_sym_alignas] = ACTIONS(6519),
    [anon_sym__Alignas] = ACTIONS(6519),
    [sym_primitive_type] = ACTIONS(6519),
    [anon_sym_enum] = ACTIONS(6519),
    [anon_sym_struct] = ACTIONS(6519),
    [anon_sym_union] = ACTIONS(6519),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6519),
    [anon_sym___typeof] = ACTIONS(6519),
    [anon_sym_typeof] = ACTIONS(6519),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6519),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6519),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6519),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_API_AVAILABLE] = ACTIONS(6519),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_API_DEPRECATED] = ACTIONS(6519),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym___deprecated_msg] = ACTIONS(6519),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6519),
    [anon_sym_BOOL] = ACTIONS(6519),
    [anon_sym_IMP] = ACTIONS(6519),
    [anon_sym_SEL] = ACTIONS(6519),
    [anon_sym_Class] = ACTIONS(6519),
    [anon_sym_id] = ACTIONS(6519),
    [anon_sym_ATdefs] = ACTIONS(6521),
  },
  [STATE(2589)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token2] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [aux_sym_preproc_else_token1] = ACTIONS(6523),
    [aux_sym_preproc_elif_token1] = ACTIONS(6523),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2590)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token2] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [aux_sym_preproc_else_token1] = ACTIONS(6523),
    [aux_sym_preproc_elif_token1] = ACTIONS(6523),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2591)] = {
    [sym_identifier] = ACTIONS(6527),
    [aux_sym_preproc_def_token1] = ACTIONS(6527),
    [aux_sym_preproc_if_token1] = ACTIONS(6527),
    [aux_sym_preproc_if_token2] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6527),
    [aux_sym_preproc_else_token1] = ACTIONS(6527),
    [aux_sym_preproc_elif_token1] = ACTIONS(6527),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6527),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6527),
    [sym_preproc_directive] = ACTIONS(6527),
    [anon_sym___extension__] = ACTIONS(6527),
    [anon_sym_extern] = ACTIONS(6527),
    [anon_sym___attribute__] = ACTIONS(6527),
    [anon_sym___attribute] = ACTIONS(6527),
    [anon_sym_noreturn] = ACTIONS(6527),
    [anon_sym_LBRACK] = ACTIONS(6529),
    [anon_sym___declspec] = ACTIONS(6527),
    [anon_sym_signed] = ACTIONS(6527),
    [anon_sym_unsigned] = ACTIONS(6527),
    [anon_sym_long] = ACTIONS(6527),
    [anon_sym_short] = ACTIONS(6527),
    [anon_sym_static] = ACTIONS(6527),
    [anon_sym_auto] = ACTIONS(6527),
    [anon_sym_register] = ACTIONS(6527),
    [anon_sym_inline] = ACTIONS(6527),
    [anon_sym___inline] = ACTIONS(6527),
    [anon_sym___inline__] = ACTIONS(6527),
    [anon_sym___forceinline] = ACTIONS(6527),
    [anon_sym_thread_local] = ACTIONS(6527),
    [anon_sym___thread] = ACTIONS(6527),
    [anon_sym_CG_EXTERN] = ACTIONS(6527),
    [anon_sym_CG_INLINE] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6527),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6527),
    [anon_sym_IBOutlet] = ACTIONS(6527),
    [anon_sym_IBInspectable] = ACTIONS(6527),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6527),
    [anon_sym_NS_INLINE] = ACTIONS(6527),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6527),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6527),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6527),
    [anon_sym_const] = ACTIONS(6527),
    [anon_sym_constexpr] = ACTIONS(6527),
    [anon_sym_volatile] = ACTIONS(6527),
    [anon_sym_restrict] = ACTIONS(6527),
    [anon_sym___restrict__] = ACTIONS(6527),
    [anon_sym__Atomic] = ACTIONS(6527),
    [anon_sym__Noreturn] = ACTIONS(6527),
    [anon_sym__Nonnull] = ACTIONS(6527),
    [anon_sym_nullable] = ACTIONS(6527),
    [anon_sym__Complex] = ACTIONS(6527),
    [anon_sym__Nullable] = ACTIONS(6527),
    [anon_sym__Nullable_result] = ACTIONS(6527),
    [anon_sym__Null_unspecified] = ACTIONS(6527),
    [anon_sym___autoreleasing] = ACTIONS(6527),
    [anon_sym___block] = ACTIONS(6527),
    [anon_sym___bridge] = ACTIONS(6527),
    [anon_sym___bridge_retained] = ACTIONS(6527),
    [anon_sym___bridge_transfer] = ACTIONS(6527),
    [anon_sym___complex] = ACTIONS(6527),
    [anon_sym___const] = ACTIONS(6527),
    [anon_sym___imag] = ACTIONS(6527),
    [anon_sym___kindof] = ACTIONS(6527),
    [anon_sym___nonnull] = ACTIONS(6527),
    [anon_sym___nullable] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6527),
    [anon_sym___real] = ACTIONS(6527),
    [anon_sym___strong] = ACTIONS(6527),
    [anon_sym___unsafe_unretained] = ACTIONS(6527),
    [anon_sym___unused] = ACTIONS(6527),
    [anon_sym___weak] = ACTIONS(6527),
    [anon_sym_alignas] = ACTIONS(6527),
    [anon_sym__Alignas] = ACTIONS(6527),
    [sym_primitive_type] = ACTIONS(6527),
    [anon_sym_enum] = ACTIONS(6527),
    [anon_sym_struct] = ACTIONS(6527),
    [anon_sym_union] = ACTIONS(6527),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6527),
    [anon_sym___typeof] = ACTIONS(6527),
    [anon_sym_typeof] = ACTIONS(6527),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6527),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6527),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6527),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_API_AVAILABLE] = ACTIONS(6527),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_API_DEPRECATED] = ACTIONS(6527),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym___deprecated_msg] = ACTIONS(6527),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6527),
    [anon_sym_BOOL] = ACTIONS(6527),
    [anon_sym_IMP] = ACTIONS(6527),
    [anon_sym_SEL] = ACTIONS(6527),
    [anon_sym_Class] = ACTIONS(6527),
    [anon_sym_id] = ACTIONS(6527),
    [anon_sym_ATdefs] = ACTIONS(6529),
  },
  [STATE(2592)] = {
    [sym_identifier] = ACTIONS(6531),
    [aux_sym_preproc_def_token1] = ACTIONS(6531),
    [aux_sym_preproc_if_token1] = ACTIONS(6531),
    [aux_sym_preproc_if_token2] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6531),
    [aux_sym_preproc_else_token1] = ACTIONS(6531),
    [aux_sym_preproc_elif_token1] = ACTIONS(6531),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6531),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6531),
    [sym_preproc_directive] = ACTIONS(6531),
    [anon_sym___extension__] = ACTIONS(6531),
    [anon_sym_extern] = ACTIONS(6531),
    [anon_sym___attribute__] = ACTIONS(6531),
    [anon_sym___attribute] = ACTIONS(6531),
    [anon_sym_noreturn] = ACTIONS(6531),
    [anon_sym_LBRACK] = ACTIONS(6533),
    [anon_sym___declspec] = ACTIONS(6531),
    [anon_sym_signed] = ACTIONS(6531),
    [anon_sym_unsigned] = ACTIONS(6531),
    [anon_sym_long] = ACTIONS(6531),
    [anon_sym_short] = ACTIONS(6531),
    [anon_sym_static] = ACTIONS(6531),
    [anon_sym_auto] = ACTIONS(6531),
    [anon_sym_register] = ACTIONS(6531),
    [anon_sym_inline] = ACTIONS(6531),
    [anon_sym___inline] = ACTIONS(6531),
    [anon_sym___inline__] = ACTIONS(6531),
    [anon_sym___forceinline] = ACTIONS(6531),
    [anon_sym_thread_local] = ACTIONS(6531),
    [anon_sym___thread] = ACTIONS(6531),
    [anon_sym_CG_EXTERN] = ACTIONS(6531),
    [anon_sym_CG_INLINE] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6531),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6531),
    [anon_sym_IBOutlet] = ACTIONS(6531),
    [anon_sym_IBInspectable] = ACTIONS(6531),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6531),
    [anon_sym_NS_INLINE] = ACTIONS(6531),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6531),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6531),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6531),
    [anon_sym_const] = ACTIONS(6531),
    [anon_sym_constexpr] = ACTIONS(6531),
    [anon_sym_volatile] = ACTIONS(6531),
    [anon_sym_restrict] = ACTIONS(6531),
    [anon_sym___restrict__] = ACTIONS(6531),
    [anon_sym__Atomic] = ACTIONS(6531),
    [anon_sym__Noreturn] = ACTIONS(6531),
    [anon_sym__Nonnull] = ACTIONS(6531),
    [anon_sym_nullable] = ACTIONS(6531),
    [anon_sym__Complex] = ACTIONS(6531),
    [anon_sym__Nullable] = ACTIONS(6531),
    [anon_sym__Nullable_result] = ACTIONS(6531),
    [anon_sym__Null_unspecified] = ACTIONS(6531),
    [anon_sym___autoreleasing] = ACTIONS(6531),
    [anon_sym___block] = ACTIONS(6531),
    [anon_sym___bridge] = ACTIONS(6531),
    [anon_sym___bridge_retained] = ACTIONS(6531),
    [anon_sym___bridge_transfer] = ACTIONS(6531),
    [anon_sym___complex] = ACTIONS(6531),
    [anon_sym___const] = ACTIONS(6531),
    [anon_sym___imag] = ACTIONS(6531),
    [anon_sym___kindof] = ACTIONS(6531),
    [anon_sym___nonnull] = ACTIONS(6531),
    [anon_sym___nullable] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6531),
    [anon_sym___real] = ACTIONS(6531),
    [anon_sym___strong] = ACTIONS(6531),
    [anon_sym___unsafe_unretained] = ACTIONS(6531),
    [anon_sym___unused] = ACTIONS(6531),
    [anon_sym___weak] = ACTIONS(6531),
    [anon_sym_alignas] = ACTIONS(6531),
    [anon_sym__Alignas] = ACTIONS(6531),
    [sym_primitive_type] = ACTIONS(6531),
    [anon_sym_enum] = ACTIONS(6531),
    [anon_sym_struct] = ACTIONS(6531),
    [anon_sym_union] = ACTIONS(6531),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6531),
    [anon_sym___typeof] = ACTIONS(6531),
    [anon_sym_typeof] = ACTIONS(6531),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6531),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6531),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6531),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_API_AVAILABLE] = ACTIONS(6531),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_API_DEPRECATED] = ACTIONS(6531),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym___deprecated_msg] = ACTIONS(6531),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6531),
    [anon_sym_BOOL] = ACTIONS(6531),
    [anon_sym_IMP] = ACTIONS(6531),
    [anon_sym_SEL] = ACTIONS(6531),
    [anon_sym_Class] = ACTIONS(6531),
    [anon_sym_id] = ACTIONS(6531),
    [anon_sym_ATdefs] = ACTIONS(6533),
  },
  [STATE(2593)] = {
    [sym_identifier] = ACTIONS(6531),
    [aux_sym_preproc_def_token1] = ACTIONS(6531),
    [aux_sym_preproc_if_token1] = ACTIONS(6531),
    [aux_sym_preproc_if_token2] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6531),
    [aux_sym_preproc_else_token1] = ACTIONS(6531),
    [aux_sym_preproc_elif_token1] = ACTIONS(6531),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6531),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6531),
    [sym_preproc_directive] = ACTIONS(6531),
    [anon_sym___extension__] = ACTIONS(6531),
    [anon_sym_extern] = ACTIONS(6531),
    [anon_sym___attribute__] = ACTIONS(6531),
    [anon_sym___attribute] = ACTIONS(6531),
    [anon_sym_noreturn] = ACTIONS(6531),
    [anon_sym_LBRACK] = ACTIONS(6533),
    [anon_sym___declspec] = ACTIONS(6531),
    [anon_sym_signed] = ACTIONS(6531),
    [anon_sym_unsigned] = ACTIONS(6531),
    [anon_sym_long] = ACTIONS(6531),
    [anon_sym_short] = ACTIONS(6531),
    [anon_sym_static] = ACTIONS(6531),
    [anon_sym_auto] = ACTIONS(6531),
    [anon_sym_register] = ACTIONS(6531),
    [anon_sym_inline] = ACTIONS(6531),
    [anon_sym___inline] = ACTIONS(6531),
    [anon_sym___inline__] = ACTIONS(6531),
    [anon_sym___forceinline] = ACTIONS(6531),
    [anon_sym_thread_local] = ACTIONS(6531),
    [anon_sym___thread] = ACTIONS(6531),
    [anon_sym_CG_EXTERN] = ACTIONS(6531),
    [anon_sym_CG_INLINE] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6531),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6531),
    [anon_sym_IBOutlet] = ACTIONS(6531),
    [anon_sym_IBInspectable] = ACTIONS(6531),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6531),
    [anon_sym_NS_INLINE] = ACTIONS(6531),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6531),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6531),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6531),
    [anon_sym_const] = ACTIONS(6531),
    [anon_sym_constexpr] = ACTIONS(6531),
    [anon_sym_volatile] = ACTIONS(6531),
    [anon_sym_restrict] = ACTIONS(6531),
    [anon_sym___restrict__] = ACTIONS(6531),
    [anon_sym__Atomic] = ACTIONS(6531),
    [anon_sym__Noreturn] = ACTIONS(6531),
    [anon_sym__Nonnull] = ACTIONS(6531),
    [anon_sym_nullable] = ACTIONS(6531),
    [anon_sym__Complex] = ACTIONS(6531),
    [anon_sym__Nullable] = ACTIONS(6531),
    [anon_sym__Nullable_result] = ACTIONS(6531),
    [anon_sym__Null_unspecified] = ACTIONS(6531),
    [anon_sym___autoreleasing] = ACTIONS(6531),
    [anon_sym___block] = ACTIONS(6531),
    [anon_sym___bridge] = ACTIONS(6531),
    [anon_sym___bridge_retained] = ACTIONS(6531),
    [anon_sym___bridge_transfer] = ACTIONS(6531),
    [anon_sym___complex] = ACTIONS(6531),
    [anon_sym___const] = ACTIONS(6531),
    [anon_sym___imag] = ACTIONS(6531),
    [anon_sym___kindof] = ACTIONS(6531),
    [anon_sym___nonnull] = ACTIONS(6531),
    [anon_sym___nullable] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6531),
    [anon_sym___real] = ACTIONS(6531),
    [anon_sym___strong] = ACTIONS(6531),
    [anon_sym___unsafe_unretained] = ACTIONS(6531),
    [anon_sym___unused] = ACTIONS(6531),
    [anon_sym___weak] = ACTIONS(6531),
    [anon_sym_alignas] = ACTIONS(6531),
    [anon_sym__Alignas] = ACTIONS(6531),
    [sym_primitive_type] = ACTIONS(6531),
    [anon_sym_enum] = ACTIONS(6531),
    [anon_sym_struct] = ACTIONS(6531),
    [anon_sym_union] = ACTIONS(6531),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6531),
    [anon_sym___typeof] = ACTIONS(6531),
    [anon_sym_typeof] = ACTIONS(6531),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6531),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6531),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6531),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_API_AVAILABLE] = ACTIONS(6531),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_API_DEPRECATED] = ACTIONS(6531),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym___deprecated_msg] = ACTIONS(6531),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6531),
    [anon_sym_BOOL] = ACTIONS(6531),
    [anon_sym_IMP] = ACTIONS(6531),
    [anon_sym_SEL] = ACTIONS(6531),
    [anon_sym_Class] = ACTIONS(6531),
    [anon_sym_id] = ACTIONS(6531),
    [anon_sym_ATdefs] = ACTIONS(6533),
  },
  [STATE(2594)] = {
    [sym_identifier] = ACTIONS(6535),
    [aux_sym_preproc_def_token1] = ACTIONS(6535),
    [aux_sym_preproc_if_token1] = ACTIONS(6535),
    [aux_sym_preproc_if_token2] = ACTIONS(6535),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6535),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6535),
    [aux_sym_preproc_else_token1] = ACTIONS(6535),
    [aux_sym_preproc_elif_token1] = ACTIONS(6535),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6535),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6535),
    [sym_preproc_directive] = ACTIONS(6535),
    [anon_sym___extension__] = ACTIONS(6535),
    [anon_sym_extern] = ACTIONS(6535),
    [anon_sym___attribute__] = ACTIONS(6535),
    [anon_sym___attribute] = ACTIONS(6535),
    [anon_sym_noreturn] = ACTIONS(6535),
    [anon_sym_LBRACK] = ACTIONS(6537),
    [anon_sym___declspec] = ACTIONS(6535),
    [anon_sym_signed] = ACTIONS(6535),
    [anon_sym_unsigned] = ACTIONS(6535),
    [anon_sym_long] = ACTIONS(6535),
    [anon_sym_short] = ACTIONS(6535),
    [anon_sym_static] = ACTIONS(6535),
    [anon_sym_auto] = ACTIONS(6535),
    [anon_sym_register] = ACTIONS(6535),
    [anon_sym_inline] = ACTIONS(6535),
    [anon_sym___inline] = ACTIONS(6535),
    [anon_sym___inline__] = ACTIONS(6535),
    [anon_sym___forceinline] = ACTIONS(6535),
    [anon_sym_thread_local] = ACTIONS(6535),
    [anon_sym___thread] = ACTIONS(6535),
    [anon_sym_CG_EXTERN] = ACTIONS(6535),
    [anon_sym_CG_INLINE] = ACTIONS(6535),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6535),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6535),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6535),
    [anon_sym_IBOutlet] = ACTIONS(6535),
    [anon_sym_IBInspectable] = ACTIONS(6535),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6535),
    [anon_sym_NS_INLINE] = ACTIONS(6535),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6535),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6535),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6535),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6535),
    [anon_sym_const] = ACTIONS(6535),
    [anon_sym_constexpr] = ACTIONS(6535),
    [anon_sym_volatile] = ACTIONS(6535),
    [anon_sym_restrict] = ACTIONS(6535),
    [anon_sym___restrict__] = ACTIONS(6535),
    [anon_sym__Atomic] = ACTIONS(6535),
    [anon_sym__Noreturn] = ACTIONS(6535),
    [anon_sym__Nonnull] = ACTIONS(6535),
    [anon_sym_nullable] = ACTIONS(6535),
    [anon_sym__Complex] = ACTIONS(6535),
    [anon_sym__Nullable] = ACTIONS(6535),
    [anon_sym__Nullable_result] = ACTIONS(6535),
    [anon_sym__Null_unspecified] = ACTIONS(6535),
    [anon_sym___autoreleasing] = ACTIONS(6535),
    [anon_sym___block] = ACTIONS(6535),
    [anon_sym___bridge] = ACTIONS(6535),
    [anon_sym___bridge_retained] = ACTIONS(6535),
    [anon_sym___bridge_transfer] = ACTIONS(6535),
    [anon_sym___complex] = ACTIONS(6535),
    [anon_sym___const] = ACTIONS(6535),
    [anon_sym___imag] = ACTIONS(6535),
    [anon_sym___kindof] = ACTIONS(6535),
    [anon_sym___nonnull] = ACTIONS(6535),
    [anon_sym___nullable] = ACTIONS(6535),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6535),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6535),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6535),
    [anon_sym___real] = ACTIONS(6535),
    [anon_sym___strong] = ACTIONS(6535),
    [anon_sym___unsafe_unretained] = ACTIONS(6535),
    [anon_sym___unused] = ACTIONS(6535),
    [anon_sym___weak] = ACTIONS(6535),
    [anon_sym_alignas] = ACTIONS(6535),
    [anon_sym__Alignas] = ACTIONS(6535),
    [sym_primitive_type] = ACTIONS(6535),
    [anon_sym_enum] = ACTIONS(6535),
    [anon_sym_struct] = ACTIONS(6535),
    [anon_sym_union] = ACTIONS(6535),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6535),
    [anon_sym___typeof] = ACTIONS(6535),
    [anon_sym_typeof] = ACTIONS(6535),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6535),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6535),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6535),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6535),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6535),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6535),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6535),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6535),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6535),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_API_AVAILABLE] = ACTIONS(6535),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_API_DEPRECATED] = ACTIONS(6535),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6535),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6535),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6535),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6535),
    [anon_sym___deprecated_msg] = ACTIONS(6535),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6535),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6535),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6535),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6535),
    [anon_sym_BOOL] = ACTIONS(6535),
    [anon_sym_IMP] = ACTIONS(6535),
    [anon_sym_SEL] = ACTIONS(6535),
    [anon_sym_Class] = ACTIONS(6535),
    [anon_sym_id] = ACTIONS(6535),
    [anon_sym_ATdefs] = ACTIONS(6537),
  },
  [STATE(2595)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token2] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [aux_sym_preproc_else_token1] = ACTIONS(6539),
    [aux_sym_preproc_elif_token1] = ACTIONS(6539),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2596)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token2] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [aux_sym_preproc_else_token1] = ACTIONS(6539),
    [aux_sym_preproc_elif_token1] = ACTIONS(6539),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2597)] = {
    [sym_identifier] = ACTIONS(6543),
    [aux_sym_preproc_def_token1] = ACTIONS(6543),
    [aux_sym_preproc_if_token1] = ACTIONS(6543),
    [aux_sym_preproc_if_token2] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6543),
    [aux_sym_preproc_else_token1] = ACTIONS(6543),
    [aux_sym_preproc_elif_token1] = ACTIONS(6543),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6543),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6543),
    [sym_preproc_directive] = ACTIONS(6543),
    [anon_sym___extension__] = ACTIONS(6543),
    [anon_sym_extern] = ACTIONS(6543),
    [anon_sym___attribute__] = ACTIONS(6543),
    [anon_sym___attribute] = ACTIONS(6543),
    [anon_sym_noreturn] = ACTIONS(6543),
    [anon_sym_LBRACK] = ACTIONS(6545),
    [anon_sym___declspec] = ACTIONS(6543),
    [anon_sym_signed] = ACTIONS(6543),
    [anon_sym_unsigned] = ACTIONS(6543),
    [anon_sym_long] = ACTIONS(6543),
    [anon_sym_short] = ACTIONS(6543),
    [anon_sym_static] = ACTIONS(6543),
    [anon_sym_auto] = ACTIONS(6543),
    [anon_sym_register] = ACTIONS(6543),
    [anon_sym_inline] = ACTIONS(6543),
    [anon_sym___inline] = ACTIONS(6543),
    [anon_sym___inline__] = ACTIONS(6543),
    [anon_sym___forceinline] = ACTIONS(6543),
    [anon_sym_thread_local] = ACTIONS(6543),
    [anon_sym___thread] = ACTIONS(6543),
    [anon_sym_CG_EXTERN] = ACTIONS(6543),
    [anon_sym_CG_INLINE] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6543),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6543),
    [anon_sym_IBOutlet] = ACTIONS(6543),
    [anon_sym_IBInspectable] = ACTIONS(6543),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6543),
    [anon_sym_NS_INLINE] = ACTIONS(6543),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6543),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6543),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6543),
    [anon_sym_const] = ACTIONS(6543),
    [anon_sym_constexpr] = ACTIONS(6543),
    [anon_sym_volatile] = ACTIONS(6543),
    [anon_sym_restrict] = ACTIONS(6543),
    [anon_sym___restrict__] = ACTIONS(6543),
    [anon_sym__Atomic] = ACTIONS(6543),
    [anon_sym__Noreturn] = ACTIONS(6543),
    [anon_sym__Nonnull] = ACTIONS(6543),
    [anon_sym_nullable] = ACTIONS(6543),
    [anon_sym__Complex] = ACTIONS(6543),
    [anon_sym__Nullable] = ACTIONS(6543),
    [anon_sym__Nullable_result] = ACTIONS(6543),
    [anon_sym__Null_unspecified] = ACTIONS(6543),
    [anon_sym___autoreleasing] = ACTIONS(6543),
    [anon_sym___block] = ACTIONS(6543),
    [anon_sym___bridge] = ACTIONS(6543),
    [anon_sym___bridge_retained] = ACTIONS(6543),
    [anon_sym___bridge_transfer] = ACTIONS(6543),
    [anon_sym___complex] = ACTIONS(6543),
    [anon_sym___const] = ACTIONS(6543),
    [anon_sym___imag] = ACTIONS(6543),
    [anon_sym___kindof] = ACTIONS(6543),
    [anon_sym___nonnull] = ACTIONS(6543),
    [anon_sym___nullable] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6543),
    [anon_sym___real] = ACTIONS(6543),
    [anon_sym___strong] = ACTIONS(6543),
    [anon_sym___unsafe_unretained] = ACTIONS(6543),
    [anon_sym___unused] = ACTIONS(6543),
    [anon_sym___weak] = ACTIONS(6543),
    [anon_sym_alignas] = ACTIONS(6543),
    [anon_sym__Alignas] = ACTIONS(6543),
    [sym_primitive_type] = ACTIONS(6543),
    [anon_sym_enum] = ACTIONS(6543),
    [anon_sym_struct] = ACTIONS(6543),
    [anon_sym_union] = ACTIONS(6543),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6543),
    [anon_sym___typeof] = ACTIONS(6543),
    [anon_sym_typeof] = ACTIONS(6543),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6543),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6543),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6543),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_API_AVAILABLE] = ACTIONS(6543),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_API_DEPRECATED] = ACTIONS(6543),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym___deprecated_msg] = ACTIONS(6543),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6543),
    [anon_sym_BOOL] = ACTIONS(6543),
    [anon_sym_IMP] = ACTIONS(6543),
    [anon_sym_SEL] = ACTIONS(6543),
    [anon_sym_Class] = ACTIONS(6543),
    [anon_sym_id] = ACTIONS(6543),
    [anon_sym_ATdefs] = ACTIONS(6545),
  },
  [STATE(2598)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6491),
    [anon_sym_COMMA] = ACTIONS(6547),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(6499),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(6547),
    [anon_sym___extension__] = ACTIONS(6547),
    [anon_sym_extern] = ACTIONS(6547),
    [anon_sym___attribute__] = ACTIONS(6547),
    [anon_sym___attribute] = ACTIONS(6549),
    [anon_sym_noreturn] = ACTIONS(6547),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(6547),
    [anon_sym_static] = ACTIONS(6547),
    [anon_sym_auto] = ACTIONS(6547),
    [anon_sym_register] = ACTIONS(6547),
    [anon_sym_inline] = ACTIONS(6547),
    [anon_sym___inline] = ACTIONS(6549),
    [anon_sym___inline__] = ACTIONS(6547),
    [anon_sym___forceinline] = ACTIONS(6547),
    [anon_sym_thread_local] = ACTIONS(6547),
    [anon_sym___thread] = ACTIONS(6547),
    [anon_sym_CG_EXTERN] = ACTIONS(6547),
    [anon_sym_CG_INLINE] = ACTIONS(6547),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6547),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6547),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6547),
    [anon_sym_IBOutlet] = ACTIONS(6547),
    [anon_sym_IBInspectable] = ACTIONS(6547),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6547),
    [anon_sym_NS_INLINE] = ACTIONS(6547),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6547),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6547),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6547),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6547),
    [anon_sym_const] = ACTIONS(6549),
    [anon_sym_constexpr] = ACTIONS(6547),
    [anon_sym_volatile] = ACTIONS(6547),
    [anon_sym_restrict] = ACTIONS(6547),
    [anon_sym___restrict__] = ACTIONS(6547),
    [anon_sym__Atomic] = ACTIONS(6547),
    [anon_sym__Noreturn] = ACTIONS(6547),
    [anon_sym__Nonnull] = ACTIONS(6547),
    [anon_sym_nullable] = ACTIONS(6547),
    [anon_sym__Complex] = ACTIONS(6547),
    [anon_sym__Nullable] = ACTIONS(6549),
    [anon_sym__Nullable_result] = ACTIONS(6547),
    [anon_sym__Null_unspecified] = ACTIONS(6547),
    [anon_sym___autoreleasing] = ACTIONS(6547),
    [anon_sym___block] = ACTIONS(6547),
    [anon_sym___bridge] = ACTIONS(6549),
    [anon_sym___bridge_retained] = ACTIONS(6547),
    [anon_sym___bridge_transfer] = ACTIONS(6547),
    [anon_sym___complex] = ACTIONS(6547),
    [anon_sym___const] = ACTIONS(6547),
    [anon_sym___imag] = ACTIONS(6547),
    [anon_sym___kindof] = ACTIONS(6547),
    [anon_sym___nonnull] = ACTIONS(6547),
    [anon_sym___nullable] = ACTIONS(6547),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6547),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6547),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6547),
    [anon_sym___real] = ACTIONS(6547),
    [anon_sym___strong] = ACTIONS(6547),
    [anon_sym___unsafe_unretained] = ACTIONS(6547),
    [anon_sym___unused] = ACTIONS(6547),
    [anon_sym___weak] = ACTIONS(6547),
    [anon_sym_alignas] = ACTIONS(6547),
    [anon_sym__Alignas] = ACTIONS(6547),
    [anon_sym_QMARK] = ACTIONS(6517),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6547),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6547),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6547),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6547),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6547),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6547),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6547),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6547),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6547),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6547),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6549),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6547),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6547),
    [anon_sym_API_AVAILABLE] = ACTIONS(6547),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6547),
    [anon_sym_API_DEPRECATED] = ACTIONS(6547),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6547),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6547),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6547),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6547),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6547),
    [anon_sym___deprecated_msg] = ACTIONS(6547),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6547),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6547),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6547),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6547),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6547),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6547),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6547),
  },
  [STATE(2599)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token2] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [aux_sym_preproc_else_token1] = ACTIONS(6539),
    [aux_sym_preproc_elif_token1] = ACTIONS(6539),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2600)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token2] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [aux_sym_preproc_else_token1] = ACTIONS(6539),
    [aux_sym_preproc_elif_token1] = ACTIONS(6539),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2601)] = {
    [sym_identifier] = ACTIONS(6543),
    [aux_sym_preproc_def_token1] = ACTIONS(6543),
    [aux_sym_preproc_if_token1] = ACTIONS(6543),
    [aux_sym_preproc_if_token2] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6543),
    [aux_sym_preproc_else_token1] = ACTIONS(6543),
    [aux_sym_preproc_elif_token1] = ACTIONS(6543),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6543),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6543),
    [sym_preproc_directive] = ACTIONS(6543),
    [anon_sym___extension__] = ACTIONS(6543),
    [anon_sym_extern] = ACTIONS(6543),
    [anon_sym___attribute__] = ACTIONS(6543),
    [anon_sym___attribute] = ACTIONS(6543),
    [anon_sym_noreturn] = ACTIONS(6543),
    [anon_sym_LBRACK] = ACTIONS(6545),
    [anon_sym___declspec] = ACTIONS(6543),
    [anon_sym_signed] = ACTIONS(6543),
    [anon_sym_unsigned] = ACTIONS(6543),
    [anon_sym_long] = ACTIONS(6543),
    [anon_sym_short] = ACTIONS(6543),
    [anon_sym_static] = ACTIONS(6543),
    [anon_sym_auto] = ACTIONS(6543),
    [anon_sym_register] = ACTIONS(6543),
    [anon_sym_inline] = ACTIONS(6543),
    [anon_sym___inline] = ACTIONS(6543),
    [anon_sym___inline__] = ACTIONS(6543),
    [anon_sym___forceinline] = ACTIONS(6543),
    [anon_sym_thread_local] = ACTIONS(6543),
    [anon_sym___thread] = ACTIONS(6543),
    [anon_sym_CG_EXTERN] = ACTIONS(6543),
    [anon_sym_CG_INLINE] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6543),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6543),
    [anon_sym_IBOutlet] = ACTIONS(6543),
    [anon_sym_IBInspectable] = ACTIONS(6543),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6543),
    [anon_sym_NS_INLINE] = ACTIONS(6543),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6543),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6543),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6543),
    [anon_sym_const] = ACTIONS(6543),
    [anon_sym_constexpr] = ACTIONS(6543),
    [anon_sym_volatile] = ACTIONS(6543),
    [anon_sym_restrict] = ACTIONS(6543),
    [anon_sym___restrict__] = ACTIONS(6543),
    [anon_sym__Atomic] = ACTIONS(6543),
    [anon_sym__Noreturn] = ACTIONS(6543),
    [anon_sym__Nonnull] = ACTIONS(6543),
    [anon_sym_nullable] = ACTIONS(6543),
    [anon_sym__Complex] = ACTIONS(6543),
    [anon_sym__Nullable] = ACTIONS(6543),
    [anon_sym__Nullable_result] = ACTIONS(6543),
    [anon_sym__Null_unspecified] = ACTIONS(6543),
    [anon_sym___autoreleasing] = ACTIONS(6543),
    [anon_sym___block] = ACTIONS(6543),
    [anon_sym___bridge] = ACTIONS(6543),
    [anon_sym___bridge_retained] = ACTIONS(6543),
    [anon_sym___bridge_transfer] = ACTIONS(6543),
    [anon_sym___complex] = ACTIONS(6543),
    [anon_sym___const] = ACTIONS(6543),
    [anon_sym___imag] = ACTIONS(6543),
    [anon_sym___kindof] = ACTIONS(6543),
    [anon_sym___nonnull] = ACTIONS(6543),
    [anon_sym___nullable] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6543),
    [anon_sym___real] = ACTIONS(6543),
    [anon_sym___strong] = ACTIONS(6543),
    [anon_sym___unsafe_unretained] = ACTIONS(6543),
    [anon_sym___unused] = ACTIONS(6543),
    [anon_sym___weak] = ACTIONS(6543),
    [anon_sym_alignas] = ACTIONS(6543),
    [anon_sym__Alignas] = ACTIONS(6543),
    [sym_primitive_type] = ACTIONS(6543),
    [anon_sym_enum] = ACTIONS(6543),
    [anon_sym_struct] = ACTIONS(6543),
    [anon_sym_union] = ACTIONS(6543),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6543),
    [anon_sym___typeof] = ACTIONS(6543),
    [anon_sym_typeof] = ACTIONS(6543),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6543),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6543),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6543),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_API_AVAILABLE] = ACTIONS(6543),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_API_DEPRECATED] = ACTIONS(6543),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym___deprecated_msg] = ACTIONS(6543),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6543),
    [anon_sym_BOOL] = ACTIONS(6543),
    [anon_sym_IMP] = ACTIONS(6543),
    [anon_sym_SEL] = ACTIONS(6543),
    [anon_sym_Class] = ACTIONS(6543),
    [anon_sym_id] = ACTIONS(6543),
    [anon_sym_ATdefs] = ACTIONS(6545),
  },
  [STATE(2602)] = {
    [sym_identifier] = ACTIONS(6551),
    [aux_sym_preproc_def_token1] = ACTIONS(6551),
    [aux_sym_preproc_if_token1] = ACTIONS(6551),
    [aux_sym_preproc_if_token2] = ACTIONS(6551),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6551),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6551),
    [aux_sym_preproc_else_token1] = ACTIONS(6551),
    [aux_sym_preproc_elif_token1] = ACTIONS(6551),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6551),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6551),
    [sym_preproc_directive] = ACTIONS(6551),
    [anon_sym___extension__] = ACTIONS(6551),
    [anon_sym_extern] = ACTIONS(6551),
    [anon_sym___attribute__] = ACTIONS(6551),
    [anon_sym___attribute] = ACTIONS(6551),
    [anon_sym_noreturn] = ACTIONS(6551),
    [anon_sym_LBRACK] = ACTIONS(6553),
    [anon_sym___declspec] = ACTIONS(6551),
    [anon_sym_signed] = ACTIONS(6551),
    [anon_sym_unsigned] = ACTIONS(6551),
    [anon_sym_long] = ACTIONS(6551),
    [anon_sym_short] = ACTIONS(6551),
    [anon_sym_static] = ACTIONS(6551),
    [anon_sym_auto] = ACTIONS(6551),
    [anon_sym_register] = ACTIONS(6551),
    [anon_sym_inline] = ACTIONS(6551),
    [anon_sym___inline] = ACTIONS(6551),
    [anon_sym___inline__] = ACTIONS(6551),
    [anon_sym___forceinline] = ACTIONS(6551),
    [anon_sym_thread_local] = ACTIONS(6551),
    [anon_sym___thread] = ACTIONS(6551),
    [anon_sym_CG_EXTERN] = ACTIONS(6551),
    [anon_sym_CG_INLINE] = ACTIONS(6551),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6551),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6551),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6551),
    [anon_sym_IBOutlet] = ACTIONS(6551),
    [anon_sym_IBInspectable] = ACTIONS(6551),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6551),
    [anon_sym_NS_INLINE] = ACTIONS(6551),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6551),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6551),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6551),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6551),
    [anon_sym_const] = ACTIONS(6551),
    [anon_sym_constexpr] = ACTIONS(6551),
    [anon_sym_volatile] = ACTIONS(6551),
    [anon_sym_restrict] = ACTIONS(6551),
    [anon_sym___restrict__] = ACTIONS(6551),
    [anon_sym__Atomic] = ACTIONS(6551),
    [anon_sym__Noreturn] = ACTIONS(6551),
    [anon_sym__Nonnull] = ACTIONS(6551),
    [anon_sym_nullable] = ACTIONS(6551),
    [anon_sym__Complex] = ACTIONS(6551),
    [anon_sym__Nullable] = ACTIONS(6551),
    [anon_sym__Nullable_result] = ACTIONS(6551),
    [anon_sym__Null_unspecified] = ACTIONS(6551),
    [anon_sym___autoreleasing] = ACTIONS(6551),
    [anon_sym___block] = ACTIONS(6551),
    [anon_sym___bridge] = ACTIONS(6551),
    [anon_sym___bridge_retained] = ACTIONS(6551),
    [anon_sym___bridge_transfer] = ACTIONS(6551),
    [anon_sym___complex] = ACTIONS(6551),
    [anon_sym___const] = ACTIONS(6551),
    [anon_sym___imag] = ACTIONS(6551),
    [anon_sym___kindof] = ACTIONS(6551),
    [anon_sym___nonnull] = ACTIONS(6551),
    [anon_sym___nullable] = ACTIONS(6551),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6551),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6551),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6551),
    [anon_sym___real] = ACTIONS(6551),
    [anon_sym___strong] = ACTIONS(6551),
    [anon_sym___unsafe_unretained] = ACTIONS(6551),
    [anon_sym___unused] = ACTIONS(6551),
    [anon_sym___weak] = ACTIONS(6551),
    [anon_sym_alignas] = ACTIONS(6551),
    [anon_sym__Alignas] = ACTIONS(6551),
    [sym_primitive_type] = ACTIONS(6551),
    [anon_sym_enum] = ACTIONS(6551),
    [anon_sym_struct] = ACTIONS(6551),
    [anon_sym_union] = ACTIONS(6551),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6551),
    [anon_sym___typeof] = ACTIONS(6551),
    [anon_sym_typeof] = ACTIONS(6551),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6551),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6551),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6551),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6551),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6551),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6551),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6551),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6551),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6551),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_API_AVAILABLE] = ACTIONS(6551),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_API_DEPRECATED] = ACTIONS(6551),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6551),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6551),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6551),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6551),
    [anon_sym___deprecated_msg] = ACTIONS(6551),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6551),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6551),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6551),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6551),
    [anon_sym_BOOL] = ACTIONS(6551),
    [anon_sym_IMP] = ACTIONS(6551),
    [anon_sym_SEL] = ACTIONS(6551),
    [anon_sym_Class] = ACTIONS(6551),
    [anon_sym_id] = ACTIONS(6551),
    [anon_sym_ATdefs] = ACTIONS(6553),
  },
  [STATE(2603)] = {
    [sym_identifier] = ACTIONS(6555),
    [aux_sym_preproc_def_token1] = ACTIONS(6555),
    [aux_sym_preproc_if_token1] = ACTIONS(6555),
    [aux_sym_preproc_if_token2] = ACTIONS(6555),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6555),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6555),
    [aux_sym_preproc_else_token1] = ACTIONS(6555),
    [aux_sym_preproc_elif_token1] = ACTIONS(6555),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6555),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6555),
    [sym_preproc_directive] = ACTIONS(6555),
    [anon_sym___extension__] = ACTIONS(6555),
    [anon_sym_extern] = ACTIONS(6555),
    [anon_sym___attribute__] = ACTIONS(6555),
    [anon_sym___attribute] = ACTIONS(6555),
    [anon_sym_noreturn] = ACTIONS(6555),
    [anon_sym_LBRACK] = ACTIONS(6557),
    [anon_sym___declspec] = ACTIONS(6555),
    [anon_sym_signed] = ACTIONS(6555),
    [anon_sym_unsigned] = ACTIONS(6555),
    [anon_sym_long] = ACTIONS(6555),
    [anon_sym_short] = ACTIONS(6555),
    [anon_sym_static] = ACTIONS(6555),
    [anon_sym_auto] = ACTIONS(6555),
    [anon_sym_register] = ACTIONS(6555),
    [anon_sym_inline] = ACTIONS(6555),
    [anon_sym___inline] = ACTIONS(6555),
    [anon_sym___inline__] = ACTIONS(6555),
    [anon_sym___forceinline] = ACTIONS(6555),
    [anon_sym_thread_local] = ACTIONS(6555),
    [anon_sym___thread] = ACTIONS(6555),
    [anon_sym_CG_EXTERN] = ACTIONS(6555),
    [anon_sym_CG_INLINE] = ACTIONS(6555),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6555),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6555),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6555),
    [anon_sym_IBOutlet] = ACTIONS(6555),
    [anon_sym_IBInspectable] = ACTIONS(6555),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6555),
    [anon_sym_NS_INLINE] = ACTIONS(6555),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6555),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6555),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6555),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6555),
    [anon_sym_const] = ACTIONS(6555),
    [anon_sym_constexpr] = ACTIONS(6555),
    [anon_sym_volatile] = ACTIONS(6555),
    [anon_sym_restrict] = ACTIONS(6555),
    [anon_sym___restrict__] = ACTIONS(6555),
    [anon_sym__Atomic] = ACTIONS(6555),
    [anon_sym__Noreturn] = ACTIONS(6555),
    [anon_sym__Nonnull] = ACTIONS(6555),
    [anon_sym_nullable] = ACTIONS(6555),
    [anon_sym__Complex] = ACTIONS(6555),
    [anon_sym__Nullable] = ACTIONS(6555),
    [anon_sym__Nullable_result] = ACTIONS(6555),
    [anon_sym__Null_unspecified] = ACTIONS(6555),
    [anon_sym___autoreleasing] = ACTIONS(6555),
    [anon_sym___block] = ACTIONS(6555),
    [anon_sym___bridge] = ACTIONS(6555),
    [anon_sym___bridge_retained] = ACTIONS(6555),
    [anon_sym___bridge_transfer] = ACTIONS(6555),
    [anon_sym___complex] = ACTIONS(6555),
    [anon_sym___const] = ACTIONS(6555),
    [anon_sym___imag] = ACTIONS(6555),
    [anon_sym___kindof] = ACTIONS(6555),
    [anon_sym___nonnull] = ACTIONS(6555),
    [anon_sym___nullable] = ACTIONS(6555),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6555),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6555),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6555),
    [anon_sym___real] = ACTIONS(6555),
    [anon_sym___strong] = ACTIONS(6555),
    [anon_sym___unsafe_unretained] = ACTIONS(6555),
    [anon_sym___unused] = ACTIONS(6555),
    [anon_sym___weak] = ACTIONS(6555),
    [anon_sym_alignas] = ACTIONS(6555),
    [anon_sym__Alignas] = ACTIONS(6555),
    [sym_primitive_type] = ACTIONS(6555),
    [anon_sym_enum] = ACTIONS(6555),
    [anon_sym_struct] = ACTIONS(6555),
    [anon_sym_union] = ACTIONS(6555),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6555),
    [anon_sym___typeof] = ACTIONS(6555),
    [anon_sym_typeof] = ACTIONS(6555),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6555),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6555),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6555),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6555),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6555),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6555),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6555),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6555),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6555),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_API_AVAILABLE] = ACTIONS(6555),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_API_DEPRECATED] = ACTIONS(6555),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6555),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6555),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6555),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6555),
    [anon_sym___deprecated_msg] = ACTIONS(6555),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6555),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6555),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6555),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6555),
    [anon_sym_BOOL] = ACTIONS(6555),
    [anon_sym_IMP] = ACTIONS(6555),
    [anon_sym_SEL] = ACTIONS(6555),
    [anon_sym_Class] = ACTIONS(6555),
    [anon_sym_id] = ACTIONS(6555),
    [anon_sym_ATdefs] = ACTIONS(6557),
  },
  [STATE(2604)] = {
    [sym_identifier] = ACTIONS(2856),
    [aux_sym_preproc_def_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token1] = ACTIONS(2856),
    [aux_sym_preproc_if_token2] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2856),
    [aux_sym_preproc_else_token1] = ACTIONS(2856),
    [aux_sym_preproc_elif_token1] = ACTIONS(2856),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2856),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2856),
    [sym_preproc_directive] = ACTIONS(2856),
    [anon_sym___extension__] = ACTIONS(2856),
    [anon_sym_extern] = ACTIONS(2856),
    [anon_sym___attribute__] = ACTIONS(2856),
    [anon_sym___attribute] = ACTIONS(2856),
    [anon_sym_noreturn] = ACTIONS(2856),
    [anon_sym_LBRACK] = ACTIONS(2858),
    [anon_sym___declspec] = ACTIONS(2856),
    [anon_sym_signed] = ACTIONS(2856),
    [anon_sym_unsigned] = ACTIONS(2856),
    [anon_sym_long] = ACTIONS(2856),
    [anon_sym_short] = ACTIONS(2856),
    [anon_sym_static] = ACTIONS(2856),
    [anon_sym_auto] = ACTIONS(2856),
    [anon_sym_register] = ACTIONS(2856),
    [anon_sym_inline] = ACTIONS(2856),
    [anon_sym___inline] = ACTIONS(2856),
    [anon_sym___inline__] = ACTIONS(2856),
    [anon_sym___forceinline] = ACTIONS(2856),
    [anon_sym_thread_local] = ACTIONS(2856),
    [anon_sym___thread] = ACTIONS(2856),
    [anon_sym_CG_EXTERN] = ACTIONS(2856),
    [anon_sym_CG_INLINE] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2856),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2856),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2856),
    [anon_sym_IBOutlet] = ACTIONS(2856),
    [anon_sym_IBInspectable] = ACTIONS(2856),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2856),
    [anon_sym_NS_INLINE] = ACTIONS(2856),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2856),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2856),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2856),
    [anon_sym_const] = ACTIONS(2856),
    [anon_sym_constexpr] = ACTIONS(2856),
    [anon_sym_volatile] = ACTIONS(2856),
    [anon_sym_restrict] = ACTIONS(2856),
    [anon_sym___restrict__] = ACTIONS(2856),
    [anon_sym__Atomic] = ACTIONS(2856),
    [anon_sym__Noreturn] = ACTIONS(2856),
    [anon_sym__Nonnull] = ACTIONS(2856),
    [anon_sym_nullable] = ACTIONS(2856),
    [anon_sym__Complex] = ACTIONS(2856),
    [anon_sym__Nullable] = ACTIONS(2856),
    [anon_sym__Nullable_result] = ACTIONS(2856),
    [anon_sym__Null_unspecified] = ACTIONS(2856),
    [anon_sym___autoreleasing] = ACTIONS(2856),
    [anon_sym___block] = ACTIONS(2856),
    [anon_sym___bridge] = ACTIONS(2856),
    [anon_sym___bridge_retained] = ACTIONS(2856),
    [anon_sym___bridge_transfer] = ACTIONS(2856),
    [anon_sym___complex] = ACTIONS(2856),
    [anon_sym___const] = ACTIONS(2856),
    [anon_sym___imag] = ACTIONS(2856),
    [anon_sym___kindof] = ACTIONS(2856),
    [anon_sym___nonnull] = ACTIONS(2856),
    [anon_sym___nullable] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2856),
    [anon_sym___real] = ACTIONS(2856),
    [anon_sym___strong] = ACTIONS(2856),
    [anon_sym___unsafe_unretained] = ACTIONS(2856),
    [anon_sym___unused] = ACTIONS(2856),
    [anon_sym___weak] = ACTIONS(2856),
    [anon_sym_alignas] = ACTIONS(2856),
    [anon_sym__Alignas] = ACTIONS(2856),
    [sym_primitive_type] = ACTIONS(2856),
    [anon_sym_enum] = ACTIONS(2856),
    [anon_sym_struct] = ACTIONS(2856),
    [anon_sym_union] = ACTIONS(2856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2856),
    [anon_sym___typeof] = ACTIONS(2856),
    [anon_sym_typeof] = ACTIONS(2856),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2856),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2856),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2856),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2856),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2856),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_API_AVAILABLE] = ACTIONS(2856),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_API_DEPRECATED] = ACTIONS(2856),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2856),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2856),
    [anon_sym___deprecated_msg] = ACTIONS(2856),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2856),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2856),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2856),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2856),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2856),
    [anon_sym_BOOL] = ACTIONS(2856),
    [anon_sym_IMP] = ACTIONS(2856),
    [anon_sym_SEL] = ACTIONS(2856),
    [anon_sym_Class] = ACTIONS(2856),
    [anon_sym_id] = ACTIONS(2856),
    [anon_sym_ATdefs] = ACTIONS(2858),
  },
  [STATE(2605)] = {
    [sym_identifier] = ACTIONS(2588),
    [aux_sym_preproc_def_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token1] = ACTIONS(2588),
    [aux_sym_preproc_if_token2] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2588),
    [aux_sym_preproc_else_token1] = ACTIONS(2588),
    [aux_sym_preproc_elif_token1] = ACTIONS(2588),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2588),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2588),
    [sym_preproc_directive] = ACTIONS(2588),
    [anon_sym___extension__] = ACTIONS(2588),
    [anon_sym_extern] = ACTIONS(2588),
    [anon_sym___attribute__] = ACTIONS(2588),
    [anon_sym___attribute] = ACTIONS(2588),
    [anon_sym_noreturn] = ACTIONS(2588),
    [anon_sym_LBRACK] = ACTIONS(2590),
    [anon_sym___declspec] = ACTIONS(2588),
    [anon_sym_signed] = ACTIONS(2588),
    [anon_sym_unsigned] = ACTIONS(2588),
    [anon_sym_long] = ACTIONS(2588),
    [anon_sym_short] = ACTIONS(2588),
    [anon_sym_static] = ACTIONS(2588),
    [anon_sym_auto] = ACTIONS(2588),
    [anon_sym_register] = ACTIONS(2588),
    [anon_sym_inline] = ACTIONS(2588),
    [anon_sym___inline] = ACTIONS(2588),
    [anon_sym___inline__] = ACTIONS(2588),
    [anon_sym___forceinline] = ACTIONS(2588),
    [anon_sym_thread_local] = ACTIONS(2588),
    [anon_sym___thread] = ACTIONS(2588),
    [anon_sym_CG_EXTERN] = ACTIONS(2588),
    [anon_sym_CG_INLINE] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2588),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2588),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2588),
    [anon_sym_IBOutlet] = ACTIONS(2588),
    [anon_sym_IBInspectable] = ACTIONS(2588),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2588),
    [anon_sym_NS_INLINE] = ACTIONS(2588),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2588),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2588),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2588),
    [anon_sym_const] = ACTIONS(2588),
    [anon_sym_constexpr] = ACTIONS(2588),
    [anon_sym_volatile] = ACTIONS(2588),
    [anon_sym_restrict] = ACTIONS(2588),
    [anon_sym___restrict__] = ACTIONS(2588),
    [anon_sym__Atomic] = ACTIONS(2588),
    [anon_sym__Noreturn] = ACTIONS(2588),
    [anon_sym__Nonnull] = ACTIONS(2588),
    [anon_sym_nullable] = ACTIONS(2588),
    [anon_sym__Complex] = ACTIONS(2588),
    [anon_sym__Nullable] = ACTIONS(2588),
    [anon_sym__Nullable_result] = ACTIONS(2588),
    [anon_sym__Null_unspecified] = ACTIONS(2588),
    [anon_sym___autoreleasing] = ACTIONS(2588),
    [anon_sym___block] = ACTIONS(2588),
    [anon_sym___bridge] = ACTIONS(2588),
    [anon_sym___bridge_retained] = ACTIONS(2588),
    [anon_sym___bridge_transfer] = ACTIONS(2588),
    [anon_sym___complex] = ACTIONS(2588),
    [anon_sym___const] = ACTIONS(2588),
    [anon_sym___imag] = ACTIONS(2588),
    [anon_sym___kindof] = ACTIONS(2588),
    [anon_sym___nonnull] = ACTIONS(2588),
    [anon_sym___nullable] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2588),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2588),
    [anon_sym___real] = ACTIONS(2588),
    [anon_sym___strong] = ACTIONS(2588),
    [anon_sym___unsafe_unretained] = ACTIONS(2588),
    [anon_sym___unused] = ACTIONS(2588),
    [anon_sym___weak] = ACTIONS(2588),
    [anon_sym_alignas] = ACTIONS(2588),
    [anon_sym__Alignas] = ACTIONS(2588),
    [sym_primitive_type] = ACTIONS(2588),
    [anon_sym_enum] = ACTIONS(2588),
    [anon_sym_struct] = ACTIONS(2588),
    [anon_sym_union] = ACTIONS(2588),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2588),
    [anon_sym___typeof] = ACTIONS(2588),
    [anon_sym_typeof] = ACTIONS(2588),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2588),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2588),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2588),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2588),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2588),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_API_AVAILABLE] = ACTIONS(2588),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_API_DEPRECATED] = ACTIONS(2588),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2588),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2588),
    [anon_sym___deprecated_msg] = ACTIONS(2588),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2588),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2588),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2588),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2588),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2588),
    [anon_sym_BOOL] = ACTIONS(2588),
    [anon_sym_IMP] = ACTIONS(2588),
    [anon_sym_SEL] = ACTIONS(2588),
    [anon_sym_Class] = ACTIONS(2588),
    [anon_sym_id] = ACTIONS(2588),
    [anon_sym_ATdefs] = ACTIONS(2590),
  },
  [STATE(2606)] = {
    [sym_identifier] = ACTIONS(2852),
    [aux_sym_preproc_def_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token1] = ACTIONS(2852),
    [aux_sym_preproc_if_token2] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2852),
    [aux_sym_preproc_else_token1] = ACTIONS(2852),
    [aux_sym_preproc_elif_token1] = ACTIONS(2852),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2852),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2852),
    [sym_preproc_directive] = ACTIONS(2852),
    [anon_sym___extension__] = ACTIONS(2852),
    [anon_sym_extern] = ACTIONS(2852),
    [anon_sym___attribute__] = ACTIONS(2852),
    [anon_sym___attribute] = ACTIONS(2852),
    [anon_sym_noreturn] = ACTIONS(2852),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym___declspec] = ACTIONS(2852),
    [anon_sym_signed] = ACTIONS(2852),
    [anon_sym_unsigned] = ACTIONS(2852),
    [anon_sym_long] = ACTIONS(2852),
    [anon_sym_short] = ACTIONS(2852),
    [anon_sym_static] = ACTIONS(2852),
    [anon_sym_auto] = ACTIONS(2852),
    [anon_sym_register] = ACTIONS(2852),
    [anon_sym_inline] = ACTIONS(2852),
    [anon_sym___inline] = ACTIONS(2852),
    [anon_sym___inline__] = ACTIONS(2852),
    [anon_sym___forceinline] = ACTIONS(2852),
    [anon_sym_thread_local] = ACTIONS(2852),
    [anon_sym___thread] = ACTIONS(2852),
    [anon_sym_CG_EXTERN] = ACTIONS(2852),
    [anon_sym_CG_INLINE] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2852),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2852),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2852),
    [anon_sym_IBOutlet] = ACTIONS(2852),
    [anon_sym_IBInspectable] = ACTIONS(2852),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2852),
    [anon_sym_NS_INLINE] = ACTIONS(2852),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2852),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2852),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2852),
    [anon_sym_const] = ACTIONS(2852),
    [anon_sym_constexpr] = ACTIONS(2852),
    [anon_sym_volatile] = ACTIONS(2852),
    [anon_sym_restrict] = ACTIONS(2852),
    [anon_sym___restrict__] = ACTIONS(2852),
    [anon_sym__Atomic] = ACTIONS(2852),
    [anon_sym__Noreturn] = ACTIONS(2852),
    [anon_sym__Nonnull] = ACTIONS(2852),
    [anon_sym_nullable] = ACTIONS(2852),
    [anon_sym__Complex] = ACTIONS(2852),
    [anon_sym__Nullable] = ACTIONS(2852),
    [anon_sym__Nullable_result] = ACTIONS(2852),
    [anon_sym__Null_unspecified] = ACTIONS(2852),
    [anon_sym___autoreleasing] = ACTIONS(2852),
    [anon_sym___block] = ACTIONS(2852),
    [anon_sym___bridge] = ACTIONS(2852),
    [anon_sym___bridge_retained] = ACTIONS(2852),
    [anon_sym___bridge_transfer] = ACTIONS(2852),
    [anon_sym___complex] = ACTIONS(2852),
    [anon_sym___const] = ACTIONS(2852),
    [anon_sym___imag] = ACTIONS(2852),
    [anon_sym___kindof] = ACTIONS(2852),
    [anon_sym___nonnull] = ACTIONS(2852),
    [anon_sym___nullable] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2852),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2852),
    [anon_sym___real] = ACTIONS(2852),
    [anon_sym___strong] = ACTIONS(2852),
    [anon_sym___unsafe_unretained] = ACTIONS(2852),
    [anon_sym___unused] = ACTIONS(2852),
    [anon_sym___weak] = ACTIONS(2852),
    [anon_sym_alignas] = ACTIONS(2852),
    [anon_sym__Alignas] = ACTIONS(2852),
    [sym_primitive_type] = ACTIONS(2852),
    [anon_sym_enum] = ACTIONS(2852),
    [anon_sym_struct] = ACTIONS(2852),
    [anon_sym_union] = ACTIONS(2852),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2852),
    [anon_sym___typeof] = ACTIONS(2852),
    [anon_sym_typeof] = ACTIONS(2852),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2852),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2852),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2852),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2852),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2852),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_API_AVAILABLE] = ACTIONS(2852),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_API_DEPRECATED] = ACTIONS(2852),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2852),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2852),
    [anon_sym___deprecated_msg] = ACTIONS(2852),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2852),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2852),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2852),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2852),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2852),
    [anon_sym_BOOL] = ACTIONS(2852),
    [anon_sym_IMP] = ACTIONS(2852),
    [anon_sym_SEL] = ACTIONS(2852),
    [anon_sym_Class] = ACTIONS(2852),
    [anon_sym_id] = ACTIONS(2852),
    [anon_sym_ATdefs] = ACTIONS(2854),
  },
  [STATE(2607)] = {
    [sym_identifier] = ACTIONS(2608),
    [aux_sym_preproc_def_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token1] = ACTIONS(2608),
    [aux_sym_preproc_if_token2] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2608),
    [aux_sym_preproc_else_token1] = ACTIONS(2608),
    [aux_sym_preproc_elif_token1] = ACTIONS(2608),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2608),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2608),
    [sym_preproc_directive] = ACTIONS(2608),
    [anon_sym___extension__] = ACTIONS(2608),
    [anon_sym_extern] = ACTIONS(2608),
    [anon_sym___attribute__] = ACTIONS(2608),
    [anon_sym___attribute] = ACTIONS(2608),
    [anon_sym_noreturn] = ACTIONS(2608),
    [anon_sym_LBRACK] = ACTIONS(2610),
    [anon_sym___declspec] = ACTIONS(2608),
    [anon_sym_signed] = ACTIONS(2608),
    [anon_sym_unsigned] = ACTIONS(2608),
    [anon_sym_long] = ACTIONS(2608),
    [anon_sym_short] = ACTIONS(2608),
    [anon_sym_static] = ACTIONS(2608),
    [anon_sym_auto] = ACTIONS(2608),
    [anon_sym_register] = ACTIONS(2608),
    [anon_sym_inline] = ACTIONS(2608),
    [anon_sym___inline] = ACTIONS(2608),
    [anon_sym___inline__] = ACTIONS(2608),
    [anon_sym___forceinline] = ACTIONS(2608),
    [anon_sym_thread_local] = ACTIONS(2608),
    [anon_sym___thread] = ACTIONS(2608),
    [anon_sym_CG_EXTERN] = ACTIONS(2608),
    [anon_sym_CG_INLINE] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2608),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2608),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2608),
    [anon_sym_IBOutlet] = ACTIONS(2608),
    [anon_sym_IBInspectable] = ACTIONS(2608),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2608),
    [anon_sym_NS_INLINE] = ACTIONS(2608),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2608),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2608),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2608),
    [anon_sym_const] = ACTIONS(2608),
    [anon_sym_constexpr] = ACTIONS(2608),
    [anon_sym_volatile] = ACTIONS(2608),
    [anon_sym_restrict] = ACTIONS(2608),
    [anon_sym___restrict__] = ACTIONS(2608),
    [anon_sym__Atomic] = ACTIONS(2608),
    [anon_sym__Noreturn] = ACTIONS(2608),
    [anon_sym__Nonnull] = ACTIONS(2608),
    [anon_sym_nullable] = ACTIONS(2608),
    [anon_sym__Complex] = ACTIONS(2608),
    [anon_sym__Nullable] = ACTIONS(2608),
    [anon_sym__Nullable_result] = ACTIONS(2608),
    [anon_sym__Null_unspecified] = ACTIONS(2608),
    [anon_sym___autoreleasing] = ACTIONS(2608),
    [anon_sym___block] = ACTIONS(2608),
    [anon_sym___bridge] = ACTIONS(2608),
    [anon_sym___bridge_retained] = ACTIONS(2608),
    [anon_sym___bridge_transfer] = ACTIONS(2608),
    [anon_sym___complex] = ACTIONS(2608),
    [anon_sym___const] = ACTIONS(2608),
    [anon_sym___imag] = ACTIONS(2608),
    [anon_sym___kindof] = ACTIONS(2608),
    [anon_sym___nonnull] = ACTIONS(2608),
    [anon_sym___nullable] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2608),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2608),
    [anon_sym___real] = ACTIONS(2608),
    [anon_sym___strong] = ACTIONS(2608),
    [anon_sym___unsafe_unretained] = ACTIONS(2608),
    [anon_sym___unused] = ACTIONS(2608),
    [anon_sym___weak] = ACTIONS(2608),
    [anon_sym_alignas] = ACTIONS(2608),
    [anon_sym__Alignas] = ACTIONS(2608),
    [sym_primitive_type] = ACTIONS(2608),
    [anon_sym_enum] = ACTIONS(2608),
    [anon_sym_struct] = ACTIONS(2608),
    [anon_sym_union] = ACTIONS(2608),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2608),
    [anon_sym___typeof] = ACTIONS(2608),
    [anon_sym_typeof] = ACTIONS(2608),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2608),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2608),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2608),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2608),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2608),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_API_AVAILABLE] = ACTIONS(2608),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_API_DEPRECATED] = ACTIONS(2608),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2608),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2608),
    [anon_sym___deprecated_msg] = ACTIONS(2608),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2608),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2608),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2608),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2608),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2608),
    [anon_sym_BOOL] = ACTIONS(2608),
    [anon_sym_IMP] = ACTIONS(2608),
    [anon_sym_SEL] = ACTIONS(2608),
    [anon_sym_Class] = ACTIONS(2608),
    [anon_sym_id] = ACTIONS(2608),
    [anon_sym_ATdefs] = ACTIONS(2610),
  },
  [STATE(2608)] = {
    [sym_identifier] = ACTIONS(2618),
    [aux_sym_preproc_def_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token1] = ACTIONS(2618),
    [aux_sym_preproc_if_token2] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2618),
    [aux_sym_preproc_else_token1] = ACTIONS(2618),
    [aux_sym_preproc_elif_token1] = ACTIONS(2618),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2618),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2618),
    [sym_preproc_directive] = ACTIONS(2618),
    [anon_sym___extension__] = ACTIONS(2618),
    [anon_sym_extern] = ACTIONS(2618),
    [anon_sym___attribute__] = ACTIONS(2618),
    [anon_sym___attribute] = ACTIONS(2618),
    [anon_sym_noreturn] = ACTIONS(2618),
    [anon_sym_LBRACK] = ACTIONS(2620),
    [anon_sym___declspec] = ACTIONS(2618),
    [anon_sym_signed] = ACTIONS(2618),
    [anon_sym_unsigned] = ACTIONS(2618),
    [anon_sym_long] = ACTIONS(2618),
    [anon_sym_short] = ACTIONS(2618),
    [anon_sym_static] = ACTIONS(2618),
    [anon_sym_auto] = ACTIONS(2618),
    [anon_sym_register] = ACTIONS(2618),
    [anon_sym_inline] = ACTIONS(2618),
    [anon_sym___inline] = ACTIONS(2618),
    [anon_sym___inline__] = ACTIONS(2618),
    [anon_sym___forceinline] = ACTIONS(2618),
    [anon_sym_thread_local] = ACTIONS(2618),
    [anon_sym___thread] = ACTIONS(2618),
    [anon_sym_CG_EXTERN] = ACTIONS(2618),
    [anon_sym_CG_INLINE] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2618),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2618),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2618),
    [anon_sym_IBOutlet] = ACTIONS(2618),
    [anon_sym_IBInspectable] = ACTIONS(2618),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2618),
    [anon_sym_NS_INLINE] = ACTIONS(2618),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2618),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2618),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2618),
    [anon_sym_const] = ACTIONS(2618),
    [anon_sym_constexpr] = ACTIONS(2618),
    [anon_sym_volatile] = ACTIONS(2618),
    [anon_sym_restrict] = ACTIONS(2618),
    [anon_sym___restrict__] = ACTIONS(2618),
    [anon_sym__Atomic] = ACTIONS(2618),
    [anon_sym__Noreturn] = ACTIONS(2618),
    [anon_sym__Nonnull] = ACTIONS(2618),
    [anon_sym_nullable] = ACTIONS(2618),
    [anon_sym__Complex] = ACTIONS(2618),
    [anon_sym__Nullable] = ACTIONS(2618),
    [anon_sym__Nullable_result] = ACTIONS(2618),
    [anon_sym__Null_unspecified] = ACTIONS(2618),
    [anon_sym___autoreleasing] = ACTIONS(2618),
    [anon_sym___block] = ACTIONS(2618),
    [anon_sym___bridge] = ACTIONS(2618),
    [anon_sym___bridge_retained] = ACTIONS(2618),
    [anon_sym___bridge_transfer] = ACTIONS(2618),
    [anon_sym___complex] = ACTIONS(2618),
    [anon_sym___const] = ACTIONS(2618),
    [anon_sym___imag] = ACTIONS(2618),
    [anon_sym___kindof] = ACTIONS(2618),
    [anon_sym___nonnull] = ACTIONS(2618),
    [anon_sym___nullable] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2618),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2618),
    [anon_sym___real] = ACTIONS(2618),
    [anon_sym___strong] = ACTIONS(2618),
    [anon_sym___unsafe_unretained] = ACTIONS(2618),
    [anon_sym___unused] = ACTIONS(2618),
    [anon_sym___weak] = ACTIONS(2618),
    [anon_sym_alignas] = ACTIONS(2618),
    [anon_sym__Alignas] = ACTIONS(2618),
    [sym_primitive_type] = ACTIONS(2618),
    [anon_sym_enum] = ACTIONS(2618),
    [anon_sym_struct] = ACTIONS(2618),
    [anon_sym_union] = ACTIONS(2618),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2618),
    [anon_sym___typeof] = ACTIONS(2618),
    [anon_sym_typeof] = ACTIONS(2618),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2618),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2618),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2618),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2618),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2618),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_API_AVAILABLE] = ACTIONS(2618),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_API_DEPRECATED] = ACTIONS(2618),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2618),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2618),
    [anon_sym___deprecated_msg] = ACTIONS(2618),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2618),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2618),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2618),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2618),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2618),
    [anon_sym_BOOL] = ACTIONS(2618),
    [anon_sym_IMP] = ACTIONS(2618),
    [anon_sym_SEL] = ACTIONS(2618),
    [anon_sym_Class] = ACTIONS(2618),
    [anon_sym_id] = ACTIONS(2618),
    [anon_sym_ATdefs] = ACTIONS(2620),
  },
  [STATE(2609)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6491),
    [anon_sym_COMMA] = ACTIONS(4821),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(6499),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4821),
    [anon_sym___extension__] = ACTIONS(4821),
    [anon_sym_extern] = ACTIONS(4821),
    [anon_sym___attribute__] = ACTIONS(4821),
    [anon_sym___attribute] = ACTIONS(4819),
    [anon_sym_noreturn] = ACTIONS(4821),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4821),
    [anon_sym_static] = ACTIONS(4821),
    [anon_sym_auto] = ACTIONS(4821),
    [anon_sym_register] = ACTIONS(4821),
    [anon_sym_inline] = ACTIONS(4821),
    [anon_sym___inline] = ACTIONS(4819),
    [anon_sym___inline__] = ACTIONS(4821),
    [anon_sym___forceinline] = ACTIONS(4821),
    [anon_sym_thread_local] = ACTIONS(4821),
    [anon_sym___thread] = ACTIONS(4821),
    [anon_sym_CG_EXTERN] = ACTIONS(4821),
    [anon_sym_CG_INLINE] = ACTIONS(4821),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4821),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4821),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4821),
    [anon_sym_IBOutlet] = ACTIONS(4821),
    [anon_sym_IBInspectable] = ACTIONS(4821),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4821),
    [anon_sym_NS_INLINE] = ACTIONS(4821),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4821),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4821),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4821),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4821),
    [anon_sym_const] = ACTIONS(4819),
    [anon_sym_constexpr] = ACTIONS(4821),
    [anon_sym_volatile] = ACTIONS(4821),
    [anon_sym_restrict] = ACTIONS(4821),
    [anon_sym___restrict__] = ACTIONS(4821),
    [anon_sym__Atomic] = ACTIONS(4821),
    [anon_sym__Noreturn] = ACTIONS(4821),
    [anon_sym__Nonnull] = ACTIONS(4821),
    [anon_sym_nullable] = ACTIONS(4821),
    [anon_sym__Complex] = ACTIONS(4821),
    [anon_sym__Nullable] = ACTIONS(4819),
    [anon_sym__Nullable_result] = ACTIONS(4821),
    [anon_sym__Null_unspecified] = ACTIONS(4821),
    [anon_sym___autoreleasing] = ACTIONS(4821),
    [anon_sym___block] = ACTIONS(4821),
    [anon_sym___bridge] = ACTIONS(4819),
    [anon_sym___bridge_retained] = ACTIONS(4821),
    [anon_sym___bridge_transfer] = ACTIONS(4821),
    [anon_sym___complex] = ACTIONS(4821),
    [anon_sym___const] = ACTIONS(4821),
    [anon_sym___imag] = ACTIONS(4821),
    [anon_sym___kindof] = ACTIONS(4821),
    [anon_sym___nonnull] = ACTIONS(4821),
    [anon_sym___nullable] = ACTIONS(4821),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4821),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4821),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4821),
    [anon_sym___real] = ACTIONS(4821),
    [anon_sym___strong] = ACTIONS(4821),
    [anon_sym___unsafe_unretained] = ACTIONS(4821),
    [anon_sym___unused] = ACTIONS(4821),
    [anon_sym___weak] = ACTIONS(4821),
    [anon_sym_alignas] = ACTIONS(4821),
    [anon_sym__Alignas] = ACTIONS(4821),
    [anon_sym_QMARK] = ACTIONS(6517),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4821),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4821),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4821),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4821),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4821),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4821),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4821),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4821),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4821),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4821),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4819),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4821),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4821),
    [anon_sym_API_AVAILABLE] = ACTIONS(4821),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4821),
    [anon_sym_API_DEPRECATED] = ACTIONS(4821),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4821),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4821),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4821),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4821),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4821),
    [anon_sym___deprecated_msg] = ACTIONS(4821),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4821),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4821),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4821),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4821),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4821),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4821),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4821),
  },
  [STATE(2610)] = {
    [sym_identifier] = ACTIONS(6559),
    [aux_sym_preproc_def_token1] = ACTIONS(6559),
    [aux_sym_preproc_if_token1] = ACTIONS(6559),
    [aux_sym_preproc_if_token2] = ACTIONS(6559),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6559),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6559),
    [aux_sym_preproc_else_token1] = ACTIONS(6559),
    [aux_sym_preproc_elif_token1] = ACTIONS(6559),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6559),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6559),
    [sym_preproc_directive] = ACTIONS(6559),
    [anon_sym___extension__] = ACTIONS(6559),
    [anon_sym_extern] = ACTIONS(6559),
    [anon_sym___attribute__] = ACTIONS(6559),
    [anon_sym___attribute] = ACTIONS(6559),
    [anon_sym_noreturn] = ACTIONS(6559),
    [anon_sym_LBRACK] = ACTIONS(6561),
    [anon_sym___declspec] = ACTIONS(6559),
    [anon_sym_signed] = ACTIONS(6559),
    [anon_sym_unsigned] = ACTIONS(6559),
    [anon_sym_long] = ACTIONS(6559),
    [anon_sym_short] = ACTIONS(6559),
    [anon_sym_static] = ACTIONS(6559),
    [anon_sym_auto] = ACTIONS(6559),
    [anon_sym_register] = ACTIONS(6559),
    [anon_sym_inline] = ACTIONS(6559),
    [anon_sym___inline] = ACTIONS(6559),
    [anon_sym___inline__] = ACTIONS(6559),
    [anon_sym___forceinline] = ACTIONS(6559),
    [anon_sym_thread_local] = ACTIONS(6559),
    [anon_sym___thread] = ACTIONS(6559),
    [anon_sym_CG_EXTERN] = ACTIONS(6559),
    [anon_sym_CG_INLINE] = ACTIONS(6559),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6559),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6559),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6559),
    [anon_sym_IBOutlet] = ACTIONS(6559),
    [anon_sym_IBInspectable] = ACTIONS(6559),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6559),
    [anon_sym_NS_INLINE] = ACTIONS(6559),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6559),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6559),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6559),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6559),
    [anon_sym_const] = ACTIONS(6559),
    [anon_sym_constexpr] = ACTIONS(6559),
    [anon_sym_volatile] = ACTIONS(6559),
    [anon_sym_restrict] = ACTIONS(6559),
    [anon_sym___restrict__] = ACTIONS(6559),
    [anon_sym__Atomic] = ACTIONS(6559),
    [anon_sym__Noreturn] = ACTIONS(6559),
    [anon_sym__Nonnull] = ACTIONS(6559),
    [anon_sym_nullable] = ACTIONS(6559),
    [anon_sym__Complex] = ACTIONS(6559),
    [anon_sym__Nullable] = ACTIONS(6559),
    [anon_sym__Nullable_result] = ACTIONS(6559),
    [anon_sym__Null_unspecified] = ACTIONS(6559),
    [anon_sym___autoreleasing] = ACTIONS(6559),
    [anon_sym___block] = ACTIONS(6559),
    [anon_sym___bridge] = ACTIONS(6559),
    [anon_sym___bridge_retained] = ACTIONS(6559),
    [anon_sym___bridge_transfer] = ACTIONS(6559),
    [anon_sym___complex] = ACTIONS(6559),
    [anon_sym___const] = ACTIONS(6559),
    [anon_sym___imag] = ACTIONS(6559),
    [anon_sym___kindof] = ACTIONS(6559),
    [anon_sym___nonnull] = ACTIONS(6559),
    [anon_sym___nullable] = ACTIONS(6559),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6559),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6559),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6559),
    [anon_sym___real] = ACTIONS(6559),
    [anon_sym___strong] = ACTIONS(6559),
    [anon_sym___unsafe_unretained] = ACTIONS(6559),
    [anon_sym___unused] = ACTIONS(6559),
    [anon_sym___weak] = ACTIONS(6559),
    [anon_sym_alignas] = ACTIONS(6559),
    [anon_sym__Alignas] = ACTIONS(6559),
    [sym_primitive_type] = ACTIONS(6559),
    [anon_sym_enum] = ACTIONS(6559),
    [anon_sym_struct] = ACTIONS(6559),
    [anon_sym_union] = ACTIONS(6559),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6559),
    [anon_sym___typeof] = ACTIONS(6559),
    [anon_sym_typeof] = ACTIONS(6559),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6559),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6559),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6559),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6559),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6559),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6559),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6559),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6559),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6559),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_API_AVAILABLE] = ACTIONS(6559),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_API_DEPRECATED] = ACTIONS(6559),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6559),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6559),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6559),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6559),
    [anon_sym___deprecated_msg] = ACTIONS(6559),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6559),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6559),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6559),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6559),
    [anon_sym_BOOL] = ACTIONS(6559),
    [anon_sym_IMP] = ACTIONS(6559),
    [anon_sym_SEL] = ACTIONS(6559),
    [anon_sym_Class] = ACTIONS(6559),
    [anon_sym_id] = ACTIONS(6559),
    [anon_sym_ATdefs] = ACTIONS(6561),
  },
  [STATE(2611)] = {
    [sym_identifier] = ACTIONS(6563),
    [aux_sym_preproc_def_token1] = ACTIONS(6563),
    [aux_sym_preproc_if_token1] = ACTIONS(6563),
    [aux_sym_preproc_if_token2] = ACTIONS(6563),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6563),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6563),
    [aux_sym_preproc_else_token1] = ACTIONS(6563),
    [aux_sym_preproc_elif_token1] = ACTIONS(6563),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6563),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6563),
    [sym_preproc_directive] = ACTIONS(6563),
    [anon_sym___extension__] = ACTIONS(6563),
    [anon_sym_extern] = ACTIONS(6563),
    [anon_sym___attribute__] = ACTIONS(6563),
    [anon_sym___attribute] = ACTIONS(6563),
    [anon_sym_noreturn] = ACTIONS(6563),
    [anon_sym_LBRACK] = ACTIONS(6565),
    [anon_sym___declspec] = ACTIONS(6563),
    [anon_sym_signed] = ACTIONS(6563),
    [anon_sym_unsigned] = ACTIONS(6563),
    [anon_sym_long] = ACTIONS(6563),
    [anon_sym_short] = ACTIONS(6563),
    [anon_sym_static] = ACTIONS(6563),
    [anon_sym_auto] = ACTIONS(6563),
    [anon_sym_register] = ACTIONS(6563),
    [anon_sym_inline] = ACTIONS(6563),
    [anon_sym___inline] = ACTIONS(6563),
    [anon_sym___inline__] = ACTIONS(6563),
    [anon_sym___forceinline] = ACTIONS(6563),
    [anon_sym_thread_local] = ACTIONS(6563),
    [anon_sym___thread] = ACTIONS(6563),
    [anon_sym_CG_EXTERN] = ACTIONS(6563),
    [anon_sym_CG_INLINE] = ACTIONS(6563),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6563),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6563),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6563),
    [anon_sym_IBOutlet] = ACTIONS(6563),
    [anon_sym_IBInspectable] = ACTIONS(6563),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6563),
    [anon_sym_NS_INLINE] = ACTIONS(6563),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6563),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6563),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6563),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6563),
    [anon_sym_const] = ACTIONS(6563),
    [anon_sym_constexpr] = ACTIONS(6563),
    [anon_sym_volatile] = ACTIONS(6563),
    [anon_sym_restrict] = ACTIONS(6563),
    [anon_sym___restrict__] = ACTIONS(6563),
    [anon_sym__Atomic] = ACTIONS(6563),
    [anon_sym__Noreturn] = ACTIONS(6563),
    [anon_sym__Nonnull] = ACTIONS(6563),
    [anon_sym_nullable] = ACTIONS(6563),
    [anon_sym__Complex] = ACTIONS(6563),
    [anon_sym__Nullable] = ACTIONS(6563),
    [anon_sym__Nullable_result] = ACTIONS(6563),
    [anon_sym__Null_unspecified] = ACTIONS(6563),
    [anon_sym___autoreleasing] = ACTIONS(6563),
    [anon_sym___block] = ACTIONS(6563),
    [anon_sym___bridge] = ACTIONS(6563),
    [anon_sym___bridge_retained] = ACTIONS(6563),
    [anon_sym___bridge_transfer] = ACTIONS(6563),
    [anon_sym___complex] = ACTIONS(6563),
    [anon_sym___const] = ACTIONS(6563),
    [anon_sym___imag] = ACTIONS(6563),
    [anon_sym___kindof] = ACTIONS(6563),
    [anon_sym___nonnull] = ACTIONS(6563),
    [anon_sym___nullable] = ACTIONS(6563),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6563),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6563),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6563),
    [anon_sym___real] = ACTIONS(6563),
    [anon_sym___strong] = ACTIONS(6563),
    [anon_sym___unsafe_unretained] = ACTIONS(6563),
    [anon_sym___unused] = ACTIONS(6563),
    [anon_sym___weak] = ACTIONS(6563),
    [anon_sym_alignas] = ACTIONS(6563),
    [anon_sym__Alignas] = ACTIONS(6563),
    [sym_primitive_type] = ACTIONS(6563),
    [anon_sym_enum] = ACTIONS(6563),
    [anon_sym_struct] = ACTIONS(6563),
    [anon_sym_union] = ACTIONS(6563),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6563),
    [anon_sym___typeof] = ACTIONS(6563),
    [anon_sym_typeof] = ACTIONS(6563),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6563),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6563),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6563),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6563),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6563),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6563),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6563),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6563),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6563),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_API_AVAILABLE] = ACTIONS(6563),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_API_DEPRECATED] = ACTIONS(6563),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6563),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6563),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6563),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6563),
    [anon_sym___deprecated_msg] = ACTIONS(6563),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6563),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6563),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6563),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6563),
    [anon_sym_BOOL] = ACTIONS(6563),
    [anon_sym_IMP] = ACTIONS(6563),
    [anon_sym_SEL] = ACTIONS(6563),
    [anon_sym_Class] = ACTIONS(6563),
    [anon_sym_id] = ACTIONS(6563),
    [anon_sym_ATdefs] = ACTIONS(6565),
  },
  [STATE(2612)] = {
    [sym_identifier] = ACTIONS(6567),
    [aux_sym_preproc_def_token1] = ACTIONS(6567),
    [aux_sym_preproc_if_token1] = ACTIONS(6567),
    [aux_sym_preproc_if_token2] = ACTIONS(6567),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6567),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6567),
    [aux_sym_preproc_else_token1] = ACTIONS(6567),
    [aux_sym_preproc_elif_token1] = ACTIONS(6567),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6567),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6567),
    [sym_preproc_directive] = ACTIONS(6567),
    [anon_sym___extension__] = ACTIONS(6567),
    [anon_sym_extern] = ACTIONS(6567),
    [anon_sym___attribute__] = ACTIONS(6567),
    [anon_sym___attribute] = ACTIONS(6567),
    [anon_sym_noreturn] = ACTIONS(6567),
    [anon_sym_LBRACK] = ACTIONS(6569),
    [anon_sym___declspec] = ACTIONS(6567),
    [anon_sym_signed] = ACTIONS(6567),
    [anon_sym_unsigned] = ACTIONS(6567),
    [anon_sym_long] = ACTIONS(6567),
    [anon_sym_short] = ACTIONS(6567),
    [anon_sym_static] = ACTIONS(6567),
    [anon_sym_auto] = ACTIONS(6567),
    [anon_sym_register] = ACTIONS(6567),
    [anon_sym_inline] = ACTIONS(6567),
    [anon_sym___inline] = ACTIONS(6567),
    [anon_sym___inline__] = ACTIONS(6567),
    [anon_sym___forceinline] = ACTIONS(6567),
    [anon_sym_thread_local] = ACTIONS(6567),
    [anon_sym___thread] = ACTIONS(6567),
    [anon_sym_CG_EXTERN] = ACTIONS(6567),
    [anon_sym_CG_INLINE] = ACTIONS(6567),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6567),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6567),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6567),
    [anon_sym_IBOutlet] = ACTIONS(6567),
    [anon_sym_IBInspectable] = ACTIONS(6567),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6567),
    [anon_sym_NS_INLINE] = ACTIONS(6567),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6567),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6567),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6567),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6567),
    [anon_sym_const] = ACTIONS(6567),
    [anon_sym_constexpr] = ACTIONS(6567),
    [anon_sym_volatile] = ACTIONS(6567),
    [anon_sym_restrict] = ACTIONS(6567),
    [anon_sym___restrict__] = ACTIONS(6567),
    [anon_sym__Atomic] = ACTIONS(6567),
    [anon_sym__Noreturn] = ACTIONS(6567),
    [anon_sym__Nonnull] = ACTIONS(6567),
    [anon_sym_nullable] = ACTIONS(6567),
    [anon_sym__Complex] = ACTIONS(6567),
    [anon_sym__Nullable] = ACTIONS(6567),
    [anon_sym__Nullable_result] = ACTIONS(6567),
    [anon_sym__Null_unspecified] = ACTIONS(6567),
    [anon_sym___autoreleasing] = ACTIONS(6567),
    [anon_sym___block] = ACTIONS(6567),
    [anon_sym___bridge] = ACTIONS(6567),
    [anon_sym___bridge_retained] = ACTIONS(6567),
    [anon_sym___bridge_transfer] = ACTIONS(6567),
    [anon_sym___complex] = ACTIONS(6567),
    [anon_sym___const] = ACTIONS(6567),
    [anon_sym___imag] = ACTIONS(6567),
    [anon_sym___kindof] = ACTIONS(6567),
    [anon_sym___nonnull] = ACTIONS(6567),
    [anon_sym___nullable] = ACTIONS(6567),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6567),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6567),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6567),
    [anon_sym___real] = ACTIONS(6567),
    [anon_sym___strong] = ACTIONS(6567),
    [anon_sym___unsafe_unretained] = ACTIONS(6567),
    [anon_sym___unused] = ACTIONS(6567),
    [anon_sym___weak] = ACTIONS(6567),
    [anon_sym_alignas] = ACTIONS(6567),
    [anon_sym__Alignas] = ACTIONS(6567),
    [sym_primitive_type] = ACTIONS(6567),
    [anon_sym_enum] = ACTIONS(6567),
    [anon_sym_struct] = ACTIONS(6567),
    [anon_sym_union] = ACTIONS(6567),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6567),
    [anon_sym___typeof] = ACTIONS(6567),
    [anon_sym_typeof] = ACTIONS(6567),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6567),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6567),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6567),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6567),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6567),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6567),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6567),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6567),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6567),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_API_AVAILABLE] = ACTIONS(6567),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_API_DEPRECATED] = ACTIONS(6567),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6567),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6567),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6567),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6567),
    [anon_sym___deprecated_msg] = ACTIONS(6567),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6567),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6567),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6567),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6567),
    [anon_sym_BOOL] = ACTIONS(6567),
    [anon_sym_IMP] = ACTIONS(6567),
    [anon_sym_SEL] = ACTIONS(6567),
    [anon_sym_Class] = ACTIONS(6567),
    [anon_sym_id] = ACTIONS(6567),
    [anon_sym_ATdefs] = ACTIONS(6569),
  },
  [STATE(2613)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token2] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [aux_sym_preproc_else_token1] = ACTIONS(6571),
    [aux_sym_preproc_elif_token1] = ACTIONS(6571),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2614)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token2] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [aux_sym_preproc_else_token1] = ACTIONS(6571),
    [aux_sym_preproc_elif_token1] = ACTIONS(6571),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2615)] = {
    [sym_identifier] = ACTIONS(6575),
    [aux_sym_preproc_def_token1] = ACTIONS(6575),
    [aux_sym_preproc_if_token1] = ACTIONS(6575),
    [aux_sym_preproc_if_token2] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6575),
    [aux_sym_preproc_else_token1] = ACTIONS(6575),
    [aux_sym_preproc_elif_token1] = ACTIONS(6575),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6575),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6575),
    [sym_preproc_directive] = ACTIONS(6575),
    [anon_sym___extension__] = ACTIONS(6575),
    [anon_sym_extern] = ACTIONS(6575),
    [anon_sym___attribute__] = ACTIONS(6575),
    [anon_sym___attribute] = ACTIONS(6575),
    [anon_sym_noreturn] = ACTIONS(6575),
    [anon_sym_LBRACK] = ACTIONS(6577),
    [anon_sym___declspec] = ACTIONS(6575),
    [anon_sym_signed] = ACTIONS(6575),
    [anon_sym_unsigned] = ACTIONS(6575),
    [anon_sym_long] = ACTIONS(6575),
    [anon_sym_short] = ACTIONS(6575),
    [anon_sym_static] = ACTIONS(6575),
    [anon_sym_auto] = ACTIONS(6575),
    [anon_sym_register] = ACTIONS(6575),
    [anon_sym_inline] = ACTIONS(6575),
    [anon_sym___inline] = ACTIONS(6575),
    [anon_sym___inline__] = ACTIONS(6575),
    [anon_sym___forceinline] = ACTIONS(6575),
    [anon_sym_thread_local] = ACTIONS(6575),
    [anon_sym___thread] = ACTIONS(6575),
    [anon_sym_CG_EXTERN] = ACTIONS(6575),
    [anon_sym_CG_INLINE] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6575),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6575),
    [anon_sym_IBOutlet] = ACTIONS(6575),
    [anon_sym_IBInspectable] = ACTIONS(6575),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6575),
    [anon_sym_NS_INLINE] = ACTIONS(6575),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6575),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6575),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6575),
    [anon_sym_const] = ACTIONS(6575),
    [anon_sym_constexpr] = ACTIONS(6575),
    [anon_sym_volatile] = ACTIONS(6575),
    [anon_sym_restrict] = ACTIONS(6575),
    [anon_sym___restrict__] = ACTIONS(6575),
    [anon_sym__Atomic] = ACTIONS(6575),
    [anon_sym__Noreturn] = ACTIONS(6575),
    [anon_sym__Nonnull] = ACTIONS(6575),
    [anon_sym_nullable] = ACTIONS(6575),
    [anon_sym__Complex] = ACTIONS(6575),
    [anon_sym__Nullable] = ACTIONS(6575),
    [anon_sym__Nullable_result] = ACTIONS(6575),
    [anon_sym__Null_unspecified] = ACTIONS(6575),
    [anon_sym___autoreleasing] = ACTIONS(6575),
    [anon_sym___block] = ACTIONS(6575),
    [anon_sym___bridge] = ACTIONS(6575),
    [anon_sym___bridge_retained] = ACTIONS(6575),
    [anon_sym___bridge_transfer] = ACTIONS(6575),
    [anon_sym___complex] = ACTIONS(6575),
    [anon_sym___const] = ACTIONS(6575),
    [anon_sym___imag] = ACTIONS(6575),
    [anon_sym___kindof] = ACTIONS(6575),
    [anon_sym___nonnull] = ACTIONS(6575),
    [anon_sym___nullable] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6575),
    [anon_sym___real] = ACTIONS(6575),
    [anon_sym___strong] = ACTIONS(6575),
    [anon_sym___unsafe_unretained] = ACTIONS(6575),
    [anon_sym___unused] = ACTIONS(6575),
    [anon_sym___weak] = ACTIONS(6575),
    [anon_sym_alignas] = ACTIONS(6575),
    [anon_sym__Alignas] = ACTIONS(6575),
    [sym_primitive_type] = ACTIONS(6575),
    [anon_sym_enum] = ACTIONS(6575),
    [anon_sym_struct] = ACTIONS(6575),
    [anon_sym_union] = ACTIONS(6575),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6575),
    [anon_sym___typeof] = ACTIONS(6575),
    [anon_sym_typeof] = ACTIONS(6575),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6575),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6575),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6575),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_API_AVAILABLE] = ACTIONS(6575),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_API_DEPRECATED] = ACTIONS(6575),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym___deprecated_msg] = ACTIONS(6575),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6575),
    [anon_sym_BOOL] = ACTIONS(6575),
    [anon_sym_IMP] = ACTIONS(6575),
    [anon_sym_SEL] = ACTIONS(6575),
    [anon_sym_Class] = ACTIONS(6575),
    [anon_sym_id] = ACTIONS(6575),
    [anon_sym_ATdefs] = ACTIONS(6577),
  },
  [STATE(2616)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6491),
    [anon_sym_COMMA] = ACTIONS(6579),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(6499),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(6579),
    [anon_sym___extension__] = ACTIONS(6579),
    [anon_sym_extern] = ACTIONS(6579),
    [anon_sym___attribute__] = ACTIONS(6579),
    [anon_sym___attribute] = ACTIONS(6581),
    [anon_sym_noreturn] = ACTIONS(6579),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(6579),
    [anon_sym_static] = ACTIONS(6579),
    [anon_sym_auto] = ACTIONS(6579),
    [anon_sym_register] = ACTIONS(6579),
    [anon_sym_inline] = ACTIONS(6579),
    [anon_sym___inline] = ACTIONS(6581),
    [anon_sym___inline__] = ACTIONS(6579),
    [anon_sym___forceinline] = ACTIONS(6579),
    [anon_sym_thread_local] = ACTIONS(6579),
    [anon_sym___thread] = ACTIONS(6579),
    [anon_sym_CG_EXTERN] = ACTIONS(6579),
    [anon_sym_CG_INLINE] = ACTIONS(6579),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6579),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6579),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6579),
    [anon_sym_IBOutlet] = ACTIONS(6579),
    [anon_sym_IBInspectable] = ACTIONS(6579),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6579),
    [anon_sym_NS_INLINE] = ACTIONS(6579),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6579),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6579),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6579),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6579),
    [anon_sym_const] = ACTIONS(6581),
    [anon_sym_constexpr] = ACTIONS(6579),
    [anon_sym_volatile] = ACTIONS(6579),
    [anon_sym_restrict] = ACTIONS(6579),
    [anon_sym___restrict__] = ACTIONS(6579),
    [anon_sym__Atomic] = ACTIONS(6579),
    [anon_sym__Noreturn] = ACTIONS(6579),
    [anon_sym__Nonnull] = ACTIONS(6579),
    [anon_sym_nullable] = ACTIONS(6579),
    [anon_sym__Complex] = ACTIONS(6579),
    [anon_sym__Nullable] = ACTIONS(6581),
    [anon_sym__Nullable_result] = ACTIONS(6579),
    [anon_sym__Null_unspecified] = ACTIONS(6579),
    [anon_sym___autoreleasing] = ACTIONS(6579),
    [anon_sym___block] = ACTIONS(6579),
    [anon_sym___bridge] = ACTIONS(6581),
    [anon_sym___bridge_retained] = ACTIONS(6579),
    [anon_sym___bridge_transfer] = ACTIONS(6579),
    [anon_sym___complex] = ACTIONS(6579),
    [anon_sym___const] = ACTIONS(6579),
    [anon_sym___imag] = ACTIONS(6579),
    [anon_sym___kindof] = ACTIONS(6579),
    [anon_sym___nonnull] = ACTIONS(6579),
    [anon_sym___nullable] = ACTIONS(6579),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6579),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6579),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6579),
    [anon_sym___real] = ACTIONS(6579),
    [anon_sym___strong] = ACTIONS(6579),
    [anon_sym___unsafe_unretained] = ACTIONS(6579),
    [anon_sym___unused] = ACTIONS(6579),
    [anon_sym___weak] = ACTIONS(6579),
    [anon_sym_alignas] = ACTIONS(6579),
    [anon_sym__Alignas] = ACTIONS(6579),
    [anon_sym_QMARK] = ACTIONS(6517),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6579),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6579),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6579),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6579),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6579),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6579),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6579),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6579),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6579),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6579),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6581),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6579),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6579),
    [anon_sym_API_AVAILABLE] = ACTIONS(6579),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6579),
    [anon_sym_API_DEPRECATED] = ACTIONS(6579),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6579),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6579),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6579),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6579),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6579),
    [anon_sym___deprecated_msg] = ACTIONS(6579),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6579),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6579),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6579),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6579),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6579),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6579),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6579),
  },
  [STATE(2617)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token2] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [aux_sym_preproc_else_token1] = ACTIONS(6583),
    [aux_sym_preproc_elif_token1] = ACTIONS(6583),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2618)] = {
    [sym_identifier] = ACTIONS(2600),
    [aux_sym_preproc_def_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token1] = ACTIONS(2600),
    [aux_sym_preproc_if_token2] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2600),
    [aux_sym_preproc_else_token1] = ACTIONS(2600),
    [aux_sym_preproc_elif_token1] = ACTIONS(2600),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2600),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2600),
    [sym_preproc_directive] = ACTIONS(2600),
    [anon_sym___extension__] = ACTIONS(2600),
    [anon_sym_extern] = ACTIONS(2600),
    [anon_sym___attribute__] = ACTIONS(2600),
    [anon_sym___attribute] = ACTIONS(2600),
    [anon_sym_noreturn] = ACTIONS(2600),
    [anon_sym_LBRACK] = ACTIONS(2602),
    [anon_sym___declspec] = ACTIONS(2600),
    [anon_sym_signed] = ACTIONS(2600),
    [anon_sym_unsigned] = ACTIONS(2600),
    [anon_sym_long] = ACTIONS(2600),
    [anon_sym_short] = ACTIONS(2600),
    [anon_sym_static] = ACTIONS(2600),
    [anon_sym_auto] = ACTIONS(2600),
    [anon_sym_register] = ACTIONS(2600),
    [anon_sym_inline] = ACTIONS(2600),
    [anon_sym___inline] = ACTIONS(2600),
    [anon_sym___inline__] = ACTIONS(2600),
    [anon_sym___forceinline] = ACTIONS(2600),
    [anon_sym_thread_local] = ACTIONS(2600),
    [anon_sym___thread] = ACTIONS(2600),
    [anon_sym_CG_EXTERN] = ACTIONS(2600),
    [anon_sym_CG_INLINE] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2600),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2600),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2600),
    [anon_sym_IBOutlet] = ACTIONS(2600),
    [anon_sym_IBInspectable] = ACTIONS(2600),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2600),
    [anon_sym_NS_INLINE] = ACTIONS(2600),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2600),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2600),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2600),
    [anon_sym_const] = ACTIONS(2600),
    [anon_sym_constexpr] = ACTIONS(2600),
    [anon_sym_volatile] = ACTIONS(2600),
    [anon_sym_restrict] = ACTIONS(2600),
    [anon_sym___restrict__] = ACTIONS(2600),
    [anon_sym__Atomic] = ACTIONS(2600),
    [anon_sym__Noreturn] = ACTIONS(2600),
    [anon_sym__Nonnull] = ACTIONS(2600),
    [anon_sym_nullable] = ACTIONS(2600),
    [anon_sym__Complex] = ACTIONS(2600),
    [anon_sym__Nullable] = ACTIONS(2600),
    [anon_sym__Nullable_result] = ACTIONS(2600),
    [anon_sym__Null_unspecified] = ACTIONS(2600),
    [anon_sym___autoreleasing] = ACTIONS(2600),
    [anon_sym___block] = ACTIONS(2600),
    [anon_sym___bridge] = ACTIONS(2600),
    [anon_sym___bridge_retained] = ACTIONS(2600),
    [anon_sym___bridge_transfer] = ACTIONS(2600),
    [anon_sym___complex] = ACTIONS(2600),
    [anon_sym___const] = ACTIONS(2600),
    [anon_sym___imag] = ACTIONS(2600),
    [anon_sym___kindof] = ACTIONS(2600),
    [anon_sym___nonnull] = ACTIONS(2600),
    [anon_sym___nullable] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2600),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2600),
    [anon_sym___real] = ACTIONS(2600),
    [anon_sym___strong] = ACTIONS(2600),
    [anon_sym___unsafe_unretained] = ACTIONS(2600),
    [anon_sym___unused] = ACTIONS(2600),
    [anon_sym___weak] = ACTIONS(2600),
    [anon_sym_alignas] = ACTIONS(2600),
    [anon_sym__Alignas] = ACTIONS(2600),
    [sym_primitive_type] = ACTIONS(2600),
    [anon_sym_enum] = ACTIONS(2600),
    [anon_sym_struct] = ACTIONS(2600),
    [anon_sym_union] = ACTIONS(2600),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2600),
    [anon_sym___typeof] = ACTIONS(2600),
    [anon_sym_typeof] = ACTIONS(2600),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2600),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2600),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2600),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2600),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2600),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_API_AVAILABLE] = ACTIONS(2600),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_API_DEPRECATED] = ACTIONS(2600),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2600),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2600),
    [anon_sym___deprecated_msg] = ACTIONS(2600),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2600),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2600),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2600),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2600),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2600),
    [anon_sym_BOOL] = ACTIONS(2600),
    [anon_sym_IMP] = ACTIONS(2600),
    [anon_sym_SEL] = ACTIONS(2600),
    [anon_sym_Class] = ACTIONS(2600),
    [anon_sym_id] = ACTIONS(2600),
    [anon_sym_ATdefs] = ACTIONS(2602),
  },
  [STATE(2619)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token2] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [aux_sym_preproc_else_token1] = ACTIONS(6571),
    [aux_sym_preproc_elif_token1] = ACTIONS(6571),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2620)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token2] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [aux_sym_preproc_else_token1] = ACTIONS(6571),
    [aux_sym_preproc_elif_token1] = ACTIONS(6571),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2621)] = {
    [sym_identifier] = ACTIONS(6575),
    [aux_sym_preproc_def_token1] = ACTIONS(6575),
    [aux_sym_preproc_if_token1] = ACTIONS(6575),
    [aux_sym_preproc_if_token2] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6575),
    [aux_sym_preproc_else_token1] = ACTIONS(6575),
    [aux_sym_preproc_elif_token1] = ACTIONS(6575),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6575),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6575),
    [sym_preproc_directive] = ACTIONS(6575),
    [anon_sym___extension__] = ACTIONS(6575),
    [anon_sym_extern] = ACTIONS(6575),
    [anon_sym___attribute__] = ACTIONS(6575),
    [anon_sym___attribute] = ACTIONS(6575),
    [anon_sym_noreturn] = ACTIONS(6575),
    [anon_sym_LBRACK] = ACTIONS(6577),
    [anon_sym___declspec] = ACTIONS(6575),
    [anon_sym_signed] = ACTIONS(6575),
    [anon_sym_unsigned] = ACTIONS(6575),
    [anon_sym_long] = ACTIONS(6575),
    [anon_sym_short] = ACTIONS(6575),
    [anon_sym_static] = ACTIONS(6575),
    [anon_sym_auto] = ACTIONS(6575),
    [anon_sym_register] = ACTIONS(6575),
    [anon_sym_inline] = ACTIONS(6575),
    [anon_sym___inline] = ACTIONS(6575),
    [anon_sym___inline__] = ACTIONS(6575),
    [anon_sym___forceinline] = ACTIONS(6575),
    [anon_sym_thread_local] = ACTIONS(6575),
    [anon_sym___thread] = ACTIONS(6575),
    [anon_sym_CG_EXTERN] = ACTIONS(6575),
    [anon_sym_CG_INLINE] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6575),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6575),
    [anon_sym_IBOutlet] = ACTIONS(6575),
    [anon_sym_IBInspectable] = ACTIONS(6575),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6575),
    [anon_sym_NS_INLINE] = ACTIONS(6575),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6575),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6575),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6575),
    [anon_sym_const] = ACTIONS(6575),
    [anon_sym_constexpr] = ACTIONS(6575),
    [anon_sym_volatile] = ACTIONS(6575),
    [anon_sym_restrict] = ACTIONS(6575),
    [anon_sym___restrict__] = ACTIONS(6575),
    [anon_sym__Atomic] = ACTIONS(6575),
    [anon_sym__Noreturn] = ACTIONS(6575),
    [anon_sym__Nonnull] = ACTIONS(6575),
    [anon_sym_nullable] = ACTIONS(6575),
    [anon_sym__Complex] = ACTIONS(6575),
    [anon_sym__Nullable] = ACTIONS(6575),
    [anon_sym__Nullable_result] = ACTIONS(6575),
    [anon_sym__Null_unspecified] = ACTIONS(6575),
    [anon_sym___autoreleasing] = ACTIONS(6575),
    [anon_sym___block] = ACTIONS(6575),
    [anon_sym___bridge] = ACTIONS(6575),
    [anon_sym___bridge_retained] = ACTIONS(6575),
    [anon_sym___bridge_transfer] = ACTIONS(6575),
    [anon_sym___complex] = ACTIONS(6575),
    [anon_sym___const] = ACTIONS(6575),
    [anon_sym___imag] = ACTIONS(6575),
    [anon_sym___kindof] = ACTIONS(6575),
    [anon_sym___nonnull] = ACTIONS(6575),
    [anon_sym___nullable] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6575),
    [anon_sym___real] = ACTIONS(6575),
    [anon_sym___strong] = ACTIONS(6575),
    [anon_sym___unsafe_unretained] = ACTIONS(6575),
    [anon_sym___unused] = ACTIONS(6575),
    [anon_sym___weak] = ACTIONS(6575),
    [anon_sym_alignas] = ACTIONS(6575),
    [anon_sym__Alignas] = ACTIONS(6575),
    [sym_primitive_type] = ACTIONS(6575),
    [anon_sym_enum] = ACTIONS(6575),
    [anon_sym_struct] = ACTIONS(6575),
    [anon_sym_union] = ACTIONS(6575),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6575),
    [anon_sym___typeof] = ACTIONS(6575),
    [anon_sym_typeof] = ACTIONS(6575),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6575),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6575),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6575),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_API_AVAILABLE] = ACTIONS(6575),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_API_DEPRECATED] = ACTIONS(6575),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym___deprecated_msg] = ACTIONS(6575),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6575),
    [anon_sym_BOOL] = ACTIONS(6575),
    [anon_sym_IMP] = ACTIONS(6575),
    [anon_sym_SEL] = ACTIONS(6575),
    [anon_sym_Class] = ACTIONS(6575),
    [anon_sym_id] = ACTIONS(6575),
    [anon_sym_ATdefs] = ACTIONS(6577),
  },
  [STATE(2622)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token2] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [aux_sym_preproc_else_token1] = ACTIONS(6583),
    [aux_sym_preproc_elif_token1] = ACTIONS(6583),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2623)] = {
    [sym_type_qualifier] = STATE(3402),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4551),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(3402),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(6483),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2624)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token2] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [aux_sym_preproc_else_token1] = ACTIONS(6583),
    [aux_sym_preproc_elif_token1] = ACTIONS(6583),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2625)] = {
    [sym_identifier] = ACTIONS(6587),
    [aux_sym_preproc_def_token1] = ACTIONS(6587),
    [aux_sym_preproc_if_token1] = ACTIONS(6587),
    [aux_sym_preproc_if_token2] = ACTIONS(6587),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6587),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6587),
    [aux_sym_preproc_else_token1] = ACTIONS(6587),
    [aux_sym_preproc_elif_token1] = ACTIONS(6587),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6587),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6587),
    [sym_preproc_directive] = ACTIONS(6587),
    [anon_sym___extension__] = ACTIONS(6587),
    [anon_sym_extern] = ACTIONS(6587),
    [anon_sym___attribute__] = ACTIONS(6587),
    [anon_sym___attribute] = ACTIONS(6587),
    [anon_sym_noreturn] = ACTIONS(6587),
    [anon_sym_LBRACK] = ACTIONS(6589),
    [anon_sym___declspec] = ACTIONS(6587),
    [anon_sym_signed] = ACTIONS(6587),
    [anon_sym_unsigned] = ACTIONS(6587),
    [anon_sym_long] = ACTIONS(6587),
    [anon_sym_short] = ACTIONS(6587),
    [anon_sym_static] = ACTIONS(6587),
    [anon_sym_auto] = ACTIONS(6587),
    [anon_sym_register] = ACTIONS(6587),
    [anon_sym_inline] = ACTIONS(6587),
    [anon_sym___inline] = ACTIONS(6587),
    [anon_sym___inline__] = ACTIONS(6587),
    [anon_sym___forceinline] = ACTIONS(6587),
    [anon_sym_thread_local] = ACTIONS(6587),
    [anon_sym___thread] = ACTIONS(6587),
    [anon_sym_CG_EXTERN] = ACTIONS(6587),
    [anon_sym_CG_INLINE] = ACTIONS(6587),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6587),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6587),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6587),
    [anon_sym_IBOutlet] = ACTIONS(6587),
    [anon_sym_IBInspectable] = ACTIONS(6587),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6587),
    [anon_sym_NS_INLINE] = ACTIONS(6587),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6587),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6587),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6587),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6587),
    [anon_sym_const] = ACTIONS(6587),
    [anon_sym_constexpr] = ACTIONS(6587),
    [anon_sym_volatile] = ACTIONS(6587),
    [anon_sym_restrict] = ACTIONS(6587),
    [anon_sym___restrict__] = ACTIONS(6587),
    [anon_sym__Atomic] = ACTIONS(6587),
    [anon_sym__Noreturn] = ACTIONS(6587),
    [anon_sym__Nonnull] = ACTIONS(6587),
    [anon_sym_nullable] = ACTIONS(6587),
    [anon_sym__Complex] = ACTIONS(6587),
    [anon_sym__Nullable] = ACTIONS(6587),
    [anon_sym__Nullable_result] = ACTIONS(6587),
    [anon_sym__Null_unspecified] = ACTIONS(6587),
    [anon_sym___autoreleasing] = ACTIONS(6587),
    [anon_sym___block] = ACTIONS(6587),
    [anon_sym___bridge] = ACTIONS(6587),
    [anon_sym___bridge_retained] = ACTIONS(6587),
    [anon_sym___bridge_transfer] = ACTIONS(6587),
    [anon_sym___complex] = ACTIONS(6587),
    [anon_sym___const] = ACTIONS(6587),
    [anon_sym___imag] = ACTIONS(6587),
    [anon_sym___kindof] = ACTIONS(6587),
    [anon_sym___nonnull] = ACTIONS(6587),
    [anon_sym___nullable] = ACTIONS(6587),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6587),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6587),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6587),
    [anon_sym___real] = ACTIONS(6587),
    [anon_sym___strong] = ACTIONS(6587),
    [anon_sym___unsafe_unretained] = ACTIONS(6587),
    [anon_sym___unused] = ACTIONS(6587),
    [anon_sym___weak] = ACTIONS(6587),
    [anon_sym_alignas] = ACTIONS(6587),
    [anon_sym__Alignas] = ACTIONS(6587),
    [sym_primitive_type] = ACTIONS(6587),
    [anon_sym_enum] = ACTIONS(6587),
    [anon_sym_struct] = ACTIONS(6587),
    [anon_sym_union] = ACTIONS(6587),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6587),
    [anon_sym___typeof] = ACTIONS(6587),
    [anon_sym_typeof] = ACTIONS(6587),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6587),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6587),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6587),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6587),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6587),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6587),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6587),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6587),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6587),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_API_AVAILABLE] = ACTIONS(6587),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_API_DEPRECATED] = ACTIONS(6587),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6587),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6587),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6587),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6587),
    [anon_sym___deprecated_msg] = ACTIONS(6587),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6587),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6587),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6587),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6587),
    [anon_sym_BOOL] = ACTIONS(6587),
    [anon_sym_IMP] = ACTIONS(6587),
    [anon_sym_SEL] = ACTIONS(6587),
    [anon_sym_Class] = ACTIONS(6587),
    [anon_sym_id] = ACTIONS(6587),
    [anon_sym_ATdefs] = ACTIONS(6589),
  },
  [STATE(2626)] = {
    [sym_identifier] = ACTIONS(6591),
    [aux_sym_preproc_def_token1] = ACTIONS(6591),
    [aux_sym_preproc_if_token1] = ACTIONS(6591),
    [aux_sym_preproc_if_token2] = ACTIONS(6591),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6591),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6591),
    [aux_sym_preproc_else_token1] = ACTIONS(6591),
    [aux_sym_preproc_elif_token1] = ACTIONS(6591),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6591),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6591),
    [sym_preproc_directive] = ACTIONS(6591),
    [anon_sym___extension__] = ACTIONS(6591),
    [anon_sym_extern] = ACTIONS(6591),
    [anon_sym___attribute__] = ACTIONS(6591),
    [anon_sym___attribute] = ACTIONS(6591),
    [anon_sym_noreturn] = ACTIONS(6591),
    [anon_sym_LBRACK] = ACTIONS(6593),
    [anon_sym___declspec] = ACTIONS(6591),
    [anon_sym_signed] = ACTIONS(6591),
    [anon_sym_unsigned] = ACTIONS(6591),
    [anon_sym_long] = ACTIONS(6591),
    [anon_sym_short] = ACTIONS(6591),
    [anon_sym_static] = ACTIONS(6591),
    [anon_sym_auto] = ACTIONS(6591),
    [anon_sym_register] = ACTIONS(6591),
    [anon_sym_inline] = ACTIONS(6591),
    [anon_sym___inline] = ACTIONS(6591),
    [anon_sym___inline__] = ACTIONS(6591),
    [anon_sym___forceinline] = ACTIONS(6591),
    [anon_sym_thread_local] = ACTIONS(6591),
    [anon_sym___thread] = ACTIONS(6591),
    [anon_sym_CG_EXTERN] = ACTIONS(6591),
    [anon_sym_CG_INLINE] = ACTIONS(6591),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6591),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6591),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6591),
    [anon_sym_IBOutlet] = ACTIONS(6591),
    [anon_sym_IBInspectable] = ACTIONS(6591),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6591),
    [anon_sym_NS_INLINE] = ACTIONS(6591),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6591),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6591),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6591),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6591),
    [anon_sym_const] = ACTIONS(6591),
    [anon_sym_constexpr] = ACTIONS(6591),
    [anon_sym_volatile] = ACTIONS(6591),
    [anon_sym_restrict] = ACTIONS(6591),
    [anon_sym___restrict__] = ACTIONS(6591),
    [anon_sym__Atomic] = ACTIONS(6591),
    [anon_sym__Noreturn] = ACTIONS(6591),
    [anon_sym__Nonnull] = ACTIONS(6591),
    [anon_sym_nullable] = ACTIONS(6591),
    [anon_sym__Complex] = ACTIONS(6591),
    [anon_sym__Nullable] = ACTIONS(6591),
    [anon_sym__Nullable_result] = ACTIONS(6591),
    [anon_sym__Null_unspecified] = ACTIONS(6591),
    [anon_sym___autoreleasing] = ACTIONS(6591),
    [anon_sym___block] = ACTIONS(6591),
    [anon_sym___bridge] = ACTIONS(6591),
    [anon_sym___bridge_retained] = ACTIONS(6591),
    [anon_sym___bridge_transfer] = ACTIONS(6591),
    [anon_sym___complex] = ACTIONS(6591),
    [anon_sym___const] = ACTIONS(6591),
    [anon_sym___imag] = ACTIONS(6591),
    [anon_sym___kindof] = ACTIONS(6591),
    [anon_sym___nonnull] = ACTIONS(6591),
    [anon_sym___nullable] = ACTIONS(6591),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6591),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6591),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6591),
    [anon_sym___real] = ACTIONS(6591),
    [anon_sym___strong] = ACTIONS(6591),
    [anon_sym___unsafe_unretained] = ACTIONS(6591),
    [anon_sym___unused] = ACTIONS(6591),
    [anon_sym___weak] = ACTIONS(6591),
    [anon_sym_alignas] = ACTIONS(6591),
    [anon_sym__Alignas] = ACTIONS(6591),
    [sym_primitive_type] = ACTIONS(6591),
    [anon_sym_enum] = ACTIONS(6591),
    [anon_sym_struct] = ACTIONS(6591),
    [anon_sym_union] = ACTIONS(6591),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6591),
    [anon_sym___typeof] = ACTIONS(6591),
    [anon_sym_typeof] = ACTIONS(6591),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6591),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6591),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6591),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6591),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6591),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6591),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6591),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6591),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6591),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_API_AVAILABLE] = ACTIONS(6591),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_API_DEPRECATED] = ACTIONS(6591),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6591),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6591),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6591),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6591),
    [anon_sym___deprecated_msg] = ACTIONS(6591),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6591),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6591),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6591),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6591),
    [anon_sym_BOOL] = ACTIONS(6591),
    [anon_sym_IMP] = ACTIONS(6591),
    [anon_sym_SEL] = ACTIONS(6591),
    [anon_sym_Class] = ACTIONS(6591),
    [anon_sym_id] = ACTIONS(6591),
    [anon_sym_ATdefs] = ACTIONS(6593),
  },
  [STATE(2627)] = {
    [sym_type_qualifier] = STATE(3402),
    [sym_alignas_qualifier] = STATE(3672),
    [sym_expression] = STATE(4542),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym__type_definition_type_repeat1] = STATE(3402),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6343),
    [anon_sym_noreturn] = ACTIONS(6345),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(6483),
    [anon_sym_constexpr] = ACTIONS(6345),
    [anon_sym_volatile] = ACTIONS(6345),
    [anon_sym_restrict] = ACTIONS(6345),
    [anon_sym___restrict__] = ACTIONS(6345),
    [anon_sym__Atomic] = ACTIONS(6345),
    [anon_sym__Noreturn] = ACTIONS(6345),
    [anon_sym__Nonnull] = ACTIONS(6345),
    [anon_sym_nullable] = ACTIONS(6345),
    [anon_sym__Complex] = ACTIONS(6345),
    [anon_sym__Nullable] = ACTIONS(6345),
    [anon_sym__Nullable_result] = ACTIONS(6345),
    [anon_sym__Null_unspecified] = ACTIONS(6345),
    [anon_sym___autoreleasing] = ACTIONS(6345),
    [anon_sym___block] = ACTIONS(6345),
    [anon_sym___bridge] = ACTIONS(6345),
    [anon_sym___bridge_retained] = ACTIONS(6345),
    [anon_sym___bridge_transfer] = ACTIONS(6345),
    [anon_sym___complex] = ACTIONS(6345),
    [anon_sym___const] = ACTIONS(6345),
    [anon_sym___imag] = ACTIONS(6353),
    [anon_sym___kindof] = ACTIONS(6345),
    [anon_sym___nonnull] = ACTIONS(6345),
    [anon_sym___nullable] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6345),
    [anon_sym___real] = ACTIONS(6353),
    [anon_sym___strong] = ACTIONS(6345),
    [anon_sym___unsafe_unretained] = ACTIONS(6345),
    [anon_sym___unused] = ACTIONS(6345),
    [anon_sym___weak] = ACTIONS(6345),
    [anon_sym_alignas] = ACTIONS(6355),
    [anon_sym__Alignas] = ACTIONS(6355),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2628)] = {
    [sym_identifier] = ACTIONS(6595),
    [aux_sym_preproc_def_token1] = ACTIONS(6595),
    [aux_sym_preproc_if_token1] = ACTIONS(6595),
    [aux_sym_preproc_if_token2] = ACTIONS(6595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6595),
    [aux_sym_preproc_else_token1] = ACTIONS(6595),
    [aux_sym_preproc_elif_token1] = ACTIONS(6595),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6595),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6595),
    [sym_preproc_directive] = ACTIONS(6595),
    [anon_sym___extension__] = ACTIONS(6595),
    [anon_sym_extern] = ACTIONS(6595),
    [anon_sym___attribute__] = ACTIONS(6595),
    [anon_sym___attribute] = ACTIONS(6595),
    [anon_sym_noreturn] = ACTIONS(6595),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym___declspec] = ACTIONS(6595),
    [anon_sym_signed] = ACTIONS(6595),
    [anon_sym_unsigned] = ACTIONS(6595),
    [anon_sym_long] = ACTIONS(6595),
    [anon_sym_short] = ACTIONS(6595),
    [anon_sym_static] = ACTIONS(6595),
    [anon_sym_auto] = ACTIONS(6595),
    [anon_sym_register] = ACTIONS(6595),
    [anon_sym_inline] = ACTIONS(6595),
    [anon_sym___inline] = ACTIONS(6595),
    [anon_sym___inline__] = ACTIONS(6595),
    [anon_sym___forceinline] = ACTIONS(6595),
    [anon_sym_thread_local] = ACTIONS(6595),
    [anon_sym___thread] = ACTIONS(6595),
    [anon_sym_CG_EXTERN] = ACTIONS(6595),
    [anon_sym_CG_INLINE] = ACTIONS(6595),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6595),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6595),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6595),
    [anon_sym_IBOutlet] = ACTIONS(6595),
    [anon_sym_IBInspectable] = ACTIONS(6595),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6595),
    [anon_sym_NS_INLINE] = ACTIONS(6595),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6595),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6595),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6595),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6595),
    [anon_sym_const] = ACTIONS(6595),
    [anon_sym_constexpr] = ACTIONS(6595),
    [anon_sym_volatile] = ACTIONS(6595),
    [anon_sym_restrict] = ACTIONS(6595),
    [anon_sym___restrict__] = ACTIONS(6595),
    [anon_sym__Atomic] = ACTIONS(6595),
    [anon_sym__Noreturn] = ACTIONS(6595),
    [anon_sym__Nonnull] = ACTIONS(6595),
    [anon_sym_nullable] = ACTIONS(6595),
    [anon_sym__Complex] = ACTIONS(6595),
    [anon_sym__Nullable] = ACTIONS(6595),
    [anon_sym__Nullable_result] = ACTIONS(6595),
    [anon_sym__Null_unspecified] = ACTIONS(6595),
    [anon_sym___autoreleasing] = ACTIONS(6595),
    [anon_sym___block] = ACTIONS(6595),
    [anon_sym___bridge] = ACTIONS(6595),
    [anon_sym___bridge_retained] = ACTIONS(6595),
    [anon_sym___bridge_transfer] = ACTIONS(6595),
    [anon_sym___complex] = ACTIONS(6595),
    [anon_sym___const] = ACTIONS(6595),
    [anon_sym___imag] = ACTIONS(6595),
    [anon_sym___kindof] = ACTIONS(6595),
    [anon_sym___nonnull] = ACTIONS(6595),
    [anon_sym___nullable] = ACTIONS(6595),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6595),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6595),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6595),
    [anon_sym___real] = ACTIONS(6595),
    [anon_sym___strong] = ACTIONS(6595),
    [anon_sym___unsafe_unretained] = ACTIONS(6595),
    [anon_sym___unused] = ACTIONS(6595),
    [anon_sym___weak] = ACTIONS(6595),
    [anon_sym_alignas] = ACTIONS(6595),
    [anon_sym__Alignas] = ACTIONS(6595),
    [sym_primitive_type] = ACTIONS(6595),
    [anon_sym_enum] = ACTIONS(6595),
    [anon_sym_struct] = ACTIONS(6595),
    [anon_sym_union] = ACTIONS(6595),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6595),
    [anon_sym___typeof] = ACTIONS(6595),
    [anon_sym_typeof] = ACTIONS(6595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6595),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6595),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6595),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6595),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6595),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6595),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6595),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6595),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6595),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_API_AVAILABLE] = ACTIONS(6595),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_API_DEPRECATED] = ACTIONS(6595),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6595),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6595),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6595),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6595),
    [anon_sym___deprecated_msg] = ACTIONS(6595),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6595),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6595),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6595),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6595),
    [anon_sym_BOOL] = ACTIONS(6595),
    [anon_sym_IMP] = ACTIONS(6595),
    [anon_sym_SEL] = ACTIONS(6595),
    [anon_sym_Class] = ACTIONS(6595),
    [anon_sym_id] = ACTIONS(6595),
    [anon_sym_ATdefs] = ACTIONS(6597),
  },
  [STATE(2629)] = {
    [sym_identifier] = ACTIONS(6599),
    [aux_sym_preproc_def_token1] = ACTIONS(6599),
    [aux_sym_preproc_if_token1] = ACTIONS(6599),
    [aux_sym_preproc_if_token2] = ACTIONS(6599),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6599),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6599),
    [aux_sym_preproc_else_token1] = ACTIONS(6599),
    [aux_sym_preproc_elif_token1] = ACTIONS(6599),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6599),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6599),
    [sym_preproc_directive] = ACTIONS(6599),
    [anon_sym___extension__] = ACTIONS(6599),
    [anon_sym_extern] = ACTIONS(6599),
    [anon_sym___attribute__] = ACTIONS(6599),
    [anon_sym___attribute] = ACTIONS(6599),
    [anon_sym_noreturn] = ACTIONS(6599),
    [anon_sym_LBRACK] = ACTIONS(6601),
    [anon_sym___declspec] = ACTIONS(6599),
    [anon_sym_signed] = ACTIONS(6599),
    [anon_sym_unsigned] = ACTIONS(6599),
    [anon_sym_long] = ACTIONS(6599),
    [anon_sym_short] = ACTIONS(6599),
    [anon_sym_static] = ACTIONS(6599),
    [anon_sym_auto] = ACTIONS(6599),
    [anon_sym_register] = ACTIONS(6599),
    [anon_sym_inline] = ACTIONS(6599),
    [anon_sym___inline] = ACTIONS(6599),
    [anon_sym___inline__] = ACTIONS(6599),
    [anon_sym___forceinline] = ACTIONS(6599),
    [anon_sym_thread_local] = ACTIONS(6599),
    [anon_sym___thread] = ACTIONS(6599),
    [anon_sym_CG_EXTERN] = ACTIONS(6599),
    [anon_sym_CG_INLINE] = ACTIONS(6599),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6599),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6599),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6599),
    [anon_sym_IBOutlet] = ACTIONS(6599),
    [anon_sym_IBInspectable] = ACTIONS(6599),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6599),
    [anon_sym_NS_INLINE] = ACTIONS(6599),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6599),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6599),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6599),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6599),
    [anon_sym_const] = ACTIONS(6599),
    [anon_sym_constexpr] = ACTIONS(6599),
    [anon_sym_volatile] = ACTIONS(6599),
    [anon_sym_restrict] = ACTIONS(6599),
    [anon_sym___restrict__] = ACTIONS(6599),
    [anon_sym__Atomic] = ACTIONS(6599),
    [anon_sym__Noreturn] = ACTIONS(6599),
    [anon_sym__Nonnull] = ACTIONS(6599),
    [anon_sym_nullable] = ACTIONS(6599),
    [anon_sym__Complex] = ACTIONS(6599),
    [anon_sym__Nullable] = ACTIONS(6599),
    [anon_sym__Nullable_result] = ACTIONS(6599),
    [anon_sym__Null_unspecified] = ACTIONS(6599),
    [anon_sym___autoreleasing] = ACTIONS(6599),
    [anon_sym___block] = ACTIONS(6599),
    [anon_sym___bridge] = ACTIONS(6599),
    [anon_sym___bridge_retained] = ACTIONS(6599),
    [anon_sym___bridge_transfer] = ACTIONS(6599),
    [anon_sym___complex] = ACTIONS(6599),
    [anon_sym___const] = ACTIONS(6599),
    [anon_sym___imag] = ACTIONS(6599),
    [anon_sym___kindof] = ACTIONS(6599),
    [anon_sym___nonnull] = ACTIONS(6599),
    [anon_sym___nullable] = ACTIONS(6599),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6599),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6599),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6599),
    [anon_sym___real] = ACTIONS(6599),
    [anon_sym___strong] = ACTIONS(6599),
    [anon_sym___unsafe_unretained] = ACTIONS(6599),
    [anon_sym___unused] = ACTIONS(6599),
    [anon_sym___weak] = ACTIONS(6599),
    [anon_sym_alignas] = ACTIONS(6599),
    [anon_sym__Alignas] = ACTIONS(6599),
    [sym_primitive_type] = ACTIONS(6599),
    [anon_sym_enum] = ACTIONS(6599),
    [anon_sym_struct] = ACTIONS(6599),
    [anon_sym_union] = ACTIONS(6599),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6599),
    [anon_sym___typeof] = ACTIONS(6599),
    [anon_sym_typeof] = ACTIONS(6599),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6599),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6599),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6599),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6599),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6599),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6599),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6599),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6599),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6599),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_API_AVAILABLE] = ACTIONS(6599),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_API_DEPRECATED] = ACTIONS(6599),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6599),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6599),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6599),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6599),
    [anon_sym___deprecated_msg] = ACTIONS(6599),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6599),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6599),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6599),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6599),
    [anon_sym_BOOL] = ACTIONS(6599),
    [anon_sym_IMP] = ACTIONS(6599),
    [anon_sym_SEL] = ACTIONS(6599),
    [anon_sym_Class] = ACTIONS(6599),
    [anon_sym_id] = ACTIONS(6599),
    [anon_sym_ATdefs] = ACTIONS(6601),
  },
  [STATE(2630)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6491),
    [anon_sym_COMMA] = ACTIONS(4468),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(6499),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4468),
    [anon_sym___extension__] = ACTIONS(4468),
    [anon_sym_extern] = ACTIONS(4468),
    [anon_sym___attribute__] = ACTIONS(4468),
    [anon_sym___attribute] = ACTIONS(4466),
    [anon_sym_noreturn] = ACTIONS(4468),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4468),
    [anon_sym_static] = ACTIONS(4468),
    [anon_sym_auto] = ACTIONS(4468),
    [anon_sym_register] = ACTIONS(4468),
    [anon_sym_inline] = ACTIONS(4468),
    [anon_sym___inline] = ACTIONS(4466),
    [anon_sym___inline__] = ACTIONS(4468),
    [anon_sym___forceinline] = ACTIONS(4468),
    [anon_sym_thread_local] = ACTIONS(4468),
    [anon_sym___thread] = ACTIONS(4468),
    [anon_sym_CG_EXTERN] = ACTIONS(4468),
    [anon_sym_CG_INLINE] = ACTIONS(4468),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4468),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4468),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4468),
    [anon_sym_IBOutlet] = ACTIONS(4468),
    [anon_sym_IBInspectable] = ACTIONS(4468),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4468),
    [anon_sym_NS_INLINE] = ACTIONS(4468),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4468),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4468),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4468),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4468),
    [anon_sym_const] = ACTIONS(4466),
    [anon_sym_constexpr] = ACTIONS(4468),
    [anon_sym_volatile] = ACTIONS(4468),
    [anon_sym_restrict] = ACTIONS(4468),
    [anon_sym___restrict__] = ACTIONS(4468),
    [anon_sym__Atomic] = ACTIONS(4468),
    [anon_sym__Noreturn] = ACTIONS(4468),
    [anon_sym__Nonnull] = ACTIONS(4468),
    [anon_sym_nullable] = ACTIONS(4468),
    [anon_sym__Complex] = ACTIONS(4468),
    [anon_sym__Nullable] = ACTIONS(4466),
    [anon_sym__Nullable_result] = ACTIONS(4468),
    [anon_sym__Null_unspecified] = ACTIONS(4468),
    [anon_sym___autoreleasing] = ACTIONS(4468),
    [anon_sym___block] = ACTIONS(4468),
    [anon_sym___bridge] = ACTIONS(4466),
    [anon_sym___bridge_retained] = ACTIONS(4468),
    [anon_sym___bridge_transfer] = ACTIONS(4468),
    [anon_sym___complex] = ACTIONS(4468),
    [anon_sym___const] = ACTIONS(4468),
    [anon_sym___imag] = ACTIONS(4468),
    [anon_sym___kindof] = ACTIONS(4468),
    [anon_sym___nonnull] = ACTIONS(4468),
    [anon_sym___nullable] = ACTIONS(4468),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4468),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4468),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4468),
    [anon_sym___real] = ACTIONS(4468),
    [anon_sym___strong] = ACTIONS(4468),
    [anon_sym___unsafe_unretained] = ACTIONS(4468),
    [anon_sym___unused] = ACTIONS(4468),
    [anon_sym___weak] = ACTIONS(4468),
    [anon_sym_alignas] = ACTIONS(4468),
    [anon_sym__Alignas] = ACTIONS(4468),
    [anon_sym_QMARK] = ACTIONS(6517),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4468),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4468),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4468),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4468),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4468),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4468),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4468),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4468),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4468),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4468),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4466),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4468),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4468),
    [anon_sym_API_AVAILABLE] = ACTIONS(4468),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4468),
    [anon_sym_API_DEPRECATED] = ACTIONS(4468),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4468),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4468),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4468),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4468),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4468),
    [anon_sym___deprecated_msg] = ACTIONS(4468),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4468),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4468),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4468),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4468),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4468),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4468),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4468),
  },
  [STATE(2631)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6491),
    [anon_sym_COMMA] = ACTIONS(4472),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(6499),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4472),
    [anon_sym___extension__] = ACTIONS(4472),
    [anon_sym_extern] = ACTIONS(4472),
    [anon_sym___attribute__] = ACTIONS(4472),
    [anon_sym___attribute] = ACTIONS(4470),
    [anon_sym_noreturn] = ACTIONS(4472),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4472),
    [anon_sym_static] = ACTIONS(4472),
    [anon_sym_auto] = ACTIONS(4472),
    [anon_sym_register] = ACTIONS(4472),
    [anon_sym_inline] = ACTIONS(4472),
    [anon_sym___inline] = ACTIONS(4470),
    [anon_sym___inline__] = ACTIONS(4472),
    [anon_sym___forceinline] = ACTIONS(4472),
    [anon_sym_thread_local] = ACTIONS(4472),
    [anon_sym___thread] = ACTIONS(4472),
    [anon_sym_CG_EXTERN] = ACTIONS(4472),
    [anon_sym_CG_INLINE] = ACTIONS(4472),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4472),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4472),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4472),
    [anon_sym_IBOutlet] = ACTIONS(4472),
    [anon_sym_IBInspectable] = ACTIONS(4472),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4472),
    [anon_sym_NS_INLINE] = ACTIONS(4472),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4472),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4472),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4472),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4472),
    [anon_sym_const] = ACTIONS(4470),
    [anon_sym_constexpr] = ACTIONS(4472),
    [anon_sym_volatile] = ACTIONS(4472),
    [anon_sym_restrict] = ACTIONS(4472),
    [anon_sym___restrict__] = ACTIONS(4472),
    [anon_sym__Atomic] = ACTIONS(4472),
    [anon_sym__Noreturn] = ACTIONS(4472),
    [anon_sym__Nonnull] = ACTIONS(4472),
    [anon_sym_nullable] = ACTIONS(4472),
    [anon_sym__Complex] = ACTIONS(4472),
    [anon_sym__Nullable] = ACTIONS(4470),
    [anon_sym__Nullable_result] = ACTIONS(4472),
    [anon_sym__Null_unspecified] = ACTIONS(4472),
    [anon_sym___autoreleasing] = ACTIONS(4472),
    [anon_sym___block] = ACTIONS(4472),
    [anon_sym___bridge] = ACTIONS(4470),
    [anon_sym___bridge_retained] = ACTIONS(4472),
    [anon_sym___bridge_transfer] = ACTIONS(4472),
    [anon_sym___complex] = ACTIONS(4472),
    [anon_sym___const] = ACTIONS(4472),
    [anon_sym___imag] = ACTIONS(4472),
    [anon_sym___kindof] = ACTIONS(4472),
    [anon_sym___nonnull] = ACTIONS(4472),
    [anon_sym___nullable] = ACTIONS(4472),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4472),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4472),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4472),
    [anon_sym___real] = ACTIONS(4472),
    [anon_sym___strong] = ACTIONS(4472),
    [anon_sym___unsafe_unretained] = ACTIONS(4472),
    [anon_sym___unused] = ACTIONS(4472),
    [anon_sym___weak] = ACTIONS(4472),
    [anon_sym_alignas] = ACTIONS(4472),
    [anon_sym__Alignas] = ACTIONS(4472),
    [anon_sym_QMARK] = ACTIONS(4472),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4472),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4472),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4472),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4472),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4472),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4472),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4472),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4472),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4472),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4472),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4470),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4472),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4472),
    [anon_sym_API_AVAILABLE] = ACTIONS(4472),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4472),
    [anon_sym_API_DEPRECATED] = ACTIONS(4472),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4472),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4472),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4472),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4472),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4472),
    [anon_sym___deprecated_msg] = ACTIONS(4472),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4472),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4472),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4472),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4472),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4472),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4472),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4472),
  },
  [STATE(2632)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(4000),
    [anon_sym_PLUS] = ACTIONS(4000),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4002),
    [anon_sym_GT_GT] = ACTIONS(4002),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2633)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2634)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2635)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2636)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2637)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2638)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2639)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2640)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4002),
    [anon_sym_COMMA] = ACTIONS(4002),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(4002),
    [anon_sym_AMP_AMP] = ACTIONS(4002),
    [anon_sym_PIPE] = ACTIONS(4000),
    [anon_sym_CARET] = ACTIONS(4002),
    [anon_sym_AMP] = ACTIONS(4000),
    [anon_sym_EQ_EQ] = ACTIONS(4002),
    [anon_sym_BANG_EQ] = ACTIONS(4002),
    [anon_sym_GT] = ACTIONS(4000),
    [anon_sym_GT_EQ] = ACTIONS(4002),
    [anon_sym_LT_EQ] = ACTIONS(4002),
    [anon_sym_LT] = ACTIONS(4000),
    [anon_sym_LT_LT] = ACTIONS(4002),
    [anon_sym_GT_GT] = ACTIONS(4002),
    [anon_sym_SEMI] = ACTIONS(4002),
    [anon_sym___extension__] = ACTIONS(4002),
    [anon_sym_extern] = ACTIONS(4002),
    [anon_sym___attribute__] = ACTIONS(4002),
    [anon_sym___attribute] = ACTIONS(4000),
    [anon_sym_noreturn] = ACTIONS(4002),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4002),
    [anon_sym_static] = ACTIONS(4002),
    [anon_sym_auto] = ACTIONS(4002),
    [anon_sym_register] = ACTIONS(4002),
    [anon_sym_inline] = ACTIONS(4002),
    [anon_sym___inline] = ACTIONS(4000),
    [anon_sym___inline__] = ACTIONS(4002),
    [anon_sym___forceinline] = ACTIONS(4002),
    [anon_sym_thread_local] = ACTIONS(4002),
    [anon_sym___thread] = ACTIONS(4002),
    [anon_sym_CG_EXTERN] = ACTIONS(4002),
    [anon_sym_CG_INLINE] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4002),
    [anon_sym_IBOutlet] = ACTIONS(4002),
    [anon_sym_IBInspectable] = ACTIONS(4002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4002),
    [anon_sym_NS_INLINE] = ACTIONS(4002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4002),
    [anon_sym_const] = ACTIONS(4000),
    [anon_sym_constexpr] = ACTIONS(4002),
    [anon_sym_volatile] = ACTIONS(4002),
    [anon_sym_restrict] = ACTIONS(4002),
    [anon_sym___restrict__] = ACTIONS(4002),
    [anon_sym__Atomic] = ACTIONS(4002),
    [anon_sym__Noreturn] = ACTIONS(4002),
    [anon_sym__Nonnull] = ACTIONS(4002),
    [anon_sym_nullable] = ACTIONS(4002),
    [anon_sym__Complex] = ACTIONS(4002),
    [anon_sym__Nullable] = ACTIONS(4000),
    [anon_sym__Nullable_result] = ACTIONS(4002),
    [anon_sym__Null_unspecified] = ACTIONS(4002),
    [anon_sym___autoreleasing] = ACTIONS(4002),
    [anon_sym___block] = ACTIONS(4002),
    [anon_sym___bridge] = ACTIONS(4000),
    [anon_sym___bridge_retained] = ACTIONS(4002),
    [anon_sym___bridge_transfer] = ACTIONS(4002),
    [anon_sym___complex] = ACTIONS(4002),
    [anon_sym___const] = ACTIONS(4002),
    [anon_sym___imag] = ACTIONS(4002),
    [anon_sym___kindof] = ACTIONS(4002),
    [anon_sym___nonnull] = ACTIONS(4002),
    [anon_sym___nullable] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4002),
    [anon_sym___real] = ACTIONS(4002),
    [anon_sym___strong] = ACTIONS(4002),
    [anon_sym___unsafe_unretained] = ACTIONS(4002),
    [anon_sym___unused] = ACTIONS(4002),
    [anon_sym___weak] = ACTIONS(4002),
    [anon_sym_alignas] = ACTIONS(4002),
    [anon_sym__Alignas] = ACTIONS(4002),
    [anon_sym_QMARK] = ACTIONS(4002),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4000),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_API_AVAILABLE] = ACTIONS(4002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_API_DEPRECATED] = ACTIONS(4002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4002),
    [anon_sym___deprecated_msg] = ACTIONS(4002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4002),
  },
  [STATE(2641)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6491),
    [anon_sym_COMMA] = ACTIONS(4480),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(6499),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4480),
    [anon_sym___extension__] = ACTIONS(4480),
    [anon_sym_extern] = ACTIONS(4480),
    [anon_sym___attribute__] = ACTIONS(4480),
    [anon_sym___attribute] = ACTIONS(4478),
    [anon_sym_noreturn] = ACTIONS(4480),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4480),
    [anon_sym_static] = ACTIONS(4480),
    [anon_sym_auto] = ACTIONS(4480),
    [anon_sym_register] = ACTIONS(4480),
    [anon_sym_inline] = ACTIONS(4480),
    [anon_sym___inline] = ACTIONS(4478),
    [anon_sym___inline__] = ACTIONS(4480),
    [anon_sym___forceinline] = ACTIONS(4480),
    [anon_sym_thread_local] = ACTIONS(4480),
    [anon_sym___thread] = ACTIONS(4480),
    [anon_sym_CG_EXTERN] = ACTIONS(4480),
    [anon_sym_CG_INLINE] = ACTIONS(4480),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4480),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4480),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4480),
    [anon_sym_IBOutlet] = ACTIONS(4480),
    [anon_sym_IBInspectable] = ACTIONS(4480),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4480),
    [anon_sym_NS_INLINE] = ACTIONS(4480),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4480),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4480),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4480),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4480),
    [anon_sym_const] = ACTIONS(4478),
    [anon_sym_constexpr] = ACTIONS(4480),
    [anon_sym_volatile] = ACTIONS(4480),
    [anon_sym_restrict] = ACTIONS(4480),
    [anon_sym___restrict__] = ACTIONS(4480),
    [anon_sym__Atomic] = ACTIONS(4480),
    [anon_sym__Noreturn] = ACTIONS(4480),
    [anon_sym__Nonnull] = ACTIONS(4480),
    [anon_sym_nullable] = ACTIONS(4480),
    [anon_sym__Complex] = ACTIONS(4480),
    [anon_sym__Nullable] = ACTIONS(4478),
    [anon_sym__Nullable_result] = ACTIONS(4480),
    [anon_sym__Null_unspecified] = ACTIONS(4480),
    [anon_sym___autoreleasing] = ACTIONS(4480),
    [anon_sym___block] = ACTIONS(4480),
    [anon_sym___bridge] = ACTIONS(4478),
    [anon_sym___bridge_retained] = ACTIONS(4480),
    [anon_sym___bridge_transfer] = ACTIONS(4480),
    [anon_sym___complex] = ACTIONS(4480),
    [anon_sym___const] = ACTIONS(4480),
    [anon_sym___imag] = ACTIONS(4480),
    [anon_sym___kindof] = ACTIONS(4480),
    [anon_sym___nonnull] = ACTIONS(4480),
    [anon_sym___nullable] = ACTIONS(4480),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4480),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4480),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4480),
    [anon_sym___real] = ACTIONS(4480),
    [anon_sym___strong] = ACTIONS(4480),
    [anon_sym___unsafe_unretained] = ACTIONS(4480),
    [anon_sym___unused] = ACTIONS(4480),
    [anon_sym___weak] = ACTIONS(4480),
    [anon_sym_alignas] = ACTIONS(4480),
    [anon_sym__Alignas] = ACTIONS(4480),
    [anon_sym_QMARK] = ACTIONS(6517),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4480),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4480),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4480),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4480),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4480),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4480),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4480),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4480),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4480),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4480),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4478),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4480),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4480),
    [anon_sym_API_AVAILABLE] = ACTIONS(4480),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4480),
    [anon_sym_API_DEPRECATED] = ACTIONS(4480),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4480),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4480),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4480),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4480),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4480),
    [anon_sym___deprecated_msg] = ACTIONS(4480),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4480),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4480),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4480),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4480),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4480),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4480),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4480),
  },
  [STATE(2642)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6491),
    [anon_sym_COMMA] = ACTIONS(4462),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(6499),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4462),
    [anon_sym___extension__] = ACTIONS(4462),
    [anon_sym_extern] = ACTIONS(4462),
    [anon_sym___attribute__] = ACTIONS(4462),
    [anon_sym___attribute] = ACTIONS(4458),
    [anon_sym_noreturn] = ACTIONS(4462),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4462),
    [anon_sym_static] = ACTIONS(4462),
    [anon_sym_auto] = ACTIONS(4462),
    [anon_sym_register] = ACTIONS(4462),
    [anon_sym_inline] = ACTIONS(4462),
    [anon_sym___inline] = ACTIONS(4458),
    [anon_sym___inline__] = ACTIONS(4462),
    [anon_sym___forceinline] = ACTIONS(4462),
    [anon_sym_thread_local] = ACTIONS(4462),
    [anon_sym___thread] = ACTIONS(4462),
    [anon_sym_CG_EXTERN] = ACTIONS(4462),
    [anon_sym_CG_INLINE] = ACTIONS(4462),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4462),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4462),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4462),
    [anon_sym_IBOutlet] = ACTIONS(4462),
    [anon_sym_IBInspectable] = ACTIONS(4462),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4462),
    [anon_sym_NS_INLINE] = ACTIONS(4462),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4462),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4462),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4462),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4462),
    [anon_sym_const] = ACTIONS(4458),
    [anon_sym_constexpr] = ACTIONS(4462),
    [anon_sym_volatile] = ACTIONS(4462),
    [anon_sym_restrict] = ACTIONS(4462),
    [anon_sym___restrict__] = ACTIONS(4462),
    [anon_sym__Atomic] = ACTIONS(4462),
    [anon_sym__Noreturn] = ACTIONS(4462),
    [anon_sym__Nonnull] = ACTIONS(4462),
    [anon_sym_nullable] = ACTIONS(4462),
    [anon_sym__Complex] = ACTIONS(4462),
    [anon_sym__Nullable] = ACTIONS(4458),
    [anon_sym__Nullable_result] = ACTIONS(4462),
    [anon_sym__Null_unspecified] = ACTIONS(4462),
    [anon_sym___autoreleasing] = ACTIONS(4462),
    [anon_sym___block] = ACTIONS(4462),
    [anon_sym___bridge] = ACTIONS(4458),
    [anon_sym___bridge_retained] = ACTIONS(4462),
    [anon_sym___bridge_transfer] = ACTIONS(4462),
    [anon_sym___complex] = ACTIONS(4462),
    [anon_sym___const] = ACTIONS(4462),
    [anon_sym___imag] = ACTIONS(4462),
    [anon_sym___kindof] = ACTIONS(4462),
    [anon_sym___nonnull] = ACTIONS(4462),
    [anon_sym___nullable] = ACTIONS(4462),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4462),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4462),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4462),
    [anon_sym___real] = ACTIONS(4462),
    [anon_sym___strong] = ACTIONS(4462),
    [anon_sym___unsafe_unretained] = ACTIONS(4462),
    [anon_sym___unused] = ACTIONS(4462),
    [anon_sym___weak] = ACTIONS(4462),
    [anon_sym_alignas] = ACTIONS(4462),
    [anon_sym__Alignas] = ACTIONS(4462),
    [anon_sym_QMARK] = ACTIONS(6517),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4462),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4462),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4462),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4462),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4462),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4462),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4462),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4462),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4462),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4462),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4458),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4462),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4462),
    [anon_sym_API_AVAILABLE] = ACTIONS(4462),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4462),
    [anon_sym_API_DEPRECATED] = ACTIONS(4462),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4462),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4462),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4462),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4462),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4462),
    [anon_sym___deprecated_msg] = ACTIONS(4462),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4462),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4462),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4462),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4462),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4462),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4462),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4462),
  },
  [STATE(2643)] = {
    [sym_identifier] = ACTIONS(6603),
    [aux_sym_preproc_def_token1] = ACTIONS(6603),
    [aux_sym_preproc_if_token1] = ACTIONS(6603),
    [aux_sym_preproc_if_token2] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6603),
    [aux_sym_preproc_else_token1] = ACTIONS(6603),
    [aux_sym_preproc_elif_token1] = ACTIONS(6603),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6603),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6603),
    [sym_preproc_directive] = ACTIONS(6603),
    [anon_sym___extension__] = ACTIONS(6603),
    [anon_sym_extern] = ACTIONS(6603),
    [anon_sym___attribute__] = ACTIONS(6603),
    [anon_sym___attribute] = ACTIONS(6603),
    [anon_sym_noreturn] = ACTIONS(6603),
    [anon_sym_LBRACK] = ACTIONS(6605),
    [anon_sym___declspec] = ACTIONS(6603),
    [anon_sym_signed] = ACTIONS(6603),
    [anon_sym_unsigned] = ACTIONS(6603),
    [anon_sym_long] = ACTIONS(6603),
    [anon_sym_short] = ACTIONS(6603),
    [anon_sym_static] = ACTIONS(6603),
    [anon_sym_auto] = ACTIONS(6603),
    [anon_sym_register] = ACTIONS(6603),
    [anon_sym_inline] = ACTIONS(6603),
    [anon_sym___inline] = ACTIONS(6603),
    [anon_sym___inline__] = ACTIONS(6603),
    [anon_sym___forceinline] = ACTIONS(6603),
    [anon_sym_thread_local] = ACTIONS(6603),
    [anon_sym___thread] = ACTIONS(6603),
    [anon_sym_CG_EXTERN] = ACTIONS(6603),
    [anon_sym_CG_INLINE] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6603),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6603),
    [anon_sym_IBOutlet] = ACTIONS(6603),
    [anon_sym_IBInspectable] = ACTIONS(6603),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6603),
    [anon_sym_NS_INLINE] = ACTIONS(6603),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6603),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6603),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6603),
    [anon_sym_const] = ACTIONS(6603),
    [anon_sym_constexpr] = ACTIONS(6603),
    [anon_sym_volatile] = ACTIONS(6603),
    [anon_sym_restrict] = ACTIONS(6603),
    [anon_sym___restrict__] = ACTIONS(6603),
    [anon_sym__Atomic] = ACTIONS(6603),
    [anon_sym__Noreturn] = ACTIONS(6603),
    [anon_sym__Nonnull] = ACTIONS(6603),
    [anon_sym_nullable] = ACTIONS(6603),
    [anon_sym__Complex] = ACTIONS(6603),
    [anon_sym__Nullable] = ACTIONS(6603),
    [anon_sym__Nullable_result] = ACTIONS(6603),
    [anon_sym__Null_unspecified] = ACTIONS(6603),
    [anon_sym___autoreleasing] = ACTIONS(6603),
    [anon_sym___block] = ACTIONS(6603),
    [anon_sym___bridge] = ACTIONS(6603),
    [anon_sym___bridge_retained] = ACTIONS(6603),
    [anon_sym___bridge_transfer] = ACTIONS(6603),
    [anon_sym___complex] = ACTIONS(6603),
    [anon_sym___const] = ACTIONS(6603),
    [anon_sym___imag] = ACTIONS(6603),
    [anon_sym___kindof] = ACTIONS(6603),
    [anon_sym___nonnull] = ACTIONS(6603),
    [anon_sym___nullable] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6603),
    [anon_sym___real] = ACTIONS(6603),
    [anon_sym___strong] = ACTIONS(6603),
    [anon_sym___unsafe_unretained] = ACTIONS(6603),
    [anon_sym___unused] = ACTIONS(6603),
    [anon_sym___weak] = ACTIONS(6603),
    [anon_sym_alignas] = ACTIONS(6603),
    [anon_sym__Alignas] = ACTIONS(6603),
    [sym_primitive_type] = ACTIONS(6603),
    [anon_sym_enum] = ACTIONS(6603),
    [anon_sym_struct] = ACTIONS(6603),
    [anon_sym_union] = ACTIONS(6603),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6603),
    [anon_sym___typeof] = ACTIONS(6603),
    [anon_sym_typeof] = ACTIONS(6603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6603),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6603),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6603),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_API_AVAILABLE] = ACTIONS(6603),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_API_DEPRECATED] = ACTIONS(6603),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym___deprecated_msg] = ACTIONS(6603),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6603),
    [anon_sym_BOOL] = ACTIONS(6603),
    [anon_sym_IMP] = ACTIONS(6603),
    [anon_sym_SEL] = ACTIONS(6603),
    [anon_sym_Class] = ACTIONS(6603),
    [anon_sym_id] = ACTIONS(6603),
    [anon_sym_ATdefs] = ACTIONS(6605),
  },
  [STATE(2644)] = {
    [sym_identifier] = ACTIONS(6519),
    [aux_sym_preproc_def_token1] = ACTIONS(6519),
    [aux_sym_preproc_if_token1] = ACTIONS(6519),
    [aux_sym_preproc_if_token2] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6519),
    [aux_sym_preproc_else_token1] = ACTIONS(6519),
    [aux_sym_preproc_elif_token1] = ACTIONS(6519),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6519),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6519),
    [sym_preproc_directive] = ACTIONS(6519),
    [anon_sym___extension__] = ACTIONS(6519),
    [anon_sym_extern] = ACTIONS(6519),
    [anon_sym___attribute__] = ACTIONS(6519),
    [anon_sym___attribute] = ACTIONS(6519),
    [anon_sym_noreturn] = ACTIONS(6519),
    [anon_sym_LBRACK] = ACTIONS(6521),
    [anon_sym___declspec] = ACTIONS(6519),
    [anon_sym_signed] = ACTIONS(6519),
    [anon_sym_unsigned] = ACTIONS(6519),
    [anon_sym_long] = ACTIONS(6519),
    [anon_sym_short] = ACTIONS(6519),
    [anon_sym_static] = ACTIONS(6519),
    [anon_sym_auto] = ACTIONS(6519),
    [anon_sym_register] = ACTIONS(6519),
    [anon_sym_inline] = ACTIONS(6519),
    [anon_sym___inline] = ACTIONS(6519),
    [anon_sym___inline__] = ACTIONS(6519),
    [anon_sym___forceinline] = ACTIONS(6519),
    [anon_sym_thread_local] = ACTIONS(6519),
    [anon_sym___thread] = ACTIONS(6519),
    [anon_sym_CG_EXTERN] = ACTIONS(6519),
    [anon_sym_CG_INLINE] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6519),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6519),
    [anon_sym_IBOutlet] = ACTIONS(6519),
    [anon_sym_IBInspectable] = ACTIONS(6519),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6519),
    [anon_sym_NS_INLINE] = ACTIONS(6519),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6519),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6519),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6519),
    [anon_sym_const] = ACTIONS(6519),
    [anon_sym_constexpr] = ACTIONS(6519),
    [anon_sym_volatile] = ACTIONS(6519),
    [anon_sym_restrict] = ACTIONS(6519),
    [anon_sym___restrict__] = ACTIONS(6519),
    [anon_sym__Atomic] = ACTIONS(6519),
    [anon_sym__Noreturn] = ACTIONS(6519),
    [anon_sym__Nonnull] = ACTIONS(6519),
    [anon_sym_nullable] = ACTIONS(6519),
    [anon_sym__Complex] = ACTIONS(6519),
    [anon_sym__Nullable] = ACTIONS(6519),
    [anon_sym__Nullable_result] = ACTIONS(6519),
    [anon_sym__Null_unspecified] = ACTIONS(6519),
    [anon_sym___autoreleasing] = ACTIONS(6519),
    [anon_sym___block] = ACTIONS(6519),
    [anon_sym___bridge] = ACTIONS(6519),
    [anon_sym___bridge_retained] = ACTIONS(6519),
    [anon_sym___bridge_transfer] = ACTIONS(6519),
    [anon_sym___complex] = ACTIONS(6519),
    [anon_sym___const] = ACTIONS(6519),
    [anon_sym___imag] = ACTIONS(6519),
    [anon_sym___kindof] = ACTIONS(6519),
    [anon_sym___nonnull] = ACTIONS(6519),
    [anon_sym___nullable] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6519),
    [anon_sym___real] = ACTIONS(6519),
    [anon_sym___strong] = ACTIONS(6519),
    [anon_sym___unsafe_unretained] = ACTIONS(6519),
    [anon_sym___unused] = ACTIONS(6519),
    [anon_sym___weak] = ACTIONS(6519),
    [anon_sym_alignas] = ACTIONS(6519),
    [anon_sym__Alignas] = ACTIONS(6519),
    [sym_primitive_type] = ACTIONS(6519),
    [anon_sym_enum] = ACTIONS(6519),
    [anon_sym_struct] = ACTIONS(6519),
    [anon_sym_union] = ACTIONS(6519),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6519),
    [anon_sym___typeof] = ACTIONS(6519),
    [anon_sym_typeof] = ACTIONS(6519),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6519),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6519),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6519),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_API_AVAILABLE] = ACTIONS(6519),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_API_DEPRECATED] = ACTIONS(6519),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym___deprecated_msg] = ACTIONS(6519),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6519),
    [anon_sym_BOOL] = ACTIONS(6519),
    [anon_sym_IMP] = ACTIONS(6519),
    [anon_sym_SEL] = ACTIONS(6519),
    [anon_sym_Class] = ACTIONS(6519),
    [anon_sym_id] = ACTIONS(6519),
    [anon_sym_ATdefs] = ACTIONS(6521),
  },
  [STATE(2645)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token2] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [aux_sym_preproc_else_token1] = ACTIONS(6523),
    [aux_sym_preproc_elif_token1] = ACTIONS(6523),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2646)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token2] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [aux_sym_preproc_else_token1] = ACTIONS(6523),
    [aux_sym_preproc_elif_token1] = ACTIONS(6523),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2647)] = {
    [sym_identifier] = ACTIONS(6527),
    [aux_sym_preproc_def_token1] = ACTIONS(6527),
    [aux_sym_preproc_if_token1] = ACTIONS(6527),
    [aux_sym_preproc_if_token2] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6527),
    [aux_sym_preproc_else_token1] = ACTIONS(6527),
    [aux_sym_preproc_elif_token1] = ACTIONS(6527),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6527),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6527),
    [sym_preproc_directive] = ACTIONS(6527),
    [anon_sym___extension__] = ACTIONS(6527),
    [anon_sym_extern] = ACTIONS(6527),
    [anon_sym___attribute__] = ACTIONS(6527),
    [anon_sym___attribute] = ACTIONS(6527),
    [anon_sym_noreturn] = ACTIONS(6527),
    [anon_sym_LBRACK] = ACTIONS(6529),
    [anon_sym___declspec] = ACTIONS(6527),
    [anon_sym_signed] = ACTIONS(6527),
    [anon_sym_unsigned] = ACTIONS(6527),
    [anon_sym_long] = ACTIONS(6527),
    [anon_sym_short] = ACTIONS(6527),
    [anon_sym_static] = ACTIONS(6527),
    [anon_sym_auto] = ACTIONS(6527),
    [anon_sym_register] = ACTIONS(6527),
    [anon_sym_inline] = ACTIONS(6527),
    [anon_sym___inline] = ACTIONS(6527),
    [anon_sym___inline__] = ACTIONS(6527),
    [anon_sym___forceinline] = ACTIONS(6527),
    [anon_sym_thread_local] = ACTIONS(6527),
    [anon_sym___thread] = ACTIONS(6527),
    [anon_sym_CG_EXTERN] = ACTIONS(6527),
    [anon_sym_CG_INLINE] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6527),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6527),
    [anon_sym_IBOutlet] = ACTIONS(6527),
    [anon_sym_IBInspectable] = ACTIONS(6527),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6527),
    [anon_sym_NS_INLINE] = ACTIONS(6527),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6527),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6527),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6527),
    [anon_sym_const] = ACTIONS(6527),
    [anon_sym_constexpr] = ACTIONS(6527),
    [anon_sym_volatile] = ACTIONS(6527),
    [anon_sym_restrict] = ACTIONS(6527),
    [anon_sym___restrict__] = ACTIONS(6527),
    [anon_sym__Atomic] = ACTIONS(6527),
    [anon_sym__Noreturn] = ACTIONS(6527),
    [anon_sym__Nonnull] = ACTIONS(6527),
    [anon_sym_nullable] = ACTIONS(6527),
    [anon_sym__Complex] = ACTIONS(6527),
    [anon_sym__Nullable] = ACTIONS(6527),
    [anon_sym__Nullable_result] = ACTIONS(6527),
    [anon_sym__Null_unspecified] = ACTIONS(6527),
    [anon_sym___autoreleasing] = ACTIONS(6527),
    [anon_sym___block] = ACTIONS(6527),
    [anon_sym___bridge] = ACTIONS(6527),
    [anon_sym___bridge_retained] = ACTIONS(6527),
    [anon_sym___bridge_transfer] = ACTIONS(6527),
    [anon_sym___complex] = ACTIONS(6527),
    [anon_sym___const] = ACTIONS(6527),
    [anon_sym___imag] = ACTIONS(6527),
    [anon_sym___kindof] = ACTIONS(6527),
    [anon_sym___nonnull] = ACTIONS(6527),
    [anon_sym___nullable] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6527),
    [anon_sym___real] = ACTIONS(6527),
    [anon_sym___strong] = ACTIONS(6527),
    [anon_sym___unsafe_unretained] = ACTIONS(6527),
    [anon_sym___unused] = ACTIONS(6527),
    [anon_sym___weak] = ACTIONS(6527),
    [anon_sym_alignas] = ACTIONS(6527),
    [anon_sym__Alignas] = ACTIONS(6527),
    [sym_primitive_type] = ACTIONS(6527),
    [anon_sym_enum] = ACTIONS(6527),
    [anon_sym_struct] = ACTIONS(6527),
    [anon_sym_union] = ACTIONS(6527),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6527),
    [anon_sym___typeof] = ACTIONS(6527),
    [anon_sym_typeof] = ACTIONS(6527),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6527),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6527),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6527),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_API_AVAILABLE] = ACTIONS(6527),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_API_DEPRECATED] = ACTIONS(6527),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym___deprecated_msg] = ACTIONS(6527),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6527),
    [anon_sym_BOOL] = ACTIONS(6527),
    [anon_sym_IMP] = ACTIONS(6527),
    [anon_sym_SEL] = ACTIONS(6527),
    [anon_sym_Class] = ACTIONS(6527),
    [anon_sym_id] = ACTIONS(6527),
    [anon_sym_ATdefs] = ACTIONS(6529),
  },
  [STATE(2648)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(4436),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(6499),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4436),
    [anon_sym___extension__] = ACTIONS(4436),
    [anon_sym_extern] = ACTIONS(4436),
    [anon_sym___attribute__] = ACTIONS(4436),
    [anon_sym___attribute] = ACTIONS(4431),
    [anon_sym_noreturn] = ACTIONS(4436),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4436),
    [anon_sym_static] = ACTIONS(4436),
    [anon_sym_auto] = ACTIONS(4436),
    [anon_sym_register] = ACTIONS(4436),
    [anon_sym_inline] = ACTIONS(4436),
    [anon_sym___inline] = ACTIONS(4431),
    [anon_sym___inline__] = ACTIONS(4436),
    [anon_sym___forceinline] = ACTIONS(4436),
    [anon_sym_thread_local] = ACTIONS(4436),
    [anon_sym___thread] = ACTIONS(4436),
    [anon_sym_CG_EXTERN] = ACTIONS(4436),
    [anon_sym_CG_INLINE] = ACTIONS(4436),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4436),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4436),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4436),
    [anon_sym_IBOutlet] = ACTIONS(4436),
    [anon_sym_IBInspectable] = ACTIONS(4436),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4436),
    [anon_sym_NS_INLINE] = ACTIONS(4436),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4436),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4436),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4436),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4436),
    [anon_sym_const] = ACTIONS(4431),
    [anon_sym_constexpr] = ACTIONS(4436),
    [anon_sym_volatile] = ACTIONS(4436),
    [anon_sym_restrict] = ACTIONS(4436),
    [anon_sym___restrict__] = ACTIONS(4436),
    [anon_sym__Atomic] = ACTIONS(4436),
    [anon_sym__Noreturn] = ACTIONS(4436),
    [anon_sym__Nonnull] = ACTIONS(4436),
    [anon_sym_nullable] = ACTIONS(4436),
    [anon_sym__Complex] = ACTIONS(4436),
    [anon_sym__Nullable] = ACTIONS(4431),
    [anon_sym__Nullable_result] = ACTIONS(4436),
    [anon_sym__Null_unspecified] = ACTIONS(4436),
    [anon_sym___autoreleasing] = ACTIONS(4436),
    [anon_sym___block] = ACTIONS(4436),
    [anon_sym___bridge] = ACTIONS(4431),
    [anon_sym___bridge_retained] = ACTIONS(4436),
    [anon_sym___bridge_transfer] = ACTIONS(4436),
    [anon_sym___complex] = ACTIONS(4436),
    [anon_sym___const] = ACTIONS(4436),
    [anon_sym___imag] = ACTIONS(4436),
    [anon_sym___kindof] = ACTIONS(4436),
    [anon_sym___nonnull] = ACTIONS(4436),
    [anon_sym___nullable] = ACTIONS(4436),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4436),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4436),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4436),
    [anon_sym___real] = ACTIONS(4436),
    [anon_sym___strong] = ACTIONS(4436),
    [anon_sym___unsafe_unretained] = ACTIONS(4436),
    [anon_sym___unused] = ACTIONS(4436),
    [anon_sym___weak] = ACTIONS(4436),
    [anon_sym_alignas] = ACTIONS(4436),
    [anon_sym__Alignas] = ACTIONS(4436),
    [anon_sym_QMARK] = ACTIONS(4436),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4436),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4436),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4436),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4436),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4436),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4436),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4436),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4436),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4436),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4436),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4431),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4436),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4436),
    [anon_sym_API_AVAILABLE] = ACTIONS(4436),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4436),
    [anon_sym_API_DEPRECATED] = ACTIONS(4436),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4436),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4436),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4436),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4436),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4436),
    [anon_sym___deprecated_msg] = ACTIONS(4436),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4436),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4436),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4436),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4436),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4436),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4436),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4436),
  },
  [STATE(2649)] = {
    [sym_argument_list] = STATE(1509),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(6491),
    [anon_sym_COMMA] = ACTIONS(4476),
    [anon_sym_LPAREN2] = ACTIONS(3980),
    [anon_sym_DASH] = ACTIONS(6493),
    [anon_sym_PLUS] = ACTIONS(6493),
    [anon_sym_STAR] = ACTIONS(6495),
    [anon_sym_SLASH] = ACTIONS(6497),
    [anon_sym_PERCENT] = ACTIONS(6495),
    [anon_sym_PIPE_PIPE] = ACTIONS(6499),
    [anon_sym_AMP_AMP] = ACTIONS(6501),
    [anon_sym_PIPE] = ACTIONS(6503),
    [anon_sym_CARET] = ACTIONS(6505),
    [anon_sym_AMP] = ACTIONS(6507),
    [anon_sym_EQ_EQ] = ACTIONS(6509),
    [anon_sym_BANG_EQ] = ACTIONS(6509),
    [anon_sym_GT] = ACTIONS(6511),
    [anon_sym_GT_EQ] = ACTIONS(6513),
    [anon_sym_LT_EQ] = ACTIONS(6513),
    [anon_sym_LT] = ACTIONS(6511),
    [anon_sym_LT_LT] = ACTIONS(6515),
    [anon_sym_GT_GT] = ACTIONS(6515),
    [anon_sym_SEMI] = ACTIONS(4476),
    [anon_sym___extension__] = ACTIONS(4476),
    [anon_sym_extern] = ACTIONS(4476),
    [anon_sym___attribute__] = ACTIONS(4476),
    [anon_sym___attribute] = ACTIONS(4474),
    [anon_sym_noreturn] = ACTIONS(4476),
    [anon_sym_LBRACK] = ACTIONS(3982),
    [anon_sym___declspec] = ACTIONS(4476),
    [anon_sym_static] = ACTIONS(4476),
    [anon_sym_auto] = ACTIONS(4476),
    [anon_sym_register] = ACTIONS(4476),
    [anon_sym_inline] = ACTIONS(4476),
    [anon_sym___inline] = ACTIONS(4474),
    [anon_sym___inline__] = ACTIONS(4476),
    [anon_sym___forceinline] = ACTIONS(4476),
    [anon_sym_thread_local] = ACTIONS(4476),
    [anon_sym___thread] = ACTIONS(4476),
    [anon_sym_CG_EXTERN] = ACTIONS(4476),
    [anon_sym_CG_INLINE] = ACTIONS(4476),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4476),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4476),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4476),
    [anon_sym_IBOutlet] = ACTIONS(4476),
    [anon_sym_IBInspectable] = ACTIONS(4476),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4476),
    [anon_sym_NS_INLINE] = ACTIONS(4476),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4476),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4476),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4476),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4476),
    [anon_sym_const] = ACTIONS(4474),
    [anon_sym_constexpr] = ACTIONS(4476),
    [anon_sym_volatile] = ACTIONS(4476),
    [anon_sym_restrict] = ACTIONS(4476),
    [anon_sym___restrict__] = ACTIONS(4476),
    [anon_sym__Atomic] = ACTIONS(4476),
    [anon_sym__Noreturn] = ACTIONS(4476),
    [anon_sym__Nonnull] = ACTIONS(4476),
    [anon_sym_nullable] = ACTIONS(4476),
    [anon_sym__Complex] = ACTIONS(4476),
    [anon_sym__Nullable] = ACTIONS(4474),
    [anon_sym__Nullable_result] = ACTIONS(4476),
    [anon_sym__Null_unspecified] = ACTIONS(4476),
    [anon_sym___autoreleasing] = ACTIONS(4476),
    [anon_sym___block] = ACTIONS(4476),
    [anon_sym___bridge] = ACTIONS(4474),
    [anon_sym___bridge_retained] = ACTIONS(4476),
    [anon_sym___bridge_transfer] = ACTIONS(4476),
    [anon_sym___complex] = ACTIONS(4476),
    [anon_sym___const] = ACTIONS(4476),
    [anon_sym___imag] = ACTIONS(4476),
    [anon_sym___kindof] = ACTIONS(4476),
    [anon_sym___nonnull] = ACTIONS(4476),
    [anon_sym___nullable] = ACTIONS(4476),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4476),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4476),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4476),
    [anon_sym___real] = ACTIONS(4476),
    [anon_sym___strong] = ACTIONS(4476),
    [anon_sym___unsafe_unretained] = ACTIONS(4476),
    [anon_sym___unused] = ACTIONS(4476),
    [anon_sym___weak] = ACTIONS(4476),
    [anon_sym_alignas] = ACTIONS(4476),
    [anon_sym__Alignas] = ACTIONS(4476),
    [anon_sym_QMARK] = ACTIONS(4476),
    [anon_sym_DASH_DASH] = ACTIONS(4621),
    [anon_sym_PLUS_PLUS] = ACTIONS(4621),
    [anon_sym_DOT] = ACTIONS(3984),
    [anon_sym_DASH_GT] = ACTIONS(3986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4476),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4476),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4476),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4476),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4476),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4476),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4476),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4476),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4476),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4476),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4474),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4476),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4476),
    [anon_sym_API_AVAILABLE] = ACTIONS(4476),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4476),
    [anon_sym_API_DEPRECATED] = ACTIONS(4476),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4476),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4476),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4476),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4476),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4476),
    [anon_sym___deprecated_msg] = ACTIONS(4476),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4476),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4476),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4476),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4476),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4476),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4476),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4476),
  },
  [STATE(2650)] = {
    [sym_identifier] = ACTIONS(6610),
    [aux_sym_preproc_def_token1] = ACTIONS(6610),
    [aux_sym_preproc_if_token1] = ACTIONS(6610),
    [aux_sym_preproc_if_token2] = ACTIONS(6610),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6610),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6610),
    [aux_sym_preproc_else_token1] = ACTIONS(6610),
    [aux_sym_preproc_elif_token1] = ACTIONS(6610),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6610),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6610),
    [sym_preproc_directive] = ACTIONS(6610),
    [anon_sym___extension__] = ACTIONS(6610),
    [anon_sym_extern] = ACTIONS(6610),
    [anon_sym___attribute__] = ACTIONS(6610),
    [anon_sym___attribute] = ACTIONS(6610),
    [anon_sym_noreturn] = ACTIONS(6610),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym___declspec] = ACTIONS(6610),
    [anon_sym_signed] = ACTIONS(6610),
    [anon_sym_unsigned] = ACTIONS(6610),
    [anon_sym_long] = ACTIONS(6610),
    [anon_sym_short] = ACTIONS(6610),
    [anon_sym_static] = ACTIONS(6610),
    [anon_sym_auto] = ACTIONS(6610),
    [anon_sym_register] = ACTIONS(6610),
    [anon_sym_inline] = ACTIONS(6610),
    [anon_sym___inline] = ACTIONS(6610),
    [anon_sym___inline__] = ACTIONS(6610),
    [anon_sym___forceinline] = ACTIONS(6610),
    [anon_sym_thread_local] = ACTIONS(6610),
    [anon_sym___thread] = ACTIONS(6610),
    [anon_sym_CG_EXTERN] = ACTIONS(6610),
    [anon_sym_CG_INLINE] = ACTIONS(6610),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6610),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6610),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6610),
    [anon_sym_IBOutlet] = ACTIONS(6610),
    [anon_sym_IBInspectable] = ACTIONS(6610),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6610),
    [anon_sym_NS_INLINE] = ACTIONS(6610),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6610),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6610),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6610),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6610),
    [anon_sym_const] = ACTIONS(6610),
    [anon_sym_constexpr] = ACTIONS(6610),
    [anon_sym_volatile] = ACTIONS(6610),
    [anon_sym_restrict] = ACTIONS(6610),
    [anon_sym___restrict__] = ACTIONS(6610),
    [anon_sym__Atomic] = ACTIONS(6610),
    [anon_sym__Noreturn] = ACTIONS(6610),
    [anon_sym__Nonnull] = ACTIONS(6610),
    [anon_sym_nullable] = ACTIONS(6610),
    [anon_sym__Complex] = ACTIONS(6610),
    [anon_sym__Nullable] = ACTIONS(6610),
    [anon_sym__Nullable_result] = ACTIONS(6610),
    [anon_sym__Null_unspecified] = ACTIONS(6610),
    [anon_sym___autoreleasing] = ACTIONS(6610),
    [anon_sym___block] = ACTIONS(6610),
    [anon_sym___bridge] = ACTIONS(6610),
    [anon_sym___bridge_retained] = ACTIONS(6610),
    [anon_sym___bridge_transfer] = ACTIONS(6610),
    [anon_sym___complex] = ACTIONS(6610),
    [anon_sym___const] = ACTIONS(6610),
    [anon_sym___imag] = ACTIONS(6610),
    [anon_sym___kindof] = ACTIONS(6610),
    [anon_sym___nonnull] = ACTIONS(6610),
    [anon_sym___nullable] = ACTIONS(6610),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6610),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6610),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6610),
    [anon_sym___real] = ACTIONS(6610),
    [anon_sym___strong] = ACTIONS(6610),
    [anon_sym___unsafe_unretained] = ACTIONS(6610),
    [anon_sym___unused] = ACTIONS(6610),
    [anon_sym___weak] = ACTIONS(6610),
    [anon_sym_alignas] = ACTIONS(6610),
    [anon_sym__Alignas] = ACTIONS(6610),
    [sym_primitive_type] = ACTIONS(6610),
    [anon_sym_enum] = ACTIONS(6610),
    [anon_sym_struct] = ACTIONS(6610),
    [anon_sym_union] = ACTIONS(6610),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6610),
    [anon_sym___typeof] = ACTIONS(6610),
    [anon_sym_typeof] = ACTIONS(6610),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6610),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6610),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6610),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6610),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6610),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6610),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6610),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6610),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6610),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_API_AVAILABLE] = ACTIONS(6610),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_API_DEPRECATED] = ACTIONS(6610),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6610),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6610),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6610),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6610),
    [anon_sym___deprecated_msg] = ACTIONS(6610),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6610),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6610),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6610),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6610),
    [anon_sym_BOOL] = ACTIONS(6610),
    [anon_sym_IMP] = ACTIONS(6610),
    [anon_sym_SEL] = ACTIONS(6610),
    [anon_sym_Class] = ACTIONS(6610),
    [anon_sym_id] = ACTIONS(6610),
    [anon_sym_ATdefs] = ACTIONS(6612),
  },
  [STATE(2651)] = {
    [sym_identifier] = ACTIONS(6603),
    [aux_sym_preproc_def_token1] = ACTIONS(6603),
    [aux_sym_preproc_if_token1] = ACTIONS(6603),
    [aux_sym_preproc_if_token2] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6603),
    [aux_sym_preproc_else_token1] = ACTIONS(6603),
    [aux_sym_preproc_elif_token1] = ACTIONS(6603),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6603),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6603),
    [sym_preproc_directive] = ACTIONS(6603),
    [anon_sym___extension__] = ACTIONS(6603),
    [anon_sym_extern] = ACTIONS(6603),
    [anon_sym___attribute__] = ACTIONS(6603),
    [anon_sym___attribute] = ACTIONS(6603),
    [anon_sym_noreturn] = ACTIONS(6603),
    [anon_sym_LBRACK] = ACTIONS(6605),
    [anon_sym___declspec] = ACTIONS(6603),
    [anon_sym_signed] = ACTIONS(6603),
    [anon_sym_unsigned] = ACTIONS(6603),
    [anon_sym_long] = ACTIONS(6603),
    [anon_sym_short] = ACTIONS(6603),
    [anon_sym_static] = ACTIONS(6603),
    [anon_sym_auto] = ACTIONS(6603),
    [anon_sym_register] = ACTIONS(6603),
    [anon_sym_inline] = ACTIONS(6603),
    [anon_sym___inline] = ACTIONS(6603),
    [anon_sym___inline__] = ACTIONS(6603),
    [anon_sym___forceinline] = ACTIONS(6603),
    [anon_sym_thread_local] = ACTIONS(6603),
    [anon_sym___thread] = ACTIONS(6603),
    [anon_sym_CG_EXTERN] = ACTIONS(6603),
    [anon_sym_CG_INLINE] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6603),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6603),
    [anon_sym_IBOutlet] = ACTIONS(6603),
    [anon_sym_IBInspectable] = ACTIONS(6603),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6603),
    [anon_sym_NS_INLINE] = ACTIONS(6603),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6603),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6603),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6603),
    [anon_sym_const] = ACTIONS(6603),
    [anon_sym_constexpr] = ACTIONS(6603),
    [anon_sym_volatile] = ACTIONS(6603),
    [anon_sym_restrict] = ACTIONS(6603),
    [anon_sym___restrict__] = ACTIONS(6603),
    [anon_sym__Atomic] = ACTIONS(6603),
    [anon_sym__Noreturn] = ACTIONS(6603),
    [anon_sym__Nonnull] = ACTIONS(6603),
    [anon_sym_nullable] = ACTIONS(6603),
    [anon_sym__Complex] = ACTIONS(6603),
    [anon_sym__Nullable] = ACTIONS(6603),
    [anon_sym__Nullable_result] = ACTIONS(6603),
    [anon_sym__Null_unspecified] = ACTIONS(6603),
    [anon_sym___autoreleasing] = ACTIONS(6603),
    [anon_sym___block] = ACTIONS(6603),
    [anon_sym___bridge] = ACTIONS(6603),
    [anon_sym___bridge_retained] = ACTIONS(6603),
    [anon_sym___bridge_transfer] = ACTIONS(6603),
    [anon_sym___complex] = ACTIONS(6603),
    [anon_sym___const] = ACTIONS(6603),
    [anon_sym___imag] = ACTIONS(6603),
    [anon_sym___kindof] = ACTIONS(6603),
    [anon_sym___nonnull] = ACTIONS(6603),
    [anon_sym___nullable] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6603),
    [anon_sym___real] = ACTIONS(6603),
    [anon_sym___strong] = ACTIONS(6603),
    [anon_sym___unsafe_unretained] = ACTIONS(6603),
    [anon_sym___unused] = ACTIONS(6603),
    [anon_sym___weak] = ACTIONS(6603),
    [anon_sym_alignas] = ACTIONS(6603),
    [anon_sym__Alignas] = ACTIONS(6603),
    [sym_primitive_type] = ACTIONS(6603),
    [anon_sym_enum] = ACTIONS(6603),
    [anon_sym_struct] = ACTIONS(6603),
    [anon_sym_union] = ACTIONS(6603),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6603),
    [anon_sym___typeof] = ACTIONS(6603),
    [anon_sym_typeof] = ACTIONS(6603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6603),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6603),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6603),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_API_AVAILABLE] = ACTIONS(6603),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_API_DEPRECATED] = ACTIONS(6603),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym___deprecated_msg] = ACTIONS(6603),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6603),
    [anon_sym_BOOL] = ACTIONS(6603),
    [anon_sym_IMP] = ACTIONS(6603),
    [anon_sym_SEL] = ACTIONS(6603),
    [anon_sym_Class] = ACTIONS(6603),
    [anon_sym_id] = ACTIONS(6603),
    [anon_sym_ATdefs] = ACTIONS(6605),
  },
  [STATE(2652)] = {
    [sym_identifier] = ACTIONS(6614),
    [aux_sym_preproc_def_token1] = ACTIONS(6614),
    [aux_sym_preproc_if_token1] = ACTIONS(6614),
    [aux_sym_preproc_if_token2] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6614),
    [aux_sym_preproc_else_token1] = ACTIONS(6614),
    [aux_sym_preproc_elif_token1] = ACTIONS(6614),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6614),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6614),
    [sym_preproc_directive] = ACTIONS(6614),
    [anon_sym___extension__] = ACTIONS(6614),
    [anon_sym_extern] = ACTIONS(6614),
    [anon_sym___attribute__] = ACTIONS(6614),
    [anon_sym___attribute] = ACTIONS(6614),
    [anon_sym_noreturn] = ACTIONS(6614),
    [anon_sym_LBRACK] = ACTIONS(6616),
    [anon_sym___declspec] = ACTIONS(6614),
    [anon_sym_signed] = ACTIONS(6614),
    [anon_sym_unsigned] = ACTIONS(6614),
    [anon_sym_long] = ACTIONS(6614),
    [anon_sym_short] = ACTIONS(6614),
    [anon_sym_static] = ACTIONS(6614),
    [anon_sym_auto] = ACTIONS(6614),
    [anon_sym_register] = ACTIONS(6614),
    [anon_sym_inline] = ACTIONS(6614),
    [anon_sym___inline] = ACTIONS(6614),
    [anon_sym___inline__] = ACTIONS(6614),
    [anon_sym___forceinline] = ACTIONS(6614),
    [anon_sym_thread_local] = ACTIONS(6614),
    [anon_sym___thread] = ACTIONS(6614),
    [anon_sym_CG_EXTERN] = ACTIONS(6614),
    [anon_sym_CG_INLINE] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6614),
    [anon_sym_IBOutlet] = ACTIONS(6614),
    [anon_sym_IBInspectable] = ACTIONS(6614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6614),
    [anon_sym_NS_INLINE] = ACTIONS(6614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6614),
    [anon_sym_const] = ACTIONS(6614),
    [anon_sym_constexpr] = ACTIONS(6614),
    [anon_sym_volatile] = ACTIONS(6614),
    [anon_sym_restrict] = ACTIONS(6614),
    [anon_sym___restrict__] = ACTIONS(6614),
    [anon_sym__Atomic] = ACTIONS(6614),
    [anon_sym__Noreturn] = ACTIONS(6614),
    [anon_sym__Nonnull] = ACTIONS(6614),
    [anon_sym_nullable] = ACTIONS(6614),
    [anon_sym__Complex] = ACTIONS(6614),
    [anon_sym__Nullable] = ACTIONS(6614),
    [anon_sym__Nullable_result] = ACTIONS(6614),
    [anon_sym__Null_unspecified] = ACTIONS(6614),
    [anon_sym___autoreleasing] = ACTIONS(6614),
    [anon_sym___block] = ACTIONS(6614),
    [anon_sym___bridge] = ACTIONS(6614),
    [anon_sym___bridge_retained] = ACTIONS(6614),
    [anon_sym___bridge_transfer] = ACTIONS(6614),
    [anon_sym___complex] = ACTIONS(6614),
    [anon_sym___const] = ACTIONS(6614),
    [anon_sym___imag] = ACTIONS(6614),
    [anon_sym___kindof] = ACTIONS(6614),
    [anon_sym___nonnull] = ACTIONS(6614),
    [anon_sym___nullable] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6614),
    [anon_sym___real] = ACTIONS(6614),
    [anon_sym___strong] = ACTIONS(6614),
    [anon_sym___unsafe_unretained] = ACTIONS(6614),
    [anon_sym___unused] = ACTIONS(6614),
    [anon_sym___weak] = ACTIONS(6614),
    [anon_sym_alignas] = ACTIONS(6614),
    [anon_sym__Alignas] = ACTIONS(6614),
    [sym_primitive_type] = ACTIONS(6614),
    [anon_sym_enum] = ACTIONS(6614),
    [anon_sym_struct] = ACTIONS(6614),
    [anon_sym_union] = ACTIONS(6614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6614),
    [anon_sym___typeof] = ACTIONS(6614),
    [anon_sym_typeof] = ACTIONS(6614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_API_AVAILABLE] = ACTIONS(6614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_API_DEPRECATED] = ACTIONS(6614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym___deprecated_msg] = ACTIONS(6614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6614),
    [anon_sym_BOOL] = ACTIONS(6614),
    [anon_sym_IMP] = ACTIONS(6614),
    [anon_sym_SEL] = ACTIONS(6614),
    [anon_sym_Class] = ACTIONS(6614),
    [anon_sym_id] = ACTIONS(6614),
    [anon_sym_ATdefs] = ACTIONS(6616),
  },
  [STATE(2653)] = {
    [sym_identifier] = ACTIONS(6614),
    [aux_sym_preproc_def_token1] = ACTIONS(6614),
    [aux_sym_preproc_if_token1] = ACTIONS(6614),
    [aux_sym_preproc_if_token2] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6614),
    [aux_sym_preproc_else_token1] = ACTIONS(6614),
    [aux_sym_preproc_elif_token1] = ACTIONS(6614),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6614),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6614),
    [sym_preproc_directive] = ACTIONS(6614),
    [anon_sym___extension__] = ACTIONS(6614),
    [anon_sym_extern] = ACTIONS(6614),
    [anon_sym___attribute__] = ACTIONS(6614),
    [anon_sym___attribute] = ACTIONS(6614),
    [anon_sym_noreturn] = ACTIONS(6614),
    [anon_sym_LBRACK] = ACTIONS(6616),
    [anon_sym___declspec] = ACTIONS(6614),
    [anon_sym_signed] = ACTIONS(6614),
    [anon_sym_unsigned] = ACTIONS(6614),
    [anon_sym_long] = ACTIONS(6614),
    [anon_sym_short] = ACTIONS(6614),
    [anon_sym_static] = ACTIONS(6614),
    [anon_sym_auto] = ACTIONS(6614),
    [anon_sym_register] = ACTIONS(6614),
    [anon_sym_inline] = ACTIONS(6614),
    [anon_sym___inline] = ACTIONS(6614),
    [anon_sym___inline__] = ACTIONS(6614),
    [anon_sym___forceinline] = ACTIONS(6614),
    [anon_sym_thread_local] = ACTIONS(6614),
    [anon_sym___thread] = ACTIONS(6614),
    [anon_sym_CG_EXTERN] = ACTIONS(6614),
    [anon_sym_CG_INLINE] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6614),
    [anon_sym_IBOutlet] = ACTIONS(6614),
    [anon_sym_IBInspectable] = ACTIONS(6614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6614),
    [anon_sym_NS_INLINE] = ACTIONS(6614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6614),
    [anon_sym_const] = ACTIONS(6614),
    [anon_sym_constexpr] = ACTIONS(6614),
    [anon_sym_volatile] = ACTIONS(6614),
    [anon_sym_restrict] = ACTIONS(6614),
    [anon_sym___restrict__] = ACTIONS(6614),
    [anon_sym__Atomic] = ACTIONS(6614),
    [anon_sym__Noreturn] = ACTIONS(6614),
    [anon_sym__Nonnull] = ACTIONS(6614),
    [anon_sym_nullable] = ACTIONS(6614),
    [anon_sym__Complex] = ACTIONS(6614),
    [anon_sym__Nullable] = ACTIONS(6614),
    [anon_sym__Nullable_result] = ACTIONS(6614),
    [anon_sym__Null_unspecified] = ACTIONS(6614),
    [anon_sym___autoreleasing] = ACTIONS(6614),
    [anon_sym___block] = ACTIONS(6614),
    [anon_sym___bridge] = ACTIONS(6614),
    [anon_sym___bridge_retained] = ACTIONS(6614),
    [anon_sym___bridge_transfer] = ACTIONS(6614),
    [anon_sym___complex] = ACTIONS(6614),
    [anon_sym___const] = ACTIONS(6614),
    [anon_sym___imag] = ACTIONS(6614),
    [anon_sym___kindof] = ACTIONS(6614),
    [anon_sym___nonnull] = ACTIONS(6614),
    [anon_sym___nullable] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6614),
    [anon_sym___real] = ACTIONS(6614),
    [anon_sym___strong] = ACTIONS(6614),
    [anon_sym___unsafe_unretained] = ACTIONS(6614),
    [anon_sym___unused] = ACTIONS(6614),
    [anon_sym___weak] = ACTIONS(6614),
    [anon_sym_alignas] = ACTIONS(6614),
    [anon_sym__Alignas] = ACTIONS(6614),
    [sym_primitive_type] = ACTIONS(6614),
    [anon_sym_enum] = ACTIONS(6614),
    [anon_sym_struct] = ACTIONS(6614),
    [anon_sym_union] = ACTIONS(6614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6614),
    [anon_sym___typeof] = ACTIONS(6614),
    [anon_sym_typeof] = ACTIONS(6614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_API_AVAILABLE] = ACTIONS(6614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_API_DEPRECATED] = ACTIONS(6614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym___deprecated_msg] = ACTIONS(6614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6614),
    [anon_sym_BOOL] = ACTIONS(6614),
    [anon_sym_IMP] = ACTIONS(6614),
    [anon_sym_SEL] = ACTIONS(6614),
    [anon_sym_Class] = ACTIONS(6614),
    [anon_sym_id] = ACTIONS(6614),
    [anon_sym_ATdefs] = ACTIONS(6616),
  },
  [STATE(2654)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token2] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [aux_sym_preproc_else_token1] = ACTIONS(6583),
    [aux_sym_preproc_elif_token1] = ACTIONS(6583),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2655)] = {
    [sym_string_literal] = STATE(1724),
    [sym_identifier] = ACTIONS(5682),
    [anon_sym___extension__] = ACTIONS(5682),
    [anon_sym_extern] = ACTIONS(5682),
    [anon_sym___attribute__] = ACTIONS(5682),
    [anon_sym___attribute] = ACTIONS(5682),
    [anon_sym_noreturn] = ACTIONS(5682),
    [anon_sym_LBRACK] = ACTIONS(5684),
    [anon_sym___declspec] = ACTIONS(5682),
    [anon_sym_signed] = ACTIONS(5682),
    [anon_sym_unsigned] = ACTIONS(5682),
    [anon_sym_long] = ACTIONS(5682),
    [anon_sym_short] = ACTIONS(5682),
    [anon_sym_static] = ACTIONS(5682),
    [anon_sym_auto] = ACTIONS(5682),
    [anon_sym_register] = ACTIONS(5682),
    [anon_sym_inline] = ACTIONS(5682),
    [anon_sym___inline] = ACTIONS(5682),
    [anon_sym___inline__] = ACTIONS(5682),
    [anon_sym___forceinline] = ACTIONS(5682),
    [anon_sym_thread_local] = ACTIONS(5682),
    [anon_sym___thread] = ACTIONS(5682),
    [anon_sym_CG_EXTERN] = ACTIONS(5682),
    [anon_sym_CG_INLINE] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5682),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5682),
    [anon_sym_IBOutlet] = ACTIONS(5682),
    [anon_sym_IBInspectable] = ACTIONS(5682),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5682),
    [anon_sym_NS_INLINE] = ACTIONS(5682),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5682),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5682),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5682),
    [anon_sym_const] = ACTIONS(5682),
    [anon_sym_constexpr] = ACTIONS(5682),
    [anon_sym_volatile] = ACTIONS(5682),
    [anon_sym_restrict] = ACTIONS(5682),
    [anon_sym___restrict__] = ACTIONS(5682),
    [anon_sym__Atomic] = ACTIONS(5682),
    [anon_sym__Noreturn] = ACTIONS(5682),
    [anon_sym__Nonnull] = ACTIONS(5682),
    [anon_sym_nullable] = ACTIONS(5682),
    [anon_sym__Complex] = ACTIONS(5682),
    [anon_sym__Nullable] = ACTIONS(5682),
    [anon_sym__Nullable_result] = ACTIONS(5682),
    [anon_sym__Null_unspecified] = ACTIONS(5682),
    [anon_sym___autoreleasing] = ACTIONS(5682),
    [anon_sym___block] = ACTIONS(5682),
    [anon_sym___bridge] = ACTIONS(5682),
    [anon_sym___bridge_retained] = ACTIONS(5682),
    [anon_sym___bridge_transfer] = ACTIONS(5682),
    [anon_sym___complex] = ACTIONS(5682),
    [anon_sym___const] = ACTIONS(5682),
    [anon_sym___imag] = ACTIONS(5682),
    [anon_sym___kindof] = ACTIONS(5682),
    [anon_sym___nonnull] = ACTIONS(5682),
    [anon_sym___nullable] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5682),
    [anon_sym___real] = ACTIONS(5682),
    [anon_sym___strong] = ACTIONS(5682),
    [anon_sym___unsafe_unretained] = ACTIONS(5682),
    [anon_sym___unused] = ACTIONS(5682),
    [anon_sym___weak] = ACTIONS(5682),
    [anon_sym_alignas] = ACTIONS(5682),
    [anon_sym__Alignas] = ACTIONS(5682),
    [sym_primitive_type] = ACTIONS(5682),
    [anon_sym_enum] = ACTIONS(5682),
    [anon_sym_struct] = ACTIONS(5682),
    [anon_sym_union] = ACTIONS(5682),
    [anon_sym_AT] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6620),
    [anon_sym_L_DQUOTE] = ACTIONS(6620),
    [anon_sym_u_DQUOTE] = ACTIONS(6620),
    [anon_sym_U_DQUOTE] = ACTIONS(6620),
    [anon_sym_u8_DQUOTE] = ACTIONS(6620),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5682),
    [anon_sym___typeof] = ACTIONS(5682),
    [anon_sym_typeof] = ACTIONS(5682),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5682),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5682),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5682),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_API_AVAILABLE] = ACTIONS(5682),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_API_DEPRECATED] = ACTIONS(5682),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym___deprecated_msg] = ACTIONS(5682),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5682),
    [anon_sym_ATprotocol] = ACTIONS(5684),
    [anon_sym_ATinterface] = ACTIONS(5684),
    [anon_sym_ATimplementation] = ACTIONS(5684),
    [anon_sym_BOOL] = ACTIONS(5682),
    [anon_sym_IMP] = ACTIONS(5682),
    [anon_sym_SEL] = ACTIONS(5682),
    [anon_sym_Class] = ACTIONS(5682),
    [anon_sym_id] = ACTIONS(5682),
  },
  [STATE(2656)] = {
    [sym_string_literal] = STATE(1714),
    [sym_identifier] = ACTIONS(5682),
    [anon_sym___extension__] = ACTIONS(5682),
    [anon_sym_extern] = ACTIONS(5682),
    [anon_sym___attribute__] = ACTIONS(5682),
    [anon_sym___attribute] = ACTIONS(5682),
    [anon_sym_noreturn] = ACTIONS(5682),
    [anon_sym_LBRACK] = ACTIONS(5684),
    [anon_sym___declspec] = ACTIONS(5682),
    [anon_sym_signed] = ACTIONS(5682),
    [anon_sym_unsigned] = ACTIONS(5682),
    [anon_sym_long] = ACTIONS(5682),
    [anon_sym_short] = ACTIONS(5682),
    [anon_sym_static] = ACTIONS(5682),
    [anon_sym_auto] = ACTIONS(5682),
    [anon_sym_register] = ACTIONS(5682),
    [anon_sym_inline] = ACTIONS(5682),
    [anon_sym___inline] = ACTIONS(5682),
    [anon_sym___inline__] = ACTIONS(5682),
    [anon_sym___forceinline] = ACTIONS(5682),
    [anon_sym_thread_local] = ACTIONS(5682),
    [anon_sym___thread] = ACTIONS(5682),
    [anon_sym_CG_EXTERN] = ACTIONS(5682),
    [anon_sym_CG_INLINE] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5682),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5682),
    [anon_sym_IBOutlet] = ACTIONS(5682),
    [anon_sym_IBInspectable] = ACTIONS(5682),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5682),
    [anon_sym_NS_INLINE] = ACTIONS(5682),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5682),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5682),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5682),
    [anon_sym_const] = ACTIONS(5682),
    [anon_sym_constexpr] = ACTIONS(5682),
    [anon_sym_volatile] = ACTIONS(5682),
    [anon_sym_restrict] = ACTIONS(5682),
    [anon_sym___restrict__] = ACTIONS(5682),
    [anon_sym__Atomic] = ACTIONS(5682),
    [anon_sym__Noreturn] = ACTIONS(5682),
    [anon_sym__Nonnull] = ACTIONS(5682),
    [anon_sym_nullable] = ACTIONS(5682),
    [anon_sym__Complex] = ACTIONS(5682),
    [anon_sym__Nullable] = ACTIONS(5682),
    [anon_sym__Nullable_result] = ACTIONS(5682),
    [anon_sym__Null_unspecified] = ACTIONS(5682),
    [anon_sym___autoreleasing] = ACTIONS(5682),
    [anon_sym___block] = ACTIONS(5682),
    [anon_sym___bridge] = ACTIONS(5682),
    [anon_sym___bridge_retained] = ACTIONS(5682),
    [anon_sym___bridge_transfer] = ACTIONS(5682),
    [anon_sym___complex] = ACTIONS(5682),
    [anon_sym___const] = ACTIONS(5682),
    [anon_sym___imag] = ACTIONS(5682),
    [anon_sym___kindof] = ACTIONS(5682),
    [anon_sym___nonnull] = ACTIONS(5682),
    [anon_sym___nullable] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5682),
    [anon_sym___real] = ACTIONS(5682),
    [anon_sym___strong] = ACTIONS(5682),
    [anon_sym___unsafe_unretained] = ACTIONS(5682),
    [anon_sym___unused] = ACTIONS(5682),
    [anon_sym___weak] = ACTIONS(5682),
    [anon_sym_alignas] = ACTIONS(5682),
    [anon_sym__Alignas] = ACTIONS(5682),
    [sym_primitive_type] = ACTIONS(5682),
    [anon_sym_enum] = ACTIONS(5682),
    [anon_sym_struct] = ACTIONS(5682),
    [anon_sym_union] = ACTIONS(5682),
    [anon_sym_AT] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6620),
    [anon_sym_L_DQUOTE] = ACTIONS(6620),
    [anon_sym_u_DQUOTE] = ACTIONS(6620),
    [anon_sym_U_DQUOTE] = ACTIONS(6620),
    [anon_sym_u8_DQUOTE] = ACTIONS(6620),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5682),
    [anon_sym___typeof] = ACTIONS(5682),
    [anon_sym_typeof] = ACTIONS(5682),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5682),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5682),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5682),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_API_AVAILABLE] = ACTIONS(5682),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_API_DEPRECATED] = ACTIONS(5682),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym___deprecated_msg] = ACTIONS(5682),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5682),
    [anon_sym_ATprotocol] = ACTIONS(5684),
    [anon_sym_ATinterface] = ACTIONS(5684),
    [anon_sym_ATimplementation] = ACTIONS(5684),
    [anon_sym_BOOL] = ACTIONS(5682),
    [anon_sym_IMP] = ACTIONS(5682),
    [anon_sym_SEL] = ACTIONS(5682),
    [anon_sym_Class] = ACTIONS(5682),
    [anon_sym_id] = ACTIONS(5682),
  },
  [STATE(2657)] = {
    [sym_string_literal] = STATE(1713),
    [sym_identifier] = ACTIONS(5682),
    [anon_sym___extension__] = ACTIONS(5682),
    [anon_sym_extern] = ACTIONS(5682),
    [anon_sym___attribute__] = ACTIONS(5682),
    [anon_sym___attribute] = ACTIONS(5682),
    [anon_sym_noreturn] = ACTIONS(5682),
    [anon_sym_LBRACK] = ACTIONS(5684),
    [anon_sym___declspec] = ACTIONS(5682),
    [anon_sym_signed] = ACTIONS(5682),
    [anon_sym_unsigned] = ACTIONS(5682),
    [anon_sym_long] = ACTIONS(5682),
    [anon_sym_short] = ACTIONS(5682),
    [anon_sym_static] = ACTIONS(5682),
    [anon_sym_auto] = ACTIONS(5682),
    [anon_sym_register] = ACTIONS(5682),
    [anon_sym_inline] = ACTIONS(5682),
    [anon_sym___inline] = ACTIONS(5682),
    [anon_sym___inline__] = ACTIONS(5682),
    [anon_sym___forceinline] = ACTIONS(5682),
    [anon_sym_thread_local] = ACTIONS(5682),
    [anon_sym___thread] = ACTIONS(5682),
    [anon_sym_CG_EXTERN] = ACTIONS(5682),
    [anon_sym_CG_INLINE] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5682),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5682),
    [anon_sym_IBOutlet] = ACTIONS(5682),
    [anon_sym_IBInspectable] = ACTIONS(5682),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5682),
    [anon_sym_NS_INLINE] = ACTIONS(5682),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5682),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5682),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5682),
    [anon_sym_const] = ACTIONS(5682),
    [anon_sym_constexpr] = ACTIONS(5682),
    [anon_sym_volatile] = ACTIONS(5682),
    [anon_sym_restrict] = ACTIONS(5682),
    [anon_sym___restrict__] = ACTIONS(5682),
    [anon_sym__Atomic] = ACTIONS(5682),
    [anon_sym__Noreturn] = ACTIONS(5682),
    [anon_sym__Nonnull] = ACTIONS(5682),
    [anon_sym_nullable] = ACTIONS(5682),
    [anon_sym__Complex] = ACTIONS(5682),
    [anon_sym__Nullable] = ACTIONS(5682),
    [anon_sym__Nullable_result] = ACTIONS(5682),
    [anon_sym__Null_unspecified] = ACTIONS(5682),
    [anon_sym___autoreleasing] = ACTIONS(5682),
    [anon_sym___block] = ACTIONS(5682),
    [anon_sym___bridge] = ACTIONS(5682),
    [anon_sym___bridge_retained] = ACTIONS(5682),
    [anon_sym___bridge_transfer] = ACTIONS(5682),
    [anon_sym___complex] = ACTIONS(5682),
    [anon_sym___const] = ACTIONS(5682),
    [anon_sym___imag] = ACTIONS(5682),
    [anon_sym___kindof] = ACTIONS(5682),
    [anon_sym___nonnull] = ACTIONS(5682),
    [anon_sym___nullable] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5682),
    [anon_sym___real] = ACTIONS(5682),
    [anon_sym___strong] = ACTIONS(5682),
    [anon_sym___unsafe_unretained] = ACTIONS(5682),
    [anon_sym___unused] = ACTIONS(5682),
    [anon_sym___weak] = ACTIONS(5682),
    [anon_sym_alignas] = ACTIONS(5682),
    [anon_sym__Alignas] = ACTIONS(5682),
    [sym_primitive_type] = ACTIONS(5682),
    [anon_sym_enum] = ACTIONS(5682),
    [anon_sym_struct] = ACTIONS(5682),
    [anon_sym_union] = ACTIONS(5682),
    [anon_sym_AT] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6620),
    [anon_sym_L_DQUOTE] = ACTIONS(6620),
    [anon_sym_u_DQUOTE] = ACTIONS(6620),
    [anon_sym_U_DQUOTE] = ACTIONS(6620),
    [anon_sym_u8_DQUOTE] = ACTIONS(6620),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5682),
    [anon_sym___typeof] = ACTIONS(5682),
    [anon_sym_typeof] = ACTIONS(5682),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5682),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5682),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5682),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_API_AVAILABLE] = ACTIONS(5682),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_API_DEPRECATED] = ACTIONS(5682),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym___deprecated_msg] = ACTIONS(5682),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5682),
    [anon_sym_ATprotocol] = ACTIONS(5684),
    [anon_sym_ATinterface] = ACTIONS(5684),
    [anon_sym_ATimplementation] = ACTIONS(5684),
    [anon_sym_BOOL] = ACTIONS(5682),
    [anon_sym_IMP] = ACTIONS(5682),
    [anon_sym_SEL] = ACTIONS(5682),
    [anon_sym_Class] = ACTIONS(5682),
    [anon_sym_id] = ACTIONS(5682),
  },
  [STATE(2658)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(2920),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2659)] = {
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(2920),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2660)] = {
    [sym_string_literal] = STATE(1721),
    [sym_identifier] = ACTIONS(5682),
    [anon_sym___extension__] = ACTIONS(5682),
    [anon_sym_extern] = ACTIONS(5682),
    [anon_sym___attribute__] = ACTIONS(5682),
    [anon_sym___attribute] = ACTIONS(5682),
    [anon_sym_noreturn] = ACTIONS(5682),
    [anon_sym_LBRACK] = ACTIONS(5684),
    [anon_sym___declspec] = ACTIONS(5682),
    [anon_sym_signed] = ACTIONS(5682),
    [anon_sym_unsigned] = ACTIONS(5682),
    [anon_sym_long] = ACTIONS(5682),
    [anon_sym_short] = ACTIONS(5682),
    [anon_sym_static] = ACTIONS(5682),
    [anon_sym_auto] = ACTIONS(5682),
    [anon_sym_register] = ACTIONS(5682),
    [anon_sym_inline] = ACTIONS(5682),
    [anon_sym___inline] = ACTIONS(5682),
    [anon_sym___inline__] = ACTIONS(5682),
    [anon_sym___forceinline] = ACTIONS(5682),
    [anon_sym_thread_local] = ACTIONS(5682),
    [anon_sym___thread] = ACTIONS(5682),
    [anon_sym_CG_EXTERN] = ACTIONS(5682),
    [anon_sym_CG_INLINE] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5682),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5682),
    [anon_sym_IBOutlet] = ACTIONS(5682),
    [anon_sym_IBInspectable] = ACTIONS(5682),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5682),
    [anon_sym_NS_INLINE] = ACTIONS(5682),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5682),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5682),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5682),
    [anon_sym_const] = ACTIONS(5682),
    [anon_sym_constexpr] = ACTIONS(5682),
    [anon_sym_volatile] = ACTIONS(5682),
    [anon_sym_restrict] = ACTIONS(5682),
    [anon_sym___restrict__] = ACTIONS(5682),
    [anon_sym__Atomic] = ACTIONS(5682),
    [anon_sym__Noreturn] = ACTIONS(5682),
    [anon_sym__Nonnull] = ACTIONS(5682),
    [anon_sym_nullable] = ACTIONS(5682),
    [anon_sym__Complex] = ACTIONS(5682),
    [anon_sym__Nullable] = ACTIONS(5682),
    [anon_sym__Nullable_result] = ACTIONS(5682),
    [anon_sym__Null_unspecified] = ACTIONS(5682),
    [anon_sym___autoreleasing] = ACTIONS(5682),
    [anon_sym___block] = ACTIONS(5682),
    [anon_sym___bridge] = ACTIONS(5682),
    [anon_sym___bridge_retained] = ACTIONS(5682),
    [anon_sym___bridge_transfer] = ACTIONS(5682),
    [anon_sym___complex] = ACTIONS(5682),
    [anon_sym___const] = ACTIONS(5682),
    [anon_sym___imag] = ACTIONS(5682),
    [anon_sym___kindof] = ACTIONS(5682),
    [anon_sym___nonnull] = ACTIONS(5682),
    [anon_sym___nullable] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5682),
    [anon_sym___real] = ACTIONS(5682),
    [anon_sym___strong] = ACTIONS(5682),
    [anon_sym___unsafe_unretained] = ACTIONS(5682),
    [anon_sym___unused] = ACTIONS(5682),
    [anon_sym___weak] = ACTIONS(5682),
    [anon_sym_alignas] = ACTIONS(5682),
    [anon_sym__Alignas] = ACTIONS(5682),
    [sym_primitive_type] = ACTIONS(5682),
    [anon_sym_enum] = ACTIONS(5682),
    [anon_sym_struct] = ACTIONS(5682),
    [anon_sym_union] = ACTIONS(5682),
    [anon_sym_AT] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6620),
    [anon_sym_L_DQUOTE] = ACTIONS(6620),
    [anon_sym_u_DQUOTE] = ACTIONS(6620),
    [anon_sym_U_DQUOTE] = ACTIONS(6620),
    [anon_sym_u8_DQUOTE] = ACTIONS(6620),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5682),
    [anon_sym___typeof] = ACTIONS(5682),
    [anon_sym_typeof] = ACTIONS(5682),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5682),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5682),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5682),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_API_AVAILABLE] = ACTIONS(5682),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_API_DEPRECATED] = ACTIONS(5682),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym___deprecated_msg] = ACTIONS(5682),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5682),
    [anon_sym_ATprotocol] = ACTIONS(5684),
    [anon_sym_ATinterface] = ACTIONS(5684),
    [anon_sym_ATimplementation] = ACTIONS(5684),
    [anon_sym_BOOL] = ACTIONS(5682),
    [anon_sym_IMP] = ACTIONS(5682),
    [anon_sym_SEL] = ACTIONS(5682),
    [anon_sym_Class] = ACTIONS(5682),
    [anon_sym_id] = ACTIONS(5682),
  },
  [STATE(2661)] = {
    [sym_string_literal] = STATE(1712),
    [sym_identifier] = ACTIONS(5682),
    [anon_sym___extension__] = ACTIONS(5682),
    [anon_sym_extern] = ACTIONS(5682),
    [anon_sym___attribute__] = ACTIONS(5682),
    [anon_sym___attribute] = ACTIONS(5682),
    [anon_sym_noreturn] = ACTIONS(5682),
    [anon_sym_LBRACK] = ACTIONS(5684),
    [anon_sym___declspec] = ACTIONS(5682),
    [anon_sym_signed] = ACTIONS(5682),
    [anon_sym_unsigned] = ACTIONS(5682),
    [anon_sym_long] = ACTIONS(5682),
    [anon_sym_short] = ACTIONS(5682),
    [anon_sym_static] = ACTIONS(5682),
    [anon_sym_auto] = ACTIONS(5682),
    [anon_sym_register] = ACTIONS(5682),
    [anon_sym_inline] = ACTIONS(5682),
    [anon_sym___inline] = ACTIONS(5682),
    [anon_sym___inline__] = ACTIONS(5682),
    [anon_sym___forceinline] = ACTIONS(5682),
    [anon_sym_thread_local] = ACTIONS(5682),
    [anon_sym___thread] = ACTIONS(5682),
    [anon_sym_CG_EXTERN] = ACTIONS(5682),
    [anon_sym_CG_INLINE] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5682),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5682),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5682),
    [anon_sym_IBOutlet] = ACTIONS(5682),
    [anon_sym_IBInspectable] = ACTIONS(5682),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5682),
    [anon_sym_NS_INLINE] = ACTIONS(5682),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5682),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5682),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5682),
    [anon_sym_const] = ACTIONS(5682),
    [anon_sym_constexpr] = ACTIONS(5682),
    [anon_sym_volatile] = ACTIONS(5682),
    [anon_sym_restrict] = ACTIONS(5682),
    [anon_sym___restrict__] = ACTIONS(5682),
    [anon_sym__Atomic] = ACTIONS(5682),
    [anon_sym__Noreturn] = ACTIONS(5682),
    [anon_sym__Nonnull] = ACTIONS(5682),
    [anon_sym_nullable] = ACTIONS(5682),
    [anon_sym__Complex] = ACTIONS(5682),
    [anon_sym__Nullable] = ACTIONS(5682),
    [anon_sym__Nullable_result] = ACTIONS(5682),
    [anon_sym__Null_unspecified] = ACTIONS(5682),
    [anon_sym___autoreleasing] = ACTIONS(5682),
    [anon_sym___block] = ACTIONS(5682),
    [anon_sym___bridge] = ACTIONS(5682),
    [anon_sym___bridge_retained] = ACTIONS(5682),
    [anon_sym___bridge_transfer] = ACTIONS(5682),
    [anon_sym___complex] = ACTIONS(5682),
    [anon_sym___const] = ACTIONS(5682),
    [anon_sym___imag] = ACTIONS(5682),
    [anon_sym___kindof] = ACTIONS(5682),
    [anon_sym___nonnull] = ACTIONS(5682),
    [anon_sym___nullable] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5682),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5682),
    [anon_sym___real] = ACTIONS(5682),
    [anon_sym___strong] = ACTIONS(5682),
    [anon_sym___unsafe_unretained] = ACTIONS(5682),
    [anon_sym___unused] = ACTIONS(5682),
    [anon_sym___weak] = ACTIONS(5682),
    [anon_sym_alignas] = ACTIONS(5682),
    [anon_sym__Alignas] = ACTIONS(5682),
    [sym_primitive_type] = ACTIONS(5682),
    [anon_sym_enum] = ACTIONS(5682),
    [anon_sym_struct] = ACTIONS(5682),
    [anon_sym_union] = ACTIONS(5682),
    [anon_sym_AT] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6620),
    [anon_sym_L_DQUOTE] = ACTIONS(6620),
    [anon_sym_u_DQUOTE] = ACTIONS(6620),
    [anon_sym_U_DQUOTE] = ACTIONS(6620),
    [anon_sym_u8_DQUOTE] = ACTIONS(6620),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5682),
    [anon_sym___typeof] = ACTIONS(5682),
    [anon_sym_typeof] = ACTIONS(5682),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5682),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5682),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5682),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5682),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5682),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_API_AVAILABLE] = ACTIONS(5682),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_API_DEPRECATED] = ACTIONS(5682),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5682),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5682),
    [anon_sym___deprecated_msg] = ACTIONS(5682),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5682),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5682),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5682),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5682),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5682),
    [anon_sym_ATprotocol] = ACTIONS(5684),
    [anon_sym_ATinterface] = ACTIONS(5684),
    [anon_sym_ATimplementation] = ACTIONS(5684),
    [anon_sym_BOOL] = ACTIONS(5682),
    [anon_sym_IMP] = ACTIONS(5682),
    [anon_sym_SEL] = ACTIONS(5682),
    [anon_sym_Class] = ACTIONS(5682),
    [anon_sym_id] = ACTIONS(5682),
  },
  [STATE(2662)] = {
    [sym_identifier] = ACTIONS(6622),
    [anon_sym_COMMA] = ACTIONS(6624),
    [anon_sym_LPAREN2] = ACTIONS(6624),
    [anon_sym_SEMI] = ACTIONS(6624),
    [anon_sym___extension__] = ACTIONS(6622),
    [anon_sym_extern] = ACTIONS(6622),
    [anon_sym___attribute__] = ACTIONS(6622),
    [anon_sym___attribute] = ACTIONS(6622),
    [anon_sym_noreturn] = ACTIONS(6622),
    [anon_sym_LBRACK] = ACTIONS(6624),
    [anon_sym___declspec] = ACTIONS(6622),
    [anon_sym_LBRACE] = ACTIONS(6624),
    [anon_sym_signed] = ACTIONS(6622),
    [anon_sym_unsigned] = ACTIONS(6622),
    [anon_sym_long] = ACTIONS(6622),
    [anon_sym_short] = ACTIONS(6622),
    [anon_sym_static] = ACTIONS(6622),
    [anon_sym_EQ] = ACTIONS(6624),
    [anon_sym_ATautoreleasepool] = ACTIONS(6624),
    [anon_sym_auto] = ACTIONS(6622),
    [anon_sym_register] = ACTIONS(6622),
    [anon_sym_inline] = ACTIONS(6622),
    [anon_sym___inline] = ACTIONS(6622),
    [anon_sym___inline__] = ACTIONS(6622),
    [anon_sym___forceinline] = ACTIONS(6622),
    [anon_sym_thread_local] = ACTIONS(6622),
    [anon_sym___thread] = ACTIONS(6622),
    [anon_sym_CG_EXTERN] = ACTIONS(6622),
    [anon_sym_CG_INLINE] = ACTIONS(6622),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6622),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6622),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6622),
    [anon_sym_IBOutlet] = ACTIONS(6622),
    [anon_sym_IBInspectable] = ACTIONS(6622),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6622),
    [anon_sym_NS_INLINE] = ACTIONS(6622),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6622),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6622),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6622),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6622),
    [anon_sym_const] = ACTIONS(6622),
    [anon_sym_constexpr] = ACTIONS(6622),
    [anon_sym_volatile] = ACTIONS(6622),
    [anon_sym_restrict] = ACTIONS(6622),
    [anon_sym___restrict__] = ACTIONS(6622),
    [anon_sym__Atomic] = ACTIONS(6622),
    [anon_sym__Noreturn] = ACTIONS(6622),
    [anon_sym__Nonnull] = ACTIONS(6622),
    [anon_sym_nullable] = ACTIONS(6622),
    [anon_sym__Complex] = ACTIONS(6622),
    [anon_sym__Nullable] = ACTIONS(6622),
    [anon_sym__Nullable_result] = ACTIONS(6622),
    [anon_sym__Null_unspecified] = ACTIONS(6622),
    [anon_sym___autoreleasing] = ACTIONS(6622),
    [anon_sym___block] = ACTIONS(6622),
    [anon_sym___bridge] = ACTIONS(6622),
    [anon_sym___bridge_retained] = ACTIONS(6622),
    [anon_sym___bridge_transfer] = ACTIONS(6622),
    [anon_sym___complex] = ACTIONS(6622),
    [anon_sym___const] = ACTIONS(6622),
    [anon_sym___imag] = ACTIONS(6622),
    [anon_sym___kindof] = ACTIONS(6622),
    [anon_sym___nonnull] = ACTIONS(6622),
    [anon_sym___nullable] = ACTIONS(6622),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6622),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6622),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6622),
    [anon_sym___real] = ACTIONS(6622),
    [anon_sym___strong] = ACTIONS(6622),
    [anon_sym___unsafe_unretained] = ACTIONS(6622),
    [anon_sym___unused] = ACTIONS(6622),
    [anon_sym___weak] = ACTIONS(6622),
    [anon_sym_alignas] = ACTIONS(6622),
    [anon_sym__Alignas] = ACTIONS(6622),
    [sym_primitive_type] = ACTIONS(6622),
    [anon_sym_enum] = ACTIONS(6622),
    [anon_sym_struct] = ACTIONS(6622),
    [anon_sym_union] = ACTIONS(6622),
    [anon_sym_asm] = ACTIONS(6622),
    [anon_sym___asm__] = ACTIONS(6622),
    [anon_sym___asm] = ACTIONS(6622),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6622),
    [anon_sym___typeof] = ACTIONS(6622),
    [anon_sym_typeof] = ACTIONS(6622),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6622),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6622),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6622),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6622),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6622),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6622),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6622),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6622),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6622),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6622),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6622),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6622),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6622),
    [anon_sym_API_AVAILABLE] = ACTIONS(6622),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6622),
    [anon_sym_API_DEPRECATED] = ACTIONS(6622),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6622),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6622),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6622),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6622),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6622),
    [anon_sym___deprecated_msg] = ACTIONS(6622),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6622),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6622),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6622),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6622),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6622),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6622),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6622),
    [anon_sym_BOOL] = ACTIONS(6622),
    [anon_sym_IMP] = ACTIONS(6622),
    [anon_sym_SEL] = ACTIONS(6622),
    [anon_sym_Class] = ACTIONS(6622),
    [anon_sym_id] = ACTIONS(6622),
  },
  [STATE(2663)] = {
    [sym_attribute_specifier] = STATE(6228),
    [sym_parameter_list] = STATE(2552),
    [aux_sym_c_method_parameter_repeat1] = STATE(2806),
    [sym_identifier] = ACTIONS(6626),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6630),
    [anon_sym_SEMI] = ACTIONS(6628),
    [anon_sym___extension__] = ACTIONS(6626),
    [anon_sym_extern] = ACTIONS(6626),
    [anon_sym___attribute__] = ACTIONS(6626),
    [anon_sym___attribute] = ACTIONS(6626),
    [anon_sym_noreturn] = ACTIONS(6626),
    [anon_sym_LBRACK] = ACTIONS(6632),
    [anon_sym___declspec] = ACTIONS(6626),
    [anon_sym_LBRACE] = ACTIONS(6628),
    [anon_sym_signed] = ACTIONS(6626),
    [anon_sym_unsigned] = ACTIONS(6626),
    [anon_sym_long] = ACTIONS(6626),
    [anon_sym_short] = ACTIONS(6626),
    [anon_sym_static] = ACTIONS(6626),
    [anon_sym_EQ] = ACTIONS(6634),
    [anon_sym_ATautoreleasepool] = ACTIONS(6628),
    [anon_sym_auto] = ACTIONS(6626),
    [anon_sym_register] = ACTIONS(6626),
    [anon_sym_inline] = ACTIONS(6626),
    [anon_sym___inline] = ACTIONS(6626),
    [anon_sym___inline__] = ACTIONS(6626),
    [anon_sym___forceinline] = ACTIONS(6626),
    [anon_sym_thread_local] = ACTIONS(6626),
    [anon_sym___thread] = ACTIONS(6626),
    [anon_sym_CG_EXTERN] = ACTIONS(6626),
    [anon_sym_CG_INLINE] = ACTIONS(6626),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6626),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6626),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6626),
    [anon_sym_IBOutlet] = ACTIONS(6626),
    [anon_sym_IBInspectable] = ACTIONS(6626),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6626),
    [anon_sym_NS_INLINE] = ACTIONS(6626),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6626),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6626),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6626),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6626),
    [anon_sym_const] = ACTIONS(6626),
    [anon_sym_constexpr] = ACTIONS(6626),
    [anon_sym_volatile] = ACTIONS(6626),
    [anon_sym_restrict] = ACTIONS(6626),
    [anon_sym___restrict__] = ACTIONS(6626),
    [anon_sym__Atomic] = ACTIONS(6626),
    [anon_sym__Noreturn] = ACTIONS(6626),
    [anon_sym__Nonnull] = ACTIONS(6626),
    [anon_sym_nullable] = ACTIONS(6626),
    [anon_sym__Complex] = ACTIONS(6626),
    [anon_sym__Nullable] = ACTIONS(6626),
    [anon_sym__Nullable_result] = ACTIONS(6626),
    [anon_sym__Null_unspecified] = ACTIONS(6626),
    [anon_sym___autoreleasing] = ACTIONS(6626),
    [anon_sym___block] = ACTIONS(6626),
    [anon_sym___bridge] = ACTIONS(6626),
    [anon_sym___bridge_retained] = ACTIONS(6626),
    [anon_sym___bridge_transfer] = ACTIONS(6626),
    [anon_sym___complex] = ACTIONS(6626),
    [anon_sym___const] = ACTIONS(6626),
    [anon_sym___imag] = ACTIONS(6626),
    [anon_sym___kindof] = ACTIONS(6626),
    [anon_sym___nonnull] = ACTIONS(6626),
    [anon_sym___nullable] = ACTIONS(6626),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6626),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6626),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6626),
    [anon_sym___real] = ACTIONS(6626),
    [anon_sym___strong] = ACTIONS(6626),
    [anon_sym___unsafe_unretained] = ACTIONS(6626),
    [anon_sym___unused] = ACTIONS(6626),
    [anon_sym___weak] = ACTIONS(6626),
    [anon_sym_alignas] = ACTIONS(6626),
    [anon_sym__Alignas] = ACTIONS(6626),
    [sym_primitive_type] = ACTIONS(6626),
    [anon_sym_enum] = ACTIONS(6626),
    [anon_sym_struct] = ACTIONS(6626),
    [anon_sym_union] = ACTIONS(6626),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6626),
    [anon_sym___typeof] = ACTIONS(6626),
    [anon_sym_typeof] = ACTIONS(6626),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6626),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6626),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6626),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6626),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6626),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6626),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6626),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6626),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6626),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6626),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6626),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6626),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6626),
    [anon_sym_API_AVAILABLE] = ACTIONS(6626),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6626),
    [anon_sym_API_DEPRECATED] = ACTIONS(6626),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6626),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6626),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6626),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6626),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6626),
    [anon_sym___deprecated_msg] = ACTIONS(6626),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6626),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6626),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6626),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6626),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6626),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6626),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6626),
    [anon_sym_BOOL] = ACTIONS(6626),
    [anon_sym_IMP] = ACTIONS(6626),
    [anon_sym_SEL] = ACTIONS(6626),
    [anon_sym_Class] = ACTIONS(6626),
    [anon_sym_id] = ACTIONS(6626),
  },
  [STATE(2664)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6343),
    [sym__type_declarator] = STATE(5246),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5698),
    [anon_sym_CARET] = ACTIONS(5700),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2665)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6343),
    [sym__type_declarator] = STATE(5216),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5698),
    [anon_sym_CARET] = ACTIONS(5700),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2666)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2220),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(6636),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2220),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2220),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2220),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2220),
    [anon_sym_GT_GT] = ACTIONS(2220),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym_RBRACK] = ACTIONS(2222),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_EQ] = ACTIONS(2220),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_COLON] = ACTIONS(2222),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_STAR_EQ] = ACTIONS(2222),
    [anon_sym_SLASH_EQ] = ACTIONS(2222),
    [anon_sym_PERCENT_EQ] = ACTIONS(2222),
    [anon_sym_PLUS_EQ] = ACTIONS(2222),
    [anon_sym_DASH_EQ] = ACTIONS(2222),
    [anon_sym_LT_LT_EQ] = ACTIONS(2222),
    [anon_sym_GT_GT_EQ] = ACTIONS(2222),
    [anon_sym_AMP_EQ] = ACTIONS(2222),
    [anon_sym_CARET_EQ] = ACTIONS(2222),
    [anon_sym_PIPE_EQ] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(2667)] = {
    [sym_identifier] = ACTIONS(6648),
    [anon_sym_COMMA] = ACTIONS(6650),
    [anon_sym_LPAREN2] = ACTIONS(6650),
    [anon_sym_SEMI] = ACTIONS(6650),
    [anon_sym___extension__] = ACTIONS(6648),
    [anon_sym_extern] = ACTIONS(6648),
    [anon_sym___attribute__] = ACTIONS(6648),
    [anon_sym___attribute] = ACTIONS(6648),
    [anon_sym_noreturn] = ACTIONS(6648),
    [anon_sym_LBRACK] = ACTIONS(6650),
    [anon_sym___declspec] = ACTIONS(6648),
    [anon_sym_LBRACE] = ACTIONS(6650),
    [anon_sym_signed] = ACTIONS(6648),
    [anon_sym_unsigned] = ACTIONS(6648),
    [anon_sym_long] = ACTIONS(6648),
    [anon_sym_short] = ACTIONS(6648),
    [anon_sym_static] = ACTIONS(6648),
    [anon_sym_EQ] = ACTIONS(6650),
    [anon_sym_ATautoreleasepool] = ACTIONS(6650),
    [anon_sym_auto] = ACTIONS(6648),
    [anon_sym_register] = ACTIONS(6648),
    [anon_sym_inline] = ACTIONS(6648),
    [anon_sym___inline] = ACTIONS(6648),
    [anon_sym___inline__] = ACTIONS(6648),
    [anon_sym___forceinline] = ACTIONS(6648),
    [anon_sym_thread_local] = ACTIONS(6648),
    [anon_sym___thread] = ACTIONS(6648),
    [anon_sym_CG_EXTERN] = ACTIONS(6648),
    [anon_sym_CG_INLINE] = ACTIONS(6648),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6648),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6648),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6648),
    [anon_sym_IBOutlet] = ACTIONS(6648),
    [anon_sym_IBInspectable] = ACTIONS(6648),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6648),
    [anon_sym_NS_INLINE] = ACTIONS(6648),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6648),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6648),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6648),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6648),
    [anon_sym_const] = ACTIONS(6648),
    [anon_sym_constexpr] = ACTIONS(6648),
    [anon_sym_volatile] = ACTIONS(6648),
    [anon_sym_restrict] = ACTIONS(6648),
    [anon_sym___restrict__] = ACTIONS(6648),
    [anon_sym__Atomic] = ACTIONS(6648),
    [anon_sym__Noreturn] = ACTIONS(6648),
    [anon_sym__Nonnull] = ACTIONS(6648),
    [anon_sym_nullable] = ACTIONS(6648),
    [anon_sym__Complex] = ACTIONS(6648),
    [anon_sym__Nullable] = ACTIONS(6648),
    [anon_sym__Nullable_result] = ACTIONS(6648),
    [anon_sym__Null_unspecified] = ACTIONS(6648),
    [anon_sym___autoreleasing] = ACTIONS(6648),
    [anon_sym___block] = ACTIONS(6648),
    [anon_sym___bridge] = ACTIONS(6648),
    [anon_sym___bridge_retained] = ACTIONS(6648),
    [anon_sym___bridge_transfer] = ACTIONS(6648),
    [anon_sym___complex] = ACTIONS(6648),
    [anon_sym___const] = ACTIONS(6648),
    [anon_sym___imag] = ACTIONS(6648),
    [anon_sym___kindof] = ACTIONS(6648),
    [anon_sym___nonnull] = ACTIONS(6648),
    [anon_sym___nullable] = ACTIONS(6648),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6648),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6648),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6648),
    [anon_sym___real] = ACTIONS(6648),
    [anon_sym___strong] = ACTIONS(6648),
    [anon_sym___unsafe_unretained] = ACTIONS(6648),
    [anon_sym___unused] = ACTIONS(6648),
    [anon_sym___weak] = ACTIONS(6648),
    [anon_sym_alignas] = ACTIONS(6648),
    [anon_sym__Alignas] = ACTIONS(6648),
    [sym_primitive_type] = ACTIONS(6648),
    [anon_sym_enum] = ACTIONS(6648),
    [anon_sym_struct] = ACTIONS(6648),
    [anon_sym_union] = ACTIONS(6648),
    [anon_sym_asm] = ACTIONS(6648),
    [anon_sym___asm__] = ACTIONS(6648),
    [anon_sym___asm] = ACTIONS(6648),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6648),
    [anon_sym___typeof] = ACTIONS(6648),
    [anon_sym_typeof] = ACTIONS(6648),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6648),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6648),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6648),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6648),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6648),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6648),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6648),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6648),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6648),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6648),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6648),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6648),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6648),
    [anon_sym_API_AVAILABLE] = ACTIONS(6648),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6648),
    [anon_sym_API_DEPRECATED] = ACTIONS(6648),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6648),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6648),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6648),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6648),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6648),
    [anon_sym___deprecated_msg] = ACTIONS(6648),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6648),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6648),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6648),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6648),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6648),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6648),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6648),
    [anon_sym_BOOL] = ACTIONS(6648),
    [anon_sym_IMP] = ACTIONS(6648),
    [anon_sym_SEL] = ACTIONS(6648),
    [anon_sym_Class] = ACTIONS(6648),
    [anon_sym_id] = ACTIONS(6648),
  },
  [STATE(2668)] = {
    [sym__declaration_modifiers] = STATE(2673),
    [sym_attribute_specifier] = STATE(2673),
    [sym_attribute_declaration] = STATE(2673),
    [sym_ms_declspec_modifier] = STATE(2673),
    [sym_ms_based_modifier] = STATE(6343),
    [sym__type_declarator] = STATE(5184),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(2673),
    [sym_type_qualifier] = STATE(2673),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2673),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2673),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5698),
    [anon_sym_CARET] = ACTIONS(5700),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2669)] = {
    [sym_identifier] = ACTIONS(6652),
    [anon_sym_COMMA] = ACTIONS(6654),
    [anon_sym_LPAREN2] = ACTIONS(6654),
    [anon_sym_SEMI] = ACTIONS(6654),
    [anon_sym___extension__] = ACTIONS(6656),
    [anon_sym_extern] = ACTIONS(6656),
    [anon_sym___attribute__] = ACTIONS(6656),
    [anon_sym___attribute] = ACTIONS(6656),
    [anon_sym_noreturn] = ACTIONS(6656),
    [anon_sym_LBRACK] = ACTIONS(6659),
    [anon_sym___declspec] = ACTIONS(6656),
    [anon_sym_LBRACE] = ACTIONS(6654),
    [anon_sym_signed] = ACTIONS(6652),
    [anon_sym_unsigned] = ACTIONS(6652),
    [anon_sym_long] = ACTIONS(6652),
    [anon_sym_short] = ACTIONS(6652),
    [anon_sym_static] = ACTIONS(6656),
    [anon_sym_EQ] = ACTIONS(6654),
    [anon_sym_ATautoreleasepool] = ACTIONS(6654),
    [anon_sym_auto] = ACTIONS(6656),
    [anon_sym_register] = ACTIONS(6656),
    [anon_sym_inline] = ACTIONS(6656),
    [anon_sym___inline] = ACTIONS(6656),
    [anon_sym___inline__] = ACTIONS(6656),
    [anon_sym___forceinline] = ACTIONS(6656),
    [anon_sym_thread_local] = ACTIONS(6656),
    [anon_sym___thread] = ACTIONS(6656),
    [anon_sym_CG_EXTERN] = ACTIONS(6656),
    [anon_sym_CG_INLINE] = ACTIONS(6656),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6656),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6656),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6656),
    [anon_sym_IBOutlet] = ACTIONS(6656),
    [anon_sym_IBInspectable] = ACTIONS(6656),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6656),
    [anon_sym_NS_INLINE] = ACTIONS(6656),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6656),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6656),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6656),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6656),
    [anon_sym_const] = ACTIONS(6656),
    [anon_sym_constexpr] = ACTIONS(6656),
    [anon_sym_volatile] = ACTIONS(6656),
    [anon_sym_restrict] = ACTIONS(6656),
    [anon_sym___restrict__] = ACTIONS(6656),
    [anon_sym__Atomic] = ACTIONS(6656),
    [anon_sym__Noreturn] = ACTIONS(6656),
    [anon_sym__Nonnull] = ACTIONS(6656),
    [anon_sym_nullable] = ACTIONS(6656),
    [anon_sym__Complex] = ACTIONS(6656),
    [anon_sym__Nullable] = ACTIONS(6656),
    [anon_sym__Nullable_result] = ACTIONS(6656),
    [anon_sym__Null_unspecified] = ACTIONS(6656),
    [anon_sym___autoreleasing] = ACTIONS(6656),
    [anon_sym___block] = ACTIONS(6656),
    [anon_sym___bridge] = ACTIONS(6656),
    [anon_sym___bridge_retained] = ACTIONS(6656),
    [anon_sym___bridge_transfer] = ACTIONS(6656),
    [anon_sym___complex] = ACTIONS(6656),
    [anon_sym___const] = ACTIONS(6656),
    [anon_sym___imag] = ACTIONS(6656),
    [anon_sym___kindof] = ACTIONS(6656),
    [anon_sym___nonnull] = ACTIONS(6656),
    [anon_sym___nullable] = ACTIONS(6656),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6656),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6656),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6656),
    [anon_sym___real] = ACTIONS(6656),
    [anon_sym___strong] = ACTIONS(6656),
    [anon_sym___unsafe_unretained] = ACTIONS(6656),
    [anon_sym___unused] = ACTIONS(6656),
    [anon_sym___weak] = ACTIONS(6656),
    [anon_sym_alignas] = ACTIONS(6656),
    [anon_sym__Alignas] = ACTIONS(6656),
    [sym_primitive_type] = ACTIONS(6652),
    [anon_sym_enum] = ACTIONS(6652),
    [anon_sym_struct] = ACTIONS(6652),
    [anon_sym_union] = ACTIONS(6652),
    [anon_sym_asm] = ACTIONS(6662),
    [anon_sym___asm__] = ACTIONS(6662),
    [anon_sym___asm] = ACTIONS(6662),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6652),
    [anon_sym___typeof] = ACTIONS(6652),
    [anon_sym_typeof] = ACTIONS(6652),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6656),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6656),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6656),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6656),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6656),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6656),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6656),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6656),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6656),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6656),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6656),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6656),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6656),
    [anon_sym_API_AVAILABLE] = ACTIONS(6656),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6656),
    [anon_sym_API_DEPRECATED] = ACTIONS(6656),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6656),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6656),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6656),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6656),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6656),
    [anon_sym___deprecated_msg] = ACTIONS(6656),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6656),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6656),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6656),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6656),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6656),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6656),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6656),
    [anon_sym_BOOL] = ACTIONS(6652),
    [anon_sym_IMP] = ACTIONS(6652),
    [anon_sym_SEL] = ACTIONS(6652),
    [anon_sym_Class] = ACTIONS(6652),
    [anon_sym_id] = ACTIONS(6652),
  },
  [STATE(2670)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6578),
    [sym__type_declarator] = STATE(3250),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5690),
    [anon_sym_CARET] = ACTIONS(5692),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2671)] = {
    [sym_identifier] = ACTIONS(6662),
    [anon_sym_COMMA] = ACTIONS(6654),
    [anon_sym_LPAREN2] = ACTIONS(6654),
    [anon_sym_SEMI] = ACTIONS(6654),
    [anon_sym___extension__] = ACTIONS(6662),
    [anon_sym_extern] = ACTIONS(6662),
    [anon_sym___attribute__] = ACTIONS(6662),
    [anon_sym___attribute] = ACTIONS(6662),
    [anon_sym_noreturn] = ACTIONS(6662),
    [anon_sym_LBRACK] = ACTIONS(6654),
    [anon_sym___declspec] = ACTIONS(6662),
    [anon_sym_LBRACE] = ACTIONS(6654),
    [anon_sym_signed] = ACTIONS(6662),
    [anon_sym_unsigned] = ACTIONS(6662),
    [anon_sym_long] = ACTIONS(6662),
    [anon_sym_short] = ACTIONS(6662),
    [anon_sym_static] = ACTIONS(6662),
    [anon_sym_EQ] = ACTIONS(6654),
    [anon_sym_ATautoreleasepool] = ACTIONS(6654),
    [anon_sym_auto] = ACTIONS(6662),
    [anon_sym_register] = ACTIONS(6662),
    [anon_sym_inline] = ACTIONS(6662),
    [anon_sym___inline] = ACTIONS(6662),
    [anon_sym___inline__] = ACTIONS(6662),
    [anon_sym___forceinline] = ACTIONS(6662),
    [anon_sym_thread_local] = ACTIONS(6662),
    [anon_sym___thread] = ACTIONS(6662),
    [anon_sym_CG_EXTERN] = ACTIONS(6662),
    [anon_sym_CG_INLINE] = ACTIONS(6662),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6662),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6662),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6662),
    [anon_sym_IBOutlet] = ACTIONS(6662),
    [anon_sym_IBInspectable] = ACTIONS(6662),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6662),
    [anon_sym_NS_INLINE] = ACTIONS(6662),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6662),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6662),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6662),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6662),
    [anon_sym_const] = ACTIONS(6662),
    [anon_sym_constexpr] = ACTIONS(6662),
    [anon_sym_volatile] = ACTIONS(6662),
    [anon_sym_restrict] = ACTIONS(6662),
    [anon_sym___restrict__] = ACTIONS(6662),
    [anon_sym__Atomic] = ACTIONS(6662),
    [anon_sym__Noreturn] = ACTIONS(6662),
    [anon_sym__Nonnull] = ACTIONS(6662),
    [anon_sym_nullable] = ACTIONS(6662),
    [anon_sym__Complex] = ACTIONS(6662),
    [anon_sym__Nullable] = ACTIONS(6662),
    [anon_sym__Nullable_result] = ACTIONS(6662),
    [anon_sym__Null_unspecified] = ACTIONS(6662),
    [anon_sym___autoreleasing] = ACTIONS(6662),
    [anon_sym___block] = ACTIONS(6662),
    [anon_sym___bridge] = ACTIONS(6662),
    [anon_sym___bridge_retained] = ACTIONS(6662),
    [anon_sym___bridge_transfer] = ACTIONS(6662),
    [anon_sym___complex] = ACTIONS(6662),
    [anon_sym___const] = ACTIONS(6662),
    [anon_sym___imag] = ACTIONS(6662),
    [anon_sym___kindof] = ACTIONS(6662),
    [anon_sym___nonnull] = ACTIONS(6662),
    [anon_sym___nullable] = ACTIONS(6662),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6662),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6662),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6662),
    [anon_sym___real] = ACTIONS(6662),
    [anon_sym___strong] = ACTIONS(6662),
    [anon_sym___unsafe_unretained] = ACTIONS(6662),
    [anon_sym___unused] = ACTIONS(6662),
    [anon_sym___weak] = ACTIONS(6662),
    [anon_sym_alignas] = ACTIONS(6662),
    [anon_sym__Alignas] = ACTIONS(6662),
    [sym_primitive_type] = ACTIONS(6662),
    [anon_sym_enum] = ACTIONS(6662),
    [anon_sym_struct] = ACTIONS(6662),
    [anon_sym_union] = ACTIONS(6662),
    [anon_sym_asm] = ACTIONS(6662),
    [anon_sym___asm__] = ACTIONS(6662),
    [anon_sym___asm] = ACTIONS(6662),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6662),
    [anon_sym___typeof] = ACTIONS(6662),
    [anon_sym_typeof] = ACTIONS(6662),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6662),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6662),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6662),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6662),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6662),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6662),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6662),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6662),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6662),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6662),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6662),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6662),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6662),
    [anon_sym_API_AVAILABLE] = ACTIONS(6662),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6662),
    [anon_sym_API_DEPRECATED] = ACTIONS(6662),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6662),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6662),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6662),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6662),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6662),
    [anon_sym___deprecated_msg] = ACTIONS(6662),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6662),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6662),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6662),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6662),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6662),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6662),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6662),
    [anon_sym_BOOL] = ACTIONS(6662),
    [anon_sym_IMP] = ACTIONS(6662),
    [anon_sym_SEL] = ACTIONS(6662),
    [anon_sym_Class] = ACTIONS(6662),
    [anon_sym_id] = ACTIONS(6662),
  },
  [STATE(2672)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6578),
    [sym__type_declarator] = STATE(3260),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5690),
    [anon_sym_CARET] = ACTIONS(5692),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2673)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6343),
    [sym__type_declarator] = STATE(5276),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5698),
    [anon_sym_CARET] = ACTIONS(5700),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2674)] = {
    [sym_identifier] = ACTIONS(6664),
    [anon_sym_COMMA] = ACTIONS(6650),
    [anon_sym_LPAREN2] = ACTIONS(6650),
    [anon_sym_SEMI] = ACTIONS(6650),
    [anon_sym___extension__] = ACTIONS(6666),
    [anon_sym_extern] = ACTIONS(6666),
    [anon_sym___attribute__] = ACTIONS(6666),
    [anon_sym___attribute] = ACTIONS(6666),
    [anon_sym_noreturn] = ACTIONS(6666),
    [anon_sym_LBRACK] = ACTIONS(6669),
    [anon_sym___declspec] = ACTIONS(6666),
    [anon_sym_LBRACE] = ACTIONS(6650),
    [anon_sym_signed] = ACTIONS(6664),
    [anon_sym_unsigned] = ACTIONS(6664),
    [anon_sym_long] = ACTIONS(6664),
    [anon_sym_short] = ACTIONS(6664),
    [anon_sym_static] = ACTIONS(6666),
    [anon_sym_EQ] = ACTIONS(6650),
    [anon_sym_ATautoreleasepool] = ACTIONS(6650),
    [anon_sym_auto] = ACTIONS(6666),
    [anon_sym_register] = ACTIONS(6666),
    [anon_sym_inline] = ACTIONS(6666),
    [anon_sym___inline] = ACTIONS(6666),
    [anon_sym___inline__] = ACTIONS(6666),
    [anon_sym___forceinline] = ACTIONS(6666),
    [anon_sym_thread_local] = ACTIONS(6666),
    [anon_sym___thread] = ACTIONS(6666),
    [anon_sym_CG_EXTERN] = ACTIONS(6666),
    [anon_sym_CG_INLINE] = ACTIONS(6666),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6666),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6666),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6666),
    [anon_sym_IBOutlet] = ACTIONS(6666),
    [anon_sym_IBInspectable] = ACTIONS(6666),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6666),
    [anon_sym_NS_INLINE] = ACTIONS(6666),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6666),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6666),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6666),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6666),
    [anon_sym_const] = ACTIONS(6666),
    [anon_sym_constexpr] = ACTIONS(6666),
    [anon_sym_volatile] = ACTIONS(6666),
    [anon_sym_restrict] = ACTIONS(6666),
    [anon_sym___restrict__] = ACTIONS(6666),
    [anon_sym__Atomic] = ACTIONS(6666),
    [anon_sym__Noreturn] = ACTIONS(6666),
    [anon_sym__Nonnull] = ACTIONS(6666),
    [anon_sym_nullable] = ACTIONS(6666),
    [anon_sym__Complex] = ACTIONS(6666),
    [anon_sym__Nullable] = ACTIONS(6666),
    [anon_sym__Nullable_result] = ACTIONS(6666),
    [anon_sym__Null_unspecified] = ACTIONS(6666),
    [anon_sym___autoreleasing] = ACTIONS(6666),
    [anon_sym___block] = ACTIONS(6666),
    [anon_sym___bridge] = ACTIONS(6666),
    [anon_sym___bridge_retained] = ACTIONS(6666),
    [anon_sym___bridge_transfer] = ACTIONS(6666),
    [anon_sym___complex] = ACTIONS(6666),
    [anon_sym___const] = ACTIONS(6666),
    [anon_sym___imag] = ACTIONS(6666),
    [anon_sym___kindof] = ACTIONS(6666),
    [anon_sym___nonnull] = ACTIONS(6666),
    [anon_sym___nullable] = ACTIONS(6666),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6666),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6666),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6666),
    [anon_sym___real] = ACTIONS(6666),
    [anon_sym___strong] = ACTIONS(6666),
    [anon_sym___unsafe_unretained] = ACTIONS(6666),
    [anon_sym___unused] = ACTIONS(6666),
    [anon_sym___weak] = ACTIONS(6666),
    [anon_sym_alignas] = ACTIONS(6666),
    [anon_sym__Alignas] = ACTIONS(6666),
    [sym_primitive_type] = ACTIONS(6664),
    [anon_sym_enum] = ACTIONS(6664),
    [anon_sym_struct] = ACTIONS(6664),
    [anon_sym_union] = ACTIONS(6664),
    [anon_sym_asm] = ACTIONS(6648),
    [anon_sym___asm__] = ACTIONS(6648),
    [anon_sym___asm] = ACTIONS(6648),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6664),
    [anon_sym___typeof] = ACTIONS(6664),
    [anon_sym_typeof] = ACTIONS(6664),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6666),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6666),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6666),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6666),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6666),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6666),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6666),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6666),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6666),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6666),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6666),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6666),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6666),
    [anon_sym_API_AVAILABLE] = ACTIONS(6666),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6666),
    [anon_sym_API_DEPRECATED] = ACTIONS(6666),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6666),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6666),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6666),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6666),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6666),
    [anon_sym___deprecated_msg] = ACTIONS(6666),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6666),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6666),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6666),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6666),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6666),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6666),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6666),
    [anon_sym_BOOL] = ACTIONS(6664),
    [anon_sym_IMP] = ACTIONS(6664),
    [anon_sym_SEL] = ACTIONS(6664),
    [anon_sym_Class] = ACTIONS(6664),
    [anon_sym_id] = ACTIONS(6664),
  },
  [STATE(2675)] = {
    [sym_keyword_declarator] = STATE(4949),
    [sym_method_parameter] = STATE(2675),
    [aux_sym_method_definition_repeat1] = STATE(2675),
    [aux_sym_keyword_selector_repeat1] = STATE(4949),
    [sym_identifier] = ACTIONS(6672),
    [anon_sym_COMMA] = ACTIONS(6675),
    [anon_sym_SEMI] = ACTIONS(6677),
    [anon_sym___extension__] = ACTIONS(6680),
    [anon_sym_extern] = ACTIONS(6680),
    [anon_sym___attribute__] = ACTIONS(6680),
    [anon_sym___attribute] = ACTIONS(6680),
    [anon_sym_noreturn] = ACTIONS(6680),
    [anon_sym_LBRACK] = ACTIONS(6675),
    [anon_sym___declspec] = ACTIONS(6680),
    [anon_sym_LBRACE] = ACTIONS(6675),
    [anon_sym_signed] = ACTIONS(6680),
    [anon_sym_unsigned] = ACTIONS(6680),
    [anon_sym_long] = ACTIONS(6680),
    [anon_sym_short] = ACTIONS(6680),
    [anon_sym_static] = ACTIONS(6680),
    [anon_sym_ATautoreleasepool] = ACTIONS(6675),
    [anon_sym_auto] = ACTIONS(6680),
    [anon_sym_register] = ACTIONS(6680),
    [anon_sym_inline] = ACTIONS(6680),
    [anon_sym___inline] = ACTIONS(6680),
    [anon_sym___inline__] = ACTIONS(6680),
    [anon_sym___forceinline] = ACTIONS(6680),
    [anon_sym_thread_local] = ACTIONS(6680),
    [anon_sym___thread] = ACTIONS(6680),
    [anon_sym_CG_EXTERN] = ACTIONS(6680),
    [anon_sym_CG_INLINE] = ACTIONS(6680),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6680),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6680),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6680),
    [anon_sym_IBOutlet] = ACTIONS(6680),
    [anon_sym_IBInspectable] = ACTIONS(6680),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6680),
    [anon_sym_NS_INLINE] = ACTIONS(6680),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6680),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6680),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6680),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6680),
    [anon_sym_const] = ACTIONS(6680),
    [anon_sym_constexpr] = ACTIONS(6680),
    [anon_sym_volatile] = ACTIONS(6680),
    [anon_sym_restrict] = ACTIONS(6680),
    [anon_sym___restrict__] = ACTIONS(6680),
    [anon_sym__Atomic] = ACTIONS(6680),
    [anon_sym__Noreturn] = ACTIONS(6680),
    [anon_sym__Nonnull] = ACTIONS(6680),
    [anon_sym_nullable] = ACTIONS(6680),
    [anon_sym__Complex] = ACTIONS(6680),
    [anon_sym__Nullable] = ACTIONS(6680),
    [anon_sym__Nullable_result] = ACTIONS(6680),
    [anon_sym__Null_unspecified] = ACTIONS(6680),
    [anon_sym___autoreleasing] = ACTIONS(6680),
    [anon_sym___block] = ACTIONS(6680),
    [anon_sym___bridge] = ACTIONS(6680),
    [anon_sym___bridge_retained] = ACTIONS(6680),
    [anon_sym___bridge_transfer] = ACTIONS(6680),
    [anon_sym___complex] = ACTIONS(6680),
    [anon_sym___const] = ACTIONS(6680),
    [anon_sym___imag] = ACTIONS(6680),
    [anon_sym___kindof] = ACTIONS(6680),
    [anon_sym___nonnull] = ACTIONS(6680),
    [anon_sym___nullable] = ACTIONS(6680),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6680),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6680),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6680),
    [anon_sym___real] = ACTIONS(6680),
    [anon_sym___strong] = ACTIONS(6680),
    [anon_sym___unsafe_unretained] = ACTIONS(6680),
    [anon_sym___unused] = ACTIONS(6680),
    [anon_sym___weak] = ACTIONS(6680),
    [anon_sym_alignas] = ACTIONS(6680),
    [anon_sym__Alignas] = ACTIONS(6680),
    [sym_primitive_type] = ACTIONS(6680),
    [anon_sym_enum] = ACTIONS(6680),
    [anon_sym_COLON] = ACTIONS(6682),
    [anon_sym_struct] = ACTIONS(6680),
    [anon_sym_union] = ACTIONS(6680),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6680),
    [anon_sym___typeof] = ACTIONS(6680),
    [anon_sym_typeof] = ACTIONS(6680),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6680),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6680),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6680),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6680),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6680),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6680),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6680),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6680),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6680),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6680),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6680),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6680),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6680),
    [anon_sym_API_AVAILABLE] = ACTIONS(6680),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6680),
    [anon_sym_API_DEPRECATED] = ACTIONS(6680),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6680),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6680),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6680),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6680),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6680),
    [anon_sym___deprecated_msg] = ACTIONS(6680),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6680),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6680),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6680),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6680),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6680),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6680),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6680),
    [anon_sym_BOOL] = ACTIONS(6680),
    [anon_sym_IMP] = ACTIONS(6680),
    [anon_sym_SEL] = ACTIONS(6680),
    [anon_sym_Class] = ACTIONS(6680),
    [anon_sym_id] = ACTIONS(6680),
  },
  [STATE(2676)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6578),
    [sym__type_declarator] = STATE(3245),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5690),
    [anon_sym_CARET] = ACTIONS(5692),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2677)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6343),
    [sym__type_declarator] = STATE(5164),
    [sym_parenthesized_type_declarator] = STATE(3256),
    [sym_pointer_type_declarator] = STATE(3256),
    [sym_function_type_declarator] = STATE(3256),
    [sym_array_type_declarator] = STATE(3256),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_type_declarator] = STATE(3256),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5686),
    [anon_sym_LPAREN2] = ACTIONS(5688),
    [anon_sym_STAR] = ACTIONS(5698),
    [anon_sym_CARET] = ACTIONS(5700),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(5694),
    [anon_sym_unsigned] = ACTIONS(5694),
    [anon_sym_long] = ACTIONS(5694),
    [anon_sym_short] = ACTIONS(5694),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(5696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2678)] = {
    [sym_attribute_specifier] = STATE(2679),
    [aux_sym_function_declarator_repeat1] = STATE(2679),
    [sym_identifier] = ACTIONS(6685),
    [anon_sym_COMMA] = ACTIONS(6687),
    [anon_sym_LPAREN2] = ACTIONS(6687),
    [anon_sym_SEMI] = ACTIONS(6687),
    [anon_sym___extension__] = ACTIONS(6685),
    [anon_sym_extern] = ACTIONS(6685),
    [anon_sym___attribute__] = ACTIONS(6685),
    [anon_sym___attribute] = ACTIONS(6685),
    [anon_sym_noreturn] = ACTIONS(6685),
    [anon_sym_LBRACK] = ACTIONS(6687),
    [anon_sym___declspec] = ACTIONS(6685),
    [anon_sym_LBRACE] = ACTIONS(6687),
    [anon_sym_signed] = ACTIONS(6685),
    [anon_sym_unsigned] = ACTIONS(6685),
    [anon_sym_long] = ACTIONS(6685),
    [anon_sym_short] = ACTIONS(6685),
    [anon_sym_static] = ACTIONS(6685),
    [anon_sym_EQ] = ACTIONS(6687),
    [anon_sym_ATautoreleasepool] = ACTIONS(6687),
    [anon_sym_auto] = ACTIONS(6685),
    [anon_sym_register] = ACTIONS(6685),
    [anon_sym_inline] = ACTIONS(6685),
    [anon_sym___inline] = ACTIONS(6685),
    [anon_sym___inline__] = ACTIONS(6685),
    [anon_sym___forceinline] = ACTIONS(6685),
    [anon_sym_thread_local] = ACTIONS(6685),
    [anon_sym___thread] = ACTIONS(6685),
    [anon_sym_CG_EXTERN] = ACTIONS(6685),
    [anon_sym_CG_INLINE] = ACTIONS(6685),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6685),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6685),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6685),
    [anon_sym_IBOutlet] = ACTIONS(6685),
    [anon_sym_IBInspectable] = ACTIONS(6685),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6685),
    [anon_sym_NS_INLINE] = ACTIONS(6685),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6685),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6685),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6685),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6685),
    [anon_sym_const] = ACTIONS(6685),
    [anon_sym_constexpr] = ACTIONS(6685),
    [anon_sym_volatile] = ACTIONS(6685),
    [anon_sym_restrict] = ACTIONS(6685),
    [anon_sym___restrict__] = ACTIONS(6685),
    [anon_sym__Atomic] = ACTIONS(6685),
    [anon_sym__Noreturn] = ACTIONS(6685),
    [anon_sym__Nonnull] = ACTIONS(6685),
    [anon_sym_nullable] = ACTIONS(6685),
    [anon_sym__Complex] = ACTIONS(6685),
    [anon_sym__Nullable] = ACTIONS(6685),
    [anon_sym__Nullable_result] = ACTIONS(6685),
    [anon_sym__Null_unspecified] = ACTIONS(6685),
    [anon_sym___autoreleasing] = ACTIONS(6685),
    [anon_sym___block] = ACTIONS(6685),
    [anon_sym___bridge] = ACTIONS(6685),
    [anon_sym___bridge_retained] = ACTIONS(6685),
    [anon_sym___bridge_transfer] = ACTIONS(6685),
    [anon_sym___complex] = ACTIONS(6685),
    [anon_sym___const] = ACTIONS(6685),
    [anon_sym___imag] = ACTIONS(6685),
    [anon_sym___kindof] = ACTIONS(6685),
    [anon_sym___nonnull] = ACTIONS(6685),
    [anon_sym___nullable] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6685),
    [anon_sym___real] = ACTIONS(6685),
    [anon_sym___strong] = ACTIONS(6685),
    [anon_sym___unsafe_unretained] = ACTIONS(6685),
    [anon_sym___unused] = ACTIONS(6685),
    [anon_sym___weak] = ACTIONS(6685),
    [anon_sym_alignas] = ACTIONS(6685),
    [anon_sym__Alignas] = ACTIONS(6685),
    [sym_primitive_type] = ACTIONS(6685),
    [anon_sym_enum] = ACTIONS(6685),
    [anon_sym_struct] = ACTIONS(6685),
    [anon_sym_union] = ACTIONS(6685),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6685),
    [anon_sym___typeof] = ACTIONS(6685),
    [anon_sym_typeof] = ACTIONS(6685),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6685),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6685),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6685),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6685),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6685),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6685),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6685),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6685),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6685),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6685),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6685),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6685),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6685),
    [anon_sym_API_AVAILABLE] = ACTIONS(6685),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6685),
    [anon_sym_API_DEPRECATED] = ACTIONS(6685),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6685),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6685),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6685),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6685),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6685),
    [anon_sym___deprecated_msg] = ACTIONS(6685),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6685),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6685),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6685),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6685),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6685),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6685),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6685),
    [anon_sym_BOOL] = ACTIONS(6685),
    [anon_sym_IMP] = ACTIONS(6685),
    [anon_sym_SEL] = ACTIONS(6685),
    [anon_sym_Class] = ACTIONS(6685),
    [anon_sym_id] = ACTIONS(6685),
  },
  [STATE(2679)] = {
    [sym_attribute_specifier] = STATE(2679),
    [aux_sym_function_declarator_repeat1] = STATE(2679),
    [sym_identifier] = ACTIONS(6689),
    [anon_sym_COMMA] = ACTIONS(6691),
    [anon_sym_LPAREN2] = ACTIONS(6691),
    [anon_sym_SEMI] = ACTIONS(6691),
    [anon_sym___extension__] = ACTIONS(6689),
    [anon_sym_extern] = ACTIONS(6689),
    [anon_sym___attribute__] = ACTIONS(6693),
    [anon_sym___attribute] = ACTIONS(6693),
    [anon_sym_noreturn] = ACTIONS(6689),
    [anon_sym_LBRACK] = ACTIONS(6691),
    [anon_sym___declspec] = ACTIONS(6689),
    [anon_sym_LBRACE] = ACTIONS(6691),
    [anon_sym_signed] = ACTIONS(6689),
    [anon_sym_unsigned] = ACTIONS(6689),
    [anon_sym_long] = ACTIONS(6689),
    [anon_sym_short] = ACTIONS(6689),
    [anon_sym_static] = ACTIONS(6689),
    [anon_sym_EQ] = ACTIONS(6691),
    [anon_sym_ATautoreleasepool] = ACTIONS(6691),
    [anon_sym_auto] = ACTIONS(6689),
    [anon_sym_register] = ACTIONS(6689),
    [anon_sym_inline] = ACTIONS(6689),
    [anon_sym___inline] = ACTIONS(6689),
    [anon_sym___inline__] = ACTIONS(6689),
    [anon_sym___forceinline] = ACTIONS(6689),
    [anon_sym_thread_local] = ACTIONS(6689),
    [anon_sym___thread] = ACTIONS(6689),
    [anon_sym_CG_EXTERN] = ACTIONS(6689),
    [anon_sym_CG_INLINE] = ACTIONS(6689),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6689),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6689),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6689),
    [anon_sym_IBOutlet] = ACTIONS(6689),
    [anon_sym_IBInspectable] = ACTIONS(6689),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6689),
    [anon_sym_NS_INLINE] = ACTIONS(6689),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6689),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6689),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6689),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6689),
    [anon_sym_const] = ACTIONS(6689),
    [anon_sym_constexpr] = ACTIONS(6689),
    [anon_sym_volatile] = ACTIONS(6689),
    [anon_sym_restrict] = ACTIONS(6689),
    [anon_sym___restrict__] = ACTIONS(6689),
    [anon_sym__Atomic] = ACTIONS(6689),
    [anon_sym__Noreturn] = ACTIONS(6689),
    [anon_sym__Nonnull] = ACTIONS(6689),
    [anon_sym_nullable] = ACTIONS(6689),
    [anon_sym__Complex] = ACTIONS(6689),
    [anon_sym__Nullable] = ACTIONS(6689),
    [anon_sym__Nullable_result] = ACTIONS(6689),
    [anon_sym__Null_unspecified] = ACTIONS(6689),
    [anon_sym___autoreleasing] = ACTIONS(6689),
    [anon_sym___block] = ACTIONS(6689),
    [anon_sym___bridge] = ACTIONS(6689),
    [anon_sym___bridge_retained] = ACTIONS(6689),
    [anon_sym___bridge_transfer] = ACTIONS(6689),
    [anon_sym___complex] = ACTIONS(6689),
    [anon_sym___const] = ACTIONS(6689),
    [anon_sym___imag] = ACTIONS(6689),
    [anon_sym___kindof] = ACTIONS(6689),
    [anon_sym___nonnull] = ACTIONS(6689),
    [anon_sym___nullable] = ACTIONS(6689),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6689),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6689),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6689),
    [anon_sym___real] = ACTIONS(6689),
    [anon_sym___strong] = ACTIONS(6689),
    [anon_sym___unsafe_unretained] = ACTIONS(6689),
    [anon_sym___unused] = ACTIONS(6689),
    [anon_sym___weak] = ACTIONS(6689),
    [anon_sym_alignas] = ACTIONS(6689),
    [anon_sym__Alignas] = ACTIONS(6689),
    [sym_primitive_type] = ACTIONS(6689),
    [anon_sym_enum] = ACTIONS(6689),
    [anon_sym_struct] = ACTIONS(6689),
    [anon_sym_union] = ACTIONS(6689),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6689),
    [anon_sym___typeof] = ACTIONS(6689),
    [anon_sym_typeof] = ACTIONS(6689),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6689),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6689),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6689),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6689),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6689),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6689),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6689),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6689),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6689),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6689),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6689),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6689),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6689),
    [anon_sym_API_AVAILABLE] = ACTIONS(6689),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6689),
    [anon_sym_API_DEPRECATED] = ACTIONS(6689),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6689),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6689),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6689),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6689),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6689),
    [anon_sym___deprecated_msg] = ACTIONS(6689),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6689),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6689),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6689),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6689),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6689),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6689),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6689),
    [anon_sym_BOOL] = ACTIONS(6689),
    [anon_sym_IMP] = ACTIONS(6689),
    [anon_sym_SEL] = ACTIONS(6689),
    [anon_sym_Class] = ACTIONS(6689),
    [anon_sym_id] = ACTIONS(6689),
  },
  [STATE(2680)] = {
    [sym_identifier] = ACTIONS(4258),
    [anon_sym_LPAREN2] = ACTIONS(4260),
    [anon_sym___extension__] = ACTIONS(4258),
    [anon_sym_extern] = ACTIONS(4258),
    [anon_sym___attribute__] = ACTIONS(4258),
    [anon_sym___attribute] = ACTIONS(4258),
    [anon_sym_noreturn] = ACTIONS(4258),
    [anon_sym_LBRACK] = ACTIONS(4260),
    [anon_sym___declspec] = ACTIONS(4258),
    [anon_sym___cdecl] = ACTIONS(4258),
    [anon_sym___clrcall] = ACTIONS(4258),
    [anon_sym___stdcall] = ACTIONS(4258),
    [anon_sym___fastcall] = ACTIONS(4258),
    [anon_sym___thiscall] = ACTIONS(4258),
    [anon_sym___vectorcall] = ACTIONS(4258),
    [anon_sym_LBRACE] = ACTIONS(4260),
    [anon_sym_signed] = ACTIONS(4258),
    [anon_sym_unsigned] = ACTIONS(4258),
    [anon_sym_long] = ACTIONS(4258),
    [anon_sym_short] = ACTIONS(4258),
    [anon_sym_static] = ACTIONS(4258),
    [anon_sym_auto] = ACTIONS(4258),
    [anon_sym_register] = ACTIONS(4258),
    [anon_sym_inline] = ACTIONS(4258),
    [anon_sym___inline] = ACTIONS(4258),
    [anon_sym___inline__] = ACTIONS(4258),
    [anon_sym___forceinline] = ACTIONS(4258),
    [anon_sym_thread_local] = ACTIONS(4258),
    [anon_sym___thread] = ACTIONS(4258),
    [anon_sym_CG_EXTERN] = ACTIONS(4258),
    [anon_sym_CG_INLINE] = ACTIONS(4258),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4258),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4258),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4258),
    [anon_sym_IBOutlet] = ACTIONS(4258),
    [anon_sym_IBInspectable] = ACTIONS(4258),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4258),
    [anon_sym_NS_INLINE] = ACTIONS(4258),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4258),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4258),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4258),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4258),
    [anon_sym_const] = ACTIONS(4258),
    [anon_sym_constexpr] = ACTIONS(4258),
    [anon_sym_volatile] = ACTIONS(4258),
    [anon_sym_restrict] = ACTIONS(4258),
    [anon_sym___restrict__] = ACTIONS(4258),
    [anon_sym__Atomic] = ACTIONS(4258),
    [anon_sym__Noreturn] = ACTIONS(4258),
    [anon_sym__Nonnull] = ACTIONS(4258),
    [anon_sym_nullable] = ACTIONS(4258),
    [anon_sym__Complex] = ACTIONS(4258),
    [anon_sym__Nullable] = ACTIONS(4258),
    [anon_sym__Nullable_result] = ACTIONS(4258),
    [anon_sym__Null_unspecified] = ACTIONS(4258),
    [anon_sym___autoreleasing] = ACTIONS(4258),
    [anon_sym___block] = ACTIONS(4258),
    [anon_sym___bridge] = ACTIONS(4258),
    [anon_sym___bridge_retained] = ACTIONS(4258),
    [anon_sym___bridge_transfer] = ACTIONS(4258),
    [anon_sym___complex] = ACTIONS(4258),
    [anon_sym___const] = ACTIONS(4258),
    [anon_sym___imag] = ACTIONS(4258),
    [anon_sym___kindof] = ACTIONS(4258),
    [anon_sym___nonnull] = ACTIONS(4258),
    [anon_sym___nullable] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4258),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4258),
    [anon_sym___real] = ACTIONS(4258),
    [anon_sym___strong] = ACTIONS(4258),
    [anon_sym___unsafe_unretained] = ACTIONS(4258),
    [anon_sym___unused] = ACTIONS(4258),
    [anon_sym___weak] = ACTIONS(4258),
    [anon_sym_alignas] = ACTIONS(4258),
    [anon_sym__Alignas] = ACTIONS(4258),
    [sym_primitive_type] = ACTIONS(4258),
    [anon_sym_enum] = ACTIONS(4258),
    [anon_sym_struct] = ACTIONS(4258),
    [anon_sym_union] = ACTIONS(4258),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4258),
    [anon_sym___typeof] = ACTIONS(4258),
    [anon_sym_typeof] = ACTIONS(4258),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4258),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4258),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4258),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4258),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4258),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4258),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4258),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_API_AVAILABLE] = ACTIONS(4258),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_API_DEPRECATED] = ACTIONS(4258),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4258),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4258),
    [anon_sym___deprecated_msg] = ACTIONS(4258),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4258),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4258),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4258),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4258),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4258),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4258),
    [anon_sym_BOOL] = ACTIONS(4258),
    [anon_sym_IMP] = ACTIONS(4258),
    [anon_sym_SEL] = ACTIONS(4258),
    [anon_sym_Class] = ACTIONS(4258),
    [anon_sym_id] = ACTIONS(4258),
  },
  [STATE(2681)] = {
    [sym_attribute_specifier] = STATE(2679),
    [aux_sym_function_declarator_repeat1] = STATE(2679),
    [sym_identifier] = ACTIONS(6696),
    [anon_sym_COMMA] = ACTIONS(6698),
    [anon_sym_LPAREN2] = ACTIONS(6698),
    [anon_sym_SEMI] = ACTIONS(6698),
    [anon_sym___extension__] = ACTIONS(6696),
    [anon_sym_extern] = ACTIONS(6696),
    [anon_sym___attribute__] = ACTIONS(6696),
    [anon_sym___attribute] = ACTIONS(6696),
    [anon_sym_noreturn] = ACTIONS(6696),
    [anon_sym_LBRACK] = ACTIONS(6698),
    [anon_sym___declspec] = ACTIONS(6696),
    [anon_sym_LBRACE] = ACTIONS(6698),
    [anon_sym_signed] = ACTIONS(6696),
    [anon_sym_unsigned] = ACTIONS(6696),
    [anon_sym_long] = ACTIONS(6696),
    [anon_sym_short] = ACTIONS(6696),
    [anon_sym_static] = ACTIONS(6696),
    [anon_sym_EQ] = ACTIONS(6698),
    [anon_sym_ATautoreleasepool] = ACTIONS(6698),
    [anon_sym_auto] = ACTIONS(6696),
    [anon_sym_register] = ACTIONS(6696),
    [anon_sym_inline] = ACTIONS(6696),
    [anon_sym___inline] = ACTIONS(6696),
    [anon_sym___inline__] = ACTIONS(6696),
    [anon_sym___forceinline] = ACTIONS(6696),
    [anon_sym_thread_local] = ACTIONS(6696),
    [anon_sym___thread] = ACTIONS(6696),
    [anon_sym_CG_EXTERN] = ACTIONS(6696),
    [anon_sym_CG_INLINE] = ACTIONS(6696),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6696),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6696),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6696),
    [anon_sym_IBOutlet] = ACTIONS(6696),
    [anon_sym_IBInspectable] = ACTIONS(6696),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6696),
    [anon_sym_NS_INLINE] = ACTIONS(6696),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6696),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6696),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6696),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6696),
    [anon_sym_const] = ACTIONS(6696),
    [anon_sym_constexpr] = ACTIONS(6696),
    [anon_sym_volatile] = ACTIONS(6696),
    [anon_sym_restrict] = ACTIONS(6696),
    [anon_sym___restrict__] = ACTIONS(6696),
    [anon_sym__Atomic] = ACTIONS(6696),
    [anon_sym__Noreturn] = ACTIONS(6696),
    [anon_sym__Nonnull] = ACTIONS(6696),
    [anon_sym_nullable] = ACTIONS(6696),
    [anon_sym__Complex] = ACTIONS(6696),
    [anon_sym__Nullable] = ACTIONS(6696),
    [anon_sym__Nullable_result] = ACTIONS(6696),
    [anon_sym__Null_unspecified] = ACTIONS(6696),
    [anon_sym___autoreleasing] = ACTIONS(6696),
    [anon_sym___block] = ACTIONS(6696),
    [anon_sym___bridge] = ACTIONS(6696),
    [anon_sym___bridge_retained] = ACTIONS(6696),
    [anon_sym___bridge_transfer] = ACTIONS(6696),
    [anon_sym___complex] = ACTIONS(6696),
    [anon_sym___const] = ACTIONS(6696),
    [anon_sym___imag] = ACTIONS(6696),
    [anon_sym___kindof] = ACTIONS(6696),
    [anon_sym___nonnull] = ACTIONS(6696),
    [anon_sym___nullable] = ACTIONS(6696),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6696),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6696),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6696),
    [anon_sym___real] = ACTIONS(6696),
    [anon_sym___strong] = ACTIONS(6696),
    [anon_sym___unsafe_unretained] = ACTIONS(6696),
    [anon_sym___unused] = ACTIONS(6696),
    [anon_sym___weak] = ACTIONS(6696),
    [anon_sym_alignas] = ACTIONS(6696),
    [anon_sym__Alignas] = ACTIONS(6696),
    [sym_primitive_type] = ACTIONS(6696),
    [anon_sym_enum] = ACTIONS(6696),
    [anon_sym_struct] = ACTIONS(6696),
    [anon_sym_union] = ACTIONS(6696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6696),
    [anon_sym___typeof] = ACTIONS(6696),
    [anon_sym_typeof] = ACTIONS(6696),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6696),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6696),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6696),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6696),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6696),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6696),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6696),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6696),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6696),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6696),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6696),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6696),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6696),
    [anon_sym_API_AVAILABLE] = ACTIONS(6696),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6696),
    [anon_sym_API_DEPRECATED] = ACTIONS(6696),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6696),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6696),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6696),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6696),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6696),
    [anon_sym___deprecated_msg] = ACTIONS(6696),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6696),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6696),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6696),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6696),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6696),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6696),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6696),
    [anon_sym_BOOL] = ACTIONS(6696),
    [anon_sym_IMP] = ACTIONS(6696),
    [anon_sym_SEL] = ACTIONS(6696),
    [anon_sym_Class] = ACTIONS(6696),
    [anon_sym_id] = ACTIONS(6696),
  },
  [STATE(2682)] = {
    [sym_attribute_specifier] = STATE(2681),
    [aux_sym_function_declarator_repeat1] = STATE(2681),
    [sym_identifier] = ACTIONS(6685),
    [anon_sym_COMMA] = ACTIONS(6687),
    [anon_sym_LPAREN2] = ACTIONS(6687),
    [anon_sym_SEMI] = ACTIONS(6687),
    [anon_sym___extension__] = ACTIONS(6685),
    [anon_sym_extern] = ACTIONS(6685),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6685),
    [anon_sym_LBRACK] = ACTIONS(6687),
    [anon_sym___declspec] = ACTIONS(6685),
    [anon_sym_LBRACE] = ACTIONS(6687),
    [anon_sym_signed] = ACTIONS(6685),
    [anon_sym_unsigned] = ACTIONS(6685),
    [anon_sym_long] = ACTIONS(6685),
    [anon_sym_short] = ACTIONS(6685),
    [anon_sym_static] = ACTIONS(6685),
    [anon_sym_EQ] = ACTIONS(6687),
    [anon_sym_ATautoreleasepool] = ACTIONS(6687),
    [anon_sym_auto] = ACTIONS(6685),
    [anon_sym_register] = ACTIONS(6685),
    [anon_sym_inline] = ACTIONS(6685),
    [anon_sym___inline] = ACTIONS(6685),
    [anon_sym___inline__] = ACTIONS(6685),
    [anon_sym___forceinline] = ACTIONS(6685),
    [anon_sym_thread_local] = ACTIONS(6685),
    [anon_sym___thread] = ACTIONS(6685),
    [anon_sym_CG_EXTERN] = ACTIONS(6685),
    [anon_sym_CG_INLINE] = ACTIONS(6685),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6685),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6685),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6685),
    [anon_sym_IBOutlet] = ACTIONS(6685),
    [anon_sym_IBInspectable] = ACTIONS(6685),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6685),
    [anon_sym_NS_INLINE] = ACTIONS(6685),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6685),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6685),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6685),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6685),
    [anon_sym_const] = ACTIONS(6685),
    [anon_sym_constexpr] = ACTIONS(6685),
    [anon_sym_volatile] = ACTIONS(6685),
    [anon_sym_restrict] = ACTIONS(6685),
    [anon_sym___restrict__] = ACTIONS(6685),
    [anon_sym__Atomic] = ACTIONS(6685),
    [anon_sym__Noreturn] = ACTIONS(6685),
    [anon_sym__Nonnull] = ACTIONS(6685),
    [anon_sym_nullable] = ACTIONS(6685),
    [anon_sym__Complex] = ACTIONS(6685),
    [anon_sym__Nullable] = ACTIONS(6685),
    [anon_sym__Nullable_result] = ACTIONS(6685),
    [anon_sym__Null_unspecified] = ACTIONS(6685),
    [anon_sym___autoreleasing] = ACTIONS(6685),
    [anon_sym___block] = ACTIONS(6685),
    [anon_sym___bridge] = ACTIONS(6685),
    [anon_sym___bridge_retained] = ACTIONS(6685),
    [anon_sym___bridge_transfer] = ACTIONS(6685),
    [anon_sym___complex] = ACTIONS(6685),
    [anon_sym___const] = ACTIONS(6685),
    [anon_sym___imag] = ACTIONS(6685),
    [anon_sym___kindof] = ACTIONS(6685),
    [anon_sym___nonnull] = ACTIONS(6685),
    [anon_sym___nullable] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6685),
    [anon_sym___real] = ACTIONS(6685),
    [anon_sym___strong] = ACTIONS(6685),
    [anon_sym___unsafe_unretained] = ACTIONS(6685),
    [anon_sym___unused] = ACTIONS(6685),
    [anon_sym___weak] = ACTIONS(6685),
    [anon_sym_alignas] = ACTIONS(6685),
    [anon_sym__Alignas] = ACTIONS(6685),
    [sym_primitive_type] = ACTIONS(6685),
    [anon_sym_enum] = ACTIONS(6685),
    [anon_sym_struct] = ACTIONS(6685),
    [anon_sym_union] = ACTIONS(6685),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6685),
    [anon_sym___typeof] = ACTIONS(6685),
    [anon_sym_typeof] = ACTIONS(6685),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6685),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6685),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6685),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6685),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6685),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6685),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6685),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6685),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6685),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6685),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6685),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6685),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6685),
    [anon_sym_API_AVAILABLE] = ACTIONS(6685),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6685),
    [anon_sym_API_DEPRECATED] = ACTIONS(6685),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6685),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6685),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6685),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6685),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6685),
    [anon_sym___deprecated_msg] = ACTIONS(6685),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6685),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6685),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6685),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6685),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6685),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6685),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6685),
    [anon_sym_BOOL] = ACTIONS(6685),
    [anon_sym_IMP] = ACTIONS(6685),
    [anon_sym_SEL] = ACTIONS(6685),
    [anon_sym_Class] = ACTIONS(6685),
    [anon_sym_id] = ACTIONS(6685),
  },
  [STATE(2683)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [anon_sym_RPAREN] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(6700),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2220),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2220),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2220),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2220),
    [anon_sym_GT_GT] = ACTIONS(2220),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_EQ] = ACTIONS(2220),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_STAR_EQ] = ACTIONS(2222),
    [anon_sym_SLASH_EQ] = ACTIONS(2222),
    [anon_sym_PERCENT_EQ] = ACTIONS(2222),
    [anon_sym_PLUS_EQ] = ACTIONS(2222),
    [anon_sym_DASH_EQ] = ACTIONS(2222),
    [anon_sym_LT_LT_EQ] = ACTIONS(2222),
    [anon_sym_GT_GT_EQ] = ACTIONS(2222),
    [anon_sym_AMP_EQ] = ACTIONS(2222),
    [anon_sym_CARET_EQ] = ACTIONS(2222),
    [anon_sym_PIPE_EQ] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(2684)] = {
    [sym_attribute_specifier] = STATE(6228),
    [sym_parameter_list] = STATE(2552),
    [sym_identifier] = ACTIONS(6714),
    [anon_sym_COMMA] = ACTIONS(6716),
    [anon_sym_LPAREN2] = ACTIONS(6630),
    [anon_sym_SEMI] = ACTIONS(6716),
    [anon_sym___extension__] = ACTIONS(6714),
    [anon_sym_extern] = ACTIONS(6714),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6714),
    [anon_sym_LBRACK] = ACTIONS(6632),
    [anon_sym___declspec] = ACTIONS(6714),
    [anon_sym_LBRACE] = ACTIONS(6716),
    [anon_sym_signed] = ACTIONS(6714),
    [anon_sym_unsigned] = ACTIONS(6714),
    [anon_sym_long] = ACTIONS(6714),
    [anon_sym_short] = ACTIONS(6714),
    [anon_sym_static] = ACTIONS(6714),
    [anon_sym_EQ] = ACTIONS(6634),
    [anon_sym_ATautoreleasepool] = ACTIONS(6716),
    [anon_sym_auto] = ACTIONS(6714),
    [anon_sym_register] = ACTIONS(6714),
    [anon_sym_inline] = ACTIONS(6714),
    [anon_sym___inline] = ACTIONS(6714),
    [anon_sym___inline__] = ACTIONS(6714),
    [anon_sym___forceinline] = ACTIONS(6714),
    [anon_sym_thread_local] = ACTIONS(6714),
    [anon_sym___thread] = ACTIONS(6714),
    [anon_sym_CG_EXTERN] = ACTIONS(6714),
    [anon_sym_CG_INLINE] = ACTIONS(6714),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6714),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6714),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6714),
    [anon_sym_IBOutlet] = ACTIONS(6714),
    [anon_sym_IBInspectable] = ACTIONS(6714),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6714),
    [anon_sym_NS_INLINE] = ACTIONS(6714),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6714),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6714),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6714),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6714),
    [anon_sym_const] = ACTIONS(6714),
    [anon_sym_constexpr] = ACTIONS(6714),
    [anon_sym_volatile] = ACTIONS(6714),
    [anon_sym_restrict] = ACTIONS(6714),
    [anon_sym___restrict__] = ACTIONS(6714),
    [anon_sym__Atomic] = ACTIONS(6714),
    [anon_sym__Noreturn] = ACTIONS(6714),
    [anon_sym__Nonnull] = ACTIONS(6714),
    [anon_sym_nullable] = ACTIONS(6714),
    [anon_sym__Complex] = ACTIONS(6714),
    [anon_sym__Nullable] = ACTIONS(6714),
    [anon_sym__Nullable_result] = ACTIONS(6714),
    [anon_sym__Null_unspecified] = ACTIONS(6714),
    [anon_sym___autoreleasing] = ACTIONS(6714),
    [anon_sym___block] = ACTIONS(6714),
    [anon_sym___bridge] = ACTIONS(6714),
    [anon_sym___bridge_retained] = ACTIONS(6714),
    [anon_sym___bridge_transfer] = ACTIONS(6714),
    [anon_sym___complex] = ACTIONS(6714),
    [anon_sym___const] = ACTIONS(6714),
    [anon_sym___imag] = ACTIONS(6714),
    [anon_sym___kindof] = ACTIONS(6714),
    [anon_sym___nonnull] = ACTIONS(6714),
    [anon_sym___nullable] = ACTIONS(6714),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6714),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6714),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6714),
    [anon_sym___real] = ACTIONS(6714),
    [anon_sym___strong] = ACTIONS(6714),
    [anon_sym___unsafe_unretained] = ACTIONS(6714),
    [anon_sym___unused] = ACTIONS(6714),
    [anon_sym___weak] = ACTIONS(6714),
    [anon_sym_alignas] = ACTIONS(6714),
    [anon_sym__Alignas] = ACTIONS(6714),
    [sym_primitive_type] = ACTIONS(6714),
    [anon_sym_enum] = ACTIONS(6714),
    [anon_sym_struct] = ACTIONS(6714),
    [anon_sym_union] = ACTIONS(6714),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6714),
    [anon_sym___typeof] = ACTIONS(6714),
    [anon_sym_typeof] = ACTIONS(6714),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6714),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6714),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6714),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6714),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6714),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6714),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6714),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6714),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6714),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6714),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6714),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6714),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6714),
    [anon_sym_API_AVAILABLE] = ACTIONS(6714),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6714),
    [anon_sym_API_DEPRECATED] = ACTIONS(6714),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6714),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6714),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6714),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6714),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6714),
    [anon_sym___deprecated_msg] = ACTIONS(6714),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6714),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6714),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6714),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6714),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6714),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6714),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6714),
    [anon_sym_BOOL] = ACTIONS(6714),
    [anon_sym_IMP] = ACTIONS(6714),
    [anon_sym_SEL] = ACTIONS(6714),
    [anon_sym_Class] = ACTIONS(6714),
    [anon_sym_id] = ACTIONS(6714),
  },
  [STATE(2685)] = {
    [sym_identifier] = ACTIONS(6543),
    [aux_sym_preproc_def_token1] = ACTIONS(6543),
    [aux_sym_preproc_if_token1] = ACTIONS(6543),
    [aux_sym_preproc_if_token2] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6543),
    [sym_preproc_directive] = ACTIONS(6543),
    [anon_sym___extension__] = ACTIONS(6543),
    [anon_sym_extern] = ACTIONS(6543),
    [anon_sym___attribute__] = ACTIONS(6543),
    [anon_sym___attribute] = ACTIONS(6543),
    [anon_sym_noreturn] = ACTIONS(6543),
    [anon_sym_LBRACK] = ACTIONS(6545),
    [anon_sym___declspec] = ACTIONS(6543),
    [anon_sym_signed] = ACTIONS(6543),
    [anon_sym_unsigned] = ACTIONS(6543),
    [anon_sym_long] = ACTIONS(6543),
    [anon_sym_short] = ACTIONS(6543),
    [anon_sym_static] = ACTIONS(6543),
    [anon_sym_auto] = ACTIONS(6543),
    [anon_sym_register] = ACTIONS(6543),
    [anon_sym_inline] = ACTIONS(6543),
    [anon_sym___inline] = ACTIONS(6543),
    [anon_sym___inline__] = ACTIONS(6543),
    [anon_sym___forceinline] = ACTIONS(6543),
    [anon_sym_thread_local] = ACTIONS(6543),
    [anon_sym___thread] = ACTIONS(6543),
    [anon_sym_CG_EXTERN] = ACTIONS(6543),
    [anon_sym_CG_INLINE] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6543),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6543),
    [anon_sym_IBOutlet] = ACTIONS(6543),
    [anon_sym_IBInspectable] = ACTIONS(6543),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6543),
    [anon_sym_NS_INLINE] = ACTIONS(6543),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6543),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6543),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6543),
    [anon_sym_const] = ACTIONS(6543),
    [anon_sym_constexpr] = ACTIONS(6543),
    [anon_sym_volatile] = ACTIONS(6543),
    [anon_sym_restrict] = ACTIONS(6543),
    [anon_sym___restrict__] = ACTIONS(6543),
    [anon_sym__Atomic] = ACTIONS(6543),
    [anon_sym__Noreturn] = ACTIONS(6543),
    [anon_sym__Nonnull] = ACTIONS(6543),
    [anon_sym_nullable] = ACTIONS(6543),
    [anon_sym__Complex] = ACTIONS(6543),
    [anon_sym__Nullable] = ACTIONS(6543),
    [anon_sym__Nullable_result] = ACTIONS(6543),
    [anon_sym__Null_unspecified] = ACTIONS(6543),
    [anon_sym___autoreleasing] = ACTIONS(6543),
    [anon_sym___block] = ACTIONS(6543),
    [anon_sym___bridge] = ACTIONS(6543),
    [anon_sym___bridge_retained] = ACTIONS(6543),
    [anon_sym___bridge_transfer] = ACTIONS(6543),
    [anon_sym___complex] = ACTIONS(6543),
    [anon_sym___const] = ACTIONS(6543),
    [anon_sym___imag] = ACTIONS(6543),
    [anon_sym___kindof] = ACTIONS(6543),
    [anon_sym___nonnull] = ACTIONS(6543),
    [anon_sym___nullable] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6543),
    [anon_sym___real] = ACTIONS(6543),
    [anon_sym___strong] = ACTIONS(6543),
    [anon_sym___unsafe_unretained] = ACTIONS(6543),
    [anon_sym___unused] = ACTIONS(6543),
    [anon_sym___weak] = ACTIONS(6543),
    [anon_sym_alignas] = ACTIONS(6543),
    [anon_sym__Alignas] = ACTIONS(6543),
    [sym_primitive_type] = ACTIONS(6543),
    [anon_sym_enum] = ACTIONS(6543),
    [anon_sym_struct] = ACTIONS(6543),
    [anon_sym_union] = ACTIONS(6543),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6543),
    [anon_sym___typeof] = ACTIONS(6543),
    [anon_sym_typeof] = ACTIONS(6543),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6543),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6543),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6543),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_API_AVAILABLE] = ACTIONS(6543),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_API_DEPRECATED] = ACTIONS(6543),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym___deprecated_msg] = ACTIONS(6543),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6543),
    [anon_sym_BOOL] = ACTIONS(6543),
    [anon_sym_IMP] = ACTIONS(6543),
    [anon_sym_SEL] = ACTIONS(6543),
    [anon_sym_Class] = ACTIONS(6543),
    [anon_sym_id] = ACTIONS(6543),
    [anon_sym_ATdefs] = ACTIONS(6545),
  },
  [STATE(2686)] = {
    [sym_identifier] = ACTIONS(6591),
    [aux_sym_preproc_def_token1] = ACTIONS(6591),
    [aux_sym_preproc_if_token1] = ACTIONS(6591),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6591),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6591),
    [sym_preproc_directive] = ACTIONS(6591),
    [anon_sym___extension__] = ACTIONS(6591),
    [anon_sym_extern] = ACTIONS(6591),
    [anon_sym___attribute__] = ACTIONS(6591),
    [anon_sym___attribute] = ACTIONS(6591),
    [anon_sym_noreturn] = ACTIONS(6591),
    [anon_sym_LBRACK] = ACTIONS(6593),
    [anon_sym___declspec] = ACTIONS(6591),
    [anon_sym_RBRACE] = ACTIONS(6593),
    [anon_sym_signed] = ACTIONS(6591),
    [anon_sym_unsigned] = ACTIONS(6591),
    [anon_sym_long] = ACTIONS(6591),
    [anon_sym_short] = ACTIONS(6591),
    [anon_sym_static] = ACTIONS(6591),
    [anon_sym_auto] = ACTIONS(6591),
    [anon_sym_register] = ACTIONS(6591),
    [anon_sym_inline] = ACTIONS(6591),
    [anon_sym___inline] = ACTIONS(6591),
    [anon_sym___inline__] = ACTIONS(6591),
    [anon_sym___forceinline] = ACTIONS(6591),
    [anon_sym_thread_local] = ACTIONS(6591),
    [anon_sym___thread] = ACTIONS(6591),
    [anon_sym_CG_EXTERN] = ACTIONS(6591),
    [anon_sym_CG_INLINE] = ACTIONS(6591),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6591),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6591),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6591),
    [anon_sym_IBOutlet] = ACTIONS(6591),
    [anon_sym_IBInspectable] = ACTIONS(6591),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6591),
    [anon_sym_NS_INLINE] = ACTIONS(6591),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6591),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6591),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6591),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6591),
    [anon_sym_const] = ACTIONS(6591),
    [anon_sym_constexpr] = ACTIONS(6591),
    [anon_sym_volatile] = ACTIONS(6591),
    [anon_sym_restrict] = ACTIONS(6591),
    [anon_sym___restrict__] = ACTIONS(6591),
    [anon_sym__Atomic] = ACTIONS(6591),
    [anon_sym__Noreturn] = ACTIONS(6591),
    [anon_sym__Nonnull] = ACTIONS(6591),
    [anon_sym_nullable] = ACTIONS(6591),
    [anon_sym__Complex] = ACTIONS(6591),
    [anon_sym__Nullable] = ACTIONS(6591),
    [anon_sym__Nullable_result] = ACTIONS(6591),
    [anon_sym__Null_unspecified] = ACTIONS(6591),
    [anon_sym___autoreleasing] = ACTIONS(6591),
    [anon_sym___block] = ACTIONS(6591),
    [anon_sym___bridge] = ACTIONS(6591),
    [anon_sym___bridge_retained] = ACTIONS(6591),
    [anon_sym___bridge_transfer] = ACTIONS(6591),
    [anon_sym___complex] = ACTIONS(6591),
    [anon_sym___const] = ACTIONS(6591),
    [anon_sym___imag] = ACTIONS(6591),
    [anon_sym___kindof] = ACTIONS(6591),
    [anon_sym___nonnull] = ACTIONS(6591),
    [anon_sym___nullable] = ACTIONS(6591),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6591),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6591),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6591),
    [anon_sym___real] = ACTIONS(6591),
    [anon_sym___strong] = ACTIONS(6591),
    [anon_sym___unsafe_unretained] = ACTIONS(6591),
    [anon_sym___unused] = ACTIONS(6591),
    [anon_sym___weak] = ACTIONS(6591),
    [anon_sym_alignas] = ACTIONS(6591),
    [anon_sym__Alignas] = ACTIONS(6591),
    [sym_primitive_type] = ACTIONS(6591),
    [anon_sym_enum] = ACTIONS(6591),
    [anon_sym_struct] = ACTIONS(6591),
    [anon_sym_union] = ACTIONS(6591),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6591),
    [anon_sym___typeof] = ACTIONS(6591),
    [anon_sym_typeof] = ACTIONS(6591),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6591),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6591),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6591),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6591),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6591),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6591),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6591),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6591),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6591),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_API_AVAILABLE] = ACTIONS(6591),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_API_DEPRECATED] = ACTIONS(6591),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6591),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6591),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6591),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6591),
    [anon_sym___deprecated_msg] = ACTIONS(6591),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6591),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6591),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6591),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6591),
    [anon_sym_BOOL] = ACTIONS(6591),
    [anon_sym_IMP] = ACTIONS(6591),
    [anon_sym_SEL] = ACTIONS(6591),
    [anon_sym_Class] = ACTIONS(6591),
    [anon_sym_id] = ACTIONS(6591),
    [anon_sym_ATdefs] = ACTIONS(6593),
  },
  [STATE(2687)] = {
    [sym_parameter_list] = STATE(2552),
    [sym_identifier] = ACTIONS(6718),
    [anon_sym_COMMA] = ACTIONS(6720),
    [anon_sym_LPAREN2] = ACTIONS(6630),
    [anon_sym_SEMI] = ACTIONS(6720),
    [anon_sym___extension__] = ACTIONS(6718),
    [anon_sym_extern] = ACTIONS(6718),
    [anon_sym___attribute__] = ACTIONS(6718),
    [anon_sym___attribute] = ACTIONS(6718),
    [anon_sym_noreturn] = ACTIONS(6718),
    [anon_sym_LBRACK] = ACTIONS(6632),
    [anon_sym___declspec] = ACTIONS(6718),
    [anon_sym_LBRACE] = ACTIONS(6720),
    [anon_sym_signed] = ACTIONS(6718),
    [anon_sym_unsigned] = ACTIONS(6718),
    [anon_sym_long] = ACTIONS(6718),
    [anon_sym_short] = ACTIONS(6718),
    [anon_sym_static] = ACTIONS(6718),
    [anon_sym_EQ] = ACTIONS(6720),
    [anon_sym_ATautoreleasepool] = ACTIONS(6720),
    [anon_sym_auto] = ACTIONS(6718),
    [anon_sym_register] = ACTIONS(6718),
    [anon_sym_inline] = ACTIONS(6718),
    [anon_sym___inline] = ACTIONS(6718),
    [anon_sym___inline__] = ACTIONS(6718),
    [anon_sym___forceinline] = ACTIONS(6718),
    [anon_sym_thread_local] = ACTIONS(6718),
    [anon_sym___thread] = ACTIONS(6718),
    [anon_sym_CG_EXTERN] = ACTIONS(6718),
    [anon_sym_CG_INLINE] = ACTIONS(6718),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6718),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6718),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6718),
    [anon_sym_IBOutlet] = ACTIONS(6718),
    [anon_sym_IBInspectable] = ACTIONS(6718),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6718),
    [anon_sym_NS_INLINE] = ACTIONS(6718),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6718),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6718),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6718),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6718),
    [anon_sym_const] = ACTIONS(6718),
    [anon_sym_constexpr] = ACTIONS(6718),
    [anon_sym_volatile] = ACTIONS(6718),
    [anon_sym_restrict] = ACTIONS(6718),
    [anon_sym___restrict__] = ACTIONS(6718),
    [anon_sym__Atomic] = ACTIONS(6718),
    [anon_sym__Noreturn] = ACTIONS(6718),
    [anon_sym__Nonnull] = ACTIONS(6718),
    [anon_sym_nullable] = ACTIONS(6718),
    [anon_sym__Complex] = ACTIONS(6718),
    [anon_sym__Nullable] = ACTIONS(6718),
    [anon_sym__Nullable_result] = ACTIONS(6718),
    [anon_sym__Null_unspecified] = ACTIONS(6718),
    [anon_sym___autoreleasing] = ACTIONS(6718),
    [anon_sym___block] = ACTIONS(6718),
    [anon_sym___bridge] = ACTIONS(6718),
    [anon_sym___bridge_retained] = ACTIONS(6718),
    [anon_sym___bridge_transfer] = ACTIONS(6718),
    [anon_sym___complex] = ACTIONS(6718),
    [anon_sym___const] = ACTIONS(6718),
    [anon_sym___imag] = ACTIONS(6718),
    [anon_sym___kindof] = ACTIONS(6718),
    [anon_sym___nonnull] = ACTIONS(6718),
    [anon_sym___nullable] = ACTIONS(6718),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6718),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6718),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6718),
    [anon_sym___real] = ACTIONS(6718),
    [anon_sym___strong] = ACTIONS(6718),
    [anon_sym___unsafe_unretained] = ACTIONS(6718),
    [anon_sym___unused] = ACTIONS(6718),
    [anon_sym___weak] = ACTIONS(6718),
    [anon_sym_alignas] = ACTIONS(6718),
    [anon_sym__Alignas] = ACTIONS(6718),
    [sym_primitive_type] = ACTIONS(6718),
    [anon_sym_enum] = ACTIONS(6718),
    [anon_sym_struct] = ACTIONS(6718),
    [anon_sym_union] = ACTIONS(6718),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6718),
    [anon_sym___typeof] = ACTIONS(6718),
    [anon_sym_typeof] = ACTIONS(6718),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6718),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6718),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6718),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6718),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6718),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6718),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6718),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6718),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6718),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6718),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6718),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6718),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6718),
    [anon_sym_API_AVAILABLE] = ACTIONS(6718),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6718),
    [anon_sym_API_DEPRECATED] = ACTIONS(6718),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6718),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6718),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6718),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6718),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6718),
    [anon_sym___deprecated_msg] = ACTIONS(6718),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6718),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6718),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6718),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6718),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6718),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6718),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6718),
    [anon_sym_BOOL] = ACTIONS(6718),
    [anon_sym_IMP] = ACTIONS(6718),
    [anon_sym_SEL] = ACTIONS(6718),
    [anon_sym_Class] = ACTIONS(6718),
    [anon_sym_id] = ACTIONS(6718),
  },
  [STATE(2688)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_RBRACE] = ACTIONS(6541),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2689)] = {
    [sym_identifier] = ACTIONS(6610),
    [aux_sym_preproc_def_token1] = ACTIONS(6610),
    [aux_sym_preproc_if_token1] = ACTIONS(6610),
    [aux_sym_preproc_if_token2] = ACTIONS(6610),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6610),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6610),
    [sym_preproc_directive] = ACTIONS(6610),
    [anon_sym___extension__] = ACTIONS(6610),
    [anon_sym_extern] = ACTIONS(6610),
    [anon_sym___attribute__] = ACTIONS(6610),
    [anon_sym___attribute] = ACTIONS(6610),
    [anon_sym_noreturn] = ACTIONS(6610),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym___declspec] = ACTIONS(6610),
    [anon_sym_signed] = ACTIONS(6610),
    [anon_sym_unsigned] = ACTIONS(6610),
    [anon_sym_long] = ACTIONS(6610),
    [anon_sym_short] = ACTIONS(6610),
    [anon_sym_static] = ACTIONS(6610),
    [anon_sym_auto] = ACTIONS(6610),
    [anon_sym_register] = ACTIONS(6610),
    [anon_sym_inline] = ACTIONS(6610),
    [anon_sym___inline] = ACTIONS(6610),
    [anon_sym___inline__] = ACTIONS(6610),
    [anon_sym___forceinline] = ACTIONS(6610),
    [anon_sym_thread_local] = ACTIONS(6610),
    [anon_sym___thread] = ACTIONS(6610),
    [anon_sym_CG_EXTERN] = ACTIONS(6610),
    [anon_sym_CG_INLINE] = ACTIONS(6610),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6610),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6610),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6610),
    [anon_sym_IBOutlet] = ACTIONS(6610),
    [anon_sym_IBInspectable] = ACTIONS(6610),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6610),
    [anon_sym_NS_INLINE] = ACTIONS(6610),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6610),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6610),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6610),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6610),
    [anon_sym_const] = ACTIONS(6610),
    [anon_sym_constexpr] = ACTIONS(6610),
    [anon_sym_volatile] = ACTIONS(6610),
    [anon_sym_restrict] = ACTIONS(6610),
    [anon_sym___restrict__] = ACTIONS(6610),
    [anon_sym__Atomic] = ACTIONS(6610),
    [anon_sym__Noreturn] = ACTIONS(6610),
    [anon_sym__Nonnull] = ACTIONS(6610),
    [anon_sym_nullable] = ACTIONS(6610),
    [anon_sym__Complex] = ACTIONS(6610),
    [anon_sym__Nullable] = ACTIONS(6610),
    [anon_sym__Nullable_result] = ACTIONS(6610),
    [anon_sym__Null_unspecified] = ACTIONS(6610),
    [anon_sym___autoreleasing] = ACTIONS(6610),
    [anon_sym___block] = ACTIONS(6610),
    [anon_sym___bridge] = ACTIONS(6610),
    [anon_sym___bridge_retained] = ACTIONS(6610),
    [anon_sym___bridge_transfer] = ACTIONS(6610),
    [anon_sym___complex] = ACTIONS(6610),
    [anon_sym___const] = ACTIONS(6610),
    [anon_sym___imag] = ACTIONS(6610),
    [anon_sym___kindof] = ACTIONS(6610),
    [anon_sym___nonnull] = ACTIONS(6610),
    [anon_sym___nullable] = ACTIONS(6610),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6610),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6610),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6610),
    [anon_sym___real] = ACTIONS(6610),
    [anon_sym___strong] = ACTIONS(6610),
    [anon_sym___unsafe_unretained] = ACTIONS(6610),
    [anon_sym___unused] = ACTIONS(6610),
    [anon_sym___weak] = ACTIONS(6610),
    [anon_sym_alignas] = ACTIONS(6610),
    [anon_sym__Alignas] = ACTIONS(6610),
    [sym_primitive_type] = ACTIONS(6610),
    [anon_sym_enum] = ACTIONS(6610),
    [anon_sym_struct] = ACTIONS(6610),
    [anon_sym_union] = ACTIONS(6610),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6610),
    [anon_sym___typeof] = ACTIONS(6610),
    [anon_sym_typeof] = ACTIONS(6610),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6610),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6610),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6610),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6610),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6610),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6610),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6610),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6610),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6610),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_API_AVAILABLE] = ACTIONS(6610),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_API_DEPRECATED] = ACTIONS(6610),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6610),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6610),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6610),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6610),
    [anon_sym___deprecated_msg] = ACTIONS(6610),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6610),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6610),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6610),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6610),
    [anon_sym_BOOL] = ACTIONS(6610),
    [anon_sym_IMP] = ACTIONS(6610),
    [anon_sym_SEL] = ACTIONS(6610),
    [anon_sym_Class] = ACTIONS(6610),
    [anon_sym_id] = ACTIONS(6610),
    [anon_sym_ATdefs] = ACTIONS(6612),
  },
  [STATE(2690)] = {
    [sym_identifier] = ACTIONS(6603),
    [aux_sym_preproc_def_token1] = ACTIONS(6603),
    [aux_sym_preproc_if_token1] = ACTIONS(6603),
    [aux_sym_preproc_if_token2] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6603),
    [sym_preproc_directive] = ACTIONS(6603),
    [anon_sym___extension__] = ACTIONS(6603),
    [anon_sym_extern] = ACTIONS(6603),
    [anon_sym___attribute__] = ACTIONS(6603),
    [anon_sym___attribute] = ACTIONS(6603),
    [anon_sym_noreturn] = ACTIONS(6603),
    [anon_sym_LBRACK] = ACTIONS(6605),
    [anon_sym___declspec] = ACTIONS(6603),
    [anon_sym_signed] = ACTIONS(6603),
    [anon_sym_unsigned] = ACTIONS(6603),
    [anon_sym_long] = ACTIONS(6603),
    [anon_sym_short] = ACTIONS(6603),
    [anon_sym_static] = ACTIONS(6603),
    [anon_sym_auto] = ACTIONS(6603),
    [anon_sym_register] = ACTIONS(6603),
    [anon_sym_inline] = ACTIONS(6603),
    [anon_sym___inline] = ACTIONS(6603),
    [anon_sym___inline__] = ACTIONS(6603),
    [anon_sym___forceinline] = ACTIONS(6603),
    [anon_sym_thread_local] = ACTIONS(6603),
    [anon_sym___thread] = ACTIONS(6603),
    [anon_sym_CG_EXTERN] = ACTIONS(6603),
    [anon_sym_CG_INLINE] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6603),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6603),
    [anon_sym_IBOutlet] = ACTIONS(6603),
    [anon_sym_IBInspectable] = ACTIONS(6603),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6603),
    [anon_sym_NS_INLINE] = ACTIONS(6603),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6603),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6603),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6603),
    [anon_sym_const] = ACTIONS(6603),
    [anon_sym_constexpr] = ACTIONS(6603),
    [anon_sym_volatile] = ACTIONS(6603),
    [anon_sym_restrict] = ACTIONS(6603),
    [anon_sym___restrict__] = ACTIONS(6603),
    [anon_sym__Atomic] = ACTIONS(6603),
    [anon_sym__Noreturn] = ACTIONS(6603),
    [anon_sym__Nonnull] = ACTIONS(6603),
    [anon_sym_nullable] = ACTIONS(6603),
    [anon_sym__Complex] = ACTIONS(6603),
    [anon_sym__Nullable] = ACTIONS(6603),
    [anon_sym__Nullable_result] = ACTIONS(6603),
    [anon_sym__Null_unspecified] = ACTIONS(6603),
    [anon_sym___autoreleasing] = ACTIONS(6603),
    [anon_sym___block] = ACTIONS(6603),
    [anon_sym___bridge] = ACTIONS(6603),
    [anon_sym___bridge_retained] = ACTIONS(6603),
    [anon_sym___bridge_transfer] = ACTIONS(6603),
    [anon_sym___complex] = ACTIONS(6603),
    [anon_sym___const] = ACTIONS(6603),
    [anon_sym___imag] = ACTIONS(6603),
    [anon_sym___kindof] = ACTIONS(6603),
    [anon_sym___nonnull] = ACTIONS(6603),
    [anon_sym___nullable] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6603),
    [anon_sym___real] = ACTIONS(6603),
    [anon_sym___strong] = ACTIONS(6603),
    [anon_sym___unsafe_unretained] = ACTIONS(6603),
    [anon_sym___unused] = ACTIONS(6603),
    [anon_sym___weak] = ACTIONS(6603),
    [anon_sym_alignas] = ACTIONS(6603),
    [anon_sym__Alignas] = ACTIONS(6603),
    [sym_primitive_type] = ACTIONS(6603),
    [anon_sym_enum] = ACTIONS(6603),
    [anon_sym_struct] = ACTIONS(6603),
    [anon_sym_union] = ACTIONS(6603),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6603),
    [anon_sym___typeof] = ACTIONS(6603),
    [anon_sym_typeof] = ACTIONS(6603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6603),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6603),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6603),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_API_AVAILABLE] = ACTIONS(6603),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_API_DEPRECATED] = ACTIONS(6603),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym___deprecated_msg] = ACTIONS(6603),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6603),
    [anon_sym_BOOL] = ACTIONS(6603),
    [anon_sym_IMP] = ACTIONS(6603),
    [anon_sym_SEL] = ACTIONS(6603),
    [anon_sym_Class] = ACTIONS(6603),
    [anon_sym_id] = ACTIONS(6603),
    [anon_sym_ATdefs] = ACTIONS(6605),
  },
  [STATE(2691)] = {
    [sym_identifier] = ACTIONS(6614),
    [aux_sym_preproc_def_token1] = ACTIONS(6614),
    [aux_sym_preproc_if_token1] = ACTIONS(6614),
    [aux_sym_preproc_if_token2] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6614),
    [sym_preproc_directive] = ACTIONS(6614),
    [anon_sym___extension__] = ACTIONS(6614),
    [anon_sym_extern] = ACTIONS(6614),
    [anon_sym___attribute__] = ACTIONS(6614),
    [anon_sym___attribute] = ACTIONS(6614),
    [anon_sym_noreturn] = ACTIONS(6614),
    [anon_sym_LBRACK] = ACTIONS(6616),
    [anon_sym___declspec] = ACTIONS(6614),
    [anon_sym_signed] = ACTIONS(6614),
    [anon_sym_unsigned] = ACTIONS(6614),
    [anon_sym_long] = ACTIONS(6614),
    [anon_sym_short] = ACTIONS(6614),
    [anon_sym_static] = ACTIONS(6614),
    [anon_sym_auto] = ACTIONS(6614),
    [anon_sym_register] = ACTIONS(6614),
    [anon_sym_inline] = ACTIONS(6614),
    [anon_sym___inline] = ACTIONS(6614),
    [anon_sym___inline__] = ACTIONS(6614),
    [anon_sym___forceinline] = ACTIONS(6614),
    [anon_sym_thread_local] = ACTIONS(6614),
    [anon_sym___thread] = ACTIONS(6614),
    [anon_sym_CG_EXTERN] = ACTIONS(6614),
    [anon_sym_CG_INLINE] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6614),
    [anon_sym_IBOutlet] = ACTIONS(6614),
    [anon_sym_IBInspectable] = ACTIONS(6614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6614),
    [anon_sym_NS_INLINE] = ACTIONS(6614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6614),
    [anon_sym_const] = ACTIONS(6614),
    [anon_sym_constexpr] = ACTIONS(6614),
    [anon_sym_volatile] = ACTIONS(6614),
    [anon_sym_restrict] = ACTIONS(6614),
    [anon_sym___restrict__] = ACTIONS(6614),
    [anon_sym__Atomic] = ACTIONS(6614),
    [anon_sym__Noreturn] = ACTIONS(6614),
    [anon_sym__Nonnull] = ACTIONS(6614),
    [anon_sym_nullable] = ACTIONS(6614),
    [anon_sym__Complex] = ACTIONS(6614),
    [anon_sym__Nullable] = ACTIONS(6614),
    [anon_sym__Nullable_result] = ACTIONS(6614),
    [anon_sym__Null_unspecified] = ACTIONS(6614),
    [anon_sym___autoreleasing] = ACTIONS(6614),
    [anon_sym___block] = ACTIONS(6614),
    [anon_sym___bridge] = ACTIONS(6614),
    [anon_sym___bridge_retained] = ACTIONS(6614),
    [anon_sym___bridge_transfer] = ACTIONS(6614),
    [anon_sym___complex] = ACTIONS(6614),
    [anon_sym___const] = ACTIONS(6614),
    [anon_sym___imag] = ACTIONS(6614),
    [anon_sym___kindof] = ACTIONS(6614),
    [anon_sym___nonnull] = ACTIONS(6614),
    [anon_sym___nullable] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6614),
    [anon_sym___real] = ACTIONS(6614),
    [anon_sym___strong] = ACTIONS(6614),
    [anon_sym___unsafe_unretained] = ACTIONS(6614),
    [anon_sym___unused] = ACTIONS(6614),
    [anon_sym___weak] = ACTIONS(6614),
    [anon_sym_alignas] = ACTIONS(6614),
    [anon_sym__Alignas] = ACTIONS(6614),
    [sym_primitive_type] = ACTIONS(6614),
    [anon_sym_enum] = ACTIONS(6614),
    [anon_sym_struct] = ACTIONS(6614),
    [anon_sym_union] = ACTIONS(6614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6614),
    [anon_sym___typeof] = ACTIONS(6614),
    [anon_sym_typeof] = ACTIONS(6614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_API_AVAILABLE] = ACTIONS(6614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_API_DEPRECATED] = ACTIONS(6614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym___deprecated_msg] = ACTIONS(6614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6614),
    [anon_sym_BOOL] = ACTIONS(6614),
    [anon_sym_IMP] = ACTIONS(6614),
    [anon_sym_SEL] = ACTIONS(6614),
    [anon_sym_Class] = ACTIONS(6614),
    [anon_sym_id] = ACTIONS(6614),
    [anon_sym_ATdefs] = ACTIONS(6616),
  },
  [STATE(2692)] = {
    [sym_identifier] = ACTIONS(6614),
    [aux_sym_preproc_def_token1] = ACTIONS(6614),
    [aux_sym_preproc_if_token1] = ACTIONS(6614),
    [aux_sym_preproc_if_token2] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6614),
    [sym_preproc_directive] = ACTIONS(6614),
    [anon_sym___extension__] = ACTIONS(6614),
    [anon_sym_extern] = ACTIONS(6614),
    [anon_sym___attribute__] = ACTIONS(6614),
    [anon_sym___attribute] = ACTIONS(6614),
    [anon_sym_noreturn] = ACTIONS(6614),
    [anon_sym_LBRACK] = ACTIONS(6616),
    [anon_sym___declspec] = ACTIONS(6614),
    [anon_sym_signed] = ACTIONS(6614),
    [anon_sym_unsigned] = ACTIONS(6614),
    [anon_sym_long] = ACTIONS(6614),
    [anon_sym_short] = ACTIONS(6614),
    [anon_sym_static] = ACTIONS(6614),
    [anon_sym_auto] = ACTIONS(6614),
    [anon_sym_register] = ACTIONS(6614),
    [anon_sym_inline] = ACTIONS(6614),
    [anon_sym___inline] = ACTIONS(6614),
    [anon_sym___inline__] = ACTIONS(6614),
    [anon_sym___forceinline] = ACTIONS(6614),
    [anon_sym_thread_local] = ACTIONS(6614),
    [anon_sym___thread] = ACTIONS(6614),
    [anon_sym_CG_EXTERN] = ACTIONS(6614),
    [anon_sym_CG_INLINE] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6614),
    [anon_sym_IBOutlet] = ACTIONS(6614),
    [anon_sym_IBInspectable] = ACTIONS(6614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6614),
    [anon_sym_NS_INLINE] = ACTIONS(6614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6614),
    [anon_sym_const] = ACTIONS(6614),
    [anon_sym_constexpr] = ACTIONS(6614),
    [anon_sym_volatile] = ACTIONS(6614),
    [anon_sym_restrict] = ACTIONS(6614),
    [anon_sym___restrict__] = ACTIONS(6614),
    [anon_sym__Atomic] = ACTIONS(6614),
    [anon_sym__Noreturn] = ACTIONS(6614),
    [anon_sym__Nonnull] = ACTIONS(6614),
    [anon_sym_nullable] = ACTIONS(6614),
    [anon_sym__Complex] = ACTIONS(6614),
    [anon_sym__Nullable] = ACTIONS(6614),
    [anon_sym__Nullable_result] = ACTIONS(6614),
    [anon_sym__Null_unspecified] = ACTIONS(6614),
    [anon_sym___autoreleasing] = ACTIONS(6614),
    [anon_sym___block] = ACTIONS(6614),
    [anon_sym___bridge] = ACTIONS(6614),
    [anon_sym___bridge_retained] = ACTIONS(6614),
    [anon_sym___bridge_transfer] = ACTIONS(6614),
    [anon_sym___complex] = ACTIONS(6614),
    [anon_sym___const] = ACTIONS(6614),
    [anon_sym___imag] = ACTIONS(6614),
    [anon_sym___kindof] = ACTIONS(6614),
    [anon_sym___nonnull] = ACTIONS(6614),
    [anon_sym___nullable] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6614),
    [anon_sym___real] = ACTIONS(6614),
    [anon_sym___strong] = ACTIONS(6614),
    [anon_sym___unsafe_unretained] = ACTIONS(6614),
    [anon_sym___unused] = ACTIONS(6614),
    [anon_sym___weak] = ACTIONS(6614),
    [anon_sym_alignas] = ACTIONS(6614),
    [anon_sym__Alignas] = ACTIONS(6614),
    [sym_primitive_type] = ACTIONS(6614),
    [anon_sym_enum] = ACTIONS(6614),
    [anon_sym_struct] = ACTIONS(6614),
    [anon_sym_union] = ACTIONS(6614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6614),
    [anon_sym___typeof] = ACTIONS(6614),
    [anon_sym_typeof] = ACTIONS(6614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_API_AVAILABLE] = ACTIONS(6614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_API_DEPRECATED] = ACTIONS(6614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym___deprecated_msg] = ACTIONS(6614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6614),
    [anon_sym_BOOL] = ACTIONS(6614),
    [anon_sym_IMP] = ACTIONS(6614),
    [anon_sym_SEL] = ACTIONS(6614),
    [anon_sym_Class] = ACTIONS(6614),
    [anon_sym_id] = ACTIONS(6614),
    [anon_sym_ATdefs] = ACTIONS(6616),
  },
  [STATE(2693)] = {
    [sym_identifier] = ACTIONS(6603),
    [aux_sym_preproc_def_token1] = ACTIONS(6603),
    [aux_sym_preproc_if_token1] = ACTIONS(6603),
    [aux_sym_preproc_if_token2] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6603),
    [sym_preproc_directive] = ACTIONS(6603),
    [anon_sym___extension__] = ACTIONS(6603),
    [anon_sym_extern] = ACTIONS(6603),
    [anon_sym___attribute__] = ACTIONS(6603),
    [anon_sym___attribute] = ACTIONS(6603),
    [anon_sym_noreturn] = ACTIONS(6603),
    [anon_sym_LBRACK] = ACTIONS(6605),
    [anon_sym___declspec] = ACTIONS(6603),
    [anon_sym_signed] = ACTIONS(6603),
    [anon_sym_unsigned] = ACTIONS(6603),
    [anon_sym_long] = ACTIONS(6603),
    [anon_sym_short] = ACTIONS(6603),
    [anon_sym_static] = ACTIONS(6603),
    [anon_sym_auto] = ACTIONS(6603),
    [anon_sym_register] = ACTIONS(6603),
    [anon_sym_inline] = ACTIONS(6603),
    [anon_sym___inline] = ACTIONS(6603),
    [anon_sym___inline__] = ACTIONS(6603),
    [anon_sym___forceinline] = ACTIONS(6603),
    [anon_sym_thread_local] = ACTIONS(6603),
    [anon_sym___thread] = ACTIONS(6603),
    [anon_sym_CG_EXTERN] = ACTIONS(6603),
    [anon_sym_CG_INLINE] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6603),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6603),
    [anon_sym_IBOutlet] = ACTIONS(6603),
    [anon_sym_IBInspectable] = ACTIONS(6603),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6603),
    [anon_sym_NS_INLINE] = ACTIONS(6603),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6603),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6603),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6603),
    [anon_sym_const] = ACTIONS(6603),
    [anon_sym_constexpr] = ACTIONS(6603),
    [anon_sym_volatile] = ACTIONS(6603),
    [anon_sym_restrict] = ACTIONS(6603),
    [anon_sym___restrict__] = ACTIONS(6603),
    [anon_sym__Atomic] = ACTIONS(6603),
    [anon_sym__Noreturn] = ACTIONS(6603),
    [anon_sym__Nonnull] = ACTIONS(6603),
    [anon_sym_nullable] = ACTIONS(6603),
    [anon_sym__Complex] = ACTIONS(6603),
    [anon_sym__Nullable] = ACTIONS(6603),
    [anon_sym__Nullable_result] = ACTIONS(6603),
    [anon_sym__Null_unspecified] = ACTIONS(6603),
    [anon_sym___autoreleasing] = ACTIONS(6603),
    [anon_sym___block] = ACTIONS(6603),
    [anon_sym___bridge] = ACTIONS(6603),
    [anon_sym___bridge_retained] = ACTIONS(6603),
    [anon_sym___bridge_transfer] = ACTIONS(6603),
    [anon_sym___complex] = ACTIONS(6603),
    [anon_sym___const] = ACTIONS(6603),
    [anon_sym___imag] = ACTIONS(6603),
    [anon_sym___kindof] = ACTIONS(6603),
    [anon_sym___nonnull] = ACTIONS(6603),
    [anon_sym___nullable] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6603),
    [anon_sym___real] = ACTIONS(6603),
    [anon_sym___strong] = ACTIONS(6603),
    [anon_sym___unsafe_unretained] = ACTIONS(6603),
    [anon_sym___unused] = ACTIONS(6603),
    [anon_sym___weak] = ACTIONS(6603),
    [anon_sym_alignas] = ACTIONS(6603),
    [anon_sym__Alignas] = ACTIONS(6603),
    [sym_primitive_type] = ACTIONS(6603),
    [anon_sym_enum] = ACTIONS(6603),
    [anon_sym_struct] = ACTIONS(6603),
    [anon_sym_union] = ACTIONS(6603),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6603),
    [anon_sym___typeof] = ACTIONS(6603),
    [anon_sym_typeof] = ACTIONS(6603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6603),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6603),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6603),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_API_AVAILABLE] = ACTIONS(6603),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_API_DEPRECATED] = ACTIONS(6603),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym___deprecated_msg] = ACTIONS(6603),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6603),
    [anon_sym_BOOL] = ACTIONS(6603),
    [anon_sym_IMP] = ACTIONS(6603),
    [anon_sym_SEL] = ACTIONS(6603),
    [anon_sym_Class] = ACTIONS(6603),
    [anon_sym_id] = ACTIONS(6603),
    [anon_sym_ATdefs] = ACTIONS(6605),
  },
  [STATE(2694)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_RBRACE] = ACTIONS(6541),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2695)] = {
    [sym_identifier] = ACTIONS(6543),
    [aux_sym_preproc_def_token1] = ACTIONS(6543),
    [aux_sym_preproc_if_token1] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6543),
    [sym_preproc_directive] = ACTIONS(6543),
    [anon_sym___extension__] = ACTIONS(6543),
    [anon_sym_extern] = ACTIONS(6543),
    [anon_sym___attribute__] = ACTIONS(6543),
    [anon_sym___attribute] = ACTIONS(6543),
    [anon_sym_noreturn] = ACTIONS(6543),
    [anon_sym_LBRACK] = ACTIONS(6545),
    [anon_sym___declspec] = ACTIONS(6543),
    [anon_sym_RBRACE] = ACTIONS(6545),
    [anon_sym_signed] = ACTIONS(6543),
    [anon_sym_unsigned] = ACTIONS(6543),
    [anon_sym_long] = ACTIONS(6543),
    [anon_sym_short] = ACTIONS(6543),
    [anon_sym_static] = ACTIONS(6543),
    [anon_sym_auto] = ACTIONS(6543),
    [anon_sym_register] = ACTIONS(6543),
    [anon_sym_inline] = ACTIONS(6543),
    [anon_sym___inline] = ACTIONS(6543),
    [anon_sym___inline__] = ACTIONS(6543),
    [anon_sym___forceinline] = ACTIONS(6543),
    [anon_sym_thread_local] = ACTIONS(6543),
    [anon_sym___thread] = ACTIONS(6543),
    [anon_sym_CG_EXTERN] = ACTIONS(6543),
    [anon_sym_CG_INLINE] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6543),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6543),
    [anon_sym_IBOutlet] = ACTIONS(6543),
    [anon_sym_IBInspectable] = ACTIONS(6543),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6543),
    [anon_sym_NS_INLINE] = ACTIONS(6543),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6543),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6543),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6543),
    [anon_sym_const] = ACTIONS(6543),
    [anon_sym_constexpr] = ACTIONS(6543),
    [anon_sym_volatile] = ACTIONS(6543),
    [anon_sym_restrict] = ACTIONS(6543),
    [anon_sym___restrict__] = ACTIONS(6543),
    [anon_sym__Atomic] = ACTIONS(6543),
    [anon_sym__Noreturn] = ACTIONS(6543),
    [anon_sym__Nonnull] = ACTIONS(6543),
    [anon_sym_nullable] = ACTIONS(6543),
    [anon_sym__Complex] = ACTIONS(6543),
    [anon_sym__Nullable] = ACTIONS(6543),
    [anon_sym__Nullable_result] = ACTIONS(6543),
    [anon_sym__Null_unspecified] = ACTIONS(6543),
    [anon_sym___autoreleasing] = ACTIONS(6543),
    [anon_sym___block] = ACTIONS(6543),
    [anon_sym___bridge] = ACTIONS(6543),
    [anon_sym___bridge_retained] = ACTIONS(6543),
    [anon_sym___bridge_transfer] = ACTIONS(6543),
    [anon_sym___complex] = ACTIONS(6543),
    [anon_sym___const] = ACTIONS(6543),
    [anon_sym___imag] = ACTIONS(6543),
    [anon_sym___kindof] = ACTIONS(6543),
    [anon_sym___nonnull] = ACTIONS(6543),
    [anon_sym___nullable] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6543),
    [anon_sym___real] = ACTIONS(6543),
    [anon_sym___strong] = ACTIONS(6543),
    [anon_sym___unsafe_unretained] = ACTIONS(6543),
    [anon_sym___unused] = ACTIONS(6543),
    [anon_sym___weak] = ACTIONS(6543),
    [anon_sym_alignas] = ACTIONS(6543),
    [anon_sym__Alignas] = ACTIONS(6543),
    [sym_primitive_type] = ACTIONS(6543),
    [anon_sym_enum] = ACTIONS(6543),
    [anon_sym_struct] = ACTIONS(6543),
    [anon_sym_union] = ACTIONS(6543),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6543),
    [anon_sym___typeof] = ACTIONS(6543),
    [anon_sym_typeof] = ACTIONS(6543),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6543),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6543),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6543),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_API_AVAILABLE] = ACTIONS(6543),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_API_DEPRECATED] = ACTIONS(6543),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym___deprecated_msg] = ACTIONS(6543),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6543),
    [anon_sym_BOOL] = ACTIONS(6543),
    [anon_sym_IMP] = ACTIONS(6543),
    [anon_sym_SEL] = ACTIONS(6543),
    [anon_sym_Class] = ACTIONS(6543),
    [anon_sym_id] = ACTIONS(6543),
    [anon_sym_ATdefs] = ACTIONS(6545),
  },
  [STATE(2696)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_RBRACE] = ACTIONS(6541),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2697)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_RBRACE] = ACTIONS(6541),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2698)] = {
    [sym_identifier] = ACTIONS(6543),
    [aux_sym_preproc_def_token1] = ACTIONS(6543),
    [aux_sym_preproc_if_token1] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6543),
    [sym_preproc_directive] = ACTIONS(6543),
    [anon_sym___extension__] = ACTIONS(6543),
    [anon_sym_extern] = ACTIONS(6543),
    [anon_sym___attribute__] = ACTIONS(6543),
    [anon_sym___attribute] = ACTIONS(6543),
    [anon_sym_noreturn] = ACTIONS(6543),
    [anon_sym_LBRACK] = ACTIONS(6545),
    [anon_sym___declspec] = ACTIONS(6543),
    [anon_sym_RBRACE] = ACTIONS(6545),
    [anon_sym_signed] = ACTIONS(6543),
    [anon_sym_unsigned] = ACTIONS(6543),
    [anon_sym_long] = ACTIONS(6543),
    [anon_sym_short] = ACTIONS(6543),
    [anon_sym_static] = ACTIONS(6543),
    [anon_sym_auto] = ACTIONS(6543),
    [anon_sym_register] = ACTIONS(6543),
    [anon_sym_inline] = ACTIONS(6543),
    [anon_sym___inline] = ACTIONS(6543),
    [anon_sym___inline__] = ACTIONS(6543),
    [anon_sym___forceinline] = ACTIONS(6543),
    [anon_sym_thread_local] = ACTIONS(6543),
    [anon_sym___thread] = ACTIONS(6543),
    [anon_sym_CG_EXTERN] = ACTIONS(6543),
    [anon_sym_CG_INLINE] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6543),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6543),
    [anon_sym_IBOutlet] = ACTIONS(6543),
    [anon_sym_IBInspectable] = ACTIONS(6543),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6543),
    [anon_sym_NS_INLINE] = ACTIONS(6543),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6543),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6543),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6543),
    [anon_sym_const] = ACTIONS(6543),
    [anon_sym_constexpr] = ACTIONS(6543),
    [anon_sym_volatile] = ACTIONS(6543),
    [anon_sym_restrict] = ACTIONS(6543),
    [anon_sym___restrict__] = ACTIONS(6543),
    [anon_sym__Atomic] = ACTIONS(6543),
    [anon_sym__Noreturn] = ACTIONS(6543),
    [anon_sym__Nonnull] = ACTIONS(6543),
    [anon_sym_nullable] = ACTIONS(6543),
    [anon_sym__Complex] = ACTIONS(6543),
    [anon_sym__Nullable] = ACTIONS(6543),
    [anon_sym__Nullable_result] = ACTIONS(6543),
    [anon_sym__Null_unspecified] = ACTIONS(6543),
    [anon_sym___autoreleasing] = ACTIONS(6543),
    [anon_sym___block] = ACTIONS(6543),
    [anon_sym___bridge] = ACTIONS(6543),
    [anon_sym___bridge_retained] = ACTIONS(6543),
    [anon_sym___bridge_transfer] = ACTIONS(6543),
    [anon_sym___complex] = ACTIONS(6543),
    [anon_sym___const] = ACTIONS(6543),
    [anon_sym___imag] = ACTIONS(6543),
    [anon_sym___kindof] = ACTIONS(6543),
    [anon_sym___nonnull] = ACTIONS(6543),
    [anon_sym___nullable] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6543),
    [anon_sym___real] = ACTIONS(6543),
    [anon_sym___strong] = ACTIONS(6543),
    [anon_sym___unsafe_unretained] = ACTIONS(6543),
    [anon_sym___unused] = ACTIONS(6543),
    [anon_sym___weak] = ACTIONS(6543),
    [anon_sym_alignas] = ACTIONS(6543),
    [anon_sym__Alignas] = ACTIONS(6543),
    [sym_primitive_type] = ACTIONS(6543),
    [anon_sym_enum] = ACTIONS(6543),
    [anon_sym_struct] = ACTIONS(6543),
    [anon_sym_union] = ACTIONS(6543),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6543),
    [anon_sym___typeof] = ACTIONS(6543),
    [anon_sym_typeof] = ACTIONS(6543),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6543),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6543),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6543),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_API_AVAILABLE] = ACTIONS(6543),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_API_DEPRECATED] = ACTIONS(6543),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym___deprecated_msg] = ACTIONS(6543),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6543),
    [anon_sym_BOOL] = ACTIONS(6543),
    [anon_sym_IMP] = ACTIONS(6543),
    [anon_sym_SEL] = ACTIONS(6543),
    [anon_sym_Class] = ACTIONS(6543),
    [anon_sym_id] = ACTIONS(6543),
    [anon_sym_ATdefs] = ACTIONS(6545),
  },
  [STATE(2699)] = {
    [sym_identifier] = ACTIONS(6599),
    [aux_sym_preproc_def_token1] = ACTIONS(6599),
    [aux_sym_preproc_if_token1] = ACTIONS(6599),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6599),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6599),
    [sym_preproc_directive] = ACTIONS(6599),
    [anon_sym___extension__] = ACTIONS(6599),
    [anon_sym_extern] = ACTIONS(6599),
    [anon_sym___attribute__] = ACTIONS(6599),
    [anon_sym___attribute] = ACTIONS(6599),
    [anon_sym_noreturn] = ACTIONS(6599),
    [anon_sym_LBRACK] = ACTIONS(6601),
    [anon_sym___declspec] = ACTIONS(6599),
    [anon_sym_RBRACE] = ACTIONS(6601),
    [anon_sym_signed] = ACTIONS(6599),
    [anon_sym_unsigned] = ACTIONS(6599),
    [anon_sym_long] = ACTIONS(6599),
    [anon_sym_short] = ACTIONS(6599),
    [anon_sym_static] = ACTIONS(6599),
    [anon_sym_auto] = ACTIONS(6599),
    [anon_sym_register] = ACTIONS(6599),
    [anon_sym_inline] = ACTIONS(6599),
    [anon_sym___inline] = ACTIONS(6599),
    [anon_sym___inline__] = ACTIONS(6599),
    [anon_sym___forceinline] = ACTIONS(6599),
    [anon_sym_thread_local] = ACTIONS(6599),
    [anon_sym___thread] = ACTIONS(6599),
    [anon_sym_CG_EXTERN] = ACTIONS(6599),
    [anon_sym_CG_INLINE] = ACTIONS(6599),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6599),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6599),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6599),
    [anon_sym_IBOutlet] = ACTIONS(6599),
    [anon_sym_IBInspectable] = ACTIONS(6599),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6599),
    [anon_sym_NS_INLINE] = ACTIONS(6599),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6599),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6599),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6599),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6599),
    [anon_sym_const] = ACTIONS(6599),
    [anon_sym_constexpr] = ACTIONS(6599),
    [anon_sym_volatile] = ACTIONS(6599),
    [anon_sym_restrict] = ACTIONS(6599),
    [anon_sym___restrict__] = ACTIONS(6599),
    [anon_sym__Atomic] = ACTIONS(6599),
    [anon_sym__Noreturn] = ACTIONS(6599),
    [anon_sym__Nonnull] = ACTIONS(6599),
    [anon_sym_nullable] = ACTIONS(6599),
    [anon_sym__Complex] = ACTIONS(6599),
    [anon_sym__Nullable] = ACTIONS(6599),
    [anon_sym__Nullable_result] = ACTIONS(6599),
    [anon_sym__Null_unspecified] = ACTIONS(6599),
    [anon_sym___autoreleasing] = ACTIONS(6599),
    [anon_sym___block] = ACTIONS(6599),
    [anon_sym___bridge] = ACTIONS(6599),
    [anon_sym___bridge_retained] = ACTIONS(6599),
    [anon_sym___bridge_transfer] = ACTIONS(6599),
    [anon_sym___complex] = ACTIONS(6599),
    [anon_sym___const] = ACTIONS(6599),
    [anon_sym___imag] = ACTIONS(6599),
    [anon_sym___kindof] = ACTIONS(6599),
    [anon_sym___nonnull] = ACTIONS(6599),
    [anon_sym___nullable] = ACTIONS(6599),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6599),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6599),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6599),
    [anon_sym___real] = ACTIONS(6599),
    [anon_sym___strong] = ACTIONS(6599),
    [anon_sym___unsafe_unretained] = ACTIONS(6599),
    [anon_sym___unused] = ACTIONS(6599),
    [anon_sym___weak] = ACTIONS(6599),
    [anon_sym_alignas] = ACTIONS(6599),
    [anon_sym__Alignas] = ACTIONS(6599),
    [sym_primitive_type] = ACTIONS(6599),
    [anon_sym_enum] = ACTIONS(6599),
    [anon_sym_struct] = ACTIONS(6599),
    [anon_sym_union] = ACTIONS(6599),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6599),
    [anon_sym___typeof] = ACTIONS(6599),
    [anon_sym_typeof] = ACTIONS(6599),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6599),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6599),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6599),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6599),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6599),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6599),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6599),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6599),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6599),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_API_AVAILABLE] = ACTIONS(6599),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_API_DEPRECATED] = ACTIONS(6599),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6599),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6599),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6599),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6599),
    [anon_sym___deprecated_msg] = ACTIONS(6599),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6599),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6599),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6599),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6599),
    [anon_sym_BOOL] = ACTIONS(6599),
    [anon_sym_IMP] = ACTIONS(6599),
    [anon_sym_SEL] = ACTIONS(6599),
    [anon_sym_Class] = ACTIONS(6599),
    [anon_sym_id] = ACTIONS(6599),
    [anon_sym_ATdefs] = ACTIONS(6601),
  },
  [STATE(2700)] = {
    [sym_identifier] = ACTIONS(6519),
    [aux_sym_preproc_def_token1] = ACTIONS(6519),
    [aux_sym_preproc_if_token1] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6519),
    [sym_preproc_directive] = ACTIONS(6519),
    [anon_sym___extension__] = ACTIONS(6519),
    [anon_sym_extern] = ACTIONS(6519),
    [anon_sym___attribute__] = ACTIONS(6519),
    [anon_sym___attribute] = ACTIONS(6519),
    [anon_sym_noreturn] = ACTIONS(6519),
    [anon_sym_LBRACK] = ACTIONS(6521),
    [anon_sym___declspec] = ACTIONS(6519),
    [anon_sym_RBRACE] = ACTIONS(6521),
    [anon_sym_signed] = ACTIONS(6519),
    [anon_sym_unsigned] = ACTIONS(6519),
    [anon_sym_long] = ACTIONS(6519),
    [anon_sym_short] = ACTIONS(6519),
    [anon_sym_static] = ACTIONS(6519),
    [anon_sym_auto] = ACTIONS(6519),
    [anon_sym_register] = ACTIONS(6519),
    [anon_sym_inline] = ACTIONS(6519),
    [anon_sym___inline] = ACTIONS(6519),
    [anon_sym___inline__] = ACTIONS(6519),
    [anon_sym___forceinline] = ACTIONS(6519),
    [anon_sym_thread_local] = ACTIONS(6519),
    [anon_sym___thread] = ACTIONS(6519),
    [anon_sym_CG_EXTERN] = ACTIONS(6519),
    [anon_sym_CG_INLINE] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6519),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6519),
    [anon_sym_IBOutlet] = ACTIONS(6519),
    [anon_sym_IBInspectable] = ACTIONS(6519),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6519),
    [anon_sym_NS_INLINE] = ACTIONS(6519),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6519),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6519),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6519),
    [anon_sym_const] = ACTIONS(6519),
    [anon_sym_constexpr] = ACTIONS(6519),
    [anon_sym_volatile] = ACTIONS(6519),
    [anon_sym_restrict] = ACTIONS(6519),
    [anon_sym___restrict__] = ACTIONS(6519),
    [anon_sym__Atomic] = ACTIONS(6519),
    [anon_sym__Noreturn] = ACTIONS(6519),
    [anon_sym__Nonnull] = ACTIONS(6519),
    [anon_sym_nullable] = ACTIONS(6519),
    [anon_sym__Complex] = ACTIONS(6519),
    [anon_sym__Nullable] = ACTIONS(6519),
    [anon_sym__Nullable_result] = ACTIONS(6519),
    [anon_sym__Null_unspecified] = ACTIONS(6519),
    [anon_sym___autoreleasing] = ACTIONS(6519),
    [anon_sym___block] = ACTIONS(6519),
    [anon_sym___bridge] = ACTIONS(6519),
    [anon_sym___bridge_retained] = ACTIONS(6519),
    [anon_sym___bridge_transfer] = ACTIONS(6519),
    [anon_sym___complex] = ACTIONS(6519),
    [anon_sym___const] = ACTIONS(6519),
    [anon_sym___imag] = ACTIONS(6519),
    [anon_sym___kindof] = ACTIONS(6519),
    [anon_sym___nonnull] = ACTIONS(6519),
    [anon_sym___nullable] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6519),
    [anon_sym___real] = ACTIONS(6519),
    [anon_sym___strong] = ACTIONS(6519),
    [anon_sym___unsafe_unretained] = ACTIONS(6519),
    [anon_sym___unused] = ACTIONS(6519),
    [anon_sym___weak] = ACTIONS(6519),
    [anon_sym_alignas] = ACTIONS(6519),
    [anon_sym__Alignas] = ACTIONS(6519),
    [sym_primitive_type] = ACTIONS(6519),
    [anon_sym_enum] = ACTIONS(6519),
    [anon_sym_struct] = ACTIONS(6519),
    [anon_sym_union] = ACTIONS(6519),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6519),
    [anon_sym___typeof] = ACTIONS(6519),
    [anon_sym_typeof] = ACTIONS(6519),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6519),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6519),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6519),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_API_AVAILABLE] = ACTIONS(6519),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_API_DEPRECATED] = ACTIONS(6519),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym___deprecated_msg] = ACTIONS(6519),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6519),
    [anon_sym_BOOL] = ACTIONS(6519),
    [anon_sym_IMP] = ACTIONS(6519),
    [anon_sym_SEL] = ACTIONS(6519),
    [anon_sym_Class] = ACTIONS(6519),
    [anon_sym_id] = ACTIONS(6519),
    [anon_sym_ATdefs] = ACTIONS(6521),
  },
  [STATE(2701)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_RBRACE] = ACTIONS(6525),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2702)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_RBRACE] = ACTIONS(6525),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2703)] = {
    [sym_identifier] = ACTIONS(6527),
    [aux_sym_preproc_def_token1] = ACTIONS(6527),
    [aux_sym_preproc_if_token1] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6527),
    [sym_preproc_directive] = ACTIONS(6527),
    [anon_sym___extension__] = ACTIONS(6527),
    [anon_sym_extern] = ACTIONS(6527),
    [anon_sym___attribute__] = ACTIONS(6527),
    [anon_sym___attribute] = ACTIONS(6527),
    [anon_sym_noreturn] = ACTIONS(6527),
    [anon_sym_LBRACK] = ACTIONS(6529),
    [anon_sym___declspec] = ACTIONS(6527),
    [anon_sym_RBRACE] = ACTIONS(6529),
    [anon_sym_signed] = ACTIONS(6527),
    [anon_sym_unsigned] = ACTIONS(6527),
    [anon_sym_long] = ACTIONS(6527),
    [anon_sym_short] = ACTIONS(6527),
    [anon_sym_static] = ACTIONS(6527),
    [anon_sym_auto] = ACTIONS(6527),
    [anon_sym_register] = ACTIONS(6527),
    [anon_sym_inline] = ACTIONS(6527),
    [anon_sym___inline] = ACTIONS(6527),
    [anon_sym___inline__] = ACTIONS(6527),
    [anon_sym___forceinline] = ACTIONS(6527),
    [anon_sym_thread_local] = ACTIONS(6527),
    [anon_sym___thread] = ACTIONS(6527),
    [anon_sym_CG_EXTERN] = ACTIONS(6527),
    [anon_sym_CG_INLINE] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6527),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6527),
    [anon_sym_IBOutlet] = ACTIONS(6527),
    [anon_sym_IBInspectable] = ACTIONS(6527),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6527),
    [anon_sym_NS_INLINE] = ACTIONS(6527),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6527),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6527),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6527),
    [anon_sym_const] = ACTIONS(6527),
    [anon_sym_constexpr] = ACTIONS(6527),
    [anon_sym_volatile] = ACTIONS(6527),
    [anon_sym_restrict] = ACTIONS(6527),
    [anon_sym___restrict__] = ACTIONS(6527),
    [anon_sym__Atomic] = ACTIONS(6527),
    [anon_sym__Noreturn] = ACTIONS(6527),
    [anon_sym__Nonnull] = ACTIONS(6527),
    [anon_sym_nullable] = ACTIONS(6527),
    [anon_sym__Complex] = ACTIONS(6527),
    [anon_sym__Nullable] = ACTIONS(6527),
    [anon_sym__Nullable_result] = ACTIONS(6527),
    [anon_sym__Null_unspecified] = ACTIONS(6527),
    [anon_sym___autoreleasing] = ACTIONS(6527),
    [anon_sym___block] = ACTIONS(6527),
    [anon_sym___bridge] = ACTIONS(6527),
    [anon_sym___bridge_retained] = ACTIONS(6527),
    [anon_sym___bridge_transfer] = ACTIONS(6527),
    [anon_sym___complex] = ACTIONS(6527),
    [anon_sym___const] = ACTIONS(6527),
    [anon_sym___imag] = ACTIONS(6527),
    [anon_sym___kindof] = ACTIONS(6527),
    [anon_sym___nonnull] = ACTIONS(6527),
    [anon_sym___nullable] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6527),
    [anon_sym___real] = ACTIONS(6527),
    [anon_sym___strong] = ACTIONS(6527),
    [anon_sym___unsafe_unretained] = ACTIONS(6527),
    [anon_sym___unused] = ACTIONS(6527),
    [anon_sym___weak] = ACTIONS(6527),
    [anon_sym_alignas] = ACTIONS(6527),
    [anon_sym__Alignas] = ACTIONS(6527),
    [sym_primitive_type] = ACTIONS(6527),
    [anon_sym_enum] = ACTIONS(6527),
    [anon_sym_struct] = ACTIONS(6527),
    [anon_sym_union] = ACTIONS(6527),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6527),
    [anon_sym___typeof] = ACTIONS(6527),
    [anon_sym_typeof] = ACTIONS(6527),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6527),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6527),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6527),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_API_AVAILABLE] = ACTIONS(6527),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_API_DEPRECATED] = ACTIONS(6527),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym___deprecated_msg] = ACTIONS(6527),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6527),
    [anon_sym_BOOL] = ACTIONS(6527),
    [anon_sym_IMP] = ACTIONS(6527),
    [anon_sym_SEL] = ACTIONS(6527),
    [anon_sym_Class] = ACTIONS(6527),
    [anon_sym_id] = ACTIONS(6527),
    [anon_sym_ATdefs] = ACTIONS(6529),
  },
  [STATE(2704)] = {
    [sym_identifier] = ACTIONS(6519),
    [aux_sym_preproc_def_token1] = ACTIONS(6519),
    [aux_sym_preproc_if_token1] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6519),
    [sym_preproc_directive] = ACTIONS(6519),
    [anon_sym___extension__] = ACTIONS(6519),
    [anon_sym_extern] = ACTIONS(6519),
    [anon_sym___attribute__] = ACTIONS(6519),
    [anon_sym___attribute] = ACTIONS(6519),
    [anon_sym_noreturn] = ACTIONS(6519),
    [anon_sym_LBRACK] = ACTIONS(6521),
    [anon_sym___declspec] = ACTIONS(6519),
    [anon_sym_RBRACE] = ACTIONS(6521),
    [anon_sym_signed] = ACTIONS(6519),
    [anon_sym_unsigned] = ACTIONS(6519),
    [anon_sym_long] = ACTIONS(6519),
    [anon_sym_short] = ACTIONS(6519),
    [anon_sym_static] = ACTIONS(6519),
    [anon_sym_auto] = ACTIONS(6519),
    [anon_sym_register] = ACTIONS(6519),
    [anon_sym_inline] = ACTIONS(6519),
    [anon_sym___inline] = ACTIONS(6519),
    [anon_sym___inline__] = ACTIONS(6519),
    [anon_sym___forceinline] = ACTIONS(6519),
    [anon_sym_thread_local] = ACTIONS(6519),
    [anon_sym___thread] = ACTIONS(6519),
    [anon_sym_CG_EXTERN] = ACTIONS(6519),
    [anon_sym_CG_INLINE] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6519),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6519),
    [anon_sym_IBOutlet] = ACTIONS(6519),
    [anon_sym_IBInspectable] = ACTIONS(6519),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6519),
    [anon_sym_NS_INLINE] = ACTIONS(6519),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6519),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6519),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6519),
    [anon_sym_const] = ACTIONS(6519),
    [anon_sym_constexpr] = ACTIONS(6519),
    [anon_sym_volatile] = ACTIONS(6519),
    [anon_sym_restrict] = ACTIONS(6519),
    [anon_sym___restrict__] = ACTIONS(6519),
    [anon_sym__Atomic] = ACTIONS(6519),
    [anon_sym__Noreturn] = ACTIONS(6519),
    [anon_sym__Nonnull] = ACTIONS(6519),
    [anon_sym_nullable] = ACTIONS(6519),
    [anon_sym__Complex] = ACTIONS(6519),
    [anon_sym__Nullable] = ACTIONS(6519),
    [anon_sym__Nullable_result] = ACTIONS(6519),
    [anon_sym__Null_unspecified] = ACTIONS(6519),
    [anon_sym___autoreleasing] = ACTIONS(6519),
    [anon_sym___block] = ACTIONS(6519),
    [anon_sym___bridge] = ACTIONS(6519),
    [anon_sym___bridge_retained] = ACTIONS(6519),
    [anon_sym___bridge_transfer] = ACTIONS(6519),
    [anon_sym___complex] = ACTIONS(6519),
    [anon_sym___const] = ACTIONS(6519),
    [anon_sym___imag] = ACTIONS(6519),
    [anon_sym___kindof] = ACTIONS(6519),
    [anon_sym___nonnull] = ACTIONS(6519),
    [anon_sym___nullable] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6519),
    [anon_sym___real] = ACTIONS(6519),
    [anon_sym___strong] = ACTIONS(6519),
    [anon_sym___unsafe_unretained] = ACTIONS(6519),
    [anon_sym___unused] = ACTIONS(6519),
    [anon_sym___weak] = ACTIONS(6519),
    [anon_sym_alignas] = ACTIONS(6519),
    [anon_sym__Alignas] = ACTIONS(6519),
    [sym_primitive_type] = ACTIONS(6519),
    [anon_sym_enum] = ACTIONS(6519),
    [anon_sym_struct] = ACTIONS(6519),
    [anon_sym_union] = ACTIONS(6519),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6519),
    [anon_sym___typeof] = ACTIONS(6519),
    [anon_sym_typeof] = ACTIONS(6519),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6519),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6519),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6519),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_API_AVAILABLE] = ACTIONS(6519),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_API_DEPRECATED] = ACTIONS(6519),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym___deprecated_msg] = ACTIONS(6519),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6519),
    [anon_sym_BOOL] = ACTIONS(6519),
    [anon_sym_IMP] = ACTIONS(6519),
    [anon_sym_SEL] = ACTIONS(6519),
    [anon_sym_Class] = ACTIONS(6519),
    [anon_sym_id] = ACTIONS(6519),
    [anon_sym_ATdefs] = ACTIONS(6521),
  },
  [STATE(2705)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_RBRACE] = ACTIONS(6525),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2706)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_RBRACE] = ACTIONS(6525),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2707)] = {
    [sym_identifier] = ACTIONS(6527),
    [aux_sym_preproc_def_token1] = ACTIONS(6527),
    [aux_sym_preproc_if_token1] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6527),
    [sym_preproc_directive] = ACTIONS(6527),
    [anon_sym___extension__] = ACTIONS(6527),
    [anon_sym_extern] = ACTIONS(6527),
    [anon_sym___attribute__] = ACTIONS(6527),
    [anon_sym___attribute] = ACTIONS(6527),
    [anon_sym_noreturn] = ACTIONS(6527),
    [anon_sym_LBRACK] = ACTIONS(6529),
    [anon_sym___declspec] = ACTIONS(6527),
    [anon_sym_RBRACE] = ACTIONS(6529),
    [anon_sym_signed] = ACTIONS(6527),
    [anon_sym_unsigned] = ACTIONS(6527),
    [anon_sym_long] = ACTIONS(6527),
    [anon_sym_short] = ACTIONS(6527),
    [anon_sym_static] = ACTIONS(6527),
    [anon_sym_auto] = ACTIONS(6527),
    [anon_sym_register] = ACTIONS(6527),
    [anon_sym_inline] = ACTIONS(6527),
    [anon_sym___inline] = ACTIONS(6527),
    [anon_sym___inline__] = ACTIONS(6527),
    [anon_sym___forceinline] = ACTIONS(6527),
    [anon_sym_thread_local] = ACTIONS(6527),
    [anon_sym___thread] = ACTIONS(6527),
    [anon_sym_CG_EXTERN] = ACTIONS(6527),
    [anon_sym_CG_INLINE] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6527),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6527),
    [anon_sym_IBOutlet] = ACTIONS(6527),
    [anon_sym_IBInspectable] = ACTIONS(6527),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6527),
    [anon_sym_NS_INLINE] = ACTIONS(6527),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6527),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6527),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6527),
    [anon_sym_const] = ACTIONS(6527),
    [anon_sym_constexpr] = ACTIONS(6527),
    [anon_sym_volatile] = ACTIONS(6527),
    [anon_sym_restrict] = ACTIONS(6527),
    [anon_sym___restrict__] = ACTIONS(6527),
    [anon_sym__Atomic] = ACTIONS(6527),
    [anon_sym__Noreturn] = ACTIONS(6527),
    [anon_sym__Nonnull] = ACTIONS(6527),
    [anon_sym_nullable] = ACTIONS(6527),
    [anon_sym__Complex] = ACTIONS(6527),
    [anon_sym__Nullable] = ACTIONS(6527),
    [anon_sym__Nullable_result] = ACTIONS(6527),
    [anon_sym__Null_unspecified] = ACTIONS(6527),
    [anon_sym___autoreleasing] = ACTIONS(6527),
    [anon_sym___block] = ACTIONS(6527),
    [anon_sym___bridge] = ACTIONS(6527),
    [anon_sym___bridge_retained] = ACTIONS(6527),
    [anon_sym___bridge_transfer] = ACTIONS(6527),
    [anon_sym___complex] = ACTIONS(6527),
    [anon_sym___const] = ACTIONS(6527),
    [anon_sym___imag] = ACTIONS(6527),
    [anon_sym___kindof] = ACTIONS(6527),
    [anon_sym___nonnull] = ACTIONS(6527),
    [anon_sym___nullable] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6527),
    [anon_sym___real] = ACTIONS(6527),
    [anon_sym___strong] = ACTIONS(6527),
    [anon_sym___unsafe_unretained] = ACTIONS(6527),
    [anon_sym___unused] = ACTIONS(6527),
    [anon_sym___weak] = ACTIONS(6527),
    [anon_sym_alignas] = ACTIONS(6527),
    [anon_sym__Alignas] = ACTIONS(6527),
    [sym_primitive_type] = ACTIONS(6527),
    [anon_sym_enum] = ACTIONS(6527),
    [anon_sym_struct] = ACTIONS(6527),
    [anon_sym_union] = ACTIONS(6527),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6527),
    [anon_sym___typeof] = ACTIONS(6527),
    [anon_sym_typeof] = ACTIONS(6527),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6527),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6527),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6527),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_API_AVAILABLE] = ACTIONS(6527),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_API_DEPRECATED] = ACTIONS(6527),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym___deprecated_msg] = ACTIONS(6527),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6527),
    [anon_sym_BOOL] = ACTIONS(6527),
    [anon_sym_IMP] = ACTIONS(6527),
    [anon_sym_SEL] = ACTIONS(6527),
    [anon_sym_Class] = ACTIONS(6527),
    [anon_sym_id] = ACTIONS(6527),
    [anon_sym_ATdefs] = ACTIONS(6529),
  },
  [STATE(2708)] = {
    [sym_parameter_list] = STATE(2552),
    [sym_identifier] = ACTIONS(6722),
    [anon_sym_COMMA] = ACTIONS(6724),
    [anon_sym_LPAREN2] = ACTIONS(6630),
    [anon_sym_SEMI] = ACTIONS(6724),
    [anon_sym___extension__] = ACTIONS(6722),
    [anon_sym_extern] = ACTIONS(6722),
    [anon_sym___attribute__] = ACTIONS(6722),
    [anon_sym___attribute] = ACTIONS(6722),
    [anon_sym_noreturn] = ACTIONS(6722),
    [anon_sym_LBRACK] = ACTIONS(6632),
    [anon_sym___declspec] = ACTIONS(6722),
    [anon_sym_LBRACE] = ACTIONS(6724),
    [anon_sym_signed] = ACTIONS(6722),
    [anon_sym_unsigned] = ACTIONS(6722),
    [anon_sym_long] = ACTIONS(6722),
    [anon_sym_short] = ACTIONS(6722),
    [anon_sym_static] = ACTIONS(6722),
    [anon_sym_EQ] = ACTIONS(6724),
    [anon_sym_ATautoreleasepool] = ACTIONS(6724),
    [anon_sym_auto] = ACTIONS(6722),
    [anon_sym_register] = ACTIONS(6722),
    [anon_sym_inline] = ACTIONS(6722),
    [anon_sym___inline] = ACTIONS(6722),
    [anon_sym___inline__] = ACTIONS(6722),
    [anon_sym___forceinline] = ACTIONS(6722),
    [anon_sym_thread_local] = ACTIONS(6722),
    [anon_sym___thread] = ACTIONS(6722),
    [anon_sym_CG_EXTERN] = ACTIONS(6722),
    [anon_sym_CG_INLINE] = ACTIONS(6722),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6722),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6722),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6722),
    [anon_sym_IBOutlet] = ACTIONS(6722),
    [anon_sym_IBInspectable] = ACTIONS(6722),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6722),
    [anon_sym_NS_INLINE] = ACTIONS(6722),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6722),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6722),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6722),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6722),
    [anon_sym_const] = ACTIONS(6722),
    [anon_sym_constexpr] = ACTIONS(6722),
    [anon_sym_volatile] = ACTIONS(6722),
    [anon_sym_restrict] = ACTIONS(6722),
    [anon_sym___restrict__] = ACTIONS(6722),
    [anon_sym__Atomic] = ACTIONS(6722),
    [anon_sym__Noreturn] = ACTIONS(6722),
    [anon_sym__Nonnull] = ACTIONS(6722),
    [anon_sym_nullable] = ACTIONS(6722),
    [anon_sym__Complex] = ACTIONS(6722),
    [anon_sym__Nullable] = ACTIONS(6722),
    [anon_sym__Nullable_result] = ACTIONS(6722),
    [anon_sym__Null_unspecified] = ACTIONS(6722),
    [anon_sym___autoreleasing] = ACTIONS(6722),
    [anon_sym___block] = ACTIONS(6722),
    [anon_sym___bridge] = ACTIONS(6722),
    [anon_sym___bridge_retained] = ACTIONS(6722),
    [anon_sym___bridge_transfer] = ACTIONS(6722),
    [anon_sym___complex] = ACTIONS(6722),
    [anon_sym___const] = ACTIONS(6722),
    [anon_sym___imag] = ACTIONS(6722),
    [anon_sym___kindof] = ACTIONS(6722),
    [anon_sym___nonnull] = ACTIONS(6722),
    [anon_sym___nullable] = ACTIONS(6722),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6722),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6722),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6722),
    [anon_sym___real] = ACTIONS(6722),
    [anon_sym___strong] = ACTIONS(6722),
    [anon_sym___unsafe_unretained] = ACTIONS(6722),
    [anon_sym___unused] = ACTIONS(6722),
    [anon_sym___weak] = ACTIONS(6722),
    [anon_sym_alignas] = ACTIONS(6722),
    [anon_sym__Alignas] = ACTIONS(6722),
    [sym_primitive_type] = ACTIONS(6722),
    [anon_sym_enum] = ACTIONS(6722),
    [anon_sym_struct] = ACTIONS(6722),
    [anon_sym_union] = ACTIONS(6722),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6722),
    [anon_sym___typeof] = ACTIONS(6722),
    [anon_sym_typeof] = ACTIONS(6722),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6722),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6722),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6722),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6722),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6722),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6722),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6722),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6722),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6722),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6722),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6722),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6722),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6722),
    [anon_sym_API_AVAILABLE] = ACTIONS(6722),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6722),
    [anon_sym_API_DEPRECATED] = ACTIONS(6722),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6722),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6722),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6722),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6722),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6722),
    [anon_sym___deprecated_msg] = ACTIONS(6722),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6722),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6722),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6722),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6722),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6722),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6722),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6722),
    [anon_sym_BOOL] = ACTIONS(6722),
    [anon_sym_IMP] = ACTIONS(6722),
    [anon_sym_SEL] = ACTIONS(6722),
    [anon_sym_Class] = ACTIONS(6722),
    [anon_sym_id] = ACTIONS(6722),
  },
  [STATE(2709)] = {
    [sym_identifier] = ACTIONS(6555),
    [aux_sym_preproc_def_token1] = ACTIONS(6555),
    [aux_sym_preproc_if_token1] = ACTIONS(6555),
    [aux_sym_preproc_if_token2] = ACTIONS(6555),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6555),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6555),
    [sym_preproc_directive] = ACTIONS(6555),
    [anon_sym___extension__] = ACTIONS(6555),
    [anon_sym_extern] = ACTIONS(6555),
    [anon_sym___attribute__] = ACTIONS(6555),
    [anon_sym___attribute] = ACTIONS(6555),
    [anon_sym_noreturn] = ACTIONS(6555),
    [anon_sym_LBRACK] = ACTIONS(6557),
    [anon_sym___declspec] = ACTIONS(6555),
    [anon_sym_signed] = ACTIONS(6555),
    [anon_sym_unsigned] = ACTIONS(6555),
    [anon_sym_long] = ACTIONS(6555),
    [anon_sym_short] = ACTIONS(6555),
    [anon_sym_static] = ACTIONS(6555),
    [anon_sym_auto] = ACTIONS(6555),
    [anon_sym_register] = ACTIONS(6555),
    [anon_sym_inline] = ACTIONS(6555),
    [anon_sym___inline] = ACTIONS(6555),
    [anon_sym___inline__] = ACTIONS(6555),
    [anon_sym___forceinline] = ACTIONS(6555),
    [anon_sym_thread_local] = ACTIONS(6555),
    [anon_sym___thread] = ACTIONS(6555),
    [anon_sym_CG_EXTERN] = ACTIONS(6555),
    [anon_sym_CG_INLINE] = ACTIONS(6555),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6555),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6555),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6555),
    [anon_sym_IBOutlet] = ACTIONS(6555),
    [anon_sym_IBInspectable] = ACTIONS(6555),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6555),
    [anon_sym_NS_INLINE] = ACTIONS(6555),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6555),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6555),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6555),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6555),
    [anon_sym_const] = ACTIONS(6555),
    [anon_sym_constexpr] = ACTIONS(6555),
    [anon_sym_volatile] = ACTIONS(6555),
    [anon_sym_restrict] = ACTIONS(6555),
    [anon_sym___restrict__] = ACTIONS(6555),
    [anon_sym__Atomic] = ACTIONS(6555),
    [anon_sym__Noreturn] = ACTIONS(6555),
    [anon_sym__Nonnull] = ACTIONS(6555),
    [anon_sym_nullable] = ACTIONS(6555),
    [anon_sym__Complex] = ACTIONS(6555),
    [anon_sym__Nullable] = ACTIONS(6555),
    [anon_sym__Nullable_result] = ACTIONS(6555),
    [anon_sym__Null_unspecified] = ACTIONS(6555),
    [anon_sym___autoreleasing] = ACTIONS(6555),
    [anon_sym___block] = ACTIONS(6555),
    [anon_sym___bridge] = ACTIONS(6555),
    [anon_sym___bridge_retained] = ACTIONS(6555),
    [anon_sym___bridge_transfer] = ACTIONS(6555),
    [anon_sym___complex] = ACTIONS(6555),
    [anon_sym___const] = ACTIONS(6555),
    [anon_sym___imag] = ACTIONS(6555),
    [anon_sym___kindof] = ACTIONS(6555),
    [anon_sym___nonnull] = ACTIONS(6555),
    [anon_sym___nullable] = ACTIONS(6555),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6555),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6555),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6555),
    [anon_sym___real] = ACTIONS(6555),
    [anon_sym___strong] = ACTIONS(6555),
    [anon_sym___unsafe_unretained] = ACTIONS(6555),
    [anon_sym___unused] = ACTIONS(6555),
    [anon_sym___weak] = ACTIONS(6555),
    [anon_sym_alignas] = ACTIONS(6555),
    [anon_sym__Alignas] = ACTIONS(6555),
    [sym_primitive_type] = ACTIONS(6555),
    [anon_sym_enum] = ACTIONS(6555),
    [anon_sym_struct] = ACTIONS(6555),
    [anon_sym_union] = ACTIONS(6555),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6555),
    [anon_sym___typeof] = ACTIONS(6555),
    [anon_sym_typeof] = ACTIONS(6555),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6555),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6555),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6555),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6555),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6555),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6555),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6555),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6555),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6555),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_API_AVAILABLE] = ACTIONS(6555),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_API_DEPRECATED] = ACTIONS(6555),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6555),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6555),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6555),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6555),
    [anon_sym___deprecated_msg] = ACTIONS(6555),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6555),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6555),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6555),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6555),
    [anon_sym_BOOL] = ACTIONS(6555),
    [anon_sym_IMP] = ACTIONS(6555),
    [anon_sym_SEL] = ACTIONS(6555),
    [anon_sym_Class] = ACTIONS(6555),
    [anon_sym_id] = ACTIONS(6555),
    [anon_sym_ATdefs] = ACTIONS(6557),
  },
  [STATE(2710)] = {
    [sym_identifier] = ACTIONS(6587),
    [aux_sym_preproc_def_token1] = ACTIONS(6587),
    [aux_sym_preproc_if_token1] = ACTIONS(6587),
    [aux_sym_preproc_if_token2] = ACTIONS(6587),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6587),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6587),
    [sym_preproc_directive] = ACTIONS(6587),
    [anon_sym___extension__] = ACTIONS(6587),
    [anon_sym_extern] = ACTIONS(6587),
    [anon_sym___attribute__] = ACTIONS(6587),
    [anon_sym___attribute] = ACTIONS(6587),
    [anon_sym_noreturn] = ACTIONS(6587),
    [anon_sym_LBRACK] = ACTIONS(6589),
    [anon_sym___declspec] = ACTIONS(6587),
    [anon_sym_signed] = ACTIONS(6587),
    [anon_sym_unsigned] = ACTIONS(6587),
    [anon_sym_long] = ACTIONS(6587),
    [anon_sym_short] = ACTIONS(6587),
    [anon_sym_static] = ACTIONS(6587),
    [anon_sym_auto] = ACTIONS(6587),
    [anon_sym_register] = ACTIONS(6587),
    [anon_sym_inline] = ACTIONS(6587),
    [anon_sym___inline] = ACTIONS(6587),
    [anon_sym___inline__] = ACTIONS(6587),
    [anon_sym___forceinline] = ACTIONS(6587),
    [anon_sym_thread_local] = ACTIONS(6587),
    [anon_sym___thread] = ACTIONS(6587),
    [anon_sym_CG_EXTERN] = ACTIONS(6587),
    [anon_sym_CG_INLINE] = ACTIONS(6587),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6587),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6587),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6587),
    [anon_sym_IBOutlet] = ACTIONS(6587),
    [anon_sym_IBInspectable] = ACTIONS(6587),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6587),
    [anon_sym_NS_INLINE] = ACTIONS(6587),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6587),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6587),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6587),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6587),
    [anon_sym_const] = ACTIONS(6587),
    [anon_sym_constexpr] = ACTIONS(6587),
    [anon_sym_volatile] = ACTIONS(6587),
    [anon_sym_restrict] = ACTIONS(6587),
    [anon_sym___restrict__] = ACTIONS(6587),
    [anon_sym__Atomic] = ACTIONS(6587),
    [anon_sym__Noreturn] = ACTIONS(6587),
    [anon_sym__Nonnull] = ACTIONS(6587),
    [anon_sym_nullable] = ACTIONS(6587),
    [anon_sym__Complex] = ACTIONS(6587),
    [anon_sym__Nullable] = ACTIONS(6587),
    [anon_sym__Nullable_result] = ACTIONS(6587),
    [anon_sym__Null_unspecified] = ACTIONS(6587),
    [anon_sym___autoreleasing] = ACTIONS(6587),
    [anon_sym___block] = ACTIONS(6587),
    [anon_sym___bridge] = ACTIONS(6587),
    [anon_sym___bridge_retained] = ACTIONS(6587),
    [anon_sym___bridge_transfer] = ACTIONS(6587),
    [anon_sym___complex] = ACTIONS(6587),
    [anon_sym___const] = ACTIONS(6587),
    [anon_sym___imag] = ACTIONS(6587),
    [anon_sym___kindof] = ACTIONS(6587),
    [anon_sym___nonnull] = ACTIONS(6587),
    [anon_sym___nullable] = ACTIONS(6587),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6587),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6587),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6587),
    [anon_sym___real] = ACTIONS(6587),
    [anon_sym___strong] = ACTIONS(6587),
    [anon_sym___unsafe_unretained] = ACTIONS(6587),
    [anon_sym___unused] = ACTIONS(6587),
    [anon_sym___weak] = ACTIONS(6587),
    [anon_sym_alignas] = ACTIONS(6587),
    [anon_sym__Alignas] = ACTIONS(6587),
    [sym_primitive_type] = ACTIONS(6587),
    [anon_sym_enum] = ACTIONS(6587),
    [anon_sym_struct] = ACTIONS(6587),
    [anon_sym_union] = ACTIONS(6587),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6587),
    [anon_sym___typeof] = ACTIONS(6587),
    [anon_sym_typeof] = ACTIONS(6587),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6587),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6587),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6587),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6587),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6587),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6587),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6587),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6587),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6587),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_API_AVAILABLE] = ACTIONS(6587),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_API_DEPRECATED] = ACTIONS(6587),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6587),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6587),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6587),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6587),
    [anon_sym___deprecated_msg] = ACTIONS(6587),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6587),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6587),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6587),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6587),
    [anon_sym_BOOL] = ACTIONS(6587),
    [anon_sym_IMP] = ACTIONS(6587),
    [anon_sym_SEL] = ACTIONS(6587),
    [anon_sym_Class] = ACTIONS(6587),
    [anon_sym_id] = ACTIONS(6587),
    [anon_sym_ATdefs] = ACTIONS(6589),
  },
  [STATE(2711)] = {
    [sym_identifier] = ACTIONS(6591),
    [aux_sym_preproc_def_token1] = ACTIONS(6591),
    [aux_sym_preproc_if_token1] = ACTIONS(6591),
    [aux_sym_preproc_if_token2] = ACTIONS(6591),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6591),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6591),
    [sym_preproc_directive] = ACTIONS(6591),
    [anon_sym___extension__] = ACTIONS(6591),
    [anon_sym_extern] = ACTIONS(6591),
    [anon_sym___attribute__] = ACTIONS(6591),
    [anon_sym___attribute] = ACTIONS(6591),
    [anon_sym_noreturn] = ACTIONS(6591),
    [anon_sym_LBRACK] = ACTIONS(6593),
    [anon_sym___declspec] = ACTIONS(6591),
    [anon_sym_signed] = ACTIONS(6591),
    [anon_sym_unsigned] = ACTIONS(6591),
    [anon_sym_long] = ACTIONS(6591),
    [anon_sym_short] = ACTIONS(6591),
    [anon_sym_static] = ACTIONS(6591),
    [anon_sym_auto] = ACTIONS(6591),
    [anon_sym_register] = ACTIONS(6591),
    [anon_sym_inline] = ACTIONS(6591),
    [anon_sym___inline] = ACTIONS(6591),
    [anon_sym___inline__] = ACTIONS(6591),
    [anon_sym___forceinline] = ACTIONS(6591),
    [anon_sym_thread_local] = ACTIONS(6591),
    [anon_sym___thread] = ACTIONS(6591),
    [anon_sym_CG_EXTERN] = ACTIONS(6591),
    [anon_sym_CG_INLINE] = ACTIONS(6591),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6591),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6591),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6591),
    [anon_sym_IBOutlet] = ACTIONS(6591),
    [anon_sym_IBInspectable] = ACTIONS(6591),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6591),
    [anon_sym_NS_INLINE] = ACTIONS(6591),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6591),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6591),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6591),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6591),
    [anon_sym_const] = ACTIONS(6591),
    [anon_sym_constexpr] = ACTIONS(6591),
    [anon_sym_volatile] = ACTIONS(6591),
    [anon_sym_restrict] = ACTIONS(6591),
    [anon_sym___restrict__] = ACTIONS(6591),
    [anon_sym__Atomic] = ACTIONS(6591),
    [anon_sym__Noreturn] = ACTIONS(6591),
    [anon_sym__Nonnull] = ACTIONS(6591),
    [anon_sym_nullable] = ACTIONS(6591),
    [anon_sym__Complex] = ACTIONS(6591),
    [anon_sym__Nullable] = ACTIONS(6591),
    [anon_sym__Nullable_result] = ACTIONS(6591),
    [anon_sym__Null_unspecified] = ACTIONS(6591),
    [anon_sym___autoreleasing] = ACTIONS(6591),
    [anon_sym___block] = ACTIONS(6591),
    [anon_sym___bridge] = ACTIONS(6591),
    [anon_sym___bridge_retained] = ACTIONS(6591),
    [anon_sym___bridge_transfer] = ACTIONS(6591),
    [anon_sym___complex] = ACTIONS(6591),
    [anon_sym___const] = ACTIONS(6591),
    [anon_sym___imag] = ACTIONS(6591),
    [anon_sym___kindof] = ACTIONS(6591),
    [anon_sym___nonnull] = ACTIONS(6591),
    [anon_sym___nullable] = ACTIONS(6591),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6591),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6591),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6591),
    [anon_sym___real] = ACTIONS(6591),
    [anon_sym___strong] = ACTIONS(6591),
    [anon_sym___unsafe_unretained] = ACTIONS(6591),
    [anon_sym___unused] = ACTIONS(6591),
    [anon_sym___weak] = ACTIONS(6591),
    [anon_sym_alignas] = ACTIONS(6591),
    [anon_sym__Alignas] = ACTIONS(6591),
    [sym_primitive_type] = ACTIONS(6591),
    [anon_sym_enum] = ACTIONS(6591),
    [anon_sym_struct] = ACTIONS(6591),
    [anon_sym_union] = ACTIONS(6591),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6591),
    [anon_sym___typeof] = ACTIONS(6591),
    [anon_sym_typeof] = ACTIONS(6591),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6591),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6591),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6591),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6591),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6591),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6591),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6591),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6591),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6591),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_API_AVAILABLE] = ACTIONS(6591),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_API_DEPRECATED] = ACTIONS(6591),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6591),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6591),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6591),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6591),
    [anon_sym___deprecated_msg] = ACTIONS(6591),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6591),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6591),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6591),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6591),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6591),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6591),
    [anon_sym_BOOL] = ACTIONS(6591),
    [anon_sym_IMP] = ACTIONS(6591),
    [anon_sym_SEL] = ACTIONS(6591),
    [anon_sym_Class] = ACTIONS(6591),
    [anon_sym_id] = ACTIONS(6591),
    [anon_sym_ATdefs] = ACTIONS(6593),
  },
  [STATE(2712)] = {
    [sym_identifier] = ACTIONS(6535),
    [aux_sym_preproc_def_token1] = ACTIONS(6535),
    [aux_sym_preproc_if_token1] = ACTIONS(6535),
    [aux_sym_preproc_if_token2] = ACTIONS(6535),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6535),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6535),
    [sym_preproc_directive] = ACTIONS(6535),
    [anon_sym___extension__] = ACTIONS(6535),
    [anon_sym_extern] = ACTIONS(6535),
    [anon_sym___attribute__] = ACTIONS(6535),
    [anon_sym___attribute] = ACTIONS(6535),
    [anon_sym_noreturn] = ACTIONS(6535),
    [anon_sym_LBRACK] = ACTIONS(6537),
    [anon_sym___declspec] = ACTIONS(6535),
    [anon_sym_signed] = ACTIONS(6535),
    [anon_sym_unsigned] = ACTIONS(6535),
    [anon_sym_long] = ACTIONS(6535),
    [anon_sym_short] = ACTIONS(6535),
    [anon_sym_static] = ACTIONS(6535),
    [anon_sym_auto] = ACTIONS(6535),
    [anon_sym_register] = ACTIONS(6535),
    [anon_sym_inline] = ACTIONS(6535),
    [anon_sym___inline] = ACTIONS(6535),
    [anon_sym___inline__] = ACTIONS(6535),
    [anon_sym___forceinline] = ACTIONS(6535),
    [anon_sym_thread_local] = ACTIONS(6535),
    [anon_sym___thread] = ACTIONS(6535),
    [anon_sym_CG_EXTERN] = ACTIONS(6535),
    [anon_sym_CG_INLINE] = ACTIONS(6535),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6535),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6535),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6535),
    [anon_sym_IBOutlet] = ACTIONS(6535),
    [anon_sym_IBInspectable] = ACTIONS(6535),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6535),
    [anon_sym_NS_INLINE] = ACTIONS(6535),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6535),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6535),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6535),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6535),
    [anon_sym_const] = ACTIONS(6535),
    [anon_sym_constexpr] = ACTIONS(6535),
    [anon_sym_volatile] = ACTIONS(6535),
    [anon_sym_restrict] = ACTIONS(6535),
    [anon_sym___restrict__] = ACTIONS(6535),
    [anon_sym__Atomic] = ACTIONS(6535),
    [anon_sym__Noreturn] = ACTIONS(6535),
    [anon_sym__Nonnull] = ACTIONS(6535),
    [anon_sym_nullable] = ACTIONS(6535),
    [anon_sym__Complex] = ACTIONS(6535),
    [anon_sym__Nullable] = ACTIONS(6535),
    [anon_sym__Nullable_result] = ACTIONS(6535),
    [anon_sym__Null_unspecified] = ACTIONS(6535),
    [anon_sym___autoreleasing] = ACTIONS(6535),
    [anon_sym___block] = ACTIONS(6535),
    [anon_sym___bridge] = ACTIONS(6535),
    [anon_sym___bridge_retained] = ACTIONS(6535),
    [anon_sym___bridge_transfer] = ACTIONS(6535),
    [anon_sym___complex] = ACTIONS(6535),
    [anon_sym___const] = ACTIONS(6535),
    [anon_sym___imag] = ACTIONS(6535),
    [anon_sym___kindof] = ACTIONS(6535),
    [anon_sym___nonnull] = ACTIONS(6535),
    [anon_sym___nullable] = ACTIONS(6535),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6535),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6535),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6535),
    [anon_sym___real] = ACTIONS(6535),
    [anon_sym___strong] = ACTIONS(6535),
    [anon_sym___unsafe_unretained] = ACTIONS(6535),
    [anon_sym___unused] = ACTIONS(6535),
    [anon_sym___weak] = ACTIONS(6535),
    [anon_sym_alignas] = ACTIONS(6535),
    [anon_sym__Alignas] = ACTIONS(6535),
    [sym_primitive_type] = ACTIONS(6535),
    [anon_sym_enum] = ACTIONS(6535),
    [anon_sym_struct] = ACTIONS(6535),
    [anon_sym_union] = ACTIONS(6535),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6535),
    [anon_sym___typeof] = ACTIONS(6535),
    [anon_sym_typeof] = ACTIONS(6535),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6535),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6535),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6535),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6535),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6535),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6535),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6535),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6535),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6535),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_API_AVAILABLE] = ACTIONS(6535),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_API_DEPRECATED] = ACTIONS(6535),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6535),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6535),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6535),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6535),
    [anon_sym___deprecated_msg] = ACTIONS(6535),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6535),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6535),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6535),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6535),
    [anon_sym_BOOL] = ACTIONS(6535),
    [anon_sym_IMP] = ACTIONS(6535),
    [anon_sym_SEL] = ACTIONS(6535),
    [anon_sym_Class] = ACTIONS(6535),
    [anon_sym_id] = ACTIONS(6535),
    [anon_sym_ATdefs] = ACTIONS(6537),
  },
  [STATE(2713)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token2] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2714)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token2] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2715)] = {
    [sym_identifier] = ACTIONS(6543),
    [aux_sym_preproc_def_token1] = ACTIONS(6543),
    [aux_sym_preproc_if_token1] = ACTIONS(6543),
    [aux_sym_preproc_if_token2] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6543),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6543),
    [sym_preproc_directive] = ACTIONS(6543),
    [anon_sym___extension__] = ACTIONS(6543),
    [anon_sym_extern] = ACTIONS(6543),
    [anon_sym___attribute__] = ACTIONS(6543),
    [anon_sym___attribute] = ACTIONS(6543),
    [anon_sym_noreturn] = ACTIONS(6543),
    [anon_sym_LBRACK] = ACTIONS(6545),
    [anon_sym___declspec] = ACTIONS(6543),
    [anon_sym_signed] = ACTIONS(6543),
    [anon_sym_unsigned] = ACTIONS(6543),
    [anon_sym_long] = ACTIONS(6543),
    [anon_sym_short] = ACTIONS(6543),
    [anon_sym_static] = ACTIONS(6543),
    [anon_sym_auto] = ACTIONS(6543),
    [anon_sym_register] = ACTIONS(6543),
    [anon_sym_inline] = ACTIONS(6543),
    [anon_sym___inline] = ACTIONS(6543),
    [anon_sym___inline__] = ACTIONS(6543),
    [anon_sym___forceinline] = ACTIONS(6543),
    [anon_sym_thread_local] = ACTIONS(6543),
    [anon_sym___thread] = ACTIONS(6543),
    [anon_sym_CG_EXTERN] = ACTIONS(6543),
    [anon_sym_CG_INLINE] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6543),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6543),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6543),
    [anon_sym_IBOutlet] = ACTIONS(6543),
    [anon_sym_IBInspectable] = ACTIONS(6543),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6543),
    [anon_sym_NS_INLINE] = ACTIONS(6543),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6543),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6543),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6543),
    [anon_sym_const] = ACTIONS(6543),
    [anon_sym_constexpr] = ACTIONS(6543),
    [anon_sym_volatile] = ACTIONS(6543),
    [anon_sym_restrict] = ACTIONS(6543),
    [anon_sym___restrict__] = ACTIONS(6543),
    [anon_sym__Atomic] = ACTIONS(6543),
    [anon_sym__Noreturn] = ACTIONS(6543),
    [anon_sym__Nonnull] = ACTIONS(6543),
    [anon_sym_nullable] = ACTIONS(6543),
    [anon_sym__Complex] = ACTIONS(6543),
    [anon_sym__Nullable] = ACTIONS(6543),
    [anon_sym__Nullable_result] = ACTIONS(6543),
    [anon_sym__Null_unspecified] = ACTIONS(6543),
    [anon_sym___autoreleasing] = ACTIONS(6543),
    [anon_sym___block] = ACTIONS(6543),
    [anon_sym___bridge] = ACTIONS(6543),
    [anon_sym___bridge_retained] = ACTIONS(6543),
    [anon_sym___bridge_transfer] = ACTIONS(6543),
    [anon_sym___complex] = ACTIONS(6543),
    [anon_sym___const] = ACTIONS(6543),
    [anon_sym___imag] = ACTIONS(6543),
    [anon_sym___kindof] = ACTIONS(6543),
    [anon_sym___nonnull] = ACTIONS(6543),
    [anon_sym___nullable] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6543),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6543),
    [anon_sym___real] = ACTIONS(6543),
    [anon_sym___strong] = ACTIONS(6543),
    [anon_sym___unsafe_unretained] = ACTIONS(6543),
    [anon_sym___unused] = ACTIONS(6543),
    [anon_sym___weak] = ACTIONS(6543),
    [anon_sym_alignas] = ACTIONS(6543),
    [anon_sym__Alignas] = ACTIONS(6543),
    [sym_primitive_type] = ACTIONS(6543),
    [anon_sym_enum] = ACTIONS(6543),
    [anon_sym_struct] = ACTIONS(6543),
    [anon_sym_union] = ACTIONS(6543),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6543),
    [anon_sym___typeof] = ACTIONS(6543),
    [anon_sym_typeof] = ACTIONS(6543),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6543),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6543),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6543),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6543),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6543),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_API_AVAILABLE] = ACTIONS(6543),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_API_DEPRECATED] = ACTIONS(6543),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6543),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6543),
    [anon_sym___deprecated_msg] = ACTIONS(6543),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6543),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6543),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6543),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6543),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6543),
    [anon_sym_BOOL] = ACTIONS(6543),
    [anon_sym_IMP] = ACTIONS(6543),
    [anon_sym_SEL] = ACTIONS(6543),
    [anon_sym_Class] = ACTIONS(6543),
    [anon_sym_id] = ACTIONS(6543),
    [anon_sym_ATdefs] = ACTIONS(6545),
  },
  [STATE(2716)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token2] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2717)] = {
    [sym_identifier] = ACTIONS(6539),
    [aux_sym_preproc_def_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token1] = ACTIONS(6539),
    [aux_sym_preproc_if_token2] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6539),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6539),
    [sym_preproc_directive] = ACTIONS(6539),
    [anon_sym___extension__] = ACTIONS(6539),
    [anon_sym_extern] = ACTIONS(6539),
    [anon_sym___attribute__] = ACTIONS(6539),
    [anon_sym___attribute] = ACTIONS(6539),
    [anon_sym_noreturn] = ACTIONS(6539),
    [anon_sym_LBRACK] = ACTIONS(6541),
    [anon_sym___declspec] = ACTIONS(6539),
    [anon_sym_signed] = ACTIONS(6539),
    [anon_sym_unsigned] = ACTIONS(6539),
    [anon_sym_long] = ACTIONS(6539),
    [anon_sym_short] = ACTIONS(6539),
    [anon_sym_static] = ACTIONS(6539),
    [anon_sym_auto] = ACTIONS(6539),
    [anon_sym_register] = ACTIONS(6539),
    [anon_sym_inline] = ACTIONS(6539),
    [anon_sym___inline] = ACTIONS(6539),
    [anon_sym___inline__] = ACTIONS(6539),
    [anon_sym___forceinline] = ACTIONS(6539),
    [anon_sym_thread_local] = ACTIONS(6539),
    [anon_sym___thread] = ACTIONS(6539),
    [anon_sym_CG_EXTERN] = ACTIONS(6539),
    [anon_sym_CG_INLINE] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6539),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6539),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6539),
    [anon_sym_IBOutlet] = ACTIONS(6539),
    [anon_sym_IBInspectable] = ACTIONS(6539),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6539),
    [anon_sym_NS_INLINE] = ACTIONS(6539),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6539),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6539),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6539),
    [anon_sym_const] = ACTIONS(6539),
    [anon_sym_constexpr] = ACTIONS(6539),
    [anon_sym_volatile] = ACTIONS(6539),
    [anon_sym_restrict] = ACTIONS(6539),
    [anon_sym___restrict__] = ACTIONS(6539),
    [anon_sym__Atomic] = ACTIONS(6539),
    [anon_sym__Noreturn] = ACTIONS(6539),
    [anon_sym__Nonnull] = ACTIONS(6539),
    [anon_sym_nullable] = ACTIONS(6539),
    [anon_sym__Complex] = ACTIONS(6539),
    [anon_sym__Nullable] = ACTIONS(6539),
    [anon_sym__Nullable_result] = ACTIONS(6539),
    [anon_sym__Null_unspecified] = ACTIONS(6539),
    [anon_sym___autoreleasing] = ACTIONS(6539),
    [anon_sym___block] = ACTIONS(6539),
    [anon_sym___bridge] = ACTIONS(6539),
    [anon_sym___bridge_retained] = ACTIONS(6539),
    [anon_sym___bridge_transfer] = ACTIONS(6539),
    [anon_sym___complex] = ACTIONS(6539),
    [anon_sym___const] = ACTIONS(6539),
    [anon_sym___imag] = ACTIONS(6539),
    [anon_sym___kindof] = ACTIONS(6539),
    [anon_sym___nonnull] = ACTIONS(6539),
    [anon_sym___nullable] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6539),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6539),
    [anon_sym___real] = ACTIONS(6539),
    [anon_sym___strong] = ACTIONS(6539),
    [anon_sym___unsafe_unretained] = ACTIONS(6539),
    [anon_sym___unused] = ACTIONS(6539),
    [anon_sym___weak] = ACTIONS(6539),
    [anon_sym_alignas] = ACTIONS(6539),
    [anon_sym__Alignas] = ACTIONS(6539),
    [sym_primitive_type] = ACTIONS(6539),
    [anon_sym_enum] = ACTIONS(6539),
    [anon_sym_struct] = ACTIONS(6539),
    [anon_sym_union] = ACTIONS(6539),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6539),
    [anon_sym___typeof] = ACTIONS(6539),
    [anon_sym_typeof] = ACTIONS(6539),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6539),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6539),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6539),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6539),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6539),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_API_AVAILABLE] = ACTIONS(6539),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_API_DEPRECATED] = ACTIONS(6539),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6539),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6539),
    [anon_sym___deprecated_msg] = ACTIONS(6539),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6539),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6539),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6539),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6539),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6539),
    [anon_sym_BOOL] = ACTIONS(6539),
    [anon_sym_IMP] = ACTIONS(6539),
    [anon_sym_SEL] = ACTIONS(6539),
    [anon_sym_Class] = ACTIONS(6539),
    [anon_sym_id] = ACTIONS(6539),
    [anon_sym_ATdefs] = ACTIONS(6541),
  },
  [STATE(2718)] = {
    [sym_identifier] = ACTIONS(6551),
    [aux_sym_preproc_def_token1] = ACTIONS(6551),
    [aux_sym_preproc_if_token1] = ACTIONS(6551),
    [aux_sym_preproc_if_token2] = ACTIONS(6551),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6551),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6551),
    [sym_preproc_directive] = ACTIONS(6551),
    [anon_sym___extension__] = ACTIONS(6551),
    [anon_sym_extern] = ACTIONS(6551),
    [anon_sym___attribute__] = ACTIONS(6551),
    [anon_sym___attribute] = ACTIONS(6551),
    [anon_sym_noreturn] = ACTIONS(6551),
    [anon_sym_LBRACK] = ACTIONS(6553),
    [anon_sym___declspec] = ACTIONS(6551),
    [anon_sym_signed] = ACTIONS(6551),
    [anon_sym_unsigned] = ACTIONS(6551),
    [anon_sym_long] = ACTIONS(6551),
    [anon_sym_short] = ACTIONS(6551),
    [anon_sym_static] = ACTIONS(6551),
    [anon_sym_auto] = ACTIONS(6551),
    [anon_sym_register] = ACTIONS(6551),
    [anon_sym_inline] = ACTIONS(6551),
    [anon_sym___inline] = ACTIONS(6551),
    [anon_sym___inline__] = ACTIONS(6551),
    [anon_sym___forceinline] = ACTIONS(6551),
    [anon_sym_thread_local] = ACTIONS(6551),
    [anon_sym___thread] = ACTIONS(6551),
    [anon_sym_CG_EXTERN] = ACTIONS(6551),
    [anon_sym_CG_INLINE] = ACTIONS(6551),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6551),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6551),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6551),
    [anon_sym_IBOutlet] = ACTIONS(6551),
    [anon_sym_IBInspectable] = ACTIONS(6551),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6551),
    [anon_sym_NS_INLINE] = ACTIONS(6551),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6551),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6551),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6551),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6551),
    [anon_sym_const] = ACTIONS(6551),
    [anon_sym_constexpr] = ACTIONS(6551),
    [anon_sym_volatile] = ACTIONS(6551),
    [anon_sym_restrict] = ACTIONS(6551),
    [anon_sym___restrict__] = ACTIONS(6551),
    [anon_sym__Atomic] = ACTIONS(6551),
    [anon_sym__Noreturn] = ACTIONS(6551),
    [anon_sym__Nonnull] = ACTIONS(6551),
    [anon_sym_nullable] = ACTIONS(6551),
    [anon_sym__Complex] = ACTIONS(6551),
    [anon_sym__Nullable] = ACTIONS(6551),
    [anon_sym__Nullable_result] = ACTIONS(6551),
    [anon_sym__Null_unspecified] = ACTIONS(6551),
    [anon_sym___autoreleasing] = ACTIONS(6551),
    [anon_sym___block] = ACTIONS(6551),
    [anon_sym___bridge] = ACTIONS(6551),
    [anon_sym___bridge_retained] = ACTIONS(6551),
    [anon_sym___bridge_transfer] = ACTIONS(6551),
    [anon_sym___complex] = ACTIONS(6551),
    [anon_sym___const] = ACTIONS(6551),
    [anon_sym___imag] = ACTIONS(6551),
    [anon_sym___kindof] = ACTIONS(6551),
    [anon_sym___nonnull] = ACTIONS(6551),
    [anon_sym___nullable] = ACTIONS(6551),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6551),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6551),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6551),
    [anon_sym___real] = ACTIONS(6551),
    [anon_sym___strong] = ACTIONS(6551),
    [anon_sym___unsafe_unretained] = ACTIONS(6551),
    [anon_sym___unused] = ACTIONS(6551),
    [anon_sym___weak] = ACTIONS(6551),
    [anon_sym_alignas] = ACTIONS(6551),
    [anon_sym__Alignas] = ACTIONS(6551),
    [sym_primitive_type] = ACTIONS(6551),
    [anon_sym_enum] = ACTIONS(6551),
    [anon_sym_struct] = ACTIONS(6551),
    [anon_sym_union] = ACTIONS(6551),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6551),
    [anon_sym___typeof] = ACTIONS(6551),
    [anon_sym_typeof] = ACTIONS(6551),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6551),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6551),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6551),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6551),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6551),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6551),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6551),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6551),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6551),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_API_AVAILABLE] = ACTIONS(6551),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_API_DEPRECATED] = ACTIONS(6551),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6551),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6551),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6551),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6551),
    [anon_sym___deprecated_msg] = ACTIONS(6551),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6551),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6551),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6551),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6551),
    [anon_sym_BOOL] = ACTIONS(6551),
    [anon_sym_IMP] = ACTIONS(6551),
    [anon_sym_SEL] = ACTIONS(6551),
    [anon_sym_Class] = ACTIONS(6551),
    [anon_sym_id] = ACTIONS(6551),
    [anon_sym_ATdefs] = ACTIONS(6553),
  },
  [STATE(2719)] = {
    [sym_identifier] = ACTIONS(6551),
    [aux_sym_preproc_def_token1] = ACTIONS(6551),
    [aux_sym_preproc_if_token1] = ACTIONS(6551),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6551),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6551),
    [sym_preproc_directive] = ACTIONS(6551),
    [anon_sym___extension__] = ACTIONS(6551),
    [anon_sym_extern] = ACTIONS(6551),
    [anon_sym___attribute__] = ACTIONS(6551),
    [anon_sym___attribute] = ACTIONS(6551),
    [anon_sym_noreturn] = ACTIONS(6551),
    [anon_sym_LBRACK] = ACTIONS(6553),
    [anon_sym___declspec] = ACTIONS(6551),
    [anon_sym_RBRACE] = ACTIONS(6553),
    [anon_sym_signed] = ACTIONS(6551),
    [anon_sym_unsigned] = ACTIONS(6551),
    [anon_sym_long] = ACTIONS(6551),
    [anon_sym_short] = ACTIONS(6551),
    [anon_sym_static] = ACTIONS(6551),
    [anon_sym_auto] = ACTIONS(6551),
    [anon_sym_register] = ACTIONS(6551),
    [anon_sym_inline] = ACTIONS(6551),
    [anon_sym___inline] = ACTIONS(6551),
    [anon_sym___inline__] = ACTIONS(6551),
    [anon_sym___forceinline] = ACTIONS(6551),
    [anon_sym_thread_local] = ACTIONS(6551),
    [anon_sym___thread] = ACTIONS(6551),
    [anon_sym_CG_EXTERN] = ACTIONS(6551),
    [anon_sym_CG_INLINE] = ACTIONS(6551),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6551),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6551),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6551),
    [anon_sym_IBOutlet] = ACTIONS(6551),
    [anon_sym_IBInspectable] = ACTIONS(6551),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6551),
    [anon_sym_NS_INLINE] = ACTIONS(6551),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6551),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6551),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6551),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6551),
    [anon_sym_const] = ACTIONS(6551),
    [anon_sym_constexpr] = ACTIONS(6551),
    [anon_sym_volatile] = ACTIONS(6551),
    [anon_sym_restrict] = ACTIONS(6551),
    [anon_sym___restrict__] = ACTIONS(6551),
    [anon_sym__Atomic] = ACTIONS(6551),
    [anon_sym__Noreturn] = ACTIONS(6551),
    [anon_sym__Nonnull] = ACTIONS(6551),
    [anon_sym_nullable] = ACTIONS(6551),
    [anon_sym__Complex] = ACTIONS(6551),
    [anon_sym__Nullable] = ACTIONS(6551),
    [anon_sym__Nullable_result] = ACTIONS(6551),
    [anon_sym__Null_unspecified] = ACTIONS(6551),
    [anon_sym___autoreleasing] = ACTIONS(6551),
    [anon_sym___block] = ACTIONS(6551),
    [anon_sym___bridge] = ACTIONS(6551),
    [anon_sym___bridge_retained] = ACTIONS(6551),
    [anon_sym___bridge_transfer] = ACTIONS(6551),
    [anon_sym___complex] = ACTIONS(6551),
    [anon_sym___const] = ACTIONS(6551),
    [anon_sym___imag] = ACTIONS(6551),
    [anon_sym___kindof] = ACTIONS(6551),
    [anon_sym___nonnull] = ACTIONS(6551),
    [anon_sym___nullable] = ACTIONS(6551),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6551),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6551),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6551),
    [anon_sym___real] = ACTIONS(6551),
    [anon_sym___strong] = ACTIONS(6551),
    [anon_sym___unsafe_unretained] = ACTIONS(6551),
    [anon_sym___unused] = ACTIONS(6551),
    [anon_sym___weak] = ACTIONS(6551),
    [anon_sym_alignas] = ACTIONS(6551),
    [anon_sym__Alignas] = ACTIONS(6551),
    [sym_primitive_type] = ACTIONS(6551),
    [anon_sym_enum] = ACTIONS(6551),
    [anon_sym_struct] = ACTIONS(6551),
    [anon_sym_union] = ACTIONS(6551),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6551),
    [anon_sym___typeof] = ACTIONS(6551),
    [anon_sym_typeof] = ACTIONS(6551),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6551),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6551),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6551),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6551),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6551),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6551),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6551),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6551),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6551),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_API_AVAILABLE] = ACTIONS(6551),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_API_DEPRECATED] = ACTIONS(6551),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6551),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6551),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6551),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6551),
    [anon_sym___deprecated_msg] = ACTIONS(6551),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6551),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6551),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6551),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6551),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6551),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6551),
    [anon_sym_BOOL] = ACTIONS(6551),
    [anon_sym_IMP] = ACTIONS(6551),
    [anon_sym_SEL] = ACTIONS(6551),
    [anon_sym_Class] = ACTIONS(6551),
    [anon_sym_id] = ACTIONS(6551),
    [anon_sym_ATdefs] = ACTIONS(6553),
  },
  [STATE(2720)] = {
    [sym_identifier] = ACTIONS(6559),
    [aux_sym_preproc_def_token1] = ACTIONS(6559),
    [aux_sym_preproc_if_token1] = ACTIONS(6559),
    [aux_sym_preproc_if_token2] = ACTIONS(6559),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6559),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6559),
    [sym_preproc_directive] = ACTIONS(6559),
    [anon_sym___extension__] = ACTIONS(6559),
    [anon_sym_extern] = ACTIONS(6559),
    [anon_sym___attribute__] = ACTIONS(6559),
    [anon_sym___attribute] = ACTIONS(6559),
    [anon_sym_noreturn] = ACTIONS(6559),
    [anon_sym_LBRACK] = ACTIONS(6561),
    [anon_sym___declspec] = ACTIONS(6559),
    [anon_sym_signed] = ACTIONS(6559),
    [anon_sym_unsigned] = ACTIONS(6559),
    [anon_sym_long] = ACTIONS(6559),
    [anon_sym_short] = ACTIONS(6559),
    [anon_sym_static] = ACTIONS(6559),
    [anon_sym_auto] = ACTIONS(6559),
    [anon_sym_register] = ACTIONS(6559),
    [anon_sym_inline] = ACTIONS(6559),
    [anon_sym___inline] = ACTIONS(6559),
    [anon_sym___inline__] = ACTIONS(6559),
    [anon_sym___forceinline] = ACTIONS(6559),
    [anon_sym_thread_local] = ACTIONS(6559),
    [anon_sym___thread] = ACTIONS(6559),
    [anon_sym_CG_EXTERN] = ACTIONS(6559),
    [anon_sym_CG_INLINE] = ACTIONS(6559),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6559),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6559),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6559),
    [anon_sym_IBOutlet] = ACTIONS(6559),
    [anon_sym_IBInspectable] = ACTIONS(6559),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6559),
    [anon_sym_NS_INLINE] = ACTIONS(6559),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6559),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6559),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6559),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6559),
    [anon_sym_const] = ACTIONS(6559),
    [anon_sym_constexpr] = ACTIONS(6559),
    [anon_sym_volatile] = ACTIONS(6559),
    [anon_sym_restrict] = ACTIONS(6559),
    [anon_sym___restrict__] = ACTIONS(6559),
    [anon_sym__Atomic] = ACTIONS(6559),
    [anon_sym__Noreturn] = ACTIONS(6559),
    [anon_sym__Nonnull] = ACTIONS(6559),
    [anon_sym_nullable] = ACTIONS(6559),
    [anon_sym__Complex] = ACTIONS(6559),
    [anon_sym__Nullable] = ACTIONS(6559),
    [anon_sym__Nullable_result] = ACTIONS(6559),
    [anon_sym__Null_unspecified] = ACTIONS(6559),
    [anon_sym___autoreleasing] = ACTIONS(6559),
    [anon_sym___block] = ACTIONS(6559),
    [anon_sym___bridge] = ACTIONS(6559),
    [anon_sym___bridge_retained] = ACTIONS(6559),
    [anon_sym___bridge_transfer] = ACTIONS(6559),
    [anon_sym___complex] = ACTIONS(6559),
    [anon_sym___const] = ACTIONS(6559),
    [anon_sym___imag] = ACTIONS(6559),
    [anon_sym___kindof] = ACTIONS(6559),
    [anon_sym___nonnull] = ACTIONS(6559),
    [anon_sym___nullable] = ACTIONS(6559),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6559),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6559),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6559),
    [anon_sym___real] = ACTIONS(6559),
    [anon_sym___strong] = ACTIONS(6559),
    [anon_sym___unsafe_unretained] = ACTIONS(6559),
    [anon_sym___unused] = ACTIONS(6559),
    [anon_sym___weak] = ACTIONS(6559),
    [anon_sym_alignas] = ACTIONS(6559),
    [anon_sym__Alignas] = ACTIONS(6559),
    [sym_primitive_type] = ACTIONS(6559),
    [anon_sym_enum] = ACTIONS(6559),
    [anon_sym_struct] = ACTIONS(6559),
    [anon_sym_union] = ACTIONS(6559),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6559),
    [anon_sym___typeof] = ACTIONS(6559),
    [anon_sym_typeof] = ACTIONS(6559),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6559),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6559),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6559),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6559),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6559),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6559),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6559),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6559),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6559),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_API_AVAILABLE] = ACTIONS(6559),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_API_DEPRECATED] = ACTIONS(6559),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6559),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6559),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6559),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6559),
    [anon_sym___deprecated_msg] = ACTIONS(6559),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6559),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6559),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6559),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6559),
    [anon_sym_BOOL] = ACTIONS(6559),
    [anon_sym_IMP] = ACTIONS(6559),
    [anon_sym_SEL] = ACTIONS(6559),
    [anon_sym_Class] = ACTIONS(6559),
    [anon_sym_id] = ACTIONS(6559),
    [anon_sym_ATdefs] = ACTIONS(6561),
  },
  [STATE(2721)] = {
    [sym_identifier] = ACTIONS(6563),
    [aux_sym_preproc_def_token1] = ACTIONS(6563),
    [aux_sym_preproc_if_token1] = ACTIONS(6563),
    [aux_sym_preproc_if_token2] = ACTIONS(6563),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6563),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6563),
    [sym_preproc_directive] = ACTIONS(6563),
    [anon_sym___extension__] = ACTIONS(6563),
    [anon_sym_extern] = ACTIONS(6563),
    [anon_sym___attribute__] = ACTIONS(6563),
    [anon_sym___attribute] = ACTIONS(6563),
    [anon_sym_noreturn] = ACTIONS(6563),
    [anon_sym_LBRACK] = ACTIONS(6565),
    [anon_sym___declspec] = ACTIONS(6563),
    [anon_sym_signed] = ACTIONS(6563),
    [anon_sym_unsigned] = ACTIONS(6563),
    [anon_sym_long] = ACTIONS(6563),
    [anon_sym_short] = ACTIONS(6563),
    [anon_sym_static] = ACTIONS(6563),
    [anon_sym_auto] = ACTIONS(6563),
    [anon_sym_register] = ACTIONS(6563),
    [anon_sym_inline] = ACTIONS(6563),
    [anon_sym___inline] = ACTIONS(6563),
    [anon_sym___inline__] = ACTIONS(6563),
    [anon_sym___forceinline] = ACTIONS(6563),
    [anon_sym_thread_local] = ACTIONS(6563),
    [anon_sym___thread] = ACTIONS(6563),
    [anon_sym_CG_EXTERN] = ACTIONS(6563),
    [anon_sym_CG_INLINE] = ACTIONS(6563),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6563),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6563),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6563),
    [anon_sym_IBOutlet] = ACTIONS(6563),
    [anon_sym_IBInspectable] = ACTIONS(6563),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6563),
    [anon_sym_NS_INLINE] = ACTIONS(6563),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6563),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6563),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6563),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6563),
    [anon_sym_const] = ACTIONS(6563),
    [anon_sym_constexpr] = ACTIONS(6563),
    [anon_sym_volatile] = ACTIONS(6563),
    [anon_sym_restrict] = ACTIONS(6563),
    [anon_sym___restrict__] = ACTIONS(6563),
    [anon_sym__Atomic] = ACTIONS(6563),
    [anon_sym__Noreturn] = ACTIONS(6563),
    [anon_sym__Nonnull] = ACTIONS(6563),
    [anon_sym_nullable] = ACTIONS(6563),
    [anon_sym__Complex] = ACTIONS(6563),
    [anon_sym__Nullable] = ACTIONS(6563),
    [anon_sym__Nullable_result] = ACTIONS(6563),
    [anon_sym__Null_unspecified] = ACTIONS(6563),
    [anon_sym___autoreleasing] = ACTIONS(6563),
    [anon_sym___block] = ACTIONS(6563),
    [anon_sym___bridge] = ACTIONS(6563),
    [anon_sym___bridge_retained] = ACTIONS(6563),
    [anon_sym___bridge_transfer] = ACTIONS(6563),
    [anon_sym___complex] = ACTIONS(6563),
    [anon_sym___const] = ACTIONS(6563),
    [anon_sym___imag] = ACTIONS(6563),
    [anon_sym___kindof] = ACTIONS(6563),
    [anon_sym___nonnull] = ACTIONS(6563),
    [anon_sym___nullable] = ACTIONS(6563),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6563),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6563),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6563),
    [anon_sym___real] = ACTIONS(6563),
    [anon_sym___strong] = ACTIONS(6563),
    [anon_sym___unsafe_unretained] = ACTIONS(6563),
    [anon_sym___unused] = ACTIONS(6563),
    [anon_sym___weak] = ACTIONS(6563),
    [anon_sym_alignas] = ACTIONS(6563),
    [anon_sym__Alignas] = ACTIONS(6563),
    [sym_primitive_type] = ACTIONS(6563),
    [anon_sym_enum] = ACTIONS(6563),
    [anon_sym_struct] = ACTIONS(6563),
    [anon_sym_union] = ACTIONS(6563),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6563),
    [anon_sym___typeof] = ACTIONS(6563),
    [anon_sym_typeof] = ACTIONS(6563),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6563),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6563),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6563),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6563),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6563),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6563),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6563),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6563),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6563),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_API_AVAILABLE] = ACTIONS(6563),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_API_DEPRECATED] = ACTIONS(6563),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6563),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6563),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6563),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6563),
    [anon_sym___deprecated_msg] = ACTIONS(6563),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6563),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6563),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6563),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6563),
    [anon_sym_BOOL] = ACTIONS(6563),
    [anon_sym_IMP] = ACTIONS(6563),
    [anon_sym_SEL] = ACTIONS(6563),
    [anon_sym_Class] = ACTIONS(6563),
    [anon_sym_id] = ACTIONS(6563),
    [anon_sym_ATdefs] = ACTIONS(6565),
  },
  [STATE(2722)] = {
    [sym_identifier] = ACTIONS(6567),
    [aux_sym_preproc_def_token1] = ACTIONS(6567),
    [aux_sym_preproc_if_token1] = ACTIONS(6567),
    [aux_sym_preproc_if_token2] = ACTIONS(6567),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6567),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6567),
    [sym_preproc_directive] = ACTIONS(6567),
    [anon_sym___extension__] = ACTIONS(6567),
    [anon_sym_extern] = ACTIONS(6567),
    [anon_sym___attribute__] = ACTIONS(6567),
    [anon_sym___attribute] = ACTIONS(6567),
    [anon_sym_noreturn] = ACTIONS(6567),
    [anon_sym_LBRACK] = ACTIONS(6569),
    [anon_sym___declspec] = ACTIONS(6567),
    [anon_sym_signed] = ACTIONS(6567),
    [anon_sym_unsigned] = ACTIONS(6567),
    [anon_sym_long] = ACTIONS(6567),
    [anon_sym_short] = ACTIONS(6567),
    [anon_sym_static] = ACTIONS(6567),
    [anon_sym_auto] = ACTIONS(6567),
    [anon_sym_register] = ACTIONS(6567),
    [anon_sym_inline] = ACTIONS(6567),
    [anon_sym___inline] = ACTIONS(6567),
    [anon_sym___inline__] = ACTIONS(6567),
    [anon_sym___forceinline] = ACTIONS(6567),
    [anon_sym_thread_local] = ACTIONS(6567),
    [anon_sym___thread] = ACTIONS(6567),
    [anon_sym_CG_EXTERN] = ACTIONS(6567),
    [anon_sym_CG_INLINE] = ACTIONS(6567),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6567),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6567),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6567),
    [anon_sym_IBOutlet] = ACTIONS(6567),
    [anon_sym_IBInspectable] = ACTIONS(6567),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6567),
    [anon_sym_NS_INLINE] = ACTIONS(6567),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6567),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6567),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6567),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6567),
    [anon_sym_const] = ACTIONS(6567),
    [anon_sym_constexpr] = ACTIONS(6567),
    [anon_sym_volatile] = ACTIONS(6567),
    [anon_sym_restrict] = ACTIONS(6567),
    [anon_sym___restrict__] = ACTIONS(6567),
    [anon_sym__Atomic] = ACTIONS(6567),
    [anon_sym__Noreturn] = ACTIONS(6567),
    [anon_sym__Nonnull] = ACTIONS(6567),
    [anon_sym_nullable] = ACTIONS(6567),
    [anon_sym__Complex] = ACTIONS(6567),
    [anon_sym__Nullable] = ACTIONS(6567),
    [anon_sym__Nullable_result] = ACTIONS(6567),
    [anon_sym__Null_unspecified] = ACTIONS(6567),
    [anon_sym___autoreleasing] = ACTIONS(6567),
    [anon_sym___block] = ACTIONS(6567),
    [anon_sym___bridge] = ACTIONS(6567),
    [anon_sym___bridge_retained] = ACTIONS(6567),
    [anon_sym___bridge_transfer] = ACTIONS(6567),
    [anon_sym___complex] = ACTIONS(6567),
    [anon_sym___const] = ACTIONS(6567),
    [anon_sym___imag] = ACTIONS(6567),
    [anon_sym___kindof] = ACTIONS(6567),
    [anon_sym___nonnull] = ACTIONS(6567),
    [anon_sym___nullable] = ACTIONS(6567),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6567),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6567),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6567),
    [anon_sym___real] = ACTIONS(6567),
    [anon_sym___strong] = ACTIONS(6567),
    [anon_sym___unsafe_unretained] = ACTIONS(6567),
    [anon_sym___unused] = ACTIONS(6567),
    [anon_sym___weak] = ACTIONS(6567),
    [anon_sym_alignas] = ACTIONS(6567),
    [anon_sym__Alignas] = ACTIONS(6567),
    [sym_primitive_type] = ACTIONS(6567),
    [anon_sym_enum] = ACTIONS(6567),
    [anon_sym_struct] = ACTIONS(6567),
    [anon_sym_union] = ACTIONS(6567),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6567),
    [anon_sym___typeof] = ACTIONS(6567),
    [anon_sym_typeof] = ACTIONS(6567),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6567),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6567),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6567),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6567),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6567),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6567),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6567),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6567),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6567),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_API_AVAILABLE] = ACTIONS(6567),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_API_DEPRECATED] = ACTIONS(6567),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6567),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6567),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6567),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6567),
    [anon_sym___deprecated_msg] = ACTIONS(6567),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6567),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6567),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6567),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6567),
    [anon_sym_BOOL] = ACTIONS(6567),
    [anon_sym_IMP] = ACTIONS(6567),
    [anon_sym_SEL] = ACTIONS(6567),
    [anon_sym_Class] = ACTIONS(6567),
    [anon_sym_id] = ACTIONS(6567),
    [anon_sym_ATdefs] = ACTIONS(6569),
  },
  [STATE(2723)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token2] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2724)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token2] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2725)] = {
    [sym_identifier] = ACTIONS(6575),
    [aux_sym_preproc_def_token1] = ACTIONS(6575),
    [aux_sym_preproc_if_token1] = ACTIONS(6575),
    [aux_sym_preproc_if_token2] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6575),
    [sym_preproc_directive] = ACTIONS(6575),
    [anon_sym___extension__] = ACTIONS(6575),
    [anon_sym_extern] = ACTIONS(6575),
    [anon_sym___attribute__] = ACTIONS(6575),
    [anon_sym___attribute] = ACTIONS(6575),
    [anon_sym_noreturn] = ACTIONS(6575),
    [anon_sym_LBRACK] = ACTIONS(6577),
    [anon_sym___declspec] = ACTIONS(6575),
    [anon_sym_signed] = ACTIONS(6575),
    [anon_sym_unsigned] = ACTIONS(6575),
    [anon_sym_long] = ACTIONS(6575),
    [anon_sym_short] = ACTIONS(6575),
    [anon_sym_static] = ACTIONS(6575),
    [anon_sym_auto] = ACTIONS(6575),
    [anon_sym_register] = ACTIONS(6575),
    [anon_sym_inline] = ACTIONS(6575),
    [anon_sym___inline] = ACTIONS(6575),
    [anon_sym___inline__] = ACTIONS(6575),
    [anon_sym___forceinline] = ACTIONS(6575),
    [anon_sym_thread_local] = ACTIONS(6575),
    [anon_sym___thread] = ACTIONS(6575),
    [anon_sym_CG_EXTERN] = ACTIONS(6575),
    [anon_sym_CG_INLINE] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6575),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6575),
    [anon_sym_IBOutlet] = ACTIONS(6575),
    [anon_sym_IBInspectable] = ACTIONS(6575),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6575),
    [anon_sym_NS_INLINE] = ACTIONS(6575),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6575),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6575),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6575),
    [anon_sym_const] = ACTIONS(6575),
    [anon_sym_constexpr] = ACTIONS(6575),
    [anon_sym_volatile] = ACTIONS(6575),
    [anon_sym_restrict] = ACTIONS(6575),
    [anon_sym___restrict__] = ACTIONS(6575),
    [anon_sym__Atomic] = ACTIONS(6575),
    [anon_sym__Noreturn] = ACTIONS(6575),
    [anon_sym__Nonnull] = ACTIONS(6575),
    [anon_sym_nullable] = ACTIONS(6575),
    [anon_sym__Complex] = ACTIONS(6575),
    [anon_sym__Nullable] = ACTIONS(6575),
    [anon_sym__Nullable_result] = ACTIONS(6575),
    [anon_sym__Null_unspecified] = ACTIONS(6575),
    [anon_sym___autoreleasing] = ACTIONS(6575),
    [anon_sym___block] = ACTIONS(6575),
    [anon_sym___bridge] = ACTIONS(6575),
    [anon_sym___bridge_retained] = ACTIONS(6575),
    [anon_sym___bridge_transfer] = ACTIONS(6575),
    [anon_sym___complex] = ACTIONS(6575),
    [anon_sym___const] = ACTIONS(6575),
    [anon_sym___imag] = ACTIONS(6575),
    [anon_sym___kindof] = ACTIONS(6575),
    [anon_sym___nonnull] = ACTIONS(6575),
    [anon_sym___nullable] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6575),
    [anon_sym___real] = ACTIONS(6575),
    [anon_sym___strong] = ACTIONS(6575),
    [anon_sym___unsafe_unretained] = ACTIONS(6575),
    [anon_sym___unused] = ACTIONS(6575),
    [anon_sym___weak] = ACTIONS(6575),
    [anon_sym_alignas] = ACTIONS(6575),
    [anon_sym__Alignas] = ACTIONS(6575),
    [sym_primitive_type] = ACTIONS(6575),
    [anon_sym_enum] = ACTIONS(6575),
    [anon_sym_struct] = ACTIONS(6575),
    [anon_sym_union] = ACTIONS(6575),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6575),
    [anon_sym___typeof] = ACTIONS(6575),
    [anon_sym_typeof] = ACTIONS(6575),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6575),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6575),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6575),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_API_AVAILABLE] = ACTIONS(6575),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_API_DEPRECATED] = ACTIONS(6575),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym___deprecated_msg] = ACTIONS(6575),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6575),
    [anon_sym_BOOL] = ACTIONS(6575),
    [anon_sym_IMP] = ACTIONS(6575),
    [anon_sym_SEL] = ACTIONS(6575),
    [anon_sym_Class] = ACTIONS(6575),
    [anon_sym_id] = ACTIONS(6575),
    [anon_sym_ATdefs] = ACTIONS(6577),
  },
  [STATE(2726)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token2] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2727)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token2] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2728)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token2] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2729)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token2] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2730)] = {
    [sym_identifier] = ACTIONS(6575),
    [aux_sym_preproc_def_token1] = ACTIONS(6575),
    [aux_sym_preproc_if_token1] = ACTIONS(6575),
    [aux_sym_preproc_if_token2] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6575),
    [sym_preproc_directive] = ACTIONS(6575),
    [anon_sym___extension__] = ACTIONS(6575),
    [anon_sym_extern] = ACTIONS(6575),
    [anon_sym___attribute__] = ACTIONS(6575),
    [anon_sym___attribute] = ACTIONS(6575),
    [anon_sym_noreturn] = ACTIONS(6575),
    [anon_sym_LBRACK] = ACTIONS(6577),
    [anon_sym___declspec] = ACTIONS(6575),
    [anon_sym_signed] = ACTIONS(6575),
    [anon_sym_unsigned] = ACTIONS(6575),
    [anon_sym_long] = ACTIONS(6575),
    [anon_sym_short] = ACTIONS(6575),
    [anon_sym_static] = ACTIONS(6575),
    [anon_sym_auto] = ACTIONS(6575),
    [anon_sym_register] = ACTIONS(6575),
    [anon_sym_inline] = ACTIONS(6575),
    [anon_sym___inline] = ACTIONS(6575),
    [anon_sym___inline__] = ACTIONS(6575),
    [anon_sym___forceinline] = ACTIONS(6575),
    [anon_sym_thread_local] = ACTIONS(6575),
    [anon_sym___thread] = ACTIONS(6575),
    [anon_sym_CG_EXTERN] = ACTIONS(6575),
    [anon_sym_CG_INLINE] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6575),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6575),
    [anon_sym_IBOutlet] = ACTIONS(6575),
    [anon_sym_IBInspectable] = ACTIONS(6575),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6575),
    [anon_sym_NS_INLINE] = ACTIONS(6575),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6575),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6575),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6575),
    [anon_sym_const] = ACTIONS(6575),
    [anon_sym_constexpr] = ACTIONS(6575),
    [anon_sym_volatile] = ACTIONS(6575),
    [anon_sym_restrict] = ACTIONS(6575),
    [anon_sym___restrict__] = ACTIONS(6575),
    [anon_sym__Atomic] = ACTIONS(6575),
    [anon_sym__Noreturn] = ACTIONS(6575),
    [anon_sym__Nonnull] = ACTIONS(6575),
    [anon_sym_nullable] = ACTIONS(6575),
    [anon_sym__Complex] = ACTIONS(6575),
    [anon_sym__Nullable] = ACTIONS(6575),
    [anon_sym__Nullable_result] = ACTIONS(6575),
    [anon_sym__Null_unspecified] = ACTIONS(6575),
    [anon_sym___autoreleasing] = ACTIONS(6575),
    [anon_sym___block] = ACTIONS(6575),
    [anon_sym___bridge] = ACTIONS(6575),
    [anon_sym___bridge_retained] = ACTIONS(6575),
    [anon_sym___bridge_transfer] = ACTIONS(6575),
    [anon_sym___complex] = ACTIONS(6575),
    [anon_sym___const] = ACTIONS(6575),
    [anon_sym___imag] = ACTIONS(6575),
    [anon_sym___kindof] = ACTIONS(6575),
    [anon_sym___nonnull] = ACTIONS(6575),
    [anon_sym___nullable] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6575),
    [anon_sym___real] = ACTIONS(6575),
    [anon_sym___strong] = ACTIONS(6575),
    [anon_sym___unsafe_unretained] = ACTIONS(6575),
    [anon_sym___unused] = ACTIONS(6575),
    [anon_sym___weak] = ACTIONS(6575),
    [anon_sym_alignas] = ACTIONS(6575),
    [anon_sym__Alignas] = ACTIONS(6575),
    [sym_primitive_type] = ACTIONS(6575),
    [anon_sym_enum] = ACTIONS(6575),
    [anon_sym_struct] = ACTIONS(6575),
    [anon_sym_union] = ACTIONS(6575),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6575),
    [anon_sym___typeof] = ACTIONS(6575),
    [anon_sym_typeof] = ACTIONS(6575),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6575),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6575),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6575),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_API_AVAILABLE] = ACTIONS(6575),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_API_DEPRECATED] = ACTIONS(6575),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym___deprecated_msg] = ACTIONS(6575),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6575),
    [anon_sym_BOOL] = ACTIONS(6575),
    [anon_sym_IMP] = ACTIONS(6575),
    [anon_sym_SEL] = ACTIONS(6575),
    [anon_sym_Class] = ACTIONS(6575),
    [anon_sym_id] = ACTIONS(6575),
    [anon_sym_ATdefs] = ACTIONS(6577),
  },
  [STATE(2731)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token2] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2732)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token2] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2733)] = {
    [sym_identifier] = ACTIONS(6599),
    [aux_sym_preproc_def_token1] = ACTIONS(6599),
    [aux_sym_preproc_if_token1] = ACTIONS(6599),
    [aux_sym_preproc_if_token2] = ACTIONS(6599),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6599),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6599),
    [sym_preproc_directive] = ACTIONS(6599),
    [anon_sym___extension__] = ACTIONS(6599),
    [anon_sym_extern] = ACTIONS(6599),
    [anon_sym___attribute__] = ACTIONS(6599),
    [anon_sym___attribute] = ACTIONS(6599),
    [anon_sym_noreturn] = ACTIONS(6599),
    [anon_sym_LBRACK] = ACTIONS(6601),
    [anon_sym___declspec] = ACTIONS(6599),
    [anon_sym_signed] = ACTIONS(6599),
    [anon_sym_unsigned] = ACTIONS(6599),
    [anon_sym_long] = ACTIONS(6599),
    [anon_sym_short] = ACTIONS(6599),
    [anon_sym_static] = ACTIONS(6599),
    [anon_sym_auto] = ACTIONS(6599),
    [anon_sym_register] = ACTIONS(6599),
    [anon_sym_inline] = ACTIONS(6599),
    [anon_sym___inline] = ACTIONS(6599),
    [anon_sym___inline__] = ACTIONS(6599),
    [anon_sym___forceinline] = ACTIONS(6599),
    [anon_sym_thread_local] = ACTIONS(6599),
    [anon_sym___thread] = ACTIONS(6599),
    [anon_sym_CG_EXTERN] = ACTIONS(6599),
    [anon_sym_CG_INLINE] = ACTIONS(6599),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6599),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6599),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6599),
    [anon_sym_IBOutlet] = ACTIONS(6599),
    [anon_sym_IBInspectable] = ACTIONS(6599),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6599),
    [anon_sym_NS_INLINE] = ACTIONS(6599),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6599),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6599),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6599),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6599),
    [anon_sym_const] = ACTIONS(6599),
    [anon_sym_constexpr] = ACTIONS(6599),
    [anon_sym_volatile] = ACTIONS(6599),
    [anon_sym_restrict] = ACTIONS(6599),
    [anon_sym___restrict__] = ACTIONS(6599),
    [anon_sym__Atomic] = ACTIONS(6599),
    [anon_sym__Noreturn] = ACTIONS(6599),
    [anon_sym__Nonnull] = ACTIONS(6599),
    [anon_sym_nullable] = ACTIONS(6599),
    [anon_sym__Complex] = ACTIONS(6599),
    [anon_sym__Nullable] = ACTIONS(6599),
    [anon_sym__Nullable_result] = ACTIONS(6599),
    [anon_sym__Null_unspecified] = ACTIONS(6599),
    [anon_sym___autoreleasing] = ACTIONS(6599),
    [anon_sym___block] = ACTIONS(6599),
    [anon_sym___bridge] = ACTIONS(6599),
    [anon_sym___bridge_retained] = ACTIONS(6599),
    [anon_sym___bridge_transfer] = ACTIONS(6599),
    [anon_sym___complex] = ACTIONS(6599),
    [anon_sym___const] = ACTIONS(6599),
    [anon_sym___imag] = ACTIONS(6599),
    [anon_sym___kindof] = ACTIONS(6599),
    [anon_sym___nonnull] = ACTIONS(6599),
    [anon_sym___nullable] = ACTIONS(6599),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6599),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6599),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6599),
    [anon_sym___real] = ACTIONS(6599),
    [anon_sym___strong] = ACTIONS(6599),
    [anon_sym___unsafe_unretained] = ACTIONS(6599),
    [anon_sym___unused] = ACTIONS(6599),
    [anon_sym___weak] = ACTIONS(6599),
    [anon_sym_alignas] = ACTIONS(6599),
    [anon_sym__Alignas] = ACTIONS(6599),
    [sym_primitive_type] = ACTIONS(6599),
    [anon_sym_enum] = ACTIONS(6599),
    [anon_sym_struct] = ACTIONS(6599),
    [anon_sym_union] = ACTIONS(6599),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6599),
    [anon_sym___typeof] = ACTIONS(6599),
    [anon_sym_typeof] = ACTIONS(6599),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6599),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6599),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6599),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6599),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6599),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6599),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6599),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6599),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6599),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_API_AVAILABLE] = ACTIONS(6599),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_API_DEPRECATED] = ACTIONS(6599),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6599),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6599),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6599),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6599),
    [anon_sym___deprecated_msg] = ACTIONS(6599),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6599),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6599),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6599),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6599),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6599),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6599),
    [anon_sym_BOOL] = ACTIONS(6599),
    [anon_sym_IMP] = ACTIONS(6599),
    [anon_sym_SEL] = ACTIONS(6599),
    [anon_sym_Class] = ACTIONS(6599),
    [anon_sym_id] = ACTIONS(6599),
    [anon_sym_ATdefs] = ACTIONS(6601),
  },
  [STATE(2734)] = {
    [sym_identifier] = ACTIONS(6519),
    [aux_sym_preproc_def_token1] = ACTIONS(6519),
    [aux_sym_preproc_if_token1] = ACTIONS(6519),
    [aux_sym_preproc_if_token2] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6519),
    [sym_preproc_directive] = ACTIONS(6519),
    [anon_sym___extension__] = ACTIONS(6519),
    [anon_sym_extern] = ACTIONS(6519),
    [anon_sym___attribute__] = ACTIONS(6519),
    [anon_sym___attribute] = ACTIONS(6519),
    [anon_sym_noreturn] = ACTIONS(6519),
    [anon_sym_LBRACK] = ACTIONS(6521),
    [anon_sym___declspec] = ACTIONS(6519),
    [anon_sym_signed] = ACTIONS(6519),
    [anon_sym_unsigned] = ACTIONS(6519),
    [anon_sym_long] = ACTIONS(6519),
    [anon_sym_short] = ACTIONS(6519),
    [anon_sym_static] = ACTIONS(6519),
    [anon_sym_auto] = ACTIONS(6519),
    [anon_sym_register] = ACTIONS(6519),
    [anon_sym_inline] = ACTIONS(6519),
    [anon_sym___inline] = ACTIONS(6519),
    [anon_sym___inline__] = ACTIONS(6519),
    [anon_sym___forceinline] = ACTIONS(6519),
    [anon_sym_thread_local] = ACTIONS(6519),
    [anon_sym___thread] = ACTIONS(6519),
    [anon_sym_CG_EXTERN] = ACTIONS(6519),
    [anon_sym_CG_INLINE] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6519),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6519),
    [anon_sym_IBOutlet] = ACTIONS(6519),
    [anon_sym_IBInspectable] = ACTIONS(6519),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6519),
    [anon_sym_NS_INLINE] = ACTIONS(6519),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6519),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6519),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6519),
    [anon_sym_const] = ACTIONS(6519),
    [anon_sym_constexpr] = ACTIONS(6519),
    [anon_sym_volatile] = ACTIONS(6519),
    [anon_sym_restrict] = ACTIONS(6519),
    [anon_sym___restrict__] = ACTIONS(6519),
    [anon_sym__Atomic] = ACTIONS(6519),
    [anon_sym__Noreturn] = ACTIONS(6519),
    [anon_sym__Nonnull] = ACTIONS(6519),
    [anon_sym_nullable] = ACTIONS(6519),
    [anon_sym__Complex] = ACTIONS(6519),
    [anon_sym__Nullable] = ACTIONS(6519),
    [anon_sym__Nullable_result] = ACTIONS(6519),
    [anon_sym__Null_unspecified] = ACTIONS(6519),
    [anon_sym___autoreleasing] = ACTIONS(6519),
    [anon_sym___block] = ACTIONS(6519),
    [anon_sym___bridge] = ACTIONS(6519),
    [anon_sym___bridge_retained] = ACTIONS(6519),
    [anon_sym___bridge_transfer] = ACTIONS(6519),
    [anon_sym___complex] = ACTIONS(6519),
    [anon_sym___const] = ACTIONS(6519),
    [anon_sym___imag] = ACTIONS(6519),
    [anon_sym___kindof] = ACTIONS(6519),
    [anon_sym___nonnull] = ACTIONS(6519),
    [anon_sym___nullable] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6519),
    [anon_sym___real] = ACTIONS(6519),
    [anon_sym___strong] = ACTIONS(6519),
    [anon_sym___unsafe_unretained] = ACTIONS(6519),
    [anon_sym___unused] = ACTIONS(6519),
    [anon_sym___weak] = ACTIONS(6519),
    [anon_sym_alignas] = ACTIONS(6519),
    [anon_sym__Alignas] = ACTIONS(6519),
    [sym_primitive_type] = ACTIONS(6519),
    [anon_sym_enum] = ACTIONS(6519),
    [anon_sym_struct] = ACTIONS(6519),
    [anon_sym_union] = ACTIONS(6519),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6519),
    [anon_sym___typeof] = ACTIONS(6519),
    [anon_sym_typeof] = ACTIONS(6519),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6519),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6519),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6519),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_API_AVAILABLE] = ACTIONS(6519),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_API_DEPRECATED] = ACTIONS(6519),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym___deprecated_msg] = ACTIONS(6519),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6519),
    [anon_sym_BOOL] = ACTIONS(6519),
    [anon_sym_IMP] = ACTIONS(6519),
    [anon_sym_SEL] = ACTIONS(6519),
    [anon_sym_Class] = ACTIONS(6519),
    [anon_sym_id] = ACTIONS(6519),
    [anon_sym_ATdefs] = ACTIONS(6521),
  },
  [STATE(2735)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token2] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2736)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token2] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2737)] = {
    [sym_identifier] = ACTIONS(6527),
    [aux_sym_preproc_def_token1] = ACTIONS(6527),
    [aux_sym_preproc_if_token1] = ACTIONS(6527),
    [aux_sym_preproc_if_token2] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6527),
    [sym_preproc_directive] = ACTIONS(6527),
    [anon_sym___extension__] = ACTIONS(6527),
    [anon_sym_extern] = ACTIONS(6527),
    [anon_sym___attribute__] = ACTIONS(6527),
    [anon_sym___attribute] = ACTIONS(6527),
    [anon_sym_noreturn] = ACTIONS(6527),
    [anon_sym_LBRACK] = ACTIONS(6529),
    [anon_sym___declspec] = ACTIONS(6527),
    [anon_sym_signed] = ACTIONS(6527),
    [anon_sym_unsigned] = ACTIONS(6527),
    [anon_sym_long] = ACTIONS(6527),
    [anon_sym_short] = ACTIONS(6527),
    [anon_sym_static] = ACTIONS(6527),
    [anon_sym_auto] = ACTIONS(6527),
    [anon_sym_register] = ACTIONS(6527),
    [anon_sym_inline] = ACTIONS(6527),
    [anon_sym___inline] = ACTIONS(6527),
    [anon_sym___inline__] = ACTIONS(6527),
    [anon_sym___forceinline] = ACTIONS(6527),
    [anon_sym_thread_local] = ACTIONS(6527),
    [anon_sym___thread] = ACTIONS(6527),
    [anon_sym_CG_EXTERN] = ACTIONS(6527),
    [anon_sym_CG_INLINE] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6527),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6527),
    [anon_sym_IBOutlet] = ACTIONS(6527),
    [anon_sym_IBInspectable] = ACTIONS(6527),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6527),
    [anon_sym_NS_INLINE] = ACTIONS(6527),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6527),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6527),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6527),
    [anon_sym_const] = ACTIONS(6527),
    [anon_sym_constexpr] = ACTIONS(6527),
    [anon_sym_volatile] = ACTIONS(6527),
    [anon_sym_restrict] = ACTIONS(6527),
    [anon_sym___restrict__] = ACTIONS(6527),
    [anon_sym__Atomic] = ACTIONS(6527),
    [anon_sym__Noreturn] = ACTIONS(6527),
    [anon_sym__Nonnull] = ACTIONS(6527),
    [anon_sym_nullable] = ACTIONS(6527),
    [anon_sym__Complex] = ACTIONS(6527),
    [anon_sym__Nullable] = ACTIONS(6527),
    [anon_sym__Nullable_result] = ACTIONS(6527),
    [anon_sym__Null_unspecified] = ACTIONS(6527),
    [anon_sym___autoreleasing] = ACTIONS(6527),
    [anon_sym___block] = ACTIONS(6527),
    [anon_sym___bridge] = ACTIONS(6527),
    [anon_sym___bridge_retained] = ACTIONS(6527),
    [anon_sym___bridge_transfer] = ACTIONS(6527),
    [anon_sym___complex] = ACTIONS(6527),
    [anon_sym___const] = ACTIONS(6527),
    [anon_sym___imag] = ACTIONS(6527),
    [anon_sym___kindof] = ACTIONS(6527),
    [anon_sym___nonnull] = ACTIONS(6527),
    [anon_sym___nullable] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6527),
    [anon_sym___real] = ACTIONS(6527),
    [anon_sym___strong] = ACTIONS(6527),
    [anon_sym___unsafe_unretained] = ACTIONS(6527),
    [anon_sym___unused] = ACTIONS(6527),
    [anon_sym___weak] = ACTIONS(6527),
    [anon_sym_alignas] = ACTIONS(6527),
    [anon_sym__Alignas] = ACTIONS(6527),
    [sym_primitive_type] = ACTIONS(6527),
    [anon_sym_enum] = ACTIONS(6527),
    [anon_sym_struct] = ACTIONS(6527),
    [anon_sym_union] = ACTIONS(6527),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6527),
    [anon_sym___typeof] = ACTIONS(6527),
    [anon_sym_typeof] = ACTIONS(6527),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6527),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6527),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6527),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_API_AVAILABLE] = ACTIONS(6527),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_API_DEPRECATED] = ACTIONS(6527),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym___deprecated_msg] = ACTIONS(6527),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6527),
    [anon_sym_BOOL] = ACTIONS(6527),
    [anon_sym_IMP] = ACTIONS(6527),
    [anon_sym_SEL] = ACTIONS(6527),
    [anon_sym_Class] = ACTIONS(6527),
    [anon_sym_id] = ACTIONS(6527),
    [anon_sym_ATdefs] = ACTIONS(6529),
  },
  [STATE(2738)] = {
    [sym_identifier] = ACTIONS(6519),
    [aux_sym_preproc_def_token1] = ACTIONS(6519),
    [aux_sym_preproc_if_token1] = ACTIONS(6519),
    [aux_sym_preproc_if_token2] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6519),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6519),
    [sym_preproc_directive] = ACTIONS(6519),
    [anon_sym___extension__] = ACTIONS(6519),
    [anon_sym_extern] = ACTIONS(6519),
    [anon_sym___attribute__] = ACTIONS(6519),
    [anon_sym___attribute] = ACTIONS(6519),
    [anon_sym_noreturn] = ACTIONS(6519),
    [anon_sym_LBRACK] = ACTIONS(6521),
    [anon_sym___declspec] = ACTIONS(6519),
    [anon_sym_signed] = ACTIONS(6519),
    [anon_sym_unsigned] = ACTIONS(6519),
    [anon_sym_long] = ACTIONS(6519),
    [anon_sym_short] = ACTIONS(6519),
    [anon_sym_static] = ACTIONS(6519),
    [anon_sym_auto] = ACTIONS(6519),
    [anon_sym_register] = ACTIONS(6519),
    [anon_sym_inline] = ACTIONS(6519),
    [anon_sym___inline] = ACTIONS(6519),
    [anon_sym___inline__] = ACTIONS(6519),
    [anon_sym___forceinline] = ACTIONS(6519),
    [anon_sym_thread_local] = ACTIONS(6519),
    [anon_sym___thread] = ACTIONS(6519),
    [anon_sym_CG_EXTERN] = ACTIONS(6519),
    [anon_sym_CG_INLINE] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6519),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6519),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6519),
    [anon_sym_IBOutlet] = ACTIONS(6519),
    [anon_sym_IBInspectable] = ACTIONS(6519),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6519),
    [anon_sym_NS_INLINE] = ACTIONS(6519),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6519),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6519),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6519),
    [anon_sym_const] = ACTIONS(6519),
    [anon_sym_constexpr] = ACTIONS(6519),
    [anon_sym_volatile] = ACTIONS(6519),
    [anon_sym_restrict] = ACTIONS(6519),
    [anon_sym___restrict__] = ACTIONS(6519),
    [anon_sym__Atomic] = ACTIONS(6519),
    [anon_sym__Noreturn] = ACTIONS(6519),
    [anon_sym__Nonnull] = ACTIONS(6519),
    [anon_sym_nullable] = ACTIONS(6519),
    [anon_sym__Complex] = ACTIONS(6519),
    [anon_sym__Nullable] = ACTIONS(6519),
    [anon_sym__Nullable_result] = ACTIONS(6519),
    [anon_sym__Null_unspecified] = ACTIONS(6519),
    [anon_sym___autoreleasing] = ACTIONS(6519),
    [anon_sym___block] = ACTIONS(6519),
    [anon_sym___bridge] = ACTIONS(6519),
    [anon_sym___bridge_retained] = ACTIONS(6519),
    [anon_sym___bridge_transfer] = ACTIONS(6519),
    [anon_sym___complex] = ACTIONS(6519),
    [anon_sym___const] = ACTIONS(6519),
    [anon_sym___imag] = ACTIONS(6519),
    [anon_sym___kindof] = ACTIONS(6519),
    [anon_sym___nonnull] = ACTIONS(6519),
    [anon_sym___nullable] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6519),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6519),
    [anon_sym___real] = ACTIONS(6519),
    [anon_sym___strong] = ACTIONS(6519),
    [anon_sym___unsafe_unretained] = ACTIONS(6519),
    [anon_sym___unused] = ACTIONS(6519),
    [anon_sym___weak] = ACTIONS(6519),
    [anon_sym_alignas] = ACTIONS(6519),
    [anon_sym__Alignas] = ACTIONS(6519),
    [sym_primitive_type] = ACTIONS(6519),
    [anon_sym_enum] = ACTIONS(6519),
    [anon_sym_struct] = ACTIONS(6519),
    [anon_sym_union] = ACTIONS(6519),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6519),
    [anon_sym___typeof] = ACTIONS(6519),
    [anon_sym_typeof] = ACTIONS(6519),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6519),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6519),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6519),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6519),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6519),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_API_AVAILABLE] = ACTIONS(6519),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_API_DEPRECATED] = ACTIONS(6519),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6519),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6519),
    [anon_sym___deprecated_msg] = ACTIONS(6519),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6519),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6519),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6519),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6519),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6519),
    [anon_sym_BOOL] = ACTIONS(6519),
    [anon_sym_IMP] = ACTIONS(6519),
    [anon_sym_SEL] = ACTIONS(6519),
    [anon_sym_Class] = ACTIONS(6519),
    [anon_sym_id] = ACTIONS(6519),
    [anon_sym_ATdefs] = ACTIONS(6521),
  },
  [STATE(2739)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token2] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2740)] = {
    [sym_identifier] = ACTIONS(6523),
    [aux_sym_preproc_def_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token1] = ACTIONS(6523),
    [aux_sym_preproc_if_token2] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6523),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6523),
    [sym_preproc_directive] = ACTIONS(6523),
    [anon_sym___extension__] = ACTIONS(6523),
    [anon_sym_extern] = ACTIONS(6523),
    [anon_sym___attribute__] = ACTIONS(6523),
    [anon_sym___attribute] = ACTIONS(6523),
    [anon_sym_noreturn] = ACTIONS(6523),
    [anon_sym_LBRACK] = ACTIONS(6525),
    [anon_sym___declspec] = ACTIONS(6523),
    [anon_sym_signed] = ACTIONS(6523),
    [anon_sym_unsigned] = ACTIONS(6523),
    [anon_sym_long] = ACTIONS(6523),
    [anon_sym_short] = ACTIONS(6523),
    [anon_sym_static] = ACTIONS(6523),
    [anon_sym_auto] = ACTIONS(6523),
    [anon_sym_register] = ACTIONS(6523),
    [anon_sym_inline] = ACTIONS(6523),
    [anon_sym___inline] = ACTIONS(6523),
    [anon_sym___inline__] = ACTIONS(6523),
    [anon_sym___forceinline] = ACTIONS(6523),
    [anon_sym_thread_local] = ACTIONS(6523),
    [anon_sym___thread] = ACTIONS(6523),
    [anon_sym_CG_EXTERN] = ACTIONS(6523),
    [anon_sym_CG_INLINE] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6523),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6523),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6523),
    [anon_sym_IBOutlet] = ACTIONS(6523),
    [anon_sym_IBInspectable] = ACTIONS(6523),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6523),
    [anon_sym_NS_INLINE] = ACTIONS(6523),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6523),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6523),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6523),
    [anon_sym_const] = ACTIONS(6523),
    [anon_sym_constexpr] = ACTIONS(6523),
    [anon_sym_volatile] = ACTIONS(6523),
    [anon_sym_restrict] = ACTIONS(6523),
    [anon_sym___restrict__] = ACTIONS(6523),
    [anon_sym__Atomic] = ACTIONS(6523),
    [anon_sym__Noreturn] = ACTIONS(6523),
    [anon_sym__Nonnull] = ACTIONS(6523),
    [anon_sym_nullable] = ACTIONS(6523),
    [anon_sym__Complex] = ACTIONS(6523),
    [anon_sym__Nullable] = ACTIONS(6523),
    [anon_sym__Nullable_result] = ACTIONS(6523),
    [anon_sym__Null_unspecified] = ACTIONS(6523),
    [anon_sym___autoreleasing] = ACTIONS(6523),
    [anon_sym___block] = ACTIONS(6523),
    [anon_sym___bridge] = ACTIONS(6523),
    [anon_sym___bridge_retained] = ACTIONS(6523),
    [anon_sym___bridge_transfer] = ACTIONS(6523),
    [anon_sym___complex] = ACTIONS(6523),
    [anon_sym___const] = ACTIONS(6523),
    [anon_sym___imag] = ACTIONS(6523),
    [anon_sym___kindof] = ACTIONS(6523),
    [anon_sym___nonnull] = ACTIONS(6523),
    [anon_sym___nullable] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6523),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6523),
    [anon_sym___real] = ACTIONS(6523),
    [anon_sym___strong] = ACTIONS(6523),
    [anon_sym___unsafe_unretained] = ACTIONS(6523),
    [anon_sym___unused] = ACTIONS(6523),
    [anon_sym___weak] = ACTIONS(6523),
    [anon_sym_alignas] = ACTIONS(6523),
    [anon_sym__Alignas] = ACTIONS(6523),
    [sym_primitive_type] = ACTIONS(6523),
    [anon_sym_enum] = ACTIONS(6523),
    [anon_sym_struct] = ACTIONS(6523),
    [anon_sym_union] = ACTIONS(6523),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6523),
    [anon_sym___typeof] = ACTIONS(6523),
    [anon_sym_typeof] = ACTIONS(6523),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6523),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6523),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6523),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6523),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6523),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_API_AVAILABLE] = ACTIONS(6523),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_API_DEPRECATED] = ACTIONS(6523),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6523),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6523),
    [anon_sym___deprecated_msg] = ACTIONS(6523),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6523),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6523),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6523),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6523),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6523),
    [anon_sym_BOOL] = ACTIONS(6523),
    [anon_sym_IMP] = ACTIONS(6523),
    [anon_sym_SEL] = ACTIONS(6523),
    [anon_sym_Class] = ACTIONS(6523),
    [anon_sym_id] = ACTIONS(6523),
    [anon_sym_ATdefs] = ACTIONS(6525),
  },
  [STATE(2741)] = {
    [sym_identifier] = ACTIONS(6527),
    [aux_sym_preproc_def_token1] = ACTIONS(6527),
    [aux_sym_preproc_if_token1] = ACTIONS(6527),
    [aux_sym_preproc_if_token2] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6527),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6527),
    [sym_preproc_directive] = ACTIONS(6527),
    [anon_sym___extension__] = ACTIONS(6527),
    [anon_sym_extern] = ACTIONS(6527),
    [anon_sym___attribute__] = ACTIONS(6527),
    [anon_sym___attribute] = ACTIONS(6527),
    [anon_sym_noreturn] = ACTIONS(6527),
    [anon_sym_LBRACK] = ACTIONS(6529),
    [anon_sym___declspec] = ACTIONS(6527),
    [anon_sym_signed] = ACTIONS(6527),
    [anon_sym_unsigned] = ACTIONS(6527),
    [anon_sym_long] = ACTIONS(6527),
    [anon_sym_short] = ACTIONS(6527),
    [anon_sym_static] = ACTIONS(6527),
    [anon_sym_auto] = ACTIONS(6527),
    [anon_sym_register] = ACTIONS(6527),
    [anon_sym_inline] = ACTIONS(6527),
    [anon_sym___inline] = ACTIONS(6527),
    [anon_sym___inline__] = ACTIONS(6527),
    [anon_sym___forceinline] = ACTIONS(6527),
    [anon_sym_thread_local] = ACTIONS(6527),
    [anon_sym___thread] = ACTIONS(6527),
    [anon_sym_CG_EXTERN] = ACTIONS(6527),
    [anon_sym_CG_INLINE] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6527),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6527),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6527),
    [anon_sym_IBOutlet] = ACTIONS(6527),
    [anon_sym_IBInspectable] = ACTIONS(6527),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6527),
    [anon_sym_NS_INLINE] = ACTIONS(6527),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6527),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6527),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6527),
    [anon_sym_const] = ACTIONS(6527),
    [anon_sym_constexpr] = ACTIONS(6527),
    [anon_sym_volatile] = ACTIONS(6527),
    [anon_sym_restrict] = ACTIONS(6527),
    [anon_sym___restrict__] = ACTIONS(6527),
    [anon_sym__Atomic] = ACTIONS(6527),
    [anon_sym__Noreturn] = ACTIONS(6527),
    [anon_sym__Nonnull] = ACTIONS(6527),
    [anon_sym_nullable] = ACTIONS(6527),
    [anon_sym__Complex] = ACTIONS(6527),
    [anon_sym__Nullable] = ACTIONS(6527),
    [anon_sym__Nullable_result] = ACTIONS(6527),
    [anon_sym__Null_unspecified] = ACTIONS(6527),
    [anon_sym___autoreleasing] = ACTIONS(6527),
    [anon_sym___block] = ACTIONS(6527),
    [anon_sym___bridge] = ACTIONS(6527),
    [anon_sym___bridge_retained] = ACTIONS(6527),
    [anon_sym___bridge_transfer] = ACTIONS(6527),
    [anon_sym___complex] = ACTIONS(6527),
    [anon_sym___const] = ACTIONS(6527),
    [anon_sym___imag] = ACTIONS(6527),
    [anon_sym___kindof] = ACTIONS(6527),
    [anon_sym___nonnull] = ACTIONS(6527),
    [anon_sym___nullable] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6527),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6527),
    [anon_sym___real] = ACTIONS(6527),
    [anon_sym___strong] = ACTIONS(6527),
    [anon_sym___unsafe_unretained] = ACTIONS(6527),
    [anon_sym___unused] = ACTIONS(6527),
    [anon_sym___weak] = ACTIONS(6527),
    [anon_sym_alignas] = ACTIONS(6527),
    [anon_sym__Alignas] = ACTIONS(6527),
    [sym_primitive_type] = ACTIONS(6527),
    [anon_sym_enum] = ACTIONS(6527),
    [anon_sym_struct] = ACTIONS(6527),
    [anon_sym_union] = ACTIONS(6527),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6527),
    [anon_sym___typeof] = ACTIONS(6527),
    [anon_sym_typeof] = ACTIONS(6527),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6527),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6527),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6527),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6527),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6527),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_API_AVAILABLE] = ACTIONS(6527),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_API_DEPRECATED] = ACTIONS(6527),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6527),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6527),
    [anon_sym___deprecated_msg] = ACTIONS(6527),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6527),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6527),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6527),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6527),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6527),
    [anon_sym_BOOL] = ACTIONS(6527),
    [anon_sym_IMP] = ACTIONS(6527),
    [anon_sym_SEL] = ACTIONS(6527),
    [anon_sym_Class] = ACTIONS(6527),
    [anon_sym_id] = ACTIONS(6527),
    [anon_sym_ATdefs] = ACTIONS(6529),
  },
  [STATE(2742)] = {
    [sym_identifier] = ACTIONS(6531),
    [aux_sym_preproc_def_token1] = ACTIONS(6531),
    [aux_sym_preproc_if_token1] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6531),
    [sym_preproc_directive] = ACTIONS(6531),
    [anon_sym___extension__] = ACTIONS(6531),
    [anon_sym_extern] = ACTIONS(6531),
    [anon_sym___attribute__] = ACTIONS(6531),
    [anon_sym___attribute] = ACTIONS(6531),
    [anon_sym_noreturn] = ACTIONS(6531),
    [anon_sym_LBRACK] = ACTIONS(6533),
    [anon_sym___declspec] = ACTIONS(6531),
    [anon_sym_RBRACE] = ACTIONS(6533),
    [anon_sym_signed] = ACTIONS(6531),
    [anon_sym_unsigned] = ACTIONS(6531),
    [anon_sym_long] = ACTIONS(6531),
    [anon_sym_short] = ACTIONS(6531),
    [anon_sym_static] = ACTIONS(6531),
    [anon_sym_auto] = ACTIONS(6531),
    [anon_sym_register] = ACTIONS(6531),
    [anon_sym_inline] = ACTIONS(6531),
    [anon_sym___inline] = ACTIONS(6531),
    [anon_sym___inline__] = ACTIONS(6531),
    [anon_sym___forceinline] = ACTIONS(6531),
    [anon_sym_thread_local] = ACTIONS(6531),
    [anon_sym___thread] = ACTIONS(6531),
    [anon_sym_CG_EXTERN] = ACTIONS(6531),
    [anon_sym_CG_INLINE] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6531),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6531),
    [anon_sym_IBOutlet] = ACTIONS(6531),
    [anon_sym_IBInspectable] = ACTIONS(6531),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6531),
    [anon_sym_NS_INLINE] = ACTIONS(6531),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6531),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6531),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6531),
    [anon_sym_const] = ACTIONS(6531),
    [anon_sym_constexpr] = ACTIONS(6531),
    [anon_sym_volatile] = ACTIONS(6531),
    [anon_sym_restrict] = ACTIONS(6531),
    [anon_sym___restrict__] = ACTIONS(6531),
    [anon_sym__Atomic] = ACTIONS(6531),
    [anon_sym__Noreturn] = ACTIONS(6531),
    [anon_sym__Nonnull] = ACTIONS(6531),
    [anon_sym_nullable] = ACTIONS(6531),
    [anon_sym__Complex] = ACTIONS(6531),
    [anon_sym__Nullable] = ACTIONS(6531),
    [anon_sym__Nullable_result] = ACTIONS(6531),
    [anon_sym__Null_unspecified] = ACTIONS(6531),
    [anon_sym___autoreleasing] = ACTIONS(6531),
    [anon_sym___block] = ACTIONS(6531),
    [anon_sym___bridge] = ACTIONS(6531),
    [anon_sym___bridge_retained] = ACTIONS(6531),
    [anon_sym___bridge_transfer] = ACTIONS(6531),
    [anon_sym___complex] = ACTIONS(6531),
    [anon_sym___const] = ACTIONS(6531),
    [anon_sym___imag] = ACTIONS(6531),
    [anon_sym___kindof] = ACTIONS(6531),
    [anon_sym___nonnull] = ACTIONS(6531),
    [anon_sym___nullable] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6531),
    [anon_sym___real] = ACTIONS(6531),
    [anon_sym___strong] = ACTIONS(6531),
    [anon_sym___unsafe_unretained] = ACTIONS(6531),
    [anon_sym___unused] = ACTIONS(6531),
    [anon_sym___weak] = ACTIONS(6531),
    [anon_sym_alignas] = ACTIONS(6531),
    [anon_sym__Alignas] = ACTIONS(6531),
    [sym_primitive_type] = ACTIONS(6531),
    [anon_sym_enum] = ACTIONS(6531),
    [anon_sym_struct] = ACTIONS(6531),
    [anon_sym_union] = ACTIONS(6531),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6531),
    [anon_sym___typeof] = ACTIONS(6531),
    [anon_sym_typeof] = ACTIONS(6531),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6531),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6531),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6531),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_API_AVAILABLE] = ACTIONS(6531),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_API_DEPRECATED] = ACTIONS(6531),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym___deprecated_msg] = ACTIONS(6531),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6531),
    [anon_sym_BOOL] = ACTIONS(6531),
    [anon_sym_IMP] = ACTIONS(6531),
    [anon_sym_SEL] = ACTIONS(6531),
    [anon_sym_Class] = ACTIONS(6531),
    [anon_sym_id] = ACTIONS(6531),
    [anon_sym_ATdefs] = ACTIONS(6533),
  },
  [STATE(2743)] = {
    [sym_identifier] = ACTIONS(6531),
    [aux_sym_preproc_def_token1] = ACTIONS(6531),
    [aux_sym_preproc_if_token1] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6531),
    [sym_preproc_directive] = ACTIONS(6531),
    [anon_sym___extension__] = ACTIONS(6531),
    [anon_sym_extern] = ACTIONS(6531),
    [anon_sym___attribute__] = ACTIONS(6531),
    [anon_sym___attribute] = ACTIONS(6531),
    [anon_sym_noreturn] = ACTIONS(6531),
    [anon_sym_LBRACK] = ACTIONS(6533),
    [anon_sym___declspec] = ACTIONS(6531),
    [anon_sym_RBRACE] = ACTIONS(6533),
    [anon_sym_signed] = ACTIONS(6531),
    [anon_sym_unsigned] = ACTIONS(6531),
    [anon_sym_long] = ACTIONS(6531),
    [anon_sym_short] = ACTIONS(6531),
    [anon_sym_static] = ACTIONS(6531),
    [anon_sym_auto] = ACTIONS(6531),
    [anon_sym_register] = ACTIONS(6531),
    [anon_sym_inline] = ACTIONS(6531),
    [anon_sym___inline] = ACTIONS(6531),
    [anon_sym___inline__] = ACTIONS(6531),
    [anon_sym___forceinline] = ACTIONS(6531),
    [anon_sym_thread_local] = ACTIONS(6531),
    [anon_sym___thread] = ACTIONS(6531),
    [anon_sym_CG_EXTERN] = ACTIONS(6531),
    [anon_sym_CG_INLINE] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6531),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6531),
    [anon_sym_IBOutlet] = ACTIONS(6531),
    [anon_sym_IBInspectable] = ACTIONS(6531),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6531),
    [anon_sym_NS_INLINE] = ACTIONS(6531),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6531),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6531),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6531),
    [anon_sym_const] = ACTIONS(6531),
    [anon_sym_constexpr] = ACTIONS(6531),
    [anon_sym_volatile] = ACTIONS(6531),
    [anon_sym_restrict] = ACTIONS(6531),
    [anon_sym___restrict__] = ACTIONS(6531),
    [anon_sym__Atomic] = ACTIONS(6531),
    [anon_sym__Noreturn] = ACTIONS(6531),
    [anon_sym__Nonnull] = ACTIONS(6531),
    [anon_sym_nullable] = ACTIONS(6531),
    [anon_sym__Complex] = ACTIONS(6531),
    [anon_sym__Nullable] = ACTIONS(6531),
    [anon_sym__Nullable_result] = ACTIONS(6531),
    [anon_sym__Null_unspecified] = ACTIONS(6531),
    [anon_sym___autoreleasing] = ACTIONS(6531),
    [anon_sym___block] = ACTIONS(6531),
    [anon_sym___bridge] = ACTIONS(6531),
    [anon_sym___bridge_retained] = ACTIONS(6531),
    [anon_sym___bridge_transfer] = ACTIONS(6531),
    [anon_sym___complex] = ACTIONS(6531),
    [anon_sym___const] = ACTIONS(6531),
    [anon_sym___imag] = ACTIONS(6531),
    [anon_sym___kindof] = ACTIONS(6531),
    [anon_sym___nonnull] = ACTIONS(6531),
    [anon_sym___nullable] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6531),
    [anon_sym___real] = ACTIONS(6531),
    [anon_sym___strong] = ACTIONS(6531),
    [anon_sym___unsafe_unretained] = ACTIONS(6531),
    [anon_sym___unused] = ACTIONS(6531),
    [anon_sym___weak] = ACTIONS(6531),
    [anon_sym_alignas] = ACTIONS(6531),
    [anon_sym__Alignas] = ACTIONS(6531),
    [sym_primitive_type] = ACTIONS(6531),
    [anon_sym_enum] = ACTIONS(6531),
    [anon_sym_struct] = ACTIONS(6531),
    [anon_sym_union] = ACTIONS(6531),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6531),
    [anon_sym___typeof] = ACTIONS(6531),
    [anon_sym_typeof] = ACTIONS(6531),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6531),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6531),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6531),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_API_AVAILABLE] = ACTIONS(6531),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_API_DEPRECATED] = ACTIONS(6531),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym___deprecated_msg] = ACTIONS(6531),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6531),
    [anon_sym_BOOL] = ACTIONS(6531),
    [anon_sym_IMP] = ACTIONS(6531),
    [anon_sym_SEL] = ACTIONS(6531),
    [anon_sym_Class] = ACTIONS(6531),
    [anon_sym_id] = ACTIONS(6531),
    [anon_sym_ATdefs] = ACTIONS(6533),
  },
  [STATE(2744)] = {
    [sym_keyword_declarator] = STATE(2744),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(6726),
    [anon_sym_COMMA] = ACTIONS(6729),
    [anon_sym_SEMI] = ACTIONS(6731),
    [anon_sym___extension__] = ACTIONS(6734),
    [anon_sym_extern] = ACTIONS(6734),
    [anon_sym___attribute__] = ACTIONS(6734),
    [anon_sym___attribute] = ACTIONS(6734),
    [anon_sym_noreturn] = ACTIONS(6734),
    [anon_sym_LBRACK] = ACTIONS(6729),
    [anon_sym___declspec] = ACTIONS(6734),
    [anon_sym_LBRACE] = ACTIONS(6729),
    [anon_sym_signed] = ACTIONS(6734),
    [anon_sym_unsigned] = ACTIONS(6734),
    [anon_sym_long] = ACTIONS(6734),
    [anon_sym_short] = ACTIONS(6734),
    [anon_sym_static] = ACTIONS(6734),
    [anon_sym_ATautoreleasepool] = ACTIONS(6729),
    [anon_sym_auto] = ACTIONS(6734),
    [anon_sym_register] = ACTIONS(6734),
    [anon_sym_inline] = ACTIONS(6734),
    [anon_sym___inline] = ACTIONS(6734),
    [anon_sym___inline__] = ACTIONS(6734),
    [anon_sym___forceinline] = ACTIONS(6734),
    [anon_sym_thread_local] = ACTIONS(6734),
    [anon_sym___thread] = ACTIONS(6734),
    [anon_sym_CG_EXTERN] = ACTIONS(6734),
    [anon_sym_CG_INLINE] = ACTIONS(6734),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6734),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6734),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6734),
    [anon_sym_IBOutlet] = ACTIONS(6734),
    [anon_sym_IBInspectable] = ACTIONS(6734),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6734),
    [anon_sym_NS_INLINE] = ACTIONS(6734),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6734),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6734),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6734),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6734),
    [anon_sym_const] = ACTIONS(6734),
    [anon_sym_constexpr] = ACTIONS(6734),
    [anon_sym_volatile] = ACTIONS(6734),
    [anon_sym_restrict] = ACTIONS(6734),
    [anon_sym___restrict__] = ACTIONS(6734),
    [anon_sym__Atomic] = ACTIONS(6734),
    [anon_sym__Noreturn] = ACTIONS(6734),
    [anon_sym__Nonnull] = ACTIONS(6734),
    [anon_sym_nullable] = ACTIONS(6734),
    [anon_sym__Complex] = ACTIONS(6734),
    [anon_sym__Nullable] = ACTIONS(6734),
    [anon_sym__Nullable_result] = ACTIONS(6734),
    [anon_sym__Null_unspecified] = ACTIONS(6734),
    [anon_sym___autoreleasing] = ACTIONS(6734),
    [anon_sym___block] = ACTIONS(6734),
    [anon_sym___bridge] = ACTIONS(6734),
    [anon_sym___bridge_retained] = ACTIONS(6734),
    [anon_sym___bridge_transfer] = ACTIONS(6734),
    [anon_sym___complex] = ACTIONS(6734),
    [anon_sym___const] = ACTIONS(6734),
    [anon_sym___imag] = ACTIONS(6734),
    [anon_sym___kindof] = ACTIONS(6734),
    [anon_sym___nonnull] = ACTIONS(6734),
    [anon_sym___nullable] = ACTIONS(6734),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6734),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6734),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6734),
    [anon_sym___real] = ACTIONS(6734),
    [anon_sym___strong] = ACTIONS(6734),
    [anon_sym___unsafe_unretained] = ACTIONS(6734),
    [anon_sym___unused] = ACTIONS(6734),
    [anon_sym___weak] = ACTIONS(6734),
    [anon_sym_alignas] = ACTIONS(6734),
    [anon_sym__Alignas] = ACTIONS(6734),
    [sym_primitive_type] = ACTIONS(6734),
    [anon_sym_enum] = ACTIONS(6734),
    [anon_sym_COLON] = ACTIONS(6729),
    [anon_sym_struct] = ACTIONS(6734),
    [anon_sym_union] = ACTIONS(6734),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6734),
    [anon_sym___typeof] = ACTIONS(6734),
    [anon_sym_typeof] = ACTIONS(6734),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6734),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6734),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6734),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6734),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6734),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6734),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6734),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6734),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6734),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6734),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6734),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6734),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6734),
    [anon_sym_API_AVAILABLE] = ACTIONS(6734),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6734),
    [anon_sym_API_DEPRECATED] = ACTIONS(6734),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6734),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6734),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6734),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6734),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6734),
    [anon_sym___deprecated_msg] = ACTIONS(6734),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6734),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6734),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6734),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6734),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6734),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6734),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6734),
    [anon_sym_BOOL] = ACTIONS(6734),
    [anon_sym_IMP] = ACTIONS(6734),
    [anon_sym_SEL] = ACTIONS(6734),
    [anon_sym_Class] = ACTIONS(6734),
    [anon_sym_id] = ACTIONS(6734),
  },
  [STATE(2745)] = {
    [sym_identifier] = ACTIONS(6531),
    [aux_sym_preproc_def_token1] = ACTIONS(6531),
    [aux_sym_preproc_if_token1] = ACTIONS(6531),
    [aux_sym_preproc_if_token2] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6531),
    [sym_preproc_directive] = ACTIONS(6531),
    [anon_sym___extension__] = ACTIONS(6531),
    [anon_sym_extern] = ACTIONS(6531),
    [anon_sym___attribute__] = ACTIONS(6531),
    [anon_sym___attribute] = ACTIONS(6531),
    [anon_sym_noreturn] = ACTIONS(6531),
    [anon_sym_LBRACK] = ACTIONS(6533),
    [anon_sym___declspec] = ACTIONS(6531),
    [anon_sym_signed] = ACTIONS(6531),
    [anon_sym_unsigned] = ACTIONS(6531),
    [anon_sym_long] = ACTIONS(6531),
    [anon_sym_short] = ACTIONS(6531),
    [anon_sym_static] = ACTIONS(6531),
    [anon_sym_auto] = ACTIONS(6531),
    [anon_sym_register] = ACTIONS(6531),
    [anon_sym_inline] = ACTIONS(6531),
    [anon_sym___inline] = ACTIONS(6531),
    [anon_sym___inline__] = ACTIONS(6531),
    [anon_sym___forceinline] = ACTIONS(6531),
    [anon_sym_thread_local] = ACTIONS(6531),
    [anon_sym___thread] = ACTIONS(6531),
    [anon_sym_CG_EXTERN] = ACTIONS(6531),
    [anon_sym_CG_INLINE] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6531),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6531),
    [anon_sym_IBOutlet] = ACTIONS(6531),
    [anon_sym_IBInspectable] = ACTIONS(6531),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6531),
    [anon_sym_NS_INLINE] = ACTIONS(6531),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6531),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6531),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6531),
    [anon_sym_const] = ACTIONS(6531),
    [anon_sym_constexpr] = ACTIONS(6531),
    [anon_sym_volatile] = ACTIONS(6531),
    [anon_sym_restrict] = ACTIONS(6531),
    [anon_sym___restrict__] = ACTIONS(6531),
    [anon_sym__Atomic] = ACTIONS(6531),
    [anon_sym__Noreturn] = ACTIONS(6531),
    [anon_sym__Nonnull] = ACTIONS(6531),
    [anon_sym_nullable] = ACTIONS(6531),
    [anon_sym__Complex] = ACTIONS(6531),
    [anon_sym__Nullable] = ACTIONS(6531),
    [anon_sym__Nullable_result] = ACTIONS(6531),
    [anon_sym__Null_unspecified] = ACTIONS(6531),
    [anon_sym___autoreleasing] = ACTIONS(6531),
    [anon_sym___block] = ACTIONS(6531),
    [anon_sym___bridge] = ACTIONS(6531),
    [anon_sym___bridge_retained] = ACTIONS(6531),
    [anon_sym___bridge_transfer] = ACTIONS(6531),
    [anon_sym___complex] = ACTIONS(6531),
    [anon_sym___const] = ACTIONS(6531),
    [anon_sym___imag] = ACTIONS(6531),
    [anon_sym___kindof] = ACTIONS(6531),
    [anon_sym___nonnull] = ACTIONS(6531),
    [anon_sym___nullable] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6531),
    [anon_sym___real] = ACTIONS(6531),
    [anon_sym___strong] = ACTIONS(6531),
    [anon_sym___unsafe_unretained] = ACTIONS(6531),
    [anon_sym___unused] = ACTIONS(6531),
    [anon_sym___weak] = ACTIONS(6531),
    [anon_sym_alignas] = ACTIONS(6531),
    [anon_sym__Alignas] = ACTIONS(6531),
    [sym_primitive_type] = ACTIONS(6531),
    [anon_sym_enum] = ACTIONS(6531),
    [anon_sym_struct] = ACTIONS(6531),
    [anon_sym_union] = ACTIONS(6531),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6531),
    [anon_sym___typeof] = ACTIONS(6531),
    [anon_sym_typeof] = ACTIONS(6531),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6531),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6531),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6531),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_API_AVAILABLE] = ACTIONS(6531),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_API_DEPRECATED] = ACTIONS(6531),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym___deprecated_msg] = ACTIONS(6531),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6531),
    [anon_sym_BOOL] = ACTIONS(6531),
    [anon_sym_IMP] = ACTIONS(6531),
    [anon_sym_SEL] = ACTIONS(6531),
    [anon_sym_Class] = ACTIONS(6531),
    [anon_sym_id] = ACTIONS(6531),
    [anon_sym_ATdefs] = ACTIONS(6533),
  },
  [STATE(2746)] = {
    [sym_identifier] = ACTIONS(6531),
    [aux_sym_preproc_def_token1] = ACTIONS(6531),
    [aux_sym_preproc_if_token1] = ACTIONS(6531),
    [aux_sym_preproc_if_token2] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6531),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6531),
    [sym_preproc_directive] = ACTIONS(6531),
    [anon_sym___extension__] = ACTIONS(6531),
    [anon_sym_extern] = ACTIONS(6531),
    [anon_sym___attribute__] = ACTIONS(6531),
    [anon_sym___attribute] = ACTIONS(6531),
    [anon_sym_noreturn] = ACTIONS(6531),
    [anon_sym_LBRACK] = ACTIONS(6533),
    [anon_sym___declspec] = ACTIONS(6531),
    [anon_sym_signed] = ACTIONS(6531),
    [anon_sym_unsigned] = ACTIONS(6531),
    [anon_sym_long] = ACTIONS(6531),
    [anon_sym_short] = ACTIONS(6531),
    [anon_sym_static] = ACTIONS(6531),
    [anon_sym_auto] = ACTIONS(6531),
    [anon_sym_register] = ACTIONS(6531),
    [anon_sym_inline] = ACTIONS(6531),
    [anon_sym___inline] = ACTIONS(6531),
    [anon_sym___inline__] = ACTIONS(6531),
    [anon_sym___forceinline] = ACTIONS(6531),
    [anon_sym_thread_local] = ACTIONS(6531),
    [anon_sym___thread] = ACTIONS(6531),
    [anon_sym_CG_EXTERN] = ACTIONS(6531),
    [anon_sym_CG_INLINE] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6531),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6531),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6531),
    [anon_sym_IBOutlet] = ACTIONS(6531),
    [anon_sym_IBInspectable] = ACTIONS(6531),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6531),
    [anon_sym_NS_INLINE] = ACTIONS(6531),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6531),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6531),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6531),
    [anon_sym_const] = ACTIONS(6531),
    [anon_sym_constexpr] = ACTIONS(6531),
    [anon_sym_volatile] = ACTIONS(6531),
    [anon_sym_restrict] = ACTIONS(6531),
    [anon_sym___restrict__] = ACTIONS(6531),
    [anon_sym__Atomic] = ACTIONS(6531),
    [anon_sym__Noreturn] = ACTIONS(6531),
    [anon_sym__Nonnull] = ACTIONS(6531),
    [anon_sym_nullable] = ACTIONS(6531),
    [anon_sym__Complex] = ACTIONS(6531),
    [anon_sym__Nullable] = ACTIONS(6531),
    [anon_sym__Nullable_result] = ACTIONS(6531),
    [anon_sym__Null_unspecified] = ACTIONS(6531),
    [anon_sym___autoreleasing] = ACTIONS(6531),
    [anon_sym___block] = ACTIONS(6531),
    [anon_sym___bridge] = ACTIONS(6531),
    [anon_sym___bridge_retained] = ACTIONS(6531),
    [anon_sym___bridge_transfer] = ACTIONS(6531),
    [anon_sym___complex] = ACTIONS(6531),
    [anon_sym___const] = ACTIONS(6531),
    [anon_sym___imag] = ACTIONS(6531),
    [anon_sym___kindof] = ACTIONS(6531),
    [anon_sym___nonnull] = ACTIONS(6531),
    [anon_sym___nullable] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6531),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6531),
    [anon_sym___real] = ACTIONS(6531),
    [anon_sym___strong] = ACTIONS(6531),
    [anon_sym___unsafe_unretained] = ACTIONS(6531),
    [anon_sym___unused] = ACTIONS(6531),
    [anon_sym___weak] = ACTIONS(6531),
    [anon_sym_alignas] = ACTIONS(6531),
    [anon_sym__Alignas] = ACTIONS(6531),
    [sym_primitive_type] = ACTIONS(6531),
    [anon_sym_enum] = ACTIONS(6531),
    [anon_sym_struct] = ACTIONS(6531),
    [anon_sym_union] = ACTIONS(6531),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6531),
    [anon_sym___typeof] = ACTIONS(6531),
    [anon_sym_typeof] = ACTIONS(6531),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6531),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6531),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6531),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6531),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6531),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_API_AVAILABLE] = ACTIONS(6531),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_API_DEPRECATED] = ACTIONS(6531),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6531),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6531),
    [anon_sym___deprecated_msg] = ACTIONS(6531),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6531),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6531),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6531),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6531),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6531),
    [anon_sym_BOOL] = ACTIONS(6531),
    [anon_sym_IMP] = ACTIONS(6531),
    [anon_sym_SEL] = ACTIONS(6531),
    [anon_sym_Class] = ACTIONS(6531),
    [anon_sym_id] = ACTIONS(6531),
    [anon_sym_ATdefs] = ACTIONS(6533),
  },
  [STATE(2747)] = {
    [sym_identifier] = ACTIONS(4262),
    [anon_sym___extension__] = ACTIONS(4262),
    [anon_sym_extern] = ACTIONS(4262),
    [anon_sym___attribute__] = ACTIONS(4262),
    [anon_sym___attribute] = ACTIONS(4262),
    [anon_sym_noreturn] = ACTIONS(4262),
    [anon_sym_LBRACK] = ACTIONS(4264),
    [anon_sym___declspec] = ACTIONS(4262),
    [anon_sym___cdecl] = ACTIONS(4262),
    [anon_sym___clrcall] = ACTIONS(4262),
    [anon_sym___stdcall] = ACTIONS(4262),
    [anon_sym___fastcall] = ACTIONS(4262),
    [anon_sym___thiscall] = ACTIONS(4262),
    [anon_sym___vectorcall] = ACTIONS(4262),
    [anon_sym_LBRACE] = ACTIONS(4264),
    [anon_sym_signed] = ACTIONS(4262),
    [anon_sym_unsigned] = ACTIONS(4262),
    [anon_sym_long] = ACTIONS(4262),
    [anon_sym_short] = ACTIONS(4262),
    [anon_sym_static] = ACTIONS(4262),
    [anon_sym_auto] = ACTIONS(4262),
    [anon_sym_register] = ACTIONS(4262),
    [anon_sym_inline] = ACTIONS(4262),
    [anon_sym___inline] = ACTIONS(4262),
    [anon_sym___inline__] = ACTIONS(4262),
    [anon_sym___forceinline] = ACTIONS(4262),
    [anon_sym_thread_local] = ACTIONS(4262),
    [anon_sym___thread] = ACTIONS(4262),
    [anon_sym_CG_EXTERN] = ACTIONS(4262),
    [anon_sym_CG_INLINE] = ACTIONS(4262),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4262),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4262),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4262),
    [anon_sym_IBOutlet] = ACTIONS(4262),
    [anon_sym_IBInspectable] = ACTIONS(4262),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4262),
    [anon_sym_NS_INLINE] = ACTIONS(4262),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4262),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4262),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4262),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4262),
    [anon_sym_const] = ACTIONS(4262),
    [anon_sym_constexpr] = ACTIONS(4262),
    [anon_sym_volatile] = ACTIONS(4262),
    [anon_sym_restrict] = ACTIONS(4262),
    [anon_sym___restrict__] = ACTIONS(4262),
    [anon_sym__Atomic] = ACTIONS(4262),
    [anon_sym__Noreturn] = ACTIONS(4262),
    [anon_sym__Nonnull] = ACTIONS(4262),
    [anon_sym_nullable] = ACTIONS(4262),
    [anon_sym__Complex] = ACTIONS(4262),
    [anon_sym__Nullable] = ACTIONS(4262),
    [anon_sym__Nullable_result] = ACTIONS(4262),
    [anon_sym__Null_unspecified] = ACTIONS(4262),
    [anon_sym___autoreleasing] = ACTIONS(4262),
    [anon_sym___block] = ACTIONS(4262),
    [anon_sym___bridge] = ACTIONS(4262),
    [anon_sym___bridge_retained] = ACTIONS(4262),
    [anon_sym___bridge_transfer] = ACTIONS(4262),
    [anon_sym___complex] = ACTIONS(4262),
    [anon_sym___const] = ACTIONS(4262),
    [anon_sym___imag] = ACTIONS(4262),
    [anon_sym___kindof] = ACTIONS(4262),
    [anon_sym___nonnull] = ACTIONS(4262),
    [anon_sym___nullable] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4262),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4262),
    [anon_sym___real] = ACTIONS(4262),
    [anon_sym___strong] = ACTIONS(4262),
    [anon_sym___unsafe_unretained] = ACTIONS(4262),
    [anon_sym___unused] = ACTIONS(4262),
    [anon_sym___weak] = ACTIONS(4262),
    [anon_sym_alignas] = ACTIONS(4262),
    [anon_sym__Alignas] = ACTIONS(4262),
    [sym_primitive_type] = ACTIONS(4262),
    [anon_sym_enum] = ACTIONS(4262),
    [anon_sym_struct] = ACTIONS(4262),
    [anon_sym_union] = ACTIONS(4262),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4262),
    [anon_sym___typeof] = ACTIONS(4262),
    [anon_sym_typeof] = ACTIONS(4262),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4262),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4262),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4262),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4262),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4262),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4262),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4262),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_API_AVAILABLE] = ACTIONS(4262),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_API_DEPRECATED] = ACTIONS(4262),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4262),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4262),
    [anon_sym___deprecated_msg] = ACTIONS(4262),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4262),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4262),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4262),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4262),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4262),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4262),
    [anon_sym_BOOL] = ACTIONS(4262),
    [anon_sym_IMP] = ACTIONS(4262),
    [anon_sym_SEL] = ACTIONS(4262),
    [anon_sym_Class] = ACTIONS(4262),
    [anon_sym_id] = ACTIONS(4262),
  },
  [STATE(2748)] = {
    [sym_parameter_list] = STATE(2552),
    [sym_identifier] = ACTIONS(6736),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6630),
    [anon_sym_SEMI] = ACTIONS(6738),
    [anon_sym___extension__] = ACTIONS(6736),
    [anon_sym_extern] = ACTIONS(6736),
    [anon_sym___attribute__] = ACTIONS(6736),
    [anon_sym___attribute] = ACTIONS(6736),
    [anon_sym_noreturn] = ACTIONS(6736),
    [anon_sym_LBRACK] = ACTIONS(6632),
    [anon_sym___declspec] = ACTIONS(6736),
    [anon_sym_LBRACE] = ACTIONS(6738),
    [anon_sym_signed] = ACTIONS(6736),
    [anon_sym_unsigned] = ACTIONS(6736),
    [anon_sym_long] = ACTIONS(6736),
    [anon_sym_short] = ACTIONS(6736),
    [anon_sym_static] = ACTIONS(6736),
    [anon_sym_EQ] = ACTIONS(6738),
    [anon_sym_ATautoreleasepool] = ACTIONS(6738),
    [anon_sym_auto] = ACTIONS(6736),
    [anon_sym_register] = ACTIONS(6736),
    [anon_sym_inline] = ACTIONS(6736),
    [anon_sym___inline] = ACTIONS(6736),
    [anon_sym___inline__] = ACTIONS(6736),
    [anon_sym___forceinline] = ACTIONS(6736),
    [anon_sym_thread_local] = ACTIONS(6736),
    [anon_sym___thread] = ACTIONS(6736),
    [anon_sym_CG_EXTERN] = ACTIONS(6736),
    [anon_sym_CG_INLINE] = ACTIONS(6736),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6736),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6736),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6736),
    [anon_sym_IBOutlet] = ACTIONS(6736),
    [anon_sym_IBInspectable] = ACTIONS(6736),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6736),
    [anon_sym_NS_INLINE] = ACTIONS(6736),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6736),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6736),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6736),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6736),
    [anon_sym_const] = ACTIONS(6736),
    [anon_sym_constexpr] = ACTIONS(6736),
    [anon_sym_volatile] = ACTIONS(6736),
    [anon_sym_restrict] = ACTIONS(6736),
    [anon_sym___restrict__] = ACTIONS(6736),
    [anon_sym__Atomic] = ACTIONS(6736),
    [anon_sym__Noreturn] = ACTIONS(6736),
    [anon_sym__Nonnull] = ACTIONS(6736),
    [anon_sym_nullable] = ACTIONS(6736),
    [anon_sym__Complex] = ACTIONS(6736),
    [anon_sym__Nullable] = ACTIONS(6736),
    [anon_sym__Nullable_result] = ACTIONS(6736),
    [anon_sym__Null_unspecified] = ACTIONS(6736),
    [anon_sym___autoreleasing] = ACTIONS(6736),
    [anon_sym___block] = ACTIONS(6736),
    [anon_sym___bridge] = ACTIONS(6736),
    [anon_sym___bridge_retained] = ACTIONS(6736),
    [anon_sym___bridge_transfer] = ACTIONS(6736),
    [anon_sym___complex] = ACTIONS(6736),
    [anon_sym___const] = ACTIONS(6736),
    [anon_sym___imag] = ACTIONS(6736),
    [anon_sym___kindof] = ACTIONS(6736),
    [anon_sym___nonnull] = ACTIONS(6736),
    [anon_sym___nullable] = ACTIONS(6736),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6736),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6736),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6736),
    [anon_sym___real] = ACTIONS(6736),
    [anon_sym___strong] = ACTIONS(6736),
    [anon_sym___unsafe_unretained] = ACTIONS(6736),
    [anon_sym___unused] = ACTIONS(6736),
    [anon_sym___weak] = ACTIONS(6736),
    [anon_sym_alignas] = ACTIONS(6736),
    [anon_sym__Alignas] = ACTIONS(6736),
    [sym_primitive_type] = ACTIONS(6736),
    [anon_sym_enum] = ACTIONS(6736),
    [anon_sym_struct] = ACTIONS(6736),
    [anon_sym_union] = ACTIONS(6736),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6736),
    [anon_sym___typeof] = ACTIONS(6736),
    [anon_sym_typeof] = ACTIONS(6736),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6736),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6736),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6736),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6736),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6736),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6736),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6736),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6736),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6736),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6736),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6736),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6736),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6736),
    [anon_sym_API_AVAILABLE] = ACTIONS(6736),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6736),
    [anon_sym_API_DEPRECATED] = ACTIONS(6736),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6736),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6736),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6736),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6736),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6736),
    [anon_sym___deprecated_msg] = ACTIONS(6736),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6736),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6736),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6736),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6736),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6736),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6736),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6736),
    [anon_sym_BOOL] = ACTIONS(6736),
    [anon_sym_IMP] = ACTIONS(6736),
    [anon_sym_SEL] = ACTIONS(6736),
    [anon_sym_Class] = ACTIONS(6736),
    [anon_sym_id] = ACTIONS(6736),
  },
  [STATE(2749)] = {
    [sym_identifier] = ACTIONS(4118),
    [anon_sym___extension__] = ACTIONS(4118),
    [anon_sym_extern] = ACTIONS(4118),
    [anon_sym___attribute__] = ACTIONS(4118),
    [anon_sym___attribute] = ACTIONS(4118),
    [anon_sym_noreturn] = ACTIONS(4118),
    [anon_sym_LBRACK] = ACTIONS(4120),
    [anon_sym___declspec] = ACTIONS(4118),
    [anon_sym___cdecl] = ACTIONS(4118),
    [anon_sym___clrcall] = ACTIONS(4118),
    [anon_sym___stdcall] = ACTIONS(4118),
    [anon_sym___fastcall] = ACTIONS(4118),
    [anon_sym___thiscall] = ACTIONS(4118),
    [anon_sym___vectorcall] = ACTIONS(4118),
    [anon_sym_LBRACE] = ACTIONS(4120),
    [anon_sym_signed] = ACTIONS(4118),
    [anon_sym_unsigned] = ACTIONS(4118),
    [anon_sym_long] = ACTIONS(4118),
    [anon_sym_short] = ACTIONS(4118),
    [anon_sym_static] = ACTIONS(4118),
    [anon_sym_auto] = ACTIONS(4118),
    [anon_sym_register] = ACTIONS(4118),
    [anon_sym_inline] = ACTIONS(4118),
    [anon_sym___inline] = ACTIONS(4118),
    [anon_sym___inline__] = ACTIONS(4118),
    [anon_sym___forceinline] = ACTIONS(4118),
    [anon_sym_thread_local] = ACTIONS(4118),
    [anon_sym___thread] = ACTIONS(4118),
    [anon_sym_CG_EXTERN] = ACTIONS(4118),
    [anon_sym_CG_INLINE] = ACTIONS(4118),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4118),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4118),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4118),
    [anon_sym_IBOutlet] = ACTIONS(4118),
    [anon_sym_IBInspectable] = ACTIONS(4118),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4118),
    [anon_sym_NS_INLINE] = ACTIONS(4118),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4118),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4118),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4118),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4118),
    [anon_sym_const] = ACTIONS(4118),
    [anon_sym_constexpr] = ACTIONS(4118),
    [anon_sym_volatile] = ACTIONS(4118),
    [anon_sym_restrict] = ACTIONS(4118),
    [anon_sym___restrict__] = ACTIONS(4118),
    [anon_sym__Atomic] = ACTIONS(4118),
    [anon_sym__Noreturn] = ACTIONS(4118),
    [anon_sym__Nonnull] = ACTIONS(4118),
    [anon_sym_nullable] = ACTIONS(4118),
    [anon_sym__Complex] = ACTIONS(4118),
    [anon_sym__Nullable] = ACTIONS(4118),
    [anon_sym__Nullable_result] = ACTIONS(4118),
    [anon_sym__Null_unspecified] = ACTIONS(4118),
    [anon_sym___autoreleasing] = ACTIONS(4118),
    [anon_sym___block] = ACTIONS(4118),
    [anon_sym___bridge] = ACTIONS(4118),
    [anon_sym___bridge_retained] = ACTIONS(4118),
    [anon_sym___bridge_transfer] = ACTIONS(4118),
    [anon_sym___complex] = ACTIONS(4118),
    [anon_sym___const] = ACTIONS(4118),
    [anon_sym___imag] = ACTIONS(4118),
    [anon_sym___kindof] = ACTIONS(4118),
    [anon_sym___nonnull] = ACTIONS(4118),
    [anon_sym___nullable] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4118),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4118),
    [anon_sym___real] = ACTIONS(4118),
    [anon_sym___strong] = ACTIONS(4118),
    [anon_sym___unsafe_unretained] = ACTIONS(4118),
    [anon_sym___unused] = ACTIONS(4118),
    [anon_sym___weak] = ACTIONS(4118),
    [anon_sym_alignas] = ACTIONS(4118),
    [anon_sym__Alignas] = ACTIONS(4118),
    [sym_primitive_type] = ACTIONS(4118),
    [anon_sym_enum] = ACTIONS(4118),
    [anon_sym_struct] = ACTIONS(4118),
    [anon_sym_union] = ACTIONS(4118),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4118),
    [anon_sym___typeof] = ACTIONS(4118),
    [anon_sym_typeof] = ACTIONS(4118),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4118),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4118),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4118),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4118),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4118),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4118),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4118),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_API_AVAILABLE] = ACTIONS(4118),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_API_DEPRECATED] = ACTIONS(4118),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4118),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4118),
    [anon_sym___deprecated_msg] = ACTIONS(4118),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4118),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4118),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4118),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4118),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4118),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4118),
    [anon_sym_BOOL] = ACTIONS(4118),
    [anon_sym_IMP] = ACTIONS(4118),
    [anon_sym_SEL] = ACTIONS(4118),
    [anon_sym_Class] = ACTIONS(4118),
    [anon_sym_id] = ACTIONS(4118),
  },
  [STATE(2750)] = {
    [sym_parameter_list] = STATE(2552),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6742),
    [anon_sym_LPAREN2] = ACTIONS(6630),
    [anon_sym_SEMI] = ACTIONS(6742),
    [anon_sym___extension__] = ACTIONS(6740),
    [anon_sym_extern] = ACTIONS(6740),
    [anon_sym___attribute__] = ACTIONS(6740),
    [anon_sym___attribute] = ACTIONS(6740),
    [anon_sym_noreturn] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6632),
    [anon_sym___declspec] = ACTIONS(6740),
    [anon_sym_LBRACE] = ACTIONS(6742),
    [anon_sym_signed] = ACTIONS(6740),
    [anon_sym_unsigned] = ACTIONS(6740),
    [anon_sym_long] = ACTIONS(6740),
    [anon_sym_short] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6742),
    [anon_sym_ATautoreleasepool] = ACTIONS(6742),
    [anon_sym_auto] = ACTIONS(6740),
    [anon_sym_register] = ACTIONS(6740),
    [anon_sym_inline] = ACTIONS(6740),
    [anon_sym___inline] = ACTIONS(6740),
    [anon_sym___inline__] = ACTIONS(6740),
    [anon_sym___forceinline] = ACTIONS(6740),
    [anon_sym_thread_local] = ACTIONS(6740),
    [anon_sym___thread] = ACTIONS(6740),
    [anon_sym_CG_EXTERN] = ACTIONS(6740),
    [anon_sym_CG_INLINE] = ACTIONS(6740),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6740),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6740),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6740),
    [anon_sym_IBOutlet] = ACTIONS(6740),
    [anon_sym_IBInspectable] = ACTIONS(6740),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6740),
    [anon_sym_NS_INLINE] = ACTIONS(6740),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6740),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6740),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6740),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6740),
    [anon_sym_const] = ACTIONS(6740),
    [anon_sym_constexpr] = ACTIONS(6740),
    [anon_sym_volatile] = ACTIONS(6740),
    [anon_sym_restrict] = ACTIONS(6740),
    [anon_sym___restrict__] = ACTIONS(6740),
    [anon_sym__Atomic] = ACTIONS(6740),
    [anon_sym__Noreturn] = ACTIONS(6740),
    [anon_sym__Nonnull] = ACTIONS(6740),
    [anon_sym_nullable] = ACTIONS(6740),
    [anon_sym__Complex] = ACTIONS(6740),
    [anon_sym__Nullable] = ACTIONS(6740),
    [anon_sym__Nullable_result] = ACTIONS(6740),
    [anon_sym__Null_unspecified] = ACTIONS(6740),
    [anon_sym___autoreleasing] = ACTIONS(6740),
    [anon_sym___block] = ACTIONS(6740),
    [anon_sym___bridge] = ACTIONS(6740),
    [anon_sym___bridge_retained] = ACTIONS(6740),
    [anon_sym___bridge_transfer] = ACTIONS(6740),
    [anon_sym___complex] = ACTIONS(6740),
    [anon_sym___const] = ACTIONS(6740),
    [anon_sym___imag] = ACTIONS(6740),
    [anon_sym___kindof] = ACTIONS(6740),
    [anon_sym___nonnull] = ACTIONS(6740),
    [anon_sym___nullable] = ACTIONS(6740),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6740),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6740),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6740),
    [anon_sym___real] = ACTIONS(6740),
    [anon_sym___strong] = ACTIONS(6740),
    [anon_sym___unsafe_unretained] = ACTIONS(6740),
    [anon_sym___unused] = ACTIONS(6740),
    [anon_sym___weak] = ACTIONS(6740),
    [anon_sym_alignas] = ACTIONS(6740),
    [anon_sym__Alignas] = ACTIONS(6740),
    [sym_primitive_type] = ACTIONS(6740),
    [anon_sym_enum] = ACTIONS(6740),
    [anon_sym_struct] = ACTIONS(6740),
    [anon_sym_union] = ACTIONS(6740),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6740),
    [anon_sym___typeof] = ACTIONS(6740),
    [anon_sym_typeof] = ACTIONS(6740),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6740),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6740),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6740),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6740),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6740),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6740),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6740),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6740),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6740),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6740),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6740),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6740),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6740),
    [anon_sym_API_AVAILABLE] = ACTIONS(6740),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6740),
    [anon_sym_API_DEPRECATED] = ACTIONS(6740),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6740),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6740),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6740),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6740),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6740),
    [anon_sym___deprecated_msg] = ACTIONS(6740),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6740),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6740),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6740),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6740),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6740),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6740),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6740),
    [anon_sym_BOOL] = ACTIONS(6740),
    [anon_sym_IMP] = ACTIONS(6740),
    [anon_sym_SEL] = ACTIONS(6740),
    [anon_sym_Class] = ACTIONS(6740),
    [anon_sym_id] = ACTIONS(6740),
  },
  [STATE(2751)] = {
    [sym_identifier] = ACTIONS(6595),
    [aux_sym_preproc_def_token1] = ACTIONS(6595),
    [aux_sym_preproc_if_token1] = ACTIONS(6595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6595),
    [sym_preproc_directive] = ACTIONS(6595),
    [anon_sym___extension__] = ACTIONS(6595),
    [anon_sym_extern] = ACTIONS(6595),
    [anon_sym___attribute__] = ACTIONS(6595),
    [anon_sym___attribute] = ACTIONS(6595),
    [anon_sym_noreturn] = ACTIONS(6595),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym___declspec] = ACTIONS(6595),
    [anon_sym_RBRACE] = ACTIONS(6597),
    [anon_sym_signed] = ACTIONS(6595),
    [anon_sym_unsigned] = ACTIONS(6595),
    [anon_sym_long] = ACTIONS(6595),
    [anon_sym_short] = ACTIONS(6595),
    [anon_sym_static] = ACTIONS(6595),
    [anon_sym_auto] = ACTIONS(6595),
    [anon_sym_register] = ACTIONS(6595),
    [anon_sym_inline] = ACTIONS(6595),
    [anon_sym___inline] = ACTIONS(6595),
    [anon_sym___inline__] = ACTIONS(6595),
    [anon_sym___forceinline] = ACTIONS(6595),
    [anon_sym_thread_local] = ACTIONS(6595),
    [anon_sym___thread] = ACTIONS(6595),
    [anon_sym_CG_EXTERN] = ACTIONS(6595),
    [anon_sym_CG_INLINE] = ACTIONS(6595),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6595),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6595),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6595),
    [anon_sym_IBOutlet] = ACTIONS(6595),
    [anon_sym_IBInspectable] = ACTIONS(6595),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6595),
    [anon_sym_NS_INLINE] = ACTIONS(6595),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6595),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6595),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6595),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6595),
    [anon_sym_const] = ACTIONS(6595),
    [anon_sym_constexpr] = ACTIONS(6595),
    [anon_sym_volatile] = ACTIONS(6595),
    [anon_sym_restrict] = ACTIONS(6595),
    [anon_sym___restrict__] = ACTIONS(6595),
    [anon_sym__Atomic] = ACTIONS(6595),
    [anon_sym__Noreturn] = ACTIONS(6595),
    [anon_sym__Nonnull] = ACTIONS(6595),
    [anon_sym_nullable] = ACTIONS(6595),
    [anon_sym__Complex] = ACTIONS(6595),
    [anon_sym__Nullable] = ACTIONS(6595),
    [anon_sym__Nullable_result] = ACTIONS(6595),
    [anon_sym__Null_unspecified] = ACTIONS(6595),
    [anon_sym___autoreleasing] = ACTIONS(6595),
    [anon_sym___block] = ACTIONS(6595),
    [anon_sym___bridge] = ACTIONS(6595),
    [anon_sym___bridge_retained] = ACTIONS(6595),
    [anon_sym___bridge_transfer] = ACTIONS(6595),
    [anon_sym___complex] = ACTIONS(6595),
    [anon_sym___const] = ACTIONS(6595),
    [anon_sym___imag] = ACTIONS(6595),
    [anon_sym___kindof] = ACTIONS(6595),
    [anon_sym___nonnull] = ACTIONS(6595),
    [anon_sym___nullable] = ACTIONS(6595),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6595),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6595),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6595),
    [anon_sym___real] = ACTIONS(6595),
    [anon_sym___strong] = ACTIONS(6595),
    [anon_sym___unsafe_unretained] = ACTIONS(6595),
    [anon_sym___unused] = ACTIONS(6595),
    [anon_sym___weak] = ACTIONS(6595),
    [anon_sym_alignas] = ACTIONS(6595),
    [anon_sym__Alignas] = ACTIONS(6595),
    [sym_primitive_type] = ACTIONS(6595),
    [anon_sym_enum] = ACTIONS(6595),
    [anon_sym_struct] = ACTIONS(6595),
    [anon_sym_union] = ACTIONS(6595),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6595),
    [anon_sym___typeof] = ACTIONS(6595),
    [anon_sym_typeof] = ACTIONS(6595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6595),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6595),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6595),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6595),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6595),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6595),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6595),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6595),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6595),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_API_AVAILABLE] = ACTIONS(6595),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_API_DEPRECATED] = ACTIONS(6595),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6595),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6595),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6595),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6595),
    [anon_sym___deprecated_msg] = ACTIONS(6595),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6595),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6595),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6595),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6595),
    [anon_sym_BOOL] = ACTIONS(6595),
    [anon_sym_IMP] = ACTIONS(6595),
    [anon_sym_SEL] = ACTIONS(6595),
    [anon_sym_Class] = ACTIONS(6595),
    [anon_sym_id] = ACTIONS(6595),
    [anon_sym_ATdefs] = ACTIONS(6597),
  },
  [STATE(2752)] = {
    [sym_identifier] = ACTIONS(6555),
    [aux_sym_preproc_def_token1] = ACTIONS(6555),
    [aux_sym_preproc_if_token1] = ACTIONS(6555),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6555),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6555),
    [sym_preproc_directive] = ACTIONS(6555),
    [anon_sym___extension__] = ACTIONS(6555),
    [anon_sym_extern] = ACTIONS(6555),
    [anon_sym___attribute__] = ACTIONS(6555),
    [anon_sym___attribute] = ACTIONS(6555),
    [anon_sym_noreturn] = ACTIONS(6555),
    [anon_sym_LBRACK] = ACTIONS(6557),
    [anon_sym___declspec] = ACTIONS(6555),
    [anon_sym_RBRACE] = ACTIONS(6557),
    [anon_sym_signed] = ACTIONS(6555),
    [anon_sym_unsigned] = ACTIONS(6555),
    [anon_sym_long] = ACTIONS(6555),
    [anon_sym_short] = ACTIONS(6555),
    [anon_sym_static] = ACTIONS(6555),
    [anon_sym_auto] = ACTIONS(6555),
    [anon_sym_register] = ACTIONS(6555),
    [anon_sym_inline] = ACTIONS(6555),
    [anon_sym___inline] = ACTIONS(6555),
    [anon_sym___inline__] = ACTIONS(6555),
    [anon_sym___forceinline] = ACTIONS(6555),
    [anon_sym_thread_local] = ACTIONS(6555),
    [anon_sym___thread] = ACTIONS(6555),
    [anon_sym_CG_EXTERN] = ACTIONS(6555),
    [anon_sym_CG_INLINE] = ACTIONS(6555),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6555),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6555),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6555),
    [anon_sym_IBOutlet] = ACTIONS(6555),
    [anon_sym_IBInspectable] = ACTIONS(6555),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6555),
    [anon_sym_NS_INLINE] = ACTIONS(6555),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6555),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6555),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6555),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6555),
    [anon_sym_const] = ACTIONS(6555),
    [anon_sym_constexpr] = ACTIONS(6555),
    [anon_sym_volatile] = ACTIONS(6555),
    [anon_sym_restrict] = ACTIONS(6555),
    [anon_sym___restrict__] = ACTIONS(6555),
    [anon_sym__Atomic] = ACTIONS(6555),
    [anon_sym__Noreturn] = ACTIONS(6555),
    [anon_sym__Nonnull] = ACTIONS(6555),
    [anon_sym_nullable] = ACTIONS(6555),
    [anon_sym__Complex] = ACTIONS(6555),
    [anon_sym__Nullable] = ACTIONS(6555),
    [anon_sym__Nullable_result] = ACTIONS(6555),
    [anon_sym__Null_unspecified] = ACTIONS(6555),
    [anon_sym___autoreleasing] = ACTIONS(6555),
    [anon_sym___block] = ACTIONS(6555),
    [anon_sym___bridge] = ACTIONS(6555),
    [anon_sym___bridge_retained] = ACTIONS(6555),
    [anon_sym___bridge_transfer] = ACTIONS(6555),
    [anon_sym___complex] = ACTIONS(6555),
    [anon_sym___const] = ACTIONS(6555),
    [anon_sym___imag] = ACTIONS(6555),
    [anon_sym___kindof] = ACTIONS(6555),
    [anon_sym___nonnull] = ACTIONS(6555),
    [anon_sym___nullable] = ACTIONS(6555),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6555),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6555),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6555),
    [anon_sym___real] = ACTIONS(6555),
    [anon_sym___strong] = ACTIONS(6555),
    [anon_sym___unsafe_unretained] = ACTIONS(6555),
    [anon_sym___unused] = ACTIONS(6555),
    [anon_sym___weak] = ACTIONS(6555),
    [anon_sym_alignas] = ACTIONS(6555),
    [anon_sym__Alignas] = ACTIONS(6555),
    [sym_primitive_type] = ACTIONS(6555),
    [anon_sym_enum] = ACTIONS(6555),
    [anon_sym_struct] = ACTIONS(6555),
    [anon_sym_union] = ACTIONS(6555),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6555),
    [anon_sym___typeof] = ACTIONS(6555),
    [anon_sym_typeof] = ACTIONS(6555),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6555),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6555),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6555),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6555),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6555),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6555),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6555),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6555),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6555),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_API_AVAILABLE] = ACTIONS(6555),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_API_DEPRECATED] = ACTIONS(6555),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6555),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6555),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6555),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6555),
    [anon_sym___deprecated_msg] = ACTIONS(6555),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6555),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6555),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6555),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6555),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6555),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6555),
    [anon_sym_BOOL] = ACTIONS(6555),
    [anon_sym_IMP] = ACTIONS(6555),
    [anon_sym_SEL] = ACTIONS(6555),
    [anon_sym_Class] = ACTIONS(6555),
    [anon_sym_id] = ACTIONS(6555),
    [anon_sym_ATdefs] = ACTIONS(6557),
  },
  [STATE(2753)] = {
    [sym_parameter_list] = STATE(2552),
    [sym_identifier] = ACTIONS(6744),
    [anon_sym_COMMA] = ACTIONS(6746),
    [anon_sym_LPAREN2] = ACTIONS(6630),
    [anon_sym_SEMI] = ACTIONS(6746),
    [anon_sym___extension__] = ACTIONS(6744),
    [anon_sym_extern] = ACTIONS(6744),
    [anon_sym___attribute__] = ACTIONS(6744),
    [anon_sym___attribute] = ACTIONS(6744),
    [anon_sym_noreturn] = ACTIONS(6744),
    [anon_sym_LBRACK] = ACTIONS(6632),
    [anon_sym___declspec] = ACTIONS(6744),
    [anon_sym_LBRACE] = ACTIONS(6746),
    [anon_sym_signed] = ACTIONS(6744),
    [anon_sym_unsigned] = ACTIONS(6744),
    [anon_sym_long] = ACTIONS(6744),
    [anon_sym_short] = ACTIONS(6744),
    [anon_sym_static] = ACTIONS(6744),
    [anon_sym_EQ] = ACTIONS(6746),
    [anon_sym_ATautoreleasepool] = ACTIONS(6746),
    [anon_sym_auto] = ACTIONS(6744),
    [anon_sym_register] = ACTIONS(6744),
    [anon_sym_inline] = ACTIONS(6744),
    [anon_sym___inline] = ACTIONS(6744),
    [anon_sym___inline__] = ACTIONS(6744),
    [anon_sym___forceinline] = ACTIONS(6744),
    [anon_sym_thread_local] = ACTIONS(6744),
    [anon_sym___thread] = ACTIONS(6744),
    [anon_sym_CG_EXTERN] = ACTIONS(6744),
    [anon_sym_CG_INLINE] = ACTIONS(6744),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6744),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6744),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6744),
    [anon_sym_IBOutlet] = ACTIONS(6744),
    [anon_sym_IBInspectable] = ACTIONS(6744),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6744),
    [anon_sym_NS_INLINE] = ACTIONS(6744),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6744),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6744),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6744),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6744),
    [anon_sym_const] = ACTIONS(6744),
    [anon_sym_constexpr] = ACTIONS(6744),
    [anon_sym_volatile] = ACTIONS(6744),
    [anon_sym_restrict] = ACTIONS(6744),
    [anon_sym___restrict__] = ACTIONS(6744),
    [anon_sym__Atomic] = ACTIONS(6744),
    [anon_sym__Noreturn] = ACTIONS(6744),
    [anon_sym__Nonnull] = ACTIONS(6744),
    [anon_sym_nullable] = ACTIONS(6744),
    [anon_sym__Complex] = ACTIONS(6744),
    [anon_sym__Nullable] = ACTIONS(6744),
    [anon_sym__Nullable_result] = ACTIONS(6744),
    [anon_sym__Null_unspecified] = ACTIONS(6744),
    [anon_sym___autoreleasing] = ACTIONS(6744),
    [anon_sym___block] = ACTIONS(6744),
    [anon_sym___bridge] = ACTIONS(6744),
    [anon_sym___bridge_retained] = ACTIONS(6744),
    [anon_sym___bridge_transfer] = ACTIONS(6744),
    [anon_sym___complex] = ACTIONS(6744),
    [anon_sym___const] = ACTIONS(6744),
    [anon_sym___imag] = ACTIONS(6744),
    [anon_sym___kindof] = ACTIONS(6744),
    [anon_sym___nonnull] = ACTIONS(6744),
    [anon_sym___nullable] = ACTIONS(6744),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6744),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6744),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6744),
    [anon_sym___real] = ACTIONS(6744),
    [anon_sym___strong] = ACTIONS(6744),
    [anon_sym___unsafe_unretained] = ACTIONS(6744),
    [anon_sym___unused] = ACTIONS(6744),
    [anon_sym___weak] = ACTIONS(6744),
    [anon_sym_alignas] = ACTIONS(6744),
    [anon_sym__Alignas] = ACTIONS(6744),
    [sym_primitive_type] = ACTIONS(6744),
    [anon_sym_enum] = ACTIONS(6744),
    [anon_sym_struct] = ACTIONS(6744),
    [anon_sym_union] = ACTIONS(6744),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6744),
    [anon_sym___typeof] = ACTIONS(6744),
    [anon_sym_typeof] = ACTIONS(6744),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6744),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6744),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6744),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6744),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6744),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6744),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6744),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6744),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6744),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6744),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6744),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6744),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6744),
    [anon_sym_API_AVAILABLE] = ACTIONS(6744),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6744),
    [anon_sym_API_DEPRECATED] = ACTIONS(6744),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6744),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6744),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6744),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6744),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6744),
    [anon_sym___deprecated_msg] = ACTIONS(6744),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6744),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6744),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6744),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6744),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6744),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6744),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6744),
    [anon_sym_BOOL] = ACTIONS(6744),
    [anon_sym_IMP] = ACTIONS(6744),
    [anon_sym_SEL] = ACTIONS(6744),
    [anon_sym_Class] = ACTIONS(6744),
    [anon_sym_id] = ACTIONS(6744),
  },
  [STATE(2754)] = {
    [sym_parameter_list] = STATE(2552),
    [sym_identifier] = ACTIONS(6748),
    [anon_sym_COMMA] = ACTIONS(6750),
    [anon_sym_LPAREN2] = ACTIONS(6630),
    [anon_sym_SEMI] = ACTIONS(6750),
    [anon_sym___extension__] = ACTIONS(6748),
    [anon_sym_extern] = ACTIONS(6748),
    [anon_sym___attribute__] = ACTIONS(6748),
    [anon_sym___attribute] = ACTIONS(6748),
    [anon_sym_noreturn] = ACTIONS(6748),
    [anon_sym_LBRACK] = ACTIONS(6632),
    [anon_sym___declspec] = ACTIONS(6748),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_signed] = ACTIONS(6748),
    [anon_sym_unsigned] = ACTIONS(6748),
    [anon_sym_long] = ACTIONS(6748),
    [anon_sym_short] = ACTIONS(6748),
    [anon_sym_static] = ACTIONS(6748),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_ATautoreleasepool] = ACTIONS(6750),
    [anon_sym_auto] = ACTIONS(6748),
    [anon_sym_register] = ACTIONS(6748),
    [anon_sym_inline] = ACTIONS(6748),
    [anon_sym___inline] = ACTIONS(6748),
    [anon_sym___inline__] = ACTIONS(6748),
    [anon_sym___forceinline] = ACTIONS(6748),
    [anon_sym_thread_local] = ACTIONS(6748),
    [anon_sym___thread] = ACTIONS(6748),
    [anon_sym_CG_EXTERN] = ACTIONS(6748),
    [anon_sym_CG_INLINE] = ACTIONS(6748),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6748),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6748),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6748),
    [anon_sym_IBOutlet] = ACTIONS(6748),
    [anon_sym_IBInspectable] = ACTIONS(6748),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6748),
    [anon_sym_NS_INLINE] = ACTIONS(6748),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6748),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6748),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6748),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6748),
    [anon_sym_const] = ACTIONS(6748),
    [anon_sym_constexpr] = ACTIONS(6748),
    [anon_sym_volatile] = ACTIONS(6748),
    [anon_sym_restrict] = ACTIONS(6748),
    [anon_sym___restrict__] = ACTIONS(6748),
    [anon_sym__Atomic] = ACTIONS(6748),
    [anon_sym__Noreturn] = ACTIONS(6748),
    [anon_sym__Nonnull] = ACTIONS(6748),
    [anon_sym_nullable] = ACTIONS(6748),
    [anon_sym__Complex] = ACTIONS(6748),
    [anon_sym__Nullable] = ACTIONS(6748),
    [anon_sym__Nullable_result] = ACTIONS(6748),
    [anon_sym__Null_unspecified] = ACTIONS(6748),
    [anon_sym___autoreleasing] = ACTIONS(6748),
    [anon_sym___block] = ACTIONS(6748),
    [anon_sym___bridge] = ACTIONS(6748),
    [anon_sym___bridge_retained] = ACTIONS(6748),
    [anon_sym___bridge_transfer] = ACTIONS(6748),
    [anon_sym___complex] = ACTIONS(6748),
    [anon_sym___const] = ACTIONS(6748),
    [anon_sym___imag] = ACTIONS(6748),
    [anon_sym___kindof] = ACTIONS(6748),
    [anon_sym___nonnull] = ACTIONS(6748),
    [anon_sym___nullable] = ACTIONS(6748),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6748),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6748),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6748),
    [anon_sym___real] = ACTIONS(6748),
    [anon_sym___strong] = ACTIONS(6748),
    [anon_sym___unsafe_unretained] = ACTIONS(6748),
    [anon_sym___unused] = ACTIONS(6748),
    [anon_sym___weak] = ACTIONS(6748),
    [anon_sym_alignas] = ACTIONS(6748),
    [anon_sym__Alignas] = ACTIONS(6748),
    [sym_primitive_type] = ACTIONS(6748),
    [anon_sym_enum] = ACTIONS(6748),
    [anon_sym_struct] = ACTIONS(6748),
    [anon_sym_union] = ACTIONS(6748),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6748),
    [anon_sym___typeof] = ACTIONS(6748),
    [anon_sym_typeof] = ACTIONS(6748),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6748),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6748),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6748),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6748),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6748),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6748),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6748),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6748),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6748),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6748),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6748),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6748),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6748),
    [anon_sym_API_AVAILABLE] = ACTIONS(6748),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6748),
    [anon_sym_API_DEPRECATED] = ACTIONS(6748),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6748),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6748),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6748),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6748),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6748),
    [anon_sym___deprecated_msg] = ACTIONS(6748),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6748),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6748),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6748),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6748),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6748),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6748),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6748),
    [anon_sym_BOOL] = ACTIONS(6748),
    [anon_sym_IMP] = ACTIONS(6748),
    [anon_sym_SEL] = ACTIONS(6748),
    [anon_sym_Class] = ACTIONS(6748),
    [anon_sym_id] = ACTIONS(6748),
  },
  [STATE(2755)] = {
    [sym_identifier] = ACTIONS(6610),
    [aux_sym_preproc_def_token1] = ACTIONS(6610),
    [aux_sym_preproc_if_token1] = ACTIONS(6610),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6610),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6610),
    [sym_preproc_directive] = ACTIONS(6610),
    [anon_sym___extension__] = ACTIONS(6610),
    [anon_sym_extern] = ACTIONS(6610),
    [anon_sym___attribute__] = ACTIONS(6610),
    [anon_sym___attribute] = ACTIONS(6610),
    [anon_sym_noreturn] = ACTIONS(6610),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym___declspec] = ACTIONS(6610),
    [anon_sym_RBRACE] = ACTIONS(6612),
    [anon_sym_signed] = ACTIONS(6610),
    [anon_sym_unsigned] = ACTIONS(6610),
    [anon_sym_long] = ACTIONS(6610),
    [anon_sym_short] = ACTIONS(6610),
    [anon_sym_static] = ACTIONS(6610),
    [anon_sym_auto] = ACTIONS(6610),
    [anon_sym_register] = ACTIONS(6610),
    [anon_sym_inline] = ACTIONS(6610),
    [anon_sym___inline] = ACTIONS(6610),
    [anon_sym___inline__] = ACTIONS(6610),
    [anon_sym___forceinline] = ACTIONS(6610),
    [anon_sym_thread_local] = ACTIONS(6610),
    [anon_sym___thread] = ACTIONS(6610),
    [anon_sym_CG_EXTERN] = ACTIONS(6610),
    [anon_sym_CG_INLINE] = ACTIONS(6610),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6610),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6610),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6610),
    [anon_sym_IBOutlet] = ACTIONS(6610),
    [anon_sym_IBInspectable] = ACTIONS(6610),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6610),
    [anon_sym_NS_INLINE] = ACTIONS(6610),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6610),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6610),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6610),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6610),
    [anon_sym_const] = ACTIONS(6610),
    [anon_sym_constexpr] = ACTIONS(6610),
    [anon_sym_volatile] = ACTIONS(6610),
    [anon_sym_restrict] = ACTIONS(6610),
    [anon_sym___restrict__] = ACTIONS(6610),
    [anon_sym__Atomic] = ACTIONS(6610),
    [anon_sym__Noreturn] = ACTIONS(6610),
    [anon_sym__Nonnull] = ACTIONS(6610),
    [anon_sym_nullable] = ACTIONS(6610),
    [anon_sym__Complex] = ACTIONS(6610),
    [anon_sym__Nullable] = ACTIONS(6610),
    [anon_sym__Nullable_result] = ACTIONS(6610),
    [anon_sym__Null_unspecified] = ACTIONS(6610),
    [anon_sym___autoreleasing] = ACTIONS(6610),
    [anon_sym___block] = ACTIONS(6610),
    [anon_sym___bridge] = ACTIONS(6610),
    [anon_sym___bridge_retained] = ACTIONS(6610),
    [anon_sym___bridge_transfer] = ACTIONS(6610),
    [anon_sym___complex] = ACTIONS(6610),
    [anon_sym___const] = ACTIONS(6610),
    [anon_sym___imag] = ACTIONS(6610),
    [anon_sym___kindof] = ACTIONS(6610),
    [anon_sym___nonnull] = ACTIONS(6610),
    [anon_sym___nullable] = ACTIONS(6610),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6610),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6610),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6610),
    [anon_sym___real] = ACTIONS(6610),
    [anon_sym___strong] = ACTIONS(6610),
    [anon_sym___unsafe_unretained] = ACTIONS(6610),
    [anon_sym___unused] = ACTIONS(6610),
    [anon_sym___weak] = ACTIONS(6610),
    [anon_sym_alignas] = ACTIONS(6610),
    [anon_sym__Alignas] = ACTIONS(6610),
    [sym_primitive_type] = ACTIONS(6610),
    [anon_sym_enum] = ACTIONS(6610),
    [anon_sym_struct] = ACTIONS(6610),
    [anon_sym_union] = ACTIONS(6610),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6610),
    [anon_sym___typeof] = ACTIONS(6610),
    [anon_sym_typeof] = ACTIONS(6610),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6610),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6610),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6610),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6610),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6610),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6610),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6610),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6610),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6610),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_API_AVAILABLE] = ACTIONS(6610),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_API_DEPRECATED] = ACTIONS(6610),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6610),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6610),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6610),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6610),
    [anon_sym___deprecated_msg] = ACTIONS(6610),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6610),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6610),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6610),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6610),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6610),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6610),
    [anon_sym_BOOL] = ACTIONS(6610),
    [anon_sym_IMP] = ACTIONS(6610),
    [anon_sym_SEL] = ACTIONS(6610),
    [anon_sym_Class] = ACTIONS(6610),
    [anon_sym_id] = ACTIONS(6610),
    [anon_sym_ATdefs] = ACTIONS(6612),
  },
  [STATE(2756)] = {
    [sym_identifier] = ACTIONS(6603),
    [aux_sym_preproc_def_token1] = ACTIONS(6603),
    [aux_sym_preproc_if_token1] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6603),
    [sym_preproc_directive] = ACTIONS(6603),
    [anon_sym___extension__] = ACTIONS(6603),
    [anon_sym_extern] = ACTIONS(6603),
    [anon_sym___attribute__] = ACTIONS(6603),
    [anon_sym___attribute] = ACTIONS(6603),
    [anon_sym_noreturn] = ACTIONS(6603),
    [anon_sym_LBRACK] = ACTIONS(6605),
    [anon_sym___declspec] = ACTIONS(6603),
    [anon_sym_RBRACE] = ACTIONS(6605),
    [anon_sym_signed] = ACTIONS(6603),
    [anon_sym_unsigned] = ACTIONS(6603),
    [anon_sym_long] = ACTIONS(6603),
    [anon_sym_short] = ACTIONS(6603),
    [anon_sym_static] = ACTIONS(6603),
    [anon_sym_auto] = ACTIONS(6603),
    [anon_sym_register] = ACTIONS(6603),
    [anon_sym_inline] = ACTIONS(6603),
    [anon_sym___inline] = ACTIONS(6603),
    [anon_sym___inline__] = ACTIONS(6603),
    [anon_sym___forceinline] = ACTIONS(6603),
    [anon_sym_thread_local] = ACTIONS(6603),
    [anon_sym___thread] = ACTIONS(6603),
    [anon_sym_CG_EXTERN] = ACTIONS(6603),
    [anon_sym_CG_INLINE] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6603),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6603),
    [anon_sym_IBOutlet] = ACTIONS(6603),
    [anon_sym_IBInspectable] = ACTIONS(6603),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6603),
    [anon_sym_NS_INLINE] = ACTIONS(6603),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6603),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6603),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6603),
    [anon_sym_const] = ACTIONS(6603),
    [anon_sym_constexpr] = ACTIONS(6603),
    [anon_sym_volatile] = ACTIONS(6603),
    [anon_sym_restrict] = ACTIONS(6603),
    [anon_sym___restrict__] = ACTIONS(6603),
    [anon_sym__Atomic] = ACTIONS(6603),
    [anon_sym__Noreturn] = ACTIONS(6603),
    [anon_sym__Nonnull] = ACTIONS(6603),
    [anon_sym_nullable] = ACTIONS(6603),
    [anon_sym__Complex] = ACTIONS(6603),
    [anon_sym__Nullable] = ACTIONS(6603),
    [anon_sym__Nullable_result] = ACTIONS(6603),
    [anon_sym__Null_unspecified] = ACTIONS(6603),
    [anon_sym___autoreleasing] = ACTIONS(6603),
    [anon_sym___block] = ACTIONS(6603),
    [anon_sym___bridge] = ACTIONS(6603),
    [anon_sym___bridge_retained] = ACTIONS(6603),
    [anon_sym___bridge_transfer] = ACTIONS(6603),
    [anon_sym___complex] = ACTIONS(6603),
    [anon_sym___const] = ACTIONS(6603),
    [anon_sym___imag] = ACTIONS(6603),
    [anon_sym___kindof] = ACTIONS(6603),
    [anon_sym___nonnull] = ACTIONS(6603),
    [anon_sym___nullable] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6603),
    [anon_sym___real] = ACTIONS(6603),
    [anon_sym___strong] = ACTIONS(6603),
    [anon_sym___unsafe_unretained] = ACTIONS(6603),
    [anon_sym___unused] = ACTIONS(6603),
    [anon_sym___weak] = ACTIONS(6603),
    [anon_sym_alignas] = ACTIONS(6603),
    [anon_sym__Alignas] = ACTIONS(6603),
    [sym_primitive_type] = ACTIONS(6603),
    [anon_sym_enum] = ACTIONS(6603),
    [anon_sym_struct] = ACTIONS(6603),
    [anon_sym_union] = ACTIONS(6603),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6603),
    [anon_sym___typeof] = ACTIONS(6603),
    [anon_sym_typeof] = ACTIONS(6603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6603),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6603),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6603),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_API_AVAILABLE] = ACTIONS(6603),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_API_DEPRECATED] = ACTIONS(6603),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym___deprecated_msg] = ACTIONS(6603),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6603),
    [anon_sym_BOOL] = ACTIONS(6603),
    [anon_sym_IMP] = ACTIONS(6603),
    [anon_sym_SEL] = ACTIONS(6603),
    [anon_sym_Class] = ACTIONS(6603),
    [anon_sym_id] = ACTIONS(6603),
    [anon_sym_ATdefs] = ACTIONS(6605),
  },
  [STATE(2757)] = {
    [sym_identifier] = ACTIONS(6614),
    [aux_sym_preproc_def_token1] = ACTIONS(6614),
    [aux_sym_preproc_if_token1] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6614),
    [sym_preproc_directive] = ACTIONS(6614),
    [anon_sym___extension__] = ACTIONS(6614),
    [anon_sym_extern] = ACTIONS(6614),
    [anon_sym___attribute__] = ACTIONS(6614),
    [anon_sym___attribute] = ACTIONS(6614),
    [anon_sym_noreturn] = ACTIONS(6614),
    [anon_sym_LBRACK] = ACTIONS(6616),
    [anon_sym___declspec] = ACTIONS(6614),
    [anon_sym_RBRACE] = ACTIONS(6616),
    [anon_sym_signed] = ACTIONS(6614),
    [anon_sym_unsigned] = ACTIONS(6614),
    [anon_sym_long] = ACTIONS(6614),
    [anon_sym_short] = ACTIONS(6614),
    [anon_sym_static] = ACTIONS(6614),
    [anon_sym_auto] = ACTIONS(6614),
    [anon_sym_register] = ACTIONS(6614),
    [anon_sym_inline] = ACTIONS(6614),
    [anon_sym___inline] = ACTIONS(6614),
    [anon_sym___inline__] = ACTIONS(6614),
    [anon_sym___forceinline] = ACTIONS(6614),
    [anon_sym_thread_local] = ACTIONS(6614),
    [anon_sym___thread] = ACTIONS(6614),
    [anon_sym_CG_EXTERN] = ACTIONS(6614),
    [anon_sym_CG_INLINE] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6614),
    [anon_sym_IBOutlet] = ACTIONS(6614),
    [anon_sym_IBInspectable] = ACTIONS(6614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6614),
    [anon_sym_NS_INLINE] = ACTIONS(6614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6614),
    [anon_sym_const] = ACTIONS(6614),
    [anon_sym_constexpr] = ACTIONS(6614),
    [anon_sym_volatile] = ACTIONS(6614),
    [anon_sym_restrict] = ACTIONS(6614),
    [anon_sym___restrict__] = ACTIONS(6614),
    [anon_sym__Atomic] = ACTIONS(6614),
    [anon_sym__Noreturn] = ACTIONS(6614),
    [anon_sym__Nonnull] = ACTIONS(6614),
    [anon_sym_nullable] = ACTIONS(6614),
    [anon_sym__Complex] = ACTIONS(6614),
    [anon_sym__Nullable] = ACTIONS(6614),
    [anon_sym__Nullable_result] = ACTIONS(6614),
    [anon_sym__Null_unspecified] = ACTIONS(6614),
    [anon_sym___autoreleasing] = ACTIONS(6614),
    [anon_sym___block] = ACTIONS(6614),
    [anon_sym___bridge] = ACTIONS(6614),
    [anon_sym___bridge_retained] = ACTIONS(6614),
    [anon_sym___bridge_transfer] = ACTIONS(6614),
    [anon_sym___complex] = ACTIONS(6614),
    [anon_sym___const] = ACTIONS(6614),
    [anon_sym___imag] = ACTIONS(6614),
    [anon_sym___kindof] = ACTIONS(6614),
    [anon_sym___nonnull] = ACTIONS(6614),
    [anon_sym___nullable] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6614),
    [anon_sym___real] = ACTIONS(6614),
    [anon_sym___strong] = ACTIONS(6614),
    [anon_sym___unsafe_unretained] = ACTIONS(6614),
    [anon_sym___unused] = ACTIONS(6614),
    [anon_sym___weak] = ACTIONS(6614),
    [anon_sym_alignas] = ACTIONS(6614),
    [anon_sym__Alignas] = ACTIONS(6614),
    [sym_primitive_type] = ACTIONS(6614),
    [anon_sym_enum] = ACTIONS(6614),
    [anon_sym_struct] = ACTIONS(6614),
    [anon_sym_union] = ACTIONS(6614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6614),
    [anon_sym___typeof] = ACTIONS(6614),
    [anon_sym_typeof] = ACTIONS(6614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_API_AVAILABLE] = ACTIONS(6614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_API_DEPRECATED] = ACTIONS(6614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym___deprecated_msg] = ACTIONS(6614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6614),
    [anon_sym_BOOL] = ACTIONS(6614),
    [anon_sym_IMP] = ACTIONS(6614),
    [anon_sym_SEL] = ACTIONS(6614),
    [anon_sym_Class] = ACTIONS(6614),
    [anon_sym_id] = ACTIONS(6614),
    [anon_sym_ATdefs] = ACTIONS(6616),
  },
  [STATE(2758)] = {
    [sym_identifier] = ACTIONS(6614),
    [aux_sym_preproc_def_token1] = ACTIONS(6614),
    [aux_sym_preproc_if_token1] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6614),
    [sym_preproc_directive] = ACTIONS(6614),
    [anon_sym___extension__] = ACTIONS(6614),
    [anon_sym_extern] = ACTIONS(6614),
    [anon_sym___attribute__] = ACTIONS(6614),
    [anon_sym___attribute] = ACTIONS(6614),
    [anon_sym_noreturn] = ACTIONS(6614),
    [anon_sym_LBRACK] = ACTIONS(6616),
    [anon_sym___declspec] = ACTIONS(6614),
    [anon_sym_RBRACE] = ACTIONS(6616),
    [anon_sym_signed] = ACTIONS(6614),
    [anon_sym_unsigned] = ACTIONS(6614),
    [anon_sym_long] = ACTIONS(6614),
    [anon_sym_short] = ACTIONS(6614),
    [anon_sym_static] = ACTIONS(6614),
    [anon_sym_auto] = ACTIONS(6614),
    [anon_sym_register] = ACTIONS(6614),
    [anon_sym_inline] = ACTIONS(6614),
    [anon_sym___inline] = ACTIONS(6614),
    [anon_sym___inline__] = ACTIONS(6614),
    [anon_sym___forceinline] = ACTIONS(6614),
    [anon_sym_thread_local] = ACTIONS(6614),
    [anon_sym___thread] = ACTIONS(6614),
    [anon_sym_CG_EXTERN] = ACTIONS(6614),
    [anon_sym_CG_INLINE] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6614),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6614),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6614),
    [anon_sym_IBOutlet] = ACTIONS(6614),
    [anon_sym_IBInspectable] = ACTIONS(6614),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6614),
    [anon_sym_NS_INLINE] = ACTIONS(6614),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6614),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6614),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6614),
    [anon_sym_const] = ACTIONS(6614),
    [anon_sym_constexpr] = ACTIONS(6614),
    [anon_sym_volatile] = ACTIONS(6614),
    [anon_sym_restrict] = ACTIONS(6614),
    [anon_sym___restrict__] = ACTIONS(6614),
    [anon_sym__Atomic] = ACTIONS(6614),
    [anon_sym__Noreturn] = ACTIONS(6614),
    [anon_sym__Nonnull] = ACTIONS(6614),
    [anon_sym_nullable] = ACTIONS(6614),
    [anon_sym__Complex] = ACTIONS(6614),
    [anon_sym__Nullable] = ACTIONS(6614),
    [anon_sym__Nullable_result] = ACTIONS(6614),
    [anon_sym__Null_unspecified] = ACTIONS(6614),
    [anon_sym___autoreleasing] = ACTIONS(6614),
    [anon_sym___block] = ACTIONS(6614),
    [anon_sym___bridge] = ACTIONS(6614),
    [anon_sym___bridge_retained] = ACTIONS(6614),
    [anon_sym___bridge_transfer] = ACTIONS(6614),
    [anon_sym___complex] = ACTIONS(6614),
    [anon_sym___const] = ACTIONS(6614),
    [anon_sym___imag] = ACTIONS(6614),
    [anon_sym___kindof] = ACTIONS(6614),
    [anon_sym___nonnull] = ACTIONS(6614),
    [anon_sym___nullable] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6614),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6614),
    [anon_sym___real] = ACTIONS(6614),
    [anon_sym___strong] = ACTIONS(6614),
    [anon_sym___unsafe_unretained] = ACTIONS(6614),
    [anon_sym___unused] = ACTIONS(6614),
    [anon_sym___weak] = ACTIONS(6614),
    [anon_sym_alignas] = ACTIONS(6614),
    [anon_sym__Alignas] = ACTIONS(6614),
    [sym_primitive_type] = ACTIONS(6614),
    [anon_sym_enum] = ACTIONS(6614),
    [anon_sym_struct] = ACTIONS(6614),
    [anon_sym_union] = ACTIONS(6614),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6614),
    [anon_sym___typeof] = ACTIONS(6614),
    [anon_sym_typeof] = ACTIONS(6614),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6614),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6614),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6614),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6614),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6614),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_API_AVAILABLE] = ACTIONS(6614),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_API_DEPRECATED] = ACTIONS(6614),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6614),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6614),
    [anon_sym___deprecated_msg] = ACTIONS(6614),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6614),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6614),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6614),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6614),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6614),
    [anon_sym_BOOL] = ACTIONS(6614),
    [anon_sym_IMP] = ACTIONS(6614),
    [anon_sym_SEL] = ACTIONS(6614),
    [anon_sym_Class] = ACTIONS(6614),
    [anon_sym_id] = ACTIONS(6614),
    [anon_sym_ATdefs] = ACTIONS(6616),
  },
  [STATE(2759)] = {
    [sym_identifier] = ACTIONS(6603),
    [aux_sym_preproc_def_token1] = ACTIONS(6603),
    [aux_sym_preproc_if_token1] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6603),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6603),
    [sym_preproc_directive] = ACTIONS(6603),
    [anon_sym___extension__] = ACTIONS(6603),
    [anon_sym_extern] = ACTIONS(6603),
    [anon_sym___attribute__] = ACTIONS(6603),
    [anon_sym___attribute] = ACTIONS(6603),
    [anon_sym_noreturn] = ACTIONS(6603),
    [anon_sym_LBRACK] = ACTIONS(6605),
    [anon_sym___declspec] = ACTIONS(6603),
    [anon_sym_RBRACE] = ACTIONS(6605),
    [anon_sym_signed] = ACTIONS(6603),
    [anon_sym_unsigned] = ACTIONS(6603),
    [anon_sym_long] = ACTIONS(6603),
    [anon_sym_short] = ACTIONS(6603),
    [anon_sym_static] = ACTIONS(6603),
    [anon_sym_auto] = ACTIONS(6603),
    [anon_sym_register] = ACTIONS(6603),
    [anon_sym_inline] = ACTIONS(6603),
    [anon_sym___inline] = ACTIONS(6603),
    [anon_sym___inline__] = ACTIONS(6603),
    [anon_sym___forceinline] = ACTIONS(6603),
    [anon_sym_thread_local] = ACTIONS(6603),
    [anon_sym___thread] = ACTIONS(6603),
    [anon_sym_CG_EXTERN] = ACTIONS(6603),
    [anon_sym_CG_INLINE] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6603),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6603),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6603),
    [anon_sym_IBOutlet] = ACTIONS(6603),
    [anon_sym_IBInspectable] = ACTIONS(6603),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6603),
    [anon_sym_NS_INLINE] = ACTIONS(6603),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6603),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6603),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6603),
    [anon_sym_const] = ACTIONS(6603),
    [anon_sym_constexpr] = ACTIONS(6603),
    [anon_sym_volatile] = ACTIONS(6603),
    [anon_sym_restrict] = ACTIONS(6603),
    [anon_sym___restrict__] = ACTIONS(6603),
    [anon_sym__Atomic] = ACTIONS(6603),
    [anon_sym__Noreturn] = ACTIONS(6603),
    [anon_sym__Nonnull] = ACTIONS(6603),
    [anon_sym_nullable] = ACTIONS(6603),
    [anon_sym__Complex] = ACTIONS(6603),
    [anon_sym__Nullable] = ACTIONS(6603),
    [anon_sym__Nullable_result] = ACTIONS(6603),
    [anon_sym__Null_unspecified] = ACTIONS(6603),
    [anon_sym___autoreleasing] = ACTIONS(6603),
    [anon_sym___block] = ACTIONS(6603),
    [anon_sym___bridge] = ACTIONS(6603),
    [anon_sym___bridge_retained] = ACTIONS(6603),
    [anon_sym___bridge_transfer] = ACTIONS(6603),
    [anon_sym___complex] = ACTIONS(6603),
    [anon_sym___const] = ACTIONS(6603),
    [anon_sym___imag] = ACTIONS(6603),
    [anon_sym___kindof] = ACTIONS(6603),
    [anon_sym___nonnull] = ACTIONS(6603),
    [anon_sym___nullable] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6603),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6603),
    [anon_sym___real] = ACTIONS(6603),
    [anon_sym___strong] = ACTIONS(6603),
    [anon_sym___unsafe_unretained] = ACTIONS(6603),
    [anon_sym___unused] = ACTIONS(6603),
    [anon_sym___weak] = ACTIONS(6603),
    [anon_sym_alignas] = ACTIONS(6603),
    [anon_sym__Alignas] = ACTIONS(6603),
    [sym_primitive_type] = ACTIONS(6603),
    [anon_sym_enum] = ACTIONS(6603),
    [anon_sym_struct] = ACTIONS(6603),
    [anon_sym_union] = ACTIONS(6603),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6603),
    [anon_sym___typeof] = ACTIONS(6603),
    [anon_sym_typeof] = ACTIONS(6603),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6603),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6603),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6603),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6603),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6603),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_API_AVAILABLE] = ACTIONS(6603),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_API_DEPRECATED] = ACTIONS(6603),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6603),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6603),
    [anon_sym___deprecated_msg] = ACTIONS(6603),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6603),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6603),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6603),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6603),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6603),
    [anon_sym_BOOL] = ACTIONS(6603),
    [anon_sym_IMP] = ACTIONS(6603),
    [anon_sym_SEL] = ACTIONS(6603),
    [anon_sym_Class] = ACTIONS(6603),
    [anon_sym_id] = ACTIONS(6603),
    [anon_sym_ATdefs] = ACTIONS(6605),
  },
  [STATE(2760)] = {
    [sym_identifier] = ACTIONS(6595),
    [aux_sym_preproc_def_token1] = ACTIONS(6595),
    [aux_sym_preproc_if_token1] = ACTIONS(6595),
    [aux_sym_preproc_if_token2] = ACTIONS(6595),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6595),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6595),
    [sym_preproc_directive] = ACTIONS(6595),
    [anon_sym___extension__] = ACTIONS(6595),
    [anon_sym_extern] = ACTIONS(6595),
    [anon_sym___attribute__] = ACTIONS(6595),
    [anon_sym___attribute] = ACTIONS(6595),
    [anon_sym_noreturn] = ACTIONS(6595),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym___declspec] = ACTIONS(6595),
    [anon_sym_signed] = ACTIONS(6595),
    [anon_sym_unsigned] = ACTIONS(6595),
    [anon_sym_long] = ACTIONS(6595),
    [anon_sym_short] = ACTIONS(6595),
    [anon_sym_static] = ACTIONS(6595),
    [anon_sym_auto] = ACTIONS(6595),
    [anon_sym_register] = ACTIONS(6595),
    [anon_sym_inline] = ACTIONS(6595),
    [anon_sym___inline] = ACTIONS(6595),
    [anon_sym___inline__] = ACTIONS(6595),
    [anon_sym___forceinline] = ACTIONS(6595),
    [anon_sym_thread_local] = ACTIONS(6595),
    [anon_sym___thread] = ACTIONS(6595),
    [anon_sym_CG_EXTERN] = ACTIONS(6595),
    [anon_sym_CG_INLINE] = ACTIONS(6595),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6595),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6595),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6595),
    [anon_sym_IBOutlet] = ACTIONS(6595),
    [anon_sym_IBInspectable] = ACTIONS(6595),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6595),
    [anon_sym_NS_INLINE] = ACTIONS(6595),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6595),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6595),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6595),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6595),
    [anon_sym_const] = ACTIONS(6595),
    [anon_sym_constexpr] = ACTIONS(6595),
    [anon_sym_volatile] = ACTIONS(6595),
    [anon_sym_restrict] = ACTIONS(6595),
    [anon_sym___restrict__] = ACTIONS(6595),
    [anon_sym__Atomic] = ACTIONS(6595),
    [anon_sym__Noreturn] = ACTIONS(6595),
    [anon_sym__Nonnull] = ACTIONS(6595),
    [anon_sym_nullable] = ACTIONS(6595),
    [anon_sym__Complex] = ACTIONS(6595),
    [anon_sym__Nullable] = ACTIONS(6595),
    [anon_sym__Nullable_result] = ACTIONS(6595),
    [anon_sym__Null_unspecified] = ACTIONS(6595),
    [anon_sym___autoreleasing] = ACTIONS(6595),
    [anon_sym___block] = ACTIONS(6595),
    [anon_sym___bridge] = ACTIONS(6595),
    [anon_sym___bridge_retained] = ACTIONS(6595),
    [anon_sym___bridge_transfer] = ACTIONS(6595),
    [anon_sym___complex] = ACTIONS(6595),
    [anon_sym___const] = ACTIONS(6595),
    [anon_sym___imag] = ACTIONS(6595),
    [anon_sym___kindof] = ACTIONS(6595),
    [anon_sym___nonnull] = ACTIONS(6595),
    [anon_sym___nullable] = ACTIONS(6595),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6595),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6595),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6595),
    [anon_sym___real] = ACTIONS(6595),
    [anon_sym___strong] = ACTIONS(6595),
    [anon_sym___unsafe_unretained] = ACTIONS(6595),
    [anon_sym___unused] = ACTIONS(6595),
    [anon_sym___weak] = ACTIONS(6595),
    [anon_sym_alignas] = ACTIONS(6595),
    [anon_sym__Alignas] = ACTIONS(6595),
    [sym_primitive_type] = ACTIONS(6595),
    [anon_sym_enum] = ACTIONS(6595),
    [anon_sym_struct] = ACTIONS(6595),
    [anon_sym_union] = ACTIONS(6595),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6595),
    [anon_sym___typeof] = ACTIONS(6595),
    [anon_sym_typeof] = ACTIONS(6595),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6595),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6595),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6595),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6595),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6595),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6595),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6595),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6595),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6595),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_API_AVAILABLE] = ACTIONS(6595),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_API_DEPRECATED] = ACTIONS(6595),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6595),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6595),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6595),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6595),
    [anon_sym___deprecated_msg] = ACTIONS(6595),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6595),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6595),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6595),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6595),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6595),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6595),
    [anon_sym_BOOL] = ACTIONS(6595),
    [anon_sym_IMP] = ACTIONS(6595),
    [anon_sym_SEL] = ACTIONS(6595),
    [anon_sym_Class] = ACTIONS(6595),
    [anon_sym_id] = ACTIONS(6595),
    [anon_sym_ATdefs] = ACTIONS(6597),
  },
  [STATE(2761)] = {
    [sym_identifier] = ACTIONS(6587),
    [aux_sym_preproc_def_token1] = ACTIONS(6587),
    [aux_sym_preproc_if_token1] = ACTIONS(6587),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6587),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6587),
    [sym_preproc_directive] = ACTIONS(6587),
    [anon_sym___extension__] = ACTIONS(6587),
    [anon_sym_extern] = ACTIONS(6587),
    [anon_sym___attribute__] = ACTIONS(6587),
    [anon_sym___attribute] = ACTIONS(6587),
    [anon_sym_noreturn] = ACTIONS(6587),
    [anon_sym_LBRACK] = ACTIONS(6589),
    [anon_sym___declspec] = ACTIONS(6587),
    [anon_sym_RBRACE] = ACTIONS(6589),
    [anon_sym_signed] = ACTIONS(6587),
    [anon_sym_unsigned] = ACTIONS(6587),
    [anon_sym_long] = ACTIONS(6587),
    [anon_sym_short] = ACTIONS(6587),
    [anon_sym_static] = ACTIONS(6587),
    [anon_sym_auto] = ACTIONS(6587),
    [anon_sym_register] = ACTIONS(6587),
    [anon_sym_inline] = ACTIONS(6587),
    [anon_sym___inline] = ACTIONS(6587),
    [anon_sym___inline__] = ACTIONS(6587),
    [anon_sym___forceinline] = ACTIONS(6587),
    [anon_sym_thread_local] = ACTIONS(6587),
    [anon_sym___thread] = ACTIONS(6587),
    [anon_sym_CG_EXTERN] = ACTIONS(6587),
    [anon_sym_CG_INLINE] = ACTIONS(6587),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6587),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6587),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6587),
    [anon_sym_IBOutlet] = ACTIONS(6587),
    [anon_sym_IBInspectable] = ACTIONS(6587),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6587),
    [anon_sym_NS_INLINE] = ACTIONS(6587),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6587),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6587),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6587),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6587),
    [anon_sym_const] = ACTIONS(6587),
    [anon_sym_constexpr] = ACTIONS(6587),
    [anon_sym_volatile] = ACTIONS(6587),
    [anon_sym_restrict] = ACTIONS(6587),
    [anon_sym___restrict__] = ACTIONS(6587),
    [anon_sym__Atomic] = ACTIONS(6587),
    [anon_sym__Noreturn] = ACTIONS(6587),
    [anon_sym__Nonnull] = ACTIONS(6587),
    [anon_sym_nullable] = ACTIONS(6587),
    [anon_sym__Complex] = ACTIONS(6587),
    [anon_sym__Nullable] = ACTIONS(6587),
    [anon_sym__Nullable_result] = ACTIONS(6587),
    [anon_sym__Null_unspecified] = ACTIONS(6587),
    [anon_sym___autoreleasing] = ACTIONS(6587),
    [anon_sym___block] = ACTIONS(6587),
    [anon_sym___bridge] = ACTIONS(6587),
    [anon_sym___bridge_retained] = ACTIONS(6587),
    [anon_sym___bridge_transfer] = ACTIONS(6587),
    [anon_sym___complex] = ACTIONS(6587),
    [anon_sym___const] = ACTIONS(6587),
    [anon_sym___imag] = ACTIONS(6587),
    [anon_sym___kindof] = ACTIONS(6587),
    [anon_sym___nonnull] = ACTIONS(6587),
    [anon_sym___nullable] = ACTIONS(6587),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6587),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6587),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6587),
    [anon_sym___real] = ACTIONS(6587),
    [anon_sym___strong] = ACTIONS(6587),
    [anon_sym___unsafe_unretained] = ACTIONS(6587),
    [anon_sym___unused] = ACTIONS(6587),
    [anon_sym___weak] = ACTIONS(6587),
    [anon_sym_alignas] = ACTIONS(6587),
    [anon_sym__Alignas] = ACTIONS(6587),
    [sym_primitive_type] = ACTIONS(6587),
    [anon_sym_enum] = ACTIONS(6587),
    [anon_sym_struct] = ACTIONS(6587),
    [anon_sym_union] = ACTIONS(6587),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6587),
    [anon_sym___typeof] = ACTIONS(6587),
    [anon_sym_typeof] = ACTIONS(6587),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6587),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6587),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6587),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6587),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6587),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6587),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6587),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6587),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6587),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_API_AVAILABLE] = ACTIONS(6587),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_API_DEPRECATED] = ACTIONS(6587),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6587),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6587),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6587),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6587),
    [anon_sym___deprecated_msg] = ACTIONS(6587),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6587),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6587),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6587),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6587),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6587),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6587),
    [anon_sym_BOOL] = ACTIONS(6587),
    [anon_sym_IMP] = ACTIONS(6587),
    [anon_sym_SEL] = ACTIONS(6587),
    [anon_sym_Class] = ACTIONS(6587),
    [anon_sym_id] = ACTIONS(6587),
    [anon_sym_ATdefs] = ACTIONS(6589),
  },
  [STATE(2762)] = {
    [sym_identifier] = ACTIONS(6559),
    [aux_sym_preproc_def_token1] = ACTIONS(6559),
    [aux_sym_preproc_if_token1] = ACTIONS(6559),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6559),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6559),
    [sym_preproc_directive] = ACTIONS(6559),
    [anon_sym___extension__] = ACTIONS(6559),
    [anon_sym_extern] = ACTIONS(6559),
    [anon_sym___attribute__] = ACTIONS(6559),
    [anon_sym___attribute] = ACTIONS(6559),
    [anon_sym_noreturn] = ACTIONS(6559),
    [anon_sym_LBRACK] = ACTIONS(6561),
    [anon_sym___declspec] = ACTIONS(6559),
    [anon_sym_RBRACE] = ACTIONS(6561),
    [anon_sym_signed] = ACTIONS(6559),
    [anon_sym_unsigned] = ACTIONS(6559),
    [anon_sym_long] = ACTIONS(6559),
    [anon_sym_short] = ACTIONS(6559),
    [anon_sym_static] = ACTIONS(6559),
    [anon_sym_auto] = ACTIONS(6559),
    [anon_sym_register] = ACTIONS(6559),
    [anon_sym_inline] = ACTIONS(6559),
    [anon_sym___inline] = ACTIONS(6559),
    [anon_sym___inline__] = ACTIONS(6559),
    [anon_sym___forceinline] = ACTIONS(6559),
    [anon_sym_thread_local] = ACTIONS(6559),
    [anon_sym___thread] = ACTIONS(6559),
    [anon_sym_CG_EXTERN] = ACTIONS(6559),
    [anon_sym_CG_INLINE] = ACTIONS(6559),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6559),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6559),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6559),
    [anon_sym_IBOutlet] = ACTIONS(6559),
    [anon_sym_IBInspectable] = ACTIONS(6559),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6559),
    [anon_sym_NS_INLINE] = ACTIONS(6559),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6559),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6559),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6559),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6559),
    [anon_sym_const] = ACTIONS(6559),
    [anon_sym_constexpr] = ACTIONS(6559),
    [anon_sym_volatile] = ACTIONS(6559),
    [anon_sym_restrict] = ACTIONS(6559),
    [anon_sym___restrict__] = ACTIONS(6559),
    [anon_sym__Atomic] = ACTIONS(6559),
    [anon_sym__Noreturn] = ACTIONS(6559),
    [anon_sym__Nonnull] = ACTIONS(6559),
    [anon_sym_nullable] = ACTIONS(6559),
    [anon_sym__Complex] = ACTIONS(6559),
    [anon_sym__Nullable] = ACTIONS(6559),
    [anon_sym__Nullable_result] = ACTIONS(6559),
    [anon_sym__Null_unspecified] = ACTIONS(6559),
    [anon_sym___autoreleasing] = ACTIONS(6559),
    [anon_sym___block] = ACTIONS(6559),
    [anon_sym___bridge] = ACTIONS(6559),
    [anon_sym___bridge_retained] = ACTIONS(6559),
    [anon_sym___bridge_transfer] = ACTIONS(6559),
    [anon_sym___complex] = ACTIONS(6559),
    [anon_sym___const] = ACTIONS(6559),
    [anon_sym___imag] = ACTIONS(6559),
    [anon_sym___kindof] = ACTIONS(6559),
    [anon_sym___nonnull] = ACTIONS(6559),
    [anon_sym___nullable] = ACTIONS(6559),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6559),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6559),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6559),
    [anon_sym___real] = ACTIONS(6559),
    [anon_sym___strong] = ACTIONS(6559),
    [anon_sym___unsafe_unretained] = ACTIONS(6559),
    [anon_sym___unused] = ACTIONS(6559),
    [anon_sym___weak] = ACTIONS(6559),
    [anon_sym_alignas] = ACTIONS(6559),
    [anon_sym__Alignas] = ACTIONS(6559),
    [sym_primitive_type] = ACTIONS(6559),
    [anon_sym_enum] = ACTIONS(6559),
    [anon_sym_struct] = ACTIONS(6559),
    [anon_sym_union] = ACTIONS(6559),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6559),
    [anon_sym___typeof] = ACTIONS(6559),
    [anon_sym_typeof] = ACTIONS(6559),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6559),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6559),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6559),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6559),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6559),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6559),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6559),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6559),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6559),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_API_AVAILABLE] = ACTIONS(6559),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_API_DEPRECATED] = ACTIONS(6559),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6559),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6559),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6559),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6559),
    [anon_sym___deprecated_msg] = ACTIONS(6559),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6559),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6559),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6559),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6559),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6559),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6559),
    [anon_sym_BOOL] = ACTIONS(6559),
    [anon_sym_IMP] = ACTIONS(6559),
    [anon_sym_SEL] = ACTIONS(6559),
    [anon_sym_Class] = ACTIONS(6559),
    [anon_sym_id] = ACTIONS(6559),
    [anon_sym_ATdefs] = ACTIONS(6561),
  },
  [STATE(2763)] = {
    [sym_identifier] = ACTIONS(6563),
    [aux_sym_preproc_def_token1] = ACTIONS(6563),
    [aux_sym_preproc_if_token1] = ACTIONS(6563),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6563),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6563),
    [sym_preproc_directive] = ACTIONS(6563),
    [anon_sym___extension__] = ACTIONS(6563),
    [anon_sym_extern] = ACTIONS(6563),
    [anon_sym___attribute__] = ACTIONS(6563),
    [anon_sym___attribute] = ACTIONS(6563),
    [anon_sym_noreturn] = ACTIONS(6563),
    [anon_sym_LBRACK] = ACTIONS(6565),
    [anon_sym___declspec] = ACTIONS(6563),
    [anon_sym_RBRACE] = ACTIONS(6565),
    [anon_sym_signed] = ACTIONS(6563),
    [anon_sym_unsigned] = ACTIONS(6563),
    [anon_sym_long] = ACTIONS(6563),
    [anon_sym_short] = ACTIONS(6563),
    [anon_sym_static] = ACTIONS(6563),
    [anon_sym_auto] = ACTIONS(6563),
    [anon_sym_register] = ACTIONS(6563),
    [anon_sym_inline] = ACTIONS(6563),
    [anon_sym___inline] = ACTIONS(6563),
    [anon_sym___inline__] = ACTIONS(6563),
    [anon_sym___forceinline] = ACTIONS(6563),
    [anon_sym_thread_local] = ACTIONS(6563),
    [anon_sym___thread] = ACTIONS(6563),
    [anon_sym_CG_EXTERN] = ACTIONS(6563),
    [anon_sym_CG_INLINE] = ACTIONS(6563),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6563),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6563),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6563),
    [anon_sym_IBOutlet] = ACTIONS(6563),
    [anon_sym_IBInspectable] = ACTIONS(6563),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6563),
    [anon_sym_NS_INLINE] = ACTIONS(6563),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6563),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6563),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6563),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6563),
    [anon_sym_const] = ACTIONS(6563),
    [anon_sym_constexpr] = ACTIONS(6563),
    [anon_sym_volatile] = ACTIONS(6563),
    [anon_sym_restrict] = ACTIONS(6563),
    [anon_sym___restrict__] = ACTIONS(6563),
    [anon_sym__Atomic] = ACTIONS(6563),
    [anon_sym__Noreturn] = ACTIONS(6563),
    [anon_sym__Nonnull] = ACTIONS(6563),
    [anon_sym_nullable] = ACTIONS(6563),
    [anon_sym__Complex] = ACTIONS(6563),
    [anon_sym__Nullable] = ACTIONS(6563),
    [anon_sym__Nullable_result] = ACTIONS(6563),
    [anon_sym__Null_unspecified] = ACTIONS(6563),
    [anon_sym___autoreleasing] = ACTIONS(6563),
    [anon_sym___block] = ACTIONS(6563),
    [anon_sym___bridge] = ACTIONS(6563),
    [anon_sym___bridge_retained] = ACTIONS(6563),
    [anon_sym___bridge_transfer] = ACTIONS(6563),
    [anon_sym___complex] = ACTIONS(6563),
    [anon_sym___const] = ACTIONS(6563),
    [anon_sym___imag] = ACTIONS(6563),
    [anon_sym___kindof] = ACTIONS(6563),
    [anon_sym___nonnull] = ACTIONS(6563),
    [anon_sym___nullable] = ACTIONS(6563),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6563),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6563),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6563),
    [anon_sym___real] = ACTIONS(6563),
    [anon_sym___strong] = ACTIONS(6563),
    [anon_sym___unsafe_unretained] = ACTIONS(6563),
    [anon_sym___unused] = ACTIONS(6563),
    [anon_sym___weak] = ACTIONS(6563),
    [anon_sym_alignas] = ACTIONS(6563),
    [anon_sym__Alignas] = ACTIONS(6563),
    [sym_primitive_type] = ACTIONS(6563),
    [anon_sym_enum] = ACTIONS(6563),
    [anon_sym_struct] = ACTIONS(6563),
    [anon_sym_union] = ACTIONS(6563),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6563),
    [anon_sym___typeof] = ACTIONS(6563),
    [anon_sym_typeof] = ACTIONS(6563),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6563),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6563),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6563),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6563),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6563),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6563),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6563),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6563),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6563),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_API_AVAILABLE] = ACTIONS(6563),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_API_DEPRECATED] = ACTIONS(6563),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6563),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6563),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6563),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6563),
    [anon_sym___deprecated_msg] = ACTIONS(6563),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6563),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6563),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6563),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6563),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6563),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6563),
    [anon_sym_BOOL] = ACTIONS(6563),
    [anon_sym_IMP] = ACTIONS(6563),
    [anon_sym_SEL] = ACTIONS(6563),
    [anon_sym_Class] = ACTIONS(6563),
    [anon_sym_id] = ACTIONS(6563),
    [anon_sym_ATdefs] = ACTIONS(6565),
  },
  [STATE(2764)] = {
    [sym_identifier] = ACTIONS(6567),
    [aux_sym_preproc_def_token1] = ACTIONS(6567),
    [aux_sym_preproc_if_token1] = ACTIONS(6567),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6567),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6567),
    [sym_preproc_directive] = ACTIONS(6567),
    [anon_sym___extension__] = ACTIONS(6567),
    [anon_sym_extern] = ACTIONS(6567),
    [anon_sym___attribute__] = ACTIONS(6567),
    [anon_sym___attribute] = ACTIONS(6567),
    [anon_sym_noreturn] = ACTIONS(6567),
    [anon_sym_LBRACK] = ACTIONS(6569),
    [anon_sym___declspec] = ACTIONS(6567),
    [anon_sym_RBRACE] = ACTIONS(6569),
    [anon_sym_signed] = ACTIONS(6567),
    [anon_sym_unsigned] = ACTIONS(6567),
    [anon_sym_long] = ACTIONS(6567),
    [anon_sym_short] = ACTIONS(6567),
    [anon_sym_static] = ACTIONS(6567),
    [anon_sym_auto] = ACTIONS(6567),
    [anon_sym_register] = ACTIONS(6567),
    [anon_sym_inline] = ACTIONS(6567),
    [anon_sym___inline] = ACTIONS(6567),
    [anon_sym___inline__] = ACTIONS(6567),
    [anon_sym___forceinline] = ACTIONS(6567),
    [anon_sym_thread_local] = ACTIONS(6567),
    [anon_sym___thread] = ACTIONS(6567),
    [anon_sym_CG_EXTERN] = ACTIONS(6567),
    [anon_sym_CG_INLINE] = ACTIONS(6567),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6567),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6567),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6567),
    [anon_sym_IBOutlet] = ACTIONS(6567),
    [anon_sym_IBInspectable] = ACTIONS(6567),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6567),
    [anon_sym_NS_INLINE] = ACTIONS(6567),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6567),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6567),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6567),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6567),
    [anon_sym_const] = ACTIONS(6567),
    [anon_sym_constexpr] = ACTIONS(6567),
    [anon_sym_volatile] = ACTIONS(6567),
    [anon_sym_restrict] = ACTIONS(6567),
    [anon_sym___restrict__] = ACTIONS(6567),
    [anon_sym__Atomic] = ACTIONS(6567),
    [anon_sym__Noreturn] = ACTIONS(6567),
    [anon_sym__Nonnull] = ACTIONS(6567),
    [anon_sym_nullable] = ACTIONS(6567),
    [anon_sym__Complex] = ACTIONS(6567),
    [anon_sym__Nullable] = ACTIONS(6567),
    [anon_sym__Nullable_result] = ACTIONS(6567),
    [anon_sym__Null_unspecified] = ACTIONS(6567),
    [anon_sym___autoreleasing] = ACTIONS(6567),
    [anon_sym___block] = ACTIONS(6567),
    [anon_sym___bridge] = ACTIONS(6567),
    [anon_sym___bridge_retained] = ACTIONS(6567),
    [anon_sym___bridge_transfer] = ACTIONS(6567),
    [anon_sym___complex] = ACTIONS(6567),
    [anon_sym___const] = ACTIONS(6567),
    [anon_sym___imag] = ACTIONS(6567),
    [anon_sym___kindof] = ACTIONS(6567),
    [anon_sym___nonnull] = ACTIONS(6567),
    [anon_sym___nullable] = ACTIONS(6567),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6567),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6567),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6567),
    [anon_sym___real] = ACTIONS(6567),
    [anon_sym___strong] = ACTIONS(6567),
    [anon_sym___unsafe_unretained] = ACTIONS(6567),
    [anon_sym___unused] = ACTIONS(6567),
    [anon_sym___weak] = ACTIONS(6567),
    [anon_sym_alignas] = ACTIONS(6567),
    [anon_sym__Alignas] = ACTIONS(6567),
    [sym_primitive_type] = ACTIONS(6567),
    [anon_sym_enum] = ACTIONS(6567),
    [anon_sym_struct] = ACTIONS(6567),
    [anon_sym_union] = ACTIONS(6567),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6567),
    [anon_sym___typeof] = ACTIONS(6567),
    [anon_sym_typeof] = ACTIONS(6567),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6567),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6567),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6567),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6567),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6567),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6567),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6567),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6567),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6567),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_API_AVAILABLE] = ACTIONS(6567),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_API_DEPRECATED] = ACTIONS(6567),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6567),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6567),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6567),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6567),
    [anon_sym___deprecated_msg] = ACTIONS(6567),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6567),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6567),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6567),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6567),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6567),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6567),
    [anon_sym_BOOL] = ACTIONS(6567),
    [anon_sym_IMP] = ACTIONS(6567),
    [anon_sym_SEL] = ACTIONS(6567),
    [anon_sym_Class] = ACTIONS(6567),
    [anon_sym_id] = ACTIONS(6567),
    [anon_sym_ATdefs] = ACTIONS(6569),
  },
  [STATE(2765)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_RBRACE] = ACTIONS(6573),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2766)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_RBRACE] = ACTIONS(6573),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2767)] = {
    [sym_identifier] = ACTIONS(6575),
    [aux_sym_preproc_def_token1] = ACTIONS(6575),
    [aux_sym_preproc_if_token1] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6575),
    [sym_preproc_directive] = ACTIONS(6575),
    [anon_sym___extension__] = ACTIONS(6575),
    [anon_sym_extern] = ACTIONS(6575),
    [anon_sym___attribute__] = ACTIONS(6575),
    [anon_sym___attribute] = ACTIONS(6575),
    [anon_sym_noreturn] = ACTIONS(6575),
    [anon_sym_LBRACK] = ACTIONS(6577),
    [anon_sym___declspec] = ACTIONS(6575),
    [anon_sym_RBRACE] = ACTIONS(6577),
    [anon_sym_signed] = ACTIONS(6575),
    [anon_sym_unsigned] = ACTIONS(6575),
    [anon_sym_long] = ACTIONS(6575),
    [anon_sym_short] = ACTIONS(6575),
    [anon_sym_static] = ACTIONS(6575),
    [anon_sym_auto] = ACTIONS(6575),
    [anon_sym_register] = ACTIONS(6575),
    [anon_sym_inline] = ACTIONS(6575),
    [anon_sym___inline] = ACTIONS(6575),
    [anon_sym___inline__] = ACTIONS(6575),
    [anon_sym___forceinline] = ACTIONS(6575),
    [anon_sym_thread_local] = ACTIONS(6575),
    [anon_sym___thread] = ACTIONS(6575),
    [anon_sym_CG_EXTERN] = ACTIONS(6575),
    [anon_sym_CG_INLINE] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6575),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6575),
    [anon_sym_IBOutlet] = ACTIONS(6575),
    [anon_sym_IBInspectable] = ACTIONS(6575),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6575),
    [anon_sym_NS_INLINE] = ACTIONS(6575),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6575),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6575),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6575),
    [anon_sym_const] = ACTIONS(6575),
    [anon_sym_constexpr] = ACTIONS(6575),
    [anon_sym_volatile] = ACTIONS(6575),
    [anon_sym_restrict] = ACTIONS(6575),
    [anon_sym___restrict__] = ACTIONS(6575),
    [anon_sym__Atomic] = ACTIONS(6575),
    [anon_sym__Noreturn] = ACTIONS(6575),
    [anon_sym__Nonnull] = ACTIONS(6575),
    [anon_sym_nullable] = ACTIONS(6575),
    [anon_sym__Complex] = ACTIONS(6575),
    [anon_sym__Nullable] = ACTIONS(6575),
    [anon_sym__Nullable_result] = ACTIONS(6575),
    [anon_sym__Null_unspecified] = ACTIONS(6575),
    [anon_sym___autoreleasing] = ACTIONS(6575),
    [anon_sym___block] = ACTIONS(6575),
    [anon_sym___bridge] = ACTIONS(6575),
    [anon_sym___bridge_retained] = ACTIONS(6575),
    [anon_sym___bridge_transfer] = ACTIONS(6575),
    [anon_sym___complex] = ACTIONS(6575),
    [anon_sym___const] = ACTIONS(6575),
    [anon_sym___imag] = ACTIONS(6575),
    [anon_sym___kindof] = ACTIONS(6575),
    [anon_sym___nonnull] = ACTIONS(6575),
    [anon_sym___nullable] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6575),
    [anon_sym___real] = ACTIONS(6575),
    [anon_sym___strong] = ACTIONS(6575),
    [anon_sym___unsafe_unretained] = ACTIONS(6575),
    [anon_sym___unused] = ACTIONS(6575),
    [anon_sym___weak] = ACTIONS(6575),
    [anon_sym_alignas] = ACTIONS(6575),
    [anon_sym__Alignas] = ACTIONS(6575),
    [sym_primitive_type] = ACTIONS(6575),
    [anon_sym_enum] = ACTIONS(6575),
    [anon_sym_struct] = ACTIONS(6575),
    [anon_sym_union] = ACTIONS(6575),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6575),
    [anon_sym___typeof] = ACTIONS(6575),
    [anon_sym_typeof] = ACTIONS(6575),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6575),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6575),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6575),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_API_AVAILABLE] = ACTIONS(6575),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_API_DEPRECATED] = ACTIONS(6575),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym___deprecated_msg] = ACTIONS(6575),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6575),
    [anon_sym_BOOL] = ACTIONS(6575),
    [anon_sym_IMP] = ACTIONS(6575),
    [anon_sym_SEL] = ACTIONS(6575),
    [anon_sym_Class] = ACTIONS(6575),
    [anon_sym_id] = ACTIONS(6575),
    [anon_sym_ATdefs] = ACTIONS(6577),
  },
  [STATE(2768)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_RBRACE] = ACTIONS(6585),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2769)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_RBRACE] = ACTIONS(6585),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2770)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_RBRACE] = ACTIONS(6573),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2771)] = {
    [sym_identifier] = ACTIONS(6571),
    [aux_sym_preproc_def_token1] = ACTIONS(6571),
    [aux_sym_preproc_if_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6571),
    [sym_preproc_directive] = ACTIONS(6571),
    [anon_sym___extension__] = ACTIONS(6571),
    [anon_sym_extern] = ACTIONS(6571),
    [anon_sym___attribute__] = ACTIONS(6571),
    [anon_sym___attribute] = ACTIONS(6571),
    [anon_sym_noreturn] = ACTIONS(6571),
    [anon_sym_LBRACK] = ACTIONS(6573),
    [anon_sym___declspec] = ACTIONS(6571),
    [anon_sym_RBRACE] = ACTIONS(6573),
    [anon_sym_signed] = ACTIONS(6571),
    [anon_sym_unsigned] = ACTIONS(6571),
    [anon_sym_long] = ACTIONS(6571),
    [anon_sym_short] = ACTIONS(6571),
    [anon_sym_static] = ACTIONS(6571),
    [anon_sym_auto] = ACTIONS(6571),
    [anon_sym_register] = ACTIONS(6571),
    [anon_sym_inline] = ACTIONS(6571),
    [anon_sym___inline] = ACTIONS(6571),
    [anon_sym___inline__] = ACTIONS(6571),
    [anon_sym___forceinline] = ACTIONS(6571),
    [anon_sym_thread_local] = ACTIONS(6571),
    [anon_sym___thread] = ACTIONS(6571),
    [anon_sym_CG_EXTERN] = ACTIONS(6571),
    [anon_sym_CG_INLINE] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6571),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6571),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6571),
    [anon_sym_IBOutlet] = ACTIONS(6571),
    [anon_sym_IBInspectable] = ACTIONS(6571),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6571),
    [anon_sym_NS_INLINE] = ACTIONS(6571),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6571),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6571),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6571),
    [anon_sym_const] = ACTIONS(6571),
    [anon_sym_constexpr] = ACTIONS(6571),
    [anon_sym_volatile] = ACTIONS(6571),
    [anon_sym_restrict] = ACTIONS(6571),
    [anon_sym___restrict__] = ACTIONS(6571),
    [anon_sym__Atomic] = ACTIONS(6571),
    [anon_sym__Noreturn] = ACTIONS(6571),
    [anon_sym__Nonnull] = ACTIONS(6571),
    [anon_sym_nullable] = ACTIONS(6571),
    [anon_sym__Complex] = ACTIONS(6571),
    [anon_sym__Nullable] = ACTIONS(6571),
    [anon_sym__Nullable_result] = ACTIONS(6571),
    [anon_sym__Null_unspecified] = ACTIONS(6571),
    [anon_sym___autoreleasing] = ACTIONS(6571),
    [anon_sym___block] = ACTIONS(6571),
    [anon_sym___bridge] = ACTIONS(6571),
    [anon_sym___bridge_retained] = ACTIONS(6571),
    [anon_sym___bridge_transfer] = ACTIONS(6571),
    [anon_sym___complex] = ACTIONS(6571),
    [anon_sym___const] = ACTIONS(6571),
    [anon_sym___imag] = ACTIONS(6571),
    [anon_sym___kindof] = ACTIONS(6571),
    [anon_sym___nonnull] = ACTIONS(6571),
    [anon_sym___nullable] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6571),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6571),
    [anon_sym___real] = ACTIONS(6571),
    [anon_sym___strong] = ACTIONS(6571),
    [anon_sym___unsafe_unretained] = ACTIONS(6571),
    [anon_sym___unused] = ACTIONS(6571),
    [anon_sym___weak] = ACTIONS(6571),
    [anon_sym_alignas] = ACTIONS(6571),
    [anon_sym__Alignas] = ACTIONS(6571),
    [sym_primitive_type] = ACTIONS(6571),
    [anon_sym_enum] = ACTIONS(6571),
    [anon_sym_struct] = ACTIONS(6571),
    [anon_sym_union] = ACTIONS(6571),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6571),
    [anon_sym___typeof] = ACTIONS(6571),
    [anon_sym_typeof] = ACTIONS(6571),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6571),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6571),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6571),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6571),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6571),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_API_AVAILABLE] = ACTIONS(6571),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_API_DEPRECATED] = ACTIONS(6571),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6571),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6571),
    [anon_sym___deprecated_msg] = ACTIONS(6571),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6571),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6571),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6571),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6571),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6571),
    [anon_sym_BOOL] = ACTIONS(6571),
    [anon_sym_IMP] = ACTIONS(6571),
    [anon_sym_SEL] = ACTIONS(6571),
    [anon_sym_Class] = ACTIONS(6571),
    [anon_sym_id] = ACTIONS(6571),
    [anon_sym_ATdefs] = ACTIONS(6573),
  },
  [STATE(2772)] = {
    [sym_identifier] = ACTIONS(6575),
    [aux_sym_preproc_def_token1] = ACTIONS(6575),
    [aux_sym_preproc_if_token1] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6575),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6575),
    [sym_preproc_directive] = ACTIONS(6575),
    [anon_sym___extension__] = ACTIONS(6575),
    [anon_sym_extern] = ACTIONS(6575),
    [anon_sym___attribute__] = ACTIONS(6575),
    [anon_sym___attribute] = ACTIONS(6575),
    [anon_sym_noreturn] = ACTIONS(6575),
    [anon_sym_LBRACK] = ACTIONS(6577),
    [anon_sym___declspec] = ACTIONS(6575),
    [anon_sym_RBRACE] = ACTIONS(6577),
    [anon_sym_signed] = ACTIONS(6575),
    [anon_sym_unsigned] = ACTIONS(6575),
    [anon_sym_long] = ACTIONS(6575),
    [anon_sym_short] = ACTIONS(6575),
    [anon_sym_static] = ACTIONS(6575),
    [anon_sym_auto] = ACTIONS(6575),
    [anon_sym_register] = ACTIONS(6575),
    [anon_sym_inline] = ACTIONS(6575),
    [anon_sym___inline] = ACTIONS(6575),
    [anon_sym___inline__] = ACTIONS(6575),
    [anon_sym___forceinline] = ACTIONS(6575),
    [anon_sym_thread_local] = ACTIONS(6575),
    [anon_sym___thread] = ACTIONS(6575),
    [anon_sym_CG_EXTERN] = ACTIONS(6575),
    [anon_sym_CG_INLINE] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6575),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6575),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6575),
    [anon_sym_IBOutlet] = ACTIONS(6575),
    [anon_sym_IBInspectable] = ACTIONS(6575),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6575),
    [anon_sym_NS_INLINE] = ACTIONS(6575),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6575),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6575),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6575),
    [anon_sym_const] = ACTIONS(6575),
    [anon_sym_constexpr] = ACTIONS(6575),
    [anon_sym_volatile] = ACTIONS(6575),
    [anon_sym_restrict] = ACTIONS(6575),
    [anon_sym___restrict__] = ACTIONS(6575),
    [anon_sym__Atomic] = ACTIONS(6575),
    [anon_sym__Noreturn] = ACTIONS(6575),
    [anon_sym__Nonnull] = ACTIONS(6575),
    [anon_sym_nullable] = ACTIONS(6575),
    [anon_sym__Complex] = ACTIONS(6575),
    [anon_sym__Nullable] = ACTIONS(6575),
    [anon_sym__Nullable_result] = ACTIONS(6575),
    [anon_sym__Null_unspecified] = ACTIONS(6575),
    [anon_sym___autoreleasing] = ACTIONS(6575),
    [anon_sym___block] = ACTIONS(6575),
    [anon_sym___bridge] = ACTIONS(6575),
    [anon_sym___bridge_retained] = ACTIONS(6575),
    [anon_sym___bridge_transfer] = ACTIONS(6575),
    [anon_sym___complex] = ACTIONS(6575),
    [anon_sym___const] = ACTIONS(6575),
    [anon_sym___imag] = ACTIONS(6575),
    [anon_sym___kindof] = ACTIONS(6575),
    [anon_sym___nonnull] = ACTIONS(6575),
    [anon_sym___nullable] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6575),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6575),
    [anon_sym___real] = ACTIONS(6575),
    [anon_sym___strong] = ACTIONS(6575),
    [anon_sym___unsafe_unretained] = ACTIONS(6575),
    [anon_sym___unused] = ACTIONS(6575),
    [anon_sym___weak] = ACTIONS(6575),
    [anon_sym_alignas] = ACTIONS(6575),
    [anon_sym__Alignas] = ACTIONS(6575),
    [sym_primitive_type] = ACTIONS(6575),
    [anon_sym_enum] = ACTIONS(6575),
    [anon_sym_struct] = ACTIONS(6575),
    [anon_sym_union] = ACTIONS(6575),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6575),
    [anon_sym___typeof] = ACTIONS(6575),
    [anon_sym_typeof] = ACTIONS(6575),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6575),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6575),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6575),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6575),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6575),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_API_AVAILABLE] = ACTIONS(6575),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_API_DEPRECATED] = ACTIONS(6575),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6575),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6575),
    [anon_sym___deprecated_msg] = ACTIONS(6575),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6575),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6575),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6575),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6575),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6575),
    [anon_sym_BOOL] = ACTIONS(6575),
    [anon_sym_IMP] = ACTIONS(6575),
    [anon_sym_SEL] = ACTIONS(6575),
    [anon_sym_Class] = ACTIONS(6575),
    [anon_sym_id] = ACTIONS(6575),
    [anon_sym_ATdefs] = ACTIONS(6577),
  },
  [STATE(2773)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_RBRACE] = ACTIONS(6585),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2774)] = {
    [sym_identifier] = ACTIONS(6583),
    [aux_sym_preproc_def_token1] = ACTIONS(6583),
    [aux_sym_preproc_if_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6583),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6583),
    [sym_preproc_directive] = ACTIONS(6583),
    [anon_sym___extension__] = ACTIONS(6583),
    [anon_sym_extern] = ACTIONS(6583),
    [anon_sym___attribute__] = ACTIONS(6583),
    [anon_sym___attribute] = ACTIONS(6583),
    [anon_sym_noreturn] = ACTIONS(6583),
    [anon_sym_LBRACK] = ACTIONS(6585),
    [anon_sym___declspec] = ACTIONS(6583),
    [anon_sym_RBRACE] = ACTIONS(6585),
    [anon_sym_signed] = ACTIONS(6583),
    [anon_sym_unsigned] = ACTIONS(6583),
    [anon_sym_long] = ACTIONS(6583),
    [anon_sym_short] = ACTIONS(6583),
    [anon_sym_static] = ACTIONS(6583),
    [anon_sym_auto] = ACTIONS(6583),
    [anon_sym_register] = ACTIONS(6583),
    [anon_sym_inline] = ACTIONS(6583),
    [anon_sym___inline] = ACTIONS(6583),
    [anon_sym___inline__] = ACTIONS(6583),
    [anon_sym___forceinline] = ACTIONS(6583),
    [anon_sym_thread_local] = ACTIONS(6583),
    [anon_sym___thread] = ACTIONS(6583),
    [anon_sym_CG_EXTERN] = ACTIONS(6583),
    [anon_sym_CG_INLINE] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6583),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6583),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6583),
    [anon_sym_IBOutlet] = ACTIONS(6583),
    [anon_sym_IBInspectable] = ACTIONS(6583),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6583),
    [anon_sym_NS_INLINE] = ACTIONS(6583),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6583),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6583),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6583),
    [anon_sym_const] = ACTIONS(6583),
    [anon_sym_constexpr] = ACTIONS(6583),
    [anon_sym_volatile] = ACTIONS(6583),
    [anon_sym_restrict] = ACTIONS(6583),
    [anon_sym___restrict__] = ACTIONS(6583),
    [anon_sym__Atomic] = ACTIONS(6583),
    [anon_sym__Noreturn] = ACTIONS(6583),
    [anon_sym__Nonnull] = ACTIONS(6583),
    [anon_sym_nullable] = ACTIONS(6583),
    [anon_sym__Complex] = ACTIONS(6583),
    [anon_sym__Nullable] = ACTIONS(6583),
    [anon_sym__Nullable_result] = ACTIONS(6583),
    [anon_sym__Null_unspecified] = ACTIONS(6583),
    [anon_sym___autoreleasing] = ACTIONS(6583),
    [anon_sym___block] = ACTIONS(6583),
    [anon_sym___bridge] = ACTIONS(6583),
    [anon_sym___bridge_retained] = ACTIONS(6583),
    [anon_sym___bridge_transfer] = ACTIONS(6583),
    [anon_sym___complex] = ACTIONS(6583),
    [anon_sym___const] = ACTIONS(6583),
    [anon_sym___imag] = ACTIONS(6583),
    [anon_sym___kindof] = ACTIONS(6583),
    [anon_sym___nonnull] = ACTIONS(6583),
    [anon_sym___nullable] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6583),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6583),
    [anon_sym___real] = ACTIONS(6583),
    [anon_sym___strong] = ACTIONS(6583),
    [anon_sym___unsafe_unretained] = ACTIONS(6583),
    [anon_sym___unused] = ACTIONS(6583),
    [anon_sym___weak] = ACTIONS(6583),
    [anon_sym_alignas] = ACTIONS(6583),
    [anon_sym__Alignas] = ACTIONS(6583),
    [sym_primitive_type] = ACTIONS(6583),
    [anon_sym_enum] = ACTIONS(6583),
    [anon_sym_struct] = ACTIONS(6583),
    [anon_sym_union] = ACTIONS(6583),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6583),
    [anon_sym___typeof] = ACTIONS(6583),
    [anon_sym_typeof] = ACTIONS(6583),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6583),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6583),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6583),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6583),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6583),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_API_AVAILABLE] = ACTIONS(6583),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_API_DEPRECATED] = ACTIONS(6583),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6583),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6583),
    [anon_sym___deprecated_msg] = ACTIONS(6583),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6583),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6583),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6583),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6583),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6583),
    [anon_sym_BOOL] = ACTIONS(6583),
    [anon_sym_IMP] = ACTIONS(6583),
    [anon_sym_SEL] = ACTIONS(6583),
    [anon_sym_Class] = ACTIONS(6583),
    [anon_sym_id] = ACTIONS(6583),
    [anon_sym_ATdefs] = ACTIONS(6585),
  },
  [STATE(2775)] = {
    [sym_identifier] = ACTIONS(6535),
    [aux_sym_preproc_def_token1] = ACTIONS(6535),
    [aux_sym_preproc_if_token1] = ACTIONS(6535),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(6535),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(6535),
    [sym_preproc_directive] = ACTIONS(6535),
    [anon_sym___extension__] = ACTIONS(6535),
    [anon_sym_extern] = ACTIONS(6535),
    [anon_sym___attribute__] = ACTIONS(6535),
    [anon_sym___attribute] = ACTIONS(6535),
    [anon_sym_noreturn] = ACTIONS(6535),
    [anon_sym_LBRACK] = ACTIONS(6537),
    [anon_sym___declspec] = ACTIONS(6535),
    [anon_sym_RBRACE] = ACTIONS(6537),
    [anon_sym_signed] = ACTIONS(6535),
    [anon_sym_unsigned] = ACTIONS(6535),
    [anon_sym_long] = ACTIONS(6535),
    [anon_sym_short] = ACTIONS(6535),
    [anon_sym_static] = ACTIONS(6535),
    [anon_sym_auto] = ACTIONS(6535),
    [anon_sym_register] = ACTIONS(6535),
    [anon_sym_inline] = ACTIONS(6535),
    [anon_sym___inline] = ACTIONS(6535),
    [anon_sym___inline__] = ACTIONS(6535),
    [anon_sym___forceinline] = ACTIONS(6535),
    [anon_sym_thread_local] = ACTIONS(6535),
    [anon_sym___thread] = ACTIONS(6535),
    [anon_sym_CG_EXTERN] = ACTIONS(6535),
    [anon_sym_CG_INLINE] = ACTIONS(6535),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6535),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6535),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6535),
    [anon_sym_IBOutlet] = ACTIONS(6535),
    [anon_sym_IBInspectable] = ACTIONS(6535),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6535),
    [anon_sym_NS_INLINE] = ACTIONS(6535),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6535),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6535),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6535),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6535),
    [anon_sym_const] = ACTIONS(6535),
    [anon_sym_constexpr] = ACTIONS(6535),
    [anon_sym_volatile] = ACTIONS(6535),
    [anon_sym_restrict] = ACTIONS(6535),
    [anon_sym___restrict__] = ACTIONS(6535),
    [anon_sym__Atomic] = ACTIONS(6535),
    [anon_sym__Noreturn] = ACTIONS(6535),
    [anon_sym__Nonnull] = ACTIONS(6535),
    [anon_sym_nullable] = ACTIONS(6535),
    [anon_sym__Complex] = ACTIONS(6535),
    [anon_sym__Nullable] = ACTIONS(6535),
    [anon_sym__Nullable_result] = ACTIONS(6535),
    [anon_sym__Null_unspecified] = ACTIONS(6535),
    [anon_sym___autoreleasing] = ACTIONS(6535),
    [anon_sym___block] = ACTIONS(6535),
    [anon_sym___bridge] = ACTIONS(6535),
    [anon_sym___bridge_retained] = ACTIONS(6535),
    [anon_sym___bridge_transfer] = ACTIONS(6535),
    [anon_sym___complex] = ACTIONS(6535),
    [anon_sym___const] = ACTIONS(6535),
    [anon_sym___imag] = ACTIONS(6535),
    [anon_sym___kindof] = ACTIONS(6535),
    [anon_sym___nonnull] = ACTIONS(6535),
    [anon_sym___nullable] = ACTIONS(6535),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6535),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6535),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6535),
    [anon_sym___real] = ACTIONS(6535),
    [anon_sym___strong] = ACTIONS(6535),
    [anon_sym___unsafe_unretained] = ACTIONS(6535),
    [anon_sym___unused] = ACTIONS(6535),
    [anon_sym___weak] = ACTIONS(6535),
    [anon_sym_alignas] = ACTIONS(6535),
    [anon_sym__Alignas] = ACTIONS(6535),
    [sym_primitive_type] = ACTIONS(6535),
    [anon_sym_enum] = ACTIONS(6535),
    [anon_sym_struct] = ACTIONS(6535),
    [anon_sym_union] = ACTIONS(6535),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6535),
    [anon_sym___typeof] = ACTIONS(6535),
    [anon_sym_typeof] = ACTIONS(6535),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6535),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6535),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6535),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6535),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6535),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6535),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6535),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6535),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6535),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_API_AVAILABLE] = ACTIONS(6535),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_API_DEPRECATED] = ACTIONS(6535),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6535),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6535),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6535),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6535),
    [anon_sym___deprecated_msg] = ACTIONS(6535),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6535),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6535),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6535),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6535),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6535),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6535),
    [anon_sym_BOOL] = ACTIONS(6535),
    [anon_sym_IMP] = ACTIONS(6535),
    [anon_sym_SEL] = ACTIONS(6535),
    [anon_sym_Class] = ACTIONS(6535),
    [anon_sym_id] = ACTIONS(6535),
    [anon_sym_ATdefs] = ACTIONS(6537),
  },
  [STATE(2776)] = {
    [sym_identifier] = ACTIONS(6752),
    [anon_sym_COMMA] = ACTIONS(6754),
    [anon_sym_LPAREN2] = ACTIONS(6754),
    [anon_sym_SEMI] = ACTIONS(6754),
    [anon_sym___extension__] = ACTIONS(6752),
    [anon_sym_extern] = ACTIONS(6752),
    [anon_sym___attribute__] = ACTIONS(6752),
    [anon_sym___attribute] = ACTIONS(6752),
    [anon_sym_noreturn] = ACTIONS(6752),
    [anon_sym_LBRACK] = ACTIONS(6754),
    [anon_sym___declspec] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6754),
    [anon_sym_signed] = ACTIONS(6752),
    [anon_sym_unsigned] = ACTIONS(6752),
    [anon_sym_long] = ACTIONS(6752),
    [anon_sym_short] = ACTIONS(6752),
    [anon_sym_static] = ACTIONS(6752),
    [anon_sym_EQ] = ACTIONS(6754),
    [anon_sym_ATautoreleasepool] = ACTIONS(6754),
    [anon_sym_auto] = ACTIONS(6752),
    [anon_sym_register] = ACTIONS(6752),
    [anon_sym_inline] = ACTIONS(6752),
    [anon_sym___inline] = ACTIONS(6752),
    [anon_sym___inline__] = ACTIONS(6752),
    [anon_sym___forceinline] = ACTIONS(6752),
    [anon_sym_thread_local] = ACTIONS(6752),
    [anon_sym___thread] = ACTIONS(6752),
    [anon_sym_CG_EXTERN] = ACTIONS(6752),
    [anon_sym_CG_INLINE] = ACTIONS(6752),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6752),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6752),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6752),
    [anon_sym_IBOutlet] = ACTIONS(6752),
    [anon_sym_IBInspectable] = ACTIONS(6752),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6752),
    [anon_sym_NS_INLINE] = ACTIONS(6752),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6752),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6752),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6752),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6752),
    [anon_sym_const] = ACTIONS(6752),
    [anon_sym_constexpr] = ACTIONS(6752),
    [anon_sym_volatile] = ACTIONS(6752),
    [anon_sym_restrict] = ACTIONS(6752),
    [anon_sym___restrict__] = ACTIONS(6752),
    [anon_sym__Atomic] = ACTIONS(6752),
    [anon_sym__Noreturn] = ACTIONS(6752),
    [anon_sym__Nonnull] = ACTIONS(6752),
    [anon_sym_nullable] = ACTIONS(6752),
    [anon_sym__Complex] = ACTIONS(6752),
    [anon_sym__Nullable] = ACTIONS(6752),
    [anon_sym__Nullable_result] = ACTIONS(6752),
    [anon_sym__Null_unspecified] = ACTIONS(6752),
    [anon_sym___autoreleasing] = ACTIONS(6752),
    [anon_sym___block] = ACTIONS(6752),
    [anon_sym___bridge] = ACTIONS(6752),
    [anon_sym___bridge_retained] = ACTIONS(6752),
    [anon_sym___bridge_transfer] = ACTIONS(6752),
    [anon_sym___complex] = ACTIONS(6752),
    [anon_sym___const] = ACTIONS(6752),
    [anon_sym___imag] = ACTIONS(6752),
    [anon_sym___kindof] = ACTIONS(6752),
    [anon_sym___nonnull] = ACTIONS(6752),
    [anon_sym___nullable] = ACTIONS(6752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6752),
    [anon_sym___real] = ACTIONS(6752),
    [anon_sym___strong] = ACTIONS(6752),
    [anon_sym___unsafe_unretained] = ACTIONS(6752),
    [anon_sym___unused] = ACTIONS(6752),
    [anon_sym___weak] = ACTIONS(6752),
    [anon_sym_alignas] = ACTIONS(6752),
    [anon_sym__Alignas] = ACTIONS(6752),
    [sym_primitive_type] = ACTIONS(6752),
    [anon_sym_enum] = ACTIONS(6752),
    [anon_sym_struct] = ACTIONS(6752),
    [anon_sym_union] = ACTIONS(6752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6752),
    [anon_sym___typeof] = ACTIONS(6752),
    [anon_sym_typeof] = ACTIONS(6752),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6752),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6752),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6752),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6752),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6752),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6752),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6752),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6752),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6752),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6752),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6752),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6752),
    [anon_sym_API_AVAILABLE] = ACTIONS(6752),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6752),
    [anon_sym_API_DEPRECATED] = ACTIONS(6752),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6752),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6752),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6752),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6752),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6752),
    [anon_sym___deprecated_msg] = ACTIONS(6752),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6752),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6752),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6752),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6752),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6752),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6752),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6752),
    [anon_sym_BOOL] = ACTIONS(6752),
    [anon_sym_IMP] = ACTIONS(6752),
    [anon_sym_SEL] = ACTIONS(6752),
    [anon_sym_Class] = ACTIONS(6752),
    [anon_sym_id] = ACTIONS(6752),
  },
  [STATE(2777)] = {
    [sym_identifier] = ACTIONS(6652),
    [anon_sym_LPAREN2] = ACTIONS(6654),
    [anon_sym___extension__] = ACTIONS(6652),
    [anon_sym_extern] = ACTIONS(6652),
    [anon_sym___attribute__] = ACTIONS(6656),
    [anon_sym___attribute] = ACTIONS(6656),
    [anon_sym_noreturn] = ACTIONS(6652),
    [anon_sym_LBRACK] = ACTIONS(6659),
    [anon_sym___declspec] = ACTIONS(6652),
    [anon_sym_LBRACE] = ACTIONS(6654),
    [anon_sym_signed] = ACTIONS(6652),
    [anon_sym_unsigned] = ACTIONS(6652),
    [anon_sym_long] = ACTIONS(6652),
    [anon_sym_short] = ACTIONS(6652),
    [anon_sym_static] = ACTIONS(6652),
    [anon_sym_ATautoreleasepool] = ACTIONS(6654),
    [anon_sym_auto] = ACTIONS(6652),
    [anon_sym_register] = ACTIONS(6652),
    [anon_sym_inline] = ACTIONS(6652),
    [anon_sym___inline] = ACTIONS(6652),
    [anon_sym___inline__] = ACTIONS(6652),
    [anon_sym___forceinline] = ACTIONS(6652),
    [anon_sym_thread_local] = ACTIONS(6652),
    [anon_sym___thread] = ACTIONS(6652),
    [anon_sym_CG_EXTERN] = ACTIONS(6652),
    [anon_sym_CG_INLINE] = ACTIONS(6652),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6652),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6652),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6652),
    [anon_sym_IBOutlet] = ACTIONS(6652),
    [anon_sym_IBInspectable] = ACTIONS(6652),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6652),
    [anon_sym_NS_INLINE] = ACTIONS(6652),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6652),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6652),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6652),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6652),
    [anon_sym_const] = ACTIONS(6652),
    [anon_sym_constexpr] = ACTIONS(6652),
    [anon_sym_volatile] = ACTIONS(6652),
    [anon_sym_restrict] = ACTIONS(6652),
    [anon_sym___restrict__] = ACTIONS(6652),
    [anon_sym__Atomic] = ACTIONS(6652),
    [anon_sym__Noreturn] = ACTIONS(6652),
    [anon_sym__Nonnull] = ACTIONS(6652),
    [anon_sym_nullable] = ACTIONS(6652),
    [anon_sym__Complex] = ACTIONS(6652),
    [anon_sym__Nullable] = ACTIONS(6652),
    [anon_sym__Nullable_result] = ACTIONS(6652),
    [anon_sym__Null_unspecified] = ACTIONS(6652),
    [anon_sym___autoreleasing] = ACTIONS(6652),
    [anon_sym___block] = ACTIONS(6652),
    [anon_sym___bridge] = ACTIONS(6652),
    [anon_sym___bridge_retained] = ACTIONS(6652),
    [anon_sym___bridge_transfer] = ACTIONS(6652),
    [anon_sym___complex] = ACTIONS(6652),
    [anon_sym___const] = ACTIONS(6652),
    [anon_sym___imag] = ACTIONS(6652),
    [anon_sym___kindof] = ACTIONS(6652),
    [anon_sym___nonnull] = ACTIONS(6652),
    [anon_sym___nullable] = ACTIONS(6652),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6652),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6652),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6652),
    [anon_sym___real] = ACTIONS(6652),
    [anon_sym___strong] = ACTIONS(6652),
    [anon_sym___unsafe_unretained] = ACTIONS(6652),
    [anon_sym___unused] = ACTIONS(6652),
    [anon_sym___weak] = ACTIONS(6652),
    [anon_sym_alignas] = ACTIONS(6652),
    [anon_sym__Alignas] = ACTIONS(6652),
    [sym_primitive_type] = ACTIONS(6652),
    [anon_sym_enum] = ACTIONS(6652),
    [anon_sym_struct] = ACTIONS(6652),
    [anon_sym_union] = ACTIONS(6652),
    [anon_sym_asm] = ACTIONS(6662),
    [anon_sym___asm__] = ACTIONS(6662),
    [anon_sym___asm] = ACTIONS(6662),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6652),
    [anon_sym___typeof] = ACTIONS(6652),
    [anon_sym_typeof] = ACTIONS(6652),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6652),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6652),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6652),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6652),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6652),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6652),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6652),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6652),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6652),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6652),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6652),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6652),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6652),
    [anon_sym_API_AVAILABLE] = ACTIONS(6652),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6652),
    [anon_sym_API_DEPRECATED] = ACTIONS(6652),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6652),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6652),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6652),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6652),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6652),
    [anon_sym___deprecated_msg] = ACTIONS(6652),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6652),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6652),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6652),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6652),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6652),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6652),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6652),
    [anon_sym_BOOL] = ACTIONS(6652),
    [anon_sym_IMP] = ACTIONS(6652),
    [anon_sym_SEL] = ACTIONS(6652),
    [anon_sym_Class] = ACTIONS(6652),
    [anon_sym_id] = ACTIONS(6652),
  },
  [STATE(2778)] = {
    [sym_identifier] = ACTIONS(6141),
    [anon_sym_COMMA] = ACTIONS(6128),
    [anon_sym_LPAREN2] = ACTIONS(6128),
    [anon_sym_SEMI] = ACTIONS(6128),
    [anon_sym___extension__] = ACTIONS(6141),
    [anon_sym_extern] = ACTIONS(6141),
    [anon_sym___attribute__] = ACTIONS(6141),
    [anon_sym___attribute] = ACTIONS(6141),
    [anon_sym_noreturn] = ACTIONS(6141),
    [anon_sym_LBRACK] = ACTIONS(6128),
    [anon_sym___declspec] = ACTIONS(6141),
    [anon_sym_LBRACE] = ACTIONS(6128),
    [anon_sym_signed] = ACTIONS(6141),
    [anon_sym_unsigned] = ACTIONS(6141),
    [anon_sym_long] = ACTIONS(6141),
    [anon_sym_short] = ACTIONS(6141),
    [anon_sym_static] = ACTIONS(6141),
    [anon_sym_EQ] = ACTIONS(6128),
    [anon_sym_ATautoreleasepool] = ACTIONS(6128),
    [anon_sym_auto] = ACTIONS(6141),
    [anon_sym_register] = ACTIONS(6141),
    [anon_sym_inline] = ACTIONS(6141),
    [anon_sym___inline] = ACTIONS(6141),
    [anon_sym___inline__] = ACTIONS(6141),
    [anon_sym___forceinline] = ACTIONS(6141),
    [anon_sym_thread_local] = ACTIONS(6141),
    [anon_sym___thread] = ACTIONS(6141),
    [anon_sym_CG_EXTERN] = ACTIONS(6141),
    [anon_sym_CG_INLINE] = ACTIONS(6141),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6141),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6141),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6141),
    [anon_sym_IBOutlet] = ACTIONS(6141),
    [anon_sym_IBInspectable] = ACTIONS(6141),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6141),
    [anon_sym_NS_INLINE] = ACTIONS(6141),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6141),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6141),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6141),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6141),
    [anon_sym_const] = ACTIONS(6141),
    [anon_sym_constexpr] = ACTIONS(6141),
    [anon_sym_volatile] = ACTIONS(6141),
    [anon_sym_restrict] = ACTIONS(6141),
    [anon_sym___restrict__] = ACTIONS(6141),
    [anon_sym__Atomic] = ACTIONS(6141),
    [anon_sym__Noreturn] = ACTIONS(6141),
    [anon_sym__Nonnull] = ACTIONS(6141),
    [anon_sym_nullable] = ACTIONS(6141),
    [anon_sym__Complex] = ACTIONS(6141),
    [anon_sym__Nullable] = ACTIONS(6141),
    [anon_sym__Nullable_result] = ACTIONS(6141),
    [anon_sym__Null_unspecified] = ACTIONS(6141),
    [anon_sym___autoreleasing] = ACTIONS(6141),
    [anon_sym___block] = ACTIONS(6141),
    [anon_sym___bridge] = ACTIONS(6141),
    [anon_sym___bridge_retained] = ACTIONS(6141),
    [anon_sym___bridge_transfer] = ACTIONS(6141),
    [anon_sym___complex] = ACTIONS(6141),
    [anon_sym___const] = ACTIONS(6141),
    [anon_sym___imag] = ACTIONS(6141),
    [anon_sym___kindof] = ACTIONS(6141),
    [anon_sym___nonnull] = ACTIONS(6141),
    [anon_sym___nullable] = ACTIONS(6141),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6141),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6141),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6141),
    [anon_sym___real] = ACTIONS(6141),
    [anon_sym___strong] = ACTIONS(6141),
    [anon_sym___unsafe_unretained] = ACTIONS(6141),
    [anon_sym___unused] = ACTIONS(6141),
    [anon_sym___weak] = ACTIONS(6141),
    [anon_sym_alignas] = ACTIONS(6141),
    [anon_sym__Alignas] = ACTIONS(6141),
    [sym_primitive_type] = ACTIONS(6141),
    [anon_sym_enum] = ACTIONS(6141),
    [anon_sym_struct] = ACTIONS(6141),
    [anon_sym_union] = ACTIONS(6141),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6141),
    [anon_sym___typeof] = ACTIONS(6141),
    [anon_sym_typeof] = ACTIONS(6141),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6141),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6141),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6141),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6141),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6141),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6141),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6141),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6141),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6141),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6141),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6141),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6141),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6141),
    [anon_sym_API_AVAILABLE] = ACTIONS(6141),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6141),
    [anon_sym_API_DEPRECATED] = ACTIONS(6141),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6141),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6141),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6141),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6141),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6141),
    [anon_sym___deprecated_msg] = ACTIONS(6141),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6141),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6141),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6141),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6141),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6141),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6141),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6141),
    [anon_sym_BOOL] = ACTIONS(6141),
    [anon_sym_IMP] = ACTIONS(6141),
    [anon_sym_SEL] = ACTIONS(6141),
    [anon_sym_Class] = ACTIONS(6141),
    [anon_sym_id] = ACTIONS(6141),
  },
  [STATE(2779)] = {
    [sym_identifier] = ACTIONS(6756),
    [anon_sym_COMMA] = ACTIONS(6758),
    [anon_sym_LPAREN2] = ACTIONS(6758),
    [anon_sym_SEMI] = ACTIONS(6758),
    [anon_sym___extension__] = ACTIONS(6756),
    [anon_sym_extern] = ACTIONS(6756),
    [anon_sym___attribute__] = ACTIONS(6756),
    [anon_sym___attribute] = ACTIONS(6756),
    [anon_sym_noreturn] = ACTIONS(6756),
    [anon_sym_LBRACK] = ACTIONS(6758),
    [anon_sym___declspec] = ACTIONS(6756),
    [anon_sym_LBRACE] = ACTIONS(6758),
    [anon_sym_signed] = ACTIONS(6756),
    [anon_sym_unsigned] = ACTIONS(6756),
    [anon_sym_long] = ACTIONS(6756),
    [anon_sym_short] = ACTIONS(6756),
    [anon_sym_static] = ACTIONS(6756),
    [anon_sym_EQ] = ACTIONS(6758),
    [anon_sym_ATautoreleasepool] = ACTIONS(6758),
    [anon_sym_auto] = ACTIONS(6756),
    [anon_sym_register] = ACTIONS(6756),
    [anon_sym_inline] = ACTIONS(6756),
    [anon_sym___inline] = ACTIONS(6756),
    [anon_sym___inline__] = ACTIONS(6756),
    [anon_sym___forceinline] = ACTIONS(6756),
    [anon_sym_thread_local] = ACTIONS(6756),
    [anon_sym___thread] = ACTIONS(6756),
    [anon_sym_CG_EXTERN] = ACTIONS(6756),
    [anon_sym_CG_INLINE] = ACTIONS(6756),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6756),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6756),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6756),
    [anon_sym_IBOutlet] = ACTIONS(6756),
    [anon_sym_IBInspectable] = ACTIONS(6756),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6756),
    [anon_sym_NS_INLINE] = ACTIONS(6756),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6756),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6756),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6756),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6756),
    [anon_sym_const] = ACTIONS(6756),
    [anon_sym_constexpr] = ACTIONS(6756),
    [anon_sym_volatile] = ACTIONS(6756),
    [anon_sym_restrict] = ACTIONS(6756),
    [anon_sym___restrict__] = ACTIONS(6756),
    [anon_sym__Atomic] = ACTIONS(6756),
    [anon_sym__Noreturn] = ACTIONS(6756),
    [anon_sym__Nonnull] = ACTIONS(6756),
    [anon_sym_nullable] = ACTIONS(6756),
    [anon_sym__Complex] = ACTIONS(6756),
    [anon_sym__Nullable] = ACTIONS(6756),
    [anon_sym__Nullable_result] = ACTIONS(6756),
    [anon_sym__Null_unspecified] = ACTIONS(6756),
    [anon_sym___autoreleasing] = ACTIONS(6756),
    [anon_sym___block] = ACTIONS(6756),
    [anon_sym___bridge] = ACTIONS(6756),
    [anon_sym___bridge_retained] = ACTIONS(6756),
    [anon_sym___bridge_transfer] = ACTIONS(6756),
    [anon_sym___complex] = ACTIONS(6756),
    [anon_sym___const] = ACTIONS(6756),
    [anon_sym___imag] = ACTIONS(6756),
    [anon_sym___kindof] = ACTIONS(6756),
    [anon_sym___nonnull] = ACTIONS(6756),
    [anon_sym___nullable] = ACTIONS(6756),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6756),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6756),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6756),
    [anon_sym___real] = ACTIONS(6756),
    [anon_sym___strong] = ACTIONS(6756),
    [anon_sym___unsafe_unretained] = ACTIONS(6756),
    [anon_sym___unused] = ACTIONS(6756),
    [anon_sym___weak] = ACTIONS(6756),
    [anon_sym_alignas] = ACTIONS(6756),
    [anon_sym__Alignas] = ACTIONS(6756),
    [sym_primitive_type] = ACTIONS(6756),
    [anon_sym_enum] = ACTIONS(6756),
    [anon_sym_struct] = ACTIONS(6756),
    [anon_sym_union] = ACTIONS(6756),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6756),
    [anon_sym___typeof] = ACTIONS(6756),
    [anon_sym_typeof] = ACTIONS(6756),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6756),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6756),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6756),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6756),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6756),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6756),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6756),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6756),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6756),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6756),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6756),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6756),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6756),
    [anon_sym_API_AVAILABLE] = ACTIONS(6756),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6756),
    [anon_sym_API_DEPRECATED] = ACTIONS(6756),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6756),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6756),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6756),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6756),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6756),
    [anon_sym___deprecated_msg] = ACTIONS(6756),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6756),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6756),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6756),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6756),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6756),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6756),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6756),
    [anon_sym_BOOL] = ACTIONS(6756),
    [anon_sym_IMP] = ACTIONS(6756),
    [anon_sym_SEL] = ACTIONS(6756),
    [anon_sym_Class] = ACTIONS(6756),
    [anon_sym_id] = ACTIONS(6756),
  },
  [STATE(2780)] = {
    [sym_identifier] = ACTIONS(6760),
    [anon_sym_COMMA] = ACTIONS(6762),
    [anon_sym_LPAREN2] = ACTIONS(6762),
    [anon_sym_SEMI] = ACTIONS(6762),
    [anon_sym___extension__] = ACTIONS(6760),
    [anon_sym_extern] = ACTIONS(6760),
    [anon_sym___attribute__] = ACTIONS(6760),
    [anon_sym___attribute] = ACTIONS(6760),
    [anon_sym_noreturn] = ACTIONS(6760),
    [anon_sym_LBRACK] = ACTIONS(6762),
    [anon_sym___declspec] = ACTIONS(6760),
    [anon_sym_LBRACE] = ACTIONS(6762),
    [anon_sym_signed] = ACTIONS(6760),
    [anon_sym_unsigned] = ACTIONS(6760),
    [anon_sym_long] = ACTIONS(6760),
    [anon_sym_short] = ACTIONS(6760),
    [anon_sym_static] = ACTIONS(6760),
    [anon_sym_EQ] = ACTIONS(6762),
    [anon_sym_ATautoreleasepool] = ACTIONS(6762),
    [anon_sym_auto] = ACTIONS(6760),
    [anon_sym_register] = ACTIONS(6760),
    [anon_sym_inline] = ACTIONS(6760),
    [anon_sym___inline] = ACTIONS(6760),
    [anon_sym___inline__] = ACTIONS(6760),
    [anon_sym___forceinline] = ACTIONS(6760),
    [anon_sym_thread_local] = ACTIONS(6760),
    [anon_sym___thread] = ACTIONS(6760),
    [anon_sym_CG_EXTERN] = ACTIONS(6760),
    [anon_sym_CG_INLINE] = ACTIONS(6760),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6760),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6760),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6760),
    [anon_sym_IBOutlet] = ACTIONS(6760),
    [anon_sym_IBInspectable] = ACTIONS(6760),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6760),
    [anon_sym_NS_INLINE] = ACTIONS(6760),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6760),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6760),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6760),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6760),
    [anon_sym_const] = ACTIONS(6760),
    [anon_sym_constexpr] = ACTIONS(6760),
    [anon_sym_volatile] = ACTIONS(6760),
    [anon_sym_restrict] = ACTIONS(6760),
    [anon_sym___restrict__] = ACTIONS(6760),
    [anon_sym__Atomic] = ACTIONS(6760),
    [anon_sym__Noreturn] = ACTIONS(6760),
    [anon_sym__Nonnull] = ACTIONS(6760),
    [anon_sym_nullable] = ACTIONS(6760),
    [anon_sym__Complex] = ACTIONS(6760),
    [anon_sym__Nullable] = ACTIONS(6760),
    [anon_sym__Nullable_result] = ACTIONS(6760),
    [anon_sym__Null_unspecified] = ACTIONS(6760),
    [anon_sym___autoreleasing] = ACTIONS(6760),
    [anon_sym___block] = ACTIONS(6760),
    [anon_sym___bridge] = ACTIONS(6760),
    [anon_sym___bridge_retained] = ACTIONS(6760),
    [anon_sym___bridge_transfer] = ACTIONS(6760),
    [anon_sym___complex] = ACTIONS(6760),
    [anon_sym___const] = ACTIONS(6760),
    [anon_sym___imag] = ACTIONS(6760),
    [anon_sym___kindof] = ACTIONS(6760),
    [anon_sym___nonnull] = ACTIONS(6760),
    [anon_sym___nullable] = ACTIONS(6760),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6760),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6760),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6760),
    [anon_sym___real] = ACTIONS(6760),
    [anon_sym___strong] = ACTIONS(6760),
    [anon_sym___unsafe_unretained] = ACTIONS(6760),
    [anon_sym___unused] = ACTIONS(6760),
    [anon_sym___weak] = ACTIONS(6760),
    [anon_sym_alignas] = ACTIONS(6760),
    [anon_sym__Alignas] = ACTIONS(6760),
    [sym_primitive_type] = ACTIONS(6760),
    [anon_sym_enum] = ACTIONS(6760),
    [anon_sym_struct] = ACTIONS(6760),
    [anon_sym_union] = ACTIONS(6760),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6760),
    [anon_sym___typeof] = ACTIONS(6760),
    [anon_sym_typeof] = ACTIONS(6760),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6760),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6760),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6760),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6760),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6760),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6760),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6760),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6760),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6760),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6760),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6760),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6760),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6760),
    [anon_sym_API_AVAILABLE] = ACTIONS(6760),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6760),
    [anon_sym_API_DEPRECATED] = ACTIONS(6760),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6760),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6760),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6760),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6760),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6760),
    [anon_sym___deprecated_msg] = ACTIONS(6760),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6760),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6760),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6760),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6760),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6760),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6760),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6760),
    [anon_sym_BOOL] = ACTIONS(6760),
    [anon_sym_IMP] = ACTIONS(6760),
    [anon_sym_SEL] = ACTIONS(6760),
    [anon_sym_Class] = ACTIONS(6760),
    [anon_sym_id] = ACTIONS(6760),
  },
  [STATE(2781)] = {
    [sym_identifier] = ACTIONS(6664),
    [anon_sym_LPAREN2] = ACTIONS(6650),
    [anon_sym___extension__] = ACTIONS(6664),
    [anon_sym_extern] = ACTIONS(6664),
    [anon_sym___attribute__] = ACTIONS(6666),
    [anon_sym___attribute] = ACTIONS(6666),
    [anon_sym_noreturn] = ACTIONS(6664),
    [anon_sym_LBRACK] = ACTIONS(6669),
    [anon_sym___declspec] = ACTIONS(6664),
    [anon_sym_LBRACE] = ACTIONS(6650),
    [anon_sym_signed] = ACTIONS(6664),
    [anon_sym_unsigned] = ACTIONS(6664),
    [anon_sym_long] = ACTIONS(6664),
    [anon_sym_short] = ACTIONS(6664),
    [anon_sym_static] = ACTIONS(6664),
    [anon_sym_ATautoreleasepool] = ACTIONS(6650),
    [anon_sym_auto] = ACTIONS(6664),
    [anon_sym_register] = ACTIONS(6664),
    [anon_sym_inline] = ACTIONS(6664),
    [anon_sym___inline] = ACTIONS(6664),
    [anon_sym___inline__] = ACTIONS(6664),
    [anon_sym___forceinline] = ACTIONS(6664),
    [anon_sym_thread_local] = ACTIONS(6664),
    [anon_sym___thread] = ACTIONS(6664),
    [anon_sym_CG_EXTERN] = ACTIONS(6664),
    [anon_sym_CG_INLINE] = ACTIONS(6664),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6664),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6664),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6664),
    [anon_sym_IBOutlet] = ACTIONS(6664),
    [anon_sym_IBInspectable] = ACTIONS(6664),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6664),
    [anon_sym_NS_INLINE] = ACTIONS(6664),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6664),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6664),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6664),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6664),
    [anon_sym_const] = ACTIONS(6664),
    [anon_sym_constexpr] = ACTIONS(6664),
    [anon_sym_volatile] = ACTIONS(6664),
    [anon_sym_restrict] = ACTIONS(6664),
    [anon_sym___restrict__] = ACTIONS(6664),
    [anon_sym__Atomic] = ACTIONS(6664),
    [anon_sym__Noreturn] = ACTIONS(6664),
    [anon_sym__Nonnull] = ACTIONS(6664),
    [anon_sym_nullable] = ACTIONS(6664),
    [anon_sym__Complex] = ACTIONS(6664),
    [anon_sym__Nullable] = ACTIONS(6664),
    [anon_sym__Nullable_result] = ACTIONS(6664),
    [anon_sym__Null_unspecified] = ACTIONS(6664),
    [anon_sym___autoreleasing] = ACTIONS(6664),
    [anon_sym___block] = ACTIONS(6664),
    [anon_sym___bridge] = ACTIONS(6664),
    [anon_sym___bridge_retained] = ACTIONS(6664),
    [anon_sym___bridge_transfer] = ACTIONS(6664),
    [anon_sym___complex] = ACTIONS(6664),
    [anon_sym___const] = ACTIONS(6664),
    [anon_sym___imag] = ACTIONS(6664),
    [anon_sym___kindof] = ACTIONS(6664),
    [anon_sym___nonnull] = ACTIONS(6664),
    [anon_sym___nullable] = ACTIONS(6664),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6664),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6664),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6664),
    [anon_sym___real] = ACTIONS(6664),
    [anon_sym___strong] = ACTIONS(6664),
    [anon_sym___unsafe_unretained] = ACTIONS(6664),
    [anon_sym___unused] = ACTIONS(6664),
    [anon_sym___weak] = ACTIONS(6664),
    [anon_sym_alignas] = ACTIONS(6664),
    [anon_sym__Alignas] = ACTIONS(6664),
    [sym_primitive_type] = ACTIONS(6664),
    [anon_sym_enum] = ACTIONS(6664),
    [anon_sym_struct] = ACTIONS(6664),
    [anon_sym_union] = ACTIONS(6664),
    [anon_sym_asm] = ACTIONS(6648),
    [anon_sym___asm__] = ACTIONS(6648),
    [anon_sym___asm] = ACTIONS(6648),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6664),
    [anon_sym___typeof] = ACTIONS(6664),
    [anon_sym_typeof] = ACTIONS(6664),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6664),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6664),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6664),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6664),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6664),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6664),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6664),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6664),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6664),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6664),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6664),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6664),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6664),
    [anon_sym_API_AVAILABLE] = ACTIONS(6664),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6664),
    [anon_sym_API_DEPRECATED] = ACTIONS(6664),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6664),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6664),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6664),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6664),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6664),
    [anon_sym___deprecated_msg] = ACTIONS(6664),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6664),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6664),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6664),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6664),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6664),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6664),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6664),
    [anon_sym_BOOL] = ACTIONS(6664),
    [anon_sym_IMP] = ACTIONS(6664),
    [anon_sym_SEL] = ACTIONS(6664),
    [anon_sym_Class] = ACTIONS(6664),
    [anon_sym_id] = ACTIONS(6664),
  },
  [STATE(2782)] = {
    [sym__declaration_modifiers] = STATE(2783),
    [sym_attribute_specifier] = STATE(2783),
    [sym_attribute_declaration] = STATE(2783),
    [sym_ms_declspec_modifier] = STATE(2783),
    [sym__abstract_declarator] = STATE(4978),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(2783),
    [sym_type_qualifier] = STATE(2783),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4271),
    [sym_availability_attribute_specifier] = STATE(2783),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2783),
    [anon_sym_COMMA] = ACTIONS(5314),
    [anon_sym_RPAREN] = ACTIONS(5314),
    [anon_sym_LPAREN2] = ACTIONS(4488),
    [anon_sym_STAR] = ACTIONS(4490),
    [anon_sym_CARET] = ACTIONS(4492),
    [anon_sym_GT] = ACTIONS(5314),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6770),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_COLON] = ACTIONS(5314),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2783)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym__abstract_declarator] = STATE(4999),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4271),
    [sym_availability_attribute_specifier] = STATE(2864),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [anon_sym_COMMA] = ACTIONS(6361),
    [anon_sym_RPAREN] = ACTIONS(6361),
    [anon_sym_LPAREN2] = ACTIONS(4488),
    [anon_sym_STAR] = ACTIONS(4490),
    [anon_sym_CARET] = ACTIONS(4492),
    [anon_sym_GT] = ACTIONS(6361),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6770),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_COLON] = ACTIONS(6361),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2784)] = {
    [sym_identifier] = ACTIONS(6780),
    [anon_sym_COMMA] = ACTIONS(6782),
    [anon_sym_LPAREN2] = ACTIONS(6782),
    [anon_sym_SEMI] = ACTIONS(6782),
    [anon_sym___extension__] = ACTIONS(6780),
    [anon_sym_extern] = ACTIONS(6780),
    [anon_sym___attribute__] = ACTIONS(6780),
    [anon_sym___attribute] = ACTIONS(6780),
    [anon_sym_noreturn] = ACTIONS(6780),
    [anon_sym_LBRACK] = ACTIONS(6782),
    [anon_sym___declspec] = ACTIONS(6780),
    [anon_sym_LBRACE] = ACTIONS(6782),
    [anon_sym_signed] = ACTIONS(6780),
    [anon_sym_unsigned] = ACTIONS(6780),
    [anon_sym_long] = ACTIONS(6780),
    [anon_sym_short] = ACTIONS(6780),
    [anon_sym_static] = ACTIONS(6780),
    [anon_sym_EQ] = ACTIONS(6782),
    [anon_sym_ATautoreleasepool] = ACTIONS(6782),
    [anon_sym_auto] = ACTIONS(6780),
    [anon_sym_register] = ACTIONS(6780),
    [anon_sym_inline] = ACTIONS(6780),
    [anon_sym___inline] = ACTIONS(6780),
    [anon_sym___inline__] = ACTIONS(6780),
    [anon_sym___forceinline] = ACTIONS(6780),
    [anon_sym_thread_local] = ACTIONS(6780),
    [anon_sym___thread] = ACTIONS(6780),
    [anon_sym_CG_EXTERN] = ACTIONS(6780),
    [anon_sym_CG_INLINE] = ACTIONS(6780),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6780),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6780),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6780),
    [anon_sym_IBOutlet] = ACTIONS(6780),
    [anon_sym_IBInspectable] = ACTIONS(6780),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6780),
    [anon_sym_NS_INLINE] = ACTIONS(6780),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6780),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6780),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6780),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6780),
    [anon_sym_const] = ACTIONS(6780),
    [anon_sym_constexpr] = ACTIONS(6780),
    [anon_sym_volatile] = ACTIONS(6780),
    [anon_sym_restrict] = ACTIONS(6780),
    [anon_sym___restrict__] = ACTIONS(6780),
    [anon_sym__Atomic] = ACTIONS(6780),
    [anon_sym__Noreturn] = ACTIONS(6780),
    [anon_sym__Nonnull] = ACTIONS(6780),
    [anon_sym_nullable] = ACTIONS(6780),
    [anon_sym__Complex] = ACTIONS(6780),
    [anon_sym__Nullable] = ACTIONS(6780),
    [anon_sym__Nullable_result] = ACTIONS(6780),
    [anon_sym__Null_unspecified] = ACTIONS(6780),
    [anon_sym___autoreleasing] = ACTIONS(6780),
    [anon_sym___block] = ACTIONS(6780),
    [anon_sym___bridge] = ACTIONS(6780),
    [anon_sym___bridge_retained] = ACTIONS(6780),
    [anon_sym___bridge_transfer] = ACTIONS(6780),
    [anon_sym___complex] = ACTIONS(6780),
    [anon_sym___const] = ACTIONS(6780),
    [anon_sym___imag] = ACTIONS(6780),
    [anon_sym___kindof] = ACTIONS(6780),
    [anon_sym___nonnull] = ACTIONS(6780),
    [anon_sym___nullable] = ACTIONS(6780),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6780),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6780),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6780),
    [anon_sym___real] = ACTIONS(6780),
    [anon_sym___strong] = ACTIONS(6780),
    [anon_sym___unsafe_unretained] = ACTIONS(6780),
    [anon_sym___unused] = ACTIONS(6780),
    [anon_sym___weak] = ACTIONS(6780),
    [anon_sym_alignas] = ACTIONS(6780),
    [anon_sym__Alignas] = ACTIONS(6780),
    [sym_primitive_type] = ACTIONS(6780),
    [anon_sym_enum] = ACTIONS(6780),
    [anon_sym_struct] = ACTIONS(6780),
    [anon_sym_union] = ACTIONS(6780),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6780),
    [anon_sym___typeof] = ACTIONS(6780),
    [anon_sym_typeof] = ACTIONS(6780),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6780),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6780),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6780),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6780),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6780),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6780),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6780),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6780),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6780),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6780),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6780),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6780),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6780),
    [anon_sym_API_AVAILABLE] = ACTIONS(6780),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6780),
    [anon_sym_API_DEPRECATED] = ACTIONS(6780),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6780),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6780),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6780),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6780),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6780),
    [anon_sym___deprecated_msg] = ACTIONS(6780),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6780),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6780),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6780),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6780),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6780),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6780),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6780),
    [anon_sym_BOOL] = ACTIONS(6780),
    [anon_sym_IMP] = ACTIONS(6780),
    [anon_sym_SEL] = ACTIONS(6780),
    [anon_sym_Class] = ACTIONS(6780),
    [anon_sym_id] = ACTIONS(6780),
  },
  [STATE(2785)] = {
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6786),
    [anon_sym_SEMI] = ACTIONS(6786),
    [anon_sym___extension__] = ACTIONS(6784),
    [anon_sym_extern] = ACTIONS(6784),
    [anon_sym___attribute__] = ACTIONS(6784),
    [anon_sym___attribute] = ACTIONS(6784),
    [anon_sym_noreturn] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6786),
    [anon_sym___declspec] = ACTIONS(6784),
    [anon_sym_LBRACE] = ACTIONS(6786),
    [anon_sym_signed] = ACTIONS(6784),
    [anon_sym_unsigned] = ACTIONS(6784),
    [anon_sym_long] = ACTIONS(6784),
    [anon_sym_short] = ACTIONS(6784),
    [anon_sym_static] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6786),
    [anon_sym_ATautoreleasepool] = ACTIONS(6786),
    [anon_sym_auto] = ACTIONS(6784),
    [anon_sym_register] = ACTIONS(6784),
    [anon_sym_inline] = ACTIONS(6784),
    [anon_sym___inline] = ACTIONS(6784),
    [anon_sym___inline__] = ACTIONS(6784),
    [anon_sym___forceinline] = ACTIONS(6784),
    [anon_sym_thread_local] = ACTIONS(6784),
    [anon_sym___thread] = ACTIONS(6784),
    [anon_sym_CG_EXTERN] = ACTIONS(6784),
    [anon_sym_CG_INLINE] = ACTIONS(6784),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6784),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6784),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6784),
    [anon_sym_IBOutlet] = ACTIONS(6784),
    [anon_sym_IBInspectable] = ACTIONS(6784),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6784),
    [anon_sym_NS_INLINE] = ACTIONS(6784),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6784),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6784),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6784),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6784),
    [anon_sym_const] = ACTIONS(6784),
    [anon_sym_constexpr] = ACTIONS(6784),
    [anon_sym_volatile] = ACTIONS(6784),
    [anon_sym_restrict] = ACTIONS(6784),
    [anon_sym___restrict__] = ACTIONS(6784),
    [anon_sym__Atomic] = ACTIONS(6784),
    [anon_sym__Noreturn] = ACTIONS(6784),
    [anon_sym__Nonnull] = ACTIONS(6784),
    [anon_sym_nullable] = ACTIONS(6784),
    [anon_sym__Complex] = ACTIONS(6784),
    [anon_sym__Nullable] = ACTIONS(6784),
    [anon_sym__Nullable_result] = ACTIONS(6784),
    [anon_sym__Null_unspecified] = ACTIONS(6784),
    [anon_sym___autoreleasing] = ACTIONS(6784),
    [anon_sym___block] = ACTIONS(6784),
    [anon_sym___bridge] = ACTIONS(6784),
    [anon_sym___bridge_retained] = ACTIONS(6784),
    [anon_sym___bridge_transfer] = ACTIONS(6784),
    [anon_sym___complex] = ACTIONS(6784),
    [anon_sym___const] = ACTIONS(6784),
    [anon_sym___imag] = ACTIONS(6784),
    [anon_sym___kindof] = ACTIONS(6784),
    [anon_sym___nonnull] = ACTIONS(6784),
    [anon_sym___nullable] = ACTIONS(6784),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6784),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6784),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6784),
    [anon_sym___real] = ACTIONS(6784),
    [anon_sym___strong] = ACTIONS(6784),
    [anon_sym___unsafe_unretained] = ACTIONS(6784),
    [anon_sym___unused] = ACTIONS(6784),
    [anon_sym___weak] = ACTIONS(6784),
    [anon_sym_alignas] = ACTIONS(6784),
    [anon_sym__Alignas] = ACTIONS(6784),
    [sym_primitive_type] = ACTIONS(6784),
    [anon_sym_enum] = ACTIONS(6784),
    [anon_sym_struct] = ACTIONS(6784),
    [anon_sym_union] = ACTIONS(6784),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6784),
    [anon_sym___typeof] = ACTIONS(6784),
    [anon_sym_typeof] = ACTIONS(6784),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6784),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6784),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6784),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6784),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6784),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6784),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6784),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6784),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6784),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6784),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6784),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6784),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6784),
    [anon_sym_API_AVAILABLE] = ACTIONS(6784),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6784),
    [anon_sym_API_DEPRECATED] = ACTIONS(6784),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6784),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6784),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6784),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6784),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6784),
    [anon_sym___deprecated_msg] = ACTIONS(6784),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6784),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6784),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6784),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6784),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6784),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6784),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6784),
    [anon_sym_BOOL] = ACTIONS(6784),
    [anon_sym_IMP] = ACTIONS(6784),
    [anon_sym_SEL] = ACTIONS(6784),
    [anon_sym_Class] = ACTIONS(6784),
    [anon_sym_id] = ACTIONS(6784),
  },
  [STATE(2786)] = {
    [sym_identifier] = ACTIONS(6788),
    [anon_sym_COMMA] = ACTIONS(6790),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_SEMI] = ACTIONS(6790),
    [anon_sym___extension__] = ACTIONS(6788),
    [anon_sym_extern] = ACTIONS(6788),
    [anon_sym___attribute__] = ACTIONS(6788),
    [anon_sym___attribute] = ACTIONS(6788),
    [anon_sym_noreturn] = ACTIONS(6788),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym___declspec] = ACTIONS(6788),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_signed] = ACTIONS(6788),
    [anon_sym_unsigned] = ACTIONS(6788),
    [anon_sym_long] = ACTIONS(6788),
    [anon_sym_short] = ACTIONS(6788),
    [anon_sym_static] = ACTIONS(6788),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_ATautoreleasepool] = ACTIONS(6790),
    [anon_sym_auto] = ACTIONS(6788),
    [anon_sym_register] = ACTIONS(6788),
    [anon_sym_inline] = ACTIONS(6788),
    [anon_sym___inline] = ACTIONS(6788),
    [anon_sym___inline__] = ACTIONS(6788),
    [anon_sym___forceinline] = ACTIONS(6788),
    [anon_sym_thread_local] = ACTIONS(6788),
    [anon_sym___thread] = ACTIONS(6788),
    [anon_sym_CG_EXTERN] = ACTIONS(6788),
    [anon_sym_CG_INLINE] = ACTIONS(6788),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6788),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6788),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6788),
    [anon_sym_IBOutlet] = ACTIONS(6788),
    [anon_sym_IBInspectable] = ACTIONS(6788),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6788),
    [anon_sym_NS_INLINE] = ACTIONS(6788),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6788),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6788),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6788),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6788),
    [anon_sym_const] = ACTIONS(6788),
    [anon_sym_constexpr] = ACTIONS(6788),
    [anon_sym_volatile] = ACTIONS(6788),
    [anon_sym_restrict] = ACTIONS(6788),
    [anon_sym___restrict__] = ACTIONS(6788),
    [anon_sym__Atomic] = ACTIONS(6788),
    [anon_sym__Noreturn] = ACTIONS(6788),
    [anon_sym__Nonnull] = ACTIONS(6788),
    [anon_sym_nullable] = ACTIONS(6788),
    [anon_sym__Complex] = ACTIONS(6788),
    [anon_sym__Nullable] = ACTIONS(6788),
    [anon_sym__Nullable_result] = ACTIONS(6788),
    [anon_sym__Null_unspecified] = ACTIONS(6788),
    [anon_sym___autoreleasing] = ACTIONS(6788),
    [anon_sym___block] = ACTIONS(6788),
    [anon_sym___bridge] = ACTIONS(6788),
    [anon_sym___bridge_retained] = ACTIONS(6788),
    [anon_sym___bridge_transfer] = ACTIONS(6788),
    [anon_sym___complex] = ACTIONS(6788),
    [anon_sym___const] = ACTIONS(6788),
    [anon_sym___imag] = ACTIONS(6788),
    [anon_sym___kindof] = ACTIONS(6788),
    [anon_sym___nonnull] = ACTIONS(6788),
    [anon_sym___nullable] = ACTIONS(6788),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6788),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6788),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6788),
    [anon_sym___real] = ACTIONS(6788),
    [anon_sym___strong] = ACTIONS(6788),
    [anon_sym___unsafe_unretained] = ACTIONS(6788),
    [anon_sym___unused] = ACTIONS(6788),
    [anon_sym___weak] = ACTIONS(6788),
    [anon_sym_alignas] = ACTIONS(6788),
    [anon_sym__Alignas] = ACTIONS(6788),
    [sym_primitive_type] = ACTIONS(6788),
    [anon_sym_enum] = ACTIONS(6788),
    [anon_sym_struct] = ACTIONS(6788),
    [anon_sym_union] = ACTIONS(6788),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6788),
    [anon_sym___typeof] = ACTIONS(6788),
    [anon_sym_typeof] = ACTIONS(6788),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6788),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6788),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6788),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6788),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6788),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6788),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6788),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6788),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6788),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6788),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6788),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6788),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6788),
    [anon_sym_API_AVAILABLE] = ACTIONS(6788),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6788),
    [anon_sym_API_DEPRECATED] = ACTIONS(6788),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6788),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6788),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6788),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6788),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6788),
    [anon_sym___deprecated_msg] = ACTIONS(6788),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6788),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6788),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6788),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6788),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6788),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6788),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6788),
    [anon_sym_BOOL] = ACTIONS(6788),
    [anon_sym_IMP] = ACTIONS(6788),
    [anon_sym_SEL] = ACTIONS(6788),
    [anon_sym_Class] = ACTIONS(6788),
    [anon_sym_id] = ACTIONS(6788),
  },
  [STATE(2787)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6222),
    [sym__field_declarator] = STATE(4865),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2788)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6257),
    [sym__field_declarator] = STATE(5028),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6401),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2789)] = {
    [aux_sym_c_method_parameter_repeat1] = STATE(2791),
    [sym_identifier] = ACTIONS(6626),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_SEMI] = ACTIONS(6628),
    [anon_sym___extension__] = ACTIONS(6626),
    [anon_sym_extern] = ACTIONS(6626),
    [anon_sym___attribute__] = ACTIONS(6626),
    [anon_sym___attribute] = ACTIONS(6626),
    [anon_sym_noreturn] = ACTIONS(6626),
    [anon_sym_LBRACK] = ACTIONS(6628),
    [anon_sym___declspec] = ACTIONS(6626),
    [anon_sym_LBRACE] = ACTIONS(6628),
    [anon_sym_signed] = ACTIONS(6626),
    [anon_sym_unsigned] = ACTIONS(6626),
    [anon_sym_long] = ACTIONS(6626),
    [anon_sym_short] = ACTIONS(6626),
    [anon_sym_static] = ACTIONS(6626),
    [anon_sym_ATautoreleasepool] = ACTIONS(6628),
    [anon_sym_auto] = ACTIONS(6626),
    [anon_sym_register] = ACTIONS(6626),
    [anon_sym_inline] = ACTIONS(6626),
    [anon_sym___inline] = ACTIONS(6626),
    [anon_sym___inline__] = ACTIONS(6626),
    [anon_sym___forceinline] = ACTIONS(6626),
    [anon_sym_thread_local] = ACTIONS(6626),
    [anon_sym___thread] = ACTIONS(6626),
    [anon_sym_CG_EXTERN] = ACTIONS(6626),
    [anon_sym_CG_INLINE] = ACTIONS(6626),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6626),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6626),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6626),
    [anon_sym_IBOutlet] = ACTIONS(6626),
    [anon_sym_IBInspectable] = ACTIONS(6626),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6626),
    [anon_sym_NS_INLINE] = ACTIONS(6626),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6626),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6626),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6626),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6626),
    [anon_sym_const] = ACTIONS(6626),
    [anon_sym_constexpr] = ACTIONS(6626),
    [anon_sym_volatile] = ACTIONS(6626),
    [anon_sym_restrict] = ACTIONS(6626),
    [anon_sym___restrict__] = ACTIONS(6626),
    [anon_sym__Atomic] = ACTIONS(6626),
    [anon_sym__Noreturn] = ACTIONS(6626),
    [anon_sym__Nonnull] = ACTIONS(6626),
    [anon_sym_nullable] = ACTIONS(6626),
    [anon_sym__Complex] = ACTIONS(6626),
    [anon_sym__Nullable] = ACTIONS(6626),
    [anon_sym__Nullable_result] = ACTIONS(6626),
    [anon_sym__Null_unspecified] = ACTIONS(6626),
    [anon_sym___autoreleasing] = ACTIONS(6626),
    [anon_sym___block] = ACTIONS(6626),
    [anon_sym___bridge] = ACTIONS(6626),
    [anon_sym___bridge_retained] = ACTIONS(6626),
    [anon_sym___bridge_transfer] = ACTIONS(6626),
    [anon_sym___complex] = ACTIONS(6626),
    [anon_sym___const] = ACTIONS(6626),
    [anon_sym___imag] = ACTIONS(6626),
    [anon_sym___kindof] = ACTIONS(6626),
    [anon_sym___nonnull] = ACTIONS(6626),
    [anon_sym___nullable] = ACTIONS(6626),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6626),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6626),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6626),
    [anon_sym___real] = ACTIONS(6626),
    [anon_sym___strong] = ACTIONS(6626),
    [anon_sym___unsafe_unretained] = ACTIONS(6626),
    [anon_sym___unused] = ACTIONS(6626),
    [anon_sym___weak] = ACTIONS(6626),
    [anon_sym_alignas] = ACTIONS(6626),
    [anon_sym__Alignas] = ACTIONS(6626),
    [sym_primitive_type] = ACTIONS(6626),
    [anon_sym_enum] = ACTIONS(6626),
    [anon_sym_struct] = ACTIONS(6626),
    [anon_sym_union] = ACTIONS(6626),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6626),
    [anon_sym___typeof] = ACTIONS(6626),
    [anon_sym_typeof] = ACTIONS(6626),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6626),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6626),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6626),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6626),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6626),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6626),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6626),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6626),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6626),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6626),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6626),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6626),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6626),
    [anon_sym_API_AVAILABLE] = ACTIONS(6626),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6626),
    [anon_sym_API_DEPRECATED] = ACTIONS(6626),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6626),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6626),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6626),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6626),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6626),
    [anon_sym___deprecated_msg] = ACTIONS(6626),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6626),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6626),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6626),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6626),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6626),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6626),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6626),
    [anon_sym_BOOL] = ACTIONS(6626),
    [anon_sym_IMP] = ACTIONS(6626),
    [anon_sym_SEL] = ACTIONS(6626),
    [anon_sym_Class] = ACTIONS(6626),
    [anon_sym_id] = ACTIONS(6626),
  },
  [STATE(2790)] = {
    [sym_identifier] = ACTIONS(6680),
    [anon_sym_COMMA] = ACTIONS(6675),
    [anon_sym_SEMI] = ACTIONS(6675),
    [anon_sym___extension__] = ACTIONS(6680),
    [anon_sym_extern] = ACTIONS(6680),
    [anon_sym___attribute__] = ACTIONS(6680),
    [anon_sym___attribute] = ACTIONS(6680),
    [anon_sym_noreturn] = ACTIONS(6680),
    [anon_sym_LBRACK] = ACTIONS(6675),
    [anon_sym___declspec] = ACTIONS(6680),
    [anon_sym_LBRACE] = ACTIONS(6675),
    [anon_sym_signed] = ACTIONS(6680),
    [anon_sym_unsigned] = ACTIONS(6680),
    [anon_sym_long] = ACTIONS(6680),
    [anon_sym_short] = ACTIONS(6680),
    [anon_sym_static] = ACTIONS(6680),
    [anon_sym_ATautoreleasepool] = ACTIONS(6675),
    [anon_sym_auto] = ACTIONS(6680),
    [anon_sym_register] = ACTIONS(6680),
    [anon_sym_inline] = ACTIONS(6680),
    [anon_sym___inline] = ACTIONS(6680),
    [anon_sym___inline__] = ACTIONS(6680),
    [anon_sym___forceinline] = ACTIONS(6680),
    [anon_sym_thread_local] = ACTIONS(6680),
    [anon_sym___thread] = ACTIONS(6680),
    [anon_sym_CG_EXTERN] = ACTIONS(6680),
    [anon_sym_CG_INLINE] = ACTIONS(6680),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6680),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6680),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6680),
    [anon_sym_IBOutlet] = ACTIONS(6680),
    [anon_sym_IBInspectable] = ACTIONS(6680),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6680),
    [anon_sym_NS_INLINE] = ACTIONS(6680),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6680),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6680),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6680),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6680),
    [anon_sym_const] = ACTIONS(6680),
    [anon_sym_constexpr] = ACTIONS(6680),
    [anon_sym_volatile] = ACTIONS(6680),
    [anon_sym_restrict] = ACTIONS(6680),
    [anon_sym___restrict__] = ACTIONS(6680),
    [anon_sym__Atomic] = ACTIONS(6680),
    [anon_sym__Noreturn] = ACTIONS(6680),
    [anon_sym__Nonnull] = ACTIONS(6680),
    [anon_sym_nullable] = ACTIONS(6680),
    [anon_sym__Complex] = ACTIONS(6680),
    [anon_sym__Nullable] = ACTIONS(6680),
    [anon_sym__Nullable_result] = ACTIONS(6680),
    [anon_sym__Null_unspecified] = ACTIONS(6680),
    [anon_sym___autoreleasing] = ACTIONS(6680),
    [anon_sym___block] = ACTIONS(6680),
    [anon_sym___bridge] = ACTIONS(6680),
    [anon_sym___bridge_retained] = ACTIONS(6680),
    [anon_sym___bridge_transfer] = ACTIONS(6680),
    [anon_sym___complex] = ACTIONS(6680),
    [anon_sym___const] = ACTIONS(6680),
    [anon_sym___imag] = ACTIONS(6680),
    [anon_sym___kindof] = ACTIONS(6680),
    [anon_sym___nonnull] = ACTIONS(6680),
    [anon_sym___nullable] = ACTIONS(6680),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6680),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6680),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6680),
    [anon_sym___real] = ACTIONS(6680),
    [anon_sym___strong] = ACTIONS(6680),
    [anon_sym___unsafe_unretained] = ACTIONS(6680),
    [anon_sym___unused] = ACTIONS(6680),
    [anon_sym___weak] = ACTIONS(6680),
    [anon_sym_alignas] = ACTIONS(6680),
    [anon_sym__Alignas] = ACTIONS(6680),
    [sym_primitive_type] = ACTIONS(6680),
    [anon_sym_enum] = ACTIONS(6680),
    [anon_sym_COLON] = ACTIONS(6675),
    [anon_sym_struct] = ACTIONS(6680),
    [anon_sym_union] = ACTIONS(6680),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6680),
    [anon_sym___typeof] = ACTIONS(6680),
    [anon_sym_typeof] = ACTIONS(6680),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6680),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6680),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6680),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6680),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6680),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6680),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6680),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6680),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6680),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6680),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6680),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6680),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6680),
    [anon_sym_API_AVAILABLE] = ACTIONS(6680),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6680),
    [anon_sym_API_DEPRECATED] = ACTIONS(6680),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6680),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6680),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6680),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6680),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6680),
    [anon_sym___deprecated_msg] = ACTIONS(6680),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6680),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6680),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6680),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6680),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6680),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6680),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6680),
    [anon_sym_BOOL] = ACTIONS(6680),
    [anon_sym_IMP] = ACTIONS(6680),
    [anon_sym_SEL] = ACTIONS(6680),
    [anon_sym_Class] = ACTIONS(6680),
    [anon_sym_id] = ACTIONS(6680),
  },
  [STATE(2791)] = {
    [aux_sym_c_method_parameter_repeat1] = STATE(2809),
    [sym_identifier] = ACTIONS(6792),
    [anon_sym_COMMA] = ACTIONS(6794),
    [anon_sym_SEMI] = ACTIONS(6794),
    [anon_sym___extension__] = ACTIONS(6792),
    [anon_sym_extern] = ACTIONS(6792),
    [anon_sym___attribute__] = ACTIONS(6792),
    [anon_sym___attribute] = ACTIONS(6792),
    [anon_sym_noreturn] = ACTIONS(6792),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym___declspec] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_signed] = ACTIONS(6792),
    [anon_sym_unsigned] = ACTIONS(6792),
    [anon_sym_long] = ACTIONS(6792),
    [anon_sym_short] = ACTIONS(6792),
    [anon_sym_static] = ACTIONS(6792),
    [anon_sym_ATautoreleasepool] = ACTIONS(6794),
    [anon_sym_auto] = ACTIONS(6792),
    [anon_sym_register] = ACTIONS(6792),
    [anon_sym_inline] = ACTIONS(6792),
    [anon_sym___inline] = ACTIONS(6792),
    [anon_sym___inline__] = ACTIONS(6792),
    [anon_sym___forceinline] = ACTIONS(6792),
    [anon_sym_thread_local] = ACTIONS(6792),
    [anon_sym___thread] = ACTIONS(6792),
    [anon_sym_CG_EXTERN] = ACTIONS(6792),
    [anon_sym_CG_INLINE] = ACTIONS(6792),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6792),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6792),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6792),
    [anon_sym_IBOutlet] = ACTIONS(6792),
    [anon_sym_IBInspectable] = ACTIONS(6792),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6792),
    [anon_sym_NS_INLINE] = ACTIONS(6792),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6792),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6792),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6792),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6792),
    [anon_sym_const] = ACTIONS(6792),
    [anon_sym_constexpr] = ACTIONS(6792),
    [anon_sym_volatile] = ACTIONS(6792),
    [anon_sym_restrict] = ACTIONS(6792),
    [anon_sym___restrict__] = ACTIONS(6792),
    [anon_sym__Atomic] = ACTIONS(6792),
    [anon_sym__Noreturn] = ACTIONS(6792),
    [anon_sym__Nonnull] = ACTIONS(6792),
    [anon_sym_nullable] = ACTIONS(6792),
    [anon_sym__Complex] = ACTIONS(6792),
    [anon_sym__Nullable] = ACTIONS(6792),
    [anon_sym__Nullable_result] = ACTIONS(6792),
    [anon_sym__Null_unspecified] = ACTIONS(6792),
    [anon_sym___autoreleasing] = ACTIONS(6792),
    [anon_sym___block] = ACTIONS(6792),
    [anon_sym___bridge] = ACTIONS(6792),
    [anon_sym___bridge_retained] = ACTIONS(6792),
    [anon_sym___bridge_transfer] = ACTIONS(6792),
    [anon_sym___complex] = ACTIONS(6792),
    [anon_sym___const] = ACTIONS(6792),
    [anon_sym___imag] = ACTIONS(6792),
    [anon_sym___kindof] = ACTIONS(6792),
    [anon_sym___nonnull] = ACTIONS(6792),
    [anon_sym___nullable] = ACTIONS(6792),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6792),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6792),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6792),
    [anon_sym___real] = ACTIONS(6792),
    [anon_sym___strong] = ACTIONS(6792),
    [anon_sym___unsafe_unretained] = ACTIONS(6792),
    [anon_sym___unused] = ACTIONS(6792),
    [anon_sym___weak] = ACTIONS(6792),
    [anon_sym_alignas] = ACTIONS(6792),
    [anon_sym__Alignas] = ACTIONS(6792),
    [sym_primitive_type] = ACTIONS(6792),
    [anon_sym_enum] = ACTIONS(6792),
    [anon_sym_struct] = ACTIONS(6792),
    [anon_sym_union] = ACTIONS(6792),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6792),
    [anon_sym___typeof] = ACTIONS(6792),
    [anon_sym_typeof] = ACTIONS(6792),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6792),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6792),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6792),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6792),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6792),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6792),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6792),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6792),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6792),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6792),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6792),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6792),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6792),
    [anon_sym_API_AVAILABLE] = ACTIONS(6792),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6792),
    [anon_sym_API_DEPRECATED] = ACTIONS(6792),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6792),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6792),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6792),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6792),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6792),
    [anon_sym___deprecated_msg] = ACTIONS(6792),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6792),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6792),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6792),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6792),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6792),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6792),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6792),
    [anon_sym_BOOL] = ACTIONS(6792),
    [anon_sym_IMP] = ACTIONS(6792),
    [anon_sym_SEL] = ACTIONS(6792),
    [anon_sym_Class] = ACTIONS(6792),
    [anon_sym_id] = ACTIONS(6792),
  },
  [STATE(2792)] = {
    [sym__declaration_modifiers] = STATE(6399),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6399),
    [sym_ms_declspec_modifier] = STATE(6399),
    [sym_compound_statement] = STATE(861),
    [sym_storage_class_specifier] = STATE(6399),
    [sym_type_qualifier] = STATE(6399),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym__old_style_parameter_list] = STATE(2906),
    [sym_gnu_asm_expression] = STATE(3160),
    [sym_availability_attribute_specifier] = STATE(6399),
    [aux_sym_declaration_repeat1] = STATE(3161),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6798),
    [anon_sym_SEMI] = ACTIONS(6800),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2793)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6257),
    [sym__field_declarator] = STATE(5082),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6401),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2794)] = {
    [sym__declaration_modifiers] = STATE(6348),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6348),
    [sym_ms_declspec_modifier] = STATE(6348),
    [sym_compound_statement] = STATE(517),
    [sym_storage_class_specifier] = STATE(6348),
    [sym_type_qualifier] = STATE(6348),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym__old_style_parameter_list] = STATE(2906),
    [sym_gnu_asm_expression] = STATE(3076),
    [sym_availability_attribute_specifier] = STATE(6348),
    [aux_sym_declaration_repeat1] = STATE(3077),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6798),
    [anon_sym_SEMI] = ACTIONS(6808),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2795)] = {
    [sym_identifier] = ACTIONS(6810),
    [anon_sym_COMMA] = ACTIONS(6812),
    [anon_sym_SEMI] = ACTIONS(6812),
    [anon_sym___extension__] = ACTIONS(6810),
    [anon_sym_extern] = ACTIONS(6810),
    [anon_sym___attribute__] = ACTIONS(6810),
    [anon_sym___attribute] = ACTIONS(6810),
    [anon_sym_noreturn] = ACTIONS(6810),
    [anon_sym_LBRACK] = ACTIONS(6812),
    [anon_sym___declspec] = ACTIONS(6810),
    [anon_sym_LBRACE] = ACTIONS(6812),
    [anon_sym_signed] = ACTIONS(6810),
    [anon_sym_unsigned] = ACTIONS(6810),
    [anon_sym_long] = ACTIONS(6810),
    [anon_sym_short] = ACTIONS(6810),
    [anon_sym_static] = ACTIONS(6810),
    [anon_sym_ATautoreleasepool] = ACTIONS(6812),
    [anon_sym_auto] = ACTIONS(6810),
    [anon_sym_register] = ACTIONS(6810),
    [anon_sym_inline] = ACTIONS(6810),
    [anon_sym___inline] = ACTIONS(6810),
    [anon_sym___inline__] = ACTIONS(6810),
    [anon_sym___forceinline] = ACTIONS(6810),
    [anon_sym_thread_local] = ACTIONS(6810),
    [anon_sym___thread] = ACTIONS(6810),
    [anon_sym_CG_EXTERN] = ACTIONS(6810),
    [anon_sym_CG_INLINE] = ACTIONS(6810),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6810),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6810),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6810),
    [anon_sym_IBOutlet] = ACTIONS(6810),
    [anon_sym_IBInspectable] = ACTIONS(6810),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6810),
    [anon_sym_NS_INLINE] = ACTIONS(6810),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6810),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6810),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6810),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6810),
    [anon_sym_const] = ACTIONS(6810),
    [anon_sym_constexpr] = ACTIONS(6810),
    [anon_sym_volatile] = ACTIONS(6810),
    [anon_sym_restrict] = ACTIONS(6810),
    [anon_sym___restrict__] = ACTIONS(6810),
    [anon_sym__Atomic] = ACTIONS(6810),
    [anon_sym__Noreturn] = ACTIONS(6810),
    [anon_sym__Nonnull] = ACTIONS(6810),
    [anon_sym_nullable] = ACTIONS(6810),
    [anon_sym__Complex] = ACTIONS(6810),
    [anon_sym__Nullable] = ACTIONS(6810),
    [anon_sym__Nullable_result] = ACTIONS(6810),
    [anon_sym__Null_unspecified] = ACTIONS(6810),
    [anon_sym___autoreleasing] = ACTIONS(6810),
    [anon_sym___block] = ACTIONS(6810),
    [anon_sym___bridge] = ACTIONS(6810),
    [anon_sym___bridge_retained] = ACTIONS(6810),
    [anon_sym___bridge_transfer] = ACTIONS(6810),
    [anon_sym___complex] = ACTIONS(6810),
    [anon_sym___const] = ACTIONS(6810),
    [anon_sym___imag] = ACTIONS(6810),
    [anon_sym___kindof] = ACTIONS(6810),
    [anon_sym___nonnull] = ACTIONS(6810),
    [anon_sym___nullable] = ACTIONS(6810),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6810),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6810),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6810),
    [anon_sym___real] = ACTIONS(6810),
    [anon_sym___strong] = ACTIONS(6810),
    [anon_sym___unsafe_unretained] = ACTIONS(6810),
    [anon_sym___unused] = ACTIONS(6810),
    [anon_sym___weak] = ACTIONS(6810),
    [anon_sym_alignas] = ACTIONS(6810),
    [anon_sym__Alignas] = ACTIONS(6810),
    [sym_primitive_type] = ACTIONS(6810),
    [anon_sym_enum] = ACTIONS(6810),
    [anon_sym_COLON] = ACTIONS(6812),
    [anon_sym_struct] = ACTIONS(6810),
    [anon_sym_union] = ACTIONS(6810),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6810),
    [anon_sym___typeof] = ACTIONS(6810),
    [anon_sym_typeof] = ACTIONS(6810),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6810),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6810),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6810),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6810),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6810),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6810),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6810),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6810),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6810),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6810),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6810),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6810),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6810),
    [anon_sym_API_AVAILABLE] = ACTIONS(6810),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6810),
    [anon_sym_API_DEPRECATED] = ACTIONS(6810),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6810),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6810),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6810),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6810),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6810),
    [anon_sym___deprecated_msg] = ACTIONS(6810),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6810),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6810),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6810),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6810),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6810),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6810),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6810),
    [anon_sym_BOOL] = ACTIONS(6810),
    [anon_sym_IMP] = ACTIONS(6810),
    [anon_sym_SEL] = ACTIONS(6810),
    [anon_sym_Class] = ACTIONS(6810),
    [anon_sym_id] = ACTIONS(6810),
  },
  [STATE(2796)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6257),
    [sym__field_declarator] = STATE(5078),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6401),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2797)] = {
    [sym__declaration_modifiers] = STATE(6444),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6444),
    [sym_ms_declspec_modifier] = STATE(6444),
    [sym_compound_statement] = STATE(1022),
    [sym_storage_class_specifier] = STATE(6444),
    [sym_type_qualifier] = STATE(6444),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym__old_style_parameter_list] = STATE(2906),
    [sym_gnu_asm_expression] = STATE(3118),
    [sym_availability_attribute_specifier] = STATE(6444),
    [aux_sym_declaration_repeat1] = STATE(3117),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6798),
    [anon_sym_SEMI] = ACTIONS(6814),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2798)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6222),
    [sym__field_declarator] = STATE(4867),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2799)] = {
    [sym__declaration_modifiers] = STATE(6102),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6102),
    [sym_ms_declspec_modifier] = STATE(6102),
    [sym_compound_statement] = STATE(280),
    [sym_storage_class_specifier] = STATE(6102),
    [sym_type_qualifier] = STATE(6102),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym__old_style_parameter_list] = STATE(2906),
    [sym_gnu_asm_expression] = STATE(3162),
    [sym_availability_attribute_specifier] = STATE(6102),
    [aux_sym_declaration_repeat1] = STATE(3163),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6798),
    [anon_sym_SEMI] = ACTIONS(6816),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2800)] = {
    [sym_identifier] = ACTIONS(6818),
    [anon_sym_COMMA] = ACTIONS(6820),
    [anon_sym_SEMI] = ACTIONS(6820),
    [anon_sym___extension__] = ACTIONS(6818),
    [anon_sym_extern] = ACTIONS(6818),
    [anon_sym___attribute__] = ACTIONS(6818),
    [anon_sym___attribute] = ACTIONS(6818),
    [anon_sym_noreturn] = ACTIONS(6818),
    [anon_sym_LBRACK] = ACTIONS(6820),
    [anon_sym___declspec] = ACTIONS(6818),
    [anon_sym_LBRACE] = ACTIONS(6820),
    [anon_sym_signed] = ACTIONS(6818),
    [anon_sym_unsigned] = ACTIONS(6818),
    [anon_sym_long] = ACTIONS(6818),
    [anon_sym_short] = ACTIONS(6818),
    [anon_sym_static] = ACTIONS(6818),
    [anon_sym_ATautoreleasepool] = ACTIONS(6820),
    [anon_sym_auto] = ACTIONS(6818),
    [anon_sym_register] = ACTIONS(6818),
    [anon_sym_inline] = ACTIONS(6818),
    [anon_sym___inline] = ACTIONS(6818),
    [anon_sym___inline__] = ACTIONS(6818),
    [anon_sym___forceinline] = ACTIONS(6818),
    [anon_sym_thread_local] = ACTIONS(6818),
    [anon_sym___thread] = ACTIONS(6818),
    [anon_sym_CG_EXTERN] = ACTIONS(6818),
    [anon_sym_CG_INLINE] = ACTIONS(6818),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6818),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6818),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6818),
    [anon_sym_IBOutlet] = ACTIONS(6818),
    [anon_sym_IBInspectable] = ACTIONS(6818),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6818),
    [anon_sym_NS_INLINE] = ACTIONS(6818),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6818),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6818),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6818),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6818),
    [anon_sym_const] = ACTIONS(6818),
    [anon_sym_constexpr] = ACTIONS(6818),
    [anon_sym_volatile] = ACTIONS(6818),
    [anon_sym_restrict] = ACTIONS(6818),
    [anon_sym___restrict__] = ACTIONS(6818),
    [anon_sym__Atomic] = ACTIONS(6818),
    [anon_sym__Noreturn] = ACTIONS(6818),
    [anon_sym__Nonnull] = ACTIONS(6818),
    [anon_sym_nullable] = ACTIONS(6818),
    [anon_sym__Complex] = ACTIONS(6818),
    [anon_sym__Nullable] = ACTIONS(6818),
    [anon_sym__Nullable_result] = ACTIONS(6818),
    [anon_sym__Null_unspecified] = ACTIONS(6818),
    [anon_sym___autoreleasing] = ACTIONS(6818),
    [anon_sym___block] = ACTIONS(6818),
    [anon_sym___bridge] = ACTIONS(6818),
    [anon_sym___bridge_retained] = ACTIONS(6818),
    [anon_sym___bridge_transfer] = ACTIONS(6818),
    [anon_sym___complex] = ACTIONS(6818),
    [anon_sym___const] = ACTIONS(6818),
    [anon_sym___imag] = ACTIONS(6818),
    [anon_sym___kindof] = ACTIONS(6818),
    [anon_sym___nonnull] = ACTIONS(6818),
    [anon_sym___nullable] = ACTIONS(6818),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6818),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6818),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6818),
    [anon_sym___real] = ACTIONS(6818),
    [anon_sym___strong] = ACTIONS(6818),
    [anon_sym___unsafe_unretained] = ACTIONS(6818),
    [anon_sym___unused] = ACTIONS(6818),
    [anon_sym___weak] = ACTIONS(6818),
    [anon_sym_alignas] = ACTIONS(6818),
    [anon_sym__Alignas] = ACTIONS(6818),
    [sym_primitive_type] = ACTIONS(6818),
    [anon_sym_enum] = ACTIONS(6818),
    [anon_sym_COLON] = ACTIONS(6820),
    [anon_sym_struct] = ACTIONS(6818),
    [anon_sym_union] = ACTIONS(6818),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6818),
    [anon_sym___typeof] = ACTIONS(6818),
    [anon_sym_typeof] = ACTIONS(6818),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6818),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6818),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6818),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6818),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6818),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6818),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6818),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6818),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6818),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6818),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6818),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6818),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6818),
    [anon_sym_API_AVAILABLE] = ACTIONS(6818),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6818),
    [anon_sym_API_DEPRECATED] = ACTIONS(6818),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6818),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6818),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6818),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6818),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6818),
    [anon_sym___deprecated_msg] = ACTIONS(6818),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6818),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6818),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6818),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6818),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6818),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6818),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6818),
    [anon_sym_BOOL] = ACTIONS(6818),
    [anon_sym_IMP] = ACTIONS(6818),
    [anon_sym_SEL] = ACTIONS(6818),
    [anon_sym_Class] = ACTIONS(6818),
    [anon_sym_id] = ACTIONS(6818),
  },
  [STATE(2801)] = {
    [sym__declaration_modifiers] = STATE(6602),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6602),
    [sym_ms_declspec_modifier] = STATE(6602),
    [sym_compound_statement] = STATE(645),
    [sym_storage_class_specifier] = STATE(6602),
    [sym_type_qualifier] = STATE(6602),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym__old_style_parameter_list] = STATE(2906),
    [sym_gnu_asm_expression] = STATE(3085),
    [sym_availability_attribute_specifier] = STATE(6602),
    [aux_sym_declaration_repeat1] = STATE(3086),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6798),
    [anon_sym_SEMI] = ACTIONS(6822),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2802)] = {
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6826),
    [anon_sym_SEMI] = ACTIONS(6826),
    [anon_sym___extension__] = ACTIONS(6824),
    [anon_sym_extern] = ACTIONS(6824),
    [anon_sym___attribute__] = ACTIONS(6824),
    [anon_sym___attribute] = ACTIONS(6824),
    [anon_sym_noreturn] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6826),
    [anon_sym___declspec] = ACTIONS(6824),
    [anon_sym_LBRACE] = ACTIONS(6826),
    [anon_sym_signed] = ACTIONS(6824),
    [anon_sym_unsigned] = ACTIONS(6824),
    [anon_sym_long] = ACTIONS(6824),
    [anon_sym_short] = ACTIONS(6824),
    [anon_sym_static] = ACTIONS(6824),
    [anon_sym_ATautoreleasepool] = ACTIONS(6826),
    [anon_sym_auto] = ACTIONS(6824),
    [anon_sym_register] = ACTIONS(6824),
    [anon_sym_inline] = ACTIONS(6824),
    [anon_sym___inline] = ACTIONS(6824),
    [anon_sym___inline__] = ACTIONS(6824),
    [anon_sym___forceinline] = ACTIONS(6824),
    [anon_sym_thread_local] = ACTIONS(6824),
    [anon_sym___thread] = ACTIONS(6824),
    [anon_sym_CG_EXTERN] = ACTIONS(6824),
    [anon_sym_CG_INLINE] = ACTIONS(6824),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6824),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6824),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6824),
    [anon_sym_IBOutlet] = ACTIONS(6824),
    [anon_sym_IBInspectable] = ACTIONS(6824),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6824),
    [anon_sym_NS_INLINE] = ACTIONS(6824),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6824),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6824),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6824),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6824),
    [anon_sym_const] = ACTIONS(6824),
    [anon_sym_constexpr] = ACTIONS(6824),
    [anon_sym_volatile] = ACTIONS(6824),
    [anon_sym_restrict] = ACTIONS(6824),
    [anon_sym___restrict__] = ACTIONS(6824),
    [anon_sym__Atomic] = ACTIONS(6824),
    [anon_sym__Noreturn] = ACTIONS(6824),
    [anon_sym__Nonnull] = ACTIONS(6824),
    [anon_sym_nullable] = ACTIONS(6824),
    [anon_sym__Complex] = ACTIONS(6824),
    [anon_sym__Nullable] = ACTIONS(6824),
    [anon_sym__Nullable_result] = ACTIONS(6824),
    [anon_sym__Null_unspecified] = ACTIONS(6824),
    [anon_sym___autoreleasing] = ACTIONS(6824),
    [anon_sym___block] = ACTIONS(6824),
    [anon_sym___bridge] = ACTIONS(6824),
    [anon_sym___bridge_retained] = ACTIONS(6824),
    [anon_sym___bridge_transfer] = ACTIONS(6824),
    [anon_sym___complex] = ACTIONS(6824),
    [anon_sym___const] = ACTIONS(6824),
    [anon_sym___imag] = ACTIONS(6824),
    [anon_sym___kindof] = ACTIONS(6824),
    [anon_sym___nonnull] = ACTIONS(6824),
    [anon_sym___nullable] = ACTIONS(6824),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6824),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6824),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6824),
    [anon_sym___real] = ACTIONS(6824),
    [anon_sym___strong] = ACTIONS(6824),
    [anon_sym___unsafe_unretained] = ACTIONS(6824),
    [anon_sym___unused] = ACTIONS(6824),
    [anon_sym___weak] = ACTIONS(6824),
    [anon_sym_alignas] = ACTIONS(6824),
    [anon_sym__Alignas] = ACTIONS(6824),
    [sym_primitive_type] = ACTIONS(6824),
    [anon_sym_enum] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6826),
    [anon_sym_struct] = ACTIONS(6824),
    [anon_sym_union] = ACTIONS(6824),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6824),
    [anon_sym___typeof] = ACTIONS(6824),
    [anon_sym_typeof] = ACTIONS(6824),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6824),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6824),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6824),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6824),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6824),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6824),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6824),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6824),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6824),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6824),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6824),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6824),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6824),
    [anon_sym_API_AVAILABLE] = ACTIONS(6824),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6824),
    [anon_sym_API_DEPRECATED] = ACTIONS(6824),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6824),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6824),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6824),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6824),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6824),
    [anon_sym___deprecated_msg] = ACTIONS(6824),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6824),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6824),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6824),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6824),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6824),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6824),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6824),
    [anon_sym_BOOL] = ACTIONS(6824),
    [anon_sym_IMP] = ACTIONS(6824),
    [anon_sym_SEL] = ACTIONS(6824),
    [anon_sym_Class] = ACTIONS(6824),
    [anon_sym_id] = ACTIONS(6824),
  },
  [STATE(2803)] = {
    [sym__declaration_modifiers] = STATE(2796),
    [sym_attribute_specifier] = STATE(2796),
    [sym_attribute_declaration] = STATE(2796),
    [sym_ms_declspec_modifier] = STATE(2796),
    [sym_ms_based_modifier] = STATE(6257),
    [sym__field_declarator] = STATE(5030),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(2796),
    [sym_type_qualifier] = STATE(2796),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2796),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2796),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6401),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2804)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6222),
    [sym__field_declarator] = STATE(4866),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2805)] = {
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_SEMI] = ACTIONS(6830),
    [anon_sym___extension__] = ACTIONS(6828),
    [anon_sym_extern] = ACTIONS(6828),
    [anon_sym___attribute__] = ACTIONS(6828),
    [anon_sym___attribute] = ACTIONS(6828),
    [anon_sym_noreturn] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6830),
    [anon_sym___declspec] = ACTIONS(6828),
    [anon_sym_LBRACE] = ACTIONS(6830),
    [anon_sym_signed] = ACTIONS(6828),
    [anon_sym_unsigned] = ACTIONS(6828),
    [anon_sym_long] = ACTIONS(6828),
    [anon_sym_short] = ACTIONS(6828),
    [anon_sym_static] = ACTIONS(6828),
    [anon_sym_ATautoreleasepool] = ACTIONS(6830),
    [anon_sym_auto] = ACTIONS(6828),
    [anon_sym_register] = ACTIONS(6828),
    [anon_sym_inline] = ACTIONS(6828),
    [anon_sym___inline] = ACTIONS(6828),
    [anon_sym___inline__] = ACTIONS(6828),
    [anon_sym___forceinline] = ACTIONS(6828),
    [anon_sym_thread_local] = ACTIONS(6828),
    [anon_sym___thread] = ACTIONS(6828),
    [anon_sym_CG_EXTERN] = ACTIONS(6828),
    [anon_sym_CG_INLINE] = ACTIONS(6828),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6828),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6828),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6828),
    [anon_sym_IBOutlet] = ACTIONS(6828),
    [anon_sym_IBInspectable] = ACTIONS(6828),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6828),
    [anon_sym_NS_INLINE] = ACTIONS(6828),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6828),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6828),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6828),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6828),
    [anon_sym_const] = ACTIONS(6828),
    [anon_sym_constexpr] = ACTIONS(6828),
    [anon_sym_volatile] = ACTIONS(6828),
    [anon_sym_restrict] = ACTIONS(6828),
    [anon_sym___restrict__] = ACTIONS(6828),
    [anon_sym__Atomic] = ACTIONS(6828),
    [anon_sym__Noreturn] = ACTIONS(6828),
    [anon_sym__Nonnull] = ACTIONS(6828),
    [anon_sym_nullable] = ACTIONS(6828),
    [anon_sym__Complex] = ACTIONS(6828),
    [anon_sym__Nullable] = ACTIONS(6828),
    [anon_sym__Nullable_result] = ACTIONS(6828),
    [anon_sym__Null_unspecified] = ACTIONS(6828),
    [anon_sym___autoreleasing] = ACTIONS(6828),
    [anon_sym___block] = ACTIONS(6828),
    [anon_sym___bridge] = ACTIONS(6828),
    [anon_sym___bridge_retained] = ACTIONS(6828),
    [anon_sym___bridge_transfer] = ACTIONS(6828),
    [anon_sym___complex] = ACTIONS(6828),
    [anon_sym___const] = ACTIONS(6828),
    [anon_sym___imag] = ACTIONS(6828),
    [anon_sym___kindof] = ACTIONS(6828),
    [anon_sym___nonnull] = ACTIONS(6828),
    [anon_sym___nullable] = ACTIONS(6828),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6828),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6828),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6828),
    [anon_sym___real] = ACTIONS(6828),
    [anon_sym___strong] = ACTIONS(6828),
    [anon_sym___unsafe_unretained] = ACTIONS(6828),
    [anon_sym___unused] = ACTIONS(6828),
    [anon_sym___weak] = ACTIONS(6828),
    [anon_sym_alignas] = ACTIONS(6828),
    [anon_sym__Alignas] = ACTIONS(6828),
    [sym_primitive_type] = ACTIONS(6828),
    [anon_sym_enum] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6830),
    [anon_sym_struct] = ACTIONS(6828),
    [anon_sym_union] = ACTIONS(6828),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6828),
    [anon_sym___typeof] = ACTIONS(6828),
    [anon_sym_typeof] = ACTIONS(6828),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6828),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6828),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6828),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6828),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6828),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6828),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6828),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6828),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6828),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6828),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6828),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6828),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6828),
    [anon_sym_API_AVAILABLE] = ACTIONS(6828),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6828),
    [anon_sym_API_DEPRECATED] = ACTIONS(6828),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6828),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6828),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6828),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6828),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6828),
    [anon_sym___deprecated_msg] = ACTIONS(6828),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6828),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6828),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6828),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6828),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6828),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6828),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6828),
    [anon_sym_BOOL] = ACTIONS(6828),
    [anon_sym_IMP] = ACTIONS(6828),
    [anon_sym_SEL] = ACTIONS(6828),
    [anon_sym_Class] = ACTIONS(6828),
    [anon_sym_id] = ACTIONS(6828),
  },
  [STATE(2806)] = {
    [aux_sym_c_method_parameter_repeat1] = STATE(2809),
    [sym_identifier] = ACTIONS(6792),
    [anon_sym_COMMA] = ACTIONS(6794),
    [anon_sym_SEMI] = ACTIONS(6794),
    [anon_sym___extension__] = ACTIONS(6792),
    [anon_sym_extern] = ACTIONS(6792),
    [anon_sym___attribute__] = ACTIONS(6792),
    [anon_sym___attribute] = ACTIONS(6792),
    [anon_sym_noreturn] = ACTIONS(6792),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym___declspec] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_signed] = ACTIONS(6792),
    [anon_sym_unsigned] = ACTIONS(6792),
    [anon_sym_long] = ACTIONS(6792),
    [anon_sym_short] = ACTIONS(6792),
    [anon_sym_static] = ACTIONS(6792),
    [anon_sym_ATautoreleasepool] = ACTIONS(6794),
    [anon_sym_auto] = ACTIONS(6792),
    [anon_sym_register] = ACTIONS(6792),
    [anon_sym_inline] = ACTIONS(6792),
    [anon_sym___inline] = ACTIONS(6792),
    [anon_sym___inline__] = ACTIONS(6792),
    [anon_sym___forceinline] = ACTIONS(6792),
    [anon_sym_thread_local] = ACTIONS(6792),
    [anon_sym___thread] = ACTIONS(6792),
    [anon_sym_CG_EXTERN] = ACTIONS(6792),
    [anon_sym_CG_INLINE] = ACTIONS(6792),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6792),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6792),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6792),
    [anon_sym_IBOutlet] = ACTIONS(6792),
    [anon_sym_IBInspectable] = ACTIONS(6792),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6792),
    [anon_sym_NS_INLINE] = ACTIONS(6792),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6792),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6792),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6792),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6792),
    [anon_sym_const] = ACTIONS(6792),
    [anon_sym_constexpr] = ACTIONS(6792),
    [anon_sym_volatile] = ACTIONS(6792),
    [anon_sym_restrict] = ACTIONS(6792),
    [anon_sym___restrict__] = ACTIONS(6792),
    [anon_sym__Atomic] = ACTIONS(6792),
    [anon_sym__Noreturn] = ACTIONS(6792),
    [anon_sym__Nonnull] = ACTIONS(6792),
    [anon_sym_nullable] = ACTIONS(6792),
    [anon_sym__Complex] = ACTIONS(6792),
    [anon_sym__Nullable] = ACTIONS(6792),
    [anon_sym__Nullable_result] = ACTIONS(6792),
    [anon_sym__Null_unspecified] = ACTIONS(6792),
    [anon_sym___autoreleasing] = ACTIONS(6792),
    [anon_sym___block] = ACTIONS(6792),
    [anon_sym___bridge] = ACTIONS(6792),
    [anon_sym___bridge_retained] = ACTIONS(6792),
    [anon_sym___bridge_transfer] = ACTIONS(6792),
    [anon_sym___complex] = ACTIONS(6792),
    [anon_sym___const] = ACTIONS(6792),
    [anon_sym___imag] = ACTIONS(6792),
    [anon_sym___kindof] = ACTIONS(6792),
    [anon_sym___nonnull] = ACTIONS(6792),
    [anon_sym___nullable] = ACTIONS(6792),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6792),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6792),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6792),
    [anon_sym___real] = ACTIONS(6792),
    [anon_sym___strong] = ACTIONS(6792),
    [anon_sym___unsafe_unretained] = ACTIONS(6792),
    [anon_sym___unused] = ACTIONS(6792),
    [anon_sym___weak] = ACTIONS(6792),
    [anon_sym_alignas] = ACTIONS(6792),
    [anon_sym__Alignas] = ACTIONS(6792),
    [sym_primitive_type] = ACTIONS(6792),
    [anon_sym_enum] = ACTIONS(6792),
    [anon_sym_struct] = ACTIONS(6792),
    [anon_sym_union] = ACTIONS(6792),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6792),
    [anon_sym___typeof] = ACTIONS(6792),
    [anon_sym_typeof] = ACTIONS(6792),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6792),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6792),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6792),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6792),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6792),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6792),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6792),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6792),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6792),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6792),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6792),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6792),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6792),
    [anon_sym_API_AVAILABLE] = ACTIONS(6792),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6792),
    [anon_sym_API_DEPRECATED] = ACTIONS(6792),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6792),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6792),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6792),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6792),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6792),
    [anon_sym___deprecated_msg] = ACTIONS(6792),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6792),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6792),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6792),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6792),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6792),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6792),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6792),
    [anon_sym_BOOL] = ACTIONS(6792),
    [anon_sym_IMP] = ACTIONS(6792),
    [anon_sym_SEL] = ACTIONS(6792),
    [anon_sym_Class] = ACTIONS(6792),
    [anon_sym_id] = ACTIONS(6792),
  },
  [STATE(2807)] = {
    [aux_sym_method_definition_repeat2] = STATE(2807),
    [sym_identifier] = ACTIONS(6832),
    [anon_sym_COMMA] = ACTIONS(6834),
    [anon_sym_SEMI] = ACTIONS(6837),
    [anon_sym___extension__] = ACTIONS(6832),
    [anon_sym_extern] = ACTIONS(6832),
    [anon_sym___attribute__] = ACTIONS(6832),
    [anon_sym___attribute] = ACTIONS(6832),
    [anon_sym_noreturn] = ACTIONS(6832),
    [anon_sym_LBRACK] = ACTIONS(6837),
    [anon_sym___declspec] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6837),
    [anon_sym_signed] = ACTIONS(6832),
    [anon_sym_unsigned] = ACTIONS(6832),
    [anon_sym_long] = ACTIONS(6832),
    [anon_sym_short] = ACTIONS(6832),
    [anon_sym_static] = ACTIONS(6832),
    [anon_sym_ATautoreleasepool] = ACTIONS(6837),
    [anon_sym_auto] = ACTIONS(6832),
    [anon_sym_register] = ACTIONS(6832),
    [anon_sym_inline] = ACTIONS(6832),
    [anon_sym___inline] = ACTIONS(6832),
    [anon_sym___inline__] = ACTIONS(6832),
    [anon_sym___forceinline] = ACTIONS(6832),
    [anon_sym_thread_local] = ACTIONS(6832),
    [anon_sym___thread] = ACTIONS(6832),
    [anon_sym_CG_EXTERN] = ACTIONS(6832),
    [anon_sym_CG_INLINE] = ACTIONS(6832),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6832),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6832),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6832),
    [anon_sym_IBOutlet] = ACTIONS(6832),
    [anon_sym_IBInspectable] = ACTIONS(6832),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6832),
    [anon_sym_NS_INLINE] = ACTIONS(6832),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6832),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6832),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6832),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6832),
    [anon_sym_const] = ACTIONS(6832),
    [anon_sym_constexpr] = ACTIONS(6832),
    [anon_sym_volatile] = ACTIONS(6832),
    [anon_sym_restrict] = ACTIONS(6832),
    [anon_sym___restrict__] = ACTIONS(6832),
    [anon_sym__Atomic] = ACTIONS(6832),
    [anon_sym__Noreturn] = ACTIONS(6832),
    [anon_sym__Nonnull] = ACTIONS(6832),
    [anon_sym_nullable] = ACTIONS(6832),
    [anon_sym__Complex] = ACTIONS(6832),
    [anon_sym__Nullable] = ACTIONS(6832),
    [anon_sym__Nullable_result] = ACTIONS(6832),
    [anon_sym__Null_unspecified] = ACTIONS(6832),
    [anon_sym___autoreleasing] = ACTIONS(6832),
    [anon_sym___block] = ACTIONS(6832),
    [anon_sym___bridge] = ACTIONS(6832),
    [anon_sym___bridge_retained] = ACTIONS(6832),
    [anon_sym___bridge_transfer] = ACTIONS(6832),
    [anon_sym___complex] = ACTIONS(6832),
    [anon_sym___const] = ACTIONS(6832),
    [anon_sym___imag] = ACTIONS(6832),
    [anon_sym___kindof] = ACTIONS(6832),
    [anon_sym___nonnull] = ACTIONS(6832),
    [anon_sym___nullable] = ACTIONS(6832),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6832),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6832),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6832),
    [anon_sym___real] = ACTIONS(6832),
    [anon_sym___strong] = ACTIONS(6832),
    [anon_sym___unsafe_unretained] = ACTIONS(6832),
    [anon_sym___unused] = ACTIONS(6832),
    [anon_sym___weak] = ACTIONS(6832),
    [anon_sym_alignas] = ACTIONS(6832),
    [anon_sym__Alignas] = ACTIONS(6832),
    [sym_primitive_type] = ACTIONS(6832),
    [anon_sym_enum] = ACTIONS(6832),
    [anon_sym_struct] = ACTIONS(6832),
    [anon_sym_union] = ACTIONS(6832),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6832),
    [anon_sym___typeof] = ACTIONS(6832),
    [anon_sym_typeof] = ACTIONS(6832),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6832),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6832),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6832),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6832),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6832),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6832),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6832),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6832),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6832),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6832),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6832),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6832),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6832),
    [anon_sym_API_AVAILABLE] = ACTIONS(6832),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6832),
    [anon_sym_API_DEPRECATED] = ACTIONS(6832),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6832),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6832),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6832),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6832),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6832),
    [anon_sym___deprecated_msg] = ACTIONS(6832),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6832),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6832),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6832),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6832),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6832),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6832),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6832),
    [anon_sym_BOOL] = ACTIONS(6832),
    [anon_sym_IMP] = ACTIONS(6832),
    [anon_sym_SEL] = ACTIONS(6832),
    [anon_sym_Class] = ACTIONS(6832),
    [anon_sym_id] = ACTIONS(6832),
  },
  [STATE(2808)] = {
    [sym_identifier] = ACTIONS(6839),
    [anon_sym_COMMA] = ACTIONS(6841),
    [anon_sym_SEMI] = ACTIONS(6841),
    [anon_sym___extension__] = ACTIONS(6839),
    [anon_sym_extern] = ACTIONS(6839),
    [anon_sym___attribute__] = ACTIONS(6839),
    [anon_sym___attribute] = ACTIONS(6839),
    [anon_sym_noreturn] = ACTIONS(6839),
    [anon_sym_LBRACK] = ACTIONS(6841),
    [anon_sym___declspec] = ACTIONS(6839),
    [anon_sym_LBRACE] = ACTIONS(6841),
    [anon_sym_signed] = ACTIONS(6839),
    [anon_sym_unsigned] = ACTIONS(6839),
    [anon_sym_long] = ACTIONS(6839),
    [anon_sym_short] = ACTIONS(6839),
    [anon_sym_static] = ACTIONS(6839),
    [anon_sym_ATautoreleasepool] = ACTIONS(6841),
    [anon_sym_auto] = ACTIONS(6839),
    [anon_sym_register] = ACTIONS(6839),
    [anon_sym_inline] = ACTIONS(6839),
    [anon_sym___inline] = ACTIONS(6839),
    [anon_sym___inline__] = ACTIONS(6839),
    [anon_sym___forceinline] = ACTIONS(6839),
    [anon_sym_thread_local] = ACTIONS(6839),
    [anon_sym___thread] = ACTIONS(6839),
    [anon_sym_CG_EXTERN] = ACTIONS(6839),
    [anon_sym_CG_INLINE] = ACTIONS(6839),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6839),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6839),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6839),
    [anon_sym_IBOutlet] = ACTIONS(6839),
    [anon_sym_IBInspectable] = ACTIONS(6839),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6839),
    [anon_sym_NS_INLINE] = ACTIONS(6839),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6839),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6839),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6839),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6839),
    [anon_sym_const] = ACTIONS(6839),
    [anon_sym_constexpr] = ACTIONS(6839),
    [anon_sym_volatile] = ACTIONS(6839),
    [anon_sym_restrict] = ACTIONS(6839),
    [anon_sym___restrict__] = ACTIONS(6839),
    [anon_sym__Atomic] = ACTIONS(6839),
    [anon_sym__Noreturn] = ACTIONS(6839),
    [anon_sym__Nonnull] = ACTIONS(6839),
    [anon_sym_nullable] = ACTIONS(6839),
    [anon_sym__Complex] = ACTIONS(6839),
    [anon_sym__Nullable] = ACTIONS(6839),
    [anon_sym__Nullable_result] = ACTIONS(6839),
    [anon_sym__Null_unspecified] = ACTIONS(6839),
    [anon_sym___autoreleasing] = ACTIONS(6839),
    [anon_sym___block] = ACTIONS(6839),
    [anon_sym___bridge] = ACTIONS(6839),
    [anon_sym___bridge_retained] = ACTIONS(6839),
    [anon_sym___bridge_transfer] = ACTIONS(6839),
    [anon_sym___complex] = ACTIONS(6839),
    [anon_sym___const] = ACTIONS(6839),
    [anon_sym___imag] = ACTIONS(6839),
    [anon_sym___kindof] = ACTIONS(6839),
    [anon_sym___nonnull] = ACTIONS(6839),
    [anon_sym___nullable] = ACTIONS(6839),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6839),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6839),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6839),
    [anon_sym___real] = ACTIONS(6839),
    [anon_sym___strong] = ACTIONS(6839),
    [anon_sym___unsafe_unretained] = ACTIONS(6839),
    [anon_sym___unused] = ACTIONS(6839),
    [anon_sym___weak] = ACTIONS(6839),
    [anon_sym_alignas] = ACTIONS(6839),
    [anon_sym__Alignas] = ACTIONS(6839),
    [sym_primitive_type] = ACTIONS(6839),
    [anon_sym_enum] = ACTIONS(6839),
    [anon_sym_COLON] = ACTIONS(6841),
    [anon_sym_struct] = ACTIONS(6839),
    [anon_sym_union] = ACTIONS(6839),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6839),
    [anon_sym___typeof] = ACTIONS(6839),
    [anon_sym_typeof] = ACTIONS(6839),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6839),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6839),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6839),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6839),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6839),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6839),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6839),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6839),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6839),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6839),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6839),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6839),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6839),
    [anon_sym_API_AVAILABLE] = ACTIONS(6839),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6839),
    [anon_sym_API_DEPRECATED] = ACTIONS(6839),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6839),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6839),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6839),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6839),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6839),
    [anon_sym___deprecated_msg] = ACTIONS(6839),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6839),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6839),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6839),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6839),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6839),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6839),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6839),
    [anon_sym_BOOL] = ACTIONS(6839),
    [anon_sym_IMP] = ACTIONS(6839),
    [anon_sym_SEL] = ACTIONS(6839),
    [anon_sym_Class] = ACTIONS(6839),
    [anon_sym_id] = ACTIONS(6839),
  },
  [STATE(2809)] = {
    [aux_sym_c_method_parameter_repeat1] = STATE(2809),
    [sym_identifier] = ACTIONS(6843),
    [anon_sym_COMMA] = ACTIONS(6845),
    [anon_sym_SEMI] = ACTIONS(6848),
    [anon_sym___extension__] = ACTIONS(6843),
    [anon_sym_extern] = ACTIONS(6843),
    [anon_sym___attribute__] = ACTIONS(6843),
    [anon_sym___attribute] = ACTIONS(6843),
    [anon_sym_noreturn] = ACTIONS(6843),
    [anon_sym_LBRACK] = ACTIONS(6848),
    [anon_sym___declspec] = ACTIONS(6843),
    [anon_sym_LBRACE] = ACTIONS(6848),
    [anon_sym_signed] = ACTIONS(6843),
    [anon_sym_unsigned] = ACTIONS(6843),
    [anon_sym_long] = ACTIONS(6843),
    [anon_sym_short] = ACTIONS(6843),
    [anon_sym_static] = ACTIONS(6843),
    [anon_sym_ATautoreleasepool] = ACTIONS(6848),
    [anon_sym_auto] = ACTIONS(6843),
    [anon_sym_register] = ACTIONS(6843),
    [anon_sym_inline] = ACTIONS(6843),
    [anon_sym___inline] = ACTIONS(6843),
    [anon_sym___inline__] = ACTIONS(6843),
    [anon_sym___forceinline] = ACTIONS(6843),
    [anon_sym_thread_local] = ACTIONS(6843),
    [anon_sym___thread] = ACTIONS(6843),
    [anon_sym_CG_EXTERN] = ACTIONS(6843),
    [anon_sym_CG_INLINE] = ACTIONS(6843),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6843),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6843),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6843),
    [anon_sym_IBOutlet] = ACTIONS(6843),
    [anon_sym_IBInspectable] = ACTIONS(6843),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6843),
    [anon_sym_NS_INLINE] = ACTIONS(6843),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6843),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6843),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6843),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6843),
    [anon_sym_const] = ACTIONS(6843),
    [anon_sym_constexpr] = ACTIONS(6843),
    [anon_sym_volatile] = ACTIONS(6843),
    [anon_sym_restrict] = ACTIONS(6843),
    [anon_sym___restrict__] = ACTIONS(6843),
    [anon_sym__Atomic] = ACTIONS(6843),
    [anon_sym__Noreturn] = ACTIONS(6843),
    [anon_sym__Nonnull] = ACTIONS(6843),
    [anon_sym_nullable] = ACTIONS(6843),
    [anon_sym__Complex] = ACTIONS(6843),
    [anon_sym__Nullable] = ACTIONS(6843),
    [anon_sym__Nullable_result] = ACTIONS(6843),
    [anon_sym__Null_unspecified] = ACTIONS(6843),
    [anon_sym___autoreleasing] = ACTIONS(6843),
    [anon_sym___block] = ACTIONS(6843),
    [anon_sym___bridge] = ACTIONS(6843),
    [anon_sym___bridge_retained] = ACTIONS(6843),
    [anon_sym___bridge_transfer] = ACTIONS(6843),
    [anon_sym___complex] = ACTIONS(6843),
    [anon_sym___const] = ACTIONS(6843),
    [anon_sym___imag] = ACTIONS(6843),
    [anon_sym___kindof] = ACTIONS(6843),
    [anon_sym___nonnull] = ACTIONS(6843),
    [anon_sym___nullable] = ACTIONS(6843),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6843),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6843),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6843),
    [anon_sym___real] = ACTIONS(6843),
    [anon_sym___strong] = ACTIONS(6843),
    [anon_sym___unsafe_unretained] = ACTIONS(6843),
    [anon_sym___unused] = ACTIONS(6843),
    [anon_sym___weak] = ACTIONS(6843),
    [anon_sym_alignas] = ACTIONS(6843),
    [anon_sym__Alignas] = ACTIONS(6843),
    [sym_primitive_type] = ACTIONS(6843),
    [anon_sym_enum] = ACTIONS(6843),
    [anon_sym_struct] = ACTIONS(6843),
    [anon_sym_union] = ACTIONS(6843),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6843),
    [anon_sym___typeof] = ACTIONS(6843),
    [anon_sym_typeof] = ACTIONS(6843),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6843),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6843),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6843),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6843),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6843),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6843),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6843),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6843),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6843),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6843),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6843),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6843),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6843),
    [anon_sym_API_AVAILABLE] = ACTIONS(6843),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6843),
    [anon_sym_API_DEPRECATED] = ACTIONS(6843),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6843),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6843),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6843),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6843),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6843),
    [anon_sym___deprecated_msg] = ACTIONS(6843),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6843),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6843),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6843),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6843),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6843),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6843),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6843),
    [anon_sym_BOOL] = ACTIONS(6843),
    [anon_sym_IMP] = ACTIONS(6843),
    [anon_sym_SEL] = ACTIONS(6843),
    [anon_sym_Class] = ACTIONS(6843),
    [anon_sym_id] = ACTIONS(6843),
  },
  [STATE(2810)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6257),
    [sym__field_declarator] = STATE(5088),
    [sym_parenthesized_field_declarator] = STATE(4889),
    [sym_attributed_field_declarator] = STATE(4889),
    [sym_pointer_field_declarator] = STATE(4889),
    [sym_function_field_declarator] = STATE(4889),
    [sym_array_field_declarator] = STATE(4889),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_field_declarator] = STATE(4889),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6363),
    [anon_sym_LPAREN2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6401),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2811)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym__abstract_declarator] = STATE(4990),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4290),
    [sym_availability_attribute_specifier] = STATE(2864),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [anon_sym_LPAREN2] = ACTIONS(4488),
    [anon_sym_STAR] = ACTIONS(6850),
    [anon_sym_CARET] = ACTIONS(6852),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6770),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(6361),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_ATautoreleasepool] = ACTIONS(6361),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2812)] = {
    [sym__declaration_modifiers] = STATE(6348),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6348),
    [sym_ms_declspec_modifier] = STATE(6348),
    [sym_compound_statement] = STATE(517),
    [sym_storage_class_specifier] = STATE(6348),
    [sym_type_qualifier] = STATE(6348),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3076),
    [sym_availability_attribute_specifier] = STATE(6348),
    [aux_sym_declaration_repeat1] = STATE(3077),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6808),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(303),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(305),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2813)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6199),
    [sym__declarator] = STATE(3053),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2814)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3829),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2815)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3866),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2816)] = {
    [sym__declaration_modifiers] = STATE(6602),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6602),
    [sym_ms_declspec_modifier] = STATE(6602),
    [sym_compound_statement] = STATE(645),
    [sym_storage_class_specifier] = STATE(6602),
    [sym_type_qualifier] = STATE(6602),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3085),
    [sym_availability_attribute_specifier] = STATE(6602),
    [aux_sym_declaration_repeat1] = STATE(3086),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6822),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(870),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(872),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2817)] = {
    [sym__declaration_modifiers] = STATE(6068),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6068),
    [sym_ms_declspec_modifier] = STATE(6068),
    [sym_compound_statement] = STATE(2259),
    [sym_storage_class_specifier] = STATE(6068),
    [sym_type_qualifier] = STATE(6068),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3107),
    [sym_availability_attribute_specifier] = STATE(6068),
    [aux_sym_declaration_repeat1] = STATE(3108),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6856),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(4649),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(4651),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2818)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6101),
    [sym__declarator] = STATE(2708),
    [sym_parenthesized_declarator] = STATE(2778),
    [sym_pointer_declarator] = STATE(2778),
    [sym_function_declarator] = STATE(2778),
    [sym_array_declarator] = STATE(2778),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(2778),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6463),
    [anon_sym_LPAREN2] = ACTIONS(6465),
    [anon_sym_STAR] = ACTIONS(6467),
    [anon_sym_CARET] = ACTIONS(6469),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2819)] = {
    [sym__declaration_modifiers] = STATE(6444),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6444),
    [sym_ms_declspec_modifier] = STATE(6444),
    [sym_compound_statement] = STATE(1022),
    [sym_storage_class_specifier] = STATE(6444),
    [sym_type_qualifier] = STATE(6444),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3118),
    [sym_availability_attribute_specifier] = STATE(6444),
    [aux_sym_declaration_repeat1] = STATE(3117),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6814),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2820)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6199),
    [sym__declarator] = STATE(3068),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2821)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3832),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2822)] = {
    [sym__declaration_modifiers] = STATE(2826),
    [sym_attribute_specifier] = STATE(2826),
    [sym_attribute_declaration] = STATE(2826),
    [sym_ms_declspec_modifier] = STATE(2826),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5206),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2826),
    [sym_type_qualifier] = STATE(2826),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2826),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2826),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2823)] = {
    [sym__declaration_modifiers] = STATE(2811),
    [sym_attribute_specifier] = STATE(2811),
    [sym_attribute_declaration] = STATE(2811),
    [sym_ms_declspec_modifier] = STATE(2811),
    [sym__abstract_declarator] = STATE(4988),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_storage_class_specifier] = STATE(2811),
    [sym_type_qualifier] = STATE(2811),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(4290),
    [sym_availability_attribute_specifier] = STATE(2811),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2811),
    [anon_sym_LPAREN2] = ACTIONS(4488),
    [anon_sym_STAR] = ACTIONS(6850),
    [anon_sym_CARET] = ACTIONS(6852),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6770),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(5314),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_ATautoreleasepool] = ACTIONS(5314),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2824)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5012),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2825)] = {
    [sym__declaration_modifiers] = STATE(6102),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6102),
    [sym_ms_declspec_modifier] = STATE(6102),
    [sym_compound_statement] = STATE(280),
    [sym_storage_class_specifier] = STATE(6102),
    [sym_type_qualifier] = STATE(6102),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3162),
    [sym_availability_attribute_specifier] = STATE(6102),
    [aux_sym_declaration_repeat1] = STATE(3163),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6816),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(183),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2826)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5287),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2827)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5058),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2828)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6101),
    [sym__declarator] = STATE(2687),
    [sym_parenthesized_declarator] = STATE(2778),
    [sym_pointer_declarator] = STATE(2778),
    [sym_function_declarator] = STATE(2778),
    [sym_array_declarator] = STATE(2778),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(2778),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6463),
    [anon_sym_LPAREN2] = ACTIONS(6465),
    [anon_sym_STAR] = ACTIONS(6467),
    [anon_sym_CARET] = ACTIONS(6469),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2829)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5085),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2830)] = {
    [sym__declaration_modifiers] = STATE(2850),
    [sym_attribute_specifier] = STATE(2850),
    [sym_attribute_declaration] = STATE(2850),
    [sym_ms_declspec_modifier] = STATE(2850),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5172),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2850),
    [sym_type_qualifier] = STATE(2850),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2850),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2850),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2831)] = {
    [sym_identifier] = ACTIONS(6714),
    [anon_sym_COMMA] = ACTIONS(6716),
    [anon_sym_SEMI] = ACTIONS(6716),
    [anon_sym___extension__] = ACTIONS(6714),
    [anon_sym_extern] = ACTIONS(6714),
    [anon_sym___attribute__] = ACTIONS(6714),
    [anon_sym___attribute] = ACTIONS(6714),
    [anon_sym_noreturn] = ACTIONS(6714),
    [anon_sym_LBRACK] = ACTIONS(6716),
    [anon_sym___declspec] = ACTIONS(6714),
    [anon_sym_LBRACE] = ACTIONS(6716),
    [anon_sym_signed] = ACTIONS(6714),
    [anon_sym_unsigned] = ACTIONS(6714),
    [anon_sym_long] = ACTIONS(6714),
    [anon_sym_short] = ACTIONS(6714),
    [anon_sym_static] = ACTIONS(6714),
    [anon_sym_ATautoreleasepool] = ACTIONS(6716),
    [anon_sym_auto] = ACTIONS(6714),
    [anon_sym_register] = ACTIONS(6714),
    [anon_sym_inline] = ACTIONS(6714),
    [anon_sym___inline] = ACTIONS(6714),
    [anon_sym___inline__] = ACTIONS(6714),
    [anon_sym___forceinline] = ACTIONS(6714),
    [anon_sym_thread_local] = ACTIONS(6714),
    [anon_sym___thread] = ACTIONS(6714),
    [anon_sym_CG_EXTERN] = ACTIONS(6714),
    [anon_sym_CG_INLINE] = ACTIONS(6714),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6714),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6714),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6714),
    [anon_sym_IBOutlet] = ACTIONS(6714),
    [anon_sym_IBInspectable] = ACTIONS(6714),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6714),
    [anon_sym_NS_INLINE] = ACTIONS(6714),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6714),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6714),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6714),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6714),
    [anon_sym_const] = ACTIONS(6714),
    [anon_sym_constexpr] = ACTIONS(6714),
    [anon_sym_volatile] = ACTIONS(6714),
    [anon_sym_restrict] = ACTIONS(6714),
    [anon_sym___restrict__] = ACTIONS(6714),
    [anon_sym__Atomic] = ACTIONS(6714),
    [anon_sym__Noreturn] = ACTIONS(6714),
    [anon_sym__Nonnull] = ACTIONS(6714),
    [anon_sym_nullable] = ACTIONS(6714),
    [anon_sym__Complex] = ACTIONS(6714),
    [anon_sym__Nullable] = ACTIONS(6714),
    [anon_sym__Nullable_result] = ACTIONS(6714),
    [anon_sym__Null_unspecified] = ACTIONS(6714),
    [anon_sym___autoreleasing] = ACTIONS(6714),
    [anon_sym___block] = ACTIONS(6714),
    [anon_sym___bridge] = ACTIONS(6714),
    [anon_sym___bridge_retained] = ACTIONS(6714),
    [anon_sym___bridge_transfer] = ACTIONS(6714),
    [anon_sym___complex] = ACTIONS(6714),
    [anon_sym___const] = ACTIONS(6714),
    [anon_sym___imag] = ACTIONS(6714),
    [anon_sym___kindof] = ACTIONS(6714),
    [anon_sym___nonnull] = ACTIONS(6714),
    [anon_sym___nullable] = ACTIONS(6714),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6714),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6714),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6714),
    [anon_sym___real] = ACTIONS(6714),
    [anon_sym___strong] = ACTIONS(6714),
    [anon_sym___unsafe_unretained] = ACTIONS(6714),
    [anon_sym___unused] = ACTIONS(6714),
    [anon_sym___weak] = ACTIONS(6714),
    [anon_sym_alignas] = ACTIONS(6714),
    [anon_sym__Alignas] = ACTIONS(6714),
    [sym_primitive_type] = ACTIONS(6714),
    [anon_sym_enum] = ACTIONS(6714),
    [anon_sym_struct] = ACTIONS(6714),
    [anon_sym_union] = ACTIONS(6714),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6714),
    [anon_sym___typeof] = ACTIONS(6714),
    [anon_sym_typeof] = ACTIONS(6714),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6714),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6714),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6714),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6714),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6714),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6714),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6714),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6714),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6714),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6714),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6714),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6714),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6714),
    [anon_sym_API_AVAILABLE] = ACTIONS(6714),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6714),
    [anon_sym_API_DEPRECATED] = ACTIONS(6714),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6714),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6714),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6714),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6714),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6714),
    [anon_sym___deprecated_msg] = ACTIONS(6714),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6714),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6714),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6714),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6714),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6714),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6714),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6714),
    [anon_sym_BOOL] = ACTIONS(6714),
    [anon_sym_IMP] = ACTIONS(6714),
    [anon_sym_SEL] = ACTIONS(6714),
    [anon_sym_Class] = ACTIONS(6714),
    [anon_sym_id] = ACTIONS(6714),
  },
  [STATE(2832)] = {
    [sym__declaration_modifiers] = STATE(6042),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6042),
    [sym_ms_declspec_modifier] = STATE(6042),
    [sym_compound_statement] = STATE(2123),
    [sym_storage_class_specifier] = STATE(6042),
    [sym_type_qualifier] = STATE(6042),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3083),
    [sym_availability_attribute_specifier] = STATE(6042),
    [aux_sym_declaration_repeat1] = STATE(3120),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6858),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(4683),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(4685),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2833)] = {
    [sym_identifier] = ACTIONS(6832),
    [anon_sym_COMMA] = ACTIONS(6837),
    [anon_sym_SEMI] = ACTIONS(6837),
    [anon_sym___extension__] = ACTIONS(6832),
    [anon_sym_extern] = ACTIONS(6832),
    [anon_sym___attribute__] = ACTIONS(6832),
    [anon_sym___attribute] = ACTIONS(6832),
    [anon_sym_noreturn] = ACTIONS(6832),
    [anon_sym_LBRACK] = ACTIONS(6837),
    [anon_sym___declspec] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6837),
    [anon_sym_signed] = ACTIONS(6832),
    [anon_sym_unsigned] = ACTIONS(6832),
    [anon_sym_long] = ACTIONS(6832),
    [anon_sym_short] = ACTIONS(6832),
    [anon_sym_static] = ACTIONS(6832),
    [anon_sym_ATautoreleasepool] = ACTIONS(6837),
    [anon_sym_auto] = ACTIONS(6832),
    [anon_sym_register] = ACTIONS(6832),
    [anon_sym_inline] = ACTIONS(6832),
    [anon_sym___inline] = ACTIONS(6832),
    [anon_sym___inline__] = ACTIONS(6832),
    [anon_sym___forceinline] = ACTIONS(6832),
    [anon_sym_thread_local] = ACTIONS(6832),
    [anon_sym___thread] = ACTIONS(6832),
    [anon_sym_CG_EXTERN] = ACTIONS(6832),
    [anon_sym_CG_INLINE] = ACTIONS(6832),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6832),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6832),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6832),
    [anon_sym_IBOutlet] = ACTIONS(6832),
    [anon_sym_IBInspectable] = ACTIONS(6832),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6832),
    [anon_sym_NS_INLINE] = ACTIONS(6832),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6832),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6832),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6832),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6832),
    [anon_sym_const] = ACTIONS(6832),
    [anon_sym_constexpr] = ACTIONS(6832),
    [anon_sym_volatile] = ACTIONS(6832),
    [anon_sym_restrict] = ACTIONS(6832),
    [anon_sym___restrict__] = ACTIONS(6832),
    [anon_sym__Atomic] = ACTIONS(6832),
    [anon_sym__Noreturn] = ACTIONS(6832),
    [anon_sym__Nonnull] = ACTIONS(6832),
    [anon_sym_nullable] = ACTIONS(6832),
    [anon_sym__Complex] = ACTIONS(6832),
    [anon_sym__Nullable] = ACTIONS(6832),
    [anon_sym__Nullable_result] = ACTIONS(6832),
    [anon_sym__Null_unspecified] = ACTIONS(6832),
    [anon_sym___autoreleasing] = ACTIONS(6832),
    [anon_sym___block] = ACTIONS(6832),
    [anon_sym___bridge] = ACTIONS(6832),
    [anon_sym___bridge_retained] = ACTIONS(6832),
    [anon_sym___bridge_transfer] = ACTIONS(6832),
    [anon_sym___complex] = ACTIONS(6832),
    [anon_sym___const] = ACTIONS(6832),
    [anon_sym___imag] = ACTIONS(6832),
    [anon_sym___kindof] = ACTIONS(6832),
    [anon_sym___nonnull] = ACTIONS(6832),
    [anon_sym___nullable] = ACTIONS(6832),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6832),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6832),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6832),
    [anon_sym___real] = ACTIONS(6832),
    [anon_sym___strong] = ACTIONS(6832),
    [anon_sym___unsafe_unretained] = ACTIONS(6832),
    [anon_sym___unused] = ACTIONS(6832),
    [anon_sym___weak] = ACTIONS(6832),
    [anon_sym_alignas] = ACTIONS(6832),
    [anon_sym__Alignas] = ACTIONS(6832),
    [sym_primitive_type] = ACTIONS(6832),
    [anon_sym_enum] = ACTIONS(6832),
    [anon_sym_struct] = ACTIONS(6832),
    [anon_sym_union] = ACTIONS(6832),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6832),
    [anon_sym___typeof] = ACTIONS(6832),
    [anon_sym_typeof] = ACTIONS(6832),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6832),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6832),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6832),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6832),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6832),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6832),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6832),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6832),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6832),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6832),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6832),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6832),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6832),
    [anon_sym_API_AVAILABLE] = ACTIONS(6832),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6832),
    [anon_sym_API_DEPRECATED] = ACTIONS(6832),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6832),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6832),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6832),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6832),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6832),
    [anon_sym___deprecated_msg] = ACTIONS(6832),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6832),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6832),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6832),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6832),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6832),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6832),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6832),
    [anon_sym_BOOL] = ACTIONS(6832),
    [anon_sym_IMP] = ACTIONS(6832),
    [anon_sym_SEL] = ACTIONS(6832),
    [anon_sym_Class] = ACTIONS(6832),
    [anon_sym_id] = ACTIONS(6832),
  },
  [STATE(2834)] = {
    [sym_identifier] = ACTIONS(4819),
    [anon_sym_COMMA] = ACTIONS(4821),
    [anon_sym_SEMI] = ACTIONS(4821),
    [anon_sym___extension__] = ACTIONS(4819),
    [anon_sym_extern] = ACTIONS(4819),
    [anon_sym___attribute__] = ACTIONS(4819),
    [anon_sym___attribute] = ACTIONS(4819),
    [anon_sym_noreturn] = ACTIONS(4819),
    [anon_sym_LBRACK] = ACTIONS(4821),
    [anon_sym___declspec] = ACTIONS(4819),
    [anon_sym_LBRACE] = ACTIONS(4821),
    [anon_sym_signed] = ACTIONS(4819),
    [anon_sym_unsigned] = ACTIONS(4819),
    [anon_sym_long] = ACTIONS(4819),
    [anon_sym_short] = ACTIONS(4819),
    [anon_sym_static] = ACTIONS(4819),
    [anon_sym_ATautoreleasepool] = ACTIONS(4821),
    [anon_sym_auto] = ACTIONS(4819),
    [anon_sym_register] = ACTIONS(4819),
    [anon_sym_inline] = ACTIONS(4819),
    [anon_sym___inline] = ACTIONS(4819),
    [anon_sym___inline__] = ACTIONS(4819),
    [anon_sym___forceinline] = ACTIONS(4819),
    [anon_sym_thread_local] = ACTIONS(4819),
    [anon_sym___thread] = ACTIONS(4819),
    [anon_sym_CG_EXTERN] = ACTIONS(4819),
    [anon_sym_CG_INLINE] = ACTIONS(4819),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4819),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4819),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4819),
    [anon_sym_IBOutlet] = ACTIONS(4819),
    [anon_sym_IBInspectable] = ACTIONS(4819),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4819),
    [anon_sym_NS_INLINE] = ACTIONS(4819),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4819),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4819),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4819),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4819),
    [anon_sym_const] = ACTIONS(4819),
    [anon_sym_constexpr] = ACTIONS(4819),
    [anon_sym_volatile] = ACTIONS(4819),
    [anon_sym_restrict] = ACTIONS(4819),
    [anon_sym___restrict__] = ACTIONS(4819),
    [anon_sym__Atomic] = ACTIONS(4819),
    [anon_sym__Noreturn] = ACTIONS(4819),
    [anon_sym__Nonnull] = ACTIONS(4819),
    [anon_sym_nullable] = ACTIONS(4819),
    [anon_sym__Complex] = ACTIONS(4819),
    [anon_sym__Nullable] = ACTIONS(4819),
    [anon_sym__Nullable_result] = ACTIONS(4819),
    [anon_sym__Null_unspecified] = ACTIONS(4819),
    [anon_sym___autoreleasing] = ACTIONS(4819),
    [anon_sym___block] = ACTIONS(4819),
    [anon_sym___bridge] = ACTIONS(4819),
    [anon_sym___bridge_retained] = ACTIONS(4819),
    [anon_sym___bridge_transfer] = ACTIONS(4819),
    [anon_sym___complex] = ACTIONS(4819),
    [anon_sym___const] = ACTIONS(4819),
    [anon_sym___imag] = ACTIONS(4819),
    [anon_sym___kindof] = ACTIONS(4819),
    [anon_sym___nonnull] = ACTIONS(4819),
    [anon_sym___nullable] = ACTIONS(4819),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4819),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4819),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4819),
    [anon_sym___real] = ACTIONS(4819),
    [anon_sym___strong] = ACTIONS(4819),
    [anon_sym___unsafe_unretained] = ACTIONS(4819),
    [anon_sym___unused] = ACTIONS(4819),
    [anon_sym___weak] = ACTIONS(4819),
    [anon_sym_alignas] = ACTIONS(4819),
    [anon_sym__Alignas] = ACTIONS(4819),
    [sym_primitive_type] = ACTIONS(4819),
    [anon_sym_enum] = ACTIONS(4819),
    [anon_sym_struct] = ACTIONS(4819),
    [anon_sym_union] = ACTIONS(4819),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4819),
    [anon_sym___typeof] = ACTIONS(4819),
    [anon_sym_typeof] = ACTIONS(4819),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4819),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4819),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4819),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4819),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4819),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4819),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4819),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4819),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4819),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4819),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4819),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4819),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4819),
    [anon_sym_API_AVAILABLE] = ACTIONS(4819),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4819),
    [anon_sym_API_DEPRECATED] = ACTIONS(4819),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4819),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4819),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4819),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4819),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4819),
    [anon_sym___deprecated_msg] = ACTIONS(4819),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4819),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4819),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4819),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4819),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4819),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4819),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4819),
    [anon_sym_BOOL] = ACTIONS(4819),
    [anon_sym_IMP] = ACTIONS(4819),
    [anon_sym_SEL] = ACTIONS(4819),
    [anon_sym_Class] = ACTIONS(4819),
    [anon_sym_id] = ACTIONS(4819),
  },
  [STATE(2835)] = {
    [sym__declaration_modifiers] = STATE(6399),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6399),
    [sym_ms_declspec_modifier] = STATE(6399),
    [sym_compound_statement] = STATE(861),
    [sym_storage_class_specifier] = STATE(6399),
    [sym_type_qualifier] = STATE(6399),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3160),
    [sym_availability_attribute_specifier] = STATE(6399),
    [aux_sym_declaration_repeat1] = STATE(3161),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6800),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(948),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2836)] = {
    [sym_identifier] = ACTIONS(2987),
    [anon_sym___extension__] = ACTIONS(2987),
    [anon_sym_typedef] = ACTIONS(2922),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2992),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2987),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2987),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2987),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2987),
    [anon_sym_union] = ACTIONS(2987),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_ATprotocol] = ACTIONS(2992),
    [anon_sym_ATinterface] = ACTIONS(2992),
    [anon_sym_ATimplementation] = ACTIONS(2992),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2987),
  },
  [STATE(2837)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6199),
    [sym__declarator] = STATE(3055),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2838)] = {
    [sym__declaration_modifiers] = STATE(6017),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6017),
    [sym_ms_declspec_modifier] = STATE(6017),
    [sym_compound_statement] = STATE(2190),
    [sym_storage_class_specifier] = STATE(6017),
    [sym_type_qualifier] = STATE(6017),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3099),
    [sym_availability_attribute_specifier] = STATE(6017),
    [aux_sym_declaration_repeat1] = STATE(3100),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6860),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_LBRACE] = ACTIONS(4659),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_ATautoreleasepool] = ACTIONS(4661),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2839)] = {
    [sym_identifier] = ACTIONS(4779),
    [anon_sym_COMMA] = ACTIONS(4783),
    [anon_sym_SEMI] = ACTIONS(4783),
    [anon_sym___extension__] = ACTIONS(4779),
    [anon_sym_extern] = ACTIONS(4779),
    [anon_sym___attribute__] = ACTIONS(4779),
    [anon_sym___attribute] = ACTIONS(4779),
    [anon_sym_noreturn] = ACTIONS(4779),
    [anon_sym_LBRACK] = ACTIONS(4783),
    [anon_sym___declspec] = ACTIONS(4779),
    [anon_sym_LBRACE] = ACTIONS(4783),
    [anon_sym_signed] = ACTIONS(4779),
    [anon_sym_unsigned] = ACTIONS(4779),
    [anon_sym_long] = ACTIONS(4779),
    [anon_sym_short] = ACTIONS(4779),
    [anon_sym_static] = ACTIONS(4779),
    [anon_sym_ATautoreleasepool] = ACTIONS(4783),
    [anon_sym_auto] = ACTIONS(4779),
    [anon_sym_register] = ACTIONS(4779),
    [anon_sym_inline] = ACTIONS(4779),
    [anon_sym___inline] = ACTIONS(4779),
    [anon_sym___inline__] = ACTIONS(4779),
    [anon_sym___forceinline] = ACTIONS(4779),
    [anon_sym_thread_local] = ACTIONS(4779),
    [anon_sym___thread] = ACTIONS(4779),
    [anon_sym_CG_EXTERN] = ACTIONS(4779),
    [anon_sym_CG_INLINE] = ACTIONS(4779),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4779),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4779),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4779),
    [anon_sym_IBOutlet] = ACTIONS(4779),
    [anon_sym_IBInspectable] = ACTIONS(4779),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4779),
    [anon_sym_NS_INLINE] = ACTIONS(4779),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4779),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4779),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4779),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4779),
    [anon_sym_const] = ACTIONS(4779),
    [anon_sym_constexpr] = ACTIONS(4779),
    [anon_sym_volatile] = ACTIONS(4779),
    [anon_sym_restrict] = ACTIONS(4779),
    [anon_sym___restrict__] = ACTIONS(4779),
    [anon_sym__Atomic] = ACTIONS(4779),
    [anon_sym__Noreturn] = ACTIONS(4779),
    [anon_sym__Nonnull] = ACTIONS(4779),
    [anon_sym_nullable] = ACTIONS(4779),
    [anon_sym__Complex] = ACTIONS(4779),
    [anon_sym__Nullable] = ACTIONS(4779),
    [anon_sym__Nullable_result] = ACTIONS(4779),
    [anon_sym__Null_unspecified] = ACTIONS(4779),
    [anon_sym___autoreleasing] = ACTIONS(4779),
    [anon_sym___block] = ACTIONS(4779),
    [anon_sym___bridge] = ACTIONS(4779),
    [anon_sym___bridge_retained] = ACTIONS(4779),
    [anon_sym___bridge_transfer] = ACTIONS(4779),
    [anon_sym___complex] = ACTIONS(4779),
    [anon_sym___const] = ACTIONS(4779),
    [anon_sym___imag] = ACTIONS(4779),
    [anon_sym___kindof] = ACTIONS(4779),
    [anon_sym___nonnull] = ACTIONS(4779),
    [anon_sym___nullable] = ACTIONS(4779),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4779),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4779),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4779),
    [anon_sym___real] = ACTIONS(4779),
    [anon_sym___strong] = ACTIONS(4779),
    [anon_sym___unsafe_unretained] = ACTIONS(4779),
    [anon_sym___unused] = ACTIONS(4779),
    [anon_sym___weak] = ACTIONS(4779),
    [anon_sym_alignas] = ACTIONS(4779),
    [anon_sym__Alignas] = ACTIONS(4779),
    [sym_primitive_type] = ACTIONS(4779),
    [anon_sym_enum] = ACTIONS(4779),
    [anon_sym_struct] = ACTIONS(4779),
    [anon_sym_union] = ACTIONS(4779),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4779),
    [anon_sym___typeof] = ACTIONS(4779),
    [anon_sym_typeof] = ACTIONS(4779),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4779),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4779),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4779),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4779),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4779),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4779),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4779),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4779),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4779),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4779),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4779),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4779),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4779),
    [anon_sym_API_AVAILABLE] = ACTIONS(4779),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4779),
    [anon_sym_API_DEPRECATED] = ACTIONS(4779),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4779),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4779),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4779),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4779),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4779),
    [anon_sym___deprecated_msg] = ACTIONS(4779),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4779),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4779),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4779),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4779),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4779),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4779),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4779),
    [anon_sym_BOOL] = ACTIONS(4779),
    [anon_sym_IMP] = ACTIONS(4779),
    [anon_sym_SEL] = ACTIONS(4779),
    [anon_sym_Class] = ACTIONS(4779),
    [anon_sym_id] = ACTIONS(4779),
  },
  [STATE(2840)] = {
    [sym_identifier] = ACTIONS(2987),
    [anon_sym___extension__] = ACTIONS(2987),
    [anon_sym_typedef] = ACTIONS(2912),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2992),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2987),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2987),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2987),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2987),
    [anon_sym_union] = ACTIONS(2987),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_ATprotocol] = ACTIONS(2992),
    [anon_sym_ATinterface] = ACTIONS(2992),
    [anon_sym_ATimplementation] = ACTIONS(2992),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2987),
  },
  [STATE(2841)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6656),
    [sym__declarator] = STATE(3215),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6487),
    [anon_sym_CARET] = ACTIONS(6489),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2842)] = {
    [sym__declaration_modifiers] = STATE(2845),
    [sym_attribute_specifier] = STATE(2845),
    [sym_attribute_declaration] = STATE(2845),
    [sym_ms_declspec_modifier] = STATE(2845),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5307),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(2845),
    [sym_type_qualifier] = STATE(2845),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2845),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2845),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2843)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6656),
    [sym__declarator] = STATE(3211),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6487),
    [anon_sym_CARET] = ACTIONS(6489),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2844)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6656),
    [sym__declarator] = STATE(3212),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3065),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5415),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6487),
    [anon_sym_CARET] = ACTIONS(6489),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2845)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5173),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2846)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6101),
    [sym__declarator] = STATE(2753),
    [sym_parenthesized_declarator] = STATE(2778),
    [sym_pointer_declarator] = STATE(2778),
    [sym_function_declarator] = STATE(2778),
    [sym_array_declarator] = STATE(2778),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(2778),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6463),
    [anon_sym_LPAREN2] = ACTIONS(6465),
    [anon_sym_STAR] = ACTIONS(6467),
    [anon_sym_CARET] = ACTIONS(6469),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2847)] = {
    [sym_identifier] = ACTIONS(2987),
    [anon_sym___extension__] = ACTIONS(2987),
    [anon_sym_typedef] = ACTIONS(2924),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2992),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2987),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2987),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2987),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2987),
    [anon_sym_union] = ACTIONS(2987),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_ATprotocol] = ACTIONS(2992),
    [anon_sym_ATinterface] = ACTIONS(2992),
    [anon_sym_ATimplementation] = ACTIONS(2992),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2987),
  },
  [STATE(2848)] = {
    [sym_identifier] = ACTIONS(2987),
    [anon_sym___extension__] = ACTIONS(2987),
    [anon_sym_typedef] = ACTIONS(2928),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2992),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2987),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2987),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2987),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2987),
    [anon_sym_union] = ACTIONS(2987),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_ATprotocol] = ACTIONS(2992),
    [anon_sym_ATinterface] = ACTIONS(2992),
    [anon_sym_ATimplementation] = ACTIONS(2992),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2987),
  },
  [STATE(2849)] = {
    [sym_identifier] = ACTIONS(2987),
    [anon_sym___extension__] = ACTIONS(2987),
    [anon_sym_typedef] = ACTIONS(2926),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2992),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2987),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2987),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2987),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2987),
    [anon_sym_union] = ACTIONS(2987),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_ATprotocol] = ACTIONS(2992),
    [anon_sym_ATinterface] = ACTIONS(2992),
    [anon_sym_ATimplementation] = ACTIONS(2992),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2987),
  },
  [STATE(2850)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_ms_based_modifier] = STATE(6198),
    [sym__declarator] = STATE(5197),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [sym_block_pointer_declarator] = STATE(3980),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(5312),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(4484),
    [anon_sym_CARET] = ACTIONS(4486),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2851)] = {
    [sym_ms_declspec_modifier] = STATE(6676),
    [sym_identifier] = ACTIONS(2914),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_typedef] = ACTIONS(6862),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATproperty] = ACTIONS(2920),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(2852)] = {
    [sym_ms_declspec_modifier] = STATE(6766),
    [sym_identifier] = ACTIONS(2914),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_typedef] = ACTIONS(6864),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATproperty] = ACTIONS(2920),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(2853)] = {
    [sym_ms_declspec_modifier] = STATE(6768),
    [sym_identifier] = ACTIONS(2914),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_typedef] = ACTIONS(6866),
    [anon_sym_extern] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_auto] = ACTIONS(2914),
    [anon_sym_register] = ACTIONS(2914),
    [anon_sym_inline] = ACTIONS(2914),
    [anon_sym___inline] = ACTIONS(2914),
    [anon_sym___inline__] = ACTIONS(2914),
    [anon_sym___forceinline] = ACTIONS(2914),
    [anon_sym_thread_local] = ACTIONS(2914),
    [anon_sym___thread] = ACTIONS(2914),
    [anon_sym_CG_EXTERN] = ACTIONS(2914),
    [anon_sym_CG_INLINE] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2914),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2914),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2914),
    [anon_sym_IBOutlet] = ACTIONS(2914),
    [anon_sym_IBInspectable] = ACTIONS(2914),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2914),
    [anon_sym_NS_INLINE] = ACTIONS(2914),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2914),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2914),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2914),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2914),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2914),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2914),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2914),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_API_AVAILABLE] = ACTIONS(2914),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_API_DEPRECATED] = ACTIONS(2914),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2914),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2914),
    [anon_sym___deprecated_msg] = ACTIONS(2914),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2914),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2914),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2914),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2914),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2914),
    [anon_sym_ATproperty] = ACTIONS(2920),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(2854)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6868),
    [anon_sym_COMMA] = ACTIONS(6870),
    [anon_sym_RPAREN] = ACTIONS(6870),
    [anon_sym_LPAREN2] = ACTIONS(6870),
    [anon_sym_STAR] = ACTIONS(6870),
    [anon_sym_CARET] = ACTIONS(6870),
    [anon_sym_SEMI] = ACTIONS(6870),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(6868),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_enum] = ACTIONS(6868),
    [anon_sym_COLON] = ACTIONS(6870),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2855)] = {
    [sym__declaration_modifiers] = STATE(6159),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6159),
    [sym_ms_declspec_modifier] = STATE(6159),
    [sym_storage_class_specifier] = STATE(6159),
    [sym_type_qualifier] = STATE(6159),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3158),
    [sym_gnu_asm_expression] = STATE(3168),
    [sym_availability_attribute_specifier] = STATE(6159),
    [aux_sym_declaration_repeat1] = STATE(3172),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6872),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_in] = ACTIONS(6874),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2856)] = {
    [sym_type_qualifier] = STATE(2487),
    [sym_alignas_qualifier] = STATE(2354),
    [aux_sym__type_definition_type_repeat1] = STATE(2487),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2995),
    [sym_identifier] = ACTIONS(6876),
    [anon_sym_COMMA] = ACTIONS(6879),
    [anon_sym_RPAREN] = ACTIONS(6879),
    [anon_sym_LPAREN2] = ACTIONS(6879),
    [anon_sym_STAR] = ACTIONS(6879),
    [anon_sym_CARET] = ACTIONS(6879),
    [anon_sym_SEMI] = ACTIONS(6879),
    [anon_sym___extension__] = ACTIONS(6881),
    [anon_sym_extern] = ACTIONS(6884),
    [anon_sym___attribute__] = ACTIONS(6884),
    [anon_sym___attribute] = ACTIONS(6884),
    [anon_sym_noreturn] = ACTIONS(6881),
    [anon_sym_LBRACK] = ACTIONS(6879),
    [anon_sym___declspec] = ACTIONS(6884),
    [anon_sym___based] = ACTIONS(6884),
    [anon_sym_signed] = ACTIONS(6886),
    [anon_sym_unsigned] = ACTIONS(6886),
    [anon_sym_long] = ACTIONS(6886),
    [anon_sym_short] = ACTIONS(6886),
    [anon_sym_static] = ACTIONS(6884),
    [anon_sym_auto] = ACTIONS(6884),
    [anon_sym_register] = ACTIONS(6884),
    [anon_sym_inline] = ACTIONS(6884),
    [anon_sym___inline] = ACTIONS(6884),
    [anon_sym___inline__] = ACTIONS(6884),
    [anon_sym___forceinline] = ACTIONS(6884),
    [anon_sym_thread_local] = ACTIONS(6884),
    [anon_sym___thread] = ACTIONS(6884),
    [anon_sym_CG_EXTERN] = ACTIONS(6884),
    [anon_sym_CG_INLINE] = ACTIONS(6884),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6884),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6884),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6884),
    [anon_sym_IBOutlet] = ACTIONS(6884),
    [anon_sym_IBInspectable] = ACTIONS(6884),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6884),
    [anon_sym_NS_INLINE] = ACTIONS(6884),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6884),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6884),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6884),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6884),
    [anon_sym_const] = ACTIONS(6881),
    [anon_sym_constexpr] = ACTIONS(6881),
    [anon_sym_volatile] = ACTIONS(6881),
    [anon_sym_restrict] = ACTIONS(6881),
    [anon_sym___restrict__] = ACTIONS(6881),
    [anon_sym__Atomic] = ACTIONS(6881),
    [anon_sym__Noreturn] = ACTIONS(6881),
    [anon_sym__Nonnull] = ACTIONS(6881),
    [anon_sym_nullable] = ACTIONS(6881),
    [anon_sym__Complex] = ACTIONS(6881),
    [anon_sym__Nullable] = ACTIONS(6881),
    [anon_sym__Nullable_result] = ACTIONS(6881),
    [anon_sym__Null_unspecified] = ACTIONS(6881),
    [anon_sym___autoreleasing] = ACTIONS(6881),
    [anon_sym___block] = ACTIONS(6881),
    [anon_sym___bridge] = ACTIONS(6881),
    [anon_sym___bridge_retained] = ACTIONS(6881),
    [anon_sym___bridge_transfer] = ACTIONS(6881),
    [anon_sym___complex] = ACTIONS(6881),
    [anon_sym___const] = ACTIONS(6881),
    [anon_sym___imag] = ACTIONS(6881),
    [anon_sym___kindof] = ACTIONS(6881),
    [anon_sym___nonnull] = ACTIONS(6881),
    [anon_sym___nullable] = ACTIONS(6881),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6881),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6881),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6881),
    [anon_sym___real] = ACTIONS(6881),
    [anon_sym___strong] = ACTIONS(6881),
    [anon_sym___unsafe_unretained] = ACTIONS(6881),
    [anon_sym___unused] = ACTIONS(6881),
    [anon_sym___weak] = ACTIONS(6881),
    [anon_sym_alignas] = ACTIONS(6888),
    [anon_sym__Alignas] = ACTIONS(6888),
    [sym_primitive_type] = ACTIONS(6891),
    [anon_sym_enum] = ACTIONS(6884),
    [anon_sym_COLON] = ACTIONS(6879),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6884),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6884),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6884),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6884),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6884),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6884),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6884),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6884),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6884),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6884),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6884),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6884),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6884),
    [anon_sym_API_AVAILABLE] = ACTIONS(6884),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6884),
    [anon_sym_API_DEPRECATED] = ACTIONS(6884),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6884),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6884),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6884),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6884),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6884),
    [anon_sym___deprecated_msg] = ACTIONS(6884),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6884),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6884),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6884),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6884),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6884),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6884),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6884),
  },
  [STATE(2857)] = {
    [sym__declaration_modifiers] = STATE(2854),
    [sym_attribute_specifier] = STATE(2854),
    [sym_attribute_declaration] = STATE(2854),
    [sym_ms_declspec_modifier] = STATE(2854),
    [sym_storage_class_specifier] = STATE(2854),
    [sym_type_qualifier] = STATE(2854),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2854),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2854),
    [sym_identifier] = ACTIONS(6893),
    [anon_sym_COMMA] = ACTIONS(6895),
    [anon_sym_RPAREN] = ACTIONS(6895),
    [anon_sym_LPAREN2] = ACTIONS(6895),
    [anon_sym_STAR] = ACTIONS(6895),
    [anon_sym_CARET] = ACTIONS(6895),
    [anon_sym_SEMI] = ACTIONS(6895),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6897),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(6893),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_enum] = ACTIONS(6893),
    [anon_sym_COLON] = ACTIONS(6895),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2858)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(23),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2222),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2222),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2222),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2222),
    [anon_sym_GT_GT] = ACTIONS(2222),
    [anon_sym_SEMI] = ACTIONS(2222),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym___attribute__] = ACTIONS(2220),
    [anon_sym___attribute] = ACTIONS(2220),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym_RBRACK] = ACTIONS(2222),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_RBRACE] = ACTIONS(2222),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(2222),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2859)] = {
    [sym__declaration_modifiers] = STATE(2866),
    [sym_attribute_specifier] = STATE(2866),
    [sym_attribute_declaration] = STATE(2866),
    [sym_ms_declspec_modifier] = STATE(2866),
    [sym_storage_class_specifier] = STATE(2866),
    [sym_type_qualifier] = STATE(2866),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2866),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2866),
    [sym_identifier] = ACTIONS(6899),
    [anon_sym_COMMA] = ACTIONS(6901),
    [anon_sym_RPAREN] = ACTIONS(6901),
    [anon_sym_LPAREN2] = ACTIONS(6901),
    [anon_sym_STAR] = ACTIONS(6901),
    [anon_sym_CARET] = ACTIONS(6901),
    [anon_sym_SEMI] = ACTIONS(6901),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6897),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(6899),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_enum] = ACTIONS(6899),
    [anon_sym_COLON] = ACTIONS(6901),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2860)] = {
    [sym__declaration_modifiers] = STATE(6159),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6159),
    [sym_ms_declspec_modifier] = STATE(6159),
    [sym_storage_class_specifier] = STATE(6159),
    [sym_type_qualifier] = STATE(6159),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3158),
    [sym_gnu_asm_expression] = STATE(3168),
    [sym_availability_attribute_specifier] = STATE(6159),
    [aux_sym_declaration_repeat1] = STATE(3172),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6872),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_in] = ACTIONS(6903),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2861)] = {
    [sym_identifier] = ACTIONS(2987),
    [anon_sym___extension__] = ACTIONS(2987),
    [anon_sym_typedef] = ACTIONS(6862),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2992),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2987),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2987),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2987),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2987),
    [anon_sym_union] = ACTIONS(2987),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_ATproperty] = ACTIONS(2992),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2987),
  },
  [STATE(2862)] = {
    [sym_type_qualifier] = STATE(2856),
    [sym_alignas_qualifier] = STATE(2354),
    [aux_sym__type_definition_type_repeat1] = STATE(2856),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2914),
    [sym_identifier] = ACTIONS(6905),
    [anon_sym_COMMA] = ACTIONS(6908),
    [anon_sym_RPAREN] = ACTIONS(6908),
    [anon_sym_LPAREN2] = ACTIONS(6908),
    [anon_sym_STAR] = ACTIONS(6908),
    [anon_sym_CARET] = ACTIONS(6908),
    [anon_sym_SEMI] = ACTIONS(6908),
    [anon_sym___extension__] = ACTIONS(6910),
    [anon_sym_extern] = ACTIONS(6913),
    [anon_sym___attribute__] = ACTIONS(6913),
    [anon_sym___attribute] = ACTIONS(6913),
    [anon_sym_noreturn] = ACTIONS(6910),
    [anon_sym_LBRACK] = ACTIONS(6908),
    [anon_sym___declspec] = ACTIONS(6913),
    [anon_sym___based] = ACTIONS(6913),
    [anon_sym_signed] = ACTIONS(6915),
    [anon_sym_unsigned] = ACTIONS(6915),
    [anon_sym_long] = ACTIONS(6915),
    [anon_sym_short] = ACTIONS(6915),
    [anon_sym_static] = ACTIONS(6913),
    [anon_sym_auto] = ACTIONS(6913),
    [anon_sym_register] = ACTIONS(6913),
    [anon_sym_inline] = ACTIONS(6913),
    [anon_sym___inline] = ACTIONS(6913),
    [anon_sym___inline__] = ACTIONS(6913),
    [anon_sym___forceinline] = ACTIONS(6913),
    [anon_sym_thread_local] = ACTIONS(6913),
    [anon_sym___thread] = ACTIONS(6913),
    [anon_sym_CG_EXTERN] = ACTIONS(6913),
    [anon_sym_CG_INLINE] = ACTIONS(6913),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6913),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6913),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6913),
    [anon_sym_IBOutlet] = ACTIONS(6913),
    [anon_sym_IBInspectable] = ACTIONS(6913),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6913),
    [anon_sym_NS_INLINE] = ACTIONS(6913),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6913),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6913),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6913),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6913),
    [anon_sym_const] = ACTIONS(6910),
    [anon_sym_constexpr] = ACTIONS(6910),
    [anon_sym_volatile] = ACTIONS(6910),
    [anon_sym_restrict] = ACTIONS(6910),
    [anon_sym___restrict__] = ACTIONS(6910),
    [anon_sym__Atomic] = ACTIONS(6910),
    [anon_sym__Noreturn] = ACTIONS(6910),
    [anon_sym__Nonnull] = ACTIONS(6910),
    [anon_sym_nullable] = ACTIONS(6910),
    [anon_sym__Complex] = ACTIONS(6910),
    [anon_sym__Nullable] = ACTIONS(6910),
    [anon_sym__Nullable_result] = ACTIONS(6910),
    [anon_sym__Null_unspecified] = ACTIONS(6910),
    [anon_sym___autoreleasing] = ACTIONS(6910),
    [anon_sym___block] = ACTIONS(6910),
    [anon_sym___bridge] = ACTIONS(6910),
    [anon_sym___bridge_retained] = ACTIONS(6910),
    [anon_sym___bridge_transfer] = ACTIONS(6910),
    [anon_sym___complex] = ACTIONS(6910),
    [anon_sym___const] = ACTIONS(6910),
    [anon_sym___imag] = ACTIONS(6910),
    [anon_sym___kindof] = ACTIONS(6910),
    [anon_sym___nonnull] = ACTIONS(6910),
    [anon_sym___nullable] = ACTIONS(6910),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6910),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6910),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6910),
    [anon_sym___real] = ACTIONS(6910),
    [anon_sym___strong] = ACTIONS(6910),
    [anon_sym___unsafe_unretained] = ACTIONS(6910),
    [anon_sym___unused] = ACTIONS(6910),
    [anon_sym___weak] = ACTIONS(6910),
    [anon_sym_alignas] = ACTIONS(6917),
    [anon_sym__Alignas] = ACTIONS(6917),
    [sym_primitive_type] = ACTIONS(6920),
    [anon_sym_enum] = ACTIONS(6913),
    [anon_sym_COLON] = ACTIONS(6908),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6913),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6913),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6913),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6913),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6913),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6913),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6913),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6913),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6913),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6913),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6913),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6913),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6913),
    [anon_sym_API_AVAILABLE] = ACTIONS(6913),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6913),
    [anon_sym_API_DEPRECATED] = ACTIONS(6913),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6913),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6913),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6913),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6913),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6913),
    [anon_sym___deprecated_msg] = ACTIONS(6913),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6913),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6913),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6913),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6913),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6913),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6913),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6913),
  },
  [STATE(2863)] = {
    [sym_identifier] = ACTIONS(2987),
    [anon_sym___extension__] = ACTIONS(2987),
    [anon_sym_typedef] = ACTIONS(6864),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2992),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2987),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2987),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2987),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2987),
    [anon_sym_union] = ACTIONS(2987),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_ATproperty] = ACTIONS(2992),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2987),
  },
  [STATE(2864)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [anon_sym_COMMA] = ACTIONS(5122),
    [anon_sym_RPAREN] = ACTIONS(5122),
    [anon_sym_LPAREN2] = ACTIONS(5122),
    [anon_sym_STAR] = ACTIONS(5122),
    [anon_sym_CARET] = ACTIONS(5122),
    [anon_sym_GT] = ACTIONS(5122),
    [anon_sym_SEMI] = ACTIONS(5122),
    [anon_sym___extension__] = ACTIONS(6922),
    [anon_sym_extern] = ACTIONS(6925),
    [anon_sym___attribute__] = ACTIONS(6928),
    [anon_sym___attribute] = ACTIONS(5130),
    [anon_sym_noreturn] = ACTIONS(6922),
    [anon_sym_LBRACK] = ACTIONS(6931),
    [anon_sym___declspec] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(5122),
    [anon_sym_static] = ACTIONS(6925),
    [anon_sym_ATautoreleasepool] = ACTIONS(5122),
    [anon_sym_auto] = ACTIONS(6925),
    [anon_sym_register] = ACTIONS(6925),
    [anon_sym_inline] = ACTIONS(6925),
    [anon_sym___inline] = ACTIONS(5127),
    [anon_sym___inline__] = ACTIONS(6925),
    [anon_sym___forceinline] = ACTIONS(6925),
    [anon_sym_thread_local] = ACTIONS(6925),
    [anon_sym___thread] = ACTIONS(6925),
    [anon_sym_CG_EXTERN] = ACTIONS(6925),
    [anon_sym_CG_INLINE] = ACTIONS(6925),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6925),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6925),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6925),
    [anon_sym_IBOutlet] = ACTIONS(6925),
    [anon_sym_IBInspectable] = ACTIONS(6925),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6925),
    [anon_sym_NS_INLINE] = ACTIONS(6925),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6925),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6925),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6925),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6925),
    [anon_sym_const] = ACTIONS(5124),
    [anon_sym_constexpr] = ACTIONS(6922),
    [anon_sym_volatile] = ACTIONS(6922),
    [anon_sym_restrict] = ACTIONS(6922),
    [anon_sym___restrict__] = ACTIONS(6922),
    [anon_sym__Atomic] = ACTIONS(6922),
    [anon_sym__Noreturn] = ACTIONS(6922),
    [anon_sym__Nonnull] = ACTIONS(6922),
    [anon_sym_nullable] = ACTIONS(6922),
    [anon_sym__Complex] = ACTIONS(6922),
    [anon_sym__Nullable] = ACTIONS(5124),
    [anon_sym__Nullable_result] = ACTIONS(6922),
    [anon_sym__Null_unspecified] = ACTIONS(6922),
    [anon_sym___autoreleasing] = ACTIONS(6922),
    [anon_sym___block] = ACTIONS(6922),
    [anon_sym___bridge] = ACTIONS(5124),
    [anon_sym___bridge_retained] = ACTIONS(6922),
    [anon_sym___bridge_transfer] = ACTIONS(6922),
    [anon_sym___complex] = ACTIONS(6922),
    [anon_sym___const] = ACTIONS(6922),
    [anon_sym___imag] = ACTIONS(6922),
    [anon_sym___kindof] = ACTIONS(6922),
    [anon_sym___nonnull] = ACTIONS(6922),
    [anon_sym___nullable] = ACTIONS(6922),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6922),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6922),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6922),
    [anon_sym___real] = ACTIONS(6922),
    [anon_sym___strong] = ACTIONS(6922),
    [anon_sym___unsafe_unretained] = ACTIONS(6922),
    [anon_sym___unused] = ACTIONS(6922),
    [anon_sym___weak] = ACTIONS(6922),
    [anon_sym_alignas] = ACTIONS(6937),
    [anon_sym__Alignas] = ACTIONS(6937),
    [anon_sym_COLON] = ACTIONS(5122),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6940),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6940),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6940),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6940),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6940),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6940),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6940),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6940),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6940),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6943),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5145),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6943),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6943),
    [anon_sym_API_AVAILABLE] = ACTIONS(6943),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6943),
    [anon_sym_API_DEPRECATED] = ACTIONS(6943),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6943),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6943),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6943),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6943),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6943),
    [anon_sym___deprecated_msg] = ACTIONS(6943),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6943),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6943),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6943),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6943),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6943),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6943),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6943),
  },
  [STATE(2865)] = {
    [sym__declaration_modifiers] = STATE(6159),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6159),
    [sym_ms_declspec_modifier] = STATE(6159),
    [sym_storage_class_specifier] = STATE(6159),
    [sym_type_qualifier] = STATE(6159),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3158),
    [sym_gnu_asm_expression] = STATE(3168),
    [sym_availability_attribute_specifier] = STATE(6159),
    [aux_sym_declaration_repeat1] = STATE(3172),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6872),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_in] = ACTIONS(6946),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2866)] = {
    [sym__declaration_modifiers] = STATE(1993),
    [sym_attribute_specifier] = STATE(1993),
    [sym_attribute_declaration] = STATE(1993),
    [sym_ms_declspec_modifier] = STATE(1993),
    [sym_storage_class_specifier] = STATE(1993),
    [sym_type_qualifier] = STATE(1993),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(1993),
    [aux_sym__declaration_specifiers_repeat1] = STATE(1993),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_RPAREN] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6950),
    [anon_sym_STAR] = ACTIONS(6950),
    [anon_sym_CARET] = ACTIONS(6950),
    [anon_sym_SEMI] = ACTIONS(6950),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(6948),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_enum] = ACTIONS(6948),
    [anon_sym_COLON] = ACTIONS(6950),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2867)] = {
    [sym__declaration_modifiers] = STATE(6159),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6159),
    [sym_ms_declspec_modifier] = STATE(6159),
    [sym_storage_class_specifier] = STATE(6159),
    [sym_type_qualifier] = STATE(6159),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3158),
    [sym_gnu_asm_expression] = STATE(3168),
    [sym_availability_attribute_specifier] = STATE(6159),
    [aux_sym_declaration_repeat1] = STATE(3172),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6872),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_in] = ACTIONS(6952),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2868)] = {
    [sym__declaration_modifiers] = STATE(6159),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6159),
    [sym_ms_declspec_modifier] = STATE(6159),
    [sym_storage_class_specifier] = STATE(6159),
    [sym_type_qualifier] = STATE(6159),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3158),
    [sym_gnu_asm_expression] = STATE(3168),
    [sym_availability_attribute_specifier] = STATE(6159),
    [aux_sym_declaration_repeat1] = STATE(3172),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6872),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_in] = ACTIONS(6954),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2869)] = {
    [sym__declaration_modifiers] = STATE(6159),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6159),
    [sym_ms_declspec_modifier] = STATE(6159),
    [sym_storage_class_specifier] = STATE(6159),
    [sym_type_qualifier] = STATE(6159),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3158),
    [sym_gnu_asm_expression] = STATE(3168),
    [sym_availability_attribute_specifier] = STATE(6159),
    [aux_sym_declaration_repeat1] = STATE(3172),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6872),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_in] = ACTIONS(6956),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2870)] = {
    [sym_identifier] = ACTIONS(2987),
    [anon_sym___extension__] = ACTIONS(2987),
    [anon_sym_typedef] = ACTIONS(6866),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2992),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2987),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2987),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2987),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2987),
    [anon_sym_union] = ACTIONS(2987),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_ATproperty] = ACTIONS(2992),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2987),
  },
  [STATE(2871)] = {
    [sym__declaration_modifiers] = STATE(6159),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6159),
    [sym_ms_declspec_modifier] = STATE(6159),
    [sym_storage_class_specifier] = STATE(6159),
    [sym_type_qualifier] = STATE(6159),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3158),
    [sym_gnu_asm_expression] = STATE(3168),
    [sym_availability_attribute_specifier] = STATE(6159),
    [aux_sym_declaration_repeat1] = STATE(3172),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6872),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_in] = ACTIONS(6958),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2872)] = {
    [sym__declaration_modifiers] = STATE(6159),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6159),
    [sym_ms_declspec_modifier] = STATE(6159),
    [sym_storage_class_specifier] = STATE(6159),
    [sym_type_qualifier] = STATE(6159),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3158),
    [sym_gnu_asm_expression] = STATE(3168),
    [sym_availability_attribute_specifier] = STATE(6159),
    [aux_sym_declaration_repeat1] = STATE(3172),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6872),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_in] = ACTIONS(6960),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2873)] = {
    [sym_attribute_specifier] = STATE(2952),
    [sym_field_declaration_list] = STATE(2926),
    [sym_identifier] = ACTIONS(6962),
    [anon_sym_COMMA] = ACTIONS(6964),
    [anon_sym_RPAREN] = ACTIONS(6964),
    [anon_sym_LPAREN2] = ACTIONS(6964),
    [anon_sym_STAR] = ACTIONS(6964),
    [anon_sym_CARET] = ACTIONS(6964),
    [anon_sym_SEMI] = ACTIONS(6964),
    [anon_sym___extension__] = ACTIONS(6962),
    [anon_sym_extern] = ACTIONS(6962),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6962),
    [anon_sym_LBRACK] = ACTIONS(6964),
    [anon_sym___declspec] = ACTIONS(6962),
    [anon_sym___based] = ACTIONS(6962),
    [anon_sym_LBRACE] = ACTIONS(6966),
    [anon_sym_signed] = ACTIONS(6962),
    [anon_sym_unsigned] = ACTIONS(6962),
    [anon_sym_long] = ACTIONS(6962),
    [anon_sym_short] = ACTIONS(6962),
    [anon_sym_static] = ACTIONS(6962),
    [anon_sym_auto] = ACTIONS(6962),
    [anon_sym_register] = ACTIONS(6962),
    [anon_sym_inline] = ACTIONS(6962),
    [anon_sym___inline] = ACTIONS(6962),
    [anon_sym___inline__] = ACTIONS(6962),
    [anon_sym___forceinline] = ACTIONS(6962),
    [anon_sym_thread_local] = ACTIONS(6962),
    [anon_sym___thread] = ACTIONS(6962),
    [anon_sym_CG_EXTERN] = ACTIONS(6962),
    [anon_sym_CG_INLINE] = ACTIONS(6962),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6962),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6962),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6962),
    [anon_sym_IBOutlet] = ACTIONS(6962),
    [anon_sym_IBInspectable] = ACTIONS(6962),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6962),
    [anon_sym_NS_INLINE] = ACTIONS(6962),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6962),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6962),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6962),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6962),
    [anon_sym_const] = ACTIONS(6962),
    [anon_sym_constexpr] = ACTIONS(6962),
    [anon_sym_volatile] = ACTIONS(6962),
    [anon_sym_restrict] = ACTIONS(6962),
    [anon_sym___restrict__] = ACTIONS(6962),
    [anon_sym__Atomic] = ACTIONS(6962),
    [anon_sym__Noreturn] = ACTIONS(6962),
    [anon_sym__Nonnull] = ACTIONS(6962),
    [anon_sym_nullable] = ACTIONS(6962),
    [anon_sym__Complex] = ACTIONS(6962),
    [anon_sym__Nullable] = ACTIONS(6962),
    [anon_sym__Nullable_result] = ACTIONS(6962),
    [anon_sym__Null_unspecified] = ACTIONS(6962),
    [anon_sym___autoreleasing] = ACTIONS(6962),
    [anon_sym___block] = ACTIONS(6962),
    [anon_sym___bridge] = ACTIONS(6962),
    [anon_sym___bridge_retained] = ACTIONS(6962),
    [anon_sym___bridge_transfer] = ACTIONS(6962),
    [anon_sym___complex] = ACTIONS(6962),
    [anon_sym___const] = ACTIONS(6962),
    [anon_sym___imag] = ACTIONS(6962),
    [anon_sym___kindof] = ACTIONS(6962),
    [anon_sym___nonnull] = ACTIONS(6962),
    [anon_sym___nullable] = ACTIONS(6962),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6962),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6962),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6962),
    [anon_sym___real] = ACTIONS(6962),
    [anon_sym___strong] = ACTIONS(6962),
    [anon_sym___unsafe_unretained] = ACTIONS(6962),
    [anon_sym___unused] = ACTIONS(6962),
    [anon_sym___weak] = ACTIONS(6962),
    [anon_sym_alignas] = ACTIONS(6962),
    [anon_sym__Alignas] = ACTIONS(6962),
    [sym_primitive_type] = ACTIONS(6962),
    [anon_sym_enum] = ACTIONS(6962),
    [anon_sym_COLON] = ACTIONS(6964),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6962),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6962),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6962),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6962),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6962),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6962),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6962),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6962),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6962),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6962),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6962),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6962),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6962),
    [anon_sym_API_AVAILABLE] = ACTIONS(6962),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6962),
    [anon_sym_API_DEPRECATED] = ACTIONS(6962),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6962),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6962),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6962),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6962),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6962),
    [anon_sym___deprecated_msg] = ACTIONS(6962),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6962),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6962),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6962),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6962),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6962),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6962),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6962),
  },
  [STATE(2874)] = {
    [sym_identifier] = ACTIONS(2987),
    [anon_sym___extension__] = ACTIONS(2987),
    [anon_sym_typedef] = ACTIONS(2956),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2992),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2987),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2987),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2987),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2987),
    [anon_sym_union] = ACTIONS(2987),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2987),
  },
  [STATE(2875)] = {
    [sym__declaration_modifiers] = STATE(6602),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6602),
    [sym_ms_declspec_modifier] = STATE(6602),
    [sym_storage_class_specifier] = STATE(6602),
    [sym_type_qualifier] = STATE(6602),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3085),
    [sym_availability_attribute_specifier] = STATE(6602),
    [aux_sym_declaration_repeat1] = STATE(3086),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6822),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2876)] = {
    [sym_attribute_specifier] = STATE(2987),
    [sym_field_declaration_list] = STATE(2929),
    [sym_identifier] = ACTIONS(6968),
    [anon_sym_COMMA] = ACTIONS(6970),
    [anon_sym_RPAREN] = ACTIONS(6970),
    [anon_sym_LPAREN2] = ACTIONS(6970),
    [anon_sym_STAR] = ACTIONS(6970),
    [anon_sym_CARET] = ACTIONS(6970),
    [anon_sym_SEMI] = ACTIONS(6970),
    [anon_sym___extension__] = ACTIONS(6968),
    [anon_sym_extern] = ACTIONS(6968),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6968),
    [anon_sym_LBRACK] = ACTIONS(6970),
    [anon_sym___declspec] = ACTIONS(6968),
    [anon_sym___based] = ACTIONS(6968),
    [anon_sym_LBRACE] = ACTIONS(6966),
    [anon_sym_signed] = ACTIONS(6968),
    [anon_sym_unsigned] = ACTIONS(6968),
    [anon_sym_long] = ACTIONS(6968),
    [anon_sym_short] = ACTIONS(6968),
    [anon_sym_static] = ACTIONS(6968),
    [anon_sym_auto] = ACTIONS(6968),
    [anon_sym_register] = ACTIONS(6968),
    [anon_sym_inline] = ACTIONS(6968),
    [anon_sym___inline] = ACTIONS(6968),
    [anon_sym___inline__] = ACTIONS(6968),
    [anon_sym___forceinline] = ACTIONS(6968),
    [anon_sym_thread_local] = ACTIONS(6968),
    [anon_sym___thread] = ACTIONS(6968),
    [anon_sym_CG_EXTERN] = ACTIONS(6968),
    [anon_sym_CG_INLINE] = ACTIONS(6968),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6968),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6968),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6968),
    [anon_sym_IBOutlet] = ACTIONS(6968),
    [anon_sym_IBInspectable] = ACTIONS(6968),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6968),
    [anon_sym_NS_INLINE] = ACTIONS(6968),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6968),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6968),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6968),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6968),
    [anon_sym_const] = ACTIONS(6968),
    [anon_sym_constexpr] = ACTIONS(6968),
    [anon_sym_volatile] = ACTIONS(6968),
    [anon_sym_restrict] = ACTIONS(6968),
    [anon_sym___restrict__] = ACTIONS(6968),
    [anon_sym__Atomic] = ACTIONS(6968),
    [anon_sym__Noreturn] = ACTIONS(6968),
    [anon_sym__Nonnull] = ACTIONS(6968),
    [anon_sym_nullable] = ACTIONS(6968),
    [anon_sym__Complex] = ACTIONS(6968),
    [anon_sym__Nullable] = ACTIONS(6968),
    [anon_sym__Nullable_result] = ACTIONS(6968),
    [anon_sym__Null_unspecified] = ACTIONS(6968),
    [anon_sym___autoreleasing] = ACTIONS(6968),
    [anon_sym___block] = ACTIONS(6968),
    [anon_sym___bridge] = ACTIONS(6968),
    [anon_sym___bridge_retained] = ACTIONS(6968),
    [anon_sym___bridge_transfer] = ACTIONS(6968),
    [anon_sym___complex] = ACTIONS(6968),
    [anon_sym___const] = ACTIONS(6968),
    [anon_sym___imag] = ACTIONS(6968),
    [anon_sym___kindof] = ACTIONS(6968),
    [anon_sym___nonnull] = ACTIONS(6968),
    [anon_sym___nullable] = ACTIONS(6968),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6968),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6968),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6968),
    [anon_sym___real] = ACTIONS(6968),
    [anon_sym___strong] = ACTIONS(6968),
    [anon_sym___unsafe_unretained] = ACTIONS(6968),
    [anon_sym___unused] = ACTIONS(6968),
    [anon_sym___weak] = ACTIONS(6968),
    [anon_sym_alignas] = ACTIONS(6968),
    [anon_sym__Alignas] = ACTIONS(6968),
    [sym_primitive_type] = ACTIONS(6968),
    [anon_sym_enum] = ACTIONS(6968),
    [anon_sym_COLON] = ACTIONS(6970),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6968),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6968),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6968),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6968),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6968),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6968),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6968),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6968),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6968),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6968),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6968),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6968),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6968),
    [anon_sym_API_AVAILABLE] = ACTIONS(6968),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6968),
    [anon_sym_API_DEPRECATED] = ACTIONS(6968),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6968),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6968),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6968),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6968),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6968),
    [anon_sym___deprecated_msg] = ACTIONS(6968),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6968),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6968),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6968),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6968),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6968),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6968),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6968),
  },
  [STATE(2877)] = {
    [sym__declaration_modifiers] = STATE(3122),
    [sym_attribute_specifier] = STATE(3122),
    [sym_attribute_declaration] = STATE(3122),
    [sym_ms_declspec_modifier] = STATE(3122),
    [sym_storage_class_specifier] = STATE(3122),
    [sym_type_qualifier] = STATE(3122),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3122),
    [sym_keyword_declarator] = STATE(3221),
    [sym_method_parameter] = STATE(3217),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3122),
    [aux_sym_keyword_selector_repeat1] = STATE(3221),
    [aux_sym_method_declaration_repeat1] = STATE(3217),
    [aux_sym_method_declaration_repeat2] = STATE(2187),
    [sym_identifier] = ACTIONS(6972),
    [anon_sym_COMMA] = ACTIONS(6974),
    [anon_sym_SEMI] = ACTIONS(6976),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6980),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2878)] = {
    [sym__declaration_modifiers] = STATE(3129),
    [sym_attribute_specifier] = STATE(3129),
    [sym_attribute_declaration] = STATE(3129),
    [sym_ms_declspec_modifier] = STATE(3129),
    [sym_storage_class_specifier] = STATE(3129),
    [sym_type_qualifier] = STATE(3129),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3129),
    [sym_keyword_declarator] = STATE(3221),
    [sym_method_parameter] = STATE(3217),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3129),
    [aux_sym_keyword_selector_repeat1] = STATE(3221),
    [aux_sym_method_declaration_repeat1] = STATE(3217),
    [aux_sym_method_declaration_repeat2] = STATE(2180),
    [sym_identifier] = ACTIONS(6972),
    [anon_sym_COMMA] = ACTIONS(6982),
    [anon_sym_SEMI] = ACTIONS(6976),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6980),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2879)] = {
    [sym__declaration_modifiers] = STATE(3149),
    [sym_attribute_specifier] = STATE(3149),
    [sym_attribute_declaration] = STATE(3149),
    [sym_ms_declspec_modifier] = STATE(3149),
    [sym_storage_class_specifier] = STATE(3149),
    [sym_type_qualifier] = STATE(3149),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3149),
    [sym_keyword_declarator] = STATE(3221),
    [sym_method_parameter] = STATE(3217),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3149),
    [aux_sym_keyword_selector_repeat1] = STATE(3221),
    [aux_sym_method_declaration_repeat1] = STATE(3217),
    [aux_sym_method_declaration_repeat2] = STATE(2133),
    [sym_identifier] = ACTIONS(6972),
    [anon_sym_COMMA] = ACTIONS(6984),
    [anon_sym_SEMI] = ACTIONS(6976),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6980),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2880)] = {
    [sym_attribute_specifier] = STATE(2990),
    [sym_field_declaration_list] = STATE(2910),
    [sym_identifier] = ACTIONS(6986),
    [anon_sym_COMMA] = ACTIONS(6988),
    [anon_sym_RPAREN] = ACTIONS(6988),
    [anon_sym_LPAREN2] = ACTIONS(6988),
    [anon_sym_STAR] = ACTIONS(6988),
    [anon_sym_CARET] = ACTIONS(6988),
    [anon_sym_SEMI] = ACTIONS(6988),
    [anon_sym___extension__] = ACTIONS(6986),
    [anon_sym_extern] = ACTIONS(6986),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6986),
    [anon_sym_LBRACK] = ACTIONS(6988),
    [anon_sym___declspec] = ACTIONS(6986),
    [anon_sym___based] = ACTIONS(6986),
    [anon_sym_LBRACE] = ACTIONS(6966),
    [anon_sym_signed] = ACTIONS(6986),
    [anon_sym_unsigned] = ACTIONS(6986),
    [anon_sym_long] = ACTIONS(6986),
    [anon_sym_short] = ACTIONS(6986),
    [anon_sym_static] = ACTIONS(6986),
    [anon_sym_auto] = ACTIONS(6986),
    [anon_sym_register] = ACTIONS(6986),
    [anon_sym_inline] = ACTIONS(6986),
    [anon_sym___inline] = ACTIONS(6986),
    [anon_sym___inline__] = ACTIONS(6986),
    [anon_sym___forceinline] = ACTIONS(6986),
    [anon_sym_thread_local] = ACTIONS(6986),
    [anon_sym___thread] = ACTIONS(6986),
    [anon_sym_CG_EXTERN] = ACTIONS(6986),
    [anon_sym_CG_INLINE] = ACTIONS(6986),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6986),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6986),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6986),
    [anon_sym_IBOutlet] = ACTIONS(6986),
    [anon_sym_IBInspectable] = ACTIONS(6986),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6986),
    [anon_sym_NS_INLINE] = ACTIONS(6986),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6986),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6986),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6986),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6986),
    [anon_sym_const] = ACTIONS(6986),
    [anon_sym_constexpr] = ACTIONS(6986),
    [anon_sym_volatile] = ACTIONS(6986),
    [anon_sym_restrict] = ACTIONS(6986),
    [anon_sym___restrict__] = ACTIONS(6986),
    [anon_sym__Atomic] = ACTIONS(6986),
    [anon_sym__Noreturn] = ACTIONS(6986),
    [anon_sym__Nonnull] = ACTIONS(6986),
    [anon_sym_nullable] = ACTIONS(6986),
    [anon_sym__Complex] = ACTIONS(6986),
    [anon_sym__Nullable] = ACTIONS(6986),
    [anon_sym__Nullable_result] = ACTIONS(6986),
    [anon_sym__Null_unspecified] = ACTIONS(6986),
    [anon_sym___autoreleasing] = ACTIONS(6986),
    [anon_sym___block] = ACTIONS(6986),
    [anon_sym___bridge] = ACTIONS(6986),
    [anon_sym___bridge_retained] = ACTIONS(6986),
    [anon_sym___bridge_transfer] = ACTIONS(6986),
    [anon_sym___complex] = ACTIONS(6986),
    [anon_sym___const] = ACTIONS(6986),
    [anon_sym___imag] = ACTIONS(6986),
    [anon_sym___kindof] = ACTIONS(6986),
    [anon_sym___nonnull] = ACTIONS(6986),
    [anon_sym___nullable] = ACTIONS(6986),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6986),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6986),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6986),
    [anon_sym___real] = ACTIONS(6986),
    [anon_sym___strong] = ACTIONS(6986),
    [anon_sym___unsafe_unretained] = ACTIONS(6986),
    [anon_sym___unused] = ACTIONS(6986),
    [anon_sym___weak] = ACTIONS(6986),
    [anon_sym_alignas] = ACTIONS(6986),
    [anon_sym__Alignas] = ACTIONS(6986),
    [sym_primitive_type] = ACTIONS(6986),
    [anon_sym_enum] = ACTIONS(6986),
    [anon_sym_COLON] = ACTIONS(6988),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6986),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6986),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6986),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6986),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6986),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6986),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6986),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6986),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6986),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6986),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6986),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6986),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6986),
    [anon_sym_API_AVAILABLE] = ACTIONS(6986),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6986),
    [anon_sym_API_DEPRECATED] = ACTIONS(6986),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6986),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6986),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6986),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6986),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6986),
    [anon_sym___deprecated_msg] = ACTIONS(6986),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6986),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6986),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6986),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6986),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6986),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6986),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6986),
  },
  [STATE(2881)] = {
    [sym__declaration_modifiers] = STATE(6509),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6509),
    [sym_ms_declspec_modifier] = STATE(6509),
    [sym_storage_class_specifier] = STATE(6509),
    [sym_type_qualifier] = STATE(6509),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3123),
    [sym_availability_attribute_specifier] = STATE(6509),
    [aux_sym_declaration_repeat1] = STATE(3113),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6990),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2882)] = {
    [sym__declaration_modifiers] = STATE(6399),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6399),
    [sym_ms_declspec_modifier] = STATE(6399),
    [sym_storage_class_specifier] = STATE(6399),
    [sym_type_qualifier] = STATE(6399),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3160),
    [sym_availability_attribute_specifier] = STATE(6399),
    [aux_sym_declaration_repeat1] = STATE(3161),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6800),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2883)] = {
    [sym__declaration_modifiers] = STATE(3137),
    [sym_attribute_specifier] = STATE(3137),
    [sym_attribute_declaration] = STATE(3137),
    [sym_ms_declspec_modifier] = STATE(3137),
    [sym_storage_class_specifier] = STATE(3137),
    [sym_type_qualifier] = STATE(3137),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3137),
    [sym_keyword_declarator] = STATE(3221),
    [sym_method_parameter] = STATE(3217),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3137),
    [aux_sym_keyword_selector_repeat1] = STATE(3221),
    [aux_sym_method_declaration_repeat1] = STATE(3217),
    [aux_sym_method_declaration_repeat2] = STATE(2172),
    [sym_identifier] = ACTIONS(6972),
    [anon_sym_COMMA] = ACTIONS(6992),
    [anon_sym_SEMI] = ACTIONS(6994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6980),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2884)] = {
    [sym_ms_unaligned_ptr_modifier] = STATE(3000),
    [sym_ms_pointer_modifier] = STATE(2884),
    [aux_sym_pointer_declarator_repeat1] = STATE(2884),
    [sym_identifier] = ACTIONS(6996),
    [anon_sym_LPAREN2] = ACTIONS(6998),
    [anon_sym_STAR] = ACTIONS(6998),
    [anon_sym_CARET] = ACTIONS(6998),
    [anon_sym___extension__] = ACTIONS(6996),
    [anon_sym_extern] = ACTIONS(6996),
    [anon_sym___attribute__] = ACTIONS(6996),
    [anon_sym___attribute] = ACTIONS(6996),
    [anon_sym_noreturn] = ACTIONS(6996),
    [anon_sym_LBRACK] = ACTIONS(6998),
    [anon_sym___declspec] = ACTIONS(6996),
    [anon_sym___based] = ACTIONS(6996),
    [sym_ms_restrict_modifier] = ACTIONS(7000),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(7000),
    [sym_ms_signed_ptr_modifier] = ACTIONS(7000),
    [anon_sym__unaligned] = ACTIONS(7003),
    [anon_sym___unaligned] = ACTIONS(7003),
    [anon_sym_signed] = ACTIONS(6996),
    [anon_sym_unsigned] = ACTIONS(6996),
    [anon_sym_long] = ACTIONS(6996),
    [anon_sym_short] = ACTIONS(6996),
    [anon_sym_static] = ACTIONS(6996),
    [anon_sym_auto] = ACTIONS(6996),
    [anon_sym_register] = ACTIONS(6996),
    [anon_sym_inline] = ACTIONS(6996),
    [anon_sym___inline] = ACTIONS(6996),
    [anon_sym___inline__] = ACTIONS(6996),
    [anon_sym___forceinline] = ACTIONS(6996),
    [anon_sym_thread_local] = ACTIONS(6996),
    [anon_sym___thread] = ACTIONS(6996),
    [anon_sym_CG_EXTERN] = ACTIONS(6996),
    [anon_sym_CG_INLINE] = ACTIONS(6996),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6996),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6996),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6996),
    [anon_sym_IBOutlet] = ACTIONS(6996),
    [anon_sym_IBInspectable] = ACTIONS(6996),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6996),
    [anon_sym_NS_INLINE] = ACTIONS(6996),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6996),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6996),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6996),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6996),
    [anon_sym_const] = ACTIONS(6996),
    [anon_sym_constexpr] = ACTIONS(6996),
    [anon_sym_volatile] = ACTIONS(6996),
    [anon_sym_restrict] = ACTIONS(6996),
    [anon_sym___restrict__] = ACTIONS(6996),
    [anon_sym__Atomic] = ACTIONS(6996),
    [anon_sym__Noreturn] = ACTIONS(6996),
    [anon_sym__Nonnull] = ACTIONS(6996),
    [anon_sym_nullable] = ACTIONS(6996),
    [anon_sym__Complex] = ACTIONS(6996),
    [anon_sym__Nullable] = ACTIONS(6996),
    [anon_sym__Nullable_result] = ACTIONS(6996),
    [anon_sym__Null_unspecified] = ACTIONS(6996),
    [anon_sym___autoreleasing] = ACTIONS(6996),
    [anon_sym___block] = ACTIONS(6996),
    [anon_sym___bridge] = ACTIONS(6996),
    [anon_sym___bridge_retained] = ACTIONS(6996),
    [anon_sym___bridge_transfer] = ACTIONS(6996),
    [anon_sym___complex] = ACTIONS(6996),
    [anon_sym___const] = ACTIONS(6996),
    [anon_sym___imag] = ACTIONS(6996),
    [anon_sym___kindof] = ACTIONS(6996),
    [anon_sym___nonnull] = ACTIONS(6996),
    [anon_sym___nullable] = ACTIONS(6996),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6996),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6996),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6996),
    [anon_sym___real] = ACTIONS(6996),
    [anon_sym___strong] = ACTIONS(6996),
    [anon_sym___unsafe_unretained] = ACTIONS(6996),
    [anon_sym___unused] = ACTIONS(6996),
    [anon_sym___weak] = ACTIONS(6996),
    [anon_sym_alignas] = ACTIONS(6996),
    [anon_sym__Alignas] = ACTIONS(6996),
    [sym_primitive_type] = ACTIONS(6996),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6996),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6996),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6996),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6996),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6996),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6996),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6996),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6996),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6996),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6996),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6996),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6996),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6996),
    [anon_sym_API_AVAILABLE] = ACTIONS(6996),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6996),
    [anon_sym_API_DEPRECATED] = ACTIONS(6996),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6996),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6996),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6996),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6996),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6996),
    [anon_sym___deprecated_msg] = ACTIONS(6996),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6996),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6996),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6996),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6996),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6996),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6996),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6996),
  },
  [STATE(2885)] = {
    [sym_attribute_specifier] = STATE(2947),
    [sym_field_declaration_list] = STATE(2917),
    [sym_identifier] = ACTIONS(7006),
    [anon_sym_COMMA] = ACTIONS(7008),
    [anon_sym_RPAREN] = ACTIONS(7008),
    [anon_sym_LPAREN2] = ACTIONS(7008),
    [anon_sym_STAR] = ACTIONS(7008),
    [anon_sym_CARET] = ACTIONS(7008),
    [anon_sym_SEMI] = ACTIONS(7008),
    [anon_sym___extension__] = ACTIONS(7006),
    [anon_sym_extern] = ACTIONS(7006),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7006),
    [anon_sym_LBRACK] = ACTIONS(7008),
    [anon_sym___declspec] = ACTIONS(7006),
    [anon_sym___based] = ACTIONS(7006),
    [anon_sym_LBRACE] = ACTIONS(6966),
    [anon_sym_signed] = ACTIONS(7006),
    [anon_sym_unsigned] = ACTIONS(7006),
    [anon_sym_long] = ACTIONS(7006),
    [anon_sym_short] = ACTIONS(7006),
    [anon_sym_static] = ACTIONS(7006),
    [anon_sym_auto] = ACTIONS(7006),
    [anon_sym_register] = ACTIONS(7006),
    [anon_sym_inline] = ACTIONS(7006),
    [anon_sym___inline] = ACTIONS(7006),
    [anon_sym___inline__] = ACTIONS(7006),
    [anon_sym___forceinline] = ACTIONS(7006),
    [anon_sym_thread_local] = ACTIONS(7006),
    [anon_sym___thread] = ACTIONS(7006),
    [anon_sym_CG_EXTERN] = ACTIONS(7006),
    [anon_sym_CG_INLINE] = ACTIONS(7006),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7006),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7006),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7006),
    [anon_sym_IBOutlet] = ACTIONS(7006),
    [anon_sym_IBInspectable] = ACTIONS(7006),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7006),
    [anon_sym_NS_INLINE] = ACTIONS(7006),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7006),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7006),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7006),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7006),
    [anon_sym_const] = ACTIONS(7006),
    [anon_sym_constexpr] = ACTIONS(7006),
    [anon_sym_volatile] = ACTIONS(7006),
    [anon_sym_restrict] = ACTIONS(7006),
    [anon_sym___restrict__] = ACTIONS(7006),
    [anon_sym__Atomic] = ACTIONS(7006),
    [anon_sym__Noreturn] = ACTIONS(7006),
    [anon_sym__Nonnull] = ACTIONS(7006),
    [anon_sym_nullable] = ACTIONS(7006),
    [anon_sym__Complex] = ACTIONS(7006),
    [anon_sym__Nullable] = ACTIONS(7006),
    [anon_sym__Nullable_result] = ACTIONS(7006),
    [anon_sym__Null_unspecified] = ACTIONS(7006),
    [anon_sym___autoreleasing] = ACTIONS(7006),
    [anon_sym___block] = ACTIONS(7006),
    [anon_sym___bridge] = ACTIONS(7006),
    [anon_sym___bridge_retained] = ACTIONS(7006),
    [anon_sym___bridge_transfer] = ACTIONS(7006),
    [anon_sym___complex] = ACTIONS(7006),
    [anon_sym___const] = ACTIONS(7006),
    [anon_sym___imag] = ACTIONS(7006),
    [anon_sym___kindof] = ACTIONS(7006),
    [anon_sym___nonnull] = ACTIONS(7006),
    [anon_sym___nullable] = ACTIONS(7006),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7006),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7006),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7006),
    [anon_sym___real] = ACTIONS(7006),
    [anon_sym___strong] = ACTIONS(7006),
    [anon_sym___unsafe_unretained] = ACTIONS(7006),
    [anon_sym___unused] = ACTIONS(7006),
    [anon_sym___weak] = ACTIONS(7006),
    [anon_sym_alignas] = ACTIONS(7006),
    [anon_sym__Alignas] = ACTIONS(7006),
    [sym_primitive_type] = ACTIONS(7006),
    [anon_sym_enum] = ACTIONS(7006),
    [anon_sym_COLON] = ACTIONS(7008),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7006),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7006),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7006),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7006),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7006),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7006),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7006),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7006),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7006),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7006),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7006),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7006),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7006),
    [anon_sym_API_AVAILABLE] = ACTIONS(7006),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7006),
    [anon_sym_API_DEPRECATED] = ACTIONS(7006),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7006),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7006),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7006),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7006),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7006),
    [anon_sym___deprecated_msg] = ACTIONS(7006),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7006),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7006),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7006),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7006),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7006),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7006),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7006),
  },
  [STATE(2886)] = {
    [sym__declaration_modifiers] = STATE(3127),
    [sym_attribute_specifier] = STATE(3127),
    [sym_attribute_declaration] = STATE(3127),
    [sym_ms_declspec_modifier] = STATE(3127),
    [sym_storage_class_specifier] = STATE(3127),
    [sym_type_qualifier] = STATE(3127),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3127),
    [sym_keyword_declarator] = STATE(3221),
    [sym_method_parameter] = STATE(3217),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3127),
    [aux_sym_keyword_selector_repeat1] = STATE(3221),
    [aux_sym_method_declaration_repeat1] = STATE(3217),
    [aux_sym_method_declaration_repeat2] = STATE(2314),
    [sym_identifier] = ACTIONS(6972),
    [anon_sym_COMMA] = ACTIONS(7010),
    [anon_sym_SEMI] = ACTIONS(7012),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6980),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2887)] = {
    [sym_identifier] = ACTIONS(7014),
    [anon_sym___extension__] = ACTIONS(7014),
    [anon_sym_extern] = ACTIONS(7014),
    [anon_sym___attribute__] = ACTIONS(7014),
    [anon_sym___attribute] = ACTIONS(7014),
    [anon_sym_noreturn] = ACTIONS(7014),
    [anon_sym_LBRACK] = ACTIONS(7016),
    [anon_sym___declspec] = ACTIONS(7014),
    [anon_sym_signed] = ACTIONS(7014),
    [anon_sym_unsigned] = ACTIONS(7014),
    [anon_sym_long] = ACTIONS(7014),
    [anon_sym_short] = ACTIONS(7014),
    [anon_sym_static] = ACTIONS(7014),
    [anon_sym_auto] = ACTIONS(7014),
    [anon_sym_register] = ACTIONS(7014),
    [anon_sym_inline] = ACTIONS(7014),
    [anon_sym___inline] = ACTIONS(7014),
    [anon_sym___inline__] = ACTIONS(7014),
    [anon_sym___forceinline] = ACTIONS(7014),
    [anon_sym_thread_local] = ACTIONS(7014),
    [anon_sym___thread] = ACTIONS(7014),
    [anon_sym_CG_EXTERN] = ACTIONS(7014),
    [anon_sym_CG_INLINE] = ACTIONS(7014),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7014),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7014),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7014),
    [anon_sym_IBOutlet] = ACTIONS(7014),
    [anon_sym_IBInspectable] = ACTIONS(7014),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7014),
    [anon_sym_NS_INLINE] = ACTIONS(7014),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7014),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7014),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7014),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7014),
    [anon_sym_const] = ACTIONS(7014),
    [anon_sym_constexpr] = ACTIONS(7014),
    [anon_sym_volatile] = ACTIONS(7014),
    [anon_sym_restrict] = ACTIONS(7014),
    [anon_sym___restrict__] = ACTIONS(7014),
    [anon_sym__Atomic] = ACTIONS(7014),
    [anon_sym__Noreturn] = ACTIONS(7014),
    [anon_sym__Nonnull] = ACTIONS(7014),
    [anon_sym_nullable] = ACTIONS(7014),
    [anon_sym__Complex] = ACTIONS(7014),
    [anon_sym__Nullable] = ACTIONS(7014),
    [anon_sym__Nullable_result] = ACTIONS(7014),
    [anon_sym__Null_unspecified] = ACTIONS(7014),
    [anon_sym___autoreleasing] = ACTIONS(7014),
    [anon_sym___block] = ACTIONS(7014),
    [anon_sym___bridge] = ACTIONS(7014),
    [anon_sym___bridge_retained] = ACTIONS(7014),
    [anon_sym___bridge_transfer] = ACTIONS(7014),
    [anon_sym___complex] = ACTIONS(7014),
    [anon_sym___const] = ACTIONS(7014),
    [anon_sym___imag] = ACTIONS(7014),
    [anon_sym___kindof] = ACTIONS(7014),
    [anon_sym___nonnull] = ACTIONS(7014),
    [anon_sym___nullable] = ACTIONS(7014),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7014),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7014),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7014),
    [anon_sym___real] = ACTIONS(7014),
    [anon_sym___strong] = ACTIONS(7014),
    [anon_sym___unsafe_unretained] = ACTIONS(7014),
    [anon_sym___unused] = ACTIONS(7014),
    [anon_sym___weak] = ACTIONS(7014),
    [anon_sym_alignas] = ACTIONS(7014),
    [anon_sym__Alignas] = ACTIONS(7014),
    [sym_primitive_type] = ACTIONS(7014),
    [anon_sym_enum] = ACTIONS(7014),
    [anon_sym_struct] = ACTIONS(7014),
    [anon_sym_union] = ACTIONS(7014),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7014),
    [anon_sym___typeof] = ACTIONS(7014),
    [anon_sym_typeof] = ACTIONS(7014),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7014),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7014),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7014),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7014),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7014),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7014),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7014),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7014),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7014),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_API_AVAILABLE] = ACTIONS(7014),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_API_DEPRECATED] = ACTIONS(7014),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7014),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7014),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7014),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7014),
    [anon_sym___deprecated_msg] = ACTIONS(7014),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7014),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7014),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7014),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7014),
    [anon_sym_ATproperty] = ACTIONS(7018),
    [anon_sym_BOOL] = ACTIONS(7014),
    [anon_sym_IMP] = ACTIONS(7014),
    [anon_sym_SEL] = ACTIONS(7014),
    [anon_sym_Class] = ACTIONS(7014),
    [anon_sym_id] = ACTIONS(7014),
  },
  [STATE(2888)] = {
    [sym__declaration_modifiers] = STATE(6042),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6042),
    [sym_ms_declspec_modifier] = STATE(6042),
    [sym_storage_class_specifier] = STATE(6042),
    [sym_type_qualifier] = STATE(6042),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3083),
    [sym_availability_attribute_specifier] = STATE(6042),
    [aux_sym_declaration_repeat1] = STATE(3120),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6858),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2889)] = {
    [sym__declaration_modifiers] = STATE(6444),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6444),
    [sym_ms_declspec_modifier] = STATE(6444),
    [sym_storage_class_specifier] = STATE(6444),
    [sym_type_qualifier] = STATE(6444),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3118),
    [sym_availability_attribute_specifier] = STATE(6444),
    [aux_sym_declaration_repeat1] = STATE(3117),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6814),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2890)] = {
    [sym__declaration_modifiers] = STATE(6348),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6348),
    [sym_ms_declspec_modifier] = STATE(6348),
    [sym_storage_class_specifier] = STATE(6348),
    [sym_type_qualifier] = STATE(6348),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3076),
    [sym_availability_attribute_specifier] = STATE(6348),
    [aux_sym_declaration_repeat1] = STATE(3077),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6808),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2891)] = {
    [sym__declaration_modifiers] = STATE(3187),
    [sym_attribute_specifier] = STATE(3187),
    [sym_attribute_declaration] = STATE(3187),
    [sym_ms_declspec_modifier] = STATE(3187),
    [sym_storage_class_specifier] = STATE(3187),
    [sym_type_qualifier] = STATE(3187),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3187),
    [sym_keyword_declarator] = STATE(3221),
    [sym_method_parameter] = STATE(3217),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3187),
    [aux_sym_keyword_selector_repeat1] = STATE(3221),
    [aux_sym_method_declaration_repeat1] = STATE(3217),
    [aux_sym_method_declaration_repeat2] = STATE(2313),
    [sym_identifier] = ACTIONS(6972),
    [anon_sym_COMMA] = ACTIONS(7020),
    [anon_sym_SEMI] = ACTIONS(7012),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6980),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2892)] = {
    [sym__declaration_modifiers] = STATE(3080),
    [sym_attribute_specifier] = STATE(3080),
    [sym_attribute_declaration] = STATE(3080),
    [sym_ms_declspec_modifier] = STATE(3080),
    [sym_storage_class_specifier] = STATE(3080),
    [sym_type_qualifier] = STATE(3080),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3080),
    [sym_keyword_declarator] = STATE(3221),
    [sym_method_parameter] = STATE(3217),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3080),
    [aux_sym_keyword_selector_repeat1] = STATE(3221),
    [aux_sym_method_declaration_repeat1] = STATE(3217),
    [aux_sym_method_declaration_repeat2] = STATE(2192),
    [sym_identifier] = ACTIONS(6972),
    [anon_sym_COMMA] = ACTIONS(7022),
    [anon_sym_SEMI] = ACTIONS(6994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6980),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2893)] = {
    [sym__declaration_modifiers] = STATE(5939),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(5939),
    [sym_ms_declspec_modifier] = STATE(5939),
    [sym_storage_class_specifier] = STATE(5939),
    [sym_type_qualifier] = STATE(5939),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3092),
    [sym_availability_attribute_specifier] = STATE(5939),
    [aux_sym_declaration_repeat1] = STATE(3093),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(7024),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2894)] = {
    [sym_identifier] = ACTIONS(7014),
    [anon_sym___extension__] = ACTIONS(7014),
    [anon_sym_extern] = ACTIONS(7014),
    [anon_sym___attribute__] = ACTIONS(7014),
    [anon_sym___attribute] = ACTIONS(7014),
    [anon_sym_noreturn] = ACTIONS(7014),
    [anon_sym_LBRACK] = ACTIONS(7016),
    [anon_sym___declspec] = ACTIONS(7014),
    [anon_sym_signed] = ACTIONS(7014),
    [anon_sym_unsigned] = ACTIONS(7014),
    [anon_sym_long] = ACTIONS(7014),
    [anon_sym_short] = ACTIONS(7014),
    [anon_sym_static] = ACTIONS(7014),
    [anon_sym_auto] = ACTIONS(7014),
    [anon_sym_register] = ACTIONS(7014),
    [anon_sym_inline] = ACTIONS(7014),
    [anon_sym___inline] = ACTIONS(7014),
    [anon_sym___inline__] = ACTIONS(7014),
    [anon_sym___forceinline] = ACTIONS(7014),
    [anon_sym_thread_local] = ACTIONS(7014),
    [anon_sym___thread] = ACTIONS(7014),
    [anon_sym_CG_EXTERN] = ACTIONS(7014),
    [anon_sym_CG_INLINE] = ACTIONS(7014),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7014),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7014),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7014),
    [anon_sym_IBOutlet] = ACTIONS(7014),
    [anon_sym_IBInspectable] = ACTIONS(7014),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7014),
    [anon_sym_NS_INLINE] = ACTIONS(7014),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7014),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7014),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7014),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7014),
    [anon_sym_const] = ACTIONS(7014),
    [anon_sym_constexpr] = ACTIONS(7014),
    [anon_sym_volatile] = ACTIONS(7014),
    [anon_sym_restrict] = ACTIONS(7014),
    [anon_sym___restrict__] = ACTIONS(7014),
    [anon_sym__Atomic] = ACTIONS(7014),
    [anon_sym__Noreturn] = ACTIONS(7014),
    [anon_sym__Nonnull] = ACTIONS(7014),
    [anon_sym_nullable] = ACTIONS(7014),
    [anon_sym__Complex] = ACTIONS(7014),
    [anon_sym__Nullable] = ACTIONS(7014),
    [anon_sym__Nullable_result] = ACTIONS(7014),
    [anon_sym__Null_unspecified] = ACTIONS(7014),
    [anon_sym___autoreleasing] = ACTIONS(7014),
    [anon_sym___block] = ACTIONS(7014),
    [anon_sym___bridge] = ACTIONS(7014),
    [anon_sym___bridge_retained] = ACTIONS(7014),
    [anon_sym___bridge_transfer] = ACTIONS(7014),
    [anon_sym___complex] = ACTIONS(7014),
    [anon_sym___const] = ACTIONS(7014),
    [anon_sym___imag] = ACTIONS(7014),
    [anon_sym___kindof] = ACTIONS(7014),
    [anon_sym___nonnull] = ACTIONS(7014),
    [anon_sym___nullable] = ACTIONS(7014),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7014),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7014),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7014),
    [anon_sym___real] = ACTIONS(7014),
    [anon_sym___strong] = ACTIONS(7014),
    [anon_sym___unsafe_unretained] = ACTIONS(7014),
    [anon_sym___unused] = ACTIONS(7014),
    [anon_sym___weak] = ACTIONS(7014),
    [anon_sym_alignas] = ACTIONS(7014),
    [anon_sym__Alignas] = ACTIONS(7014),
    [sym_primitive_type] = ACTIONS(7014),
    [anon_sym_enum] = ACTIONS(7014),
    [anon_sym_struct] = ACTIONS(7014),
    [anon_sym_union] = ACTIONS(7014),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7014),
    [anon_sym___typeof] = ACTIONS(7014),
    [anon_sym_typeof] = ACTIONS(7014),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7014),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7014),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7014),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7014),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7014),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7014),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7014),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7014),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7014),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_API_AVAILABLE] = ACTIONS(7014),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_API_DEPRECATED] = ACTIONS(7014),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7014),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7014),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7014),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7014),
    [anon_sym___deprecated_msg] = ACTIONS(7014),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7014),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7014),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7014),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7014),
    [anon_sym_ATproperty] = ACTIONS(7026),
    [anon_sym_BOOL] = ACTIONS(7014),
    [anon_sym_IMP] = ACTIONS(7014),
    [anon_sym_SEL] = ACTIONS(7014),
    [anon_sym_Class] = ACTIONS(7014),
    [anon_sym_id] = ACTIONS(7014),
  },
  [STATE(2895)] = {
    [sym__declaration_modifiers] = STATE(3133),
    [sym_attribute_specifier] = STATE(3133),
    [sym_attribute_declaration] = STATE(3133),
    [sym_ms_declspec_modifier] = STATE(3133),
    [sym_storage_class_specifier] = STATE(3133),
    [sym_type_qualifier] = STATE(3133),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3133),
    [sym_keyword_declarator] = STATE(3221),
    [sym_method_parameter] = STATE(3217),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3133),
    [aux_sym_keyword_selector_repeat1] = STATE(3221),
    [aux_sym_method_declaration_repeat1] = STATE(3217),
    [aux_sym_method_declaration_repeat2] = STATE(2280),
    [sym_identifier] = ACTIONS(6972),
    [anon_sym_COMMA] = ACTIONS(7028),
    [anon_sym_SEMI] = ACTIONS(7012),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6980),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2896)] = {
    [sym__declaration_modifiers] = STATE(3184),
    [sym_attribute_specifier] = STATE(3184),
    [sym_attribute_declaration] = STATE(3184),
    [sym_ms_declspec_modifier] = STATE(3184),
    [sym_storage_class_specifier] = STATE(3184),
    [sym_type_qualifier] = STATE(3184),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3184),
    [sym_keyword_declarator] = STATE(3221),
    [sym_method_parameter] = STATE(3217),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3184),
    [aux_sym_keyword_selector_repeat1] = STATE(3221),
    [aux_sym_method_declaration_repeat1] = STATE(3217),
    [aux_sym_method_declaration_repeat2] = STATE(2167),
    [sym_identifier] = ACTIONS(6972),
    [anon_sym_COMMA] = ACTIONS(7030),
    [anon_sym_SEMI] = ACTIONS(6994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6980),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2897)] = {
    [sym_attribute_specifier] = STATE(2935),
    [sym_field_declaration_list] = STATE(2919),
    [sym_identifier] = ACTIONS(7032),
    [anon_sym_COMMA] = ACTIONS(7034),
    [anon_sym_RPAREN] = ACTIONS(7034),
    [anon_sym_LPAREN2] = ACTIONS(7034),
    [anon_sym_STAR] = ACTIONS(7034),
    [anon_sym_CARET] = ACTIONS(7034),
    [anon_sym_SEMI] = ACTIONS(7034),
    [anon_sym___extension__] = ACTIONS(7032),
    [anon_sym_extern] = ACTIONS(7032),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7032),
    [anon_sym_LBRACK] = ACTIONS(7034),
    [anon_sym___declspec] = ACTIONS(7032),
    [anon_sym___based] = ACTIONS(7032),
    [anon_sym_LBRACE] = ACTIONS(6966),
    [anon_sym_signed] = ACTIONS(7032),
    [anon_sym_unsigned] = ACTIONS(7032),
    [anon_sym_long] = ACTIONS(7032),
    [anon_sym_short] = ACTIONS(7032),
    [anon_sym_static] = ACTIONS(7032),
    [anon_sym_auto] = ACTIONS(7032),
    [anon_sym_register] = ACTIONS(7032),
    [anon_sym_inline] = ACTIONS(7032),
    [anon_sym___inline] = ACTIONS(7032),
    [anon_sym___inline__] = ACTIONS(7032),
    [anon_sym___forceinline] = ACTIONS(7032),
    [anon_sym_thread_local] = ACTIONS(7032),
    [anon_sym___thread] = ACTIONS(7032),
    [anon_sym_CG_EXTERN] = ACTIONS(7032),
    [anon_sym_CG_INLINE] = ACTIONS(7032),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7032),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7032),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7032),
    [anon_sym_IBOutlet] = ACTIONS(7032),
    [anon_sym_IBInspectable] = ACTIONS(7032),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7032),
    [anon_sym_NS_INLINE] = ACTIONS(7032),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7032),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7032),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7032),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7032),
    [anon_sym_const] = ACTIONS(7032),
    [anon_sym_constexpr] = ACTIONS(7032),
    [anon_sym_volatile] = ACTIONS(7032),
    [anon_sym_restrict] = ACTIONS(7032),
    [anon_sym___restrict__] = ACTIONS(7032),
    [anon_sym__Atomic] = ACTIONS(7032),
    [anon_sym__Noreturn] = ACTIONS(7032),
    [anon_sym__Nonnull] = ACTIONS(7032),
    [anon_sym_nullable] = ACTIONS(7032),
    [anon_sym__Complex] = ACTIONS(7032),
    [anon_sym__Nullable] = ACTIONS(7032),
    [anon_sym__Nullable_result] = ACTIONS(7032),
    [anon_sym__Null_unspecified] = ACTIONS(7032),
    [anon_sym___autoreleasing] = ACTIONS(7032),
    [anon_sym___block] = ACTIONS(7032),
    [anon_sym___bridge] = ACTIONS(7032),
    [anon_sym___bridge_retained] = ACTIONS(7032),
    [anon_sym___bridge_transfer] = ACTIONS(7032),
    [anon_sym___complex] = ACTIONS(7032),
    [anon_sym___const] = ACTIONS(7032),
    [anon_sym___imag] = ACTIONS(7032),
    [anon_sym___kindof] = ACTIONS(7032),
    [anon_sym___nonnull] = ACTIONS(7032),
    [anon_sym___nullable] = ACTIONS(7032),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7032),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7032),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7032),
    [anon_sym___real] = ACTIONS(7032),
    [anon_sym___strong] = ACTIONS(7032),
    [anon_sym___unsafe_unretained] = ACTIONS(7032),
    [anon_sym___unused] = ACTIONS(7032),
    [anon_sym___weak] = ACTIONS(7032),
    [anon_sym_alignas] = ACTIONS(7032),
    [anon_sym__Alignas] = ACTIONS(7032),
    [sym_primitive_type] = ACTIONS(7032),
    [anon_sym_enum] = ACTIONS(7032),
    [anon_sym_COLON] = ACTIONS(7034),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7032),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7032),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7032),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7032),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7032),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7032),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7032),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7032),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7032),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7032),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7032),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7032),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7032),
    [anon_sym_API_AVAILABLE] = ACTIONS(7032),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7032),
    [anon_sym_API_DEPRECATED] = ACTIONS(7032),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7032),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7032),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7032),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7032),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7032),
    [anon_sym___deprecated_msg] = ACTIONS(7032),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7032),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7032),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7032),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7032),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7032),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7032),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7032),
  },
  [STATE(2898)] = {
    [sym_identifier] = ACTIONS(2987),
    [anon_sym___extension__] = ACTIONS(2987),
    [anon_sym_typedef] = ACTIONS(2946),
    [anon_sym_extern] = ACTIONS(2987),
    [anon_sym___attribute__] = ACTIONS(2987),
    [anon_sym___attribute] = ACTIONS(2987),
    [anon_sym_noreturn] = ACTIONS(2987),
    [anon_sym_LBRACK] = ACTIONS(2992),
    [anon_sym___declspec] = ACTIONS(2987),
    [anon_sym_signed] = ACTIONS(2987),
    [anon_sym_unsigned] = ACTIONS(2987),
    [anon_sym_long] = ACTIONS(2987),
    [anon_sym_short] = ACTIONS(2987),
    [anon_sym_static] = ACTIONS(2987),
    [anon_sym_auto] = ACTIONS(2987),
    [anon_sym_register] = ACTIONS(2987),
    [anon_sym_inline] = ACTIONS(2987),
    [anon_sym___inline] = ACTIONS(2987),
    [anon_sym___inline__] = ACTIONS(2987),
    [anon_sym___forceinline] = ACTIONS(2987),
    [anon_sym_thread_local] = ACTIONS(2987),
    [anon_sym___thread] = ACTIONS(2987),
    [anon_sym_CG_EXTERN] = ACTIONS(2987),
    [anon_sym_CG_INLINE] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2987),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2987),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2987),
    [anon_sym_IBOutlet] = ACTIONS(2987),
    [anon_sym_IBInspectable] = ACTIONS(2987),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2987),
    [anon_sym_NS_INLINE] = ACTIONS(2987),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2987),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2987),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2987),
    [anon_sym_const] = ACTIONS(2987),
    [anon_sym_constexpr] = ACTIONS(2987),
    [anon_sym_volatile] = ACTIONS(2987),
    [anon_sym_restrict] = ACTIONS(2987),
    [anon_sym___restrict__] = ACTIONS(2987),
    [anon_sym__Atomic] = ACTIONS(2987),
    [anon_sym__Noreturn] = ACTIONS(2987),
    [anon_sym__Nonnull] = ACTIONS(2987),
    [anon_sym_nullable] = ACTIONS(2987),
    [anon_sym__Complex] = ACTIONS(2987),
    [anon_sym__Nullable] = ACTIONS(2987),
    [anon_sym__Nullable_result] = ACTIONS(2987),
    [anon_sym__Null_unspecified] = ACTIONS(2987),
    [anon_sym___autoreleasing] = ACTIONS(2987),
    [anon_sym___block] = ACTIONS(2987),
    [anon_sym___bridge] = ACTIONS(2987),
    [anon_sym___bridge_retained] = ACTIONS(2987),
    [anon_sym___bridge_transfer] = ACTIONS(2987),
    [anon_sym___complex] = ACTIONS(2987),
    [anon_sym___const] = ACTIONS(2987),
    [anon_sym___imag] = ACTIONS(2987),
    [anon_sym___kindof] = ACTIONS(2987),
    [anon_sym___nonnull] = ACTIONS(2987),
    [anon_sym___nullable] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2987),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2987),
    [anon_sym___real] = ACTIONS(2987),
    [anon_sym___strong] = ACTIONS(2987),
    [anon_sym___unsafe_unretained] = ACTIONS(2987),
    [anon_sym___unused] = ACTIONS(2987),
    [anon_sym___weak] = ACTIONS(2987),
    [anon_sym_alignas] = ACTIONS(2987),
    [anon_sym__Alignas] = ACTIONS(2987),
    [sym_primitive_type] = ACTIONS(2987),
    [anon_sym_enum] = ACTIONS(2987),
    [anon_sym_struct] = ACTIONS(2987),
    [anon_sym_union] = ACTIONS(2987),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2987),
    [anon_sym___typeof] = ACTIONS(2987),
    [anon_sym_typeof] = ACTIONS(2987),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2987),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2987),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2987),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2987),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2987),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_API_AVAILABLE] = ACTIONS(2987),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_API_DEPRECATED] = ACTIONS(2987),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2987),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2987),
    [anon_sym___deprecated_msg] = ACTIONS(2987),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2987),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2987),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2987),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2987),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2987),
    [anon_sym_BOOL] = ACTIONS(2987),
    [anon_sym_IMP] = ACTIONS(2987),
    [anon_sym_SEL] = ACTIONS(2987),
    [anon_sym_Class] = ACTIONS(2987),
    [anon_sym_id] = ACTIONS(2987),
  },
  [STATE(2899)] = {
    [sym__declaration_modifiers] = STATE(6102),
    [sym_attribute_specifier] = STATE(5761),
    [sym_attribute_declaration] = STATE(6102),
    [sym_ms_declspec_modifier] = STATE(6102),
    [sym_storage_class_specifier] = STATE(6102),
    [sym_type_qualifier] = STATE(6102),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3162),
    [sym_availability_attribute_specifier] = STATE(6102),
    [aux_sym_declaration_repeat1] = STATE(3163),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6816),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(2900)] = {
    [sym_attribute_specifier] = STATE(2948),
    [sym_field_declaration_list] = STATE(2927),
    [sym_identifier] = ACTIONS(7036),
    [anon_sym_COMMA] = ACTIONS(7038),
    [anon_sym_RPAREN] = ACTIONS(7038),
    [anon_sym_LPAREN2] = ACTIONS(7038),
    [anon_sym_STAR] = ACTIONS(7038),
    [anon_sym_CARET] = ACTIONS(7038),
    [anon_sym_SEMI] = ACTIONS(7038),
    [anon_sym___extension__] = ACTIONS(7036),
    [anon_sym_extern] = ACTIONS(7036),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7036),
    [anon_sym_LBRACK] = ACTIONS(7038),
    [anon_sym___declspec] = ACTIONS(7036),
    [anon_sym___based] = ACTIONS(7036),
    [anon_sym_LBRACE] = ACTIONS(6966),
    [anon_sym_signed] = ACTIONS(7036),
    [anon_sym_unsigned] = ACTIONS(7036),
    [anon_sym_long] = ACTIONS(7036),
    [anon_sym_short] = ACTIONS(7036),
    [anon_sym_static] = ACTIONS(7036),
    [anon_sym_auto] = ACTIONS(7036),
    [anon_sym_register] = ACTIONS(7036),
    [anon_sym_inline] = ACTIONS(7036),
    [anon_sym___inline] = ACTIONS(7036),
    [anon_sym___inline__] = ACTIONS(7036),
    [anon_sym___forceinline] = ACTIONS(7036),
    [anon_sym_thread_local] = ACTIONS(7036),
    [anon_sym___thread] = ACTIONS(7036),
    [anon_sym_CG_EXTERN] = ACTIONS(7036),
    [anon_sym_CG_INLINE] = ACTIONS(7036),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7036),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7036),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7036),
    [anon_sym_IBOutlet] = ACTIONS(7036),
    [anon_sym_IBInspectable] = ACTIONS(7036),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7036),
    [anon_sym_NS_INLINE] = ACTIONS(7036),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7036),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7036),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7036),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7036),
    [anon_sym_const] = ACTIONS(7036),
    [anon_sym_constexpr] = ACTIONS(7036),
    [anon_sym_volatile] = ACTIONS(7036),
    [anon_sym_restrict] = ACTIONS(7036),
    [anon_sym___restrict__] = ACTIONS(7036),
    [anon_sym__Atomic] = ACTIONS(7036),
    [anon_sym__Noreturn] = ACTIONS(7036),
    [anon_sym__Nonnull] = ACTIONS(7036),
    [anon_sym_nullable] = ACTIONS(7036),
    [anon_sym__Complex] = ACTIONS(7036),
    [anon_sym__Nullable] = ACTIONS(7036),
    [anon_sym__Nullable_result] = ACTIONS(7036),
    [anon_sym__Null_unspecified] = ACTIONS(7036),
    [anon_sym___autoreleasing] = ACTIONS(7036),
    [anon_sym___block] = ACTIONS(7036),
    [anon_sym___bridge] = ACTIONS(7036),
    [anon_sym___bridge_retained] = ACTIONS(7036),
    [anon_sym___bridge_transfer] = ACTIONS(7036),
    [anon_sym___complex] = ACTIONS(7036),
    [anon_sym___const] = ACTIONS(7036),
    [anon_sym___imag] = ACTIONS(7036),
    [anon_sym___kindof] = ACTIONS(7036),
    [anon_sym___nonnull] = ACTIONS(7036),
    [anon_sym___nullable] = ACTIONS(7036),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7036),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7036),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7036),
    [anon_sym___real] = ACTIONS(7036),
    [anon_sym___strong] = ACTIONS(7036),
    [anon_sym___unsafe_unretained] = ACTIONS(7036),
    [anon_sym___unused] = ACTIONS(7036),
    [anon_sym___weak] = ACTIONS(7036),
    [anon_sym_alignas] = ACTIONS(7036),
    [anon_sym__Alignas] = ACTIONS(7036),
    [sym_primitive_type] = ACTIONS(7036),
    [anon_sym_enum] = ACTIONS(7036),
    [anon_sym_COLON] = ACTIONS(7038),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7036),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7036),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7036),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7036),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7036),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7036),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7036),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7036),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7036),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7036),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7036),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7036),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7036),
    [anon_sym_API_AVAILABLE] = ACTIONS(7036),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7036),
    [anon_sym_API_DEPRECATED] = ACTIONS(7036),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7036),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7036),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7036),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7036),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7036),
    [anon_sym___deprecated_msg] = ACTIONS(7036),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7036),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7036),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7036),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7036),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7036),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7036),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7036),
  },
  [STATE(2901)] = {
    [sym_identifier] = ACTIONS(7014),
    [anon_sym___extension__] = ACTIONS(7014),
    [anon_sym_extern] = ACTIONS(7014),
    [anon_sym___attribute__] = ACTIONS(7014),
    [anon_sym___attribute] = ACTIONS(7014),
    [anon_sym_noreturn] = ACTIONS(7014),
    [anon_sym_LBRACK] = ACTIONS(7016),
    [anon_sym___declspec] = ACTIONS(7014),
    [anon_sym_signed] = ACTIONS(7014),
    [anon_sym_unsigned] = ACTIONS(7014),
    [anon_sym_long] = ACTIONS(7014),
    [anon_sym_short] = ACTIONS(7014),
    [anon_sym_static] = ACTIONS(7014),
    [anon_sym_auto] = ACTIONS(7014),
    [anon_sym_register] = ACTIONS(7014),
    [anon_sym_inline] = ACTIONS(7014),
    [anon_sym___inline] = ACTIONS(7014),
    [anon_sym___inline__] = ACTIONS(7014),
    [anon_sym___forceinline] = ACTIONS(7014),
    [anon_sym_thread_local] = ACTIONS(7014),
    [anon_sym___thread] = ACTIONS(7014),
    [anon_sym_CG_EXTERN] = ACTIONS(7014),
    [anon_sym_CG_INLINE] = ACTIONS(7014),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7014),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7014),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7014),
    [anon_sym_IBOutlet] = ACTIONS(7014),
    [anon_sym_IBInspectable] = ACTIONS(7014),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7014),
    [anon_sym_NS_INLINE] = ACTIONS(7014),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7014),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7014),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7014),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7014),
    [anon_sym_const] = ACTIONS(7014),
    [anon_sym_constexpr] = ACTIONS(7014),
    [anon_sym_volatile] = ACTIONS(7014),
    [anon_sym_restrict] = ACTIONS(7014),
    [anon_sym___restrict__] = ACTIONS(7014),
    [anon_sym__Atomic] = ACTIONS(7014),
    [anon_sym__Noreturn] = ACTIONS(7014),
    [anon_sym__Nonnull] = ACTIONS(7014),
    [anon_sym_nullable] = ACTIONS(7014),
    [anon_sym__Complex] = ACTIONS(7014),
    [anon_sym__Nullable] = ACTIONS(7014),
    [anon_sym__Nullable_result] = ACTIONS(7014),
    [anon_sym__Null_unspecified] = ACTIONS(7014),
    [anon_sym___autoreleasing] = ACTIONS(7014),
    [anon_sym___block] = ACTIONS(7014),
    [anon_sym___bridge] = ACTIONS(7014),
    [anon_sym___bridge_retained] = ACTIONS(7014),
    [anon_sym___bridge_transfer] = ACTIONS(7014),
    [anon_sym___complex] = ACTIONS(7014),
    [anon_sym___const] = ACTIONS(7014),
    [anon_sym___imag] = ACTIONS(7014),
    [anon_sym___kindof] = ACTIONS(7014),
    [anon_sym___nonnull] = ACTIONS(7014),
    [anon_sym___nullable] = ACTIONS(7014),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7014),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7014),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7014),
    [anon_sym___real] = ACTIONS(7014),
    [anon_sym___strong] = ACTIONS(7014),
    [anon_sym___unsafe_unretained] = ACTIONS(7014),
    [anon_sym___unused] = ACTIONS(7014),
    [anon_sym___weak] = ACTIONS(7014),
    [anon_sym_alignas] = ACTIONS(7014),
    [anon_sym__Alignas] = ACTIONS(7014),
    [sym_primitive_type] = ACTIONS(7014),
    [anon_sym_enum] = ACTIONS(7014),
    [anon_sym_struct] = ACTIONS(7014),
    [anon_sym_union] = ACTIONS(7014),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7014),
    [anon_sym___typeof] = ACTIONS(7014),
    [anon_sym_typeof] = ACTIONS(7014),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7014),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7014),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7014),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7014),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7014),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7014),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7014),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7014),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7014),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_API_AVAILABLE] = ACTIONS(7014),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_API_DEPRECATED] = ACTIONS(7014),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7014),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7014),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7014),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7014),
    [anon_sym___deprecated_msg] = ACTIONS(7014),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7014),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7014),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7014),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7014),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7014),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7014),
    [anon_sym_ATproperty] = ACTIONS(7040),
    [anon_sym_BOOL] = ACTIONS(7014),
    [anon_sym_IMP] = ACTIONS(7014),
    [anon_sym_SEL] = ACTIONS(7014),
    [anon_sym_Class] = ACTIONS(7014),
    [anon_sym_id] = ACTIONS(7014),
  },
  [STATE(2902)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_COMMA] = ACTIONS(6134),
    [anon_sym_RPAREN] = ACTIONS(6134),
    [anon_sym_LPAREN2] = ACTIONS(7042),
    [anon_sym_STAR] = ACTIONS(6134),
    [anon_sym_CARET] = ACTIONS(6134),
    [anon_sym_LT] = ACTIONS(7045),
    [anon_sym_SEMI] = ACTIONS(6134),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(6134),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_enum] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(6134),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(2903)] = {
    [sym_identifier] = ACTIONS(6652),
    [anon_sym___extension__] = ACTIONS(6652),
    [anon_sym_extern] = ACTIONS(6652),
    [anon_sym___attribute__] = ACTIONS(6652),
    [anon_sym___attribute] = ACTIONS(6652),
    [anon_sym_noreturn] = ACTIONS(6652),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym___declspec] = ACTIONS(6652),
    [anon_sym_signed] = ACTIONS(6652),
    [anon_sym_unsigned] = ACTIONS(6652),
    [anon_sym_long] = ACTIONS(6652),
    [anon_sym_short] = ACTIONS(6652),
    [anon_sym_static] = ACTIONS(6652),
    [anon_sym_auto] = ACTIONS(6652),
    [anon_sym_register] = ACTIONS(6652),
    [anon_sym_inline] = ACTIONS(6652),
    [anon_sym___inline] = ACTIONS(6652),
    [anon_sym___inline__] = ACTIONS(6652),
    [anon_sym___forceinline] = ACTIONS(6652),
    [anon_sym_thread_local] = ACTIONS(6652),
    [anon_sym___thread] = ACTIONS(6652),
    [anon_sym_CG_EXTERN] = ACTIONS(6652),
    [anon_sym_CG_INLINE] = ACTIONS(6652),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6652),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6652),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6652),
    [anon_sym_IBOutlet] = ACTIONS(6652),
    [anon_sym_IBInspectable] = ACTIONS(6652),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6652),
    [anon_sym_NS_INLINE] = ACTIONS(6652),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6652),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6652),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6652),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6652),
    [anon_sym_const] = ACTIONS(6652),
    [anon_sym_constexpr] = ACTIONS(6652),
    [anon_sym_volatile] = ACTIONS(6652),
    [anon_sym_restrict] = ACTIONS(6652),
    [anon_sym___restrict__] = ACTIONS(6652),
    [anon_sym__Atomic] = ACTIONS(6652),
    [anon_sym__Noreturn] = ACTIONS(6652),
    [anon_sym__Nonnull] = ACTIONS(6652),
    [anon_sym_nullable] = ACTIONS(6652),
    [anon_sym__Complex] = ACTIONS(6652),
    [anon_sym__Nullable] = ACTIONS(6652),
    [anon_sym__Nullable_result] = ACTIONS(6652),
    [anon_sym__Null_unspecified] = ACTIONS(6652),
    [anon_sym___autoreleasing] = ACTIONS(6652),
    [anon_sym___block] = ACTIONS(6652),
    [anon_sym___bridge] = ACTIONS(6652),
    [anon_sym___bridge_retained] = ACTIONS(6652),
    [anon_sym___bridge_transfer] = ACTIONS(6652),
    [anon_sym___complex] = ACTIONS(6652),
    [anon_sym___const] = ACTIONS(6652),
    [anon_sym___imag] = ACTIONS(6652),
    [anon_sym___kindof] = ACTIONS(6652),
    [anon_sym___nonnull] = ACTIONS(6652),
    [anon_sym___nullable] = ACTIONS(6652),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6652),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6652),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6652),
    [anon_sym___real] = ACTIONS(6652),
    [anon_sym___strong] = ACTIONS(6652),
    [anon_sym___unsafe_unretained] = ACTIONS(6652),
    [anon_sym___unused] = ACTIONS(6652),
    [anon_sym___weak] = ACTIONS(6652),
    [anon_sym_alignas] = ACTIONS(6652),
    [anon_sym__Alignas] = ACTIONS(6652),
    [sym_primitive_type] = ACTIONS(6652),
    [anon_sym_enum] = ACTIONS(6652),
    [anon_sym_struct] = ACTIONS(6652),
    [anon_sym_union] = ACTIONS(6652),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6652),
    [anon_sym___typeof] = ACTIONS(6652),
    [anon_sym_typeof] = ACTIONS(6652),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6652),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6652),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6652),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6652),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6652),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6652),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6652),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6652),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6652),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6652),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6652),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6652),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6652),
    [anon_sym_API_AVAILABLE] = ACTIONS(6652),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6652),
    [anon_sym_API_DEPRECATED] = ACTIONS(6652),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6652),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6652),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6652),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6652),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6652),
    [anon_sym___deprecated_msg] = ACTIONS(6652),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6652),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6652),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6652),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6652),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6652),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6652),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6652),
    [anon_sym_BOOL] = ACTIONS(6652),
    [anon_sym_IMP] = ACTIONS(6652),
    [anon_sym_SEL] = ACTIONS(6652),
    [anon_sym_Class] = ACTIONS(6652),
    [anon_sym_id] = ACTIONS(6652),
  },
  [STATE(2904)] = {
    [sym_identifier] = ACTIONS(7049),
    [anon_sym___extension__] = ACTIONS(7049),
    [anon_sym_extern] = ACTIONS(7049),
    [anon_sym___attribute__] = ACTIONS(7049),
    [anon_sym___attribute] = ACTIONS(7049),
    [anon_sym_noreturn] = ACTIONS(7049),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym___declspec] = ACTIONS(7049),
    [anon_sym_signed] = ACTIONS(7049),
    [anon_sym_unsigned] = ACTIONS(7049),
    [anon_sym_long] = ACTIONS(7049),
    [anon_sym_short] = ACTIONS(7049),
    [anon_sym_static] = ACTIONS(7049),
    [anon_sym_auto] = ACTIONS(7049),
    [anon_sym_register] = ACTIONS(7049),
    [anon_sym_inline] = ACTIONS(7049),
    [anon_sym___inline] = ACTIONS(7049),
    [anon_sym___inline__] = ACTIONS(7049),
    [anon_sym___forceinline] = ACTIONS(7049),
    [anon_sym_thread_local] = ACTIONS(7049),
    [anon_sym___thread] = ACTIONS(7049),
    [anon_sym_CG_EXTERN] = ACTIONS(7049),
    [anon_sym_CG_INLINE] = ACTIONS(7049),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7049),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7049),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7049),
    [anon_sym_IBOutlet] = ACTIONS(7049),
    [anon_sym_IBInspectable] = ACTIONS(7049),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7049),
    [anon_sym_NS_INLINE] = ACTIONS(7049),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7049),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7049),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7049),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7049),
    [anon_sym_const] = ACTIONS(7049),
    [anon_sym_constexpr] = ACTIONS(7049),
    [anon_sym_volatile] = ACTIONS(7049),
    [anon_sym_restrict] = ACTIONS(7049),
    [anon_sym___restrict__] = ACTIONS(7049),
    [anon_sym__Atomic] = ACTIONS(7049),
    [anon_sym__Noreturn] = ACTIONS(7049),
    [anon_sym__Nonnull] = ACTIONS(7049),
    [anon_sym_nullable] = ACTIONS(7049),
    [anon_sym__Complex] = ACTIONS(7049),
    [anon_sym__Nullable] = ACTIONS(7049),
    [anon_sym__Nullable_result] = ACTIONS(7049),
    [anon_sym__Null_unspecified] = ACTIONS(7049),
    [anon_sym___autoreleasing] = ACTIONS(7049),
    [anon_sym___block] = ACTIONS(7049),
    [anon_sym___bridge] = ACTIONS(7049),
    [anon_sym___bridge_retained] = ACTIONS(7049),
    [anon_sym___bridge_transfer] = ACTIONS(7049),
    [anon_sym___complex] = ACTIONS(7049),
    [anon_sym___const] = ACTIONS(7049),
    [anon_sym___imag] = ACTIONS(7049),
    [anon_sym___kindof] = ACTIONS(7049),
    [anon_sym___nonnull] = ACTIONS(7049),
    [anon_sym___nullable] = ACTIONS(7049),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7049),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7049),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7049),
    [anon_sym___real] = ACTIONS(7049),
    [anon_sym___strong] = ACTIONS(7049),
    [anon_sym___unsafe_unretained] = ACTIONS(7049),
    [anon_sym___unused] = ACTIONS(7049),
    [anon_sym___weak] = ACTIONS(7049),
    [anon_sym_alignas] = ACTIONS(7049),
    [anon_sym__Alignas] = ACTIONS(7049),
    [sym_primitive_type] = ACTIONS(7049),
    [anon_sym_enum] = ACTIONS(7049),
    [anon_sym_struct] = ACTIONS(7049),
    [anon_sym_union] = ACTIONS(7049),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7049),
    [anon_sym___typeof] = ACTIONS(7049),
    [anon_sym_typeof] = ACTIONS(7049),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7049),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7049),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7049),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7049),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7049),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7049),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7049),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7049),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7049),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7049),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7049),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7049),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7049),
    [anon_sym_API_AVAILABLE] = ACTIONS(7049),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7049),
    [anon_sym_API_DEPRECATED] = ACTIONS(7049),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7049),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7049),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7049),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7049),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7049),
    [anon_sym___deprecated_msg] = ACTIONS(7049),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7049),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7049),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7049),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7049),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7049),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7049),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7049),
    [anon_sym_BOOL] = ACTIONS(7049),
    [anon_sym_IMP] = ACTIONS(7049),
    [anon_sym_SEL] = ACTIONS(7049),
    [anon_sym_Class] = ACTIONS(7049),
    [anon_sym_id] = ACTIONS(7049),
  },
  [STATE(2905)] = {
    [aux_sym_generic_specifier_repeat1] = STATE(2905),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_RPAREN] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7055),
    [anon_sym_STAR] = ACTIONS(7055),
    [anon_sym_CARET] = ACTIONS(7055),
    [anon_sym_LT] = ACTIONS(7057),
    [anon_sym_SEMI] = ACTIONS(7055),
    [anon_sym___extension__] = ACTIONS(7053),
    [anon_sym_extern] = ACTIONS(7053),
    [anon_sym___attribute__] = ACTIONS(7053),
    [anon_sym___attribute] = ACTIONS(7053),
    [anon_sym_noreturn] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7055),
    [anon_sym___declspec] = ACTIONS(7053),
    [anon_sym___based] = ACTIONS(7053),
    [anon_sym_signed] = ACTIONS(7053),
    [anon_sym_unsigned] = ACTIONS(7053),
    [anon_sym_long] = ACTIONS(7053),
    [anon_sym_short] = ACTIONS(7053),
    [anon_sym_static] = ACTIONS(7053),
    [anon_sym_auto] = ACTIONS(7053),
    [anon_sym_register] = ACTIONS(7053),
    [anon_sym_inline] = ACTIONS(7053),
    [anon_sym___inline] = ACTIONS(7053),
    [anon_sym___inline__] = ACTIONS(7053),
    [anon_sym___forceinline] = ACTIONS(7053),
    [anon_sym_thread_local] = ACTIONS(7053),
    [anon_sym___thread] = ACTIONS(7053),
    [anon_sym_CG_EXTERN] = ACTIONS(7053),
    [anon_sym_CG_INLINE] = ACTIONS(7053),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7053),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7053),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7053),
    [anon_sym_IBOutlet] = ACTIONS(7053),
    [anon_sym_IBInspectable] = ACTIONS(7053),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7053),
    [anon_sym_NS_INLINE] = ACTIONS(7053),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7053),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7053),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7053),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7053),
    [anon_sym_const] = ACTIONS(7053),
    [anon_sym_constexpr] = ACTIONS(7053),
    [anon_sym_volatile] = ACTIONS(7053),
    [anon_sym_restrict] = ACTIONS(7053),
    [anon_sym___restrict__] = ACTIONS(7053),
    [anon_sym__Atomic] = ACTIONS(7053),
    [anon_sym__Noreturn] = ACTIONS(7053),
    [anon_sym__Nonnull] = ACTIONS(7053),
    [anon_sym_nullable] = ACTIONS(7053),
    [anon_sym__Complex] = ACTIONS(7053),
    [anon_sym__Nullable] = ACTIONS(7053),
    [anon_sym__Nullable_result] = ACTIONS(7053),
    [anon_sym__Null_unspecified] = ACTIONS(7053),
    [anon_sym___autoreleasing] = ACTIONS(7053),
    [anon_sym___block] = ACTIONS(7053),
    [anon_sym___bridge] = ACTIONS(7053),
    [anon_sym___bridge_retained] = ACTIONS(7053),
    [anon_sym___bridge_transfer] = ACTIONS(7053),
    [anon_sym___complex] = ACTIONS(7053),
    [anon_sym___const] = ACTIONS(7053),
    [anon_sym___imag] = ACTIONS(7053),
    [anon_sym___kindof] = ACTIONS(7053),
    [anon_sym___nonnull] = ACTIONS(7053),
    [anon_sym___nullable] = ACTIONS(7053),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7053),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7053),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7053),
    [anon_sym___real] = ACTIONS(7053),
    [anon_sym___strong] = ACTIONS(7053),
    [anon_sym___unsafe_unretained] = ACTIONS(7053),
    [anon_sym___unused] = ACTIONS(7053),
    [anon_sym___weak] = ACTIONS(7053),
    [anon_sym_alignas] = ACTIONS(7053),
    [anon_sym__Alignas] = ACTIONS(7053),
    [sym_primitive_type] = ACTIONS(7053),
    [anon_sym_enum] = ACTIONS(7053),
    [anon_sym_COLON] = ACTIONS(7055),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7053),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7053),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7053),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7053),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7053),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7053),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7053),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7053),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7053),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7053),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7053),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7053),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7053),
    [anon_sym_API_AVAILABLE] = ACTIONS(7053),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7053),
    [anon_sym_API_DEPRECATED] = ACTIONS(7053),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7053),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7053),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7053),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7053),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7053),
    [anon_sym___deprecated_msg] = ACTIONS(7053),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7053),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7053),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7053),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7053),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7053),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7053),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7053),
  },
  [STATE(2906)] = {
    [sym_identifier] = ACTIONS(7060),
    [anon_sym___extension__] = ACTIONS(7060),
    [anon_sym_extern] = ACTIONS(7060),
    [anon_sym___attribute__] = ACTIONS(7060),
    [anon_sym___attribute] = ACTIONS(7060),
    [anon_sym_noreturn] = ACTIONS(7060),
    [anon_sym_LBRACK] = ACTIONS(7062),
    [anon_sym___declspec] = ACTIONS(7060),
    [anon_sym_signed] = ACTIONS(7060),
    [anon_sym_unsigned] = ACTIONS(7060),
    [anon_sym_long] = ACTIONS(7060),
    [anon_sym_short] = ACTIONS(7060),
    [anon_sym_static] = ACTIONS(7060),
    [anon_sym_auto] = ACTIONS(7060),
    [anon_sym_register] = ACTIONS(7060),
    [anon_sym_inline] = ACTIONS(7060),
    [anon_sym___inline] = ACTIONS(7060),
    [anon_sym___inline__] = ACTIONS(7060),
    [anon_sym___forceinline] = ACTIONS(7060),
    [anon_sym_thread_local] = ACTIONS(7060),
    [anon_sym___thread] = ACTIONS(7060),
    [anon_sym_CG_EXTERN] = ACTIONS(7060),
    [anon_sym_CG_INLINE] = ACTIONS(7060),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7060),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7060),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7060),
    [anon_sym_IBOutlet] = ACTIONS(7060),
    [anon_sym_IBInspectable] = ACTIONS(7060),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7060),
    [anon_sym_NS_INLINE] = ACTIONS(7060),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7060),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7060),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7060),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7060),
    [anon_sym_const] = ACTIONS(7060),
    [anon_sym_constexpr] = ACTIONS(7060),
    [anon_sym_volatile] = ACTIONS(7060),
    [anon_sym_restrict] = ACTIONS(7060),
    [anon_sym___restrict__] = ACTIONS(7060),
    [anon_sym__Atomic] = ACTIONS(7060),
    [anon_sym__Noreturn] = ACTIONS(7060),
    [anon_sym__Nonnull] = ACTIONS(7060),
    [anon_sym_nullable] = ACTIONS(7060),
    [anon_sym__Complex] = ACTIONS(7060),
    [anon_sym__Nullable] = ACTIONS(7060),
    [anon_sym__Nullable_result] = ACTIONS(7060),
    [anon_sym__Null_unspecified] = ACTIONS(7060),
    [anon_sym___autoreleasing] = ACTIONS(7060),
    [anon_sym___block] = ACTIONS(7060),
    [anon_sym___bridge] = ACTIONS(7060),
    [anon_sym___bridge_retained] = ACTIONS(7060),
    [anon_sym___bridge_transfer] = ACTIONS(7060),
    [anon_sym___complex] = ACTIONS(7060),
    [anon_sym___const] = ACTIONS(7060),
    [anon_sym___imag] = ACTIONS(7060),
    [anon_sym___kindof] = ACTIONS(7060),
    [anon_sym___nonnull] = ACTIONS(7060),
    [anon_sym___nullable] = ACTIONS(7060),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7060),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7060),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7060),
    [anon_sym___real] = ACTIONS(7060),
    [anon_sym___strong] = ACTIONS(7060),
    [anon_sym___unsafe_unretained] = ACTIONS(7060),
    [anon_sym___unused] = ACTIONS(7060),
    [anon_sym___weak] = ACTIONS(7060),
    [anon_sym_alignas] = ACTIONS(7060),
    [anon_sym__Alignas] = ACTIONS(7060),
    [sym_primitive_type] = ACTIONS(7060),
    [anon_sym_enum] = ACTIONS(7060),
    [anon_sym_struct] = ACTIONS(7060),
    [anon_sym_union] = ACTIONS(7060),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7060),
    [anon_sym___typeof] = ACTIONS(7060),
    [anon_sym_typeof] = ACTIONS(7060),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7060),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7060),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7060),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7060),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7060),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7060),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7060),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7060),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7060),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7060),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7060),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7060),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7060),
    [anon_sym_API_AVAILABLE] = ACTIONS(7060),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7060),
    [anon_sym_API_DEPRECATED] = ACTIONS(7060),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7060),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7060),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7060),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7060),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7060),
    [anon_sym___deprecated_msg] = ACTIONS(7060),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7060),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7060),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7060),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7060),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7060),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7060),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7060),
    [anon_sym_BOOL] = ACTIONS(7060),
    [anon_sym_IMP] = ACTIONS(7060),
    [anon_sym_SEL] = ACTIONS(7060),
    [anon_sym_Class] = ACTIONS(7060),
    [anon_sym_id] = ACTIONS(7060),
  },
  [STATE(2907)] = {
    [sym_identifier] = ACTIONS(7064),
    [anon_sym___extension__] = ACTIONS(7064),
    [anon_sym_extern] = ACTIONS(7064),
    [anon_sym___attribute__] = ACTIONS(7064),
    [anon_sym___attribute] = ACTIONS(7064),
    [anon_sym_noreturn] = ACTIONS(7064),
    [anon_sym_LBRACK] = ACTIONS(7066),
    [anon_sym___declspec] = ACTIONS(7064),
    [anon_sym_signed] = ACTIONS(7064),
    [anon_sym_unsigned] = ACTIONS(7064),
    [anon_sym_long] = ACTIONS(7064),
    [anon_sym_short] = ACTIONS(7064),
    [anon_sym_static] = ACTIONS(7064),
    [anon_sym_auto] = ACTIONS(7064),
    [anon_sym_register] = ACTIONS(7064),
    [anon_sym_inline] = ACTIONS(7064),
    [anon_sym___inline] = ACTIONS(7064),
    [anon_sym___inline__] = ACTIONS(7064),
    [anon_sym___forceinline] = ACTIONS(7064),
    [anon_sym_thread_local] = ACTIONS(7064),
    [anon_sym___thread] = ACTIONS(7064),
    [anon_sym_CG_EXTERN] = ACTIONS(7064),
    [anon_sym_CG_INLINE] = ACTIONS(7064),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7064),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7064),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7064),
    [anon_sym_IBOutlet] = ACTIONS(7064),
    [anon_sym_IBInspectable] = ACTIONS(7064),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7064),
    [anon_sym_NS_INLINE] = ACTIONS(7064),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7064),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7064),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7064),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7064),
    [anon_sym_const] = ACTIONS(7064),
    [anon_sym_constexpr] = ACTIONS(7064),
    [anon_sym_volatile] = ACTIONS(7064),
    [anon_sym_restrict] = ACTIONS(7064),
    [anon_sym___restrict__] = ACTIONS(7064),
    [anon_sym__Atomic] = ACTIONS(7064),
    [anon_sym__Noreturn] = ACTIONS(7064),
    [anon_sym__Nonnull] = ACTIONS(7064),
    [anon_sym_nullable] = ACTIONS(7064),
    [anon_sym__Complex] = ACTIONS(7064),
    [anon_sym__Nullable] = ACTIONS(7064),
    [anon_sym__Nullable_result] = ACTIONS(7064),
    [anon_sym__Null_unspecified] = ACTIONS(7064),
    [anon_sym___autoreleasing] = ACTIONS(7064),
    [anon_sym___block] = ACTIONS(7064),
    [anon_sym___bridge] = ACTIONS(7064),
    [anon_sym___bridge_retained] = ACTIONS(7064),
    [anon_sym___bridge_transfer] = ACTIONS(7064),
    [anon_sym___complex] = ACTIONS(7064),
    [anon_sym___const] = ACTIONS(7064),
    [anon_sym___imag] = ACTIONS(7064),
    [anon_sym___kindof] = ACTIONS(7064),
    [anon_sym___nonnull] = ACTIONS(7064),
    [anon_sym___nullable] = ACTIONS(7064),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7064),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7064),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7064),
    [anon_sym___real] = ACTIONS(7064),
    [anon_sym___strong] = ACTIONS(7064),
    [anon_sym___unsafe_unretained] = ACTIONS(7064),
    [anon_sym___unused] = ACTIONS(7064),
    [anon_sym___weak] = ACTIONS(7064),
    [anon_sym_alignas] = ACTIONS(7064),
    [anon_sym__Alignas] = ACTIONS(7064),
    [sym_primitive_type] = ACTIONS(7064),
    [anon_sym_enum] = ACTIONS(7064),
    [anon_sym_struct] = ACTIONS(7064),
    [anon_sym_union] = ACTIONS(7064),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7064),
    [anon_sym___typeof] = ACTIONS(7064),
    [anon_sym_typeof] = ACTIONS(7064),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7064),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7064),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7064),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7064),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7064),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7064),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7064),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7064),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7064),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7064),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7064),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7064),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7064),
    [anon_sym_API_AVAILABLE] = ACTIONS(7064),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7064),
    [anon_sym_API_DEPRECATED] = ACTIONS(7064),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7064),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7064),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7064),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7064),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7064),
    [anon_sym___deprecated_msg] = ACTIONS(7064),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7064),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7064),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7064),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7064),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7064),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7064),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7064),
    [anon_sym_BOOL] = ACTIONS(7064),
    [anon_sym_IMP] = ACTIONS(7064),
    [anon_sym_SEL] = ACTIONS(7064),
    [anon_sym_Class] = ACTIONS(7064),
    [anon_sym_id] = ACTIONS(7064),
  },
  [STATE(2908)] = {
    [aux_sym_generic_specifier_repeat1] = STATE(2905),
    [sym_identifier] = ACTIONS(7068),
    [anon_sym_COMMA] = ACTIONS(7070),
    [anon_sym_RPAREN] = ACTIONS(7070),
    [anon_sym_LPAREN2] = ACTIONS(7070),
    [anon_sym_STAR] = ACTIONS(7070),
    [anon_sym_CARET] = ACTIONS(7070),
    [anon_sym_LT] = ACTIONS(7045),
    [anon_sym_SEMI] = ACTIONS(7070),
    [anon_sym___extension__] = ACTIONS(7068),
    [anon_sym_extern] = ACTIONS(7068),
    [anon_sym___attribute__] = ACTIONS(7068),
    [anon_sym___attribute] = ACTIONS(7068),
    [anon_sym_noreturn] = ACTIONS(7068),
    [anon_sym_LBRACK] = ACTIONS(7070),
    [anon_sym___declspec] = ACTIONS(7068),
    [anon_sym___based] = ACTIONS(7068),
    [anon_sym_signed] = ACTIONS(7068),
    [anon_sym_unsigned] = ACTIONS(7068),
    [anon_sym_long] = ACTIONS(7068),
    [anon_sym_short] = ACTIONS(7068),
    [anon_sym_static] = ACTIONS(7068),
    [anon_sym_auto] = ACTIONS(7068),
    [anon_sym_register] = ACTIONS(7068),
    [anon_sym_inline] = ACTIONS(7068),
    [anon_sym___inline] = ACTIONS(7068),
    [anon_sym___inline__] = ACTIONS(7068),
    [anon_sym___forceinline] = ACTIONS(7068),
    [anon_sym_thread_local] = ACTIONS(7068),
    [anon_sym___thread] = ACTIONS(7068),
    [anon_sym_CG_EXTERN] = ACTIONS(7068),
    [anon_sym_CG_INLINE] = ACTIONS(7068),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7068),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7068),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7068),
    [anon_sym_IBOutlet] = ACTIONS(7068),
    [anon_sym_IBInspectable] = ACTIONS(7068),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7068),
    [anon_sym_NS_INLINE] = ACTIONS(7068),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7068),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7068),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7068),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7068),
    [anon_sym_const] = ACTIONS(7068),
    [anon_sym_constexpr] = ACTIONS(7068),
    [anon_sym_volatile] = ACTIONS(7068),
    [anon_sym_restrict] = ACTIONS(7068),
    [anon_sym___restrict__] = ACTIONS(7068),
    [anon_sym__Atomic] = ACTIONS(7068),
    [anon_sym__Noreturn] = ACTIONS(7068),
    [anon_sym__Nonnull] = ACTIONS(7068),
    [anon_sym_nullable] = ACTIONS(7068),
    [anon_sym__Complex] = ACTIONS(7068),
    [anon_sym__Nullable] = ACTIONS(7068),
    [anon_sym__Nullable_result] = ACTIONS(7068),
    [anon_sym__Null_unspecified] = ACTIONS(7068),
    [anon_sym___autoreleasing] = ACTIONS(7068),
    [anon_sym___block] = ACTIONS(7068),
    [anon_sym___bridge] = ACTIONS(7068),
    [anon_sym___bridge_retained] = ACTIONS(7068),
    [anon_sym___bridge_transfer] = ACTIONS(7068),
    [anon_sym___complex] = ACTIONS(7068),
    [anon_sym___const] = ACTIONS(7068),
    [anon_sym___imag] = ACTIONS(7068),
    [anon_sym___kindof] = ACTIONS(7068),
    [anon_sym___nonnull] = ACTIONS(7068),
    [anon_sym___nullable] = ACTIONS(7068),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7068),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7068),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7068),
    [anon_sym___real] = ACTIONS(7068),
    [anon_sym___strong] = ACTIONS(7068),
    [anon_sym___unsafe_unretained] = ACTIONS(7068),
    [anon_sym___unused] = ACTIONS(7068),
    [anon_sym___weak] = ACTIONS(7068),
    [anon_sym_alignas] = ACTIONS(7068),
    [anon_sym__Alignas] = ACTIONS(7068),
    [sym_primitive_type] = ACTIONS(7068),
    [anon_sym_enum] = ACTIONS(7068),
    [anon_sym_COLON] = ACTIONS(7070),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7068),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7068),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7068),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7068),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7068),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7068),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7068),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7068),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7068),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7068),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7068),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7068),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7068),
    [anon_sym_API_AVAILABLE] = ACTIONS(7068),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7068),
    [anon_sym_API_DEPRECATED] = ACTIONS(7068),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7068),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7068),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7068),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7068),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7068),
    [anon_sym___deprecated_msg] = ACTIONS(7068),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7068),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7068),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7068),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7068),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7068),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7068),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7068),
  },
  [STATE(2909)] = {
    [sym_protocol_reference_list] = STATE(2964),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_RPAREN] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym_SEMI] = ACTIONS(6310),
    [anon_sym___extension__] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [anon_sym___attribute__] = ACTIONS(6308),
    [anon_sym___attribute] = ACTIONS(6308),
    [anon_sym_noreturn] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6310),
    [anon_sym___declspec] = ACTIONS(6308),
    [anon_sym___based] = ACTIONS(6308),
    [anon_sym_signed] = ACTIONS(6308),
    [anon_sym_unsigned] = ACTIONS(6308),
    [anon_sym_long] = ACTIONS(6308),
    [anon_sym_short] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_auto] = ACTIONS(6308),
    [anon_sym_register] = ACTIONS(6308),
    [anon_sym_inline] = ACTIONS(6308),
    [anon_sym___inline] = ACTIONS(6308),
    [anon_sym___inline__] = ACTIONS(6308),
    [anon_sym___forceinline] = ACTIONS(6308),
    [anon_sym_thread_local] = ACTIONS(6308),
    [anon_sym___thread] = ACTIONS(6308),
    [anon_sym_CG_EXTERN] = ACTIONS(6308),
    [anon_sym_CG_INLINE] = ACTIONS(6308),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6308),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6308),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6308),
    [anon_sym_IBOutlet] = ACTIONS(6308),
    [anon_sym_IBInspectable] = ACTIONS(6308),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6308),
    [anon_sym_NS_INLINE] = ACTIONS(6308),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6308),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6308),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6308),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6308),
    [anon_sym_const] = ACTIONS(6308),
    [anon_sym_constexpr] = ACTIONS(6308),
    [anon_sym_volatile] = ACTIONS(6308),
    [anon_sym_restrict] = ACTIONS(6308),
    [anon_sym___restrict__] = ACTIONS(6308),
    [anon_sym__Atomic] = ACTIONS(6308),
    [anon_sym__Noreturn] = ACTIONS(6308),
    [anon_sym__Nonnull] = ACTIONS(6308),
    [anon_sym_nullable] = ACTIONS(6308),
    [anon_sym__Complex] = ACTIONS(6308),
    [anon_sym__Nullable] = ACTIONS(6308),
    [anon_sym__Nullable_result] = ACTIONS(6308),
    [anon_sym__Null_unspecified] = ACTIONS(6308),
    [anon_sym___autoreleasing] = ACTIONS(6308),
    [anon_sym___block] = ACTIONS(6308),
    [anon_sym___bridge] = ACTIONS(6308),
    [anon_sym___bridge_retained] = ACTIONS(6308),
    [anon_sym___bridge_transfer] = ACTIONS(6308),
    [anon_sym___complex] = ACTIONS(6308),
    [anon_sym___const] = ACTIONS(6308),
    [anon_sym___imag] = ACTIONS(6308),
    [anon_sym___kindof] = ACTIONS(6308),
    [anon_sym___nonnull] = ACTIONS(6308),
    [anon_sym___nullable] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6308),
    [anon_sym___real] = ACTIONS(6308),
    [anon_sym___strong] = ACTIONS(6308),
    [anon_sym___unsafe_unretained] = ACTIONS(6308),
    [anon_sym___unused] = ACTIONS(6308),
    [anon_sym___weak] = ACTIONS(6308),
    [anon_sym_alignas] = ACTIONS(6308),
    [anon_sym__Alignas] = ACTIONS(6308),
    [sym_primitive_type] = ACTIONS(6308),
    [anon_sym_enum] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6310),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6308),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6308),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6308),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6308),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6308),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6308),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6308),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_API_AVAILABLE] = ACTIONS(6308),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_API_DEPRECATED] = ACTIONS(6308),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6308),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6308),
    [anon_sym___deprecated_msg] = ACTIONS(6308),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6308),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6308),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6308),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6308),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6308),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6308),
  },
  [STATE(2910)] = {
    [sym_attribute_specifier] = STATE(2979),
    [sym_identifier] = ACTIONS(7072),
    [anon_sym_COMMA] = ACTIONS(7074),
    [anon_sym_RPAREN] = ACTIONS(7074),
    [anon_sym_LPAREN2] = ACTIONS(7074),
    [anon_sym_STAR] = ACTIONS(7074),
    [anon_sym_CARET] = ACTIONS(7074),
    [anon_sym_SEMI] = ACTIONS(7074),
    [anon_sym___extension__] = ACTIONS(7072),
    [anon_sym_extern] = ACTIONS(7072),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7072),
    [anon_sym_LBRACK] = ACTIONS(7074),
    [anon_sym___declspec] = ACTIONS(7072),
    [anon_sym___based] = ACTIONS(7072),
    [anon_sym_signed] = ACTIONS(7072),
    [anon_sym_unsigned] = ACTIONS(7072),
    [anon_sym_long] = ACTIONS(7072),
    [anon_sym_short] = ACTIONS(7072),
    [anon_sym_static] = ACTIONS(7072),
    [anon_sym_auto] = ACTIONS(7072),
    [anon_sym_register] = ACTIONS(7072),
    [anon_sym_inline] = ACTIONS(7072),
    [anon_sym___inline] = ACTIONS(7072),
    [anon_sym___inline__] = ACTIONS(7072),
    [anon_sym___forceinline] = ACTIONS(7072),
    [anon_sym_thread_local] = ACTIONS(7072),
    [anon_sym___thread] = ACTIONS(7072),
    [anon_sym_CG_EXTERN] = ACTIONS(7072),
    [anon_sym_CG_INLINE] = ACTIONS(7072),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7072),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7072),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7072),
    [anon_sym_IBOutlet] = ACTIONS(7072),
    [anon_sym_IBInspectable] = ACTIONS(7072),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7072),
    [anon_sym_NS_INLINE] = ACTIONS(7072),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7072),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7072),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7072),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7072),
    [anon_sym_const] = ACTIONS(7072),
    [anon_sym_constexpr] = ACTIONS(7072),
    [anon_sym_volatile] = ACTIONS(7072),
    [anon_sym_restrict] = ACTIONS(7072),
    [anon_sym___restrict__] = ACTIONS(7072),
    [anon_sym__Atomic] = ACTIONS(7072),
    [anon_sym__Noreturn] = ACTIONS(7072),
    [anon_sym__Nonnull] = ACTIONS(7072),
    [anon_sym_nullable] = ACTIONS(7072),
    [anon_sym__Complex] = ACTIONS(7072),
    [anon_sym__Nullable] = ACTIONS(7072),
    [anon_sym__Nullable_result] = ACTIONS(7072),
    [anon_sym__Null_unspecified] = ACTIONS(7072),
    [anon_sym___autoreleasing] = ACTIONS(7072),
    [anon_sym___block] = ACTIONS(7072),
    [anon_sym___bridge] = ACTIONS(7072),
    [anon_sym___bridge_retained] = ACTIONS(7072),
    [anon_sym___bridge_transfer] = ACTIONS(7072),
    [anon_sym___complex] = ACTIONS(7072),
    [anon_sym___const] = ACTIONS(7072),
    [anon_sym___imag] = ACTIONS(7072),
    [anon_sym___kindof] = ACTIONS(7072),
    [anon_sym___nonnull] = ACTIONS(7072),
    [anon_sym___nullable] = ACTIONS(7072),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7072),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7072),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7072),
    [anon_sym___real] = ACTIONS(7072),
    [anon_sym___strong] = ACTIONS(7072),
    [anon_sym___unsafe_unretained] = ACTIONS(7072),
    [anon_sym___unused] = ACTIONS(7072),
    [anon_sym___weak] = ACTIONS(7072),
    [anon_sym_alignas] = ACTIONS(7072),
    [anon_sym__Alignas] = ACTIONS(7072),
    [sym_primitive_type] = ACTIONS(7072),
    [anon_sym_enum] = ACTIONS(7072),
    [anon_sym_COLON] = ACTIONS(7074),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7072),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7072),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7072),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7072),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7072),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7072),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7072),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7072),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7072),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7072),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7072),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7072),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7072),
    [anon_sym_API_AVAILABLE] = ACTIONS(7072),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7072),
    [anon_sym_API_DEPRECATED] = ACTIONS(7072),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7072),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7072),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7072),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7072),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7072),
    [anon_sym___deprecated_msg] = ACTIONS(7072),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7072),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7072),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7072),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7072),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7072),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7072),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7072),
  },
  [STATE(2911)] = {
    [sym_attribute_specifier] = STATE(2957),
    [sym_identifier] = ACTIONS(7076),
    [anon_sym_COMMA] = ACTIONS(7078),
    [anon_sym_RPAREN] = ACTIONS(7078),
    [anon_sym_LPAREN2] = ACTIONS(7078),
    [anon_sym_STAR] = ACTIONS(7078),
    [anon_sym_CARET] = ACTIONS(7078),
    [anon_sym_SEMI] = ACTIONS(7078),
    [anon_sym___extension__] = ACTIONS(7076),
    [anon_sym_extern] = ACTIONS(7076),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7076),
    [anon_sym_LBRACK] = ACTIONS(7078),
    [anon_sym___declspec] = ACTIONS(7076),
    [anon_sym___based] = ACTIONS(7076),
    [anon_sym_signed] = ACTIONS(7076),
    [anon_sym_unsigned] = ACTIONS(7076),
    [anon_sym_long] = ACTIONS(7076),
    [anon_sym_short] = ACTIONS(7076),
    [anon_sym_static] = ACTIONS(7076),
    [anon_sym_auto] = ACTIONS(7076),
    [anon_sym_register] = ACTIONS(7076),
    [anon_sym_inline] = ACTIONS(7076),
    [anon_sym___inline] = ACTIONS(7076),
    [anon_sym___inline__] = ACTIONS(7076),
    [anon_sym___forceinline] = ACTIONS(7076),
    [anon_sym_thread_local] = ACTIONS(7076),
    [anon_sym___thread] = ACTIONS(7076),
    [anon_sym_CG_EXTERN] = ACTIONS(7076),
    [anon_sym_CG_INLINE] = ACTIONS(7076),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7076),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7076),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7076),
    [anon_sym_IBOutlet] = ACTIONS(7076),
    [anon_sym_IBInspectable] = ACTIONS(7076),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7076),
    [anon_sym_NS_INLINE] = ACTIONS(7076),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7076),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7076),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7076),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7076),
    [anon_sym_const] = ACTIONS(7076),
    [anon_sym_constexpr] = ACTIONS(7076),
    [anon_sym_volatile] = ACTIONS(7076),
    [anon_sym_restrict] = ACTIONS(7076),
    [anon_sym___restrict__] = ACTIONS(7076),
    [anon_sym__Atomic] = ACTIONS(7076),
    [anon_sym__Noreturn] = ACTIONS(7076),
    [anon_sym__Nonnull] = ACTIONS(7076),
    [anon_sym_nullable] = ACTIONS(7076),
    [anon_sym__Complex] = ACTIONS(7076),
    [anon_sym__Nullable] = ACTIONS(7076),
    [anon_sym__Nullable_result] = ACTIONS(7076),
    [anon_sym__Null_unspecified] = ACTIONS(7076),
    [anon_sym___autoreleasing] = ACTIONS(7076),
    [anon_sym___block] = ACTIONS(7076),
    [anon_sym___bridge] = ACTIONS(7076),
    [anon_sym___bridge_retained] = ACTIONS(7076),
    [anon_sym___bridge_transfer] = ACTIONS(7076),
    [anon_sym___complex] = ACTIONS(7076),
    [anon_sym___const] = ACTIONS(7076),
    [anon_sym___imag] = ACTIONS(7076),
    [anon_sym___kindof] = ACTIONS(7076),
    [anon_sym___nonnull] = ACTIONS(7076),
    [anon_sym___nullable] = ACTIONS(7076),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7076),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7076),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7076),
    [anon_sym___real] = ACTIONS(7076),
    [anon_sym___strong] = ACTIONS(7076),
    [anon_sym___unsafe_unretained] = ACTIONS(7076),
    [anon_sym___unused] = ACTIONS(7076),
    [anon_sym___weak] = ACTIONS(7076),
    [anon_sym_alignas] = ACTIONS(7076),
    [anon_sym__Alignas] = ACTIONS(7076),
    [sym_primitive_type] = ACTIONS(7076),
    [anon_sym_enum] = ACTIONS(7076),
    [anon_sym_COLON] = ACTIONS(7078),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7076),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7076),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7076),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7076),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7076),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7076),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7076),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7076),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7076),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7076),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7076),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7076),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7076),
    [anon_sym_API_AVAILABLE] = ACTIONS(7076),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7076),
    [anon_sym_API_DEPRECATED] = ACTIONS(7076),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7076),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7076),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7076),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7076),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7076),
    [anon_sym___deprecated_msg] = ACTIONS(7076),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7076),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7076),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7076),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7076),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7076),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7076),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7076),
  },
  [STATE(2912)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2912),
    [sym_identifier] = ACTIONS(7080),
    [anon_sym_COMMA] = ACTIONS(7082),
    [anon_sym_RPAREN] = ACTIONS(7082),
    [anon_sym_LPAREN2] = ACTIONS(7082),
    [anon_sym_STAR] = ACTIONS(7082),
    [anon_sym_CARET] = ACTIONS(7082),
    [anon_sym_SEMI] = ACTIONS(7082),
    [anon_sym___extension__] = ACTIONS(7080),
    [anon_sym_extern] = ACTIONS(7080),
    [anon_sym___attribute__] = ACTIONS(7080),
    [anon_sym___attribute] = ACTIONS(7080),
    [anon_sym_noreturn] = ACTIONS(7080),
    [anon_sym_LBRACK] = ACTIONS(7082),
    [anon_sym___declspec] = ACTIONS(7080),
    [anon_sym___based] = ACTIONS(7080),
    [anon_sym_signed] = ACTIONS(7084),
    [anon_sym_unsigned] = ACTIONS(7084),
    [anon_sym_long] = ACTIONS(7084),
    [anon_sym_short] = ACTIONS(7084),
    [anon_sym_static] = ACTIONS(7080),
    [anon_sym_auto] = ACTIONS(7080),
    [anon_sym_register] = ACTIONS(7080),
    [anon_sym_inline] = ACTIONS(7080),
    [anon_sym___inline] = ACTIONS(7080),
    [anon_sym___inline__] = ACTIONS(7080),
    [anon_sym___forceinline] = ACTIONS(7080),
    [anon_sym_thread_local] = ACTIONS(7080),
    [anon_sym___thread] = ACTIONS(7080),
    [anon_sym_CG_EXTERN] = ACTIONS(7080),
    [anon_sym_CG_INLINE] = ACTIONS(7080),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7080),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7080),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7080),
    [anon_sym_IBOutlet] = ACTIONS(7080),
    [anon_sym_IBInspectable] = ACTIONS(7080),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7080),
    [anon_sym_NS_INLINE] = ACTIONS(7080),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7080),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7080),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7080),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7080),
    [anon_sym_const] = ACTIONS(7080),
    [anon_sym_constexpr] = ACTIONS(7080),
    [anon_sym_volatile] = ACTIONS(7080),
    [anon_sym_restrict] = ACTIONS(7080),
    [anon_sym___restrict__] = ACTIONS(7080),
    [anon_sym__Atomic] = ACTIONS(7080),
    [anon_sym__Noreturn] = ACTIONS(7080),
    [anon_sym__Nonnull] = ACTIONS(7080),
    [anon_sym_nullable] = ACTIONS(7080),
    [anon_sym__Complex] = ACTIONS(7080),
    [anon_sym__Nullable] = ACTIONS(7080),
    [anon_sym__Nullable_result] = ACTIONS(7080),
    [anon_sym__Null_unspecified] = ACTIONS(7080),
    [anon_sym___autoreleasing] = ACTIONS(7080),
    [anon_sym___block] = ACTIONS(7080),
    [anon_sym___bridge] = ACTIONS(7080),
    [anon_sym___bridge_retained] = ACTIONS(7080),
    [anon_sym___bridge_transfer] = ACTIONS(7080),
    [anon_sym___complex] = ACTIONS(7080),
    [anon_sym___const] = ACTIONS(7080),
    [anon_sym___imag] = ACTIONS(7080),
    [anon_sym___kindof] = ACTIONS(7080),
    [anon_sym___nonnull] = ACTIONS(7080),
    [anon_sym___nullable] = ACTIONS(7080),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7080),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7080),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7080),
    [anon_sym___real] = ACTIONS(7080),
    [anon_sym___strong] = ACTIONS(7080),
    [anon_sym___unsafe_unretained] = ACTIONS(7080),
    [anon_sym___unused] = ACTIONS(7080),
    [anon_sym___weak] = ACTIONS(7080),
    [anon_sym_alignas] = ACTIONS(7080),
    [anon_sym__Alignas] = ACTIONS(7080),
    [sym_primitive_type] = ACTIONS(7080),
    [anon_sym_enum] = ACTIONS(7080),
    [anon_sym_COLON] = ACTIONS(7082),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7080),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7080),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7080),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7080),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7080),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7080),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7080),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7080),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7080),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7080),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7080),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7080),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7080),
    [anon_sym_API_AVAILABLE] = ACTIONS(7080),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7080),
    [anon_sym_API_DEPRECATED] = ACTIONS(7080),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7080),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7080),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7080),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7080),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7080),
    [anon_sym___deprecated_msg] = ACTIONS(7080),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7080),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7080),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7080),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7080),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7080),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7080),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7080),
  },
  [STATE(2913)] = {
    [sym__declaration_modifiers] = STATE(4888),
    [sym_attribute_specifier] = STATE(4888),
    [sym_attribute_declaration] = STATE(4888),
    [sym_ms_declspec_modifier] = STATE(4888),
    [sym_storage_class_specifier] = STATE(4888),
    [sym_type_qualifier] = STATE(4888),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_enumerator] = STATE(3038),
    [sym_preproc_ifdef_in_enumerator] = STATE(4927),
    [sym_availability_attribute_specifier] = STATE(4888),
    [aux_sym_enumerator_list_repeat1] = STATE(4927),
    [sym_identifier] = ACTIONS(7087),
    [anon_sym_COMMA] = ACTIONS(7089),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(7091),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(7091),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_RBRACE] = ACTIONS(7093),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2914)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2912),
    [sym_identifier] = ACTIONS(7095),
    [anon_sym_COMMA] = ACTIONS(7098),
    [anon_sym_RPAREN] = ACTIONS(7098),
    [anon_sym_LPAREN2] = ACTIONS(7098),
    [anon_sym_STAR] = ACTIONS(7098),
    [anon_sym_CARET] = ACTIONS(7098),
    [anon_sym_SEMI] = ACTIONS(7098),
    [anon_sym___extension__] = ACTIONS(7095),
    [anon_sym_extern] = ACTIONS(7095),
    [anon_sym___attribute__] = ACTIONS(7095),
    [anon_sym___attribute] = ACTIONS(7095),
    [anon_sym_noreturn] = ACTIONS(7095),
    [anon_sym_LBRACK] = ACTIONS(7098),
    [anon_sym___declspec] = ACTIONS(7095),
    [anon_sym___based] = ACTIONS(7095),
    [anon_sym_signed] = ACTIONS(7084),
    [anon_sym_unsigned] = ACTIONS(7084),
    [anon_sym_long] = ACTIONS(7084),
    [anon_sym_short] = ACTIONS(7084),
    [anon_sym_static] = ACTIONS(7095),
    [anon_sym_auto] = ACTIONS(7095),
    [anon_sym_register] = ACTIONS(7095),
    [anon_sym_inline] = ACTIONS(7095),
    [anon_sym___inline] = ACTIONS(7095),
    [anon_sym___inline__] = ACTIONS(7095),
    [anon_sym___forceinline] = ACTIONS(7095),
    [anon_sym_thread_local] = ACTIONS(7095),
    [anon_sym___thread] = ACTIONS(7095),
    [anon_sym_CG_EXTERN] = ACTIONS(7095),
    [anon_sym_CG_INLINE] = ACTIONS(7095),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7095),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7095),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7095),
    [anon_sym_IBOutlet] = ACTIONS(7095),
    [anon_sym_IBInspectable] = ACTIONS(7095),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7095),
    [anon_sym_NS_INLINE] = ACTIONS(7095),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7095),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7095),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7095),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7095),
    [anon_sym_const] = ACTIONS(7095),
    [anon_sym_constexpr] = ACTIONS(7095),
    [anon_sym_volatile] = ACTIONS(7095),
    [anon_sym_restrict] = ACTIONS(7095),
    [anon_sym___restrict__] = ACTIONS(7095),
    [anon_sym__Atomic] = ACTIONS(7095),
    [anon_sym__Noreturn] = ACTIONS(7095),
    [anon_sym__Nonnull] = ACTIONS(7095),
    [anon_sym_nullable] = ACTIONS(7095),
    [anon_sym__Complex] = ACTIONS(7095),
    [anon_sym__Nullable] = ACTIONS(7095),
    [anon_sym__Nullable_result] = ACTIONS(7095),
    [anon_sym__Null_unspecified] = ACTIONS(7095),
    [anon_sym___autoreleasing] = ACTIONS(7095),
    [anon_sym___block] = ACTIONS(7095),
    [anon_sym___bridge] = ACTIONS(7095),
    [anon_sym___bridge_retained] = ACTIONS(7095),
    [anon_sym___bridge_transfer] = ACTIONS(7095),
    [anon_sym___complex] = ACTIONS(7095),
    [anon_sym___const] = ACTIONS(7095),
    [anon_sym___imag] = ACTIONS(7095),
    [anon_sym___kindof] = ACTIONS(7095),
    [anon_sym___nonnull] = ACTIONS(7095),
    [anon_sym___nullable] = ACTIONS(7095),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7095),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7095),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7095),
    [anon_sym___real] = ACTIONS(7095),
    [anon_sym___strong] = ACTIONS(7095),
    [anon_sym___unsafe_unretained] = ACTIONS(7095),
    [anon_sym___unused] = ACTIONS(7095),
    [anon_sym___weak] = ACTIONS(7095),
    [anon_sym_alignas] = ACTIONS(7095),
    [anon_sym__Alignas] = ACTIONS(7095),
    [sym_primitive_type] = ACTIONS(7080),
    [anon_sym_enum] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7098),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7095),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7095),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7095),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7095),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7095),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7095),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7095),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7095),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7095),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7095),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7095),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7095),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7095),
    [anon_sym_API_AVAILABLE] = ACTIONS(7095),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7095),
    [anon_sym_API_DEPRECATED] = ACTIONS(7095),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7095),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7095),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7095),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7095),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7095),
    [anon_sym___deprecated_msg] = ACTIONS(7095),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7095),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7095),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7095),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7095),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7095),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7095),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7095),
  },
  [STATE(2915)] = {
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_RPAREN] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7103),
    [anon_sym_STAR] = ACTIONS(7103),
    [anon_sym_CARET] = ACTIONS(7103),
    [anon_sym_LT] = ACTIONS(7103),
    [anon_sym_SEMI] = ACTIONS(7103),
    [anon_sym___extension__] = ACTIONS(7101),
    [anon_sym_extern] = ACTIONS(7101),
    [anon_sym___attribute__] = ACTIONS(7101),
    [anon_sym___attribute] = ACTIONS(7101),
    [anon_sym_noreturn] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7103),
    [anon_sym___declspec] = ACTIONS(7101),
    [anon_sym___based] = ACTIONS(7101),
    [anon_sym_signed] = ACTIONS(7101),
    [anon_sym_unsigned] = ACTIONS(7101),
    [anon_sym_long] = ACTIONS(7101),
    [anon_sym_short] = ACTIONS(7101),
    [anon_sym_static] = ACTIONS(7101),
    [anon_sym_auto] = ACTIONS(7101),
    [anon_sym_register] = ACTIONS(7101),
    [anon_sym_inline] = ACTIONS(7101),
    [anon_sym___inline] = ACTIONS(7101),
    [anon_sym___inline__] = ACTIONS(7101),
    [anon_sym___forceinline] = ACTIONS(7101),
    [anon_sym_thread_local] = ACTIONS(7101),
    [anon_sym___thread] = ACTIONS(7101),
    [anon_sym_CG_EXTERN] = ACTIONS(7101),
    [anon_sym_CG_INLINE] = ACTIONS(7101),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7101),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7101),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7101),
    [anon_sym_IBOutlet] = ACTIONS(7101),
    [anon_sym_IBInspectable] = ACTIONS(7101),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7101),
    [anon_sym_NS_INLINE] = ACTIONS(7101),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7101),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7101),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7101),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7101),
    [anon_sym_const] = ACTIONS(7101),
    [anon_sym_constexpr] = ACTIONS(7101),
    [anon_sym_volatile] = ACTIONS(7101),
    [anon_sym_restrict] = ACTIONS(7101),
    [anon_sym___restrict__] = ACTIONS(7101),
    [anon_sym__Atomic] = ACTIONS(7101),
    [anon_sym__Noreturn] = ACTIONS(7101),
    [anon_sym__Nonnull] = ACTIONS(7101),
    [anon_sym_nullable] = ACTIONS(7101),
    [anon_sym__Complex] = ACTIONS(7101),
    [anon_sym__Nullable] = ACTIONS(7101),
    [anon_sym__Nullable_result] = ACTIONS(7101),
    [anon_sym__Null_unspecified] = ACTIONS(7101),
    [anon_sym___autoreleasing] = ACTIONS(7101),
    [anon_sym___block] = ACTIONS(7101),
    [anon_sym___bridge] = ACTIONS(7101),
    [anon_sym___bridge_retained] = ACTIONS(7101),
    [anon_sym___bridge_transfer] = ACTIONS(7101),
    [anon_sym___complex] = ACTIONS(7101),
    [anon_sym___const] = ACTIONS(7101),
    [anon_sym___imag] = ACTIONS(7101),
    [anon_sym___kindof] = ACTIONS(7101),
    [anon_sym___nonnull] = ACTIONS(7101),
    [anon_sym___nullable] = ACTIONS(7101),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7101),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7101),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7101),
    [anon_sym___real] = ACTIONS(7101),
    [anon_sym___strong] = ACTIONS(7101),
    [anon_sym___unsafe_unretained] = ACTIONS(7101),
    [anon_sym___unused] = ACTIONS(7101),
    [anon_sym___weak] = ACTIONS(7101),
    [anon_sym_alignas] = ACTIONS(7101),
    [anon_sym__Alignas] = ACTIONS(7101),
    [sym_primitive_type] = ACTIONS(7101),
    [anon_sym_enum] = ACTIONS(7101),
    [anon_sym_COLON] = ACTIONS(7103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7101),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7101),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7101),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7101),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7101),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7101),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7101),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7101),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7101),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7101),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7101),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7101),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7101),
    [anon_sym_API_AVAILABLE] = ACTIONS(7101),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7101),
    [anon_sym_API_DEPRECATED] = ACTIONS(7101),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7101),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7101),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7101),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7101),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7101),
    [anon_sym___deprecated_msg] = ACTIONS(7101),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7101),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7101),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7101),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7101),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7101),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7101),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7101),
  },
  [STATE(2916)] = {
    [sym__declaration_modifiers] = STATE(2509),
    [sym_attribute_specifier] = STATE(2509),
    [sym_attribute_declaration] = STATE(2509),
    [sym_ms_declspec_modifier] = STATE(2509),
    [sym_storage_class_specifier] = STATE(2509),
    [sym_type_qualifier] = STATE(2509),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2509),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2509),
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_COMMA] = ACTIONS(6254),
    [anon_sym_SEMI] = ACTIONS(6254),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6254),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(2917)] = {
    [sym_attribute_specifier] = STATE(2934),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_RPAREN] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7107),
    [anon_sym_STAR] = ACTIONS(7107),
    [anon_sym_CARET] = ACTIONS(7107),
    [anon_sym_SEMI] = ACTIONS(7107),
    [anon_sym___extension__] = ACTIONS(7105),
    [anon_sym_extern] = ACTIONS(7105),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7107),
    [anon_sym___declspec] = ACTIONS(7105),
    [anon_sym___based] = ACTIONS(7105),
    [anon_sym_signed] = ACTIONS(7105),
    [anon_sym_unsigned] = ACTIONS(7105),
    [anon_sym_long] = ACTIONS(7105),
    [anon_sym_short] = ACTIONS(7105),
    [anon_sym_static] = ACTIONS(7105),
    [anon_sym_auto] = ACTIONS(7105),
    [anon_sym_register] = ACTIONS(7105),
    [anon_sym_inline] = ACTIONS(7105),
    [anon_sym___inline] = ACTIONS(7105),
    [anon_sym___inline__] = ACTIONS(7105),
    [anon_sym___forceinline] = ACTIONS(7105),
    [anon_sym_thread_local] = ACTIONS(7105),
    [anon_sym___thread] = ACTIONS(7105),
    [anon_sym_CG_EXTERN] = ACTIONS(7105),
    [anon_sym_CG_INLINE] = ACTIONS(7105),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7105),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7105),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7105),
    [anon_sym_IBOutlet] = ACTIONS(7105),
    [anon_sym_IBInspectable] = ACTIONS(7105),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7105),
    [anon_sym_NS_INLINE] = ACTIONS(7105),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7105),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7105),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7105),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7105),
    [anon_sym_const] = ACTIONS(7105),
    [anon_sym_constexpr] = ACTIONS(7105),
    [anon_sym_volatile] = ACTIONS(7105),
    [anon_sym_restrict] = ACTIONS(7105),
    [anon_sym___restrict__] = ACTIONS(7105),
    [anon_sym__Atomic] = ACTIONS(7105),
    [anon_sym__Noreturn] = ACTIONS(7105),
    [anon_sym__Nonnull] = ACTIONS(7105),
    [anon_sym_nullable] = ACTIONS(7105),
    [anon_sym__Complex] = ACTIONS(7105),
    [anon_sym__Nullable] = ACTIONS(7105),
    [anon_sym__Nullable_result] = ACTIONS(7105),
    [anon_sym__Null_unspecified] = ACTIONS(7105),
    [anon_sym___autoreleasing] = ACTIONS(7105),
    [anon_sym___block] = ACTIONS(7105),
    [anon_sym___bridge] = ACTIONS(7105),
    [anon_sym___bridge_retained] = ACTIONS(7105),
    [anon_sym___bridge_transfer] = ACTIONS(7105),
    [anon_sym___complex] = ACTIONS(7105),
    [anon_sym___const] = ACTIONS(7105),
    [anon_sym___imag] = ACTIONS(7105),
    [anon_sym___kindof] = ACTIONS(7105),
    [anon_sym___nonnull] = ACTIONS(7105),
    [anon_sym___nullable] = ACTIONS(7105),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7105),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7105),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7105),
    [anon_sym___real] = ACTIONS(7105),
    [anon_sym___strong] = ACTIONS(7105),
    [anon_sym___unsafe_unretained] = ACTIONS(7105),
    [anon_sym___unused] = ACTIONS(7105),
    [anon_sym___weak] = ACTIONS(7105),
    [anon_sym_alignas] = ACTIONS(7105),
    [anon_sym__Alignas] = ACTIONS(7105),
    [sym_primitive_type] = ACTIONS(7105),
    [anon_sym_enum] = ACTIONS(7105),
    [anon_sym_COLON] = ACTIONS(7107),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7105),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7105),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7105),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7105),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7105),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7105),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7105),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7105),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7105),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7105),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7105),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7105),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7105),
    [anon_sym_API_AVAILABLE] = ACTIONS(7105),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7105),
    [anon_sym_API_DEPRECATED] = ACTIONS(7105),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7105),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7105),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7105),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7105),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7105),
    [anon_sym___deprecated_msg] = ACTIONS(7105),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7105),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7105),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7105),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7105),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7105),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7105),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7105),
  },
  [STATE(2918)] = {
    [sym_attribute_specifier] = STATE(2951),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_RPAREN] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7111),
    [anon_sym_STAR] = ACTIONS(7111),
    [anon_sym_CARET] = ACTIONS(7111),
    [anon_sym_SEMI] = ACTIONS(7111),
    [anon_sym___extension__] = ACTIONS(7109),
    [anon_sym_extern] = ACTIONS(7109),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7111),
    [anon_sym___declspec] = ACTIONS(7109),
    [anon_sym___based] = ACTIONS(7109),
    [anon_sym_signed] = ACTIONS(7109),
    [anon_sym_unsigned] = ACTIONS(7109),
    [anon_sym_long] = ACTIONS(7109),
    [anon_sym_short] = ACTIONS(7109),
    [anon_sym_static] = ACTIONS(7109),
    [anon_sym_auto] = ACTIONS(7109),
    [anon_sym_register] = ACTIONS(7109),
    [anon_sym_inline] = ACTIONS(7109),
    [anon_sym___inline] = ACTIONS(7109),
    [anon_sym___inline__] = ACTIONS(7109),
    [anon_sym___forceinline] = ACTIONS(7109),
    [anon_sym_thread_local] = ACTIONS(7109),
    [anon_sym___thread] = ACTIONS(7109),
    [anon_sym_CG_EXTERN] = ACTIONS(7109),
    [anon_sym_CG_INLINE] = ACTIONS(7109),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7109),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7109),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7109),
    [anon_sym_IBOutlet] = ACTIONS(7109),
    [anon_sym_IBInspectable] = ACTIONS(7109),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7109),
    [anon_sym_NS_INLINE] = ACTIONS(7109),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7109),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7109),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7109),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7109),
    [anon_sym_const] = ACTIONS(7109),
    [anon_sym_constexpr] = ACTIONS(7109),
    [anon_sym_volatile] = ACTIONS(7109),
    [anon_sym_restrict] = ACTIONS(7109),
    [anon_sym___restrict__] = ACTIONS(7109),
    [anon_sym__Atomic] = ACTIONS(7109),
    [anon_sym__Noreturn] = ACTIONS(7109),
    [anon_sym__Nonnull] = ACTIONS(7109),
    [anon_sym_nullable] = ACTIONS(7109),
    [anon_sym__Complex] = ACTIONS(7109),
    [anon_sym__Nullable] = ACTIONS(7109),
    [anon_sym__Nullable_result] = ACTIONS(7109),
    [anon_sym__Null_unspecified] = ACTIONS(7109),
    [anon_sym___autoreleasing] = ACTIONS(7109),
    [anon_sym___block] = ACTIONS(7109),
    [anon_sym___bridge] = ACTIONS(7109),
    [anon_sym___bridge_retained] = ACTIONS(7109),
    [anon_sym___bridge_transfer] = ACTIONS(7109),
    [anon_sym___complex] = ACTIONS(7109),
    [anon_sym___const] = ACTIONS(7109),
    [anon_sym___imag] = ACTIONS(7109),
    [anon_sym___kindof] = ACTIONS(7109),
    [anon_sym___nonnull] = ACTIONS(7109),
    [anon_sym___nullable] = ACTIONS(7109),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7109),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7109),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7109),
    [anon_sym___real] = ACTIONS(7109),
    [anon_sym___strong] = ACTIONS(7109),
    [anon_sym___unsafe_unretained] = ACTIONS(7109),
    [anon_sym___unused] = ACTIONS(7109),
    [anon_sym___weak] = ACTIONS(7109),
    [anon_sym_alignas] = ACTIONS(7109),
    [anon_sym__Alignas] = ACTIONS(7109),
    [sym_primitive_type] = ACTIONS(7109),
    [anon_sym_enum] = ACTIONS(7109),
    [anon_sym_COLON] = ACTIONS(7111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7109),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7109),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7109),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7109),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7109),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7109),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7109),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7109),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7109),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7109),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7109),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7109),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7109),
    [anon_sym_API_AVAILABLE] = ACTIONS(7109),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7109),
    [anon_sym_API_DEPRECATED] = ACTIONS(7109),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7109),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7109),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7109),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7109),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7109),
    [anon_sym___deprecated_msg] = ACTIONS(7109),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7109),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7109),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7109),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7109),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7109),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7109),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7109),
  },
  [STATE(2919)] = {
    [sym_attribute_specifier] = STATE(2986),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_RPAREN] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7115),
    [anon_sym_STAR] = ACTIONS(7115),
    [anon_sym_CARET] = ACTIONS(7115),
    [anon_sym_SEMI] = ACTIONS(7115),
    [anon_sym___extension__] = ACTIONS(7113),
    [anon_sym_extern] = ACTIONS(7113),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7115),
    [anon_sym___declspec] = ACTIONS(7113),
    [anon_sym___based] = ACTIONS(7113),
    [anon_sym_signed] = ACTIONS(7113),
    [anon_sym_unsigned] = ACTIONS(7113),
    [anon_sym_long] = ACTIONS(7113),
    [anon_sym_short] = ACTIONS(7113),
    [anon_sym_static] = ACTIONS(7113),
    [anon_sym_auto] = ACTIONS(7113),
    [anon_sym_register] = ACTIONS(7113),
    [anon_sym_inline] = ACTIONS(7113),
    [anon_sym___inline] = ACTIONS(7113),
    [anon_sym___inline__] = ACTIONS(7113),
    [anon_sym___forceinline] = ACTIONS(7113),
    [anon_sym_thread_local] = ACTIONS(7113),
    [anon_sym___thread] = ACTIONS(7113),
    [anon_sym_CG_EXTERN] = ACTIONS(7113),
    [anon_sym_CG_INLINE] = ACTIONS(7113),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7113),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7113),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7113),
    [anon_sym_IBOutlet] = ACTIONS(7113),
    [anon_sym_IBInspectable] = ACTIONS(7113),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7113),
    [anon_sym_NS_INLINE] = ACTIONS(7113),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7113),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7113),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7113),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7113),
    [anon_sym_const] = ACTIONS(7113),
    [anon_sym_constexpr] = ACTIONS(7113),
    [anon_sym_volatile] = ACTIONS(7113),
    [anon_sym_restrict] = ACTIONS(7113),
    [anon_sym___restrict__] = ACTIONS(7113),
    [anon_sym__Atomic] = ACTIONS(7113),
    [anon_sym__Noreturn] = ACTIONS(7113),
    [anon_sym__Nonnull] = ACTIONS(7113),
    [anon_sym_nullable] = ACTIONS(7113),
    [anon_sym__Complex] = ACTIONS(7113),
    [anon_sym__Nullable] = ACTIONS(7113),
    [anon_sym__Nullable_result] = ACTIONS(7113),
    [anon_sym__Null_unspecified] = ACTIONS(7113),
    [anon_sym___autoreleasing] = ACTIONS(7113),
    [anon_sym___block] = ACTIONS(7113),
    [anon_sym___bridge] = ACTIONS(7113),
    [anon_sym___bridge_retained] = ACTIONS(7113),
    [anon_sym___bridge_transfer] = ACTIONS(7113),
    [anon_sym___complex] = ACTIONS(7113),
    [anon_sym___const] = ACTIONS(7113),
    [anon_sym___imag] = ACTIONS(7113),
    [anon_sym___kindof] = ACTIONS(7113),
    [anon_sym___nonnull] = ACTIONS(7113),
    [anon_sym___nullable] = ACTIONS(7113),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7113),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7113),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7113),
    [anon_sym___real] = ACTIONS(7113),
    [anon_sym___strong] = ACTIONS(7113),
    [anon_sym___unsafe_unretained] = ACTIONS(7113),
    [anon_sym___unused] = ACTIONS(7113),
    [anon_sym___weak] = ACTIONS(7113),
    [anon_sym_alignas] = ACTIONS(7113),
    [anon_sym__Alignas] = ACTIONS(7113),
    [sym_primitive_type] = ACTIONS(7113),
    [anon_sym_enum] = ACTIONS(7113),
    [anon_sym_COLON] = ACTIONS(7115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7113),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7113),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7113),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7113),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7113),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7113),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7113),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7113),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7113),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7113),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7113),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7113),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7113),
    [anon_sym_API_AVAILABLE] = ACTIONS(7113),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7113),
    [anon_sym_API_DEPRECATED] = ACTIONS(7113),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7113),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7113),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7113),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7113),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7113),
    [anon_sym___deprecated_msg] = ACTIONS(7113),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7113),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7113),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7113),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7113),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7113),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7113),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7113),
  },
  [STATE(2920)] = {
    [sym__declaration_modifiers] = STATE(2503),
    [sym_attribute_specifier] = STATE(2503),
    [sym_attribute_declaration] = STATE(2503),
    [sym_ms_declspec_modifier] = STATE(2503),
    [sym_storage_class_specifier] = STATE(2503),
    [sym_type_qualifier] = STATE(2503),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2503),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2503),
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_COMMA] = ACTIONS(6258),
    [anon_sym_SEMI] = ACTIONS(6258),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [anon_sym_COLON] = ACTIONS(6258),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(2921)] = {
    [sym_attribute_specifier] = STATE(2988),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_RPAREN] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7119),
    [anon_sym_STAR] = ACTIONS(7119),
    [anon_sym_CARET] = ACTIONS(7119),
    [anon_sym_SEMI] = ACTIONS(7119),
    [anon_sym___extension__] = ACTIONS(7117),
    [anon_sym_extern] = ACTIONS(7117),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7119),
    [anon_sym___declspec] = ACTIONS(7117),
    [anon_sym___based] = ACTIONS(7117),
    [anon_sym_signed] = ACTIONS(7117),
    [anon_sym_unsigned] = ACTIONS(7117),
    [anon_sym_long] = ACTIONS(7117),
    [anon_sym_short] = ACTIONS(7117),
    [anon_sym_static] = ACTIONS(7117),
    [anon_sym_auto] = ACTIONS(7117),
    [anon_sym_register] = ACTIONS(7117),
    [anon_sym_inline] = ACTIONS(7117),
    [anon_sym___inline] = ACTIONS(7117),
    [anon_sym___inline__] = ACTIONS(7117),
    [anon_sym___forceinline] = ACTIONS(7117),
    [anon_sym_thread_local] = ACTIONS(7117),
    [anon_sym___thread] = ACTIONS(7117),
    [anon_sym_CG_EXTERN] = ACTIONS(7117),
    [anon_sym_CG_INLINE] = ACTIONS(7117),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7117),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7117),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7117),
    [anon_sym_IBOutlet] = ACTIONS(7117),
    [anon_sym_IBInspectable] = ACTIONS(7117),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7117),
    [anon_sym_NS_INLINE] = ACTIONS(7117),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7117),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7117),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7117),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7117),
    [anon_sym_const] = ACTIONS(7117),
    [anon_sym_constexpr] = ACTIONS(7117),
    [anon_sym_volatile] = ACTIONS(7117),
    [anon_sym_restrict] = ACTIONS(7117),
    [anon_sym___restrict__] = ACTIONS(7117),
    [anon_sym__Atomic] = ACTIONS(7117),
    [anon_sym__Noreturn] = ACTIONS(7117),
    [anon_sym__Nonnull] = ACTIONS(7117),
    [anon_sym_nullable] = ACTIONS(7117),
    [anon_sym__Complex] = ACTIONS(7117),
    [anon_sym__Nullable] = ACTIONS(7117),
    [anon_sym__Nullable_result] = ACTIONS(7117),
    [anon_sym__Null_unspecified] = ACTIONS(7117),
    [anon_sym___autoreleasing] = ACTIONS(7117),
    [anon_sym___block] = ACTIONS(7117),
    [anon_sym___bridge] = ACTIONS(7117),
    [anon_sym___bridge_retained] = ACTIONS(7117),
    [anon_sym___bridge_transfer] = ACTIONS(7117),
    [anon_sym___complex] = ACTIONS(7117),
    [anon_sym___const] = ACTIONS(7117),
    [anon_sym___imag] = ACTIONS(7117),
    [anon_sym___kindof] = ACTIONS(7117),
    [anon_sym___nonnull] = ACTIONS(7117),
    [anon_sym___nullable] = ACTIONS(7117),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7117),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7117),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7117),
    [anon_sym___real] = ACTIONS(7117),
    [anon_sym___strong] = ACTIONS(7117),
    [anon_sym___unsafe_unretained] = ACTIONS(7117),
    [anon_sym___unused] = ACTIONS(7117),
    [anon_sym___weak] = ACTIONS(7117),
    [anon_sym_alignas] = ACTIONS(7117),
    [anon_sym__Alignas] = ACTIONS(7117),
    [sym_primitive_type] = ACTIONS(7117),
    [anon_sym_enum] = ACTIONS(7117),
    [anon_sym_COLON] = ACTIONS(7119),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7117),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7117),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7117),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7117),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7117),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7117),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7117),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7117),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7117),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7117),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7117),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7117),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7117),
    [anon_sym_API_AVAILABLE] = ACTIONS(7117),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7117),
    [anon_sym_API_DEPRECATED] = ACTIONS(7117),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7117),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7117),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7117),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7117),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7117),
    [anon_sym___deprecated_msg] = ACTIONS(7117),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7117),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7117),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7117),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7117),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7117),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7117),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7117),
  },
  [STATE(2922)] = {
    [sym_attribute_specifier] = STATE(2942),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_RPAREN] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7123),
    [anon_sym_STAR] = ACTIONS(7123),
    [anon_sym_CARET] = ACTIONS(7123),
    [anon_sym_SEMI] = ACTIONS(7123),
    [anon_sym___extension__] = ACTIONS(7121),
    [anon_sym_extern] = ACTIONS(7121),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7123),
    [anon_sym___declspec] = ACTIONS(7121),
    [anon_sym___based] = ACTIONS(7121),
    [anon_sym_signed] = ACTIONS(7121),
    [anon_sym_unsigned] = ACTIONS(7121),
    [anon_sym_long] = ACTIONS(7121),
    [anon_sym_short] = ACTIONS(7121),
    [anon_sym_static] = ACTIONS(7121),
    [anon_sym_auto] = ACTIONS(7121),
    [anon_sym_register] = ACTIONS(7121),
    [anon_sym_inline] = ACTIONS(7121),
    [anon_sym___inline] = ACTIONS(7121),
    [anon_sym___inline__] = ACTIONS(7121),
    [anon_sym___forceinline] = ACTIONS(7121),
    [anon_sym_thread_local] = ACTIONS(7121),
    [anon_sym___thread] = ACTIONS(7121),
    [anon_sym_CG_EXTERN] = ACTIONS(7121),
    [anon_sym_CG_INLINE] = ACTIONS(7121),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7121),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7121),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7121),
    [anon_sym_IBOutlet] = ACTIONS(7121),
    [anon_sym_IBInspectable] = ACTIONS(7121),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7121),
    [anon_sym_NS_INLINE] = ACTIONS(7121),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7121),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7121),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7121),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7121),
    [anon_sym_const] = ACTIONS(7121),
    [anon_sym_constexpr] = ACTIONS(7121),
    [anon_sym_volatile] = ACTIONS(7121),
    [anon_sym_restrict] = ACTIONS(7121),
    [anon_sym___restrict__] = ACTIONS(7121),
    [anon_sym__Atomic] = ACTIONS(7121),
    [anon_sym__Noreturn] = ACTIONS(7121),
    [anon_sym__Nonnull] = ACTIONS(7121),
    [anon_sym_nullable] = ACTIONS(7121),
    [anon_sym__Complex] = ACTIONS(7121),
    [anon_sym__Nullable] = ACTIONS(7121),
    [anon_sym__Nullable_result] = ACTIONS(7121),
    [anon_sym__Null_unspecified] = ACTIONS(7121),
    [anon_sym___autoreleasing] = ACTIONS(7121),
    [anon_sym___block] = ACTIONS(7121),
    [anon_sym___bridge] = ACTIONS(7121),
    [anon_sym___bridge_retained] = ACTIONS(7121),
    [anon_sym___bridge_transfer] = ACTIONS(7121),
    [anon_sym___complex] = ACTIONS(7121),
    [anon_sym___const] = ACTIONS(7121),
    [anon_sym___imag] = ACTIONS(7121),
    [anon_sym___kindof] = ACTIONS(7121),
    [anon_sym___nonnull] = ACTIONS(7121),
    [anon_sym___nullable] = ACTIONS(7121),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7121),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7121),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7121),
    [anon_sym___real] = ACTIONS(7121),
    [anon_sym___strong] = ACTIONS(7121),
    [anon_sym___unsafe_unretained] = ACTIONS(7121),
    [anon_sym___unused] = ACTIONS(7121),
    [anon_sym___weak] = ACTIONS(7121),
    [anon_sym_alignas] = ACTIONS(7121),
    [anon_sym__Alignas] = ACTIONS(7121),
    [sym_primitive_type] = ACTIONS(7121),
    [anon_sym_enum] = ACTIONS(7121),
    [anon_sym_COLON] = ACTIONS(7123),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7121),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7121),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7121),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7121),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7121),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7121),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7121),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7121),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7121),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7121),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7121),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7121),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7121),
    [anon_sym_API_AVAILABLE] = ACTIONS(7121),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7121),
    [anon_sym_API_DEPRECATED] = ACTIONS(7121),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7121),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7121),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7121),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7121),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7121),
    [anon_sym___deprecated_msg] = ACTIONS(7121),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7121),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7121),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7121),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7121),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7121),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7121),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7121),
  },
  [STATE(2923)] = {
    [sym__declaration_modifiers] = STATE(4906),
    [sym_attribute_specifier] = STATE(4906),
    [sym_attribute_declaration] = STATE(4906),
    [sym_ms_declspec_modifier] = STATE(4906),
    [sym_storage_class_specifier] = STATE(4906),
    [sym_type_qualifier] = STATE(4906),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_enumerator] = STATE(3038),
    [sym_preproc_ifdef_in_enumerator] = STATE(4914),
    [sym_availability_attribute_specifier] = STATE(4906),
    [aux_sym_enumerator_list_repeat1] = STATE(4914),
    [sym_identifier] = ACTIONS(7087),
    [anon_sym_COMMA] = ACTIONS(7125),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(7091),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(7091),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_RBRACE] = ACTIONS(7127),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2924)] = {
    [sym_attribute_specifier] = STATE(2972),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_RPAREN] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7131),
    [anon_sym_STAR] = ACTIONS(7131),
    [anon_sym_CARET] = ACTIONS(7131),
    [anon_sym_SEMI] = ACTIONS(7131),
    [anon_sym___extension__] = ACTIONS(7129),
    [anon_sym_extern] = ACTIONS(7129),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7131),
    [anon_sym___declspec] = ACTIONS(7129),
    [anon_sym___based] = ACTIONS(7129),
    [anon_sym_signed] = ACTIONS(7129),
    [anon_sym_unsigned] = ACTIONS(7129),
    [anon_sym_long] = ACTIONS(7129),
    [anon_sym_short] = ACTIONS(7129),
    [anon_sym_static] = ACTIONS(7129),
    [anon_sym_auto] = ACTIONS(7129),
    [anon_sym_register] = ACTIONS(7129),
    [anon_sym_inline] = ACTIONS(7129),
    [anon_sym___inline] = ACTIONS(7129),
    [anon_sym___inline__] = ACTIONS(7129),
    [anon_sym___forceinline] = ACTIONS(7129),
    [anon_sym_thread_local] = ACTIONS(7129),
    [anon_sym___thread] = ACTIONS(7129),
    [anon_sym_CG_EXTERN] = ACTIONS(7129),
    [anon_sym_CG_INLINE] = ACTIONS(7129),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7129),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7129),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7129),
    [anon_sym_IBOutlet] = ACTIONS(7129),
    [anon_sym_IBInspectable] = ACTIONS(7129),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7129),
    [anon_sym_NS_INLINE] = ACTIONS(7129),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7129),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7129),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7129),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7129),
    [anon_sym_const] = ACTIONS(7129),
    [anon_sym_constexpr] = ACTIONS(7129),
    [anon_sym_volatile] = ACTIONS(7129),
    [anon_sym_restrict] = ACTIONS(7129),
    [anon_sym___restrict__] = ACTIONS(7129),
    [anon_sym__Atomic] = ACTIONS(7129),
    [anon_sym__Noreturn] = ACTIONS(7129),
    [anon_sym__Nonnull] = ACTIONS(7129),
    [anon_sym_nullable] = ACTIONS(7129),
    [anon_sym__Complex] = ACTIONS(7129),
    [anon_sym__Nullable] = ACTIONS(7129),
    [anon_sym__Nullable_result] = ACTIONS(7129),
    [anon_sym__Null_unspecified] = ACTIONS(7129),
    [anon_sym___autoreleasing] = ACTIONS(7129),
    [anon_sym___block] = ACTIONS(7129),
    [anon_sym___bridge] = ACTIONS(7129),
    [anon_sym___bridge_retained] = ACTIONS(7129),
    [anon_sym___bridge_transfer] = ACTIONS(7129),
    [anon_sym___complex] = ACTIONS(7129),
    [anon_sym___const] = ACTIONS(7129),
    [anon_sym___imag] = ACTIONS(7129),
    [anon_sym___kindof] = ACTIONS(7129),
    [anon_sym___nonnull] = ACTIONS(7129),
    [anon_sym___nullable] = ACTIONS(7129),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7129),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7129),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7129),
    [anon_sym___real] = ACTIONS(7129),
    [anon_sym___strong] = ACTIONS(7129),
    [anon_sym___unsafe_unretained] = ACTIONS(7129),
    [anon_sym___unused] = ACTIONS(7129),
    [anon_sym___weak] = ACTIONS(7129),
    [anon_sym_alignas] = ACTIONS(7129),
    [anon_sym__Alignas] = ACTIONS(7129),
    [sym_primitive_type] = ACTIONS(7129),
    [anon_sym_enum] = ACTIONS(7129),
    [anon_sym_COLON] = ACTIONS(7131),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7129),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7129),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7129),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7129),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7129),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7129),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7129),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7129),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7129),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7129),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7129),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7129),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7129),
    [anon_sym_API_AVAILABLE] = ACTIONS(7129),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7129),
    [anon_sym_API_DEPRECATED] = ACTIONS(7129),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7129),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7129),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7129),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7129),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7129),
    [anon_sym___deprecated_msg] = ACTIONS(7129),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7129),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7129),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7129),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7129),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7129),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7129),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7129),
  },
  [STATE(2925)] = {
    [sym_attribute_specifier] = STATE(2991),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_RPAREN] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7135),
    [anon_sym_STAR] = ACTIONS(7135),
    [anon_sym_CARET] = ACTIONS(7135),
    [anon_sym_SEMI] = ACTIONS(7135),
    [anon_sym___extension__] = ACTIONS(7133),
    [anon_sym_extern] = ACTIONS(7133),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7135),
    [anon_sym___declspec] = ACTIONS(7133),
    [anon_sym___based] = ACTIONS(7133),
    [anon_sym_signed] = ACTIONS(7133),
    [anon_sym_unsigned] = ACTIONS(7133),
    [anon_sym_long] = ACTIONS(7133),
    [anon_sym_short] = ACTIONS(7133),
    [anon_sym_static] = ACTIONS(7133),
    [anon_sym_auto] = ACTIONS(7133),
    [anon_sym_register] = ACTIONS(7133),
    [anon_sym_inline] = ACTIONS(7133),
    [anon_sym___inline] = ACTIONS(7133),
    [anon_sym___inline__] = ACTIONS(7133),
    [anon_sym___forceinline] = ACTIONS(7133),
    [anon_sym_thread_local] = ACTIONS(7133),
    [anon_sym___thread] = ACTIONS(7133),
    [anon_sym_CG_EXTERN] = ACTIONS(7133),
    [anon_sym_CG_INLINE] = ACTIONS(7133),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7133),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7133),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7133),
    [anon_sym_IBOutlet] = ACTIONS(7133),
    [anon_sym_IBInspectable] = ACTIONS(7133),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7133),
    [anon_sym_NS_INLINE] = ACTIONS(7133),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7133),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7133),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7133),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7133),
    [anon_sym_const] = ACTIONS(7133),
    [anon_sym_constexpr] = ACTIONS(7133),
    [anon_sym_volatile] = ACTIONS(7133),
    [anon_sym_restrict] = ACTIONS(7133),
    [anon_sym___restrict__] = ACTIONS(7133),
    [anon_sym__Atomic] = ACTIONS(7133),
    [anon_sym__Noreturn] = ACTIONS(7133),
    [anon_sym__Nonnull] = ACTIONS(7133),
    [anon_sym_nullable] = ACTIONS(7133),
    [anon_sym__Complex] = ACTIONS(7133),
    [anon_sym__Nullable] = ACTIONS(7133),
    [anon_sym__Nullable_result] = ACTIONS(7133),
    [anon_sym__Null_unspecified] = ACTIONS(7133),
    [anon_sym___autoreleasing] = ACTIONS(7133),
    [anon_sym___block] = ACTIONS(7133),
    [anon_sym___bridge] = ACTIONS(7133),
    [anon_sym___bridge_retained] = ACTIONS(7133),
    [anon_sym___bridge_transfer] = ACTIONS(7133),
    [anon_sym___complex] = ACTIONS(7133),
    [anon_sym___const] = ACTIONS(7133),
    [anon_sym___imag] = ACTIONS(7133),
    [anon_sym___kindof] = ACTIONS(7133),
    [anon_sym___nonnull] = ACTIONS(7133),
    [anon_sym___nullable] = ACTIONS(7133),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7133),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7133),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7133),
    [anon_sym___real] = ACTIONS(7133),
    [anon_sym___strong] = ACTIONS(7133),
    [anon_sym___unsafe_unretained] = ACTIONS(7133),
    [anon_sym___unused] = ACTIONS(7133),
    [anon_sym___weak] = ACTIONS(7133),
    [anon_sym_alignas] = ACTIONS(7133),
    [anon_sym__Alignas] = ACTIONS(7133),
    [sym_primitive_type] = ACTIONS(7133),
    [anon_sym_enum] = ACTIONS(7133),
    [anon_sym_COLON] = ACTIONS(7135),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7133),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7133),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7133),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7133),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7133),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7133),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7133),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7133),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7133),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7133),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7133),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7133),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7133),
    [anon_sym_API_AVAILABLE] = ACTIONS(7133),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7133),
    [anon_sym_API_DEPRECATED] = ACTIONS(7133),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7133),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7133),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7133),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7133),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7133),
    [anon_sym___deprecated_msg] = ACTIONS(7133),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7133),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7133),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7133),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7133),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7133),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7133),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7133),
  },
  [STATE(2926)] = {
    [sym_attribute_specifier] = STATE(2946),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_RPAREN] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7139),
    [anon_sym_STAR] = ACTIONS(7139),
    [anon_sym_CARET] = ACTIONS(7139),
    [anon_sym_SEMI] = ACTIONS(7139),
    [anon_sym___extension__] = ACTIONS(7137),
    [anon_sym_extern] = ACTIONS(7137),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7139),
    [anon_sym___declspec] = ACTIONS(7137),
    [anon_sym___based] = ACTIONS(7137),
    [anon_sym_signed] = ACTIONS(7137),
    [anon_sym_unsigned] = ACTIONS(7137),
    [anon_sym_long] = ACTIONS(7137),
    [anon_sym_short] = ACTIONS(7137),
    [anon_sym_static] = ACTIONS(7137),
    [anon_sym_auto] = ACTIONS(7137),
    [anon_sym_register] = ACTIONS(7137),
    [anon_sym_inline] = ACTIONS(7137),
    [anon_sym___inline] = ACTIONS(7137),
    [anon_sym___inline__] = ACTIONS(7137),
    [anon_sym___forceinline] = ACTIONS(7137),
    [anon_sym_thread_local] = ACTIONS(7137),
    [anon_sym___thread] = ACTIONS(7137),
    [anon_sym_CG_EXTERN] = ACTIONS(7137),
    [anon_sym_CG_INLINE] = ACTIONS(7137),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7137),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7137),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7137),
    [anon_sym_IBOutlet] = ACTIONS(7137),
    [anon_sym_IBInspectable] = ACTIONS(7137),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7137),
    [anon_sym_NS_INLINE] = ACTIONS(7137),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7137),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7137),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7137),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7137),
    [anon_sym_const] = ACTIONS(7137),
    [anon_sym_constexpr] = ACTIONS(7137),
    [anon_sym_volatile] = ACTIONS(7137),
    [anon_sym_restrict] = ACTIONS(7137),
    [anon_sym___restrict__] = ACTIONS(7137),
    [anon_sym__Atomic] = ACTIONS(7137),
    [anon_sym__Noreturn] = ACTIONS(7137),
    [anon_sym__Nonnull] = ACTIONS(7137),
    [anon_sym_nullable] = ACTIONS(7137),
    [anon_sym__Complex] = ACTIONS(7137),
    [anon_sym__Nullable] = ACTIONS(7137),
    [anon_sym__Nullable_result] = ACTIONS(7137),
    [anon_sym__Null_unspecified] = ACTIONS(7137),
    [anon_sym___autoreleasing] = ACTIONS(7137),
    [anon_sym___block] = ACTIONS(7137),
    [anon_sym___bridge] = ACTIONS(7137),
    [anon_sym___bridge_retained] = ACTIONS(7137),
    [anon_sym___bridge_transfer] = ACTIONS(7137),
    [anon_sym___complex] = ACTIONS(7137),
    [anon_sym___const] = ACTIONS(7137),
    [anon_sym___imag] = ACTIONS(7137),
    [anon_sym___kindof] = ACTIONS(7137),
    [anon_sym___nonnull] = ACTIONS(7137),
    [anon_sym___nullable] = ACTIONS(7137),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7137),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7137),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7137),
    [anon_sym___real] = ACTIONS(7137),
    [anon_sym___strong] = ACTIONS(7137),
    [anon_sym___unsafe_unretained] = ACTIONS(7137),
    [anon_sym___unused] = ACTIONS(7137),
    [anon_sym___weak] = ACTIONS(7137),
    [anon_sym_alignas] = ACTIONS(7137),
    [anon_sym__Alignas] = ACTIONS(7137),
    [sym_primitive_type] = ACTIONS(7137),
    [anon_sym_enum] = ACTIONS(7137),
    [anon_sym_COLON] = ACTIONS(7139),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7137),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7137),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7137),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7137),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7137),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7137),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7137),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7137),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7137),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7137),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7137),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7137),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7137),
    [anon_sym_API_AVAILABLE] = ACTIONS(7137),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7137),
    [anon_sym_API_DEPRECATED] = ACTIONS(7137),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7137),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7137),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7137),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7137),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7137),
    [anon_sym___deprecated_msg] = ACTIONS(7137),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7137),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7137),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7137),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7137),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7137),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7137),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7137),
  },
  [STATE(2927)] = {
    [sym_attribute_specifier] = STATE(2989),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_RPAREN] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7143),
    [anon_sym_STAR] = ACTIONS(7143),
    [anon_sym_CARET] = ACTIONS(7143),
    [anon_sym_SEMI] = ACTIONS(7143),
    [anon_sym___extension__] = ACTIONS(7141),
    [anon_sym_extern] = ACTIONS(7141),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7143),
    [anon_sym___declspec] = ACTIONS(7141),
    [anon_sym___based] = ACTIONS(7141),
    [anon_sym_signed] = ACTIONS(7141),
    [anon_sym_unsigned] = ACTIONS(7141),
    [anon_sym_long] = ACTIONS(7141),
    [anon_sym_short] = ACTIONS(7141),
    [anon_sym_static] = ACTIONS(7141),
    [anon_sym_auto] = ACTIONS(7141),
    [anon_sym_register] = ACTIONS(7141),
    [anon_sym_inline] = ACTIONS(7141),
    [anon_sym___inline] = ACTIONS(7141),
    [anon_sym___inline__] = ACTIONS(7141),
    [anon_sym___forceinline] = ACTIONS(7141),
    [anon_sym_thread_local] = ACTIONS(7141),
    [anon_sym___thread] = ACTIONS(7141),
    [anon_sym_CG_EXTERN] = ACTIONS(7141),
    [anon_sym_CG_INLINE] = ACTIONS(7141),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7141),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7141),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7141),
    [anon_sym_IBOutlet] = ACTIONS(7141),
    [anon_sym_IBInspectable] = ACTIONS(7141),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7141),
    [anon_sym_NS_INLINE] = ACTIONS(7141),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7141),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7141),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7141),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7141),
    [anon_sym_const] = ACTIONS(7141),
    [anon_sym_constexpr] = ACTIONS(7141),
    [anon_sym_volatile] = ACTIONS(7141),
    [anon_sym_restrict] = ACTIONS(7141),
    [anon_sym___restrict__] = ACTIONS(7141),
    [anon_sym__Atomic] = ACTIONS(7141),
    [anon_sym__Noreturn] = ACTIONS(7141),
    [anon_sym__Nonnull] = ACTIONS(7141),
    [anon_sym_nullable] = ACTIONS(7141),
    [anon_sym__Complex] = ACTIONS(7141),
    [anon_sym__Nullable] = ACTIONS(7141),
    [anon_sym__Nullable_result] = ACTIONS(7141),
    [anon_sym__Null_unspecified] = ACTIONS(7141),
    [anon_sym___autoreleasing] = ACTIONS(7141),
    [anon_sym___block] = ACTIONS(7141),
    [anon_sym___bridge] = ACTIONS(7141),
    [anon_sym___bridge_retained] = ACTIONS(7141),
    [anon_sym___bridge_transfer] = ACTIONS(7141),
    [anon_sym___complex] = ACTIONS(7141),
    [anon_sym___const] = ACTIONS(7141),
    [anon_sym___imag] = ACTIONS(7141),
    [anon_sym___kindof] = ACTIONS(7141),
    [anon_sym___nonnull] = ACTIONS(7141),
    [anon_sym___nullable] = ACTIONS(7141),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7141),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7141),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7141),
    [anon_sym___real] = ACTIONS(7141),
    [anon_sym___strong] = ACTIONS(7141),
    [anon_sym___unsafe_unretained] = ACTIONS(7141),
    [anon_sym___unused] = ACTIONS(7141),
    [anon_sym___weak] = ACTIONS(7141),
    [anon_sym_alignas] = ACTIONS(7141),
    [anon_sym__Alignas] = ACTIONS(7141),
    [sym_primitive_type] = ACTIONS(7141),
    [anon_sym_enum] = ACTIONS(7141),
    [anon_sym_COLON] = ACTIONS(7143),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7141),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7141),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7141),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7141),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7141),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7141),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7141),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7141),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7141),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7141),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7141),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7141),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7141),
    [anon_sym_API_AVAILABLE] = ACTIONS(7141),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7141),
    [anon_sym_API_DEPRECATED] = ACTIONS(7141),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7141),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7141),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7141),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7141),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7141),
    [anon_sym___deprecated_msg] = ACTIONS(7141),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7141),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7141),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7141),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7141),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7141),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7141),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7141),
  },
  [STATE(2928)] = {
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_RPAREN] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7147),
    [anon_sym_STAR] = ACTIONS(7147),
    [anon_sym_CARET] = ACTIONS(7147),
    [anon_sym_LT] = ACTIONS(7147),
    [anon_sym_SEMI] = ACTIONS(7147),
    [anon_sym___extension__] = ACTIONS(7145),
    [anon_sym_extern] = ACTIONS(7145),
    [anon_sym___attribute__] = ACTIONS(7145),
    [anon_sym___attribute] = ACTIONS(7145),
    [anon_sym_noreturn] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7147),
    [anon_sym___declspec] = ACTIONS(7145),
    [anon_sym___based] = ACTIONS(7145),
    [anon_sym_signed] = ACTIONS(7145),
    [anon_sym_unsigned] = ACTIONS(7145),
    [anon_sym_long] = ACTIONS(7145),
    [anon_sym_short] = ACTIONS(7145),
    [anon_sym_static] = ACTIONS(7145),
    [anon_sym_auto] = ACTIONS(7145),
    [anon_sym_register] = ACTIONS(7145),
    [anon_sym_inline] = ACTIONS(7145),
    [anon_sym___inline] = ACTIONS(7145),
    [anon_sym___inline__] = ACTIONS(7145),
    [anon_sym___forceinline] = ACTIONS(7145),
    [anon_sym_thread_local] = ACTIONS(7145),
    [anon_sym___thread] = ACTIONS(7145),
    [anon_sym_CG_EXTERN] = ACTIONS(7145),
    [anon_sym_CG_INLINE] = ACTIONS(7145),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7145),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7145),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7145),
    [anon_sym_IBOutlet] = ACTIONS(7145),
    [anon_sym_IBInspectable] = ACTIONS(7145),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7145),
    [anon_sym_NS_INLINE] = ACTIONS(7145),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7145),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7145),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7145),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7145),
    [anon_sym_const] = ACTIONS(7145),
    [anon_sym_constexpr] = ACTIONS(7145),
    [anon_sym_volatile] = ACTIONS(7145),
    [anon_sym_restrict] = ACTIONS(7145),
    [anon_sym___restrict__] = ACTIONS(7145),
    [anon_sym__Atomic] = ACTIONS(7145),
    [anon_sym__Noreturn] = ACTIONS(7145),
    [anon_sym__Nonnull] = ACTIONS(7145),
    [anon_sym_nullable] = ACTIONS(7145),
    [anon_sym__Complex] = ACTIONS(7145),
    [anon_sym__Nullable] = ACTIONS(7145),
    [anon_sym__Nullable_result] = ACTIONS(7145),
    [anon_sym__Null_unspecified] = ACTIONS(7145),
    [anon_sym___autoreleasing] = ACTIONS(7145),
    [anon_sym___block] = ACTIONS(7145),
    [anon_sym___bridge] = ACTIONS(7145),
    [anon_sym___bridge_retained] = ACTIONS(7145),
    [anon_sym___bridge_transfer] = ACTIONS(7145),
    [anon_sym___complex] = ACTIONS(7145),
    [anon_sym___const] = ACTIONS(7145),
    [anon_sym___imag] = ACTIONS(7145),
    [anon_sym___kindof] = ACTIONS(7145),
    [anon_sym___nonnull] = ACTIONS(7145),
    [anon_sym___nullable] = ACTIONS(7145),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7145),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7145),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7145),
    [anon_sym___real] = ACTIONS(7145),
    [anon_sym___strong] = ACTIONS(7145),
    [anon_sym___unsafe_unretained] = ACTIONS(7145),
    [anon_sym___unused] = ACTIONS(7145),
    [anon_sym___weak] = ACTIONS(7145),
    [anon_sym_alignas] = ACTIONS(7145),
    [anon_sym__Alignas] = ACTIONS(7145),
    [sym_primitive_type] = ACTIONS(7145),
    [anon_sym_enum] = ACTIONS(7145),
    [anon_sym_COLON] = ACTIONS(7147),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7145),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7145),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7145),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7145),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7145),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7145),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7145),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7145),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7145),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7145),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7145),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7145),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7145),
    [anon_sym_API_AVAILABLE] = ACTIONS(7145),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7145),
    [anon_sym_API_DEPRECATED] = ACTIONS(7145),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7145),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7145),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7145),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7145),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7145),
    [anon_sym___deprecated_msg] = ACTIONS(7145),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7145),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7145),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7145),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7145),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7145),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7145),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7145),
  },
  [STATE(2929)] = {
    [sym_attribute_specifier] = STATE(2978),
    [sym_identifier] = ACTIONS(7149),
    [anon_sym_COMMA] = ACTIONS(7151),
    [anon_sym_RPAREN] = ACTIONS(7151),
    [anon_sym_LPAREN2] = ACTIONS(7151),
    [anon_sym_STAR] = ACTIONS(7151),
    [anon_sym_CARET] = ACTIONS(7151),
    [anon_sym_SEMI] = ACTIONS(7151),
    [anon_sym___extension__] = ACTIONS(7149),
    [anon_sym_extern] = ACTIONS(7149),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7149),
    [anon_sym_LBRACK] = ACTIONS(7151),
    [anon_sym___declspec] = ACTIONS(7149),
    [anon_sym___based] = ACTIONS(7149),
    [anon_sym_signed] = ACTIONS(7149),
    [anon_sym_unsigned] = ACTIONS(7149),
    [anon_sym_long] = ACTIONS(7149),
    [anon_sym_short] = ACTIONS(7149),
    [anon_sym_static] = ACTIONS(7149),
    [anon_sym_auto] = ACTIONS(7149),
    [anon_sym_register] = ACTIONS(7149),
    [anon_sym_inline] = ACTIONS(7149),
    [anon_sym___inline] = ACTIONS(7149),
    [anon_sym___inline__] = ACTIONS(7149),
    [anon_sym___forceinline] = ACTIONS(7149),
    [anon_sym_thread_local] = ACTIONS(7149),
    [anon_sym___thread] = ACTIONS(7149),
    [anon_sym_CG_EXTERN] = ACTIONS(7149),
    [anon_sym_CG_INLINE] = ACTIONS(7149),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7149),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7149),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7149),
    [anon_sym_IBOutlet] = ACTIONS(7149),
    [anon_sym_IBInspectable] = ACTIONS(7149),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7149),
    [anon_sym_NS_INLINE] = ACTIONS(7149),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7149),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7149),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7149),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7149),
    [anon_sym_const] = ACTIONS(7149),
    [anon_sym_constexpr] = ACTIONS(7149),
    [anon_sym_volatile] = ACTIONS(7149),
    [anon_sym_restrict] = ACTIONS(7149),
    [anon_sym___restrict__] = ACTIONS(7149),
    [anon_sym__Atomic] = ACTIONS(7149),
    [anon_sym__Noreturn] = ACTIONS(7149),
    [anon_sym__Nonnull] = ACTIONS(7149),
    [anon_sym_nullable] = ACTIONS(7149),
    [anon_sym__Complex] = ACTIONS(7149),
    [anon_sym__Nullable] = ACTIONS(7149),
    [anon_sym__Nullable_result] = ACTIONS(7149),
    [anon_sym__Null_unspecified] = ACTIONS(7149),
    [anon_sym___autoreleasing] = ACTIONS(7149),
    [anon_sym___block] = ACTIONS(7149),
    [anon_sym___bridge] = ACTIONS(7149),
    [anon_sym___bridge_retained] = ACTIONS(7149),
    [anon_sym___bridge_transfer] = ACTIONS(7149),
    [anon_sym___complex] = ACTIONS(7149),
    [anon_sym___const] = ACTIONS(7149),
    [anon_sym___imag] = ACTIONS(7149),
    [anon_sym___kindof] = ACTIONS(7149),
    [anon_sym___nonnull] = ACTIONS(7149),
    [anon_sym___nullable] = ACTIONS(7149),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7149),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7149),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7149),
    [anon_sym___real] = ACTIONS(7149),
    [anon_sym___strong] = ACTIONS(7149),
    [anon_sym___unsafe_unretained] = ACTIONS(7149),
    [anon_sym___unused] = ACTIONS(7149),
    [anon_sym___weak] = ACTIONS(7149),
    [anon_sym_alignas] = ACTIONS(7149),
    [anon_sym__Alignas] = ACTIONS(7149),
    [sym_primitive_type] = ACTIONS(7149),
    [anon_sym_enum] = ACTIONS(7149),
    [anon_sym_COLON] = ACTIONS(7151),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7149),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7149),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7149),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7149),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7149),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7149),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7149),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7149),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7149),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7149),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7149),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7149),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7149),
    [anon_sym_API_AVAILABLE] = ACTIONS(7149),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7149),
    [anon_sym_API_DEPRECATED] = ACTIONS(7149),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7149),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7149),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7149),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7149),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7149),
    [anon_sym___deprecated_msg] = ACTIONS(7149),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7149),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7149),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7149),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7149),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7149),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7149),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7149),
  },
  [STATE(2930)] = {
    [sym_identifier] = ACTIONS(7153),
    [anon_sym_COMMA] = ACTIONS(7155),
    [anon_sym_RPAREN] = ACTIONS(7155),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_STAR] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_SEMI] = ACTIONS(7155),
    [anon_sym___extension__] = ACTIONS(7153),
    [anon_sym_extern] = ACTIONS(7153),
    [anon_sym___attribute__] = ACTIONS(7153),
    [anon_sym___attribute] = ACTIONS(7153),
    [anon_sym_noreturn] = ACTIONS(7153),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym___declspec] = ACTIONS(7153),
    [anon_sym___based] = ACTIONS(7153),
    [anon_sym_signed] = ACTIONS(7153),
    [anon_sym_unsigned] = ACTIONS(7153),
    [anon_sym_long] = ACTIONS(7153),
    [anon_sym_short] = ACTIONS(7153),
    [anon_sym_static] = ACTIONS(7153),
    [anon_sym_auto] = ACTIONS(7153),
    [anon_sym_register] = ACTIONS(7153),
    [anon_sym_inline] = ACTIONS(7153),
    [anon_sym___inline] = ACTIONS(7153),
    [anon_sym___inline__] = ACTIONS(7153),
    [anon_sym___forceinline] = ACTIONS(7153),
    [anon_sym_thread_local] = ACTIONS(7153),
    [anon_sym___thread] = ACTIONS(7153),
    [anon_sym_CG_EXTERN] = ACTIONS(7153),
    [anon_sym_CG_INLINE] = ACTIONS(7153),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7153),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7153),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7153),
    [anon_sym_IBOutlet] = ACTIONS(7153),
    [anon_sym_IBInspectable] = ACTIONS(7153),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7153),
    [anon_sym_NS_INLINE] = ACTIONS(7153),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7153),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7153),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7153),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7153),
    [anon_sym_const] = ACTIONS(7153),
    [anon_sym_constexpr] = ACTIONS(7153),
    [anon_sym_volatile] = ACTIONS(7153),
    [anon_sym_restrict] = ACTIONS(7153),
    [anon_sym___restrict__] = ACTIONS(7153),
    [anon_sym__Atomic] = ACTIONS(7153),
    [anon_sym__Noreturn] = ACTIONS(7153),
    [anon_sym__Nonnull] = ACTIONS(7153),
    [anon_sym_nullable] = ACTIONS(7153),
    [anon_sym__Complex] = ACTIONS(7153),
    [anon_sym__Nullable] = ACTIONS(7153),
    [anon_sym__Nullable_result] = ACTIONS(7153),
    [anon_sym__Null_unspecified] = ACTIONS(7153),
    [anon_sym___autoreleasing] = ACTIONS(7153),
    [anon_sym___block] = ACTIONS(7153),
    [anon_sym___bridge] = ACTIONS(7153),
    [anon_sym___bridge_retained] = ACTIONS(7153),
    [anon_sym___bridge_transfer] = ACTIONS(7153),
    [anon_sym___complex] = ACTIONS(7153),
    [anon_sym___const] = ACTIONS(7153),
    [anon_sym___imag] = ACTIONS(7153),
    [anon_sym___kindof] = ACTIONS(7153),
    [anon_sym___nonnull] = ACTIONS(7153),
    [anon_sym___nullable] = ACTIONS(7153),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7153),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7153),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7153),
    [anon_sym___real] = ACTIONS(7153),
    [anon_sym___strong] = ACTIONS(7153),
    [anon_sym___unsafe_unretained] = ACTIONS(7153),
    [anon_sym___unused] = ACTIONS(7153),
    [anon_sym___weak] = ACTIONS(7153),
    [anon_sym_alignas] = ACTIONS(7153),
    [anon_sym__Alignas] = ACTIONS(7153),
    [sym_primitive_type] = ACTIONS(7153),
    [anon_sym_enum] = ACTIONS(7153),
    [anon_sym_COLON] = ACTIONS(7155),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7153),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7153),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7153),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7153),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7153),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7153),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7153),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7153),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7153),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7153),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7153),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7153),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7153),
    [anon_sym_API_AVAILABLE] = ACTIONS(7153),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7153),
    [anon_sym_API_DEPRECATED] = ACTIONS(7153),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7153),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7153),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7153),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7153),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7153),
    [anon_sym___deprecated_msg] = ACTIONS(7153),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7153),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7153),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7153),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7153),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7153),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7153),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7153),
  },
  [STATE(2931)] = {
    [sym_identifier] = ACTIONS(7157),
    [anon_sym_COMMA] = ACTIONS(7159),
    [anon_sym_RPAREN] = ACTIONS(7159),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_STAR] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_SEMI] = ACTIONS(7159),
    [anon_sym___extension__] = ACTIONS(7157),
    [anon_sym_extern] = ACTIONS(7157),
    [anon_sym___attribute__] = ACTIONS(7157),
    [anon_sym___attribute] = ACTIONS(7157),
    [anon_sym_noreturn] = ACTIONS(7157),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym___declspec] = ACTIONS(7157),
    [anon_sym___based] = ACTIONS(7157),
    [anon_sym_signed] = ACTIONS(7157),
    [anon_sym_unsigned] = ACTIONS(7157),
    [anon_sym_long] = ACTIONS(7157),
    [anon_sym_short] = ACTIONS(7157),
    [anon_sym_static] = ACTIONS(7157),
    [anon_sym_auto] = ACTIONS(7157),
    [anon_sym_register] = ACTIONS(7157),
    [anon_sym_inline] = ACTIONS(7157),
    [anon_sym___inline] = ACTIONS(7157),
    [anon_sym___inline__] = ACTIONS(7157),
    [anon_sym___forceinline] = ACTIONS(7157),
    [anon_sym_thread_local] = ACTIONS(7157),
    [anon_sym___thread] = ACTIONS(7157),
    [anon_sym_CG_EXTERN] = ACTIONS(7157),
    [anon_sym_CG_INLINE] = ACTIONS(7157),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7157),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7157),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7157),
    [anon_sym_IBOutlet] = ACTIONS(7157),
    [anon_sym_IBInspectable] = ACTIONS(7157),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7157),
    [anon_sym_NS_INLINE] = ACTIONS(7157),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7157),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7157),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7157),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7157),
    [anon_sym_const] = ACTIONS(7157),
    [anon_sym_constexpr] = ACTIONS(7157),
    [anon_sym_volatile] = ACTIONS(7157),
    [anon_sym_restrict] = ACTIONS(7157),
    [anon_sym___restrict__] = ACTIONS(7157),
    [anon_sym__Atomic] = ACTIONS(7157),
    [anon_sym__Noreturn] = ACTIONS(7157),
    [anon_sym__Nonnull] = ACTIONS(7157),
    [anon_sym_nullable] = ACTIONS(7157),
    [anon_sym__Complex] = ACTIONS(7157),
    [anon_sym__Nullable] = ACTIONS(7157),
    [anon_sym__Nullable_result] = ACTIONS(7157),
    [anon_sym__Null_unspecified] = ACTIONS(7157),
    [anon_sym___autoreleasing] = ACTIONS(7157),
    [anon_sym___block] = ACTIONS(7157),
    [anon_sym___bridge] = ACTIONS(7157),
    [anon_sym___bridge_retained] = ACTIONS(7157),
    [anon_sym___bridge_transfer] = ACTIONS(7157),
    [anon_sym___complex] = ACTIONS(7157),
    [anon_sym___const] = ACTIONS(7157),
    [anon_sym___imag] = ACTIONS(7157),
    [anon_sym___kindof] = ACTIONS(7157),
    [anon_sym___nonnull] = ACTIONS(7157),
    [anon_sym___nullable] = ACTIONS(7157),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7157),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7157),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7157),
    [anon_sym___real] = ACTIONS(7157),
    [anon_sym___strong] = ACTIONS(7157),
    [anon_sym___unsafe_unretained] = ACTIONS(7157),
    [anon_sym___unused] = ACTIONS(7157),
    [anon_sym___weak] = ACTIONS(7157),
    [anon_sym_alignas] = ACTIONS(7157),
    [anon_sym__Alignas] = ACTIONS(7157),
    [sym_primitive_type] = ACTIONS(7157),
    [anon_sym_enum] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7159),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7157),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7157),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7157),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7157),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7157),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7157),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7157),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7157),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7157),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7157),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7157),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7157),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7157),
    [anon_sym_API_AVAILABLE] = ACTIONS(7157),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7157),
    [anon_sym_API_DEPRECATED] = ACTIONS(7157),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7157),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7157),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7157),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7157),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7157),
    [anon_sym___deprecated_msg] = ACTIONS(7157),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7157),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7157),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7157),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7157),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7157),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7157),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7157),
  },
  [STATE(2932)] = {
    [sym_identifier] = ACTIONS(7161),
    [anon_sym_COMMA] = ACTIONS(7163),
    [anon_sym_RPAREN] = ACTIONS(7163),
    [anon_sym_LPAREN2] = ACTIONS(7163),
    [anon_sym_STAR] = ACTIONS(7163),
    [anon_sym_CARET] = ACTIONS(7163),
    [anon_sym_SEMI] = ACTIONS(7163),
    [anon_sym___extension__] = ACTIONS(7161),
    [anon_sym_extern] = ACTIONS(7161),
    [anon_sym___attribute__] = ACTIONS(7161),
    [anon_sym___attribute] = ACTIONS(7161),
    [anon_sym_noreturn] = ACTIONS(7161),
    [anon_sym_LBRACK] = ACTIONS(7163),
    [anon_sym___declspec] = ACTIONS(7161),
    [anon_sym___based] = ACTIONS(7161),
    [anon_sym_signed] = ACTIONS(7161),
    [anon_sym_unsigned] = ACTIONS(7161),
    [anon_sym_long] = ACTIONS(7161),
    [anon_sym_short] = ACTIONS(7161),
    [anon_sym_static] = ACTIONS(7161),
    [anon_sym_auto] = ACTIONS(7161),
    [anon_sym_register] = ACTIONS(7161),
    [anon_sym_inline] = ACTIONS(7161),
    [anon_sym___inline] = ACTIONS(7161),
    [anon_sym___inline__] = ACTIONS(7161),
    [anon_sym___forceinline] = ACTIONS(7161),
    [anon_sym_thread_local] = ACTIONS(7161),
    [anon_sym___thread] = ACTIONS(7161),
    [anon_sym_CG_EXTERN] = ACTIONS(7161),
    [anon_sym_CG_INLINE] = ACTIONS(7161),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7161),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7161),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7161),
    [anon_sym_IBOutlet] = ACTIONS(7161),
    [anon_sym_IBInspectable] = ACTIONS(7161),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7161),
    [anon_sym_NS_INLINE] = ACTIONS(7161),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7161),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7161),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7161),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7161),
    [anon_sym_const] = ACTIONS(7161),
    [anon_sym_constexpr] = ACTIONS(7161),
    [anon_sym_volatile] = ACTIONS(7161),
    [anon_sym_restrict] = ACTIONS(7161),
    [anon_sym___restrict__] = ACTIONS(7161),
    [anon_sym__Atomic] = ACTIONS(7161),
    [anon_sym__Noreturn] = ACTIONS(7161),
    [anon_sym__Nonnull] = ACTIONS(7161),
    [anon_sym_nullable] = ACTIONS(7161),
    [anon_sym__Complex] = ACTIONS(7161),
    [anon_sym__Nullable] = ACTIONS(7161),
    [anon_sym__Nullable_result] = ACTIONS(7161),
    [anon_sym__Null_unspecified] = ACTIONS(7161),
    [anon_sym___autoreleasing] = ACTIONS(7161),
    [anon_sym___block] = ACTIONS(7161),
    [anon_sym___bridge] = ACTIONS(7161),
    [anon_sym___bridge_retained] = ACTIONS(7161),
    [anon_sym___bridge_transfer] = ACTIONS(7161),
    [anon_sym___complex] = ACTIONS(7161),
    [anon_sym___const] = ACTIONS(7161),
    [anon_sym___imag] = ACTIONS(7161),
    [anon_sym___kindof] = ACTIONS(7161),
    [anon_sym___nonnull] = ACTIONS(7161),
    [anon_sym___nullable] = ACTIONS(7161),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7161),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7161),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7161),
    [anon_sym___real] = ACTIONS(7161),
    [anon_sym___strong] = ACTIONS(7161),
    [anon_sym___unsafe_unretained] = ACTIONS(7161),
    [anon_sym___unused] = ACTIONS(7161),
    [anon_sym___weak] = ACTIONS(7161),
    [anon_sym_alignas] = ACTIONS(7161),
    [anon_sym__Alignas] = ACTIONS(7161),
    [sym_primitive_type] = ACTIONS(7161),
    [anon_sym_enum] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7163),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7161),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7161),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7161),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7161),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7161),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7161),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7161),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7161),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7161),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7161),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7161),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7161),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7161),
    [anon_sym_API_AVAILABLE] = ACTIONS(7161),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7161),
    [anon_sym_API_DEPRECATED] = ACTIONS(7161),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7161),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7161),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7161),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7161),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7161),
    [anon_sym___deprecated_msg] = ACTIONS(7161),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7161),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7161),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7161),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7161),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7161),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7161),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7161),
  },
  [STATE(2933)] = {
    [sym__declaration_modifiers] = STATE(2866),
    [sym_attribute_specifier] = STATE(2866),
    [sym_attribute_declaration] = STATE(2866),
    [sym_ms_declspec_modifier] = STATE(2866),
    [sym_storage_class_specifier] = STATE(2866),
    [sym_type_qualifier] = STATE(2866),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2866),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2866),
    [sym_identifier] = ACTIONS(6899),
    [anon_sym_LPAREN2] = ACTIONS(6901),
    [anon_sym_STAR] = ACTIONS(6901),
    [anon_sym_CARET] = ACTIONS(6901),
    [anon_sym_SEMI] = ACTIONS(7165),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6897),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(6899),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2934)] = {
    [sym_identifier] = ACTIONS(7167),
    [anon_sym_COMMA] = ACTIONS(7169),
    [anon_sym_RPAREN] = ACTIONS(7169),
    [anon_sym_LPAREN2] = ACTIONS(7169),
    [anon_sym_STAR] = ACTIONS(7169),
    [anon_sym_CARET] = ACTIONS(7169),
    [anon_sym_SEMI] = ACTIONS(7169),
    [anon_sym___extension__] = ACTIONS(7167),
    [anon_sym_extern] = ACTIONS(7167),
    [anon_sym___attribute__] = ACTIONS(7167),
    [anon_sym___attribute] = ACTIONS(7167),
    [anon_sym_noreturn] = ACTIONS(7167),
    [anon_sym_LBRACK] = ACTIONS(7169),
    [anon_sym___declspec] = ACTIONS(7167),
    [anon_sym___based] = ACTIONS(7167),
    [anon_sym_signed] = ACTIONS(7167),
    [anon_sym_unsigned] = ACTIONS(7167),
    [anon_sym_long] = ACTIONS(7167),
    [anon_sym_short] = ACTIONS(7167),
    [anon_sym_static] = ACTIONS(7167),
    [anon_sym_auto] = ACTIONS(7167),
    [anon_sym_register] = ACTIONS(7167),
    [anon_sym_inline] = ACTIONS(7167),
    [anon_sym___inline] = ACTIONS(7167),
    [anon_sym___inline__] = ACTIONS(7167),
    [anon_sym___forceinline] = ACTIONS(7167),
    [anon_sym_thread_local] = ACTIONS(7167),
    [anon_sym___thread] = ACTIONS(7167),
    [anon_sym_CG_EXTERN] = ACTIONS(7167),
    [anon_sym_CG_INLINE] = ACTIONS(7167),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7167),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7167),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7167),
    [anon_sym_IBOutlet] = ACTIONS(7167),
    [anon_sym_IBInspectable] = ACTIONS(7167),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7167),
    [anon_sym_NS_INLINE] = ACTIONS(7167),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7167),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7167),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7167),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7167),
    [anon_sym_const] = ACTIONS(7167),
    [anon_sym_constexpr] = ACTIONS(7167),
    [anon_sym_volatile] = ACTIONS(7167),
    [anon_sym_restrict] = ACTIONS(7167),
    [anon_sym___restrict__] = ACTIONS(7167),
    [anon_sym__Atomic] = ACTIONS(7167),
    [anon_sym__Noreturn] = ACTIONS(7167),
    [anon_sym__Nonnull] = ACTIONS(7167),
    [anon_sym_nullable] = ACTIONS(7167),
    [anon_sym__Complex] = ACTIONS(7167),
    [anon_sym__Nullable] = ACTIONS(7167),
    [anon_sym__Nullable_result] = ACTIONS(7167),
    [anon_sym__Null_unspecified] = ACTIONS(7167),
    [anon_sym___autoreleasing] = ACTIONS(7167),
    [anon_sym___block] = ACTIONS(7167),
    [anon_sym___bridge] = ACTIONS(7167),
    [anon_sym___bridge_retained] = ACTIONS(7167),
    [anon_sym___bridge_transfer] = ACTIONS(7167),
    [anon_sym___complex] = ACTIONS(7167),
    [anon_sym___const] = ACTIONS(7167),
    [anon_sym___imag] = ACTIONS(7167),
    [anon_sym___kindof] = ACTIONS(7167),
    [anon_sym___nonnull] = ACTIONS(7167),
    [anon_sym___nullable] = ACTIONS(7167),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7167),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7167),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7167),
    [anon_sym___real] = ACTIONS(7167),
    [anon_sym___strong] = ACTIONS(7167),
    [anon_sym___unsafe_unretained] = ACTIONS(7167),
    [anon_sym___unused] = ACTIONS(7167),
    [anon_sym___weak] = ACTIONS(7167),
    [anon_sym_alignas] = ACTIONS(7167),
    [anon_sym__Alignas] = ACTIONS(7167),
    [sym_primitive_type] = ACTIONS(7167),
    [anon_sym_enum] = ACTIONS(7167),
    [anon_sym_COLON] = ACTIONS(7169),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7167),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7167),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7167),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7167),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7167),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7167),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7167),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7167),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7167),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7167),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7167),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7167),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7167),
    [anon_sym_API_AVAILABLE] = ACTIONS(7167),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7167),
    [anon_sym_API_DEPRECATED] = ACTIONS(7167),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7167),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7167),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7167),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7167),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7167),
    [anon_sym___deprecated_msg] = ACTIONS(7167),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7167),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7167),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7167),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7167),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7167),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7167),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7167),
  },
  [STATE(2935)] = {
    [sym_identifier] = ACTIONS(7171),
    [anon_sym_COMMA] = ACTIONS(7173),
    [anon_sym_RPAREN] = ACTIONS(7173),
    [anon_sym_LPAREN2] = ACTIONS(7173),
    [anon_sym_STAR] = ACTIONS(7173),
    [anon_sym_CARET] = ACTIONS(7173),
    [anon_sym_SEMI] = ACTIONS(7173),
    [anon_sym___extension__] = ACTIONS(7171),
    [anon_sym_extern] = ACTIONS(7171),
    [anon_sym___attribute__] = ACTIONS(7171),
    [anon_sym___attribute] = ACTIONS(7171),
    [anon_sym_noreturn] = ACTIONS(7171),
    [anon_sym_LBRACK] = ACTIONS(7173),
    [anon_sym___declspec] = ACTIONS(7171),
    [anon_sym___based] = ACTIONS(7171),
    [anon_sym_signed] = ACTIONS(7171),
    [anon_sym_unsigned] = ACTIONS(7171),
    [anon_sym_long] = ACTIONS(7171),
    [anon_sym_short] = ACTIONS(7171),
    [anon_sym_static] = ACTIONS(7171),
    [anon_sym_auto] = ACTIONS(7171),
    [anon_sym_register] = ACTIONS(7171),
    [anon_sym_inline] = ACTIONS(7171),
    [anon_sym___inline] = ACTIONS(7171),
    [anon_sym___inline__] = ACTIONS(7171),
    [anon_sym___forceinline] = ACTIONS(7171),
    [anon_sym_thread_local] = ACTIONS(7171),
    [anon_sym___thread] = ACTIONS(7171),
    [anon_sym_CG_EXTERN] = ACTIONS(7171),
    [anon_sym_CG_INLINE] = ACTIONS(7171),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7171),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7171),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7171),
    [anon_sym_IBOutlet] = ACTIONS(7171),
    [anon_sym_IBInspectable] = ACTIONS(7171),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7171),
    [anon_sym_NS_INLINE] = ACTIONS(7171),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7171),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7171),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7171),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7171),
    [anon_sym_const] = ACTIONS(7171),
    [anon_sym_constexpr] = ACTIONS(7171),
    [anon_sym_volatile] = ACTIONS(7171),
    [anon_sym_restrict] = ACTIONS(7171),
    [anon_sym___restrict__] = ACTIONS(7171),
    [anon_sym__Atomic] = ACTIONS(7171),
    [anon_sym__Noreturn] = ACTIONS(7171),
    [anon_sym__Nonnull] = ACTIONS(7171),
    [anon_sym_nullable] = ACTIONS(7171),
    [anon_sym__Complex] = ACTIONS(7171),
    [anon_sym__Nullable] = ACTIONS(7171),
    [anon_sym__Nullable_result] = ACTIONS(7171),
    [anon_sym__Null_unspecified] = ACTIONS(7171),
    [anon_sym___autoreleasing] = ACTIONS(7171),
    [anon_sym___block] = ACTIONS(7171),
    [anon_sym___bridge] = ACTIONS(7171),
    [anon_sym___bridge_retained] = ACTIONS(7171),
    [anon_sym___bridge_transfer] = ACTIONS(7171),
    [anon_sym___complex] = ACTIONS(7171),
    [anon_sym___const] = ACTIONS(7171),
    [anon_sym___imag] = ACTIONS(7171),
    [anon_sym___kindof] = ACTIONS(7171),
    [anon_sym___nonnull] = ACTIONS(7171),
    [anon_sym___nullable] = ACTIONS(7171),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7171),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7171),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7171),
    [anon_sym___real] = ACTIONS(7171),
    [anon_sym___strong] = ACTIONS(7171),
    [anon_sym___unsafe_unretained] = ACTIONS(7171),
    [anon_sym___unused] = ACTIONS(7171),
    [anon_sym___weak] = ACTIONS(7171),
    [anon_sym_alignas] = ACTIONS(7171),
    [anon_sym__Alignas] = ACTIONS(7171),
    [sym_primitive_type] = ACTIONS(7171),
    [anon_sym_enum] = ACTIONS(7171),
    [anon_sym_COLON] = ACTIONS(7173),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7171),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7171),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7171),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7171),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7171),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7171),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7171),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7171),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7171),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7171),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7171),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7171),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7171),
    [anon_sym_API_AVAILABLE] = ACTIONS(7171),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7171),
    [anon_sym_API_DEPRECATED] = ACTIONS(7171),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7171),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7171),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7171),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7171),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7171),
    [anon_sym___deprecated_msg] = ACTIONS(7171),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7171),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7171),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7171),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7171),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7171),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7171),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7171),
  },
  [STATE(2936)] = {
    [sym_identifier] = ACTIONS(7175),
    [anon_sym_COMMA] = ACTIONS(7177),
    [anon_sym_RPAREN] = ACTIONS(7177),
    [anon_sym_LPAREN2] = ACTIONS(7177),
    [anon_sym_STAR] = ACTIONS(7177),
    [anon_sym_CARET] = ACTIONS(7177),
    [anon_sym_SEMI] = ACTIONS(7177),
    [anon_sym___extension__] = ACTIONS(7175),
    [anon_sym_extern] = ACTIONS(7175),
    [anon_sym___attribute__] = ACTIONS(7175),
    [anon_sym___attribute] = ACTIONS(7175),
    [anon_sym_noreturn] = ACTIONS(7175),
    [anon_sym_LBRACK] = ACTIONS(7177),
    [anon_sym___declspec] = ACTIONS(7175),
    [anon_sym___based] = ACTIONS(7175),
    [anon_sym_signed] = ACTIONS(7175),
    [anon_sym_unsigned] = ACTIONS(7175),
    [anon_sym_long] = ACTIONS(7175),
    [anon_sym_short] = ACTIONS(7175),
    [anon_sym_static] = ACTIONS(7175),
    [anon_sym_auto] = ACTIONS(7175),
    [anon_sym_register] = ACTIONS(7175),
    [anon_sym_inline] = ACTIONS(7175),
    [anon_sym___inline] = ACTIONS(7175),
    [anon_sym___inline__] = ACTIONS(7175),
    [anon_sym___forceinline] = ACTIONS(7175),
    [anon_sym_thread_local] = ACTIONS(7175),
    [anon_sym___thread] = ACTIONS(7175),
    [anon_sym_CG_EXTERN] = ACTIONS(7175),
    [anon_sym_CG_INLINE] = ACTIONS(7175),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7175),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7175),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7175),
    [anon_sym_IBOutlet] = ACTIONS(7175),
    [anon_sym_IBInspectable] = ACTIONS(7175),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7175),
    [anon_sym_NS_INLINE] = ACTIONS(7175),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7175),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7175),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7175),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7175),
    [anon_sym_const] = ACTIONS(7175),
    [anon_sym_constexpr] = ACTIONS(7175),
    [anon_sym_volatile] = ACTIONS(7175),
    [anon_sym_restrict] = ACTIONS(7175),
    [anon_sym___restrict__] = ACTIONS(7175),
    [anon_sym__Atomic] = ACTIONS(7175),
    [anon_sym__Noreturn] = ACTIONS(7175),
    [anon_sym__Nonnull] = ACTIONS(7175),
    [anon_sym_nullable] = ACTIONS(7175),
    [anon_sym__Complex] = ACTIONS(7175),
    [anon_sym__Nullable] = ACTIONS(7175),
    [anon_sym__Nullable_result] = ACTIONS(7175),
    [anon_sym__Null_unspecified] = ACTIONS(7175),
    [anon_sym___autoreleasing] = ACTIONS(7175),
    [anon_sym___block] = ACTIONS(7175),
    [anon_sym___bridge] = ACTIONS(7175),
    [anon_sym___bridge_retained] = ACTIONS(7175),
    [anon_sym___bridge_transfer] = ACTIONS(7175),
    [anon_sym___complex] = ACTIONS(7175),
    [anon_sym___const] = ACTIONS(7175),
    [anon_sym___imag] = ACTIONS(7175),
    [anon_sym___kindof] = ACTIONS(7175),
    [anon_sym___nonnull] = ACTIONS(7175),
    [anon_sym___nullable] = ACTIONS(7175),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7175),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7175),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7175),
    [anon_sym___real] = ACTIONS(7175),
    [anon_sym___strong] = ACTIONS(7175),
    [anon_sym___unsafe_unretained] = ACTIONS(7175),
    [anon_sym___unused] = ACTIONS(7175),
    [anon_sym___weak] = ACTIONS(7175),
    [anon_sym_alignas] = ACTIONS(7175),
    [anon_sym__Alignas] = ACTIONS(7175),
    [sym_primitive_type] = ACTIONS(7175),
    [anon_sym_enum] = ACTIONS(7175),
    [anon_sym_COLON] = ACTIONS(7177),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7175),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7175),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7175),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7175),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7175),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7175),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7175),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7175),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7175),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7175),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7175),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7175),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7175),
    [anon_sym_API_AVAILABLE] = ACTIONS(7175),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7175),
    [anon_sym_API_DEPRECATED] = ACTIONS(7175),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7175),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7175),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7175),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7175),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7175),
    [anon_sym___deprecated_msg] = ACTIONS(7175),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7175),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7175),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7175),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7175),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7175),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7175),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7175),
  },
  [STATE(2937)] = {
    [sym_identifier] = ACTIONS(7179),
    [anon_sym_COMMA] = ACTIONS(7181),
    [anon_sym_RPAREN] = ACTIONS(7181),
    [anon_sym_LPAREN2] = ACTIONS(7181),
    [anon_sym_STAR] = ACTIONS(7181),
    [anon_sym_CARET] = ACTIONS(7181),
    [anon_sym_SEMI] = ACTIONS(7181),
    [anon_sym___extension__] = ACTIONS(7179),
    [anon_sym_extern] = ACTIONS(7179),
    [anon_sym___attribute__] = ACTIONS(7179),
    [anon_sym___attribute] = ACTIONS(7179),
    [anon_sym_noreturn] = ACTIONS(7179),
    [anon_sym_LBRACK] = ACTIONS(7181),
    [anon_sym___declspec] = ACTIONS(7179),
    [anon_sym___based] = ACTIONS(7179),
    [anon_sym_signed] = ACTIONS(7179),
    [anon_sym_unsigned] = ACTIONS(7179),
    [anon_sym_long] = ACTIONS(7179),
    [anon_sym_short] = ACTIONS(7179),
    [anon_sym_static] = ACTIONS(7179),
    [anon_sym_auto] = ACTIONS(7179),
    [anon_sym_register] = ACTIONS(7179),
    [anon_sym_inline] = ACTIONS(7179),
    [anon_sym___inline] = ACTIONS(7179),
    [anon_sym___inline__] = ACTIONS(7179),
    [anon_sym___forceinline] = ACTIONS(7179),
    [anon_sym_thread_local] = ACTIONS(7179),
    [anon_sym___thread] = ACTIONS(7179),
    [anon_sym_CG_EXTERN] = ACTIONS(7179),
    [anon_sym_CG_INLINE] = ACTIONS(7179),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7179),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7179),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7179),
    [anon_sym_IBOutlet] = ACTIONS(7179),
    [anon_sym_IBInspectable] = ACTIONS(7179),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7179),
    [anon_sym_NS_INLINE] = ACTIONS(7179),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7179),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7179),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7179),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7179),
    [anon_sym_const] = ACTIONS(7179),
    [anon_sym_constexpr] = ACTIONS(7179),
    [anon_sym_volatile] = ACTIONS(7179),
    [anon_sym_restrict] = ACTIONS(7179),
    [anon_sym___restrict__] = ACTIONS(7179),
    [anon_sym__Atomic] = ACTIONS(7179),
    [anon_sym__Noreturn] = ACTIONS(7179),
    [anon_sym__Nonnull] = ACTIONS(7179),
    [anon_sym_nullable] = ACTIONS(7179),
    [anon_sym__Complex] = ACTIONS(7179),
    [anon_sym__Nullable] = ACTIONS(7179),
    [anon_sym__Nullable_result] = ACTIONS(7179),
    [anon_sym__Null_unspecified] = ACTIONS(7179),
    [anon_sym___autoreleasing] = ACTIONS(7179),
    [anon_sym___block] = ACTIONS(7179),
    [anon_sym___bridge] = ACTIONS(7179),
    [anon_sym___bridge_retained] = ACTIONS(7179),
    [anon_sym___bridge_transfer] = ACTIONS(7179),
    [anon_sym___complex] = ACTIONS(7179),
    [anon_sym___const] = ACTIONS(7179),
    [anon_sym___imag] = ACTIONS(7179),
    [anon_sym___kindof] = ACTIONS(7179),
    [anon_sym___nonnull] = ACTIONS(7179),
    [anon_sym___nullable] = ACTIONS(7179),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7179),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7179),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7179),
    [anon_sym___real] = ACTIONS(7179),
    [anon_sym___strong] = ACTIONS(7179),
    [anon_sym___unsafe_unretained] = ACTIONS(7179),
    [anon_sym___unused] = ACTIONS(7179),
    [anon_sym___weak] = ACTIONS(7179),
    [anon_sym_alignas] = ACTIONS(7179),
    [anon_sym__Alignas] = ACTIONS(7179),
    [sym_primitive_type] = ACTIONS(7179),
    [anon_sym_enum] = ACTIONS(7179),
    [anon_sym_COLON] = ACTIONS(7181),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7179),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7179),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7179),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7179),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7179),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7179),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7179),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7179),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7179),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7179),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7179),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7179),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7179),
    [anon_sym_API_AVAILABLE] = ACTIONS(7179),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7179),
    [anon_sym_API_DEPRECATED] = ACTIONS(7179),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7179),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7179),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7179),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7179),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7179),
    [anon_sym___deprecated_msg] = ACTIONS(7179),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7179),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7179),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7179),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7179),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7179),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7179),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7179),
  },
  [STATE(2938)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2982),
    [sym_identifier] = ACTIONS(7183),
    [anon_sym_COMMA] = ACTIONS(7185),
    [anon_sym_RPAREN] = ACTIONS(7185),
    [anon_sym_LPAREN2] = ACTIONS(7185),
    [anon_sym_STAR] = ACTIONS(7185),
    [anon_sym_CARET] = ACTIONS(7185),
    [anon_sym_SEMI] = ACTIONS(7185),
    [anon_sym___extension__] = ACTIONS(7183),
    [anon_sym_extern] = ACTIONS(7183),
    [anon_sym___attribute__] = ACTIONS(7183),
    [anon_sym___attribute] = ACTIONS(7183),
    [anon_sym_noreturn] = ACTIONS(7183),
    [anon_sym_LBRACK] = ACTIONS(7185),
    [anon_sym___declspec] = ACTIONS(7183),
    [anon_sym___based] = ACTIONS(7183),
    [anon_sym_signed] = ACTIONS(7187),
    [anon_sym_unsigned] = ACTIONS(7187),
    [anon_sym_long] = ACTIONS(7187),
    [anon_sym_short] = ACTIONS(7187),
    [anon_sym_static] = ACTIONS(7183),
    [anon_sym_auto] = ACTIONS(7183),
    [anon_sym_register] = ACTIONS(7183),
    [anon_sym_inline] = ACTIONS(7183),
    [anon_sym___inline] = ACTIONS(7183),
    [anon_sym___inline__] = ACTIONS(7183),
    [anon_sym___forceinline] = ACTIONS(7183),
    [anon_sym_thread_local] = ACTIONS(7183),
    [anon_sym___thread] = ACTIONS(7183),
    [anon_sym_CG_EXTERN] = ACTIONS(7183),
    [anon_sym_CG_INLINE] = ACTIONS(7183),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7183),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7183),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7183),
    [anon_sym_IBOutlet] = ACTIONS(7183),
    [anon_sym_IBInspectable] = ACTIONS(7183),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7183),
    [anon_sym_NS_INLINE] = ACTIONS(7183),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7183),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7183),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7183),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7183),
    [anon_sym_const] = ACTIONS(7183),
    [anon_sym_constexpr] = ACTIONS(7183),
    [anon_sym_volatile] = ACTIONS(7183),
    [anon_sym_restrict] = ACTIONS(7183),
    [anon_sym___restrict__] = ACTIONS(7183),
    [anon_sym__Atomic] = ACTIONS(7183),
    [anon_sym__Noreturn] = ACTIONS(7183),
    [anon_sym__Nonnull] = ACTIONS(7183),
    [anon_sym_nullable] = ACTIONS(7183),
    [anon_sym__Complex] = ACTIONS(7183),
    [anon_sym__Nullable] = ACTIONS(7183),
    [anon_sym__Nullable_result] = ACTIONS(7183),
    [anon_sym__Null_unspecified] = ACTIONS(7183),
    [anon_sym___autoreleasing] = ACTIONS(7183),
    [anon_sym___block] = ACTIONS(7183),
    [anon_sym___bridge] = ACTIONS(7183),
    [anon_sym___bridge_retained] = ACTIONS(7183),
    [anon_sym___bridge_transfer] = ACTIONS(7183),
    [anon_sym___complex] = ACTIONS(7183),
    [anon_sym___const] = ACTIONS(7183),
    [anon_sym___imag] = ACTIONS(7183),
    [anon_sym___kindof] = ACTIONS(7183),
    [anon_sym___nonnull] = ACTIONS(7183),
    [anon_sym___nullable] = ACTIONS(7183),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7183),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7183),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7183),
    [anon_sym___real] = ACTIONS(7183),
    [anon_sym___strong] = ACTIONS(7183),
    [anon_sym___unsafe_unretained] = ACTIONS(7183),
    [anon_sym___unused] = ACTIONS(7183),
    [anon_sym___weak] = ACTIONS(7183),
    [anon_sym_alignas] = ACTIONS(7183),
    [anon_sym__Alignas] = ACTIONS(7183),
    [anon_sym_enum] = ACTIONS(7183),
    [anon_sym_COLON] = ACTIONS(7185),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7183),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7183),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7183),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7183),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7183),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7183),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7183),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7183),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7183),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7183),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7183),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7183),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7183),
    [anon_sym_API_AVAILABLE] = ACTIONS(7183),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7183),
    [anon_sym_API_DEPRECATED] = ACTIONS(7183),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7183),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7183),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7183),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7183),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7183),
    [anon_sym___deprecated_msg] = ACTIONS(7183),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7183),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7183),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7183),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7183),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7183),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7183),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7183),
  },
  [STATE(2939)] = {
    [sym_identifier] = ACTIONS(7189),
    [anon_sym_COMMA] = ACTIONS(7191),
    [anon_sym_RPAREN] = ACTIONS(7191),
    [anon_sym_LPAREN2] = ACTIONS(7191),
    [anon_sym_STAR] = ACTIONS(7191),
    [anon_sym_CARET] = ACTIONS(7191),
    [anon_sym_SEMI] = ACTIONS(7191),
    [anon_sym___extension__] = ACTIONS(7189),
    [anon_sym_extern] = ACTIONS(7189),
    [anon_sym___attribute__] = ACTIONS(7189),
    [anon_sym___attribute] = ACTIONS(7189),
    [anon_sym_noreturn] = ACTIONS(7189),
    [anon_sym_LBRACK] = ACTIONS(7191),
    [anon_sym___declspec] = ACTIONS(7189),
    [anon_sym___based] = ACTIONS(7189),
    [anon_sym_signed] = ACTIONS(7189),
    [anon_sym_unsigned] = ACTIONS(7189),
    [anon_sym_long] = ACTIONS(7189),
    [anon_sym_short] = ACTIONS(7189),
    [anon_sym_static] = ACTIONS(7189),
    [anon_sym_auto] = ACTIONS(7189),
    [anon_sym_register] = ACTIONS(7189),
    [anon_sym_inline] = ACTIONS(7189),
    [anon_sym___inline] = ACTIONS(7189),
    [anon_sym___inline__] = ACTIONS(7189),
    [anon_sym___forceinline] = ACTIONS(7189),
    [anon_sym_thread_local] = ACTIONS(7189),
    [anon_sym___thread] = ACTIONS(7189),
    [anon_sym_CG_EXTERN] = ACTIONS(7189),
    [anon_sym_CG_INLINE] = ACTIONS(7189),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7189),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7189),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7189),
    [anon_sym_IBOutlet] = ACTIONS(7189),
    [anon_sym_IBInspectable] = ACTIONS(7189),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7189),
    [anon_sym_NS_INLINE] = ACTIONS(7189),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7189),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7189),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7189),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7189),
    [anon_sym_const] = ACTIONS(7189),
    [anon_sym_constexpr] = ACTIONS(7189),
    [anon_sym_volatile] = ACTIONS(7189),
    [anon_sym_restrict] = ACTIONS(7189),
    [anon_sym___restrict__] = ACTIONS(7189),
    [anon_sym__Atomic] = ACTIONS(7189),
    [anon_sym__Noreturn] = ACTIONS(7189),
    [anon_sym__Nonnull] = ACTIONS(7189),
    [anon_sym_nullable] = ACTIONS(7189),
    [anon_sym__Complex] = ACTIONS(7189),
    [anon_sym__Nullable] = ACTIONS(7189),
    [anon_sym__Nullable_result] = ACTIONS(7189),
    [anon_sym__Null_unspecified] = ACTIONS(7189),
    [anon_sym___autoreleasing] = ACTIONS(7189),
    [anon_sym___block] = ACTIONS(7189),
    [anon_sym___bridge] = ACTIONS(7189),
    [anon_sym___bridge_retained] = ACTIONS(7189),
    [anon_sym___bridge_transfer] = ACTIONS(7189),
    [anon_sym___complex] = ACTIONS(7189),
    [anon_sym___const] = ACTIONS(7189),
    [anon_sym___imag] = ACTIONS(7189),
    [anon_sym___kindof] = ACTIONS(7189),
    [anon_sym___nonnull] = ACTIONS(7189),
    [anon_sym___nullable] = ACTIONS(7189),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7189),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7189),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7189),
    [anon_sym___real] = ACTIONS(7189),
    [anon_sym___strong] = ACTIONS(7189),
    [anon_sym___unsafe_unretained] = ACTIONS(7189),
    [anon_sym___unused] = ACTIONS(7189),
    [anon_sym___weak] = ACTIONS(7189),
    [anon_sym_alignas] = ACTIONS(7189),
    [anon_sym__Alignas] = ACTIONS(7189),
    [sym_primitive_type] = ACTIONS(7189),
    [anon_sym_enum] = ACTIONS(7189),
    [anon_sym_COLON] = ACTIONS(7191),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7189),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7189),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7189),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7189),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7189),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7189),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7189),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7189),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7189),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7189),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7189),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7189),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7189),
    [anon_sym_API_AVAILABLE] = ACTIONS(7189),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7189),
    [anon_sym_API_DEPRECATED] = ACTIONS(7189),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7189),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7189),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7189),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7189),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7189),
    [anon_sym___deprecated_msg] = ACTIONS(7189),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7189),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7189),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7189),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7189),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7189),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7189),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7189),
  },
  [STATE(2940)] = {
    [sym_identifier] = ACTIONS(7193),
    [anon_sym_COMMA] = ACTIONS(7195),
    [anon_sym_RPAREN] = ACTIONS(7195),
    [anon_sym_LPAREN2] = ACTIONS(7195),
    [anon_sym_STAR] = ACTIONS(7195),
    [anon_sym_CARET] = ACTIONS(7195),
    [anon_sym_SEMI] = ACTIONS(7195),
    [anon_sym___extension__] = ACTIONS(7193),
    [anon_sym_extern] = ACTIONS(7193),
    [anon_sym___attribute__] = ACTIONS(7193),
    [anon_sym___attribute] = ACTIONS(7193),
    [anon_sym_noreturn] = ACTIONS(7193),
    [anon_sym_LBRACK] = ACTIONS(7195),
    [anon_sym___declspec] = ACTIONS(7193),
    [anon_sym___based] = ACTIONS(7193),
    [anon_sym_signed] = ACTIONS(7193),
    [anon_sym_unsigned] = ACTIONS(7193),
    [anon_sym_long] = ACTIONS(7193),
    [anon_sym_short] = ACTIONS(7193),
    [anon_sym_static] = ACTIONS(7193),
    [anon_sym_auto] = ACTIONS(7193),
    [anon_sym_register] = ACTIONS(7193),
    [anon_sym_inline] = ACTIONS(7193),
    [anon_sym___inline] = ACTIONS(7193),
    [anon_sym___inline__] = ACTIONS(7193),
    [anon_sym___forceinline] = ACTIONS(7193),
    [anon_sym_thread_local] = ACTIONS(7193),
    [anon_sym___thread] = ACTIONS(7193),
    [anon_sym_CG_EXTERN] = ACTIONS(7193),
    [anon_sym_CG_INLINE] = ACTIONS(7193),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7193),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7193),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7193),
    [anon_sym_IBOutlet] = ACTIONS(7193),
    [anon_sym_IBInspectable] = ACTIONS(7193),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7193),
    [anon_sym_NS_INLINE] = ACTIONS(7193),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7193),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7193),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7193),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7193),
    [anon_sym_const] = ACTIONS(7193),
    [anon_sym_constexpr] = ACTIONS(7193),
    [anon_sym_volatile] = ACTIONS(7193),
    [anon_sym_restrict] = ACTIONS(7193),
    [anon_sym___restrict__] = ACTIONS(7193),
    [anon_sym__Atomic] = ACTIONS(7193),
    [anon_sym__Noreturn] = ACTIONS(7193),
    [anon_sym__Nonnull] = ACTIONS(7193),
    [anon_sym_nullable] = ACTIONS(7193),
    [anon_sym__Complex] = ACTIONS(7193),
    [anon_sym__Nullable] = ACTIONS(7193),
    [anon_sym__Nullable_result] = ACTIONS(7193),
    [anon_sym__Null_unspecified] = ACTIONS(7193),
    [anon_sym___autoreleasing] = ACTIONS(7193),
    [anon_sym___block] = ACTIONS(7193),
    [anon_sym___bridge] = ACTIONS(7193),
    [anon_sym___bridge_retained] = ACTIONS(7193),
    [anon_sym___bridge_transfer] = ACTIONS(7193),
    [anon_sym___complex] = ACTIONS(7193),
    [anon_sym___const] = ACTIONS(7193),
    [anon_sym___imag] = ACTIONS(7193),
    [anon_sym___kindof] = ACTIONS(7193),
    [anon_sym___nonnull] = ACTIONS(7193),
    [anon_sym___nullable] = ACTIONS(7193),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7193),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7193),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7193),
    [anon_sym___real] = ACTIONS(7193),
    [anon_sym___strong] = ACTIONS(7193),
    [anon_sym___unsafe_unretained] = ACTIONS(7193),
    [anon_sym___unused] = ACTIONS(7193),
    [anon_sym___weak] = ACTIONS(7193),
    [anon_sym_alignas] = ACTIONS(7193),
    [anon_sym__Alignas] = ACTIONS(7193),
    [sym_primitive_type] = ACTIONS(7193),
    [anon_sym_enum] = ACTIONS(7193),
    [anon_sym_COLON] = ACTIONS(7195),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7193),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7193),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7193),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7193),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7193),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7193),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7193),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7193),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7193),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7193),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7193),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7193),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7193),
    [anon_sym_API_AVAILABLE] = ACTIONS(7193),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7193),
    [anon_sym_API_DEPRECATED] = ACTIONS(7193),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7193),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7193),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7193),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7193),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7193),
    [anon_sym___deprecated_msg] = ACTIONS(7193),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7193),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7193),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7193),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7193),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7193),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7193),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7193),
  },
  [STATE(2941)] = {
    [sym_identifier] = ACTIONS(7197),
    [anon_sym_COMMA] = ACTIONS(7199),
    [anon_sym_RPAREN] = ACTIONS(7199),
    [anon_sym_LPAREN2] = ACTIONS(7199),
    [anon_sym_STAR] = ACTIONS(7199),
    [anon_sym_CARET] = ACTIONS(7199),
    [anon_sym_SEMI] = ACTIONS(7199),
    [anon_sym___extension__] = ACTIONS(7197),
    [anon_sym_extern] = ACTIONS(7197),
    [anon_sym___attribute__] = ACTIONS(7197),
    [anon_sym___attribute] = ACTIONS(7197),
    [anon_sym_noreturn] = ACTIONS(7197),
    [anon_sym_LBRACK] = ACTIONS(7199),
    [anon_sym___declspec] = ACTIONS(7197),
    [anon_sym___based] = ACTIONS(7197),
    [anon_sym_signed] = ACTIONS(7197),
    [anon_sym_unsigned] = ACTIONS(7197),
    [anon_sym_long] = ACTIONS(7197),
    [anon_sym_short] = ACTIONS(7197),
    [anon_sym_static] = ACTIONS(7197),
    [anon_sym_auto] = ACTIONS(7197),
    [anon_sym_register] = ACTIONS(7197),
    [anon_sym_inline] = ACTIONS(7197),
    [anon_sym___inline] = ACTIONS(7197),
    [anon_sym___inline__] = ACTIONS(7197),
    [anon_sym___forceinline] = ACTIONS(7197),
    [anon_sym_thread_local] = ACTIONS(7197),
    [anon_sym___thread] = ACTIONS(7197),
    [anon_sym_CG_EXTERN] = ACTIONS(7197),
    [anon_sym_CG_INLINE] = ACTIONS(7197),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7197),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7197),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7197),
    [anon_sym_IBOutlet] = ACTIONS(7197),
    [anon_sym_IBInspectable] = ACTIONS(7197),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7197),
    [anon_sym_NS_INLINE] = ACTIONS(7197),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7197),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7197),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7197),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7197),
    [anon_sym_const] = ACTIONS(7197),
    [anon_sym_constexpr] = ACTIONS(7197),
    [anon_sym_volatile] = ACTIONS(7197),
    [anon_sym_restrict] = ACTIONS(7197),
    [anon_sym___restrict__] = ACTIONS(7197),
    [anon_sym__Atomic] = ACTIONS(7197),
    [anon_sym__Noreturn] = ACTIONS(7197),
    [anon_sym__Nonnull] = ACTIONS(7197),
    [anon_sym_nullable] = ACTIONS(7197),
    [anon_sym__Complex] = ACTIONS(7197),
    [anon_sym__Nullable] = ACTIONS(7197),
    [anon_sym__Nullable_result] = ACTIONS(7197),
    [anon_sym__Null_unspecified] = ACTIONS(7197),
    [anon_sym___autoreleasing] = ACTIONS(7197),
    [anon_sym___block] = ACTIONS(7197),
    [anon_sym___bridge] = ACTIONS(7197),
    [anon_sym___bridge_retained] = ACTIONS(7197),
    [anon_sym___bridge_transfer] = ACTIONS(7197),
    [anon_sym___complex] = ACTIONS(7197),
    [anon_sym___const] = ACTIONS(7197),
    [anon_sym___imag] = ACTIONS(7197),
    [anon_sym___kindof] = ACTIONS(7197),
    [anon_sym___nonnull] = ACTIONS(7197),
    [anon_sym___nullable] = ACTIONS(7197),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7197),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7197),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7197),
    [anon_sym___real] = ACTIONS(7197),
    [anon_sym___strong] = ACTIONS(7197),
    [anon_sym___unsafe_unretained] = ACTIONS(7197),
    [anon_sym___unused] = ACTIONS(7197),
    [anon_sym___weak] = ACTIONS(7197),
    [anon_sym_alignas] = ACTIONS(7197),
    [anon_sym__Alignas] = ACTIONS(7197),
    [sym_primitive_type] = ACTIONS(7197),
    [anon_sym_enum] = ACTIONS(7197),
    [anon_sym_COLON] = ACTIONS(7199),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7197),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7197),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7197),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7197),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7197),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7197),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7197),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7197),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7197),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7197),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7197),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7197),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7197),
    [anon_sym_API_AVAILABLE] = ACTIONS(7197),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7197),
    [anon_sym_API_DEPRECATED] = ACTIONS(7197),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7197),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7197),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7197),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7197),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7197),
    [anon_sym___deprecated_msg] = ACTIONS(7197),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7197),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7197),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7197),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7197),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7197),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7197),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7197),
  },
  [STATE(2942)] = {
    [sym_identifier] = ACTIONS(7201),
    [anon_sym_COMMA] = ACTIONS(7203),
    [anon_sym_RPAREN] = ACTIONS(7203),
    [anon_sym_LPAREN2] = ACTIONS(7203),
    [anon_sym_STAR] = ACTIONS(7203),
    [anon_sym_CARET] = ACTIONS(7203),
    [anon_sym_SEMI] = ACTIONS(7203),
    [anon_sym___extension__] = ACTIONS(7201),
    [anon_sym_extern] = ACTIONS(7201),
    [anon_sym___attribute__] = ACTIONS(7201),
    [anon_sym___attribute] = ACTIONS(7201),
    [anon_sym_noreturn] = ACTIONS(7201),
    [anon_sym_LBRACK] = ACTIONS(7203),
    [anon_sym___declspec] = ACTIONS(7201),
    [anon_sym___based] = ACTIONS(7201),
    [anon_sym_signed] = ACTIONS(7201),
    [anon_sym_unsigned] = ACTIONS(7201),
    [anon_sym_long] = ACTIONS(7201),
    [anon_sym_short] = ACTIONS(7201),
    [anon_sym_static] = ACTIONS(7201),
    [anon_sym_auto] = ACTIONS(7201),
    [anon_sym_register] = ACTIONS(7201),
    [anon_sym_inline] = ACTIONS(7201),
    [anon_sym___inline] = ACTIONS(7201),
    [anon_sym___inline__] = ACTIONS(7201),
    [anon_sym___forceinline] = ACTIONS(7201),
    [anon_sym_thread_local] = ACTIONS(7201),
    [anon_sym___thread] = ACTIONS(7201),
    [anon_sym_CG_EXTERN] = ACTIONS(7201),
    [anon_sym_CG_INLINE] = ACTIONS(7201),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7201),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7201),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7201),
    [anon_sym_IBOutlet] = ACTIONS(7201),
    [anon_sym_IBInspectable] = ACTIONS(7201),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7201),
    [anon_sym_NS_INLINE] = ACTIONS(7201),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7201),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7201),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7201),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7201),
    [anon_sym_const] = ACTIONS(7201),
    [anon_sym_constexpr] = ACTIONS(7201),
    [anon_sym_volatile] = ACTIONS(7201),
    [anon_sym_restrict] = ACTIONS(7201),
    [anon_sym___restrict__] = ACTIONS(7201),
    [anon_sym__Atomic] = ACTIONS(7201),
    [anon_sym__Noreturn] = ACTIONS(7201),
    [anon_sym__Nonnull] = ACTIONS(7201),
    [anon_sym_nullable] = ACTIONS(7201),
    [anon_sym__Complex] = ACTIONS(7201),
    [anon_sym__Nullable] = ACTIONS(7201),
    [anon_sym__Nullable_result] = ACTIONS(7201),
    [anon_sym__Null_unspecified] = ACTIONS(7201),
    [anon_sym___autoreleasing] = ACTIONS(7201),
    [anon_sym___block] = ACTIONS(7201),
    [anon_sym___bridge] = ACTIONS(7201),
    [anon_sym___bridge_retained] = ACTIONS(7201),
    [anon_sym___bridge_transfer] = ACTIONS(7201),
    [anon_sym___complex] = ACTIONS(7201),
    [anon_sym___const] = ACTIONS(7201),
    [anon_sym___imag] = ACTIONS(7201),
    [anon_sym___kindof] = ACTIONS(7201),
    [anon_sym___nonnull] = ACTIONS(7201),
    [anon_sym___nullable] = ACTIONS(7201),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7201),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7201),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7201),
    [anon_sym___real] = ACTIONS(7201),
    [anon_sym___strong] = ACTIONS(7201),
    [anon_sym___unsafe_unretained] = ACTIONS(7201),
    [anon_sym___unused] = ACTIONS(7201),
    [anon_sym___weak] = ACTIONS(7201),
    [anon_sym_alignas] = ACTIONS(7201),
    [anon_sym__Alignas] = ACTIONS(7201),
    [sym_primitive_type] = ACTIONS(7201),
    [anon_sym_enum] = ACTIONS(7201),
    [anon_sym_COLON] = ACTIONS(7203),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7201),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7201),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7201),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7201),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7201),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7201),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7201),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7201),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7201),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7201),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7201),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7201),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7201),
    [anon_sym_API_AVAILABLE] = ACTIONS(7201),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7201),
    [anon_sym_API_DEPRECATED] = ACTIONS(7201),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7201),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7201),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7201),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7201),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7201),
    [anon_sym___deprecated_msg] = ACTIONS(7201),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7201),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7201),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7201),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7201),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7201),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7201),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7201),
  },
  [STATE(2943)] = {
    [sym_identifier] = ACTIONS(7205),
    [anon_sym_COMMA] = ACTIONS(7207),
    [anon_sym_RPAREN] = ACTIONS(7207),
    [anon_sym_LPAREN2] = ACTIONS(7207),
    [anon_sym_STAR] = ACTIONS(7207),
    [anon_sym_CARET] = ACTIONS(7207),
    [anon_sym_SEMI] = ACTIONS(7207),
    [anon_sym___extension__] = ACTIONS(7205),
    [anon_sym_extern] = ACTIONS(7205),
    [anon_sym___attribute__] = ACTIONS(7205),
    [anon_sym___attribute] = ACTIONS(7205),
    [anon_sym_noreturn] = ACTIONS(7205),
    [anon_sym_LBRACK] = ACTIONS(7207),
    [anon_sym___declspec] = ACTIONS(7205),
    [anon_sym___based] = ACTIONS(7205),
    [anon_sym_signed] = ACTIONS(7205),
    [anon_sym_unsigned] = ACTIONS(7205),
    [anon_sym_long] = ACTIONS(7205),
    [anon_sym_short] = ACTIONS(7205),
    [anon_sym_static] = ACTIONS(7205),
    [anon_sym_auto] = ACTIONS(7205),
    [anon_sym_register] = ACTIONS(7205),
    [anon_sym_inline] = ACTIONS(7205),
    [anon_sym___inline] = ACTIONS(7205),
    [anon_sym___inline__] = ACTIONS(7205),
    [anon_sym___forceinline] = ACTIONS(7205),
    [anon_sym_thread_local] = ACTIONS(7205),
    [anon_sym___thread] = ACTIONS(7205),
    [anon_sym_CG_EXTERN] = ACTIONS(7205),
    [anon_sym_CG_INLINE] = ACTIONS(7205),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7205),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7205),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7205),
    [anon_sym_IBOutlet] = ACTIONS(7205),
    [anon_sym_IBInspectable] = ACTIONS(7205),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7205),
    [anon_sym_NS_INLINE] = ACTIONS(7205),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7205),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7205),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7205),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7205),
    [anon_sym_const] = ACTIONS(7205),
    [anon_sym_constexpr] = ACTIONS(7205),
    [anon_sym_volatile] = ACTIONS(7205),
    [anon_sym_restrict] = ACTIONS(7205),
    [anon_sym___restrict__] = ACTIONS(7205),
    [anon_sym__Atomic] = ACTIONS(7205),
    [anon_sym__Noreturn] = ACTIONS(7205),
    [anon_sym__Nonnull] = ACTIONS(7205),
    [anon_sym_nullable] = ACTIONS(7205),
    [anon_sym__Complex] = ACTIONS(7205),
    [anon_sym__Nullable] = ACTIONS(7205),
    [anon_sym__Nullable_result] = ACTIONS(7205),
    [anon_sym__Null_unspecified] = ACTIONS(7205),
    [anon_sym___autoreleasing] = ACTIONS(7205),
    [anon_sym___block] = ACTIONS(7205),
    [anon_sym___bridge] = ACTIONS(7205),
    [anon_sym___bridge_retained] = ACTIONS(7205),
    [anon_sym___bridge_transfer] = ACTIONS(7205),
    [anon_sym___complex] = ACTIONS(7205),
    [anon_sym___const] = ACTIONS(7205),
    [anon_sym___imag] = ACTIONS(7205),
    [anon_sym___kindof] = ACTIONS(7205),
    [anon_sym___nonnull] = ACTIONS(7205),
    [anon_sym___nullable] = ACTIONS(7205),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7205),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7205),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7205),
    [anon_sym___real] = ACTIONS(7205),
    [anon_sym___strong] = ACTIONS(7205),
    [anon_sym___unsafe_unretained] = ACTIONS(7205),
    [anon_sym___unused] = ACTIONS(7205),
    [anon_sym___weak] = ACTIONS(7205),
    [anon_sym_alignas] = ACTIONS(7205),
    [anon_sym__Alignas] = ACTIONS(7205),
    [sym_primitive_type] = ACTIONS(7205),
    [anon_sym_enum] = ACTIONS(7205),
    [anon_sym_COLON] = ACTIONS(7207),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7205),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7205),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7205),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7205),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7205),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7205),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7205),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7205),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7205),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7205),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7205),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7205),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7205),
    [anon_sym_API_AVAILABLE] = ACTIONS(7205),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7205),
    [anon_sym_API_DEPRECATED] = ACTIONS(7205),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7205),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7205),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7205),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7205),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7205),
    [anon_sym___deprecated_msg] = ACTIONS(7205),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7205),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7205),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7205),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7205),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7205),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7205),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7205),
  },
  [STATE(2944)] = {
    [sym_identifier] = ACTIONS(7209),
    [anon_sym_COMMA] = ACTIONS(7211),
    [anon_sym_RPAREN] = ACTIONS(7211),
    [anon_sym_LPAREN2] = ACTIONS(7211),
    [anon_sym_STAR] = ACTIONS(7211),
    [anon_sym_CARET] = ACTIONS(7211),
    [anon_sym_SEMI] = ACTIONS(7211),
    [anon_sym___extension__] = ACTIONS(7209),
    [anon_sym_extern] = ACTIONS(7209),
    [anon_sym___attribute__] = ACTIONS(7209),
    [anon_sym___attribute] = ACTIONS(7209),
    [anon_sym_noreturn] = ACTIONS(7209),
    [anon_sym_LBRACK] = ACTIONS(7211),
    [anon_sym___declspec] = ACTIONS(7209),
    [anon_sym___based] = ACTIONS(7209),
    [anon_sym_signed] = ACTIONS(7209),
    [anon_sym_unsigned] = ACTIONS(7209),
    [anon_sym_long] = ACTIONS(7209),
    [anon_sym_short] = ACTIONS(7209),
    [anon_sym_static] = ACTIONS(7209),
    [anon_sym_auto] = ACTIONS(7209),
    [anon_sym_register] = ACTIONS(7209),
    [anon_sym_inline] = ACTIONS(7209),
    [anon_sym___inline] = ACTIONS(7209),
    [anon_sym___inline__] = ACTIONS(7209),
    [anon_sym___forceinline] = ACTIONS(7209),
    [anon_sym_thread_local] = ACTIONS(7209),
    [anon_sym___thread] = ACTIONS(7209),
    [anon_sym_CG_EXTERN] = ACTIONS(7209),
    [anon_sym_CG_INLINE] = ACTIONS(7209),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7209),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7209),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7209),
    [anon_sym_IBOutlet] = ACTIONS(7209),
    [anon_sym_IBInspectable] = ACTIONS(7209),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7209),
    [anon_sym_NS_INLINE] = ACTIONS(7209),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7209),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7209),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7209),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7209),
    [anon_sym_const] = ACTIONS(7209),
    [anon_sym_constexpr] = ACTIONS(7209),
    [anon_sym_volatile] = ACTIONS(7209),
    [anon_sym_restrict] = ACTIONS(7209),
    [anon_sym___restrict__] = ACTIONS(7209),
    [anon_sym__Atomic] = ACTIONS(7209),
    [anon_sym__Noreturn] = ACTIONS(7209),
    [anon_sym__Nonnull] = ACTIONS(7209),
    [anon_sym_nullable] = ACTIONS(7209),
    [anon_sym__Complex] = ACTIONS(7209),
    [anon_sym__Nullable] = ACTIONS(7209),
    [anon_sym__Nullable_result] = ACTIONS(7209),
    [anon_sym__Null_unspecified] = ACTIONS(7209),
    [anon_sym___autoreleasing] = ACTIONS(7209),
    [anon_sym___block] = ACTIONS(7209),
    [anon_sym___bridge] = ACTIONS(7209),
    [anon_sym___bridge_retained] = ACTIONS(7209),
    [anon_sym___bridge_transfer] = ACTIONS(7209),
    [anon_sym___complex] = ACTIONS(7209),
    [anon_sym___const] = ACTIONS(7209),
    [anon_sym___imag] = ACTIONS(7209),
    [anon_sym___kindof] = ACTIONS(7209),
    [anon_sym___nonnull] = ACTIONS(7209),
    [anon_sym___nullable] = ACTIONS(7209),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7209),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7209),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7209),
    [anon_sym___real] = ACTIONS(7209),
    [anon_sym___strong] = ACTIONS(7209),
    [anon_sym___unsafe_unretained] = ACTIONS(7209),
    [anon_sym___unused] = ACTIONS(7209),
    [anon_sym___weak] = ACTIONS(7209),
    [anon_sym_alignas] = ACTIONS(7209),
    [anon_sym__Alignas] = ACTIONS(7209),
    [sym_primitive_type] = ACTIONS(7209),
    [anon_sym_enum] = ACTIONS(7209),
    [anon_sym_COLON] = ACTIONS(7211),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7209),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7209),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7209),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7209),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7209),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7209),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7209),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7209),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7209),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7209),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7209),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7209),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7209),
    [anon_sym_API_AVAILABLE] = ACTIONS(7209),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7209),
    [anon_sym_API_DEPRECATED] = ACTIONS(7209),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7209),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7209),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7209),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7209),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7209),
    [anon_sym___deprecated_msg] = ACTIONS(7209),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7209),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7209),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7209),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7209),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7209),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7209),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7209),
  },
  [STATE(2945)] = {
    [sym_identifier] = ACTIONS(7213),
    [anon_sym_COMMA] = ACTIONS(7215),
    [anon_sym_RPAREN] = ACTIONS(7215),
    [anon_sym_LPAREN2] = ACTIONS(7215),
    [anon_sym_STAR] = ACTIONS(7215),
    [anon_sym_CARET] = ACTIONS(7215),
    [anon_sym_SEMI] = ACTIONS(7215),
    [anon_sym___extension__] = ACTIONS(7213),
    [anon_sym_extern] = ACTIONS(7213),
    [anon_sym___attribute__] = ACTIONS(7213),
    [anon_sym___attribute] = ACTIONS(7213),
    [anon_sym_noreturn] = ACTIONS(7213),
    [anon_sym_LBRACK] = ACTIONS(7215),
    [anon_sym___declspec] = ACTIONS(7213),
    [anon_sym___based] = ACTIONS(7213),
    [anon_sym_signed] = ACTIONS(7213),
    [anon_sym_unsigned] = ACTIONS(7213),
    [anon_sym_long] = ACTIONS(7213),
    [anon_sym_short] = ACTIONS(7213),
    [anon_sym_static] = ACTIONS(7213),
    [anon_sym_auto] = ACTIONS(7213),
    [anon_sym_register] = ACTIONS(7213),
    [anon_sym_inline] = ACTIONS(7213),
    [anon_sym___inline] = ACTIONS(7213),
    [anon_sym___inline__] = ACTIONS(7213),
    [anon_sym___forceinline] = ACTIONS(7213),
    [anon_sym_thread_local] = ACTIONS(7213),
    [anon_sym___thread] = ACTIONS(7213),
    [anon_sym_CG_EXTERN] = ACTIONS(7213),
    [anon_sym_CG_INLINE] = ACTIONS(7213),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7213),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7213),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7213),
    [anon_sym_IBOutlet] = ACTIONS(7213),
    [anon_sym_IBInspectable] = ACTIONS(7213),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7213),
    [anon_sym_NS_INLINE] = ACTIONS(7213),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7213),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7213),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7213),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7213),
    [anon_sym_const] = ACTIONS(7213),
    [anon_sym_constexpr] = ACTIONS(7213),
    [anon_sym_volatile] = ACTIONS(7213),
    [anon_sym_restrict] = ACTIONS(7213),
    [anon_sym___restrict__] = ACTIONS(7213),
    [anon_sym__Atomic] = ACTIONS(7213),
    [anon_sym__Noreturn] = ACTIONS(7213),
    [anon_sym__Nonnull] = ACTIONS(7213),
    [anon_sym_nullable] = ACTIONS(7213),
    [anon_sym__Complex] = ACTIONS(7213),
    [anon_sym__Nullable] = ACTIONS(7213),
    [anon_sym__Nullable_result] = ACTIONS(7213),
    [anon_sym__Null_unspecified] = ACTIONS(7213),
    [anon_sym___autoreleasing] = ACTIONS(7213),
    [anon_sym___block] = ACTIONS(7213),
    [anon_sym___bridge] = ACTIONS(7213),
    [anon_sym___bridge_retained] = ACTIONS(7213),
    [anon_sym___bridge_transfer] = ACTIONS(7213),
    [anon_sym___complex] = ACTIONS(7213),
    [anon_sym___const] = ACTIONS(7213),
    [anon_sym___imag] = ACTIONS(7213),
    [anon_sym___kindof] = ACTIONS(7213),
    [anon_sym___nonnull] = ACTIONS(7213),
    [anon_sym___nullable] = ACTIONS(7213),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7213),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7213),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7213),
    [anon_sym___real] = ACTIONS(7213),
    [anon_sym___strong] = ACTIONS(7213),
    [anon_sym___unsafe_unretained] = ACTIONS(7213),
    [anon_sym___unused] = ACTIONS(7213),
    [anon_sym___weak] = ACTIONS(7213),
    [anon_sym_alignas] = ACTIONS(7213),
    [anon_sym__Alignas] = ACTIONS(7213),
    [sym_primitive_type] = ACTIONS(7213),
    [anon_sym_enum] = ACTIONS(7213),
    [anon_sym_COLON] = ACTIONS(7215),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7213),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7213),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7213),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7213),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7213),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7213),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7213),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7213),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7213),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7213),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7213),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7213),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7213),
    [anon_sym_API_AVAILABLE] = ACTIONS(7213),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7213),
    [anon_sym_API_DEPRECATED] = ACTIONS(7213),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7213),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7213),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7213),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7213),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7213),
    [anon_sym___deprecated_msg] = ACTIONS(7213),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7213),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7213),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7213),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7213),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7213),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7213),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7213),
  },
  [STATE(2946)] = {
    [sym_identifier] = ACTIONS(7217),
    [anon_sym_COMMA] = ACTIONS(7219),
    [anon_sym_RPAREN] = ACTIONS(7219),
    [anon_sym_LPAREN2] = ACTIONS(7219),
    [anon_sym_STAR] = ACTIONS(7219),
    [anon_sym_CARET] = ACTIONS(7219),
    [anon_sym_SEMI] = ACTIONS(7219),
    [anon_sym___extension__] = ACTIONS(7217),
    [anon_sym_extern] = ACTIONS(7217),
    [anon_sym___attribute__] = ACTIONS(7217),
    [anon_sym___attribute] = ACTIONS(7217),
    [anon_sym_noreturn] = ACTIONS(7217),
    [anon_sym_LBRACK] = ACTIONS(7219),
    [anon_sym___declspec] = ACTIONS(7217),
    [anon_sym___based] = ACTIONS(7217),
    [anon_sym_signed] = ACTIONS(7217),
    [anon_sym_unsigned] = ACTIONS(7217),
    [anon_sym_long] = ACTIONS(7217),
    [anon_sym_short] = ACTIONS(7217),
    [anon_sym_static] = ACTIONS(7217),
    [anon_sym_auto] = ACTIONS(7217),
    [anon_sym_register] = ACTIONS(7217),
    [anon_sym_inline] = ACTIONS(7217),
    [anon_sym___inline] = ACTIONS(7217),
    [anon_sym___inline__] = ACTIONS(7217),
    [anon_sym___forceinline] = ACTIONS(7217),
    [anon_sym_thread_local] = ACTIONS(7217),
    [anon_sym___thread] = ACTIONS(7217),
    [anon_sym_CG_EXTERN] = ACTIONS(7217),
    [anon_sym_CG_INLINE] = ACTIONS(7217),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7217),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7217),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7217),
    [anon_sym_IBOutlet] = ACTIONS(7217),
    [anon_sym_IBInspectable] = ACTIONS(7217),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7217),
    [anon_sym_NS_INLINE] = ACTIONS(7217),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7217),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7217),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7217),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7217),
    [anon_sym_const] = ACTIONS(7217),
    [anon_sym_constexpr] = ACTIONS(7217),
    [anon_sym_volatile] = ACTIONS(7217),
    [anon_sym_restrict] = ACTIONS(7217),
    [anon_sym___restrict__] = ACTIONS(7217),
    [anon_sym__Atomic] = ACTIONS(7217),
    [anon_sym__Noreturn] = ACTIONS(7217),
    [anon_sym__Nonnull] = ACTIONS(7217),
    [anon_sym_nullable] = ACTIONS(7217),
    [anon_sym__Complex] = ACTIONS(7217),
    [anon_sym__Nullable] = ACTIONS(7217),
    [anon_sym__Nullable_result] = ACTIONS(7217),
    [anon_sym__Null_unspecified] = ACTIONS(7217),
    [anon_sym___autoreleasing] = ACTIONS(7217),
    [anon_sym___block] = ACTIONS(7217),
    [anon_sym___bridge] = ACTIONS(7217),
    [anon_sym___bridge_retained] = ACTIONS(7217),
    [anon_sym___bridge_transfer] = ACTIONS(7217),
    [anon_sym___complex] = ACTIONS(7217),
    [anon_sym___const] = ACTIONS(7217),
    [anon_sym___imag] = ACTIONS(7217),
    [anon_sym___kindof] = ACTIONS(7217),
    [anon_sym___nonnull] = ACTIONS(7217),
    [anon_sym___nullable] = ACTIONS(7217),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7217),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7217),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7217),
    [anon_sym___real] = ACTIONS(7217),
    [anon_sym___strong] = ACTIONS(7217),
    [anon_sym___unsafe_unretained] = ACTIONS(7217),
    [anon_sym___unused] = ACTIONS(7217),
    [anon_sym___weak] = ACTIONS(7217),
    [anon_sym_alignas] = ACTIONS(7217),
    [anon_sym__Alignas] = ACTIONS(7217),
    [sym_primitive_type] = ACTIONS(7217),
    [anon_sym_enum] = ACTIONS(7217),
    [anon_sym_COLON] = ACTIONS(7219),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7217),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7217),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7217),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7217),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7217),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7217),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7217),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7217),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7217),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7217),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7217),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7217),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7217),
    [anon_sym_API_AVAILABLE] = ACTIONS(7217),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7217),
    [anon_sym_API_DEPRECATED] = ACTIONS(7217),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7217),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7217),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7217),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7217),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7217),
    [anon_sym___deprecated_msg] = ACTIONS(7217),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7217),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7217),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7217),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7217),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7217),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7217),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7217),
  },
  [STATE(2947)] = {
    [sym_identifier] = ACTIONS(7221),
    [anon_sym_COMMA] = ACTIONS(7223),
    [anon_sym_RPAREN] = ACTIONS(7223),
    [anon_sym_LPAREN2] = ACTIONS(7223),
    [anon_sym_STAR] = ACTIONS(7223),
    [anon_sym_CARET] = ACTIONS(7223),
    [anon_sym_SEMI] = ACTIONS(7223),
    [anon_sym___extension__] = ACTIONS(7221),
    [anon_sym_extern] = ACTIONS(7221),
    [anon_sym___attribute__] = ACTIONS(7221),
    [anon_sym___attribute] = ACTIONS(7221),
    [anon_sym_noreturn] = ACTIONS(7221),
    [anon_sym_LBRACK] = ACTIONS(7223),
    [anon_sym___declspec] = ACTIONS(7221),
    [anon_sym___based] = ACTIONS(7221),
    [anon_sym_signed] = ACTIONS(7221),
    [anon_sym_unsigned] = ACTIONS(7221),
    [anon_sym_long] = ACTIONS(7221),
    [anon_sym_short] = ACTIONS(7221),
    [anon_sym_static] = ACTIONS(7221),
    [anon_sym_auto] = ACTIONS(7221),
    [anon_sym_register] = ACTIONS(7221),
    [anon_sym_inline] = ACTIONS(7221),
    [anon_sym___inline] = ACTIONS(7221),
    [anon_sym___inline__] = ACTIONS(7221),
    [anon_sym___forceinline] = ACTIONS(7221),
    [anon_sym_thread_local] = ACTIONS(7221),
    [anon_sym___thread] = ACTIONS(7221),
    [anon_sym_CG_EXTERN] = ACTIONS(7221),
    [anon_sym_CG_INLINE] = ACTIONS(7221),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7221),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7221),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7221),
    [anon_sym_IBOutlet] = ACTIONS(7221),
    [anon_sym_IBInspectable] = ACTIONS(7221),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7221),
    [anon_sym_NS_INLINE] = ACTIONS(7221),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7221),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7221),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7221),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7221),
    [anon_sym_const] = ACTIONS(7221),
    [anon_sym_constexpr] = ACTIONS(7221),
    [anon_sym_volatile] = ACTIONS(7221),
    [anon_sym_restrict] = ACTIONS(7221),
    [anon_sym___restrict__] = ACTIONS(7221),
    [anon_sym__Atomic] = ACTIONS(7221),
    [anon_sym__Noreturn] = ACTIONS(7221),
    [anon_sym__Nonnull] = ACTIONS(7221),
    [anon_sym_nullable] = ACTIONS(7221),
    [anon_sym__Complex] = ACTIONS(7221),
    [anon_sym__Nullable] = ACTIONS(7221),
    [anon_sym__Nullable_result] = ACTIONS(7221),
    [anon_sym__Null_unspecified] = ACTIONS(7221),
    [anon_sym___autoreleasing] = ACTIONS(7221),
    [anon_sym___block] = ACTIONS(7221),
    [anon_sym___bridge] = ACTIONS(7221),
    [anon_sym___bridge_retained] = ACTIONS(7221),
    [anon_sym___bridge_transfer] = ACTIONS(7221),
    [anon_sym___complex] = ACTIONS(7221),
    [anon_sym___const] = ACTIONS(7221),
    [anon_sym___imag] = ACTIONS(7221),
    [anon_sym___kindof] = ACTIONS(7221),
    [anon_sym___nonnull] = ACTIONS(7221),
    [anon_sym___nullable] = ACTIONS(7221),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7221),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7221),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7221),
    [anon_sym___real] = ACTIONS(7221),
    [anon_sym___strong] = ACTIONS(7221),
    [anon_sym___unsafe_unretained] = ACTIONS(7221),
    [anon_sym___unused] = ACTIONS(7221),
    [anon_sym___weak] = ACTIONS(7221),
    [anon_sym_alignas] = ACTIONS(7221),
    [anon_sym__Alignas] = ACTIONS(7221),
    [sym_primitive_type] = ACTIONS(7221),
    [anon_sym_enum] = ACTIONS(7221),
    [anon_sym_COLON] = ACTIONS(7223),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7221),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7221),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7221),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7221),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7221),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7221),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7221),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7221),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7221),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7221),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7221),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7221),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7221),
    [anon_sym_API_AVAILABLE] = ACTIONS(7221),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7221),
    [anon_sym_API_DEPRECATED] = ACTIONS(7221),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7221),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7221),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7221),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7221),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7221),
    [anon_sym___deprecated_msg] = ACTIONS(7221),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7221),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7221),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7221),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7221),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7221),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7221),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7221),
  },
  [STATE(2948)] = {
    [sym_identifier] = ACTIONS(7225),
    [anon_sym_COMMA] = ACTIONS(7227),
    [anon_sym_RPAREN] = ACTIONS(7227),
    [anon_sym_LPAREN2] = ACTIONS(7227),
    [anon_sym_STAR] = ACTIONS(7227),
    [anon_sym_CARET] = ACTIONS(7227),
    [anon_sym_SEMI] = ACTIONS(7227),
    [anon_sym___extension__] = ACTIONS(7225),
    [anon_sym_extern] = ACTIONS(7225),
    [anon_sym___attribute__] = ACTIONS(7225),
    [anon_sym___attribute] = ACTIONS(7225),
    [anon_sym_noreturn] = ACTIONS(7225),
    [anon_sym_LBRACK] = ACTIONS(7227),
    [anon_sym___declspec] = ACTIONS(7225),
    [anon_sym___based] = ACTIONS(7225),
    [anon_sym_signed] = ACTIONS(7225),
    [anon_sym_unsigned] = ACTIONS(7225),
    [anon_sym_long] = ACTIONS(7225),
    [anon_sym_short] = ACTIONS(7225),
    [anon_sym_static] = ACTIONS(7225),
    [anon_sym_auto] = ACTIONS(7225),
    [anon_sym_register] = ACTIONS(7225),
    [anon_sym_inline] = ACTIONS(7225),
    [anon_sym___inline] = ACTIONS(7225),
    [anon_sym___inline__] = ACTIONS(7225),
    [anon_sym___forceinline] = ACTIONS(7225),
    [anon_sym_thread_local] = ACTIONS(7225),
    [anon_sym___thread] = ACTIONS(7225),
    [anon_sym_CG_EXTERN] = ACTIONS(7225),
    [anon_sym_CG_INLINE] = ACTIONS(7225),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7225),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7225),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7225),
    [anon_sym_IBOutlet] = ACTIONS(7225),
    [anon_sym_IBInspectable] = ACTIONS(7225),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7225),
    [anon_sym_NS_INLINE] = ACTIONS(7225),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7225),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7225),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7225),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7225),
    [anon_sym_const] = ACTIONS(7225),
    [anon_sym_constexpr] = ACTIONS(7225),
    [anon_sym_volatile] = ACTIONS(7225),
    [anon_sym_restrict] = ACTIONS(7225),
    [anon_sym___restrict__] = ACTIONS(7225),
    [anon_sym__Atomic] = ACTIONS(7225),
    [anon_sym__Noreturn] = ACTIONS(7225),
    [anon_sym__Nonnull] = ACTIONS(7225),
    [anon_sym_nullable] = ACTIONS(7225),
    [anon_sym__Complex] = ACTIONS(7225),
    [anon_sym__Nullable] = ACTIONS(7225),
    [anon_sym__Nullable_result] = ACTIONS(7225),
    [anon_sym__Null_unspecified] = ACTIONS(7225),
    [anon_sym___autoreleasing] = ACTIONS(7225),
    [anon_sym___block] = ACTIONS(7225),
    [anon_sym___bridge] = ACTIONS(7225),
    [anon_sym___bridge_retained] = ACTIONS(7225),
    [anon_sym___bridge_transfer] = ACTIONS(7225),
    [anon_sym___complex] = ACTIONS(7225),
    [anon_sym___const] = ACTIONS(7225),
    [anon_sym___imag] = ACTIONS(7225),
    [anon_sym___kindof] = ACTIONS(7225),
    [anon_sym___nonnull] = ACTIONS(7225),
    [anon_sym___nullable] = ACTIONS(7225),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7225),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7225),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7225),
    [anon_sym___real] = ACTIONS(7225),
    [anon_sym___strong] = ACTIONS(7225),
    [anon_sym___unsafe_unretained] = ACTIONS(7225),
    [anon_sym___unused] = ACTIONS(7225),
    [anon_sym___weak] = ACTIONS(7225),
    [anon_sym_alignas] = ACTIONS(7225),
    [anon_sym__Alignas] = ACTIONS(7225),
    [sym_primitive_type] = ACTIONS(7225),
    [anon_sym_enum] = ACTIONS(7225),
    [anon_sym_COLON] = ACTIONS(7227),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7225),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7225),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7225),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7225),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7225),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7225),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7225),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7225),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7225),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7225),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7225),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7225),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7225),
    [anon_sym_API_AVAILABLE] = ACTIONS(7225),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7225),
    [anon_sym_API_DEPRECATED] = ACTIONS(7225),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7225),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7225),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7225),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7225),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7225),
    [anon_sym___deprecated_msg] = ACTIONS(7225),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7225),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7225),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7225),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7225),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7225),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7225),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7225),
  },
  [STATE(2949)] = {
    [sym_identifier] = ACTIONS(7229),
    [anon_sym_COMMA] = ACTIONS(7231),
    [anon_sym_RPAREN] = ACTIONS(7231),
    [anon_sym_LPAREN2] = ACTIONS(7231),
    [anon_sym_STAR] = ACTIONS(7231),
    [anon_sym_CARET] = ACTIONS(7231),
    [anon_sym_SEMI] = ACTIONS(7231),
    [anon_sym___extension__] = ACTIONS(7229),
    [anon_sym_extern] = ACTIONS(7229),
    [anon_sym___attribute__] = ACTIONS(7229),
    [anon_sym___attribute] = ACTIONS(7229),
    [anon_sym_noreturn] = ACTIONS(7229),
    [anon_sym_LBRACK] = ACTIONS(7231),
    [anon_sym___declspec] = ACTIONS(7229),
    [anon_sym___based] = ACTIONS(7229),
    [anon_sym_signed] = ACTIONS(7229),
    [anon_sym_unsigned] = ACTIONS(7229),
    [anon_sym_long] = ACTIONS(7229),
    [anon_sym_short] = ACTIONS(7229),
    [anon_sym_static] = ACTIONS(7229),
    [anon_sym_auto] = ACTIONS(7229),
    [anon_sym_register] = ACTIONS(7229),
    [anon_sym_inline] = ACTIONS(7229),
    [anon_sym___inline] = ACTIONS(7229),
    [anon_sym___inline__] = ACTIONS(7229),
    [anon_sym___forceinline] = ACTIONS(7229),
    [anon_sym_thread_local] = ACTIONS(7229),
    [anon_sym___thread] = ACTIONS(7229),
    [anon_sym_CG_EXTERN] = ACTIONS(7229),
    [anon_sym_CG_INLINE] = ACTIONS(7229),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7229),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7229),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7229),
    [anon_sym_IBOutlet] = ACTIONS(7229),
    [anon_sym_IBInspectable] = ACTIONS(7229),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7229),
    [anon_sym_NS_INLINE] = ACTIONS(7229),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7229),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7229),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7229),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7229),
    [anon_sym_const] = ACTIONS(7229),
    [anon_sym_constexpr] = ACTIONS(7229),
    [anon_sym_volatile] = ACTIONS(7229),
    [anon_sym_restrict] = ACTIONS(7229),
    [anon_sym___restrict__] = ACTIONS(7229),
    [anon_sym__Atomic] = ACTIONS(7229),
    [anon_sym__Noreturn] = ACTIONS(7229),
    [anon_sym__Nonnull] = ACTIONS(7229),
    [anon_sym_nullable] = ACTIONS(7229),
    [anon_sym__Complex] = ACTIONS(7229),
    [anon_sym__Nullable] = ACTIONS(7229),
    [anon_sym__Nullable_result] = ACTIONS(7229),
    [anon_sym__Null_unspecified] = ACTIONS(7229),
    [anon_sym___autoreleasing] = ACTIONS(7229),
    [anon_sym___block] = ACTIONS(7229),
    [anon_sym___bridge] = ACTIONS(7229),
    [anon_sym___bridge_retained] = ACTIONS(7229),
    [anon_sym___bridge_transfer] = ACTIONS(7229),
    [anon_sym___complex] = ACTIONS(7229),
    [anon_sym___const] = ACTIONS(7229),
    [anon_sym___imag] = ACTIONS(7229),
    [anon_sym___kindof] = ACTIONS(7229),
    [anon_sym___nonnull] = ACTIONS(7229),
    [anon_sym___nullable] = ACTIONS(7229),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7229),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7229),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7229),
    [anon_sym___real] = ACTIONS(7229),
    [anon_sym___strong] = ACTIONS(7229),
    [anon_sym___unsafe_unretained] = ACTIONS(7229),
    [anon_sym___unused] = ACTIONS(7229),
    [anon_sym___weak] = ACTIONS(7229),
    [anon_sym_alignas] = ACTIONS(7229),
    [anon_sym__Alignas] = ACTIONS(7229),
    [sym_primitive_type] = ACTIONS(7229),
    [anon_sym_enum] = ACTIONS(7229),
    [anon_sym_COLON] = ACTIONS(7231),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7229),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7229),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7229),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7229),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7229),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7229),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7229),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7229),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7229),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7229),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7229),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7229),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7229),
    [anon_sym_API_AVAILABLE] = ACTIONS(7229),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7229),
    [anon_sym_API_DEPRECATED] = ACTIONS(7229),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7229),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7229),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7229),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7229),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7229),
    [anon_sym___deprecated_msg] = ACTIONS(7229),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7229),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7229),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7229),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7229),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7229),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7229),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7229),
  },
  [STATE(2950)] = {
    [sym_identifier] = ACTIONS(7233),
    [anon_sym_COMMA] = ACTIONS(7235),
    [anon_sym_RPAREN] = ACTIONS(7235),
    [anon_sym_LPAREN2] = ACTIONS(7235),
    [anon_sym_STAR] = ACTIONS(7235),
    [anon_sym_CARET] = ACTIONS(7235),
    [anon_sym_SEMI] = ACTIONS(7235),
    [anon_sym___extension__] = ACTIONS(7233),
    [anon_sym_extern] = ACTIONS(7233),
    [anon_sym___attribute__] = ACTIONS(7233),
    [anon_sym___attribute] = ACTIONS(7233),
    [anon_sym_noreturn] = ACTIONS(7233),
    [anon_sym_LBRACK] = ACTIONS(7235),
    [anon_sym___declspec] = ACTIONS(7233),
    [anon_sym___based] = ACTIONS(7233),
    [anon_sym_signed] = ACTIONS(7233),
    [anon_sym_unsigned] = ACTIONS(7233),
    [anon_sym_long] = ACTIONS(7233),
    [anon_sym_short] = ACTIONS(7233),
    [anon_sym_static] = ACTIONS(7233),
    [anon_sym_auto] = ACTIONS(7233),
    [anon_sym_register] = ACTIONS(7233),
    [anon_sym_inline] = ACTIONS(7233),
    [anon_sym___inline] = ACTIONS(7233),
    [anon_sym___inline__] = ACTIONS(7233),
    [anon_sym___forceinline] = ACTIONS(7233),
    [anon_sym_thread_local] = ACTIONS(7233),
    [anon_sym___thread] = ACTIONS(7233),
    [anon_sym_CG_EXTERN] = ACTIONS(7233),
    [anon_sym_CG_INLINE] = ACTIONS(7233),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7233),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7233),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7233),
    [anon_sym_IBOutlet] = ACTIONS(7233),
    [anon_sym_IBInspectable] = ACTIONS(7233),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7233),
    [anon_sym_NS_INLINE] = ACTIONS(7233),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7233),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7233),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7233),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7233),
    [anon_sym_const] = ACTIONS(7233),
    [anon_sym_constexpr] = ACTIONS(7233),
    [anon_sym_volatile] = ACTIONS(7233),
    [anon_sym_restrict] = ACTIONS(7233),
    [anon_sym___restrict__] = ACTIONS(7233),
    [anon_sym__Atomic] = ACTIONS(7233),
    [anon_sym__Noreturn] = ACTIONS(7233),
    [anon_sym__Nonnull] = ACTIONS(7233),
    [anon_sym_nullable] = ACTIONS(7233),
    [anon_sym__Complex] = ACTIONS(7233),
    [anon_sym__Nullable] = ACTIONS(7233),
    [anon_sym__Nullable_result] = ACTIONS(7233),
    [anon_sym__Null_unspecified] = ACTIONS(7233),
    [anon_sym___autoreleasing] = ACTIONS(7233),
    [anon_sym___block] = ACTIONS(7233),
    [anon_sym___bridge] = ACTIONS(7233),
    [anon_sym___bridge_retained] = ACTIONS(7233),
    [anon_sym___bridge_transfer] = ACTIONS(7233),
    [anon_sym___complex] = ACTIONS(7233),
    [anon_sym___const] = ACTIONS(7233),
    [anon_sym___imag] = ACTIONS(7233),
    [anon_sym___kindof] = ACTIONS(7233),
    [anon_sym___nonnull] = ACTIONS(7233),
    [anon_sym___nullable] = ACTIONS(7233),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7233),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7233),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7233),
    [anon_sym___real] = ACTIONS(7233),
    [anon_sym___strong] = ACTIONS(7233),
    [anon_sym___unsafe_unretained] = ACTIONS(7233),
    [anon_sym___unused] = ACTIONS(7233),
    [anon_sym___weak] = ACTIONS(7233),
    [anon_sym_alignas] = ACTIONS(7233),
    [anon_sym__Alignas] = ACTIONS(7233),
    [sym_primitive_type] = ACTIONS(7233),
    [anon_sym_enum] = ACTIONS(7233),
    [anon_sym_COLON] = ACTIONS(7235),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7233),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7233),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7233),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7233),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7233),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7233),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7233),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7233),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7233),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7233),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7233),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7233),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7233),
    [anon_sym_API_AVAILABLE] = ACTIONS(7233),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7233),
    [anon_sym_API_DEPRECATED] = ACTIONS(7233),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7233),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7233),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7233),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7233),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7233),
    [anon_sym___deprecated_msg] = ACTIONS(7233),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7233),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7233),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7233),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7233),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7233),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7233),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7233),
  },
  [STATE(2951)] = {
    [sym_identifier] = ACTIONS(7237),
    [anon_sym_COMMA] = ACTIONS(7239),
    [anon_sym_RPAREN] = ACTIONS(7239),
    [anon_sym_LPAREN2] = ACTIONS(7239),
    [anon_sym_STAR] = ACTIONS(7239),
    [anon_sym_CARET] = ACTIONS(7239),
    [anon_sym_SEMI] = ACTIONS(7239),
    [anon_sym___extension__] = ACTIONS(7237),
    [anon_sym_extern] = ACTIONS(7237),
    [anon_sym___attribute__] = ACTIONS(7237),
    [anon_sym___attribute] = ACTIONS(7237),
    [anon_sym_noreturn] = ACTIONS(7237),
    [anon_sym_LBRACK] = ACTIONS(7239),
    [anon_sym___declspec] = ACTIONS(7237),
    [anon_sym___based] = ACTIONS(7237),
    [anon_sym_signed] = ACTIONS(7237),
    [anon_sym_unsigned] = ACTIONS(7237),
    [anon_sym_long] = ACTIONS(7237),
    [anon_sym_short] = ACTIONS(7237),
    [anon_sym_static] = ACTIONS(7237),
    [anon_sym_auto] = ACTIONS(7237),
    [anon_sym_register] = ACTIONS(7237),
    [anon_sym_inline] = ACTIONS(7237),
    [anon_sym___inline] = ACTIONS(7237),
    [anon_sym___inline__] = ACTIONS(7237),
    [anon_sym___forceinline] = ACTIONS(7237),
    [anon_sym_thread_local] = ACTIONS(7237),
    [anon_sym___thread] = ACTIONS(7237),
    [anon_sym_CG_EXTERN] = ACTIONS(7237),
    [anon_sym_CG_INLINE] = ACTIONS(7237),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7237),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7237),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7237),
    [anon_sym_IBOutlet] = ACTIONS(7237),
    [anon_sym_IBInspectable] = ACTIONS(7237),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7237),
    [anon_sym_NS_INLINE] = ACTIONS(7237),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7237),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7237),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7237),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7237),
    [anon_sym_const] = ACTIONS(7237),
    [anon_sym_constexpr] = ACTIONS(7237),
    [anon_sym_volatile] = ACTIONS(7237),
    [anon_sym_restrict] = ACTIONS(7237),
    [anon_sym___restrict__] = ACTIONS(7237),
    [anon_sym__Atomic] = ACTIONS(7237),
    [anon_sym__Noreturn] = ACTIONS(7237),
    [anon_sym__Nonnull] = ACTIONS(7237),
    [anon_sym_nullable] = ACTIONS(7237),
    [anon_sym__Complex] = ACTIONS(7237),
    [anon_sym__Nullable] = ACTIONS(7237),
    [anon_sym__Nullable_result] = ACTIONS(7237),
    [anon_sym__Null_unspecified] = ACTIONS(7237),
    [anon_sym___autoreleasing] = ACTIONS(7237),
    [anon_sym___block] = ACTIONS(7237),
    [anon_sym___bridge] = ACTIONS(7237),
    [anon_sym___bridge_retained] = ACTIONS(7237),
    [anon_sym___bridge_transfer] = ACTIONS(7237),
    [anon_sym___complex] = ACTIONS(7237),
    [anon_sym___const] = ACTIONS(7237),
    [anon_sym___imag] = ACTIONS(7237),
    [anon_sym___kindof] = ACTIONS(7237),
    [anon_sym___nonnull] = ACTIONS(7237),
    [anon_sym___nullable] = ACTIONS(7237),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7237),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7237),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7237),
    [anon_sym___real] = ACTIONS(7237),
    [anon_sym___strong] = ACTIONS(7237),
    [anon_sym___unsafe_unretained] = ACTIONS(7237),
    [anon_sym___unused] = ACTIONS(7237),
    [anon_sym___weak] = ACTIONS(7237),
    [anon_sym_alignas] = ACTIONS(7237),
    [anon_sym__Alignas] = ACTIONS(7237),
    [sym_primitive_type] = ACTIONS(7237),
    [anon_sym_enum] = ACTIONS(7237),
    [anon_sym_COLON] = ACTIONS(7239),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7237),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7237),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7237),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7237),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7237),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7237),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7237),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7237),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7237),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7237),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7237),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7237),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7237),
    [anon_sym_API_AVAILABLE] = ACTIONS(7237),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7237),
    [anon_sym_API_DEPRECATED] = ACTIONS(7237),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7237),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7237),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7237),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7237),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7237),
    [anon_sym___deprecated_msg] = ACTIONS(7237),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7237),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7237),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7237),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7237),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7237),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7237),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7237),
  },
  [STATE(2952)] = {
    [sym_identifier] = ACTIONS(7241),
    [anon_sym_COMMA] = ACTIONS(7243),
    [anon_sym_RPAREN] = ACTIONS(7243),
    [anon_sym_LPAREN2] = ACTIONS(7243),
    [anon_sym_STAR] = ACTIONS(7243),
    [anon_sym_CARET] = ACTIONS(7243),
    [anon_sym_SEMI] = ACTIONS(7243),
    [anon_sym___extension__] = ACTIONS(7241),
    [anon_sym_extern] = ACTIONS(7241),
    [anon_sym___attribute__] = ACTIONS(7241),
    [anon_sym___attribute] = ACTIONS(7241),
    [anon_sym_noreturn] = ACTIONS(7241),
    [anon_sym_LBRACK] = ACTIONS(7243),
    [anon_sym___declspec] = ACTIONS(7241),
    [anon_sym___based] = ACTIONS(7241),
    [anon_sym_signed] = ACTIONS(7241),
    [anon_sym_unsigned] = ACTIONS(7241),
    [anon_sym_long] = ACTIONS(7241),
    [anon_sym_short] = ACTIONS(7241),
    [anon_sym_static] = ACTIONS(7241),
    [anon_sym_auto] = ACTIONS(7241),
    [anon_sym_register] = ACTIONS(7241),
    [anon_sym_inline] = ACTIONS(7241),
    [anon_sym___inline] = ACTIONS(7241),
    [anon_sym___inline__] = ACTIONS(7241),
    [anon_sym___forceinline] = ACTIONS(7241),
    [anon_sym_thread_local] = ACTIONS(7241),
    [anon_sym___thread] = ACTIONS(7241),
    [anon_sym_CG_EXTERN] = ACTIONS(7241),
    [anon_sym_CG_INLINE] = ACTIONS(7241),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7241),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7241),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7241),
    [anon_sym_IBOutlet] = ACTIONS(7241),
    [anon_sym_IBInspectable] = ACTIONS(7241),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7241),
    [anon_sym_NS_INLINE] = ACTIONS(7241),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7241),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7241),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7241),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7241),
    [anon_sym_const] = ACTIONS(7241),
    [anon_sym_constexpr] = ACTIONS(7241),
    [anon_sym_volatile] = ACTIONS(7241),
    [anon_sym_restrict] = ACTIONS(7241),
    [anon_sym___restrict__] = ACTIONS(7241),
    [anon_sym__Atomic] = ACTIONS(7241),
    [anon_sym__Noreturn] = ACTIONS(7241),
    [anon_sym__Nonnull] = ACTIONS(7241),
    [anon_sym_nullable] = ACTIONS(7241),
    [anon_sym__Complex] = ACTIONS(7241),
    [anon_sym__Nullable] = ACTIONS(7241),
    [anon_sym__Nullable_result] = ACTIONS(7241),
    [anon_sym__Null_unspecified] = ACTIONS(7241),
    [anon_sym___autoreleasing] = ACTIONS(7241),
    [anon_sym___block] = ACTIONS(7241),
    [anon_sym___bridge] = ACTIONS(7241),
    [anon_sym___bridge_retained] = ACTIONS(7241),
    [anon_sym___bridge_transfer] = ACTIONS(7241),
    [anon_sym___complex] = ACTIONS(7241),
    [anon_sym___const] = ACTIONS(7241),
    [anon_sym___imag] = ACTIONS(7241),
    [anon_sym___kindof] = ACTIONS(7241),
    [anon_sym___nonnull] = ACTIONS(7241),
    [anon_sym___nullable] = ACTIONS(7241),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7241),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7241),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7241),
    [anon_sym___real] = ACTIONS(7241),
    [anon_sym___strong] = ACTIONS(7241),
    [anon_sym___unsafe_unretained] = ACTIONS(7241),
    [anon_sym___unused] = ACTIONS(7241),
    [anon_sym___weak] = ACTIONS(7241),
    [anon_sym_alignas] = ACTIONS(7241),
    [anon_sym__Alignas] = ACTIONS(7241),
    [sym_primitive_type] = ACTIONS(7241),
    [anon_sym_enum] = ACTIONS(7241),
    [anon_sym_COLON] = ACTIONS(7243),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7241),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7241),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7241),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7241),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7241),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7241),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7241),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7241),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7241),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7241),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7241),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7241),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7241),
    [anon_sym_API_AVAILABLE] = ACTIONS(7241),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7241),
    [anon_sym_API_DEPRECATED] = ACTIONS(7241),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7241),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7241),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7241),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7241),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7241),
    [anon_sym___deprecated_msg] = ACTIONS(7241),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7241),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7241),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7241),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7241),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7241),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7241),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7241),
  },
  [STATE(2953)] = {
    [sym_identifier] = ACTIONS(7245),
    [anon_sym_COMMA] = ACTIONS(7247),
    [anon_sym_RPAREN] = ACTIONS(7247),
    [anon_sym_LPAREN2] = ACTIONS(7247),
    [anon_sym_STAR] = ACTIONS(7247),
    [anon_sym_CARET] = ACTIONS(7247),
    [anon_sym_SEMI] = ACTIONS(7247),
    [anon_sym___extension__] = ACTIONS(7245),
    [anon_sym_extern] = ACTIONS(7245),
    [anon_sym___attribute__] = ACTIONS(7245),
    [anon_sym___attribute] = ACTIONS(7245),
    [anon_sym_noreturn] = ACTIONS(7245),
    [anon_sym_LBRACK] = ACTIONS(7247),
    [anon_sym___declspec] = ACTIONS(7245),
    [anon_sym___based] = ACTIONS(7245),
    [anon_sym_signed] = ACTIONS(7245),
    [anon_sym_unsigned] = ACTIONS(7245),
    [anon_sym_long] = ACTIONS(7245),
    [anon_sym_short] = ACTIONS(7245),
    [anon_sym_static] = ACTIONS(7245),
    [anon_sym_auto] = ACTIONS(7245),
    [anon_sym_register] = ACTIONS(7245),
    [anon_sym_inline] = ACTIONS(7245),
    [anon_sym___inline] = ACTIONS(7245),
    [anon_sym___inline__] = ACTIONS(7245),
    [anon_sym___forceinline] = ACTIONS(7245),
    [anon_sym_thread_local] = ACTIONS(7245),
    [anon_sym___thread] = ACTIONS(7245),
    [anon_sym_CG_EXTERN] = ACTIONS(7245),
    [anon_sym_CG_INLINE] = ACTIONS(7245),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7245),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7245),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7245),
    [anon_sym_IBOutlet] = ACTIONS(7245),
    [anon_sym_IBInspectable] = ACTIONS(7245),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7245),
    [anon_sym_NS_INLINE] = ACTIONS(7245),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7245),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7245),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7245),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7245),
    [anon_sym_const] = ACTIONS(7245),
    [anon_sym_constexpr] = ACTIONS(7245),
    [anon_sym_volatile] = ACTIONS(7245),
    [anon_sym_restrict] = ACTIONS(7245),
    [anon_sym___restrict__] = ACTIONS(7245),
    [anon_sym__Atomic] = ACTIONS(7245),
    [anon_sym__Noreturn] = ACTIONS(7245),
    [anon_sym__Nonnull] = ACTIONS(7245),
    [anon_sym_nullable] = ACTIONS(7245),
    [anon_sym__Complex] = ACTIONS(7245),
    [anon_sym__Nullable] = ACTIONS(7245),
    [anon_sym__Nullable_result] = ACTIONS(7245),
    [anon_sym__Null_unspecified] = ACTIONS(7245),
    [anon_sym___autoreleasing] = ACTIONS(7245),
    [anon_sym___block] = ACTIONS(7245),
    [anon_sym___bridge] = ACTIONS(7245),
    [anon_sym___bridge_retained] = ACTIONS(7245),
    [anon_sym___bridge_transfer] = ACTIONS(7245),
    [anon_sym___complex] = ACTIONS(7245),
    [anon_sym___const] = ACTIONS(7245),
    [anon_sym___imag] = ACTIONS(7245),
    [anon_sym___kindof] = ACTIONS(7245),
    [anon_sym___nonnull] = ACTIONS(7245),
    [anon_sym___nullable] = ACTIONS(7245),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7245),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7245),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7245),
    [anon_sym___real] = ACTIONS(7245),
    [anon_sym___strong] = ACTIONS(7245),
    [anon_sym___unsafe_unretained] = ACTIONS(7245),
    [anon_sym___unused] = ACTIONS(7245),
    [anon_sym___weak] = ACTIONS(7245),
    [anon_sym_alignas] = ACTIONS(7245),
    [anon_sym__Alignas] = ACTIONS(7245),
    [sym_primitive_type] = ACTIONS(7245),
    [anon_sym_enum] = ACTIONS(7245),
    [anon_sym_COLON] = ACTIONS(7247),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7245),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7245),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7245),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7245),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7245),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7245),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7245),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7245),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7245),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7245),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7245),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7245),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7245),
    [anon_sym_API_AVAILABLE] = ACTIONS(7245),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7245),
    [anon_sym_API_DEPRECATED] = ACTIONS(7245),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7245),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7245),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7245),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7245),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7245),
    [anon_sym___deprecated_msg] = ACTIONS(7245),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7245),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7245),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7245),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7245),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7245),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7245),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7245),
  },
  [STATE(2954)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2220),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(7249),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2222),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2222),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2222),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2222),
    [anon_sym_GT_GT] = ACTIONS(2222),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym_RBRACK] = ACTIONS(2222),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_COLON] = ACTIONS(2222),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(2955)] = {
    [sym_identifier] = ACTIONS(7261),
    [anon_sym_COMMA] = ACTIONS(7263),
    [anon_sym_RPAREN] = ACTIONS(7263),
    [anon_sym_LPAREN2] = ACTIONS(7263),
    [anon_sym_STAR] = ACTIONS(7263),
    [anon_sym_CARET] = ACTIONS(7263),
    [anon_sym_SEMI] = ACTIONS(7263),
    [anon_sym___extension__] = ACTIONS(7261),
    [anon_sym_extern] = ACTIONS(7261),
    [anon_sym___attribute__] = ACTIONS(7261),
    [anon_sym___attribute] = ACTIONS(7261),
    [anon_sym_noreturn] = ACTIONS(7261),
    [anon_sym_LBRACK] = ACTIONS(7263),
    [anon_sym___declspec] = ACTIONS(7261),
    [anon_sym___based] = ACTIONS(7261),
    [anon_sym_signed] = ACTIONS(7261),
    [anon_sym_unsigned] = ACTIONS(7261),
    [anon_sym_long] = ACTIONS(7261),
    [anon_sym_short] = ACTIONS(7261),
    [anon_sym_static] = ACTIONS(7261),
    [anon_sym_auto] = ACTIONS(7261),
    [anon_sym_register] = ACTIONS(7261),
    [anon_sym_inline] = ACTIONS(7261),
    [anon_sym___inline] = ACTIONS(7261),
    [anon_sym___inline__] = ACTIONS(7261),
    [anon_sym___forceinline] = ACTIONS(7261),
    [anon_sym_thread_local] = ACTIONS(7261),
    [anon_sym___thread] = ACTIONS(7261),
    [anon_sym_CG_EXTERN] = ACTIONS(7261),
    [anon_sym_CG_INLINE] = ACTIONS(7261),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7261),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7261),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7261),
    [anon_sym_IBOutlet] = ACTIONS(7261),
    [anon_sym_IBInspectable] = ACTIONS(7261),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7261),
    [anon_sym_NS_INLINE] = ACTIONS(7261),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7261),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7261),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7261),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7261),
    [anon_sym_const] = ACTIONS(7261),
    [anon_sym_constexpr] = ACTIONS(7261),
    [anon_sym_volatile] = ACTIONS(7261),
    [anon_sym_restrict] = ACTIONS(7261),
    [anon_sym___restrict__] = ACTIONS(7261),
    [anon_sym__Atomic] = ACTIONS(7261),
    [anon_sym__Noreturn] = ACTIONS(7261),
    [anon_sym__Nonnull] = ACTIONS(7261),
    [anon_sym_nullable] = ACTIONS(7261),
    [anon_sym__Complex] = ACTIONS(7261),
    [anon_sym__Nullable] = ACTIONS(7261),
    [anon_sym__Nullable_result] = ACTIONS(7261),
    [anon_sym__Null_unspecified] = ACTIONS(7261),
    [anon_sym___autoreleasing] = ACTIONS(7261),
    [anon_sym___block] = ACTIONS(7261),
    [anon_sym___bridge] = ACTIONS(7261),
    [anon_sym___bridge_retained] = ACTIONS(7261),
    [anon_sym___bridge_transfer] = ACTIONS(7261),
    [anon_sym___complex] = ACTIONS(7261),
    [anon_sym___const] = ACTIONS(7261),
    [anon_sym___imag] = ACTIONS(7261),
    [anon_sym___kindof] = ACTIONS(7261),
    [anon_sym___nonnull] = ACTIONS(7261),
    [anon_sym___nullable] = ACTIONS(7261),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7261),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7261),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7261),
    [anon_sym___real] = ACTIONS(7261),
    [anon_sym___strong] = ACTIONS(7261),
    [anon_sym___unsafe_unretained] = ACTIONS(7261),
    [anon_sym___unused] = ACTIONS(7261),
    [anon_sym___weak] = ACTIONS(7261),
    [anon_sym_alignas] = ACTIONS(7261),
    [anon_sym__Alignas] = ACTIONS(7261),
    [sym_primitive_type] = ACTIONS(7261),
    [anon_sym_enum] = ACTIONS(7261),
    [anon_sym_COLON] = ACTIONS(7263),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7261),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7261),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7261),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7261),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7261),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7261),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7261),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7261),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7261),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7261),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7261),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7261),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7261),
    [anon_sym_API_AVAILABLE] = ACTIONS(7261),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7261),
    [anon_sym_API_DEPRECATED] = ACTIONS(7261),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7261),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7261),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7261),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7261),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7261),
    [anon_sym___deprecated_msg] = ACTIONS(7261),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7261),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7261),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7261),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7261),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7261),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7261),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7261),
  },
  [STATE(2956)] = {
    [sym_identifier] = ACTIONS(7265),
    [anon_sym_COMMA] = ACTIONS(7267),
    [anon_sym_RPAREN] = ACTIONS(7267),
    [anon_sym_LPAREN2] = ACTIONS(7267),
    [anon_sym_STAR] = ACTIONS(7267),
    [anon_sym_CARET] = ACTIONS(7267),
    [anon_sym_SEMI] = ACTIONS(7267),
    [anon_sym___extension__] = ACTIONS(7265),
    [anon_sym_extern] = ACTIONS(7265),
    [anon_sym___attribute__] = ACTIONS(7265),
    [anon_sym___attribute] = ACTIONS(7265),
    [anon_sym_noreturn] = ACTIONS(7265),
    [anon_sym_LBRACK] = ACTIONS(7267),
    [anon_sym___declspec] = ACTIONS(7265),
    [anon_sym___based] = ACTIONS(7265),
    [anon_sym_signed] = ACTIONS(7265),
    [anon_sym_unsigned] = ACTIONS(7265),
    [anon_sym_long] = ACTIONS(7265),
    [anon_sym_short] = ACTIONS(7265),
    [anon_sym_static] = ACTIONS(7265),
    [anon_sym_auto] = ACTIONS(7265),
    [anon_sym_register] = ACTIONS(7265),
    [anon_sym_inline] = ACTIONS(7265),
    [anon_sym___inline] = ACTIONS(7265),
    [anon_sym___inline__] = ACTIONS(7265),
    [anon_sym___forceinline] = ACTIONS(7265),
    [anon_sym_thread_local] = ACTIONS(7265),
    [anon_sym___thread] = ACTIONS(7265),
    [anon_sym_CG_EXTERN] = ACTIONS(7265),
    [anon_sym_CG_INLINE] = ACTIONS(7265),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7265),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7265),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7265),
    [anon_sym_IBOutlet] = ACTIONS(7265),
    [anon_sym_IBInspectable] = ACTIONS(7265),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7265),
    [anon_sym_NS_INLINE] = ACTIONS(7265),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7265),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7265),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7265),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7265),
    [anon_sym_const] = ACTIONS(7265),
    [anon_sym_constexpr] = ACTIONS(7265),
    [anon_sym_volatile] = ACTIONS(7265),
    [anon_sym_restrict] = ACTIONS(7265),
    [anon_sym___restrict__] = ACTIONS(7265),
    [anon_sym__Atomic] = ACTIONS(7265),
    [anon_sym__Noreturn] = ACTIONS(7265),
    [anon_sym__Nonnull] = ACTIONS(7265),
    [anon_sym_nullable] = ACTIONS(7265),
    [anon_sym__Complex] = ACTIONS(7265),
    [anon_sym__Nullable] = ACTIONS(7265),
    [anon_sym__Nullable_result] = ACTIONS(7265),
    [anon_sym__Null_unspecified] = ACTIONS(7265),
    [anon_sym___autoreleasing] = ACTIONS(7265),
    [anon_sym___block] = ACTIONS(7265),
    [anon_sym___bridge] = ACTIONS(7265),
    [anon_sym___bridge_retained] = ACTIONS(7265),
    [anon_sym___bridge_transfer] = ACTIONS(7265),
    [anon_sym___complex] = ACTIONS(7265),
    [anon_sym___const] = ACTIONS(7265),
    [anon_sym___imag] = ACTIONS(7265),
    [anon_sym___kindof] = ACTIONS(7265),
    [anon_sym___nonnull] = ACTIONS(7265),
    [anon_sym___nullable] = ACTIONS(7265),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7265),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7265),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7265),
    [anon_sym___real] = ACTIONS(7265),
    [anon_sym___strong] = ACTIONS(7265),
    [anon_sym___unsafe_unretained] = ACTIONS(7265),
    [anon_sym___unused] = ACTIONS(7265),
    [anon_sym___weak] = ACTIONS(7265),
    [anon_sym_alignas] = ACTIONS(7265),
    [anon_sym__Alignas] = ACTIONS(7265),
    [sym_primitive_type] = ACTIONS(7265),
    [anon_sym_enum] = ACTIONS(7265),
    [anon_sym_COLON] = ACTIONS(7267),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7265),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7265),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7265),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7265),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7265),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7265),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7265),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7265),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7265),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7265),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7265),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7265),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7265),
    [anon_sym_API_AVAILABLE] = ACTIONS(7265),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7265),
    [anon_sym_API_DEPRECATED] = ACTIONS(7265),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7265),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7265),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7265),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7265),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7265),
    [anon_sym___deprecated_msg] = ACTIONS(7265),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7265),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7265),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7265),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7265),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7265),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7265),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7265),
  },
  [STATE(2957)] = {
    [sym_identifier] = ACTIONS(7269),
    [anon_sym_COMMA] = ACTIONS(7271),
    [anon_sym_RPAREN] = ACTIONS(7271),
    [anon_sym_LPAREN2] = ACTIONS(7271),
    [anon_sym_STAR] = ACTIONS(7271),
    [anon_sym_CARET] = ACTIONS(7271),
    [anon_sym_SEMI] = ACTIONS(7271),
    [anon_sym___extension__] = ACTIONS(7269),
    [anon_sym_extern] = ACTIONS(7269),
    [anon_sym___attribute__] = ACTIONS(7269),
    [anon_sym___attribute] = ACTIONS(7269),
    [anon_sym_noreturn] = ACTIONS(7269),
    [anon_sym_LBRACK] = ACTIONS(7271),
    [anon_sym___declspec] = ACTIONS(7269),
    [anon_sym___based] = ACTIONS(7269),
    [anon_sym_signed] = ACTIONS(7269),
    [anon_sym_unsigned] = ACTIONS(7269),
    [anon_sym_long] = ACTIONS(7269),
    [anon_sym_short] = ACTIONS(7269),
    [anon_sym_static] = ACTIONS(7269),
    [anon_sym_auto] = ACTIONS(7269),
    [anon_sym_register] = ACTIONS(7269),
    [anon_sym_inline] = ACTIONS(7269),
    [anon_sym___inline] = ACTIONS(7269),
    [anon_sym___inline__] = ACTIONS(7269),
    [anon_sym___forceinline] = ACTIONS(7269),
    [anon_sym_thread_local] = ACTIONS(7269),
    [anon_sym___thread] = ACTIONS(7269),
    [anon_sym_CG_EXTERN] = ACTIONS(7269),
    [anon_sym_CG_INLINE] = ACTIONS(7269),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7269),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7269),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7269),
    [anon_sym_IBOutlet] = ACTIONS(7269),
    [anon_sym_IBInspectable] = ACTIONS(7269),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7269),
    [anon_sym_NS_INLINE] = ACTIONS(7269),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7269),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7269),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7269),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7269),
    [anon_sym_const] = ACTIONS(7269),
    [anon_sym_constexpr] = ACTIONS(7269),
    [anon_sym_volatile] = ACTIONS(7269),
    [anon_sym_restrict] = ACTIONS(7269),
    [anon_sym___restrict__] = ACTIONS(7269),
    [anon_sym__Atomic] = ACTIONS(7269),
    [anon_sym__Noreturn] = ACTIONS(7269),
    [anon_sym__Nonnull] = ACTIONS(7269),
    [anon_sym_nullable] = ACTIONS(7269),
    [anon_sym__Complex] = ACTIONS(7269),
    [anon_sym__Nullable] = ACTIONS(7269),
    [anon_sym__Nullable_result] = ACTIONS(7269),
    [anon_sym__Null_unspecified] = ACTIONS(7269),
    [anon_sym___autoreleasing] = ACTIONS(7269),
    [anon_sym___block] = ACTIONS(7269),
    [anon_sym___bridge] = ACTIONS(7269),
    [anon_sym___bridge_retained] = ACTIONS(7269),
    [anon_sym___bridge_transfer] = ACTIONS(7269),
    [anon_sym___complex] = ACTIONS(7269),
    [anon_sym___const] = ACTIONS(7269),
    [anon_sym___imag] = ACTIONS(7269),
    [anon_sym___kindof] = ACTIONS(7269),
    [anon_sym___nonnull] = ACTIONS(7269),
    [anon_sym___nullable] = ACTIONS(7269),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7269),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7269),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7269),
    [anon_sym___real] = ACTIONS(7269),
    [anon_sym___strong] = ACTIONS(7269),
    [anon_sym___unsafe_unretained] = ACTIONS(7269),
    [anon_sym___unused] = ACTIONS(7269),
    [anon_sym___weak] = ACTIONS(7269),
    [anon_sym_alignas] = ACTIONS(7269),
    [anon_sym__Alignas] = ACTIONS(7269),
    [sym_primitive_type] = ACTIONS(7269),
    [anon_sym_enum] = ACTIONS(7269),
    [anon_sym_COLON] = ACTIONS(7271),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7269),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7269),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7269),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7269),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7269),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7269),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7269),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7269),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7269),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7269),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7269),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7269),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7269),
    [anon_sym_API_AVAILABLE] = ACTIONS(7269),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7269),
    [anon_sym_API_DEPRECATED] = ACTIONS(7269),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7269),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7269),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7269),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7269),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7269),
    [anon_sym___deprecated_msg] = ACTIONS(7269),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7269),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7269),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7269),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7269),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7269),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7269),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7269),
  },
  [STATE(2958)] = {
    [sym_identifier] = ACTIONS(7273),
    [anon_sym_COMMA] = ACTIONS(7275),
    [anon_sym_RPAREN] = ACTIONS(7275),
    [anon_sym_LPAREN2] = ACTIONS(7275),
    [anon_sym_STAR] = ACTIONS(7275),
    [anon_sym_CARET] = ACTIONS(7275),
    [anon_sym_SEMI] = ACTIONS(7275),
    [anon_sym___extension__] = ACTIONS(7273),
    [anon_sym_extern] = ACTIONS(7273),
    [anon_sym___attribute__] = ACTIONS(7273),
    [anon_sym___attribute] = ACTIONS(7273),
    [anon_sym_noreturn] = ACTIONS(7273),
    [anon_sym_LBRACK] = ACTIONS(7275),
    [anon_sym___declspec] = ACTIONS(7273),
    [anon_sym___based] = ACTIONS(7273),
    [anon_sym_signed] = ACTIONS(7273),
    [anon_sym_unsigned] = ACTIONS(7273),
    [anon_sym_long] = ACTIONS(7273),
    [anon_sym_short] = ACTIONS(7273),
    [anon_sym_static] = ACTIONS(7273),
    [anon_sym_auto] = ACTIONS(7273),
    [anon_sym_register] = ACTIONS(7273),
    [anon_sym_inline] = ACTIONS(7273),
    [anon_sym___inline] = ACTIONS(7273),
    [anon_sym___inline__] = ACTIONS(7273),
    [anon_sym___forceinline] = ACTIONS(7273),
    [anon_sym_thread_local] = ACTIONS(7273),
    [anon_sym___thread] = ACTIONS(7273),
    [anon_sym_CG_EXTERN] = ACTIONS(7273),
    [anon_sym_CG_INLINE] = ACTIONS(7273),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7273),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7273),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7273),
    [anon_sym_IBOutlet] = ACTIONS(7273),
    [anon_sym_IBInspectable] = ACTIONS(7273),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7273),
    [anon_sym_NS_INLINE] = ACTIONS(7273),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7273),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7273),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7273),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7273),
    [anon_sym_const] = ACTIONS(7273),
    [anon_sym_constexpr] = ACTIONS(7273),
    [anon_sym_volatile] = ACTIONS(7273),
    [anon_sym_restrict] = ACTIONS(7273),
    [anon_sym___restrict__] = ACTIONS(7273),
    [anon_sym__Atomic] = ACTIONS(7273),
    [anon_sym__Noreturn] = ACTIONS(7273),
    [anon_sym__Nonnull] = ACTIONS(7273),
    [anon_sym_nullable] = ACTIONS(7273),
    [anon_sym__Complex] = ACTIONS(7273),
    [anon_sym__Nullable] = ACTIONS(7273),
    [anon_sym__Nullable_result] = ACTIONS(7273),
    [anon_sym__Null_unspecified] = ACTIONS(7273),
    [anon_sym___autoreleasing] = ACTIONS(7273),
    [anon_sym___block] = ACTIONS(7273),
    [anon_sym___bridge] = ACTIONS(7273),
    [anon_sym___bridge_retained] = ACTIONS(7273),
    [anon_sym___bridge_transfer] = ACTIONS(7273),
    [anon_sym___complex] = ACTIONS(7273),
    [anon_sym___const] = ACTIONS(7273),
    [anon_sym___imag] = ACTIONS(7273),
    [anon_sym___kindof] = ACTIONS(7273),
    [anon_sym___nonnull] = ACTIONS(7273),
    [anon_sym___nullable] = ACTIONS(7273),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7273),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7273),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7273),
    [anon_sym___real] = ACTIONS(7273),
    [anon_sym___strong] = ACTIONS(7273),
    [anon_sym___unsafe_unretained] = ACTIONS(7273),
    [anon_sym___unused] = ACTIONS(7273),
    [anon_sym___weak] = ACTIONS(7273),
    [anon_sym_alignas] = ACTIONS(7273),
    [anon_sym__Alignas] = ACTIONS(7273),
    [sym_primitive_type] = ACTIONS(7273),
    [anon_sym_enum] = ACTIONS(7273),
    [anon_sym_COLON] = ACTIONS(7275),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7273),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7273),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7273),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7273),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7273),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7273),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7273),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7273),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7273),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7273),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7273),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7273),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7273),
    [anon_sym_API_AVAILABLE] = ACTIONS(7273),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7273),
    [anon_sym_API_DEPRECATED] = ACTIONS(7273),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7273),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7273),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7273),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7273),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7273),
    [anon_sym___deprecated_msg] = ACTIONS(7273),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7273),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7273),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7273),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7273),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7273),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7273),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7273),
  },
  [STATE(2959)] = {
    [sym__declaration_modifiers] = STATE(2866),
    [sym_attribute_specifier] = STATE(2866),
    [sym_attribute_declaration] = STATE(2866),
    [sym_ms_declspec_modifier] = STATE(2866),
    [sym_storage_class_specifier] = STATE(2866),
    [sym_type_qualifier] = STATE(2866),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2866),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2866),
    [sym_identifier] = ACTIONS(6899),
    [anon_sym_LPAREN2] = ACTIONS(6901),
    [anon_sym_STAR] = ACTIONS(6901),
    [anon_sym_CARET] = ACTIONS(6901),
    [anon_sym_SEMI] = ACTIONS(7277),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6897),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(6899),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2960)] = {
    [sym_identifier] = ACTIONS(7279),
    [anon_sym_COMMA] = ACTIONS(7281),
    [anon_sym_RPAREN] = ACTIONS(7281),
    [anon_sym_LPAREN2] = ACTIONS(7281),
    [anon_sym_STAR] = ACTIONS(7281),
    [anon_sym_CARET] = ACTIONS(7281),
    [anon_sym_SEMI] = ACTIONS(7281),
    [anon_sym___extension__] = ACTIONS(7279),
    [anon_sym_extern] = ACTIONS(7279),
    [anon_sym___attribute__] = ACTIONS(7279),
    [anon_sym___attribute] = ACTIONS(7279),
    [anon_sym_noreturn] = ACTIONS(7279),
    [anon_sym_LBRACK] = ACTIONS(7281),
    [anon_sym___declspec] = ACTIONS(7279),
    [anon_sym___based] = ACTIONS(7279),
    [anon_sym_signed] = ACTIONS(7279),
    [anon_sym_unsigned] = ACTIONS(7279),
    [anon_sym_long] = ACTIONS(7279),
    [anon_sym_short] = ACTIONS(7279),
    [anon_sym_static] = ACTIONS(7279),
    [anon_sym_auto] = ACTIONS(7279),
    [anon_sym_register] = ACTIONS(7279),
    [anon_sym_inline] = ACTIONS(7279),
    [anon_sym___inline] = ACTIONS(7279),
    [anon_sym___inline__] = ACTIONS(7279),
    [anon_sym___forceinline] = ACTIONS(7279),
    [anon_sym_thread_local] = ACTIONS(7279),
    [anon_sym___thread] = ACTIONS(7279),
    [anon_sym_CG_EXTERN] = ACTIONS(7279),
    [anon_sym_CG_INLINE] = ACTIONS(7279),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7279),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7279),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7279),
    [anon_sym_IBOutlet] = ACTIONS(7279),
    [anon_sym_IBInspectable] = ACTIONS(7279),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7279),
    [anon_sym_NS_INLINE] = ACTIONS(7279),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7279),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7279),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7279),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7279),
    [anon_sym_const] = ACTIONS(7279),
    [anon_sym_constexpr] = ACTIONS(7279),
    [anon_sym_volatile] = ACTIONS(7279),
    [anon_sym_restrict] = ACTIONS(7279),
    [anon_sym___restrict__] = ACTIONS(7279),
    [anon_sym__Atomic] = ACTIONS(7279),
    [anon_sym__Noreturn] = ACTIONS(7279),
    [anon_sym__Nonnull] = ACTIONS(7279),
    [anon_sym_nullable] = ACTIONS(7279),
    [anon_sym__Complex] = ACTIONS(7279),
    [anon_sym__Nullable] = ACTIONS(7279),
    [anon_sym__Nullable_result] = ACTIONS(7279),
    [anon_sym__Null_unspecified] = ACTIONS(7279),
    [anon_sym___autoreleasing] = ACTIONS(7279),
    [anon_sym___block] = ACTIONS(7279),
    [anon_sym___bridge] = ACTIONS(7279),
    [anon_sym___bridge_retained] = ACTIONS(7279),
    [anon_sym___bridge_transfer] = ACTIONS(7279),
    [anon_sym___complex] = ACTIONS(7279),
    [anon_sym___const] = ACTIONS(7279),
    [anon_sym___imag] = ACTIONS(7279),
    [anon_sym___kindof] = ACTIONS(7279),
    [anon_sym___nonnull] = ACTIONS(7279),
    [anon_sym___nullable] = ACTIONS(7279),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7279),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7279),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7279),
    [anon_sym___real] = ACTIONS(7279),
    [anon_sym___strong] = ACTIONS(7279),
    [anon_sym___unsafe_unretained] = ACTIONS(7279),
    [anon_sym___unused] = ACTIONS(7279),
    [anon_sym___weak] = ACTIONS(7279),
    [anon_sym_alignas] = ACTIONS(7279),
    [anon_sym__Alignas] = ACTIONS(7279),
    [sym_primitive_type] = ACTIONS(7279),
    [anon_sym_enum] = ACTIONS(7279),
    [anon_sym_COLON] = ACTIONS(7281),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7279),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7279),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7279),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7279),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7279),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7279),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7279),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7279),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7279),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7279),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7279),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7279),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7279),
    [anon_sym_API_AVAILABLE] = ACTIONS(7279),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7279),
    [anon_sym_API_DEPRECATED] = ACTIONS(7279),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7279),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7279),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7279),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7279),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7279),
    [anon_sym___deprecated_msg] = ACTIONS(7279),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7279),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7279),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7279),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7279),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7279),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7279),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7279),
  },
  [STATE(2961)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2912),
    [sym_identifier] = ACTIONS(7283),
    [anon_sym_COMMA] = ACTIONS(7285),
    [anon_sym_RPAREN] = ACTIONS(7285),
    [anon_sym_LPAREN2] = ACTIONS(7285),
    [anon_sym_STAR] = ACTIONS(7285),
    [anon_sym_CARET] = ACTIONS(7285),
    [anon_sym_SEMI] = ACTIONS(7285),
    [anon_sym___extension__] = ACTIONS(7283),
    [anon_sym_extern] = ACTIONS(7283),
    [anon_sym___attribute__] = ACTIONS(7283),
    [anon_sym___attribute] = ACTIONS(7283),
    [anon_sym_noreturn] = ACTIONS(7283),
    [anon_sym_LBRACK] = ACTIONS(7285),
    [anon_sym___declspec] = ACTIONS(7283),
    [anon_sym___based] = ACTIONS(7283),
    [anon_sym_signed] = ACTIONS(7287),
    [anon_sym_unsigned] = ACTIONS(7287),
    [anon_sym_long] = ACTIONS(7287),
    [anon_sym_short] = ACTIONS(7287),
    [anon_sym_static] = ACTIONS(7283),
    [anon_sym_auto] = ACTIONS(7283),
    [anon_sym_register] = ACTIONS(7283),
    [anon_sym_inline] = ACTIONS(7283),
    [anon_sym___inline] = ACTIONS(7283),
    [anon_sym___inline__] = ACTIONS(7283),
    [anon_sym___forceinline] = ACTIONS(7283),
    [anon_sym_thread_local] = ACTIONS(7283),
    [anon_sym___thread] = ACTIONS(7283),
    [anon_sym_CG_EXTERN] = ACTIONS(7283),
    [anon_sym_CG_INLINE] = ACTIONS(7283),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7283),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7283),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7283),
    [anon_sym_IBOutlet] = ACTIONS(7283),
    [anon_sym_IBInspectable] = ACTIONS(7283),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7283),
    [anon_sym_NS_INLINE] = ACTIONS(7283),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7283),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7283),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7283),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7283),
    [anon_sym_const] = ACTIONS(7283),
    [anon_sym_constexpr] = ACTIONS(7283),
    [anon_sym_volatile] = ACTIONS(7283),
    [anon_sym_restrict] = ACTIONS(7283),
    [anon_sym___restrict__] = ACTIONS(7283),
    [anon_sym__Atomic] = ACTIONS(7283),
    [anon_sym__Noreturn] = ACTIONS(7283),
    [anon_sym__Nonnull] = ACTIONS(7283),
    [anon_sym_nullable] = ACTIONS(7283),
    [anon_sym__Complex] = ACTIONS(7283),
    [anon_sym__Nullable] = ACTIONS(7283),
    [anon_sym__Nullable_result] = ACTIONS(7283),
    [anon_sym__Null_unspecified] = ACTIONS(7283),
    [anon_sym___autoreleasing] = ACTIONS(7283),
    [anon_sym___block] = ACTIONS(7283),
    [anon_sym___bridge] = ACTIONS(7283),
    [anon_sym___bridge_retained] = ACTIONS(7283),
    [anon_sym___bridge_transfer] = ACTIONS(7283),
    [anon_sym___complex] = ACTIONS(7283),
    [anon_sym___const] = ACTIONS(7283),
    [anon_sym___imag] = ACTIONS(7283),
    [anon_sym___kindof] = ACTIONS(7283),
    [anon_sym___nonnull] = ACTIONS(7283),
    [anon_sym___nullable] = ACTIONS(7283),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7283),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7283),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7283),
    [anon_sym___real] = ACTIONS(7283),
    [anon_sym___strong] = ACTIONS(7283),
    [anon_sym___unsafe_unretained] = ACTIONS(7283),
    [anon_sym___unused] = ACTIONS(7283),
    [anon_sym___weak] = ACTIONS(7283),
    [anon_sym_alignas] = ACTIONS(7283),
    [anon_sym__Alignas] = ACTIONS(7283),
    [anon_sym_enum] = ACTIONS(7283),
    [anon_sym_COLON] = ACTIONS(7285),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7283),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7283),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7283),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7283),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7283),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7283),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7283),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7283),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7283),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7283),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7283),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7283),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7283),
    [anon_sym_API_AVAILABLE] = ACTIONS(7283),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7283),
    [anon_sym_API_DEPRECATED] = ACTIONS(7283),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7283),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7283),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7283),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7283),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7283),
    [anon_sym___deprecated_msg] = ACTIONS(7283),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7283),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7283),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7283),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7283),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7283),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7283),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7283),
  },
  [STATE(2962)] = {
    [sym_method_parameter] = STATE(2790),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_LPAREN2] = ACTIONS(7042),
    [anon_sym_STAR] = ACTIONS(6134),
    [anon_sym_CARET] = ACTIONS(6134),
    [anon_sym_LT] = ACTIONS(7045),
    [anon_sym_SEMI] = ACTIONS(7289),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(6134),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(4653),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(2963)] = {
    [sym_identifier] = ACTIONS(7291),
    [anon_sym_COMMA] = ACTIONS(7293),
    [anon_sym_RPAREN] = ACTIONS(7293),
    [anon_sym_LPAREN2] = ACTIONS(7293),
    [anon_sym_STAR] = ACTIONS(7293),
    [anon_sym_CARET] = ACTIONS(7293),
    [anon_sym_SEMI] = ACTIONS(7293),
    [anon_sym___extension__] = ACTIONS(7291),
    [anon_sym_extern] = ACTIONS(7291),
    [anon_sym___attribute__] = ACTIONS(7291),
    [anon_sym___attribute] = ACTIONS(7291),
    [anon_sym_noreturn] = ACTIONS(7291),
    [anon_sym_LBRACK] = ACTIONS(7293),
    [anon_sym___declspec] = ACTIONS(7291),
    [anon_sym___based] = ACTIONS(7291),
    [anon_sym_signed] = ACTIONS(7291),
    [anon_sym_unsigned] = ACTIONS(7291),
    [anon_sym_long] = ACTIONS(7291),
    [anon_sym_short] = ACTIONS(7291),
    [anon_sym_static] = ACTIONS(7291),
    [anon_sym_auto] = ACTIONS(7291),
    [anon_sym_register] = ACTIONS(7291),
    [anon_sym_inline] = ACTIONS(7291),
    [anon_sym___inline] = ACTIONS(7291),
    [anon_sym___inline__] = ACTIONS(7291),
    [anon_sym___forceinline] = ACTIONS(7291),
    [anon_sym_thread_local] = ACTIONS(7291),
    [anon_sym___thread] = ACTIONS(7291),
    [anon_sym_CG_EXTERN] = ACTIONS(7291),
    [anon_sym_CG_INLINE] = ACTIONS(7291),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7291),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7291),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7291),
    [anon_sym_IBOutlet] = ACTIONS(7291),
    [anon_sym_IBInspectable] = ACTIONS(7291),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7291),
    [anon_sym_NS_INLINE] = ACTIONS(7291),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7291),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7291),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7291),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7291),
    [anon_sym_const] = ACTIONS(7291),
    [anon_sym_constexpr] = ACTIONS(7291),
    [anon_sym_volatile] = ACTIONS(7291),
    [anon_sym_restrict] = ACTIONS(7291),
    [anon_sym___restrict__] = ACTIONS(7291),
    [anon_sym__Atomic] = ACTIONS(7291),
    [anon_sym__Noreturn] = ACTIONS(7291),
    [anon_sym__Nonnull] = ACTIONS(7291),
    [anon_sym_nullable] = ACTIONS(7291),
    [anon_sym__Complex] = ACTIONS(7291),
    [anon_sym__Nullable] = ACTIONS(7291),
    [anon_sym__Nullable_result] = ACTIONS(7291),
    [anon_sym__Null_unspecified] = ACTIONS(7291),
    [anon_sym___autoreleasing] = ACTIONS(7291),
    [anon_sym___block] = ACTIONS(7291),
    [anon_sym___bridge] = ACTIONS(7291),
    [anon_sym___bridge_retained] = ACTIONS(7291),
    [anon_sym___bridge_transfer] = ACTIONS(7291),
    [anon_sym___complex] = ACTIONS(7291),
    [anon_sym___const] = ACTIONS(7291),
    [anon_sym___imag] = ACTIONS(7291),
    [anon_sym___kindof] = ACTIONS(7291),
    [anon_sym___nonnull] = ACTIONS(7291),
    [anon_sym___nullable] = ACTIONS(7291),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7291),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7291),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7291),
    [anon_sym___real] = ACTIONS(7291),
    [anon_sym___strong] = ACTIONS(7291),
    [anon_sym___unsafe_unretained] = ACTIONS(7291),
    [anon_sym___unused] = ACTIONS(7291),
    [anon_sym___weak] = ACTIONS(7291),
    [anon_sym_alignas] = ACTIONS(7291),
    [anon_sym__Alignas] = ACTIONS(7291),
    [sym_primitive_type] = ACTIONS(7291),
    [anon_sym_enum] = ACTIONS(7291),
    [anon_sym_COLON] = ACTIONS(7293),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7291),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7291),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7291),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7291),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7291),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7291),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7291),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7291),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7291),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7291),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7291),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7291),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7291),
    [anon_sym_API_AVAILABLE] = ACTIONS(7291),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7291),
    [anon_sym_API_DEPRECATED] = ACTIONS(7291),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7291),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7291),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7291),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7291),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7291),
    [anon_sym___deprecated_msg] = ACTIONS(7291),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7291),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7291),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7291),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7291),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7291),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7291),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7291),
  },
  [STATE(2964)] = {
    [sym_identifier] = ACTIONS(7295),
    [anon_sym_COMMA] = ACTIONS(7297),
    [anon_sym_RPAREN] = ACTIONS(7297),
    [anon_sym_LPAREN2] = ACTIONS(7297),
    [anon_sym_STAR] = ACTIONS(7297),
    [anon_sym_CARET] = ACTIONS(7297),
    [anon_sym_SEMI] = ACTIONS(7297),
    [anon_sym___extension__] = ACTIONS(7295),
    [anon_sym_extern] = ACTIONS(7295),
    [anon_sym___attribute__] = ACTIONS(7295),
    [anon_sym___attribute] = ACTIONS(7295),
    [anon_sym_noreturn] = ACTIONS(7295),
    [anon_sym_LBRACK] = ACTIONS(7297),
    [anon_sym___declspec] = ACTIONS(7295),
    [anon_sym___based] = ACTIONS(7295),
    [anon_sym_signed] = ACTIONS(7295),
    [anon_sym_unsigned] = ACTIONS(7295),
    [anon_sym_long] = ACTIONS(7295),
    [anon_sym_short] = ACTIONS(7295),
    [anon_sym_static] = ACTIONS(7295),
    [anon_sym_auto] = ACTIONS(7295),
    [anon_sym_register] = ACTIONS(7295),
    [anon_sym_inline] = ACTIONS(7295),
    [anon_sym___inline] = ACTIONS(7295),
    [anon_sym___inline__] = ACTIONS(7295),
    [anon_sym___forceinline] = ACTIONS(7295),
    [anon_sym_thread_local] = ACTIONS(7295),
    [anon_sym___thread] = ACTIONS(7295),
    [anon_sym_CG_EXTERN] = ACTIONS(7295),
    [anon_sym_CG_INLINE] = ACTIONS(7295),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7295),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7295),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7295),
    [anon_sym_IBOutlet] = ACTIONS(7295),
    [anon_sym_IBInspectable] = ACTIONS(7295),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7295),
    [anon_sym_NS_INLINE] = ACTIONS(7295),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7295),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7295),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7295),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7295),
    [anon_sym_const] = ACTIONS(7295),
    [anon_sym_constexpr] = ACTIONS(7295),
    [anon_sym_volatile] = ACTIONS(7295),
    [anon_sym_restrict] = ACTIONS(7295),
    [anon_sym___restrict__] = ACTIONS(7295),
    [anon_sym__Atomic] = ACTIONS(7295),
    [anon_sym__Noreturn] = ACTIONS(7295),
    [anon_sym__Nonnull] = ACTIONS(7295),
    [anon_sym_nullable] = ACTIONS(7295),
    [anon_sym__Complex] = ACTIONS(7295),
    [anon_sym__Nullable] = ACTIONS(7295),
    [anon_sym__Nullable_result] = ACTIONS(7295),
    [anon_sym__Null_unspecified] = ACTIONS(7295),
    [anon_sym___autoreleasing] = ACTIONS(7295),
    [anon_sym___block] = ACTIONS(7295),
    [anon_sym___bridge] = ACTIONS(7295),
    [anon_sym___bridge_retained] = ACTIONS(7295),
    [anon_sym___bridge_transfer] = ACTIONS(7295),
    [anon_sym___complex] = ACTIONS(7295),
    [anon_sym___const] = ACTIONS(7295),
    [anon_sym___imag] = ACTIONS(7295),
    [anon_sym___kindof] = ACTIONS(7295),
    [anon_sym___nonnull] = ACTIONS(7295),
    [anon_sym___nullable] = ACTIONS(7295),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7295),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7295),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7295),
    [anon_sym___real] = ACTIONS(7295),
    [anon_sym___strong] = ACTIONS(7295),
    [anon_sym___unsafe_unretained] = ACTIONS(7295),
    [anon_sym___unused] = ACTIONS(7295),
    [anon_sym___weak] = ACTIONS(7295),
    [anon_sym_alignas] = ACTIONS(7295),
    [anon_sym__Alignas] = ACTIONS(7295),
    [sym_primitive_type] = ACTIONS(7295),
    [anon_sym_enum] = ACTIONS(7295),
    [anon_sym_COLON] = ACTIONS(7297),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7295),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7295),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7295),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7295),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7295),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7295),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7295),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7295),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7295),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7295),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7295),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7295),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7295),
    [anon_sym_API_AVAILABLE] = ACTIONS(7295),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7295),
    [anon_sym_API_DEPRECATED] = ACTIONS(7295),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7295),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7295),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7295),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7295),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7295),
    [anon_sym___deprecated_msg] = ACTIONS(7295),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7295),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7295),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7295),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7295),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7295),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7295),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7295),
  },
  [STATE(2965)] = {
    [sym_identifier] = ACTIONS(7299),
    [anon_sym_COMMA] = ACTIONS(7301),
    [anon_sym_RPAREN] = ACTIONS(7301),
    [anon_sym_LPAREN2] = ACTIONS(7301),
    [anon_sym_STAR] = ACTIONS(7301),
    [anon_sym_CARET] = ACTIONS(7301),
    [anon_sym_SEMI] = ACTIONS(7301),
    [anon_sym___extension__] = ACTIONS(7299),
    [anon_sym_extern] = ACTIONS(7299),
    [anon_sym___attribute__] = ACTIONS(7299),
    [anon_sym___attribute] = ACTIONS(7299),
    [anon_sym_noreturn] = ACTIONS(7299),
    [anon_sym_LBRACK] = ACTIONS(7301),
    [anon_sym___declspec] = ACTIONS(7299),
    [anon_sym___based] = ACTIONS(7299),
    [anon_sym_signed] = ACTIONS(7299),
    [anon_sym_unsigned] = ACTIONS(7299),
    [anon_sym_long] = ACTIONS(7299),
    [anon_sym_short] = ACTIONS(7299),
    [anon_sym_static] = ACTIONS(7299),
    [anon_sym_auto] = ACTIONS(7299),
    [anon_sym_register] = ACTIONS(7299),
    [anon_sym_inline] = ACTIONS(7299),
    [anon_sym___inline] = ACTIONS(7299),
    [anon_sym___inline__] = ACTIONS(7299),
    [anon_sym___forceinline] = ACTIONS(7299),
    [anon_sym_thread_local] = ACTIONS(7299),
    [anon_sym___thread] = ACTIONS(7299),
    [anon_sym_CG_EXTERN] = ACTIONS(7299),
    [anon_sym_CG_INLINE] = ACTIONS(7299),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7299),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7299),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7299),
    [anon_sym_IBOutlet] = ACTIONS(7299),
    [anon_sym_IBInspectable] = ACTIONS(7299),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7299),
    [anon_sym_NS_INLINE] = ACTIONS(7299),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7299),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7299),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7299),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7299),
    [anon_sym_const] = ACTIONS(7299),
    [anon_sym_constexpr] = ACTIONS(7299),
    [anon_sym_volatile] = ACTIONS(7299),
    [anon_sym_restrict] = ACTIONS(7299),
    [anon_sym___restrict__] = ACTIONS(7299),
    [anon_sym__Atomic] = ACTIONS(7299),
    [anon_sym__Noreturn] = ACTIONS(7299),
    [anon_sym__Nonnull] = ACTIONS(7299),
    [anon_sym_nullable] = ACTIONS(7299),
    [anon_sym__Complex] = ACTIONS(7299),
    [anon_sym__Nullable] = ACTIONS(7299),
    [anon_sym__Nullable_result] = ACTIONS(7299),
    [anon_sym__Null_unspecified] = ACTIONS(7299),
    [anon_sym___autoreleasing] = ACTIONS(7299),
    [anon_sym___block] = ACTIONS(7299),
    [anon_sym___bridge] = ACTIONS(7299),
    [anon_sym___bridge_retained] = ACTIONS(7299),
    [anon_sym___bridge_transfer] = ACTIONS(7299),
    [anon_sym___complex] = ACTIONS(7299),
    [anon_sym___const] = ACTIONS(7299),
    [anon_sym___imag] = ACTIONS(7299),
    [anon_sym___kindof] = ACTIONS(7299),
    [anon_sym___nonnull] = ACTIONS(7299),
    [anon_sym___nullable] = ACTIONS(7299),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7299),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7299),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7299),
    [anon_sym___real] = ACTIONS(7299),
    [anon_sym___strong] = ACTIONS(7299),
    [anon_sym___unsafe_unretained] = ACTIONS(7299),
    [anon_sym___unused] = ACTIONS(7299),
    [anon_sym___weak] = ACTIONS(7299),
    [anon_sym_alignas] = ACTIONS(7299),
    [anon_sym__Alignas] = ACTIONS(7299),
    [sym_primitive_type] = ACTIONS(7299),
    [anon_sym_enum] = ACTIONS(7299),
    [anon_sym_COLON] = ACTIONS(7301),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7299),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7299),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7299),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7299),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7299),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7299),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7299),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7299),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7299),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7299),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7299),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7299),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7299),
    [anon_sym_API_AVAILABLE] = ACTIONS(7299),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7299),
    [anon_sym_API_DEPRECATED] = ACTIONS(7299),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7299),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7299),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7299),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7299),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7299),
    [anon_sym___deprecated_msg] = ACTIONS(7299),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7299),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7299),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7299),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7299),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7299),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7299),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7299),
  },
  [STATE(2966)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2912),
    [sym_identifier] = ACTIONS(7303),
    [anon_sym_COMMA] = ACTIONS(7305),
    [anon_sym_RPAREN] = ACTIONS(7305),
    [anon_sym_LPAREN2] = ACTIONS(7305),
    [anon_sym_STAR] = ACTIONS(7305),
    [anon_sym_CARET] = ACTIONS(7305),
    [anon_sym_SEMI] = ACTIONS(7305),
    [anon_sym___extension__] = ACTIONS(7303),
    [anon_sym_extern] = ACTIONS(7303),
    [anon_sym___attribute__] = ACTIONS(7303),
    [anon_sym___attribute] = ACTIONS(7303),
    [anon_sym_noreturn] = ACTIONS(7303),
    [anon_sym_LBRACK] = ACTIONS(7305),
    [anon_sym___declspec] = ACTIONS(7303),
    [anon_sym___based] = ACTIONS(7303),
    [anon_sym_signed] = ACTIONS(7287),
    [anon_sym_unsigned] = ACTIONS(7287),
    [anon_sym_long] = ACTIONS(7287),
    [anon_sym_short] = ACTIONS(7287),
    [anon_sym_static] = ACTIONS(7303),
    [anon_sym_auto] = ACTIONS(7303),
    [anon_sym_register] = ACTIONS(7303),
    [anon_sym_inline] = ACTIONS(7303),
    [anon_sym___inline] = ACTIONS(7303),
    [anon_sym___inline__] = ACTIONS(7303),
    [anon_sym___forceinline] = ACTIONS(7303),
    [anon_sym_thread_local] = ACTIONS(7303),
    [anon_sym___thread] = ACTIONS(7303),
    [anon_sym_CG_EXTERN] = ACTIONS(7303),
    [anon_sym_CG_INLINE] = ACTIONS(7303),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7303),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7303),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7303),
    [anon_sym_IBOutlet] = ACTIONS(7303),
    [anon_sym_IBInspectable] = ACTIONS(7303),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7303),
    [anon_sym_NS_INLINE] = ACTIONS(7303),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7303),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7303),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7303),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7303),
    [anon_sym_const] = ACTIONS(7303),
    [anon_sym_constexpr] = ACTIONS(7303),
    [anon_sym_volatile] = ACTIONS(7303),
    [anon_sym_restrict] = ACTIONS(7303),
    [anon_sym___restrict__] = ACTIONS(7303),
    [anon_sym__Atomic] = ACTIONS(7303),
    [anon_sym__Noreturn] = ACTIONS(7303),
    [anon_sym__Nonnull] = ACTIONS(7303),
    [anon_sym_nullable] = ACTIONS(7303),
    [anon_sym__Complex] = ACTIONS(7303),
    [anon_sym__Nullable] = ACTIONS(7303),
    [anon_sym__Nullable_result] = ACTIONS(7303),
    [anon_sym__Null_unspecified] = ACTIONS(7303),
    [anon_sym___autoreleasing] = ACTIONS(7303),
    [anon_sym___block] = ACTIONS(7303),
    [anon_sym___bridge] = ACTIONS(7303),
    [anon_sym___bridge_retained] = ACTIONS(7303),
    [anon_sym___bridge_transfer] = ACTIONS(7303),
    [anon_sym___complex] = ACTIONS(7303),
    [anon_sym___const] = ACTIONS(7303),
    [anon_sym___imag] = ACTIONS(7303),
    [anon_sym___kindof] = ACTIONS(7303),
    [anon_sym___nonnull] = ACTIONS(7303),
    [anon_sym___nullable] = ACTIONS(7303),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7303),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7303),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7303),
    [anon_sym___real] = ACTIONS(7303),
    [anon_sym___strong] = ACTIONS(7303),
    [anon_sym___unsafe_unretained] = ACTIONS(7303),
    [anon_sym___unused] = ACTIONS(7303),
    [anon_sym___weak] = ACTIONS(7303),
    [anon_sym_alignas] = ACTIONS(7303),
    [anon_sym__Alignas] = ACTIONS(7303),
    [anon_sym_enum] = ACTIONS(7303),
    [anon_sym_COLON] = ACTIONS(7305),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7303),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7303),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7303),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7303),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7303),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7303),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7303),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7303),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7303),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7303),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7303),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7303),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7303),
    [anon_sym_API_AVAILABLE] = ACTIONS(7303),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7303),
    [anon_sym_API_DEPRECATED] = ACTIONS(7303),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7303),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7303),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7303),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7303),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7303),
    [anon_sym___deprecated_msg] = ACTIONS(7303),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7303),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7303),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7303),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7303),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7303),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7303),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7303),
  },
  [STATE(2967)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2912),
    [sym_identifier] = ACTIONS(7307),
    [anon_sym_COMMA] = ACTIONS(7309),
    [anon_sym_RPAREN] = ACTIONS(7309),
    [anon_sym_LPAREN2] = ACTIONS(7309),
    [anon_sym_STAR] = ACTIONS(7309),
    [anon_sym_CARET] = ACTIONS(7309),
    [anon_sym_SEMI] = ACTIONS(7309),
    [anon_sym___extension__] = ACTIONS(7307),
    [anon_sym_extern] = ACTIONS(7307),
    [anon_sym___attribute__] = ACTIONS(7307),
    [anon_sym___attribute] = ACTIONS(7307),
    [anon_sym_noreturn] = ACTIONS(7307),
    [anon_sym_LBRACK] = ACTIONS(7309),
    [anon_sym___declspec] = ACTIONS(7307),
    [anon_sym___based] = ACTIONS(7307),
    [anon_sym_signed] = ACTIONS(7287),
    [anon_sym_unsigned] = ACTIONS(7287),
    [anon_sym_long] = ACTIONS(7287),
    [anon_sym_short] = ACTIONS(7287),
    [anon_sym_static] = ACTIONS(7307),
    [anon_sym_auto] = ACTIONS(7307),
    [anon_sym_register] = ACTIONS(7307),
    [anon_sym_inline] = ACTIONS(7307),
    [anon_sym___inline] = ACTIONS(7307),
    [anon_sym___inline__] = ACTIONS(7307),
    [anon_sym___forceinline] = ACTIONS(7307),
    [anon_sym_thread_local] = ACTIONS(7307),
    [anon_sym___thread] = ACTIONS(7307),
    [anon_sym_CG_EXTERN] = ACTIONS(7307),
    [anon_sym_CG_INLINE] = ACTIONS(7307),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7307),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7307),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7307),
    [anon_sym_IBOutlet] = ACTIONS(7307),
    [anon_sym_IBInspectable] = ACTIONS(7307),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7307),
    [anon_sym_NS_INLINE] = ACTIONS(7307),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7307),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7307),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7307),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7307),
    [anon_sym_const] = ACTIONS(7307),
    [anon_sym_constexpr] = ACTIONS(7307),
    [anon_sym_volatile] = ACTIONS(7307),
    [anon_sym_restrict] = ACTIONS(7307),
    [anon_sym___restrict__] = ACTIONS(7307),
    [anon_sym__Atomic] = ACTIONS(7307),
    [anon_sym__Noreturn] = ACTIONS(7307),
    [anon_sym__Nonnull] = ACTIONS(7307),
    [anon_sym_nullable] = ACTIONS(7307),
    [anon_sym__Complex] = ACTIONS(7307),
    [anon_sym__Nullable] = ACTIONS(7307),
    [anon_sym__Nullable_result] = ACTIONS(7307),
    [anon_sym__Null_unspecified] = ACTIONS(7307),
    [anon_sym___autoreleasing] = ACTIONS(7307),
    [anon_sym___block] = ACTIONS(7307),
    [anon_sym___bridge] = ACTIONS(7307),
    [anon_sym___bridge_retained] = ACTIONS(7307),
    [anon_sym___bridge_transfer] = ACTIONS(7307),
    [anon_sym___complex] = ACTIONS(7307),
    [anon_sym___const] = ACTIONS(7307),
    [anon_sym___imag] = ACTIONS(7307),
    [anon_sym___kindof] = ACTIONS(7307),
    [anon_sym___nonnull] = ACTIONS(7307),
    [anon_sym___nullable] = ACTIONS(7307),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7307),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7307),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7307),
    [anon_sym___real] = ACTIONS(7307),
    [anon_sym___strong] = ACTIONS(7307),
    [anon_sym___unsafe_unretained] = ACTIONS(7307),
    [anon_sym___unused] = ACTIONS(7307),
    [anon_sym___weak] = ACTIONS(7307),
    [anon_sym_alignas] = ACTIONS(7307),
    [anon_sym__Alignas] = ACTIONS(7307),
    [anon_sym_enum] = ACTIONS(7307),
    [anon_sym_COLON] = ACTIONS(7309),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7307),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7307),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7307),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7307),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7307),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7307),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7307),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7307),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7307),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7307),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7307),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7307),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7307),
    [anon_sym_API_AVAILABLE] = ACTIONS(7307),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7307),
    [anon_sym_API_DEPRECATED] = ACTIONS(7307),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7307),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7307),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7307),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7307),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7307),
    [anon_sym___deprecated_msg] = ACTIONS(7307),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7307),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7307),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7307),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7307),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7307),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7307),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7307),
  },
  [STATE(2968)] = {
    [sym_identifier] = ACTIONS(7311),
    [anon_sym_COMMA] = ACTIONS(7313),
    [anon_sym_RPAREN] = ACTIONS(7313),
    [anon_sym_LPAREN2] = ACTIONS(7313),
    [anon_sym_STAR] = ACTIONS(7313),
    [anon_sym_CARET] = ACTIONS(7313),
    [anon_sym_SEMI] = ACTIONS(7313),
    [anon_sym___extension__] = ACTIONS(7311),
    [anon_sym_extern] = ACTIONS(7311),
    [anon_sym___attribute__] = ACTIONS(7311),
    [anon_sym___attribute] = ACTIONS(7311),
    [anon_sym_noreturn] = ACTIONS(7311),
    [anon_sym_LBRACK] = ACTIONS(7313),
    [anon_sym___declspec] = ACTIONS(7311),
    [anon_sym___based] = ACTIONS(7311),
    [anon_sym_signed] = ACTIONS(7311),
    [anon_sym_unsigned] = ACTIONS(7311),
    [anon_sym_long] = ACTIONS(7311),
    [anon_sym_short] = ACTIONS(7311),
    [anon_sym_static] = ACTIONS(7311),
    [anon_sym_auto] = ACTIONS(7311),
    [anon_sym_register] = ACTIONS(7311),
    [anon_sym_inline] = ACTIONS(7311),
    [anon_sym___inline] = ACTIONS(7311),
    [anon_sym___inline__] = ACTIONS(7311),
    [anon_sym___forceinline] = ACTIONS(7311),
    [anon_sym_thread_local] = ACTIONS(7311),
    [anon_sym___thread] = ACTIONS(7311),
    [anon_sym_CG_EXTERN] = ACTIONS(7311),
    [anon_sym_CG_INLINE] = ACTIONS(7311),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7311),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7311),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7311),
    [anon_sym_IBOutlet] = ACTIONS(7311),
    [anon_sym_IBInspectable] = ACTIONS(7311),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7311),
    [anon_sym_NS_INLINE] = ACTIONS(7311),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7311),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7311),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7311),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7311),
    [anon_sym_const] = ACTIONS(7311),
    [anon_sym_constexpr] = ACTIONS(7311),
    [anon_sym_volatile] = ACTIONS(7311),
    [anon_sym_restrict] = ACTIONS(7311),
    [anon_sym___restrict__] = ACTIONS(7311),
    [anon_sym__Atomic] = ACTIONS(7311),
    [anon_sym__Noreturn] = ACTIONS(7311),
    [anon_sym__Nonnull] = ACTIONS(7311),
    [anon_sym_nullable] = ACTIONS(7311),
    [anon_sym__Complex] = ACTIONS(7311),
    [anon_sym__Nullable] = ACTIONS(7311),
    [anon_sym__Nullable_result] = ACTIONS(7311),
    [anon_sym__Null_unspecified] = ACTIONS(7311),
    [anon_sym___autoreleasing] = ACTIONS(7311),
    [anon_sym___block] = ACTIONS(7311),
    [anon_sym___bridge] = ACTIONS(7311),
    [anon_sym___bridge_retained] = ACTIONS(7311),
    [anon_sym___bridge_transfer] = ACTIONS(7311),
    [anon_sym___complex] = ACTIONS(7311),
    [anon_sym___const] = ACTIONS(7311),
    [anon_sym___imag] = ACTIONS(7311),
    [anon_sym___kindof] = ACTIONS(7311),
    [anon_sym___nonnull] = ACTIONS(7311),
    [anon_sym___nullable] = ACTIONS(7311),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7311),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7311),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7311),
    [anon_sym___real] = ACTIONS(7311),
    [anon_sym___strong] = ACTIONS(7311),
    [anon_sym___unsafe_unretained] = ACTIONS(7311),
    [anon_sym___unused] = ACTIONS(7311),
    [anon_sym___weak] = ACTIONS(7311),
    [anon_sym_alignas] = ACTIONS(7311),
    [anon_sym__Alignas] = ACTIONS(7311),
    [sym_primitive_type] = ACTIONS(7311),
    [anon_sym_enum] = ACTIONS(7311),
    [anon_sym_COLON] = ACTIONS(7313),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7311),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7311),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7311),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7311),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7311),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7311),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7311),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7311),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7311),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7311),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7311),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7311),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7311),
    [anon_sym_API_AVAILABLE] = ACTIONS(7311),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7311),
    [anon_sym_API_DEPRECATED] = ACTIONS(7311),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7311),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7311),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7311),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7311),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7311),
    [anon_sym___deprecated_msg] = ACTIONS(7311),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7311),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7311),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7311),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7311),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7311),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7311),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7311),
  },
  [STATE(2969)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2985),
    [sym_identifier] = ACTIONS(7315),
    [anon_sym_COMMA] = ACTIONS(7317),
    [anon_sym_RPAREN] = ACTIONS(7317),
    [anon_sym_LPAREN2] = ACTIONS(7317),
    [anon_sym_STAR] = ACTIONS(7317),
    [anon_sym_CARET] = ACTIONS(7317),
    [anon_sym_SEMI] = ACTIONS(7317),
    [anon_sym___extension__] = ACTIONS(7315),
    [anon_sym_extern] = ACTIONS(7315),
    [anon_sym___attribute__] = ACTIONS(7315),
    [anon_sym___attribute] = ACTIONS(7315),
    [anon_sym_noreturn] = ACTIONS(7315),
    [anon_sym_LBRACK] = ACTIONS(7317),
    [anon_sym___declspec] = ACTIONS(7315),
    [anon_sym___based] = ACTIONS(7315),
    [anon_sym_signed] = ACTIONS(7319),
    [anon_sym_unsigned] = ACTIONS(7319),
    [anon_sym_long] = ACTIONS(7319),
    [anon_sym_short] = ACTIONS(7319),
    [anon_sym_static] = ACTIONS(7315),
    [anon_sym_auto] = ACTIONS(7315),
    [anon_sym_register] = ACTIONS(7315),
    [anon_sym_inline] = ACTIONS(7315),
    [anon_sym___inline] = ACTIONS(7315),
    [anon_sym___inline__] = ACTIONS(7315),
    [anon_sym___forceinline] = ACTIONS(7315),
    [anon_sym_thread_local] = ACTIONS(7315),
    [anon_sym___thread] = ACTIONS(7315),
    [anon_sym_CG_EXTERN] = ACTIONS(7315),
    [anon_sym_CG_INLINE] = ACTIONS(7315),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7315),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7315),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7315),
    [anon_sym_IBOutlet] = ACTIONS(7315),
    [anon_sym_IBInspectable] = ACTIONS(7315),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7315),
    [anon_sym_NS_INLINE] = ACTIONS(7315),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7315),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7315),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7315),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7315),
    [anon_sym_const] = ACTIONS(7315),
    [anon_sym_constexpr] = ACTIONS(7315),
    [anon_sym_volatile] = ACTIONS(7315),
    [anon_sym_restrict] = ACTIONS(7315),
    [anon_sym___restrict__] = ACTIONS(7315),
    [anon_sym__Atomic] = ACTIONS(7315),
    [anon_sym__Noreturn] = ACTIONS(7315),
    [anon_sym__Nonnull] = ACTIONS(7315),
    [anon_sym_nullable] = ACTIONS(7315),
    [anon_sym__Complex] = ACTIONS(7315),
    [anon_sym__Nullable] = ACTIONS(7315),
    [anon_sym__Nullable_result] = ACTIONS(7315),
    [anon_sym__Null_unspecified] = ACTIONS(7315),
    [anon_sym___autoreleasing] = ACTIONS(7315),
    [anon_sym___block] = ACTIONS(7315),
    [anon_sym___bridge] = ACTIONS(7315),
    [anon_sym___bridge_retained] = ACTIONS(7315),
    [anon_sym___bridge_transfer] = ACTIONS(7315),
    [anon_sym___complex] = ACTIONS(7315),
    [anon_sym___const] = ACTIONS(7315),
    [anon_sym___imag] = ACTIONS(7315),
    [anon_sym___kindof] = ACTIONS(7315),
    [anon_sym___nonnull] = ACTIONS(7315),
    [anon_sym___nullable] = ACTIONS(7315),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7315),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7315),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7315),
    [anon_sym___real] = ACTIONS(7315),
    [anon_sym___strong] = ACTIONS(7315),
    [anon_sym___unsafe_unretained] = ACTIONS(7315),
    [anon_sym___unused] = ACTIONS(7315),
    [anon_sym___weak] = ACTIONS(7315),
    [anon_sym_alignas] = ACTIONS(7315),
    [anon_sym__Alignas] = ACTIONS(7315),
    [anon_sym_enum] = ACTIONS(7315),
    [anon_sym_COLON] = ACTIONS(7317),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7315),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7315),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7315),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7315),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7315),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7315),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7315),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7315),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7315),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7315),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7315),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7315),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7315),
    [anon_sym_API_AVAILABLE] = ACTIONS(7315),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7315),
    [anon_sym_API_DEPRECATED] = ACTIONS(7315),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7315),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7315),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7315),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7315),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7315),
    [anon_sym___deprecated_msg] = ACTIONS(7315),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7315),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7315),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7315),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7315),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7315),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7315),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7315),
  },
  [STATE(2970)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym__old_style_parameter_list_repeat1] = STATE(5701),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_COMMA] = ACTIONS(7321),
    [anon_sym_RPAREN] = ACTIONS(7324),
    [anon_sym_LPAREN2] = ACTIONS(7042),
    [anon_sym_STAR] = ACTIONS(6134),
    [anon_sym_CARET] = ACTIONS(6134),
    [anon_sym_LT] = ACTIONS(7045),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(6134),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(2971)] = {
    [sym_identifier] = ACTIONS(5074),
    [anon_sym_COMMA] = ACTIONS(5070),
    [anon_sym_RPAREN] = ACTIONS(5070),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym_SEMI] = ACTIONS(5070),
    [anon_sym___extension__] = ACTIONS(5074),
    [anon_sym_extern] = ACTIONS(5074),
    [anon_sym___attribute__] = ACTIONS(5074),
    [anon_sym___attribute] = ACTIONS(5074),
    [anon_sym_noreturn] = ACTIONS(5074),
    [anon_sym_LBRACK] = ACTIONS(5070),
    [anon_sym___declspec] = ACTIONS(5074),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym_signed] = ACTIONS(5074),
    [anon_sym_unsigned] = ACTIONS(5074),
    [anon_sym_long] = ACTIONS(5074),
    [anon_sym_short] = ACTIONS(5074),
    [anon_sym_static] = ACTIONS(5074),
    [anon_sym_auto] = ACTIONS(5074),
    [anon_sym_register] = ACTIONS(5074),
    [anon_sym_inline] = ACTIONS(5074),
    [anon_sym___inline] = ACTIONS(5074),
    [anon_sym___inline__] = ACTIONS(5074),
    [anon_sym___forceinline] = ACTIONS(5074),
    [anon_sym_thread_local] = ACTIONS(5074),
    [anon_sym___thread] = ACTIONS(5074),
    [anon_sym_CG_EXTERN] = ACTIONS(5074),
    [anon_sym_CG_INLINE] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5074),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5074),
    [anon_sym_IBOutlet] = ACTIONS(5074),
    [anon_sym_IBInspectable] = ACTIONS(5074),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5074),
    [anon_sym_NS_INLINE] = ACTIONS(5074),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5074),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5074),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5074),
    [anon_sym_const] = ACTIONS(5074),
    [anon_sym_constexpr] = ACTIONS(5074),
    [anon_sym_volatile] = ACTIONS(5074),
    [anon_sym_restrict] = ACTIONS(5074),
    [anon_sym___restrict__] = ACTIONS(5074),
    [anon_sym__Atomic] = ACTIONS(5074),
    [anon_sym__Noreturn] = ACTIONS(5074),
    [anon_sym__Nonnull] = ACTIONS(5074),
    [anon_sym_nullable] = ACTIONS(5074),
    [anon_sym__Complex] = ACTIONS(5074),
    [anon_sym__Nullable] = ACTIONS(5074),
    [anon_sym__Nullable_result] = ACTIONS(5074),
    [anon_sym__Null_unspecified] = ACTIONS(5074),
    [anon_sym___autoreleasing] = ACTIONS(5074),
    [anon_sym___block] = ACTIONS(5074),
    [anon_sym___bridge] = ACTIONS(5074),
    [anon_sym___bridge_retained] = ACTIONS(5074),
    [anon_sym___bridge_transfer] = ACTIONS(5074),
    [anon_sym___complex] = ACTIONS(5074),
    [anon_sym___const] = ACTIONS(5074),
    [anon_sym___imag] = ACTIONS(5074),
    [anon_sym___kindof] = ACTIONS(5074),
    [anon_sym___nonnull] = ACTIONS(5074),
    [anon_sym___nullable] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5074),
    [anon_sym___real] = ACTIONS(5074),
    [anon_sym___strong] = ACTIONS(5074),
    [anon_sym___unsafe_unretained] = ACTIONS(5074),
    [anon_sym___unused] = ACTIONS(5074),
    [anon_sym___weak] = ACTIONS(5074),
    [anon_sym_alignas] = ACTIONS(5074),
    [anon_sym__Alignas] = ACTIONS(5074),
    [sym_primitive_type] = ACTIONS(5074),
    [anon_sym_enum] = ACTIONS(5074),
    [anon_sym_COLON] = ACTIONS(5070),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5074),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5074),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5074),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_API_AVAILABLE] = ACTIONS(5074),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_API_DEPRECATED] = ACTIONS(5074),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym___deprecated_msg] = ACTIONS(5074),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5074),
  },
  [STATE(2972)] = {
    [sym_identifier] = ACTIONS(7327),
    [anon_sym_COMMA] = ACTIONS(7329),
    [anon_sym_RPAREN] = ACTIONS(7329),
    [anon_sym_LPAREN2] = ACTIONS(7329),
    [anon_sym_STAR] = ACTIONS(7329),
    [anon_sym_CARET] = ACTIONS(7329),
    [anon_sym_SEMI] = ACTIONS(7329),
    [anon_sym___extension__] = ACTIONS(7327),
    [anon_sym_extern] = ACTIONS(7327),
    [anon_sym___attribute__] = ACTIONS(7327),
    [anon_sym___attribute] = ACTIONS(7327),
    [anon_sym_noreturn] = ACTIONS(7327),
    [anon_sym_LBRACK] = ACTIONS(7329),
    [anon_sym___declspec] = ACTIONS(7327),
    [anon_sym___based] = ACTIONS(7327),
    [anon_sym_signed] = ACTIONS(7327),
    [anon_sym_unsigned] = ACTIONS(7327),
    [anon_sym_long] = ACTIONS(7327),
    [anon_sym_short] = ACTIONS(7327),
    [anon_sym_static] = ACTIONS(7327),
    [anon_sym_auto] = ACTIONS(7327),
    [anon_sym_register] = ACTIONS(7327),
    [anon_sym_inline] = ACTIONS(7327),
    [anon_sym___inline] = ACTIONS(7327),
    [anon_sym___inline__] = ACTIONS(7327),
    [anon_sym___forceinline] = ACTIONS(7327),
    [anon_sym_thread_local] = ACTIONS(7327),
    [anon_sym___thread] = ACTIONS(7327),
    [anon_sym_CG_EXTERN] = ACTIONS(7327),
    [anon_sym_CG_INLINE] = ACTIONS(7327),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7327),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7327),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7327),
    [anon_sym_IBOutlet] = ACTIONS(7327),
    [anon_sym_IBInspectable] = ACTIONS(7327),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7327),
    [anon_sym_NS_INLINE] = ACTIONS(7327),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7327),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7327),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7327),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7327),
    [anon_sym_const] = ACTIONS(7327),
    [anon_sym_constexpr] = ACTIONS(7327),
    [anon_sym_volatile] = ACTIONS(7327),
    [anon_sym_restrict] = ACTIONS(7327),
    [anon_sym___restrict__] = ACTIONS(7327),
    [anon_sym__Atomic] = ACTIONS(7327),
    [anon_sym__Noreturn] = ACTIONS(7327),
    [anon_sym__Nonnull] = ACTIONS(7327),
    [anon_sym_nullable] = ACTIONS(7327),
    [anon_sym__Complex] = ACTIONS(7327),
    [anon_sym__Nullable] = ACTIONS(7327),
    [anon_sym__Nullable_result] = ACTIONS(7327),
    [anon_sym__Null_unspecified] = ACTIONS(7327),
    [anon_sym___autoreleasing] = ACTIONS(7327),
    [anon_sym___block] = ACTIONS(7327),
    [anon_sym___bridge] = ACTIONS(7327),
    [anon_sym___bridge_retained] = ACTIONS(7327),
    [anon_sym___bridge_transfer] = ACTIONS(7327),
    [anon_sym___complex] = ACTIONS(7327),
    [anon_sym___const] = ACTIONS(7327),
    [anon_sym___imag] = ACTIONS(7327),
    [anon_sym___kindof] = ACTIONS(7327),
    [anon_sym___nonnull] = ACTIONS(7327),
    [anon_sym___nullable] = ACTIONS(7327),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7327),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7327),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7327),
    [anon_sym___real] = ACTIONS(7327),
    [anon_sym___strong] = ACTIONS(7327),
    [anon_sym___unsafe_unretained] = ACTIONS(7327),
    [anon_sym___unused] = ACTIONS(7327),
    [anon_sym___weak] = ACTIONS(7327),
    [anon_sym_alignas] = ACTIONS(7327),
    [anon_sym__Alignas] = ACTIONS(7327),
    [sym_primitive_type] = ACTIONS(7327),
    [anon_sym_enum] = ACTIONS(7327),
    [anon_sym_COLON] = ACTIONS(7329),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7327),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7327),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7327),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7327),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7327),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7327),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7327),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7327),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7327),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7327),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7327),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7327),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7327),
    [anon_sym_API_AVAILABLE] = ACTIONS(7327),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7327),
    [anon_sym_API_DEPRECATED] = ACTIONS(7327),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7327),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7327),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7327),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7327),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7327),
    [anon_sym___deprecated_msg] = ACTIONS(7327),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7327),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7327),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7327),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7327),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7327),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7327),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7327),
  },
  [STATE(2973)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2999),
    [sym_identifier] = ACTIONS(5074),
    [anon_sym_COMMA] = ACTIONS(5070),
    [anon_sym_RPAREN] = ACTIONS(5070),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym_SEMI] = ACTIONS(5070),
    [anon_sym___extension__] = ACTIONS(5074),
    [anon_sym_extern] = ACTIONS(5074),
    [anon_sym___attribute__] = ACTIONS(5074),
    [anon_sym___attribute] = ACTIONS(5074),
    [anon_sym_noreturn] = ACTIONS(5074),
    [anon_sym_LBRACK] = ACTIONS(5070),
    [anon_sym___declspec] = ACTIONS(5074),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym_signed] = ACTIONS(7331),
    [anon_sym_unsigned] = ACTIONS(7331),
    [anon_sym_long] = ACTIONS(7331),
    [anon_sym_short] = ACTIONS(7331),
    [anon_sym_static] = ACTIONS(5074),
    [anon_sym_auto] = ACTIONS(5074),
    [anon_sym_register] = ACTIONS(5074),
    [anon_sym_inline] = ACTIONS(5074),
    [anon_sym___inline] = ACTIONS(5074),
    [anon_sym___inline__] = ACTIONS(5074),
    [anon_sym___forceinline] = ACTIONS(5074),
    [anon_sym_thread_local] = ACTIONS(5074),
    [anon_sym___thread] = ACTIONS(5074),
    [anon_sym_CG_EXTERN] = ACTIONS(5074),
    [anon_sym_CG_INLINE] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5074),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5074),
    [anon_sym_IBOutlet] = ACTIONS(5074),
    [anon_sym_IBInspectable] = ACTIONS(5074),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5074),
    [anon_sym_NS_INLINE] = ACTIONS(5074),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5074),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5074),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5074),
    [anon_sym_const] = ACTIONS(5074),
    [anon_sym_constexpr] = ACTIONS(5074),
    [anon_sym_volatile] = ACTIONS(5074),
    [anon_sym_restrict] = ACTIONS(5074),
    [anon_sym___restrict__] = ACTIONS(5074),
    [anon_sym__Atomic] = ACTIONS(5074),
    [anon_sym__Noreturn] = ACTIONS(5074),
    [anon_sym__Nonnull] = ACTIONS(5074),
    [anon_sym_nullable] = ACTIONS(5074),
    [anon_sym__Complex] = ACTIONS(5074),
    [anon_sym__Nullable] = ACTIONS(5074),
    [anon_sym__Nullable_result] = ACTIONS(5074),
    [anon_sym__Null_unspecified] = ACTIONS(5074),
    [anon_sym___autoreleasing] = ACTIONS(5074),
    [anon_sym___block] = ACTIONS(5074),
    [anon_sym___bridge] = ACTIONS(5074),
    [anon_sym___bridge_retained] = ACTIONS(5074),
    [anon_sym___bridge_transfer] = ACTIONS(5074),
    [anon_sym___complex] = ACTIONS(5074),
    [anon_sym___const] = ACTIONS(5074),
    [anon_sym___imag] = ACTIONS(5074),
    [anon_sym___kindof] = ACTIONS(5074),
    [anon_sym___nonnull] = ACTIONS(5074),
    [anon_sym___nullable] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5074),
    [anon_sym___real] = ACTIONS(5074),
    [anon_sym___strong] = ACTIONS(5074),
    [anon_sym___unsafe_unretained] = ACTIONS(5074),
    [anon_sym___unused] = ACTIONS(5074),
    [anon_sym___weak] = ACTIONS(5074),
    [anon_sym_alignas] = ACTIONS(5074),
    [anon_sym__Alignas] = ACTIONS(5074),
    [anon_sym_enum] = ACTIONS(5074),
    [anon_sym_COLON] = ACTIONS(5070),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5074),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5074),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5074),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_API_AVAILABLE] = ACTIONS(5074),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_API_DEPRECATED] = ACTIONS(5074),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym___deprecated_msg] = ACTIONS(5074),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5074),
  },
  [STATE(2974)] = {
    [sym_identifier] = ACTIONS(7333),
    [anon_sym_COMMA] = ACTIONS(7335),
    [anon_sym_RPAREN] = ACTIONS(7335),
    [anon_sym_LPAREN2] = ACTIONS(7335),
    [anon_sym_STAR] = ACTIONS(7335),
    [anon_sym_CARET] = ACTIONS(7335),
    [anon_sym_SEMI] = ACTIONS(7335),
    [anon_sym___extension__] = ACTIONS(7333),
    [anon_sym_extern] = ACTIONS(7333),
    [anon_sym___attribute__] = ACTIONS(7333),
    [anon_sym___attribute] = ACTIONS(7333),
    [anon_sym_noreturn] = ACTIONS(7333),
    [anon_sym_LBRACK] = ACTIONS(7335),
    [anon_sym___declspec] = ACTIONS(7333),
    [anon_sym___based] = ACTIONS(7333),
    [anon_sym_signed] = ACTIONS(7333),
    [anon_sym_unsigned] = ACTIONS(7333),
    [anon_sym_long] = ACTIONS(7333),
    [anon_sym_short] = ACTIONS(7333),
    [anon_sym_static] = ACTIONS(7333),
    [anon_sym_auto] = ACTIONS(7333),
    [anon_sym_register] = ACTIONS(7333),
    [anon_sym_inline] = ACTIONS(7333),
    [anon_sym___inline] = ACTIONS(7333),
    [anon_sym___inline__] = ACTIONS(7333),
    [anon_sym___forceinline] = ACTIONS(7333),
    [anon_sym_thread_local] = ACTIONS(7333),
    [anon_sym___thread] = ACTIONS(7333),
    [anon_sym_CG_EXTERN] = ACTIONS(7333),
    [anon_sym_CG_INLINE] = ACTIONS(7333),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7333),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7333),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7333),
    [anon_sym_IBOutlet] = ACTIONS(7333),
    [anon_sym_IBInspectable] = ACTIONS(7333),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7333),
    [anon_sym_NS_INLINE] = ACTIONS(7333),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7333),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7333),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7333),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7333),
    [anon_sym_const] = ACTIONS(7333),
    [anon_sym_constexpr] = ACTIONS(7333),
    [anon_sym_volatile] = ACTIONS(7333),
    [anon_sym_restrict] = ACTIONS(7333),
    [anon_sym___restrict__] = ACTIONS(7333),
    [anon_sym__Atomic] = ACTIONS(7333),
    [anon_sym__Noreturn] = ACTIONS(7333),
    [anon_sym__Nonnull] = ACTIONS(7333),
    [anon_sym_nullable] = ACTIONS(7333),
    [anon_sym__Complex] = ACTIONS(7333),
    [anon_sym__Nullable] = ACTIONS(7333),
    [anon_sym__Nullable_result] = ACTIONS(7333),
    [anon_sym__Null_unspecified] = ACTIONS(7333),
    [anon_sym___autoreleasing] = ACTIONS(7333),
    [anon_sym___block] = ACTIONS(7333),
    [anon_sym___bridge] = ACTIONS(7333),
    [anon_sym___bridge_retained] = ACTIONS(7333),
    [anon_sym___bridge_transfer] = ACTIONS(7333),
    [anon_sym___complex] = ACTIONS(7333),
    [anon_sym___const] = ACTIONS(7333),
    [anon_sym___imag] = ACTIONS(7333),
    [anon_sym___kindof] = ACTIONS(7333),
    [anon_sym___nonnull] = ACTIONS(7333),
    [anon_sym___nullable] = ACTIONS(7333),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7333),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7333),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7333),
    [anon_sym___real] = ACTIONS(7333),
    [anon_sym___strong] = ACTIONS(7333),
    [anon_sym___unsafe_unretained] = ACTIONS(7333),
    [anon_sym___unused] = ACTIONS(7333),
    [anon_sym___weak] = ACTIONS(7333),
    [anon_sym_alignas] = ACTIONS(7333),
    [anon_sym__Alignas] = ACTIONS(7333),
    [sym_primitive_type] = ACTIONS(7333),
    [anon_sym_enum] = ACTIONS(7333),
    [anon_sym_COLON] = ACTIONS(7335),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7333),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7333),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7333),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7333),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7333),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7333),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7333),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7333),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7333),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7333),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7333),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7333),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7333),
    [anon_sym_API_AVAILABLE] = ACTIONS(7333),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7333),
    [anon_sym_API_DEPRECATED] = ACTIONS(7333),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7333),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7333),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7333),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7333),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7333),
    [anon_sym___deprecated_msg] = ACTIONS(7333),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7333),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7333),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7333),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7333),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7333),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7333),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7333),
  },
  [STATE(2975)] = {
    [sym_identifier] = ACTIONS(7337),
    [anon_sym_COMMA] = ACTIONS(7339),
    [anon_sym_RPAREN] = ACTIONS(7339),
    [anon_sym_LPAREN2] = ACTIONS(7339),
    [anon_sym_STAR] = ACTIONS(7339),
    [anon_sym_CARET] = ACTIONS(7339),
    [anon_sym_SEMI] = ACTIONS(7339),
    [anon_sym___extension__] = ACTIONS(7337),
    [anon_sym_extern] = ACTIONS(7337),
    [anon_sym___attribute__] = ACTIONS(7337),
    [anon_sym___attribute] = ACTIONS(7337),
    [anon_sym_noreturn] = ACTIONS(7337),
    [anon_sym_LBRACK] = ACTIONS(7339),
    [anon_sym___declspec] = ACTIONS(7337),
    [anon_sym___based] = ACTIONS(7337),
    [anon_sym_signed] = ACTIONS(7337),
    [anon_sym_unsigned] = ACTIONS(7337),
    [anon_sym_long] = ACTIONS(7337),
    [anon_sym_short] = ACTIONS(7337),
    [anon_sym_static] = ACTIONS(7337),
    [anon_sym_auto] = ACTIONS(7337),
    [anon_sym_register] = ACTIONS(7337),
    [anon_sym_inline] = ACTIONS(7337),
    [anon_sym___inline] = ACTIONS(7337),
    [anon_sym___inline__] = ACTIONS(7337),
    [anon_sym___forceinline] = ACTIONS(7337),
    [anon_sym_thread_local] = ACTIONS(7337),
    [anon_sym___thread] = ACTIONS(7337),
    [anon_sym_CG_EXTERN] = ACTIONS(7337),
    [anon_sym_CG_INLINE] = ACTIONS(7337),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7337),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7337),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7337),
    [anon_sym_IBOutlet] = ACTIONS(7337),
    [anon_sym_IBInspectable] = ACTIONS(7337),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7337),
    [anon_sym_NS_INLINE] = ACTIONS(7337),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7337),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7337),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7337),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7337),
    [anon_sym_const] = ACTIONS(7337),
    [anon_sym_constexpr] = ACTIONS(7337),
    [anon_sym_volatile] = ACTIONS(7337),
    [anon_sym_restrict] = ACTIONS(7337),
    [anon_sym___restrict__] = ACTIONS(7337),
    [anon_sym__Atomic] = ACTIONS(7337),
    [anon_sym__Noreturn] = ACTIONS(7337),
    [anon_sym__Nonnull] = ACTIONS(7337),
    [anon_sym_nullable] = ACTIONS(7337),
    [anon_sym__Complex] = ACTIONS(7337),
    [anon_sym__Nullable] = ACTIONS(7337),
    [anon_sym__Nullable_result] = ACTIONS(7337),
    [anon_sym__Null_unspecified] = ACTIONS(7337),
    [anon_sym___autoreleasing] = ACTIONS(7337),
    [anon_sym___block] = ACTIONS(7337),
    [anon_sym___bridge] = ACTIONS(7337),
    [anon_sym___bridge_retained] = ACTIONS(7337),
    [anon_sym___bridge_transfer] = ACTIONS(7337),
    [anon_sym___complex] = ACTIONS(7337),
    [anon_sym___const] = ACTIONS(7337),
    [anon_sym___imag] = ACTIONS(7337),
    [anon_sym___kindof] = ACTIONS(7337),
    [anon_sym___nonnull] = ACTIONS(7337),
    [anon_sym___nullable] = ACTIONS(7337),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7337),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7337),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7337),
    [anon_sym___real] = ACTIONS(7337),
    [anon_sym___strong] = ACTIONS(7337),
    [anon_sym___unsafe_unretained] = ACTIONS(7337),
    [anon_sym___unused] = ACTIONS(7337),
    [anon_sym___weak] = ACTIONS(7337),
    [anon_sym_alignas] = ACTIONS(7337),
    [anon_sym__Alignas] = ACTIONS(7337),
    [sym_primitive_type] = ACTIONS(7337),
    [anon_sym_enum] = ACTIONS(7337),
    [anon_sym_COLON] = ACTIONS(7339),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7337),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7337),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7337),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7337),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7337),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7337),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7337),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7337),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7337),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7337),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7337),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7337),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7337),
    [anon_sym_API_AVAILABLE] = ACTIONS(7337),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7337),
    [anon_sym_API_DEPRECATED] = ACTIONS(7337),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7337),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7337),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7337),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7337),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7337),
    [anon_sym___deprecated_msg] = ACTIONS(7337),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7337),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7337),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7337),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7337),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7337),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7337),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7337),
  },
  [STATE(2976)] = {
    [sym_identifier] = ACTIONS(7341),
    [anon_sym_COMMA] = ACTIONS(7343),
    [anon_sym_RPAREN] = ACTIONS(7343),
    [anon_sym_LPAREN2] = ACTIONS(7343),
    [anon_sym_STAR] = ACTIONS(7343),
    [anon_sym_CARET] = ACTIONS(7343),
    [anon_sym_SEMI] = ACTIONS(7343),
    [anon_sym___extension__] = ACTIONS(7341),
    [anon_sym_extern] = ACTIONS(7341),
    [anon_sym___attribute__] = ACTIONS(7341),
    [anon_sym___attribute] = ACTIONS(7341),
    [anon_sym_noreturn] = ACTIONS(7341),
    [anon_sym_LBRACK] = ACTIONS(7343),
    [anon_sym___declspec] = ACTIONS(7341),
    [anon_sym___based] = ACTIONS(7341),
    [anon_sym_signed] = ACTIONS(7341),
    [anon_sym_unsigned] = ACTIONS(7341),
    [anon_sym_long] = ACTIONS(7341),
    [anon_sym_short] = ACTIONS(7341),
    [anon_sym_static] = ACTIONS(7341),
    [anon_sym_auto] = ACTIONS(7341),
    [anon_sym_register] = ACTIONS(7341),
    [anon_sym_inline] = ACTIONS(7341),
    [anon_sym___inline] = ACTIONS(7341),
    [anon_sym___inline__] = ACTIONS(7341),
    [anon_sym___forceinline] = ACTIONS(7341),
    [anon_sym_thread_local] = ACTIONS(7341),
    [anon_sym___thread] = ACTIONS(7341),
    [anon_sym_CG_EXTERN] = ACTIONS(7341),
    [anon_sym_CG_INLINE] = ACTIONS(7341),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7341),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7341),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7341),
    [anon_sym_IBOutlet] = ACTIONS(7341),
    [anon_sym_IBInspectable] = ACTIONS(7341),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7341),
    [anon_sym_NS_INLINE] = ACTIONS(7341),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7341),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7341),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7341),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7341),
    [anon_sym_const] = ACTIONS(7341),
    [anon_sym_constexpr] = ACTIONS(7341),
    [anon_sym_volatile] = ACTIONS(7341),
    [anon_sym_restrict] = ACTIONS(7341),
    [anon_sym___restrict__] = ACTIONS(7341),
    [anon_sym__Atomic] = ACTIONS(7341),
    [anon_sym__Noreturn] = ACTIONS(7341),
    [anon_sym__Nonnull] = ACTIONS(7341),
    [anon_sym_nullable] = ACTIONS(7341),
    [anon_sym__Complex] = ACTIONS(7341),
    [anon_sym__Nullable] = ACTIONS(7341),
    [anon_sym__Nullable_result] = ACTIONS(7341),
    [anon_sym__Null_unspecified] = ACTIONS(7341),
    [anon_sym___autoreleasing] = ACTIONS(7341),
    [anon_sym___block] = ACTIONS(7341),
    [anon_sym___bridge] = ACTIONS(7341),
    [anon_sym___bridge_retained] = ACTIONS(7341),
    [anon_sym___bridge_transfer] = ACTIONS(7341),
    [anon_sym___complex] = ACTIONS(7341),
    [anon_sym___const] = ACTIONS(7341),
    [anon_sym___imag] = ACTIONS(7341),
    [anon_sym___kindof] = ACTIONS(7341),
    [anon_sym___nonnull] = ACTIONS(7341),
    [anon_sym___nullable] = ACTIONS(7341),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7341),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7341),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7341),
    [anon_sym___real] = ACTIONS(7341),
    [anon_sym___strong] = ACTIONS(7341),
    [anon_sym___unsafe_unretained] = ACTIONS(7341),
    [anon_sym___unused] = ACTIONS(7341),
    [anon_sym___weak] = ACTIONS(7341),
    [anon_sym_alignas] = ACTIONS(7341),
    [anon_sym__Alignas] = ACTIONS(7341),
    [sym_primitive_type] = ACTIONS(7341),
    [anon_sym_enum] = ACTIONS(7341),
    [anon_sym_COLON] = ACTIONS(7343),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7341),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7341),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7341),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7341),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7341),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7341),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7341),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7341),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7341),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7341),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7341),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7341),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7341),
    [anon_sym_API_AVAILABLE] = ACTIONS(7341),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7341),
    [anon_sym_API_DEPRECATED] = ACTIONS(7341),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7341),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7341),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7341),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7341),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7341),
    [anon_sym___deprecated_msg] = ACTIONS(7341),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7341),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7341),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7341),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7341),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7341),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7341),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7341),
  },
  [STATE(2977)] = {
    [sym_identifier] = ACTIONS(7345),
    [anon_sym_COMMA] = ACTIONS(7347),
    [anon_sym_RPAREN] = ACTIONS(7347),
    [anon_sym_LPAREN2] = ACTIONS(7347),
    [anon_sym_STAR] = ACTIONS(7347),
    [anon_sym_CARET] = ACTIONS(7347),
    [anon_sym_SEMI] = ACTIONS(7347),
    [anon_sym___extension__] = ACTIONS(7345),
    [anon_sym_extern] = ACTIONS(7345),
    [anon_sym___attribute__] = ACTIONS(7345),
    [anon_sym___attribute] = ACTIONS(7345),
    [anon_sym_noreturn] = ACTIONS(7345),
    [anon_sym_LBRACK] = ACTIONS(7347),
    [anon_sym___declspec] = ACTIONS(7345),
    [anon_sym___based] = ACTIONS(7345),
    [anon_sym_signed] = ACTIONS(7345),
    [anon_sym_unsigned] = ACTIONS(7345),
    [anon_sym_long] = ACTIONS(7345),
    [anon_sym_short] = ACTIONS(7345),
    [anon_sym_static] = ACTIONS(7345),
    [anon_sym_auto] = ACTIONS(7345),
    [anon_sym_register] = ACTIONS(7345),
    [anon_sym_inline] = ACTIONS(7345),
    [anon_sym___inline] = ACTIONS(7345),
    [anon_sym___inline__] = ACTIONS(7345),
    [anon_sym___forceinline] = ACTIONS(7345),
    [anon_sym_thread_local] = ACTIONS(7345),
    [anon_sym___thread] = ACTIONS(7345),
    [anon_sym_CG_EXTERN] = ACTIONS(7345),
    [anon_sym_CG_INLINE] = ACTIONS(7345),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7345),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7345),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7345),
    [anon_sym_IBOutlet] = ACTIONS(7345),
    [anon_sym_IBInspectable] = ACTIONS(7345),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7345),
    [anon_sym_NS_INLINE] = ACTIONS(7345),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7345),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7345),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7345),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7345),
    [anon_sym_const] = ACTIONS(7345),
    [anon_sym_constexpr] = ACTIONS(7345),
    [anon_sym_volatile] = ACTIONS(7345),
    [anon_sym_restrict] = ACTIONS(7345),
    [anon_sym___restrict__] = ACTIONS(7345),
    [anon_sym__Atomic] = ACTIONS(7345),
    [anon_sym__Noreturn] = ACTIONS(7345),
    [anon_sym__Nonnull] = ACTIONS(7345),
    [anon_sym_nullable] = ACTIONS(7345),
    [anon_sym__Complex] = ACTIONS(7345),
    [anon_sym__Nullable] = ACTIONS(7345),
    [anon_sym__Nullable_result] = ACTIONS(7345),
    [anon_sym__Null_unspecified] = ACTIONS(7345),
    [anon_sym___autoreleasing] = ACTIONS(7345),
    [anon_sym___block] = ACTIONS(7345),
    [anon_sym___bridge] = ACTIONS(7345),
    [anon_sym___bridge_retained] = ACTIONS(7345),
    [anon_sym___bridge_transfer] = ACTIONS(7345),
    [anon_sym___complex] = ACTIONS(7345),
    [anon_sym___const] = ACTIONS(7345),
    [anon_sym___imag] = ACTIONS(7345),
    [anon_sym___kindof] = ACTIONS(7345),
    [anon_sym___nonnull] = ACTIONS(7345),
    [anon_sym___nullable] = ACTIONS(7345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7345),
    [anon_sym___real] = ACTIONS(7345),
    [anon_sym___strong] = ACTIONS(7345),
    [anon_sym___unsafe_unretained] = ACTIONS(7345),
    [anon_sym___unused] = ACTIONS(7345),
    [anon_sym___weak] = ACTIONS(7345),
    [anon_sym_alignas] = ACTIONS(7345),
    [anon_sym__Alignas] = ACTIONS(7345),
    [sym_primitive_type] = ACTIONS(7345),
    [anon_sym_enum] = ACTIONS(7345),
    [anon_sym_COLON] = ACTIONS(7347),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7345),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7345),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7345),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7345),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7345),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7345),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7345),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7345),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7345),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7345),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7345),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7345),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7345),
    [anon_sym_API_AVAILABLE] = ACTIONS(7345),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7345),
    [anon_sym_API_DEPRECATED] = ACTIONS(7345),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7345),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7345),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7345),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7345),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7345),
    [anon_sym___deprecated_msg] = ACTIONS(7345),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7345),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7345),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7345),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7345),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7345),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7345),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7345),
  },
  [STATE(2978)] = {
    [sym_identifier] = ACTIONS(7349),
    [anon_sym_COMMA] = ACTIONS(7351),
    [anon_sym_RPAREN] = ACTIONS(7351),
    [anon_sym_LPAREN2] = ACTIONS(7351),
    [anon_sym_STAR] = ACTIONS(7351),
    [anon_sym_CARET] = ACTIONS(7351),
    [anon_sym_SEMI] = ACTIONS(7351),
    [anon_sym___extension__] = ACTIONS(7349),
    [anon_sym_extern] = ACTIONS(7349),
    [anon_sym___attribute__] = ACTIONS(7349),
    [anon_sym___attribute] = ACTIONS(7349),
    [anon_sym_noreturn] = ACTIONS(7349),
    [anon_sym_LBRACK] = ACTIONS(7351),
    [anon_sym___declspec] = ACTIONS(7349),
    [anon_sym___based] = ACTIONS(7349),
    [anon_sym_signed] = ACTIONS(7349),
    [anon_sym_unsigned] = ACTIONS(7349),
    [anon_sym_long] = ACTIONS(7349),
    [anon_sym_short] = ACTIONS(7349),
    [anon_sym_static] = ACTIONS(7349),
    [anon_sym_auto] = ACTIONS(7349),
    [anon_sym_register] = ACTIONS(7349),
    [anon_sym_inline] = ACTIONS(7349),
    [anon_sym___inline] = ACTIONS(7349),
    [anon_sym___inline__] = ACTIONS(7349),
    [anon_sym___forceinline] = ACTIONS(7349),
    [anon_sym_thread_local] = ACTIONS(7349),
    [anon_sym___thread] = ACTIONS(7349),
    [anon_sym_CG_EXTERN] = ACTIONS(7349),
    [anon_sym_CG_INLINE] = ACTIONS(7349),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7349),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7349),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7349),
    [anon_sym_IBOutlet] = ACTIONS(7349),
    [anon_sym_IBInspectable] = ACTIONS(7349),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7349),
    [anon_sym_NS_INLINE] = ACTIONS(7349),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7349),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7349),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7349),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7349),
    [anon_sym_const] = ACTIONS(7349),
    [anon_sym_constexpr] = ACTIONS(7349),
    [anon_sym_volatile] = ACTIONS(7349),
    [anon_sym_restrict] = ACTIONS(7349),
    [anon_sym___restrict__] = ACTIONS(7349),
    [anon_sym__Atomic] = ACTIONS(7349),
    [anon_sym__Noreturn] = ACTIONS(7349),
    [anon_sym__Nonnull] = ACTIONS(7349),
    [anon_sym_nullable] = ACTIONS(7349),
    [anon_sym__Complex] = ACTIONS(7349),
    [anon_sym__Nullable] = ACTIONS(7349),
    [anon_sym__Nullable_result] = ACTIONS(7349),
    [anon_sym__Null_unspecified] = ACTIONS(7349),
    [anon_sym___autoreleasing] = ACTIONS(7349),
    [anon_sym___block] = ACTIONS(7349),
    [anon_sym___bridge] = ACTIONS(7349),
    [anon_sym___bridge_retained] = ACTIONS(7349),
    [anon_sym___bridge_transfer] = ACTIONS(7349),
    [anon_sym___complex] = ACTIONS(7349),
    [anon_sym___const] = ACTIONS(7349),
    [anon_sym___imag] = ACTIONS(7349),
    [anon_sym___kindof] = ACTIONS(7349),
    [anon_sym___nonnull] = ACTIONS(7349),
    [anon_sym___nullable] = ACTIONS(7349),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7349),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7349),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7349),
    [anon_sym___real] = ACTIONS(7349),
    [anon_sym___strong] = ACTIONS(7349),
    [anon_sym___unsafe_unretained] = ACTIONS(7349),
    [anon_sym___unused] = ACTIONS(7349),
    [anon_sym___weak] = ACTIONS(7349),
    [anon_sym_alignas] = ACTIONS(7349),
    [anon_sym__Alignas] = ACTIONS(7349),
    [sym_primitive_type] = ACTIONS(7349),
    [anon_sym_enum] = ACTIONS(7349),
    [anon_sym_COLON] = ACTIONS(7351),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7349),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7349),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7349),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7349),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7349),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7349),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7349),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7349),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7349),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7349),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7349),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7349),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7349),
    [anon_sym_API_AVAILABLE] = ACTIONS(7349),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7349),
    [anon_sym_API_DEPRECATED] = ACTIONS(7349),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7349),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7349),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7349),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7349),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7349),
    [anon_sym___deprecated_msg] = ACTIONS(7349),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7349),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7349),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7349),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7349),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7349),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7349),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7349),
  },
  [STATE(2979)] = {
    [sym_identifier] = ACTIONS(7353),
    [anon_sym_COMMA] = ACTIONS(7355),
    [anon_sym_RPAREN] = ACTIONS(7355),
    [anon_sym_LPAREN2] = ACTIONS(7355),
    [anon_sym_STAR] = ACTIONS(7355),
    [anon_sym_CARET] = ACTIONS(7355),
    [anon_sym_SEMI] = ACTIONS(7355),
    [anon_sym___extension__] = ACTIONS(7353),
    [anon_sym_extern] = ACTIONS(7353),
    [anon_sym___attribute__] = ACTIONS(7353),
    [anon_sym___attribute] = ACTIONS(7353),
    [anon_sym_noreturn] = ACTIONS(7353),
    [anon_sym_LBRACK] = ACTIONS(7355),
    [anon_sym___declspec] = ACTIONS(7353),
    [anon_sym___based] = ACTIONS(7353),
    [anon_sym_signed] = ACTIONS(7353),
    [anon_sym_unsigned] = ACTIONS(7353),
    [anon_sym_long] = ACTIONS(7353),
    [anon_sym_short] = ACTIONS(7353),
    [anon_sym_static] = ACTIONS(7353),
    [anon_sym_auto] = ACTIONS(7353),
    [anon_sym_register] = ACTIONS(7353),
    [anon_sym_inline] = ACTIONS(7353),
    [anon_sym___inline] = ACTIONS(7353),
    [anon_sym___inline__] = ACTIONS(7353),
    [anon_sym___forceinline] = ACTIONS(7353),
    [anon_sym_thread_local] = ACTIONS(7353),
    [anon_sym___thread] = ACTIONS(7353),
    [anon_sym_CG_EXTERN] = ACTIONS(7353),
    [anon_sym_CG_INLINE] = ACTIONS(7353),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7353),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7353),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7353),
    [anon_sym_IBOutlet] = ACTIONS(7353),
    [anon_sym_IBInspectable] = ACTIONS(7353),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7353),
    [anon_sym_NS_INLINE] = ACTIONS(7353),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7353),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7353),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7353),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7353),
    [anon_sym_const] = ACTIONS(7353),
    [anon_sym_constexpr] = ACTIONS(7353),
    [anon_sym_volatile] = ACTIONS(7353),
    [anon_sym_restrict] = ACTIONS(7353),
    [anon_sym___restrict__] = ACTIONS(7353),
    [anon_sym__Atomic] = ACTIONS(7353),
    [anon_sym__Noreturn] = ACTIONS(7353),
    [anon_sym__Nonnull] = ACTIONS(7353),
    [anon_sym_nullable] = ACTIONS(7353),
    [anon_sym__Complex] = ACTIONS(7353),
    [anon_sym__Nullable] = ACTIONS(7353),
    [anon_sym__Nullable_result] = ACTIONS(7353),
    [anon_sym__Null_unspecified] = ACTIONS(7353),
    [anon_sym___autoreleasing] = ACTIONS(7353),
    [anon_sym___block] = ACTIONS(7353),
    [anon_sym___bridge] = ACTIONS(7353),
    [anon_sym___bridge_retained] = ACTIONS(7353),
    [anon_sym___bridge_transfer] = ACTIONS(7353),
    [anon_sym___complex] = ACTIONS(7353),
    [anon_sym___const] = ACTIONS(7353),
    [anon_sym___imag] = ACTIONS(7353),
    [anon_sym___kindof] = ACTIONS(7353),
    [anon_sym___nonnull] = ACTIONS(7353),
    [anon_sym___nullable] = ACTIONS(7353),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7353),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7353),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7353),
    [anon_sym___real] = ACTIONS(7353),
    [anon_sym___strong] = ACTIONS(7353),
    [anon_sym___unsafe_unretained] = ACTIONS(7353),
    [anon_sym___unused] = ACTIONS(7353),
    [anon_sym___weak] = ACTIONS(7353),
    [anon_sym_alignas] = ACTIONS(7353),
    [anon_sym__Alignas] = ACTIONS(7353),
    [sym_primitive_type] = ACTIONS(7353),
    [anon_sym_enum] = ACTIONS(7353),
    [anon_sym_COLON] = ACTIONS(7355),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7353),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7353),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7353),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7353),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7353),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7353),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7353),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7353),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7353),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7353),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7353),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7353),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7353),
    [anon_sym_API_AVAILABLE] = ACTIONS(7353),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7353),
    [anon_sym_API_DEPRECATED] = ACTIONS(7353),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7353),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7353),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7353),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7353),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7353),
    [anon_sym___deprecated_msg] = ACTIONS(7353),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7353),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7353),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7353),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7353),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7353),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7353),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7353),
  },
  [STATE(2980)] = {
    [sym_identifier] = ACTIONS(7357),
    [anon_sym_COMMA] = ACTIONS(7359),
    [anon_sym_RPAREN] = ACTIONS(7359),
    [anon_sym_LPAREN2] = ACTIONS(7359),
    [anon_sym_STAR] = ACTIONS(7359),
    [anon_sym_CARET] = ACTIONS(7359),
    [anon_sym_SEMI] = ACTIONS(7359),
    [anon_sym___extension__] = ACTIONS(7357),
    [anon_sym_extern] = ACTIONS(7357),
    [anon_sym___attribute__] = ACTIONS(7357),
    [anon_sym___attribute] = ACTIONS(7357),
    [anon_sym_noreturn] = ACTIONS(7357),
    [anon_sym_LBRACK] = ACTIONS(7359),
    [anon_sym___declspec] = ACTIONS(7357),
    [anon_sym___based] = ACTIONS(7357),
    [anon_sym_signed] = ACTIONS(7357),
    [anon_sym_unsigned] = ACTIONS(7357),
    [anon_sym_long] = ACTIONS(7357),
    [anon_sym_short] = ACTIONS(7357),
    [anon_sym_static] = ACTIONS(7357),
    [anon_sym_auto] = ACTIONS(7357),
    [anon_sym_register] = ACTIONS(7357),
    [anon_sym_inline] = ACTIONS(7357),
    [anon_sym___inline] = ACTIONS(7357),
    [anon_sym___inline__] = ACTIONS(7357),
    [anon_sym___forceinline] = ACTIONS(7357),
    [anon_sym_thread_local] = ACTIONS(7357),
    [anon_sym___thread] = ACTIONS(7357),
    [anon_sym_CG_EXTERN] = ACTIONS(7357),
    [anon_sym_CG_INLINE] = ACTIONS(7357),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7357),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7357),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7357),
    [anon_sym_IBOutlet] = ACTIONS(7357),
    [anon_sym_IBInspectable] = ACTIONS(7357),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7357),
    [anon_sym_NS_INLINE] = ACTIONS(7357),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7357),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7357),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7357),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7357),
    [anon_sym_const] = ACTIONS(7357),
    [anon_sym_constexpr] = ACTIONS(7357),
    [anon_sym_volatile] = ACTIONS(7357),
    [anon_sym_restrict] = ACTIONS(7357),
    [anon_sym___restrict__] = ACTIONS(7357),
    [anon_sym__Atomic] = ACTIONS(7357),
    [anon_sym__Noreturn] = ACTIONS(7357),
    [anon_sym__Nonnull] = ACTIONS(7357),
    [anon_sym_nullable] = ACTIONS(7357),
    [anon_sym__Complex] = ACTIONS(7357),
    [anon_sym__Nullable] = ACTIONS(7357),
    [anon_sym__Nullable_result] = ACTIONS(7357),
    [anon_sym__Null_unspecified] = ACTIONS(7357),
    [anon_sym___autoreleasing] = ACTIONS(7357),
    [anon_sym___block] = ACTIONS(7357),
    [anon_sym___bridge] = ACTIONS(7357),
    [anon_sym___bridge_retained] = ACTIONS(7357),
    [anon_sym___bridge_transfer] = ACTIONS(7357),
    [anon_sym___complex] = ACTIONS(7357),
    [anon_sym___const] = ACTIONS(7357),
    [anon_sym___imag] = ACTIONS(7357),
    [anon_sym___kindof] = ACTIONS(7357),
    [anon_sym___nonnull] = ACTIONS(7357),
    [anon_sym___nullable] = ACTIONS(7357),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7357),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7357),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7357),
    [anon_sym___real] = ACTIONS(7357),
    [anon_sym___strong] = ACTIONS(7357),
    [anon_sym___unsafe_unretained] = ACTIONS(7357),
    [anon_sym___unused] = ACTIONS(7357),
    [anon_sym___weak] = ACTIONS(7357),
    [anon_sym_alignas] = ACTIONS(7357),
    [anon_sym__Alignas] = ACTIONS(7357),
    [sym_primitive_type] = ACTIONS(7357),
    [anon_sym_enum] = ACTIONS(7357),
    [anon_sym_COLON] = ACTIONS(7359),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7357),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7357),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7357),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7357),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7357),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7357),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7357),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7357),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7357),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7357),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7357),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7357),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7357),
    [anon_sym_API_AVAILABLE] = ACTIONS(7357),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7357),
    [anon_sym_API_DEPRECATED] = ACTIONS(7357),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7357),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7357),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7357),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7357),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7357),
    [anon_sym___deprecated_msg] = ACTIONS(7357),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7357),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7357),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7357),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7357),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7357),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7357),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7357),
  },
  [STATE(2981)] = {
    [sym_identifier] = ACTIONS(7361),
    [anon_sym_COMMA] = ACTIONS(7363),
    [anon_sym_RPAREN] = ACTIONS(7363),
    [anon_sym_LPAREN2] = ACTIONS(7363),
    [anon_sym_STAR] = ACTIONS(7363),
    [anon_sym_CARET] = ACTIONS(7363),
    [anon_sym_SEMI] = ACTIONS(7363),
    [anon_sym___extension__] = ACTIONS(7361),
    [anon_sym_extern] = ACTIONS(7361),
    [anon_sym___attribute__] = ACTIONS(7361),
    [anon_sym___attribute] = ACTIONS(7361),
    [anon_sym_noreturn] = ACTIONS(7361),
    [anon_sym_LBRACK] = ACTIONS(7363),
    [anon_sym___declspec] = ACTIONS(7361),
    [anon_sym___based] = ACTIONS(7361),
    [anon_sym_signed] = ACTIONS(7361),
    [anon_sym_unsigned] = ACTIONS(7361),
    [anon_sym_long] = ACTIONS(7361),
    [anon_sym_short] = ACTIONS(7361),
    [anon_sym_static] = ACTIONS(7361),
    [anon_sym_auto] = ACTIONS(7361),
    [anon_sym_register] = ACTIONS(7361),
    [anon_sym_inline] = ACTIONS(7361),
    [anon_sym___inline] = ACTIONS(7361),
    [anon_sym___inline__] = ACTIONS(7361),
    [anon_sym___forceinline] = ACTIONS(7361),
    [anon_sym_thread_local] = ACTIONS(7361),
    [anon_sym___thread] = ACTIONS(7361),
    [anon_sym_CG_EXTERN] = ACTIONS(7361),
    [anon_sym_CG_INLINE] = ACTIONS(7361),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7361),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7361),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7361),
    [anon_sym_IBOutlet] = ACTIONS(7361),
    [anon_sym_IBInspectable] = ACTIONS(7361),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7361),
    [anon_sym_NS_INLINE] = ACTIONS(7361),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7361),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7361),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7361),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7361),
    [anon_sym_const] = ACTIONS(7361),
    [anon_sym_constexpr] = ACTIONS(7361),
    [anon_sym_volatile] = ACTIONS(7361),
    [anon_sym_restrict] = ACTIONS(7361),
    [anon_sym___restrict__] = ACTIONS(7361),
    [anon_sym__Atomic] = ACTIONS(7361),
    [anon_sym__Noreturn] = ACTIONS(7361),
    [anon_sym__Nonnull] = ACTIONS(7361),
    [anon_sym_nullable] = ACTIONS(7361),
    [anon_sym__Complex] = ACTIONS(7361),
    [anon_sym__Nullable] = ACTIONS(7361),
    [anon_sym__Nullable_result] = ACTIONS(7361),
    [anon_sym__Null_unspecified] = ACTIONS(7361),
    [anon_sym___autoreleasing] = ACTIONS(7361),
    [anon_sym___block] = ACTIONS(7361),
    [anon_sym___bridge] = ACTIONS(7361),
    [anon_sym___bridge_retained] = ACTIONS(7361),
    [anon_sym___bridge_transfer] = ACTIONS(7361),
    [anon_sym___complex] = ACTIONS(7361),
    [anon_sym___const] = ACTIONS(7361),
    [anon_sym___imag] = ACTIONS(7361),
    [anon_sym___kindof] = ACTIONS(7361),
    [anon_sym___nonnull] = ACTIONS(7361),
    [anon_sym___nullable] = ACTIONS(7361),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7361),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7361),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7361),
    [anon_sym___real] = ACTIONS(7361),
    [anon_sym___strong] = ACTIONS(7361),
    [anon_sym___unsafe_unretained] = ACTIONS(7361),
    [anon_sym___unused] = ACTIONS(7361),
    [anon_sym___weak] = ACTIONS(7361),
    [anon_sym_alignas] = ACTIONS(7361),
    [anon_sym__Alignas] = ACTIONS(7361),
    [sym_primitive_type] = ACTIONS(7361),
    [anon_sym_enum] = ACTIONS(7361),
    [anon_sym_COLON] = ACTIONS(7363),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7361),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7361),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7361),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7361),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7361),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7361),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7361),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7361),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7361),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7361),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7361),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7361),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7361),
    [anon_sym_API_AVAILABLE] = ACTIONS(7361),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7361),
    [anon_sym_API_DEPRECATED] = ACTIONS(7361),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7361),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7361),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7361),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7361),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7361),
    [anon_sym___deprecated_msg] = ACTIONS(7361),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7361),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7361),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7361),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7361),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7361),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7361),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7361),
  },
  [STATE(2982)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2912),
    [sym_identifier] = ACTIONS(7365),
    [anon_sym_COMMA] = ACTIONS(7367),
    [anon_sym_RPAREN] = ACTIONS(7367),
    [anon_sym_LPAREN2] = ACTIONS(7367),
    [anon_sym_STAR] = ACTIONS(7367),
    [anon_sym_CARET] = ACTIONS(7367),
    [anon_sym_SEMI] = ACTIONS(7367),
    [anon_sym___extension__] = ACTIONS(7365),
    [anon_sym_extern] = ACTIONS(7365),
    [anon_sym___attribute__] = ACTIONS(7365),
    [anon_sym___attribute] = ACTIONS(7365),
    [anon_sym_noreturn] = ACTIONS(7365),
    [anon_sym_LBRACK] = ACTIONS(7367),
    [anon_sym___declspec] = ACTIONS(7365),
    [anon_sym___based] = ACTIONS(7365),
    [anon_sym_signed] = ACTIONS(7287),
    [anon_sym_unsigned] = ACTIONS(7287),
    [anon_sym_long] = ACTIONS(7287),
    [anon_sym_short] = ACTIONS(7287),
    [anon_sym_static] = ACTIONS(7365),
    [anon_sym_auto] = ACTIONS(7365),
    [anon_sym_register] = ACTIONS(7365),
    [anon_sym_inline] = ACTIONS(7365),
    [anon_sym___inline] = ACTIONS(7365),
    [anon_sym___inline__] = ACTIONS(7365),
    [anon_sym___forceinline] = ACTIONS(7365),
    [anon_sym_thread_local] = ACTIONS(7365),
    [anon_sym___thread] = ACTIONS(7365),
    [anon_sym_CG_EXTERN] = ACTIONS(7365),
    [anon_sym_CG_INLINE] = ACTIONS(7365),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7365),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7365),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7365),
    [anon_sym_IBOutlet] = ACTIONS(7365),
    [anon_sym_IBInspectable] = ACTIONS(7365),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7365),
    [anon_sym_NS_INLINE] = ACTIONS(7365),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7365),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7365),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7365),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7365),
    [anon_sym_const] = ACTIONS(7365),
    [anon_sym_constexpr] = ACTIONS(7365),
    [anon_sym_volatile] = ACTIONS(7365),
    [anon_sym_restrict] = ACTIONS(7365),
    [anon_sym___restrict__] = ACTIONS(7365),
    [anon_sym__Atomic] = ACTIONS(7365),
    [anon_sym__Noreturn] = ACTIONS(7365),
    [anon_sym__Nonnull] = ACTIONS(7365),
    [anon_sym_nullable] = ACTIONS(7365),
    [anon_sym__Complex] = ACTIONS(7365),
    [anon_sym__Nullable] = ACTIONS(7365),
    [anon_sym__Nullable_result] = ACTIONS(7365),
    [anon_sym__Null_unspecified] = ACTIONS(7365),
    [anon_sym___autoreleasing] = ACTIONS(7365),
    [anon_sym___block] = ACTIONS(7365),
    [anon_sym___bridge] = ACTIONS(7365),
    [anon_sym___bridge_retained] = ACTIONS(7365),
    [anon_sym___bridge_transfer] = ACTIONS(7365),
    [anon_sym___complex] = ACTIONS(7365),
    [anon_sym___const] = ACTIONS(7365),
    [anon_sym___imag] = ACTIONS(7365),
    [anon_sym___kindof] = ACTIONS(7365),
    [anon_sym___nonnull] = ACTIONS(7365),
    [anon_sym___nullable] = ACTIONS(7365),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7365),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7365),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7365),
    [anon_sym___real] = ACTIONS(7365),
    [anon_sym___strong] = ACTIONS(7365),
    [anon_sym___unsafe_unretained] = ACTIONS(7365),
    [anon_sym___unused] = ACTIONS(7365),
    [anon_sym___weak] = ACTIONS(7365),
    [anon_sym_alignas] = ACTIONS(7365),
    [anon_sym__Alignas] = ACTIONS(7365),
    [anon_sym_enum] = ACTIONS(7365),
    [anon_sym_COLON] = ACTIONS(7367),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7365),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7365),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7365),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7365),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7365),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7365),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7365),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7365),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7365),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7365),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7365),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7365),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7365),
    [anon_sym_API_AVAILABLE] = ACTIONS(7365),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7365),
    [anon_sym_API_DEPRECATED] = ACTIONS(7365),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7365),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7365),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7365),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7365),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7365),
    [anon_sym___deprecated_msg] = ACTIONS(7365),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7365),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7365),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7365),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7365),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7365),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7365),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7365),
  },
  [STATE(2983)] = {
    [sym_identifier] = ACTIONS(7369),
    [anon_sym_COMMA] = ACTIONS(7371),
    [anon_sym_RPAREN] = ACTIONS(7371),
    [anon_sym_LPAREN2] = ACTIONS(7371),
    [anon_sym_STAR] = ACTIONS(7371),
    [anon_sym_CARET] = ACTIONS(7371),
    [anon_sym_SEMI] = ACTIONS(7371),
    [anon_sym___extension__] = ACTIONS(7369),
    [anon_sym_extern] = ACTIONS(7369),
    [anon_sym___attribute__] = ACTIONS(7369),
    [anon_sym___attribute] = ACTIONS(7369),
    [anon_sym_noreturn] = ACTIONS(7369),
    [anon_sym_LBRACK] = ACTIONS(7371),
    [anon_sym___declspec] = ACTIONS(7369),
    [anon_sym___based] = ACTIONS(7369),
    [anon_sym_signed] = ACTIONS(7369),
    [anon_sym_unsigned] = ACTIONS(7369),
    [anon_sym_long] = ACTIONS(7369),
    [anon_sym_short] = ACTIONS(7369),
    [anon_sym_static] = ACTIONS(7369),
    [anon_sym_auto] = ACTIONS(7369),
    [anon_sym_register] = ACTIONS(7369),
    [anon_sym_inline] = ACTIONS(7369),
    [anon_sym___inline] = ACTIONS(7369),
    [anon_sym___inline__] = ACTIONS(7369),
    [anon_sym___forceinline] = ACTIONS(7369),
    [anon_sym_thread_local] = ACTIONS(7369),
    [anon_sym___thread] = ACTIONS(7369),
    [anon_sym_CG_EXTERN] = ACTIONS(7369),
    [anon_sym_CG_INLINE] = ACTIONS(7369),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7369),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7369),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7369),
    [anon_sym_IBOutlet] = ACTIONS(7369),
    [anon_sym_IBInspectable] = ACTIONS(7369),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7369),
    [anon_sym_NS_INLINE] = ACTIONS(7369),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7369),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7369),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7369),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7369),
    [anon_sym_const] = ACTIONS(7369),
    [anon_sym_constexpr] = ACTIONS(7369),
    [anon_sym_volatile] = ACTIONS(7369),
    [anon_sym_restrict] = ACTIONS(7369),
    [anon_sym___restrict__] = ACTIONS(7369),
    [anon_sym__Atomic] = ACTIONS(7369),
    [anon_sym__Noreturn] = ACTIONS(7369),
    [anon_sym__Nonnull] = ACTIONS(7369),
    [anon_sym_nullable] = ACTIONS(7369),
    [anon_sym__Complex] = ACTIONS(7369),
    [anon_sym__Nullable] = ACTIONS(7369),
    [anon_sym__Nullable_result] = ACTIONS(7369),
    [anon_sym__Null_unspecified] = ACTIONS(7369),
    [anon_sym___autoreleasing] = ACTIONS(7369),
    [anon_sym___block] = ACTIONS(7369),
    [anon_sym___bridge] = ACTIONS(7369),
    [anon_sym___bridge_retained] = ACTIONS(7369),
    [anon_sym___bridge_transfer] = ACTIONS(7369),
    [anon_sym___complex] = ACTIONS(7369),
    [anon_sym___const] = ACTIONS(7369),
    [anon_sym___imag] = ACTIONS(7369),
    [anon_sym___kindof] = ACTIONS(7369),
    [anon_sym___nonnull] = ACTIONS(7369),
    [anon_sym___nullable] = ACTIONS(7369),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7369),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7369),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7369),
    [anon_sym___real] = ACTIONS(7369),
    [anon_sym___strong] = ACTIONS(7369),
    [anon_sym___unsafe_unretained] = ACTIONS(7369),
    [anon_sym___unused] = ACTIONS(7369),
    [anon_sym___weak] = ACTIONS(7369),
    [anon_sym_alignas] = ACTIONS(7369),
    [anon_sym__Alignas] = ACTIONS(7369),
    [sym_primitive_type] = ACTIONS(7369),
    [anon_sym_enum] = ACTIONS(7369),
    [anon_sym_COLON] = ACTIONS(7371),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7369),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7369),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7369),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7369),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7369),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7369),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7369),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7369),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7369),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7369),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7369),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7369),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7369),
    [anon_sym_API_AVAILABLE] = ACTIONS(7369),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7369),
    [anon_sym_API_DEPRECATED] = ACTIONS(7369),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7369),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7369),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7369),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7369),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7369),
    [anon_sym___deprecated_msg] = ACTIONS(7369),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7369),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7369),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7369),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7369),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7369),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7369),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7369),
  },
  [STATE(2984)] = {
    [sym_identifier] = ACTIONS(7373),
    [anon_sym_COMMA] = ACTIONS(7375),
    [anon_sym_RPAREN] = ACTIONS(7375),
    [anon_sym_LPAREN2] = ACTIONS(7375),
    [anon_sym_STAR] = ACTIONS(7375),
    [anon_sym_CARET] = ACTIONS(7375),
    [anon_sym_SEMI] = ACTIONS(7375),
    [anon_sym___extension__] = ACTIONS(7373),
    [anon_sym_extern] = ACTIONS(7373),
    [anon_sym___attribute__] = ACTIONS(7373),
    [anon_sym___attribute] = ACTIONS(7373),
    [anon_sym_noreturn] = ACTIONS(7373),
    [anon_sym_LBRACK] = ACTIONS(7375),
    [anon_sym___declspec] = ACTIONS(7373),
    [anon_sym___based] = ACTIONS(7373),
    [anon_sym_signed] = ACTIONS(7373),
    [anon_sym_unsigned] = ACTIONS(7373),
    [anon_sym_long] = ACTIONS(7373),
    [anon_sym_short] = ACTIONS(7373),
    [anon_sym_static] = ACTIONS(7373),
    [anon_sym_auto] = ACTIONS(7373),
    [anon_sym_register] = ACTIONS(7373),
    [anon_sym_inline] = ACTIONS(7373),
    [anon_sym___inline] = ACTIONS(7373),
    [anon_sym___inline__] = ACTIONS(7373),
    [anon_sym___forceinline] = ACTIONS(7373),
    [anon_sym_thread_local] = ACTIONS(7373),
    [anon_sym___thread] = ACTIONS(7373),
    [anon_sym_CG_EXTERN] = ACTIONS(7373),
    [anon_sym_CG_INLINE] = ACTIONS(7373),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7373),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7373),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7373),
    [anon_sym_IBOutlet] = ACTIONS(7373),
    [anon_sym_IBInspectable] = ACTIONS(7373),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7373),
    [anon_sym_NS_INLINE] = ACTIONS(7373),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7373),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7373),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7373),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7373),
    [anon_sym_const] = ACTIONS(7373),
    [anon_sym_constexpr] = ACTIONS(7373),
    [anon_sym_volatile] = ACTIONS(7373),
    [anon_sym_restrict] = ACTIONS(7373),
    [anon_sym___restrict__] = ACTIONS(7373),
    [anon_sym__Atomic] = ACTIONS(7373),
    [anon_sym__Noreturn] = ACTIONS(7373),
    [anon_sym__Nonnull] = ACTIONS(7373),
    [anon_sym_nullable] = ACTIONS(7373),
    [anon_sym__Complex] = ACTIONS(7373),
    [anon_sym__Nullable] = ACTIONS(7373),
    [anon_sym__Nullable_result] = ACTIONS(7373),
    [anon_sym__Null_unspecified] = ACTIONS(7373),
    [anon_sym___autoreleasing] = ACTIONS(7373),
    [anon_sym___block] = ACTIONS(7373),
    [anon_sym___bridge] = ACTIONS(7373),
    [anon_sym___bridge_retained] = ACTIONS(7373),
    [anon_sym___bridge_transfer] = ACTIONS(7373),
    [anon_sym___complex] = ACTIONS(7373),
    [anon_sym___const] = ACTIONS(7373),
    [anon_sym___imag] = ACTIONS(7373),
    [anon_sym___kindof] = ACTIONS(7373),
    [anon_sym___nonnull] = ACTIONS(7373),
    [anon_sym___nullable] = ACTIONS(7373),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7373),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7373),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7373),
    [anon_sym___real] = ACTIONS(7373),
    [anon_sym___strong] = ACTIONS(7373),
    [anon_sym___unsafe_unretained] = ACTIONS(7373),
    [anon_sym___unused] = ACTIONS(7373),
    [anon_sym___weak] = ACTIONS(7373),
    [anon_sym_alignas] = ACTIONS(7373),
    [anon_sym__Alignas] = ACTIONS(7373),
    [sym_primitive_type] = ACTIONS(7373),
    [anon_sym_enum] = ACTIONS(7373),
    [anon_sym_COLON] = ACTIONS(7375),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7373),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7373),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7373),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7373),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7373),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7373),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7373),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7373),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7373),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7373),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7373),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7373),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7373),
    [anon_sym_API_AVAILABLE] = ACTIONS(7373),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7373),
    [anon_sym_API_DEPRECATED] = ACTIONS(7373),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7373),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7373),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7373),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7373),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7373),
    [anon_sym___deprecated_msg] = ACTIONS(7373),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7373),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7373),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7373),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7373),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7373),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7373),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7373),
  },
  [STATE(2985)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2912),
    [sym_identifier] = ACTIONS(7377),
    [anon_sym_COMMA] = ACTIONS(7379),
    [anon_sym_RPAREN] = ACTIONS(7379),
    [anon_sym_LPAREN2] = ACTIONS(7379),
    [anon_sym_STAR] = ACTIONS(7379),
    [anon_sym_CARET] = ACTIONS(7379),
    [anon_sym_SEMI] = ACTIONS(7379),
    [anon_sym___extension__] = ACTIONS(7377),
    [anon_sym_extern] = ACTIONS(7377),
    [anon_sym___attribute__] = ACTIONS(7377),
    [anon_sym___attribute] = ACTIONS(7377),
    [anon_sym_noreturn] = ACTIONS(7377),
    [anon_sym_LBRACK] = ACTIONS(7379),
    [anon_sym___declspec] = ACTIONS(7377),
    [anon_sym___based] = ACTIONS(7377),
    [anon_sym_signed] = ACTIONS(7287),
    [anon_sym_unsigned] = ACTIONS(7287),
    [anon_sym_long] = ACTIONS(7287),
    [anon_sym_short] = ACTIONS(7287),
    [anon_sym_static] = ACTIONS(7377),
    [anon_sym_auto] = ACTIONS(7377),
    [anon_sym_register] = ACTIONS(7377),
    [anon_sym_inline] = ACTIONS(7377),
    [anon_sym___inline] = ACTIONS(7377),
    [anon_sym___inline__] = ACTIONS(7377),
    [anon_sym___forceinline] = ACTIONS(7377),
    [anon_sym_thread_local] = ACTIONS(7377),
    [anon_sym___thread] = ACTIONS(7377),
    [anon_sym_CG_EXTERN] = ACTIONS(7377),
    [anon_sym_CG_INLINE] = ACTIONS(7377),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7377),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7377),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7377),
    [anon_sym_IBOutlet] = ACTIONS(7377),
    [anon_sym_IBInspectable] = ACTIONS(7377),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7377),
    [anon_sym_NS_INLINE] = ACTIONS(7377),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7377),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7377),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7377),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7377),
    [anon_sym_const] = ACTIONS(7377),
    [anon_sym_constexpr] = ACTIONS(7377),
    [anon_sym_volatile] = ACTIONS(7377),
    [anon_sym_restrict] = ACTIONS(7377),
    [anon_sym___restrict__] = ACTIONS(7377),
    [anon_sym__Atomic] = ACTIONS(7377),
    [anon_sym__Noreturn] = ACTIONS(7377),
    [anon_sym__Nonnull] = ACTIONS(7377),
    [anon_sym_nullable] = ACTIONS(7377),
    [anon_sym__Complex] = ACTIONS(7377),
    [anon_sym__Nullable] = ACTIONS(7377),
    [anon_sym__Nullable_result] = ACTIONS(7377),
    [anon_sym__Null_unspecified] = ACTIONS(7377),
    [anon_sym___autoreleasing] = ACTIONS(7377),
    [anon_sym___block] = ACTIONS(7377),
    [anon_sym___bridge] = ACTIONS(7377),
    [anon_sym___bridge_retained] = ACTIONS(7377),
    [anon_sym___bridge_transfer] = ACTIONS(7377),
    [anon_sym___complex] = ACTIONS(7377),
    [anon_sym___const] = ACTIONS(7377),
    [anon_sym___imag] = ACTIONS(7377),
    [anon_sym___kindof] = ACTIONS(7377),
    [anon_sym___nonnull] = ACTIONS(7377),
    [anon_sym___nullable] = ACTIONS(7377),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7377),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7377),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7377),
    [anon_sym___real] = ACTIONS(7377),
    [anon_sym___strong] = ACTIONS(7377),
    [anon_sym___unsafe_unretained] = ACTIONS(7377),
    [anon_sym___unused] = ACTIONS(7377),
    [anon_sym___weak] = ACTIONS(7377),
    [anon_sym_alignas] = ACTIONS(7377),
    [anon_sym__Alignas] = ACTIONS(7377),
    [anon_sym_enum] = ACTIONS(7377),
    [anon_sym_COLON] = ACTIONS(7379),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7377),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7377),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7377),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7377),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7377),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7377),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7377),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7377),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7377),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7377),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7377),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7377),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7377),
    [anon_sym_API_AVAILABLE] = ACTIONS(7377),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7377),
    [anon_sym_API_DEPRECATED] = ACTIONS(7377),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7377),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7377),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7377),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7377),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7377),
    [anon_sym___deprecated_msg] = ACTIONS(7377),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7377),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7377),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7377),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7377),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7377),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7377),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7377),
  },
  [STATE(2986)] = {
    [sym_identifier] = ACTIONS(7381),
    [anon_sym_COMMA] = ACTIONS(7383),
    [anon_sym_RPAREN] = ACTIONS(7383),
    [anon_sym_LPAREN2] = ACTIONS(7383),
    [anon_sym_STAR] = ACTIONS(7383),
    [anon_sym_CARET] = ACTIONS(7383),
    [anon_sym_SEMI] = ACTIONS(7383),
    [anon_sym___extension__] = ACTIONS(7381),
    [anon_sym_extern] = ACTIONS(7381),
    [anon_sym___attribute__] = ACTIONS(7381),
    [anon_sym___attribute] = ACTIONS(7381),
    [anon_sym_noreturn] = ACTIONS(7381),
    [anon_sym_LBRACK] = ACTIONS(7383),
    [anon_sym___declspec] = ACTIONS(7381),
    [anon_sym___based] = ACTIONS(7381),
    [anon_sym_signed] = ACTIONS(7381),
    [anon_sym_unsigned] = ACTIONS(7381),
    [anon_sym_long] = ACTIONS(7381),
    [anon_sym_short] = ACTIONS(7381),
    [anon_sym_static] = ACTIONS(7381),
    [anon_sym_auto] = ACTIONS(7381),
    [anon_sym_register] = ACTIONS(7381),
    [anon_sym_inline] = ACTIONS(7381),
    [anon_sym___inline] = ACTIONS(7381),
    [anon_sym___inline__] = ACTIONS(7381),
    [anon_sym___forceinline] = ACTIONS(7381),
    [anon_sym_thread_local] = ACTIONS(7381),
    [anon_sym___thread] = ACTIONS(7381),
    [anon_sym_CG_EXTERN] = ACTIONS(7381),
    [anon_sym_CG_INLINE] = ACTIONS(7381),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7381),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7381),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7381),
    [anon_sym_IBOutlet] = ACTIONS(7381),
    [anon_sym_IBInspectable] = ACTIONS(7381),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7381),
    [anon_sym_NS_INLINE] = ACTIONS(7381),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7381),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7381),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7381),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7381),
    [anon_sym_const] = ACTIONS(7381),
    [anon_sym_constexpr] = ACTIONS(7381),
    [anon_sym_volatile] = ACTIONS(7381),
    [anon_sym_restrict] = ACTIONS(7381),
    [anon_sym___restrict__] = ACTIONS(7381),
    [anon_sym__Atomic] = ACTIONS(7381),
    [anon_sym__Noreturn] = ACTIONS(7381),
    [anon_sym__Nonnull] = ACTIONS(7381),
    [anon_sym_nullable] = ACTIONS(7381),
    [anon_sym__Complex] = ACTIONS(7381),
    [anon_sym__Nullable] = ACTIONS(7381),
    [anon_sym__Nullable_result] = ACTIONS(7381),
    [anon_sym__Null_unspecified] = ACTIONS(7381),
    [anon_sym___autoreleasing] = ACTIONS(7381),
    [anon_sym___block] = ACTIONS(7381),
    [anon_sym___bridge] = ACTIONS(7381),
    [anon_sym___bridge_retained] = ACTIONS(7381),
    [anon_sym___bridge_transfer] = ACTIONS(7381),
    [anon_sym___complex] = ACTIONS(7381),
    [anon_sym___const] = ACTIONS(7381),
    [anon_sym___imag] = ACTIONS(7381),
    [anon_sym___kindof] = ACTIONS(7381),
    [anon_sym___nonnull] = ACTIONS(7381),
    [anon_sym___nullable] = ACTIONS(7381),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7381),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7381),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7381),
    [anon_sym___real] = ACTIONS(7381),
    [anon_sym___strong] = ACTIONS(7381),
    [anon_sym___unsafe_unretained] = ACTIONS(7381),
    [anon_sym___unused] = ACTIONS(7381),
    [anon_sym___weak] = ACTIONS(7381),
    [anon_sym_alignas] = ACTIONS(7381),
    [anon_sym__Alignas] = ACTIONS(7381),
    [sym_primitive_type] = ACTIONS(7381),
    [anon_sym_enum] = ACTIONS(7381),
    [anon_sym_COLON] = ACTIONS(7383),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7381),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7381),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7381),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7381),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7381),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7381),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7381),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7381),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7381),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7381),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7381),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7381),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7381),
    [anon_sym_API_AVAILABLE] = ACTIONS(7381),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7381),
    [anon_sym_API_DEPRECATED] = ACTIONS(7381),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7381),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7381),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7381),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7381),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7381),
    [anon_sym___deprecated_msg] = ACTIONS(7381),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7381),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7381),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7381),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7381),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7381),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7381),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7381),
  },
  [STATE(2987)] = {
    [sym_identifier] = ACTIONS(7385),
    [anon_sym_COMMA] = ACTIONS(7387),
    [anon_sym_RPAREN] = ACTIONS(7387),
    [anon_sym_LPAREN2] = ACTIONS(7387),
    [anon_sym_STAR] = ACTIONS(7387),
    [anon_sym_CARET] = ACTIONS(7387),
    [anon_sym_SEMI] = ACTIONS(7387),
    [anon_sym___extension__] = ACTIONS(7385),
    [anon_sym_extern] = ACTIONS(7385),
    [anon_sym___attribute__] = ACTIONS(7385),
    [anon_sym___attribute] = ACTIONS(7385),
    [anon_sym_noreturn] = ACTIONS(7385),
    [anon_sym_LBRACK] = ACTIONS(7387),
    [anon_sym___declspec] = ACTIONS(7385),
    [anon_sym___based] = ACTIONS(7385),
    [anon_sym_signed] = ACTIONS(7385),
    [anon_sym_unsigned] = ACTIONS(7385),
    [anon_sym_long] = ACTIONS(7385),
    [anon_sym_short] = ACTIONS(7385),
    [anon_sym_static] = ACTIONS(7385),
    [anon_sym_auto] = ACTIONS(7385),
    [anon_sym_register] = ACTIONS(7385),
    [anon_sym_inline] = ACTIONS(7385),
    [anon_sym___inline] = ACTIONS(7385),
    [anon_sym___inline__] = ACTIONS(7385),
    [anon_sym___forceinline] = ACTIONS(7385),
    [anon_sym_thread_local] = ACTIONS(7385),
    [anon_sym___thread] = ACTIONS(7385),
    [anon_sym_CG_EXTERN] = ACTIONS(7385),
    [anon_sym_CG_INLINE] = ACTIONS(7385),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7385),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7385),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7385),
    [anon_sym_IBOutlet] = ACTIONS(7385),
    [anon_sym_IBInspectable] = ACTIONS(7385),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7385),
    [anon_sym_NS_INLINE] = ACTIONS(7385),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7385),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7385),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7385),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7385),
    [anon_sym_const] = ACTIONS(7385),
    [anon_sym_constexpr] = ACTIONS(7385),
    [anon_sym_volatile] = ACTIONS(7385),
    [anon_sym_restrict] = ACTIONS(7385),
    [anon_sym___restrict__] = ACTIONS(7385),
    [anon_sym__Atomic] = ACTIONS(7385),
    [anon_sym__Noreturn] = ACTIONS(7385),
    [anon_sym__Nonnull] = ACTIONS(7385),
    [anon_sym_nullable] = ACTIONS(7385),
    [anon_sym__Complex] = ACTIONS(7385),
    [anon_sym__Nullable] = ACTIONS(7385),
    [anon_sym__Nullable_result] = ACTIONS(7385),
    [anon_sym__Null_unspecified] = ACTIONS(7385),
    [anon_sym___autoreleasing] = ACTIONS(7385),
    [anon_sym___block] = ACTIONS(7385),
    [anon_sym___bridge] = ACTIONS(7385),
    [anon_sym___bridge_retained] = ACTIONS(7385),
    [anon_sym___bridge_transfer] = ACTIONS(7385),
    [anon_sym___complex] = ACTIONS(7385),
    [anon_sym___const] = ACTIONS(7385),
    [anon_sym___imag] = ACTIONS(7385),
    [anon_sym___kindof] = ACTIONS(7385),
    [anon_sym___nonnull] = ACTIONS(7385),
    [anon_sym___nullable] = ACTIONS(7385),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7385),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7385),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7385),
    [anon_sym___real] = ACTIONS(7385),
    [anon_sym___strong] = ACTIONS(7385),
    [anon_sym___unsafe_unretained] = ACTIONS(7385),
    [anon_sym___unused] = ACTIONS(7385),
    [anon_sym___weak] = ACTIONS(7385),
    [anon_sym_alignas] = ACTIONS(7385),
    [anon_sym__Alignas] = ACTIONS(7385),
    [sym_primitive_type] = ACTIONS(7385),
    [anon_sym_enum] = ACTIONS(7385),
    [anon_sym_COLON] = ACTIONS(7387),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7385),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7385),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7385),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7385),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7385),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7385),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7385),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7385),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7385),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7385),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7385),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7385),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7385),
    [anon_sym_API_AVAILABLE] = ACTIONS(7385),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7385),
    [anon_sym_API_DEPRECATED] = ACTIONS(7385),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7385),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7385),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7385),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7385),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7385),
    [anon_sym___deprecated_msg] = ACTIONS(7385),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7385),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7385),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7385),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7385),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7385),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7385),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7385),
  },
  [STATE(2988)] = {
    [sym_identifier] = ACTIONS(7389),
    [anon_sym_COMMA] = ACTIONS(7391),
    [anon_sym_RPAREN] = ACTIONS(7391),
    [anon_sym_LPAREN2] = ACTIONS(7391),
    [anon_sym_STAR] = ACTIONS(7391),
    [anon_sym_CARET] = ACTIONS(7391),
    [anon_sym_SEMI] = ACTIONS(7391),
    [anon_sym___extension__] = ACTIONS(7389),
    [anon_sym_extern] = ACTIONS(7389),
    [anon_sym___attribute__] = ACTIONS(7389),
    [anon_sym___attribute] = ACTIONS(7389),
    [anon_sym_noreturn] = ACTIONS(7389),
    [anon_sym_LBRACK] = ACTIONS(7391),
    [anon_sym___declspec] = ACTIONS(7389),
    [anon_sym___based] = ACTIONS(7389),
    [anon_sym_signed] = ACTIONS(7389),
    [anon_sym_unsigned] = ACTIONS(7389),
    [anon_sym_long] = ACTIONS(7389),
    [anon_sym_short] = ACTIONS(7389),
    [anon_sym_static] = ACTIONS(7389),
    [anon_sym_auto] = ACTIONS(7389),
    [anon_sym_register] = ACTIONS(7389),
    [anon_sym_inline] = ACTIONS(7389),
    [anon_sym___inline] = ACTIONS(7389),
    [anon_sym___inline__] = ACTIONS(7389),
    [anon_sym___forceinline] = ACTIONS(7389),
    [anon_sym_thread_local] = ACTIONS(7389),
    [anon_sym___thread] = ACTIONS(7389),
    [anon_sym_CG_EXTERN] = ACTIONS(7389),
    [anon_sym_CG_INLINE] = ACTIONS(7389),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7389),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7389),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7389),
    [anon_sym_IBOutlet] = ACTIONS(7389),
    [anon_sym_IBInspectable] = ACTIONS(7389),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7389),
    [anon_sym_NS_INLINE] = ACTIONS(7389),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7389),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7389),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7389),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7389),
    [anon_sym_const] = ACTIONS(7389),
    [anon_sym_constexpr] = ACTIONS(7389),
    [anon_sym_volatile] = ACTIONS(7389),
    [anon_sym_restrict] = ACTIONS(7389),
    [anon_sym___restrict__] = ACTIONS(7389),
    [anon_sym__Atomic] = ACTIONS(7389),
    [anon_sym__Noreturn] = ACTIONS(7389),
    [anon_sym__Nonnull] = ACTIONS(7389),
    [anon_sym_nullable] = ACTIONS(7389),
    [anon_sym__Complex] = ACTIONS(7389),
    [anon_sym__Nullable] = ACTIONS(7389),
    [anon_sym__Nullable_result] = ACTIONS(7389),
    [anon_sym__Null_unspecified] = ACTIONS(7389),
    [anon_sym___autoreleasing] = ACTIONS(7389),
    [anon_sym___block] = ACTIONS(7389),
    [anon_sym___bridge] = ACTIONS(7389),
    [anon_sym___bridge_retained] = ACTIONS(7389),
    [anon_sym___bridge_transfer] = ACTIONS(7389),
    [anon_sym___complex] = ACTIONS(7389),
    [anon_sym___const] = ACTIONS(7389),
    [anon_sym___imag] = ACTIONS(7389),
    [anon_sym___kindof] = ACTIONS(7389),
    [anon_sym___nonnull] = ACTIONS(7389),
    [anon_sym___nullable] = ACTIONS(7389),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7389),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7389),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7389),
    [anon_sym___real] = ACTIONS(7389),
    [anon_sym___strong] = ACTIONS(7389),
    [anon_sym___unsafe_unretained] = ACTIONS(7389),
    [anon_sym___unused] = ACTIONS(7389),
    [anon_sym___weak] = ACTIONS(7389),
    [anon_sym_alignas] = ACTIONS(7389),
    [anon_sym__Alignas] = ACTIONS(7389),
    [sym_primitive_type] = ACTIONS(7389),
    [anon_sym_enum] = ACTIONS(7389),
    [anon_sym_COLON] = ACTIONS(7391),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7389),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7389),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7389),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7389),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7389),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7389),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7389),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7389),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7389),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7389),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7389),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7389),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7389),
    [anon_sym_API_AVAILABLE] = ACTIONS(7389),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7389),
    [anon_sym_API_DEPRECATED] = ACTIONS(7389),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7389),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7389),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7389),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7389),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7389),
    [anon_sym___deprecated_msg] = ACTIONS(7389),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7389),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7389),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7389),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7389),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7389),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7389),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7389),
  },
  [STATE(2989)] = {
    [sym_identifier] = ACTIONS(7393),
    [anon_sym_COMMA] = ACTIONS(7395),
    [anon_sym_RPAREN] = ACTIONS(7395),
    [anon_sym_LPAREN2] = ACTIONS(7395),
    [anon_sym_STAR] = ACTIONS(7395),
    [anon_sym_CARET] = ACTIONS(7395),
    [anon_sym_SEMI] = ACTIONS(7395),
    [anon_sym___extension__] = ACTIONS(7393),
    [anon_sym_extern] = ACTIONS(7393),
    [anon_sym___attribute__] = ACTIONS(7393),
    [anon_sym___attribute] = ACTIONS(7393),
    [anon_sym_noreturn] = ACTIONS(7393),
    [anon_sym_LBRACK] = ACTIONS(7395),
    [anon_sym___declspec] = ACTIONS(7393),
    [anon_sym___based] = ACTIONS(7393),
    [anon_sym_signed] = ACTIONS(7393),
    [anon_sym_unsigned] = ACTIONS(7393),
    [anon_sym_long] = ACTIONS(7393),
    [anon_sym_short] = ACTIONS(7393),
    [anon_sym_static] = ACTIONS(7393),
    [anon_sym_auto] = ACTIONS(7393),
    [anon_sym_register] = ACTIONS(7393),
    [anon_sym_inline] = ACTIONS(7393),
    [anon_sym___inline] = ACTIONS(7393),
    [anon_sym___inline__] = ACTIONS(7393),
    [anon_sym___forceinline] = ACTIONS(7393),
    [anon_sym_thread_local] = ACTIONS(7393),
    [anon_sym___thread] = ACTIONS(7393),
    [anon_sym_CG_EXTERN] = ACTIONS(7393),
    [anon_sym_CG_INLINE] = ACTIONS(7393),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7393),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7393),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7393),
    [anon_sym_IBOutlet] = ACTIONS(7393),
    [anon_sym_IBInspectable] = ACTIONS(7393),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7393),
    [anon_sym_NS_INLINE] = ACTIONS(7393),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7393),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7393),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7393),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7393),
    [anon_sym_const] = ACTIONS(7393),
    [anon_sym_constexpr] = ACTIONS(7393),
    [anon_sym_volatile] = ACTIONS(7393),
    [anon_sym_restrict] = ACTIONS(7393),
    [anon_sym___restrict__] = ACTIONS(7393),
    [anon_sym__Atomic] = ACTIONS(7393),
    [anon_sym__Noreturn] = ACTIONS(7393),
    [anon_sym__Nonnull] = ACTIONS(7393),
    [anon_sym_nullable] = ACTIONS(7393),
    [anon_sym__Complex] = ACTIONS(7393),
    [anon_sym__Nullable] = ACTIONS(7393),
    [anon_sym__Nullable_result] = ACTIONS(7393),
    [anon_sym__Null_unspecified] = ACTIONS(7393),
    [anon_sym___autoreleasing] = ACTIONS(7393),
    [anon_sym___block] = ACTIONS(7393),
    [anon_sym___bridge] = ACTIONS(7393),
    [anon_sym___bridge_retained] = ACTIONS(7393),
    [anon_sym___bridge_transfer] = ACTIONS(7393),
    [anon_sym___complex] = ACTIONS(7393),
    [anon_sym___const] = ACTIONS(7393),
    [anon_sym___imag] = ACTIONS(7393),
    [anon_sym___kindof] = ACTIONS(7393),
    [anon_sym___nonnull] = ACTIONS(7393),
    [anon_sym___nullable] = ACTIONS(7393),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7393),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7393),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7393),
    [anon_sym___real] = ACTIONS(7393),
    [anon_sym___strong] = ACTIONS(7393),
    [anon_sym___unsafe_unretained] = ACTIONS(7393),
    [anon_sym___unused] = ACTIONS(7393),
    [anon_sym___weak] = ACTIONS(7393),
    [anon_sym_alignas] = ACTIONS(7393),
    [anon_sym__Alignas] = ACTIONS(7393),
    [sym_primitive_type] = ACTIONS(7393),
    [anon_sym_enum] = ACTIONS(7393),
    [anon_sym_COLON] = ACTIONS(7395),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7393),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7393),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7393),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7393),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7393),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7393),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7393),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7393),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7393),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7393),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7393),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7393),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7393),
    [anon_sym_API_AVAILABLE] = ACTIONS(7393),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7393),
    [anon_sym_API_DEPRECATED] = ACTIONS(7393),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7393),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7393),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7393),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7393),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7393),
    [anon_sym___deprecated_msg] = ACTIONS(7393),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7393),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7393),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7393),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7393),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7393),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7393),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7393),
  },
  [STATE(2990)] = {
    [sym_identifier] = ACTIONS(7397),
    [anon_sym_COMMA] = ACTIONS(7399),
    [anon_sym_RPAREN] = ACTIONS(7399),
    [anon_sym_LPAREN2] = ACTIONS(7399),
    [anon_sym_STAR] = ACTIONS(7399),
    [anon_sym_CARET] = ACTIONS(7399),
    [anon_sym_SEMI] = ACTIONS(7399),
    [anon_sym___extension__] = ACTIONS(7397),
    [anon_sym_extern] = ACTIONS(7397),
    [anon_sym___attribute__] = ACTIONS(7397),
    [anon_sym___attribute] = ACTIONS(7397),
    [anon_sym_noreturn] = ACTIONS(7397),
    [anon_sym_LBRACK] = ACTIONS(7399),
    [anon_sym___declspec] = ACTIONS(7397),
    [anon_sym___based] = ACTIONS(7397),
    [anon_sym_signed] = ACTIONS(7397),
    [anon_sym_unsigned] = ACTIONS(7397),
    [anon_sym_long] = ACTIONS(7397),
    [anon_sym_short] = ACTIONS(7397),
    [anon_sym_static] = ACTIONS(7397),
    [anon_sym_auto] = ACTIONS(7397),
    [anon_sym_register] = ACTIONS(7397),
    [anon_sym_inline] = ACTIONS(7397),
    [anon_sym___inline] = ACTIONS(7397),
    [anon_sym___inline__] = ACTIONS(7397),
    [anon_sym___forceinline] = ACTIONS(7397),
    [anon_sym_thread_local] = ACTIONS(7397),
    [anon_sym___thread] = ACTIONS(7397),
    [anon_sym_CG_EXTERN] = ACTIONS(7397),
    [anon_sym_CG_INLINE] = ACTIONS(7397),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7397),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7397),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7397),
    [anon_sym_IBOutlet] = ACTIONS(7397),
    [anon_sym_IBInspectable] = ACTIONS(7397),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7397),
    [anon_sym_NS_INLINE] = ACTIONS(7397),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7397),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7397),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7397),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7397),
    [anon_sym_const] = ACTIONS(7397),
    [anon_sym_constexpr] = ACTIONS(7397),
    [anon_sym_volatile] = ACTIONS(7397),
    [anon_sym_restrict] = ACTIONS(7397),
    [anon_sym___restrict__] = ACTIONS(7397),
    [anon_sym__Atomic] = ACTIONS(7397),
    [anon_sym__Noreturn] = ACTIONS(7397),
    [anon_sym__Nonnull] = ACTIONS(7397),
    [anon_sym_nullable] = ACTIONS(7397),
    [anon_sym__Complex] = ACTIONS(7397),
    [anon_sym__Nullable] = ACTIONS(7397),
    [anon_sym__Nullable_result] = ACTIONS(7397),
    [anon_sym__Null_unspecified] = ACTIONS(7397),
    [anon_sym___autoreleasing] = ACTIONS(7397),
    [anon_sym___block] = ACTIONS(7397),
    [anon_sym___bridge] = ACTIONS(7397),
    [anon_sym___bridge_retained] = ACTIONS(7397),
    [anon_sym___bridge_transfer] = ACTIONS(7397),
    [anon_sym___complex] = ACTIONS(7397),
    [anon_sym___const] = ACTIONS(7397),
    [anon_sym___imag] = ACTIONS(7397),
    [anon_sym___kindof] = ACTIONS(7397),
    [anon_sym___nonnull] = ACTIONS(7397),
    [anon_sym___nullable] = ACTIONS(7397),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7397),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7397),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7397),
    [anon_sym___real] = ACTIONS(7397),
    [anon_sym___strong] = ACTIONS(7397),
    [anon_sym___unsafe_unretained] = ACTIONS(7397),
    [anon_sym___unused] = ACTIONS(7397),
    [anon_sym___weak] = ACTIONS(7397),
    [anon_sym_alignas] = ACTIONS(7397),
    [anon_sym__Alignas] = ACTIONS(7397),
    [sym_primitive_type] = ACTIONS(7397),
    [anon_sym_enum] = ACTIONS(7397),
    [anon_sym_COLON] = ACTIONS(7399),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7397),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7397),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7397),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7397),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7397),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7397),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7397),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7397),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7397),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7397),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7397),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7397),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7397),
    [anon_sym_API_AVAILABLE] = ACTIONS(7397),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7397),
    [anon_sym_API_DEPRECATED] = ACTIONS(7397),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7397),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7397),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7397),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7397),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7397),
    [anon_sym___deprecated_msg] = ACTIONS(7397),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7397),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7397),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7397),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7397),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7397),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7397),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7397),
  },
  [STATE(2991)] = {
    [sym_identifier] = ACTIONS(7401),
    [anon_sym_COMMA] = ACTIONS(7403),
    [anon_sym_RPAREN] = ACTIONS(7403),
    [anon_sym_LPAREN2] = ACTIONS(7403),
    [anon_sym_STAR] = ACTIONS(7403),
    [anon_sym_CARET] = ACTIONS(7403),
    [anon_sym_SEMI] = ACTIONS(7403),
    [anon_sym___extension__] = ACTIONS(7401),
    [anon_sym_extern] = ACTIONS(7401),
    [anon_sym___attribute__] = ACTIONS(7401),
    [anon_sym___attribute] = ACTIONS(7401),
    [anon_sym_noreturn] = ACTIONS(7401),
    [anon_sym_LBRACK] = ACTIONS(7403),
    [anon_sym___declspec] = ACTIONS(7401),
    [anon_sym___based] = ACTIONS(7401),
    [anon_sym_signed] = ACTIONS(7401),
    [anon_sym_unsigned] = ACTIONS(7401),
    [anon_sym_long] = ACTIONS(7401),
    [anon_sym_short] = ACTIONS(7401),
    [anon_sym_static] = ACTIONS(7401),
    [anon_sym_auto] = ACTIONS(7401),
    [anon_sym_register] = ACTIONS(7401),
    [anon_sym_inline] = ACTIONS(7401),
    [anon_sym___inline] = ACTIONS(7401),
    [anon_sym___inline__] = ACTIONS(7401),
    [anon_sym___forceinline] = ACTIONS(7401),
    [anon_sym_thread_local] = ACTIONS(7401),
    [anon_sym___thread] = ACTIONS(7401),
    [anon_sym_CG_EXTERN] = ACTIONS(7401),
    [anon_sym_CG_INLINE] = ACTIONS(7401),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7401),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7401),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7401),
    [anon_sym_IBOutlet] = ACTIONS(7401),
    [anon_sym_IBInspectable] = ACTIONS(7401),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7401),
    [anon_sym_NS_INLINE] = ACTIONS(7401),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7401),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7401),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7401),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7401),
    [anon_sym_const] = ACTIONS(7401),
    [anon_sym_constexpr] = ACTIONS(7401),
    [anon_sym_volatile] = ACTIONS(7401),
    [anon_sym_restrict] = ACTIONS(7401),
    [anon_sym___restrict__] = ACTIONS(7401),
    [anon_sym__Atomic] = ACTIONS(7401),
    [anon_sym__Noreturn] = ACTIONS(7401),
    [anon_sym__Nonnull] = ACTIONS(7401),
    [anon_sym_nullable] = ACTIONS(7401),
    [anon_sym__Complex] = ACTIONS(7401),
    [anon_sym__Nullable] = ACTIONS(7401),
    [anon_sym__Nullable_result] = ACTIONS(7401),
    [anon_sym__Null_unspecified] = ACTIONS(7401),
    [anon_sym___autoreleasing] = ACTIONS(7401),
    [anon_sym___block] = ACTIONS(7401),
    [anon_sym___bridge] = ACTIONS(7401),
    [anon_sym___bridge_retained] = ACTIONS(7401),
    [anon_sym___bridge_transfer] = ACTIONS(7401),
    [anon_sym___complex] = ACTIONS(7401),
    [anon_sym___const] = ACTIONS(7401),
    [anon_sym___imag] = ACTIONS(7401),
    [anon_sym___kindof] = ACTIONS(7401),
    [anon_sym___nonnull] = ACTIONS(7401),
    [anon_sym___nullable] = ACTIONS(7401),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7401),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7401),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7401),
    [anon_sym___real] = ACTIONS(7401),
    [anon_sym___strong] = ACTIONS(7401),
    [anon_sym___unsafe_unretained] = ACTIONS(7401),
    [anon_sym___unused] = ACTIONS(7401),
    [anon_sym___weak] = ACTIONS(7401),
    [anon_sym_alignas] = ACTIONS(7401),
    [anon_sym__Alignas] = ACTIONS(7401),
    [sym_primitive_type] = ACTIONS(7401),
    [anon_sym_enum] = ACTIONS(7401),
    [anon_sym_COLON] = ACTIONS(7403),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7401),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7401),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7401),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7401),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7401),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7401),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7401),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7401),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7401),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7401),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7401),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7401),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7401),
    [anon_sym_API_AVAILABLE] = ACTIONS(7401),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7401),
    [anon_sym_API_DEPRECATED] = ACTIONS(7401),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7401),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7401),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7401),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7401),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7401),
    [anon_sym___deprecated_msg] = ACTIONS(7401),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7401),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7401),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7401),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7401),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7401),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7401),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7401),
  },
  [STATE(2992)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2966),
    [sym_identifier] = ACTIONS(7405),
    [anon_sym_COMMA] = ACTIONS(7407),
    [anon_sym_RPAREN] = ACTIONS(7407),
    [anon_sym_LPAREN2] = ACTIONS(7407),
    [anon_sym_STAR] = ACTIONS(7407),
    [anon_sym_CARET] = ACTIONS(7407),
    [anon_sym_SEMI] = ACTIONS(7407),
    [anon_sym___extension__] = ACTIONS(7405),
    [anon_sym_extern] = ACTIONS(7405),
    [anon_sym___attribute__] = ACTIONS(7405),
    [anon_sym___attribute] = ACTIONS(7405),
    [anon_sym_noreturn] = ACTIONS(7405),
    [anon_sym_LBRACK] = ACTIONS(7407),
    [anon_sym___declspec] = ACTIONS(7405),
    [anon_sym___based] = ACTIONS(7405),
    [anon_sym_signed] = ACTIONS(7409),
    [anon_sym_unsigned] = ACTIONS(7409),
    [anon_sym_long] = ACTIONS(7409),
    [anon_sym_short] = ACTIONS(7409),
    [anon_sym_static] = ACTIONS(7405),
    [anon_sym_auto] = ACTIONS(7405),
    [anon_sym_register] = ACTIONS(7405),
    [anon_sym_inline] = ACTIONS(7405),
    [anon_sym___inline] = ACTIONS(7405),
    [anon_sym___inline__] = ACTIONS(7405),
    [anon_sym___forceinline] = ACTIONS(7405),
    [anon_sym_thread_local] = ACTIONS(7405),
    [anon_sym___thread] = ACTIONS(7405),
    [anon_sym_CG_EXTERN] = ACTIONS(7405),
    [anon_sym_CG_INLINE] = ACTIONS(7405),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7405),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7405),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7405),
    [anon_sym_IBOutlet] = ACTIONS(7405),
    [anon_sym_IBInspectable] = ACTIONS(7405),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7405),
    [anon_sym_NS_INLINE] = ACTIONS(7405),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7405),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7405),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7405),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7405),
    [anon_sym_const] = ACTIONS(7405),
    [anon_sym_constexpr] = ACTIONS(7405),
    [anon_sym_volatile] = ACTIONS(7405),
    [anon_sym_restrict] = ACTIONS(7405),
    [anon_sym___restrict__] = ACTIONS(7405),
    [anon_sym__Atomic] = ACTIONS(7405),
    [anon_sym__Noreturn] = ACTIONS(7405),
    [anon_sym__Nonnull] = ACTIONS(7405),
    [anon_sym_nullable] = ACTIONS(7405),
    [anon_sym__Complex] = ACTIONS(7405),
    [anon_sym__Nullable] = ACTIONS(7405),
    [anon_sym__Nullable_result] = ACTIONS(7405),
    [anon_sym__Null_unspecified] = ACTIONS(7405),
    [anon_sym___autoreleasing] = ACTIONS(7405),
    [anon_sym___block] = ACTIONS(7405),
    [anon_sym___bridge] = ACTIONS(7405),
    [anon_sym___bridge_retained] = ACTIONS(7405),
    [anon_sym___bridge_transfer] = ACTIONS(7405),
    [anon_sym___complex] = ACTIONS(7405),
    [anon_sym___const] = ACTIONS(7405),
    [anon_sym___imag] = ACTIONS(7405),
    [anon_sym___kindof] = ACTIONS(7405),
    [anon_sym___nonnull] = ACTIONS(7405),
    [anon_sym___nullable] = ACTIONS(7405),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7405),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7405),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7405),
    [anon_sym___real] = ACTIONS(7405),
    [anon_sym___strong] = ACTIONS(7405),
    [anon_sym___unsafe_unretained] = ACTIONS(7405),
    [anon_sym___unused] = ACTIONS(7405),
    [anon_sym___weak] = ACTIONS(7405),
    [anon_sym_alignas] = ACTIONS(7405),
    [anon_sym__Alignas] = ACTIONS(7405),
    [anon_sym_enum] = ACTIONS(7405),
    [anon_sym_COLON] = ACTIONS(7407),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7405),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7405),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7405),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7405),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7405),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7405),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7405),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7405),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7405),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7405),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7405),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7405),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7405),
    [anon_sym_API_AVAILABLE] = ACTIONS(7405),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7405),
    [anon_sym_API_DEPRECATED] = ACTIONS(7405),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7405),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7405),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7405),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7405),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7405),
    [anon_sym___deprecated_msg] = ACTIONS(7405),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7405),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7405),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7405),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7405),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7405),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7405),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7405),
  },
  [STATE(2993)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2967),
    [sym_identifier] = ACTIONS(7411),
    [anon_sym_COMMA] = ACTIONS(7413),
    [anon_sym_RPAREN] = ACTIONS(7413),
    [anon_sym_LPAREN2] = ACTIONS(7413),
    [anon_sym_STAR] = ACTIONS(7413),
    [anon_sym_CARET] = ACTIONS(7413),
    [anon_sym_SEMI] = ACTIONS(7413),
    [anon_sym___extension__] = ACTIONS(7411),
    [anon_sym_extern] = ACTIONS(7411),
    [anon_sym___attribute__] = ACTIONS(7411),
    [anon_sym___attribute] = ACTIONS(7411),
    [anon_sym_noreturn] = ACTIONS(7411),
    [anon_sym_LBRACK] = ACTIONS(7413),
    [anon_sym___declspec] = ACTIONS(7411),
    [anon_sym___based] = ACTIONS(7411),
    [anon_sym_signed] = ACTIONS(7415),
    [anon_sym_unsigned] = ACTIONS(7415),
    [anon_sym_long] = ACTIONS(7415),
    [anon_sym_short] = ACTIONS(7415),
    [anon_sym_static] = ACTIONS(7411),
    [anon_sym_auto] = ACTIONS(7411),
    [anon_sym_register] = ACTIONS(7411),
    [anon_sym_inline] = ACTIONS(7411),
    [anon_sym___inline] = ACTIONS(7411),
    [anon_sym___inline__] = ACTIONS(7411),
    [anon_sym___forceinline] = ACTIONS(7411),
    [anon_sym_thread_local] = ACTIONS(7411),
    [anon_sym___thread] = ACTIONS(7411),
    [anon_sym_CG_EXTERN] = ACTIONS(7411),
    [anon_sym_CG_INLINE] = ACTIONS(7411),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7411),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7411),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7411),
    [anon_sym_IBOutlet] = ACTIONS(7411),
    [anon_sym_IBInspectable] = ACTIONS(7411),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7411),
    [anon_sym_NS_INLINE] = ACTIONS(7411),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7411),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7411),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7411),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7411),
    [anon_sym_const] = ACTIONS(7411),
    [anon_sym_constexpr] = ACTIONS(7411),
    [anon_sym_volatile] = ACTIONS(7411),
    [anon_sym_restrict] = ACTIONS(7411),
    [anon_sym___restrict__] = ACTIONS(7411),
    [anon_sym__Atomic] = ACTIONS(7411),
    [anon_sym__Noreturn] = ACTIONS(7411),
    [anon_sym__Nonnull] = ACTIONS(7411),
    [anon_sym_nullable] = ACTIONS(7411),
    [anon_sym__Complex] = ACTIONS(7411),
    [anon_sym__Nullable] = ACTIONS(7411),
    [anon_sym__Nullable_result] = ACTIONS(7411),
    [anon_sym__Null_unspecified] = ACTIONS(7411),
    [anon_sym___autoreleasing] = ACTIONS(7411),
    [anon_sym___block] = ACTIONS(7411),
    [anon_sym___bridge] = ACTIONS(7411),
    [anon_sym___bridge_retained] = ACTIONS(7411),
    [anon_sym___bridge_transfer] = ACTIONS(7411),
    [anon_sym___complex] = ACTIONS(7411),
    [anon_sym___const] = ACTIONS(7411),
    [anon_sym___imag] = ACTIONS(7411),
    [anon_sym___kindof] = ACTIONS(7411),
    [anon_sym___nonnull] = ACTIONS(7411),
    [anon_sym___nullable] = ACTIONS(7411),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7411),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7411),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7411),
    [anon_sym___real] = ACTIONS(7411),
    [anon_sym___strong] = ACTIONS(7411),
    [anon_sym___unsafe_unretained] = ACTIONS(7411),
    [anon_sym___unused] = ACTIONS(7411),
    [anon_sym___weak] = ACTIONS(7411),
    [anon_sym_alignas] = ACTIONS(7411),
    [anon_sym__Alignas] = ACTIONS(7411),
    [anon_sym_enum] = ACTIONS(7411),
    [anon_sym_COLON] = ACTIONS(7413),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7411),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7411),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7411),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7411),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7411),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7411),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7411),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7411),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7411),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7411),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7411),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7411),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7411),
    [anon_sym_API_AVAILABLE] = ACTIONS(7411),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7411),
    [anon_sym_API_DEPRECATED] = ACTIONS(7411),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7411),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7411),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7411),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7411),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7411),
    [anon_sym___deprecated_msg] = ACTIONS(7411),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7411),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7411),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7411),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7411),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7411),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7411),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7411),
  },
  [STATE(2994)] = {
    [sym_identifier] = ACTIONS(7417),
    [anon_sym_COMMA] = ACTIONS(7419),
    [anon_sym_RPAREN] = ACTIONS(7419),
    [anon_sym_LPAREN2] = ACTIONS(7419),
    [anon_sym_STAR] = ACTIONS(7419),
    [anon_sym_CARET] = ACTIONS(7419),
    [anon_sym_SEMI] = ACTIONS(7419),
    [anon_sym___extension__] = ACTIONS(7417),
    [anon_sym_extern] = ACTIONS(7417),
    [anon_sym___attribute__] = ACTIONS(7417),
    [anon_sym___attribute] = ACTIONS(7417),
    [anon_sym_noreturn] = ACTIONS(7417),
    [anon_sym_LBRACK] = ACTIONS(7419),
    [anon_sym___declspec] = ACTIONS(7417),
    [anon_sym___based] = ACTIONS(7417),
    [anon_sym_signed] = ACTIONS(7417),
    [anon_sym_unsigned] = ACTIONS(7417),
    [anon_sym_long] = ACTIONS(7417),
    [anon_sym_short] = ACTIONS(7417),
    [anon_sym_static] = ACTIONS(7417),
    [anon_sym_auto] = ACTIONS(7417),
    [anon_sym_register] = ACTIONS(7417),
    [anon_sym_inline] = ACTIONS(7417),
    [anon_sym___inline] = ACTIONS(7417),
    [anon_sym___inline__] = ACTIONS(7417),
    [anon_sym___forceinline] = ACTIONS(7417),
    [anon_sym_thread_local] = ACTIONS(7417),
    [anon_sym___thread] = ACTIONS(7417),
    [anon_sym_CG_EXTERN] = ACTIONS(7417),
    [anon_sym_CG_INLINE] = ACTIONS(7417),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7417),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7417),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7417),
    [anon_sym_IBOutlet] = ACTIONS(7417),
    [anon_sym_IBInspectable] = ACTIONS(7417),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7417),
    [anon_sym_NS_INLINE] = ACTIONS(7417),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7417),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7417),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7417),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7417),
    [anon_sym_const] = ACTIONS(7417),
    [anon_sym_constexpr] = ACTIONS(7417),
    [anon_sym_volatile] = ACTIONS(7417),
    [anon_sym_restrict] = ACTIONS(7417),
    [anon_sym___restrict__] = ACTIONS(7417),
    [anon_sym__Atomic] = ACTIONS(7417),
    [anon_sym__Noreturn] = ACTIONS(7417),
    [anon_sym__Nonnull] = ACTIONS(7417),
    [anon_sym_nullable] = ACTIONS(7417),
    [anon_sym__Complex] = ACTIONS(7417),
    [anon_sym__Nullable] = ACTIONS(7417),
    [anon_sym__Nullable_result] = ACTIONS(7417),
    [anon_sym__Null_unspecified] = ACTIONS(7417),
    [anon_sym___autoreleasing] = ACTIONS(7417),
    [anon_sym___block] = ACTIONS(7417),
    [anon_sym___bridge] = ACTIONS(7417),
    [anon_sym___bridge_retained] = ACTIONS(7417),
    [anon_sym___bridge_transfer] = ACTIONS(7417),
    [anon_sym___complex] = ACTIONS(7417),
    [anon_sym___const] = ACTIONS(7417),
    [anon_sym___imag] = ACTIONS(7417),
    [anon_sym___kindof] = ACTIONS(7417),
    [anon_sym___nonnull] = ACTIONS(7417),
    [anon_sym___nullable] = ACTIONS(7417),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7417),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7417),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7417),
    [anon_sym___real] = ACTIONS(7417),
    [anon_sym___strong] = ACTIONS(7417),
    [anon_sym___unsafe_unretained] = ACTIONS(7417),
    [anon_sym___unused] = ACTIONS(7417),
    [anon_sym___weak] = ACTIONS(7417),
    [anon_sym_alignas] = ACTIONS(7417),
    [anon_sym__Alignas] = ACTIONS(7417),
    [sym_primitive_type] = ACTIONS(7417),
    [anon_sym_enum] = ACTIONS(7417),
    [anon_sym_COLON] = ACTIONS(7419),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7417),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7417),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7417),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7417),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7417),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7417),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7417),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7417),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7417),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7417),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7417),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7417),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7417),
    [anon_sym_API_AVAILABLE] = ACTIONS(7417),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7417),
    [anon_sym_API_DEPRECATED] = ACTIONS(7417),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7417),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7417),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7417),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7417),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7417),
    [anon_sym___deprecated_msg] = ACTIONS(7417),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7417),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7417),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7417),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7417),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7417),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7417),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7417),
  },
  [STATE(2995)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2912),
    [sym_identifier] = ACTIONS(7421),
    [anon_sym_COMMA] = ACTIONS(7423),
    [anon_sym_RPAREN] = ACTIONS(7423),
    [anon_sym_LPAREN2] = ACTIONS(7423),
    [anon_sym_STAR] = ACTIONS(7423),
    [anon_sym_CARET] = ACTIONS(7423),
    [anon_sym_SEMI] = ACTIONS(7423),
    [anon_sym___extension__] = ACTIONS(7421),
    [anon_sym_extern] = ACTIONS(7421),
    [anon_sym___attribute__] = ACTIONS(7421),
    [anon_sym___attribute] = ACTIONS(7421),
    [anon_sym_noreturn] = ACTIONS(7421),
    [anon_sym_LBRACK] = ACTIONS(7423),
    [anon_sym___declspec] = ACTIONS(7421),
    [anon_sym___based] = ACTIONS(7421),
    [anon_sym_signed] = ACTIONS(7287),
    [anon_sym_unsigned] = ACTIONS(7287),
    [anon_sym_long] = ACTIONS(7287),
    [anon_sym_short] = ACTIONS(7287),
    [anon_sym_static] = ACTIONS(7421),
    [anon_sym_auto] = ACTIONS(7421),
    [anon_sym_register] = ACTIONS(7421),
    [anon_sym_inline] = ACTIONS(7421),
    [anon_sym___inline] = ACTIONS(7421),
    [anon_sym___inline__] = ACTIONS(7421),
    [anon_sym___forceinline] = ACTIONS(7421),
    [anon_sym_thread_local] = ACTIONS(7421),
    [anon_sym___thread] = ACTIONS(7421),
    [anon_sym_CG_EXTERN] = ACTIONS(7421),
    [anon_sym_CG_INLINE] = ACTIONS(7421),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7421),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7421),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7421),
    [anon_sym_IBOutlet] = ACTIONS(7421),
    [anon_sym_IBInspectable] = ACTIONS(7421),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7421),
    [anon_sym_NS_INLINE] = ACTIONS(7421),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7421),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7421),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7421),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7421),
    [anon_sym_const] = ACTIONS(7421),
    [anon_sym_constexpr] = ACTIONS(7421),
    [anon_sym_volatile] = ACTIONS(7421),
    [anon_sym_restrict] = ACTIONS(7421),
    [anon_sym___restrict__] = ACTIONS(7421),
    [anon_sym__Atomic] = ACTIONS(7421),
    [anon_sym__Noreturn] = ACTIONS(7421),
    [anon_sym__Nonnull] = ACTIONS(7421),
    [anon_sym_nullable] = ACTIONS(7421),
    [anon_sym__Complex] = ACTIONS(7421),
    [anon_sym__Nullable] = ACTIONS(7421),
    [anon_sym__Nullable_result] = ACTIONS(7421),
    [anon_sym__Null_unspecified] = ACTIONS(7421),
    [anon_sym___autoreleasing] = ACTIONS(7421),
    [anon_sym___block] = ACTIONS(7421),
    [anon_sym___bridge] = ACTIONS(7421),
    [anon_sym___bridge_retained] = ACTIONS(7421),
    [anon_sym___bridge_transfer] = ACTIONS(7421),
    [anon_sym___complex] = ACTIONS(7421),
    [anon_sym___const] = ACTIONS(7421),
    [anon_sym___imag] = ACTIONS(7421),
    [anon_sym___kindof] = ACTIONS(7421),
    [anon_sym___nonnull] = ACTIONS(7421),
    [anon_sym___nullable] = ACTIONS(7421),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7421),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7421),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7421),
    [anon_sym___real] = ACTIONS(7421),
    [anon_sym___strong] = ACTIONS(7421),
    [anon_sym___unsafe_unretained] = ACTIONS(7421),
    [anon_sym___unused] = ACTIONS(7421),
    [anon_sym___weak] = ACTIONS(7421),
    [anon_sym_alignas] = ACTIONS(7421),
    [anon_sym__Alignas] = ACTIONS(7421),
    [anon_sym_enum] = ACTIONS(7421),
    [anon_sym_COLON] = ACTIONS(7423),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7421),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7421),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7421),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7421),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7421),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7421),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7421),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7421),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7421),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7421),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7421),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7421),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7421),
    [anon_sym_API_AVAILABLE] = ACTIONS(7421),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7421),
    [anon_sym_API_DEPRECATED] = ACTIONS(7421),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7421),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7421),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7421),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7421),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7421),
    [anon_sym___deprecated_msg] = ACTIONS(7421),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7421),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7421),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7421),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7421),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7421),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7421),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7421),
  },
  [STATE(2996)] = {
    [sym_identifier] = ACTIONS(7425),
    [anon_sym_COMMA] = ACTIONS(7427),
    [anon_sym_RPAREN] = ACTIONS(7427),
    [anon_sym_LPAREN2] = ACTIONS(7427),
    [anon_sym_STAR] = ACTIONS(7427),
    [anon_sym_CARET] = ACTIONS(7427),
    [anon_sym_SEMI] = ACTIONS(7427),
    [anon_sym___extension__] = ACTIONS(7425),
    [anon_sym_extern] = ACTIONS(7425),
    [anon_sym___attribute__] = ACTIONS(7425),
    [anon_sym___attribute] = ACTIONS(7425),
    [anon_sym_noreturn] = ACTIONS(7425),
    [anon_sym_LBRACK] = ACTIONS(7427),
    [anon_sym___declspec] = ACTIONS(7425),
    [anon_sym___based] = ACTIONS(7425),
    [anon_sym_signed] = ACTIONS(7425),
    [anon_sym_unsigned] = ACTIONS(7425),
    [anon_sym_long] = ACTIONS(7425),
    [anon_sym_short] = ACTIONS(7425),
    [anon_sym_static] = ACTIONS(7425),
    [anon_sym_auto] = ACTIONS(7425),
    [anon_sym_register] = ACTIONS(7425),
    [anon_sym_inline] = ACTIONS(7425),
    [anon_sym___inline] = ACTIONS(7425),
    [anon_sym___inline__] = ACTIONS(7425),
    [anon_sym___forceinline] = ACTIONS(7425),
    [anon_sym_thread_local] = ACTIONS(7425),
    [anon_sym___thread] = ACTIONS(7425),
    [anon_sym_CG_EXTERN] = ACTIONS(7425),
    [anon_sym_CG_INLINE] = ACTIONS(7425),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7425),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7425),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7425),
    [anon_sym_IBOutlet] = ACTIONS(7425),
    [anon_sym_IBInspectable] = ACTIONS(7425),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7425),
    [anon_sym_NS_INLINE] = ACTIONS(7425),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7425),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7425),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7425),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7425),
    [anon_sym_const] = ACTIONS(7425),
    [anon_sym_constexpr] = ACTIONS(7425),
    [anon_sym_volatile] = ACTIONS(7425),
    [anon_sym_restrict] = ACTIONS(7425),
    [anon_sym___restrict__] = ACTIONS(7425),
    [anon_sym__Atomic] = ACTIONS(7425),
    [anon_sym__Noreturn] = ACTIONS(7425),
    [anon_sym__Nonnull] = ACTIONS(7425),
    [anon_sym_nullable] = ACTIONS(7425),
    [anon_sym__Complex] = ACTIONS(7425),
    [anon_sym__Nullable] = ACTIONS(7425),
    [anon_sym__Nullable_result] = ACTIONS(7425),
    [anon_sym__Null_unspecified] = ACTIONS(7425),
    [anon_sym___autoreleasing] = ACTIONS(7425),
    [anon_sym___block] = ACTIONS(7425),
    [anon_sym___bridge] = ACTIONS(7425),
    [anon_sym___bridge_retained] = ACTIONS(7425),
    [anon_sym___bridge_transfer] = ACTIONS(7425),
    [anon_sym___complex] = ACTIONS(7425),
    [anon_sym___const] = ACTIONS(7425),
    [anon_sym___imag] = ACTIONS(7425),
    [anon_sym___kindof] = ACTIONS(7425),
    [anon_sym___nonnull] = ACTIONS(7425),
    [anon_sym___nullable] = ACTIONS(7425),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7425),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7425),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7425),
    [anon_sym___real] = ACTIONS(7425),
    [anon_sym___strong] = ACTIONS(7425),
    [anon_sym___unsafe_unretained] = ACTIONS(7425),
    [anon_sym___unused] = ACTIONS(7425),
    [anon_sym___weak] = ACTIONS(7425),
    [anon_sym_alignas] = ACTIONS(7425),
    [anon_sym__Alignas] = ACTIONS(7425),
    [sym_primitive_type] = ACTIONS(7425),
    [anon_sym_enum] = ACTIONS(7425),
    [anon_sym_COLON] = ACTIONS(7427),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7425),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7425),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7425),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7425),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7425),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7425),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7425),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7425),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7425),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7425),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7425),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7425),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7425),
    [anon_sym_API_AVAILABLE] = ACTIONS(7425),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7425),
    [anon_sym_API_DEPRECATED] = ACTIONS(7425),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7425),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7425),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7425),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7425),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7425),
    [anon_sym___deprecated_msg] = ACTIONS(7425),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7425),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7425),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7425),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7425),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7425),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7425),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7425),
  },
  [STATE(2997)] = {
    [sym_identifier] = ACTIONS(7429),
    [anon_sym_COMMA] = ACTIONS(7431),
    [anon_sym_RPAREN] = ACTIONS(7431),
    [anon_sym_LPAREN2] = ACTIONS(7431),
    [anon_sym_STAR] = ACTIONS(7431),
    [anon_sym_CARET] = ACTIONS(7431),
    [anon_sym_SEMI] = ACTIONS(7431),
    [anon_sym___extension__] = ACTIONS(7429),
    [anon_sym_extern] = ACTIONS(7429),
    [anon_sym___attribute__] = ACTIONS(7429),
    [anon_sym___attribute] = ACTIONS(7429),
    [anon_sym_noreturn] = ACTIONS(7429),
    [anon_sym_LBRACK] = ACTIONS(7431),
    [anon_sym___declspec] = ACTIONS(7429),
    [anon_sym___based] = ACTIONS(7429),
    [anon_sym_signed] = ACTIONS(7429),
    [anon_sym_unsigned] = ACTIONS(7429),
    [anon_sym_long] = ACTIONS(7429),
    [anon_sym_short] = ACTIONS(7429),
    [anon_sym_static] = ACTIONS(7429),
    [anon_sym_auto] = ACTIONS(7429),
    [anon_sym_register] = ACTIONS(7429),
    [anon_sym_inline] = ACTIONS(7429),
    [anon_sym___inline] = ACTIONS(7429),
    [anon_sym___inline__] = ACTIONS(7429),
    [anon_sym___forceinline] = ACTIONS(7429),
    [anon_sym_thread_local] = ACTIONS(7429),
    [anon_sym___thread] = ACTIONS(7429),
    [anon_sym_CG_EXTERN] = ACTIONS(7429),
    [anon_sym_CG_INLINE] = ACTIONS(7429),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7429),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7429),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7429),
    [anon_sym_IBOutlet] = ACTIONS(7429),
    [anon_sym_IBInspectable] = ACTIONS(7429),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7429),
    [anon_sym_NS_INLINE] = ACTIONS(7429),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7429),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7429),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7429),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7429),
    [anon_sym_const] = ACTIONS(7429),
    [anon_sym_constexpr] = ACTIONS(7429),
    [anon_sym_volatile] = ACTIONS(7429),
    [anon_sym_restrict] = ACTIONS(7429),
    [anon_sym___restrict__] = ACTIONS(7429),
    [anon_sym__Atomic] = ACTIONS(7429),
    [anon_sym__Noreturn] = ACTIONS(7429),
    [anon_sym__Nonnull] = ACTIONS(7429),
    [anon_sym_nullable] = ACTIONS(7429),
    [anon_sym__Complex] = ACTIONS(7429),
    [anon_sym__Nullable] = ACTIONS(7429),
    [anon_sym__Nullable_result] = ACTIONS(7429),
    [anon_sym__Null_unspecified] = ACTIONS(7429),
    [anon_sym___autoreleasing] = ACTIONS(7429),
    [anon_sym___block] = ACTIONS(7429),
    [anon_sym___bridge] = ACTIONS(7429),
    [anon_sym___bridge_retained] = ACTIONS(7429),
    [anon_sym___bridge_transfer] = ACTIONS(7429),
    [anon_sym___complex] = ACTIONS(7429),
    [anon_sym___const] = ACTIONS(7429),
    [anon_sym___imag] = ACTIONS(7429),
    [anon_sym___kindof] = ACTIONS(7429),
    [anon_sym___nonnull] = ACTIONS(7429),
    [anon_sym___nullable] = ACTIONS(7429),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7429),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7429),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7429),
    [anon_sym___real] = ACTIONS(7429),
    [anon_sym___strong] = ACTIONS(7429),
    [anon_sym___unsafe_unretained] = ACTIONS(7429),
    [anon_sym___unused] = ACTIONS(7429),
    [anon_sym___weak] = ACTIONS(7429),
    [anon_sym_alignas] = ACTIONS(7429),
    [anon_sym__Alignas] = ACTIONS(7429),
    [sym_primitive_type] = ACTIONS(7429),
    [anon_sym_enum] = ACTIONS(7429),
    [anon_sym_COLON] = ACTIONS(7431),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7429),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7429),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7429),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7429),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7429),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7429),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7429),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7429),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7429),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7429),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7429),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7429),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7429),
    [anon_sym_API_AVAILABLE] = ACTIONS(7429),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7429),
    [anon_sym_API_DEPRECATED] = ACTIONS(7429),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7429),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7429),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7429),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7429),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7429),
    [anon_sym___deprecated_msg] = ACTIONS(7429),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7429),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7429),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7429),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7429),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7429),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7429),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7429),
  },
  [STATE(2998)] = {
    [sym__declaration_modifiers] = STATE(2866),
    [sym_attribute_specifier] = STATE(2866),
    [sym_attribute_declaration] = STATE(2866),
    [sym_ms_declspec_modifier] = STATE(2866),
    [sym_storage_class_specifier] = STATE(2866),
    [sym_type_qualifier] = STATE(2866),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2866),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2866),
    [sym_identifier] = ACTIONS(6899),
    [anon_sym_LPAREN2] = ACTIONS(6901),
    [anon_sym_STAR] = ACTIONS(6901),
    [anon_sym_CARET] = ACTIONS(6901),
    [anon_sym_SEMI] = ACTIONS(7433),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6897),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(6899),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(2999)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2912),
    [sym_identifier] = ACTIONS(7435),
    [anon_sym_COMMA] = ACTIONS(7437),
    [anon_sym_RPAREN] = ACTIONS(7437),
    [anon_sym_LPAREN2] = ACTIONS(7437),
    [anon_sym_STAR] = ACTIONS(7437),
    [anon_sym_CARET] = ACTIONS(7437),
    [anon_sym_SEMI] = ACTIONS(7437),
    [anon_sym___extension__] = ACTIONS(7435),
    [anon_sym_extern] = ACTIONS(7435),
    [anon_sym___attribute__] = ACTIONS(7435),
    [anon_sym___attribute] = ACTIONS(7435),
    [anon_sym_noreturn] = ACTIONS(7435),
    [anon_sym_LBRACK] = ACTIONS(7437),
    [anon_sym___declspec] = ACTIONS(7435),
    [anon_sym___based] = ACTIONS(7435),
    [anon_sym_signed] = ACTIONS(7287),
    [anon_sym_unsigned] = ACTIONS(7287),
    [anon_sym_long] = ACTIONS(7287),
    [anon_sym_short] = ACTIONS(7287),
    [anon_sym_static] = ACTIONS(7435),
    [anon_sym_auto] = ACTIONS(7435),
    [anon_sym_register] = ACTIONS(7435),
    [anon_sym_inline] = ACTIONS(7435),
    [anon_sym___inline] = ACTIONS(7435),
    [anon_sym___inline__] = ACTIONS(7435),
    [anon_sym___forceinline] = ACTIONS(7435),
    [anon_sym_thread_local] = ACTIONS(7435),
    [anon_sym___thread] = ACTIONS(7435),
    [anon_sym_CG_EXTERN] = ACTIONS(7435),
    [anon_sym_CG_INLINE] = ACTIONS(7435),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7435),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7435),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7435),
    [anon_sym_IBOutlet] = ACTIONS(7435),
    [anon_sym_IBInspectable] = ACTIONS(7435),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7435),
    [anon_sym_NS_INLINE] = ACTIONS(7435),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7435),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7435),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7435),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7435),
    [anon_sym_const] = ACTIONS(7435),
    [anon_sym_constexpr] = ACTIONS(7435),
    [anon_sym_volatile] = ACTIONS(7435),
    [anon_sym_restrict] = ACTIONS(7435),
    [anon_sym___restrict__] = ACTIONS(7435),
    [anon_sym__Atomic] = ACTIONS(7435),
    [anon_sym__Noreturn] = ACTIONS(7435),
    [anon_sym__Nonnull] = ACTIONS(7435),
    [anon_sym_nullable] = ACTIONS(7435),
    [anon_sym__Complex] = ACTIONS(7435),
    [anon_sym__Nullable] = ACTIONS(7435),
    [anon_sym__Nullable_result] = ACTIONS(7435),
    [anon_sym__Null_unspecified] = ACTIONS(7435),
    [anon_sym___autoreleasing] = ACTIONS(7435),
    [anon_sym___block] = ACTIONS(7435),
    [anon_sym___bridge] = ACTIONS(7435),
    [anon_sym___bridge_retained] = ACTIONS(7435),
    [anon_sym___bridge_transfer] = ACTIONS(7435),
    [anon_sym___complex] = ACTIONS(7435),
    [anon_sym___const] = ACTIONS(7435),
    [anon_sym___imag] = ACTIONS(7435),
    [anon_sym___kindof] = ACTIONS(7435),
    [anon_sym___nonnull] = ACTIONS(7435),
    [anon_sym___nullable] = ACTIONS(7435),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7435),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7435),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7435),
    [anon_sym___real] = ACTIONS(7435),
    [anon_sym___strong] = ACTIONS(7435),
    [anon_sym___unsafe_unretained] = ACTIONS(7435),
    [anon_sym___unused] = ACTIONS(7435),
    [anon_sym___weak] = ACTIONS(7435),
    [anon_sym_alignas] = ACTIONS(7435),
    [anon_sym__Alignas] = ACTIONS(7435),
    [anon_sym_enum] = ACTIONS(7435),
    [anon_sym_COLON] = ACTIONS(7437),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7435),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7435),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7435),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7435),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7435),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7435),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7435),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7435),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7435),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7435),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7435),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7435),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7435),
    [anon_sym_API_AVAILABLE] = ACTIONS(7435),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7435),
    [anon_sym_API_DEPRECATED] = ACTIONS(7435),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7435),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7435),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7435),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7435),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7435),
    [anon_sym___deprecated_msg] = ACTIONS(7435),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7435),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7435),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7435),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7435),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7435),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7435),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7435),
  },
  [STATE(3000)] = {
    [sym_identifier] = ACTIONS(7439),
    [anon_sym_LPAREN2] = ACTIONS(7441),
    [anon_sym_STAR] = ACTIONS(7441),
    [anon_sym_CARET] = ACTIONS(7441),
    [anon_sym___extension__] = ACTIONS(7439),
    [anon_sym_extern] = ACTIONS(7439),
    [anon_sym___attribute__] = ACTIONS(7439),
    [anon_sym___attribute] = ACTIONS(7439),
    [anon_sym_noreturn] = ACTIONS(7439),
    [anon_sym_LBRACK] = ACTIONS(7441),
    [anon_sym___declspec] = ACTIONS(7439),
    [anon_sym___based] = ACTIONS(7439),
    [sym_ms_restrict_modifier] = ACTIONS(7439),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(7439),
    [sym_ms_signed_ptr_modifier] = ACTIONS(7439),
    [anon_sym__unaligned] = ACTIONS(7439),
    [anon_sym___unaligned] = ACTIONS(7439),
    [anon_sym_signed] = ACTIONS(7439),
    [anon_sym_unsigned] = ACTIONS(7439),
    [anon_sym_long] = ACTIONS(7439),
    [anon_sym_short] = ACTIONS(7439),
    [anon_sym_static] = ACTIONS(7439),
    [anon_sym_auto] = ACTIONS(7439),
    [anon_sym_register] = ACTIONS(7439),
    [anon_sym_inline] = ACTIONS(7439),
    [anon_sym___inline] = ACTIONS(7439),
    [anon_sym___inline__] = ACTIONS(7439),
    [anon_sym___forceinline] = ACTIONS(7439),
    [anon_sym_thread_local] = ACTIONS(7439),
    [anon_sym___thread] = ACTIONS(7439),
    [anon_sym_CG_EXTERN] = ACTIONS(7439),
    [anon_sym_CG_INLINE] = ACTIONS(7439),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7439),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7439),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7439),
    [anon_sym_IBOutlet] = ACTIONS(7439),
    [anon_sym_IBInspectable] = ACTIONS(7439),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7439),
    [anon_sym_NS_INLINE] = ACTIONS(7439),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7439),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7439),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7439),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7439),
    [anon_sym_const] = ACTIONS(7439),
    [anon_sym_constexpr] = ACTIONS(7439),
    [anon_sym_volatile] = ACTIONS(7439),
    [anon_sym_restrict] = ACTIONS(7439),
    [anon_sym___restrict__] = ACTIONS(7439),
    [anon_sym__Atomic] = ACTIONS(7439),
    [anon_sym__Noreturn] = ACTIONS(7439),
    [anon_sym__Nonnull] = ACTIONS(7439),
    [anon_sym_nullable] = ACTIONS(7439),
    [anon_sym__Complex] = ACTIONS(7439),
    [anon_sym__Nullable] = ACTIONS(7439),
    [anon_sym__Nullable_result] = ACTIONS(7439),
    [anon_sym__Null_unspecified] = ACTIONS(7439),
    [anon_sym___autoreleasing] = ACTIONS(7439),
    [anon_sym___block] = ACTIONS(7439),
    [anon_sym___bridge] = ACTIONS(7439),
    [anon_sym___bridge_retained] = ACTIONS(7439),
    [anon_sym___bridge_transfer] = ACTIONS(7439),
    [anon_sym___complex] = ACTIONS(7439),
    [anon_sym___const] = ACTIONS(7439),
    [anon_sym___imag] = ACTIONS(7439),
    [anon_sym___kindof] = ACTIONS(7439),
    [anon_sym___nonnull] = ACTIONS(7439),
    [anon_sym___nullable] = ACTIONS(7439),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7439),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7439),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7439),
    [anon_sym___real] = ACTIONS(7439),
    [anon_sym___strong] = ACTIONS(7439),
    [anon_sym___unsafe_unretained] = ACTIONS(7439),
    [anon_sym___unused] = ACTIONS(7439),
    [anon_sym___weak] = ACTIONS(7439),
    [anon_sym_alignas] = ACTIONS(7439),
    [anon_sym__Alignas] = ACTIONS(7439),
    [sym_primitive_type] = ACTIONS(7439),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7439),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7439),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7439),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7439),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7439),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7439),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7439),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7439),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7439),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7439),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7439),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7439),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7439),
    [anon_sym_API_AVAILABLE] = ACTIONS(7439),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7439),
    [anon_sym_API_DEPRECATED] = ACTIONS(7439),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7439),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7439),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7439),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7439),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7439),
    [anon_sym___deprecated_msg] = ACTIONS(7439),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7439),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7439),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7439),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7439),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7439),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7439),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7439),
  },
  [STATE(3001)] = {
    [sym_identifier] = ACTIONS(7443),
    [anon_sym_LPAREN2] = ACTIONS(7445),
    [anon_sym_STAR] = ACTIONS(7445),
    [anon_sym_CARET] = ACTIONS(7445),
    [anon_sym___extension__] = ACTIONS(7443),
    [anon_sym_extern] = ACTIONS(7443),
    [anon_sym___attribute__] = ACTIONS(7443),
    [anon_sym___attribute] = ACTIONS(7443),
    [anon_sym_noreturn] = ACTIONS(7443),
    [anon_sym_LBRACK] = ACTIONS(7445),
    [anon_sym___declspec] = ACTIONS(7443),
    [anon_sym___based] = ACTIONS(7443),
    [sym_ms_restrict_modifier] = ACTIONS(7443),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(7443),
    [sym_ms_signed_ptr_modifier] = ACTIONS(7443),
    [anon_sym__unaligned] = ACTIONS(7443),
    [anon_sym___unaligned] = ACTIONS(7443),
    [anon_sym_signed] = ACTIONS(7443),
    [anon_sym_unsigned] = ACTIONS(7443),
    [anon_sym_long] = ACTIONS(7443),
    [anon_sym_short] = ACTIONS(7443),
    [anon_sym_static] = ACTIONS(7443),
    [anon_sym_auto] = ACTIONS(7443),
    [anon_sym_register] = ACTIONS(7443),
    [anon_sym_inline] = ACTIONS(7443),
    [anon_sym___inline] = ACTIONS(7443),
    [anon_sym___inline__] = ACTIONS(7443),
    [anon_sym___forceinline] = ACTIONS(7443),
    [anon_sym_thread_local] = ACTIONS(7443),
    [anon_sym___thread] = ACTIONS(7443),
    [anon_sym_CG_EXTERN] = ACTIONS(7443),
    [anon_sym_CG_INLINE] = ACTIONS(7443),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7443),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7443),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7443),
    [anon_sym_IBOutlet] = ACTIONS(7443),
    [anon_sym_IBInspectable] = ACTIONS(7443),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7443),
    [anon_sym_NS_INLINE] = ACTIONS(7443),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7443),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7443),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7443),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7443),
    [anon_sym_const] = ACTIONS(7443),
    [anon_sym_constexpr] = ACTIONS(7443),
    [anon_sym_volatile] = ACTIONS(7443),
    [anon_sym_restrict] = ACTIONS(7443),
    [anon_sym___restrict__] = ACTIONS(7443),
    [anon_sym__Atomic] = ACTIONS(7443),
    [anon_sym__Noreturn] = ACTIONS(7443),
    [anon_sym__Nonnull] = ACTIONS(7443),
    [anon_sym_nullable] = ACTIONS(7443),
    [anon_sym__Complex] = ACTIONS(7443),
    [anon_sym__Nullable] = ACTIONS(7443),
    [anon_sym__Nullable_result] = ACTIONS(7443),
    [anon_sym__Null_unspecified] = ACTIONS(7443),
    [anon_sym___autoreleasing] = ACTIONS(7443),
    [anon_sym___block] = ACTIONS(7443),
    [anon_sym___bridge] = ACTIONS(7443),
    [anon_sym___bridge_retained] = ACTIONS(7443),
    [anon_sym___bridge_transfer] = ACTIONS(7443),
    [anon_sym___complex] = ACTIONS(7443),
    [anon_sym___const] = ACTIONS(7443),
    [anon_sym___imag] = ACTIONS(7443),
    [anon_sym___kindof] = ACTIONS(7443),
    [anon_sym___nonnull] = ACTIONS(7443),
    [anon_sym___nullable] = ACTIONS(7443),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7443),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7443),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7443),
    [anon_sym___real] = ACTIONS(7443),
    [anon_sym___strong] = ACTIONS(7443),
    [anon_sym___unsafe_unretained] = ACTIONS(7443),
    [anon_sym___unused] = ACTIONS(7443),
    [anon_sym___weak] = ACTIONS(7443),
    [anon_sym_alignas] = ACTIONS(7443),
    [anon_sym__Alignas] = ACTIONS(7443),
    [sym_primitive_type] = ACTIONS(7443),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7443),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7443),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7443),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7443),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7443),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7443),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7443),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7443),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7443),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7443),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7443),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7443),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7443),
    [anon_sym_API_AVAILABLE] = ACTIONS(7443),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7443),
    [anon_sym_API_DEPRECATED] = ACTIONS(7443),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7443),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7443),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7443),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7443),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7443),
    [anon_sym___deprecated_msg] = ACTIONS(7443),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7443),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7443),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7443),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7443),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7443),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7443),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7443),
  },
  [STATE(3002)] = {
    [sym_identifier] = ACTIONS(7447),
    [anon_sym_COMMA] = ACTIONS(7449),
    [anon_sym_RPAREN] = ACTIONS(7449),
    [anon_sym_LPAREN2] = ACTIONS(7449),
    [anon_sym_STAR] = ACTIONS(7449),
    [anon_sym_CARET] = ACTIONS(7449),
    [anon_sym_SEMI] = ACTIONS(7449),
    [anon_sym___extension__] = ACTIONS(7447),
    [anon_sym_extern] = ACTIONS(7447),
    [anon_sym___attribute__] = ACTIONS(7447),
    [anon_sym___attribute] = ACTIONS(7447),
    [anon_sym_noreturn] = ACTIONS(7447),
    [anon_sym_LBRACK] = ACTIONS(7449),
    [anon_sym___declspec] = ACTIONS(7447),
    [anon_sym___based] = ACTIONS(7447),
    [anon_sym_signed] = ACTIONS(7447),
    [anon_sym_unsigned] = ACTIONS(7447),
    [anon_sym_long] = ACTIONS(7447),
    [anon_sym_short] = ACTIONS(7447),
    [anon_sym_static] = ACTIONS(7447),
    [anon_sym_auto] = ACTIONS(7447),
    [anon_sym_register] = ACTIONS(7447),
    [anon_sym_inline] = ACTIONS(7447),
    [anon_sym___inline] = ACTIONS(7447),
    [anon_sym___inline__] = ACTIONS(7447),
    [anon_sym___forceinline] = ACTIONS(7447),
    [anon_sym_thread_local] = ACTIONS(7447),
    [anon_sym___thread] = ACTIONS(7447),
    [anon_sym_CG_EXTERN] = ACTIONS(7447),
    [anon_sym_CG_INLINE] = ACTIONS(7447),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7447),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7447),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7447),
    [anon_sym_IBOutlet] = ACTIONS(7447),
    [anon_sym_IBInspectable] = ACTIONS(7447),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7447),
    [anon_sym_NS_INLINE] = ACTIONS(7447),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7447),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7447),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7447),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7447),
    [anon_sym_const] = ACTIONS(7447),
    [anon_sym_constexpr] = ACTIONS(7447),
    [anon_sym_volatile] = ACTIONS(7447),
    [anon_sym_restrict] = ACTIONS(7447),
    [anon_sym___restrict__] = ACTIONS(7447),
    [anon_sym__Atomic] = ACTIONS(7447),
    [anon_sym__Noreturn] = ACTIONS(7447),
    [anon_sym__Nonnull] = ACTIONS(7447),
    [anon_sym_nullable] = ACTIONS(7447),
    [anon_sym__Complex] = ACTIONS(7447),
    [anon_sym__Nullable] = ACTIONS(7447),
    [anon_sym__Nullable_result] = ACTIONS(7447),
    [anon_sym__Null_unspecified] = ACTIONS(7447),
    [anon_sym___autoreleasing] = ACTIONS(7447),
    [anon_sym___block] = ACTIONS(7447),
    [anon_sym___bridge] = ACTIONS(7447),
    [anon_sym___bridge_retained] = ACTIONS(7447),
    [anon_sym___bridge_transfer] = ACTIONS(7447),
    [anon_sym___complex] = ACTIONS(7447),
    [anon_sym___const] = ACTIONS(7447),
    [anon_sym___imag] = ACTIONS(7447),
    [anon_sym___kindof] = ACTIONS(7447),
    [anon_sym___nonnull] = ACTIONS(7447),
    [anon_sym___nullable] = ACTIONS(7447),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7447),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7447),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7447),
    [anon_sym___real] = ACTIONS(7447),
    [anon_sym___strong] = ACTIONS(7447),
    [anon_sym___unsafe_unretained] = ACTIONS(7447),
    [anon_sym___unused] = ACTIONS(7447),
    [anon_sym___weak] = ACTIONS(7447),
    [anon_sym_alignas] = ACTIONS(7447),
    [anon_sym__Alignas] = ACTIONS(7447),
    [sym_primitive_type] = ACTIONS(7447),
    [anon_sym_enum] = ACTIONS(7447),
    [anon_sym_COLON] = ACTIONS(7449),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7447),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7447),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7447),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7447),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7447),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7447),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7447),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7447),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7447),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7447),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7447),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7447),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7447),
    [anon_sym_API_AVAILABLE] = ACTIONS(7447),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7447),
    [anon_sym_API_DEPRECATED] = ACTIONS(7447),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7447),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7447),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7447),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7447),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7447),
    [anon_sym___deprecated_msg] = ACTIONS(7447),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7447),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7447),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7447),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7447),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7447),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7447),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7447),
  },
  [STATE(3003)] = {
    [sym_identifier] = ACTIONS(7451),
    [anon_sym_COMMA] = ACTIONS(7453),
    [anon_sym_RPAREN] = ACTIONS(7453),
    [anon_sym_LPAREN2] = ACTIONS(7453),
    [anon_sym_STAR] = ACTIONS(7453),
    [anon_sym_CARET] = ACTIONS(7453),
    [anon_sym_SEMI] = ACTIONS(7453),
    [anon_sym___extension__] = ACTIONS(7451),
    [anon_sym_extern] = ACTIONS(7451),
    [anon_sym___attribute__] = ACTIONS(7451),
    [anon_sym___attribute] = ACTIONS(7451),
    [anon_sym_noreturn] = ACTIONS(7451),
    [anon_sym_LBRACK] = ACTIONS(7453),
    [anon_sym___declspec] = ACTIONS(7451),
    [anon_sym___based] = ACTIONS(7451),
    [anon_sym_signed] = ACTIONS(7451),
    [anon_sym_unsigned] = ACTIONS(7451),
    [anon_sym_long] = ACTIONS(7451),
    [anon_sym_short] = ACTIONS(7451),
    [anon_sym_static] = ACTIONS(7451),
    [anon_sym_auto] = ACTIONS(7451),
    [anon_sym_register] = ACTIONS(7451),
    [anon_sym_inline] = ACTIONS(7451),
    [anon_sym___inline] = ACTIONS(7451),
    [anon_sym___inline__] = ACTIONS(7451),
    [anon_sym___forceinline] = ACTIONS(7451),
    [anon_sym_thread_local] = ACTIONS(7451),
    [anon_sym___thread] = ACTIONS(7451),
    [anon_sym_CG_EXTERN] = ACTIONS(7451),
    [anon_sym_CG_INLINE] = ACTIONS(7451),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7451),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7451),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7451),
    [anon_sym_IBOutlet] = ACTIONS(7451),
    [anon_sym_IBInspectable] = ACTIONS(7451),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7451),
    [anon_sym_NS_INLINE] = ACTIONS(7451),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7451),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7451),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7451),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7451),
    [anon_sym_const] = ACTIONS(7451),
    [anon_sym_constexpr] = ACTIONS(7451),
    [anon_sym_volatile] = ACTIONS(7451),
    [anon_sym_restrict] = ACTIONS(7451),
    [anon_sym___restrict__] = ACTIONS(7451),
    [anon_sym__Atomic] = ACTIONS(7451),
    [anon_sym__Noreturn] = ACTIONS(7451),
    [anon_sym__Nonnull] = ACTIONS(7451),
    [anon_sym_nullable] = ACTIONS(7451),
    [anon_sym__Complex] = ACTIONS(7451),
    [anon_sym__Nullable] = ACTIONS(7451),
    [anon_sym__Nullable_result] = ACTIONS(7451),
    [anon_sym__Null_unspecified] = ACTIONS(7451),
    [anon_sym___autoreleasing] = ACTIONS(7451),
    [anon_sym___block] = ACTIONS(7451),
    [anon_sym___bridge] = ACTIONS(7451),
    [anon_sym___bridge_retained] = ACTIONS(7451),
    [anon_sym___bridge_transfer] = ACTIONS(7451),
    [anon_sym___complex] = ACTIONS(7451),
    [anon_sym___const] = ACTIONS(7451),
    [anon_sym___imag] = ACTIONS(7451),
    [anon_sym___kindof] = ACTIONS(7451),
    [anon_sym___nonnull] = ACTIONS(7451),
    [anon_sym___nullable] = ACTIONS(7451),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7451),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7451),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7451),
    [anon_sym___real] = ACTIONS(7451),
    [anon_sym___strong] = ACTIONS(7451),
    [anon_sym___unsafe_unretained] = ACTIONS(7451),
    [anon_sym___unused] = ACTIONS(7451),
    [anon_sym___weak] = ACTIONS(7451),
    [anon_sym_alignas] = ACTIONS(7451),
    [anon_sym__Alignas] = ACTIONS(7451),
    [sym_primitive_type] = ACTIONS(7451),
    [anon_sym_enum] = ACTIONS(7451),
    [anon_sym_COLON] = ACTIONS(7453),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7451),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7451),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7451),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7451),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7451),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7451),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7451),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7451),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7451),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7451),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7451),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7451),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7451),
    [anon_sym_API_AVAILABLE] = ACTIONS(7451),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7451),
    [anon_sym_API_DEPRECATED] = ACTIONS(7451),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7451),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7451),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7451),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7451),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7451),
    [anon_sym___deprecated_msg] = ACTIONS(7451),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7451),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7451),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7451),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7451),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7451),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7451),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7451),
  },
  [STATE(3004)] = {
    [sym_identifier] = ACTIONS(7455),
    [anon_sym_COMMA] = ACTIONS(7457),
    [anon_sym_RPAREN] = ACTIONS(7457),
    [anon_sym_LPAREN2] = ACTIONS(7457),
    [anon_sym_STAR] = ACTIONS(7457),
    [anon_sym_CARET] = ACTIONS(7457),
    [anon_sym_SEMI] = ACTIONS(7457),
    [anon_sym___extension__] = ACTIONS(7455),
    [anon_sym_extern] = ACTIONS(7455),
    [anon_sym___attribute__] = ACTIONS(7455),
    [anon_sym___attribute] = ACTIONS(7455),
    [anon_sym_noreturn] = ACTIONS(7455),
    [anon_sym_LBRACK] = ACTIONS(7457),
    [anon_sym___declspec] = ACTIONS(7455),
    [anon_sym___based] = ACTIONS(7455),
    [anon_sym_signed] = ACTIONS(7455),
    [anon_sym_unsigned] = ACTIONS(7455),
    [anon_sym_long] = ACTIONS(7455),
    [anon_sym_short] = ACTIONS(7455),
    [anon_sym_static] = ACTIONS(7455),
    [anon_sym_auto] = ACTIONS(7455),
    [anon_sym_register] = ACTIONS(7455),
    [anon_sym_inline] = ACTIONS(7455),
    [anon_sym___inline] = ACTIONS(7455),
    [anon_sym___inline__] = ACTIONS(7455),
    [anon_sym___forceinline] = ACTIONS(7455),
    [anon_sym_thread_local] = ACTIONS(7455),
    [anon_sym___thread] = ACTIONS(7455),
    [anon_sym_CG_EXTERN] = ACTIONS(7455),
    [anon_sym_CG_INLINE] = ACTIONS(7455),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7455),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7455),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7455),
    [anon_sym_IBOutlet] = ACTIONS(7455),
    [anon_sym_IBInspectable] = ACTIONS(7455),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7455),
    [anon_sym_NS_INLINE] = ACTIONS(7455),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7455),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7455),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7455),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7455),
    [anon_sym_const] = ACTIONS(7455),
    [anon_sym_constexpr] = ACTIONS(7455),
    [anon_sym_volatile] = ACTIONS(7455),
    [anon_sym_restrict] = ACTIONS(7455),
    [anon_sym___restrict__] = ACTIONS(7455),
    [anon_sym__Atomic] = ACTIONS(7455),
    [anon_sym__Noreturn] = ACTIONS(7455),
    [anon_sym__Nonnull] = ACTIONS(7455),
    [anon_sym_nullable] = ACTIONS(7455),
    [anon_sym__Complex] = ACTIONS(7455),
    [anon_sym__Nullable] = ACTIONS(7455),
    [anon_sym__Nullable_result] = ACTIONS(7455),
    [anon_sym__Null_unspecified] = ACTIONS(7455),
    [anon_sym___autoreleasing] = ACTIONS(7455),
    [anon_sym___block] = ACTIONS(7455),
    [anon_sym___bridge] = ACTIONS(7455),
    [anon_sym___bridge_retained] = ACTIONS(7455),
    [anon_sym___bridge_transfer] = ACTIONS(7455),
    [anon_sym___complex] = ACTIONS(7455),
    [anon_sym___const] = ACTIONS(7455),
    [anon_sym___imag] = ACTIONS(7455),
    [anon_sym___kindof] = ACTIONS(7455),
    [anon_sym___nonnull] = ACTIONS(7455),
    [anon_sym___nullable] = ACTIONS(7455),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7455),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7455),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7455),
    [anon_sym___real] = ACTIONS(7455),
    [anon_sym___strong] = ACTIONS(7455),
    [anon_sym___unsafe_unretained] = ACTIONS(7455),
    [anon_sym___unused] = ACTIONS(7455),
    [anon_sym___weak] = ACTIONS(7455),
    [anon_sym_alignas] = ACTIONS(7455),
    [anon_sym__Alignas] = ACTIONS(7455),
    [sym_primitive_type] = ACTIONS(7455),
    [anon_sym_enum] = ACTIONS(7455),
    [anon_sym_COLON] = ACTIONS(7457),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7455),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7455),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7455),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7455),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7455),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7455),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7455),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7455),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7455),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7455),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7455),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7455),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7455),
    [anon_sym_API_AVAILABLE] = ACTIONS(7455),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7455),
    [anon_sym_API_DEPRECATED] = ACTIONS(7455),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7455),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7455),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7455),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7455),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7455),
    [anon_sym___deprecated_msg] = ACTIONS(7455),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7455),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7455),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7455),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7455),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7455),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7455),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7455),
  },
  [STATE(3005)] = {
    [sym__declaration_modifiers] = STATE(2866),
    [sym_attribute_specifier] = STATE(2866),
    [sym_attribute_declaration] = STATE(2866),
    [sym_ms_declspec_modifier] = STATE(2866),
    [sym_storage_class_specifier] = STATE(2866),
    [sym_type_qualifier] = STATE(2866),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2866),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2866),
    [sym_identifier] = ACTIONS(6899),
    [anon_sym_LPAREN2] = ACTIONS(6901),
    [anon_sym_STAR] = ACTIONS(6901),
    [anon_sym_CARET] = ACTIONS(6901),
    [anon_sym_SEMI] = ACTIONS(7459),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6897),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(6899),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(3006)] = {
    [sym__declaration_modifiers] = STATE(2866),
    [sym_attribute_specifier] = STATE(2866),
    [sym_attribute_declaration] = STATE(2866),
    [sym_ms_declspec_modifier] = STATE(2866),
    [sym_storage_class_specifier] = STATE(2866),
    [sym_type_qualifier] = STATE(2866),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2866),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2866),
    [sym_identifier] = ACTIONS(6899),
    [anon_sym_LPAREN2] = ACTIONS(6901),
    [anon_sym_STAR] = ACTIONS(6901),
    [anon_sym_CARET] = ACTIONS(6901),
    [anon_sym_SEMI] = ACTIONS(7461),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(6897),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym___based] = ACTIONS(6899),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(3007)] = {
    [sym_attribute_specifier] = STATE(3021),
    [sym_gnu_asm_expression] = STATE(3020),
    [aux_sym_function_declarator_repeat1] = STATE(3021),
    [anon_sym_COMMA] = ACTIONS(6473),
    [anon_sym_RPAREN] = ACTIONS(6473),
    [anon_sym_LPAREN2] = ACTIONS(6473),
    [anon_sym_SEMI] = ACTIONS(6473),
    [anon_sym___extension__] = ACTIONS(6473),
    [anon_sym_extern] = ACTIONS(6473),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6473),
    [anon_sym_LBRACK] = ACTIONS(6473),
    [anon_sym___declspec] = ACTIONS(6473),
    [anon_sym_LBRACE] = ACTIONS(6473),
    [anon_sym_static] = ACTIONS(6473),
    [anon_sym_EQ] = ACTIONS(6473),
    [anon_sym_ATautoreleasepool] = ACTIONS(6473),
    [anon_sym_auto] = ACTIONS(6473),
    [anon_sym_register] = ACTIONS(6473),
    [anon_sym_inline] = ACTIONS(6473),
    [anon_sym___inline] = ACTIONS(6471),
    [anon_sym___inline__] = ACTIONS(6473),
    [anon_sym___forceinline] = ACTIONS(6473),
    [anon_sym_thread_local] = ACTIONS(6473),
    [anon_sym___thread] = ACTIONS(6473),
    [anon_sym_CG_EXTERN] = ACTIONS(6473),
    [anon_sym_CG_INLINE] = ACTIONS(6473),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6473),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6473),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6473),
    [anon_sym_IBOutlet] = ACTIONS(6473),
    [anon_sym_IBInspectable] = ACTIONS(6473),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6473),
    [anon_sym_NS_INLINE] = ACTIONS(6473),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6473),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6473),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6473),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6473),
    [anon_sym_const] = ACTIONS(6471),
    [anon_sym_constexpr] = ACTIONS(6473),
    [anon_sym_volatile] = ACTIONS(6473),
    [anon_sym_restrict] = ACTIONS(6473),
    [anon_sym___restrict__] = ACTIONS(6473),
    [anon_sym__Atomic] = ACTIONS(6473),
    [anon_sym__Noreturn] = ACTIONS(6473),
    [anon_sym__Nonnull] = ACTIONS(6473),
    [anon_sym_nullable] = ACTIONS(6473),
    [anon_sym__Complex] = ACTIONS(6473),
    [anon_sym__Nullable] = ACTIONS(6471),
    [anon_sym__Nullable_result] = ACTIONS(6473),
    [anon_sym__Null_unspecified] = ACTIONS(6473),
    [anon_sym___autoreleasing] = ACTIONS(6473),
    [anon_sym___block] = ACTIONS(6473),
    [anon_sym___bridge] = ACTIONS(6471),
    [anon_sym___bridge_retained] = ACTIONS(6473),
    [anon_sym___bridge_transfer] = ACTIONS(6473),
    [anon_sym___complex] = ACTIONS(6473),
    [anon_sym___const] = ACTIONS(6473),
    [anon_sym___imag] = ACTIONS(6473),
    [anon_sym___kindof] = ACTIONS(6473),
    [anon_sym___nonnull] = ACTIONS(6473),
    [anon_sym___nullable] = ACTIONS(6473),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6473),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6473),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6473),
    [anon_sym___real] = ACTIONS(6473),
    [anon_sym___strong] = ACTIONS(6473),
    [anon_sym___unsafe_unretained] = ACTIONS(6473),
    [anon_sym___unused] = ACTIONS(6473),
    [anon_sym___weak] = ACTIONS(6473),
    [anon_sym_alignas] = ACTIONS(6473),
    [anon_sym__Alignas] = ACTIONS(6473),
    [anon_sym_COLON] = ACTIONS(6473),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6473),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6473),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6473),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6473),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6473),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6473),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6473),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6473),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6473),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6473),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6471),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6473),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6473),
    [anon_sym_API_AVAILABLE] = ACTIONS(6473),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6473),
    [anon_sym_API_DEPRECATED] = ACTIONS(6473),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6473),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6473),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6473),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6473),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6473),
    [anon_sym___deprecated_msg] = ACTIONS(6473),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6473),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6473),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6473),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6473),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6473),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6473),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6473),
  },
  [STATE(3008)] = {
    [sym__declaration_modifiers] = STATE(5112),
    [sym_attribute_specifier] = STATE(5112),
    [sym_attribute_declaration] = STATE(5112),
    [sym_ms_declspec_modifier] = STATE(5112),
    [sym_storage_class_specifier] = STATE(5112),
    [sym_type_qualifier] = STATE(5112),
    [sym_alignas_qualifier] = STATE(5135),
    [sym_availability_attribute_specifier] = STATE(5112),
    [sym_method_type] = STATE(3072),
    [sym_identifier] = ACTIONS(7463),
    [anon_sym_LPAREN2] = ACTIONS(5308),
    [anon_sym___extension__] = ACTIONS(7465),
    [anon_sym_extern] = ACTIONS(7467),
    [anon_sym___attribute__] = ACTIONS(7469),
    [anon_sym___attribute] = ACTIONS(7469),
    [anon_sym_noreturn] = ACTIONS(7465),
    [anon_sym_LBRACK] = ACTIONS(7471),
    [anon_sym___declspec] = ACTIONS(7473),
    [anon_sym_static] = ACTIONS(7467),
    [anon_sym_auto] = ACTIONS(7467),
    [anon_sym_register] = ACTIONS(7467),
    [anon_sym_inline] = ACTIONS(7467),
    [anon_sym___inline] = ACTIONS(7467),
    [anon_sym___inline__] = ACTIONS(7467),
    [anon_sym___forceinline] = ACTIONS(7467),
    [anon_sym_thread_local] = ACTIONS(7467),
    [anon_sym___thread] = ACTIONS(7467),
    [anon_sym_CG_EXTERN] = ACTIONS(7467),
    [anon_sym_CG_INLINE] = ACTIONS(7467),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7467),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7467),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7467),
    [anon_sym_IBOutlet] = ACTIONS(7467),
    [anon_sym_IBInspectable] = ACTIONS(7467),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7467),
    [anon_sym_NS_INLINE] = ACTIONS(7467),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7467),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7467),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7467),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7467),
    [anon_sym_const] = ACTIONS(7475),
    [anon_sym_constexpr] = ACTIONS(7465),
    [anon_sym_volatile] = ACTIONS(7465),
    [anon_sym_restrict] = ACTIONS(7465),
    [anon_sym___restrict__] = ACTIONS(7465),
    [anon_sym__Atomic] = ACTIONS(7465),
    [anon_sym__Noreturn] = ACTIONS(7465),
    [anon_sym__Nonnull] = ACTIONS(7465),
    [anon_sym_nullable] = ACTIONS(7465),
    [anon_sym__Complex] = ACTIONS(7465),
    [anon_sym__Nullable] = ACTIONS(7465),
    [anon_sym__Nullable_result] = ACTIONS(7465),
    [anon_sym__Null_unspecified] = ACTIONS(7465),
    [anon_sym___autoreleasing] = ACTIONS(7465),
    [anon_sym___block] = ACTIONS(7465),
    [anon_sym___bridge] = ACTIONS(7465),
    [anon_sym___bridge_retained] = ACTIONS(7465),
    [anon_sym___bridge_transfer] = ACTIONS(7465),
    [anon_sym___complex] = ACTIONS(7465),
    [anon_sym___const] = ACTIONS(7465),
    [anon_sym___imag] = ACTIONS(7465),
    [anon_sym___kindof] = ACTIONS(7465),
    [anon_sym___nonnull] = ACTIONS(7465),
    [anon_sym___nullable] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7465),
    [anon_sym___real] = ACTIONS(7465),
    [anon_sym___strong] = ACTIONS(7465),
    [anon_sym___unsafe_unretained] = ACTIONS(7465),
    [anon_sym___unused] = ACTIONS(7465),
    [anon_sym___weak] = ACTIONS(7465),
    [anon_sym_alignas] = ACTIONS(7477),
    [anon_sym__Alignas] = ACTIONS(7477),
    [anon_sym_struct] = ACTIONS(7479),
    [anon_sym_in] = ACTIONS(7479),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7481),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7481),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7481),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7481),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7481),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7481),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7481),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7481),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7481),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7483),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7483),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7483),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_API_AVAILABLE] = ACTIONS(7483),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7483),
    [anon_sym_API_DEPRECATED] = ACTIONS(7483),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7483),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7483),
    [anon_sym___deprecated_msg] = ACTIONS(7483),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7483),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7483),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7483),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7483),
    [anon_sym_id] = ACTIONS(7479),
  },
  [STATE(3009)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2222),
    [anon_sym_COMMA] = ACTIONS(2222),
    [anon_sym_RPAREN] = ACTIONS(2222),
    [anon_sym_LPAREN2] = ACTIONS(2222),
    [anon_sym_BANG] = ACTIONS(4342),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(2220),
    [anon_sym_PLUS] = ACTIONS(2220),
    [anon_sym_STAR] = ACTIONS(2222),
    [anon_sym_SLASH] = ACTIONS(2220),
    [anon_sym_PERCENT] = ACTIONS(2222),
    [anon_sym_PIPE_PIPE] = ACTIONS(2222),
    [anon_sym_AMP_AMP] = ACTIONS(2222),
    [anon_sym_PIPE] = ACTIONS(2220),
    [anon_sym_CARET] = ACTIONS(2222),
    [anon_sym_AMP] = ACTIONS(2220),
    [anon_sym_EQ_EQ] = ACTIONS(2222),
    [anon_sym_BANG_EQ] = ACTIONS(2222),
    [anon_sym_GT] = ACTIONS(2220),
    [anon_sym_GT_EQ] = ACTIONS(2222),
    [anon_sym_LT_EQ] = ACTIONS(2222),
    [anon_sym_LT] = ACTIONS(2220),
    [anon_sym_LT_LT] = ACTIONS(2222),
    [anon_sym_GT_GT] = ACTIONS(2222),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2222),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_QMARK] = ACTIONS(2222),
    [anon_sym_DASH_DASH] = ACTIONS(2222),
    [anon_sym_PLUS_PLUS] = ACTIONS(2222),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(2220),
    [anon_sym_DASH_GT] = ACTIONS(2222),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3010)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_COMMA] = ACTIONS(7485),
    [anon_sym_RPAREN] = ACTIONS(7485),
    [anon_sym_LPAREN2] = ACTIONS(7042),
    [anon_sym_STAR] = ACTIONS(6134),
    [anon_sym_CARET] = ACTIONS(6134),
    [anon_sym_LT] = ACTIONS(7045),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(6134),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(3011)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_COMMA] = ACTIONS(6134),
    [anon_sym_RPAREN] = ACTIONS(6143),
    [anon_sym_LPAREN2] = ACTIONS(7488),
    [anon_sym_STAR] = ACTIONS(6134),
    [anon_sym_CARET] = ACTIONS(6134),
    [anon_sym_LT] = ACTIONS(7045),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(6143),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(3012)] = {
    [sym__declaration_modifiers] = STATE(5112),
    [sym_attribute_specifier] = STATE(5112),
    [sym_attribute_declaration] = STATE(5112),
    [sym_ms_declspec_modifier] = STATE(5112),
    [sym_storage_class_specifier] = STATE(5112),
    [sym_type_qualifier] = STATE(5112),
    [sym_alignas_qualifier] = STATE(5135),
    [sym_availability_attribute_specifier] = STATE(5112),
    [sym_method_type] = STATE(3057),
    [sym_identifier] = ACTIONS(7463),
    [anon_sym_LPAREN2] = ACTIONS(5308),
    [anon_sym___extension__] = ACTIONS(7465),
    [anon_sym_extern] = ACTIONS(7467),
    [anon_sym___attribute__] = ACTIONS(7469),
    [anon_sym___attribute] = ACTIONS(7469),
    [anon_sym_noreturn] = ACTIONS(7465),
    [anon_sym_LBRACK] = ACTIONS(7471),
    [anon_sym___declspec] = ACTIONS(7473),
    [anon_sym_static] = ACTIONS(7467),
    [anon_sym_auto] = ACTIONS(7467),
    [anon_sym_register] = ACTIONS(7467),
    [anon_sym_inline] = ACTIONS(7467),
    [anon_sym___inline] = ACTIONS(7467),
    [anon_sym___inline__] = ACTIONS(7467),
    [anon_sym___forceinline] = ACTIONS(7467),
    [anon_sym_thread_local] = ACTIONS(7467),
    [anon_sym___thread] = ACTIONS(7467),
    [anon_sym_CG_EXTERN] = ACTIONS(7467),
    [anon_sym_CG_INLINE] = ACTIONS(7467),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7467),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7467),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7467),
    [anon_sym_IBOutlet] = ACTIONS(7467),
    [anon_sym_IBInspectable] = ACTIONS(7467),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7467),
    [anon_sym_NS_INLINE] = ACTIONS(7467),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7467),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7467),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7467),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7467),
    [anon_sym_const] = ACTIONS(7492),
    [anon_sym_constexpr] = ACTIONS(7465),
    [anon_sym_volatile] = ACTIONS(7465),
    [anon_sym_restrict] = ACTIONS(7465),
    [anon_sym___restrict__] = ACTIONS(7465),
    [anon_sym__Atomic] = ACTIONS(7465),
    [anon_sym__Noreturn] = ACTIONS(7465),
    [anon_sym__Nonnull] = ACTIONS(7465),
    [anon_sym_nullable] = ACTIONS(7465),
    [anon_sym__Complex] = ACTIONS(7465),
    [anon_sym__Nullable] = ACTIONS(7465),
    [anon_sym__Nullable_result] = ACTIONS(7465),
    [anon_sym__Null_unspecified] = ACTIONS(7465),
    [anon_sym___autoreleasing] = ACTIONS(7465),
    [anon_sym___block] = ACTIONS(7465),
    [anon_sym___bridge] = ACTIONS(7465),
    [anon_sym___bridge_retained] = ACTIONS(7465),
    [anon_sym___bridge_transfer] = ACTIONS(7465),
    [anon_sym___complex] = ACTIONS(7465),
    [anon_sym___const] = ACTIONS(7465),
    [anon_sym___imag] = ACTIONS(7465),
    [anon_sym___kindof] = ACTIONS(7465),
    [anon_sym___nonnull] = ACTIONS(7465),
    [anon_sym___nullable] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7465),
    [anon_sym___real] = ACTIONS(7465),
    [anon_sym___strong] = ACTIONS(7465),
    [anon_sym___unsafe_unretained] = ACTIONS(7465),
    [anon_sym___unused] = ACTIONS(7465),
    [anon_sym___weak] = ACTIONS(7465),
    [anon_sym_alignas] = ACTIONS(7477),
    [anon_sym__Alignas] = ACTIONS(7477),
    [anon_sym_struct] = ACTIONS(7479),
    [anon_sym_in] = ACTIONS(7479),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7481),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7481),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7481),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7481),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7481),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7481),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7481),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7481),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7481),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7483),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7483),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7483),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_API_AVAILABLE] = ACTIONS(7483),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7483),
    [anon_sym_API_DEPRECATED] = ACTIONS(7483),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7483),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7483),
    [anon_sym___deprecated_msg] = ACTIONS(7483),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7483),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7483),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7483),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7483),
    [anon_sym_id] = ACTIONS(7479),
  },
  [STATE(3013)] = {
    [sym__declaration_modifiers] = STATE(3195),
    [sym_attribute_specifier] = STATE(3195),
    [sym_attribute_declaration] = STATE(3195),
    [sym_ms_declspec_modifier] = STATE(3195),
    [sym_storage_class_specifier] = STATE(3195),
    [sym_type_qualifier] = STATE(3195),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3195),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3195),
    [aux_sym_method_definition_repeat2] = STATE(3265),
    [aux_sym_method_declaration_repeat2] = STATE(2270),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7496),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3014)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_LPAREN2] = ACTIONS(7042),
    [anon_sym_STAR] = ACTIONS(6134),
    [anon_sym_CARET] = ACTIONS(6134),
    [anon_sym_LT] = ACTIONS(7045),
    [anon_sym_SEMI] = ACTIONS(7289),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(6134),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(3015)] = {
    [sym__declaration_modifiers] = STATE(3184),
    [sym_attribute_specifier] = STATE(3184),
    [sym_attribute_declaration] = STATE(3184),
    [sym_ms_declspec_modifier] = STATE(3184),
    [sym_storage_class_specifier] = STATE(3184),
    [sym_type_qualifier] = STATE(3184),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3184),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3184),
    [aux_sym_method_definition_repeat2] = STATE(3035),
    [aux_sym_method_declaration_repeat2] = STATE(2167),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7498),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3016)] = {
    [sym__declaration_modifiers] = STATE(3179),
    [sym_attribute_specifier] = STATE(3179),
    [sym_attribute_declaration] = STATE(3179),
    [sym_ms_declspec_modifier] = STATE(3179),
    [sym_storage_class_specifier] = STATE(3179),
    [sym_type_qualifier] = STATE(3179),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3179),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3179),
    [aux_sym_method_definition_repeat2] = STATE(3265),
    [aux_sym_method_declaration_repeat2] = STATE(2183),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7500),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3017)] = {
    [sym__declaration_modifiers] = STATE(3191),
    [sym_attribute_specifier] = STATE(3191),
    [sym_attribute_declaration] = STATE(3191),
    [sym_ms_declspec_modifier] = STATE(3191),
    [sym_storage_class_specifier] = STATE(3191),
    [sym_type_qualifier] = STATE(3191),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3191),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3191),
    [aux_sym_method_definition_repeat2] = STATE(3013),
    [aux_sym_method_declaration_repeat2] = STATE(2290),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7502),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3018)] = {
    [sym__declaration_modifiers] = STATE(3122),
    [sym_attribute_specifier] = STATE(3122),
    [sym_attribute_declaration] = STATE(3122),
    [sym_ms_declspec_modifier] = STATE(3122),
    [sym_storage_class_specifier] = STATE(3122),
    [sym_type_qualifier] = STATE(3122),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3122),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3122),
    [aux_sym_method_definition_repeat2] = STATE(3016),
    [aux_sym_method_declaration_repeat2] = STATE(2187),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7504),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3019)] = {
    [sym__declaration_modifiers] = STATE(3179),
    [sym_attribute_specifier] = STATE(3179),
    [sym_attribute_declaration] = STATE(3179),
    [sym_ms_declspec_modifier] = STATE(3179),
    [sym_storage_class_specifier] = STATE(3179),
    [sym_type_qualifier] = STATE(3179),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3179),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3179),
    [aux_sym_method_definition_repeat2] = STATE(3031),
    [aux_sym_method_declaration_repeat2] = STATE(2183),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7500),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3020)] = {
    [sym_attribute_specifier] = STATE(3028),
    [aux_sym_function_declarator_repeat1] = STATE(3028),
    [anon_sym_COMMA] = ACTIONS(6687),
    [anon_sym_RPAREN] = ACTIONS(6687),
    [anon_sym_LPAREN2] = ACTIONS(6687),
    [anon_sym_SEMI] = ACTIONS(6687),
    [anon_sym___extension__] = ACTIONS(6687),
    [anon_sym_extern] = ACTIONS(6687),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6687),
    [anon_sym_LBRACK] = ACTIONS(6687),
    [anon_sym___declspec] = ACTIONS(6687),
    [anon_sym_LBRACE] = ACTIONS(6687),
    [anon_sym_static] = ACTIONS(6687),
    [anon_sym_EQ] = ACTIONS(6687),
    [anon_sym_ATautoreleasepool] = ACTIONS(6687),
    [anon_sym_auto] = ACTIONS(6687),
    [anon_sym_register] = ACTIONS(6687),
    [anon_sym_inline] = ACTIONS(6687),
    [anon_sym___inline] = ACTIONS(6685),
    [anon_sym___inline__] = ACTIONS(6687),
    [anon_sym___forceinline] = ACTIONS(6687),
    [anon_sym_thread_local] = ACTIONS(6687),
    [anon_sym___thread] = ACTIONS(6687),
    [anon_sym_CG_EXTERN] = ACTIONS(6687),
    [anon_sym_CG_INLINE] = ACTIONS(6687),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6687),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6687),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6687),
    [anon_sym_IBOutlet] = ACTIONS(6687),
    [anon_sym_IBInspectable] = ACTIONS(6687),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6687),
    [anon_sym_NS_INLINE] = ACTIONS(6687),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6687),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6687),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6687),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6687),
    [anon_sym_const] = ACTIONS(6685),
    [anon_sym_constexpr] = ACTIONS(6687),
    [anon_sym_volatile] = ACTIONS(6687),
    [anon_sym_restrict] = ACTIONS(6687),
    [anon_sym___restrict__] = ACTIONS(6687),
    [anon_sym__Atomic] = ACTIONS(6687),
    [anon_sym__Noreturn] = ACTIONS(6687),
    [anon_sym__Nonnull] = ACTIONS(6687),
    [anon_sym_nullable] = ACTIONS(6687),
    [anon_sym__Complex] = ACTIONS(6687),
    [anon_sym__Nullable] = ACTIONS(6685),
    [anon_sym__Nullable_result] = ACTIONS(6687),
    [anon_sym__Null_unspecified] = ACTIONS(6687),
    [anon_sym___autoreleasing] = ACTIONS(6687),
    [anon_sym___block] = ACTIONS(6687),
    [anon_sym___bridge] = ACTIONS(6685),
    [anon_sym___bridge_retained] = ACTIONS(6687),
    [anon_sym___bridge_transfer] = ACTIONS(6687),
    [anon_sym___complex] = ACTIONS(6687),
    [anon_sym___const] = ACTIONS(6687),
    [anon_sym___imag] = ACTIONS(6687),
    [anon_sym___kindof] = ACTIONS(6687),
    [anon_sym___nonnull] = ACTIONS(6687),
    [anon_sym___nullable] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6687),
    [anon_sym___real] = ACTIONS(6687),
    [anon_sym___strong] = ACTIONS(6687),
    [anon_sym___unsafe_unretained] = ACTIONS(6687),
    [anon_sym___unused] = ACTIONS(6687),
    [anon_sym___weak] = ACTIONS(6687),
    [anon_sym_alignas] = ACTIONS(6687),
    [anon_sym__Alignas] = ACTIONS(6687),
    [anon_sym_COLON] = ACTIONS(6687),
    [anon_sym_asm] = ACTIONS(6687),
    [anon_sym___asm__] = ACTIONS(6687),
    [anon_sym___asm] = ACTIONS(6685),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6687),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6687),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6687),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6687),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6687),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6687),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6685),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_API_AVAILABLE] = ACTIONS(6687),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_API_DEPRECATED] = ACTIONS(6687),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6687),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym___deprecated_msg] = ACTIONS(6687),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6687),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6687),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6687),
  },
  [STATE(3021)] = {
    [sym_attribute_specifier] = STATE(3029),
    [aux_sym_function_declarator_repeat1] = STATE(3029),
    [anon_sym_COMMA] = ACTIONS(6687),
    [anon_sym_RPAREN] = ACTIONS(6687),
    [anon_sym_LPAREN2] = ACTIONS(6687),
    [anon_sym_SEMI] = ACTIONS(6687),
    [anon_sym___extension__] = ACTIONS(6687),
    [anon_sym_extern] = ACTIONS(6687),
    [anon_sym___attribute__] = ACTIONS(6687),
    [anon_sym___attribute] = ACTIONS(6685),
    [anon_sym_noreturn] = ACTIONS(6687),
    [anon_sym_LBRACK] = ACTIONS(6687),
    [anon_sym___declspec] = ACTIONS(6687),
    [anon_sym_LBRACE] = ACTIONS(6687),
    [anon_sym_static] = ACTIONS(6687),
    [anon_sym_EQ] = ACTIONS(6687),
    [anon_sym_ATautoreleasepool] = ACTIONS(6687),
    [anon_sym_auto] = ACTIONS(6687),
    [anon_sym_register] = ACTIONS(6687),
    [anon_sym_inline] = ACTIONS(6687),
    [anon_sym___inline] = ACTIONS(6685),
    [anon_sym___inline__] = ACTIONS(6687),
    [anon_sym___forceinline] = ACTIONS(6687),
    [anon_sym_thread_local] = ACTIONS(6687),
    [anon_sym___thread] = ACTIONS(6687),
    [anon_sym_CG_EXTERN] = ACTIONS(6687),
    [anon_sym_CG_INLINE] = ACTIONS(6687),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6687),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6687),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6687),
    [anon_sym_IBOutlet] = ACTIONS(6687),
    [anon_sym_IBInspectable] = ACTIONS(6687),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6687),
    [anon_sym_NS_INLINE] = ACTIONS(6687),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6687),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6687),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6687),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6687),
    [anon_sym_const] = ACTIONS(6685),
    [anon_sym_constexpr] = ACTIONS(6687),
    [anon_sym_volatile] = ACTIONS(6687),
    [anon_sym_restrict] = ACTIONS(6687),
    [anon_sym___restrict__] = ACTIONS(6687),
    [anon_sym__Atomic] = ACTIONS(6687),
    [anon_sym__Noreturn] = ACTIONS(6687),
    [anon_sym__Nonnull] = ACTIONS(6687),
    [anon_sym_nullable] = ACTIONS(6687),
    [anon_sym__Complex] = ACTIONS(6687),
    [anon_sym__Nullable] = ACTIONS(6685),
    [anon_sym__Nullable_result] = ACTIONS(6687),
    [anon_sym__Null_unspecified] = ACTIONS(6687),
    [anon_sym___autoreleasing] = ACTIONS(6687),
    [anon_sym___block] = ACTIONS(6687),
    [anon_sym___bridge] = ACTIONS(6685),
    [anon_sym___bridge_retained] = ACTIONS(6687),
    [anon_sym___bridge_transfer] = ACTIONS(6687),
    [anon_sym___complex] = ACTIONS(6687),
    [anon_sym___const] = ACTIONS(6687),
    [anon_sym___imag] = ACTIONS(6687),
    [anon_sym___kindof] = ACTIONS(6687),
    [anon_sym___nonnull] = ACTIONS(6687),
    [anon_sym___nullable] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6687),
    [anon_sym___real] = ACTIONS(6687),
    [anon_sym___strong] = ACTIONS(6687),
    [anon_sym___unsafe_unretained] = ACTIONS(6687),
    [anon_sym___unused] = ACTIONS(6687),
    [anon_sym___weak] = ACTIONS(6687),
    [anon_sym_alignas] = ACTIONS(6687),
    [anon_sym__Alignas] = ACTIONS(6687),
    [anon_sym_COLON] = ACTIONS(6687),
    [anon_sym_asm] = ACTIONS(6687),
    [anon_sym___asm__] = ACTIONS(6687),
    [anon_sym___asm] = ACTIONS(6685),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6687),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6687),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6687),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6687),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6687),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6687),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6685),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_API_AVAILABLE] = ACTIONS(6687),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_API_DEPRECATED] = ACTIONS(6687),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6687),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym___deprecated_msg] = ACTIONS(6687),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6687),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6687),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6687),
  },
  [STATE(3022)] = {
    [sym__declaration_modifiers] = STATE(5104),
    [sym_attribute_specifier] = STATE(5104),
    [sym_attribute_declaration] = STATE(5104),
    [sym_ms_declspec_modifier] = STATE(5104),
    [sym_storage_class_specifier] = STATE(5104),
    [sym_type_qualifier] = STATE(5104),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5104),
    [sym_identifier] = ACTIONS(7506),
    [anon_sym_COMMA] = ACTIONS(7508),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(7508),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(7508),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_RBRACE] = ACTIONS(7508),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(3023)] = {
    [sym__declaration_modifiers] = STATE(3170),
    [sym_attribute_specifier] = STATE(3170),
    [sym_attribute_declaration] = STATE(3170),
    [sym_ms_declspec_modifier] = STATE(3170),
    [sym_storage_class_specifier] = STATE(3170),
    [sym_type_qualifier] = STATE(3170),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3170),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3170),
    [aux_sym_method_definition_repeat2] = STATE(3265),
    [aux_sym_method_declaration_repeat2] = STATE(2100),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7510),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3024)] = {
    [sym__declaration_modifiers] = STATE(3191),
    [sym_attribute_specifier] = STATE(3191),
    [sym_attribute_declaration] = STATE(3191),
    [sym_ms_declspec_modifier] = STATE(3191),
    [sym_storage_class_specifier] = STATE(3191),
    [sym_type_qualifier] = STATE(3191),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3191),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3191),
    [aux_sym_method_definition_repeat2] = STATE(3265),
    [aux_sym_method_declaration_repeat2] = STATE(2290),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7502),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3025)] = {
    [sym_attribute_specifier] = STATE(2960),
    [sym_enumerator_list] = STATE(3151),
    [sym_identifier] = ACTIONS(7512),
    [anon_sym_COMMA] = ACTIONS(7514),
    [anon_sym_RPAREN] = ACTIONS(7514),
    [anon_sym_LPAREN2] = ACTIONS(7514),
    [anon_sym_STAR] = ACTIONS(7514),
    [anon_sym_CARET] = ACTIONS(7514),
    [anon_sym_SEMI] = ACTIONS(7514),
    [anon_sym___extension__] = ACTIONS(7512),
    [anon_sym_extern] = ACTIONS(7512),
    [anon_sym___attribute__] = ACTIONS(7516),
    [anon_sym___attribute] = ACTIONS(7516),
    [anon_sym_noreturn] = ACTIONS(7512),
    [anon_sym_LBRACK] = ACTIONS(7514),
    [anon_sym___declspec] = ACTIONS(7512),
    [anon_sym___based] = ACTIONS(7512),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7512),
    [anon_sym_auto] = ACTIONS(7512),
    [anon_sym_register] = ACTIONS(7512),
    [anon_sym_inline] = ACTIONS(7512),
    [anon_sym___inline] = ACTIONS(7512),
    [anon_sym___inline__] = ACTIONS(7512),
    [anon_sym___forceinline] = ACTIONS(7512),
    [anon_sym_thread_local] = ACTIONS(7512),
    [anon_sym___thread] = ACTIONS(7512),
    [anon_sym_CG_EXTERN] = ACTIONS(7512),
    [anon_sym_CG_INLINE] = ACTIONS(7512),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7512),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7512),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7512),
    [anon_sym_IBOutlet] = ACTIONS(7512),
    [anon_sym_IBInspectable] = ACTIONS(7512),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7512),
    [anon_sym_NS_INLINE] = ACTIONS(7512),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7512),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7512),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7512),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7512),
    [anon_sym_const] = ACTIONS(7512),
    [anon_sym_constexpr] = ACTIONS(7512),
    [anon_sym_volatile] = ACTIONS(7512),
    [anon_sym_restrict] = ACTIONS(7512),
    [anon_sym___restrict__] = ACTIONS(7512),
    [anon_sym__Atomic] = ACTIONS(7512),
    [anon_sym__Noreturn] = ACTIONS(7512),
    [anon_sym__Nonnull] = ACTIONS(7512),
    [anon_sym_nullable] = ACTIONS(7512),
    [anon_sym__Complex] = ACTIONS(7512),
    [anon_sym__Nullable] = ACTIONS(7512),
    [anon_sym__Nullable_result] = ACTIONS(7512),
    [anon_sym__Null_unspecified] = ACTIONS(7512),
    [anon_sym___autoreleasing] = ACTIONS(7512),
    [anon_sym___block] = ACTIONS(7512),
    [anon_sym___bridge] = ACTIONS(7512),
    [anon_sym___bridge_retained] = ACTIONS(7512),
    [anon_sym___bridge_transfer] = ACTIONS(7512),
    [anon_sym___complex] = ACTIONS(7512),
    [anon_sym___const] = ACTIONS(7512),
    [anon_sym___imag] = ACTIONS(7512),
    [anon_sym___kindof] = ACTIONS(7512),
    [anon_sym___nonnull] = ACTIONS(7512),
    [anon_sym___nullable] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7512),
    [anon_sym___real] = ACTIONS(7512),
    [anon_sym___strong] = ACTIONS(7512),
    [anon_sym___unsafe_unretained] = ACTIONS(7512),
    [anon_sym___unused] = ACTIONS(7512),
    [anon_sym___weak] = ACTIONS(7512),
    [anon_sym_alignas] = ACTIONS(7512),
    [anon_sym__Alignas] = ACTIONS(7512),
    [anon_sym_enum] = ACTIONS(7512),
    [anon_sym_COLON] = ACTIONS(7514),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7512),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7512),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7512),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7512),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7512),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7512),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7512),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7512),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7512),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7512),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7512),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7512),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7512),
    [anon_sym_API_AVAILABLE] = ACTIONS(7512),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7512),
    [anon_sym_API_DEPRECATED] = ACTIONS(7512),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7512),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7512),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7512),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7512),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7512),
    [anon_sym___deprecated_msg] = ACTIONS(7512),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7512),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7512),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7512),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7512),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7512),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7512),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7512),
  },
  [STATE(3026)] = {
    [sym_attribute_specifier] = STATE(2963),
    [sym_enumerator_list] = STATE(3157),
    [sym_identifier] = ACTIONS(7521),
    [anon_sym_COMMA] = ACTIONS(7523),
    [anon_sym_RPAREN] = ACTIONS(7523),
    [anon_sym_LPAREN2] = ACTIONS(7523),
    [anon_sym_STAR] = ACTIONS(7523),
    [anon_sym_CARET] = ACTIONS(7523),
    [anon_sym_SEMI] = ACTIONS(7523),
    [anon_sym___extension__] = ACTIONS(7521),
    [anon_sym_extern] = ACTIONS(7521),
    [anon_sym___attribute__] = ACTIONS(7525),
    [anon_sym___attribute] = ACTIONS(7525),
    [anon_sym_noreturn] = ACTIONS(7521),
    [anon_sym_LBRACK] = ACTIONS(7523),
    [anon_sym___declspec] = ACTIONS(7521),
    [anon_sym___based] = ACTIONS(7521),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7521),
    [anon_sym_auto] = ACTIONS(7521),
    [anon_sym_register] = ACTIONS(7521),
    [anon_sym_inline] = ACTIONS(7521),
    [anon_sym___inline] = ACTIONS(7521),
    [anon_sym___inline__] = ACTIONS(7521),
    [anon_sym___forceinline] = ACTIONS(7521),
    [anon_sym_thread_local] = ACTIONS(7521),
    [anon_sym___thread] = ACTIONS(7521),
    [anon_sym_CG_EXTERN] = ACTIONS(7521),
    [anon_sym_CG_INLINE] = ACTIONS(7521),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7521),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7521),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7521),
    [anon_sym_IBOutlet] = ACTIONS(7521),
    [anon_sym_IBInspectable] = ACTIONS(7521),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7521),
    [anon_sym_NS_INLINE] = ACTIONS(7521),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7521),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7521),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7521),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7521),
    [anon_sym_const] = ACTIONS(7521),
    [anon_sym_constexpr] = ACTIONS(7521),
    [anon_sym_volatile] = ACTIONS(7521),
    [anon_sym_restrict] = ACTIONS(7521),
    [anon_sym___restrict__] = ACTIONS(7521),
    [anon_sym__Atomic] = ACTIONS(7521),
    [anon_sym__Noreturn] = ACTIONS(7521),
    [anon_sym__Nonnull] = ACTIONS(7521),
    [anon_sym_nullable] = ACTIONS(7521),
    [anon_sym__Complex] = ACTIONS(7521),
    [anon_sym__Nullable] = ACTIONS(7521),
    [anon_sym__Nullable_result] = ACTIONS(7521),
    [anon_sym__Null_unspecified] = ACTIONS(7521),
    [anon_sym___autoreleasing] = ACTIONS(7521),
    [anon_sym___block] = ACTIONS(7521),
    [anon_sym___bridge] = ACTIONS(7521),
    [anon_sym___bridge_retained] = ACTIONS(7521),
    [anon_sym___bridge_transfer] = ACTIONS(7521),
    [anon_sym___complex] = ACTIONS(7521),
    [anon_sym___const] = ACTIONS(7521),
    [anon_sym___imag] = ACTIONS(7521),
    [anon_sym___kindof] = ACTIONS(7521),
    [anon_sym___nonnull] = ACTIONS(7521),
    [anon_sym___nullable] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7521),
    [anon_sym___real] = ACTIONS(7521),
    [anon_sym___strong] = ACTIONS(7521),
    [anon_sym___unsafe_unretained] = ACTIONS(7521),
    [anon_sym___unused] = ACTIONS(7521),
    [anon_sym___weak] = ACTIONS(7521),
    [anon_sym_alignas] = ACTIONS(7521),
    [anon_sym__Alignas] = ACTIONS(7521),
    [anon_sym_enum] = ACTIONS(7521),
    [anon_sym_COLON] = ACTIONS(7523),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7521),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7521),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7521),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7521),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7521),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7521),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7521),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7521),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7521),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7521),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7521),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7521),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7521),
    [anon_sym_API_AVAILABLE] = ACTIONS(7521),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7521),
    [anon_sym_API_DEPRECATED] = ACTIONS(7521),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7521),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7521),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7521),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7521),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7521),
    [anon_sym___deprecated_msg] = ACTIONS(7521),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7521),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7521),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7521),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7521),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7521),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7521),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7521),
  },
  [STATE(3027)] = {
    [sym__declaration_modifiers] = STATE(3195),
    [sym_attribute_specifier] = STATE(3195),
    [sym_attribute_declaration] = STATE(3195),
    [sym_ms_declspec_modifier] = STATE(3195),
    [sym_storage_class_specifier] = STATE(3195),
    [sym_type_qualifier] = STATE(3195),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3195),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3195),
    [aux_sym_method_definition_repeat2] = STATE(3043),
    [aux_sym_method_declaration_repeat2] = STATE(2270),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7496),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3028)] = {
    [sym_attribute_specifier] = STATE(3029),
    [aux_sym_function_declarator_repeat1] = STATE(3029),
    [anon_sym_COMMA] = ACTIONS(6698),
    [anon_sym_RPAREN] = ACTIONS(6698),
    [anon_sym_LPAREN2] = ACTIONS(6698),
    [anon_sym_SEMI] = ACTIONS(6698),
    [anon_sym___extension__] = ACTIONS(6698),
    [anon_sym_extern] = ACTIONS(6698),
    [anon_sym___attribute__] = ACTIONS(6698),
    [anon_sym___attribute] = ACTIONS(6696),
    [anon_sym_noreturn] = ACTIONS(6698),
    [anon_sym_LBRACK] = ACTIONS(6698),
    [anon_sym___declspec] = ACTIONS(6698),
    [anon_sym_LBRACE] = ACTIONS(6698),
    [anon_sym_static] = ACTIONS(6698),
    [anon_sym_EQ] = ACTIONS(6698),
    [anon_sym_ATautoreleasepool] = ACTIONS(6698),
    [anon_sym_auto] = ACTIONS(6698),
    [anon_sym_register] = ACTIONS(6698),
    [anon_sym_inline] = ACTIONS(6698),
    [anon_sym___inline] = ACTIONS(6696),
    [anon_sym___inline__] = ACTIONS(6698),
    [anon_sym___forceinline] = ACTIONS(6698),
    [anon_sym_thread_local] = ACTIONS(6698),
    [anon_sym___thread] = ACTIONS(6698),
    [anon_sym_CG_EXTERN] = ACTIONS(6698),
    [anon_sym_CG_INLINE] = ACTIONS(6698),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6698),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6698),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6698),
    [anon_sym_IBOutlet] = ACTIONS(6698),
    [anon_sym_IBInspectable] = ACTIONS(6698),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6698),
    [anon_sym_NS_INLINE] = ACTIONS(6698),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6698),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6698),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6698),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6698),
    [anon_sym_const] = ACTIONS(6696),
    [anon_sym_constexpr] = ACTIONS(6698),
    [anon_sym_volatile] = ACTIONS(6698),
    [anon_sym_restrict] = ACTIONS(6698),
    [anon_sym___restrict__] = ACTIONS(6698),
    [anon_sym__Atomic] = ACTIONS(6698),
    [anon_sym__Noreturn] = ACTIONS(6698),
    [anon_sym__Nonnull] = ACTIONS(6698),
    [anon_sym_nullable] = ACTIONS(6698),
    [anon_sym__Complex] = ACTIONS(6698),
    [anon_sym__Nullable] = ACTIONS(6696),
    [anon_sym__Nullable_result] = ACTIONS(6698),
    [anon_sym__Null_unspecified] = ACTIONS(6698),
    [anon_sym___autoreleasing] = ACTIONS(6698),
    [anon_sym___block] = ACTIONS(6698),
    [anon_sym___bridge] = ACTIONS(6696),
    [anon_sym___bridge_retained] = ACTIONS(6698),
    [anon_sym___bridge_transfer] = ACTIONS(6698),
    [anon_sym___complex] = ACTIONS(6698),
    [anon_sym___const] = ACTIONS(6698),
    [anon_sym___imag] = ACTIONS(6698),
    [anon_sym___kindof] = ACTIONS(6698),
    [anon_sym___nonnull] = ACTIONS(6698),
    [anon_sym___nullable] = ACTIONS(6698),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6698),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6698),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6698),
    [anon_sym___real] = ACTIONS(6698),
    [anon_sym___strong] = ACTIONS(6698),
    [anon_sym___unsafe_unretained] = ACTIONS(6698),
    [anon_sym___unused] = ACTIONS(6698),
    [anon_sym___weak] = ACTIONS(6698),
    [anon_sym_alignas] = ACTIONS(6698),
    [anon_sym__Alignas] = ACTIONS(6698),
    [anon_sym_COLON] = ACTIONS(6698),
    [anon_sym_asm] = ACTIONS(6698),
    [anon_sym___asm__] = ACTIONS(6698),
    [anon_sym___asm] = ACTIONS(6696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6698),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6698),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6698),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6698),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6698),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6698),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6698),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6698),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6698),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6698),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6696),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6698),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6698),
    [anon_sym_API_AVAILABLE] = ACTIONS(6698),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6698),
    [anon_sym_API_DEPRECATED] = ACTIONS(6698),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6698),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6698),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6698),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6698),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6698),
    [anon_sym___deprecated_msg] = ACTIONS(6698),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6698),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6698),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6698),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6698),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6698),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6698),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6698),
  },
  [STATE(3029)] = {
    [sym_attribute_specifier] = STATE(3029),
    [aux_sym_function_declarator_repeat1] = STATE(3029),
    [anon_sym_COMMA] = ACTIONS(6691),
    [anon_sym_RPAREN] = ACTIONS(6691),
    [anon_sym_LPAREN2] = ACTIONS(6691),
    [anon_sym_SEMI] = ACTIONS(6691),
    [anon_sym___extension__] = ACTIONS(6691),
    [anon_sym_extern] = ACTIONS(6691),
    [anon_sym___attribute__] = ACTIONS(7528),
    [anon_sym___attribute] = ACTIONS(6693),
    [anon_sym_noreturn] = ACTIONS(6691),
    [anon_sym_LBRACK] = ACTIONS(6691),
    [anon_sym___declspec] = ACTIONS(6691),
    [anon_sym_LBRACE] = ACTIONS(6691),
    [anon_sym_static] = ACTIONS(6691),
    [anon_sym_EQ] = ACTIONS(6691),
    [anon_sym_ATautoreleasepool] = ACTIONS(6691),
    [anon_sym_auto] = ACTIONS(6691),
    [anon_sym_register] = ACTIONS(6691),
    [anon_sym_inline] = ACTIONS(6691),
    [anon_sym___inline] = ACTIONS(6689),
    [anon_sym___inline__] = ACTIONS(6691),
    [anon_sym___forceinline] = ACTIONS(6691),
    [anon_sym_thread_local] = ACTIONS(6691),
    [anon_sym___thread] = ACTIONS(6691),
    [anon_sym_CG_EXTERN] = ACTIONS(6691),
    [anon_sym_CG_INLINE] = ACTIONS(6691),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6691),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6691),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6691),
    [anon_sym_IBOutlet] = ACTIONS(6691),
    [anon_sym_IBInspectable] = ACTIONS(6691),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6691),
    [anon_sym_NS_INLINE] = ACTIONS(6691),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6691),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6691),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6691),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6691),
    [anon_sym_const] = ACTIONS(6689),
    [anon_sym_constexpr] = ACTIONS(6691),
    [anon_sym_volatile] = ACTIONS(6691),
    [anon_sym_restrict] = ACTIONS(6691),
    [anon_sym___restrict__] = ACTIONS(6691),
    [anon_sym__Atomic] = ACTIONS(6691),
    [anon_sym__Noreturn] = ACTIONS(6691),
    [anon_sym__Nonnull] = ACTIONS(6691),
    [anon_sym_nullable] = ACTIONS(6691),
    [anon_sym__Complex] = ACTIONS(6691),
    [anon_sym__Nullable] = ACTIONS(6689),
    [anon_sym__Nullable_result] = ACTIONS(6691),
    [anon_sym__Null_unspecified] = ACTIONS(6691),
    [anon_sym___autoreleasing] = ACTIONS(6691),
    [anon_sym___block] = ACTIONS(6691),
    [anon_sym___bridge] = ACTIONS(6689),
    [anon_sym___bridge_retained] = ACTIONS(6691),
    [anon_sym___bridge_transfer] = ACTIONS(6691),
    [anon_sym___complex] = ACTIONS(6691),
    [anon_sym___const] = ACTIONS(6691),
    [anon_sym___imag] = ACTIONS(6691),
    [anon_sym___kindof] = ACTIONS(6691),
    [anon_sym___nonnull] = ACTIONS(6691),
    [anon_sym___nullable] = ACTIONS(6691),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6691),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6691),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6691),
    [anon_sym___real] = ACTIONS(6691),
    [anon_sym___strong] = ACTIONS(6691),
    [anon_sym___unsafe_unretained] = ACTIONS(6691),
    [anon_sym___unused] = ACTIONS(6691),
    [anon_sym___weak] = ACTIONS(6691),
    [anon_sym_alignas] = ACTIONS(6691),
    [anon_sym__Alignas] = ACTIONS(6691),
    [anon_sym_COLON] = ACTIONS(6691),
    [anon_sym_asm] = ACTIONS(6691),
    [anon_sym___asm__] = ACTIONS(6691),
    [anon_sym___asm] = ACTIONS(6689),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6691),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6691),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6691),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6691),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6691),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6691),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6691),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6691),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6691),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6691),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6689),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6691),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6691),
    [anon_sym_API_AVAILABLE] = ACTIONS(6691),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6691),
    [anon_sym_API_DEPRECATED] = ACTIONS(6691),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6691),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6691),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6691),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6691),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6691),
    [anon_sym___deprecated_msg] = ACTIONS(6691),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6691),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6691),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6691),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6691),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6691),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6691),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6691),
  },
  [STATE(3030)] = {
    [sym__declaration_modifiers] = STATE(3131),
    [sym_attribute_specifier] = STATE(3131),
    [sym_attribute_declaration] = STATE(3131),
    [sym_ms_declspec_modifier] = STATE(3131),
    [sym_storage_class_specifier] = STATE(3131),
    [sym_type_qualifier] = STATE(3131),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3131),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3131),
    [aux_sym_method_definition_repeat2] = STATE(3037),
    [aux_sym_method_declaration_repeat2] = STATE(2191),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7531),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3031)] = {
    [sym__declaration_modifiers] = STATE(3196),
    [sym_attribute_specifier] = STATE(3196),
    [sym_attribute_declaration] = STATE(3196),
    [sym_ms_declspec_modifier] = STATE(3196),
    [sym_storage_class_specifier] = STATE(3196),
    [sym_type_qualifier] = STATE(3196),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3196),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3196),
    [aux_sym_method_definition_repeat2] = STATE(3265),
    [aux_sym_method_declaration_repeat2] = STATE(2200),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7533),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3032)] = {
    [sym_attribute_specifier] = STATE(2958),
    [sym_enumerator_list] = STATE(3167),
    [sym_identifier] = ACTIONS(7535),
    [anon_sym_COMMA] = ACTIONS(7537),
    [anon_sym_RPAREN] = ACTIONS(7537),
    [anon_sym_LPAREN2] = ACTIONS(7537),
    [anon_sym_STAR] = ACTIONS(7537),
    [anon_sym_CARET] = ACTIONS(7537),
    [anon_sym_SEMI] = ACTIONS(7537),
    [anon_sym___extension__] = ACTIONS(7535),
    [anon_sym_extern] = ACTIONS(7535),
    [anon_sym___attribute__] = ACTIONS(7539),
    [anon_sym___attribute] = ACTIONS(7539),
    [anon_sym_noreturn] = ACTIONS(7535),
    [anon_sym_LBRACK] = ACTIONS(7537),
    [anon_sym___declspec] = ACTIONS(7535),
    [anon_sym___based] = ACTIONS(7535),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7535),
    [anon_sym_auto] = ACTIONS(7535),
    [anon_sym_register] = ACTIONS(7535),
    [anon_sym_inline] = ACTIONS(7535),
    [anon_sym___inline] = ACTIONS(7535),
    [anon_sym___inline__] = ACTIONS(7535),
    [anon_sym___forceinline] = ACTIONS(7535),
    [anon_sym_thread_local] = ACTIONS(7535),
    [anon_sym___thread] = ACTIONS(7535),
    [anon_sym_CG_EXTERN] = ACTIONS(7535),
    [anon_sym_CG_INLINE] = ACTIONS(7535),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7535),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7535),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7535),
    [anon_sym_IBOutlet] = ACTIONS(7535),
    [anon_sym_IBInspectable] = ACTIONS(7535),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7535),
    [anon_sym_NS_INLINE] = ACTIONS(7535),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7535),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7535),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7535),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7535),
    [anon_sym_const] = ACTIONS(7535),
    [anon_sym_constexpr] = ACTIONS(7535),
    [anon_sym_volatile] = ACTIONS(7535),
    [anon_sym_restrict] = ACTIONS(7535),
    [anon_sym___restrict__] = ACTIONS(7535),
    [anon_sym__Atomic] = ACTIONS(7535),
    [anon_sym__Noreturn] = ACTIONS(7535),
    [anon_sym__Nonnull] = ACTIONS(7535),
    [anon_sym_nullable] = ACTIONS(7535),
    [anon_sym__Complex] = ACTIONS(7535),
    [anon_sym__Nullable] = ACTIONS(7535),
    [anon_sym__Nullable_result] = ACTIONS(7535),
    [anon_sym__Null_unspecified] = ACTIONS(7535),
    [anon_sym___autoreleasing] = ACTIONS(7535),
    [anon_sym___block] = ACTIONS(7535),
    [anon_sym___bridge] = ACTIONS(7535),
    [anon_sym___bridge_retained] = ACTIONS(7535),
    [anon_sym___bridge_transfer] = ACTIONS(7535),
    [anon_sym___complex] = ACTIONS(7535),
    [anon_sym___const] = ACTIONS(7535),
    [anon_sym___imag] = ACTIONS(7535),
    [anon_sym___kindof] = ACTIONS(7535),
    [anon_sym___nonnull] = ACTIONS(7535),
    [anon_sym___nullable] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7535),
    [anon_sym___real] = ACTIONS(7535),
    [anon_sym___strong] = ACTIONS(7535),
    [anon_sym___unsafe_unretained] = ACTIONS(7535),
    [anon_sym___unused] = ACTIONS(7535),
    [anon_sym___weak] = ACTIONS(7535),
    [anon_sym_alignas] = ACTIONS(7535),
    [anon_sym__Alignas] = ACTIONS(7535),
    [anon_sym_enum] = ACTIONS(7535),
    [anon_sym_COLON] = ACTIONS(7537),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7535),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7535),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7535),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7535),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7535),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7535),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7535),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7535),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7535),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7535),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7535),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7535),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7535),
    [anon_sym_API_AVAILABLE] = ACTIONS(7535),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7535),
    [anon_sym_API_DEPRECATED] = ACTIONS(7535),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7535),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7535),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7535),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7535),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7535),
    [anon_sym___deprecated_msg] = ACTIONS(7535),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7535),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7535),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7535),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7535),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7535),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7535),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7535),
  },
  [STATE(3033)] = {
    [sym__declaration_modifiers] = STATE(6362),
    [sym_attribute_specifier] = STATE(6362),
    [sym_attribute_declaration] = STATE(6362),
    [sym_ms_declspec_modifier] = STATE(6362),
    [sym_storage_class_specifier] = STATE(6362),
    [sym_type_qualifier] = STATE(6362),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3223),
    [sym_availability_attribute_specifier] = STATE(6362),
    [aux_sym__type_definition_declarators_repeat1] = STATE(3155),
    [anon_sym_COMMA] = ACTIONS(7542),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(7544),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(7546),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3034)] = {
    [sym_attribute_specifier] = STATE(2965),
    [sym_enumerator_list] = STATE(3181),
    [sym_identifier] = ACTIONS(7548),
    [anon_sym_COMMA] = ACTIONS(7550),
    [anon_sym_RPAREN] = ACTIONS(7550),
    [anon_sym_LPAREN2] = ACTIONS(7550),
    [anon_sym_STAR] = ACTIONS(7550),
    [anon_sym_CARET] = ACTIONS(7550),
    [anon_sym_SEMI] = ACTIONS(7550),
    [anon_sym___extension__] = ACTIONS(7548),
    [anon_sym_extern] = ACTIONS(7548),
    [anon_sym___attribute__] = ACTIONS(7552),
    [anon_sym___attribute] = ACTIONS(7552),
    [anon_sym_noreturn] = ACTIONS(7548),
    [anon_sym_LBRACK] = ACTIONS(7550),
    [anon_sym___declspec] = ACTIONS(7548),
    [anon_sym___based] = ACTIONS(7548),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7548),
    [anon_sym_auto] = ACTIONS(7548),
    [anon_sym_register] = ACTIONS(7548),
    [anon_sym_inline] = ACTIONS(7548),
    [anon_sym___inline] = ACTIONS(7548),
    [anon_sym___inline__] = ACTIONS(7548),
    [anon_sym___forceinline] = ACTIONS(7548),
    [anon_sym_thread_local] = ACTIONS(7548),
    [anon_sym___thread] = ACTIONS(7548),
    [anon_sym_CG_EXTERN] = ACTIONS(7548),
    [anon_sym_CG_INLINE] = ACTIONS(7548),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7548),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7548),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7548),
    [anon_sym_IBOutlet] = ACTIONS(7548),
    [anon_sym_IBInspectable] = ACTIONS(7548),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7548),
    [anon_sym_NS_INLINE] = ACTIONS(7548),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7548),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7548),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7548),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7548),
    [anon_sym_const] = ACTIONS(7548),
    [anon_sym_constexpr] = ACTIONS(7548),
    [anon_sym_volatile] = ACTIONS(7548),
    [anon_sym_restrict] = ACTIONS(7548),
    [anon_sym___restrict__] = ACTIONS(7548),
    [anon_sym__Atomic] = ACTIONS(7548),
    [anon_sym__Noreturn] = ACTIONS(7548),
    [anon_sym__Nonnull] = ACTIONS(7548),
    [anon_sym_nullable] = ACTIONS(7548),
    [anon_sym__Complex] = ACTIONS(7548),
    [anon_sym__Nullable] = ACTIONS(7548),
    [anon_sym__Nullable_result] = ACTIONS(7548),
    [anon_sym__Null_unspecified] = ACTIONS(7548),
    [anon_sym___autoreleasing] = ACTIONS(7548),
    [anon_sym___block] = ACTIONS(7548),
    [anon_sym___bridge] = ACTIONS(7548),
    [anon_sym___bridge_retained] = ACTIONS(7548),
    [anon_sym___bridge_transfer] = ACTIONS(7548),
    [anon_sym___complex] = ACTIONS(7548),
    [anon_sym___const] = ACTIONS(7548),
    [anon_sym___imag] = ACTIONS(7548),
    [anon_sym___kindof] = ACTIONS(7548),
    [anon_sym___nonnull] = ACTIONS(7548),
    [anon_sym___nullable] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7548),
    [anon_sym___real] = ACTIONS(7548),
    [anon_sym___strong] = ACTIONS(7548),
    [anon_sym___unsafe_unretained] = ACTIONS(7548),
    [anon_sym___unused] = ACTIONS(7548),
    [anon_sym___weak] = ACTIONS(7548),
    [anon_sym_alignas] = ACTIONS(7548),
    [anon_sym__Alignas] = ACTIONS(7548),
    [anon_sym_enum] = ACTIONS(7548),
    [anon_sym_COLON] = ACTIONS(7550),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7548),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7548),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7548),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7548),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7548),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7548),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7548),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7548),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7548),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7548),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7548),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7548),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7548),
    [anon_sym_API_AVAILABLE] = ACTIONS(7548),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7548),
    [anon_sym_API_DEPRECATED] = ACTIONS(7548),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7548),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7548),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7548),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7548),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7548),
    [anon_sym___deprecated_msg] = ACTIONS(7548),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7548),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7548),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7548),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7548),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7548),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7548),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7548),
  },
  [STATE(3035)] = {
    [sym__declaration_modifiers] = STATE(3131),
    [sym_attribute_specifier] = STATE(3131),
    [sym_attribute_declaration] = STATE(3131),
    [sym_ms_declspec_modifier] = STATE(3131),
    [sym_storage_class_specifier] = STATE(3131),
    [sym_type_qualifier] = STATE(3131),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3131),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3131),
    [aux_sym_method_definition_repeat2] = STATE(3265),
    [aux_sym_method_declaration_repeat2] = STATE(2191),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7531),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3036)] = {
    [sym__declaration_modifiers] = STATE(3196),
    [sym_attribute_specifier] = STATE(3196),
    [sym_attribute_declaration] = STATE(3196),
    [sym_ms_declspec_modifier] = STATE(3196),
    [sym_storage_class_specifier] = STATE(3196),
    [sym_type_qualifier] = STATE(3196),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3196),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3196),
    [aux_sym_method_definition_repeat2] = STATE(3044),
    [aux_sym_method_declaration_repeat2] = STATE(2200),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7533),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3037)] = {
    [sym__declaration_modifiers] = STATE(3148),
    [sym_attribute_specifier] = STATE(3148),
    [sym_attribute_declaration] = STATE(3148),
    [sym_ms_declspec_modifier] = STATE(3148),
    [sym_storage_class_specifier] = STATE(3148),
    [sym_type_qualifier] = STATE(3148),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3148),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3148),
    [aux_sym_method_definition_repeat2] = STATE(3265),
    [aux_sym_method_declaration_repeat2] = STATE(2121),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7555),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3038)] = {
    [sym__declaration_modifiers] = STATE(5121),
    [sym_attribute_specifier] = STATE(5121),
    [sym_attribute_declaration] = STATE(5121),
    [sym_ms_declspec_modifier] = STATE(5121),
    [sym_storage_class_specifier] = STATE(5121),
    [sym_type_qualifier] = STATE(5121),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5121),
    [sym_identifier] = ACTIONS(7557),
    [anon_sym_COMMA] = ACTIONS(7559),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(7559),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(7559),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_extern] = ACTIONS(51),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_RBRACE] = ACTIONS(7559),
    [anon_sym_static] = ACTIONS(51),
    [anon_sym_auto] = ACTIONS(51),
    [anon_sym_register] = ACTIONS(51),
    [anon_sym_inline] = ACTIONS(51),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(51),
    [anon_sym___forceinline] = ACTIONS(51),
    [anon_sym_thread_local] = ACTIONS(51),
    [anon_sym___thread] = ACTIONS(51),
    [anon_sym_CG_EXTERN] = ACTIONS(51),
    [anon_sym_CG_INLINE] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(51),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(51),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(51),
    [anon_sym_IBOutlet] = ACTIONS(51),
    [anon_sym_IBInspectable] = ACTIONS(51),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(51),
    [anon_sym_NS_INLINE] = ACTIONS(51),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(51),
    [anon_sym_OBJC_EXPORT] = ACTIONS(51),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(51),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(51),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(125),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(125),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(125),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(125),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(125),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(125),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(125),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(127),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_API_AVAILABLE] = ACTIONS(127),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_API_DEPRECATED] = ACTIONS(127),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(127),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(127),
    [anon_sym___deprecated_msg] = ACTIONS(127),
    [anon_sym___deprecated_enum_msg] = ACTIONS(127),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(127),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(127),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(127),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(127),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(127),
  },
  [STATE(3039)] = {
    [sym_attribute_specifier] = STATE(2977),
    [sym_enumerator_list] = STATE(3128),
    [sym_identifier] = ACTIONS(7561),
    [anon_sym_COMMA] = ACTIONS(7563),
    [anon_sym_RPAREN] = ACTIONS(7563),
    [anon_sym_LPAREN2] = ACTIONS(7563),
    [anon_sym_STAR] = ACTIONS(7563),
    [anon_sym_CARET] = ACTIONS(7563),
    [anon_sym_SEMI] = ACTIONS(7563),
    [anon_sym___extension__] = ACTIONS(7561),
    [anon_sym_extern] = ACTIONS(7561),
    [anon_sym___attribute__] = ACTIONS(7565),
    [anon_sym___attribute] = ACTIONS(7565),
    [anon_sym_noreturn] = ACTIONS(7561),
    [anon_sym_LBRACK] = ACTIONS(7563),
    [anon_sym___declspec] = ACTIONS(7561),
    [anon_sym___based] = ACTIONS(7561),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7561),
    [anon_sym_auto] = ACTIONS(7561),
    [anon_sym_register] = ACTIONS(7561),
    [anon_sym_inline] = ACTIONS(7561),
    [anon_sym___inline] = ACTIONS(7561),
    [anon_sym___inline__] = ACTIONS(7561),
    [anon_sym___forceinline] = ACTIONS(7561),
    [anon_sym_thread_local] = ACTIONS(7561),
    [anon_sym___thread] = ACTIONS(7561),
    [anon_sym_CG_EXTERN] = ACTIONS(7561),
    [anon_sym_CG_INLINE] = ACTIONS(7561),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7561),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7561),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7561),
    [anon_sym_IBOutlet] = ACTIONS(7561),
    [anon_sym_IBInspectable] = ACTIONS(7561),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7561),
    [anon_sym_NS_INLINE] = ACTIONS(7561),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7561),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7561),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7561),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7561),
    [anon_sym_const] = ACTIONS(7561),
    [anon_sym_constexpr] = ACTIONS(7561),
    [anon_sym_volatile] = ACTIONS(7561),
    [anon_sym_restrict] = ACTIONS(7561),
    [anon_sym___restrict__] = ACTIONS(7561),
    [anon_sym__Atomic] = ACTIONS(7561),
    [anon_sym__Noreturn] = ACTIONS(7561),
    [anon_sym__Nonnull] = ACTIONS(7561),
    [anon_sym_nullable] = ACTIONS(7561),
    [anon_sym__Complex] = ACTIONS(7561),
    [anon_sym__Nullable] = ACTIONS(7561),
    [anon_sym__Nullable_result] = ACTIONS(7561),
    [anon_sym__Null_unspecified] = ACTIONS(7561),
    [anon_sym___autoreleasing] = ACTIONS(7561),
    [anon_sym___block] = ACTIONS(7561),
    [anon_sym___bridge] = ACTIONS(7561),
    [anon_sym___bridge_retained] = ACTIONS(7561),
    [anon_sym___bridge_transfer] = ACTIONS(7561),
    [anon_sym___complex] = ACTIONS(7561),
    [anon_sym___const] = ACTIONS(7561),
    [anon_sym___imag] = ACTIONS(7561),
    [anon_sym___kindof] = ACTIONS(7561),
    [anon_sym___nonnull] = ACTIONS(7561),
    [anon_sym___nullable] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7561),
    [anon_sym___real] = ACTIONS(7561),
    [anon_sym___strong] = ACTIONS(7561),
    [anon_sym___unsafe_unretained] = ACTIONS(7561),
    [anon_sym___unused] = ACTIONS(7561),
    [anon_sym___weak] = ACTIONS(7561),
    [anon_sym_alignas] = ACTIONS(7561),
    [anon_sym__Alignas] = ACTIONS(7561),
    [anon_sym_enum] = ACTIONS(7561),
    [anon_sym_COLON] = ACTIONS(7563),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7561),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7561),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7561),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7561),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7561),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7561),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7561),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7561),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7561),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7561),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7561),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7561),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7561),
    [anon_sym_API_AVAILABLE] = ACTIONS(7561),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7561),
    [anon_sym_API_DEPRECATED] = ACTIONS(7561),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7561),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7561),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7561),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7561),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7561),
    [anon_sym___deprecated_msg] = ACTIONS(7561),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7561),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7561),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7561),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7561),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7561),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7561),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7561),
  },
  [STATE(3040)] = {
    [sym_attribute_specifier] = STATE(2980),
    [sym_enumerator_list] = STATE(3134),
    [sym_identifier] = ACTIONS(7568),
    [anon_sym_COMMA] = ACTIONS(7570),
    [anon_sym_RPAREN] = ACTIONS(7570),
    [anon_sym_LPAREN2] = ACTIONS(7570),
    [anon_sym_STAR] = ACTIONS(7570),
    [anon_sym_CARET] = ACTIONS(7570),
    [anon_sym_SEMI] = ACTIONS(7570),
    [anon_sym___extension__] = ACTIONS(7568),
    [anon_sym_extern] = ACTIONS(7568),
    [anon_sym___attribute__] = ACTIONS(7572),
    [anon_sym___attribute] = ACTIONS(7572),
    [anon_sym_noreturn] = ACTIONS(7568),
    [anon_sym_LBRACK] = ACTIONS(7570),
    [anon_sym___declspec] = ACTIONS(7568),
    [anon_sym___based] = ACTIONS(7568),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7568),
    [anon_sym_auto] = ACTIONS(7568),
    [anon_sym_register] = ACTIONS(7568),
    [anon_sym_inline] = ACTIONS(7568),
    [anon_sym___inline] = ACTIONS(7568),
    [anon_sym___inline__] = ACTIONS(7568),
    [anon_sym___forceinline] = ACTIONS(7568),
    [anon_sym_thread_local] = ACTIONS(7568),
    [anon_sym___thread] = ACTIONS(7568),
    [anon_sym_CG_EXTERN] = ACTIONS(7568),
    [anon_sym_CG_INLINE] = ACTIONS(7568),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7568),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7568),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7568),
    [anon_sym_IBOutlet] = ACTIONS(7568),
    [anon_sym_IBInspectable] = ACTIONS(7568),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7568),
    [anon_sym_NS_INLINE] = ACTIONS(7568),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7568),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7568),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7568),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7568),
    [anon_sym_const] = ACTIONS(7568),
    [anon_sym_constexpr] = ACTIONS(7568),
    [anon_sym_volatile] = ACTIONS(7568),
    [anon_sym_restrict] = ACTIONS(7568),
    [anon_sym___restrict__] = ACTIONS(7568),
    [anon_sym__Atomic] = ACTIONS(7568),
    [anon_sym__Noreturn] = ACTIONS(7568),
    [anon_sym__Nonnull] = ACTIONS(7568),
    [anon_sym_nullable] = ACTIONS(7568),
    [anon_sym__Complex] = ACTIONS(7568),
    [anon_sym__Nullable] = ACTIONS(7568),
    [anon_sym__Nullable_result] = ACTIONS(7568),
    [anon_sym__Null_unspecified] = ACTIONS(7568),
    [anon_sym___autoreleasing] = ACTIONS(7568),
    [anon_sym___block] = ACTIONS(7568),
    [anon_sym___bridge] = ACTIONS(7568),
    [anon_sym___bridge_retained] = ACTIONS(7568),
    [anon_sym___bridge_transfer] = ACTIONS(7568),
    [anon_sym___complex] = ACTIONS(7568),
    [anon_sym___const] = ACTIONS(7568),
    [anon_sym___imag] = ACTIONS(7568),
    [anon_sym___kindof] = ACTIONS(7568),
    [anon_sym___nonnull] = ACTIONS(7568),
    [anon_sym___nullable] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7568),
    [anon_sym___real] = ACTIONS(7568),
    [anon_sym___strong] = ACTIONS(7568),
    [anon_sym___unsafe_unretained] = ACTIONS(7568),
    [anon_sym___unused] = ACTIONS(7568),
    [anon_sym___weak] = ACTIONS(7568),
    [anon_sym_alignas] = ACTIONS(7568),
    [anon_sym__Alignas] = ACTIONS(7568),
    [anon_sym_enum] = ACTIONS(7568),
    [anon_sym_COLON] = ACTIONS(7570),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7568),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7568),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7568),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7568),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7568),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7568),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7568),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7568),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7568),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7568),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7568),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7568),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7568),
    [anon_sym_API_AVAILABLE] = ACTIONS(7568),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7568),
    [anon_sym_API_DEPRECATED] = ACTIONS(7568),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7568),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7568),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7568),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7568),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7568),
    [anon_sym___deprecated_msg] = ACTIONS(7568),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7568),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7568),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7568),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7568),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7568),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7568),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7568),
  },
  [STATE(3041)] = {
    [sym__declaration_modifiers] = STATE(3148),
    [sym_attribute_specifier] = STATE(3148),
    [sym_attribute_declaration] = STATE(3148),
    [sym_ms_declspec_modifier] = STATE(3148),
    [sym_storage_class_specifier] = STATE(3148),
    [sym_type_qualifier] = STATE(3148),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3148),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3148),
    [aux_sym_method_definition_repeat2] = STATE(3023),
    [aux_sym_method_declaration_repeat2] = STATE(2121),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7555),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3042)] = {
    [sym__declaration_modifiers] = STATE(3127),
    [sym_attribute_specifier] = STATE(3127),
    [sym_attribute_declaration] = STATE(3127),
    [sym_ms_declspec_modifier] = STATE(3127),
    [sym_storage_class_specifier] = STATE(3127),
    [sym_type_qualifier] = STATE(3127),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3127),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3127),
    [aux_sym_method_definition_repeat2] = STATE(3024),
    [aux_sym_method_declaration_repeat2] = STATE(2314),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7575),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3043)] = {
    [sym__declaration_modifiers] = STATE(3074),
    [sym_attribute_specifier] = STATE(3074),
    [sym_attribute_declaration] = STATE(3074),
    [sym_ms_declspec_modifier] = STATE(3074),
    [sym_storage_class_specifier] = STATE(3074),
    [sym_type_qualifier] = STATE(3074),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3074),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3074),
    [aux_sym_method_definition_repeat2] = STATE(3265),
    [aux_sym_method_declaration_repeat2] = STATE(2296),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7577),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3044)] = {
    [sym__declaration_modifiers] = STATE(3142),
    [sym_attribute_specifier] = STATE(3142),
    [sym_attribute_declaration] = STATE(3142),
    [sym_ms_declspec_modifier] = STATE(3142),
    [sym_storage_class_specifier] = STATE(3142),
    [sym_type_qualifier] = STATE(3142),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3142),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3142),
    [aux_sym_method_definition_repeat2] = STATE(3265),
    [aux_sym_method_declaration_repeat2] = STATE(2102),
    [anon_sym_COMMA] = ACTIONS(7494),
    [anon_sym_SEMI] = ACTIONS(7579),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3045)] = {
    [sym_attribute_specifier] = STATE(2983),
    [sym_enumerator_list] = STATE(3147),
    [sym_identifier] = ACTIONS(7581),
    [anon_sym_COMMA] = ACTIONS(7583),
    [anon_sym_RPAREN] = ACTIONS(7583),
    [anon_sym_LPAREN2] = ACTIONS(7583),
    [anon_sym_STAR] = ACTIONS(7583),
    [anon_sym_CARET] = ACTIONS(7583),
    [anon_sym_SEMI] = ACTIONS(7583),
    [anon_sym___extension__] = ACTIONS(7581),
    [anon_sym_extern] = ACTIONS(7581),
    [anon_sym___attribute__] = ACTIONS(7585),
    [anon_sym___attribute] = ACTIONS(7585),
    [anon_sym_noreturn] = ACTIONS(7581),
    [anon_sym_LBRACK] = ACTIONS(7583),
    [anon_sym___declspec] = ACTIONS(7581),
    [anon_sym___based] = ACTIONS(7581),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7581),
    [anon_sym_auto] = ACTIONS(7581),
    [anon_sym_register] = ACTIONS(7581),
    [anon_sym_inline] = ACTIONS(7581),
    [anon_sym___inline] = ACTIONS(7581),
    [anon_sym___inline__] = ACTIONS(7581),
    [anon_sym___forceinline] = ACTIONS(7581),
    [anon_sym_thread_local] = ACTIONS(7581),
    [anon_sym___thread] = ACTIONS(7581),
    [anon_sym_CG_EXTERN] = ACTIONS(7581),
    [anon_sym_CG_INLINE] = ACTIONS(7581),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7581),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7581),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7581),
    [anon_sym_IBOutlet] = ACTIONS(7581),
    [anon_sym_IBInspectable] = ACTIONS(7581),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7581),
    [anon_sym_NS_INLINE] = ACTIONS(7581),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7581),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7581),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7581),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7581),
    [anon_sym_const] = ACTIONS(7581),
    [anon_sym_constexpr] = ACTIONS(7581),
    [anon_sym_volatile] = ACTIONS(7581),
    [anon_sym_restrict] = ACTIONS(7581),
    [anon_sym___restrict__] = ACTIONS(7581),
    [anon_sym__Atomic] = ACTIONS(7581),
    [anon_sym__Noreturn] = ACTIONS(7581),
    [anon_sym__Nonnull] = ACTIONS(7581),
    [anon_sym_nullable] = ACTIONS(7581),
    [anon_sym__Complex] = ACTIONS(7581),
    [anon_sym__Nullable] = ACTIONS(7581),
    [anon_sym__Nullable_result] = ACTIONS(7581),
    [anon_sym__Null_unspecified] = ACTIONS(7581),
    [anon_sym___autoreleasing] = ACTIONS(7581),
    [anon_sym___block] = ACTIONS(7581),
    [anon_sym___bridge] = ACTIONS(7581),
    [anon_sym___bridge_retained] = ACTIONS(7581),
    [anon_sym___bridge_transfer] = ACTIONS(7581),
    [anon_sym___complex] = ACTIONS(7581),
    [anon_sym___const] = ACTIONS(7581),
    [anon_sym___imag] = ACTIONS(7581),
    [anon_sym___kindof] = ACTIONS(7581),
    [anon_sym___nonnull] = ACTIONS(7581),
    [anon_sym___nullable] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7581),
    [anon_sym___real] = ACTIONS(7581),
    [anon_sym___strong] = ACTIONS(7581),
    [anon_sym___unsafe_unretained] = ACTIONS(7581),
    [anon_sym___unused] = ACTIONS(7581),
    [anon_sym___weak] = ACTIONS(7581),
    [anon_sym_alignas] = ACTIONS(7581),
    [anon_sym__Alignas] = ACTIONS(7581),
    [anon_sym_COLON] = ACTIONS(7588),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7581),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7581),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7581),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7581),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7581),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7581),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7581),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7581),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7581),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7581),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7581),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7581),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7581),
    [anon_sym_API_AVAILABLE] = ACTIONS(7581),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7581),
    [anon_sym_API_DEPRECATED] = ACTIONS(7581),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7581),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7581),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7581),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7581),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7581),
    [anon_sym___deprecated_msg] = ACTIONS(7581),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7581),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7581),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7581),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7581),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7581),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7581),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7581),
  },
  [STATE(3046)] = {
    [sym__declaration_modifiers] = STATE(5800),
    [sym_attribute_specifier] = STATE(5800),
    [sym_attribute_declaration] = STATE(5800),
    [sym_ms_declspec_modifier] = STATE(5800),
    [sym_storage_class_specifier] = STATE(5800),
    [sym_type_qualifier] = STATE(5800),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_parameter_list] = STATE(3007),
    [sym_availability_attribute_specifier] = STATE(5800),
    [anon_sym_COMMA] = ACTIONS(7590),
    [anon_sym_RPAREN] = ACTIONS(7590),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6802),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3047)] = {
    [sym_attribute_specifier] = STATE(3047),
    [sym_type_qualifier] = STATE(3047),
    [sym_alignas_qualifier] = STATE(2354),
    [aux_sym_function_field_declarator_repeat1] = STATE(3047),
    [anon_sym_COMMA] = ACTIONS(7592),
    [anon_sym_RPAREN] = ACTIONS(7592),
    [anon_sym_LPAREN2] = ACTIONS(7592),
    [anon_sym_GT] = ACTIONS(7592),
    [anon_sym_SEMI] = ACTIONS(7592),
    [anon_sym___extension__] = ACTIONS(7594),
    [anon_sym_extern] = ACTIONS(7592),
    [anon_sym___attribute__] = ACTIONS(7597),
    [anon_sym___attribute] = ACTIONS(7600),
    [anon_sym_noreturn] = ACTIONS(7594),
    [anon_sym_LBRACK] = ACTIONS(7592),
    [anon_sym___declspec] = ACTIONS(7592),
    [anon_sym_LBRACE] = ACTIONS(7592),
    [anon_sym_static] = ACTIONS(7592),
    [anon_sym_ATautoreleasepool] = ACTIONS(7592),
    [anon_sym_auto] = ACTIONS(7592),
    [anon_sym_register] = ACTIONS(7592),
    [anon_sym_inline] = ACTIONS(7592),
    [anon_sym___inline] = ACTIONS(7603),
    [anon_sym___inline__] = ACTIONS(7592),
    [anon_sym___forceinline] = ACTIONS(7592),
    [anon_sym_thread_local] = ACTIONS(7592),
    [anon_sym___thread] = ACTIONS(7592),
    [anon_sym_CG_EXTERN] = ACTIONS(7592),
    [anon_sym_CG_INLINE] = ACTIONS(7592),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7592),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7592),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7592),
    [anon_sym_IBOutlet] = ACTIONS(7592),
    [anon_sym_IBInspectable] = ACTIONS(7592),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7592),
    [anon_sym_NS_INLINE] = ACTIONS(7592),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7592),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7592),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7592),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7592),
    [anon_sym_const] = ACTIONS(7605),
    [anon_sym_constexpr] = ACTIONS(7594),
    [anon_sym_volatile] = ACTIONS(7594),
    [anon_sym_restrict] = ACTIONS(7594),
    [anon_sym___restrict__] = ACTIONS(7594),
    [anon_sym__Atomic] = ACTIONS(7594),
    [anon_sym__Noreturn] = ACTIONS(7594),
    [anon_sym__Nonnull] = ACTIONS(7594),
    [anon_sym_nullable] = ACTIONS(7594),
    [anon_sym__Complex] = ACTIONS(7594),
    [anon_sym__Nullable] = ACTIONS(7605),
    [anon_sym__Nullable_result] = ACTIONS(7594),
    [anon_sym__Null_unspecified] = ACTIONS(7594),
    [anon_sym___autoreleasing] = ACTIONS(7594),
    [anon_sym___block] = ACTIONS(7594),
    [anon_sym___bridge] = ACTIONS(7605),
    [anon_sym___bridge_retained] = ACTIONS(7594),
    [anon_sym___bridge_transfer] = ACTIONS(7594),
    [anon_sym___complex] = ACTIONS(7594),
    [anon_sym___const] = ACTIONS(7594),
    [anon_sym___imag] = ACTIONS(7594),
    [anon_sym___kindof] = ACTIONS(7594),
    [anon_sym___nonnull] = ACTIONS(7594),
    [anon_sym___nullable] = ACTIONS(7594),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7594),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7594),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7594),
    [anon_sym___real] = ACTIONS(7594),
    [anon_sym___strong] = ACTIONS(7594),
    [anon_sym___unsafe_unretained] = ACTIONS(7594),
    [anon_sym___unused] = ACTIONS(7594),
    [anon_sym___weak] = ACTIONS(7594),
    [anon_sym_alignas] = ACTIONS(7608),
    [anon_sym__Alignas] = ACTIONS(7608),
    [anon_sym_COLON] = ACTIONS(7592),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7592),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7592),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7592),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7592),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7592),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7592),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7592),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7592),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7592),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7592),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7603),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7592),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7592),
    [anon_sym_API_AVAILABLE] = ACTIONS(7592),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7592),
    [anon_sym_API_DEPRECATED] = ACTIONS(7592),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7592),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7592),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7592),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7592),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7592),
    [anon_sym___deprecated_msg] = ACTIONS(7592),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7592),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7592),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7592),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7592),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7592),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7592),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7592),
  },
  [STATE(3048)] = {
    [anon_sym_COMMA] = ACTIONS(6650),
    [anon_sym_RPAREN] = ACTIONS(6650),
    [anon_sym_LPAREN2] = ACTIONS(6650),
    [anon_sym_SEMI] = ACTIONS(6650),
    [anon_sym___extension__] = ACTIONS(6650),
    [anon_sym_extern] = ACTIONS(6650),
    [anon_sym___attribute__] = ACTIONS(6650),
    [anon_sym___attribute] = ACTIONS(6648),
    [anon_sym_noreturn] = ACTIONS(6650),
    [anon_sym_LBRACK] = ACTIONS(6650),
    [anon_sym___declspec] = ACTIONS(6650),
    [anon_sym_LBRACE] = ACTIONS(6650),
    [anon_sym_static] = ACTIONS(6650),
    [anon_sym_EQ] = ACTIONS(6650),
    [anon_sym_ATautoreleasepool] = ACTIONS(6650),
    [anon_sym_auto] = ACTIONS(6650),
    [anon_sym_register] = ACTIONS(6650),
    [anon_sym_inline] = ACTIONS(6650),
    [anon_sym___inline] = ACTIONS(6648),
    [anon_sym___inline__] = ACTIONS(6650),
    [anon_sym___forceinline] = ACTIONS(6650),
    [anon_sym_thread_local] = ACTIONS(6650),
    [anon_sym___thread] = ACTIONS(6650),
    [anon_sym_CG_EXTERN] = ACTIONS(6650),
    [anon_sym_CG_INLINE] = ACTIONS(6650),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6650),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6650),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6650),
    [anon_sym_IBOutlet] = ACTIONS(6650),
    [anon_sym_IBInspectable] = ACTIONS(6650),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6650),
    [anon_sym_NS_INLINE] = ACTIONS(6650),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6650),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6650),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6650),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6650),
    [anon_sym_const] = ACTIONS(6648),
    [anon_sym_constexpr] = ACTIONS(6650),
    [anon_sym_volatile] = ACTIONS(6650),
    [anon_sym_restrict] = ACTIONS(6650),
    [anon_sym___restrict__] = ACTIONS(6650),
    [anon_sym__Atomic] = ACTIONS(6650),
    [anon_sym__Noreturn] = ACTIONS(6650),
    [anon_sym__Nonnull] = ACTIONS(6650),
    [anon_sym_nullable] = ACTIONS(6650),
    [anon_sym__Complex] = ACTIONS(6650),
    [anon_sym__Nullable] = ACTIONS(6648),
    [anon_sym__Nullable_result] = ACTIONS(6650),
    [anon_sym__Null_unspecified] = ACTIONS(6650),
    [anon_sym___autoreleasing] = ACTIONS(6650),
    [anon_sym___block] = ACTIONS(6650),
    [anon_sym___bridge] = ACTIONS(6648),
    [anon_sym___bridge_retained] = ACTIONS(6650),
    [anon_sym___bridge_transfer] = ACTIONS(6650),
    [anon_sym___complex] = ACTIONS(6650),
    [anon_sym___const] = ACTIONS(6650),
    [anon_sym___imag] = ACTIONS(6650),
    [anon_sym___kindof] = ACTIONS(6650),
    [anon_sym___nonnull] = ACTIONS(6650),
    [anon_sym___nullable] = ACTIONS(6650),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6650),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6650),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6650),
    [anon_sym___real] = ACTIONS(6650),
    [anon_sym___strong] = ACTIONS(6650),
    [anon_sym___unsafe_unretained] = ACTIONS(6650),
    [anon_sym___unused] = ACTIONS(6650),
    [anon_sym___weak] = ACTIONS(6650),
    [anon_sym_alignas] = ACTIONS(6650),
    [anon_sym__Alignas] = ACTIONS(6650),
    [anon_sym_COLON] = ACTIONS(6650),
    [anon_sym_in] = ACTIONS(6648),
    [anon_sym_asm] = ACTIONS(6650),
    [anon_sym___asm__] = ACTIONS(6650),
    [anon_sym___asm] = ACTIONS(6648),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6650),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6650),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6650),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6650),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6650),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6650),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6650),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6650),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6650),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6650),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6648),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6650),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6650),
    [anon_sym_API_AVAILABLE] = ACTIONS(6650),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6650),
    [anon_sym_API_DEPRECATED] = ACTIONS(6650),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6650),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6650),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6650),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6650),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6650),
    [anon_sym___deprecated_msg] = ACTIONS(6650),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6650),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6650),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6650),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6650),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6650),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6650),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6650),
  },
  [STATE(3049)] = {
    [sym_parameter_list] = STATE(3007),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_RPAREN] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6738),
    [anon_sym___extension__] = ACTIONS(6738),
    [anon_sym_extern] = ACTIONS(6738),
    [anon_sym___attribute__] = ACTIONS(6738),
    [anon_sym___attribute] = ACTIONS(6736),
    [anon_sym_noreturn] = ACTIONS(6738),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6738),
    [anon_sym_static] = ACTIONS(6738),
    [anon_sym_EQ] = ACTIONS(6738),
    [anon_sym_ATautoreleasepool] = ACTIONS(6738),
    [anon_sym_auto] = ACTIONS(6738),
    [anon_sym_register] = ACTIONS(6738),
    [anon_sym_inline] = ACTIONS(6738),
    [anon_sym___inline] = ACTIONS(6736),
    [anon_sym___inline__] = ACTIONS(6738),
    [anon_sym___forceinline] = ACTIONS(6738),
    [anon_sym_thread_local] = ACTIONS(6738),
    [anon_sym___thread] = ACTIONS(6738),
    [anon_sym_CG_EXTERN] = ACTIONS(6738),
    [anon_sym_CG_INLINE] = ACTIONS(6738),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6738),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6738),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6738),
    [anon_sym_IBOutlet] = ACTIONS(6738),
    [anon_sym_IBInspectable] = ACTIONS(6738),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6738),
    [anon_sym_NS_INLINE] = ACTIONS(6738),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6738),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6738),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6738),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6738),
    [anon_sym_const] = ACTIONS(6736),
    [anon_sym_constexpr] = ACTIONS(6738),
    [anon_sym_volatile] = ACTIONS(6738),
    [anon_sym_restrict] = ACTIONS(6738),
    [anon_sym___restrict__] = ACTIONS(6738),
    [anon_sym__Atomic] = ACTIONS(6738),
    [anon_sym__Noreturn] = ACTIONS(6738),
    [anon_sym__Nonnull] = ACTIONS(6738),
    [anon_sym_nullable] = ACTIONS(6738),
    [anon_sym__Complex] = ACTIONS(6738),
    [anon_sym__Nullable] = ACTIONS(6736),
    [anon_sym__Nullable_result] = ACTIONS(6738),
    [anon_sym__Null_unspecified] = ACTIONS(6738),
    [anon_sym___autoreleasing] = ACTIONS(6738),
    [anon_sym___block] = ACTIONS(6738),
    [anon_sym___bridge] = ACTIONS(6736),
    [anon_sym___bridge_retained] = ACTIONS(6738),
    [anon_sym___bridge_transfer] = ACTIONS(6738),
    [anon_sym___complex] = ACTIONS(6738),
    [anon_sym___const] = ACTIONS(6738),
    [anon_sym___imag] = ACTIONS(6738),
    [anon_sym___kindof] = ACTIONS(6738),
    [anon_sym___nonnull] = ACTIONS(6738),
    [anon_sym___nullable] = ACTIONS(6738),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6738),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6738),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6738),
    [anon_sym___real] = ACTIONS(6738),
    [anon_sym___strong] = ACTIONS(6738),
    [anon_sym___unsafe_unretained] = ACTIONS(6738),
    [anon_sym___unused] = ACTIONS(6738),
    [anon_sym___weak] = ACTIONS(6738),
    [anon_sym_alignas] = ACTIONS(6738),
    [anon_sym__Alignas] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6738),
    [anon_sym_asm] = ACTIONS(6738),
    [anon_sym___asm__] = ACTIONS(6738),
    [anon_sym___asm] = ACTIONS(6736),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6738),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6738),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6738),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6738),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6738),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6738),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6738),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6738),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6738),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6738),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6736),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6738),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6738),
    [anon_sym_API_AVAILABLE] = ACTIONS(6738),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6738),
    [anon_sym_API_DEPRECATED] = ACTIONS(6738),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6738),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6738),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6738),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6738),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6738),
    [anon_sym___deprecated_msg] = ACTIONS(6738),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6738),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6738),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6738),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6738),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6738),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6738),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6738),
  },
  [STATE(3050)] = {
    [sym_parameter_list] = STATE(3007),
    [anon_sym_COMMA] = ACTIONS(6742),
    [anon_sym_RPAREN] = ACTIONS(6742),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6742),
    [anon_sym___extension__] = ACTIONS(6742),
    [anon_sym_extern] = ACTIONS(6742),
    [anon_sym___attribute__] = ACTIONS(6742),
    [anon_sym___attribute] = ACTIONS(6740),
    [anon_sym_noreturn] = ACTIONS(6742),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6742),
    [anon_sym_LBRACE] = ACTIONS(6742),
    [anon_sym_static] = ACTIONS(6742),
    [anon_sym_EQ] = ACTIONS(6742),
    [anon_sym_ATautoreleasepool] = ACTIONS(6742),
    [anon_sym_auto] = ACTIONS(6742),
    [anon_sym_register] = ACTIONS(6742),
    [anon_sym_inline] = ACTIONS(6742),
    [anon_sym___inline] = ACTIONS(6740),
    [anon_sym___inline__] = ACTIONS(6742),
    [anon_sym___forceinline] = ACTIONS(6742),
    [anon_sym_thread_local] = ACTIONS(6742),
    [anon_sym___thread] = ACTIONS(6742),
    [anon_sym_CG_EXTERN] = ACTIONS(6742),
    [anon_sym_CG_INLINE] = ACTIONS(6742),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6742),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6742),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6742),
    [anon_sym_IBOutlet] = ACTIONS(6742),
    [anon_sym_IBInspectable] = ACTIONS(6742),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6742),
    [anon_sym_NS_INLINE] = ACTIONS(6742),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6742),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6742),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6742),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6742),
    [anon_sym_const] = ACTIONS(6740),
    [anon_sym_constexpr] = ACTIONS(6742),
    [anon_sym_volatile] = ACTIONS(6742),
    [anon_sym_restrict] = ACTIONS(6742),
    [anon_sym___restrict__] = ACTIONS(6742),
    [anon_sym__Atomic] = ACTIONS(6742),
    [anon_sym__Noreturn] = ACTIONS(6742),
    [anon_sym__Nonnull] = ACTIONS(6742),
    [anon_sym_nullable] = ACTIONS(6742),
    [anon_sym__Complex] = ACTIONS(6742),
    [anon_sym__Nullable] = ACTIONS(6740),
    [anon_sym__Nullable_result] = ACTIONS(6742),
    [anon_sym__Null_unspecified] = ACTIONS(6742),
    [anon_sym___autoreleasing] = ACTIONS(6742),
    [anon_sym___block] = ACTIONS(6742),
    [anon_sym___bridge] = ACTIONS(6740),
    [anon_sym___bridge_retained] = ACTIONS(6742),
    [anon_sym___bridge_transfer] = ACTIONS(6742),
    [anon_sym___complex] = ACTIONS(6742),
    [anon_sym___const] = ACTIONS(6742),
    [anon_sym___imag] = ACTIONS(6742),
    [anon_sym___kindof] = ACTIONS(6742),
    [anon_sym___nonnull] = ACTIONS(6742),
    [anon_sym___nullable] = ACTIONS(6742),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6742),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6742),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6742),
    [anon_sym___real] = ACTIONS(6742),
    [anon_sym___strong] = ACTIONS(6742),
    [anon_sym___unsafe_unretained] = ACTIONS(6742),
    [anon_sym___unused] = ACTIONS(6742),
    [anon_sym___weak] = ACTIONS(6742),
    [anon_sym_alignas] = ACTIONS(6742),
    [anon_sym__Alignas] = ACTIONS(6742),
    [anon_sym_COLON] = ACTIONS(6742),
    [anon_sym_asm] = ACTIONS(6742),
    [anon_sym___asm__] = ACTIONS(6742),
    [anon_sym___asm] = ACTIONS(6740),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6742),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6742),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6742),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6742),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6742),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6742),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6742),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6742),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6742),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6742),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6740),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6742),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6742),
    [anon_sym_API_AVAILABLE] = ACTIONS(6742),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6742),
    [anon_sym_API_DEPRECATED] = ACTIONS(6742),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6742),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6742),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6742),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6742),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6742),
    [anon_sym___deprecated_msg] = ACTIONS(6742),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6742),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6742),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6742),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6742),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6742),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6742),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6742),
  },
  [STATE(3051)] = {
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_RPAREN] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6786),
    [anon_sym_SEMI] = ACTIONS(6786),
    [anon_sym___extension__] = ACTIONS(6786),
    [anon_sym_extern] = ACTIONS(6786),
    [anon_sym___attribute__] = ACTIONS(6786),
    [anon_sym___attribute] = ACTIONS(6784),
    [anon_sym_noreturn] = ACTIONS(6786),
    [anon_sym_LBRACK] = ACTIONS(6786),
    [anon_sym___declspec] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6786),
    [anon_sym_static] = ACTIONS(6786),
    [anon_sym_EQ] = ACTIONS(6786),
    [anon_sym_ATautoreleasepool] = ACTIONS(6786),
    [anon_sym_auto] = ACTIONS(6786),
    [anon_sym_register] = ACTIONS(6786),
    [anon_sym_inline] = ACTIONS(6786),
    [anon_sym___inline] = ACTIONS(6784),
    [anon_sym___inline__] = ACTIONS(6786),
    [anon_sym___forceinline] = ACTIONS(6786),
    [anon_sym_thread_local] = ACTIONS(6786),
    [anon_sym___thread] = ACTIONS(6786),
    [anon_sym_CG_EXTERN] = ACTIONS(6786),
    [anon_sym_CG_INLINE] = ACTIONS(6786),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6786),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6786),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6786),
    [anon_sym_IBOutlet] = ACTIONS(6786),
    [anon_sym_IBInspectable] = ACTIONS(6786),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6786),
    [anon_sym_NS_INLINE] = ACTIONS(6786),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6786),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6786),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6786),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6786),
    [anon_sym_const] = ACTIONS(6784),
    [anon_sym_constexpr] = ACTIONS(6786),
    [anon_sym_volatile] = ACTIONS(6786),
    [anon_sym_restrict] = ACTIONS(6786),
    [anon_sym___restrict__] = ACTIONS(6786),
    [anon_sym__Atomic] = ACTIONS(6786),
    [anon_sym__Noreturn] = ACTIONS(6786),
    [anon_sym__Nonnull] = ACTIONS(6786),
    [anon_sym_nullable] = ACTIONS(6786),
    [anon_sym__Complex] = ACTIONS(6786),
    [anon_sym__Nullable] = ACTIONS(6784),
    [anon_sym__Nullable_result] = ACTIONS(6786),
    [anon_sym__Null_unspecified] = ACTIONS(6786),
    [anon_sym___autoreleasing] = ACTIONS(6786),
    [anon_sym___block] = ACTIONS(6786),
    [anon_sym___bridge] = ACTIONS(6784),
    [anon_sym___bridge_retained] = ACTIONS(6786),
    [anon_sym___bridge_transfer] = ACTIONS(6786),
    [anon_sym___complex] = ACTIONS(6786),
    [anon_sym___const] = ACTIONS(6786),
    [anon_sym___imag] = ACTIONS(6786),
    [anon_sym___kindof] = ACTIONS(6786),
    [anon_sym___nonnull] = ACTIONS(6786),
    [anon_sym___nullable] = ACTIONS(6786),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6786),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6786),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6786),
    [anon_sym___real] = ACTIONS(6786),
    [anon_sym___strong] = ACTIONS(6786),
    [anon_sym___unsafe_unretained] = ACTIONS(6786),
    [anon_sym___unused] = ACTIONS(6786),
    [anon_sym___weak] = ACTIONS(6786),
    [anon_sym_alignas] = ACTIONS(6786),
    [anon_sym__Alignas] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6786),
    [anon_sym_in] = ACTIONS(6784),
    [anon_sym_asm] = ACTIONS(6786),
    [anon_sym___asm__] = ACTIONS(6786),
    [anon_sym___asm] = ACTIONS(6784),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6786),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6786),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6786),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6786),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6786),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6786),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6786),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6786),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6786),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6786),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6784),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6786),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6786),
    [anon_sym_API_AVAILABLE] = ACTIONS(6786),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6786),
    [anon_sym_API_DEPRECATED] = ACTIONS(6786),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6786),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6786),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6786),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6786),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6786),
    [anon_sym___deprecated_msg] = ACTIONS(6786),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6786),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6786),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6786),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6786),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6786),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6786),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6786),
  },
  [STATE(3052)] = {
    [anon_sym_COMMA] = ACTIONS(6654),
    [anon_sym_RPAREN] = ACTIONS(6654),
    [anon_sym_LPAREN2] = ACTIONS(6654),
    [anon_sym_SEMI] = ACTIONS(6654),
    [anon_sym___extension__] = ACTIONS(6654),
    [anon_sym_extern] = ACTIONS(6654),
    [anon_sym___attribute__] = ACTIONS(6654),
    [anon_sym___attribute] = ACTIONS(6662),
    [anon_sym_noreturn] = ACTIONS(6654),
    [anon_sym_LBRACK] = ACTIONS(6654),
    [anon_sym___declspec] = ACTIONS(6654),
    [anon_sym_LBRACE] = ACTIONS(6654),
    [anon_sym_static] = ACTIONS(6654),
    [anon_sym_EQ] = ACTIONS(6654),
    [anon_sym_ATautoreleasepool] = ACTIONS(6654),
    [anon_sym_auto] = ACTIONS(6654),
    [anon_sym_register] = ACTIONS(6654),
    [anon_sym_inline] = ACTIONS(6654),
    [anon_sym___inline] = ACTIONS(6662),
    [anon_sym___inline__] = ACTIONS(6654),
    [anon_sym___forceinline] = ACTIONS(6654),
    [anon_sym_thread_local] = ACTIONS(6654),
    [anon_sym___thread] = ACTIONS(6654),
    [anon_sym_CG_EXTERN] = ACTIONS(6654),
    [anon_sym_CG_INLINE] = ACTIONS(6654),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6654),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6654),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6654),
    [anon_sym_IBOutlet] = ACTIONS(6654),
    [anon_sym_IBInspectable] = ACTIONS(6654),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6654),
    [anon_sym_NS_INLINE] = ACTIONS(6654),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6654),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6654),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6654),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6654),
    [anon_sym_const] = ACTIONS(6662),
    [anon_sym_constexpr] = ACTIONS(6654),
    [anon_sym_volatile] = ACTIONS(6654),
    [anon_sym_restrict] = ACTIONS(6654),
    [anon_sym___restrict__] = ACTIONS(6654),
    [anon_sym__Atomic] = ACTIONS(6654),
    [anon_sym__Noreturn] = ACTIONS(6654),
    [anon_sym__Nonnull] = ACTIONS(6654),
    [anon_sym_nullable] = ACTIONS(6654),
    [anon_sym__Complex] = ACTIONS(6654),
    [anon_sym__Nullable] = ACTIONS(6662),
    [anon_sym__Nullable_result] = ACTIONS(6654),
    [anon_sym__Null_unspecified] = ACTIONS(6654),
    [anon_sym___autoreleasing] = ACTIONS(6654),
    [anon_sym___block] = ACTIONS(6654),
    [anon_sym___bridge] = ACTIONS(6662),
    [anon_sym___bridge_retained] = ACTIONS(6654),
    [anon_sym___bridge_transfer] = ACTIONS(6654),
    [anon_sym___complex] = ACTIONS(6654),
    [anon_sym___const] = ACTIONS(6654),
    [anon_sym___imag] = ACTIONS(6654),
    [anon_sym___kindof] = ACTIONS(6654),
    [anon_sym___nonnull] = ACTIONS(6654),
    [anon_sym___nullable] = ACTIONS(6654),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6654),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6654),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6654),
    [anon_sym___real] = ACTIONS(6654),
    [anon_sym___strong] = ACTIONS(6654),
    [anon_sym___unsafe_unretained] = ACTIONS(6654),
    [anon_sym___unused] = ACTIONS(6654),
    [anon_sym___weak] = ACTIONS(6654),
    [anon_sym_alignas] = ACTIONS(6654),
    [anon_sym__Alignas] = ACTIONS(6654),
    [anon_sym_COLON] = ACTIONS(6654),
    [anon_sym_in] = ACTIONS(6662),
    [anon_sym_asm] = ACTIONS(6654),
    [anon_sym___asm__] = ACTIONS(6654),
    [anon_sym___asm] = ACTIONS(6662),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6654),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6654),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6654),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6654),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6654),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6654),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6654),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6654),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6654),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6654),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6662),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6654),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6654),
    [anon_sym_API_AVAILABLE] = ACTIONS(6654),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6654),
    [anon_sym_API_DEPRECATED] = ACTIONS(6654),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6654),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6654),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6654),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6654),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6654),
    [anon_sym___deprecated_msg] = ACTIONS(6654),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6654),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6654),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6654),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6654),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6654),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6654),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6654),
  },
  [STATE(3053)] = {
    [sym_parameter_list] = STATE(3007),
    [anon_sym_COMMA] = ACTIONS(6746),
    [anon_sym_RPAREN] = ACTIONS(6746),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6746),
    [anon_sym___extension__] = ACTIONS(6746),
    [anon_sym_extern] = ACTIONS(6746),
    [anon_sym___attribute__] = ACTIONS(6746),
    [anon_sym___attribute] = ACTIONS(6744),
    [anon_sym_noreturn] = ACTIONS(6746),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6746),
    [anon_sym_LBRACE] = ACTIONS(6746),
    [anon_sym_static] = ACTIONS(6746),
    [anon_sym_EQ] = ACTIONS(6746),
    [anon_sym_ATautoreleasepool] = ACTIONS(6746),
    [anon_sym_auto] = ACTIONS(6746),
    [anon_sym_register] = ACTIONS(6746),
    [anon_sym_inline] = ACTIONS(6746),
    [anon_sym___inline] = ACTIONS(6744),
    [anon_sym___inline__] = ACTIONS(6746),
    [anon_sym___forceinline] = ACTIONS(6746),
    [anon_sym_thread_local] = ACTIONS(6746),
    [anon_sym___thread] = ACTIONS(6746),
    [anon_sym_CG_EXTERN] = ACTIONS(6746),
    [anon_sym_CG_INLINE] = ACTIONS(6746),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6746),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6746),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6746),
    [anon_sym_IBOutlet] = ACTIONS(6746),
    [anon_sym_IBInspectable] = ACTIONS(6746),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6746),
    [anon_sym_NS_INLINE] = ACTIONS(6746),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6746),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6746),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6746),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6746),
    [anon_sym_const] = ACTIONS(6744),
    [anon_sym_constexpr] = ACTIONS(6746),
    [anon_sym_volatile] = ACTIONS(6746),
    [anon_sym_restrict] = ACTIONS(6746),
    [anon_sym___restrict__] = ACTIONS(6746),
    [anon_sym__Atomic] = ACTIONS(6746),
    [anon_sym__Noreturn] = ACTIONS(6746),
    [anon_sym__Nonnull] = ACTIONS(6746),
    [anon_sym_nullable] = ACTIONS(6746),
    [anon_sym__Complex] = ACTIONS(6746),
    [anon_sym__Nullable] = ACTIONS(6744),
    [anon_sym__Nullable_result] = ACTIONS(6746),
    [anon_sym__Null_unspecified] = ACTIONS(6746),
    [anon_sym___autoreleasing] = ACTIONS(6746),
    [anon_sym___block] = ACTIONS(6746),
    [anon_sym___bridge] = ACTIONS(6744),
    [anon_sym___bridge_retained] = ACTIONS(6746),
    [anon_sym___bridge_transfer] = ACTIONS(6746),
    [anon_sym___complex] = ACTIONS(6746),
    [anon_sym___const] = ACTIONS(6746),
    [anon_sym___imag] = ACTIONS(6746),
    [anon_sym___kindof] = ACTIONS(6746),
    [anon_sym___nonnull] = ACTIONS(6746),
    [anon_sym___nullable] = ACTIONS(6746),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6746),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6746),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6746),
    [anon_sym___real] = ACTIONS(6746),
    [anon_sym___strong] = ACTIONS(6746),
    [anon_sym___unsafe_unretained] = ACTIONS(6746),
    [anon_sym___unused] = ACTIONS(6746),
    [anon_sym___weak] = ACTIONS(6746),
    [anon_sym_alignas] = ACTIONS(6746),
    [anon_sym__Alignas] = ACTIONS(6746),
    [anon_sym_COLON] = ACTIONS(6746),
    [anon_sym_asm] = ACTIONS(6746),
    [anon_sym___asm__] = ACTIONS(6746),
    [anon_sym___asm] = ACTIONS(6744),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6746),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6746),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6746),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6746),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6746),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6746),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6746),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6746),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6746),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6746),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6744),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6746),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6746),
    [anon_sym_API_AVAILABLE] = ACTIONS(6746),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6746),
    [anon_sym_API_DEPRECATED] = ACTIONS(6746),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6746),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6746),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6746),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6746),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6746),
    [anon_sym___deprecated_msg] = ACTIONS(6746),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6746),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6746),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6746),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6746),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6746),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6746),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6746),
  },
  [STATE(3054)] = {
    [sym_parameter_list] = STATE(3007),
    [anon_sym_COMMA] = ACTIONS(6750),
    [anon_sym_RPAREN] = ACTIONS(6750),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6750),
    [anon_sym___extension__] = ACTIONS(6750),
    [anon_sym_extern] = ACTIONS(6750),
    [anon_sym___attribute__] = ACTIONS(6750),
    [anon_sym___attribute] = ACTIONS(6748),
    [anon_sym_noreturn] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6750),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_ATautoreleasepool] = ACTIONS(6750),
    [anon_sym_auto] = ACTIONS(6750),
    [anon_sym_register] = ACTIONS(6750),
    [anon_sym_inline] = ACTIONS(6750),
    [anon_sym___inline] = ACTIONS(6748),
    [anon_sym___inline__] = ACTIONS(6750),
    [anon_sym___forceinline] = ACTIONS(6750),
    [anon_sym_thread_local] = ACTIONS(6750),
    [anon_sym___thread] = ACTIONS(6750),
    [anon_sym_CG_EXTERN] = ACTIONS(6750),
    [anon_sym_CG_INLINE] = ACTIONS(6750),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6750),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6750),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6750),
    [anon_sym_IBOutlet] = ACTIONS(6750),
    [anon_sym_IBInspectable] = ACTIONS(6750),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6750),
    [anon_sym_NS_INLINE] = ACTIONS(6750),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6750),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6750),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6750),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6750),
    [anon_sym_const] = ACTIONS(6748),
    [anon_sym_constexpr] = ACTIONS(6750),
    [anon_sym_volatile] = ACTIONS(6750),
    [anon_sym_restrict] = ACTIONS(6750),
    [anon_sym___restrict__] = ACTIONS(6750),
    [anon_sym__Atomic] = ACTIONS(6750),
    [anon_sym__Noreturn] = ACTIONS(6750),
    [anon_sym__Nonnull] = ACTIONS(6750),
    [anon_sym_nullable] = ACTIONS(6750),
    [anon_sym__Complex] = ACTIONS(6750),
    [anon_sym__Nullable] = ACTIONS(6748),
    [anon_sym__Nullable_result] = ACTIONS(6750),
    [anon_sym__Null_unspecified] = ACTIONS(6750),
    [anon_sym___autoreleasing] = ACTIONS(6750),
    [anon_sym___block] = ACTIONS(6750),
    [anon_sym___bridge] = ACTIONS(6748),
    [anon_sym___bridge_retained] = ACTIONS(6750),
    [anon_sym___bridge_transfer] = ACTIONS(6750),
    [anon_sym___complex] = ACTIONS(6750),
    [anon_sym___const] = ACTIONS(6750),
    [anon_sym___imag] = ACTIONS(6750),
    [anon_sym___kindof] = ACTIONS(6750),
    [anon_sym___nonnull] = ACTIONS(6750),
    [anon_sym___nullable] = ACTIONS(6750),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6750),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6750),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6750),
    [anon_sym___real] = ACTIONS(6750),
    [anon_sym___strong] = ACTIONS(6750),
    [anon_sym___unsafe_unretained] = ACTIONS(6750),
    [anon_sym___unused] = ACTIONS(6750),
    [anon_sym___weak] = ACTIONS(6750),
    [anon_sym_alignas] = ACTIONS(6750),
    [anon_sym__Alignas] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_asm] = ACTIONS(6750),
    [anon_sym___asm__] = ACTIONS(6750),
    [anon_sym___asm] = ACTIONS(6748),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6750),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6750),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6750),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6750),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6750),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6750),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6750),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6750),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6750),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6750),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6748),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6750),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6750),
    [anon_sym_API_AVAILABLE] = ACTIONS(6750),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6750),
    [anon_sym_API_DEPRECATED] = ACTIONS(6750),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6750),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6750),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6750),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6750),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6750),
    [anon_sym___deprecated_msg] = ACTIONS(6750),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6750),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6750),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6750),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6750),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6750),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6750),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6750),
  },
  [STATE(3055)] = {
    [sym_parameter_list] = STATE(3007),
    [anon_sym_COMMA] = ACTIONS(6724),
    [anon_sym_RPAREN] = ACTIONS(6724),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6724),
    [anon_sym___extension__] = ACTIONS(6724),
    [anon_sym_extern] = ACTIONS(6724),
    [anon_sym___attribute__] = ACTIONS(6724),
    [anon_sym___attribute] = ACTIONS(6722),
    [anon_sym_noreturn] = ACTIONS(6724),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6724),
    [anon_sym_LBRACE] = ACTIONS(6724),
    [anon_sym_static] = ACTIONS(6724),
    [anon_sym_EQ] = ACTIONS(6724),
    [anon_sym_ATautoreleasepool] = ACTIONS(6724),
    [anon_sym_auto] = ACTIONS(6724),
    [anon_sym_register] = ACTIONS(6724),
    [anon_sym_inline] = ACTIONS(6724),
    [anon_sym___inline] = ACTIONS(6722),
    [anon_sym___inline__] = ACTIONS(6724),
    [anon_sym___forceinline] = ACTIONS(6724),
    [anon_sym_thread_local] = ACTIONS(6724),
    [anon_sym___thread] = ACTIONS(6724),
    [anon_sym_CG_EXTERN] = ACTIONS(6724),
    [anon_sym_CG_INLINE] = ACTIONS(6724),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6724),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6724),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6724),
    [anon_sym_IBOutlet] = ACTIONS(6724),
    [anon_sym_IBInspectable] = ACTIONS(6724),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6724),
    [anon_sym_NS_INLINE] = ACTIONS(6724),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6724),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6724),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6724),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6724),
    [anon_sym_const] = ACTIONS(6722),
    [anon_sym_constexpr] = ACTIONS(6724),
    [anon_sym_volatile] = ACTIONS(6724),
    [anon_sym_restrict] = ACTIONS(6724),
    [anon_sym___restrict__] = ACTIONS(6724),
    [anon_sym__Atomic] = ACTIONS(6724),
    [anon_sym__Noreturn] = ACTIONS(6724),
    [anon_sym__Nonnull] = ACTIONS(6724),
    [anon_sym_nullable] = ACTIONS(6724),
    [anon_sym__Complex] = ACTIONS(6724),
    [anon_sym__Nullable] = ACTIONS(6722),
    [anon_sym__Nullable_result] = ACTIONS(6724),
    [anon_sym__Null_unspecified] = ACTIONS(6724),
    [anon_sym___autoreleasing] = ACTIONS(6724),
    [anon_sym___block] = ACTIONS(6724),
    [anon_sym___bridge] = ACTIONS(6722),
    [anon_sym___bridge_retained] = ACTIONS(6724),
    [anon_sym___bridge_transfer] = ACTIONS(6724),
    [anon_sym___complex] = ACTIONS(6724),
    [anon_sym___const] = ACTIONS(6724),
    [anon_sym___imag] = ACTIONS(6724),
    [anon_sym___kindof] = ACTIONS(6724),
    [anon_sym___nonnull] = ACTIONS(6724),
    [anon_sym___nullable] = ACTIONS(6724),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6724),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6724),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6724),
    [anon_sym___real] = ACTIONS(6724),
    [anon_sym___strong] = ACTIONS(6724),
    [anon_sym___unsafe_unretained] = ACTIONS(6724),
    [anon_sym___unused] = ACTIONS(6724),
    [anon_sym___weak] = ACTIONS(6724),
    [anon_sym_alignas] = ACTIONS(6724),
    [anon_sym__Alignas] = ACTIONS(6724),
    [anon_sym_COLON] = ACTIONS(6724),
    [anon_sym_asm] = ACTIONS(6724),
    [anon_sym___asm__] = ACTIONS(6724),
    [anon_sym___asm] = ACTIONS(6722),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6724),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6724),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6724),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6724),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6724),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6724),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6724),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6724),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6724),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6724),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6722),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6724),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6724),
    [anon_sym_API_AVAILABLE] = ACTIONS(6724),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6724),
    [anon_sym_API_DEPRECATED] = ACTIONS(6724),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6724),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6724),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6724),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6724),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6724),
    [anon_sym___deprecated_msg] = ACTIONS(6724),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6724),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6724),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6724),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6724),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6724),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6724),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6724),
  },
  [STATE(3056)] = {
    [anon_sym_COMMA] = ACTIONS(6758),
    [anon_sym_RPAREN] = ACTIONS(6758),
    [anon_sym_LPAREN2] = ACTIONS(6758),
    [anon_sym_SEMI] = ACTIONS(6758),
    [anon_sym___extension__] = ACTIONS(6758),
    [anon_sym_extern] = ACTIONS(6758),
    [anon_sym___attribute__] = ACTIONS(6758),
    [anon_sym___attribute] = ACTIONS(6756),
    [anon_sym_noreturn] = ACTIONS(6758),
    [anon_sym_LBRACK] = ACTIONS(6758),
    [anon_sym___declspec] = ACTIONS(6758),
    [anon_sym_LBRACE] = ACTIONS(6758),
    [anon_sym_static] = ACTIONS(6758),
    [anon_sym_EQ] = ACTIONS(6758),
    [anon_sym_ATautoreleasepool] = ACTIONS(6758),
    [anon_sym_auto] = ACTIONS(6758),
    [anon_sym_register] = ACTIONS(6758),
    [anon_sym_inline] = ACTIONS(6758),
    [anon_sym___inline] = ACTIONS(6756),
    [anon_sym___inline__] = ACTIONS(6758),
    [anon_sym___forceinline] = ACTIONS(6758),
    [anon_sym_thread_local] = ACTIONS(6758),
    [anon_sym___thread] = ACTIONS(6758),
    [anon_sym_CG_EXTERN] = ACTIONS(6758),
    [anon_sym_CG_INLINE] = ACTIONS(6758),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6758),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6758),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6758),
    [anon_sym_IBOutlet] = ACTIONS(6758),
    [anon_sym_IBInspectable] = ACTIONS(6758),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6758),
    [anon_sym_NS_INLINE] = ACTIONS(6758),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6758),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6758),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6758),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6758),
    [anon_sym_const] = ACTIONS(6756),
    [anon_sym_constexpr] = ACTIONS(6758),
    [anon_sym_volatile] = ACTIONS(6758),
    [anon_sym_restrict] = ACTIONS(6758),
    [anon_sym___restrict__] = ACTIONS(6758),
    [anon_sym__Atomic] = ACTIONS(6758),
    [anon_sym__Noreturn] = ACTIONS(6758),
    [anon_sym__Nonnull] = ACTIONS(6758),
    [anon_sym_nullable] = ACTIONS(6758),
    [anon_sym__Complex] = ACTIONS(6758),
    [anon_sym__Nullable] = ACTIONS(6756),
    [anon_sym__Nullable_result] = ACTIONS(6758),
    [anon_sym__Null_unspecified] = ACTIONS(6758),
    [anon_sym___autoreleasing] = ACTIONS(6758),
    [anon_sym___block] = ACTIONS(6758),
    [anon_sym___bridge] = ACTIONS(6756),
    [anon_sym___bridge_retained] = ACTIONS(6758),
    [anon_sym___bridge_transfer] = ACTIONS(6758),
    [anon_sym___complex] = ACTIONS(6758),
    [anon_sym___const] = ACTIONS(6758),
    [anon_sym___imag] = ACTIONS(6758),
    [anon_sym___kindof] = ACTIONS(6758),
    [anon_sym___nonnull] = ACTIONS(6758),
    [anon_sym___nullable] = ACTIONS(6758),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6758),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6758),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6758),
    [anon_sym___real] = ACTIONS(6758),
    [anon_sym___strong] = ACTIONS(6758),
    [anon_sym___unsafe_unretained] = ACTIONS(6758),
    [anon_sym___unused] = ACTIONS(6758),
    [anon_sym___weak] = ACTIONS(6758),
    [anon_sym_alignas] = ACTIONS(6758),
    [anon_sym__Alignas] = ACTIONS(6758),
    [anon_sym_COLON] = ACTIONS(6758),
    [anon_sym_in] = ACTIONS(6756),
    [anon_sym_asm] = ACTIONS(6758),
    [anon_sym___asm__] = ACTIONS(6758),
    [anon_sym___asm] = ACTIONS(6756),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6758),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6758),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6758),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6758),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6758),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6758),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6758),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6758),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6758),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6758),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6756),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6758),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6758),
    [anon_sym_API_AVAILABLE] = ACTIONS(6758),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6758),
    [anon_sym_API_DEPRECATED] = ACTIONS(6758),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6758),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6758),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6758),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6758),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6758),
    [anon_sym___deprecated_msg] = ACTIONS(6758),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6758),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6758),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6758),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6758),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6758),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6758),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6758),
  },
  [STATE(3057)] = {
    [sym__declaration_modifiers] = STATE(5143),
    [sym_attribute_specifier] = STATE(5143),
    [sym_attribute_declaration] = STATE(5143),
    [sym_ms_declspec_modifier] = STATE(5143),
    [sym_storage_class_specifier] = STATE(5143),
    [sym_type_qualifier] = STATE(5143),
    [sym_alignas_qualifier] = STATE(5135),
    [sym_availability_attribute_specifier] = STATE(5143),
    [sym_identifier] = ACTIONS(7613),
    [anon_sym___extension__] = ACTIONS(7465),
    [anon_sym_extern] = ACTIONS(7467),
    [anon_sym___attribute__] = ACTIONS(7469),
    [anon_sym___attribute] = ACTIONS(7469),
    [anon_sym_noreturn] = ACTIONS(7465),
    [anon_sym_LBRACK] = ACTIONS(7471),
    [anon_sym___declspec] = ACTIONS(7473),
    [anon_sym_static] = ACTIONS(7467),
    [anon_sym_auto] = ACTIONS(7467),
    [anon_sym_register] = ACTIONS(7467),
    [anon_sym_inline] = ACTIONS(7467),
    [anon_sym___inline] = ACTIONS(7467),
    [anon_sym___inline__] = ACTIONS(7467),
    [anon_sym___forceinline] = ACTIONS(7467),
    [anon_sym_thread_local] = ACTIONS(7467),
    [anon_sym___thread] = ACTIONS(7467),
    [anon_sym_CG_EXTERN] = ACTIONS(7467),
    [anon_sym_CG_INLINE] = ACTIONS(7467),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7467),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7467),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7467),
    [anon_sym_IBOutlet] = ACTIONS(7467),
    [anon_sym_IBInspectable] = ACTIONS(7467),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7467),
    [anon_sym_NS_INLINE] = ACTIONS(7467),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7467),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7467),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7467),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7467),
    [anon_sym_const] = ACTIONS(7615),
    [anon_sym_constexpr] = ACTIONS(7465),
    [anon_sym_volatile] = ACTIONS(7465),
    [anon_sym_restrict] = ACTIONS(7465),
    [anon_sym___restrict__] = ACTIONS(7465),
    [anon_sym__Atomic] = ACTIONS(7465),
    [anon_sym__Noreturn] = ACTIONS(7465),
    [anon_sym__Nonnull] = ACTIONS(7465),
    [anon_sym_nullable] = ACTIONS(7465),
    [anon_sym__Complex] = ACTIONS(7465),
    [anon_sym__Nullable] = ACTIONS(7465),
    [anon_sym__Nullable_result] = ACTIONS(7465),
    [anon_sym__Null_unspecified] = ACTIONS(7465),
    [anon_sym___autoreleasing] = ACTIONS(7465),
    [anon_sym___block] = ACTIONS(7465),
    [anon_sym___bridge] = ACTIONS(7465),
    [anon_sym___bridge_retained] = ACTIONS(7465),
    [anon_sym___bridge_transfer] = ACTIONS(7465),
    [anon_sym___complex] = ACTIONS(7465),
    [anon_sym___const] = ACTIONS(7465),
    [anon_sym___imag] = ACTIONS(7465),
    [anon_sym___kindof] = ACTIONS(7465),
    [anon_sym___nonnull] = ACTIONS(7465),
    [anon_sym___nullable] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7465),
    [anon_sym___real] = ACTIONS(7465),
    [anon_sym___strong] = ACTIONS(7465),
    [anon_sym___unsafe_unretained] = ACTIONS(7465),
    [anon_sym___unused] = ACTIONS(7465),
    [anon_sym___weak] = ACTIONS(7465),
    [anon_sym_alignas] = ACTIONS(7477),
    [anon_sym__Alignas] = ACTIONS(7477),
    [anon_sym_struct] = ACTIONS(7617),
    [anon_sym_in] = ACTIONS(7617),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7481),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7481),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7481),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7481),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7481),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7481),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7481),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7481),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7481),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7483),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7483),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7483),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_API_AVAILABLE] = ACTIONS(7483),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7483),
    [anon_sym_API_DEPRECATED] = ACTIONS(7483),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7483),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7483),
    [anon_sym___deprecated_msg] = ACTIONS(7483),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7483),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7483),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7483),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7483),
    [anon_sym_id] = ACTIONS(7617),
  },
  [STATE(3058)] = {
    [sym_attribute_specifier] = STATE(2940),
    [sym_enumerator_list] = STATE(3159),
    [sym_identifier] = ACTIONS(7619),
    [anon_sym_COMMA] = ACTIONS(7621),
    [anon_sym_LPAREN2] = ACTIONS(7621),
    [anon_sym_STAR] = ACTIONS(7621),
    [anon_sym_CARET] = ACTIONS(7621),
    [anon_sym_SEMI] = ACTIONS(7621),
    [anon_sym___extension__] = ACTIONS(7619),
    [anon_sym_extern] = ACTIONS(7619),
    [anon_sym___attribute__] = ACTIONS(7623),
    [anon_sym___attribute] = ACTIONS(7623),
    [anon_sym_noreturn] = ACTIONS(7619),
    [anon_sym_LBRACK] = ACTIONS(7621),
    [anon_sym___declspec] = ACTIONS(7619),
    [anon_sym___based] = ACTIONS(7619),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7619),
    [anon_sym_auto] = ACTIONS(7619),
    [anon_sym_register] = ACTIONS(7619),
    [anon_sym_inline] = ACTIONS(7619),
    [anon_sym___inline] = ACTIONS(7619),
    [anon_sym___inline__] = ACTIONS(7619),
    [anon_sym___forceinline] = ACTIONS(7619),
    [anon_sym_thread_local] = ACTIONS(7619),
    [anon_sym___thread] = ACTIONS(7619),
    [anon_sym_CG_EXTERN] = ACTIONS(7619),
    [anon_sym_CG_INLINE] = ACTIONS(7619),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7619),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7619),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7619),
    [anon_sym_IBOutlet] = ACTIONS(7619),
    [anon_sym_IBInspectable] = ACTIONS(7619),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7619),
    [anon_sym_NS_INLINE] = ACTIONS(7619),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7619),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7619),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7619),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7619),
    [anon_sym_const] = ACTIONS(7619),
    [anon_sym_constexpr] = ACTIONS(7619),
    [anon_sym_volatile] = ACTIONS(7619),
    [anon_sym_restrict] = ACTIONS(7619),
    [anon_sym___restrict__] = ACTIONS(7619),
    [anon_sym__Atomic] = ACTIONS(7619),
    [anon_sym__Noreturn] = ACTIONS(7619),
    [anon_sym__Nonnull] = ACTIONS(7619),
    [anon_sym_nullable] = ACTIONS(7619),
    [anon_sym__Complex] = ACTIONS(7619),
    [anon_sym__Nullable] = ACTIONS(7619),
    [anon_sym__Nullable_result] = ACTIONS(7619),
    [anon_sym__Null_unspecified] = ACTIONS(7619),
    [anon_sym___autoreleasing] = ACTIONS(7619),
    [anon_sym___block] = ACTIONS(7619),
    [anon_sym___bridge] = ACTIONS(7619),
    [anon_sym___bridge_retained] = ACTIONS(7619),
    [anon_sym___bridge_transfer] = ACTIONS(7619),
    [anon_sym___complex] = ACTIONS(7619),
    [anon_sym___const] = ACTIONS(7619),
    [anon_sym___imag] = ACTIONS(7619),
    [anon_sym___kindof] = ACTIONS(7619),
    [anon_sym___nonnull] = ACTIONS(7619),
    [anon_sym___nullable] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7619),
    [anon_sym___real] = ACTIONS(7619),
    [anon_sym___strong] = ACTIONS(7619),
    [anon_sym___unsafe_unretained] = ACTIONS(7619),
    [anon_sym___unused] = ACTIONS(7619),
    [anon_sym___weak] = ACTIONS(7619),
    [anon_sym_alignas] = ACTIONS(7619),
    [anon_sym__Alignas] = ACTIONS(7619),
    [anon_sym_enum] = ACTIONS(7619),
    [anon_sym_COLON] = ACTIONS(7626),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7619),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7619),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7619),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7619),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7619),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7619),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7619),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7619),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7619),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7619),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7619),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7619),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7619),
    [anon_sym_API_AVAILABLE] = ACTIONS(7619),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7619),
    [anon_sym_API_DEPRECATED] = ACTIONS(7619),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7619),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7619),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7619),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7619),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7619),
    [anon_sym___deprecated_msg] = ACTIONS(7619),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7619),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7619),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7619),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7619),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7619),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7619),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7619),
  },
  [STATE(3059)] = {
    [anon_sym_COMMA] = ACTIONS(6790),
    [anon_sym_RPAREN] = ACTIONS(6790),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_SEMI] = ACTIONS(6790),
    [anon_sym___extension__] = ACTIONS(6790),
    [anon_sym_extern] = ACTIONS(6790),
    [anon_sym___attribute__] = ACTIONS(6790),
    [anon_sym___attribute] = ACTIONS(6788),
    [anon_sym_noreturn] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym___declspec] = ACTIONS(6790),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_static] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_ATautoreleasepool] = ACTIONS(6790),
    [anon_sym_auto] = ACTIONS(6790),
    [anon_sym_register] = ACTIONS(6790),
    [anon_sym_inline] = ACTIONS(6790),
    [anon_sym___inline] = ACTIONS(6788),
    [anon_sym___inline__] = ACTIONS(6790),
    [anon_sym___forceinline] = ACTIONS(6790),
    [anon_sym_thread_local] = ACTIONS(6790),
    [anon_sym___thread] = ACTIONS(6790),
    [anon_sym_CG_EXTERN] = ACTIONS(6790),
    [anon_sym_CG_INLINE] = ACTIONS(6790),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6790),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6790),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6790),
    [anon_sym_IBOutlet] = ACTIONS(6790),
    [anon_sym_IBInspectable] = ACTIONS(6790),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6790),
    [anon_sym_NS_INLINE] = ACTIONS(6790),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6790),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6790),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6790),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6790),
    [anon_sym_const] = ACTIONS(6788),
    [anon_sym_constexpr] = ACTIONS(6790),
    [anon_sym_volatile] = ACTIONS(6790),
    [anon_sym_restrict] = ACTIONS(6790),
    [anon_sym___restrict__] = ACTIONS(6790),
    [anon_sym__Atomic] = ACTIONS(6790),
    [anon_sym__Noreturn] = ACTIONS(6790),
    [anon_sym__Nonnull] = ACTIONS(6790),
    [anon_sym_nullable] = ACTIONS(6790),
    [anon_sym__Complex] = ACTIONS(6790),
    [anon_sym__Nullable] = ACTIONS(6788),
    [anon_sym__Nullable_result] = ACTIONS(6790),
    [anon_sym__Null_unspecified] = ACTIONS(6790),
    [anon_sym___autoreleasing] = ACTIONS(6790),
    [anon_sym___block] = ACTIONS(6790),
    [anon_sym___bridge] = ACTIONS(6788),
    [anon_sym___bridge_retained] = ACTIONS(6790),
    [anon_sym___bridge_transfer] = ACTIONS(6790),
    [anon_sym___complex] = ACTIONS(6790),
    [anon_sym___const] = ACTIONS(6790),
    [anon_sym___imag] = ACTIONS(6790),
    [anon_sym___kindof] = ACTIONS(6790),
    [anon_sym___nonnull] = ACTIONS(6790),
    [anon_sym___nullable] = ACTIONS(6790),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6790),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6790),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6790),
    [anon_sym___real] = ACTIONS(6790),
    [anon_sym___strong] = ACTIONS(6790),
    [anon_sym___unsafe_unretained] = ACTIONS(6790),
    [anon_sym___unused] = ACTIONS(6790),
    [anon_sym___weak] = ACTIONS(6790),
    [anon_sym_alignas] = ACTIONS(6790),
    [anon_sym__Alignas] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_in] = ACTIONS(6788),
    [anon_sym_asm] = ACTIONS(6790),
    [anon_sym___asm__] = ACTIONS(6790),
    [anon_sym___asm] = ACTIONS(6788),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6790),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6790),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6790),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6790),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6790),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6790),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6790),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6790),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6790),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6790),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6788),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6790),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6790),
    [anon_sym_API_AVAILABLE] = ACTIONS(6790),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6790),
    [anon_sym_API_DEPRECATED] = ACTIONS(6790),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6790),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6790),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6790),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6790),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6790),
    [anon_sym___deprecated_msg] = ACTIONS(6790),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6790),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6790),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6790),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6790),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6790),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6790),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6790),
  },
  [STATE(3060)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_LPAREN2] = ACTIONS(7629),
    [anon_sym_STAR] = ACTIONS(6134),
    [anon_sym_CARET] = ACTIONS(6134),
    [anon_sym_LT] = ACTIONS(7045),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(6134),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(3061)] = {
    [sym_attribute_specifier] = STATE(3004),
    [sym_enumerator_list] = STATE(3198),
    [sym_identifier] = ACTIONS(7632),
    [anon_sym_COMMA] = ACTIONS(7634),
    [anon_sym_LPAREN2] = ACTIONS(7634),
    [anon_sym_STAR] = ACTIONS(7634),
    [anon_sym_CARET] = ACTIONS(7634),
    [anon_sym_SEMI] = ACTIONS(7634),
    [anon_sym___extension__] = ACTIONS(7632),
    [anon_sym_extern] = ACTIONS(7632),
    [anon_sym___attribute__] = ACTIONS(7636),
    [anon_sym___attribute] = ACTIONS(7636),
    [anon_sym_noreturn] = ACTIONS(7632),
    [anon_sym_LBRACK] = ACTIONS(7634),
    [anon_sym___declspec] = ACTIONS(7632),
    [anon_sym___based] = ACTIONS(7632),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7632),
    [anon_sym_auto] = ACTIONS(7632),
    [anon_sym_register] = ACTIONS(7632),
    [anon_sym_inline] = ACTIONS(7632),
    [anon_sym___inline] = ACTIONS(7632),
    [anon_sym___inline__] = ACTIONS(7632),
    [anon_sym___forceinline] = ACTIONS(7632),
    [anon_sym_thread_local] = ACTIONS(7632),
    [anon_sym___thread] = ACTIONS(7632),
    [anon_sym_CG_EXTERN] = ACTIONS(7632),
    [anon_sym_CG_INLINE] = ACTIONS(7632),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7632),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7632),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7632),
    [anon_sym_IBOutlet] = ACTIONS(7632),
    [anon_sym_IBInspectable] = ACTIONS(7632),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7632),
    [anon_sym_NS_INLINE] = ACTIONS(7632),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7632),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7632),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7632),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7632),
    [anon_sym_const] = ACTIONS(7632),
    [anon_sym_constexpr] = ACTIONS(7632),
    [anon_sym_volatile] = ACTIONS(7632),
    [anon_sym_restrict] = ACTIONS(7632),
    [anon_sym___restrict__] = ACTIONS(7632),
    [anon_sym__Atomic] = ACTIONS(7632),
    [anon_sym__Noreturn] = ACTIONS(7632),
    [anon_sym__Nonnull] = ACTIONS(7632),
    [anon_sym_nullable] = ACTIONS(7632),
    [anon_sym__Complex] = ACTIONS(7632),
    [anon_sym__Nullable] = ACTIONS(7632),
    [anon_sym__Nullable_result] = ACTIONS(7632),
    [anon_sym__Null_unspecified] = ACTIONS(7632),
    [anon_sym___autoreleasing] = ACTIONS(7632),
    [anon_sym___block] = ACTIONS(7632),
    [anon_sym___bridge] = ACTIONS(7632),
    [anon_sym___bridge_retained] = ACTIONS(7632),
    [anon_sym___bridge_transfer] = ACTIONS(7632),
    [anon_sym___complex] = ACTIONS(7632),
    [anon_sym___const] = ACTIONS(7632),
    [anon_sym___imag] = ACTIONS(7632),
    [anon_sym___kindof] = ACTIONS(7632),
    [anon_sym___nonnull] = ACTIONS(7632),
    [anon_sym___nullable] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7632),
    [anon_sym___real] = ACTIONS(7632),
    [anon_sym___strong] = ACTIONS(7632),
    [anon_sym___unsafe_unretained] = ACTIONS(7632),
    [anon_sym___unused] = ACTIONS(7632),
    [anon_sym___weak] = ACTIONS(7632),
    [anon_sym_alignas] = ACTIONS(7632),
    [anon_sym__Alignas] = ACTIONS(7632),
    [anon_sym_enum] = ACTIONS(7632),
    [anon_sym_COLON] = ACTIONS(7639),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7632),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7632),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7632),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7632),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7632),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7632),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7632),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7632),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7632),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7632),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7632),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7632),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7632),
    [anon_sym_API_AVAILABLE] = ACTIONS(7632),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7632),
    [anon_sym_API_DEPRECATED] = ACTIONS(7632),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7632),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7632),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7632),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7632),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7632),
    [anon_sym___deprecated_msg] = ACTIONS(7632),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7632),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7632),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7632),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7632),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7632),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7632),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7632),
  },
  [STATE(3062)] = {
    [sym_attribute_specifier] = STATE(2940),
    [sym_enumerator_list] = STATE(3159),
    [sym_identifier] = ACTIONS(7619),
    [anon_sym_COMMA] = ACTIONS(7621),
    [anon_sym_RPAREN] = ACTIONS(7621),
    [anon_sym_LPAREN2] = ACTIONS(7621),
    [anon_sym_STAR] = ACTIONS(7621),
    [anon_sym_CARET] = ACTIONS(7621),
    [anon_sym_SEMI] = ACTIONS(7621),
    [anon_sym___extension__] = ACTIONS(7619),
    [anon_sym_extern] = ACTIONS(7619),
    [anon_sym___attribute__] = ACTIONS(7623),
    [anon_sym___attribute] = ACTIONS(7623),
    [anon_sym_noreturn] = ACTIONS(7619),
    [anon_sym_LBRACK] = ACTIONS(7621),
    [anon_sym___declspec] = ACTIONS(7619),
    [anon_sym___based] = ACTIONS(7619),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7619),
    [anon_sym_auto] = ACTIONS(7619),
    [anon_sym_register] = ACTIONS(7619),
    [anon_sym_inline] = ACTIONS(7619),
    [anon_sym___inline] = ACTIONS(7619),
    [anon_sym___inline__] = ACTIONS(7619),
    [anon_sym___forceinline] = ACTIONS(7619),
    [anon_sym_thread_local] = ACTIONS(7619),
    [anon_sym___thread] = ACTIONS(7619),
    [anon_sym_CG_EXTERN] = ACTIONS(7619),
    [anon_sym_CG_INLINE] = ACTIONS(7619),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7619),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7619),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7619),
    [anon_sym_IBOutlet] = ACTIONS(7619),
    [anon_sym_IBInspectable] = ACTIONS(7619),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7619),
    [anon_sym_NS_INLINE] = ACTIONS(7619),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7619),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7619),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7619),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7619),
    [anon_sym_const] = ACTIONS(7619),
    [anon_sym_constexpr] = ACTIONS(7619),
    [anon_sym_volatile] = ACTIONS(7619),
    [anon_sym_restrict] = ACTIONS(7619),
    [anon_sym___restrict__] = ACTIONS(7619),
    [anon_sym__Atomic] = ACTIONS(7619),
    [anon_sym__Noreturn] = ACTIONS(7619),
    [anon_sym__Nonnull] = ACTIONS(7619),
    [anon_sym_nullable] = ACTIONS(7619),
    [anon_sym__Complex] = ACTIONS(7619),
    [anon_sym__Nullable] = ACTIONS(7619),
    [anon_sym__Nullable_result] = ACTIONS(7619),
    [anon_sym__Null_unspecified] = ACTIONS(7619),
    [anon_sym___autoreleasing] = ACTIONS(7619),
    [anon_sym___block] = ACTIONS(7619),
    [anon_sym___bridge] = ACTIONS(7619),
    [anon_sym___bridge_retained] = ACTIONS(7619),
    [anon_sym___bridge_transfer] = ACTIONS(7619),
    [anon_sym___complex] = ACTIONS(7619),
    [anon_sym___const] = ACTIONS(7619),
    [anon_sym___imag] = ACTIONS(7619),
    [anon_sym___kindof] = ACTIONS(7619),
    [anon_sym___nonnull] = ACTIONS(7619),
    [anon_sym___nullable] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7619),
    [anon_sym___real] = ACTIONS(7619),
    [anon_sym___strong] = ACTIONS(7619),
    [anon_sym___unsafe_unretained] = ACTIONS(7619),
    [anon_sym___unused] = ACTIONS(7619),
    [anon_sym___weak] = ACTIONS(7619),
    [anon_sym_alignas] = ACTIONS(7619),
    [anon_sym__Alignas] = ACTIONS(7619),
    [anon_sym_COLON] = ACTIONS(7642),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7619),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7619),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7619),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7619),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7619),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7619),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7619),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7619),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7619),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7619),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7619),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7619),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7619),
    [anon_sym_API_AVAILABLE] = ACTIONS(7619),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7619),
    [anon_sym_API_DEPRECATED] = ACTIONS(7619),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7619),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7619),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7619),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7619),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7619),
    [anon_sym___deprecated_msg] = ACTIONS(7619),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7619),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7619),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7619),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7619),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7619),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7619),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7619),
  },
  [STATE(3063)] = {
    [sym_attribute_specifier] = STATE(2983),
    [sym_enumerator_list] = STATE(3147),
    [sym_identifier] = ACTIONS(7581),
    [anon_sym_COMMA] = ACTIONS(7583),
    [anon_sym_LPAREN2] = ACTIONS(7583),
    [anon_sym_STAR] = ACTIONS(7583),
    [anon_sym_CARET] = ACTIONS(7583),
    [anon_sym_SEMI] = ACTIONS(7583),
    [anon_sym___extension__] = ACTIONS(7581),
    [anon_sym_extern] = ACTIONS(7581),
    [anon_sym___attribute__] = ACTIONS(7585),
    [anon_sym___attribute] = ACTIONS(7585),
    [anon_sym_noreturn] = ACTIONS(7581),
    [anon_sym_LBRACK] = ACTIONS(7583),
    [anon_sym___declspec] = ACTIONS(7581),
    [anon_sym___based] = ACTIONS(7581),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7581),
    [anon_sym_auto] = ACTIONS(7581),
    [anon_sym_register] = ACTIONS(7581),
    [anon_sym_inline] = ACTIONS(7581),
    [anon_sym___inline] = ACTIONS(7581),
    [anon_sym___inline__] = ACTIONS(7581),
    [anon_sym___forceinline] = ACTIONS(7581),
    [anon_sym_thread_local] = ACTIONS(7581),
    [anon_sym___thread] = ACTIONS(7581),
    [anon_sym_CG_EXTERN] = ACTIONS(7581),
    [anon_sym_CG_INLINE] = ACTIONS(7581),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7581),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7581),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7581),
    [anon_sym_IBOutlet] = ACTIONS(7581),
    [anon_sym_IBInspectable] = ACTIONS(7581),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7581),
    [anon_sym_NS_INLINE] = ACTIONS(7581),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7581),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7581),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7581),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7581),
    [anon_sym_const] = ACTIONS(7581),
    [anon_sym_constexpr] = ACTIONS(7581),
    [anon_sym_volatile] = ACTIONS(7581),
    [anon_sym_restrict] = ACTIONS(7581),
    [anon_sym___restrict__] = ACTIONS(7581),
    [anon_sym__Atomic] = ACTIONS(7581),
    [anon_sym__Noreturn] = ACTIONS(7581),
    [anon_sym__Nonnull] = ACTIONS(7581),
    [anon_sym_nullable] = ACTIONS(7581),
    [anon_sym__Complex] = ACTIONS(7581),
    [anon_sym__Nullable] = ACTIONS(7581),
    [anon_sym__Nullable_result] = ACTIONS(7581),
    [anon_sym__Null_unspecified] = ACTIONS(7581),
    [anon_sym___autoreleasing] = ACTIONS(7581),
    [anon_sym___block] = ACTIONS(7581),
    [anon_sym___bridge] = ACTIONS(7581),
    [anon_sym___bridge_retained] = ACTIONS(7581),
    [anon_sym___bridge_transfer] = ACTIONS(7581),
    [anon_sym___complex] = ACTIONS(7581),
    [anon_sym___const] = ACTIONS(7581),
    [anon_sym___imag] = ACTIONS(7581),
    [anon_sym___kindof] = ACTIONS(7581),
    [anon_sym___nonnull] = ACTIONS(7581),
    [anon_sym___nullable] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7581),
    [anon_sym___real] = ACTIONS(7581),
    [anon_sym___strong] = ACTIONS(7581),
    [anon_sym___unsafe_unretained] = ACTIONS(7581),
    [anon_sym___unused] = ACTIONS(7581),
    [anon_sym___weak] = ACTIONS(7581),
    [anon_sym_alignas] = ACTIONS(7581),
    [anon_sym__Alignas] = ACTIONS(7581),
    [anon_sym_enum] = ACTIONS(7581),
    [anon_sym_COLON] = ACTIONS(7644),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7581),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7581),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7581),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7581),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7581),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7581),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7581),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7581),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7581),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7581),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7581),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7581),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7581),
    [anon_sym_API_AVAILABLE] = ACTIONS(7581),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7581),
    [anon_sym_API_DEPRECATED] = ACTIONS(7581),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7581),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7581),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7581),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7581),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7581),
    [anon_sym___deprecated_msg] = ACTIONS(7581),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7581),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7581),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7581),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7581),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7581),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7581),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7581),
  },
  [STATE(3064)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_LPAREN2] = ACTIONS(7647),
    [anon_sym_STAR] = ACTIONS(6134),
    [anon_sym_CARET] = ACTIONS(6134),
    [anon_sym_LT] = ACTIONS(7045),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_extern] = ACTIONS(4494),
    [anon_sym___attribute__] = ACTIONS(4494),
    [anon_sym___attribute] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(6134),
    [anon_sym___declspec] = ACTIONS(4494),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(4509),
    [anon_sym_unsigned] = ACTIONS(4509),
    [anon_sym_long] = ACTIONS(4509),
    [anon_sym_short] = ACTIONS(4509),
    [anon_sym_static] = ACTIONS(4494),
    [anon_sym_auto] = ACTIONS(4494),
    [anon_sym_register] = ACTIONS(4494),
    [anon_sym_inline] = ACTIONS(4494),
    [anon_sym___inline] = ACTIONS(4494),
    [anon_sym___inline__] = ACTIONS(4494),
    [anon_sym___forceinline] = ACTIONS(4494),
    [anon_sym_thread_local] = ACTIONS(4494),
    [anon_sym___thread] = ACTIONS(4494),
    [anon_sym_CG_EXTERN] = ACTIONS(4494),
    [anon_sym_CG_INLINE] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(4494),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(4494),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(4494),
    [anon_sym_IBOutlet] = ACTIONS(4494),
    [anon_sym_IBInspectable] = ACTIONS(4494),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(4494),
    [anon_sym_NS_INLINE] = ACTIONS(4494),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(4494),
    [anon_sym_OBJC_EXPORT] = ACTIONS(4494),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(4494),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(4494),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(4494),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(4494),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(4494),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE] = ACTIONS(4494),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_API_AVAILABLE] = ACTIONS(4494),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_API_DEPRECATED] = ACTIONS(4494),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(4494),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(4494),
    [anon_sym___deprecated_msg] = ACTIONS(4494),
    [anon_sym___deprecated_enum_msg] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(4494),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(4494),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(4494),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(4494),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(4494),
  },
  [STATE(3065)] = {
    [anon_sym_COMMA] = ACTIONS(6128),
    [anon_sym_RPAREN] = ACTIONS(6128),
    [anon_sym_LPAREN2] = ACTIONS(6128),
    [anon_sym_SEMI] = ACTIONS(6128),
    [anon_sym___extension__] = ACTIONS(6128),
    [anon_sym_extern] = ACTIONS(6128),
    [anon_sym___attribute__] = ACTIONS(6128),
    [anon_sym___attribute] = ACTIONS(6141),
    [anon_sym_noreturn] = ACTIONS(6128),
    [anon_sym_LBRACK] = ACTIONS(6128),
    [anon_sym___declspec] = ACTIONS(6128),
    [anon_sym_LBRACE] = ACTIONS(6128),
    [anon_sym_static] = ACTIONS(6128),
    [anon_sym_EQ] = ACTIONS(6128),
    [anon_sym_ATautoreleasepool] = ACTIONS(6128),
    [anon_sym_auto] = ACTIONS(6128),
    [anon_sym_register] = ACTIONS(6128),
    [anon_sym_inline] = ACTIONS(6128),
    [anon_sym___inline] = ACTIONS(6141),
    [anon_sym___inline__] = ACTIONS(6128),
    [anon_sym___forceinline] = ACTIONS(6128),
    [anon_sym_thread_local] = ACTIONS(6128),
    [anon_sym___thread] = ACTIONS(6128),
    [anon_sym_CG_EXTERN] = ACTIONS(6128),
    [anon_sym_CG_INLINE] = ACTIONS(6128),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6128),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6128),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6128),
    [anon_sym_IBOutlet] = ACTIONS(6128),
    [anon_sym_IBInspectable] = ACTIONS(6128),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6128),
    [anon_sym_NS_INLINE] = ACTIONS(6128),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6128),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6128),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6128),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6128),
    [anon_sym_const] = ACTIONS(6141),
    [anon_sym_constexpr] = ACTIONS(6128),
    [anon_sym_volatile] = ACTIONS(6128),
    [anon_sym_restrict] = ACTIONS(6128),
    [anon_sym___restrict__] = ACTIONS(6128),
    [anon_sym__Atomic] = ACTIONS(6128),
    [anon_sym__Noreturn] = ACTIONS(6128),
    [anon_sym__Nonnull] = ACTIONS(6128),
    [anon_sym_nullable] = ACTIONS(6128),
    [anon_sym__Complex] = ACTIONS(6128),
    [anon_sym__Nullable] = ACTIONS(6141),
    [anon_sym__Nullable_result] = ACTIONS(6128),
    [anon_sym__Null_unspecified] = ACTIONS(6128),
    [anon_sym___autoreleasing] = ACTIONS(6128),
    [anon_sym___block] = ACTIONS(6128),
    [anon_sym___bridge] = ACTIONS(6141),
    [anon_sym___bridge_retained] = ACTIONS(6128),
    [anon_sym___bridge_transfer] = ACTIONS(6128),
    [anon_sym___complex] = ACTIONS(6128),
    [anon_sym___const] = ACTIONS(6128),
    [anon_sym___imag] = ACTIONS(6128),
    [anon_sym___kindof] = ACTIONS(6128),
    [anon_sym___nonnull] = ACTIONS(6128),
    [anon_sym___nullable] = ACTIONS(6128),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6128),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6128),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6128),
    [anon_sym___real] = ACTIONS(6128),
    [anon_sym___strong] = ACTIONS(6128),
    [anon_sym___unsafe_unretained] = ACTIONS(6128),
    [anon_sym___unused] = ACTIONS(6128),
    [anon_sym___weak] = ACTIONS(6128),
    [anon_sym_alignas] = ACTIONS(6128),
    [anon_sym__Alignas] = ACTIONS(6128),
    [anon_sym_COLON] = ACTIONS(6128),
    [anon_sym_in] = ACTIONS(6141),
    [anon_sym_asm] = ACTIONS(6128),
    [anon_sym___asm__] = ACTIONS(6128),
    [anon_sym___asm] = ACTIONS(6141),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6128),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6128),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6128),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6128),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6128),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6128),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6128),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6128),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6128),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6128),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6141),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6128),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6128),
    [anon_sym_API_AVAILABLE] = ACTIONS(6128),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6128),
    [anon_sym_API_DEPRECATED] = ACTIONS(6128),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6128),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6128),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6128),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6128),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6128),
    [anon_sym___deprecated_msg] = ACTIONS(6128),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6128),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6128),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6128),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6128),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6128),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6128),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6128),
  },
  [STATE(3066)] = {
    [anon_sym_COMMA] = ACTIONS(6624),
    [anon_sym_RPAREN] = ACTIONS(6624),
    [anon_sym_LPAREN2] = ACTIONS(6624),
    [anon_sym_SEMI] = ACTIONS(6624),
    [anon_sym___extension__] = ACTIONS(6624),
    [anon_sym_extern] = ACTIONS(6624),
    [anon_sym___attribute__] = ACTIONS(6624),
    [anon_sym___attribute] = ACTIONS(6622),
    [anon_sym_noreturn] = ACTIONS(6624),
    [anon_sym_LBRACK] = ACTIONS(6624),
    [anon_sym___declspec] = ACTIONS(6624),
    [anon_sym_LBRACE] = ACTIONS(6624),
    [anon_sym_static] = ACTIONS(6624),
    [anon_sym_EQ] = ACTIONS(6624),
    [anon_sym_ATautoreleasepool] = ACTIONS(6624),
    [anon_sym_auto] = ACTIONS(6624),
    [anon_sym_register] = ACTIONS(6624),
    [anon_sym_inline] = ACTIONS(6624),
    [anon_sym___inline] = ACTIONS(6622),
    [anon_sym___inline__] = ACTIONS(6624),
    [anon_sym___forceinline] = ACTIONS(6624),
    [anon_sym_thread_local] = ACTIONS(6624),
    [anon_sym___thread] = ACTIONS(6624),
    [anon_sym_CG_EXTERN] = ACTIONS(6624),
    [anon_sym_CG_INLINE] = ACTIONS(6624),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6624),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6624),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6624),
    [anon_sym_IBOutlet] = ACTIONS(6624),
    [anon_sym_IBInspectable] = ACTIONS(6624),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6624),
    [anon_sym_NS_INLINE] = ACTIONS(6624),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6624),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6624),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6624),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6624),
    [anon_sym_const] = ACTIONS(6622),
    [anon_sym_constexpr] = ACTIONS(6624),
    [anon_sym_volatile] = ACTIONS(6624),
    [anon_sym_restrict] = ACTIONS(6624),
    [anon_sym___restrict__] = ACTIONS(6624),
    [anon_sym__Atomic] = ACTIONS(6624),
    [anon_sym__Noreturn] = ACTIONS(6624),
    [anon_sym__Nonnull] = ACTIONS(6624),
    [anon_sym_nullable] = ACTIONS(6624),
    [anon_sym__Complex] = ACTIONS(6624),
    [anon_sym__Nullable] = ACTIONS(6622),
    [anon_sym__Nullable_result] = ACTIONS(6624),
    [anon_sym__Null_unspecified] = ACTIONS(6624),
    [anon_sym___autoreleasing] = ACTIONS(6624),
    [anon_sym___block] = ACTIONS(6624),
    [anon_sym___bridge] = ACTIONS(6622),
    [anon_sym___bridge_retained] = ACTIONS(6624),
    [anon_sym___bridge_transfer] = ACTIONS(6624),
    [anon_sym___complex] = ACTIONS(6624),
    [anon_sym___const] = ACTIONS(6624),
    [anon_sym___imag] = ACTIONS(6624),
    [anon_sym___kindof] = ACTIONS(6624),
    [anon_sym___nonnull] = ACTIONS(6624),
    [anon_sym___nullable] = ACTIONS(6624),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6624),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6624),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6624),
    [anon_sym___real] = ACTIONS(6624),
    [anon_sym___strong] = ACTIONS(6624),
    [anon_sym___unsafe_unretained] = ACTIONS(6624),
    [anon_sym___unused] = ACTIONS(6624),
    [anon_sym___weak] = ACTIONS(6624),
    [anon_sym_alignas] = ACTIONS(6624),
    [anon_sym__Alignas] = ACTIONS(6624),
    [anon_sym_COLON] = ACTIONS(6624),
    [anon_sym_in] = ACTIONS(6622),
    [anon_sym_asm] = ACTIONS(6624),
    [anon_sym___asm__] = ACTIONS(6624),
    [anon_sym___asm] = ACTIONS(6622),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6624),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6624),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6624),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6624),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6624),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6624),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6624),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6624),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6624),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6624),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6622),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6624),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6624),
    [anon_sym_API_AVAILABLE] = ACTIONS(6624),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6624),
    [anon_sym_API_DEPRECATED] = ACTIONS(6624),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6624),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6624),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6624),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6624),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6624),
    [anon_sym___deprecated_msg] = ACTIONS(6624),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6624),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6624),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6624),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6624),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6624),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6624),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6624),
  },
  [STATE(3067)] = {
    [anon_sym_COMMA] = ACTIONS(6754),
    [anon_sym_RPAREN] = ACTIONS(6754),
    [anon_sym_LPAREN2] = ACTIONS(6754),
    [anon_sym_SEMI] = ACTIONS(6754),
    [anon_sym___extension__] = ACTIONS(6754),
    [anon_sym_extern] = ACTIONS(6754),
    [anon_sym___attribute__] = ACTIONS(6754),
    [anon_sym___attribute] = ACTIONS(6752),
    [anon_sym_noreturn] = ACTIONS(6754),
    [anon_sym_LBRACK] = ACTIONS(6754),
    [anon_sym___declspec] = ACTIONS(6754),
    [anon_sym_LBRACE] = ACTIONS(6754),
    [anon_sym_static] = ACTIONS(6754),
    [anon_sym_EQ] = ACTIONS(6754),
    [anon_sym_ATautoreleasepool] = ACTIONS(6754),
    [anon_sym_auto] = ACTIONS(6754),
    [anon_sym_register] = ACTIONS(6754),
    [anon_sym_inline] = ACTIONS(6754),
    [anon_sym___inline] = ACTIONS(6752),
    [anon_sym___inline__] = ACTIONS(6754),
    [anon_sym___forceinline] = ACTIONS(6754),
    [anon_sym_thread_local] = ACTIONS(6754),
    [anon_sym___thread] = ACTIONS(6754),
    [anon_sym_CG_EXTERN] = ACTIONS(6754),
    [anon_sym_CG_INLINE] = ACTIONS(6754),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6754),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6754),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6754),
    [anon_sym_IBOutlet] = ACTIONS(6754),
    [anon_sym_IBInspectable] = ACTIONS(6754),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6754),
    [anon_sym_NS_INLINE] = ACTIONS(6754),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6754),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6754),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6754),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6754),
    [anon_sym_const] = ACTIONS(6752),
    [anon_sym_constexpr] = ACTIONS(6754),
    [anon_sym_volatile] = ACTIONS(6754),
    [anon_sym_restrict] = ACTIONS(6754),
    [anon_sym___restrict__] = ACTIONS(6754),
    [anon_sym__Atomic] = ACTIONS(6754),
    [anon_sym__Noreturn] = ACTIONS(6754),
    [anon_sym__Nonnull] = ACTIONS(6754),
    [anon_sym_nullable] = ACTIONS(6754),
    [anon_sym__Complex] = ACTIONS(6754),
    [anon_sym__Nullable] = ACTIONS(6752),
    [anon_sym__Nullable_result] = ACTIONS(6754),
    [anon_sym__Null_unspecified] = ACTIONS(6754),
    [anon_sym___autoreleasing] = ACTIONS(6754),
    [anon_sym___block] = ACTIONS(6754),
    [anon_sym___bridge] = ACTIONS(6752),
    [anon_sym___bridge_retained] = ACTIONS(6754),
    [anon_sym___bridge_transfer] = ACTIONS(6754),
    [anon_sym___complex] = ACTIONS(6754),
    [anon_sym___const] = ACTIONS(6754),
    [anon_sym___imag] = ACTIONS(6754),
    [anon_sym___kindof] = ACTIONS(6754),
    [anon_sym___nonnull] = ACTIONS(6754),
    [anon_sym___nullable] = ACTIONS(6754),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6754),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6754),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6754),
    [anon_sym___real] = ACTIONS(6754),
    [anon_sym___strong] = ACTIONS(6754),
    [anon_sym___unsafe_unretained] = ACTIONS(6754),
    [anon_sym___unused] = ACTIONS(6754),
    [anon_sym___weak] = ACTIONS(6754),
    [anon_sym_alignas] = ACTIONS(6754),
    [anon_sym__Alignas] = ACTIONS(6754),
    [anon_sym_COLON] = ACTIONS(6754),
    [anon_sym_in] = ACTIONS(6752),
    [anon_sym_asm] = ACTIONS(6754),
    [anon_sym___asm__] = ACTIONS(6754),
    [anon_sym___asm] = ACTIONS(6752),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6754),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6754),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6754),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6754),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6754),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6754),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6754),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6754),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6754),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6754),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6752),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6754),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6754),
    [anon_sym_API_AVAILABLE] = ACTIONS(6754),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6754),
    [anon_sym_API_DEPRECATED] = ACTIONS(6754),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6754),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6754),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6754),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6754),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6754),
    [anon_sym___deprecated_msg] = ACTIONS(6754),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6754),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6754),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6754),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6754),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6754),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6754),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6754),
  },
  [STATE(3068)] = {
    [sym_parameter_list] = STATE(3007),
    [anon_sym_COMMA] = ACTIONS(6720),
    [anon_sym_RPAREN] = ACTIONS(6720),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6720),
    [anon_sym___extension__] = ACTIONS(6720),
    [anon_sym_extern] = ACTIONS(6720),
    [anon_sym___attribute__] = ACTIONS(6720),
    [anon_sym___attribute] = ACTIONS(6718),
    [anon_sym_noreturn] = ACTIONS(6720),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6720),
    [anon_sym_LBRACE] = ACTIONS(6720),
    [anon_sym_static] = ACTIONS(6720),
    [anon_sym_EQ] = ACTIONS(6720),
    [anon_sym_ATautoreleasepool] = ACTIONS(6720),
    [anon_sym_auto] = ACTIONS(6720),
    [anon_sym_register] = ACTIONS(6720),
    [anon_sym_inline] = ACTIONS(6720),
    [anon_sym___inline] = ACTIONS(6718),
    [anon_sym___inline__] = ACTIONS(6720),
    [anon_sym___forceinline] = ACTIONS(6720),
    [anon_sym_thread_local] = ACTIONS(6720),
    [anon_sym___thread] = ACTIONS(6720),
    [anon_sym_CG_EXTERN] = ACTIONS(6720),
    [anon_sym_CG_INLINE] = ACTIONS(6720),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6720),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6720),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6720),
    [anon_sym_IBOutlet] = ACTIONS(6720),
    [anon_sym_IBInspectable] = ACTIONS(6720),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6720),
    [anon_sym_NS_INLINE] = ACTIONS(6720),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6720),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6720),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6720),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6720),
    [anon_sym_const] = ACTIONS(6718),
    [anon_sym_constexpr] = ACTIONS(6720),
    [anon_sym_volatile] = ACTIONS(6720),
    [anon_sym_restrict] = ACTIONS(6720),
    [anon_sym___restrict__] = ACTIONS(6720),
    [anon_sym__Atomic] = ACTIONS(6720),
    [anon_sym__Noreturn] = ACTIONS(6720),
    [anon_sym__Nonnull] = ACTIONS(6720),
    [anon_sym_nullable] = ACTIONS(6720),
    [anon_sym__Complex] = ACTIONS(6720),
    [anon_sym__Nullable] = ACTIONS(6718),
    [anon_sym__Nullable_result] = ACTIONS(6720),
    [anon_sym__Null_unspecified] = ACTIONS(6720),
    [anon_sym___autoreleasing] = ACTIONS(6720),
    [anon_sym___block] = ACTIONS(6720),
    [anon_sym___bridge] = ACTIONS(6718),
    [anon_sym___bridge_retained] = ACTIONS(6720),
    [anon_sym___bridge_transfer] = ACTIONS(6720),
    [anon_sym___complex] = ACTIONS(6720),
    [anon_sym___const] = ACTIONS(6720),
    [anon_sym___imag] = ACTIONS(6720),
    [anon_sym___kindof] = ACTIONS(6720),
    [anon_sym___nonnull] = ACTIONS(6720),
    [anon_sym___nullable] = ACTIONS(6720),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6720),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6720),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6720),
    [anon_sym___real] = ACTIONS(6720),
    [anon_sym___strong] = ACTIONS(6720),
    [anon_sym___unsafe_unretained] = ACTIONS(6720),
    [anon_sym___unused] = ACTIONS(6720),
    [anon_sym___weak] = ACTIONS(6720),
    [anon_sym_alignas] = ACTIONS(6720),
    [anon_sym__Alignas] = ACTIONS(6720),
    [anon_sym_COLON] = ACTIONS(6720),
    [anon_sym_asm] = ACTIONS(6720),
    [anon_sym___asm__] = ACTIONS(6720),
    [anon_sym___asm] = ACTIONS(6718),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6720),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6720),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6720),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6720),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6720),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6720),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6720),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6720),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6720),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6720),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6718),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6720),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6720),
    [anon_sym_API_AVAILABLE] = ACTIONS(6720),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6720),
    [anon_sym_API_DEPRECATED] = ACTIONS(6720),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6720),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6720),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6720),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6720),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6720),
    [anon_sym___deprecated_msg] = ACTIONS(6720),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6720),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6720),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6720),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6720),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6720),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6720),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6720),
  },
  [STATE(3069)] = {
    [sym_attribute_specifier] = STATE(3004),
    [sym_enumerator_list] = STATE(3198),
    [sym_identifier] = ACTIONS(7632),
    [anon_sym_COMMA] = ACTIONS(7634),
    [anon_sym_RPAREN] = ACTIONS(7634),
    [anon_sym_LPAREN2] = ACTIONS(7634),
    [anon_sym_STAR] = ACTIONS(7634),
    [anon_sym_CARET] = ACTIONS(7634),
    [anon_sym_SEMI] = ACTIONS(7634),
    [anon_sym___extension__] = ACTIONS(7632),
    [anon_sym_extern] = ACTIONS(7632),
    [anon_sym___attribute__] = ACTIONS(7636),
    [anon_sym___attribute] = ACTIONS(7636),
    [anon_sym_noreturn] = ACTIONS(7632),
    [anon_sym_LBRACK] = ACTIONS(7634),
    [anon_sym___declspec] = ACTIONS(7632),
    [anon_sym___based] = ACTIONS(7632),
    [anon_sym_LBRACE] = ACTIONS(7519),
    [anon_sym_static] = ACTIONS(7632),
    [anon_sym_auto] = ACTIONS(7632),
    [anon_sym_register] = ACTIONS(7632),
    [anon_sym_inline] = ACTIONS(7632),
    [anon_sym___inline] = ACTIONS(7632),
    [anon_sym___inline__] = ACTIONS(7632),
    [anon_sym___forceinline] = ACTIONS(7632),
    [anon_sym_thread_local] = ACTIONS(7632),
    [anon_sym___thread] = ACTIONS(7632),
    [anon_sym_CG_EXTERN] = ACTIONS(7632),
    [anon_sym_CG_INLINE] = ACTIONS(7632),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7632),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7632),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7632),
    [anon_sym_IBOutlet] = ACTIONS(7632),
    [anon_sym_IBInspectable] = ACTIONS(7632),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7632),
    [anon_sym_NS_INLINE] = ACTIONS(7632),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7632),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7632),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7632),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7632),
    [anon_sym_const] = ACTIONS(7632),
    [anon_sym_constexpr] = ACTIONS(7632),
    [anon_sym_volatile] = ACTIONS(7632),
    [anon_sym_restrict] = ACTIONS(7632),
    [anon_sym___restrict__] = ACTIONS(7632),
    [anon_sym__Atomic] = ACTIONS(7632),
    [anon_sym__Noreturn] = ACTIONS(7632),
    [anon_sym__Nonnull] = ACTIONS(7632),
    [anon_sym_nullable] = ACTIONS(7632),
    [anon_sym__Complex] = ACTIONS(7632),
    [anon_sym__Nullable] = ACTIONS(7632),
    [anon_sym__Nullable_result] = ACTIONS(7632),
    [anon_sym__Null_unspecified] = ACTIONS(7632),
    [anon_sym___autoreleasing] = ACTIONS(7632),
    [anon_sym___block] = ACTIONS(7632),
    [anon_sym___bridge] = ACTIONS(7632),
    [anon_sym___bridge_retained] = ACTIONS(7632),
    [anon_sym___bridge_transfer] = ACTIONS(7632),
    [anon_sym___complex] = ACTIONS(7632),
    [anon_sym___const] = ACTIONS(7632),
    [anon_sym___imag] = ACTIONS(7632),
    [anon_sym___kindof] = ACTIONS(7632),
    [anon_sym___nonnull] = ACTIONS(7632),
    [anon_sym___nullable] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7632),
    [anon_sym___real] = ACTIONS(7632),
    [anon_sym___strong] = ACTIONS(7632),
    [anon_sym___unsafe_unretained] = ACTIONS(7632),
    [anon_sym___unused] = ACTIONS(7632),
    [anon_sym___weak] = ACTIONS(7632),
    [anon_sym_alignas] = ACTIONS(7632),
    [anon_sym__Alignas] = ACTIONS(7632),
    [anon_sym_COLON] = ACTIONS(7650),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7632),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7632),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7632),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7632),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7632),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7632),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7632),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7632),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7632),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7632),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7632),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7632),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7632),
    [anon_sym_API_AVAILABLE] = ACTIONS(7632),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7632),
    [anon_sym_API_DEPRECATED] = ACTIONS(7632),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7632),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7632),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7632),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7632),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7632),
    [anon_sym___deprecated_msg] = ACTIONS(7632),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7632),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7632),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7632),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7632),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7632),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7632),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7632),
  },
  [STATE(3070)] = {
    [anon_sym_COMMA] = ACTIONS(6782),
    [anon_sym_RPAREN] = ACTIONS(6782),
    [anon_sym_LPAREN2] = ACTIONS(6782),
    [anon_sym_SEMI] = ACTIONS(6782),
    [anon_sym___extension__] = ACTIONS(6782),
    [anon_sym_extern] = ACTIONS(6782),
    [anon_sym___attribute__] = ACTIONS(6782),
    [anon_sym___attribute] = ACTIONS(6780),
    [anon_sym_noreturn] = ACTIONS(6782),
    [anon_sym_LBRACK] = ACTIONS(6782),
    [anon_sym___declspec] = ACTIONS(6782),
    [anon_sym_LBRACE] = ACTIONS(6782),
    [anon_sym_static] = ACTIONS(6782),
    [anon_sym_EQ] = ACTIONS(6782),
    [anon_sym_ATautoreleasepool] = ACTIONS(6782),
    [anon_sym_auto] = ACTIONS(6782),
    [anon_sym_register] = ACTIONS(6782),
    [anon_sym_inline] = ACTIONS(6782),
    [anon_sym___inline] = ACTIONS(6780),
    [anon_sym___inline__] = ACTIONS(6782),
    [anon_sym___forceinline] = ACTIONS(6782),
    [anon_sym_thread_local] = ACTIONS(6782),
    [anon_sym___thread] = ACTIONS(6782),
    [anon_sym_CG_EXTERN] = ACTIONS(6782),
    [anon_sym_CG_INLINE] = ACTIONS(6782),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6782),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6782),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6782),
    [anon_sym_IBOutlet] = ACTIONS(6782),
    [anon_sym_IBInspectable] = ACTIONS(6782),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6782),
    [anon_sym_NS_INLINE] = ACTIONS(6782),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6782),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6782),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6782),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6782),
    [anon_sym_const] = ACTIONS(6780),
    [anon_sym_constexpr] = ACTIONS(6782),
    [anon_sym_volatile] = ACTIONS(6782),
    [anon_sym_restrict] = ACTIONS(6782),
    [anon_sym___restrict__] = ACTIONS(6782),
    [anon_sym__Atomic] = ACTIONS(6782),
    [anon_sym__Noreturn] = ACTIONS(6782),
    [anon_sym__Nonnull] = ACTIONS(6782),
    [anon_sym_nullable] = ACTIONS(6782),
    [anon_sym__Complex] = ACTIONS(6782),
    [anon_sym__Nullable] = ACTIONS(6780),
    [anon_sym__Nullable_result] = ACTIONS(6782),
    [anon_sym__Null_unspecified] = ACTIONS(6782),
    [anon_sym___autoreleasing] = ACTIONS(6782),
    [anon_sym___block] = ACTIONS(6782),
    [anon_sym___bridge] = ACTIONS(6780),
    [anon_sym___bridge_retained] = ACTIONS(6782),
    [anon_sym___bridge_transfer] = ACTIONS(6782),
    [anon_sym___complex] = ACTIONS(6782),
    [anon_sym___const] = ACTIONS(6782),
    [anon_sym___imag] = ACTIONS(6782),
    [anon_sym___kindof] = ACTIONS(6782),
    [anon_sym___nonnull] = ACTIONS(6782),
    [anon_sym___nullable] = ACTIONS(6782),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6782),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6782),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6782),
    [anon_sym___real] = ACTIONS(6782),
    [anon_sym___strong] = ACTIONS(6782),
    [anon_sym___unsafe_unretained] = ACTIONS(6782),
    [anon_sym___unused] = ACTIONS(6782),
    [anon_sym___weak] = ACTIONS(6782),
    [anon_sym_alignas] = ACTIONS(6782),
    [anon_sym__Alignas] = ACTIONS(6782),
    [anon_sym_COLON] = ACTIONS(6782),
    [anon_sym_in] = ACTIONS(6780),
    [anon_sym_asm] = ACTIONS(6782),
    [anon_sym___asm__] = ACTIONS(6782),
    [anon_sym___asm] = ACTIONS(6780),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6782),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6782),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6782),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6782),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6782),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6782),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6782),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6782),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6782),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6782),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6780),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6782),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6782),
    [anon_sym_API_AVAILABLE] = ACTIONS(6782),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6782),
    [anon_sym_API_DEPRECATED] = ACTIONS(6782),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6782),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6782),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6782),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6782),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6782),
    [anon_sym___deprecated_msg] = ACTIONS(6782),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6782),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6782),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6782),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6782),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6782),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6782),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6782),
  },
  [STATE(3071)] = {
    [anon_sym_COMMA] = ACTIONS(6762),
    [anon_sym_RPAREN] = ACTIONS(6762),
    [anon_sym_LPAREN2] = ACTIONS(6762),
    [anon_sym_SEMI] = ACTIONS(6762),
    [anon_sym___extension__] = ACTIONS(6762),
    [anon_sym_extern] = ACTIONS(6762),
    [anon_sym___attribute__] = ACTIONS(6762),
    [anon_sym___attribute] = ACTIONS(6760),
    [anon_sym_noreturn] = ACTIONS(6762),
    [anon_sym_LBRACK] = ACTIONS(6762),
    [anon_sym___declspec] = ACTIONS(6762),
    [anon_sym_LBRACE] = ACTIONS(6762),
    [anon_sym_static] = ACTIONS(6762),
    [anon_sym_EQ] = ACTIONS(6762),
    [anon_sym_ATautoreleasepool] = ACTIONS(6762),
    [anon_sym_auto] = ACTIONS(6762),
    [anon_sym_register] = ACTIONS(6762),
    [anon_sym_inline] = ACTIONS(6762),
    [anon_sym___inline] = ACTIONS(6760),
    [anon_sym___inline__] = ACTIONS(6762),
    [anon_sym___forceinline] = ACTIONS(6762),
    [anon_sym_thread_local] = ACTIONS(6762),
    [anon_sym___thread] = ACTIONS(6762),
    [anon_sym_CG_EXTERN] = ACTIONS(6762),
    [anon_sym_CG_INLINE] = ACTIONS(6762),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6762),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6762),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6762),
    [anon_sym_IBOutlet] = ACTIONS(6762),
    [anon_sym_IBInspectable] = ACTIONS(6762),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6762),
    [anon_sym_NS_INLINE] = ACTIONS(6762),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6762),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6762),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6762),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6762),
    [anon_sym_const] = ACTIONS(6760),
    [anon_sym_constexpr] = ACTIONS(6762),
    [anon_sym_volatile] = ACTIONS(6762),
    [anon_sym_restrict] = ACTIONS(6762),
    [anon_sym___restrict__] = ACTIONS(6762),
    [anon_sym__Atomic] = ACTIONS(6762),
    [anon_sym__Noreturn] = ACTIONS(6762),
    [anon_sym__Nonnull] = ACTIONS(6762),
    [anon_sym_nullable] = ACTIONS(6762),
    [anon_sym__Complex] = ACTIONS(6762),
    [anon_sym__Nullable] = ACTIONS(6760),
    [anon_sym__Nullable_result] = ACTIONS(6762),
    [anon_sym__Null_unspecified] = ACTIONS(6762),
    [anon_sym___autoreleasing] = ACTIONS(6762),
    [anon_sym___block] = ACTIONS(6762),
    [anon_sym___bridge] = ACTIONS(6760),
    [anon_sym___bridge_retained] = ACTIONS(6762),
    [anon_sym___bridge_transfer] = ACTIONS(6762),
    [anon_sym___complex] = ACTIONS(6762),
    [anon_sym___const] = ACTIONS(6762),
    [anon_sym___imag] = ACTIONS(6762),
    [anon_sym___kindof] = ACTIONS(6762),
    [anon_sym___nonnull] = ACTIONS(6762),
    [anon_sym___nullable] = ACTIONS(6762),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6762),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6762),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6762),
    [anon_sym___real] = ACTIONS(6762),
    [anon_sym___strong] = ACTIONS(6762),
    [anon_sym___unsafe_unretained] = ACTIONS(6762),
    [anon_sym___unused] = ACTIONS(6762),
    [anon_sym___weak] = ACTIONS(6762),
    [anon_sym_alignas] = ACTIONS(6762),
    [anon_sym__Alignas] = ACTIONS(6762),
    [anon_sym_COLON] = ACTIONS(6762),
    [anon_sym_in] = ACTIONS(6760),
    [anon_sym_asm] = ACTIONS(6762),
    [anon_sym___asm__] = ACTIONS(6762),
    [anon_sym___asm] = ACTIONS(6760),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6762),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6762),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6762),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6762),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6762),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6762),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6762),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6762),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6762),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6762),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6760),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6762),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6762),
    [anon_sym_API_AVAILABLE] = ACTIONS(6762),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6762),
    [anon_sym_API_DEPRECATED] = ACTIONS(6762),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6762),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6762),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6762),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6762),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6762),
    [anon_sym___deprecated_msg] = ACTIONS(6762),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6762),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6762),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6762),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6762),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6762),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6762),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6762),
  },
  [STATE(3072)] = {
    [sym__declaration_modifiers] = STATE(5143),
    [sym_attribute_specifier] = STATE(5143),
    [sym_attribute_declaration] = STATE(5143),
    [sym_ms_declspec_modifier] = STATE(5143),
    [sym_storage_class_specifier] = STATE(5143),
    [sym_type_qualifier] = STATE(5143),
    [sym_alignas_qualifier] = STATE(5135),
    [sym_availability_attribute_specifier] = STATE(5143),
    [sym_identifier] = ACTIONS(7613),
    [anon_sym___extension__] = ACTIONS(7465),
    [anon_sym_extern] = ACTIONS(7467),
    [anon_sym___attribute__] = ACTIONS(7469),
    [anon_sym___attribute] = ACTIONS(7469),
    [anon_sym_noreturn] = ACTIONS(7465),
    [anon_sym_LBRACK] = ACTIONS(7471),
    [anon_sym___declspec] = ACTIONS(7473),
    [anon_sym_static] = ACTIONS(7467),
    [anon_sym_auto] = ACTIONS(7467),
    [anon_sym_register] = ACTIONS(7467),
    [anon_sym_inline] = ACTIONS(7467),
    [anon_sym___inline] = ACTIONS(7467),
    [anon_sym___inline__] = ACTIONS(7467),
    [anon_sym___forceinline] = ACTIONS(7467),
    [anon_sym_thread_local] = ACTIONS(7467),
    [anon_sym___thread] = ACTIONS(7467),
    [anon_sym_CG_EXTERN] = ACTIONS(7467),
    [anon_sym_CG_INLINE] = ACTIONS(7467),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7467),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7467),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7467),
    [anon_sym_IBOutlet] = ACTIONS(7467),
    [anon_sym_IBInspectable] = ACTIONS(7467),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7467),
    [anon_sym_NS_INLINE] = ACTIONS(7467),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7467),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7467),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7467),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7467),
    [anon_sym_const] = ACTIONS(7652),
    [anon_sym_constexpr] = ACTIONS(7465),
    [anon_sym_volatile] = ACTIONS(7465),
    [anon_sym_restrict] = ACTIONS(7465),
    [anon_sym___restrict__] = ACTIONS(7465),
    [anon_sym__Atomic] = ACTIONS(7465),
    [anon_sym__Noreturn] = ACTIONS(7465),
    [anon_sym__Nonnull] = ACTIONS(7465),
    [anon_sym_nullable] = ACTIONS(7465),
    [anon_sym__Complex] = ACTIONS(7465),
    [anon_sym__Nullable] = ACTIONS(7465),
    [anon_sym__Nullable_result] = ACTIONS(7465),
    [anon_sym__Null_unspecified] = ACTIONS(7465),
    [anon_sym___autoreleasing] = ACTIONS(7465),
    [anon_sym___block] = ACTIONS(7465),
    [anon_sym___bridge] = ACTIONS(7465),
    [anon_sym___bridge_retained] = ACTIONS(7465),
    [anon_sym___bridge_transfer] = ACTIONS(7465),
    [anon_sym___complex] = ACTIONS(7465),
    [anon_sym___const] = ACTIONS(7465),
    [anon_sym___imag] = ACTIONS(7465),
    [anon_sym___kindof] = ACTIONS(7465),
    [anon_sym___nonnull] = ACTIONS(7465),
    [anon_sym___nullable] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7465),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7465),
    [anon_sym___real] = ACTIONS(7465),
    [anon_sym___strong] = ACTIONS(7465),
    [anon_sym___unsafe_unretained] = ACTIONS(7465),
    [anon_sym___unused] = ACTIONS(7465),
    [anon_sym___weak] = ACTIONS(7465),
    [anon_sym_alignas] = ACTIONS(7477),
    [anon_sym__Alignas] = ACTIONS(7477),
    [anon_sym_struct] = ACTIONS(7617),
    [anon_sym_in] = ACTIONS(7617),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7481),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7481),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7481),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7481),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7481),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7481),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7481),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7481),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7481),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7483),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7483),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7483),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_API_AVAILABLE] = ACTIONS(7483),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7483),
    [anon_sym_API_DEPRECATED] = ACTIONS(7483),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7483),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7483),
    [anon_sym___deprecated_msg] = ACTIONS(7483),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7483),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7483),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7483),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7483),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7483),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7483),
    [anon_sym_id] = ACTIONS(7617),
  },
  [STATE(3073)] = {
    [sym__declaration_modifiers] = STATE(6462),
    [sym_attribute_specifier] = STATE(6462),
    [sym_attribute_declaration] = STATE(6462),
    [sym_ms_declspec_modifier] = STATE(6462),
    [sym_storage_class_specifier] = STATE(6462),
    [sym_type_qualifier] = STATE(6462),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6462),
    [aux_sym_struct_declaration_repeat2] = STATE(3193),
    [anon_sym_COMMA] = ACTIONS(7654),
    [anon_sym_SEMI] = ACTIONS(7656),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3074)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2310),
    [anon_sym_SEMI] = ACTIONS(7658),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3075)] = {
    [sym__declaration_modifiers] = STATE(6369),
    [sym_attribute_specifier] = STATE(6369),
    [sym_attribute_declaration] = STATE(6369),
    [sym_ms_declspec_modifier] = STATE(6369),
    [sym_storage_class_specifier] = STATE(6369),
    [sym_type_qualifier] = STATE(6369),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6369),
    [aux_sym_declaration_repeat1] = STATE(3078),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7660),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3076)] = {
    [sym__declaration_modifiers] = STATE(6460),
    [sym_attribute_specifier] = STATE(6460),
    [sym_attribute_declaration] = STATE(6460),
    [sym_ms_declspec_modifier] = STATE(6460),
    [sym_storage_class_specifier] = STATE(6460),
    [sym_type_qualifier] = STATE(6460),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6460),
    [aux_sym_declaration_repeat1] = STATE(3081),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7662),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3077)] = {
    [sym__declaration_modifiers] = STATE(6466),
    [sym_attribute_specifier] = STATE(6466),
    [sym_attribute_declaration] = STATE(6466),
    [sym_ms_declspec_modifier] = STATE(6466),
    [sym_storage_class_specifier] = STATE(6466),
    [sym_type_qualifier] = STATE(6466),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6466),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7664),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3078)] = {
    [sym__declaration_modifiers] = STATE(6470),
    [sym_attribute_specifier] = STATE(6470),
    [sym_attribute_declaration] = STATE(6470),
    [sym_ms_declspec_modifier] = STATE(6470),
    [sym_storage_class_specifier] = STATE(6470),
    [sym_type_qualifier] = STATE(6470),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6470),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7666),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3079)] = {
    [sym__declaration_modifiers] = STATE(6472),
    [sym_attribute_specifier] = STATE(6472),
    [sym_attribute_declaration] = STATE(6472),
    [sym_ms_declspec_modifier] = STATE(6472),
    [sym_storage_class_specifier] = STATE(6472),
    [sym_type_qualifier] = STATE(6472),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6472),
    [aux_sym_declaration_repeat1] = STATE(3082),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7668),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3080)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2172),
    [anon_sym_SEMI] = ACTIONS(7670),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3081)] = {
    [sym__declaration_modifiers] = STATE(6493),
    [sym_attribute_specifier] = STATE(6493),
    [sym_attribute_declaration] = STATE(6493),
    [sym_ms_declspec_modifier] = STATE(6493),
    [sym_storage_class_specifier] = STATE(6493),
    [sym_type_qualifier] = STATE(6493),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6493),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7672),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3082)] = {
    [sym__declaration_modifiers] = STATE(6498),
    [sym_attribute_specifier] = STATE(6498),
    [sym_attribute_declaration] = STATE(6498),
    [sym_ms_declspec_modifier] = STATE(6498),
    [sym_storage_class_specifier] = STATE(6498),
    [sym_type_qualifier] = STATE(6498),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6498),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7674),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3083)] = {
    [sym__declaration_modifiers] = STATE(6554),
    [sym_attribute_specifier] = STATE(6554),
    [sym_attribute_declaration] = STATE(6554),
    [sym_ms_declspec_modifier] = STATE(6554),
    [sym_storage_class_specifier] = STATE(6554),
    [sym_type_qualifier] = STATE(6554),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6554),
    [aux_sym_declaration_repeat1] = STATE(3124),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7676),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3084)] = {
    [sym__declaration_modifiers] = STATE(6627),
    [sym_attribute_specifier] = STATE(6627),
    [sym_attribute_declaration] = STATE(6627),
    [sym_ms_declspec_modifier] = STATE(6627),
    [sym_storage_class_specifier] = STATE(6627),
    [sym_type_qualifier] = STATE(6627),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6627),
    [aux_sym_declaration_repeat1] = STATE(3087),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7678),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3085)] = {
    [sym__declaration_modifiers] = STATE(6667),
    [sym_attribute_specifier] = STATE(6667),
    [sym_attribute_declaration] = STATE(6667),
    [sym_ms_declspec_modifier] = STATE(6667),
    [sym_storage_class_specifier] = STATE(6667),
    [sym_type_qualifier] = STATE(6667),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6667),
    [aux_sym_declaration_repeat1] = STATE(3089),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7680),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3086)] = {
    [sym__declaration_modifiers] = STATE(6687),
    [sym_attribute_specifier] = STATE(6687),
    [sym_attribute_declaration] = STATE(6687),
    [sym_ms_declspec_modifier] = STATE(6687),
    [sym_storage_class_specifier] = STATE(6687),
    [sym_type_qualifier] = STATE(6687),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6687),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7682),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3087)] = {
    [sym__declaration_modifiers] = STATE(6691),
    [sym_attribute_specifier] = STATE(6691),
    [sym_attribute_declaration] = STATE(6691),
    [sym_ms_declspec_modifier] = STATE(6691),
    [sym_storage_class_specifier] = STATE(6691),
    [sym_type_qualifier] = STATE(6691),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6691),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7684),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3088)] = {
    [sym__declaration_modifiers] = STATE(6705),
    [sym_attribute_specifier] = STATE(6705),
    [sym_attribute_declaration] = STATE(6705),
    [sym_ms_declspec_modifier] = STATE(6705),
    [sym_storage_class_specifier] = STATE(6705),
    [sym_type_qualifier] = STATE(6705),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6705),
    [aux_sym_declaration_repeat1] = STATE(3090),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7686),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3089)] = {
    [sym__declaration_modifiers] = STATE(6775),
    [sym_attribute_specifier] = STATE(6775),
    [sym_attribute_declaration] = STATE(6775),
    [sym_ms_declspec_modifier] = STATE(6775),
    [sym_storage_class_specifier] = STATE(6775),
    [sym_type_qualifier] = STATE(6775),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6775),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7688),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3090)] = {
    [sym__declaration_modifiers] = STATE(6778),
    [sym_attribute_specifier] = STATE(6778),
    [sym_attribute_declaration] = STATE(6778),
    [sym_ms_declspec_modifier] = STATE(6778),
    [sym_storage_class_specifier] = STATE(6778),
    [sym_type_qualifier] = STATE(6778),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6778),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7690),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3091)] = {
    [sym__declaration_modifiers] = STATE(5945),
    [sym_attribute_specifier] = STATE(5945),
    [sym_attribute_declaration] = STATE(5945),
    [sym_ms_declspec_modifier] = STATE(5945),
    [sym_storage_class_specifier] = STATE(5945),
    [sym_type_qualifier] = STATE(5945),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5945),
    [aux_sym_declaration_repeat1] = STATE(3094),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7692),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3092)] = {
    [sym__declaration_modifiers] = STATE(5970),
    [sym_attribute_specifier] = STATE(5970),
    [sym_attribute_declaration] = STATE(5970),
    [sym_ms_declspec_modifier] = STATE(5970),
    [sym_storage_class_specifier] = STATE(5970),
    [sym_type_qualifier] = STATE(5970),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5970),
    [aux_sym_declaration_repeat1] = STATE(3096),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7694),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3093)] = {
    [sym__declaration_modifiers] = STATE(5972),
    [sym_attribute_specifier] = STATE(5972),
    [sym_attribute_declaration] = STATE(5972),
    [sym_ms_declspec_modifier] = STATE(5972),
    [sym_storage_class_specifier] = STATE(5972),
    [sym_type_qualifier] = STATE(5972),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5972),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7696),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3094)] = {
    [sym__declaration_modifiers] = STATE(5976),
    [sym_attribute_specifier] = STATE(5976),
    [sym_attribute_declaration] = STATE(5976),
    [sym_ms_declspec_modifier] = STATE(5976),
    [sym_storage_class_specifier] = STATE(5976),
    [sym_type_qualifier] = STATE(5976),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5976),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7698),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3095)] = {
    [sym__declaration_modifiers] = STATE(5977),
    [sym_attribute_specifier] = STATE(5977),
    [sym_attribute_declaration] = STATE(5977),
    [sym_ms_declspec_modifier] = STATE(5977),
    [sym_storage_class_specifier] = STATE(5977),
    [sym_type_qualifier] = STATE(5977),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5977),
    [aux_sym_declaration_repeat1] = STATE(3097),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7700),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3096)] = {
    [sym__declaration_modifiers] = STATE(5988),
    [sym_attribute_specifier] = STATE(5988),
    [sym_attribute_declaration] = STATE(5988),
    [sym_ms_declspec_modifier] = STATE(5988),
    [sym_storage_class_specifier] = STATE(5988),
    [sym_type_qualifier] = STATE(5988),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5988),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7702),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3097)] = {
    [sym__declaration_modifiers] = STATE(5991),
    [sym_attribute_specifier] = STATE(5991),
    [sym_attribute_declaration] = STATE(5991),
    [sym_ms_declspec_modifier] = STATE(5991),
    [sym_storage_class_specifier] = STATE(5991),
    [sym_type_qualifier] = STATE(5991),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5991),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7704),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3098)] = {
    [sym__declaration_modifiers] = STATE(6021),
    [sym_attribute_specifier] = STATE(6021),
    [sym_attribute_declaration] = STATE(6021),
    [sym_ms_declspec_modifier] = STATE(6021),
    [sym_storage_class_specifier] = STATE(6021),
    [sym_type_qualifier] = STATE(6021),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6021),
    [aux_sym_declaration_repeat1] = STATE(3101),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7706),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3099)] = {
    [sym__declaration_modifiers] = STATE(6031),
    [sym_attribute_specifier] = STATE(6031),
    [sym_attribute_declaration] = STATE(6031),
    [sym_ms_declspec_modifier] = STATE(6031),
    [sym_storage_class_specifier] = STATE(6031),
    [sym_type_qualifier] = STATE(6031),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6031),
    [aux_sym_declaration_repeat1] = STATE(3103),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7708),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3100)] = {
    [sym__declaration_modifiers] = STATE(6035),
    [sym_attribute_specifier] = STATE(6035),
    [sym_attribute_declaration] = STATE(6035),
    [sym_ms_declspec_modifier] = STATE(6035),
    [sym_storage_class_specifier] = STATE(6035),
    [sym_type_qualifier] = STATE(6035),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6035),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7710),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3101)] = {
    [sym__declaration_modifiers] = STATE(6037),
    [sym_attribute_specifier] = STATE(6037),
    [sym_attribute_declaration] = STATE(6037),
    [sym_ms_declspec_modifier] = STATE(6037),
    [sym_storage_class_specifier] = STATE(6037),
    [sym_type_qualifier] = STATE(6037),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6037),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7712),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3102)] = {
    [sym__declaration_modifiers] = STATE(6038),
    [sym_attribute_specifier] = STATE(6038),
    [sym_attribute_declaration] = STATE(6038),
    [sym_ms_declspec_modifier] = STATE(6038),
    [sym_storage_class_specifier] = STATE(6038),
    [sym_type_qualifier] = STATE(6038),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6038),
    [aux_sym_declaration_repeat1] = STATE(3104),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7714),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3103)] = {
    [sym__declaration_modifiers] = STATE(6050),
    [sym_attribute_specifier] = STATE(6050),
    [sym_attribute_declaration] = STATE(6050),
    [sym_ms_declspec_modifier] = STATE(6050),
    [sym_storage_class_specifier] = STATE(6050),
    [sym_type_qualifier] = STATE(6050),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6050),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7716),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3104)] = {
    [sym__declaration_modifiers] = STATE(6055),
    [sym_attribute_specifier] = STATE(6055),
    [sym_attribute_declaration] = STATE(6055),
    [sym_ms_declspec_modifier] = STATE(6055),
    [sym_storage_class_specifier] = STATE(6055),
    [sym_type_qualifier] = STATE(6055),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6055),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7718),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3105)] = {
    [sym__declaration_modifiers] = STATE(6682),
    [sym_attribute_specifier] = STATE(6682),
    [sym_attribute_declaration] = STATE(6682),
    [sym_ms_declspec_modifier] = STATE(6682),
    [sym_storage_class_specifier] = STATE(6682),
    [sym_type_qualifier] = STATE(6682),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6682),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7720),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3106)] = {
    [sym__declaration_modifiers] = STATE(6076),
    [sym_attribute_specifier] = STATE(6076),
    [sym_attribute_declaration] = STATE(6076),
    [sym_ms_declspec_modifier] = STATE(6076),
    [sym_storage_class_specifier] = STATE(6076),
    [sym_type_qualifier] = STATE(6076),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6076),
    [aux_sym_declaration_repeat1] = STATE(3109),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7722),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3107)] = {
    [sym__declaration_modifiers] = STATE(6081),
    [sym_attribute_specifier] = STATE(6081),
    [sym_attribute_declaration] = STATE(6081),
    [sym_ms_declspec_modifier] = STATE(6081),
    [sym_storage_class_specifier] = STATE(6081),
    [sym_type_qualifier] = STATE(6081),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6081),
    [aux_sym_declaration_repeat1] = STATE(3111),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7724),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3108)] = {
    [sym__declaration_modifiers] = STATE(6083),
    [sym_attribute_specifier] = STATE(6083),
    [sym_attribute_declaration] = STATE(6083),
    [sym_ms_declspec_modifier] = STATE(6083),
    [sym_storage_class_specifier] = STATE(6083),
    [sym_type_qualifier] = STATE(6083),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6083),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7726),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3109)] = {
    [sym__declaration_modifiers] = STATE(6084),
    [sym_attribute_specifier] = STATE(6084),
    [sym_attribute_declaration] = STATE(6084),
    [sym_ms_declspec_modifier] = STATE(6084),
    [sym_storage_class_specifier] = STATE(6084),
    [sym_type_qualifier] = STATE(6084),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6084),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7728),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3110)] = {
    [sym__declaration_modifiers] = STATE(6085),
    [sym_attribute_specifier] = STATE(6085),
    [sym_attribute_declaration] = STATE(6085),
    [sym_ms_declspec_modifier] = STATE(6085),
    [sym_storage_class_specifier] = STATE(6085),
    [sym_type_qualifier] = STATE(6085),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6085),
    [aux_sym_declaration_repeat1] = STATE(3112),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7730),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3111)] = {
    [sym__declaration_modifiers] = STATE(6091),
    [sym_attribute_specifier] = STATE(6091),
    [sym_attribute_declaration] = STATE(6091),
    [sym_ms_declspec_modifier] = STATE(6091),
    [sym_storage_class_specifier] = STATE(6091),
    [sym_type_qualifier] = STATE(6091),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6091),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7732),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3112)] = {
    [sym__declaration_modifiers] = STATE(6096),
    [sym_attribute_specifier] = STATE(6096),
    [sym_attribute_declaration] = STATE(6096),
    [sym_ms_declspec_modifier] = STATE(6096),
    [sym_storage_class_specifier] = STATE(6096),
    [sym_type_qualifier] = STATE(6096),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6096),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7734),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3113)] = {
    [sym__declaration_modifiers] = STATE(6614),
    [sym_attribute_specifier] = STATE(6614),
    [sym_attribute_declaration] = STATE(6614),
    [sym_ms_declspec_modifier] = STATE(6614),
    [sym_storage_class_specifier] = STATE(6614),
    [sym_type_qualifier] = STATE(6614),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6614),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7736),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3114)] = {
    [sym__declaration_modifiers] = STATE(6615),
    [sym_attribute_specifier] = STATE(6615),
    [sym_attribute_declaration] = STATE(6615),
    [sym_ms_declspec_modifier] = STATE(6615),
    [sym_storage_class_specifier] = STATE(6615),
    [sym_type_qualifier] = STATE(6615),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6615),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7738),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3115)] = {
    [sym__declaration_modifiers] = STATE(6619),
    [sym_attribute_specifier] = STATE(6619),
    [sym_attribute_declaration] = STATE(6619),
    [sym_ms_declspec_modifier] = STATE(6619),
    [sym_storage_class_specifier] = STATE(6619),
    [sym_type_qualifier] = STATE(6619),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6619),
    [aux_sym_declaration_repeat1] = STATE(3141),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7740),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3116)] = {
    [sym__declaration_modifiers] = STATE(6683),
    [sym_attribute_specifier] = STATE(6683),
    [sym_attribute_declaration] = STATE(6683),
    [sym_ms_declspec_modifier] = STATE(6683),
    [sym_storage_class_specifier] = STATE(6683),
    [sym_type_qualifier] = STATE(6683),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6683),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7742),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3117)] = {
    [sym__declaration_modifiers] = STATE(6372),
    [sym_attribute_specifier] = STATE(6372),
    [sym_attribute_declaration] = STATE(6372),
    [sym_ms_declspec_modifier] = STATE(6372),
    [sym_storage_class_specifier] = STATE(6372),
    [sym_type_qualifier] = STATE(6372),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6372),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7744),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3118)] = {
    [sym__declaration_modifiers] = STATE(6347),
    [sym_attribute_specifier] = STATE(6347),
    [sym_attribute_declaration] = STATE(6347),
    [sym_ms_declspec_modifier] = STATE(6347),
    [sym_storage_class_specifier] = STATE(6347),
    [sym_type_qualifier] = STATE(6347),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6347),
    [aux_sym_declaration_repeat1] = STATE(3154),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7746),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3119)] = {
    [sym__declaration_modifiers] = STATE(6405),
    [sym_attribute_specifier] = STATE(6405),
    [sym_attribute_declaration] = STATE(6405),
    [sym_ms_declspec_modifier] = STATE(6405),
    [sym_storage_class_specifier] = STATE(6405),
    [sym_type_qualifier] = STATE(6405),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6405),
    [aux_sym_struct_declaration_repeat2] = STATE(3276),
    [anon_sym_COMMA] = ACTIONS(7654),
    [anon_sym_SEMI] = ACTIONS(7748),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3120)] = {
    [sym__declaration_modifiers] = STATE(6563),
    [sym_attribute_specifier] = STATE(6563),
    [sym_attribute_declaration] = STATE(6563),
    [sym_ms_declspec_modifier] = STATE(6563),
    [sym_storage_class_specifier] = STATE(6563),
    [sym_type_qualifier] = STATE(6563),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6563),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7750),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3121)] = {
    [sym__declaration_modifiers] = STATE(6776),
    [sym_attribute_specifier] = STATE(6776),
    [sym_attribute_declaration] = STATE(6776),
    [sym_ms_declspec_modifier] = STATE(6776),
    [sym_storage_class_specifier] = STATE(6776),
    [sym_type_qualifier] = STATE(6776),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6776),
    [aux_sym_struct_declaration_repeat2] = STATE(3119),
    [anon_sym_COMMA] = ACTIONS(7654),
    [anon_sym_SEMI] = ACTIONS(7752),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3122)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2183),
    [anon_sym_SEMI] = ACTIONS(7500),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3123)] = {
    [sym__declaration_modifiers] = STATE(6611),
    [sym_attribute_specifier] = STATE(6611),
    [sym_attribute_declaration] = STATE(6611),
    [sym_ms_declspec_modifier] = STATE(6611),
    [sym_storage_class_specifier] = STATE(6611),
    [sym_type_qualifier] = STATE(6611),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6611),
    [aux_sym_declaration_repeat1] = STATE(3132),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7754),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3124)] = {
    [sym__declaration_modifiers] = STATE(6290),
    [sym_attribute_specifier] = STATE(6290),
    [sym_attribute_declaration] = STATE(6290),
    [sym_ms_declspec_modifier] = STATE(6290),
    [sym_storage_class_specifier] = STATE(6290),
    [sym_type_qualifier] = STATE(6290),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6290),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7756),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3125)] = {
    [sym__declaration_modifiers] = STATE(3195),
    [sym_attribute_specifier] = STATE(3195),
    [sym_attribute_declaration] = STATE(3195),
    [sym_ms_declspec_modifier] = STATE(3195),
    [sym_storage_class_specifier] = STATE(3195),
    [sym_type_qualifier] = STATE(3195),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3195),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3195),
    [aux_sym_method_declaration_repeat2] = STATE(2270),
    [anon_sym_SEMI] = ACTIONS(7496),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3126)] = {
    [sym__declaration_modifiers] = STATE(6314),
    [sym_attribute_specifier] = STATE(6314),
    [sym_attribute_declaration] = STATE(6314),
    [sym_ms_declspec_modifier] = STATE(6314),
    [sym_storage_class_specifier] = STATE(6314),
    [sym_type_qualifier] = STATE(6314),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6314),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7758),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3127)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2290),
    [anon_sym_SEMI] = ACTIONS(7502),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3128)] = {
    [sym_attribute_specifier] = STATE(2941),
    [sym_identifier] = ACTIONS(7760),
    [anon_sym_COMMA] = ACTIONS(7762),
    [anon_sym_RPAREN] = ACTIONS(7762),
    [anon_sym_LPAREN2] = ACTIONS(7762),
    [anon_sym_STAR] = ACTIONS(7762),
    [anon_sym_CARET] = ACTIONS(7762),
    [anon_sym_SEMI] = ACTIONS(7762),
    [anon_sym___extension__] = ACTIONS(7760),
    [anon_sym_extern] = ACTIONS(7760),
    [anon_sym___attribute__] = ACTIONS(7764),
    [anon_sym___attribute] = ACTIONS(7764),
    [anon_sym_noreturn] = ACTIONS(7760),
    [anon_sym_LBRACK] = ACTIONS(7762),
    [anon_sym___declspec] = ACTIONS(7760),
    [anon_sym___based] = ACTIONS(7760),
    [anon_sym_static] = ACTIONS(7760),
    [anon_sym_auto] = ACTIONS(7760),
    [anon_sym_register] = ACTIONS(7760),
    [anon_sym_inline] = ACTIONS(7760),
    [anon_sym___inline] = ACTIONS(7760),
    [anon_sym___inline__] = ACTIONS(7760),
    [anon_sym___forceinline] = ACTIONS(7760),
    [anon_sym_thread_local] = ACTIONS(7760),
    [anon_sym___thread] = ACTIONS(7760),
    [anon_sym_CG_EXTERN] = ACTIONS(7760),
    [anon_sym_CG_INLINE] = ACTIONS(7760),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7760),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7760),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7760),
    [anon_sym_IBOutlet] = ACTIONS(7760),
    [anon_sym_IBInspectable] = ACTIONS(7760),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7760),
    [anon_sym_NS_INLINE] = ACTIONS(7760),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7760),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7760),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7760),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7760),
    [anon_sym_const] = ACTIONS(7760),
    [anon_sym_constexpr] = ACTIONS(7760),
    [anon_sym_volatile] = ACTIONS(7760),
    [anon_sym_restrict] = ACTIONS(7760),
    [anon_sym___restrict__] = ACTIONS(7760),
    [anon_sym__Atomic] = ACTIONS(7760),
    [anon_sym__Noreturn] = ACTIONS(7760),
    [anon_sym__Nonnull] = ACTIONS(7760),
    [anon_sym_nullable] = ACTIONS(7760),
    [anon_sym__Complex] = ACTIONS(7760),
    [anon_sym__Nullable] = ACTIONS(7760),
    [anon_sym__Nullable_result] = ACTIONS(7760),
    [anon_sym__Null_unspecified] = ACTIONS(7760),
    [anon_sym___autoreleasing] = ACTIONS(7760),
    [anon_sym___block] = ACTIONS(7760),
    [anon_sym___bridge] = ACTIONS(7760),
    [anon_sym___bridge_retained] = ACTIONS(7760),
    [anon_sym___bridge_transfer] = ACTIONS(7760),
    [anon_sym___complex] = ACTIONS(7760),
    [anon_sym___const] = ACTIONS(7760),
    [anon_sym___imag] = ACTIONS(7760),
    [anon_sym___kindof] = ACTIONS(7760),
    [anon_sym___nonnull] = ACTIONS(7760),
    [anon_sym___nullable] = ACTIONS(7760),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7760),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7760),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7760),
    [anon_sym___real] = ACTIONS(7760),
    [anon_sym___strong] = ACTIONS(7760),
    [anon_sym___unsafe_unretained] = ACTIONS(7760),
    [anon_sym___unused] = ACTIONS(7760),
    [anon_sym___weak] = ACTIONS(7760),
    [anon_sym_alignas] = ACTIONS(7760),
    [anon_sym__Alignas] = ACTIONS(7760),
    [anon_sym_enum] = ACTIONS(7760),
    [anon_sym_COLON] = ACTIONS(7762),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7760),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7760),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7760),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7760),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7760),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7760),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7760),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7760),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7760),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7760),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7760),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7760),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7760),
    [anon_sym_API_AVAILABLE] = ACTIONS(7760),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7760),
    [anon_sym_API_DEPRECATED] = ACTIONS(7760),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7760),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7760),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7760),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7760),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7760),
    [anon_sym___deprecated_msg] = ACTIONS(7760),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7760),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7760),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7760),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7760),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7760),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7760),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7760),
  },
  [STATE(3129)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2133),
    [anon_sym_SEMI] = ACTIONS(7767),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3130)] = {
    [sym__declaration_modifiers] = STATE(6696),
    [sym_attribute_specifier] = STATE(6696),
    [sym_attribute_declaration] = STATE(6696),
    [sym_ms_declspec_modifier] = STATE(6696),
    [sym_storage_class_specifier] = STATE(6696),
    [sym_type_qualifier] = STATE(6696),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6696),
    [aux_sym_struct_declaration_repeat2] = STATE(3152),
    [anon_sym_COMMA] = ACTIONS(7654),
    [anon_sym_SEMI] = ACTIONS(7769),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3131)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2121),
    [anon_sym_SEMI] = ACTIONS(7555),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3132)] = {
    [sym__declaration_modifiers] = STATE(6677),
    [sym_attribute_specifier] = STATE(6677),
    [sym_attribute_declaration] = STATE(6677),
    [sym_ms_declspec_modifier] = STATE(6677),
    [sym_storage_class_specifier] = STATE(6677),
    [sym_type_qualifier] = STATE(6677),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6677),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7771),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3133)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2313),
    [anon_sym_SEMI] = ACTIONS(7773),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3134)] = {
    [sym_attribute_specifier] = STATE(2944),
    [sym_identifier] = ACTIONS(7775),
    [anon_sym_COMMA] = ACTIONS(7777),
    [anon_sym_RPAREN] = ACTIONS(7777),
    [anon_sym_LPAREN2] = ACTIONS(7777),
    [anon_sym_STAR] = ACTIONS(7777),
    [anon_sym_CARET] = ACTIONS(7777),
    [anon_sym_SEMI] = ACTIONS(7777),
    [anon_sym___extension__] = ACTIONS(7775),
    [anon_sym_extern] = ACTIONS(7775),
    [anon_sym___attribute__] = ACTIONS(7779),
    [anon_sym___attribute] = ACTIONS(7779),
    [anon_sym_noreturn] = ACTIONS(7775),
    [anon_sym_LBRACK] = ACTIONS(7777),
    [anon_sym___declspec] = ACTIONS(7775),
    [anon_sym___based] = ACTIONS(7775),
    [anon_sym_static] = ACTIONS(7775),
    [anon_sym_auto] = ACTIONS(7775),
    [anon_sym_register] = ACTIONS(7775),
    [anon_sym_inline] = ACTIONS(7775),
    [anon_sym___inline] = ACTIONS(7775),
    [anon_sym___inline__] = ACTIONS(7775),
    [anon_sym___forceinline] = ACTIONS(7775),
    [anon_sym_thread_local] = ACTIONS(7775),
    [anon_sym___thread] = ACTIONS(7775),
    [anon_sym_CG_EXTERN] = ACTIONS(7775),
    [anon_sym_CG_INLINE] = ACTIONS(7775),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7775),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7775),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7775),
    [anon_sym_IBOutlet] = ACTIONS(7775),
    [anon_sym_IBInspectable] = ACTIONS(7775),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7775),
    [anon_sym_NS_INLINE] = ACTIONS(7775),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7775),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7775),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7775),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7775),
    [anon_sym_const] = ACTIONS(7775),
    [anon_sym_constexpr] = ACTIONS(7775),
    [anon_sym_volatile] = ACTIONS(7775),
    [anon_sym_restrict] = ACTIONS(7775),
    [anon_sym___restrict__] = ACTIONS(7775),
    [anon_sym__Atomic] = ACTIONS(7775),
    [anon_sym__Noreturn] = ACTIONS(7775),
    [anon_sym__Nonnull] = ACTIONS(7775),
    [anon_sym_nullable] = ACTIONS(7775),
    [anon_sym__Complex] = ACTIONS(7775),
    [anon_sym__Nullable] = ACTIONS(7775),
    [anon_sym__Nullable_result] = ACTIONS(7775),
    [anon_sym__Null_unspecified] = ACTIONS(7775),
    [anon_sym___autoreleasing] = ACTIONS(7775),
    [anon_sym___block] = ACTIONS(7775),
    [anon_sym___bridge] = ACTIONS(7775),
    [anon_sym___bridge_retained] = ACTIONS(7775),
    [anon_sym___bridge_transfer] = ACTIONS(7775),
    [anon_sym___complex] = ACTIONS(7775),
    [anon_sym___const] = ACTIONS(7775),
    [anon_sym___imag] = ACTIONS(7775),
    [anon_sym___kindof] = ACTIONS(7775),
    [anon_sym___nonnull] = ACTIONS(7775),
    [anon_sym___nullable] = ACTIONS(7775),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7775),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7775),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7775),
    [anon_sym___real] = ACTIONS(7775),
    [anon_sym___strong] = ACTIONS(7775),
    [anon_sym___unsafe_unretained] = ACTIONS(7775),
    [anon_sym___unused] = ACTIONS(7775),
    [anon_sym___weak] = ACTIONS(7775),
    [anon_sym_alignas] = ACTIONS(7775),
    [anon_sym__Alignas] = ACTIONS(7775),
    [anon_sym_enum] = ACTIONS(7775),
    [anon_sym_COLON] = ACTIONS(7777),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7775),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7775),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7775),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7775),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7775),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7775),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7775),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7775),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7775),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7775),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7775),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7775),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7775),
    [anon_sym_API_AVAILABLE] = ACTIONS(7775),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7775),
    [anon_sym_API_DEPRECATED] = ACTIONS(7775),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7775),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7775),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7775),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7775),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7775),
    [anon_sym___deprecated_msg] = ACTIONS(7775),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7775),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7775),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7775),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7775),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7775),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7775),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7775),
  },
  [STATE(3135)] = {
    [sym_attribute_specifier] = STATE(2950),
    [sym_identifier] = ACTIONS(7782),
    [anon_sym_COMMA] = ACTIONS(7784),
    [anon_sym_RPAREN] = ACTIONS(7784),
    [anon_sym_LPAREN2] = ACTIONS(7784),
    [anon_sym_STAR] = ACTIONS(7784),
    [anon_sym_CARET] = ACTIONS(7784),
    [anon_sym_SEMI] = ACTIONS(7784),
    [anon_sym___extension__] = ACTIONS(7782),
    [anon_sym_extern] = ACTIONS(7782),
    [anon_sym___attribute__] = ACTIONS(7786),
    [anon_sym___attribute] = ACTIONS(7786),
    [anon_sym_noreturn] = ACTIONS(7782),
    [anon_sym_LBRACK] = ACTIONS(7784),
    [anon_sym___declspec] = ACTIONS(7782),
    [anon_sym___based] = ACTIONS(7782),
    [anon_sym_static] = ACTIONS(7782),
    [anon_sym_auto] = ACTIONS(7782),
    [anon_sym_register] = ACTIONS(7782),
    [anon_sym_inline] = ACTIONS(7782),
    [anon_sym___inline] = ACTIONS(7782),
    [anon_sym___inline__] = ACTIONS(7782),
    [anon_sym___forceinline] = ACTIONS(7782),
    [anon_sym_thread_local] = ACTIONS(7782),
    [anon_sym___thread] = ACTIONS(7782),
    [anon_sym_CG_EXTERN] = ACTIONS(7782),
    [anon_sym_CG_INLINE] = ACTIONS(7782),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7782),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7782),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7782),
    [anon_sym_IBOutlet] = ACTIONS(7782),
    [anon_sym_IBInspectable] = ACTIONS(7782),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7782),
    [anon_sym_NS_INLINE] = ACTIONS(7782),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7782),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7782),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7782),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7782),
    [anon_sym_const] = ACTIONS(7782),
    [anon_sym_constexpr] = ACTIONS(7782),
    [anon_sym_volatile] = ACTIONS(7782),
    [anon_sym_restrict] = ACTIONS(7782),
    [anon_sym___restrict__] = ACTIONS(7782),
    [anon_sym__Atomic] = ACTIONS(7782),
    [anon_sym__Noreturn] = ACTIONS(7782),
    [anon_sym__Nonnull] = ACTIONS(7782),
    [anon_sym_nullable] = ACTIONS(7782),
    [anon_sym__Complex] = ACTIONS(7782),
    [anon_sym__Nullable] = ACTIONS(7782),
    [anon_sym__Nullable_result] = ACTIONS(7782),
    [anon_sym__Null_unspecified] = ACTIONS(7782),
    [anon_sym___autoreleasing] = ACTIONS(7782),
    [anon_sym___block] = ACTIONS(7782),
    [anon_sym___bridge] = ACTIONS(7782),
    [anon_sym___bridge_retained] = ACTIONS(7782),
    [anon_sym___bridge_transfer] = ACTIONS(7782),
    [anon_sym___complex] = ACTIONS(7782),
    [anon_sym___const] = ACTIONS(7782),
    [anon_sym___imag] = ACTIONS(7782),
    [anon_sym___kindof] = ACTIONS(7782),
    [anon_sym___nonnull] = ACTIONS(7782),
    [anon_sym___nullable] = ACTIONS(7782),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7782),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7782),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7782),
    [anon_sym___real] = ACTIONS(7782),
    [anon_sym___strong] = ACTIONS(7782),
    [anon_sym___unsafe_unretained] = ACTIONS(7782),
    [anon_sym___unused] = ACTIONS(7782),
    [anon_sym___weak] = ACTIONS(7782),
    [anon_sym_alignas] = ACTIONS(7782),
    [anon_sym__Alignas] = ACTIONS(7782),
    [anon_sym_enum] = ACTIONS(7782),
    [anon_sym_COLON] = ACTIONS(7784),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7782),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7782),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7782),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7782),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7782),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7782),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7782),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7782),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7782),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7782),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7782),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7782),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7782),
    [anon_sym_API_AVAILABLE] = ACTIONS(7782),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7782),
    [anon_sym_API_DEPRECATED] = ACTIONS(7782),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7782),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7782),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7782),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7782),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7782),
    [anon_sym___deprecated_msg] = ACTIONS(7782),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7782),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7782),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7782),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7782),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7782),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7782),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7782),
  },
  [STATE(3136)] = {
    [sym_attribute_specifier] = STATE(2984),
    [sym_identifier] = ACTIONS(7789),
    [anon_sym_COMMA] = ACTIONS(7791),
    [anon_sym_RPAREN] = ACTIONS(7791),
    [anon_sym_LPAREN2] = ACTIONS(7791),
    [anon_sym_STAR] = ACTIONS(7791),
    [anon_sym_CARET] = ACTIONS(7791),
    [anon_sym_SEMI] = ACTIONS(7791),
    [anon_sym___extension__] = ACTIONS(7789),
    [anon_sym_extern] = ACTIONS(7789),
    [anon_sym___attribute__] = ACTIONS(7793),
    [anon_sym___attribute] = ACTIONS(7793),
    [anon_sym_noreturn] = ACTIONS(7789),
    [anon_sym_LBRACK] = ACTIONS(7791),
    [anon_sym___declspec] = ACTIONS(7789),
    [anon_sym___based] = ACTIONS(7789),
    [anon_sym_static] = ACTIONS(7789),
    [anon_sym_auto] = ACTIONS(7789),
    [anon_sym_register] = ACTIONS(7789),
    [anon_sym_inline] = ACTIONS(7789),
    [anon_sym___inline] = ACTIONS(7789),
    [anon_sym___inline__] = ACTIONS(7789),
    [anon_sym___forceinline] = ACTIONS(7789),
    [anon_sym_thread_local] = ACTIONS(7789),
    [anon_sym___thread] = ACTIONS(7789),
    [anon_sym_CG_EXTERN] = ACTIONS(7789),
    [anon_sym_CG_INLINE] = ACTIONS(7789),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7789),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7789),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7789),
    [anon_sym_IBOutlet] = ACTIONS(7789),
    [anon_sym_IBInspectable] = ACTIONS(7789),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7789),
    [anon_sym_NS_INLINE] = ACTIONS(7789),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7789),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7789),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7789),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7789),
    [anon_sym_const] = ACTIONS(7789),
    [anon_sym_constexpr] = ACTIONS(7789),
    [anon_sym_volatile] = ACTIONS(7789),
    [anon_sym_restrict] = ACTIONS(7789),
    [anon_sym___restrict__] = ACTIONS(7789),
    [anon_sym__Atomic] = ACTIONS(7789),
    [anon_sym__Noreturn] = ACTIONS(7789),
    [anon_sym__Nonnull] = ACTIONS(7789),
    [anon_sym_nullable] = ACTIONS(7789),
    [anon_sym__Complex] = ACTIONS(7789),
    [anon_sym__Nullable] = ACTIONS(7789),
    [anon_sym__Nullable_result] = ACTIONS(7789),
    [anon_sym__Null_unspecified] = ACTIONS(7789),
    [anon_sym___autoreleasing] = ACTIONS(7789),
    [anon_sym___block] = ACTIONS(7789),
    [anon_sym___bridge] = ACTIONS(7789),
    [anon_sym___bridge_retained] = ACTIONS(7789),
    [anon_sym___bridge_transfer] = ACTIONS(7789),
    [anon_sym___complex] = ACTIONS(7789),
    [anon_sym___const] = ACTIONS(7789),
    [anon_sym___imag] = ACTIONS(7789),
    [anon_sym___kindof] = ACTIONS(7789),
    [anon_sym___nonnull] = ACTIONS(7789),
    [anon_sym___nullable] = ACTIONS(7789),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7789),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7789),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7789),
    [anon_sym___real] = ACTIONS(7789),
    [anon_sym___strong] = ACTIONS(7789),
    [anon_sym___unsafe_unretained] = ACTIONS(7789),
    [anon_sym___unused] = ACTIONS(7789),
    [anon_sym___weak] = ACTIONS(7789),
    [anon_sym_alignas] = ACTIONS(7789),
    [anon_sym__Alignas] = ACTIONS(7789),
    [anon_sym_enum] = ACTIONS(7789),
    [anon_sym_COLON] = ACTIONS(7791),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7789),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7789),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7789),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7789),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7789),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7789),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7789),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7789),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7789),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7789),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7789),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7789),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7789),
    [anon_sym_API_AVAILABLE] = ACTIONS(7789),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7789),
    [anon_sym_API_DEPRECATED] = ACTIONS(7789),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7789),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7789),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7789),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7789),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7789),
    [anon_sym___deprecated_msg] = ACTIONS(7789),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7789),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7789),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7789),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7789),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7789),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7789),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7789),
  },
  [STATE(3137)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2167),
    [anon_sym_SEMI] = ACTIONS(7498),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3138)] = {
    [sym__declaration_modifiers] = STATE(6566),
    [sym_attribute_specifier] = STATE(6566),
    [sym_attribute_declaration] = STATE(6566),
    [sym_ms_declspec_modifier] = STATE(6566),
    [sym_storage_class_specifier] = STATE(6566),
    [sym_type_qualifier] = STATE(6566),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6566),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7796),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3139)] = {
    [sym__declaration_modifiers] = STATE(6582),
    [sym_attribute_specifier] = STATE(6582),
    [sym_attribute_declaration] = STATE(6582),
    [sym_ms_declspec_modifier] = STATE(6582),
    [sym_storage_class_specifier] = STATE(6582),
    [sym_type_qualifier] = STATE(6582),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6582),
    [aux_sym_declaration_repeat1] = STATE(3126),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7798),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3140)] = {
    [sym__declaration_modifiers] = STATE(3148),
    [sym_attribute_specifier] = STATE(3148),
    [sym_attribute_declaration] = STATE(3148),
    [sym_ms_declspec_modifier] = STATE(3148),
    [sym_storage_class_specifier] = STATE(3148),
    [sym_type_qualifier] = STATE(3148),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3148),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3148),
    [aux_sym_method_declaration_repeat2] = STATE(2121),
    [anon_sym_SEMI] = ACTIONS(7555),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3141)] = {
    [sym__declaration_modifiers] = STATE(6684),
    [sym_attribute_specifier] = STATE(6684),
    [sym_attribute_declaration] = STATE(6684),
    [sym_ms_declspec_modifier] = STATE(6684),
    [sym_storage_class_specifier] = STATE(6684),
    [sym_type_qualifier] = STATE(6684),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6684),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7800),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3142)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2105),
    [anon_sym_SEMI] = ACTIONS(7802),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3143)] = {
    [sym__declaration_modifiers] = STATE(6408),
    [sym_attribute_specifier] = STATE(6408),
    [sym_attribute_declaration] = STATE(6408),
    [sym_ms_declspec_modifier] = STATE(6408),
    [sym_storage_class_specifier] = STATE(6408),
    [sym_type_qualifier] = STATE(6408),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6408),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7804),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3144)] = {
    [sym_attribute_specifier] = STATE(2943),
    [sym_identifier] = ACTIONS(7806),
    [anon_sym_COMMA] = ACTIONS(7808),
    [anon_sym_RPAREN] = ACTIONS(7808),
    [anon_sym_LPAREN2] = ACTIONS(7808),
    [anon_sym_STAR] = ACTIONS(7808),
    [anon_sym_CARET] = ACTIONS(7808),
    [anon_sym_SEMI] = ACTIONS(7808),
    [anon_sym___extension__] = ACTIONS(7806),
    [anon_sym_extern] = ACTIONS(7806),
    [anon_sym___attribute__] = ACTIONS(7810),
    [anon_sym___attribute] = ACTIONS(7810),
    [anon_sym_noreturn] = ACTIONS(7806),
    [anon_sym_LBRACK] = ACTIONS(7808),
    [anon_sym___declspec] = ACTIONS(7806),
    [anon_sym___based] = ACTIONS(7806),
    [anon_sym_static] = ACTIONS(7806),
    [anon_sym_auto] = ACTIONS(7806),
    [anon_sym_register] = ACTIONS(7806),
    [anon_sym_inline] = ACTIONS(7806),
    [anon_sym___inline] = ACTIONS(7806),
    [anon_sym___inline__] = ACTIONS(7806),
    [anon_sym___forceinline] = ACTIONS(7806),
    [anon_sym_thread_local] = ACTIONS(7806),
    [anon_sym___thread] = ACTIONS(7806),
    [anon_sym_CG_EXTERN] = ACTIONS(7806),
    [anon_sym_CG_INLINE] = ACTIONS(7806),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7806),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7806),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7806),
    [anon_sym_IBOutlet] = ACTIONS(7806),
    [anon_sym_IBInspectable] = ACTIONS(7806),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7806),
    [anon_sym_NS_INLINE] = ACTIONS(7806),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7806),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7806),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7806),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7806),
    [anon_sym_const] = ACTIONS(7806),
    [anon_sym_constexpr] = ACTIONS(7806),
    [anon_sym_volatile] = ACTIONS(7806),
    [anon_sym_restrict] = ACTIONS(7806),
    [anon_sym___restrict__] = ACTIONS(7806),
    [anon_sym__Atomic] = ACTIONS(7806),
    [anon_sym__Noreturn] = ACTIONS(7806),
    [anon_sym__Nonnull] = ACTIONS(7806),
    [anon_sym_nullable] = ACTIONS(7806),
    [anon_sym__Complex] = ACTIONS(7806),
    [anon_sym__Nullable] = ACTIONS(7806),
    [anon_sym__Nullable_result] = ACTIONS(7806),
    [anon_sym__Null_unspecified] = ACTIONS(7806),
    [anon_sym___autoreleasing] = ACTIONS(7806),
    [anon_sym___block] = ACTIONS(7806),
    [anon_sym___bridge] = ACTIONS(7806),
    [anon_sym___bridge_retained] = ACTIONS(7806),
    [anon_sym___bridge_transfer] = ACTIONS(7806),
    [anon_sym___complex] = ACTIONS(7806),
    [anon_sym___const] = ACTIONS(7806),
    [anon_sym___imag] = ACTIONS(7806),
    [anon_sym___kindof] = ACTIONS(7806),
    [anon_sym___nonnull] = ACTIONS(7806),
    [anon_sym___nullable] = ACTIONS(7806),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7806),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7806),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7806),
    [anon_sym___real] = ACTIONS(7806),
    [anon_sym___strong] = ACTIONS(7806),
    [anon_sym___unsafe_unretained] = ACTIONS(7806),
    [anon_sym___unused] = ACTIONS(7806),
    [anon_sym___weak] = ACTIONS(7806),
    [anon_sym_alignas] = ACTIONS(7806),
    [anon_sym__Alignas] = ACTIONS(7806),
    [anon_sym_enum] = ACTIONS(7806),
    [anon_sym_COLON] = ACTIONS(7808),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7806),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7806),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7806),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7806),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7806),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7806),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7806),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7806),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7806),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7806),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7806),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7806),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7806),
    [anon_sym_API_AVAILABLE] = ACTIONS(7806),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7806),
    [anon_sym_API_DEPRECATED] = ACTIONS(7806),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7806),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7806),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7806),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7806),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7806),
    [anon_sym___deprecated_msg] = ACTIONS(7806),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7806),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7806),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7806),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7806),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7806),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7806),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7806),
  },
  [STATE(3145)] = {
    [sym__declaration_modifiers] = STATE(6074),
    [sym_attribute_specifier] = STATE(6074),
    [sym_attribute_declaration] = STATE(6074),
    [sym_ms_declspec_modifier] = STATE(6074),
    [sym_storage_class_specifier] = STATE(6074),
    [sym_type_qualifier] = STATE(6074),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6074),
    [aux_sym_declaration_repeat1] = STATE(3138),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7813),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3146)] = {
    [sym__declaration_modifiers] = STATE(6109),
    [sym_attribute_specifier] = STATE(6109),
    [sym_attribute_declaration] = STATE(6109),
    [sym_ms_declspec_modifier] = STATE(6109),
    [sym_storage_class_specifier] = STATE(6109),
    [sym_type_qualifier] = STATE(6109),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6109),
    [aux_sym_declaration_repeat1] = STATE(3165),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7815),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3147)] = {
    [sym_attribute_specifier] = STATE(2968),
    [sym_identifier] = ACTIONS(7817),
    [anon_sym_COMMA] = ACTIONS(7819),
    [anon_sym_RPAREN] = ACTIONS(7819),
    [anon_sym_LPAREN2] = ACTIONS(7819),
    [anon_sym_STAR] = ACTIONS(7819),
    [anon_sym_CARET] = ACTIONS(7819),
    [anon_sym_SEMI] = ACTIONS(7819),
    [anon_sym___extension__] = ACTIONS(7817),
    [anon_sym_extern] = ACTIONS(7817),
    [anon_sym___attribute__] = ACTIONS(7821),
    [anon_sym___attribute] = ACTIONS(7821),
    [anon_sym_noreturn] = ACTIONS(7817),
    [anon_sym_LBRACK] = ACTIONS(7819),
    [anon_sym___declspec] = ACTIONS(7817),
    [anon_sym___based] = ACTIONS(7817),
    [anon_sym_static] = ACTIONS(7817),
    [anon_sym_auto] = ACTIONS(7817),
    [anon_sym_register] = ACTIONS(7817),
    [anon_sym_inline] = ACTIONS(7817),
    [anon_sym___inline] = ACTIONS(7817),
    [anon_sym___inline__] = ACTIONS(7817),
    [anon_sym___forceinline] = ACTIONS(7817),
    [anon_sym_thread_local] = ACTIONS(7817),
    [anon_sym___thread] = ACTIONS(7817),
    [anon_sym_CG_EXTERN] = ACTIONS(7817),
    [anon_sym_CG_INLINE] = ACTIONS(7817),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7817),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7817),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7817),
    [anon_sym_IBOutlet] = ACTIONS(7817),
    [anon_sym_IBInspectable] = ACTIONS(7817),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7817),
    [anon_sym_NS_INLINE] = ACTIONS(7817),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7817),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7817),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7817),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7817),
    [anon_sym_const] = ACTIONS(7817),
    [anon_sym_constexpr] = ACTIONS(7817),
    [anon_sym_volatile] = ACTIONS(7817),
    [anon_sym_restrict] = ACTIONS(7817),
    [anon_sym___restrict__] = ACTIONS(7817),
    [anon_sym__Atomic] = ACTIONS(7817),
    [anon_sym__Noreturn] = ACTIONS(7817),
    [anon_sym__Nonnull] = ACTIONS(7817),
    [anon_sym_nullable] = ACTIONS(7817),
    [anon_sym__Complex] = ACTIONS(7817),
    [anon_sym__Nullable] = ACTIONS(7817),
    [anon_sym__Nullable_result] = ACTIONS(7817),
    [anon_sym__Null_unspecified] = ACTIONS(7817),
    [anon_sym___autoreleasing] = ACTIONS(7817),
    [anon_sym___block] = ACTIONS(7817),
    [anon_sym___bridge] = ACTIONS(7817),
    [anon_sym___bridge_retained] = ACTIONS(7817),
    [anon_sym___bridge_transfer] = ACTIONS(7817),
    [anon_sym___complex] = ACTIONS(7817),
    [anon_sym___const] = ACTIONS(7817),
    [anon_sym___imag] = ACTIONS(7817),
    [anon_sym___kindof] = ACTIONS(7817),
    [anon_sym___nonnull] = ACTIONS(7817),
    [anon_sym___nullable] = ACTIONS(7817),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7817),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7817),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7817),
    [anon_sym___real] = ACTIONS(7817),
    [anon_sym___strong] = ACTIONS(7817),
    [anon_sym___unsafe_unretained] = ACTIONS(7817),
    [anon_sym___unused] = ACTIONS(7817),
    [anon_sym___weak] = ACTIONS(7817),
    [anon_sym_alignas] = ACTIONS(7817),
    [anon_sym__Alignas] = ACTIONS(7817),
    [anon_sym_enum] = ACTIONS(7817),
    [anon_sym_COLON] = ACTIONS(7819),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7817),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7817),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7817),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7817),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7817),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7817),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7817),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7817),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7817),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7817),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7817),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7817),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7817),
    [anon_sym_API_AVAILABLE] = ACTIONS(7817),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7817),
    [anon_sym_API_DEPRECATED] = ACTIONS(7817),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7817),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7817),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7817),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7817),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7817),
    [anon_sym___deprecated_msg] = ACTIONS(7817),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7817),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7817),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7817),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7817),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7817),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7817),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7817),
  },
  [STATE(3148)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2100),
    [anon_sym_SEMI] = ACTIONS(7510),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3149)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2187),
    [anon_sym_SEMI] = ACTIONS(7504),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3150)] = {
    [sym__declaration_modifiers] = STATE(6514),
    [sym_attribute_specifier] = STATE(6514),
    [sym_attribute_declaration] = STATE(6514),
    [sym_ms_declspec_modifier] = STATE(6514),
    [sym_storage_class_specifier] = STATE(6514),
    [sym_type_qualifier] = STATE(6514),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6514),
    [aux_sym_declaration_repeat1] = STATE(3114),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7824),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3151)] = {
    [sym_attribute_specifier] = STATE(2996),
    [sym_identifier] = ACTIONS(7826),
    [anon_sym_COMMA] = ACTIONS(7828),
    [anon_sym_RPAREN] = ACTIONS(7828),
    [anon_sym_LPAREN2] = ACTIONS(7828),
    [anon_sym_STAR] = ACTIONS(7828),
    [anon_sym_CARET] = ACTIONS(7828),
    [anon_sym_SEMI] = ACTIONS(7828),
    [anon_sym___extension__] = ACTIONS(7826),
    [anon_sym_extern] = ACTIONS(7826),
    [anon_sym___attribute__] = ACTIONS(7830),
    [anon_sym___attribute] = ACTIONS(7830),
    [anon_sym_noreturn] = ACTIONS(7826),
    [anon_sym_LBRACK] = ACTIONS(7828),
    [anon_sym___declspec] = ACTIONS(7826),
    [anon_sym___based] = ACTIONS(7826),
    [anon_sym_static] = ACTIONS(7826),
    [anon_sym_auto] = ACTIONS(7826),
    [anon_sym_register] = ACTIONS(7826),
    [anon_sym_inline] = ACTIONS(7826),
    [anon_sym___inline] = ACTIONS(7826),
    [anon_sym___inline__] = ACTIONS(7826),
    [anon_sym___forceinline] = ACTIONS(7826),
    [anon_sym_thread_local] = ACTIONS(7826),
    [anon_sym___thread] = ACTIONS(7826),
    [anon_sym_CG_EXTERN] = ACTIONS(7826),
    [anon_sym_CG_INLINE] = ACTIONS(7826),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7826),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7826),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7826),
    [anon_sym_IBOutlet] = ACTIONS(7826),
    [anon_sym_IBInspectable] = ACTIONS(7826),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7826),
    [anon_sym_NS_INLINE] = ACTIONS(7826),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7826),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7826),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7826),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7826),
    [anon_sym_const] = ACTIONS(7826),
    [anon_sym_constexpr] = ACTIONS(7826),
    [anon_sym_volatile] = ACTIONS(7826),
    [anon_sym_restrict] = ACTIONS(7826),
    [anon_sym___restrict__] = ACTIONS(7826),
    [anon_sym__Atomic] = ACTIONS(7826),
    [anon_sym__Noreturn] = ACTIONS(7826),
    [anon_sym__Nonnull] = ACTIONS(7826),
    [anon_sym_nullable] = ACTIONS(7826),
    [anon_sym__Complex] = ACTIONS(7826),
    [anon_sym__Nullable] = ACTIONS(7826),
    [anon_sym__Nullable_result] = ACTIONS(7826),
    [anon_sym__Null_unspecified] = ACTIONS(7826),
    [anon_sym___autoreleasing] = ACTIONS(7826),
    [anon_sym___block] = ACTIONS(7826),
    [anon_sym___bridge] = ACTIONS(7826),
    [anon_sym___bridge_retained] = ACTIONS(7826),
    [anon_sym___bridge_transfer] = ACTIONS(7826),
    [anon_sym___complex] = ACTIONS(7826),
    [anon_sym___const] = ACTIONS(7826),
    [anon_sym___imag] = ACTIONS(7826),
    [anon_sym___kindof] = ACTIONS(7826),
    [anon_sym___nonnull] = ACTIONS(7826),
    [anon_sym___nullable] = ACTIONS(7826),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7826),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7826),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7826),
    [anon_sym___real] = ACTIONS(7826),
    [anon_sym___strong] = ACTIONS(7826),
    [anon_sym___unsafe_unretained] = ACTIONS(7826),
    [anon_sym___unused] = ACTIONS(7826),
    [anon_sym___weak] = ACTIONS(7826),
    [anon_sym_alignas] = ACTIONS(7826),
    [anon_sym__Alignas] = ACTIONS(7826),
    [anon_sym_enum] = ACTIONS(7826),
    [anon_sym_COLON] = ACTIONS(7828),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7826),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7826),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7826),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7826),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7826),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7826),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7826),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7826),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7826),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7826),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7826),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7826),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7826),
    [anon_sym_API_AVAILABLE] = ACTIONS(7826),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7826),
    [anon_sym_API_DEPRECATED] = ACTIONS(7826),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7826),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7826),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7826),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7826),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7826),
    [anon_sym___deprecated_msg] = ACTIONS(7826),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7826),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7826),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7826),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7826),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7826),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7826),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7826),
  },
  [STATE(3152)] = {
    [sym__declaration_modifiers] = STATE(6780),
    [sym_attribute_specifier] = STATE(6780),
    [sym_attribute_declaration] = STATE(6780),
    [sym_ms_declspec_modifier] = STATE(6780),
    [sym_storage_class_specifier] = STATE(6780),
    [sym_type_qualifier] = STATE(6780),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6780),
    [aux_sym_struct_declaration_repeat2] = STATE(3276),
    [anon_sym_COMMA] = ACTIONS(7654),
    [anon_sym_SEMI] = ACTIONS(7833),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3153)] = {
    [sym__declaration_modifiers] = STATE(6240),
    [sym_attribute_specifier] = STATE(6240),
    [sym_attribute_declaration] = STATE(6240),
    [sym_ms_declspec_modifier] = STATE(6240),
    [sym_storage_class_specifier] = STATE(6240),
    [sym_type_qualifier] = STATE(6240),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6240),
    [aux_sym_declaration_repeat1] = STATE(3173),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7835),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3154)] = {
    [sym__declaration_modifiers] = STATE(6126),
    [sym_attribute_specifier] = STATE(6126),
    [sym_attribute_declaration] = STATE(6126),
    [sym_ms_declspec_modifier] = STATE(6126),
    [sym_storage_class_specifier] = STATE(6126),
    [sym_type_qualifier] = STATE(6126),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6126),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7837),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3155)] = {
    [sym__declaration_modifiers] = STATE(6297),
    [sym_attribute_specifier] = STATE(6297),
    [sym_attribute_declaration] = STATE(6297),
    [sym_ms_declspec_modifier] = STATE(6297),
    [sym_storage_class_specifier] = STATE(6297),
    [sym_type_qualifier] = STATE(6297),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6297),
    [aux_sym__type_definition_declarators_repeat1] = STATE(3266),
    [anon_sym_COMMA] = ACTIONS(7542),
    [anon_sym_SEMI] = ACTIONS(7839),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3156)] = {
    [sym__declaration_modifiers] = STATE(6403),
    [sym_attribute_specifier] = STATE(6403),
    [sym_attribute_declaration] = STATE(6403),
    [sym_ms_declspec_modifier] = STATE(6403),
    [sym_storage_class_specifier] = STATE(6403),
    [sym_type_qualifier] = STATE(6403),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6403),
    [aux_sym_declaration_repeat1] = STATE(3169),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7841),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3157)] = {
    [sym_attribute_specifier] = STATE(2997),
    [sym_identifier] = ACTIONS(7843),
    [anon_sym_COMMA] = ACTIONS(7845),
    [anon_sym_RPAREN] = ACTIONS(7845),
    [anon_sym_LPAREN2] = ACTIONS(7845),
    [anon_sym_STAR] = ACTIONS(7845),
    [anon_sym_CARET] = ACTIONS(7845),
    [anon_sym_SEMI] = ACTIONS(7845),
    [anon_sym___extension__] = ACTIONS(7843),
    [anon_sym_extern] = ACTIONS(7843),
    [anon_sym___attribute__] = ACTIONS(7847),
    [anon_sym___attribute] = ACTIONS(7847),
    [anon_sym_noreturn] = ACTIONS(7843),
    [anon_sym_LBRACK] = ACTIONS(7845),
    [anon_sym___declspec] = ACTIONS(7843),
    [anon_sym___based] = ACTIONS(7843),
    [anon_sym_static] = ACTIONS(7843),
    [anon_sym_auto] = ACTIONS(7843),
    [anon_sym_register] = ACTIONS(7843),
    [anon_sym_inline] = ACTIONS(7843),
    [anon_sym___inline] = ACTIONS(7843),
    [anon_sym___inline__] = ACTIONS(7843),
    [anon_sym___forceinline] = ACTIONS(7843),
    [anon_sym_thread_local] = ACTIONS(7843),
    [anon_sym___thread] = ACTIONS(7843),
    [anon_sym_CG_EXTERN] = ACTIONS(7843),
    [anon_sym_CG_INLINE] = ACTIONS(7843),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7843),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7843),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7843),
    [anon_sym_IBOutlet] = ACTIONS(7843),
    [anon_sym_IBInspectable] = ACTIONS(7843),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7843),
    [anon_sym_NS_INLINE] = ACTIONS(7843),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7843),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7843),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7843),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7843),
    [anon_sym_const] = ACTIONS(7843),
    [anon_sym_constexpr] = ACTIONS(7843),
    [anon_sym_volatile] = ACTIONS(7843),
    [anon_sym_restrict] = ACTIONS(7843),
    [anon_sym___restrict__] = ACTIONS(7843),
    [anon_sym__Atomic] = ACTIONS(7843),
    [anon_sym__Noreturn] = ACTIONS(7843),
    [anon_sym__Nonnull] = ACTIONS(7843),
    [anon_sym_nullable] = ACTIONS(7843),
    [anon_sym__Complex] = ACTIONS(7843),
    [anon_sym__Nullable] = ACTIONS(7843),
    [anon_sym__Nullable_result] = ACTIONS(7843),
    [anon_sym__Null_unspecified] = ACTIONS(7843),
    [anon_sym___autoreleasing] = ACTIONS(7843),
    [anon_sym___block] = ACTIONS(7843),
    [anon_sym___bridge] = ACTIONS(7843),
    [anon_sym___bridge_retained] = ACTIONS(7843),
    [anon_sym___bridge_transfer] = ACTIONS(7843),
    [anon_sym___complex] = ACTIONS(7843),
    [anon_sym___const] = ACTIONS(7843),
    [anon_sym___imag] = ACTIONS(7843),
    [anon_sym___kindof] = ACTIONS(7843),
    [anon_sym___nonnull] = ACTIONS(7843),
    [anon_sym___nullable] = ACTIONS(7843),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7843),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7843),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7843),
    [anon_sym___real] = ACTIONS(7843),
    [anon_sym___strong] = ACTIONS(7843),
    [anon_sym___unsafe_unretained] = ACTIONS(7843),
    [anon_sym___unused] = ACTIONS(7843),
    [anon_sym___weak] = ACTIONS(7843),
    [anon_sym_alignas] = ACTIONS(7843),
    [anon_sym__Alignas] = ACTIONS(7843),
    [anon_sym_enum] = ACTIONS(7843),
    [anon_sym_COLON] = ACTIONS(7845),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7843),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7843),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7843),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7843),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7843),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7843),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7843),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7843),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7843),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7843),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7843),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7843),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7843),
    [anon_sym_API_AVAILABLE] = ACTIONS(7843),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7843),
    [anon_sym_API_DEPRECATED] = ACTIONS(7843),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7843),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7843),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7843),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7843),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7843),
    [anon_sym___deprecated_msg] = ACTIONS(7843),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7843),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7843),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7843),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7843),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7843),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7843),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7843),
  },
  [STATE(3158)] = {
    [sym_attribute_specifier] = STATE(3202),
    [sym_gnu_asm_expression] = STATE(3208),
    [aux_sym_function_declarator_repeat1] = STATE(3202),
    [anon_sym_COMMA] = ACTIONS(6473),
    [anon_sym_LPAREN2] = ACTIONS(6473),
    [anon_sym_SEMI] = ACTIONS(6473),
    [anon_sym___extension__] = ACTIONS(6473),
    [anon_sym_extern] = ACTIONS(6473),
    [anon_sym___attribute__] = ACTIONS(7850),
    [anon_sym___attribute] = ACTIONS(7852),
    [anon_sym_noreturn] = ACTIONS(6473),
    [anon_sym_LBRACK] = ACTIONS(6473),
    [anon_sym___declspec] = ACTIONS(6473),
    [anon_sym_static] = ACTIONS(6473),
    [anon_sym_EQ] = ACTIONS(6473),
    [anon_sym_auto] = ACTIONS(6473),
    [anon_sym_register] = ACTIONS(6473),
    [anon_sym_inline] = ACTIONS(6473),
    [anon_sym___inline] = ACTIONS(6471),
    [anon_sym___inline__] = ACTIONS(6473),
    [anon_sym___forceinline] = ACTIONS(6473),
    [anon_sym_thread_local] = ACTIONS(6473),
    [anon_sym___thread] = ACTIONS(6473),
    [anon_sym_CG_EXTERN] = ACTIONS(6473),
    [anon_sym_CG_INLINE] = ACTIONS(6473),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6473),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6473),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6473),
    [anon_sym_IBOutlet] = ACTIONS(6473),
    [anon_sym_IBInspectable] = ACTIONS(6473),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6473),
    [anon_sym_NS_INLINE] = ACTIONS(6473),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6473),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6473),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6473),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6473),
    [anon_sym_const] = ACTIONS(6471),
    [anon_sym_constexpr] = ACTIONS(6473),
    [anon_sym_volatile] = ACTIONS(6473),
    [anon_sym_restrict] = ACTIONS(6473),
    [anon_sym___restrict__] = ACTIONS(6473),
    [anon_sym__Atomic] = ACTIONS(6473),
    [anon_sym__Noreturn] = ACTIONS(6473),
    [anon_sym__Nonnull] = ACTIONS(6473),
    [anon_sym_nullable] = ACTIONS(6473),
    [anon_sym__Complex] = ACTIONS(6473),
    [anon_sym__Nullable] = ACTIONS(6471),
    [anon_sym__Nullable_result] = ACTIONS(6473),
    [anon_sym__Null_unspecified] = ACTIONS(6473),
    [anon_sym___autoreleasing] = ACTIONS(6473),
    [anon_sym___block] = ACTIONS(6473),
    [anon_sym___bridge] = ACTIONS(6471),
    [anon_sym___bridge_retained] = ACTIONS(6473),
    [anon_sym___bridge_transfer] = ACTIONS(6473),
    [anon_sym___complex] = ACTIONS(6473),
    [anon_sym___const] = ACTIONS(6473),
    [anon_sym___imag] = ACTIONS(6473),
    [anon_sym___kindof] = ACTIONS(6473),
    [anon_sym___nonnull] = ACTIONS(6473),
    [anon_sym___nullable] = ACTIONS(6473),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6473),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6473),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6473),
    [anon_sym___real] = ACTIONS(6473),
    [anon_sym___strong] = ACTIONS(6473),
    [anon_sym___unsafe_unretained] = ACTIONS(6473),
    [anon_sym___unused] = ACTIONS(6473),
    [anon_sym___weak] = ACTIONS(6473),
    [anon_sym_alignas] = ACTIONS(6473),
    [anon_sym__Alignas] = ACTIONS(6473),
    [anon_sym_in] = ACTIONS(6471),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6473),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6473),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6473),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6473),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6473),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6473),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6473),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6473),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6473),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6473),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6471),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6473),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6473),
    [anon_sym_API_AVAILABLE] = ACTIONS(6473),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6473),
    [anon_sym_API_DEPRECATED] = ACTIONS(6473),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6473),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6473),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6473),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6473),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6473),
    [anon_sym___deprecated_msg] = ACTIONS(6473),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6473),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6473),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6473),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6473),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6473),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6473),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6473),
  },
  [STATE(3159)] = {
    [sym_attribute_specifier] = STATE(3003),
    [sym_identifier] = ACTIONS(7854),
    [anon_sym_COMMA] = ACTIONS(7856),
    [anon_sym_RPAREN] = ACTIONS(7856),
    [anon_sym_LPAREN2] = ACTIONS(7856),
    [anon_sym_STAR] = ACTIONS(7856),
    [anon_sym_CARET] = ACTIONS(7856),
    [anon_sym_SEMI] = ACTIONS(7856),
    [anon_sym___extension__] = ACTIONS(7854),
    [anon_sym_extern] = ACTIONS(7854),
    [anon_sym___attribute__] = ACTIONS(7858),
    [anon_sym___attribute] = ACTIONS(7858),
    [anon_sym_noreturn] = ACTIONS(7854),
    [anon_sym_LBRACK] = ACTIONS(7856),
    [anon_sym___declspec] = ACTIONS(7854),
    [anon_sym___based] = ACTIONS(7854),
    [anon_sym_static] = ACTIONS(7854),
    [anon_sym_auto] = ACTIONS(7854),
    [anon_sym_register] = ACTIONS(7854),
    [anon_sym_inline] = ACTIONS(7854),
    [anon_sym___inline] = ACTIONS(7854),
    [anon_sym___inline__] = ACTIONS(7854),
    [anon_sym___forceinline] = ACTIONS(7854),
    [anon_sym_thread_local] = ACTIONS(7854),
    [anon_sym___thread] = ACTIONS(7854),
    [anon_sym_CG_EXTERN] = ACTIONS(7854),
    [anon_sym_CG_INLINE] = ACTIONS(7854),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7854),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7854),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7854),
    [anon_sym_IBOutlet] = ACTIONS(7854),
    [anon_sym_IBInspectable] = ACTIONS(7854),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7854),
    [anon_sym_NS_INLINE] = ACTIONS(7854),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7854),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7854),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7854),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7854),
    [anon_sym_const] = ACTIONS(7854),
    [anon_sym_constexpr] = ACTIONS(7854),
    [anon_sym_volatile] = ACTIONS(7854),
    [anon_sym_restrict] = ACTIONS(7854),
    [anon_sym___restrict__] = ACTIONS(7854),
    [anon_sym__Atomic] = ACTIONS(7854),
    [anon_sym__Noreturn] = ACTIONS(7854),
    [anon_sym__Nonnull] = ACTIONS(7854),
    [anon_sym_nullable] = ACTIONS(7854),
    [anon_sym__Complex] = ACTIONS(7854),
    [anon_sym__Nullable] = ACTIONS(7854),
    [anon_sym__Nullable_result] = ACTIONS(7854),
    [anon_sym__Null_unspecified] = ACTIONS(7854),
    [anon_sym___autoreleasing] = ACTIONS(7854),
    [anon_sym___block] = ACTIONS(7854),
    [anon_sym___bridge] = ACTIONS(7854),
    [anon_sym___bridge_retained] = ACTIONS(7854),
    [anon_sym___bridge_transfer] = ACTIONS(7854),
    [anon_sym___complex] = ACTIONS(7854),
    [anon_sym___const] = ACTIONS(7854),
    [anon_sym___imag] = ACTIONS(7854),
    [anon_sym___kindof] = ACTIONS(7854),
    [anon_sym___nonnull] = ACTIONS(7854),
    [anon_sym___nullable] = ACTIONS(7854),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7854),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7854),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7854),
    [anon_sym___real] = ACTIONS(7854),
    [anon_sym___strong] = ACTIONS(7854),
    [anon_sym___unsafe_unretained] = ACTIONS(7854),
    [anon_sym___unused] = ACTIONS(7854),
    [anon_sym___weak] = ACTIONS(7854),
    [anon_sym_alignas] = ACTIONS(7854),
    [anon_sym__Alignas] = ACTIONS(7854),
    [anon_sym_enum] = ACTIONS(7854),
    [anon_sym_COLON] = ACTIONS(7856),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7854),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7854),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7854),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7854),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7854),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7854),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7854),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7854),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7854),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7854),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7854),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7854),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7854),
    [anon_sym_API_AVAILABLE] = ACTIONS(7854),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7854),
    [anon_sym_API_DEPRECATED] = ACTIONS(7854),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7854),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7854),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7854),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7854),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7854),
    [anon_sym___deprecated_msg] = ACTIONS(7854),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7854),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7854),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7854),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7854),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7854),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7854),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7854),
  },
  [STATE(3160)] = {
    [sym__declaration_modifiers] = STATE(6547),
    [sym_attribute_specifier] = STATE(6547),
    [sym_attribute_declaration] = STATE(6547),
    [sym_ms_declspec_modifier] = STATE(6547),
    [sym_storage_class_specifier] = STATE(6547),
    [sym_type_qualifier] = STATE(6547),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6547),
    [aux_sym_declaration_repeat1] = STATE(3105),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7861),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3161)] = {
    [sym__declaration_modifiers] = STATE(6549),
    [sym_attribute_specifier] = STATE(6549),
    [sym_attribute_declaration] = STATE(6549),
    [sym_ms_declspec_modifier] = STATE(6549),
    [sym_storage_class_specifier] = STATE(6549),
    [sym_type_qualifier] = STATE(6549),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6549),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7863),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3162)] = {
    [sym__declaration_modifiers] = STATE(6207),
    [sym_attribute_specifier] = STATE(6207),
    [sym_attribute_declaration] = STATE(6207),
    [sym_ms_declspec_modifier] = STATE(6207),
    [sym_storage_class_specifier] = STATE(6207),
    [sym_type_qualifier] = STATE(6207),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6207),
    [aux_sym_declaration_repeat1] = STATE(3182),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7865),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3163)] = {
    [sym__declaration_modifiers] = STATE(6211),
    [sym_attribute_specifier] = STATE(6211),
    [sym_attribute_declaration] = STATE(6211),
    [sym_ms_declspec_modifier] = STATE(6211),
    [sym_storage_class_specifier] = STATE(6211),
    [sym_type_qualifier] = STATE(6211),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6211),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7867),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3164)] = {
    [sym__declaration_modifiers] = STATE(6422),
    [sym_attribute_specifier] = STATE(6422),
    [sym_attribute_declaration] = STATE(6422),
    [sym_ms_declspec_modifier] = STATE(6422),
    [sym_storage_class_specifier] = STATE(6422),
    [sym_type_qualifier] = STATE(6422),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6422),
    [aux_sym_declaration_repeat1] = STATE(3197),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7869),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3165)] = {
    [sym__declaration_modifiers] = STATE(6213),
    [sym_attribute_specifier] = STATE(6213),
    [sym_attribute_declaration] = STATE(6213),
    [sym_ms_declspec_modifier] = STATE(6213),
    [sym_storage_class_specifier] = STATE(6213),
    [sym_type_qualifier] = STATE(6213),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6213),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7871),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3166)] = {
    [sym__declaration_modifiers] = STATE(6214),
    [sym_attribute_specifier] = STATE(6214),
    [sym_attribute_declaration] = STATE(6214),
    [sym_ms_declspec_modifier] = STATE(6214),
    [sym_storage_class_specifier] = STATE(6214),
    [sym_type_qualifier] = STATE(6214),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6214),
    [aux_sym_declaration_repeat1] = STATE(3183),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7873),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3167)] = {
    [sym_attribute_specifier] = STATE(2953),
    [sym_identifier] = ACTIONS(7875),
    [anon_sym_COMMA] = ACTIONS(7877),
    [anon_sym_RPAREN] = ACTIONS(7877),
    [anon_sym_LPAREN2] = ACTIONS(7877),
    [anon_sym_STAR] = ACTIONS(7877),
    [anon_sym_CARET] = ACTIONS(7877),
    [anon_sym_SEMI] = ACTIONS(7877),
    [anon_sym___extension__] = ACTIONS(7875),
    [anon_sym_extern] = ACTIONS(7875),
    [anon_sym___attribute__] = ACTIONS(7879),
    [anon_sym___attribute] = ACTIONS(7879),
    [anon_sym_noreturn] = ACTIONS(7875),
    [anon_sym_LBRACK] = ACTIONS(7877),
    [anon_sym___declspec] = ACTIONS(7875),
    [anon_sym___based] = ACTIONS(7875),
    [anon_sym_static] = ACTIONS(7875),
    [anon_sym_auto] = ACTIONS(7875),
    [anon_sym_register] = ACTIONS(7875),
    [anon_sym_inline] = ACTIONS(7875),
    [anon_sym___inline] = ACTIONS(7875),
    [anon_sym___inline__] = ACTIONS(7875),
    [anon_sym___forceinline] = ACTIONS(7875),
    [anon_sym_thread_local] = ACTIONS(7875),
    [anon_sym___thread] = ACTIONS(7875),
    [anon_sym_CG_EXTERN] = ACTIONS(7875),
    [anon_sym_CG_INLINE] = ACTIONS(7875),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7875),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7875),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7875),
    [anon_sym_IBOutlet] = ACTIONS(7875),
    [anon_sym_IBInspectable] = ACTIONS(7875),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7875),
    [anon_sym_NS_INLINE] = ACTIONS(7875),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7875),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7875),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7875),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7875),
    [anon_sym_const] = ACTIONS(7875),
    [anon_sym_constexpr] = ACTIONS(7875),
    [anon_sym_volatile] = ACTIONS(7875),
    [anon_sym_restrict] = ACTIONS(7875),
    [anon_sym___restrict__] = ACTIONS(7875),
    [anon_sym__Atomic] = ACTIONS(7875),
    [anon_sym__Noreturn] = ACTIONS(7875),
    [anon_sym__Nonnull] = ACTIONS(7875),
    [anon_sym_nullable] = ACTIONS(7875),
    [anon_sym__Complex] = ACTIONS(7875),
    [anon_sym__Nullable] = ACTIONS(7875),
    [anon_sym__Nullable_result] = ACTIONS(7875),
    [anon_sym__Null_unspecified] = ACTIONS(7875),
    [anon_sym___autoreleasing] = ACTIONS(7875),
    [anon_sym___block] = ACTIONS(7875),
    [anon_sym___bridge] = ACTIONS(7875),
    [anon_sym___bridge_retained] = ACTIONS(7875),
    [anon_sym___bridge_transfer] = ACTIONS(7875),
    [anon_sym___complex] = ACTIONS(7875),
    [anon_sym___const] = ACTIONS(7875),
    [anon_sym___imag] = ACTIONS(7875),
    [anon_sym___kindof] = ACTIONS(7875),
    [anon_sym___nonnull] = ACTIONS(7875),
    [anon_sym___nullable] = ACTIONS(7875),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7875),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7875),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7875),
    [anon_sym___real] = ACTIONS(7875),
    [anon_sym___strong] = ACTIONS(7875),
    [anon_sym___unsafe_unretained] = ACTIONS(7875),
    [anon_sym___unused] = ACTIONS(7875),
    [anon_sym___weak] = ACTIONS(7875),
    [anon_sym_alignas] = ACTIONS(7875),
    [anon_sym__Alignas] = ACTIONS(7875),
    [anon_sym_enum] = ACTIONS(7875),
    [anon_sym_COLON] = ACTIONS(7877),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7875),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7875),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7875),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7875),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7875),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7875),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7875),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7875),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7875),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7875),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7875),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7875),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7875),
    [anon_sym_API_AVAILABLE] = ACTIONS(7875),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7875),
    [anon_sym_API_DEPRECATED] = ACTIONS(7875),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7875),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7875),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7875),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7875),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7875),
    [anon_sym___deprecated_msg] = ACTIONS(7875),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7875),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7875),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7875),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7875),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7875),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7875),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7875),
  },
  [STATE(3168)] = {
    [sym__declaration_modifiers] = STATE(5902),
    [sym_attribute_specifier] = STATE(5902),
    [sym_attribute_declaration] = STATE(5902),
    [sym_ms_declspec_modifier] = STATE(5902),
    [sym_storage_class_specifier] = STATE(5902),
    [sym_type_qualifier] = STATE(5902),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5902),
    [aux_sym_declaration_repeat1] = STATE(3186),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7882),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3169)] = {
    [sym__declaration_modifiers] = STATE(6550),
    [sym_attribute_specifier] = STATE(6550),
    [sym_attribute_declaration] = STATE(6550),
    [sym_ms_declspec_modifier] = STATE(6550),
    [sym_storage_class_specifier] = STATE(6550),
    [sym_type_qualifier] = STATE(6550),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6550),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7884),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3170)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2137),
    [anon_sym_SEMI] = ACTIONS(7886),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3171)] = {
    [sym__declaration_modifiers] = STATE(6551),
    [sym_attribute_specifier] = STATE(6551),
    [sym_attribute_declaration] = STATE(6551),
    [sym_ms_declspec_modifier] = STATE(6551),
    [sym_storage_class_specifier] = STATE(6551),
    [sym_type_qualifier] = STATE(6551),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6551),
    [aux_sym_declaration_repeat1] = STATE(3116),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7888),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3172)] = {
    [sym__declaration_modifiers] = STATE(5908),
    [sym_attribute_specifier] = STATE(5908),
    [sym_attribute_declaration] = STATE(5908),
    [sym_ms_declspec_modifier] = STATE(5908),
    [sym_storage_class_specifier] = STATE(5908),
    [sym_type_qualifier] = STATE(5908),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5908),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7890),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3173)] = {
    [sym__declaration_modifiers] = STATE(5912),
    [sym_attribute_specifier] = STATE(5912),
    [sym_attribute_declaration] = STATE(5912),
    [sym_ms_declspec_modifier] = STATE(5912),
    [sym_storage_class_specifier] = STATE(5912),
    [sym_type_qualifier] = STATE(5912),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5912),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7892),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3174)] = {
    [sym__declaration_modifiers] = STATE(5922),
    [sym_attribute_specifier] = STATE(5922),
    [sym_attribute_declaration] = STATE(5922),
    [sym_ms_declspec_modifier] = STATE(5922),
    [sym_storage_class_specifier] = STATE(5922),
    [sym_type_qualifier] = STATE(5922),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(5922),
    [aux_sym_declaration_repeat1] = STATE(3200),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7894),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3175)] = {
    [sym__declaration_modifiers] = STATE(3184),
    [sym_attribute_specifier] = STATE(3184),
    [sym_attribute_declaration] = STATE(3184),
    [sym_ms_declspec_modifier] = STATE(3184),
    [sym_storage_class_specifier] = STATE(3184),
    [sym_type_qualifier] = STATE(3184),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3184),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3184),
    [aux_sym_method_declaration_repeat2] = STATE(2167),
    [anon_sym_SEMI] = ACTIONS(7498),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3176)] = {
    [sym__declaration_modifiers] = STATE(3179),
    [sym_attribute_specifier] = STATE(3179),
    [sym_attribute_declaration] = STATE(3179),
    [sym_ms_declspec_modifier] = STATE(3179),
    [sym_storage_class_specifier] = STATE(3179),
    [sym_type_qualifier] = STATE(3179),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3179),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3179),
    [aux_sym_method_declaration_repeat2] = STATE(2183),
    [anon_sym_SEMI] = ACTIONS(7500),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3177)] = {
    [sym__declaration_modifiers] = STATE(6558),
    [sym_attribute_specifier] = STATE(6558),
    [sym_attribute_declaration] = STATE(6558),
    [sym_ms_declspec_modifier] = STATE(6558),
    [sym_storage_class_specifier] = STATE(6558),
    [sym_type_qualifier] = STATE(6558),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6558),
    [aux_sym_declaration_repeat1] = STATE(3143),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7896),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3178)] = {
    [sym_attribute_specifier] = STATE(2931),
    [sym_identifier] = ACTIONS(7898),
    [anon_sym_COMMA] = ACTIONS(7900),
    [anon_sym_RPAREN] = ACTIONS(7900),
    [anon_sym_LPAREN2] = ACTIONS(7900),
    [anon_sym_STAR] = ACTIONS(7900),
    [anon_sym_CARET] = ACTIONS(7900),
    [anon_sym_SEMI] = ACTIONS(7900),
    [anon_sym___extension__] = ACTIONS(7898),
    [anon_sym_extern] = ACTIONS(7898),
    [anon_sym___attribute__] = ACTIONS(7902),
    [anon_sym___attribute] = ACTIONS(7902),
    [anon_sym_noreturn] = ACTIONS(7898),
    [anon_sym_LBRACK] = ACTIONS(7900),
    [anon_sym___declspec] = ACTIONS(7898),
    [anon_sym___based] = ACTIONS(7898),
    [anon_sym_static] = ACTIONS(7898),
    [anon_sym_auto] = ACTIONS(7898),
    [anon_sym_register] = ACTIONS(7898),
    [anon_sym_inline] = ACTIONS(7898),
    [anon_sym___inline] = ACTIONS(7898),
    [anon_sym___inline__] = ACTIONS(7898),
    [anon_sym___forceinline] = ACTIONS(7898),
    [anon_sym_thread_local] = ACTIONS(7898),
    [anon_sym___thread] = ACTIONS(7898),
    [anon_sym_CG_EXTERN] = ACTIONS(7898),
    [anon_sym_CG_INLINE] = ACTIONS(7898),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7898),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7898),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7898),
    [anon_sym_IBOutlet] = ACTIONS(7898),
    [anon_sym_IBInspectable] = ACTIONS(7898),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7898),
    [anon_sym_NS_INLINE] = ACTIONS(7898),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7898),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7898),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7898),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7898),
    [anon_sym_const] = ACTIONS(7898),
    [anon_sym_constexpr] = ACTIONS(7898),
    [anon_sym_volatile] = ACTIONS(7898),
    [anon_sym_restrict] = ACTIONS(7898),
    [anon_sym___restrict__] = ACTIONS(7898),
    [anon_sym__Atomic] = ACTIONS(7898),
    [anon_sym__Noreturn] = ACTIONS(7898),
    [anon_sym__Nonnull] = ACTIONS(7898),
    [anon_sym_nullable] = ACTIONS(7898),
    [anon_sym__Complex] = ACTIONS(7898),
    [anon_sym__Nullable] = ACTIONS(7898),
    [anon_sym__Nullable_result] = ACTIONS(7898),
    [anon_sym__Null_unspecified] = ACTIONS(7898),
    [anon_sym___autoreleasing] = ACTIONS(7898),
    [anon_sym___block] = ACTIONS(7898),
    [anon_sym___bridge] = ACTIONS(7898),
    [anon_sym___bridge_retained] = ACTIONS(7898),
    [anon_sym___bridge_transfer] = ACTIONS(7898),
    [anon_sym___complex] = ACTIONS(7898),
    [anon_sym___const] = ACTIONS(7898),
    [anon_sym___imag] = ACTIONS(7898),
    [anon_sym___kindof] = ACTIONS(7898),
    [anon_sym___nonnull] = ACTIONS(7898),
    [anon_sym___nullable] = ACTIONS(7898),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7898),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7898),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7898),
    [anon_sym___real] = ACTIONS(7898),
    [anon_sym___strong] = ACTIONS(7898),
    [anon_sym___unsafe_unretained] = ACTIONS(7898),
    [anon_sym___unused] = ACTIONS(7898),
    [anon_sym___weak] = ACTIONS(7898),
    [anon_sym_alignas] = ACTIONS(7898),
    [anon_sym__Alignas] = ACTIONS(7898),
    [anon_sym_enum] = ACTIONS(7898),
    [anon_sym_COLON] = ACTIONS(7900),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7898),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7898),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7898),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7898),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7898),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7898),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7898),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7898),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7898),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7898),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7898),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7898),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7898),
    [anon_sym_API_AVAILABLE] = ACTIONS(7898),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7898),
    [anon_sym_API_DEPRECATED] = ACTIONS(7898),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7898),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7898),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7898),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7898),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7898),
    [anon_sym___deprecated_msg] = ACTIONS(7898),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7898),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7898),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7898),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7898),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7898),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7898),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7898),
  },
  [STATE(3179)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2200),
    [anon_sym_SEMI] = ACTIONS(7533),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3180)] = {
    [sym__declaration_modifiers] = STATE(3131),
    [sym_attribute_specifier] = STATE(3131),
    [sym_attribute_declaration] = STATE(3131),
    [sym_ms_declspec_modifier] = STATE(3131),
    [sym_storage_class_specifier] = STATE(3131),
    [sym_type_qualifier] = STATE(3131),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3131),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3131),
    [aux_sym_method_declaration_repeat2] = STATE(2191),
    [anon_sym_SEMI] = ACTIONS(7531),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3181)] = {
    [sym_attribute_specifier] = STATE(2955),
    [sym_identifier] = ACTIONS(7905),
    [anon_sym_COMMA] = ACTIONS(7907),
    [anon_sym_RPAREN] = ACTIONS(7907),
    [anon_sym_LPAREN2] = ACTIONS(7907),
    [anon_sym_STAR] = ACTIONS(7907),
    [anon_sym_CARET] = ACTIONS(7907),
    [anon_sym_SEMI] = ACTIONS(7907),
    [anon_sym___extension__] = ACTIONS(7905),
    [anon_sym_extern] = ACTIONS(7905),
    [anon_sym___attribute__] = ACTIONS(7909),
    [anon_sym___attribute] = ACTIONS(7909),
    [anon_sym_noreturn] = ACTIONS(7905),
    [anon_sym_LBRACK] = ACTIONS(7907),
    [anon_sym___declspec] = ACTIONS(7905),
    [anon_sym___based] = ACTIONS(7905),
    [anon_sym_static] = ACTIONS(7905),
    [anon_sym_auto] = ACTIONS(7905),
    [anon_sym_register] = ACTIONS(7905),
    [anon_sym_inline] = ACTIONS(7905),
    [anon_sym___inline] = ACTIONS(7905),
    [anon_sym___inline__] = ACTIONS(7905),
    [anon_sym___forceinline] = ACTIONS(7905),
    [anon_sym_thread_local] = ACTIONS(7905),
    [anon_sym___thread] = ACTIONS(7905),
    [anon_sym_CG_EXTERN] = ACTIONS(7905),
    [anon_sym_CG_INLINE] = ACTIONS(7905),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7905),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7905),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7905),
    [anon_sym_IBOutlet] = ACTIONS(7905),
    [anon_sym_IBInspectable] = ACTIONS(7905),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7905),
    [anon_sym_NS_INLINE] = ACTIONS(7905),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7905),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7905),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7905),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7905),
    [anon_sym_const] = ACTIONS(7905),
    [anon_sym_constexpr] = ACTIONS(7905),
    [anon_sym_volatile] = ACTIONS(7905),
    [anon_sym_restrict] = ACTIONS(7905),
    [anon_sym___restrict__] = ACTIONS(7905),
    [anon_sym__Atomic] = ACTIONS(7905),
    [anon_sym__Noreturn] = ACTIONS(7905),
    [anon_sym__Nonnull] = ACTIONS(7905),
    [anon_sym_nullable] = ACTIONS(7905),
    [anon_sym__Complex] = ACTIONS(7905),
    [anon_sym__Nullable] = ACTIONS(7905),
    [anon_sym__Nullable_result] = ACTIONS(7905),
    [anon_sym__Null_unspecified] = ACTIONS(7905),
    [anon_sym___autoreleasing] = ACTIONS(7905),
    [anon_sym___block] = ACTIONS(7905),
    [anon_sym___bridge] = ACTIONS(7905),
    [anon_sym___bridge_retained] = ACTIONS(7905),
    [anon_sym___bridge_transfer] = ACTIONS(7905),
    [anon_sym___complex] = ACTIONS(7905),
    [anon_sym___const] = ACTIONS(7905),
    [anon_sym___imag] = ACTIONS(7905),
    [anon_sym___kindof] = ACTIONS(7905),
    [anon_sym___nonnull] = ACTIONS(7905),
    [anon_sym___nullable] = ACTIONS(7905),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7905),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7905),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7905),
    [anon_sym___real] = ACTIONS(7905),
    [anon_sym___strong] = ACTIONS(7905),
    [anon_sym___unsafe_unretained] = ACTIONS(7905),
    [anon_sym___unused] = ACTIONS(7905),
    [anon_sym___weak] = ACTIONS(7905),
    [anon_sym_alignas] = ACTIONS(7905),
    [anon_sym__Alignas] = ACTIONS(7905),
    [anon_sym_enum] = ACTIONS(7905),
    [anon_sym_COLON] = ACTIONS(7907),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7905),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7905),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7905),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7905),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7905),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7905),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7905),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7905),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7905),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7905),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7905),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7905),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7905),
    [anon_sym_API_AVAILABLE] = ACTIONS(7905),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7905),
    [anon_sym_API_DEPRECATED] = ACTIONS(7905),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7905),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7905),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7905),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7905),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7905),
    [anon_sym___deprecated_msg] = ACTIONS(7905),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7905),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7905),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7905),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7905),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7905),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7905),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7905),
  },
  [STATE(3182)] = {
    [sym__declaration_modifiers] = STATE(6276),
    [sym_attribute_specifier] = STATE(6276),
    [sym_attribute_declaration] = STATE(6276),
    [sym_ms_declspec_modifier] = STATE(6276),
    [sym_storage_class_specifier] = STATE(6276),
    [sym_type_qualifier] = STATE(6276),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6276),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7912),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3183)] = {
    [sym__declaration_modifiers] = STATE(6278),
    [sym_attribute_specifier] = STATE(6278),
    [sym_attribute_declaration] = STATE(6278),
    [sym_ms_declspec_modifier] = STATE(6278),
    [sym_storage_class_specifier] = STATE(6278),
    [sym_type_qualifier] = STATE(6278),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6278),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7914),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3184)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2191),
    [anon_sym_SEMI] = ACTIONS(7531),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3185)] = {
    [sym__declaration_modifiers] = STATE(6279),
    [sym_attribute_specifier] = STATE(6279),
    [sym_attribute_declaration] = STATE(6279),
    [sym_ms_declspec_modifier] = STATE(6279),
    [sym_storage_class_specifier] = STATE(6279),
    [sym_type_qualifier] = STATE(6279),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6279),
    [aux_sym_struct_declaration_repeat2] = STATE(3194),
    [anon_sym_COMMA] = ACTIONS(7654),
    [anon_sym_SEMI] = ACTIONS(7916),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3186)] = {
    [sym__declaration_modifiers] = STATE(6039),
    [sym_attribute_specifier] = STATE(6039),
    [sym_attribute_declaration] = STATE(6039),
    [sym_ms_declspec_modifier] = STATE(6039),
    [sym_storage_class_specifier] = STATE(6039),
    [sym_type_qualifier] = STATE(6039),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6039),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7918),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3187)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2314),
    [anon_sym_SEMI] = ACTIONS(7575),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3188)] = {
    [sym__declaration_modifiers] = STATE(3196),
    [sym_attribute_specifier] = STATE(3196),
    [sym_attribute_declaration] = STATE(3196),
    [sym_ms_declspec_modifier] = STATE(3196),
    [sym_storage_class_specifier] = STATE(3196),
    [sym_type_qualifier] = STATE(3196),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3196),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3196),
    [aux_sym_method_declaration_repeat2] = STATE(2200),
    [anon_sym_SEMI] = ACTIONS(7533),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3189)] = {
    [sym_attribute_specifier] = STATE(2956),
    [sym_identifier] = ACTIONS(7920),
    [anon_sym_COMMA] = ACTIONS(7922),
    [anon_sym_RPAREN] = ACTIONS(7922),
    [anon_sym_LPAREN2] = ACTIONS(7922),
    [anon_sym_STAR] = ACTIONS(7922),
    [anon_sym_CARET] = ACTIONS(7922),
    [anon_sym_SEMI] = ACTIONS(7922),
    [anon_sym___extension__] = ACTIONS(7920),
    [anon_sym_extern] = ACTIONS(7920),
    [anon_sym___attribute__] = ACTIONS(7924),
    [anon_sym___attribute] = ACTIONS(7924),
    [anon_sym_noreturn] = ACTIONS(7920),
    [anon_sym_LBRACK] = ACTIONS(7922),
    [anon_sym___declspec] = ACTIONS(7920),
    [anon_sym___based] = ACTIONS(7920),
    [anon_sym_static] = ACTIONS(7920),
    [anon_sym_auto] = ACTIONS(7920),
    [anon_sym_register] = ACTIONS(7920),
    [anon_sym_inline] = ACTIONS(7920),
    [anon_sym___inline] = ACTIONS(7920),
    [anon_sym___inline__] = ACTIONS(7920),
    [anon_sym___forceinline] = ACTIONS(7920),
    [anon_sym_thread_local] = ACTIONS(7920),
    [anon_sym___thread] = ACTIONS(7920),
    [anon_sym_CG_EXTERN] = ACTIONS(7920),
    [anon_sym_CG_INLINE] = ACTIONS(7920),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7920),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7920),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7920),
    [anon_sym_IBOutlet] = ACTIONS(7920),
    [anon_sym_IBInspectable] = ACTIONS(7920),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7920),
    [anon_sym_NS_INLINE] = ACTIONS(7920),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7920),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7920),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7920),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7920),
    [anon_sym_const] = ACTIONS(7920),
    [anon_sym_constexpr] = ACTIONS(7920),
    [anon_sym_volatile] = ACTIONS(7920),
    [anon_sym_restrict] = ACTIONS(7920),
    [anon_sym___restrict__] = ACTIONS(7920),
    [anon_sym__Atomic] = ACTIONS(7920),
    [anon_sym__Noreturn] = ACTIONS(7920),
    [anon_sym__Nonnull] = ACTIONS(7920),
    [anon_sym_nullable] = ACTIONS(7920),
    [anon_sym__Complex] = ACTIONS(7920),
    [anon_sym__Nullable] = ACTIONS(7920),
    [anon_sym__Nullable_result] = ACTIONS(7920),
    [anon_sym__Null_unspecified] = ACTIONS(7920),
    [anon_sym___autoreleasing] = ACTIONS(7920),
    [anon_sym___block] = ACTIONS(7920),
    [anon_sym___bridge] = ACTIONS(7920),
    [anon_sym___bridge_retained] = ACTIONS(7920),
    [anon_sym___bridge_transfer] = ACTIONS(7920),
    [anon_sym___complex] = ACTIONS(7920),
    [anon_sym___const] = ACTIONS(7920),
    [anon_sym___imag] = ACTIONS(7920),
    [anon_sym___kindof] = ACTIONS(7920),
    [anon_sym___nonnull] = ACTIONS(7920),
    [anon_sym___nullable] = ACTIONS(7920),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7920),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7920),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7920),
    [anon_sym___real] = ACTIONS(7920),
    [anon_sym___strong] = ACTIONS(7920),
    [anon_sym___unsafe_unretained] = ACTIONS(7920),
    [anon_sym___unused] = ACTIONS(7920),
    [anon_sym___weak] = ACTIONS(7920),
    [anon_sym_alignas] = ACTIONS(7920),
    [anon_sym__Alignas] = ACTIONS(7920),
    [anon_sym_enum] = ACTIONS(7920),
    [anon_sym_COLON] = ACTIONS(7922),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7920),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7920),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7920),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7920),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7920),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7920),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7920),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7920),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7920),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7920),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7920),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7920),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7920),
    [anon_sym_API_AVAILABLE] = ACTIONS(7920),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7920),
    [anon_sym_API_DEPRECATED] = ACTIONS(7920),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7920),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7920),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7920),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7920),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7920),
    [anon_sym___deprecated_msg] = ACTIONS(7920),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7920),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7920),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7920),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7920),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7920),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7920),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7920),
  },
  [STATE(3190)] = {
    [sym__declaration_modifiers] = STATE(3127),
    [sym_attribute_specifier] = STATE(3127),
    [sym_attribute_declaration] = STATE(3127),
    [sym_ms_declspec_modifier] = STATE(3127),
    [sym_storage_class_specifier] = STATE(3127),
    [sym_type_qualifier] = STATE(3127),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3127),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3127),
    [aux_sym_method_declaration_repeat2] = STATE(2314),
    [anon_sym_SEMI] = ACTIONS(7575),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3191)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2270),
    [anon_sym_SEMI] = ACTIONS(7496),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3192)] = {
    [sym_attribute_specifier] = STATE(2976),
    [sym_identifier] = ACTIONS(7927),
    [anon_sym_COMMA] = ACTIONS(7929),
    [anon_sym_RPAREN] = ACTIONS(7929),
    [anon_sym_LPAREN2] = ACTIONS(7929),
    [anon_sym_STAR] = ACTIONS(7929),
    [anon_sym_CARET] = ACTIONS(7929),
    [anon_sym_SEMI] = ACTIONS(7929),
    [anon_sym___extension__] = ACTIONS(7927),
    [anon_sym_extern] = ACTIONS(7927),
    [anon_sym___attribute__] = ACTIONS(7931),
    [anon_sym___attribute] = ACTIONS(7931),
    [anon_sym_noreturn] = ACTIONS(7927),
    [anon_sym_LBRACK] = ACTIONS(7929),
    [anon_sym___declspec] = ACTIONS(7927),
    [anon_sym___based] = ACTIONS(7927),
    [anon_sym_static] = ACTIONS(7927),
    [anon_sym_auto] = ACTIONS(7927),
    [anon_sym_register] = ACTIONS(7927),
    [anon_sym_inline] = ACTIONS(7927),
    [anon_sym___inline] = ACTIONS(7927),
    [anon_sym___inline__] = ACTIONS(7927),
    [anon_sym___forceinline] = ACTIONS(7927),
    [anon_sym_thread_local] = ACTIONS(7927),
    [anon_sym___thread] = ACTIONS(7927),
    [anon_sym_CG_EXTERN] = ACTIONS(7927),
    [anon_sym_CG_INLINE] = ACTIONS(7927),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7927),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7927),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7927),
    [anon_sym_IBOutlet] = ACTIONS(7927),
    [anon_sym_IBInspectable] = ACTIONS(7927),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7927),
    [anon_sym_NS_INLINE] = ACTIONS(7927),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7927),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7927),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7927),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7927),
    [anon_sym_const] = ACTIONS(7927),
    [anon_sym_constexpr] = ACTIONS(7927),
    [anon_sym_volatile] = ACTIONS(7927),
    [anon_sym_restrict] = ACTIONS(7927),
    [anon_sym___restrict__] = ACTIONS(7927),
    [anon_sym__Atomic] = ACTIONS(7927),
    [anon_sym__Noreturn] = ACTIONS(7927),
    [anon_sym__Nonnull] = ACTIONS(7927),
    [anon_sym_nullable] = ACTIONS(7927),
    [anon_sym__Complex] = ACTIONS(7927),
    [anon_sym__Nullable] = ACTIONS(7927),
    [anon_sym__Nullable_result] = ACTIONS(7927),
    [anon_sym__Null_unspecified] = ACTIONS(7927),
    [anon_sym___autoreleasing] = ACTIONS(7927),
    [anon_sym___block] = ACTIONS(7927),
    [anon_sym___bridge] = ACTIONS(7927),
    [anon_sym___bridge_retained] = ACTIONS(7927),
    [anon_sym___bridge_transfer] = ACTIONS(7927),
    [anon_sym___complex] = ACTIONS(7927),
    [anon_sym___const] = ACTIONS(7927),
    [anon_sym___imag] = ACTIONS(7927),
    [anon_sym___kindof] = ACTIONS(7927),
    [anon_sym___nonnull] = ACTIONS(7927),
    [anon_sym___nullable] = ACTIONS(7927),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7927),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7927),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7927),
    [anon_sym___real] = ACTIONS(7927),
    [anon_sym___strong] = ACTIONS(7927),
    [anon_sym___unsafe_unretained] = ACTIONS(7927),
    [anon_sym___unused] = ACTIONS(7927),
    [anon_sym___weak] = ACTIONS(7927),
    [anon_sym_alignas] = ACTIONS(7927),
    [anon_sym__Alignas] = ACTIONS(7927),
    [anon_sym_enum] = ACTIONS(7927),
    [anon_sym_COLON] = ACTIONS(7929),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7927),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7927),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7927),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7927),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7927),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7927),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7927),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7927),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7927),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7927),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7927),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7927),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7927),
    [anon_sym_API_AVAILABLE] = ACTIONS(7927),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7927),
    [anon_sym_API_DEPRECATED] = ACTIONS(7927),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7927),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7927),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7927),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7927),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7927),
    [anon_sym___deprecated_msg] = ACTIONS(7927),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7927),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7927),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7927),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7927),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7927),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7927),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7927),
  },
  [STATE(3193)] = {
    [sym__declaration_modifiers] = STATE(6262),
    [sym_attribute_specifier] = STATE(6262),
    [sym_attribute_declaration] = STATE(6262),
    [sym_ms_declspec_modifier] = STATE(6262),
    [sym_storage_class_specifier] = STATE(6262),
    [sym_type_qualifier] = STATE(6262),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6262),
    [aux_sym_struct_declaration_repeat2] = STATE(3276),
    [anon_sym_COMMA] = ACTIONS(7654),
    [anon_sym_SEMI] = ACTIONS(7934),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3194)] = {
    [sym__declaration_modifiers] = STATE(6318),
    [sym_attribute_specifier] = STATE(6318),
    [sym_attribute_declaration] = STATE(6318),
    [sym_ms_declspec_modifier] = STATE(6318),
    [sym_storage_class_specifier] = STATE(6318),
    [sym_type_qualifier] = STATE(6318),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6318),
    [aux_sym_struct_declaration_repeat2] = STATE(3276),
    [anon_sym_COMMA] = ACTIONS(7654),
    [anon_sym_SEMI] = ACTIONS(7936),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3195)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2296),
    [anon_sym_SEMI] = ACTIONS(7577),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3196)] = {
    [sym__declaration_modifiers] = STATE(2864),
    [sym_attribute_specifier] = STATE(2864),
    [sym_attribute_declaration] = STATE(2864),
    [sym_ms_declspec_modifier] = STATE(2864),
    [sym_storage_class_specifier] = STATE(2864),
    [sym_type_qualifier] = STATE(2864),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(2864),
    [aux_sym__declaration_specifiers_repeat1] = STATE(2864),
    [aux_sym_method_declaration_repeat2] = STATE(2102),
    [anon_sym_SEMI] = ACTIONS(7579),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3197)] = {
    [sym__declaration_modifiers] = STATE(6295),
    [sym_attribute_specifier] = STATE(6295),
    [sym_attribute_declaration] = STATE(6295),
    [sym_ms_declspec_modifier] = STATE(6295),
    [sym_storage_class_specifier] = STATE(6295),
    [sym_type_qualifier] = STATE(6295),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6295),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7938),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3198)] = {
    [sym_attribute_specifier] = STATE(2981),
    [sym_identifier] = ACTIONS(7940),
    [anon_sym_COMMA] = ACTIONS(7942),
    [anon_sym_RPAREN] = ACTIONS(7942),
    [anon_sym_LPAREN2] = ACTIONS(7942),
    [anon_sym_STAR] = ACTIONS(7942),
    [anon_sym_CARET] = ACTIONS(7942),
    [anon_sym_SEMI] = ACTIONS(7942),
    [anon_sym___extension__] = ACTIONS(7940),
    [anon_sym_extern] = ACTIONS(7940),
    [anon_sym___attribute__] = ACTIONS(7944),
    [anon_sym___attribute] = ACTIONS(7944),
    [anon_sym_noreturn] = ACTIONS(7940),
    [anon_sym_LBRACK] = ACTIONS(7942),
    [anon_sym___declspec] = ACTIONS(7940),
    [anon_sym___based] = ACTIONS(7940),
    [anon_sym_static] = ACTIONS(7940),
    [anon_sym_auto] = ACTIONS(7940),
    [anon_sym_register] = ACTIONS(7940),
    [anon_sym_inline] = ACTIONS(7940),
    [anon_sym___inline] = ACTIONS(7940),
    [anon_sym___inline__] = ACTIONS(7940),
    [anon_sym___forceinline] = ACTIONS(7940),
    [anon_sym_thread_local] = ACTIONS(7940),
    [anon_sym___thread] = ACTIONS(7940),
    [anon_sym_CG_EXTERN] = ACTIONS(7940),
    [anon_sym_CG_INLINE] = ACTIONS(7940),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7940),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7940),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7940),
    [anon_sym_IBOutlet] = ACTIONS(7940),
    [anon_sym_IBInspectable] = ACTIONS(7940),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7940),
    [anon_sym_NS_INLINE] = ACTIONS(7940),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7940),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7940),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7940),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7940),
    [anon_sym_const] = ACTIONS(7940),
    [anon_sym_constexpr] = ACTIONS(7940),
    [anon_sym_volatile] = ACTIONS(7940),
    [anon_sym_restrict] = ACTIONS(7940),
    [anon_sym___restrict__] = ACTIONS(7940),
    [anon_sym__Atomic] = ACTIONS(7940),
    [anon_sym__Noreturn] = ACTIONS(7940),
    [anon_sym__Nonnull] = ACTIONS(7940),
    [anon_sym_nullable] = ACTIONS(7940),
    [anon_sym__Complex] = ACTIONS(7940),
    [anon_sym__Nullable] = ACTIONS(7940),
    [anon_sym__Nullable_result] = ACTIONS(7940),
    [anon_sym__Null_unspecified] = ACTIONS(7940),
    [anon_sym___autoreleasing] = ACTIONS(7940),
    [anon_sym___block] = ACTIONS(7940),
    [anon_sym___bridge] = ACTIONS(7940),
    [anon_sym___bridge_retained] = ACTIONS(7940),
    [anon_sym___bridge_transfer] = ACTIONS(7940),
    [anon_sym___complex] = ACTIONS(7940),
    [anon_sym___const] = ACTIONS(7940),
    [anon_sym___imag] = ACTIONS(7940),
    [anon_sym___kindof] = ACTIONS(7940),
    [anon_sym___nonnull] = ACTIONS(7940),
    [anon_sym___nullable] = ACTIONS(7940),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7940),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7940),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7940),
    [anon_sym___real] = ACTIONS(7940),
    [anon_sym___strong] = ACTIONS(7940),
    [anon_sym___unsafe_unretained] = ACTIONS(7940),
    [anon_sym___unused] = ACTIONS(7940),
    [anon_sym___weak] = ACTIONS(7940),
    [anon_sym_alignas] = ACTIONS(7940),
    [anon_sym__Alignas] = ACTIONS(7940),
    [anon_sym_enum] = ACTIONS(7940),
    [anon_sym_COLON] = ACTIONS(7942),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7940),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7940),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7940),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7940),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7940),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7940),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7940),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7940),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7940),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7940),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7940),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7940),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7940),
    [anon_sym_API_AVAILABLE] = ACTIONS(7940),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7940),
    [anon_sym_API_DEPRECATED] = ACTIONS(7940),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7940),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7940),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7940),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7940),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7940),
    [anon_sym___deprecated_msg] = ACTIONS(7940),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7940),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7940),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7940),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7940),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7940),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7940),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7940),
  },
  [STATE(3199)] = {
    [sym__declaration_modifiers] = STATE(3191),
    [sym_attribute_specifier] = STATE(3191),
    [sym_attribute_declaration] = STATE(3191),
    [sym_ms_declspec_modifier] = STATE(3191),
    [sym_storage_class_specifier] = STATE(3191),
    [sym_type_qualifier] = STATE(3191),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3191),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3191),
    [aux_sym_method_declaration_repeat2] = STATE(2290),
    [anon_sym_SEMI] = ACTIONS(7502),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3200)] = {
    [sym__declaration_modifiers] = STATE(6061),
    [sym_attribute_specifier] = STATE(6061),
    [sym_attribute_declaration] = STATE(6061),
    [sym_ms_declspec_modifier] = STATE(6061),
    [sym_storage_class_specifier] = STATE(6061),
    [sym_type_qualifier] = STATE(6061),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(6061),
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_SEMI] = ACTIONS(7947),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3201)] = {
    [sym__declaration_modifiers] = STATE(3122),
    [sym_attribute_specifier] = STATE(3122),
    [sym_attribute_declaration] = STATE(3122),
    [sym_ms_declspec_modifier] = STATE(3122),
    [sym_storage_class_specifier] = STATE(3122),
    [sym_type_qualifier] = STATE(3122),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_availability_attribute_specifier] = STATE(3122),
    [aux_sym__declaration_specifiers_repeat1] = STATE(3122),
    [aux_sym_method_declaration_repeat2] = STATE(2187),
    [anon_sym_SEMI] = ACTIONS(7504),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(6766),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(6978),
    [anon_sym___declspec] = ACTIONS(6772),
    [anon_sym_static] = ACTIONS(6766),
    [anon_sym_auto] = ACTIONS(6766),
    [anon_sym_register] = ACTIONS(6766),
    [anon_sym_inline] = ACTIONS(6766),
    [anon_sym___inline] = ACTIONS(51),
    [anon_sym___inline__] = ACTIONS(6766),
    [anon_sym___forceinline] = ACTIONS(6766),
    [anon_sym_thread_local] = ACTIONS(6766),
    [anon_sym___thread] = ACTIONS(6766),
    [anon_sym_CG_EXTERN] = ACTIONS(6766),
    [anon_sym_CG_INLINE] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6766),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6766),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6766),
    [anon_sym_IBOutlet] = ACTIONS(6766),
    [anon_sym_IBInspectable] = ACTIONS(6766),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6766),
    [anon_sym_NS_INLINE] = ACTIONS(6766),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6766),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6766),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6766),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6766),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6776),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6776),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6776),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6776),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6776),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6776),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE] = ACTIONS(127),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_API_AVAILABLE] = ACTIONS(6778),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_API_DEPRECATED] = ACTIONS(6778),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6778),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6778),
    [anon_sym___deprecated_msg] = ACTIONS(6778),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6778),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6778),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6778),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6778),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6778),
  },
  [STATE(3202)] = {
    [sym_attribute_specifier] = STATE(3204),
    [aux_sym_function_declarator_repeat1] = STATE(3204),
    [anon_sym_COMMA] = ACTIONS(6687),
    [anon_sym_LPAREN2] = ACTIONS(6687),
    [anon_sym_SEMI] = ACTIONS(6687),
    [anon_sym___extension__] = ACTIONS(6687),
    [anon_sym_extern] = ACTIONS(6687),
    [anon_sym___attribute__] = ACTIONS(6687),
    [anon_sym___attribute] = ACTIONS(6685),
    [anon_sym_noreturn] = ACTIONS(6687),
    [anon_sym_LBRACK] = ACTIONS(6687),
    [anon_sym___declspec] = ACTIONS(6687),
    [anon_sym_static] = ACTIONS(6687),
    [anon_sym_EQ] = ACTIONS(6687),
    [anon_sym_auto] = ACTIONS(6687),
    [anon_sym_register] = ACTIONS(6687),
    [anon_sym_inline] = ACTIONS(6687),
    [anon_sym___inline] = ACTIONS(6685),
    [anon_sym___inline__] = ACTIONS(6687),
    [anon_sym___forceinline] = ACTIONS(6687),
    [anon_sym_thread_local] = ACTIONS(6687),
    [anon_sym___thread] = ACTIONS(6687),
    [anon_sym_CG_EXTERN] = ACTIONS(6687),
    [anon_sym_CG_INLINE] = ACTIONS(6687),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6687),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6687),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6687),
    [anon_sym_IBOutlet] = ACTIONS(6687),
    [anon_sym_IBInspectable] = ACTIONS(6687),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6687),
    [anon_sym_NS_INLINE] = ACTIONS(6687),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6687),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6687),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6687),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6687),
    [anon_sym_const] = ACTIONS(6685),
    [anon_sym_constexpr] = ACTIONS(6687),
    [anon_sym_volatile] = ACTIONS(6687),
    [anon_sym_restrict] = ACTIONS(6687),
    [anon_sym___restrict__] = ACTIONS(6687),
    [anon_sym__Atomic] = ACTIONS(6687),
    [anon_sym__Noreturn] = ACTIONS(6687),
    [anon_sym__Nonnull] = ACTIONS(6687),
    [anon_sym_nullable] = ACTIONS(6687),
    [anon_sym__Complex] = ACTIONS(6687),
    [anon_sym__Nullable] = ACTIONS(6685),
    [anon_sym__Nullable_result] = ACTIONS(6687),
    [anon_sym__Null_unspecified] = ACTIONS(6687),
    [anon_sym___autoreleasing] = ACTIONS(6687),
    [anon_sym___block] = ACTIONS(6687),
    [anon_sym___bridge] = ACTIONS(6685),
    [anon_sym___bridge_retained] = ACTIONS(6687),
    [anon_sym___bridge_transfer] = ACTIONS(6687),
    [anon_sym___complex] = ACTIONS(6687),
    [anon_sym___const] = ACTIONS(6687),
    [anon_sym___imag] = ACTIONS(6687),
    [anon_sym___kindof] = ACTIONS(6687),
    [anon_sym___nonnull] = ACTIONS(6687),
    [anon_sym___nullable] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6687),
    [anon_sym___real] = ACTIONS(6687),
    [anon_sym___strong] = ACTIONS(6687),
    [anon_sym___unsafe_unretained] = ACTIONS(6687),
    [anon_sym___unused] = ACTIONS(6687),
    [anon_sym___weak] = ACTIONS(6687),
    [anon_sym_alignas] = ACTIONS(6687),
    [anon_sym__Alignas] = ACTIONS(6687),
    [anon_sym_in] = ACTIONS(6685),
    [anon_sym_asm] = ACTIONS(6687),
    [anon_sym___asm__] = ACTIONS(6687),
    [anon_sym___asm] = ACTIONS(6685),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6687),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6687),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6687),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6687),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6687),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6687),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6685),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_API_AVAILABLE] = ACTIONS(6687),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_API_DEPRECATED] = ACTIONS(6687),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6687),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym___deprecated_msg] = ACTIONS(6687),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6687),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6687),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6687),
  },
  [STATE(3203)] = {
    [sym_attribute_specifier] = STATE(3204),
    [aux_sym_function_declarator_repeat1] = STATE(3204),
    [anon_sym_COMMA] = ACTIONS(6698),
    [anon_sym_LPAREN2] = ACTIONS(6698),
    [anon_sym_SEMI] = ACTIONS(6698),
    [anon_sym___extension__] = ACTIONS(6698),
    [anon_sym_extern] = ACTIONS(6698),
    [anon_sym___attribute__] = ACTIONS(6698),
    [anon_sym___attribute] = ACTIONS(6696),
    [anon_sym_noreturn] = ACTIONS(6698),
    [anon_sym_LBRACK] = ACTIONS(6698),
    [anon_sym___declspec] = ACTIONS(6698),
    [anon_sym_static] = ACTIONS(6698),
    [anon_sym_EQ] = ACTIONS(6698),
    [anon_sym_auto] = ACTIONS(6698),
    [anon_sym_register] = ACTIONS(6698),
    [anon_sym_inline] = ACTIONS(6698),
    [anon_sym___inline] = ACTIONS(6696),
    [anon_sym___inline__] = ACTIONS(6698),
    [anon_sym___forceinline] = ACTIONS(6698),
    [anon_sym_thread_local] = ACTIONS(6698),
    [anon_sym___thread] = ACTIONS(6698),
    [anon_sym_CG_EXTERN] = ACTIONS(6698),
    [anon_sym_CG_INLINE] = ACTIONS(6698),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6698),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6698),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6698),
    [anon_sym_IBOutlet] = ACTIONS(6698),
    [anon_sym_IBInspectable] = ACTIONS(6698),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6698),
    [anon_sym_NS_INLINE] = ACTIONS(6698),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6698),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6698),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6698),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6698),
    [anon_sym_const] = ACTIONS(6696),
    [anon_sym_constexpr] = ACTIONS(6698),
    [anon_sym_volatile] = ACTIONS(6698),
    [anon_sym_restrict] = ACTIONS(6698),
    [anon_sym___restrict__] = ACTIONS(6698),
    [anon_sym__Atomic] = ACTIONS(6698),
    [anon_sym__Noreturn] = ACTIONS(6698),
    [anon_sym__Nonnull] = ACTIONS(6698),
    [anon_sym_nullable] = ACTIONS(6698),
    [anon_sym__Complex] = ACTIONS(6698),
    [anon_sym__Nullable] = ACTIONS(6696),
    [anon_sym__Nullable_result] = ACTIONS(6698),
    [anon_sym__Null_unspecified] = ACTIONS(6698),
    [anon_sym___autoreleasing] = ACTIONS(6698),
    [anon_sym___block] = ACTIONS(6698),
    [anon_sym___bridge] = ACTIONS(6696),
    [anon_sym___bridge_retained] = ACTIONS(6698),
    [anon_sym___bridge_transfer] = ACTIONS(6698),
    [anon_sym___complex] = ACTIONS(6698),
    [anon_sym___const] = ACTIONS(6698),
    [anon_sym___imag] = ACTIONS(6698),
    [anon_sym___kindof] = ACTIONS(6698),
    [anon_sym___nonnull] = ACTIONS(6698),
    [anon_sym___nullable] = ACTIONS(6698),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6698),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6698),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6698),
    [anon_sym___real] = ACTIONS(6698),
    [anon_sym___strong] = ACTIONS(6698),
    [anon_sym___unsafe_unretained] = ACTIONS(6698),
    [anon_sym___unused] = ACTIONS(6698),
    [anon_sym___weak] = ACTIONS(6698),
    [anon_sym_alignas] = ACTIONS(6698),
    [anon_sym__Alignas] = ACTIONS(6698),
    [anon_sym_in] = ACTIONS(6696),
    [anon_sym_asm] = ACTIONS(6698),
    [anon_sym___asm__] = ACTIONS(6698),
    [anon_sym___asm] = ACTIONS(6696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6698),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6698),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6698),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6698),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6698),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6698),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6698),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6698),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6698),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6698),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6696),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6698),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6698),
    [anon_sym_API_AVAILABLE] = ACTIONS(6698),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6698),
    [anon_sym_API_DEPRECATED] = ACTIONS(6698),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6698),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6698),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6698),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6698),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6698),
    [anon_sym___deprecated_msg] = ACTIONS(6698),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6698),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6698),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6698),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6698),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6698),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6698),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6698),
  },
  [STATE(3204)] = {
    [sym_attribute_specifier] = STATE(3204),
    [aux_sym_function_declarator_repeat1] = STATE(3204),
    [anon_sym_COMMA] = ACTIONS(6691),
    [anon_sym_LPAREN2] = ACTIONS(6691),
    [anon_sym_SEMI] = ACTIONS(6691),
    [anon_sym___extension__] = ACTIONS(6691),
    [anon_sym_extern] = ACTIONS(6691),
    [anon_sym___attribute__] = ACTIONS(7949),
    [anon_sym___attribute] = ACTIONS(7952),
    [anon_sym_noreturn] = ACTIONS(6691),
    [anon_sym_LBRACK] = ACTIONS(6691),
    [anon_sym___declspec] = ACTIONS(6691),
    [anon_sym_static] = ACTIONS(6691),
    [anon_sym_EQ] = ACTIONS(6691),
    [anon_sym_auto] = ACTIONS(6691),
    [anon_sym_register] = ACTIONS(6691),
    [anon_sym_inline] = ACTIONS(6691),
    [anon_sym___inline] = ACTIONS(6689),
    [anon_sym___inline__] = ACTIONS(6691),
    [anon_sym___forceinline] = ACTIONS(6691),
    [anon_sym_thread_local] = ACTIONS(6691),
    [anon_sym___thread] = ACTIONS(6691),
    [anon_sym_CG_EXTERN] = ACTIONS(6691),
    [anon_sym_CG_INLINE] = ACTIONS(6691),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6691),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6691),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6691),
    [anon_sym_IBOutlet] = ACTIONS(6691),
    [anon_sym_IBInspectable] = ACTIONS(6691),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6691),
    [anon_sym_NS_INLINE] = ACTIONS(6691),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6691),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6691),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6691),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6691),
    [anon_sym_const] = ACTIONS(6689),
    [anon_sym_constexpr] = ACTIONS(6691),
    [anon_sym_volatile] = ACTIONS(6691),
    [anon_sym_restrict] = ACTIONS(6691),
    [anon_sym___restrict__] = ACTIONS(6691),
    [anon_sym__Atomic] = ACTIONS(6691),
    [anon_sym__Noreturn] = ACTIONS(6691),
    [anon_sym__Nonnull] = ACTIONS(6691),
    [anon_sym_nullable] = ACTIONS(6691),
    [anon_sym__Complex] = ACTIONS(6691),
    [anon_sym__Nullable] = ACTIONS(6689),
    [anon_sym__Nullable_result] = ACTIONS(6691),
    [anon_sym__Null_unspecified] = ACTIONS(6691),
    [anon_sym___autoreleasing] = ACTIONS(6691),
    [anon_sym___block] = ACTIONS(6691),
    [anon_sym___bridge] = ACTIONS(6689),
    [anon_sym___bridge_retained] = ACTIONS(6691),
    [anon_sym___bridge_transfer] = ACTIONS(6691),
    [anon_sym___complex] = ACTIONS(6691),
    [anon_sym___const] = ACTIONS(6691),
    [anon_sym___imag] = ACTIONS(6691),
    [anon_sym___kindof] = ACTIONS(6691),
    [anon_sym___nonnull] = ACTIONS(6691),
    [anon_sym___nullable] = ACTIONS(6691),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6691),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6691),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6691),
    [anon_sym___real] = ACTIONS(6691),
    [anon_sym___strong] = ACTIONS(6691),
    [anon_sym___unsafe_unretained] = ACTIONS(6691),
    [anon_sym___unused] = ACTIONS(6691),
    [anon_sym___weak] = ACTIONS(6691),
    [anon_sym_alignas] = ACTIONS(6691),
    [anon_sym__Alignas] = ACTIONS(6691),
    [anon_sym_in] = ACTIONS(6689),
    [anon_sym_asm] = ACTIONS(6691),
    [anon_sym___asm__] = ACTIONS(6691),
    [anon_sym___asm] = ACTIONS(6689),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6691),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6691),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6691),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6691),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6691),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6691),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6691),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6691),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6691),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6691),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6689),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6691),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6691),
    [anon_sym_API_AVAILABLE] = ACTIONS(6691),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6691),
    [anon_sym_API_DEPRECATED] = ACTIONS(6691),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6691),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6691),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6691),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6691),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6691),
    [anon_sym___deprecated_msg] = ACTIONS(6691),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6691),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6691),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6691),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6691),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6691),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6691),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6691),
  },
  [STATE(3205)] = {
    [sym_attribute_specifier] = STATE(3287),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym__abstract_declarator] = STATE(5033),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_parameter_list] = STATE(4271),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_protocol_reference_list] = STATE(4794),
    [sym_specifier_qualifier] = STATE(3287),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3287),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_COMMA] = ACTIONS(7957),
    [anon_sym_RPAREN] = ACTIONS(7957),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(5316),
    [anon_sym_CARET] = ACTIONS(5318),
    [anon_sym_GT] = ACTIONS(7957),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(7963),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3206)] = {
    [sym_attribute_specifier] = STATE(6552),
    [sym_parameter_list] = STATE(3007),
    [sym_gnu_asm_expression] = STATE(3286),
    [anon_sym_COMMA] = ACTIONS(7983),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(7983),
    [anon_sym___extension__] = ACTIONS(7983),
    [anon_sym_extern] = ACTIONS(7983),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(7983),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(7983),
    [anon_sym_static] = ACTIONS(7983),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(7983),
    [anon_sym_register] = ACTIONS(7983),
    [anon_sym_inline] = ACTIONS(7983),
    [anon_sym___inline] = ACTIONS(7985),
    [anon_sym___inline__] = ACTIONS(7983),
    [anon_sym___forceinline] = ACTIONS(7983),
    [anon_sym_thread_local] = ACTIONS(7983),
    [anon_sym___thread] = ACTIONS(7983),
    [anon_sym_CG_EXTERN] = ACTIONS(7983),
    [anon_sym_CG_INLINE] = ACTIONS(7983),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7983),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7983),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7983),
    [anon_sym_IBOutlet] = ACTIONS(7983),
    [anon_sym_IBInspectable] = ACTIONS(7983),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7983),
    [anon_sym_NS_INLINE] = ACTIONS(7983),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7983),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7983),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7983),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7983),
    [anon_sym_const] = ACTIONS(7985),
    [anon_sym_constexpr] = ACTIONS(7983),
    [anon_sym_volatile] = ACTIONS(7983),
    [anon_sym_restrict] = ACTIONS(7983),
    [anon_sym___restrict__] = ACTIONS(7983),
    [anon_sym__Atomic] = ACTIONS(7983),
    [anon_sym__Noreturn] = ACTIONS(7983),
    [anon_sym__Nonnull] = ACTIONS(7983),
    [anon_sym_nullable] = ACTIONS(7983),
    [anon_sym__Complex] = ACTIONS(7983),
    [anon_sym__Nullable] = ACTIONS(7985),
    [anon_sym__Nullable_result] = ACTIONS(7983),
    [anon_sym__Null_unspecified] = ACTIONS(7983),
    [anon_sym___autoreleasing] = ACTIONS(7983),
    [anon_sym___block] = ACTIONS(7983),
    [anon_sym___bridge] = ACTIONS(7985),
    [anon_sym___bridge_retained] = ACTIONS(7983),
    [anon_sym___bridge_transfer] = ACTIONS(7983),
    [anon_sym___complex] = ACTIONS(7983),
    [anon_sym___const] = ACTIONS(7983),
    [anon_sym___imag] = ACTIONS(7983),
    [anon_sym___kindof] = ACTIONS(7983),
    [anon_sym___nonnull] = ACTIONS(7983),
    [anon_sym___nullable] = ACTIONS(7983),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7983),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7983),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7983),
    [anon_sym___real] = ACTIONS(7983),
    [anon_sym___strong] = ACTIONS(7983),
    [anon_sym___unsafe_unretained] = ACTIONS(7983),
    [anon_sym___unused] = ACTIONS(7983),
    [anon_sym___weak] = ACTIONS(7983),
    [anon_sym_alignas] = ACTIONS(7983),
    [anon_sym__Alignas] = ACTIONS(7983),
    [anon_sym_asm] = ACTIONS(6806),
    [anon_sym___asm__] = ACTIONS(6806),
    [anon_sym___asm] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7983),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7983),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7983),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7983),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7983),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7983),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7983),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7983),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7983),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7983),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7985),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7983),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7983),
    [anon_sym_API_AVAILABLE] = ACTIONS(7983),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7983),
    [anon_sym_API_DEPRECATED] = ACTIONS(7983),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7983),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7983),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7983),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7983),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7983),
    [anon_sym___deprecated_msg] = ACTIONS(7983),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7983),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7983),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7983),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7983),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7983),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7983),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7983),
  },
  [STATE(3207)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_LPAREN2] = ACTIONS(2920),
    [anon_sym_BANG] = ACTIONS(2914),
    [anon_sym_TILDE] = ACTIONS(2920),
    [anon_sym_DASH] = ACTIONS(2914),
    [anon_sym_PLUS] = ACTIONS(2914),
    [anon_sym_STAR] = ACTIONS(2920),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4044),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(2920),
    [anon_sym_AMP] = ACTIONS(2914),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(4042),
    [anon_sym_LT_LT] = ACTIONS(4044),
    [anon_sym_GT_GT] = ACTIONS(4044),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym_RBRACK] = ACTIONS(2920),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(2920),
    [anon_sym_PLUS_PLUS] = ACTIONS(2920),
    [anon_sym_sizeof] = ACTIONS(2914),
    [anon_sym___alignof__] = ACTIONS(2914),
    [anon_sym___alignof] = ACTIONS(2914),
    [anon_sym__alignof] = ACTIONS(2914),
    [anon_sym_alignof] = ACTIONS(2914),
    [anon_sym__Alignof] = ACTIONS(2914),
    [anon_sym_offsetof] = ACTIONS(2914),
    [anon_sym__Generic] = ACTIONS(2914),
    [anon_sym_asm] = ACTIONS(2914),
    [anon_sym___asm__] = ACTIONS(2914),
    [anon_sym___asm] = ACTIONS(2914),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_number_literal] = ACTIONS(2920),
    [anon_sym_L_SQUOTE] = ACTIONS(2920),
    [anon_sym_u_SQUOTE] = ACTIONS(2920),
    [anon_sym_U_SQUOTE] = ACTIONS(2920),
    [anon_sym_u8_SQUOTE] = ACTIONS(2920),
    [anon_sym_SQUOTE] = ACTIONS(2920),
    [anon_sym_AT] = ACTIONS(2914),
    [anon_sym_DQUOTE] = ACTIONS(2920),
    [anon_sym_L_DQUOTE] = ACTIONS(2920),
    [anon_sym_u_DQUOTE] = ACTIONS(2920),
    [anon_sym_U_DQUOTE] = ACTIONS(2920),
    [anon_sym_u8_DQUOTE] = ACTIONS(2920),
    [sym_true] = ACTIONS(2914),
    [sym_false] = ACTIONS(2914),
    [anon_sym_NULL] = ACTIONS(2914),
    [anon_sym_nullptr] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2920),
    [anon_sym_ATavailable] = ACTIONS(2920),
    [anon_sym___builtin_available] = ACTIONS(2914),
    [anon_sym_va_arg] = ACTIONS(2914),
    [anon_sym_ATencode] = ACTIONS(2920),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(3208)] = {
    [sym_attribute_specifier] = STATE(3203),
    [aux_sym_function_declarator_repeat1] = STATE(3203),
    [anon_sym_COMMA] = ACTIONS(6687),
    [anon_sym_LPAREN2] = ACTIONS(6687),
    [anon_sym_SEMI] = ACTIONS(6687),
    [anon_sym___extension__] = ACTIONS(6687),
    [anon_sym_extern] = ACTIONS(6687),
    [anon_sym___attribute__] = ACTIONS(7850),
    [anon_sym___attribute] = ACTIONS(7852),
    [anon_sym_noreturn] = ACTIONS(6687),
    [anon_sym_LBRACK] = ACTIONS(6687),
    [anon_sym___declspec] = ACTIONS(6687),
    [anon_sym_static] = ACTIONS(6687),
    [anon_sym_EQ] = ACTIONS(6687),
    [anon_sym_auto] = ACTIONS(6687),
    [anon_sym_register] = ACTIONS(6687),
    [anon_sym_inline] = ACTIONS(6687),
    [anon_sym___inline] = ACTIONS(6685),
    [anon_sym___inline__] = ACTIONS(6687),
    [anon_sym___forceinline] = ACTIONS(6687),
    [anon_sym_thread_local] = ACTIONS(6687),
    [anon_sym___thread] = ACTIONS(6687),
    [anon_sym_CG_EXTERN] = ACTIONS(6687),
    [anon_sym_CG_INLINE] = ACTIONS(6687),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6687),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6687),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6687),
    [anon_sym_IBOutlet] = ACTIONS(6687),
    [anon_sym_IBInspectable] = ACTIONS(6687),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6687),
    [anon_sym_NS_INLINE] = ACTIONS(6687),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6687),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6687),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6687),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6687),
    [anon_sym_const] = ACTIONS(6685),
    [anon_sym_constexpr] = ACTIONS(6687),
    [anon_sym_volatile] = ACTIONS(6687),
    [anon_sym_restrict] = ACTIONS(6687),
    [anon_sym___restrict__] = ACTIONS(6687),
    [anon_sym__Atomic] = ACTIONS(6687),
    [anon_sym__Noreturn] = ACTIONS(6687),
    [anon_sym__Nonnull] = ACTIONS(6687),
    [anon_sym_nullable] = ACTIONS(6687),
    [anon_sym__Complex] = ACTIONS(6687),
    [anon_sym__Nullable] = ACTIONS(6685),
    [anon_sym__Nullable_result] = ACTIONS(6687),
    [anon_sym__Null_unspecified] = ACTIONS(6687),
    [anon_sym___autoreleasing] = ACTIONS(6687),
    [anon_sym___block] = ACTIONS(6687),
    [anon_sym___bridge] = ACTIONS(6685),
    [anon_sym___bridge_retained] = ACTIONS(6687),
    [anon_sym___bridge_transfer] = ACTIONS(6687),
    [anon_sym___complex] = ACTIONS(6687),
    [anon_sym___const] = ACTIONS(6687),
    [anon_sym___imag] = ACTIONS(6687),
    [anon_sym___kindof] = ACTIONS(6687),
    [anon_sym___nonnull] = ACTIONS(6687),
    [anon_sym___nullable] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6687),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6687),
    [anon_sym___real] = ACTIONS(6687),
    [anon_sym___strong] = ACTIONS(6687),
    [anon_sym___unsafe_unretained] = ACTIONS(6687),
    [anon_sym___unused] = ACTIONS(6687),
    [anon_sym___weak] = ACTIONS(6687),
    [anon_sym_alignas] = ACTIONS(6687),
    [anon_sym__Alignas] = ACTIONS(6687),
    [anon_sym_in] = ACTIONS(6685),
    [anon_sym_asm] = ACTIONS(6687),
    [anon_sym___asm__] = ACTIONS(6687),
    [anon_sym___asm] = ACTIONS(6685),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6687),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6687),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6687),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6687),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6687),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6687),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6685),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_API_AVAILABLE] = ACTIONS(6687),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_API_DEPRECATED] = ACTIONS(6687),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6687),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6687),
    [anon_sym___deprecated_msg] = ACTIONS(6687),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6687),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6687),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6687),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6687),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6687),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6687),
  },
  [STATE(3209)] = {
    [sym_parameter_list] = STATE(3158),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6738),
    [anon_sym___extension__] = ACTIONS(6738),
    [anon_sym_extern] = ACTIONS(6738),
    [anon_sym___attribute__] = ACTIONS(6738),
    [anon_sym___attribute] = ACTIONS(6736),
    [anon_sym_noreturn] = ACTIONS(6738),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6738),
    [anon_sym_static] = ACTIONS(6738),
    [anon_sym_EQ] = ACTIONS(6738),
    [anon_sym_auto] = ACTIONS(6738),
    [anon_sym_register] = ACTIONS(6738),
    [anon_sym_inline] = ACTIONS(6738),
    [anon_sym___inline] = ACTIONS(6736),
    [anon_sym___inline__] = ACTIONS(6738),
    [anon_sym___forceinline] = ACTIONS(6738),
    [anon_sym_thread_local] = ACTIONS(6738),
    [anon_sym___thread] = ACTIONS(6738),
    [anon_sym_CG_EXTERN] = ACTIONS(6738),
    [anon_sym_CG_INLINE] = ACTIONS(6738),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6738),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6738),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6738),
    [anon_sym_IBOutlet] = ACTIONS(6738),
    [anon_sym_IBInspectable] = ACTIONS(6738),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6738),
    [anon_sym_NS_INLINE] = ACTIONS(6738),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6738),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6738),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6738),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6738),
    [anon_sym_const] = ACTIONS(6736),
    [anon_sym_constexpr] = ACTIONS(6738),
    [anon_sym_volatile] = ACTIONS(6738),
    [anon_sym_restrict] = ACTIONS(6738),
    [anon_sym___restrict__] = ACTIONS(6738),
    [anon_sym__Atomic] = ACTIONS(6738),
    [anon_sym__Noreturn] = ACTIONS(6738),
    [anon_sym__Nonnull] = ACTIONS(6738),
    [anon_sym_nullable] = ACTIONS(6738),
    [anon_sym__Complex] = ACTIONS(6738),
    [anon_sym__Nullable] = ACTIONS(6736),
    [anon_sym__Nullable_result] = ACTIONS(6738),
    [anon_sym__Null_unspecified] = ACTIONS(6738),
    [anon_sym___autoreleasing] = ACTIONS(6738),
    [anon_sym___block] = ACTIONS(6738),
    [anon_sym___bridge] = ACTIONS(6736),
    [anon_sym___bridge_retained] = ACTIONS(6738),
    [anon_sym___bridge_transfer] = ACTIONS(6738),
    [anon_sym___complex] = ACTIONS(6738),
    [anon_sym___const] = ACTIONS(6738),
    [anon_sym___imag] = ACTIONS(6738),
    [anon_sym___kindof] = ACTIONS(6738),
    [anon_sym___nonnull] = ACTIONS(6738),
    [anon_sym___nullable] = ACTIONS(6738),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6738),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6738),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6738),
    [anon_sym___real] = ACTIONS(6738),
    [anon_sym___strong] = ACTIONS(6738),
    [anon_sym___unsafe_unretained] = ACTIONS(6738),
    [anon_sym___unused] = ACTIONS(6738),
    [anon_sym___weak] = ACTIONS(6738),
    [anon_sym_alignas] = ACTIONS(6738),
    [anon_sym__Alignas] = ACTIONS(6738),
    [anon_sym_in] = ACTIONS(6736),
    [anon_sym_asm] = ACTIONS(6738),
    [anon_sym___asm__] = ACTIONS(6738),
    [anon_sym___asm] = ACTIONS(6736),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6738),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6738),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6738),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6738),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6738),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6738),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6738),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6738),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6738),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6738),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6736),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6738),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6738),
    [anon_sym_API_AVAILABLE] = ACTIONS(6738),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6738),
    [anon_sym_API_DEPRECATED] = ACTIONS(6738),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6738),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6738),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6738),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6738),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6738),
    [anon_sym___deprecated_msg] = ACTIONS(6738),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6738),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6738),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6738),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6738),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6738),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6738),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6738),
  },
  [STATE(3210)] = {
    [sym_parameter_list] = STATE(3158),
    [anon_sym_COMMA] = ACTIONS(6742),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6742),
    [anon_sym___extension__] = ACTIONS(6742),
    [anon_sym_extern] = ACTIONS(6742),
    [anon_sym___attribute__] = ACTIONS(6742),
    [anon_sym___attribute] = ACTIONS(6740),
    [anon_sym_noreturn] = ACTIONS(6742),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6742),
    [anon_sym_static] = ACTIONS(6742),
    [anon_sym_EQ] = ACTIONS(6742),
    [anon_sym_auto] = ACTIONS(6742),
    [anon_sym_register] = ACTIONS(6742),
    [anon_sym_inline] = ACTIONS(6742),
    [anon_sym___inline] = ACTIONS(6740),
    [anon_sym___inline__] = ACTIONS(6742),
    [anon_sym___forceinline] = ACTIONS(6742),
    [anon_sym_thread_local] = ACTIONS(6742),
    [anon_sym___thread] = ACTIONS(6742),
    [anon_sym_CG_EXTERN] = ACTIONS(6742),
    [anon_sym_CG_INLINE] = ACTIONS(6742),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6742),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6742),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6742),
    [anon_sym_IBOutlet] = ACTIONS(6742),
    [anon_sym_IBInspectable] = ACTIONS(6742),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6742),
    [anon_sym_NS_INLINE] = ACTIONS(6742),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6742),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6742),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6742),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6742),
    [anon_sym_const] = ACTIONS(6740),
    [anon_sym_constexpr] = ACTIONS(6742),
    [anon_sym_volatile] = ACTIONS(6742),
    [anon_sym_restrict] = ACTIONS(6742),
    [anon_sym___restrict__] = ACTIONS(6742),
    [anon_sym__Atomic] = ACTIONS(6742),
    [anon_sym__Noreturn] = ACTIONS(6742),
    [anon_sym__Nonnull] = ACTIONS(6742),
    [anon_sym_nullable] = ACTIONS(6742),
    [anon_sym__Complex] = ACTIONS(6742),
    [anon_sym__Nullable] = ACTIONS(6740),
    [anon_sym__Nullable_result] = ACTIONS(6742),
    [anon_sym__Null_unspecified] = ACTIONS(6742),
    [anon_sym___autoreleasing] = ACTIONS(6742),
    [anon_sym___block] = ACTIONS(6742),
    [anon_sym___bridge] = ACTIONS(6740),
    [anon_sym___bridge_retained] = ACTIONS(6742),
    [anon_sym___bridge_transfer] = ACTIONS(6742),
    [anon_sym___complex] = ACTIONS(6742),
    [anon_sym___const] = ACTIONS(6742),
    [anon_sym___imag] = ACTIONS(6742),
    [anon_sym___kindof] = ACTIONS(6742),
    [anon_sym___nonnull] = ACTIONS(6742),
    [anon_sym___nullable] = ACTIONS(6742),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6742),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6742),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6742),
    [anon_sym___real] = ACTIONS(6742),
    [anon_sym___strong] = ACTIONS(6742),
    [anon_sym___unsafe_unretained] = ACTIONS(6742),
    [anon_sym___unused] = ACTIONS(6742),
    [anon_sym___weak] = ACTIONS(6742),
    [anon_sym_alignas] = ACTIONS(6742),
    [anon_sym__Alignas] = ACTIONS(6742),
    [anon_sym_in] = ACTIONS(6740),
    [anon_sym_asm] = ACTIONS(6742),
    [anon_sym___asm__] = ACTIONS(6742),
    [anon_sym___asm] = ACTIONS(6740),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6742),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6742),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6742),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6742),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6742),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6742),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6742),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6742),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6742),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6742),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6740),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6742),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6742),
    [anon_sym_API_AVAILABLE] = ACTIONS(6742),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6742),
    [anon_sym_API_DEPRECATED] = ACTIONS(6742),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6742),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6742),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6742),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6742),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6742),
    [anon_sym___deprecated_msg] = ACTIONS(6742),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6742),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6742),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6742),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6742),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6742),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6742),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6742),
  },
  [STATE(3211)] = {
    [sym_parameter_list] = STATE(3158),
    [anon_sym_COMMA] = ACTIONS(6720),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6720),
    [anon_sym___extension__] = ACTIONS(6720),
    [anon_sym_extern] = ACTIONS(6720),
    [anon_sym___attribute__] = ACTIONS(6720),
    [anon_sym___attribute] = ACTIONS(6718),
    [anon_sym_noreturn] = ACTIONS(6720),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6720),
    [anon_sym_static] = ACTIONS(6720),
    [anon_sym_EQ] = ACTIONS(6720),
    [anon_sym_auto] = ACTIONS(6720),
    [anon_sym_register] = ACTIONS(6720),
    [anon_sym_inline] = ACTIONS(6720),
    [anon_sym___inline] = ACTIONS(6718),
    [anon_sym___inline__] = ACTIONS(6720),
    [anon_sym___forceinline] = ACTIONS(6720),
    [anon_sym_thread_local] = ACTIONS(6720),
    [anon_sym___thread] = ACTIONS(6720),
    [anon_sym_CG_EXTERN] = ACTIONS(6720),
    [anon_sym_CG_INLINE] = ACTIONS(6720),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6720),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6720),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6720),
    [anon_sym_IBOutlet] = ACTIONS(6720),
    [anon_sym_IBInspectable] = ACTIONS(6720),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6720),
    [anon_sym_NS_INLINE] = ACTIONS(6720),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6720),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6720),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6720),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6720),
    [anon_sym_const] = ACTIONS(6718),
    [anon_sym_constexpr] = ACTIONS(6720),
    [anon_sym_volatile] = ACTIONS(6720),
    [anon_sym_restrict] = ACTIONS(6720),
    [anon_sym___restrict__] = ACTIONS(6720),
    [anon_sym__Atomic] = ACTIONS(6720),
    [anon_sym__Noreturn] = ACTIONS(6720),
    [anon_sym__Nonnull] = ACTIONS(6720),
    [anon_sym_nullable] = ACTIONS(6720),
    [anon_sym__Complex] = ACTIONS(6720),
    [anon_sym__Nullable] = ACTIONS(6718),
    [anon_sym__Nullable_result] = ACTIONS(6720),
    [anon_sym__Null_unspecified] = ACTIONS(6720),
    [anon_sym___autoreleasing] = ACTIONS(6720),
    [anon_sym___block] = ACTIONS(6720),
    [anon_sym___bridge] = ACTIONS(6718),
    [anon_sym___bridge_retained] = ACTIONS(6720),
    [anon_sym___bridge_transfer] = ACTIONS(6720),
    [anon_sym___complex] = ACTIONS(6720),
    [anon_sym___const] = ACTIONS(6720),
    [anon_sym___imag] = ACTIONS(6720),
    [anon_sym___kindof] = ACTIONS(6720),
    [anon_sym___nonnull] = ACTIONS(6720),
    [anon_sym___nullable] = ACTIONS(6720),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6720),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6720),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6720),
    [anon_sym___real] = ACTIONS(6720),
    [anon_sym___strong] = ACTIONS(6720),
    [anon_sym___unsafe_unretained] = ACTIONS(6720),
    [anon_sym___unused] = ACTIONS(6720),
    [anon_sym___weak] = ACTIONS(6720),
    [anon_sym_alignas] = ACTIONS(6720),
    [anon_sym__Alignas] = ACTIONS(6720),
    [anon_sym_in] = ACTIONS(6718),
    [anon_sym_asm] = ACTIONS(6720),
    [anon_sym___asm__] = ACTIONS(6720),
    [anon_sym___asm] = ACTIONS(6718),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6720),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6720),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6720),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6720),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6720),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6720),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6720),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6720),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6720),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6720),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6718),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6720),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6720),
    [anon_sym_API_AVAILABLE] = ACTIONS(6720),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6720),
    [anon_sym_API_DEPRECATED] = ACTIONS(6720),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6720),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6720),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6720),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6720),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6720),
    [anon_sym___deprecated_msg] = ACTIONS(6720),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6720),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6720),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6720),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6720),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6720),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6720),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6720),
  },
  [STATE(3212)] = {
    [sym_parameter_list] = STATE(3158),
    [anon_sym_COMMA] = ACTIONS(6724),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6724),
    [anon_sym___extension__] = ACTIONS(6724),
    [anon_sym_extern] = ACTIONS(6724),
    [anon_sym___attribute__] = ACTIONS(6724),
    [anon_sym___attribute] = ACTIONS(6722),
    [anon_sym_noreturn] = ACTIONS(6724),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6724),
    [anon_sym_static] = ACTIONS(6724),
    [anon_sym_EQ] = ACTIONS(6724),
    [anon_sym_auto] = ACTIONS(6724),
    [anon_sym_register] = ACTIONS(6724),
    [anon_sym_inline] = ACTIONS(6724),
    [anon_sym___inline] = ACTIONS(6722),
    [anon_sym___inline__] = ACTIONS(6724),
    [anon_sym___forceinline] = ACTIONS(6724),
    [anon_sym_thread_local] = ACTIONS(6724),
    [anon_sym___thread] = ACTIONS(6724),
    [anon_sym_CG_EXTERN] = ACTIONS(6724),
    [anon_sym_CG_INLINE] = ACTIONS(6724),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6724),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6724),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6724),
    [anon_sym_IBOutlet] = ACTIONS(6724),
    [anon_sym_IBInspectable] = ACTIONS(6724),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6724),
    [anon_sym_NS_INLINE] = ACTIONS(6724),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6724),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6724),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6724),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6724),
    [anon_sym_const] = ACTIONS(6722),
    [anon_sym_constexpr] = ACTIONS(6724),
    [anon_sym_volatile] = ACTIONS(6724),
    [anon_sym_restrict] = ACTIONS(6724),
    [anon_sym___restrict__] = ACTIONS(6724),
    [anon_sym__Atomic] = ACTIONS(6724),
    [anon_sym__Noreturn] = ACTIONS(6724),
    [anon_sym__Nonnull] = ACTIONS(6724),
    [anon_sym_nullable] = ACTIONS(6724),
    [anon_sym__Complex] = ACTIONS(6724),
    [anon_sym__Nullable] = ACTIONS(6722),
    [anon_sym__Nullable_result] = ACTIONS(6724),
    [anon_sym__Null_unspecified] = ACTIONS(6724),
    [anon_sym___autoreleasing] = ACTIONS(6724),
    [anon_sym___block] = ACTIONS(6724),
    [anon_sym___bridge] = ACTIONS(6722),
    [anon_sym___bridge_retained] = ACTIONS(6724),
    [anon_sym___bridge_transfer] = ACTIONS(6724),
    [anon_sym___complex] = ACTIONS(6724),
    [anon_sym___const] = ACTIONS(6724),
    [anon_sym___imag] = ACTIONS(6724),
    [anon_sym___kindof] = ACTIONS(6724),
    [anon_sym___nonnull] = ACTIONS(6724),
    [anon_sym___nullable] = ACTIONS(6724),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6724),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6724),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6724),
    [anon_sym___real] = ACTIONS(6724),
    [anon_sym___strong] = ACTIONS(6724),
    [anon_sym___unsafe_unretained] = ACTIONS(6724),
    [anon_sym___unused] = ACTIONS(6724),
    [anon_sym___weak] = ACTIONS(6724),
    [anon_sym_alignas] = ACTIONS(6724),
    [anon_sym__Alignas] = ACTIONS(6724),
    [anon_sym_in] = ACTIONS(6722),
    [anon_sym_asm] = ACTIONS(6724),
    [anon_sym___asm__] = ACTIONS(6724),
    [anon_sym___asm] = ACTIONS(6722),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6724),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6724),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6724),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6724),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6724),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6724),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6724),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6724),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6724),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6724),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6722),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6724),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6724),
    [anon_sym_API_AVAILABLE] = ACTIONS(6724),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6724),
    [anon_sym_API_DEPRECATED] = ACTIONS(6724),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6724),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6724),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6724),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6724),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6724),
    [anon_sym___deprecated_msg] = ACTIONS(6724),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6724),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6724),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6724),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6724),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6724),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6724),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6724),
  },
  [STATE(3213)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_LPAREN2] = ACTIONS(2920),
    [anon_sym_BANG] = ACTIONS(2914),
    [anon_sym_TILDE] = ACTIONS(2920),
    [anon_sym_DASH] = ACTIONS(2914),
    [anon_sym_PLUS] = ACTIONS(2914),
    [anon_sym_STAR] = ACTIONS(2920),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4044),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(2920),
    [anon_sym_AMP] = ACTIONS(2914),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(4042),
    [anon_sym_LT_LT] = ACTIONS(4044),
    [anon_sym_GT_GT] = ACTIONS(4044),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym_RBRACK] = ACTIONS(4044),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(2920),
    [anon_sym_PLUS_PLUS] = ACTIONS(2920),
    [anon_sym_sizeof] = ACTIONS(2914),
    [anon_sym___alignof__] = ACTIONS(2914),
    [anon_sym___alignof] = ACTIONS(2914),
    [anon_sym__alignof] = ACTIONS(2914),
    [anon_sym_alignof] = ACTIONS(2914),
    [anon_sym__Alignof] = ACTIONS(2914),
    [anon_sym_offsetof] = ACTIONS(2914),
    [anon_sym__Generic] = ACTIONS(2914),
    [anon_sym_asm] = ACTIONS(2914),
    [anon_sym___asm__] = ACTIONS(2914),
    [anon_sym___asm] = ACTIONS(2914),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_number_literal] = ACTIONS(2920),
    [anon_sym_L_SQUOTE] = ACTIONS(2920),
    [anon_sym_u_SQUOTE] = ACTIONS(2920),
    [anon_sym_U_SQUOTE] = ACTIONS(2920),
    [anon_sym_u8_SQUOTE] = ACTIONS(2920),
    [anon_sym_SQUOTE] = ACTIONS(2920),
    [anon_sym_AT] = ACTIONS(2914),
    [anon_sym_DQUOTE] = ACTIONS(2920),
    [anon_sym_L_DQUOTE] = ACTIONS(2920),
    [anon_sym_u_DQUOTE] = ACTIONS(2920),
    [anon_sym_U_DQUOTE] = ACTIONS(2920),
    [anon_sym_u8_DQUOTE] = ACTIONS(2920),
    [sym_true] = ACTIONS(2914),
    [sym_false] = ACTIONS(2914),
    [anon_sym_NULL] = ACTIONS(2914),
    [anon_sym_nullptr] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2920),
    [anon_sym_ATavailable] = ACTIONS(2920),
    [anon_sym___builtin_available] = ACTIONS(2914),
    [anon_sym_va_arg] = ACTIONS(2914),
    [anon_sym_ATencode] = ACTIONS(2920),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(3214)] = {
    [sym_attribute_specifier] = STATE(3292),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym__abstract_declarator] = STATE(4974),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_abstract_parenthesized_declarator] = STATE(4856),
    [sym_pointer_declarator] = STATE(3980),
    [sym_abstract_pointer_declarator] = STATE(4856),
    [sym_function_declarator] = STATE(3980),
    [sym_abstract_function_declarator] = STATE(4856),
    [sym_array_declarator] = STATE(3980),
    [sym_abstract_array_declarator] = STATE(4856),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_parameter_list] = STATE(4290),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_protocol_reference_list] = STATE(4788),
    [sym_specifier_qualifier] = STATE(3292),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_abstract_block_pointer_declarator] = STATE(4856),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3292),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4084),
    [anon_sym_STAR] = ACTIONS(5427),
    [anon_sym_CARET] = ACTIONS(5429),
    [anon_sym_LT] = ACTIONS(3048),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(7963),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_LBRACE] = ACTIONS(7957),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_ATautoreleasepool] = ACTIONS(7957),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7987),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3215)] = {
    [sym_parameter_list] = STATE(3158),
    [anon_sym_COMMA] = ACTIONS(6746),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6746),
    [anon_sym___extension__] = ACTIONS(6746),
    [anon_sym_extern] = ACTIONS(6746),
    [anon_sym___attribute__] = ACTIONS(6746),
    [anon_sym___attribute] = ACTIONS(6744),
    [anon_sym_noreturn] = ACTIONS(6746),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6746),
    [anon_sym_static] = ACTIONS(6746),
    [anon_sym_EQ] = ACTIONS(6746),
    [anon_sym_auto] = ACTIONS(6746),
    [anon_sym_register] = ACTIONS(6746),
    [anon_sym_inline] = ACTIONS(6746),
    [anon_sym___inline] = ACTIONS(6744),
    [anon_sym___inline__] = ACTIONS(6746),
    [anon_sym___forceinline] = ACTIONS(6746),
    [anon_sym_thread_local] = ACTIONS(6746),
    [anon_sym___thread] = ACTIONS(6746),
    [anon_sym_CG_EXTERN] = ACTIONS(6746),
    [anon_sym_CG_INLINE] = ACTIONS(6746),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6746),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6746),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6746),
    [anon_sym_IBOutlet] = ACTIONS(6746),
    [anon_sym_IBInspectable] = ACTIONS(6746),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6746),
    [anon_sym_NS_INLINE] = ACTIONS(6746),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6746),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6746),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6746),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6746),
    [anon_sym_const] = ACTIONS(6744),
    [anon_sym_constexpr] = ACTIONS(6746),
    [anon_sym_volatile] = ACTIONS(6746),
    [anon_sym_restrict] = ACTIONS(6746),
    [anon_sym___restrict__] = ACTIONS(6746),
    [anon_sym__Atomic] = ACTIONS(6746),
    [anon_sym__Noreturn] = ACTIONS(6746),
    [anon_sym__Nonnull] = ACTIONS(6746),
    [anon_sym_nullable] = ACTIONS(6746),
    [anon_sym__Complex] = ACTIONS(6746),
    [anon_sym__Nullable] = ACTIONS(6744),
    [anon_sym__Nullable_result] = ACTIONS(6746),
    [anon_sym__Null_unspecified] = ACTIONS(6746),
    [anon_sym___autoreleasing] = ACTIONS(6746),
    [anon_sym___block] = ACTIONS(6746),
    [anon_sym___bridge] = ACTIONS(6744),
    [anon_sym___bridge_retained] = ACTIONS(6746),
    [anon_sym___bridge_transfer] = ACTIONS(6746),
    [anon_sym___complex] = ACTIONS(6746),
    [anon_sym___const] = ACTIONS(6746),
    [anon_sym___imag] = ACTIONS(6746),
    [anon_sym___kindof] = ACTIONS(6746),
    [anon_sym___nonnull] = ACTIONS(6746),
    [anon_sym___nullable] = ACTIONS(6746),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6746),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6746),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6746),
    [anon_sym___real] = ACTIONS(6746),
    [anon_sym___strong] = ACTIONS(6746),
    [anon_sym___unsafe_unretained] = ACTIONS(6746),
    [anon_sym___unused] = ACTIONS(6746),
    [anon_sym___weak] = ACTIONS(6746),
    [anon_sym_alignas] = ACTIONS(6746),
    [anon_sym__Alignas] = ACTIONS(6746),
    [anon_sym_in] = ACTIONS(6744),
    [anon_sym_asm] = ACTIONS(6746),
    [anon_sym___asm__] = ACTIONS(6746),
    [anon_sym___asm] = ACTIONS(6744),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6746),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6746),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6746),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6746),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6746),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6746),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6746),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6746),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6746),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6746),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6744),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6746),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6746),
    [anon_sym_API_AVAILABLE] = ACTIONS(6746),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6746),
    [anon_sym_API_DEPRECATED] = ACTIONS(6746),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6746),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6746),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6746),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6746),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6746),
    [anon_sym___deprecated_msg] = ACTIONS(6746),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6746),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6746),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6746),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6746),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6746),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6746),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6746),
  },
  [STATE(3216)] = {
    [sym_parameter_list] = STATE(3158),
    [anon_sym_COMMA] = ACTIONS(6750),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6750),
    [anon_sym___extension__] = ACTIONS(6750),
    [anon_sym_extern] = ACTIONS(6750),
    [anon_sym___attribute__] = ACTIONS(6750),
    [anon_sym___attribute] = ACTIONS(6748),
    [anon_sym_noreturn] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_auto] = ACTIONS(6750),
    [anon_sym_register] = ACTIONS(6750),
    [anon_sym_inline] = ACTIONS(6750),
    [anon_sym___inline] = ACTIONS(6748),
    [anon_sym___inline__] = ACTIONS(6750),
    [anon_sym___forceinline] = ACTIONS(6750),
    [anon_sym_thread_local] = ACTIONS(6750),
    [anon_sym___thread] = ACTIONS(6750),
    [anon_sym_CG_EXTERN] = ACTIONS(6750),
    [anon_sym_CG_INLINE] = ACTIONS(6750),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6750),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6750),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6750),
    [anon_sym_IBOutlet] = ACTIONS(6750),
    [anon_sym_IBInspectable] = ACTIONS(6750),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6750),
    [anon_sym_NS_INLINE] = ACTIONS(6750),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6750),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6750),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6750),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6750),
    [anon_sym_const] = ACTIONS(6748),
    [anon_sym_constexpr] = ACTIONS(6750),
    [anon_sym_volatile] = ACTIONS(6750),
    [anon_sym_restrict] = ACTIONS(6750),
    [anon_sym___restrict__] = ACTIONS(6750),
    [anon_sym__Atomic] = ACTIONS(6750),
    [anon_sym__Noreturn] = ACTIONS(6750),
    [anon_sym__Nonnull] = ACTIONS(6750),
    [anon_sym_nullable] = ACTIONS(6750),
    [anon_sym__Complex] = ACTIONS(6750),
    [anon_sym__Nullable] = ACTIONS(6748),
    [anon_sym__Nullable_result] = ACTIONS(6750),
    [anon_sym__Null_unspecified] = ACTIONS(6750),
    [anon_sym___autoreleasing] = ACTIONS(6750),
    [anon_sym___block] = ACTIONS(6750),
    [anon_sym___bridge] = ACTIONS(6748),
    [anon_sym___bridge_retained] = ACTIONS(6750),
    [anon_sym___bridge_transfer] = ACTIONS(6750),
    [anon_sym___complex] = ACTIONS(6750),
    [anon_sym___const] = ACTIONS(6750),
    [anon_sym___imag] = ACTIONS(6750),
    [anon_sym___kindof] = ACTIONS(6750),
    [anon_sym___nonnull] = ACTIONS(6750),
    [anon_sym___nullable] = ACTIONS(6750),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6750),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6750),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6750),
    [anon_sym___real] = ACTIONS(6750),
    [anon_sym___strong] = ACTIONS(6750),
    [anon_sym___unsafe_unretained] = ACTIONS(6750),
    [anon_sym___unused] = ACTIONS(6750),
    [anon_sym___weak] = ACTIONS(6750),
    [anon_sym_alignas] = ACTIONS(6750),
    [anon_sym__Alignas] = ACTIONS(6750),
    [anon_sym_in] = ACTIONS(6748),
    [anon_sym_asm] = ACTIONS(6750),
    [anon_sym___asm__] = ACTIONS(6750),
    [anon_sym___asm] = ACTIONS(6748),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6750),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6750),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6750),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6750),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6750),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6750),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6750),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6750),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6750),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6750),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6748),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6750),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6750),
    [anon_sym_API_AVAILABLE] = ACTIONS(6750),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6750),
    [anon_sym_API_DEPRECATED] = ACTIONS(6750),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6750),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6750),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6750),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6750),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6750),
    [anon_sym___deprecated_msg] = ACTIONS(6750),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6750),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6750),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6750),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6750),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6750),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6750),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6750),
  },
  [STATE(3217)] = {
    [sym_keyword_declarator] = STATE(3221),
    [sym_method_parameter] = STATE(3217),
    [aux_sym_keyword_selector_repeat1] = STATE(3221),
    [aux_sym_method_declaration_repeat1] = STATE(3217),
    [sym_identifier] = ACTIONS(7989),
    [anon_sym_COMMA] = ACTIONS(7992),
    [anon_sym_SEMI] = ACTIONS(7994),
    [anon_sym___extension__] = ACTIONS(7997),
    [anon_sym_extern] = ACTIONS(7997),
    [anon_sym___attribute__] = ACTIONS(7997),
    [anon_sym___attribute] = ACTIONS(7997),
    [anon_sym_noreturn] = ACTIONS(7997),
    [anon_sym_LBRACK] = ACTIONS(7992),
    [anon_sym___declspec] = ACTIONS(7997),
    [anon_sym_static] = ACTIONS(7997),
    [anon_sym_auto] = ACTIONS(7997),
    [anon_sym_register] = ACTIONS(7997),
    [anon_sym_inline] = ACTIONS(7997),
    [anon_sym___inline] = ACTIONS(7997),
    [anon_sym___inline__] = ACTIONS(7997),
    [anon_sym___forceinline] = ACTIONS(7997),
    [anon_sym_thread_local] = ACTIONS(7997),
    [anon_sym___thread] = ACTIONS(7997),
    [anon_sym_CG_EXTERN] = ACTIONS(7997),
    [anon_sym_CG_INLINE] = ACTIONS(7997),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7997),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7997),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7997),
    [anon_sym_IBOutlet] = ACTIONS(7997),
    [anon_sym_IBInspectable] = ACTIONS(7997),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7997),
    [anon_sym_NS_INLINE] = ACTIONS(7997),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7997),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7997),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7997),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7997),
    [anon_sym_const] = ACTIONS(7997),
    [anon_sym_constexpr] = ACTIONS(7997),
    [anon_sym_volatile] = ACTIONS(7997),
    [anon_sym_restrict] = ACTIONS(7997),
    [anon_sym___restrict__] = ACTIONS(7997),
    [anon_sym__Atomic] = ACTIONS(7997),
    [anon_sym__Noreturn] = ACTIONS(7997),
    [anon_sym__Nonnull] = ACTIONS(7997),
    [anon_sym_nullable] = ACTIONS(7997),
    [anon_sym__Complex] = ACTIONS(7997),
    [anon_sym__Nullable] = ACTIONS(7997),
    [anon_sym__Nullable_result] = ACTIONS(7997),
    [anon_sym__Null_unspecified] = ACTIONS(7997),
    [anon_sym___autoreleasing] = ACTIONS(7997),
    [anon_sym___block] = ACTIONS(7997),
    [anon_sym___bridge] = ACTIONS(7997),
    [anon_sym___bridge_retained] = ACTIONS(7997),
    [anon_sym___bridge_transfer] = ACTIONS(7997),
    [anon_sym___complex] = ACTIONS(7997),
    [anon_sym___const] = ACTIONS(7997),
    [anon_sym___imag] = ACTIONS(7997),
    [anon_sym___kindof] = ACTIONS(7997),
    [anon_sym___nonnull] = ACTIONS(7997),
    [anon_sym___nullable] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7997),
    [anon_sym___real] = ACTIONS(7997),
    [anon_sym___strong] = ACTIONS(7997),
    [anon_sym___unsafe_unretained] = ACTIONS(7997),
    [anon_sym___unused] = ACTIONS(7997),
    [anon_sym___weak] = ACTIONS(7997),
    [anon_sym_alignas] = ACTIONS(7997),
    [anon_sym__Alignas] = ACTIONS(7997),
    [anon_sym_COLON] = ACTIONS(7999),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7997),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7997),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7997),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7997),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7997),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7997),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7997),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_API_AVAILABLE] = ACTIONS(7997),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_API_DEPRECATED] = ACTIONS(7997),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7997),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym___deprecated_msg] = ACTIONS(7997),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7997),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7997),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7997),
  },
  [STATE(3218)] = {
    [anon_sym_COMMA] = ACTIONS(2714),
    [anon_sym_LPAREN2] = ACTIONS(2714),
    [anon_sym_SEMI] = ACTIONS(2714),
    [anon_sym___extension__] = ACTIONS(2714),
    [anon_sym_extern] = ACTIONS(2714),
    [anon_sym___attribute__] = ACTIONS(2714),
    [anon_sym___attribute] = ACTIONS(2712),
    [anon_sym_noreturn] = ACTIONS(2714),
    [anon_sym_LBRACK] = ACTIONS(2714),
    [anon_sym___declspec] = ACTIONS(2714),
    [anon_sym_static] = ACTIONS(2714),
    [anon_sym_EQ] = ACTIONS(2714),
    [anon_sym_auto] = ACTIONS(2714),
    [anon_sym_register] = ACTIONS(2714),
    [anon_sym_inline] = ACTIONS(2714),
    [anon_sym___inline] = ACTIONS(2712),
    [anon_sym___inline__] = ACTIONS(2714),
    [anon_sym___forceinline] = ACTIONS(2714),
    [anon_sym_thread_local] = ACTIONS(2714),
    [anon_sym___thread] = ACTIONS(2714),
    [anon_sym_CG_EXTERN] = ACTIONS(2714),
    [anon_sym_CG_INLINE] = ACTIONS(2714),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2714),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2714),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2714),
    [anon_sym_IBOutlet] = ACTIONS(2714),
    [anon_sym_IBInspectable] = ACTIONS(2714),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2714),
    [anon_sym_NS_INLINE] = ACTIONS(2714),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2714),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2714),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2714),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2714),
    [anon_sym_const] = ACTIONS(2712),
    [anon_sym_constexpr] = ACTIONS(2714),
    [anon_sym_volatile] = ACTIONS(2714),
    [anon_sym_restrict] = ACTIONS(2714),
    [anon_sym___restrict__] = ACTIONS(2714),
    [anon_sym__Atomic] = ACTIONS(2714),
    [anon_sym__Noreturn] = ACTIONS(2714),
    [anon_sym__Nonnull] = ACTIONS(2714),
    [anon_sym_nullable] = ACTIONS(2714),
    [anon_sym__Complex] = ACTIONS(2714),
    [anon_sym__Nullable] = ACTIONS(2712),
    [anon_sym__Nullable_result] = ACTIONS(2714),
    [anon_sym__Null_unspecified] = ACTIONS(2714),
    [anon_sym___autoreleasing] = ACTIONS(2714),
    [anon_sym___block] = ACTIONS(2714),
    [anon_sym___bridge] = ACTIONS(2712),
    [anon_sym___bridge_retained] = ACTIONS(2714),
    [anon_sym___bridge_transfer] = ACTIONS(2714),
    [anon_sym___complex] = ACTIONS(2714),
    [anon_sym___const] = ACTIONS(2714),
    [anon_sym___imag] = ACTIONS(2714),
    [anon_sym___kindof] = ACTIONS(2714),
    [anon_sym___nonnull] = ACTIONS(2714),
    [anon_sym___nullable] = ACTIONS(2714),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2714),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2714),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2714),
    [anon_sym___real] = ACTIONS(2714),
    [anon_sym___strong] = ACTIONS(2714),
    [anon_sym___unsafe_unretained] = ACTIONS(2714),
    [anon_sym___unused] = ACTIONS(2714),
    [anon_sym___weak] = ACTIONS(2714),
    [anon_sym_alignas] = ACTIONS(2714),
    [anon_sym__Alignas] = ACTIONS(2714),
    [anon_sym_in] = ACTIONS(2712),
    [anon_sym_asm] = ACTIONS(2714),
    [anon_sym___asm__] = ACTIONS(2714),
    [anon_sym___asm] = ACTIONS(2712),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2714),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2714),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2714),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2714),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2714),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2714),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2714),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2714),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2714),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2714),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2712),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2714),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2714),
    [anon_sym_API_AVAILABLE] = ACTIONS(2714),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2714),
    [anon_sym_API_DEPRECATED] = ACTIONS(2714),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2714),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2714),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2714),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2714),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2714),
    [anon_sym___deprecated_msg] = ACTIONS(2714),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2714),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2714),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2714),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2714),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2714),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2714),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2714),
  },
  [STATE(3219)] = {
    [anon_sym_COMMA] = ACTIONS(2830),
    [anon_sym_LPAREN2] = ACTIONS(2830),
    [anon_sym_SEMI] = ACTIONS(2830),
    [anon_sym___extension__] = ACTIONS(2830),
    [anon_sym_extern] = ACTIONS(2830),
    [anon_sym___attribute__] = ACTIONS(2830),
    [anon_sym___attribute] = ACTIONS(2828),
    [anon_sym_noreturn] = ACTIONS(2830),
    [anon_sym_LBRACK] = ACTIONS(2830),
    [anon_sym___declspec] = ACTIONS(2830),
    [anon_sym_static] = ACTIONS(2830),
    [anon_sym_EQ] = ACTIONS(2830),
    [anon_sym_auto] = ACTIONS(2830),
    [anon_sym_register] = ACTIONS(2830),
    [anon_sym_inline] = ACTIONS(2830),
    [anon_sym___inline] = ACTIONS(2828),
    [anon_sym___inline__] = ACTIONS(2830),
    [anon_sym___forceinline] = ACTIONS(2830),
    [anon_sym_thread_local] = ACTIONS(2830),
    [anon_sym___thread] = ACTIONS(2830),
    [anon_sym_CG_EXTERN] = ACTIONS(2830),
    [anon_sym_CG_INLINE] = ACTIONS(2830),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2830),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2830),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2830),
    [anon_sym_IBOutlet] = ACTIONS(2830),
    [anon_sym_IBInspectable] = ACTIONS(2830),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2830),
    [anon_sym_NS_INLINE] = ACTIONS(2830),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2830),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2830),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2830),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2830),
    [anon_sym_const] = ACTIONS(2828),
    [anon_sym_constexpr] = ACTIONS(2830),
    [anon_sym_volatile] = ACTIONS(2830),
    [anon_sym_restrict] = ACTIONS(2830),
    [anon_sym___restrict__] = ACTIONS(2830),
    [anon_sym__Atomic] = ACTIONS(2830),
    [anon_sym__Noreturn] = ACTIONS(2830),
    [anon_sym__Nonnull] = ACTIONS(2830),
    [anon_sym_nullable] = ACTIONS(2830),
    [anon_sym__Complex] = ACTIONS(2830),
    [anon_sym__Nullable] = ACTIONS(2828),
    [anon_sym__Nullable_result] = ACTIONS(2830),
    [anon_sym__Null_unspecified] = ACTIONS(2830),
    [anon_sym___autoreleasing] = ACTIONS(2830),
    [anon_sym___block] = ACTIONS(2830),
    [anon_sym___bridge] = ACTIONS(2828),
    [anon_sym___bridge_retained] = ACTIONS(2830),
    [anon_sym___bridge_transfer] = ACTIONS(2830),
    [anon_sym___complex] = ACTIONS(2830),
    [anon_sym___const] = ACTIONS(2830),
    [anon_sym___imag] = ACTIONS(2830),
    [anon_sym___kindof] = ACTIONS(2830),
    [anon_sym___nonnull] = ACTIONS(2830),
    [anon_sym___nullable] = ACTIONS(2830),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2830),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2830),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2830),
    [anon_sym___real] = ACTIONS(2830),
    [anon_sym___strong] = ACTIONS(2830),
    [anon_sym___unsafe_unretained] = ACTIONS(2830),
    [anon_sym___unused] = ACTIONS(2830),
    [anon_sym___weak] = ACTIONS(2830),
    [anon_sym_alignas] = ACTIONS(2830),
    [anon_sym__Alignas] = ACTIONS(2830),
    [anon_sym_in] = ACTIONS(2828),
    [anon_sym_asm] = ACTIONS(2830),
    [anon_sym___asm__] = ACTIONS(2830),
    [anon_sym___asm] = ACTIONS(2828),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2830),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2830),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2830),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2830),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2830),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2830),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2830),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2830),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2830),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2830),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2828),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2830),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2830),
    [anon_sym_API_AVAILABLE] = ACTIONS(2830),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2830),
    [anon_sym_API_DEPRECATED] = ACTIONS(2830),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2830),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2830),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2830),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2830),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2830),
    [anon_sym___deprecated_msg] = ACTIONS(2830),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2830),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2830),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2830),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2830),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2830),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2830),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2830),
  },
  [STATE(3220)] = {
    [anon_sym_COMMA] = ACTIONS(2594),
    [anon_sym_LPAREN2] = ACTIONS(2594),
    [anon_sym_SEMI] = ACTIONS(2594),
    [anon_sym___extension__] = ACTIONS(2594),
    [anon_sym_extern] = ACTIONS(2594),
    [anon_sym___attribute__] = ACTIONS(2594),
    [anon_sym___attribute] = ACTIONS(2592),
    [anon_sym_noreturn] = ACTIONS(2594),
    [anon_sym_LBRACK] = ACTIONS(2594),
    [anon_sym___declspec] = ACTIONS(2594),
    [anon_sym_static] = ACTIONS(2594),
    [anon_sym_EQ] = ACTIONS(2594),
    [anon_sym_auto] = ACTIONS(2594),
    [anon_sym_register] = ACTIONS(2594),
    [anon_sym_inline] = ACTIONS(2594),
    [anon_sym___inline] = ACTIONS(2592),
    [anon_sym___inline__] = ACTIONS(2594),
    [anon_sym___forceinline] = ACTIONS(2594),
    [anon_sym_thread_local] = ACTIONS(2594),
    [anon_sym___thread] = ACTIONS(2594),
    [anon_sym_CG_EXTERN] = ACTIONS(2594),
    [anon_sym_CG_INLINE] = ACTIONS(2594),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(2594),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(2594),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(2594),
    [anon_sym_IBOutlet] = ACTIONS(2594),
    [anon_sym_IBInspectable] = ACTIONS(2594),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(2594),
    [anon_sym_NS_INLINE] = ACTIONS(2594),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(2594),
    [anon_sym_OBJC_EXPORT] = ACTIONS(2594),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(2594),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(2594),
    [anon_sym_const] = ACTIONS(2592),
    [anon_sym_constexpr] = ACTIONS(2594),
    [anon_sym_volatile] = ACTIONS(2594),
    [anon_sym_restrict] = ACTIONS(2594),
    [anon_sym___restrict__] = ACTIONS(2594),
    [anon_sym__Atomic] = ACTIONS(2594),
    [anon_sym__Noreturn] = ACTIONS(2594),
    [anon_sym__Nonnull] = ACTIONS(2594),
    [anon_sym_nullable] = ACTIONS(2594),
    [anon_sym__Complex] = ACTIONS(2594),
    [anon_sym__Nullable] = ACTIONS(2592),
    [anon_sym__Nullable_result] = ACTIONS(2594),
    [anon_sym__Null_unspecified] = ACTIONS(2594),
    [anon_sym___autoreleasing] = ACTIONS(2594),
    [anon_sym___block] = ACTIONS(2594),
    [anon_sym___bridge] = ACTIONS(2592),
    [anon_sym___bridge_retained] = ACTIONS(2594),
    [anon_sym___bridge_transfer] = ACTIONS(2594),
    [anon_sym___complex] = ACTIONS(2594),
    [anon_sym___const] = ACTIONS(2594),
    [anon_sym___imag] = ACTIONS(2594),
    [anon_sym___kindof] = ACTIONS(2594),
    [anon_sym___nonnull] = ACTIONS(2594),
    [anon_sym___nullable] = ACTIONS(2594),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2594),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2594),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2594),
    [anon_sym___real] = ACTIONS(2594),
    [anon_sym___strong] = ACTIONS(2594),
    [anon_sym___unsafe_unretained] = ACTIONS(2594),
    [anon_sym___unused] = ACTIONS(2594),
    [anon_sym___weak] = ACTIONS(2594),
    [anon_sym_alignas] = ACTIONS(2594),
    [anon_sym__Alignas] = ACTIONS(2594),
    [anon_sym_in] = ACTIONS(2592),
    [anon_sym_asm] = ACTIONS(2594),
    [anon_sym___asm__] = ACTIONS(2594),
    [anon_sym___asm] = ACTIONS(2592),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(2594),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(2594),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(2594),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(2594),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(2594),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(2594),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(2594),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(2594),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(2594),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(2594),
    [anon_sym_NS_AVAILABLE] = ACTIONS(2592),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(2594),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(2594),
    [anon_sym_API_AVAILABLE] = ACTIONS(2594),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(2594),
    [anon_sym_API_DEPRECATED] = ACTIONS(2594),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(2594),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(2594),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(2594),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(2594),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(2594),
    [anon_sym___deprecated_msg] = ACTIONS(2594),
    [anon_sym___deprecated_enum_msg] = ACTIONS(2594),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(2594),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(2594),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(2594),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(2594),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(2594),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(2594),
  },
  [STATE(3221)] = {
    [sym_attribute_specifier] = STATE(3239),
    [sym_keyword_declarator] = STATE(2744),
    [sym_method_parameter] = STATE(3247),
    [aux_sym_keyword_selector_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(8002),
    [anon_sym_COMMA] = ACTIONS(8004),
    [anon_sym_SEMI] = ACTIONS(8007),
    [anon_sym___extension__] = ACTIONS(8002),
    [anon_sym_extern] = ACTIONS(8002),
    [anon_sym___attribute__] = ACTIONS(8009),
    [anon_sym___attribute] = ACTIONS(8009),
    [anon_sym_noreturn] = ACTIONS(8002),
    [anon_sym_LBRACK] = ACTIONS(8007),
    [anon_sym___declspec] = ACTIONS(8002),
    [anon_sym_static] = ACTIONS(8002),
    [anon_sym_auto] = ACTIONS(8002),
    [anon_sym_register] = ACTIONS(8002),
    [anon_sym_inline] = ACTIONS(8002),
    [anon_sym___inline] = ACTIONS(8002),
    [anon_sym___inline__] = ACTIONS(8002),
    [anon_sym___forceinline] = ACTIONS(8002),
    [anon_sym_thread_local] = ACTIONS(8002),
    [anon_sym___thread] = ACTIONS(8002),
    [anon_sym_CG_EXTERN] = ACTIONS(8002),
    [anon_sym_CG_INLINE] = ACTIONS(8002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8002),
    [anon_sym_IBOutlet] = ACTIONS(8002),
    [anon_sym_IBInspectable] = ACTIONS(8002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8002),
    [anon_sym_NS_INLINE] = ACTIONS(8002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8002),
    [anon_sym_const] = ACTIONS(8002),
    [anon_sym_constexpr] = ACTIONS(8002),
    [anon_sym_volatile] = ACTIONS(8002),
    [anon_sym_restrict] = ACTIONS(8002),
    [anon_sym___restrict__] = ACTIONS(8002),
    [anon_sym__Atomic] = ACTIONS(8002),
    [anon_sym__Noreturn] = ACTIONS(8002),
    [anon_sym__Nonnull] = ACTIONS(8002),
    [anon_sym_nullable] = ACTIONS(8002),
    [anon_sym__Complex] = ACTIONS(8002),
    [anon_sym__Nullable] = ACTIONS(8002),
    [anon_sym__Nullable_result] = ACTIONS(8002),
    [anon_sym__Null_unspecified] = ACTIONS(8002),
    [anon_sym___autoreleasing] = ACTIONS(8002),
    [anon_sym___block] = ACTIONS(8002),
    [anon_sym___bridge] = ACTIONS(8002),
    [anon_sym___bridge_retained] = ACTIONS(8002),
    [anon_sym___bridge_transfer] = ACTIONS(8002),
    [anon_sym___complex] = ACTIONS(8002),
    [anon_sym___const] = ACTIONS(8002),
    [anon_sym___imag] = ACTIONS(8002),
    [anon_sym___kindof] = ACTIONS(8002),
    [anon_sym___nonnull] = ACTIONS(8002),
    [anon_sym___nullable] = ACTIONS(8002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8002),
    [anon_sym___real] = ACTIONS(8002),
    [anon_sym___strong] = ACTIONS(8002),
    [anon_sym___unsafe_unretained] = ACTIONS(8002),
    [anon_sym___unused] = ACTIONS(8002),
    [anon_sym___weak] = ACTIONS(8002),
    [anon_sym_alignas] = ACTIONS(8002),
    [anon_sym__Alignas] = ACTIONS(8002),
    [anon_sym_COLON] = ACTIONS(8012),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8002),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8002),
    [anon_sym_API_AVAILABLE] = ACTIONS(8002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8002),
    [anon_sym_API_DEPRECATED] = ACTIONS(8002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8002),
    [anon_sym___deprecated_msg] = ACTIONS(8002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8002),
  },
  [STATE(3222)] = {
    [sym_attribute_specifier] = STATE(6552),
    [sym_parameter_list] = STATE(3007),
    [aux_sym_c_method_parameter_repeat1] = STATE(3267),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6628),
    [anon_sym___extension__] = ACTIONS(6628),
    [anon_sym_extern] = ACTIONS(6628),
    [anon_sym___attribute__] = ACTIONS(6628),
    [anon_sym___attribute] = ACTIONS(6626),
    [anon_sym_noreturn] = ACTIONS(6628),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6628),
    [anon_sym_static] = ACTIONS(6628),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6628),
    [anon_sym_register] = ACTIONS(6628),
    [anon_sym_inline] = ACTIONS(6628),
    [anon_sym___inline] = ACTIONS(6626),
    [anon_sym___inline__] = ACTIONS(6628),
    [anon_sym___forceinline] = ACTIONS(6628),
    [anon_sym_thread_local] = ACTIONS(6628),
    [anon_sym___thread] = ACTIONS(6628),
    [anon_sym_CG_EXTERN] = ACTIONS(6628),
    [anon_sym_CG_INLINE] = ACTIONS(6628),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6628),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6628),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6628),
    [anon_sym_IBOutlet] = ACTIONS(6628),
    [anon_sym_IBInspectable] = ACTIONS(6628),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6628),
    [anon_sym_NS_INLINE] = ACTIONS(6628),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6628),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6628),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6628),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6628),
    [anon_sym_const] = ACTIONS(6626),
    [anon_sym_constexpr] = ACTIONS(6628),
    [anon_sym_volatile] = ACTIONS(6628),
    [anon_sym_restrict] = ACTIONS(6628),
    [anon_sym___restrict__] = ACTIONS(6628),
    [anon_sym__Atomic] = ACTIONS(6628),
    [anon_sym__Noreturn] = ACTIONS(6628),
    [anon_sym__Nonnull] = ACTIONS(6628),
    [anon_sym_nullable] = ACTIONS(6628),
    [anon_sym__Complex] = ACTIONS(6628),
    [anon_sym__Nullable] = ACTIONS(6626),
    [anon_sym__Nullable_result] = ACTIONS(6628),
    [anon_sym__Null_unspecified] = ACTIONS(6628),
    [anon_sym___autoreleasing] = ACTIONS(6628),
    [anon_sym___block] = ACTIONS(6628),
    [anon_sym___bridge] = ACTIONS(6626),
    [anon_sym___bridge_retained] = ACTIONS(6628),
    [anon_sym___bridge_transfer] = ACTIONS(6628),
    [anon_sym___complex] = ACTIONS(6628),
    [anon_sym___const] = ACTIONS(6628),
    [anon_sym___imag] = ACTIONS(6628),
    [anon_sym___kindof] = ACTIONS(6628),
    [anon_sym___nonnull] = ACTIONS(6628),
    [anon_sym___nullable] = ACTIONS(6628),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6628),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6628),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6628),
    [anon_sym___real] = ACTIONS(6628),
    [anon_sym___strong] = ACTIONS(6628),
    [anon_sym___unsafe_unretained] = ACTIONS(6628),
    [anon_sym___unused] = ACTIONS(6628),
    [anon_sym___weak] = ACTIONS(6628),
    [anon_sym_alignas] = ACTIONS(6628),
    [anon_sym__Alignas] = ACTIONS(6628),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6628),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6628),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6628),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6628),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6628),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6628),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6628),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6628),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6628),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6628),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6626),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6628),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6628),
    [anon_sym_API_AVAILABLE] = ACTIONS(6628),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6628),
    [anon_sym_API_DEPRECATED] = ACTIONS(6628),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6628),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6628),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6628),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6628),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6628),
    [anon_sym___deprecated_msg] = ACTIONS(6628),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6628),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6628),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6628),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6628),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6628),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6628),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6628),
  },
  [STATE(3223)] = {
    [sym_attribute_specifier] = STATE(3224),
    [sym_type_qualifier] = STATE(3224),
    [sym_alignas_qualifier] = STATE(2354),
    [aux_sym_function_field_declarator_repeat1] = STATE(3224),
    [anon_sym_COMMA] = ACTIONS(8015),
    [anon_sym_LPAREN2] = ACTIONS(8015),
    [anon_sym_SEMI] = ACTIONS(8015),
    [anon_sym___extension__] = ACTIONS(6764),
    [anon_sym_extern] = ACTIONS(8015),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6764),
    [anon_sym_LBRACK] = ACTIONS(8015),
    [anon_sym___declspec] = ACTIONS(8015),
    [anon_sym_static] = ACTIONS(8015),
    [anon_sym_auto] = ACTIONS(8015),
    [anon_sym_register] = ACTIONS(8015),
    [anon_sym_inline] = ACTIONS(8015),
    [anon_sym___inline] = ACTIONS(8017),
    [anon_sym___inline__] = ACTIONS(8015),
    [anon_sym___forceinline] = ACTIONS(8015),
    [anon_sym_thread_local] = ACTIONS(8015),
    [anon_sym___thread] = ACTIONS(8015),
    [anon_sym_CG_EXTERN] = ACTIONS(8015),
    [anon_sym_CG_INLINE] = ACTIONS(8015),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8015),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8015),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8015),
    [anon_sym_IBOutlet] = ACTIONS(8015),
    [anon_sym_IBInspectable] = ACTIONS(8015),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8015),
    [anon_sym_NS_INLINE] = ACTIONS(8015),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8015),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8015),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8015),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8015),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(6764),
    [anon_sym_volatile] = ACTIONS(6764),
    [anon_sym_restrict] = ACTIONS(6764),
    [anon_sym___restrict__] = ACTIONS(6764),
    [anon_sym__Atomic] = ACTIONS(6764),
    [anon_sym__Noreturn] = ACTIONS(6764),
    [anon_sym__Nonnull] = ACTIONS(6764),
    [anon_sym_nullable] = ACTIONS(6764),
    [anon_sym__Complex] = ACTIONS(6764),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(6764),
    [anon_sym__Null_unspecified] = ACTIONS(6764),
    [anon_sym___autoreleasing] = ACTIONS(6764),
    [anon_sym___block] = ACTIONS(6764),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(6764),
    [anon_sym___bridge_transfer] = ACTIONS(6764),
    [anon_sym___complex] = ACTIONS(6764),
    [anon_sym___const] = ACTIONS(6764),
    [anon_sym___imag] = ACTIONS(6764),
    [anon_sym___kindof] = ACTIONS(6764),
    [anon_sym___nonnull] = ACTIONS(6764),
    [anon_sym___nullable] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6764),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6764),
    [anon_sym___real] = ACTIONS(6764),
    [anon_sym___strong] = ACTIONS(6764),
    [anon_sym___unsafe_unretained] = ACTIONS(6764),
    [anon_sym___unused] = ACTIONS(6764),
    [anon_sym___weak] = ACTIONS(6764),
    [anon_sym_alignas] = ACTIONS(6774),
    [anon_sym__Alignas] = ACTIONS(6774),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8015),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8015),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8015),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8015),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8015),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8015),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8015),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8015),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8015),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8015),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8017),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8015),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8015),
    [anon_sym_API_AVAILABLE] = ACTIONS(8015),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8015),
    [anon_sym_API_DEPRECATED] = ACTIONS(8015),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8015),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8015),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8015),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8015),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8015),
    [anon_sym___deprecated_msg] = ACTIONS(8015),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8015),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8015),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8015),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8015),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8015),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8015),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8015),
  },
  [STATE(3224)] = {
    [sym_attribute_specifier] = STATE(3047),
    [sym_type_qualifier] = STATE(3047),
    [sym_alignas_qualifier] = STATE(2354),
    [aux_sym_function_field_declarator_repeat1] = STATE(3047),
    [anon_sym_COMMA] = ACTIONS(8019),
    [anon_sym_LPAREN2] = ACTIONS(8019),
    [anon_sym_SEMI] = ACTIONS(8019),
    [anon_sym___extension__] = ACTIONS(8019),
    [anon_sym_extern] = ACTIONS(8019),
    [anon_sym___attribute__] = ACTIONS(8019),
    [anon_sym___attribute] = ACTIONS(8021),
    [anon_sym_noreturn] = ACTIONS(8019),
    [anon_sym_LBRACK] = ACTIONS(8019),
    [anon_sym___declspec] = ACTIONS(8019),
    [anon_sym_static] = ACTIONS(8019),
    [anon_sym_auto] = ACTIONS(8019),
    [anon_sym_register] = ACTIONS(8019),
    [anon_sym_inline] = ACTIONS(8019),
    [anon_sym___inline] = ACTIONS(8021),
    [anon_sym___inline__] = ACTIONS(8019),
    [anon_sym___forceinline] = ACTIONS(8019),
    [anon_sym_thread_local] = ACTIONS(8019),
    [anon_sym___thread] = ACTIONS(8019),
    [anon_sym_CG_EXTERN] = ACTIONS(8019),
    [anon_sym_CG_INLINE] = ACTIONS(8019),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8019),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8019),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8019),
    [anon_sym_IBOutlet] = ACTIONS(8019),
    [anon_sym_IBInspectable] = ACTIONS(8019),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8019),
    [anon_sym_NS_INLINE] = ACTIONS(8019),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8019),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8019),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8019),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8019),
    [anon_sym_const] = ACTIONS(8021),
    [anon_sym_constexpr] = ACTIONS(8019),
    [anon_sym_volatile] = ACTIONS(8019),
    [anon_sym_restrict] = ACTIONS(8019),
    [anon_sym___restrict__] = ACTIONS(8019),
    [anon_sym__Atomic] = ACTIONS(8019),
    [anon_sym__Noreturn] = ACTIONS(8019),
    [anon_sym__Nonnull] = ACTIONS(8019),
    [anon_sym_nullable] = ACTIONS(8019),
    [anon_sym__Complex] = ACTIONS(8019),
    [anon_sym__Nullable] = ACTIONS(8021),
    [anon_sym__Nullable_result] = ACTIONS(8019),
    [anon_sym__Null_unspecified] = ACTIONS(8019),
    [anon_sym___autoreleasing] = ACTIONS(8019),
    [anon_sym___block] = ACTIONS(8019),
    [anon_sym___bridge] = ACTIONS(8021),
    [anon_sym___bridge_retained] = ACTIONS(8019),
    [anon_sym___bridge_transfer] = ACTIONS(8019),
    [anon_sym___complex] = ACTIONS(8019),
    [anon_sym___const] = ACTIONS(8019),
    [anon_sym___imag] = ACTIONS(8019),
    [anon_sym___kindof] = ACTIONS(8019),
    [anon_sym___nonnull] = ACTIONS(8019),
    [anon_sym___nullable] = ACTIONS(8019),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8019),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8019),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8019),
    [anon_sym___real] = ACTIONS(8019),
    [anon_sym___strong] = ACTIONS(8019),
    [anon_sym___unsafe_unretained] = ACTIONS(8019),
    [anon_sym___unused] = ACTIONS(8019),
    [anon_sym___weak] = ACTIONS(8019),
    [anon_sym_alignas] = ACTIONS(8019),
    [anon_sym__Alignas] = ACTIONS(8019),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8019),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8019),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8019),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8019),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8019),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8019),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8019),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8019),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8019),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8019),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8021),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8019),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8019),
    [anon_sym_API_AVAILABLE] = ACTIONS(8019),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8019),
    [anon_sym_API_DEPRECATED] = ACTIONS(8019),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8019),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8019),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8019),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8019),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8019),
    [anon_sym___deprecated_msg] = ACTIONS(8019),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8019),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8019),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8019),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8019),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8019),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8019),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8019),
  },
  [STATE(3225)] = {
    [sym_attribute_specifier] = STATE(3240),
    [sym_method_parameter] = STATE(3254),
    [sym_identifier] = ACTIONS(7997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(8023),
    [anon_sym_COMMA] = ACTIONS(7992),
    [anon_sym_SEMI] = ACTIONS(7992),
    [anon_sym___extension__] = ACTIONS(7997),
    [anon_sym_extern] = ACTIONS(7997),
    [anon_sym___attribute__] = ACTIONS(8025),
    [anon_sym___attribute] = ACTIONS(8025),
    [anon_sym_noreturn] = ACTIONS(7997),
    [anon_sym_LBRACK] = ACTIONS(7992),
    [anon_sym___declspec] = ACTIONS(7997),
    [anon_sym_static] = ACTIONS(7997),
    [anon_sym_auto] = ACTIONS(7997),
    [anon_sym_register] = ACTIONS(7997),
    [anon_sym_inline] = ACTIONS(7997),
    [anon_sym___inline] = ACTIONS(7997),
    [anon_sym___inline__] = ACTIONS(7997),
    [anon_sym___forceinline] = ACTIONS(7997),
    [anon_sym_thread_local] = ACTIONS(7997),
    [anon_sym___thread] = ACTIONS(7997),
    [anon_sym_CG_EXTERN] = ACTIONS(7997),
    [anon_sym_CG_INLINE] = ACTIONS(7997),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7997),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7997),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7997),
    [anon_sym_IBOutlet] = ACTIONS(7997),
    [anon_sym_IBInspectable] = ACTIONS(7997),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7997),
    [anon_sym_NS_INLINE] = ACTIONS(7997),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7997),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7997),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7997),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7997),
    [anon_sym_const] = ACTIONS(7997),
    [anon_sym_constexpr] = ACTIONS(7997),
    [anon_sym_volatile] = ACTIONS(7997),
    [anon_sym_restrict] = ACTIONS(7997),
    [anon_sym___restrict__] = ACTIONS(7997),
    [anon_sym__Atomic] = ACTIONS(7997),
    [anon_sym__Noreturn] = ACTIONS(7997),
    [anon_sym__Nonnull] = ACTIONS(7997),
    [anon_sym_nullable] = ACTIONS(7997),
    [anon_sym__Complex] = ACTIONS(7997),
    [anon_sym__Nullable] = ACTIONS(7997),
    [anon_sym__Nullable_result] = ACTIONS(7997),
    [anon_sym__Null_unspecified] = ACTIONS(7997),
    [anon_sym___autoreleasing] = ACTIONS(7997),
    [anon_sym___block] = ACTIONS(7997),
    [anon_sym___bridge] = ACTIONS(7997),
    [anon_sym___bridge_retained] = ACTIONS(7997),
    [anon_sym___bridge_transfer] = ACTIONS(7997),
    [anon_sym___complex] = ACTIONS(7997),
    [anon_sym___const] = ACTIONS(7997),
    [anon_sym___imag] = ACTIONS(7997),
    [anon_sym___kindof] = ACTIONS(7997),
    [anon_sym___nonnull] = ACTIONS(7997),
    [anon_sym___nullable] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7997),
    [anon_sym___real] = ACTIONS(7997),
    [anon_sym___strong] = ACTIONS(7997),
    [anon_sym___unsafe_unretained] = ACTIONS(7997),
    [anon_sym___unused] = ACTIONS(7997),
    [anon_sym___weak] = ACTIONS(7997),
    [anon_sym_alignas] = ACTIONS(7997),
    [anon_sym__Alignas] = ACTIONS(7997),
    [anon_sym_COLON] = ACTIONS(7999),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7997),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7997),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7997),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7997),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7997),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7997),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7997),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_API_AVAILABLE] = ACTIONS(7997),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_API_DEPRECATED] = ACTIONS(7997),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7997),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym___deprecated_msg] = ACTIONS(7997),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7997),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7997),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7997),
  },
  [STATE(3226)] = {
    [sym_identifier] = ACTIONS(8028),
    [anon_sym_SEMI] = ACTIONS(8030),
    [anon_sym___extension__] = ACTIONS(8028),
    [anon_sym_extern] = ACTIONS(8028),
    [anon_sym___attribute__] = ACTIONS(8028),
    [anon_sym___attribute] = ACTIONS(8028),
    [anon_sym_noreturn] = ACTIONS(8028),
    [anon_sym_LBRACK] = ACTIONS(8030),
    [anon_sym___declspec] = ACTIONS(8028),
    [anon_sym_static] = ACTIONS(8028),
    [anon_sym_auto] = ACTIONS(8028),
    [anon_sym_register] = ACTIONS(8028),
    [anon_sym_inline] = ACTIONS(8028),
    [anon_sym___inline] = ACTIONS(8028),
    [anon_sym___inline__] = ACTIONS(8028),
    [anon_sym___forceinline] = ACTIONS(8028),
    [anon_sym_thread_local] = ACTIONS(8028),
    [anon_sym___thread] = ACTIONS(8028),
    [anon_sym_CG_EXTERN] = ACTIONS(8028),
    [anon_sym_CG_INLINE] = ACTIONS(8028),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8028),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8028),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8028),
    [anon_sym_IBOutlet] = ACTIONS(8028),
    [anon_sym_IBInspectable] = ACTIONS(8028),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8028),
    [anon_sym_NS_INLINE] = ACTIONS(8028),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8028),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8028),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8028),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8028),
    [anon_sym_const] = ACTIONS(8028),
    [anon_sym_constexpr] = ACTIONS(8028),
    [anon_sym_volatile] = ACTIONS(8028),
    [anon_sym_restrict] = ACTIONS(8028),
    [anon_sym___restrict__] = ACTIONS(8028),
    [anon_sym__Atomic] = ACTIONS(8028),
    [anon_sym__Noreturn] = ACTIONS(8028),
    [anon_sym__Nonnull] = ACTIONS(8028),
    [anon_sym_nullable] = ACTIONS(8028),
    [anon_sym__Complex] = ACTIONS(8028),
    [anon_sym__Nullable] = ACTIONS(8028),
    [anon_sym__Nullable_result] = ACTIONS(8028),
    [anon_sym__Null_unspecified] = ACTIONS(8028),
    [anon_sym___autoreleasing] = ACTIONS(8028),
    [anon_sym___block] = ACTIONS(8028),
    [anon_sym___bridge] = ACTIONS(8028),
    [anon_sym___bridge_retained] = ACTIONS(8028),
    [anon_sym___bridge_transfer] = ACTIONS(8028),
    [anon_sym___complex] = ACTIONS(8028),
    [anon_sym___const] = ACTIONS(8028),
    [anon_sym___imag] = ACTIONS(8028),
    [anon_sym___kindof] = ACTIONS(8028),
    [anon_sym___nonnull] = ACTIONS(8028),
    [anon_sym___nullable] = ACTIONS(8028),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8028),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8028),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8028),
    [anon_sym___real] = ACTIONS(8028),
    [anon_sym___strong] = ACTIONS(8028),
    [anon_sym___unsafe_unretained] = ACTIONS(8028),
    [anon_sym___unused] = ACTIONS(8028),
    [anon_sym___weak] = ACTIONS(8028),
    [anon_sym_alignas] = ACTIONS(8028),
    [anon_sym__Alignas] = ACTIONS(8028),
    [anon_sym_COLON] = ACTIONS(8030),
    [anon_sym_struct] = ACTIONS(8028),
    [anon_sym_in] = ACTIONS(8028),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8028),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8028),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8028),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8028),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8028),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8028),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8028),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8028),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8028),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8028),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8028),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8028),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8028),
    [anon_sym_API_AVAILABLE] = ACTIONS(8028),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8028),
    [anon_sym_API_DEPRECATED] = ACTIONS(8028),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8028),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8028),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8028),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8028),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8028),
    [anon_sym___deprecated_msg] = ACTIONS(8028),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8028),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8028),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8028),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8028),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8028),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8028),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8028),
    [anon_sym_id] = ACTIONS(8028),
  },
  [STATE(3227)] = {
    [sym_identifier] = ACTIONS(5074),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym_SEMI] = ACTIONS(8032),
    [anon_sym___extension__] = ACTIONS(5074),
    [anon_sym_extern] = ACTIONS(5074),
    [anon_sym___attribute__] = ACTIONS(5074),
    [anon_sym___attribute] = ACTIONS(5074),
    [anon_sym_noreturn] = ACTIONS(5074),
    [anon_sym_LBRACK] = ACTIONS(5070),
    [anon_sym___declspec] = ACTIONS(5074),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym_static] = ACTIONS(5074),
    [anon_sym_auto] = ACTIONS(5074),
    [anon_sym_register] = ACTIONS(5074),
    [anon_sym_inline] = ACTIONS(5074),
    [anon_sym___inline] = ACTIONS(5074),
    [anon_sym___inline__] = ACTIONS(5074),
    [anon_sym___forceinline] = ACTIONS(5074),
    [anon_sym_thread_local] = ACTIONS(5074),
    [anon_sym___thread] = ACTIONS(5074),
    [anon_sym_CG_EXTERN] = ACTIONS(5074),
    [anon_sym_CG_INLINE] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5074),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5074),
    [anon_sym_IBOutlet] = ACTIONS(5074),
    [anon_sym_IBInspectable] = ACTIONS(5074),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5074),
    [anon_sym_NS_INLINE] = ACTIONS(5074),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5074),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5074),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5074),
    [anon_sym_const] = ACTIONS(5074),
    [anon_sym_constexpr] = ACTIONS(5074),
    [anon_sym_volatile] = ACTIONS(5074),
    [anon_sym_restrict] = ACTIONS(5074),
    [anon_sym___restrict__] = ACTIONS(5074),
    [anon_sym__Atomic] = ACTIONS(5074),
    [anon_sym__Noreturn] = ACTIONS(5074),
    [anon_sym__Nonnull] = ACTIONS(5074),
    [anon_sym_nullable] = ACTIONS(5074),
    [anon_sym__Complex] = ACTIONS(5074),
    [anon_sym__Nullable] = ACTIONS(5074),
    [anon_sym__Nullable_result] = ACTIONS(5074),
    [anon_sym__Null_unspecified] = ACTIONS(5074),
    [anon_sym___autoreleasing] = ACTIONS(5074),
    [anon_sym___block] = ACTIONS(5074),
    [anon_sym___bridge] = ACTIONS(5074),
    [anon_sym___bridge_retained] = ACTIONS(5074),
    [anon_sym___bridge_transfer] = ACTIONS(5074),
    [anon_sym___complex] = ACTIONS(5074),
    [anon_sym___const] = ACTIONS(5074),
    [anon_sym___imag] = ACTIONS(5074),
    [anon_sym___kindof] = ACTIONS(5074),
    [anon_sym___nonnull] = ACTIONS(5074),
    [anon_sym___nullable] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5074),
    [anon_sym___real] = ACTIONS(5074),
    [anon_sym___strong] = ACTIONS(5074),
    [anon_sym___unsafe_unretained] = ACTIONS(5074),
    [anon_sym___unused] = ACTIONS(5074),
    [anon_sym___weak] = ACTIONS(5074),
    [anon_sym_alignas] = ACTIONS(5074),
    [anon_sym__Alignas] = ACTIONS(5074),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5074),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5074),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5074),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_API_AVAILABLE] = ACTIONS(5074),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_API_DEPRECATED] = ACTIONS(5074),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym___deprecated_msg] = ACTIONS(5074),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5074),
  },
  [STATE(3228)] = {
    [sym_identifier] = ACTIONS(5074),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym_SEMI] = ACTIONS(8034),
    [anon_sym___extension__] = ACTIONS(5074),
    [anon_sym_extern] = ACTIONS(5074),
    [anon_sym___attribute__] = ACTIONS(5074),
    [anon_sym___attribute] = ACTIONS(5074),
    [anon_sym_noreturn] = ACTIONS(5074),
    [anon_sym_LBRACK] = ACTIONS(5070),
    [anon_sym___declspec] = ACTIONS(5074),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym_static] = ACTIONS(5074),
    [anon_sym_auto] = ACTIONS(5074),
    [anon_sym_register] = ACTIONS(5074),
    [anon_sym_inline] = ACTIONS(5074),
    [anon_sym___inline] = ACTIONS(5074),
    [anon_sym___inline__] = ACTIONS(5074),
    [anon_sym___forceinline] = ACTIONS(5074),
    [anon_sym_thread_local] = ACTIONS(5074),
    [anon_sym___thread] = ACTIONS(5074),
    [anon_sym_CG_EXTERN] = ACTIONS(5074),
    [anon_sym_CG_INLINE] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5074),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5074),
    [anon_sym_IBOutlet] = ACTIONS(5074),
    [anon_sym_IBInspectable] = ACTIONS(5074),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5074),
    [anon_sym_NS_INLINE] = ACTIONS(5074),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5074),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5074),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5074),
    [anon_sym_const] = ACTIONS(5074),
    [anon_sym_constexpr] = ACTIONS(5074),
    [anon_sym_volatile] = ACTIONS(5074),
    [anon_sym_restrict] = ACTIONS(5074),
    [anon_sym___restrict__] = ACTIONS(5074),
    [anon_sym__Atomic] = ACTIONS(5074),
    [anon_sym__Noreturn] = ACTIONS(5074),
    [anon_sym__Nonnull] = ACTIONS(5074),
    [anon_sym_nullable] = ACTIONS(5074),
    [anon_sym__Complex] = ACTIONS(5074),
    [anon_sym__Nullable] = ACTIONS(5074),
    [anon_sym__Nullable_result] = ACTIONS(5074),
    [anon_sym__Null_unspecified] = ACTIONS(5074),
    [anon_sym___autoreleasing] = ACTIONS(5074),
    [anon_sym___block] = ACTIONS(5074),
    [anon_sym___bridge] = ACTIONS(5074),
    [anon_sym___bridge_retained] = ACTIONS(5074),
    [anon_sym___bridge_transfer] = ACTIONS(5074),
    [anon_sym___complex] = ACTIONS(5074),
    [anon_sym___const] = ACTIONS(5074),
    [anon_sym___imag] = ACTIONS(5074),
    [anon_sym___kindof] = ACTIONS(5074),
    [anon_sym___nonnull] = ACTIONS(5074),
    [anon_sym___nullable] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5074),
    [anon_sym___real] = ACTIONS(5074),
    [anon_sym___strong] = ACTIONS(5074),
    [anon_sym___unsafe_unretained] = ACTIONS(5074),
    [anon_sym___unused] = ACTIONS(5074),
    [anon_sym___weak] = ACTIONS(5074),
    [anon_sym_alignas] = ACTIONS(5074),
    [anon_sym__Alignas] = ACTIONS(5074),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5074),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5074),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5074),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_API_AVAILABLE] = ACTIONS(5074),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_API_DEPRECATED] = ACTIONS(5074),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym___deprecated_msg] = ACTIONS(5074),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5074),
  },
  [STATE(3229)] = {
    [sym_attribute_specifier] = STATE(3239),
    [sym_method_parameter] = STATE(3247),
    [sym_identifier] = ACTIONS(8002),
    [anon_sym_COMMA] = ACTIONS(8007),
    [anon_sym_SEMI] = ACTIONS(8007),
    [anon_sym___extension__] = ACTIONS(8002),
    [anon_sym_extern] = ACTIONS(8002),
    [anon_sym___attribute__] = ACTIONS(8009),
    [anon_sym___attribute] = ACTIONS(8009),
    [anon_sym_noreturn] = ACTIONS(8002),
    [anon_sym_LBRACK] = ACTIONS(8007),
    [anon_sym___declspec] = ACTIONS(8002),
    [anon_sym_static] = ACTIONS(8002),
    [anon_sym_auto] = ACTIONS(8002),
    [anon_sym_register] = ACTIONS(8002),
    [anon_sym_inline] = ACTIONS(8002),
    [anon_sym___inline] = ACTIONS(8002),
    [anon_sym___inline__] = ACTIONS(8002),
    [anon_sym___forceinline] = ACTIONS(8002),
    [anon_sym_thread_local] = ACTIONS(8002),
    [anon_sym___thread] = ACTIONS(8002),
    [anon_sym_CG_EXTERN] = ACTIONS(8002),
    [anon_sym_CG_INLINE] = ACTIONS(8002),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8002),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8002),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8002),
    [anon_sym_IBOutlet] = ACTIONS(8002),
    [anon_sym_IBInspectable] = ACTIONS(8002),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8002),
    [anon_sym_NS_INLINE] = ACTIONS(8002),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8002),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8002),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8002),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8002),
    [anon_sym_const] = ACTIONS(8002),
    [anon_sym_constexpr] = ACTIONS(8002),
    [anon_sym_volatile] = ACTIONS(8002),
    [anon_sym_restrict] = ACTIONS(8002),
    [anon_sym___restrict__] = ACTIONS(8002),
    [anon_sym__Atomic] = ACTIONS(8002),
    [anon_sym__Noreturn] = ACTIONS(8002),
    [anon_sym__Nonnull] = ACTIONS(8002),
    [anon_sym_nullable] = ACTIONS(8002),
    [anon_sym__Complex] = ACTIONS(8002),
    [anon_sym__Nullable] = ACTIONS(8002),
    [anon_sym__Nullable_result] = ACTIONS(8002),
    [anon_sym__Null_unspecified] = ACTIONS(8002),
    [anon_sym___autoreleasing] = ACTIONS(8002),
    [anon_sym___block] = ACTIONS(8002),
    [anon_sym___bridge] = ACTIONS(8002),
    [anon_sym___bridge_retained] = ACTIONS(8002),
    [anon_sym___bridge_transfer] = ACTIONS(8002),
    [anon_sym___complex] = ACTIONS(8002),
    [anon_sym___const] = ACTIONS(8002),
    [anon_sym___imag] = ACTIONS(8002),
    [anon_sym___kindof] = ACTIONS(8002),
    [anon_sym___nonnull] = ACTIONS(8002),
    [anon_sym___nullable] = ACTIONS(8002),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8002),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8002),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8002),
    [anon_sym___real] = ACTIONS(8002),
    [anon_sym___strong] = ACTIONS(8002),
    [anon_sym___unsafe_unretained] = ACTIONS(8002),
    [anon_sym___unused] = ACTIONS(8002),
    [anon_sym___weak] = ACTIONS(8002),
    [anon_sym_alignas] = ACTIONS(8002),
    [anon_sym__Alignas] = ACTIONS(8002),
    [anon_sym_COLON] = ACTIONS(8012),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8002),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8002),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8002),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8002),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8002),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8002),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8002),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8002),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8002),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8002),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8002),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8002),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8002),
    [anon_sym_API_AVAILABLE] = ACTIONS(8002),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8002),
    [anon_sym_API_DEPRECATED] = ACTIONS(8002),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8002),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8002),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8002),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8002),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8002),
    [anon_sym___deprecated_msg] = ACTIONS(8002),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8002),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8002),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8002),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8002),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8002),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8002),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8002),
  },
  [STATE(3230)] = {
    [sym_identifier] = ACTIONS(5074),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym_SEMI] = ACTIONS(8036),
    [anon_sym___extension__] = ACTIONS(5074),
    [anon_sym_extern] = ACTIONS(5074),
    [anon_sym___attribute__] = ACTIONS(5074),
    [anon_sym___attribute] = ACTIONS(5074),
    [anon_sym_noreturn] = ACTIONS(5074),
    [anon_sym_LBRACK] = ACTIONS(5070),
    [anon_sym___declspec] = ACTIONS(5074),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym_static] = ACTIONS(5074),
    [anon_sym_auto] = ACTIONS(5074),
    [anon_sym_register] = ACTIONS(5074),
    [anon_sym_inline] = ACTIONS(5074),
    [anon_sym___inline] = ACTIONS(5074),
    [anon_sym___inline__] = ACTIONS(5074),
    [anon_sym___forceinline] = ACTIONS(5074),
    [anon_sym_thread_local] = ACTIONS(5074),
    [anon_sym___thread] = ACTIONS(5074),
    [anon_sym_CG_EXTERN] = ACTIONS(5074),
    [anon_sym_CG_INLINE] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5074),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5074),
    [anon_sym_IBOutlet] = ACTIONS(5074),
    [anon_sym_IBInspectable] = ACTIONS(5074),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5074),
    [anon_sym_NS_INLINE] = ACTIONS(5074),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5074),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5074),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5074),
    [anon_sym_const] = ACTIONS(5074),
    [anon_sym_constexpr] = ACTIONS(5074),
    [anon_sym_volatile] = ACTIONS(5074),
    [anon_sym_restrict] = ACTIONS(5074),
    [anon_sym___restrict__] = ACTIONS(5074),
    [anon_sym__Atomic] = ACTIONS(5074),
    [anon_sym__Noreturn] = ACTIONS(5074),
    [anon_sym__Nonnull] = ACTIONS(5074),
    [anon_sym_nullable] = ACTIONS(5074),
    [anon_sym__Complex] = ACTIONS(5074),
    [anon_sym__Nullable] = ACTIONS(5074),
    [anon_sym__Nullable_result] = ACTIONS(5074),
    [anon_sym__Null_unspecified] = ACTIONS(5074),
    [anon_sym___autoreleasing] = ACTIONS(5074),
    [anon_sym___block] = ACTIONS(5074),
    [anon_sym___bridge] = ACTIONS(5074),
    [anon_sym___bridge_retained] = ACTIONS(5074),
    [anon_sym___bridge_transfer] = ACTIONS(5074),
    [anon_sym___complex] = ACTIONS(5074),
    [anon_sym___const] = ACTIONS(5074),
    [anon_sym___imag] = ACTIONS(5074),
    [anon_sym___kindof] = ACTIONS(5074),
    [anon_sym___nonnull] = ACTIONS(5074),
    [anon_sym___nullable] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5074),
    [anon_sym___real] = ACTIONS(5074),
    [anon_sym___strong] = ACTIONS(5074),
    [anon_sym___unsafe_unretained] = ACTIONS(5074),
    [anon_sym___unused] = ACTIONS(5074),
    [anon_sym___weak] = ACTIONS(5074),
    [anon_sym_alignas] = ACTIONS(5074),
    [anon_sym__Alignas] = ACTIONS(5074),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5074),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5074),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5074),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_API_AVAILABLE] = ACTIONS(5074),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_API_DEPRECATED] = ACTIONS(5074),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym___deprecated_msg] = ACTIONS(5074),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5074),
  },
  [STATE(3231)] = {
    [sym_identifier] = ACTIONS(5074),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym_SEMI] = ACTIONS(8038),
    [anon_sym___extension__] = ACTIONS(5074),
    [anon_sym_extern] = ACTIONS(5074),
    [anon_sym___attribute__] = ACTIONS(5074),
    [anon_sym___attribute] = ACTIONS(5074),
    [anon_sym_noreturn] = ACTIONS(5074),
    [anon_sym_LBRACK] = ACTIONS(5070),
    [anon_sym___declspec] = ACTIONS(5074),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym_static] = ACTIONS(5074),
    [anon_sym_auto] = ACTIONS(5074),
    [anon_sym_register] = ACTIONS(5074),
    [anon_sym_inline] = ACTIONS(5074),
    [anon_sym___inline] = ACTIONS(5074),
    [anon_sym___inline__] = ACTIONS(5074),
    [anon_sym___forceinline] = ACTIONS(5074),
    [anon_sym_thread_local] = ACTIONS(5074),
    [anon_sym___thread] = ACTIONS(5074),
    [anon_sym_CG_EXTERN] = ACTIONS(5074),
    [anon_sym_CG_INLINE] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5074),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5074),
    [anon_sym_IBOutlet] = ACTIONS(5074),
    [anon_sym_IBInspectable] = ACTIONS(5074),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5074),
    [anon_sym_NS_INLINE] = ACTIONS(5074),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5074),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5074),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5074),
    [anon_sym_const] = ACTIONS(5074),
    [anon_sym_constexpr] = ACTIONS(5074),
    [anon_sym_volatile] = ACTIONS(5074),
    [anon_sym_restrict] = ACTIONS(5074),
    [anon_sym___restrict__] = ACTIONS(5074),
    [anon_sym__Atomic] = ACTIONS(5074),
    [anon_sym__Noreturn] = ACTIONS(5074),
    [anon_sym__Nonnull] = ACTIONS(5074),
    [anon_sym_nullable] = ACTIONS(5074),
    [anon_sym__Complex] = ACTIONS(5074),
    [anon_sym__Nullable] = ACTIONS(5074),
    [anon_sym__Nullable_result] = ACTIONS(5074),
    [anon_sym__Null_unspecified] = ACTIONS(5074),
    [anon_sym___autoreleasing] = ACTIONS(5074),
    [anon_sym___block] = ACTIONS(5074),
    [anon_sym___bridge] = ACTIONS(5074),
    [anon_sym___bridge_retained] = ACTIONS(5074),
    [anon_sym___bridge_transfer] = ACTIONS(5074),
    [anon_sym___complex] = ACTIONS(5074),
    [anon_sym___const] = ACTIONS(5074),
    [anon_sym___imag] = ACTIONS(5074),
    [anon_sym___kindof] = ACTIONS(5074),
    [anon_sym___nonnull] = ACTIONS(5074),
    [anon_sym___nullable] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5074),
    [anon_sym___real] = ACTIONS(5074),
    [anon_sym___strong] = ACTIONS(5074),
    [anon_sym___unsafe_unretained] = ACTIONS(5074),
    [anon_sym___unused] = ACTIONS(5074),
    [anon_sym___weak] = ACTIONS(5074),
    [anon_sym_alignas] = ACTIONS(5074),
    [anon_sym__Alignas] = ACTIONS(5074),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5074),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5074),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5074),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_API_AVAILABLE] = ACTIONS(5074),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_API_DEPRECATED] = ACTIONS(5074),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym___deprecated_msg] = ACTIONS(5074),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5074),
  },
  [STATE(3232)] = {
    [sym_identifier] = ACTIONS(8040),
    [anon_sym_COMMA] = ACTIONS(8042),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(8042),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(8042),
    [anon_sym___extension__] = ACTIONS(8040),
    [anon_sym_extern] = ACTIONS(8040),
    [anon_sym___attribute__] = ACTIONS(8040),
    [anon_sym___attribute] = ACTIONS(8040),
    [anon_sym_noreturn] = ACTIONS(8040),
    [anon_sym_LBRACK] = ACTIONS(8042),
    [anon_sym___declspec] = ACTIONS(8040),
    [anon_sym_RBRACE] = ACTIONS(8042),
    [anon_sym_static] = ACTIONS(8040),
    [anon_sym_EQ] = ACTIONS(8044),
    [anon_sym_auto] = ACTIONS(8040),
    [anon_sym_register] = ACTIONS(8040),
    [anon_sym_inline] = ACTIONS(8040),
    [anon_sym___inline] = ACTIONS(8040),
    [anon_sym___inline__] = ACTIONS(8040),
    [anon_sym___forceinline] = ACTIONS(8040),
    [anon_sym_thread_local] = ACTIONS(8040),
    [anon_sym___thread] = ACTIONS(8040),
    [anon_sym_CG_EXTERN] = ACTIONS(8040),
    [anon_sym_CG_INLINE] = ACTIONS(8040),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8040),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8040),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8040),
    [anon_sym_IBOutlet] = ACTIONS(8040),
    [anon_sym_IBInspectable] = ACTIONS(8040),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8040),
    [anon_sym_NS_INLINE] = ACTIONS(8040),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8040),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8040),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8040),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8040),
    [anon_sym_const] = ACTIONS(8040),
    [anon_sym_constexpr] = ACTIONS(8040),
    [anon_sym_volatile] = ACTIONS(8040),
    [anon_sym_restrict] = ACTIONS(8040),
    [anon_sym___restrict__] = ACTIONS(8040),
    [anon_sym__Atomic] = ACTIONS(8040),
    [anon_sym__Noreturn] = ACTIONS(8040),
    [anon_sym__Nonnull] = ACTIONS(8040),
    [anon_sym_nullable] = ACTIONS(8040),
    [anon_sym__Complex] = ACTIONS(8040),
    [anon_sym__Nullable] = ACTIONS(8040),
    [anon_sym__Nullable_result] = ACTIONS(8040),
    [anon_sym__Null_unspecified] = ACTIONS(8040),
    [anon_sym___autoreleasing] = ACTIONS(8040),
    [anon_sym___block] = ACTIONS(8040),
    [anon_sym___bridge] = ACTIONS(8040),
    [anon_sym___bridge_retained] = ACTIONS(8040),
    [anon_sym___bridge_transfer] = ACTIONS(8040),
    [anon_sym___complex] = ACTIONS(8040),
    [anon_sym___const] = ACTIONS(8040),
    [anon_sym___imag] = ACTIONS(8040),
    [anon_sym___kindof] = ACTIONS(8040),
    [anon_sym___nonnull] = ACTIONS(8040),
    [anon_sym___nullable] = ACTIONS(8040),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8040),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8040),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8040),
    [anon_sym___real] = ACTIONS(8040),
    [anon_sym___strong] = ACTIONS(8040),
    [anon_sym___unsafe_unretained] = ACTIONS(8040),
    [anon_sym___unused] = ACTIONS(8040),
    [anon_sym___weak] = ACTIONS(8040),
    [anon_sym_alignas] = ACTIONS(8040),
    [anon_sym__Alignas] = ACTIONS(8040),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8040),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8040),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8040),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8040),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8040),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8040),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8040),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8040),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8040),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8040),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8040),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8040),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8040),
    [anon_sym_API_AVAILABLE] = ACTIONS(8040),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8040),
    [anon_sym_API_DEPRECATED] = ACTIONS(8040),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8040),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8040),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8040),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8040),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8040),
    [anon_sym___deprecated_msg] = ACTIONS(8040),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8040),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8040),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8040),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8040),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8040),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8040),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8040),
  },
  [STATE(3233)] = {
    [sym_attribute_specifier] = STATE(6552),
    [sym_parameter_list] = STATE(3007),
    [anon_sym_COMMA] = ACTIONS(6716),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(6716),
    [anon_sym___extension__] = ACTIONS(6716),
    [anon_sym_extern] = ACTIONS(6716),
    [anon_sym___attribute__] = ACTIONS(6768),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(6716),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(6716),
    [anon_sym_static] = ACTIONS(6716),
    [anon_sym_EQ] = ACTIONS(6804),
    [anon_sym_auto] = ACTIONS(6716),
    [anon_sym_register] = ACTIONS(6716),
    [anon_sym_inline] = ACTIONS(6716),
    [anon_sym___inline] = ACTIONS(6714),
    [anon_sym___inline__] = ACTIONS(6716),
    [anon_sym___forceinline] = ACTIONS(6716),
    [anon_sym_thread_local] = ACTIONS(6716),
    [anon_sym___thread] = ACTIONS(6716),
    [anon_sym_CG_EXTERN] = ACTIONS(6716),
    [anon_sym_CG_INLINE] = ACTIONS(6716),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6716),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6716),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6716),
    [anon_sym_IBOutlet] = ACTIONS(6716),
    [anon_sym_IBInspectable] = ACTIONS(6716),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6716),
    [anon_sym_NS_INLINE] = ACTIONS(6716),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6716),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6716),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6716),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6716),
    [anon_sym_const] = ACTIONS(6714),
    [anon_sym_constexpr] = ACTIONS(6716),
    [anon_sym_volatile] = ACTIONS(6716),
    [anon_sym_restrict] = ACTIONS(6716),
    [anon_sym___restrict__] = ACTIONS(6716),
    [anon_sym__Atomic] = ACTIONS(6716),
    [anon_sym__Noreturn] = ACTIONS(6716),
    [anon_sym__Nonnull] = ACTIONS(6716),
    [anon_sym_nullable] = ACTIONS(6716),
    [anon_sym__Complex] = ACTIONS(6716),
    [anon_sym__Nullable] = ACTIONS(6714),
    [anon_sym__Nullable_result] = ACTIONS(6716),
    [anon_sym__Null_unspecified] = ACTIONS(6716),
    [anon_sym___autoreleasing] = ACTIONS(6716),
    [anon_sym___block] = ACTIONS(6716),
    [anon_sym___bridge] = ACTIONS(6714),
    [anon_sym___bridge_retained] = ACTIONS(6716),
    [anon_sym___bridge_transfer] = ACTIONS(6716),
    [anon_sym___complex] = ACTIONS(6716),
    [anon_sym___const] = ACTIONS(6716),
    [anon_sym___imag] = ACTIONS(6716),
    [anon_sym___kindof] = ACTIONS(6716),
    [anon_sym___nonnull] = ACTIONS(6716),
    [anon_sym___nullable] = ACTIONS(6716),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6716),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6716),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6716),
    [anon_sym___real] = ACTIONS(6716),
    [anon_sym___strong] = ACTIONS(6716),
    [anon_sym___unsafe_unretained] = ACTIONS(6716),
    [anon_sym___unused] = ACTIONS(6716),
    [anon_sym___weak] = ACTIONS(6716),
    [anon_sym_alignas] = ACTIONS(6716),
    [anon_sym__Alignas] = ACTIONS(6716),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6716),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6716),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6716),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6716),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6716),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6716),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6716),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6716),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6716),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6716),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6714),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6716),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6716),
    [anon_sym_API_AVAILABLE] = ACTIONS(6716),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6716),
    [anon_sym_API_DEPRECATED] = ACTIONS(6716),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6716),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6716),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6716),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6716),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6716),
    [anon_sym___deprecated_msg] = ACTIONS(6716),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6716),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6716),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6716),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6716),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6716),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6716),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6716),
  },
  [STATE(3234)] = {
    [sym_identifier] = ACTIONS(8046),
    [anon_sym_SEMI] = ACTIONS(8048),
    [anon_sym___extension__] = ACTIONS(8046),
    [anon_sym_extern] = ACTIONS(8046),
    [anon_sym___attribute__] = ACTIONS(8046),
    [anon_sym___attribute] = ACTIONS(8046),
    [anon_sym_noreturn] = ACTIONS(8046),
    [anon_sym_LBRACK] = ACTIONS(8048),
    [anon_sym___declspec] = ACTIONS(8046),
    [anon_sym_static] = ACTIONS(8046),
    [anon_sym_auto] = ACTIONS(8046),
    [anon_sym_register] = ACTIONS(8046),
    [anon_sym_inline] = ACTIONS(8046),
    [anon_sym___inline] = ACTIONS(8046),
    [anon_sym___inline__] = ACTIONS(8046),
    [anon_sym___forceinline] = ACTIONS(8046),
    [anon_sym_thread_local] = ACTIONS(8046),
    [anon_sym___thread] = ACTIONS(8046),
    [anon_sym_CG_EXTERN] = ACTIONS(8046),
    [anon_sym_CG_INLINE] = ACTIONS(8046),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8046),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8046),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8046),
    [anon_sym_IBOutlet] = ACTIONS(8046),
    [anon_sym_IBInspectable] = ACTIONS(8046),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8046),
    [anon_sym_NS_INLINE] = ACTIONS(8046),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8046),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8046),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8046),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8046),
    [anon_sym_const] = ACTIONS(8046),
    [anon_sym_constexpr] = ACTIONS(8046),
    [anon_sym_volatile] = ACTIONS(8046),
    [anon_sym_restrict] = ACTIONS(8046),
    [anon_sym___restrict__] = ACTIONS(8046),
    [anon_sym__Atomic] = ACTIONS(8046),
    [anon_sym__Noreturn] = ACTIONS(8046),
    [anon_sym__Nonnull] = ACTIONS(8046),
    [anon_sym_nullable] = ACTIONS(8046),
    [anon_sym__Complex] = ACTIONS(8046),
    [anon_sym__Nullable] = ACTIONS(8046),
    [anon_sym__Nullable_result] = ACTIONS(8046),
    [anon_sym__Null_unspecified] = ACTIONS(8046),
    [anon_sym___autoreleasing] = ACTIONS(8046),
    [anon_sym___block] = ACTIONS(8046),
    [anon_sym___bridge] = ACTIONS(8046),
    [anon_sym___bridge_retained] = ACTIONS(8046),
    [anon_sym___bridge_transfer] = ACTIONS(8046),
    [anon_sym___complex] = ACTIONS(8046),
    [anon_sym___const] = ACTIONS(8046),
    [anon_sym___imag] = ACTIONS(8046),
    [anon_sym___kindof] = ACTIONS(8046),
    [anon_sym___nonnull] = ACTIONS(8046),
    [anon_sym___nullable] = ACTIONS(8046),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8046),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8046),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8046),
    [anon_sym___real] = ACTIONS(8046),
    [anon_sym___strong] = ACTIONS(8046),
    [anon_sym___unsafe_unretained] = ACTIONS(8046),
    [anon_sym___unused] = ACTIONS(8046),
    [anon_sym___weak] = ACTIONS(8046),
    [anon_sym_alignas] = ACTIONS(8046),
    [anon_sym__Alignas] = ACTIONS(8046),
    [anon_sym_COLON] = ACTIONS(8048),
    [anon_sym_struct] = ACTIONS(8046),
    [anon_sym_in] = ACTIONS(8046),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8046),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8046),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8046),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8046),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8046),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8046),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8046),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8046),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8046),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8046),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8046),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8046),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8046),
    [anon_sym_API_AVAILABLE] = ACTIONS(8046),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8046),
    [anon_sym_API_DEPRECATED] = ACTIONS(8046),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8046),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8046),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8046),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8046),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8046),
    [anon_sym___deprecated_msg] = ACTIONS(8046),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8046),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8046),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8046),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8046),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8046),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8046),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8046),
    [anon_sym_id] = ACTIONS(8046),
  },
  [STATE(3235)] = {
    [sym_identifier] = ACTIONS(5074),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym_SEMI] = ACTIONS(8050),
    [anon_sym___extension__] = ACTIONS(5074),
    [anon_sym_extern] = ACTIONS(5074),
    [anon_sym___attribute__] = ACTIONS(5074),
    [anon_sym___attribute] = ACTIONS(5074),
    [anon_sym_noreturn] = ACTIONS(5074),
    [anon_sym_LBRACK] = ACTIONS(5070),
    [anon_sym___declspec] = ACTIONS(5074),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym_static] = ACTIONS(5074),
    [anon_sym_auto] = ACTIONS(5074),
    [anon_sym_register] = ACTIONS(5074),
    [anon_sym_inline] = ACTIONS(5074),
    [anon_sym___inline] = ACTIONS(5074),
    [anon_sym___inline__] = ACTIONS(5074),
    [anon_sym___forceinline] = ACTIONS(5074),
    [anon_sym_thread_local] = ACTIONS(5074),
    [anon_sym___thread] = ACTIONS(5074),
    [anon_sym_CG_EXTERN] = ACTIONS(5074),
    [anon_sym_CG_INLINE] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5074),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5074),
    [anon_sym_IBOutlet] = ACTIONS(5074),
    [anon_sym_IBInspectable] = ACTIONS(5074),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5074),
    [anon_sym_NS_INLINE] = ACTIONS(5074),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5074),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5074),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5074),
    [anon_sym_const] = ACTIONS(5074),
    [anon_sym_constexpr] = ACTIONS(5074),
    [anon_sym_volatile] = ACTIONS(5074),
    [anon_sym_restrict] = ACTIONS(5074),
    [anon_sym___restrict__] = ACTIONS(5074),
    [anon_sym__Atomic] = ACTIONS(5074),
    [anon_sym__Noreturn] = ACTIONS(5074),
    [anon_sym__Nonnull] = ACTIONS(5074),
    [anon_sym_nullable] = ACTIONS(5074),
    [anon_sym__Complex] = ACTIONS(5074),
    [anon_sym__Nullable] = ACTIONS(5074),
    [anon_sym__Nullable_result] = ACTIONS(5074),
    [anon_sym__Null_unspecified] = ACTIONS(5074),
    [anon_sym___autoreleasing] = ACTIONS(5074),
    [anon_sym___block] = ACTIONS(5074),
    [anon_sym___bridge] = ACTIONS(5074),
    [anon_sym___bridge_retained] = ACTIONS(5074),
    [anon_sym___bridge_transfer] = ACTIONS(5074),
    [anon_sym___complex] = ACTIONS(5074),
    [anon_sym___const] = ACTIONS(5074),
    [anon_sym___imag] = ACTIONS(5074),
    [anon_sym___kindof] = ACTIONS(5074),
    [anon_sym___nonnull] = ACTIONS(5074),
    [anon_sym___nullable] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5074),
    [anon_sym___real] = ACTIONS(5074),
    [anon_sym___strong] = ACTIONS(5074),
    [anon_sym___unsafe_unretained] = ACTIONS(5074),
    [anon_sym___unused] = ACTIONS(5074),
    [anon_sym___weak] = ACTIONS(5074),
    [anon_sym_alignas] = ACTIONS(5074),
    [anon_sym__Alignas] = ACTIONS(5074),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5074),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5074),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5074),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_API_AVAILABLE] = ACTIONS(5074),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_API_DEPRECATED] = ACTIONS(5074),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym___deprecated_msg] = ACTIONS(5074),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5074),
  },
  [STATE(3236)] = {
    [sym_attribute_specifier] = STATE(3241),
    [sym_method_parameter] = STATE(3259),
    [sym_identifier] = ACTIONS(8052),
    [anon_sym_COMMA] = ACTIONS(8054),
    [anon_sym_SEMI] = ACTIONS(8054),
    [anon_sym___extension__] = ACTIONS(8052),
    [anon_sym_extern] = ACTIONS(8052),
    [anon_sym___attribute__] = ACTIONS(8056),
    [anon_sym___attribute] = ACTIONS(8056),
    [anon_sym_noreturn] = ACTIONS(8052),
    [anon_sym_LBRACK] = ACTIONS(8054),
    [anon_sym___declspec] = ACTIONS(8052),
    [anon_sym_static] = ACTIONS(8052),
    [anon_sym_auto] = ACTIONS(8052),
    [anon_sym_register] = ACTIONS(8052),
    [anon_sym_inline] = ACTIONS(8052),
    [anon_sym___inline] = ACTIONS(8052),
    [anon_sym___inline__] = ACTIONS(8052),
    [anon_sym___forceinline] = ACTIONS(8052),
    [anon_sym_thread_local] = ACTIONS(8052),
    [anon_sym___thread] = ACTIONS(8052),
    [anon_sym_CG_EXTERN] = ACTIONS(8052),
    [anon_sym_CG_INLINE] = ACTIONS(8052),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8052),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8052),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8052),
    [anon_sym_IBOutlet] = ACTIONS(8052),
    [anon_sym_IBInspectable] = ACTIONS(8052),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8052),
    [anon_sym_NS_INLINE] = ACTIONS(8052),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8052),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8052),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8052),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8052),
    [anon_sym_const] = ACTIONS(8052),
    [anon_sym_constexpr] = ACTIONS(8052),
    [anon_sym_volatile] = ACTIONS(8052),
    [anon_sym_restrict] = ACTIONS(8052),
    [anon_sym___restrict__] = ACTIONS(8052),
    [anon_sym__Atomic] = ACTIONS(8052),
    [anon_sym__Noreturn] = ACTIONS(8052),
    [anon_sym__Nonnull] = ACTIONS(8052),
    [anon_sym_nullable] = ACTIONS(8052),
    [anon_sym__Complex] = ACTIONS(8052),
    [anon_sym__Nullable] = ACTIONS(8052),
    [anon_sym__Nullable_result] = ACTIONS(8052),
    [anon_sym__Null_unspecified] = ACTIONS(8052),
    [anon_sym___autoreleasing] = ACTIONS(8052),
    [anon_sym___block] = ACTIONS(8052),
    [anon_sym___bridge] = ACTIONS(8052),
    [anon_sym___bridge_retained] = ACTIONS(8052),
    [anon_sym___bridge_transfer] = ACTIONS(8052),
    [anon_sym___complex] = ACTIONS(8052),
    [anon_sym___const] = ACTIONS(8052),
    [anon_sym___imag] = ACTIONS(8052),
    [anon_sym___kindof] = ACTIONS(8052),
    [anon_sym___nonnull] = ACTIONS(8052),
    [anon_sym___nullable] = ACTIONS(8052),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8052),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8052),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8052),
    [anon_sym___real] = ACTIONS(8052),
    [anon_sym___strong] = ACTIONS(8052),
    [anon_sym___unsafe_unretained] = ACTIONS(8052),
    [anon_sym___unused] = ACTIONS(8052),
    [anon_sym___weak] = ACTIONS(8052),
    [anon_sym_alignas] = ACTIONS(8052),
    [anon_sym__Alignas] = ACTIONS(8052),
    [anon_sym_COLON] = ACTIONS(8059),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8052),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8052),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8052),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8052),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8052),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8052),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8052),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8052),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8052),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_API_AVAILABLE] = ACTIONS(8052),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_API_DEPRECATED] = ACTIONS(8052),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8052),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8052),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8052),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8052),
    [anon_sym___deprecated_msg] = ACTIONS(8052),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8052),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8052),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8052),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8052),
  },
  [STATE(3237)] = {
    [sym_identifier] = ACTIONS(5074),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym_SEMI] = ACTIONS(8062),
    [anon_sym___extension__] = ACTIONS(5074),
    [anon_sym_extern] = ACTIONS(5074),
    [anon_sym___attribute__] = ACTIONS(5074),
    [anon_sym___attribute] = ACTIONS(5074),
    [anon_sym_noreturn] = ACTIONS(5074),
    [anon_sym_LBRACK] = ACTIONS(5070),
    [anon_sym___declspec] = ACTIONS(5074),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym_static] = ACTIONS(5074),
    [anon_sym_auto] = ACTIONS(5074),
    [anon_sym_register] = ACTIONS(5074),
    [anon_sym_inline] = ACTIONS(5074),
    [anon_sym___inline] = ACTIONS(5074),
    [anon_sym___inline__] = ACTIONS(5074),
    [anon_sym___forceinline] = ACTIONS(5074),
    [anon_sym_thread_local] = ACTIONS(5074),
    [anon_sym___thread] = ACTIONS(5074),
    [anon_sym_CG_EXTERN] = ACTIONS(5074),
    [anon_sym_CG_INLINE] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(5074),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(5074),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(5074),
    [anon_sym_IBOutlet] = ACTIONS(5074),
    [anon_sym_IBInspectable] = ACTIONS(5074),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(5074),
    [anon_sym_NS_INLINE] = ACTIONS(5074),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(5074),
    [anon_sym_OBJC_EXPORT] = ACTIONS(5074),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(5074),
    [anon_sym_const] = ACTIONS(5074),
    [anon_sym_constexpr] = ACTIONS(5074),
    [anon_sym_volatile] = ACTIONS(5074),
    [anon_sym_restrict] = ACTIONS(5074),
    [anon_sym___restrict__] = ACTIONS(5074),
    [anon_sym__Atomic] = ACTIONS(5074),
    [anon_sym__Noreturn] = ACTIONS(5074),
    [anon_sym__Nonnull] = ACTIONS(5074),
    [anon_sym_nullable] = ACTIONS(5074),
    [anon_sym__Complex] = ACTIONS(5074),
    [anon_sym__Nullable] = ACTIONS(5074),
    [anon_sym__Nullable_result] = ACTIONS(5074),
    [anon_sym__Null_unspecified] = ACTIONS(5074),
    [anon_sym___autoreleasing] = ACTIONS(5074),
    [anon_sym___block] = ACTIONS(5074),
    [anon_sym___bridge] = ACTIONS(5074),
    [anon_sym___bridge_retained] = ACTIONS(5074),
    [anon_sym___bridge_transfer] = ACTIONS(5074),
    [anon_sym___complex] = ACTIONS(5074),
    [anon_sym___const] = ACTIONS(5074),
    [anon_sym___imag] = ACTIONS(5074),
    [anon_sym___kindof] = ACTIONS(5074),
    [anon_sym___nonnull] = ACTIONS(5074),
    [anon_sym___nullable] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5074),
    [anon_sym___real] = ACTIONS(5074),
    [anon_sym___strong] = ACTIONS(5074),
    [anon_sym___unsafe_unretained] = ACTIONS(5074),
    [anon_sym___unused] = ACTIONS(5074),
    [anon_sym___weak] = ACTIONS(5074),
    [anon_sym_alignas] = ACTIONS(5074),
    [anon_sym__Alignas] = ACTIONS(5074),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(5074),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(5074),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(5074),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(5074),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE] = ACTIONS(5074),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_API_AVAILABLE] = ACTIONS(5074),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_API_DEPRECATED] = ACTIONS(5074),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(5074),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(5074),
    [anon_sym___deprecated_msg] = ACTIONS(5074),
    [anon_sym___deprecated_enum_msg] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(5074),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(5074),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(5074),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(5074),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(5074),
  },
  [STATE(3238)] = {
    [sym_identifier] = ACTIONS(8064),
    [anon_sym_SEMI] = ACTIONS(8066),
    [anon_sym___extension__] = ACTIONS(8064),
    [anon_sym_extern] = ACTIONS(8064),
    [anon_sym___attribute__] = ACTIONS(8064),
    [anon_sym___attribute] = ACTIONS(8064),
    [anon_sym_noreturn] = ACTIONS(8064),
    [anon_sym_LBRACK] = ACTIONS(8066),
    [anon_sym___declspec] = ACTIONS(8064),
    [anon_sym_static] = ACTIONS(8064),
    [anon_sym_auto] = ACTIONS(8064),
    [anon_sym_register] = ACTIONS(8064),
    [anon_sym_inline] = ACTIONS(8064),
    [anon_sym___inline] = ACTIONS(8064),
    [anon_sym___inline__] = ACTIONS(8064),
    [anon_sym___forceinline] = ACTIONS(8064),
    [anon_sym_thread_local] = ACTIONS(8064),
    [anon_sym___thread] = ACTIONS(8064),
    [anon_sym_CG_EXTERN] = ACTIONS(8064),
    [anon_sym_CG_INLINE] = ACTIONS(8064),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8064),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8064),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8064),
    [anon_sym_IBOutlet] = ACTIONS(8064),
    [anon_sym_IBInspectable] = ACTIONS(8064),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8064),
    [anon_sym_NS_INLINE] = ACTIONS(8064),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8064),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8064),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8064),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8064),
    [anon_sym_const] = ACTIONS(8064),
    [anon_sym_constexpr] = ACTIONS(8064),
    [anon_sym_volatile] = ACTIONS(8064),
    [anon_sym_restrict] = ACTIONS(8064),
    [anon_sym___restrict__] = ACTIONS(8064),
    [anon_sym__Atomic] = ACTIONS(8064),
    [anon_sym__Noreturn] = ACTIONS(8064),
    [anon_sym__Nonnull] = ACTIONS(8064),
    [anon_sym_nullable] = ACTIONS(8064),
    [anon_sym__Complex] = ACTIONS(8064),
    [anon_sym__Nullable] = ACTIONS(8064),
    [anon_sym__Nullable_result] = ACTIONS(8064),
    [anon_sym__Null_unspecified] = ACTIONS(8064),
    [anon_sym___autoreleasing] = ACTIONS(8064),
    [anon_sym___block] = ACTIONS(8064),
    [anon_sym___bridge] = ACTIONS(8064),
    [anon_sym___bridge_retained] = ACTIONS(8064),
    [anon_sym___bridge_transfer] = ACTIONS(8064),
    [anon_sym___complex] = ACTIONS(8064),
    [anon_sym___const] = ACTIONS(8064),
    [anon_sym___imag] = ACTIONS(8064),
    [anon_sym___kindof] = ACTIONS(8064),
    [anon_sym___nonnull] = ACTIONS(8064),
    [anon_sym___nullable] = ACTIONS(8064),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8064),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8064),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8064),
    [anon_sym___real] = ACTIONS(8064),
    [anon_sym___strong] = ACTIONS(8064),
    [anon_sym___unsafe_unretained] = ACTIONS(8064),
    [anon_sym___unused] = ACTIONS(8064),
    [anon_sym___weak] = ACTIONS(8064),
    [anon_sym_alignas] = ACTIONS(8064),
    [anon_sym__Alignas] = ACTIONS(8064),
    [anon_sym_COLON] = ACTIONS(8066),
    [anon_sym_struct] = ACTIONS(8064),
    [anon_sym_in] = ACTIONS(8064),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8064),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8064),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8064),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8064),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8064),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8064),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8064),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8064),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8064),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8064),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8064),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8064),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8064),
    [anon_sym_API_AVAILABLE] = ACTIONS(8064),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8064),
    [anon_sym_API_DEPRECATED] = ACTIONS(8064),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8064),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8064),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8064),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8064),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8064),
    [anon_sym___deprecated_msg] = ACTIONS(8064),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8064),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8064),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8064),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8064),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8064),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8064),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8064),
    [anon_sym_id] = ACTIONS(8064),
  },
  [STATE(3239)] = {
    [sym_method_parameter] = STATE(3254),
    [sym_identifier] = ACTIONS(7997),
    [anon_sym_COMMA] = ACTIONS(7992),
    [anon_sym_SEMI] = ACTIONS(7992),
    [anon_sym___extension__] = ACTIONS(7997),
    [anon_sym_extern] = ACTIONS(7997),
    [anon_sym___attribute__] = ACTIONS(7997),
    [anon_sym___attribute] = ACTIONS(7997),
    [anon_sym_noreturn] = ACTIONS(7997),
    [anon_sym_LBRACK] = ACTIONS(7992),
    [anon_sym___declspec] = ACTIONS(7997),
    [anon_sym_static] = ACTIONS(7997),
    [anon_sym_auto] = ACTIONS(7997),
    [anon_sym_register] = ACTIONS(7997),
    [anon_sym_inline] = ACTIONS(7997),
    [anon_sym___inline] = ACTIONS(7997),
    [anon_sym___inline__] = ACTIONS(7997),
    [anon_sym___forceinline] = ACTIONS(7997),
    [anon_sym_thread_local] = ACTIONS(7997),
    [anon_sym___thread] = ACTIONS(7997),
    [anon_sym_CG_EXTERN] = ACTIONS(7997),
    [anon_sym_CG_INLINE] = ACTIONS(7997),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7997),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7997),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7997),
    [anon_sym_IBOutlet] = ACTIONS(7997),
    [anon_sym_IBInspectable] = ACTIONS(7997),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7997),
    [anon_sym_NS_INLINE] = ACTIONS(7997),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7997),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7997),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7997),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7997),
    [anon_sym_const] = ACTIONS(7997),
    [anon_sym_constexpr] = ACTIONS(7997),
    [anon_sym_volatile] = ACTIONS(7997),
    [anon_sym_restrict] = ACTIONS(7997),
    [anon_sym___restrict__] = ACTIONS(7997),
    [anon_sym__Atomic] = ACTIONS(7997),
    [anon_sym__Noreturn] = ACTIONS(7997),
    [anon_sym__Nonnull] = ACTIONS(7997),
    [anon_sym_nullable] = ACTIONS(7997),
    [anon_sym__Complex] = ACTIONS(7997),
    [anon_sym__Nullable] = ACTIONS(7997),
    [anon_sym__Nullable_result] = ACTIONS(7997),
    [anon_sym__Null_unspecified] = ACTIONS(7997),
    [anon_sym___autoreleasing] = ACTIONS(7997),
    [anon_sym___block] = ACTIONS(7997),
    [anon_sym___bridge] = ACTIONS(7997),
    [anon_sym___bridge_retained] = ACTIONS(7997),
    [anon_sym___bridge_transfer] = ACTIONS(7997),
    [anon_sym___complex] = ACTIONS(7997),
    [anon_sym___const] = ACTIONS(7997),
    [anon_sym___imag] = ACTIONS(7997),
    [anon_sym___kindof] = ACTIONS(7997),
    [anon_sym___nonnull] = ACTIONS(7997),
    [anon_sym___nullable] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7997),
    [anon_sym___real] = ACTIONS(7997),
    [anon_sym___strong] = ACTIONS(7997),
    [anon_sym___unsafe_unretained] = ACTIONS(7997),
    [anon_sym___unused] = ACTIONS(7997),
    [anon_sym___weak] = ACTIONS(7997),
    [anon_sym_alignas] = ACTIONS(7997),
    [anon_sym__Alignas] = ACTIONS(7997),
    [anon_sym_COLON] = ACTIONS(7999),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7997),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7997),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7997),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7997),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7997),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7997),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7997),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_API_AVAILABLE] = ACTIONS(7997),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_API_DEPRECATED] = ACTIONS(7997),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7997),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym___deprecated_msg] = ACTIONS(7997),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7997),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7997),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7997),
  },
  [STATE(3240)] = {
    [sym_method_parameter] = STATE(3259),
    [sym_identifier] = ACTIONS(8052),
    [anon_sym_COMMA] = ACTIONS(8054),
    [anon_sym_SEMI] = ACTIONS(8054),
    [anon_sym___extension__] = ACTIONS(8052),
    [anon_sym_extern] = ACTIONS(8052),
    [anon_sym___attribute__] = ACTIONS(8052),
    [anon_sym___attribute] = ACTIONS(8052),
    [anon_sym_noreturn] = ACTIONS(8052),
    [anon_sym_LBRACK] = ACTIONS(8054),
    [anon_sym___declspec] = ACTIONS(8052),
    [anon_sym_static] = ACTIONS(8052),
    [anon_sym_auto] = ACTIONS(8052),
    [anon_sym_register] = ACTIONS(8052),
    [anon_sym_inline] = ACTIONS(8052),
    [anon_sym___inline] = ACTIONS(8052),
    [anon_sym___inline__] = ACTIONS(8052),
    [anon_sym___forceinline] = ACTIONS(8052),
    [anon_sym_thread_local] = ACTIONS(8052),
    [anon_sym___thread] = ACTIONS(8052),
    [anon_sym_CG_EXTERN] = ACTIONS(8052),
    [anon_sym_CG_INLINE] = ACTIONS(8052),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8052),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8052),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8052),
    [anon_sym_IBOutlet] = ACTIONS(8052),
    [anon_sym_IBInspectable] = ACTIONS(8052),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8052),
    [anon_sym_NS_INLINE] = ACTIONS(8052),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8052),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8052),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8052),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8052),
    [anon_sym_const] = ACTIONS(8052),
    [anon_sym_constexpr] = ACTIONS(8052),
    [anon_sym_volatile] = ACTIONS(8052),
    [anon_sym_restrict] = ACTIONS(8052),
    [anon_sym___restrict__] = ACTIONS(8052),
    [anon_sym__Atomic] = ACTIONS(8052),
    [anon_sym__Noreturn] = ACTIONS(8052),
    [anon_sym__Nonnull] = ACTIONS(8052),
    [anon_sym_nullable] = ACTIONS(8052),
    [anon_sym__Complex] = ACTIONS(8052),
    [anon_sym__Nullable] = ACTIONS(8052),
    [anon_sym__Nullable_result] = ACTIONS(8052),
    [anon_sym__Null_unspecified] = ACTIONS(8052),
    [anon_sym___autoreleasing] = ACTIONS(8052),
    [anon_sym___block] = ACTIONS(8052),
    [anon_sym___bridge] = ACTIONS(8052),
    [anon_sym___bridge_retained] = ACTIONS(8052),
    [anon_sym___bridge_transfer] = ACTIONS(8052),
    [anon_sym___complex] = ACTIONS(8052),
    [anon_sym___const] = ACTIONS(8052),
    [anon_sym___imag] = ACTIONS(8052),
    [anon_sym___kindof] = ACTIONS(8052),
    [anon_sym___nonnull] = ACTIONS(8052),
    [anon_sym___nullable] = ACTIONS(8052),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8052),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8052),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8052),
    [anon_sym___real] = ACTIONS(8052),
    [anon_sym___strong] = ACTIONS(8052),
    [anon_sym___unsafe_unretained] = ACTIONS(8052),
    [anon_sym___unused] = ACTIONS(8052),
    [anon_sym___weak] = ACTIONS(8052),
    [anon_sym_alignas] = ACTIONS(8052),
    [anon_sym__Alignas] = ACTIONS(8052),
    [anon_sym_COLON] = ACTIONS(8059),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8052),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8052),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8052),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8052),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8052),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8052),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8052),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8052),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8052),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_API_AVAILABLE] = ACTIONS(8052),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_API_DEPRECATED] = ACTIONS(8052),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8052),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8052),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8052),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8052),
    [anon_sym___deprecated_msg] = ACTIONS(8052),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8052),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8052),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8052),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8052),
  },
  [STATE(3241)] = {
    [sym_method_parameter] = STATE(3249),
    [sym_identifier] = ACTIONS(8068),
    [anon_sym_COMMA] = ACTIONS(8070),
    [anon_sym_SEMI] = ACTIONS(8070),
    [anon_sym___extension__] = ACTIONS(8068),
    [anon_sym_extern] = ACTIONS(8068),
    [anon_sym___attribute__] = ACTIONS(8068),
    [anon_sym___attribute] = ACTIONS(8068),
    [anon_sym_noreturn] = ACTIONS(8068),
    [anon_sym_LBRACK] = ACTIONS(8070),
    [anon_sym___declspec] = ACTIONS(8068),
    [anon_sym_static] = ACTIONS(8068),
    [anon_sym_auto] = ACTIONS(8068),
    [anon_sym_register] = ACTIONS(8068),
    [anon_sym_inline] = ACTIONS(8068),
    [anon_sym___inline] = ACTIONS(8068),
    [anon_sym___inline__] = ACTIONS(8068),
    [anon_sym___forceinline] = ACTIONS(8068),
    [anon_sym_thread_local] = ACTIONS(8068),
    [anon_sym___thread] = ACTIONS(8068),
    [anon_sym_CG_EXTERN] = ACTIONS(8068),
    [anon_sym_CG_INLINE] = ACTIONS(8068),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8068),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8068),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8068),
    [anon_sym_IBOutlet] = ACTIONS(8068),
    [anon_sym_IBInspectable] = ACTIONS(8068),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8068),
    [anon_sym_NS_INLINE] = ACTIONS(8068),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8068),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8068),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8068),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8068),
    [anon_sym_const] = ACTIONS(8068),
    [anon_sym_constexpr] = ACTIONS(8068),
    [anon_sym_volatile] = ACTIONS(8068),
    [anon_sym_restrict] = ACTIONS(8068),
    [anon_sym___restrict__] = ACTIONS(8068),
    [anon_sym__Atomic] = ACTIONS(8068),
    [anon_sym__Noreturn] = ACTIONS(8068),
    [anon_sym__Nonnull] = ACTIONS(8068),
    [anon_sym_nullable] = ACTIONS(8068),
    [anon_sym__Complex] = ACTIONS(8068),
    [anon_sym__Nullable] = ACTIONS(8068),
    [anon_sym__Nullable_result] = ACTIONS(8068),
    [anon_sym__Null_unspecified] = ACTIONS(8068),
    [anon_sym___autoreleasing] = ACTIONS(8068),
    [anon_sym___block] = ACTIONS(8068),
    [anon_sym___bridge] = ACTIONS(8068),
    [anon_sym___bridge_retained] = ACTIONS(8068),
    [anon_sym___bridge_transfer] = ACTIONS(8068),
    [anon_sym___complex] = ACTIONS(8068),
    [anon_sym___const] = ACTIONS(8068),
    [anon_sym___imag] = ACTIONS(8068),
    [anon_sym___kindof] = ACTIONS(8068),
    [anon_sym___nonnull] = ACTIONS(8068),
    [anon_sym___nullable] = ACTIONS(8068),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8068),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8068),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8068),
    [anon_sym___real] = ACTIONS(8068),
    [anon_sym___strong] = ACTIONS(8068),
    [anon_sym___unsafe_unretained] = ACTIONS(8068),
    [anon_sym___unused] = ACTIONS(8068),
    [anon_sym___weak] = ACTIONS(8068),
    [anon_sym_alignas] = ACTIONS(8068),
    [anon_sym__Alignas] = ACTIONS(8068),
    [anon_sym_COLON] = ACTIONS(8072),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8068),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8068),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8068),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8068),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8068),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8068),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8068),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8068),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8068),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8068),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8068),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8068),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8068),
    [anon_sym_API_AVAILABLE] = ACTIONS(8068),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8068),
    [anon_sym_API_DEPRECATED] = ACTIONS(8068),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8068),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8068),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8068),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8068),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8068),
    [anon_sym___deprecated_msg] = ACTIONS(8068),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8068),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8068),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8068),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8068),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8068),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8068),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8068),
  },
  [STATE(3242)] = {
    [sym_parameter_list] = STATE(3007),
    [anon_sym_COMMA] = ACTIONS(8075),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(8075),
    [anon_sym___extension__] = ACTIONS(8075),
    [anon_sym_extern] = ACTIONS(8075),
    [anon_sym___attribute__] = ACTIONS(8075),
    [anon_sym___attribute] = ACTIONS(8077),
    [anon_sym_noreturn] = ACTIONS(8075),
    [anon_sym_LBRACK] = ACTIONS(7611),
    [anon_sym___declspec] = ACTIONS(8075),
    [anon_sym_static] = ACTIONS(8075),
    [anon_sym_auto] = ACTIONS(8075),
    [anon_sym_register] = ACTIONS(8075),
    [anon_sym_inline] = ACTIONS(8075),
    [anon_sym___inline] = ACTIONS(8077),
    [anon_sym___inline__] = ACTIONS(8075),
    [anon_sym___forceinline] = ACTIONS(8075),
    [anon_sym_thread_local] = ACTIONS(8075),
    [anon_sym___thread] = ACTIONS(8075),
    [anon_sym_CG_EXTERN] = ACTIONS(8075),
    [anon_sym_CG_INLINE] = ACTIONS(8075),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8075),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8075),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8075),
    [anon_sym_IBOutlet] = ACTIONS(8075),
    [anon_sym_IBInspectable] = ACTIONS(8075),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8075),
    [anon_sym_NS_INLINE] = ACTIONS(8075),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8075),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8075),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8075),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8075),
    [anon_sym_const] = ACTIONS(8077),
    [anon_sym_constexpr] = ACTIONS(8075),
    [anon_sym_volatile] = ACTIONS(8075),
    [anon_sym_restrict] = ACTIONS(8075),
    [anon_sym___restrict__] = ACTIONS(8075),
    [anon_sym__Atomic] = ACTIONS(8075),
    [anon_sym__Noreturn] = ACTIONS(8075),
    [anon_sym__Nonnull] = ACTIONS(8075),
    [anon_sym_nullable] = ACTIONS(8075),
    [anon_sym__Complex] = ACTIONS(8075),
    [anon_sym__Nullable] = ACTIONS(8077),
    [anon_sym__Nullable_result] = ACTIONS(8075),
    [anon_sym__Null_unspecified] = ACTIONS(8075),
    [anon_sym___autoreleasing] = ACTIONS(8075),
    [anon_sym___block] = ACTIONS(8075),
    [anon_sym___bridge] = ACTIONS(8077),
    [anon_sym___bridge_retained] = ACTIONS(8075),
    [anon_sym___bridge_transfer] = ACTIONS(8075),
    [anon_sym___complex] = ACTIONS(8075),
    [anon_sym___const] = ACTIONS(8075),
    [anon_sym___imag] = ACTIONS(8075),
    [anon_sym___kindof] = ACTIONS(8075),
    [anon_sym___nonnull] = ACTIONS(8075),
    [anon_sym___nullable] = ACTIONS(8075),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8075),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8075),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8075),
    [anon_sym___real] = ACTIONS(8075),
    [anon_sym___strong] = ACTIONS(8075),
    [anon_sym___unsafe_unretained] = ACTIONS(8075),
    [anon_sym___unused] = ACTIONS(8075),
    [anon_sym___weak] = ACTIONS(8075),
    [anon_sym_alignas] = ACTIONS(8075),
    [anon_sym__Alignas] = ACTIONS(8075),
    [anon_sym_COLON] = ACTIONS(8079),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8075),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8075),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8075),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8075),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8075),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8075),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8075),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8075),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8075),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8075),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8077),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8075),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8075),
    [anon_sym_API_AVAILABLE] = ACTIONS(8075),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8075),
    [anon_sym_API_DEPRECATED] = ACTIONS(8075),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8075),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8075),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8075),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8075),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8075),
    [anon_sym___deprecated_msg] = ACTIONS(8075),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8075),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8075),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8075),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8075),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8075),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8075),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8075),
  },
  [STATE(3243)] = {
    [anon_sym_COMMA] = ACTIONS(8081),
    [anon_sym_RPAREN] = ACTIONS(8081),
    [anon_sym_LPAREN2] = ACTIONS(8081),
    [anon_sym_SEMI] = ACTIONS(8081),
    [anon_sym___extension__] = ACTIONS(8081),
    [anon_sym_extern] = ACTIONS(8081),
    [anon_sym___attribute__] = ACTIONS(8081),
    [anon_sym___attribute] = ACTIONS(8083),
    [anon_sym_noreturn] = ACTIONS(8081),
    [anon_sym_LBRACK] = ACTIONS(8081),
    [anon_sym___declspec] = ACTIONS(8081),
    [anon_sym_static] = ACTIONS(8081),
    [anon_sym_auto] = ACTIONS(8081),
    [anon_sym_register] = ACTIONS(8081),
    [anon_sym_inline] = ACTIONS(8081),
    [anon_sym___inline] = ACTIONS(8083),
    [anon_sym___inline__] = ACTIONS(8081),
    [anon_sym___forceinline] = ACTIONS(8081),
    [anon_sym_thread_local] = ACTIONS(8081),
    [anon_sym___thread] = ACTIONS(8081),
    [anon_sym_CG_EXTERN] = ACTIONS(8081),
    [anon_sym_CG_INLINE] = ACTIONS(8081),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8081),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8081),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8081),
    [anon_sym_IBOutlet] = ACTIONS(8081),
    [anon_sym_IBInspectable] = ACTIONS(8081),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8081),
    [anon_sym_NS_INLINE] = ACTIONS(8081),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8081),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8081),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8081),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8081),
    [anon_sym_const] = ACTIONS(8083),
    [anon_sym_constexpr] = ACTIONS(8081),
    [anon_sym_volatile] = ACTIONS(8081),
    [anon_sym_restrict] = ACTIONS(8081),
    [anon_sym___restrict__] = ACTIONS(8081),
    [anon_sym__Atomic] = ACTIONS(8081),
    [anon_sym__Noreturn] = ACTIONS(8081),
    [anon_sym__Nonnull] = ACTIONS(8081),
    [anon_sym_nullable] = ACTIONS(8081),
    [anon_sym__Complex] = ACTIONS(8081),
    [anon_sym__Nullable] = ACTIONS(8083),
    [anon_sym__Nullable_result] = ACTIONS(8081),
    [anon_sym__Null_unspecified] = ACTIONS(8081),
    [anon_sym___autoreleasing] = ACTIONS(8081),
    [anon_sym___block] = ACTIONS(8081),
    [anon_sym___bridge] = ACTIONS(8083),
    [anon_sym___bridge_retained] = ACTIONS(8081),
    [anon_sym___bridge_transfer] = ACTIONS(8081),
    [anon_sym___complex] = ACTIONS(8081),
    [anon_sym___const] = ACTIONS(8081),
    [anon_sym___imag] = ACTIONS(8081),
    [anon_sym___kindof] = ACTIONS(8081),
    [anon_sym___nonnull] = ACTIONS(8081),
    [anon_sym___nullable] = ACTIONS(8081),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8081),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8081),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8081),
    [anon_sym___real] = ACTIONS(8081),
    [anon_sym___strong] = ACTIONS(8081),
    [anon_sym___unsafe_unretained] = ACTIONS(8081),
    [anon_sym___unused] = ACTIONS(8081),
    [anon_sym___weak] = ACTIONS(8081),
    [anon_sym_alignas] = ACTIONS(8081),
    [anon_sym__Alignas] = ACTIONS(8081),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8081),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8081),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8081),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8081),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8081),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8081),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8081),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8081),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8081),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8081),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8083),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8081),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8081),
    [anon_sym_API_AVAILABLE] = ACTIONS(8081),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8081),
    [anon_sym_API_DEPRECATED] = ACTIONS(8081),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8081),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8081),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8081),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8081),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8081),
    [anon_sym___deprecated_msg] = ACTIONS(8081),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8081),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8081),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8081),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8081),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8081),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8081),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8081),
  },
  [STATE(3244)] = {
    [sym_parameter_list] = STATE(3223),
    [anon_sym_COMMA] = ACTIONS(8085),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(8085),
    [anon_sym___extension__] = ACTIONS(8085),
    [anon_sym_extern] = ACTIONS(8085),
    [anon_sym___attribute__] = ACTIONS(8085),
    [anon_sym___attribute] = ACTIONS(8087),
    [anon_sym_noreturn] = ACTIONS(8085),
    [anon_sym_LBRACK] = ACTIONS(8089),
    [anon_sym___declspec] = ACTIONS(8085),
    [anon_sym_static] = ACTIONS(8085),
    [anon_sym_auto] = ACTIONS(8085),
    [anon_sym_register] = ACTIONS(8085),
    [anon_sym_inline] = ACTIONS(8085),
    [anon_sym___inline] = ACTIONS(8087),
    [anon_sym___inline__] = ACTIONS(8085),
    [anon_sym___forceinline] = ACTIONS(8085),
    [anon_sym_thread_local] = ACTIONS(8085),
    [anon_sym___thread] = ACTIONS(8085),
    [anon_sym_CG_EXTERN] = ACTIONS(8085),
    [anon_sym_CG_INLINE] = ACTIONS(8085),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8085),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8085),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8085),
    [anon_sym_IBOutlet] = ACTIONS(8085),
    [anon_sym_IBInspectable] = ACTIONS(8085),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8085),
    [anon_sym_NS_INLINE] = ACTIONS(8085),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8085),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8085),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8085),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8085),
    [anon_sym_const] = ACTIONS(8087),
    [anon_sym_constexpr] = ACTIONS(8085),
    [anon_sym_volatile] = ACTIONS(8085),
    [anon_sym_restrict] = ACTIONS(8085),
    [anon_sym___restrict__] = ACTIONS(8085),
    [anon_sym__Atomic] = ACTIONS(8085),
    [anon_sym__Noreturn] = ACTIONS(8085),
    [anon_sym__Nonnull] = ACTIONS(8085),
    [anon_sym_nullable] = ACTIONS(8085),
    [anon_sym__Complex] = ACTIONS(8085),
    [anon_sym__Nullable] = ACTIONS(8087),
    [anon_sym__Nullable_result] = ACTIONS(8085),
    [anon_sym__Null_unspecified] = ACTIONS(8085),
    [anon_sym___autoreleasing] = ACTIONS(8085),
    [anon_sym___block] = ACTIONS(8085),
    [anon_sym___bridge] = ACTIONS(8087),
    [anon_sym___bridge_retained] = ACTIONS(8085),
    [anon_sym___bridge_transfer] = ACTIONS(8085),
    [anon_sym___complex] = ACTIONS(8085),
    [anon_sym___const] = ACTIONS(8085),
    [anon_sym___imag] = ACTIONS(8085),
    [anon_sym___kindof] = ACTIONS(8085),
    [anon_sym___nonnull] = ACTIONS(8085),
    [anon_sym___nullable] = ACTIONS(8085),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8085),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8085),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8085),
    [anon_sym___real] = ACTIONS(8085),
    [anon_sym___strong] = ACTIONS(8085),
    [anon_sym___unsafe_unretained] = ACTIONS(8085),
    [anon_sym___unused] = ACTIONS(8085),
    [anon_sym___weak] = ACTIONS(8085),
    [anon_sym_alignas] = ACTIONS(8085),
    [anon_sym__Alignas] = ACTIONS(8085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8085),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8085),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8085),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8085),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8085),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8085),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8085),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8085),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8085),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8085),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8087),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8085),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8085),
    [anon_sym_API_AVAILABLE] = ACTIONS(8085),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8085),
    [anon_sym_API_DEPRECATED] = ACTIONS(8085),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8085),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8085),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8085),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8085),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8085),
    [anon_sym___deprecated_msg] = ACTIONS(8085),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8085),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8085),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8085),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8085),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8085),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8085),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8085),
  },
  [STATE(3245)] = {
    [sym_parameter_list] = STATE(3223),
    [anon_sym_COMMA] = ACTIONS(8091),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(8091),
    [anon_sym___extension__] = ACTIONS(8091),
    [anon_sym_extern] = ACTIONS(8091),
    [anon_sym___attribute__] = ACTIONS(8091),
    [anon_sym___attribute] = ACTIONS(8093),
    [anon_sym_noreturn] = ACTIONS(8091),
    [anon_sym_LBRACK] = ACTIONS(8089),
    [anon_sym___declspec] = ACTIONS(8091),
    [anon_sym_static] = ACTIONS(8091),
    [anon_sym_auto] = ACTIONS(8091),
    [anon_sym_register] = ACTIONS(8091),
    [anon_sym_inline] = ACTIONS(8091),
    [anon_sym___inline] = ACTIONS(8093),
    [anon_sym___inline__] = ACTIONS(8091),
    [anon_sym___forceinline] = ACTIONS(8091),
    [anon_sym_thread_local] = ACTIONS(8091),
    [anon_sym___thread] = ACTIONS(8091),
    [anon_sym_CG_EXTERN] = ACTIONS(8091),
    [anon_sym_CG_INLINE] = ACTIONS(8091),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8091),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8091),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8091),
    [anon_sym_IBOutlet] = ACTIONS(8091),
    [anon_sym_IBInspectable] = ACTIONS(8091),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8091),
    [anon_sym_NS_INLINE] = ACTIONS(8091),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8091),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8091),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8091),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8091),
    [anon_sym_const] = ACTIONS(8093),
    [anon_sym_constexpr] = ACTIONS(8091),
    [anon_sym_volatile] = ACTIONS(8091),
    [anon_sym_restrict] = ACTIONS(8091),
    [anon_sym___restrict__] = ACTIONS(8091),
    [anon_sym__Atomic] = ACTIONS(8091),
    [anon_sym__Noreturn] = ACTIONS(8091),
    [anon_sym__Nonnull] = ACTIONS(8091),
    [anon_sym_nullable] = ACTIONS(8091),
    [anon_sym__Complex] = ACTIONS(8091),
    [anon_sym__Nullable] = ACTIONS(8093),
    [anon_sym__Nullable_result] = ACTIONS(8091),
    [anon_sym__Null_unspecified] = ACTIONS(8091),
    [anon_sym___autoreleasing] = ACTIONS(8091),
    [anon_sym___block] = ACTIONS(8091),
    [anon_sym___bridge] = ACTIONS(8093),
    [anon_sym___bridge_retained] = ACTIONS(8091),
    [anon_sym___bridge_transfer] = ACTIONS(8091),
    [anon_sym___complex] = ACTIONS(8091),
    [anon_sym___const] = ACTIONS(8091),
    [anon_sym___imag] = ACTIONS(8091),
    [anon_sym___kindof] = ACTIONS(8091),
    [anon_sym___nonnull] = ACTIONS(8091),
    [anon_sym___nullable] = ACTIONS(8091),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8091),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8091),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8091),
    [anon_sym___real] = ACTIONS(8091),
    [anon_sym___strong] = ACTIONS(8091),
    [anon_sym___unsafe_unretained] = ACTIONS(8091),
    [anon_sym___unused] = ACTIONS(8091),
    [anon_sym___weak] = ACTIONS(8091),
    [anon_sym_alignas] = ACTIONS(8091),
    [anon_sym__Alignas] = ACTIONS(8091),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8091),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8091),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8091),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8091),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8091),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8091),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8091),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8091),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8091),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8091),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8093),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8091),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8091),
    [anon_sym_API_AVAILABLE] = ACTIONS(8091),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8091),
    [anon_sym_API_DEPRECATED] = ACTIONS(8091),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8091),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8091),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8091),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8091),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8091),
    [anon_sym___deprecated_msg] = ACTIONS(8091),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8091),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8091),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8091),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8091),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8091),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8091),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8091),
  },
  [STATE(3246)] = {
    [sym_parameter_list] = STATE(3223),
    [anon_sym_COMMA] = ACTIONS(8095),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(8095),
    [anon_sym___extension__] = ACTIONS(8095),
    [anon_sym_extern] = ACTIONS(8095),
    [anon_sym___attribute__] = ACTIONS(8095),
    [anon_sym___attribute] = ACTIONS(8097),
    [anon_sym_noreturn] = ACTIONS(8095),
    [anon_sym_LBRACK] = ACTIONS(8089),
    [anon_sym___declspec] = ACTIONS(8095),
    [anon_sym_static] = ACTIONS(8095),
    [anon_sym_auto] = ACTIONS(8095),
    [anon_sym_register] = ACTIONS(8095),
    [anon_sym_inline] = ACTIONS(8095),
    [anon_sym___inline] = ACTIONS(8097),
    [anon_sym___inline__] = ACTIONS(8095),
    [anon_sym___forceinline] = ACTIONS(8095),
    [anon_sym_thread_local] = ACTIONS(8095),
    [anon_sym___thread] = ACTIONS(8095),
    [anon_sym_CG_EXTERN] = ACTIONS(8095),
    [anon_sym_CG_INLINE] = ACTIONS(8095),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8095),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8095),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8095),
    [anon_sym_IBOutlet] = ACTIONS(8095),
    [anon_sym_IBInspectable] = ACTIONS(8095),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8095),
    [anon_sym_NS_INLINE] = ACTIONS(8095),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8095),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8095),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8095),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8095),
    [anon_sym_const] = ACTIONS(8097),
    [anon_sym_constexpr] = ACTIONS(8095),
    [anon_sym_volatile] = ACTIONS(8095),
    [anon_sym_restrict] = ACTIONS(8095),
    [anon_sym___restrict__] = ACTIONS(8095),
    [anon_sym__Atomic] = ACTIONS(8095),
    [anon_sym__Noreturn] = ACTIONS(8095),
    [anon_sym__Nonnull] = ACTIONS(8095),
    [anon_sym_nullable] = ACTIONS(8095),
    [anon_sym__Complex] = ACTIONS(8095),
    [anon_sym__Nullable] = ACTIONS(8097),
    [anon_sym__Nullable_result] = ACTIONS(8095),
    [anon_sym__Null_unspecified] = ACTIONS(8095),
    [anon_sym___autoreleasing] = ACTIONS(8095),
    [anon_sym___block] = ACTIONS(8095),
    [anon_sym___bridge] = ACTIONS(8097),
    [anon_sym___bridge_retained] = ACTIONS(8095),
    [anon_sym___bridge_transfer] = ACTIONS(8095),
    [anon_sym___complex] = ACTIONS(8095),
    [anon_sym___const] = ACTIONS(8095),
    [anon_sym___imag] = ACTIONS(8095),
    [anon_sym___kindof] = ACTIONS(8095),
    [anon_sym___nonnull] = ACTIONS(8095),
    [anon_sym___nullable] = ACTIONS(8095),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8095),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8095),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8095),
    [anon_sym___real] = ACTIONS(8095),
    [anon_sym___strong] = ACTIONS(8095),
    [anon_sym___unsafe_unretained] = ACTIONS(8095),
    [anon_sym___unused] = ACTIONS(8095),
    [anon_sym___weak] = ACTIONS(8095),
    [anon_sym_alignas] = ACTIONS(8095),
    [anon_sym__Alignas] = ACTIONS(8095),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8095),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8095),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8095),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8095),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8095),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8095),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8095),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8095),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8095),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8095),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8097),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8095),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8095),
    [anon_sym_API_AVAILABLE] = ACTIONS(8095),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8095),
    [anon_sym_API_DEPRECATED] = ACTIONS(8095),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8095),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8095),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8095),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8095),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8095),
    [anon_sym___deprecated_msg] = ACTIONS(8095),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8095),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8095),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8095),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8095),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8095),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8095),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8095),
  },
  [STATE(3247)] = {
    [sym_identifier] = ACTIONS(7997),
    [anon_sym_COMMA] = ACTIONS(7992),
    [anon_sym_SEMI] = ACTIONS(7992),
    [anon_sym___extension__] = ACTIONS(7997),
    [anon_sym_extern] = ACTIONS(7997),
    [anon_sym___attribute__] = ACTIONS(7997),
    [anon_sym___attribute] = ACTIONS(7997),
    [anon_sym_noreturn] = ACTIONS(7997),
    [anon_sym_LBRACK] = ACTIONS(7992),
    [anon_sym___declspec] = ACTIONS(7997),
    [anon_sym_static] = ACTIONS(7997),
    [anon_sym_auto] = ACTIONS(7997),
    [anon_sym_register] = ACTIONS(7997),
    [anon_sym_inline] = ACTIONS(7997),
    [anon_sym___inline] = ACTIONS(7997),
    [anon_sym___inline__] = ACTIONS(7997),
    [anon_sym___forceinline] = ACTIONS(7997),
    [anon_sym_thread_local] = ACTIONS(7997),
    [anon_sym___thread] = ACTIONS(7997),
    [anon_sym_CG_EXTERN] = ACTIONS(7997),
    [anon_sym_CG_INLINE] = ACTIONS(7997),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7997),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7997),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7997),
    [anon_sym_IBOutlet] = ACTIONS(7997),
    [anon_sym_IBInspectable] = ACTIONS(7997),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7997),
    [anon_sym_NS_INLINE] = ACTIONS(7997),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7997),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7997),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7997),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7997),
    [anon_sym_const] = ACTIONS(7997),
    [anon_sym_constexpr] = ACTIONS(7997),
    [anon_sym_volatile] = ACTIONS(7997),
    [anon_sym_restrict] = ACTIONS(7997),
    [anon_sym___restrict__] = ACTIONS(7997),
    [anon_sym__Atomic] = ACTIONS(7997),
    [anon_sym__Noreturn] = ACTIONS(7997),
    [anon_sym__Nonnull] = ACTIONS(7997),
    [anon_sym_nullable] = ACTIONS(7997),
    [anon_sym__Complex] = ACTIONS(7997),
    [anon_sym__Nullable] = ACTIONS(7997),
    [anon_sym__Nullable_result] = ACTIONS(7997),
    [anon_sym__Null_unspecified] = ACTIONS(7997),
    [anon_sym___autoreleasing] = ACTIONS(7997),
    [anon_sym___block] = ACTIONS(7997),
    [anon_sym___bridge] = ACTIONS(7997),
    [anon_sym___bridge_retained] = ACTIONS(7997),
    [anon_sym___bridge_transfer] = ACTIONS(7997),
    [anon_sym___complex] = ACTIONS(7997),
    [anon_sym___const] = ACTIONS(7997),
    [anon_sym___imag] = ACTIONS(7997),
    [anon_sym___kindof] = ACTIONS(7997),
    [anon_sym___nonnull] = ACTIONS(7997),
    [anon_sym___nullable] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7997),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7997),
    [anon_sym___real] = ACTIONS(7997),
    [anon_sym___strong] = ACTIONS(7997),
    [anon_sym___unsafe_unretained] = ACTIONS(7997),
    [anon_sym___unused] = ACTIONS(7997),
    [anon_sym___weak] = ACTIONS(7997),
    [anon_sym_alignas] = ACTIONS(7997),
    [anon_sym__Alignas] = ACTIONS(7997),
    [anon_sym_COLON] = ACTIONS(7992),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7997),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7997),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7997),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7997),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7997),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7997),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7997),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_API_AVAILABLE] = ACTIONS(7997),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_API_DEPRECATED] = ACTIONS(7997),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7997),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7997),
    [anon_sym___deprecated_msg] = ACTIONS(7997),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7997),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7997),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7997),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7997),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7997),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7997),
  },
  [STATE(3248)] = {
    [sym_attribute_specifier] = STATE(3990),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_compound_statement] = STATE(1662),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_parameter_list] = STATE(5008),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3214),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(4905),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3214),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(8099),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym_LT] = ACTIONS(8101),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8101),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_LBRACE] = ACTIONS(8103),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_ATautoreleasepool] = ACTIONS(8105),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7987),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3249)] = {
    [sym_identifier] = ACTIONS(8107),
    [anon_sym_COMMA] = ACTIONS(8109),
    [anon_sym_SEMI] = ACTIONS(8109),
    [anon_sym___extension__] = ACTIONS(8107),
    [anon_sym_extern] = ACTIONS(8107),
    [anon_sym___attribute__] = ACTIONS(8107),
    [anon_sym___attribute] = ACTIONS(8107),
    [anon_sym_noreturn] = ACTIONS(8107),
    [anon_sym_LBRACK] = ACTIONS(8109),
    [anon_sym___declspec] = ACTIONS(8107),
    [anon_sym_static] = ACTIONS(8107),
    [anon_sym_auto] = ACTIONS(8107),
    [anon_sym_register] = ACTIONS(8107),
    [anon_sym_inline] = ACTIONS(8107),
    [anon_sym___inline] = ACTIONS(8107),
    [anon_sym___inline__] = ACTIONS(8107),
    [anon_sym___forceinline] = ACTIONS(8107),
    [anon_sym_thread_local] = ACTIONS(8107),
    [anon_sym___thread] = ACTIONS(8107),
    [anon_sym_CG_EXTERN] = ACTIONS(8107),
    [anon_sym_CG_INLINE] = ACTIONS(8107),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8107),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8107),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8107),
    [anon_sym_IBOutlet] = ACTIONS(8107),
    [anon_sym_IBInspectable] = ACTIONS(8107),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8107),
    [anon_sym_NS_INLINE] = ACTIONS(8107),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8107),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8107),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8107),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8107),
    [anon_sym_const] = ACTIONS(8107),
    [anon_sym_constexpr] = ACTIONS(8107),
    [anon_sym_volatile] = ACTIONS(8107),
    [anon_sym_restrict] = ACTIONS(8107),
    [anon_sym___restrict__] = ACTIONS(8107),
    [anon_sym__Atomic] = ACTIONS(8107),
    [anon_sym__Noreturn] = ACTIONS(8107),
    [anon_sym__Nonnull] = ACTIONS(8107),
    [anon_sym_nullable] = ACTIONS(8107),
    [anon_sym__Complex] = ACTIONS(8107),
    [anon_sym__Nullable] = ACTIONS(8107),
    [anon_sym__Nullable_result] = ACTIONS(8107),
    [anon_sym__Null_unspecified] = ACTIONS(8107),
    [anon_sym___autoreleasing] = ACTIONS(8107),
    [anon_sym___block] = ACTIONS(8107),
    [anon_sym___bridge] = ACTIONS(8107),
    [anon_sym___bridge_retained] = ACTIONS(8107),
    [anon_sym___bridge_transfer] = ACTIONS(8107),
    [anon_sym___complex] = ACTIONS(8107),
    [anon_sym___const] = ACTIONS(8107),
    [anon_sym___imag] = ACTIONS(8107),
    [anon_sym___kindof] = ACTIONS(8107),
    [anon_sym___nonnull] = ACTIONS(8107),
    [anon_sym___nullable] = ACTIONS(8107),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8107),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8107),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8107),
    [anon_sym___real] = ACTIONS(8107),
    [anon_sym___strong] = ACTIONS(8107),
    [anon_sym___unsafe_unretained] = ACTIONS(8107),
    [anon_sym___unused] = ACTIONS(8107),
    [anon_sym___weak] = ACTIONS(8107),
    [anon_sym_alignas] = ACTIONS(8107),
    [anon_sym__Alignas] = ACTIONS(8107),
    [anon_sym_COLON] = ACTIONS(8109),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8107),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8107),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8107),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8107),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8107),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8107),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8107),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8107),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8107),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8107),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8107),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8107),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8107),
    [anon_sym_API_AVAILABLE] = ACTIONS(8107),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8107),
    [anon_sym_API_DEPRECATED] = ACTIONS(8107),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8107),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8107),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8107),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8107),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8107),
    [anon_sym___deprecated_msg] = ACTIONS(8107),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8107),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8107),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8107),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8107),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8107),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8107),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8107),
  },
  [STATE(3250)] = {
    [sym_parameter_list] = STATE(3223),
    [anon_sym_COMMA] = ACTIONS(8111),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(8111),
    [anon_sym___extension__] = ACTIONS(8111),
    [anon_sym_extern] = ACTIONS(8111),
    [anon_sym___attribute__] = ACTIONS(8111),
    [anon_sym___attribute] = ACTIONS(8113),
    [anon_sym_noreturn] = ACTIONS(8111),
    [anon_sym_LBRACK] = ACTIONS(8089),
    [anon_sym___declspec] = ACTIONS(8111),
    [anon_sym_static] = ACTIONS(8111),
    [anon_sym_auto] = ACTIONS(8111),
    [anon_sym_register] = ACTIONS(8111),
    [anon_sym_inline] = ACTIONS(8111),
    [anon_sym___inline] = ACTIONS(8113),
    [anon_sym___inline__] = ACTIONS(8111),
    [anon_sym___forceinline] = ACTIONS(8111),
    [anon_sym_thread_local] = ACTIONS(8111),
    [anon_sym___thread] = ACTIONS(8111),
    [anon_sym_CG_EXTERN] = ACTIONS(8111),
    [anon_sym_CG_INLINE] = ACTIONS(8111),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8111),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8111),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8111),
    [anon_sym_IBOutlet] = ACTIONS(8111),
    [anon_sym_IBInspectable] = ACTIONS(8111),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8111),
    [anon_sym_NS_INLINE] = ACTIONS(8111),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8111),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8111),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8111),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8111),
    [anon_sym_const] = ACTIONS(8113),
    [anon_sym_constexpr] = ACTIONS(8111),
    [anon_sym_volatile] = ACTIONS(8111),
    [anon_sym_restrict] = ACTIONS(8111),
    [anon_sym___restrict__] = ACTIONS(8111),
    [anon_sym__Atomic] = ACTIONS(8111),
    [anon_sym__Noreturn] = ACTIONS(8111),
    [anon_sym__Nonnull] = ACTIONS(8111),
    [anon_sym_nullable] = ACTIONS(8111),
    [anon_sym__Complex] = ACTIONS(8111),
    [anon_sym__Nullable] = ACTIONS(8113),
    [anon_sym__Nullable_result] = ACTIONS(8111),
    [anon_sym__Null_unspecified] = ACTIONS(8111),
    [anon_sym___autoreleasing] = ACTIONS(8111),
    [anon_sym___block] = ACTIONS(8111),
    [anon_sym___bridge] = ACTIONS(8113),
    [anon_sym___bridge_retained] = ACTIONS(8111),
    [anon_sym___bridge_transfer] = ACTIONS(8111),
    [anon_sym___complex] = ACTIONS(8111),
    [anon_sym___const] = ACTIONS(8111),
    [anon_sym___imag] = ACTIONS(8111),
    [anon_sym___kindof] = ACTIONS(8111),
    [anon_sym___nonnull] = ACTIONS(8111),
    [anon_sym___nullable] = ACTIONS(8111),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8111),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8111),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8111),
    [anon_sym___real] = ACTIONS(8111),
    [anon_sym___strong] = ACTIONS(8111),
    [anon_sym___unsafe_unretained] = ACTIONS(8111),
    [anon_sym___unused] = ACTIONS(8111),
    [anon_sym___weak] = ACTIONS(8111),
    [anon_sym_alignas] = ACTIONS(8111),
    [anon_sym__Alignas] = ACTIONS(8111),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8111),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8111),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8111),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8111),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8111),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8111),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8111),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8111),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8111),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8111),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8113),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8111),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8111),
    [anon_sym_API_AVAILABLE] = ACTIONS(8111),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8111),
    [anon_sym_API_DEPRECATED] = ACTIONS(8111),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8111),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8111),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8111),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8111),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8111),
    [anon_sym___deprecated_msg] = ACTIONS(8111),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8111),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8111),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8111),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8111),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8111),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8111),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8111),
  },
  [STATE(3251)] = {
    [anon_sym_COMMA] = ACTIONS(8115),
    [anon_sym_RPAREN] = ACTIONS(8115),
    [anon_sym_LPAREN2] = ACTIONS(8115),
    [anon_sym_SEMI] = ACTIONS(8115),
    [anon_sym___extension__] = ACTIONS(8115),
    [anon_sym_extern] = ACTIONS(8115),
    [anon_sym___attribute__] = ACTIONS(8115),
    [anon_sym___attribute] = ACTIONS(8117),
    [anon_sym_noreturn] = ACTIONS(8115),
    [anon_sym_LBRACK] = ACTIONS(8115),
    [anon_sym___declspec] = ACTIONS(8115),
    [anon_sym_static] = ACTIONS(8115),
    [anon_sym_auto] = ACTIONS(8115),
    [anon_sym_register] = ACTIONS(8115),
    [anon_sym_inline] = ACTIONS(8115),
    [anon_sym___inline] = ACTIONS(8117),
    [anon_sym___inline__] = ACTIONS(8115),
    [anon_sym___forceinline] = ACTIONS(8115),
    [anon_sym_thread_local] = ACTIONS(8115),
    [anon_sym___thread] = ACTIONS(8115),
    [anon_sym_CG_EXTERN] = ACTIONS(8115),
    [anon_sym_CG_INLINE] = ACTIONS(8115),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8115),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8115),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8115),
    [anon_sym_IBOutlet] = ACTIONS(8115),
    [anon_sym_IBInspectable] = ACTIONS(8115),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8115),
    [anon_sym_NS_INLINE] = ACTIONS(8115),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8115),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8115),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8115),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8115),
    [anon_sym_const] = ACTIONS(8117),
    [anon_sym_constexpr] = ACTIONS(8115),
    [anon_sym_volatile] = ACTIONS(8115),
    [anon_sym_restrict] = ACTIONS(8115),
    [anon_sym___restrict__] = ACTIONS(8115),
    [anon_sym__Atomic] = ACTIONS(8115),
    [anon_sym__Noreturn] = ACTIONS(8115),
    [anon_sym__Nonnull] = ACTIONS(8115),
    [anon_sym_nullable] = ACTIONS(8115),
    [anon_sym__Complex] = ACTIONS(8115),
    [anon_sym__Nullable] = ACTIONS(8117),
    [anon_sym__Nullable_result] = ACTIONS(8115),
    [anon_sym__Null_unspecified] = ACTIONS(8115),
    [anon_sym___autoreleasing] = ACTIONS(8115),
    [anon_sym___block] = ACTIONS(8115),
    [anon_sym___bridge] = ACTIONS(8117),
    [anon_sym___bridge_retained] = ACTIONS(8115),
    [anon_sym___bridge_transfer] = ACTIONS(8115),
    [anon_sym___complex] = ACTIONS(8115),
    [anon_sym___const] = ACTIONS(8115),
    [anon_sym___imag] = ACTIONS(8115),
    [anon_sym___kindof] = ACTIONS(8115),
    [anon_sym___nonnull] = ACTIONS(8115),
    [anon_sym___nullable] = ACTIONS(8115),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8115),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8115),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8115),
    [anon_sym___real] = ACTIONS(8115),
    [anon_sym___strong] = ACTIONS(8115),
    [anon_sym___unsafe_unretained] = ACTIONS(8115),
    [anon_sym___unused] = ACTIONS(8115),
    [anon_sym___weak] = ACTIONS(8115),
    [anon_sym_alignas] = ACTIONS(8115),
    [anon_sym__Alignas] = ACTIONS(8115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8115),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8115),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8115),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8115),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8115),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8115),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8115),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8115),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8115),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8115),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8117),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8115),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8115),
    [anon_sym_API_AVAILABLE] = ACTIONS(8115),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8115),
    [anon_sym_API_DEPRECATED] = ACTIONS(8115),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8115),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8115),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8115),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8115),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8115),
    [anon_sym___deprecated_msg] = ACTIONS(8115),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8115),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8115),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8115),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8115),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8115),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8115),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8115),
  },
  [STATE(3252)] = {
    [anon_sym_COMMA] = ACTIONS(8119),
    [anon_sym_RPAREN] = ACTIONS(8119),
    [anon_sym_LPAREN2] = ACTIONS(8119),
    [anon_sym_SEMI] = ACTIONS(8119),
    [anon_sym___extension__] = ACTIONS(8119),
    [anon_sym_extern] = ACTIONS(8119),
    [anon_sym___attribute__] = ACTIONS(8119),
    [anon_sym___attribute] = ACTIONS(8121),
    [anon_sym_noreturn] = ACTIONS(8119),
    [anon_sym_LBRACK] = ACTIONS(8119),
    [anon_sym___declspec] = ACTIONS(8119),
    [anon_sym_static] = ACTIONS(8119),
    [anon_sym_auto] = ACTIONS(8119),
    [anon_sym_register] = ACTIONS(8119),
    [anon_sym_inline] = ACTIONS(8119),
    [anon_sym___inline] = ACTIONS(8121),
    [anon_sym___inline__] = ACTIONS(8119),
    [anon_sym___forceinline] = ACTIONS(8119),
    [anon_sym_thread_local] = ACTIONS(8119),
    [anon_sym___thread] = ACTIONS(8119),
    [anon_sym_CG_EXTERN] = ACTIONS(8119),
    [anon_sym_CG_INLINE] = ACTIONS(8119),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8119),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8119),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8119),
    [anon_sym_IBOutlet] = ACTIONS(8119),
    [anon_sym_IBInspectable] = ACTIONS(8119),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8119),
    [anon_sym_NS_INLINE] = ACTIONS(8119),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8119),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8119),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8119),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8119),
    [anon_sym_const] = ACTIONS(8121),
    [anon_sym_constexpr] = ACTIONS(8119),
    [anon_sym_volatile] = ACTIONS(8119),
    [anon_sym_restrict] = ACTIONS(8119),
    [anon_sym___restrict__] = ACTIONS(8119),
    [anon_sym__Atomic] = ACTIONS(8119),
    [anon_sym__Noreturn] = ACTIONS(8119),
    [anon_sym__Nonnull] = ACTIONS(8119),
    [anon_sym_nullable] = ACTIONS(8119),
    [anon_sym__Complex] = ACTIONS(8119),
    [anon_sym__Nullable] = ACTIONS(8121),
    [anon_sym__Nullable_result] = ACTIONS(8119),
    [anon_sym__Null_unspecified] = ACTIONS(8119),
    [anon_sym___autoreleasing] = ACTIONS(8119),
    [anon_sym___block] = ACTIONS(8119),
    [anon_sym___bridge] = ACTIONS(8121),
    [anon_sym___bridge_retained] = ACTIONS(8119),
    [anon_sym___bridge_transfer] = ACTIONS(8119),
    [anon_sym___complex] = ACTIONS(8119),
    [anon_sym___const] = ACTIONS(8119),
    [anon_sym___imag] = ACTIONS(8119),
    [anon_sym___kindof] = ACTIONS(8119),
    [anon_sym___nonnull] = ACTIONS(8119),
    [anon_sym___nullable] = ACTIONS(8119),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8119),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8119),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8119),
    [anon_sym___real] = ACTIONS(8119),
    [anon_sym___strong] = ACTIONS(8119),
    [anon_sym___unsafe_unretained] = ACTIONS(8119),
    [anon_sym___unused] = ACTIONS(8119),
    [anon_sym___weak] = ACTIONS(8119),
    [anon_sym_alignas] = ACTIONS(8119),
    [anon_sym__Alignas] = ACTIONS(8119),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8119),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8119),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8119),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8119),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8119),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8119),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8119),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8119),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8119),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8119),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8121),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8119),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8119),
    [anon_sym_API_AVAILABLE] = ACTIONS(8119),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8119),
    [anon_sym_API_DEPRECATED] = ACTIONS(8119),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8119),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8119),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8119),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8119),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8119),
    [anon_sym___deprecated_msg] = ACTIONS(8119),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8119),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8119),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8119),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8119),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8119),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8119),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8119),
  },
  [STATE(3253)] = {
    [anon_sym_COMMA] = ACTIONS(8123),
    [anon_sym_RPAREN] = ACTIONS(8123),
    [anon_sym_LPAREN2] = ACTIONS(8123),
    [anon_sym_SEMI] = ACTIONS(8123),
    [anon_sym___extension__] = ACTIONS(8123),
    [anon_sym_extern] = ACTIONS(8123),
    [anon_sym___attribute__] = ACTIONS(8123),
    [anon_sym___attribute] = ACTIONS(8125),
    [anon_sym_noreturn] = ACTIONS(8123),
    [anon_sym_LBRACK] = ACTIONS(8123),
    [anon_sym___declspec] = ACTIONS(8123),
    [anon_sym_static] = ACTIONS(8123),
    [anon_sym_auto] = ACTIONS(8123),
    [anon_sym_register] = ACTIONS(8123),
    [anon_sym_inline] = ACTIONS(8123),
    [anon_sym___inline] = ACTIONS(8125),
    [anon_sym___inline__] = ACTIONS(8123),
    [anon_sym___forceinline] = ACTIONS(8123),
    [anon_sym_thread_local] = ACTIONS(8123),
    [anon_sym___thread] = ACTIONS(8123),
    [anon_sym_CG_EXTERN] = ACTIONS(8123),
    [anon_sym_CG_INLINE] = ACTIONS(8123),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8123),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8123),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8123),
    [anon_sym_IBOutlet] = ACTIONS(8123),
    [anon_sym_IBInspectable] = ACTIONS(8123),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8123),
    [anon_sym_NS_INLINE] = ACTIONS(8123),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8123),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8123),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8123),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8123),
    [anon_sym_const] = ACTIONS(8125),
    [anon_sym_constexpr] = ACTIONS(8123),
    [anon_sym_volatile] = ACTIONS(8123),
    [anon_sym_restrict] = ACTIONS(8123),
    [anon_sym___restrict__] = ACTIONS(8123),
    [anon_sym__Atomic] = ACTIONS(8123),
    [anon_sym__Noreturn] = ACTIONS(8123),
    [anon_sym__Nonnull] = ACTIONS(8123),
    [anon_sym_nullable] = ACTIONS(8123),
    [anon_sym__Complex] = ACTIONS(8123),
    [anon_sym__Nullable] = ACTIONS(8125),
    [anon_sym__Nullable_result] = ACTIONS(8123),
    [anon_sym__Null_unspecified] = ACTIONS(8123),
    [anon_sym___autoreleasing] = ACTIONS(8123),
    [anon_sym___block] = ACTIONS(8123),
    [anon_sym___bridge] = ACTIONS(8125),
    [anon_sym___bridge_retained] = ACTIONS(8123),
    [anon_sym___bridge_transfer] = ACTIONS(8123),
    [anon_sym___complex] = ACTIONS(8123),
    [anon_sym___const] = ACTIONS(8123),
    [anon_sym___imag] = ACTIONS(8123),
    [anon_sym___kindof] = ACTIONS(8123),
    [anon_sym___nonnull] = ACTIONS(8123),
    [anon_sym___nullable] = ACTIONS(8123),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8123),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8123),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8123),
    [anon_sym___real] = ACTIONS(8123),
    [anon_sym___strong] = ACTIONS(8123),
    [anon_sym___unsafe_unretained] = ACTIONS(8123),
    [anon_sym___unused] = ACTIONS(8123),
    [anon_sym___weak] = ACTIONS(8123),
    [anon_sym_alignas] = ACTIONS(8123),
    [anon_sym__Alignas] = ACTIONS(8123),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8123),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8123),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8123),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8123),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8123),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8123),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8123),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8123),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8123),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8123),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8125),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8123),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8123),
    [anon_sym_API_AVAILABLE] = ACTIONS(8123),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8123),
    [anon_sym_API_DEPRECATED] = ACTIONS(8123),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8123),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8123),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8123),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8123),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8123),
    [anon_sym___deprecated_msg] = ACTIONS(8123),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8123),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8123),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8123),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8123),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8123),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8123),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8123),
  },
  [STATE(3254)] = {
    [sym_identifier] = ACTIONS(8052),
    [anon_sym_COMMA] = ACTIONS(8054),
    [anon_sym_SEMI] = ACTIONS(8054),
    [anon_sym___extension__] = ACTIONS(8052),
    [anon_sym_extern] = ACTIONS(8052),
    [anon_sym___attribute__] = ACTIONS(8052),
    [anon_sym___attribute] = ACTIONS(8052),
    [anon_sym_noreturn] = ACTIONS(8052),
    [anon_sym_LBRACK] = ACTIONS(8054),
    [anon_sym___declspec] = ACTIONS(8052),
    [anon_sym_static] = ACTIONS(8052),
    [anon_sym_auto] = ACTIONS(8052),
    [anon_sym_register] = ACTIONS(8052),
    [anon_sym_inline] = ACTIONS(8052),
    [anon_sym___inline] = ACTIONS(8052),
    [anon_sym___inline__] = ACTIONS(8052),
    [anon_sym___forceinline] = ACTIONS(8052),
    [anon_sym_thread_local] = ACTIONS(8052),
    [anon_sym___thread] = ACTIONS(8052),
    [anon_sym_CG_EXTERN] = ACTIONS(8052),
    [anon_sym_CG_INLINE] = ACTIONS(8052),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8052),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8052),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8052),
    [anon_sym_IBOutlet] = ACTIONS(8052),
    [anon_sym_IBInspectable] = ACTIONS(8052),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8052),
    [anon_sym_NS_INLINE] = ACTIONS(8052),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8052),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8052),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8052),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8052),
    [anon_sym_const] = ACTIONS(8052),
    [anon_sym_constexpr] = ACTIONS(8052),
    [anon_sym_volatile] = ACTIONS(8052),
    [anon_sym_restrict] = ACTIONS(8052),
    [anon_sym___restrict__] = ACTIONS(8052),
    [anon_sym__Atomic] = ACTIONS(8052),
    [anon_sym__Noreturn] = ACTIONS(8052),
    [anon_sym__Nonnull] = ACTIONS(8052),
    [anon_sym_nullable] = ACTIONS(8052),
    [anon_sym__Complex] = ACTIONS(8052),
    [anon_sym__Nullable] = ACTIONS(8052),
    [anon_sym__Nullable_result] = ACTIONS(8052),
    [anon_sym__Null_unspecified] = ACTIONS(8052),
    [anon_sym___autoreleasing] = ACTIONS(8052),
    [anon_sym___block] = ACTIONS(8052),
    [anon_sym___bridge] = ACTIONS(8052),
    [anon_sym___bridge_retained] = ACTIONS(8052),
    [anon_sym___bridge_transfer] = ACTIONS(8052),
    [anon_sym___complex] = ACTIONS(8052),
    [anon_sym___const] = ACTIONS(8052),
    [anon_sym___imag] = ACTIONS(8052),
    [anon_sym___kindof] = ACTIONS(8052),
    [anon_sym___nonnull] = ACTIONS(8052),
    [anon_sym___nullable] = ACTIONS(8052),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8052),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8052),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8052),
    [anon_sym___real] = ACTIONS(8052),
    [anon_sym___strong] = ACTIONS(8052),
    [anon_sym___unsafe_unretained] = ACTIONS(8052),
    [anon_sym___unused] = ACTIONS(8052),
    [anon_sym___weak] = ACTIONS(8052),
    [anon_sym_alignas] = ACTIONS(8052),
    [anon_sym__Alignas] = ACTIONS(8052),
    [anon_sym_COLON] = ACTIONS(8054),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8052),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8052),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8052),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8052),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8052),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8052),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8052),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8052),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8052),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_API_AVAILABLE] = ACTIONS(8052),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_API_DEPRECATED] = ACTIONS(8052),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8052),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8052),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8052),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8052),
    [anon_sym___deprecated_msg] = ACTIONS(8052),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8052),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8052),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8052),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8052),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8052),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8052),
  },
  [STATE(3255)] = {
    [sym_attribute_specifier] = STATE(3996),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_compound_statement] = STATE(1499),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_parameter_list] = STATE(5076),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3214),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(4908),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3214),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(8099),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym_LT] = ACTIONS(8101),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8101),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_LBRACE] = ACTIONS(8127),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_ATautoreleasepool] = ACTIONS(8129),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7987),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3256)] = {
    [anon_sym_COMMA] = ACTIONS(8131),
    [anon_sym_RPAREN] = ACTIONS(8131),
    [anon_sym_LPAREN2] = ACTIONS(8131),
    [anon_sym_SEMI] = ACTIONS(8131),
    [anon_sym___extension__] = ACTIONS(8131),
    [anon_sym_extern] = ACTIONS(8131),
    [anon_sym___attribute__] = ACTIONS(8131),
    [anon_sym___attribute] = ACTIONS(8133),
    [anon_sym_noreturn] = ACTIONS(8131),
    [anon_sym_LBRACK] = ACTIONS(8131),
    [anon_sym___declspec] = ACTIONS(8131),
    [anon_sym_static] = ACTIONS(8131),
    [anon_sym_auto] = ACTIONS(8131),
    [anon_sym_register] = ACTIONS(8131),
    [anon_sym_inline] = ACTIONS(8131),
    [anon_sym___inline] = ACTIONS(8133),
    [anon_sym___inline__] = ACTIONS(8131),
    [anon_sym___forceinline] = ACTIONS(8131),
    [anon_sym_thread_local] = ACTIONS(8131),
    [anon_sym___thread] = ACTIONS(8131),
    [anon_sym_CG_EXTERN] = ACTIONS(8131),
    [anon_sym_CG_INLINE] = ACTIONS(8131),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8131),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8131),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8131),
    [anon_sym_IBOutlet] = ACTIONS(8131),
    [anon_sym_IBInspectable] = ACTIONS(8131),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8131),
    [anon_sym_NS_INLINE] = ACTIONS(8131),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8131),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8131),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8131),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8131),
    [anon_sym_const] = ACTIONS(8133),
    [anon_sym_constexpr] = ACTIONS(8131),
    [anon_sym_volatile] = ACTIONS(8131),
    [anon_sym_restrict] = ACTIONS(8131),
    [anon_sym___restrict__] = ACTIONS(8131),
    [anon_sym__Atomic] = ACTIONS(8131),
    [anon_sym__Noreturn] = ACTIONS(8131),
    [anon_sym__Nonnull] = ACTIONS(8131),
    [anon_sym_nullable] = ACTIONS(8131),
    [anon_sym__Complex] = ACTIONS(8131),
    [anon_sym__Nullable] = ACTIONS(8133),
    [anon_sym__Nullable_result] = ACTIONS(8131),
    [anon_sym__Null_unspecified] = ACTIONS(8131),
    [anon_sym___autoreleasing] = ACTIONS(8131),
    [anon_sym___block] = ACTIONS(8131),
    [anon_sym___bridge] = ACTIONS(8133),
    [anon_sym___bridge_retained] = ACTIONS(8131),
    [anon_sym___bridge_transfer] = ACTIONS(8131),
    [anon_sym___complex] = ACTIONS(8131),
    [anon_sym___const] = ACTIONS(8131),
    [anon_sym___imag] = ACTIONS(8131),
    [anon_sym___kindof] = ACTIONS(8131),
    [anon_sym___nonnull] = ACTIONS(8131),
    [anon_sym___nullable] = ACTIONS(8131),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8131),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8131),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8131),
    [anon_sym___real] = ACTIONS(8131),
    [anon_sym___strong] = ACTIONS(8131),
    [anon_sym___unsafe_unretained] = ACTIONS(8131),
    [anon_sym___unused] = ACTIONS(8131),
    [anon_sym___weak] = ACTIONS(8131),
    [anon_sym_alignas] = ACTIONS(8131),
    [anon_sym__Alignas] = ACTIONS(8131),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8131),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8131),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8131),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8131),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8131),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8131),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8131),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8131),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8131),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8131),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8133),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8131),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8131),
    [anon_sym_API_AVAILABLE] = ACTIONS(8131),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8131),
    [anon_sym_API_DEPRECATED] = ACTIONS(8131),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8131),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8131),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8131),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8131),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8131),
    [anon_sym___deprecated_msg] = ACTIONS(8131),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8131),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8131),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8131),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8131),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8131),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8131),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8131),
  },
  [STATE(3257)] = {
    [anon_sym_COMMA] = ACTIONS(8135),
    [anon_sym_RPAREN] = ACTIONS(8135),
    [anon_sym_LPAREN2] = ACTIONS(8135),
    [anon_sym_SEMI] = ACTIONS(8135),
    [anon_sym___extension__] = ACTIONS(8135),
    [anon_sym_extern] = ACTIONS(8135),
    [anon_sym___attribute__] = ACTIONS(8135),
    [anon_sym___attribute] = ACTIONS(8137),
    [anon_sym_noreturn] = ACTIONS(8135),
    [anon_sym_LBRACK] = ACTIONS(8135),
    [anon_sym___declspec] = ACTIONS(8135),
    [anon_sym_static] = ACTIONS(8135),
    [anon_sym_auto] = ACTIONS(8135),
    [anon_sym_register] = ACTIONS(8135),
    [anon_sym_inline] = ACTIONS(8135),
    [anon_sym___inline] = ACTIONS(8137),
    [anon_sym___inline__] = ACTIONS(8135),
    [anon_sym___forceinline] = ACTIONS(8135),
    [anon_sym_thread_local] = ACTIONS(8135),
    [anon_sym___thread] = ACTIONS(8135),
    [anon_sym_CG_EXTERN] = ACTIONS(8135),
    [anon_sym_CG_INLINE] = ACTIONS(8135),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8135),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8135),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8135),
    [anon_sym_IBOutlet] = ACTIONS(8135),
    [anon_sym_IBInspectable] = ACTIONS(8135),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8135),
    [anon_sym_NS_INLINE] = ACTIONS(8135),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8135),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8135),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8135),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8135),
    [anon_sym_const] = ACTIONS(8137),
    [anon_sym_constexpr] = ACTIONS(8135),
    [anon_sym_volatile] = ACTIONS(8135),
    [anon_sym_restrict] = ACTIONS(8135),
    [anon_sym___restrict__] = ACTIONS(8135),
    [anon_sym__Atomic] = ACTIONS(8135),
    [anon_sym__Noreturn] = ACTIONS(8135),
    [anon_sym__Nonnull] = ACTIONS(8135),
    [anon_sym_nullable] = ACTIONS(8135),
    [anon_sym__Complex] = ACTIONS(8135),
    [anon_sym__Nullable] = ACTIONS(8137),
    [anon_sym__Nullable_result] = ACTIONS(8135),
    [anon_sym__Null_unspecified] = ACTIONS(8135),
    [anon_sym___autoreleasing] = ACTIONS(8135),
    [anon_sym___block] = ACTIONS(8135),
    [anon_sym___bridge] = ACTIONS(8137),
    [anon_sym___bridge_retained] = ACTIONS(8135),
    [anon_sym___bridge_transfer] = ACTIONS(8135),
    [anon_sym___complex] = ACTIONS(8135),
    [anon_sym___const] = ACTIONS(8135),
    [anon_sym___imag] = ACTIONS(8135),
    [anon_sym___kindof] = ACTIONS(8135),
    [anon_sym___nonnull] = ACTIONS(8135),
    [anon_sym___nullable] = ACTIONS(8135),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8135),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8135),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8135),
    [anon_sym___real] = ACTIONS(8135),
    [anon_sym___strong] = ACTIONS(8135),
    [anon_sym___unsafe_unretained] = ACTIONS(8135),
    [anon_sym___unused] = ACTIONS(8135),
    [anon_sym___weak] = ACTIONS(8135),
    [anon_sym_alignas] = ACTIONS(8135),
    [anon_sym__Alignas] = ACTIONS(8135),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8135),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8135),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8135),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8135),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8135),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8135),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8135),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8135),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8135),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8135),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8137),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8135),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8135),
    [anon_sym_API_AVAILABLE] = ACTIONS(8135),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8135),
    [anon_sym_API_DEPRECATED] = ACTIONS(8135),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8135),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8135),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8135),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8135),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8135),
    [anon_sym___deprecated_msg] = ACTIONS(8135),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8135),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8135),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8135),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8135),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8135),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8135),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8135),
  },
  [STATE(3258)] = {
    [anon_sym_COMMA] = ACTIONS(8139),
    [anon_sym_RPAREN] = ACTIONS(8139),
    [anon_sym_LPAREN2] = ACTIONS(8139),
    [anon_sym_SEMI] = ACTIONS(8139),
    [anon_sym___extension__] = ACTIONS(8139),
    [anon_sym_extern] = ACTIONS(8139),
    [anon_sym___attribute__] = ACTIONS(8139),
    [anon_sym___attribute] = ACTIONS(8141),
    [anon_sym_noreturn] = ACTIONS(8139),
    [anon_sym_LBRACK] = ACTIONS(8139),
    [anon_sym___declspec] = ACTIONS(8139),
    [anon_sym_static] = ACTIONS(8139),
    [anon_sym_auto] = ACTIONS(8139),
    [anon_sym_register] = ACTIONS(8139),
    [anon_sym_inline] = ACTIONS(8139),
    [anon_sym___inline] = ACTIONS(8141),
    [anon_sym___inline__] = ACTIONS(8139),
    [anon_sym___forceinline] = ACTIONS(8139),
    [anon_sym_thread_local] = ACTIONS(8139),
    [anon_sym___thread] = ACTIONS(8139),
    [anon_sym_CG_EXTERN] = ACTIONS(8139),
    [anon_sym_CG_INLINE] = ACTIONS(8139),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8139),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8139),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8139),
    [anon_sym_IBOutlet] = ACTIONS(8139),
    [anon_sym_IBInspectable] = ACTIONS(8139),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8139),
    [anon_sym_NS_INLINE] = ACTIONS(8139),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8139),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8139),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8139),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8139),
    [anon_sym_const] = ACTIONS(8141),
    [anon_sym_constexpr] = ACTIONS(8139),
    [anon_sym_volatile] = ACTIONS(8139),
    [anon_sym_restrict] = ACTIONS(8139),
    [anon_sym___restrict__] = ACTIONS(8139),
    [anon_sym__Atomic] = ACTIONS(8139),
    [anon_sym__Noreturn] = ACTIONS(8139),
    [anon_sym__Nonnull] = ACTIONS(8139),
    [anon_sym_nullable] = ACTIONS(8139),
    [anon_sym__Complex] = ACTIONS(8139),
    [anon_sym__Nullable] = ACTIONS(8141),
    [anon_sym__Nullable_result] = ACTIONS(8139),
    [anon_sym__Null_unspecified] = ACTIONS(8139),
    [anon_sym___autoreleasing] = ACTIONS(8139),
    [anon_sym___block] = ACTIONS(8139),
    [anon_sym___bridge] = ACTIONS(8141),
    [anon_sym___bridge_retained] = ACTIONS(8139),
    [anon_sym___bridge_transfer] = ACTIONS(8139),
    [anon_sym___complex] = ACTIONS(8139),
    [anon_sym___const] = ACTIONS(8139),
    [anon_sym___imag] = ACTIONS(8139),
    [anon_sym___kindof] = ACTIONS(8139),
    [anon_sym___nonnull] = ACTIONS(8139),
    [anon_sym___nullable] = ACTIONS(8139),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8139),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8139),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8139),
    [anon_sym___real] = ACTIONS(8139),
    [anon_sym___strong] = ACTIONS(8139),
    [anon_sym___unsafe_unretained] = ACTIONS(8139),
    [anon_sym___unused] = ACTIONS(8139),
    [anon_sym___weak] = ACTIONS(8139),
    [anon_sym_alignas] = ACTIONS(8139),
    [anon_sym__Alignas] = ACTIONS(8139),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8139),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8139),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8139),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8139),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8139),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8139),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8139),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8139),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8139),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8139),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8141),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8139),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8139),
    [anon_sym_API_AVAILABLE] = ACTIONS(8139),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8139),
    [anon_sym_API_DEPRECATED] = ACTIONS(8139),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8139),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8139),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8139),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8139),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8139),
    [anon_sym___deprecated_msg] = ACTIONS(8139),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8139),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8139),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8139),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8139),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8139),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8139),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8139),
  },
  [STATE(3259)] = {
    [sym_identifier] = ACTIONS(8068),
    [anon_sym_COMMA] = ACTIONS(8070),
    [anon_sym_SEMI] = ACTIONS(8070),
    [anon_sym___extension__] = ACTIONS(8068),
    [anon_sym_extern] = ACTIONS(8068),
    [anon_sym___attribute__] = ACTIONS(8068),
    [anon_sym___attribute] = ACTIONS(8068),
    [anon_sym_noreturn] = ACTIONS(8068),
    [anon_sym_LBRACK] = ACTIONS(8070),
    [anon_sym___declspec] = ACTIONS(8068),
    [anon_sym_static] = ACTIONS(8068),
    [anon_sym_auto] = ACTIONS(8068),
    [anon_sym_register] = ACTIONS(8068),
    [anon_sym_inline] = ACTIONS(8068),
    [anon_sym___inline] = ACTIONS(8068),
    [anon_sym___inline__] = ACTIONS(8068),
    [anon_sym___forceinline] = ACTIONS(8068),
    [anon_sym_thread_local] = ACTIONS(8068),
    [anon_sym___thread] = ACTIONS(8068),
    [anon_sym_CG_EXTERN] = ACTIONS(8068),
    [anon_sym_CG_INLINE] = ACTIONS(8068),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8068),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8068),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8068),
    [anon_sym_IBOutlet] = ACTIONS(8068),
    [anon_sym_IBInspectable] = ACTIONS(8068),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8068),
    [anon_sym_NS_INLINE] = ACTIONS(8068),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8068),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8068),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8068),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8068),
    [anon_sym_const] = ACTIONS(8068),
    [anon_sym_constexpr] = ACTIONS(8068),
    [anon_sym_volatile] = ACTIONS(8068),
    [anon_sym_restrict] = ACTIONS(8068),
    [anon_sym___restrict__] = ACTIONS(8068),
    [anon_sym__Atomic] = ACTIONS(8068),
    [anon_sym__Noreturn] = ACTIONS(8068),
    [anon_sym__Nonnull] = ACTIONS(8068),
    [anon_sym_nullable] = ACTIONS(8068),
    [anon_sym__Complex] = ACTIONS(8068),
    [anon_sym__Nullable] = ACTIONS(8068),
    [anon_sym__Nullable_result] = ACTIONS(8068),
    [anon_sym__Null_unspecified] = ACTIONS(8068),
    [anon_sym___autoreleasing] = ACTIONS(8068),
    [anon_sym___block] = ACTIONS(8068),
    [anon_sym___bridge] = ACTIONS(8068),
    [anon_sym___bridge_retained] = ACTIONS(8068),
    [anon_sym___bridge_transfer] = ACTIONS(8068),
    [anon_sym___complex] = ACTIONS(8068),
    [anon_sym___const] = ACTIONS(8068),
    [anon_sym___imag] = ACTIONS(8068),
    [anon_sym___kindof] = ACTIONS(8068),
    [anon_sym___nonnull] = ACTIONS(8068),
    [anon_sym___nullable] = ACTIONS(8068),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8068),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8068),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8068),
    [anon_sym___real] = ACTIONS(8068),
    [anon_sym___strong] = ACTIONS(8068),
    [anon_sym___unsafe_unretained] = ACTIONS(8068),
    [anon_sym___unused] = ACTIONS(8068),
    [anon_sym___weak] = ACTIONS(8068),
    [anon_sym_alignas] = ACTIONS(8068),
    [anon_sym__Alignas] = ACTIONS(8068),
    [anon_sym_COLON] = ACTIONS(8070),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8068),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8068),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8068),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8068),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8068),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8068),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8068),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8068),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8068),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8068),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8068),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8068),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8068),
    [anon_sym_API_AVAILABLE] = ACTIONS(8068),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8068),
    [anon_sym_API_DEPRECATED] = ACTIONS(8068),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8068),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8068),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8068),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8068),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8068),
    [anon_sym___deprecated_msg] = ACTIONS(8068),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8068),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8068),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8068),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8068),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8068),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8068),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8068),
  },
  [STATE(3260)] = {
    [sym_parameter_list] = STATE(3223),
    [anon_sym_COMMA] = ACTIONS(8143),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(8143),
    [anon_sym___extension__] = ACTIONS(8143),
    [anon_sym_extern] = ACTIONS(8143),
    [anon_sym___attribute__] = ACTIONS(8143),
    [anon_sym___attribute] = ACTIONS(8145),
    [anon_sym_noreturn] = ACTIONS(8143),
    [anon_sym_LBRACK] = ACTIONS(8089),
    [anon_sym___declspec] = ACTIONS(8143),
    [anon_sym_static] = ACTIONS(8143),
    [anon_sym_auto] = ACTIONS(8143),
    [anon_sym_register] = ACTIONS(8143),
    [anon_sym_inline] = ACTIONS(8143),
    [anon_sym___inline] = ACTIONS(8145),
    [anon_sym___inline__] = ACTIONS(8143),
    [anon_sym___forceinline] = ACTIONS(8143),
    [anon_sym_thread_local] = ACTIONS(8143),
    [anon_sym___thread] = ACTIONS(8143),
    [anon_sym_CG_EXTERN] = ACTIONS(8143),
    [anon_sym_CG_INLINE] = ACTIONS(8143),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8143),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8143),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8143),
    [anon_sym_IBOutlet] = ACTIONS(8143),
    [anon_sym_IBInspectable] = ACTIONS(8143),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8143),
    [anon_sym_NS_INLINE] = ACTIONS(8143),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8143),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8143),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8143),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8143),
    [anon_sym_const] = ACTIONS(8145),
    [anon_sym_constexpr] = ACTIONS(8143),
    [anon_sym_volatile] = ACTIONS(8143),
    [anon_sym_restrict] = ACTIONS(8143),
    [anon_sym___restrict__] = ACTIONS(8143),
    [anon_sym__Atomic] = ACTIONS(8143),
    [anon_sym__Noreturn] = ACTIONS(8143),
    [anon_sym__Nonnull] = ACTIONS(8143),
    [anon_sym_nullable] = ACTIONS(8143),
    [anon_sym__Complex] = ACTIONS(8143),
    [anon_sym__Nullable] = ACTIONS(8145),
    [anon_sym__Nullable_result] = ACTIONS(8143),
    [anon_sym__Null_unspecified] = ACTIONS(8143),
    [anon_sym___autoreleasing] = ACTIONS(8143),
    [anon_sym___block] = ACTIONS(8143),
    [anon_sym___bridge] = ACTIONS(8145),
    [anon_sym___bridge_retained] = ACTIONS(8143),
    [anon_sym___bridge_transfer] = ACTIONS(8143),
    [anon_sym___complex] = ACTIONS(8143),
    [anon_sym___const] = ACTIONS(8143),
    [anon_sym___imag] = ACTIONS(8143),
    [anon_sym___kindof] = ACTIONS(8143),
    [anon_sym___nonnull] = ACTIONS(8143),
    [anon_sym___nullable] = ACTIONS(8143),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8143),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8143),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8143),
    [anon_sym___real] = ACTIONS(8143),
    [anon_sym___strong] = ACTIONS(8143),
    [anon_sym___unsafe_unretained] = ACTIONS(8143),
    [anon_sym___unused] = ACTIONS(8143),
    [anon_sym___weak] = ACTIONS(8143),
    [anon_sym_alignas] = ACTIONS(8143),
    [anon_sym__Alignas] = ACTIONS(8143),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8143),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8143),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8143),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8143),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8143),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8143),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8143),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8143),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8143),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8143),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8145),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8143),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8143),
    [anon_sym_API_AVAILABLE] = ACTIONS(8143),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8143),
    [anon_sym_API_DEPRECATED] = ACTIONS(8143),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8143),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8143),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8143),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8143),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8143),
    [anon_sym___deprecated_msg] = ACTIONS(8143),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8143),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8143),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8143),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8143),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8143),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8143),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8143),
  },
  [STATE(3261)] = {
    [anon_sym_COMMA] = ACTIONS(8147),
    [anon_sym_RPAREN] = ACTIONS(8147),
    [anon_sym_LPAREN2] = ACTIONS(8147),
    [anon_sym_SEMI] = ACTIONS(8147),
    [anon_sym___extension__] = ACTIONS(8147),
    [anon_sym_extern] = ACTIONS(8147),
    [anon_sym___attribute__] = ACTIONS(8147),
    [anon_sym___attribute] = ACTIONS(8149),
    [anon_sym_noreturn] = ACTIONS(8147),
    [anon_sym_LBRACK] = ACTIONS(8147),
    [anon_sym___declspec] = ACTIONS(8147),
    [anon_sym_static] = ACTIONS(8147),
    [anon_sym_auto] = ACTIONS(8147),
    [anon_sym_register] = ACTIONS(8147),
    [anon_sym_inline] = ACTIONS(8147),
    [anon_sym___inline] = ACTIONS(8149),
    [anon_sym___inline__] = ACTIONS(8147),
    [anon_sym___forceinline] = ACTIONS(8147),
    [anon_sym_thread_local] = ACTIONS(8147),
    [anon_sym___thread] = ACTIONS(8147),
    [anon_sym_CG_EXTERN] = ACTIONS(8147),
    [anon_sym_CG_INLINE] = ACTIONS(8147),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8147),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8147),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8147),
    [anon_sym_IBOutlet] = ACTIONS(8147),
    [anon_sym_IBInspectable] = ACTIONS(8147),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8147),
    [anon_sym_NS_INLINE] = ACTIONS(8147),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8147),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8147),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8147),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8147),
    [anon_sym_const] = ACTIONS(8149),
    [anon_sym_constexpr] = ACTIONS(8147),
    [anon_sym_volatile] = ACTIONS(8147),
    [anon_sym_restrict] = ACTIONS(8147),
    [anon_sym___restrict__] = ACTIONS(8147),
    [anon_sym__Atomic] = ACTIONS(8147),
    [anon_sym__Noreturn] = ACTIONS(8147),
    [anon_sym__Nonnull] = ACTIONS(8147),
    [anon_sym_nullable] = ACTIONS(8147),
    [anon_sym__Complex] = ACTIONS(8147),
    [anon_sym__Nullable] = ACTIONS(8149),
    [anon_sym__Nullable_result] = ACTIONS(8147),
    [anon_sym__Null_unspecified] = ACTIONS(8147),
    [anon_sym___autoreleasing] = ACTIONS(8147),
    [anon_sym___block] = ACTIONS(8147),
    [anon_sym___bridge] = ACTIONS(8149),
    [anon_sym___bridge_retained] = ACTIONS(8147),
    [anon_sym___bridge_transfer] = ACTIONS(8147),
    [anon_sym___complex] = ACTIONS(8147),
    [anon_sym___const] = ACTIONS(8147),
    [anon_sym___imag] = ACTIONS(8147),
    [anon_sym___kindof] = ACTIONS(8147),
    [anon_sym___nonnull] = ACTIONS(8147),
    [anon_sym___nullable] = ACTIONS(8147),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8147),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8147),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8147),
    [anon_sym___real] = ACTIONS(8147),
    [anon_sym___strong] = ACTIONS(8147),
    [anon_sym___unsafe_unretained] = ACTIONS(8147),
    [anon_sym___unused] = ACTIONS(8147),
    [anon_sym___weak] = ACTIONS(8147),
    [anon_sym_alignas] = ACTIONS(8147),
    [anon_sym__Alignas] = ACTIONS(8147),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8147),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8147),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8147),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8147),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8147),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8147),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8147),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8147),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8147),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8147),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8149),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8147),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8147),
    [anon_sym_API_AVAILABLE] = ACTIONS(8147),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8147),
    [anon_sym_API_DEPRECATED] = ACTIONS(8147),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8147),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8147),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8147),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8147),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8147),
    [anon_sym___deprecated_msg] = ACTIONS(8147),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8147),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8147),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8147),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8147),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8147),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8147),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8147),
  },
  [STATE(3262)] = {
    [sym_parameter_list] = STATE(3223),
    [anon_sym_COMMA] = ACTIONS(8151),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(8151),
    [anon_sym___extension__] = ACTIONS(8151),
    [anon_sym_extern] = ACTIONS(8151),
    [anon_sym___attribute__] = ACTIONS(8151),
    [anon_sym___attribute] = ACTIONS(8153),
    [anon_sym_noreturn] = ACTIONS(8151),
    [anon_sym_LBRACK] = ACTIONS(8089),
    [anon_sym___declspec] = ACTIONS(8151),
    [anon_sym_static] = ACTIONS(8151),
    [anon_sym_auto] = ACTIONS(8151),
    [anon_sym_register] = ACTIONS(8151),
    [anon_sym_inline] = ACTIONS(8151),
    [anon_sym___inline] = ACTIONS(8153),
    [anon_sym___inline__] = ACTIONS(8151),
    [anon_sym___forceinline] = ACTIONS(8151),
    [anon_sym_thread_local] = ACTIONS(8151),
    [anon_sym___thread] = ACTIONS(8151),
    [anon_sym_CG_EXTERN] = ACTIONS(8151),
    [anon_sym_CG_INLINE] = ACTIONS(8151),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8151),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8151),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8151),
    [anon_sym_IBOutlet] = ACTIONS(8151),
    [anon_sym_IBInspectable] = ACTIONS(8151),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8151),
    [anon_sym_NS_INLINE] = ACTIONS(8151),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8151),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8151),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8151),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8151),
    [anon_sym_const] = ACTIONS(8153),
    [anon_sym_constexpr] = ACTIONS(8151),
    [anon_sym_volatile] = ACTIONS(8151),
    [anon_sym_restrict] = ACTIONS(8151),
    [anon_sym___restrict__] = ACTIONS(8151),
    [anon_sym__Atomic] = ACTIONS(8151),
    [anon_sym__Noreturn] = ACTIONS(8151),
    [anon_sym__Nonnull] = ACTIONS(8151),
    [anon_sym_nullable] = ACTIONS(8151),
    [anon_sym__Complex] = ACTIONS(8151),
    [anon_sym__Nullable] = ACTIONS(8153),
    [anon_sym__Nullable_result] = ACTIONS(8151),
    [anon_sym__Null_unspecified] = ACTIONS(8151),
    [anon_sym___autoreleasing] = ACTIONS(8151),
    [anon_sym___block] = ACTIONS(8151),
    [anon_sym___bridge] = ACTIONS(8153),
    [anon_sym___bridge_retained] = ACTIONS(8151),
    [anon_sym___bridge_transfer] = ACTIONS(8151),
    [anon_sym___complex] = ACTIONS(8151),
    [anon_sym___const] = ACTIONS(8151),
    [anon_sym___imag] = ACTIONS(8151),
    [anon_sym___kindof] = ACTIONS(8151),
    [anon_sym___nonnull] = ACTIONS(8151),
    [anon_sym___nullable] = ACTIONS(8151),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8151),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8151),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8151),
    [anon_sym___real] = ACTIONS(8151),
    [anon_sym___strong] = ACTIONS(8151),
    [anon_sym___unsafe_unretained] = ACTIONS(8151),
    [anon_sym___unused] = ACTIONS(8151),
    [anon_sym___weak] = ACTIONS(8151),
    [anon_sym_alignas] = ACTIONS(8151),
    [anon_sym__Alignas] = ACTIONS(8151),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8151),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8151),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8151),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8151),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8151),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8151),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8151),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8151),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8151),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8151),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8153),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8151),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8151),
    [anon_sym_API_AVAILABLE] = ACTIONS(8151),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8151),
    [anon_sym_API_DEPRECATED] = ACTIONS(8151),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8151),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8151),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8151),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8151),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8151),
    [anon_sym___deprecated_msg] = ACTIONS(8151),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8151),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8151),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8151),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8151),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8151),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8151),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8151),
  },
  [STATE(3263)] = {
    [sym_parameter_list] = STATE(3223),
    [anon_sym_COMMA] = ACTIONS(8155),
    [anon_sym_LPAREN2] = ACTIONS(6854),
    [anon_sym_SEMI] = ACTIONS(8155),
    [anon_sym___extension__] = ACTIONS(8155),
    [anon_sym_extern] = ACTIONS(8155),
    [anon_sym___attribute__] = ACTIONS(8155),
    [anon_sym___attribute] = ACTIONS(8157),
    [anon_sym_noreturn] = ACTIONS(8155),
    [anon_sym_LBRACK] = ACTIONS(8089),
    [anon_sym___declspec] = ACTIONS(8155),
    [anon_sym_static] = ACTIONS(8155),
    [anon_sym_auto] = ACTIONS(8155),
    [anon_sym_register] = ACTIONS(8155),
    [anon_sym_inline] = ACTIONS(8155),
    [anon_sym___inline] = ACTIONS(8157),
    [anon_sym___inline__] = ACTIONS(8155),
    [anon_sym___forceinline] = ACTIONS(8155),
    [anon_sym_thread_local] = ACTIONS(8155),
    [anon_sym___thread] = ACTIONS(8155),
    [anon_sym_CG_EXTERN] = ACTIONS(8155),
    [anon_sym_CG_INLINE] = ACTIONS(8155),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8155),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8155),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8155),
    [anon_sym_IBOutlet] = ACTIONS(8155),
    [anon_sym_IBInspectable] = ACTIONS(8155),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8155),
    [anon_sym_NS_INLINE] = ACTIONS(8155),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8155),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8155),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8155),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8155),
    [anon_sym_const] = ACTIONS(8157),
    [anon_sym_constexpr] = ACTIONS(8155),
    [anon_sym_volatile] = ACTIONS(8155),
    [anon_sym_restrict] = ACTIONS(8155),
    [anon_sym___restrict__] = ACTIONS(8155),
    [anon_sym__Atomic] = ACTIONS(8155),
    [anon_sym__Noreturn] = ACTIONS(8155),
    [anon_sym__Nonnull] = ACTIONS(8155),
    [anon_sym_nullable] = ACTIONS(8155),
    [anon_sym__Complex] = ACTIONS(8155),
    [anon_sym__Nullable] = ACTIONS(8157),
    [anon_sym__Nullable_result] = ACTIONS(8155),
    [anon_sym__Null_unspecified] = ACTIONS(8155),
    [anon_sym___autoreleasing] = ACTIONS(8155),
    [anon_sym___block] = ACTIONS(8155),
    [anon_sym___bridge] = ACTIONS(8157),
    [anon_sym___bridge_retained] = ACTIONS(8155),
    [anon_sym___bridge_transfer] = ACTIONS(8155),
    [anon_sym___complex] = ACTIONS(8155),
    [anon_sym___const] = ACTIONS(8155),
    [anon_sym___imag] = ACTIONS(8155),
    [anon_sym___kindof] = ACTIONS(8155),
    [anon_sym___nonnull] = ACTIONS(8155),
    [anon_sym___nullable] = ACTIONS(8155),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8155),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8155),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8155),
    [anon_sym___real] = ACTIONS(8155),
    [anon_sym___strong] = ACTIONS(8155),
    [anon_sym___unsafe_unretained] = ACTIONS(8155),
    [anon_sym___unused] = ACTIONS(8155),
    [anon_sym___weak] = ACTIONS(8155),
    [anon_sym_alignas] = ACTIONS(8155),
    [anon_sym__Alignas] = ACTIONS(8155),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8155),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8155),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8155),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8155),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8155),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8155),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8155),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8155),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8155),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8155),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8157),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8155),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8155),
    [anon_sym_API_AVAILABLE] = ACTIONS(8155),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8155),
    [anon_sym_API_DEPRECATED] = ACTIONS(8155),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8155),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8155),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8155),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8155),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8155),
    [anon_sym___deprecated_msg] = ACTIONS(8155),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8155),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8155),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8155),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8155),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8155),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8155),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8155),
  },
  [STATE(3264)] = {
    [anon_sym_COMMA] = ACTIONS(8159),
    [anon_sym_RPAREN] = ACTIONS(8159),
    [anon_sym_LPAREN2] = ACTIONS(8159),
    [anon_sym_SEMI] = ACTIONS(8159),
    [anon_sym___extension__] = ACTIONS(8159),
    [anon_sym_extern] = ACTIONS(8159),
    [anon_sym___attribute__] = ACTIONS(8159),
    [anon_sym___attribute] = ACTIONS(8161),
    [anon_sym_noreturn] = ACTIONS(8159),
    [anon_sym_LBRACK] = ACTIONS(8159),
    [anon_sym___declspec] = ACTIONS(8159),
    [anon_sym_static] = ACTIONS(8159),
    [anon_sym_auto] = ACTIONS(8159),
    [anon_sym_register] = ACTIONS(8159),
    [anon_sym_inline] = ACTIONS(8159),
    [anon_sym___inline] = ACTIONS(8161),
    [anon_sym___inline__] = ACTIONS(8159),
    [anon_sym___forceinline] = ACTIONS(8159),
    [anon_sym_thread_local] = ACTIONS(8159),
    [anon_sym___thread] = ACTIONS(8159),
    [anon_sym_CG_EXTERN] = ACTIONS(8159),
    [anon_sym_CG_INLINE] = ACTIONS(8159),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8159),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8159),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8159),
    [anon_sym_IBOutlet] = ACTIONS(8159),
    [anon_sym_IBInspectable] = ACTIONS(8159),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8159),
    [anon_sym_NS_INLINE] = ACTIONS(8159),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8159),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8159),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8159),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8159),
    [anon_sym_const] = ACTIONS(8161),
    [anon_sym_constexpr] = ACTIONS(8159),
    [anon_sym_volatile] = ACTIONS(8159),
    [anon_sym_restrict] = ACTIONS(8159),
    [anon_sym___restrict__] = ACTIONS(8159),
    [anon_sym__Atomic] = ACTIONS(8159),
    [anon_sym__Noreturn] = ACTIONS(8159),
    [anon_sym__Nonnull] = ACTIONS(8159),
    [anon_sym_nullable] = ACTIONS(8159),
    [anon_sym__Complex] = ACTIONS(8159),
    [anon_sym__Nullable] = ACTIONS(8161),
    [anon_sym__Nullable_result] = ACTIONS(8159),
    [anon_sym__Null_unspecified] = ACTIONS(8159),
    [anon_sym___autoreleasing] = ACTIONS(8159),
    [anon_sym___block] = ACTIONS(8159),
    [anon_sym___bridge] = ACTIONS(8161),
    [anon_sym___bridge_retained] = ACTIONS(8159),
    [anon_sym___bridge_transfer] = ACTIONS(8159),
    [anon_sym___complex] = ACTIONS(8159),
    [anon_sym___const] = ACTIONS(8159),
    [anon_sym___imag] = ACTIONS(8159),
    [anon_sym___kindof] = ACTIONS(8159),
    [anon_sym___nonnull] = ACTIONS(8159),
    [anon_sym___nullable] = ACTIONS(8159),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8159),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8159),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8159),
    [anon_sym___real] = ACTIONS(8159),
    [anon_sym___strong] = ACTIONS(8159),
    [anon_sym___unsafe_unretained] = ACTIONS(8159),
    [anon_sym___unused] = ACTIONS(8159),
    [anon_sym___weak] = ACTIONS(8159),
    [anon_sym_alignas] = ACTIONS(8159),
    [anon_sym__Alignas] = ACTIONS(8159),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8159),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8159),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8159),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8159),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8159),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8159),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8159),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8159),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8159),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8159),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8161),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8159),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8159),
    [anon_sym_API_AVAILABLE] = ACTIONS(8159),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8159),
    [anon_sym_API_DEPRECATED] = ACTIONS(8159),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8159),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8159),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8159),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8159),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8159),
    [anon_sym___deprecated_msg] = ACTIONS(8159),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8159),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8159),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8159),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8159),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8159),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8159),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8159),
  },
  [STATE(3265)] = {
    [aux_sym_method_definition_repeat2] = STATE(3265),
    [anon_sym_COMMA] = ACTIONS(8163),
    [anon_sym_SEMI] = ACTIONS(6837),
    [anon_sym___extension__] = ACTIONS(6837),
    [anon_sym_extern] = ACTIONS(6837),
    [anon_sym___attribute__] = ACTIONS(6837),
    [anon_sym___attribute] = ACTIONS(6832),
    [anon_sym_noreturn] = ACTIONS(6837),
    [anon_sym_LBRACK] = ACTIONS(6837),
    [anon_sym___declspec] = ACTIONS(6837),
    [anon_sym_static] = ACTIONS(6837),
    [anon_sym_auto] = ACTIONS(6837),
    [anon_sym_register] = ACTIONS(6837),
    [anon_sym_inline] = ACTIONS(6837),
    [anon_sym___inline] = ACTIONS(6832),
    [anon_sym___inline__] = ACTIONS(6837),
    [anon_sym___forceinline] = ACTIONS(6837),
    [anon_sym_thread_local] = ACTIONS(6837),
    [anon_sym___thread] = ACTIONS(6837),
    [anon_sym_CG_EXTERN] = ACTIONS(6837),
    [anon_sym_CG_INLINE] = ACTIONS(6837),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6837),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6837),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6837),
    [anon_sym_IBOutlet] = ACTIONS(6837),
    [anon_sym_IBInspectable] = ACTIONS(6837),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6837),
    [anon_sym_NS_INLINE] = ACTIONS(6837),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6837),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6837),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6837),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6837),
    [anon_sym_const] = ACTIONS(6832),
    [anon_sym_constexpr] = ACTIONS(6837),
    [anon_sym_volatile] = ACTIONS(6837),
    [anon_sym_restrict] = ACTIONS(6837),
    [anon_sym___restrict__] = ACTIONS(6837),
    [anon_sym__Atomic] = ACTIONS(6837),
    [anon_sym__Noreturn] = ACTIONS(6837),
    [anon_sym__Nonnull] = ACTIONS(6837),
    [anon_sym_nullable] = ACTIONS(6837),
    [anon_sym__Complex] = ACTIONS(6837),
    [anon_sym__Nullable] = ACTIONS(6832),
    [anon_sym__Nullable_result] = ACTIONS(6837),
    [anon_sym__Null_unspecified] = ACTIONS(6837),
    [anon_sym___autoreleasing] = ACTIONS(6837),
    [anon_sym___block] = ACTIONS(6837),
    [anon_sym___bridge] = ACTIONS(6832),
    [anon_sym___bridge_retained] = ACTIONS(6837),
    [anon_sym___bridge_transfer] = ACTIONS(6837),
    [anon_sym___complex] = ACTIONS(6837),
    [anon_sym___const] = ACTIONS(6837),
    [anon_sym___imag] = ACTIONS(6837),
    [anon_sym___kindof] = ACTIONS(6837),
    [anon_sym___nonnull] = ACTIONS(6837),
    [anon_sym___nullable] = ACTIONS(6837),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6837),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6837),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6837),
    [anon_sym___real] = ACTIONS(6837),
    [anon_sym___strong] = ACTIONS(6837),
    [anon_sym___unsafe_unretained] = ACTIONS(6837),
    [anon_sym___unused] = ACTIONS(6837),
    [anon_sym___weak] = ACTIONS(6837),
    [anon_sym_alignas] = ACTIONS(6837),
    [anon_sym__Alignas] = ACTIONS(6837),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6837),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6837),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6837),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6837),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6837),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6837),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6837),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6837),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6837),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6837),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6832),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6837),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6837),
    [anon_sym_API_AVAILABLE] = ACTIONS(6837),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6837),
    [anon_sym_API_DEPRECATED] = ACTIONS(6837),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6837),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6837),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6837),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6837),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6837),
    [anon_sym___deprecated_msg] = ACTIONS(6837),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6837),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6837),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6837),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6837),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6837),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6837),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6837),
  },
  [STATE(3266)] = {
    [aux_sym__type_definition_declarators_repeat1] = STATE(3266),
    [anon_sym_COMMA] = ACTIONS(8166),
    [anon_sym_SEMI] = ACTIONS(8169),
    [anon_sym___extension__] = ACTIONS(8169),
    [anon_sym_extern] = ACTIONS(8169),
    [anon_sym___attribute__] = ACTIONS(8169),
    [anon_sym___attribute] = ACTIONS(8171),
    [anon_sym_noreturn] = ACTIONS(8169),
    [anon_sym_LBRACK] = ACTIONS(8169),
    [anon_sym___declspec] = ACTIONS(8169),
    [anon_sym_static] = ACTIONS(8169),
    [anon_sym_auto] = ACTIONS(8169),
    [anon_sym_register] = ACTIONS(8169),
    [anon_sym_inline] = ACTIONS(8169),
    [anon_sym___inline] = ACTIONS(8171),
    [anon_sym___inline__] = ACTIONS(8169),
    [anon_sym___forceinline] = ACTIONS(8169),
    [anon_sym_thread_local] = ACTIONS(8169),
    [anon_sym___thread] = ACTIONS(8169),
    [anon_sym_CG_EXTERN] = ACTIONS(8169),
    [anon_sym_CG_INLINE] = ACTIONS(8169),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8169),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8169),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8169),
    [anon_sym_IBOutlet] = ACTIONS(8169),
    [anon_sym_IBInspectable] = ACTIONS(8169),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8169),
    [anon_sym_NS_INLINE] = ACTIONS(8169),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8169),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8169),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8169),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8169),
    [anon_sym_const] = ACTIONS(8171),
    [anon_sym_constexpr] = ACTIONS(8169),
    [anon_sym_volatile] = ACTIONS(8169),
    [anon_sym_restrict] = ACTIONS(8169),
    [anon_sym___restrict__] = ACTIONS(8169),
    [anon_sym__Atomic] = ACTIONS(8169),
    [anon_sym__Noreturn] = ACTIONS(8169),
    [anon_sym__Nonnull] = ACTIONS(8169),
    [anon_sym_nullable] = ACTIONS(8169),
    [anon_sym__Complex] = ACTIONS(8169),
    [anon_sym__Nullable] = ACTIONS(8171),
    [anon_sym__Nullable_result] = ACTIONS(8169),
    [anon_sym__Null_unspecified] = ACTIONS(8169),
    [anon_sym___autoreleasing] = ACTIONS(8169),
    [anon_sym___block] = ACTIONS(8169),
    [anon_sym___bridge] = ACTIONS(8171),
    [anon_sym___bridge_retained] = ACTIONS(8169),
    [anon_sym___bridge_transfer] = ACTIONS(8169),
    [anon_sym___complex] = ACTIONS(8169),
    [anon_sym___const] = ACTIONS(8169),
    [anon_sym___imag] = ACTIONS(8169),
    [anon_sym___kindof] = ACTIONS(8169),
    [anon_sym___nonnull] = ACTIONS(8169),
    [anon_sym___nullable] = ACTIONS(8169),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8169),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8169),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8169),
    [anon_sym___real] = ACTIONS(8169),
    [anon_sym___strong] = ACTIONS(8169),
    [anon_sym___unsafe_unretained] = ACTIONS(8169),
    [anon_sym___unused] = ACTIONS(8169),
    [anon_sym___weak] = ACTIONS(8169),
    [anon_sym_alignas] = ACTIONS(8169),
    [anon_sym__Alignas] = ACTIONS(8169),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8169),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8169),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8169),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8169),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8169),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8169),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8169),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8169),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8169),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8169),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8171),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8169),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8169),
    [anon_sym_API_AVAILABLE] = ACTIONS(8169),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8169),
    [anon_sym_API_DEPRECATED] = ACTIONS(8169),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8169),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8169),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8169),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8169),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8169),
    [anon_sym___deprecated_msg] = ACTIONS(8169),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8169),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8169),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8169),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8169),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8169),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8169),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8169),
  },
  [STATE(3267)] = {
    [aux_sym_c_method_parameter_repeat1] = STATE(3278),
    [anon_sym_COMMA] = ACTIONS(6794),
    [anon_sym_SEMI] = ACTIONS(6794),
    [anon_sym___extension__] = ACTIONS(6794),
    [anon_sym_extern] = ACTIONS(6794),
    [anon_sym___attribute__] = ACTIONS(6794),
    [anon_sym___attribute] = ACTIONS(6792),
    [anon_sym_noreturn] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym___declspec] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_auto] = ACTIONS(6794),
    [anon_sym_register] = ACTIONS(6794),
    [anon_sym_inline] = ACTIONS(6794),
    [anon_sym___inline] = ACTIONS(6792),
    [anon_sym___inline__] = ACTIONS(6794),
    [anon_sym___forceinline] = ACTIONS(6794),
    [anon_sym_thread_local] = ACTIONS(6794),
    [anon_sym___thread] = ACTIONS(6794),
    [anon_sym_CG_EXTERN] = ACTIONS(6794),
    [anon_sym_CG_INLINE] = ACTIONS(6794),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6794),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6794),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6794),
    [anon_sym_IBOutlet] = ACTIONS(6794),
    [anon_sym_IBInspectable] = ACTIONS(6794),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6794),
    [anon_sym_NS_INLINE] = ACTIONS(6794),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6794),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6794),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6794),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6794),
    [anon_sym_const] = ACTIONS(6792),
    [anon_sym_constexpr] = ACTIONS(6794),
    [anon_sym_volatile] = ACTIONS(6794),
    [anon_sym_restrict] = ACTIONS(6794),
    [anon_sym___restrict__] = ACTIONS(6794),
    [anon_sym__Atomic] = ACTIONS(6794),
    [anon_sym__Noreturn] = ACTIONS(6794),
    [anon_sym__Nonnull] = ACTIONS(6794),
    [anon_sym_nullable] = ACTIONS(6794),
    [anon_sym__Complex] = ACTIONS(6794),
    [anon_sym__Nullable] = ACTIONS(6792),
    [anon_sym__Nullable_result] = ACTIONS(6794),
    [anon_sym__Null_unspecified] = ACTIONS(6794),
    [anon_sym___autoreleasing] = ACTIONS(6794),
    [anon_sym___block] = ACTIONS(6794),
    [anon_sym___bridge] = ACTIONS(6792),
    [anon_sym___bridge_retained] = ACTIONS(6794),
    [anon_sym___bridge_transfer] = ACTIONS(6794),
    [anon_sym___complex] = ACTIONS(6794),
    [anon_sym___const] = ACTIONS(6794),
    [anon_sym___imag] = ACTIONS(6794),
    [anon_sym___kindof] = ACTIONS(6794),
    [anon_sym___nonnull] = ACTIONS(6794),
    [anon_sym___nullable] = ACTIONS(6794),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6794),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6794),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6794),
    [anon_sym___real] = ACTIONS(6794),
    [anon_sym___strong] = ACTIONS(6794),
    [anon_sym___unsafe_unretained] = ACTIONS(6794),
    [anon_sym___unused] = ACTIONS(6794),
    [anon_sym___weak] = ACTIONS(6794),
    [anon_sym_alignas] = ACTIONS(6794),
    [anon_sym__Alignas] = ACTIONS(6794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6794),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6794),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6794),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6794),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6794),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6794),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6794),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6794),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6794),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6794),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6792),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6794),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6794),
    [anon_sym_API_AVAILABLE] = ACTIONS(6794),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6794),
    [anon_sym_API_DEPRECATED] = ACTIONS(6794),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6794),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6794),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6794),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6794),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6794),
    [anon_sym___deprecated_msg] = ACTIONS(6794),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6794),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6794),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6794),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6794),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6794),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6794),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6794),
  },
  [STATE(3268)] = {
    [aux_sym_declaration_repeat1] = STATE(3268),
    [anon_sym_COMMA] = ACTIONS(8173),
    [anon_sym_SEMI] = ACTIONS(8176),
    [anon_sym___extension__] = ACTIONS(8176),
    [anon_sym_extern] = ACTIONS(8176),
    [anon_sym___attribute__] = ACTIONS(8176),
    [anon_sym___attribute] = ACTIONS(8178),
    [anon_sym_noreturn] = ACTIONS(8176),
    [anon_sym_LBRACK] = ACTIONS(8176),
    [anon_sym___declspec] = ACTIONS(8176),
    [anon_sym_static] = ACTIONS(8176),
    [anon_sym_auto] = ACTIONS(8176),
    [anon_sym_register] = ACTIONS(8176),
    [anon_sym_inline] = ACTIONS(8176),
    [anon_sym___inline] = ACTIONS(8178),
    [anon_sym___inline__] = ACTIONS(8176),
    [anon_sym___forceinline] = ACTIONS(8176),
    [anon_sym_thread_local] = ACTIONS(8176),
    [anon_sym___thread] = ACTIONS(8176),
    [anon_sym_CG_EXTERN] = ACTIONS(8176),
    [anon_sym_CG_INLINE] = ACTIONS(8176),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8176),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8176),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8176),
    [anon_sym_IBOutlet] = ACTIONS(8176),
    [anon_sym_IBInspectable] = ACTIONS(8176),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8176),
    [anon_sym_NS_INLINE] = ACTIONS(8176),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8176),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8176),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8176),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8176),
    [anon_sym_const] = ACTIONS(8178),
    [anon_sym_constexpr] = ACTIONS(8176),
    [anon_sym_volatile] = ACTIONS(8176),
    [anon_sym_restrict] = ACTIONS(8176),
    [anon_sym___restrict__] = ACTIONS(8176),
    [anon_sym__Atomic] = ACTIONS(8176),
    [anon_sym__Noreturn] = ACTIONS(8176),
    [anon_sym__Nonnull] = ACTIONS(8176),
    [anon_sym_nullable] = ACTIONS(8176),
    [anon_sym__Complex] = ACTIONS(8176),
    [anon_sym__Nullable] = ACTIONS(8178),
    [anon_sym__Nullable_result] = ACTIONS(8176),
    [anon_sym__Null_unspecified] = ACTIONS(8176),
    [anon_sym___autoreleasing] = ACTIONS(8176),
    [anon_sym___block] = ACTIONS(8176),
    [anon_sym___bridge] = ACTIONS(8178),
    [anon_sym___bridge_retained] = ACTIONS(8176),
    [anon_sym___bridge_transfer] = ACTIONS(8176),
    [anon_sym___complex] = ACTIONS(8176),
    [anon_sym___const] = ACTIONS(8176),
    [anon_sym___imag] = ACTIONS(8176),
    [anon_sym___kindof] = ACTIONS(8176),
    [anon_sym___nonnull] = ACTIONS(8176),
    [anon_sym___nullable] = ACTIONS(8176),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8176),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8176),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8176),
    [anon_sym___real] = ACTIONS(8176),
    [anon_sym___strong] = ACTIONS(8176),
    [anon_sym___unsafe_unretained] = ACTIONS(8176),
    [anon_sym___unused] = ACTIONS(8176),
    [anon_sym___weak] = ACTIONS(8176),
    [anon_sym_alignas] = ACTIONS(8176),
    [anon_sym__Alignas] = ACTIONS(8176),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8176),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8176),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8176),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8176),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8176),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8176),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8176),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8176),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8176),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8176),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8178),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8176),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8176),
    [anon_sym_API_AVAILABLE] = ACTIONS(8176),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8176),
    [anon_sym_API_DEPRECATED] = ACTIONS(8176),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8176),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8176),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8176),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8176),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8176),
    [anon_sym___deprecated_msg] = ACTIONS(8176),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8176),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8176),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8176),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8176),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8176),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8176),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8176),
  },
  [STATE(3269)] = {
    [sym_expression] = STATE(4401),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(5661),
    [sym_initializer_pair] = STATE(5661),
    [sym_subscript_designator] = STATE(4995),
    [sym_subscript_range_designator] = STATE(4995),
    [sym_field_designator] = STATE(4995),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_initializer_pair_repeat1] = STATE(4995),
    [sym_identifier] = ACTIONS(8180),
    [anon_sym_COMMA] = ACTIONS(8182),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8184),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_RBRACE] = ACTIONS(8186),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(8188),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3270)] = {
    [aux_sym_c_method_parameter_repeat1] = STATE(3278),
    [anon_sym_COMMA] = ACTIONS(6794),
    [anon_sym_SEMI] = ACTIONS(6794),
    [anon_sym___extension__] = ACTIONS(6794),
    [anon_sym_extern] = ACTIONS(6794),
    [anon_sym___attribute__] = ACTIONS(6794),
    [anon_sym___attribute] = ACTIONS(6792),
    [anon_sym_noreturn] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym___declspec] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_auto] = ACTIONS(6794),
    [anon_sym_register] = ACTIONS(6794),
    [anon_sym_inline] = ACTIONS(6794),
    [anon_sym___inline] = ACTIONS(6792),
    [anon_sym___inline__] = ACTIONS(6794),
    [anon_sym___forceinline] = ACTIONS(6794),
    [anon_sym_thread_local] = ACTIONS(6794),
    [anon_sym___thread] = ACTIONS(6794),
    [anon_sym_CG_EXTERN] = ACTIONS(6794),
    [anon_sym_CG_INLINE] = ACTIONS(6794),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6794),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6794),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6794),
    [anon_sym_IBOutlet] = ACTIONS(6794),
    [anon_sym_IBInspectable] = ACTIONS(6794),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6794),
    [anon_sym_NS_INLINE] = ACTIONS(6794),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6794),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6794),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6794),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6794),
    [anon_sym_const] = ACTIONS(6792),
    [anon_sym_constexpr] = ACTIONS(6794),
    [anon_sym_volatile] = ACTIONS(6794),
    [anon_sym_restrict] = ACTIONS(6794),
    [anon_sym___restrict__] = ACTIONS(6794),
    [anon_sym__Atomic] = ACTIONS(6794),
    [anon_sym__Noreturn] = ACTIONS(6794),
    [anon_sym__Nonnull] = ACTIONS(6794),
    [anon_sym_nullable] = ACTIONS(6794),
    [anon_sym__Complex] = ACTIONS(6794),
    [anon_sym__Nullable] = ACTIONS(6792),
    [anon_sym__Nullable_result] = ACTIONS(6794),
    [anon_sym__Null_unspecified] = ACTIONS(6794),
    [anon_sym___autoreleasing] = ACTIONS(6794),
    [anon_sym___block] = ACTIONS(6794),
    [anon_sym___bridge] = ACTIONS(6792),
    [anon_sym___bridge_retained] = ACTIONS(6794),
    [anon_sym___bridge_transfer] = ACTIONS(6794),
    [anon_sym___complex] = ACTIONS(6794),
    [anon_sym___const] = ACTIONS(6794),
    [anon_sym___imag] = ACTIONS(6794),
    [anon_sym___kindof] = ACTIONS(6794),
    [anon_sym___nonnull] = ACTIONS(6794),
    [anon_sym___nullable] = ACTIONS(6794),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6794),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6794),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6794),
    [anon_sym___real] = ACTIONS(6794),
    [anon_sym___strong] = ACTIONS(6794),
    [anon_sym___unsafe_unretained] = ACTIONS(6794),
    [anon_sym___unused] = ACTIONS(6794),
    [anon_sym___weak] = ACTIONS(6794),
    [anon_sym_alignas] = ACTIONS(6794),
    [anon_sym__Alignas] = ACTIONS(6794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6794),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6794),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6794),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6794),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6794),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6794),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6794),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6794),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6794),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6794),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6792),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6794),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6794),
    [anon_sym_API_AVAILABLE] = ACTIONS(6794),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6794),
    [anon_sym_API_DEPRECATED] = ACTIONS(6794),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6794),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6794),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6794),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6794),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6794),
    [anon_sym___deprecated_msg] = ACTIONS(6794),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6794),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6794),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6794),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6794),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6794),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6794),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6794),
  },
  [STATE(3271)] = {
    [sym_preproc_if] = STATE(4060),
    [sym_preproc_ifdef] = STATE(4060),
    [sym_attribute_specifier] = STATE(3330),
    [sym_attribute_declaration] = STATE(3330),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_instance_variable] = STATE(3279),
    [sym_visibility_specification] = STATE(4060),
    [sym_struct_declaration] = STATE(4060),
    [sym_atomic_declaration] = STATE(4060),
    [sym_specifier_qualifier] = STATE(3328),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_instance_variables_repeat1] = STATE(3279),
    [aux_sym_struct_declaration_repeat1] = STATE(3328),
    [sym_identifier] = ACTIONS(8190),
    [aux_sym_preproc_if_token1] = ACTIONS(8192),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(8194),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(8194),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_RBRACE] = ACTIONS(8198),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8200),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_ATprivate] = ACTIONS(8204),
    [anon_sym_ATprotected] = ACTIONS(8204),
    [anon_sym_ATpackage] = ACTIONS(8204),
    [anon_sym_ATpublic] = ACTIONS(8204),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3272)] = {
    [sym_expression] = STATE(4411),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(5457),
    [sym_initializer_pair] = STATE(5457),
    [sym_subscript_designator] = STATE(4995),
    [sym_subscript_range_designator] = STATE(4995),
    [sym_field_designator] = STATE(4995),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_initializer_pair_repeat1] = STATE(4995),
    [sym_identifier] = ACTIONS(8180),
    [anon_sym_COMMA] = ACTIONS(8206),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8184),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_RBRACE] = ACTIONS(8208),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(8188),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3273)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_parameterized_arguments] = STATE(5900),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5900),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_COMMA] = ACTIONS(8101),
    [anon_sym_RPAREN] = ACTIONS(8101),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8101),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3274)] = {
    [aux_sym_c_method_parameter_repeat1] = STATE(3270),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_SEMI] = ACTIONS(6628),
    [anon_sym___extension__] = ACTIONS(6628),
    [anon_sym_extern] = ACTIONS(6628),
    [anon_sym___attribute__] = ACTIONS(6628),
    [anon_sym___attribute] = ACTIONS(6626),
    [anon_sym_noreturn] = ACTIONS(6628),
    [anon_sym_LBRACK] = ACTIONS(6628),
    [anon_sym___declspec] = ACTIONS(6628),
    [anon_sym_static] = ACTIONS(6628),
    [anon_sym_auto] = ACTIONS(6628),
    [anon_sym_register] = ACTIONS(6628),
    [anon_sym_inline] = ACTIONS(6628),
    [anon_sym___inline] = ACTIONS(6626),
    [anon_sym___inline__] = ACTIONS(6628),
    [anon_sym___forceinline] = ACTIONS(6628),
    [anon_sym_thread_local] = ACTIONS(6628),
    [anon_sym___thread] = ACTIONS(6628),
    [anon_sym_CG_EXTERN] = ACTIONS(6628),
    [anon_sym_CG_INLINE] = ACTIONS(6628),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6628),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6628),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6628),
    [anon_sym_IBOutlet] = ACTIONS(6628),
    [anon_sym_IBInspectable] = ACTIONS(6628),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6628),
    [anon_sym_NS_INLINE] = ACTIONS(6628),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6628),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6628),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6628),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6628),
    [anon_sym_const] = ACTIONS(6626),
    [anon_sym_constexpr] = ACTIONS(6628),
    [anon_sym_volatile] = ACTIONS(6628),
    [anon_sym_restrict] = ACTIONS(6628),
    [anon_sym___restrict__] = ACTIONS(6628),
    [anon_sym__Atomic] = ACTIONS(6628),
    [anon_sym__Noreturn] = ACTIONS(6628),
    [anon_sym__Nonnull] = ACTIONS(6628),
    [anon_sym_nullable] = ACTIONS(6628),
    [anon_sym__Complex] = ACTIONS(6628),
    [anon_sym__Nullable] = ACTIONS(6626),
    [anon_sym__Nullable_result] = ACTIONS(6628),
    [anon_sym__Null_unspecified] = ACTIONS(6628),
    [anon_sym___autoreleasing] = ACTIONS(6628),
    [anon_sym___block] = ACTIONS(6628),
    [anon_sym___bridge] = ACTIONS(6626),
    [anon_sym___bridge_retained] = ACTIONS(6628),
    [anon_sym___bridge_transfer] = ACTIONS(6628),
    [anon_sym___complex] = ACTIONS(6628),
    [anon_sym___const] = ACTIONS(6628),
    [anon_sym___imag] = ACTIONS(6628),
    [anon_sym___kindof] = ACTIONS(6628),
    [anon_sym___nonnull] = ACTIONS(6628),
    [anon_sym___nullable] = ACTIONS(6628),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6628),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6628),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6628),
    [anon_sym___real] = ACTIONS(6628),
    [anon_sym___strong] = ACTIONS(6628),
    [anon_sym___unsafe_unretained] = ACTIONS(6628),
    [anon_sym___unused] = ACTIONS(6628),
    [anon_sym___weak] = ACTIONS(6628),
    [anon_sym_alignas] = ACTIONS(6628),
    [anon_sym__Alignas] = ACTIONS(6628),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6628),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6628),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6628),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6628),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6628),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6628),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6628),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6628),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6628),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6628),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6626),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6628),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6628),
    [anon_sym_API_AVAILABLE] = ACTIONS(6628),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6628),
    [anon_sym_API_DEPRECATED] = ACTIONS(6628),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6628),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6628),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6628),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6628),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6628),
    [anon_sym___deprecated_msg] = ACTIONS(6628),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6628),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6628),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6628),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6628),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6628),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6628),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6628),
  },
  [STATE(3275)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_parameterized_arguments] = STATE(5575),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5575),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_COMMA] = ACTIONS(8101),
    [anon_sym_RPAREN] = ACTIONS(8101),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8101),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3276)] = {
    [aux_sym_struct_declaration_repeat2] = STATE(3276),
    [anon_sym_COMMA] = ACTIONS(8210),
    [anon_sym_SEMI] = ACTIONS(8213),
    [anon_sym___extension__] = ACTIONS(8213),
    [anon_sym_extern] = ACTIONS(8213),
    [anon_sym___attribute__] = ACTIONS(8213),
    [anon_sym___attribute] = ACTIONS(8215),
    [anon_sym_noreturn] = ACTIONS(8213),
    [anon_sym_LBRACK] = ACTIONS(8213),
    [anon_sym___declspec] = ACTIONS(8213),
    [anon_sym_static] = ACTIONS(8213),
    [anon_sym_auto] = ACTIONS(8213),
    [anon_sym_register] = ACTIONS(8213),
    [anon_sym_inline] = ACTIONS(8213),
    [anon_sym___inline] = ACTIONS(8215),
    [anon_sym___inline__] = ACTIONS(8213),
    [anon_sym___forceinline] = ACTIONS(8213),
    [anon_sym_thread_local] = ACTIONS(8213),
    [anon_sym___thread] = ACTIONS(8213),
    [anon_sym_CG_EXTERN] = ACTIONS(8213),
    [anon_sym_CG_INLINE] = ACTIONS(8213),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8213),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8213),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8213),
    [anon_sym_IBOutlet] = ACTIONS(8213),
    [anon_sym_IBInspectable] = ACTIONS(8213),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8213),
    [anon_sym_NS_INLINE] = ACTIONS(8213),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8213),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8213),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8213),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8213),
    [anon_sym_const] = ACTIONS(8215),
    [anon_sym_constexpr] = ACTIONS(8213),
    [anon_sym_volatile] = ACTIONS(8213),
    [anon_sym_restrict] = ACTIONS(8213),
    [anon_sym___restrict__] = ACTIONS(8213),
    [anon_sym__Atomic] = ACTIONS(8213),
    [anon_sym__Noreturn] = ACTIONS(8213),
    [anon_sym__Nonnull] = ACTIONS(8213),
    [anon_sym_nullable] = ACTIONS(8213),
    [anon_sym__Complex] = ACTIONS(8213),
    [anon_sym__Nullable] = ACTIONS(8215),
    [anon_sym__Nullable_result] = ACTIONS(8213),
    [anon_sym__Null_unspecified] = ACTIONS(8213),
    [anon_sym___autoreleasing] = ACTIONS(8213),
    [anon_sym___block] = ACTIONS(8213),
    [anon_sym___bridge] = ACTIONS(8215),
    [anon_sym___bridge_retained] = ACTIONS(8213),
    [anon_sym___bridge_transfer] = ACTIONS(8213),
    [anon_sym___complex] = ACTIONS(8213),
    [anon_sym___const] = ACTIONS(8213),
    [anon_sym___imag] = ACTIONS(8213),
    [anon_sym___kindof] = ACTIONS(8213),
    [anon_sym___nonnull] = ACTIONS(8213),
    [anon_sym___nullable] = ACTIONS(8213),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8213),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8213),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8213),
    [anon_sym___real] = ACTIONS(8213),
    [anon_sym___strong] = ACTIONS(8213),
    [anon_sym___unsafe_unretained] = ACTIONS(8213),
    [anon_sym___unused] = ACTIONS(8213),
    [anon_sym___weak] = ACTIONS(8213),
    [anon_sym_alignas] = ACTIONS(8213),
    [anon_sym__Alignas] = ACTIONS(8213),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8213),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8213),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8213),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8213),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8213),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8213),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8213),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8213),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8213),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8213),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8215),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8213),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8213),
    [anon_sym_API_AVAILABLE] = ACTIONS(8213),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8213),
    [anon_sym_API_DEPRECATED] = ACTIONS(8213),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8213),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8213),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8213),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8213),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8213),
    [anon_sym___deprecated_msg] = ACTIONS(8213),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8213),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8213),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8213),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8213),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8213),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8213),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8213),
  },
  [STATE(3277)] = {
    [sym_preproc_if] = STATE(4060),
    [sym_preproc_ifdef] = STATE(4060),
    [sym_attribute_specifier] = STATE(3330),
    [sym_attribute_declaration] = STATE(3330),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_instance_variable] = STATE(3277),
    [sym_visibility_specification] = STATE(4060),
    [sym_struct_declaration] = STATE(4060),
    [sym_atomic_declaration] = STATE(4060),
    [sym_specifier_qualifier] = STATE(3328),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_instance_variables_repeat1] = STATE(3277),
    [aux_sym_struct_declaration_repeat1] = STATE(3328),
    [sym_identifier] = ACTIONS(8217),
    [aux_sym_preproc_if_token1] = ACTIONS(8220),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(8223),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(8223),
    [anon_sym___extension__] = ACTIONS(8226),
    [anon_sym___attribute__] = ACTIONS(8229),
    [anon_sym___attribute] = ACTIONS(8229),
    [anon_sym_noreturn] = ACTIONS(8226),
    [anon_sym_LBRACK] = ACTIONS(8232),
    [anon_sym_RBRACE] = ACTIONS(8235),
    [anon_sym_signed] = ACTIONS(8237),
    [anon_sym_unsigned] = ACTIONS(8237),
    [anon_sym_long] = ACTIONS(8237),
    [anon_sym_short] = ACTIONS(8237),
    [anon_sym_const] = ACTIONS(8226),
    [anon_sym_constexpr] = ACTIONS(8226),
    [anon_sym_volatile] = ACTIONS(8226),
    [anon_sym_restrict] = ACTIONS(8226),
    [anon_sym___restrict__] = ACTIONS(8226),
    [anon_sym__Atomic] = ACTIONS(8240),
    [anon_sym__Noreturn] = ACTIONS(8226),
    [anon_sym__Nonnull] = ACTIONS(8226),
    [anon_sym_nullable] = ACTIONS(8226),
    [anon_sym__Complex] = ACTIONS(8226),
    [anon_sym__Nullable] = ACTIONS(8226),
    [anon_sym__Nullable_result] = ACTIONS(8226),
    [anon_sym__Null_unspecified] = ACTIONS(8226),
    [anon_sym___autoreleasing] = ACTIONS(8226),
    [anon_sym___block] = ACTIONS(8226),
    [anon_sym___bridge] = ACTIONS(8226),
    [anon_sym___bridge_retained] = ACTIONS(8226),
    [anon_sym___bridge_transfer] = ACTIONS(8226),
    [anon_sym___complex] = ACTIONS(8226),
    [anon_sym___const] = ACTIONS(8226),
    [anon_sym___imag] = ACTIONS(8226),
    [anon_sym___kindof] = ACTIONS(8226),
    [anon_sym___nonnull] = ACTIONS(8226),
    [anon_sym___nullable] = ACTIONS(8226),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8226),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8226),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8226),
    [anon_sym___real] = ACTIONS(8226),
    [anon_sym___strong] = ACTIONS(8226),
    [anon_sym___unsafe_unretained] = ACTIONS(8226),
    [anon_sym___unused] = ACTIONS(8226),
    [anon_sym___weak] = ACTIONS(8226),
    [anon_sym_alignas] = ACTIONS(8243),
    [anon_sym__Alignas] = ACTIONS(8243),
    [sym_primitive_type] = ACTIONS(8246),
    [anon_sym_enum] = ACTIONS(8249),
    [anon_sym_struct] = ACTIONS(8252),
    [anon_sym_union] = ACTIONS(8255),
    [anon_sym_in] = ACTIONS(8258),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8261),
    [anon_sym___typeof] = ACTIONS(8261),
    [anon_sym_typeof] = ACTIONS(8261),
    [anon_sym_ATprivate] = ACTIONS(8264),
    [anon_sym_ATprotected] = ACTIONS(8264),
    [anon_sym_ATpackage] = ACTIONS(8264),
    [anon_sym_ATpublic] = ACTIONS(8264),
    [anon_sym_BOOL] = ACTIONS(8267),
    [anon_sym_IMP] = ACTIONS(8267),
    [anon_sym_SEL] = ACTIONS(8267),
    [anon_sym_Class] = ACTIONS(8267),
    [anon_sym_id] = ACTIONS(8267),
    [anon_sym_out] = ACTIONS(8258),
    [anon_sym_inout] = ACTIONS(8258),
    [anon_sym_bycopy] = ACTIONS(8258),
    [anon_sym_byref] = ACTIONS(8258),
    [anon_sym_oneway] = ACTIONS(8258),
  },
  [STATE(3278)] = {
    [aux_sym_c_method_parameter_repeat1] = STATE(3278),
    [anon_sym_COMMA] = ACTIONS(8270),
    [anon_sym_SEMI] = ACTIONS(6848),
    [anon_sym___extension__] = ACTIONS(6848),
    [anon_sym_extern] = ACTIONS(6848),
    [anon_sym___attribute__] = ACTIONS(6848),
    [anon_sym___attribute] = ACTIONS(6843),
    [anon_sym_noreturn] = ACTIONS(6848),
    [anon_sym_LBRACK] = ACTIONS(6848),
    [anon_sym___declspec] = ACTIONS(6848),
    [anon_sym_static] = ACTIONS(6848),
    [anon_sym_auto] = ACTIONS(6848),
    [anon_sym_register] = ACTIONS(6848),
    [anon_sym_inline] = ACTIONS(6848),
    [anon_sym___inline] = ACTIONS(6843),
    [anon_sym___inline__] = ACTIONS(6848),
    [anon_sym___forceinline] = ACTIONS(6848),
    [anon_sym_thread_local] = ACTIONS(6848),
    [anon_sym___thread] = ACTIONS(6848),
    [anon_sym_CG_EXTERN] = ACTIONS(6848),
    [anon_sym_CG_INLINE] = ACTIONS(6848),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(6848),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(6848),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(6848),
    [anon_sym_IBOutlet] = ACTIONS(6848),
    [anon_sym_IBInspectable] = ACTIONS(6848),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(6848),
    [anon_sym_NS_INLINE] = ACTIONS(6848),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(6848),
    [anon_sym_OBJC_EXPORT] = ACTIONS(6848),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(6848),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(6848),
    [anon_sym_const] = ACTIONS(6843),
    [anon_sym_constexpr] = ACTIONS(6848),
    [anon_sym_volatile] = ACTIONS(6848),
    [anon_sym_restrict] = ACTIONS(6848),
    [anon_sym___restrict__] = ACTIONS(6848),
    [anon_sym__Atomic] = ACTIONS(6848),
    [anon_sym__Noreturn] = ACTIONS(6848),
    [anon_sym__Nonnull] = ACTIONS(6848),
    [anon_sym_nullable] = ACTIONS(6848),
    [anon_sym__Complex] = ACTIONS(6848),
    [anon_sym__Nullable] = ACTIONS(6843),
    [anon_sym__Nullable_result] = ACTIONS(6848),
    [anon_sym__Null_unspecified] = ACTIONS(6848),
    [anon_sym___autoreleasing] = ACTIONS(6848),
    [anon_sym___block] = ACTIONS(6848),
    [anon_sym___bridge] = ACTIONS(6843),
    [anon_sym___bridge_retained] = ACTIONS(6848),
    [anon_sym___bridge_transfer] = ACTIONS(6848),
    [anon_sym___complex] = ACTIONS(6848),
    [anon_sym___const] = ACTIONS(6848),
    [anon_sym___imag] = ACTIONS(6848),
    [anon_sym___kindof] = ACTIONS(6848),
    [anon_sym___nonnull] = ACTIONS(6848),
    [anon_sym___nullable] = ACTIONS(6848),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6848),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6848),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6848),
    [anon_sym___real] = ACTIONS(6848),
    [anon_sym___strong] = ACTIONS(6848),
    [anon_sym___unsafe_unretained] = ACTIONS(6848),
    [anon_sym___unused] = ACTIONS(6848),
    [anon_sym___weak] = ACTIONS(6848),
    [anon_sym_alignas] = ACTIONS(6848),
    [anon_sym__Alignas] = ACTIONS(6848),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(6848),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(6848),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(6848),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(6848),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(6848),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(6848),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(6848),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(6848),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(6848),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(6848),
    [anon_sym_NS_AVAILABLE] = ACTIONS(6843),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(6848),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(6848),
    [anon_sym_API_AVAILABLE] = ACTIONS(6848),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(6848),
    [anon_sym_API_DEPRECATED] = ACTIONS(6848),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(6848),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(6848),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(6848),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(6848),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(6848),
    [anon_sym___deprecated_msg] = ACTIONS(6848),
    [anon_sym___deprecated_enum_msg] = ACTIONS(6848),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(6848),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(6848),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(6848),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(6848),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(6848),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(6848),
  },
  [STATE(3279)] = {
    [sym_preproc_if] = STATE(4060),
    [sym_preproc_ifdef] = STATE(4060),
    [sym_attribute_specifier] = STATE(3330),
    [sym_attribute_declaration] = STATE(3330),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_instance_variable] = STATE(3277),
    [sym_visibility_specification] = STATE(4060),
    [sym_struct_declaration] = STATE(4060),
    [sym_atomic_declaration] = STATE(4060),
    [sym_specifier_qualifier] = STATE(3328),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_instance_variables_repeat1] = STATE(3277),
    [aux_sym_struct_declaration_repeat1] = STATE(3328),
    [sym_identifier] = ACTIONS(8190),
    [aux_sym_preproc_if_token1] = ACTIONS(8192),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(8194),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(8194),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_RBRACE] = ACTIONS(8273),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8200),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_ATprivate] = ACTIONS(8204),
    [anon_sym_ATprotected] = ACTIONS(8204),
    [anon_sym_ATpackage] = ACTIONS(8204),
    [anon_sym_ATpublic] = ACTIONS(8204),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3280)] = {
    [sym_expression] = STATE(4474),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(5852),
    [sym_initializer_pair] = STATE(5852),
    [sym_subscript_designator] = STATE(4995),
    [sym_subscript_range_designator] = STATE(4995),
    [sym_field_designator] = STATE(4995),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_initializer_pair_repeat1] = STATE(4995),
    [sym_identifier] = ACTIONS(8180),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8184),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_RBRACE] = ACTIONS(8275),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(8188),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3281)] = {
    [sym_expression] = STATE(4474),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(5852),
    [sym_initializer_pair] = STATE(5852),
    [sym_subscript_designator] = STATE(4995),
    [sym_subscript_range_designator] = STATE(4995),
    [sym_field_designator] = STATE(4995),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_initializer_pair_repeat1] = STATE(4995),
    [sym_identifier] = ACTIONS(8180),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8184),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_RBRACE] = ACTIONS(8277),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(8188),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3282)] = {
    [sym_expression] = STATE(4474),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(5852),
    [sym_initializer_pair] = STATE(5852),
    [sym_subscript_designator] = STATE(4995),
    [sym_subscript_range_designator] = STATE(4995),
    [sym_field_designator] = STATE(4995),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_initializer_pair_repeat1] = STATE(4995),
    [sym_identifier] = ACTIONS(8180),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8184),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_RBRACE] = ACTIONS(8279),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(8188),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3283)] = {
    [sym_attribute_specifier] = STATE(3248),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_compound_statement] = STATE(1650),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_parameter_list] = STATE(5077),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3214),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(4945),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3214),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(8099),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_LBRACE] = ACTIONS(8103),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_ATautoreleasepool] = ACTIONS(8105),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7987),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3284)] = {
    [sym_expression] = STATE(4474),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(5852),
    [sym_initializer_pair] = STATE(5852),
    [sym_subscript_designator] = STATE(4995),
    [sym_subscript_range_designator] = STATE(4995),
    [sym_field_designator] = STATE(4995),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_initializer_pair_repeat1] = STATE(4995),
    [sym_identifier] = ACTIONS(8180),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8184),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_RBRACE] = ACTIONS(8281),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(8188),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3285)] = {
    [anon_sym_COMMA] = ACTIONS(8213),
    [anon_sym_SEMI] = ACTIONS(8213),
    [anon_sym___extension__] = ACTIONS(8213),
    [anon_sym_extern] = ACTIONS(8213),
    [anon_sym___attribute__] = ACTIONS(8213),
    [anon_sym___attribute] = ACTIONS(8215),
    [anon_sym_noreturn] = ACTIONS(8213),
    [anon_sym_LBRACK] = ACTIONS(8213),
    [anon_sym___declspec] = ACTIONS(8213),
    [anon_sym_static] = ACTIONS(8213),
    [anon_sym_auto] = ACTIONS(8213),
    [anon_sym_register] = ACTIONS(8213),
    [anon_sym_inline] = ACTIONS(8213),
    [anon_sym___inline] = ACTIONS(8215),
    [anon_sym___inline__] = ACTIONS(8213),
    [anon_sym___forceinline] = ACTIONS(8213),
    [anon_sym_thread_local] = ACTIONS(8213),
    [anon_sym___thread] = ACTIONS(8213),
    [anon_sym_CG_EXTERN] = ACTIONS(8213),
    [anon_sym_CG_INLINE] = ACTIONS(8213),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8213),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8213),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8213),
    [anon_sym_IBOutlet] = ACTIONS(8213),
    [anon_sym_IBInspectable] = ACTIONS(8213),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8213),
    [anon_sym_NS_INLINE] = ACTIONS(8213),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8213),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8213),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8213),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8213),
    [anon_sym_const] = ACTIONS(8215),
    [anon_sym_constexpr] = ACTIONS(8213),
    [anon_sym_volatile] = ACTIONS(8213),
    [anon_sym_restrict] = ACTIONS(8213),
    [anon_sym___restrict__] = ACTIONS(8213),
    [anon_sym__Atomic] = ACTIONS(8213),
    [anon_sym__Noreturn] = ACTIONS(8213),
    [anon_sym__Nonnull] = ACTIONS(8213),
    [anon_sym_nullable] = ACTIONS(8213),
    [anon_sym__Complex] = ACTIONS(8213),
    [anon_sym__Nullable] = ACTIONS(8215),
    [anon_sym__Nullable_result] = ACTIONS(8213),
    [anon_sym__Null_unspecified] = ACTIONS(8213),
    [anon_sym___autoreleasing] = ACTIONS(8213),
    [anon_sym___block] = ACTIONS(8213),
    [anon_sym___bridge] = ACTIONS(8215),
    [anon_sym___bridge_retained] = ACTIONS(8213),
    [anon_sym___bridge_transfer] = ACTIONS(8213),
    [anon_sym___complex] = ACTIONS(8213),
    [anon_sym___const] = ACTIONS(8213),
    [anon_sym___imag] = ACTIONS(8213),
    [anon_sym___kindof] = ACTIONS(8213),
    [anon_sym___nonnull] = ACTIONS(8213),
    [anon_sym___nullable] = ACTIONS(8213),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8213),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8213),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8213),
    [anon_sym___real] = ACTIONS(8213),
    [anon_sym___strong] = ACTIONS(8213),
    [anon_sym___unsafe_unretained] = ACTIONS(8213),
    [anon_sym___unused] = ACTIONS(8213),
    [anon_sym___weak] = ACTIONS(8213),
    [anon_sym_alignas] = ACTIONS(8213),
    [anon_sym__Alignas] = ACTIONS(8213),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8213),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8213),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8213),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8213),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8213),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8213),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8213),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8213),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8213),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8213),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8215),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8213),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8213),
    [anon_sym_API_AVAILABLE] = ACTIONS(8213),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8213),
    [anon_sym_API_DEPRECATED] = ACTIONS(8213),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8213),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8213),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8213),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8213),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8213),
    [anon_sym___deprecated_msg] = ACTIONS(8213),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8213),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8213),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8213),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8213),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8213),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8213),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8213),
  },
  [STATE(3286)] = {
    [anon_sym_COMMA] = ACTIONS(8283),
    [anon_sym_SEMI] = ACTIONS(8283),
    [anon_sym___extension__] = ACTIONS(8283),
    [anon_sym_extern] = ACTIONS(8283),
    [anon_sym___attribute__] = ACTIONS(8283),
    [anon_sym___attribute] = ACTIONS(8285),
    [anon_sym_noreturn] = ACTIONS(8283),
    [anon_sym_LBRACK] = ACTIONS(8283),
    [anon_sym___declspec] = ACTIONS(8283),
    [anon_sym_static] = ACTIONS(8283),
    [anon_sym_auto] = ACTIONS(8283),
    [anon_sym_register] = ACTIONS(8283),
    [anon_sym_inline] = ACTIONS(8283),
    [anon_sym___inline] = ACTIONS(8285),
    [anon_sym___inline__] = ACTIONS(8283),
    [anon_sym___forceinline] = ACTIONS(8283),
    [anon_sym_thread_local] = ACTIONS(8283),
    [anon_sym___thread] = ACTIONS(8283),
    [anon_sym_CG_EXTERN] = ACTIONS(8283),
    [anon_sym_CG_INLINE] = ACTIONS(8283),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(8283),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(8283),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(8283),
    [anon_sym_IBOutlet] = ACTIONS(8283),
    [anon_sym_IBInspectable] = ACTIONS(8283),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(8283),
    [anon_sym_NS_INLINE] = ACTIONS(8283),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(8283),
    [anon_sym_OBJC_EXPORT] = ACTIONS(8283),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(8283),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(8283),
    [anon_sym_const] = ACTIONS(8285),
    [anon_sym_constexpr] = ACTIONS(8283),
    [anon_sym_volatile] = ACTIONS(8283),
    [anon_sym_restrict] = ACTIONS(8283),
    [anon_sym___restrict__] = ACTIONS(8283),
    [anon_sym__Atomic] = ACTIONS(8283),
    [anon_sym__Noreturn] = ACTIONS(8283),
    [anon_sym__Nonnull] = ACTIONS(8283),
    [anon_sym_nullable] = ACTIONS(8283),
    [anon_sym__Complex] = ACTIONS(8283),
    [anon_sym__Nullable] = ACTIONS(8285),
    [anon_sym__Nullable_result] = ACTIONS(8283),
    [anon_sym__Null_unspecified] = ACTIONS(8283),
    [anon_sym___autoreleasing] = ACTIONS(8283),
    [anon_sym___block] = ACTIONS(8283),
    [anon_sym___bridge] = ACTIONS(8285),
    [anon_sym___bridge_retained] = ACTIONS(8283),
    [anon_sym___bridge_transfer] = ACTIONS(8283),
    [anon_sym___complex] = ACTIONS(8283),
    [anon_sym___const] = ACTIONS(8283),
    [anon_sym___imag] = ACTIONS(8283),
    [anon_sym___kindof] = ACTIONS(8283),
    [anon_sym___nonnull] = ACTIONS(8283),
    [anon_sym___nullable] = ACTIONS(8283),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8283),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8283),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8283),
    [anon_sym___real] = ACTIONS(8283),
    [anon_sym___strong] = ACTIONS(8283),
    [anon_sym___unsafe_unretained] = ACTIONS(8283),
    [anon_sym___unused] = ACTIONS(8283),
    [anon_sym___weak] = ACTIONS(8283),
    [anon_sym_alignas] = ACTIONS(8283),
    [anon_sym__Alignas] = ACTIONS(8283),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(8283),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(8283),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(8283),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(8283),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(8283),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(8283),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(8283),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(8283),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(8283),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(8283),
    [anon_sym_NS_AVAILABLE] = ACTIONS(8285),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(8283),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(8283),
    [anon_sym_API_AVAILABLE] = ACTIONS(8283),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(8283),
    [anon_sym_API_DEPRECATED] = ACTIONS(8283),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(8283),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(8283),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(8283),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(8283),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(8283),
    [anon_sym___deprecated_msg] = ACTIONS(8283),
    [anon_sym___deprecated_enum_msg] = ACTIONS(8283),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(8283),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(8283),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(8283),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(8283),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(8283),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(8283),
  },
  [STATE(3287)] = {
    [sym_attribute_specifier] = STATE(3287),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3287),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3287),
    [sym_identifier] = ACTIONS(8287),
    [anon_sym_COMMA] = ACTIONS(8290),
    [anon_sym_RPAREN] = ACTIONS(8290),
    [anon_sym_LPAREN2] = ACTIONS(8292),
    [anon_sym_STAR] = ACTIONS(8295),
    [anon_sym_CARET] = ACTIONS(8298),
    [anon_sym_GT] = ACTIONS(8290),
    [anon_sym_LT] = ACTIONS(8290),
    [anon_sym___extension__] = ACTIONS(8301),
    [anon_sym___attribute__] = ACTIONS(8304),
    [anon_sym___attribute] = ACTIONS(8304),
    [anon_sym_noreturn] = ACTIONS(8301),
    [anon_sym_LBRACK] = ACTIONS(8290),
    [anon_sym___based] = ACTIONS(8307),
    [anon_sym_signed] = ACTIONS(8310),
    [anon_sym_unsigned] = ACTIONS(8310),
    [anon_sym_long] = ACTIONS(8310),
    [anon_sym_short] = ACTIONS(8310),
    [anon_sym_const] = ACTIONS(8301),
    [anon_sym_constexpr] = ACTIONS(8301),
    [anon_sym_volatile] = ACTIONS(8301),
    [anon_sym_restrict] = ACTIONS(8301),
    [anon_sym___restrict__] = ACTIONS(8301),
    [anon_sym__Atomic] = ACTIONS(8301),
    [anon_sym__Noreturn] = ACTIONS(8301),
    [anon_sym__Nonnull] = ACTIONS(8301),
    [anon_sym_nullable] = ACTIONS(8301),
    [anon_sym__Complex] = ACTIONS(8301),
    [anon_sym__Nullable] = ACTIONS(8301),
    [anon_sym__Nullable_result] = ACTIONS(8301),
    [anon_sym__Null_unspecified] = ACTIONS(8301),
    [anon_sym___autoreleasing] = ACTIONS(8301),
    [anon_sym___block] = ACTIONS(8301),
    [anon_sym___bridge] = ACTIONS(8301),
    [anon_sym___bridge_retained] = ACTIONS(8301),
    [anon_sym___bridge_transfer] = ACTIONS(8301),
    [anon_sym___complex] = ACTIONS(8301),
    [anon_sym___const] = ACTIONS(8301),
    [anon_sym___imag] = ACTIONS(8301),
    [anon_sym___kindof] = ACTIONS(8301),
    [anon_sym___nonnull] = ACTIONS(8301),
    [anon_sym___nullable] = ACTIONS(8301),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8301),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8301),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8301),
    [anon_sym___real] = ACTIONS(8301),
    [anon_sym___strong] = ACTIONS(8301),
    [anon_sym___unsafe_unretained] = ACTIONS(8301),
    [anon_sym___unused] = ACTIONS(8301),
    [anon_sym___weak] = ACTIONS(8301),
    [anon_sym_alignas] = ACTIONS(8313),
    [anon_sym__Alignas] = ACTIONS(8313),
    [sym_primitive_type] = ACTIONS(8316),
    [anon_sym_enum] = ACTIONS(8319),
    [anon_sym_struct] = ACTIONS(8322),
    [anon_sym_union] = ACTIONS(8325),
    [anon_sym_in] = ACTIONS(8328),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8331),
    [anon_sym___typeof] = ACTIONS(8331),
    [anon_sym_typeof] = ACTIONS(8331),
    [anon_sym_BOOL] = ACTIONS(8334),
    [anon_sym_IMP] = ACTIONS(8334),
    [anon_sym_SEL] = ACTIONS(8334),
    [anon_sym_Class] = ACTIONS(8334),
    [anon_sym_id] = ACTIONS(8334),
    [anon_sym_out] = ACTIONS(8328),
    [anon_sym_inout] = ACTIONS(8328),
    [anon_sym_bycopy] = ACTIONS(8328),
    [anon_sym_byref] = ACTIONS(8328),
    [anon_sym_oneway] = ACTIONS(8328),
  },
  [STATE(3288)] = {
    [sym_attribute_specifier] = STATE(3255),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_compound_statement] = STATE(1477),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_parameter_list] = STATE(5070),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3214),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(4897),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3214),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(8099),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_LBRACE] = ACTIONS(8127),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_ATautoreleasepool] = ACTIONS(8129),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7987),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3289)] = {
    [anon_sym_COMMA] = ACTIONS(7983),
    [anon_sym_SEMI] = ACTIONS(7983),
    [anon_sym___extension__] = ACTIONS(7983),
    [anon_sym_extern] = ACTIONS(7983),
    [anon_sym___attribute__] = ACTIONS(7983),
    [anon_sym___attribute] = ACTIONS(7985),
    [anon_sym_noreturn] = ACTIONS(7983),
    [anon_sym_LBRACK] = ACTIONS(7983),
    [anon_sym___declspec] = ACTIONS(7983),
    [anon_sym_static] = ACTIONS(7983),
    [anon_sym_auto] = ACTIONS(7983),
    [anon_sym_register] = ACTIONS(7983),
    [anon_sym_inline] = ACTIONS(7983),
    [anon_sym___inline] = ACTIONS(7985),
    [anon_sym___inline__] = ACTIONS(7983),
    [anon_sym___forceinline] = ACTIONS(7983),
    [anon_sym_thread_local] = ACTIONS(7983),
    [anon_sym___thread] = ACTIONS(7983),
    [anon_sym_CG_EXTERN] = ACTIONS(7983),
    [anon_sym_CG_INLINE] = ACTIONS(7983),
    [anon_sym_FOUNDATION_EXPORT] = ACTIONS(7983),
    [anon_sym_FOUNDATION_EXTERN] = ACTIONS(7983),
    [anon_sym_FOUNDATION_STATIC_INLINE] = ACTIONS(7983),
    [anon_sym_IBOutlet] = ACTIONS(7983),
    [anon_sym_IBInspectable] = ACTIONS(7983),
    [anon_sym_IB_DESIGNABLE] = ACTIONS(7983),
    [anon_sym_NS_INLINE] = ACTIONS(7983),
    [anon_sym_NS_VALID_UNTIL_END_OF_SCOPE] = ACTIONS(7983),
    [anon_sym_OBJC_EXPORT] = ACTIONS(7983),
    [anon_sym_OBJC_ROOT_CLASS] = ACTIONS(7983),
    [anon_sym_UIKIT_EXTERN] = ACTIONS(7983),
    [anon_sym_const] = ACTIONS(7985),
    [anon_sym_constexpr] = ACTIONS(7983),
    [anon_sym_volatile] = ACTIONS(7983),
    [anon_sym_restrict] = ACTIONS(7983),
    [anon_sym___restrict__] = ACTIONS(7983),
    [anon_sym__Atomic] = ACTIONS(7983),
    [anon_sym__Noreturn] = ACTIONS(7983),
    [anon_sym__Nonnull] = ACTIONS(7983),
    [anon_sym_nullable] = ACTIONS(7983),
    [anon_sym__Complex] = ACTIONS(7983),
    [anon_sym__Nullable] = ACTIONS(7985),
    [anon_sym__Nullable_result] = ACTIONS(7983),
    [anon_sym__Null_unspecified] = ACTIONS(7983),
    [anon_sym___autoreleasing] = ACTIONS(7983),
    [anon_sym___block] = ACTIONS(7983),
    [anon_sym___bridge] = ACTIONS(7985),
    [anon_sym___bridge_retained] = ACTIONS(7983),
    [anon_sym___bridge_transfer] = ACTIONS(7983),
    [anon_sym___complex] = ACTIONS(7983),
    [anon_sym___const] = ACTIONS(7983),
    [anon_sym___imag] = ACTIONS(7983),
    [anon_sym___kindof] = ACTIONS(7983),
    [anon_sym___nonnull] = ACTIONS(7983),
    [anon_sym___nullable] = ACTIONS(7983),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7983),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7983),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7983),
    [anon_sym___real] = ACTIONS(7983),
    [anon_sym___strong] = ACTIONS(7983),
    [anon_sym___unsafe_unretained] = ACTIONS(7983),
    [anon_sym___unused] = ACTIONS(7983),
    [anon_sym___weak] = ACTIONS(7983),
    [anon_sym_alignas] = ACTIONS(7983),
    [anon_sym__Alignas] = ACTIONS(7983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_NS_AUTOMATED_REFCOUNT_UNAVAILABLE] = ACTIONS(7983),
    [anon_sym_NS_ROOT_CLASS] = ACTIONS(7983),
    [anon_sym_NS_UNAVAILABLE] = ACTIONS(7983),
    [anon_sym_NS_REQUIRES_NIL_TERMINATION] = ACTIONS(7983),
    [anon_sym_CF_RETURNS_RETAINED] = ACTIONS(7983),
    [anon_sym_CF_RETURNS_NOT_RETAINED] = ACTIONS(7983),
    [anon_sym_DEPRECATED_ATTRIBUTE] = ACTIONS(7983),
    [anon_sym_UI_APPEARANCE_SELECTOR] = ACTIONS(7983),
    [anon_sym_UNAVAILABLE_ATTRIBUTE] = ACTIONS(7983),
    [anon_sym_CF_FORMAT_FUNCTION] = ACTIONS(7983),
    [anon_sym_NS_AVAILABLE] = ACTIONS(7985),
    [anon_sym___IOS_AVAILABLE] = ACTIONS(7983),
    [anon_sym_NS_AVAILABLE_IOS] = ACTIONS(7983),
    [anon_sym_API_AVAILABLE] = ACTIONS(7983),
    [anon_sym_API_UNAVAILABLE] = ACTIONS(7983),
    [anon_sym_API_DEPRECATED] = ACTIONS(7983),
    [anon_sym_NS_ENUM_AVAILABLE_IOS] = ACTIONS(7983),
    [anon_sym_NS_DEPRECATED_IOS] = ACTIONS(7983),
    [anon_sym_NS_ENUM_DEPRECATED_IOS] = ACTIONS(7983),
    [anon_sym_NS_FORMAT_FUNCTION] = ACTIONS(7983),
    [anon_sym_DEPRECATED_MSG_ATTRIBUTE] = ACTIONS(7983),
    [anon_sym___deprecated_msg] = ACTIONS(7983),
    [anon_sym___deprecated_enum_msg] = ACTIONS(7983),
    [anon_sym_NS_SWIFT_NAME] = ACTIONS(7983),
    [anon_sym_NS_SWIFT_UNAVAILABLE] = ACTIONS(7983),
    [anon_sym_NS_EXTENSION_UNAVAILABLE_IOS] = ACTIONS(7983),
    [anon_sym_NS_CLASS_AVAILABLE_IOS] = ACTIONS(7983),
    [anon_sym_NS_CLASS_DEPRECATED_IOS] = ACTIONS(7983),
    [anon_sym___OSX_AVAILABLE_STARTING] = ACTIONS(7983),
  },
  [STATE(3290)] = {
    [sym_expression] = STATE(4405),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_version] = STATE(5365),
    [sym_platform] = STATE(5523),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(219),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [sym_version_number] = ACTIONS(8337),
    [anon_sym_ios] = ACTIONS(8339),
    [anon_sym_tvos] = ACTIONS(8339),
    [anon_sym_macos] = ACTIONS(8339),
    [anon_sym_macosx] = ACTIONS(8339),
    [anon_sym_watchos] = ACTIONS(8339),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3291)] = {
    [sym_expression] = STATE(4472),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_version] = STATE(5892),
    [sym_platform] = STATE(5523),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(219),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [sym_version_number] = ACTIONS(8337),
    [anon_sym_ios] = ACTIONS(8339),
    [anon_sym_tvos] = ACTIONS(8339),
    [anon_sym_macos] = ACTIONS(8339),
    [anon_sym_macosx] = ACTIONS(8339),
    [anon_sym_watchos] = ACTIONS(8339),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3292)] = {
    [sym_attribute_specifier] = STATE(3292),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3292),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3292),
    [sym_identifier] = ACTIONS(8287),
    [anon_sym_LPAREN2] = ACTIONS(8292),
    [anon_sym_STAR] = ACTIONS(8295),
    [anon_sym_CARET] = ACTIONS(8298),
    [anon_sym_LT] = ACTIONS(8290),
    [anon_sym___extension__] = ACTIONS(8301),
    [anon_sym___attribute__] = ACTIONS(8304),
    [anon_sym___attribute] = ACTIONS(8304),
    [anon_sym_noreturn] = ACTIONS(8301),
    [anon_sym_LBRACK] = ACTIONS(8290),
    [anon_sym___based] = ACTIONS(8307),
    [anon_sym_LBRACE] = ACTIONS(8290),
    [anon_sym_signed] = ACTIONS(8310),
    [anon_sym_unsigned] = ACTIONS(8310),
    [anon_sym_long] = ACTIONS(8310),
    [anon_sym_short] = ACTIONS(8310),
    [anon_sym_ATautoreleasepool] = ACTIONS(8290),
    [anon_sym_const] = ACTIONS(8301),
    [anon_sym_constexpr] = ACTIONS(8301),
    [anon_sym_volatile] = ACTIONS(8301),
    [anon_sym_restrict] = ACTIONS(8301),
    [anon_sym___restrict__] = ACTIONS(8301),
    [anon_sym__Atomic] = ACTIONS(8301),
    [anon_sym__Noreturn] = ACTIONS(8301),
    [anon_sym__Nonnull] = ACTIONS(8301),
    [anon_sym_nullable] = ACTIONS(8301),
    [anon_sym__Complex] = ACTIONS(8301),
    [anon_sym__Nullable] = ACTIONS(8301),
    [anon_sym__Nullable_result] = ACTIONS(8301),
    [anon_sym__Null_unspecified] = ACTIONS(8301),
    [anon_sym___autoreleasing] = ACTIONS(8301),
    [anon_sym___block] = ACTIONS(8301),
    [anon_sym___bridge] = ACTIONS(8301),
    [anon_sym___bridge_retained] = ACTIONS(8301),
    [anon_sym___bridge_transfer] = ACTIONS(8301),
    [anon_sym___complex] = ACTIONS(8301),
    [anon_sym___const] = ACTIONS(8301),
    [anon_sym___imag] = ACTIONS(8301),
    [anon_sym___kindof] = ACTIONS(8301),
    [anon_sym___nonnull] = ACTIONS(8301),
    [anon_sym___nullable] = ACTIONS(8301),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8301),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8301),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8301),
    [anon_sym___real] = ACTIONS(8301),
    [anon_sym___strong] = ACTIONS(8301),
    [anon_sym___unsafe_unretained] = ACTIONS(8301),
    [anon_sym___unused] = ACTIONS(8301),
    [anon_sym___weak] = ACTIONS(8301),
    [anon_sym_alignas] = ACTIONS(8313),
    [anon_sym__Alignas] = ACTIONS(8313),
    [sym_primitive_type] = ACTIONS(8316),
    [anon_sym_enum] = ACTIONS(8341),
    [anon_sym_struct] = ACTIONS(8322),
    [anon_sym_union] = ACTIONS(8325),
    [anon_sym_in] = ACTIONS(8328),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8331),
    [anon_sym___typeof] = ACTIONS(8331),
    [anon_sym_typeof] = ACTIONS(8331),
    [anon_sym_BOOL] = ACTIONS(8334),
    [anon_sym_IMP] = ACTIONS(8334),
    [anon_sym_SEL] = ACTIONS(8334),
    [anon_sym_Class] = ACTIONS(8334),
    [anon_sym_id] = ACTIONS(8334),
    [anon_sym_out] = ACTIONS(8328),
    [anon_sym_inout] = ACTIONS(8328),
    [anon_sym_bycopy] = ACTIONS(8328),
    [anon_sym_byref] = ACTIONS(8328),
    [anon_sym_oneway] = ACTIONS(8328),
  },
  [STATE(3293)] = {
    [sym_expression] = STATE(4474),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(5852),
    [sym_initializer_pair] = STATE(5852),
    [sym_subscript_designator] = STATE(4995),
    [sym_subscript_range_designator] = STATE(4995),
    [sym_field_designator] = STATE(4995),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [aux_sym_initializer_pair_repeat1] = STATE(4995),
    [sym_identifier] = ACTIONS(8180),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8184),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [anon_sym_DOT] = ACTIONS(8188),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3294)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5378),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(8344),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym___covariant] = ACTIONS(8346),
    [anon_sym___contravariant] = ACTIONS(8346),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3295)] = {
    [sym_attribute_specifier] = STATE(3275),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_parameterized_arguments] = STATE(5647),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5647),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3296)] = {
    [sym_attribute_specifier] = STATE(3273),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_parameterized_arguments] = STATE(5841),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5841),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym_LT] = ACTIONS(3008),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3297)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5981),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(8348),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3298)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(6041),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(8350),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3299)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(6515),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(8352),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3300)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(6631),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(8354),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3301)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(6273),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(8356),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3302)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5904),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(8358),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3303)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(6118),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(8360),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3304)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(6086),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(8362),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3305)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5826),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3306)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5827),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3307)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5611),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3308)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5785),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3309)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5765),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3310)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5668),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3311)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5414),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3312)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5763),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3313)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5628),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3314)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(6401),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3315)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5399),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3316)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5404),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3317)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(6396),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3318)] = {
    [sym_attribute_specifier] = STATE(3205),
    [sym_ms_based_modifier] = STATE(6497),
    [sym__declarator] = STATE(3943),
    [sym_parenthesized_declarator] = STATE(3980),
    [sym_pointer_declarator] = STATE(3980),
    [sym_function_declarator] = STATE(3980),
    [sym_array_declarator] = STATE(3980),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3205),
    [sym_block_pointer_declarator] = STATE(3980),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [sym_type_name] = STATE(5588),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_type_name_repeat1] = STATE(3205),
    [sym_identifier] = ACTIONS(7955),
    [anon_sym_LPAREN2] = ACTIONS(4482),
    [anon_sym_STAR] = ACTIONS(6459),
    [anon_sym_CARET] = ACTIONS(6461),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(7971),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3319)] = {
    [sym_compound_statement] = STATE(6116),
    [sym_expression] = STATE(4461),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6116),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3320)] = {
    [sym_expression] = STATE(4402),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_typeof_specifier] = STATE(5620),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3321)] = {
    [sym_compound_statement] = STATE(6174),
    [sym_expression] = STATE(4453),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6174),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(47),
    [anon_sym_ATautoreleasepool] = ACTIONS(53),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3322)] = {
    [sym_expression] = STATE(4463),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_typeof_specifier] = STATE(5824),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3323)] = {
    [sym_expression] = STATE(4458),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3324)] = {
    [sym_ms_based_modifier] = STATE(6199),
    [sym__declarator] = STATE(3242),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3761),
    [sym_struct_declarator] = STATE(3130),
    [sym_block_pointer_declarator] = STATE(3065),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3761),
    [sym_identifier] = ACTIONS(8364),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_COLON] = ACTIONS(8366),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3325)] = {
    [sym_type_qualifier] = STATE(3325),
    [sym_alignas_qualifier] = STATE(3672),
    [aux_sym_array_declarator_repeat1] = STATE(3325),
    [sym_identifier] = ACTIONS(8368),
    [anon_sym_LPAREN2] = ACTIONS(8370),
    [anon_sym_BANG] = ACTIONS(8370),
    [anon_sym_TILDE] = ACTIONS(8370),
    [anon_sym_DASH] = ACTIONS(8368),
    [anon_sym_PLUS] = ACTIONS(8368),
    [anon_sym_STAR] = ACTIONS(8370),
    [anon_sym_CARET] = ACTIONS(8370),
    [anon_sym_AMP] = ACTIONS(8370),
    [anon_sym___extension__] = ACTIONS(8372),
    [anon_sym_noreturn] = ACTIONS(8372),
    [anon_sym_LBRACK] = ACTIONS(8370),
    [anon_sym_RBRACK] = ACTIONS(8370),
    [anon_sym_static] = ACTIONS(8375),
    [anon_sym_const] = ACTIONS(8372),
    [anon_sym_constexpr] = ACTIONS(8372),
    [anon_sym_volatile] = ACTIONS(8372),
    [anon_sym_restrict] = ACTIONS(8372),
    [anon_sym___restrict__] = ACTIONS(8372),
    [anon_sym__Atomic] = ACTIONS(8372),
    [anon_sym__Noreturn] = ACTIONS(8372),
    [anon_sym__Nonnull] = ACTIONS(8372),
    [anon_sym_nullable] = ACTIONS(8372),
    [anon_sym__Complex] = ACTIONS(8372),
    [anon_sym__Nullable] = ACTIONS(8372),
    [anon_sym__Nullable_result] = ACTIONS(8372),
    [anon_sym__Null_unspecified] = ACTIONS(8372),
    [anon_sym___autoreleasing] = ACTIONS(8372),
    [anon_sym___block] = ACTIONS(8372),
    [anon_sym___bridge] = ACTIONS(8372),
    [anon_sym___bridge_retained] = ACTIONS(8372),
    [anon_sym___bridge_transfer] = ACTIONS(8372),
    [anon_sym___complex] = ACTIONS(8372),
    [anon_sym___const] = ACTIONS(8372),
    [anon_sym___imag] = ACTIONS(8372),
    [anon_sym___kindof] = ACTIONS(8372),
    [anon_sym___nonnull] = ACTIONS(8372),
    [anon_sym___nullable] = ACTIONS(8372),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8372),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8372),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8372),
    [anon_sym___real] = ACTIONS(8372),
    [anon_sym___strong] = ACTIONS(8372),
    [anon_sym___unsafe_unretained] = ACTIONS(8372),
    [anon_sym___unused] = ACTIONS(8372),
    [anon_sym___weak] = ACTIONS(8372),
    [anon_sym_alignas] = ACTIONS(8378),
    [anon_sym__Alignas] = ACTIONS(8378),
    [anon_sym_struct] = ACTIONS(8368),
    [anon_sym_in] = ACTIONS(8368),
    [anon_sym_DASH_DASH] = ACTIONS(8370),
    [anon_sym_PLUS_PLUS] = ACTIONS(8370),
    [anon_sym_sizeof] = ACTIONS(8368),
    [anon_sym___alignof__] = ACTIONS(8368),
    [anon_sym___alignof] = ACTIONS(8368),
    [anon_sym__alignof] = ACTIONS(8368),
    [anon_sym_alignof] = ACTIONS(8368),
    [anon_sym__Alignof] = ACTIONS(8368),
    [anon_sym_offsetof] = ACTIONS(8368),
    [anon_sym__Generic] = ACTIONS(8368),
    [anon_sym_asm] = ACTIONS(8368),
    [anon_sym___asm__] = ACTIONS(8368),
    [anon_sym___asm] = ACTIONS(8368),
    [sym_number_literal] = ACTIONS(8370),
    [anon_sym_L_SQUOTE] = ACTIONS(8370),
    [anon_sym_u_SQUOTE] = ACTIONS(8370),
    [anon_sym_U_SQUOTE] = ACTIONS(8370),
    [anon_sym_u8_SQUOTE] = ACTIONS(8370),
    [anon_sym_SQUOTE] = ACTIONS(8370),
    [anon_sym_AT] = ACTIONS(8368),
    [anon_sym_DQUOTE] = ACTIONS(8370),
    [anon_sym_L_DQUOTE] = ACTIONS(8370),
    [anon_sym_u_DQUOTE] = ACTIONS(8370),
    [anon_sym_U_DQUOTE] = ACTIONS(8370),
    [anon_sym_u8_DQUOTE] = ACTIONS(8370),
    [sym_true] = ACTIONS(8368),
    [sym_false] = ACTIONS(8368),
    [anon_sym_NULL] = ACTIONS(8368),
    [anon_sym_nullptr] = ACTIONS(8368),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(8370),
    [anon_sym_ATavailable] = ACTIONS(8370),
    [anon_sym___builtin_available] = ACTIONS(8368),
    [anon_sym_va_arg] = ACTIONS(8368),
    [anon_sym_ATencode] = ACTIONS(8370),
    [anon_sym_id] = ACTIONS(8368),
  },
  [STATE(3326)] = {
    [sym_ms_based_modifier] = STATE(6199),
    [sym__declarator] = STATE(3242),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3761),
    [sym_struct_declarator] = STATE(3185),
    [sym_block_pointer_declarator] = STATE(3065),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3761),
    [sym_identifier] = ACTIONS(8364),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_COLON] = ACTIONS(8366),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3327)] = {
    [sym_ms_based_modifier] = STATE(6199),
    [sym__declarator] = STATE(3242),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3761),
    [sym_struct_declarator] = STATE(3073),
    [sym_block_pointer_declarator] = STATE(3065),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3761),
    [sym_identifier] = ACTIONS(8364),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_COLON] = ACTIONS(8366),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3328)] = {
    [sym_ms_based_modifier] = STATE(6199),
    [sym__declarator] = STATE(3242),
    [sym_parenthesized_declarator] = STATE(3065),
    [sym_pointer_declarator] = STATE(3065),
    [sym_function_declarator] = STATE(3065),
    [sym_array_declarator] = STATE(3065),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3761),
    [sym_struct_declarator] = STATE(3121),
    [sym_block_pointer_declarator] = STATE(3065),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3761),
    [sym_identifier] = ACTIONS(8364),
    [anon_sym_LPAREN2] = ACTIONS(6475),
    [anon_sym_STAR] = ACTIONS(6477),
    [anon_sym_CARET] = ACTIONS(6479),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym___based] = ACTIONS(4092),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(7959),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_COLON] = ACTIONS(8366),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3329)] = {
    [sym_string_literal] = STATE(1916),
    [aux_sym_sized_type_specifier_repeat1] = STATE(2961),
    [aux_sym_generic_specifier_repeat1] = STATE(2908),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4026),
    [anon_sym_COMMA] = ACTIONS(4026),
    [anon_sym_RPAREN] = ACTIONS(4506),
    [anon_sym_LPAREN2] = ACTIONS(4496),
    [anon_sym_DASH] = ACTIONS(4024),
    [anon_sym_PLUS] = ACTIONS(4024),
    [anon_sym_STAR] = ACTIONS(4500),
    [anon_sym_SLASH] = ACTIONS(4024),
    [anon_sym_PERCENT] = ACTIONS(4024),
    [anon_sym_PIPE_PIPE] = ACTIONS(4026),
    [anon_sym_AMP_AMP] = ACTIONS(4026),
    [anon_sym_PIPE] = ACTIONS(4024),
    [anon_sym_CARET] = ACTIONS(4500),
    [anon_sym_AMP] = ACTIONS(4024),
    [anon_sym_EQ_EQ] = ACTIONS(4026),
    [anon_sym_BANG_EQ] = ACTIONS(4026),
    [anon_sym_GT] = ACTIONS(4024),
    [anon_sym_GT_EQ] = ACTIONS(4026),
    [anon_sym_LT_EQ] = ACTIONS(4026),
    [anon_sym_LT] = ACTIONS(4503),
    [anon_sym_LT_LT] = ACTIONS(4024),
    [anon_sym_GT_GT] = ACTIONS(4024),
    [anon_sym___extension__] = ACTIONS(6134),
    [anon_sym_noreturn] = ACTIONS(6134),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym_signed] = ACTIONS(8381),
    [anon_sym_unsigned] = ACTIONS(8381),
    [anon_sym_long] = ACTIONS(8381),
    [anon_sym_short] = ACTIONS(8381),
    [anon_sym_EQ] = ACTIONS(8383),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(6134),
    [anon_sym_volatile] = ACTIONS(6134),
    [anon_sym_restrict] = ACTIONS(6134),
    [anon_sym___restrict__] = ACTIONS(6134),
    [anon_sym__Atomic] = ACTIONS(6134),
    [anon_sym__Noreturn] = ACTIONS(6134),
    [anon_sym__Nonnull] = ACTIONS(6134),
    [anon_sym_nullable] = ACTIONS(6134),
    [anon_sym__Complex] = ACTIONS(6134),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(6134),
    [anon_sym__Null_unspecified] = ACTIONS(6134),
    [anon_sym___autoreleasing] = ACTIONS(6134),
    [anon_sym___block] = ACTIONS(6134),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(6134),
    [anon_sym___bridge_transfer] = ACTIONS(6134),
    [anon_sym___complex] = ACTIONS(6134),
    [anon_sym___const] = ACTIONS(6134),
    [anon_sym___imag] = ACTIONS(6134),
    [anon_sym___kindof] = ACTIONS(6134),
    [anon_sym___nonnull] = ACTIONS(6134),
    [anon_sym___nullable] = ACTIONS(6134),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6134),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6134),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6134),
    [anon_sym___real] = ACTIONS(6134),
    [anon_sym___strong] = ACTIONS(6134),
    [anon_sym___unsafe_unretained] = ACTIONS(6134),
    [anon_sym___unused] = ACTIONS(6134),
    [anon_sym___weak] = ACTIONS(6134),
    [anon_sym_alignas] = ACTIONS(6134),
    [anon_sym__Alignas] = ACTIONS(6134),
    [anon_sym_QMARK] = ACTIONS(4026),
    [anon_sym_STAR_EQ] = ACTIONS(8385),
    [anon_sym_SLASH_EQ] = ACTIONS(8385),
    [anon_sym_PERCENT_EQ] = ACTIONS(8385),
    [anon_sym_PLUS_EQ] = ACTIONS(8385),
    [anon_sym_DASH_EQ] = ACTIONS(8385),
    [anon_sym_LT_LT_EQ] = ACTIONS(8385),
    [anon_sym_GT_GT_EQ] = ACTIONS(8385),
    [anon_sym_AMP_EQ] = ACTIONS(8385),
    [anon_sym_CARET_EQ] = ACTIONS(8385),
    [anon_sym_PIPE_EQ] = ACTIONS(8385),
    [anon_sym_DASH_DASH] = ACTIONS(4026),
    [anon_sym_PLUS_PLUS] = ACTIONS(4026),
    [anon_sym_DOT] = ACTIONS(4024),
    [anon_sym_DASH_GT] = ACTIONS(4026),
    [anon_sym_AT] = ACTIONS(4517),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(3330)] = {
    [sym_preproc_if] = STATE(4060),
    [sym_preproc_ifdef] = STATE(4060),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_instance_variable] = STATE(4047),
    [sym_visibility_specification] = STATE(4060),
    [sym_struct_declaration] = STATE(4060),
    [sym_atomic_declaration] = STATE(4060),
    [sym_specifier_qualifier] = STATE(3328),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3328),
    [sym_identifier] = ACTIONS(8190),
    [aux_sym_preproc_if_token1] = ACTIONS(8192),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(8194),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(8194),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8200),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_ATprivate] = ACTIONS(8204),
    [anon_sym_ATprotected] = ACTIONS(8204),
    [anon_sym_ATpackage] = ACTIONS(8204),
    [anon_sym_ATpublic] = ACTIONS(8204),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3331)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3332)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3333)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3334)] = {
    [sym_expression] = STATE(4473),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(5848),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3335)] = {
    [sym_expression] = STATE(4477),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6107),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8395),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3336)] = {
    [sym_expression] = STATE(4530),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_dictionary_pair] = STATE(5395),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACE] = ACTIONS(8397),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3337)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3338)] = {
    [sym_expression] = STATE(4466),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6419),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8399),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3339)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3340)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3341)] = {
    [sym_expression] = STATE(4481),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6706),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8425),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3342)] = {
    [sym_expression] = STATE(4484),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6425),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8427),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3343)] = {
    [sym_expression] = STATE(4480),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6699),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_RPAREN] = ACTIONS(8429),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3344)] = {
    [sym_expression] = STATE(4432),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6140),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8431),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3345)] = {
    [sym_expression] = STATE(4441),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6590),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8433),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3346)] = {
    [sym_expression] = STATE(1733),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_initializer_list] = STATE(2834),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3347)] = {
    [sym_expression] = STATE(4460),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6599),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8445),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3348)] = {
    [sym_expression] = STATE(4530),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_dictionary_pair] = STATE(5898),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACE] = ACTIONS(8447),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3349)] = {
    [sym_expression] = STATE(4380),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5837),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(8455),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3350)] = {
    [sym_expression] = STATE(1728),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_initializer_list] = STATE(2839),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3351)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3352)] = {
    [sym_expression] = STATE(4455),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6123),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8463),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3353)] = {
    [sym_expression] = STATE(4450),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6250),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8465),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3354)] = {
    [sym_expression] = STATE(4376),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(8471),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_class] = ACTIONS(8473),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3355)] = {
    [sym_expression] = STATE(4530),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_dictionary_pair] = STATE(5587),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACE] = ACTIONS(8475),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3356)] = {
    [sym_expression] = STATE(4451),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6531),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8477),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3357)] = {
    [sym_expression] = STATE(4420),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6232),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8479),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3358)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3359)] = {
    [sym_expression] = STATE(4376),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(8471),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_class] = ACTIONS(8489),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3360)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3361)] = {
    [sym_expression] = STATE(4376),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(8471),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_class] = ACTIONS(8501),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3362)] = {
    [sym_expression] = STATE(4448),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5952),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_RPAREN] = ACTIONS(8503),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3363)] = {
    [sym_attribute] = STATE(5468),
    [sym_expression] = STATE(4442),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8505),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3364)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3365)] = {
    [sym_expression] = STATE(4475),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6661),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8507),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3366)] = {
    [sym_expression] = STATE(4443),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6289),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8509),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3367)] = {
    [sym_expression] = STATE(4454),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5968),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_RPAREN] = ACTIONS(8511),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3368)] = {
    [sym_expression] = STATE(2587),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_initializer_list] = STATE(2839),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3369)] = {
    [sym_expression] = STATE(4478),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6686),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8513),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3370)] = {
    [sym_expression] = STATE(4456),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6134),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8515),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3371)] = {
    [sym_expression] = STATE(4449),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5958),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8517),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3372)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3373)] = {
    [sym_expression] = STATE(4467),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6386),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8527),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3374)] = {
    [sym_expression] = STATE(4438),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6355),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8529),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3375)] = {
    [sym_expression] = STATE(4530),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_dictionary_pair] = STATE(5898),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACE] = ACTIONS(8531),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3376)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_initializer_list] = STATE(1470),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_LBRACE] = ACTIONS(2298),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3377)] = {
    [sym_expression] = STATE(4440),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6367),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8537),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3378)] = {
    [sym_expression] = STATE(4482),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6728),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8539),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3379)] = {
    [sym_expression] = STATE(4479),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6693),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_RPAREN] = ACTIONS(8541),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3380)] = {
    [sym_expression] = STATE(4471),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6525),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_COLON] = ACTIONS(8543),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3381)] = {
    [sym_expression] = STATE(4530),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_dictionary_pair] = STATE(5898),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACE] = ACTIONS(8545),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3382)] = {
    [sym_expression] = STATE(2609),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_initializer_list] = STATE(2834),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3383)] = {
    [sym_expression] = STATE(4530),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_dictionary_pair] = STATE(5898),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACE] = ACTIONS(8547),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3384)] = {
    [sym_expression] = STATE(4469),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6316),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8549),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3385)] = {
    [sym_expression] = STATE(4376),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(8471),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_class] = ACTIONS(8551),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3386)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_initializer_list] = STATE(1656),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3387)] = {
    [sym_expression] = STATE(4376),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(8471),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_class] = ACTIONS(8555),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3388)] = {
    [sym_attribute] = STATE(5398),
    [sym_expression] = STATE(4442),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8505),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3389)] = {
    [sym_attribute] = STATE(5444),
    [sym_expression] = STATE(4442),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8505),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3390)] = {
    [sym_expression] = STATE(4476),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_initializer_list] = STATE(5867),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_LBRACE] = ACTIONS(2888),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3391)] = {
    [sym_expression] = STATE(4452),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(6538),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_RPAREN] = ACTIONS(8557),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3392)] = {
    [sym_attribute] = STATE(5467),
    [sym_expression] = STATE(4442),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8505),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3393)] = {
    [sym_expression] = STATE(4378),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(8559),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3394)] = {
    [sym_expression] = STATE(4445),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5933),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_RPAREN] = ACTIONS(8561),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3395)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8563),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3396)] = {
    [sym_expression] = STATE(4506),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_RPAREN] = ACTIONS(8565),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3397)] = {
    [sym_expression] = STATE(4483),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(8567),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3398)] = {
    [sym_expression] = STATE(1561),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8569),
    [anon_sym_LBRACE] = ACTIONS(8571),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(8573),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3399)] = {
    [sym_expression] = STATE(4545),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8575),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3400)] = {
    [sym_expression] = STATE(4293),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8569),
    [anon_sym_LBRACE] = ACTIONS(8571),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(8577),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3401)] = {
    [sym_expression] = STATE(4399),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_RPAREN] = ACTIONS(8579),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3402)] = {
    [sym_type_qualifier] = STATE(3402),
    [sym_alignas_qualifier] = STATE(3672),
    [aux_sym__type_definition_type_repeat1] = STATE(3402),
    [sym_identifier] = ACTIONS(6292),
    [anon_sym_LPAREN2] = ACTIONS(6294),
    [anon_sym_BANG] = ACTIONS(6294),
    [anon_sym_TILDE] = ACTIONS(6294),
    [anon_sym_DASH] = ACTIONS(6292),
    [anon_sym_PLUS] = ACTIONS(6292),
    [anon_sym_STAR] = ACTIONS(6294),
    [anon_sym_CARET] = ACTIONS(6294),
    [anon_sym_AMP] = ACTIONS(6294),
    [anon_sym___extension__] = ACTIONS(8581),
    [anon_sym_noreturn] = ACTIONS(8581),
    [anon_sym_LBRACK] = ACTIONS(6294),
    [anon_sym_const] = ACTIONS(8581),
    [anon_sym_constexpr] = ACTIONS(8581),
    [anon_sym_volatile] = ACTIONS(8581),
    [anon_sym_restrict] = ACTIONS(8581),
    [anon_sym___restrict__] = ACTIONS(8581),
    [anon_sym__Atomic] = ACTIONS(8581),
    [anon_sym__Noreturn] = ACTIONS(8581),
    [anon_sym__Nonnull] = ACTIONS(8581),
    [anon_sym_nullable] = ACTIONS(8581),
    [anon_sym__Complex] = ACTIONS(8581),
    [anon_sym__Nullable] = ACTIONS(8581),
    [anon_sym__Nullable_result] = ACTIONS(8581),
    [anon_sym__Null_unspecified] = ACTIONS(8581),
    [anon_sym___autoreleasing] = ACTIONS(8581),
    [anon_sym___block] = ACTIONS(8581),
    [anon_sym___bridge] = ACTIONS(8581),
    [anon_sym___bridge_retained] = ACTIONS(8581),
    [anon_sym___bridge_transfer] = ACTIONS(8581),
    [anon_sym___complex] = ACTIONS(8581),
    [anon_sym___const] = ACTIONS(8581),
    [anon_sym___imag] = ACTIONS(8581),
    [anon_sym___kindof] = ACTIONS(8581),
    [anon_sym___nonnull] = ACTIONS(8581),
    [anon_sym___nullable] = ACTIONS(8581),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8581),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8581),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8581),
    [anon_sym___real] = ACTIONS(8581),
    [anon_sym___strong] = ACTIONS(8581),
    [anon_sym___unsafe_unretained] = ACTIONS(8581),
    [anon_sym___unused] = ACTIONS(8581),
    [anon_sym___weak] = ACTIONS(8581),
    [anon_sym_alignas] = ACTIONS(8584),
    [anon_sym__Alignas] = ACTIONS(8584),
    [anon_sym_struct] = ACTIONS(6292),
    [anon_sym_in] = ACTIONS(6292),
    [anon_sym_DASH_DASH] = ACTIONS(6294),
    [anon_sym_PLUS_PLUS] = ACTIONS(6294),
    [anon_sym_sizeof] = ACTIONS(6292),
    [anon_sym___alignof__] = ACTIONS(6292),
    [anon_sym___alignof] = ACTIONS(6292),
    [anon_sym__alignof] = ACTIONS(6292),
    [anon_sym_alignof] = ACTIONS(6292),
    [anon_sym__Alignof] = ACTIONS(6292),
    [anon_sym_offsetof] = ACTIONS(6292),
    [anon_sym__Generic] = ACTIONS(6292),
    [anon_sym_asm] = ACTIONS(6292),
    [anon_sym___asm__] = ACTIONS(6292),
    [anon_sym___asm] = ACTIONS(6292),
    [sym_number_literal] = ACTIONS(6294),
    [anon_sym_L_SQUOTE] = ACTIONS(6294),
    [anon_sym_u_SQUOTE] = ACTIONS(6294),
    [anon_sym_U_SQUOTE] = ACTIONS(6294),
    [anon_sym_u8_SQUOTE] = ACTIONS(6294),
    [anon_sym_SQUOTE] = ACTIONS(6294),
    [anon_sym_AT] = ACTIONS(6292),
    [anon_sym_DQUOTE] = ACTIONS(6294),
    [anon_sym_L_DQUOTE] = ACTIONS(6294),
    [anon_sym_u_DQUOTE] = ACTIONS(6294),
    [anon_sym_U_DQUOTE] = ACTIONS(6294),
    [anon_sym_u8_DQUOTE] = ACTIONS(6294),
    [sym_true] = ACTIONS(6292),
    [sym_false] = ACTIONS(6292),
    [anon_sym_NULL] = ACTIONS(6292),
    [anon_sym_nullptr] = ACTIONS(6292),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6294),
    [anon_sym_ATavailable] = ACTIONS(6294),
    [anon_sym___builtin_available] = ACTIONS(6292),
    [anon_sym_va_arg] = ACTIONS(6292),
    [anon_sym_ATencode] = ACTIONS(6294),
    [anon_sym_id] = ACTIONS(6292),
  },
  [STATE(3403)] = {
    [sym_expression] = STATE(4442),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3404)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8587),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3405)] = {
    [sym_expression] = STATE(4483),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(8589),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3406)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8591),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3407)] = {
    [sym_expression] = STATE(4554),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8593),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3408)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8595),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3409)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8597),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3410)] = {
    [sym_expression] = STATE(4415),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3411)] = {
    [sym_expression] = STATE(4442),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3412)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8599),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3413)] = {
    [sym_expression] = STATE(4483),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(8601),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3414)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8603),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3415)] = {
    [sym_expression] = STATE(4493),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8605),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3416)] = {
    [sym_expression] = STATE(4408),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_RPAREN] = ACTIONS(8607),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3417)] = {
    [sym_expression] = STATE(4497),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8609),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3418)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8611),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3419)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8613),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3420)] = {
    [sym_expression] = STATE(4459),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(8471),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3421)] = {
    [sym_expression] = STATE(4396),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(8471),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3422)] = {
    [sym_expression] = STATE(1763),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(8615),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3423)] = {
    [sym_expression] = STATE(4533),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8617),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3424)] = {
    [sym_expression] = STATE(4501),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8619),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3425)] = {
    [sym_expression] = STATE(4465),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5837),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3426)] = {
    [sym_expression] = STATE(4376),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(8471),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3427)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8621),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3428)] = {
    [sym_expression] = STATE(4398),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5651),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3429)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8623),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3430)] = {
    [sym_expression] = STATE(4522),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(8625),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3431)] = {
    [sym_expression] = STATE(2506),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(8471),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3432)] = {
    [sym_expression] = STATE(1913),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8569),
    [anon_sym_LBRACE] = ACTIONS(8571),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(8577),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3433)] = {
    [sym_expression] = STATE(4483),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_RBRACK] = ACTIONS(8627),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3434)] = {
    [sym_expression] = STATE(2649),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(8467),
    [anon_sym_LBRACE] = ACTIONS(8469),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(8471),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3435)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8629),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3436)] = {
    [sym_expression] = STATE(4530),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_dictionary_pair] = STATE(5898),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3437)] = {
    [sym_expression] = STATE(4442),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(8631),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3438)] = {
    [sym_expression] = STATE(4457),
    [sym__string] = STATE(1641),
    [sym_comma_expression] = STATE(5651),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3439)] = {
    [sym_expression] = STATE(4528),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_RPAREN] = ACTIONS(8633),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3440)] = {
    [sym_expression] = STATE(4346),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8569),
    [anon_sym_LBRACE] = ACTIONS(8571),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(8577),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3441)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8635),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3442)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_RBRACK] = ACTIONS(8637),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3443)] = {
    [sym_expression] = STATE(4442),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(8639),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3444)] = {
    [sym_expression] = STATE(2078),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8569),
    [anon_sym_LBRACE] = ACTIONS(8571),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(8577),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3445)] = {
    [sym_expression] = STATE(4500),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_COMMA] = ACTIONS(8641),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3446)] = {
    [sym_expression] = STATE(4442),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_generic_specifier] = STATE(5842),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(6415),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3447)] = {
    [sym_expression] = STATE(4321),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8569),
    [anon_sym_LBRACE] = ACTIONS(8571),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(8577),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3448)] = {
    [sym_expression] = STATE(4385),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3449)] = {
    [sym_expression] = STATE(1745),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3450)] = {
    [sym_expression] = STATE(1566),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3451)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3452)] = {
    [sym_expression] = STATE(4397),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3453)] = {
    [sym_expression] = STATE(4365),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3454)] = {
    [sym_expression] = STATE(1749),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3455)] = {
    [sym_expression] = STATE(1750),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3456)] = {
    [sym_expression] = STATE(1615),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3457)] = {
    [sym_expression] = STATE(4305),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3458)] = {
    [sym_expression] = STATE(1625),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(8643),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3459)] = {
    [sym_expression] = STATE(1623),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3460)] = {
    [sym_expression] = STATE(1625),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8645),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3461)] = {
    [sym_expression] = STATE(4307),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3462)] = {
    [sym_expression] = STATE(1439),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3463)] = {
    [sym_expression] = STATE(4301),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3464)] = {
    [sym_expression] = STATE(4309),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3465)] = {
    [sym_expression] = STATE(1746),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3466)] = {
    [sym_expression] = STATE(4310),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3467)] = {
    [sym_expression] = STATE(4292),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3468)] = {
    [sym_expression] = STATE(4304),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3469)] = {
    [sym_expression] = STATE(4298),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3470)] = {
    [sym_expression] = STATE(4303),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3471)] = {
    [sym_expression] = STATE(4295),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3472)] = {
    [sym_expression] = STATE(1615),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3473)] = {
    [sym_expression] = STATE(1751),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3474)] = {
    [sym_expression] = STATE(1556),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3475)] = {
    [sym_expression] = STATE(1558),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3476)] = {
    [sym_expression] = STATE(1553),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3477)] = {
    [sym_expression] = STATE(1439),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3478)] = {
    [sym_expression] = STATE(1560),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3479)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3480)] = {
    [sym_expression] = STATE(1570),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3481)] = {
    [sym_expression] = STATE(4366),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3482)] = {
    [sym_expression] = STATE(4437),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3483)] = {
    [sym_expression] = STATE(1626),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3484)] = {
    [sym_identifier] = ACTIONS(8647),
    [anon_sym_LPAREN2] = ACTIONS(8650),
    [anon_sym_BANG] = ACTIONS(8650),
    [anon_sym_TILDE] = ACTIONS(8650),
    [anon_sym_DASH] = ACTIONS(8647),
    [anon_sym_PLUS] = ACTIONS(8647),
    [anon_sym_STAR] = ACTIONS(8650),
    [anon_sym_CARET] = ACTIONS(8650),
    [anon_sym_AMP] = ACTIONS(8650),
    [anon_sym___extension__] = ACTIONS(8647),
    [anon_sym_noreturn] = ACTIONS(8647),
    [anon_sym_LBRACK] = ACTIONS(8650),
    [anon_sym_RBRACK] = ACTIONS(8653),
    [anon_sym_static] = ACTIONS(8655),
    [anon_sym_const] = ACTIONS(8647),
    [anon_sym_constexpr] = ACTIONS(8647),
    [anon_sym_volatile] = ACTIONS(8647),
    [anon_sym_restrict] = ACTIONS(8647),
    [anon_sym___restrict__] = ACTIONS(8647),
    [anon_sym__Atomic] = ACTIONS(8647),
    [anon_sym__Noreturn] = ACTIONS(8647),
    [anon_sym__Nonnull] = ACTIONS(8647),
    [anon_sym_nullable] = ACTIONS(8647),
    [anon_sym__Complex] = ACTIONS(8647),
    [anon_sym__Nullable] = ACTIONS(8647),
    [anon_sym__Nullable_result] = ACTIONS(8647),
    [anon_sym__Null_unspecified] = ACTIONS(8647),
    [anon_sym___autoreleasing] = ACTIONS(8647),
    [anon_sym___block] = ACTIONS(8647),
    [anon_sym___bridge] = ACTIONS(8647),
    [anon_sym___bridge_retained] = ACTIONS(8647),
    [anon_sym___bridge_transfer] = ACTIONS(8647),
    [anon_sym___complex] = ACTIONS(8647),
    [anon_sym___const] = ACTIONS(8647),
    [anon_sym___imag] = ACTIONS(8647),
    [anon_sym___kindof] = ACTIONS(8647),
    [anon_sym___nonnull] = ACTIONS(8647),
    [anon_sym___nullable] = ACTIONS(8647),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8647),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8647),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8647),
    [anon_sym___real] = ACTIONS(8647),
    [anon_sym___strong] = ACTIONS(8647),
    [anon_sym___unsafe_unretained] = ACTIONS(8647),
    [anon_sym___unused] = ACTIONS(8647),
    [anon_sym___weak] = ACTIONS(8647),
    [anon_sym_alignas] = ACTIONS(8647),
    [anon_sym__Alignas] = ACTIONS(8647),
    [anon_sym_struct] = ACTIONS(8647),
    [anon_sym_in] = ACTIONS(8647),
    [anon_sym_DASH_DASH] = ACTIONS(8650),
    [anon_sym_PLUS_PLUS] = ACTIONS(8650),
    [anon_sym_sizeof] = ACTIONS(8647),
    [anon_sym___alignof__] = ACTIONS(8647),
    [anon_sym___alignof] = ACTIONS(8647),
    [anon_sym__alignof] = ACTIONS(8647),
    [anon_sym_alignof] = ACTIONS(8647),
    [anon_sym__Alignof] = ACTIONS(8647),
    [anon_sym_offsetof] = ACTIONS(8647),
    [anon_sym__Generic] = ACTIONS(8647),
    [anon_sym_asm] = ACTIONS(8647),
    [anon_sym___asm__] = ACTIONS(8647),
    [anon_sym___asm] = ACTIONS(8647),
    [sym_number_literal] = ACTIONS(8650),
    [anon_sym_L_SQUOTE] = ACTIONS(8650),
    [anon_sym_u_SQUOTE] = ACTIONS(8650),
    [anon_sym_U_SQUOTE] = ACTIONS(8650),
    [anon_sym_u8_SQUOTE] = ACTIONS(8650),
    [anon_sym_SQUOTE] = ACTIONS(8650),
    [anon_sym_AT] = ACTIONS(8647),
    [anon_sym_DQUOTE] = ACTIONS(8650),
    [anon_sym_L_DQUOTE] = ACTIONS(8650),
    [anon_sym_u_DQUOTE] = ACTIONS(8650),
    [anon_sym_U_DQUOTE] = ACTIONS(8650),
    [anon_sym_u8_DQUOTE] = ACTIONS(8650),
    [sym_true] = ACTIONS(8647),
    [sym_false] = ACTIONS(8647),
    [anon_sym_NULL] = ACTIONS(8647),
    [anon_sym_nullptr] = ACTIONS(8647),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(8650),
    [anon_sym_ATavailable] = ACTIONS(8650),
    [anon_sym___builtin_available] = ACTIONS(8647),
    [anon_sym_va_arg] = ACTIONS(8647),
    [anon_sym_ATencode] = ACTIONS(8650),
    [anon_sym_id] = ACTIONS(8647),
  },
  [STATE(3485)] = {
    [sym_expression] = STATE(1436),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3486)] = {
    [sym_expression] = STATE(1441),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3487)] = {
    [sym_expression] = STATE(1442),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8657),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3488)] = {
    [sym_expression] = STATE(1563),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3489)] = {
    [sym_expression] = STATE(1564),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3490)] = {
    [sym_expression] = STATE(4490),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3491)] = {
    [sym_expression] = STATE(2499),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3492)] = {
    [sym_expression] = STATE(1565),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3493)] = {
    [sym_expression] = STATE(4367),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3494)] = {
    [sym_expression] = STATE(1567),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3495)] = {
    [sym_expression] = STATE(1758),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3496)] = {
    [sym_expression] = STATE(1747),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3497)] = {
    [sym_expression] = STATE(4379),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3498)] = {
    [sym_expression] = STATE(1436),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3499)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3500)] = {
    [sym_expression] = STATE(1441),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3501)] = {
    [sym_expression] = STATE(1555),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3502)] = {
    [sym_expression] = STATE(4320),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3503)] = {
    [sym_expression] = STATE(1437),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3504)] = {
    [sym_expression] = STATE(1442),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8659),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3505)] = {
    [sym_expression] = STATE(4297),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3506)] = {
    [sym_expression] = STATE(4368),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3507)] = {
    [sym_expression] = STATE(4324),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3508)] = {
    [sym_expression] = STATE(4328),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3509)] = {
    [sym_expression] = STATE(1439),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3510)] = {
    [sym_expression] = STATE(4317),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3511)] = {
    [sym_expression] = STATE(4319),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3512)] = {
    [sym_expression] = STATE(4329),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3513)] = {
    [sym_expression] = STATE(4323),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3514)] = {
    [sym_expression] = STATE(4325),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3515)] = {
    [sym_expression] = STATE(4369),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3516)] = {
    [sym_expression] = STATE(4316),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3517)] = {
    [sym_expression] = STATE(4364),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3518)] = {
    [sym_expression] = STATE(4370),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3519)] = {
    [sym_expression] = STATE(4371),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3520)] = {
    [sym_expression] = STATE(4322),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3521)] = {
    [sym_expression] = STATE(4434),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3522)] = {
    [sym_expression] = STATE(4327),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3523)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3524)] = {
    [sym_expression] = STATE(4521),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3525)] = {
    [sym_expression] = STATE(4315),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3526)] = {
    [sym_expression] = STATE(1437),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3527)] = {
    [sym_expression] = STATE(4326),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3528)] = {
    [sym_expression] = STATE(4552),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3529)] = {
    [sym_expression] = STATE(1436),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3530)] = {
    [sym_expression] = STATE(4561),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3531)] = {
    [sym_expression] = STATE(4462),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3532)] = {
    [sym_expression] = STATE(4403),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3533)] = {
    [sym_expression] = STATE(1615),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3534)] = {
    [sym_expression] = STATE(1442),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8661),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3535)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3536)] = {
    [sym_expression] = STATE(4395),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3537)] = {
    [sym_expression] = STATE(4496),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3538)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3539)] = {
    [sym_expression] = STATE(1623),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3540)] = {
    [sym_expression] = STATE(1625),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8663),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3541)] = {
    [sym_expression] = STATE(4526),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3542)] = {
    [sym_expression] = STATE(4529),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3543)] = {
    [sym_expression] = STATE(1559),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3544)] = {
    [sym_expression] = STATE(1734),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3545)] = {
    [sym_expression] = STATE(4539),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3546)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3547)] = {
    [sym_expression] = STATE(1441),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8533),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3548)] = {
    [sym_expression] = STATE(1442),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8665),
    [anon_sym_BANG] = ACTIONS(6638),
    [anon_sym_TILDE] = ACTIONS(6638),
    [anon_sym_DASH] = ACTIONS(6636),
    [anon_sym_PLUS] = ACTIONS(6636),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(6640),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6642),
    [anon_sym___real] = ACTIONS(6642),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8535),
    [anon_sym_PLUS_PLUS] = ACTIONS(8535),
    [anon_sym_sizeof] = ACTIONS(6644),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6646),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3549)] = {
    [sym_expression] = STATE(4550),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3550)] = {
    [sym_expression] = STATE(4372),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3551)] = {
    [sym_expression] = STATE(4491),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3552)] = {
    [sym_expression] = STATE(1625),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8667),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3553)] = {
    [sym_expression] = STATE(1752),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3554)] = {
    [sym_expression] = STATE(1562),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3555)] = {
    [sym_expression] = STATE(4458),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3556)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3557)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3558)] = {
    [sym_expression] = STATE(4421),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3559)] = {
    [sym_expression] = STATE(4422),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3560)] = {
    [sym_expression] = STATE(4423),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3561)] = {
    [sym_expression] = STATE(1626),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3562)] = {
    [sym_expression] = STATE(4424),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3563)] = {
    [sym_expression] = STATE(4417),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3564)] = {
    [sym_expression] = STATE(4426),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3565)] = {
    [sym_expression] = STATE(4427),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3566)] = {
    [sym_expression] = STATE(4428),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3567)] = {
    [sym_expression] = STATE(4429),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3568)] = {
    [sym_expression] = STATE(4430),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3569)] = {
    [sym_expression] = STATE(4431),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3570)] = {
    [sym_expression] = STATE(4375),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3571)] = {
    [sym_expression] = STATE(1748),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3572)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3573)] = {
    [sym_expression] = STATE(4409),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3574)] = {
    [sym_expression] = STATE(4433),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3575)] = {
    [sym_expression] = STATE(4503),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3576)] = {
    [sym_expression] = STATE(4435),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3577)] = {
    [sym_expression] = STATE(4483),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3578)] = {
    [sym_expression] = STATE(4504),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3579)] = {
    [sym_expression] = STATE(4406),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3580)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1440),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(8401),
    [anon_sym_LPAREN2] = ACTIONS(8403),
    [anon_sym_BANG] = ACTIONS(2226),
    [anon_sym_TILDE] = ACTIONS(2226),
    [anon_sym_DASH] = ACTIONS(2224),
    [anon_sym_PLUS] = ACTIONS(2224),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8407),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8409),
    [anon_sym___real] = ACTIONS(8409),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8411),
    [anon_sym_PLUS_PLUS] = ACTIONS(8411),
    [anon_sym_sizeof] = ACTIONS(2230),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2244),
    [anon_sym_DQUOTE] = ACTIONS(2246),
    [anon_sym_L_DQUOTE] = ACTIONS(2246),
    [anon_sym_u_DQUOTE] = ACTIONS(2246),
    [anon_sym_U_DQUOTE] = ACTIONS(2246),
    [anon_sym_u8_DQUOTE] = ACTIONS(2246),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3581)] = {
    [sym_expression] = STATE(4383),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3582)] = {
    [sym_expression] = STATE(4384),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3583)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3584)] = {
    [sym_expression] = STATE(1442),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8669),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3585)] = {
    [sym_expression] = STATE(2505),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3586)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3587)] = {
    [sym_expression] = STATE(2513),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3588)] = {
    [sym_expression] = STATE(2484),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3589)] = {
    [sym_expression] = STATE(2485),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3590)] = {
    [sym_expression] = STATE(1626),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3591)] = {
    [sym_expression] = STATE(2486),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3592)] = {
    [sym_expression] = STATE(2488),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3593)] = {
    [sym_expression] = STATE(2489),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3594)] = {
    [sym_expression] = STATE(2490),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3595)] = {
    [sym_expression] = STATE(2491),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3596)] = {
    [sym_expression] = STATE(2493),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3597)] = {
    [sym_expression] = STATE(2494),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3598)] = {
    [sym_expression] = STATE(2495),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3599)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3600)] = {
    [sym_expression] = STATE(4386),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3601)] = {
    [sym_expression] = STATE(1626),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3602)] = {
    [sym_expression] = STATE(4387),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3603)] = {
    [sym_expression] = STATE(4388),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3604)] = {
    [sym_expression] = STATE(2496),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3605)] = {
    [sym_expression] = STATE(4389),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3606)] = {
    [sym_expression] = STATE(2497),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3607)] = {
    [sym_expression] = STATE(4547),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3608)] = {
    [sym_expression] = STATE(1912),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3609)] = {
    [sym_expression] = STATE(1437),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3610)] = {
    [sym_expression] = STATE(1900),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3611)] = {
    [sym_expression] = STATE(1901),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3612)] = {
    [sym_expression] = STATE(1439),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3613)] = {
    [sym_expression] = STATE(1902),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3614)] = {
    [sym_expression] = STATE(1903),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3615)] = {
    [sym_expression] = STATE(1904),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3616)] = {
    [sym_expression] = STATE(1905),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3617)] = {
    [sym_expression] = STATE(1906),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3618)] = {
    [sym_expression] = STATE(1907),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3619)] = {
    [sym_expression] = STATE(1908),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3620)] = {
    [sym_expression] = STATE(4390),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3621)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3622)] = {
    [sym_expression] = STATE(4391),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3623)] = {
    [sym_expression] = STATE(4302),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3624)] = {
    [sym_expression] = STATE(4541),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3625)] = {
    [sym_expression] = STATE(4382),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3626)] = {
    [sym_expression] = STATE(4392),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3627)] = {
    [sym_expression] = STATE(1910),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3628)] = {
    [sym_expression] = STATE(4373),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3629)] = {
    [sym_expression] = STATE(1911),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3630)] = {
    [sym_expression] = STATE(1615),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3631)] = {
    [sym_expression] = STATE(1623),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8415),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3632)] = {
    [sym_expression] = STATE(1625),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1918),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1918),
    [sym_call_expression] = STATE(1918),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1918),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1918),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8413),
    [anon_sym_LPAREN2] = ACTIONS(8671),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2884),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8419),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8421),
    [anon_sym___real] = ACTIONS(8421),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8423),
    [anon_sym_PLUS_PLUS] = ACTIONS(8423),
    [anon_sym_sizeof] = ACTIONS(2890),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3633)] = {
    [sym_expression] = STATE(4393),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3634)] = {
    [sym_expression] = STATE(1436),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3635)] = {
    [sym_expression] = STATE(1441),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3636)] = {
    [sym_expression] = STATE(4363),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3637)] = {
    [sym_identifier] = ACTIONS(5666),
    [anon_sym_LPAREN2] = ACTIONS(5668),
    [anon_sym_BANG] = ACTIONS(5668),
    [anon_sym_TILDE] = ACTIONS(5668),
    [anon_sym_DASH] = ACTIONS(5666),
    [anon_sym_PLUS] = ACTIONS(5666),
    [anon_sym_STAR] = ACTIONS(5668),
    [anon_sym_CARET] = ACTIONS(5668),
    [anon_sym_AMP] = ACTIONS(5668),
    [anon_sym___extension__] = ACTIONS(5666),
    [anon_sym_noreturn] = ACTIONS(5666),
    [anon_sym_LBRACK] = ACTIONS(5668),
    [anon_sym_RBRACK] = ACTIONS(5668),
    [anon_sym_static] = ACTIONS(5666),
    [anon_sym_const] = ACTIONS(5666),
    [anon_sym_constexpr] = ACTIONS(5666),
    [anon_sym_volatile] = ACTIONS(5666),
    [anon_sym_restrict] = ACTIONS(5666),
    [anon_sym___restrict__] = ACTIONS(5666),
    [anon_sym__Atomic] = ACTIONS(5666),
    [anon_sym__Noreturn] = ACTIONS(5666),
    [anon_sym__Nonnull] = ACTIONS(5666),
    [anon_sym_nullable] = ACTIONS(5666),
    [anon_sym__Complex] = ACTIONS(5666),
    [anon_sym__Nullable] = ACTIONS(5666),
    [anon_sym__Nullable_result] = ACTIONS(5666),
    [anon_sym__Null_unspecified] = ACTIONS(5666),
    [anon_sym___autoreleasing] = ACTIONS(5666),
    [anon_sym___block] = ACTIONS(5666),
    [anon_sym___bridge] = ACTIONS(5666),
    [anon_sym___bridge_retained] = ACTIONS(5666),
    [anon_sym___bridge_transfer] = ACTIONS(5666),
    [anon_sym___complex] = ACTIONS(5666),
    [anon_sym___const] = ACTIONS(5666),
    [anon_sym___imag] = ACTIONS(5666),
    [anon_sym___kindof] = ACTIONS(5666),
    [anon_sym___nonnull] = ACTIONS(5666),
    [anon_sym___nullable] = ACTIONS(5666),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5666),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5666),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5666),
    [anon_sym___real] = ACTIONS(5666),
    [anon_sym___strong] = ACTIONS(5666),
    [anon_sym___unsafe_unretained] = ACTIONS(5666),
    [anon_sym___unused] = ACTIONS(5666),
    [anon_sym___weak] = ACTIONS(5666),
    [anon_sym_alignas] = ACTIONS(5666),
    [anon_sym__Alignas] = ACTIONS(5666),
    [anon_sym_struct] = ACTIONS(5666),
    [anon_sym_in] = ACTIONS(5666),
    [anon_sym_DASH_DASH] = ACTIONS(5668),
    [anon_sym_PLUS_PLUS] = ACTIONS(5668),
    [anon_sym_sizeof] = ACTIONS(5666),
    [anon_sym___alignof__] = ACTIONS(5666),
    [anon_sym___alignof] = ACTIONS(5666),
    [anon_sym__alignof] = ACTIONS(5666),
    [anon_sym_alignof] = ACTIONS(5666),
    [anon_sym__Alignof] = ACTIONS(5666),
    [anon_sym_offsetof] = ACTIONS(5666),
    [anon_sym__Generic] = ACTIONS(5666),
    [anon_sym_asm] = ACTIONS(5666),
    [anon_sym___asm__] = ACTIONS(5666),
    [anon_sym___asm] = ACTIONS(5666),
    [sym_number_literal] = ACTIONS(5668),
    [anon_sym_L_SQUOTE] = ACTIONS(5668),
    [anon_sym_u_SQUOTE] = ACTIONS(5668),
    [anon_sym_U_SQUOTE] = ACTIONS(5668),
    [anon_sym_u8_SQUOTE] = ACTIONS(5668),
    [anon_sym_SQUOTE] = ACTIONS(5668),
    [anon_sym_AT] = ACTIONS(5666),
    [anon_sym_DQUOTE] = ACTIONS(5668),
    [anon_sym_L_DQUOTE] = ACTIONS(5668),
    [anon_sym_u_DQUOTE] = ACTIONS(5668),
    [anon_sym_U_DQUOTE] = ACTIONS(5668),
    [anon_sym_u8_DQUOTE] = ACTIONS(5668),
    [sym_true] = ACTIONS(5666),
    [sym_false] = ACTIONS(5666),
    [anon_sym_NULL] = ACTIONS(5666),
    [anon_sym_nullptr] = ACTIONS(5666),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(5668),
    [anon_sym_ATavailable] = ACTIONS(5668),
    [anon_sym___builtin_available] = ACTIONS(5666),
    [anon_sym_va_arg] = ACTIONS(5666),
    [anon_sym_ATencode] = ACTIONS(5668),
    [anon_sym_id] = ACTIONS(5666),
  },
  [STATE(3638)] = {
    [sym_expression] = STATE(4300),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3639)] = {
    [sym_expression] = STATE(2648),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3640)] = {
    [sym_expression] = STATE(1617),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3641)] = {
    [sym_expression] = STATE(2630),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3642)] = {
    [sym_expression] = STATE(2631),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3643)] = {
    [sym_expression] = STATE(2632),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3644)] = {
    [sym_expression] = STATE(1626),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3645)] = {
    [sym_expression] = STATE(2633),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3646)] = {
    [sym_expression] = STATE(2634),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3647)] = {
    [sym_expression] = STATE(2635),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3648)] = {
    [sym_expression] = STATE(2636),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3649)] = {
    [sym_expression] = STATE(2637),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3650)] = {
    [sym_expression] = STATE(2638),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3651)] = {
    [sym_expression] = STATE(2639),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3652)] = {
    [sym_expression] = STATE(2640),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3653)] = {
    [sym_expression] = STATE(4498),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3654)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3655)] = {
    [sym_expression] = STATE(4410),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3656)] = {
    [sym_expression] = STATE(1437),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8387),
    [anon_sym_BANG] = ACTIONS(6445),
    [anon_sym_TILDE] = ACTIONS(6445),
    [anon_sym_DASH] = ACTIONS(6443),
    [anon_sym_PLUS] = ACTIONS(6443),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(6447),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6449),
    [anon_sym___real] = ACTIONS(6449),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8393),
    [anon_sym_PLUS_PLUS] = ACTIONS(8393),
    [anon_sym_sizeof] = ACTIONS(6451),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6453),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3657)] = {
    [sym_expression] = STATE(2641),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3658)] = {
    [sym_expression] = STATE(4509),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3659)] = {
    [sym_expression] = STATE(2642),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3660)] = {
    [sym_expression] = STATE(4510),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3661)] = {
    [sym_expression] = STATE(4523),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3662)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3663)] = {
    [sym_expression] = STATE(4404),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3664)] = {
    [sym_expression] = STATE(4499),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3665)] = {
    [sym_expression] = STATE(4412),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3666)] = {
    [sym_expression] = STATE(4511),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3667)] = {
    [sym_expression] = STATE(4512),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3668)] = {
    [sym_expression] = STATE(4377),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3669)] = {
    [sym_expression] = STATE(4413),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3670)] = {
    [sym_expression] = STATE(4514),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3671)] = {
    [sym_expression] = STATE(4515),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3672)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_LPAREN2] = ACTIONS(2920),
    [anon_sym_BANG] = ACTIONS(2920),
    [anon_sym_TILDE] = ACTIONS(2920),
    [anon_sym_DASH] = ACTIONS(2914),
    [anon_sym_PLUS] = ACTIONS(2914),
    [anon_sym_STAR] = ACTIONS(2920),
    [anon_sym_CARET] = ACTIONS(2920),
    [anon_sym_AMP] = ACTIONS(2920),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym_RBRACK] = ACTIONS(2920),
    [anon_sym_static] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [anon_sym_DASH_DASH] = ACTIONS(2920),
    [anon_sym_PLUS_PLUS] = ACTIONS(2920),
    [anon_sym_sizeof] = ACTIONS(2914),
    [anon_sym___alignof__] = ACTIONS(2914),
    [anon_sym___alignof] = ACTIONS(2914),
    [anon_sym__alignof] = ACTIONS(2914),
    [anon_sym_alignof] = ACTIONS(2914),
    [anon_sym__Alignof] = ACTIONS(2914),
    [anon_sym_offsetof] = ACTIONS(2914),
    [anon_sym__Generic] = ACTIONS(2914),
    [anon_sym_asm] = ACTIONS(2914),
    [anon_sym___asm__] = ACTIONS(2914),
    [anon_sym___asm] = ACTIONS(2914),
    [sym_number_literal] = ACTIONS(2920),
    [anon_sym_L_SQUOTE] = ACTIONS(2920),
    [anon_sym_u_SQUOTE] = ACTIONS(2920),
    [anon_sym_U_SQUOTE] = ACTIONS(2920),
    [anon_sym_u8_SQUOTE] = ACTIONS(2920),
    [anon_sym_SQUOTE] = ACTIONS(2920),
    [anon_sym_AT] = ACTIONS(2914),
    [anon_sym_DQUOTE] = ACTIONS(2920),
    [anon_sym_L_DQUOTE] = ACTIONS(2920),
    [anon_sym_u_DQUOTE] = ACTIONS(2920),
    [anon_sym_U_DQUOTE] = ACTIONS(2920),
    [anon_sym_u8_DQUOTE] = ACTIONS(2920),
    [sym_true] = ACTIONS(2914),
    [sym_false] = ACTIONS(2914),
    [anon_sym_NULL] = ACTIONS(2914),
    [anon_sym_nullptr] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2920),
    [anon_sym_ATavailable] = ACTIONS(2920),
    [anon_sym___builtin_available] = ACTIONS(2914),
    [anon_sym_va_arg] = ACTIONS(2914),
    [anon_sym_ATencode] = ACTIONS(2920),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(3673)] = {
    [sym_expression] = STATE(4414),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3674)] = {
    [sym_expression] = STATE(4516),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3675)] = {
    [sym_expression] = STATE(4517),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3676)] = {
    [sym_expression] = STATE(4400),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3677)] = {
    [sym_expression] = STATE(4416),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3678)] = {
    [sym_expression] = STATE(1615),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3679)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3680)] = {
    [sym_expression] = STATE(4549),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3681)] = {
    [sym_expression] = STATE(1623),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3682)] = {
    [sym_expression] = STATE(1625),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(8673),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3683)] = {
    [sym_expression] = STATE(4562),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3684)] = {
    [sym_expression] = STATE(4520),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3685)] = {
    [sym_expression] = STATE(4557),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3686)] = {
    [sym_expression] = STATE(4374),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3687)] = {
    [sym_expression] = STATE(4343),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3688)] = {
    [sym_expression] = STATE(1437),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3689)] = {
    [sym_expression] = STATE(4345),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3690)] = {
    [sym_expression] = STATE(4347),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3691)] = {
    [sym_expression] = STATE(1439),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3692)] = {
    [sym_expression] = STATE(4349),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3693)] = {
    [sym_expression] = STATE(4350),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3694)] = {
    [sym_expression] = STATE(4351),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3695)] = {
    [sym_expression] = STATE(4352),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3696)] = {
    [sym_expression] = STATE(4353),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3697)] = {
    [sym_expression] = STATE(4354),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3698)] = {
    [sym_expression] = STATE(4355),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3699)] = {
    [sym_expression] = STATE(4356),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3700)] = {
    [sym_expression] = STATE(4394),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3701)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3702)] = {
    [sym_expression] = STATE(4381),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4341),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4341),
    [sym_call_expression] = STATE(4341),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4341),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4341),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2072),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8553),
    [anon_sym_LPAREN2] = ACTIONS(8451),
    [anon_sym_BANG] = ACTIONS(7251),
    [anon_sym_TILDE] = ACTIONS(7251),
    [anon_sym_DASH] = ACTIONS(7249),
    [anon_sym_PLUS] = ACTIONS(7249),
    [anon_sym_STAR] = ACTIONS(8453),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8453),
    [anon_sym___extension__] = ACTIONS(7253),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(7255),
    [anon_sym___real] = ACTIONS(7255),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8457),
    [anon_sym_PLUS_PLUS] = ACTIONS(8457),
    [anon_sym_sizeof] = ACTIONS(7257),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(7259),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3703)] = {
    [sym_expression] = STATE(4357),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3704)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3705)] = {
    [sym_expression] = STATE(4358),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3706)] = {
    [sym_expression] = STATE(1436),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3707)] = {
    [sym_expression] = STATE(1441),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8459),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3708)] = {
    [sym_expression] = STATE(1442),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8675),
    [anon_sym_BANG] = ACTIONS(6702),
    [anon_sym_TILDE] = ACTIONS(6702),
    [anon_sym_DASH] = ACTIONS(6700),
    [anon_sym_PLUS] = ACTIONS(6700),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(6704),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(6706),
    [anon_sym___real] = ACTIONS(6706),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8461),
    [anon_sym_PLUS_PLUS] = ACTIONS(8461),
    [anon_sym_sizeof] = ACTIONS(6708),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(6710),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(6712),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3709)] = {
    [sym_expression] = STATE(4559),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3710)] = {
    [sym_expression] = STATE(4487),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3711)] = {
    [sym_expression] = STATE(1615),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3712)] = {
    [sym_expression] = STATE(1443),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3713)] = {
    [sym_expression] = STATE(1619),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3714)] = {
    [sym_expression] = STATE(1623),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3715)] = {
    [sym_expression] = STATE(4436),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3716)] = {
    [sym_expression] = STATE(2098),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3717)] = {
    [sym_expression] = STATE(1437),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3718)] = {
    [sym_expression] = STATE(2087),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3719)] = {
    [sym_expression] = STATE(2088),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3720)] = {
    [sym_expression] = STATE(1439),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3721)] = {
    [sym_expression] = STATE(2089),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3722)] = {
    [sym_expression] = STATE(2090),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3723)] = {
    [sym_expression] = STATE(2091),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3724)] = {
    [sym_expression] = STATE(2092),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3725)] = {
    [sym_expression] = STATE(2093),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3726)] = {
    [sym_expression] = STATE(2094),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3727)] = {
    [sym_expression] = STATE(2095),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3728)] = {
    [sym_expression] = STATE(2096),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3729)] = {
    [sym_expression] = STATE(4531),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3730)] = {
    [sym_expression] = STATE(1444),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3731)] = {
    [sym_expression] = STATE(2598),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3732)] = {
    [sym_expression] = STATE(1759),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3733)] = {
    [sym_expression] = STATE(2097),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3734)] = {
    [sym_expression] = STATE(2058),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3735)] = {
    [sym_expression] = STATE(1436),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3736)] = {
    [sym_expression] = STATE(1441),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1710),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8519),
    [anon_sym_BANG] = ACTIONS(2332),
    [anon_sym_TILDE] = ACTIONS(2332),
    [anon_sym_DASH] = ACTIONS(2330),
    [anon_sym_PLUS] = ACTIONS(2330),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8521),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8523),
    [anon_sym___real] = ACTIONS(8523),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8525),
    [anon_sym_PLUS_PLUS] = ACTIONS(8525),
    [anon_sym_sizeof] = ACTIONS(2334),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2336),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3737)] = {
    [sym_expression] = STATE(4505),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4359),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4359),
    [sym_call_expression] = STATE(4359),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4359),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4359),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(5336),
    [anon_sym_LPAREN2] = ACTIONS(4338),
    [anon_sym_BANG] = ACTIONS(4340),
    [anon_sym_TILDE] = ACTIONS(4340),
    [anon_sym_DASH] = ACTIONS(4342),
    [anon_sym_PLUS] = ACTIONS(4342),
    [anon_sym_STAR] = ACTIONS(4344),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(4344),
    [anon_sym___extension__] = ACTIONS(5338),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(5340),
    [anon_sym___real] = ACTIONS(5340),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(4356),
    [anon_sym_PLUS_PLUS] = ACTIONS(4356),
    [anon_sym_sizeof] = ACTIONS(4358),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(4360),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(4362),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3738)] = {
    [sym_expression] = STATE(1741),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3739)] = {
    [sym_expression] = STATE(4524),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3740)] = {
    [sym_expression] = STATE(2616),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(2181),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(2181),
    [sym_call_expression] = STATE(2181),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(2181),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(2181),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2898),
    [anon_sym_LPAREN2] = ACTIONS(8491),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2900),
    [anon_sym_PLUS] = ACTIONS(2900),
    [anon_sym_STAR] = ACTIONS(8493),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8493),
    [anon_sym___extension__] = ACTIONS(8495),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8497),
    [anon_sym___real] = ACTIONS(8497),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8499),
    [anon_sym_PLUS_PLUS] = ACTIONS(8499),
    [anon_sym_sizeof] = ACTIONS(2904),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3741)] = {
    [sym_expression] = STATE(1743),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3742)] = {
    [sym_expression] = STATE(1623),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3743)] = {
    [sym_expression] = STATE(1766),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3744)] = {
    [sym_expression] = STATE(1626),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(1572),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(1572),
    [sym_call_expression] = STATE(1572),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(1572),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(1572),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(1624),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(8435),
    [anon_sym_LPAREN2] = ACTIONS(8437),
    [anon_sym_BANG] = ACTIONS(2284),
    [anon_sym_TILDE] = ACTIONS(2284),
    [anon_sym_DASH] = ACTIONS(2282),
    [anon_sym_PLUS] = ACTIONS(2282),
    [anon_sym_STAR] = ACTIONS(8405),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(8405),
    [anon_sym___extension__] = ACTIONS(8439),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(8441),
    [anon_sym___real] = ACTIONS(8441),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(8443),
    [anon_sym_PLUS_PLUS] = ACTIONS(8443),
    [anon_sym_sizeof] = ACTIONS(2286),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(2288),
    [anon_sym___asm__] = ACTIONS(2288),
    [anon_sym___asm] = ACTIONS(2288),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(2290),
    [anon_sym_DQUOTE] = ACTIONS(2292),
    [anon_sym_L_DQUOTE] = ACTIONS(2292),
    [anon_sym_u_DQUOTE] = ACTIONS(2292),
    [anon_sym_U_DQUOTE] = ACTIONS(2292),
    [anon_sym_u8_DQUOTE] = ACTIONS(2292),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3745)] = {
    [sym_expression] = STATE(4486),
    [sym__string] = STATE(1641),
    [sym_conditional_expression] = STATE(1641),
    [sym_assignment_expression] = STATE(1641),
    [sym_pointer_expression] = STATE(4312),
    [sym_unary_expression] = STATE(1641),
    [sym_binary_expression] = STATE(1641),
    [sym_update_expression] = STATE(1641),
    [sym_cast_expression] = STATE(1641),
    [sym_sizeof_expression] = STATE(1641),
    [sym_alignof_expression] = STATE(1641),
    [sym_offsetof_expression] = STATE(1641),
    [sym_generic_expression] = STATE(1641),
    [sym_subscript_expression] = STATE(4312),
    [sym_call_expression] = STATE(4312),
    [sym_gnu_asm_expression] = STATE(1641),
    [sym_extension_expression] = STATE(1641),
    [sym_field_expression] = STATE(4312),
    [sym_compound_literal_expression] = STATE(1641),
    [sym_parenthesized_expression] = STATE(4312),
    [sym_char_literal] = STATE(1641),
    [sym_concatenated_string] = STATE(1641),
    [sym_string_literal] = STATE(2069),
    [sym_null] = STATE(1641),
    [sym_selector_expression] = STATE(1641),
    [sym_available_expression] = STATE(1641),
    [sym_range_expression] = STATE(1641),
    [sym_block_literal] = STATE(1641),
    [sym_message_expression] = STATE(1641),
    [sym_va_arg_expression] = STATE(1641),
    [sym_encode_expression] = STATE(1641),
    [sym_at_expression] = STATE(1641),
    [sym_dictionary_literal] = STATE(1641),
    [sym_array_literal] = STATE(1641),
    [sym_identifier] = ACTIONS(2908),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_CARET] = ACTIONS(27),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2916),
    [anon_sym_const] = ACTIONS(203),
    [anon_sym___imag] = ACTIONS(2918),
    [anon_sym___real] = ACTIONS(2918),
    [anon_sym_struct] = ACTIONS(203),
    [anon_sym_in] = ACTIONS(203),
    [anon_sym_DASH_DASH] = ACTIONS(93),
    [anon_sym_PLUS_PLUS] = ACTIONS(93),
    [anon_sym_sizeof] = ACTIONS(95),
    [anon_sym___alignof__] = ACTIONS(97),
    [anon_sym___alignof] = ACTIONS(97),
    [anon_sym__alignof] = ACTIONS(97),
    [anon_sym_alignof] = ACTIONS(97),
    [anon_sym__Alignof] = ACTIONS(97),
    [anon_sym_offsetof] = ACTIONS(99),
    [anon_sym__Generic] = ACTIONS(101),
    [anon_sym_asm] = ACTIONS(103),
    [anon_sym___asm__] = ACTIONS(103),
    [anon_sym___asm] = ACTIONS(103),
    [sym_number_literal] = ACTIONS(215),
    [anon_sym_L_SQUOTE] = ACTIONS(107),
    [anon_sym_u_SQUOTE] = ACTIONS(107),
    [anon_sym_U_SQUOTE] = ACTIONS(107),
    [anon_sym_u8_SQUOTE] = ACTIONS(107),
    [anon_sym_SQUOTE] = ACTIONS(107),
    [anon_sym_AT] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(111),
    [anon_sym_L_DQUOTE] = ACTIONS(111),
    [anon_sym_u_DQUOTE] = ACTIONS(111),
    [anon_sym_U_DQUOTE] = ACTIONS(111),
    [anon_sym_u8_DQUOTE] = ACTIONS(111),
    [sym_true] = ACTIONS(219),
    [sym_false] = ACTIONS(219),
    [anon_sym_NULL] = ACTIONS(115),
    [anon_sym_nullptr] = ACTIONS(115),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(137),
    [anon_sym_ATavailable] = ACTIONS(139),
    [anon_sym___builtin_available] = ACTIONS(141),
    [anon_sym_va_arg] = ACTIONS(143),
    [anon_sym_ATencode] = ACTIONS(145),
    [anon_sym_id] = ACTIONS(203),
  },
  [STATE(3746)] = {
    [sym_expression] = STATE(1909),
    [sym__string] = STATE(1465),
    [sym_conditional_expression] = STATE(1465),
    [sym_assignment_expression] = STATE(1465),
    [sym_pointer_expression] = STATE(1454),
    [sym_unary_expression] = STATE(1465),
    [sym_binary_expression] = STATE(1465),
    [sym_update_expression] = STATE(1465),
    [sym_cast_expression] = STATE(1465),
    [sym_sizeof_expression] = STATE(1465),
    [sym_alignof_expression] = STATE(1465),
    [sym_offsetof_expression] = STATE(1465),
    [sym_generic_expression] = STATE(1465),
    [sym_subscript_expression] = STATE(1454),
    [sym_call_expression] = STATE(1454),
    [sym_gnu_asm_expression] = STATE(1465),
    [sym_extension_expression] = STATE(1465),
    [sym_field_expression] = STATE(1454),
    [sym_compound_literal_expression] = STATE(1465),
    [sym_parenthesized_expression] = STATE(1454),
    [sym_char_literal] = STATE(1465),
    [sym_concatenated_string] = STATE(1465),
    [sym_string_literal] = STATE(1700),
    [sym_null] = STATE(1465),
    [sym_selector_expression] = STATE(1465),
    [sym_available_expression] = STATE(1465),
    [sym_range_expression] = STATE(1465),
    [sym_block_literal] = STATE(1465),
    [sym_message_expression] = STATE(1465),
    [sym_va_arg_expression] = STATE(1465),
    [sym_encode_expression] = STATE(1465),
    [sym_at_expression] = STATE(1465),
    [sym_dictionary_literal] = STATE(1465),
    [sym_array_literal] = STATE(1465),
    [sym_identifier] = ACTIONS(2328),
    [anon_sym_LPAREN2] = ACTIONS(8481),
    [anon_sym_BANG] = ACTIONS(2296),
    [anon_sym_TILDE] = ACTIONS(2296),
    [anon_sym_DASH] = ACTIONS(2294),
    [anon_sym_PLUS] = ACTIONS(2294),
    [anon_sym_STAR] = ACTIONS(8417),
    [anon_sym_CARET] = ACTIONS(8389),
    [anon_sym_AMP] = ACTIONS(8417),
    [anon_sym___extension__] = ACTIONS(8483),
    [anon_sym_LBRACK] = ACTIONS(8391),
    [anon_sym_const] = ACTIONS(2228),
    [anon_sym___imag] = ACTIONS(8485),
    [anon_sym___real] = ACTIONS(8485),
    [anon_sym_struct] = ACTIONS(2228),
    [anon_sym_in] = ACTIONS(2228),
    [anon_sym_DASH_DASH] = ACTIONS(8487),
    [anon_sym_PLUS_PLUS] = ACTIONS(8487),
    [anon_sym_sizeof] = ACTIONS(2300),
    [anon_sym___alignof__] = ACTIONS(2232),
    [anon_sym___alignof] = ACTIONS(2232),
    [anon_sym__alignof] = ACTIONS(2232),
    [anon_sym_alignof] = ACTIONS(2232),
    [anon_sym__Alignof] = ACTIONS(2232),
    [anon_sym_offsetof] = ACTIONS(2234),
    [anon_sym__Generic] = ACTIONS(2236),
    [anon_sym_asm] = ACTIONS(2238),
    [anon_sym___asm__] = ACTIONS(2238),
    [anon_sym___asm] = ACTIONS(2238),
    [sym_number_literal] = ACTIONS(2240),
    [anon_sym_L_SQUOTE] = ACTIONS(2242),
    [anon_sym_u_SQUOTE] = ACTIONS(2242),
    [anon_sym_U_SQUOTE] = ACTIONS(2242),
    [anon_sym_u8_SQUOTE] = ACTIONS(2242),
    [anon_sym_SQUOTE] = ACTIONS(2242),
    [anon_sym_AT] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [anon_sym_L_DQUOTE] = ACTIONS(2304),
    [anon_sym_u_DQUOTE] = ACTIONS(2304),
    [anon_sym_U_DQUOTE] = ACTIONS(2304),
    [anon_sym_u8_DQUOTE] = ACTIONS(2304),
    [sym_true] = ACTIONS(2248),
    [sym_false] = ACTIONS(2248),
    [anon_sym_NULL] = ACTIONS(2250),
    [anon_sym_nullptr] = ACTIONS(2250),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATselector] = ACTIONS(2252),
    [anon_sym_ATavailable] = ACTIONS(2254),
    [anon_sym___builtin_available] = ACTIONS(2256),
    [anon_sym_va_arg] = ACTIONS(2258),
    [anon_sym_ATencode] = ACTIONS(2260),
    [anon_sym_id] = ACTIONS(2228),
  },
  [STATE(3747)] = {
    [sym_attribute_specifier] = STATE(3767),
    [sym_attribute_declaration] = STATE(3767),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_property_attributes_declaration] = STATE(3758),
    [sym_struct_declaration] = STATE(2409),
    [sym_atomic_declaration] = STATE(2409),
    [sym_specifier_qualifier] = STATE(3326),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3326),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym_LPAREN2] = ACTIONS(8677),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8679),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3748)] = {
    [sym_attribute_specifier] = STATE(3765),
    [sym_attribute_declaration] = STATE(3765),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_property_attributes_declaration] = STATE(3755),
    [sym_struct_declaration] = STATE(2288),
    [sym_atomic_declaration] = STATE(2288),
    [sym_specifier_qualifier] = STATE(3327),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3327),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym_LPAREN2] = ACTIONS(8677),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8681),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3749)] = {
    [sym_attribute_specifier] = STATE(3773),
    [sym_attribute_declaration] = STATE(3773),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_property_attributes_declaration] = STATE(3753),
    [sym_struct_declaration] = STATE(2245),
    [sym_atomic_declaration] = STATE(2245),
    [sym_specifier_qualifier] = STATE(3324),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3324),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym_LPAREN2] = ACTIONS(8677),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8683),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3750)] = {
    [sym_attribute_specifier] = STATE(3766),
    [sym_attribute_declaration] = STATE(3766),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_property_attributes_declaration] = STATE(3756),
    [sym_struct_declaration] = STATE(2267),
    [sym_atomic_declaration] = STATE(2267),
    [sym_specifier_qualifier] = STATE(3327),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3327),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym_LPAREN2] = ACTIONS(8677),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8681),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3751)] = {
    [sym_attribute_specifier] = STATE(3768),
    [sym_attribute_declaration] = STATE(3768),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_property_attributes_declaration] = STATE(3757),
    [sym_struct_declaration] = STATE(2403),
    [sym_atomic_declaration] = STATE(2403),
    [sym_specifier_qualifier] = STATE(3326),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3326),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym_LPAREN2] = ACTIONS(8677),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8679),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3752)] = {
    [sym_attribute_specifier] = STATE(3776),
    [sym_attribute_declaration] = STATE(3776),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_property_attributes_declaration] = STATE(3754),
    [sym_struct_declaration] = STATE(2249),
    [sym_atomic_declaration] = STATE(2249),
    [sym_specifier_qualifier] = STATE(3324),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3324),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym_LPAREN2] = ACTIONS(8677),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8683),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3753)] = {
    [sym_attribute_specifier] = STATE(3778),
    [sym_attribute_declaration] = STATE(3778),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2335),
    [sym_atomic_declaration] = STATE(2335),
    [sym_specifier_qualifier] = STATE(3324),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3324),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8683),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3754)] = {
    [sym_attribute_specifier] = STATE(3773),
    [sym_attribute_declaration] = STATE(3773),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2245),
    [sym_atomic_declaration] = STATE(2245),
    [sym_specifier_qualifier] = STATE(3324),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3324),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8683),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3755)] = {
    [sym_attribute_specifier] = STATE(3766),
    [sym_attribute_declaration] = STATE(3766),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2267),
    [sym_atomic_declaration] = STATE(2267),
    [sym_specifier_qualifier] = STATE(3327),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3327),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8681),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3756)] = {
    [sym_attribute_specifier] = STATE(3770),
    [sym_attribute_declaration] = STATE(3770),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2276),
    [sym_atomic_declaration] = STATE(2276),
    [sym_specifier_qualifier] = STATE(3327),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3327),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8681),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3757)] = {
    [sym_attribute_specifier] = STATE(3769),
    [sym_attribute_declaration] = STATE(3769),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2420),
    [sym_atomic_declaration] = STATE(2420),
    [sym_specifier_qualifier] = STATE(3326),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3326),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8679),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3758)] = {
    [sym_attribute_specifier] = STATE(3768),
    [sym_attribute_declaration] = STATE(3768),
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2403),
    [sym_atomic_declaration] = STATE(2403),
    [sym_specifier_qualifier] = STATE(3326),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3326),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_LBRACK] = ACTIONS(8196),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8679),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3759)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_COMMA] = ACTIONS(4044),
    [anon_sym_RPAREN] = ACTIONS(4044),
    [anon_sym_LPAREN2] = ACTIONS(4044),
    [anon_sym_DASH] = ACTIONS(4042),
    [anon_sym_PLUS] = ACTIONS(4042),
    [anon_sym_STAR] = ACTIONS(4044),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4044),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(4044),
    [anon_sym_AMP] = ACTIONS(4042),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(4042),
    [anon_sym_LT_LT] = ACTIONS(4044),
    [anon_sym_GT_GT] = ACTIONS(4044),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(4044),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(4044),
    [anon_sym_PLUS_PLUS] = ACTIONS(4044),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(3760)] = {
    [sym_identifier] = ACTIONS(2828),
    [anon_sym_COMMA] = ACTIONS(2830),
    [anon_sym_RPAREN] = ACTIONS(2830),
    [aux_sym_preproc_if_token1] = ACTIONS(2828),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2830),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2830),
    [anon_sym_LPAREN2] = ACTIONS(2830),
    [anon_sym_STAR] = ACTIONS(2830),
    [anon_sym_CARET] = ACTIONS(2830),
    [anon_sym_GT] = ACTIONS(2830),
    [anon_sym_LT] = ACTIONS(2830),
    [anon_sym___extension__] = ACTIONS(2828),
    [anon_sym___attribute__] = ACTIONS(2828),
    [anon_sym___attribute] = ACTIONS(2828),
    [anon_sym_noreturn] = ACTIONS(2828),
    [anon_sym_LBRACK] = ACTIONS(2830),
    [anon_sym___based] = ACTIONS(2828),
    [anon_sym_LBRACE] = ACTIONS(2830),
    [anon_sym_signed] = ACTIONS(2828),
    [anon_sym_unsigned] = ACTIONS(2828),
    [anon_sym_long] = ACTIONS(2828),
    [anon_sym_short] = ACTIONS(2828),
    [anon_sym_ATautoreleasepool] = ACTIONS(2830),
    [anon_sym_const] = ACTIONS(2828),
    [anon_sym_constexpr] = ACTIONS(2828),
    [anon_sym_volatile] = ACTIONS(2828),
    [anon_sym_restrict] = ACTIONS(2828),
    [anon_sym___restrict__] = ACTIONS(2828),
    [anon_sym__Atomic] = ACTIONS(2828),
    [anon_sym__Noreturn] = ACTIONS(2828),
    [anon_sym__Nonnull] = ACTIONS(2828),
    [anon_sym_nullable] = ACTIONS(2828),
    [anon_sym__Complex] = ACTIONS(2828),
    [anon_sym__Nullable] = ACTIONS(2828),
    [anon_sym__Nullable_result] = ACTIONS(2828),
    [anon_sym__Null_unspecified] = ACTIONS(2828),
    [anon_sym___autoreleasing] = ACTIONS(2828),
    [anon_sym___block] = ACTIONS(2828),
    [anon_sym___bridge] = ACTIONS(2828),
    [anon_sym___bridge_retained] = ACTIONS(2828),
    [anon_sym___bridge_transfer] = ACTIONS(2828),
    [anon_sym___complex] = ACTIONS(2828),
    [anon_sym___const] = ACTIONS(2828),
    [anon_sym___imag] = ACTIONS(2828),
    [anon_sym___kindof] = ACTIONS(2828),
    [anon_sym___nonnull] = ACTIONS(2828),
    [anon_sym___nullable] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2828),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2828),
    [anon_sym___real] = ACTIONS(2828),
    [anon_sym___strong] = ACTIONS(2828),
    [anon_sym___unsafe_unretained] = ACTIONS(2828),
    [anon_sym___unused] = ACTIONS(2828),
    [anon_sym___weak] = ACTIONS(2828),
    [anon_sym_alignas] = ACTIONS(2828),
    [anon_sym__Alignas] = ACTIONS(2828),
    [sym_primitive_type] = ACTIONS(2828),
    [anon_sym_enum] = ACTIONS(2828),
    [anon_sym_COLON] = ACTIONS(2830),
    [anon_sym_struct] = ACTIONS(2828),
    [anon_sym_union] = ACTIONS(2828),
    [anon_sym_in] = ACTIONS(2828),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2828),
    [anon_sym___typeof] = ACTIONS(2828),
    [anon_sym_typeof] = ACTIONS(2828),
    [anon_sym_ATprivate] = ACTIONS(2830),
    [anon_sym_ATprotected] = ACTIONS(2830),
    [anon_sym_ATpackage] = ACTIONS(2830),
    [anon_sym_ATpublic] = ACTIONS(2830),
    [anon_sym_BOOL] = ACTIONS(2828),
    [anon_sym_IMP] = ACTIONS(2828),
    [anon_sym_SEL] = ACTIONS(2828),
    [anon_sym_Class] = ACTIONS(2828),
    [anon_sym_id] = ACTIONS(2828),
    [anon_sym_out] = ACTIONS(2828),
    [anon_sym_inout] = ACTIONS(2828),
    [anon_sym_bycopy] = ACTIONS(2828),
    [anon_sym_byref] = ACTIONS(2828),
    [anon_sym_oneway] = ACTIONS(2828),
  },
  [STATE(3761)] = {
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_specifier_qualifier] = STATE(3761),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3761),
    [sym_identifier] = ACTIONS(8685),
    [anon_sym_LPAREN2] = ACTIONS(8688),
    [anon_sym_STAR] = ACTIONS(8688),
    [anon_sym_CARET] = ACTIONS(8688),
    [anon_sym___extension__] = ACTIONS(8690),
    [anon_sym_noreturn] = ACTIONS(8690),
    [anon_sym___based] = ACTIONS(8693),
    [anon_sym_signed] = ACTIONS(8695),
    [anon_sym_unsigned] = ACTIONS(8695),
    [anon_sym_long] = ACTIONS(8695),
    [anon_sym_short] = ACTIONS(8695),
    [anon_sym_const] = ACTIONS(8690),
    [anon_sym_constexpr] = ACTIONS(8690),
    [anon_sym_volatile] = ACTIONS(8690),
    [anon_sym_restrict] = ACTIONS(8690),
    [anon_sym___restrict__] = ACTIONS(8690),
    [anon_sym__Atomic] = ACTIONS(8690),
    [anon_sym__Noreturn] = ACTIONS(8690),
    [anon_sym__Nonnull] = ACTIONS(8690),
    [anon_sym_nullable] = ACTIONS(8690),
    [anon_sym__Complex] = ACTIONS(8690),
    [anon_sym__Nullable] = ACTIONS(8690),
    [anon_sym__Nullable_result] = ACTIONS(8690),
    [anon_sym__Null_unspecified] = ACTIONS(8690),
    [anon_sym___autoreleasing] = ACTIONS(8690),
    [anon_sym___block] = ACTIONS(8690),
    [anon_sym___bridge] = ACTIONS(8690),
    [anon_sym___bridge_retained] = ACTIONS(8690),
    [anon_sym___bridge_transfer] = ACTIONS(8690),
    [anon_sym___complex] = ACTIONS(8690),
    [anon_sym___const] = ACTIONS(8690),
    [anon_sym___imag] = ACTIONS(8690),
    [anon_sym___kindof] = ACTIONS(8690),
    [anon_sym___nonnull] = ACTIONS(8690),
    [anon_sym___nullable] = ACTIONS(8690),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8690),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8690),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8690),
    [anon_sym___real] = ACTIONS(8690),
    [anon_sym___strong] = ACTIONS(8690),
    [anon_sym___unsafe_unretained] = ACTIONS(8690),
    [anon_sym___unused] = ACTIONS(8690),
    [anon_sym___weak] = ACTIONS(8690),
    [anon_sym_alignas] = ACTIONS(8698),
    [anon_sym__Alignas] = ACTIONS(8698),
    [sym_primitive_type] = ACTIONS(8701),
    [anon_sym_enum] = ACTIONS(8704),
    [anon_sym_COLON] = ACTIONS(8688),
    [anon_sym_struct] = ACTIONS(8707),
    [anon_sym_union] = ACTIONS(8710),
    [anon_sym_in] = ACTIONS(8713),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8716),
    [anon_sym___typeof] = ACTIONS(8716),
    [anon_sym_typeof] = ACTIONS(8716),
    [anon_sym_BOOL] = ACTIONS(8719),
    [anon_sym_IMP] = ACTIONS(8719),
    [anon_sym_SEL] = ACTIONS(8719),
    [anon_sym_Class] = ACTIONS(8719),
    [anon_sym_id] = ACTIONS(8719),
    [anon_sym_out] = ACTIONS(8713),
    [anon_sym_inout] = ACTIONS(8713),
    [anon_sym_bycopy] = ACTIONS(8713),
    [anon_sym_byref] = ACTIONS(8713),
    [anon_sym_oneway] = ACTIONS(8713),
  },
  [STATE(3762)] = {
    [sym_identifier] = ACTIONS(2712),
    [anon_sym_COMMA] = ACTIONS(2714),
    [anon_sym_RPAREN] = ACTIONS(2714),
    [aux_sym_preproc_if_token1] = ACTIONS(2712),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2714),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2714),
    [anon_sym_LPAREN2] = ACTIONS(2714),
    [anon_sym_STAR] = ACTIONS(2714),
    [anon_sym_CARET] = ACTIONS(2714),
    [anon_sym_GT] = ACTIONS(2714),
    [anon_sym_LT] = ACTIONS(2714),
    [anon_sym___extension__] = ACTIONS(2712),
    [anon_sym___attribute__] = ACTIONS(2712),
    [anon_sym___attribute] = ACTIONS(2712),
    [anon_sym_noreturn] = ACTIONS(2712),
    [anon_sym_LBRACK] = ACTIONS(2714),
    [anon_sym___based] = ACTIONS(2712),
    [anon_sym_LBRACE] = ACTIONS(2714),
    [anon_sym_signed] = ACTIONS(2712),
    [anon_sym_unsigned] = ACTIONS(2712),
    [anon_sym_long] = ACTIONS(2712),
    [anon_sym_short] = ACTIONS(2712),
    [anon_sym_ATautoreleasepool] = ACTIONS(2714),
    [anon_sym_const] = ACTIONS(2712),
    [anon_sym_constexpr] = ACTIONS(2712),
    [anon_sym_volatile] = ACTIONS(2712),
    [anon_sym_restrict] = ACTIONS(2712),
    [anon_sym___restrict__] = ACTIONS(2712),
    [anon_sym__Atomic] = ACTIONS(2712),
    [anon_sym__Noreturn] = ACTIONS(2712),
    [anon_sym__Nonnull] = ACTIONS(2712),
    [anon_sym_nullable] = ACTIONS(2712),
    [anon_sym__Complex] = ACTIONS(2712),
    [anon_sym__Nullable] = ACTIONS(2712),
    [anon_sym__Nullable_result] = ACTIONS(2712),
    [anon_sym__Null_unspecified] = ACTIONS(2712),
    [anon_sym___autoreleasing] = ACTIONS(2712),
    [anon_sym___block] = ACTIONS(2712),
    [anon_sym___bridge] = ACTIONS(2712),
    [anon_sym___bridge_retained] = ACTIONS(2712),
    [anon_sym___bridge_transfer] = ACTIONS(2712),
    [anon_sym___complex] = ACTIONS(2712),
    [anon_sym___const] = ACTIONS(2712),
    [anon_sym___imag] = ACTIONS(2712),
    [anon_sym___kindof] = ACTIONS(2712),
    [anon_sym___nonnull] = ACTIONS(2712),
    [anon_sym___nullable] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2712),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2712),
    [anon_sym___real] = ACTIONS(2712),
    [anon_sym___strong] = ACTIONS(2712),
    [anon_sym___unsafe_unretained] = ACTIONS(2712),
    [anon_sym___unused] = ACTIONS(2712),
    [anon_sym___weak] = ACTIONS(2712),
    [anon_sym_alignas] = ACTIONS(2712),
    [anon_sym__Alignas] = ACTIONS(2712),
    [sym_primitive_type] = ACTIONS(2712),
    [anon_sym_enum] = ACTIONS(2712),
    [anon_sym_COLON] = ACTIONS(2714),
    [anon_sym_struct] = ACTIONS(2712),
    [anon_sym_union] = ACTIONS(2712),
    [anon_sym_in] = ACTIONS(2712),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2712),
    [anon_sym___typeof] = ACTIONS(2712),
    [anon_sym_typeof] = ACTIONS(2712),
    [anon_sym_ATprivate] = ACTIONS(2714),
    [anon_sym_ATprotected] = ACTIONS(2714),
    [anon_sym_ATpackage] = ACTIONS(2714),
    [anon_sym_ATpublic] = ACTIONS(2714),
    [anon_sym_BOOL] = ACTIONS(2712),
    [anon_sym_IMP] = ACTIONS(2712),
    [anon_sym_SEL] = ACTIONS(2712),
    [anon_sym_Class] = ACTIONS(2712),
    [anon_sym_id] = ACTIONS(2712),
    [anon_sym_out] = ACTIONS(2712),
    [anon_sym_inout] = ACTIONS(2712),
    [anon_sym_bycopy] = ACTIONS(2712),
    [anon_sym_byref] = ACTIONS(2712),
    [anon_sym_oneway] = ACTIONS(2712),
  },
  [STATE(3763)] = {
    [sym_identifier] = ACTIONS(2592),
    [anon_sym_COMMA] = ACTIONS(2594),
    [anon_sym_RPAREN] = ACTIONS(2594),
    [aux_sym_preproc_if_token1] = ACTIONS(2592),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2594),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2594),
    [anon_sym_LPAREN2] = ACTIONS(2594),
    [anon_sym_STAR] = ACTIONS(2594),
    [anon_sym_CARET] = ACTIONS(2594),
    [anon_sym_GT] = ACTIONS(2594),
    [anon_sym_LT] = ACTIONS(2594),
    [anon_sym___extension__] = ACTIONS(2592),
    [anon_sym___attribute__] = ACTIONS(2592),
    [anon_sym___attribute] = ACTIONS(2592),
    [anon_sym_noreturn] = ACTIONS(2592),
    [anon_sym_LBRACK] = ACTIONS(2594),
    [anon_sym___based] = ACTIONS(2592),
    [anon_sym_LBRACE] = ACTIONS(2594),
    [anon_sym_signed] = ACTIONS(2592),
    [anon_sym_unsigned] = ACTIONS(2592),
    [anon_sym_long] = ACTIONS(2592),
    [anon_sym_short] = ACTIONS(2592),
    [anon_sym_ATautoreleasepool] = ACTIONS(2594),
    [anon_sym_const] = ACTIONS(2592),
    [anon_sym_constexpr] = ACTIONS(2592),
    [anon_sym_volatile] = ACTIONS(2592),
    [anon_sym_restrict] = ACTIONS(2592),
    [anon_sym___restrict__] = ACTIONS(2592),
    [anon_sym__Atomic] = ACTIONS(2592),
    [anon_sym__Noreturn] = ACTIONS(2592),
    [anon_sym__Nonnull] = ACTIONS(2592),
    [anon_sym_nullable] = ACTIONS(2592),
    [anon_sym__Complex] = ACTIONS(2592),
    [anon_sym__Nullable] = ACTIONS(2592),
    [anon_sym__Nullable_result] = ACTIONS(2592),
    [anon_sym__Null_unspecified] = ACTIONS(2592),
    [anon_sym___autoreleasing] = ACTIONS(2592),
    [anon_sym___block] = ACTIONS(2592),
    [anon_sym___bridge] = ACTIONS(2592),
    [anon_sym___bridge_retained] = ACTIONS(2592),
    [anon_sym___bridge_transfer] = ACTIONS(2592),
    [anon_sym___complex] = ACTIONS(2592),
    [anon_sym___const] = ACTIONS(2592),
    [anon_sym___imag] = ACTIONS(2592),
    [anon_sym___kindof] = ACTIONS(2592),
    [anon_sym___nonnull] = ACTIONS(2592),
    [anon_sym___nullable] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2592),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2592),
    [anon_sym___real] = ACTIONS(2592),
    [anon_sym___strong] = ACTIONS(2592),
    [anon_sym___unsafe_unretained] = ACTIONS(2592),
    [anon_sym___unused] = ACTIONS(2592),
    [anon_sym___weak] = ACTIONS(2592),
    [anon_sym_alignas] = ACTIONS(2592),
    [anon_sym__Alignas] = ACTIONS(2592),
    [sym_primitive_type] = ACTIONS(2592),
    [anon_sym_enum] = ACTIONS(2592),
    [anon_sym_COLON] = ACTIONS(2594),
    [anon_sym_struct] = ACTIONS(2592),
    [anon_sym_union] = ACTIONS(2592),
    [anon_sym_in] = ACTIONS(2592),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2592),
    [anon_sym___typeof] = ACTIONS(2592),
    [anon_sym_typeof] = ACTIONS(2592),
    [anon_sym_ATprivate] = ACTIONS(2594),
    [anon_sym_ATprotected] = ACTIONS(2594),
    [anon_sym_ATpackage] = ACTIONS(2594),
    [anon_sym_ATpublic] = ACTIONS(2594),
    [anon_sym_BOOL] = ACTIONS(2592),
    [anon_sym_IMP] = ACTIONS(2592),
    [anon_sym_SEL] = ACTIONS(2592),
    [anon_sym_Class] = ACTIONS(2592),
    [anon_sym_id] = ACTIONS(2592),
    [anon_sym_out] = ACTIONS(2592),
    [anon_sym_inout] = ACTIONS(2592),
    [anon_sym_bycopy] = ACTIONS(2592),
    [anon_sym_byref] = ACTIONS(2592),
    [anon_sym_oneway] = ACTIONS(2592),
  },
  [STATE(3764)] = {
    [sym_attribute_specifier] = STATE(3807),
    [sym_gnu_asm_expression] = STATE(3806),
    [aux_sym_function_declarator_repeat1] = STATE(3807),
    [sym_identifier] = ACTIONS(6471),
    [anon_sym_COMMA] = ACTIONS(6473),
    [anon_sym_RPAREN] = ACTIONS(6473),
    [anon_sym_LPAREN2] = ACTIONS(6473),
    [anon_sym_STAR] = ACTIONS(6473),
    [anon_sym_CARET] = ACTIONS(6473),
    [anon_sym_GT] = ACTIONS(6473),
    [anon_sym_LT] = ACTIONS(6473),
    [anon_sym___extension__] = ACTIONS(6471),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(6471),
    [anon_sym_LBRACK] = ACTIONS(6473),
    [anon_sym___based] = ACTIONS(6471),
    [anon_sym_LBRACE] = ACTIONS(6473),
    [anon_sym_signed] = ACTIONS(6471),
    [anon_sym_unsigned] = ACTIONS(6471),
    [anon_sym_long] = ACTIONS(6471),
    [anon_sym_short] = ACTIONS(6471),
    [anon_sym_ATautoreleasepool] = ACTIONS(6473),
    [anon_sym_const] = ACTIONS(6471),
    [anon_sym_constexpr] = ACTIONS(6471),
    [anon_sym_volatile] = ACTIONS(6471),
    [anon_sym_restrict] = ACTIONS(6471),
    [anon_sym___restrict__] = ACTIONS(6471),
    [anon_sym__Atomic] = ACTIONS(6471),
    [anon_sym__Noreturn] = ACTIONS(6471),
    [anon_sym__Nonnull] = ACTIONS(6471),
    [anon_sym_nullable] = ACTIONS(6471),
    [anon_sym__Complex] = ACTIONS(6471),
    [anon_sym__Nullable] = ACTIONS(6471),
    [anon_sym__Nullable_result] = ACTIONS(6471),
    [anon_sym__Null_unspecified] = ACTIONS(6471),
    [anon_sym___autoreleasing] = ACTIONS(6471),
    [anon_sym___block] = ACTIONS(6471),
    [anon_sym___bridge] = ACTIONS(6471),
    [anon_sym___bridge_retained] = ACTIONS(6471),
    [anon_sym___bridge_transfer] = ACTIONS(6471),
    [anon_sym___complex] = ACTIONS(6471),
    [anon_sym___const] = ACTIONS(6471),
    [anon_sym___imag] = ACTIONS(6471),
    [anon_sym___kindof] = ACTIONS(6471),
    [anon_sym___nonnull] = ACTIONS(6471),
    [anon_sym___nullable] = ACTIONS(6471),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6471),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6471),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6471),
    [anon_sym___real] = ACTIONS(6471),
    [anon_sym___strong] = ACTIONS(6471),
    [anon_sym___unsafe_unretained] = ACTIONS(6471),
    [anon_sym___unused] = ACTIONS(6471),
    [anon_sym___weak] = ACTIONS(6471),
    [anon_sym_alignas] = ACTIONS(6471),
    [anon_sym__Alignas] = ACTIONS(6471),
    [sym_primitive_type] = ACTIONS(6471),
    [anon_sym_enum] = ACTIONS(6471),
    [anon_sym_struct] = ACTIONS(6471),
    [anon_sym_union] = ACTIONS(6471),
    [anon_sym_in] = ACTIONS(6471),
    [anon_sym_asm] = ACTIONS(8722),
    [anon_sym___asm__] = ACTIONS(8722),
    [anon_sym___asm] = ACTIONS(8722),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6471),
    [anon_sym___typeof] = ACTIONS(6471),
    [anon_sym_typeof] = ACTIONS(6471),
    [anon_sym_BOOL] = ACTIONS(6471),
    [anon_sym_IMP] = ACTIONS(6471),
    [anon_sym_SEL] = ACTIONS(6471),
    [anon_sym_Class] = ACTIONS(6471),
    [anon_sym_id] = ACTIONS(6471),
    [anon_sym_out] = ACTIONS(6471),
    [anon_sym_inout] = ACTIONS(6471),
    [anon_sym_bycopy] = ACTIONS(6471),
    [anon_sym_byref] = ACTIONS(6471),
    [anon_sym_oneway] = ACTIONS(6471),
  },
  [STATE(3765)] = {
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2267),
    [sym_atomic_declaration] = STATE(2267),
    [sym_specifier_qualifier] = STATE(3327),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3327),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8681),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3766)] = {
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2276),
    [sym_atomic_declaration] = STATE(2276),
    [sym_specifier_qualifier] = STATE(3327),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3327),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8681),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3767)] = {
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2403),
    [sym_atomic_declaration] = STATE(2403),
    [sym_specifier_qualifier] = STATE(3326),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3326),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8679),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3768)] = {
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2420),
    [sym_atomic_declaration] = STATE(2420),
    [sym_specifier_qualifier] = STATE(3326),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3326),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8679),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3769)] = {
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2423),
    [sym_atomic_declaration] = STATE(2423),
    [sym_specifier_qualifier] = STATE(3326),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3326),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8679),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3770)] = {
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2248),
    [sym_atomic_declaration] = STATE(2248),
    [sym_specifier_qualifier] = STATE(3327),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3327),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8681),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3771)] = {
    [sym_type_qualifier] = STATE(3772),
    [sym_alignas_qualifier] = STATE(3836),
    [aux_sym__type_definition_type_repeat1] = STATE(3772),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3787),
    [sym_identifier] = ACTIONS(8724),
    [anon_sym_COMMA] = ACTIONS(6908),
    [anon_sym_RPAREN] = ACTIONS(6908),
    [anon_sym_LPAREN2] = ACTIONS(6908),
    [anon_sym_STAR] = ACTIONS(6908),
    [anon_sym_CARET] = ACTIONS(6908),
    [anon_sym_GT] = ACTIONS(6908),
    [anon_sym_LT] = ACTIONS(6908),
    [anon_sym___extension__] = ACTIONS(8727),
    [anon_sym___attribute__] = ACTIONS(6913),
    [anon_sym___attribute] = ACTIONS(6913),
    [anon_sym_noreturn] = ACTIONS(8727),
    [anon_sym_LBRACK] = ACTIONS(6908),
    [anon_sym___based] = ACTIONS(6913),
    [anon_sym_LBRACE] = ACTIONS(6908),
    [anon_sym_signed] = ACTIONS(8730),
    [anon_sym_unsigned] = ACTIONS(8730),
    [anon_sym_long] = ACTIONS(8730),
    [anon_sym_short] = ACTIONS(8730),
    [anon_sym_ATautoreleasepool] = ACTIONS(6908),
    [anon_sym_const] = ACTIONS(8727),
    [anon_sym_constexpr] = ACTIONS(8727),
    [anon_sym_volatile] = ACTIONS(8727),
    [anon_sym_restrict] = ACTIONS(8727),
    [anon_sym___restrict__] = ACTIONS(8727),
    [anon_sym__Atomic] = ACTIONS(8727),
    [anon_sym__Noreturn] = ACTIONS(8727),
    [anon_sym__Nonnull] = ACTIONS(8727),
    [anon_sym_nullable] = ACTIONS(8727),
    [anon_sym__Complex] = ACTIONS(8727),
    [anon_sym__Nullable] = ACTIONS(8727),
    [anon_sym__Nullable_result] = ACTIONS(8727),
    [anon_sym__Null_unspecified] = ACTIONS(8727),
    [anon_sym___autoreleasing] = ACTIONS(8727),
    [anon_sym___block] = ACTIONS(8727),
    [anon_sym___bridge] = ACTIONS(8727),
    [anon_sym___bridge_retained] = ACTIONS(8727),
    [anon_sym___bridge_transfer] = ACTIONS(8727),
    [anon_sym___complex] = ACTIONS(8727),
    [anon_sym___const] = ACTIONS(8727),
    [anon_sym___imag] = ACTIONS(8727),
    [anon_sym___kindof] = ACTIONS(8727),
    [anon_sym___nonnull] = ACTIONS(8727),
    [anon_sym___nullable] = ACTIONS(8727),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8727),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8727),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8727),
    [anon_sym___real] = ACTIONS(8727),
    [anon_sym___strong] = ACTIONS(8727),
    [anon_sym___unsafe_unretained] = ACTIONS(8727),
    [anon_sym___unused] = ACTIONS(8727),
    [anon_sym___weak] = ACTIONS(8727),
    [anon_sym_alignas] = ACTIONS(8733),
    [anon_sym__Alignas] = ACTIONS(8733),
    [sym_primitive_type] = ACTIONS(8736),
    [anon_sym_enum] = ACTIONS(6913),
    [anon_sym_COLON] = ACTIONS(6908),
    [anon_sym_struct] = ACTIONS(6913),
    [anon_sym_union] = ACTIONS(6913),
    [anon_sym_in] = ACTIONS(6913),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6913),
    [anon_sym___typeof] = ACTIONS(6913),
    [anon_sym_typeof] = ACTIONS(6913),
    [anon_sym_BOOL] = ACTIONS(6913),
    [anon_sym_IMP] = ACTIONS(6913),
    [anon_sym_SEL] = ACTIONS(6913),
    [anon_sym_Class] = ACTIONS(6913),
    [anon_sym_id] = ACTIONS(6913),
    [anon_sym_out] = ACTIONS(6913),
    [anon_sym_inout] = ACTIONS(6913),
    [anon_sym_bycopy] = ACTIONS(6913),
    [anon_sym_byref] = ACTIONS(6913),
    [anon_sym_oneway] = ACTIONS(6913),
  },
  [STATE(3772)] = {
    [sym_type_qualifier] = STATE(3779),
    [sym_alignas_qualifier] = STATE(3836),
    [aux_sym__type_definition_type_repeat1] = STATE(3779),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3818),
    [sym_identifier] = ACTIONS(8739),
    [anon_sym_COMMA] = ACTIONS(6879),
    [anon_sym_RPAREN] = ACTIONS(6879),
    [anon_sym_LPAREN2] = ACTIONS(6879),
    [anon_sym_STAR] = ACTIONS(6879),
    [anon_sym_CARET] = ACTIONS(6879),
    [anon_sym_GT] = ACTIONS(6879),
    [anon_sym_LT] = ACTIONS(6879),
    [anon_sym___extension__] = ACTIONS(8742),
    [anon_sym___attribute__] = ACTIONS(6884),
    [anon_sym___attribute] = ACTIONS(6884),
    [anon_sym_noreturn] = ACTIONS(8742),
    [anon_sym_LBRACK] = ACTIONS(6879),
    [anon_sym___based] = ACTIONS(6884),
    [anon_sym_LBRACE] = ACTIONS(6879),
    [anon_sym_signed] = ACTIONS(8745),
    [anon_sym_unsigned] = ACTIONS(8745),
    [anon_sym_long] = ACTIONS(8745),
    [anon_sym_short] = ACTIONS(8745),
    [anon_sym_ATautoreleasepool] = ACTIONS(6879),
    [anon_sym_const] = ACTIONS(8742),
    [anon_sym_constexpr] = ACTIONS(8742),
    [anon_sym_volatile] = ACTIONS(8742),
    [anon_sym_restrict] = ACTIONS(8742),
    [anon_sym___restrict__] = ACTIONS(8742),
    [anon_sym__Atomic] = ACTIONS(8742),
    [anon_sym__Noreturn] = ACTIONS(8742),
    [anon_sym__Nonnull] = ACTIONS(8742),
    [anon_sym_nullable] = ACTIONS(8742),
    [anon_sym__Complex] = ACTIONS(8742),
    [anon_sym__Nullable] = ACTIONS(8742),
    [anon_sym__Nullable_result] = ACTIONS(8742),
    [anon_sym__Null_unspecified] = ACTIONS(8742),
    [anon_sym___autoreleasing] = ACTIONS(8742),
    [anon_sym___block] = ACTIONS(8742),
    [anon_sym___bridge] = ACTIONS(8742),
    [anon_sym___bridge_retained] = ACTIONS(8742),
    [anon_sym___bridge_transfer] = ACTIONS(8742),
    [anon_sym___complex] = ACTIONS(8742),
    [anon_sym___const] = ACTIONS(8742),
    [anon_sym___imag] = ACTIONS(8742),
    [anon_sym___kindof] = ACTIONS(8742),
    [anon_sym___nonnull] = ACTIONS(8742),
    [anon_sym___nullable] = ACTIONS(8742),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8742),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8742),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8742),
    [anon_sym___real] = ACTIONS(8742),
    [anon_sym___strong] = ACTIONS(8742),
    [anon_sym___unsafe_unretained] = ACTIONS(8742),
    [anon_sym___unused] = ACTIONS(8742),
    [anon_sym___weak] = ACTIONS(8742),
    [anon_sym_alignas] = ACTIONS(8748),
    [anon_sym__Alignas] = ACTIONS(8748),
    [sym_primitive_type] = ACTIONS(8751),
    [anon_sym_enum] = ACTIONS(6884),
    [anon_sym_COLON] = ACTIONS(6879),
    [anon_sym_struct] = ACTIONS(6884),
    [anon_sym_union] = ACTIONS(6884),
    [anon_sym_in] = ACTIONS(6884),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6884),
    [anon_sym___typeof] = ACTIONS(6884),
    [anon_sym_typeof] = ACTIONS(6884),
    [anon_sym_BOOL] = ACTIONS(6884),
    [anon_sym_IMP] = ACTIONS(6884),
    [anon_sym_SEL] = ACTIONS(6884),
    [anon_sym_Class] = ACTIONS(6884),
    [anon_sym_id] = ACTIONS(6884),
    [anon_sym_out] = ACTIONS(6884),
    [anon_sym_inout] = ACTIONS(6884),
    [anon_sym_bycopy] = ACTIONS(6884),
    [anon_sym_byref] = ACTIONS(6884),
    [anon_sym_oneway] = ACTIONS(6884),
  },
  [STATE(3773)] = {
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2335),
    [sym_atomic_declaration] = STATE(2335),
    [sym_specifier_qualifier] = STATE(3324),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3324),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8683),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3774)] = {
    [sym_identifier] = ACTIONS(6622),
    [anon_sym_COMMA] = ACTIONS(6624),
    [anon_sym_RPAREN] = ACTIONS(6624),
    [anon_sym_LPAREN2] = ACTIONS(6624),
    [anon_sym_STAR] = ACTIONS(6624),
    [anon_sym_CARET] = ACTIONS(6624),
    [anon_sym_GT] = ACTIONS(6624),
    [anon_sym_LT] = ACTIONS(6624),
    [anon_sym_SEMI] = ACTIONS(6624),
    [anon_sym___extension__] = ACTIONS(6622),
    [anon_sym___attribute__] = ACTIONS(6622),
    [anon_sym___attribute] = ACTIONS(6622),
    [anon_sym_noreturn] = ACTIONS(6622),
    [anon_sym_LBRACK] = ACTIONS(6624),
    [anon_sym___based] = ACTIONS(6622),
    [anon_sym_LBRACE] = ACTIONS(6624),
    [anon_sym_signed] = ACTIONS(6622),
    [anon_sym_unsigned] = ACTIONS(6622),
    [anon_sym_long] = ACTIONS(6622),
    [anon_sym_short] = ACTIONS(6622),
    [anon_sym_ATautoreleasepool] = ACTIONS(6624),
    [anon_sym_const] = ACTIONS(6622),
    [anon_sym_constexpr] = ACTIONS(6622),
    [anon_sym_volatile] = ACTIONS(6622),
    [anon_sym_restrict] = ACTIONS(6622),
    [anon_sym___restrict__] = ACTIONS(6622),
    [anon_sym__Atomic] = ACTIONS(6622),
    [anon_sym__Noreturn] = ACTIONS(6622),
    [anon_sym__Nonnull] = ACTIONS(6622),
    [anon_sym_nullable] = ACTIONS(6622),
    [anon_sym__Complex] = ACTIONS(6622),
    [anon_sym__Nullable] = ACTIONS(6622),
    [anon_sym__Nullable_result] = ACTIONS(6622),
    [anon_sym__Null_unspecified] = ACTIONS(6622),
    [anon_sym___autoreleasing] = ACTIONS(6622),
    [anon_sym___block] = ACTIONS(6622),
    [anon_sym___bridge] = ACTIONS(6622),
    [anon_sym___bridge_retained] = ACTIONS(6622),
    [anon_sym___bridge_transfer] = ACTIONS(6622),
    [anon_sym___complex] = ACTIONS(6622),
    [anon_sym___const] = ACTIONS(6622),
    [anon_sym___imag] = ACTIONS(6622),
    [anon_sym___kindof] = ACTIONS(6622),
    [anon_sym___nonnull] = ACTIONS(6622),
    [anon_sym___nullable] = ACTIONS(6622),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6622),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6622),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6622),
    [anon_sym___real] = ACTIONS(6622),
    [anon_sym___strong] = ACTIONS(6622),
    [anon_sym___unsafe_unretained] = ACTIONS(6622),
    [anon_sym___unused] = ACTIONS(6622),
    [anon_sym___weak] = ACTIONS(6622),
    [anon_sym_alignas] = ACTIONS(6622),
    [anon_sym__Alignas] = ACTIONS(6622),
    [sym_primitive_type] = ACTIONS(6622),
    [anon_sym_enum] = ACTIONS(6622),
    [anon_sym_COLON] = ACTIONS(6624),
    [anon_sym_struct] = ACTIONS(6622),
    [anon_sym_union] = ACTIONS(6622),
    [anon_sym_in] = ACTIONS(6622),
    [anon_sym_asm] = ACTIONS(6622),
    [anon_sym___asm__] = ACTIONS(6622),
    [anon_sym___asm] = ACTIONS(6622),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6622),
    [anon_sym___typeof] = ACTIONS(6622),
    [anon_sym_typeof] = ACTIONS(6622),
    [anon_sym_BOOL] = ACTIONS(6622),
    [anon_sym_IMP] = ACTIONS(6622),
    [anon_sym_SEL] = ACTIONS(6622),
    [anon_sym_Class] = ACTIONS(6622),
    [anon_sym_id] = ACTIONS(6622),
    [anon_sym_out] = ACTIONS(6622),
    [anon_sym_inout] = ACTIONS(6622),
    [anon_sym_bycopy] = ACTIONS(6622),
    [anon_sym_byref] = ACTIONS(6622),
    [anon_sym_oneway] = ACTIONS(6622),
  },
  [STATE(3775)] = {
    [sym_identifier] = ACTIONS(6648),
    [anon_sym_COMMA] = ACTIONS(6650),
    [anon_sym_RPAREN] = ACTIONS(6650),
    [anon_sym_LPAREN2] = ACTIONS(6650),
    [anon_sym_STAR] = ACTIONS(6650),
    [anon_sym_CARET] = ACTIONS(6650),
    [anon_sym_GT] = ACTIONS(6650),
    [anon_sym_LT] = ACTIONS(6650),
    [anon_sym_SEMI] = ACTIONS(6650),
    [anon_sym___extension__] = ACTIONS(6648),
    [anon_sym___attribute__] = ACTIONS(6648),
    [anon_sym___attribute] = ACTIONS(6648),
    [anon_sym_noreturn] = ACTIONS(6648),
    [anon_sym_LBRACK] = ACTIONS(6650),
    [anon_sym___based] = ACTIONS(6648),
    [anon_sym_LBRACE] = ACTIONS(6650),
    [anon_sym_signed] = ACTIONS(6648),
    [anon_sym_unsigned] = ACTIONS(6648),
    [anon_sym_long] = ACTIONS(6648),
    [anon_sym_short] = ACTIONS(6648),
    [anon_sym_ATautoreleasepool] = ACTIONS(6650),
    [anon_sym_const] = ACTIONS(6648),
    [anon_sym_constexpr] = ACTIONS(6648),
    [anon_sym_volatile] = ACTIONS(6648),
    [anon_sym_restrict] = ACTIONS(6648),
    [anon_sym___restrict__] = ACTIONS(6648),
    [anon_sym__Atomic] = ACTIONS(6648),
    [anon_sym__Noreturn] = ACTIONS(6648),
    [anon_sym__Nonnull] = ACTIONS(6648),
    [anon_sym_nullable] = ACTIONS(6648),
    [anon_sym__Complex] = ACTIONS(6648),
    [anon_sym__Nullable] = ACTIONS(6648),
    [anon_sym__Nullable_result] = ACTIONS(6648),
    [anon_sym__Null_unspecified] = ACTIONS(6648),
    [anon_sym___autoreleasing] = ACTIONS(6648),
    [anon_sym___block] = ACTIONS(6648),
    [anon_sym___bridge] = ACTIONS(6648),
    [anon_sym___bridge_retained] = ACTIONS(6648),
    [anon_sym___bridge_transfer] = ACTIONS(6648),
    [anon_sym___complex] = ACTIONS(6648),
    [anon_sym___const] = ACTIONS(6648),
    [anon_sym___imag] = ACTIONS(6648),
    [anon_sym___kindof] = ACTIONS(6648),
    [anon_sym___nonnull] = ACTIONS(6648),
    [anon_sym___nullable] = ACTIONS(6648),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6648),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6648),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6648),
    [anon_sym___real] = ACTIONS(6648),
    [anon_sym___strong] = ACTIONS(6648),
    [anon_sym___unsafe_unretained] = ACTIONS(6648),
    [anon_sym___unused] = ACTIONS(6648),
    [anon_sym___weak] = ACTIONS(6648),
    [anon_sym_alignas] = ACTIONS(6648),
    [anon_sym__Alignas] = ACTIONS(6648),
    [sym_primitive_type] = ACTIONS(6648),
    [anon_sym_enum] = ACTIONS(6648),
    [anon_sym_COLON] = ACTIONS(6650),
    [anon_sym_struct] = ACTIONS(6648),
    [anon_sym_union] = ACTIONS(6648),
    [anon_sym_in] = ACTIONS(6648),
    [anon_sym_asm] = ACTIONS(6648),
    [anon_sym___asm__] = ACTIONS(6648),
    [anon_sym___asm] = ACTIONS(6648),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6648),
    [anon_sym___typeof] = ACTIONS(6648),
    [anon_sym_typeof] = ACTIONS(6648),
    [anon_sym_BOOL] = ACTIONS(6648),
    [anon_sym_IMP] = ACTIONS(6648),
    [anon_sym_SEL] = ACTIONS(6648),
    [anon_sym_Class] = ACTIONS(6648),
    [anon_sym_id] = ACTIONS(6648),
    [anon_sym_out] = ACTIONS(6648),
    [anon_sym_inout] = ACTIONS(6648),
    [anon_sym_bycopy] = ACTIONS(6648),
    [anon_sym_byref] = ACTIONS(6648),
    [anon_sym_oneway] = ACTIONS(6648),
  },
  [STATE(3776)] = {
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2245),
    [sym_atomic_declaration] = STATE(2245),
    [sym_specifier_qualifier] = STATE(3324),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3324),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8683),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3777)] = {
    [sym_identifier] = ACTIONS(6662),
    [anon_sym_COMMA] = ACTIONS(6654),
    [anon_sym_RPAREN] = ACTIONS(6654),
    [anon_sym_LPAREN2] = ACTIONS(6654),
    [anon_sym_STAR] = ACTIONS(6654),
    [anon_sym_CARET] = ACTIONS(6654),
    [anon_sym_GT] = ACTIONS(6654),
    [anon_sym_LT] = ACTIONS(6654),
    [anon_sym_SEMI] = ACTIONS(6654),
    [anon_sym___extension__] = ACTIONS(6662),
    [anon_sym___attribute__] = ACTIONS(6662),
    [anon_sym___attribute] = ACTIONS(6662),
    [anon_sym_noreturn] = ACTIONS(6662),
    [anon_sym_LBRACK] = ACTIONS(6654),
    [anon_sym___based] = ACTIONS(6662),
    [anon_sym_LBRACE] = ACTIONS(6654),
    [anon_sym_signed] = ACTIONS(6662),
    [anon_sym_unsigned] = ACTIONS(6662),
    [anon_sym_long] = ACTIONS(6662),
    [anon_sym_short] = ACTIONS(6662),
    [anon_sym_ATautoreleasepool] = ACTIONS(6654),
    [anon_sym_const] = ACTIONS(6662),
    [anon_sym_constexpr] = ACTIONS(6662),
    [anon_sym_volatile] = ACTIONS(6662),
    [anon_sym_restrict] = ACTIONS(6662),
    [anon_sym___restrict__] = ACTIONS(6662),
    [anon_sym__Atomic] = ACTIONS(6662),
    [anon_sym__Noreturn] = ACTIONS(6662),
    [anon_sym__Nonnull] = ACTIONS(6662),
    [anon_sym_nullable] = ACTIONS(6662),
    [anon_sym__Complex] = ACTIONS(6662),
    [anon_sym__Nullable] = ACTIONS(6662),
    [anon_sym__Nullable_result] = ACTIONS(6662),
    [anon_sym__Null_unspecified] = ACTIONS(6662),
    [anon_sym___autoreleasing] = ACTIONS(6662),
    [anon_sym___block] = ACTIONS(6662),
    [anon_sym___bridge] = ACTIONS(6662),
    [anon_sym___bridge_retained] = ACTIONS(6662),
    [anon_sym___bridge_transfer] = ACTIONS(6662),
    [anon_sym___complex] = ACTIONS(6662),
    [anon_sym___const] = ACTIONS(6662),
    [anon_sym___imag] = ACTIONS(6662),
    [anon_sym___kindof] = ACTIONS(6662),
    [anon_sym___nonnull] = ACTIONS(6662),
    [anon_sym___nullable] = ACTIONS(6662),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6662),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6662),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6662),
    [anon_sym___real] = ACTIONS(6662),
    [anon_sym___strong] = ACTIONS(6662),
    [anon_sym___unsafe_unretained] = ACTIONS(6662),
    [anon_sym___unused] = ACTIONS(6662),
    [anon_sym___weak] = ACTIONS(6662),
    [anon_sym_alignas] = ACTIONS(6662),
    [anon_sym__Alignas] = ACTIONS(6662),
    [sym_primitive_type] = ACTIONS(6662),
    [anon_sym_enum] = ACTIONS(6662),
    [anon_sym_COLON] = ACTIONS(6654),
    [anon_sym_struct] = ACTIONS(6662),
    [anon_sym_union] = ACTIONS(6662),
    [anon_sym_in] = ACTIONS(6662),
    [anon_sym_asm] = ACTIONS(6662),
    [anon_sym___asm__] = ACTIONS(6662),
    [anon_sym___asm] = ACTIONS(6662),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6662),
    [anon_sym___typeof] = ACTIONS(6662),
    [anon_sym_typeof] = ACTIONS(6662),
    [anon_sym_BOOL] = ACTIONS(6662),
    [anon_sym_IMP] = ACTIONS(6662),
    [anon_sym_SEL] = ACTIONS(6662),
    [anon_sym_Class] = ACTIONS(6662),
    [anon_sym_id] = ACTIONS(6662),
    [anon_sym_out] = ACTIONS(6662),
    [anon_sym_inout] = ACTIONS(6662),
    [anon_sym_bycopy] = ACTIONS(6662),
    [anon_sym_byref] = ACTIONS(6662),
    [anon_sym_oneway] = ACTIONS(6662),
  },
  [STATE(3778)] = {
    [sym_type_qualifier] = STATE(3944),
    [sym_alignas_qualifier] = STATE(3836),
    [sym_type_specifier] = STATE(3974),
    [sym_sized_type_specifier] = STATE(3870),
    [sym_enum_specifier] = STATE(3870),
    [sym_struct_specifier] = STATE(3870),
    [sym_union_specifier] = STATE(3870),
    [sym_macro_type_specifier] = STATE(3870),
    [sym_typeof_specifier] = STATE(3870),
    [sym_struct_declaration] = STATE(2327),
    [sym_atomic_declaration] = STATE(2327),
    [sym_specifier_qualifier] = STATE(3324),
    [sym_generic_specifier] = STATE(3870),
    [sym_typedefed_specifier] = STATE(3870),
    [sym_array_type_specifier] = STATE(3870),
    [sym_protocol_qualifier] = STATE(3944),
    [aux_sym_sized_type_specifier_repeat1] = STATE(3771),
    [aux_sym_struct_declaration_repeat1] = STATE(3324),
    [sym_identifier] = ACTIONS(8190),
    [anon_sym___extension__] = ACTIONS(7959),
    [anon_sym_noreturn] = ACTIONS(7959),
    [anon_sym_signed] = ACTIONS(7965),
    [anon_sym_unsigned] = ACTIONS(7965),
    [anon_sym_long] = ACTIONS(7965),
    [anon_sym_short] = ACTIONS(7965),
    [anon_sym_const] = ACTIONS(7959),
    [anon_sym_constexpr] = ACTIONS(7959),
    [anon_sym_volatile] = ACTIONS(7959),
    [anon_sym_restrict] = ACTIONS(7959),
    [anon_sym___restrict__] = ACTIONS(7959),
    [anon_sym__Atomic] = ACTIONS(8683),
    [anon_sym__Noreturn] = ACTIONS(7959),
    [anon_sym__Nonnull] = ACTIONS(7959),
    [anon_sym_nullable] = ACTIONS(7959),
    [anon_sym__Complex] = ACTIONS(7959),
    [anon_sym__Nullable] = ACTIONS(7959),
    [anon_sym__Nullable_result] = ACTIONS(7959),
    [anon_sym__Null_unspecified] = ACTIONS(7959),
    [anon_sym___autoreleasing] = ACTIONS(7959),
    [anon_sym___block] = ACTIONS(7959),
    [anon_sym___bridge] = ACTIONS(7959),
    [anon_sym___bridge_retained] = ACTIONS(7959),
    [anon_sym___bridge_transfer] = ACTIONS(7959),
    [anon_sym___complex] = ACTIONS(7959),
    [anon_sym___const] = ACTIONS(7959),
    [anon_sym___imag] = ACTIONS(7959),
    [anon_sym___kindof] = ACTIONS(7959),
    [anon_sym___nonnull] = ACTIONS(7959),
    [anon_sym___nullable] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7959),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7959),
    [anon_sym___real] = ACTIONS(7959),
    [anon_sym___strong] = ACTIONS(7959),
    [anon_sym___unsafe_unretained] = ACTIONS(7959),
    [anon_sym___unused] = ACTIONS(7959),
    [anon_sym___weak] = ACTIONS(7959),
    [anon_sym_alignas] = ACTIONS(7967),
    [anon_sym__Alignas] = ACTIONS(7967),
    [sym_primitive_type] = ACTIONS(7969),
    [anon_sym_enum] = ACTIONS(8202),
    [anon_sym_struct] = ACTIONS(7973),
    [anon_sym_union] = ACTIONS(7975),
    [anon_sym_in] = ACTIONS(7977),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7979),
    [anon_sym___typeof] = ACTIONS(7979),
    [anon_sym_typeof] = ACTIONS(7979),
    [anon_sym_BOOL] = ACTIONS(7981),
    [anon_sym_IMP] = ACTIONS(7981),
    [anon_sym_SEL] = ACTIONS(7981),
    [anon_sym_Class] = ACTIONS(7981),
    [anon_sym_id] = ACTIONS(7981),
    [anon_sym_out] = ACTIONS(7977),
    [anon_sym_inout] = ACTIONS(7977),
    [anon_sym_bycopy] = ACTIONS(7977),
    [anon_sym_byref] = ACTIONS(7977),
    [anon_sym_oneway] = ACTIONS(7977),
  },
  [STATE(3779)] = {
    [sym_type_qualifier] = STATE(3779),
    [sym_alignas_qualifier] = STATE(3836),
    [aux_sym__type_definition_type_repeat1] = STATE(3779),
    [sym_identifier] = ACTIONS(6292),
    [anon_sym_COMMA] = ACTIONS(6294),
    [anon_sym_RPAREN] = ACTIONS(6294),
    [anon_sym_LPAREN2] = ACTIONS(6294),
    [anon_sym_STAR] = ACTIONS(6294),
    [anon_sym_CARET] = ACTIONS(6294),
    [anon_sym_GT] = ACTIONS(6294),
    [anon_sym_LT] = ACTIONS(6294),
    [anon_sym___extension__] = ACTIONS(8754),
    [anon_sym___attribute__] = ACTIONS(6292),
    [anon_sym___attribute] = ACTIONS(6292),
    [anon_sym_noreturn] = ACTIONS(8754),
    [anon_sym_LBRACK] = ACTIONS(6294),
    [anon_sym___based] = ACTIONS(6292),
    [anon_sym_LBRACE] = ACTIONS(6294),
    [anon_sym_signed] = ACTIONS(6292),
    [anon_sym_unsigned] = ACTIONS(6292),
    [anon_sym_long] = ACTIONS(6292),
    [anon_sym_short] = ACTIONS(6292),
    [anon_sym_ATautoreleasepool] = ACTIONS(6294),
    [anon_sym_const] = ACTIONS(8754),
    [anon_sym_constexpr] = ACTIONS(8754),
    [anon_sym_volatile] = ACTIONS(8754),
    [anon_sym_restrict] = ACTIONS(8754),
    [anon_sym___restrict__] = ACTIONS(8754),
    [anon_sym__Atomic] = ACTIONS(8754),
    [anon_sym__Noreturn] = ACTIONS(8754),
    [anon_sym__Nonnull] = ACTIONS(8754),
    [anon_sym_nullable] = ACTIONS(8754),
    [anon_sym__Complex] = ACTIONS(8754),
    [anon_sym__Nullable] = ACTIONS(8754),
    [anon_sym__Nullable_result] = ACTIONS(8754),
    [anon_sym__Null_unspecified] = ACTIONS(8754),
    [anon_sym___autoreleasing] = ACTIONS(8754),
    [anon_sym___block] = ACTIONS(8754),
    [anon_sym___bridge] = ACTIONS(8754),
    [anon_sym___bridge_retained] = ACTIONS(8754),
    [anon_sym___bridge_transfer] = ACTIONS(8754),
    [anon_sym___complex] = ACTIONS(8754),
    [anon_sym___const] = ACTIONS(8754),
    [anon_sym___imag] = ACTIONS(8754),
    [anon_sym___kindof] = ACTIONS(8754),
    [anon_sym___nonnull] = ACTIONS(8754),
    [anon_sym___nullable] = ACTIONS(8754),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8754),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8754),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8754),
    [anon_sym___real] = ACTIONS(8754),
    [anon_sym___strong] = ACTIONS(8754),
    [anon_sym___unsafe_unretained] = ACTIONS(8754),
    [anon_sym___unused] = ACTIONS(8754),
    [anon_sym___weak] = ACTIONS(8754),
    [anon_sym_alignas] = ACTIONS(8757),
    [anon_sym__Alignas] = ACTIONS(8757),
    [sym_primitive_type] = ACTIONS(6292),
    [anon_sym_enum] = ACTIONS(6292),
    [anon_sym_COLON] = ACTIONS(6294),
    [anon_sym_struct] = ACTIONS(6292),
    [anon_sym_union] = ACTIONS(6292),
    [anon_sym_in] = ACTIONS(6292),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6292),
    [anon_sym___typeof] = ACTIONS(6292),
    [anon_sym_typeof] = ACTIONS(6292),
    [anon_sym_BOOL] = ACTIONS(6292),
    [anon_sym_IMP] = ACTIONS(6292),
    [anon_sym_SEL] = ACTIONS(6292),
    [anon_sym_Class] = ACTIONS(6292),
    [anon_sym_id] = ACTIONS(6292),
    [anon_sym_out] = ACTIONS(6292),
    [anon_sym_inout] = ACTIONS(6292),
    [anon_sym_bycopy] = ACTIONS(6292),
    [anon_sym_byref] = ACTIONS(6292),
    [anon_sym_oneway] = ACTIONS(6292),
  },
  [STATE(3780)] = {
    [sym_attribute_specifier] = STATE(3842),
    [sym_field_declaration_list] = STATE(3793),
    [sym_identifier] = ACTIONS(7006),
    [anon_sym_COMMA] = ACTIONS(7008),
    [anon_sym_RPAREN] = ACTIONS(7008),
    [anon_sym_LPAREN2] = ACTIONS(7008),
    [anon_sym_STAR] = ACTIONS(7008),
    [anon_sym_CARET] = ACTIONS(7008),
    [anon_sym_GT] = ACTIONS(7008),
    [anon_sym_LT] = ACTIONS(7008),
    [anon_sym___extension__] = ACTIONS(7006),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7006),
    [anon_sym_LBRACK] = ACTIONS(7008),
    [anon_sym___based] = ACTIONS(7006),
    [anon_sym_LBRACE] = ACTIONS(8760),
    [anon_sym_signed] = ACTIONS(7006),
    [anon_sym_unsigned] = ACTIONS(7006),
    [anon_sym_long] = ACTIONS(7006),
    [anon_sym_short] = ACTIONS(7006),
    [anon_sym_ATautoreleasepool] = ACTIONS(7008),
    [anon_sym_const] = ACTIONS(7006),
    [anon_sym_constexpr] = ACTIONS(7006),
    [anon_sym_volatile] = ACTIONS(7006),
    [anon_sym_restrict] = ACTIONS(7006),
    [anon_sym___restrict__] = ACTIONS(7006),
    [anon_sym__Atomic] = ACTIONS(7006),
    [anon_sym__Noreturn] = ACTIONS(7006),
    [anon_sym__Nonnull] = ACTIONS(7006),
    [anon_sym_nullable] = ACTIONS(7006),
    [anon_sym__Complex] = ACTIONS(7006),
    [anon_sym__Nullable] = ACTIONS(7006),
    [anon_sym__Nullable_result] = ACTIONS(7006),
    [anon_sym__Null_unspecified] = ACTIONS(7006),
    [anon_sym___autoreleasing] = ACTIONS(7006),
    [anon_sym___block] = ACTIONS(7006),
    [anon_sym___bridge] = ACTIONS(7006),
    [anon_sym___bridge_retained] = ACTIONS(7006),
    [anon_sym___bridge_transfer] = ACTIONS(7006),
    [anon_sym___complex] = ACTIONS(7006),
    [anon_sym___const] = ACTIONS(7006),
    [anon_sym___imag] = ACTIONS(7006),
    [anon_sym___kindof] = ACTIONS(7006),
    [anon_sym___nonnull] = ACTIONS(7006),
    [anon_sym___nullable] = ACTIONS(7006),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7006),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7006),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7006),
    [anon_sym___real] = ACTIONS(7006),
    [anon_sym___strong] = ACTIONS(7006),
    [anon_sym___unsafe_unretained] = ACTIONS(7006),
    [anon_sym___unused] = ACTIONS(7006),
    [anon_sym___weak] = ACTIONS(7006),
    [anon_sym_alignas] = ACTIONS(7006),
    [anon_sym__Alignas] = ACTIONS(7006),
    [sym_primitive_type] = ACTIONS(7006),
    [anon_sym_enum] = ACTIONS(7006),
    [anon_sym_COLON] = ACTIONS(7008),
    [anon_sym_struct] = ACTIONS(7006),
    [anon_sym_union] = ACTIONS(7006),
    [anon_sym_in] = ACTIONS(7006),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7006),
    [anon_sym___typeof] = ACTIONS(7006),
    [anon_sym_typeof] = ACTIONS(7006),
    [anon_sym_BOOL] = ACTIONS(7006),
    [anon_sym_IMP] = ACTIONS(7006),
    [anon_sym_SEL] = ACTIONS(7006),
    [anon_sym_Class] = ACTIONS(7006),
    [anon_sym_id] = ACTIONS(7006),
    [anon_sym_out] = ACTIONS(7006),
    [anon_sym_inout] = ACTIONS(7006),
    [anon_sym_bycopy] = ACTIONS(7006),
    [anon_sym_byref] = ACTIONS(7006),
    [anon_sym_oneway] = ACTIONS(7006),
  },
  [STATE(3781)] = {
    [sym_attribute_specifier] = STATE(3847),
    [sym_field_declaration_list] = STATE(3796),
    [sym_identifier] = ACTIONS(6962),
    [anon_sym_COMMA] = ACTIONS(6964),
    [anon_sym_RPAREN] = ACTIONS(6964),
    [anon_sym_LPAREN2] = ACTIONS(6964),
    [anon_sym_STAR] = ACTIONS(6964),
    [anon_sym_CARET] = ACTIONS(6964),
    [anon_sym_GT] = ACTIONS(6964),
    [anon_sym_LT] = ACTIONS(6964),
    [anon_sym___extension__] = ACTIONS(6962),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(6962),
    [anon_sym_LBRACK] = ACTIONS(6964),
    [anon_sym___based] = ACTIONS(6962),
    [anon_sym_LBRACE] = ACTIONS(8760),
    [anon_sym_signed] = ACTIONS(6962),
    [anon_sym_unsigned] = ACTIONS(6962),
    [anon_sym_long] = ACTIONS(6962),
    [anon_sym_short] = ACTIONS(6962),
    [anon_sym_ATautoreleasepool] = ACTIONS(6964),
    [anon_sym_const] = ACTIONS(6962),
    [anon_sym_constexpr] = ACTIONS(6962),
    [anon_sym_volatile] = ACTIONS(6962),
    [anon_sym_restrict] = ACTIONS(6962),
    [anon_sym___restrict__] = ACTIONS(6962),
    [anon_sym__Atomic] = ACTIONS(6962),
    [anon_sym__Noreturn] = ACTIONS(6962),
    [anon_sym__Nonnull] = ACTIONS(6962),
    [anon_sym_nullable] = ACTIONS(6962),
    [anon_sym__Complex] = ACTIONS(6962),
    [anon_sym__Nullable] = ACTIONS(6962),
    [anon_sym__Nullable_result] = ACTIONS(6962),
    [anon_sym__Null_unspecified] = ACTIONS(6962),
    [anon_sym___autoreleasing] = ACTIONS(6962),
    [anon_sym___block] = ACTIONS(6962),
    [anon_sym___bridge] = ACTIONS(6962),
    [anon_sym___bridge_retained] = ACTIONS(6962),
    [anon_sym___bridge_transfer] = ACTIONS(6962),
    [anon_sym___complex] = ACTIONS(6962),
    [anon_sym___const] = ACTIONS(6962),
    [anon_sym___imag] = ACTIONS(6962),
    [anon_sym___kindof] = ACTIONS(6962),
    [anon_sym___nonnull] = ACTIONS(6962),
    [anon_sym___nullable] = ACTIONS(6962),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6962),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6962),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6962),
    [anon_sym___real] = ACTIONS(6962),
    [anon_sym___strong] = ACTIONS(6962),
    [anon_sym___unsafe_unretained] = ACTIONS(6962),
    [anon_sym___unused] = ACTIONS(6962),
    [anon_sym___weak] = ACTIONS(6962),
    [anon_sym_alignas] = ACTIONS(6962),
    [anon_sym__Alignas] = ACTIONS(6962),
    [sym_primitive_type] = ACTIONS(6962),
    [anon_sym_enum] = ACTIONS(6962),
    [anon_sym_COLON] = ACTIONS(6964),
    [anon_sym_struct] = ACTIONS(6962),
    [anon_sym_union] = ACTIONS(6962),
    [anon_sym_in] = ACTIONS(6962),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6962),
    [anon_sym___typeof] = ACTIONS(6962),
    [anon_sym_typeof] = ACTIONS(6962),
    [anon_sym_BOOL] = ACTIONS(6962),
    [anon_sym_IMP] = ACTIONS(6962),
    [anon_sym_SEL] = ACTIONS(6962),
    [anon_sym_Class] = ACTIONS(6962),
    [anon_sym_id] = ACTIONS(6962),
    [anon_sym_out] = ACTIONS(6962),
    [anon_sym_inout] = ACTIONS(6962),
    [anon_sym_bycopy] = ACTIONS(6962),
    [anon_sym_byref] = ACTIONS(6962),
    [anon_sym_oneway] = ACTIONS(6962),
  },
  [STATE(3782)] = {
    [sym_attribute_specifier] = STATE(3967),
    [sym_field_declaration_list] = STATE(3821),
    [sym_identifier] = ACTIONS(6968),
    [anon_sym_COMMA] = ACTIONS(6970),
    [anon_sym_RPAREN] = ACTIONS(6970),
    [anon_sym_LPAREN2] = ACTIONS(6970),
    [anon_sym_STAR] = ACTIONS(6970),
    [anon_sym_CARET] = ACTIONS(6970),
    [anon_sym_GT] = ACTIONS(6970),
    [anon_sym_LT] = ACTIONS(6970),
    [anon_sym___extension__] = ACTIONS(6968),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(6968),
    [anon_sym_LBRACK] = ACTIONS(6970),
    [anon_sym___based] = ACTIONS(6968),
    [anon_sym_LBRACE] = ACTIONS(8760),
    [anon_sym_signed] = ACTIONS(6968),
    [anon_sym_unsigned] = ACTIONS(6968),
    [anon_sym_long] = ACTIONS(6968),
    [anon_sym_short] = ACTIONS(6968),
    [anon_sym_ATautoreleasepool] = ACTIONS(6970),
    [anon_sym_const] = ACTIONS(6968),
    [anon_sym_constexpr] = ACTIONS(6968),
    [anon_sym_volatile] = ACTIONS(6968),
    [anon_sym_restrict] = ACTIONS(6968),
    [anon_sym___restrict__] = ACTIONS(6968),
    [anon_sym__Atomic] = ACTIONS(6968),
    [anon_sym__Noreturn] = ACTIONS(6968),
    [anon_sym__Nonnull] = ACTIONS(6968),
    [anon_sym_nullable] = ACTIONS(6968),
    [anon_sym__Complex] = ACTIONS(6968),
    [anon_sym__Nullable] = ACTIONS(6968),
    [anon_sym__Nullable_result] = ACTIONS(6968),
    [anon_sym__Null_unspecified] = ACTIONS(6968),
    [anon_sym___autoreleasing] = ACTIONS(6968),
    [anon_sym___block] = ACTIONS(6968),
    [anon_sym___bridge] = ACTIONS(6968),
    [anon_sym___bridge_retained] = ACTIONS(6968),
    [anon_sym___bridge_transfer] = ACTIONS(6968),
    [anon_sym___complex] = ACTIONS(6968),
    [anon_sym___const] = ACTIONS(6968),
    [anon_sym___imag] = ACTIONS(6968),
    [anon_sym___kindof] = ACTIONS(6968),
    [anon_sym___nonnull] = ACTIONS(6968),
    [anon_sym___nullable] = ACTIONS(6968),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6968),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6968),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6968),
    [anon_sym___real] = ACTIONS(6968),
    [anon_sym___strong] = ACTIONS(6968),
    [anon_sym___unsafe_unretained] = ACTIONS(6968),
    [anon_sym___unused] = ACTIONS(6968),
    [anon_sym___weak] = ACTIONS(6968),
    [anon_sym_alignas] = ACTIONS(6968),
    [anon_sym__Alignas] = ACTIONS(6968),
    [sym_primitive_type] = ACTIONS(6968),
    [anon_sym_enum] = ACTIONS(6968),
    [anon_sym_COLON] = ACTIONS(6970),
    [anon_sym_struct] = ACTIONS(6968),
    [anon_sym_union] = ACTIONS(6968),
    [anon_sym_in] = ACTIONS(6968),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6968),
    [anon_sym___typeof] = ACTIONS(6968),
    [anon_sym_typeof] = ACTIONS(6968),
    [anon_sym_BOOL] = ACTIONS(6968),
    [anon_sym_IMP] = ACTIONS(6968),
    [anon_sym_SEL] = ACTIONS(6968),
    [anon_sym_Class] = ACTIONS(6968),
    [anon_sym_id] = ACTIONS(6968),
    [anon_sym_out] = ACTIONS(6968),
    [anon_sym_inout] = ACTIONS(6968),
    [anon_sym_bycopy] = ACTIONS(6968),
    [anon_sym_byref] = ACTIONS(6968),
    [anon_sym_oneway] = ACTIONS(6968),
  },
  [STATE(3783)] = {
    [sym_attribute_specifier] = STATE(3971),
    [sym_field_declaration_list] = STATE(3822),
    [sym_identifier] = ACTIONS(6986),
    [anon_sym_COMMA] = ACTIONS(6988),
    [anon_sym_RPAREN] = ACTIONS(6988),
    [anon_sym_LPAREN2] = ACTIONS(6988),
    [anon_sym_STAR] = ACTIONS(6988),
    [anon_sym_CARET] = ACTIONS(6988),
    [anon_sym_GT] = ACTIONS(6988),
    [anon_sym_LT] = ACTIONS(6988),
    [anon_sym___extension__] = ACTIONS(6986),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(6986),
    [anon_sym_LBRACK] = ACTIONS(6988),
    [anon_sym___based] = ACTIONS(6986),
    [anon_sym_LBRACE] = ACTIONS(8760),
    [anon_sym_signed] = ACTIONS(6986),
    [anon_sym_unsigned] = ACTIONS(6986),
    [anon_sym_long] = ACTIONS(6986),
    [anon_sym_short] = ACTIONS(6986),
    [anon_sym_ATautoreleasepool] = ACTIONS(6988),
    [anon_sym_const] = ACTIONS(6986),
    [anon_sym_constexpr] = ACTIONS(6986),
    [anon_sym_volatile] = ACTIONS(6986),
    [anon_sym_restrict] = ACTIONS(6986),
    [anon_sym___restrict__] = ACTIONS(6986),
    [anon_sym__Atomic] = ACTIONS(6986),
    [anon_sym__Noreturn] = ACTIONS(6986),
    [anon_sym__Nonnull] = ACTIONS(6986),
    [anon_sym_nullable] = ACTIONS(6986),
    [anon_sym__Complex] = ACTIONS(6986),
    [anon_sym__Nullable] = ACTIONS(6986),
    [anon_sym__Nullable_result] = ACTIONS(6986),
    [anon_sym__Null_unspecified] = ACTIONS(6986),
    [anon_sym___autoreleasing] = ACTIONS(6986),
    [anon_sym___block] = ACTIONS(6986),
    [anon_sym___bridge] = ACTIONS(6986),
    [anon_sym___bridge_retained] = ACTIONS(6986),
    [anon_sym___bridge_transfer] = ACTIONS(6986),
    [anon_sym___complex] = ACTIONS(6986),
    [anon_sym___const] = ACTIONS(6986),
    [anon_sym___imag] = ACTIONS(6986),
    [anon_sym___kindof] = ACTIONS(6986),
    [anon_sym___nonnull] = ACTIONS(6986),
    [anon_sym___nullable] = ACTIONS(6986),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6986),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6986),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6986),
    [anon_sym___real] = ACTIONS(6986),
    [anon_sym___strong] = ACTIONS(6986),
    [anon_sym___unsafe_unretained] = ACTIONS(6986),
    [anon_sym___unused] = ACTIONS(6986),
    [anon_sym___weak] = ACTIONS(6986),
    [anon_sym_alignas] = ACTIONS(6986),
    [anon_sym__Alignas] = ACTIONS(6986),
    [sym_primitive_type] = ACTIONS(6986),
    [anon_sym_enum] = ACTIONS(6986),
    [anon_sym_COLON] = ACTIONS(6988),
    [anon_sym_struct] = ACTIONS(6986),
    [anon_sym_union] = ACTIONS(6986),
    [anon_sym_in] = ACTIONS(6986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6986),
    [anon_sym___typeof] = ACTIONS(6986),
    [anon_sym_typeof] = ACTIONS(6986),
    [anon_sym_BOOL] = ACTIONS(6986),
    [anon_sym_IMP] = ACTIONS(6986),
    [anon_sym_SEL] = ACTIONS(6986),
    [anon_sym_Class] = ACTIONS(6986),
    [anon_sym_id] = ACTIONS(6986),
    [anon_sym_out] = ACTIONS(6986),
    [anon_sym_inout] = ACTIONS(6986),
    [anon_sym_bycopy] = ACTIONS(6986),
    [anon_sym_byref] = ACTIONS(6986),
    [anon_sym_oneway] = ACTIONS(6986),
  },
  [STATE(3784)] = {
    [sym_attribute_specifier] = STATE(3878),
    [sym_field_declaration_list] = STATE(3802),
    [sym_identifier] = ACTIONS(7032),
    [anon_sym_COMMA] = ACTIONS(7034),
    [anon_sym_RPAREN] = ACTIONS(7034),
    [anon_sym_LPAREN2] = ACTIONS(7034),
    [anon_sym_STAR] = ACTIONS(7034),
    [anon_sym_CARET] = ACTIONS(7034),
    [anon_sym_GT] = ACTIONS(7034),
    [anon_sym_LT] = ACTIONS(7034),
    [anon_sym___extension__] = ACTIONS(7032),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7032),
    [anon_sym_LBRACK] = ACTIONS(7034),
    [anon_sym___based] = ACTIONS(7032),
    [anon_sym_LBRACE] = ACTIONS(8760),
    [anon_sym_signed] = ACTIONS(7032),
    [anon_sym_unsigned] = ACTIONS(7032),
    [anon_sym_long] = ACTIONS(7032),
    [anon_sym_short] = ACTIONS(7032),
    [anon_sym_ATautoreleasepool] = ACTIONS(7034),
    [anon_sym_const] = ACTIONS(7032),
    [anon_sym_constexpr] = ACTIONS(7032),
    [anon_sym_volatile] = ACTIONS(7032),
    [anon_sym_restrict] = ACTIONS(7032),
    [anon_sym___restrict__] = ACTIONS(7032),
    [anon_sym__Atomic] = ACTIONS(7032),
    [anon_sym__Noreturn] = ACTIONS(7032),
    [anon_sym__Nonnull] = ACTIONS(7032),
    [anon_sym_nullable] = ACTIONS(7032),
    [anon_sym__Complex] = ACTIONS(7032),
    [anon_sym__Nullable] = ACTIONS(7032),
    [anon_sym__Nullable_result] = ACTIONS(7032),
    [anon_sym__Null_unspecified] = ACTIONS(7032),
    [anon_sym___autoreleasing] = ACTIONS(7032),
    [anon_sym___block] = ACTIONS(7032),
    [anon_sym___bridge] = ACTIONS(7032),
    [anon_sym___bridge_retained] = ACTIONS(7032),
    [anon_sym___bridge_transfer] = ACTIONS(7032),
    [anon_sym___complex] = ACTIONS(7032),
    [anon_sym___const] = ACTIONS(7032),
    [anon_sym___imag] = ACTIONS(7032),
    [anon_sym___kindof] = ACTIONS(7032),
    [anon_sym___nonnull] = ACTIONS(7032),
    [anon_sym___nullable] = ACTIONS(7032),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7032),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7032),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7032),
    [anon_sym___real] = ACTIONS(7032),
    [anon_sym___strong] = ACTIONS(7032),
    [anon_sym___unsafe_unretained] = ACTIONS(7032),
    [anon_sym___unused] = ACTIONS(7032),
    [anon_sym___weak] = ACTIONS(7032),
    [anon_sym_alignas] = ACTIONS(7032),
    [anon_sym__Alignas] = ACTIONS(7032),
    [sym_primitive_type] = ACTIONS(7032),
    [anon_sym_enum] = ACTIONS(7032),
    [anon_sym_COLON] = ACTIONS(7034),
    [anon_sym_struct] = ACTIONS(7032),
    [anon_sym_union] = ACTIONS(7032),
    [anon_sym_in] = ACTIONS(7032),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7032),
    [anon_sym___typeof] = ACTIONS(7032),
    [anon_sym_typeof] = ACTIONS(7032),
    [anon_sym_BOOL] = ACTIONS(7032),
    [anon_sym_IMP] = ACTIONS(7032),
    [anon_sym_SEL] = ACTIONS(7032),
    [anon_sym_Class] = ACTIONS(7032),
    [anon_sym_id] = ACTIONS(7032),
    [anon_sym_out] = ACTIONS(7032),
    [anon_sym_inout] = ACTIONS(7032),
    [anon_sym_bycopy] = ACTIONS(7032),
    [anon_sym_byref] = ACTIONS(7032),
    [anon_sym_oneway] = ACTIONS(7032),
  },
  [STATE(3785)] = {
    [sym_attribute_specifier] = STATE(3856),
    [sym_field_declaration_list] = STATE(3794),
    [sym_identifier] = ACTIONS(7036),
    [anon_sym_COMMA] = ACTIONS(7038),
    [anon_sym_RPAREN] = ACTIONS(7038),
    [anon_sym_LPAREN2] = ACTIONS(7038),
    [anon_sym_STAR] = ACTIONS(7038),
    [anon_sym_CARET] = ACTIONS(7038),
    [anon_sym_GT] = ACTIONS(7038),
    [anon_sym_LT] = ACTIONS(7038),
    [anon_sym___extension__] = ACTIONS(7036),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7036),
    [anon_sym_LBRACK] = ACTIONS(7038),
    [anon_sym___based] = ACTIONS(7036),
    [anon_sym_LBRACE] = ACTIONS(8760),
    [anon_sym_signed] = ACTIONS(7036),
    [anon_sym_unsigned] = ACTIONS(7036),
    [anon_sym_long] = ACTIONS(7036),
    [anon_sym_short] = ACTIONS(7036),
    [anon_sym_ATautoreleasepool] = ACTIONS(7038),
    [anon_sym_const] = ACTIONS(7036),
    [anon_sym_constexpr] = ACTIONS(7036),
    [anon_sym_volatile] = ACTIONS(7036),
    [anon_sym_restrict] = ACTIONS(7036),
    [anon_sym___restrict__] = ACTIONS(7036),
    [anon_sym__Atomic] = ACTIONS(7036),
    [anon_sym__Noreturn] = ACTIONS(7036),
    [anon_sym__Nonnull] = ACTIONS(7036),
    [anon_sym_nullable] = ACTIONS(7036),
    [anon_sym__Complex] = ACTIONS(7036),
    [anon_sym__Nullable] = ACTIONS(7036),
    [anon_sym__Nullable_result] = ACTIONS(7036),
    [anon_sym__Null_unspecified] = ACTIONS(7036),
    [anon_sym___autoreleasing] = ACTIONS(7036),
    [anon_sym___block] = ACTIONS(7036),
    [anon_sym___bridge] = ACTIONS(7036),
    [anon_sym___bridge_retained] = ACTIONS(7036),
    [anon_sym___bridge_transfer] = ACTIONS(7036),
    [anon_sym___complex] = ACTIONS(7036),
    [anon_sym___const] = ACTIONS(7036),
    [anon_sym___imag] = ACTIONS(7036),
    [anon_sym___kindof] = ACTIONS(7036),
    [anon_sym___nonnull] = ACTIONS(7036),
    [anon_sym___nullable] = ACTIONS(7036),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7036),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7036),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7036),
    [anon_sym___real] = ACTIONS(7036),
    [anon_sym___strong] = ACTIONS(7036),
    [anon_sym___unsafe_unretained] = ACTIONS(7036),
    [anon_sym___unused] = ACTIONS(7036),
    [anon_sym___weak] = ACTIONS(7036),
    [anon_sym_alignas] = ACTIONS(7036),
    [anon_sym__Alignas] = ACTIONS(7036),
    [sym_primitive_type] = ACTIONS(7036),
    [anon_sym_enum] = ACTIONS(7036),
    [anon_sym_COLON] = ACTIONS(7038),
    [anon_sym_struct] = ACTIONS(7036),
    [anon_sym_union] = ACTIONS(7036),
    [anon_sym_in] = ACTIONS(7036),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7036),
    [anon_sym___typeof] = ACTIONS(7036),
    [anon_sym_typeof] = ACTIONS(7036),
    [anon_sym_BOOL] = ACTIONS(7036),
    [anon_sym_IMP] = ACTIONS(7036),
    [anon_sym_SEL] = ACTIONS(7036),
    [anon_sym_Class] = ACTIONS(7036),
    [anon_sym_id] = ACTIONS(7036),
    [anon_sym_out] = ACTIONS(7036),
    [anon_sym_inout] = ACTIONS(7036),
    [anon_sym_bycopy] = ACTIONS(7036),
    [anon_sym_byref] = ACTIONS(7036),
    [anon_sym_oneway] = ACTIONS(7036),
  },
  [STATE(3786)] = {
    [sym_protocol_reference_list] = STATE(3947),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_RPAREN] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_GT] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(8762),
    [anon_sym___extension__] = ACTIONS(6308),
    [anon_sym___attribute__] = ACTIONS(6308),
    [anon_sym___attribute] = ACTIONS(6308),
    [anon_sym_noreturn] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6310),
    [anon_sym___based] = ACTIONS(6308),
    [anon_sym_LBRACE] = ACTIONS(6310),
    [anon_sym_signed] = ACTIONS(6308),
    [anon_sym_unsigned] = ACTIONS(6308),
    [anon_sym_long] = ACTIONS(6308),
    [anon_sym_short] = ACTIONS(6308),
    [anon_sym_ATautoreleasepool] = ACTIONS(6310),
    [anon_sym_const] = ACTIONS(6308),
    [anon_sym_constexpr] = ACTIONS(6308),
    [anon_sym_volatile] = ACTIONS(6308),
    [anon_sym_restrict] = ACTIONS(6308),
    [anon_sym___restrict__] = ACTIONS(6308),
    [anon_sym__Atomic] = ACTIONS(6308),
    [anon_sym__Noreturn] = ACTIONS(6308),
    [anon_sym__Nonnull] = ACTIONS(6308),
    [anon_sym_nullable] = ACTIONS(6308),
    [anon_sym__Complex] = ACTIONS(6308),
    [anon_sym__Nullable] = ACTIONS(6308),
    [anon_sym__Nullable_result] = ACTIONS(6308),
    [anon_sym__Null_unspecified] = ACTIONS(6308),
    [anon_sym___autoreleasing] = ACTIONS(6308),
    [anon_sym___block] = ACTIONS(6308),
    [anon_sym___bridge] = ACTIONS(6308),
    [anon_sym___bridge_retained] = ACTIONS(6308),
    [anon_sym___bridge_transfer] = ACTIONS(6308),
    [anon_sym___complex] = ACTIONS(6308),
    [anon_sym___const] = ACTIONS(6308),
    [anon_sym___imag] = ACTIONS(6308),
    [anon_sym___kindof] = ACTIONS(6308),
    [anon_sym___nonnull] = ACTIONS(6308),
    [anon_sym___nullable] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6308),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6308),
    [anon_sym___real] = ACTIONS(6308),
    [anon_sym___strong] = ACTIONS(6308),
    [anon_sym___unsafe_unretained] = ACTIONS(6308),
    [anon_sym___unused] = ACTIONS(6308),
    [anon_sym___weak] = ACTIONS(6308),
    [anon_sym_alignas] = ACTIONS(6308),
    [anon_sym__Alignas] = ACTIONS(6308),
    [sym_primitive_type] = ACTIONS(6308),
    [anon_sym_enum] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6310),
    [anon_sym_struct] = ACTIONS(6308),
    [anon_sym_union] = ACTIONS(6308),
    [anon_sym_in] = ACTIONS(6308),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6308),
    [anon_sym___typeof] = ACTIONS(6308),
    [anon_sym_typeof] = ACTIONS(6308),
    [anon_sym_BOOL] = ACTIONS(6308),
    [anon_sym_IMP] = ACTIONS(6308),
    [anon_sym_SEL] = ACTIONS(6308),
    [anon_sym_Class] = ACTIONS(6308),
    [anon_sym_id] = ACTIONS(6308),
    [anon_sym_out] = ACTIONS(6308),
    [anon_sym_inout] = ACTIONS(6308),
    [anon_sym_bycopy] = ACTIONS(6308),
    [anon_sym_byref] = ACTIONS(6308),
    [anon_sym_oneway] = ACTIONS(6308),
  },
  [STATE(3787)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3792),
    [sym_identifier] = ACTIONS(7095),
    [anon_sym_COMMA] = ACTIONS(7098),
    [anon_sym_RPAREN] = ACTIONS(7098),
    [anon_sym_LPAREN2] = ACTIONS(7098),
    [anon_sym_STAR] = ACTIONS(7098),
    [anon_sym_CARET] = ACTIONS(7098),
    [anon_sym_GT] = ACTIONS(7098),
    [anon_sym_LT] = ACTIONS(7098),
    [anon_sym___extension__] = ACTIONS(7095),
    [anon_sym___attribute__] = ACTIONS(7095),
    [anon_sym___attribute] = ACTIONS(7095),
    [anon_sym_noreturn] = ACTIONS(7095),
    [anon_sym_LBRACK] = ACTIONS(7098),
    [anon_sym___based] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7098),
    [anon_sym_signed] = ACTIONS(8764),
    [anon_sym_unsigned] = ACTIONS(8764),
    [anon_sym_long] = ACTIONS(8764),
    [anon_sym_short] = ACTIONS(8764),
    [anon_sym_ATautoreleasepool] = ACTIONS(7098),
    [anon_sym_const] = ACTIONS(7095),
    [anon_sym_constexpr] = ACTIONS(7095),
    [anon_sym_volatile] = ACTIONS(7095),
    [anon_sym_restrict] = ACTIONS(7095),
    [anon_sym___restrict__] = ACTIONS(7095),
    [anon_sym__Atomic] = ACTIONS(7095),
    [anon_sym__Noreturn] = ACTIONS(7095),
    [anon_sym__Nonnull] = ACTIONS(7095),
    [anon_sym_nullable] = ACTIONS(7095),
    [anon_sym__Complex] = ACTIONS(7095),
    [anon_sym__Nullable] = ACTIONS(7095),
    [anon_sym__Nullable_result] = ACTIONS(7095),
    [anon_sym__Null_unspecified] = ACTIONS(7095),
    [anon_sym___autoreleasing] = ACTIONS(7095),
    [anon_sym___block] = ACTIONS(7095),
    [anon_sym___bridge] = ACTIONS(7095),
    [anon_sym___bridge_retained] = ACTIONS(7095),
    [anon_sym___bridge_transfer] = ACTIONS(7095),
    [anon_sym___complex] = ACTIONS(7095),
    [anon_sym___const] = ACTIONS(7095),
    [anon_sym___imag] = ACTIONS(7095),
    [anon_sym___kindof] = ACTIONS(7095),
    [anon_sym___nonnull] = ACTIONS(7095),
    [anon_sym___nullable] = ACTIONS(7095),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7095),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7095),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7095),
    [anon_sym___real] = ACTIONS(7095),
    [anon_sym___strong] = ACTIONS(7095),
    [anon_sym___unsafe_unretained] = ACTIONS(7095),
    [anon_sym___unused] = ACTIONS(7095),
    [anon_sym___weak] = ACTIONS(7095),
    [anon_sym_alignas] = ACTIONS(7095),
    [anon_sym__Alignas] = ACTIONS(7095),
    [sym_primitive_type] = ACTIONS(7095),
    [anon_sym_enum] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7098),
    [anon_sym_struct] = ACTIONS(7095),
    [anon_sym_union] = ACTIONS(7095),
    [anon_sym_in] = ACTIONS(7095),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7095),
    [anon_sym___typeof] = ACTIONS(7095),
    [anon_sym_typeof] = ACTIONS(7095),
    [anon_sym_BOOL] = ACTIONS(7095),
    [anon_sym_IMP] = ACTIONS(7095),
    [anon_sym_SEL] = ACTIONS(7095),
    [anon_sym_Class] = ACTIONS(7095),
    [anon_sym_id] = ACTIONS(7095),
    [anon_sym_out] = ACTIONS(7095),
    [anon_sym_inout] = ACTIONS(7095),
    [anon_sym_bycopy] = ACTIONS(7095),
    [anon_sym_byref] = ACTIONS(7095),
    [anon_sym_oneway] = ACTIONS(7095),
  },
  [STATE(3788)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3814),
    [sym_identifier] = ACTIONS(7183),
    [anon_sym_COMMA] = ACTIONS(7185),
    [anon_sym_RPAREN] = ACTIONS(7185),
    [anon_sym_LPAREN2] = ACTIONS(7185),
    [anon_sym_STAR] = ACTIONS(7185),
    [anon_sym_CARET] = ACTIONS(7185),
    [anon_sym_GT] = ACTIONS(7185),
    [anon_sym_LT] = ACTIONS(7185),
    [anon_sym___extension__] = ACTIONS(7183),
    [anon_sym___attribute__] = ACTIONS(7183),
    [anon_sym___attribute] = ACTIONS(7183),
    [anon_sym_noreturn] = ACTIONS(7183),
    [anon_sym_LBRACK] = ACTIONS(7185),
    [anon_sym___based] = ACTIONS(7183),
    [anon_sym_LBRACE] = ACTIONS(7185),
    [anon_sym_signed] = ACTIONS(8768),
    [anon_sym_unsigned] = ACTIONS(8768),
    [anon_sym_long] = ACTIONS(8768),
    [anon_sym_short] = ACTIONS(8768),
    [anon_sym_ATautoreleasepool] = ACTIONS(7185),
    [anon_sym_const] = ACTIONS(7183),
    [anon_sym_constexpr] = ACTIONS(7183),
    [anon_sym_volatile] = ACTIONS(7183),
    [anon_sym_restrict] = ACTIONS(7183),
    [anon_sym___restrict__] = ACTIONS(7183),
    [anon_sym__Atomic] = ACTIONS(7183),
    [anon_sym__Noreturn] = ACTIONS(7183),
    [anon_sym__Nonnull] = ACTIONS(7183),
    [anon_sym_nullable] = ACTIONS(7183),
    [anon_sym__Complex] = ACTIONS(7183),
    [anon_sym__Nullable] = ACTIONS(7183),
    [anon_sym__Nullable_result] = ACTIONS(7183),
    [anon_sym__Null_unspecified] = ACTIONS(7183),
    [anon_sym___autoreleasing] = ACTIONS(7183),
    [anon_sym___block] = ACTIONS(7183),
    [anon_sym___bridge] = ACTIONS(7183),
    [anon_sym___bridge_retained] = ACTIONS(7183),
    [anon_sym___bridge_transfer] = ACTIONS(7183),
    [anon_sym___complex] = ACTIONS(7183),
    [anon_sym___const] = ACTIONS(7183),
    [anon_sym___imag] = ACTIONS(7183),
    [anon_sym___kindof] = ACTIONS(7183),
    [anon_sym___nonnull] = ACTIONS(7183),
    [anon_sym___nullable] = ACTIONS(7183),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7183),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7183),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7183),
    [anon_sym___real] = ACTIONS(7183),
    [anon_sym___strong] = ACTIONS(7183),
    [anon_sym___unsafe_unretained] = ACTIONS(7183),
    [anon_sym___unused] = ACTIONS(7183),
    [anon_sym___weak] = ACTIONS(7183),
    [anon_sym_alignas] = ACTIONS(7183),
    [anon_sym__Alignas] = ACTIONS(7183),
    [sym_primitive_type] = ACTIONS(7183),
    [anon_sym_enum] = ACTIONS(7183),
    [anon_sym_COLON] = ACTIONS(7185),
    [anon_sym_struct] = ACTIONS(7183),
    [anon_sym_union] = ACTIONS(7183),
    [anon_sym_in] = ACTIONS(7183),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7183),
    [anon_sym___typeof] = ACTIONS(7183),
    [anon_sym_typeof] = ACTIONS(7183),
    [anon_sym_BOOL] = ACTIONS(7183),
    [anon_sym_IMP] = ACTIONS(7183),
    [anon_sym_SEL] = ACTIONS(7183),
    [anon_sym_Class] = ACTIONS(7183),
    [anon_sym_id] = ACTIONS(7183),
    [anon_sym_out] = ACTIONS(7183),
    [anon_sym_inout] = ACTIONS(7183),
    [anon_sym_bycopy] = ACTIONS(7183),
    [anon_sym_byref] = ACTIONS(7183),
    [anon_sym_oneway] = ACTIONS(7183),
  },
  [STATE(3789)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3815),
    [sym_identifier] = ACTIONS(7315),
    [anon_sym_COMMA] = ACTIONS(7317),
    [anon_sym_RPAREN] = ACTIONS(7317),
    [anon_sym_LPAREN2] = ACTIONS(7317),
    [anon_sym_STAR] = ACTIONS(7317),
    [anon_sym_CARET] = ACTIONS(7317),
    [anon_sym_GT] = ACTIONS(7317),
    [anon_sym_LT] = ACTIONS(7317),
    [anon_sym___extension__] = ACTIONS(7315),
    [anon_sym___attribute__] = ACTIONS(7315),
    [anon_sym___attribute] = ACTIONS(7315),
    [anon_sym_noreturn] = ACTIONS(7315),
    [anon_sym_LBRACK] = ACTIONS(7317),
    [anon_sym___based] = ACTIONS(7315),
    [anon_sym_LBRACE] = ACTIONS(7317),
    [anon_sym_signed] = ACTIONS(8771),
    [anon_sym_unsigned] = ACTIONS(8771),
    [anon_sym_long] = ACTIONS(8771),
    [anon_sym_short] = ACTIONS(8771),
    [anon_sym_ATautoreleasepool] = ACTIONS(7317),
    [anon_sym_const] = ACTIONS(7315),
    [anon_sym_constexpr] = ACTIONS(7315),
    [anon_sym_volatile] = ACTIONS(7315),
    [anon_sym_restrict] = ACTIONS(7315),
    [anon_sym___restrict__] = ACTIONS(7315),
    [anon_sym__Atomic] = ACTIONS(7315),
    [anon_sym__Noreturn] = ACTIONS(7315),
    [anon_sym__Nonnull] = ACTIONS(7315),
    [anon_sym_nullable] = ACTIONS(7315),
    [anon_sym__Complex] = ACTIONS(7315),
    [anon_sym__Nullable] = ACTIONS(7315),
    [anon_sym__Nullable_result] = ACTIONS(7315),
    [anon_sym__Null_unspecified] = ACTIONS(7315),
    [anon_sym___autoreleasing] = ACTIONS(7315),
    [anon_sym___block] = ACTIONS(7315),
    [anon_sym___bridge] = ACTIONS(7315),
    [anon_sym___bridge_retained] = ACTIONS(7315),
    [anon_sym___bridge_transfer] = ACTIONS(7315),
    [anon_sym___complex] = ACTIONS(7315),
    [anon_sym___const] = ACTIONS(7315),
    [anon_sym___imag] = ACTIONS(7315),
    [anon_sym___kindof] = ACTIONS(7315),
    [anon_sym___nonnull] = ACTIONS(7315),
    [anon_sym___nullable] = ACTIONS(7315),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7315),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7315),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7315),
    [anon_sym___real] = ACTIONS(7315),
    [anon_sym___strong] = ACTIONS(7315),
    [anon_sym___unsafe_unretained] = ACTIONS(7315),
    [anon_sym___unused] = ACTIONS(7315),
    [anon_sym___weak] = ACTIONS(7315),
    [anon_sym_alignas] = ACTIONS(7315),
    [anon_sym__Alignas] = ACTIONS(7315),
    [sym_primitive_type] = ACTIONS(7315),
    [anon_sym_enum] = ACTIONS(7315),
    [anon_sym_COLON] = ACTIONS(7317),
    [anon_sym_struct] = ACTIONS(7315),
    [anon_sym_union] = ACTIONS(7315),
    [anon_sym_in] = ACTIONS(7315),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7315),
    [anon_sym___typeof] = ACTIONS(7315),
    [anon_sym_typeof] = ACTIONS(7315),
    [anon_sym_BOOL] = ACTIONS(7315),
    [anon_sym_IMP] = ACTIONS(7315),
    [anon_sym_SEL] = ACTIONS(7315),
    [anon_sym_Class] = ACTIONS(7315),
    [anon_sym_id] = ACTIONS(7315),
    [anon_sym_out] = ACTIONS(7315),
    [anon_sym_inout] = ACTIONS(7315),
    [anon_sym_bycopy] = ACTIONS(7315),
    [anon_sym_byref] = ACTIONS(7315),
    [anon_sym_oneway] = ACTIONS(7315),
  },
  [STATE(3790)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3799),
    [sym_identifier] = ACTIONS(5074),
    [anon_sym_COMMA] = ACTIONS(5070),
    [anon_sym_RPAREN] = ACTIONS(5070),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym_GT] = ACTIONS(5070),
    [anon_sym_LT] = ACTIONS(5070),
    [anon_sym___extension__] = ACTIONS(5074),
    [anon_sym___attribute__] = ACTIONS(5074),
    [anon_sym___attribute] = ACTIONS(5074),
    [anon_sym_noreturn] = ACTIONS(5074),
    [anon_sym_LBRACK] = ACTIONS(5070),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym_LBRACE] = ACTIONS(5070),
    [anon_sym_signed] = ACTIONS(8774),
    [anon_sym_unsigned] = ACTIONS(8774),
    [anon_sym_long] = ACTIONS(8774),
    [anon_sym_short] = ACTIONS(8774),
    [anon_sym_ATautoreleasepool] = ACTIONS(5070),
    [anon_sym_const] = ACTIONS(5074),
    [anon_sym_constexpr] = ACTIONS(5074),
    [anon_sym_volatile] = ACTIONS(5074),
    [anon_sym_restrict] = ACTIONS(5074),
    [anon_sym___restrict__] = ACTIONS(5074),
    [anon_sym__Atomic] = ACTIONS(5074),
    [anon_sym__Noreturn] = ACTIONS(5074),
    [anon_sym__Nonnull] = ACTIONS(5074),
    [anon_sym_nullable] = ACTIONS(5074),
    [anon_sym__Complex] = ACTIONS(5074),
    [anon_sym__Nullable] = ACTIONS(5074),
    [anon_sym__Nullable_result] = ACTIONS(5074),
    [anon_sym__Null_unspecified] = ACTIONS(5074),
    [anon_sym___autoreleasing] = ACTIONS(5074),
    [anon_sym___block] = ACTIONS(5074),
    [anon_sym___bridge] = ACTIONS(5074),
    [anon_sym___bridge_retained] = ACTIONS(5074),
    [anon_sym___bridge_transfer] = ACTIONS(5074),
    [anon_sym___complex] = ACTIONS(5074),
    [anon_sym___const] = ACTIONS(5074),
    [anon_sym___imag] = ACTIONS(5074),
    [anon_sym___kindof] = ACTIONS(5074),
    [anon_sym___nonnull] = ACTIONS(5074),
    [anon_sym___nullable] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5074),
    [anon_sym___real] = ACTIONS(5074),
    [anon_sym___strong] = ACTIONS(5074),
    [anon_sym___unsafe_unretained] = ACTIONS(5074),
    [anon_sym___unused] = ACTIONS(5074),
    [anon_sym___weak] = ACTIONS(5074),
    [anon_sym_alignas] = ACTIONS(5074),
    [anon_sym__Alignas] = ACTIONS(5074),
    [sym_primitive_type] = ACTIONS(5074),
    [anon_sym_enum] = ACTIONS(5074),
    [anon_sym_COLON] = ACTIONS(5070),
    [anon_sym_struct] = ACTIONS(5074),
    [anon_sym_union] = ACTIONS(5074),
    [anon_sym_in] = ACTIONS(5074),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5074),
    [anon_sym___typeof] = ACTIONS(5074),
    [anon_sym_typeof] = ACTIONS(5074),
    [anon_sym_BOOL] = ACTIONS(5074),
    [anon_sym_IMP] = ACTIONS(5074),
    [anon_sym_SEL] = ACTIONS(5074),
    [anon_sym_Class] = ACTIONS(5074),
    [anon_sym_id] = ACTIONS(5074),
    [anon_sym_out] = ACTIONS(5074),
    [anon_sym_inout] = ACTIONS(5074),
    [anon_sym_bycopy] = ACTIONS(5074),
    [anon_sym_byref] = ACTIONS(5074),
    [anon_sym_oneway] = ACTIONS(5074),
  },
  [STATE(3791)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3812),
    [aux_sym_generic_specifier_repeat1] = STATE(3813),
    [sym_identifier] = ACTIONS(6138),
    [anon_sym_COMMA] = ACTIONS(6143),
    [anon_sym_RPAREN] = ACTIONS(6143),
    [anon_sym_LPAREN2] = ACTIONS(6130),
    [anon_sym_STAR] = ACTIONS(6143),
    [anon_sym_CARET] = ACTIONS(6143),
    [anon_sym_GT] = ACTIONS(6143),
    [anon_sym_LT] = ACTIONS(8777),
    [anon_sym___extension__] = ACTIONS(6138),
    [anon_sym___attribute__] = ACTIONS(6138),
    [anon_sym___attribute] = ACTIONS(6138),
    [anon_sym_noreturn] = ACTIONS(6138),
    [anon_sym_LBRACK] = ACTIONS(6143),
    [anon_sym___based] = ACTIONS(6138),
    [anon_sym_LBRACE] = ACTIONS(6143),
    [anon_sym_signed] = ACTIONS(8781),
    [anon_sym_unsigned] = ACTIONS(8781),
    [anon_sym_long] = ACTIONS(8781),
    [anon_sym_short] = ACTIONS(8781),
    [anon_sym_ATautoreleasepool] = ACTIONS(6143),
    [anon_sym_const] = ACTIONS(6138),
    [anon_sym_constexpr] = ACTIONS(6138),
    [anon_sym_volatile] = ACTIONS(6138),
    [anon_sym_restrict] = ACTIONS(6138),
    [anon_sym___restrict__] = ACTIONS(6138),
    [anon_sym__Atomic] = ACTIONS(6138),
    [anon_sym__Noreturn] = ACTIONS(6138),
    [anon_sym__Nonnull] = ACTIONS(6138),
    [anon_sym_nullable] = ACTIONS(6138),
    [anon_sym__Complex] = ACTIONS(6138),
    [anon_sym__Nullable] = ACTIONS(6138),
    [anon_sym__Nullable_result] = ACTIONS(6138),
    [anon_sym__Null_unspecified] = ACTIONS(6138),
    [anon_sym___autoreleasing] = ACTIONS(6138),
    [anon_sym___block] = ACTIONS(6138),
    [anon_sym___bridge] = ACTIONS(6138),
    [anon_sym___bridge_retained] = ACTIONS(6138),
    [anon_sym___bridge_transfer] = ACTIONS(6138),
    [anon_sym___complex] = ACTIONS(6138),
    [anon_sym___const] = ACTIONS(6138),
    [anon_sym___imag] = ACTIONS(6138),
    [anon_sym___kindof] = ACTIONS(6138),
    [anon_sym___nonnull] = ACTIONS(6138),
    [anon_sym___nullable] = ACTIONS(6138),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6138),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6138),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6138),
    [anon_sym___real] = ACTIONS(6138),
    [anon_sym___strong] = ACTIONS(6138),
    [anon_sym___unsafe_unretained] = ACTIONS(6138),
    [anon_sym___unused] = ACTIONS(6138),
    [anon_sym___weak] = ACTIONS(6138),
    [anon_sym_alignas] = ACTIONS(6138),
    [anon_sym__Alignas] = ACTIONS(6138),
    [sym_primitive_type] = ACTIONS(6138),
    [anon_sym_enum] = ACTIONS(6138),
    [anon_sym_struct] = ACTIONS(6138),
    [anon_sym_union] = ACTIONS(6138),
    [anon_sym_in] = ACTIONS(6138),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6138),
    [anon_sym___typeof] = ACTIONS(6138),
    [anon_sym_typeof] = ACTIONS(6138),
    [anon_sym_BOOL] = ACTIONS(6138),
    [anon_sym_IMP] = ACTIONS(6138),
    [anon_sym_SEL] = ACTIONS(6138),
    [anon_sym_Class] = ACTIONS(6138),
    [anon_sym_id] = ACTIONS(6138),
    [anon_sym_out] = ACTIONS(6138),
    [anon_sym_inout] = ACTIONS(6138),
    [anon_sym_bycopy] = ACTIONS(6138),
    [anon_sym_byref] = ACTIONS(6138),
    [anon_sym_oneway] = ACTIONS(6138),
  },
  [STATE(3792)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3792),
    [sym_identifier] = ACTIONS(7080),
    [anon_sym_COMMA] = ACTIONS(7082),
    [anon_sym_RPAREN] = ACTIONS(7082),
    [anon_sym_LPAREN2] = ACTIONS(7082),
    [anon_sym_STAR] = ACTIONS(7082),
    [anon_sym_CARET] = ACTIONS(7082),
    [anon_sym_GT] = ACTIONS(7082),
    [anon_sym_LT] = ACTIONS(7082),
    [anon_sym___extension__] = ACTIONS(7080),
    [anon_sym___attribute__] = ACTIONS(7080),
    [anon_sym___attribute] = ACTIONS(7080),
    [anon_sym_noreturn] = ACTIONS(7080),
    [anon_sym_LBRACK] = ACTIONS(7082),
    [anon_sym___based] = ACTIONS(7080),
    [anon_sym_LBRACE] = ACTIONS(7082),
    [anon_sym_signed] = ACTIONS(8785),
    [anon_sym_unsigned] = ACTIONS(8785),
    [anon_sym_long] = ACTIONS(8785),
    [anon_sym_short] = ACTIONS(8785),
    [anon_sym_ATautoreleasepool] = ACTIONS(7082),
    [anon_sym_const] = ACTIONS(7080),
    [anon_sym_constexpr] = ACTIONS(7080),
    [anon_sym_volatile] = ACTIONS(7080),
    [anon_sym_restrict] = ACTIONS(7080),
    [anon_sym___restrict__] = ACTIONS(7080),
    [anon_sym__Atomic] = ACTIONS(7080),
    [anon_sym__Noreturn] = ACTIONS(7080),
    [anon_sym__Nonnull] = ACTIONS(7080),
    [anon_sym_nullable] = ACTIONS(7080),
    [anon_sym__Complex] = ACTIONS(7080),
    [anon_sym__Nullable] = ACTIONS(7080),
    [anon_sym__Nullable_result] = ACTIONS(7080),
    [anon_sym__Null_unspecified] = ACTIONS(7080),
    [anon_sym___autoreleasing] = ACTIONS(7080),
    [anon_sym___block] = ACTIONS(7080),
    [anon_sym___bridge] = ACTIONS(7080),
    [anon_sym___bridge_retained] = ACTIONS(7080),
    [anon_sym___bridge_transfer] = ACTIONS(7080),
    [anon_sym___complex] = ACTIONS(7080),
    [anon_sym___const] = ACTIONS(7080),
    [anon_sym___imag] = ACTIONS(7080),
    [anon_sym___kindof] = ACTIONS(7080),
    [anon_sym___nonnull] = ACTIONS(7080),
    [anon_sym___nullable] = ACTIONS(7080),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7080),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7080),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7080),
    [anon_sym___real] = ACTIONS(7080),
    [anon_sym___strong] = ACTIONS(7080),
    [anon_sym___unsafe_unretained] = ACTIONS(7080),
    [anon_sym___unused] = ACTIONS(7080),
    [anon_sym___weak] = ACTIONS(7080),
    [anon_sym_alignas] = ACTIONS(7080),
    [anon_sym__Alignas] = ACTIONS(7080),
    [sym_primitive_type] = ACTIONS(7080),
    [anon_sym_enum] = ACTIONS(7080),
    [anon_sym_COLON] = ACTIONS(7082),
    [anon_sym_struct] = ACTIONS(7080),
    [anon_sym_union] = ACTIONS(7080),
    [anon_sym_in] = ACTIONS(7080),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7080),
    [anon_sym___typeof] = ACTIONS(7080),
    [anon_sym_typeof] = ACTIONS(7080),
    [anon_sym_BOOL] = ACTIONS(7080),
    [anon_sym_IMP] = ACTIONS(7080),
    [anon_sym_SEL] = ACTIONS(7080),
    [anon_sym_Class] = ACTIONS(7080),
    [anon_sym_id] = ACTIONS(7080),
    [anon_sym_out] = ACTIONS(7080),
    [anon_sym_inout] = ACTIONS(7080),
    [anon_sym_bycopy] = ACTIONS(7080),
    [anon_sym_byref] = ACTIONS(7080),
    [anon_sym_oneway] = ACTIONS(7080),
  },
  [STATE(3793)] = {
    [sym_attribute_specifier] = STATE(3867),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_RPAREN] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7107),
    [anon_sym_STAR] = ACTIONS(7107),
    [anon_sym_CARET] = ACTIONS(7107),
    [anon_sym_GT] = ACTIONS(7107),
    [anon_sym_LT] = ACTIONS(7107),
    [anon_sym___extension__] = ACTIONS(7105),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7107),
    [anon_sym___based] = ACTIONS(7105),
    [anon_sym_LBRACE] = ACTIONS(7107),
    [anon_sym_signed] = ACTIONS(7105),
    [anon_sym_unsigned] = ACTIONS(7105),
    [anon_sym_long] = ACTIONS(7105),
    [anon_sym_short] = ACTIONS(7105),
    [anon_sym_ATautoreleasepool] = ACTIONS(7107),
    [anon_sym_const] = ACTIONS(7105),
    [anon_sym_constexpr] = ACTIONS(7105),
    [anon_sym_volatile] = ACTIONS(7105),
    [anon_sym_restrict] = ACTIONS(7105),
    [anon_sym___restrict__] = ACTIONS(7105),
    [anon_sym__Atomic] = ACTIONS(7105),
    [anon_sym__Noreturn] = ACTIONS(7105),
    [anon_sym__Nonnull] = ACTIONS(7105),
    [anon_sym_nullable] = ACTIONS(7105),
    [anon_sym__Complex] = ACTIONS(7105),
    [anon_sym__Nullable] = ACTIONS(7105),
    [anon_sym__Nullable_result] = ACTIONS(7105),
    [anon_sym__Null_unspecified] = ACTIONS(7105),
    [anon_sym___autoreleasing] = ACTIONS(7105),
    [anon_sym___block] = ACTIONS(7105),
    [anon_sym___bridge] = ACTIONS(7105),
    [anon_sym___bridge_retained] = ACTIONS(7105),
    [anon_sym___bridge_transfer] = ACTIONS(7105),
    [anon_sym___complex] = ACTIONS(7105),
    [anon_sym___const] = ACTIONS(7105),
    [anon_sym___imag] = ACTIONS(7105),
    [anon_sym___kindof] = ACTIONS(7105),
    [anon_sym___nonnull] = ACTIONS(7105),
    [anon_sym___nullable] = ACTIONS(7105),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7105),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7105),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7105),
    [anon_sym___real] = ACTIONS(7105),
    [anon_sym___strong] = ACTIONS(7105),
    [anon_sym___unsafe_unretained] = ACTIONS(7105),
    [anon_sym___unused] = ACTIONS(7105),
    [anon_sym___weak] = ACTIONS(7105),
    [anon_sym_alignas] = ACTIONS(7105),
    [anon_sym__Alignas] = ACTIONS(7105),
    [sym_primitive_type] = ACTIONS(7105),
    [anon_sym_enum] = ACTIONS(7105),
    [anon_sym_COLON] = ACTIONS(7107),
    [anon_sym_struct] = ACTIONS(7105),
    [anon_sym_union] = ACTIONS(7105),
    [anon_sym_in] = ACTIONS(7105),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7105),
    [anon_sym___typeof] = ACTIONS(7105),
    [anon_sym_typeof] = ACTIONS(7105),
    [anon_sym_BOOL] = ACTIONS(7105),
    [anon_sym_IMP] = ACTIONS(7105),
    [anon_sym_SEL] = ACTIONS(7105),
    [anon_sym_Class] = ACTIONS(7105),
    [anon_sym_id] = ACTIONS(7105),
    [anon_sym_out] = ACTIONS(7105),
    [anon_sym_inout] = ACTIONS(7105),
    [anon_sym_bycopy] = ACTIONS(7105),
    [anon_sym_byref] = ACTIONS(7105),
    [anon_sym_oneway] = ACTIONS(7105),
  },
  [STATE(3794)] = {
    [sym_attribute_specifier] = STATE(3970),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_RPAREN] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7143),
    [anon_sym_STAR] = ACTIONS(7143),
    [anon_sym_CARET] = ACTIONS(7143),
    [anon_sym_GT] = ACTIONS(7143),
    [anon_sym_LT] = ACTIONS(7143),
    [anon_sym___extension__] = ACTIONS(7141),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7143),
    [anon_sym___based] = ACTIONS(7141),
    [anon_sym_LBRACE] = ACTIONS(7143),
    [anon_sym_signed] = ACTIONS(7141),
    [anon_sym_unsigned] = ACTIONS(7141),
    [anon_sym_long] = ACTIONS(7141),
    [anon_sym_short] = ACTIONS(7141),
    [anon_sym_ATautoreleasepool] = ACTIONS(7143),
    [anon_sym_const] = ACTIONS(7141),
    [anon_sym_constexpr] = ACTIONS(7141),
    [anon_sym_volatile] = ACTIONS(7141),
    [anon_sym_restrict] = ACTIONS(7141),
    [anon_sym___restrict__] = ACTIONS(7141),
    [anon_sym__Atomic] = ACTIONS(7141),
    [anon_sym__Noreturn] = ACTIONS(7141),
    [anon_sym__Nonnull] = ACTIONS(7141),
    [anon_sym_nullable] = ACTIONS(7141),
    [anon_sym__Complex] = ACTIONS(7141),
    [anon_sym__Nullable] = ACTIONS(7141),
    [anon_sym__Nullable_result] = ACTIONS(7141),
    [anon_sym__Null_unspecified] = ACTIONS(7141),
    [anon_sym___autoreleasing] = ACTIONS(7141),
    [anon_sym___block] = ACTIONS(7141),
    [anon_sym___bridge] = ACTIONS(7141),
    [anon_sym___bridge_retained] = ACTIONS(7141),
    [anon_sym___bridge_transfer] = ACTIONS(7141),
    [anon_sym___complex] = ACTIONS(7141),
    [anon_sym___const] = ACTIONS(7141),
    [anon_sym___imag] = ACTIONS(7141),
    [anon_sym___kindof] = ACTIONS(7141),
    [anon_sym___nonnull] = ACTIONS(7141),
    [anon_sym___nullable] = ACTIONS(7141),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7141),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7141),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7141),
    [anon_sym___real] = ACTIONS(7141),
    [anon_sym___strong] = ACTIONS(7141),
    [anon_sym___unsafe_unretained] = ACTIONS(7141),
    [anon_sym___unused] = ACTIONS(7141),
    [anon_sym___weak] = ACTIONS(7141),
    [anon_sym_alignas] = ACTIONS(7141),
    [anon_sym__Alignas] = ACTIONS(7141),
    [sym_primitive_type] = ACTIONS(7141),
    [anon_sym_enum] = ACTIONS(7141),
    [anon_sym_COLON] = ACTIONS(7143),
    [anon_sym_struct] = ACTIONS(7141),
    [anon_sym_union] = ACTIONS(7141),
    [anon_sym_in] = ACTIONS(7141),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7141),
    [anon_sym___typeof] = ACTIONS(7141),
    [anon_sym_typeof] = ACTIONS(7141),
    [anon_sym_BOOL] = ACTIONS(7141),
    [anon_sym_IMP] = ACTIONS(7141),
    [anon_sym_SEL] = ACTIONS(7141),
    [anon_sym_Class] = ACTIONS(7141),
    [anon_sym_id] = ACTIONS(7141),
    [anon_sym_out] = ACTIONS(7141),
    [anon_sym_inout] = ACTIONS(7141),
    [anon_sym_bycopy] = ACTIONS(7141),
    [anon_sym_byref] = ACTIONS(7141),
    [anon_sym_oneway] = ACTIONS(7141),
  },
  [STATE(3795)] = {
    [sym_attribute_specifier] = STATE(3854),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_RPAREN] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7123),
    [anon_sym_STAR] = ACTIONS(7123),
    [anon_sym_CARET] = ACTIONS(7123),
    [anon_sym_GT] = ACTIONS(7123),
    [anon_sym_LT] = ACTIONS(7123),
    [anon_sym___extension__] = ACTIONS(7121),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7123),
    [anon_sym___based] = ACTIONS(7121),
    [anon_sym_LBRACE] = ACTIONS(7123),
    [anon_sym_signed] = ACTIONS(7121),
    [anon_sym_unsigned] = ACTIONS(7121),
    [anon_sym_long] = ACTIONS(7121),
    [anon_sym_short] = ACTIONS(7121),
    [anon_sym_ATautoreleasepool] = ACTIONS(7123),
    [anon_sym_const] = ACTIONS(7121),
    [anon_sym_constexpr] = ACTIONS(7121),
    [anon_sym_volatile] = ACTIONS(7121),
    [anon_sym_restrict] = ACTIONS(7121),
    [anon_sym___restrict__] = ACTIONS(7121),
    [anon_sym__Atomic] = ACTIONS(7121),
    [anon_sym__Noreturn] = ACTIONS(7121),
    [anon_sym__Nonnull] = ACTIONS(7121),
    [anon_sym_nullable] = ACTIONS(7121),
    [anon_sym__Complex] = ACTIONS(7121),
    [anon_sym__Nullable] = ACTIONS(7121),
    [anon_sym__Nullable_result] = ACTIONS(7121),
    [anon_sym__Null_unspecified] = ACTIONS(7121),
    [anon_sym___autoreleasing] = ACTIONS(7121),
    [anon_sym___block] = ACTIONS(7121),
    [anon_sym___bridge] = ACTIONS(7121),
    [anon_sym___bridge_retained] = ACTIONS(7121),
    [anon_sym___bridge_transfer] = ACTIONS(7121),
    [anon_sym___complex] = ACTIONS(7121),
    [anon_sym___const] = ACTIONS(7121),
    [anon_sym___imag] = ACTIONS(7121),
    [anon_sym___kindof] = ACTIONS(7121),
    [anon_sym___nonnull] = ACTIONS(7121),
    [anon_sym___nullable] = ACTIONS(7121),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7121),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7121),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7121),
    [anon_sym___real] = ACTIONS(7121),
    [anon_sym___strong] = ACTIONS(7121),
    [anon_sym___unsafe_unretained] = ACTIONS(7121),
    [anon_sym___unused] = ACTIONS(7121),
    [anon_sym___weak] = ACTIONS(7121),
    [anon_sym_alignas] = ACTIONS(7121),
    [anon_sym__Alignas] = ACTIONS(7121),
    [sym_primitive_type] = ACTIONS(7121),
    [anon_sym_enum] = ACTIONS(7121),
    [anon_sym_COLON] = ACTIONS(7123),
    [anon_sym_struct] = ACTIONS(7121),
    [anon_sym_union] = ACTIONS(7121),
    [anon_sym_in] = ACTIONS(7121),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7121),
    [anon_sym___typeof] = ACTIONS(7121),
    [anon_sym_typeof] = ACTIONS(7121),
    [anon_sym_BOOL] = ACTIONS(7121),
    [anon_sym_IMP] = ACTIONS(7121),
    [anon_sym_SEL] = ACTIONS(7121),
    [anon_sym_Class] = ACTIONS(7121),
    [anon_sym_id] = ACTIONS(7121),
    [anon_sym_out] = ACTIONS(7121),
    [anon_sym_inout] = ACTIONS(7121),
    [anon_sym_bycopy] = ACTIONS(7121),
    [anon_sym_byref] = ACTIONS(7121),
    [anon_sym_oneway] = ACTIONS(7121),
  },
  [STATE(3796)] = {
    [sym_attribute_specifier] = STATE(3855),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_RPAREN] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7139),
    [anon_sym_STAR] = ACTIONS(7139),
    [anon_sym_CARET] = ACTIONS(7139),
    [anon_sym_GT] = ACTIONS(7139),
    [anon_sym_LT] = ACTIONS(7139),
    [anon_sym___extension__] = ACTIONS(7137),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7139),
    [anon_sym___based] = ACTIONS(7137),
    [anon_sym_LBRACE] = ACTIONS(7139),
    [anon_sym_signed] = ACTIONS(7137),
    [anon_sym_unsigned] = ACTIONS(7137),
    [anon_sym_long] = ACTIONS(7137),
    [anon_sym_short] = ACTIONS(7137),
    [anon_sym_ATautoreleasepool] = ACTIONS(7139),
    [anon_sym_const] = ACTIONS(7137),
    [anon_sym_constexpr] = ACTIONS(7137),
    [anon_sym_volatile] = ACTIONS(7137),
    [anon_sym_restrict] = ACTIONS(7137),
    [anon_sym___restrict__] = ACTIONS(7137),
    [anon_sym__Atomic] = ACTIONS(7137),
    [anon_sym__Noreturn] = ACTIONS(7137),
    [anon_sym__Nonnull] = ACTIONS(7137),
    [anon_sym_nullable] = ACTIONS(7137),
    [anon_sym__Complex] = ACTIONS(7137),
    [anon_sym__Nullable] = ACTIONS(7137),
    [anon_sym__Nullable_result] = ACTIONS(7137),
    [anon_sym__Null_unspecified] = ACTIONS(7137),
    [anon_sym___autoreleasing] = ACTIONS(7137),
    [anon_sym___block] = ACTIONS(7137),
    [anon_sym___bridge] = ACTIONS(7137),
    [anon_sym___bridge_retained] = ACTIONS(7137),
    [anon_sym___bridge_transfer] = ACTIONS(7137),
    [anon_sym___complex] = ACTIONS(7137),
    [anon_sym___const] = ACTIONS(7137),
    [anon_sym___imag] = ACTIONS(7137),
    [anon_sym___kindof] = ACTIONS(7137),
    [anon_sym___nonnull] = ACTIONS(7137),
    [anon_sym___nullable] = ACTIONS(7137),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7137),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7137),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7137),
    [anon_sym___real] = ACTIONS(7137),
    [anon_sym___strong] = ACTIONS(7137),
    [anon_sym___unsafe_unretained] = ACTIONS(7137),
    [anon_sym___unused] = ACTIONS(7137),
    [anon_sym___weak] = ACTIONS(7137),
    [anon_sym_alignas] = ACTIONS(7137),
    [anon_sym__Alignas] = ACTIONS(7137),
    [sym_primitive_type] = ACTIONS(7137),
    [anon_sym_enum] = ACTIONS(7137),
    [anon_sym_COLON] = ACTIONS(7139),
    [anon_sym_struct] = ACTIONS(7137),
    [anon_sym_union] = ACTIONS(7137),
    [anon_sym_in] = ACTIONS(7137),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7137),
    [anon_sym___typeof] = ACTIONS(7137),
    [anon_sym_typeof] = ACTIONS(7137),
    [anon_sym_BOOL] = ACTIONS(7137),
    [anon_sym_IMP] = ACTIONS(7137),
    [anon_sym_SEL] = ACTIONS(7137),
    [anon_sym_Class] = ACTIONS(7137),
    [anon_sym_id] = ACTIONS(7137),
    [anon_sym_out] = ACTIONS(7137),
    [anon_sym_inout] = ACTIONS(7137),
    [anon_sym_bycopy] = ACTIONS(7137),
    [anon_sym_byref] = ACTIONS(7137),
    [anon_sym_oneway] = ACTIONS(7137),
  },
  [STATE(3797)] = {
    [sym_attribute_specifier] = STATE(3969),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_RPAREN] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7119),
    [anon_sym_STAR] = ACTIONS(7119),
    [anon_sym_CARET] = ACTIONS(7119),
    [anon_sym_GT] = ACTIONS(7119),
    [anon_sym_LT] = ACTIONS(7119),
    [anon_sym___extension__] = ACTIONS(7117),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7119),
    [anon_sym___based] = ACTIONS(7117),
    [anon_sym_LBRACE] = ACTIONS(7119),
    [anon_sym_signed] = ACTIONS(7117),
    [anon_sym_unsigned] = ACTIONS(7117),
    [anon_sym_long] = ACTIONS(7117),
    [anon_sym_short] = ACTIONS(7117),
    [anon_sym_ATautoreleasepool] = ACTIONS(7119),
    [anon_sym_const] = ACTIONS(7117),
    [anon_sym_constexpr] = ACTIONS(7117),
    [anon_sym_volatile] = ACTIONS(7117),
    [anon_sym_restrict] = ACTIONS(7117),
    [anon_sym___restrict__] = ACTIONS(7117),
    [anon_sym__Atomic] = ACTIONS(7117),
    [anon_sym__Noreturn] = ACTIONS(7117),
    [anon_sym__Nonnull] = ACTIONS(7117),
    [anon_sym_nullable] = ACTIONS(7117),
    [anon_sym__Complex] = ACTIONS(7117),
    [anon_sym__Nullable] = ACTIONS(7117),
    [anon_sym__Nullable_result] = ACTIONS(7117),
    [anon_sym__Null_unspecified] = ACTIONS(7117),
    [anon_sym___autoreleasing] = ACTIONS(7117),
    [anon_sym___block] = ACTIONS(7117),
    [anon_sym___bridge] = ACTIONS(7117),
    [anon_sym___bridge_retained] = ACTIONS(7117),
    [anon_sym___bridge_transfer] = ACTIONS(7117),
    [anon_sym___complex] = ACTIONS(7117),
    [anon_sym___const] = ACTIONS(7117),
    [anon_sym___imag] = ACTIONS(7117),
    [anon_sym___kindof] = ACTIONS(7117),
    [anon_sym___nonnull] = ACTIONS(7117),
    [anon_sym___nullable] = ACTIONS(7117),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7117),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7117),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7117),
    [anon_sym___real] = ACTIONS(7117),
    [anon_sym___strong] = ACTIONS(7117),
    [anon_sym___unsafe_unretained] = ACTIONS(7117),
    [anon_sym___unused] = ACTIONS(7117),
    [anon_sym___weak] = ACTIONS(7117),
    [anon_sym_alignas] = ACTIONS(7117),
    [anon_sym__Alignas] = ACTIONS(7117),
    [sym_primitive_type] = ACTIONS(7117),
    [anon_sym_enum] = ACTIONS(7117),
    [anon_sym_COLON] = ACTIONS(7119),
    [anon_sym_struct] = ACTIONS(7117),
    [anon_sym_union] = ACTIONS(7117),
    [anon_sym_in] = ACTIONS(7117),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7117),
    [anon_sym___typeof] = ACTIONS(7117),
    [anon_sym_typeof] = ACTIONS(7117),
    [anon_sym_BOOL] = ACTIONS(7117),
    [anon_sym_IMP] = ACTIONS(7117),
    [anon_sym_SEL] = ACTIONS(7117),
    [anon_sym_Class] = ACTIONS(7117),
    [anon_sym_id] = ACTIONS(7117),
    [anon_sym_out] = ACTIONS(7117),
    [anon_sym_inout] = ACTIONS(7117),
    [anon_sym_bycopy] = ACTIONS(7117),
    [anon_sym_byref] = ACTIONS(7117),
    [anon_sym_oneway] = ACTIONS(7117),
  },
  [STATE(3798)] = {
    [sym_attribute_specifier] = STATE(3862),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_RPAREN] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7131),
    [anon_sym_STAR] = ACTIONS(7131),
    [anon_sym_CARET] = ACTIONS(7131),
    [anon_sym_GT] = ACTIONS(7131),
    [anon_sym_LT] = ACTIONS(7131),
    [anon_sym___extension__] = ACTIONS(7129),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7131),
    [anon_sym___based] = ACTIONS(7129),
    [anon_sym_LBRACE] = ACTIONS(7131),
    [anon_sym_signed] = ACTIONS(7129),
    [anon_sym_unsigned] = ACTIONS(7129),
    [anon_sym_long] = ACTIONS(7129),
    [anon_sym_short] = ACTIONS(7129),
    [anon_sym_ATautoreleasepool] = ACTIONS(7131),
    [anon_sym_const] = ACTIONS(7129),
    [anon_sym_constexpr] = ACTIONS(7129),
    [anon_sym_volatile] = ACTIONS(7129),
    [anon_sym_restrict] = ACTIONS(7129),
    [anon_sym___restrict__] = ACTIONS(7129),
    [anon_sym__Atomic] = ACTIONS(7129),
    [anon_sym__Noreturn] = ACTIONS(7129),
    [anon_sym__Nonnull] = ACTIONS(7129),
    [anon_sym_nullable] = ACTIONS(7129),
    [anon_sym__Complex] = ACTIONS(7129),
    [anon_sym__Nullable] = ACTIONS(7129),
    [anon_sym__Nullable_result] = ACTIONS(7129),
    [anon_sym__Null_unspecified] = ACTIONS(7129),
    [anon_sym___autoreleasing] = ACTIONS(7129),
    [anon_sym___block] = ACTIONS(7129),
    [anon_sym___bridge] = ACTIONS(7129),
    [anon_sym___bridge_retained] = ACTIONS(7129),
    [anon_sym___bridge_transfer] = ACTIONS(7129),
    [anon_sym___complex] = ACTIONS(7129),
    [anon_sym___const] = ACTIONS(7129),
    [anon_sym___imag] = ACTIONS(7129),
    [anon_sym___kindof] = ACTIONS(7129),
    [anon_sym___nonnull] = ACTIONS(7129),
    [anon_sym___nullable] = ACTIONS(7129),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7129),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7129),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7129),
    [anon_sym___real] = ACTIONS(7129),
    [anon_sym___strong] = ACTIONS(7129),
    [anon_sym___unsafe_unretained] = ACTIONS(7129),
    [anon_sym___unused] = ACTIONS(7129),
    [anon_sym___weak] = ACTIONS(7129),
    [anon_sym_alignas] = ACTIONS(7129),
    [anon_sym__Alignas] = ACTIONS(7129),
    [sym_primitive_type] = ACTIONS(7129),
    [anon_sym_enum] = ACTIONS(7129),
    [anon_sym_COLON] = ACTIONS(7131),
    [anon_sym_struct] = ACTIONS(7129),
    [anon_sym_union] = ACTIONS(7129),
    [anon_sym_in] = ACTIONS(7129),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7129),
    [anon_sym___typeof] = ACTIONS(7129),
    [anon_sym_typeof] = ACTIONS(7129),
    [anon_sym_BOOL] = ACTIONS(7129),
    [anon_sym_IMP] = ACTIONS(7129),
    [anon_sym_SEL] = ACTIONS(7129),
    [anon_sym_Class] = ACTIONS(7129),
    [anon_sym_id] = ACTIONS(7129),
    [anon_sym_out] = ACTIONS(7129),
    [anon_sym_inout] = ACTIONS(7129),
    [anon_sym_bycopy] = ACTIONS(7129),
    [anon_sym_byref] = ACTIONS(7129),
    [anon_sym_oneway] = ACTIONS(7129),
  },
  [STATE(3799)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3792),
    [sym_identifier] = ACTIONS(7435),
    [anon_sym_COMMA] = ACTIONS(7437),
    [anon_sym_RPAREN] = ACTIONS(7437),
    [anon_sym_LPAREN2] = ACTIONS(7437),
    [anon_sym_STAR] = ACTIONS(7437),
    [anon_sym_CARET] = ACTIONS(7437),
    [anon_sym_GT] = ACTIONS(7437),
    [anon_sym_LT] = ACTIONS(7437),
    [anon_sym___extension__] = ACTIONS(7435),
    [anon_sym___attribute__] = ACTIONS(7435),
    [anon_sym___attribute] = ACTIONS(7435),
    [anon_sym_noreturn] = ACTIONS(7435),
    [anon_sym_LBRACK] = ACTIONS(7437),
    [anon_sym___based] = ACTIONS(7435),
    [anon_sym_LBRACE] = ACTIONS(7437),
    [anon_sym_signed] = ACTIONS(8788),
    [anon_sym_unsigned] = ACTIONS(8788),
    [anon_sym_long] = ACTIONS(8788),
    [anon_sym_short] = ACTIONS(8788),
    [anon_sym_ATautoreleasepool] = ACTIONS(7437),
    [anon_sym_const] = ACTIONS(7435),
    [anon_sym_constexpr] = ACTIONS(7435),
    [anon_sym_volatile] = ACTIONS(7435),
    [anon_sym_restrict] = ACTIONS(7435),
    [anon_sym___restrict__] = ACTIONS(7435),
    [anon_sym__Atomic] = ACTIONS(7435),
    [anon_sym__Noreturn] = ACTIONS(7435),
    [anon_sym__Nonnull] = ACTIONS(7435),
    [anon_sym_nullable] = ACTIONS(7435),
    [anon_sym__Complex] = ACTIONS(7435),
    [anon_sym__Nullable] = ACTIONS(7435),
    [anon_sym__Nullable_result] = ACTIONS(7435),
    [anon_sym__Null_unspecified] = ACTIONS(7435),
    [anon_sym___autoreleasing] = ACTIONS(7435),
    [anon_sym___block] = ACTIONS(7435),
    [anon_sym___bridge] = ACTIONS(7435),
    [anon_sym___bridge_retained] = ACTIONS(7435),
    [anon_sym___bridge_transfer] = ACTIONS(7435),
    [anon_sym___complex] = ACTIONS(7435),
    [anon_sym___const] = ACTIONS(7435),
    [anon_sym___imag] = ACTIONS(7435),
    [anon_sym___kindof] = ACTIONS(7435),
    [anon_sym___nonnull] = ACTIONS(7435),
    [anon_sym___nullable] = ACTIONS(7435),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7435),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7435),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7435),
    [anon_sym___real] = ACTIONS(7435),
    [anon_sym___strong] = ACTIONS(7435),
    [anon_sym___unsafe_unretained] = ACTIONS(7435),
    [anon_sym___unused] = ACTIONS(7435),
    [anon_sym___weak] = ACTIONS(7435),
    [anon_sym_alignas] = ACTIONS(7435),
    [anon_sym__Alignas] = ACTIONS(7435),
    [sym_primitive_type] = ACTIONS(7435),
    [anon_sym_enum] = ACTIONS(7435),
    [anon_sym_COLON] = ACTIONS(7437),
    [anon_sym_struct] = ACTIONS(7435),
    [anon_sym_union] = ACTIONS(7435),
    [anon_sym_in] = ACTIONS(7435),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7435),
    [anon_sym___typeof] = ACTIONS(7435),
    [anon_sym_typeof] = ACTIONS(7435),
    [anon_sym_BOOL] = ACTIONS(7435),
    [anon_sym_IMP] = ACTIONS(7435),
    [anon_sym_SEL] = ACTIONS(7435),
    [anon_sym_Class] = ACTIONS(7435),
    [anon_sym_id] = ACTIONS(7435),
    [anon_sym_out] = ACTIONS(7435),
    [anon_sym_inout] = ACTIONS(7435),
    [anon_sym_bycopy] = ACTIONS(7435),
    [anon_sym_byref] = ACTIONS(7435),
    [anon_sym_oneway] = ACTIONS(7435),
  },
  [STATE(3800)] = {
    [sym_attribute_specifier] = STATE(3839),
    [sym_enumerator_list] = STATE(3961),
    [sym_identifier] = ACTIONS(7619),
    [anon_sym_COMMA] = ACTIONS(7621),
    [anon_sym_RPAREN] = ACTIONS(7621),
    [anon_sym_LPAREN2] = ACTIONS(7621),
    [anon_sym_STAR] = ACTIONS(7621),
    [anon_sym_CARET] = ACTIONS(7621),
    [anon_sym_GT] = ACTIONS(7621),
    [anon_sym_LT] = ACTIONS(7621),
    [anon_sym___extension__] = ACTIONS(7619),
    [anon_sym___attribute__] = ACTIONS(8791),
    [anon_sym___attribute] = ACTIONS(8791),
    [anon_sym_noreturn] = ACTIONS(7619),
    [anon_sym_LBRACK] = ACTIONS(7621),
    [anon_sym___based] = ACTIONS(7619),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7619),
    [anon_sym_unsigned] = ACTIONS(7619),
    [anon_sym_long] = ACTIONS(7619),
    [anon_sym_short] = ACTIONS(7619),
    [anon_sym_const] = ACTIONS(7619),
    [anon_sym_constexpr] = ACTIONS(7619),
    [anon_sym_volatile] = ACTIONS(7619),
    [anon_sym_restrict] = ACTIONS(7619),
    [anon_sym___restrict__] = ACTIONS(7619),
    [anon_sym__Atomic] = ACTIONS(7619),
    [anon_sym__Noreturn] = ACTIONS(7619),
    [anon_sym__Nonnull] = ACTIONS(7619),
    [anon_sym_nullable] = ACTIONS(7619),
    [anon_sym__Complex] = ACTIONS(7619),
    [anon_sym__Nullable] = ACTIONS(7619),
    [anon_sym__Nullable_result] = ACTIONS(7619),
    [anon_sym__Null_unspecified] = ACTIONS(7619),
    [anon_sym___autoreleasing] = ACTIONS(7619),
    [anon_sym___block] = ACTIONS(7619),
    [anon_sym___bridge] = ACTIONS(7619),
    [anon_sym___bridge_retained] = ACTIONS(7619),
    [anon_sym___bridge_transfer] = ACTIONS(7619),
    [anon_sym___complex] = ACTIONS(7619),
    [anon_sym___const] = ACTIONS(7619),
    [anon_sym___imag] = ACTIONS(7619),
    [anon_sym___kindof] = ACTIONS(7619),
    [anon_sym___nonnull] = ACTIONS(7619),
    [anon_sym___nullable] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7619),
    [anon_sym___real] = ACTIONS(7619),
    [anon_sym___strong] = ACTIONS(7619),
    [anon_sym___unsafe_unretained] = ACTIONS(7619),
    [anon_sym___unused] = ACTIONS(7619),
    [anon_sym___weak] = ACTIONS(7619),
    [anon_sym_alignas] = ACTIONS(7619),
    [anon_sym__Alignas] = ACTIONS(7619),
    [sym_primitive_type] = ACTIONS(7619),
    [anon_sym_enum] = ACTIONS(7619),
    [anon_sym_COLON] = ACTIONS(8796),
    [anon_sym_struct] = ACTIONS(7619),
    [anon_sym_union] = ACTIONS(7619),
    [anon_sym_in] = ACTIONS(7619),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7619),
    [anon_sym___typeof] = ACTIONS(7619),
    [anon_sym_typeof] = ACTIONS(7619),
    [anon_sym_BOOL] = ACTIONS(7619),
    [anon_sym_IMP] = ACTIONS(7619),
    [anon_sym_SEL] = ACTIONS(7619),
    [anon_sym_Class] = ACTIONS(7619),
    [anon_sym_id] = ACTIONS(7619),
    [anon_sym_out] = ACTIONS(7619),
    [anon_sym_inout] = ACTIONS(7619),
    [anon_sym_bycopy] = ACTIONS(7619),
    [anon_sym_byref] = ACTIONS(7619),
    [anon_sym_oneway] = ACTIONS(7619),
  },
  [STATE(3801)] = {
    [sym_attribute_specifier] = STATE(3846),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_RPAREN] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7111),
    [anon_sym_STAR] = ACTIONS(7111),
    [anon_sym_CARET] = ACTIONS(7111),
    [anon_sym_GT] = ACTIONS(7111),
    [anon_sym_LT] = ACTIONS(7111),
    [anon_sym___extension__] = ACTIONS(7109),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7111),
    [anon_sym___based] = ACTIONS(7109),
    [anon_sym_LBRACE] = ACTIONS(7111),
    [anon_sym_signed] = ACTIONS(7109),
    [anon_sym_unsigned] = ACTIONS(7109),
    [anon_sym_long] = ACTIONS(7109),
    [anon_sym_short] = ACTIONS(7109),
    [anon_sym_ATautoreleasepool] = ACTIONS(7111),
    [anon_sym_const] = ACTIONS(7109),
    [anon_sym_constexpr] = ACTIONS(7109),
    [anon_sym_volatile] = ACTIONS(7109),
    [anon_sym_restrict] = ACTIONS(7109),
    [anon_sym___restrict__] = ACTIONS(7109),
    [anon_sym__Atomic] = ACTIONS(7109),
    [anon_sym__Noreturn] = ACTIONS(7109),
    [anon_sym__Nonnull] = ACTIONS(7109),
    [anon_sym_nullable] = ACTIONS(7109),
    [anon_sym__Complex] = ACTIONS(7109),
    [anon_sym__Nullable] = ACTIONS(7109),
    [anon_sym__Nullable_result] = ACTIONS(7109),
    [anon_sym__Null_unspecified] = ACTIONS(7109),
    [anon_sym___autoreleasing] = ACTIONS(7109),
    [anon_sym___block] = ACTIONS(7109),
    [anon_sym___bridge] = ACTIONS(7109),
    [anon_sym___bridge_retained] = ACTIONS(7109),
    [anon_sym___bridge_transfer] = ACTIONS(7109),
    [anon_sym___complex] = ACTIONS(7109),
    [anon_sym___const] = ACTIONS(7109),
    [anon_sym___imag] = ACTIONS(7109),
    [anon_sym___kindof] = ACTIONS(7109),
    [anon_sym___nonnull] = ACTIONS(7109),
    [anon_sym___nullable] = ACTIONS(7109),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7109),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7109),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7109),
    [anon_sym___real] = ACTIONS(7109),
    [anon_sym___strong] = ACTIONS(7109),
    [anon_sym___unsafe_unretained] = ACTIONS(7109),
    [anon_sym___unused] = ACTIONS(7109),
    [anon_sym___weak] = ACTIONS(7109),
    [anon_sym_alignas] = ACTIONS(7109),
    [anon_sym__Alignas] = ACTIONS(7109),
    [sym_primitive_type] = ACTIONS(7109),
    [anon_sym_enum] = ACTIONS(7109),
    [anon_sym_COLON] = ACTIONS(7111),
    [anon_sym_struct] = ACTIONS(7109),
    [anon_sym_union] = ACTIONS(7109),
    [anon_sym_in] = ACTIONS(7109),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7109),
    [anon_sym___typeof] = ACTIONS(7109),
    [anon_sym_typeof] = ACTIONS(7109),
    [anon_sym_BOOL] = ACTIONS(7109),
    [anon_sym_IMP] = ACTIONS(7109),
    [anon_sym_SEL] = ACTIONS(7109),
    [anon_sym_Class] = ACTIONS(7109),
    [anon_sym_id] = ACTIONS(7109),
    [anon_sym_out] = ACTIONS(7109),
    [anon_sym_inout] = ACTIONS(7109),
    [anon_sym_bycopy] = ACTIONS(7109),
    [anon_sym_byref] = ACTIONS(7109),
    [anon_sym_oneway] = ACTIONS(7109),
  },
  [STATE(3802)] = {
    [sym_attribute_specifier] = STATE(3966),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_RPAREN] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7115),
    [anon_sym_STAR] = ACTIONS(7115),
    [anon_sym_CARET] = ACTIONS(7115),
    [anon_sym_GT] = ACTIONS(7115),
    [anon_sym_LT] = ACTIONS(7115),
    [anon_sym___extension__] = ACTIONS(7113),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7115),
    [anon_sym___based] = ACTIONS(7113),
    [anon_sym_LBRACE] = ACTIONS(7115),
    [anon_sym_signed] = ACTIONS(7113),
    [anon_sym_unsigned] = ACTIONS(7113),
    [anon_sym_long] = ACTIONS(7113),
    [anon_sym_short] = ACTIONS(7113),
    [anon_sym_ATautoreleasepool] = ACTIONS(7115),
    [anon_sym_const] = ACTIONS(7113),
    [anon_sym_constexpr] = ACTIONS(7113),
    [anon_sym_volatile] = ACTIONS(7113),
    [anon_sym_restrict] = ACTIONS(7113),
    [anon_sym___restrict__] = ACTIONS(7113),
    [anon_sym__Atomic] = ACTIONS(7113),
    [anon_sym__Noreturn] = ACTIONS(7113),
    [anon_sym__Nonnull] = ACTIONS(7113),
    [anon_sym_nullable] = ACTIONS(7113),
    [anon_sym__Complex] = ACTIONS(7113),
    [anon_sym__Nullable] = ACTIONS(7113),
    [anon_sym__Nullable_result] = ACTIONS(7113),
    [anon_sym__Null_unspecified] = ACTIONS(7113),
    [anon_sym___autoreleasing] = ACTIONS(7113),
    [anon_sym___block] = ACTIONS(7113),
    [anon_sym___bridge] = ACTIONS(7113),
    [anon_sym___bridge_retained] = ACTIONS(7113),
    [anon_sym___bridge_transfer] = ACTIONS(7113),
    [anon_sym___complex] = ACTIONS(7113),
    [anon_sym___const] = ACTIONS(7113),
    [anon_sym___imag] = ACTIONS(7113),
    [anon_sym___kindof] = ACTIONS(7113),
    [anon_sym___nonnull] = ACTIONS(7113),
    [anon_sym___nullable] = ACTIONS(7113),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7113),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7113),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7113),
    [anon_sym___real] = ACTIONS(7113),
    [anon_sym___strong] = ACTIONS(7113),
    [anon_sym___unsafe_unretained] = ACTIONS(7113),
    [anon_sym___unused] = ACTIONS(7113),
    [anon_sym___weak] = ACTIONS(7113),
    [anon_sym_alignas] = ACTIONS(7113),
    [anon_sym__Alignas] = ACTIONS(7113),
    [sym_primitive_type] = ACTIONS(7113),
    [anon_sym_enum] = ACTIONS(7113),
    [anon_sym_COLON] = ACTIONS(7115),
    [anon_sym_struct] = ACTIONS(7113),
    [anon_sym_union] = ACTIONS(7113),
    [anon_sym_in] = ACTIONS(7113),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7113),
    [anon_sym___typeof] = ACTIONS(7113),
    [anon_sym_typeof] = ACTIONS(7113),
    [anon_sym_BOOL] = ACTIONS(7113),
    [anon_sym_IMP] = ACTIONS(7113),
    [anon_sym_SEL] = ACTIONS(7113),
    [anon_sym_Class] = ACTIONS(7113),
    [anon_sym_id] = ACTIONS(7113),
    [anon_sym_out] = ACTIONS(7113),
    [anon_sym_inout] = ACTIONS(7113),
    [anon_sym_bycopy] = ACTIONS(7113),
    [anon_sym_byref] = ACTIONS(7113),
    [anon_sym_oneway] = ACTIONS(7113),
  },
  [STATE(3803)] = {
    [sym_attribute_specifier] = STATE(3803),
    [aux_sym_function_declarator_repeat1] = STATE(3803),
    [sym_identifier] = ACTIONS(6689),
    [anon_sym_COMMA] = ACTIONS(6691),
    [anon_sym_RPAREN] = ACTIONS(6691),
    [anon_sym_LPAREN2] = ACTIONS(6691),
    [anon_sym_STAR] = ACTIONS(6691),
    [anon_sym_CARET] = ACTIONS(6691),
    [anon_sym_GT] = ACTIONS(6691),
    [anon_sym_LT] = ACTIONS(6691),
    [anon_sym___extension__] = ACTIONS(6689),
    [anon_sym___attribute__] = ACTIONS(8798),
    [anon_sym___attribute] = ACTIONS(8798),
    [anon_sym_noreturn] = ACTIONS(6689),
    [anon_sym_LBRACK] = ACTIONS(6691),
    [anon_sym___based] = ACTIONS(6689),
    [anon_sym_LBRACE] = ACTIONS(6691),
    [anon_sym_signed] = ACTIONS(6689),
    [anon_sym_unsigned] = ACTIONS(6689),
    [anon_sym_long] = ACTIONS(6689),
    [anon_sym_short] = ACTIONS(6689),
    [anon_sym_ATautoreleasepool] = ACTIONS(6691),
    [anon_sym_const] = ACTIONS(6689),
    [anon_sym_constexpr] = ACTIONS(6689),
    [anon_sym_volatile] = ACTIONS(6689),
    [anon_sym_restrict] = ACTIONS(6689),
    [anon_sym___restrict__] = ACTIONS(6689),
    [anon_sym__Atomic] = ACTIONS(6689),
    [anon_sym__Noreturn] = ACTIONS(6689),
    [anon_sym__Nonnull] = ACTIONS(6689),
    [anon_sym_nullable] = ACTIONS(6689),
    [anon_sym__Complex] = ACTIONS(6689),
    [anon_sym__Nullable] = ACTIONS(6689),
    [anon_sym__Nullable_result] = ACTIONS(6689),
    [anon_sym__Null_unspecified] = ACTIONS(6689),
    [anon_sym___autoreleasing] = ACTIONS(6689),
    [anon_sym___block] = ACTIONS(6689),
    [anon_sym___bridge] = ACTIONS(6689),
    [anon_sym___bridge_retained] = ACTIONS(6689),
    [anon_sym___bridge_transfer] = ACTIONS(6689),
    [anon_sym___complex] = ACTIONS(6689),
    [anon_sym___const] = ACTIONS(6689),
    [anon_sym___imag] = ACTIONS(6689),
    [anon_sym___kindof] = ACTIONS(6689),
    [anon_sym___nonnull] = ACTIONS(6689),
    [anon_sym___nullable] = ACTIONS(6689),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6689),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6689),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6689),
    [anon_sym___real] = ACTIONS(6689),
    [anon_sym___strong] = ACTIONS(6689),
    [anon_sym___unsafe_unretained] = ACTIONS(6689),
    [anon_sym___unused] = ACTIONS(6689),
    [anon_sym___weak] = ACTIONS(6689),
    [anon_sym_alignas] = ACTIONS(6689),
    [anon_sym__Alignas] = ACTIONS(6689),
    [sym_primitive_type] = ACTIONS(6689),
    [anon_sym_enum] = ACTIONS(6689),
    [anon_sym_struct] = ACTIONS(6689),
    [anon_sym_union] = ACTIONS(6689),
    [anon_sym_in] = ACTIONS(6689),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6689),
    [anon_sym___typeof] = ACTIONS(6689),
    [anon_sym_typeof] = ACTIONS(6689),
    [anon_sym_BOOL] = ACTIONS(6689),
    [anon_sym_IMP] = ACTIONS(6689),
    [anon_sym_SEL] = ACTIONS(6689),
    [anon_sym_Class] = ACTIONS(6689),
    [anon_sym_id] = ACTIONS(6689),
    [anon_sym_out] = ACTIONS(6689),
    [anon_sym_inout] = ACTIONS(6689),
    [anon_sym_bycopy] = ACTIONS(6689),
    [anon_sym_byref] = ACTIONS(6689),
    [anon_sym_oneway] = ACTIONS(6689),
  },
  [STATE(3804)] = {
    [sym_attribute_specifier] = STATE(3973),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_RPAREN] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7135),
    [anon_sym_STAR] = ACTIONS(7135),
    [anon_sym_CARET] = ACTIONS(7135),
    [anon_sym_GT] = ACTIONS(7135),
    [anon_sym_LT] = ACTIONS(7135),
    [anon_sym___extension__] = ACTIONS(7133),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7135),
    [anon_sym___based] = ACTIONS(7133),
    [anon_sym_LBRACE] = ACTIONS(7135),
    [anon_sym_signed] = ACTIONS(7133),
    [anon_sym_unsigned] = ACTIONS(7133),
    [anon_sym_long] = ACTIONS(7133),
    [anon_sym_short] = ACTIONS(7133),
    [anon_sym_ATautoreleasepool] = ACTIONS(7135),
    [anon_sym_const] = ACTIONS(7133),
    [anon_sym_constexpr] = ACTIONS(7133),
    [anon_sym_volatile] = ACTIONS(7133),
    [anon_sym_restrict] = ACTIONS(7133),
    [anon_sym___restrict__] = ACTIONS(7133),
    [anon_sym__Atomic] = ACTIONS(7133),
    [anon_sym__Noreturn] = ACTIONS(7133),
    [anon_sym__Nonnull] = ACTIONS(7133),
    [anon_sym_nullable] = ACTIONS(7133),
    [anon_sym__Complex] = ACTIONS(7133),
    [anon_sym__Nullable] = ACTIONS(7133),
    [anon_sym__Nullable_result] = ACTIONS(7133),
    [anon_sym__Null_unspecified] = ACTIONS(7133),
    [anon_sym___autoreleasing] = ACTIONS(7133),
    [anon_sym___block] = ACTIONS(7133),
    [anon_sym___bridge] = ACTIONS(7133),
    [anon_sym___bridge_retained] = ACTIONS(7133),
    [anon_sym___bridge_transfer] = ACTIONS(7133),
    [anon_sym___complex] = ACTIONS(7133),
    [anon_sym___const] = ACTIONS(7133),
    [anon_sym___imag] = ACTIONS(7133),
    [anon_sym___kindof] = ACTIONS(7133),
    [anon_sym___nonnull] = ACTIONS(7133),
    [anon_sym___nullable] = ACTIONS(7133),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7133),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7133),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7133),
    [anon_sym___real] = ACTIONS(7133),
    [anon_sym___strong] = ACTIONS(7133),
    [anon_sym___unsafe_unretained] = ACTIONS(7133),
    [anon_sym___unused] = ACTIONS(7133),
    [anon_sym___weak] = ACTIONS(7133),
    [anon_sym_alignas] = ACTIONS(7133),
    [anon_sym__Alignas] = ACTIONS(7133),
    [sym_primitive_type] = ACTIONS(7133),
    [anon_sym_enum] = ACTIONS(7133),
    [anon_sym_COLON] = ACTIONS(7135),
    [anon_sym_struct] = ACTIONS(7133),
    [anon_sym_union] = ACTIONS(7133),
    [anon_sym_in] = ACTIONS(7133),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7133),
    [anon_sym___typeof] = ACTIONS(7133),
    [anon_sym_typeof] = ACTIONS(7133),
    [anon_sym_BOOL] = ACTIONS(7133),
    [anon_sym_IMP] = ACTIONS(7133),
    [anon_sym_SEL] = ACTIONS(7133),
    [anon_sym_Class] = ACTIONS(7133),
    [anon_sym_id] = ACTIONS(7133),
    [anon_sym_out] = ACTIONS(7133),
    [anon_sym_inout] = ACTIONS(7133),
    [anon_sym_bycopy] = ACTIONS(7133),
    [anon_sym_byref] = ACTIONS(7133),
    [anon_sym_oneway] = ACTIONS(7133),
  },
  [STATE(3805)] = {
    [sym_attribute_specifier] = STATE(3857),
    [sym_identifier] = ACTIONS(7076),
    [anon_sym_COMMA] = ACTIONS(7078),
    [anon_sym_RPAREN] = ACTIONS(7078),
    [anon_sym_LPAREN2] = ACTIONS(7078),
    [anon_sym_STAR] = ACTIONS(7078),
    [anon_sym_CARET] = ACTIONS(7078),
    [anon_sym_GT] = ACTIONS(7078),
    [anon_sym_LT] = ACTIONS(7078),
    [anon_sym___extension__] = ACTIONS(7076),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7076),
    [anon_sym_LBRACK] = ACTIONS(7078),
    [anon_sym___based] = ACTIONS(7076),
    [anon_sym_LBRACE] = ACTIONS(7078),
    [anon_sym_signed] = ACTIONS(7076),
    [anon_sym_unsigned] = ACTIONS(7076),
    [anon_sym_long] = ACTIONS(7076),
    [anon_sym_short] = ACTIONS(7076),
    [anon_sym_ATautoreleasepool] = ACTIONS(7078),
    [anon_sym_const] = ACTIONS(7076),
    [anon_sym_constexpr] = ACTIONS(7076),
    [anon_sym_volatile] = ACTIONS(7076),
    [anon_sym_restrict] = ACTIONS(7076),
    [anon_sym___restrict__] = ACTIONS(7076),
    [anon_sym__Atomic] = ACTIONS(7076),
    [anon_sym__Noreturn] = ACTIONS(7076),
    [anon_sym__Nonnull] = ACTIONS(7076),
    [anon_sym_nullable] = ACTIONS(7076),
    [anon_sym__Complex] = ACTIONS(7076),
    [anon_sym__Nullable] = ACTIONS(7076),
    [anon_sym__Nullable_result] = ACTIONS(7076),
    [anon_sym__Null_unspecified] = ACTIONS(7076),
    [anon_sym___autoreleasing] = ACTIONS(7076),
    [anon_sym___block] = ACTIONS(7076),
    [anon_sym___bridge] = ACTIONS(7076),
    [anon_sym___bridge_retained] = ACTIONS(7076),
    [anon_sym___bridge_transfer] = ACTIONS(7076),
    [anon_sym___complex] = ACTIONS(7076),
    [anon_sym___const] = ACTIONS(7076),
    [anon_sym___imag] = ACTIONS(7076),
    [anon_sym___kindof] = ACTIONS(7076),
    [anon_sym___nonnull] = ACTIONS(7076),
    [anon_sym___nullable] = ACTIONS(7076),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7076),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7076),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7076),
    [anon_sym___real] = ACTIONS(7076),
    [anon_sym___strong] = ACTIONS(7076),
    [anon_sym___unsafe_unretained] = ACTIONS(7076),
    [anon_sym___unused] = ACTIONS(7076),
    [anon_sym___weak] = ACTIONS(7076),
    [anon_sym_alignas] = ACTIONS(7076),
    [anon_sym__Alignas] = ACTIONS(7076),
    [sym_primitive_type] = ACTIONS(7076),
    [anon_sym_enum] = ACTIONS(7076),
    [anon_sym_COLON] = ACTIONS(7078),
    [anon_sym_struct] = ACTIONS(7076),
    [anon_sym_union] = ACTIONS(7076),
    [anon_sym_in] = ACTIONS(7076),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7076),
    [anon_sym___typeof] = ACTIONS(7076),
    [anon_sym_typeof] = ACTIONS(7076),
    [anon_sym_BOOL] = ACTIONS(7076),
    [anon_sym_IMP] = ACTIONS(7076),
    [anon_sym_SEL] = ACTIONS(7076),
    [anon_sym_Class] = ACTIONS(7076),
    [anon_sym_id] = ACTIONS(7076),
    [anon_sym_out] = ACTIONS(7076),
    [anon_sym_inout] = ACTIONS(7076),
    [anon_sym_bycopy] = ACTIONS(7076),
    [anon_sym_byref] = ACTIONS(7076),
    [anon_sym_oneway] = ACTIONS(7076),
  },
  [STATE(3806)] = {
    [sym_attribute_specifier] = STATE(3823),
    [aux_sym_function_declarator_repeat1] = STATE(3823),
    [sym_identifier] = ACTIONS(6685),
    [anon_sym_COMMA] = ACTIONS(6687),
    [anon_sym_RPAREN] = ACTIONS(6687),
    [anon_sym_LPAREN2] = ACTIONS(6687),
    [anon_sym_STAR] = ACTIONS(6687),
    [anon_sym_CARET] = ACTIONS(6687),
    [anon_sym_GT] = ACTIONS(6687),
    [anon_sym_LT] = ACTIONS(6687),
    [anon_sym___extension__] = ACTIONS(6685),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(6685),
    [anon_sym_LBRACK] = ACTIONS(6687),
    [anon_sym___based] = ACTIONS(6685),
    [anon_sym_LBRACE] = ACTIONS(6687),
    [anon_sym_signed] = ACTIONS(6685),
    [anon_sym_unsigned] = ACTIONS(6685),
    [anon_sym_long] = ACTIONS(6685),
    [anon_sym_short] = ACTIONS(6685),
    [anon_sym_ATautoreleasepool] = ACTIONS(6687),
    [anon_sym_const] = ACTIONS(6685),
    [anon_sym_constexpr] = ACTIONS(6685),
    [anon_sym_volatile] = ACTIONS(6685),
    [anon_sym_restrict] = ACTIONS(6685),
    [anon_sym___restrict__] = ACTIONS(6685),
    [anon_sym__Atomic] = ACTIONS(6685),
    [anon_sym__Noreturn] = ACTIONS(6685),
    [anon_sym__Nonnull] = ACTIONS(6685),
    [anon_sym_nullable] = ACTIONS(6685),
    [anon_sym__Complex] = ACTIONS(6685),
    [anon_sym__Nullable] = ACTIONS(6685),
    [anon_sym__Nullable_result] = ACTIONS(6685),
    [anon_sym__Null_unspecified] = ACTIONS(6685),
    [anon_sym___autoreleasing] = ACTIONS(6685),
    [anon_sym___block] = ACTIONS(6685),
    [anon_sym___bridge] = ACTIONS(6685),
    [anon_sym___bridge_retained] = ACTIONS(6685),
    [anon_sym___bridge_transfer] = ACTIONS(6685),
    [anon_sym___complex] = ACTIONS(6685),
    [anon_sym___const] = ACTIONS(6685),
    [anon_sym___imag] = ACTIONS(6685),
    [anon_sym___kindof] = ACTIONS(6685),
    [anon_sym___nonnull] = ACTIONS(6685),
    [anon_sym___nullable] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6685),
    [anon_sym___real] = ACTIONS(6685),
    [anon_sym___strong] = ACTIONS(6685),
    [anon_sym___unsafe_unretained] = ACTIONS(6685),
    [anon_sym___unused] = ACTIONS(6685),
    [anon_sym___weak] = ACTIONS(6685),
    [anon_sym_alignas] = ACTIONS(6685),
    [anon_sym__Alignas] = ACTIONS(6685),
    [sym_primitive_type] = ACTIONS(6685),
    [anon_sym_enum] = ACTIONS(6685),
    [anon_sym_struct] = ACTIONS(6685),
    [anon_sym_union] = ACTIONS(6685),
    [anon_sym_in] = ACTIONS(6685),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6685),
    [anon_sym___typeof] = ACTIONS(6685),
    [anon_sym_typeof] = ACTIONS(6685),
    [anon_sym_BOOL] = ACTIONS(6685),
    [anon_sym_IMP] = ACTIONS(6685),
    [anon_sym_SEL] = ACTIONS(6685),
    [anon_sym_Class] = ACTIONS(6685),
    [anon_sym_id] = ACTIONS(6685),
    [anon_sym_out] = ACTIONS(6685),
    [anon_sym_inout] = ACTIONS(6685),
    [anon_sym_bycopy] = ACTIONS(6685),
    [anon_sym_byref] = ACTIONS(6685),
    [anon_sym_oneway] = ACTIONS(6685),
  },
  [STATE(3807)] = {
    [sym_attribute_specifier] = STATE(3803),
    [aux_sym_function_declarator_repeat1] = STATE(3803),
    [sym_identifier] = ACTIONS(6685),
    [anon_sym_COMMA] = ACTIONS(6687),
    [anon_sym_RPAREN] = ACTIONS(6687),
    [anon_sym_LPAREN2] = ACTIONS(6687),
    [anon_sym_STAR] = ACTIONS(6687),
    [anon_sym_CARET] = ACTIONS(6687),
    [anon_sym_GT] = ACTIONS(6687),
    [anon_sym_LT] = ACTIONS(6687),
    [anon_sym___extension__] = ACTIONS(6685),
    [anon_sym___attribute__] = ACTIONS(6685),
    [anon_sym___attribute] = ACTIONS(6685),
    [anon_sym_noreturn] = ACTIONS(6685),
    [anon_sym_LBRACK] = ACTIONS(6687),
    [anon_sym___based] = ACTIONS(6685),
    [anon_sym_LBRACE] = ACTIONS(6687),
    [anon_sym_signed] = ACTIONS(6685),
    [anon_sym_unsigned] = ACTIONS(6685),
    [anon_sym_long] = ACTIONS(6685),
    [anon_sym_short] = ACTIONS(6685),
    [anon_sym_ATautoreleasepool] = ACTIONS(6687),
    [anon_sym_const] = ACTIONS(6685),
    [anon_sym_constexpr] = ACTIONS(6685),
    [anon_sym_volatile] = ACTIONS(6685),
    [anon_sym_restrict] = ACTIONS(6685),
    [anon_sym___restrict__] = ACTIONS(6685),
    [anon_sym__Atomic] = ACTIONS(6685),
    [anon_sym__Noreturn] = ACTIONS(6685),
    [anon_sym__Nonnull] = ACTIONS(6685),
    [anon_sym_nullable] = ACTIONS(6685),
    [anon_sym__Complex] = ACTIONS(6685),
    [anon_sym__Nullable] = ACTIONS(6685),
    [anon_sym__Nullable_result] = ACTIONS(6685),
    [anon_sym__Null_unspecified] = ACTIONS(6685),
    [anon_sym___autoreleasing] = ACTIONS(6685),
    [anon_sym___block] = ACTIONS(6685),
    [anon_sym___bridge] = ACTIONS(6685),
    [anon_sym___bridge_retained] = ACTIONS(6685),
    [anon_sym___bridge_transfer] = ACTIONS(6685),
    [anon_sym___complex] = ACTIONS(6685),
    [anon_sym___const] = ACTIONS(6685),
    [anon_sym___imag] = ACTIONS(6685),
    [anon_sym___kindof] = ACTIONS(6685),
    [anon_sym___nonnull] = ACTIONS(6685),
    [anon_sym___nullable] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6685),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6685),
    [anon_sym___real] = ACTIONS(6685),
    [anon_sym___strong] = ACTIONS(6685),
    [anon_sym___unsafe_unretained] = ACTIONS(6685),
    [anon_sym___unused] = ACTIONS(6685),
    [anon_sym___weak] = ACTIONS(6685),
    [anon_sym_alignas] = ACTIONS(6685),
    [anon_sym__Alignas] = ACTIONS(6685),
    [sym_primitive_type] = ACTIONS(6685),
    [anon_sym_enum] = ACTIONS(6685),
    [anon_sym_struct] = ACTIONS(6685),
    [anon_sym_union] = ACTIONS(6685),
    [anon_sym_in] = ACTIONS(6685),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6685),
    [anon_sym___typeof] = ACTIONS(6685),
    [anon_sym_typeof] = ACTIONS(6685),
    [anon_sym_BOOL] = ACTIONS(6685),
    [anon_sym_IMP] = ACTIONS(6685),
    [anon_sym_SEL] = ACTIONS(6685),
    [anon_sym_Class] = ACTIONS(6685),
    [anon_sym_id] = ACTIONS(6685),
    [anon_sym_out] = ACTIONS(6685),
    [anon_sym_inout] = ACTIONS(6685),
    [anon_sym_bycopy] = ACTIONS(6685),
    [anon_sym_byref] = ACTIONS(6685),
    [anon_sym_oneway] = ACTIONS(6685),
  },
  [STATE(3808)] = {
    [aux_sym_generic_specifier_repeat1] = STATE(3808),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_RPAREN] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7055),
    [anon_sym_STAR] = ACTIONS(7055),
    [anon_sym_CARET] = ACTIONS(7055),
    [anon_sym_GT] = ACTIONS(7055),
    [anon_sym_LT] = ACTIONS(8801),
    [anon_sym___extension__] = ACTIONS(7053),
    [anon_sym___attribute__] = ACTIONS(7053),
    [anon_sym___attribute] = ACTIONS(7053),
    [anon_sym_noreturn] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7055),
    [anon_sym___based] = ACTIONS(7053),
    [anon_sym_LBRACE] = ACTIONS(7055),
    [anon_sym_signed] = ACTIONS(7053),
    [anon_sym_unsigned] = ACTIONS(7053),
    [anon_sym_long] = ACTIONS(7053),
    [anon_sym_short] = ACTIONS(7053),
    [anon_sym_ATautoreleasepool] = ACTIONS(7055),
    [anon_sym_const] = ACTIONS(7053),
    [anon_sym_constexpr] = ACTIONS(7053),
    [anon_sym_volatile] = ACTIONS(7053),
    [anon_sym_restrict] = ACTIONS(7053),
    [anon_sym___restrict__] = ACTIONS(7053),
    [anon_sym__Atomic] = ACTIONS(7053),
    [anon_sym__Noreturn] = ACTIONS(7053),
    [anon_sym__Nonnull] = ACTIONS(7053),
    [anon_sym_nullable] = ACTIONS(7053),
    [anon_sym__Complex] = ACTIONS(7053),
    [anon_sym__Nullable] = ACTIONS(7053),
    [anon_sym__Nullable_result] = ACTIONS(7053),
    [anon_sym__Null_unspecified] = ACTIONS(7053),
    [anon_sym___autoreleasing] = ACTIONS(7053),
    [anon_sym___block] = ACTIONS(7053),
    [anon_sym___bridge] = ACTIONS(7053),
    [anon_sym___bridge_retained] = ACTIONS(7053),
    [anon_sym___bridge_transfer] = ACTIONS(7053),
    [anon_sym___complex] = ACTIONS(7053),
    [anon_sym___const] = ACTIONS(7053),
    [anon_sym___imag] = ACTIONS(7053),
    [anon_sym___kindof] = ACTIONS(7053),
    [anon_sym___nonnull] = ACTIONS(7053),
    [anon_sym___nullable] = ACTIONS(7053),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7053),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7053),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7053),
    [anon_sym___real] = ACTIONS(7053),
    [anon_sym___strong] = ACTIONS(7053),
    [anon_sym___unsafe_unretained] = ACTIONS(7053),
    [anon_sym___unused] = ACTIONS(7053),
    [anon_sym___weak] = ACTIONS(7053),
    [anon_sym_alignas] = ACTIONS(7053),
    [anon_sym__Alignas] = ACTIONS(7053),
    [sym_primitive_type] = ACTIONS(7053),
    [anon_sym_enum] = ACTIONS(7053),
    [anon_sym_COLON] = ACTIONS(7055),
    [anon_sym_struct] = ACTIONS(7053),
    [anon_sym_union] = ACTIONS(7053),
    [anon_sym_in] = ACTIONS(7053),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7053),
    [anon_sym___typeof] = ACTIONS(7053),
    [anon_sym_typeof] = ACTIONS(7053),
    [anon_sym_BOOL] = ACTIONS(7053),
    [anon_sym_IMP] = ACTIONS(7053),
    [anon_sym_SEL] = ACTIONS(7053),
    [anon_sym_Class] = ACTIONS(7053),
    [anon_sym_id] = ACTIONS(7053),
    [anon_sym_out] = ACTIONS(7053),
    [anon_sym_inout] = ACTIONS(7053),
    [anon_sym_bycopy] = ACTIONS(7053),
    [anon_sym_byref] = ACTIONS(7053),
    [anon_sym_oneway] = ACTIONS(7053),
  },
  [STATE(3809)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3792),
    [sym_identifier] = ACTIONS(7303),
    [anon_sym_COMMA] = ACTIONS(7305),
    [anon_sym_RPAREN] = ACTIONS(7305),
    [anon_sym_LPAREN2] = ACTIONS(7305),
    [anon_sym_STAR] = ACTIONS(7305),
    [anon_sym_CARET] = ACTIONS(7305),
    [anon_sym_GT] = ACTIONS(7305),
    [anon_sym_LT] = ACTIONS(7305),
    [anon_sym___extension__] = ACTIONS(7303),
    [anon_sym___attribute__] = ACTIONS(7303),
    [anon_sym___attribute] = ACTIONS(7303),
    [anon_sym_noreturn] = ACTIONS(7303),
    [anon_sym_LBRACK] = ACTIONS(7305),
    [anon_sym___based] = ACTIONS(7303),
    [anon_sym_LBRACE] = ACTIONS(7305),
    [anon_sym_signed] = ACTIONS(8804),
    [anon_sym_unsigned] = ACTIONS(8804),
    [anon_sym_long] = ACTIONS(8804),
    [anon_sym_short] = ACTIONS(8804),
    [anon_sym_ATautoreleasepool] = ACTIONS(7305),
    [anon_sym_const] = ACTIONS(7303),
    [anon_sym_constexpr] = ACTIONS(7303),
    [anon_sym_volatile] = ACTIONS(7303),
    [anon_sym_restrict] = ACTIONS(7303),
    [anon_sym___restrict__] = ACTIONS(7303),
    [anon_sym__Atomic] = ACTIONS(7303),
    [anon_sym__Noreturn] = ACTIONS(7303),
    [anon_sym__Nonnull] = ACTIONS(7303),
    [anon_sym_nullable] = ACTIONS(7303),
    [anon_sym__Complex] = ACTIONS(7303),
    [anon_sym__Nullable] = ACTIONS(7303),
    [anon_sym__Nullable_result] = ACTIONS(7303),
    [anon_sym__Null_unspecified] = ACTIONS(7303),
    [anon_sym___autoreleasing] = ACTIONS(7303),
    [anon_sym___block] = ACTIONS(7303),
    [anon_sym___bridge] = ACTIONS(7303),
    [anon_sym___bridge_retained] = ACTIONS(7303),
    [anon_sym___bridge_transfer] = ACTIONS(7303),
    [anon_sym___complex] = ACTIONS(7303),
    [anon_sym___const] = ACTIONS(7303),
    [anon_sym___imag] = ACTIONS(7303),
    [anon_sym___kindof] = ACTIONS(7303),
    [anon_sym___nonnull] = ACTIONS(7303),
    [anon_sym___nullable] = ACTIONS(7303),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7303),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7303),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7303),
    [anon_sym___real] = ACTIONS(7303),
    [anon_sym___strong] = ACTIONS(7303),
    [anon_sym___unsafe_unretained] = ACTIONS(7303),
    [anon_sym___unused] = ACTIONS(7303),
    [anon_sym___weak] = ACTIONS(7303),
    [anon_sym_alignas] = ACTIONS(7303),
    [anon_sym__Alignas] = ACTIONS(7303),
    [sym_primitive_type] = ACTIONS(7303),
    [anon_sym_enum] = ACTIONS(7303),
    [anon_sym_COLON] = ACTIONS(7305),
    [anon_sym_struct] = ACTIONS(7303),
    [anon_sym_union] = ACTIONS(7303),
    [anon_sym_in] = ACTIONS(7303),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7303),
    [anon_sym___typeof] = ACTIONS(7303),
    [anon_sym_typeof] = ACTIONS(7303),
    [anon_sym_BOOL] = ACTIONS(7303),
    [anon_sym_IMP] = ACTIONS(7303),
    [anon_sym_SEL] = ACTIONS(7303),
    [anon_sym_Class] = ACTIONS(7303),
    [anon_sym_id] = ACTIONS(7303),
    [anon_sym_out] = ACTIONS(7303),
    [anon_sym_inout] = ACTIONS(7303),
    [anon_sym_bycopy] = ACTIONS(7303),
    [anon_sym_byref] = ACTIONS(7303),
    [anon_sym_oneway] = ACTIONS(7303),
  },
  [STATE(3810)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3792),
    [sym_identifier] = ACTIONS(7307),
    [anon_sym_COMMA] = ACTIONS(7309),
    [anon_sym_RPAREN] = ACTIONS(7309),
    [anon_sym_LPAREN2] = ACTIONS(7309),
    [anon_sym_STAR] = ACTIONS(7309),
    [anon_sym_CARET] = ACTIONS(7309),
    [anon_sym_GT] = ACTIONS(7309),
    [anon_sym_LT] = ACTIONS(7309),
    [anon_sym___extension__] = ACTIONS(7307),
    [anon_sym___attribute__] = ACTIONS(7307),
    [anon_sym___attribute] = ACTIONS(7307),
    [anon_sym_noreturn] = ACTIONS(7307),
    [anon_sym_LBRACK] = ACTIONS(7309),
    [anon_sym___based] = ACTIONS(7307),
    [anon_sym_LBRACE] = ACTIONS(7309),
    [anon_sym_signed] = ACTIONS(8807),
    [anon_sym_unsigned] = ACTIONS(8807),
    [anon_sym_long] = ACTIONS(8807),
    [anon_sym_short] = ACTIONS(8807),
    [anon_sym_ATautoreleasepool] = ACTIONS(7309),
    [anon_sym_const] = ACTIONS(7307),
    [anon_sym_constexpr] = ACTIONS(7307),
    [anon_sym_volatile] = ACTIONS(7307),
    [anon_sym_restrict] = ACTIONS(7307),
    [anon_sym___restrict__] = ACTIONS(7307),
    [anon_sym__Atomic] = ACTIONS(7307),
    [anon_sym__Noreturn] = ACTIONS(7307),
    [anon_sym__Nonnull] = ACTIONS(7307),
    [anon_sym_nullable] = ACTIONS(7307),
    [anon_sym__Complex] = ACTIONS(7307),
    [anon_sym__Nullable] = ACTIONS(7307),
    [anon_sym__Nullable_result] = ACTIONS(7307),
    [anon_sym__Null_unspecified] = ACTIONS(7307),
    [anon_sym___autoreleasing] = ACTIONS(7307),
    [anon_sym___block] = ACTIONS(7307),
    [anon_sym___bridge] = ACTIONS(7307),
    [anon_sym___bridge_retained] = ACTIONS(7307),
    [anon_sym___bridge_transfer] = ACTIONS(7307),
    [anon_sym___complex] = ACTIONS(7307),
    [anon_sym___const] = ACTIONS(7307),
    [anon_sym___imag] = ACTIONS(7307),
    [anon_sym___kindof] = ACTIONS(7307),
    [anon_sym___nonnull] = ACTIONS(7307),
    [anon_sym___nullable] = ACTIONS(7307),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7307),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7307),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7307),
    [anon_sym___real] = ACTIONS(7307),
    [anon_sym___strong] = ACTIONS(7307),
    [anon_sym___unsafe_unretained] = ACTIONS(7307),
    [anon_sym___unused] = ACTIONS(7307),
    [anon_sym___weak] = ACTIONS(7307),
    [anon_sym_alignas] = ACTIONS(7307),
    [anon_sym__Alignas] = ACTIONS(7307),
    [sym_primitive_type] = ACTIONS(7307),
    [anon_sym_enum] = ACTIONS(7307),
    [anon_sym_COLON] = ACTIONS(7309),
    [anon_sym_struct] = ACTIONS(7307),
    [anon_sym_union] = ACTIONS(7307),
    [anon_sym_in] = ACTIONS(7307),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7307),
    [anon_sym___typeof] = ACTIONS(7307),
    [anon_sym_typeof] = ACTIONS(7307),
    [anon_sym_BOOL] = ACTIONS(7307),
    [anon_sym_IMP] = ACTIONS(7307),
    [anon_sym_SEL] = ACTIONS(7307),
    [anon_sym_Class] = ACTIONS(7307),
    [anon_sym_id] = ACTIONS(7307),
    [anon_sym_out] = ACTIONS(7307),
    [anon_sym_inout] = ACTIONS(7307),
    [anon_sym_bycopy] = ACTIONS(7307),
    [anon_sym_byref] = ACTIONS(7307),
    [anon_sym_oneway] = ACTIONS(7307),
  },
  [STATE(3811)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3812),
    [aux_sym_parameterized_arguments_repeat1] = STATE(5126),
    [aux_sym_parameterized_arguments_repeat2] = STATE(5618),
    [aux_sym_generic_specifier_repeat1] = STATE(3813),
    [sym_identifier] = ACTIONS(6138),
    [anon_sym_COMMA] = ACTIONS(6143),
    [anon_sym_LPAREN2] = ACTIONS(6130),
    [anon_sym_STAR] = ACTIONS(6143),
    [anon_sym_CARET] = ACTIONS(6143),
    [anon_sym_GT] = ACTIONS(6143),
    [anon_sym_LT] = ACTIONS(8777),
    [anon_sym___extension__] = ACTIONS(6138),
    [anon_sym___attribute__] = ACTIONS(6138),
    [anon_sym___attribute] = ACTIONS(6138),
    [anon_sym_noreturn] = ACTIONS(6138),
    [anon_sym_LBRACK] = ACTIONS(6143),
    [anon_sym___based] = ACTIONS(6138),
    [anon_sym_signed] = ACTIONS(8781),
    [anon_sym_unsigned] = ACTIONS(8781),
    [anon_sym_long] = ACTIONS(8781),
    [anon_sym_short] = ACTIONS(8781),
    [anon_sym_const] = ACTIONS(6138),
    [anon_sym_constexpr] = ACTIONS(6138),
    [anon_sym_volatile] = ACTIONS(6138),
    [anon_sym_restrict] = ACTIONS(6138),
    [anon_sym___restrict__] = ACTIONS(6138),
    [anon_sym__Atomic] = ACTIONS(6138),
    [anon_sym__Noreturn] = ACTIONS(6138),
    [anon_sym__Nonnull] = ACTIONS(6138),
    [anon_sym_nullable] = ACTIONS(6138),
    [anon_sym__Complex] = ACTIONS(6138),
    [anon_sym__Nullable] = ACTIONS(6138),
    [anon_sym__Nullable_result] = ACTIONS(6138),
    [anon_sym__Null_unspecified] = ACTIONS(6138),
    [anon_sym___autoreleasing] = ACTIONS(6138),
    [anon_sym___block] = ACTIONS(6138),
    [anon_sym___bridge] = ACTIONS(6138),
    [anon_sym___bridge_retained] = ACTIONS(6138),
    [anon_sym___bridge_transfer] = ACTIONS(6138),
    [anon_sym___complex] = ACTIONS(6138),
    [anon_sym___const] = ACTIONS(6138),
    [anon_sym___imag] = ACTIONS(6138),
    [anon_sym___kindof] = ACTIONS(6138),
    [anon_sym___nonnull] = ACTIONS(6138),
    [anon_sym___nullable] = ACTIONS(6138),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6138),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6138),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6138),
    [anon_sym___real] = ACTIONS(6138),
    [anon_sym___strong] = ACTIONS(6138),
    [anon_sym___unsafe_unretained] = ACTIONS(6138),
    [anon_sym___unused] = ACTIONS(6138),
    [anon_sym___weak] = ACTIONS(6138),
    [anon_sym_alignas] = ACTIONS(6138),
    [anon_sym__Alignas] = ACTIONS(6138),
    [sym_primitive_type] = ACTIONS(6138),
    [anon_sym_enum] = ACTIONS(6138),
    [anon_sym_COLON] = ACTIONS(8810),
    [anon_sym_struct] = ACTIONS(6138),
    [anon_sym_union] = ACTIONS(6138),
    [anon_sym_in] = ACTIONS(6138),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6138),
    [anon_sym___typeof] = ACTIONS(6138),
    [anon_sym_typeof] = ACTIONS(6138),
    [anon_sym_BOOL] = ACTIONS(6138),
    [anon_sym_IMP] = ACTIONS(6138),
    [anon_sym_SEL] = ACTIONS(6138),
    [anon_sym_Class] = ACTIONS(6138),
    [anon_sym_id] = ACTIONS(6138),
    [anon_sym_out] = ACTIONS(6138),
    [anon_sym_inout] = ACTIONS(6138),
    [anon_sym_bycopy] = ACTIONS(6138),
    [anon_sym_byref] = ACTIONS(6138),
    [anon_sym_oneway] = ACTIONS(6138),
  },
  [STATE(3812)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3792),
    [sym_identifier] = ACTIONS(7283),
    [anon_sym_COMMA] = ACTIONS(7285),
    [anon_sym_RPAREN] = ACTIONS(7285),
    [anon_sym_LPAREN2] = ACTIONS(7285),
    [anon_sym_STAR] = ACTIONS(7285),
    [anon_sym_CARET] = ACTIONS(7285),
    [anon_sym_GT] = ACTIONS(7285),
    [anon_sym_LT] = ACTIONS(7285),
    [anon_sym___extension__] = ACTIONS(7283),
    [anon_sym___attribute__] = ACTIONS(7283),
    [anon_sym___attribute] = ACTIONS(7283),
    [anon_sym_noreturn] = ACTIONS(7283),
    [anon_sym_LBRACK] = ACTIONS(7285),
    [anon_sym___based] = ACTIONS(7283),
    [anon_sym_LBRACE] = ACTIONS(7285),
    [anon_sym_signed] = ACTIONS(8812),
    [anon_sym_unsigned] = ACTIONS(8812),
    [anon_sym_long] = ACTIONS(8812),
    [anon_sym_short] = ACTIONS(8812),
    [anon_sym_ATautoreleasepool] = ACTIONS(7285),
    [anon_sym_const] = ACTIONS(7283),
    [anon_sym_constexpr] = ACTIONS(7283),
    [anon_sym_volatile] = ACTIONS(7283),
    [anon_sym_restrict] = ACTIONS(7283),
    [anon_sym___restrict__] = ACTIONS(7283),
    [anon_sym__Atomic] = ACTIONS(7283),
    [anon_sym__Noreturn] = ACTIONS(7283),
    [anon_sym__Nonnull] = ACTIONS(7283),
    [anon_sym_nullable] = ACTIONS(7283),
    [anon_sym__Complex] = ACTIONS(7283),
    [anon_sym__Nullable] = ACTIONS(7283),
    [anon_sym__Nullable_result] = ACTIONS(7283),
    [anon_sym__Null_unspecified] = ACTIONS(7283),
    [anon_sym___autoreleasing] = ACTIONS(7283),
    [anon_sym___block] = ACTIONS(7283),
    [anon_sym___bridge] = ACTIONS(7283),
    [anon_sym___bridge_retained] = ACTIONS(7283),
    [anon_sym___bridge_transfer] = ACTIONS(7283),
    [anon_sym___complex] = ACTIONS(7283),
    [anon_sym___const] = ACTIONS(7283),
    [anon_sym___imag] = ACTIONS(7283),
    [anon_sym___kindof] = ACTIONS(7283),
    [anon_sym___nonnull] = ACTIONS(7283),
    [anon_sym___nullable] = ACTIONS(7283),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7283),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7283),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7283),
    [anon_sym___real] = ACTIONS(7283),
    [anon_sym___strong] = ACTIONS(7283),
    [anon_sym___unsafe_unretained] = ACTIONS(7283),
    [anon_sym___unused] = ACTIONS(7283),
    [anon_sym___weak] = ACTIONS(7283),
    [anon_sym_alignas] = ACTIONS(7283),
    [anon_sym__Alignas] = ACTIONS(7283),
    [sym_primitive_type] = ACTIONS(7283),
    [anon_sym_enum] = ACTIONS(7283),
    [anon_sym_COLON] = ACTIONS(7285),
    [anon_sym_struct] = ACTIONS(7283),
    [anon_sym_union] = ACTIONS(7283),
    [anon_sym_in] = ACTIONS(7283),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7283),
    [anon_sym___typeof] = ACTIONS(7283),
    [anon_sym_typeof] = ACTIONS(7283),
    [anon_sym_BOOL] = ACTIONS(7283),
    [anon_sym_IMP] = ACTIONS(7283),
    [anon_sym_SEL] = ACTIONS(7283),
    [anon_sym_Class] = ACTIONS(7283),
    [anon_sym_id] = ACTIONS(7283),
    [anon_sym_out] = ACTIONS(7283),
    [anon_sym_inout] = ACTIONS(7283),
    [anon_sym_bycopy] = ACTIONS(7283),
    [anon_sym_byref] = ACTIONS(7283),
    [anon_sym_oneway] = ACTIONS(7283),
  },
  [STATE(3813)] = {
    [aux_sym_generic_specifier_repeat1] = STATE(3808),
    [sym_identifier] = ACTIONS(7068),
    [anon_sym_COMMA] = ACTIONS(7070),
    [anon_sym_RPAREN] = ACTIONS(7070),
    [anon_sym_LPAREN2] = ACTIONS(7070),
    [anon_sym_STAR] = ACTIONS(7070),
    [anon_sym_CARET] = ACTIONS(7070),
    [anon_sym_GT] = ACTIONS(7070),
    [anon_sym_LT] = ACTIONS(6136),
    [anon_sym___extension__] = ACTIONS(7068),
    [anon_sym___attribute__] = ACTIONS(7068),
    [anon_sym___attribute] = ACTIONS(7068),
    [anon_sym_noreturn] = ACTIONS(7068),
    [anon_sym_LBRACK] = ACTIONS(7070),
    [anon_sym___based] = ACTIONS(7068),
    [anon_sym_LBRACE] = ACTIONS(7070),
    [anon_sym_signed] = ACTIONS(7068),
    [anon_sym_unsigned] = ACTIONS(7068),
    [anon_sym_long] = ACTIONS(7068),
    [anon_sym_short] = ACTIONS(7068),
    [anon_sym_ATautoreleasepool] = ACTIONS(7070),
    [anon_sym_const] = ACTIONS(7068),
    [anon_sym_constexpr] = ACTIONS(7068),
    [anon_sym_volatile] = ACTIONS(7068),
    [anon_sym_restrict] = ACTIONS(7068),
    [anon_sym___restrict__] = ACTIONS(7068),
    [anon_sym__Atomic] = ACTIONS(7068),
    [anon_sym__Noreturn] = ACTIONS(7068),
    [anon_sym__Nonnull] = ACTIONS(7068),
    [anon_sym_nullable] = ACTIONS(7068),
    [anon_sym__Complex] = ACTIONS(7068),
    [anon_sym__Nullable] = ACTIONS(7068),
    [anon_sym__Nullable_result] = ACTIONS(7068),
    [anon_sym__Null_unspecified] = ACTIONS(7068),
    [anon_sym___autoreleasing] = ACTIONS(7068),
    [anon_sym___block] = ACTIONS(7068),
    [anon_sym___bridge] = ACTIONS(7068),
    [anon_sym___bridge_retained] = ACTIONS(7068),
    [anon_sym___bridge_transfer] = ACTIONS(7068),
    [anon_sym___complex] = ACTIONS(7068),
    [anon_sym___const] = ACTIONS(7068),
    [anon_sym___imag] = ACTIONS(7068),
    [anon_sym___kindof] = ACTIONS(7068),
    [anon_sym___nonnull] = ACTIONS(7068),
    [anon_sym___nullable] = ACTIONS(7068),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7068),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7068),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7068),
    [anon_sym___real] = ACTIONS(7068),
    [anon_sym___strong] = ACTIONS(7068),
    [anon_sym___unsafe_unretained] = ACTIONS(7068),
    [anon_sym___unused] = ACTIONS(7068),
    [anon_sym___weak] = ACTIONS(7068),
    [anon_sym_alignas] = ACTIONS(7068),
    [anon_sym__Alignas] = ACTIONS(7068),
    [sym_primitive_type] = ACTIONS(7068),
    [anon_sym_enum] = ACTIONS(7068),
    [anon_sym_COLON] = ACTIONS(7070),
    [anon_sym_struct] = ACTIONS(7068),
    [anon_sym_union] = ACTIONS(7068),
    [anon_sym_in] = ACTIONS(7068),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7068),
    [anon_sym___typeof] = ACTIONS(7068),
    [anon_sym_typeof] = ACTIONS(7068),
    [anon_sym_BOOL] = ACTIONS(7068),
    [anon_sym_IMP] = ACTIONS(7068),
    [anon_sym_SEL] = ACTIONS(7068),
    [anon_sym_Class] = ACTIONS(7068),
    [anon_sym_id] = ACTIONS(7068),
    [anon_sym_out] = ACTIONS(7068),
    [anon_sym_inout] = ACTIONS(7068),
    [anon_sym_bycopy] = ACTIONS(7068),
    [anon_sym_byref] = ACTIONS(7068),
    [anon_sym_oneway] = ACTIONS(7068),
  },
  [STATE(3814)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3792),
    [sym_identifier] = ACTIONS(7365),
    [anon_sym_COMMA] = ACTIONS(7367),
    [anon_sym_RPAREN] = ACTIONS(7367),
    [anon_sym_LPAREN2] = ACTIONS(7367),
    [anon_sym_STAR] = ACTIONS(7367),
    [anon_sym_CARET] = ACTIONS(7367),
    [anon_sym_GT] = ACTIONS(7367),
    [anon_sym_LT] = ACTIONS(7367),
    [anon_sym___extension__] = ACTIONS(7365),
    [anon_sym___attribute__] = ACTIONS(7365),
    [anon_sym___attribute] = ACTIONS(7365),
    [anon_sym_noreturn] = ACTIONS(7365),
    [anon_sym_LBRACK] = ACTIONS(7367),
    [anon_sym___based] = ACTIONS(7365),
    [anon_sym_LBRACE] = ACTIONS(7367),
    [anon_sym_signed] = ACTIONS(8815),
    [anon_sym_unsigned] = ACTIONS(8815),
    [anon_sym_long] = ACTIONS(8815),
    [anon_sym_short] = ACTIONS(8815),
    [anon_sym_ATautoreleasepool] = ACTIONS(7367),
    [anon_sym_const] = ACTIONS(7365),
    [anon_sym_constexpr] = ACTIONS(7365),
    [anon_sym_volatile] = ACTIONS(7365),
    [anon_sym_restrict] = ACTIONS(7365),
    [anon_sym___restrict__] = ACTIONS(7365),
    [anon_sym__Atomic] = ACTIONS(7365),
    [anon_sym__Noreturn] = ACTIONS(7365),
    [anon_sym__Nonnull] = ACTIONS(7365),
    [anon_sym_nullable] = ACTIONS(7365),
    [anon_sym__Complex] = ACTIONS(7365),
    [anon_sym__Nullable] = ACTIONS(7365),
    [anon_sym__Nullable_result] = ACTIONS(7365),
    [anon_sym__Null_unspecified] = ACTIONS(7365),
    [anon_sym___autoreleasing] = ACTIONS(7365),
    [anon_sym___block] = ACTIONS(7365),
    [anon_sym___bridge] = ACTIONS(7365),
    [anon_sym___bridge_retained] = ACTIONS(7365),
    [anon_sym___bridge_transfer] = ACTIONS(7365),
    [anon_sym___complex] = ACTIONS(7365),
    [anon_sym___const] = ACTIONS(7365),
    [anon_sym___imag] = ACTIONS(7365),
    [anon_sym___kindof] = ACTIONS(7365),
    [anon_sym___nonnull] = ACTIONS(7365),
    [anon_sym___nullable] = ACTIONS(7365),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7365),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7365),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7365),
    [anon_sym___real] = ACTIONS(7365),
    [anon_sym___strong] = ACTIONS(7365),
    [anon_sym___unsafe_unretained] = ACTIONS(7365),
    [anon_sym___unused] = ACTIONS(7365),
    [anon_sym___weak] = ACTIONS(7365),
    [anon_sym_alignas] = ACTIONS(7365),
    [anon_sym__Alignas] = ACTIONS(7365),
    [sym_primitive_type] = ACTIONS(7365),
    [anon_sym_enum] = ACTIONS(7365),
    [anon_sym_COLON] = ACTIONS(7367),
    [anon_sym_struct] = ACTIONS(7365),
    [anon_sym_union] = ACTIONS(7365),
    [anon_sym_in] = ACTIONS(7365),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7365),
    [anon_sym___typeof] = ACTIONS(7365),
    [anon_sym_typeof] = ACTIONS(7365),
    [anon_sym_BOOL] = ACTIONS(7365),
    [anon_sym_IMP] = ACTIONS(7365),
    [anon_sym_SEL] = ACTIONS(7365),
    [anon_sym_Class] = ACTIONS(7365),
    [anon_sym_id] = ACTIONS(7365),
    [anon_sym_out] = ACTIONS(7365),
    [anon_sym_inout] = ACTIONS(7365),
    [anon_sym_bycopy] = ACTIONS(7365),
    [anon_sym_byref] = ACTIONS(7365),
    [anon_sym_oneway] = ACTIONS(7365),
  },
  [STATE(3815)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3792),
    [sym_identifier] = ACTIONS(7377),
    [anon_sym_COMMA] = ACTIONS(7379),
    [anon_sym_RPAREN] = ACTIONS(7379),
    [anon_sym_LPAREN2] = ACTIONS(7379),
    [anon_sym_STAR] = ACTIONS(7379),
    [anon_sym_CARET] = ACTIONS(7379),
    [anon_sym_GT] = ACTIONS(7379),
    [anon_sym_LT] = ACTIONS(7379),
    [anon_sym___extension__] = ACTIONS(7377),
    [anon_sym___attribute__] = ACTIONS(7377),
    [anon_sym___attribute] = ACTIONS(7377),
    [anon_sym_noreturn] = ACTIONS(7377),
    [anon_sym_LBRACK] = ACTIONS(7379),
    [anon_sym___based] = ACTIONS(7377),
    [anon_sym_LBRACE] = ACTIONS(7379),
    [anon_sym_signed] = ACTIONS(8818),
    [anon_sym_unsigned] = ACTIONS(8818),
    [anon_sym_long] = ACTIONS(8818),
    [anon_sym_short] = ACTIONS(8818),
    [anon_sym_ATautoreleasepool] = ACTIONS(7379),
    [anon_sym_const] = ACTIONS(7377),
    [anon_sym_constexpr] = ACTIONS(7377),
    [anon_sym_volatile] = ACTIONS(7377),
    [anon_sym_restrict] = ACTIONS(7377),
    [anon_sym___restrict__] = ACTIONS(7377),
    [anon_sym__Atomic] = ACTIONS(7377),
    [anon_sym__Noreturn] = ACTIONS(7377),
    [anon_sym__Nonnull] = ACTIONS(7377),
    [anon_sym_nullable] = ACTIONS(7377),
    [anon_sym__Complex] = ACTIONS(7377),
    [anon_sym__Nullable] = ACTIONS(7377),
    [anon_sym__Nullable_result] = ACTIONS(7377),
    [anon_sym__Null_unspecified] = ACTIONS(7377),
    [anon_sym___autoreleasing] = ACTIONS(7377),
    [anon_sym___block] = ACTIONS(7377),
    [anon_sym___bridge] = ACTIONS(7377),
    [anon_sym___bridge_retained] = ACTIONS(7377),
    [anon_sym___bridge_transfer] = ACTIONS(7377),
    [anon_sym___complex] = ACTIONS(7377),
    [anon_sym___const] = ACTIONS(7377),
    [anon_sym___imag] = ACTIONS(7377),
    [anon_sym___kindof] = ACTIONS(7377),
    [anon_sym___nonnull] = ACTIONS(7377),
    [anon_sym___nullable] = ACTIONS(7377),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7377),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7377),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7377),
    [anon_sym___real] = ACTIONS(7377),
    [anon_sym___strong] = ACTIONS(7377),
    [anon_sym___unsafe_unretained] = ACTIONS(7377),
    [anon_sym___unused] = ACTIONS(7377),
    [anon_sym___weak] = ACTIONS(7377),
    [anon_sym_alignas] = ACTIONS(7377),
    [anon_sym__Alignas] = ACTIONS(7377),
    [sym_primitive_type] = ACTIONS(7377),
    [anon_sym_enum] = ACTIONS(7377),
    [anon_sym_COLON] = ACTIONS(7379),
    [anon_sym_struct] = ACTIONS(7377),
    [anon_sym_union] = ACTIONS(7377),
    [anon_sym_in] = ACTIONS(7377),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7377),
    [anon_sym___typeof] = ACTIONS(7377),
    [anon_sym_typeof] = ACTIONS(7377),
    [anon_sym_BOOL] = ACTIONS(7377),
    [anon_sym_IMP] = ACTIONS(7377),
    [anon_sym_SEL] = ACTIONS(7377),
    [anon_sym_Class] = ACTIONS(7377),
    [anon_sym_id] = ACTIONS(7377),
    [anon_sym_out] = ACTIONS(7377),
    [anon_sym_inout] = ACTIONS(7377),
    [anon_sym_bycopy] = ACTIONS(7377),
    [anon_sym_byref] = ACTIONS(7377),
    [anon_sym_oneway] = ACTIONS(7377),
  },
  [STATE(3816)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3809),
    [sym_identifier] = ACTIONS(7405),
    [anon_sym_COMMA] = ACTIONS(7407),
    [anon_sym_RPAREN] = ACTIONS(7407),
    [anon_sym_LPAREN2] = ACTIONS(7407),
    [anon_sym_STAR] = ACTIONS(7407),
    [anon_sym_CARET] = ACTIONS(7407),
    [anon_sym_GT] = ACTIONS(7407),
    [anon_sym_LT] = ACTIONS(7407),
    [anon_sym___extension__] = ACTIONS(7405),
    [anon_sym___attribute__] = ACTIONS(7405),
    [anon_sym___attribute] = ACTIONS(7405),
    [anon_sym_noreturn] = ACTIONS(7405),
    [anon_sym_LBRACK] = ACTIONS(7407),
    [anon_sym___based] = ACTIONS(7405),
    [anon_sym_LBRACE] = ACTIONS(7407),
    [anon_sym_signed] = ACTIONS(8821),
    [anon_sym_unsigned] = ACTIONS(8821),
    [anon_sym_long] = ACTIONS(8821),
    [anon_sym_short] = ACTIONS(8821),
    [anon_sym_ATautoreleasepool] = ACTIONS(7407),
    [anon_sym_const] = ACTIONS(7405),
    [anon_sym_constexpr] = ACTIONS(7405),
    [anon_sym_volatile] = ACTIONS(7405),
    [anon_sym_restrict] = ACTIONS(7405),
    [anon_sym___restrict__] = ACTIONS(7405),
    [anon_sym__Atomic] = ACTIONS(7405),
    [anon_sym__Noreturn] = ACTIONS(7405),
    [anon_sym__Nonnull] = ACTIONS(7405),
    [anon_sym_nullable] = ACTIONS(7405),
    [anon_sym__Complex] = ACTIONS(7405),
    [anon_sym__Nullable] = ACTIONS(7405),
    [anon_sym__Nullable_result] = ACTIONS(7405),
    [anon_sym__Null_unspecified] = ACTIONS(7405),
    [anon_sym___autoreleasing] = ACTIONS(7405),
    [anon_sym___block] = ACTIONS(7405),
    [anon_sym___bridge] = ACTIONS(7405),
    [anon_sym___bridge_retained] = ACTIONS(7405),
    [anon_sym___bridge_transfer] = ACTIONS(7405),
    [anon_sym___complex] = ACTIONS(7405),
    [anon_sym___const] = ACTIONS(7405),
    [anon_sym___imag] = ACTIONS(7405),
    [anon_sym___kindof] = ACTIONS(7405),
    [anon_sym___nonnull] = ACTIONS(7405),
    [anon_sym___nullable] = ACTIONS(7405),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7405),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7405),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7405),
    [anon_sym___real] = ACTIONS(7405),
    [anon_sym___strong] = ACTIONS(7405),
    [anon_sym___unsafe_unretained] = ACTIONS(7405),
    [anon_sym___unused] = ACTIONS(7405),
    [anon_sym___weak] = ACTIONS(7405),
    [anon_sym_alignas] = ACTIONS(7405),
    [anon_sym__Alignas] = ACTIONS(7405),
    [sym_primitive_type] = ACTIONS(7405),
    [anon_sym_enum] = ACTIONS(7405),
    [anon_sym_COLON] = ACTIONS(7407),
    [anon_sym_struct] = ACTIONS(7405),
    [anon_sym_union] = ACTIONS(7405),
    [anon_sym_in] = ACTIONS(7405),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7405),
    [anon_sym___typeof] = ACTIONS(7405),
    [anon_sym_typeof] = ACTIONS(7405),
    [anon_sym_BOOL] = ACTIONS(7405),
    [anon_sym_IMP] = ACTIONS(7405),
    [anon_sym_SEL] = ACTIONS(7405),
    [anon_sym_Class] = ACTIONS(7405),
    [anon_sym_id] = ACTIONS(7405),
    [anon_sym_out] = ACTIONS(7405),
    [anon_sym_inout] = ACTIONS(7405),
    [anon_sym_bycopy] = ACTIONS(7405),
    [anon_sym_byref] = ACTIONS(7405),
    [anon_sym_oneway] = ACTIONS(7405),
  },
  [STATE(3817)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3810),
    [sym_identifier] = ACTIONS(7411),
    [anon_sym_COMMA] = ACTIONS(7413),
    [anon_sym_RPAREN] = ACTIONS(7413),
    [anon_sym_LPAREN2] = ACTIONS(7413),
    [anon_sym_STAR] = ACTIONS(7413),
    [anon_sym_CARET] = ACTIONS(7413),
    [anon_sym_GT] = ACTIONS(7413),
    [anon_sym_LT] = ACTIONS(7413),
    [anon_sym___extension__] = ACTIONS(7411),
    [anon_sym___attribute__] = ACTIONS(7411),
    [anon_sym___attribute] = ACTIONS(7411),
    [anon_sym_noreturn] = ACTIONS(7411),
    [anon_sym_LBRACK] = ACTIONS(7413),
    [anon_sym___based] = ACTIONS(7411),
    [anon_sym_LBRACE] = ACTIONS(7413),
    [anon_sym_signed] = ACTIONS(8824),
    [anon_sym_unsigned] = ACTIONS(8824),
    [anon_sym_long] = ACTIONS(8824),
    [anon_sym_short] = ACTIONS(8824),
    [anon_sym_ATautoreleasepool] = ACTIONS(7413),
    [anon_sym_const] = ACTIONS(7411),
    [anon_sym_constexpr] = ACTIONS(7411),
    [anon_sym_volatile] = ACTIONS(7411),
    [anon_sym_restrict] = ACTIONS(7411),
    [anon_sym___restrict__] = ACTIONS(7411),
    [anon_sym__Atomic] = ACTIONS(7411),
    [anon_sym__Noreturn] = ACTIONS(7411),
    [anon_sym__Nonnull] = ACTIONS(7411),
    [anon_sym_nullable] = ACTIONS(7411),
    [anon_sym__Complex] = ACTIONS(7411),
    [anon_sym__Nullable] = ACTIONS(7411),
    [anon_sym__Nullable_result] = ACTIONS(7411),
    [anon_sym__Null_unspecified] = ACTIONS(7411),
    [anon_sym___autoreleasing] = ACTIONS(7411),
    [anon_sym___block] = ACTIONS(7411),
    [anon_sym___bridge] = ACTIONS(7411),
    [anon_sym___bridge_retained] = ACTIONS(7411),
    [anon_sym___bridge_transfer] = ACTIONS(7411),
    [anon_sym___complex] = ACTIONS(7411),
    [anon_sym___const] = ACTIONS(7411),
    [anon_sym___imag] = ACTIONS(7411),
    [anon_sym___kindof] = ACTIONS(7411),
    [anon_sym___nonnull] = ACTIONS(7411),
    [anon_sym___nullable] = ACTIONS(7411),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7411),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7411),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7411),
    [anon_sym___real] = ACTIONS(7411),
    [anon_sym___strong] = ACTIONS(7411),
    [anon_sym___unsafe_unretained] = ACTIONS(7411),
    [anon_sym___unused] = ACTIONS(7411),
    [anon_sym___weak] = ACTIONS(7411),
    [anon_sym_alignas] = ACTIONS(7411),
    [anon_sym__Alignas] = ACTIONS(7411),
    [sym_primitive_type] = ACTIONS(7411),
    [anon_sym_enum] = ACTIONS(7411),
    [anon_sym_COLON] = ACTIONS(7413),
    [anon_sym_struct] = ACTIONS(7411),
    [anon_sym_union] = ACTIONS(7411),
    [anon_sym_in] = ACTIONS(7411),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7411),
    [anon_sym___typeof] = ACTIONS(7411),
    [anon_sym_typeof] = ACTIONS(7411),
    [anon_sym_BOOL] = ACTIONS(7411),
    [anon_sym_IMP] = ACTIONS(7411),
    [anon_sym_SEL] = ACTIONS(7411),
    [anon_sym_Class] = ACTIONS(7411),
    [anon_sym_id] = ACTIONS(7411),
    [anon_sym_out] = ACTIONS(7411),
    [anon_sym_inout] = ACTIONS(7411),
    [anon_sym_bycopy] = ACTIONS(7411),
    [anon_sym_byref] = ACTIONS(7411),
    [anon_sym_oneway] = ACTIONS(7411),
  },
  [STATE(3818)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3792),
    [sym_identifier] = ACTIONS(7421),
    [anon_sym_COMMA] = ACTIONS(7423),
    [anon_sym_RPAREN] = ACTIONS(7423),
    [anon_sym_LPAREN2] = ACTIONS(7423),
    [anon_sym_STAR] = ACTIONS(7423),
    [anon_sym_CARET] = ACTIONS(7423),
    [anon_sym_GT] = ACTIONS(7423),
    [anon_sym_LT] = ACTIONS(7423),
    [anon_sym___extension__] = ACTIONS(7421),
    [anon_sym___attribute__] = ACTIONS(7421),
    [anon_sym___attribute] = ACTIONS(7421),
    [anon_sym_noreturn] = ACTIONS(7421),
    [anon_sym_LBRACK] = ACTIONS(7423),
    [anon_sym___based] = ACTIONS(7421),
    [anon_sym_LBRACE] = ACTIONS(7423),
    [anon_sym_signed] = ACTIONS(8827),
    [anon_sym_unsigned] = ACTIONS(8827),
    [anon_sym_long] = ACTIONS(8827),
    [anon_sym_short] = ACTIONS(8827),
    [anon_sym_ATautoreleasepool] = ACTIONS(7423),
    [anon_sym_const] = ACTIONS(7421),
    [anon_sym_constexpr] = ACTIONS(7421),
    [anon_sym_volatile] = ACTIONS(7421),
    [anon_sym_restrict] = ACTIONS(7421),
    [anon_sym___restrict__] = ACTIONS(7421),
    [anon_sym__Atomic] = ACTIONS(7421),
    [anon_sym__Noreturn] = ACTIONS(7421),
    [anon_sym__Nonnull] = ACTIONS(7421),
    [anon_sym_nullable] = ACTIONS(7421),
    [anon_sym__Complex] = ACTIONS(7421),
    [anon_sym__Nullable] = ACTIONS(7421),
    [anon_sym__Nullable_result] = ACTIONS(7421),
    [anon_sym__Null_unspecified] = ACTIONS(7421),
    [anon_sym___autoreleasing] = ACTIONS(7421),
    [anon_sym___block] = ACTIONS(7421),
    [anon_sym___bridge] = ACTIONS(7421),
    [anon_sym___bridge_retained] = ACTIONS(7421),
    [anon_sym___bridge_transfer] = ACTIONS(7421),
    [anon_sym___complex] = ACTIONS(7421),
    [anon_sym___const] = ACTIONS(7421),
    [anon_sym___imag] = ACTIONS(7421),
    [anon_sym___kindof] = ACTIONS(7421),
    [anon_sym___nonnull] = ACTIONS(7421),
    [anon_sym___nullable] = ACTIONS(7421),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7421),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7421),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7421),
    [anon_sym___real] = ACTIONS(7421),
    [anon_sym___strong] = ACTIONS(7421),
    [anon_sym___unsafe_unretained] = ACTIONS(7421),
    [anon_sym___unused] = ACTIONS(7421),
    [anon_sym___weak] = ACTIONS(7421),
    [anon_sym_alignas] = ACTIONS(7421),
    [anon_sym__Alignas] = ACTIONS(7421),
    [sym_primitive_type] = ACTIONS(7421),
    [anon_sym_enum] = ACTIONS(7421),
    [anon_sym_COLON] = ACTIONS(7423),
    [anon_sym_struct] = ACTIONS(7421),
    [anon_sym_union] = ACTIONS(7421),
    [anon_sym_in] = ACTIONS(7421),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7421),
    [anon_sym___typeof] = ACTIONS(7421),
    [anon_sym_typeof] = ACTIONS(7421),
    [anon_sym_BOOL] = ACTIONS(7421),
    [anon_sym_IMP] = ACTIONS(7421),
    [anon_sym_SEL] = ACTIONS(7421),
    [anon_sym_Class] = ACTIONS(7421),
    [anon_sym_id] = ACTIONS(7421),
    [anon_sym_out] = ACTIONS(7421),
    [anon_sym_inout] = ACTIONS(7421),
    [anon_sym_bycopy] = ACTIONS(7421),
    [anon_sym_byref] = ACTIONS(7421),
    [anon_sym_oneway] = ACTIONS(7421),
  },
  [STATE(3819)] = {
    [sym_attribute_specifier] = STATE(3963),
    [sym_enumerator_list] = STATE(3844),
    [sym_identifier] = ACTIONS(7581),
    [anon_sym_COMMA] = ACTIONS(7583),
    [anon_sym_RPAREN] = ACTIONS(7583),
    [anon_sym_LPAREN2] = ACTIONS(7583),
    [anon_sym_STAR] = ACTIONS(7583),
    [anon_sym_CARET] = ACTIONS(7583),
    [anon_sym_GT] = ACTIONS(7583),
    [anon_sym_LT] = ACTIONS(7583),
    [anon_sym___extension__] = ACTIONS(7581),
    [anon_sym___attribute__] = ACTIONS(8830),
    [anon_sym___attribute] = ACTIONS(8830),
    [anon_sym_noreturn] = ACTIONS(7581),
    [anon_sym_LBRACK] = ACTIONS(7583),
    [anon_sym___based] = ACTIONS(7581),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7581),
    [anon_sym_unsigned] = ACTIONS(7581),
    [anon_sym_long] = ACTIONS(7581),
    [anon_sym_short] = ACTIONS(7581),
    [anon_sym_const] = ACTIONS(7581),
    [anon_sym_constexpr] = ACTIONS(7581),
    [anon_sym_volatile] = ACTIONS(7581),
    [anon_sym_restrict] = ACTIONS(7581),
    [anon_sym___restrict__] = ACTIONS(7581),
    [anon_sym__Atomic] = ACTIONS(7581),
    [anon_sym__Noreturn] = ACTIONS(7581),
    [anon_sym__Nonnull] = ACTIONS(7581),
    [anon_sym_nullable] = ACTIONS(7581),
    [anon_sym__Complex] = ACTIONS(7581),
    [anon_sym__Nullable] = ACTIONS(7581),
    [anon_sym__Nullable_result] = ACTIONS(7581),
    [anon_sym__Null_unspecified] = ACTIONS(7581),
    [anon_sym___autoreleasing] = ACTIONS(7581),
    [anon_sym___block] = ACTIONS(7581),
    [anon_sym___bridge] = ACTIONS(7581),
    [anon_sym___bridge_retained] = ACTIONS(7581),
    [anon_sym___bridge_transfer] = ACTIONS(7581),
    [anon_sym___complex] = ACTIONS(7581),
    [anon_sym___const] = ACTIONS(7581),
    [anon_sym___imag] = ACTIONS(7581),
    [anon_sym___kindof] = ACTIONS(7581),
    [anon_sym___nonnull] = ACTIONS(7581),
    [anon_sym___nullable] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7581),
    [anon_sym___real] = ACTIONS(7581),
    [anon_sym___strong] = ACTIONS(7581),
    [anon_sym___unsafe_unretained] = ACTIONS(7581),
    [anon_sym___unused] = ACTIONS(7581),
    [anon_sym___weak] = ACTIONS(7581),
    [anon_sym_alignas] = ACTIONS(7581),
    [anon_sym__Alignas] = ACTIONS(7581),
    [sym_primitive_type] = ACTIONS(7581),
    [anon_sym_enum] = ACTIONS(7581),
    [anon_sym_COLON] = ACTIONS(8833),
    [anon_sym_struct] = ACTIONS(7581),
    [anon_sym_union] = ACTIONS(7581),
    [anon_sym_in] = ACTIONS(7581),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7581),
    [anon_sym___typeof] = ACTIONS(7581),
    [anon_sym_typeof] = ACTIONS(7581),
    [anon_sym_BOOL] = ACTIONS(7581),
    [anon_sym_IMP] = ACTIONS(7581),
    [anon_sym_SEL] = ACTIONS(7581),
    [anon_sym_Class] = ACTIONS(7581),
    [anon_sym_id] = ACTIONS(7581),
    [anon_sym_out] = ACTIONS(7581),
    [anon_sym_inout] = ACTIONS(7581),
    [anon_sym_bycopy] = ACTIONS(7581),
    [anon_sym_byref] = ACTIONS(7581),
    [anon_sym_oneway] = ACTIONS(7581),
  },
  [STATE(3820)] = {
    [sym_attribute_specifier] = STATE(3860),
    [sym_enumerator_list] = STATE(3890),
    [sym_identifier] = ACTIONS(7632),
    [anon_sym_COMMA] = ACTIONS(7634),
    [anon_sym_RPAREN] = ACTIONS(7634),
    [anon_sym_LPAREN2] = ACTIONS(7634),
    [anon_sym_STAR] = ACTIONS(7634),
    [anon_sym_CARET] = ACTIONS(7634),
    [anon_sym_GT] = ACTIONS(7634),
    [anon_sym_LT] = ACTIONS(7634),
    [anon_sym___extension__] = ACTIONS(7632),
    [anon_sym___attribute__] = ACTIONS(8835),
    [anon_sym___attribute] = ACTIONS(8835),
    [anon_sym_noreturn] = ACTIONS(7632),
    [anon_sym_LBRACK] = ACTIONS(7634),
    [anon_sym___based] = ACTIONS(7632),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7632),
    [anon_sym_unsigned] = ACTIONS(7632),
    [anon_sym_long] = ACTIONS(7632),
    [anon_sym_short] = ACTIONS(7632),
    [anon_sym_const] = ACTIONS(7632),
    [anon_sym_constexpr] = ACTIONS(7632),
    [anon_sym_volatile] = ACTIONS(7632),
    [anon_sym_restrict] = ACTIONS(7632),
    [anon_sym___restrict__] = ACTIONS(7632),
    [anon_sym__Atomic] = ACTIONS(7632),
    [anon_sym__Noreturn] = ACTIONS(7632),
    [anon_sym__Nonnull] = ACTIONS(7632),
    [anon_sym_nullable] = ACTIONS(7632),
    [anon_sym__Complex] = ACTIONS(7632),
    [anon_sym__Nullable] = ACTIONS(7632),
    [anon_sym__Nullable_result] = ACTIONS(7632),
    [anon_sym__Null_unspecified] = ACTIONS(7632),
    [anon_sym___autoreleasing] = ACTIONS(7632),
    [anon_sym___block] = ACTIONS(7632),
    [anon_sym___bridge] = ACTIONS(7632),
    [anon_sym___bridge_retained] = ACTIONS(7632),
    [anon_sym___bridge_transfer] = ACTIONS(7632),
    [anon_sym___complex] = ACTIONS(7632),
    [anon_sym___const] = ACTIONS(7632),
    [anon_sym___imag] = ACTIONS(7632),
    [anon_sym___kindof] = ACTIONS(7632),
    [anon_sym___nonnull] = ACTIONS(7632),
    [anon_sym___nullable] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7632),
    [anon_sym___real] = ACTIONS(7632),
    [anon_sym___strong] = ACTIONS(7632),
    [anon_sym___unsafe_unretained] = ACTIONS(7632),
    [anon_sym___unused] = ACTIONS(7632),
    [anon_sym___weak] = ACTIONS(7632),
    [anon_sym_alignas] = ACTIONS(7632),
    [anon_sym__Alignas] = ACTIONS(7632),
    [sym_primitive_type] = ACTIONS(7632),
    [anon_sym_enum] = ACTIONS(7632),
    [anon_sym_COLON] = ACTIONS(8838),
    [anon_sym_struct] = ACTIONS(7632),
    [anon_sym_union] = ACTIONS(7632),
    [anon_sym_in] = ACTIONS(7632),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7632),
    [anon_sym___typeof] = ACTIONS(7632),
    [anon_sym_typeof] = ACTIONS(7632),
    [anon_sym_BOOL] = ACTIONS(7632),
    [anon_sym_IMP] = ACTIONS(7632),
    [anon_sym_SEL] = ACTIONS(7632),
    [anon_sym_Class] = ACTIONS(7632),
    [anon_sym_id] = ACTIONS(7632),
    [anon_sym_out] = ACTIONS(7632),
    [anon_sym_inout] = ACTIONS(7632),
    [anon_sym_bycopy] = ACTIONS(7632),
    [anon_sym_byref] = ACTIONS(7632),
    [anon_sym_oneway] = ACTIONS(7632),
  },
  [STATE(3821)] = {
    [sym_attribute_specifier] = STATE(3965),
    [sym_identifier] = ACTIONS(7149),
    [anon_sym_COMMA] = ACTIONS(7151),
    [anon_sym_RPAREN] = ACTIONS(7151),
    [anon_sym_LPAREN2] = ACTIONS(7151),
    [anon_sym_STAR] = ACTIONS(7151),
    [anon_sym_CARET] = ACTIONS(7151),
    [anon_sym_GT] = ACTIONS(7151),
    [anon_sym_LT] = ACTIONS(7151),
    [anon_sym___extension__] = ACTIONS(7149),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7149),
    [anon_sym_LBRACK] = ACTIONS(7151),
    [anon_sym___based] = ACTIONS(7149),
    [anon_sym_LBRACE] = ACTIONS(7151),
    [anon_sym_signed] = ACTIONS(7149),
    [anon_sym_unsigned] = ACTIONS(7149),
    [anon_sym_long] = ACTIONS(7149),
    [anon_sym_short] = ACTIONS(7149),
    [anon_sym_ATautoreleasepool] = ACTIONS(7151),
    [anon_sym_const] = ACTIONS(7149),
    [anon_sym_constexpr] = ACTIONS(7149),
    [anon_sym_volatile] = ACTIONS(7149),
    [anon_sym_restrict] = ACTIONS(7149),
    [anon_sym___restrict__] = ACTIONS(7149),
    [anon_sym__Atomic] = ACTIONS(7149),
    [anon_sym__Noreturn] = ACTIONS(7149),
    [anon_sym__Nonnull] = ACTIONS(7149),
    [anon_sym_nullable] = ACTIONS(7149),
    [anon_sym__Complex] = ACTIONS(7149),
    [anon_sym__Nullable] = ACTIONS(7149),
    [anon_sym__Nullable_result] = ACTIONS(7149),
    [anon_sym__Null_unspecified] = ACTIONS(7149),
    [anon_sym___autoreleasing] = ACTIONS(7149),
    [anon_sym___block] = ACTIONS(7149),
    [anon_sym___bridge] = ACTIONS(7149),
    [anon_sym___bridge_retained] = ACTIONS(7149),
    [anon_sym___bridge_transfer] = ACTIONS(7149),
    [anon_sym___complex] = ACTIONS(7149),
    [anon_sym___const] = ACTIONS(7149),
    [anon_sym___imag] = ACTIONS(7149),
    [anon_sym___kindof] = ACTIONS(7149),
    [anon_sym___nonnull] = ACTIONS(7149),
    [anon_sym___nullable] = ACTIONS(7149),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7149),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7149),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7149),
    [anon_sym___real] = ACTIONS(7149),
    [anon_sym___strong] = ACTIONS(7149),
    [anon_sym___unsafe_unretained] = ACTIONS(7149),
    [anon_sym___unused] = ACTIONS(7149),
    [anon_sym___weak] = ACTIONS(7149),
    [anon_sym_alignas] = ACTIONS(7149),
    [anon_sym__Alignas] = ACTIONS(7149),
    [sym_primitive_type] = ACTIONS(7149),
    [anon_sym_enum] = ACTIONS(7149),
    [anon_sym_COLON] = ACTIONS(7151),
    [anon_sym_struct] = ACTIONS(7149),
    [anon_sym_union] = ACTIONS(7149),
    [anon_sym_in] = ACTIONS(7149),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7149),
    [anon_sym___typeof] = ACTIONS(7149),
    [anon_sym_typeof] = ACTIONS(7149),
    [anon_sym_BOOL] = ACTIONS(7149),
    [anon_sym_IMP] = ACTIONS(7149),
    [anon_sym_SEL] = ACTIONS(7149),
    [anon_sym_Class] = ACTIONS(7149),
    [anon_sym_id] = ACTIONS(7149),
    [anon_sym_out] = ACTIONS(7149),
    [anon_sym_inout] = ACTIONS(7149),
    [anon_sym_bycopy] = ACTIONS(7149),
    [anon_sym_byref] = ACTIONS(7149),
    [anon_sym_oneway] = ACTIONS(7149),
  },
  [STATE(3822)] = {
    [sym_attribute_specifier] = STATE(3960),
    [sym_identifier] = ACTIONS(7072),
    [anon_sym_COMMA] = ACTIONS(7074),
    [anon_sym_RPAREN] = ACTIONS(7074),
    [anon_sym_LPAREN2] = ACTIONS(7074),
    [anon_sym_STAR] = ACTIONS(7074),
    [anon_sym_CARET] = ACTIONS(7074),
    [anon_sym_GT] = ACTIONS(7074),
    [anon_sym_LT] = ACTIONS(7074),
    [anon_sym___extension__] = ACTIONS(7072),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7072),
    [anon_sym_LBRACK] = ACTIONS(7074),
    [anon_sym___based] = ACTIONS(7072),
    [anon_sym_LBRACE] = ACTIONS(7074),
    [anon_sym_signed] = ACTIONS(7072),
    [anon_sym_unsigned] = ACTIONS(7072),
    [anon_sym_long] = ACTIONS(7072),
    [anon_sym_short] = ACTIONS(7072),
    [anon_sym_ATautoreleasepool] = ACTIONS(7074),
    [anon_sym_const] = ACTIONS(7072),
    [anon_sym_constexpr] = ACTIONS(7072),
    [anon_sym_volatile] = ACTIONS(7072),
    [anon_sym_restrict] = ACTIONS(7072),
    [anon_sym___restrict__] = ACTIONS(7072),
    [anon_sym__Atomic] = ACTIONS(7072),
    [anon_sym__Noreturn] = ACTIONS(7072),
    [anon_sym__Nonnull] = ACTIONS(7072),
    [anon_sym_nullable] = ACTIONS(7072),
    [anon_sym__Complex] = ACTIONS(7072),
    [anon_sym__Nullable] = ACTIONS(7072),
    [anon_sym__Nullable_result] = ACTIONS(7072),
    [anon_sym__Null_unspecified] = ACTIONS(7072),
    [anon_sym___autoreleasing] = ACTIONS(7072),
    [anon_sym___block] = ACTIONS(7072),
    [anon_sym___bridge] = ACTIONS(7072),
    [anon_sym___bridge_retained] = ACTIONS(7072),
    [anon_sym___bridge_transfer] = ACTIONS(7072),
    [anon_sym___complex] = ACTIONS(7072),
    [anon_sym___const] = ACTIONS(7072),
    [anon_sym___imag] = ACTIONS(7072),
    [anon_sym___kindof] = ACTIONS(7072),
    [anon_sym___nonnull] = ACTIONS(7072),
    [anon_sym___nullable] = ACTIONS(7072),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7072),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7072),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7072),
    [anon_sym___real] = ACTIONS(7072),
    [anon_sym___strong] = ACTIONS(7072),
    [anon_sym___unsafe_unretained] = ACTIONS(7072),
    [anon_sym___unused] = ACTIONS(7072),
    [anon_sym___weak] = ACTIONS(7072),
    [anon_sym_alignas] = ACTIONS(7072),
    [anon_sym__Alignas] = ACTIONS(7072),
    [sym_primitive_type] = ACTIONS(7072),
    [anon_sym_enum] = ACTIONS(7072),
    [anon_sym_COLON] = ACTIONS(7074),
    [anon_sym_struct] = ACTIONS(7072),
    [anon_sym_union] = ACTIONS(7072),
    [anon_sym_in] = ACTIONS(7072),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7072),
    [anon_sym___typeof] = ACTIONS(7072),
    [anon_sym_typeof] = ACTIONS(7072),
    [anon_sym_BOOL] = ACTIONS(7072),
    [anon_sym_IMP] = ACTIONS(7072),
    [anon_sym_SEL] = ACTIONS(7072),
    [anon_sym_Class] = ACTIONS(7072),
    [anon_sym_id] = ACTIONS(7072),
    [anon_sym_out] = ACTIONS(7072),
    [anon_sym_inout] = ACTIONS(7072),
    [anon_sym_bycopy] = ACTIONS(7072),
    [anon_sym_byref] = ACTIONS(7072),
    [anon_sym_oneway] = ACTIONS(7072),
  },
  [STATE(3823)] = {
    [sym_attribute_specifier] = STATE(3803),
    [aux_sym_function_declarator_repeat1] = STATE(3803),
    [sym_identifier] = ACTIONS(6696),
    [anon_sym_COMMA] = ACTIONS(6698),
    [anon_sym_RPAREN] = ACTIONS(6698),
    [anon_sym_LPAREN2] = ACTIONS(6698),
    [anon_sym_STAR] = ACTIONS(6698),
    [anon_sym_CARET] = ACTIONS(6698),
    [anon_sym_GT] = ACTIONS(6698),
    [anon_sym_LT] = ACTIONS(6698),
    [anon_sym___extension__] = ACTIONS(6696),
    [anon_sym___attribute__] = ACTIONS(6696),
    [anon_sym___attribute] = ACTIONS(6696),
    [anon_sym_noreturn] = ACTIONS(6696),
    [anon_sym_LBRACK] = ACTIONS(6698),
    [anon_sym___based] = ACTIONS(6696),
    [anon_sym_LBRACE] = ACTIONS(6698),
    [anon_sym_signed] = ACTIONS(6696),
    [anon_sym_unsigned] = ACTIONS(6696),
    [anon_sym_long] = ACTIONS(6696),
    [anon_sym_short] = ACTIONS(6696),
    [anon_sym_ATautoreleasepool] = ACTIONS(6698),
    [anon_sym_const] = ACTIONS(6696),
    [anon_sym_constexpr] = ACTIONS(6696),
    [anon_sym_volatile] = ACTIONS(6696),
    [anon_sym_restrict] = ACTIONS(6696),
    [anon_sym___restrict__] = ACTIONS(6696),
    [anon_sym__Atomic] = ACTIONS(6696),
    [anon_sym__Noreturn] = ACTIONS(6696),
    [anon_sym__Nonnull] = ACTIONS(6696),
    [anon_sym_nullable] = ACTIONS(6696),
    [anon_sym__Complex] = ACTIONS(6696),
    [anon_sym__Nullable] = ACTIONS(6696),
    [anon_sym__Nullable_result] = ACTIONS(6696),
    [anon_sym__Null_unspecified] = ACTIONS(6696),
    [anon_sym___autoreleasing] = ACTIONS(6696),
    [anon_sym___block] = ACTIONS(6696),
    [anon_sym___bridge] = ACTIONS(6696),
    [anon_sym___bridge_retained] = ACTIONS(6696),
    [anon_sym___bridge_transfer] = ACTIONS(6696),
    [anon_sym___complex] = ACTIONS(6696),
    [anon_sym___const] = ACTIONS(6696),
    [anon_sym___imag] = ACTIONS(6696),
    [anon_sym___kindof] = ACTIONS(6696),
    [anon_sym___nonnull] = ACTIONS(6696),
    [anon_sym___nullable] = ACTIONS(6696),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6696),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6696),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6696),
    [anon_sym___real] = ACTIONS(6696),
    [anon_sym___strong] = ACTIONS(6696),
    [anon_sym___unsafe_unretained] = ACTIONS(6696),
    [anon_sym___unused] = ACTIONS(6696),
    [anon_sym___weak] = ACTIONS(6696),
    [anon_sym_alignas] = ACTIONS(6696),
    [anon_sym__Alignas] = ACTIONS(6696),
    [sym_primitive_type] = ACTIONS(6696),
    [anon_sym_enum] = ACTIONS(6696),
    [anon_sym_struct] = ACTIONS(6696),
    [anon_sym_union] = ACTIONS(6696),
    [anon_sym_in] = ACTIONS(6696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6696),
    [anon_sym___typeof] = ACTIONS(6696),
    [anon_sym_typeof] = ACTIONS(6696),
    [anon_sym_BOOL] = ACTIONS(6696),
    [anon_sym_IMP] = ACTIONS(6696),
    [anon_sym_SEL] = ACTIONS(6696),
    [anon_sym_Class] = ACTIONS(6696),
    [anon_sym_id] = ACTIONS(6696),
    [anon_sym_out] = ACTIONS(6696),
    [anon_sym_inout] = ACTIONS(6696),
    [anon_sym_bycopy] = ACTIONS(6696),
    [anon_sym_byref] = ACTIONS(6696),
    [anon_sym_oneway] = ACTIONS(6696),
  },
  [STATE(3824)] = {
    [sym_identifier] = ACTIONS(7361),
    [anon_sym_COMMA] = ACTIONS(7363),
    [anon_sym_RPAREN] = ACTIONS(7363),
    [anon_sym_LPAREN2] = ACTIONS(7363),
    [anon_sym_STAR] = ACTIONS(7363),
    [anon_sym_CARET] = ACTIONS(7363),
    [anon_sym_GT] = ACTIONS(7363),
    [anon_sym_LT] = ACTIONS(7363),
    [anon_sym___extension__] = ACTIONS(7361),
    [anon_sym___attribute__] = ACTIONS(7361),
    [anon_sym___attribute] = ACTIONS(7361),
    [anon_sym_noreturn] = ACTIONS(7361),
    [anon_sym_LBRACK] = ACTIONS(7363),
    [anon_sym___based] = ACTIONS(7361),
    [anon_sym_LBRACE] = ACTIONS(7363),
    [anon_sym_signed] = ACTIONS(7361),
    [anon_sym_unsigned] = ACTIONS(7361),
    [anon_sym_long] = ACTIONS(7361),
    [anon_sym_short] = ACTIONS(7361),
    [anon_sym_ATautoreleasepool] = ACTIONS(7363),
    [anon_sym_const] = ACTIONS(7361),
    [anon_sym_constexpr] = ACTIONS(7361),
    [anon_sym_volatile] = ACTIONS(7361),
    [anon_sym_restrict] = ACTIONS(7361),
    [anon_sym___restrict__] = ACTIONS(7361),
    [anon_sym__Atomic] = ACTIONS(7361),
    [anon_sym__Noreturn] = ACTIONS(7361),
    [anon_sym__Nonnull] = ACTIONS(7361),
    [anon_sym_nullable] = ACTIONS(7361),
    [anon_sym__Complex] = ACTIONS(7361),
    [anon_sym__Nullable] = ACTIONS(7361),
    [anon_sym__Nullable_result] = ACTIONS(7361),
    [anon_sym__Null_unspecified] = ACTIONS(7361),
    [anon_sym___autoreleasing] = ACTIONS(7361),
    [anon_sym___block] = ACTIONS(7361),
    [anon_sym___bridge] = ACTIONS(7361),
    [anon_sym___bridge_retained] = ACTIONS(7361),
    [anon_sym___bridge_transfer] = ACTIONS(7361),
    [anon_sym___complex] = ACTIONS(7361),
    [anon_sym___const] = ACTIONS(7361),
    [anon_sym___imag] = ACTIONS(7361),
    [anon_sym___kindof] = ACTIONS(7361),
    [anon_sym___nonnull] = ACTIONS(7361),
    [anon_sym___nullable] = ACTIONS(7361),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7361),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7361),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7361),
    [anon_sym___real] = ACTIONS(7361),
    [anon_sym___strong] = ACTIONS(7361),
    [anon_sym___unsafe_unretained] = ACTIONS(7361),
    [anon_sym___unused] = ACTIONS(7361),
    [anon_sym___weak] = ACTIONS(7361),
    [anon_sym_alignas] = ACTIONS(7361),
    [anon_sym__Alignas] = ACTIONS(7361),
    [sym_primitive_type] = ACTIONS(7361),
    [anon_sym_enum] = ACTIONS(7361),
    [anon_sym_COLON] = ACTIONS(7363),
    [anon_sym_struct] = ACTIONS(7361),
    [anon_sym_union] = ACTIONS(7361),
    [anon_sym_in] = ACTIONS(7361),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7361),
    [anon_sym___typeof] = ACTIONS(7361),
    [anon_sym_typeof] = ACTIONS(7361),
    [anon_sym_BOOL] = ACTIONS(7361),
    [anon_sym_IMP] = ACTIONS(7361),
    [anon_sym_SEL] = ACTIONS(7361),
    [anon_sym_Class] = ACTIONS(7361),
    [anon_sym_id] = ACTIONS(7361),
    [anon_sym_out] = ACTIONS(7361),
    [anon_sym_inout] = ACTIONS(7361),
    [anon_sym_bycopy] = ACTIONS(7361),
    [anon_sym_byref] = ACTIONS(7361),
    [anon_sym_oneway] = ACTIONS(7361),
  },
  [STATE(3825)] = {
    [sym__type_definition_type] = STATE(4747),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4001),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3826)] = {
    [sym__type_definition_type] = STATE(4750),
    [sym_attribute_specifier] = STATE(3828),
    [sym_ms_declspec_modifier] = STATE(4001),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3827)] = {
    [sym_identifier] = ACTIONS(8846),
    [anon_sym_COMMA] = ACTIONS(8848),
    [anon_sym_RPAREN] = ACTIONS(8848),
    [anon_sym_LPAREN2] = ACTIONS(8848),
    [anon_sym_STAR] = ACTIONS(8848),
    [anon_sym_CARET] = ACTIONS(8848),
    [anon_sym_GT] = ACTIONS(8848),
    [anon_sym_LT] = ACTIONS(8848),
    [anon_sym___extension__] = ACTIONS(8846),
    [anon_sym___attribute__] = ACTIONS(8846),
    [anon_sym___attribute] = ACTIONS(8846),
    [anon_sym_noreturn] = ACTIONS(8846),
    [anon_sym_LBRACK] = ACTIONS(8848),
    [anon_sym___based] = ACTIONS(8846),
    [anon_sym_LBRACE] = ACTIONS(8848),
    [anon_sym_signed] = ACTIONS(8846),
    [anon_sym_unsigned] = ACTIONS(8846),
    [anon_sym_long] = ACTIONS(8846),
    [anon_sym_short] = ACTIONS(8846),
    [anon_sym_ATautoreleasepool] = ACTIONS(8848),
    [anon_sym_const] = ACTIONS(8846),
    [anon_sym_constexpr] = ACTIONS(8846),
    [anon_sym_volatile] = ACTIONS(8846),
    [anon_sym_restrict] = ACTIONS(8846),
    [anon_sym___restrict__] = ACTIONS(8846),
    [anon_sym__Atomic] = ACTIONS(8846),
    [anon_sym__Noreturn] = ACTIONS(8846),
    [anon_sym__Nonnull] = ACTIONS(8846),
    [anon_sym_nullable] = ACTIONS(8846),
    [anon_sym__Complex] = ACTIONS(8846),
    [anon_sym__Nullable] = ACTIONS(8846),
    [anon_sym__Nullable_result] = ACTIONS(8846),
    [anon_sym__Null_unspecified] = ACTIONS(8846),
    [anon_sym___autoreleasing] = ACTIONS(8846),
    [anon_sym___block] = ACTIONS(8846),
    [anon_sym___bridge] = ACTIONS(8846),
    [anon_sym___bridge_retained] = ACTIONS(8846),
    [anon_sym___bridge_transfer] = ACTIONS(8846),
    [anon_sym___complex] = ACTIONS(8846),
    [anon_sym___const] = ACTIONS(8846),
    [anon_sym___imag] = ACTIONS(8846),
    [anon_sym___kindof] = ACTIONS(8846),
    [anon_sym___nonnull] = ACTIONS(8846),
    [anon_sym___nullable] = ACTIONS(8846),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8846),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8846),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8846),
    [anon_sym___real] = ACTIONS(8846),
    [anon_sym___strong] = ACTIONS(8846),
    [anon_sym___unsafe_unretained] = ACTIONS(8846),
    [anon_sym___unused] = ACTIONS(8846),
    [anon_sym___weak] = ACTIONS(8846),
    [anon_sym_alignas] = ACTIONS(8846),
    [anon_sym__Alignas] = ACTIONS(8846),
    [sym_primitive_type] = ACTIONS(8846),
    [anon_sym_enum] = ACTIONS(8846),
    [anon_sym_COLON] = ACTIONS(8848),
    [anon_sym_struct] = ACTIONS(8846),
    [anon_sym_union] = ACTIONS(8846),
    [anon_sym_in] = ACTIONS(8846),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8846),
    [anon_sym___typeof] = ACTIONS(8846),
    [anon_sym_typeof] = ACTIONS(8846),
    [anon_sym_BOOL] = ACTIONS(8846),
    [anon_sym_IMP] = ACTIONS(8846),
    [anon_sym_SEL] = ACTIONS(8846),
    [anon_sym_Class] = ACTIONS(8846),
    [anon_sym_id] = ACTIONS(8846),
    [anon_sym_out] = ACTIONS(8846),
    [anon_sym_inout] = ACTIONS(8846),
    [anon_sym_bycopy] = ACTIONS(8846),
    [anon_sym_byref] = ACTIONS(8846),
    [anon_sym_oneway] = ACTIONS(8846),
  },
  [STATE(3828)] = {
    [sym__type_definition_type] = STATE(4755),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4009),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3829)] = {
    [sym_parameter_list] = STATE(3764),
    [sym_identifier] = ACTIONS(6722),
    [anon_sym_COMMA] = ACTIONS(6724),
    [anon_sym_RPAREN] = ACTIONS(6724),
    [anon_sym_LPAREN2] = ACTIONS(8850),
    [anon_sym_STAR] = ACTIONS(6724),
    [anon_sym_CARET] = ACTIONS(6724),
    [anon_sym_GT] = ACTIONS(6724),
    [anon_sym_LT] = ACTIONS(6724),
    [anon_sym___extension__] = ACTIONS(6722),
    [anon_sym___attribute__] = ACTIONS(6722),
    [anon_sym___attribute] = ACTIONS(6722),
    [anon_sym_noreturn] = ACTIONS(6722),
    [anon_sym_LBRACK] = ACTIONS(8852),
    [anon_sym___based] = ACTIONS(6722),
    [anon_sym_LBRACE] = ACTIONS(6724),
    [anon_sym_signed] = ACTIONS(6722),
    [anon_sym_unsigned] = ACTIONS(6722),
    [anon_sym_long] = ACTIONS(6722),
    [anon_sym_short] = ACTIONS(6722),
    [anon_sym_ATautoreleasepool] = ACTIONS(6724),
    [anon_sym_const] = ACTIONS(6722),
    [anon_sym_constexpr] = ACTIONS(6722),
    [anon_sym_volatile] = ACTIONS(6722),
    [anon_sym_restrict] = ACTIONS(6722),
    [anon_sym___restrict__] = ACTIONS(6722),
    [anon_sym__Atomic] = ACTIONS(6722),
    [anon_sym__Noreturn] = ACTIONS(6722),
    [anon_sym__Nonnull] = ACTIONS(6722),
    [anon_sym_nullable] = ACTIONS(6722),
    [anon_sym__Complex] = ACTIONS(6722),
    [anon_sym__Nullable] = ACTIONS(6722),
    [anon_sym__Nullable_result] = ACTIONS(6722),
    [anon_sym__Null_unspecified] = ACTIONS(6722),
    [anon_sym___autoreleasing] = ACTIONS(6722),
    [anon_sym___block] = ACTIONS(6722),
    [anon_sym___bridge] = ACTIONS(6722),
    [anon_sym___bridge_retained] = ACTIONS(6722),
    [anon_sym___bridge_transfer] = ACTIONS(6722),
    [anon_sym___complex] = ACTIONS(6722),
    [anon_sym___const] = ACTIONS(6722),
    [anon_sym___imag] = ACTIONS(6722),
    [anon_sym___kindof] = ACTIONS(6722),
    [anon_sym___nonnull] = ACTIONS(6722),
    [anon_sym___nullable] = ACTIONS(6722),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6722),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6722),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6722),
    [anon_sym___real] = ACTIONS(6722),
    [anon_sym___strong] = ACTIONS(6722),
    [anon_sym___unsafe_unretained] = ACTIONS(6722),
    [anon_sym___unused] = ACTIONS(6722),
    [anon_sym___weak] = ACTIONS(6722),
    [anon_sym_alignas] = ACTIONS(6722),
    [anon_sym__Alignas] = ACTIONS(6722),
    [sym_primitive_type] = ACTIONS(6722),
    [anon_sym_enum] = ACTIONS(6722),
    [anon_sym_struct] = ACTIONS(6722),
    [anon_sym_union] = ACTIONS(6722),
    [anon_sym_in] = ACTIONS(6722),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6722),
    [anon_sym___typeof] = ACTIONS(6722),
    [anon_sym_typeof] = ACTIONS(6722),
    [anon_sym_BOOL] = ACTIONS(6722),
    [anon_sym_IMP] = ACTIONS(6722),
    [anon_sym_SEL] = ACTIONS(6722),
    [anon_sym_Class] = ACTIONS(6722),
    [anon_sym_id] = ACTIONS(6722),
    [anon_sym_out] = ACTIONS(6722),
    [anon_sym_inout] = ACTIONS(6722),
    [anon_sym_bycopy] = ACTIONS(6722),
    [anon_sym_byref] = ACTIONS(6722),
    [anon_sym_oneway] = ACTIONS(6722),
  },
  [STATE(3830)] = {
    [sym_identifier] = ACTIONS(4856),
    [anon_sym_COMMA] = ACTIONS(4858),
    [anon_sym_RPAREN] = ACTIONS(4858),
    [anon_sym_LPAREN2] = ACTIONS(4858),
    [anon_sym_STAR] = ACTIONS(4858),
    [anon_sym_CARET] = ACTIONS(4858),
    [anon_sym_GT] = ACTIONS(4858),
    [anon_sym_LT] = ACTIONS(4858),
    [anon_sym___extension__] = ACTIONS(4856),
    [anon_sym___attribute__] = ACTIONS(4856),
    [anon_sym___attribute] = ACTIONS(4856),
    [anon_sym_noreturn] = ACTIONS(4856),
    [anon_sym_LBRACK] = ACTIONS(4858),
    [anon_sym___based] = ACTIONS(4856),
    [anon_sym_LBRACE] = ACTIONS(4858),
    [anon_sym_signed] = ACTIONS(4856),
    [anon_sym_unsigned] = ACTIONS(4856),
    [anon_sym_long] = ACTIONS(4856),
    [anon_sym_short] = ACTIONS(4856),
    [anon_sym_ATautoreleasepool] = ACTIONS(4858),
    [anon_sym_const] = ACTIONS(4856),
    [anon_sym_constexpr] = ACTIONS(4856),
    [anon_sym_volatile] = ACTIONS(4856),
    [anon_sym_restrict] = ACTIONS(4856),
    [anon_sym___restrict__] = ACTIONS(4856),
    [anon_sym__Atomic] = ACTIONS(4856),
    [anon_sym__Noreturn] = ACTIONS(4856),
    [anon_sym__Nonnull] = ACTIONS(4856),
    [anon_sym_nullable] = ACTIONS(4856),
    [anon_sym__Complex] = ACTIONS(4856),
    [anon_sym__Nullable] = ACTIONS(4856),
    [anon_sym__Nullable_result] = ACTIONS(4856),
    [anon_sym__Null_unspecified] = ACTIONS(4856),
    [anon_sym___autoreleasing] = ACTIONS(4856),
    [anon_sym___block] = ACTIONS(4856),
    [anon_sym___bridge] = ACTIONS(4856),
    [anon_sym___bridge_retained] = ACTIONS(4856),
    [anon_sym___bridge_transfer] = ACTIONS(4856),
    [anon_sym___complex] = ACTIONS(4856),
    [anon_sym___const] = ACTIONS(4856),
    [anon_sym___imag] = ACTIONS(4856),
    [anon_sym___kindof] = ACTIONS(4856),
    [anon_sym___nonnull] = ACTIONS(4856),
    [anon_sym___nullable] = ACTIONS(4856),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4856),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4856),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4856),
    [anon_sym___real] = ACTIONS(4856),
    [anon_sym___strong] = ACTIONS(4856),
    [anon_sym___unsafe_unretained] = ACTIONS(4856),
    [anon_sym___unused] = ACTIONS(4856),
    [anon_sym___weak] = ACTIONS(4856),
    [anon_sym_alignas] = ACTIONS(4856),
    [anon_sym__Alignas] = ACTIONS(4856),
    [sym_primitive_type] = ACTIONS(4856),
    [anon_sym_enum] = ACTIONS(4856),
    [anon_sym_COLON] = ACTIONS(4858),
    [anon_sym_struct] = ACTIONS(4856),
    [anon_sym_union] = ACTIONS(4856),
    [anon_sym_in] = ACTIONS(4856),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4856),
    [anon_sym___typeof] = ACTIONS(4856),
    [anon_sym_typeof] = ACTIONS(4856),
    [anon_sym_BOOL] = ACTIONS(4856),
    [anon_sym_IMP] = ACTIONS(4856),
    [anon_sym_SEL] = ACTIONS(4856),
    [anon_sym_Class] = ACTIONS(4856),
    [anon_sym_id] = ACTIONS(4856),
    [anon_sym_out] = ACTIONS(4856),
    [anon_sym_inout] = ACTIONS(4856),
    [anon_sym_bycopy] = ACTIONS(4856),
    [anon_sym_byref] = ACTIONS(4856),
    [anon_sym_oneway] = ACTIONS(4856),
  },
  [STATE(3831)] = {
    [sym_parameter_list] = STATE(3764),
    [sym_identifier] = ACTIONS(6736),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_RPAREN] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(8850),
    [anon_sym_STAR] = ACTIONS(6738),
    [anon_sym_CARET] = ACTIONS(6738),
    [anon_sym_GT] = ACTIONS(6738),
    [anon_sym_LT] = ACTIONS(6738),
    [anon_sym___extension__] = ACTIONS(6736),
    [anon_sym___attribute__] = ACTIONS(6736),
    [anon_sym___attribute] = ACTIONS(6736),
    [anon_sym_noreturn] = ACTIONS(6736),
    [anon_sym_LBRACK] = ACTIONS(8852),
    [anon_sym___based] = ACTIONS(6736),
    [anon_sym_LBRACE] = ACTIONS(6738),
    [anon_sym_signed] = ACTIONS(6736),
    [anon_sym_unsigned] = ACTIONS(6736),
    [anon_sym_long] = ACTIONS(6736),
    [anon_sym_short] = ACTIONS(6736),
    [anon_sym_ATautoreleasepool] = ACTIONS(6738),
    [anon_sym_const] = ACTIONS(6736),
    [anon_sym_constexpr] = ACTIONS(6736),
    [anon_sym_volatile] = ACTIONS(6736),
    [anon_sym_restrict] = ACTIONS(6736),
    [anon_sym___restrict__] = ACTIONS(6736),
    [anon_sym__Atomic] = ACTIONS(6736),
    [anon_sym__Noreturn] = ACTIONS(6736),
    [anon_sym__Nonnull] = ACTIONS(6736),
    [anon_sym_nullable] = ACTIONS(6736),
    [anon_sym__Complex] = ACTIONS(6736),
    [anon_sym__Nullable] = ACTIONS(6736),
    [anon_sym__Nullable_result] = ACTIONS(6736),
    [anon_sym__Null_unspecified] = ACTIONS(6736),
    [anon_sym___autoreleasing] = ACTIONS(6736),
    [anon_sym___block] = ACTIONS(6736),
    [anon_sym___bridge] = ACTIONS(6736),
    [anon_sym___bridge_retained] = ACTIONS(6736),
    [anon_sym___bridge_transfer] = ACTIONS(6736),
    [anon_sym___complex] = ACTIONS(6736),
    [anon_sym___const] = ACTIONS(6736),
    [anon_sym___imag] = ACTIONS(6736),
    [anon_sym___kindof] = ACTIONS(6736),
    [anon_sym___nonnull] = ACTIONS(6736),
    [anon_sym___nullable] = ACTIONS(6736),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6736),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6736),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6736),
    [anon_sym___real] = ACTIONS(6736),
    [anon_sym___strong] = ACTIONS(6736),
    [anon_sym___unsafe_unretained] = ACTIONS(6736),
    [anon_sym___unused] = ACTIONS(6736),
    [anon_sym___weak] = ACTIONS(6736),
    [anon_sym_alignas] = ACTIONS(6736),
    [anon_sym__Alignas] = ACTIONS(6736),
    [sym_primitive_type] = ACTIONS(6736),
    [anon_sym_enum] = ACTIONS(6736),
    [anon_sym_struct] = ACTIONS(6736),
    [anon_sym_union] = ACTIONS(6736),
    [anon_sym_in] = ACTIONS(6736),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6736),
    [anon_sym___typeof] = ACTIONS(6736),
    [anon_sym_typeof] = ACTIONS(6736),
    [anon_sym_BOOL] = ACTIONS(6736),
    [anon_sym_IMP] = ACTIONS(6736),
    [anon_sym_SEL] = ACTIONS(6736),
    [anon_sym_Class] = ACTIONS(6736),
    [anon_sym_id] = ACTIONS(6736),
    [anon_sym_out] = ACTIONS(6736),
    [anon_sym_inout] = ACTIONS(6736),
    [anon_sym_bycopy] = ACTIONS(6736),
    [anon_sym_byref] = ACTIONS(6736),
    [anon_sym_oneway] = ACTIONS(6736),
  },
  [STATE(3832)] = {
    [sym_parameter_list] = STATE(3764),
    [sym_identifier] = ACTIONS(6718),
    [anon_sym_COMMA] = ACTIONS(6720),
    [anon_sym_RPAREN] = ACTIONS(6720),
    [anon_sym_LPAREN2] = ACTIONS(8850),
    [anon_sym_STAR] = ACTIONS(6720),
    [anon_sym_CARET] = ACTIONS(6720),
    [anon_sym_GT] = ACTIONS(6720),
    [anon_sym_LT] = ACTIONS(6720),
    [anon_sym___extension__] = ACTIONS(6718),
    [anon_sym___attribute__] = ACTIONS(6718),
    [anon_sym___attribute] = ACTIONS(6718),
    [anon_sym_noreturn] = ACTIONS(6718),
    [anon_sym_LBRACK] = ACTIONS(8852),
    [anon_sym___based] = ACTIONS(6718),
    [anon_sym_LBRACE] = ACTIONS(6720),
    [anon_sym_signed] = ACTIONS(6718),
    [anon_sym_unsigned] = ACTIONS(6718),
    [anon_sym_long] = ACTIONS(6718),
    [anon_sym_short] = ACTIONS(6718),
    [anon_sym_ATautoreleasepool] = ACTIONS(6720),
    [anon_sym_const] = ACTIONS(6718),
    [anon_sym_constexpr] = ACTIONS(6718),
    [anon_sym_volatile] = ACTIONS(6718),
    [anon_sym_restrict] = ACTIONS(6718),
    [anon_sym___restrict__] = ACTIONS(6718),
    [anon_sym__Atomic] = ACTIONS(6718),
    [anon_sym__Noreturn] = ACTIONS(6718),
    [anon_sym__Nonnull] = ACTIONS(6718),
    [anon_sym_nullable] = ACTIONS(6718),
    [anon_sym__Complex] = ACTIONS(6718),
    [anon_sym__Nullable] = ACTIONS(6718),
    [anon_sym__Nullable_result] = ACTIONS(6718),
    [anon_sym__Null_unspecified] = ACTIONS(6718),
    [anon_sym___autoreleasing] = ACTIONS(6718),
    [anon_sym___block] = ACTIONS(6718),
    [anon_sym___bridge] = ACTIONS(6718),
    [anon_sym___bridge_retained] = ACTIONS(6718),
    [anon_sym___bridge_transfer] = ACTIONS(6718),
    [anon_sym___complex] = ACTIONS(6718),
    [anon_sym___const] = ACTIONS(6718),
    [anon_sym___imag] = ACTIONS(6718),
    [anon_sym___kindof] = ACTIONS(6718),
    [anon_sym___nonnull] = ACTIONS(6718),
    [anon_sym___nullable] = ACTIONS(6718),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6718),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6718),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6718),
    [anon_sym___real] = ACTIONS(6718),
    [anon_sym___strong] = ACTIONS(6718),
    [anon_sym___unsafe_unretained] = ACTIONS(6718),
    [anon_sym___unused] = ACTIONS(6718),
    [anon_sym___weak] = ACTIONS(6718),
    [anon_sym_alignas] = ACTIONS(6718),
    [anon_sym__Alignas] = ACTIONS(6718),
    [sym_primitive_type] = ACTIONS(6718),
    [anon_sym_enum] = ACTIONS(6718),
    [anon_sym_struct] = ACTIONS(6718),
    [anon_sym_union] = ACTIONS(6718),
    [anon_sym_in] = ACTIONS(6718),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6718),
    [anon_sym___typeof] = ACTIONS(6718),
    [anon_sym_typeof] = ACTIONS(6718),
    [anon_sym_BOOL] = ACTIONS(6718),
    [anon_sym_IMP] = ACTIONS(6718),
    [anon_sym_SEL] = ACTIONS(6718),
    [anon_sym_Class] = ACTIONS(6718),
    [anon_sym_id] = ACTIONS(6718),
    [anon_sym_out] = ACTIONS(6718),
    [anon_sym_inout] = ACTIONS(6718),
    [anon_sym_bycopy] = ACTIONS(6718),
    [anon_sym_byref] = ACTIONS(6718),
    [anon_sym_oneway] = ACTIONS(6718),
  },
  [STATE(3833)] = {
    [sym__type_definition_type] = STATE(4687),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4035),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3834)] = {
    [sym_attribute_specifier] = STATE(3904),
    [sym_enumerator_list] = STATE(3883),
    [sym_identifier] = ACTIONS(7535),
    [anon_sym_COMMA] = ACTIONS(7537),
    [anon_sym_RPAREN] = ACTIONS(7537),
    [anon_sym_LPAREN2] = ACTIONS(7537),
    [anon_sym_STAR] = ACTIONS(7537),
    [anon_sym_CARET] = ACTIONS(7537),
    [anon_sym_GT] = ACTIONS(7537),
    [anon_sym_LT] = ACTIONS(7537),
    [anon_sym___extension__] = ACTIONS(7535),
    [anon_sym___attribute__] = ACTIONS(8854),
    [anon_sym___attribute] = ACTIONS(8854),
    [anon_sym_noreturn] = ACTIONS(7535),
    [anon_sym_LBRACK] = ACTIONS(7537),
    [anon_sym___based] = ACTIONS(7535),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7535),
    [anon_sym_unsigned] = ACTIONS(7535),
    [anon_sym_long] = ACTIONS(7535),
    [anon_sym_short] = ACTIONS(7535),
    [anon_sym_const] = ACTIONS(7535),
    [anon_sym_constexpr] = ACTIONS(7535),
    [anon_sym_volatile] = ACTIONS(7535),
    [anon_sym_restrict] = ACTIONS(7535),
    [anon_sym___restrict__] = ACTIONS(7535),
    [anon_sym__Atomic] = ACTIONS(7535),
    [anon_sym__Noreturn] = ACTIONS(7535),
    [anon_sym__Nonnull] = ACTIONS(7535),
    [anon_sym_nullable] = ACTIONS(7535),
    [anon_sym__Complex] = ACTIONS(7535),
    [anon_sym__Nullable] = ACTIONS(7535),
    [anon_sym__Nullable_result] = ACTIONS(7535),
    [anon_sym__Null_unspecified] = ACTIONS(7535),
    [anon_sym___autoreleasing] = ACTIONS(7535),
    [anon_sym___block] = ACTIONS(7535),
    [anon_sym___bridge] = ACTIONS(7535),
    [anon_sym___bridge_retained] = ACTIONS(7535),
    [anon_sym___bridge_transfer] = ACTIONS(7535),
    [anon_sym___complex] = ACTIONS(7535),
    [anon_sym___const] = ACTIONS(7535),
    [anon_sym___imag] = ACTIONS(7535),
    [anon_sym___kindof] = ACTIONS(7535),
    [anon_sym___nonnull] = ACTIONS(7535),
    [anon_sym___nullable] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7535),
    [anon_sym___real] = ACTIONS(7535),
    [anon_sym___strong] = ACTIONS(7535),
    [anon_sym___unsafe_unretained] = ACTIONS(7535),
    [anon_sym___unused] = ACTIONS(7535),
    [anon_sym___weak] = ACTIONS(7535),
    [anon_sym_alignas] = ACTIONS(7535),
    [anon_sym__Alignas] = ACTIONS(7535),
    [sym_primitive_type] = ACTIONS(7535),
    [anon_sym_enum] = ACTIONS(7535),
    [anon_sym_struct] = ACTIONS(7535),
    [anon_sym_union] = ACTIONS(7535),
    [anon_sym_in] = ACTIONS(7535),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7535),
    [anon_sym___typeof] = ACTIONS(7535),
    [anon_sym_typeof] = ACTIONS(7535),
    [anon_sym_BOOL] = ACTIONS(7535),
    [anon_sym_IMP] = ACTIONS(7535),
    [anon_sym_SEL] = ACTIONS(7535),
    [anon_sym_Class] = ACTIONS(7535),
    [anon_sym_id] = ACTIONS(7535),
    [anon_sym_out] = ACTIONS(7535),
    [anon_sym_inout] = ACTIONS(7535),
    [anon_sym_bycopy] = ACTIONS(7535),
    [anon_sym_byref] = ACTIONS(7535),
    [anon_sym_oneway] = ACTIONS(7535),
  },
  [STATE(3835)] = {
    [sym_attribute_specifier] = STATE(3942),
    [sym_enumerator_list] = STATE(3885),
    [sym_identifier] = ACTIONS(7548),
    [anon_sym_COMMA] = ACTIONS(7550),
    [anon_sym_RPAREN] = ACTIONS(7550),
    [anon_sym_LPAREN2] = ACTIONS(7550),
    [anon_sym_STAR] = ACTIONS(7550),
    [anon_sym_CARET] = ACTIONS(7550),
    [anon_sym_GT] = ACTIONS(7550),
    [anon_sym_LT] = ACTIONS(7550),
    [anon_sym___extension__] = ACTIONS(7548),
    [anon_sym___attribute__] = ACTIONS(8857),
    [anon_sym___attribute] = ACTIONS(8857),
    [anon_sym_noreturn] = ACTIONS(7548),
    [anon_sym_LBRACK] = ACTIONS(7550),
    [anon_sym___based] = ACTIONS(7548),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7548),
    [anon_sym_unsigned] = ACTIONS(7548),
    [anon_sym_long] = ACTIONS(7548),
    [anon_sym_short] = ACTIONS(7548),
    [anon_sym_const] = ACTIONS(7548),
    [anon_sym_constexpr] = ACTIONS(7548),
    [anon_sym_volatile] = ACTIONS(7548),
    [anon_sym_restrict] = ACTIONS(7548),
    [anon_sym___restrict__] = ACTIONS(7548),
    [anon_sym__Atomic] = ACTIONS(7548),
    [anon_sym__Noreturn] = ACTIONS(7548),
    [anon_sym__Nonnull] = ACTIONS(7548),
    [anon_sym_nullable] = ACTIONS(7548),
    [anon_sym__Complex] = ACTIONS(7548),
    [anon_sym__Nullable] = ACTIONS(7548),
    [anon_sym__Nullable_result] = ACTIONS(7548),
    [anon_sym__Null_unspecified] = ACTIONS(7548),
    [anon_sym___autoreleasing] = ACTIONS(7548),
    [anon_sym___block] = ACTIONS(7548),
    [anon_sym___bridge] = ACTIONS(7548),
    [anon_sym___bridge_retained] = ACTIONS(7548),
    [anon_sym___bridge_transfer] = ACTIONS(7548),
    [anon_sym___complex] = ACTIONS(7548),
    [anon_sym___const] = ACTIONS(7548),
    [anon_sym___imag] = ACTIONS(7548),
    [anon_sym___kindof] = ACTIONS(7548),
    [anon_sym___nonnull] = ACTIONS(7548),
    [anon_sym___nullable] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7548),
    [anon_sym___real] = ACTIONS(7548),
    [anon_sym___strong] = ACTIONS(7548),
    [anon_sym___unsafe_unretained] = ACTIONS(7548),
    [anon_sym___unused] = ACTIONS(7548),
    [anon_sym___weak] = ACTIONS(7548),
    [anon_sym_alignas] = ACTIONS(7548),
    [anon_sym__Alignas] = ACTIONS(7548),
    [sym_primitive_type] = ACTIONS(7548),
    [anon_sym_enum] = ACTIONS(7548),
    [anon_sym_struct] = ACTIONS(7548),
    [anon_sym_union] = ACTIONS(7548),
    [anon_sym_in] = ACTIONS(7548),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7548),
    [anon_sym___typeof] = ACTIONS(7548),
    [anon_sym_typeof] = ACTIONS(7548),
    [anon_sym_BOOL] = ACTIONS(7548),
    [anon_sym_IMP] = ACTIONS(7548),
    [anon_sym_SEL] = ACTIONS(7548),
    [anon_sym_Class] = ACTIONS(7548),
    [anon_sym_id] = ACTIONS(7548),
    [anon_sym_out] = ACTIONS(7548),
    [anon_sym_inout] = ACTIONS(7548),
    [anon_sym_bycopy] = ACTIONS(7548),
    [anon_sym_byref] = ACTIONS(7548),
    [anon_sym_oneway] = ACTIONS(7548),
  },
  [STATE(3836)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_COMMA] = ACTIONS(2920),
    [anon_sym_RPAREN] = ACTIONS(2920),
    [anon_sym_LPAREN2] = ACTIONS(2920),
    [anon_sym_STAR] = ACTIONS(2920),
    [anon_sym_CARET] = ACTIONS(2920),
    [anon_sym_GT] = ACTIONS(2920),
    [anon_sym_LT] = ACTIONS(2920),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym___attribute__] = ACTIONS(2914),
    [anon_sym___attribute] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym___based] = ACTIONS(2914),
    [anon_sym_LBRACE] = ACTIONS(2920),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_ATautoreleasepool] = ACTIONS(2920),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_COLON] = ACTIONS(2920),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
    [anon_sym_out] = ACTIONS(2914),
    [anon_sym_inout] = ACTIONS(2914),
    [anon_sym_bycopy] = ACTIONS(2914),
    [anon_sym_byref] = ACTIONS(2914),
    [anon_sym_oneway] = ACTIONS(2914),
  },
  [STATE(3837)] = {
    [sym__type_definition_type] = STATE(4723),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4043),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3838)] = {
    [sym_identifier] = ACTIONS(7179),
    [anon_sym_COMMA] = ACTIONS(7181),
    [anon_sym_RPAREN] = ACTIONS(7181),
    [anon_sym_LPAREN2] = ACTIONS(7181),
    [anon_sym_STAR] = ACTIONS(7181),
    [anon_sym_CARET] = ACTIONS(7181),
    [anon_sym_GT] = ACTIONS(7181),
    [anon_sym_LT] = ACTIONS(7181),
    [anon_sym___extension__] = ACTIONS(7179),
    [anon_sym___attribute__] = ACTIONS(7179),
    [anon_sym___attribute] = ACTIONS(7179),
    [anon_sym_noreturn] = ACTIONS(7179),
    [anon_sym_LBRACK] = ACTIONS(7181),
    [anon_sym___based] = ACTIONS(7179),
    [anon_sym_LBRACE] = ACTIONS(7181),
    [anon_sym_signed] = ACTIONS(7179),
    [anon_sym_unsigned] = ACTIONS(7179),
    [anon_sym_long] = ACTIONS(7179),
    [anon_sym_short] = ACTIONS(7179),
    [anon_sym_ATautoreleasepool] = ACTIONS(7181),
    [anon_sym_const] = ACTIONS(7179),
    [anon_sym_constexpr] = ACTIONS(7179),
    [anon_sym_volatile] = ACTIONS(7179),
    [anon_sym_restrict] = ACTIONS(7179),
    [anon_sym___restrict__] = ACTIONS(7179),
    [anon_sym__Atomic] = ACTIONS(7179),
    [anon_sym__Noreturn] = ACTIONS(7179),
    [anon_sym__Nonnull] = ACTIONS(7179),
    [anon_sym_nullable] = ACTIONS(7179),
    [anon_sym__Complex] = ACTIONS(7179),
    [anon_sym__Nullable] = ACTIONS(7179),
    [anon_sym__Nullable_result] = ACTIONS(7179),
    [anon_sym__Null_unspecified] = ACTIONS(7179),
    [anon_sym___autoreleasing] = ACTIONS(7179),
    [anon_sym___block] = ACTIONS(7179),
    [anon_sym___bridge] = ACTIONS(7179),
    [anon_sym___bridge_retained] = ACTIONS(7179),
    [anon_sym___bridge_transfer] = ACTIONS(7179),
    [anon_sym___complex] = ACTIONS(7179),
    [anon_sym___const] = ACTIONS(7179),
    [anon_sym___imag] = ACTIONS(7179),
    [anon_sym___kindof] = ACTIONS(7179),
    [anon_sym___nonnull] = ACTIONS(7179),
    [anon_sym___nullable] = ACTIONS(7179),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7179),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7179),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7179),
    [anon_sym___real] = ACTIONS(7179),
    [anon_sym___strong] = ACTIONS(7179),
    [anon_sym___unsafe_unretained] = ACTIONS(7179),
    [anon_sym___unused] = ACTIONS(7179),
    [anon_sym___weak] = ACTIONS(7179),
    [anon_sym_alignas] = ACTIONS(7179),
    [anon_sym__Alignas] = ACTIONS(7179),
    [sym_primitive_type] = ACTIONS(7179),
    [anon_sym_enum] = ACTIONS(7179),
    [anon_sym_COLON] = ACTIONS(7181),
    [anon_sym_struct] = ACTIONS(7179),
    [anon_sym_union] = ACTIONS(7179),
    [anon_sym_in] = ACTIONS(7179),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7179),
    [anon_sym___typeof] = ACTIONS(7179),
    [anon_sym_typeof] = ACTIONS(7179),
    [anon_sym_BOOL] = ACTIONS(7179),
    [anon_sym_IMP] = ACTIONS(7179),
    [anon_sym_SEL] = ACTIONS(7179),
    [anon_sym_Class] = ACTIONS(7179),
    [anon_sym_id] = ACTIONS(7179),
    [anon_sym_out] = ACTIONS(7179),
    [anon_sym_inout] = ACTIONS(7179),
    [anon_sym_bycopy] = ACTIONS(7179),
    [anon_sym_byref] = ACTIONS(7179),
    [anon_sym_oneway] = ACTIONS(7179),
  },
  [STATE(3839)] = {
    [sym_identifier] = ACTIONS(7193),
    [anon_sym_COMMA] = ACTIONS(7195),
    [anon_sym_RPAREN] = ACTIONS(7195),
    [anon_sym_LPAREN2] = ACTIONS(7195),
    [anon_sym_STAR] = ACTIONS(7195),
    [anon_sym_CARET] = ACTIONS(7195),
    [anon_sym_GT] = ACTIONS(7195),
    [anon_sym_LT] = ACTIONS(7195),
    [anon_sym___extension__] = ACTIONS(7193),
    [anon_sym___attribute__] = ACTIONS(7193),
    [anon_sym___attribute] = ACTIONS(7193),
    [anon_sym_noreturn] = ACTIONS(7193),
    [anon_sym_LBRACK] = ACTIONS(7195),
    [anon_sym___based] = ACTIONS(7193),
    [anon_sym_LBRACE] = ACTIONS(7195),
    [anon_sym_signed] = ACTIONS(7193),
    [anon_sym_unsigned] = ACTIONS(7193),
    [anon_sym_long] = ACTIONS(7193),
    [anon_sym_short] = ACTIONS(7193),
    [anon_sym_ATautoreleasepool] = ACTIONS(7195),
    [anon_sym_const] = ACTIONS(7193),
    [anon_sym_constexpr] = ACTIONS(7193),
    [anon_sym_volatile] = ACTIONS(7193),
    [anon_sym_restrict] = ACTIONS(7193),
    [anon_sym___restrict__] = ACTIONS(7193),
    [anon_sym__Atomic] = ACTIONS(7193),
    [anon_sym__Noreturn] = ACTIONS(7193),
    [anon_sym__Nonnull] = ACTIONS(7193),
    [anon_sym_nullable] = ACTIONS(7193),
    [anon_sym__Complex] = ACTIONS(7193),
    [anon_sym__Nullable] = ACTIONS(7193),
    [anon_sym__Nullable_result] = ACTIONS(7193),
    [anon_sym__Null_unspecified] = ACTIONS(7193),
    [anon_sym___autoreleasing] = ACTIONS(7193),
    [anon_sym___block] = ACTIONS(7193),
    [anon_sym___bridge] = ACTIONS(7193),
    [anon_sym___bridge_retained] = ACTIONS(7193),
    [anon_sym___bridge_transfer] = ACTIONS(7193),
    [anon_sym___complex] = ACTIONS(7193),
    [anon_sym___const] = ACTIONS(7193),
    [anon_sym___imag] = ACTIONS(7193),
    [anon_sym___kindof] = ACTIONS(7193),
    [anon_sym___nonnull] = ACTIONS(7193),
    [anon_sym___nullable] = ACTIONS(7193),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7193),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7193),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7193),
    [anon_sym___real] = ACTIONS(7193),
    [anon_sym___strong] = ACTIONS(7193),
    [anon_sym___unsafe_unretained] = ACTIONS(7193),
    [anon_sym___unused] = ACTIONS(7193),
    [anon_sym___weak] = ACTIONS(7193),
    [anon_sym_alignas] = ACTIONS(7193),
    [anon_sym__Alignas] = ACTIONS(7193),
    [sym_primitive_type] = ACTIONS(7193),
    [anon_sym_enum] = ACTIONS(7193),
    [anon_sym_COLON] = ACTIONS(7195),
    [anon_sym_struct] = ACTIONS(7193),
    [anon_sym_union] = ACTIONS(7193),
    [anon_sym_in] = ACTIONS(7193),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7193),
    [anon_sym___typeof] = ACTIONS(7193),
    [anon_sym_typeof] = ACTIONS(7193),
    [anon_sym_BOOL] = ACTIONS(7193),
    [anon_sym_IMP] = ACTIONS(7193),
    [anon_sym_SEL] = ACTIONS(7193),
    [anon_sym_Class] = ACTIONS(7193),
    [anon_sym_id] = ACTIONS(7193),
    [anon_sym_out] = ACTIONS(7193),
    [anon_sym_inout] = ACTIONS(7193),
    [anon_sym_bycopy] = ACTIONS(7193),
    [anon_sym_byref] = ACTIONS(7193),
    [anon_sym_oneway] = ACTIONS(7193),
  },
  [STATE(3840)] = {
    [sym_identifier] = ACTIONS(7205),
    [anon_sym_COMMA] = ACTIONS(7207),
    [anon_sym_RPAREN] = ACTIONS(7207),
    [anon_sym_LPAREN2] = ACTIONS(7207),
    [anon_sym_STAR] = ACTIONS(7207),
    [anon_sym_CARET] = ACTIONS(7207),
    [anon_sym_GT] = ACTIONS(7207),
    [anon_sym_LT] = ACTIONS(7207),
    [anon_sym___extension__] = ACTIONS(7205),
    [anon_sym___attribute__] = ACTIONS(7205),
    [anon_sym___attribute] = ACTIONS(7205),
    [anon_sym_noreturn] = ACTIONS(7205),
    [anon_sym_LBRACK] = ACTIONS(7207),
    [anon_sym___based] = ACTIONS(7205),
    [anon_sym_LBRACE] = ACTIONS(7207),
    [anon_sym_signed] = ACTIONS(7205),
    [anon_sym_unsigned] = ACTIONS(7205),
    [anon_sym_long] = ACTIONS(7205),
    [anon_sym_short] = ACTIONS(7205),
    [anon_sym_ATautoreleasepool] = ACTIONS(7207),
    [anon_sym_const] = ACTIONS(7205),
    [anon_sym_constexpr] = ACTIONS(7205),
    [anon_sym_volatile] = ACTIONS(7205),
    [anon_sym_restrict] = ACTIONS(7205),
    [anon_sym___restrict__] = ACTIONS(7205),
    [anon_sym__Atomic] = ACTIONS(7205),
    [anon_sym__Noreturn] = ACTIONS(7205),
    [anon_sym__Nonnull] = ACTIONS(7205),
    [anon_sym_nullable] = ACTIONS(7205),
    [anon_sym__Complex] = ACTIONS(7205),
    [anon_sym__Nullable] = ACTIONS(7205),
    [anon_sym__Nullable_result] = ACTIONS(7205),
    [anon_sym__Null_unspecified] = ACTIONS(7205),
    [anon_sym___autoreleasing] = ACTIONS(7205),
    [anon_sym___block] = ACTIONS(7205),
    [anon_sym___bridge] = ACTIONS(7205),
    [anon_sym___bridge_retained] = ACTIONS(7205),
    [anon_sym___bridge_transfer] = ACTIONS(7205),
    [anon_sym___complex] = ACTIONS(7205),
    [anon_sym___const] = ACTIONS(7205),
    [anon_sym___imag] = ACTIONS(7205),
    [anon_sym___kindof] = ACTIONS(7205),
    [anon_sym___nonnull] = ACTIONS(7205),
    [anon_sym___nullable] = ACTIONS(7205),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7205),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7205),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7205),
    [anon_sym___real] = ACTIONS(7205),
    [anon_sym___strong] = ACTIONS(7205),
    [anon_sym___unsafe_unretained] = ACTIONS(7205),
    [anon_sym___unused] = ACTIONS(7205),
    [anon_sym___weak] = ACTIONS(7205),
    [anon_sym_alignas] = ACTIONS(7205),
    [anon_sym__Alignas] = ACTIONS(7205),
    [sym_primitive_type] = ACTIONS(7205),
    [anon_sym_enum] = ACTIONS(7205),
    [anon_sym_COLON] = ACTIONS(7207),
    [anon_sym_struct] = ACTIONS(7205),
    [anon_sym_union] = ACTIONS(7205),
    [anon_sym_in] = ACTIONS(7205),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7205),
    [anon_sym___typeof] = ACTIONS(7205),
    [anon_sym_typeof] = ACTIONS(7205),
    [anon_sym_BOOL] = ACTIONS(7205),
    [anon_sym_IMP] = ACTIONS(7205),
    [anon_sym_SEL] = ACTIONS(7205),
    [anon_sym_Class] = ACTIONS(7205),
    [anon_sym_id] = ACTIONS(7205),
    [anon_sym_out] = ACTIONS(7205),
    [anon_sym_inout] = ACTIONS(7205),
    [anon_sym_bycopy] = ACTIONS(7205),
    [anon_sym_byref] = ACTIONS(7205),
    [anon_sym_oneway] = ACTIONS(7205),
  },
  [STATE(3841)] = {
    [sym_identifier] = ACTIONS(7213),
    [anon_sym_COMMA] = ACTIONS(7215),
    [anon_sym_RPAREN] = ACTIONS(7215),
    [anon_sym_LPAREN2] = ACTIONS(7215),
    [anon_sym_STAR] = ACTIONS(7215),
    [anon_sym_CARET] = ACTIONS(7215),
    [anon_sym_GT] = ACTIONS(7215),
    [anon_sym_LT] = ACTIONS(7215),
    [anon_sym___extension__] = ACTIONS(7213),
    [anon_sym___attribute__] = ACTIONS(7213),
    [anon_sym___attribute] = ACTIONS(7213),
    [anon_sym_noreturn] = ACTIONS(7213),
    [anon_sym_LBRACK] = ACTIONS(7215),
    [anon_sym___based] = ACTIONS(7213),
    [anon_sym_LBRACE] = ACTIONS(7215),
    [anon_sym_signed] = ACTIONS(7213),
    [anon_sym_unsigned] = ACTIONS(7213),
    [anon_sym_long] = ACTIONS(7213),
    [anon_sym_short] = ACTIONS(7213),
    [anon_sym_ATautoreleasepool] = ACTIONS(7215),
    [anon_sym_const] = ACTIONS(7213),
    [anon_sym_constexpr] = ACTIONS(7213),
    [anon_sym_volatile] = ACTIONS(7213),
    [anon_sym_restrict] = ACTIONS(7213),
    [anon_sym___restrict__] = ACTIONS(7213),
    [anon_sym__Atomic] = ACTIONS(7213),
    [anon_sym__Noreturn] = ACTIONS(7213),
    [anon_sym__Nonnull] = ACTIONS(7213),
    [anon_sym_nullable] = ACTIONS(7213),
    [anon_sym__Complex] = ACTIONS(7213),
    [anon_sym__Nullable] = ACTIONS(7213),
    [anon_sym__Nullable_result] = ACTIONS(7213),
    [anon_sym__Null_unspecified] = ACTIONS(7213),
    [anon_sym___autoreleasing] = ACTIONS(7213),
    [anon_sym___block] = ACTIONS(7213),
    [anon_sym___bridge] = ACTIONS(7213),
    [anon_sym___bridge_retained] = ACTIONS(7213),
    [anon_sym___bridge_transfer] = ACTIONS(7213),
    [anon_sym___complex] = ACTIONS(7213),
    [anon_sym___const] = ACTIONS(7213),
    [anon_sym___imag] = ACTIONS(7213),
    [anon_sym___kindof] = ACTIONS(7213),
    [anon_sym___nonnull] = ACTIONS(7213),
    [anon_sym___nullable] = ACTIONS(7213),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7213),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7213),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7213),
    [anon_sym___real] = ACTIONS(7213),
    [anon_sym___strong] = ACTIONS(7213),
    [anon_sym___unsafe_unretained] = ACTIONS(7213),
    [anon_sym___unused] = ACTIONS(7213),
    [anon_sym___weak] = ACTIONS(7213),
    [anon_sym_alignas] = ACTIONS(7213),
    [anon_sym__Alignas] = ACTIONS(7213),
    [sym_primitive_type] = ACTIONS(7213),
    [anon_sym_enum] = ACTIONS(7213),
    [anon_sym_COLON] = ACTIONS(7215),
    [anon_sym_struct] = ACTIONS(7213),
    [anon_sym_union] = ACTIONS(7213),
    [anon_sym_in] = ACTIONS(7213),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7213),
    [anon_sym___typeof] = ACTIONS(7213),
    [anon_sym_typeof] = ACTIONS(7213),
    [anon_sym_BOOL] = ACTIONS(7213),
    [anon_sym_IMP] = ACTIONS(7213),
    [anon_sym_SEL] = ACTIONS(7213),
    [anon_sym_Class] = ACTIONS(7213),
    [anon_sym_id] = ACTIONS(7213),
    [anon_sym_out] = ACTIONS(7213),
    [anon_sym_inout] = ACTIONS(7213),
    [anon_sym_bycopy] = ACTIONS(7213),
    [anon_sym_byref] = ACTIONS(7213),
    [anon_sym_oneway] = ACTIONS(7213),
  },
  [STATE(3842)] = {
    [sym_identifier] = ACTIONS(7221),
    [anon_sym_COMMA] = ACTIONS(7223),
    [anon_sym_RPAREN] = ACTIONS(7223),
    [anon_sym_LPAREN2] = ACTIONS(7223),
    [anon_sym_STAR] = ACTIONS(7223),
    [anon_sym_CARET] = ACTIONS(7223),
    [anon_sym_GT] = ACTIONS(7223),
    [anon_sym_LT] = ACTIONS(7223),
    [anon_sym___extension__] = ACTIONS(7221),
    [anon_sym___attribute__] = ACTIONS(7221),
    [anon_sym___attribute] = ACTIONS(7221),
    [anon_sym_noreturn] = ACTIONS(7221),
    [anon_sym_LBRACK] = ACTIONS(7223),
    [anon_sym___based] = ACTIONS(7221),
    [anon_sym_LBRACE] = ACTIONS(7223),
    [anon_sym_signed] = ACTIONS(7221),
    [anon_sym_unsigned] = ACTIONS(7221),
    [anon_sym_long] = ACTIONS(7221),
    [anon_sym_short] = ACTIONS(7221),
    [anon_sym_ATautoreleasepool] = ACTIONS(7223),
    [anon_sym_const] = ACTIONS(7221),
    [anon_sym_constexpr] = ACTIONS(7221),
    [anon_sym_volatile] = ACTIONS(7221),
    [anon_sym_restrict] = ACTIONS(7221),
    [anon_sym___restrict__] = ACTIONS(7221),
    [anon_sym__Atomic] = ACTIONS(7221),
    [anon_sym__Noreturn] = ACTIONS(7221),
    [anon_sym__Nonnull] = ACTIONS(7221),
    [anon_sym_nullable] = ACTIONS(7221),
    [anon_sym__Complex] = ACTIONS(7221),
    [anon_sym__Nullable] = ACTIONS(7221),
    [anon_sym__Nullable_result] = ACTIONS(7221),
    [anon_sym__Null_unspecified] = ACTIONS(7221),
    [anon_sym___autoreleasing] = ACTIONS(7221),
    [anon_sym___block] = ACTIONS(7221),
    [anon_sym___bridge] = ACTIONS(7221),
    [anon_sym___bridge_retained] = ACTIONS(7221),
    [anon_sym___bridge_transfer] = ACTIONS(7221),
    [anon_sym___complex] = ACTIONS(7221),
    [anon_sym___const] = ACTIONS(7221),
    [anon_sym___imag] = ACTIONS(7221),
    [anon_sym___kindof] = ACTIONS(7221),
    [anon_sym___nonnull] = ACTIONS(7221),
    [anon_sym___nullable] = ACTIONS(7221),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7221),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7221),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7221),
    [anon_sym___real] = ACTIONS(7221),
    [anon_sym___strong] = ACTIONS(7221),
    [anon_sym___unsafe_unretained] = ACTIONS(7221),
    [anon_sym___unused] = ACTIONS(7221),
    [anon_sym___weak] = ACTIONS(7221),
    [anon_sym_alignas] = ACTIONS(7221),
    [anon_sym__Alignas] = ACTIONS(7221),
    [sym_primitive_type] = ACTIONS(7221),
    [anon_sym_enum] = ACTIONS(7221),
    [anon_sym_COLON] = ACTIONS(7223),
    [anon_sym_struct] = ACTIONS(7221),
    [anon_sym_union] = ACTIONS(7221),
    [anon_sym_in] = ACTIONS(7221),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7221),
    [anon_sym___typeof] = ACTIONS(7221),
    [anon_sym_typeof] = ACTIONS(7221),
    [anon_sym_BOOL] = ACTIONS(7221),
    [anon_sym_IMP] = ACTIONS(7221),
    [anon_sym_SEL] = ACTIONS(7221),
    [anon_sym_Class] = ACTIONS(7221),
    [anon_sym_id] = ACTIONS(7221),
    [anon_sym_out] = ACTIONS(7221),
    [anon_sym_inout] = ACTIONS(7221),
    [anon_sym_bycopy] = ACTIONS(7221),
    [anon_sym_byref] = ACTIONS(7221),
    [anon_sym_oneway] = ACTIONS(7221),
  },
  [STATE(3843)] = {
    [sym_identifier] = ACTIONS(5666),
    [anon_sym_COMMA] = ACTIONS(5668),
    [anon_sym_RPAREN] = ACTIONS(5668),
    [anon_sym_LPAREN2] = ACTIONS(5668),
    [anon_sym_STAR] = ACTIONS(5668),
    [anon_sym_CARET] = ACTIONS(5668),
    [anon_sym_GT] = ACTIONS(5668),
    [anon_sym_LT] = ACTIONS(5668),
    [anon_sym___extension__] = ACTIONS(5666),
    [anon_sym___attribute__] = ACTIONS(5666),
    [anon_sym___attribute] = ACTIONS(5666),
    [anon_sym_noreturn] = ACTIONS(5666),
    [anon_sym_LBRACK] = ACTIONS(5668),
    [anon_sym___based] = ACTIONS(5666),
    [anon_sym_LBRACE] = ACTIONS(5668),
    [anon_sym_signed] = ACTIONS(5666),
    [anon_sym_unsigned] = ACTIONS(5666),
    [anon_sym_long] = ACTIONS(5666),
    [anon_sym_short] = ACTIONS(5666),
    [anon_sym_ATautoreleasepool] = ACTIONS(5668),
    [anon_sym_const] = ACTIONS(5666),
    [anon_sym_constexpr] = ACTIONS(5666),
    [anon_sym_volatile] = ACTIONS(5666),
    [anon_sym_restrict] = ACTIONS(5666),
    [anon_sym___restrict__] = ACTIONS(5666),
    [anon_sym__Atomic] = ACTIONS(5666),
    [anon_sym__Noreturn] = ACTIONS(5666),
    [anon_sym__Nonnull] = ACTIONS(5666),
    [anon_sym_nullable] = ACTIONS(5666),
    [anon_sym__Complex] = ACTIONS(5666),
    [anon_sym__Nullable] = ACTIONS(5666),
    [anon_sym__Nullable_result] = ACTIONS(5666),
    [anon_sym__Null_unspecified] = ACTIONS(5666),
    [anon_sym___autoreleasing] = ACTIONS(5666),
    [anon_sym___block] = ACTIONS(5666),
    [anon_sym___bridge] = ACTIONS(5666),
    [anon_sym___bridge_retained] = ACTIONS(5666),
    [anon_sym___bridge_transfer] = ACTIONS(5666),
    [anon_sym___complex] = ACTIONS(5666),
    [anon_sym___const] = ACTIONS(5666),
    [anon_sym___imag] = ACTIONS(5666),
    [anon_sym___kindof] = ACTIONS(5666),
    [anon_sym___nonnull] = ACTIONS(5666),
    [anon_sym___nullable] = ACTIONS(5666),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5666),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5666),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5666),
    [anon_sym___real] = ACTIONS(5666),
    [anon_sym___strong] = ACTIONS(5666),
    [anon_sym___unsafe_unretained] = ACTIONS(5666),
    [anon_sym___unused] = ACTIONS(5666),
    [anon_sym___weak] = ACTIONS(5666),
    [anon_sym_alignas] = ACTIONS(5666),
    [anon_sym__Alignas] = ACTIONS(5666),
    [sym_primitive_type] = ACTIONS(5666),
    [anon_sym_enum] = ACTIONS(5666),
    [anon_sym_COLON] = ACTIONS(5668),
    [anon_sym_struct] = ACTIONS(5666),
    [anon_sym_union] = ACTIONS(5666),
    [anon_sym_in] = ACTIONS(5666),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5666),
    [anon_sym___typeof] = ACTIONS(5666),
    [anon_sym_typeof] = ACTIONS(5666),
    [anon_sym_BOOL] = ACTIONS(5666),
    [anon_sym_IMP] = ACTIONS(5666),
    [anon_sym_SEL] = ACTIONS(5666),
    [anon_sym_Class] = ACTIONS(5666),
    [anon_sym_id] = ACTIONS(5666),
    [anon_sym_out] = ACTIONS(5666),
    [anon_sym_inout] = ACTIONS(5666),
    [anon_sym_bycopy] = ACTIONS(5666),
    [anon_sym_byref] = ACTIONS(5666),
    [anon_sym_oneway] = ACTIONS(5666),
  },
  [STATE(3844)] = {
    [sym_attribute_specifier] = STATE(3877),
    [sym_identifier] = ACTIONS(7817),
    [anon_sym_COMMA] = ACTIONS(7819),
    [anon_sym_RPAREN] = ACTIONS(7819),
    [anon_sym_LPAREN2] = ACTIONS(7819),
    [anon_sym_STAR] = ACTIONS(7819),
    [anon_sym_CARET] = ACTIONS(7819),
    [anon_sym_GT] = ACTIONS(7819),
    [anon_sym_LT] = ACTIONS(7819),
    [anon_sym___extension__] = ACTIONS(7817),
    [anon_sym___attribute__] = ACTIONS(8860),
    [anon_sym___attribute] = ACTIONS(8860),
    [anon_sym_noreturn] = ACTIONS(7817),
    [anon_sym_LBRACK] = ACTIONS(7819),
    [anon_sym___based] = ACTIONS(7817),
    [anon_sym_LBRACE] = ACTIONS(7819),
    [anon_sym_signed] = ACTIONS(7817),
    [anon_sym_unsigned] = ACTIONS(7817),
    [anon_sym_long] = ACTIONS(7817),
    [anon_sym_short] = ACTIONS(7817),
    [anon_sym_ATautoreleasepool] = ACTIONS(7819),
    [anon_sym_const] = ACTIONS(7817),
    [anon_sym_constexpr] = ACTIONS(7817),
    [anon_sym_volatile] = ACTIONS(7817),
    [anon_sym_restrict] = ACTIONS(7817),
    [anon_sym___restrict__] = ACTIONS(7817),
    [anon_sym__Atomic] = ACTIONS(7817),
    [anon_sym__Noreturn] = ACTIONS(7817),
    [anon_sym__Nonnull] = ACTIONS(7817),
    [anon_sym_nullable] = ACTIONS(7817),
    [anon_sym__Complex] = ACTIONS(7817),
    [anon_sym__Nullable] = ACTIONS(7817),
    [anon_sym__Nullable_result] = ACTIONS(7817),
    [anon_sym__Null_unspecified] = ACTIONS(7817),
    [anon_sym___autoreleasing] = ACTIONS(7817),
    [anon_sym___block] = ACTIONS(7817),
    [anon_sym___bridge] = ACTIONS(7817),
    [anon_sym___bridge_retained] = ACTIONS(7817),
    [anon_sym___bridge_transfer] = ACTIONS(7817),
    [anon_sym___complex] = ACTIONS(7817),
    [anon_sym___const] = ACTIONS(7817),
    [anon_sym___imag] = ACTIONS(7817),
    [anon_sym___kindof] = ACTIONS(7817),
    [anon_sym___nonnull] = ACTIONS(7817),
    [anon_sym___nullable] = ACTIONS(7817),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7817),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7817),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7817),
    [anon_sym___real] = ACTIONS(7817),
    [anon_sym___strong] = ACTIONS(7817),
    [anon_sym___unsafe_unretained] = ACTIONS(7817),
    [anon_sym___unused] = ACTIONS(7817),
    [anon_sym___weak] = ACTIONS(7817),
    [anon_sym_alignas] = ACTIONS(7817),
    [anon_sym__Alignas] = ACTIONS(7817),
    [sym_primitive_type] = ACTIONS(7817),
    [anon_sym_enum] = ACTIONS(7817),
    [anon_sym_struct] = ACTIONS(7817),
    [anon_sym_union] = ACTIONS(7817),
    [anon_sym_in] = ACTIONS(7817),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7817),
    [anon_sym___typeof] = ACTIONS(7817),
    [anon_sym_typeof] = ACTIONS(7817),
    [anon_sym_BOOL] = ACTIONS(7817),
    [anon_sym_IMP] = ACTIONS(7817),
    [anon_sym_SEL] = ACTIONS(7817),
    [anon_sym_Class] = ACTIONS(7817),
    [anon_sym_id] = ACTIONS(7817),
    [anon_sym_out] = ACTIONS(7817),
    [anon_sym_inout] = ACTIONS(7817),
    [anon_sym_bycopy] = ACTIONS(7817),
    [anon_sym_byref] = ACTIONS(7817),
    [anon_sym_oneway] = ACTIONS(7817),
  },
  [STATE(3845)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4044),
    [anon_sym_COMMA] = ACTIONS(4044),
    [anon_sym_RPAREN] = ACTIONS(4044),
    [anon_sym_LPAREN2] = ACTIONS(2920),
    [anon_sym_BANG] = ACTIONS(2914),
    [anon_sym_TILDE] = ACTIONS(2920),
    [anon_sym_DASH] = ACTIONS(2914),
    [anon_sym_PLUS] = ACTIONS(2914),
    [anon_sym_STAR] = ACTIONS(2920),
    [anon_sym_SLASH] = ACTIONS(4042),
    [anon_sym_PERCENT] = ACTIONS(4044),
    [anon_sym_PIPE_PIPE] = ACTIONS(4044),
    [anon_sym_AMP_AMP] = ACTIONS(4044),
    [anon_sym_PIPE] = ACTIONS(4042),
    [anon_sym_CARET] = ACTIONS(2920),
    [anon_sym_AMP] = ACTIONS(2914),
    [anon_sym_EQ_EQ] = ACTIONS(4044),
    [anon_sym_BANG_EQ] = ACTIONS(4044),
    [anon_sym_GT] = ACTIONS(4042),
    [anon_sym_GT_EQ] = ACTIONS(4044),
    [anon_sym_LT_EQ] = ACTIONS(4044),
    [anon_sym_LT] = ACTIONS(4042),
    [anon_sym_LT_LT] = ACTIONS(4044),
    [anon_sym_GT_GT] = ACTIONS(4044),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2920),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [anon_sym_QMARK] = ACTIONS(4044),
    [anon_sym_DASH_DASH] = ACTIONS(2920),
    [anon_sym_PLUS_PLUS] = ACTIONS(2920),
    [anon_sym_sizeof] = ACTIONS(2914),
    [anon_sym___alignof__] = ACTIONS(2914),
    [anon_sym___alignof] = ACTIONS(2914),
    [anon_sym__alignof] = ACTIONS(2914),
    [anon_sym_alignof] = ACTIONS(2914),
    [anon_sym__Alignof] = ACTIONS(2914),
    [anon_sym_offsetof] = ACTIONS(2914),
    [anon_sym__Generic] = ACTIONS(2914),
    [anon_sym_asm] = ACTIONS(2914),
    [anon_sym___asm__] = ACTIONS(2914),
    [anon_sym___asm] = ACTIONS(2914),
    [anon_sym_DOT] = ACTIONS(4042),
    [anon_sym_DASH_GT] = ACTIONS(4044),
    [sym_number_literal] = ACTIONS(2920),
    [anon_sym_L_SQUOTE] = ACTIONS(2920),
    [anon_sym_u_SQUOTE] = ACTIONS(2920),
    [anon_sym_U_SQUOTE] = ACTIONS(2920),
    [anon_sym_u8_SQUOTE] = ACTIONS(2920),
    [anon_sym_SQUOTE] = ACTIONS(2920),
    [anon_sym_AT] = ACTIONS(2914),
    [anon_sym_DQUOTE] = ACTIONS(2920),
    [anon_sym_L_DQUOTE] = ACTIONS(2920),
    [anon_sym_u_DQUOTE] = ACTIONS(2920),
    [anon_sym_U_DQUOTE] = ACTIONS(2920),
    [anon_sym_u8_DQUOTE] = ACTIONS(2920),
    [sym_true] = ACTIONS(2914),
    [sym_false] = ACTIONS(2914),
    [anon_sym_NULL] = ACTIONS(2914),
    [anon_sym_nullptr] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_ATselector] = ACTIONS(2920),
    [anon_sym_ATavailable] = ACTIONS(2920),
    [anon_sym___builtin_available] = ACTIONS(2914),
    [anon_sym_va_arg] = ACTIONS(2914),
    [anon_sym_ATencode] = ACTIONS(2920),
    [anon_sym_id] = ACTIONS(2914),
  },
  [STATE(3846)] = {
    [sym_identifier] = ACTIONS(7237),
    [anon_sym_COMMA] = ACTIONS(7239),
    [anon_sym_RPAREN] = ACTIONS(7239),
    [anon_sym_LPAREN2] = ACTIONS(7239),
    [anon_sym_STAR] = ACTIONS(7239),
    [anon_sym_CARET] = ACTIONS(7239),
    [anon_sym_GT] = ACTIONS(7239),
    [anon_sym_LT] = ACTIONS(7239),
    [anon_sym___extension__] = ACTIONS(7237),
    [anon_sym___attribute__] = ACTIONS(7237),
    [anon_sym___attribute] = ACTIONS(7237),
    [anon_sym_noreturn] = ACTIONS(7237),
    [anon_sym_LBRACK] = ACTIONS(7239),
    [anon_sym___based] = ACTIONS(7237),
    [anon_sym_LBRACE] = ACTIONS(7239),
    [anon_sym_signed] = ACTIONS(7237),
    [anon_sym_unsigned] = ACTIONS(7237),
    [anon_sym_long] = ACTIONS(7237),
    [anon_sym_short] = ACTIONS(7237),
    [anon_sym_ATautoreleasepool] = ACTIONS(7239),
    [anon_sym_const] = ACTIONS(7237),
    [anon_sym_constexpr] = ACTIONS(7237),
    [anon_sym_volatile] = ACTIONS(7237),
    [anon_sym_restrict] = ACTIONS(7237),
    [anon_sym___restrict__] = ACTIONS(7237),
    [anon_sym__Atomic] = ACTIONS(7237),
    [anon_sym__Noreturn] = ACTIONS(7237),
    [anon_sym__Nonnull] = ACTIONS(7237),
    [anon_sym_nullable] = ACTIONS(7237),
    [anon_sym__Complex] = ACTIONS(7237),
    [anon_sym__Nullable] = ACTIONS(7237),
    [anon_sym__Nullable_result] = ACTIONS(7237),
    [anon_sym__Null_unspecified] = ACTIONS(7237),
    [anon_sym___autoreleasing] = ACTIONS(7237),
    [anon_sym___block] = ACTIONS(7237),
    [anon_sym___bridge] = ACTIONS(7237),
    [anon_sym___bridge_retained] = ACTIONS(7237),
    [anon_sym___bridge_transfer] = ACTIONS(7237),
    [anon_sym___complex] = ACTIONS(7237),
    [anon_sym___const] = ACTIONS(7237),
    [anon_sym___imag] = ACTIONS(7237),
    [anon_sym___kindof] = ACTIONS(7237),
    [anon_sym___nonnull] = ACTIONS(7237),
    [anon_sym___nullable] = ACTIONS(7237),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7237),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7237),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7237),
    [anon_sym___real] = ACTIONS(7237),
    [anon_sym___strong] = ACTIONS(7237),
    [anon_sym___unsafe_unretained] = ACTIONS(7237),
    [anon_sym___unused] = ACTIONS(7237),
    [anon_sym___weak] = ACTIONS(7237),
    [anon_sym_alignas] = ACTIONS(7237),
    [anon_sym__Alignas] = ACTIONS(7237),
    [sym_primitive_type] = ACTIONS(7237),
    [anon_sym_enum] = ACTIONS(7237),
    [anon_sym_COLON] = ACTIONS(7239),
    [anon_sym_struct] = ACTIONS(7237),
    [anon_sym_union] = ACTIONS(7237),
    [anon_sym_in] = ACTIONS(7237),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7237),
    [anon_sym___typeof] = ACTIONS(7237),
    [anon_sym_typeof] = ACTIONS(7237),
    [anon_sym_BOOL] = ACTIONS(7237),
    [anon_sym_IMP] = ACTIONS(7237),
    [anon_sym_SEL] = ACTIONS(7237),
    [anon_sym_Class] = ACTIONS(7237),
    [anon_sym_id] = ACTIONS(7237),
    [anon_sym_out] = ACTIONS(7237),
    [anon_sym_inout] = ACTIONS(7237),
    [anon_sym_bycopy] = ACTIONS(7237),
    [anon_sym_byref] = ACTIONS(7237),
    [anon_sym_oneway] = ACTIONS(7237),
  },
  [STATE(3847)] = {
    [sym_identifier] = ACTIONS(7241),
    [anon_sym_COMMA] = ACTIONS(7243),
    [anon_sym_RPAREN] = ACTIONS(7243),
    [anon_sym_LPAREN2] = ACTIONS(7243),
    [anon_sym_STAR] = ACTIONS(7243),
    [anon_sym_CARET] = ACTIONS(7243),
    [anon_sym_GT] = ACTIONS(7243),
    [anon_sym_LT] = ACTIONS(7243),
    [anon_sym___extension__] = ACTIONS(7241),
    [anon_sym___attribute__] = ACTIONS(7241),
    [anon_sym___attribute] = ACTIONS(7241),
    [anon_sym_noreturn] = ACTIONS(7241),
    [anon_sym_LBRACK] = ACTIONS(7243),
    [anon_sym___based] = ACTIONS(7241),
    [anon_sym_LBRACE] = ACTIONS(7243),
    [anon_sym_signed] = ACTIONS(7241),
    [anon_sym_unsigned] = ACTIONS(7241),
    [anon_sym_long] = ACTIONS(7241),
    [anon_sym_short] = ACTIONS(7241),
    [anon_sym_ATautoreleasepool] = ACTIONS(7243),
    [anon_sym_const] = ACTIONS(7241),
    [anon_sym_constexpr] = ACTIONS(7241),
    [anon_sym_volatile] = ACTIONS(7241),
    [anon_sym_restrict] = ACTIONS(7241),
    [anon_sym___restrict__] = ACTIONS(7241),
    [anon_sym__Atomic] = ACTIONS(7241),
    [anon_sym__Noreturn] = ACTIONS(7241),
    [anon_sym__Nonnull] = ACTIONS(7241),
    [anon_sym_nullable] = ACTIONS(7241),
    [anon_sym__Complex] = ACTIONS(7241),
    [anon_sym__Nullable] = ACTIONS(7241),
    [anon_sym__Nullable_result] = ACTIONS(7241),
    [anon_sym__Null_unspecified] = ACTIONS(7241),
    [anon_sym___autoreleasing] = ACTIONS(7241),
    [anon_sym___block] = ACTIONS(7241),
    [anon_sym___bridge] = ACTIONS(7241),
    [anon_sym___bridge_retained] = ACTIONS(7241),
    [anon_sym___bridge_transfer] = ACTIONS(7241),
    [anon_sym___complex] = ACTIONS(7241),
    [anon_sym___const] = ACTIONS(7241),
    [anon_sym___imag] = ACTIONS(7241),
    [anon_sym___kindof] = ACTIONS(7241),
    [anon_sym___nonnull] = ACTIONS(7241),
    [anon_sym___nullable] = ACTIONS(7241),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7241),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7241),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7241),
    [anon_sym___real] = ACTIONS(7241),
    [anon_sym___strong] = ACTIONS(7241),
    [anon_sym___unsafe_unretained] = ACTIONS(7241),
    [anon_sym___unused] = ACTIONS(7241),
    [anon_sym___weak] = ACTIONS(7241),
    [anon_sym_alignas] = ACTIONS(7241),
    [anon_sym__Alignas] = ACTIONS(7241),
    [sym_primitive_type] = ACTIONS(7241),
    [anon_sym_enum] = ACTIONS(7241),
    [anon_sym_COLON] = ACTIONS(7243),
    [anon_sym_struct] = ACTIONS(7241),
    [anon_sym_union] = ACTIONS(7241),
    [anon_sym_in] = ACTIONS(7241),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7241),
    [anon_sym___typeof] = ACTIONS(7241),
    [anon_sym_typeof] = ACTIONS(7241),
    [anon_sym_BOOL] = ACTIONS(7241),
    [anon_sym_IMP] = ACTIONS(7241),
    [anon_sym_SEL] = ACTIONS(7241),
    [anon_sym_Class] = ACTIONS(7241),
    [anon_sym_id] = ACTIONS(7241),
    [anon_sym_out] = ACTIONS(7241),
    [anon_sym_inout] = ACTIONS(7241),
    [anon_sym_bycopy] = ACTIONS(7241),
    [anon_sym_byref] = ACTIONS(7241),
    [anon_sym_oneway] = ACTIONS(7241),
  },
  [STATE(3848)] = {
    [sym_identifier] = ACTIONS(7153),
    [anon_sym_COMMA] = ACTIONS(7155),
    [anon_sym_RPAREN] = ACTIONS(7155),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_STAR] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_GT] = ACTIONS(7155),
    [anon_sym_LT] = ACTIONS(7155),
    [anon_sym___extension__] = ACTIONS(7153),
    [anon_sym___attribute__] = ACTIONS(7153),
    [anon_sym___attribute] = ACTIONS(7153),
    [anon_sym_noreturn] = ACTIONS(7153),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym___based] = ACTIONS(7153),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_signed] = ACTIONS(7153),
    [anon_sym_unsigned] = ACTIONS(7153),
    [anon_sym_long] = ACTIONS(7153),
    [anon_sym_short] = ACTIONS(7153),
    [anon_sym_ATautoreleasepool] = ACTIONS(7155),
    [anon_sym_const] = ACTIONS(7153),
    [anon_sym_constexpr] = ACTIONS(7153),
    [anon_sym_volatile] = ACTIONS(7153),
    [anon_sym_restrict] = ACTIONS(7153),
    [anon_sym___restrict__] = ACTIONS(7153),
    [anon_sym__Atomic] = ACTIONS(7153),
    [anon_sym__Noreturn] = ACTIONS(7153),
    [anon_sym__Nonnull] = ACTIONS(7153),
    [anon_sym_nullable] = ACTIONS(7153),
    [anon_sym__Complex] = ACTIONS(7153),
    [anon_sym__Nullable] = ACTIONS(7153),
    [anon_sym__Nullable_result] = ACTIONS(7153),
    [anon_sym__Null_unspecified] = ACTIONS(7153),
    [anon_sym___autoreleasing] = ACTIONS(7153),
    [anon_sym___block] = ACTIONS(7153),
    [anon_sym___bridge] = ACTIONS(7153),
    [anon_sym___bridge_retained] = ACTIONS(7153),
    [anon_sym___bridge_transfer] = ACTIONS(7153),
    [anon_sym___complex] = ACTIONS(7153),
    [anon_sym___const] = ACTIONS(7153),
    [anon_sym___imag] = ACTIONS(7153),
    [anon_sym___kindof] = ACTIONS(7153),
    [anon_sym___nonnull] = ACTIONS(7153),
    [anon_sym___nullable] = ACTIONS(7153),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7153),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7153),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7153),
    [anon_sym___real] = ACTIONS(7153),
    [anon_sym___strong] = ACTIONS(7153),
    [anon_sym___unsafe_unretained] = ACTIONS(7153),
    [anon_sym___unused] = ACTIONS(7153),
    [anon_sym___weak] = ACTIONS(7153),
    [anon_sym_alignas] = ACTIONS(7153),
    [anon_sym__Alignas] = ACTIONS(7153),
    [sym_primitive_type] = ACTIONS(7153),
    [anon_sym_enum] = ACTIONS(7153),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_struct] = ACTIONS(7153),
    [anon_sym_union] = ACTIONS(7153),
    [anon_sym_in] = ACTIONS(7153),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7153),
    [anon_sym___typeof] = ACTIONS(7153),
    [anon_sym_typeof] = ACTIONS(7153),
    [anon_sym_BOOL] = ACTIONS(7153),
    [anon_sym_IMP] = ACTIONS(7153),
    [anon_sym_SEL] = ACTIONS(7153),
    [anon_sym_Class] = ACTIONS(7153),
    [anon_sym_id] = ACTIONS(7153),
    [anon_sym_out] = ACTIONS(7153),
    [anon_sym_inout] = ACTIONS(7153),
    [anon_sym_bycopy] = ACTIONS(7153),
    [anon_sym_byref] = ACTIONS(7153),
    [anon_sym_oneway] = ACTIONS(7153),
  },
  [STATE(3849)] = {
    [sym_identifier] = ACTIONS(4826),
    [anon_sym_COMMA] = ACTIONS(4828),
    [anon_sym_RPAREN] = ACTIONS(4828),
    [anon_sym_LPAREN2] = ACTIONS(4828),
    [anon_sym_STAR] = ACTIONS(4828),
    [anon_sym_CARET] = ACTIONS(4828),
    [anon_sym_GT] = ACTIONS(4828),
    [anon_sym_LT] = ACTIONS(4828),
    [anon_sym___extension__] = ACTIONS(4826),
    [anon_sym___attribute__] = ACTIONS(4826),
    [anon_sym___attribute] = ACTIONS(4826),
    [anon_sym_noreturn] = ACTIONS(4826),
    [anon_sym_LBRACK] = ACTIONS(4828),
    [anon_sym___based] = ACTIONS(4826),
    [anon_sym_LBRACE] = ACTIONS(4828),
    [anon_sym_signed] = ACTIONS(4826),
    [anon_sym_unsigned] = ACTIONS(4826),
    [anon_sym_long] = ACTIONS(4826),
    [anon_sym_short] = ACTIONS(4826),
    [anon_sym_ATautoreleasepool] = ACTIONS(4828),
    [anon_sym_const] = ACTIONS(4826),
    [anon_sym_constexpr] = ACTIONS(4826),
    [anon_sym_volatile] = ACTIONS(4826),
    [anon_sym_restrict] = ACTIONS(4826),
    [anon_sym___restrict__] = ACTIONS(4826),
    [anon_sym__Atomic] = ACTIONS(4826),
    [anon_sym__Noreturn] = ACTIONS(4826),
    [anon_sym__Nonnull] = ACTIONS(4826),
    [anon_sym_nullable] = ACTIONS(4826),
    [anon_sym__Complex] = ACTIONS(4826),
    [anon_sym__Nullable] = ACTIONS(4826),
    [anon_sym__Nullable_result] = ACTIONS(4826),
    [anon_sym__Null_unspecified] = ACTIONS(4826),
    [anon_sym___autoreleasing] = ACTIONS(4826),
    [anon_sym___block] = ACTIONS(4826),
    [anon_sym___bridge] = ACTIONS(4826),
    [anon_sym___bridge_retained] = ACTIONS(4826),
    [anon_sym___bridge_transfer] = ACTIONS(4826),
    [anon_sym___complex] = ACTIONS(4826),
    [anon_sym___const] = ACTIONS(4826),
    [anon_sym___imag] = ACTIONS(4826),
    [anon_sym___kindof] = ACTIONS(4826),
    [anon_sym___nonnull] = ACTIONS(4826),
    [anon_sym___nullable] = ACTIONS(4826),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4826),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4826),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4826),
    [anon_sym___real] = ACTIONS(4826),
    [anon_sym___strong] = ACTIONS(4826),
    [anon_sym___unsafe_unretained] = ACTIONS(4826),
    [anon_sym___unused] = ACTIONS(4826),
    [anon_sym___weak] = ACTIONS(4826),
    [anon_sym_alignas] = ACTIONS(4826),
    [anon_sym__Alignas] = ACTIONS(4826),
    [sym_primitive_type] = ACTIONS(4826),
    [anon_sym_enum] = ACTIONS(4826),
    [anon_sym_COLON] = ACTIONS(4828),
    [anon_sym_struct] = ACTIONS(4826),
    [anon_sym_union] = ACTIONS(4826),
    [anon_sym_in] = ACTIONS(4826),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4826),
    [anon_sym___typeof] = ACTIONS(4826),
    [anon_sym_typeof] = ACTIONS(4826),
    [anon_sym_BOOL] = ACTIONS(4826),
    [anon_sym_IMP] = ACTIONS(4826),
    [anon_sym_SEL] = ACTIONS(4826),
    [anon_sym_Class] = ACTIONS(4826),
    [anon_sym_id] = ACTIONS(4826),
    [anon_sym_out] = ACTIONS(4826),
    [anon_sym_inout] = ACTIONS(4826),
    [anon_sym_bycopy] = ACTIONS(4826),
    [anon_sym_byref] = ACTIONS(4826),
    [anon_sym_oneway] = ACTIONS(4826),
  },
  [STATE(3850)] = {
    [sym_attribute_specifier] = STATE(3950),
    [sym_identifier] = ACTIONS(7826),
    [anon_sym_COMMA] = ACTIONS(7828),
    [anon_sym_RPAREN] = ACTIONS(7828),
    [anon_sym_LPAREN2] = ACTIONS(7828),
    [anon_sym_STAR] = ACTIONS(7828),
    [anon_sym_CARET] = ACTIONS(7828),
    [anon_sym_GT] = ACTIONS(7828),
    [anon_sym_LT] = ACTIONS(7828),
    [anon_sym___extension__] = ACTIONS(7826),
    [anon_sym___attribute__] = ACTIONS(8863),
    [anon_sym___attribute] = ACTIONS(8863),
    [anon_sym_noreturn] = ACTIONS(7826),
    [anon_sym_LBRACK] = ACTIONS(7828),
    [anon_sym___based] = ACTIONS(7826),
    [anon_sym_LBRACE] = ACTIONS(7828),
    [anon_sym_signed] = ACTIONS(7826),
    [anon_sym_unsigned] = ACTIONS(7826),
    [anon_sym_long] = ACTIONS(7826),
    [anon_sym_short] = ACTIONS(7826),
    [anon_sym_ATautoreleasepool] = ACTIONS(7828),
    [anon_sym_const] = ACTIONS(7826),
    [anon_sym_constexpr] = ACTIONS(7826),
    [anon_sym_volatile] = ACTIONS(7826),
    [anon_sym_restrict] = ACTIONS(7826),
    [anon_sym___restrict__] = ACTIONS(7826),
    [anon_sym__Atomic] = ACTIONS(7826),
    [anon_sym__Noreturn] = ACTIONS(7826),
    [anon_sym__Nonnull] = ACTIONS(7826),
    [anon_sym_nullable] = ACTIONS(7826),
    [anon_sym__Complex] = ACTIONS(7826),
    [anon_sym__Nullable] = ACTIONS(7826),
    [anon_sym__Nullable_result] = ACTIONS(7826),
    [anon_sym__Null_unspecified] = ACTIONS(7826),
    [anon_sym___autoreleasing] = ACTIONS(7826),
    [anon_sym___block] = ACTIONS(7826),
    [anon_sym___bridge] = ACTIONS(7826),
    [anon_sym___bridge_retained] = ACTIONS(7826),
    [anon_sym___bridge_transfer] = ACTIONS(7826),
    [anon_sym___complex] = ACTIONS(7826),
    [anon_sym___const] = ACTIONS(7826),
    [anon_sym___imag] = ACTIONS(7826),
    [anon_sym___kindof] = ACTIONS(7826),
    [anon_sym___nonnull] = ACTIONS(7826),
    [anon_sym___nullable] = ACTIONS(7826),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7826),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7826),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7826),
    [anon_sym___real] = ACTIONS(7826),
    [anon_sym___strong] = ACTIONS(7826),
    [anon_sym___unsafe_unretained] = ACTIONS(7826),
    [anon_sym___unused] = ACTIONS(7826),
    [anon_sym___weak] = ACTIONS(7826),
    [anon_sym_alignas] = ACTIONS(7826),
    [anon_sym__Alignas] = ACTIONS(7826),
    [sym_primitive_type] = ACTIONS(7826),
    [anon_sym_enum] = ACTIONS(7826),
    [anon_sym_struct] = ACTIONS(7826),
    [anon_sym_union] = ACTIONS(7826),
    [anon_sym_in] = ACTIONS(7826),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7826),
    [anon_sym___typeof] = ACTIONS(7826),
    [anon_sym_typeof] = ACTIONS(7826),
    [anon_sym_BOOL] = ACTIONS(7826),
    [anon_sym_IMP] = ACTIONS(7826),
    [anon_sym_SEL] = ACTIONS(7826),
    [anon_sym_Class] = ACTIONS(7826),
    [anon_sym_id] = ACTIONS(7826),
    [anon_sym_out] = ACTIONS(7826),
    [anon_sym_inout] = ACTIONS(7826),
    [anon_sym_bycopy] = ACTIONS(7826),
    [anon_sym_byref] = ACTIONS(7826),
    [anon_sym_oneway] = ACTIONS(7826),
  },
  [STATE(3851)] = {
    [sym_attribute_specifier] = STATE(3972),
    [sym_identifier] = ACTIONS(7843),
    [anon_sym_COMMA] = ACTIONS(7845),
    [anon_sym_RPAREN] = ACTIONS(7845),
    [anon_sym_LPAREN2] = ACTIONS(7845),
    [anon_sym_STAR] = ACTIONS(7845),
    [anon_sym_CARET] = ACTIONS(7845),
    [anon_sym_GT] = ACTIONS(7845),
    [anon_sym_LT] = ACTIONS(7845),
    [anon_sym___extension__] = ACTIONS(7843),
    [anon_sym___attribute__] = ACTIONS(8866),
    [anon_sym___attribute] = ACTIONS(8866),
    [anon_sym_noreturn] = ACTIONS(7843),
    [anon_sym_LBRACK] = ACTIONS(7845),
    [anon_sym___based] = ACTIONS(7843),
    [anon_sym_LBRACE] = ACTIONS(7845),
    [anon_sym_signed] = ACTIONS(7843),
    [anon_sym_unsigned] = ACTIONS(7843),
    [anon_sym_long] = ACTIONS(7843),
    [anon_sym_short] = ACTIONS(7843),
    [anon_sym_ATautoreleasepool] = ACTIONS(7845),
    [anon_sym_const] = ACTIONS(7843),
    [anon_sym_constexpr] = ACTIONS(7843),
    [anon_sym_volatile] = ACTIONS(7843),
    [anon_sym_restrict] = ACTIONS(7843),
    [anon_sym___restrict__] = ACTIONS(7843),
    [anon_sym__Atomic] = ACTIONS(7843),
    [anon_sym__Noreturn] = ACTIONS(7843),
    [anon_sym__Nonnull] = ACTIONS(7843),
    [anon_sym_nullable] = ACTIONS(7843),
    [anon_sym__Complex] = ACTIONS(7843),
    [anon_sym__Nullable] = ACTIONS(7843),
    [anon_sym__Nullable_result] = ACTIONS(7843),
    [anon_sym__Null_unspecified] = ACTIONS(7843),
    [anon_sym___autoreleasing] = ACTIONS(7843),
    [anon_sym___block] = ACTIONS(7843),
    [anon_sym___bridge] = ACTIONS(7843),
    [anon_sym___bridge_retained] = ACTIONS(7843),
    [anon_sym___bridge_transfer] = ACTIONS(7843),
    [anon_sym___complex] = ACTIONS(7843),
    [anon_sym___const] = ACTIONS(7843),
    [anon_sym___imag] = ACTIONS(7843),
    [anon_sym___kindof] = ACTIONS(7843),
    [anon_sym___nonnull] = ACTIONS(7843),
    [anon_sym___nullable] = ACTIONS(7843),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7843),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7843),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7843),
    [anon_sym___real] = ACTIONS(7843),
    [anon_sym___strong] = ACTIONS(7843),
    [anon_sym___unsafe_unretained] = ACTIONS(7843),
    [anon_sym___unused] = ACTIONS(7843),
    [anon_sym___weak] = ACTIONS(7843),
    [anon_sym_alignas] = ACTIONS(7843),
    [anon_sym__Alignas] = ACTIONS(7843),
    [sym_primitive_type] = ACTIONS(7843),
    [anon_sym_enum] = ACTIONS(7843),
    [anon_sym_struct] = ACTIONS(7843),
    [anon_sym_union] = ACTIONS(7843),
    [anon_sym_in] = ACTIONS(7843),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7843),
    [anon_sym___typeof] = ACTIONS(7843),
    [anon_sym_typeof] = ACTIONS(7843),
    [anon_sym_BOOL] = ACTIONS(7843),
    [anon_sym_IMP] = ACTIONS(7843),
    [anon_sym_SEL] = ACTIONS(7843),
    [anon_sym_Class] = ACTIONS(7843),
    [anon_sym_id] = ACTIONS(7843),
    [anon_sym_out] = ACTIONS(7843),
    [anon_sym_inout] = ACTIONS(7843),
    [anon_sym_bycopy] = ACTIONS(7843),
    [anon_sym_byref] = ACTIONS(7843),
    [anon_sym_oneway] = ACTIONS(7843),
  },
  [STATE(3852)] = {
    [sym_identifier] = ACTIONS(7229),
    [anon_sym_COMMA] = ACTIONS(7231),
    [anon_sym_RPAREN] = ACTIONS(7231),
    [anon_sym_LPAREN2] = ACTIONS(7231),
    [anon_sym_STAR] = ACTIONS(7231),
    [anon_sym_CARET] = ACTIONS(7231),
    [anon_sym_GT] = ACTIONS(7231),
    [anon_sym_LT] = ACTIONS(7231),
    [anon_sym___extension__] = ACTIONS(7229),
    [anon_sym___attribute__] = ACTIONS(7229),
    [anon_sym___attribute] = ACTIONS(7229),
    [anon_sym_noreturn] = ACTIONS(7229),
    [anon_sym_LBRACK] = ACTIONS(7231),
    [anon_sym___based] = ACTIONS(7229),
    [anon_sym_LBRACE] = ACTIONS(7231),
    [anon_sym_signed] = ACTIONS(7229),
    [anon_sym_unsigned] = ACTIONS(7229),
    [anon_sym_long] = ACTIONS(7229),
    [anon_sym_short] = ACTIONS(7229),
    [anon_sym_ATautoreleasepool] = ACTIONS(7231),
    [anon_sym_const] = ACTIONS(7229),
    [anon_sym_constexpr] = ACTIONS(7229),
    [anon_sym_volatile] = ACTIONS(7229),
    [anon_sym_restrict] = ACTIONS(7229),
    [anon_sym___restrict__] = ACTIONS(7229),
    [anon_sym__Atomic] = ACTIONS(7229),
    [anon_sym__Noreturn] = ACTIONS(7229),
    [anon_sym__Nonnull] = ACTIONS(7229),
    [anon_sym_nullable] = ACTIONS(7229),
    [anon_sym__Complex] = ACTIONS(7229),
    [anon_sym__Nullable] = ACTIONS(7229),
    [anon_sym__Nullable_result] = ACTIONS(7229),
    [anon_sym__Null_unspecified] = ACTIONS(7229),
    [anon_sym___autoreleasing] = ACTIONS(7229),
    [anon_sym___block] = ACTIONS(7229),
    [anon_sym___bridge] = ACTIONS(7229),
    [anon_sym___bridge_retained] = ACTIONS(7229),
    [anon_sym___bridge_transfer] = ACTIONS(7229),
    [anon_sym___complex] = ACTIONS(7229),
    [anon_sym___const] = ACTIONS(7229),
    [anon_sym___imag] = ACTIONS(7229),
    [anon_sym___kindof] = ACTIONS(7229),
    [anon_sym___nonnull] = ACTIONS(7229),
    [anon_sym___nullable] = ACTIONS(7229),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7229),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7229),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7229),
    [anon_sym___real] = ACTIONS(7229),
    [anon_sym___strong] = ACTIONS(7229),
    [anon_sym___unsafe_unretained] = ACTIONS(7229),
    [anon_sym___unused] = ACTIONS(7229),
    [anon_sym___weak] = ACTIONS(7229),
    [anon_sym_alignas] = ACTIONS(7229),
    [anon_sym__Alignas] = ACTIONS(7229),
    [sym_primitive_type] = ACTIONS(7229),
    [anon_sym_enum] = ACTIONS(7229),
    [anon_sym_COLON] = ACTIONS(7231),
    [anon_sym_struct] = ACTIONS(7229),
    [anon_sym_union] = ACTIONS(7229),
    [anon_sym_in] = ACTIONS(7229),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7229),
    [anon_sym___typeof] = ACTIONS(7229),
    [anon_sym_typeof] = ACTIONS(7229),
    [anon_sym_BOOL] = ACTIONS(7229),
    [anon_sym_IMP] = ACTIONS(7229),
    [anon_sym_SEL] = ACTIONS(7229),
    [anon_sym_Class] = ACTIONS(7229),
    [anon_sym_id] = ACTIONS(7229),
    [anon_sym_out] = ACTIONS(7229),
    [anon_sym_inout] = ACTIONS(7229),
    [anon_sym_bycopy] = ACTIONS(7229),
    [anon_sym_byref] = ACTIONS(7229),
    [anon_sym_oneway] = ACTIONS(7229),
  },
  [STATE(3853)] = {
    [sym_identifier] = ACTIONS(7447),
    [anon_sym_COMMA] = ACTIONS(7449),
    [anon_sym_RPAREN] = ACTIONS(7449),
    [anon_sym_LPAREN2] = ACTIONS(7449),
    [anon_sym_STAR] = ACTIONS(7449),
    [anon_sym_CARET] = ACTIONS(7449),
    [anon_sym_GT] = ACTIONS(7449),
    [anon_sym_LT] = ACTIONS(7449),
    [anon_sym___extension__] = ACTIONS(7447),
    [anon_sym___attribute__] = ACTIONS(7447),
    [anon_sym___attribute] = ACTIONS(7447),
    [anon_sym_noreturn] = ACTIONS(7447),
    [anon_sym_LBRACK] = ACTIONS(7449),
    [anon_sym___based] = ACTIONS(7447),
    [anon_sym_LBRACE] = ACTIONS(7449),
    [anon_sym_signed] = ACTIONS(7447),
    [anon_sym_unsigned] = ACTIONS(7447),
    [anon_sym_long] = ACTIONS(7447),
    [anon_sym_short] = ACTIONS(7447),
    [anon_sym_ATautoreleasepool] = ACTIONS(7449),
    [anon_sym_const] = ACTIONS(7447),
    [anon_sym_constexpr] = ACTIONS(7447),
    [anon_sym_volatile] = ACTIONS(7447),
    [anon_sym_restrict] = ACTIONS(7447),
    [anon_sym___restrict__] = ACTIONS(7447),
    [anon_sym__Atomic] = ACTIONS(7447),
    [anon_sym__Noreturn] = ACTIONS(7447),
    [anon_sym__Nonnull] = ACTIONS(7447),
    [anon_sym_nullable] = ACTIONS(7447),
    [anon_sym__Complex] = ACTIONS(7447),
    [anon_sym__Nullable] = ACTIONS(7447),
    [anon_sym__Nullable_result] = ACTIONS(7447),
    [anon_sym__Null_unspecified] = ACTIONS(7447),
    [anon_sym___autoreleasing] = ACTIONS(7447),
    [anon_sym___block] = ACTIONS(7447),
    [anon_sym___bridge] = ACTIONS(7447),
    [anon_sym___bridge_retained] = ACTIONS(7447),
    [anon_sym___bridge_transfer] = ACTIONS(7447),
    [anon_sym___complex] = ACTIONS(7447),
    [anon_sym___const] = ACTIONS(7447),
    [anon_sym___imag] = ACTIONS(7447),
    [anon_sym___kindof] = ACTIONS(7447),
    [anon_sym___nonnull] = ACTIONS(7447),
    [anon_sym___nullable] = ACTIONS(7447),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7447),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7447),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7447),
    [anon_sym___real] = ACTIONS(7447),
    [anon_sym___strong] = ACTIONS(7447),
    [anon_sym___unsafe_unretained] = ACTIONS(7447),
    [anon_sym___unused] = ACTIONS(7447),
    [anon_sym___weak] = ACTIONS(7447),
    [anon_sym_alignas] = ACTIONS(7447),
    [anon_sym__Alignas] = ACTIONS(7447),
    [sym_primitive_type] = ACTIONS(7447),
    [anon_sym_enum] = ACTIONS(7447),
    [anon_sym_COLON] = ACTIONS(7449),
    [anon_sym_struct] = ACTIONS(7447),
    [anon_sym_union] = ACTIONS(7447),
    [anon_sym_in] = ACTIONS(7447),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7447),
    [anon_sym___typeof] = ACTIONS(7447),
    [anon_sym_typeof] = ACTIONS(7447),
    [anon_sym_BOOL] = ACTIONS(7447),
    [anon_sym_IMP] = ACTIONS(7447),
    [anon_sym_SEL] = ACTIONS(7447),
    [anon_sym_Class] = ACTIONS(7447),
    [anon_sym_id] = ACTIONS(7447),
    [anon_sym_out] = ACTIONS(7447),
    [anon_sym_inout] = ACTIONS(7447),
    [anon_sym_bycopy] = ACTIONS(7447),
    [anon_sym_byref] = ACTIONS(7447),
    [anon_sym_oneway] = ACTIONS(7447),
  },
  [STATE(3854)] = {
    [sym_identifier] = ACTIONS(7201),
    [anon_sym_COMMA] = ACTIONS(7203),
    [anon_sym_RPAREN] = ACTIONS(7203),
    [anon_sym_LPAREN2] = ACTIONS(7203),
    [anon_sym_STAR] = ACTIONS(7203),
    [anon_sym_CARET] = ACTIONS(7203),
    [anon_sym_GT] = ACTIONS(7203),
    [anon_sym_LT] = ACTIONS(7203),
    [anon_sym___extension__] = ACTIONS(7201),
    [anon_sym___attribute__] = ACTIONS(7201),
    [anon_sym___attribute] = ACTIONS(7201),
    [anon_sym_noreturn] = ACTIONS(7201),
    [anon_sym_LBRACK] = ACTIONS(7203),
    [anon_sym___based] = ACTIONS(7201),
    [anon_sym_LBRACE] = ACTIONS(7203),
    [anon_sym_signed] = ACTIONS(7201),
    [anon_sym_unsigned] = ACTIONS(7201),
    [anon_sym_long] = ACTIONS(7201),
    [anon_sym_short] = ACTIONS(7201),
    [anon_sym_ATautoreleasepool] = ACTIONS(7203),
    [anon_sym_const] = ACTIONS(7201),
    [anon_sym_constexpr] = ACTIONS(7201),
    [anon_sym_volatile] = ACTIONS(7201),
    [anon_sym_restrict] = ACTIONS(7201),
    [anon_sym___restrict__] = ACTIONS(7201),
    [anon_sym__Atomic] = ACTIONS(7201),
    [anon_sym__Noreturn] = ACTIONS(7201),
    [anon_sym__Nonnull] = ACTIONS(7201),
    [anon_sym_nullable] = ACTIONS(7201),
    [anon_sym__Complex] = ACTIONS(7201),
    [anon_sym__Nullable] = ACTIONS(7201),
    [anon_sym__Nullable_result] = ACTIONS(7201),
    [anon_sym__Null_unspecified] = ACTIONS(7201),
    [anon_sym___autoreleasing] = ACTIONS(7201),
    [anon_sym___block] = ACTIONS(7201),
    [anon_sym___bridge] = ACTIONS(7201),
    [anon_sym___bridge_retained] = ACTIONS(7201),
    [anon_sym___bridge_transfer] = ACTIONS(7201),
    [anon_sym___complex] = ACTIONS(7201),
    [anon_sym___const] = ACTIONS(7201),
    [anon_sym___imag] = ACTIONS(7201),
    [anon_sym___kindof] = ACTIONS(7201),
    [anon_sym___nonnull] = ACTIONS(7201),
    [anon_sym___nullable] = ACTIONS(7201),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7201),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7201),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7201),
    [anon_sym___real] = ACTIONS(7201),
    [anon_sym___strong] = ACTIONS(7201),
    [anon_sym___unsafe_unretained] = ACTIONS(7201),
    [anon_sym___unused] = ACTIONS(7201),
    [anon_sym___weak] = ACTIONS(7201),
    [anon_sym_alignas] = ACTIONS(7201),
    [anon_sym__Alignas] = ACTIONS(7201),
    [sym_primitive_type] = ACTIONS(7201),
    [anon_sym_enum] = ACTIONS(7201),
    [anon_sym_COLON] = ACTIONS(7203),
    [anon_sym_struct] = ACTIONS(7201),
    [anon_sym_union] = ACTIONS(7201),
    [anon_sym_in] = ACTIONS(7201),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7201),
    [anon_sym___typeof] = ACTIONS(7201),
    [anon_sym_typeof] = ACTIONS(7201),
    [anon_sym_BOOL] = ACTIONS(7201),
    [anon_sym_IMP] = ACTIONS(7201),
    [anon_sym_SEL] = ACTIONS(7201),
    [anon_sym_Class] = ACTIONS(7201),
    [anon_sym_id] = ACTIONS(7201),
    [anon_sym_out] = ACTIONS(7201),
    [anon_sym_inout] = ACTIONS(7201),
    [anon_sym_bycopy] = ACTIONS(7201),
    [anon_sym_byref] = ACTIONS(7201),
    [anon_sym_oneway] = ACTIONS(7201),
  },
  [STATE(3855)] = {
    [sym_identifier] = ACTIONS(7217),
    [anon_sym_COMMA] = ACTIONS(7219),
    [anon_sym_RPAREN] = ACTIONS(7219),
    [anon_sym_LPAREN2] = ACTIONS(7219),
    [anon_sym_STAR] = ACTIONS(7219),
    [anon_sym_CARET] = ACTIONS(7219),
    [anon_sym_GT] = ACTIONS(7219),
    [anon_sym_LT] = ACTIONS(7219),
    [anon_sym___extension__] = ACTIONS(7217),
    [anon_sym___attribute__] = ACTIONS(7217),
    [anon_sym___attribute] = ACTIONS(7217),
    [anon_sym_noreturn] = ACTIONS(7217),
    [anon_sym_LBRACK] = ACTIONS(7219),
    [anon_sym___based] = ACTIONS(7217),
    [anon_sym_LBRACE] = ACTIONS(7219),
    [anon_sym_signed] = ACTIONS(7217),
    [anon_sym_unsigned] = ACTIONS(7217),
    [anon_sym_long] = ACTIONS(7217),
    [anon_sym_short] = ACTIONS(7217),
    [anon_sym_ATautoreleasepool] = ACTIONS(7219),
    [anon_sym_const] = ACTIONS(7217),
    [anon_sym_constexpr] = ACTIONS(7217),
    [anon_sym_volatile] = ACTIONS(7217),
    [anon_sym_restrict] = ACTIONS(7217),
    [anon_sym___restrict__] = ACTIONS(7217),
    [anon_sym__Atomic] = ACTIONS(7217),
    [anon_sym__Noreturn] = ACTIONS(7217),
    [anon_sym__Nonnull] = ACTIONS(7217),
    [anon_sym_nullable] = ACTIONS(7217),
    [anon_sym__Complex] = ACTIONS(7217),
    [anon_sym__Nullable] = ACTIONS(7217),
    [anon_sym__Nullable_result] = ACTIONS(7217),
    [anon_sym__Null_unspecified] = ACTIONS(7217),
    [anon_sym___autoreleasing] = ACTIONS(7217),
    [anon_sym___block] = ACTIONS(7217),
    [anon_sym___bridge] = ACTIONS(7217),
    [anon_sym___bridge_retained] = ACTIONS(7217),
    [anon_sym___bridge_transfer] = ACTIONS(7217),
    [anon_sym___complex] = ACTIONS(7217),
    [anon_sym___const] = ACTIONS(7217),
    [anon_sym___imag] = ACTIONS(7217),
    [anon_sym___kindof] = ACTIONS(7217),
    [anon_sym___nonnull] = ACTIONS(7217),
    [anon_sym___nullable] = ACTIONS(7217),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7217),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7217),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7217),
    [anon_sym___real] = ACTIONS(7217),
    [anon_sym___strong] = ACTIONS(7217),
    [anon_sym___unsafe_unretained] = ACTIONS(7217),
    [anon_sym___unused] = ACTIONS(7217),
    [anon_sym___weak] = ACTIONS(7217),
    [anon_sym_alignas] = ACTIONS(7217),
    [anon_sym__Alignas] = ACTIONS(7217),
    [sym_primitive_type] = ACTIONS(7217),
    [anon_sym_enum] = ACTIONS(7217),
    [anon_sym_COLON] = ACTIONS(7219),
    [anon_sym_struct] = ACTIONS(7217),
    [anon_sym_union] = ACTIONS(7217),
    [anon_sym_in] = ACTIONS(7217),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7217),
    [anon_sym___typeof] = ACTIONS(7217),
    [anon_sym_typeof] = ACTIONS(7217),
    [anon_sym_BOOL] = ACTIONS(7217),
    [anon_sym_IMP] = ACTIONS(7217),
    [anon_sym_SEL] = ACTIONS(7217),
    [anon_sym_Class] = ACTIONS(7217),
    [anon_sym_id] = ACTIONS(7217),
    [anon_sym_out] = ACTIONS(7217),
    [anon_sym_inout] = ACTIONS(7217),
    [anon_sym_bycopy] = ACTIONS(7217),
    [anon_sym_byref] = ACTIONS(7217),
    [anon_sym_oneway] = ACTIONS(7217),
  },
  [STATE(3856)] = {
    [sym_identifier] = ACTIONS(7225),
    [anon_sym_COMMA] = ACTIONS(7227),
    [anon_sym_RPAREN] = ACTIONS(7227),
    [anon_sym_LPAREN2] = ACTIONS(7227),
    [anon_sym_STAR] = ACTIONS(7227),
    [anon_sym_CARET] = ACTIONS(7227),
    [anon_sym_GT] = ACTIONS(7227),
    [anon_sym_LT] = ACTIONS(7227),
    [anon_sym___extension__] = ACTIONS(7225),
    [anon_sym___attribute__] = ACTIONS(7225),
    [anon_sym___attribute] = ACTIONS(7225),
    [anon_sym_noreturn] = ACTIONS(7225),
    [anon_sym_LBRACK] = ACTIONS(7227),
    [anon_sym___based] = ACTIONS(7225),
    [anon_sym_LBRACE] = ACTIONS(7227),
    [anon_sym_signed] = ACTIONS(7225),
    [anon_sym_unsigned] = ACTIONS(7225),
    [anon_sym_long] = ACTIONS(7225),
    [anon_sym_short] = ACTIONS(7225),
    [anon_sym_ATautoreleasepool] = ACTIONS(7227),
    [anon_sym_const] = ACTIONS(7225),
    [anon_sym_constexpr] = ACTIONS(7225),
    [anon_sym_volatile] = ACTIONS(7225),
    [anon_sym_restrict] = ACTIONS(7225),
    [anon_sym___restrict__] = ACTIONS(7225),
    [anon_sym__Atomic] = ACTIONS(7225),
    [anon_sym__Noreturn] = ACTIONS(7225),
    [anon_sym__Nonnull] = ACTIONS(7225),
    [anon_sym_nullable] = ACTIONS(7225),
    [anon_sym__Complex] = ACTIONS(7225),
    [anon_sym__Nullable] = ACTIONS(7225),
    [anon_sym__Nullable_result] = ACTIONS(7225),
    [anon_sym__Null_unspecified] = ACTIONS(7225),
    [anon_sym___autoreleasing] = ACTIONS(7225),
    [anon_sym___block] = ACTIONS(7225),
    [anon_sym___bridge] = ACTIONS(7225),
    [anon_sym___bridge_retained] = ACTIONS(7225),
    [anon_sym___bridge_transfer] = ACTIONS(7225),
    [anon_sym___complex] = ACTIONS(7225),
    [anon_sym___const] = ACTIONS(7225),
    [anon_sym___imag] = ACTIONS(7225),
    [anon_sym___kindof] = ACTIONS(7225),
    [anon_sym___nonnull] = ACTIONS(7225),
    [anon_sym___nullable] = ACTIONS(7225),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7225),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7225),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7225),
    [anon_sym___real] = ACTIONS(7225),
    [anon_sym___strong] = ACTIONS(7225),
    [anon_sym___unsafe_unretained] = ACTIONS(7225),
    [anon_sym___unused] = ACTIONS(7225),
    [anon_sym___weak] = ACTIONS(7225),
    [anon_sym_alignas] = ACTIONS(7225),
    [anon_sym__Alignas] = ACTIONS(7225),
    [sym_primitive_type] = ACTIONS(7225),
    [anon_sym_enum] = ACTIONS(7225),
    [anon_sym_COLON] = ACTIONS(7227),
    [anon_sym_struct] = ACTIONS(7225),
    [anon_sym_union] = ACTIONS(7225),
    [anon_sym_in] = ACTIONS(7225),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7225),
    [anon_sym___typeof] = ACTIONS(7225),
    [anon_sym_typeof] = ACTIONS(7225),
    [anon_sym_BOOL] = ACTIONS(7225),
    [anon_sym_IMP] = ACTIONS(7225),
    [anon_sym_SEL] = ACTIONS(7225),
    [anon_sym_Class] = ACTIONS(7225),
    [anon_sym_id] = ACTIONS(7225),
    [anon_sym_out] = ACTIONS(7225),
    [anon_sym_inout] = ACTIONS(7225),
    [anon_sym_bycopy] = ACTIONS(7225),
    [anon_sym_byref] = ACTIONS(7225),
    [anon_sym_oneway] = ACTIONS(7225),
  },
  [STATE(3857)] = {
    [sym_identifier] = ACTIONS(7269),
    [anon_sym_COMMA] = ACTIONS(7271),
    [anon_sym_RPAREN] = ACTIONS(7271),
    [anon_sym_LPAREN2] = ACTIONS(7271),
    [anon_sym_STAR] = ACTIONS(7271),
    [anon_sym_CARET] = ACTIONS(7271),
    [anon_sym_GT] = ACTIONS(7271),
    [anon_sym_LT] = ACTIONS(7271),
    [anon_sym___extension__] = ACTIONS(7269),
    [anon_sym___attribute__] = ACTIONS(7269),
    [anon_sym___attribute] = ACTIONS(7269),
    [anon_sym_noreturn] = ACTIONS(7269),
    [anon_sym_LBRACK] = ACTIONS(7271),
    [anon_sym___based] = ACTIONS(7269),
    [anon_sym_LBRACE] = ACTIONS(7271),
    [anon_sym_signed] = ACTIONS(7269),
    [anon_sym_unsigned] = ACTIONS(7269),
    [anon_sym_long] = ACTIONS(7269),
    [anon_sym_short] = ACTIONS(7269),
    [anon_sym_ATautoreleasepool] = ACTIONS(7271),
    [anon_sym_const] = ACTIONS(7269),
    [anon_sym_constexpr] = ACTIONS(7269),
    [anon_sym_volatile] = ACTIONS(7269),
    [anon_sym_restrict] = ACTIONS(7269),
    [anon_sym___restrict__] = ACTIONS(7269),
    [anon_sym__Atomic] = ACTIONS(7269),
    [anon_sym__Noreturn] = ACTIONS(7269),
    [anon_sym__Nonnull] = ACTIONS(7269),
    [anon_sym_nullable] = ACTIONS(7269),
    [anon_sym__Complex] = ACTIONS(7269),
    [anon_sym__Nullable] = ACTIONS(7269),
    [anon_sym__Nullable_result] = ACTIONS(7269),
    [anon_sym__Null_unspecified] = ACTIONS(7269),
    [anon_sym___autoreleasing] = ACTIONS(7269),
    [anon_sym___block] = ACTIONS(7269),
    [anon_sym___bridge] = ACTIONS(7269),
    [anon_sym___bridge_retained] = ACTIONS(7269),
    [anon_sym___bridge_transfer] = ACTIONS(7269),
    [anon_sym___complex] = ACTIONS(7269),
    [anon_sym___const] = ACTIONS(7269),
    [anon_sym___imag] = ACTIONS(7269),
    [anon_sym___kindof] = ACTIONS(7269),
    [anon_sym___nonnull] = ACTIONS(7269),
    [anon_sym___nullable] = ACTIONS(7269),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7269),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7269),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7269),
    [anon_sym___real] = ACTIONS(7269),
    [anon_sym___strong] = ACTIONS(7269),
    [anon_sym___unsafe_unretained] = ACTIONS(7269),
    [anon_sym___unused] = ACTIONS(7269),
    [anon_sym___weak] = ACTIONS(7269),
    [anon_sym_alignas] = ACTIONS(7269),
    [anon_sym__Alignas] = ACTIONS(7269),
    [sym_primitive_type] = ACTIONS(7269),
    [anon_sym_enum] = ACTIONS(7269),
    [anon_sym_COLON] = ACTIONS(7271),
    [anon_sym_struct] = ACTIONS(7269),
    [anon_sym_union] = ACTIONS(7269),
    [anon_sym_in] = ACTIONS(7269),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7269),
    [anon_sym___typeof] = ACTIONS(7269),
    [anon_sym_typeof] = ACTIONS(7269),
    [anon_sym_BOOL] = ACTIONS(7269),
    [anon_sym_IMP] = ACTIONS(7269),
    [anon_sym_SEL] = ACTIONS(7269),
    [anon_sym_Class] = ACTIONS(7269),
    [anon_sym_id] = ACTIONS(7269),
    [anon_sym_out] = ACTIONS(7269),
    [anon_sym_inout] = ACTIONS(7269),
    [anon_sym_bycopy] = ACTIONS(7269),
    [anon_sym_byref] = ACTIONS(7269),
    [anon_sym_oneway] = ACTIONS(7269),
  },
  [STATE(3858)] = {
    [sym__type_definition_type] = STATE(4728),
    [sym_attribute_specifier] = STATE(3833),
    [sym_ms_declspec_modifier] = STATE(4043),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3859)] = {
    [sym_identifier] = ACTIONS(7451),
    [anon_sym_COMMA] = ACTIONS(7453),
    [anon_sym_RPAREN] = ACTIONS(7453),
    [anon_sym_LPAREN2] = ACTIONS(7453),
    [anon_sym_STAR] = ACTIONS(7453),
    [anon_sym_CARET] = ACTIONS(7453),
    [anon_sym_GT] = ACTIONS(7453),
    [anon_sym_LT] = ACTIONS(7453),
    [anon_sym___extension__] = ACTIONS(7451),
    [anon_sym___attribute__] = ACTIONS(7451),
    [anon_sym___attribute] = ACTIONS(7451),
    [anon_sym_noreturn] = ACTIONS(7451),
    [anon_sym_LBRACK] = ACTIONS(7453),
    [anon_sym___based] = ACTIONS(7451),
    [anon_sym_LBRACE] = ACTIONS(7453),
    [anon_sym_signed] = ACTIONS(7451),
    [anon_sym_unsigned] = ACTIONS(7451),
    [anon_sym_long] = ACTIONS(7451),
    [anon_sym_short] = ACTIONS(7451),
    [anon_sym_ATautoreleasepool] = ACTIONS(7453),
    [anon_sym_const] = ACTIONS(7451),
    [anon_sym_constexpr] = ACTIONS(7451),
    [anon_sym_volatile] = ACTIONS(7451),
    [anon_sym_restrict] = ACTIONS(7451),
    [anon_sym___restrict__] = ACTIONS(7451),
    [anon_sym__Atomic] = ACTIONS(7451),
    [anon_sym__Noreturn] = ACTIONS(7451),
    [anon_sym__Nonnull] = ACTIONS(7451),
    [anon_sym_nullable] = ACTIONS(7451),
    [anon_sym__Complex] = ACTIONS(7451),
    [anon_sym__Nullable] = ACTIONS(7451),
    [anon_sym__Nullable_result] = ACTIONS(7451),
    [anon_sym__Null_unspecified] = ACTIONS(7451),
    [anon_sym___autoreleasing] = ACTIONS(7451),
    [anon_sym___block] = ACTIONS(7451),
    [anon_sym___bridge] = ACTIONS(7451),
    [anon_sym___bridge_retained] = ACTIONS(7451),
    [anon_sym___bridge_transfer] = ACTIONS(7451),
    [anon_sym___complex] = ACTIONS(7451),
    [anon_sym___const] = ACTIONS(7451),
    [anon_sym___imag] = ACTIONS(7451),
    [anon_sym___kindof] = ACTIONS(7451),
    [anon_sym___nonnull] = ACTIONS(7451),
    [anon_sym___nullable] = ACTIONS(7451),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7451),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7451),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7451),
    [anon_sym___real] = ACTIONS(7451),
    [anon_sym___strong] = ACTIONS(7451),
    [anon_sym___unsafe_unretained] = ACTIONS(7451),
    [anon_sym___unused] = ACTIONS(7451),
    [anon_sym___weak] = ACTIONS(7451),
    [anon_sym_alignas] = ACTIONS(7451),
    [anon_sym__Alignas] = ACTIONS(7451),
    [sym_primitive_type] = ACTIONS(7451),
    [anon_sym_enum] = ACTIONS(7451),
    [anon_sym_COLON] = ACTIONS(7453),
    [anon_sym_struct] = ACTIONS(7451),
    [anon_sym_union] = ACTIONS(7451),
    [anon_sym_in] = ACTIONS(7451),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7451),
    [anon_sym___typeof] = ACTIONS(7451),
    [anon_sym_typeof] = ACTIONS(7451),
    [anon_sym_BOOL] = ACTIONS(7451),
    [anon_sym_IMP] = ACTIONS(7451),
    [anon_sym_SEL] = ACTIONS(7451),
    [anon_sym_Class] = ACTIONS(7451),
    [anon_sym_id] = ACTIONS(7451),
    [anon_sym_out] = ACTIONS(7451),
    [anon_sym_inout] = ACTIONS(7451),
    [anon_sym_bycopy] = ACTIONS(7451),
    [anon_sym_byref] = ACTIONS(7451),
    [anon_sym_oneway] = ACTIONS(7451),
  },
  [STATE(3860)] = {
    [sym_identifier] = ACTIONS(7455),
    [anon_sym_COMMA] = ACTIONS(7457),
    [anon_sym_RPAREN] = ACTIONS(7457),
    [anon_sym_LPAREN2] = ACTIONS(7457),
    [anon_sym_STAR] = ACTIONS(7457),
    [anon_sym_CARET] = ACTIONS(7457),
    [anon_sym_GT] = ACTIONS(7457),
    [anon_sym_LT] = ACTIONS(7457),
    [anon_sym___extension__] = ACTIONS(7455),
    [anon_sym___attribute__] = ACTIONS(7455),
    [anon_sym___attribute] = ACTIONS(7455),
    [anon_sym_noreturn] = ACTIONS(7455),
    [anon_sym_LBRACK] = ACTIONS(7457),
    [anon_sym___based] = ACTIONS(7455),
    [anon_sym_LBRACE] = ACTIONS(7457),
    [anon_sym_signed] = ACTIONS(7455),
    [anon_sym_unsigned] = ACTIONS(7455),
    [anon_sym_long] = ACTIONS(7455),
    [anon_sym_short] = ACTIONS(7455),
    [anon_sym_ATautoreleasepool] = ACTIONS(7457),
    [anon_sym_const] = ACTIONS(7455),
    [anon_sym_constexpr] = ACTIONS(7455),
    [anon_sym_volatile] = ACTIONS(7455),
    [anon_sym_restrict] = ACTIONS(7455),
    [anon_sym___restrict__] = ACTIONS(7455),
    [anon_sym__Atomic] = ACTIONS(7455),
    [anon_sym__Noreturn] = ACTIONS(7455),
    [anon_sym__Nonnull] = ACTIONS(7455),
    [anon_sym_nullable] = ACTIONS(7455),
    [anon_sym__Complex] = ACTIONS(7455),
    [anon_sym__Nullable] = ACTIONS(7455),
    [anon_sym__Nullable_result] = ACTIONS(7455),
    [anon_sym__Null_unspecified] = ACTIONS(7455),
    [anon_sym___autoreleasing] = ACTIONS(7455),
    [anon_sym___block] = ACTIONS(7455),
    [anon_sym___bridge] = ACTIONS(7455),
    [anon_sym___bridge_retained] = ACTIONS(7455),
    [anon_sym___bridge_transfer] = ACTIONS(7455),
    [anon_sym___complex] = ACTIONS(7455),
    [anon_sym___const] = ACTIONS(7455),
    [anon_sym___imag] = ACTIONS(7455),
    [anon_sym___kindof] = ACTIONS(7455),
    [anon_sym___nonnull] = ACTIONS(7455),
    [anon_sym___nullable] = ACTIONS(7455),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7455),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7455),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7455),
    [anon_sym___real] = ACTIONS(7455),
    [anon_sym___strong] = ACTIONS(7455),
    [anon_sym___unsafe_unretained] = ACTIONS(7455),
    [anon_sym___unused] = ACTIONS(7455),
    [anon_sym___weak] = ACTIONS(7455),
    [anon_sym_alignas] = ACTIONS(7455),
    [anon_sym__Alignas] = ACTIONS(7455),
    [sym_primitive_type] = ACTIONS(7455),
    [anon_sym_enum] = ACTIONS(7455),
    [anon_sym_COLON] = ACTIONS(7457),
    [anon_sym_struct] = ACTIONS(7455),
    [anon_sym_union] = ACTIONS(7455),
    [anon_sym_in] = ACTIONS(7455),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7455),
    [anon_sym___typeof] = ACTIONS(7455),
    [anon_sym_typeof] = ACTIONS(7455),
    [anon_sym_BOOL] = ACTIONS(7455),
    [anon_sym_IMP] = ACTIONS(7455),
    [anon_sym_SEL] = ACTIONS(7455),
    [anon_sym_Class] = ACTIONS(7455),
    [anon_sym_id] = ACTIONS(7455),
    [anon_sym_out] = ACTIONS(7455),
    [anon_sym_inout] = ACTIONS(7455),
    [anon_sym_bycopy] = ACTIONS(7455),
    [anon_sym_byref] = ACTIONS(7455),
    [anon_sym_oneway] = ACTIONS(7455),
  },
  [STATE(3861)] = {
    [sym_identifier] = ACTIONS(7157),
    [anon_sym_COMMA] = ACTIONS(7159),
    [anon_sym_RPAREN] = ACTIONS(7159),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_STAR] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_GT] = ACTIONS(7159),
    [anon_sym_LT] = ACTIONS(7159),
    [anon_sym___extension__] = ACTIONS(7157),
    [anon_sym___attribute__] = ACTIONS(7157),
    [anon_sym___attribute] = ACTIONS(7157),
    [anon_sym_noreturn] = ACTIONS(7157),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym___based] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_signed] = ACTIONS(7157),
    [anon_sym_unsigned] = ACTIONS(7157),
    [anon_sym_long] = ACTIONS(7157),
    [anon_sym_short] = ACTIONS(7157),
    [anon_sym_ATautoreleasepool] = ACTIONS(7159),
    [anon_sym_const] = ACTIONS(7157),
    [anon_sym_constexpr] = ACTIONS(7157),
    [anon_sym_volatile] = ACTIONS(7157),
    [anon_sym_restrict] = ACTIONS(7157),
    [anon_sym___restrict__] = ACTIONS(7157),
    [anon_sym__Atomic] = ACTIONS(7157),
    [anon_sym__Noreturn] = ACTIONS(7157),
    [anon_sym__Nonnull] = ACTIONS(7157),
    [anon_sym_nullable] = ACTIONS(7157),
    [anon_sym__Complex] = ACTIONS(7157),
    [anon_sym__Nullable] = ACTIONS(7157),
    [anon_sym__Nullable_result] = ACTIONS(7157),
    [anon_sym__Null_unspecified] = ACTIONS(7157),
    [anon_sym___autoreleasing] = ACTIONS(7157),
    [anon_sym___block] = ACTIONS(7157),
    [anon_sym___bridge] = ACTIONS(7157),
    [anon_sym___bridge_retained] = ACTIONS(7157),
    [anon_sym___bridge_transfer] = ACTIONS(7157),
    [anon_sym___complex] = ACTIONS(7157),
    [anon_sym___const] = ACTIONS(7157),
    [anon_sym___imag] = ACTIONS(7157),
    [anon_sym___kindof] = ACTIONS(7157),
    [anon_sym___nonnull] = ACTIONS(7157),
    [anon_sym___nullable] = ACTIONS(7157),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7157),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7157),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7157),
    [anon_sym___real] = ACTIONS(7157),
    [anon_sym___strong] = ACTIONS(7157),
    [anon_sym___unsafe_unretained] = ACTIONS(7157),
    [anon_sym___unused] = ACTIONS(7157),
    [anon_sym___weak] = ACTIONS(7157),
    [anon_sym_alignas] = ACTIONS(7157),
    [anon_sym__Alignas] = ACTIONS(7157),
    [sym_primitive_type] = ACTIONS(7157),
    [anon_sym_enum] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_struct] = ACTIONS(7157),
    [anon_sym_union] = ACTIONS(7157),
    [anon_sym_in] = ACTIONS(7157),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7157),
    [anon_sym___typeof] = ACTIONS(7157),
    [anon_sym_typeof] = ACTIONS(7157),
    [anon_sym_BOOL] = ACTIONS(7157),
    [anon_sym_IMP] = ACTIONS(7157),
    [anon_sym_SEL] = ACTIONS(7157),
    [anon_sym_Class] = ACTIONS(7157),
    [anon_sym_id] = ACTIONS(7157),
    [anon_sym_out] = ACTIONS(7157),
    [anon_sym_inout] = ACTIONS(7157),
    [anon_sym_bycopy] = ACTIONS(7157),
    [anon_sym_byref] = ACTIONS(7157),
    [anon_sym_oneway] = ACTIONS(7157),
  },
  [STATE(3862)] = {
    [sym_identifier] = ACTIONS(7327),
    [anon_sym_COMMA] = ACTIONS(7329),
    [anon_sym_RPAREN] = ACTIONS(7329),
    [anon_sym_LPAREN2] = ACTIONS(7329),
    [anon_sym_STAR] = ACTIONS(7329),
    [anon_sym_CARET] = ACTIONS(7329),
    [anon_sym_GT] = ACTIONS(7329),
    [anon_sym_LT] = ACTIONS(7329),
    [anon_sym___extension__] = ACTIONS(7327),
    [anon_sym___attribute__] = ACTIONS(7327),
    [anon_sym___attribute] = ACTIONS(7327),
    [anon_sym_noreturn] = ACTIONS(7327),
    [anon_sym_LBRACK] = ACTIONS(7329),
    [anon_sym___based] = ACTIONS(7327),
    [anon_sym_LBRACE] = ACTIONS(7329),
    [anon_sym_signed] = ACTIONS(7327),
    [anon_sym_unsigned] = ACTIONS(7327),
    [anon_sym_long] = ACTIONS(7327),
    [anon_sym_short] = ACTIONS(7327),
    [anon_sym_ATautoreleasepool] = ACTIONS(7329),
    [anon_sym_const] = ACTIONS(7327),
    [anon_sym_constexpr] = ACTIONS(7327),
    [anon_sym_volatile] = ACTIONS(7327),
    [anon_sym_restrict] = ACTIONS(7327),
    [anon_sym___restrict__] = ACTIONS(7327),
    [anon_sym__Atomic] = ACTIONS(7327),
    [anon_sym__Noreturn] = ACTIONS(7327),
    [anon_sym__Nonnull] = ACTIONS(7327),
    [anon_sym_nullable] = ACTIONS(7327),
    [anon_sym__Complex] = ACTIONS(7327),
    [anon_sym__Nullable] = ACTIONS(7327),
    [anon_sym__Nullable_result] = ACTIONS(7327),
    [anon_sym__Null_unspecified] = ACTIONS(7327),
    [anon_sym___autoreleasing] = ACTIONS(7327),
    [anon_sym___block] = ACTIONS(7327),
    [anon_sym___bridge] = ACTIONS(7327),
    [anon_sym___bridge_retained] = ACTIONS(7327),
    [anon_sym___bridge_transfer] = ACTIONS(7327),
    [anon_sym___complex] = ACTIONS(7327),
    [anon_sym___const] = ACTIONS(7327),
    [anon_sym___imag] = ACTIONS(7327),
    [anon_sym___kindof] = ACTIONS(7327),
    [anon_sym___nonnull] = ACTIONS(7327),
    [anon_sym___nullable] = ACTIONS(7327),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7327),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7327),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7327),
    [anon_sym___real] = ACTIONS(7327),
    [anon_sym___strong] = ACTIONS(7327),
    [anon_sym___unsafe_unretained] = ACTIONS(7327),
    [anon_sym___unused] = ACTIONS(7327),
    [anon_sym___weak] = ACTIONS(7327),
    [anon_sym_alignas] = ACTIONS(7327),
    [anon_sym__Alignas] = ACTIONS(7327),
    [sym_primitive_type] = ACTIONS(7327),
    [anon_sym_enum] = ACTIONS(7327),
    [anon_sym_COLON] = ACTIONS(7329),
    [anon_sym_struct] = ACTIONS(7327),
    [anon_sym_union] = ACTIONS(7327),
    [anon_sym_in] = ACTIONS(7327),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7327),
    [anon_sym___typeof] = ACTIONS(7327),
    [anon_sym_typeof] = ACTIONS(7327),
    [anon_sym_BOOL] = ACTIONS(7327),
    [anon_sym_IMP] = ACTIONS(7327),
    [anon_sym_SEL] = ACTIONS(7327),
    [anon_sym_Class] = ACTIONS(7327),
    [anon_sym_id] = ACTIONS(7327),
    [anon_sym_out] = ACTIONS(7327),
    [anon_sym_inout] = ACTIONS(7327),
    [anon_sym_bycopy] = ACTIONS(7327),
    [anon_sym_byref] = ACTIONS(7327),
    [anon_sym_oneway] = ACTIONS(7327),
  },
  [STATE(3863)] = {
    [sym_parameter_list] = STATE(3764),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6742),
    [anon_sym_RPAREN] = ACTIONS(6742),
    [anon_sym_LPAREN2] = ACTIONS(8850),
    [anon_sym_STAR] = ACTIONS(6742),
    [anon_sym_CARET] = ACTIONS(6742),
    [anon_sym_GT] = ACTIONS(6742),
    [anon_sym_LT] = ACTIONS(6742),
    [anon_sym___extension__] = ACTIONS(6740),
    [anon_sym___attribute__] = ACTIONS(6740),
    [anon_sym___attribute] = ACTIONS(6740),
    [anon_sym_noreturn] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(8852),
    [anon_sym___based] = ACTIONS(6740),
    [anon_sym_LBRACE] = ACTIONS(6742),
    [anon_sym_signed] = ACTIONS(6740),
    [anon_sym_unsigned] = ACTIONS(6740),
    [anon_sym_long] = ACTIONS(6740),
    [anon_sym_short] = ACTIONS(6740),
    [anon_sym_ATautoreleasepool] = ACTIONS(6742),
    [anon_sym_const] = ACTIONS(6740),
    [anon_sym_constexpr] = ACTIONS(6740),
    [anon_sym_volatile] = ACTIONS(6740),
    [anon_sym_restrict] = ACTIONS(6740),
    [anon_sym___restrict__] = ACTIONS(6740),
    [anon_sym__Atomic] = ACTIONS(6740),
    [anon_sym__Noreturn] = ACTIONS(6740),
    [anon_sym__Nonnull] = ACTIONS(6740),
    [anon_sym_nullable] = ACTIONS(6740),
    [anon_sym__Complex] = ACTIONS(6740),
    [anon_sym__Nullable] = ACTIONS(6740),
    [anon_sym__Nullable_result] = ACTIONS(6740),
    [anon_sym__Null_unspecified] = ACTIONS(6740),
    [anon_sym___autoreleasing] = ACTIONS(6740),
    [anon_sym___block] = ACTIONS(6740),
    [anon_sym___bridge] = ACTIONS(6740),
    [anon_sym___bridge_retained] = ACTIONS(6740),
    [anon_sym___bridge_transfer] = ACTIONS(6740),
    [anon_sym___complex] = ACTIONS(6740),
    [anon_sym___const] = ACTIONS(6740),
    [anon_sym___imag] = ACTIONS(6740),
    [anon_sym___kindof] = ACTIONS(6740),
    [anon_sym___nonnull] = ACTIONS(6740),
    [anon_sym___nullable] = ACTIONS(6740),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6740),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6740),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6740),
    [anon_sym___real] = ACTIONS(6740),
    [anon_sym___strong] = ACTIONS(6740),
    [anon_sym___unsafe_unretained] = ACTIONS(6740),
    [anon_sym___unused] = ACTIONS(6740),
    [anon_sym___weak] = ACTIONS(6740),
    [anon_sym_alignas] = ACTIONS(6740),
    [anon_sym__Alignas] = ACTIONS(6740),
    [sym_primitive_type] = ACTIONS(6740),
    [anon_sym_enum] = ACTIONS(6740),
    [anon_sym_struct] = ACTIONS(6740),
    [anon_sym_union] = ACTIONS(6740),
    [anon_sym_in] = ACTIONS(6740),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6740),
    [anon_sym___typeof] = ACTIONS(6740),
    [anon_sym_typeof] = ACTIONS(6740),
    [anon_sym_BOOL] = ACTIONS(6740),
    [anon_sym_IMP] = ACTIONS(6740),
    [anon_sym_SEL] = ACTIONS(6740),
    [anon_sym_Class] = ACTIONS(6740),
    [anon_sym_id] = ACTIONS(6740),
    [anon_sym_out] = ACTIONS(6740),
    [anon_sym_inout] = ACTIONS(6740),
    [anon_sym_bycopy] = ACTIONS(6740),
    [anon_sym_byref] = ACTIONS(6740),
    [anon_sym_oneway] = ACTIONS(6740),
  },
  [STATE(3864)] = {
    [sym_identifier] = ACTIONS(7161),
    [anon_sym_COMMA] = ACTIONS(7163),
    [anon_sym_RPAREN] = ACTIONS(7163),
    [anon_sym_LPAREN2] = ACTIONS(7163),
    [anon_sym_STAR] = ACTIONS(7163),
    [anon_sym_CARET] = ACTIONS(7163),
    [anon_sym_GT] = ACTIONS(7163),
    [anon_sym_LT] = ACTIONS(7163),
    [anon_sym___extension__] = ACTIONS(7161),
    [anon_sym___attribute__] = ACTIONS(7161),
    [anon_sym___attribute] = ACTIONS(7161),
    [anon_sym_noreturn] = ACTIONS(7161),
    [anon_sym_LBRACK] = ACTIONS(7163),
    [anon_sym___based] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7163),
    [anon_sym_signed] = ACTIONS(7161),
    [anon_sym_unsigned] = ACTIONS(7161),
    [anon_sym_long] = ACTIONS(7161),
    [anon_sym_short] = ACTIONS(7161),
    [anon_sym_ATautoreleasepool] = ACTIONS(7163),
    [anon_sym_const] = ACTIONS(7161),
    [anon_sym_constexpr] = ACTIONS(7161),
    [anon_sym_volatile] = ACTIONS(7161),
    [anon_sym_restrict] = ACTIONS(7161),
    [anon_sym___restrict__] = ACTIONS(7161),
    [anon_sym__Atomic] = ACTIONS(7161),
    [anon_sym__Noreturn] = ACTIONS(7161),
    [anon_sym__Nonnull] = ACTIONS(7161),
    [anon_sym_nullable] = ACTIONS(7161),
    [anon_sym__Complex] = ACTIONS(7161),
    [anon_sym__Nullable] = ACTIONS(7161),
    [anon_sym__Nullable_result] = ACTIONS(7161),
    [anon_sym__Null_unspecified] = ACTIONS(7161),
    [anon_sym___autoreleasing] = ACTIONS(7161),
    [anon_sym___block] = ACTIONS(7161),
    [anon_sym___bridge] = ACTIONS(7161),
    [anon_sym___bridge_retained] = ACTIONS(7161),
    [anon_sym___bridge_transfer] = ACTIONS(7161),
    [anon_sym___complex] = ACTIONS(7161),
    [anon_sym___const] = ACTIONS(7161),
    [anon_sym___imag] = ACTIONS(7161),
    [anon_sym___kindof] = ACTIONS(7161),
    [anon_sym___nonnull] = ACTIONS(7161),
    [anon_sym___nullable] = ACTIONS(7161),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7161),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7161),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7161),
    [anon_sym___real] = ACTIONS(7161),
    [anon_sym___strong] = ACTIONS(7161),
    [anon_sym___unsafe_unretained] = ACTIONS(7161),
    [anon_sym___unused] = ACTIONS(7161),
    [anon_sym___weak] = ACTIONS(7161),
    [anon_sym_alignas] = ACTIONS(7161),
    [anon_sym__Alignas] = ACTIONS(7161),
    [sym_primitive_type] = ACTIONS(7161),
    [anon_sym_enum] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7163),
    [anon_sym_struct] = ACTIONS(7161),
    [anon_sym_union] = ACTIONS(7161),
    [anon_sym_in] = ACTIONS(7161),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7161),
    [anon_sym___typeof] = ACTIONS(7161),
    [anon_sym_typeof] = ACTIONS(7161),
    [anon_sym_BOOL] = ACTIONS(7161),
    [anon_sym_IMP] = ACTIONS(7161),
    [anon_sym_SEL] = ACTIONS(7161),
    [anon_sym_Class] = ACTIONS(7161),
    [anon_sym_id] = ACTIONS(7161),
    [anon_sym_out] = ACTIONS(7161),
    [anon_sym_inout] = ACTIONS(7161),
    [anon_sym_bycopy] = ACTIONS(7161),
    [anon_sym_byref] = ACTIONS(7161),
    [anon_sym_oneway] = ACTIONS(7161),
  },
  [STATE(3865)] = {
    [sym_identifier] = ACTIONS(7175),
    [anon_sym_COMMA] = ACTIONS(7177),
    [anon_sym_RPAREN] = ACTIONS(7177),
    [anon_sym_LPAREN2] = ACTIONS(7177),
    [anon_sym_STAR] = ACTIONS(7177),
    [anon_sym_CARET] = ACTIONS(7177),
    [anon_sym_GT] = ACTIONS(7177),
    [anon_sym_LT] = ACTIONS(7177),
    [anon_sym___extension__] = ACTIONS(7175),
    [anon_sym___attribute__] = ACTIONS(7175),
    [anon_sym___attribute] = ACTIONS(7175),
    [anon_sym_noreturn] = ACTIONS(7175),
    [anon_sym_LBRACK] = ACTIONS(7177),
    [anon_sym___based] = ACTIONS(7175),
    [anon_sym_LBRACE] = ACTIONS(7177),
    [anon_sym_signed] = ACTIONS(7175),
    [anon_sym_unsigned] = ACTIONS(7175),
    [anon_sym_long] = ACTIONS(7175),
    [anon_sym_short] = ACTIONS(7175),
    [anon_sym_ATautoreleasepool] = ACTIONS(7177),
    [anon_sym_const] = ACTIONS(7175),
    [anon_sym_constexpr] = ACTIONS(7175),
    [anon_sym_volatile] = ACTIONS(7175),
    [anon_sym_restrict] = ACTIONS(7175),
    [anon_sym___restrict__] = ACTIONS(7175),
    [anon_sym__Atomic] = ACTIONS(7175),
    [anon_sym__Noreturn] = ACTIONS(7175),
    [anon_sym__Nonnull] = ACTIONS(7175),
    [anon_sym_nullable] = ACTIONS(7175),
    [anon_sym__Complex] = ACTIONS(7175),
    [anon_sym__Nullable] = ACTIONS(7175),
    [anon_sym__Nullable_result] = ACTIONS(7175),
    [anon_sym__Null_unspecified] = ACTIONS(7175),
    [anon_sym___autoreleasing] = ACTIONS(7175),
    [anon_sym___block] = ACTIONS(7175),
    [anon_sym___bridge] = ACTIONS(7175),
    [anon_sym___bridge_retained] = ACTIONS(7175),
    [anon_sym___bridge_transfer] = ACTIONS(7175),
    [anon_sym___complex] = ACTIONS(7175),
    [anon_sym___const] = ACTIONS(7175),
    [anon_sym___imag] = ACTIONS(7175),
    [anon_sym___kindof] = ACTIONS(7175),
    [anon_sym___nonnull] = ACTIONS(7175),
    [anon_sym___nullable] = ACTIONS(7175),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7175),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7175),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7175),
    [anon_sym___real] = ACTIONS(7175),
    [anon_sym___strong] = ACTIONS(7175),
    [anon_sym___unsafe_unretained] = ACTIONS(7175),
    [anon_sym___unused] = ACTIONS(7175),
    [anon_sym___weak] = ACTIONS(7175),
    [anon_sym_alignas] = ACTIONS(7175),
    [anon_sym__Alignas] = ACTIONS(7175),
    [sym_primitive_type] = ACTIONS(7175),
    [anon_sym_enum] = ACTIONS(7175),
    [anon_sym_COLON] = ACTIONS(7177),
    [anon_sym_struct] = ACTIONS(7175),
    [anon_sym_union] = ACTIONS(7175),
    [anon_sym_in] = ACTIONS(7175),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7175),
    [anon_sym___typeof] = ACTIONS(7175),
    [anon_sym_typeof] = ACTIONS(7175),
    [anon_sym_BOOL] = ACTIONS(7175),
    [anon_sym_IMP] = ACTIONS(7175),
    [anon_sym_SEL] = ACTIONS(7175),
    [anon_sym_Class] = ACTIONS(7175),
    [anon_sym_id] = ACTIONS(7175),
    [anon_sym_out] = ACTIONS(7175),
    [anon_sym_inout] = ACTIONS(7175),
    [anon_sym_bycopy] = ACTIONS(7175),
    [anon_sym_byref] = ACTIONS(7175),
    [anon_sym_oneway] = ACTIONS(7175),
  },
  [STATE(3866)] = {
    [sym_parameter_list] = STATE(3764),
    [sym_identifier] = ACTIONS(6744),
    [anon_sym_COMMA] = ACTIONS(6746),
    [anon_sym_RPAREN] = ACTIONS(6746),
    [anon_sym_LPAREN2] = ACTIONS(8850),
    [anon_sym_STAR] = ACTIONS(6746),
    [anon_sym_CARET] = ACTIONS(6746),
    [anon_sym_GT] = ACTIONS(6746),
    [anon_sym_LT] = ACTIONS(6746),
    [anon_sym___extension__] = ACTIONS(6744),
    [anon_sym___attribute__] = ACTIONS(6744),
    [anon_sym___attribute] = ACTIONS(6744),
    [anon_sym_noreturn] = ACTIONS(6744),
    [anon_sym_LBRACK] = ACTIONS(8852),
    [anon_sym___based] = ACTIONS(6744),
    [anon_sym_LBRACE] = ACTIONS(6746),
    [anon_sym_signed] = ACTIONS(6744),
    [anon_sym_unsigned] = ACTIONS(6744),
    [anon_sym_long] = ACTIONS(6744),
    [anon_sym_short] = ACTIONS(6744),
    [anon_sym_ATautoreleasepool] = ACTIONS(6746),
    [anon_sym_const] = ACTIONS(6744),
    [anon_sym_constexpr] = ACTIONS(6744),
    [anon_sym_volatile] = ACTIONS(6744),
    [anon_sym_restrict] = ACTIONS(6744),
    [anon_sym___restrict__] = ACTIONS(6744),
    [anon_sym__Atomic] = ACTIONS(6744),
    [anon_sym__Noreturn] = ACTIONS(6744),
    [anon_sym__Nonnull] = ACTIONS(6744),
    [anon_sym_nullable] = ACTIONS(6744),
    [anon_sym__Complex] = ACTIONS(6744),
    [anon_sym__Nullable] = ACTIONS(6744),
    [anon_sym__Nullable_result] = ACTIONS(6744),
    [anon_sym__Null_unspecified] = ACTIONS(6744),
    [anon_sym___autoreleasing] = ACTIONS(6744),
    [anon_sym___block] = ACTIONS(6744),
    [anon_sym___bridge] = ACTIONS(6744),
    [anon_sym___bridge_retained] = ACTIONS(6744),
    [anon_sym___bridge_transfer] = ACTIONS(6744),
    [anon_sym___complex] = ACTIONS(6744),
    [anon_sym___const] = ACTIONS(6744),
    [anon_sym___imag] = ACTIONS(6744),
    [anon_sym___kindof] = ACTIONS(6744),
    [anon_sym___nonnull] = ACTIONS(6744),
    [anon_sym___nullable] = ACTIONS(6744),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6744),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6744),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6744),
    [anon_sym___real] = ACTIONS(6744),
    [anon_sym___strong] = ACTIONS(6744),
    [anon_sym___unsafe_unretained] = ACTIONS(6744),
    [anon_sym___unused] = ACTIONS(6744),
    [anon_sym___weak] = ACTIONS(6744),
    [anon_sym_alignas] = ACTIONS(6744),
    [anon_sym__Alignas] = ACTIONS(6744),
    [sym_primitive_type] = ACTIONS(6744),
    [anon_sym_enum] = ACTIONS(6744),
    [anon_sym_struct] = ACTIONS(6744),
    [anon_sym_union] = ACTIONS(6744),
    [anon_sym_in] = ACTIONS(6744),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6744),
    [anon_sym___typeof] = ACTIONS(6744),
    [anon_sym_typeof] = ACTIONS(6744),
    [anon_sym_BOOL] = ACTIONS(6744),
    [anon_sym_IMP] = ACTIONS(6744),
    [anon_sym_SEL] = ACTIONS(6744),
    [anon_sym_Class] = ACTIONS(6744),
    [anon_sym_id] = ACTIONS(6744),
    [anon_sym_out] = ACTIONS(6744),
    [anon_sym_inout] = ACTIONS(6744),
    [anon_sym_bycopy] = ACTIONS(6744),
    [anon_sym_byref] = ACTIONS(6744),
    [anon_sym_oneway] = ACTIONS(6744),
  },
  [STATE(3867)] = {
    [sym_identifier] = ACTIONS(7167),
    [anon_sym_COMMA] = ACTIONS(7169),
    [anon_sym_RPAREN] = ACTIONS(7169),
    [anon_sym_LPAREN2] = ACTIONS(7169),
    [anon_sym_STAR] = ACTIONS(7169),
    [anon_sym_CARET] = ACTIONS(7169),
    [anon_sym_GT] = ACTIONS(7169),
    [anon_sym_LT] = ACTIONS(7169),
    [anon_sym___extension__] = ACTIONS(7167),
    [anon_sym___attribute__] = ACTIONS(7167),
    [anon_sym___attribute] = ACTIONS(7167),
    [anon_sym_noreturn] = ACTIONS(7167),
    [anon_sym_LBRACK] = ACTIONS(7169),
    [anon_sym___based] = ACTIONS(7167),
    [anon_sym_LBRACE] = ACTIONS(7169),
    [anon_sym_signed] = ACTIONS(7167),
    [anon_sym_unsigned] = ACTIONS(7167),
    [anon_sym_long] = ACTIONS(7167),
    [anon_sym_short] = ACTIONS(7167),
    [anon_sym_ATautoreleasepool] = ACTIONS(7169),
    [anon_sym_const] = ACTIONS(7167),
    [anon_sym_constexpr] = ACTIONS(7167),
    [anon_sym_volatile] = ACTIONS(7167),
    [anon_sym_restrict] = ACTIONS(7167),
    [anon_sym___restrict__] = ACTIONS(7167),
    [anon_sym__Atomic] = ACTIONS(7167),
    [anon_sym__Noreturn] = ACTIONS(7167),
    [anon_sym__Nonnull] = ACTIONS(7167),
    [anon_sym_nullable] = ACTIONS(7167),
    [anon_sym__Complex] = ACTIONS(7167),
    [anon_sym__Nullable] = ACTIONS(7167),
    [anon_sym__Nullable_result] = ACTIONS(7167),
    [anon_sym__Null_unspecified] = ACTIONS(7167),
    [anon_sym___autoreleasing] = ACTIONS(7167),
    [anon_sym___block] = ACTIONS(7167),
    [anon_sym___bridge] = ACTIONS(7167),
    [anon_sym___bridge_retained] = ACTIONS(7167),
    [anon_sym___bridge_transfer] = ACTIONS(7167),
    [anon_sym___complex] = ACTIONS(7167),
    [anon_sym___const] = ACTIONS(7167),
    [anon_sym___imag] = ACTIONS(7167),
    [anon_sym___kindof] = ACTIONS(7167),
    [anon_sym___nonnull] = ACTIONS(7167),
    [anon_sym___nullable] = ACTIONS(7167),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7167),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7167),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7167),
    [anon_sym___real] = ACTIONS(7167),
    [anon_sym___strong] = ACTIONS(7167),
    [anon_sym___unsafe_unretained] = ACTIONS(7167),
    [anon_sym___unused] = ACTIONS(7167),
    [anon_sym___weak] = ACTIONS(7167),
    [anon_sym_alignas] = ACTIONS(7167),
    [anon_sym__Alignas] = ACTIONS(7167),
    [sym_primitive_type] = ACTIONS(7167),
    [anon_sym_enum] = ACTIONS(7167),
    [anon_sym_COLON] = ACTIONS(7169),
    [anon_sym_struct] = ACTIONS(7167),
    [anon_sym_union] = ACTIONS(7167),
    [anon_sym_in] = ACTIONS(7167),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7167),
    [anon_sym___typeof] = ACTIONS(7167),
    [anon_sym_typeof] = ACTIONS(7167),
    [anon_sym_BOOL] = ACTIONS(7167),
    [anon_sym_IMP] = ACTIONS(7167),
    [anon_sym_SEL] = ACTIONS(7167),
    [anon_sym_Class] = ACTIONS(7167),
    [anon_sym_id] = ACTIONS(7167),
    [anon_sym_out] = ACTIONS(7167),
    [anon_sym_inout] = ACTIONS(7167),
    [anon_sym_bycopy] = ACTIONS(7167),
    [anon_sym_byref] = ACTIONS(7167),
    [anon_sym_oneway] = ACTIONS(7167),
  },
  [STATE(3868)] = {
    [sym_attribute_specifier] = STATE(3861),
    [sym_identifier] = ACTIONS(7898),
    [anon_sym_COMMA] = ACTIONS(7900),
    [anon_sym_RPAREN] = ACTIONS(7900),
    [anon_sym_LPAREN2] = ACTIONS(7900),
    [anon_sym_STAR] = ACTIONS(7900),
    [anon_sym_CARET] = ACTIONS(7900),
    [anon_sym_GT] = ACTIONS(7900),
    [anon_sym_LT] = ACTIONS(7900),
    [anon_sym___extension__] = ACTIONS(7898),
    [anon_sym___attribute__] = ACTIONS(8869),
    [anon_sym___attribute] = ACTIONS(8869),
    [anon_sym_noreturn] = ACTIONS(7898),
    [anon_sym_LBRACK] = ACTIONS(7900),
    [anon_sym___based] = ACTIONS(7898),
    [anon_sym_LBRACE] = ACTIONS(7900),
    [anon_sym_signed] = ACTIONS(7898),
    [anon_sym_unsigned] = ACTIONS(7898),
    [anon_sym_long] = ACTIONS(7898),
    [anon_sym_short] = ACTIONS(7898),
    [anon_sym_ATautoreleasepool] = ACTIONS(7900),
    [anon_sym_const] = ACTIONS(7898),
    [anon_sym_constexpr] = ACTIONS(7898),
    [anon_sym_volatile] = ACTIONS(7898),
    [anon_sym_restrict] = ACTIONS(7898),
    [anon_sym___restrict__] = ACTIONS(7898),
    [anon_sym__Atomic] = ACTIONS(7898),
    [anon_sym__Noreturn] = ACTIONS(7898),
    [anon_sym__Nonnull] = ACTIONS(7898),
    [anon_sym_nullable] = ACTIONS(7898),
    [anon_sym__Complex] = ACTIONS(7898),
    [anon_sym__Nullable] = ACTIONS(7898),
    [anon_sym__Nullable_result] = ACTIONS(7898),
    [anon_sym__Null_unspecified] = ACTIONS(7898),
    [anon_sym___autoreleasing] = ACTIONS(7898),
    [anon_sym___block] = ACTIONS(7898),
    [anon_sym___bridge] = ACTIONS(7898),
    [anon_sym___bridge_retained] = ACTIONS(7898),
    [anon_sym___bridge_transfer] = ACTIONS(7898),
    [anon_sym___complex] = ACTIONS(7898),
    [anon_sym___const] = ACTIONS(7898),
    [anon_sym___imag] = ACTIONS(7898),
    [anon_sym___kindof] = ACTIONS(7898),
    [anon_sym___nonnull] = ACTIONS(7898),
    [anon_sym___nullable] = ACTIONS(7898),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7898),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7898),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7898),
    [anon_sym___real] = ACTIONS(7898),
    [anon_sym___strong] = ACTIONS(7898),
    [anon_sym___unsafe_unretained] = ACTIONS(7898),
    [anon_sym___unused] = ACTIONS(7898),
    [anon_sym___weak] = ACTIONS(7898),
    [anon_sym_alignas] = ACTIONS(7898),
    [anon_sym__Alignas] = ACTIONS(7898),
    [sym_primitive_type] = ACTIONS(7898),
    [anon_sym_enum] = ACTIONS(7898),
    [anon_sym_struct] = ACTIONS(7898),
    [anon_sym_union] = ACTIONS(7898),
    [anon_sym_in] = ACTIONS(7898),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7898),
    [anon_sym___typeof] = ACTIONS(7898),
    [anon_sym_typeof] = ACTIONS(7898),
    [anon_sym_BOOL] = ACTIONS(7898),
    [anon_sym_IMP] = ACTIONS(7898),
    [anon_sym_SEL] = ACTIONS(7898),
    [anon_sym_Class] = ACTIONS(7898),
    [anon_sym_id] = ACTIONS(7898),
    [anon_sym_out] = ACTIONS(7898),
    [anon_sym_inout] = ACTIONS(7898),
    [anon_sym_bycopy] = ACTIONS(7898),
    [anon_sym_byref] = ACTIONS(7898),
    [anon_sym_oneway] = ACTIONS(7898),
  },
  [STATE(3869)] = {
    [sym_identifier] = ACTIONS(7189),
    [anon_sym_COMMA] = ACTIONS(7191),
    [anon_sym_RPAREN] = ACTIONS(7191),
    [anon_sym_LPAREN2] = ACTIONS(7191),
    [anon_sym_STAR] = ACTIONS(7191),
    [anon_sym_CARET] = ACTIONS(7191),
    [anon_sym_GT] = ACTIONS(7191),
    [anon_sym_LT] = ACTIONS(7191),
    [anon_sym___extension__] = ACTIONS(7189),
    [anon_sym___attribute__] = ACTIONS(7189),
    [anon_sym___attribute] = ACTIONS(7189),
    [anon_sym_noreturn] = ACTIONS(7189),
    [anon_sym_LBRACK] = ACTIONS(7191),
    [anon_sym___based] = ACTIONS(7189),
    [anon_sym_LBRACE] = ACTIONS(7191),
    [anon_sym_signed] = ACTIONS(7189),
    [anon_sym_unsigned] = ACTIONS(7189),
    [anon_sym_long] = ACTIONS(7189),
    [anon_sym_short] = ACTIONS(7189),
    [anon_sym_ATautoreleasepool] = ACTIONS(7191),
    [anon_sym_const] = ACTIONS(7189),
    [anon_sym_constexpr] = ACTIONS(7189),
    [anon_sym_volatile] = ACTIONS(7189),
    [anon_sym_restrict] = ACTIONS(7189),
    [anon_sym___restrict__] = ACTIONS(7189),
    [anon_sym__Atomic] = ACTIONS(7189),
    [anon_sym__Noreturn] = ACTIONS(7189),
    [anon_sym__Nonnull] = ACTIONS(7189),
    [anon_sym_nullable] = ACTIONS(7189),
    [anon_sym__Complex] = ACTIONS(7189),
    [anon_sym__Nullable] = ACTIONS(7189),
    [anon_sym__Nullable_result] = ACTIONS(7189),
    [anon_sym__Null_unspecified] = ACTIONS(7189),
    [anon_sym___autoreleasing] = ACTIONS(7189),
    [anon_sym___block] = ACTIONS(7189),
    [anon_sym___bridge] = ACTIONS(7189),
    [anon_sym___bridge_retained] = ACTIONS(7189),
    [anon_sym___bridge_transfer] = ACTIONS(7189),
    [anon_sym___complex] = ACTIONS(7189),
    [anon_sym___const] = ACTIONS(7189),
    [anon_sym___imag] = ACTIONS(7189),
    [anon_sym___kindof] = ACTIONS(7189),
    [anon_sym___nonnull] = ACTIONS(7189),
    [anon_sym___nullable] = ACTIONS(7189),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7189),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7189),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7189),
    [anon_sym___real] = ACTIONS(7189),
    [anon_sym___strong] = ACTIONS(7189),
    [anon_sym___unsafe_unretained] = ACTIONS(7189),
    [anon_sym___unused] = ACTIONS(7189),
    [anon_sym___weak] = ACTIONS(7189),
    [anon_sym_alignas] = ACTIONS(7189),
    [anon_sym__Alignas] = ACTIONS(7189),
    [sym_primitive_type] = ACTIONS(7189),
    [anon_sym_enum] = ACTIONS(7189),
    [anon_sym_COLON] = ACTIONS(7191),
    [anon_sym_struct] = ACTIONS(7189),
    [anon_sym_union] = ACTIONS(7189),
    [anon_sym_in] = ACTIONS(7189),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7189),
    [anon_sym___typeof] = ACTIONS(7189),
    [anon_sym_typeof] = ACTIONS(7189),
    [anon_sym_BOOL] = ACTIONS(7189),
    [anon_sym_IMP] = ACTIONS(7189),
    [anon_sym_SEL] = ACTIONS(7189),
    [anon_sym_Class] = ACTIONS(7189),
    [anon_sym_id] = ACTIONS(7189),
    [anon_sym_out] = ACTIONS(7189),
    [anon_sym_inout] = ACTIONS(7189),
    [anon_sym_bycopy] = ACTIONS(7189),
    [anon_sym_byref] = ACTIONS(7189),
    [anon_sym_oneway] = ACTIONS(7189),
  },
  [STATE(3870)] = {
    [sym_identifier] = ACTIONS(5074),
    [anon_sym_COMMA] = ACTIONS(5070),
    [anon_sym_RPAREN] = ACTIONS(5070),
    [anon_sym_LPAREN2] = ACTIONS(5070),
    [anon_sym_STAR] = ACTIONS(5070),
    [anon_sym_CARET] = ACTIONS(5070),
    [anon_sym_GT] = ACTIONS(5070),
    [anon_sym_LT] = ACTIONS(5070),
    [anon_sym___extension__] = ACTIONS(5074),
    [anon_sym___attribute__] = ACTIONS(5074),
    [anon_sym___attribute] = ACTIONS(5074),
    [anon_sym_noreturn] = ACTIONS(5074),
    [anon_sym_LBRACK] = ACTIONS(5070),
    [anon_sym___based] = ACTIONS(5074),
    [anon_sym_LBRACE] = ACTIONS(5070),
    [anon_sym_signed] = ACTIONS(5074),
    [anon_sym_unsigned] = ACTIONS(5074),
    [anon_sym_long] = ACTIONS(5074),
    [anon_sym_short] = ACTIONS(5074),
    [anon_sym_ATautoreleasepool] = ACTIONS(5070),
    [anon_sym_const] = ACTIONS(5074),
    [anon_sym_constexpr] = ACTIONS(5074),
    [anon_sym_volatile] = ACTIONS(5074),
    [anon_sym_restrict] = ACTIONS(5074),
    [anon_sym___restrict__] = ACTIONS(5074),
    [anon_sym__Atomic] = ACTIONS(5074),
    [anon_sym__Noreturn] = ACTIONS(5074),
    [anon_sym__Nonnull] = ACTIONS(5074),
    [anon_sym_nullable] = ACTIONS(5074),
    [anon_sym__Complex] = ACTIONS(5074),
    [anon_sym__Nullable] = ACTIONS(5074),
    [anon_sym__Nullable_result] = ACTIONS(5074),
    [anon_sym__Null_unspecified] = ACTIONS(5074),
    [anon_sym___autoreleasing] = ACTIONS(5074),
    [anon_sym___block] = ACTIONS(5074),
    [anon_sym___bridge] = ACTIONS(5074),
    [anon_sym___bridge_retained] = ACTIONS(5074),
    [anon_sym___bridge_transfer] = ACTIONS(5074),
    [anon_sym___complex] = ACTIONS(5074),
    [anon_sym___const] = ACTIONS(5074),
    [anon_sym___imag] = ACTIONS(5074),
    [anon_sym___kindof] = ACTIONS(5074),
    [anon_sym___nonnull] = ACTIONS(5074),
    [anon_sym___nullable] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5074),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5074),
    [anon_sym___real] = ACTIONS(5074),
    [anon_sym___strong] = ACTIONS(5074),
    [anon_sym___unsafe_unretained] = ACTIONS(5074),
    [anon_sym___unused] = ACTIONS(5074),
    [anon_sym___weak] = ACTIONS(5074),
    [anon_sym_alignas] = ACTIONS(5074),
    [anon_sym__Alignas] = ACTIONS(5074),
    [sym_primitive_type] = ACTIONS(5074),
    [anon_sym_enum] = ACTIONS(5074),
    [anon_sym_COLON] = ACTIONS(5070),
    [anon_sym_struct] = ACTIONS(5074),
    [anon_sym_union] = ACTIONS(5074),
    [anon_sym_in] = ACTIONS(5074),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5074),
    [anon_sym___typeof] = ACTIONS(5074),
    [anon_sym_typeof] = ACTIONS(5074),
    [anon_sym_BOOL] = ACTIONS(5074),
    [anon_sym_IMP] = ACTIONS(5074),
    [anon_sym_SEL] = ACTIONS(5074),
    [anon_sym_Class] = ACTIONS(5074),
    [anon_sym_id] = ACTIONS(5074),
    [anon_sym_out] = ACTIONS(5074),
    [anon_sym_inout] = ACTIONS(5074),
    [anon_sym_bycopy] = ACTIONS(5074),
    [anon_sym_byref] = ACTIONS(5074),
    [anon_sym_oneway] = ACTIONS(5074),
  },
  [STATE(3871)] = {
    [sym_attribute_specifier] = STATE(3892),
    [sym_enumerator_list] = STATE(3850),
    [sym_identifier] = ACTIONS(7512),
    [anon_sym_COMMA] = ACTIONS(7514),
    [anon_sym_RPAREN] = ACTIONS(7514),
    [anon_sym_LPAREN2] = ACTIONS(7514),
    [anon_sym_STAR] = ACTIONS(7514),
    [anon_sym_CARET] = ACTIONS(7514),
    [anon_sym_GT] = ACTIONS(7514),
    [anon_sym_LT] = ACTIONS(7514),
    [anon_sym___extension__] = ACTIONS(7512),
    [anon_sym___attribute__] = ACTIONS(8872),
    [anon_sym___attribute] = ACTIONS(8872),
    [anon_sym_noreturn] = ACTIONS(7512),
    [anon_sym_LBRACK] = ACTIONS(7514),
    [anon_sym___based] = ACTIONS(7512),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7512),
    [anon_sym_unsigned] = ACTIONS(7512),
    [anon_sym_long] = ACTIONS(7512),
    [anon_sym_short] = ACTIONS(7512),
    [anon_sym_const] = ACTIONS(7512),
    [anon_sym_constexpr] = ACTIONS(7512),
    [anon_sym_volatile] = ACTIONS(7512),
    [anon_sym_restrict] = ACTIONS(7512),
    [anon_sym___restrict__] = ACTIONS(7512),
    [anon_sym__Atomic] = ACTIONS(7512),
    [anon_sym__Noreturn] = ACTIONS(7512),
    [anon_sym__Nonnull] = ACTIONS(7512),
    [anon_sym_nullable] = ACTIONS(7512),
    [anon_sym__Complex] = ACTIONS(7512),
    [anon_sym__Nullable] = ACTIONS(7512),
    [anon_sym__Nullable_result] = ACTIONS(7512),
    [anon_sym__Null_unspecified] = ACTIONS(7512),
    [anon_sym___autoreleasing] = ACTIONS(7512),
    [anon_sym___block] = ACTIONS(7512),
    [anon_sym___bridge] = ACTIONS(7512),
    [anon_sym___bridge_retained] = ACTIONS(7512),
    [anon_sym___bridge_transfer] = ACTIONS(7512),
    [anon_sym___complex] = ACTIONS(7512),
    [anon_sym___const] = ACTIONS(7512),
    [anon_sym___imag] = ACTIONS(7512),
    [anon_sym___kindof] = ACTIONS(7512),
    [anon_sym___nonnull] = ACTIONS(7512),
    [anon_sym___nullable] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7512),
    [anon_sym___real] = ACTIONS(7512),
    [anon_sym___strong] = ACTIONS(7512),
    [anon_sym___unsafe_unretained] = ACTIONS(7512),
    [anon_sym___unused] = ACTIONS(7512),
    [anon_sym___weak] = ACTIONS(7512),
    [anon_sym_alignas] = ACTIONS(7512),
    [anon_sym__Alignas] = ACTIONS(7512),
    [sym_primitive_type] = ACTIONS(7512),
    [anon_sym_enum] = ACTIONS(7512),
    [anon_sym_struct] = ACTIONS(7512),
    [anon_sym_union] = ACTIONS(7512),
    [anon_sym_in] = ACTIONS(7512),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7512),
    [anon_sym___typeof] = ACTIONS(7512),
    [anon_sym_typeof] = ACTIONS(7512),
    [anon_sym_BOOL] = ACTIONS(7512),
    [anon_sym_IMP] = ACTIONS(7512),
    [anon_sym_SEL] = ACTIONS(7512),
    [anon_sym_Class] = ACTIONS(7512),
    [anon_sym_id] = ACTIONS(7512),
    [anon_sym_out] = ACTIONS(7512),
    [anon_sym_inout] = ACTIONS(7512),
    [anon_sym_bycopy] = ACTIONS(7512),
    [anon_sym_byref] = ACTIONS(7512),
    [anon_sym_oneway] = ACTIONS(7512),
  },
  [STATE(3872)] = {
    [sym_attribute_specifier] = STATE(3894),
    [sym_enumerator_list] = STATE(3851),
    [sym_identifier] = ACTIONS(7521),
    [anon_sym_COMMA] = ACTIONS(7523),
    [anon_sym_RPAREN] = ACTIONS(7523),
    [anon_sym_LPAREN2] = ACTIONS(7523),
    [anon_sym_STAR] = ACTIONS(7523),
    [anon_sym_CARET] = ACTIONS(7523),
    [anon_sym_GT] = ACTIONS(7523),
    [anon_sym_LT] = ACTIONS(7523),
    [anon_sym___extension__] = ACTIONS(7521),
    [anon_sym___attribute__] = ACTIONS(8875),
    [anon_sym___attribute] = ACTIONS(8875),
    [anon_sym_noreturn] = ACTIONS(7521),
    [anon_sym_LBRACK] = ACTIONS(7523),
    [anon_sym___based] = ACTIONS(7521),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7521),
    [anon_sym_unsigned] = ACTIONS(7521),
    [anon_sym_long] = ACTIONS(7521),
    [anon_sym_short] = ACTIONS(7521),
    [anon_sym_const] = ACTIONS(7521),
    [anon_sym_constexpr] = ACTIONS(7521),
    [anon_sym_volatile] = ACTIONS(7521),
    [anon_sym_restrict] = ACTIONS(7521),
    [anon_sym___restrict__] = ACTIONS(7521),
    [anon_sym__Atomic] = ACTIONS(7521),
    [anon_sym__Noreturn] = ACTIONS(7521),
    [anon_sym__Nonnull] = ACTIONS(7521),
    [anon_sym_nullable] = ACTIONS(7521),
    [anon_sym__Complex] = ACTIONS(7521),
    [anon_sym__Nullable] = ACTIONS(7521),
    [anon_sym__Nullable_result] = ACTIONS(7521),
    [anon_sym__Null_unspecified] = ACTIONS(7521),
    [anon_sym___autoreleasing] = ACTIONS(7521),
    [anon_sym___block] = ACTIONS(7521),
    [anon_sym___bridge] = ACTIONS(7521),
    [anon_sym___bridge_retained] = ACTIONS(7521),
    [anon_sym___bridge_transfer] = ACTIONS(7521),
    [anon_sym___complex] = ACTIONS(7521),
    [anon_sym___const] = ACTIONS(7521),
    [anon_sym___imag] = ACTIONS(7521),
    [anon_sym___kindof] = ACTIONS(7521),
    [anon_sym___nonnull] = ACTIONS(7521),
    [anon_sym___nullable] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7521),
    [anon_sym___real] = ACTIONS(7521),
    [anon_sym___strong] = ACTIONS(7521),
    [anon_sym___unsafe_unretained] = ACTIONS(7521),
    [anon_sym___unused] = ACTIONS(7521),
    [anon_sym___weak] = ACTIONS(7521),
    [anon_sym_alignas] = ACTIONS(7521),
    [anon_sym__Alignas] = ACTIONS(7521),
    [sym_primitive_type] = ACTIONS(7521),
    [anon_sym_enum] = ACTIONS(7521),
    [anon_sym_struct] = ACTIONS(7521),
    [anon_sym_union] = ACTIONS(7521),
    [anon_sym_in] = ACTIONS(7521),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7521),
    [anon_sym___typeof] = ACTIONS(7521),
    [anon_sym_typeof] = ACTIONS(7521),
    [anon_sym_BOOL] = ACTIONS(7521),
    [anon_sym_IMP] = ACTIONS(7521),
    [anon_sym_SEL] = ACTIONS(7521),
    [anon_sym_Class] = ACTIONS(7521),
    [anon_sym_id] = ACTIONS(7521),
    [anon_sym_out] = ACTIONS(7521),
    [anon_sym_inout] = ACTIONS(7521),
    [anon_sym_bycopy] = ACTIONS(7521),
    [anon_sym_byref] = ACTIONS(7521),
    [anon_sym_oneway] = ACTIONS(7521),
  },
  [STATE(3873)] = {
    [sym_identifier] = ACTIONS(7245),
    [anon_sym_COMMA] = ACTIONS(7247),
    [anon_sym_RPAREN] = ACTIONS(7247),
    [anon_sym_LPAREN2] = ACTIONS(7247),
    [anon_sym_STAR] = ACTIONS(7247),
    [anon_sym_CARET] = ACTIONS(7247),
    [anon_sym_GT] = ACTIONS(7247),
    [anon_sym_LT] = ACTIONS(7247),
    [anon_sym___extension__] = ACTIONS(7245),
    [anon_sym___attribute__] = ACTIONS(7245),
    [anon_sym___attribute] = ACTIONS(7245),
    [anon_sym_noreturn] = ACTIONS(7245),
    [anon_sym_LBRACK] = ACTIONS(7247),
    [anon_sym___based] = ACTIONS(7245),
    [anon_sym_LBRACE] = ACTIONS(7247),
    [anon_sym_signed] = ACTIONS(7245),
    [anon_sym_unsigned] = ACTIONS(7245),
    [anon_sym_long] = ACTIONS(7245),
    [anon_sym_short] = ACTIONS(7245),
    [anon_sym_ATautoreleasepool] = ACTIONS(7247),
    [anon_sym_const] = ACTIONS(7245),
    [anon_sym_constexpr] = ACTIONS(7245),
    [anon_sym_volatile] = ACTIONS(7245),
    [anon_sym_restrict] = ACTIONS(7245),
    [anon_sym___restrict__] = ACTIONS(7245),
    [anon_sym__Atomic] = ACTIONS(7245),
    [anon_sym__Noreturn] = ACTIONS(7245),
    [anon_sym__Nonnull] = ACTIONS(7245),
    [anon_sym_nullable] = ACTIONS(7245),
    [anon_sym__Complex] = ACTIONS(7245),
    [anon_sym__Nullable] = ACTIONS(7245),
    [anon_sym__Nullable_result] = ACTIONS(7245),
    [anon_sym__Null_unspecified] = ACTIONS(7245),
    [anon_sym___autoreleasing] = ACTIONS(7245),
    [anon_sym___block] = ACTIONS(7245),
    [anon_sym___bridge] = ACTIONS(7245),
    [anon_sym___bridge_retained] = ACTIONS(7245),
    [anon_sym___bridge_transfer] = ACTIONS(7245),
    [anon_sym___complex] = ACTIONS(7245),
    [anon_sym___const] = ACTIONS(7245),
    [anon_sym___imag] = ACTIONS(7245),
    [anon_sym___kindof] = ACTIONS(7245),
    [anon_sym___nonnull] = ACTIONS(7245),
    [anon_sym___nullable] = ACTIONS(7245),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7245),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7245),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7245),
    [anon_sym___real] = ACTIONS(7245),
    [anon_sym___strong] = ACTIONS(7245),
    [anon_sym___unsafe_unretained] = ACTIONS(7245),
    [anon_sym___unused] = ACTIONS(7245),
    [anon_sym___weak] = ACTIONS(7245),
    [anon_sym_alignas] = ACTIONS(7245),
    [anon_sym__Alignas] = ACTIONS(7245),
    [sym_primitive_type] = ACTIONS(7245),
    [anon_sym_enum] = ACTIONS(7245),
    [anon_sym_COLON] = ACTIONS(7247),
    [anon_sym_struct] = ACTIONS(7245),
    [anon_sym_union] = ACTIONS(7245),
    [anon_sym_in] = ACTIONS(7245),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7245),
    [anon_sym___typeof] = ACTIONS(7245),
    [anon_sym_typeof] = ACTIONS(7245),
    [anon_sym_BOOL] = ACTIONS(7245),
    [anon_sym_IMP] = ACTIONS(7245),
    [anon_sym_SEL] = ACTIONS(7245),
    [anon_sym_Class] = ACTIONS(7245),
    [anon_sym_id] = ACTIONS(7245),
    [anon_sym_out] = ACTIONS(7245),
    [anon_sym_inout] = ACTIONS(7245),
    [anon_sym_bycopy] = ACTIONS(7245),
    [anon_sym_byref] = ACTIONS(7245),
    [anon_sym_oneway] = ACTIONS(7245),
  },
  [STATE(3874)] = {
    [sym_identifier] = ACTIONS(7261),
    [anon_sym_COMMA] = ACTIONS(7263),
    [anon_sym_RPAREN] = ACTIONS(7263),
    [anon_sym_LPAREN2] = ACTIONS(7263),
    [anon_sym_STAR] = ACTIONS(7263),
    [anon_sym_CARET] = ACTIONS(7263),
    [anon_sym_GT] = ACTIONS(7263),
    [anon_sym_LT] = ACTIONS(7263),
    [anon_sym___extension__] = ACTIONS(7261),
    [anon_sym___attribute__] = ACTIONS(7261),
    [anon_sym___attribute] = ACTIONS(7261),
    [anon_sym_noreturn] = ACTIONS(7261),
    [anon_sym_LBRACK] = ACTIONS(7263),
    [anon_sym___based] = ACTIONS(7261),
    [anon_sym_LBRACE] = ACTIONS(7263),
    [anon_sym_signed] = ACTIONS(7261),
    [anon_sym_unsigned] = ACTIONS(7261),
    [anon_sym_long] = ACTIONS(7261),
    [anon_sym_short] = ACTIONS(7261),
    [anon_sym_ATautoreleasepool] = ACTIONS(7263),
    [anon_sym_const] = ACTIONS(7261),
    [anon_sym_constexpr] = ACTIONS(7261),
    [anon_sym_volatile] = ACTIONS(7261),
    [anon_sym_restrict] = ACTIONS(7261),
    [anon_sym___restrict__] = ACTIONS(7261),
    [anon_sym__Atomic] = ACTIONS(7261),
    [anon_sym__Noreturn] = ACTIONS(7261),
    [anon_sym__Nonnull] = ACTIONS(7261),
    [anon_sym_nullable] = ACTIONS(7261),
    [anon_sym__Complex] = ACTIONS(7261),
    [anon_sym__Nullable] = ACTIONS(7261),
    [anon_sym__Nullable_result] = ACTIONS(7261),
    [anon_sym__Null_unspecified] = ACTIONS(7261),
    [anon_sym___autoreleasing] = ACTIONS(7261),
    [anon_sym___block] = ACTIONS(7261),
    [anon_sym___bridge] = ACTIONS(7261),
    [anon_sym___bridge_retained] = ACTIONS(7261),
    [anon_sym___bridge_transfer] = ACTIONS(7261),
    [anon_sym___complex] = ACTIONS(7261),
    [anon_sym___const] = ACTIONS(7261),
    [anon_sym___imag] = ACTIONS(7261),
    [anon_sym___kindof] = ACTIONS(7261),
    [anon_sym___nonnull] = ACTIONS(7261),
    [anon_sym___nullable] = ACTIONS(7261),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7261),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7261),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7261),
    [anon_sym___real] = ACTIONS(7261),
    [anon_sym___strong] = ACTIONS(7261),
    [anon_sym___unsafe_unretained] = ACTIONS(7261),
    [anon_sym___unused] = ACTIONS(7261),
    [anon_sym___weak] = ACTIONS(7261),
    [anon_sym_alignas] = ACTIONS(7261),
    [anon_sym__Alignas] = ACTIONS(7261),
    [sym_primitive_type] = ACTIONS(7261),
    [anon_sym_enum] = ACTIONS(7261),
    [anon_sym_COLON] = ACTIONS(7263),
    [anon_sym_struct] = ACTIONS(7261),
    [anon_sym_union] = ACTIONS(7261),
    [anon_sym_in] = ACTIONS(7261),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7261),
    [anon_sym___typeof] = ACTIONS(7261),
    [anon_sym_typeof] = ACTIONS(7261),
    [anon_sym_BOOL] = ACTIONS(7261),
    [anon_sym_IMP] = ACTIONS(7261),
    [anon_sym_SEL] = ACTIONS(7261),
    [anon_sym_Class] = ACTIONS(7261),
    [anon_sym_id] = ACTIONS(7261),
    [anon_sym_out] = ACTIONS(7261),
    [anon_sym_inout] = ACTIONS(7261),
    [anon_sym_bycopy] = ACTIONS(7261),
    [anon_sym_byref] = ACTIONS(7261),
    [anon_sym_oneway] = ACTIONS(7261),
  },
  [STATE(3875)] = {
    [sym__type_definition_type] = STATE(4766),
    [sym_attribute_specifier] = STATE(3881),
    [sym_ms_declspec_modifier] = STATE(4024),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3876)] = {
    [sym_identifier] = ACTIONS(7197),
    [anon_sym_COMMA] = ACTIONS(7199),
    [anon_sym_RPAREN] = ACTIONS(7199),
    [anon_sym_LPAREN2] = ACTIONS(7199),
    [anon_sym_STAR] = ACTIONS(7199),
    [anon_sym_CARET] = ACTIONS(7199),
    [anon_sym_GT] = ACTIONS(7199),
    [anon_sym_LT] = ACTIONS(7199),
    [anon_sym___extension__] = ACTIONS(7197),
    [anon_sym___attribute__] = ACTIONS(7197),
    [anon_sym___attribute] = ACTIONS(7197),
    [anon_sym_noreturn] = ACTIONS(7197),
    [anon_sym_LBRACK] = ACTIONS(7199),
    [anon_sym___based] = ACTIONS(7197),
    [anon_sym_LBRACE] = ACTIONS(7199),
    [anon_sym_signed] = ACTIONS(7197),
    [anon_sym_unsigned] = ACTIONS(7197),
    [anon_sym_long] = ACTIONS(7197),
    [anon_sym_short] = ACTIONS(7197),
    [anon_sym_ATautoreleasepool] = ACTIONS(7199),
    [anon_sym_const] = ACTIONS(7197),
    [anon_sym_constexpr] = ACTIONS(7197),
    [anon_sym_volatile] = ACTIONS(7197),
    [anon_sym_restrict] = ACTIONS(7197),
    [anon_sym___restrict__] = ACTIONS(7197),
    [anon_sym__Atomic] = ACTIONS(7197),
    [anon_sym__Noreturn] = ACTIONS(7197),
    [anon_sym__Nonnull] = ACTIONS(7197),
    [anon_sym_nullable] = ACTIONS(7197),
    [anon_sym__Complex] = ACTIONS(7197),
    [anon_sym__Nullable] = ACTIONS(7197),
    [anon_sym__Nullable_result] = ACTIONS(7197),
    [anon_sym__Null_unspecified] = ACTIONS(7197),
    [anon_sym___autoreleasing] = ACTIONS(7197),
    [anon_sym___block] = ACTIONS(7197),
    [anon_sym___bridge] = ACTIONS(7197),
    [anon_sym___bridge_retained] = ACTIONS(7197),
    [anon_sym___bridge_transfer] = ACTIONS(7197),
    [anon_sym___complex] = ACTIONS(7197),
    [anon_sym___const] = ACTIONS(7197),
    [anon_sym___imag] = ACTIONS(7197),
    [anon_sym___kindof] = ACTIONS(7197),
    [anon_sym___nonnull] = ACTIONS(7197),
    [anon_sym___nullable] = ACTIONS(7197),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7197),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7197),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7197),
    [anon_sym___real] = ACTIONS(7197),
    [anon_sym___strong] = ACTIONS(7197),
    [anon_sym___unsafe_unretained] = ACTIONS(7197),
    [anon_sym___unused] = ACTIONS(7197),
    [anon_sym___weak] = ACTIONS(7197),
    [anon_sym_alignas] = ACTIONS(7197),
    [anon_sym__Alignas] = ACTIONS(7197),
    [sym_primitive_type] = ACTIONS(7197),
    [anon_sym_enum] = ACTIONS(7197),
    [anon_sym_COLON] = ACTIONS(7199),
    [anon_sym_struct] = ACTIONS(7197),
    [anon_sym_union] = ACTIONS(7197),
    [anon_sym_in] = ACTIONS(7197),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7197),
    [anon_sym___typeof] = ACTIONS(7197),
    [anon_sym_typeof] = ACTIONS(7197),
    [anon_sym_BOOL] = ACTIONS(7197),
    [anon_sym_IMP] = ACTIONS(7197),
    [anon_sym_SEL] = ACTIONS(7197),
    [anon_sym_Class] = ACTIONS(7197),
    [anon_sym_id] = ACTIONS(7197),
    [anon_sym_out] = ACTIONS(7197),
    [anon_sym_inout] = ACTIONS(7197),
    [anon_sym_bycopy] = ACTIONS(7197),
    [anon_sym_byref] = ACTIONS(7197),
    [anon_sym_oneway] = ACTIONS(7197),
  },
  [STATE(3877)] = {
    [sym_identifier] = ACTIONS(7311),
    [anon_sym_COMMA] = ACTIONS(7313),
    [anon_sym_RPAREN] = ACTIONS(7313),
    [anon_sym_LPAREN2] = ACTIONS(7313),
    [anon_sym_STAR] = ACTIONS(7313),
    [anon_sym_CARET] = ACTIONS(7313),
    [anon_sym_GT] = ACTIONS(7313),
    [anon_sym_LT] = ACTIONS(7313),
    [anon_sym___extension__] = ACTIONS(7311),
    [anon_sym___attribute__] = ACTIONS(7311),
    [anon_sym___attribute] = ACTIONS(7311),
    [anon_sym_noreturn] = ACTIONS(7311),
    [anon_sym_LBRACK] = ACTIONS(7313),
    [anon_sym___based] = ACTIONS(7311),
    [anon_sym_LBRACE] = ACTIONS(7313),
    [anon_sym_signed] = ACTIONS(7311),
    [anon_sym_unsigned] = ACTIONS(7311),
    [anon_sym_long] = ACTIONS(7311),
    [anon_sym_short] = ACTIONS(7311),
    [anon_sym_ATautoreleasepool] = ACTIONS(7313),
    [anon_sym_const] = ACTIONS(7311),
    [anon_sym_constexpr] = ACTIONS(7311),
    [anon_sym_volatile] = ACTIONS(7311),
    [anon_sym_restrict] = ACTIONS(7311),
    [anon_sym___restrict__] = ACTIONS(7311),
    [anon_sym__Atomic] = ACTIONS(7311),
    [anon_sym__Noreturn] = ACTIONS(7311),
    [anon_sym__Nonnull] = ACTIONS(7311),
    [anon_sym_nullable] = ACTIONS(7311),
    [anon_sym__Complex] = ACTIONS(7311),
    [anon_sym__Nullable] = ACTIONS(7311),
    [anon_sym__Nullable_result] = ACTIONS(7311),
    [anon_sym__Null_unspecified] = ACTIONS(7311),
    [anon_sym___autoreleasing] = ACTIONS(7311),
    [anon_sym___block] = ACTIONS(7311),
    [anon_sym___bridge] = ACTIONS(7311),
    [anon_sym___bridge_retained] = ACTIONS(7311),
    [anon_sym___bridge_transfer] = ACTIONS(7311),
    [anon_sym___complex] = ACTIONS(7311),
    [anon_sym___const] = ACTIONS(7311),
    [anon_sym___imag] = ACTIONS(7311),
    [anon_sym___kindof] = ACTIONS(7311),
    [anon_sym___nonnull] = ACTIONS(7311),
    [anon_sym___nullable] = ACTIONS(7311),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7311),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7311),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7311),
    [anon_sym___real] = ACTIONS(7311),
    [anon_sym___strong] = ACTIONS(7311),
    [anon_sym___unsafe_unretained] = ACTIONS(7311),
    [anon_sym___unused] = ACTIONS(7311),
    [anon_sym___weak] = ACTIONS(7311),
    [anon_sym_alignas] = ACTIONS(7311),
    [anon_sym__Alignas] = ACTIONS(7311),
    [sym_primitive_type] = ACTIONS(7311),
    [anon_sym_enum] = ACTIONS(7311),
    [anon_sym_COLON] = ACTIONS(7313),
    [anon_sym_struct] = ACTIONS(7311),
    [anon_sym_union] = ACTIONS(7311),
    [anon_sym_in] = ACTIONS(7311),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7311),
    [anon_sym___typeof] = ACTIONS(7311),
    [anon_sym_typeof] = ACTIONS(7311),
    [anon_sym_BOOL] = ACTIONS(7311),
    [anon_sym_IMP] = ACTIONS(7311),
    [anon_sym_SEL] = ACTIONS(7311),
    [anon_sym_Class] = ACTIONS(7311),
    [anon_sym_id] = ACTIONS(7311),
    [anon_sym_out] = ACTIONS(7311),
    [anon_sym_inout] = ACTIONS(7311),
    [anon_sym_bycopy] = ACTIONS(7311),
    [anon_sym_byref] = ACTIONS(7311),
    [anon_sym_oneway] = ACTIONS(7311),
  },
  [STATE(3878)] = {
    [sym_identifier] = ACTIONS(7171),
    [anon_sym_COMMA] = ACTIONS(7173),
    [anon_sym_RPAREN] = ACTIONS(7173),
    [anon_sym_LPAREN2] = ACTIONS(7173),
    [anon_sym_STAR] = ACTIONS(7173),
    [anon_sym_CARET] = ACTIONS(7173),
    [anon_sym_GT] = ACTIONS(7173),
    [anon_sym_LT] = ACTIONS(7173),
    [anon_sym___extension__] = ACTIONS(7171),
    [anon_sym___attribute__] = ACTIONS(7171),
    [anon_sym___attribute] = ACTIONS(7171),
    [anon_sym_noreturn] = ACTIONS(7171),
    [anon_sym_LBRACK] = ACTIONS(7173),
    [anon_sym___based] = ACTIONS(7171),
    [anon_sym_LBRACE] = ACTIONS(7173),
    [anon_sym_signed] = ACTIONS(7171),
    [anon_sym_unsigned] = ACTIONS(7171),
    [anon_sym_long] = ACTIONS(7171),
    [anon_sym_short] = ACTIONS(7171),
    [anon_sym_ATautoreleasepool] = ACTIONS(7173),
    [anon_sym_const] = ACTIONS(7171),
    [anon_sym_constexpr] = ACTIONS(7171),
    [anon_sym_volatile] = ACTIONS(7171),
    [anon_sym_restrict] = ACTIONS(7171),
    [anon_sym___restrict__] = ACTIONS(7171),
    [anon_sym__Atomic] = ACTIONS(7171),
    [anon_sym__Noreturn] = ACTIONS(7171),
    [anon_sym__Nonnull] = ACTIONS(7171),
    [anon_sym_nullable] = ACTIONS(7171),
    [anon_sym__Complex] = ACTIONS(7171),
    [anon_sym__Nullable] = ACTIONS(7171),
    [anon_sym__Nullable_result] = ACTIONS(7171),
    [anon_sym__Null_unspecified] = ACTIONS(7171),
    [anon_sym___autoreleasing] = ACTIONS(7171),
    [anon_sym___block] = ACTIONS(7171),
    [anon_sym___bridge] = ACTIONS(7171),
    [anon_sym___bridge_retained] = ACTIONS(7171),
    [anon_sym___bridge_transfer] = ACTIONS(7171),
    [anon_sym___complex] = ACTIONS(7171),
    [anon_sym___const] = ACTIONS(7171),
    [anon_sym___imag] = ACTIONS(7171),
    [anon_sym___kindof] = ACTIONS(7171),
    [anon_sym___nonnull] = ACTIONS(7171),
    [anon_sym___nullable] = ACTIONS(7171),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7171),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7171),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7171),
    [anon_sym___real] = ACTIONS(7171),
    [anon_sym___strong] = ACTIONS(7171),
    [anon_sym___unsafe_unretained] = ACTIONS(7171),
    [anon_sym___unused] = ACTIONS(7171),
    [anon_sym___weak] = ACTIONS(7171),
    [anon_sym_alignas] = ACTIONS(7171),
    [anon_sym__Alignas] = ACTIONS(7171),
    [sym_primitive_type] = ACTIONS(7171),
    [anon_sym_enum] = ACTIONS(7171),
    [anon_sym_COLON] = ACTIONS(7173),
    [anon_sym_struct] = ACTIONS(7171),
    [anon_sym_union] = ACTIONS(7171),
    [anon_sym_in] = ACTIONS(7171),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7171),
    [anon_sym___typeof] = ACTIONS(7171),
    [anon_sym_typeof] = ACTIONS(7171),
    [anon_sym_BOOL] = ACTIONS(7171),
    [anon_sym_IMP] = ACTIONS(7171),
    [anon_sym_SEL] = ACTIONS(7171),
    [anon_sym_Class] = ACTIONS(7171),
    [anon_sym_id] = ACTIONS(7171),
    [anon_sym_out] = ACTIONS(7171),
    [anon_sym_inout] = ACTIONS(7171),
    [anon_sym_bycopy] = ACTIONS(7171),
    [anon_sym_byref] = ACTIONS(7171),
    [anon_sym_oneway] = ACTIONS(7171),
  },
  [STATE(3879)] = {
    [sym_identifier] = ACTIONS(7209),
    [anon_sym_COMMA] = ACTIONS(7211),
    [anon_sym_RPAREN] = ACTIONS(7211),
    [anon_sym_LPAREN2] = ACTIONS(7211),
    [anon_sym_STAR] = ACTIONS(7211),
    [anon_sym_CARET] = ACTIONS(7211),
    [anon_sym_GT] = ACTIONS(7211),
    [anon_sym_LT] = ACTIONS(7211),
    [anon_sym___extension__] = ACTIONS(7209),
    [anon_sym___attribute__] = ACTIONS(7209),
    [anon_sym___attribute] = ACTIONS(7209),
    [anon_sym_noreturn] = ACTIONS(7209),
    [anon_sym_LBRACK] = ACTIONS(7211),
    [anon_sym___based] = ACTIONS(7209),
    [anon_sym_LBRACE] = ACTIONS(7211),
    [anon_sym_signed] = ACTIONS(7209),
    [anon_sym_unsigned] = ACTIONS(7209),
    [anon_sym_long] = ACTIONS(7209),
    [anon_sym_short] = ACTIONS(7209),
    [anon_sym_ATautoreleasepool] = ACTIONS(7211),
    [anon_sym_const] = ACTIONS(7209),
    [anon_sym_constexpr] = ACTIONS(7209),
    [anon_sym_volatile] = ACTIONS(7209),
    [anon_sym_restrict] = ACTIONS(7209),
    [anon_sym___restrict__] = ACTIONS(7209),
    [anon_sym__Atomic] = ACTIONS(7209),
    [anon_sym__Noreturn] = ACTIONS(7209),
    [anon_sym__Nonnull] = ACTIONS(7209),
    [anon_sym_nullable] = ACTIONS(7209),
    [anon_sym__Complex] = ACTIONS(7209),
    [anon_sym__Nullable] = ACTIONS(7209),
    [anon_sym__Nullable_result] = ACTIONS(7209),
    [anon_sym__Null_unspecified] = ACTIONS(7209),
    [anon_sym___autoreleasing] = ACTIONS(7209),
    [anon_sym___block] = ACTIONS(7209),
    [anon_sym___bridge] = ACTIONS(7209),
    [anon_sym___bridge_retained] = ACTIONS(7209),
    [anon_sym___bridge_transfer] = ACTIONS(7209),
    [anon_sym___complex] = ACTIONS(7209),
    [anon_sym___const] = ACTIONS(7209),
    [anon_sym___imag] = ACTIONS(7209),
    [anon_sym___kindof] = ACTIONS(7209),
    [anon_sym___nonnull] = ACTIONS(7209),
    [anon_sym___nullable] = ACTIONS(7209),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7209),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7209),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7209),
    [anon_sym___real] = ACTIONS(7209),
    [anon_sym___strong] = ACTIONS(7209),
    [anon_sym___unsafe_unretained] = ACTIONS(7209),
    [anon_sym___unused] = ACTIONS(7209),
    [anon_sym___weak] = ACTIONS(7209),
    [anon_sym_alignas] = ACTIONS(7209),
    [anon_sym__Alignas] = ACTIONS(7209),
    [sym_primitive_type] = ACTIONS(7209),
    [anon_sym_enum] = ACTIONS(7209),
    [anon_sym_COLON] = ACTIONS(7211),
    [anon_sym_struct] = ACTIONS(7209),
    [anon_sym_union] = ACTIONS(7209),
    [anon_sym_in] = ACTIONS(7209),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7209),
    [anon_sym___typeof] = ACTIONS(7209),
    [anon_sym_typeof] = ACTIONS(7209),
    [anon_sym_BOOL] = ACTIONS(7209),
    [anon_sym_IMP] = ACTIONS(7209),
    [anon_sym_SEL] = ACTIONS(7209),
    [anon_sym_Class] = ACTIONS(7209),
    [anon_sym_id] = ACTIONS(7209),
    [anon_sym_out] = ACTIONS(7209),
    [anon_sym_inout] = ACTIONS(7209),
    [anon_sym_bycopy] = ACTIONS(7209),
    [anon_sym_byref] = ACTIONS(7209),
    [anon_sym_oneway] = ACTIONS(7209),
  },
  [STATE(3880)] = {
    [sym__type_definition_type] = STATE(4769),
    [sym_attribute_specifier] = STATE(3886),
    [sym_ms_declspec_modifier] = STATE(4038),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3881)] = {
    [sym__type_definition_type] = STATE(4771),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4038),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3882)] = {
    [sym_attribute_specifier] = STATE(3956),
    [sym_identifier] = ACTIONS(7927),
    [anon_sym_COMMA] = ACTIONS(7929),
    [anon_sym_RPAREN] = ACTIONS(7929),
    [anon_sym_LPAREN2] = ACTIONS(7929),
    [anon_sym_STAR] = ACTIONS(7929),
    [anon_sym_CARET] = ACTIONS(7929),
    [anon_sym_GT] = ACTIONS(7929),
    [anon_sym_LT] = ACTIONS(7929),
    [anon_sym___extension__] = ACTIONS(7927),
    [anon_sym___attribute__] = ACTIONS(8878),
    [anon_sym___attribute] = ACTIONS(8878),
    [anon_sym_noreturn] = ACTIONS(7927),
    [anon_sym_LBRACK] = ACTIONS(7929),
    [anon_sym___based] = ACTIONS(7927),
    [anon_sym_LBRACE] = ACTIONS(7929),
    [anon_sym_signed] = ACTIONS(7927),
    [anon_sym_unsigned] = ACTIONS(7927),
    [anon_sym_long] = ACTIONS(7927),
    [anon_sym_short] = ACTIONS(7927),
    [anon_sym_ATautoreleasepool] = ACTIONS(7929),
    [anon_sym_const] = ACTIONS(7927),
    [anon_sym_constexpr] = ACTIONS(7927),
    [anon_sym_volatile] = ACTIONS(7927),
    [anon_sym_restrict] = ACTIONS(7927),
    [anon_sym___restrict__] = ACTIONS(7927),
    [anon_sym__Atomic] = ACTIONS(7927),
    [anon_sym__Noreturn] = ACTIONS(7927),
    [anon_sym__Nonnull] = ACTIONS(7927),
    [anon_sym_nullable] = ACTIONS(7927),
    [anon_sym__Complex] = ACTIONS(7927),
    [anon_sym__Nullable] = ACTIONS(7927),
    [anon_sym__Nullable_result] = ACTIONS(7927),
    [anon_sym__Null_unspecified] = ACTIONS(7927),
    [anon_sym___autoreleasing] = ACTIONS(7927),
    [anon_sym___block] = ACTIONS(7927),
    [anon_sym___bridge] = ACTIONS(7927),
    [anon_sym___bridge_retained] = ACTIONS(7927),
    [anon_sym___bridge_transfer] = ACTIONS(7927),
    [anon_sym___complex] = ACTIONS(7927),
    [anon_sym___const] = ACTIONS(7927),
    [anon_sym___imag] = ACTIONS(7927),
    [anon_sym___kindof] = ACTIONS(7927),
    [anon_sym___nonnull] = ACTIONS(7927),
    [anon_sym___nullable] = ACTIONS(7927),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7927),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7927),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7927),
    [anon_sym___real] = ACTIONS(7927),
    [anon_sym___strong] = ACTIONS(7927),
    [anon_sym___unsafe_unretained] = ACTIONS(7927),
    [anon_sym___unused] = ACTIONS(7927),
    [anon_sym___weak] = ACTIONS(7927),
    [anon_sym_alignas] = ACTIONS(7927),
    [anon_sym__Alignas] = ACTIONS(7927),
    [sym_primitive_type] = ACTIONS(7927),
    [anon_sym_enum] = ACTIONS(7927),
    [anon_sym_struct] = ACTIONS(7927),
    [anon_sym_union] = ACTIONS(7927),
    [anon_sym_in] = ACTIONS(7927),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7927),
    [anon_sym___typeof] = ACTIONS(7927),
    [anon_sym_typeof] = ACTIONS(7927),
    [anon_sym_BOOL] = ACTIONS(7927),
    [anon_sym_IMP] = ACTIONS(7927),
    [anon_sym_SEL] = ACTIONS(7927),
    [anon_sym_Class] = ACTIONS(7927),
    [anon_sym_id] = ACTIONS(7927),
    [anon_sym_out] = ACTIONS(7927),
    [anon_sym_inout] = ACTIONS(7927),
    [anon_sym_bycopy] = ACTIONS(7927),
    [anon_sym_byref] = ACTIONS(7927),
    [anon_sym_oneway] = ACTIONS(7927),
  },
  [STATE(3883)] = {
    [sym_attribute_specifier] = STATE(3873),
    [sym_identifier] = ACTIONS(7875),
    [anon_sym_COMMA] = ACTIONS(7877),
    [anon_sym_RPAREN] = ACTIONS(7877),
    [anon_sym_LPAREN2] = ACTIONS(7877),
    [anon_sym_STAR] = ACTIONS(7877),
    [anon_sym_CARET] = ACTIONS(7877),
    [anon_sym_GT] = ACTIONS(7877),
    [anon_sym_LT] = ACTIONS(7877),
    [anon_sym___extension__] = ACTIONS(7875),
    [anon_sym___attribute__] = ACTIONS(8881),
    [anon_sym___attribute] = ACTIONS(8881),
    [anon_sym_noreturn] = ACTIONS(7875),
    [anon_sym_LBRACK] = ACTIONS(7877),
    [anon_sym___based] = ACTIONS(7875),
    [anon_sym_LBRACE] = ACTIONS(7877),
    [anon_sym_signed] = ACTIONS(7875),
    [anon_sym_unsigned] = ACTIONS(7875),
    [anon_sym_long] = ACTIONS(7875),
    [anon_sym_short] = ACTIONS(7875),
    [anon_sym_ATautoreleasepool] = ACTIONS(7877),
    [anon_sym_const] = ACTIONS(7875),
    [anon_sym_constexpr] = ACTIONS(7875),
    [anon_sym_volatile] = ACTIONS(7875),
    [anon_sym_restrict] = ACTIONS(7875),
    [anon_sym___restrict__] = ACTIONS(7875),
    [anon_sym__Atomic] = ACTIONS(7875),
    [anon_sym__Noreturn] = ACTIONS(7875),
    [anon_sym__Nonnull] = ACTIONS(7875),
    [anon_sym_nullable] = ACTIONS(7875),
    [anon_sym__Complex] = ACTIONS(7875),
    [anon_sym__Nullable] = ACTIONS(7875),
    [anon_sym__Nullable_result] = ACTIONS(7875),
    [anon_sym__Null_unspecified] = ACTIONS(7875),
    [anon_sym___autoreleasing] = ACTIONS(7875),
    [anon_sym___block] = ACTIONS(7875),
    [anon_sym___bridge] = ACTIONS(7875),
    [anon_sym___bridge_retained] = ACTIONS(7875),
    [anon_sym___bridge_transfer] = ACTIONS(7875),
    [anon_sym___complex] = ACTIONS(7875),
    [anon_sym___const] = ACTIONS(7875),
    [anon_sym___imag] = ACTIONS(7875),
    [anon_sym___kindof] = ACTIONS(7875),
    [anon_sym___nonnull] = ACTIONS(7875),
    [anon_sym___nullable] = ACTIONS(7875),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7875),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7875),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7875),
    [anon_sym___real] = ACTIONS(7875),
    [anon_sym___strong] = ACTIONS(7875),
    [anon_sym___unsafe_unretained] = ACTIONS(7875),
    [anon_sym___unused] = ACTIONS(7875),
    [anon_sym___weak] = ACTIONS(7875),
    [anon_sym_alignas] = ACTIONS(7875),
    [anon_sym__Alignas] = ACTIONS(7875),
    [sym_primitive_type] = ACTIONS(7875),
    [anon_sym_enum] = ACTIONS(7875),
    [anon_sym_struct] = ACTIONS(7875),
    [anon_sym_union] = ACTIONS(7875),
    [anon_sym_in] = ACTIONS(7875),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7875),
    [anon_sym___typeof] = ACTIONS(7875),
    [anon_sym_typeof] = ACTIONS(7875),
    [anon_sym_BOOL] = ACTIONS(7875),
    [anon_sym_IMP] = ACTIONS(7875),
    [anon_sym_SEL] = ACTIONS(7875),
    [anon_sym_Class] = ACTIONS(7875),
    [anon_sym_id] = ACTIONS(7875),
    [anon_sym_out] = ACTIONS(7875),
    [anon_sym_inout] = ACTIONS(7875),
    [anon_sym_bycopy] = ACTIONS(7875),
    [anon_sym_byref] = ACTIONS(7875),
    [anon_sym_oneway] = ACTIONS(7875),
  },
  [STATE(3884)] = {
    [sym_identifier] = ACTIONS(7333),
    [anon_sym_COMMA] = ACTIONS(7335),
    [anon_sym_RPAREN] = ACTIONS(7335),
    [anon_sym_LPAREN2] = ACTIONS(7335),
    [anon_sym_STAR] = ACTIONS(7335),
    [anon_sym_CARET] = ACTIONS(7335),
    [anon_sym_GT] = ACTIONS(7335),
    [anon_sym_LT] = ACTIONS(7335),
    [anon_sym___extension__] = ACTIONS(7333),
    [anon_sym___attribute__] = ACTIONS(7333),
    [anon_sym___attribute] = ACTIONS(7333),
    [anon_sym_noreturn] = ACTIONS(7333),
    [anon_sym_LBRACK] = ACTIONS(7335),
    [anon_sym___based] = ACTIONS(7333),
    [anon_sym_LBRACE] = ACTIONS(7335),
    [anon_sym_signed] = ACTIONS(7333),
    [anon_sym_unsigned] = ACTIONS(7333),
    [anon_sym_long] = ACTIONS(7333),
    [anon_sym_short] = ACTIONS(7333),
    [anon_sym_ATautoreleasepool] = ACTIONS(7335),
    [anon_sym_const] = ACTIONS(7333),
    [anon_sym_constexpr] = ACTIONS(7333),
    [anon_sym_volatile] = ACTIONS(7333),
    [anon_sym_restrict] = ACTIONS(7333),
    [anon_sym___restrict__] = ACTIONS(7333),
    [anon_sym__Atomic] = ACTIONS(7333),
    [anon_sym__Noreturn] = ACTIONS(7333),
    [anon_sym__Nonnull] = ACTIONS(7333),
    [anon_sym_nullable] = ACTIONS(7333),
    [anon_sym__Complex] = ACTIONS(7333),
    [anon_sym__Nullable] = ACTIONS(7333),
    [anon_sym__Nullable_result] = ACTIONS(7333),
    [anon_sym__Null_unspecified] = ACTIONS(7333),
    [anon_sym___autoreleasing] = ACTIONS(7333),
    [anon_sym___block] = ACTIONS(7333),
    [anon_sym___bridge] = ACTIONS(7333),
    [anon_sym___bridge_retained] = ACTIONS(7333),
    [anon_sym___bridge_transfer] = ACTIONS(7333),
    [anon_sym___complex] = ACTIONS(7333),
    [anon_sym___const] = ACTIONS(7333),
    [anon_sym___imag] = ACTIONS(7333),
    [anon_sym___kindof] = ACTIONS(7333),
    [anon_sym___nonnull] = ACTIONS(7333),
    [anon_sym___nullable] = ACTIONS(7333),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7333),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7333),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7333),
    [anon_sym___real] = ACTIONS(7333),
    [anon_sym___strong] = ACTIONS(7333),
    [anon_sym___unsafe_unretained] = ACTIONS(7333),
    [anon_sym___unused] = ACTIONS(7333),
    [anon_sym___weak] = ACTIONS(7333),
    [anon_sym_alignas] = ACTIONS(7333),
    [anon_sym__Alignas] = ACTIONS(7333),
    [sym_primitive_type] = ACTIONS(7333),
    [anon_sym_enum] = ACTIONS(7333),
    [anon_sym_COLON] = ACTIONS(7335),
    [anon_sym_struct] = ACTIONS(7333),
    [anon_sym_union] = ACTIONS(7333),
    [anon_sym_in] = ACTIONS(7333),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7333),
    [anon_sym___typeof] = ACTIONS(7333),
    [anon_sym_typeof] = ACTIONS(7333),
    [anon_sym_BOOL] = ACTIONS(7333),
    [anon_sym_IMP] = ACTIONS(7333),
    [anon_sym_SEL] = ACTIONS(7333),
    [anon_sym_Class] = ACTIONS(7333),
    [anon_sym_id] = ACTIONS(7333),
    [anon_sym_out] = ACTIONS(7333),
    [anon_sym_inout] = ACTIONS(7333),
    [anon_sym_bycopy] = ACTIONS(7333),
    [anon_sym_byref] = ACTIONS(7333),
    [anon_sym_oneway] = ACTIONS(7333),
  },
  [STATE(3885)] = {
    [sym_attribute_specifier] = STATE(3874),
    [sym_identifier] = ACTIONS(7905),
    [anon_sym_COMMA] = ACTIONS(7907),
    [anon_sym_RPAREN] = ACTIONS(7907),
    [anon_sym_LPAREN2] = ACTIONS(7907),
    [anon_sym_STAR] = ACTIONS(7907),
    [anon_sym_CARET] = ACTIONS(7907),
    [anon_sym_GT] = ACTIONS(7907),
    [anon_sym_LT] = ACTIONS(7907),
    [anon_sym___extension__] = ACTIONS(7905),
    [anon_sym___attribute__] = ACTIONS(8884),
    [anon_sym___attribute] = ACTIONS(8884),
    [anon_sym_noreturn] = ACTIONS(7905),
    [anon_sym_LBRACK] = ACTIONS(7907),
    [anon_sym___based] = ACTIONS(7905),
    [anon_sym_LBRACE] = ACTIONS(7907),
    [anon_sym_signed] = ACTIONS(7905),
    [anon_sym_unsigned] = ACTIONS(7905),
    [anon_sym_long] = ACTIONS(7905),
    [anon_sym_short] = ACTIONS(7905),
    [anon_sym_ATautoreleasepool] = ACTIONS(7907),
    [anon_sym_const] = ACTIONS(7905),
    [anon_sym_constexpr] = ACTIONS(7905),
    [anon_sym_volatile] = ACTIONS(7905),
    [anon_sym_restrict] = ACTIONS(7905),
    [anon_sym___restrict__] = ACTIONS(7905),
    [anon_sym__Atomic] = ACTIONS(7905),
    [anon_sym__Noreturn] = ACTIONS(7905),
    [anon_sym__Nonnull] = ACTIONS(7905),
    [anon_sym_nullable] = ACTIONS(7905),
    [anon_sym__Complex] = ACTIONS(7905),
    [anon_sym__Nullable] = ACTIONS(7905),
    [anon_sym__Nullable_result] = ACTIONS(7905),
    [anon_sym__Null_unspecified] = ACTIONS(7905),
    [anon_sym___autoreleasing] = ACTIONS(7905),
    [anon_sym___block] = ACTIONS(7905),
    [anon_sym___bridge] = ACTIONS(7905),
    [anon_sym___bridge_retained] = ACTIONS(7905),
    [anon_sym___bridge_transfer] = ACTIONS(7905),
    [anon_sym___complex] = ACTIONS(7905),
    [anon_sym___const] = ACTIONS(7905),
    [anon_sym___imag] = ACTIONS(7905),
    [anon_sym___kindof] = ACTIONS(7905),
    [anon_sym___nonnull] = ACTIONS(7905),
    [anon_sym___nullable] = ACTIONS(7905),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7905),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7905),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7905),
    [anon_sym___real] = ACTIONS(7905),
    [anon_sym___strong] = ACTIONS(7905),
    [anon_sym___unsafe_unretained] = ACTIONS(7905),
    [anon_sym___unused] = ACTIONS(7905),
    [anon_sym___weak] = ACTIONS(7905),
    [anon_sym_alignas] = ACTIONS(7905),
    [anon_sym__Alignas] = ACTIONS(7905),
    [sym_primitive_type] = ACTIONS(7905),
    [anon_sym_enum] = ACTIONS(7905),
    [anon_sym_struct] = ACTIONS(7905),
    [anon_sym_union] = ACTIONS(7905),
    [anon_sym_in] = ACTIONS(7905),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7905),
    [anon_sym___typeof] = ACTIONS(7905),
    [anon_sym_typeof] = ACTIONS(7905),
    [anon_sym_BOOL] = ACTIONS(7905),
    [anon_sym_IMP] = ACTIONS(7905),
    [anon_sym_SEL] = ACTIONS(7905),
    [anon_sym_Class] = ACTIONS(7905),
    [anon_sym_id] = ACTIONS(7905),
    [anon_sym_out] = ACTIONS(7905),
    [anon_sym_inout] = ACTIONS(7905),
    [anon_sym_bycopy] = ACTIONS(7905),
    [anon_sym_byref] = ACTIONS(7905),
    [anon_sym_oneway] = ACTIONS(7905),
  },
  [STATE(3886)] = {
    [sym__type_definition_type] = STATE(4779),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4016),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3887)] = {
    [sym__type_definition_type] = STATE(4781),
    [sym_attribute_specifier] = STATE(3893),
    [sym_ms_declspec_modifier] = STATE(4016),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3888)] = {
    [sym_attribute_specifier] = STATE(3889),
    [sym_identifier] = ACTIONS(7920),
    [anon_sym_COMMA] = ACTIONS(7922),
    [anon_sym_RPAREN] = ACTIONS(7922),
    [anon_sym_LPAREN2] = ACTIONS(7922),
    [anon_sym_STAR] = ACTIONS(7922),
    [anon_sym_CARET] = ACTIONS(7922),
    [anon_sym_GT] = ACTIONS(7922),
    [anon_sym_LT] = ACTIONS(7922),
    [anon_sym___extension__] = ACTIONS(7920),
    [anon_sym___attribute__] = ACTIONS(8887),
    [anon_sym___attribute] = ACTIONS(8887),
    [anon_sym_noreturn] = ACTIONS(7920),
    [anon_sym_LBRACK] = ACTIONS(7922),
    [anon_sym___based] = ACTIONS(7920),
    [anon_sym_LBRACE] = ACTIONS(7922),
    [anon_sym_signed] = ACTIONS(7920),
    [anon_sym_unsigned] = ACTIONS(7920),
    [anon_sym_long] = ACTIONS(7920),
    [anon_sym_short] = ACTIONS(7920),
    [anon_sym_ATautoreleasepool] = ACTIONS(7922),
    [anon_sym_const] = ACTIONS(7920),
    [anon_sym_constexpr] = ACTIONS(7920),
    [anon_sym_volatile] = ACTIONS(7920),
    [anon_sym_restrict] = ACTIONS(7920),
    [anon_sym___restrict__] = ACTIONS(7920),
    [anon_sym__Atomic] = ACTIONS(7920),
    [anon_sym__Noreturn] = ACTIONS(7920),
    [anon_sym__Nonnull] = ACTIONS(7920),
    [anon_sym_nullable] = ACTIONS(7920),
    [anon_sym__Complex] = ACTIONS(7920),
    [anon_sym__Nullable] = ACTIONS(7920),
    [anon_sym__Nullable_result] = ACTIONS(7920),
    [anon_sym__Null_unspecified] = ACTIONS(7920),
    [anon_sym___autoreleasing] = ACTIONS(7920),
    [anon_sym___block] = ACTIONS(7920),
    [anon_sym___bridge] = ACTIONS(7920),
    [anon_sym___bridge_retained] = ACTIONS(7920),
    [anon_sym___bridge_transfer] = ACTIONS(7920),
    [anon_sym___complex] = ACTIONS(7920),
    [anon_sym___const] = ACTIONS(7920),
    [anon_sym___imag] = ACTIONS(7920),
    [anon_sym___kindof] = ACTIONS(7920),
    [anon_sym___nonnull] = ACTIONS(7920),
    [anon_sym___nullable] = ACTIONS(7920),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7920),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7920),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7920),
    [anon_sym___real] = ACTIONS(7920),
    [anon_sym___strong] = ACTIONS(7920),
    [anon_sym___unsafe_unretained] = ACTIONS(7920),
    [anon_sym___unused] = ACTIONS(7920),
    [anon_sym___weak] = ACTIONS(7920),
    [anon_sym_alignas] = ACTIONS(7920),
    [anon_sym__Alignas] = ACTIONS(7920),
    [sym_primitive_type] = ACTIONS(7920),
    [anon_sym_enum] = ACTIONS(7920),
    [anon_sym_struct] = ACTIONS(7920),
    [anon_sym_union] = ACTIONS(7920),
    [anon_sym_in] = ACTIONS(7920),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7920),
    [anon_sym___typeof] = ACTIONS(7920),
    [anon_sym_typeof] = ACTIONS(7920),
    [anon_sym_BOOL] = ACTIONS(7920),
    [anon_sym_IMP] = ACTIONS(7920),
    [anon_sym_SEL] = ACTIONS(7920),
    [anon_sym_Class] = ACTIONS(7920),
    [anon_sym_id] = ACTIONS(7920),
    [anon_sym_out] = ACTIONS(7920),
    [anon_sym_inout] = ACTIONS(7920),
    [anon_sym_bycopy] = ACTIONS(7920),
    [anon_sym_byref] = ACTIONS(7920),
    [anon_sym_oneway] = ACTIONS(7920),
  },
  [STATE(3889)] = {
    [sym_identifier] = ACTIONS(7265),
    [anon_sym_COMMA] = ACTIONS(7267),
    [anon_sym_RPAREN] = ACTIONS(7267),
    [anon_sym_LPAREN2] = ACTIONS(7267),
    [anon_sym_STAR] = ACTIONS(7267),
    [anon_sym_CARET] = ACTIONS(7267),
    [anon_sym_GT] = ACTIONS(7267),
    [anon_sym_LT] = ACTIONS(7267),
    [anon_sym___extension__] = ACTIONS(7265),
    [anon_sym___attribute__] = ACTIONS(7265),
    [anon_sym___attribute] = ACTIONS(7265),
    [anon_sym_noreturn] = ACTIONS(7265),
    [anon_sym_LBRACK] = ACTIONS(7267),
    [anon_sym___based] = ACTIONS(7265),
    [anon_sym_LBRACE] = ACTIONS(7267),
    [anon_sym_signed] = ACTIONS(7265),
    [anon_sym_unsigned] = ACTIONS(7265),
    [anon_sym_long] = ACTIONS(7265),
    [anon_sym_short] = ACTIONS(7265),
    [anon_sym_ATautoreleasepool] = ACTIONS(7267),
    [anon_sym_const] = ACTIONS(7265),
    [anon_sym_constexpr] = ACTIONS(7265),
    [anon_sym_volatile] = ACTIONS(7265),
    [anon_sym_restrict] = ACTIONS(7265),
    [anon_sym___restrict__] = ACTIONS(7265),
    [anon_sym__Atomic] = ACTIONS(7265),
    [anon_sym__Noreturn] = ACTIONS(7265),
    [anon_sym__Nonnull] = ACTIONS(7265),
    [anon_sym_nullable] = ACTIONS(7265),
    [anon_sym__Complex] = ACTIONS(7265),
    [anon_sym__Nullable] = ACTIONS(7265),
    [anon_sym__Nullable_result] = ACTIONS(7265),
    [anon_sym__Null_unspecified] = ACTIONS(7265),
    [anon_sym___autoreleasing] = ACTIONS(7265),
    [anon_sym___block] = ACTIONS(7265),
    [anon_sym___bridge] = ACTIONS(7265),
    [anon_sym___bridge_retained] = ACTIONS(7265),
    [anon_sym___bridge_transfer] = ACTIONS(7265),
    [anon_sym___complex] = ACTIONS(7265),
    [anon_sym___const] = ACTIONS(7265),
    [anon_sym___imag] = ACTIONS(7265),
    [anon_sym___kindof] = ACTIONS(7265),
    [anon_sym___nonnull] = ACTIONS(7265),
    [anon_sym___nullable] = ACTIONS(7265),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7265),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7265),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7265),
    [anon_sym___real] = ACTIONS(7265),
    [anon_sym___strong] = ACTIONS(7265),
    [anon_sym___unsafe_unretained] = ACTIONS(7265),
    [anon_sym___unused] = ACTIONS(7265),
    [anon_sym___weak] = ACTIONS(7265),
    [anon_sym_alignas] = ACTIONS(7265),
    [anon_sym__Alignas] = ACTIONS(7265),
    [sym_primitive_type] = ACTIONS(7265),
    [anon_sym_enum] = ACTIONS(7265),
    [anon_sym_COLON] = ACTIONS(7267),
    [anon_sym_struct] = ACTIONS(7265),
    [anon_sym_union] = ACTIONS(7265),
    [anon_sym_in] = ACTIONS(7265),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7265),
    [anon_sym___typeof] = ACTIONS(7265),
    [anon_sym_typeof] = ACTIONS(7265),
    [anon_sym_BOOL] = ACTIONS(7265),
    [anon_sym_IMP] = ACTIONS(7265),
    [anon_sym_SEL] = ACTIONS(7265),
    [anon_sym_Class] = ACTIONS(7265),
    [anon_sym_id] = ACTIONS(7265),
    [anon_sym_out] = ACTIONS(7265),
    [anon_sym_inout] = ACTIONS(7265),
    [anon_sym_bycopy] = ACTIONS(7265),
    [anon_sym_byref] = ACTIONS(7265),
    [anon_sym_oneway] = ACTIONS(7265),
  },
  [STATE(3890)] = {
    [sym_attribute_specifier] = STATE(3824),
    [sym_identifier] = ACTIONS(7940),
    [anon_sym_COMMA] = ACTIONS(7942),
    [anon_sym_RPAREN] = ACTIONS(7942),
    [anon_sym_LPAREN2] = ACTIONS(7942),
    [anon_sym_STAR] = ACTIONS(7942),
    [anon_sym_CARET] = ACTIONS(7942),
    [anon_sym_GT] = ACTIONS(7942),
    [anon_sym_LT] = ACTIONS(7942),
    [anon_sym___extension__] = ACTIONS(7940),
    [anon_sym___attribute__] = ACTIONS(8890),
    [anon_sym___attribute] = ACTIONS(8890),
    [anon_sym_noreturn] = ACTIONS(7940),
    [anon_sym_LBRACK] = ACTIONS(7942),
    [anon_sym___based] = ACTIONS(7940),
    [anon_sym_LBRACE] = ACTIONS(7942),
    [anon_sym_signed] = ACTIONS(7940),
    [anon_sym_unsigned] = ACTIONS(7940),
    [anon_sym_long] = ACTIONS(7940),
    [anon_sym_short] = ACTIONS(7940),
    [anon_sym_ATautoreleasepool] = ACTIONS(7942),
    [anon_sym_const] = ACTIONS(7940),
    [anon_sym_constexpr] = ACTIONS(7940),
    [anon_sym_volatile] = ACTIONS(7940),
    [anon_sym_restrict] = ACTIONS(7940),
    [anon_sym___restrict__] = ACTIONS(7940),
    [anon_sym__Atomic] = ACTIONS(7940),
    [anon_sym__Noreturn] = ACTIONS(7940),
    [anon_sym__Nonnull] = ACTIONS(7940),
    [anon_sym_nullable] = ACTIONS(7940),
    [anon_sym__Complex] = ACTIONS(7940),
    [anon_sym__Nullable] = ACTIONS(7940),
    [anon_sym__Nullable_result] = ACTIONS(7940),
    [anon_sym__Null_unspecified] = ACTIONS(7940),
    [anon_sym___autoreleasing] = ACTIONS(7940),
    [anon_sym___block] = ACTIONS(7940),
    [anon_sym___bridge] = ACTIONS(7940),
    [anon_sym___bridge_retained] = ACTIONS(7940),
    [anon_sym___bridge_transfer] = ACTIONS(7940),
    [anon_sym___complex] = ACTIONS(7940),
    [anon_sym___const] = ACTIONS(7940),
    [anon_sym___imag] = ACTIONS(7940),
    [anon_sym___kindof] = ACTIONS(7940),
    [anon_sym___nonnull] = ACTIONS(7940),
    [anon_sym___nullable] = ACTIONS(7940),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7940),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7940),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7940),
    [anon_sym___real] = ACTIONS(7940),
    [anon_sym___strong] = ACTIONS(7940),
    [anon_sym___unsafe_unretained] = ACTIONS(7940),
    [anon_sym___unused] = ACTIONS(7940),
    [anon_sym___weak] = ACTIONS(7940),
    [anon_sym_alignas] = ACTIONS(7940),
    [anon_sym__Alignas] = ACTIONS(7940),
    [sym_primitive_type] = ACTIONS(7940),
    [anon_sym_enum] = ACTIONS(7940),
    [anon_sym_struct] = ACTIONS(7940),
    [anon_sym_union] = ACTIONS(7940),
    [anon_sym_in] = ACTIONS(7940),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7940),
    [anon_sym___typeof] = ACTIONS(7940),
    [anon_sym_typeof] = ACTIONS(7940),
    [anon_sym_BOOL] = ACTIONS(7940),
    [anon_sym_IMP] = ACTIONS(7940),
    [anon_sym_SEL] = ACTIONS(7940),
    [anon_sym_Class] = ACTIONS(7940),
    [anon_sym_id] = ACTIONS(7940),
    [anon_sym_out] = ACTIONS(7940),
    [anon_sym_inout] = ACTIONS(7940),
    [anon_sym_bycopy] = ACTIONS(7940),
    [anon_sym_byref] = ACTIONS(7940),
    [anon_sym_oneway] = ACTIONS(7940),
  },
  [STATE(3891)] = {
    [sym_attribute_specifier] = STATE(3964),
    [sym_identifier] = ACTIONS(7789),
    [anon_sym_COMMA] = ACTIONS(7791),
    [anon_sym_RPAREN] = ACTIONS(7791),
    [anon_sym_LPAREN2] = ACTIONS(7791),
    [anon_sym_STAR] = ACTIONS(7791),
    [anon_sym_CARET] = ACTIONS(7791),
    [anon_sym_GT] = ACTIONS(7791),
    [anon_sym_LT] = ACTIONS(7791),
    [anon_sym___extension__] = ACTIONS(7789),
    [anon_sym___attribute__] = ACTIONS(8893),
    [anon_sym___attribute] = ACTIONS(8893),
    [anon_sym_noreturn] = ACTIONS(7789),
    [anon_sym_LBRACK] = ACTIONS(7791),
    [anon_sym___based] = ACTIONS(7789),
    [anon_sym_LBRACE] = ACTIONS(7791),
    [anon_sym_signed] = ACTIONS(7789),
    [anon_sym_unsigned] = ACTIONS(7789),
    [anon_sym_long] = ACTIONS(7789),
    [anon_sym_short] = ACTIONS(7789),
    [anon_sym_ATautoreleasepool] = ACTIONS(7791),
    [anon_sym_const] = ACTIONS(7789),
    [anon_sym_constexpr] = ACTIONS(7789),
    [anon_sym_volatile] = ACTIONS(7789),
    [anon_sym_restrict] = ACTIONS(7789),
    [anon_sym___restrict__] = ACTIONS(7789),
    [anon_sym__Atomic] = ACTIONS(7789),
    [anon_sym__Noreturn] = ACTIONS(7789),
    [anon_sym__Nonnull] = ACTIONS(7789),
    [anon_sym_nullable] = ACTIONS(7789),
    [anon_sym__Complex] = ACTIONS(7789),
    [anon_sym__Nullable] = ACTIONS(7789),
    [anon_sym__Nullable_result] = ACTIONS(7789),
    [anon_sym__Null_unspecified] = ACTIONS(7789),
    [anon_sym___autoreleasing] = ACTIONS(7789),
    [anon_sym___block] = ACTIONS(7789),
    [anon_sym___bridge] = ACTIONS(7789),
    [anon_sym___bridge_retained] = ACTIONS(7789),
    [anon_sym___bridge_transfer] = ACTIONS(7789),
    [anon_sym___complex] = ACTIONS(7789),
    [anon_sym___const] = ACTIONS(7789),
    [anon_sym___imag] = ACTIONS(7789),
    [anon_sym___kindof] = ACTIONS(7789),
    [anon_sym___nonnull] = ACTIONS(7789),
    [anon_sym___nullable] = ACTIONS(7789),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7789),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7789),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7789),
    [anon_sym___real] = ACTIONS(7789),
    [anon_sym___strong] = ACTIONS(7789),
    [anon_sym___unsafe_unretained] = ACTIONS(7789),
    [anon_sym___unused] = ACTIONS(7789),
    [anon_sym___weak] = ACTIONS(7789),
    [anon_sym_alignas] = ACTIONS(7789),
    [anon_sym__Alignas] = ACTIONS(7789),
    [sym_primitive_type] = ACTIONS(7789),
    [anon_sym_enum] = ACTIONS(7789),
    [anon_sym_struct] = ACTIONS(7789),
    [anon_sym_union] = ACTIONS(7789),
    [anon_sym_in] = ACTIONS(7789),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7789),
    [anon_sym___typeof] = ACTIONS(7789),
    [anon_sym_typeof] = ACTIONS(7789),
    [anon_sym_BOOL] = ACTIONS(7789),
    [anon_sym_IMP] = ACTIONS(7789),
    [anon_sym_SEL] = ACTIONS(7789),
    [anon_sym_Class] = ACTIONS(7789),
    [anon_sym_id] = ACTIONS(7789),
    [anon_sym_out] = ACTIONS(7789),
    [anon_sym_inout] = ACTIONS(7789),
    [anon_sym_bycopy] = ACTIONS(7789),
    [anon_sym_byref] = ACTIONS(7789),
    [anon_sym_oneway] = ACTIONS(7789),
  },
  [STATE(3892)] = {
    [sym_identifier] = ACTIONS(7279),
    [anon_sym_COMMA] = ACTIONS(7281),
    [anon_sym_RPAREN] = ACTIONS(7281),
    [anon_sym_LPAREN2] = ACTIONS(7281),
    [anon_sym_STAR] = ACTIONS(7281),
    [anon_sym_CARET] = ACTIONS(7281),
    [anon_sym_GT] = ACTIONS(7281),
    [anon_sym_LT] = ACTIONS(7281),
    [anon_sym___extension__] = ACTIONS(7279),
    [anon_sym___attribute__] = ACTIONS(7279),
    [anon_sym___attribute] = ACTIONS(7279),
    [anon_sym_noreturn] = ACTIONS(7279),
    [anon_sym_LBRACK] = ACTIONS(7281),
    [anon_sym___based] = ACTIONS(7279),
    [anon_sym_LBRACE] = ACTIONS(7281),
    [anon_sym_signed] = ACTIONS(7279),
    [anon_sym_unsigned] = ACTIONS(7279),
    [anon_sym_long] = ACTIONS(7279),
    [anon_sym_short] = ACTIONS(7279),
    [anon_sym_ATautoreleasepool] = ACTIONS(7281),
    [anon_sym_const] = ACTIONS(7279),
    [anon_sym_constexpr] = ACTIONS(7279),
    [anon_sym_volatile] = ACTIONS(7279),
    [anon_sym_restrict] = ACTIONS(7279),
    [anon_sym___restrict__] = ACTIONS(7279),
    [anon_sym__Atomic] = ACTIONS(7279),
    [anon_sym__Noreturn] = ACTIONS(7279),
    [anon_sym__Nonnull] = ACTIONS(7279),
    [anon_sym_nullable] = ACTIONS(7279),
    [anon_sym__Complex] = ACTIONS(7279),
    [anon_sym__Nullable] = ACTIONS(7279),
    [anon_sym__Nullable_result] = ACTIONS(7279),
    [anon_sym__Null_unspecified] = ACTIONS(7279),
    [anon_sym___autoreleasing] = ACTIONS(7279),
    [anon_sym___block] = ACTIONS(7279),
    [anon_sym___bridge] = ACTIONS(7279),
    [anon_sym___bridge_retained] = ACTIONS(7279),
    [anon_sym___bridge_transfer] = ACTIONS(7279),
    [anon_sym___complex] = ACTIONS(7279),
    [anon_sym___const] = ACTIONS(7279),
    [anon_sym___imag] = ACTIONS(7279),
    [anon_sym___kindof] = ACTIONS(7279),
    [anon_sym___nonnull] = ACTIONS(7279),
    [anon_sym___nullable] = ACTIONS(7279),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7279),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7279),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7279),
    [anon_sym___real] = ACTIONS(7279),
    [anon_sym___strong] = ACTIONS(7279),
    [anon_sym___unsafe_unretained] = ACTIONS(7279),
    [anon_sym___unused] = ACTIONS(7279),
    [anon_sym___weak] = ACTIONS(7279),
    [anon_sym_alignas] = ACTIONS(7279),
    [anon_sym__Alignas] = ACTIONS(7279),
    [sym_primitive_type] = ACTIONS(7279),
    [anon_sym_enum] = ACTIONS(7279),
    [anon_sym_COLON] = ACTIONS(7281),
    [anon_sym_struct] = ACTIONS(7279),
    [anon_sym_union] = ACTIONS(7279),
    [anon_sym_in] = ACTIONS(7279),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7279),
    [anon_sym___typeof] = ACTIONS(7279),
    [anon_sym_typeof] = ACTIONS(7279),
    [anon_sym_BOOL] = ACTIONS(7279),
    [anon_sym_IMP] = ACTIONS(7279),
    [anon_sym_SEL] = ACTIONS(7279),
    [anon_sym_Class] = ACTIONS(7279),
    [anon_sym_id] = ACTIONS(7279),
    [anon_sym_out] = ACTIONS(7279),
    [anon_sym_inout] = ACTIONS(7279),
    [anon_sym_bycopy] = ACTIONS(7279),
    [anon_sym_byref] = ACTIONS(7279),
    [anon_sym_oneway] = ACTIONS(7279),
  },
  [STATE(3893)] = {
    [sym__type_definition_type] = STATE(4785),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4023),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3894)] = {
    [sym_identifier] = ACTIONS(7291),
    [anon_sym_COMMA] = ACTIONS(7293),
    [anon_sym_RPAREN] = ACTIONS(7293),
    [anon_sym_LPAREN2] = ACTIONS(7293),
    [anon_sym_STAR] = ACTIONS(7293),
    [anon_sym_CARET] = ACTIONS(7293),
    [anon_sym_GT] = ACTIONS(7293),
    [anon_sym_LT] = ACTIONS(7293),
    [anon_sym___extension__] = ACTIONS(7291),
    [anon_sym___attribute__] = ACTIONS(7291),
    [anon_sym___attribute] = ACTIONS(7291),
    [anon_sym_noreturn] = ACTIONS(7291),
    [anon_sym_LBRACK] = ACTIONS(7293),
    [anon_sym___based] = ACTIONS(7291),
    [anon_sym_LBRACE] = ACTIONS(7293),
    [anon_sym_signed] = ACTIONS(7291),
    [anon_sym_unsigned] = ACTIONS(7291),
    [anon_sym_long] = ACTIONS(7291),
    [anon_sym_short] = ACTIONS(7291),
    [anon_sym_ATautoreleasepool] = ACTIONS(7293),
    [anon_sym_const] = ACTIONS(7291),
    [anon_sym_constexpr] = ACTIONS(7291),
    [anon_sym_volatile] = ACTIONS(7291),
    [anon_sym_restrict] = ACTIONS(7291),
    [anon_sym___restrict__] = ACTIONS(7291),
    [anon_sym__Atomic] = ACTIONS(7291),
    [anon_sym__Noreturn] = ACTIONS(7291),
    [anon_sym__Nonnull] = ACTIONS(7291),
    [anon_sym_nullable] = ACTIONS(7291),
    [anon_sym__Complex] = ACTIONS(7291),
    [anon_sym__Nullable] = ACTIONS(7291),
    [anon_sym__Nullable_result] = ACTIONS(7291),
    [anon_sym__Null_unspecified] = ACTIONS(7291),
    [anon_sym___autoreleasing] = ACTIONS(7291),
    [anon_sym___block] = ACTIONS(7291),
    [anon_sym___bridge] = ACTIONS(7291),
    [anon_sym___bridge_retained] = ACTIONS(7291),
    [anon_sym___bridge_transfer] = ACTIONS(7291),
    [anon_sym___complex] = ACTIONS(7291),
    [anon_sym___const] = ACTIONS(7291),
    [anon_sym___imag] = ACTIONS(7291),
    [anon_sym___kindof] = ACTIONS(7291),
    [anon_sym___nonnull] = ACTIONS(7291),
    [anon_sym___nullable] = ACTIONS(7291),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7291),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7291),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7291),
    [anon_sym___real] = ACTIONS(7291),
    [anon_sym___strong] = ACTIONS(7291),
    [anon_sym___unsafe_unretained] = ACTIONS(7291),
    [anon_sym___unused] = ACTIONS(7291),
    [anon_sym___weak] = ACTIONS(7291),
    [anon_sym_alignas] = ACTIONS(7291),
    [anon_sym__Alignas] = ACTIONS(7291),
    [sym_primitive_type] = ACTIONS(7291),
    [anon_sym_enum] = ACTIONS(7291),
    [anon_sym_COLON] = ACTIONS(7293),
    [anon_sym_struct] = ACTIONS(7291),
    [anon_sym_union] = ACTIONS(7291),
    [anon_sym_in] = ACTIONS(7291),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7291),
    [anon_sym___typeof] = ACTIONS(7291),
    [anon_sym_typeof] = ACTIONS(7291),
    [anon_sym_BOOL] = ACTIONS(7291),
    [anon_sym_IMP] = ACTIONS(7291),
    [anon_sym_SEL] = ACTIONS(7291),
    [anon_sym_Class] = ACTIONS(7291),
    [anon_sym_id] = ACTIONS(7291),
    [anon_sym_out] = ACTIONS(7291),
    [anon_sym_inout] = ACTIONS(7291),
    [anon_sym_bycopy] = ACTIONS(7291),
    [anon_sym_byref] = ACTIONS(7291),
    [anon_sym_oneway] = ACTIONS(7291),
  },
  [STATE(3895)] = {
    [sym__type_definition_type] = STATE(4757),
    [sym_attribute_specifier] = STATE(3897),
    [sym_ms_declspec_modifier] = STATE(4030),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3896)] = {
    [sym__type_definition_type] = STATE(4722),
    [sym_attribute_specifier] = STATE(3898),
    [sym_ms_declspec_modifier] = STATE(4036),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3897)] = {
    [sym__type_definition_type] = STATE(4724),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4036),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3898)] = {
    [sym__type_definition_type] = STATE(4732),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4040),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3899)] = {
    [sym__type_definition_type] = STATE(4737),
    [sym_attribute_specifier] = STATE(3900),
    [sym_ms_declspec_modifier] = STATE(4040),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3900)] = {
    [sym__type_definition_type] = STATE(4739),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4013),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3901)] = {
    [sym_identifier] = ACTIONS(7233),
    [anon_sym_COMMA] = ACTIONS(7235),
    [anon_sym_RPAREN] = ACTIONS(7235),
    [anon_sym_LPAREN2] = ACTIONS(7235),
    [anon_sym_STAR] = ACTIONS(7235),
    [anon_sym_CARET] = ACTIONS(7235),
    [anon_sym_GT] = ACTIONS(7235),
    [anon_sym_LT] = ACTIONS(7235),
    [anon_sym___extension__] = ACTIONS(7233),
    [anon_sym___attribute__] = ACTIONS(7233),
    [anon_sym___attribute] = ACTIONS(7233),
    [anon_sym_noreturn] = ACTIONS(7233),
    [anon_sym_LBRACK] = ACTIONS(7235),
    [anon_sym___based] = ACTIONS(7233),
    [anon_sym_LBRACE] = ACTIONS(7235),
    [anon_sym_signed] = ACTIONS(7233),
    [anon_sym_unsigned] = ACTIONS(7233),
    [anon_sym_long] = ACTIONS(7233),
    [anon_sym_short] = ACTIONS(7233),
    [anon_sym_ATautoreleasepool] = ACTIONS(7235),
    [anon_sym_const] = ACTIONS(7233),
    [anon_sym_constexpr] = ACTIONS(7233),
    [anon_sym_volatile] = ACTIONS(7233),
    [anon_sym_restrict] = ACTIONS(7233),
    [anon_sym___restrict__] = ACTIONS(7233),
    [anon_sym__Atomic] = ACTIONS(7233),
    [anon_sym__Noreturn] = ACTIONS(7233),
    [anon_sym__Nonnull] = ACTIONS(7233),
    [anon_sym_nullable] = ACTIONS(7233),
    [anon_sym__Complex] = ACTIONS(7233),
    [anon_sym__Nullable] = ACTIONS(7233),
    [anon_sym__Nullable_result] = ACTIONS(7233),
    [anon_sym__Null_unspecified] = ACTIONS(7233),
    [anon_sym___autoreleasing] = ACTIONS(7233),
    [anon_sym___block] = ACTIONS(7233),
    [anon_sym___bridge] = ACTIONS(7233),
    [anon_sym___bridge_retained] = ACTIONS(7233),
    [anon_sym___bridge_transfer] = ACTIONS(7233),
    [anon_sym___complex] = ACTIONS(7233),
    [anon_sym___const] = ACTIONS(7233),
    [anon_sym___imag] = ACTIONS(7233),
    [anon_sym___kindof] = ACTIONS(7233),
    [anon_sym___nonnull] = ACTIONS(7233),
    [anon_sym___nullable] = ACTIONS(7233),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7233),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7233),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7233),
    [anon_sym___real] = ACTIONS(7233),
    [anon_sym___strong] = ACTIONS(7233),
    [anon_sym___unsafe_unretained] = ACTIONS(7233),
    [anon_sym___unused] = ACTIONS(7233),
    [anon_sym___weak] = ACTIONS(7233),
    [anon_sym_alignas] = ACTIONS(7233),
    [anon_sym__Alignas] = ACTIONS(7233),
    [sym_primitive_type] = ACTIONS(7233),
    [anon_sym_enum] = ACTIONS(7233),
    [anon_sym_COLON] = ACTIONS(7235),
    [anon_sym_struct] = ACTIONS(7233),
    [anon_sym_union] = ACTIONS(7233),
    [anon_sym_in] = ACTIONS(7233),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7233),
    [anon_sym___typeof] = ACTIONS(7233),
    [anon_sym_typeof] = ACTIONS(7233),
    [anon_sym_BOOL] = ACTIONS(7233),
    [anon_sym_IMP] = ACTIONS(7233),
    [anon_sym_SEL] = ACTIONS(7233),
    [anon_sym_Class] = ACTIONS(7233),
    [anon_sym_id] = ACTIONS(7233),
    [anon_sym_out] = ACTIONS(7233),
    [anon_sym_inout] = ACTIONS(7233),
    [anon_sym_bycopy] = ACTIONS(7233),
    [anon_sym_byref] = ACTIONS(7233),
    [anon_sym_oneway] = ACTIONS(7233),
  },
  [STATE(3902)] = {
    [sym_parameter_list] = STATE(3764),
    [sym_identifier] = ACTIONS(6748),
    [anon_sym_COMMA] = ACTIONS(6750),
    [anon_sym_RPAREN] = ACTIONS(6750),
    [anon_sym_LPAREN2] = ACTIONS(8850),
    [anon_sym_STAR] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_GT] = ACTIONS(6750),
    [anon_sym_LT] = ACTIONS(6750),
    [anon_sym___extension__] = ACTIONS(6748),
    [anon_sym___attribute__] = ACTIONS(6748),
    [anon_sym___attribute] = ACTIONS(6748),
    [anon_sym_noreturn] = ACTIONS(6748),
    [anon_sym_LBRACK] = ACTIONS(8852),
    [anon_sym___based] = ACTIONS(6748),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_signed] = ACTIONS(6748),
    [anon_sym_unsigned] = ACTIONS(6748),
    [anon_sym_long] = ACTIONS(6748),
    [anon_sym_short] = ACTIONS(6748),
    [anon_sym_ATautoreleasepool] = ACTIONS(6750),
    [anon_sym_const] = ACTIONS(6748),
    [anon_sym_constexpr] = ACTIONS(6748),
    [anon_sym_volatile] = ACTIONS(6748),
    [anon_sym_restrict] = ACTIONS(6748),
    [anon_sym___restrict__] = ACTIONS(6748),
    [anon_sym__Atomic] = ACTIONS(6748),
    [anon_sym__Noreturn] = ACTIONS(6748),
    [anon_sym__Nonnull] = ACTIONS(6748),
    [anon_sym_nullable] = ACTIONS(6748),
    [anon_sym__Complex] = ACTIONS(6748),
    [anon_sym__Nullable] = ACTIONS(6748),
    [anon_sym__Nullable_result] = ACTIONS(6748),
    [anon_sym__Null_unspecified] = ACTIONS(6748),
    [anon_sym___autoreleasing] = ACTIONS(6748),
    [anon_sym___block] = ACTIONS(6748),
    [anon_sym___bridge] = ACTIONS(6748),
    [anon_sym___bridge_retained] = ACTIONS(6748),
    [anon_sym___bridge_transfer] = ACTIONS(6748),
    [anon_sym___complex] = ACTIONS(6748),
    [anon_sym___const] = ACTIONS(6748),
    [anon_sym___imag] = ACTIONS(6748),
    [anon_sym___kindof] = ACTIONS(6748),
    [anon_sym___nonnull] = ACTIONS(6748),
    [anon_sym___nullable] = ACTIONS(6748),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6748),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6748),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6748),
    [anon_sym___real] = ACTIONS(6748),
    [anon_sym___strong] = ACTIONS(6748),
    [anon_sym___unsafe_unretained] = ACTIONS(6748),
    [anon_sym___unused] = ACTIONS(6748),
    [anon_sym___weak] = ACTIONS(6748),
    [anon_sym_alignas] = ACTIONS(6748),
    [anon_sym__Alignas] = ACTIONS(6748),
    [sym_primitive_type] = ACTIONS(6748),
    [anon_sym_enum] = ACTIONS(6748),
    [anon_sym_struct] = ACTIONS(6748),
    [anon_sym_union] = ACTIONS(6748),
    [anon_sym_in] = ACTIONS(6748),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6748),
    [anon_sym___typeof] = ACTIONS(6748),
    [anon_sym_typeof] = ACTIONS(6748),
    [anon_sym_BOOL] = ACTIONS(6748),
    [anon_sym_IMP] = ACTIONS(6748),
    [anon_sym_SEL] = ACTIONS(6748),
    [anon_sym_Class] = ACTIONS(6748),
    [anon_sym_id] = ACTIONS(6748),
    [anon_sym_out] = ACTIONS(6748),
    [anon_sym_inout] = ACTIONS(6748),
    [anon_sym_bycopy] = ACTIONS(6748),
    [anon_sym_byref] = ACTIONS(6748),
    [anon_sym_oneway] = ACTIONS(6748),
  },
  [STATE(3903)] = {
    [sym__type_definition_type] = STATE(4727),
    [sym_attribute_specifier] = STATE(3949),
    [sym_ms_declspec_modifier] = STATE(4008),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3904)] = {
    [sym_identifier] = ACTIONS(7273),
    [anon_sym_COMMA] = ACTIONS(7275),
    [anon_sym_RPAREN] = ACTIONS(7275),
    [anon_sym_LPAREN2] = ACTIONS(7275),
    [anon_sym_STAR] = ACTIONS(7275),
    [anon_sym_CARET] = ACTIONS(7275),
    [anon_sym_GT] = ACTIONS(7275),
    [anon_sym_LT] = ACTIONS(7275),
    [anon_sym___extension__] = ACTIONS(7273),
    [anon_sym___attribute__] = ACTIONS(7273),
    [anon_sym___attribute] = ACTIONS(7273),
    [anon_sym_noreturn] = ACTIONS(7273),
    [anon_sym_LBRACK] = ACTIONS(7275),
    [anon_sym___based] = ACTIONS(7273),
    [anon_sym_LBRACE] = ACTIONS(7275),
    [anon_sym_signed] = ACTIONS(7273),
    [anon_sym_unsigned] = ACTIONS(7273),
    [anon_sym_long] = ACTIONS(7273),
    [anon_sym_short] = ACTIONS(7273),
    [anon_sym_ATautoreleasepool] = ACTIONS(7275),
    [anon_sym_const] = ACTIONS(7273),
    [anon_sym_constexpr] = ACTIONS(7273),
    [anon_sym_volatile] = ACTIONS(7273),
    [anon_sym_restrict] = ACTIONS(7273),
    [anon_sym___restrict__] = ACTIONS(7273),
    [anon_sym__Atomic] = ACTIONS(7273),
    [anon_sym__Noreturn] = ACTIONS(7273),
    [anon_sym__Nonnull] = ACTIONS(7273),
    [anon_sym_nullable] = ACTIONS(7273),
    [anon_sym__Complex] = ACTIONS(7273),
    [anon_sym__Nullable] = ACTIONS(7273),
    [anon_sym__Nullable_result] = ACTIONS(7273),
    [anon_sym__Null_unspecified] = ACTIONS(7273),
    [anon_sym___autoreleasing] = ACTIONS(7273),
    [anon_sym___block] = ACTIONS(7273),
    [anon_sym___bridge] = ACTIONS(7273),
    [anon_sym___bridge_retained] = ACTIONS(7273),
    [anon_sym___bridge_transfer] = ACTIONS(7273),
    [anon_sym___complex] = ACTIONS(7273),
    [anon_sym___const] = ACTIONS(7273),
    [anon_sym___imag] = ACTIONS(7273),
    [anon_sym___kindof] = ACTIONS(7273),
    [anon_sym___nonnull] = ACTIONS(7273),
    [anon_sym___nullable] = ACTIONS(7273),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7273),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7273),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7273),
    [anon_sym___real] = ACTIONS(7273),
    [anon_sym___strong] = ACTIONS(7273),
    [anon_sym___unsafe_unretained] = ACTIONS(7273),
    [anon_sym___unused] = ACTIONS(7273),
    [anon_sym___weak] = ACTIONS(7273),
    [anon_sym_alignas] = ACTIONS(7273),
    [anon_sym__Alignas] = ACTIONS(7273),
    [sym_primitive_type] = ACTIONS(7273),
    [anon_sym_enum] = ACTIONS(7273),
    [anon_sym_COLON] = ACTIONS(7275),
    [anon_sym_struct] = ACTIONS(7273),
    [anon_sym_union] = ACTIONS(7273),
    [anon_sym_in] = ACTIONS(7273),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7273),
    [anon_sym___typeof] = ACTIONS(7273),
    [anon_sym_typeof] = ACTIONS(7273),
    [anon_sym_BOOL] = ACTIONS(7273),
    [anon_sym_IMP] = ACTIONS(7273),
    [anon_sym_SEL] = ACTIONS(7273),
    [anon_sym_Class] = ACTIONS(7273),
    [anon_sym_id] = ACTIONS(7273),
    [anon_sym_out] = ACTIONS(7273),
    [anon_sym_inout] = ACTIONS(7273),
    [anon_sym_bycopy] = ACTIONS(7273),
    [anon_sym_byref] = ACTIONS(7273),
    [anon_sym_oneway] = ACTIONS(7273),
  },
  [STATE(3905)] = {
    [sym__type_definition_type] = STATE(4742),
    [sym_attribute_specifier] = STATE(3907),
    [sym_ms_declspec_modifier] = STATE(4021),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3906)] = {
    [sym__type_definition_type] = STATE(4743),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4021),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3907)] = {
    [sym__type_definition_type] = STATE(4745),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4020),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3908)] = {
    [sym__type_definition_type] = STATE(4748),
    [sym_attribute_specifier] = STATE(3910),
    [sym_ms_declspec_modifier] = STATE(4020),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3909)] = {
    [sym_attribute_specifier] = STATE(3876),
    [sym_identifier] = ACTIONS(7760),
    [anon_sym_COMMA] = ACTIONS(7762),
    [anon_sym_RPAREN] = ACTIONS(7762),
    [anon_sym_LPAREN2] = ACTIONS(7762),
    [anon_sym_STAR] = ACTIONS(7762),
    [anon_sym_CARET] = ACTIONS(7762),
    [anon_sym_GT] = ACTIONS(7762),
    [anon_sym_LT] = ACTIONS(7762),
    [anon_sym___extension__] = ACTIONS(7760),
    [anon_sym___attribute__] = ACTIONS(8896),
    [anon_sym___attribute] = ACTIONS(8896),
    [anon_sym_noreturn] = ACTIONS(7760),
    [anon_sym_LBRACK] = ACTIONS(7762),
    [anon_sym___based] = ACTIONS(7760),
    [anon_sym_LBRACE] = ACTIONS(7762),
    [anon_sym_signed] = ACTIONS(7760),
    [anon_sym_unsigned] = ACTIONS(7760),
    [anon_sym_long] = ACTIONS(7760),
    [anon_sym_short] = ACTIONS(7760),
    [anon_sym_ATautoreleasepool] = ACTIONS(7762),
    [anon_sym_const] = ACTIONS(7760),
    [anon_sym_constexpr] = ACTIONS(7760),
    [anon_sym_volatile] = ACTIONS(7760),
    [anon_sym_restrict] = ACTIONS(7760),
    [anon_sym___restrict__] = ACTIONS(7760),
    [anon_sym__Atomic] = ACTIONS(7760),
    [anon_sym__Noreturn] = ACTIONS(7760),
    [anon_sym__Nonnull] = ACTIONS(7760),
    [anon_sym_nullable] = ACTIONS(7760),
    [anon_sym__Complex] = ACTIONS(7760),
    [anon_sym__Nullable] = ACTIONS(7760),
    [anon_sym__Nullable_result] = ACTIONS(7760),
    [anon_sym__Null_unspecified] = ACTIONS(7760),
    [anon_sym___autoreleasing] = ACTIONS(7760),
    [anon_sym___block] = ACTIONS(7760),
    [anon_sym___bridge] = ACTIONS(7760),
    [anon_sym___bridge_retained] = ACTIONS(7760),
    [anon_sym___bridge_transfer] = ACTIONS(7760),
    [anon_sym___complex] = ACTIONS(7760),
    [anon_sym___const] = ACTIONS(7760),
    [anon_sym___imag] = ACTIONS(7760),
    [anon_sym___kindof] = ACTIONS(7760),
    [anon_sym___nonnull] = ACTIONS(7760),
    [anon_sym___nullable] = ACTIONS(7760),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7760),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7760),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7760),
    [anon_sym___real] = ACTIONS(7760),
    [anon_sym___strong] = ACTIONS(7760),
    [anon_sym___unsafe_unretained] = ACTIONS(7760),
    [anon_sym___unused] = ACTIONS(7760),
    [anon_sym___weak] = ACTIONS(7760),
    [anon_sym_alignas] = ACTIONS(7760),
    [anon_sym__Alignas] = ACTIONS(7760),
    [sym_primitive_type] = ACTIONS(7760),
    [anon_sym_enum] = ACTIONS(7760),
    [anon_sym_struct] = ACTIONS(7760),
    [anon_sym_union] = ACTIONS(7760),
    [anon_sym_in] = ACTIONS(7760),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7760),
    [anon_sym___typeof] = ACTIONS(7760),
    [anon_sym_typeof] = ACTIONS(7760),
    [anon_sym_BOOL] = ACTIONS(7760),
    [anon_sym_IMP] = ACTIONS(7760),
    [anon_sym_SEL] = ACTIONS(7760),
    [anon_sym_Class] = ACTIONS(7760),
    [anon_sym_id] = ACTIONS(7760),
    [anon_sym_out] = ACTIONS(7760),
    [anon_sym_inout] = ACTIONS(7760),
    [anon_sym_bycopy] = ACTIONS(7760),
    [anon_sym_byref] = ACTIONS(7760),
    [anon_sym_oneway] = ACTIONS(7760),
  },
  [STATE(3910)] = {
    [sym__type_definition_type] = STATE(4752),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4002),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3911)] = {
    [sym__type_definition_type] = STATE(4759),
    [sym_attribute_specifier] = STATE(3913),
    [sym_ms_declspec_modifier] = STATE(4007),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3912)] = {
    [sym__type_definition_type] = STATE(4760),
    [sym_attribute_specifier] = STATE(3914),
    [sym_ms_declspec_modifier] = STATE(4011),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3913)] = {
    [sym__type_definition_type] = STATE(4761),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4011),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3914)] = {
    [sym__type_definition_type] = STATE(4763),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4000),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3915)] = {
    [sym__type_definition_type] = STATE(4765),
    [sym_attribute_specifier] = STATE(3916),
    [sym_ms_declspec_modifier] = STATE(4000),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3916)] = {
    [sym__type_definition_type] = STATE(4768),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4018),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3917)] = {
    [sym_attribute_specifier] = STATE(3879),
    [sym_identifier] = ACTIONS(7775),
    [anon_sym_COMMA] = ACTIONS(7777),
    [anon_sym_RPAREN] = ACTIONS(7777),
    [anon_sym_LPAREN2] = ACTIONS(7777),
    [anon_sym_STAR] = ACTIONS(7777),
    [anon_sym_CARET] = ACTIONS(7777),
    [anon_sym_GT] = ACTIONS(7777),
    [anon_sym_LT] = ACTIONS(7777),
    [anon_sym___extension__] = ACTIONS(7775),
    [anon_sym___attribute__] = ACTIONS(8899),
    [anon_sym___attribute] = ACTIONS(8899),
    [anon_sym_noreturn] = ACTIONS(7775),
    [anon_sym_LBRACK] = ACTIONS(7777),
    [anon_sym___based] = ACTIONS(7775),
    [anon_sym_LBRACE] = ACTIONS(7777),
    [anon_sym_signed] = ACTIONS(7775),
    [anon_sym_unsigned] = ACTIONS(7775),
    [anon_sym_long] = ACTIONS(7775),
    [anon_sym_short] = ACTIONS(7775),
    [anon_sym_ATautoreleasepool] = ACTIONS(7777),
    [anon_sym_const] = ACTIONS(7775),
    [anon_sym_constexpr] = ACTIONS(7775),
    [anon_sym_volatile] = ACTIONS(7775),
    [anon_sym_restrict] = ACTIONS(7775),
    [anon_sym___restrict__] = ACTIONS(7775),
    [anon_sym__Atomic] = ACTIONS(7775),
    [anon_sym__Noreturn] = ACTIONS(7775),
    [anon_sym__Nonnull] = ACTIONS(7775),
    [anon_sym_nullable] = ACTIONS(7775),
    [anon_sym__Complex] = ACTIONS(7775),
    [anon_sym__Nullable] = ACTIONS(7775),
    [anon_sym__Nullable_result] = ACTIONS(7775),
    [anon_sym__Null_unspecified] = ACTIONS(7775),
    [anon_sym___autoreleasing] = ACTIONS(7775),
    [anon_sym___block] = ACTIONS(7775),
    [anon_sym___bridge] = ACTIONS(7775),
    [anon_sym___bridge_retained] = ACTIONS(7775),
    [anon_sym___bridge_transfer] = ACTIONS(7775),
    [anon_sym___complex] = ACTIONS(7775),
    [anon_sym___const] = ACTIONS(7775),
    [anon_sym___imag] = ACTIONS(7775),
    [anon_sym___kindof] = ACTIONS(7775),
    [anon_sym___nonnull] = ACTIONS(7775),
    [anon_sym___nullable] = ACTIONS(7775),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7775),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7775),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7775),
    [anon_sym___real] = ACTIONS(7775),
    [anon_sym___strong] = ACTIONS(7775),
    [anon_sym___unsafe_unretained] = ACTIONS(7775),
    [anon_sym___unused] = ACTIONS(7775),
    [anon_sym___weak] = ACTIONS(7775),
    [anon_sym_alignas] = ACTIONS(7775),
    [anon_sym__Alignas] = ACTIONS(7775),
    [sym_primitive_type] = ACTIONS(7775),
    [anon_sym_enum] = ACTIONS(7775),
    [anon_sym_struct] = ACTIONS(7775),
    [anon_sym_union] = ACTIONS(7775),
    [anon_sym_in] = ACTIONS(7775),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7775),
    [anon_sym___typeof] = ACTIONS(7775),
    [anon_sym_typeof] = ACTIONS(7775),
    [anon_sym_BOOL] = ACTIONS(7775),
    [anon_sym_IMP] = ACTIONS(7775),
    [anon_sym_SEL] = ACTIONS(7775),
    [anon_sym_Class] = ACTIONS(7775),
    [anon_sym_id] = ACTIONS(7775),
    [anon_sym_out] = ACTIONS(7775),
    [anon_sym_inout] = ACTIONS(7775),
    [anon_sym_bycopy] = ACTIONS(7775),
    [anon_sym_byref] = ACTIONS(7775),
    [anon_sym_oneway] = ACTIONS(7775),
  },
  [STATE(3918)] = {
    [sym__type_definition_type] = STATE(4773),
    [sym_attribute_specifier] = STATE(3920),
    [sym_ms_declspec_modifier] = STATE(4025),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3919)] = {
    [sym__type_definition_type] = STATE(4774),
    [sym_attribute_specifier] = STATE(3921),
    [sym_ms_declspec_modifier] = STATE(4027),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3920)] = {
    [sym__type_definition_type] = STATE(4776),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4027),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3921)] = {
    [sym__type_definition_type] = STATE(4782),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4029),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3922)] = {
    [sym__type_definition_type] = STATE(4688),
    [sym_attribute_specifier] = STATE(3923),
    [sym_ms_declspec_modifier] = STATE(4029),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3923)] = {
    [sym__type_definition_type] = STATE(4690),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4032),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3924)] = {
    [sym__type_definition_type] = STATE(4692),
    [sym_attribute_specifier] = STATE(3926),
    [sym_ms_declspec_modifier] = STATE(4037),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3925)] = {
    [sym__type_definition_type] = STATE(4693),
    [sym_attribute_specifier] = STATE(3927),
    [sym_ms_declspec_modifier] = STATE(4039),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3926)] = {
    [sym__type_definition_type] = STATE(4694),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4039),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3927)] = {
    [sym__type_definition_type] = STATE(4696),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4041),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3928)] = {
    [sym__type_definition_type] = STATE(4786),
    [sym_attribute_specifier] = STATE(3929),
    [sym_ms_declspec_modifier] = STATE(4041),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3929)] = {
    [sym__type_definition_type] = STATE(4700),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4044),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3930)] = {
    [sym__type_definition_type] = STATE(4702),
    [sym_attribute_specifier] = STATE(3932),
    [sym_ms_declspec_modifier] = STATE(4006),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3931)] = {
    [sym__type_definition_type] = STATE(4703),
    [sym_attribute_specifier] = STATE(3933),
    [sym_ms_declspec_modifier] = STATE(4031),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3932)] = {
    [sym__type_definition_type] = STATE(4704),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4031),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3933)] = {
    [sym__type_definition_type] = STATE(4706),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4042),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3934)] = {
    [sym__type_definition_type] = STATE(4708),
    [sym_attribute_specifier] = STATE(3935),
    [sym_ms_declspec_modifier] = STATE(4042),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3935)] = {
    [sym__type_definition_type] = STATE(4710),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4005),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3936)] = {
    [sym__type_definition_type] = STATE(4712),
    [sym_attribute_specifier] = STATE(3938),
    [sym_ms_declspec_modifier] = STATE(4017),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3937)] = {
    [sym__type_definition_type] = STATE(4713),
    [sym_attribute_specifier] = STATE(3939),
    [sym_ms_declspec_modifier] = STATE(4045),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3938)] = {
    [sym__type_definition_type] = STATE(4714),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4045),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3939)] = {
    [sym__type_definition_type] = STATE(4716),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4022),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3940)] = {
    [sym__type_definition_type] = STATE(4718),
    [sym_attribute_specifier] = STATE(3941),
    [sym_ms_declspec_modifier] = STATE(4022),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3941)] = {
    [sym__type_definition_type] = STATE(4720),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4033),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3942)] = {
    [sym_identifier] = ACTIONS(7299),
    [anon_sym_COMMA] = ACTIONS(7301),
    [anon_sym_RPAREN] = ACTIONS(7301),
    [anon_sym_LPAREN2] = ACTIONS(7301),
    [anon_sym_STAR] = ACTIONS(7301),
    [anon_sym_CARET] = ACTIONS(7301),
    [anon_sym_GT] = ACTIONS(7301),
    [anon_sym_LT] = ACTIONS(7301),
    [anon_sym___extension__] = ACTIONS(7299),
    [anon_sym___attribute__] = ACTIONS(7299),
    [anon_sym___attribute] = ACTIONS(7299),
    [anon_sym_noreturn] = ACTIONS(7299),
    [anon_sym_LBRACK] = ACTIONS(7301),
    [anon_sym___based] = ACTIONS(7299),
    [anon_sym_LBRACE] = ACTIONS(7301),
    [anon_sym_signed] = ACTIONS(7299),
    [anon_sym_unsigned] = ACTIONS(7299),
    [anon_sym_long] = ACTIONS(7299),
    [anon_sym_short] = ACTIONS(7299),
    [anon_sym_ATautoreleasepool] = ACTIONS(7301),
    [anon_sym_const] = ACTIONS(7299),
    [anon_sym_constexpr] = ACTIONS(7299),
    [anon_sym_volatile] = ACTIONS(7299),
    [anon_sym_restrict] = ACTIONS(7299),
    [anon_sym___restrict__] = ACTIONS(7299),
    [anon_sym__Atomic] = ACTIONS(7299),
    [anon_sym__Noreturn] = ACTIONS(7299),
    [anon_sym__Nonnull] = ACTIONS(7299),
    [anon_sym_nullable] = ACTIONS(7299),
    [anon_sym__Complex] = ACTIONS(7299),
    [anon_sym__Nullable] = ACTIONS(7299),
    [anon_sym__Nullable_result] = ACTIONS(7299),
    [anon_sym__Null_unspecified] = ACTIONS(7299),
    [anon_sym___autoreleasing] = ACTIONS(7299),
    [anon_sym___block] = ACTIONS(7299),
    [anon_sym___bridge] = ACTIONS(7299),
    [anon_sym___bridge_retained] = ACTIONS(7299),
    [anon_sym___bridge_transfer] = ACTIONS(7299),
    [anon_sym___complex] = ACTIONS(7299),
    [anon_sym___const] = ACTIONS(7299),
    [anon_sym___imag] = ACTIONS(7299),
    [anon_sym___kindof] = ACTIONS(7299),
    [anon_sym___nonnull] = ACTIONS(7299),
    [anon_sym___nullable] = ACTIONS(7299),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7299),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7299),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7299),
    [anon_sym___real] = ACTIONS(7299),
    [anon_sym___strong] = ACTIONS(7299),
    [anon_sym___unsafe_unretained] = ACTIONS(7299),
    [anon_sym___unused] = ACTIONS(7299),
    [anon_sym___weak] = ACTIONS(7299),
    [anon_sym_alignas] = ACTIONS(7299),
    [anon_sym__Alignas] = ACTIONS(7299),
    [sym_primitive_type] = ACTIONS(7299),
    [anon_sym_enum] = ACTIONS(7299),
    [anon_sym_COLON] = ACTIONS(7301),
    [anon_sym_struct] = ACTIONS(7299),
    [anon_sym_union] = ACTIONS(7299),
    [anon_sym_in] = ACTIONS(7299),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7299),
    [anon_sym___typeof] = ACTIONS(7299),
    [anon_sym_typeof] = ACTIONS(7299),
    [anon_sym_BOOL] = ACTIONS(7299),
    [anon_sym_IMP] = ACTIONS(7299),
    [anon_sym_SEL] = ACTIONS(7299),
    [anon_sym_Class] = ACTIONS(7299),
    [anon_sym_id] = ACTIONS(7299),
    [anon_sym_out] = ACTIONS(7299),
    [anon_sym_inout] = ACTIONS(7299),
    [anon_sym_bycopy] = ACTIONS(7299),
    [anon_sym_byref] = ACTIONS(7299),
    [anon_sym_oneway] = ACTIONS(7299),
  },
  [STATE(3943)] = {
    [sym_parameter_list] = STATE(3764),
    [sym_identifier] = ACTIONS(8902),
    [anon_sym_COMMA] = ACTIONS(8101),
    [anon_sym_RPAREN] = ACTIONS(8101),
    [anon_sym_LPAREN2] = ACTIONS(8850),
    [anon_sym_STAR] = ACTIONS(8101),
    [anon_sym_CARET] = ACTIONS(8101),
    [anon_sym_GT] = ACTIONS(8101),
    [anon_sym_LT] = ACTIONS(8101),
    [anon_sym___extension__] = ACTIONS(8902),
    [anon_sym___attribute__] = ACTIONS(8902),
    [anon_sym___attribute] = ACTIONS(8902),
    [anon_sym_noreturn] = ACTIONS(8902),
    [anon_sym_LBRACK] = ACTIONS(8852),
    [anon_sym___based] = ACTIONS(8902),
    [anon_sym_LBRACE] = ACTIONS(8101),
    [anon_sym_signed] = ACTIONS(8902),
    [anon_sym_unsigned] = ACTIONS(8902),
    [anon_sym_long] = ACTIONS(8902),
    [anon_sym_short] = ACTIONS(8902),
    [anon_sym_ATautoreleasepool] = ACTIONS(8101),
    [anon_sym_const] = ACTIONS(8902),
    [anon_sym_constexpr] = ACTIONS(8902),
    [anon_sym_volatile] = ACTIONS(8902),
    [anon_sym_restrict] = ACTIONS(8902),
    [anon_sym___restrict__] = ACTIONS(8902),
    [anon_sym__Atomic] = ACTIONS(8902),
    [anon_sym__Noreturn] = ACTIONS(8902),
    [anon_sym__Nonnull] = ACTIONS(8902),
    [anon_sym_nullable] = ACTIONS(8902),
    [anon_sym__Complex] = ACTIONS(8902),
    [anon_sym__Nullable] = ACTIONS(8902),
    [anon_sym__Nullable_result] = ACTIONS(8902),
    [anon_sym__Null_unspecified] = ACTIONS(8902),
    [anon_sym___autoreleasing] = ACTIONS(8902),
    [anon_sym___block] = ACTIONS(8902),
    [anon_sym___bridge] = ACTIONS(8902),
    [anon_sym___bridge_retained] = ACTIONS(8902),
    [anon_sym___bridge_transfer] = ACTIONS(8902),
    [anon_sym___complex] = ACTIONS(8902),
    [anon_sym___const] = ACTIONS(8902),
    [anon_sym___imag] = ACTIONS(8902),
    [anon_sym___kindof] = ACTIONS(8902),
    [anon_sym___nonnull] = ACTIONS(8902),
    [anon_sym___nullable] = ACTIONS(8902),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8902),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8902),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8902),
    [anon_sym___real] = ACTIONS(8902),
    [anon_sym___strong] = ACTIONS(8902),
    [anon_sym___unsafe_unretained] = ACTIONS(8902),
    [anon_sym___unused] = ACTIONS(8902),
    [anon_sym___weak] = ACTIONS(8902),
    [anon_sym_alignas] = ACTIONS(8902),
    [anon_sym__Alignas] = ACTIONS(8902),
    [sym_primitive_type] = ACTIONS(8902),
    [anon_sym_enum] = ACTIONS(8902),
    [anon_sym_struct] = ACTIONS(8902),
    [anon_sym_union] = ACTIONS(8902),
    [anon_sym_in] = ACTIONS(8902),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8902),
    [anon_sym___typeof] = ACTIONS(8902),
    [anon_sym_typeof] = ACTIONS(8902),
    [anon_sym_BOOL] = ACTIONS(8902),
    [anon_sym_IMP] = ACTIONS(8902),
    [anon_sym_SEL] = ACTIONS(8902),
    [anon_sym_Class] = ACTIONS(8902),
    [anon_sym_id] = ACTIONS(8902),
    [anon_sym_out] = ACTIONS(8902),
    [anon_sym_inout] = ACTIONS(8902),
    [anon_sym_bycopy] = ACTIONS(8902),
    [anon_sym_byref] = ACTIONS(8902),
    [anon_sym_oneway] = ACTIONS(8902),
  },
  [STATE(3944)] = {
    [sym_identifier] = ACTIONS(8904),
    [anon_sym_COMMA] = ACTIONS(8906),
    [anon_sym_RPAREN] = ACTIONS(8906),
    [anon_sym_LPAREN2] = ACTIONS(8906),
    [anon_sym_STAR] = ACTIONS(8906),
    [anon_sym_CARET] = ACTIONS(8906),
    [anon_sym_GT] = ACTIONS(8906),
    [anon_sym_LT] = ACTIONS(8906),
    [anon_sym___extension__] = ACTIONS(8904),
    [anon_sym___attribute__] = ACTIONS(8904),
    [anon_sym___attribute] = ACTIONS(8904),
    [anon_sym_noreturn] = ACTIONS(8904),
    [anon_sym_LBRACK] = ACTIONS(8906),
    [anon_sym___based] = ACTIONS(8904),
    [anon_sym_LBRACE] = ACTIONS(8906),
    [anon_sym_signed] = ACTIONS(8904),
    [anon_sym_unsigned] = ACTIONS(8904),
    [anon_sym_long] = ACTIONS(8904),
    [anon_sym_short] = ACTIONS(8904),
    [anon_sym_ATautoreleasepool] = ACTIONS(8906),
    [anon_sym_const] = ACTIONS(8904),
    [anon_sym_constexpr] = ACTIONS(8904),
    [anon_sym_volatile] = ACTIONS(8904),
    [anon_sym_restrict] = ACTIONS(8904),
    [anon_sym___restrict__] = ACTIONS(8904),
    [anon_sym__Atomic] = ACTIONS(8904),
    [anon_sym__Noreturn] = ACTIONS(8904),
    [anon_sym__Nonnull] = ACTIONS(8904),
    [anon_sym_nullable] = ACTIONS(8904),
    [anon_sym__Complex] = ACTIONS(8904),
    [anon_sym__Nullable] = ACTIONS(8904),
    [anon_sym__Nullable_result] = ACTIONS(8904),
    [anon_sym__Null_unspecified] = ACTIONS(8904),
    [anon_sym___autoreleasing] = ACTIONS(8904),
    [anon_sym___block] = ACTIONS(8904),
    [anon_sym___bridge] = ACTIONS(8904),
    [anon_sym___bridge_retained] = ACTIONS(8904),
    [anon_sym___bridge_transfer] = ACTIONS(8904),
    [anon_sym___complex] = ACTIONS(8904),
    [anon_sym___const] = ACTIONS(8904),
    [anon_sym___imag] = ACTIONS(8904),
    [anon_sym___kindof] = ACTIONS(8904),
    [anon_sym___nonnull] = ACTIONS(8904),
    [anon_sym___nullable] = ACTIONS(8904),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8904),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8904),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8904),
    [anon_sym___real] = ACTIONS(8904),
    [anon_sym___strong] = ACTIONS(8904),
    [anon_sym___unsafe_unretained] = ACTIONS(8904),
    [anon_sym___unused] = ACTIONS(8904),
    [anon_sym___weak] = ACTIONS(8904),
    [anon_sym_alignas] = ACTIONS(8904),
    [anon_sym__Alignas] = ACTIONS(8904),
    [sym_primitive_type] = ACTIONS(8904),
    [anon_sym_enum] = ACTIONS(8904),
    [anon_sym_COLON] = ACTIONS(8906),
    [anon_sym_struct] = ACTIONS(8904),
    [anon_sym_union] = ACTIONS(8904),
    [anon_sym_in] = ACTIONS(8904),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8904),
    [anon_sym___typeof] = ACTIONS(8904),
    [anon_sym_typeof] = ACTIONS(8904),
    [anon_sym_BOOL] = ACTIONS(8904),
    [anon_sym_IMP] = ACTIONS(8904),
    [anon_sym_SEL] = ACTIONS(8904),
    [anon_sym_Class] = ACTIONS(8904),
    [anon_sym_id] = ACTIONS(8904),
    [anon_sym_out] = ACTIONS(8904),
    [anon_sym_inout] = ACTIONS(8904),
    [anon_sym_bycopy] = ACTIONS(8904),
    [anon_sym_byref] = ACTIONS(8904),
    [anon_sym_oneway] = ACTIONS(8904),
  },
  [STATE(3945)] = {
    [sym_identifier] = ACTIONS(4926),
    [anon_sym_COMMA] = ACTIONS(4928),
    [anon_sym_RPAREN] = ACTIONS(4928),
    [anon_sym_LPAREN2] = ACTIONS(4928),
    [anon_sym_STAR] = ACTIONS(4928),
    [anon_sym_CARET] = ACTIONS(4928),
    [anon_sym_GT] = ACTIONS(4928),
    [anon_sym_LT] = ACTIONS(4928),
    [anon_sym___extension__] = ACTIONS(4926),
    [anon_sym___attribute__] = ACTIONS(4926),
    [anon_sym___attribute] = ACTIONS(4926),
    [anon_sym_noreturn] = ACTIONS(4926),
    [anon_sym_LBRACK] = ACTIONS(4928),
    [anon_sym___based] = ACTIONS(4926),
    [anon_sym_LBRACE] = ACTIONS(4928),
    [anon_sym_signed] = ACTIONS(4926),
    [anon_sym_unsigned] = ACTIONS(4926),
    [anon_sym_long] = ACTIONS(4926),
    [anon_sym_short] = ACTIONS(4926),
    [anon_sym_ATautoreleasepool] = ACTIONS(4928),
    [anon_sym_const] = ACTIONS(4926),
    [anon_sym_constexpr] = ACTIONS(4926),
    [anon_sym_volatile] = ACTIONS(4926),
    [anon_sym_restrict] = ACTIONS(4926),
    [anon_sym___restrict__] = ACTIONS(4926),
    [anon_sym__Atomic] = ACTIONS(4926),
    [anon_sym__Noreturn] = ACTIONS(4926),
    [anon_sym__Nonnull] = ACTIONS(4926),
    [anon_sym_nullable] = ACTIONS(4926),
    [anon_sym__Complex] = ACTIONS(4926),
    [anon_sym__Nullable] = ACTIONS(4926),
    [anon_sym__Nullable_result] = ACTIONS(4926),
    [anon_sym__Null_unspecified] = ACTIONS(4926),
    [anon_sym___autoreleasing] = ACTIONS(4926),
    [anon_sym___block] = ACTIONS(4926),
    [anon_sym___bridge] = ACTIONS(4926),
    [anon_sym___bridge_retained] = ACTIONS(4926),
    [anon_sym___bridge_transfer] = ACTIONS(4926),
    [anon_sym___complex] = ACTIONS(4926),
    [anon_sym___const] = ACTIONS(4926),
    [anon_sym___imag] = ACTIONS(4926),
    [anon_sym___kindof] = ACTIONS(4926),
    [anon_sym___nonnull] = ACTIONS(4926),
    [anon_sym___nullable] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4926),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4926),
    [anon_sym___real] = ACTIONS(4926),
    [anon_sym___strong] = ACTIONS(4926),
    [anon_sym___unsafe_unretained] = ACTIONS(4926),
    [anon_sym___unused] = ACTIONS(4926),
    [anon_sym___weak] = ACTIONS(4926),
    [anon_sym_alignas] = ACTIONS(4926),
    [anon_sym__Alignas] = ACTIONS(4926),
    [sym_primitive_type] = ACTIONS(4926),
    [anon_sym_enum] = ACTIONS(4926),
    [anon_sym_COLON] = ACTIONS(4928),
    [anon_sym_struct] = ACTIONS(4926),
    [anon_sym_union] = ACTIONS(4926),
    [anon_sym_in] = ACTIONS(4926),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4926),
    [anon_sym___typeof] = ACTIONS(4926),
    [anon_sym_typeof] = ACTIONS(4926),
    [anon_sym_BOOL] = ACTIONS(4926),
    [anon_sym_IMP] = ACTIONS(4926),
    [anon_sym_SEL] = ACTIONS(4926),
    [anon_sym_Class] = ACTIONS(4926),
    [anon_sym_id] = ACTIONS(4926),
    [anon_sym_out] = ACTIONS(4926),
    [anon_sym_inout] = ACTIONS(4926),
    [anon_sym_bycopy] = ACTIONS(4926),
    [anon_sym_byref] = ACTIONS(4926),
    [anon_sym_oneway] = ACTIONS(4926),
  },
  [STATE(3946)] = {
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_RPAREN] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7147),
    [anon_sym_STAR] = ACTIONS(7147),
    [anon_sym_CARET] = ACTIONS(7147),
    [anon_sym_GT] = ACTIONS(7147),
    [anon_sym_LT] = ACTIONS(7147),
    [anon_sym___extension__] = ACTIONS(7145),
    [anon_sym___attribute__] = ACTIONS(7145),
    [anon_sym___attribute] = ACTIONS(7145),
    [anon_sym_noreturn] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7147),
    [anon_sym___based] = ACTIONS(7145),
    [anon_sym_LBRACE] = ACTIONS(7147),
    [anon_sym_signed] = ACTIONS(7145),
    [anon_sym_unsigned] = ACTIONS(7145),
    [anon_sym_long] = ACTIONS(7145),
    [anon_sym_short] = ACTIONS(7145),
    [anon_sym_ATautoreleasepool] = ACTIONS(7147),
    [anon_sym_const] = ACTIONS(7145),
    [anon_sym_constexpr] = ACTIONS(7145),
    [anon_sym_volatile] = ACTIONS(7145),
    [anon_sym_restrict] = ACTIONS(7145),
    [anon_sym___restrict__] = ACTIONS(7145),
    [anon_sym__Atomic] = ACTIONS(7145),
    [anon_sym__Noreturn] = ACTIONS(7145),
    [anon_sym__Nonnull] = ACTIONS(7145),
    [anon_sym_nullable] = ACTIONS(7145),
    [anon_sym__Complex] = ACTIONS(7145),
    [anon_sym__Nullable] = ACTIONS(7145),
    [anon_sym__Nullable_result] = ACTIONS(7145),
    [anon_sym__Null_unspecified] = ACTIONS(7145),
    [anon_sym___autoreleasing] = ACTIONS(7145),
    [anon_sym___block] = ACTIONS(7145),
    [anon_sym___bridge] = ACTIONS(7145),
    [anon_sym___bridge_retained] = ACTIONS(7145),
    [anon_sym___bridge_transfer] = ACTIONS(7145),
    [anon_sym___complex] = ACTIONS(7145),
    [anon_sym___const] = ACTIONS(7145),
    [anon_sym___imag] = ACTIONS(7145),
    [anon_sym___kindof] = ACTIONS(7145),
    [anon_sym___nonnull] = ACTIONS(7145),
    [anon_sym___nullable] = ACTIONS(7145),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7145),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7145),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7145),
    [anon_sym___real] = ACTIONS(7145),
    [anon_sym___strong] = ACTIONS(7145),
    [anon_sym___unsafe_unretained] = ACTIONS(7145),
    [anon_sym___unused] = ACTIONS(7145),
    [anon_sym___weak] = ACTIONS(7145),
    [anon_sym_alignas] = ACTIONS(7145),
    [anon_sym__Alignas] = ACTIONS(7145),
    [sym_primitive_type] = ACTIONS(7145),
    [anon_sym_enum] = ACTIONS(7145),
    [anon_sym_COLON] = ACTIONS(7147),
    [anon_sym_struct] = ACTIONS(7145),
    [anon_sym_union] = ACTIONS(7145),
    [anon_sym_in] = ACTIONS(7145),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7145),
    [anon_sym___typeof] = ACTIONS(7145),
    [anon_sym_typeof] = ACTIONS(7145),
    [anon_sym_BOOL] = ACTIONS(7145),
    [anon_sym_IMP] = ACTIONS(7145),
    [anon_sym_SEL] = ACTIONS(7145),
    [anon_sym_Class] = ACTIONS(7145),
    [anon_sym_id] = ACTIONS(7145),
    [anon_sym_out] = ACTIONS(7145),
    [anon_sym_inout] = ACTIONS(7145),
    [anon_sym_bycopy] = ACTIONS(7145),
    [anon_sym_byref] = ACTIONS(7145),
    [anon_sym_oneway] = ACTIONS(7145),
  },
  [STATE(3947)] = {
    [sym_identifier] = ACTIONS(7295),
    [anon_sym_COMMA] = ACTIONS(7297),
    [anon_sym_RPAREN] = ACTIONS(7297),
    [anon_sym_LPAREN2] = ACTIONS(7297),
    [anon_sym_STAR] = ACTIONS(7297),
    [anon_sym_CARET] = ACTIONS(7297),
    [anon_sym_GT] = ACTIONS(7297),
    [anon_sym_LT] = ACTIONS(7297),
    [anon_sym___extension__] = ACTIONS(7295),
    [anon_sym___attribute__] = ACTIONS(7295),
    [anon_sym___attribute] = ACTIONS(7295),
    [anon_sym_noreturn] = ACTIONS(7295),
    [anon_sym_LBRACK] = ACTIONS(7297),
    [anon_sym___based] = ACTIONS(7295),
    [anon_sym_LBRACE] = ACTIONS(7297),
    [anon_sym_signed] = ACTIONS(7295),
    [anon_sym_unsigned] = ACTIONS(7295),
    [anon_sym_long] = ACTIONS(7295),
    [anon_sym_short] = ACTIONS(7295),
    [anon_sym_ATautoreleasepool] = ACTIONS(7297),
    [anon_sym_const] = ACTIONS(7295),
    [anon_sym_constexpr] = ACTIONS(7295),
    [anon_sym_volatile] = ACTIONS(7295),
    [anon_sym_restrict] = ACTIONS(7295),
    [anon_sym___restrict__] = ACTIONS(7295),
    [anon_sym__Atomic] = ACTIONS(7295),
    [anon_sym__Noreturn] = ACTIONS(7295),
    [anon_sym__Nonnull] = ACTIONS(7295),
    [anon_sym_nullable] = ACTIONS(7295),
    [anon_sym__Complex] = ACTIONS(7295),
    [anon_sym__Nullable] = ACTIONS(7295),
    [anon_sym__Nullable_result] = ACTIONS(7295),
    [anon_sym__Null_unspecified] = ACTIONS(7295),
    [anon_sym___autoreleasing] = ACTIONS(7295),
    [anon_sym___block] = ACTIONS(7295),
    [anon_sym___bridge] = ACTIONS(7295),
    [anon_sym___bridge_retained] = ACTIONS(7295),
    [anon_sym___bridge_transfer] = ACTIONS(7295),
    [anon_sym___complex] = ACTIONS(7295),
    [anon_sym___const] = ACTIONS(7295),
    [anon_sym___imag] = ACTIONS(7295),
    [anon_sym___kindof] = ACTIONS(7295),
    [anon_sym___nonnull] = ACTIONS(7295),
    [anon_sym___nullable] = ACTIONS(7295),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7295),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7295),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7295),
    [anon_sym___real] = ACTIONS(7295),
    [anon_sym___strong] = ACTIONS(7295),
    [anon_sym___unsafe_unretained] = ACTIONS(7295),
    [anon_sym___unused] = ACTIONS(7295),
    [anon_sym___weak] = ACTIONS(7295),
    [anon_sym_alignas] = ACTIONS(7295),
    [anon_sym__Alignas] = ACTIONS(7295),
    [sym_primitive_type] = ACTIONS(7295),
    [anon_sym_enum] = ACTIONS(7295),
    [anon_sym_COLON] = ACTIONS(7297),
    [anon_sym_struct] = ACTIONS(7295),
    [anon_sym_union] = ACTIONS(7295),
    [anon_sym_in] = ACTIONS(7295),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7295),
    [anon_sym___typeof] = ACTIONS(7295),
    [anon_sym_typeof] = ACTIONS(7295),
    [anon_sym_BOOL] = ACTIONS(7295),
    [anon_sym_IMP] = ACTIONS(7295),
    [anon_sym_SEL] = ACTIONS(7295),
    [anon_sym_Class] = ACTIONS(7295),
    [anon_sym_id] = ACTIONS(7295),
    [anon_sym_out] = ACTIONS(7295),
    [anon_sym_inout] = ACTIONS(7295),
    [anon_sym_bycopy] = ACTIONS(7295),
    [anon_sym_byref] = ACTIONS(7295),
    [anon_sym_oneway] = ACTIONS(7295),
  },
  [STATE(3948)] = {
    [sym__type_definition_type] = STATE(4734),
    [sym_attribute_specifier] = STATE(3825),
    [sym_ms_declspec_modifier] = STATE(4014),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3949)] = {
    [sym__type_definition_type] = STATE(4735),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4014),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3950)] = {
    [sym_identifier] = ACTIONS(7425),
    [anon_sym_COMMA] = ACTIONS(7427),
    [anon_sym_RPAREN] = ACTIONS(7427),
    [anon_sym_LPAREN2] = ACTIONS(7427),
    [anon_sym_STAR] = ACTIONS(7427),
    [anon_sym_CARET] = ACTIONS(7427),
    [anon_sym_GT] = ACTIONS(7427),
    [anon_sym_LT] = ACTIONS(7427),
    [anon_sym___extension__] = ACTIONS(7425),
    [anon_sym___attribute__] = ACTIONS(7425),
    [anon_sym___attribute] = ACTIONS(7425),
    [anon_sym_noreturn] = ACTIONS(7425),
    [anon_sym_LBRACK] = ACTIONS(7427),
    [anon_sym___based] = ACTIONS(7425),
    [anon_sym_LBRACE] = ACTIONS(7427),
    [anon_sym_signed] = ACTIONS(7425),
    [anon_sym_unsigned] = ACTIONS(7425),
    [anon_sym_long] = ACTIONS(7425),
    [anon_sym_short] = ACTIONS(7425),
    [anon_sym_ATautoreleasepool] = ACTIONS(7427),
    [anon_sym_const] = ACTIONS(7425),
    [anon_sym_constexpr] = ACTIONS(7425),
    [anon_sym_volatile] = ACTIONS(7425),
    [anon_sym_restrict] = ACTIONS(7425),
    [anon_sym___restrict__] = ACTIONS(7425),
    [anon_sym__Atomic] = ACTIONS(7425),
    [anon_sym__Noreturn] = ACTIONS(7425),
    [anon_sym__Nonnull] = ACTIONS(7425),
    [anon_sym_nullable] = ACTIONS(7425),
    [anon_sym__Complex] = ACTIONS(7425),
    [anon_sym__Nullable] = ACTIONS(7425),
    [anon_sym__Nullable_result] = ACTIONS(7425),
    [anon_sym__Null_unspecified] = ACTIONS(7425),
    [anon_sym___autoreleasing] = ACTIONS(7425),
    [anon_sym___block] = ACTIONS(7425),
    [anon_sym___bridge] = ACTIONS(7425),
    [anon_sym___bridge_retained] = ACTIONS(7425),
    [anon_sym___bridge_transfer] = ACTIONS(7425),
    [anon_sym___complex] = ACTIONS(7425),
    [anon_sym___const] = ACTIONS(7425),
    [anon_sym___imag] = ACTIONS(7425),
    [anon_sym___kindof] = ACTIONS(7425),
    [anon_sym___nonnull] = ACTIONS(7425),
    [anon_sym___nullable] = ACTIONS(7425),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7425),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7425),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7425),
    [anon_sym___real] = ACTIONS(7425),
    [anon_sym___strong] = ACTIONS(7425),
    [anon_sym___unsafe_unretained] = ACTIONS(7425),
    [anon_sym___unused] = ACTIONS(7425),
    [anon_sym___weak] = ACTIONS(7425),
    [anon_sym_alignas] = ACTIONS(7425),
    [anon_sym__Alignas] = ACTIONS(7425),
    [sym_primitive_type] = ACTIONS(7425),
    [anon_sym_enum] = ACTIONS(7425),
    [anon_sym_COLON] = ACTIONS(7427),
    [anon_sym_struct] = ACTIONS(7425),
    [anon_sym_union] = ACTIONS(7425),
    [anon_sym_in] = ACTIONS(7425),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7425),
    [anon_sym___typeof] = ACTIONS(7425),
    [anon_sym_typeof] = ACTIONS(7425),
    [anon_sym_BOOL] = ACTIONS(7425),
    [anon_sym_IMP] = ACTIONS(7425),
    [anon_sym_SEL] = ACTIONS(7425),
    [anon_sym_Class] = ACTIONS(7425),
    [anon_sym_id] = ACTIONS(7425),
    [anon_sym_out] = ACTIONS(7425),
    [anon_sym_inout] = ACTIONS(7425),
    [anon_sym_bycopy] = ACTIONS(7425),
    [anon_sym_byref] = ACTIONS(7425),
    [anon_sym_oneway] = ACTIONS(7425),
  },
  [STATE(3951)] = {
    [sym_attribute_specifier] = STATE(3957),
    [sym_enumerator_list] = STATE(3909),
    [sym_identifier] = ACTIONS(7561),
    [anon_sym_COMMA] = ACTIONS(7563),
    [anon_sym_RPAREN] = ACTIONS(7563),
    [anon_sym_LPAREN2] = ACTIONS(7563),
    [anon_sym_STAR] = ACTIONS(7563),
    [anon_sym_CARET] = ACTIONS(7563),
    [anon_sym_GT] = ACTIONS(7563),
    [anon_sym_LT] = ACTIONS(7563),
    [anon_sym___extension__] = ACTIONS(7561),
    [anon_sym___attribute__] = ACTIONS(8908),
    [anon_sym___attribute] = ACTIONS(8908),
    [anon_sym_noreturn] = ACTIONS(7561),
    [anon_sym_LBRACK] = ACTIONS(7563),
    [anon_sym___based] = ACTIONS(7561),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7561),
    [anon_sym_unsigned] = ACTIONS(7561),
    [anon_sym_long] = ACTIONS(7561),
    [anon_sym_short] = ACTIONS(7561),
    [anon_sym_const] = ACTIONS(7561),
    [anon_sym_constexpr] = ACTIONS(7561),
    [anon_sym_volatile] = ACTIONS(7561),
    [anon_sym_restrict] = ACTIONS(7561),
    [anon_sym___restrict__] = ACTIONS(7561),
    [anon_sym__Atomic] = ACTIONS(7561),
    [anon_sym__Noreturn] = ACTIONS(7561),
    [anon_sym__Nonnull] = ACTIONS(7561),
    [anon_sym_nullable] = ACTIONS(7561),
    [anon_sym__Complex] = ACTIONS(7561),
    [anon_sym__Nullable] = ACTIONS(7561),
    [anon_sym__Nullable_result] = ACTIONS(7561),
    [anon_sym__Null_unspecified] = ACTIONS(7561),
    [anon_sym___autoreleasing] = ACTIONS(7561),
    [anon_sym___block] = ACTIONS(7561),
    [anon_sym___bridge] = ACTIONS(7561),
    [anon_sym___bridge_retained] = ACTIONS(7561),
    [anon_sym___bridge_transfer] = ACTIONS(7561),
    [anon_sym___complex] = ACTIONS(7561),
    [anon_sym___const] = ACTIONS(7561),
    [anon_sym___imag] = ACTIONS(7561),
    [anon_sym___kindof] = ACTIONS(7561),
    [anon_sym___nonnull] = ACTIONS(7561),
    [anon_sym___nullable] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7561),
    [anon_sym___real] = ACTIONS(7561),
    [anon_sym___strong] = ACTIONS(7561),
    [anon_sym___unsafe_unretained] = ACTIONS(7561),
    [anon_sym___unused] = ACTIONS(7561),
    [anon_sym___weak] = ACTIONS(7561),
    [anon_sym_alignas] = ACTIONS(7561),
    [anon_sym__Alignas] = ACTIONS(7561),
    [sym_primitive_type] = ACTIONS(7561),
    [anon_sym_enum] = ACTIONS(7561),
    [anon_sym_struct] = ACTIONS(7561),
    [anon_sym_union] = ACTIONS(7561),
    [anon_sym_in] = ACTIONS(7561),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7561),
    [anon_sym___typeof] = ACTIONS(7561),
    [anon_sym_typeof] = ACTIONS(7561),
    [anon_sym_BOOL] = ACTIONS(7561),
    [anon_sym_IMP] = ACTIONS(7561),
    [anon_sym_SEL] = ACTIONS(7561),
    [anon_sym_Class] = ACTIONS(7561),
    [anon_sym_id] = ACTIONS(7561),
    [anon_sym_out] = ACTIONS(7561),
    [anon_sym_inout] = ACTIONS(7561),
    [anon_sym_bycopy] = ACTIONS(7561),
    [anon_sym_byref] = ACTIONS(7561),
    [anon_sym_oneway] = ACTIONS(7561),
  },
  [STATE(3952)] = {
    [sym_attribute_specifier] = STATE(3959),
    [sym_enumerator_list] = STATE(3917),
    [sym_identifier] = ACTIONS(7568),
    [anon_sym_COMMA] = ACTIONS(7570),
    [anon_sym_RPAREN] = ACTIONS(7570),
    [anon_sym_LPAREN2] = ACTIONS(7570),
    [anon_sym_STAR] = ACTIONS(7570),
    [anon_sym_CARET] = ACTIONS(7570),
    [anon_sym_GT] = ACTIONS(7570),
    [anon_sym_LT] = ACTIONS(7570),
    [anon_sym___extension__] = ACTIONS(7568),
    [anon_sym___attribute__] = ACTIONS(8911),
    [anon_sym___attribute] = ACTIONS(8911),
    [anon_sym_noreturn] = ACTIONS(7568),
    [anon_sym_LBRACK] = ACTIONS(7570),
    [anon_sym___based] = ACTIONS(7568),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7568),
    [anon_sym_unsigned] = ACTIONS(7568),
    [anon_sym_long] = ACTIONS(7568),
    [anon_sym_short] = ACTIONS(7568),
    [anon_sym_const] = ACTIONS(7568),
    [anon_sym_constexpr] = ACTIONS(7568),
    [anon_sym_volatile] = ACTIONS(7568),
    [anon_sym_restrict] = ACTIONS(7568),
    [anon_sym___restrict__] = ACTIONS(7568),
    [anon_sym__Atomic] = ACTIONS(7568),
    [anon_sym__Noreturn] = ACTIONS(7568),
    [anon_sym__Nonnull] = ACTIONS(7568),
    [anon_sym_nullable] = ACTIONS(7568),
    [anon_sym__Complex] = ACTIONS(7568),
    [anon_sym__Nullable] = ACTIONS(7568),
    [anon_sym__Nullable_result] = ACTIONS(7568),
    [anon_sym__Null_unspecified] = ACTIONS(7568),
    [anon_sym___autoreleasing] = ACTIONS(7568),
    [anon_sym___block] = ACTIONS(7568),
    [anon_sym___bridge] = ACTIONS(7568),
    [anon_sym___bridge_retained] = ACTIONS(7568),
    [anon_sym___bridge_transfer] = ACTIONS(7568),
    [anon_sym___complex] = ACTIONS(7568),
    [anon_sym___const] = ACTIONS(7568),
    [anon_sym___imag] = ACTIONS(7568),
    [anon_sym___kindof] = ACTIONS(7568),
    [anon_sym___nonnull] = ACTIONS(7568),
    [anon_sym___nullable] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7568),
    [anon_sym___real] = ACTIONS(7568),
    [anon_sym___strong] = ACTIONS(7568),
    [anon_sym___unsafe_unretained] = ACTIONS(7568),
    [anon_sym___unused] = ACTIONS(7568),
    [anon_sym___weak] = ACTIONS(7568),
    [anon_sym_alignas] = ACTIONS(7568),
    [anon_sym__Alignas] = ACTIONS(7568),
    [sym_primitive_type] = ACTIONS(7568),
    [anon_sym_enum] = ACTIONS(7568),
    [anon_sym_struct] = ACTIONS(7568),
    [anon_sym_union] = ACTIONS(7568),
    [anon_sym_in] = ACTIONS(7568),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7568),
    [anon_sym___typeof] = ACTIONS(7568),
    [anon_sym_typeof] = ACTIONS(7568),
    [anon_sym_BOOL] = ACTIONS(7568),
    [anon_sym_IMP] = ACTIONS(7568),
    [anon_sym_SEL] = ACTIONS(7568),
    [anon_sym_Class] = ACTIONS(7568),
    [anon_sym_id] = ACTIONS(7568),
    [anon_sym_out] = ACTIONS(7568),
    [anon_sym_inout] = ACTIONS(7568),
    [anon_sym_bycopy] = ACTIONS(7568),
    [anon_sym_byref] = ACTIONS(7568),
    [anon_sym_oneway] = ACTIONS(7568),
  },
  [STATE(3953)] = {
    [sym_attribute_specifier] = STATE(3901),
    [sym_identifier] = ACTIONS(7782),
    [anon_sym_COMMA] = ACTIONS(7784),
    [anon_sym_RPAREN] = ACTIONS(7784),
    [anon_sym_LPAREN2] = ACTIONS(7784),
    [anon_sym_STAR] = ACTIONS(7784),
    [anon_sym_CARET] = ACTIONS(7784),
    [anon_sym_GT] = ACTIONS(7784),
    [anon_sym_LT] = ACTIONS(7784),
    [anon_sym___extension__] = ACTIONS(7782),
    [anon_sym___attribute__] = ACTIONS(8914),
    [anon_sym___attribute] = ACTIONS(8914),
    [anon_sym_noreturn] = ACTIONS(7782),
    [anon_sym_LBRACK] = ACTIONS(7784),
    [anon_sym___based] = ACTIONS(7782),
    [anon_sym_LBRACE] = ACTIONS(7784),
    [anon_sym_signed] = ACTIONS(7782),
    [anon_sym_unsigned] = ACTIONS(7782),
    [anon_sym_long] = ACTIONS(7782),
    [anon_sym_short] = ACTIONS(7782),
    [anon_sym_ATautoreleasepool] = ACTIONS(7784),
    [anon_sym_const] = ACTIONS(7782),
    [anon_sym_constexpr] = ACTIONS(7782),
    [anon_sym_volatile] = ACTIONS(7782),
    [anon_sym_restrict] = ACTIONS(7782),
    [anon_sym___restrict__] = ACTIONS(7782),
    [anon_sym__Atomic] = ACTIONS(7782),
    [anon_sym__Noreturn] = ACTIONS(7782),
    [anon_sym__Nonnull] = ACTIONS(7782),
    [anon_sym_nullable] = ACTIONS(7782),
    [anon_sym__Complex] = ACTIONS(7782),
    [anon_sym__Nullable] = ACTIONS(7782),
    [anon_sym__Nullable_result] = ACTIONS(7782),
    [anon_sym__Null_unspecified] = ACTIONS(7782),
    [anon_sym___autoreleasing] = ACTIONS(7782),
    [anon_sym___block] = ACTIONS(7782),
    [anon_sym___bridge] = ACTIONS(7782),
    [anon_sym___bridge_retained] = ACTIONS(7782),
    [anon_sym___bridge_transfer] = ACTIONS(7782),
    [anon_sym___complex] = ACTIONS(7782),
    [anon_sym___const] = ACTIONS(7782),
    [anon_sym___imag] = ACTIONS(7782),
    [anon_sym___kindof] = ACTIONS(7782),
    [anon_sym___nonnull] = ACTIONS(7782),
    [anon_sym___nullable] = ACTIONS(7782),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7782),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7782),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7782),
    [anon_sym___real] = ACTIONS(7782),
    [anon_sym___strong] = ACTIONS(7782),
    [anon_sym___unsafe_unretained] = ACTIONS(7782),
    [anon_sym___unused] = ACTIONS(7782),
    [anon_sym___weak] = ACTIONS(7782),
    [anon_sym_alignas] = ACTIONS(7782),
    [anon_sym__Alignas] = ACTIONS(7782),
    [sym_primitive_type] = ACTIONS(7782),
    [anon_sym_enum] = ACTIONS(7782),
    [anon_sym_struct] = ACTIONS(7782),
    [anon_sym_union] = ACTIONS(7782),
    [anon_sym_in] = ACTIONS(7782),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7782),
    [anon_sym___typeof] = ACTIONS(7782),
    [anon_sym_typeof] = ACTIONS(7782),
    [anon_sym_BOOL] = ACTIONS(7782),
    [anon_sym_IMP] = ACTIONS(7782),
    [anon_sym_SEL] = ACTIONS(7782),
    [anon_sym_Class] = ACTIONS(7782),
    [anon_sym_id] = ACTIONS(7782),
    [anon_sym_out] = ACTIONS(7782),
    [anon_sym_inout] = ACTIONS(7782),
    [anon_sym_bycopy] = ACTIONS(7782),
    [anon_sym_byref] = ACTIONS(7782),
    [anon_sym_oneway] = ACTIONS(7782),
  },
  [STATE(3954)] = {
    [sym__type_definition_type] = STATE(4730),
    [sym_attribute_specifier] = STATE(3962),
    [sym_ms_declspec_modifier] = STATE(4004),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3955)] = {
    [sym_identifier] = ACTIONS(7337),
    [anon_sym_COMMA] = ACTIONS(7339),
    [anon_sym_RPAREN] = ACTIONS(7339),
    [anon_sym_LPAREN2] = ACTIONS(7339),
    [anon_sym_STAR] = ACTIONS(7339),
    [anon_sym_CARET] = ACTIONS(7339),
    [anon_sym_GT] = ACTIONS(7339),
    [anon_sym_LT] = ACTIONS(7339),
    [anon_sym___extension__] = ACTIONS(7337),
    [anon_sym___attribute__] = ACTIONS(7337),
    [anon_sym___attribute] = ACTIONS(7337),
    [anon_sym_noreturn] = ACTIONS(7337),
    [anon_sym_LBRACK] = ACTIONS(7339),
    [anon_sym___based] = ACTIONS(7337),
    [anon_sym_LBRACE] = ACTIONS(7339),
    [anon_sym_signed] = ACTIONS(7337),
    [anon_sym_unsigned] = ACTIONS(7337),
    [anon_sym_long] = ACTIONS(7337),
    [anon_sym_short] = ACTIONS(7337),
    [anon_sym_ATautoreleasepool] = ACTIONS(7339),
    [anon_sym_const] = ACTIONS(7337),
    [anon_sym_constexpr] = ACTIONS(7337),
    [anon_sym_volatile] = ACTIONS(7337),
    [anon_sym_restrict] = ACTIONS(7337),
    [anon_sym___restrict__] = ACTIONS(7337),
    [anon_sym__Atomic] = ACTIONS(7337),
    [anon_sym__Noreturn] = ACTIONS(7337),
    [anon_sym__Nonnull] = ACTIONS(7337),
    [anon_sym_nullable] = ACTIONS(7337),
    [anon_sym__Complex] = ACTIONS(7337),
    [anon_sym__Nullable] = ACTIONS(7337),
    [anon_sym__Nullable_result] = ACTIONS(7337),
    [anon_sym__Null_unspecified] = ACTIONS(7337),
    [anon_sym___autoreleasing] = ACTIONS(7337),
    [anon_sym___block] = ACTIONS(7337),
    [anon_sym___bridge] = ACTIONS(7337),
    [anon_sym___bridge_retained] = ACTIONS(7337),
    [anon_sym___bridge_transfer] = ACTIONS(7337),
    [anon_sym___complex] = ACTIONS(7337),
    [anon_sym___const] = ACTIONS(7337),
    [anon_sym___imag] = ACTIONS(7337),
    [anon_sym___kindof] = ACTIONS(7337),
    [anon_sym___nonnull] = ACTIONS(7337),
    [anon_sym___nullable] = ACTIONS(7337),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7337),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7337),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7337),
    [anon_sym___real] = ACTIONS(7337),
    [anon_sym___strong] = ACTIONS(7337),
    [anon_sym___unsafe_unretained] = ACTIONS(7337),
    [anon_sym___unused] = ACTIONS(7337),
    [anon_sym___weak] = ACTIONS(7337),
    [anon_sym_alignas] = ACTIONS(7337),
    [anon_sym__Alignas] = ACTIONS(7337),
    [sym_primitive_type] = ACTIONS(7337),
    [anon_sym_enum] = ACTIONS(7337),
    [anon_sym_COLON] = ACTIONS(7339),
    [anon_sym_struct] = ACTIONS(7337),
    [anon_sym_union] = ACTIONS(7337),
    [anon_sym_in] = ACTIONS(7337),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7337),
    [anon_sym___typeof] = ACTIONS(7337),
    [anon_sym_typeof] = ACTIONS(7337),
    [anon_sym_BOOL] = ACTIONS(7337),
    [anon_sym_IMP] = ACTIONS(7337),
    [anon_sym_SEL] = ACTIONS(7337),
    [anon_sym_Class] = ACTIONS(7337),
    [anon_sym_id] = ACTIONS(7337),
    [anon_sym_out] = ACTIONS(7337),
    [anon_sym_inout] = ACTIONS(7337),
    [anon_sym_bycopy] = ACTIONS(7337),
    [anon_sym_byref] = ACTIONS(7337),
    [anon_sym_oneway] = ACTIONS(7337),
  },
  [STATE(3956)] = {
    [sym_identifier] = ACTIONS(7341),
    [anon_sym_COMMA] = ACTIONS(7343),
    [anon_sym_RPAREN] = ACTIONS(7343),
    [anon_sym_LPAREN2] = ACTIONS(7343),
    [anon_sym_STAR] = ACTIONS(7343),
    [anon_sym_CARET] = ACTIONS(7343),
    [anon_sym_GT] = ACTIONS(7343),
    [anon_sym_LT] = ACTIONS(7343),
    [anon_sym___extension__] = ACTIONS(7341),
    [anon_sym___attribute__] = ACTIONS(7341),
    [anon_sym___attribute] = ACTIONS(7341),
    [anon_sym_noreturn] = ACTIONS(7341),
    [anon_sym_LBRACK] = ACTIONS(7343),
    [anon_sym___based] = ACTIONS(7341),
    [anon_sym_LBRACE] = ACTIONS(7343),
    [anon_sym_signed] = ACTIONS(7341),
    [anon_sym_unsigned] = ACTIONS(7341),
    [anon_sym_long] = ACTIONS(7341),
    [anon_sym_short] = ACTIONS(7341),
    [anon_sym_ATautoreleasepool] = ACTIONS(7343),
    [anon_sym_const] = ACTIONS(7341),
    [anon_sym_constexpr] = ACTIONS(7341),
    [anon_sym_volatile] = ACTIONS(7341),
    [anon_sym_restrict] = ACTIONS(7341),
    [anon_sym___restrict__] = ACTIONS(7341),
    [anon_sym__Atomic] = ACTIONS(7341),
    [anon_sym__Noreturn] = ACTIONS(7341),
    [anon_sym__Nonnull] = ACTIONS(7341),
    [anon_sym_nullable] = ACTIONS(7341),
    [anon_sym__Complex] = ACTIONS(7341),
    [anon_sym__Nullable] = ACTIONS(7341),
    [anon_sym__Nullable_result] = ACTIONS(7341),
    [anon_sym__Null_unspecified] = ACTIONS(7341),
    [anon_sym___autoreleasing] = ACTIONS(7341),
    [anon_sym___block] = ACTIONS(7341),
    [anon_sym___bridge] = ACTIONS(7341),
    [anon_sym___bridge_retained] = ACTIONS(7341),
    [anon_sym___bridge_transfer] = ACTIONS(7341),
    [anon_sym___complex] = ACTIONS(7341),
    [anon_sym___const] = ACTIONS(7341),
    [anon_sym___imag] = ACTIONS(7341),
    [anon_sym___kindof] = ACTIONS(7341),
    [anon_sym___nonnull] = ACTIONS(7341),
    [anon_sym___nullable] = ACTIONS(7341),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7341),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7341),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7341),
    [anon_sym___real] = ACTIONS(7341),
    [anon_sym___strong] = ACTIONS(7341),
    [anon_sym___unsafe_unretained] = ACTIONS(7341),
    [anon_sym___unused] = ACTIONS(7341),
    [anon_sym___weak] = ACTIONS(7341),
    [anon_sym_alignas] = ACTIONS(7341),
    [anon_sym__Alignas] = ACTIONS(7341),
    [sym_primitive_type] = ACTIONS(7341),
    [anon_sym_enum] = ACTIONS(7341),
    [anon_sym_COLON] = ACTIONS(7343),
    [anon_sym_struct] = ACTIONS(7341),
    [anon_sym_union] = ACTIONS(7341),
    [anon_sym_in] = ACTIONS(7341),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7341),
    [anon_sym___typeof] = ACTIONS(7341),
    [anon_sym_typeof] = ACTIONS(7341),
    [anon_sym_BOOL] = ACTIONS(7341),
    [anon_sym_IMP] = ACTIONS(7341),
    [anon_sym_SEL] = ACTIONS(7341),
    [anon_sym_Class] = ACTIONS(7341),
    [anon_sym_id] = ACTIONS(7341),
    [anon_sym_out] = ACTIONS(7341),
    [anon_sym_inout] = ACTIONS(7341),
    [anon_sym_bycopy] = ACTIONS(7341),
    [anon_sym_byref] = ACTIONS(7341),
    [anon_sym_oneway] = ACTIONS(7341),
  },
  [STATE(3957)] = {
    [sym_identifier] = ACTIONS(7345),
    [anon_sym_COMMA] = ACTIONS(7347),
    [anon_sym_RPAREN] = ACTIONS(7347),
    [anon_sym_LPAREN2] = ACTIONS(7347),
    [anon_sym_STAR] = ACTIONS(7347),
    [anon_sym_CARET] = ACTIONS(7347),
    [anon_sym_GT] = ACTIONS(7347),
    [anon_sym_LT] = ACTIONS(7347),
    [anon_sym___extension__] = ACTIONS(7345),
    [anon_sym___attribute__] = ACTIONS(7345),
    [anon_sym___attribute] = ACTIONS(7345),
    [anon_sym_noreturn] = ACTIONS(7345),
    [anon_sym_LBRACK] = ACTIONS(7347),
    [anon_sym___based] = ACTIONS(7345),
    [anon_sym_LBRACE] = ACTIONS(7347),
    [anon_sym_signed] = ACTIONS(7345),
    [anon_sym_unsigned] = ACTIONS(7345),
    [anon_sym_long] = ACTIONS(7345),
    [anon_sym_short] = ACTIONS(7345),
    [anon_sym_ATautoreleasepool] = ACTIONS(7347),
    [anon_sym_const] = ACTIONS(7345),
    [anon_sym_constexpr] = ACTIONS(7345),
    [anon_sym_volatile] = ACTIONS(7345),
    [anon_sym_restrict] = ACTIONS(7345),
    [anon_sym___restrict__] = ACTIONS(7345),
    [anon_sym__Atomic] = ACTIONS(7345),
    [anon_sym__Noreturn] = ACTIONS(7345),
    [anon_sym__Nonnull] = ACTIONS(7345),
    [anon_sym_nullable] = ACTIONS(7345),
    [anon_sym__Complex] = ACTIONS(7345),
    [anon_sym__Nullable] = ACTIONS(7345),
    [anon_sym__Nullable_result] = ACTIONS(7345),
    [anon_sym__Null_unspecified] = ACTIONS(7345),
    [anon_sym___autoreleasing] = ACTIONS(7345),
    [anon_sym___block] = ACTIONS(7345),
    [anon_sym___bridge] = ACTIONS(7345),
    [anon_sym___bridge_retained] = ACTIONS(7345),
    [anon_sym___bridge_transfer] = ACTIONS(7345),
    [anon_sym___complex] = ACTIONS(7345),
    [anon_sym___const] = ACTIONS(7345),
    [anon_sym___imag] = ACTIONS(7345),
    [anon_sym___kindof] = ACTIONS(7345),
    [anon_sym___nonnull] = ACTIONS(7345),
    [anon_sym___nullable] = ACTIONS(7345),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7345),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7345),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7345),
    [anon_sym___real] = ACTIONS(7345),
    [anon_sym___strong] = ACTIONS(7345),
    [anon_sym___unsafe_unretained] = ACTIONS(7345),
    [anon_sym___unused] = ACTIONS(7345),
    [anon_sym___weak] = ACTIONS(7345),
    [anon_sym_alignas] = ACTIONS(7345),
    [anon_sym__Alignas] = ACTIONS(7345),
    [sym_primitive_type] = ACTIONS(7345),
    [anon_sym_enum] = ACTIONS(7345),
    [anon_sym_COLON] = ACTIONS(7347),
    [anon_sym_struct] = ACTIONS(7345),
    [anon_sym_union] = ACTIONS(7345),
    [anon_sym_in] = ACTIONS(7345),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7345),
    [anon_sym___typeof] = ACTIONS(7345),
    [anon_sym_typeof] = ACTIONS(7345),
    [anon_sym_BOOL] = ACTIONS(7345),
    [anon_sym_IMP] = ACTIONS(7345),
    [anon_sym_SEL] = ACTIONS(7345),
    [anon_sym_Class] = ACTIONS(7345),
    [anon_sym_id] = ACTIONS(7345),
    [anon_sym_out] = ACTIONS(7345),
    [anon_sym_inout] = ACTIONS(7345),
    [anon_sym_bycopy] = ACTIONS(7345),
    [anon_sym_byref] = ACTIONS(7345),
    [anon_sym_oneway] = ACTIONS(7345),
  },
  [STATE(3958)] = {
    [sym_attribute_specifier] = STATE(3840),
    [sym_identifier] = ACTIONS(7806),
    [anon_sym_COMMA] = ACTIONS(7808),
    [anon_sym_RPAREN] = ACTIONS(7808),
    [anon_sym_LPAREN2] = ACTIONS(7808),
    [anon_sym_STAR] = ACTIONS(7808),
    [anon_sym_CARET] = ACTIONS(7808),
    [anon_sym_GT] = ACTIONS(7808),
    [anon_sym_LT] = ACTIONS(7808),
    [anon_sym___extension__] = ACTIONS(7806),
    [anon_sym___attribute__] = ACTIONS(8917),
    [anon_sym___attribute] = ACTIONS(8917),
    [anon_sym_noreturn] = ACTIONS(7806),
    [anon_sym_LBRACK] = ACTIONS(7808),
    [anon_sym___based] = ACTIONS(7806),
    [anon_sym_LBRACE] = ACTIONS(7808),
    [anon_sym_signed] = ACTIONS(7806),
    [anon_sym_unsigned] = ACTIONS(7806),
    [anon_sym_long] = ACTIONS(7806),
    [anon_sym_short] = ACTIONS(7806),
    [anon_sym_ATautoreleasepool] = ACTIONS(7808),
    [anon_sym_const] = ACTIONS(7806),
    [anon_sym_constexpr] = ACTIONS(7806),
    [anon_sym_volatile] = ACTIONS(7806),
    [anon_sym_restrict] = ACTIONS(7806),
    [anon_sym___restrict__] = ACTIONS(7806),
    [anon_sym__Atomic] = ACTIONS(7806),
    [anon_sym__Noreturn] = ACTIONS(7806),
    [anon_sym__Nonnull] = ACTIONS(7806),
    [anon_sym_nullable] = ACTIONS(7806),
    [anon_sym__Complex] = ACTIONS(7806),
    [anon_sym__Nullable] = ACTIONS(7806),
    [anon_sym__Nullable_result] = ACTIONS(7806),
    [anon_sym__Null_unspecified] = ACTIONS(7806),
    [anon_sym___autoreleasing] = ACTIONS(7806),
    [anon_sym___block] = ACTIONS(7806),
    [anon_sym___bridge] = ACTIONS(7806),
    [anon_sym___bridge_retained] = ACTIONS(7806),
    [anon_sym___bridge_transfer] = ACTIONS(7806),
    [anon_sym___complex] = ACTIONS(7806),
    [anon_sym___const] = ACTIONS(7806),
    [anon_sym___imag] = ACTIONS(7806),
    [anon_sym___kindof] = ACTIONS(7806),
    [anon_sym___nonnull] = ACTIONS(7806),
    [anon_sym___nullable] = ACTIONS(7806),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7806),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7806),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7806),
    [anon_sym___real] = ACTIONS(7806),
    [anon_sym___strong] = ACTIONS(7806),
    [anon_sym___unsafe_unretained] = ACTIONS(7806),
    [anon_sym___unused] = ACTIONS(7806),
    [anon_sym___weak] = ACTIONS(7806),
    [anon_sym_alignas] = ACTIONS(7806),
    [anon_sym__Alignas] = ACTIONS(7806),
    [sym_primitive_type] = ACTIONS(7806),
    [anon_sym_enum] = ACTIONS(7806),
    [anon_sym_struct] = ACTIONS(7806),
    [anon_sym_union] = ACTIONS(7806),
    [anon_sym_in] = ACTIONS(7806),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7806),
    [anon_sym___typeof] = ACTIONS(7806),
    [anon_sym_typeof] = ACTIONS(7806),
    [anon_sym_BOOL] = ACTIONS(7806),
    [anon_sym_IMP] = ACTIONS(7806),
    [anon_sym_SEL] = ACTIONS(7806),
    [anon_sym_Class] = ACTIONS(7806),
    [anon_sym_id] = ACTIONS(7806),
    [anon_sym_out] = ACTIONS(7806),
    [anon_sym_inout] = ACTIONS(7806),
    [anon_sym_bycopy] = ACTIONS(7806),
    [anon_sym_byref] = ACTIONS(7806),
    [anon_sym_oneway] = ACTIONS(7806),
  },
  [STATE(3959)] = {
    [sym_identifier] = ACTIONS(7357),
    [anon_sym_COMMA] = ACTIONS(7359),
    [anon_sym_RPAREN] = ACTIONS(7359),
    [anon_sym_LPAREN2] = ACTIONS(7359),
    [anon_sym_STAR] = ACTIONS(7359),
    [anon_sym_CARET] = ACTIONS(7359),
    [anon_sym_GT] = ACTIONS(7359),
    [anon_sym_LT] = ACTIONS(7359),
    [anon_sym___extension__] = ACTIONS(7357),
    [anon_sym___attribute__] = ACTIONS(7357),
    [anon_sym___attribute] = ACTIONS(7357),
    [anon_sym_noreturn] = ACTIONS(7357),
    [anon_sym_LBRACK] = ACTIONS(7359),
    [anon_sym___based] = ACTIONS(7357),
    [anon_sym_LBRACE] = ACTIONS(7359),
    [anon_sym_signed] = ACTIONS(7357),
    [anon_sym_unsigned] = ACTIONS(7357),
    [anon_sym_long] = ACTIONS(7357),
    [anon_sym_short] = ACTIONS(7357),
    [anon_sym_ATautoreleasepool] = ACTIONS(7359),
    [anon_sym_const] = ACTIONS(7357),
    [anon_sym_constexpr] = ACTIONS(7357),
    [anon_sym_volatile] = ACTIONS(7357),
    [anon_sym_restrict] = ACTIONS(7357),
    [anon_sym___restrict__] = ACTIONS(7357),
    [anon_sym__Atomic] = ACTIONS(7357),
    [anon_sym__Noreturn] = ACTIONS(7357),
    [anon_sym__Nonnull] = ACTIONS(7357),
    [anon_sym_nullable] = ACTIONS(7357),
    [anon_sym__Complex] = ACTIONS(7357),
    [anon_sym__Nullable] = ACTIONS(7357),
    [anon_sym__Nullable_result] = ACTIONS(7357),
    [anon_sym__Null_unspecified] = ACTIONS(7357),
    [anon_sym___autoreleasing] = ACTIONS(7357),
    [anon_sym___block] = ACTIONS(7357),
    [anon_sym___bridge] = ACTIONS(7357),
    [anon_sym___bridge_retained] = ACTIONS(7357),
    [anon_sym___bridge_transfer] = ACTIONS(7357),
    [anon_sym___complex] = ACTIONS(7357),
    [anon_sym___const] = ACTIONS(7357),
    [anon_sym___imag] = ACTIONS(7357),
    [anon_sym___kindof] = ACTIONS(7357),
    [anon_sym___nonnull] = ACTIONS(7357),
    [anon_sym___nullable] = ACTIONS(7357),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7357),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7357),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7357),
    [anon_sym___real] = ACTIONS(7357),
    [anon_sym___strong] = ACTIONS(7357),
    [anon_sym___unsafe_unretained] = ACTIONS(7357),
    [anon_sym___unused] = ACTIONS(7357),
    [anon_sym___weak] = ACTIONS(7357),
    [anon_sym_alignas] = ACTIONS(7357),
    [anon_sym__Alignas] = ACTIONS(7357),
    [sym_primitive_type] = ACTIONS(7357),
    [anon_sym_enum] = ACTIONS(7357),
    [anon_sym_COLON] = ACTIONS(7359),
    [anon_sym_struct] = ACTIONS(7357),
    [anon_sym_union] = ACTIONS(7357),
    [anon_sym_in] = ACTIONS(7357),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7357),
    [anon_sym___typeof] = ACTIONS(7357),
    [anon_sym_typeof] = ACTIONS(7357),
    [anon_sym_BOOL] = ACTIONS(7357),
    [anon_sym_IMP] = ACTIONS(7357),
    [anon_sym_SEL] = ACTIONS(7357),
    [anon_sym_Class] = ACTIONS(7357),
    [anon_sym_id] = ACTIONS(7357),
    [anon_sym_out] = ACTIONS(7357),
    [anon_sym_inout] = ACTIONS(7357),
    [anon_sym_bycopy] = ACTIONS(7357),
    [anon_sym_byref] = ACTIONS(7357),
    [anon_sym_oneway] = ACTIONS(7357),
  },
  [STATE(3960)] = {
    [sym_identifier] = ACTIONS(7353),
    [anon_sym_COMMA] = ACTIONS(7355),
    [anon_sym_RPAREN] = ACTIONS(7355),
    [anon_sym_LPAREN2] = ACTIONS(7355),
    [anon_sym_STAR] = ACTIONS(7355),
    [anon_sym_CARET] = ACTIONS(7355),
    [anon_sym_GT] = ACTIONS(7355),
    [anon_sym_LT] = ACTIONS(7355),
    [anon_sym___extension__] = ACTIONS(7353),
    [anon_sym___attribute__] = ACTIONS(7353),
    [anon_sym___attribute] = ACTIONS(7353),
    [anon_sym_noreturn] = ACTIONS(7353),
    [anon_sym_LBRACK] = ACTIONS(7355),
    [anon_sym___based] = ACTIONS(7353),
    [anon_sym_LBRACE] = ACTIONS(7355),
    [anon_sym_signed] = ACTIONS(7353),
    [anon_sym_unsigned] = ACTIONS(7353),
    [anon_sym_long] = ACTIONS(7353),
    [anon_sym_short] = ACTIONS(7353),
    [anon_sym_ATautoreleasepool] = ACTIONS(7355),
    [anon_sym_const] = ACTIONS(7353),
    [anon_sym_constexpr] = ACTIONS(7353),
    [anon_sym_volatile] = ACTIONS(7353),
    [anon_sym_restrict] = ACTIONS(7353),
    [anon_sym___restrict__] = ACTIONS(7353),
    [anon_sym__Atomic] = ACTIONS(7353),
    [anon_sym__Noreturn] = ACTIONS(7353),
    [anon_sym__Nonnull] = ACTIONS(7353),
    [anon_sym_nullable] = ACTIONS(7353),
    [anon_sym__Complex] = ACTIONS(7353),
    [anon_sym__Nullable] = ACTIONS(7353),
    [anon_sym__Nullable_result] = ACTIONS(7353),
    [anon_sym__Null_unspecified] = ACTIONS(7353),
    [anon_sym___autoreleasing] = ACTIONS(7353),
    [anon_sym___block] = ACTIONS(7353),
    [anon_sym___bridge] = ACTIONS(7353),
    [anon_sym___bridge_retained] = ACTIONS(7353),
    [anon_sym___bridge_transfer] = ACTIONS(7353),
    [anon_sym___complex] = ACTIONS(7353),
    [anon_sym___const] = ACTIONS(7353),
    [anon_sym___imag] = ACTIONS(7353),
    [anon_sym___kindof] = ACTIONS(7353),
    [anon_sym___nonnull] = ACTIONS(7353),
    [anon_sym___nullable] = ACTIONS(7353),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7353),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7353),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7353),
    [anon_sym___real] = ACTIONS(7353),
    [anon_sym___strong] = ACTIONS(7353),
    [anon_sym___unsafe_unretained] = ACTIONS(7353),
    [anon_sym___unused] = ACTIONS(7353),
    [anon_sym___weak] = ACTIONS(7353),
    [anon_sym_alignas] = ACTIONS(7353),
    [anon_sym__Alignas] = ACTIONS(7353),
    [sym_primitive_type] = ACTIONS(7353),
    [anon_sym_enum] = ACTIONS(7353),
    [anon_sym_COLON] = ACTIONS(7355),
    [anon_sym_struct] = ACTIONS(7353),
    [anon_sym_union] = ACTIONS(7353),
    [anon_sym_in] = ACTIONS(7353),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7353),
    [anon_sym___typeof] = ACTIONS(7353),
    [anon_sym_typeof] = ACTIONS(7353),
    [anon_sym_BOOL] = ACTIONS(7353),
    [anon_sym_IMP] = ACTIONS(7353),
    [anon_sym_SEL] = ACTIONS(7353),
    [anon_sym_Class] = ACTIONS(7353),
    [anon_sym_id] = ACTIONS(7353),
    [anon_sym_out] = ACTIONS(7353),
    [anon_sym_inout] = ACTIONS(7353),
    [anon_sym_bycopy] = ACTIONS(7353),
    [anon_sym_byref] = ACTIONS(7353),
    [anon_sym_oneway] = ACTIONS(7353),
  },
  [STATE(3961)] = {
    [sym_attribute_specifier] = STATE(3859),
    [sym_identifier] = ACTIONS(7854),
    [anon_sym_COMMA] = ACTIONS(7856),
    [anon_sym_RPAREN] = ACTIONS(7856),
    [anon_sym_LPAREN2] = ACTIONS(7856),
    [anon_sym_STAR] = ACTIONS(7856),
    [anon_sym_CARET] = ACTIONS(7856),
    [anon_sym_GT] = ACTIONS(7856),
    [anon_sym_LT] = ACTIONS(7856),
    [anon_sym___extension__] = ACTIONS(7854),
    [anon_sym___attribute__] = ACTIONS(8920),
    [anon_sym___attribute] = ACTIONS(8920),
    [anon_sym_noreturn] = ACTIONS(7854),
    [anon_sym_LBRACK] = ACTIONS(7856),
    [anon_sym___based] = ACTIONS(7854),
    [anon_sym_LBRACE] = ACTIONS(7856),
    [anon_sym_signed] = ACTIONS(7854),
    [anon_sym_unsigned] = ACTIONS(7854),
    [anon_sym_long] = ACTIONS(7854),
    [anon_sym_short] = ACTIONS(7854),
    [anon_sym_ATautoreleasepool] = ACTIONS(7856),
    [anon_sym_const] = ACTIONS(7854),
    [anon_sym_constexpr] = ACTIONS(7854),
    [anon_sym_volatile] = ACTIONS(7854),
    [anon_sym_restrict] = ACTIONS(7854),
    [anon_sym___restrict__] = ACTIONS(7854),
    [anon_sym__Atomic] = ACTIONS(7854),
    [anon_sym__Noreturn] = ACTIONS(7854),
    [anon_sym__Nonnull] = ACTIONS(7854),
    [anon_sym_nullable] = ACTIONS(7854),
    [anon_sym__Complex] = ACTIONS(7854),
    [anon_sym__Nullable] = ACTIONS(7854),
    [anon_sym__Nullable_result] = ACTIONS(7854),
    [anon_sym__Null_unspecified] = ACTIONS(7854),
    [anon_sym___autoreleasing] = ACTIONS(7854),
    [anon_sym___block] = ACTIONS(7854),
    [anon_sym___bridge] = ACTIONS(7854),
    [anon_sym___bridge_retained] = ACTIONS(7854),
    [anon_sym___bridge_transfer] = ACTIONS(7854),
    [anon_sym___complex] = ACTIONS(7854),
    [anon_sym___const] = ACTIONS(7854),
    [anon_sym___imag] = ACTIONS(7854),
    [anon_sym___kindof] = ACTIONS(7854),
    [anon_sym___nonnull] = ACTIONS(7854),
    [anon_sym___nullable] = ACTIONS(7854),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7854),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7854),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7854),
    [anon_sym___real] = ACTIONS(7854),
    [anon_sym___strong] = ACTIONS(7854),
    [anon_sym___unsafe_unretained] = ACTIONS(7854),
    [anon_sym___unused] = ACTIONS(7854),
    [anon_sym___weak] = ACTIONS(7854),
    [anon_sym_alignas] = ACTIONS(7854),
    [anon_sym__Alignas] = ACTIONS(7854),
    [sym_primitive_type] = ACTIONS(7854),
    [anon_sym_enum] = ACTIONS(7854),
    [anon_sym_struct] = ACTIONS(7854),
    [anon_sym_union] = ACTIONS(7854),
    [anon_sym_in] = ACTIONS(7854),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7854),
    [anon_sym___typeof] = ACTIONS(7854),
    [anon_sym_typeof] = ACTIONS(7854),
    [anon_sym_BOOL] = ACTIONS(7854),
    [anon_sym_IMP] = ACTIONS(7854),
    [anon_sym_SEL] = ACTIONS(7854),
    [anon_sym_Class] = ACTIONS(7854),
    [anon_sym_id] = ACTIONS(7854),
    [anon_sym_out] = ACTIONS(7854),
    [anon_sym_inout] = ACTIONS(7854),
    [anon_sym_bycopy] = ACTIONS(7854),
    [anon_sym_byref] = ACTIONS(7854),
    [anon_sym_oneway] = ACTIONS(7854),
  },
  [STATE(3962)] = {
    [sym__type_definition_type] = STATE(4775),
    [sym_attribute_specifier] = STATE(4553),
    [sym_ms_declspec_modifier] = STATE(4026),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4189),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3963)] = {
    [sym_identifier] = ACTIONS(7369),
    [anon_sym_COMMA] = ACTIONS(7371),
    [anon_sym_RPAREN] = ACTIONS(7371),
    [anon_sym_LPAREN2] = ACTIONS(7371),
    [anon_sym_STAR] = ACTIONS(7371),
    [anon_sym_CARET] = ACTIONS(7371),
    [anon_sym_GT] = ACTIONS(7371),
    [anon_sym_LT] = ACTIONS(7371),
    [anon_sym___extension__] = ACTIONS(7369),
    [anon_sym___attribute__] = ACTIONS(7369),
    [anon_sym___attribute] = ACTIONS(7369),
    [anon_sym_noreturn] = ACTIONS(7369),
    [anon_sym_LBRACK] = ACTIONS(7371),
    [anon_sym___based] = ACTIONS(7369),
    [anon_sym_LBRACE] = ACTIONS(7371),
    [anon_sym_signed] = ACTIONS(7369),
    [anon_sym_unsigned] = ACTIONS(7369),
    [anon_sym_long] = ACTIONS(7369),
    [anon_sym_short] = ACTIONS(7369),
    [anon_sym_ATautoreleasepool] = ACTIONS(7371),
    [anon_sym_const] = ACTIONS(7369),
    [anon_sym_constexpr] = ACTIONS(7369),
    [anon_sym_volatile] = ACTIONS(7369),
    [anon_sym_restrict] = ACTIONS(7369),
    [anon_sym___restrict__] = ACTIONS(7369),
    [anon_sym__Atomic] = ACTIONS(7369),
    [anon_sym__Noreturn] = ACTIONS(7369),
    [anon_sym__Nonnull] = ACTIONS(7369),
    [anon_sym_nullable] = ACTIONS(7369),
    [anon_sym__Complex] = ACTIONS(7369),
    [anon_sym__Nullable] = ACTIONS(7369),
    [anon_sym__Nullable_result] = ACTIONS(7369),
    [anon_sym__Null_unspecified] = ACTIONS(7369),
    [anon_sym___autoreleasing] = ACTIONS(7369),
    [anon_sym___block] = ACTIONS(7369),
    [anon_sym___bridge] = ACTIONS(7369),
    [anon_sym___bridge_retained] = ACTIONS(7369),
    [anon_sym___bridge_transfer] = ACTIONS(7369),
    [anon_sym___complex] = ACTIONS(7369),
    [anon_sym___const] = ACTIONS(7369),
    [anon_sym___imag] = ACTIONS(7369),
    [anon_sym___kindof] = ACTIONS(7369),
    [anon_sym___nonnull] = ACTIONS(7369),
    [anon_sym___nullable] = ACTIONS(7369),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7369),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7369),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7369),
    [anon_sym___real] = ACTIONS(7369),
    [anon_sym___strong] = ACTIONS(7369),
    [anon_sym___unsafe_unretained] = ACTIONS(7369),
    [anon_sym___unused] = ACTIONS(7369),
    [anon_sym___weak] = ACTIONS(7369),
    [anon_sym_alignas] = ACTIONS(7369),
    [anon_sym__Alignas] = ACTIONS(7369),
    [sym_primitive_type] = ACTIONS(7369),
    [anon_sym_enum] = ACTIONS(7369),
    [anon_sym_COLON] = ACTIONS(7371),
    [anon_sym_struct] = ACTIONS(7369),
    [anon_sym_union] = ACTIONS(7369),
    [anon_sym_in] = ACTIONS(7369),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7369),
    [anon_sym___typeof] = ACTIONS(7369),
    [anon_sym_typeof] = ACTIONS(7369),
    [anon_sym_BOOL] = ACTIONS(7369),
    [anon_sym_IMP] = ACTIONS(7369),
    [anon_sym_SEL] = ACTIONS(7369),
    [anon_sym_Class] = ACTIONS(7369),
    [anon_sym_id] = ACTIONS(7369),
    [anon_sym_out] = ACTIONS(7369),
    [anon_sym_inout] = ACTIONS(7369),
    [anon_sym_bycopy] = ACTIONS(7369),
    [anon_sym_byref] = ACTIONS(7369),
    [anon_sym_oneway] = ACTIONS(7369),
  },
  [STATE(3964)] = {
    [sym_identifier] = ACTIONS(7373),
    [anon_sym_COMMA] = ACTIONS(7375),
    [anon_sym_RPAREN] = ACTIONS(7375),
    [anon_sym_LPAREN2] = ACTIONS(7375),
    [anon_sym_STAR] = ACTIONS(7375),
    [anon_sym_CARET] = ACTIONS(7375),
    [anon_sym_GT] = ACTIONS(7375),
    [anon_sym_LT] = ACTIONS(7375),
    [anon_sym___extension__] = ACTIONS(7373),
    [anon_sym___attribute__] = ACTIONS(7373),
    [anon_sym___attribute] = ACTIONS(7373),
    [anon_sym_noreturn] = ACTIONS(7373),
    [anon_sym_LBRACK] = ACTIONS(7375),
    [anon_sym___based] = ACTIONS(7373),
    [anon_sym_LBRACE] = ACTIONS(7375),
    [anon_sym_signed] = ACTIONS(7373),
    [anon_sym_unsigned] = ACTIONS(7373),
    [anon_sym_long] = ACTIONS(7373),
    [anon_sym_short] = ACTIONS(7373),
    [anon_sym_ATautoreleasepool] = ACTIONS(7375),
    [anon_sym_const] = ACTIONS(7373),
    [anon_sym_constexpr] = ACTIONS(7373),
    [anon_sym_volatile] = ACTIONS(7373),
    [anon_sym_restrict] = ACTIONS(7373),
    [anon_sym___restrict__] = ACTIONS(7373),
    [anon_sym__Atomic] = ACTIONS(7373),
    [anon_sym__Noreturn] = ACTIONS(7373),
    [anon_sym__Nonnull] = ACTIONS(7373),
    [anon_sym_nullable] = ACTIONS(7373),
    [anon_sym__Complex] = ACTIONS(7373),
    [anon_sym__Nullable] = ACTIONS(7373),
    [anon_sym__Nullable_result] = ACTIONS(7373),
    [anon_sym__Null_unspecified] = ACTIONS(7373),
    [anon_sym___autoreleasing] = ACTIONS(7373),
    [anon_sym___block] = ACTIONS(7373),
    [anon_sym___bridge] = ACTIONS(7373),
    [anon_sym___bridge_retained] = ACTIONS(7373),
    [anon_sym___bridge_transfer] = ACTIONS(7373),
    [anon_sym___complex] = ACTIONS(7373),
    [anon_sym___const] = ACTIONS(7373),
    [anon_sym___imag] = ACTIONS(7373),
    [anon_sym___kindof] = ACTIONS(7373),
    [anon_sym___nonnull] = ACTIONS(7373),
    [anon_sym___nullable] = ACTIONS(7373),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7373),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7373),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7373),
    [anon_sym___real] = ACTIONS(7373),
    [anon_sym___strong] = ACTIONS(7373),
    [anon_sym___unsafe_unretained] = ACTIONS(7373),
    [anon_sym___unused] = ACTIONS(7373),
    [anon_sym___weak] = ACTIONS(7373),
    [anon_sym_alignas] = ACTIONS(7373),
    [anon_sym__Alignas] = ACTIONS(7373),
    [sym_primitive_type] = ACTIONS(7373),
    [anon_sym_enum] = ACTIONS(7373),
    [anon_sym_COLON] = ACTIONS(7375),
    [anon_sym_struct] = ACTIONS(7373),
    [anon_sym_union] = ACTIONS(7373),
    [anon_sym_in] = ACTIONS(7373),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7373),
    [anon_sym___typeof] = ACTIONS(7373),
    [anon_sym_typeof] = ACTIONS(7373),
    [anon_sym_BOOL] = ACTIONS(7373),
    [anon_sym_IMP] = ACTIONS(7373),
    [anon_sym_SEL] = ACTIONS(7373),
    [anon_sym_Class] = ACTIONS(7373),
    [anon_sym_id] = ACTIONS(7373),
    [anon_sym_out] = ACTIONS(7373),
    [anon_sym_inout] = ACTIONS(7373),
    [anon_sym_bycopy] = ACTIONS(7373),
    [anon_sym_byref] = ACTIONS(7373),
    [anon_sym_oneway] = ACTIONS(7373),
  },
  [STATE(3965)] = {
    [sym_identifier] = ACTIONS(7349),
    [anon_sym_COMMA] = ACTIONS(7351),
    [anon_sym_RPAREN] = ACTIONS(7351),
    [anon_sym_LPAREN2] = ACTIONS(7351),
    [anon_sym_STAR] = ACTIONS(7351),
    [anon_sym_CARET] = ACTIONS(7351),
    [anon_sym_GT] = ACTIONS(7351),
    [anon_sym_LT] = ACTIONS(7351),
    [anon_sym___extension__] = ACTIONS(7349),
    [anon_sym___attribute__] = ACTIONS(7349),
    [anon_sym___attribute] = ACTIONS(7349),
    [anon_sym_noreturn] = ACTIONS(7349),
    [anon_sym_LBRACK] = ACTIONS(7351),
    [anon_sym___based] = ACTIONS(7349),
    [anon_sym_LBRACE] = ACTIONS(7351),
    [anon_sym_signed] = ACTIONS(7349),
    [anon_sym_unsigned] = ACTIONS(7349),
    [anon_sym_long] = ACTIONS(7349),
    [anon_sym_short] = ACTIONS(7349),
    [anon_sym_ATautoreleasepool] = ACTIONS(7351),
    [anon_sym_const] = ACTIONS(7349),
    [anon_sym_constexpr] = ACTIONS(7349),
    [anon_sym_volatile] = ACTIONS(7349),
    [anon_sym_restrict] = ACTIONS(7349),
    [anon_sym___restrict__] = ACTIONS(7349),
    [anon_sym__Atomic] = ACTIONS(7349),
    [anon_sym__Noreturn] = ACTIONS(7349),
    [anon_sym__Nonnull] = ACTIONS(7349),
    [anon_sym_nullable] = ACTIONS(7349),
    [anon_sym__Complex] = ACTIONS(7349),
    [anon_sym__Nullable] = ACTIONS(7349),
    [anon_sym__Nullable_result] = ACTIONS(7349),
    [anon_sym__Null_unspecified] = ACTIONS(7349),
    [anon_sym___autoreleasing] = ACTIONS(7349),
    [anon_sym___block] = ACTIONS(7349),
    [anon_sym___bridge] = ACTIONS(7349),
    [anon_sym___bridge_retained] = ACTIONS(7349),
    [anon_sym___bridge_transfer] = ACTIONS(7349),
    [anon_sym___complex] = ACTIONS(7349),
    [anon_sym___const] = ACTIONS(7349),
    [anon_sym___imag] = ACTIONS(7349),
    [anon_sym___kindof] = ACTIONS(7349),
    [anon_sym___nonnull] = ACTIONS(7349),
    [anon_sym___nullable] = ACTIONS(7349),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7349),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7349),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7349),
    [anon_sym___real] = ACTIONS(7349),
    [anon_sym___strong] = ACTIONS(7349),
    [anon_sym___unsafe_unretained] = ACTIONS(7349),
    [anon_sym___unused] = ACTIONS(7349),
    [anon_sym___weak] = ACTIONS(7349),
    [anon_sym_alignas] = ACTIONS(7349),
    [anon_sym__Alignas] = ACTIONS(7349),
    [sym_primitive_type] = ACTIONS(7349),
    [anon_sym_enum] = ACTIONS(7349),
    [anon_sym_COLON] = ACTIONS(7351),
    [anon_sym_struct] = ACTIONS(7349),
    [anon_sym_union] = ACTIONS(7349),
    [anon_sym_in] = ACTIONS(7349),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7349),
    [anon_sym___typeof] = ACTIONS(7349),
    [anon_sym_typeof] = ACTIONS(7349),
    [anon_sym_BOOL] = ACTIONS(7349),
    [anon_sym_IMP] = ACTIONS(7349),
    [anon_sym_SEL] = ACTIONS(7349),
    [anon_sym_Class] = ACTIONS(7349),
    [anon_sym_id] = ACTIONS(7349),
    [anon_sym_out] = ACTIONS(7349),
    [anon_sym_inout] = ACTIONS(7349),
    [anon_sym_bycopy] = ACTIONS(7349),
    [anon_sym_byref] = ACTIONS(7349),
    [anon_sym_oneway] = ACTIONS(7349),
  },
  [STATE(3966)] = {
    [sym_identifier] = ACTIONS(7381),
    [anon_sym_COMMA] = ACTIONS(7383),
    [anon_sym_RPAREN] = ACTIONS(7383),
    [anon_sym_LPAREN2] = ACTIONS(7383),
    [anon_sym_STAR] = ACTIONS(7383),
    [anon_sym_CARET] = ACTIONS(7383),
    [anon_sym_GT] = ACTIONS(7383),
    [anon_sym_LT] = ACTIONS(7383),
    [anon_sym___extension__] = ACTIONS(7381),
    [anon_sym___attribute__] = ACTIONS(7381),
    [anon_sym___attribute] = ACTIONS(7381),
    [anon_sym_noreturn] = ACTIONS(7381),
    [anon_sym_LBRACK] = ACTIONS(7383),
    [anon_sym___based] = ACTIONS(7381),
    [anon_sym_LBRACE] = ACTIONS(7383),
    [anon_sym_signed] = ACTIONS(7381),
    [anon_sym_unsigned] = ACTIONS(7381),
    [anon_sym_long] = ACTIONS(7381),
    [anon_sym_short] = ACTIONS(7381),
    [anon_sym_ATautoreleasepool] = ACTIONS(7383),
    [anon_sym_const] = ACTIONS(7381),
    [anon_sym_constexpr] = ACTIONS(7381),
    [anon_sym_volatile] = ACTIONS(7381),
    [anon_sym_restrict] = ACTIONS(7381),
    [anon_sym___restrict__] = ACTIONS(7381),
    [anon_sym__Atomic] = ACTIONS(7381),
    [anon_sym__Noreturn] = ACTIONS(7381),
    [anon_sym__Nonnull] = ACTIONS(7381),
    [anon_sym_nullable] = ACTIONS(7381),
    [anon_sym__Complex] = ACTIONS(7381),
    [anon_sym__Nullable] = ACTIONS(7381),
    [anon_sym__Nullable_result] = ACTIONS(7381),
    [anon_sym__Null_unspecified] = ACTIONS(7381),
    [anon_sym___autoreleasing] = ACTIONS(7381),
    [anon_sym___block] = ACTIONS(7381),
    [anon_sym___bridge] = ACTIONS(7381),
    [anon_sym___bridge_retained] = ACTIONS(7381),
    [anon_sym___bridge_transfer] = ACTIONS(7381),
    [anon_sym___complex] = ACTIONS(7381),
    [anon_sym___const] = ACTIONS(7381),
    [anon_sym___imag] = ACTIONS(7381),
    [anon_sym___kindof] = ACTIONS(7381),
    [anon_sym___nonnull] = ACTIONS(7381),
    [anon_sym___nullable] = ACTIONS(7381),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7381),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7381),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7381),
    [anon_sym___real] = ACTIONS(7381),
    [anon_sym___strong] = ACTIONS(7381),
    [anon_sym___unsafe_unretained] = ACTIONS(7381),
    [anon_sym___unused] = ACTIONS(7381),
    [anon_sym___weak] = ACTIONS(7381),
    [anon_sym_alignas] = ACTIONS(7381),
    [anon_sym__Alignas] = ACTIONS(7381),
    [sym_primitive_type] = ACTIONS(7381),
    [anon_sym_enum] = ACTIONS(7381),
    [anon_sym_COLON] = ACTIONS(7383),
    [anon_sym_struct] = ACTIONS(7381),
    [anon_sym_union] = ACTIONS(7381),
    [anon_sym_in] = ACTIONS(7381),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7381),
    [anon_sym___typeof] = ACTIONS(7381),
    [anon_sym_typeof] = ACTIONS(7381),
    [anon_sym_BOOL] = ACTIONS(7381),
    [anon_sym_IMP] = ACTIONS(7381),
    [anon_sym_SEL] = ACTIONS(7381),
    [anon_sym_Class] = ACTIONS(7381),
    [anon_sym_id] = ACTIONS(7381),
    [anon_sym_out] = ACTIONS(7381),
    [anon_sym_inout] = ACTIONS(7381),
    [anon_sym_bycopy] = ACTIONS(7381),
    [anon_sym_byref] = ACTIONS(7381),
    [anon_sym_oneway] = ACTIONS(7381),
  },
  [STATE(3967)] = {
    [sym_identifier] = ACTIONS(7385),
    [anon_sym_COMMA] = ACTIONS(7387),
    [anon_sym_RPAREN] = ACTIONS(7387),
    [anon_sym_LPAREN2] = ACTIONS(7387),
    [anon_sym_STAR] = ACTIONS(7387),
    [anon_sym_CARET] = ACTIONS(7387),
    [anon_sym_GT] = ACTIONS(7387),
    [anon_sym_LT] = ACTIONS(7387),
    [anon_sym___extension__] = ACTIONS(7385),
    [anon_sym___attribute__] = ACTIONS(7385),
    [anon_sym___attribute] = ACTIONS(7385),
    [anon_sym_noreturn] = ACTIONS(7385),
    [anon_sym_LBRACK] = ACTIONS(7387),
    [anon_sym___based] = ACTIONS(7385),
    [anon_sym_LBRACE] = ACTIONS(7387),
    [anon_sym_signed] = ACTIONS(7385),
    [anon_sym_unsigned] = ACTIONS(7385),
    [anon_sym_long] = ACTIONS(7385),
    [anon_sym_short] = ACTIONS(7385),
    [anon_sym_ATautoreleasepool] = ACTIONS(7387),
    [anon_sym_const] = ACTIONS(7385),
    [anon_sym_constexpr] = ACTIONS(7385),
    [anon_sym_volatile] = ACTIONS(7385),
    [anon_sym_restrict] = ACTIONS(7385),
    [anon_sym___restrict__] = ACTIONS(7385),
    [anon_sym__Atomic] = ACTIONS(7385),
    [anon_sym__Noreturn] = ACTIONS(7385),
    [anon_sym__Nonnull] = ACTIONS(7385),
    [anon_sym_nullable] = ACTIONS(7385),
    [anon_sym__Complex] = ACTIONS(7385),
    [anon_sym__Nullable] = ACTIONS(7385),
    [anon_sym__Nullable_result] = ACTIONS(7385),
    [anon_sym__Null_unspecified] = ACTIONS(7385),
    [anon_sym___autoreleasing] = ACTIONS(7385),
    [anon_sym___block] = ACTIONS(7385),
    [anon_sym___bridge] = ACTIONS(7385),
    [anon_sym___bridge_retained] = ACTIONS(7385),
    [anon_sym___bridge_transfer] = ACTIONS(7385),
    [anon_sym___complex] = ACTIONS(7385),
    [anon_sym___const] = ACTIONS(7385),
    [anon_sym___imag] = ACTIONS(7385),
    [anon_sym___kindof] = ACTIONS(7385),
    [anon_sym___nonnull] = ACTIONS(7385),
    [anon_sym___nullable] = ACTIONS(7385),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7385),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7385),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7385),
    [anon_sym___real] = ACTIONS(7385),
    [anon_sym___strong] = ACTIONS(7385),
    [anon_sym___unsafe_unretained] = ACTIONS(7385),
    [anon_sym___unused] = ACTIONS(7385),
    [anon_sym___weak] = ACTIONS(7385),
    [anon_sym_alignas] = ACTIONS(7385),
    [anon_sym__Alignas] = ACTIONS(7385),
    [sym_primitive_type] = ACTIONS(7385),
    [anon_sym_enum] = ACTIONS(7385),
    [anon_sym_COLON] = ACTIONS(7387),
    [anon_sym_struct] = ACTIONS(7385),
    [anon_sym_union] = ACTIONS(7385),
    [anon_sym_in] = ACTIONS(7385),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7385),
    [anon_sym___typeof] = ACTIONS(7385),
    [anon_sym_typeof] = ACTIONS(7385),
    [anon_sym_BOOL] = ACTIONS(7385),
    [anon_sym_IMP] = ACTIONS(7385),
    [anon_sym_SEL] = ACTIONS(7385),
    [anon_sym_Class] = ACTIONS(7385),
    [anon_sym_id] = ACTIONS(7385),
    [anon_sym_out] = ACTIONS(7385),
    [anon_sym_inout] = ACTIONS(7385),
    [anon_sym_bycopy] = ACTIONS(7385),
    [anon_sym_byref] = ACTIONS(7385),
    [anon_sym_oneway] = ACTIONS(7385),
  },
  [STATE(3968)] = {
    [sym_identifier] = ACTIONS(7417),
    [anon_sym_COMMA] = ACTIONS(7419),
    [anon_sym_RPAREN] = ACTIONS(7419),
    [anon_sym_LPAREN2] = ACTIONS(7419),
    [anon_sym_STAR] = ACTIONS(7419),
    [anon_sym_CARET] = ACTIONS(7419),
    [anon_sym_GT] = ACTIONS(7419),
    [anon_sym_LT] = ACTIONS(7419),
    [anon_sym___extension__] = ACTIONS(7417),
    [anon_sym___attribute__] = ACTIONS(7417),
    [anon_sym___attribute] = ACTIONS(7417),
    [anon_sym_noreturn] = ACTIONS(7417),
    [anon_sym_LBRACK] = ACTIONS(7419),
    [anon_sym___based] = ACTIONS(7417),
    [anon_sym_LBRACE] = ACTIONS(7419),
    [anon_sym_signed] = ACTIONS(7417),
    [anon_sym_unsigned] = ACTIONS(7417),
    [anon_sym_long] = ACTIONS(7417),
    [anon_sym_short] = ACTIONS(7417),
    [anon_sym_ATautoreleasepool] = ACTIONS(7419),
    [anon_sym_const] = ACTIONS(7417),
    [anon_sym_constexpr] = ACTIONS(7417),
    [anon_sym_volatile] = ACTIONS(7417),
    [anon_sym_restrict] = ACTIONS(7417),
    [anon_sym___restrict__] = ACTIONS(7417),
    [anon_sym__Atomic] = ACTIONS(7417),
    [anon_sym__Noreturn] = ACTIONS(7417),
    [anon_sym__Nonnull] = ACTIONS(7417),
    [anon_sym_nullable] = ACTIONS(7417),
    [anon_sym__Complex] = ACTIONS(7417),
    [anon_sym__Nullable] = ACTIONS(7417),
    [anon_sym__Nullable_result] = ACTIONS(7417),
    [anon_sym__Null_unspecified] = ACTIONS(7417),
    [anon_sym___autoreleasing] = ACTIONS(7417),
    [anon_sym___block] = ACTIONS(7417),
    [anon_sym___bridge] = ACTIONS(7417),
    [anon_sym___bridge_retained] = ACTIONS(7417),
    [anon_sym___bridge_transfer] = ACTIONS(7417),
    [anon_sym___complex] = ACTIONS(7417),
    [anon_sym___const] = ACTIONS(7417),
    [anon_sym___imag] = ACTIONS(7417),
    [anon_sym___kindof] = ACTIONS(7417),
    [anon_sym___nonnull] = ACTIONS(7417),
    [anon_sym___nullable] = ACTIONS(7417),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7417),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7417),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7417),
    [anon_sym___real] = ACTIONS(7417),
    [anon_sym___strong] = ACTIONS(7417),
    [anon_sym___unsafe_unretained] = ACTIONS(7417),
    [anon_sym___unused] = ACTIONS(7417),
    [anon_sym___weak] = ACTIONS(7417),
    [anon_sym_alignas] = ACTIONS(7417),
    [anon_sym__Alignas] = ACTIONS(7417),
    [sym_primitive_type] = ACTIONS(7417),
    [anon_sym_enum] = ACTIONS(7417),
    [anon_sym_COLON] = ACTIONS(7419),
    [anon_sym_struct] = ACTIONS(7417),
    [anon_sym_union] = ACTIONS(7417),
    [anon_sym_in] = ACTIONS(7417),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7417),
    [anon_sym___typeof] = ACTIONS(7417),
    [anon_sym_typeof] = ACTIONS(7417),
    [anon_sym_BOOL] = ACTIONS(7417),
    [anon_sym_IMP] = ACTIONS(7417),
    [anon_sym_SEL] = ACTIONS(7417),
    [anon_sym_Class] = ACTIONS(7417),
    [anon_sym_id] = ACTIONS(7417),
    [anon_sym_out] = ACTIONS(7417),
    [anon_sym_inout] = ACTIONS(7417),
    [anon_sym_bycopy] = ACTIONS(7417),
    [anon_sym_byref] = ACTIONS(7417),
    [anon_sym_oneway] = ACTIONS(7417),
  },
  [STATE(3969)] = {
    [sym_identifier] = ACTIONS(7389),
    [anon_sym_COMMA] = ACTIONS(7391),
    [anon_sym_RPAREN] = ACTIONS(7391),
    [anon_sym_LPAREN2] = ACTIONS(7391),
    [anon_sym_STAR] = ACTIONS(7391),
    [anon_sym_CARET] = ACTIONS(7391),
    [anon_sym_GT] = ACTIONS(7391),
    [anon_sym_LT] = ACTIONS(7391),
    [anon_sym___extension__] = ACTIONS(7389),
    [anon_sym___attribute__] = ACTIONS(7389),
    [anon_sym___attribute] = ACTIONS(7389),
    [anon_sym_noreturn] = ACTIONS(7389),
    [anon_sym_LBRACK] = ACTIONS(7391),
    [anon_sym___based] = ACTIONS(7389),
    [anon_sym_LBRACE] = ACTIONS(7391),
    [anon_sym_signed] = ACTIONS(7389),
    [anon_sym_unsigned] = ACTIONS(7389),
    [anon_sym_long] = ACTIONS(7389),
    [anon_sym_short] = ACTIONS(7389),
    [anon_sym_ATautoreleasepool] = ACTIONS(7391),
    [anon_sym_const] = ACTIONS(7389),
    [anon_sym_constexpr] = ACTIONS(7389),
    [anon_sym_volatile] = ACTIONS(7389),
    [anon_sym_restrict] = ACTIONS(7389),
    [anon_sym___restrict__] = ACTIONS(7389),
    [anon_sym__Atomic] = ACTIONS(7389),
    [anon_sym__Noreturn] = ACTIONS(7389),
    [anon_sym__Nonnull] = ACTIONS(7389),
    [anon_sym_nullable] = ACTIONS(7389),
    [anon_sym__Complex] = ACTIONS(7389),
    [anon_sym__Nullable] = ACTIONS(7389),
    [anon_sym__Nullable_result] = ACTIONS(7389),
    [anon_sym__Null_unspecified] = ACTIONS(7389),
    [anon_sym___autoreleasing] = ACTIONS(7389),
    [anon_sym___block] = ACTIONS(7389),
    [anon_sym___bridge] = ACTIONS(7389),
    [anon_sym___bridge_retained] = ACTIONS(7389),
    [anon_sym___bridge_transfer] = ACTIONS(7389),
    [anon_sym___complex] = ACTIONS(7389),
    [anon_sym___const] = ACTIONS(7389),
    [anon_sym___imag] = ACTIONS(7389),
    [anon_sym___kindof] = ACTIONS(7389),
    [anon_sym___nonnull] = ACTIONS(7389),
    [anon_sym___nullable] = ACTIONS(7389),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7389),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7389),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7389),
    [anon_sym___real] = ACTIONS(7389),
    [anon_sym___strong] = ACTIONS(7389),
    [anon_sym___unsafe_unretained] = ACTIONS(7389),
    [anon_sym___unused] = ACTIONS(7389),
    [anon_sym___weak] = ACTIONS(7389),
    [anon_sym_alignas] = ACTIONS(7389),
    [anon_sym__Alignas] = ACTIONS(7389),
    [sym_primitive_type] = ACTIONS(7389),
    [anon_sym_enum] = ACTIONS(7389),
    [anon_sym_COLON] = ACTIONS(7391),
    [anon_sym_struct] = ACTIONS(7389),
    [anon_sym_union] = ACTIONS(7389),
    [anon_sym_in] = ACTIONS(7389),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7389),
    [anon_sym___typeof] = ACTIONS(7389),
    [anon_sym_typeof] = ACTIONS(7389),
    [anon_sym_BOOL] = ACTIONS(7389),
    [anon_sym_IMP] = ACTIONS(7389),
    [anon_sym_SEL] = ACTIONS(7389),
    [anon_sym_Class] = ACTIONS(7389),
    [anon_sym_id] = ACTIONS(7389),
    [anon_sym_out] = ACTIONS(7389),
    [anon_sym_inout] = ACTIONS(7389),
    [anon_sym_bycopy] = ACTIONS(7389),
    [anon_sym_byref] = ACTIONS(7389),
    [anon_sym_oneway] = ACTIONS(7389),
  },
  [STATE(3970)] = {
    [sym_identifier] = ACTIONS(7393),
    [anon_sym_COMMA] = ACTIONS(7395),
    [anon_sym_RPAREN] = ACTIONS(7395),
    [anon_sym_LPAREN2] = ACTIONS(7395),
    [anon_sym_STAR] = ACTIONS(7395),
    [anon_sym_CARET] = ACTIONS(7395),
    [anon_sym_GT] = ACTIONS(7395),
    [anon_sym_LT] = ACTIONS(7395),
    [anon_sym___extension__] = ACTIONS(7393),
    [anon_sym___attribute__] = ACTIONS(7393),
    [anon_sym___attribute] = ACTIONS(7393),
    [anon_sym_noreturn] = ACTIONS(7393),
    [anon_sym_LBRACK] = ACTIONS(7395),
    [anon_sym___based] = ACTIONS(7393),
    [anon_sym_LBRACE] = ACTIONS(7395),
    [anon_sym_signed] = ACTIONS(7393),
    [anon_sym_unsigned] = ACTIONS(7393),
    [anon_sym_long] = ACTIONS(7393),
    [anon_sym_short] = ACTIONS(7393),
    [anon_sym_ATautoreleasepool] = ACTIONS(7395),
    [anon_sym_const] = ACTIONS(7393),
    [anon_sym_constexpr] = ACTIONS(7393),
    [anon_sym_volatile] = ACTIONS(7393),
    [anon_sym_restrict] = ACTIONS(7393),
    [anon_sym___restrict__] = ACTIONS(7393),
    [anon_sym__Atomic] = ACTIONS(7393),
    [anon_sym__Noreturn] = ACTIONS(7393),
    [anon_sym__Nonnull] = ACTIONS(7393),
    [anon_sym_nullable] = ACTIONS(7393),
    [anon_sym__Complex] = ACTIONS(7393),
    [anon_sym__Nullable] = ACTIONS(7393),
    [anon_sym__Nullable_result] = ACTIONS(7393),
    [anon_sym__Null_unspecified] = ACTIONS(7393),
    [anon_sym___autoreleasing] = ACTIONS(7393),
    [anon_sym___block] = ACTIONS(7393),
    [anon_sym___bridge] = ACTIONS(7393),
    [anon_sym___bridge_retained] = ACTIONS(7393),
    [anon_sym___bridge_transfer] = ACTIONS(7393),
    [anon_sym___complex] = ACTIONS(7393),
    [anon_sym___const] = ACTIONS(7393),
    [anon_sym___imag] = ACTIONS(7393),
    [anon_sym___kindof] = ACTIONS(7393),
    [anon_sym___nonnull] = ACTIONS(7393),
    [anon_sym___nullable] = ACTIONS(7393),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7393),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7393),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7393),
    [anon_sym___real] = ACTIONS(7393),
    [anon_sym___strong] = ACTIONS(7393),
    [anon_sym___unsafe_unretained] = ACTIONS(7393),
    [anon_sym___unused] = ACTIONS(7393),
    [anon_sym___weak] = ACTIONS(7393),
    [anon_sym_alignas] = ACTIONS(7393),
    [anon_sym__Alignas] = ACTIONS(7393),
    [sym_primitive_type] = ACTIONS(7393),
    [anon_sym_enum] = ACTIONS(7393),
    [anon_sym_COLON] = ACTIONS(7395),
    [anon_sym_struct] = ACTIONS(7393),
    [anon_sym_union] = ACTIONS(7393),
    [anon_sym_in] = ACTIONS(7393),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7393),
    [anon_sym___typeof] = ACTIONS(7393),
    [anon_sym_typeof] = ACTIONS(7393),
    [anon_sym_BOOL] = ACTIONS(7393),
    [anon_sym_IMP] = ACTIONS(7393),
    [anon_sym_SEL] = ACTIONS(7393),
    [anon_sym_Class] = ACTIONS(7393),
    [anon_sym_id] = ACTIONS(7393),
    [anon_sym_out] = ACTIONS(7393),
    [anon_sym_inout] = ACTIONS(7393),
    [anon_sym_bycopy] = ACTIONS(7393),
    [anon_sym_byref] = ACTIONS(7393),
    [anon_sym_oneway] = ACTIONS(7393),
  },
  [STATE(3971)] = {
    [sym_identifier] = ACTIONS(7397),
    [anon_sym_COMMA] = ACTIONS(7399),
    [anon_sym_RPAREN] = ACTIONS(7399),
    [anon_sym_LPAREN2] = ACTIONS(7399),
    [anon_sym_STAR] = ACTIONS(7399),
    [anon_sym_CARET] = ACTIONS(7399),
    [anon_sym_GT] = ACTIONS(7399),
    [anon_sym_LT] = ACTIONS(7399),
    [anon_sym___extension__] = ACTIONS(7397),
    [anon_sym___attribute__] = ACTIONS(7397),
    [anon_sym___attribute] = ACTIONS(7397),
    [anon_sym_noreturn] = ACTIONS(7397),
    [anon_sym_LBRACK] = ACTIONS(7399),
    [anon_sym___based] = ACTIONS(7397),
    [anon_sym_LBRACE] = ACTIONS(7399),
    [anon_sym_signed] = ACTIONS(7397),
    [anon_sym_unsigned] = ACTIONS(7397),
    [anon_sym_long] = ACTIONS(7397),
    [anon_sym_short] = ACTIONS(7397),
    [anon_sym_ATautoreleasepool] = ACTIONS(7399),
    [anon_sym_const] = ACTIONS(7397),
    [anon_sym_constexpr] = ACTIONS(7397),
    [anon_sym_volatile] = ACTIONS(7397),
    [anon_sym_restrict] = ACTIONS(7397),
    [anon_sym___restrict__] = ACTIONS(7397),
    [anon_sym__Atomic] = ACTIONS(7397),
    [anon_sym__Noreturn] = ACTIONS(7397),
    [anon_sym__Nonnull] = ACTIONS(7397),
    [anon_sym_nullable] = ACTIONS(7397),
    [anon_sym__Complex] = ACTIONS(7397),
    [anon_sym__Nullable] = ACTIONS(7397),
    [anon_sym__Nullable_result] = ACTIONS(7397),
    [anon_sym__Null_unspecified] = ACTIONS(7397),
    [anon_sym___autoreleasing] = ACTIONS(7397),
    [anon_sym___block] = ACTIONS(7397),
    [anon_sym___bridge] = ACTIONS(7397),
    [anon_sym___bridge_retained] = ACTIONS(7397),
    [anon_sym___bridge_transfer] = ACTIONS(7397),
    [anon_sym___complex] = ACTIONS(7397),
    [anon_sym___const] = ACTIONS(7397),
    [anon_sym___imag] = ACTIONS(7397),
    [anon_sym___kindof] = ACTIONS(7397),
    [anon_sym___nonnull] = ACTIONS(7397),
    [anon_sym___nullable] = ACTIONS(7397),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7397),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7397),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7397),
    [anon_sym___real] = ACTIONS(7397),
    [anon_sym___strong] = ACTIONS(7397),
    [anon_sym___unsafe_unretained] = ACTIONS(7397),
    [anon_sym___unused] = ACTIONS(7397),
    [anon_sym___weak] = ACTIONS(7397),
    [anon_sym_alignas] = ACTIONS(7397),
    [anon_sym__Alignas] = ACTIONS(7397),
    [sym_primitive_type] = ACTIONS(7397),
    [anon_sym_enum] = ACTIONS(7397),
    [anon_sym_COLON] = ACTIONS(7399),
    [anon_sym_struct] = ACTIONS(7397),
    [anon_sym_union] = ACTIONS(7397),
    [anon_sym_in] = ACTIONS(7397),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7397),
    [anon_sym___typeof] = ACTIONS(7397),
    [anon_sym_typeof] = ACTIONS(7397),
    [anon_sym_BOOL] = ACTIONS(7397),
    [anon_sym_IMP] = ACTIONS(7397),
    [anon_sym_SEL] = ACTIONS(7397),
    [anon_sym_Class] = ACTIONS(7397),
    [anon_sym_id] = ACTIONS(7397),
    [anon_sym_out] = ACTIONS(7397),
    [anon_sym_inout] = ACTIONS(7397),
    [anon_sym_bycopy] = ACTIONS(7397),
    [anon_sym_byref] = ACTIONS(7397),
    [anon_sym_oneway] = ACTIONS(7397),
  },
  [STATE(3972)] = {
    [sym_identifier] = ACTIONS(7429),
    [anon_sym_COMMA] = ACTIONS(7431),
    [anon_sym_RPAREN] = ACTIONS(7431),
    [anon_sym_LPAREN2] = ACTIONS(7431),
    [anon_sym_STAR] = ACTIONS(7431),
    [anon_sym_CARET] = ACTIONS(7431),
    [anon_sym_GT] = ACTIONS(7431),
    [anon_sym_LT] = ACTIONS(7431),
    [anon_sym___extension__] = ACTIONS(7429),
    [anon_sym___attribute__] = ACTIONS(7429),
    [anon_sym___attribute] = ACTIONS(7429),
    [anon_sym_noreturn] = ACTIONS(7429),
    [anon_sym_LBRACK] = ACTIONS(7431),
    [anon_sym___based] = ACTIONS(7429),
    [anon_sym_LBRACE] = ACTIONS(7431),
    [anon_sym_signed] = ACTIONS(7429),
    [anon_sym_unsigned] = ACTIONS(7429),
    [anon_sym_long] = ACTIONS(7429),
    [anon_sym_short] = ACTIONS(7429),
    [anon_sym_ATautoreleasepool] = ACTIONS(7431),
    [anon_sym_const] = ACTIONS(7429),
    [anon_sym_constexpr] = ACTIONS(7429),
    [anon_sym_volatile] = ACTIONS(7429),
    [anon_sym_restrict] = ACTIONS(7429),
    [anon_sym___restrict__] = ACTIONS(7429),
    [anon_sym__Atomic] = ACTIONS(7429),
    [anon_sym__Noreturn] = ACTIONS(7429),
    [anon_sym__Nonnull] = ACTIONS(7429),
    [anon_sym_nullable] = ACTIONS(7429),
    [anon_sym__Complex] = ACTIONS(7429),
    [anon_sym__Nullable] = ACTIONS(7429),
    [anon_sym__Nullable_result] = ACTIONS(7429),
    [anon_sym__Null_unspecified] = ACTIONS(7429),
    [anon_sym___autoreleasing] = ACTIONS(7429),
    [anon_sym___block] = ACTIONS(7429),
    [anon_sym___bridge] = ACTIONS(7429),
    [anon_sym___bridge_retained] = ACTIONS(7429),
    [anon_sym___bridge_transfer] = ACTIONS(7429),
    [anon_sym___complex] = ACTIONS(7429),
    [anon_sym___const] = ACTIONS(7429),
    [anon_sym___imag] = ACTIONS(7429),
    [anon_sym___kindof] = ACTIONS(7429),
    [anon_sym___nonnull] = ACTIONS(7429),
    [anon_sym___nullable] = ACTIONS(7429),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7429),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7429),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7429),
    [anon_sym___real] = ACTIONS(7429),
    [anon_sym___strong] = ACTIONS(7429),
    [anon_sym___unsafe_unretained] = ACTIONS(7429),
    [anon_sym___unused] = ACTIONS(7429),
    [anon_sym___weak] = ACTIONS(7429),
    [anon_sym_alignas] = ACTIONS(7429),
    [anon_sym__Alignas] = ACTIONS(7429),
    [sym_primitive_type] = ACTIONS(7429),
    [anon_sym_enum] = ACTIONS(7429),
    [anon_sym_COLON] = ACTIONS(7431),
    [anon_sym_struct] = ACTIONS(7429),
    [anon_sym_union] = ACTIONS(7429),
    [anon_sym_in] = ACTIONS(7429),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7429),
    [anon_sym___typeof] = ACTIONS(7429),
    [anon_sym_typeof] = ACTIONS(7429),
    [anon_sym_BOOL] = ACTIONS(7429),
    [anon_sym_IMP] = ACTIONS(7429),
    [anon_sym_SEL] = ACTIONS(7429),
    [anon_sym_Class] = ACTIONS(7429),
    [anon_sym_id] = ACTIONS(7429),
    [anon_sym_out] = ACTIONS(7429),
    [anon_sym_inout] = ACTIONS(7429),
    [anon_sym_bycopy] = ACTIONS(7429),
    [anon_sym_byref] = ACTIONS(7429),
    [anon_sym_oneway] = ACTIONS(7429),
  },
  [STATE(3973)] = {
    [sym_identifier] = ACTIONS(7401),
    [anon_sym_COMMA] = ACTIONS(7403),
    [anon_sym_RPAREN] = ACTIONS(7403),
    [anon_sym_LPAREN2] = ACTIONS(7403),
    [anon_sym_STAR] = ACTIONS(7403),
    [anon_sym_CARET] = ACTIONS(7403),
    [anon_sym_GT] = ACTIONS(7403),
    [anon_sym_LT] = ACTIONS(7403),
    [anon_sym___extension__] = ACTIONS(7401),
    [anon_sym___attribute__] = ACTIONS(7401),
    [anon_sym___attribute] = ACTIONS(7401),
    [anon_sym_noreturn] = ACTIONS(7401),
    [anon_sym_LBRACK] = ACTIONS(7403),
    [anon_sym___based] = ACTIONS(7401),
    [anon_sym_LBRACE] = ACTIONS(7403),
    [anon_sym_signed] = ACTIONS(7401),
    [anon_sym_unsigned] = ACTIONS(7401),
    [anon_sym_long] = ACTIONS(7401),
    [anon_sym_short] = ACTIONS(7401),
    [anon_sym_ATautoreleasepool] = ACTIONS(7403),
    [anon_sym_const] = ACTIONS(7401),
    [anon_sym_constexpr] = ACTIONS(7401),
    [anon_sym_volatile] = ACTIONS(7401),
    [anon_sym_restrict] = ACTIONS(7401),
    [anon_sym___restrict__] = ACTIONS(7401),
    [anon_sym__Atomic] = ACTIONS(7401),
    [anon_sym__Noreturn] = ACTIONS(7401),
    [anon_sym__Nonnull] = ACTIONS(7401),
    [anon_sym_nullable] = ACTIONS(7401),
    [anon_sym__Complex] = ACTIONS(7401),
    [anon_sym__Nullable] = ACTIONS(7401),
    [anon_sym__Nullable_result] = ACTIONS(7401),
    [anon_sym__Null_unspecified] = ACTIONS(7401),
    [anon_sym___autoreleasing] = ACTIONS(7401),
    [anon_sym___block] = ACTIONS(7401),
    [anon_sym___bridge] = ACTIONS(7401),
    [anon_sym___bridge_retained] = ACTIONS(7401),
    [anon_sym___bridge_transfer] = ACTIONS(7401),
    [anon_sym___complex] = ACTIONS(7401),
    [anon_sym___const] = ACTIONS(7401),
    [anon_sym___imag] = ACTIONS(7401),
    [anon_sym___kindof] = ACTIONS(7401),
    [anon_sym___nonnull] = ACTIONS(7401),
    [anon_sym___nullable] = ACTIONS(7401),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7401),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7401),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7401),
    [anon_sym___real] = ACTIONS(7401),
    [anon_sym___strong] = ACTIONS(7401),
    [anon_sym___unsafe_unretained] = ACTIONS(7401),
    [anon_sym___unused] = ACTIONS(7401),
    [anon_sym___weak] = ACTIONS(7401),
    [anon_sym_alignas] = ACTIONS(7401),
    [anon_sym__Alignas] = ACTIONS(7401),
    [sym_primitive_type] = ACTIONS(7401),
    [anon_sym_enum] = ACTIONS(7401),
    [anon_sym_COLON] = ACTIONS(7403),
    [anon_sym_struct] = ACTIONS(7401),
    [anon_sym_union] = ACTIONS(7401),
    [anon_sym_in] = ACTIONS(7401),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7401),
    [anon_sym___typeof] = ACTIONS(7401),
    [anon_sym_typeof] = ACTIONS(7401),
    [anon_sym_BOOL] = ACTIONS(7401),
    [anon_sym_IMP] = ACTIONS(7401),
    [anon_sym_SEL] = ACTIONS(7401),
    [anon_sym_Class] = ACTIONS(7401),
    [anon_sym_id] = ACTIONS(7401),
    [anon_sym_out] = ACTIONS(7401),
    [anon_sym_inout] = ACTIONS(7401),
    [anon_sym_bycopy] = ACTIONS(7401),
    [anon_sym_byref] = ACTIONS(7401),
    [anon_sym_oneway] = ACTIONS(7401),
  },
  [STATE(3974)] = {
    [sym_identifier] = ACTIONS(8904),
    [anon_sym_COMMA] = ACTIONS(8906),
    [anon_sym_RPAREN] = ACTIONS(8906),
    [anon_sym_LPAREN2] = ACTIONS(8906),
    [anon_sym_STAR] = ACTIONS(8906),
    [anon_sym_CARET] = ACTIONS(8906),
    [anon_sym_GT] = ACTIONS(8906),
    [anon_sym_LT] = ACTIONS(8906),
    [anon_sym___extension__] = ACTIONS(8904),
    [anon_sym___attribute__] = ACTIONS(8904),
    [anon_sym___attribute] = ACTIONS(8904),
    [anon_sym_noreturn] = ACTIONS(8904),
    [anon_sym_LBRACK] = ACTIONS(8923),
    [anon_sym___based] = ACTIONS(8904),
    [anon_sym_LBRACE] = ACTIONS(8906),
    [anon_sym_signed] = ACTIONS(8904),
    [anon_sym_unsigned] = ACTIONS(8904),
    [anon_sym_long] = ACTIONS(8904),
    [anon_sym_short] = ACTIONS(8904),
    [anon_sym_ATautoreleasepool] = ACTIONS(8906),
    [anon_sym_const] = ACTIONS(8904),
    [anon_sym_constexpr] = ACTIONS(8904),
    [anon_sym_volatile] = ACTIONS(8904),
    [anon_sym_restrict] = ACTIONS(8904),
    [anon_sym___restrict__] = ACTIONS(8904),
    [anon_sym__Atomic] = ACTIONS(8904),
    [anon_sym__Noreturn] = ACTIONS(8904),
    [anon_sym__Nonnull] = ACTIONS(8904),
    [anon_sym_nullable] = ACTIONS(8904),
    [anon_sym__Complex] = ACTIONS(8904),
    [anon_sym__Nullable] = ACTIONS(8904),
    [anon_sym__Nullable_result] = ACTIONS(8904),
    [anon_sym__Null_unspecified] = ACTIONS(8904),
    [anon_sym___autoreleasing] = ACTIONS(8904),
    [anon_sym___block] = ACTIONS(8904),
    [anon_sym___bridge] = ACTIONS(8904),
    [anon_sym___bridge_retained] = ACTIONS(8904),
    [anon_sym___bridge_transfer] = ACTIONS(8904),
    [anon_sym___complex] = ACTIONS(8904),
    [anon_sym___const] = ACTIONS(8904),
    [anon_sym___imag] = ACTIONS(8904),
    [anon_sym___kindof] = ACTIONS(8904),
    [anon_sym___nonnull] = ACTIONS(8904),
    [anon_sym___nullable] = ACTIONS(8904),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8904),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8904),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8904),
    [anon_sym___real] = ACTIONS(8904),
    [anon_sym___strong] = ACTIONS(8904),
    [anon_sym___unsafe_unretained] = ACTIONS(8904),
    [anon_sym___unused] = ACTIONS(8904),
    [anon_sym___weak] = ACTIONS(8904),
    [anon_sym_alignas] = ACTIONS(8904),
    [anon_sym__Alignas] = ACTIONS(8904),
    [sym_primitive_type] = ACTIONS(8904),
    [anon_sym_enum] = ACTIONS(8904),
    [anon_sym_COLON] = ACTIONS(8906),
    [anon_sym_struct] = ACTIONS(8904),
    [anon_sym_union] = ACTIONS(8904),
    [anon_sym_in] = ACTIONS(8904),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8904),
    [anon_sym___typeof] = ACTIONS(8904),
    [anon_sym_typeof] = ACTIONS(8904),
    [anon_sym_BOOL] = ACTIONS(8904),
    [anon_sym_IMP] = ACTIONS(8904),
    [anon_sym_SEL] = ACTIONS(8904),
    [anon_sym_Class] = ACTIONS(8904),
    [anon_sym_id] = ACTIONS(8904),
    [anon_sym_out] = ACTIONS(8904),
    [anon_sym_inout] = ACTIONS(8904),
    [anon_sym_bycopy] = ACTIONS(8904),
    [anon_sym_byref] = ACTIONS(8904),
    [anon_sym_oneway] = ACTIONS(8904),
  },
  [STATE(3975)] = {
    [sym__type_definition_type] = STATE(4758),
    [sym_attribute_specifier] = STATE(3837),
    [sym_ms_declspec_modifier] = STATE(4026),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3976)] = {
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_RPAREN] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7103),
    [anon_sym_STAR] = ACTIONS(7103),
    [anon_sym_CARET] = ACTIONS(7103),
    [anon_sym_GT] = ACTIONS(7103),
    [anon_sym_LT] = ACTIONS(7103),
    [anon_sym___extension__] = ACTIONS(7101),
    [anon_sym___attribute__] = ACTIONS(7101),
    [anon_sym___attribute] = ACTIONS(7101),
    [anon_sym_noreturn] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7103),
    [anon_sym___based] = ACTIONS(7101),
    [anon_sym_LBRACE] = ACTIONS(7103),
    [anon_sym_signed] = ACTIONS(7101),
    [anon_sym_unsigned] = ACTIONS(7101),
    [anon_sym_long] = ACTIONS(7101),
    [anon_sym_short] = ACTIONS(7101),
    [anon_sym_ATautoreleasepool] = ACTIONS(7103),
    [anon_sym_const] = ACTIONS(7101),
    [anon_sym_constexpr] = ACTIONS(7101),
    [anon_sym_volatile] = ACTIONS(7101),
    [anon_sym_restrict] = ACTIONS(7101),
    [anon_sym___restrict__] = ACTIONS(7101),
    [anon_sym__Atomic] = ACTIONS(7101),
    [anon_sym__Noreturn] = ACTIONS(7101),
    [anon_sym__Nonnull] = ACTIONS(7101),
    [anon_sym_nullable] = ACTIONS(7101),
    [anon_sym__Complex] = ACTIONS(7101),
    [anon_sym__Nullable] = ACTIONS(7101),
    [anon_sym__Nullable_result] = ACTIONS(7101),
    [anon_sym__Null_unspecified] = ACTIONS(7101),
    [anon_sym___autoreleasing] = ACTIONS(7101),
    [anon_sym___block] = ACTIONS(7101),
    [anon_sym___bridge] = ACTIONS(7101),
    [anon_sym___bridge_retained] = ACTIONS(7101),
    [anon_sym___bridge_transfer] = ACTIONS(7101),
    [anon_sym___complex] = ACTIONS(7101),
    [anon_sym___const] = ACTIONS(7101),
    [anon_sym___imag] = ACTIONS(7101),
    [anon_sym___kindof] = ACTIONS(7101),
    [anon_sym___nonnull] = ACTIONS(7101),
    [anon_sym___nullable] = ACTIONS(7101),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7101),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7101),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7101),
    [anon_sym___real] = ACTIONS(7101),
    [anon_sym___strong] = ACTIONS(7101),
    [anon_sym___unsafe_unretained] = ACTIONS(7101),
    [anon_sym___unused] = ACTIONS(7101),
    [anon_sym___weak] = ACTIONS(7101),
    [anon_sym_alignas] = ACTIONS(7101),
    [anon_sym__Alignas] = ACTIONS(7101),
    [sym_primitive_type] = ACTIONS(7101),
    [anon_sym_enum] = ACTIONS(7101),
    [anon_sym_COLON] = ACTIONS(7103),
    [anon_sym_struct] = ACTIONS(7101),
    [anon_sym_union] = ACTIONS(7101),
    [anon_sym_in] = ACTIONS(7101),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7101),
    [anon_sym___typeof] = ACTIONS(7101),
    [anon_sym_typeof] = ACTIONS(7101),
    [anon_sym_BOOL] = ACTIONS(7101),
    [anon_sym_IMP] = ACTIONS(7101),
    [anon_sym_SEL] = ACTIONS(7101),
    [anon_sym_Class] = ACTIONS(7101),
    [anon_sym_id] = ACTIONS(7101),
    [anon_sym_out] = ACTIONS(7101),
    [anon_sym_inout] = ACTIONS(7101),
    [anon_sym_bycopy] = ACTIONS(7101),
    [anon_sym_byref] = ACTIONS(7101),
    [anon_sym_oneway] = ACTIONS(7101),
  },
  [STATE(3977)] = {
    [sym__type_definition_type] = STATE(4741),
    [sym_attribute_specifier] = STATE(3906),
    [sym_ms_declspec_modifier] = STATE(4010),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym___declspec] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(3978)] = {
    [sym_identifier] = ACTIONS(6788),
    [anon_sym_COMMA] = ACTIONS(6790),
    [anon_sym_RPAREN] = ACTIONS(6790),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_STAR] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_GT] = ACTIONS(6790),
    [anon_sym_LT] = ACTIONS(6790),
    [anon_sym___extension__] = ACTIONS(6788),
    [anon_sym___attribute__] = ACTIONS(6788),
    [anon_sym___attribute] = ACTIONS(6788),
    [anon_sym_noreturn] = ACTIONS(6788),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym___based] = ACTIONS(6788),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_signed] = ACTIONS(6788),
    [anon_sym_unsigned] = ACTIONS(6788),
    [anon_sym_long] = ACTIONS(6788),
    [anon_sym_short] = ACTIONS(6788),
    [anon_sym_ATautoreleasepool] = ACTIONS(6790),
    [anon_sym_const] = ACTIONS(6788),
    [anon_sym_constexpr] = ACTIONS(6788),
    [anon_sym_volatile] = ACTIONS(6788),
    [anon_sym_restrict] = ACTIONS(6788),
    [anon_sym___restrict__] = ACTIONS(6788),
    [anon_sym__Atomic] = ACTIONS(6788),
    [anon_sym__Noreturn] = ACTIONS(6788),
    [anon_sym__Nonnull] = ACTIONS(6788),
    [anon_sym_nullable] = ACTIONS(6788),
    [anon_sym__Complex] = ACTIONS(6788),
    [anon_sym__Nullable] = ACTIONS(6788),
    [anon_sym__Nullable_result] = ACTIONS(6788),
    [anon_sym__Null_unspecified] = ACTIONS(6788),
    [anon_sym___autoreleasing] = ACTIONS(6788),
    [anon_sym___block] = ACTIONS(6788),
    [anon_sym___bridge] = ACTIONS(6788),
    [anon_sym___bridge_retained] = ACTIONS(6788),
    [anon_sym___bridge_transfer] = ACTIONS(6788),
    [anon_sym___complex] = ACTIONS(6788),
    [anon_sym___const] = ACTIONS(6788),
    [anon_sym___imag] = ACTIONS(6788),
    [anon_sym___kindof] = ACTIONS(6788),
    [anon_sym___nonnull] = ACTIONS(6788),
    [anon_sym___nullable] = ACTIONS(6788),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6788),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6788),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6788),
    [anon_sym___real] = ACTIONS(6788),
    [anon_sym___strong] = ACTIONS(6788),
    [anon_sym___unsafe_unretained] = ACTIONS(6788),
    [anon_sym___unused] = ACTIONS(6788),
    [anon_sym___weak] = ACTIONS(6788),
    [anon_sym_alignas] = ACTIONS(6788),
    [anon_sym__Alignas] = ACTIONS(6788),
    [sym_primitive_type] = ACTIONS(6788),
    [anon_sym_enum] = ACTIONS(6788),
    [anon_sym_struct] = ACTIONS(6788),
    [anon_sym_union] = ACTIONS(6788),
    [anon_sym_in] = ACTIONS(6788),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6788),
    [anon_sym___typeof] = ACTIONS(6788),
    [anon_sym_typeof] = ACTIONS(6788),
    [anon_sym_BOOL] = ACTIONS(6788),
    [anon_sym_IMP] = ACTIONS(6788),
    [anon_sym_SEL] = ACTIONS(6788),
    [anon_sym_Class] = ACTIONS(6788),
    [anon_sym_id] = ACTIONS(6788),
    [anon_sym_out] = ACTIONS(6788),
    [anon_sym_inout] = ACTIONS(6788),
    [anon_sym_bycopy] = ACTIONS(6788),
    [anon_sym_byref] = ACTIONS(6788),
    [anon_sym_oneway] = ACTIONS(6788),
  },
  [STATE(3979)] = {
    [sym_identifier] = ACTIONS(4224),
    [anon_sym_COMMA] = ACTIONS(4226),
    [anon_sym_RPAREN] = ACTIONS(4226),
    [anon_sym_LPAREN2] = ACTIONS(4226),
    [anon_sym_STAR] = ACTIONS(4226),
    [anon_sym_CARET] = ACTIONS(4226),
    [anon_sym_GT] = ACTIONS(4226),
    [anon_sym_LT] = ACTIONS(4226),
    [anon_sym___extension__] = ACTIONS(4224),
    [anon_sym___attribute__] = ACTIONS(4224),
    [anon_sym___attribute] = ACTIONS(4224),
    [anon_sym_noreturn] = ACTIONS(4224),
    [anon_sym_LBRACK] = ACTIONS(4226),
    [anon_sym___based] = ACTIONS(4224),
    [anon_sym_LBRACE] = ACTIONS(4226),
    [anon_sym_signed] = ACTIONS(4224),
    [anon_sym_unsigned] = ACTIONS(4224),
    [anon_sym_long] = ACTIONS(4224),
    [anon_sym_short] = ACTIONS(4224),
    [anon_sym_ATautoreleasepool] = ACTIONS(4226),
    [anon_sym_const] = ACTIONS(4224),
    [anon_sym_constexpr] = ACTIONS(4224),
    [anon_sym_volatile] = ACTIONS(4224),
    [anon_sym_restrict] = ACTIONS(4224),
    [anon_sym___restrict__] = ACTIONS(4224),
    [anon_sym__Atomic] = ACTIONS(4224),
    [anon_sym__Noreturn] = ACTIONS(4224),
    [anon_sym__Nonnull] = ACTIONS(4224),
    [anon_sym_nullable] = ACTIONS(4224),
    [anon_sym__Complex] = ACTIONS(4224),
    [anon_sym__Nullable] = ACTIONS(4224),
    [anon_sym__Nullable_result] = ACTIONS(4224),
    [anon_sym__Null_unspecified] = ACTIONS(4224),
    [anon_sym___autoreleasing] = ACTIONS(4224),
    [anon_sym___block] = ACTIONS(4224),
    [anon_sym___bridge] = ACTIONS(4224),
    [anon_sym___bridge_retained] = ACTIONS(4224),
    [anon_sym___bridge_transfer] = ACTIONS(4224),
    [anon_sym___complex] = ACTIONS(4224),
    [anon_sym___const] = ACTIONS(4224),
    [anon_sym___imag] = ACTIONS(4224),
    [anon_sym___kindof] = ACTIONS(4224),
    [anon_sym___nonnull] = ACTIONS(4224),
    [anon_sym___nullable] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4224),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4224),
    [anon_sym___real] = ACTIONS(4224),
    [anon_sym___strong] = ACTIONS(4224),
    [anon_sym___unsafe_unretained] = ACTIONS(4224),
    [anon_sym___unused] = ACTIONS(4224),
    [anon_sym___weak] = ACTIONS(4224),
    [anon_sym_alignas] = ACTIONS(4224),
    [anon_sym__Alignas] = ACTIONS(4224),
    [sym_primitive_type] = ACTIONS(4224),
    [anon_sym_enum] = ACTIONS(4224),
    [anon_sym_struct] = ACTIONS(4224),
    [anon_sym_union] = ACTIONS(4224),
    [anon_sym_in] = ACTIONS(4224),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4224),
    [anon_sym___typeof] = ACTIONS(4224),
    [anon_sym_typeof] = ACTIONS(4224),
    [anon_sym_BOOL] = ACTIONS(4224),
    [anon_sym_IMP] = ACTIONS(4224),
    [anon_sym_SEL] = ACTIONS(4224),
    [anon_sym_Class] = ACTIONS(4224),
    [anon_sym_id] = ACTIONS(4224),
    [anon_sym_out] = ACTIONS(4224),
    [anon_sym_inout] = ACTIONS(4224),
    [anon_sym_bycopy] = ACTIONS(4224),
    [anon_sym_byref] = ACTIONS(4224),
    [anon_sym_oneway] = ACTIONS(4224),
  },
  [STATE(3980)] = {
    [sym_identifier] = ACTIONS(6141),
    [anon_sym_COMMA] = ACTIONS(6128),
    [anon_sym_RPAREN] = ACTIONS(6128),
    [anon_sym_LPAREN2] = ACTIONS(6128),
    [anon_sym_STAR] = ACTIONS(6128),
    [anon_sym_CARET] = ACTIONS(6128),
    [anon_sym_GT] = ACTIONS(6128),
    [anon_sym_LT] = ACTIONS(6128),
    [anon_sym___extension__] = ACTIONS(6141),
    [anon_sym___attribute__] = ACTIONS(6141),
    [anon_sym___attribute] = ACTIONS(6141),
    [anon_sym_noreturn] = ACTIONS(6141),
    [anon_sym_LBRACK] = ACTIONS(6128),
    [anon_sym___based] = ACTIONS(6141),
    [anon_sym_LBRACE] = ACTIONS(6128),
    [anon_sym_signed] = ACTIONS(6141),
    [anon_sym_unsigned] = ACTIONS(6141),
    [anon_sym_long] = ACTIONS(6141),
    [anon_sym_short] = ACTIONS(6141),
    [anon_sym_ATautoreleasepool] = ACTIONS(6128),
    [anon_sym_const] = ACTIONS(6141),
    [anon_sym_constexpr] = ACTIONS(6141),
    [anon_sym_volatile] = ACTIONS(6141),
    [anon_sym_restrict] = ACTIONS(6141),
    [anon_sym___restrict__] = ACTIONS(6141),
    [anon_sym__Atomic] = ACTIONS(6141),
    [anon_sym__Noreturn] = ACTIONS(6141),
    [anon_sym__Nonnull] = ACTIONS(6141),
    [anon_sym_nullable] = ACTIONS(6141),
    [anon_sym__Complex] = ACTIONS(6141),
    [anon_sym__Nullable] = ACTIONS(6141),
    [anon_sym__Nullable_result] = ACTIONS(6141),
    [anon_sym__Null_unspecified] = ACTIONS(6141),
    [anon_sym___autoreleasing] = ACTIONS(6141),
    [anon_sym___block] = ACTIONS(6141),
    [anon_sym___bridge] = ACTIONS(6141),
    [anon_sym___bridge_retained] = ACTIONS(6141),
    [anon_sym___bridge_transfer] = ACTIONS(6141),
    [anon_sym___complex] = ACTIONS(6141),
    [anon_sym___const] = ACTIONS(6141),
    [anon_sym___imag] = ACTIONS(6141),
    [anon_sym___kindof] = ACTIONS(6141),
    [anon_sym___nonnull] = ACTIONS(6141),
    [anon_sym___nullable] = ACTIONS(6141),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6141),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6141),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6141),
    [anon_sym___real] = ACTIONS(6141),
    [anon_sym___strong] = ACTIONS(6141),
    [anon_sym___unsafe_unretained] = ACTIONS(6141),
    [anon_sym___unused] = ACTIONS(6141),
    [anon_sym___weak] = ACTIONS(6141),
    [anon_sym_alignas] = ACTIONS(6141),
    [anon_sym__Alignas] = ACTIONS(6141),
    [sym_primitive_type] = ACTIONS(6141),
    [anon_sym_enum] = ACTIONS(6141),
    [anon_sym_struct] = ACTIONS(6141),
    [anon_sym_union] = ACTIONS(6141),
    [anon_sym_in] = ACTIONS(6141),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6141),
    [anon_sym___typeof] = ACTIONS(6141),
    [anon_sym_typeof] = ACTIONS(6141),
    [anon_sym_BOOL] = ACTIONS(6141),
    [anon_sym_IMP] = ACTIONS(6141),
    [anon_sym_SEL] = ACTIONS(6141),
    [anon_sym_Class] = ACTIONS(6141),
    [anon_sym_id] = ACTIONS(6141),
    [anon_sym_out] = ACTIONS(6141),
    [anon_sym_inout] = ACTIONS(6141),
    [anon_sym_bycopy] = ACTIONS(6141),
    [anon_sym_byref] = ACTIONS(6141),
    [anon_sym_oneway] = ACTIONS(6141),
  },
  [STATE(3981)] = {
    [sym_identifier] = ACTIONS(4168),
    [anon_sym_COMMA] = ACTIONS(4170),
    [anon_sym_RPAREN] = ACTIONS(4170),
    [anon_sym_LPAREN2] = ACTIONS(4170),
    [anon_sym_STAR] = ACTIONS(4170),
    [anon_sym_CARET] = ACTIONS(4170),
    [anon_sym_GT] = ACTIONS(4170),
    [anon_sym_LT] = ACTIONS(4170),
    [anon_sym___extension__] = ACTIONS(4168),
    [anon_sym___attribute__] = ACTIONS(4168),
    [anon_sym___attribute] = ACTIONS(4168),
    [anon_sym_noreturn] = ACTIONS(4168),
    [anon_sym_LBRACK] = ACTIONS(4170),
    [anon_sym___based] = ACTIONS(4168),
    [anon_sym_LBRACE] = ACTIONS(4170),
    [anon_sym_signed] = ACTIONS(4168),
    [anon_sym_unsigned] = ACTIONS(4168),
    [anon_sym_long] = ACTIONS(4168),
    [anon_sym_short] = ACTIONS(4168),
    [anon_sym_ATautoreleasepool] = ACTIONS(4170),
    [anon_sym_const] = ACTIONS(4168),
    [anon_sym_constexpr] = ACTIONS(4168),
    [anon_sym_volatile] = ACTIONS(4168),
    [anon_sym_restrict] = ACTIONS(4168),
    [anon_sym___restrict__] = ACTIONS(4168),
    [anon_sym__Atomic] = ACTIONS(4168),
    [anon_sym__Noreturn] = ACTIONS(4168),
    [anon_sym__Nonnull] = ACTIONS(4168),
    [anon_sym_nullable] = ACTIONS(4168),
    [anon_sym__Complex] = ACTIONS(4168),
    [anon_sym__Nullable] = ACTIONS(4168),
    [anon_sym__Nullable_result] = ACTIONS(4168),
    [anon_sym__Null_unspecified] = ACTIONS(4168),
    [anon_sym___autoreleasing] = ACTIONS(4168),
    [anon_sym___block] = ACTIONS(4168),
    [anon_sym___bridge] = ACTIONS(4168),
    [anon_sym___bridge_retained] = ACTIONS(4168),
    [anon_sym___bridge_transfer] = ACTIONS(4168),
    [anon_sym___complex] = ACTIONS(4168),
    [anon_sym___const] = ACTIONS(4168),
    [anon_sym___imag] = ACTIONS(4168),
    [anon_sym___kindof] = ACTIONS(4168),
    [anon_sym___nonnull] = ACTIONS(4168),
    [anon_sym___nullable] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4168),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4168),
    [anon_sym___real] = ACTIONS(4168),
    [anon_sym___strong] = ACTIONS(4168),
    [anon_sym___unsafe_unretained] = ACTIONS(4168),
    [anon_sym___unused] = ACTIONS(4168),
    [anon_sym___weak] = ACTIONS(4168),
    [anon_sym_alignas] = ACTIONS(4168),
    [anon_sym__Alignas] = ACTIONS(4168),
    [sym_primitive_type] = ACTIONS(4168),
    [anon_sym_enum] = ACTIONS(4168),
    [anon_sym_struct] = ACTIONS(4168),
    [anon_sym_union] = ACTIONS(4168),
    [anon_sym_in] = ACTIONS(4168),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4168),
    [anon_sym___typeof] = ACTIONS(4168),
    [anon_sym_typeof] = ACTIONS(4168),
    [anon_sym_BOOL] = ACTIONS(4168),
    [anon_sym_IMP] = ACTIONS(4168),
    [anon_sym_SEL] = ACTIONS(4168),
    [anon_sym_Class] = ACTIONS(4168),
    [anon_sym_id] = ACTIONS(4168),
    [anon_sym_out] = ACTIONS(4168),
    [anon_sym_inout] = ACTIONS(4168),
    [anon_sym_bycopy] = ACTIONS(4168),
    [anon_sym_byref] = ACTIONS(4168),
    [anon_sym_oneway] = ACTIONS(4168),
  },
  [STATE(3982)] = {
    [sym_identifier] = ACTIONS(6752),
    [anon_sym_COMMA] = ACTIONS(6754),
    [anon_sym_RPAREN] = ACTIONS(6754),
    [anon_sym_LPAREN2] = ACTIONS(6754),
    [anon_sym_STAR] = ACTIONS(6754),
    [anon_sym_CARET] = ACTIONS(6754),
    [anon_sym_GT] = ACTIONS(6754),
    [anon_sym_LT] = ACTIONS(6754),
    [anon_sym___extension__] = ACTIONS(6752),
    [anon_sym___attribute__] = ACTIONS(6752),
    [anon_sym___attribute] = ACTIONS(6752),
    [anon_sym_noreturn] = ACTIONS(6752),
    [anon_sym_LBRACK] = ACTIONS(6754),
    [anon_sym___based] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6754),
    [anon_sym_signed] = ACTIONS(6752),
    [anon_sym_unsigned] = ACTIONS(6752),
    [anon_sym_long] = ACTIONS(6752),
    [anon_sym_short] = ACTIONS(6752),
    [anon_sym_ATautoreleasepool] = ACTIONS(6754),
    [anon_sym_const] = ACTIONS(6752),
    [anon_sym_constexpr] = ACTIONS(6752),
    [anon_sym_volatile] = ACTIONS(6752),
    [anon_sym_restrict] = ACTIONS(6752),
    [anon_sym___restrict__] = ACTIONS(6752),
    [anon_sym__Atomic] = ACTIONS(6752),
    [anon_sym__Noreturn] = ACTIONS(6752),
    [anon_sym__Nonnull] = ACTIONS(6752),
    [anon_sym_nullable] = ACTIONS(6752),
    [anon_sym__Complex] = ACTIONS(6752),
    [anon_sym__Nullable] = ACTIONS(6752),
    [anon_sym__Nullable_result] = ACTIONS(6752),
    [anon_sym__Null_unspecified] = ACTIONS(6752),
    [anon_sym___autoreleasing] = ACTIONS(6752),
    [anon_sym___block] = ACTIONS(6752),
    [anon_sym___bridge] = ACTIONS(6752),
    [anon_sym___bridge_retained] = ACTIONS(6752),
    [anon_sym___bridge_transfer] = ACTIONS(6752),
    [anon_sym___complex] = ACTIONS(6752),
    [anon_sym___const] = ACTIONS(6752),
    [anon_sym___imag] = ACTIONS(6752),
    [anon_sym___kindof] = ACTIONS(6752),
    [anon_sym___nonnull] = ACTIONS(6752),
    [anon_sym___nullable] = ACTIONS(6752),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6752),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6752),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6752),
    [anon_sym___real] = ACTIONS(6752),
    [anon_sym___strong] = ACTIONS(6752),
    [anon_sym___unsafe_unretained] = ACTIONS(6752),
    [anon_sym___unused] = ACTIONS(6752),
    [anon_sym___weak] = ACTIONS(6752),
    [anon_sym_alignas] = ACTIONS(6752),
    [anon_sym__Alignas] = ACTIONS(6752),
    [sym_primitive_type] = ACTIONS(6752),
    [anon_sym_enum] = ACTIONS(6752),
    [anon_sym_struct] = ACTIONS(6752),
    [anon_sym_union] = ACTIONS(6752),
    [anon_sym_in] = ACTIONS(6752),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6752),
    [anon_sym___typeof] = ACTIONS(6752),
    [anon_sym_typeof] = ACTIONS(6752),
    [anon_sym_BOOL] = ACTIONS(6752),
    [anon_sym_IMP] = ACTIONS(6752),
    [anon_sym_SEL] = ACTIONS(6752),
    [anon_sym_Class] = ACTIONS(6752),
    [anon_sym_id] = ACTIONS(6752),
    [anon_sym_out] = ACTIONS(6752),
    [anon_sym_inout] = ACTIONS(6752),
    [anon_sym_bycopy] = ACTIONS(6752),
    [anon_sym_byref] = ACTIONS(6752),
    [anon_sym_oneway] = ACTIONS(6752),
  },
  [STATE(3983)] = {
    [sym_identifier] = ACTIONS(4054),
    [anon_sym_COMMA] = ACTIONS(4056),
    [anon_sym_RPAREN] = ACTIONS(4056),
    [anon_sym_LPAREN2] = ACTIONS(4056),
    [anon_sym_STAR] = ACTIONS(4056),
    [anon_sym_CARET] = ACTIONS(4056),
    [anon_sym_GT] = ACTIONS(4056),
    [anon_sym_LT] = ACTIONS(4056),
    [anon_sym___extension__] = ACTIONS(4054),
    [anon_sym___attribute__] = ACTIONS(4054),
    [anon_sym___attribute] = ACTIONS(4054),
    [anon_sym_noreturn] = ACTIONS(4054),
    [anon_sym_LBRACK] = ACTIONS(4056),
    [anon_sym___based] = ACTIONS(4054),
    [anon_sym_LBRACE] = ACTIONS(4056),
    [anon_sym_signed] = ACTIONS(4054),
    [anon_sym_unsigned] = ACTIONS(4054),
    [anon_sym_long] = ACTIONS(4054),
    [anon_sym_short] = ACTIONS(4054),
    [anon_sym_ATautoreleasepool] = ACTIONS(4056),
    [anon_sym_const] = ACTIONS(4054),
    [anon_sym_constexpr] = ACTIONS(4054),
    [anon_sym_volatile] = ACTIONS(4054),
    [anon_sym_restrict] = ACTIONS(4054),
    [anon_sym___restrict__] = ACTIONS(4054),
    [anon_sym__Atomic] = ACTIONS(4054),
    [anon_sym__Noreturn] = ACTIONS(4054),
    [anon_sym__Nonnull] = ACTIONS(4054),
    [anon_sym_nullable] = ACTIONS(4054),
    [anon_sym__Complex] = ACTIONS(4054),
    [anon_sym__Nullable] = ACTIONS(4054),
    [anon_sym__Nullable_result] = ACTIONS(4054),
    [anon_sym__Null_unspecified] = ACTIONS(4054),
    [anon_sym___autoreleasing] = ACTIONS(4054),
    [anon_sym___block] = ACTIONS(4054),
    [anon_sym___bridge] = ACTIONS(4054),
    [anon_sym___bridge_retained] = ACTIONS(4054),
    [anon_sym___bridge_transfer] = ACTIONS(4054),
    [anon_sym___complex] = ACTIONS(4054),
    [anon_sym___const] = ACTIONS(4054),
    [anon_sym___imag] = ACTIONS(4054),
    [anon_sym___kindof] = ACTIONS(4054),
    [anon_sym___nonnull] = ACTIONS(4054),
    [anon_sym___nullable] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4054),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4054),
    [anon_sym___real] = ACTIONS(4054),
    [anon_sym___strong] = ACTIONS(4054),
    [anon_sym___unsafe_unretained] = ACTIONS(4054),
    [anon_sym___unused] = ACTIONS(4054),
    [anon_sym___weak] = ACTIONS(4054),
    [anon_sym_alignas] = ACTIONS(4054),
    [anon_sym__Alignas] = ACTIONS(4054),
    [sym_primitive_type] = ACTIONS(4054),
    [anon_sym_enum] = ACTIONS(4054),
    [anon_sym_struct] = ACTIONS(4054),
    [anon_sym_union] = ACTIONS(4054),
    [anon_sym_in] = ACTIONS(4054),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4054),
    [anon_sym___typeof] = ACTIONS(4054),
    [anon_sym_typeof] = ACTIONS(4054),
    [anon_sym_BOOL] = ACTIONS(4054),
    [anon_sym_IMP] = ACTIONS(4054),
    [anon_sym_SEL] = ACTIONS(4054),
    [anon_sym_Class] = ACTIONS(4054),
    [anon_sym_id] = ACTIONS(4054),
    [anon_sym_out] = ACTIONS(4054),
    [anon_sym_inout] = ACTIONS(4054),
    [anon_sym_bycopy] = ACTIONS(4054),
    [anon_sym_byref] = ACTIONS(4054),
    [anon_sym_oneway] = ACTIONS(4054),
  },
  [STATE(3984)] = {
    [sym_identifier] = ACTIONS(4058),
    [anon_sym_COMMA] = ACTIONS(4060),
    [anon_sym_RPAREN] = ACTIONS(4060),
    [anon_sym_LPAREN2] = ACTIONS(4060),
    [anon_sym_STAR] = ACTIONS(4060),
    [anon_sym_CARET] = ACTIONS(4060),
    [anon_sym_GT] = ACTIONS(4060),
    [anon_sym_LT] = ACTIONS(4060),
    [anon_sym___extension__] = ACTIONS(4058),
    [anon_sym___attribute__] = ACTIONS(4058),
    [anon_sym___attribute] = ACTIONS(4058),
    [anon_sym_noreturn] = ACTIONS(4058),
    [anon_sym_LBRACK] = ACTIONS(4060),
    [anon_sym___based] = ACTIONS(4058),
    [anon_sym_LBRACE] = ACTIONS(4060),
    [anon_sym_signed] = ACTIONS(4058),
    [anon_sym_unsigned] = ACTIONS(4058),
    [anon_sym_long] = ACTIONS(4058),
    [anon_sym_short] = ACTIONS(4058),
    [anon_sym_ATautoreleasepool] = ACTIONS(4060),
    [anon_sym_const] = ACTIONS(4058),
    [anon_sym_constexpr] = ACTIONS(4058),
    [anon_sym_volatile] = ACTIONS(4058),
    [anon_sym_restrict] = ACTIONS(4058),
    [anon_sym___restrict__] = ACTIONS(4058),
    [anon_sym__Atomic] = ACTIONS(4058),
    [anon_sym__Noreturn] = ACTIONS(4058),
    [anon_sym__Nonnull] = ACTIONS(4058),
    [anon_sym_nullable] = ACTIONS(4058),
    [anon_sym__Complex] = ACTIONS(4058),
    [anon_sym__Nullable] = ACTIONS(4058),
    [anon_sym__Nullable_result] = ACTIONS(4058),
    [anon_sym__Null_unspecified] = ACTIONS(4058),
    [anon_sym___autoreleasing] = ACTIONS(4058),
    [anon_sym___block] = ACTIONS(4058),
    [anon_sym___bridge] = ACTIONS(4058),
    [anon_sym___bridge_retained] = ACTIONS(4058),
    [anon_sym___bridge_transfer] = ACTIONS(4058),
    [anon_sym___complex] = ACTIONS(4058),
    [anon_sym___const] = ACTIONS(4058),
    [anon_sym___imag] = ACTIONS(4058),
    [anon_sym___kindof] = ACTIONS(4058),
    [anon_sym___nonnull] = ACTIONS(4058),
    [anon_sym___nullable] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4058),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4058),
    [anon_sym___real] = ACTIONS(4058),
    [anon_sym___strong] = ACTIONS(4058),
    [anon_sym___unsafe_unretained] = ACTIONS(4058),
    [anon_sym___unused] = ACTIONS(4058),
    [anon_sym___weak] = ACTIONS(4058),
    [anon_sym_alignas] = ACTIONS(4058),
    [anon_sym__Alignas] = ACTIONS(4058),
    [sym_primitive_type] = ACTIONS(4058),
    [anon_sym_enum] = ACTIONS(4058),
    [anon_sym_struct] = ACTIONS(4058),
    [anon_sym_union] = ACTIONS(4058),
    [anon_sym_in] = ACTIONS(4058),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4058),
    [anon_sym___typeof] = ACTIONS(4058),
    [anon_sym_typeof] = ACTIONS(4058),
    [anon_sym_BOOL] = ACTIONS(4058),
    [anon_sym_IMP] = ACTIONS(4058),
    [anon_sym_SEL] = ACTIONS(4058),
    [anon_sym_Class] = ACTIONS(4058),
    [anon_sym_id] = ACTIONS(4058),
    [anon_sym_out] = ACTIONS(4058),
    [anon_sym_inout] = ACTIONS(4058),
    [anon_sym_bycopy] = ACTIONS(4058),
    [anon_sym_byref] = ACTIONS(4058),
    [anon_sym_oneway] = ACTIONS(4058),
  },
  [STATE(3985)] = {
    [sym_attribute_specifier] = STATE(3963),
    [sym_enumerator_list] = STATE(3844),
    [sym_identifier] = ACTIONS(7581),
    [anon_sym_LPAREN2] = ACTIONS(7583),
    [anon_sym_STAR] = ACTIONS(7583),
    [anon_sym_CARET] = ACTIONS(7583),
    [anon_sym_LT] = ACTIONS(7583),
    [anon_sym___extension__] = ACTIONS(7581),
    [anon_sym___attribute__] = ACTIONS(8830),
    [anon_sym___attribute] = ACTIONS(8830),
    [anon_sym_noreturn] = ACTIONS(7581),
    [anon_sym_LBRACK] = ACTIONS(7583),
    [anon_sym___based] = ACTIONS(7581),
    [anon_sym_LBRACE] = ACTIONS(8925),
    [anon_sym_signed] = ACTIONS(7581),
    [anon_sym_unsigned] = ACTIONS(7581),
    [anon_sym_long] = ACTIONS(7581),
    [anon_sym_short] = ACTIONS(7581),
    [anon_sym_ATautoreleasepool] = ACTIONS(7583),
    [anon_sym_const] = ACTIONS(7581),
    [anon_sym_constexpr] = ACTIONS(7581),
    [anon_sym_volatile] = ACTIONS(7581),
    [anon_sym_restrict] = ACTIONS(7581),
    [anon_sym___restrict__] = ACTIONS(7581),
    [anon_sym__Atomic] = ACTIONS(7581),
    [anon_sym__Noreturn] = ACTIONS(7581),
    [anon_sym__Nonnull] = ACTIONS(7581),
    [anon_sym_nullable] = ACTIONS(7581),
    [anon_sym__Complex] = ACTIONS(7581),
    [anon_sym__Nullable] = ACTIONS(7581),
    [anon_sym__Nullable_result] = ACTIONS(7581),
    [anon_sym__Null_unspecified] = ACTIONS(7581),
    [anon_sym___autoreleasing] = ACTIONS(7581),
    [anon_sym___block] = ACTIONS(7581),
    [anon_sym___bridge] = ACTIONS(7581),
    [anon_sym___bridge_retained] = ACTIONS(7581),
    [anon_sym___bridge_transfer] = ACTIONS(7581),
    [anon_sym___complex] = ACTIONS(7581),
    [anon_sym___const] = ACTIONS(7581),
    [anon_sym___imag] = ACTIONS(7581),
    [anon_sym___kindof] = ACTIONS(7581),
    [anon_sym___nonnull] = ACTIONS(7581),
    [anon_sym___nullable] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7581),
    [anon_sym___real] = ACTIONS(7581),
    [anon_sym___strong] = ACTIONS(7581),
    [anon_sym___unsafe_unretained] = ACTIONS(7581),
    [anon_sym___unused] = ACTIONS(7581),
    [anon_sym___weak] = ACTIONS(7581),
    [anon_sym_alignas] = ACTIONS(7581),
    [anon_sym__Alignas] = ACTIONS(7581),
    [sym_primitive_type] = ACTIONS(7581),
    [anon_sym_enum] = ACTIONS(7581),
    [anon_sym_COLON] = ACTIONS(8928),
    [anon_sym_struct] = ACTIONS(7581),
    [anon_sym_union] = ACTIONS(7581),
    [anon_sym_in] = ACTIONS(7581),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7581),
    [anon_sym___typeof] = ACTIONS(7581),
    [anon_sym_typeof] = ACTIONS(7581),
    [anon_sym_BOOL] = ACTIONS(7581),
    [anon_sym_IMP] = ACTIONS(7581),
    [anon_sym_SEL] = ACTIONS(7581),
    [anon_sym_Class] = ACTIONS(7581),
    [anon_sym_id] = ACTIONS(7581),
    [anon_sym_out] = ACTIONS(7581),
    [anon_sym_inout] = ACTIONS(7581),
    [anon_sym_bycopy] = ACTIONS(7581),
    [anon_sym_byref] = ACTIONS(7581),
    [anon_sym_oneway] = ACTIONS(7581),
  },
  [STATE(3986)] = {
    [sym_identifier] = ACTIONS(6756),
    [anon_sym_COMMA] = ACTIONS(6758),
    [anon_sym_RPAREN] = ACTIONS(6758),
    [anon_sym_LPAREN2] = ACTIONS(6758),
    [anon_sym_STAR] = ACTIONS(6758),
    [anon_sym_CARET] = ACTIONS(6758),
    [anon_sym_GT] = ACTIONS(6758),
    [anon_sym_LT] = ACTIONS(6758),
    [anon_sym___extension__] = ACTIONS(6756),
    [anon_sym___attribute__] = ACTIONS(6756),
    [anon_sym___attribute] = ACTIONS(6756),
    [anon_sym_noreturn] = ACTIONS(6756),
    [anon_sym_LBRACK] = ACTIONS(6758),
    [anon_sym___based] = ACTIONS(6756),
    [anon_sym_LBRACE] = ACTIONS(6758),
    [anon_sym_signed] = ACTIONS(6756),
    [anon_sym_unsigned] = ACTIONS(6756),
    [anon_sym_long] = ACTIONS(6756),
    [anon_sym_short] = ACTIONS(6756),
    [anon_sym_ATautoreleasepool] = ACTIONS(6758),
    [anon_sym_const] = ACTIONS(6756),
    [anon_sym_constexpr] = ACTIONS(6756),
    [anon_sym_volatile] = ACTIONS(6756),
    [anon_sym_restrict] = ACTIONS(6756),
    [anon_sym___restrict__] = ACTIONS(6756),
    [anon_sym__Atomic] = ACTIONS(6756),
    [anon_sym__Noreturn] = ACTIONS(6756),
    [anon_sym__Nonnull] = ACTIONS(6756),
    [anon_sym_nullable] = ACTIONS(6756),
    [anon_sym__Complex] = ACTIONS(6756),
    [anon_sym__Nullable] = ACTIONS(6756),
    [anon_sym__Nullable_result] = ACTIONS(6756),
    [anon_sym__Null_unspecified] = ACTIONS(6756),
    [anon_sym___autoreleasing] = ACTIONS(6756),
    [anon_sym___block] = ACTIONS(6756),
    [anon_sym___bridge] = ACTIONS(6756),
    [anon_sym___bridge_retained] = ACTIONS(6756),
    [anon_sym___bridge_transfer] = ACTIONS(6756),
    [anon_sym___complex] = ACTIONS(6756),
    [anon_sym___const] = ACTIONS(6756),
    [anon_sym___imag] = ACTIONS(6756),
    [anon_sym___kindof] = ACTIONS(6756),
    [anon_sym___nonnull] = ACTIONS(6756),
    [anon_sym___nullable] = ACTIONS(6756),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6756),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6756),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6756),
    [anon_sym___real] = ACTIONS(6756),
    [anon_sym___strong] = ACTIONS(6756),
    [anon_sym___unsafe_unretained] = ACTIONS(6756),
    [anon_sym___unused] = ACTIONS(6756),
    [anon_sym___weak] = ACTIONS(6756),
    [anon_sym_alignas] = ACTIONS(6756),
    [anon_sym__Alignas] = ACTIONS(6756),
    [sym_primitive_type] = ACTIONS(6756),
    [anon_sym_enum] = ACTIONS(6756),
    [anon_sym_struct] = ACTIONS(6756),
    [anon_sym_union] = ACTIONS(6756),
    [anon_sym_in] = ACTIONS(6756),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6756),
    [anon_sym___typeof] = ACTIONS(6756),
    [anon_sym_typeof] = ACTIONS(6756),
    [anon_sym_BOOL] = ACTIONS(6756),
    [anon_sym_IMP] = ACTIONS(6756),
    [anon_sym_SEL] = ACTIONS(6756),
    [anon_sym_Class] = ACTIONS(6756),
    [anon_sym_id] = ACTIONS(6756),
    [anon_sym_out] = ACTIONS(6756),
    [anon_sym_inout] = ACTIONS(6756),
    [anon_sym_bycopy] = ACTIONS(6756),
    [anon_sym_byref] = ACTIONS(6756),
    [anon_sym_oneway] = ACTIONS(6756),
  },
  [STATE(3987)] = {
    [sym_identifier] = ACTIONS(6780),
    [anon_sym_COMMA] = ACTIONS(6782),
    [anon_sym_RPAREN] = ACTIONS(6782),
    [anon_sym_LPAREN2] = ACTIONS(6782),
    [anon_sym_STAR] = ACTIONS(6782),
    [anon_sym_CARET] = ACTIONS(6782),
    [anon_sym_GT] = ACTIONS(6782),
    [anon_sym_LT] = ACTIONS(6782),
    [anon_sym___extension__] = ACTIONS(6780),
    [anon_sym___attribute__] = ACTIONS(6780),
    [anon_sym___attribute] = ACTIONS(6780),
    [anon_sym_noreturn] = ACTIONS(6780),
    [anon_sym_LBRACK] = ACTIONS(6782),
    [anon_sym___based] = ACTIONS(6780),
    [anon_sym_LBRACE] = ACTIONS(6782),
    [anon_sym_signed] = ACTIONS(6780),
    [anon_sym_unsigned] = ACTIONS(6780),
    [anon_sym_long] = ACTIONS(6780),
    [anon_sym_short] = ACTIONS(6780),
    [anon_sym_ATautoreleasepool] = ACTIONS(6782),
    [anon_sym_const] = ACTIONS(6780),
    [anon_sym_constexpr] = ACTIONS(6780),
    [anon_sym_volatile] = ACTIONS(6780),
    [anon_sym_restrict] = ACTIONS(6780),
    [anon_sym___restrict__] = ACTIONS(6780),
    [anon_sym__Atomic] = ACTIONS(6780),
    [anon_sym__Noreturn] = ACTIONS(6780),
    [anon_sym__Nonnull] = ACTIONS(6780),
    [anon_sym_nullable] = ACTIONS(6780),
    [anon_sym__Complex] = ACTIONS(6780),
    [anon_sym__Nullable] = ACTIONS(6780),
    [anon_sym__Nullable_result] = ACTIONS(6780),
    [anon_sym__Null_unspecified] = ACTIONS(6780),
    [anon_sym___autoreleasing] = ACTIONS(6780),
    [anon_sym___block] = ACTIONS(6780),
    [anon_sym___bridge] = ACTIONS(6780),
    [anon_sym___bridge_retained] = ACTIONS(6780),
    [anon_sym___bridge_transfer] = ACTIONS(6780),
    [anon_sym___complex] = ACTIONS(6780),
    [anon_sym___const] = ACTIONS(6780),
    [anon_sym___imag] = ACTIONS(6780),
    [anon_sym___kindof] = ACTIONS(6780),
    [anon_sym___nonnull] = ACTIONS(6780),
    [anon_sym___nullable] = ACTIONS(6780),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6780),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6780),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6780),
    [anon_sym___real] = ACTIONS(6780),
    [anon_sym___strong] = ACTIONS(6780),
    [anon_sym___unsafe_unretained] = ACTIONS(6780),
    [anon_sym___unused] = ACTIONS(6780),
    [anon_sym___weak] = ACTIONS(6780),
    [anon_sym_alignas] = ACTIONS(6780),
    [anon_sym__Alignas] = ACTIONS(6780),
    [sym_primitive_type] = ACTIONS(6780),
    [anon_sym_enum] = ACTIONS(6780),
    [anon_sym_struct] = ACTIONS(6780),
    [anon_sym_union] = ACTIONS(6780),
    [anon_sym_in] = ACTIONS(6780),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6780),
    [anon_sym___typeof] = ACTIONS(6780),
    [anon_sym_typeof] = ACTIONS(6780),
    [anon_sym_BOOL] = ACTIONS(6780),
    [anon_sym_IMP] = ACTIONS(6780),
    [anon_sym_SEL] = ACTIONS(6780),
    [anon_sym_Class] = ACTIONS(6780),
    [anon_sym_id] = ACTIONS(6780),
    [anon_sym_out] = ACTIONS(6780),
    [anon_sym_inout] = ACTIONS(6780),
    [anon_sym_bycopy] = ACTIONS(6780),
    [anon_sym_byref] = ACTIONS(6780),
    [anon_sym_oneway] = ACTIONS(6780),
  },
  [STATE(3988)] = {
    [sym_identifier] = ACTIONS(4266),
    [anon_sym_COMMA] = ACTIONS(4268),
    [anon_sym_RPAREN] = ACTIONS(4268),
    [anon_sym_LPAREN2] = ACTIONS(4268),
    [anon_sym_STAR] = ACTIONS(4268),
    [anon_sym_CARET] = ACTIONS(4268),
    [anon_sym_GT] = ACTIONS(4268),
    [anon_sym_LT] = ACTIONS(4268),
    [anon_sym___extension__] = ACTIONS(4266),
    [anon_sym___attribute__] = ACTIONS(4266),
    [anon_sym___attribute] = ACTIONS(4266),
    [anon_sym_noreturn] = ACTIONS(4266),
    [anon_sym_LBRACK] = ACTIONS(4268),
    [anon_sym___based] = ACTIONS(4266),
    [anon_sym_LBRACE] = ACTIONS(4268),
    [anon_sym_signed] = ACTIONS(4266),
    [anon_sym_unsigned] = ACTIONS(4266),
    [anon_sym_long] = ACTIONS(4266),
    [anon_sym_short] = ACTIONS(4266),
    [anon_sym_ATautoreleasepool] = ACTIONS(4268),
    [anon_sym_const] = ACTIONS(4266),
    [anon_sym_constexpr] = ACTIONS(4266),
    [anon_sym_volatile] = ACTIONS(4266),
    [anon_sym_restrict] = ACTIONS(4266),
    [anon_sym___restrict__] = ACTIONS(4266),
    [anon_sym__Atomic] = ACTIONS(4266),
    [anon_sym__Noreturn] = ACTIONS(4266),
    [anon_sym__Nonnull] = ACTIONS(4266),
    [anon_sym_nullable] = ACTIONS(4266),
    [anon_sym__Complex] = ACTIONS(4266),
    [anon_sym__Nullable] = ACTIONS(4266),
    [anon_sym__Nullable_result] = ACTIONS(4266),
    [anon_sym__Null_unspecified] = ACTIONS(4266),
    [anon_sym___autoreleasing] = ACTIONS(4266),
    [anon_sym___block] = ACTIONS(4266),
    [anon_sym___bridge] = ACTIONS(4266),
    [anon_sym___bridge_retained] = ACTIONS(4266),
    [anon_sym___bridge_transfer] = ACTIONS(4266),
    [anon_sym___complex] = ACTIONS(4266),
    [anon_sym___const] = ACTIONS(4266),
    [anon_sym___imag] = ACTIONS(4266),
    [anon_sym___kindof] = ACTIONS(4266),
    [anon_sym___nonnull] = ACTIONS(4266),
    [anon_sym___nullable] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4266),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4266),
    [anon_sym___real] = ACTIONS(4266),
    [anon_sym___strong] = ACTIONS(4266),
    [anon_sym___unsafe_unretained] = ACTIONS(4266),
    [anon_sym___unused] = ACTIONS(4266),
    [anon_sym___weak] = ACTIONS(4266),
    [anon_sym_alignas] = ACTIONS(4266),
    [anon_sym__Alignas] = ACTIONS(4266),
    [sym_primitive_type] = ACTIONS(4266),
    [anon_sym_enum] = ACTIONS(4266),
    [anon_sym_struct] = ACTIONS(4266),
    [anon_sym_union] = ACTIONS(4266),
    [anon_sym_in] = ACTIONS(4266),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4266),
    [anon_sym___typeof] = ACTIONS(4266),
    [anon_sym_typeof] = ACTIONS(4266),
    [anon_sym_BOOL] = ACTIONS(4266),
    [anon_sym_IMP] = ACTIONS(4266),
    [anon_sym_SEL] = ACTIONS(4266),
    [anon_sym_Class] = ACTIONS(4266),
    [anon_sym_id] = ACTIONS(4266),
    [anon_sym_out] = ACTIONS(4266),
    [anon_sym_inout] = ACTIONS(4266),
    [anon_sym_bycopy] = ACTIONS(4266),
    [anon_sym_byref] = ACTIONS(4266),
    [anon_sym_oneway] = ACTIONS(4266),
  },
  [STATE(3989)] = {
    [sym_identifier] = ACTIONS(4274),
    [anon_sym_COMMA] = ACTIONS(4276),
    [anon_sym_RPAREN] = ACTIONS(4276),
    [anon_sym_LPAREN2] = ACTIONS(4276),
    [anon_sym_STAR] = ACTIONS(4276),
    [anon_sym_CARET] = ACTIONS(4276),
    [anon_sym_GT] = ACTIONS(4276),
    [anon_sym_LT] = ACTIONS(4276),
    [anon_sym___extension__] = ACTIONS(4274),
    [anon_sym___attribute__] = ACTIONS(4274),
    [anon_sym___attribute] = ACTIONS(4274),
    [anon_sym_noreturn] = ACTIONS(4274),
    [anon_sym_LBRACK] = ACTIONS(4276),
    [anon_sym___based] = ACTIONS(4274),
    [anon_sym_LBRACE] = ACTIONS(4276),
    [anon_sym_signed] = ACTIONS(4274),
    [anon_sym_unsigned] = ACTIONS(4274),
    [anon_sym_long] = ACTIONS(4274),
    [anon_sym_short] = ACTIONS(4274),
    [anon_sym_ATautoreleasepool] = ACTIONS(4276),
    [anon_sym_const] = ACTIONS(4274),
    [anon_sym_constexpr] = ACTIONS(4274),
    [anon_sym_volatile] = ACTIONS(4274),
    [anon_sym_restrict] = ACTIONS(4274),
    [anon_sym___restrict__] = ACTIONS(4274),
    [anon_sym__Atomic] = ACTIONS(4274),
    [anon_sym__Noreturn] = ACTIONS(4274),
    [anon_sym__Nonnull] = ACTIONS(4274),
    [anon_sym_nullable] = ACTIONS(4274),
    [anon_sym__Complex] = ACTIONS(4274),
    [anon_sym__Nullable] = ACTIONS(4274),
    [anon_sym__Nullable_result] = ACTIONS(4274),
    [anon_sym__Null_unspecified] = ACTIONS(4274),
    [anon_sym___autoreleasing] = ACTIONS(4274),
    [anon_sym___block] = ACTIONS(4274),
    [anon_sym___bridge] = ACTIONS(4274),
    [anon_sym___bridge_retained] = ACTIONS(4274),
    [anon_sym___bridge_transfer] = ACTIONS(4274),
    [anon_sym___complex] = ACTIONS(4274),
    [anon_sym___const] = ACTIONS(4274),
    [anon_sym___imag] = ACTIONS(4274),
    [anon_sym___kindof] = ACTIONS(4274),
    [anon_sym___nonnull] = ACTIONS(4274),
    [anon_sym___nullable] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4274),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4274),
    [anon_sym___real] = ACTIONS(4274),
    [anon_sym___strong] = ACTIONS(4274),
    [anon_sym___unsafe_unretained] = ACTIONS(4274),
    [anon_sym___unused] = ACTIONS(4274),
    [anon_sym___weak] = ACTIONS(4274),
    [anon_sym_alignas] = ACTIONS(4274),
    [anon_sym__Alignas] = ACTIONS(4274),
    [sym_primitive_type] = ACTIONS(4274),
    [anon_sym_enum] = ACTIONS(4274),
    [anon_sym_struct] = ACTIONS(4274),
    [anon_sym_union] = ACTIONS(4274),
    [anon_sym_in] = ACTIONS(4274),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4274),
    [anon_sym___typeof] = ACTIONS(4274),
    [anon_sym_typeof] = ACTIONS(4274),
    [anon_sym_BOOL] = ACTIONS(4274),
    [anon_sym_IMP] = ACTIONS(4274),
    [anon_sym_SEL] = ACTIONS(4274),
    [anon_sym_Class] = ACTIONS(4274),
    [anon_sym_id] = ACTIONS(4274),
    [anon_sym_out] = ACTIONS(4274),
    [anon_sym_inout] = ACTIONS(4274),
    [anon_sym_bycopy] = ACTIONS(4274),
    [anon_sym_byref] = ACTIONS(4274),
    [anon_sym_oneway] = ACTIONS(4274),
  },
  [STATE(3990)] = {
    [sym_attribute_specifier] = STATE(5696),
    [sym_compound_statement] = STATE(1661),
    [sym_parameter_list] = STATE(5093),
    [sym_identifier] = ACTIONS(8902),
    [anon_sym_LPAREN2] = ACTIONS(8850),
    [anon_sym_STAR] = ACTIONS(8101),
    [anon_sym_CARET] = ACTIONS(8101),
    [anon_sym_LT] = ACTIONS(8101),
    [anon_sym___extension__] = ACTIONS(8902),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(8902),
    [anon_sym_LBRACK] = ACTIONS(8101),
    [anon_sym___based] = ACTIONS(8902),
    [anon_sym_LBRACE] = ACTIONS(8103),
    [anon_sym_signed] = ACTIONS(8902),
    [anon_sym_unsigned] = ACTIONS(8902),
    [anon_sym_long] = ACTIONS(8902),
    [anon_sym_short] = ACTIONS(8902),
    [anon_sym_ATautoreleasepool] = ACTIONS(8105),
    [anon_sym_const] = ACTIONS(8902),
    [anon_sym_constexpr] = ACTIONS(8902),
    [anon_sym_volatile] = ACTIONS(8902),
    [anon_sym_restrict] = ACTIONS(8902),
    [anon_sym___restrict__] = ACTIONS(8902),
    [anon_sym__Atomic] = ACTIONS(8902),
    [anon_sym__Noreturn] = ACTIONS(8902),
    [anon_sym__Nonnull] = ACTIONS(8902),
    [anon_sym_nullable] = ACTIONS(8902),
    [anon_sym__Complex] = ACTIONS(8902),
    [anon_sym__Nullable] = ACTIONS(8902),
    [anon_sym__Nullable_result] = ACTIONS(8902),
    [anon_sym__Null_unspecified] = ACTIONS(8902),
    [anon_sym___autoreleasing] = ACTIONS(8902),
    [anon_sym___block] = ACTIONS(8902),
    [anon_sym___bridge] = ACTIONS(8902),
    [anon_sym___bridge_retained] = ACTIONS(8902),
    [anon_sym___bridge_transfer] = ACTIONS(8902),
    [anon_sym___complex] = ACTIONS(8902),
    [anon_sym___const] = ACTIONS(8902),
    [anon_sym___imag] = ACTIONS(8902),
    [anon_sym___kindof] = ACTIONS(8902),
    [anon_sym___nonnull] = ACTIONS(8902),
    [anon_sym___nullable] = ACTIONS(8902),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8902),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8902),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8902),
    [anon_sym___real] = ACTIONS(8902),
    [anon_sym___strong] = ACTIONS(8902),
    [anon_sym___unsafe_unretained] = ACTIONS(8902),
    [anon_sym___unused] = ACTIONS(8902),
    [anon_sym___weak] = ACTIONS(8902),
    [anon_sym_alignas] = ACTIONS(8902),
    [anon_sym__Alignas] = ACTIONS(8902),
    [sym_primitive_type] = ACTIONS(8902),
    [anon_sym_enum] = ACTIONS(8902),
    [anon_sym_struct] = ACTIONS(8902),
    [anon_sym_union] = ACTIONS(8902),
    [anon_sym_in] = ACTIONS(8902),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8902),
    [anon_sym___typeof] = ACTIONS(8902),
    [anon_sym_typeof] = ACTIONS(8902),
    [anon_sym_BOOL] = ACTIONS(8902),
    [anon_sym_IMP] = ACTIONS(8902),
    [anon_sym_SEL] = ACTIONS(8902),
    [anon_sym_Class] = ACTIONS(8902),
    [anon_sym_id] = ACTIONS(8902),
    [anon_sym_out] = ACTIONS(8902),
    [anon_sym_inout] = ACTIONS(8902),
    [anon_sym_bycopy] = ACTIONS(8902),
    [anon_sym_byref] = ACTIONS(8902),
    [anon_sym_oneway] = ACTIONS(8902),
  },
  [STATE(3991)] = {
    [sym_identifier] = ACTIONS(4050),
    [anon_sym_COMMA] = ACTIONS(4052),
    [anon_sym_RPAREN] = ACTIONS(4052),
    [anon_sym_LPAREN2] = ACTIONS(4052),
    [anon_sym_STAR] = ACTIONS(4052),
    [anon_sym_CARET] = ACTIONS(4052),
    [anon_sym_GT] = ACTIONS(4052),
    [anon_sym_LT] = ACTIONS(4052),
    [anon_sym___extension__] = ACTIONS(4050),
    [anon_sym___attribute__] = ACTIONS(4050),
    [anon_sym___attribute] = ACTIONS(4050),
    [anon_sym_noreturn] = ACTIONS(4050),
    [anon_sym_LBRACK] = ACTIONS(4052),
    [anon_sym___based] = ACTIONS(4050),
    [anon_sym_LBRACE] = ACTIONS(4052),
    [anon_sym_signed] = ACTIONS(4050),
    [anon_sym_unsigned] = ACTIONS(4050),
    [anon_sym_long] = ACTIONS(4050),
    [anon_sym_short] = ACTIONS(4050),
    [anon_sym_ATautoreleasepool] = ACTIONS(4052),
    [anon_sym_const] = ACTIONS(4050),
    [anon_sym_constexpr] = ACTIONS(4050),
    [anon_sym_volatile] = ACTIONS(4050),
    [anon_sym_restrict] = ACTIONS(4050),
    [anon_sym___restrict__] = ACTIONS(4050),
    [anon_sym__Atomic] = ACTIONS(4050),
    [anon_sym__Noreturn] = ACTIONS(4050),
    [anon_sym__Nonnull] = ACTIONS(4050),
    [anon_sym_nullable] = ACTIONS(4050),
    [anon_sym__Complex] = ACTIONS(4050),
    [anon_sym__Nullable] = ACTIONS(4050),
    [anon_sym__Nullable_result] = ACTIONS(4050),
    [anon_sym__Null_unspecified] = ACTIONS(4050),
    [anon_sym___autoreleasing] = ACTIONS(4050),
    [anon_sym___block] = ACTIONS(4050),
    [anon_sym___bridge] = ACTIONS(4050),
    [anon_sym___bridge_retained] = ACTIONS(4050),
    [anon_sym___bridge_transfer] = ACTIONS(4050),
    [anon_sym___complex] = ACTIONS(4050),
    [anon_sym___const] = ACTIONS(4050),
    [anon_sym___imag] = ACTIONS(4050),
    [anon_sym___kindof] = ACTIONS(4050),
    [anon_sym___nonnull] = ACTIONS(4050),
    [anon_sym___nullable] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4050),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4050),
    [anon_sym___real] = ACTIONS(4050),
    [anon_sym___strong] = ACTIONS(4050),
    [anon_sym___unsafe_unretained] = ACTIONS(4050),
    [anon_sym___unused] = ACTIONS(4050),
    [anon_sym___weak] = ACTIONS(4050),
    [anon_sym_alignas] = ACTIONS(4050),
    [anon_sym__Alignas] = ACTIONS(4050),
    [sym_primitive_type] = ACTIONS(4050),
    [anon_sym_enum] = ACTIONS(4050),
    [anon_sym_struct] = ACTIONS(4050),
    [anon_sym_union] = ACTIONS(4050),
    [anon_sym_in] = ACTIONS(4050),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4050),
    [anon_sym___typeof] = ACTIONS(4050),
    [anon_sym_typeof] = ACTIONS(4050),
    [anon_sym_BOOL] = ACTIONS(4050),
    [anon_sym_IMP] = ACTIONS(4050),
    [anon_sym_SEL] = ACTIONS(4050),
    [anon_sym_Class] = ACTIONS(4050),
    [anon_sym_id] = ACTIONS(4050),
    [anon_sym_out] = ACTIONS(4050),
    [anon_sym_inout] = ACTIONS(4050),
    [anon_sym_bycopy] = ACTIONS(4050),
    [anon_sym_byref] = ACTIONS(4050),
    [anon_sym_oneway] = ACTIONS(4050),
  },
  [STATE(3992)] = {
    [sym_identifier] = ACTIONS(6760),
    [anon_sym_COMMA] = ACTIONS(6762),
    [anon_sym_RPAREN] = ACTIONS(6762),
    [anon_sym_LPAREN2] = ACTIONS(6762),
    [anon_sym_STAR] = ACTIONS(6762),
    [anon_sym_CARET] = ACTIONS(6762),
    [anon_sym_GT] = ACTIONS(6762),
    [anon_sym_LT] = ACTIONS(6762),
    [anon_sym___extension__] = ACTIONS(6760),
    [anon_sym___attribute__] = ACTIONS(6760),
    [anon_sym___attribute] = ACTIONS(6760),
    [anon_sym_noreturn] = ACTIONS(6760),
    [anon_sym_LBRACK] = ACTIONS(6762),
    [anon_sym___based] = ACTIONS(6760),
    [anon_sym_LBRACE] = ACTIONS(6762),
    [anon_sym_signed] = ACTIONS(6760),
    [anon_sym_unsigned] = ACTIONS(6760),
    [anon_sym_long] = ACTIONS(6760),
    [anon_sym_short] = ACTIONS(6760),
    [anon_sym_ATautoreleasepool] = ACTIONS(6762),
    [anon_sym_const] = ACTIONS(6760),
    [anon_sym_constexpr] = ACTIONS(6760),
    [anon_sym_volatile] = ACTIONS(6760),
    [anon_sym_restrict] = ACTIONS(6760),
    [anon_sym___restrict__] = ACTIONS(6760),
    [anon_sym__Atomic] = ACTIONS(6760),
    [anon_sym__Noreturn] = ACTIONS(6760),
    [anon_sym__Nonnull] = ACTIONS(6760),
    [anon_sym_nullable] = ACTIONS(6760),
    [anon_sym__Complex] = ACTIONS(6760),
    [anon_sym__Nullable] = ACTIONS(6760),
    [anon_sym__Nullable_result] = ACTIONS(6760),
    [anon_sym__Null_unspecified] = ACTIONS(6760),
    [anon_sym___autoreleasing] = ACTIONS(6760),
    [anon_sym___block] = ACTIONS(6760),
    [anon_sym___bridge] = ACTIONS(6760),
    [anon_sym___bridge_retained] = ACTIONS(6760),
    [anon_sym___bridge_transfer] = ACTIONS(6760),
    [anon_sym___complex] = ACTIONS(6760),
    [anon_sym___const] = ACTIONS(6760),
    [anon_sym___imag] = ACTIONS(6760),
    [anon_sym___kindof] = ACTIONS(6760),
    [anon_sym___nonnull] = ACTIONS(6760),
    [anon_sym___nullable] = ACTIONS(6760),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6760),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6760),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6760),
    [anon_sym___real] = ACTIONS(6760),
    [anon_sym___strong] = ACTIONS(6760),
    [anon_sym___unsafe_unretained] = ACTIONS(6760),
    [anon_sym___unused] = ACTIONS(6760),
    [anon_sym___weak] = ACTIONS(6760),
    [anon_sym_alignas] = ACTIONS(6760),
    [anon_sym__Alignas] = ACTIONS(6760),
    [sym_primitive_type] = ACTIONS(6760),
    [anon_sym_enum] = ACTIONS(6760),
    [anon_sym_struct] = ACTIONS(6760),
    [anon_sym_union] = ACTIONS(6760),
    [anon_sym_in] = ACTIONS(6760),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6760),
    [anon_sym___typeof] = ACTIONS(6760),
    [anon_sym_typeof] = ACTIONS(6760),
    [anon_sym_BOOL] = ACTIONS(6760),
    [anon_sym_IMP] = ACTIONS(6760),
    [anon_sym_SEL] = ACTIONS(6760),
    [anon_sym_Class] = ACTIONS(6760),
    [anon_sym_id] = ACTIONS(6760),
    [anon_sym_out] = ACTIONS(6760),
    [anon_sym_inout] = ACTIONS(6760),
    [anon_sym_bycopy] = ACTIONS(6760),
    [anon_sym_byref] = ACTIONS(6760),
    [anon_sym_oneway] = ACTIONS(6760),
  },
  [STATE(3993)] = {
    [sym_identifier] = ACTIONS(4062),
    [anon_sym_COMMA] = ACTIONS(4064),
    [anon_sym_RPAREN] = ACTIONS(4064),
    [anon_sym_LPAREN2] = ACTIONS(4064),
    [anon_sym_STAR] = ACTIONS(4064),
    [anon_sym_CARET] = ACTIONS(4064),
    [anon_sym_GT] = ACTIONS(4064),
    [anon_sym_LT] = ACTIONS(4064),
    [anon_sym___extension__] = ACTIONS(4062),
    [anon_sym___attribute__] = ACTIONS(4062),
    [anon_sym___attribute] = ACTIONS(4062),
    [anon_sym_noreturn] = ACTIONS(4062),
    [anon_sym_LBRACK] = ACTIONS(4064),
    [anon_sym___based] = ACTIONS(4062),
    [anon_sym_LBRACE] = ACTIONS(4064),
    [anon_sym_signed] = ACTIONS(4062),
    [anon_sym_unsigned] = ACTIONS(4062),
    [anon_sym_long] = ACTIONS(4062),
    [anon_sym_short] = ACTIONS(4062),
    [anon_sym_ATautoreleasepool] = ACTIONS(4064),
    [anon_sym_const] = ACTIONS(4062),
    [anon_sym_constexpr] = ACTIONS(4062),
    [anon_sym_volatile] = ACTIONS(4062),
    [anon_sym_restrict] = ACTIONS(4062),
    [anon_sym___restrict__] = ACTIONS(4062),
    [anon_sym__Atomic] = ACTIONS(4062),
    [anon_sym__Noreturn] = ACTIONS(4062),
    [anon_sym__Nonnull] = ACTIONS(4062),
    [anon_sym_nullable] = ACTIONS(4062),
    [anon_sym__Complex] = ACTIONS(4062),
    [anon_sym__Nullable] = ACTIONS(4062),
    [anon_sym__Nullable_result] = ACTIONS(4062),
    [anon_sym__Null_unspecified] = ACTIONS(4062),
    [anon_sym___autoreleasing] = ACTIONS(4062),
    [anon_sym___block] = ACTIONS(4062),
    [anon_sym___bridge] = ACTIONS(4062),
    [anon_sym___bridge_retained] = ACTIONS(4062),
    [anon_sym___bridge_transfer] = ACTIONS(4062),
    [anon_sym___complex] = ACTIONS(4062),
    [anon_sym___const] = ACTIONS(4062),
    [anon_sym___imag] = ACTIONS(4062),
    [anon_sym___kindof] = ACTIONS(4062),
    [anon_sym___nonnull] = ACTIONS(4062),
    [anon_sym___nullable] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4062),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4062),
    [anon_sym___real] = ACTIONS(4062),
    [anon_sym___strong] = ACTIONS(4062),
    [anon_sym___unsafe_unretained] = ACTIONS(4062),
    [anon_sym___unused] = ACTIONS(4062),
    [anon_sym___weak] = ACTIONS(4062),
    [anon_sym_alignas] = ACTIONS(4062),
    [anon_sym__Alignas] = ACTIONS(4062),
    [sym_primitive_type] = ACTIONS(4062),
    [anon_sym_enum] = ACTIONS(4062),
    [anon_sym_struct] = ACTIONS(4062),
    [anon_sym_union] = ACTIONS(4062),
    [anon_sym_in] = ACTIONS(4062),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4062),
    [anon_sym___typeof] = ACTIONS(4062),
    [anon_sym_typeof] = ACTIONS(4062),
    [anon_sym_BOOL] = ACTIONS(4062),
    [anon_sym_IMP] = ACTIONS(4062),
    [anon_sym_SEL] = ACTIONS(4062),
    [anon_sym_Class] = ACTIONS(4062),
    [anon_sym_id] = ACTIONS(4062),
    [anon_sym_out] = ACTIONS(4062),
    [anon_sym_inout] = ACTIONS(4062),
    [anon_sym_bycopy] = ACTIONS(4062),
    [anon_sym_byref] = ACTIONS(4062),
    [anon_sym_oneway] = ACTIONS(4062),
  },
  [STATE(3994)] = {
    [sym_attribute_specifier] = STATE(3839),
    [sym_enumerator_list] = STATE(3961),
    [sym_identifier] = ACTIONS(7619),
    [anon_sym_LPAREN2] = ACTIONS(7621),
    [anon_sym_STAR] = ACTIONS(7621),
    [anon_sym_CARET] = ACTIONS(7621),
    [anon_sym_LT] = ACTIONS(7621),
    [anon_sym___extension__] = ACTIONS(7619),
    [anon_sym___attribute__] = ACTIONS(8791),
    [anon_sym___attribute] = ACTIONS(8791),
    [anon_sym_noreturn] = ACTIONS(7619),
    [anon_sym_LBRACK] = ACTIONS(7621),
    [anon_sym___based] = ACTIONS(7619),
    [anon_sym_LBRACE] = ACTIONS(8930),
    [anon_sym_signed] = ACTIONS(7619),
    [anon_sym_unsigned] = ACTIONS(7619),
    [anon_sym_long] = ACTIONS(7619),
    [anon_sym_short] = ACTIONS(7619),
    [anon_sym_ATautoreleasepool] = ACTIONS(7621),
    [anon_sym_const] = ACTIONS(7619),
    [anon_sym_constexpr] = ACTIONS(7619),
    [anon_sym_volatile] = ACTIONS(7619),
    [anon_sym_restrict] = ACTIONS(7619),
    [anon_sym___restrict__] = ACTIONS(7619),
    [anon_sym__Atomic] = ACTIONS(7619),
    [anon_sym__Noreturn] = ACTIONS(7619),
    [anon_sym__Nonnull] = ACTIONS(7619),
    [anon_sym_nullable] = ACTIONS(7619),
    [anon_sym__Complex] = ACTIONS(7619),
    [anon_sym__Nullable] = ACTIONS(7619),
    [anon_sym__Nullable_result] = ACTIONS(7619),
    [anon_sym__Null_unspecified] = ACTIONS(7619),
    [anon_sym___autoreleasing] = ACTIONS(7619),
    [anon_sym___block] = ACTIONS(7619),
    [anon_sym___bridge] = ACTIONS(7619),
    [anon_sym___bridge_retained] = ACTIONS(7619),
    [anon_sym___bridge_transfer] = ACTIONS(7619),
    [anon_sym___complex] = ACTIONS(7619),
    [anon_sym___const] = ACTIONS(7619),
    [anon_sym___imag] = ACTIONS(7619),
    [anon_sym___kindof] = ACTIONS(7619),
    [anon_sym___nonnull] = ACTIONS(7619),
    [anon_sym___nullable] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7619),
    [anon_sym___real] = ACTIONS(7619),
    [anon_sym___strong] = ACTIONS(7619),
    [anon_sym___unsafe_unretained] = ACTIONS(7619),
    [anon_sym___unused] = ACTIONS(7619),
    [anon_sym___weak] = ACTIONS(7619),
    [anon_sym_alignas] = ACTIONS(7619),
    [anon_sym__Alignas] = ACTIONS(7619),
    [sym_primitive_type] = ACTIONS(7619),
    [anon_sym_enum] = ACTIONS(7619),
    [anon_sym_COLON] = ACTIONS(8933),
    [anon_sym_struct] = ACTIONS(7619),
    [anon_sym_union] = ACTIONS(7619),
    [anon_sym_in] = ACTIONS(7619),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7619),
    [anon_sym___typeof] = ACTIONS(7619),
    [anon_sym_typeof] = ACTIONS(7619),
    [anon_sym_BOOL] = ACTIONS(7619),
    [anon_sym_IMP] = ACTIONS(7619),
    [anon_sym_SEL] = ACTIONS(7619),
    [anon_sym_Class] = ACTIONS(7619),
    [anon_sym_id] = ACTIONS(7619),
    [anon_sym_out] = ACTIONS(7619),
    [anon_sym_inout] = ACTIONS(7619),
    [anon_sym_bycopy] = ACTIONS(7619),
    [anon_sym_byref] = ACTIONS(7619),
    [anon_sym_oneway] = ACTIONS(7619),
  },
  [STATE(3995)] = {
    [sym_identifier] = ACTIONS(4110),
    [anon_sym_COMMA] = ACTIONS(4112),
    [anon_sym_RPAREN] = ACTIONS(4112),
    [anon_sym_LPAREN2] = ACTIONS(4112),
    [anon_sym_STAR] = ACTIONS(4112),
    [anon_sym_CARET] = ACTIONS(4112),
    [anon_sym_GT] = ACTIONS(4112),
    [anon_sym_LT] = ACTIONS(4112),
    [anon_sym___extension__] = ACTIONS(4110),
    [anon_sym___attribute__] = ACTIONS(4110),
    [anon_sym___attribute] = ACTIONS(4110),
    [anon_sym_noreturn] = ACTIONS(4110),
    [anon_sym_LBRACK] = ACTIONS(4112),
    [anon_sym___based] = ACTIONS(4110),
    [anon_sym_LBRACE] = ACTIONS(4112),
    [anon_sym_signed] = ACTIONS(4110),
    [anon_sym_unsigned] = ACTIONS(4110),
    [anon_sym_long] = ACTIONS(4110),
    [anon_sym_short] = ACTIONS(4110),
    [anon_sym_ATautoreleasepool] = ACTIONS(4112),
    [anon_sym_const] = ACTIONS(4110),
    [anon_sym_constexpr] = ACTIONS(4110),
    [anon_sym_volatile] = ACTIONS(4110),
    [anon_sym_restrict] = ACTIONS(4110),
    [anon_sym___restrict__] = ACTIONS(4110),
    [anon_sym__Atomic] = ACTIONS(4110),
    [anon_sym__Noreturn] = ACTIONS(4110),
    [anon_sym__Nonnull] = ACTIONS(4110),
    [anon_sym_nullable] = ACTIONS(4110),
    [anon_sym__Complex] = ACTIONS(4110),
    [anon_sym__Nullable] = ACTIONS(4110),
    [anon_sym__Nullable_result] = ACTIONS(4110),
    [anon_sym__Null_unspecified] = ACTIONS(4110),
    [anon_sym___autoreleasing] = ACTIONS(4110),
    [anon_sym___block] = ACTIONS(4110),
    [anon_sym___bridge] = ACTIONS(4110),
    [anon_sym___bridge_retained] = ACTIONS(4110),
    [anon_sym___bridge_transfer] = ACTIONS(4110),
    [anon_sym___complex] = ACTIONS(4110),
    [anon_sym___const] = ACTIONS(4110),
    [anon_sym___imag] = ACTIONS(4110),
    [anon_sym___kindof] = ACTIONS(4110),
    [anon_sym___nonnull] = ACTIONS(4110),
    [anon_sym___nullable] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4110),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4110),
    [anon_sym___real] = ACTIONS(4110),
    [anon_sym___strong] = ACTIONS(4110),
    [anon_sym___unsafe_unretained] = ACTIONS(4110),
    [anon_sym___unused] = ACTIONS(4110),
    [anon_sym___weak] = ACTIONS(4110),
    [anon_sym_alignas] = ACTIONS(4110),
    [anon_sym__Alignas] = ACTIONS(4110),
    [sym_primitive_type] = ACTIONS(4110),
    [anon_sym_enum] = ACTIONS(4110),
    [anon_sym_struct] = ACTIONS(4110),
    [anon_sym_union] = ACTIONS(4110),
    [anon_sym_in] = ACTIONS(4110),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4110),
    [anon_sym___typeof] = ACTIONS(4110),
    [anon_sym_typeof] = ACTIONS(4110),
    [anon_sym_BOOL] = ACTIONS(4110),
    [anon_sym_IMP] = ACTIONS(4110),
    [anon_sym_SEL] = ACTIONS(4110),
    [anon_sym_Class] = ACTIONS(4110),
    [anon_sym_id] = ACTIONS(4110),
    [anon_sym_out] = ACTIONS(4110),
    [anon_sym_inout] = ACTIONS(4110),
    [anon_sym_bycopy] = ACTIONS(4110),
    [anon_sym_byref] = ACTIONS(4110),
    [anon_sym_oneway] = ACTIONS(4110),
  },
  [STATE(3996)] = {
    [sym_attribute_specifier] = STATE(5621),
    [sym_compound_statement] = STATE(1471),
    [sym_parameter_list] = STATE(5014),
    [sym_identifier] = ACTIONS(8902),
    [anon_sym_LPAREN2] = ACTIONS(8850),
    [anon_sym_STAR] = ACTIONS(8101),
    [anon_sym_CARET] = ACTIONS(8101),
    [anon_sym_LT] = ACTIONS(8101),
    [anon_sym___extension__] = ACTIONS(8902),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(8902),
    [anon_sym_LBRACK] = ACTIONS(8101),
    [anon_sym___based] = ACTIONS(8902),
    [anon_sym_LBRACE] = ACTIONS(8127),
    [anon_sym_signed] = ACTIONS(8902),
    [anon_sym_unsigned] = ACTIONS(8902),
    [anon_sym_long] = ACTIONS(8902),
    [anon_sym_short] = ACTIONS(8902),
    [anon_sym_ATautoreleasepool] = ACTIONS(8129),
    [anon_sym_const] = ACTIONS(8902),
    [anon_sym_constexpr] = ACTIONS(8902),
    [anon_sym_volatile] = ACTIONS(8902),
    [anon_sym_restrict] = ACTIONS(8902),
    [anon_sym___restrict__] = ACTIONS(8902),
    [anon_sym__Atomic] = ACTIONS(8902),
    [anon_sym__Noreturn] = ACTIONS(8902),
    [anon_sym__Nonnull] = ACTIONS(8902),
    [anon_sym_nullable] = ACTIONS(8902),
    [anon_sym__Complex] = ACTIONS(8902),
    [anon_sym__Nullable] = ACTIONS(8902),
    [anon_sym__Nullable_result] = ACTIONS(8902),
    [anon_sym__Null_unspecified] = ACTIONS(8902),
    [anon_sym___autoreleasing] = ACTIONS(8902),
    [anon_sym___block] = ACTIONS(8902),
    [anon_sym___bridge] = ACTIONS(8902),
    [anon_sym___bridge_retained] = ACTIONS(8902),
    [anon_sym___bridge_transfer] = ACTIONS(8902),
    [anon_sym___complex] = ACTIONS(8902),
    [anon_sym___const] = ACTIONS(8902),
    [anon_sym___imag] = ACTIONS(8902),
    [anon_sym___kindof] = ACTIONS(8902),
    [anon_sym___nonnull] = ACTIONS(8902),
    [anon_sym___nullable] = ACTIONS(8902),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8902),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8902),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8902),
    [anon_sym___real] = ACTIONS(8902),
    [anon_sym___strong] = ACTIONS(8902),
    [anon_sym___unsafe_unretained] = ACTIONS(8902),
    [anon_sym___unused] = ACTIONS(8902),
    [anon_sym___weak] = ACTIONS(8902),
    [anon_sym_alignas] = ACTIONS(8902),
    [anon_sym__Alignas] = ACTIONS(8902),
    [sym_primitive_type] = ACTIONS(8902),
    [anon_sym_enum] = ACTIONS(8902),
    [anon_sym_struct] = ACTIONS(8902),
    [anon_sym_union] = ACTIONS(8902),
    [anon_sym_in] = ACTIONS(8902),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8902),
    [anon_sym___typeof] = ACTIONS(8902),
    [anon_sym_typeof] = ACTIONS(8902),
    [anon_sym_BOOL] = ACTIONS(8902),
    [anon_sym_IMP] = ACTIONS(8902),
    [anon_sym_SEL] = ACTIONS(8902),
    [anon_sym_Class] = ACTIONS(8902),
    [anon_sym_id] = ACTIONS(8902),
    [anon_sym_out] = ACTIONS(8902),
    [anon_sym_inout] = ACTIONS(8902),
    [anon_sym_bycopy] = ACTIONS(8902),
    [anon_sym_byref] = ACTIONS(8902),
    [anon_sym_oneway] = ACTIONS(8902),
  },
  [STATE(3997)] = {
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_RPAREN] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6786),
    [anon_sym_STAR] = ACTIONS(6786),
    [anon_sym_CARET] = ACTIONS(6786),
    [anon_sym_GT] = ACTIONS(6786),
    [anon_sym_LT] = ACTIONS(6786),
    [anon_sym___extension__] = ACTIONS(6784),
    [anon_sym___attribute__] = ACTIONS(6784),
    [anon_sym___attribute] = ACTIONS(6784),
    [anon_sym_noreturn] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6786),
    [anon_sym___based] = ACTIONS(6784),
    [anon_sym_LBRACE] = ACTIONS(6786),
    [anon_sym_signed] = ACTIONS(6784),
    [anon_sym_unsigned] = ACTIONS(6784),
    [anon_sym_long] = ACTIONS(6784),
    [anon_sym_short] = ACTIONS(6784),
    [anon_sym_ATautoreleasepool] = ACTIONS(6786),
    [anon_sym_const] = ACTIONS(6784),
    [anon_sym_constexpr] = ACTIONS(6784),
    [anon_sym_volatile] = ACTIONS(6784),
    [anon_sym_restrict] = ACTIONS(6784),
    [anon_sym___restrict__] = ACTIONS(6784),
    [anon_sym__Atomic] = ACTIONS(6784),
    [anon_sym__Noreturn] = ACTIONS(6784),
    [anon_sym__Nonnull] = ACTIONS(6784),
    [anon_sym_nullable] = ACTIONS(6784),
    [anon_sym__Complex] = ACTIONS(6784),
    [anon_sym__Nullable] = ACTIONS(6784),
    [anon_sym__Nullable_result] = ACTIONS(6784),
    [anon_sym__Null_unspecified] = ACTIONS(6784),
    [anon_sym___autoreleasing] = ACTIONS(6784),
    [anon_sym___block] = ACTIONS(6784),
    [anon_sym___bridge] = ACTIONS(6784),
    [anon_sym___bridge_retained] = ACTIONS(6784),
    [anon_sym___bridge_transfer] = ACTIONS(6784),
    [anon_sym___complex] = ACTIONS(6784),
    [anon_sym___const] = ACTIONS(6784),
    [anon_sym___imag] = ACTIONS(6784),
    [anon_sym___kindof] = ACTIONS(6784),
    [anon_sym___nonnull] = ACTIONS(6784),
    [anon_sym___nullable] = ACTIONS(6784),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(6784),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(6784),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(6784),
    [anon_sym___real] = ACTIONS(6784),
    [anon_sym___strong] = ACTIONS(6784),
    [anon_sym___unsafe_unretained] = ACTIONS(6784),
    [anon_sym___unused] = ACTIONS(6784),
    [anon_sym___weak] = ACTIONS(6784),
    [anon_sym_alignas] = ACTIONS(6784),
    [anon_sym__Alignas] = ACTIONS(6784),
    [sym_primitive_type] = ACTIONS(6784),
    [anon_sym_enum] = ACTIONS(6784),
    [anon_sym_struct] = ACTIONS(6784),
    [anon_sym_union] = ACTIONS(6784),
    [anon_sym_in] = ACTIONS(6784),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(6784),
    [anon_sym___typeof] = ACTIONS(6784),
    [anon_sym_typeof] = ACTIONS(6784),
    [anon_sym_BOOL] = ACTIONS(6784),
    [anon_sym_IMP] = ACTIONS(6784),
    [anon_sym_SEL] = ACTIONS(6784),
    [anon_sym_Class] = ACTIONS(6784),
    [anon_sym_id] = ACTIONS(6784),
    [anon_sym_out] = ACTIONS(6784),
    [anon_sym_inout] = ACTIONS(6784),
    [anon_sym_bycopy] = ACTIONS(6784),
    [anon_sym_byref] = ACTIONS(6784),
    [anon_sym_oneway] = ACTIONS(6784),
  },
  [STATE(3998)] = {
    [sym_identifier] = ACTIONS(4220),
    [anon_sym_COMMA] = ACTIONS(4222),
    [anon_sym_RPAREN] = ACTIONS(4222),
    [anon_sym_LPAREN2] = ACTIONS(4222),
    [anon_sym_STAR] = ACTIONS(4222),
    [anon_sym_CARET] = ACTIONS(4222),
    [anon_sym_GT] = ACTIONS(4222),
    [anon_sym_LT] = ACTIONS(4222),
    [anon_sym___extension__] = ACTIONS(4220),
    [anon_sym___attribute__] = ACTIONS(4220),
    [anon_sym___attribute] = ACTIONS(4220),
    [anon_sym_noreturn] = ACTIONS(4220),
    [anon_sym_LBRACK] = ACTIONS(4222),
    [anon_sym___based] = ACTIONS(4220),
    [anon_sym_LBRACE] = ACTIONS(4222),
    [anon_sym_signed] = ACTIONS(4220),
    [anon_sym_unsigned] = ACTIONS(4220),
    [anon_sym_long] = ACTIONS(4220),
    [anon_sym_short] = ACTIONS(4220),
    [anon_sym_ATautoreleasepool] = ACTIONS(4222),
    [anon_sym_const] = ACTIONS(4220),
    [anon_sym_constexpr] = ACTIONS(4220),
    [anon_sym_volatile] = ACTIONS(4220),
    [anon_sym_restrict] = ACTIONS(4220),
    [anon_sym___restrict__] = ACTIONS(4220),
    [anon_sym__Atomic] = ACTIONS(4220),
    [anon_sym__Noreturn] = ACTIONS(4220),
    [anon_sym__Nonnull] = ACTIONS(4220),
    [anon_sym_nullable] = ACTIONS(4220),
    [anon_sym__Complex] = ACTIONS(4220),
    [anon_sym__Nullable] = ACTIONS(4220),
    [anon_sym__Nullable_result] = ACTIONS(4220),
    [anon_sym__Null_unspecified] = ACTIONS(4220),
    [anon_sym___autoreleasing] = ACTIONS(4220),
    [anon_sym___block] = ACTIONS(4220),
    [anon_sym___bridge] = ACTIONS(4220),
    [anon_sym___bridge_retained] = ACTIONS(4220),
    [anon_sym___bridge_transfer] = ACTIONS(4220),
    [anon_sym___complex] = ACTIONS(4220),
    [anon_sym___const] = ACTIONS(4220),
    [anon_sym___imag] = ACTIONS(4220),
    [anon_sym___kindof] = ACTIONS(4220),
    [anon_sym___nonnull] = ACTIONS(4220),
    [anon_sym___nullable] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4220),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4220),
    [anon_sym___real] = ACTIONS(4220),
    [anon_sym___strong] = ACTIONS(4220),
    [anon_sym___unsafe_unretained] = ACTIONS(4220),
    [anon_sym___unused] = ACTIONS(4220),
    [anon_sym___weak] = ACTIONS(4220),
    [anon_sym_alignas] = ACTIONS(4220),
    [anon_sym__Alignas] = ACTIONS(4220),
    [sym_primitive_type] = ACTIONS(4220),
    [anon_sym_enum] = ACTIONS(4220),
    [anon_sym_struct] = ACTIONS(4220),
    [anon_sym_union] = ACTIONS(4220),
    [anon_sym_in] = ACTIONS(4220),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4220),
    [anon_sym___typeof] = ACTIONS(4220),
    [anon_sym_typeof] = ACTIONS(4220),
    [anon_sym_BOOL] = ACTIONS(4220),
    [anon_sym_IMP] = ACTIONS(4220),
    [anon_sym_SEL] = ACTIONS(4220),
    [anon_sym_Class] = ACTIONS(4220),
    [anon_sym_id] = ACTIONS(4220),
    [anon_sym_out] = ACTIONS(4220),
    [anon_sym_inout] = ACTIONS(4220),
    [anon_sym_bycopy] = ACTIONS(4220),
    [anon_sym_byref] = ACTIONS(4220),
    [anon_sym_oneway] = ACTIONS(4220),
  },
  [STATE(3999)] = {
    [sym_attribute_specifier] = STATE(3860),
    [sym_enumerator_list] = STATE(3890),
    [sym_identifier] = ACTIONS(7632),
    [anon_sym_LPAREN2] = ACTIONS(7634),
    [anon_sym_STAR] = ACTIONS(7634),
    [anon_sym_CARET] = ACTIONS(7634),
    [anon_sym_LT] = ACTIONS(7634),
    [anon_sym___extension__] = ACTIONS(7632),
    [anon_sym___attribute__] = ACTIONS(8835),
    [anon_sym___attribute] = ACTIONS(8835),
    [anon_sym_noreturn] = ACTIONS(7632),
    [anon_sym_LBRACK] = ACTIONS(7634),
    [anon_sym___based] = ACTIONS(7632),
    [anon_sym_LBRACE] = ACTIONS(8935),
    [anon_sym_signed] = ACTIONS(7632),
    [anon_sym_unsigned] = ACTIONS(7632),
    [anon_sym_long] = ACTIONS(7632),
    [anon_sym_short] = ACTIONS(7632),
    [anon_sym_ATautoreleasepool] = ACTIONS(7634),
    [anon_sym_const] = ACTIONS(7632),
    [anon_sym_constexpr] = ACTIONS(7632),
    [anon_sym_volatile] = ACTIONS(7632),
    [anon_sym_restrict] = ACTIONS(7632),
    [anon_sym___restrict__] = ACTIONS(7632),
    [anon_sym__Atomic] = ACTIONS(7632),
    [anon_sym__Noreturn] = ACTIONS(7632),
    [anon_sym__Nonnull] = ACTIONS(7632),
    [anon_sym_nullable] = ACTIONS(7632),
    [anon_sym__Complex] = ACTIONS(7632),
    [anon_sym__Nullable] = ACTIONS(7632),
    [anon_sym__Nullable_result] = ACTIONS(7632),
    [anon_sym__Null_unspecified] = ACTIONS(7632),
    [anon_sym___autoreleasing] = ACTIONS(7632),
    [anon_sym___block] = ACTIONS(7632),
    [anon_sym___bridge] = ACTIONS(7632),
    [anon_sym___bridge_retained] = ACTIONS(7632),
    [anon_sym___bridge_transfer] = ACTIONS(7632),
    [anon_sym___complex] = ACTIONS(7632),
    [anon_sym___const] = ACTIONS(7632),
    [anon_sym___imag] = ACTIONS(7632),
    [anon_sym___kindof] = ACTIONS(7632),
    [anon_sym___nonnull] = ACTIONS(7632),
    [anon_sym___nullable] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7632),
    [anon_sym___real] = ACTIONS(7632),
    [anon_sym___strong] = ACTIONS(7632),
    [anon_sym___unsafe_unretained] = ACTIONS(7632),
    [anon_sym___unused] = ACTIONS(7632),
    [anon_sym___weak] = ACTIONS(7632),
    [anon_sym_alignas] = ACTIONS(7632),
    [anon_sym__Alignas] = ACTIONS(7632),
    [sym_primitive_type] = ACTIONS(7632),
    [anon_sym_enum] = ACTIONS(7632),
    [anon_sym_COLON] = ACTIONS(8938),
    [anon_sym_struct] = ACTIONS(7632),
    [anon_sym_union] = ACTIONS(7632),
    [anon_sym_in] = ACTIONS(7632),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7632),
    [anon_sym___typeof] = ACTIONS(7632),
    [anon_sym_typeof] = ACTIONS(7632),
    [anon_sym_BOOL] = ACTIONS(7632),
    [anon_sym_IMP] = ACTIONS(7632),
    [anon_sym_SEL] = ACTIONS(7632),
    [anon_sym_Class] = ACTIONS(7632),
    [anon_sym_id] = ACTIONS(7632),
    [anon_sym_out] = ACTIONS(7632),
    [anon_sym_inout] = ACTIONS(7632),
    [anon_sym_bycopy] = ACTIONS(7632),
    [anon_sym_byref] = ACTIONS(7632),
    [anon_sym_oneway] = ACTIONS(7632),
  },
  [STATE(4000)] = {
    [sym__type_definition_type] = STATE(4767),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4001)] = {
    [sym__type_definition_type] = STATE(4753),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4002)] = {
    [sym__type_definition_type] = STATE(4754),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4003)] = {
    [sym_attribute_specifier] = STATE(3892),
    [sym_enumerator_list] = STATE(3850),
    [sym_identifier] = ACTIONS(7512),
    [anon_sym_LPAREN2] = ACTIONS(7514),
    [anon_sym_STAR] = ACTIONS(7514),
    [anon_sym_CARET] = ACTIONS(7514),
    [anon_sym_LT] = ACTIONS(7514),
    [anon_sym___extension__] = ACTIONS(7512),
    [anon_sym___attribute__] = ACTIONS(8872),
    [anon_sym___attribute] = ACTIONS(8872),
    [anon_sym_noreturn] = ACTIONS(7512),
    [anon_sym_LBRACK] = ACTIONS(7514),
    [anon_sym___based] = ACTIONS(7512),
    [anon_sym_LBRACE] = ACTIONS(8940),
    [anon_sym_signed] = ACTIONS(7512),
    [anon_sym_unsigned] = ACTIONS(7512),
    [anon_sym_long] = ACTIONS(7512),
    [anon_sym_short] = ACTIONS(7512),
    [anon_sym_ATautoreleasepool] = ACTIONS(7514),
    [anon_sym_const] = ACTIONS(7512),
    [anon_sym_constexpr] = ACTIONS(7512),
    [anon_sym_volatile] = ACTIONS(7512),
    [anon_sym_restrict] = ACTIONS(7512),
    [anon_sym___restrict__] = ACTIONS(7512),
    [anon_sym__Atomic] = ACTIONS(7512),
    [anon_sym__Noreturn] = ACTIONS(7512),
    [anon_sym__Nonnull] = ACTIONS(7512),
    [anon_sym_nullable] = ACTIONS(7512),
    [anon_sym__Complex] = ACTIONS(7512),
    [anon_sym__Nullable] = ACTIONS(7512),
    [anon_sym__Nullable_result] = ACTIONS(7512),
    [anon_sym__Null_unspecified] = ACTIONS(7512),
    [anon_sym___autoreleasing] = ACTIONS(7512),
    [anon_sym___block] = ACTIONS(7512),
    [anon_sym___bridge] = ACTIONS(7512),
    [anon_sym___bridge_retained] = ACTIONS(7512),
    [anon_sym___bridge_transfer] = ACTIONS(7512),
    [anon_sym___complex] = ACTIONS(7512),
    [anon_sym___const] = ACTIONS(7512),
    [anon_sym___imag] = ACTIONS(7512),
    [anon_sym___kindof] = ACTIONS(7512),
    [anon_sym___nonnull] = ACTIONS(7512),
    [anon_sym___nullable] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7512),
    [anon_sym___real] = ACTIONS(7512),
    [anon_sym___strong] = ACTIONS(7512),
    [anon_sym___unsafe_unretained] = ACTIONS(7512),
    [anon_sym___unused] = ACTIONS(7512),
    [anon_sym___weak] = ACTIONS(7512),
    [anon_sym_alignas] = ACTIONS(7512),
    [anon_sym__Alignas] = ACTIONS(7512),
    [sym_primitive_type] = ACTIONS(7512),
    [anon_sym_enum] = ACTIONS(7512),
    [anon_sym_struct] = ACTIONS(7512),
    [anon_sym_union] = ACTIONS(7512),
    [anon_sym_in] = ACTIONS(7512),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7512),
    [anon_sym___typeof] = ACTIONS(7512),
    [anon_sym_typeof] = ACTIONS(7512),
    [anon_sym_BOOL] = ACTIONS(7512),
    [anon_sym_IMP] = ACTIONS(7512),
    [anon_sym_SEL] = ACTIONS(7512),
    [anon_sym_Class] = ACTIONS(7512),
    [anon_sym_id] = ACTIONS(7512),
    [anon_sym_out] = ACTIONS(7512),
    [anon_sym_inout] = ACTIONS(7512),
    [anon_sym_bycopy] = ACTIONS(7512),
    [anon_sym_byref] = ACTIONS(7512),
    [anon_sym_oneway] = ACTIONS(7512),
  },
  [STATE(4004)] = {
    [sym__type_definition_type] = STATE(4778),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4005)] = {
    [sym__type_definition_type] = STATE(4711),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4006)] = {
    [sym__type_definition_type] = STATE(4705),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4007)] = {
    [sym__type_definition_type] = STATE(4762),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4008)] = {
    [sym__type_definition_type] = STATE(4736),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4009)] = {
    [sym__type_definition_type] = STATE(4756),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4010)] = {
    [sym__type_definition_type] = STATE(4744),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4011)] = {
    [sym__type_definition_type] = STATE(4764),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4012)] = {
    [sym_attribute_specifier] = STATE(3894),
    [sym_enumerator_list] = STATE(3851),
    [sym_identifier] = ACTIONS(7521),
    [anon_sym_LPAREN2] = ACTIONS(7523),
    [anon_sym_STAR] = ACTIONS(7523),
    [anon_sym_CARET] = ACTIONS(7523),
    [anon_sym_LT] = ACTIONS(7523),
    [anon_sym___extension__] = ACTIONS(7521),
    [anon_sym___attribute__] = ACTIONS(8875),
    [anon_sym___attribute] = ACTIONS(8875),
    [anon_sym_noreturn] = ACTIONS(7521),
    [anon_sym_LBRACK] = ACTIONS(7523),
    [anon_sym___based] = ACTIONS(7521),
    [anon_sym_LBRACE] = ACTIONS(8943),
    [anon_sym_signed] = ACTIONS(7521),
    [anon_sym_unsigned] = ACTIONS(7521),
    [anon_sym_long] = ACTIONS(7521),
    [anon_sym_short] = ACTIONS(7521),
    [anon_sym_ATautoreleasepool] = ACTIONS(7523),
    [anon_sym_const] = ACTIONS(7521),
    [anon_sym_constexpr] = ACTIONS(7521),
    [anon_sym_volatile] = ACTIONS(7521),
    [anon_sym_restrict] = ACTIONS(7521),
    [anon_sym___restrict__] = ACTIONS(7521),
    [anon_sym__Atomic] = ACTIONS(7521),
    [anon_sym__Noreturn] = ACTIONS(7521),
    [anon_sym__Nonnull] = ACTIONS(7521),
    [anon_sym_nullable] = ACTIONS(7521),
    [anon_sym__Complex] = ACTIONS(7521),
    [anon_sym__Nullable] = ACTIONS(7521),
    [anon_sym__Nullable_result] = ACTIONS(7521),
    [anon_sym__Null_unspecified] = ACTIONS(7521),
    [anon_sym___autoreleasing] = ACTIONS(7521),
    [anon_sym___block] = ACTIONS(7521),
    [anon_sym___bridge] = ACTIONS(7521),
    [anon_sym___bridge_retained] = ACTIONS(7521),
    [anon_sym___bridge_transfer] = ACTIONS(7521),
    [anon_sym___complex] = ACTIONS(7521),
    [anon_sym___const] = ACTIONS(7521),
    [anon_sym___imag] = ACTIONS(7521),
    [anon_sym___kindof] = ACTIONS(7521),
    [anon_sym___nonnull] = ACTIONS(7521),
    [anon_sym___nullable] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7521),
    [anon_sym___real] = ACTIONS(7521),
    [anon_sym___strong] = ACTIONS(7521),
    [anon_sym___unsafe_unretained] = ACTIONS(7521),
    [anon_sym___unused] = ACTIONS(7521),
    [anon_sym___weak] = ACTIONS(7521),
    [anon_sym_alignas] = ACTIONS(7521),
    [anon_sym__Alignas] = ACTIONS(7521),
    [sym_primitive_type] = ACTIONS(7521),
    [anon_sym_enum] = ACTIONS(7521),
    [anon_sym_struct] = ACTIONS(7521),
    [anon_sym_union] = ACTIONS(7521),
    [anon_sym_in] = ACTIONS(7521),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7521),
    [anon_sym___typeof] = ACTIONS(7521),
    [anon_sym_typeof] = ACTIONS(7521),
    [anon_sym_BOOL] = ACTIONS(7521),
    [anon_sym_IMP] = ACTIONS(7521),
    [anon_sym_SEL] = ACTIONS(7521),
    [anon_sym_Class] = ACTIONS(7521),
    [anon_sym_id] = ACTIONS(7521),
    [anon_sym_out] = ACTIONS(7521),
    [anon_sym_inout] = ACTIONS(7521),
    [anon_sym_bycopy] = ACTIONS(7521),
    [anon_sym_byref] = ACTIONS(7521),
    [anon_sym_oneway] = ACTIONS(7521),
  },
  [STATE(4013)] = {
    [sym__type_definition_type] = STATE(4740),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4014)] = {
    [sym__type_definition_type] = STATE(4749),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4015)] = {
    [sym_attribute_specifier] = STATE(3904),
    [sym_enumerator_list] = STATE(3883),
    [sym_identifier] = ACTIONS(7535),
    [anon_sym_LPAREN2] = ACTIONS(7537),
    [anon_sym_STAR] = ACTIONS(7537),
    [anon_sym_CARET] = ACTIONS(7537),
    [anon_sym_LT] = ACTIONS(7537),
    [anon_sym___extension__] = ACTIONS(7535),
    [anon_sym___attribute__] = ACTIONS(8854),
    [anon_sym___attribute] = ACTIONS(8854),
    [anon_sym_noreturn] = ACTIONS(7535),
    [anon_sym_LBRACK] = ACTIONS(7537),
    [anon_sym___based] = ACTIONS(7535),
    [anon_sym_LBRACE] = ACTIONS(8946),
    [anon_sym_signed] = ACTIONS(7535),
    [anon_sym_unsigned] = ACTIONS(7535),
    [anon_sym_long] = ACTIONS(7535),
    [anon_sym_short] = ACTIONS(7535),
    [anon_sym_ATautoreleasepool] = ACTIONS(7537),
    [anon_sym_const] = ACTIONS(7535),
    [anon_sym_constexpr] = ACTIONS(7535),
    [anon_sym_volatile] = ACTIONS(7535),
    [anon_sym_restrict] = ACTIONS(7535),
    [anon_sym___restrict__] = ACTIONS(7535),
    [anon_sym__Atomic] = ACTIONS(7535),
    [anon_sym__Noreturn] = ACTIONS(7535),
    [anon_sym__Nonnull] = ACTIONS(7535),
    [anon_sym_nullable] = ACTIONS(7535),
    [anon_sym__Complex] = ACTIONS(7535),
    [anon_sym__Nullable] = ACTIONS(7535),
    [anon_sym__Nullable_result] = ACTIONS(7535),
    [anon_sym__Null_unspecified] = ACTIONS(7535),
    [anon_sym___autoreleasing] = ACTIONS(7535),
    [anon_sym___block] = ACTIONS(7535),
    [anon_sym___bridge] = ACTIONS(7535),
    [anon_sym___bridge_retained] = ACTIONS(7535),
    [anon_sym___bridge_transfer] = ACTIONS(7535),
    [anon_sym___complex] = ACTIONS(7535),
    [anon_sym___const] = ACTIONS(7535),
    [anon_sym___imag] = ACTIONS(7535),
    [anon_sym___kindof] = ACTIONS(7535),
    [anon_sym___nonnull] = ACTIONS(7535),
    [anon_sym___nullable] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7535),
    [anon_sym___real] = ACTIONS(7535),
    [anon_sym___strong] = ACTIONS(7535),
    [anon_sym___unsafe_unretained] = ACTIONS(7535),
    [anon_sym___unused] = ACTIONS(7535),
    [anon_sym___weak] = ACTIONS(7535),
    [anon_sym_alignas] = ACTIONS(7535),
    [anon_sym__Alignas] = ACTIONS(7535),
    [sym_primitive_type] = ACTIONS(7535),
    [anon_sym_enum] = ACTIONS(7535),
    [anon_sym_struct] = ACTIONS(7535),
    [anon_sym_union] = ACTIONS(7535),
    [anon_sym_in] = ACTIONS(7535),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7535),
    [anon_sym___typeof] = ACTIONS(7535),
    [anon_sym_typeof] = ACTIONS(7535),
    [anon_sym_BOOL] = ACTIONS(7535),
    [anon_sym_IMP] = ACTIONS(7535),
    [anon_sym_SEL] = ACTIONS(7535),
    [anon_sym_Class] = ACTIONS(7535),
    [anon_sym_id] = ACTIONS(7535),
    [anon_sym_out] = ACTIONS(7535),
    [anon_sym_inout] = ACTIONS(7535),
    [anon_sym_bycopy] = ACTIONS(7535),
    [anon_sym_byref] = ACTIONS(7535),
    [anon_sym_oneway] = ACTIONS(7535),
  },
  [STATE(4016)] = {
    [sym__type_definition_type] = STATE(4784),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4017)] = {
    [sym__type_definition_type] = STATE(4715),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4018)] = {
    [sym__type_definition_type] = STATE(4770),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4019)] = {
    [sym_attribute_specifier] = STATE(3957),
    [sym_enumerator_list] = STATE(3909),
    [sym_identifier] = ACTIONS(7561),
    [anon_sym_LPAREN2] = ACTIONS(7563),
    [anon_sym_STAR] = ACTIONS(7563),
    [anon_sym_CARET] = ACTIONS(7563),
    [anon_sym_LT] = ACTIONS(7563),
    [anon_sym___extension__] = ACTIONS(7561),
    [anon_sym___attribute__] = ACTIONS(8908),
    [anon_sym___attribute] = ACTIONS(8908),
    [anon_sym_noreturn] = ACTIONS(7561),
    [anon_sym_LBRACK] = ACTIONS(7563),
    [anon_sym___based] = ACTIONS(7561),
    [anon_sym_LBRACE] = ACTIONS(8949),
    [anon_sym_signed] = ACTIONS(7561),
    [anon_sym_unsigned] = ACTIONS(7561),
    [anon_sym_long] = ACTIONS(7561),
    [anon_sym_short] = ACTIONS(7561),
    [anon_sym_ATautoreleasepool] = ACTIONS(7563),
    [anon_sym_const] = ACTIONS(7561),
    [anon_sym_constexpr] = ACTIONS(7561),
    [anon_sym_volatile] = ACTIONS(7561),
    [anon_sym_restrict] = ACTIONS(7561),
    [anon_sym___restrict__] = ACTIONS(7561),
    [anon_sym__Atomic] = ACTIONS(7561),
    [anon_sym__Noreturn] = ACTIONS(7561),
    [anon_sym__Nonnull] = ACTIONS(7561),
    [anon_sym_nullable] = ACTIONS(7561),
    [anon_sym__Complex] = ACTIONS(7561),
    [anon_sym__Nullable] = ACTIONS(7561),
    [anon_sym__Nullable_result] = ACTIONS(7561),
    [anon_sym__Null_unspecified] = ACTIONS(7561),
    [anon_sym___autoreleasing] = ACTIONS(7561),
    [anon_sym___block] = ACTIONS(7561),
    [anon_sym___bridge] = ACTIONS(7561),
    [anon_sym___bridge_retained] = ACTIONS(7561),
    [anon_sym___bridge_transfer] = ACTIONS(7561),
    [anon_sym___complex] = ACTIONS(7561),
    [anon_sym___const] = ACTIONS(7561),
    [anon_sym___imag] = ACTIONS(7561),
    [anon_sym___kindof] = ACTIONS(7561),
    [anon_sym___nonnull] = ACTIONS(7561),
    [anon_sym___nullable] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7561),
    [anon_sym___real] = ACTIONS(7561),
    [anon_sym___strong] = ACTIONS(7561),
    [anon_sym___unsafe_unretained] = ACTIONS(7561),
    [anon_sym___unused] = ACTIONS(7561),
    [anon_sym___weak] = ACTIONS(7561),
    [anon_sym_alignas] = ACTIONS(7561),
    [anon_sym__Alignas] = ACTIONS(7561),
    [sym_primitive_type] = ACTIONS(7561),
    [anon_sym_enum] = ACTIONS(7561),
    [anon_sym_struct] = ACTIONS(7561),
    [anon_sym_union] = ACTIONS(7561),
    [anon_sym_in] = ACTIONS(7561),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7561),
    [anon_sym___typeof] = ACTIONS(7561),
    [anon_sym_typeof] = ACTIONS(7561),
    [anon_sym_BOOL] = ACTIONS(7561),
    [anon_sym_IMP] = ACTIONS(7561),
    [anon_sym_SEL] = ACTIONS(7561),
    [anon_sym_Class] = ACTIONS(7561),
    [anon_sym_id] = ACTIONS(7561),
    [anon_sym_out] = ACTIONS(7561),
    [anon_sym_inout] = ACTIONS(7561),
    [anon_sym_bycopy] = ACTIONS(7561),
    [anon_sym_byref] = ACTIONS(7561),
    [anon_sym_oneway] = ACTIONS(7561),
  },
  [STATE(4020)] = {
    [sym__type_definition_type] = STATE(4751),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4021)] = {
    [sym__type_definition_type] = STATE(4746),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4022)] = {
    [sym__type_definition_type] = STATE(4719),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4023)] = {
    [sym__type_definition_type] = STATE(4698),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4024)] = {
    [sym__type_definition_type] = STATE(4772),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4025)] = {
    [sym__type_definition_type] = STATE(4777),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4026)] = {
    [sym__type_definition_type] = STATE(4726),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4027)] = {
    [sym__type_definition_type] = STATE(4783),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4028)] = {
    [sym_attribute_specifier] = STATE(3942),
    [sym_enumerator_list] = STATE(3885),
    [sym_identifier] = ACTIONS(7548),
    [anon_sym_LPAREN2] = ACTIONS(7550),
    [anon_sym_STAR] = ACTIONS(7550),
    [anon_sym_CARET] = ACTIONS(7550),
    [anon_sym_LT] = ACTIONS(7550),
    [anon_sym___extension__] = ACTIONS(7548),
    [anon_sym___attribute__] = ACTIONS(8857),
    [anon_sym___attribute] = ACTIONS(8857),
    [anon_sym_noreturn] = ACTIONS(7548),
    [anon_sym_LBRACK] = ACTIONS(7550),
    [anon_sym___based] = ACTIONS(7548),
    [anon_sym_LBRACE] = ACTIONS(8952),
    [anon_sym_signed] = ACTIONS(7548),
    [anon_sym_unsigned] = ACTIONS(7548),
    [anon_sym_long] = ACTIONS(7548),
    [anon_sym_short] = ACTIONS(7548),
    [anon_sym_ATautoreleasepool] = ACTIONS(7550),
    [anon_sym_const] = ACTIONS(7548),
    [anon_sym_constexpr] = ACTIONS(7548),
    [anon_sym_volatile] = ACTIONS(7548),
    [anon_sym_restrict] = ACTIONS(7548),
    [anon_sym___restrict__] = ACTIONS(7548),
    [anon_sym__Atomic] = ACTIONS(7548),
    [anon_sym__Noreturn] = ACTIONS(7548),
    [anon_sym__Nonnull] = ACTIONS(7548),
    [anon_sym_nullable] = ACTIONS(7548),
    [anon_sym__Complex] = ACTIONS(7548),
    [anon_sym__Nullable] = ACTIONS(7548),
    [anon_sym__Nullable_result] = ACTIONS(7548),
    [anon_sym__Null_unspecified] = ACTIONS(7548),
    [anon_sym___autoreleasing] = ACTIONS(7548),
    [anon_sym___block] = ACTIONS(7548),
    [anon_sym___bridge] = ACTIONS(7548),
    [anon_sym___bridge_retained] = ACTIONS(7548),
    [anon_sym___bridge_transfer] = ACTIONS(7548),
    [anon_sym___complex] = ACTIONS(7548),
    [anon_sym___const] = ACTIONS(7548),
    [anon_sym___imag] = ACTIONS(7548),
    [anon_sym___kindof] = ACTIONS(7548),
    [anon_sym___nonnull] = ACTIONS(7548),
    [anon_sym___nullable] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7548),
    [anon_sym___real] = ACTIONS(7548),
    [anon_sym___strong] = ACTIONS(7548),
    [anon_sym___unsafe_unretained] = ACTIONS(7548),
    [anon_sym___unused] = ACTIONS(7548),
    [anon_sym___weak] = ACTIONS(7548),
    [anon_sym_alignas] = ACTIONS(7548),
    [anon_sym__Alignas] = ACTIONS(7548),
    [sym_primitive_type] = ACTIONS(7548),
    [anon_sym_enum] = ACTIONS(7548),
    [anon_sym_struct] = ACTIONS(7548),
    [anon_sym_union] = ACTIONS(7548),
    [anon_sym_in] = ACTIONS(7548),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7548),
    [anon_sym___typeof] = ACTIONS(7548),
    [anon_sym_typeof] = ACTIONS(7548),
    [anon_sym_BOOL] = ACTIONS(7548),
    [anon_sym_IMP] = ACTIONS(7548),
    [anon_sym_SEL] = ACTIONS(7548),
    [anon_sym_Class] = ACTIONS(7548),
    [anon_sym_id] = ACTIONS(7548),
    [anon_sym_out] = ACTIONS(7548),
    [anon_sym_inout] = ACTIONS(7548),
    [anon_sym_bycopy] = ACTIONS(7548),
    [anon_sym_byref] = ACTIONS(7548),
    [anon_sym_oneway] = ACTIONS(7548),
  },
  [STATE(4029)] = {
    [sym__type_definition_type] = STATE(4689),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4030)] = {
    [sym__type_definition_type] = STATE(4725),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4031)] = {
    [sym__type_definition_type] = STATE(4707),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4032)] = {
    [sym__type_definition_type] = STATE(4691),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4033)] = {
    [sym__type_definition_type] = STATE(4721),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4034)] = {
    [sym_attribute_specifier] = STATE(3959),
    [sym_enumerator_list] = STATE(3917),
    [sym_identifier] = ACTIONS(7568),
    [anon_sym_LPAREN2] = ACTIONS(7570),
    [anon_sym_STAR] = ACTIONS(7570),
    [anon_sym_CARET] = ACTIONS(7570),
    [anon_sym_LT] = ACTIONS(7570),
    [anon_sym___extension__] = ACTIONS(7568),
    [anon_sym___attribute__] = ACTIONS(8911),
    [anon_sym___attribute] = ACTIONS(8911),
    [anon_sym_noreturn] = ACTIONS(7568),
    [anon_sym_LBRACK] = ACTIONS(7570),
    [anon_sym___based] = ACTIONS(7568),
    [anon_sym_LBRACE] = ACTIONS(8955),
    [anon_sym_signed] = ACTIONS(7568),
    [anon_sym_unsigned] = ACTIONS(7568),
    [anon_sym_long] = ACTIONS(7568),
    [anon_sym_short] = ACTIONS(7568),
    [anon_sym_ATautoreleasepool] = ACTIONS(7570),
    [anon_sym_const] = ACTIONS(7568),
    [anon_sym_constexpr] = ACTIONS(7568),
    [anon_sym_volatile] = ACTIONS(7568),
    [anon_sym_restrict] = ACTIONS(7568),
    [anon_sym___restrict__] = ACTIONS(7568),
    [anon_sym__Atomic] = ACTIONS(7568),
    [anon_sym__Noreturn] = ACTIONS(7568),
    [anon_sym__Nonnull] = ACTIONS(7568),
    [anon_sym_nullable] = ACTIONS(7568),
    [anon_sym__Complex] = ACTIONS(7568),
    [anon_sym__Nullable] = ACTIONS(7568),
    [anon_sym__Nullable_result] = ACTIONS(7568),
    [anon_sym__Null_unspecified] = ACTIONS(7568),
    [anon_sym___autoreleasing] = ACTIONS(7568),
    [anon_sym___block] = ACTIONS(7568),
    [anon_sym___bridge] = ACTIONS(7568),
    [anon_sym___bridge_retained] = ACTIONS(7568),
    [anon_sym___bridge_transfer] = ACTIONS(7568),
    [anon_sym___complex] = ACTIONS(7568),
    [anon_sym___const] = ACTIONS(7568),
    [anon_sym___imag] = ACTIONS(7568),
    [anon_sym___kindof] = ACTIONS(7568),
    [anon_sym___nonnull] = ACTIONS(7568),
    [anon_sym___nullable] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7568),
    [anon_sym___real] = ACTIONS(7568),
    [anon_sym___strong] = ACTIONS(7568),
    [anon_sym___unsafe_unretained] = ACTIONS(7568),
    [anon_sym___unused] = ACTIONS(7568),
    [anon_sym___weak] = ACTIONS(7568),
    [anon_sym_alignas] = ACTIONS(7568),
    [anon_sym__Alignas] = ACTIONS(7568),
    [sym_primitive_type] = ACTIONS(7568),
    [anon_sym_enum] = ACTIONS(7568),
    [anon_sym_struct] = ACTIONS(7568),
    [anon_sym_union] = ACTIONS(7568),
    [anon_sym_in] = ACTIONS(7568),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7568),
    [anon_sym___typeof] = ACTIONS(7568),
    [anon_sym_typeof] = ACTIONS(7568),
    [anon_sym_BOOL] = ACTIONS(7568),
    [anon_sym_IMP] = ACTIONS(7568),
    [anon_sym_SEL] = ACTIONS(7568),
    [anon_sym_Class] = ACTIONS(7568),
    [anon_sym_id] = ACTIONS(7568),
    [anon_sym_out] = ACTIONS(7568),
    [anon_sym_inout] = ACTIONS(7568),
    [anon_sym_bycopy] = ACTIONS(7568),
    [anon_sym_byref] = ACTIONS(7568),
    [anon_sym_oneway] = ACTIONS(7568),
  },
  [STATE(4035)] = {
    [sym__type_definition_type] = STATE(4729),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4036)] = {
    [sym__type_definition_type] = STATE(4733),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4037)] = {
    [sym__type_definition_type] = STATE(4695),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4038)] = {
    [sym__type_definition_type] = STATE(4780),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4039)] = {
    [sym__type_definition_type] = STATE(4697),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4040)] = {
    [sym__type_definition_type] = STATE(4738),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4041)] = {
    [sym__type_definition_type] = STATE(4699),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4042)] = {
    [sym__type_definition_type] = STATE(4709),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4043)] = {
    [sym__type_definition_type] = STATE(4731),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4044)] = {
    [sym__type_definition_type] = STATE(4701),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4045)] = {
    [sym__type_definition_type] = STATE(4717),
    [sym_attribute_specifier] = STATE(4553),
    [sym_type_qualifier] = STATE(4066),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4179),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4066),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4046)] = {
    [sym_attribute_specifier] = STATE(3904),
    [sym_enumerator_list] = STATE(4078),
    [sym_identifier] = ACTIONS(7535),
    [anon_sym_LPAREN2] = ACTIONS(7537),
    [anon_sym_STAR] = ACTIONS(7537),
    [anon_sym_CARET] = ACTIONS(7537),
    [anon_sym___extension__] = ACTIONS(7535),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7535),
    [anon_sym_LBRACK] = ACTIONS(7537),
    [anon_sym___based] = ACTIONS(7535),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7535),
    [anon_sym_unsigned] = ACTIONS(7535),
    [anon_sym_long] = ACTIONS(7535),
    [anon_sym_short] = ACTIONS(7535),
    [anon_sym_const] = ACTIONS(7535),
    [anon_sym_constexpr] = ACTIONS(7535),
    [anon_sym_volatile] = ACTIONS(7535),
    [anon_sym_restrict] = ACTIONS(7535),
    [anon_sym___restrict__] = ACTIONS(7535),
    [anon_sym__Atomic] = ACTIONS(7535),
    [anon_sym__Noreturn] = ACTIONS(7535),
    [anon_sym__Nonnull] = ACTIONS(7535),
    [anon_sym_nullable] = ACTIONS(7535),
    [anon_sym__Complex] = ACTIONS(7535),
    [anon_sym__Nullable] = ACTIONS(7535),
    [anon_sym__Nullable_result] = ACTIONS(7535),
    [anon_sym__Null_unspecified] = ACTIONS(7535),
    [anon_sym___autoreleasing] = ACTIONS(7535),
    [anon_sym___block] = ACTIONS(7535),
    [anon_sym___bridge] = ACTIONS(7535),
    [anon_sym___bridge_retained] = ACTIONS(7535),
    [anon_sym___bridge_transfer] = ACTIONS(7535),
    [anon_sym___complex] = ACTIONS(7535),
    [anon_sym___const] = ACTIONS(7535),
    [anon_sym___imag] = ACTIONS(7535),
    [anon_sym___kindof] = ACTIONS(7535),
    [anon_sym___nonnull] = ACTIONS(7535),
    [anon_sym___nullable] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7535),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7535),
    [anon_sym___real] = ACTIONS(7535),
    [anon_sym___strong] = ACTIONS(7535),
    [anon_sym___unsafe_unretained] = ACTIONS(7535),
    [anon_sym___unused] = ACTIONS(7535),
    [anon_sym___weak] = ACTIONS(7535),
    [anon_sym_alignas] = ACTIONS(7535),
    [anon_sym__Alignas] = ACTIONS(7535),
    [sym_primitive_type] = ACTIONS(7535),
    [anon_sym_enum] = ACTIONS(7535),
    [anon_sym_COLON] = ACTIONS(7537),
    [anon_sym_struct] = ACTIONS(7535),
    [anon_sym_union] = ACTIONS(7535),
    [anon_sym_in] = ACTIONS(7535),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7535),
    [anon_sym___typeof] = ACTIONS(7535),
    [anon_sym_typeof] = ACTIONS(7535),
    [anon_sym_BOOL] = ACTIONS(7535),
    [anon_sym_IMP] = ACTIONS(7535),
    [anon_sym_SEL] = ACTIONS(7535),
    [anon_sym_Class] = ACTIONS(7535),
    [anon_sym_id] = ACTIONS(7535),
    [anon_sym_out] = ACTIONS(7535),
    [anon_sym_inout] = ACTIONS(7535),
    [anon_sym_bycopy] = ACTIONS(7535),
    [anon_sym_byref] = ACTIONS(7535),
    [anon_sym_oneway] = ACTIONS(7535),
  },
  [STATE(4047)] = {
    [sym_identifier] = ACTIONS(8958),
    [aux_sym_preproc_if_token1] = ACTIONS(8958),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(8235),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(8235),
    [anon_sym___extension__] = ACTIONS(8958),
    [anon_sym___attribute__] = ACTIONS(8958),
    [anon_sym___attribute] = ACTIONS(8958),
    [anon_sym_noreturn] = ACTIONS(8958),
    [anon_sym_LBRACK] = ACTIONS(8235),
    [anon_sym_RBRACE] = ACTIONS(8235),
    [anon_sym_signed] = ACTIONS(8958),
    [anon_sym_unsigned] = ACTIONS(8958),
    [anon_sym_long] = ACTIONS(8958),
    [anon_sym_short] = ACTIONS(8958),
    [anon_sym_const] = ACTIONS(8958),
    [anon_sym_constexpr] = ACTIONS(8958),
    [anon_sym_volatile] = ACTIONS(8958),
    [anon_sym_restrict] = ACTIONS(8958),
    [anon_sym___restrict__] = ACTIONS(8958),
    [anon_sym__Atomic] = ACTIONS(8958),
    [anon_sym__Noreturn] = ACTIONS(8958),
    [anon_sym__Nonnull] = ACTIONS(8958),
    [anon_sym_nullable] = ACTIONS(8958),
    [anon_sym__Complex] = ACTIONS(8958),
    [anon_sym__Nullable] = ACTIONS(8958),
    [anon_sym__Nullable_result] = ACTIONS(8958),
    [anon_sym__Null_unspecified] = ACTIONS(8958),
    [anon_sym___autoreleasing] = ACTIONS(8958),
    [anon_sym___block] = ACTIONS(8958),
    [anon_sym___bridge] = ACTIONS(8958),
    [anon_sym___bridge_retained] = ACTIONS(8958),
    [anon_sym___bridge_transfer] = ACTIONS(8958),
    [anon_sym___complex] = ACTIONS(8958),
    [anon_sym___const] = ACTIONS(8958),
    [anon_sym___imag] = ACTIONS(8958),
    [anon_sym___kindof] = ACTIONS(8958),
    [anon_sym___nonnull] = ACTIONS(8958),
    [anon_sym___nullable] = ACTIONS(8958),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8958),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8958),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8958),
    [anon_sym___real] = ACTIONS(8958),
    [anon_sym___strong] = ACTIONS(8958),
    [anon_sym___unsafe_unretained] = ACTIONS(8958),
    [anon_sym___unused] = ACTIONS(8958),
    [anon_sym___weak] = ACTIONS(8958),
    [anon_sym_alignas] = ACTIONS(8958),
    [anon_sym__Alignas] = ACTIONS(8958),
    [sym_primitive_type] = ACTIONS(8958),
    [anon_sym_enum] = ACTIONS(8958),
    [anon_sym_struct] = ACTIONS(8958),
    [anon_sym_union] = ACTIONS(8958),
    [anon_sym_in] = ACTIONS(8958),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8958),
    [anon_sym___typeof] = ACTIONS(8958),
    [anon_sym_typeof] = ACTIONS(8958),
    [anon_sym_ATprivate] = ACTIONS(8235),
    [anon_sym_ATprotected] = ACTIONS(8235),
    [anon_sym_ATpackage] = ACTIONS(8235),
    [anon_sym_ATpublic] = ACTIONS(8235),
    [anon_sym_BOOL] = ACTIONS(8958),
    [anon_sym_IMP] = ACTIONS(8958),
    [anon_sym_SEL] = ACTIONS(8958),
    [anon_sym_Class] = ACTIONS(8958),
    [anon_sym_id] = ACTIONS(8958),
    [anon_sym_out] = ACTIONS(8958),
    [anon_sym_inout] = ACTIONS(8958),
    [anon_sym_bycopy] = ACTIONS(8958),
    [anon_sym_byref] = ACTIONS(8958),
    [anon_sym_oneway] = ACTIONS(8958),
  },
  [STATE(4048)] = {
    [sym_identifier] = ACTIONS(2768),
    [aux_sym_preproc_if_token1] = ACTIONS(2768),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2770),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2770),
    [anon_sym___extension__] = ACTIONS(2768),
    [anon_sym___attribute__] = ACTIONS(2768),
    [anon_sym___attribute] = ACTIONS(2768),
    [anon_sym_noreturn] = ACTIONS(2768),
    [anon_sym_LBRACK] = ACTIONS(2770),
    [anon_sym_RBRACE] = ACTIONS(2770),
    [anon_sym_signed] = ACTIONS(2768),
    [anon_sym_unsigned] = ACTIONS(2768),
    [anon_sym_long] = ACTIONS(2768),
    [anon_sym_short] = ACTIONS(2768),
    [anon_sym_const] = ACTIONS(2768),
    [anon_sym_constexpr] = ACTIONS(2768),
    [anon_sym_volatile] = ACTIONS(2768),
    [anon_sym_restrict] = ACTIONS(2768),
    [anon_sym___restrict__] = ACTIONS(2768),
    [anon_sym__Atomic] = ACTIONS(2768),
    [anon_sym__Noreturn] = ACTIONS(2768),
    [anon_sym__Nonnull] = ACTIONS(2768),
    [anon_sym_nullable] = ACTIONS(2768),
    [anon_sym__Complex] = ACTIONS(2768),
    [anon_sym__Nullable] = ACTIONS(2768),
    [anon_sym__Nullable_result] = ACTIONS(2768),
    [anon_sym__Null_unspecified] = ACTIONS(2768),
    [anon_sym___autoreleasing] = ACTIONS(2768),
    [anon_sym___block] = ACTIONS(2768),
    [anon_sym___bridge] = ACTIONS(2768),
    [anon_sym___bridge_retained] = ACTIONS(2768),
    [anon_sym___bridge_transfer] = ACTIONS(2768),
    [anon_sym___complex] = ACTIONS(2768),
    [anon_sym___const] = ACTIONS(2768),
    [anon_sym___imag] = ACTIONS(2768),
    [anon_sym___kindof] = ACTIONS(2768),
    [anon_sym___nonnull] = ACTIONS(2768),
    [anon_sym___nullable] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2768),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2768),
    [anon_sym___real] = ACTIONS(2768),
    [anon_sym___strong] = ACTIONS(2768),
    [anon_sym___unsafe_unretained] = ACTIONS(2768),
    [anon_sym___unused] = ACTIONS(2768),
    [anon_sym___weak] = ACTIONS(2768),
    [anon_sym_alignas] = ACTIONS(2768),
    [anon_sym__Alignas] = ACTIONS(2768),
    [sym_primitive_type] = ACTIONS(2768),
    [anon_sym_enum] = ACTIONS(2768),
    [anon_sym_struct] = ACTIONS(2768),
    [anon_sym_union] = ACTIONS(2768),
    [anon_sym_in] = ACTIONS(2768),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2768),
    [anon_sym___typeof] = ACTIONS(2768),
    [anon_sym_typeof] = ACTIONS(2768),
    [anon_sym_ATprivate] = ACTIONS(2770),
    [anon_sym_ATprotected] = ACTIONS(2770),
    [anon_sym_ATpackage] = ACTIONS(2770),
    [anon_sym_ATpublic] = ACTIONS(2770),
    [anon_sym_BOOL] = ACTIONS(2768),
    [anon_sym_IMP] = ACTIONS(2768),
    [anon_sym_SEL] = ACTIONS(2768),
    [anon_sym_Class] = ACTIONS(2768),
    [anon_sym_id] = ACTIONS(2768),
    [anon_sym_out] = ACTIONS(2768),
    [anon_sym_inout] = ACTIONS(2768),
    [anon_sym_bycopy] = ACTIONS(2768),
    [anon_sym_byref] = ACTIONS(2768),
    [anon_sym_oneway] = ACTIONS(2768),
  },
  [STATE(4049)] = {
    [sym_identifier] = ACTIONS(2772),
    [aux_sym_preproc_if_token1] = ACTIONS(2772),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2774),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2774),
    [anon_sym___extension__] = ACTIONS(2772),
    [anon_sym___attribute__] = ACTIONS(2772),
    [anon_sym___attribute] = ACTIONS(2772),
    [anon_sym_noreturn] = ACTIONS(2772),
    [anon_sym_LBRACK] = ACTIONS(2774),
    [anon_sym_RBRACE] = ACTIONS(2774),
    [anon_sym_signed] = ACTIONS(2772),
    [anon_sym_unsigned] = ACTIONS(2772),
    [anon_sym_long] = ACTIONS(2772),
    [anon_sym_short] = ACTIONS(2772),
    [anon_sym_const] = ACTIONS(2772),
    [anon_sym_constexpr] = ACTIONS(2772),
    [anon_sym_volatile] = ACTIONS(2772),
    [anon_sym_restrict] = ACTIONS(2772),
    [anon_sym___restrict__] = ACTIONS(2772),
    [anon_sym__Atomic] = ACTIONS(2772),
    [anon_sym__Noreturn] = ACTIONS(2772),
    [anon_sym__Nonnull] = ACTIONS(2772),
    [anon_sym_nullable] = ACTIONS(2772),
    [anon_sym__Complex] = ACTIONS(2772),
    [anon_sym__Nullable] = ACTIONS(2772),
    [anon_sym__Nullable_result] = ACTIONS(2772),
    [anon_sym__Null_unspecified] = ACTIONS(2772),
    [anon_sym___autoreleasing] = ACTIONS(2772),
    [anon_sym___block] = ACTIONS(2772),
    [anon_sym___bridge] = ACTIONS(2772),
    [anon_sym___bridge_retained] = ACTIONS(2772),
    [anon_sym___bridge_transfer] = ACTIONS(2772),
    [anon_sym___complex] = ACTIONS(2772),
    [anon_sym___const] = ACTIONS(2772),
    [anon_sym___imag] = ACTIONS(2772),
    [anon_sym___kindof] = ACTIONS(2772),
    [anon_sym___nonnull] = ACTIONS(2772),
    [anon_sym___nullable] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2772),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2772),
    [anon_sym___real] = ACTIONS(2772),
    [anon_sym___strong] = ACTIONS(2772),
    [anon_sym___unsafe_unretained] = ACTIONS(2772),
    [anon_sym___unused] = ACTIONS(2772),
    [anon_sym___weak] = ACTIONS(2772),
    [anon_sym_alignas] = ACTIONS(2772),
    [anon_sym__Alignas] = ACTIONS(2772),
    [sym_primitive_type] = ACTIONS(2772),
    [anon_sym_enum] = ACTIONS(2772),
    [anon_sym_struct] = ACTIONS(2772),
    [anon_sym_union] = ACTIONS(2772),
    [anon_sym_in] = ACTIONS(2772),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2772),
    [anon_sym___typeof] = ACTIONS(2772),
    [anon_sym_typeof] = ACTIONS(2772),
    [anon_sym_ATprivate] = ACTIONS(2774),
    [anon_sym_ATprotected] = ACTIONS(2774),
    [anon_sym_ATpackage] = ACTIONS(2774),
    [anon_sym_ATpublic] = ACTIONS(2774),
    [anon_sym_BOOL] = ACTIONS(2772),
    [anon_sym_IMP] = ACTIONS(2772),
    [anon_sym_SEL] = ACTIONS(2772),
    [anon_sym_Class] = ACTIONS(2772),
    [anon_sym_id] = ACTIONS(2772),
    [anon_sym_out] = ACTIONS(2772),
    [anon_sym_inout] = ACTIONS(2772),
    [anon_sym_bycopy] = ACTIONS(2772),
    [anon_sym_byref] = ACTIONS(2772),
    [anon_sym_oneway] = ACTIONS(2772),
  },
  [STATE(4050)] = {
    [sym_identifier] = ACTIONS(5577),
    [aux_sym_preproc_if_token1] = ACTIONS(5577),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5579),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5579),
    [anon_sym___extension__] = ACTIONS(5577),
    [anon_sym___attribute__] = ACTIONS(5577),
    [anon_sym___attribute] = ACTIONS(5577),
    [anon_sym_noreturn] = ACTIONS(5577),
    [anon_sym_LBRACK] = ACTIONS(5579),
    [anon_sym_RBRACE] = ACTIONS(5579),
    [anon_sym_signed] = ACTIONS(5577),
    [anon_sym_unsigned] = ACTIONS(5577),
    [anon_sym_long] = ACTIONS(5577),
    [anon_sym_short] = ACTIONS(5577),
    [anon_sym_const] = ACTIONS(5577),
    [anon_sym_constexpr] = ACTIONS(5577),
    [anon_sym_volatile] = ACTIONS(5577),
    [anon_sym_restrict] = ACTIONS(5577),
    [anon_sym___restrict__] = ACTIONS(5577),
    [anon_sym__Atomic] = ACTIONS(5577),
    [anon_sym__Noreturn] = ACTIONS(5577),
    [anon_sym__Nonnull] = ACTIONS(5577),
    [anon_sym_nullable] = ACTIONS(5577),
    [anon_sym__Complex] = ACTIONS(5577),
    [anon_sym__Nullable] = ACTIONS(5577),
    [anon_sym__Nullable_result] = ACTIONS(5577),
    [anon_sym__Null_unspecified] = ACTIONS(5577),
    [anon_sym___autoreleasing] = ACTIONS(5577),
    [anon_sym___block] = ACTIONS(5577),
    [anon_sym___bridge] = ACTIONS(5577),
    [anon_sym___bridge_retained] = ACTIONS(5577),
    [anon_sym___bridge_transfer] = ACTIONS(5577),
    [anon_sym___complex] = ACTIONS(5577),
    [anon_sym___const] = ACTIONS(5577),
    [anon_sym___imag] = ACTIONS(5577),
    [anon_sym___kindof] = ACTIONS(5577),
    [anon_sym___nonnull] = ACTIONS(5577),
    [anon_sym___nullable] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5577),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5577),
    [anon_sym___real] = ACTIONS(5577),
    [anon_sym___strong] = ACTIONS(5577),
    [anon_sym___unsafe_unretained] = ACTIONS(5577),
    [anon_sym___unused] = ACTIONS(5577),
    [anon_sym___weak] = ACTIONS(5577),
    [anon_sym_alignas] = ACTIONS(5577),
    [anon_sym__Alignas] = ACTIONS(5577),
    [sym_primitive_type] = ACTIONS(5577),
    [anon_sym_enum] = ACTIONS(5577),
    [anon_sym_struct] = ACTIONS(5577),
    [anon_sym_union] = ACTIONS(5577),
    [anon_sym_in] = ACTIONS(5577),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5577),
    [anon_sym___typeof] = ACTIONS(5577),
    [anon_sym_typeof] = ACTIONS(5577),
    [anon_sym_ATprivate] = ACTIONS(5579),
    [anon_sym_ATprotected] = ACTIONS(5579),
    [anon_sym_ATpackage] = ACTIONS(5579),
    [anon_sym_ATpublic] = ACTIONS(5579),
    [anon_sym_BOOL] = ACTIONS(5577),
    [anon_sym_IMP] = ACTIONS(5577),
    [anon_sym_SEL] = ACTIONS(5577),
    [anon_sym_Class] = ACTIONS(5577),
    [anon_sym_id] = ACTIONS(5577),
    [anon_sym_out] = ACTIONS(5577),
    [anon_sym_inout] = ACTIONS(5577),
    [anon_sym_bycopy] = ACTIONS(5577),
    [anon_sym_byref] = ACTIONS(5577),
    [anon_sym_oneway] = ACTIONS(5577),
  },
  [STATE(4051)] = {
    [sym_identifier] = ACTIONS(5628),
    [aux_sym_preproc_if_token1] = ACTIONS(5628),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5630),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5630),
    [anon_sym___extension__] = ACTIONS(5628),
    [anon_sym___attribute__] = ACTIONS(5628),
    [anon_sym___attribute] = ACTIONS(5628),
    [anon_sym_noreturn] = ACTIONS(5628),
    [anon_sym_LBRACK] = ACTIONS(5630),
    [anon_sym_RBRACE] = ACTIONS(5630),
    [anon_sym_signed] = ACTIONS(5628),
    [anon_sym_unsigned] = ACTIONS(5628),
    [anon_sym_long] = ACTIONS(5628),
    [anon_sym_short] = ACTIONS(5628),
    [anon_sym_const] = ACTIONS(5628),
    [anon_sym_constexpr] = ACTIONS(5628),
    [anon_sym_volatile] = ACTIONS(5628),
    [anon_sym_restrict] = ACTIONS(5628),
    [anon_sym___restrict__] = ACTIONS(5628),
    [anon_sym__Atomic] = ACTIONS(5628),
    [anon_sym__Noreturn] = ACTIONS(5628),
    [anon_sym__Nonnull] = ACTIONS(5628),
    [anon_sym_nullable] = ACTIONS(5628),
    [anon_sym__Complex] = ACTIONS(5628),
    [anon_sym__Nullable] = ACTIONS(5628),
    [anon_sym__Nullable_result] = ACTIONS(5628),
    [anon_sym__Null_unspecified] = ACTIONS(5628),
    [anon_sym___autoreleasing] = ACTIONS(5628),
    [anon_sym___block] = ACTIONS(5628),
    [anon_sym___bridge] = ACTIONS(5628),
    [anon_sym___bridge_retained] = ACTIONS(5628),
    [anon_sym___bridge_transfer] = ACTIONS(5628),
    [anon_sym___complex] = ACTIONS(5628),
    [anon_sym___const] = ACTIONS(5628),
    [anon_sym___imag] = ACTIONS(5628),
    [anon_sym___kindof] = ACTIONS(5628),
    [anon_sym___nonnull] = ACTIONS(5628),
    [anon_sym___nullable] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5628),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5628),
    [anon_sym___real] = ACTIONS(5628),
    [anon_sym___strong] = ACTIONS(5628),
    [anon_sym___unsafe_unretained] = ACTIONS(5628),
    [anon_sym___unused] = ACTIONS(5628),
    [anon_sym___weak] = ACTIONS(5628),
    [anon_sym_alignas] = ACTIONS(5628),
    [anon_sym__Alignas] = ACTIONS(5628),
    [sym_primitive_type] = ACTIONS(5628),
    [anon_sym_enum] = ACTIONS(5628),
    [anon_sym_struct] = ACTIONS(5628),
    [anon_sym_union] = ACTIONS(5628),
    [anon_sym_in] = ACTIONS(5628),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5628),
    [anon_sym___typeof] = ACTIONS(5628),
    [anon_sym_typeof] = ACTIONS(5628),
    [anon_sym_ATprivate] = ACTIONS(5630),
    [anon_sym_ATprotected] = ACTIONS(5630),
    [anon_sym_ATpackage] = ACTIONS(5630),
    [anon_sym_ATpublic] = ACTIONS(5630),
    [anon_sym_BOOL] = ACTIONS(5628),
    [anon_sym_IMP] = ACTIONS(5628),
    [anon_sym_SEL] = ACTIONS(5628),
    [anon_sym_Class] = ACTIONS(5628),
    [anon_sym_id] = ACTIONS(5628),
    [anon_sym_out] = ACTIONS(5628),
    [anon_sym_inout] = ACTIONS(5628),
    [anon_sym_bycopy] = ACTIONS(5628),
    [anon_sym_byref] = ACTIONS(5628),
    [anon_sym_oneway] = ACTIONS(5628),
  },
  [STATE(4052)] = {
    [sym_identifier] = ACTIONS(5573),
    [aux_sym_preproc_if_token1] = ACTIONS(5573),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5575),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5575),
    [anon_sym___extension__] = ACTIONS(5573),
    [anon_sym___attribute__] = ACTIONS(5573),
    [anon_sym___attribute] = ACTIONS(5573),
    [anon_sym_noreturn] = ACTIONS(5573),
    [anon_sym_LBRACK] = ACTIONS(5575),
    [anon_sym_RBRACE] = ACTIONS(5575),
    [anon_sym_signed] = ACTIONS(5573),
    [anon_sym_unsigned] = ACTIONS(5573),
    [anon_sym_long] = ACTIONS(5573),
    [anon_sym_short] = ACTIONS(5573),
    [anon_sym_const] = ACTIONS(5573),
    [anon_sym_constexpr] = ACTIONS(5573),
    [anon_sym_volatile] = ACTIONS(5573),
    [anon_sym_restrict] = ACTIONS(5573),
    [anon_sym___restrict__] = ACTIONS(5573),
    [anon_sym__Atomic] = ACTIONS(5573),
    [anon_sym__Noreturn] = ACTIONS(5573),
    [anon_sym__Nonnull] = ACTIONS(5573),
    [anon_sym_nullable] = ACTIONS(5573),
    [anon_sym__Complex] = ACTIONS(5573),
    [anon_sym__Nullable] = ACTIONS(5573),
    [anon_sym__Nullable_result] = ACTIONS(5573),
    [anon_sym__Null_unspecified] = ACTIONS(5573),
    [anon_sym___autoreleasing] = ACTIONS(5573),
    [anon_sym___block] = ACTIONS(5573),
    [anon_sym___bridge] = ACTIONS(5573),
    [anon_sym___bridge_retained] = ACTIONS(5573),
    [anon_sym___bridge_transfer] = ACTIONS(5573),
    [anon_sym___complex] = ACTIONS(5573),
    [anon_sym___const] = ACTIONS(5573),
    [anon_sym___imag] = ACTIONS(5573),
    [anon_sym___kindof] = ACTIONS(5573),
    [anon_sym___nonnull] = ACTIONS(5573),
    [anon_sym___nullable] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5573),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5573),
    [anon_sym___real] = ACTIONS(5573),
    [anon_sym___strong] = ACTIONS(5573),
    [anon_sym___unsafe_unretained] = ACTIONS(5573),
    [anon_sym___unused] = ACTIONS(5573),
    [anon_sym___weak] = ACTIONS(5573),
    [anon_sym_alignas] = ACTIONS(5573),
    [anon_sym__Alignas] = ACTIONS(5573),
    [sym_primitive_type] = ACTIONS(5573),
    [anon_sym_enum] = ACTIONS(5573),
    [anon_sym_struct] = ACTIONS(5573),
    [anon_sym_union] = ACTIONS(5573),
    [anon_sym_in] = ACTIONS(5573),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5573),
    [anon_sym___typeof] = ACTIONS(5573),
    [anon_sym_typeof] = ACTIONS(5573),
    [anon_sym_ATprivate] = ACTIONS(5575),
    [anon_sym_ATprotected] = ACTIONS(5575),
    [anon_sym_ATpackage] = ACTIONS(5575),
    [anon_sym_ATpublic] = ACTIONS(5575),
    [anon_sym_BOOL] = ACTIONS(5573),
    [anon_sym_IMP] = ACTIONS(5573),
    [anon_sym_SEL] = ACTIONS(5573),
    [anon_sym_Class] = ACTIONS(5573),
    [anon_sym_id] = ACTIONS(5573),
    [anon_sym_out] = ACTIONS(5573),
    [anon_sym_inout] = ACTIONS(5573),
    [anon_sym_bycopy] = ACTIONS(5573),
    [anon_sym_byref] = ACTIONS(5573),
    [anon_sym_oneway] = ACTIONS(5573),
  },
  [STATE(4053)] = {
    [sym_attribute_specifier] = STATE(3892),
    [sym_enumerator_list] = STATE(4098),
    [sym_identifier] = ACTIONS(7512),
    [anon_sym_LPAREN2] = ACTIONS(7514),
    [anon_sym_STAR] = ACTIONS(7514),
    [anon_sym_CARET] = ACTIONS(7514),
    [anon_sym___extension__] = ACTIONS(7512),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7512),
    [anon_sym_LBRACK] = ACTIONS(7514),
    [anon_sym___based] = ACTIONS(7512),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7512),
    [anon_sym_unsigned] = ACTIONS(7512),
    [anon_sym_long] = ACTIONS(7512),
    [anon_sym_short] = ACTIONS(7512),
    [anon_sym_const] = ACTIONS(7512),
    [anon_sym_constexpr] = ACTIONS(7512),
    [anon_sym_volatile] = ACTIONS(7512),
    [anon_sym_restrict] = ACTIONS(7512),
    [anon_sym___restrict__] = ACTIONS(7512),
    [anon_sym__Atomic] = ACTIONS(7512),
    [anon_sym__Noreturn] = ACTIONS(7512),
    [anon_sym__Nonnull] = ACTIONS(7512),
    [anon_sym_nullable] = ACTIONS(7512),
    [anon_sym__Complex] = ACTIONS(7512),
    [anon_sym__Nullable] = ACTIONS(7512),
    [anon_sym__Nullable_result] = ACTIONS(7512),
    [anon_sym__Null_unspecified] = ACTIONS(7512),
    [anon_sym___autoreleasing] = ACTIONS(7512),
    [anon_sym___block] = ACTIONS(7512),
    [anon_sym___bridge] = ACTIONS(7512),
    [anon_sym___bridge_retained] = ACTIONS(7512),
    [anon_sym___bridge_transfer] = ACTIONS(7512),
    [anon_sym___complex] = ACTIONS(7512),
    [anon_sym___const] = ACTIONS(7512),
    [anon_sym___imag] = ACTIONS(7512),
    [anon_sym___kindof] = ACTIONS(7512),
    [anon_sym___nonnull] = ACTIONS(7512),
    [anon_sym___nullable] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7512),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7512),
    [anon_sym___real] = ACTIONS(7512),
    [anon_sym___strong] = ACTIONS(7512),
    [anon_sym___unsafe_unretained] = ACTIONS(7512),
    [anon_sym___unused] = ACTIONS(7512),
    [anon_sym___weak] = ACTIONS(7512),
    [anon_sym_alignas] = ACTIONS(7512),
    [anon_sym__Alignas] = ACTIONS(7512),
    [sym_primitive_type] = ACTIONS(7512),
    [anon_sym_enum] = ACTIONS(7512),
    [anon_sym_COLON] = ACTIONS(7514),
    [anon_sym_struct] = ACTIONS(7512),
    [anon_sym_union] = ACTIONS(7512),
    [anon_sym_in] = ACTIONS(7512),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7512),
    [anon_sym___typeof] = ACTIONS(7512),
    [anon_sym_typeof] = ACTIONS(7512),
    [anon_sym_BOOL] = ACTIONS(7512),
    [anon_sym_IMP] = ACTIONS(7512),
    [anon_sym_SEL] = ACTIONS(7512),
    [anon_sym_Class] = ACTIONS(7512),
    [anon_sym_id] = ACTIONS(7512),
    [anon_sym_out] = ACTIONS(7512),
    [anon_sym_inout] = ACTIONS(7512),
    [anon_sym_bycopy] = ACTIONS(7512),
    [anon_sym_byref] = ACTIONS(7512),
    [anon_sym_oneway] = ACTIONS(7512),
  },
  [STATE(4054)] = {
    [sym_identifier] = ACTIONS(8960),
    [aux_sym_preproc_if_token1] = ACTIONS(8960),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(8962),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(8962),
    [anon_sym___extension__] = ACTIONS(8960),
    [anon_sym___attribute__] = ACTIONS(8960),
    [anon_sym___attribute] = ACTIONS(8960),
    [anon_sym_noreturn] = ACTIONS(8960),
    [anon_sym_LBRACK] = ACTIONS(8962),
    [anon_sym_RBRACE] = ACTIONS(8962),
    [anon_sym_signed] = ACTIONS(8960),
    [anon_sym_unsigned] = ACTIONS(8960),
    [anon_sym_long] = ACTIONS(8960),
    [anon_sym_short] = ACTIONS(8960),
    [anon_sym_const] = ACTIONS(8960),
    [anon_sym_constexpr] = ACTIONS(8960),
    [anon_sym_volatile] = ACTIONS(8960),
    [anon_sym_restrict] = ACTIONS(8960),
    [anon_sym___restrict__] = ACTIONS(8960),
    [anon_sym__Atomic] = ACTIONS(8960),
    [anon_sym__Noreturn] = ACTIONS(8960),
    [anon_sym__Nonnull] = ACTIONS(8960),
    [anon_sym_nullable] = ACTIONS(8960),
    [anon_sym__Complex] = ACTIONS(8960),
    [anon_sym__Nullable] = ACTIONS(8960),
    [anon_sym__Nullable_result] = ACTIONS(8960),
    [anon_sym__Null_unspecified] = ACTIONS(8960),
    [anon_sym___autoreleasing] = ACTIONS(8960),
    [anon_sym___block] = ACTIONS(8960),
    [anon_sym___bridge] = ACTIONS(8960),
    [anon_sym___bridge_retained] = ACTIONS(8960),
    [anon_sym___bridge_transfer] = ACTIONS(8960),
    [anon_sym___complex] = ACTIONS(8960),
    [anon_sym___const] = ACTIONS(8960),
    [anon_sym___imag] = ACTIONS(8960),
    [anon_sym___kindof] = ACTIONS(8960),
    [anon_sym___nonnull] = ACTIONS(8960),
    [anon_sym___nullable] = ACTIONS(8960),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8960),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8960),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8960),
    [anon_sym___real] = ACTIONS(8960),
    [anon_sym___strong] = ACTIONS(8960),
    [anon_sym___unsafe_unretained] = ACTIONS(8960),
    [anon_sym___unused] = ACTIONS(8960),
    [anon_sym___weak] = ACTIONS(8960),
    [anon_sym_alignas] = ACTIONS(8960),
    [anon_sym__Alignas] = ACTIONS(8960),
    [sym_primitive_type] = ACTIONS(8960),
    [anon_sym_enum] = ACTIONS(8960),
    [anon_sym_struct] = ACTIONS(8960),
    [anon_sym_union] = ACTIONS(8960),
    [anon_sym_in] = ACTIONS(8960),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8960),
    [anon_sym___typeof] = ACTIONS(8960),
    [anon_sym_typeof] = ACTIONS(8960),
    [anon_sym_ATprivate] = ACTIONS(8962),
    [anon_sym_ATprotected] = ACTIONS(8962),
    [anon_sym_ATpackage] = ACTIONS(8962),
    [anon_sym_ATpublic] = ACTIONS(8962),
    [anon_sym_BOOL] = ACTIONS(8960),
    [anon_sym_IMP] = ACTIONS(8960),
    [anon_sym_SEL] = ACTIONS(8960),
    [anon_sym_Class] = ACTIONS(8960),
    [anon_sym_id] = ACTIONS(8960),
    [anon_sym_out] = ACTIONS(8960),
    [anon_sym_inout] = ACTIONS(8960),
    [anon_sym_bycopy] = ACTIONS(8960),
    [anon_sym_byref] = ACTIONS(8960),
    [anon_sym_oneway] = ACTIONS(8960),
  },
  [STATE(4055)] = {
    [sym_attribute_specifier] = STATE(3894),
    [sym_enumerator_list] = STATE(4100),
    [sym_identifier] = ACTIONS(7521),
    [anon_sym_LPAREN2] = ACTIONS(7523),
    [anon_sym_STAR] = ACTIONS(7523),
    [anon_sym_CARET] = ACTIONS(7523),
    [anon_sym___extension__] = ACTIONS(7521),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7521),
    [anon_sym_LBRACK] = ACTIONS(7523),
    [anon_sym___based] = ACTIONS(7521),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7521),
    [anon_sym_unsigned] = ACTIONS(7521),
    [anon_sym_long] = ACTIONS(7521),
    [anon_sym_short] = ACTIONS(7521),
    [anon_sym_const] = ACTIONS(7521),
    [anon_sym_constexpr] = ACTIONS(7521),
    [anon_sym_volatile] = ACTIONS(7521),
    [anon_sym_restrict] = ACTIONS(7521),
    [anon_sym___restrict__] = ACTIONS(7521),
    [anon_sym__Atomic] = ACTIONS(7521),
    [anon_sym__Noreturn] = ACTIONS(7521),
    [anon_sym__Nonnull] = ACTIONS(7521),
    [anon_sym_nullable] = ACTIONS(7521),
    [anon_sym__Complex] = ACTIONS(7521),
    [anon_sym__Nullable] = ACTIONS(7521),
    [anon_sym__Nullable_result] = ACTIONS(7521),
    [anon_sym__Null_unspecified] = ACTIONS(7521),
    [anon_sym___autoreleasing] = ACTIONS(7521),
    [anon_sym___block] = ACTIONS(7521),
    [anon_sym___bridge] = ACTIONS(7521),
    [anon_sym___bridge_retained] = ACTIONS(7521),
    [anon_sym___bridge_transfer] = ACTIONS(7521),
    [anon_sym___complex] = ACTIONS(7521),
    [anon_sym___const] = ACTIONS(7521),
    [anon_sym___imag] = ACTIONS(7521),
    [anon_sym___kindof] = ACTIONS(7521),
    [anon_sym___nonnull] = ACTIONS(7521),
    [anon_sym___nullable] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7521),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7521),
    [anon_sym___real] = ACTIONS(7521),
    [anon_sym___strong] = ACTIONS(7521),
    [anon_sym___unsafe_unretained] = ACTIONS(7521),
    [anon_sym___unused] = ACTIONS(7521),
    [anon_sym___weak] = ACTIONS(7521),
    [anon_sym_alignas] = ACTIONS(7521),
    [anon_sym__Alignas] = ACTIONS(7521),
    [sym_primitive_type] = ACTIONS(7521),
    [anon_sym_enum] = ACTIONS(7521),
    [anon_sym_COLON] = ACTIONS(7523),
    [anon_sym_struct] = ACTIONS(7521),
    [anon_sym_union] = ACTIONS(7521),
    [anon_sym_in] = ACTIONS(7521),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7521),
    [anon_sym___typeof] = ACTIONS(7521),
    [anon_sym_typeof] = ACTIONS(7521),
    [anon_sym_BOOL] = ACTIONS(7521),
    [anon_sym_IMP] = ACTIONS(7521),
    [anon_sym_SEL] = ACTIONS(7521),
    [anon_sym_Class] = ACTIONS(7521),
    [anon_sym_id] = ACTIONS(7521),
    [anon_sym_out] = ACTIONS(7521),
    [anon_sym_inout] = ACTIONS(7521),
    [anon_sym_bycopy] = ACTIONS(7521),
    [anon_sym_byref] = ACTIONS(7521),
    [anon_sym_oneway] = ACTIONS(7521),
  },
  [STATE(4056)] = {
    [sym_identifier] = ACTIONS(2704),
    [aux_sym_preproc_if_token1] = ACTIONS(2704),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2706),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2706),
    [anon_sym___extension__] = ACTIONS(2704),
    [anon_sym___attribute__] = ACTIONS(2704),
    [anon_sym___attribute] = ACTIONS(2704),
    [anon_sym_noreturn] = ACTIONS(2704),
    [anon_sym_LBRACK] = ACTIONS(2706),
    [anon_sym_RBRACE] = ACTIONS(2706),
    [anon_sym_signed] = ACTIONS(2704),
    [anon_sym_unsigned] = ACTIONS(2704),
    [anon_sym_long] = ACTIONS(2704),
    [anon_sym_short] = ACTIONS(2704),
    [anon_sym_const] = ACTIONS(2704),
    [anon_sym_constexpr] = ACTIONS(2704),
    [anon_sym_volatile] = ACTIONS(2704),
    [anon_sym_restrict] = ACTIONS(2704),
    [anon_sym___restrict__] = ACTIONS(2704),
    [anon_sym__Atomic] = ACTIONS(2704),
    [anon_sym__Noreturn] = ACTIONS(2704),
    [anon_sym__Nonnull] = ACTIONS(2704),
    [anon_sym_nullable] = ACTIONS(2704),
    [anon_sym__Complex] = ACTIONS(2704),
    [anon_sym__Nullable] = ACTIONS(2704),
    [anon_sym__Nullable_result] = ACTIONS(2704),
    [anon_sym__Null_unspecified] = ACTIONS(2704),
    [anon_sym___autoreleasing] = ACTIONS(2704),
    [anon_sym___block] = ACTIONS(2704),
    [anon_sym___bridge] = ACTIONS(2704),
    [anon_sym___bridge_retained] = ACTIONS(2704),
    [anon_sym___bridge_transfer] = ACTIONS(2704),
    [anon_sym___complex] = ACTIONS(2704),
    [anon_sym___const] = ACTIONS(2704),
    [anon_sym___imag] = ACTIONS(2704),
    [anon_sym___kindof] = ACTIONS(2704),
    [anon_sym___nonnull] = ACTIONS(2704),
    [anon_sym___nullable] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2704),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2704),
    [anon_sym___real] = ACTIONS(2704),
    [anon_sym___strong] = ACTIONS(2704),
    [anon_sym___unsafe_unretained] = ACTIONS(2704),
    [anon_sym___unused] = ACTIONS(2704),
    [anon_sym___weak] = ACTIONS(2704),
    [anon_sym_alignas] = ACTIONS(2704),
    [anon_sym__Alignas] = ACTIONS(2704),
    [sym_primitive_type] = ACTIONS(2704),
    [anon_sym_enum] = ACTIONS(2704),
    [anon_sym_struct] = ACTIONS(2704),
    [anon_sym_union] = ACTIONS(2704),
    [anon_sym_in] = ACTIONS(2704),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2704),
    [anon_sym___typeof] = ACTIONS(2704),
    [anon_sym_typeof] = ACTIONS(2704),
    [anon_sym_ATprivate] = ACTIONS(2706),
    [anon_sym_ATprotected] = ACTIONS(2706),
    [anon_sym_ATpackage] = ACTIONS(2706),
    [anon_sym_ATpublic] = ACTIONS(2706),
    [anon_sym_BOOL] = ACTIONS(2704),
    [anon_sym_IMP] = ACTIONS(2704),
    [anon_sym_SEL] = ACTIONS(2704),
    [anon_sym_Class] = ACTIONS(2704),
    [anon_sym_id] = ACTIONS(2704),
    [anon_sym_out] = ACTIONS(2704),
    [anon_sym_inout] = ACTIONS(2704),
    [anon_sym_bycopy] = ACTIONS(2704),
    [anon_sym_byref] = ACTIONS(2704),
    [anon_sym_oneway] = ACTIONS(2704),
  },
  [STATE(4057)] = {
    [sym_identifier] = ACTIONS(2716),
    [aux_sym_preproc_if_token1] = ACTIONS(2716),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2718),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2718),
    [anon_sym___extension__] = ACTIONS(2716),
    [anon_sym___attribute__] = ACTIONS(2716),
    [anon_sym___attribute] = ACTIONS(2716),
    [anon_sym_noreturn] = ACTIONS(2716),
    [anon_sym_LBRACK] = ACTIONS(2718),
    [anon_sym_RBRACE] = ACTIONS(2718),
    [anon_sym_signed] = ACTIONS(2716),
    [anon_sym_unsigned] = ACTIONS(2716),
    [anon_sym_long] = ACTIONS(2716),
    [anon_sym_short] = ACTIONS(2716),
    [anon_sym_const] = ACTIONS(2716),
    [anon_sym_constexpr] = ACTIONS(2716),
    [anon_sym_volatile] = ACTIONS(2716),
    [anon_sym_restrict] = ACTIONS(2716),
    [anon_sym___restrict__] = ACTIONS(2716),
    [anon_sym__Atomic] = ACTIONS(2716),
    [anon_sym__Noreturn] = ACTIONS(2716),
    [anon_sym__Nonnull] = ACTIONS(2716),
    [anon_sym_nullable] = ACTIONS(2716),
    [anon_sym__Complex] = ACTIONS(2716),
    [anon_sym__Nullable] = ACTIONS(2716),
    [anon_sym__Nullable_result] = ACTIONS(2716),
    [anon_sym__Null_unspecified] = ACTIONS(2716),
    [anon_sym___autoreleasing] = ACTIONS(2716),
    [anon_sym___block] = ACTIONS(2716),
    [anon_sym___bridge] = ACTIONS(2716),
    [anon_sym___bridge_retained] = ACTIONS(2716),
    [anon_sym___bridge_transfer] = ACTIONS(2716),
    [anon_sym___complex] = ACTIONS(2716),
    [anon_sym___const] = ACTIONS(2716),
    [anon_sym___imag] = ACTIONS(2716),
    [anon_sym___kindof] = ACTIONS(2716),
    [anon_sym___nonnull] = ACTIONS(2716),
    [anon_sym___nullable] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2716),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2716),
    [anon_sym___real] = ACTIONS(2716),
    [anon_sym___strong] = ACTIONS(2716),
    [anon_sym___unsafe_unretained] = ACTIONS(2716),
    [anon_sym___unused] = ACTIONS(2716),
    [anon_sym___weak] = ACTIONS(2716),
    [anon_sym_alignas] = ACTIONS(2716),
    [anon_sym__Alignas] = ACTIONS(2716),
    [sym_primitive_type] = ACTIONS(2716),
    [anon_sym_enum] = ACTIONS(2716),
    [anon_sym_struct] = ACTIONS(2716),
    [anon_sym_union] = ACTIONS(2716),
    [anon_sym_in] = ACTIONS(2716),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2716),
    [anon_sym___typeof] = ACTIONS(2716),
    [anon_sym_typeof] = ACTIONS(2716),
    [anon_sym_ATprivate] = ACTIONS(2718),
    [anon_sym_ATprotected] = ACTIONS(2718),
    [anon_sym_ATpackage] = ACTIONS(2718),
    [anon_sym_ATpublic] = ACTIONS(2718),
    [anon_sym_BOOL] = ACTIONS(2716),
    [anon_sym_IMP] = ACTIONS(2716),
    [anon_sym_SEL] = ACTIONS(2716),
    [anon_sym_Class] = ACTIONS(2716),
    [anon_sym_id] = ACTIONS(2716),
    [anon_sym_out] = ACTIONS(2716),
    [anon_sym_inout] = ACTIONS(2716),
    [anon_sym_bycopy] = ACTIONS(2716),
    [anon_sym_byref] = ACTIONS(2716),
    [anon_sym_oneway] = ACTIONS(2716),
  },
  [STATE(4058)] = {
    [sym_identifier] = ACTIONS(2634),
    [aux_sym_preproc_if_token1] = ACTIONS(2634),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2636),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2636),
    [anon_sym___extension__] = ACTIONS(2634),
    [anon_sym___attribute__] = ACTIONS(2634),
    [anon_sym___attribute] = ACTIONS(2634),
    [anon_sym_noreturn] = ACTIONS(2634),
    [anon_sym_LBRACK] = ACTIONS(2636),
    [anon_sym_RBRACE] = ACTIONS(2636),
    [anon_sym_signed] = ACTIONS(2634),
    [anon_sym_unsigned] = ACTIONS(2634),
    [anon_sym_long] = ACTIONS(2634),
    [anon_sym_short] = ACTIONS(2634),
    [anon_sym_const] = ACTIONS(2634),
    [anon_sym_constexpr] = ACTIONS(2634),
    [anon_sym_volatile] = ACTIONS(2634),
    [anon_sym_restrict] = ACTIONS(2634),
    [anon_sym___restrict__] = ACTIONS(2634),
    [anon_sym__Atomic] = ACTIONS(2634),
    [anon_sym__Noreturn] = ACTIONS(2634),
    [anon_sym__Nonnull] = ACTIONS(2634),
    [anon_sym_nullable] = ACTIONS(2634),
    [anon_sym__Complex] = ACTIONS(2634),
    [anon_sym__Nullable] = ACTIONS(2634),
    [anon_sym__Nullable_result] = ACTIONS(2634),
    [anon_sym__Null_unspecified] = ACTIONS(2634),
    [anon_sym___autoreleasing] = ACTIONS(2634),
    [anon_sym___block] = ACTIONS(2634),
    [anon_sym___bridge] = ACTIONS(2634),
    [anon_sym___bridge_retained] = ACTIONS(2634),
    [anon_sym___bridge_transfer] = ACTIONS(2634),
    [anon_sym___complex] = ACTIONS(2634),
    [anon_sym___const] = ACTIONS(2634),
    [anon_sym___imag] = ACTIONS(2634),
    [anon_sym___kindof] = ACTIONS(2634),
    [anon_sym___nonnull] = ACTIONS(2634),
    [anon_sym___nullable] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2634),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2634),
    [anon_sym___real] = ACTIONS(2634),
    [anon_sym___strong] = ACTIONS(2634),
    [anon_sym___unsafe_unretained] = ACTIONS(2634),
    [anon_sym___unused] = ACTIONS(2634),
    [anon_sym___weak] = ACTIONS(2634),
    [anon_sym_alignas] = ACTIONS(2634),
    [anon_sym__Alignas] = ACTIONS(2634),
    [sym_primitive_type] = ACTIONS(2634),
    [anon_sym_enum] = ACTIONS(2634),
    [anon_sym_struct] = ACTIONS(2634),
    [anon_sym_union] = ACTIONS(2634),
    [anon_sym_in] = ACTIONS(2634),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2634),
    [anon_sym___typeof] = ACTIONS(2634),
    [anon_sym_typeof] = ACTIONS(2634),
    [anon_sym_ATprivate] = ACTIONS(2636),
    [anon_sym_ATprotected] = ACTIONS(2636),
    [anon_sym_ATpackage] = ACTIONS(2636),
    [anon_sym_ATpublic] = ACTIONS(2636),
    [anon_sym_BOOL] = ACTIONS(2634),
    [anon_sym_IMP] = ACTIONS(2634),
    [anon_sym_SEL] = ACTIONS(2634),
    [anon_sym_Class] = ACTIONS(2634),
    [anon_sym_id] = ACTIONS(2634),
    [anon_sym_out] = ACTIONS(2634),
    [anon_sym_inout] = ACTIONS(2634),
    [anon_sym_bycopy] = ACTIONS(2634),
    [anon_sym_byref] = ACTIONS(2634),
    [anon_sym_oneway] = ACTIONS(2634),
  },
  [STATE(4059)] = {
    [sym_identifier] = ACTIONS(2668),
    [aux_sym_preproc_if_token1] = ACTIONS(2668),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2670),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2670),
    [anon_sym___extension__] = ACTIONS(2668),
    [anon_sym___attribute__] = ACTIONS(2668),
    [anon_sym___attribute] = ACTIONS(2668),
    [anon_sym_noreturn] = ACTIONS(2668),
    [anon_sym_LBRACK] = ACTIONS(2670),
    [anon_sym_RBRACE] = ACTIONS(2670),
    [anon_sym_signed] = ACTIONS(2668),
    [anon_sym_unsigned] = ACTIONS(2668),
    [anon_sym_long] = ACTIONS(2668),
    [anon_sym_short] = ACTIONS(2668),
    [anon_sym_const] = ACTIONS(2668),
    [anon_sym_constexpr] = ACTIONS(2668),
    [anon_sym_volatile] = ACTIONS(2668),
    [anon_sym_restrict] = ACTIONS(2668),
    [anon_sym___restrict__] = ACTIONS(2668),
    [anon_sym__Atomic] = ACTIONS(2668),
    [anon_sym__Noreturn] = ACTIONS(2668),
    [anon_sym__Nonnull] = ACTIONS(2668),
    [anon_sym_nullable] = ACTIONS(2668),
    [anon_sym__Complex] = ACTIONS(2668),
    [anon_sym__Nullable] = ACTIONS(2668),
    [anon_sym__Nullable_result] = ACTIONS(2668),
    [anon_sym__Null_unspecified] = ACTIONS(2668),
    [anon_sym___autoreleasing] = ACTIONS(2668),
    [anon_sym___block] = ACTIONS(2668),
    [anon_sym___bridge] = ACTIONS(2668),
    [anon_sym___bridge_retained] = ACTIONS(2668),
    [anon_sym___bridge_transfer] = ACTIONS(2668),
    [anon_sym___complex] = ACTIONS(2668),
    [anon_sym___const] = ACTIONS(2668),
    [anon_sym___imag] = ACTIONS(2668),
    [anon_sym___kindof] = ACTIONS(2668),
    [anon_sym___nonnull] = ACTIONS(2668),
    [anon_sym___nullable] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2668),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2668),
    [anon_sym___real] = ACTIONS(2668),
    [anon_sym___strong] = ACTIONS(2668),
    [anon_sym___unsafe_unretained] = ACTIONS(2668),
    [anon_sym___unused] = ACTIONS(2668),
    [anon_sym___weak] = ACTIONS(2668),
    [anon_sym_alignas] = ACTIONS(2668),
    [anon_sym__Alignas] = ACTIONS(2668),
    [sym_primitive_type] = ACTIONS(2668),
    [anon_sym_enum] = ACTIONS(2668),
    [anon_sym_struct] = ACTIONS(2668),
    [anon_sym_union] = ACTIONS(2668),
    [anon_sym_in] = ACTIONS(2668),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2668),
    [anon_sym___typeof] = ACTIONS(2668),
    [anon_sym_typeof] = ACTIONS(2668),
    [anon_sym_ATprivate] = ACTIONS(2670),
    [anon_sym_ATprotected] = ACTIONS(2670),
    [anon_sym_ATpackage] = ACTIONS(2670),
    [anon_sym_ATpublic] = ACTIONS(2670),
    [anon_sym_BOOL] = ACTIONS(2668),
    [anon_sym_IMP] = ACTIONS(2668),
    [anon_sym_SEL] = ACTIONS(2668),
    [anon_sym_Class] = ACTIONS(2668),
    [anon_sym_id] = ACTIONS(2668),
    [anon_sym_out] = ACTIONS(2668),
    [anon_sym_inout] = ACTIONS(2668),
    [anon_sym_bycopy] = ACTIONS(2668),
    [anon_sym_byref] = ACTIONS(2668),
    [anon_sym_oneway] = ACTIONS(2668),
  },
  [STATE(4060)] = {
    [sym_identifier] = ACTIONS(8964),
    [aux_sym_preproc_if_token1] = ACTIONS(8964),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(8966),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(8966),
    [anon_sym___extension__] = ACTIONS(8964),
    [anon_sym___attribute__] = ACTIONS(8964),
    [anon_sym___attribute] = ACTIONS(8964),
    [anon_sym_noreturn] = ACTIONS(8964),
    [anon_sym_LBRACK] = ACTIONS(8966),
    [anon_sym_RBRACE] = ACTIONS(8966),
    [anon_sym_signed] = ACTIONS(8964),
    [anon_sym_unsigned] = ACTIONS(8964),
    [anon_sym_long] = ACTIONS(8964),
    [anon_sym_short] = ACTIONS(8964),
    [anon_sym_const] = ACTIONS(8964),
    [anon_sym_constexpr] = ACTIONS(8964),
    [anon_sym_volatile] = ACTIONS(8964),
    [anon_sym_restrict] = ACTIONS(8964),
    [anon_sym___restrict__] = ACTIONS(8964),
    [anon_sym__Atomic] = ACTIONS(8964),
    [anon_sym__Noreturn] = ACTIONS(8964),
    [anon_sym__Nonnull] = ACTIONS(8964),
    [anon_sym_nullable] = ACTIONS(8964),
    [anon_sym__Complex] = ACTIONS(8964),
    [anon_sym__Nullable] = ACTIONS(8964),
    [anon_sym__Nullable_result] = ACTIONS(8964),
    [anon_sym__Null_unspecified] = ACTIONS(8964),
    [anon_sym___autoreleasing] = ACTIONS(8964),
    [anon_sym___block] = ACTIONS(8964),
    [anon_sym___bridge] = ACTIONS(8964),
    [anon_sym___bridge_retained] = ACTIONS(8964),
    [anon_sym___bridge_transfer] = ACTIONS(8964),
    [anon_sym___complex] = ACTIONS(8964),
    [anon_sym___const] = ACTIONS(8964),
    [anon_sym___imag] = ACTIONS(8964),
    [anon_sym___kindof] = ACTIONS(8964),
    [anon_sym___nonnull] = ACTIONS(8964),
    [anon_sym___nullable] = ACTIONS(8964),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(8964),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(8964),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(8964),
    [anon_sym___real] = ACTIONS(8964),
    [anon_sym___strong] = ACTIONS(8964),
    [anon_sym___unsafe_unretained] = ACTIONS(8964),
    [anon_sym___unused] = ACTIONS(8964),
    [anon_sym___weak] = ACTIONS(8964),
    [anon_sym_alignas] = ACTIONS(8964),
    [anon_sym__Alignas] = ACTIONS(8964),
    [sym_primitive_type] = ACTIONS(8964),
    [anon_sym_enum] = ACTIONS(8964),
    [anon_sym_struct] = ACTIONS(8964),
    [anon_sym_union] = ACTIONS(8964),
    [anon_sym_in] = ACTIONS(8964),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(8964),
    [anon_sym___typeof] = ACTIONS(8964),
    [anon_sym_typeof] = ACTIONS(8964),
    [anon_sym_ATprivate] = ACTIONS(8966),
    [anon_sym_ATprotected] = ACTIONS(8966),
    [anon_sym_ATpackage] = ACTIONS(8966),
    [anon_sym_ATpublic] = ACTIONS(8966),
    [anon_sym_BOOL] = ACTIONS(8964),
    [anon_sym_IMP] = ACTIONS(8964),
    [anon_sym_SEL] = ACTIONS(8964),
    [anon_sym_Class] = ACTIONS(8964),
    [anon_sym_id] = ACTIONS(8964),
    [anon_sym_out] = ACTIONS(8964),
    [anon_sym_inout] = ACTIONS(8964),
    [anon_sym_bycopy] = ACTIONS(8964),
    [anon_sym_byref] = ACTIONS(8964),
    [anon_sym_oneway] = ACTIONS(8964),
  },
  [STATE(4061)] = {
    [sym_identifier] = ACTIONS(2824),
    [aux_sym_preproc_if_token1] = ACTIONS(2824),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2826),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2826),
    [anon_sym___extension__] = ACTIONS(2824),
    [anon_sym___attribute__] = ACTIONS(2824),
    [anon_sym___attribute] = ACTIONS(2824),
    [anon_sym_noreturn] = ACTIONS(2824),
    [anon_sym_LBRACK] = ACTIONS(2826),
    [anon_sym_RBRACE] = ACTIONS(2826),
    [anon_sym_signed] = ACTIONS(2824),
    [anon_sym_unsigned] = ACTIONS(2824),
    [anon_sym_long] = ACTIONS(2824),
    [anon_sym_short] = ACTIONS(2824),
    [anon_sym_const] = ACTIONS(2824),
    [anon_sym_constexpr] = ACTIONS(2824),
    [anon_sym_volatile] = ACTIONS(2824),
    [anon_sym_restrict] = ACTIONS(2824),
    [anon_sym___restrict__] = ACTIONS(2824),
    [anon_sym__Atomic] = ACTIONS(2824),
    [anon_sym__Noreturn] = ACTIONS(2824),
    [anon_sym__Nonnull] = ACTIONS(2824),
    [anon_sym_nullable] = ACTIONS(2824),
    [anon_sym__Complex] = ACTIONS(2824),
    [anon_sym__Nullable] = ACTIONS(2824),
    [anon_sym__Nullable_result] = ACTIONS(2824),
    [anon_sym__Null_unspecified] = ACTIONS(2824),
    [anon_sym___autoreleasing] = ACTIONS(2824),
    [anon_sym___block] = ACTIONS(2824),
    [anon_sym___bridge] = ACTIONS(2824),
    [anon_sym___bridge_retained] = ACTIONS(2824),
    [anon_sym___bridge_transfer] = ACTIONS(2824),
    [anon_sym___complex] = ACTIONS(2824),
    [anon_sym___const] = ACTIONS(2824),
    [anon_sym___imag] = ACTIONS(2824),
    [anon_sym___kindof] = ACTIONS(2824),
    [anon_sym___nonnull] = ACTIONS(2824),
    [anon_sym___nullable] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2824),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2824),
    [anon_sym___real] = ACTIONS(2824),
    [anon_sym___strong] = ACTIONS(2824),
    [anon_sym___unsafe_unretained] = ACTIONS(2824),
    [anon_sym___unused] = ACTIONS(2824),
    [anon_sym___weak] = ACTIONS(2824),
    [anon_sym_alignas] = ACTIONS(2824),
    [anon_sym__Alignas] = ACTIONS(2824),
    [sym_primitive_type] = ACTIONS(2824),
    [anon_sym_enum] = ACTIONS(2824),
    [anon_sym_struct] = ACTIONS(2824),
    [anon_sym_union] = ACTIONS(2824),
    [anon_sym_in] = ACTIONS(2824),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2824),
    [anon_sym___typeof] = ACTIONS(2824),
    [anon_sym_typeof] = ACTIONS(2824),
    [anon_sym_ATprivate] = ACTIONS(2826),
    [anon_sym_ATprotected] = ACTIONS(2826),
    [anon_sym_ATpackage] = ACTIONS(2826),
    [anon_sym_ATpublic] = ACTIONS(2826),
    [anon_sym_BOOL] = ACTIONS(2824),
    [anon_sym_IMP] = ACTIONS(2824),
    [anon_sym_SEL] = ACTIONS(2824),
    [anon_sym_Class] = ACTIONS(2824),
    [anon_sym_id] = ACTIONS(2824),
    [anon_sym_out] = ACTIONS(2824),
    [anon_sym_inout] = ACTIONS(2824),
    [anon_sym_bycopy] = ACTIONS(2824),
    [anon_sym_byref] = ACTIONS(2824),
    [anon_sym_oneway] = ACTIONS(2824),
  },
  [STATE(4062)] = {
    [sym_attribute_specifier] = STATE(3957),
    [sym_enumerator_list] = STATE(4101),
    [sym_identifier] = ACTIONS(7561),
    [anon_sym_LPAREN2] = ACTIONS(7563),
    [anon_sym_STAR] = ACTIONS(7563),
    [anon_sym_CARET] = ACTIONS(7563),
    [anon_sym___extension__] = ACTIONS(7561),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7561),
    [anon_sym_LBRACK] = ACTIONS(7563),
    [anon_sym___based] = ACTIONS(7561),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7561),
    [anon_sym_unsigned] = ACTIONS(7561),
    [anon_sym_long] = ACTIONS(7561),
    [anon_sym_short] = ACTIONS(7561),
    [anon_sym_const] = ACTIONS(7561),
    [anon_sym_constexpr] = ACTIONS(7561),
    [anon_sym_volatile] = ACTIONS(7561),
    [anon_sym_restrict] = ACTIONS(7561),
    [anon_sym___restrict__] = ACTIONS(7561),
    [anon_sym__Atomic] = ACTIONS(7561),
    [anon_sym__Noreturn] = ACTIONS(7561),
    [anon_sym__Nonnull] = ACTIONS(7561),
    [anon_sym_nullable] = ACTIONS(7561),
    [anon_sym__Complex] = ACTIONS(7561),
    [anon_sym__Nullable] = ACTIONS(7561),
    [anon_sym__Nullable_result] = ACTIONS(7561),
    [anon_sym__Null_unspecified] = ACTIONS(7561),
    [anon_sym___autoreleasing] = ACTIONS(7561),
    [anon_sym___block] = ACTIONS(7561),
    [anon_sym___bridge] = ACTIONS(7561),
    [anon_sym___bridge_retained] = ACTIONS(7561),
    [anon_sym___bridge_transfer] = ACTIONS(7561),
    [anon_sym___complex] = ACTIONS(7561),
    [anon_sym___const] = ACTIONS(7561),
    [anon_sym___imag] = ACTIONS(7561),
    [anon_sym___kindof] = ACTIONS(7561),
    [anon_sym___nonnull] = ACTIONS(7561),
    [anon_sym___nullable] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7561),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7561),
    [anon_sym___real] = ACTIONS(7561),
    [anon_sym___strong] = ACTIONS(7561),
    [anon_sym___unsafe_unretained] = ACTIONS(7561),
    [anon_sym___unused] = ACTIONS(7561),
    [anon_sym___weak] = ACTIONS(7561),
    [anon_sym_alignas] = ACTIONS(7561),
    [anon_sym__Alignas] = ACTIONS(7561),
    [sym_primitive_type] = ACTIONS(7561),
    [anon_sym_enum] = ACTIONS(7561),
    [anon_sym_COLON] = ACTIONS(7563),
    [anon_sym_struct] = ACTIONS(7561),
    [anon_sym_union] = ACTIONS(7561),
    [anon_sym_in] = ACTIONS(7561),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7561),
    [anon_sym___typeof] = ACTIONS(7561),
    [anon_sym_typeof] = ACTIONS(7561),
    [anon_sym_BOOL] = ACTIONS(7561),
    [anon_sym_IMP] = ACTIONS(7561),
    [anon_sym_SEL] = ACTIONS(7561),
    [anon_sym_Class] = ACTIONS(7561),
    [anon_sym_id] = ACTIONS(7561),
    [anon_sym_out] = ACTIONS(7561),
    [anon_sym_inout] = ACTIONS(7561),
    [anon_sym_bycopy] = ACTIONS(7561),
    [anon_sym_byref] = ACTIONS(7561),
    [anon_sym_oneway] = ACTIONS(7561),
  },
  [STATE(4063)] = {
    [sym_identifier] = ACTIONS(2604),
    [aux_sym_preproc_if_token1] = ACTIONS(2604),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2606),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2606),
    [anon_sym___extension__] = ACTIONS(2604),
    [anon_sym___attribute__] = ACTIONS(2604),
    [anon_sym___attribute] = ACTIONS(2604),
    [anon_sym_noreturn] = ACTIONS(2604),
    [anon_sym_LBRACK] = ACTIONS(2606),
    [anon_sym_RBRACE] = ACTIONS(2606),
    [anon_sym_signed] = ACTIONS(2604),
    [anon_sym_unsigned] = ACTIONS(2604),
    [anon_sym_long] = ACTIONS(2604),
    [anon_sym_short] = ACTIONS(2604),
    [anon_sym_const] = ACTIONS(2604),
    [anon_sym_constexpr] = ACTIONS(2604),
    [anon_sym_volatile] = ACTIONS(2604),
    [anon_sym_restrict] = ACTIONS(2604),
    [anon_sym___restrict__] = ACTIONS(2604),
    [anon_sym__Atomic] = ACTIONS(2604),
    [anon_sym__Noreturn] = ACTIONS(2604),
    [anon_sym__Nonnull] = ACTIONS(2604),
    [anon_sym_nullable] = ACTIONS(2604),
    [anon_sym__Complex] = ACTIONS(2604),
    [anon_sym__Nullable] = ACTIONS(2604),
    [anon_sym__Nullable_result] = ACTIONS(2604),
    [anon_sym__Null_unspecified] = ACTIONS(2604),
    [anon_sym___autoreleasing] = ACTIONS(2604),
    [anon_sym___block] = ACTIONS(2604),
    [anon_sym___bridge] = ACTIONS(2604),
    [anon_sym___bridge_retained] = ACTIONS(2604),
    [anon_sym___bridge_transfer] = ACTIONS(2604),
    [anon_sym___complex] = ACTIONS(2604),
    [anon_sym___const] = ACTIONS(2604),
    [anon_sym___imag] = ACTIONS(2604),
    [anon_sym___kindof] = ACTIONS(2604),
    [anon_sym___nonnull] = ACTIONS(2604),
    [anon_sym___nullable] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2604),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2604),
    [anon_sym___real] = ACTIONS(2604),
    [anon_sym___strong] = ACTIONS(2604),
    [anon_sym___unsafe_unretained] = ACTIONS(2604),
    [anon_sym___unused] = ACTIONS(2604),
    [anon_sym___weak] = ACTIONS(2604),
    [anon_sym_alignas] = ACTIONS(2604),
    [anon_sym__Alignas] = ACTIONS(2604),
    [sym_primitive_type] = ACTIONS(2604),
    [anon_sym_enum] = ACTIONS(2604),
    [anon_sym_struct] = ACTIONS(2604),
    [anon_sym_union] = ACTIONS(2604),
    [anon_sym_in] = ACTIONS(2604),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2604),
    [anon_sym___typeof] = ACTIONS(2604),
    [anon_sym_typeof] = ACTIONS(2604),
    [anon_sym_ATprivate] = ACTIONS(2606),
    [anon_sym_ATprotected] = ACTIONS(2606),
    [anon_sym_ATpackage] = ACTIONS(2606),
    [anon_sym_ATpublic] = ACTIONS(2606),
    [anon_sym_BOOL] = ACTIONS(2604),
    [anon_sym_IMP] = ACTIONS(2604),
    [anon_sym_SEL] = ACTIONS(2604),
    [anon_sym_Class] = ACTIONS(2604),
    [anon_sym_id] = ACTIONS(2604),
    [anon_sym_out] = ACTIONS(2604),
    [anon_sym_inout] = ACTIONS(2604),
    [anon_sym_bycopy] = ACTIONS(2604),
    [anon_sym_byref] = ACTIONS(2604),
    [anon_sym_oneway] = ACTIONS(2604),
  },
  [STATE(4064)] = {
    [sym_attribute_specifier] = STATE(3942),
    [sym_enumerator_list] = STATE(4079),
    [sym_identifier] = ACTIONS(7548),
    [anon_sym_LPAREN2] = ACTIONS(7550),
    [anon_sym_STAR] = ACTIONS(7550),
    [anon_sym_CARET] = ACTIONS(7550),
    [anon_sym___extension__] = ACTIONS(7548),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7548),
    [anon_sym_LBRACK] = ACTIONS(7550),
    [anon_sym___based] = ACTIONS(7548),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7548),
    [anon_sym_unsigned] = ACTIONS(7548),
    [anon_sym_long] = ACTIONS(7548),
    [anon_sym_short] = ACTIONS(7548),
    [anon_sym_const] = ACTIONS(7548),
    [anon_sym_constexpr] = ACTIONS(7548),
    [anon_sym_volatile] = ACTIONS(7548),
    [anon_sym_restrict] = ACTIONS(7548),
    [anon_sym___restrict__] = ACTIONS(7548),
    [anon_sym__Atomic] = ACTIONS(7548),
    [anon_sym__Noreturn] = ACTIONS(7548),
    [anon_sym__Nonnull] = ACTIONS(7548),
    [anon_sym_nullable] = ACTIONS(7548),
    [anon_sym__Complex] = ACTIONS(7548),
    [anon_sym__Nullable] = ACTIONS(7548),
    [anon_sym__Nullable_result] = ACTIONS(7548),
    [anon_sym__Null_unspecified] = ACTIONS(7548),
    [anon_sym___autoreleasing] = ACTIONS(7548),
    [anon_sym___block] = ACTIONS(7548),
    [anon_sym___bridge] = ACTIONS(7548),
    [anon_sym___bridge_retained] = ACTIONS(7548),
    [anon_sym___bridge_transfer] = ACTIONS(7548),
    [anon_sym___complex] = ACTIONS(7548),
    [anon_sym___const] = ACTIONS(7548),
    [anon_sym___imag] = ACTIONS(7548),
    [anon_sym___kindof] = ACTIONS(7548),
    [anon_sym___nonnull] = ACTIONS(7548),
    [anon_sym___nullable] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7548),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7548),
    [anon_sym___real] = ACTIONS(7548),
    [anon_sym___strong] = ACTIONS(7548),
    [anon_sym___unsafe_unretained] = ACTIONS(7548),
    [anon_sym___unused] = ACTIONS(7548),
    [anon_sym___weak] = ACTIONS(7548),
    [anon_sym_alignas] = ACTIONS(7548),
    [anon_sym__Alignas] = ACTIONS(7548),
    [sym_primitive_type] = ACTIONS(7548),
    [anon_sym_enum] = ACTIONS(7548),
    [anon_sym_COLON] = ACTIONS(7550),
    [anon_sym_struct] = ACTIONS(7548),
    [anon_sym_union] = ACTIONS(7548),
    [anon_sym_in] = ACTIONS(7548),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7548),
    [anon_sym___typeof] = ACTIONS(7548),
    [anon_sym_typeof] = ACTIONS(7548),
    [anon_sym_BOOL] = ACTIONS(7548),
    [anon_sym_IMP] = ACTIONS(7548),
    [anon_sym_SEL] = ACTIONS(7548),
    [anon_sym_Class] = ACTIONS(7548),
    [anon_sym_id] = ACTIONS(7548),
    [anon_sym_out] = ACTIONS(7548),
    [anon_sym_inout] = ACTIONS(7548),
    [anon_sym_bycopy] = ACTIONS(7548),
    [anon_sym_byref] = ACTIONS(7548),
    [anon_sym_oneway] = ACTIONS(7548),
  },
  [STATE(4065)] = {
    [sym_attribute_specifier] = STATE(3839),
    [sym_enumerator_list] = STATE(4090),
    [sym_identifier] = ACTIONS(7619),
    [anon_sym_LPAREN2] = ACTIONS(7621),
    [anon_sym_STAR] = ACTIONS(7621),
    [anon_sym_CARET] = ACTIONS(7621),
    [anon_sym___extension__] = ACTIONS(7619),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7619),
    [anon_sym_LBRACK] = ACTIONS(7621),
    [anon_sym___based] = ACTIONS(7619),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7619),
    [anon_sym_unsigned] = ACTIONS(7619),
    [anon_sym_long] = ACTIONS(7619),
    [anon_sym_short] = ACTIONS(7619),
    [anon_sym_const] = ACTIONS(7619),
    [anon_sym_constexpr] = ACTIONS(7619),
    [anon_sym_volatile] = ACTIONS(7619),
    [anon_sym_restrict] = ACTIONS(7619),
    [anon_sym___restrict__] = ACTIONS(7619),
    [anon_sym__Atomic] = ACTIONS(7619),
    [anon_sym__Noreturn] = ACTIONS(7619),
    [anon_sym__Nonnull] = ACTIONS(7619),
    [anon_sym_nullable] = ACTIONS(7619),
    [anon_sym__Complex] = ACTIONS(7619),
    [anon_sym__Nullable] = ACTIONS(7619),
    [anon_sym__Nullable_result] = ACTIONS(7619),
    [anon_sym__Null_unspecified] = ACTIONS(7619),
    [anon_sym___autoreleasing] = ACTIONS(7619),
    [anon_sym___block] = ACTIONS(7619),
    [anon_sym___bridge] = ACTIONS(7619),
    [anon_sym___bridge_retained] = ACTIONS(7619),
    [anon_sym___bridge_transfer] = ACTIONS(7619),
    [anon_sym___complex] = ACTIONS(7619),
    [anon_sym___const] = ACTIONS(7619),
    [anon_sym___imag] = ACTIONS(7619),
    [anon_sym___kindof] = ACTIONS(7619),
    [anon_sym___nonnull] = ACTIONS(7619),
    [anon_sym___nullable] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7619),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7619),
    [anon_sym___real] = ACTIONS(7619),
    [anon_sym___strong] = ACTIONS(7619),
    [anon_sym___unsafe_unretained] = ACTIONS(7619),
    [anon_sym___unused] = ACTIONS(7619),
    [anon_sym___weak] = ACTIONS(7619),
    [anon_sym_alignas] = ACTIONS(7619),
    [anon_sym__Alignas] = ACTIONS(7619),
    [sym_primitive_type] = ACTIONS(7619),
    [anon_sym_enum] = ACTIONS(7619),
    [anon_sym_COLON] = ACTIONS(8968),
    [anon_sym_struct] = ACTIONS(7619),
    [anon_sym_union] = ACTIONS(7619),
    [anon_sym_in] = ACTIONS(7619),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7619),
    [anon_sym___typeof] = ACTIONS(7619),
    [anon_sym_typeof] = ACTIONS(7619),
    [anon_sym_BOOL] = ACTIONS(7619),
    [anon_sym_IMP] = ACTIONS(7619),
    [anon_sym_SEL] = ACTIONS(7619),
    [anon_sym_Class] = ACTIONS(7619),
    [anon_sym_id] = ACTIONS(7619),
    [anon_sym_out] = ACTIONS(7619),
    [anon_sym_inout] = ACTIONS(7619),
    [anon_sym_bycopy] = ACTIONS(7619),
    [anon_sym_byref] = ACTIONS(7619),
    [anon_sym_oneway] = ACTIONS(7619),
  },
  [STATE(4066)] = {
    [sym_attribute_specifier] = STATE(4558),
    [sym_type_qualifier] = STATE(2487),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4181),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(2487),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4185),
    [sym_identifier] = ACTIONS(8840),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym___attribute__] = ACTIONS(37),
    [anon_sym___attribute] = ACTIONS(37),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(8842),
    [anon_sym_unsigned] = ACTIONS(8842),
    [anon_sym_long] = ACTIONS(8842),
    [anon_sym_short] = ACTIONS(8842),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(8844),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4067)] = {
    [sym_attribute_specifier] = STATE(3860),
    [sym_enumerator_list] = STATE(4099),
    [sym_identifier] = ACTIONS(7632),
    [anon_sym_LPAREN2] = ACTIONS(7634),
    [anon_sym_STAR] = ACTIONS(7634),
    [anon_sym_CARET] = ACTIONS(7634),
    [anon_sym___extension__] = ACTIONS(7632),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7632),
    [anon_sym_LBRACK] = ACTIONS(7634),
    [anon_sym___based] = ACTIONS(7632),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7632),
    [anon_sym_unsigned] = ACTIONS(7632),
    [anon_sym_long] = ACTIONS(7632),
    [anon_sym_short] = ACTIONS(7632),
    [anon_sym_const] = ACTIONS(7632),
    [anon_sym_constexpr] = ACTIONS(7632),
    [anon_sym_volatile] = ACTIONS(7632),
    [anon_sym_restrict] = ACTIONS(7632),
    [anon_sym___restrict__] = ACTIONS(7632),
    [anon_sym__Atomic] = ACTIONS(7632),
    [anon_sym__Noreturn] = ACTIONS(7632),
    [anon_sym__Nonnull] = ACTIONS(7632),
    [anon_sym_nullable] = ACTIONS(7632),
    [anon_sym__Complex] = ACTIONS(7632),
    [anon_sym__Nullable] = ACTIONS(7632),
    [anon_sym__Nullable_result] = ACTIONS(7632),
    [anon_sym__Null_unspecified] = ACTIONS(7632),
    [anon_sym___autoreleasing] = ACTIONS(7632),
    [anon_sym___block] = ACTIONS(7632),
    [anon_sym___bridge] = ACTIONS(7632),
    [anon_sym___bridge_retained] = ACTIONS(7632),
    [anon_sym___bridge_transfer] = ACTIONS(7632),
    [anon_sym___complex] = ACTIONS(7632),
    [anon_sym___const] = ACTIONS(7632),
    [anon_sym___imag] = ACTIONS(7632),
    [anon_sym___kindof] = ACTIONS(7632),
    [anon_sym___nonnull] = ACTIONS(7632),
    [anon_sym___nullable] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7632),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7632),
    [anon_sym___real] = ACTIONS(7632),
    [anon_sym___strong] = ACTIONS(7632),
    [anon_sym___unsafe_unretained] = ACTIONS(7632),
    [anon_sym___unused] = ACTIONS(7632),
    [anon_sym___weak] = ACTIONS(7632),
    [anon_sym_alignas] = ACTIONS(7632),
    [anon_sym__Alignas] = ACTIONS(7632),
    [sym_primitive_type] = ACTIONS(7632),
    [anon_sym_enum] = ACTIONS(7632),
    [anon_sym_COLON] = ACTIONS(8971),
    [anon_sym_struct] = ACTIONS(7632),
    [anon_sym_union] = ACTIONS(7632),
    [anon_sym_in] = ACTIONS(7632),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7632),
    [anon_sym___typeof] = ACTIONS(7632),
    [anon_sym_typeof] = ACTIONS(7632),
    [anon_sym_BOOL] = ACTIONS(7632),
    [anon_sym_IMP] = ACTIONS(7632),
    [anon_sym_SEL] = ACTIONS(7632),
    [anon_sym_Class] = ACTIONS(7632),
    [anon_sym_id] = ACTIONS(7632),
    [anon_sym_out] = ACTIONS(7632),
    [anon_sym_inout] = ACTIONS(7632),
    [anon_sym_bycopy] = ACTIONS(7632),
    [anon_sym_byref] = ACTIONS(7632),
    [anon_sym_oneway] = ACTIONS(7632),
  },
  [STATE(4068)] = {
    [sym_attribute_specifier] = STATE(3959),
    [sym_enumerator_list] = STATE(4102),
    [sym_identifier] = ACTIONS(7568),
    [anon_sym_LPAREN2] = ACTIONS(7570),
    [anon_sym_STAR] = ACTIONS(7570),
    [anon_sym_CARET] = ACTIONS(7570),
    [anon_sym___extension__] = ACTIONS(7568),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7568),
    [anon_sym_LBRACK] = ACTIONS(7570),
    [anon_sym___based] = ACTIONS(7568),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7568),
    [anon_sym_unsigned] = ACTIONS(7568),
    [anon_sym_long] = ACTIONS(7568),
    [anon_sym_short] = ACTIONS(7568),
    [anon_sym_const] = ACTIONS(7568),
    [anon_sym_constexpr] = ACTIONS(7568),
    [anon_sym_volatile] = ACTIONS(7568),
    [anon_sym_restrict] = ACTIONS(7568),
    [anon_sym___restrict__] = ACTIONS(7568),
    [anon_sym__Atomic] = ACTIONS(7568),
    [anon_sym__Noreturn] = ACTIONS(7568),
    [anon_sym__Nonnull] = ACTIONS(7568),
    [anon_sym_nullable] = ACTIONS(7568),
    [anon_sym__Complex] = ACTIONS(7568),
    [anon_sym__Nullable] = ACTIONS(7568),
    [anon_sym__Nullable_result] = ACTIONS(7568),
    [anon_sym__Null_unspecified] = ACTIONS(7568),
    [anon_sym___autoreleasing] = ACTIONS(7568),
    [anon_sym___block] = ACTIONS(7568),
    [anon_sym___bridge] = ACTIONS(7568),
    [anon_sym___bridge_retained] = ACTIONS(7568),
    [anon_sym___bridge_transfer] = ACTIONS(7568),
    [anon_sym___complex] = ACTIONS(7568),
    [anon_sym___const] = ACTIONS(7568),
    [anon_sym___imag] = ACTIONS(7568),
    [anon_sym___kindof] = ACTIONS(7568),
    [anon_sym___nonnull] = ACTIONS(7568),
    [anon_sym___nullable] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7568),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7568),
    [anon_sym___real] = ACTIONS(7568),
    [anon_sym___strong] = ACTIONS(7568),
    [anon_sym___unsafe_unretained] = ACTIONS(7568),
    [anon_sym___unused] = ACTIONS(7568),
    [anon_sym___weak] = ACTIONS(7568),
    [anon_sym_alignas] = ACTIONS(7568),
    [anon_sym__Alignas] = ACTIONS(7568),
    [sym_primitive_type] = ACTIONS(7568),
    [anon_sym_enum] = ACTIONS(7568),
    [anon_sym_COLON] = ACTIONS(7570),
    [anon_sym_struct] = ACTIONS(7568),
    [anon_sym_union] = ACTIONS(7568),
    [anon_sym_in] = ACTIONS(7568),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7568),
    [anon_sym___typeof] = ACTIONS(7568),
    [anon_sym_typeof] = ACTIONS(7568),
    [anon_sym_BOOL] = ACTIONS(7568),
    [anon_sym_IMP] = ACTIONS(7568),
    [anon_sym_SEL] = ACTIONS(7568),
    [anon_sym_Class] = ACTIONS(7568),
    [anon_sym_id] = ACTIONS(7568),
    [anon_sym_out] = ACTIONS(7568),
    [anon_sym_inout] = ACTIONS(7568),
    [anon_sym_bycopy] = ACTIONS(7568),
    [anon_sym_byref] = ACTIONS(7568),
    [anon_sym_oneway] = ACTIONS(7568),
  },
  [STATE(4069)] = {
    [sym_attribute_specifier] = STATE(3963),
    [sym_enumerator_list] = STATE(4085),
    [sym_identifier] = ACTIONS(7581),
    [anon_sym_LPAREN2] = ACTIONS(7583),
    [anon_sym_STAR] = ACTIONS(7583),
    [anon_sym_CARET] = ACTIONS(7583),
    [anon_sym___extension__] = ACTIONS(7581),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7581),
    [anon_sym_LBRACK] = ACTIONS(7583),
    [anon_sym___based] = ACTIONS(7581),
    [anon_sym_LBRACE] = ACTIONS(8794),
    [anon_sym_signed] = ACTIONS(7581),
    [anon_sym_unsigned] = ACTIONS(7581),
    [anon_sym_long] = ACTIONS(7581),
    [anon_sym_short] = ACTIONS(7581),
    [anon_sym_const] = ACTIONS(7581),
    [anon_sym_constexpr] = ACTIONS(7581),
    [anon_sym_volatile] = ACTIONS(7581),
    [anon_sym_restrict] = ACTIONS(7581),
    [anon_sym___restrict__] = ACTIONS(7581),
    [anon_sym__Atomic] = ACTIONS(7581),
    [anon_sym__Noreturn] = ACTIONS(7581),
    [anon_sym__Nonnull] = ACTIONS(7581),
    [anon_sym_nullable] = ACTIONS(7581),
    [anon_sym__Complex] = ACTIONS(7581),
    [anon_sym__Nullable] = ACTIONS(7581),
    [anon_sym__Nullable_result] = ACTIONS(7581),
    [anon_sym__Null_unspecified] = ACTIONS(7581),
    [anon_sym___autoreleasing] = ACTIONS(7581),
    [anon_sym___block] = ACTIONS(7581),
    [anon_sym___bridge] = ACTIONS(7581),
    [anon_sym___bridge_retained] = ACTIONS(7581),
    [anon_sym___bridge_transfer] = ACTIONS(7581),
    [anon_sym___complex] = ACTIONS(7581),
    [anon_sym___const] = ACTIONS(7581),
    [anon_sym___imag] = ACTIONS(7581),
    [anon_sym___kindof] = ACTIONS(7581),
    [anon_sym___nonnull] = ACTIONS(7581),
    [anon_sym___nullable] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7581),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7581),
    [anon_sym___real] = ACTIONS(7581),
    [anon_sym___strong] = ACTIONS(7581),
    [anon_sym___unsafe_unretained] = ACTIONS(7581),
    [anon_sym___unused] = ACTIONS(7581),
    [anon_sym___weak] = ACTIONS(7581),
    [anon_sym_alignas] = ACTIONS(7581),
    [anon_sym__Alignas] = ACTIONS(7581),
    [sym_primitive_type] = ACTIONS(7581),
    [anon_sym_enum] = ACTIONS(7581),
    [anon_sym_COLON] = ACTIONS(8974),
    [anon_sym_struct] = ACTIONS(7581),
    [anon_sym_union] = ACTIONS(7581),
    [anon_sym_in] = ACTIONS(7581),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7581),
    [anon_sym___typeof] = ACTIONS(7581),
    [anon_sym_typeof] = ACTIONS(7581),
    [anon_sym_BOOL] = ACTIONS(7581),
    [anon_sym_IMP] = ACTIONS(7581),
    [anon_sym_SEL] = ACTIONS(7581),
    [anon_sym_Class] = ACTIONS(7581),
    [anon_sym_id] = ACTIONS(7581),
    [anon_sym_out] = ACTIONS(7581),
    [anon_sym_inout] = ACTIONS(7581),
    [anon_sym_bycopy] = ACTIONS(7581),
    [anon_sym_byref] = ACTIONS(7581),
    [anon_sym_oneway] = ACTIONS(7581),
  },
  [STATE(4070)] = {
    [sym_identifier] = ACTIONS(5569),
    [aux_sym_preproc_if_token1] = ACTIONS(5569),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(5571),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(5571),
    [anon_sym___extension__] = ACTIONS(5569),
    [anon_sym___attribute__] = ACTIONS(5569),
    [anon_sym___attribute] = ACTIONS(5569),
    [anon_sym_noreturn] = ACTIONS(5569),
    [anon_sym_LBRACK] = ACTIONS(5571),
    [anon_sym_RBRACE] = ACTIONS(5571),
    [anon_sym_signed] = ACTIONS(5569),
    [anon_sym_unsigned] = ACTIONS(5569),
    [anon_sym_long] = ACTIONS(5569),
    [anon_sym_short] = ACTIONS(5569),
    [anon_sym_const] = ACTIONS(5569),
    [anon_sym_constexpr] = ACTIONS(5569),
    [anon_sym_volatile] = ACTIONS(5569),
    [anon_sym_restrict] = ACTIONS(5569),
    [anon_sym___restrict__] = ACTIONS(5569),
    [anon_sym__Atomic] = ACTIONS(5569),
    [anon_sym__Noreturn] = ACTIONS(5569),
    [anon_sym__Nonnull] = ACTIONS(5569),
    [anon_sym_nullable] = ACTIONS(5569),
    [anon_sym__Complex] = ACTIONS(5569),
    [anon_sym__Nullable] = ACTIONS(5569),
    [anon_sym__Nullable_result] = ACTIONS(5569),
    [anon_sym__Null_unspecified] = ACTIONS(5569),
    [anon_sym___autoreleasing] = ACTIONS(5569),
    [anon_sym___block] = ACTIONS(5569),
    [anon_sym___bridge] = ACTIONS(5569),
    [anon_sym___bridge_retained] = ACTIONS(5569),
    [anon_sym___bridge_transfer] = ACTIONS(5569),
    [anon_sym___complex] = ACTIONS(5569),
    [anon_sym___const] = ACTIONS(5569),
    [anon_sym___imag] = ACTIONS(5569),
    [anon_sym___kindof] = ACTIONS(5569),
    [anon_sym___nonnull] = ACTIONS(5569),
    [anon_sym___nullable] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(5569),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(5569),
    [anon_sym___real] = ACTIONS(5569),
    [anon_sym___strong] = ACTIONS(5569),
    [anon_sym___unsafe_unretained] = ACTIONS(5569),
    [anon_sym___unused] = ACTIONS(5569),
    [anon_sym___weak] = ACTIONS(5569),
    [anon_sym_alignas] = ACTIONS(5569),
    [anon_sym__Alignas] = ACTIONS(5569),
    [sym_primitive_type] = ACTIONS(5569),
    [anon_sym_enum] = ACTIONS(5569),
    [anon_sym_struct] = ACTIONS(5569),
    [anon_sym_union] = ACTIONS(5569),
    [anon_sym_in] = ACTIONS(5569),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(5569),
    [anon_sym___typeof] = ACTIONS(5569),
    [anon_sym_typeof] = ACTIONS(5569),
    [anon_sym_ATprivate] = ACTIONS(5571),
    [anon_sym_ATprotected] = ACTIONS(5571),
    [anon_sym_ATpackage] = ACTIONS(5571),
    [anon_sym_ATpublic] = ACTIONS(5571),
    [anon_sym_BOOL] = ACTIONS(5569),
    [anon_sym_IMP] = ACTIONS(5569),
    [anon_sym_SEL] = ACTIONS(5569),
    [anon_sym_Class] = ACTIONS(5569),
    [anon_sym_id] = ACTIONS(5569),
    [anon_sym_out] = ACTIONS(5569),
    [anon_sym_inout] = ACTIONS(5569),
    [anon_sym_bycopy] = ACTIONS(5569),
    [anon_sym_byref] = ACTIONS(5569),
    [anon_sym_oneway] = ACTIONS(5569),
  },
  [STATE(4071)] = {
    [sym_identifier] = ACTIONS(2958),
    [anon_sym_LPAREN2] = ACTIONS(2960),
    [anon_sym_BANG] = ACTIONS(2960),
    [anon_sym_TILDE] = ACTIONS(2960),
    [anon_sym_DASH] = ACTIONS(2958),
    [anon_sym_PLUS] = ACTIONS(2958),
    [anon_sym_STAR] = ACTIONS(2960),
    [anon_sym_CARET] = ACTIONS(2960),
    [anon_sym_AMP] = ACTIONS(2960),
    [anon_sym_SEMI] = ACTIONS(2960),
    [anon_sym___extension__] = ACTIONS(2958),
    [anon_sym_LBRACK] = ACTIONS(2960),
    [anon_sym_LBRACE] = ACTIONS(2960),
    [anon_sym_ATautoreleasepool] = ACTIONS(2960),
    [anon_sym_const] = ACTIONS(2958),
    [anon_sym___imag] = ACTIONS(2958),
    [anon_sym___real] = ACTIONS(2958),
    [anon_sym_struct] = ACTIONS(2958),
    [anon_sym_if] = ACTIONS(2958),
    [anon_sym_switch] = ACTIONS(2958),
    [anon_sym_case] = ACTIONS(2958),
    [anon_sym_default] = ACTIONS(2958),
    [anon_sym_while] = ACTIONS(2958),
    [anon_sym_do] = ACTIONS(2958),
    [anon_sym_for] = ACTIONS(2958),
    [anon_sym_in] = ACTIONS(2958),
    [anon_sym_return] = ACTIONS(2958),
    [anon_sym_break] = ACTIONS(2958),
    [anon_sym_continue] = ACTIONS(2958),
    [anon_sym_goto] = ACTIONS(2958),
    [anon_sym_DASH_DASH] = ACTIONS(2960),
    [anon_sym_PLUS_PLUS] = ACTIONS(2960),
    [anon_sym_sizeof] = ACTIONS(2958),
    [anon_sym___alignof__] = ACTIONS(2958),
    [anon_sym___alignof] = ACTIONS(2958),
    [anon_sym__alignof] = ACTIONS(2958),
    [anon_sym_alignof] = ACTIONS(2958),
    [anon_sym__Alignof] = ACTIONS(2958),
    [anon_sym_offsetof] = ACTIONS(2958),
    [anon_sym__Generic] = ACTIONS(2958),
    [anon_sym_asm] = ACTIONS(2958),
    [anon_sym___asm__] = ACTIONS(2958),
    [anon_sym___asm] = ACTIONS(2958),
    [sym_number_literal] = ACTIONS(2960),
    [anon_sym_L_SQUOTE] = ACTIONS(2960),
    [anon_sym_u_SQUOTE] = ACTIONS(2960),
    [anon_sym_U_SQUOTE] = ACTIONS(2960),
    [anon_sym_u8_SQUOTE] = ACTIONS(2960),
    [anon_sym_SQUOTE] = ACTIONS(2960),
    [anon_sym_AT] = ACTIONS(2958),
    [anon_sym_DQUOTE] = ACTIONS(2960),
    [anon_sym_L_DQUOTE] = ACTIONS(2960),
    [anon_sym_u_DQUOTE] = ACTIONS(2960),
    [anon_sym_U_DQUOTE] = ACTIONS(2960),
    [anon_sym_u8_DQUOTE] = ACTIONS(2960),
    [sym_true] = ACTIONS(2958),
    [sym_false] = ACTIONS(2958),
    [anon_sym_NULL] = ACTIONS(2958),
    [anon_sym_nullptr] = ACTIONS(2958),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2960),
    [anon_sym___try] = ACTIONS(2958),
    [anon_sym_ATthrow] = ACTIONS(2960),
    [anon_sym_ATselector] = ACTIONS(2960),
    [anon_sym_ATavailable] = ACTIONS(2960),
    [anon_sym___builtin_available] = ACTIONS(2958),
    [anon_sym_va_arg] = ACTIONS(2958),
    [anon_sym_ATencode] = ACTIONS(2960),
    [anon_sym_ATsynchronized] = ACTIONS(2960),
    [anon_sym_id] = ACTIONS(2958),
  },
  [STATE(4072)] = {
    [sym_identifier] = ACTIONS(2962),
    [anon_sym_LPAREN2] = ACTIONS(2964),
    [anon_sym_BANG] = ACTIONS(2964),
    [anon_sym_TILDE] = ACTIONS(2964),
    [anon_sym_DASH] = ACTIONS(2962),
    [anon_sym_PLUS] = ACTIONS(2962),
    [anon_sym_STAR] = ACTIONS(2964),
    [anon_sym_CARET] = ACTIONS(2964),
    [anon_sym_AMP] = ACTIONS(2964),
    [anon_sym_SEMI] = ACTIONS(2964),
    [anon_sym___extension__] = ACTIONS(2962),
    [anon_sym_LBRACK] = ACTIONS(2964),
    [anon_sym_LBRACE] = ACTIONS(2964),
    [anon_sym_ATautoreleasepool] = ACTIONS(2964),
    [anon_sym_const] = ACTIONS(2962),
    [anon_sym___imag] = ACTIONS(2962),
    [anon_sym___real] = ACTIONS(2962),
    [anon_sym_struct] = ACTIONS(2962),
    [anon_sym_if] = ACTIONS(2962),
    [anon_sym_switch] = ACTIONS(2962),
    [anon_sym_case] = ACTIONS(2962),
    [anon_sym_default] = ACTIONS(2962),
    [anon_sym_while] = ACTIONS(2962),
    [anon_sym_do] = ACTIONS(2962),
    [anon_sym_for] = ACTIONS(2962),
    [anon_sym_in] = ACTIONS(2962),
    [anon_sym_return] = ACTIONS(2962),
    [anon_sym_break] = ACTIONS(2962),
    [anon_sym_continue] = ACTIONS(2962),
    [anon_sym_goto] = ACTIONS(2962),
    [anon_sym_DASH_DASH] = ACTIONS(2964),
    [anon_sym_PLUS_PLUS] = ACTIONS(2964),
    [anon_sym_sizeof] = ACTIONS(2962),
    [anon_sym___alignof__] = ACTIONS(2962),
    [anon_sym___alignof] = ACTIONS(2962),
    [anon_sym__alignof] = ACTIONS(2962),
    [anon_sym_alignof] = ACTIONS(2962),
    [anon_sym__Alignof] = ACTIONS(2962),
    [anon_sym_offsetof] = ACTIONS(2962),
    [anon_sym__Generic] = ACTIONS(2962),
    [anon_sym_asm] = ACTIONS(2962),
    [anon_sym___asm__] = ACTIONS(2962),
    [anon_sym___asm] = ACTIONS(2962),
    [sym_number_literal] = ACTIONS(2964),
    [anon_sym_L_SQUOTE] = ACTIONS(2964),
    [anon_sym_u_SQUOTE] = ACTIONS(2964),
    [anon_sym_U_SQUOTE] = ACTIONS(2964),
    [anon_sym_u8_SQUOTE] = ACTIONS(2964),
    [anon_sym_SQUOTE] = ACTIONS(2964),
    [anon_sym_AT] = ACTIONS(2962),
    [anon_sym_DQUOTE] = ACTIONS(2964),
    [anon_sym_L_DQUOTE] = ACTIONS(2964),
    [anon_sym_u_DQUOTE] = ACTIONS(2964),
    [anon_sym_U_DQUOTE] = ACTIONS(2964),
    [anon_sym_u8_DQUOTE] = ACTIONS(2964),
    [sym_true] = ACTIONS(2962),
    [sym_false] = ACTIONS(2962),
    [anon_sym_NULL] = ACTIONS(2962),
    [anon_sym_nullptr] = ACTIONS(2962),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(2964),
    [anon_sym___try] = ACTIONS(2962),
    [anon_sym_ATthrow] = ACTIONS(2964),
    [anon_sym_ATselector] = ACTIONS(2964),
    [anon_sym_ATavailable] = ACTIONS(2964),
    [anon_sym___builtin_available] = ACTIONS(2962),
    [anon_sym_va_arg] = ACTIONS(2962),
    [anon_sym_ATencode] = ACTIONS(2964),
    [anon_sym_ATsynchronized] = ACTIONS(2964),
    [anon_sym_id] = ACTIONS(2962),
  },
  [STATE(4073)] = {
    [sym_identifier] = ACTIONS(4232),
    [anon_sym_LPAREN2] = ACTIONS(4234),
    [anon_sym_BANG] = ACTIONS(4234),
    [anon_sym_TILDE] = ACTIONS(4234),
    [anon_sym_DASH] = ACTIONS(4232),
    [anon_sym_PLUS] = ACTIONS(4232),
    [anon_sym_STAR] = ACTIONS(4234),
    [anon_sym_CARET] = ACTIONS(4234),
    [anon_sym_AMP] = ACTIONS(4234),
    [anon_sym_SEMI] = ACTIONS(4234),
    [anon_sym___extension__] = ACTIONS(4232),
    [anon_sym_LBRACK] = ACTIONS(4234),
    [anon_sym_LBRACE] = ACTIONS(4234),
    [anon_sym_ATautoreleasepool] = ACTIONS(4234),
    [anon_sym_const] = ACTIONS(4232),
    [anon_sym___imag] = ACTIONS(4232),
    [anon_sym___real] = ACTIONS(4232),
    [anon_sym_struct] = ACTIONS(4232),
    [anon_sym_if] = ACTIONS(4232),
    [anon_sym_switch] = ACTIONS(4232),
    [anon_sym_case] = ACTIONS(4232),
    [anon_sym_default] = ACTIONS(4232),
    [anon_sym_while] = ACTIONS(4232),
    [anon_sym_do] = ACTIONS(4232),
    [anon_sym_for] = ACTIONS(4232),
    [anon_sym_in] = ACTIONS(4232),
    [anon_sym_return] = ACTIONS(4232),
    [anon_sym_break] = ACTIONS(4232),
    [anon_sym_continue] = ACTIONS(4232),
    [anon_sym_goto] = ACTIONS(4232),
    [anon_sym_DASH_DASH] = ACTIONS(4234),
    [anon_sym_PLUS_PLUS] = ACTIONS(4234),
    [anon_sym_sizeof] = ACTIONS(4232),
    [anon_sym___alignof__] = ACTIONS(4232),
    [anon_sym___alignof] = ACTIONS(4232),
    [anon_sym__alignof] = ACTIONS(4232),
    [anon_sym_alignof] = ACTIONS(4232),
    [anon_sym__Alignof] = ACTIONS(4232),
    [anon_sym_offsetof] = ACTIONS(4232),
    [anon_sym__Generic] = ACTIONS(4232),
    [anon_sym_asm] = ACTIONS(4232),
    [anon_sym___asm__] = ACTIONS(4232),
    [anon_sym___asm] = ACTIONS(4232),
    [sym_number_literal] = ACTIONS(4234),
    [anon_sym_L_SQUOTE] = ACTIONS(4234),
    [anon_sym_u_SQUOTE] = ACTIONS(4234),
    [anon_sym_U_SQUOTE] = ACTIONS(4234),
    [anon_sym_u8_SQUOTE] = ACTIONS(4234),
    [anon_sym_SQUOTE] = ACTIONS(4234),
    [anon_sym_AT] = ACTIONS(4232),
    [anon_sym_DQUOTE] = ACTIONS(4234),
    [anon_sym_L_DQUOTE] = ACTIONS(4234),
    [anon_sym_u_DQUOTE] = ACTIONS(4234),
    [anon_sym_U_DQUOTE] = ACTIONS(4234),
    [anon_sym_u8_DQUOTE] = ACTIONS(4234),
    [sym_true] = ACTIONS(4232),
    [sym_false] = ACTIONS(4232),
    [anon_sym_NULL] = ACTIONS(4232),
    [anon_sym_nullptr] = ACTIONS(4232),
    [sym_comment] = ACTIONS(3),
    [anon_sym_ATtry] = ACTIONS(4234),
    [anon_sym___try] = ACTIONS(4232),
    [anon_sym_ATthrow] = ACTIONS(4234),
    [anon_sym_ATselector] = ACTIONS(4234),
    [anon_sym_ATavailable] = ACTIONS(4234),
    [anon_sym___builtin_available] = ACTIONS(4232),
    [anon_sym_va_arg] = ACTIONS(4232),
    [anon_sym_ATencode] = ACTIONS(4234),
    [anon_sym_ATsynchronized] = ACTIONS(4234),
    [anon_sym_id] = ACTIONS(4232),
  },
  [STATE(4074)] = {
    [sym_attribute_specifier] = STATE(3964),
    [sym_identifier] = ACTIONS(7789),
    [anon_sym_LPAREN2] = ACTIONS(7791),
    [anon_sym_STAR] = ACTIONS(7791),
    [anon_sym_CARET] = ACTIONS(7791),
    [anon_sym___extension__] = ACTIONS(7789),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7789),
    [anon_sym_LBRACK] = ACTIONS(7791),
    [anon_sym___based] = ACTIONS(7789),
    [anon_sym_signed] = ACTIONS(7789),
    [anon_sym_unsigned] = ACTIONS(7789),
    [anon_sym_long] = ACTIONS(7789),
    [anon_sym_short] = ACTIONS(7789),
    [anon_sym_const] = ACTIONS(7789),
    [anon_sym_constexpr] = ACTIONS(7789),
    [anon_sym_volatile] = ACTIONS(7789),
    [anon_sym_restrict] = ACTIONS(7789),
    [anon_sym___restrict__] = ACTIONS(7789),
    [anon_sym__Atomic] = ACTIONS(7789),
    [anon_sym__Noreturn] = ACTIONS(7789),
    [anon_sym__Nonnull] = ACTIONS(7789),
    [anon_sym_nullable] = ACTIONS(7789),
    [anon_sym__Complex] = ACTIONS(7789),
    [anon_sym__Nullable] = ACTIONS(7789),
    [anon_sym__Nullable_result] = ACTIONS(7789),
    [anon_sym__Null_unspecified] = ACTIONS(7789),
    [anon_sym___autoreleasing] = ACTIONS(7789),
    [anon_sym___block] = ACTIONS(7789),
    [anon_sym___bridge] = ACTIONS(7789),
    [anon_sym___bridge_retained] = ACTIONS(7789),
    [anon_sym___bridge_transfer] = ACTIONS(7789),
    [anon_sym___complex] = ACTIONS(7789),
    [anon_sym___const] = ACTIONS(7789),
    [anon_sym___imag] = ACTIONS(7789),
    [anon_sym___kindof] = ACTIONS(7789),
    [anon_sym___nonnull] = ACTIONS(7789),
    [anon_sym___nullable] = ACTIONS(7789),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7789),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7789),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7789),
    [anon_sym___real] = ACTIONS(7789),
    [anon_sym___strong] = ACTIONS(7789),
    [anon_sym___unsafe_unretained] = ACTIONS(7789),
    [anon_sym___unused] = ACTIONS(7789),
    [anon_sym___weak] = ACTIONS(7789),
    [anon_sym_alignas] = ACTIONS(7789),
    [anon_sym__Alignas] = ACTIONS(7789),
    [sym_primitive_type] = ACTIONS(7789),
    [anon_sym_enum] = ACTIONS(7789),
    [anon_sym_COLON] = ACTIONS(7791),
    [anon_sym_struct] = ACTIONS(7789),
    [anon_sym_union] = ACTIONS(7789),
    [anon_sym_in] = ACTIONS(7789),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7789),
    [anon_sym___typeof] = ACTIONS(7789),
    [anon_sym_typeof] = ACTIONS(7789),
    [anon_sym_BOOL] = ACTIONS(7789),
    [anon_sym_IMP] = ACTIONS(7789),
    [anon_sym_SEL] = ACTIONS(7789),
    [anon_sym_Class] = ACTIONS(7789),
    [anon_sym_id] = ACTIONS(7789),
    [anon_sym_out] = ACTIONS(7789),
    [anon_sym_inout] = ACTIONS(7789),
    [anon_sym_bycopy] = ACTIONS(7789),
    [anon_sym_byref] = ACTIONS(7789),
    [anon_sym_oneway] = ACTIONS(7789),
  },
  [STATE(4075)] = {
    [sym_attribute_specifier] = STATE(3901),
    [sym_identifier] = ACTIONS(7782),
    [anon_sym_LPAREN2] = ACTIONS(7784),
    [anon_sym_STAR] = ACTIONS(7784),
    [anon_sym_CARET] = ACTIONS(7784),
    [anon_sym___extension__] = ACTIONS(7782),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7782),
    [anon_sym_LBRACK] = ACTIONS(7784),
    [anon_sym___based] = ACTIONS(7782),
    [anon_sym_signed] = ACTIONS(7782),
    [anon_sym_unsigned] = ACTIONS(7782),
    [anon_sym_long] = ACTIONS(7782),
    [anon_sym_short] = ACTIONS(7782),
    [anon_sym_const] = ACTIONS(7782),
    [anon_sym_constexpr] = ACTIONS(7782),
    [anon_sym_volatile] = ACTIONS(7782),
    [anon_sym_restrict] = ACTIONS(7782),
    [anon_sym___restrict__] = ACTIONS(7782),
    [anon_sym__Atomic] = ACTIONS(7782),
    [anon_sym__Noreturn] = ACTIONS(7782),
    [anon_sym__Nonnull] = ACTIONS(7782),
    [anon_sym_nullable] = ACTIONS(7782),
    [anon_sym__Complex] = ACTIONS(7782),
    [anon_sym__Nullable] = ACTIONS(7782),
    [anon_sym__Nullable_result] = ACTIONS(7782),
    [anon_sym__Null_unspecified] = ACTIONS(7782),
    [anon_sym___autoreleasing] = ACTIONS(7782),
    [anon_sym___block] = ACTIONS(7782),
    [anon_sym___bridge] = ACTIONS(7782),
    [anon_sym___bridge_retained] = ACTIONS(7782),
    [anon_sym___bridge_transfer] = ACTIONS(7782),
    [anon_sym___complex] = ACTIONS(7782),
    [anon_sym___const] = ACTIONS(7782),
    [anon_sym___imag] = ACTIONS(7782),
    [anon_sym___kindof] = ACTIONS(7782),
    [anon_sym___nonnull] = ACTIONS(7782),
    [anon_sym___nullable] = ACTIONS(7782),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7782),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7782),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7782),
    [anon_sym___real] = ACTIONS(7782),
    [anon_sym___strong] = ACTIONS(7782),
    [anon_sym___unsafe_unretained] = ACTIONS(7782),
    [anon_sym___unused] = ACTIONS(7782),
    [anon_sym___weak] = ACTIONS(7782),
    [anon_sym_alignas] = ACTIONS(7782),
    [anon_sym__Alignas] = ACTIONS(7782),
    [sym_primitive_type] = ACTIONS(7782),
    [anon_sym_enum] = ACTIONS(7782),
    [anon_sym_COLON] = ACTIONS(7784),
    [anon_sym_struct] = ACTIONS(7782),
    [anon_sym_union] = ACTIONS(7782),
    [anon_sym_in] = ACTIONS(7782),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7782),
    [anon_sym___typeof] = ACTIONS(7782),
    [anon_sym_typeof] = ACTIONS(7782),
    [anon_sym_BOOL] = ACTIONS(7782),
    [anon_sym_IMP] = ACTIONS(7782),
    [anon_sym_SEL] = ACTIONS(7782),
    [anon_sym_Class] = ACTIONS(7782),
    [anon_sym_id] = ACTIONS(7782),
    [anon_sym_out] = ACTIONS(7782),
    [anon_sym_inout] = ACTIONS(7782),
    [anon_sym_bycopy] = ACTIONS(7782),
    [anon_sym_byref] = ACTIONS(7782),
    [anon_sym_oneway] = ACTIONS(7782),
  },
  [STATE(4076)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6236),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4077)] = {
    [sym_type_qualifier] = STATE(4104),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6709),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4104),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(8977),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4078)] = {
    [sym_attribute_specifier] = STATE(3873),
    [sym_identifier] = ACTIONS(7875),
    [anon_sym_LPAREN2] = ACTIONS(7877),
    [anon_sym_STAR] = ACTIONS(7877),
    [anon_sym_CARET] = ACTIONS(7877),
    [anon_sym___extension__] = ACTIONS(7875),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7875),
    [anon_sym_LBRACK] = ACTIONS(7877),
    [anon_sym___based] = ACTIONS(7875),
    [anon_sym_signed] = ACTIONS(7875),
    [anon_sym_unsigned] = ACTIONS(7875),
    [anon_sym_long] = ACTIONS(7875),
    [anon_sym_short] = ACTIONS(7875),
    [anon_sym_const] = ACTIONS(7875),
    [anon_sym_constexpr] = ACTIONS(7875),
    [anon_sym_volatile] = ACTIONS(7875),
    [anon_sym_restrict] = ACTIONS(7875),
    [anon_sym___restrict__] = ACTIONS(7875),
    [anon_sym__Atomic] = ACTIONS(7875),
    [anon_sym__Noreturn] = ACTIONS(7875),
    [anon_sym__Nonnull] = ACTIONS(7875),
    [anon_sym_nullable] = ACTIONS(7875),
    [anon_sym__Complex] = ACTIONS(7875),
    [anon_sym__Nullable] = ACTIONS(7875),
    [anon_sym__Nullable_result] = ACTIONS(7875),
    [anon_sym__Null_unspecified] = ACTIONS(7875),
    [anon_sym___autoreleasing] = ACTIONS(7875),
    [anon_sym___block] = ACTIONS(7875),
    [anon_sym___bridge] = ACTIONS(7875),
    [anon_sym___bridge_retained] = ACTIONS(7875),
    [anon_sym___bridge_transfer] = ACTIONS(7875),
    [anon_sym___complex] = ACTIONS(7875),
    [anon_sym___const] = ACTIONS(7875),
    [anon_sym___imag] = ACTIONS(7875),
    [anon_sym___kindof] = ACTIONS(7875),
    [anon_sym___nonnull] = ACTIONS(7875),
    [anon_sym___nullable] = ACTIONS(7875),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7875),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7875),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7875),
    [anon_sym___real] = ACTIONS(7875),
    [anon_sym___strong] = ACTIONS(7875),
    [anon_sym___unsafe_unretained] = ACTIONS(7875),
    [anon_sym___unused] = ACTIONS(7875),
    [anon_sym___weak] = ACTIONS(7875),
    [anon_sym_alignas] = ACTIONS(7875),
    [anon_sym__Alignas] = ACTIONS(7875),
    [sym_primitive_type] = ACTIONS(7875),
    [anon_sym_enum] = ACTIONS(7875),
    [anon_sym_COLON] = ACTIONS(7877),
    [anon_sym_struct] = ACTIONS(7875),
    [anon_sym_union] = ACTIONS(7875),
    [anon_sym_in] = ACTIONS(7875),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7875),
    [anon_sym___typeof] = ACTIONS(7875),
    [anon_sym_typeof] = ACTIONS(7875),
    [anon_sym_BOOL] = ACTIONS(7875),
    [anon_sym_IMP] = ACTIONS(7875),
    [anon_sym_SEL] = ACTIONS(7875),
    [anon_sym_Class] = ACTIONS(7875),
    [anon_sym_id] = ACTIONS(7875),
    [anon_sym_out] = ACTIONS(7875),
    [anon_sym_inout] = ACTIONS(7875),
    [anon_sym_bycopy] = ACTIONS(7875),
    [anon_sym_byref] = ACTIONS(7875),
    [anon_sym_oneway] = ACTIONS(7875),
  },
  [STATE(4079)] = {
    [sym_attribute_specifier] = STATE(3874),
    [sym_identifier] = ACTIONS(7905),
    [anon_sym_LPAREN2] = ACTIONS(7907),
    [anon_sym_STAR] = ACTIONS(7907),
    [anon_sym_CARET] = ACTIONS(7907),
    [anon_sym___extension__] = ACTIONS(7905),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7905),
    [anon_sym_LBRACK] = ACTIONS(7907),
    [anon_sym___based] = ACTIONS(7905),
    [anon_sym_signed] = ACTIONS(7905),
    [anon_sym_unsigned] = ACTIONS(7905),
    [anon_sym_long] = ACTIONS(7905),
    [anon_sym_short] = ACTIONS(7905),
    [anon_sym_const] = ACTIONS(7905),
    [anon_sym_constexpr] = ACTIONS(7905),
    [anon_sym_volatile] = ACTIONS(7905),
    [anon_sym_restrict] = ACTIONS(7905),
    [anon_sym___restrict__] = ACTIONS(7905),
    [anon_sym__Atomic] = ACTIONS(7905),
    [anon_sym__Noreturn] = ACTIONS(7905),
    [anon_sym__Nonnull] = ACTIONS(7905),
    [anon_sym_nullable] = ACTIONS(7905),
    [anon_sym__Complex] = ACTIONS(7905),
    [anon_sym__Nullable] = ACTIONS(7905),
    [anon_sym__Nullable_result] = ACTIONS(7905),
    [anon_sym__Null_unspecified] = ACTIONS(7905),
    [anon_sym___autoreleasing] = ACTIONS(7905),
    [anon_sym___block] = ACTIONS(7905),
    [anon_sym___bridge] = ACTIONS(7905),
    [anon_sym___bridge_retained] = ACTIONS(7905),
    [anon_sym___bridge_transfer] = ACTIONS(7905),
    [anon_sym___complex] = ACTIONS(7905),
    [anon_sym___const] = ACTIONS(7905),
    [anon_sym___imag] = ACTIONS(7905),
    [anon_sym___kindof] = ACTIONS(7905),
    [anon_sym___nonnull] = ACTIONS(7905),
    [anon_sym___nullable] = ACTIONS(7905),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7905),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7905),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7905),
    [anon_sym___real] = ACTIONS(7905),
    [anon_sym___strong] = ACTIONS(7905),
    [anon_sym___unsafe_unretained] = ACTIONS(7905),
    [anon_sym___unused] = ACTIONS(7905),
    [anon_sym___weak] = ACTIONS(7905),
    [anon_sym_alignas] = ACTIONS(7905),
    [anon_sym__Alignas] = ACTIONS(7905),
    [sym_primitive_type] = ACTIONS(7905),
    [anon_sym_enum] = ACTIONS(7905),
    [anon_sym_COLON] = ACTIONS(7907),
    [anon_sym_struct] = ACTIONS(7905),
    [anon_sym_union] = ACTIONS(7905),
    [anon_sym_in] = ACTIONS(7905),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7905),
    [anon_sym___typeof] = ACTIONS(7905),
    [anon_sym_typeof] = ACTIONS(7905),
    [anon_sym_BOOL] = ACTIONS(7905),
    [anon_sym_IMP] = ACTIONS(7905),
    [anon_sym_SEL] = ACTIONS(7905),
    [anon_sym_Class] = ACTIONS(7905),
    [anon_sym_id] = ACTIONS(7905),
    [anon_sym_out] = ACTIONS(7905),
    [anon_sym_inout] = ACTIONS(7905),
    [anon_sym_bycopy] = ACTIONS(7905),
    [anon_sym_byref] = ACTIONS(7905),
    [anon_sym_oneway] = ACTIONS(7905),
  },
  [STATE(4080)] = {
    [sym_attribute_specifier] = STATE(3840),
    [sym_identifier] = ACTIONS(7806),
    [anon_sym_LPAREN2] = ACTIONS(7808),
    [anon_sym_STAR] = ACTIONS(7808),
    [anon_sym_CARET] = ACTIONS(7808),
    [anon_sym___extension__] = ACTIONS(7806),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7806),
    [anon_sym_LBRACK] = ACTIONS(7808),
    [anon_sym___based] = ACTIONS(7806),
    [anon_sym_signed] = ACTIONS(7806),
    [anon_sym_unsigned] = ACTIONS(7806),
    [anon_sym_long] = ACTIONS(7806),
    [anon_sym_short] = ACTIONS(7806),
    [anon_sym_const] = ACTIONS(7806),
    [anon_sym_constexpr] = ACTIONS(7806),
    [anon_sym_volatile] = ACTIONS(7806),
    [anon_sym_restrict] = ACTIONS(7806),
    [anon_sym___restrict__] = ACTIONS(7806),
    [anon_sym__Atomic] = ACTIONS(7806),
    [anon_sym__Noreturn] = ACTIONS(7806),
    [anon_sym__Nonnull] = ACTIONS(7806),
    [anon_sym_nullable] = ACTIONS(7806),
    [anon_sym__Complex] = ACTIONS(7806),
    [anon_sym__Nullable] = ACTIONS(7806),
    [anon_sym__Nullable_result] = ACTIONS(7806),
    [anon_sym__Null_unspecified] = ACTIONS(7806),
    [anon_sym___autoreleasing] = ACTIONS(7806),
    [anon_sym___block] = ACTIONS(7806),
    [anon_sym___bridge] = ACTIONS(7806),
    [anon_sym___bridge_retained] = ACTIONS(7806),
    [anon_sym___bridge_transfer] = ACTIONS(7806),
    [anon_sym___complex] = ACTIONS(7806),
    [anon_sym___const] = ACTIONS(7806),
    [anon_sym___imag] = ACTIONS(7806),
    [anon_sym___kindof] = ACTIONS(7806),
    [anon_sym___nonnull] = ACTIONS(7806),
    [anon_sym___nullable] = ACTIONS(7806),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7806),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7806),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7806),
    [anon_sym___real] = ACTIONS(7806),
    [anon_sym___strong] = ACTIONS(7806),
    [anon_sym___unsafe_unretained] = ACTIONS(7806),
    [anon_sym___unused] = ACTIONS(7806),
    [anon_sym___weak] = ACTIONS(7806),
    [anon_sym_alignas] = ACTIONS(7806),
    [anon_sym__Alignas] = ACTIONS(7806),
    [sym_primitive_type] = ACTIONS(7806),
    [anon_sym_enum] = ACTIONS(7806),
    [anon_sym_COLON] = ACTIONS(7808),
    [anon_sym_struct] = ACTIONS(7806),
    [anon_sym_union] = ACTIONS(7806),
    [anon_sym_in] = ACTIONS(7806),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7806),
    [anon_sym___typeof] = ACTIONS(7806),
    [anon_sym_typeof] = ACTIONS(7806),
    [anon_sym_BOOL] = ACTIONS(7806),
    [anon_sym_IMP] = ACTIONS(7806),
    [anon_sym_SEL] = ACTIONS(7806),
    [anon_sym_Class] = ACTIONS(7806),
    [anon_sym_id] = ACTIONS(7806),
    [anon_sym_out] = ACTIONS(7806),
    [anon_sym_inout] = ACTIONS(7806),
    [anon_sym_bycopy] = ACTIONS(7806),
    [anon_sym_byref] = ACTIONS(7806),
    [anon_sym_oneway] = ACTIONS(7806),
  },
  [STATE(4081)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6439),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4082)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6078),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4083)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6565),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4084)] = {
    [sym_type_qualifier] = STATE(4104),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(5914),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4104),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(8977),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4085)] = {
    [sym_attribute_specifier] = STATE(3877),
    [sym_identifier] = ACTIONS(7817),
    [anon_sym_LPAREN2] = ACTIONS(7819),
    [anon_sym_STAR] = ACTIONS(7819),
    [anon_sym_CARET] = ACTIONS(7819),
    [anon_sym___extension__] = ACTIONS(7817),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7817),
    [anon_sym_LBRACK] = ACTIONS(7819),
    [anon_sym___based] = ACTIONS(7817),
    [anon_sym_signed] = ACTIONS(7817),
    [anon_sym_unsigned] = ACTIONS(7817),
    [anon_sym_long] = ACTIONS(7817),
    [anon_sym_short] = ACTIONS(7817),
    [anon_sym_const] = ACTIONS(7817),
    [anon_sym_constexpr] = ACTIONS(7817),
    [anon_sym_volatile] = ACTIONS(7817),
    [anon_sym_restrict] = ACTIONS(7817),
    [anon_sym___restrict__] = ACTIONS(7817),
    [anon_sym__Atomic] = ACTIONS(7817),
    [anon_sym__Noreturn] = ACTIONS(7817),
    [anon_sym__Nonnull] = ACTIONS(7817),
    [anon_sym_nullable] = ACTIONS(7817),
    [anon_sym__Complex] = ACTIONS(7817),
    [anon_sym__Nullable] = ACTIONS(7817),
    [anon_sym__Nullable_result] = ACTIONS(7817),
    [anon_sym__Null_unspecified] = ACTIONS(7817),
    [anon_sym___autoreleasing] = ACTIONS(7817),
    [anon_sym___block] = ACTIONS(7817),
    [anon_sym___bridge] = ACTIONS(7817),
    [anon_sym___bridge_retained] = ACTIONS(7817),
    [anon_sym___bridge_transfer] = ACTIONS(7817),
    [anon_sym___complex] = ACTIONS(7817),
    [anon_sym___const] = ACTIONS(7817),
    [anon_sym___imag] = ACTIONS(7817),
    [anon_sym___kindof] = ACTIONS(7817),
    [anon_sym___nonnull] = ACTIONS(7817),
    [anon_sym___nullable] = ACTIONS(7817),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7817),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7817),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7817),
    [anon_sym___real] = ACTIONS(7817),
    [anon_sym___strong] = ACTIONS(7817),
    [anon_sym___unsafe_unretained] = ACTIONS(7817),
    [anon_sym___unused] = ACTIONS(7817),
    [anon_sym___weak] = ACTIONS(7817),
    [anon_sym_alignas] = ACTIONS(7817),
    [anon_sym__Alignas] = ACTIONS(7817),
    [sym_primitive_type] = ACTIONS(7817),
    [anon_sym_enum] = ACTIONS(7817),
    [anon_sym_COLON] = ACTIONS(7819),
    [anon_sym_struct] = ACTIONS(7817),
    [anon_sym_union] = ACTIONS(7817),
    [anon_sym_in] = ACTIONS(7817),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7817),
    [anon_sym___typeof] = ACTIONS(7817),
    [anon_sym_typeof] = ACTIONS(7817),
    [anon_sym_BOOL] = ACTIONS(7817),
    [anon_sym_IMP] = ACTIONS(7817),
    [anon_sym_SEL] = ACTIONS(7817),
    [anon_sym_Class] = ACTIONS(7817),
    [anon_sym_id] = ACTIONS(7817),
    [anon_sym_out] = ACTIONS(7817),
    [anon_sym_inout] = ACTIONS(7817),
    [anon_sym_bycopy] = ACTIONS(7817),
    [anon_sym_byref] = ACTIONS(7817),
    [anon_sym_oneway] = ACTIONS(7817),
  },
  [STATE(4086)] = {
    [sym_type_qualifier] = STATE(4104),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6268),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4104),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(8977),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4087)] = {
    [aux_sym_sized_type_specifier_repeat1] = STATE(3812),
    [aux_sym_generic_specifier_repeat1] = STATE(3813),
    [sym_identifier] = ACTIONS(4494),
    [anon_sym_LPAREN2] = ACTIONS(8979),
    [anon_sym_STAR] = ACTIONS(6134),
    [anon_sym_CARET] = ACTIONS(6134),
    [anon_sym_LT] = ACTIONS(6136),
    [anon_sym___extension__] = ACTIONS(4494),
    [anon_sym_noreturn] = ACTIONS(4494),
    [anon_sym_LBRACK] = ACTIONS(6134),
    [anon_sym___based] = ACTIONS(4494),
    [anon_sym_signed] = ACTIONS(6146),
    [anon_sym_unsigned] = ACTIONS(6146),
    [anon_sym_long] = ACTIONS(6146),
    [anon_sym_short] = ACTIONS(6146),
    [anon_sym_const] = ACTIONS(4494),
    [anon_sym_constexpr] = ACTIONS(4494),
    [anon_sym_volatile] = ACTIONS(4494),
    [anon_sym_restrict] = ACTIONS(4494),
    [anon_sym___restrict__] = ACTIONS(4494),
    [anon_sym__Atomic] = ACTIONS(4494),
    [anon_sym__Noreturn] = ACTIONS(4494),
    [anon_sym__Nonnull] = ACTIONS(4494),
    [anon_sym_nullable] = ACTIONS(4494),
    [anon_sym__Complex] = ACTIONS(4494),
    [anon_sym__Nullable] = ACTIONS(4494),
    [anon_sym__Nullable_result] = ACTIONS(4494),
    [anon_sym__Null_unspecified] = ACTIONS(4494),
    [anon_sym___autoreleasing] = ACTIONS(4494),
    [anon_sym___block] = ACTIONS(4494),
    [anon_sym___bridge] = ACTIONS(4494),
    [anon_sym___bridge_retained] = ACTIONS(4494),
    [anon_sym___bridge_transfer] = ACTIONS(4494),
    [anon_sym___complex] = ACTIONS(4494),
    [anon_sym___const] = ACTIONS(4494),
    [anon_sym___imag] = ACTIONS(4494),
    [anon_sym___kindof] = ACTIONS(4494),
    [anon_sym___nonnull] = ACTIONS(4494),
    [anon_sym___nullable] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(4494),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(4494),
    [anon_sym___real] = ACTIONS(4494),
    [anon_sym___strong] = ACTIONS(4494),
    [anon_sym___unsafe_unretained] = ACTIONS(4494),
    [anon_sym___unused] = ACTIONS(4494),
    [anon_sym___weak] = ACTIONS(4494),
    [anon_sym_alignas] = ACTIONS(4494),
    [anon_sym__Alignas] = ACTIONS(4494),
    [sym_primitive_type] = ACTIONS(4494),
    [anon_sym_enum] = ACTIONS(4494),
    [anon_sym_COLON] = ACTIONS(6134),
    [anon_sym_struct] = ACTIONS(4494),
    [anon_sym_union] = ACTIONS(4494),
    [anon_sym_in] = ACTIONS(4494),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(4494),
    [anon_sym___typeof] = ACTIONS(4494),
    [anon_sym_typeof] = ACTIONS(4494),
    [anon_sym_BOOL] = ACTIONS(4494),
    [anon_sym_IMP] = ACTIONS(4494),
    [anon_sym_SEL] = ACTIONS(4494),
    [anon_sym_Class] = ACTIONS(4494),
    [anon_sym_id] = ACTIONS(4494),
    [anon_sym_out] = ACTIONS(4494),
    [anon_sym_inout] = ACTIONS(4494),
    [anon_sym_bycopy] = ACTIONS(4494),
    [anon_sym_byref] = ACTIONS(4494),
    [anon_sym_oneway] = ACTIONS(4494),
  },
  [STATE(4088)] = {
    [sym_attribute_specifier] = STATE(3889),
    [sym_identifier] = ACTIONS(7920),
    [anon_sym_LPAREN2] = ACTIONS(7922),
    [anon_sym_STAR] = ACTIONS(7922),
    [anon_sym_CARET] = ACTIONS(7922),
    [anon_sym___extension__] = ACTIONS(7920),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7920),
    [anon_sym_LBRACK] = ACTIONS(7922),
    [anon_sym___based] = ACTIONS(7920),
    [anon_sym_signed] = ACTIONS(7920),
    [anon_sym_unsigned] = ACTIONS(7920),
    [anon_sym_long] = ACTIONS(7920),
    [anon_sym_short] = ACTIONS(7920),
    [anon_sym_const] = ACTIONS(7920),
    [anon_sym_constexpr] = ACTIONS(7920),
    [anon_sym_volatile] = ACTIONS(7920),
    [anon_sym_restrict] = ACTIONS(7920),
    [anon_sym___restrict__] = ACTIONS(7920),
    [anon_sym__Atomic] = ACTIONS(7920),
    [anon_sym__Noreturn] = ACTIONS(7920),
    [anon_sym__Nonnull] = ACTIONS(7920),
    [anon_sym_nullable] = ACTIONS(7920),
    [anon_sym__Complex] = ACTIONS(7920),
    [anon_sym__Nullable] = ACTIONS(7920),
    [anon_sym__Nullable_result] = ACTIONS(7920),
    [anon_sym__Null_unspecified] = ACTIONS(7920),
    [anon_sym___autoreleasing] = ACTIONS(7920),
    [anon_sym___block] = ACTIONS(7920),
    [anon_sym___bridge] = ACTIONS(7920),
    [anon_sym___bridge_retained] = ACTIONS(7920),
    [anon_sym___bridge_transfer] = ACTIONS(7920),
    [anon_sym___complex] = ACTIONS(7920),
    [anon_sym___const] = ACTIONS(7920),
    [anon_sym___imag] = ACTIONS(7920),
    [anon_sym___kindof] = ACTIONS(7920),
    [anon_sym___nonnull] = ACTIONS(7920),
    [anon_sym___nullable] = ACTIONS(7920),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7920),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7920),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7920),
    [anon_sym___real] = ACTIONS(7920),
    [anon_sym___strong] = ACTIONS(7920),
    [anon_sym___unsafe_unretained] = ACTIONS(7920),
    [anon_sym___unused] = ACTIONS(7920),
    [anon_sym___weak] = ACTIONS(7920),
    [anon_sym_alignas] = ACTIONS(7920),
    [anon_sym__Alignas] = ACTIONS(7920),
    [sym_primitive_type] = ACTIONS(7920),
    [anon_sym_enum] = ACTIONS(7920),
    [anon_sym_COLON] = ACTIONS(7922),
    [anon_sym_struct] = ACTIONS(7920),
    [anon_sym_union] = ACTIONS(7920),
    [anon_sym_in] = ACTIONS(7920),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7920),
    [anon_sym___typeof] = ACTIONS(7920),
    [anon_sym_typeof] = ACTIONS(7920),
    [anon_sym_BOOL] = ACTIONS(7920),
    [anon_sym_IMP] = ACTIONS(7920),
    [anon_sym_SEL] = ACTIONS(7920),
    [anon_sym_Class] = ACTIONS(7920),
    [anon_sym_id] = ACTIONS(7920),
    [anon_sym_out] = ACTIONS(7920),
    [anon_sym_inout] = ACTIONS(7920),
    [anon_sym_bycopy] = ACTIONS(7920),
    [anon_sym_byref] = ACTIONS(7920),
    [anon_sym_oneway] = ACTIONS(7920),
  },
  [STATE(4089)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6681),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4090)] = {
    [sym_attribute_specifier] = STATE(3859),
    [sym_identifier] = ACTIONS(7854),
    [anon_sym_LPAREN2] = ACTIONS(7856),
    [anon_sym_STAR] = ACTIONS(7856),
    [anon_sym_CARET] = ACTIONS(7856),
    [anon_sym___extension__] = ACTIONS(7854),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7854),
    [anon_sym_LBRACK] = ACTIONS(7856),
    [anon_sym___based] = ACTIONS(7854),
    [anon_sym_signed] = ACTIONS(7854),
    [anon_sym_unsigned] = ACTIONS(7854),
    [anon_sym_long] = ACTIONS(7854),
    [anon_sym_short] = ACTIONS(7854),
    [anon_sym_const] = ACTIONS(7854),
    [anon_sym_constexpr] = ACTIONS(7854),
    [anon_sym_volatile] = ACTIONS(7854),
    [anon_sym_restrict] = ACTIONS(7854),
    [anon_sym___restrict__] = ACTIONS(7854),
    [anon_sym__Atomic] = ACTIONS(7854),
    [anon_sym__Noreturn] = ACTIONS(7854),
    [anon_sym__Nonnull] = ACTIONS(7854),
    [anon_sym_nullable] = ACTIONS(7854),
    [anon_sym__Complex] = ACTIONS(7854),
    [anon_sym__Nullable] = ACTIONS(7854),
    [anon_sym__Nullable_result] = ACTIONS(7854),
    [anon_sym__Null_unspecified] = ACTIONS(7854),
    [anon_sym___autoreleasing] = ACTIONS(7854),
    [anon_sym___block] = ACTIONS(7854),
    [anon_sym___bridge] = ACTIONS(7854),
    [anon_sym___bridge_retained] = ACTIONS(7854),
    [anon_sym___bridge_transfer] = ACTIONS(7854),
    [anon_sym___complex] = ACTIONS(7854),
    [anon_sym___const] = ACTIONS(7854),
    [anon_sym___imag] = ACTIONS(7854),
    [anon_sym___kindof] = ACTIONS(7854),
    [anon_sym___nonnull] = ACTIONS(7854),
    [anon_sym___nullable] = ACTIONS(7854),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7854),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7854),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7854),
    [anon_sym___real] = ACTIONS(7854),
    [anon_sym___strong] = ACTIONS(7854),
    [anon_sym___unsafe_unretained] = ACTIONS(7854),
    [anon_sym___unused] = ACTIONS(7854),
    [anon_sym___weak] = ACTIONS(7854),
    [anon_sym_alignas] = ACTIONS(7854),
    [anon_sym__Alignas] = ACTIONS(7854),
    [sym_primitive_type] = ACTIONS(7854),
    [anon_sym_enum] = ACTIONS(7854),
    [anon_sym_COLON] = ACTIONS(7856),
    [anon_sym_struct] = ACTIONS(7854),
    [anon_sym_union] = ACTIONS(7854),
    [anon_sym_in] = ACTIONS(7854),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7854),
    [anon_sym___typeof] = ACTIONS(7854),
    [anon_sym_typeof] = ACTIONS(7854),
    [anon_sym_BOOL] = ACTIONS(7854),
    [anon_sym_IMP] = ACTIONS(7854),
    [anon_sym_SEL] = ACTIONS(7854),
    [anon_sym_Class] = ACTIONS(7854),
    [anon_sym_id] = ACTIONS(7854),
    [anon_sym_out] = ACTIONS(7854),
    [anon_sym_inout] = ACTIONS(7854),
    [anon_sym_bycopy] = ACTIONS(7854),
    [anon_sym_byref] = ACTIONS(7854),
    [anon_sym_oneway] = ACTIONS(7854),
  },
  [STATE(4091)] = {
    [sym_attribute_specifier] = STATE(3861),
    [sym_identifier] = ACTIONS(7898),
    [anon_sym_LPAREN2] = ACTIONS(7900),
    [anon_sym_STAR] = ACTIONS(7900),
    [anon_sym_CARET] = ACTIONS(7900),
    [anon_sym___extension__] = ACTIONS(7898),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7898),
    [anon_sym_LBRACK] = ACTIONS(7900),
    [anon_sym___based] = ACTIONS(7898),
    [anon_sym_signed] = ACTIONS(7898),
    [anon_sym_unsigned] = ACTIONS(7898),
    [anon_sym_long] = ACTIONS(7898),
    [anon_sym_short] = ACTIONS(7898),
    [anon_sym_const] = ACTIONS(7898),
    [anon_sym_constexpr] = ACTIONS(7898),
    [anon_sym_volatile] = ACTIONS(7898),
    [anon_sym_restrict] = ACTIONS(7898),
    [anon_sym___restrict__] = ACTIONS(7898),
    [anon_sym__Atomic] = ACTIONS(7898),
    [anon_sym__Noreturn] = ACTIONS(7898),
    [anon_sym__Nonnull] = ACTIONS(7898),
    [anon_sym_nullable] = ACTIONS(7898),
    [anon_sym__Complex] = ACTIONS(7898),
    [anon_sym__Nullable] = ACTIONS(7898),
    [anon_sym__Nullable_result] = ACTIONS(7898),
    [anon_sym__Null_unspecified] = ACTIONS(7898),
    [anon_sym___autoreleasing] = ACTIONS(7898),
    [anon_sym___block] = ACTIONS(7898),
    [anon_sym___bridge] = ACTIONS(7898),
    [anon_sym___bridge_retained] = ACTIONS(7898),
    [anon_sym___bridge_transfer] = ACTIONS(7898),
    [anon_sym___complex] = ACTIONS(7898),
    [anon_sym___const] = ACTIONS(7898),
    [anon_sym___imag] = ACTIONS(7898),
    [anon_sym___kindof] = ACTIONS(7898),
    [anon_sym___nonnull] = ACTIONS(7898),
    [anon_sym___nullable] = ACTIONS(7898),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7898),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7898),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7898),
    [anon_sym___real] = ACTIONS(7898),
    [anon_sym___strong] = ACTIONS(7898),
    [anon_sym___unsafe_unretained] = ACTIONS(7898),
    [anon_sym___unused] = ACTIONS(7898),
    [anon_sym___weak] = ACTIONS(7898),
    [anon_sym_alignas] = ACTIONS(7898),
    [anon_sym__Alignas] = ACTIONS(7898),
    [sym_primitive_type] = ACTIONS(7898),
    [anon_sym_enum] = ACTIONS(7898),
    [anon_sym_COLON] = ACTIONS(7900),
    [anon_sym_struct] = ACTIONS(7898),
    [anon_sym_union] = ACTIONS(7898),
    [anon_sym_in] = ACTIONS(7898),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7898),
    [anon_sym___typeof] = ACTIONS(7898),
    [anon_sym_typeof] = ACTIONS(7898),
    [anon_sym_BOOL] = ACTIONS(7898),
    [anon_sym_IMP] = ACTIONS(7898),
    [anon_sym_SEL] = ACTIONS(7898),
    [anon_sym_Class] = ACTIONS(7898),
    [anon_sym_id] = ACTIONS(7898),
    [anon_sym_out] = ACTIONS(7898),
    [anon_sym_inout] = ACTIONS(7898),
    [anon_sym_bycopy] = ACTIONS(7898),
    [anon_sym_byref] = ACTIONS(7898),
    [anon_sym_oneway] = ACTIONS(7898),
  },
  [STATE(4092)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6628),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4093)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6368),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4094)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6411),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4095)] = {
    [sym_attribute_specifier] = STATE(3956),
    [sym_identifier] = ACTIONS(7927),
    [anon_sym_LPAREN2] = ACTIONS(7929),
    [anon_sym_STAR] = ACTIONS(7929),
    [anon_sym_CARET] = ACTIONS(7929),
    [anon_sym___extension__] = ACTIONS(7927),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7927),
    [anon_sym_LBRACK] = ACTIONS(7929),
    [anon_sym___based] = ACTIONS(7927),
    [anon_sym_signed] = ACTIONS(7927),
    [anon_sym_unsigned] = ACTIONS(7927),
    [anon_sym_long] = ACTIONS(7927),
    [anon_sym_short] = ACTIONS(7927),
    [anon_sym_const] = ACTIONS(7927),
    [anon_sym_constexpr] = ACTIONS(7927),
    [anon_sym_volatile] = ACTIONS(7927),
    [anon_sym_restrict] = ACTIONS(7927),
    [anon_sym___restrict__] = ACTIONS(7927),
    [anon_sym__Atomic] = ACTIONS(7927),
    [anon_sym__Noreturn] = ACTIONS(7927),
    [anon_sym__Nonnull] = ACTIONS(7927),
    [anon_sym_nullable] = ACTIONS(7927),
    [anon_sym__Complex] = ACTIONS(7927),
    [anon_sym__Nullable] = ACTIONS(7927),
    [anon_sym__Nullable_result] = ACTIONS(7927),
    [anon_sym__Null_unspecified] = ACTIONS(7927),
    [anon_sym___autoreleasing] = ACTIONS(7927),
    [anon_sym___block] = ACTIONS(7927),
    [anon_sym___bridge] = ACTIONS(7927),
    [anon_sym___bridge_retained] = ACTIONS(7927),
    [anon_sym___bridge_transfer] = ACTIONS(7927),
    [anon_sym___complex] = ACTIONS(7927),
    [anon_sym___const] = ACTIONS(7927),
    [anon_sym___imag] = ACTIONS(7927),
    [anon_sym___kindof] = ACTIONS(7927),
    [anon_sym___nonnull] = ACTIONS(7927),
    [anon_sym___nullable] = ACTIONS(7927),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7927),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7927),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7927),
    [anon_sym___real] = ACTIONS(7927),
    [anon_sym___strong] = ACTIONS(7927),
    [anon_sym___unsafe_unretained] = ACTIONS(7927),
    [anon_sym___unused] = ACTIONS(7927),
    [anon_sym___weak] = ACTIONS(7927),
    [anon_sym_alignas] = ACTIONS(7927),
    [anon_sym__Alignas] = ACTIONS(7927),
    [sym_primitive_type] = ACTIONS(7927),
    [anon_sym_enum] = ACTIONS(7927),
    [anon_sym_COLON] = ACTIONS(7929),
    [anon_sym_struct] = ACTIONS(7927),
    [anon_sym_union] = ACTIONS(7927),
    [anon_sym_in] = ACTIONS(7927),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7927),
    [anon_sym___typeof] = ACTIONS(7927),
    [anon_sym_typeof] = ACTIONS(7927),
    [anon_sym_BOOL] = ACTIONS(7927),
    [anon_sym_IMP] = ACTIONS(7927),
    [anon_sym_SEL] = ACTIONS(7927),
    [anon_sym_Class] = ACTIONS(7927),
    [anon_sym_id] = ACTIONS(7927),
    [anon_sym_out] = ACTIONS(7927),
    [anon_sym_inout] = ACTIONS(7927),
    [anon_sym_bycopy] = ACTIONS(7927),
    [anon_sym_byref] = ACTIONS(7927),
    [anon_sym_oneway] = ACTIONS(7927),
  },
  [STATE(4096)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6358),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4097)] = {
    [sym_type_qualifier] = STATE(4103),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4167),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_type_descriptor] = STATE(6000),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(4103),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4098)] = {
    [sym_attribute_specifier] = STATE(3950),
    [sym_identifier] = ACTIONS(7826),
    [anon_sym_LPAREN2] = ACTIONS(7828),
    [anon_sym_STAR] = ACTIONS(7828),
    [anon_sym_CARET] = ACTIONS(7828),
    [anon_sym___extension__] = ACTIONS(7826),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7826),
    [anon_sym_LBRACK] = ACTIONS(7828),
    [anon_sym___based] = ACTIONS(7826),
    [anon_sym_signed] = ACTIONS(7826),
    [anon_sym_unsigned] = ACTIONS(7826),
    [anon_sym_long] = ACTIONS(7826),
    [anon_sym_short] = ACTIONS(7826),
    [anon_sym_const] = ACTIONS(7826),
    [anon_sym_constexpr] = ACTIONS(7826),
    [anon_sym_volatile] = ACTIONS(7826),
    [anon_sym_restrict] = ACTIONS(7826),
    [anon_sym___restrict__] = ACTIONS(7826),
    [anon_sym__Atomic] = ACTIONS(7826),
    [anon_sym__Noreturn] = ACTIONS(7826),
    [anon_sym__Nonnull] = ACTIONS(7826),
    [anon_sym_nullable] = ACTIONS(7826),
    [anon_sym__Complex] = ACTIONS(7826),
    [anon_sym__Nullable] = ACTIONS(7826),
    [anon_sym__Nullable_result] = ACTIONS(7826),
    [anon_sym__Null_unspecified] = ACTIONS(7826),
    [anon_sym___autoreleasing] = ACTIONS(7826),
    [anon_sym___block] = ACTIONS(7826),
    [anon_sym___bridge] = ACTIONS(7826),
    [anon_sym___bridge_retained] = ACTIONS(7826),
    [anon_sym___bridge_transfer] = ACTIONS(7826),
    [anon_sym___complex] = ACTIONS(7826),
    [anon_sym___const] = ACTIONS(7826),
    [anon_sym___imag] = ACTIONS(7826),
    [anon_sym___kindof] = ACTIONS(7826),
    [anon_sym___nonnull] = ACTIONS(7826),
    [anon_sym___nullable] = ACTIONS(7826),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7826),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7826),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7826),
    [anon_sym___real] = ACTIONS(7826),
    [anon_sym___strong] = ACTIONS(7826),
    [anon_sym___unsafe_unretained] = ACTIONS(7826),
    [anon_sym___unused] = ACTIONS(7826),
    [anon_sym___weak] = ACTIONS(7826),
    [anon_sym_alignas] = ACTIONS(7826),
    [anon_sym__Alignas] = ACTIONS(7826),
    [sym_primitive_type] = ACTIONS(7826),
    [anon_sym_enum] = ACTIONS(7826),
    [anon_sym_COLON] = ACTIONS(7828),
    [anon_sym_struct] = ACTIONS(7826),
    [anon_sym_union] = ACTIONS(7826),
    [anon_sym_in] = ACTIONS(7826),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7826),
    [anon_sym___typeof] = ACTIONS(7826),
    [anon_sym_typeof] = ACTIONS(7826),
    [anon_sym_BOOL] = ACTIONS(7826),
    [anon_sym_IMP] = ACTIONS(7826),
    [anon_sym_SEL] = ACTIONS(7826),
    [anon_sym_Class] = ACTIONS(7826),
    [anon_sym_id] = ACTIONS(7826),
    [anon_sym_out] = ACTIONS(7826),
    [anon_sym_inout] = ACTIONS(7826),
    [anon_sym_bycopy] = ACTIONS(7826),
    [anon_sym_byref] = ACTIONS(7826),
    [anon_sym_oneway] = ACTIONS(7826),
  },
  [STATE(4099)] = {
    [sym_attribute_specifier] = STATE(3824),
    [sym_identifier] = ACTIONS(7940),
    [anon_sym_LPAREN2] = ACTIONS(7942),
    [anon_sym_STAR] = ACTIONS(7942),
    [anon_sym_CARET] = ACTIONS(7942),
    [anon_sym___extension__] = ACTIONS(7940),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7940),
    [anon_sym_LBRACK] = ACTIONS(7942),
    [anon_sym___based] = ACTIONS(7940),
    [anon_sym_signed] = ACTIONS(7940),
    [anon_sym_unsigned] = ACTIONS(7940),
    [anon_sym_long] = ACTIONS(7940),
    [anon_sym_short] = ACTIONS(7940),
    [anon_sym_const] = ACTIONS(7940),
    [anon_sym_constexpr] = ACTIONS(7940),
    [anon_sym_volatile] = ACTIONS(7940),
    [anon_sym_restrict] = ACTIONS(7940),
    [anon_sym___restrict__] = ACTIONS(7940),
    [anon_sym__Atomic] = ACTIONS(7940),
    [anon_sym__Noreturn] = ACTIONS(7940),
    [anon_sym__Nonnull] = ACTIONS(7940),
    [anon_sym_nullable] = ACTIONS(7940),
    [anon_sym__Complex] = ACTIONS(7940),
    [anon_sym__Nullable] = ACTIONS(7940),
    [anon_sym__Nullable_result] = ACTIONS(7940),
    [anon_sym__Null_unspecified] = ACTIONS(7940),
    [anon_sym___autoreleasing] = ACTIONS(7940),
    [anon_sym___block] = ACTIONS(7940),
    [anon_sym___bridge] = ACTIONS(7940),
    [anon_sym___bridge_retained] = ACTIONS(7940),
    [anon_sym___bridge_transfer] = ACTIONS(7940),
    [anon_sym___complex] = ACTIONS(7940),
    [anon_sym___const] = ACTIONS(7940),
    [anon_sym___imag] = ACTIONS(7940),
    [anon_sym___kindof] = ACTIONS(7940),
    [anon_sym___nonnull] = ACTIONS(7940),
    [anon_sym___nullable] = ACTIONS(7940),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7940),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7940),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7940),
    [anon_sym___real] = ACTIONS(7940),
    [anon_sym___strong] = ACTIONS(7940),
    [anon_sym___unsafe_unretained] = ACTIONS(7940),
    [anon_sym___unused] = ACTIONS(7940),
    [anon_sym___weak] = ACTIONS(7940),
    [anon_sym_alignas] = ACTIONS(7940),
    [anon_sym__Alignas] = ACTIONS(7940),
    [sym_primitive_type] = ACTIONS(7940),
    [anon_sym_enum] = ACTIONS(7940),
    [anon_sym_COLON] = ACTIONS(7942),
    [anon_sym_struct] = ACTIONS(7940),
    [anon_sym_union] = ACTIONS(7940),
    [anon_sym_in] = ACTIONS(7940),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7940),
    [anon_sym___typeof] = ACTIONS(7940),
    [anon_sym_typeof] = ACTIONS(7940),
    [anon_sym_BOOL] = ACTIONS(7940),
    [anon_sym_IMP] = ACTIONS(7940),
    [anon_sym_SEL] = ACTIONS(7940),
    [anon_sym_Class] = ACTIONS(7940),
    [anon_sym_id] = ACTIONS(7940),
    [anon_sym_out] = ACTIONS(7940),
    [anon_sym_inout] = ACTIONS(7940),
    [anon_sym_bycopy] = ACTIONS(7940),
    [anon_sym_byref] = ACTIONS(7940),
    [anon_sym_oneway] = ACTIONS(7940),
  },
  [STATE(4100)] = {
    [sym_attribute_specifier] = STATE(3972),
    [sym_identifier] = ACTIONS(7843),
    [anon_sym_LPAREN2] = ACTIONS(7845),
    [anon_sym_STAR] = ACTIONS(7845),
    [anon_sym_CARET] = ACTIONS(7845),
    [anon_sym___extension__] = ACTIONS(7843),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7843),
    [anon_sym_LBRACK] = ACTIONS(7845),
    [anon_sym___based] = ACTIONS(7843),
    [anon_sym_signed] = ACTIONS(7843),
    [anon_sym_unsigned] = ACTIONS(7843),
    [anon_sym_long] = ACTIONS(7843),
    [anon_sym_short] = ACTIONS(7843),
    [anon_sym_const] = ACTIONS(7843),
    [anon_sym_constexpr] = ACTIONS(7843),
    [anon_sym_volatile] = ACTIONS(7843),
    [anon_sym_restrict] = ACTIONS(7843),
    [anon_sym___restrict__] = ACTIONS(7843),
    [anon_sym__Atomic] = ACTIONS(7843),
    [anon_sym__Noreturn] = ACTIONS(7843),
    [anon_sym__Nonnull] = ACTIONS(7843),
    [anon_sym_nullable] = ACTIONS(7843),
    [anon_sym__Complex] = ACTIONS(7843),
    [anon_sym__Nullable] = ACTIONS(7843),
    [anon_sym__Nullable_result] = ACTIONS(7843),
    [anon_sym__Null_unspecified] = ACTIONS(7843),
    [anon_sym___autoreleasing] = ACTIONS(7843),
    [anon_sym___block] = ACTIONS(7843),
    [anon_sym___bridge] = ACTIONS(7843),
    [anon_sym___bridge_retained] = ACTIONS(7843),
    [anon_sym___bridge_transfer] = ACTIONS(7843),
    [anon_sym___complex] = ACTIONS(7843),
    [anon_sym___const] = ACTIONS(7843),
    [anon_sym___imag] = ACTIONS(7843),
    [anon_sym___kindof] = ACTIONS(7843),
    [anon_sym___nonnull] = ACTIONS(7843),
    [anon_sym___nullable] = ACTIONS(7843),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7843),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7843),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7843),
    [anon_sym___real] = ACTIONS(7843),
    [anon_sym___strong] = ACTIONS(7843),
    [anon_sym___unsafe_unretained] = ACTIONS(7843),
    [anon_sym___unused] = ACTIONS(7843),
    [anon_sym___weak] = ACTIONS(7843),
    [anon_sym_alignas] = ACTIONS(7843),
    [anon_sym__Alignas] = ACTIONS(7843),
    [sym_primitive_type] = ACTIONS(7843),
    [anon_sym_enum] = ACTIONS(7843),
    [anon_sym_COLON] = ACTIONS(7845),
    [anon_sym_struct] = ACTIONS(7843),
    [anon_sym_union] = ACTIONS(7843),
    [anon_sym_in] = ACTIONS(7843),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7843),
    [anon_sym___typeof] = ACTIONS(7843),
    [anon_sym_typeof] = ACTIONS(7843),
    [anon_sym_BOOL] = ACTIONS(7843),
    [anon_sym_IMP] = ACTIONS(7843),
    [anon_sym_SEL] = ACTIONS(7843),
    [anon_sym_Class] = ACTIONS(7843),
    [anon_sym_id] = ACTIONS(7843),
    [anon_sym_out] = ACTIONS(7843),
    [anon_sym_inout] = ACTIONS(7843),
    [anon_sym_bycopy] = ACTIONS(7843),
    [anon_sym_byref] = ACTIONS(7843),
    [anon_sym_oneway] = ACTIONS(7843),
  },
  [STATE(4101)] = {
    [sym_attribute_specifier] = STATE(3876),
    [sym_identifier] = ACTIONS(7760),
    [anon_sym_LPAREN2] = ACTIONS(7762),
    [anon_sym_STAR] = ACTIONS(7762),
    [anon_sym_CARET] = ACTIONS(7762),
    [anon_sym___extension__] = ACTIONS(7760),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7760),
    [anon_sym_LBRACK] = ACTIONS(7762),
    [anon_sym___based] = ACTIONS(7760),
    [anon_sym_signed] = ACTIONS(7760),
    [anon_sym_unsigned] = ACTIONS(7760),
    [anon_sym_long] = ACTIONS(7760),
    [anon_sym_short] = ACTIONS(7760),
    [anon_sym_const] = ACTIONS(7760),
    [anon_sym_constexpr] = ACTIONS(7760),
    [anon_sym_volatile] = ACTIONS(7760),
    [anon_sym_restrict] = ACTIONS(7760),
    [anon_sym___restrict__] = ACTIONS(7760),
    [anon_sym__Atomic] = ACTIONS(7760),
    [anon_sym__Noreturn] = ACTIONS(7760),
    [anon_sym__Nonnull] = ACTIONS(7760),
    [anon_sym_nullable] = ACTIONS(7760),
    [anon_sym__Complex] = ACTIONS(7760),
    [anon_sym__Nullable] = ACTIONS(7760),
    [anon_sym__Nullable_result] = ACTIONS(7760),
    [anon_sym__Null_unspecified] = ACTIONS(7760),
    [anon_sym___autoreleasing] = ACTIONS(7760),
    [anon_sym___block] = ACTIONS(7760),
    [anon_sym___bridge] = ACTIONS(7760),
    [anon_sym___bridge_retained] = ACTIONS(7760),
    [anon_sym___bridge_transfer] = ACTIONS(7760),
    [anon_sym___complex] = ACTIONS(7760),
    [anon_sym___const] = ACTIONS(7760),
    [anon_sym___imag] = ACTIONS(7760),
    [anon_sym___kindof] = ACTIONS(7760),
    [anon_sym___nonnull] = ACTIONS(7760),
    [anon_sym___nullable] = ACTIONS(7760),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7760),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7760),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7760),
    [anon_sym___real] = ACTIONS(7760),
    [anon_sym___strong] = ACTIONS(7760),
    [anon_sym___unsafe_unretained] = ACTIONS(7760),
    [anon_sym___unused] = ACTIONS(7760),
    [anon_sym___weak] = ACTIONS(7760),
    [anon_sym_alignas] = ACTIONS(7760),
    [anon_sym__Alignas] = ACTIONS(7760),
    [sym_primitive_type] = ACTIONS(7760),
    [anon_sym_enum] = ACTIONS(7760),
    [anon_sym_COLON] = ACTIONS(7762),
    [anon_sym_struct] = ACTIONS(7760),
    [anon_sym_union] = ACTIONS(7760),
    [anon_sym_in] = ACTIONS(7760),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7760),
    [anon_sym___typeof] = ACTIONS(7760),
    [anon_sym_typeof] = ACTIONS(7760),
    [anon_sym_BOOL] = ACTIONS(7760),
    [anon_sym_IMP] = ACTIONS(7760),
    [anon_sym_SEL] = ACTIONS(7760),
    [anon_sym_Class] = ACTIONS(7760),
    [anon_sym_id] = ACTIONS(7760),
    [anon_sym_out] = ACTIONS(7760),
    [anon_sym_inout] = ACTIONS(7760),
    [anon_sym_bycopy] = ACTIONS(7760),
    [anon_sym_byref] = ACTIONS(7760),
    [anon_sym_oneway] = ACTIONS(7760),
  },
  [STATE(4102)] = {
    [sym_attribute_specifier] = STATE(3879),
    [sym_identifier] = ACTIONS(7775),
    [anon_sym_LPAREN2] = ACTIONS(7777),
    [anon_sym_STAR] = ACTIONS(7777),
    [anon_sym_CARET] = ACTIONS(7777),
    [anon_sym___extension__] = ACTIONS(7775),
    [anon_sym___attribute__] = ACTIONS(7961),
    [anon_sym___attribute] = ACTIONS(7961),
    [anon_sym_noreturn] = ACTIONS(7775),
    [anon_sym_LBRACK] = ACTIONS(7777),
    [anon_sym___based] = ACTIONS(7775),
    [anon_sym_signed] = ACTIONS(7775),
    [anon_sym_unsigned] = ACTIONS(7775),
    [anon_sym_long] = ACTIONS(7775),
    [anon_sym_short] = ACTIONS(7775),
    [anon_sym_const] = ACTIONS(7775),
    [anon_sym_constexpr] = ACTIONS(7775),
    [anon_sym_volatile] = ACTIONS(7775),
    [anon_sym_restrict] = ACTIONS(7775),
    [anon_sym___restrict__] = ACTIONS(7775),
    [anon_sym__Atomic] = ACTIONS(7775),
    [anon_sym__Noreturn] = ACTIONS(7775),
    [anon_sym__Nonnull] = ACTIONS(7775),
    [anon_sym_nullable] = ACTIONS(7775),
    [anon_sym__Complex] = ACTIONS(7775),
    [anon_sym__Nullable] = ACTIONS(7775),
    [anon_sym__Nullable_result] = ACTIONS(7775),
    [anon_sym__Null_unspecified] = ACTIONS(7775),
    [anon_sym___autoreleasing] = ACTIONS(7775),
    [anon_sym___block] = ACTIONS(7775),
    [anon_sym___bridge] = ACTIONS(7775),
    [anon_sym___bridge_retained] = ACTIONS(7775),
    [anon_sym___bridge_transfer] = ACTIONS(7775),
    [anon_sym___complex] = ACTIONS(7775),
    [anon_sym___const] = ACTIONS(7775),
    [anon_sym___imag] = ACTIONS(7775),
    [anon_sym___kindof] = ACTIONS(7775),
    [anon_sym___nonnull] = ACTIONS(7775),
    [anon_sym___nullable] = ACTIONS(7775),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(7775),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(7775),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(7775),
    [anon_sym___real] = ACTIONS(7775),
    [anon_sym___strong] = ACTIONS(7775),
    [anon_sym___unsafe_unretained] = ACTIONS(7775),
    [anon_sym___unused] = ACTIONS(7775),
    [anon_sym___weak] = ACTIONS(7775),
    [anon_sym_alignas] = ACTIONS(7775),
    [anon_sym__Alignas] = ACTIONS(7775),
    [sym_primitive_type] = ACTIONS(7775),
    [anon_sym_enum] = ACTIONS(7775),
    [anon_sym_COLON] = ACTIONS(7777),
    [anon_sym_struct] = ACTIONS(7775),
    [anon_sym_union] = ACTIONS(7775),
    [anon_sym_in] = ACTIONS(7775),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(7775),
    [anon_sym___typeof] = ACTIONS(7775),
    [anon_sym_typeof] = ACTIONS(7775),
    [anon_sym_BOOL] = ACTIONS(7775),
    [anon_sym_IMP] = ACTIONS(7775),
    [anon_sym_SEL] = ACTIONS(7775),
    [anon_sym_Class] = ACTIONS(7775),
    [anon_sym_id] = ACTIONS(7775),
    [anon_sym_out] = ACTIONS(7775),
    [anon_sym_inout] = ACTIONS(7775),
    [anon_sym_bycopy] = ACTIONS(7775),
    [anon_sym_byref] = ACTIONS(7775),
    [anon_sym_oneway] = ACTIONS(7775),
  },
  [STATE(4103)] = {
    [sym_type_qualifier] = STATE(2487),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4163),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(2487),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(4354),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4104)] = {
    [sym_type_qualifier] = STATE(2487),
    [sym_alignas_qualifier] = STATE(2354),
    [sym_type_specifier] = STATE(4163),
    [sym_sized_type_specifier] = STATE(2971),
    [sym_enum_specifier] = STATE(2971),
    [sym_struct_specifier] = STATE(2971),
    [sym_union_specifier] = STATE(2971),
    [sym_macro_type_specifier] = STATE(2971),
    [sym_typeof_specifier] = STATE(2971),
    [sym_generic_specifier] = STATE(2971),
    [sym_typedefed_specifier] = STATE(2971),
    [sym_array_type_specifier] = STATE(2971),
    [aux_sym__type_definition_type_repeat1] = STATE(2487),
    [aux_sym_sized_type_specifier_repeat1] = STATE(4165),
    [sym_identifier] = ACTIONS(2994),
    [anon_sym___extension__] = ACTIONS(39),
    [anon_sym_noreturn] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(4348),
    [anon_sym_unsigned] = ACTIONS(4348),
    [anon_sym_long] = ACTIONS(4348),
    [anon_sym_short] = ACTIONS(4348),
    [anon_sym_const] = ACTIONS(39),
    [anon_sym_constexpr] = ACTIONS(39),
    [anon_sym_volatile] = ACTIONS(39),
    [anon_sym_restrict] = ACTIONS(39),
    [anon_sym___restrict__] = ACTIONS(39),
    [anon_sym__Atomic] = ACTIONS(39),
    [anon_sym__Noreturn] = ACTIONS(39),
    [anon_sym__Nonnull] = ACTIONS(39),
    [anon_sym_nullable] = ACTIONS(39),
    [anon_sym__Complex] = ACTIONS(39),
    [anon_sym__Nullable] = ACTIONS(39),
    [anon_sym__Nullable_result] = ACTIONS(39),
    [anon_sym__Null_unspecified] = ACTIONS(39),
    [anon_sym___autoreleasing] = ACTIONS(39),
    [anon_sym___block] = ACTIONS(39),
    [anon_sym___bridge] = ACTIONS(39),
    [anon_sym___bridge_retained] = ACTIONS(39),
    [anon_sym___bridge_transfer] = ACTIONS(39),
    [anon_sym___complex] = ACTIONS(39),
    [anon_sym___const] = ACTIONS(39),
    [anon_sym___imag] = ACTIONS(39),
    [anon_sym___kindof] = ACTIONS(39),
    [anon_sym___nonnull] = ACTIONS(39),
    [anon_sym___nullable] = ACTIONS(39),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(39),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(39),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(39),
    [anon_sym___real] = ACTIONS(39),
    [anon_sym___strong] = ACTIONS(39),
    [anon_sym___unsafe_unretained] = ACTIONS(39),
    [anon_sym___unused] = ACTIONS(39),
    [anon_sym___weak] = ACTIONS(39),
    [anon_sym_alignas] = ACTIONS(59),
    [anon_sym__Alignas] = ACTIONS(59),
    [sym_primitive_type] = ACTIONS(61),
    [anon_sym_enum] = ACTIONS(8977),
    [anon_sym_struct] = ACTIONS(3020),
    [anon_sym_union] = ACTIONS(67),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(117),
    [anon_sym___typeof] = ACTIONS(117),
    [anon_sym_typeof] = ACTIONS(117),
    [anon_sym_BOOL] = ACTIONS(147),
    [anon_sym_IMP] = ACTIONS(147),
    [anon_sym_SEL] = ACTIONS(147),
    [anon_sym_Class] = ACTIONS(147),
    [anon_sym_id] = ACTIONS(147),
  },
  [STATE(4105)] = {
    [sym_identifier] = ACTIONS(2958),
    [aux_sym_preproc_if_token1] = ACTIONS(2958),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2960),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2960),
    [anon_sym___extension__] = ACTIONS(2958),
    [anon_sym_noreturn] = ACTIONS(2958),
    [anon_sym_signed] = ACTIONS(2958),
    [anon_sym_unsigned] = ACTIONS(2958),
    [anon_sym_long] = ACTIONS(2958),
    [anon_sym_short] = ACTIONS(2958),
    [anon_sym_const] = ACTIONS(2958),
    [anon_sym_constexpr] = ACTIONS(2958),
    [anon_sym_volatile] = ACTIONS(2958),
    [anon_sym_restrict] = ACTIONS(2958),
    [anon_sym___restrict__] = ACTIONS(2958),
    [anon_sym__Atomic] = ACTIONS(2958),
    [anon_sym__Noreturn] = ACTIONS(2958),
    [anon_sym__Nonnull] = ACTIONS(2958),
    [anon_sym_nullable] = ACTIONS(2958),
    [anon_sym__Complex] = ACTIONS(2958),
    [anon_sym__Nullable] = ACTIONS(2958),
    [anon_sym__Nullable_result] = ACTIONS(2958),
    [anon_sym__Null_unspecified] = ACTIONS(2958),
    [anon_sym___autoreleasing] = ACTIONS(2958),
    [anon_sym___block] = ACTIONS(2958),
    [anon_sym___bridge] = ACTIONS(2958),
    [anon_sym___bridge_retained] = ACTIONS(2958),
    [anon_sym___bridge_transfer] = ACTIONS(2958),
    [anon_sym___complex] = ACTIONS(2958),
    [anon_sym___const] = ACTIONS(2958),
    [anon_sym___imag] = ACTIONS(2958),
    [anon_sym___kindof] = ACTIONS(2958),
    [anon_sym___nonnull] = ACTIONS(2958),
    [anon_sym___nullable] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2958),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2958),
    [anon_sym___real] = ACTIONS(2958),
    [anon_sym___strong] = ACTIONS(2958),
    [anon_sym___unsafe_unretained] = ACTIONS(2958),
    [anon_sym___unused] = ACTIONS(2958),
    [anon_sym___weak] = ACTIONS(2958),
    [anon_sym_alignas] = ACTIONS(2958),
    [anon_sym__Alignas] = ACTIONS(2958),
    [sym_primitive_type] = ACTIONS(2958),
    [anon_sym_enum] = ACTIONS(2958),
    [anon_sym_struct] = ACTIONS(2958),
    [anon_sym_union] = ACTIONS(2958),
    [anon_sym_in] = ACTIONS(2958),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2958),
    [anon_sym___typeof] = ACTIONS(2958),
    [anon_sym_typeof] = ACTIONS(2958),
    [anon_sym_ATprivate] = ACTIONS(2960),
    [anon_sym_ATprotected] = ACTIONS(2960),
    [anon_sym_ATpackage] = ACTIONS(2960),
    [anon_sym_ATpublic] = ACTIONS(2960),
    [anon_sym_BOOL] = ACTIONS(2958),
    [anon_sym_IMP] = ACTIONS(2958),
    [anon_sym_SEL] = ACTIONS(2958),
    [anon_sym_Class] = ACTIONS(2958),
    [anon_sym_id] = ACTIONS(2958),
    [anon_sym_out] = ACTIONS(2958),
    [anon_sym_inout] = ACTIONS(2958),
    [anon_sym_bycopy] = ACTIONS(2958),
    [anon_sym_byref] = ACTIONS(2958),
    [anon_sym_oneway] = ACTIONS(2958),
  },
  [STATE(4106)] = {
    [sym_identifier] = ACTIONS(2962),
    [aux_sym_preproc_if_token1] = ACTIONS(2962),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2964),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2964),
    [anon_sym___extension__] = ACTIONS(2962),
    [anon_sym_noreturn] = ACTIONS(2962),
    [anon_sym_signed] = ACTIONS(2962),
    [anon_sym_unsigned] = ACTIONS(2962),
    [anon_sym_long] = ACTIONS(2962),
    [anon_sym_short] = ACTIONS(2962),
    [anon_sym_const] = ACTIONS(2962),
    [anon_sym_constexpr] = ACTIONS(2962),
    [anon_sym_volatile] = ACTIONS(2962),
    [anon_sym_restrict] = ACTIONS(2962),
    [anon_sym___restrict__] = ACTIONS(2962),
    [anon_sym__Atomic] = ACTIONS(2962),
    [anon_sym__Noreturn] = ACTIONS(2962),
    [anon_sym__Nonnull] = ACTIONS(2962),
    [anon_sym_nullable] = ACTIONS(2962),
    [anon_sym__Complex] = ACTIONS(2962),
    [anon_sym__Nullable] = ACTIONS(2962),
    [anon_sym__Nullable_result] = ACTIONS(2962),
    [anon_sym__Null_unspecified] = ACTIONS(2962),
    [anon_sym___autoreleasing] = ACTIONS(2962),
    [anon_sym___block] = ACTIONS(2962),
    [anon_sym___bridge] = ACTIONS(2962),
    [anon_sym___bridge_retained] = ACTIONS(2962),
    [anon_sym___bridge_transfer] = ACTIONS(2962),
    [anon_sym___complex] = ACTIONS(2962),
    [anon_sym___const] = ACTIONS(2962),
    [anon_sym___imag] = ACTIONS(2962),
    [anon_sym___kindof] = ACTIONS(2962),
    [anon_sym___nonnull] = ACTIONS(2962),
    [anon_sym___nullable] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2962),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2962),
    [anon_sym___real] = ACTIONS(2962),
    [anon_sym___strong] = ACTIONS(2962),
    [anon_sym___unsafe_unretained] = ACTIONS(2962),
    [anon_sym___unused] = ACTIONS(2962),
    [anon_sym___weak] = ACTIONS(2962),
    [anon_sym_alignas] = ACTIONS(2962),
    [anon_sym__Alignas] = ACTIONS(2962),
    [sym_primitive_type] = ACTIONS(2962),
    [anon_sym_enum] = ACTIONS(2962),
    [anon_sym_struct] = ACTIONS(2962),
    [anon_sym_union] = ACTIONS(2962),
    [anon_sym_in] = ACTIONS(2962),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2962),
    [anon_sym___typeof] = ACTIONS(2962),
    [anon_sym_typeof] = ACTIONS(2962),
    [anon_sym_ATprivate] = ACTIONS(2964),
    [anon_sym_ATprotected] = ACTIONS(2964),
    [anon_sym_ATpackage] = ACTIONS(2964),
    [anon_sym_ATpublic] = ACTIONS(2964),
    [anon_sym_BOOL] = ACTIONS(2962),
    [anon_sym_IMP] = ACTIONS(2962),
    [anon_sym_SEL] = ACTIONS(2962),
    [anon_sym_Class] = ACTIONS(2962),
    [anon_sym_id] = ACTIONS(2962),
    [anon_sym_out] = ACTIONS(2962),
    [anon_sym_inout] = ACTIONS(2962),
    [anon_sym_bycopy] = ACTIONS(2962),
    [anon_sym_byref] = ACTIONS(2962),
    [anon_sym_oneway] = ACTIONS(2962),
  },
  [STATE(4107)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_LPAREN2] = ACTIONS(8982),
    [anon_sym_STAR] = ACTIONS(2920),
    [anon_sym_CARET] = ACTIONS(2920),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym___based] = ACTIONS(2914),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_COLON] = ACTIONS(2920),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
    [anon_sym_out] = ACTIONS(2914),
    [anon_sym_inout] = ACTIONS(2914),
    [anon_sym_bycopy] = ACTIONS(2914),
    [anon_sym_byref] = ACTIONS(2914),
    [anon_sym_oneway] = ACTIONS(2914),
  },
  [STATE(4108)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_LPAREN2] = ACTIONS(8984),
    [anon_sym_STAR] = ACTIONS(2920),
    [anon_sym_CARET] = ACTIONS(2920),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym___based] = ACTIONS(2914),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_COLON] = ACTIONS(2920),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
    [anon_sym_out] = ACTIONS(2914),
    [anon_sym_inout] = ACTIONS(2914),
    [anon_sym_bycopy] = ACTIONS(2914),
    [anon_sym_byref] = ACTIONS(2914),
    [anon_sym_oneway] = ACTIONS(2914),
  },
  [STATE(4109)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_LPAREN2] = ACTIONS(8986),
    [anon_sym_STAR] = ACTIONS(2920),
    [anon_sym_CARET] = ACTIONS(2920),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym___based] = ACTIONS(2914),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_COLON] = ACTIONS(2920),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
    [anon_sym_out] = ACTIONS(2914),
    [anon_sym_inout] = ACTIONS(2914),
    [anon_sym_bycopy] = ACTIONS(2914),
    [anon_sym_byref] = ACTIONS(2914),
    [anon_sym_oneway] = ACTIONS(2914),
  },
  [STATE(4110)] = {
    [sym_identifier] = ACTIONS(2914),
    [anon_sym_LPAREN2] = ACTIONS(8988),
    [anon_sym_STAR] = ACTIONS(2920),
    [anon_sym_CARET] = ACTIONS(2920),
    [anon_sym___extension__] = ACTIONS(2914),
    [anon_sym_noreturn] = ACTIONS(2914),
    [anon_sym___based] = ACTIONS(2914),
    [anon_sym_signed] = ACTIONS(2914),
    [anon_sym_unsigned] = ACTIONS(2914),
    [anon_sym_long] = ACTIONS(2914),
    [anon_sym_short] = ACTIONS(2914),
    [anon_sym_const] = ACTIONS(2914),
    [anon_sym_constexpr] = ACTIONS(2914),
    [anon_sym_volatile] = ACTIONS(2914),
    [anon_sym_restrict] = ACTIONS(2914),
    [anon_sym___restrict__] = ACTIONS(2914),
    [anon_sym__Atomic] = ACTIONS(2914),
    [anon_sym__Noreturn] = ACTIONS(2914),
    [anon_sym__Nonnull] = ACTIONS(2914),
    [anon_sym_nullable] = ACTIONS(2914),
    [anon_sym__Complex] = ACTIONS(2914),
    [anon_sym__Nullable] = ACTIONS(2914),
    [anon_sym__Nullable_result] = ACTIONS(2914),
    [anon_sym__Null_unspecified] = ACTIONS(2914),
    [anon_sym___autoreleasing] = ACTIONS(2914),
    [anon_sym___block] = ACTIONS(2914),
    [anon_sym___bridge] = ACTIONS(2914),
    [anon_sym___bridge_retained] = ACTIONS(2914),
    [anon_sym___bridge_transfer] = ACTIONS(2914),
    [anon_sym___complex] = ACTIONS(2914),
    [anon_sym___const] = ACTIONS(2914),
    [anon_sym___imag] = ACTIONS(2914),
    [anon_sym___kindof] = ACTIONS(2914),
    [anon_sym___nonnull] = ACTIONS(2914),
    [anon_sym___nullable] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_class_ro] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_isa_pointer] = ACTIONS(2914),
    [anon_sym___ptrauth_objc_super_pointer] = ACTIONS(2914),
    [anon_sym___real] = ACTIONS(2914),
    [anon_sym___strong] = ACTIONS(2914),
    [anon_sym___unsafe_unretained] = ACTIONS(2914),
    [anon_sym___unused] = ACTIONS(2914),
    [anon_sym___weak] = ACTIONS(2914),
    [anon_sym_alignas] = ACTIONS(2914),
    [anon_sym__Alignas] = ACTIONS(2914),
    [sym_primitive_type] = ACTIONS(2914),
    [anon_sym_enum] = ACTIONS(2914),
    [anon_sym_COLON] = ACTIONS(2920),
    [anon_sym_struct] = ACTIONS(2914),
    [anon_sym_union] = ACTIONS(2914),
    [anon_sym_in] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___typeof__] = ACTIONS(2914),
    [anon_sym___typeof] = ACTIONS(2914),
    [anon_sym_typeof] = ACTIONS(2914),
    [anon_sym_BOOL] = ACTIONS(2914),
    [anon_sym_IMP] = ACTIONS(2914),
    [anon_sym_SEL] = ACTIONS(2914),
    [anon_sym_Class] = ACTIONS(2914),
    [anon_sym_id] = ACTIONS(2914),
    [anon_sym_out] = ACTIONS(2914),
    [anon_sym_inout] = ACTIONS(2914),
    [anon_sym_bycopy] = ACTIONS(2914),
    [anon_sym_byref] = ACTIONS(2914),
    [anon_sym_oneway] = ACTIONS(2914),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_LPAREN2,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5312), 1,
      sym_identifier,
    ACTIONS(5427), 1,
      anon_sym_STAR,
    ACTIONS(5429), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3863), 1,
      sym__declarator,
    STATE(4290), 1,
      sym_parameter_list,
    STATE(4989), 1,
      sym__abstract_declarator,
    STATE(6497), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(8990), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(8992), 2,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
    STATE(4112), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [103] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_LPAREN2,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5312), 1,
      sym_identifier,
    ACTIONS(5427), 1,
      anon_sym_STAR,
    ACTIONS(5429), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3902), 1,
      sym__declarator,
    STATE(4290), 1,
      sym_parameter_list,
    STATE(4991), 1,
      sym__abstract_declarator,
    STATE(6497), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(8994), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(8996), 2,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [206] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9000), 1,
      anon_sym_LBRACK,
    ACTIONS(8998), 61,
      anon_sym___extension__,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_noreturn,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      anon_sym_in,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym_BOOL,
      anon_sym_IMP,
      anon_sym_SEL,
      anon_sym_Class,
      anon_sym_id,
      anon_sym_out,
      anon_sym_inout,
      anon_sym_bycopy,
      anon_sym_byref,
      anon_sym_oneway,
  [276] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_LPAREN2,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5312), 1,
      sym_identifier,
    ACTIONS(5316), 1,
      anon_sym_STAR,
    ACTIONS(5318), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3902), 1,
      sym__declarator,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(5000), 1,
      sym__abstract_declarator,
    STATE(6497), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(8996), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [376] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9004), 1,
      anon_sym_LBRACK,
    ACTIONS(9002), 61,
      anon_sym___extension__,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_noreturn,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      anon_sym_in,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym_BOOL,
      anon_sym_IMP,
      anon_sym_SEL,
      anon_sym_Class,
      anon_sym_id,
      anon_sym_out,
      anon_sym_inout,
      anon_sym_bycopy,
      anon_sym_byref,
      anon_sym_oneway,
  [446] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_LPAREN2,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5312), 1,
      sym_identifier,
    ACTIONS(5316), 1,
      anon_sym_STAR,
    ACTIONS(5318), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3863), 1,
      sym__declarator,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(4979), 1,
      sym__abstract_declarator,
    STATE(6497), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4114), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(8992), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [546] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9008), 1,
      anon_sym_LBRACK,
    ACTIONS(9006), 61,
      anon_sym___extension__,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_noreturn,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      anon_sym_in,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym_BOOL,
      anon_sym_IMP,
      anon_sym_SEL,
      anon_sym_Class,
      anon_sym_id,
      anon_sym_out,
      anon_sym_inout,
      anon_sym_bycopy,
      anon_sym_byref,
      anon_sym_oneway,
  [616] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(5417), 1,
      anon_sym_LPAREN2,
    ACTIONS(5419), 1,
      anon_sym_STAR,
    ACTIONS(5421), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3054), 1,
      sym__declarator,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(5000), 1,
      sym__abstract_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(8996), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [715] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(5417), 1,
      anon_sym_LPAREN2,
    ACTIONS(5419), 1,
      anon_sym_STAR,
    ACTIONS(5421), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3050), 1,
      sym__declarator,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(4979), 1,
      sym__abstract_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(8992), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(4118), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [814] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_LPAREN2,
    ACTIONS(4086), 1,
      anon_sym_STAR,
    ACTIONS(4088), 1,
      anon_sym_CARET,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5312), 1,
      sym_identifier,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    ACTIONS(8996), 1,
      anon_sym_RPAREN,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(5000), 1,
      sym__abstract_declarator,
    STATE(5013), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [912] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_LPAREN2,
    ACTIONS(4086), 1,
      anon_sym_STAR,
    ACTIONS(4088), 1,
      anon_sym_CARET,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5312), 1,
      sym_identifier,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    ACTIONS(8992), 1,
      anon_sym_RPAREN,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(4979), 1,
      sym__abstract_declarator,
    STATE(5032), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4120), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [1010] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9012), 1,
      anon_sym_RPAREN,
    STATE(5453), 1,
      aux_sym_attribute_specifier_repeat1,
    ACTIONS(2920), 23,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2914), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [1081] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9014), 1,
      anon_sym_RPAREN,
    STATE(5510), 1,
      aux_sym_attribute_specifier_repeat1,
    ACTIONS(2920), 23,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2914), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [1152] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9016), 1,
      anon_sym_RPAREN,
    STATE(5473), 1,
      aux_sym_attribute_specifier_repeat1,
    ACTIONS(2920), 23,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2914), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [1223] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9018), 1,
      anon_sym_RPAREN,
    STATE(5502), 1,
      aux_sym_attribute_specifier_repeat1,
    ACTIONS(2920), 23,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2914), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [1294] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9020), 1,
      anon_sym_RPAREN,
    STATE(5416), 1,
      aux_sym_attribute_specifier_repeat1,
    ACTIONS(2920), 23,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2914), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [1365] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9022), 1,
      anon_sym_RPAREN,
    STATE(5574), 1,
      aux_sym_attribute_specifier_repeat1,
    ACTIONS(2920), 23,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2914), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [1436] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9024), 1,
      anon_sym_RPAREN,
    STATE(5491), 1,
      aux_sym_attribute_specifier_repeat1,
    ACTIONS(2920), 23,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2914), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [1507] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3244), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4131), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [1595] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5696), 1,
      sym_primitive_type,
    ACTIONS(5698), 1,
      anon_sym_STAR,
    ACTIONS(5700), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(5205), 1,
      sym__type_declarator,
    STATE(6343), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4137), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [1683] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3263), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [1771] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(2960), 1,
      sym_attribute_specifier,
    STATE(4153), 1,
      sym_enumerator_list,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7514), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7512), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [1843] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(9026), 1,
      anon_sym_COLON,
    STATE(2940), 1,
      sym_attribute_specifier,
    STATE(4162), 1,
      sym_enumerator_list,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7621), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(7619), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [1917] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(2963), 1,
      sym_attribute_specifier,
    STATE(4146), 1,
      sym_enumerator_list,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7523), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7521), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [1989] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(2958), 1,
      sym_attribute_specifier,
    STATE(4145), 1,
      sym_enumerator_list,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7537), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7535), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2061] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(2965), 1,
      sym_attribute_specifier,
    STATE(4147), 1,
      sym_enumerator_list,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7550), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7548), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2133] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5696), 1,
      sym_primitive_type,
    ACTIONS(5698), 1,
      anon_sym_STAR,
    ACTIONS(5700), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(5203), 1,
      sym__type_declarator,
    STATE(6343), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [2221] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(9028), 1,
      anon_sym_COLON,
    STATE(2983), 1,
      sym_attribute_specifier,
    STATE(4156), 1,
      sym_enumerator_list,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7583), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(7581), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2295] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(9030), 1,
      anon_sym_COLON,
    STATE(3004), 1,
      sym_attribute_specifier,
    STATE(4157), 1,
      sym_enumerator_list,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7634), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(7632), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2369] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(2977), 1,
      sym_attribute_specifier,
    STATE(4152), 1,
      sym_enumerator_list,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7563), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7561), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2441] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(2980), 1,
      sym_attribute_specifier,
    STATE(4154), 1,
      sym_enumerator_list,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7570), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7568), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2513] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4488), 1,
      anon_sym_LPAREN2,
    ACTIONS(4490), 1,
      anon_sym_STAR,
    ACTIONS(4492), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(4979), 1,
      sym__abstract_declarator,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4159), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(8992), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
      anon_sym_COLON,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [2597] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2976), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7929), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7927), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2663] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2956), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7922), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7920), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2729] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2953), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7877), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7875), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2795] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2997), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7845), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7843), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2861] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2955), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7907), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7905), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2927] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2943), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7808), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7806), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [2993] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2920), 23,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2914), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [3055] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5668), 23,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(5666), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [3117] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2984), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7791), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7789), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [3183] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2941), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7762), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7760), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [3249] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2996), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7828), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7826), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [3315] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2944), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7777), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7775), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [3381] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2950), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7784), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7782), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [3447] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2968), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7819), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7817), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [3513] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2981), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7942), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7940), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [3579] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4488), 1,
      anon_sym_LPAREN2,
    ACTIONS(6850), 1,
      anon_sym_STAR,
    ACTIONS(6852), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    ACTIONS(8994), 1,
      anon_sym___attribute,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4290), 1,
      sym_parameter_list,
    STATE(4991), 1,
      sym__abstract_declarator,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(8996), 3,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [3665] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4488), 1,
      anon_sym_LPAREN2,
    ACTIONS(4490), 1,
      anon_sym_STAR,
    ACTIONS(4492), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(5000), 1,
      sym__abstract_declarator,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(8996), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
      anon_sym_COLON,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [3749] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4488), 1,
      anon_sym_LPAREN2,
    ACTIONS(6850), 1,
      anon_sym_STAR,
    ACTIONS(6852), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    ACTIONS(8990), 1,
      anon_sym___attribute,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4290), 1,
      sym_parameter_list,
    STATE(4989), 1,
      sym__abstract_declarator,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4158), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(8992), 3,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [3835] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2931), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7900), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7898), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [3901] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(3003), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(7856), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7854), 44,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [3967] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4488), 1,
      anon_sym_LPAREN2,
    ACTIONS(4490), 1,
      anon_sym_STAR,
    ACTIONS(4492), 1,
      anon_sym_CARET,
    ACTIONS(9034), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(5048), 1,
      sym__abstract_declarator,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4166), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(9032), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [4050] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6891), 1,
      sym_primitive_type,
    ACTIONS(9036), 1,
      sym_identifier,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2995), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(6888), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(6886), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(6879), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(6881), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [4125] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6920), 1,
      sym_primitive_type,
    ACTIONS(9038), 1,
      sym_identifier,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4237), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(6917), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4164), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9040), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(6908), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(6910), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [4200] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4488), 1,
      anon_sym_LPAREN2,
    ACTIONS(4490), 1,
      anon_sym_STAR,
    ACTIONS(4492), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(5073), 1,
      sym__abstract_declarator,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(9042), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [4283] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4488), 1,
      anon_sym_LPAREN2,
    ACTIONS(4490), 1,
      anon_sym_STAR,
    ACTIONS(4492), 1,
      anon_sym_CARET,
    ACTIONS(9034), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(5069), 1,
      sym__abstract_declarator,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4168), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(9044), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [4366] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4488), 1,
      anon_sym_LPAREN2,
    ACTIONS(4490), 1,
      anon_sym_STAR,
    ACTIONS(4492), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(5043), 1,
      sym__abstract_declarator,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(9046), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [4449] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2548), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2546), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [4509] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2440), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2438), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [4569] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2444), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2442), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [4629] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(6363), 1,
      sym_identifier,
    ACTIONS(6365), 1,
      anon_sym_LPAREN2,
    ACTIONS(6401), 1,
      anon_sym_STAR,
    ACTIONS(6403), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(5065), 1,
      sym__field_declarator,
    STATE(6257), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4182), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(4889), 6,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
      sym_block_pointer_field_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [4709] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2426), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2424), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [4769] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2418), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2416), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [4829] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2476), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2474), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [4889] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2444), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2442), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [4949] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2426), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2424), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [5009] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2422), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2420), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [5069] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(9052), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4259), 1,
      sym_ms_declspec_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4249), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9050), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9048), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [5143] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2017), 1,
      sym__old_style_function_declarator,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2797), 1,
      sym__declarator,
    STATE(3177), 1,
      sym_init_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(6720), 1,
      sym_type_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [5227] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(9052), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4258), 1,
      sym_ms_declspec_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4246), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9056), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9054), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [5301] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(6363), 1,
      sym_identifier,
    ACTIONS(6365), 1,
      anon_sym_LPAREN2,
    ACTIONS(6401), 1,
      anon_sym_STAR,
    ACTIONS(6403), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(5089), 1,
      sym__field_declarator,
    STATE(6257), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(4889), 6,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
      sym_block_pointer_field_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [5381] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2440), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2438), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [5441] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(1994), 1,
      sym__old_style_function_declarator,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2799), 1,
      sym__declarator,
    STATE(3146), 1,
      sym_init_declarator,
    STATE(5946), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [5525] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9058), 1,
      sym_identifier,
    ACTIONS(9064), 1,
      sym_primitive_type,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4245), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(6913), 2,
      anon_sym___declspec,
      anon_sym___based,
    ACTIONS(6917), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4196), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(6908), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9061), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(6910), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [5601] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2048), 1,
      sym__old_style_function_declarator,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2792), 1,
      sym__declarator,
    STATE(3156), 1,
      sym_init_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(6296), 1,
      sym_type_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [5685] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2011), 1,
      sym__old_style_function_declarator,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2794), 1,
      sym__declarator,
    STATE(3075), 1,
      sym_init_declarator,
    STATE(6098), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [5769] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2448), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2446), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [5829] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(9052), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4242), 1,
      sym_ms_declspec_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4247), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9070), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9067), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [5903] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2476), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2474), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [5963] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2015), 1,
      sym__old_style_function_declarator,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2801), 1,
      sym__declarator,
    STATE(3084), 1,
      sym_init_declarator,
    STATE(6124), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [6047] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2548), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2546), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [6107] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(6363), 1,
      sym_identifier,
    ACTIONS(6365), 1,
      anon_sym_LPAREN2,
    ACTIONS(6367), 1,
      anon_sym_STAR,
    ACTIONS(6369), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4854), 1,
      sym__field_declarator,
    STATE(6222), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4199), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(4889), 6,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
      sym_block_pointer_field_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [6187] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2418), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2416), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [6247] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(9052), 1,
      anon_sym_LBRACK,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4243), 1,
      sym_ms_declspec_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4250), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9075), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9073), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [6321] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9077), 1,
      sym_identifier,
    ACTIONS(9083), 1,
      sym_primitive_type,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4244), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(6884), 2,
      anon_sym___declspec,
      anon_sym___based,
    ACTIONS(6888), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(6879), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9080), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(6881), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [6397] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2448), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2446), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [6457] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2422), 24,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_ATselector,
      anon_sym_ATavailable,
      anon_sym_ATencode,
    ACTIONS(2420), 28,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym___imag,
      anon_sym___real,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_AT,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
      anon_sym___builtin_available,
      anon_sym_va_arg,
      anon_sym_id,
  [6517] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(6363), 1,
      sym_identifier,
    ACTIONS(6365), 1,
      anon_sym_LPAREN2,
    ACTIONS(6367), 1,
      anon_sym_STAR,
    ACTIONS(6369), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4845), 1,
      sym__field_declarator,
    STATE(6222), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(4889), 6,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
      sym_block_pointer_field_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [6597] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7042), 1,
      anon_sym_LPAREN2,
    ACTIONS(7045), 1,
      anon_sym_LT,
    STATE(2908), 1,
      aux_sym_generic_specifier_repeat1,
    STATE(4256), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(6134), 3,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9086), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(4494), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [6666] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2816), 1,
      sym__declarator,
    STATE(3084), 1,
      sym_init_declarator,
    STATE(6124), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [6747] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2890), 1,
      sym__declarator,
    STATE(3075), 1,
      sym_init_declarator,
    STATE(6098), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [6828] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6487), 1,
      anon_sym_STAR,
    ACTIONS(6489), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2860), 1,
      sym__declarator,
    STATE(3153), 1,
      sym_init_declarator,
    STATE(6058), 1,
      sym_type_qualifier,
    STATE(6656), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [6909] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3050), 1,
      sym__declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4214), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [6988] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2888), 1,
      sym__declarator,
    STATE(3145), 1,
      sym_init_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(6637), 1,
      sym_type_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7069] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2819), 1,
      sym__declarator,
    STATE(3177), 1,
      sym_init_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(6720), 1,
      sym_type_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7150] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6487), 1,
      anon_sym_STAR,
    ACTIONS(6489), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3216), 1,
      sym__declarator,
    STATE(6656), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7229] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6487), 1,
      anon_sym_STAR,
    ACTIONS(6489), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2872), 1,
      sym__declarator,
    STATE(3153), 1,
      sym_init_declarator,
    STATE(6058), 1,
      sym_type_qualifier,
    STATE(6656), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7310] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(6463), 1,
      sym_identifier,
    ACTIONS(6465), 1,
      anon_sym_LPAREN2,
    ACTIONS(6467), 1,
      anon_sym_STAR,
    ACTIONS(6469), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2754), 1,
      sym__declarator,
    STATE(6101), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(2778), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7389] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2825), 1,
      sym__declarator,
    STATE(3146), 1,
      sym_init_declarator,
    STATE(5946), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7470] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2893), 1,
      sym__declarator,
    STATE(3091), 1,
      sym_init_declarator,
    STATE(6141), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7551] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2881), 1,
      sym__declarator,
    STATE(3150), 1,
      sym_init_declarator,
    STATE(6010), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7632] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2882), 1,
      sym__declarator,
    STATE(3156), 1,
      sym_init_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(6296), 1,
      sym_type_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7713] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3054), 1,
      sym__declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7792] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3206), 1,
      sym__declarator,
    STATE(3289), 1,
      sym_init_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(6284), 1,
      sym_type_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7873] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6487), 1,
      anon_sym_STAR,
    ACTIONS(6489), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2867), 1,
      sym__declarator,
    STATE(3153), 1,
      sym_init_declarator,
    STATE(6058), 1,
      sym_type_qualifier,
    STATE(6656), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [7954] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6487), 1,
      anon_sym_STAR,
    ACTIONS(6489), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3210), 1,
      sym__declarator,
    STATE(6656), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4207), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8033] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2832), 1,
      sym__declarator,
    STATE(3145), 1,
      sym_init_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(6637), 1,
      sym_type_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8114] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(6463), 1,
      sym_identifier,
    ACTIONS(6465), 1,
      anon_sym_LPAREN2,
    ACTIONS(6467), 1,
      anon_sym_STAR,
    ACTIONS(6469), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2750), 1,
      sym__declarator,
    STATE(6101), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4209), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(2778), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8193] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4517), 1,
      anon_sym_AT,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 14,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute,
      anon_sym_DOT,
    ACTIONS(4026), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [8262] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2817), 1,
      sym__declarator,
    STATE(3106), 1,
      sym_init_declarator,
    STATE(6164), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8343] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(5032), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4225), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8422] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2812), 1,
      sym__declarator,
    STATE(3075), 1,
      sym_init_declarator,
    STATE(6098), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8503] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(5312), 1,
      sym_identifier,
    ACTIONS(6459), 1,
      anon_sym_STAR,
    ACTIONS(6461), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3902), 1,
      sym__declarator,
    STATE(6497), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8582] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(5013), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8661] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(5312), 1,
      sym_identifier,
    ACTIONS(6459), 1,
      anon_sym_STAR,
    ACTIONS(6461), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(3863), 1,
      sym__declarator,
    STATE(6497), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4224), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8740] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6487), 1,
      anon_sym_STAR,
    ACTIONS(6489), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2855), 1,
      sym__declarator,
    STATE(3153), 1,
      sym_init_declarator,
    STATE(6058), 1,
      sym_type_qualifier,
    STATE(6656), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8821] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2875), 1,
      sym__declarator,
    STATE(3084), 1,
      sym_init_declarator,
    STATE(6124), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8902] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2838), 1,
      sym__declarator,
    STATE(3098), 1,
      sym_init_declarator,
    STATE(6153), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [8983] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6487), 1,
      anon_sym_STAR,
    ACTIONS(6489), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2865), 1,
      sym__declarator,
    STATE(3153), 1,
      sym_init_declarator,
    STATE(6058), 1,
      sym_type_qualifier,
    STATE(6656), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [9064] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2899), 1,
      sym__declarator,
    STATE(3146), 1,
      sym_init_declarator,
    STATE(5946), 1,
      sym_type_qualifier,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [9145] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6487), 1,
      anon_sym_STAR,
    ACTIONS(6489), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2868), 1,
      sym__declarator,
    STATE(3153), 1,
      sym_init_declarator,
    STATE(6058), 1,
      sym_type_qualifier,
    STATE(6656), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [9226] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2889), 1,
      sym__declarator,
    STATE(3177), 1,
      sym_init_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(6720), 1,
      sym_type_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [9307] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6487), 1,
      anon_sym_STAR,
    ACTIONS(6489), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2869), 1,
      sym__declarator,
    STATE(3153), 1,
      sym_init_declarator,
    STATE(6058), 1,
      sym_type_qualifier,
    STATE(6656), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [9388] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6487), 1,
      anon_sym_STAR,
    ACTIONS(6489), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2871), 1,
      sym__declarator,
    STATE(3153), 1,
      sym_init_declarator,
    STATE(6058), 1,
      sym_type_qualifier,
    STATE(6656), 1,
      sym_ms_based_modifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [9469] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2835), 1,
      sym__declarator,
    STATE(3156), 1,
      sym_init_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(6296), 1,
      sym_type_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [9550] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2912), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7080), 2,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(7084), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7098), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7095), 36,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [9614] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(4261), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7413), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9089), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7411), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [9675] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2912), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7437), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9092), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7435), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [9736] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9098), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9095), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [9801] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(4257), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7407), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9101), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7405), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [9862] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4240), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9107), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9104), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [9927] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4254), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9112), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9110), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [9992] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2912), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7423), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9114), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7421), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [10053] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2912), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7098), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9117), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7095), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [10114] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9123), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9121), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [10179] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9107), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9104), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [10244] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(4263), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7317), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9125), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7315), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [10305] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9130), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9128), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [10370] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9112), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9110), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [10435] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(4262), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7185), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9132), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7183), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [10496] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9137), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9135), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [10561] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9141), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9139), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [10626] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9145), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9143), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [10691] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4503), 1,
      anon_sym_LT,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(9149), 1,
      anon_sym_LPAREN2,
    ACTIONS(9152), 1,
      anon_sym_COLON_COLON,
    ACTIONS(9154), 1,
      anon_sym_EQ,
    STATE(1916), 1,
      sym_string_literal,
    STATE(2908), 1,
      aux_sym_generic_specifier_repeat1,
    ACTIONS(9147), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(9156), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4026), 12,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      sym_identifier,
  [10768] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2912), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7285), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9158), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7283), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [10829] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2912), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7305), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9161), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7303), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [10890] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4253), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9123), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9121), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [10955] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(4252), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9130), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
    ACTIONS(9128), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [11020] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(4239), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(5070), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9164), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(5074), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [11081] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2912), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7309), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9167), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7307), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [11142] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2912), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7367), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9170), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7365), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [11203] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2912), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7379), 4,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym_LBRACK,
    ACTIONS(9173), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7377), 40,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [11264] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6891), 1,
      sym_primitive_type,
    ACTIONS(9036), 1,
      sym_identifier,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(2995), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(6879), 2,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
    STATE(2487), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(6886), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [11334] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(9154), 1,
      anon_sym_EQ,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(9156), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 14,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(4026), 16,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [11400] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    STATE(3047), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(9176), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [11468] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4503), 1,
      anon_sym_LT,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(9154), 1,
      anon_sym_EQ,
    STATE(1916), 1,
      sym_string_literal,
    STATE(2908), 1,
      aux_sym_generic_specifier_repeat1,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(9156), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [11538] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    STATE(3047), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(9178), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [11606] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6920), 1,
      sym_primitive_type,
    ACTIONS(9038), 1,
      sym_identifier,
    STATE(2354), 1,
      sym_alignas_qualifier,
    STATE(4296), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(59), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(6908), 2,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
    STATE(4264), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(9180), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(39), 34,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [11676] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    STATE(4266), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(9182), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [11744] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    STATE(4268), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(9184), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [11812] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(4564), 1,
      anon_sym_COLON,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [11879] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    STATE(4274), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(9186), 5,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [11946] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9190), 1,
      anon_sym___attribute,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    STATE(3047), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(9188), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [12011] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(4531), 1,
      anon_sym_COLON,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [12078] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(8383), 1,
      anon_sym_EQ,
    ACTIONS(9192), 1,
      anon_sym_LT2,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(8385), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [12145] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(4535), 1,
      anon_sym_COLON,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [12212] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(4560), 1,
      anon_sym_COLON,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [12279] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(4529), 1,
      anon_sym_COLON,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [12346] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4513), 1,
      anon_sym_COLON,
    ACTIONS(4517), 1,
      anon_sym_AT,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [12413] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(4566), 1,
      anon_sym_COLON,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [12480] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(9194), 1,
      anon_sym_COLON,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACE,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [12547] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(4519), 1,
      anon_sym_COLON,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [12614] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(9196), 1,
      anon_sym_COLON,
    STATE(2983), 1,
      sym_attribute_specifier,
    STATE(4156), 1,
      sym_enumerator_list,
    ACTIONS(7581), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(7583), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [12680] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(9199), 1,
      anon_sym_COLON,
    STATE(3004), 1,
      sym_attribute_specifier,
    STATE(4157), 1,
      sym_enumerator_list,
    ACTIONS(7632), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(7634), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [12746] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(9202), 1,
      anon_sym_COLON,
    STATE(2940), 1,
      sym_attribute_specifier,
    STATE(4162), 1,
      sym_enumerator_list,
    ACTIONS(7619), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(7621), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [12812] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9205), 1,
      anon_sym___attribute,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    STATE(3047), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(9176), 5,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [12876] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(8383), 1,
      anon_sym_EQ,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(8385), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [12940] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    STATE(4287), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(9182), 4,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [13006] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    STATE(4291), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(9184), 4,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [13072] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9207), 1,
      anon_sym___attribute,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    STATE(3047), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(9178), 5,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [13136] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9213), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 4,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(4002), 20,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [13215] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9213), 1,
      anon_sym_AMP,
    ACTIONS(9223), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9225), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9227), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9229), 1,
      anon_sym_PIPE,
    ACTIONS(9231), 1,
      anon_sym_CARET,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(4474), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4476), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [13304] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(9186), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
    STATE(4306), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [13369] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 9,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(4002), 24,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [13438] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2912), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(7098), 2,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
    ACTIONS(7084), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(7080), 38,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [13495] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9213), 1,
      anon_sym_AMP,
    ACTIONS(9223), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9225), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9227), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9229), 1,
      anon_sym_PIPE,
    ACTIONS(9231), 1,
      anon_sym_CARET,
    ACTIONS(9233), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(4478), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4480), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [13586] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 5,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(4002), 22,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [13661] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9213), 1,
      anon_sym_AMP,
    ACTIONS(9223), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9225), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9227), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9229), 1,
      anon_sym_PIPE,
    ACTIONS(9231), 1,
      anon_sym_CARET,
    ACTIONS(9233), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(4466), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4468), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [13752] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9213), 1,
      anon_sym_AMP,
    ACTIONS(9225), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9227), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9229), 1,
      anon_sym_PIPE,
    ACTIONS(9231), 1,
      anon_sym_CARET,
    ACTIONS(9235), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(4431), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4436), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [13841] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9213), 1,
      anon_sym_AMP,
    ACTIONS(9227), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9229), 1,
      anon_sym_PIPE,
    ACTIONS(9231), 1,
      anon_sym_CARET,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(4000), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [13926] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9213), 1,
      anon_sym_AMP,
    ACTIONS(9223), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9225), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9227), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9229), 1,
      anon_sym_PIPE,
    ACTIONS(9231), 1,
      anon_sym_CARET,
    ACTIONS(9233), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(4458), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4462), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [14017] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 7,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(4002), 24,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [14088] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 5,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(4002), 20,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [14165] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9213), 1,
      anon_sym_AMP,
    ACTIONS(9223), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9225), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9227), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9229), 1,
      anon_sym_PIPE,
    ACTIONS(9231), 1,
      anon_sym_CARET,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(4470), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4472), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [14254] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(9188), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
    STATE(3047), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [14319] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 11,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(4002), 24,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [14386] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(8019), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
    STATE(3047), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [14451] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9213), 1,
      anon_sym_AMP,
    ACTIONS(9229), 1,
      anon_sym_PIPE,
    ACTIONS(9231), 1,
      anon_sym_CARET,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(4000), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 20,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [14534] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9213), 1,
      anon_sym_AMP,
    ACTIONS(9231), 1,
      anon_sym_CARET,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9209), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9215), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9217), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9219), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4000), 3,
      anon_sym_PIPE,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(9211), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 20,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [14615] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(2354), 1,
      sym_alignas_qualifier,
    ACTIONS(6774), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(39), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(8015), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
    STATE(4308), 3,
      sym_attribute_specifier,
      sym_type_qualifier,
      aux_sym_function_field_declarator_repeat1,
    ACTIONS(6764), 31,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
  [14680] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 14,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute,
      anon_sym_DOT,
    ACTIONS(4026), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [14736] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7333), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(9238), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7335), 35,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [14789] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7175), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(9240), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(7177), 35,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [14842] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4478), 1,
      anon_sym_EQ,
    ACTIONS(9242), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9252), 1,
      anon_sym_PIPE,
    ACTIONS(9254), 1,
      anon_sym_CARET,
    ACTIONS(9256), 1,
      anon_sym_AMP,
    ACTIONS(9266), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4480), 14,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [14930] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 4,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ,
    ACTIONS(4002), 20,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15002] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4000), 1,
      anon_sym_EQ,
    ACTIONS(9250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9252), 1,
      anon_sym_PIPE,
    ACTIONS(9254), 1,
      anon_sym_CARET,
    ACTIONS(9256), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 17,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15084] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4466), 1,
      anon_sym_EQ,
    ACTIONS(9242), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9252), 1,
      anon_sym_PIPE,
    ACTIONS(9254), 1,
      anon_sym_CARET,
    ACTIONS(9256), 1,
      anon_sym_AMP,
    ACTIONS(9266), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4468), 14,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15172] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4000), 1,
      anon_sym_EQ,
    ACTIONS(9252), 1,
      anon_sym_PIPE,
    ACTIONS(9254), 1,
      anon_sym_CARET,
    ACTIONS(9256), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 18,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15252] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4431), 1,
      anon_sym_EQ,
    ACTIONS(9248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9252), 1,
      anon_sym_PIPE,
    ACTIONS(9254), 1,
      anon_sym_CARET,
    ACTIONS(9256), 1,
      anon_sym_AMP,
    ACTIONS(9268), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4436), 15,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15338] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4474), 1,
      anon_sym_EQ,
    ACTIONS(9242), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9252), 1,
      anon_sym_PIPE,
    ACTIONS(9254), 1,
      anon_sym_CARET,
    ACTIONS(9256), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4476), 15,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15424] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 6,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_EQ,
    ACTIONS(4002), 22,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15492] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9256), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4000), 3,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_EQ,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 18,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15568] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4470), 1,
      anon_sym_EQ,
    ACTIONS(9242), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9252), 1,
      anon_sym_PIPE,
    ACTIONS(9254), 1,
      anon_sym_CARET,
    ACTIONS(9256), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4472), 15,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15654] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 4,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ,
    ACTIONS(4002), 18,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15728] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4458), 1,
      anon_sym_EQ,
    ACTIONS(9242), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9248), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9250), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9252), 1,
      anon_sym_PIPE,
    ACTIONS(9254), 1,
      anon_sym_CARET,
    ACTIONS(9256), 1,
      anon_sym_AMP,
    ACTIONS(9266), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4462), 14,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15816] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 8,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(4002), 22,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15882] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 10,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(4002), 22,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [15946] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9254), 1,
      anon_sym_CARET,
    ACTIONS(9256), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(4000), 2,
      anon_sym_PIPE,
      anon_sym_EQ,
    ACTIONS(9244), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9258), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9260), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9262), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9264), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9246), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 18,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [16024] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9271), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16075] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9273), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16126] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9275), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16177] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9277), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16228] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9279), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16279] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9281), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16330] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9283), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16381] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9285), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16432] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9287), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16483] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9289), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16534] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9291), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16585] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9154), 1,
      anon_sym_EQ,
    ACTIONS(9156), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 17,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      sym_identifier,
  [16638] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9293), 1,
      anon_sym_RPAREN,
    ACTIONS(5074), 3,
      anon_sym_const,
      anon_sym__Nullable,
      anon_sym___bridge,
    ACTIONS(5070), 37,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_CARET,
      anon_sym___extension__,
      anon_sym_noreturn,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym__Nonnull,
      anon_sym_nullable,
      anon_sym__Complex,
      anon_sym__Nullable_result,
      anon_sym__Null_unspecified,
      anon_sym___autoreleasing,
      anon_sym___block,
      anon_sym___bridge_retained,
      anon_sym___bridge_transfer,
      anon_sym___complex,
      anon_sym___const,
      anon_sym___imag,
      anon_sym___kindof,
      anon_sym___nonnull,
      anon_sym___nullable,
      anon_sym___ptrauth_objc_class_ro,
      anon_sym___ptrauth_objc_isa_pointer,
      anon_sym___ptrauth_objc_super_pointer,
      anon_sym___real,
      anon_sym___strong,
      anon_sym___unsafe_unretained,
      anon_sym___unused,
      anon_sym___weak,
      anon_sym_alignas,
      anon_sym__Alignas,
  [16689] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4431), 1,
      anon_sym_EQ,
    ACTIONS(9295), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9302), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9304), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9306), 1,
      anon_sym_PIPE,
    ACTIONS(9308), 1,
      anon_sym_CARET,
    ACTIONS(9310), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4436), 13,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [16773] = 4,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1435), 1,
      aux_sym_selector_expression_repeat2,
    ACTIONS(3994), 14,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
      anon_sym_DOT,
    ACTIONS(3996), 25,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [16823] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4470), 1,
      anon_sym_EQ,
    ACTIONS(9302), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9304), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9306), 1,
      anon_sym_PIPE,
    ACTIONS(9308), 1,
      anon_sym_CARET,
    ACTIONS(9310), 1,
      anon_sym_AMP,
    ACTIONS(9320), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4472), 13,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [16907] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4474), 1,
      anon_sym_EQ,
    ACTIONS(9302), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9304), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9306), 1,
      anon_sym_PIPE,
    ACTIONS(9308), 1,
      anon_sym_CARET,
    ACTIONS(9310), 1,
      anon_sym_AMP,
    ACTIONS(9320), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4476), 13,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [16991] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 10,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(4002), 20,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17053] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4466), 1,
      anon_sym_EQ,
    ACTIONS(9302), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9304), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9306), 1,
      anon_sym_PIPE,
    ACTIONS(9308), 1,
      anon_sym_CARET,
    ACTIONS(9310), 1,
      anon_sym_AMP,
    ACTIONS(9320), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9322), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4468), 12,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17139] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4000), 1,
      anon_sym_EQ,
    ACTIONS(9304), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9306), 1,
      anon_sym_PIPE,
    ACTIONS(9308), 1,
      anon_sym_CARET,
    ACTIONS(9310), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17219] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4000), 1,
      anon_sym_EQ,
    ACTIONS(9306), 1,
      anon_sym_PIPE,
    ACTIONS(9308), 1,
      anon_sym_CARET,
    ACTIONS(9310), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 16,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17297] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9308), 1,
      anon_sym_CARET,
    ACTIONS(9310), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(4000), 2,
      anon_sym_PIPE,
      anon_sym_EQ,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 16,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17373] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9310), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4000), 3,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_EQ,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4002), 16,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17447] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 4,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ,
    ACTIONS(4002), 16,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17519] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 4,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ,
    ACTIONS(4002), 18,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17589] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 6,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_EQ,
    ACTIONS(4002), 20,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17655] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4000), 8,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(4002), 20,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17719] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4478), 1,
      anon_sym_EQ,
    ACTIONS(9302), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9304), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9306), 1,
      anon_sym_PIPE,
    ACTIONS(9308), 1,
      anon_sym_CARET,
    ACTIONS(9310), 1,
      anon_sym_AMP,
    ACTIONS(9320), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9322), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4480), 12,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17805] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4458), 1,
      anon_sym_EQ,
    ACTIONS(9302), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9304), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9306), 1,
      anon_sym_PIPE,
    ACTIONS(9308), 1,
      anon_sym_CARET,
    ACTIONS(9310), 1,
      anon_sym_AMP,
    ACTIONS(9320), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9322), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(3992), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9298), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9312), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9314), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9316), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9318), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(4462), 12,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [17891] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8383), 1,
      anon_sym_EQ,
    ACTIONS(8385), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [17942] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_EQ,
    ACTIONS(9324), 1,
      anon_sym_SEMI,
    ACTIONS(4515), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(4024), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(4026), 13,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [17994] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(6966), 1,
      anon_sym_LBRACE,
    ACTIONS(9326), 1,
      sym_identifier,
    STATE(2918), 1,
      sym_field_declaration_list,
    STATE(5125), 1,
      sym_attribute_specifier,
    STATE(5558), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(4042), 8,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_DOT,
    ACTIONS(4044), 21,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18053] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(4882), 1,
      anon_sym_SEMI,
    ACTIONS(6966), 1,
      anon_sym_LBRACE,
    ACTIONS(9326), 1,
      sym_identifier,
    STATE(2918), 1,
      sym_field_declaration_list,
    STATE(5125), 1,
      sym_attribute_specifier,
    STATE(5558), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(4042), 8,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_DOT,
    ACTIONS(4044), 18,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18112] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4000), 5,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute,
    ACTIONS(4002), 17,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18172] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4478), 1,
      anon_sym___attribute,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4480), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
  [18254] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4470), 1,
      anon_sym___attribute,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4472), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18334] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4000), 7,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute,
    ACTIONS(4002), 17,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18392] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4000), 1,
      anon_sym___attribute,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 9,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18468] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4000), 1,
      anon_sym___attribute,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 10,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18542] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4000), 2,
      anon_sym_PIPE,
      anon_sym___attribute,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 10,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18614] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4000), 2,
      anon_sym_PIPE,
      anon_sym___attribute,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 11,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18684] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4000), 3,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym___attribute,
    ACTIONS(4002), 11,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18752] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4000), 3,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym___attribute,
    ACTIONS(4002), 13,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18818] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4000), 5,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute,
    ACTIONS(4002), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18880] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4431), 1,
      anon_sym___attribute,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9356), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4436), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [18960] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4466), 1,
      anon_sym___attribute,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4468), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
  [19042] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4474), 1,
      anon_sym___attribute,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4476), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [19122] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4458), 1,
      anon_sym___attribute,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4462), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON,
  [19204] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9359), 1,
      sym_identifier,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9363), 1,
      anon_sym_COMMA,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9389), 1,
      anon_sym_RBRACK,
    ACTIONS(9391), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5415), 1,
      aux_sym_attribute_repeat1,
    STATE(5559), 1,
      aux_sym_message_expression_repeat2,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [19290] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9395), 1,
      anon_sym___attribute,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9393), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [19370] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9359), 1,
      sym_identifier,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9391), 1,
      anon_sym_QMARK,
    ACTIONS(9397), 1,
      anon_sym_COMMA,
    ACTIONS(9399), 1,
      anon_sym_RBRACK,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5577), 1,
      aux_sym_message_expression_repeat2,
    STATE(5638), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [19456] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9391), 1,
      anon_sym_QMARK,
    ACTIONS(9403), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5146), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9401), 3,
      anon_sym_RBRACK,
      anon_sym_COLON,
      sym_identifier,
  [19538] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4000), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 11,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [19601] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9391), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4462), 4,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      sym_identifier,
  [19678] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9391), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4468), 4,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      sym_identifier,
  [19755] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4472), 5,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [19830] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4000), 6,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(4002), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [19885] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 7,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [19956] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 8,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [20025] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4000), 1,
      anon_sym_PIPE,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 8,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [20094] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4000), 1,
      anon_sym_PIPE,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 9,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [20161] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4000), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 9,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [20226] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4000), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(4002), 13,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [20285] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4000), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(4002), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [20342] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9391), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4480), 4,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      sym_identifier,
  [20419] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9405), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4436), 5,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [20494] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4476), 5,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      sym_identifier,
  [20569] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9391), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9408), 4,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
      sym_identifier,
  [20646] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9412), 2,
      anon_sym_SEMI,
      anon_sym_COLON,
  [20724] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(9418), 1,
      anon_sym_RPAREN,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5351), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [20804] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9446), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5511), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [20884] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9448), 1,
      anon_sym_COMMA,
    ACTIONS(9450), 1,
      anon_sym_RBRACE,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5697), 1,
      aux_sym_initializer_list_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [20964] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9452), 1,
      anon_sym_COMMA,
    ACTIONS(9454), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5483), 1,
      aux_sym_argument_list_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21044] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9456), 1,
      anon_sym_COMMA,
    ACTIONS(9458), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5438), 1,
      aux_sym_generic_expression_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21124] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9456), 1,
      anon_sym_COMMA,
    ACTIONS(9460), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5420), 1,
      aux_sym_generic_expression_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21204] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9462), 1,
      anon_sym_COMMA,
    ACTIONS(9464), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5432), 1,
      aux_sym_availability_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21284] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9466), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5608), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21364] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9452), 1,
      anon_sym_COMMA,
    ACTIONS(9468), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5617), 1,
      aux_sym_argument_list_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21444] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9470), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5387), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21524] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9472), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5417), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21604] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9474), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5454), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21684] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9476), 1,
      anon_sym_COMMA,
    ACTIONS(9478), 1,
      anon_sym_RBRACE,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5691), 1,
      aux_sym_initializer_list_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21764] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9480), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5474), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21844] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9482), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5492), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [21924] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9484), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5503), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [22004] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9359), 1,
      sym_identifier,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9391), 1,
      anon_sym_QMARK,
    ACTIONS(9486), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9488), 1,
      anon_sym_RBRACK,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5559), 1,
      aux_sym_message_expression_repeat2,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [22084] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9490), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5518), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [22164] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 6,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
  [22231] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9492), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [22308] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9494), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [22385] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9496), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [22462] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4468), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [22537] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4472), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_QMARK,
  [22610] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4000), 6,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(4002), 13,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_QMARK,
  [22663] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 5,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
  [22732] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9498), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [22809] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4000), 1,
      anon_sym_PIPE,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 6,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
  [22876] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4000), 1,
      anon_sym_PIPE,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 7,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_QMARK,
  [22941] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4000), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 7,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_QMARK,
  [23004] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4000), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4002), 9,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK,
  [23065] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4000), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(4002), 11,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_QMARK,
  [23122] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4000), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(4002), 13,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_QMARK,
  [23177] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9500), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [23254] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4480), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [23329] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9408), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [23404] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4462), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [23479] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9502), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [23554] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9504), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [23629] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9506), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [23706] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9508), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [23783] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9510), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [23860] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9512), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [23937] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9359), 1,
      sym_identifier,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9391), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5559), 1,
      aux_sym_message_expression_repeat2,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24014] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9514), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24091] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9516), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [24166] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9518), 1,
      anon_sym_COMMA,
    ACTIONS(9520), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24243] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9522), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24320] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9524), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24397] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9518), 1,
      anon_sym_COMMA,
    ACTIONS(9526), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24474] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9528), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24551] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9530), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24628] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9532), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24705] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9518), 1,
      anon_sym_COMMA,
    ACTIONS(9534), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24782] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9518), 1,
      anon_sym_COMMA,
    ACTIONS(9536), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24859] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9518), 1,
      anon_sym_COMMA,
    ACTIONS(9538), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [24936] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9540), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [25013] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9542), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [25090] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9412), 1,
      anon_sym_RPAREN,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9518), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [25167] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9544), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4436), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_QMARK,
  [25240] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(4476), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_QMARK,
  [25313] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9547), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [25390] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9518), 1,
      anon_sym_COMMA,
    ACTIONS(9549), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [25467] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9551), 1,
      anon_sym_RBRACK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4472), 2,
      anon_sym_QMARK,
      sym_identifier,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [25542] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9553), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [25617] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9555), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [25694] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9359), 1,
      sym_identifier,
    ACTIONS(9361), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9369), 1,
      anon_sym_SLASH,
    ACTIONS(9371), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9373), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9375), 1,
      anon_sym_PIPE,
    ACTIONS(9377), 1,
      anon_sym_CARET,
    ACTIONS(9379), 1,
      anon_sym_AMP,
    ACTIONS(9391), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    STATE(5577), 1,
      aux_sym_message_expression_repeat2,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9365), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9367), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9381), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9383), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9385), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9387), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [25771] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9557), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [25848] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9559), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [25925] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9561), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [26002] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9563), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [26079] = 4,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1620), 1,
      aux_sym_selector_expression_repeat2,
    ACTIONS(3994), 8,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_DOT,
    ACTIONS(3996), 20,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [26118] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9565), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [26195] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9567), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [26270] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9569), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [26345] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9571), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [26420] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9573), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [26497] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9575), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [26572] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9577), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [26649] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9579), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [26726] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9518), 1,
      anon_sym_COMMA,
    ACTIONS(9581), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [26803] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9518), 1,
      anon_sym_COMMA,
    ACTIONS(9583), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [26880] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9585), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [26957] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9587), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27034] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9408), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [27109] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9410), 1,
      anon_sym_COMMA,
    ACTIONS(9589), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27186] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8591), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27260] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9591), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27334] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9488), 1,
      anon_sym_RBRACK,
    ACTIONS(9593), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27408] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9595), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27482] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9597), 1,
      anon_sym_RBRACK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27556] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(4472), 1,
      anon_sym_QMARK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9551), 1,
      anon_sym_RBRACK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27630] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9599), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27704] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      sym_primitive_type,
    ACTIONS(67), 1,
      anon_sym_union,
    ACTIONS(3020), 1,
      anon_sym_struct,
    ACTIONS(4354), 1,
      anon_sym_enum,
    ACTIONS(9601), 1,
      sym_identifier,
    STATE(4269), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(5847), 1,
      sym_type_specifier,
    ACTIONS(117), 3,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
    ACTIONS(9603), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(147), 5,
      anon_sym_BOOL,
      anon_sym_IMP,
      anon_sym_SEL,
      anon_sym_Class,
      anon_sym_id,
    STATE(2971), 9,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
      sym_typeof_specifier,
      sym_generic_specifier,
      sym_typedefed_specifier,
      sym_array_type_specifier,
  [27758] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9605), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27832] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8637), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27906] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8599), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [27980] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9607), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28054] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9609), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28128] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9611), 1,
      anon_sym_RBRACK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28202] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9613), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28276] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9615), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28350] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9617), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28424] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9619), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28498] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9621), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28572] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9623), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28646] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9625), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28720] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9627), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28794] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8563), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28868] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8595), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [28942] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9629), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29016] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9631), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29090] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9633), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29164] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9635), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29238] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8629), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29312] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9637), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29386] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9639), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29460] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9641), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29534] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9643), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29608] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8621), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29682] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8603), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29756] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9645), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29830] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9647), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29904] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9649), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [29978] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9651), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30052] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9653), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30126] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8635), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30200] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9655), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30274] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8597), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30348] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9657), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30422] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9659), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30496] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9661), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30570] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9663), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30644] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8611), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30718] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9665), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30792] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9667), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30866] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9669), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [30940] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9671), 1,
      anon_sym_RBRACK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31014] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8623), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31088] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9673), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31162] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9675), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31236] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8613), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31310] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9677), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31384] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9679), 1,
      anon_sym_RBRACK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31458] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9681), 1,
      anon_sym_RBRACK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31532] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      sym_primitive_type,
    ACTIONS(67), 1,
      anon_sym_union,
    ACTIONS(3020), 1,
      anon_sym_struct,
    ACTIONS(4354), 1,
      anon_sym_enum,
    ACTIONS(9601), 1,
      sym_identifier,
    STATE(4269), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(5861), 1,
      sym_type_specifier,
    ACTIONS(117), 3,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
    ACTIONS(9603), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(147), 5,
      anon_sym_BOOL,
      anon_sym_IMP,
      anon_sym_SEL,
      anon_sym_Class,
      anon_sym_id,
    STATE(2971), 9,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
      sym_typeof_specifier,
      sym_generic_specifier,
      sym_typedefed_specifier,
      sym_array_type_specifier,
  [31586] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9683), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31660] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(8587), 1,
      anon_sym_RBRACK,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31734] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(6316), 1,
      anon_sym_COMMA,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31808] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9685), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31882] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9687), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [31956] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9689), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [32030] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9691), 1,
      anon_sym_RBRACK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [32104] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9693), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [32178] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(67), 1,
      anon_sym_union,
    ACTIONS(3020), 1,
      anon_sym_struct,
    ACTIONS(4354), 1,
      anon_sym_enum,
    ACTIONS(8840), 1,
      sym_identifier,
    ACTIONS(8844), 1,
      sym_primitive_type,
    STATE(4181), 1,
      sym_type_specifier,
    STATE(4185), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(117), 3,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
    ACTIONS(8842), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(147), 5,
      anon_sym_BOOL,
      anon_sym_IMP,
      anon_sym_SEL,
      anon_sym_Class,
      anon_sym_id,
    STATE(2971), 9,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
      sym_typeof_specifier,
      sym_generic_specifier,
      sym_typedefed_specifier,
      sym_array_type_specifier,
  [32232] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9695), 1,
      anon_sym_SEMI,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [32306] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      sym_primitive_type,
    ACTIONS(67), 1,
      anon_sym_union,
    ACTIONS(3020), 1,
      anon_sym_struct,
    ACTIONS(4354), 1,
      anon_sym_enum,
    ACTIONS(9601), 1,
      sym_identifier,
    STATE(4269), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(5758), 1,
      sym_type_specifier,
    ACTIONS(117), 3,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
    ACTIONS(9603), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(147), 5,
      anon_sym_BOOL,
      anon_sym_IMP,
      anon_sym_SEL,
      anon_sym_Class,
      anon_sym_id,
    STATE(2971), 9,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
      sym_typeof_specifier,
      sym_generic_specifier,
      sym_typedefed_specifier,
      sym_array_type_specifier,
  [32360] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9697), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [32434] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9699), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [32508] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(67), 1,
      anon_sym_union,
    ACTIONS(3020), 1,
      anon_sym_struct,
    ACTIONS(4354), 1,
      anon_sym_enum,
    ACTIONS(8840), 1,
      sym_identifier,
    ACTIONS(8844), 1,
      sym_primitive_type,
    STATE(4185), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(4195), 1,
      sym_type_specifier,
    ACTIONS(117), 3,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
    ACTIONS(8842), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(147), 5,
      anon_sym_BOOL,
      anon_sym_IMP,
      anon_sym_SEL,
      anon_sym_Class,
      anon_sym_id,
    STATE(2971), 9,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
      sym_typeof_specifier,
      sym_generic_specifier,
      sym_typedefed_specifier,
      sym_array_type_specifier,
  [32562] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9701), 1,
      anon_sym_COLON,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [32636] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      sym_primitive_type,
    ACTIONS(67), 1,
      anon_sym_union,
    ACTIONS(3020), 1,
      anon_sym_struct,
    ACTIONS(4354), 1,
      anon_sym_enum,
    ACTIONS(9601), 1,
      sym_identifier,
    STATE(4269), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(5772), 1,
      sym_type_specifier,
    ACTIONS(117), 3,
      anon_sym___typeof__,
      anon_sym___typeof,
      anon_sym_typeof,
    ACTIONS(9603), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(147), 5,
      anon_sym_BOOL,
      anon_sym_IMP,
      anon_sym_SEL,
      anon_sym_Class,
      anon_sym_id,
    STATE(2971), 9,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
      sym_typeof_specifier,
      sym_generic_specifier,
      sym_typedefed_specifier,
      sym_array_type_specifier,
  [32690] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9414), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9424), 1,
      anon_sym_SLASH,
    ACTIONS(9426), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9428), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9430), 1,
      anon_sym_PIPE,
    ACTIONS(9432), 1,
      anon_sym_CARET,
    ACTIONS(9434), 1,
      anon_sym_AMP,
    ACTIONS(9444), 1,
      anon_sym_QMARK,
    ACTIONS(9703), 1,
      anon_sym_RPAREN,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9420), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9422), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9436), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9438), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9440), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9442), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [32764] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    ACTIONS(9705), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [32838] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4882), 1,
      anon_sym_SEMI,
    ACTIONS(4042), 8,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_DOT,
    ACTIONS(4044), 18,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [32875] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9324), 1,
      anon_sym_SEMI,
    ACTIONS(4024), 8,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_DOT,
    ACTIONS(4026), 18,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [32912] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    ACTIONS(3982), 1,
      anon_sym_LBRACK,
    ACTIONS(3984), 1,
      anon_sym_DOT,
    ACTIONS(3986), 1,
      anon_sym_DASH_GT,
    ACTIONS(9332), 1,
      anon_sym_SLASH,
    ACTIONS(9334), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(9336), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9338), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9340), 1,
      anon_sym_PIPE,
    ACTIONS(9342), 1,
      anon_sym_CARET,
    ACTIONS(9344), 1,
      anon_sym_AMP,
    ACTIONS(9354), 1,
      anon_sym_QMARK,
    STATE(1509), 1,
      sym_argument_list,
    ACTIONS(4621), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(9328), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9330), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9346), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9348), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9350), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9352), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [32983] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(5417), 1,
      anon_sym_LPAREN2,
    ACTIONS(5419), 1,
      anon_sym_STAR,
    ACTIONS(5421), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(3046), 1,
      sym__declarator,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(5154), 1,
      sym__abstract_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    ACTIONS(9707), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
  [33035] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9711), 1,
      anon_sym_RPAREN,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9721), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4583), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33079] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9727), 1,
      anon_sym_LPAREN2,
    STATE(4626), 1,
      sym_preproc_argument_list,
    ACTIONS(9729), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9725), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [33113] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9731), 1,
      anon_sym_RPAREN,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9733), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4606), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33157] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9735), 1,
      anon_sym_RPAREN,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9737), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4579), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33201] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9739), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4651), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33242] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9751), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4671), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33283] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9755), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4678), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33324] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9757), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4652), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33365] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9759), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4682), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33406] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9761), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4653), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33447] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9763), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4654), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33488] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9765), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4685), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33529] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9767), 1,
      anon_sym_COMMA,
    ACTIONS(9769), 1,
      anon_sym_RPAREN,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9777), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9779), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9781), 1,
      anon_sym_PIPE,
    ACTIONS(9783), 1,
      anon_sym_CARET,
    ACTIONS(9785), 1,
      anon_sym_AMP,
    STATE(5392), 1,
      aux_sym_preproc_argument_list_repeat1,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [33584] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9795), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4640), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33625] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9797), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4664), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33666] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9799), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4666), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33707] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9767), 1,
      anon_sym_COMMA,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9777), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9779), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9781), 1,
      anon_sym_PIPE,
    ACTIONS(9783), 1,
      anon_sym_CARET,
    ACTIONS(9785), 1,
      anon_sym_AMP,
    ACTIONS(9801), 1,
      anon_sym_RPAREN,
    STATE(5660), 1,
      aux_sym_preproc_argument_list_repeat1,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [33762] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9803), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4658), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33803] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9805), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4660), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33844] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9807), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4647), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33885] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9809), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4680), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33926] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9811), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4661), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [33967] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9813), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4657), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34008] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9815), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4675), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34049] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9817), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4632), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34090] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9819), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4633), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34131] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9821), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4636), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34172] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9823), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4637), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34213] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9825), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4646), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34254] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9827), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4679), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34295] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9829), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4641), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34336] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9831), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4681), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34377] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9833), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4642), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34418] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9835), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4643), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34459] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9837), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4655), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34500] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9839), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4669), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34541] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4302), 1,
      anon_sym_enum,
    ACTIONS(6363), 1,
      sym_identifier,
    ACTIONS(6365), 1,
      anon_sym_LPAREN2,
    ACTIONS(6367), 1,
      anon_sym_STAR,
    ACTIONS(6369), 1,
      anon_sym_CARET,
    ACTIONS(9841), 1,
      anon_sym_SEMI,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    STATE(4795), 1,
      sym__field_declarator,
    STATE(4891), 1,
      sym_enum_specifier,
    STATE(5280), 1,
      sym_bitfield_clause,
    STATE(6145), 1,
      sym_attribute_specifier,
    STATE(6222), 1,
      sym_ms_based_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(4889), 6,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
      sym_block_pointer_field_declarator,
  [34596] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4302), 1,
      anon_sym_enum,
    ACTIONS(6363), 1,
      sym_identifier,
    ACTIONS(6365), 1,
      anon_sym_LPAREN2,
    ACTIONS(6367), 1,
      anon_sym_STAR,
    ACTIONS(6369), 1,
      anon_sym_CARET,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(9845), 1,
      anon_sym_SEMI,
    STATE(4797), 1,
      sym__field_declarator,
    STATE(4892), 1,
      sym_enum_specifier,
    STATE(5270), 1,
      sym_bitfield_clause,
    STATE(6222), 1,
      sym_ms_based_modifier,
    STATE(6483), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(4889), 6,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
      sym_block_pointer_field_declarator,
  [34651] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9847), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4644), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34692] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9767), 1,
      anon_sym_COMMA,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9777), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9779), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9781), 1,
      anon_sym_PIPE,
    ACTIONS(9783), 1,
      anon_sym_CARET,
    ACTIONS(9785), 1,
      anon_sym_AMP,
    ACTIONS(9849), 1,
      anon_sym_RPAREN,
    STATE(5322), 1,
      aux_sym_preproc_argument_list_repeat1,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [34747] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9851), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4645), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34788] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9853), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4659), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34829] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9855), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4624), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34870] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9857), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4686), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34911] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9859), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4683), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34952] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9709), 1,
      sym_identifier,
    ACTIONS(9713), 1,
      anon_sym_LPAREN2,
    ACTIONS(9715), 1,
      anon_sym_defined,
    ACTIONS(9717), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9719), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9861), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9723), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4630), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [34993] = 5,
    ACTIONS(9725), 1,
      anon_sym_LF,
    ACTIONS(9863), 1,
      anon_sym_LPAREN2,
    ACTIONS(9865), 1,
      sym_comment,
    STATE(4668), 1,
      sym_preproc_argument_list,
    ACTIONS(9729), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35026] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9867), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4649), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [35067] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9869), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4670), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [35108] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9871), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4662), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [35149] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9873), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4648), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [35190] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4302), 1,
      anon_sym_enum,
    ACTIONS(6363), 1,
      sym_identifier,
    ACTIONS(6365), 1,
      anon_sym_LPAREN2,
    ACTIONS(6367), 1,
      anon_sym_STAR,
    ACTIONS(6369), 1,
      anon_sym_CARET,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(9875), 1,
      anon_sym_SEMI,
    STATE(4800), 1,
      sym__field_declarator,
    STATE(4929), 1,
      sym_enum_specifier,
    STATE(5198), 1,
      sym_bitfield_clause,
    STATE(6222), 1,
      sym_ms_based_modifier,
    STATE(6519), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(4889), 6,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
      sym_block_pointer_field_declarator,
  [35245] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9877), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4650), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [35286] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9879), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4673), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [35327] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9881), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4676), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [35368] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9741), 1,
      sym_identifier,
    ACTIONS(9743), 1,
      anon_sym_LPAREN2,
    ACTIONS(9745), 1,
      anon_sym_defined,
    ACTIONS(9747), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(9749), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9883), 2,
      sym_number_literal,
      sym_system_lib_string,
    ACTIONS(9753), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(4677), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [35409] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4122), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(4124), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35437] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9887), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9885), 11,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35471] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9891), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9889), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35499] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9895), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9893), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35527] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9899), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9897), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35555] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(9901), 1,
      sym_identifier,
    ACTIONS(9903), 1,
      sym_version_number,
    ACTIONS(9905), 1,
      anon_sym_COLON2,
    STATE(4832), 1,
      sym_string_literal,
    STATE(5052), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5523), 1,
      sym_platform,
    STATE(5592), 3,
      sym_concatenated_string,
      sym_version,
      sym_method_identifier,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(8339), 5,
      anon_sym_ios,
      anon_sym_tvos,
      anon_sym_macos,
      anon_sym_macosx,
      anon_sym_watchos,
  [35599] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9909), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9907), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35627] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9777), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9779), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9781), 1,
      anon_sym_PIPE,
    ACTIONS(9783), 1,
      anon_sym_CARET,
    ACTIONS(9785), 1,
      anon_sym_AMP,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9911), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [35677] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9915), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9913), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35705] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9887), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9885), 13,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35737] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9887), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9885), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35765] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(9903), 1,
      sym_version_number,
    ACTIONS(9905), 1,
      anon_sym_COLON2,
    ACTIONS(9917), 1,
      sym_identifier,
    STATE(4825), 1,
      sym_string_literal,
    STATE(5052), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5523), 1,
      sym_platform,
    STATE(5515), 3,
      sym_concatenated_string,
      sym_version,
      sym_method_identifier,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(8339), 5,
      anon_sym_ios,
      anon_sym_tvos,
      anon_sym_macos,
      anon_sym_macosx,
      anon_sym_watchos,
  [35809] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9921), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9919), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [35837] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9779), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9781), 1,
      anon_sym_PIPE,
    ACTIONS(9783), 1,
      anon_sym_CARET,
    ACTIONS(9785), 1,
      anon_sym_AMP,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9885), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
  [35885] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9781), 1,
      anon_sym_PIPE,
    ACTIONS(9783), 1,
      anon_sym_CARET,
    ACTIONS(9785), 1,
      anon_sym_AMP,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9885), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [35931] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(9903), 1,
      sym_version_number,
    ACTIONS(9905), 1,
      anon_sym_COLON2,
    ACTIONS(9923), 1,
      sym_identifier,
    STATE(4870), 1,
      sym_string_literal,
    STATE(5052), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5523), 1,
      sym_platform,
    STATE(5794), 3,
      sym_concatenated_string,
      sym_version,
      sym_method_identifier,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
    ACTIONS(8339), 5,
      anon_sym_ios,
      anon_sym_tvos,
      anon_sym_macos,
      anon_sym_macosx,
      anon_sym_watchos,
  [35975] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9927), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9925), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [36003] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9931), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9929), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [36031] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9783), 1,
      anon_sym_CARET,
    ACTIONS(9785), 1,
      anon_sym_AMP,
    ACTIONS(9887), 1,
      anon_sym_PIPE,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9885), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [36077] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9785), 1,
      anon_sym_AMP,
    ACTIONS(9887), 1,
      anon_sym_PIPE,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9885), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [36121] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9887), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(9885), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [36163] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9887), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(9885), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [36203] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9887), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9885), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [36239] = 9,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9885), 1,
      anon_sym_LF,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9887), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36278] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9945), 1,
      anon_sym_LF,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36323] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9955), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36368] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9957), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36413] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9959), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36458] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9777), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9779), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9781), 1,
      anon_sym_PIPE,
    ACTIONS(9783), 1,
      anon_sym_CARET,
    ACTIONS(9785), 1,
      anon_sym_AMP,
    ACTIONS(9961), 1,
      anon_sym_RPAREN,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [36507] = 8,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9885), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
    ACTIONS(9887), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
  [36544] = 7,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9885), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
    ACTIONS(9887), 7,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [36579] = 6,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9885), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9887), 11,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36612] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9963), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36657] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9897), 1,
      anon_sym_LF,
    ACTIONS(9899), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [36684] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9965), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36729] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9967), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36774] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9969), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36819] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9885), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9887), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [36850] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9971), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36895] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9973), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [36940] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9889), 1,
      anon_sym_LF,
    ACTIONS(9891), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [36967] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9975), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37012] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9907), 1,
      anon_sym_LF,
    ACTIONS(9909), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [37039] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9977), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37084] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9913), 1,
      anon_sym_LF,
    ACTIONS(9915), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [37111] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9893), 1,
      anon_sym_LF,
    ACTIONS(9895), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [37138] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9775), 1,
      anon_sym_SLASH,
    ACTIONS(9777), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9779), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9781), 1,
      anon_sym_PIPE,
    ACTIONS(9783), 1,
      anon_sym_CARET,
    ACTIONS(9785), 1,
      anon_sym_AMP,
    ACTIONS(9979), 1,
      anon_sym_RPAREN,
    ACTIONS(9771), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9773), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(9787), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9789), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(9791), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(9793), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [37187] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9885), 1,
      anon_sym_LF,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9887), 15,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [37216] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9981), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37261] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9925), 1,
      anon_sym_LF,
    ACTIONS(9927), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [37288] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9885), 1,
      anon_sym_LF,
    ACTIONS(9887), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [37315] = 3,
    ACTIONS(4124), 1,
      anon_sym_LF,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(4122), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [37342] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9983), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37387] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9885), 1,
      anon_sym_LF,
    ACTIONS(9887), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37432] = 11,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9885), 1,
      anon_sym_LF,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9887), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37475] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9985), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37520] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9987), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37565] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9989), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37610] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9991), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37655] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9993), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37700] = 10,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9885), 1,
      anon_sym_LF,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9887), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37741] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9919), 1,
      anon_sym_LF,
    ACTIONS(9921), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [37768] = 12,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9937), 1,
      anon_sym_AMP,
    ACTIONS(9947), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(9949), 1,
      anon_sym_AMP_AMP,
    ACTIONS(9951), 1,
      anon_sym_PIPE,
    ACTIONS(9953), 1,
      anon_sym_CARET,
    ACTIONS(9995), 1,
      anon_sym_LF,
    ACTIONS(9933), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(9939), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(9943), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(9935), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(9941), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [37813] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9929), 1,
      anon_sym_LF,
    ACTIONS(9931), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [37840] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6662), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [37884] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6489), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [37928] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6510), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [37972] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6516), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38016] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6527), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38060] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6589), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38104] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6640), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38148] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6641), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38192] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6642), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38236] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6732), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38280] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6744), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38324] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6332), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38368] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6783), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38412] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6785), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38456] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5905), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38500] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5935), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38544] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5959), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38588] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5963), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38632] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5966), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38676] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5980), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38720] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5982), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38764] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5984), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38808] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5992), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38852] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5994), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38896] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6002), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38940] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6012), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [38984] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6025), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39028] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6028), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39072] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6029), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39116] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6047), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39160] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6048), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39204] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6049), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39248] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6056), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39292] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6057), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39336] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6059), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39380] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6152), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39424] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6256), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39468] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6154), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39512] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6155), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39556] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6260), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39600] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6329), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39644] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6267), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39688] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6711), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39732] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6577), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39776] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6130), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39820] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6249), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39864] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6251), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39908] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6494), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39952] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6500), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [39996] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6502), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40040] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6252), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40084] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6305), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40128] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6309), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40172] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6328), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40216] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6471), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40260] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6572), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40304] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6575), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40348] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6580), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40392] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6650), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40436] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6652), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40480] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6616), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40524] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6655), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40568] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6617), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40612] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6618), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40656] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6698), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40700] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6701), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40744] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6730), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40788] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6727), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40832] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6731), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40876] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6354), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40920] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6036), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [40964] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5909), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41008] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6064), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41052] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6530), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41096] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6533), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41140] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6535), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41184] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5971), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41228] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(5978), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41272] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6004), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41316] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6722), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41360] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6065), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41404] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6082), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41448] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6424), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41492] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6129), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41536] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6432), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41580] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6434), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41624] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6293), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41668] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6421), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41712] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6414), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41756] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6426), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41800] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6430), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41844] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6479), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41888] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6243), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41932] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6244), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [41976] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6272), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [42020] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6481), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [42064] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6484), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [42108] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6070), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [42152] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6073), 1,
      sym__type_definition_declarators,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [42196] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3033), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    STATE(6767), 1,
      sym__type_definition_declarators,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [42240] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5686), 1,
      sym_identifier,
    ACTIONS(5688), 1,
      anon_sym_LPAREN2,
    ACTIONS(5690), 1,
      anon_sym_STAR,
    ACTIONS(5692), 1,
      anon_sym_CARET,
    ACTIONS(5696), 1,
      sym_primitive_type,
    STATE(3262), 1,
      sym__type_declarator,
    STATE(6578), 1,
      sym_ms_based_modifier,
    ACTIONS(5694), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(3256), 5,
      sym_parenthesized_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
      sym_block_pointer_type_declarator,
  [42281] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4488), 1,
      anon_sym_LPAREN2,
    ACTIONS(6850), 1,
      anon_sym_STAR,
    ACTIONS(6852), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    ACTIONS(9999), 1,
      anon_sym___attribute,
    STATE(4290), 1,
      sym_parameter_list,
    STATE(4966), 1,
      sym__abstract_declarator,
    ACTIONS(9997), 3,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
  [42318] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4302), 1,
      anon_sym_enum,
    ACTIONS(6363), 1,
      sym_identifier,
    ACTIONS(6365), 1,
      anon_sym_LPAREN2,
    ACTIONS(6367), 1,
      anon_sym_STAR,
    ACTIONS(6369), 1,
      anon_sym_CARET,
    STATE(4829), 1,
      sym__field_declarator,
    STATE(5021), 1,
      sym_enum_specifier,
    STATE(6222), 1,
      sym_ms_based_modifier,
    STATE(4889), 6,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
      sym_block_pointer_field_declarator,
  [42357] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3014), 1,
      anon_sym_LBRACK,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(6972), 1,
      sym_identifier,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    STATE(4817), 1,
      sym_method_type,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(2878), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
    STATE(4987), 2,
      sym_attribute_specifier,
      sym_attribute_declaration,
  [42395] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3014), 1,
      anon_sym_LBRACK,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(6972), 1,
      sym_identifier,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    STATE(4819), 1,
      sym_method_type,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(2895), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
    STATE(4956), 2,
      sym_attribute_specifier,
      sym_attribute_declaration,
  [42433] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    ACTIONS(8366), 1,
      anon_sym_COLON,
    STATE(3242), 1,
      sym__declarator,
    STATE(3285), 1,
      sym_struct_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [42471] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3014), 1,
      anon_sym_LBRACK,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(6972), 1,
      sym_identifier,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    STATE(4816), 1,
      sym_method_type,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(2892), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
    STATE(4960), 2,
      sym_attribute_specifier,
      sym_attribute_declaration,
  [42509] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4488), 1,
      anon_sym_LPAREN2,
    ACTIONS(4490), 1,
      anon_sym_STAR,
    ACTIONS(4492), 1,
      anon_sym_CARET,
    ACTIONS(7963), 1,
      anon_sym_LBRACK,
    STATE(4271), 1,
      sym_parameter_list,
    STATE(5020), 1,
      sym__abstract_declarator,
    ACTIONS(9997), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
    STATE(4856), 5,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
      sym_abstract_block_pointer_declarator,
  [42543] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10005), 1,
      anon_sym_SEMI,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    STATE(4273), 1,
      sym_parameter_list,
    STATE(4933), 1,
      sym_bitfield_clause,
    STATE(4934), 1,
      aux_sym_field_declaration_repeat1,
    STATE(6562), 1,
      sym_attribute_specifier,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [42584] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(2051), 1,
      sym__old_style_function_declarator,
    STATE(4964), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [42619] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10009), 1,
      anon_sym_SEMI,
    STATE(4273), 1,
      sym_parameter_list,
    STATE(4895), 1,
      sym_bitfield_clause,
    STATE(4900), 1,
      aux_sym_field_declaration_repeat1,
    STATE(6125), 1,
      sym_attribute_specifier,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [42660] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(2010), 1,
      sym__old_style_function_declarator,
    STATE(4962), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [42695] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(2831), 1,
      sym_init_declarator,
    STATE(3233), 1,
      sym__declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [42730] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10011), 1,
      anon_sym_SEMI,
    STATE(4273), 1,
      sym_parameter_list,
    STATE(4940), 1,
      sym_bitfield_clause,
    STATE(4941), 1,
      aux_sym_field_declaration_repeat1,
    STATE(6648), 1,
      sym_attribute_specifier,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [42771] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(6475), 1,
      anon_sym_LPAREN2,
    ACTIONS(6477), 1,
      anon_sym_STAR,
    ACTIONS(6479), 1,
      anon_sym_CARET,
    STATE(3222), 1,
      sym__declarator,
    STATE(3274), 1,
      sym_init_declarator,
    STATE(6199), 1,
      sym_ms_based_modifier,
    STATE(3065), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [42806] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(10015), 1,
      anon_sym_RPAREN,
    ACTIONS(10017), 1,
      anon_sym_LPAREN2,
    ACTIONS(10019), 1,
      anon_sym_COLON2,
    STATE(1916), 1,
      sym_string_literal,
    STATE(5035), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5487), 1,
      aux_sym_availability_attribute_specifier_repeat1,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [42841] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(1995), 1,
      sym__old_style_function_declarator,
    STATE(4996), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [42876] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(103), 1,
      anon_sym___asm,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(4880), 1,
      sym_gnu_asm_expression,
    ACTIONS(6806), 2,
      anon_sym_asm,
      anon_sym___asm__,
    STATE(4886), 2,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(6473), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [42907] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(6463), 1,
      sym_identifier,
    ACTIONS(6465), 1,
      anon_sym_LPAREN2,
    ACTIONS(6467), 1,
      anon_sym_STAR,
    ACTIONS(6469), 1,
      anon_sym_CARET,
    STATE(2684), 1,
      sym__declarator,
    STATE(2831), 1,
      sym_init_declarator,
    STATE(6101), 1,
      sym_ms_based_modifier,
    STATE(2778), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [42942] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(2012), 1,
      sym__old_style_function_declarator,
    STATE(4971), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [42977] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(2016), 1,
      sym__old_style_function_declarator,
    STATE(4959), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [43012] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(6463), 1,
      sym_identifier,
    ACTIONS(6465), 1,
      anon_sym_LPAREN2,
    ACTIONS(6467), 1,
      anon_sym_STAR,
    ACTIONS(6469), 1,
      anon_sym_CARET,
    STATE(2663), 1,
      sym__declarator,
    STATE(2789), 1,
      sym_init_declarator,
    STATE(6101), 1,
      sym_ms_based_modifier,
    STATE(2778), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [43047] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(10019), 1,
      anon_sym_COLON2,
    ACTIONS(10021), 1,
      anon_sym_RPAREN,
    ACTIONS(10023), 1,
      anon_sym_LPAREN2,
    STATE(1916), 1,
      sym_string_literal,
    STATE(5035), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5643), 1,
      aux_sym_availability_attribute_specifier_repeat1,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [43082] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(5029), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [43114] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(5047), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [43146] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(5090), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [43178] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(5006), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [43210] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(5080), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [43242] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10025), 1,
      sym_identifier,
    STATE(2069), 1,
      sym_string_literal,
    ACTIONS(10027), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(5182), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [43270] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3014), 1,
      anon_sym_LBRACK,
    ACTIONS(6972), 1,
      sym_identifier,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(2883), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
    STATE(4969), 2,
      sym_attribute_specifier,
      sym_attribute_declaration,
  [43302] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3014), 1,
      anon_sym_LBRACK,
    ACTIONS(6972), 1,
      sym_identifier,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(2879), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
    STATE(4984), 2,
      sym_attribute_specifier,
      sym_attribute_declaration,
  [43334] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(5005), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [43366] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3014), 1,
      anon_sym_LBRACK,
    ACTIONS(6972), 1,
      sym_identifier,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(2891), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
    STATE(4965), 2,
      sym_attribute_specifier,
      sym_attribute_declaration,
  [43398] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(5041), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [43430] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10019), 1,
      anon_sym_COLON2,
    ACTIONS(10031), 1,
      anon_sym_LPAREN2,
    STATE(1916), 1,
      sym_string_literal,
    STATE(5035), 1,
      aux_sym_method_identifier_repeat1,
    ACTIONS(10029), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [43460] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4092), 1,
      anon_sym___based,
    ACTIONS(4482), 1,
      anon_sym_LPAREN2,
    ACTIONS(4484), 1,
      anon_sym_STAR,
    ACTIONS(4486), 1,
      anon_sym_CARET,
    ACTIONS(5312), 1,
      sym_identifier,
    STATE(5019), 1,
      sym__declarator,
    STATE(6198), 1,
      sym_ms_based_modifier,
    STATE(3980), 5,
      sym_parenthesized_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
      sym_block_pointer_declarator,
  [43492] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10033), 1,
      sym_identifier,
    STATE(1692), 1,
      sym_method_parameter,
    STATE(4874), 1,
      sym_method_type,
    STATE(5094), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1725), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [43525] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10035), 1,
      sym_identifier,
    ACTIONS(10037), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(10039), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(10041), 1,
      aux_sym_preproc_elif_token1,
    STATE(4831), 1,
      aux_sym_preproc_ifdef_in_enumerator_repeat1,
    STATE(6436), 1,
      sym_enumerator,
    ACTIONS(10043), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(6435), 3,
      sym_preproc_elifdef,
      sym_preproc_else_in_enumerator,
      sym_preproc_elif_in_enumerator,
  [43556] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(5324), 1,
      sym_identifier,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(10015), 1,
      anon_sym_RPAREN,
    STATE(1916), 1,
      sym_string_literal,
    STATE(5487), 1,
      aux_sym_availability_attribute_specifier_repeat1,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [43585] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8339), 1,
      anon_sym_macos,
    ACTIONS(9903), 1,
      sym_version_number,
    ACTIONS(10045), 1,
      anon_sym_COMMA,
    ACTIONS(10047), 1,
      anon_sym_RPAREN,
    STATE(5523), 1,
      sym_platform,
    STATE(5535), 1,
      aux_sym_available_expression_repeat1,
    STATE(5538), 1,
      sym_version,
    ACTIONS(10049), 4,
      anon_sym_ios,
      anon_sym_tvos,
      anon_sym_macosx,
      anon_sym_watchos,
  [43616] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10051), 1,
      sym_identifier,
    STATE(1674), 1,
      sym_method_parameter,
    STATE(4873), 1,
      sym_method_type,
    STATE(5092), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1717), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [43649] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10053), 1,
      sym_identifier,
    STATE(1685), 1,
      sym_method_parameter,
    STATE(4877), 1,
      sym_method_type,
    STATE(5064), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1719), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [43682] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10057), 1,
      anon_sym___attribute,
    ACTIONS(10059), 1,
      anon_sym_COLON,
    STATE(4273), 1,
      sym_parameter_list,
    STATE(5139), 1,
      sym_bitfield_clause,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
    ACTIONS(10055), 3,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [43713] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10064), 1,
      anon_sym_LBRACK,
    ACTIONS(10066), 1,
      anon_sym_AT,
    STATE(5291), 1,
      sym_gnu_asm_output_operand,
    STATE(6714), 1,
      sym_string_literal,
    ACTIONS(10062), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [43740] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10035), 1,
      sym_identifier,
    ACTIONS(10039), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(10041), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(10068), 1,
      aux_sym_preproc_if_token2,
    STATE(4910), 1,
      aux_sym_preproc_ifdef_in_enumerator_repeat1,
    STATE(6436), 1,
      sym_enumerator,
    ACTIONS(10043), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(6067), 3,
      sym_preproc_elifdef,
      sym_preproc_else_in_enumerator,
      sym_preproc_elif_in_enumerator,
  [43771] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(5324), 1,
      sym_identifier,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(10021), 1,
      anon_sym_RPAREN,
    STATE(1916), 1,
      sym_string_literal,
    STATE(5643), 1,
      aux_sym_availability_attribute_specifier_repeat1,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [43800] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8339), 1,
      anon_sym_macos,
    ACTIONS(9903), 1,
      sym_version_number,
    ACTIONS(10045), 1,
      anon_sym_COMMA,
    ACTIONS(10070), 1,
      anon_sym_RPAREN,
    STATE(5523), 1,
      sym_platform,
    STATE(5644), 1,
      aux_sym_available_expression_repeat1,
    STATE(5645), 1,
      sym_version,
    ACTIONS(10049), 4,
      anon_sym_ios,
      anon_sym_tvos,
      anon_sym_macosx,
      anon_sym_watchos,
  [43831] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10066), 1,
      anon_sym_AT,
    ACTIONS(10074), 1,
      anon_sym_LBRACK,
    STATE(5294), 1,
      sym_gnu_asm_input_operand,
    STATE(6011), 1,
      sym_string_literal,
    ACTIONS(10072), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [43858] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10078), 1,
      anon_sym___attribute,
    ACTIONS(10076), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [43876] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10080), 1,
      anon_sym___attribute,
    ACTIONS(9238), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [43894] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10082), 1,
      sym_identifier,
    ACTIONS(10084), 1,
      anon_sym_AT,
    ACTIONS(10088), 1,
      sym_system_lib_string,
    STATE(6474), 2,
      sym_preproc_call_expression,
      sym_string_literal,
    ACTIONS(10086), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [43918] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10092), 1,
      anon_sym___attribute,
    ACTIONS(10090), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [43936] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10094), 1,
      anon_sym___attribute,
    ACTIONS(9240), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [43954] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10098), 1,
      anon_sym___attribute,
    ACTIONS(10096), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [43972] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10084), 1,
      anon_sym_AT,
    ACTIONS(10100), 1,
      sym_identifier,
    ACTIONS(10102), 1,
      sym_system_lib_string,
    STATE(6095), 2,
      sym_preproc_call_expression,
      sym_string_literal,
    ACTIONS(10086), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [43996] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10106), 1,
      anon_sym___attribute,
    ACTIONS(10104), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [44014] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10025), 1,
      sym_identifier,
    STATE(2069), 1,
      sym_string_literal,
    STATE(5469), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44038] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7045), 1,
      anon_sym_LT,
    ACTIONS(10108), 1,
      anon_sym_LPAREN2,
    STATE(2908), 1,
      aux_sym_generic_specifier_repeat1,
    STATE(2961), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(6134), 2,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
    ACTIONS(8381), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
  [44064] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10112), 1,
      anon_sym___attribute,
    STATE(4273), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
    ACTIONS(10110), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [44090] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10025), 1,
      sym_identifier,
    STATE(2069), 1,
      sym_string_literal,
    STATE(5636), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44114] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10116), 1,
      anon_sym___attribute,
    ACTIONS(10114), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [44132] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8339), 1,
      anon_sym_macos,
    ACTIONS(9903), 1,
      sym_version_number,
    STATE(5523), 1,
      sym_platform,
    STATE(5780), 1,
      sym_version,
    ACTIONS(10118), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(10049), 4,
      anon_sym_ios,
      anon_sym_tvos,
      anon_sym_macosx,
      anon_sym_watchos,
  [44158] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10025), 1,
      sym_identifier,
    STATE(2069), 1,
      sym_string_literal,
    STATE(5465), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44182] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10025), 1,
      sym_identifier,
    STATE(2069), 1,
      sym_string_literal,
    STATE(5498), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44206] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10122), 1,
      anon_sym___attribute,
    STATE(4273), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
    ACTIONS(10120), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [44232] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10025), 1,
      sym_identifier,
    STATE(2069), 1,
      sym_string_literal,
    STATE(5586), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44256] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10084), 1,
      anon_sym_AT,
    ACTIONS(10124), 1,
      sym_identifier,
    ACTIONS(10126), 1,
      sym_system_lib_string,
    STATE(6389), 2,
      sym_preproc_call_expression,
      sym_string_literal,
    ACTIONS(10086), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44280] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10130), 1,
      anon_sym___attribute,
    STATE(4273), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
    ACTIONS(10128), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [44306] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10025), 1,
      sym_identifier,
    STATE(2069), 1,
      sym_string_literal,
    STATE(5631), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44330] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10134), 1,
      anon_sym___attribute,
    ACTIONS(10132), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [44348] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10084), 1,
      anon_sym_AT,
    ACTIONS(10136), 1,
      sym_identifier,
    ACTIONS(10138), 1,
      sym_system_lib_string,
    STATE(6758), 2,
      sym_preproc_call_expression,
      sym_string_literal,
    ACTIONS(10086), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44372] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10025), 1,
      sym_identifier,
    STATE(2069), 1,
      sym_string_literal,
    STATE(5323), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44396] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10084), 1,
      anon_sym_AT,
    ACTIONS(10140), 1,
      sym_identifier,
    ACTIONS(10142), 1,
      sym_system_lib_string,
    STATE(6043), 2,
      sym_preproc_call_expression,
      sym_string_literal,
    ACTIONS(10086), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44420] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10146), 1,
      anon_sym___attribute,
    ACTIONS(10144), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [44438] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10150), 1,
      anon_sym___attribute,
    ACTIONS(10152), 1,
      anon_sym_LBRACK,
    STATE(4861), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
    ACTIONS(10148), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [44460] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10025), 1,
      sym_identifier,
    STATE(2069), 1,
      sym_string_literal,
    STATE(5407), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44484] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6978), 1,
      anon_sym_LBRACK,
    ACTIONS(10157), 1,
      anon_sym___attribute,
    STATE(4861), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
    ACTIONS(10155), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [44506] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10161), 1,
      anon_sym___attribute,
    ACTIONS(10159), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [44524] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10165), 1,
      anon_sym___attribute,
    STATE(4273), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
    ACTIONS(10163), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [44550] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10169), 1,
      anon_sym___attribute,
    STATE(4273), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
    ACTIONS(10167), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [44576] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10173), 1,
      anon_sym___attribute,
    STATE(4273), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
    ACTIONS(10171), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [44602] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(10025), 1,
      sym_identifier,
    STATE(2069), 1,
      sym_string_literal,
    STATE(5527), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44626] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10177), 1,
      anon_sym___attribute,
    ACTIONS(10175), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_GT,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
      anon_sym_COLON,
  [44644] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    ACTIONS(5324), 1,
      sym_identifier,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(10029), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44668] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10179), 1,
      sym_identifier,
    ACTIONS(10181), 1,
      anon_sym_COLON,
    STATE(3144), 1,
      sym_enumerator_list,
    STATE(5022), 1,
      sym_attribute_specifier,
    STATE(5207), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [44697] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10183), 1,
      sym_identifier,
    ACTIONS(10185), 1,
      anon_sym_COLON,
    STATE(4148), 1,
      sym_enumerator_list,
    STATE(5091), 1,
      sym_attribute_specifier,
    STATE(5217), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [44726] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10187), 1,
      sym_identifier,
    STATE(1686), 1,
      sym_method_parameter,
    STATE(5045), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1722), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [44753] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10189), 1,
      sym_identifier,
    STATE(1703), 1,
      sym_method_parameter,
    STATE(5061), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1718), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [44780] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10185), 1,
      anon_sym_COLON,
    ACTIONS(10191), 1,
      sym_identifier,
    STATE(4148), 1,
      sym_enumerator_list,
    STATE(5040), 1,
      sym_attribute_specifier,
    STATE(5240), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [44809] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10181), 1,
      anon_sym_COLON,
    ACTIONS(10193), 1,
      sym_identifier,
    STATE(3144), 1,
      sym_enumerator_list,
    STATE(5053), 1,
      sym_attribute_specifier,
    STATE(5250), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [44838] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10195), 1,
      sym_identifier,
    STATE(1704), 1,
      sym_method_parameter,
    STATE(5038), 1,
      sym_attribute_specifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1723), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [44865] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6693), 1,
      anon_sym___attribute,
    ACTIONS(7528), 1,
      anon_sym___attribute__,
    STATE(4878), 2,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(6691), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [44886] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10197), 1,
      sym_identifier,
    ACTIONS(10199), 1,
      anon_sym_COLON,
    STATE(3958), 1,
      sym_enumerator_list,
    STATE(5068), 1,
      sym_attribute_specifier,
    STATE(5262), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [44915] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(4883), 2,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(6687), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [44936] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10064), 1,
      anon_sym_LBRACK,
    ACTIONS(10066), 1,
      anon_sym_AT,
    STATE(5353), 1,
      sym_gnu_asm_output_operand,
    STATE(6714), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44959] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10066), 1,
      anon_sym_AT,
    ACTIONS(10074), 1,
      anon_sym_LBRACK,
    STATE(5424), 1,
      sym_gnu_asm_input_operand,
    STATE(6011), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [44982] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(4878), 2,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(6698), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [45003] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10199), 1,
      anon_sym_COLON,
    ACTIONS(10201), 1,
      sym_identifier,
    STATE(3958), 1,
      sym_enumerator_list,
    STATE(5037), 1,
      sym_attribute_specifier,
    STATE(5180), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [45032] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10203), 1,
      sym_identifier,
    ACTIONS(10205), 1,
      anon_sym_COLON,
    STATE(4080), 1,
      sym_enumerator_list,
    STATE(5023), 1,
      sym_attribute_specifier,
    STATE(5168), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [45061] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    STATE(4878), 2,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(6687), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [45082] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(788), 1,
      sym_finally_clause,
    ACTIONS(2308), 2,
      anon_sym_else,
      anon_sym_while,
    ACTIONS(10207), 2,
      anon_sym_ATcatch,
      anon_sym___catch,
    ACTIONS(10209), 2,
      anon_sym_ATfinally,
      anon_sym___finally,
    STATE(4938), 2,
      sym_catch_clause,
      aux_sym_try_statement_repeat1,
  [45105] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10213), 1,
      anon_sym_COMMA,
    ACTIONS(10215), 1,
      anon_sym_RBRACE,
    STATE(3038), 1,
      sym_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(4939), 2,
      sym_preproc_ifdef_in_enumerator,
      aux_sym_enumerator_list_repeat1,
  [45129] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10219), 1,
      anon_sym___attribute,
    ACTIONS(10217), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [45145] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(8760), 1,
      anon_sym_LBRACE,
    ACTIONS(10221), 1,
      sym_identifier,
    STATE(3805), 1,
      sym_field_declaration_list,
    STATE(5138), 1,
      sym_attribute_specifier,
    STATE(5536), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [45171] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10223), 1,
      anon_sym_SEMI,
    STATE(4935), 1,
      sym_bitfield_clause,
    STATE(4936), 1,
      aux_sym_field_declaration_repeat1,
    STATE(6569), 1,
      sym_attribute_specifier,
  [45199] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10225), 1,
      anon_sym_SEMI,
    STATE(4901), 1,
      sym_bitfield_clause,
    STATE(4903), 1,
      aux_sym_field_declaration_repeat1,
    STATE(6187), 1,
      sym_attribute_specifier,
  [45227] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10227), 1,
      anon_sym_LPAREN2,
    ACTIONS(10229), 1,
      anon_sym_LBRACE,
    STATE(4997), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [45247] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10035), 1,
      sym_identifier,
    ACTIONS(10039), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(10233), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(10235), 1,
      aux_sym_preproc_elif_token1,
    STATE(4917), 1,
      aux_sym_preproc_ifdef_in_enumerator_repeat1,
    STATE(6625), 1,
      sym_enumerator,
    STATE(6265), 2,
      sym_preproc_else_in_enumerator,
      sym_preproc_elif_in_enumerator,
  [45273] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10237), 1,
      anon_sym_SEMI,
    STATE(4912), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5286), 1,
      sym_bitfield_clause,
    STATE(6112), 1,
      sym_attribute_specifier,
  [45301] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(6966), 1,
      anon_sym_LBRACE,
    ACTIONS(9326), 1,
      sym_identifier,
    STATE(2918), 1,
      sym_field_declaration_list,
    STATE(5125), 1,
      sym_attribute_specifier,
    STATE(5558), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [45327] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(1499), 1,
      sym_compound_statement,
    STATE(4909), 1,
      sym_attribute_specifier,
    STATE(5076), 1,
      sym_parameter_list,
  [45355] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10227), 1,
      anon_sym_LPAREN2,
    ACTIONS(10239), 1,
      anon_sym_LBRACE,
    STATE(4997), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [45375] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10243), 1,
      anon_sym___attribute,
    ACTIONS(10241), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [45391] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10245), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5167), 1,
      sym_bitfield_clause,
    STATE(6133), 1,
      sym_attribute_specifier,
  [45419] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10247), 1,
      anon_sym_SEMI,
    STATE(4913), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5174), 1,
      sym_bitfield_clause,
    STATE(6160), 1,
      sym_attribute_specifier,
  [45447] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10251), 1,
      anon_sym___attribute,
    ACTIONS(10249), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [45463] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10253), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5220), 1,
      sym_bitfield_clause,
    STATE(6223), 1,
      sym_attribute_specifier,
  [45491] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10257), 1,
      anon_sym___attribute,
    ACTIONS(10255), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [45507] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(1661), 1,
      sym_compound_statement,
    STATE(4948), 1,
      sym_attribute_specifier,
    STATE(5093), 1,
      sym_parameter_list,
  [45535] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10259), 1,
      anon_sym_COMMA,
    ACTIONS(10261), 1,
      anon_sym_RBRACE,
    STATE(3038), 1,
      sym_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(4920), 2,
      sym_preproc_ifdef_in_enumerator,
      aux_sym_enumerator_list_repeat1,
  [45559] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(8760), 1,
      anon_sym_LBRACE,
    ACTIONS(10263), 1,
      sym_identifier,
    STATE(3801), 1,
      sym_field_declaration_list,
    STATE(5137), 1,
      sym_attribute_specifier,
    STATE(5534), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [45585] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(1471), 1,
      sym_compound_statement,
    STATE(4923), 1,
      sym_attribute_specifier,
    STATE(5014), 1,
      sym_parameter_list,
  [45613] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(1471), 1,
      sym_compound_statement,
    STATE(5014), 1,
      sym_parameter_list,
    STATE(5621), 1,
      sym_attribute_specifier,
  [45641] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10265), 1,
      sym_identifier,
    ACTIONS(10270), 1,
      aux_sym_preproc_elif_token1,
    STATE(4910), 1,
      aux_sym_preproc_ifdef_in_enumerator_repeat1,
    STATE(6436), 1,
      sym_enumerator,
    ACTIONS(10268), 4,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
  [45663] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7089), 1,
      anon_sym_COMMA,
    ACTIONS(7093), 1,
      anon_sym_RBRACE,
    ACTIONS(10211), 1,
      sym_identifier,
    STATE(3038), 1,
      sym_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(4927), 2,
      sym_preproc_ifdef_in_enumerator,
      aux_sym_enumerator_list_repeat1,
  [45687] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10272), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5222), 1,
      sym_bitfield_clause,
    STATE(6346), 1,
      sym_attribute_specifier,
  [45715] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10274), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5225), 1,
      sym_bitfield_clause,
    STATE(6385), 1,
      sym_attribute_specifier,
  [45743] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10259), 1,
      anon_sym_COMMA,
    ACTIONS(10261), 1,
      anon_sym_RBRACE,
    STATE(3038), 1,
      sym_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(4921), 2,
      sym_preproc_ifdef_in_enumerator,
      aux_sym_enumerator_list_repeat1,
  [45767] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10227), 1,
      anon_sym_LPAREN2,
    ACTIONS(10276), 1,
      anon_sym_LBRACE,
    STATE(4997), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [45787] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10278), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5302), 1,
      sym_bitfield_clause,
    STATE(5940), 1,
      sym_attribute_specifier,
  [45815] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10035), 1,
      sym_identifier,
    ACTIONS(10039), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(10235), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(10280), 1,
      aux_sym_preproc_if_token2,
    STATE(5050), 1,
      aux_sym_preproc_ifdef_in_enumerator_repeat1,
    STATE(6625), 1,
      sym_enumerator,
    STATE(5910), 2,
      sym_preproc_else_in_enumerator,
      sym_preproc_elif_in_enumerator,
  [45841] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10284), 1,
      anon_sym___attribute,
    ACTIONS(10282), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [45857] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(1661), 1,
      sym_compound_statement,
    STATE(5093), 1,
      sym_parameter_list,
    STATE(5696), 1,
      sym_attribute_specifier,
  [45885] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10286), 1,
      anon_sym_COMMA,
    ACTIONS(10288), 1,
      anon_sym_RBRACE,
    STATE(3038), 1,
      sym_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(4921), 2,
      sym_preproc_ifdef_in_enumerator,
      aux_sym_enumerator_list_repeat1,
  [45909] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7508), 1,
      anon_sym_RBRACE,
    ACTIONS(10290), 1,
      sym_identifier,
    ACTIONS(10293), 1,
      anon_sym_COMMA,
    STATE(3038), 1,
      sym_enumerator,
    ACTIONS(10296), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(4921), 2,
      sym_preproc_ifdef_in_enumerator,
      aux_sym_enumerator_list_repeat1,
  [45933] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10227), 1,
      anon_sym_LPAREN2,
    ACTIONS(10299), 1,
      anon_sym_LBRACE,
    STATE(4997), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [45953] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(1501), 1,
      sym_compound_statement,
    STATE(5059), 1,
      sym_parameter_list,
    STATE(5662), 1,
      sym_attribute_specifier,
  [45981] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10227), 1,
      anon_sym_LPAREN2,
    ACTIONS(10301), 1,
      anon_sym_LBRACE,
    STATE(4997), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [46001] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10305), 1,
      anon_sym___attribute,
    ACTIONS(10303), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [46017] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10227), 1,
      anon_sym_LPAREN2,
    ACTIONS(10307), 1,
      anon_sym_LBRACE,
    STATE(4997), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [46037] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10213), 1,
      anon_sym_COMMA,
    ACTIONS(10215), 1,
      anon_sym_RBRACE,
    STATE(3038), 1,
      sym_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(4921), 2,
      sym_preproc_ifdef_in_enumerator,
      aux_sym_enumerator_list_repeat1,
  [46061] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10311), 1,
      anon_sym___attribute,
    ACTIONS(10309), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [46077] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10313), 1,
      anon_sym_SEMI,
    STATE(4942), 1,
      sym_bitfield_clause,
    STATE(4943), 1,
      aux_sym_field_declaration_repeat1,
    STATE(6654), 1,
      sym_attribute_specifier,
  [46105] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10227), 1,
      anon_sym_LPAREN2,
    ACTIONS(10315), 1,
      anon_sym_LBRACE,
    STATE(4997), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [46125] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10317), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5236), 1,
      sym_bitfield_clause,
    STATE(6201), 1,
      sym_attribute_specifier,
  [46153] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(6966), 1,
      anon_sym_LBRACE,
    ACTIONS(10319), 1,
      sym_identifier,
    STATE(2911), 1,
      sym_field_declaration_list,
    STATE(5140), 1,
      sym_attribute_specifier,
    STATE(5609), 1,
      sym_ms_declspec_modifier,
    ACTIONS(37), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [46179] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10321), 1,
      anon_sym_SEMI,
    STATE(4931), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5210), 1,
      sym_bitfield_clause,
    STATE(6604), 1,
      sym_attribute_specifier,
  [46207] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10323), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5228), 1,
      sym_bitfield_clause,
    STATE(6665), 1,
      sym_attribute_specifier,
  [46235] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10325), 1,
      anon_sym_SEMI,
    STATE(4937), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5238), 1,
      sym_bitfield_clause,
    STATE(6717), 1,
      sym_attribute_specifier,
  [46263] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10327), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5244), 1,
      sym_bitfield_clause,
    STATE(6737), 1,
      sym_attribute_specifier,
  [46291] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10329), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5267), 1,
      sym_bitfield_clause,
    STATE(6379), 1,
      sym_attribute_specifier,
  [46319] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10331), 2,
      anon_sym_ATcatch,
      anon_sym___catch,
    STATE(4938), 2,
      sym_catch_clause,
      aux_sym_try_statement_repeat1,
    ACTIONS(2320), 4,
      anon_sym_else,
      anon_sym_while,
      anon_sym_ATfinally,
      anon_sym___finally,
  [46337] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10334), 1,
      anon_sym_COMMA,
    ACTIONS(10336), 1,
      anon_sym_RBRACE,
    STATE(3038), 1,
      sym_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(4921), 2,
      sym_preproc_ifdef_in_enumerator,
      aux_sym_enumerator_list_repeat1,
  [46361] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10338), 1,
      anon_sym_SEMI,
    STATE(4947), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5301), 1,
      sym_bitfield_clause,
    STATE(6748), 1,
      sym_attribute_specifier,
  [46389] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10340), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5303), 1,
      sym_bitfield_clause,
    STATE(6752), 1,
      sym_attribute_specifier,
  [46417] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10342), 1,
      anon_sym_SEMI,
    STATE(4916), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5304), 1,
      sym_bitfield_clause,
    STATE(6756), 1,
      sym_attribute_specifier,
  [46445] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10344), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5306), 1,
      sym_bitfield_clause,
    STATE(6759), 1,
      sym_attribute_specifier,
  [46473] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7125), 1,
      anon_sym_COMMA,
    ACTIONS(7127), 1,
      anon_sym_RBRACE,
    ACTIONS(10211), 1,
      sym_identifier,
    STATE(3038), 1,
      sym_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(4914), 2,
      sym_preproc_ifdef_in_enumerator,
      aux_sym_enumerator_list_repeat1,
  [46497] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(1662), 1,
      sym_compound_statement,
    STATE(4919), 1,
      sym_attribute_specifier,
    STATE(5008), 1,
      sym_parameter_list,
  [46525] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10348), 1,
      anon_sym___attribute,
    ACTIONS(10346), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [46541] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(9843), 1,
      anon_sym_COLON,
    ACTIONS(10003), 1,
      anon_sym_COMMA,
    ACTIONS(10350), 1,
      anon_sym_SEMI,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    STATE(5311), 1,
      sym_bitfield_clause,
    STATE(6518), 1,
      sym_attribute_specifier,
  [46569] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(1652), 1,
      sym_compound_statement,
    STATE(5067), 1,
      sym_parameter_list,
    STATE(5711), 1,
      sym_attribute_specifier,
  [46597] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10352), 1,
      sym_identifier,
    ACTIONS(10354), 1,
      anon_sym_COMMA,
    STATE(2790), 1,
      sym_method_parameter,
    STATE(2744), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [46620] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10084), 1,
      anon_sym_AT,
    STATE(5742), 1,
      sym_string_literal,
    ACTIONS(10086), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [46637] = 8,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10356), 1,
      sym_identifier,
    ACTIONS(10358), 1,
      anon_sym_LPAREN2,
    ACTIONS(10360), 1,
      aux_sym_selector_expression_token1,
    ACTIONS(10362), 1,
      anon_sym_COLON2,
    STATE(5052), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5132), 1,
      aux_sym_selector_expression_repeat1,
    STATE(5754), 1,
      sym_method_identifier,
  [46662] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10227), 1,
      anon_sym_LPAREN2,
    STATE(4997), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [46679] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1140), 1,
      sym_finally_clause,
    ACTIONS(2966), 2,
      anon_sym_ATcatch,
      anon_sym___catch,
    ACTIONS(2970), 2,
      anon_sym_ATfinally,
      anon_sym___finally,
    STATE(1076), 2,
      sym_catch_clause,
      aux_sym_try_statement_repeat1,
  [46698] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10084), 1,
      anon_sym_AT,
    STATE(5783), 1,
      sym_string_literal,
    ACTIONS(10086), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [46715] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10066), 1,
      anon_sym_AT,
    STATE(6507), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [46732] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10364), 1,
      sym_identifier,
    STATE(2891), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [46753] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(964), 1,
      sym_finally_clause,
    ACTIONS(2868), 2,
      anon_sym_ATcatch,
      anon_sym___catch,
    ACTIONS(2872), 2,
      anon_sym_ATfinally,
      anon_sym___finally,
    STATE(462), 2,
      sym_catch_clause,
      aux_sym_try_statement_repeat1,
  [46772] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(414), 1,
      sym_finally_clause,
    ACTIONS(2338), 2,
      anon_sym_ATcatch,
      anon_sym___catch,
    ACTIONS(2342), 2,
      anon_sym_ATfinally,
      anon_sym___finally,
    STATE(182), 2,
      sym_catch_clause,
      aux_sym_try_statement_repeat1,
  [46791] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(870), 1,
      anon_sym_LBRACE,
    ACTIONS(872), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(10366), 1,
      anon_sym_LPAREN2,
    STATE(649), 1,
      sym_compound_statement,
    STATE(2906), 1,
      sym__old_style_parameter_list,
    STATE(4804), 1,
      sym_parameter_list,
  [46816] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10364), 1,
      sym_identifier,
    STATE(2883), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [46837] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10368), 1,
      anon_sym_LPAREN2,
    STATE(4963), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [46854] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(10366), 1,
      anon_sym_LPAREN2,
    STATE(1048), 1,
      sym_compound_statement,
    STATE(2906), 1,
      sym__old_style_parameter_list,
    STATE(4804), 1,
      sym_parameter_list,
  [46879] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10370), 1,
      anon_sym_LPAREN2,
    STATE(4986), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [46896] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(944), 1,
      anon_sym_LBRACE,
    ACTIONS(948), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(10366), 1,
      anon_sym_LPAREN2,
    STATE(878), 1,
      sym_compound_statement,
    STATE(2906), 1,
      sym__old_style_parameter_list,
    STATE(4804), 1,
      sym_parameter_list,
  [46921] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10364), 1,
      sym_identifier,
    STATE(2886), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [46942] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10374), 1,
      anon_sym___attribute,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4289), 1,
      sym_parameter_list,
    ACTIONS(10372), 3,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [46963] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(685), 1,
      sym_finally_clause,
    ACTIONS(2400), 2,
      anon_sym_ATcatch,
      anon_sym___catch,
    ACTIONS(2404), 2,
      anon_sym_ATfinally,
      anon_sym___finally,
    STATE(198), 2,
      sym_catch_clause,
      aux_sym_try_statement_repeat1,
  [46982] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10378), 1,
      anon_sym_LPAREN2,
    STATE(4972), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [46999] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10364), 1,
      sym_identifier,
    STATE(2896), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [47020] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10084), 1,
      anon_sym_AT,
    STATE(5854), 1,
      sym_string_literal,
    ACTIONS(10086), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [47037] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(303), 1,
      anon_sym_LBRACE,
    ACTIONS(305), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(10366), 1,
      anon_sym_LPAREN2,
    STATE(445), 1,
      sym_compound_statement,
    STATE(2906), 1,
      sym__old_style_parameter_list,
    STATE(4804), 1,
      sym_parameter_list,
  [47062] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10380), 1,
      anon_sym_LPAREN2,
    STATE(4986), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [47079] = 8,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10358), 1,
      anon_sym_LPAREN2,
    ACTIONS(10362), 1,
      anon_sym_COLON2,
    ACTIONS(10382), 1,
      sym_identifier,
    ACTIONS(10384), 1,
      aux_sym_selector_expression_token1,
    STATE(5052), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5132), 1,
      aux_sym_selector_expression_repeat1,
    STATE(5714), 1,
      sym_method_identifier,
  [47104] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(9999), 1,
      anon_sym___attribute,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4289), 1,
      sym_parameter_list,
    ACTIONS(9997), 3,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [47125] = 8,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10358), 1,
      anon_sym_LPAREN2,
    ACTIONS(10362), 1,
      anon_sym_COLON2,
    ACTIONS(10386), 1,
      sym_identifier,
    ACTIONS(10388), 1,
      aux_sym_selector_expression_token1,
    STATE(5052), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5132), 1,
      aux_sym_selector_expression_repeat1,
    STATE(5736), 1,
      sym_method_identifier,
  [47150] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10390), 1,
      anon_sym_LBRACK,
    ACTIONS(10393), 1,
      anon_sym_EQ,
    ACTIONS(10395), 1,
      anon_sym_DOT,
    STATE(4976), 4,
      sym_subscript_designator,
      sym_subscript_range_designator,
      sym_field_designator,
      aux_sym_initializer_pair_repeat1,
  [47169] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(253), 1,
      sym_finally_clause,
    ACTIONS(2310), 2,
      anon_sym_ATcatch,
      anon_sym___catch,
    ACTIONS(2314), 2,
      anon_sym_ATfinally,
      anon_sym___finally,
    STATE(179), 2,
      sym_catch_clause,
      aux_sym_try_statement_repeat1,
  [47188] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(10398), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
      anon_sym_COLON,
  [47207] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(10400), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
      anon_sym_COLON,
  [47226] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10402), 1,
      anon_sym_COMMA,
    ACTIONS(10404), 1,
      anon_sym_RBRACE,
    STATE(2923), 1,
      sym_enumerator,
    STATE(4944), 1,
      sym_preproc_ifdef_in_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
  [47249] = 8,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10358), 1,
      anon_sym_LPAREN2,
    ACTIONS(10362), 1,
      anon_sym_COLON2,
    ACTIONS(10406), 1,
      sym_identifier,
    ACTIONS(10408), 1,
      aux_sym_selector_expression_token1,
    STATE(5052), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5132), 1,
      aux_sym_selector_expression_repeat1,
    STATE(5726), 1,
      sym_method_identifier,
  [47274] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(685), 1,
      sym_finally_clause,
    ACTIONS(10207), 2,
      anon_sym_ATcatch,
      anon_sym___catch,
    ACTIONS(10209), 2,
      anon_sym_ATfinally,
      anon_sym___finally,
    STATE(4887), 2,
      sym_catch_clause,
      aux_sym_try_statement_repeat1,
  [47293] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(586), 1,
      sym_finally_clause,
    ACTIONS(2372), 2,
      anon_sym_ATcatch,
      anon_sym___catch,
    ACTIONS(2376), 2,
      anon_sym_ATfinally,
      anon_sym___finally,
    STATE(189), 2,
      sym_catch_clause,
      aux_sym_try_statement_repeat1,
  [47312] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10364), 1,
      sym_identifier,
    STATE(2877), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [47333] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4517), 1,
      anon_sym_AT,
    STATE(1916), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [47350] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10410), 1,
      anon_sym_LPAREN2,
    STATE(4986), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10412), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [47367] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10364), 1,
      sym_identifier,
    STATE(2879), 2,
      sym_method_parameter,
      aux_sym_method_declaration_repeat1,
    STATE(3221), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [47388] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    ACTIONS(10415), 1,
      anon_sym___attribute,
    STATE(4289), 1,
      sym_parameter_list,
    ACTIONS(10398), 3,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [47409] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    ACTIONS(10417), 1,
      anon_sym___attribute,
    STATE(4289), 1,
      sym_parameter_list,
    ACTIONS(10400), 3,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [47430] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    ACTIONS(10421), 1,
      anon_sym___attribute,
    STATE(4289), 1,
      sym_parameter_list,
    ACTIONS(10419), 3,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [47451] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    ACTIONS(10425), 1,
      anon_sym___attribute,
    STATE(4289), 1,
      sym_parameter_list,
    ACTIONS(10423), 3,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [47472] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10427), 1,
      anon_sym_LPAREN2,
    STATE(4998), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [47489] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10066), 1,
      anon_sym_AT,
    STATE(5931), 1,
      sym_string_literal,
    ACTIONS(111), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [47506] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(739), 1,
      sym_finally_clause,
    ACTIONS(2386), 2,
      anon_sym_ATcatch,
      anon_sym___catch,
    ACTIONS(2390), 2,
      anon_sym_ATfinally,
      anon_sym___finally,
    STATE(196), 2,
      sym_catch_clause,
      aux_sym_try_statement_repeat1,
  [47525] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10429), 1,
      anon_sym_LBRACK,
    ACTIONS(10431), 1,
      anon_sym_EQ,
    ACTIONS(10433), 1,
      anon_sym_DOT,
    STATE(4976), 4,
      sym_subscript_designator,
      sym_subscript_range_designator,
      sym_field_designator,
      aux_sym_initializer_pair_repeat1,
  [47544] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(181), 1,
      anon_sym_LBRACE,
    ACTIONS(183), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(10366), 1,
      anon_sym_LPAREN2,
    STATE(283), 1,
      sym_compound_statement,
    STATE(2906), 1,
      sym__old_style_parameter_list,
    STATE(4804), 1,
      sym_parameter_list,
  [47569] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10435), 1,
      anon_sym_LPAREN2,
    STATE(4986), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [47586] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10437), 1,
      anon_sym_LPAREN2,
    STATE(4986), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(10231), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [47603] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(10419), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
      anon_sym_COLON,
  [47622] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(10423), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
      anon_sym_COLON,
  [47641] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10084), 1,
      anon_sym_AT,
    STATE(5873), 1,
      sym_string_literal,
    ACTIONS(10086), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [47658] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10439), 1,
      anon_sym_COMMA,
    ACTIONS(10441), 1,
      anon_sym_RBRACE,
    STATE(2913), 1,
      sym_enumerator,
    STATE(4911), 1,
      sym_preproc_ifdef_in_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
  [47681] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10084), 1,
      anon_sym_AT,
    STATE(5730), 1,
      sym_string_literal,
    ACTIONS(10086), 5,
      anon_sym_DQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
  [47698] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2315), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [47720] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(944), 1,
      anon_sym_LBRACE,
    ACTIONS(948), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(878), 1,
      sym_compound_statement,
    STATE(4804), 1,
      sym_parameter_list,
  [47742] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(2263), 1,
      sym_compound_statement,
    STATE(4804), 1,
      sym_parameter_list,
  [47764] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2262), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [47786] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(1661), 1,
      sym_compound_statement,
    STATE(5696), 1,
      sym_attribute_specifier,
  [47808] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2309), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [47830] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10445), 1,
      sym_identifier,
    STATE(2309), 1,
      sym_compound_statement,
    STATE(6496), 1,
      sym_method_type,
  [47852] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2106), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [47874] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(4804), 1,
      sym_parameter_list,
    ACTIONS(6746), 3,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [47892] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(4804), 1,
      sym_parameter_list,
    ACTIONS(6750), 3,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [47910] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    STATE(1501), 1,
      sym_compound_statement,
    STATE(5662), 1,
      sym_attribute_specifier,
  [47932] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10445), 1,
      sym_identifier,
    STATE(2186), 1,
      sym_compound_statement,
    STATE(6496), 1,
      sym_method_type,
  [47954] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2312), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [47976] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2107), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [47998] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2283), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [48020] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(2142), 1,
      sym_compound_statement,
    STATE(4804), 1,
      sym_parameter_list,
  [48042] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(10372), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
  [48060] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10057), 1,
      anon_sym___attribute,
    ACTIONS(10059), 1,
      anon_sym_COLON,
    STATE(5139), 1,
      sym_bitfield_clause,
    ACTIONS(10055), 3,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [48078] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10447), 1,
      sym_identifier,
    ACTIONS(10449), 1,
      anon_sym_COLON,
    STATE(3178), 1,
      sym_enumerator_list,
    STATE(5283), 1,
      sym_ms_declspec_modifier,
  [48100] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10451), 1,
      sym_identifier,
    ACTIONS(10453), 1,
      anon_sym_COLON,
    STATE(4091), 1,
      sym_enumerator_list,
    STATE(5185), 1,
      sym_ms_declspec_modifier,
  [48122] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(4804), 1,
      sym_parameter_list,
    ACTIONS(6738), 3,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [48140] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10445), 1,
      sym_identifier,
    STATE(2268), 1,
      sym_compound_statement,
    STATE(6496), 1,
      sym_method_type,
  [48162] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10215), 1,
      anon_sym_RBRACE,
    STATE(3022), 1,
      sym_enumerator,
    STATE(5121), 1,
      sym_preproc_ifdef_in_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
  [48182] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10445), 1,
      sym_identifier,
    STATE(2334), 1,
      sym_compound_statement,
    STATE(6496), 1,
      sym_method_type,
  [48204] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10171), 1,
      anon_sym_RPAREN,
    STATE(4294), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [48224] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(870), 1,
      anon_sym_LBRACE,
    ACTIONS(872), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(649), 1,
      sym_compound_statement,
    STATE(4804), 1,
      sym_parameter_list,
  [48246] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10455), 1,
      anon_sym_RPAREN,
    STATE(4294), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [48266] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2334), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [48288] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(4804), 1,
      sym_parameter_list,
    ACTIONS(6742), 3,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [48306] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(9997), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT,
  [48324] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10445), 1,
      sym_identifier,
    STATE(2262), 1,
      sym_compound_statement,
    STATE(6496), 1,
      sym_method_type,
  [48346] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10457), 1,
      sym_identifier,
    ACTIONS(10461), 1,
      anon_sym_COLON2,
    STATE(5113), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5229), 1,
      aux_sym_method_identifier_repeat2,
    ACTIONS(10459), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [48366] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10463), 1,
      anon_sym_RBRACE,
    STATE(3022), 1,
      sym_enumerator,
    STATE(5121), 1,
      sym_preproc_ifdef_in_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
  [48386] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10465), 1,
      sym_identifier,
    ACTIONS(10467), 1,
      anon_sym_COLON,
    STATE(3868), 1,
      sym_enumerator_list,
    STATE(5221), 1,
      sym_ms_declspec_modifier,
  [48408] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10469), 1,
      sym_identifier,
    STATE(1684), 1,
      sym_method_parameter,
    STATE(1715), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [48428] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2268), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [48450] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10471), 1,
      sym_identifier,
    ACTIONS(10473), 1,
      anon_sym_COLON,
    STATE(4161), 1,
      sym_enumerator_list,
    STATE(5261), 1,
      sym_ms_declspec_modifier,
  [48472] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(1048), 1,
      sym_compound_statement,
    STATE(4804), 1,
      sym_parameter_list,
  [48494] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10261), 1,
      anon_sym_RBRACE,
    STATE(3022), 1,
      sym_enumerator,
    STATE(5121), 1,
      sym_preproc_ifdef_in_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
  [48514] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(10475), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [48532] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2297), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [48554] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10477), 1,
      sym_identifier,
    STATE(1691), 1,
      sym_method_parameter,
    STATE(1720), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [48574] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10445), 1,
      sym_identifier,
    STATE(2163), 1,
      sym_compound_statement,
    STATE(6496), 1,
      sym_method_type,
  [48596] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(2141), 1,
      sym_compound_statement,
    STATE(4804), 1,
      sym_parameter_list,
  [48618] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(10479), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [48636] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2163), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [48658] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10265), 1,
      sym_identifier,
    STATE(5050), 1,
      aux_sym_preproc_ifdef_in_enumerator_repeat1,
    STATE(6625), 1,
      sym_enumerator,
    ACTIONS(10268), 3,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
  [48676] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10481), 1,
      anon_sym_COMMA,
    ACTIONS(10486), 1,
      anon_sym___attribute,
    STATE(5051), 1,
      aux_sym_field_declaration_repeat1,
    ACTIONS(10484), 3,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [48694] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10457), 1,
      sym_identifier,
    ACTIONS(10461), 1,
      anon_sym_COLON2,
    STATE(5113), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5256), 1,
      aux_sym_method_identifier_repeat2,
    ACTIONS(10488), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [48714] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10449), 1,
      anon_sym_COLON,
    ACTIONS(10490), 1,
      sym_identifier,
    STATE(3178), 1,
      sym_enumerator_list,
    STATE(5212), 1,
      sym_ms_declspec_modifier,
  [48736] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2252), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [48758] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10445), 1,
      sym_identifier,
    STATE(2297), 1,
      sym_compound_statement,
    STATE(6496), 1,
      sym_method_type,
  [48780] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2260), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [48802] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2186), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [48824] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(4804), 1,
      sym_parameter_list,
    ACTIONS(6720), 3,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [48842] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    STATE(1483), 1,
      sym_compound_statement,
    STATE(5698), 1,
      sym_attribute_specifier,
  [48864] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2345), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [48886] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10492), 1,
      sym_identifier,
    STATE(1675), 1,
      sym_method_parameter,
    STATE(1716), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [48906] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2271), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [48928] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10120), 1,
      anon_sym_RPAREN,
    STATE(4294), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [48948] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10494), 1,
      sym_identifier,
    STATE(1704), 1,
      sym_method_parameter,
    STATE(1723), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [48968] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10128), 1,
      anon_sym_RPAREN,
    STATE(4294), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [48988] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10336), 1,
      anon_sym_RBRACE,
    STATE(3022), 1,
      sym_enumerator,
    STATE(5121), 1,
      sym_preproc_ifdef_in_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
  [49008] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(1640), 1,
      sym_compound_statement,
    STATE(5709), 1,
      sym_attribute_specifier,
  [49030] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10467), 1,
      anon_sym_COLON,
    ACTIONS(10496), 1,
      sym_identifier,
    STATE(3868), 1,
      sym_enumerator_list,
    STATE(5215), 1,
      sym_ms_declspec_modifier,
  [49052] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(10498), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [49070] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    STATE(1499), 1,
      sym_compound_statement,
    STATE(5571), 1,
      sym_attribute_specifier,
  [49092] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10445), 1,
      sym_identifier,
    STATE(2252), 1,
      sym_compound_statement,
    STATE(6496), 1,
      sym_method_type,
  [49114] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2308), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [49136] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(10500), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [49154] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10270), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(10268), 5,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_identifier,
  [49168] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2101), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [49190] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    STATE(1471), 1,
      sym_compound_statement,
    STATE(5621), 1,
      sym_attribute_specifier,
  [49212] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(1662), 1,
      sym_compound_statement,
    STATE(5606), 1,
      sym_attribute_specifier,
  [49234] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10502), 1,
      anon_sym_RPAREN,
    STATE(4294), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [49254] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10288), 1,
      anon_sym_RBRACE,
    STATE(3022), 1,
      sym_enumerator,
    STATE(5121), 1,
      sym_preproc_ifdef_in_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
  [49274] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(181), 1,
      anon_sym_LBRACE,
    ACTIONS(183), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(283), 1,
      sym_compound_statement,
    STATE(4804), 1,
      sym_parameter_list,
  [49296] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2277), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [49318] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10163), 1,
      anon_sym_RPAREN,
    STATE(4294), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [49338] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2140), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [49360] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10445), 1,
      sym_identifier,
    STATE(2140), 1,
      sym_compound_statement,
    STATE(6496), 1,
      sym_method_type,
  [49382] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(4804), 1,
      sym_parameter_list,
    ACTIONS(6724), 3,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_ATautoreleasepool,
  [49400] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    ACTIONS(10504), 1,
      anon_sym_RBRACE,
    STATE(3022), 1,
      sym_enumerator,
    STATE(5121), 1,
      sym_preproc_ifdef_in_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
  [49420] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(2104), 1,
      sym_compound_statement,
    STATE(6560), 1,
      sym_method_type,
  [49442] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10167), 1,
      anon_sym_RPAREN,
    STATE(4294), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [49462] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10007), 1,
      anon_sym_LBRACK,
    ACTIONS(10110), 1,
      anon_sym_RPAREN,
    STATE(4294), 1,
      sym_parameter_list,
    STATE(4863), 2,
      sym_attribute_declaration,
      aux_sym_attributed_field_declarator_repeat1,
  [49482] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(303), 1,
      anon_sym_LBRACE,
    ACTIONS(305), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    STATE(445), 1,
      sym_compound_statement,
    STATE(4804), 1,
      sym_parameter_list,
  [49504] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10473), 1,
      anon_sym_COLON,
    ACTIONS(10506), 1,
      sym_identifier,
    STATE(4161), 1,
      sym_enumerator_list,
    STATE(5275), 1,
      sym_ms_declspec_modifier,
  [49526] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10508), 1,
      sym_identifier,
    STATE(1686), 1,
      sym_method_parameter,
    STATE(1722), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [49546] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(1652), 1,
      sym_compound_statement,
    STATE(5711), 1,
      sym_attribute_specifier,
  [49568] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(10001), 1,
      anon_sym_SEMI,
    ACTIONS(10510), 1,
      sym_identifier,
    STATE(1703), 1,
      sym_method_parameter,
    STATE(1718), 2,
      sym_keyword_declarator,
      aux_sym_keyword_selector_repeat1,
  [49588] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3008), 1,
      anon_sym_LT,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(10514), 1,
      anon_sym_SEMI,
    STATE(5434), 1,
      sym_parameterized_arguments,
    STATE(5439), 1,
      aux_sym_class_declaration_repeat1,
  [49607] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5666), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [49618] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10516), 1,
      anon_sym_ATend,
    ACTIONS(10518), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49633] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5682), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [49644] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3008), 1,
      anon_sym_LT,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(10521), 1,
      anon_sym_SEMI,
    STATE(5640), 1,
      sym_parameterized_arguments,
    STATE(5641), 1,
      aux_sym_class_declaration_repeat1,
  [49663] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5654), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [49674] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2828), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [49685] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2962), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [49696] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10523), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49711] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10525), 5,
      anon_sym_COMMA,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      anon_sym_RBRACE,
      sym_identifier,
  [49722] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5678), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [49733] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3072), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49748] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2592), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [49759] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10211), 1,
      sym_identifier,
    STATE(3022), 1,
      sym_enumerator,
    STATE(5121), 1,
      sym_preproc_ifdef_in_enumerator,
    ACTIONS(7091), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
  [49776] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5670), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [49787] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3329), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49802] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3309), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49817] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7613), 1,
      sym_identifier,
    ACTIONS(7617), 4,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_id,
  [49830] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10529), 1,
      anon_sym_COLON2,
    STATE(5113), 1,
      aux_sym_method_identifier_repeat1,
    ACTIONS(10527), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_identifier,
  [49845] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10532), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49860] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3331), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49875] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3341), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49890] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10534), 5,
      anon_sym_COMMA,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      anon_sym_RBRACE,
      sym_identifier,
  [49901] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9905), 1,
      anon_sym_COLON2,
    ACTIONS(10536), 1,
      sym_identifier,
    ACTIONS(10538), 1,
      anon_sym_RPAREN,
    STATE(5052), 1,
      aux_sym_method_identifier_repeat1,
    STATE(5936), 1,
      sym_method_identifier,
  [49920] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10540), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49935] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3008), 1,
      anon_sym_LT,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(10542), 1,
      anon_sym_SEMI,
    STATE(5528), 1,
      sym_parameterized_arguments,
    STATE(5529), 1,
      aux_sym_class_declaration_repeat1,
  [49954] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7508), 5,
      anon_sym_COMMA,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      anon_sym_RBRACE,
      sym_identifier,
  [49965] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10544), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49980] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3311), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [49995] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10546), 5,
      anon_sym_COMMA,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      anon_sym_RBRACE,
      sym_identifier,
  [50006] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(6966), 1,
      anon_sym_LBRACE,
    ACTIONS(10548), 1,
      sym_identifier,
    STATE(2922), 1,
      sym_field_declaration_list,
    STATE(5411), 1,
      sym_ms_declspec_modifier,
  [50025] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10550), 1,
      anon_sym_COMMA,
    ACTIONS(10552), 1,
      anon_sym_GT,
    ACTIONS(10554), 1,
      anon_sym_COLON,
    STATE(5247), 1,
      aux_sym_parameterized_arguments_repeat1,
    STATE(5410), 1,
      aux_sym_parameterized_arguments_repeat2,
  [50044] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10550), 1,
      anon_sym_COMMA,
    ACTIONS(10556), 1,
      anon_sym_GT,
    ACTIONS(10558), 1,
      anon_sym_COLON,
    STATE(5148), 1,
      aux_sym_parameterized_arguments_repeat1,
    STATE(5419), 1,
      aux_sym_parameterized_arguments_repeat2,
  [50063] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3080), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [50078] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3333), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [50093] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10560), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [50108] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [50123] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10564), 1,
      anon_sym_LPAREN2,
    STATE(5132), 1,
      aux_sym_selector_expression_repeat1,
    ACTIONS(10562), 3,
      sym_identifier,
      aux_sym_selector_expression_token1,
      anon_sym_COLON2,
  [50138] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2712), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [50149] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3297), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [50164] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2914), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [50175] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3066), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [50190] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(8760), 1,
      anon_sym_LBRACE,
    ACTIONS(10567), 1,
      sym_identifier,
    STATE(3795), 1,
      sym_field_declaration_list,
    STATE(5582), 1,
      sym_ms_declspec_modifier,
  [50209] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(8760), 1,
      anon_sym_LBRACE,
    ACTIONS(10569), 1,
      sym_identifier,
    STATE(3798), 1,
      sym_field_declaration_list,
    STATE(5583), 1,
      sym_ms_declspec_modifier,
  [50228] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10573), 1,
      anon_sym___attribute,
    ACTIONS(10571), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [50241] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym___declspec,
    ACTIONS(6966), 1,
      anon_sym_LBRACE,
    ACTIONS(10575), 1,
      sym_identifier,
    STATE(2924), 1,
      sym_field_declaration_list,
    STATE(5433), 1,
      sym_ms_declspec_modifier,
  [50260] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3295), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [50275] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10461), 1,
      anon_sym_COLON2,
    STATE(5113), 1,
      aux_sym_method_identifier_repeat1,
    ACTIONS(10577), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_identifier,
  [50290] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10579), 1,
      sym_identifier,
    ACTIONS(10581), 4,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      anon_sym_id,
  [50303] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9905), 1,
      anon_sym_COLON2,
    ACTIONS(10536), 1,
      sym_identifier,
    ACTIONS(10583), 1,
      anon_sym_RPAREN,
    STATE(5052), 1,
      aux_sym_method_identifier_repeat1,
    STATE(6495), 1,
      sym_method_identifier,
  [50322] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10585), 5,
      anon_sym_LPAREN2,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [50333] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9403), 1,
      anon_sym_COMMA,
    STATE(5151), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(10587), 3,
      anon_sym_RBRACK,
      anon_sym_COLON,
      sym_identifier,
  [50348] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2958), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [50359] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10550), 1,
      anon_sym_COMMA,
    ACTIONS(10589), 1,
      anon_sym_GT,
    ACTIONS(10591), 1,
      anon_sym_COLON,
    STATE(5247), 1,
      aux_sym_parameterized_arguments_repeat1,
    STATE(5649), 1,
      aux_sym_parameterized_arguments_repeat2,
  [50378] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3008), 1,
      anon_sym_LT,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(10593), 1,
      anon_sym_SEMI,
    STATE(5632), 1,
      sym_parameterized_arguments,
    STATE(5633), 1,
      aux_sym_class_declaration_repeat1,
  [50397] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3104), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [50412] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10595), 1,
      anon_sym_COMMA,
    STATE(5151), 1,
      aux_sym_attribute_repeat1,
    ACTIONS(9408), 3,
      anon_sym_RBRACK,
      anon_sym_COLON,
      sym_identifier,
  [50427] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5646), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [50438] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3315), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [50453] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    STATE(4270), 1,
      sym_parameter_list,
    ACTIONS(7590), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [50470] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3008), 1,
      anon_sym_LT,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(10598), 1,
      anon_sym_SEMI,
    STATE(5657), 1,
      sym_parameterized_arguments,
    STATE(5658), 1,
      aux_sym_class_declaration_repeat1,
  [50489] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10600), 5,
      anon_sym_COMMA,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      anon_sym_RBRACE,
      sym_identifier,
  [50500] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5662), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [50511] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3323), 1,
      anon_sym_ATend,
    ACTIONS(3054), 2,
      anon_sym_AToptional,
      anon_sym_ATrequired,
    STATE(5097), 2,
      sym_qualified_protocol_interface_declaration,
      aux_sym_protocol_declaration_repeat1,
  [50526] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10602), 5,
      anon_sym_COMMA,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      anon_sym_RBRACE,
      sym_identifier,
  [50537] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9905), 1,
      anon_sym_COLON2,
    ACTIONS(10536), 1,
      sym_identifier,
    ACTIONS(10604), 1,
      anon_sym_RPAREN,
    STATE(5052), 1,
      aux_sym_method_identifier_repeat1,
    STATE(6311), 1,
      sym_method_identifier,
  [50556] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5658), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [50567] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5674), 5,
      anon_sym_const,
      anon_sym_struct,
      anon_sym_in,
      sym_identifier,
      anon_sym_id,
  [50578] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(10608), 1,
      anon_sym_SEMI,
    ACTIONS(10610), 1,
      anon_sym_EQ,
    STATE(5374), 1,
      aux_sym_property_implementation_repeat1,
  [50594] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8089), 1,
      anon_sym_LBRACK,
    ACTIONS(8143), 1,
      anon_sym_RPAREN,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(4311), 1,
      sym_parameter_list,
  [50610] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10612), 1,
      anon_sym_LPAREN2,
    ACTIONS(10614), 1,
      anon_sym_LBRACE,
    ACTIONS(10616), 1,
      anon_sym_ATautoreleasepool,
    STATE(192), 1,
      sym_compound_statement,
  [50626] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10618), 1,
      anon_sym_DQUOTE,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [50642] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10624), 1,
      anon_sym_SEMI,
    STATE(6374), 1,
      sym_attribute_specifier,
  [50658] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10453), 1,
      anon_sym_COLON,
    ACTIONS(10626), 1,
      sym_identifier,
    STATE(4091), 1,
      sym_enumerator_list,
  [50674] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10628), 1,
      anon_sym_COLON,
    STATE(5169), 1,
      aux_sym_message_expression_repeat1,
    ACTIONS(9401), 2,
      anon_sym_RBRACK,
      sym_identifier,
  [50688] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9462), 1,
      anon_sym_COMMA,
    ACTIONS(10631), 1,
      anon_sym_RPAREN,
    ACTIONS(10633), 1,
      anon_sym_EQ,
    STATE(5532), 1,
      aux_sym_availability_repeat1,
  [50704] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10635), 1,
      anon_sym_COMMA,
    ACTIONS(10637), 1,
      anon_sym_RPAREN,
    STATE(5619), 1,
      aux_sym_parameter_list_repeat1,
    STATE(5701), 1,
      aux_sym__old_style_parameter_list_repeat1,
  [50720] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(10639), 1,
      anon_sym_RPAREN,
    STATE(4804), 1,
      sym_parameter_list,
  [50736] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(10641), 1,
      anon_sym_RPAREN,
    STATE(4804), 1,
      sym_parameter_list,
  [50752] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10643), 1,
      anon_sym_SEMI,
    STATE(6383), 1,
      sym_attribute_specifier,
  [50768] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10645), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [50784] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(10647), 1,
      anon_sym_SEMI,
    ACTIONS(10649), 1,
      anon_sym_EQ,
    STATE(5539), 1,
      aux_sym_property_implementation_repeat1,
  [50800] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10651), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [50816] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10653), 1,
      anon_sym_COMMA,
    ACTIONS(10656), 1,
      anon_sym_GT,
    ACTIONS(10658), 1,
      anon_sym_COLON,
    STATE(5247), 1,
      aux_sym_parameterized_arguments_repeat1,
  [50832] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10660), 1,
      anon_sym_DQUOTE,
    ACTIONS(10662), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10664), 1,
      sym_escape_sequence,
    STATE(5298), 1,
      aux_sym_string_literal_repeat1,
  [50848] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10467), 1,
      anon_sym_COLON,
    ACTIONS(10666), 1,
      sym_identifier,
    STATE(3868), 1,
      sym_enumerator_list,
  [50864] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10019), 1,
      anon_sym_COLON2,
    ACTIONS(10668), 1,
      anon_sym_RPAREN,
    STATE(1622), 1,
      aux_sym_selector_expression_repeat2,
    STATE(5035), 1,
      aux_sym_method_identifier_repeat1,
  [50880] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10670), 1,
      anon_sym_COMMA,
    STATE(5271), 1,
      aux_sym_gnu_asm_clobber_list_repeat1,
    ACTIONS(10672), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [50894] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    ACTIONS(10674), 1,
      anon_sym_RPAREN,
    STATE(4270), 1,
      sym_parameter_list,
  [50910] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8089), 1,
      anon_sym_LBRACK,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10676), 1,
      anon_sym_RPAREN,
    STATE(4311), 1,
      sym_parameter_list,
  [50926] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10678), 1,
      sym_identifier,
    ACTIONS(10680), 1,
      anon_sym_COLON,
    STATE(4074), 1,
      sym_enumerator_list,
  [50942] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10682), 1,
      anon_sym_COMMA,
    ACTIONS(10685), 1,
      anon_sym_GT,
    ACTIONS(10687), 1,
      anon_sym_COLON,
    STATE(5247), 1,
      aux_sym_parameterized_arguments_repeat1,
  [50958] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8089), 1,
      anon_sym_LBRACK,
    ACTIONS(8095), 1,
      anon_sym_RPAREN,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(4311), 1,
      sym_parameter_list,
  [50974] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10689), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(10691), 1,
      anon_sym_COMMA,
    ACTIONS(10693), 1,
      anon_sym_RPAREN,
    STATE(5293), 1,
      aux_sym_preproc_params_repeat1,
  [50990] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10695), 1,
      anon_sym_COMMA,
    ACTIONS(10698), 1,
      anon_sym_GT,
    ACTIONS(10700), 1,
      anon_sym_COLON,
    STATE(5178), 1,
      aux_sym_parameterized_arguments_repeat1,
  [51006] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10702), 1,
      anon_sym_LPAREN2,
    ACTIONS(10704), 1,
      anon_sym_LBRACE,
    ACTIONS(10706), 1,
      anon_sym_ATautoreleasepool,
    STATE(723), 1,
      sym_compound_statement,
  [51022] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10708), 1,
      anon_sym_DQUOTE,
    ACTIONS(10710), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10712), 1,
      sym_escape_sequence,
    STATE(5219), 1,
      aux_sym_string_literal_repeat1,
  [51038] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(10714), 1,
      anon_sym_SEMI,
    ACTIONS(10716), 1,
      anon_sym_EQ,
    STATE(5327), 1,
      aux_sym_property_implementation_repeat1,
  [51054] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10718), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [51070] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3008), 1,
      anon_sym_LT,
    STATE(5750), 1,
      sym_parameterized_arguments,
    ACTIONS(10720), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [51084] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10722), 1,
      anon_sym_DQUOTE,
    ACTIONS(10724), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10726), 1,
      sym_escape_sequence,
    STATE(5218), 1,
      aux_sym_string_literal_repeat1,
  [51100] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10728), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [51116] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(10730), 1,
      anon_sym_RPAREN,
    STATE(4804), 1,
      sym_parameter_list,
  [51132] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10732), 1,
      anon_sym_SEMI,
    STATE(6658), 1,
      sym_attribute_specifier,
  [51148] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10734), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(10736), 1,
      anon_sym_LPAREN,
    ACTIONS(10738), 1,
      sym_preproc_arg,
    STATE(5766), 1,
      sym_preproc_params,
  [51164] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10740), 1,
      anon_sym_COMMA,
    ACTIONS(10744), 1,
      anon_sym_GT,
    ACTIONS(10747), 1,
      anon_sym_COLON,
    STATE(5178), 1,
      aux_sym_parameterized_arguments_repeat1,
  [51180] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10752), 1,
      anon_sym_COLON,
    STATE(5169), 1,
      aux_sym_message_expression_repeat1,
    ACTIONS(10750), 2,
      anon_sym_RBRACK,
      sym_identifier,
  [51194] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10754), 1,
      anon_sym_SQUOTE,
    STATE(5232), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(10756), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [51208] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8089), 1,
      anon_sym_LBRACK,
    ACTIONS(8155), 1,
      anon_sym_RPAREN,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(4311), 1,
      sym_parameter_list,
  [51224] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10635), 1,
      anon_sym_COMMA,
    ACTIONS(10758), 1,
      anon_sym_RPAREN,
    STATE(5687), 1,
      aux_sym_parameter_list_repeat1,
    STATE(5701), 1,
      aux_sym__old_style_parameter_list_repeat1,
  [51240] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8085), 1,
      anon_sym_RPAREN,
    ACTIONS(8089), 1,
      anon_sym_LBRACK,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(4311), 1,
      sym_parameter_list,
  [51256] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(10760), 1,
      anon_sym_RPAREN,
    STATE(4804), 1,
      sym_parameter_list,
  [51272] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10449), 1,
      anon_sym_COLON,
    ACTIONS(10762), 1,
      sym_identifier,
    STATE(3178), 1,
      sym_enumerator_list,
  [51288] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10752), 1,
      anon_sym_COLON,
    STATE(5201), 1,
      aux_sym_message_expression_repeat1,
    ACTIONS(10764), 2,
      anon_sym_RBRACK,
      sym_identifier,
  [51302] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10766), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [51318] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10768), 1,
      anon_sym_SEMI,
    STATE(6173), 1,
      sym_attribute_specifier,
  [51334] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10770), 1,
      anon_sym_LPAREN2,
    ACTIONS(10772), 1,
      anon_sym_LBRACE,
    ACTIONS(10774), 1,
      anon_sym_ATautoreleasepool,
    STATE(211), 1,
      sym_compound_statement,
  [51350] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10776), 1,
      sym_identifier,
    ACTIONS(10778), 1,
      anon_sym_COLON,
    STATE(3136), 1,
      sym_enumerator_list,
  [51366] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10780), 1,
      anon_sym_DQUOTE,
    ACTIONS(10782), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10784), 1,
      sym_escape_sequence,
    STATE(5259), 1,
      aux_sym_string_literal_repeat1,
  [51382] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10736), 1,
      anon_sym_LPAREN,
    ACTIONS(10786), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(10788), 1,
      sym_preproc_arg,
    STATE(5899), 1,
      sym_preproc_params,
  [51398] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10790), 1,
      sym_identifier,
    ACTIONS(10792), 1,
      anon_sym_COLON,
    STATE(3891), 1,
      sym_enumerator_list,
  [51414] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8089), 1,
      anon_sym_LBRACK,
    ACTIONS(8091), 1,
      anon_sym_RPAREN,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(4311), 1,
      sym_parameter_list,
  [51430] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10473), 1,
      anon_sym_COLON,
    ACTIONS(10794), 1,
      sym_identifier,
    STATE(4161), 1,
      sym_enumerator_list,
  [51446] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10796), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [51462] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10660), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [51478] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10798), 1,
      anon_sym_SEMI,
    STATE(6391), 1,
      sym_attribute_specifier,
  [51494] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10792), 1,
      anon_sym_COLON,
    ACTIONS(10800), 1,
      sym_identifier,
    STATE(3891), 1,
      sym_enumerator_list,
  [51510] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10802), 1,
      anon_sym_SEMI,
    STATE(6443), 1,
      sym_attribute_specifier,
  [51526] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10804), 1,
      anon_sym_LPAREN2,
    ACTIONS(10806), 1,
      anon_sym_LBRACE,
    ACTIONS(10808), 1,
      anon_sym_ATautoreleasepool,
    STATE(187), 1,
      sym_compound_statement,
  [51542] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9152), 1,
      anon_sym_COLON_COLON,
    ACTIONS(10810), 1,
      anon_sym_LPAREN2,
    ACTIONS(9147), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [51556] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10812), 1,
      anon_sym_SEMI,
    STATE(6453), 1,
      sym_attribute_specifier,
  [51572] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10268), 4,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      sym_identifier,
  [51582] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10814), 1,
      anon_sym_COMMA,
    ACTIONS(10817), 1,
      anon_sym_GT,
    ACTIONS(10819), 1,
      anon_sym_COLON,
    STATE(5186), 1,
      aux_sym_parameterized_arguments_repeat1,
  [51598] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10821), 1,
      anon_sym_SEMI,
    STATE(6270), 1,
      sym_attribute_specifier,
  [51614] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10457), 1,
      sym_identifier,
    STATE(5235), 1,
      aux_sym_method_identifier_repeat2,
    ACTIONS(10823), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [51628] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10825), 1,
      anon_sym_LPAREN2,
    ACTIONS(10827), 1,
      anon_sym_LBRACE,
    ACTIONS(10829), 1,
      anon_sym_ATautoreleasepool,
    STATE(438), 1,
      sym_compound_statement,
  [51644] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10736), 1,
      anon_sym_LPAREN,
    ACTIONS(10831), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(10833), 1,
      sym_preproc_arg,
    STATE(5792), 1,
      sym_preproc_params,
  [51660] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10835), 1,
      anon_sym_SQUOTE,
    STATE(5232), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(10837), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [51674] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10840), 1,
      anon_sym_DQUOTE,
    ACTIONS(10842), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10845), 1,
      sym_escape_sequence,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [51690] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10848), 1,
      anon_sym_COMMA,
    ACTIONS(10852), 1,
      anon_sym_GT,
    ACTIONS(10855), 1,
      anon_sym_COLON,
    STATE(5186), 1,
      aux_sym_parameterized_arguments_repeat1,
  [51706] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10858), 1,
      sym_identifier,
    STATE(5235), 1,
      aux_sym_method_identifier_repeat2,
    ACTIONS(10577), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [51720] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10861), 1,
      anon_sym_SEMI,
    STATE(6324), 1,
      sym_attribute_specifier,
  [51736] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10863), 1,
      anon_sym_SQUOTE,
    STATE(5232), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(10756), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [51750] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10865), 1,
      anon_sym_SEMI,
    STATE(6351), 1,
      sym_attribute_specifier,
  [51766] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10867), 1,
      anon_sym_DQUOTE,
    ACTIONS(10869), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10871), 1,
      sym_escape_sequence,
    STATE(5260), 1,
      aux_sym_string_literal_repeat1,
  [51782] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10473), 1,
      anon_sym_COLON,
    ACTIONS(10873), 1,
      sym_identifier,
    STATE(4161), 1,
      sym_enumerator_list,
  [51798] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10875), 1,
      anon_sym_DQUOTE,
    ACTIONS(10877), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10879), 1,
      sym_escape_sequence,
    STATE(5285), 1,
      aux_sym_string_literal_repeat1,
  [51814] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    ACTIONS(10881), 1,
      anon_sym_LPAREN2,
    STATE(438), 1,
      sym_compound_statement,
  [51830] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10883), 1,
      anon_sym_COMMA,
    STATE(5265), 1,
      aux_sym_gnu_asm_input_operand_list_repeat1,
    ACTIONS(10885), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [51844] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10887), 1,
      anon_sym_SEMI,
    STATE(6473), 1,
      sym_attribute_specifier,
  [51860] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(10889), 1,
      anon_sym_SEMI,
    ACTIONS(10891), 1,
      anon_sym_EQ,
    STATE(5670), 1,
      aux_sym_property_implementation_repeat1,
  [51876] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8089), 1,
      anon_sym_LBRACK,
    ACTIONS(8111), 1,
      anon_sym_RPAREN,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    STATE(4311), 1,
      sym_parameter_list,
  [51892] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10893), 1,
      anon_sym_COMMA,
    STATE(5247), 1,
      aux_sym_parameterized_arguments_repeat1,
    ACTIONS(10896), 2,
      anon_sym_GT,
      anon_sym_COLON,
  [51906] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10898), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [51922] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10736), 1,
      anon_sym_LPAREN,
    ACTIONS(10900), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(10902), 1,
      sym_preproc_arg,
    STATE(5894), 1,
      sym_preproc_params,
  [51938] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10449), 1,
      anon_sym_COLON,
    ACTIONS(10904), 1,
      sym_identifier,
    STATE(3178), 1,
      sym_enumerator_list,
  [51954] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10645), 1,
      anon_sym_DQUOTE,
    ACTIONS(10906), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10908), 1,
      sym_escape_sequence,
    STATE(5196), 1,
      aux_sym_string_literal_repeat1,
  [51970] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10910), 1,
      anon_sym_SQUOTE,
    STATE(5232), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(10756), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [51984] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10912), 1,
      anon_sym_LPAREN2,
    ACTIONS(10914), 1,
      anon_sym_LBRACE,
    ACTIONS(10916), 1,
      anon_sym_ATautoreleasepool,
    STATE(427), 1,
      sym_compound_statement,
  [52000] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10918), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [52016] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10019), 1,
      anon_sym_COLON2,
    ACTIONS(10920), 1,
      anon_sym_RPAREN,
    STATE(4344), 1,
      aux_sym_selector_expression_repeat2,
    STATE(5035), 1,
      aux_sym_method_identifier_repeat1,
  [52032] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10457), 1,
      sym_identifier,
    STATE(5235), 1,
      aux_sym_method_identifier_repeat2,
    ACTIONS(10459), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [52046] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10766), 1,
      anon_sym_DQUOTE,
    ACTIONS(10922), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10924), 1,
      sym_escape_sequence,
    STATE(5248), 1,
      aux_sym_string_literal_repeat1,
  [52062] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10926), 1,
      anon_sym_DQUOTE,
    ACTIONS(10928), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10930), 1,
      sym_escape_sequence,
    STATE(5177), 1,
      aux_sym_string_literal_repeat1,
  [52078] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10932), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [52094] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10934), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [52110] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10936), 1,
      sym_identifier,
    ACTIONS(10938), 1,
      anon_sym_COLON,
    STATE(4151), 1,
      sym_enumerator_list,
  [52126] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    ACTIONS(10467), 1,
      anon_sym_COLON,
    ACTIONS(10940), 1,
      sym_identifier,
    STATE(3868), 1,
      sym_enumerator_list,
  [52142] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10942), 1,
      anon_sym_DQUOTE,
    ACTIONS(10944), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10946), 1,
      sym_escape_sequence,
    STATE(5209), 1,
      aux_sym_string_literal_repeat1,
  [52158] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10766), 1,
      anon_sym_DQUOTE,
    ACTIONS(10948), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10950), 1,
      sym_escape_sequence,
    STATE(5166), 1,
      aux_sym_string_literal_repeat1,
  [52174] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10952), 1,
      anon_sym_COMMA,
    STATE(5265), 1,
      aux_sym_gnu_asm_input_operand_list_repeat1,
    ACTIONS(10955), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [52188] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10035), 1,
      sym_identifier,
    ACTIONS(10957), 1,
      aux_sym_preproc_if_token2,
    STATE(4910), 1,
      aux_sym_preproc_ifdef_in_enumerator_repeat1,
    STATE(6436), 1,
      sym_enumerator,
  [52204] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10959), 1,
      anon_sym_SEMI,
    STATE(6335), 1,
      sym_attribute_specifier,
  [52220] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10961), 1,
      anon_sym_LPAREN2,
    ACTIONS(10963), 1,
      anon_sym_LBRACE,
    ACTIONS(10965), 1,
      anon_sym_ATautoreleasepool,
    STATE(1082), 1,
      sym_compound_statement,
  [52236] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10019), 1,
      anon_sym_COLON2,
    ACTIONS(10967), 1,
      anon_sym_RPAREN,
    STATE(1438), 1,
      aux_sym_selector_expression_repeat2,
    STATE(5035), 1,
      aux_sym_method_identifier_repeat1,
  [52252] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10969), 1,
      anon_sym_SEMI,
    STATE(6215), 1,
      sym_attribute_specifier,
  [52268] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10670), 1,
      anon_sym_COMMA,
    STATE(5310), 1,
      aux_sym_gnu_asm_clobber_list_repeat1,
    ACTIONS(10971), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [52282] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10973), 1,
      anon_sym_COMMA,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
    ACTIONS(10976), 2,
      anon_sym_GT,
      anon_sym_SEMI,
  [52296] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10035), 1,
      sym_identifier,
    ACTIONS(10978), 1,
      aux_sym_preproc_if_token2,
    STATE(5266), 1,
      aux_sym_preproc_ifdef_in_enumerator_repeat1,
    STATE(6436), 1,
      sym_enumerator,
  [52312] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10736), 1,
      anon_sym_LPAREN,
    ACTIONS(10980), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(10982), 1,
      sym_preproc_arg,
    STATE(5739), 1,
      sym_preproc_params,
  [52328] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10938), 1,
      anon_sym_COLON,
    ACTIONS(10984), 1,
      sym_identifier,
    STATE(4151), 1,
      sym_enumerator_list,
  [52344] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8089), 1,
      anon_sym_LBRACK,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10986), 1,
      anon_sym_RPAREN,
    STATE(4311), 1,
      sym_parameter_list,
  [52360] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(10376), 1,
      anon_sym_LBRACK,
    ACTIONS(10988), 1,
      anon_sym_RPAREN,
    STATE(4270), 1,
      sym_parameter_list,
  [52376] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(10990), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [52392] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10992), 1,
      anon_sym_SQUOTE,
    STATE(5232), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(10756), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [52406] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(10994), 1,
      anon_sym_SEMI,
    STATE(6573), 1,
      sym_attribute_specifier,
  [52422] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10736), 1,
      anon_sym_LPAREN,
    ACTIONS(10996), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(10998), 1,
      sym_preproc_arg,
    STATE(5883), 1,
      sym_preproc_params,
  [52438] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11000), 1,
      anon_sym_DQUOTE,
    ACTIONS(11002), 1,
      aux_sym_string_literal_token1,
    ACTIONS(11004), 1,
      sym_escape_sequence,
    STATE(5278), 1,
      aux_sym_string_literal_repeat1,
  [52454] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    ACTIONS(10778), 1,
      anon_sym_COLON,
    ACTIONS(11006), 1,
      sym_identifier,
    STATE(3136), 1,
      sym_enumerator_list,
  [52470] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(11008), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [52486] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(11010), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [52502] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(11012), 1,
      anon_sym_SEMI,
    STATE(6344), 1,
      sym_attribute_specifier,
  [52518] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(11014), 1,
      anon_sym_RPAREN,
    STATE(4804), 1,
      sym_parameter_list,
  [52534] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10691), 1,
      anon_sym_COMMA,
    ACTIONS(11016), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(11018), 1,
      anon_sym_RPAREN,
    STATE(5188), 1,
      aux_sym_preproc_params_repeat1,
  [52550] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11020), 1,
      anon_sym_COMMA,
    STATE(5300), 1,
      aux_sym_gnu_asm_output_operand_list_repeat1,
    ACTIONS(11022), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [52564] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10736), 1,
      anon_sym_LPAREN,
    ACTIONS(11024), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11026), 1,
      sym_preproc_arg,
    STATE(5879), 1,
      sym_preproc_params,
  [52580] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11020), 1,
      anon_sym_COMMA,
    STATE(5289), 1,
      aux_sym_gnu_asm_output_operand_list_repeat1,
    ACTIONS(11028), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [52594] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10736), 1,
      anon_sym_LPAREN,
    ACTIONS(11030), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11032), 1,
      sym_preproc_arg,
    STATE(5786), 1,
      sym_preproc_params,
  [52610] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11036), 1,
      anon_sym_COMMA,
    STATE(5293), 1,
      aux_sym_preproc_params_repeat1,
    ACTIONS(11034), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_RPAREN,
  [52624] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10883), 1,
      anon_sym_COMMA,
    STATE(5243), 1,
      aux_sym_gnu_asm_input_operand_list_repeat1,
    ACTIONS(11039), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [52638] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10019), 1,
      anon_sym_COLON2,
    ACTIONS(11041), 1,
      anon_sym_RPAREN,
    STATE(4470), 1,
      aux_sym_selector_expression_repeat2,
    STATE(5035), 1,
      aux_sym_method_identifier_repeat1,
  [52654] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10934), 1,
      anon_sym_DQUOTE,
    ACTIONS(11043), 1,
      aux_sym_string_literal_token1,
    ACTIONS(11045), 1,
      sym_escape_sequence,
    STATE(5193), 1,
      aux_sym_string_literal_repeat1,
  [52670] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10948), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10950), 1,
      sym_escape_sequence,
    ACTIONS(11010), 1,
      anon_sym_DQUOTE,
    STATE(5166), 1,
      aux_sym_string_literal_repeat1,
  [52686] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10620), 1,
      aux_sym_string_literal_token1,
    ACTIONS(10622), 1,
      sym_escape_sequence,
    ACTIONS(11047), 1,
      anon_sym_DQUOTE,
    STATE(5233), 1,
      aux_sym_string_literal_repeat1,
  [52702] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11049), 1,
      anon_sym_DQUOTE,
    ACTIONS(11051), 1,
      aux_sym_string_literal_token1,
    ACTIONS(11053), 1,
      sym_escape_sequence,
    STATE(5175), 1,
      aux_sym_string_literal_repeat1,
  [52718] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11055), 1,
      anon_sym_COMMA,
    STATE(5300), 1,
      aux_sym_gnu_asm_output_operand_list_repeat1,
    ACTIONS(11058), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [52732] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(11060), 1,
      anon_sym_SEMI,
    STATE(6416), 1,
      sym_attribute_specifier,
  [52748] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(11062), 1,
      anon_sym_SEMI,
    STATE(6544), 1,
      sym_attribute_specifier,
  [52764] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(11064), 1,
      anon_sym_SEMI,
    STATE(6692), 1,
      sym_attribute_specifier,
  [52780] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(11066), 1,
      anon_sym_SEMI,
    STATE(5913), 1,
      sym_attribute_specifier,
  [52796] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11068), 1,
      anon_sym_SEMI,
    ACTIONS(11070), 1,
      anon_sym_EQ,
    STATE(5652), 1,
      aux_sym_property_implementation_repeat1,
  [52812] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(11072), 1,
      anon_sym_SEMI,
    STATE(5967), 1,
      sym_attribute_specifier,
  [52828] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8850), 1,
      anon_sym_LPAREN2,
    ACTIONS(8852), 1,
      anon_sym_LBRACK,
    ACTIONS(11074), 1,
      anon_sym_RPAREN,
    STATE(4804), 1,
      sym_parameter_list,
  [52844] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10736), 1,
      anon_sym_LPAREN,
    ACTIONS(11076), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11078), 1,
      sym_preproc_arg,
    STATE(5838), 1,
      sym_preproc_params,
  [52860] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11080), 1,
      anon_sym_SEMI,
    ACTIONS(11082), 1,
      anon_sym_EQ,
    STATE(5316), 1,
      aux_sym_property_implementation_repeat1,
  [52876] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11084), 1,
      anon_sym_COMMA,
    STATE(5310), 1,
      aux_sym_gnu_asm_clobber_list_repeat1,
    ACTIONS(11087), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [52890] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym___attribute,
    ACTIONS(6768), 1,
      anon_sym___attribute__,
    ACTIONS(11089), 1,
      anon_sym_SEMI,
    STATE(6542), 1,
      sym_attribute_specifier,
  [52906] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10932), 1,
      anon_sym_DQUOTE,
    ACTIONS(11091), 1,
      aux_sym_string_literal_token1,
    ACTIONS(11093), 1,
      sym_escape_sequence,
    STATE(5284), 1,
      aux_sym_string_literal_repeat1,
  [52922] = 5,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11095), 1,
      anon_sym_DQUOTE,
    ACTIONS(11097), 1,
      aux_sym_string_literal_token1,
    ACTIONS(11099), 1,
      sym_escape_sequence,
    STATE(5254), 1,
      aux_sym_string_literal_repeat1,
  [52938] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(944), 1,
      anon_sym_LBRACE,
    ACTIONS(948), 1,
      anon_sym_ATautoreleasepool,
    STATE(787), 1,
      sym_compound_statement,
  [52951] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(870), 1,
      anon_sym_LBRACE,
    ACTIONS(872), 1,
      anon_sym_ATautoreleasepool,
    STATE(573), 1,
      sym_compound_statement,
  [52964] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11101), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [52977] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9452), 1,
      anon_sym_COMMA,
    ACTIONS(9468), 1,
      anon_sym_RPAREN,
    STATE(5617), 1,
      aux_sym_argument_list_repeat1,
  [52990] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11101), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53003] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11101), 1,
      anon_sym_SEMI,
    STATE(5369), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53016] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11103), 1,
      anon_sym_RPAREN,
    ACTIONS(11105), 1,
      anon_sym_COLON,
    STATE(6629), 1,
      sym_gnu_asm_goto_list,
  [53029] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11107), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53042] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9767), 1,
      anon_sym_COMMA,
    ACTIONS(11109), 1,
      anon_sym_RPAREN,
    STATE(5520), 1,
      aux_sym_preproc_argument_list_repeat1,
  [53055] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11111), 1,
      anon_sym_RPAREN,
    ACTIONS(11113), 1,
      anon_sym_COLON,
    STATE(5403), 1,
      sym_gnu_asm_output_operand_list,
  [53068] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(10714), 1,
      anon_sym_SEMI,
    STATE(5328), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53081] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      anon_sym_ATautoreleasepool,
    STATE(936), 1,
      sym_compound_statement,
  [53094] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(303), 1,
      anon_sym_LBRACE,
    ACTIONS(305), 1,
      anon_sym_ATautoreleasepool,
    STATE(428), 1,
      sym_compound_statement,
  [53107] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11115), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [53120] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11115), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53133] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11115), 1,
      anon_sym_SEMI,
    STATE(5332), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53146] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(303), 1,
      anon_sym_LBRACE,
    ACTIONS(305), 1,
      anon_sym_ATautoreleasepool,
    STATE(436), 1,
      sym_compound_statement,
  [53159] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11117), 1,
      anon_sym_SEMI,
    STATE(5334), 1,
      aux_sym_property_implementation_repeat1,
  [53172] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11117), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53185] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(303), 1,
      anon_sym_LBRACE,
    ACTIONS(305), 1,
      anon_sym_ATautoreleasepool,
    STATE(362), 1,
      sym_compound_statement,
  [53198] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11119), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [53211] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10772), 1,
      anon_sym_LBRACE,
    ACTIONS(10774), 1,
      anon_sym_ATautoreleasepool,
    STATE(213), 1,
      sym_compound_statement,
  [53224] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11123), 1,
      anon_sym_EQ,
    ACTIONS(11121), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [53235] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2108), 1,
      sym_compound_statement,
  [53248] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11125), 1,
      anon_sym_COMMA,
    ACTIONS(11127), 1,
      anon_sym_RPAREN,
    STATE(5339), 1,
      aux_sym_property_attributes_declaration_repeat1,
  [53261] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11125), 1,
      anon_sym_COMMA,
    ACTIONS(11129), 1,
      anon_sym_RPAREN,
    STATE(5603), 1,
      aux_sym_property_attributes_declaration_repeat1,
  [53274] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(11131), 1,
      anon_sym_RPAREN,
    STATE(5421), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [53287] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11133), 1,
      anon_sym_RPAREN,
    ACTIONS(11135), 1,
      anon_sym_COLON,
    STATE(5386), 1,
      sym_gnu_asm_clobber_list,
  [53300] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11137), 1,
      anon_sym_RPAREN,
    ACTIONS(11139), 1,
      anon_sym_COLON,
    STATE(5388), 1,
      sym_gnu_asm_input_operand_list,
  [53313] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2188), 1,
      anon_sym_LBRACE,
    ACTIONS(2190), 1,
      anon_sym_ATautoreleasepool,
    STATE(1131), 1,
      sym_compound_statement,
  [53326] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11141), 1,
      anon_sym_SEMI,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [53339] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11143), 1,
      sym_identifier,
    ACTIONS(11146), 1,
      anon_sym_RBRACK,
    STATE(5345), 1,
      aux_sym_message_expression_repeat2,
  [53352] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(870), 1,
      anon_sym_LBRACE,
    ACTIONS(872), 1,
      anon_sym_ATautoreleasepool,
    STATE(533), 1,
      sym_compound_statement,
  [53365] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(870), 1,
      anon_sym_LBRACE,
    ACTIONS(872), 1,
      anon_sym_ATautoreleasepool,
    STATE(542), 1,
      sym_compound_statement,
  [53378] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(870), 1,
      anon_sym_LBRACE,
    ACTIONS(872), 1,
      anon_sym_ATautoreleasepool,
    STATE(545), 1,
      sym_compound_statement,
  [53391] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10963), 1,
      anon_sym_LBRACE,
    ACTIONS(10965), 1,
      anon_sym_ATautoreleasepool,
    STATE(1078), 1,
      sym_compound_statement,
  [53404] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11148), 1,
      anon_sym_COMMA,
    ACTIONS(11151), 1,
      anon_sym_RPAREN,
    STATE(5350), 1,
      aux_sym_attribute_specifier_repeat1,
  [53417] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(11153), 1,
      anon_sym_RPAREN,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [53430] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11155), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11157), 1,
      anon_sym_LPAREN2,
    STATE(6607), 1,
      sym_preproc_argument_list,
  [53443] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11159), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [53452] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2188), 1,
      anon_sym_LBRACE,
    ACTIONS(2190), 1,
      anon_sym_ATautoreleasepool,
    STATE(1148), 1,
      sym_compound_statement,
  [53465] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2188), 1,
      anon_sym_LBRACE,
    ACTIONS(2190), 1,
      anon_sym_ATautoreleasepool,
    STATE(1153), 1,
      sym_compound_statement,
  [53478] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2188), 1,
      anon_sym_LBRACE,
    ACTIONS(2190), 1,
      anon_sym_ATautoreleasepool,
    STATE(1156), 1,
      sym_compound_statement,
  [53491] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(944), 1,
      anon_sym_LBRACE,
    ACTIONS(948), 1,
      anon_sym_ATautoreleasepool,
    STATE(815), 1,
      sym_compound_statement,
  [53504] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11105), 1,
      anon_sym_COLON,
    ACTIONS(11161), 1,
      anon_sym_RPAREN,
    STATE(6537), 1,
      sym_gnu_asm_goto_list,
  [53517] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11163), 1,
      anon_sym_COMMA,
    ACTIONS(11165), 1,
      anon_sym_RPAREN,
    STATE(5634), 1,
      aux_sym_generic_arguments_repeat1,
  [53530] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(11167), 1,
      anon_sym_RPAREN,
    STATE(5421), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [53543] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11169), 1,
      anon_sym_COMMA,
    ACTIONS(11172), 1,
      anon_sym_RPAREN,
    STATE(5361), 1,
      aux_sym_version_repeat1,
  [53556] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2262), 1,
      sym_compound_statement,
  [53569] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10443), 1,
      sym_identifier,
    STATE(6560), 1,
      sym_method_type,
  [53582] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11176), 1,
      anon_sym_EQ,
    ACTIONS(11174), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [53593] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9462), 1,
      anon_sym_COMMA,
    ACTIONS(9464), 1,
      anon_sym_RPAREN,
    STATE(5432), 1,
      aux_sym_availability_repeat1,
  [53606] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11174), 1,
      anon_sym_RPAREN,
    ACTIONS(11178), 1,
      anon_sym_COMMA,
    STATE(5366), 1,
      aux_sym_availability_repeat1,
  [53619] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2109), 1,
      sym_compound_statement,
  [53632] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11181), 1,
      anon_sym_SEMI,
    STATE(5396), 1,
      aux_sym_property_implementation_repeat1,
  [53645] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11181), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53658] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2101), 1,
      sym_compound_statement,
  [53671] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      anon_sym_ATautoreleasepool,
    STATE(944), 1,
      sym_compound_statement,
  [53684] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11183), 1,
      sym_identifier,
    ACTIONS(11185), 1,
      anon_sym_RPAREN,
    STATE(5338), 1,
      sym_property_attribute,
  [53697] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2260), 1,
      sym_compound_statement,
  [53710] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11187), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [53723] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11187), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53736] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11187), 1,
      anon_sym_SEMI,
    STATE(5497), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53749] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10704), 1,
      anon_sym_LBRACE,
    ACTIONS(10706), 1,
      anon_sym_ATautoreleasepool,
    STATE(763), 1,
      sym_compound_statement,
  [53762] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11189), 1,
      anon_sym_COMMA,
    ACTIONS(11191), 1,
      anon_sym_GT,
    STATE(5624), 1,
      aux_sym_argument_list_repeat2,
  [53775] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2110), 1,
      sym_compound_statement,
  [53788] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11193), 3,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_COLON,
  [53797] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(4967), 1,
      sym_compound_statement,
  [53810] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11195), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [53823] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    ACTIONS(11199), 1,
      anon_sym_GT,
    STATE(5413), 1,
      aux_sym_parameterized_arguments_repeat2,
  [53836] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    STATE(5252), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(11201), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [53847] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(11203), 1,
      anon_sym_RPAREN,
    STATE(5508), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [53860] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11105), 1,
      anon_sym_COLON,
    ACTIONS(11205), 1,
      anon_sym_RPAREN,
    STATE(6395), 1,
      sym_gnu_asm_goto_list,
  [53873] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(11207), 1,
      anon_sym_RPAREN,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [53886] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11135), 1,
      anon_sym_COLON,
    ACTIONS(11209), 1,
      anon_sym_RPAREN,
    STATE(5400), 1,
      sym_gnu_asm_clobber_list,
  [53899] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2111), 1,
      sym_compound_statement,
  [53912] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(4957), 1,
      sym_compound_statement,
  [53925] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2297), 1,
      sym_compound_statement,
  [53938] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9767), 1,
      anon_sym_COMMA,
    ACTIONS(11211), 1,
      anon_sym_RPAREN,
    STATE(5520), 1,
      aux_sym_preproc_argument_list_repeat1,
  [53951] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(181), 1,
      anon_sym_LBRACE,
    ACTIONS(183), 1,
      anon_sym_ATautoreleasepool,
    STATE(243), 1,
      sym_compound_statement,
  [53964] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2112), 1,
      sym_compound_statement,
  [53977] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11213), 1,
      anon_sym_COMMA,
    ACTIONS(11215), 1,
      anon_sym_RBRACE,
    STATE(5423), 1,
      aux_sym_dictionary_literal_repeat1,
  [53990] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11217), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [54003] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10806), 1,
      anon_sym_LBRACE,
    ACTIONS(10808), 1,
      anon_sym_ATautoreleasepool,
    STATE(188), 1,
      sym_compound_statement,
  [54016] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11221), 1,
      anon_sym_RBRACK,
    STATE(5406), 1,
      aux_sym_attribute_declaration_repeat1,
  [54029] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11189), 1,
      anon_sym_COMMA,
    ACTIONS(11223), 1,
      anon_sym_GT,
    STATE(5522), 1,
      aux_sym_argument_list_repeat2,
  [54042] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11105), 1,
      anon_sym_COLON,
    ACTIONS(11225), 1,
      anon_sym_RPAREN,
    STATE(6456), 1,
      sym_gnu_asm_goto_list,
  [54055] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2345), 1,
      sym_compound_statement,
  [54068] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11227), 1,
      anon_sym_SEMI,
    ACTIONS(11229), 1,
      anon_sym_DOT,
    STATE(5477), 1,
      aux_sym_module_import_repeat1,
  [54081] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11139), 1,
      anon_sym_COLON,
    ACTIONS(11231), 1,
      anon_sym_RPAREN,
    STATE(5692), 1,
      sym_gnu_asm_input_operand_list,
  [54094] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    ACTIONS(11233), 1,
      anon_sym_GT,
    STATE(5604), 1,
      aux_sym_parameterized_arguments_repeat2,
  [54107] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9020), 1,
      anon_sym_RPAREN,
    STATE(5416), 1,
      aux_sym_attribute_specifier_repeat1,
  [54120] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11235), 1,
      anon_sym_RBRACK,
    STATE(5593), 1,
      aux_sym_attribute_declaration_repeat1,
  [54133] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11113), 1,
      anon_sym_COLON,
    ACTIONS(11237), 1,
      anon_sym_RPAREN,
    STATE(5695), 1,
      sym_gnu_asm_output_operand_list,
  [54146] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11163), 1,
      anon_sym_COMMA,
    ACTIONS(11239), 1,
      anon_sym_RPAREN,
    STATE(5634), 1,
      aux_sym_generic_arguments_repeat1,
  [54159] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2308), 1,
      sym_compound_statement,
  [54172] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    ACTIONS(11233), 1,
      anon_sym_GT,
    STATE(5413), 1,
      aux_sym_parameterized_arguments_repeat2,
  [54185] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6966), 1,
      anon_sym_LBRACE,
    ACTIONS(11241), 1,
      sym_identifier,
    STATE(2921), 1,
      sym_field_declaration_list,
  [54198] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11243), 1,
      sym_identifier,
    ACTIONS(11245), 2,
      anon_sym___covariant,
      anon_sym___contravariant,
  [54209] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11247), 1,
      anon_sym_COMMA,
    ACTIONS(11250), 1,
      anon_sym_GT,
    STATE(5413), 1,
      aux_sym_parameterized_arguments_repeat2,
  [54222] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11189), 1,
      anon_sym_COMMA,
    ACTIONS(11252), 1,
      anon_sym_GT,
    STATE(5446), 1,
      aux_sym_argument_list_repeat2,
  [54235] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8601), 1,
      anon_sym_RBRACK,
    ACTIONS(11254), 1,
      anon_sym_COMMA,
    STATE(5703), 1,
      aux_sym_attribute_repeat1,
  [54248] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(11256), 1,
      anon_sym_RPAREN,
    STATE(5350), 1,
      aux_sym_attribute_specifier_repeat1,
  [54261] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(11258), 1,
      anon_sym_RPAREN,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [54274] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2312), 1,
      sym_compound_statement,
  [54287] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10589), 1,
      anon_sym_GT,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    STATE(5413), 1,
      aux_sym_parameterized_arguments_repeat2,
  [54300] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9456), 1,
      anon_sym_COMMA,
    ACTIONS(11260), 1,
      anon_sym_RPAREN,
    STATE(5460), 1,
      aux_sym_generic_expression_repeat1,
  [54313] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10029), 1,
      anon_sym_RPAREN,
    ACTIONS(11262), 1,
      anon_sym_COMMA,
    STATE(5421), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [54326] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11265), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [54335] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8547), 1,
      anon_sym_RBRACE,
    ACTIONS(11267), 1,
      anon_sym_COMMA,
    STATE(5525), 1,
      aux_sym_dictionary_literal_repeat1,
  [54348] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11269), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [54357] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(409), 1,
      sym_compound_statement,
  [54370] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11271), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [54383] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2315), 1,
      sym_compound_statement,
  [54396] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11273), 1,
      anon_sym_COMMA,
    ACTIONS(11275), 1,
      anon_sym_RPAREN,
    STATE(5476), 1,
      aux_sym_gnu_asm_goto_list_repeat1,
  [54409] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11277), 1,
      anon_sym_COMMA,
    ACTIONS(11280), 1,
      anon_sym_GT,
    STATE(5429), 1,
      aux_sym_argument_list_repeat2,
  [54422] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2252), 1,
      sym_compound_statement,
  [54435] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2339), 1,
      sym_compound_statement,
  [54448] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9462), 1,
      anon_sym_COMMA,
    ACTIONS(11282), 1,
      anon_sym_RPAREN,
    STATE(5366), 1,
      aux_sym_availability_repeat1,
  [54461] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6966), 1,
      anon_sym_LBRACE,
    ACTIONS(11284), 1,
      sym_identifier,
    STATE(2925), 1,
      sym_field_declaration_list,
  [54474] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11286), 1,
      anon_sym_SEMI,
    STATE(5533), 1,
      aux_sym_class_declaration_repeat1,
  [54487] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    STATE(5237), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(11288), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [54498] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(4977), 1,
      sym_compound_statement,
  [54511] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2342), 1,
      sym_compound_statement,
  [54524] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9456), 1,
      anon_sym_COMMA,
    ACTIONS(11290), 1,
      anon_sym_RPAREN,
    STATE(5460), 1,
      aux_sym_generic_expression_repeat1,
  [54537] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11286), 1,
      anon_sym_SEMI,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [54550] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2295), 1,
      sym_compound_statement,
  [54563] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(10608), 1,
      anon_sym_SEMI,
    STATE(5375), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [54576] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2250), 1,
      sym_compound_statement,
  [54589] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_ATautoreleasepool,
    STATE(2256), 1,
      sym_compound_statement,
  [54602] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11292), 1,
      anon_sym_RBRACK,
    STATE(5449), 1,
      aux_sym_attribute_declaration_repeat1,
  [54615] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2271), 1,
      sym_compound_statement,
  [54628] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11189), 1,
      anon_sym_COMMA,
    ACTIONS(11294), 1,
      anon_sym_GT,
    STATE(5429), 1,
      aux_sym_argument_list_repeat2,
  [54641] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11296), 1,
      anon_sym_COMMA,
    ACTIONS(11298), 1,
      anon_sym_RPAREN,
    STATE(5597), 1,
      aux_sym_parameter_list_repeat1,
  [54654] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9012), 1,
      anon_sym_RPAREN,
    STATE(5453), 1,
      aux_sym_attribute_specifier_repeat1,
  [54667] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11300), 1,
      anon_sym_RBRACK,
    STATE(5593), 1,
      aux_sym_attribute_declaration_repeat1,
  [54680] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(11302), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(2805), 1,
      sym_method_parameter,
  [54693] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11157), 1,
      anon_sym_LPAREN2,
    ACTIONS(11304), 1,
      aux_sym_preproc_include_token3,
    STATE(6607), 1,
      sym_preproc_argument_list,
  [54706] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11229), 1,
      anon_sym_DOT,
    ACTIONS(11306), 1,
      anon_sym_SEMI,
    STATE(5462), 1,
      aux_sym_module_import_repeat1,
  [54719] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(11308), 1,
      anon_sym_RPAREN,
    STATE(5350), 1,
      aux_sym_attribute_specifier_repeat1,
  [54732] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(11310), 1,
      anon_sym_RPAREN,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [54745] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11296), 1,
      anon_sym_COMMA,
    ACTIONS(11312), 1,
      anon_sym_RPAREN,
    STATE(5687), 1,
      aux_sym_parameter_list_repeat1,
  [54758] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2104), 1,
      sym_compound_statement,
  [54771] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9476), 1,
      anon_sym_COMMA,
    ACTIONS(9478), 1,
      anon_sym_RBRACE,
    STATE(5691), 1,
      aux_sym_initializer_list_repeat1,
  [54784] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11314), 3,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_DOT,
  [54793] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11316), 1,
      anon_sym_COMMA,
    ACTIONS(11318), 1,
      anon_sym_RPAREN,
    STATE(5513), 1,
      aux_sym_version_repeat1,
  [54806] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11320), 1,
      anon_sym_COMMA,
    ACTIONS(11323), 1,
      anon_sym_RPAREN,
    STATE(5460), 1,
      aux_sym_generic_expression_repeat1,
  [54819] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    STATE(5279), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(11325), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [54830] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11327), 1,
      anon_sym_SEMI,
    ACTIONS(11329), 1,
      anon_sym_DOT,
    STATE(5462), 1,
      aux_sym_module_import_repeat1,
  [54843] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(4994), 1,
      sym_compound_statement,
  [54856] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11332), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [54865] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11334), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [54874] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1934), 1,
      anon_sym_LBRACE,
    ACTIONS(1936), 1,
      anon_sym_ATautoreleasepool,
    STATE(722), 1,
      sym_compound_statement,
  [54887] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11336), 1,
      anon_sym_RBRACK,
    STATE(5702), 1,
      aux_sym_attribute_declaration_repeat1,
  [54900] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11338), 1,
      anon_sym_RBRACK,
    STATE(5471), 1,
      aux_sym_attribute_declaration_repeat1,
  [54913] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11113), 1,
      anon_sym_COLON,
    ACTIONS(11340), 1,
      anon_sym_RPAREN,
    STATE(5630), 1,
      sym_gnu_asm_output_operand_list,
  [54926] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9016), 1,
      anon_sym_RPAREN,
    STATE(5473), 1,
      aux_sym_attribute_specifier_repeat1,
  [54939] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11342), 1,
      anon_sym_RBRACK,
    STATE(5593), 1,
      aux_sym_attribute_declaration_repeat1,
  [54952] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11229), 1,
      anon_sym_DOT,
    ACTIONS(11344), 1,
      anon_sym_SEMI,
    STATE(5501), 1,
      aux_sym_module_import_repeat1,
  [54965] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(11346), 1,
      anon_sym_RPAREN,
    STATE(5350), 1,
      aux_sym_attribute_specifier_repeat1,
  [54978] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(11348), 1,
      anon_sym_RPAREN,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [54991] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9516), 1,
      anon_sym_RPAREN,
    ACTIONS(11350), 1,
      anon_sym_COMMA,
    STATE(5475), 1,
      aux_sym_argument_list_repeat1,
  [55004] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11273), 1,
      anon_sym_COMMA,
    ACTIONS(11353), 1,
      anon_sym_RPAREN,
    STATE(5545), 1,
      aux_sym_gnu_asm_goto_list_repeat1,
  [55017] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11229), 1,
      anon_sym_DOT,
    ACTIONS(11355), 1,
      anon_sym_SEMI,
    STATE(5462), 1,
      aux_sym_module_import_repeat1,
  [55030] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(4958), 1,
      sym_compound_statement,
  [55043] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11359), 1,
      anon_sym_EQ,
    ACTIONS(11357), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [55054] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11361), 1,
      anon_sym_COMMA,
    ACTIONS(11364), 1,
      anon_sym_RPAREN,
    STATE(5480), 1,
      aux_sym_parameter_list_repeat1,
  [55067] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(3102), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [55080] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11366), 1,
      anon_sym_SEMI,
    STATE(5650), 1,
      aux_sym_property_implementation_repeat1,
  [55093] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9452), 1,
      anon_sym_COMMA,
    ACTIONS(11368), 1,
      anon_sym_RPAREN,
    STATE(5475), 1,
      aux_sym_argument_list_repeat1,
  [55106] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11370), 1,
      anon_sym_RBRACK,
    STATE(5489), 1,
      aux_sym_attribute_declaration_repeat1,
  [55119] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11105), 1,
      anon_sym_COLON,
    ACTIONS(11372), 1,
      anon_sym_RPAREN,
    STATE(5919), 1,
      sym_gnu_asm_goto_list,
  [55132] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10720), 1,
      anon_sym_SEMI,
    ACTIONS(11374), 1,
      anon_sym_COMMA,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [55145] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(11377), 1,
      anon_sym_RPAREN,
    STATE(5421), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [55158] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9024), 1,
      anon_sym_RPAREN,
    STATE(5491), 1,
      aux_sym_attribute_specifier_repeat1,
  [55171] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11379), 1,
      anon_sym_RBRACK,
    STATE(5593), 1,
      aux_sym_attribute_declaration_repeat1,
  [55184] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11135), 1,
      anon_sym_COLON,
    ACTIONS(11381), 1,
      anon_sym_RPAREN,
    STATE(5358), 1,
      sym_gnu_asm_clobber_list,
  [55197] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(11383), 1,
      anon_sym_RPAREN,
    STATE(5350), 1,
      aux_sym_attribute_specifier_repeat1,
  [55210] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(11385), 1,
      anon_sym_RPAREN,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [55223] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11157), 1,
      anon_sym_LPAREN2,
    ACTIONS(11387), 1,
      aux_sym_preproc_include_token3,
    STATE(6607), 1,
      sym_preproc_argument_list,
  [55236] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(4983), 1,
      sym_compound_statement,
  [55249] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11357), 1,
      anon_sym_SEMI,
    ACTIONS(11389), 1,
      anon_sym_COMMA,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [55262] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11392), 1,
      anon_sym_RBRACK,
    STATE(5500), 1,
      aux_sym_attribute_declaration_repeat1,
  [55275] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11366), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [55288] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11113), 1,
      anon_sym_COLON,
    ACTIONS(11394), 1,
      anon_sym_RPAREN,
    STATE(5524), 1,
      sym_gnu_asm_output_operand_list,
  [55301] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9018), 1,
      anon_sym_RPAREN,
    STATE(5502), 1,
      aux_sym_attribute_specifier_repeat1,
  [55314] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11396), 1,
      anon_sym_RBRACK,
    STATE(5593), 1,
      aux_sym_attribute_declaration_repeat1,
  [55327] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11229), 1,
      anon_sym_DOT,
    ACTIONS(11398), 1,
      anon_sym_SEMI,
    STATE(5462), 1,
      aux_sym_module_import_repeat1,
  [55340] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(11400), 1,
      anon_sym_RPAREN,
    STATE(5350), 1,
      aux_sym_attribute_specifier_repeat1,
  [55353] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(11402), 1,
      anon_sym_RPAREN,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [55366] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(3064), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [55379] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(4953), 1,
      sym_compound_statement,
  [55392] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9014), 1,
      anon_sym_RPAREN,
    STATE(5510), 1,
      aux_sym_attribute_specifier_repeat1,
  [55405] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10976), 3,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_SEMI,
  [55414] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(11404), 1,
      anon_sym_RPAREN,
    STATE(5421), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [55427] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2106), 1,
      sym_compound_statement,
  [55440] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(11406), 1,
      anon_sym_RPAREN,
    STATE(5350), 1,
      aux_sym_attribute_specifier_repeat1,
  [55453] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(11408), 1,
      anon_sym_RPAREN,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [55466] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(11404), 1,
      anon_sym_RPAREN,
    STATE(5360), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [55479] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11316), 1,
      anon_sym_COMMA,
    ACTIONS(11410), 1,
      anon_sym_RPAREN,
    STATE(5361), 1,
      aux_sym_version_repeat1,
  [55492] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11412), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
  [55501] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(10015), 1,
      anon_sym_RPAREN,
    STATE(5487), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [55514] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11157), 1,
      anon_sym_LPAREN2,
    ACTIONS(11414), 1,
      aux_sym_preproc_include_token3,
    STATE(6607), 1,
      sym_preproc_argument_list,
  [55527] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(10647), 1,
      anon_sym_SEMI,
    STATE(5540), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [55540] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(11416), 1,
      anon_sym_RPAREN,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [55553] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11418), 1,
      anon_sym_COMMA,
    ACTIONS(11421), 1,
      anon_sym_RPAREN,
    STATE(5519), 1,
      aux_sym__old_style_parameter_list_repeat1,
  [55566] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9911), 1,
      anon_sym_RPAREN,
    ACTIONS(11423), 1,
      anon_sym_COMMA,
    STATE(5520), 1,
      aux_sym_preproc_argument_list_repeat1,
  [55579] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(181), 1,
      anon_sym_LBRACE,
    ACTIONS(183), 1,
      anon_sym_ATautoreleasepool,
    STATE(261), 1,
      sym_compound_statement,
  [55592] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11189), 1,
      anon_sym_COMMA,
    ACTIONS(11426), 1,
      anon_sym_GT,
    STATE(5429), 1,
      aux_sym_argument_list_repeat2,
  [55605] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11430), 1,
      anon_sym_LPAREN2,
    ACTIONS(11428), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [55616] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11139), 1,
      anon_sym_COLON,
    ACTIONS(11432), 1,
      anon_sym_RPAREN,
    STATE(5548), 1,
      sym_gnu_asm_input_operand_list,
  [55629] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11434), 1,
      anon_sym_COMMA,
    ACTIONS(11437), 1,
      anon_sym_RBRACE,
    STATE(5525), 1,
      aux_sym_dictionary_literal_repeat1,
  [55642] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2277), 1,
      sym_compound_statement,
  [55655] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11113), 1,
      anon_sym_COLON,
    ACTIONS(11439), 1,
      anon_sym_RPAREN,
    STATE(5550), 1,
      sym_gnu_asm_output_operand_list,
  [55668] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11441), 1,
      anon_sym_SEMI,
    STATE(5552), 1,
      aux_sym_class_declaration_repeat1,
  [55681] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11441), 1,
      anon_sym_SEMI,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [55694] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(4982), 1,
      sym_compound_statement,
  [55707] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11034), 3,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [55716] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9462), 1,
      anon_sym_COMMA,
    ACTIONS(11443), 1,
      anon_sym_RPAREN,
    STATE(5366), 1,
      aux_sym_availability_repeat1,
  [55729] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11445), 1,
      anon_sym_SEMI,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [55742] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8760), 1,
      anon_sym_LBRACE,
    ACTIONS(11447), 1,
      sym_identifier,
    STATE(3795), 1,
      sym_field_declaration_list,
  [55755] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10045), 1,
      anon_sym_COMMA,
    ACTIONS(11449), 1,
      anon_sym_RPAREN,
    STATE(5585), 1,
      aux_sym_available_expression_repeat1,
  [55768] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8760), 1,
      anon_sym_LBRACE,
    ACTIONS(11451), 1,
      sym_identifier,
    STATE(3798), 1,
      sym_field_declaration_list,
  [55781] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(944), 1,
      anon_sym_LBRACE,
    ACTIONS(948), 1,
      anon_sym_ATautoreleasepool,
    STATE(827), 1,
      sym_compound_statement,
  [55794] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10045), 1,
      anon_sym_COMMA,
    ACTIONS(11449), 1,
      anon_sym_RPAREN,
    STATE(5601), 1,
      aux_sym_available_expression_repeat1,
  [55807] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11453), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [55820] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11453), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [55833] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11453), 1,
      anon_sym_SEMI,
    STATE(5563), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [55846] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11455), 1,
      anon_sym_GT,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [55859] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11457), 1,
      anon_sym_COMMA,
    ACTIONS(11459), 1,
      anon_sym_RPAREN,
    STATE(5553), 1,
      aux_sym_method_type_repeat1,
  [55872] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11461), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [55885] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11463), 1,
      anon_sym_COMMA,
    ACTIONS(11466), 1,
      anon_sym_RPAREN,
    STATE(5545), 1,
      aux_sym_gnu_asm_goto_list_repeat1,
  [55898] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(181), 1,
      anon_sym_LBRACE,
    ACTIONS(183), 1,
      anon_sym_ATautoreleasepool,
    STATE(267), 1,
      sym_compound_statement,
  [55911] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11229), 1,
      anon_sym_DOT,
    ACTIONS(11468), 1,
      anon_sym_SEMI,
    STATE(5590), 1,
      aux_sym_module_import_repeat1,
  [55924] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11135), 1,
      anon_sym_COLON,
    ACTIONS(11470), 1,
      anon_sym_RPAREN,
    STATE(5569), 1,
      sym_gnu_asm_clobber_list,
  [55937] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2283), 1,
      sym_compound_statement,
  [55950] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11139), 1,
      anon_sym_COLON,
    ACTIONS(11472), 1,
      anon_sym_RPAREN,
    STATE(5570), 1,
      sym_gnu_asm_input_operand_list,
  [55963] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9408), 1,
      anon_sym_RPAREN,
    ACTIONS(11474), 1,
      anon_sym_COMMA,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [55976] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11477), 1,
      anon_sym_SEMI,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [55989] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11479), 1,
      anon_sym_COMMA,
    ACTIONS(11482), 1,
      anon_sym_RPAREN,
    STATE(5553), 1,
      aux_sym_method_type_repeat1,
  [56002] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11484), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [56011] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11080), 1,
      anon_sym_SEMI,
    STATE(5318), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [56024] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2289), 1,
      sym_compound_statement,
  [56037] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(3078), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [56050] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6966), 1,
      anon_sym_LBRACE,
    ACTIONS(11486), 1,
      sym_identifier,
    STATE(2922), 1,
      sym_field_declaration_list,
  [56063] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9359), 1,
      sym_identifier,
    ACTIONS(11488), 1,
      anon_sym_RBRACK,
    STATE(5345), 1,
      aux_sym_message_expression_repeat2,
  [56076] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11490), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [56085] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11492), 1,
      anon_sym_SEMI,
    STATE(5572), 1,
      aux_sym_property_implementation_repeat1,
  [56098] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2299), 1,
      sym_compound_statement,
  [56111] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11492), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [56124] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11494), 1,
      anon_sym_COMMA,
    ACTIONS(11497), 1,
      anon_sym_RPAREN,
    STATE(5564), 1,
      aux_sym_generic_arguments_repeat1,
  [56137] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2301), 1,
      sym_compound_statement,
  [56150] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(181), 1,
      anon_sym_LBRACE,
    ACTIONS(183), 1,
      anon_sym_ATautoreleasepool,
    STATE(268), 1,
      sym_compound_statement,
  [56163] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11296), 1,
      anon_sym_COMMA,
    ACTIONS(11499), 1,
      anon_sym_RPAREN,
    STATE(5619), 1,
      aux_sym_parameter_list_repeat1,
  [56176] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2328), 1,
      sym_compound_statement,
  [56189] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11105), 1,
      anon_sym_COLON,
    ACTIONS(11501), 1,
      anon_sym_RPAREN,
    STATE(6338), 1,
      sym_gnu_asm_goto_list,
  [56202] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11135), 1,
      anon_sym_COLON,
    ACTIONS(11503), 1,
      anon_sym_RPAREN,
    STATE(5581), 1,
      sym_gnu_asm_clobber_list,
  [56215] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    STATE(1471), 1,
      sym_compound_statement,
  [56228] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11505), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [56241] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2332), 1,
      sym_compound_statement,
  [56254] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(11507), 1,
      anon_sym_RPAREN,
    STATE(5350), 1,
      aux_sym_attribute_specifier_repeat1,
  [56267] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11457), 1,
      anon_sym_COMMA,
    ACTIONS(11509), 1,
      anon_sym_RPAREN,
    STATE(5543), 1,
      aux_sym_method_type_repeat1,
  [56280] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11513), 1,
      anon_sym_LPAREN2,
    ACTIONS(11511), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [56291] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9359), 1,
      sym_identifier,
    ACTIONS(11515), 1,
      anon_sym_RBRACK,
    STATE(5345), 1,
      aux_sym_message_expression_repeat2,
  [56304] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9359), 1,
      sym_identifier,
    ACTIONS(11517), 1,
      anon_sym_RBRACK,
    STATE(5345), 1,
      aux_sym_message_expression_repeat2,
  [56317] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10914), 1,
      anon_sym_LBRACE,
    ACTIONS(10916), 1,
      anon_sym_ATautoreleasepool,
    STATE(382), 1,
      sym_compound_statement,
  [56330] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11519), 1,
      sym_identifier,
    ACTIONS(11521), 2,
      anon_sym___covariant,
      anon_sym___contravariant,
  [56341] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11105), 1,
      anon_sym_COLON,
    ACTIONS(11523), 1,
      anon_sym_RPAREN,
    STATE(6345), 1,
      sym_gnu_asm_goto_list,
  [56354] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8760), 1,
      anon_sym_LBRACE,
    ACTIONS(11525), 1,
      sym_identifier,
    STATE(3797), 1,
      sym_field_declaration_list,
  [56367] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8760), 1,
      anon_sym_LBRACE,
    ACTIONS(11527), 1,
      sym_identifier,
    STATE(3804), 1,
      sym_field_declaration_list,
  [56380] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2396), 1,
      anon_sym_while,
    ACTIONS(11529), 1,
      anon_sym_else,
    STATE(947), 1,
      sym_else_clause,
  [56393] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10118), 1,
      anon_sym_RPAREN,
    ACTIONS(11531), 1,
      anon_sym_COMMA,
    STATE(5585), 1,
      aux_sym_available_expression_repeat1,
  [56406] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11113), 1,
      anon_sym_COLON,
    ACTIONS(11534), 1,
      anon_sym_RPAREN,
    STATE(5635), 1,
      sym_gnu_asm_output_operand_list,
  [56419] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11536), 1,
      anon_sym_COMMA,
    ACTIONS(11538), 1,
      anon_sym_RBRACE,
    STATE(5639), 1,
      aux_sym_dictionary_literal_repeat1,
  [56432] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11189), 1,
      anon_sym_COMMA,
    ACTIONS(11540), 1,
      anon_sym_GT,
    STATE(5642), 1,
      aux_sym_argument_list_repeat2,
  [56445] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11457), 1,
      anon_sym_COMMA,
    ACTIONS(11509), 1,
      anon_sym_RPAREN,
    STATE(5553), 1,
      aux_sym_method_type_repeat1,
  [56458] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11229), 1,
      anon_sym_DOT,
    ACTIONS(11542), 1,
      anon_sym_SEMI,
    STATE(5462), 1,
      aux_sym_module_import_repeat1,
  [56471] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10045), 1,
      anon_sym_COMMA,
    ACTIONS(10047), 1,
      anon_sym_RPAREN,
    STATE(5535), 1,
      aux_sym_available_expression_repeat1,
  [56484] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(10021), 1,
      anon_sym_RPAREN,
    STATE(5643), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [56497] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11544), 1,
      anon_sym_COMMA,
    ACTIONS(11547), 1,
      anon_sym_RBRACK,
    STATE(5593), 1,
      aux_sym_attribute_declaration_repeat1,
  [56510] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(3070), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [56523] = 4,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11157), 1,
      anon_sym_LPAREN2,
    ACTIONS(11549), 1,
      aux_sym_preproc_include_token3,
    STATE(6607), 1,
      sym_preproc_argument_list,
  [56536] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10045), 1,
      anon_sym_COMMA,
    ACTIONS(10070), 1,
      anon_sym_RPAREN,
    STATE(5644), 1,
      aux_sym_available_expression_repeat1,
  [56549] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11296), 1,
      anon_sym_COMMA,
    ACTIONS(11551), 1,
      anon_sym_RPAREN,
    STATE(5480), 1,
      aux_sym_parameter_list_repeat1,
  [56562] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11553), 1,
      anon_sym_GT,
    STATE(5646), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [56575] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11555), 1,
      anon_sym_GT,
    STATE(5542), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [56588] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11559), 1,
      anon_sym_COLON,
    ACTIONS(11557), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [56599] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10045), 1,
      anon_sym_COMMA,
    ACTIONS(11561), 1,
      anon_sym_RPAREN,
    STATE(5585), 1,
      aux_sym_available_expression_repeat1,
  [56612] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(303), 1,
      anon_sym_LBRACE,
    ACTIONS(305), 1,
      anon_sym_ATautoreleasepool,
    STATE(395), 1,
      sym_compound_statement,
  [56625] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11563), 1,
      anon_sym_COMMA,
    ACTIONS(11566), 1,
      anon_sym_RPAREN,
    STATE(5603), 1,
      aux_sym_property_attributes_declaration_repeat1,
  [56638] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    ACTIONS(11568), 1,
      anon_sym_GT,
    STATE(5413), 1,
      aux_sym_parameterized_arguments_repeat2,
  [56651] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11229), 1,
      anon_sym_DOT,
    ACTIONS(11570), 1,
      anon_sym_SEMI,
    STATE(5626), 1,
      aux_sym_module_import_repeat1,
  [56664] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(1661), 1,
      sym_compound_statement,
  [56677] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(944), 1,
      anon_sym_LBRACE,
    ACTIONS(948), 1,
      anon_sym_ATautoreleasepool,
    STATE(811), 1,
      sym_compound_statement,
  [56690] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9416), 1,
      anon_sym_COMMA,
    ACTIONS(11572), 1,
      anon_sym_RPAREN,
    STATE(5551), 1,
      aux_sym_attribute_repeat1,
  [56703] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6966), 1,
      anon_sym_LBRACE,
    ACTIONS(11574), 1,
      sym_identifier,
    STATE(2924), 1,
      sym_field_declaration_list,
  [56716] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    STATE(5202), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(11576), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [56727] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    ACTIONS(11568), 1,
      anon_sym_GT,
    STATE(5664), 1,
      aux_sym_parameterized_arguments_repeat2,
  [56740] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11229), 1,
      anon_sym_DOT,
    ACTIONS(11578), 1,
      anon_sym_SEMI,
    STATE(5452), 1,
      aux_sym_module_import_repeat1,
  [56753] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11068), 1,
      anon_sym_SEMI,
    STATE(5654), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [56766] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4683), 1,
      anon_sym_LBRACE,
    ACTIONS(4685), 1,
      anon_sym_ATautoreleasepool,
    STATE(2309), 1,
      sym_compound_statement,
  [56779] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11580), 1,
      sym_identifier,
    ACTIONS(11582), 2,
      anon_sym___covariant,
      anon_sym___contravariant,
  [56790] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      anon_sym_ATautoreleasepool,
    STATE(926), 1,
      sym_compound_statement,
  [56803] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9452), 1,
      anon_sym_COMMA,
    ACTIONS(9454), 1,
      anon_sym_RPAREN,
    STATE(5475), 1,
      aux_sym_argument_list_repeat1,
  [56816] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10552), 1,
      anon_sym_GT,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    STATE(5413), 1,
      aux_sym_parameterized_arguments_repeat2,
  [56829] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11296), 1,
      anon_sym_COMMA,
    ACTIONS(11584), 1,
      anon_sym_RPAREN,
    STATE(5480), 1,
      aux_sym_parameter_list_repeat1,
  [56842] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9452), 1,
      anon_sym_COMMA,
    ACTIONS(9454), 1,
      anon_sym_RPAREN,
    STATE(5483), 1,
      aux_sym_argument_list_repeat1,
  [56855] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    STATE(1501), 1,
      sym_compound_statement,
  [56868] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      anon_sym_ATautoreleasepool,
    STATE(975), 1,
      sym_compound_statement,
  [56881] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9359), 1,
      sym_identifier,
    ACTIONS(11586), 1,
      anon_sym_RBRACK,
    STATE(5345), 1,
      aux_sym_message_expression_repeat2,
  [56894] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11189), 1,
      anon_sym_COMMA,
    ACTIONS(11588), 1,
      anon_sym_GT,
    STATE(5429), 1,
      aux_sym_argument_list_repeat2,
  [56907] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2107), 1,
      sym_compound_statement,
  [56920] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11229), 1,
      anon_sym_DOT,
    ACTIONS(11590), 1,
      anon_sym_SEMI,
    STATE(5462), 1,
      aux_sym_module_import_repeat1,
  [56933] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(3082), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [56946] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    ACTIONS(11592), 1,
      anon_sym_GT,
    STATE(5667), 1,
      aux_sym_parameterized_arguments_repeat2,
  [56959] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      anon_sym_ATautoreleasepool,
    STATE(787), 1,
      sym_compound_statement,
  [56972] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11139), 1,
      anon_sym_COLON,
    ACTIONS(11594), 1,
      anon_sym_RPAREN,
    STATE(5341), 1,
      sym_gnu_asm_input_operand_list,
  [56985] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11113), 1,
      anon_sym_COLON,
    ACTIONS(11596), 1,
      anon_sym_RPAREN,
    STATE(5342), 1,
      sym_gnu_asm_output_operand_list,
  [56998] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11598), 1,
      anon_sym_SEMI,
    STATE(5344), 1,
      aux_sym_class_declaration_repeat1,
  [57011] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11598), 1,
      anon_sym_SEMI,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [57024] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11163), 1,
      anon_sym_COMMA,
    ACTIONS(11600), 1,
      anon_sym_RPAREN,
    STATE(5564), 1,
      aux_sym_generic_arguments_repeat1,
  [57037] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11139), 1,
      anon_sym_COLON,
    ACTIONS(11602), 1,
      anon_sym_RPAREN,
    STATE(5676), 1,
      sym_gnu_asm_input_operand_list,
  [57050] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11113), 1,
      anon_sym_COLON,
    ACTIONS(11604), 1,
      anon_sym_RPAREN,
    STATE(5677), 1,
      sym_gnu_asm_output_operand_list,
  [57063] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    ACTIONS(7289), 1,
      anon_sym_SEMI,
    STATE(2790), 1,
      sym_method_parameter,
  [57076] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8567), 1,
      anon_sym_RBRACK,
    ACTIONS(11606), 1,
      anon_sym_COMMA,
    STATE(5703), 1,
      aux_sym_attribute_repeat1,
  [57089] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8545), 1,
      anon_sym_RBRACE,
    ACTIONS(11608), 1,
      anon_sym_COMMA,
    STATE(5525), 1,
      aux_sym_dictionary_literal_repeat1,
  [57102] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11610), 1,
      anon_sym_SEMI,
    STATE(5679), 1,
      aux_sym_class_declaration_repeat1,
  [57115] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11610), 1,
      anon_sym_SEMI,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [57128] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11189), 1,
      anon_sym_COMMA,
    ACTIONS(11612), 1,
      anon_sym_GT,
    STATE(5429), 1,
      aux_sym_argument_list_repeat2,
  [57141] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(11614), 1,
      anon_sym_RPAREN,
    STATE(5421), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [57154] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10045), 1,
      anon_sym_COMMA,
    ACTIONS(11616), 1,
      anon_sym_RPAREN,
    STATE(5585), 1,
      aux_sym_available_expression_repeat1,
  [57167] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10045), 1,
      anon_sym_COMMA,
    ACTIONS(11616), 1,
      anon_sym_RPAREN,
    STATE(5681), 1,
      aux_sym_available_expression_repeat1,
  [57180] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11618), 1,
      anon_sym_GT,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [57193] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11457), 1,
      anon_sym_COMMA,
    ACTIONS(11620), 1,
      anon_sym_RPAREN,
    STATE(5589), 1,
      aux_sym_method_type_repeat1,
  [57206] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(10889), 1,
      anon_sym_SEMI,
    STATE(5671), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [57219] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    ACTIONS(11592), 1,
      anon_sym_GT,
    STATE(5413), 1,
      aux_sym_parameterized_arguments_repeat2,
  [57232] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11622), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [57245] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9412), 3,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_COLON,
  [57254] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11624), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [57267] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1934), 1,
      anon_sym_LBRACE,
    ACTIONS(1936), 1,
      anon_sym_ATautoreleasepool,
    STATE(757), 1,
      sym_compound_statement,
  [57280] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11624), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [57293] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11624), 1,
      anon_sym_SEMI,
    STATE(5694), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [57306] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11626), 3,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_COLON,
  [57315] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11628), 1,
      anon_sym_SEMI,
    STATE(5678), 1,
      aux_sym_class_declaration_repeat1,
  [57328] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11628), 1,
      anon_sym_SEMI,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [57341] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11630), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [57354] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9767), 1,
      anon_sym_COMMA,
    ACTIONS(11632), 1,
      anon_sym_RPAREN,
    STATE(5520), 1,
      aux_sym_preproc_argument_list_repeat1,
  [57367] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9448), 1,
      anon_sym_COMMA,
    ACTIONS(9450), 1,
      anon_sym_RBRACE,
    STATE(5697), 1,
      aux_sym_initializer_list_repeat1,
  [57380] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    STATE(1483), 1,
      sym_compound_statement,
  [57393] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      anon_sym_ATautoreleasepool,
    STATE(811), 1,
      sym_compound_statement,
  [57406] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    ACTIONS(11634), 1,
      anon_sym_GT,
    STATE(5413), 1,
      aux_sym_parameterized_arguments_repeat2,
  [57419] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9010), 1,
      anon_sym_COMMA,
    ACTIONS(9022), 1,
      anon_sym_RPAREN,
    STATE(5574), 1,
      aux_sym_attribute_specifier_repeat1,
  [57432] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(11636), 1,
      sym_identifier,
    STATE(5787), 1,
      sym_variadic_parameter,
  [57445] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    ACTIONS(11638), 1,
      anon_sym_GT,
    STATE(5413), 1,
      aux_sym_parameterized_arguments_repeat2,
  [57458] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11197), 1,
      anon_sym_COMMA,
    ACTIONS(11638), 1,
      anon_sym_GT,
    STATE(5383), 1,
      aux_sym_parameterized_arguments_repeat2,
  [57471] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11640), 3,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_DOT,
  [57480] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11642), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [57493] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11642), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [57506] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11642), 1,
      anon_sym_SEMI,
    STATE(5686), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [57519] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5308), 1,
      anon_sym_LPAREN2,
    ACTIONS(10445), 1,
      sym_identifier,
    STATE(6496), 1,
      sym_method_type,
  [57532] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11644), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [57545] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1934), 1,
      anon_sym_LBRACE,
    ACTIONS(1936), 1,
      anon_sym_ATautoreleasepool,
    STATE(785), 1,
      sym_compound_statement,
  [57558] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11135), 1,
      anon_sym_COLON,
    ACTIONS(11646), 1,
      anon_sym_RPAREN,
    STATE(5704), 1,
      sym_gnu_asm_clobber_list,
  [57571] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11139), 1,
      anon_sym_COLON,
    ACTIONS(11648), 1,
      anon_sym_RPAREN,
    STATE(5705), 1,
      sym_gnu_asm_input_operand_list,
  [57584] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11650), 1,
      anon_sym_SEMI,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [57597] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_COMMA,
    ACTIONS(11652), 1,
      anon_sym_SEMI,
    STATE(5486), 1,
      aux_sym_class_declaration_repeat1,
  [57610] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(11654), 1,
      anon_sym_RPAREN,
    STATE(5707), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [57623] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10045), 1,
      anon_sym_COMMA,
    ACTIONS(11656), 1,
      anon_sym_RPAREN,
    STATE(5585), 1,
      aux_sym_available_expression_repeat1,
  [57636] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11660), 1,
      anon_sym_RPAREN,
    ACTIONS(11658), 2,
      anon_sym_DOT_DOT_DOT,
      sym_identifier,
  [57647] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9571), 1,
      anon_sym_RBRACE,
    ACTIONS(11662), 1,
      anon_sym_COMMA,
    STATE(5683), 1,
      aux_sym_initializer_list_repeat1,
  [57660] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11665), 3,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_DOT,
  [57669] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11667), 1,
      anon_sym_SEMI,
    STATE(5690), 1,
      aux_sym_property_implementation_repeat1,
  [57682] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11667), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [57695] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11296), 1,
      anon_sym_COMMA,
    ACTIONS(11669), 1,
      anon_sym_RPAREN,
    STATE(5480), 1,
      aux_sym_parameter_list_repeat1,
  [57708] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1934), 1,
      anon_sym_LBRACE,
    ACTIONS(1936), 1,
      anon_sym_ATautoreleasepool,
    STATE(791), 1,
      sym_compound_statement,
  [57721] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2140), 1,
      sym_compound_statement,
  [57734] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11671), 1,
      anon_sym_SEMI,
    STATE(5495), 1,
      aux_sym_property_implementation_repeat1,
  [57747] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8281), 1,
      anon_sym_RBRACE,
    ACTIONS(11673), 1,
      anon_sym_COMMA,
    STATE(5683), 1,
      aux_sym_initializer_list_repeat1,
  [57760] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11135), 1,
      anon_sym_COLON,
    ACTIONS(11675), 1,
      anon_sym_RPAREN,
    STATE(5485), 1,
      sym_gnu_asm_clobber_list,
  [57773] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10606), 1,
      anon_sym_COMMA,
    ACTIONS(11677), 1,
      anon_sym_SEMI,
    STATE(5426), 1,
      aux_sym_property_implementation_repeat1,
  [57786] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(11677), 1,
      anon_sym_SEMI,
    STATE(5272), 1,
      aux_sym_protocol_forward_declaration_repeat1,
  [57799] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11139), 1,
      anon_sym_COLON,
    ACTIONS(11679), 1,
      anon_sym_RPAREN,
    STATE(5490), 1,
      sym_gnu_asm_input_operand_list,
  [57812] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(1652), 1,
      sym_compound_statement,
  [57825] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8277), 1,
      anon_sym_RBRACE,
    ACTIONS(11681), 1,
      anon_sym_COMMA,
    STATE(5683), 1,
      aux_sym_initializer_list_repeat1,
  [57838] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8127), 1,
      anon_sym_LBRACE,
    ACTIONS(8129), 1,
      anon_sym_ATautoreleasepool,
    STATE(1486), 1,
      sym_compound_statement,
  [57851] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10614), 1,
      anon_sym_LBRACE,
    ACTIONS(10616), 1,
      anon_sym_ATautoreleasepool,
    STATE(191), 1,
      sym_compound_statement,
  [57864] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      anon_sym_ATautoreleasepool,
    STATE(815), 1,
      sym_compound_statement,
  [57877] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11683), 1,
      anon_sym_COMMA,
    ACTIONS(11685), 1,
      anon_sym_RPAREN,
    STATE(5519), 1,
      aux_sym__old_style_parameter_list_repeat1,
  [57890] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11219), 1,
      anon_sym_COMMA,
    ACTIONS(11687), 1,
      anon_sym_RBRACK,
    STATE(5593), 1,
      aux_sym_attribute_declaration_repeat1,
  [57903] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9408), 1,
      anon_sym_RBRACK,
    ACTIONS(11689), 1,
      anon_sym_COMMA,
    STATE(5703), 1,
      aux_sym_attribute_repeat1,
  [57916] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11105), 1,
      anon_sym_COLON,
    ACTIONS(11692), 1,
      anon_sym_RPAREN,
    STATE(6417), 1,
      sym_gnu_asm_goto_list,
  [57929] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11135), 1,
      anon_sym_COLON,
    ACTIONS(11694), 1,
      anon_sym_RPAREN,
    STATE(5320), 1,
      sym_gnu_asm_clobber_list,
  [57942] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10827), 1,
      anon_sym_LBRACE,
    ACTIONS(10829), 1,
      anon_sym_ATautoreleasepool,
    STATE(409), 1,
      sym_compound_statement,
  [57955] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(11696), 1,
      anon_sym_RPAREN,
    STATE(5421), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [57968] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10013), 1,
      anon_sym_COMMA,
    ACTIONS(11696), 1,
      anon_sym_RPAREN,
    STATE(5340), 1,
      aux_sym_availability_attribute_specifier_repeat1,
  [57981] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(1655), 1,
      sym_compound_statement,
  [57994] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4659), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_ATautoreleasepool,
    STATE(2186), 1,
      sym_compound_statement,
  [58007] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8103), 1,
      anon_sym_LBRACE,
    ACTIONS(8105), 1,
      anon_sym_ATautoreleasepool,
    STATE(1640), 1,
      sym_compound_statement,
  [58020] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(3192), 1,
      sym_enumerator_list,
  [58030] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11698), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58038] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10967), 1,
      anon_sym_RPAREN,
    STATE(1438), 1,
      aux_sym_selector_expression_repeat2,
  [58048] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11547), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [58056] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2440), 1,
      sym_parenthesized_expression,
  [58066] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(4143), 1,
      sym_enumerator_list,
  [58076] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11702), 1,
      sym_identifier,
    ACTIONS(11704), 1,
      sym_primitive_type,
  [58086] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11706), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11708), 1,
      sym_preproc_arg,
  [58096] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(5315), 1,
      sym_parenthesized_expression,
  [58106] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2448), 1,
      sym_parenthesized_expression,
  [58116] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2441), 1,
      sym_parenthesized_expression,
  [58126] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11712), 2,
      sym_number_literal,
      sym_identifier,
  [58134] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11714), 1,
      anon_sym_LPAREN2,
    STATE(6114), 1,
      sym_argument_list,
  [58144] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8042), 1,
      anon_sym_COMMA,
    ACTIONS(11716), 1,
      anon_sym_EQ,
  [58154] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10668), 1,
      anon_sym_RPAREN,
    STATE(1622), 1,
      aux_sym_selector_expression_repeat2,
  [58164] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(6526), 1,
      sym_parenthesized_expression,
  [58174] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11718), 1,
      sym_identifier,
    ACTIONS(11720), 1,
      anon_sym_LPARENclass_RPAREN,
  [58184] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    STATE(4095), 1,
      sym_enumerator_list,
  [58194] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11722), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11724), 1,
      sym_number_literal,
  [58204] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11726), 1,
      anon_sym_LPAREN2,
    STATE(4973), 1,
      aux_sym_selector_expression_repeat1,
  [58214] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11728), 1,
      sym_identifier,
    ACTIONS(11730), 1,
      sym_primitive_type,
  [58224] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11732), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11734), 1,
      sym_preproc_arg,
  [58234] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11736), 1,
      sym_identifier,
    ACTIONS(11738), 1,
      anon_sym_LPAREN2,
  [58244] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(6670), 1,
      sym_parenthesized_expression,
  [58254] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11041), 1,
      anon_sym_RPAREN,
    STATE(4470), 1,
      aux_sym_selector_expression_repeat2,
  [58264] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11740), 1,
      sym_identifier,
    STATE(5467), 1,
      sym_attribute,
  [58274] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11742), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11744), 1,
      sym_preproc_arg,
  [58284] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11746), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11748), 1,
      sym_preproc_arg,
  [58294] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    STATE(3888), 1,
      sym_enumerator_list,
  [58304] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11750), 1,
      sym_identifier,
    ACTIONS(11752), 1,
      anon_sym_LPARENclass_RPAREN,
  [58314] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11754), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11756), 1,
      sym_number_literal,
  [58324] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11758), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11760), 1,
      sym_preproc_arg,
  [58334] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11566), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58342] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11172), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58350] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11762), 1,
      anon_sym_COLON2,
    STATE(5142), 1,
      aux_sym_method_identifier_repeat1,
  [58360] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11740), 1,
      sym_identifier,
    STATE(5496), 1,
      sym_attribute,
  [58370] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(6511), 1,
      sym_parenthesized_expression,
  [58380] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11030), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11032), 1,
      sym_preproc_arg,
  [58390] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11764), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [58398] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(5343), 1,
      sym_parenthesized_expression,
  [58408] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11766), 1,
      anon_sym_LPAREN2,
    STATE(6301), 1,
      sym_argument_list,
  [58418] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11768), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [58426] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10920), 1,
      anon_sym_RPAREN,
    STATE(4344), 1,
      aux_sym_selector_expression_repeat2,
  [58436] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(6458), 1,
      sym_parenthesized_expression,
  [58446] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10118), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58454] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10786), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(10788), 1,
      sym_preproc_arg,
  [58464] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9052), 1,
      anon_sym_LBRACK,
    ACTIONS(11770), 1,
      anon_sym_RPAREN,
  [58474] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11772), 1,
      sym_identifier,
    ACTIONS(11774), 1,
      anon_sym_RPAREN,
  [58484] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11776), 1,
      anon_sym_LPAREN2,
    STATE(4951), 1,
      aux_sym_selector_expression_repeat1,
  [58494] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2992), 1,
      anon_sym_SEMI,
    ACTIONS(11778), 1,
      anon_sym_EQ,
  [58504] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11780), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58512] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11782), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [58520] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11784), 1,
      anon_sym_LPAREN2,
    STATE(5954), 1,
      sym_argument_list,
  [58530] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11786), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [58538] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11788), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11790), 1,
      sym_preproc_arg,
  [58548] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11792), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [58556] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11794), 1,
      sym_identifier,
    ACTIONS(11796), 1,
      sym_primitive_type,
  [58566] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    STATE(3953), 1,
      sym_enumerator_list,
  [58576] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2449), 1,
      sym_parenthesized_expression,
  [58586] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11798), 2,
      sym_number_literal,
      sym_identifier,
  [58594] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9052), 1,
      anon_sym_LBRACK,
    ACTIONS(11800), 1,
      anon_sym_RPAREN,
  [58604] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11802), 1,
      sym_identifier,
    ACTIONS(11804), 1,
      sym_primitive_type,
  [58614] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11806), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58622] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(5537), 1,
      sym_parenthesized_expression,
  [58632] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(3189), 1,
      sym_enumerator_list,
  [58642] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11808), 1,
      sym_identifier,
    ACTIONS(11810), 1,
      sym_primitive_type,
  [58652] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11812), 1,
      sym_identifier,
    ACTIONS(11814), 1,
      sym_primitive_type,
  [58662] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11428), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58670] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11816), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58678] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11183), 1,
      sym_identifier,
    STATE(5744), 1,
      sym_property_attribute,
  [58688] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2455), 1,
      sym_parenthesized_expression,
  [58698] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11818), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11820), 1,
      sym_number_literal,
  [58708] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11327), 2,
      anon_sym_SEMI,
      anon_sym_DOT,
  [58716] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11822), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [58724] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11824), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11826), 1,
      sym_preproc_arg,
  [58734] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11421), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58742] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(4155), 1,
      sym_enumerator_list,
  [58752] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11828), 1,
      sym_identifier,
    ACTIONS(11830), 1,
      sym_primitive_type,
  [58762] = 3,
    ACTIONS(4262), 1,
      sym_number_literal,
    ACTIONS(4264), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(9865), 1,
      sym_comment,
  [58772] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(6390), 1,
      sym_parenthesized_expression,
  [58782] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11832), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11834), 1,
      sym_preproc_arg,
  [58792] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11836), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58800] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10029), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58808] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11839), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11841), 1,
      sym_number_literal,
  [58818] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11843), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [58826] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    STATE(4075), 1,
      sym_enumerator_list,
  [58836] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11845), 1,
      sym_identifier,
    ACTIONS(11847), 1,
      sym_primitive_type,
  [58846] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11849), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [58854] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11851), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58862] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11853), 1,
      anon_sym_LPAREN2,
    STATE(4975), 1,
      aux_sym_selector_expression_repeat1,
  [58872] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11151), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58880] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2426), 1,
      sym_parenthesized_expression,
  [58890] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11855), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11857), 1,
      sym_number_literal,
  [58900] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(5466), 1,
      sym_parenthesized_expression,
  [58910] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_COLON,
    STATE(2808), 1,
      sym_method_parameter,
  [58920] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11364), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58928] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2431), 1,
      sym_parenthesized_expression,
  [58938] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9516), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58946] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11859), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11861), 1,
      sym_number_literal,
  [58956] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11863), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [58964] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11865), 1,
      anon_sym_LPAREN2,
    STATE(6769), 1,
      sym_argument_list,
  [58974] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11867), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11869), 1,
      sym_preproc_arg,
  [58984] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11871), 1,
      sym_identifier,
    ACTIONS(11873), 1,
      anon_sym_STAR,
  [58994] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11875), 1,
      sym_identifier,
    ACTIONS(11877), 1,
      anon_sym_STAR,
  [59004] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2463), 1,
      sym_parenthesized_expression,
  [59014] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11879), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11881), 1,
      sym_preproc_arg,
  [59024] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11883), 1,
      anon_sym_LPAREN2,
    STATE(6040), 1,
      sym_argument_list,
  [59034] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(5622), 1,
      sym_parenthesized_expression,
  [59044] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11885), 1,
      sym_identifier,
    ACTIONS(11887), 1,
      sym_primitive_type,
  [59054] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11889), 1,
      anon_sym_LPAREN2,
    STATE(6331), 1,
      sym_argument_list,
  [59064] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11891), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11893), 1,
      sym_preproc_arg,
  [59074] = 3,
    ACTIONS(4118), 1,
      sym_number_literal,
    ACTIONS(4120), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(9865), 1,
      sym_comment,
  [59084] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9553), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [59092] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11895), 1,
      sym_identifier,
    ACTIONS(11897), 1,
      sym_primitive_type,
  [59102] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11280), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [59110] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11899), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [59118] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    STATE(3882), 1,
      sym_enumerator_list,
  [59128] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(10980), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(10982), 1,
      sym_preproc_arg,
  [59138] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11901), 1,
      sym_identifier,
    ACTIONS(11903), 1,
      sym_primitive_type,
  [59148] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11905), 1,
      anon_sym_LPAREN2,
    STATE(4981), 1,
      aux_sym_selector_expression_repeat1,
  [59158] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11907), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11909), 1,
      sym_number_literal,
  [59168] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11911), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [59176] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11913), 1,
      sym_identifier,
    ACTIONS(11915), 1,
      anon_sym_LPARENclass_RPAREN,
  [59186] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11740), 1,
      sym_identifier,
    STATE(5444), 1,
      sym_attribute,
  [59196] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11740), 1,
      sym_identifier,
    STATE(5484), 1,
      sym_attribute,
  [59206] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9359), 1,
      sym_identifier,
    STATE(5578), 1,
      aux_sym_message_expression_repeat2,
  [59216] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11917), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11919), 1,
      sym_preproc_arg,
  [59226] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(6182), 1,
      sym_parenthesized_expression,
  [59236] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2430), 1,
      sym_parenthesized_expression,
  [59246] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11482), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [59254] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9359), 1,
      sym_identifier,
    STATE(5623), 1,
      aux_sym_message_expression_repeat2,
  [59264] = 3,
    ACTIONS(4258), 1,
      sym_number_literal,
    ACTIONS(4260), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(9865), 1,
      sym_comment,
  [59274] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2437), 1,
      sym_parenthesized_expression,
  [59284] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11774), 1,
      anon_sym_RPAREN,
    ACTIONS(11921), 1,
      sym_identifier,
  [59294] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8794), 1,
      anon_sym_LBRACE,
    STATE(4088), 1,
      sym_enumerator_list,
  [59304] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9052), 1,
      anon_sym_LBRACK,
    ACTIONS(11923), 1,
      anon_sym_RPAREN,
  [59314] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9569), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [59322] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(4144), 1,
      sym_enumerator_list,
  [59332] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11925), 1,
      sym_identifier,
    ACTIONS(11927), 1,
      sym_primitive_type,
  [59342] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11929), 1,
      sym_identifier,
    ACTIONS(11931), 1,
      sym_primitive_type,
  [59352] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9571), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [59360] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11933), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11935), 1,
      sym_preproc_arg,
  [59370] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11937), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11939), 1,
      sym_number_literal,
  [59380] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11941), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [59388] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10019), 1,
      anon_sym_COLON2,
    STATE(5035), 1,
      aux_sym_method_identifier_repeat1,
  [59398] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11943), 1,
      sym_identifier,
    ACTIONS(11945), 1,
      sym_primitive_type,
  [59408] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11947), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11949), 1,
      sym_number_literal,
  [59418] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11951), 1,
      sym_identifier,
    ACTIONS(11953), 1,
      anon_sym_LPAREN2,
  [59428] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7519), 1,
      anon_sym_LBRACE,
    STATE(3135), 1,
      sym_enumerator_list,
  [59438] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9052), 1,
      anon_sym_LBRACK,
    ACTIONS(11955), 1,
      anon_sym_RPAREN,
  [59448] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11957), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [59456] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2462), 1,
      sym_parenthesized_expression,
  [59466] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11959), 1,
      sym_identifier,
    ACTIONS(11961), 1,
      anon_sym_STAR,
  [59476] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11740), 1,
      sym_identifier,
    STATE(5715), 1,
      sym_attribute,
  [59486] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2471), 1,
      sym_parenthesized_expression,
  [59496] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9575), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [59504] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(5602), 1,
      sym_parenthesized_expression,
  [59514] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2453), 1,
      sym_parenthesized_expression,
  [59524] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11963), 1,
      sym_identifier,
    ACTIONS(11965), 1,
      anon_sym_LPARENclass_RPAREN,
  [59534] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11967), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11969), 1,
      sym_preproc_arg,
  [59544] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_LPAREN2,
    STATE(6218), 1,
      sym_argument_list,
  [59554] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11971), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11973), 1,
      sym_number_literal,
  [59564] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11975), 1,
      anon_sym_LPAREN2,
    STATE(6475), 1,
      sym_argument_list,
  [59574] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2439), 1,
      sym_parenthesized_expression,
  [59584] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11977), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [59592] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11979), 1,
      sym_identifier,
    ACTIONS(11981), 1,
      sym_primitive_type,
  [59602] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11983), 1,
      sym_identifier,
    ACTIONS(11985), 1,
      anon_sym_LPARENclass_RPAREN,
  [59612] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11987), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11989), 1,
      sym_preproc_arg,
  [59622] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11991), 1,
      sym_identifier,
    ACTIONS(11993), 1,
      anon_sym_RPAREN,
  [59632] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11995), 2,
      anon_sym_DOT_DOT_DOT,
      sym_identifier,
  [59640] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11997), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(11999), 1,
      sym_preproc_arg,
  [59650] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12001), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(12003), 1,
      sym_preproc_arg,
  [59660] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12005), 1,
      sym_identifier,
    ACTIONS(12007), 1,
      anon_sym_LPARENclass_RPAREN,
  [59670] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12009), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(12011), 1,
      sym_preproc_arg,
  [59680] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2433), 1,
      sym_parenthesized_expression,
  [59690] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12013), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [59698] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12015), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [59706] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(6598), 1,
      sym_parenthesized_expression,
  [59716] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12017), 2,
      anon_sym_noreturn,
      anon_sym_nothrow,
  [59724] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12019), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(12021), 1,
      sym_preproc_arg,
  [59734] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9567), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [59742] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11710), 1,
      anon_sym_LPAREN2,
    STATE(5393), 1,
      sym_parenthesized_expression,
  [59752] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12023), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(12025), 1,
      sym_preproc_arg,
  [59762] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12027), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(12029), 1,
      sym_preproc_arg,
  [59772] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11700), 1,
      anon_sym_LPAREN2,
    STATE(2434), 1,
      sym_parenthesized_expression,
  [59782] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12031), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [59790] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11437), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [59798] = 3,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12033), 1,
      aux_sym_preproc_include_token3,
    ACTIONS(12035), 1,
      sym_preproc_arg,
  [59808] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12037), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [59816] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9595), 1,
      anon_sym_RPAREN,
  [59823] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12039), 1,
      anon_sym_SEMI,
  [59830] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9289), 1,
      anon_sym_RPAREN,
  [59837] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12041), 1,
      anon_sym_RPAREN,
  [59844] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12043), 1,
      anon_sym_SEMI,
  [59851] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12045), 1,
      sym_identifier,
  [59858] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12047), 1,
      aux_sym_preproc_if_token2,
  [59865] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12049), 1,
      anon_sym_SEMI,
  [59872] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12051), 1,
      anon_sym_SEMI,
  [59879] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12053), 1,
      aux_sym_preproc_if_token2,
  [59886] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12055), 1,
      aux_sym_preproc_if_token2,
  [59893] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12057), 1,
      anon_sym_SEMI,
  [59900] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12059), 1,
      anon_sym_SEMI,
  [59907] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12061), 1,
      anon_sym_COLON,
  [59914] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9522), 1,
      anon_sym_SEMI,
  [59921] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12063), 1,
      sym_identifier,
  [59928] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12065), 1,
      sym_identifier,
  [59935] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12067), 1,
      aux_sym_preproc_include_token3,
  [59942] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12069), 1,
      anon_sym_RPAREN,
  [59949] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12071), 1,
      anon_sym_LBRACE,
  [59956] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12073), 1,
      anon_sym_RPAREN,
  [59963] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12075), 1,
      anon_sym_SEMI,
  [59970] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12077), 1,
      aux_sym_preproc_include_token3,
  [59977] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12079), 1,
      anon_sym_SEMI,
  [59984] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12081), 1,
      sym_identifier,
  [59991] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12083), 1,
      aux_sym_preproc_if_token2,
  [59998] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12085), 1,
      sym_identifier,
  [60005] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12087), 1,
      sym_identifier,
  [60012] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12089), 1,
      sym_identifier,
  [60019] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12091), 1,
      sym_identifier,
  [60026] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12093), 1,
      anon_sym_LPAREN2,
  [60033] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12095), 1,
      anon_sym_DQUOTE,
  [60040] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9520), 1,
      anon_sym_RPAREN,
  [60047] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12097), 1,
      anon_sym_RPAREN,
  [60054] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12099), 1,
      anon_sym_SEMI,
  [60061] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12101), 1,
      anon_sym_RPAREN,
  [60068] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12103), 1,
      anon_sym_RPAREN,
  [60075] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12105), 1,
      sym_identifier,
  [60082] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12107), 1,
      anon_sym_SEMI,
  [60089] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12109), 1,
      anon_sym_SEMI,
  [60096] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12111), 1,
      sym_identifier,
  [60103] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12113), 1,
      aux_sym_preproc_include_token3,
  [60110] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9619), 1,
      anon_sym_RPAREN,
  [60117] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12115), 1,
      anon_sym_LPAREN2,
  [60124] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12117), 1,
      anon_sym_SEMI,
  [60131] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12119), 1,
      sym_identifier,
  [60138] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9468), 1,
      anon_sym_RPAREN,
  [60145] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12121), 1,
      sym_identifier,
  [60152] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12123), 1,
      sym_identifier,
  [60159] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12125), 1,
      aux_sym_preproc_if_token2,
  [60166] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12127), 1,
      sym_identifier,
  [60173] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9526), 1,
      anon_sym_RPAREN,
  [60180] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12129), 1,
      aux_sym_preproc_include_token3,
  [60187] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12131), 1,
      anon_sym_RPAREN,
  [60194] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12133), 1,
      aux_sym_preproc_if_token2,
  [60201] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12135), 1,
      aux_sym_ms_asm_block_token1,
  [60208] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12137), 1,
      sym_identifier,
  [60215] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9528), 1,
      anon_sym_SEMI,
  [60222] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12139), 1,
      anon_sym_SEMI,
  [60229] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12141), 1,
      anon_sym_RPAREN,
  [60236] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12143), 1,
      anon_sym_SEMI,
  [60243] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12145), 1,
      anon_sym_RPAREN,
  [60250] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12147), 1,
      anon_sym_SEMI,
  [60257] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12149), 1,
      anon_sym_RPAREN,
  [60264] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12151), 1,
      sym_identifier,
  [60271] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12153), 1,
      anon_sym_SEMI,
  [60278] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12155), 1,
      anon_sym_SEMI,
  [60285] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9538), 1,
      anon_sym_RPAREN,
  [60292] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12157), 1,
      anon_sym_COLON,
  [60299] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12159), 1,
      anon_sym_SEMI,
  [60306] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12161), 1,
      anon_sym_SEMI,
  [60313] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12163), 1,
      anon_sym_SEMI,
  [60320] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12165), 1,
      sym_identifier,
  [60327] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12167), 1,
      sym_identifier,
  [60334] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12169), 1,
      aux_sym_preproc_include_token3,
  [60341] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12171), 1,
      anon_sym_SEMI,
  [60348] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12173), 1,
      anon_sym_SEMI,
  [60355] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12175), 1,
      anon_sym_SEMI,
  [60362] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12177), 1,
      sym_identifier,
  [60369] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12179), 1,
      anon_sym_SEMI,
  [60376] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12181), 1,
      anon_sym_RPAREN,
  [60383] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12183), 1,
      anon_sym_SEMI,
  [60390] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12185), 1,
      sym_identifier,
  [60397] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12187), 1,
      anon_sym_SEMI,
  [60404] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12189), 1,
      anon_sym_RBRACK,
  [60411] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12191), 1,
      sym_identifier,
  [60418] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12193), 1,
      sym_identifier,
  [60425] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12195), 1,
      anon_sym_SEMI,
  [60432] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12197), 1,
      anon_sym_SEMI,
  [60439] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12199), 1,
      anon_sym_LBRACE,
  [60446] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12201), 1,
      anon_sym_SEMI,
  [60453] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12203), 1,
      anon_sym_SEMI,
  [60460] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12205), 1,
      anon_sym_SEMI,
  [60467] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12207), 1,
      anon_sym_SEMI,
  [60474] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12209), 1,
      sym_identifier,
  [60481] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12211), 1,
      sym_identifier,
  [60488] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12213), 1,
      sym_identifier,
  [60495] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12215), 1,
      sym_identifier,
  [60502] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12217), 1,
      sym_identifier,
  [60509] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12219), 1,
      anon_sym_RPAREN,
  [60516] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12221), 1,
      anon_sym_DQUOTE,
  [60523] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12223), 1,
      anon_sym_SEMI,
  [60530] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12225), 1,
      anon_sym_RPAREN,
  [60537] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12227), 1,
      anon_sym_SEMI,
  [60544] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12229), 1,
      anon_sym_SEMI,
  [60551] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12231), 1,
      anon_sym_RPAREN,
  [60558] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12233), 1,
      sym_identifier,
  [60565] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12235), 1,
      sym_identifier,
  [60572] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12237), 1,
      anon_sym_RPAREN,
  [60579] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12239), 1,
      sym_identifier,
  [60586] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12241), 1,
      anon_sym_LPAREN2,
  [60593] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12243), 1,
      anon_sym_SEMI,
  [60600] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12245), 1,
      sym_identifier,
  [60607] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9287), 1,
      anon_sym_RPAREN,
  [60614] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12247), 1,
      sym_identifier,
  [60621] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12249), 1,
      anon_sym_RPAREN,
  [60628] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12251), 1,
      anon_sym_SEMI,
  [60635] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12253), 1,
      anon_sym_RBRACK,
  [60642] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12255), 1,
      anon_sym_RPAREN,
  [60649] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12257), 1,
      aux_sym_ms_asm_block_token1,
  [60656] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12259), 1,
      anon_sym_SEMI,
  [60663] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12261), 1,
      aux_sym_preproc_if_token2,
  [60670] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12263), 1,
      anon_sym_RPAREN,
  [60677] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12265), 1,
      sym_identifier,
  [60684] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12267), 1,
      anon_sym_SEMI,
  [60691] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12269), 1,
      sym_identifier,
  [60698] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12271), 1,
      aux_sym_preproc_include_token3,
  [60705] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12273), 1,
      anon_sym_SEMI,
  [60712] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12275), 1,
      anon_sym_SEMI,
  [60719] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12277), 1,
      anon_sym_RPAREN,
  [60726] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12279), 1,
      anon_sym_SEMI,
  [60733] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12281), 1,
      anon_sym_LPAREN2,
  [60740] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12283), 1,
      aux_sym_preproc_if_token2,
  [60747] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12285), 1,
      sym_identifier,
  [60754] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12287), 1,
      anon_sym_SEMI,
  [60761] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12289), 1,
      anon_sym_SEMI,
  [60768] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12291), 1,
      anon_sym_SEMI,
  [60775] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12293), 1,
      anon_sym_SEMI,
  [60782] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12295), 1,
      anon_sym_SEMI,
  [60789] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12297), 1,
      anon_sym_RPAREN,
  [60796] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12299), 1,
      anon_sym_RPAREN,
  [60803] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12301), 1,
      anon_sym_SEMI,
  [60810] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11387), 1,
      aux_sym_preproc_include_token3,
  [60817] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12303), 1,
      sym_identifier,
  [60824] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12305), 1,
      sym_identifier,
  [60831] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12307), 1,
      anon_sym_LBRACE,
  [60838] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12309), 1,
      anon_sym_SEMI,
  [60845] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12311), 1,
      anon_sym_SEMI,
  [60852] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12313), 1,
      anon_sym_SEMI,
  [60859] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12315), 1,
      anon_sym_SEMI,
  [60866] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12317), 1,
      sym_identifier,
  [60873] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12319), 1,
      sym_identifier,
  [60880] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12321), 1,
      anon_sym_DQUOTE,
  [60887] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12323), 1,
      anon_sym_DQUOTE,
  [60894] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12325), 1,
      anon_sym_SEMI,
  [60901] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12327), 1,
      anon_sym_SEMI,
  [60908] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12329), 1,
      anon_sym_SEMI,
  [60915] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12331), 1,
      sym_identifier,
  [60922] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12333), 1,
      anon_sym_SEMI,
  [60929] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12335), 1,
      sym_identifier,
  [60936] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12337), 1,
      anon_sym_SEMI,
  [60943] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9685), 1,
      anon_sym_RPAREN,
  [60950] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12339), 1,
      anon_sym_COLON,
  [60957] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12341), 1,
      anon_sym_SEMI,
  [60964] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12343), 1,
      anon_sym_SEMI,
  [60971] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12345), 1,
      aux_sym_ms_asm_block_token1,
  [60978] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12347), 1,
      aux_sym_preproc_if_token2,
  [60985] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12349), 1,
      anon_sym_SEMI,
  [60992] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12351), 1,
      anon_sym_RPAREN,
  [60999] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12353), 1,
      anon_sym_SEMI,
  [61006] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12355), 1,
      sym_identifier,
  [61013] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12357), 1,
      anon_sym_RPAREN,
  [61020] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12359), 1,
      anon_sym_SEMI,
  [61027] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12361), 1,
      anon_sym_SEMI,
  [61034] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12363), 1,
      sym_identifier,
  [61041] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12365), 1,
      anon_sym_SEMI,
  [61048] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12367), 1,
      anon_sym_RPAREN,
  [61055] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12369), 1,
      anon_sym_RPAREN,
  [61062] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12371), 1,
      sym_identifier,
  [61069] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12373), 1,
      anon_sym_while,
  [61076] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12375), 1,
      anon_sym_SEMI,
  [61083] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12377), 1,
      anon_sym_SEMI,
  [61090] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12379), 1,
      anon_sym_SEMI,
  [61097] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12381), 1,
      anon_sym_SEMI,
  [61104] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12383), 1,
      anon_sym_SEMI,
  [61111] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12385), 1,
      anon_sym_RPAREN,
  [61118] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12387), 1,
      anon_sym_LPAREN2,
  [61125] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12389), 1,
      sym_identifier,
  [61132] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12391), 1,
      sym_identifier,
  [61139] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12393), 1,
      anon_sym_LBRACE,
  [61146] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12395), 1,
      anon_sym_SEMI,
  [61153] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12397), 1,
      sym_identifier,
  [61160] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12399), 1,
      sym_identifier,
  [61167] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12401), 1,
      anon_sym_DQUOTE,
  [61174] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11414), 1,
      aux_sym_preproc_include_token3,
  [61181] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12403), 1,
      anon_sym_SEMI,
  [61188] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12405), 1,
      aux_sym_preproc_if_token2,
  [61195] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12407), 1,
      sym_identifier,
  [61202] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12409), 1,
      aux_sym_preproc_if_token2,
  [61209] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12411), 1,
      anon_sym_LPAREN2,
  [61216] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12413), 1,
      anon_sym_STAR,
  [61223] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12415), 1,
      anon_sym_SEMI,
  [61230] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12417), 1,
      aux_sym_ms_asm_block_token1,
  [61237] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12419), 1,
      aux_sym_preproc_include_token3,
  [61244] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12421), 1,
      aux_sym_preproc_if_token2,
  [61251] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12423), 1,
      anon_sym_SEMI,
  [61258] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9577), 1,
      anon_sym_COLON,
  [61265] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12425), 1,
      sym_identifier,
  [61272] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12427), 1,
      anon_sym_SEMI,
  [61279] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12429), 1,
      sym_identifier,
  [61286] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12431), 1,
      anon_sym_RPAREN,
  [61293] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12433), 1,
      anon_sym_SEMI,
  [61300] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12435), 1,
      aux_sym_preproc_if_token2,
  [61307] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12437), 1,
      anon_sym_RPAREN,
  [61314] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12439), 1,
      sym_identifier,
  [61321] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9549), 1,
      anon_sym_RPAREN,
  [61328] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9492), 1,
      anon_sym_SEMI,
  [61335] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12441), 1,
      anon_sym_RPAREN,
  [61342] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12443), 1,
      sym_identifier,
  [61349] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12445), 1,
      sym_identifier,
  [61356] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12447), 1,
      anon_sym_LBRACE,
  [61363] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12449), 1,
      sym_identifier,
  [61370] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9540), 1,
      anon_sym_COLON,
  [61377] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12451), 1,
      sym_identifier,
  [61384] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12453), 1,
      anon_sym_SEMI,
  [61391] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12455), 1,
      anon_sym_SEMI,
  [61398] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12457), 1,
      sym_identifier,
  [61405] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12459), 1,
      sym_identifier,
  [61412] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12461), 1,
      anon_sym_SEMI,
  [61419] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12463), 1,
      anon_sym_SEMI,
  [61426] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12465), 1,
      anon_sym_RPAREN,
  [61433] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12467), 1,
      anon_sym_SEMI,
  [61440] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12469), 1,
      anon_sym_SEMI,
  [61447] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9542), 1,
      anon_sym_SEMI,
  [61454] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12471), 1,
      sym_identifier,
  [61461] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12473), 1,
      aux_sym_preproc_if_token2,
  [61468] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12475), 1,
      sym_identifier,
  [61475] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12477), 1,
      sym_identifier,
  [61482] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12479), 1,
      anon_sym_LBRACE,
  [61489] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9500), 1,
      anon_sym_SEMI,
  [61496] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12481), 1,
      sym_identifier,
  [61503] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12483), 1,
      anon_sym_RPAREN,
  [61510] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12485), 1,
      aux_sym_preproc_include_token3,
  [61517] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12487), 1,
      sym_identifier,
  [61524] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12489), 1,
      anon_sym_SEMI,
  [61531] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12491), 1,
      aux_sym_preproc_if_token2,
  [61538] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12493), 1,
      aux_sym_preproc_include_token3,
  [61545] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12495), 1,
      aux_sym_preproc_if_token2,
  [61552] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9897), 1,
      aux_sym_preproc_include_token3,
  [61559] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12497), 1,
      aux_sym_preproc_if_token2,
  [61566] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12499), 1,
      anon_sym_LBRACE,
  [61573] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12501), 1,
      anon_sym_SEMI,
  [61580] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12503), 1,
      sym_identifier,
  [61587] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12505), 1,
      anon_sym_SEMI,
  [61594] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12507), 1,
      anon_sym_SEMI,
  [61601] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12509), 1,
      anon_sym_RPAREN,
  [61608] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12511), 1,
      anon_sym_RBRACK,
  [61615] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12513), 1,
      anon_sym_RBRACE,
  [61622] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12515), 1,
      anon_sym_SEMI,
  [61629] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12517), 1,
      anon_sym_SEMI,
  [61636] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12519), 1,
      anon_sym_SEMI,
  [61643] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12521), 1,
      anon_sym_SEMI,
  [61650] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12523), 1,
      anon_sym_LBRACE,
  [61657] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12525), 1,
      sym_identifier,
  [61664] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12527), 1,
      anon_sym_LBRACE,
  [61671] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12529), 1,
      anon_sym_LBRACE,
  [61678] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12531), 1,
      anon_sym_LBRACE,
  [61685] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12533), 1,
      anon_sym_LBRACE,
  [61692] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12535), 1,
      anon_sym_LBRACE,
  [61699] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12537), 1,
      anon_sym_LBRACE,
  [61706] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12539), 1,
      anon_sym_LBRACE,
  [61713] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12541), 1,
      anon_sym_LBRACE,
  [61720] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12543), 1,
      anon_sym_SEMI,
  [61727] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9536), 1,
      anon_sym_RPAREN,
  [61734] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9561), 1,
      anon_sym_SEMI,
  [61741] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12545), 1,
      anon_sym_LPAREN2,
  [61748] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12547), 1,
      anon_sym_LPAREN2,
  [61755] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12549), 1,
      anon_sym_LPAREN2,
  [61762] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12551), 1,
      anon_sym_LPAREN2,
  [61769] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12553), 1,
      sym_identifier,
  [61776] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12555), 1,
      anon_sym_LPAREN2,
  [61783] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12557), 1,
      anon_sym_SEMI,
  [61790] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12559), 1,
      anon_sym_RPAREN,
  [61797] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12561), 1,
      anon_sym_LPAREN2,
  [61804] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12563), 1,
      sym_number_literal,
  [61811] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12565), 1,
      anon_sym_LPAREN2,
  [61818] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12567), 1,
      anon_sym_SEMI,
  [61825] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12569), 1,
      anon_sym_LPAREN2,
  [61832] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12571), 1,
      anon_sym_LPAREN2,
  [61839] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12573), 1,
      aux_sym_preproc_if_token2,
  [61846] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9283), 1,
      anon_sym_RPAREN,
  [61853] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12575), 1,
      sym_identifier,
  [61860] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12577), 1,
      sym_identifier,
  [61867] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12579), 1,
      sym_identifier,
  [61874] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12581), 1,
      anon_sym_SEMI,
  [61881] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12583), 1,
      aux_sym_preproc_if_token2,
  [61888] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12585), 1,
      anon_sym_LBRACE,
  [61895] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12587), 1,
      anon_sym_STAR,
  [61902] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12589), 1,
      anon_sym_STAR,
  [61909] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12591), 1,
      aux_sym_preproc_include_token3,
  [61916] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12593), 1,
      anon_sym_SEMI,
  [61923] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12595), 1,
      sym_identifier,
  [61930] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12597), 1,
      anon_sym_RPAREN,
  [61937] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12599), 1,
      anon_sym_SEMI,
  [61944] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12601), 1,
      anon_sym_while,
  [61951] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9524), 1,
      anon_sym_SEMI,
  [61958] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12603), 1,
      anon_sym_SEMI,
  [61965] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12605), 1,
      sym_identifier,
  [61972] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12607), 1,
      anon_sym_LBRACK,
  [61979] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12609), 1,
      sym_identifier,
  [61986] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12611), 1,
      anon_sym_SEMI,
  [61993] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9494), 1,
      anon_sym_SEMI,
  [62000] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12613), 1,
      anon_sym_SEMI,
  [62007] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12615), 1,
      anon_sym_SEMI,
  [62014] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12617), 1,
      anon_sym_SEMI,
  [62021] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12619), 1,
      sym_identifier,
  [62028] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12621), 1,
      anon_sym_RPAREN,
  [62035] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12623), 1,
      anon_sym_STAR,
  [62042] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12625), 1,
      sym_identifier,
  [62049] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12627), 1,
      sym_identifier,
  [62056] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12629), 1,
      sym_identifier,
  [62063] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12631), 1,
      anon_sym_STAR,
  [62070] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12633), 1,
      anon_sym_SEMI,
  [62077] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12635), 1,
      anon_sym_DQUOTE,
  [62084] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12637), 1,
      sym_identifier,
  [62091] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12639), 1,
      aux_sym_preproc_if_token2,
  [62098] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12641), 1,
      anon_sym_SEMI,
  [62105] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12643), 1,
      anon_sym_EQ,
  [62112] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12645), 1,
      aux_sym_preproc_if_token2,
  [62119] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12647), 1,
      anon_sym_RPAREN,
  [62126] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12649), 1,
      sym_identifier,
  [62133] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9496), 1,
      anon_sym_SEMI,
  [62140] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12651), 1,
      anon_sym_LPAREN2,
  [62147] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12653), 1,
      anon_sym_LPAREN2,
  [62154] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12655), 1,
      anon_sym_SEMI,
  [62161] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12657), 1,
      anon_sym_COMMA,
  [62168] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12659), 1,
      aux_sym_preproc_if_token2,
  [62175] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12661), 1,
      anon_sym_RPAREN,
  [62182] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9281), 1,
      anon_sym_RPAREN,
  [62189] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12663), 1,
      anon_sym_SEMI,
  [62196] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12665), 1,
      aux_sym_preproc_if_token2,
  [62203] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12667), 1,
      anon_sym_LPAREN2,
  [62210] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12669), 1,
      anon_sym_SEMI,
  [62217] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12671), 1,
      anon_sym_SEMI,
  [62224] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12673), 1,
      aux_sym_preproc_include_token3,
  [62231] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12675), 1,
      aux_sym_preproc_if_token2,
  [62238] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12677), 1,
      aux_sym_preproc_include_token3,
  [62245] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12679), 1,
      aux_sym_preproc_if_token2,
  [62252] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12681), 1,
      anon_sym_SEMI,
  [62259] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9530), 1,
      anon_sym_SEMI,
  [62266] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12683), 1,
      anon_sym_SEMI,
  [62273] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12685), 1,
      anon_sym_SEMI,
  [62280] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12687), 1,
      anon_sym_RPAREN,
  [62287] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12689), 1,
      anon_sym_RPAREN,
  [62294] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12691), 1,
      anon_sym_RBRACK,
  [62301] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12693), 1,
      anon_sym_SEMI,
  [62308] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12695), 1,
      anon_sym_STAR,
  [62315] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12697), 1,
      anon_sym_SEMI,
  [62322] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12699), 1,
      sym_identifier,
  [62329] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12701), 1,
      anon_sym_SEMI,
  [62336] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12703), 1,
      aux_sym_preproc_include_token3,
  [62343] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12705), 1,
      anon_sym_SEMI,
  [62350] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12707), 1,
      sym_identifier,
  [62357] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12709), 1,
      aux_sym_preproc_include_token3,
  [62364] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12711), 1,
      aux_sym_preproc_if_token2,
  [62371] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12713), 1,
      sym_identifier,
  [62378] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12715), 1,
      anon_sym_SEMI,
  [62385] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12717), 1,
      anon_sym_COLON,
  [62392] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12719), 1,
      sym_identifier,
  [62399] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12721), 1,
      anon_sym_SEMI,
  [62406] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12723), 1,
      aux_sym_preproc_if_token2,
  [62413] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12725), 1,
      anon_sym_SEMI,
  [62420] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12727), 1,
      anon_sym_RPAREN,
  [62427] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12729), 1,
      aux_sym_preproc_include_token3,
  [62434] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12731), 1,
      sym_identifier,
  [62441] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12733), 1,
      anon_sym_SEMI,
  [62448] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12735), 1,
      anon_sym_LPAREN2,
  [62455] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12737), 1,
      anon_sym_SEMI,
  [62462] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7936), 1,
      anon_sym_SEMI,
  [62469] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12739), 1,
      sym_identifier,
  [62476] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9919), 1,
      aux_sym_preproc_include_token3,
  [62483] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12741), 1,
      aux_sym_preproc_if_token2,
  [62490] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12743), 1,
      anon_sym_RPAREN,
  [62497] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12745), 1,
      sym_identifier,
  [62504] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9889), 1,
      aux_sym_preproc_include_token3,
  [62511] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12747), 1,
      aux_sym_preproc_include_token3,
  [62518] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12749), 1,
      anon_sym_LPAREN2,
  [62525] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12751), 1,
      aux_sym_preproc_if_token2,
  [62532] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9514), 1,
      anon_sym_SEMI,
  [62539] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12753), 1,
      anon_sym_SEMI,
  [62546] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12755), 1,
      aux_sym_preproc_if_token2,
  [62553] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12757), 1,
      sym_identifier,
  [62560] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12759), 1,
      anon_sym_SEMI,
  [62567] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12761), 1,
      aux_sym_preproc_include_token3,
  [62574] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12763), 1,
      anon_sym_SEMI,
  [62581] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12765), 1,
      sym_identifier,
  [62588] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12767), 1,
      anon_sym_SEMI,
  [62595] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12769), 1,
      anon_sym_RPAREN,
  [62602] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12771), 1,
      anon_sym_LPAREN2,
  [62609] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12773), 1,
      aux_sym_preproc_if_token2,
  [62616] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12775), 1,
      anon_sym_RPAREN,
  [62623] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12777), 1,
      aux_sym_preproc_if_token2,
  [62630] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12779), 1,
      aux_sym_preproc_include_token3,
  [62637] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10693), 1,
      anon_sym_RPAREN,
  [62644] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12781), 1,
      anon_sym_SEMI,
  [62651] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12783), 1,
      sym_identifier,
  [62658] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12785), 1,
      aux_sym_preproc_if_token2,
  [62665] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12787), 1,
      anon_sym_LPAREN2,
  [62672] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12789), 1,
      anon_sym_SEMI,
  [62679] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12791), 1,
      anon_sym_RPAREN,
  [62686] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12793), 1,
      anon_sym_RPAREN,
  [62693] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12795), 1,
      sym_identifier,
  [62700] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11499), 1,
      anon_sym_RPAREN,
  [62707] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12797), 1,
      anon_sym_SEMI,
  [62714] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12799), 1,
      aux_sym_preproc_if_token2,
  [62721] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9563), 1,
      anon_sym_SEMI,
  [62728] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12801), 1,
      sym_identifier,
  [62735] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12803), 1,
      anon_sym_SEMI,
  [62742] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12805), 1,
      sym_identifier,
  [62749] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12807), 1,
      anon_sym_RPAREN,
  [62756] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12809), 1,
      sym_identifier,
  [62763] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12811), 1,
      sym_identifier,
  [62770] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12813), 1,
      anon_sym_RPAREN,
  [62777] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12815), 1,
      anon_sym_SEMI,
  [62784] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12817), 1,
      anon_sym_SEMI,
  [62791] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12819), 1,
      sym_identifier,
  [62798] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12821), 1,
      sym_identifier,
  [62805] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12823), 1,
      anon_sym_SEMI,
  [62812] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12825), 1,
      anon_sym_SEMI,
  [62819] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12827), 1,
      aux_sym_preproc_if_token2,
  [62826] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12829), 1,
      anon_sym_RPAREN,
  [62833] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12831), 1,
      anon_sym_SEMI,
  [62840] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9285), 1,
      anon_sym_RPAREN,
  [62847] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12833), 1,
      anon_sym_RPAREN,
  [62854] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12835), 1,
      anon_sym_SEMI,
  [62861] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12837), 1,
      anon_sym_RPAREN,
  [62868] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12839), 1,
      sym_number_literal,
  [62875] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12841), 1,
      anon_sym_RPAREN,
  [62882] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9667), 1,
      anon_sym_RPAREN,
  [62889] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7093), 1,
      anon_sym_RBRACE,
  [62896] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12843), 1,
      aux_sym_preproc_if_token2,
  [62903] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12845), 1,
      anon_sym_SEMI,
  [62910] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12847), 1,
      anon_sym_STAR,
  [62917] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12849), 1,
      anon_sym_SEMI,
  [62924] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12851), 1,
      anon_sym_RPAREN,
  [62931] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12853), 1,
      anon_sym_SEMI,
  [62938] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12855), 1,
      anon_sym_SEMI,
  [62945] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12857), 1,
      anon_sym_SEMI,
  [62952] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12859), 1,
      aux_sym_ms_asm_block_token1,
  [62959] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12861), 1,
      anon_sym_RPAREN,
  [62966] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12863), 1,
      anon_sym_SEMI,
  [62973] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9291), 1,
      anon_sym_RPAREN,
  [62980] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12865), 1,
      anon_sym_RBRACK,
  [62987] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12867), 1,
      anon_sym_SEMI,
  [62994] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9506), 1,
      anon_sym_COLON,
  [63001] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12869), 1,
      anon_sym_RPAREN,
  [63008] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12871), 1,
      anon_sym_RPAREN,
  [63015] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12873), 1,
      anon_sym_RPAREN,
  [63022] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12875), 1,
      anon_sym_RPAREN,
  [63029] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12877), 1,
      sym_identifier,
  [63036] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9277), 1,
      anon_sym_RPAREN,
  [63043] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12879), 1,
      anon_sym_SEMI,
  [63050] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12881), 1,
      sym_identifier,
  [63057] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12883), 1,
      anon_sym_COLON,
  [63064] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12885), 1,
      aux_sym_preproc_if_token2,
  [63071] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12887), 1,
      anon_sym_SEMI,
  [63078] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9510), 1,
      anon_sym_COLON,
  [63085] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12889), 1,
      anon_sym_RPAREN,
  [63092] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12891), 1,
      anon_sym_SEMI,
  [63099] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12893), 1,
      anon_sym_SEMI,
  [63106] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12895), 1,
      anon_sym_RPAREN,
  [63113] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12897), 1,
      anon_sym_SEMI,
  [63120] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9669), 1,
      anon_sym_RPAREN,
  [63127] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12899), 1,
      anon_sym_SEMI,
  [63134] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12901), 1,
      anon_sym_LPAREN2,
  [63141] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12903), 1,
      anon_sym_LPAREN2,
  [63148] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12905), 1,
      anon_sym_LPAREN2,
  [63155] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12907), 1,
      aux_sym_preproc_include_token3,
  [63162] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12909), 1,
      anon_sym_SEMI,
  [63169] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12911), 1,
      sym_number_literal,
  [63176] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12913), 1,
      sym_identifier,
  [63183] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12915), 1,
      aux_sym_preproc_if_token2,
  [63190] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12917), 1,
      anon_sym_SEMI,
  [63197] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12919), 1,
      aux_sym_preproc_if_token2,
  [63204] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12921), 1,
      anon_sym_SEMI,
  [63211] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9559), 1,
      anon_sym_COLON,
  [63218] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7127), 1,
      anon_sym_RBRACE,
  [63225] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12923), 1,
      sym_identifier,
  [63232] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11549), 1,
      aux_sym_preproc_include_token3,
  [63239] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12925), 1,
      anon_sym_SEMI,
  [63246] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12927), 1,
      anon_sym_SEMI,
  [63253] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12929), 1,
      anon_sym_while,
  [63260] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12931), 1,
      anon_sym_typedef,
  [63267] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12933), 1,
      aux_sym_preproc_if_token2,
  [63274] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12935), 1,
      anon_sym_RPAREN,
  [63281] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12937), 1,
      anon_sym_RPAREN,
  [63288] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9498), 1,
      anon_sym_SEMI,
  [63295] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12939), 1,
      aux_sym_preproc_include_token3,
  [63302] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12941), 1,
      anon_sym_SEMI,
  [63309] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12943), 1,
      anon_sym_RPAREN,
  [63316] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12945), 1,
      anon_sym_RPAREN,
  [63323] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9293), 1,
      anon_sym_RPAREN,
  [63330] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12947), 1,
      anon_sym_SEMI,
  [63337] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12949), 1,
      anon_sym_SEMI,
  [63344] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12951), 1,
      anon_sym_SEMI,
  [63351] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12953), 1,
      sym_identifier,
  [63358] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9555), 1,
      anon_sym_SEMI,
  [63365] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12955), 1,
      anon_sym_SEMI,
  [63372] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12957), 1,
      aux_sym_preproc_if_token2,
  [63379] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12959), 1,
      sym_identifier,
  [63386] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12961), 1,
      anon_sym_RPAREN,
  [63393] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12963), 1,
      sym_identifier,
  [63400] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12965), 1,
      sym_identifier,
  [63407] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12967), 1,
      anon_sym_SEMI,
  [63414] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12969), 1,
      sym_identifier,
  [63421] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12971), 1,
      anon_sym_SEMI,
  [63428] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12973), 1,
      anon_sym_RPAREN,
  [63435] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12975), 1,
      anon_sym_LPAREN2,
  [63442] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9557), 1,
      anon_sym_SEMI,
  [63449] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12977), 1,
      anon_sym_SEMI,
  [63456] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12979), 1,
      anon_sym_SEMI,
  [63463] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12981), 1,
      anon_sym_SEMI,
  [63470] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12983), 1,
      sym_identifier,
  [63477] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12985), 1,
      anon_sym_SEMI,
  [63484] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9589), 1,
      anon_sym_COLON,
  [63491] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12987), 1,
      anon_sym_SEMI,
  [63498] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12989), 1,
      aux_sym_preproc_include_token3,
  [63505] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12991), 1,
      aux_sym_preproc_if_token2,
  [63512] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(12993), 1,
      aux_sym_preproc_include_token3,
  [63519] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12995), 1,
      anon_sym_SEMI,
  [63526] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12997), 1,
      anon_sym_RBRACE,
  [63533] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12999), 1,
      anon_sym_SEMI,
  [63540] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13001), 1,
      sym_identifier,
  [63547] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13003), 1,
      anon_sym_SEMI,
  [63554] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13005), 1,
      aux_sym_preproc_if_token2,
  [63561] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13007), 1,
      anon_sym_COMMA,
  [63568] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13009), 1,
      anon_sym_RPAREN,
  [63575] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13011), 1,
      aux_sym_preproc_include_token3,
  [63582] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13013), 1,
      anon_sym_RPAREN,
  [63589] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13015), 1,
      sym_identifier,
  [63596] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13017), 1,
      anon_sym_LPAREN2,
  [63603] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13019), 1,
      anon_sym_RBRACK,
  [63610] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13021), 1,
      anon_sym_SEMI,
  [63617] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13023), 1,
      anon_sym_SEMI,
  [63624] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13025), 1,
      sym_identifier,
  [63631] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13027), 1,
      anon_sym_LPAREN2,
  [63638] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13029), 1,
      anon_sym_LPAREN2,
  [63645] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13031), 1,
      aux_sym_preproc_if_token2,
  [63652] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13033), 1,
      sym_identifier,
  [63659] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13035), 1,
      sym_number_literal,
  [63666] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13037), 1,
      anon_sym_LPAREN2,
  [63673] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13039), 1,
      sym_identifier,
  [63680] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13041), 1,
      anon_sym_SEMI,
  [63687] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13043), 1,
      aux_sym_preproc_if_token2,
  [63694] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13045), 1,
      aux_sym_preproc_include_token3,
  [63701] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13047), 1,
      anon_sym_RPAREN,
  [63708] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13049), 1,
      anon_sym_RBRACE,
  [63715] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13051), 1,
      anon_sym_SEMI,
  [63722] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13053), 1,
      anon_sym_while,
  [63729] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13055), 1,
      anon_sym_SEMI,
  [63736] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13057), 1,
      aux_sym_preproc_include_token3,
  [63743] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7934), 1,
      anon_sym_SEMI,
  [63750] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13059), 1,
      aux_sym_preproc_include_token3,
  [63757] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13061), 1,
      anon_sym_RPAREN,
  [63764] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13063), 1,
      aux_sym_preproc_if_token2,
  [63771] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13065), 1,
      anon_sym_SEMI,
  [63778] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13067), 1,
      aux_sym_preproc_if_token2,
  [63785] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13069), 1,
      sym_identifier,
  [63792] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13071), 1,
      aux_sym_preproc_if_token2,
  [63799] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13073), 1,
      anon_sym_SEMI,
  [63806] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13075), 1,
      anon_sym_SEMI,
  [63813] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13077), 1,
      anon_sym_SEMI,
  [63820] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13079), 1,
      anon_sym_SEMI,
  [63827] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11155), 1,
      aux_sym_preproc_include_token3,
  [63834] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13081), 1,
      anon_sym_RPAREN,
  [63841] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13083), 1,
      sym_identifier,
  [63848] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13085), 1,
      anon_sym_LPAREN2,
  [63855] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13087), 1,
      anon_sym_LPAREN2,
  [63862] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13089), 1,
      anon_sym_SEMI,
  [63869] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13091), 1,
      sym_number_literal,
  [63876] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13093), 1,
      anon_sym_SEMI,
  [63883] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13095), 1,
      aux_sym_preproc_if_token2,
  [63890] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13097), 1,
      anon_sym_SEMI,
  [63897] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13099), 1,
      anon_sym_SEMI,
  [63904] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9697), 1,
      anon_sym_RPAREN,
  [63911] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13101), 1,
      anon_sym_LBRACE,
  [63918] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13103), 1,
      anon_sym_while,
  [63925] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13105), 1,
      sym_identifier,
  [63932] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13107), 1,
      anon_sym_SEMI,
  [63939] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13109), 1,
      anon_sym_RPAREN,
  [63946] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13111), 1,
      anon_sym_RPAREN,
  [63953] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13113), 1,
      aux_sym_preproc_if_token2,
  [63960] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13115), 1,
      anon_sym_SEMI,
  [63967] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13117), 1,
      anon_sym_SEMI,
  [63974] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13119), 1,
      anon_sym_RPAREN,
  [63981] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13121), 1,
      sym_identifier,
  [63988] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13123), 1,
      anon_sym_STAR,
  [63995] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13125), 1,
      anon_sym_SEMI,
  [64002] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13127), 1,
      sym_identifier,
  [64009] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13129), 1,
      anon_sym_SEMI,
  [64016] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13131), 1,
      sym_identifier,
  [64023] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13133), 1,
      anon_sym_SEMI,
  [64030] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13135), 1,
      anon_sym_LPAREN2,
  [64037] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13137), 1,
      aux_sym_preproc_if_token2,
  [64044] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13139), 1,
      anon_sym_RBRACK,
  [64051] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13141), 1,
      anon_sym_RBRACE,
  [64058] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13143), 1,
      anon_sym_LPAREN2,
  [64065] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13145), 1,
      anon_sym_while,
  [64072] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13147), 1,
      anon_sym_SEMI,
  [64079] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13149), 1,
      anon_sym_SEMI,
  [64086] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13151), 1,
      anon_sym_SEMI,
  [64093] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13153), 1,
      sym_identifier,
  [64100] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13155), 1,
      sym_identifier,
  [64107] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13157), 1,
      anon_sym_SEMI,
  [64114] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13159), 1,
      anon_sym_RPAREN,
  [64121] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13161), 1,
      anon_sym_SEMI,
  [64128] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13163), 1,
      sym_identifier,
  [64135] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13165), 1,
      anon_sym_SEMI,
  [64142] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13167), 1,
      anon_sym_SEMI,
  [64149] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13169), 1,
      anon_sym_RPAREN,
  [64156] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13171), 1,
      sym_identifier,
  [64163] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13173), 1,
      anon_sym_RBRACK,
  [64170] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13175), 1,
      anon_sym_LPAREN2,
  [64177] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13177), 1,
      anon_sym_RPAREN,
  [64184] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9565), 1,
      anon_sym_COLON,
  [64191] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13179), 1,
      anon_sym_SEMI,
  [64198] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13181), 1,
      anon_sym_SEMI,
  [64205] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13183), 1,
      anon_sym_while,
  [64212] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13185), 1,
      aux_sym_preproc_if_token2,
  [64219] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13187), 1,
      anon_sym_SEMI,
  [64226] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9532), 1,
      anon_sym_COLON,
  [64233] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13189), 1,
      anon_sym_RPAREN,
  [64240] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13191), 1,
      anon_sym_SEMI,
  [64247] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13193), 1,
      ts_builtin_sym_end,
  [64254] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13195), 1,
      anon_sym_SEMI,
  [64261] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13197), 1,
      aux_sym_preproc_include_token3,
  [64268] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13199), 1,
      anon_sym_RPAREN,
  [64275] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9534), 1,
      anon_sym_RPAREN,
  [64282] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13201), 1,
      anon_sym_RBRACK,
  [64289] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13203), 1,
      aux_sym_preproc_include_token3,
  [64296] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13205), 1,
      sym_identifier,
  [64303] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13207), 1,
      anon_sym_SEMI,
  [64310] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13209), 1,
      anon_sym_RPAREN,
  [64317] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13211), 1,
      anon_sym_SEMI,
  [64324] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13213), 1,
      anon_sym_COMMA,
  [64331] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13215), 1,
      anon_sym_RBRACE,
  [64338] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13217), 1,
      anon_sym_SEMI,
  [64345] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13219), 1,
      sym_identifier,
  [64352] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13221), 1,
      anon_sym_SEMI,
  [64359] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13223), 1,
      anon_sym_SEMI,
  [64366] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13225), 1,
      anon_sym_SEMI,
  [64373] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11778), 1,
      anon_sym_EQ,
  [64380] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13227), 1,
      anon_sym_RPAREN,
  [64387] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13229), 1,
      anon_sym_SEMI,
  [64394] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13231), 1,
      aux_sym_preproc_include_token3,
  [64401] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13233), 1,
      sym_identifier,
  [64408] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13235), 1,
      sym_identifier,
  [64415] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13237), 1,
      anon_sym_SEMI,
  [64422] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13239), 1,
      sym_identifier,
  [64429] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(10445), 1,
      sym_identifier,
  [64436] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13241), 1,
      aux_sym_preproc_if_token2,
  [64443] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13243), 1,
      anon_sym_SEMI,
  [64450] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13245), 1,
      anon_sym_SEMI,
  [64457] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13247), 1,
      sym_identifier,
  [64464] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13249), 1,
      anon_sym_COMMA,
  [64471] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13251), 1,
      anon_sym_SEMI,
  [64478] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13253), 1,
      aux_sym_preproc_include_token3,
  [64485] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13255), 1,
      aux_sym_preproc_if_token2,
  [64492] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13257), 1,
      anon_sym_SEMI,
  [64499] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13259), 1,
      aux_sym_preproc_if_token2,
  [64506] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13261), 1,
      aux_sym_preproc_if_token2,
  [64513] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13263), 1,
      anon_sym_SEMI,
  [64520] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13265), 1,
      anon_sym_SEMI,
  [64527] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13267), 1,
      anon_sym_LBRACE,
  [64534] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13269), 1,
      anon_sym_SEMI,
  [64541] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13271), 1,
      anon_sym_LPAREN2,
  [64548] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13273), 1,
      anon_sym_SEMI,
  [64555] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13275), 1,
      anon_sym_STAR,
  [64562] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13277), 1,
      aux_sym_preproc_if_token2,
  [64569] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13279), 1,
      anon_sym_SEMI,
  [64576] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13281), 1,
      anon_sym_RPAREN,
  [64583] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13283), 1,
      anon_sym_SEMI,
  [64590] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9271), 1,
      anon_sym_RPAREN,
  [64597] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13285), 1,
      anon_sym_RBRACK,
  [64604] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13287), 1,
      anon_sym_typedef,
  [64611] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13289), 1,
      anon_sym_RBRACE,
  [64618] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13291), 1,
      anon_sym_LPAREN2,
  [64625] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13293), 1,
      aux_sym_preproc_include_token3,
  [64632] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13295), 1,
      anon_sym_SEMI,
  [64639] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9512), 1,
      anon_sym_SEMI,
  [64646] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13297), 1,
      sym_identifier,
  [64653] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13299), 1,
      anon_sym_LBRACK,
  [64660] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13301), 1,
      anon_sym_RPAREN,
  [64667] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13303), 1,
      sym_identifier,
  [64674] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13305), 1,
      sym_identifier,
  [64681] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13307), 1,
      sym_identifier,
  [64688] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13309), 1,
      anon_sym_RPAREN,
  [64695] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13311), 1,
      anon_sym_SEMI,
  [64702] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9547), 1,
      anon_sym_COLON,
  [64709] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13313), 1,
      sym_identifier,
  [64716] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13315), 1,
      sym_identifier,
  [64723] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13317), 1,
      anon_sym_SEMI,
  [64730] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13319), 1,
      anon_sym_DQUOTE,
  [64737] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13321), 1,
      anon_sym_SEMI,
  [64744] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13323), 1,
      aux_sym_preproc_if_token2,
  [64751] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13325), 1,
      sym_identifier,
  [64758] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(9893), 1,
      aux_sym_preproc_include_token3,
  [64765] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9450), 1,
      anon_sym_RBRACE,
  [64772] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13327), 1,
      anon_sym_SEMI,
  [64779] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13329), 1,
      aux_sym_preproc_if_token2,
  [64786] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13331), 1,
      anon_sym_SEMI,
  [64793] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13333), 1,
      anon_sym_RBRACK,
  [64800] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13335), 1,
      sym_identifier,
  [64807] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13337), 1,
      anon_sym_SEMI,
  [64814] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13339), 1,
      anon_sym_SEMI,
  [64821] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13341), 1,
      anon_sym_SEMI,
  [64828] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13343), 1,
      anon_sym_SEMI,
  [64835] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13345), 1,
      anon_sym_SEMI,
  [64842] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13347), 1,
      anon_sym_SEMI,
  [64849] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13349), 1,
      anon_sym_RPAREN,
  [64856] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13351), 1,
      anon_sym_RPAREN,
  [64863] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13353), 1,
      anon_sym_RBRACK,
  [64870] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13355), 1,
      anon_sym_LPAREN2,
  [64877] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13357), 1,
      sym_identifier,
  [64884] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13359), 1,
      anon_sym_COMMA,
  [64891] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13361), 1,
      sym_identifier,
  [64898] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13363), 1,
      anon_sym_SEMI,
  [64905] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13365), 1,
      anon_sym_RPAREN,
  [64912] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13367), 1,
      anon_sym_RPAREN,
  [64919] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13369), 1,
      sym_identifier,
  [64926] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13371), 1,
      anon_sym_RPAREN,
  [64933] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13373), 1,
      aux_sym_preproc_include_token3,
  [64940] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13375), 1,
      aux_sym_ms_asm_block_token1,
  [64947] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13377), 1,
      aux_sym_preproc_include_token3,
  [64954] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13379), 1,
      aux_sym_preproc_if_token2,
  [64961] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13381), 1,
      aux_sym_preproc_if_token2,
  [64968] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13383), 1,
      sym_identifier,
  [64975] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9673), 1,
      anon_sym_RPAREN,
  [64982] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13385), 1,
      sym_identifier,
  [64989] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13387), 1,
      anon_sym_SEMI,
  [64996] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13389), 1,
      anon_sym_SEMI,
  [65003] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13391), 1,
      anon_sym_SEMI,
  [65010] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13393), 1,
      sym_identifier,
  [65017] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13395), 1,
      aux_sym_preproc_if_token2,
  [65024] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13397), 1,
      anon_sym_LPAREN2,
  [65031] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13399), 1,
      anon_sym_RPAREN,
  [65038] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13401), 1,
      sym_identifier,
  [65045] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13403), 1,
      anon_sym_SEMI,
  [65052] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13405), 1,
      aux_sym_preproc_if_token2,
  [65059] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13407), 1,
      anon_sym_SEMI,
  [65066] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13409), 1,
      anon_sym_RPAREN,
  [65073] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13411), 1,
      anon_sym_SEMI,
  [65080] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9279), 1,
      anon_sym_RPAREN,
  [65087] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13413), 1,
      anon_sym_SEMI,
  [65094] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13415), 1,
      anon_sym_SEMI,
  [65101] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13417), 1,
      anon_sym_STAR,
  [65108] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13419), 1,
      anon_sym_RPAREN,
  [65115] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13421), 1,
      anon_sym_SEMI,
  [65122] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13423), 1,
      anon_sym_RPAREN,
  [65129] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13425), 1,
      anon_sym_RBRACK,
  [65136] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9573), 1,
      anon_sym_COLON,
  [65143] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13427), 1,
      anon_sym_SEMI,
  [65150] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13429), 1,
      anon_sym_RPAREN,
  [65157] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13431), 1,
      sym_identifier,
  [65164] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13433), 1,
      anon_sym_SEMI,
  [65171] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13435), 1,
      aux_sym_preproc_if_token2,
  [65178] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13437), 1,
      anon_sym_SEMI,
  [65185] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13439), 1,
      anon_sym_LPAREN2,
  [65192] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13441), 1,
      aux_sym_preproc_include_token3,
  [65199] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13443), 1,
      anon_sym_SEMI,
  [65206] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13445), 1,
      aux_sym_ms_asm_block_token1,
  [65213] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13447), 1,
      anon_sym_RPAREN,
  [65220] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13449), 1,
      aux_sym_preproc_include_token3,
  [65227] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9273), 1,
      anon_sym_RPAREN,
  [65234] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13451), 1,
      sym_identifier,
  [65241] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13453), 1,
      anon_sym_typedef,
  [65248] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13455), 1,
      anon_sym_SEMI,
  [65255] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13457), 1,
      anon_sym_LPAREN2,
  [65262] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13459), 1,
      sym_identifier,
  [65269] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13461), 1,
      anon_sym_LBRACK,
  [65276] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13463), 1,
      anon_sym_RPAREN,
  [65283] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13465), 1,
      anon_sym_SEMI,
  [65290] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13467), 1,
      anon_sym_SEMI,
  [65297] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13469), 1,
      anon_sym_SEMI,
  [65304] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13471), 1,
      sym_identifier,
  [65311] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9579), 1,
      anon_sym_COLON,
  [65318] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13473), 1,
      anon_sym_SEMI,
  [65325] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13475), 1,
      anon_sym_RPAREN,
  [65332] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13477), 1,
      aux_sym_preproc_if_token2,
  [65339] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13479), 1,
      anon_sym_LPAREN2,
  [65346] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13481), 1,
      anon_sym_SEMI,
  [65353] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13483), 1,
      anon_sym_SEMI,
  [65360] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9581), 1,
      anon_sym_RPAREN,
  [65367] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11298), 1,
      anon_sym_RPAREN,
  [65374] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13485), 1,
      aux_sym_preproc_include_token3,
  [65381] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7833), 1,
      anon_sym_SEMI,
  [65388] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13487), 1,
      anon_sym_RBRACK,
  [65395] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13489), 1,
      anon_sym_SEMI,
  [65402] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9583), 1,
      anon_sym_RPAREN,
  [65409] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13491), 1,
      anon_sym_COLON,
  [65416] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13493), 1,
      anon_sym_SEMI,
  [65423] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13495), 1,
      anon_sym_LPAREN2,
  [65430] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13497), 1,
      anon_sym_RPAREN,
  [65437] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13499), 1,
      aux_sym_preproc_if_token2,
  [65444] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13501), 1,
      anon_sym_SEMI,
  [65451] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9585), 1,
      anon_sym_SEMI,
  [65458] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13503), 1,
      anon_sym_RPAREN,
  [65465] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13505), 1,
      anon_sym_RPAREN,
  [65472] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13507), 1,
      anon_sym_COLON,
  [65479] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13509), 1,
      sym_identifier,
  [65486] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13511), 1,
      anon_sym_SEMI,
  [65493] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13513), 1,
      sym_identifier,
  [65500] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13515), 1,
      anon_sym_SEMI,
  [65507] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13517), 1,
      anon_sym_LPAREN2,
  [65514] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13519), 1,
      anon_sym_COLON,
  [65521] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13521), 1,
      anon_sym_SEMI,
  [65528] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13523), 1,
      anon_sym_SEMI,
  [65535] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13525), 1,
      anon_sym_RBRACE,
  [65542] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13527), 1,
      sym_identifier,
  [65549] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13529), 1,
      sym_identifier,
  [65556] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13531), 1,
      sym_identifier,
  [65563] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13533), 1,
      anon_sym_SEMI,
  [65570] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13535), 1,
      anon_sym_RPAREN,
  [65577] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13537), 1,
      sym_identifier,
  [65584] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9275), 1,
      anon_sym_RPAREN,
  [65591] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9478), 1,
      anon_sym_RBRACE,
  [65598] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13539), 1,
      anon_sym_SEMI,
  [65605] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9587), 1,
      anon_sym_COLON,
  [65612] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13541), 1,
      aux_sym_preproc_include_token3,
  [65619] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13543), 1,
      anon_sym_SEMI,
  [65626] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13545), 1,
      anon_sym_SEMI,
  [65633] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13547), 1,
      anon_sym_SEMI,
  [65640] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13549), 1,
      anon_sym_LPAREN2,
  [65647] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13551), 1,
      anon_sym_RPAREN,
  [65654] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13553), 1,
      anon_sym_typedef,
  [65661] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13555), 1,
      anon_sym_LPAREN2,
  [65668] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13557), 1,
      anon_sym_SEMI,
  [65675] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13559), 1,
      sym_identifier,
  [65682] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7289), 1,
      anon_sym_SEMI,
  [65689] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13561), 1,
      aux_sym_preproc_if_token2,
  [65696] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13563), 1,
      anon_sym_LBRACK,
  [65703] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13565), 1,
      aux_sym_preproc_if_token2,
  [65710] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13567), 1,
      anon_sym_LPAREN2,
  [65717] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13569), 1,
      anon_sym_SEMI,
  [65724] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13571), 1,
      anon_sym_typedef,
  [65731] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13573), 1,
      anon_sym_LPAREN2,
  [65738] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9508), 1,
      anon_sym_SEMI,
  [65745] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13575), 1,
      anon_sym_SEMI,
  [65752] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13577), 1,
      anon_sym_COLON,
  [65759] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13579), 1,
      sym_identifier,
  [65766] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13581), 1,
      anon_sym_LPAREN2,
  [65773] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13583), 1,
      anon_sym_SEMI,
  [65780] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13585), 1,
      anon_sym_typedef,
  [65787] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13587), 1,
      anon_sym_LPAREN2,
  [65794] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13589), 1,
      anon_sym_SEMI,
  [65801] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13591), 1,
      anon_sym_SEMI,
  [65808] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13593), 1,
      anon_sym_SEMI,
  [65815] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(11304), 1,
      aux_sym_preproc_include_token3,
  [65822] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13595), 1,
      anon_sym_SEMI,
  [65829] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13597), 1,
      anon_sym_typedef,
  [65836] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13599), 1,
      anon_sym_LPAREN2,
  [65843] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11312), 1,
      anon_sym_RPAREN,
  [65850] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13601), 1,
      anon_sym_typedef,
  [65857] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13603), 1,
      anon_sym_LPAREN2,
  [65864] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13605), 1,
      sym_identifier,
  [65871] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13607), 1,
      anon_sym_typedef,
  [65878] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13609), 1,
      anon_sym_SEMI,
  [65885] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13611), 1,
      anon_sym_typedef,
  [65892] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13613), 1,
      anon_sym_RPAREN,
  [65899] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13615), 1,
      anon_sym_LPAREN2,
  [65906] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13617), 1,
      anon_sym_LPAREN2,
  [65913] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13619), 1,
      sym_identifier,
  [65920] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13621), 1,
      sym_identifier,
  [65927] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13623), 1,
      aux_sym_preproc_include_token3,
  [65934] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13625), 1,
      anon_sym_SEMI,
  [65941] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7748), 1,
      anon_sym_SEMI,
  [65948] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13627), 1,
      aux_sym_preproc_if_token2,
  [65955] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13629), 1,
      anon_sym_SEMI,
  [65962] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13631), 1,
      aux_sym_preproc_include_token3,
  [65969] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13633), 1,
      anon_sym_SEMI,
  [65976] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13635), 1,
      anon_sym_RPAREN,
  [65983] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13637), 1,
      aux_sym_preproc_if_token2,
  [65990] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13639), 1,
      anon_sym_SEMI,
  [65997] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13641), 1,
      sym_identifier,
  [66004] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13643), 1,
      anon_sym_SEMI,
  [66011] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13645), 1,
      anon_sym_RPAREN,
  [66018] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13647), 1,
      anon_sym_LPAREN2,
  [66025] = 2,
    ACTIONS(9865), 1,
      sym_comment,
    ACTIONS(13649), 1,
      aux_sym_preproc_include_token3,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(4111)] = 0,
  [SMALL_STATE(4112)] = 103,
  [SMALL_STATE(4113)] = 206,
  [SMALL_STATE(4114)] = 276,
  [SMALL_STATE(4115)] = 376,
  [SMALL_STATE(4116)] = 446,
  [SMALL_STATE(4117)] = 546,
  [SMALL_STATE(4118)] = 616,
  [SMALL_STATE(4119)] = 715,
  [SMALL_STATE(4120)] = 814,
  [SMALL_STATE(4121)] = 912,
  [SMALL_STATE(4122)] = 1010,
  [SMALL_STATE(4123)] = 1081,
  [SMALL_STATE(4124)] = 1152,
  [SMALL_STATE(4125)] = 1223,
  [SMALL_STATE(4126)] = 1294,
  [SMALL_STATE(4127)] = 1365,
  [SMALL_STATE(4128)] = 1436,
  [SMALL_STATE(4129)] = 1507,
  [SMALL_STATE(4130)] = 1595,
  [SMALL_STATE(4131)] = 1683,
  [SMALL_STATE(4132)] = 1771,
  [SMALL_STATE(4133)] = 1843,
  [SMALL_STATE(4134)] = 1917,
  [SMALL_STATE(4135)] = 1989,
  [SMALL_STATE(4136)] = 2061,
  [SMALL_STATE(4137)] = 2133,
  [SMALL_STATE(4138)] = 2221,
  [SMALL_STATE(4139)] = 2295,
  [SMALL_STATE(4140)] = 2369,
  [SMALL_STATE(4141)] = 2441,
  [SMALL_STATE(4142)] = 2513,
  [SMALL_STATE(4143)] = 2597,
  [SMALL_STATE(4144)] = 2663,
  [SMALL_STATE(4145)] = 2729,
  [SMALL_STATE(4146)] = 2795,
  [SMALL_STATE(4147)] = 2861,
  [SMALL_STATE(4148)] = 2927,
  [SMALL_STATE(4149)] = 2993,
  [SMALL_STATE(4150)] = 3055,
  [SMALL_STATE(4151)] = 3117,
  [SMALL_STATE(4152)] = 3183,
  [SMALL_STATE(4153)] = 3249,
  [SMALL_STATE(4154)] = 3315,
  [SMALL_STATE(4155)] = 3381,
  [SMALL_STATE(4156)] = 3447,
  [SMALL_STATE(4157)] = 3513,
  [SMALL_STATE(4158)] = 3579,
  [SMALL_STATE(4159)] = 3665,
  [SMALL_STATE(4160)] = 3749,
  [SMALL_STATE(4161)] = 3835,
  [SMALL_STATE(4162)] = 3901,
  [SMALL_STATE(4163)] = 3967,
  [SMALL_STATE(4164)] = 4050,
  [SMALL_STATE(4165)] = 4125,
  [SMALL_STATE(4166)] = 4200,
  [SMALL_STATE(4167)] = 4283,
  [SMALL_STATE(4168)] = 4366,
  [SMALL_STATE(4169)] = 4449,
  [SMALL_STATE(4170)] = 4509,
  [SMALL_STATE(4171)] = 4569,
  [SMALL_STATE(4172)] = 4629,
  [SMALL_STATE(4173)] = 4709,
  [SMALL_STATE(4174)] = 4769,
  [SMALL_STATE(4175)] = 4829,
  [SMALL_STATE(4176)] = 4889,
  [SMALL_STATE(4177)] = 4949,
  [SMALL_STATE(4178)] = 5009,
  [SMALL_STATE(4179)] = 5069,
  [SMALL_STATE(4180)] = 5143,
  [SMALL_STATE(4181)] = 5227,
  [SMALL_STATE(4182)] = 5301,
  [SMALL_STATE(4183)] = 5381,
  [SMALL_STATE(4184)] = 5441,
  [SMALL_STATE(4185)] = 5525,
  [SMALL_STATE(4186)] = 5601,
  [SMALL_STATE(4187)] = 5685,
  [SMALL_STATE(4188)] = 5769,
  [SMALL_STATE(4189)] = 5829,
  [SMALL_STATE(4190)] = 5903,
  [SMALL_STATE(4191)] = 5963,
  [SMALL_STATE(4192)] = 6047,
  [SMALL_STATE(4193)] = 6107,
  [SMALL_STATE(4194)] = 6187,
  [SMALL_STATE(4195)] = 6247,
  [SMALL_STATE(4196)] = 6321,
  [SMALL_STATE(4197)] = 6397,
  [SMALL_STATE(4198)] = 6457,
  [SMALL_STATE(4199)] = 6517,
  [SMALL_STATE(4200)] = 6597,
  [SMALL_STATE(4201)] = 6666,
  [SMALL_STATE(4202)] = 6747,
  [SMALL_STATE(4203)] = 6828,
  [SMALL_STATE(4204)] = 6909,
  [SMALL_STATE(4205)] = 6988,
  [SMALL_STATE(4206)] = 7069,
  [SMALL_STATE(4207)] = 7150,
  [SMALL_STATE(4208)] = 7229,
  [SMALL_STATE(4209)] = 7310,
  [SMALL_STATE(4210)] = 7389,
  [SMALL_STATE(4211)] = 7470,
  [SMALL_STATE(4212)] = 7551,
  [SMALL_STATE(4213)] = 7632,
  [SMALL_STATE(4214)] = 7713,
  [SMALL_STATE(4215)] = 7792,
  [SMALL_STATE(4216)] = 7873,
  [SMALL_STATE(4217)] = 7954,
  [SMALL_STATE(4218)] = 8033,
  [SMALL_STATE(4219)] = 8114,
  [SMALL_STATE(4220)] = 8193,
  [SMALL_STATE(4221)] = 8262,
  [SMALL_STATE(4222)] = 8343,
  [SMALL_STATE(4223)] = 8422,
  [SMALL_STATE(4224)] = 8503,
  [SMALL_STATE(4225)] = 8582,
  [SMALL_STATE(4226)] = 8661,
  [SMALL_STATE(4227)] = 8740,
  [SMALL_STATE(4228)] = 8821,
  [SMALL_STATE(4229)] = 8902,
  [SMALL_STATE(4230)] = 8983,
  [SMALL_STATE(4231)] = 9064,
  [SMALL_STATE(4232)] = 9145,
  [SMALL_STATE(4233)] = 9226,
  [SMALL_STATE(4234)] = 9307,
  [SMALL_STATE(4235)] = 9388,
  [SMALL_STATE(4236)] = 9469,
  [SMALL_STATE(4237)] = 9550,
  [SMALL_STATE(4238)] = 9614,
  [SMALL_STATE(4239)] = 9675,
  [SMALL_STATE(4240)] = 9736,
  [SMALL_STATE(4241)] = 9801,
  [SMALL_STATE(4242)] = 9862,
  [SMALL_STATE(4243)] = 9927,
  [SMALL_STATE(4244)] = 9992,
  [SMALL_STATE(4245)] = 10053,
  [SMALL_STATE(4246)] = 10114,
  [SMALL_STATE(4247)] = 10179,
  [SMALL_STATE(4248)] = 10244,
  [SMALL_STATE(4249)] = 10305,
  [SMALL_STATE(4250)] = 10370,
  [SMALL_STATE(4251)] = 10435,
  [SMALL_STATE(4252)] = 10496,
  [SMALL_STATE(4253)] = 10561,
  [SMALL_STATE(4254)] = 10626,
  [SMALL_STATE(4255)] = 10691,
  [SMALL_STATE(4256)] = 10768,
  [SMALL_STATE(4257)] = 10829,
  [SMALL_STATE(4258)] = 10890,
  [SMALL_STATE(4259)] = 10955,
  [SMALL_STATE(4260)] = 11020,
  [SMALL_STATE(4261)] = 11081,
  [SMALL_STATE(4262)] = 11142,
  [SMALL_STATE(4263)] = 11203,
  [SMALL_STATE(4264)] = 11264,
  [SMALL_STATE(4265)] = 11334,
  [SMALL_STATE(4266)] = 11400,
  [SMALL_STATE(4267)] = 11468,
  [SMALL_STATE(4268)] = 11538,
  [SMALL_STATE(4269)] = 11606,
  [SMALL_STATE(4270)] = 11676,
  [SMALL_STATE(4271)] = 11744,
  [SMALL_STATE(4272)] = 11812,
  [SMALL_STATE(4273)] = 11879,
  [SMALL_STATE(4274)] = 11946,
  [SMALL_STATE(4275)] = 12011,
  [SMALL_STATE(4276)] = 12078,
  [SMALL_STATE(4277)] = 12145,
  [SMALL_STATE(4278)] = 12212,
  [SMALL_STATE(4279)] = 12279,
  [SMALL_STATE(4280)] = 12346,
  [SMALL_STATE(4281)] = 12413,
  [SMALL_STATE(4282)] = 12480,
  [SMALL_STATE(4283)] = 12547,
  [SMALL_STATE(4284)] = 12614,
  [SMALL_STATE(4285)] = 12680,
  [SMALL_STATE(4286)] = 12746,
  [SMALL_STATE(4287)] = 12812,
  [SMALL_STATE(4288)] = 12876,
  [SMALL_STATE(4289)] = 12940,
  [SMALL_STATE(4290)] = 13006,
  [SMALL_STATE(4291)] = 13072,
  [SMALL_STATE(4292)] = 13136,
  [SMALL_STATE(4293)] = 13215,
  [SMALL_STATE(4294)] = 13304,
  [SMALL_STATE(4295)] = 13369,
  [SMALL_STATE(4296)] = 13438,
  [SMALL_STATE(4297)] = 13495,
  [SMALL_STATE(4298)] = 13586,
  [SMALL_STATE(4299)] = 13661,
  [SMALL_STATE(4300)] = 13752,
  [SMALL_STATE(4301)] = 13841,
  [SMALL_STATE(4302)] = 13926,
  [SMALL_STATE(4303)] = 14017,
  [SMALL_STATE(4304)] = 14088,
  [SMALL_STATE(4305)] = 14165,
  [SMALL_STATE(4306)] = 14254,
  [SMALL_STATE(4307)] = 14319,
  [SMALL_STATE(4308)] = 14386,
  [SMALL_STATE(4309)] = 14451,
  [SMALL_STATE(4310)] = 14534,
  [SMALL_STATE(4311)] = 14615,
  [SMALL_STATE(4312)] = 14680,
  [SMALL_STATE(4313)] = 14736,
  [SMALL_STATE(4314)] = 14789,
  [SMALL_STATE(4315)] = 14842,
  [SMALL_STATE(4316)] = 14930,
  [SMALL_STATE(4317)] = 15002,
  [SMALL_STATE(4318)] = 15084,
  [SMALL_STATE(4319)] = 15172,
  [SMALL_STATE(4320)] = 15252,
  [SMALL_STATE(4321)] = 15338,
  [SMALL_STATE(4322)] = 15424,
  [SMALL_STATE(4323)] = 15492,
  [SMALL_STATE(4324)] = 15568,
  [SMALL_STATE(4325)] = 15654,
  [SMALL_STATE(4326)] = 15728,
  [SMALL_STATE(4327)] = 15816,
  [SMALL_STATE(4328)] = 15882,
  [SMALL_STATE(4329)] = 15946,
  [SMALL_STATE(4330)] = 16024,
  [SMALL_STATE(4331)] = 16075,
  [SMALL_STATE(4332)] = 16126,
  [SMALL_STATE(4333)] = 16177,
  [SMALL_STATE(4334)] = 16228,
  [SMALL_STATE(4335)] = 16279,
  [SMALL_STATE(4336)] = 16330,
  [SMALL_STATE(4337)] = 16381,
  [SMALL_STATE(4338)] = 16432,
  [SMALL_STATE(4339)] = 16483,
  [SMALL_STATE(4340)] = 16534,
  [SMALL_STATE(4341)] = 16585,
  [SMALL_STATE(4342)] = 16638,
  [SMALL_STATE(4343)] = 16689,
  [SMALL_STATE(4344)] = 16773,
  [SMALL_STATE(4345)] = 16823,
  [SMALL_STATE(4346)] = 16907,
  [SMALL_STATE(4347)] = 16991,
  [SMALL_STATE(4348)] = 17053,
  [SMALL_STATE(4349)] = 17139,
  [SMALL_STATE(4350)] = 17219,
  [SMALL_STATE(4351)] = 17297,
  [SMALL_STATE(4352)] = 17373,
  [SMALL_STATE(4353)] = 17447,
  [SMALL_STATE(4354)] = 17519,
  [SMALL_STATE(4355)] = 17589,
  [SMALL_STATE(4356)] = 17655,
  [SMALL_STATE(4357)] = 17719,
  [SMALL_STATE(4358)] = 17805,
  [SMALL_STATE(4359)] = 17891,
  [SMALL_STATE(4360)] = 17942,
  [SMALL_STATE(4361)] = 17994,
  [SMALL_STATE(4362)] = 18053,
  [SMALL_STATE(4363)] = 18112,
  [SMALL_STATE(4364)] = 18172,
  [SMALL_STATE(4365)] = 18254,
  [SMALL_STATE(4366)] = 18334,
  [SMALL_STATE(4367)] = 18392,
  [SMALL_STATE(4368)] = 18468,
  [SMALL_STATE(4369)] = 18542,
  [SMALL_STATE(4370)] = 18614,
  [SMALL_STATE(4371)] = 18684,
  [SMALL_STATE(4372)] = 18752,
  [SMALL_STATE(4373)] = 18818,
  [SMALL_STATE(4374)] = 18880,
  [SMALL_STATE(4375)] = 18960,
  [SMALL_STATE(4376)] = 19042,
  [SMALL_STATE(4377)] = 19122,
  [SMALL_STATE(4378)] = 19204,
  [SMALL_STATE(4379)] = 19290,
  [SMALL_STATE(4380)] = 19370,
  [SMALL_STATE(4381)] = 19456,
  [SMALL_STATE(4382)] = 19538,
  [SMALL_STATE(4383)] = 19601,
  [SMALL_STATE(4384)] = 19678,
  [SMALL_STATE(4385)] = 19755,
  [SMALL_STATE(4386)] = 19830,
  [SMALL_STATE(4387)] = 19885,
  [SMALL_STATE(4388)] = 19956,
  [SMALL_STATE(4389)] = 20025,
  [SMALL_STATE(4390)] = 20094,
  [SMALL_STATE(4391)] = 20161,
  [SMALL_STATE(4392)] = 20226,
  [SMALL_STATE(4393)] = 20285,
  [SMALL_STATE(4394)] = 20342,
  [SMALL_STATE(4395)] = 20419,
  [SMALL_STATE(4396)] = 20494,
  [SMALL_STATE(4397)] = 20569,
  [SMALL_STATE(4398)] = 20646,
  [SMALL_STATE(4399)] = 20724,
  [SMALL_STATE(4400)] = 20804,
  [SMALL_STATE(4401)] = 20884,
  [SMALL_STATE(4402)] = 20964,
  [SMALL_STATE(4403)] = 21044,
  [SMALL_STATE(4404)] = 21124,
  [SMALL_STATE(4405)] = 21204,
  [SMALL_STATE(4406)] = 21284,
  [SMALL_STATE(4407)] = 21364,
  [SMALL_STATE(4408)] = 21444,
  [SMALL_STATE(4409)] = 21524,
  [SMALL_STATE(4410)] = 21604,
  [SMALL_STATE(4411)] = 21684,
  [SMALL_STATE(4412)] = 21764,
  [SMALL_STATE(4413)] = 21844,
  [SMALL_STATE(4414)] = 21924,
  [SMALL_STATE(4415)] = 22004,
  [SMALL_STATE(4416)] = 22084,
  [SMALL_STATE(4417)] = 22164,
  [SMALL_STATE(4418)] = 22231,
  [SMALL_STATE(4419)] = 22308,
  [SMALL_STATE(4420)] = 22385,
  [SMALL_STATE(4421)] = 22462,
  [SMALL_STATE(4422)] = 22537,
  [SMALL_STATE(4423)] = 22610,
  [SMALL_STATE(4424)] = 22663,
  [SMALL_STATE(4425)] = 22732,
  [SMALL_STATE(4426)] = 22809,
  [SMALL_STATE(4427)] = 22876,
  [SMALL_STATE(4428)] = 22941,
  [SMALL_STATE(4429)] = 23004,
  [SMALL_STATE(4430)] = 23065,
  [SMALL_STATE(4431)] = 23122,
  [SMALL_STATE(4432)] = 23177,
  [SMALL_STATE(4433)] = 23254,
  [SMALL_STATE(4434)] = 23329,
  [SMALL_STATE(4435)] = 23404,
  [SMALL_STATE(4436)] = 23479,
  [SMALL_STATE(4437)] = 23554,
  [SMALL_STATE(4438)] = 23629,
  [SMALL_STATE(4439)] = 23706,
  [SMALL_STATE(4440)] = 23783,
  [SMALL_STATE(4441)] = 23860,
  [SMALL_STATE(4442)] = 23937,
  [SMALL_STATE(4443)] = 24014,
  [SMALL_STATE(4444)] = 24091,
  [SMALL_STATE(4445)] = 24166,
  [SMALL_STATE(4446)] = 24243,
  [SMALL_STATE(4447)] = 24320,
  [SMALL_STATE(4448)] = 24397,
  [SMALL_STATE(4449)] = 24474,
  [SMALL_STATE(4450)] = 24551,
  [SMALL_STATE(4451)] = 24628,
  [SMALL_STATE(4452)] = 24705,
  [SMALL_STATE(4453)] = 24782,
  [SMALL_STATE(4454)] = 24859,
  [SMALL_STATE(4455)] = 24936,
  [SMALL_STATE(4456)] = 25013,
  [SMALL_STATE(4457)] = 25090,
  [SMALL_STATE(4458)] = 25167,
  [SMALL_STATE(4459)] = 25240,
  [SMALL_STATE(4460)] = 25313,
  [SMALL_STATE(4461)] = 25390,
  [SMALL_STATE(4462)] = 25467,
  [SMALL_STATE(4463)] = 25542,
  [SMALL_STATE(4464)] = 25617,
  [SMALL_STATE(4465)] = 25694,
  [SMALL_STATE(4466)] = 25771,
  [SMALL_STATE(4467)] = 25848,
  [SMALL_STATE(4468)] = 25925,
  [SMALL_STATE(4469)] = 26002,
  [SMALL_STATE(4470)] = 26079,
  [SMALL_STATE(4471)] = 26118,
  [SMALL_STATE(4472)] = 26195,
  [SMALL_STATE(4473)] = 26270,
  [SMALL_STATE(4474)] = 26345,
  [SMALL_STATE(4475)] = 26420,
  [SMALL_STATE(4476)] = 26497,
  [SMALL_STATE(4477)] = 26572,
  [SMALL_STATE(4478)] = 26649,
  [SMALL_STATE(4479)] = 26726,
  [SMALL_STATE(4480)] = 26803,
  [SMALL_STATE(4481)] = 26880,
  [SMALL_STATE(4482)] = 26957,
  [SMALL_STATE(4483)] = 27034,
  [SMALL_STATE(4484)] = 27109,
  [SMALL_STATE(4485)] = 27186,
  [SMALL_STATE(4486)] = 27260,
  [SMALL_STATE(4487)] = 27334,
  [SMALL_STATE(4488)] = 27408,
  [SMALL_STATE(4489)] = 27482,
  [SMALL_STATE(4490)] = 27556,
  [SMALL_STATE(4491)] = 27630,
  [SMALL_STATE(4492)] = 27704,
  [SMALL_STATE(4493)] = 27758,
  [SMALL_STATE(4494)] = 27832,
  [SMALL_STATE(4495)] = 27906,
  [SMALL_STATE(4496)] = 27980,
  [SMALL_STATE(4497)] = 28054,
  [SMALL_STATE(4498)] = 28128,
  [SMALL_STATE(4499)] = 28202,
  [SMALL_STATE(4500)] = 28276,
  [SMALL_STATE(4501)] = 28350,
  [SMALL_STATE(4502)] = 28424,
  [SMALL_STATE(4503)] = 28498,
  [SMALL_STATE(4504)] = 28572,
  [SMALL_STATE(4505)] = 28646,
  [SMALL_STATE(4506)] = 28720,
  [SMALL_STATE(4507)] = 28794,
  [SMALL_STATE(4508)] = 28868,
  [SMALL_STATE(4509)] = 28942,
  [SMALL_STATE(4510)] = 29016,
  [SMALL_STATE(4511)] = 29090,
  [SMALL_STATE(4512)] = 29164,
  [SMALL_STATE(4513)] = 29238,
  [SMALL_STATE(4514)] = 29312,
  [SMALL_STATE(4515)] = 29386,
  [SMALL_STATE(4516)] = 29460,
  [SMALL_STATE(4517)] = 29534,
  [SMALL_STATE(4518)] = 29608,
  [SMALL_STATE(4519)] = 29682,
  [SMALL_STATE(4520)] = 29756,
  [SMALL_STATE(4521)] = 29830,
  [SMALL_STATE(4522)] = 29904,
  [SMALL_STATE(4523)] = 29978,
  [SMALL_STATE(4524)] = 30052,
  [SMALL_STATE(4525)] = 30126,
  [SMALL_STATE(4526)] = 30200,
  [SMALL_STATE(4527)] = 30274,
  [SMALL_STATE(4528)] = 30348,
  [SMALL_STATE(4529)] = 30422,
  [SMALL_STATE(4530)] = 30496,
  [SMALL_STATE(4531)] = 30570,
  [SMALL_STATE(4532)] = 30644,
  [SMALL_STATE(4533)] = 30718,
  [SMALL_STATE(4534)] = 30792,
  [SMALL_STATE(4535)] = 30866,
  [SMALL_STATE(4536)] = 30940,
  [SMALL_STATE(4537)] = 31014,
  [SMALL_STATE(4538)] = 31088,
  [SMALL_STATE(4539)] = 31162,
  [SMALL_STATE(4540)] = 31236,
  [SMALL_STATE(4541)] = 31310,
  [SMALL_STATE(4542)] = 31384,
  [SMALL_STATE(4543)] = 31458,
  [SMALL_STATE(4544)] = 31532,
  [SMALL_STATE(4545)] = 31586,
  [SMALL_STATE(4546)] = 31660,
  [SMALL_STATE(4547)] = 31734,
  [SMALL_STATE(4548)] = 31808,
  [SMALL_STATE(4549)] = 31882,
  [SMALL_STATE(4550)] = 31956,
  [SMALL_STATE(4551)] = 32030,
  [SMALL_STATE(4552)] = 32104,
  [SMALL_STATE(4553)] = 32178,
  [SMALL_STATE(4554)] = 32232,
  [SMALL_STATE(4555)] = 32306,
  [SMALL_STATE(4556)] = 32360,
  [SMALL_STATE(4557)] = 32434,
  [SMALL_STATE(4558)] = 32508,
  [SMALL_STATE(4559)] = 32562,
  [SMALL_STATE(4560)] = 32636,
  [SMALL_STATE(4561)] = 32690,
  [SMALL_STATE(4562)] = 32764,
  [SMALL_STATE(4563)] = 32838,
  [SMALL_STATE(4564)] = 32875,
  [SMALL_STATE(4565)] = 32912,
  [SMALL_STATE(4566)] = 32983,
  [SMALL_STATE(4567)] = 33035,
  [SMALL_STATE(4568)] = 33079,
  [SMALL_STATE(4569)] = 33113,
  [SMALL_STATE(4570)] = 33157,
  [SMALL_STATE(4571)] = 33201,
  [SMALL_STATE(4572)] = 33242,
  [SMALL_STATE(4573)] = 33283,
  [SMALL_STATE(4574)] = 33324,
  [SMALL_STATE(4575)] = 33365,
  [SMALL_STATE(4576)] = 33406,
  [SMALL_STATE(4577)] = 33447,
  [SMALL_STATE(4578)] = 33488,
  [SMALL_STATE(4579)] = 33529,
  [SMALL_STATE(4580)] = 33584,
  [SMALL_STATE(4581)] = 33625,
  [SMALL_STATE(4582)] = 33666,
  [SMALL_STATE(4583)] = 33707,
  [SMALL_STATE(4584)] = 33762,
  [SMALL_STATE(4585)] = 33803,
  [SMALL_STATE(4586)] = 33844,
  [SMALL_STATE(4587)] = 33885,
  [SMALL_STATE(4588)] = 33926,
  [SMALL_STATE(4589)] = 33967,
  [SMALL_STATE(4590)] = 34008,
  [SMALL_STATE(4591)] = 34049,
  [SMALL_STATE(4592)] = 34090,
  [SMALL_STATE(4593)] = 34131,
  [SMALL_STATE(4594)] = 34172,
  [SMALL_STATE(4595)] = 34213,
  [SMALL_STATE(4596)] = 34254,
  [SMALL_STATE(4597)] = 34295,
  [SMALL_STATE(4598)] = 34336,
  [SMALL_STATE(4599)] = 34377,
  [SMALL_STATE(4600)] = 34418,
  [SMALL_STATE(4601)] = 34459,
  [SMALL_STATE(4602)] = 34500,
  [SMALL_STATE(4603)] = 34541,
  [SMALL_STATE(4604)] = 34596,
  [SMALL_STATE(4605)] = 34651,
  [SMALL_STATE(4606)] = 34692,
  [SMALL_STATE(4607)] = 34747,
  [SMALL_STATE(4608)] = 34788,
  [SMALL_STATE(4609)] = 34829,
  [SMALL_STATE(4610)] = 34870,
  [SMALL_STATE(4611)] = 34911,
  [SMALL_STATE(4612)] = 34952,
  [SMALL_STATE(4613)] = 34993,
  [SMALL_STATE(4614)] = 35026,
  [SMALL_STATE(4615)] = 35067,
  [SMALL_STATE(4616)] = 35108,
  [SMALL_STATE(4617)] = 35149,
  [SMALL_STATE(4618)] = 35190,
  [SMALL_STATE(4619)] = 35245,
  [SMALL_STATE(4620)] = 35286,
  [SMALL_STATE(4621)] = 35327,
  [SMALL_STATE(4622)] = 35368,
  [SMALL_STATE(4623)] = 35409,
  [SMALL_STATE(4624)] = 35437,
  [SMALL_STATE(4625)] = 35471,
  [SMALL_STATE(4626)] = 35499,
  [SMALL_STATE(4627)] = 35527,
  [SMALL_STATE(4628)] = 35555,
  [SMALL_STATE(4629)] = 35599,
  [SMALL_STATE(4630)] = 35627,
  [SMALL_STATE(4631)] = 35677,
  [SMALL_STATE(4632)] = 35705,
  [SMALL_STATE(4633)] = 35737,
  [SMALL_STATE(4634)] = 35765,
  [SMALL_STATE(4635)] = 35809,
  [SMALL_STATE(4636)] = 35837,
  [SMALL_STATE(4637)] = 35885,
  [SMALL_STATE(4638)] = 35931,
  [SMALL_STATE(4639)] = 35975,
  [SMALL_STATE(4640)] = 36003,
  [SMALL_STATE(4641)] = 36031,
  [SMALL_STATE(4642)] = 36077,
  [SMALL_STATE(4643)] = 36121,
  [SMALL_STATE(4644)] = 36163,
  [SMALL_STATE(4645)] = 36203,
  [SMALL_STATE(4646)] = 36239,
  [SMALL_STATE(4647)] = 36278,
  [SMALL_STATE(4648)] = 36323,
  [SMALL_STATE(4649)] = 36368,
  [SMALL_STATE(4650)] = 36413,
  [SMALL_STATE(4651)] = 36458,
  [SMALL_STATE(4652)] = 36507,
  [SMALL_STATE(4653)] = 36544,
  [SMALL_STATE(4654)] = 36579,
  [SMALL_STATE(4655)] = 36612,
  [SMALL_STATE(4656)] = 36657,
  [SMALL_STATE(4657)] = 36684,
  [SMALL_STATE(4658)] = 36729,
  [SMALL_STATE(4659)] = 36774,
  [SMALL_STATE(4660)] = 36819,
  [SMALL_STATE(4661)] = 36850,
  [SMALL_STATE(4662)] = 36895,
  [SMALL_STATE(4663)] = 36940,
  [SMALL_STATE(4664)] = 36967,
  [SMALL_STATE(4665)] = 37012,
  [SMALL_STATE(4666)] = 37039,
  [SMALL_STATE(4667)] = 37084,
  [SMALL_STATE(4668)] = 37111,
  [SMALL_STATE(4669)] = 37138,
  [SMALL_STATE(4670)] = 37187,
  [SMALL_STATE(4671)] = 37216,
  [SMALL_STATE(4672)] = 37261,
  [SMALL_STATE(4673)] = 37288,
  [SMALL_STATE(4674)] = 37315,
  [SMALL_STATE(4675)] = 37342,
  [SMALL_STATE(4676)] = 37387,
  [SMALL_STATE(4677)] = 37432,
  [SMALL_STATE(4678)] = 37475,
  [SMALL_STATE(4679)] = 37520,
  [SMALL_STATE(4680)] = 37565,
  [SMALL_STATE(4681)] = 37610,
  [SMALL_STATE(4682)] = 37655,
  [SMALL_STATE(4683)] = 37700,
  [SMALL_STATE(4684)] = 37741,
  [SMALL_STATE(4685)] = 37768,
  [SMALL_STATE(4686)] = 37813,
  [SMALL_STATE(4687)] = 37840,
  [SMALL_STATE(4688)] = 37884,
  [SMALL_STATE(4689)] = 37928,
  [SMALL_STATE(4690)] = 37972,
  [SMALL_STATE(4691)] = 38016,
  [SMALL_STATE(4692)] = 38060,
  [SMALL_STATE(4693)] = 38104,
  [SMALL_STATE(4694)] = 38148,
  [SMALL_STATE(4695)] = 38192,
  [SMALL_STATE(4696)] = 38236,
  [SMALL_STATE(4697)] = 38280,
  [SMALL_STATE(4698)] = 38324,
  [SMALL_STATE(4699)] = 38368,
  [SMALL_STATE(4700)] = 38412,
  [SMALL_STATE(4701)] = 38456,
  [SMALL_STATE(4702)] = 38500,
  [SMALL_STATE(4703)] = 38544,
  [SMALL_STATE(4704)] = 38588,
  [SMALL_STATE(4705)] = 38632,
  [SMALL_STATE(4706)] = 38676,
  [SMALL_STATE(4707)] = 38720,
  [SMALL_STATE(4708)] = 38764,
  [SMALL_STATE(4709)] = 38808,
  [SMALL_STATE(4710)] = 38852,
  [SMALL_STATE(4711)] = 38896,
  [SMALL_STATE(4712)] = 38940,
  [SMALL_STATE(4713)] = 38984,
  [SMALL_STATE(4714)] = 39028,
  [SMALL_STATE(4715)] = 39072,
  [SMALL_STATE(4716)] = 39116,
  [SMALL_STATE(4717)] = 39160,
  [SMALL_STATE(4718)] = 39204,
  [SMALL_STATE(4719)] = 39248,
  [SMALL_STATE(4720)] = 39292,
  [SMALL_STATE(4721)] = 39336,
  [SMALL_STATE(4722)] = 39380,
  [SMALL_STATE(4723)] = 39424,
  [SMALL_STATE(4724)] = 39468,
  [SMALL_STATE(4725)] = 39512,
  [SMALL_STATE(4726)] = 39556,
  [SMALL_STATE(4727)] = 39600,
  [SMALL_STATE(4728)] = 39644,
  [SMALL_STATE(4729)] = 39688,
  [SMALL_STATE(4730)] = 39732,
  [SMALL_STATE(4731)] = 39776,
  [SMALL_STATE(4732)] = 39820,
  [SMALL_STATE(4733)] = 39864,
  [SMALL_STATE(4734)] = 39908,
  [SMALL_STATE(4735)] = 39952,
  [SMALL_STATE(4736)] = 39996,
  [SMALL_STATE(4737)] = 40040,
  [SMALL_STATE(4738)] = 40084,
  [SMALL_STATE(4739)] = 40128,
  [SMALL_STATE(4740)] = 40172,
  [SMALL_STATE(4741)] = 40216,
  [SMALL_STATE(4742)] = 40260,
  [SMALL_STATE(4743)] = 40304,
  [SMALL_STATE(4744)] = 40348,
  [SMALL_STATE(4745)] = 40392,
  [SMALL_STATE(4746)] = 40436,
  [SMALL_STATE(4747)] = 40480,
  [SMALL_STATE(4748)] = 40524,
  [SMALL_STATE(4749)] = 40568,
  [SMALL_STATE(4750)] = 40612,
  [SMALL_STATE(4751)] = 40656,
  [SMALL_STATE(4752)] = 40700,
  [SMALL_STATE(4753)] = 40744,
  [SMALL_STATE(4754)] = 40788,
  [SMALL_STATE(4755)] = 40832,
  [SMALL_STATE(4756)] = 40876,
  [SMALL_STATE(4757)] = 40920,
  [SMALL_STATE(4758)] = 40964,
  [SMALL_STATE(4759)] = 41008,
  [SMALL_STATE(4760)] = 41052,
  [SMALL_STATE(4761)] = 41096,
  [SMALL_STATE(4762)] = 41140,
  [SMALL_STATE(4763)] = 41184,
  [SMALL_STATE(4764)] = 41228,
  [SMALL_STATE(4765)] = 41272,
  [SMALL_STATE(4766)] = 41316,
  [SMALL_STATE(4767)] = 41360,
  [SMALL_STATE(4768)] = 41404,
  [SMALL_STATE(4769)] = 41448,
  [SMALL_STATE(4770)] = 41492,
  [SMALL_STATE(4771)] = 41536,
  [SMALL_STATE(4772)] = 41580,
  [SMALL_STATE(4773)] = 41624,
  [SMALL_STATE(4774)] = 41668,
  [SMALL_STATE(4775)] = 41712,
  [SMALL_STATE(4776)] = 41756,
  [SMALL_STATE(4777)] = 41800,
  [SMALL_STATE(4778)] = 41844,
  [SMALL_STATE(4779)] = 41888,
  [SMALL_STATE(4780)] = 41932,
  [SMALL_STATE(4781)] = 41976,
  [SMALL_STATE(4782)] = 42020,
  [SMALL_STATE(4783)] = 42064,
  [SMALL_STATE(4784)] = 42108,
  [SMALL_STATE(4785)] = 42152,
  [SMALL_STATE(4786)] = 42196,
  [SMALL_STATE(4787)] = 42240,
  [SMALL_STATE(4788)] = 42281,
  [SMALL_STATE(4789)] = 42318,
  [SMALL_STATE(4790)] = 42357,
  [SMALL_STATE(4791)] = 42395,
  [SMALL_STATE(4792)] = 42433,
  [SMALL_STATE(4793)] = 42471,
  [SMALL_STATE(4794)] = 42509,
  [SMALL_STATE(4795)] = 42543,
  [SMALL_STATE(4796)] = 42584,
  [SMALL_STATE(4797)] = 42619,
  [SMALL_STATE(4798)] = 42660,
  [SMALL_STATE(4799)] = 42695,
  [SMALL_STATE(4800)] = 42730,
  [SMALL_STATE(4801)] = 42771,
  [SMALL_STATE(4802)] = 42806,
  [SMALL_STATE(4803)] = 42841,
  [SMALL_STATE(4804)] = 42876,
  [SMALL_STATE(4805)] = 42907,
  [SMALL_STATE(4806)] = 42942,
  [SMALL_STATE(4807)] = 42977,
  [SMALL_STATE(4808)] = 43012,
  [SMALL_STATE(4809)] = 43047,
  [SMALL_STATE(4810)] = 43082,
  [SMALL_STATE(4811)] = 43114,
  [SMALL_STATE(4812)] = 43146,
  [SMALL_STATE(4813)] = 43178,
  [SMALL_STATE(4814)] = 43210,
  [SMALL_STATE(4815)] = 43242,
  [SMALL_STATE(4816)] = 43270,
  [SMALL_STATE(4817)] = 43302,
  [SMALL_STATE(4818)] = 43334,
  [SMALL_STATE(4819)] = 43366,
  [SMALL_STATE(4820)] = 43398,
  [SMALL_STATE(4821)] = 43430,
  [SMALL_STATE(4822)] = 43460,
  [SMALL_STATE(4823)] = 43492,
  [SMALL_STATE(4824)] = 43525,
  [SMALL_STATE(4825)] = 43556,
  [SMALL_STATE(4826)] = 43585,
  [SMALL_STATE(4827)] = 43616,
  [SMALL_STATE(4828)] = 43649,
  [SMALL_STATE(4829)] = 43682,
  [SMALL_STATE(4830)] = 43713,
  [SMALL_STATE(4831)] = 43740,
  [SMALL_STATE(4832)] = 43771,
  [SMALL_STATE(4833)] = 43800,
  [SMALL_STATE(4834)] = 43831,
  [SMALL_STATE(4835)] = 43858,
  [SMALL_STATE(4836)] = 43876,
  [SMALL_STATE(4837)] = 43894,
  [SMALL_STATE(4838)] = 43918,
  [SMALL_STATE(4839)] = 43936,
  [SMALL_STATE(4840)] = 43954,
  [SMALL_STATE(4841)] = 43972,
  [SMALL_STATE(4842)] = 43996,
  [SMALL_STATE(4843)] = 44014,
  [SMALL_STATE(4844)] = 44038,
  [SMALL_STATE(4845)] = 44064,
  [SMALL_STATE(4846)] = 44090,
  [SMALL_STATE(4847)] = 44114,
  [SMALL_STATE(4848)] = 44132,
  [SMALL_STATE(4849)] = 44158,
  [SMALL_STATE(4850)] = 44182,
  [SMALL_STATE(4851)] = 44206,
  [SMALL_STATE(4852)] = 44232,
  [SMALL_STATE(4853)] = 44256,
  [SMALL_STATE(4854)] = 44280,
  [SMALL_STATE(4855)] = 44306,
  [SMALL_STATE(4856)] = 44330,
  [SMALL_STATE(4857)] = 44348,
  [SMALL_STATE(4858)] = 44372,
  [SMALL_STATE(4859)] = 44396,
  [SMALL_STATE(4860)] = 44420,
  [SMALL_STATE(4861)] = 44438,
  [SMALL_STATE(4862)] = 44460,
  [SMALL_STATE(4863)] = 44484,
  [SMALL_STATE(4864)] = 44506,
  [SMALL_STATE(4865)] = 44524,
  [SMALL_STATE(4866)] = 44550,
  [SMALL_STATE(4867)] = 44576,
  [SMALL_STATE(4868)] = 44602,
  [SMALL_STATE(4869)] = 44626,
  [SMALL_STATE(4870)] = 44644,
  [SMALL_STATE(4871)] = 44668,
  [SMALL_STATE(4872)] = 44697,
  [SMALL_STATE(4873)] = 44726,
  [SMALL_STATE(4874)] = 44753,
  [SMALL_STATE(4875)] = 44780,
  [SMALL_STATE(4876)] = 44809,
  [SMALL_STATE(4877)] = 44838,
  [SMALL_STATE(4878)] = 44865,
  [SMALL_STATE(4879)] = 44886,
  [SMALL_STATE(4880)] = 44915,
  [SMALL_STATE(4881)] = 44936,
  [SMALL_STATE(4882)] = 44959,
  [SMALL_STATE(4883)] = 44982,
  [SMALL_STATE(4884)] = 45003,
  [SMALL_STATE(4885)] = 45032,
  [SMALL_STATE(4886)] = 45061,
  [SMALL_STATE(4887)] = 45082,
  [SMALL_STATE(4888)] = 45105,
  [SMALL_STATE(4889)] = 45129,
  [SMALL_STATE(4890)] = 45145,
  [SMALL_STATE(4891)] = 45171,
  [SMALL_STATE(4892)] = 45199,
  [SMALL_STATE(4893)] = 45227,
  [SMALL_STATE(4894)] = 45247,
  [SMALL_STATE(4895)] = 45273,
  [SMALL_STATE(4896)] = 45301,
  [SMALL_STATE(4897)] = 45327,
  [SMALL_STATE(4898)] = 45355,
  [SMALL_STATE(4899)] = 45375,
  [SMALL_STATE(4900)] = 45391,
  [SMALL_STATE(4901)] = 45419,
  [SMALL_STATE(4902)] = 45447,
  [SMALL_STATE(4903)] = 45463,
  [SMALL_STATE(4904)] = 45491,
  [SMALL_STATE(4905)] = 45507,
  [SMALL_STATE(4906)] = 45535,
  [SMALL_STATE(4907)] = 45559,
  [SMALL_STATE(4908)] = 45585,
  [SMALL_STATE(4909)] = 45613,
  [SMALL_STATE(4910)] = 45641,
  [SMALL_STATE(4911)] = 45663,
  [SMALL_STATE(4912)] = 45687,
  [SMALL_STATE(4913)] = 45715,
  [SMALL_STATE(4914)] = 45743,
  [SMALL_STATE(4915)] = 45767,
  [SMALL_STATE(4916)] = 45787,
  [SMALL_STATE(4917)] = 45815,
  [SMALL_STATE(4918)] = 45841,
  [SMALL_STATE(4919)] = 45857,
  [SMALL_STATE(4920)] = 45885,
  [SMALL_STATE(4921)] = 45909,
  [SMALL_STATE(4922)] = 45933,
  [SMALL_STATE(4923)] = 45953,
  [SMALL_STATE(4924)] = 45981,
  [SMALL_STATE(4925)] = 46001,
  [SMALL_STATE(4926)] = 46017,
  [SMALL_STATE(4927)] = 46037,
  [SMALL_STATE(4928)] = 46061,
  [SMALL_STATE(4929)] = 46077,
  [SMALL_STATE(4930)] = 46105,
  [SMALL_STATE(4931)] = 46125,
  [SMALL_STATE(4932)] = 46153,
  [SMALL_STATE(4933)] = 46179,
  [SMALL_STATE(4934)] = 46207,
  [SMALL_STATE(4935)] = 46235,
  [SMALL_STATE(4936)] = 46263,
  [SMALL_STATE(4937)] = 46291,
  [SMALL_STATE(4938)] = 46319,
  [SMALL_STATE(4939)] = 46337,
  [SMALL_STATE(4940)] = 46361,
  [SMALL_STATE(4941)] = 46389,
  [SMALL_STATE(4942)] = 46417,
  [SMALL_STATE(4943)] = 46445,
  [SMALL_STATE(4944)] = 46473,
  [SMALL_STATE(4945)] = 46497,
  [SMALL_STATE(4946)] = 46525,
  [SMALL_STATE(4947)] = 46541,
  [SMALL_STATE(4948)] = 46569,
  [SMALL_STATE(4949)] = 46597,
  [SMALL_STATE(4950)] = 46620,
  [SMALL_STATE(4951)] = 46637,
  [SMALL_STATE(4952)] = 46662,
  [SMALL_STATE(4953)] = 46679,
  [SMALL_STATE(4954)] = 46698,
  [SMALL_STATE(4955)] = 46715,
  [SMALL_STATE(4956)] = 46732,
  [SMALL_STATE(4957)] = 46753,
  [SMALL_STATE(4958)] = 46772,
  [SMALL_STATE(4959)] = 46791,
  [SMALL_STATE(4960)] = 46816,
  [SMALL_STATE(4961)] = 46837,
  [SMALL_STATE(4962)] = 46854,
  [SMALL_STATE(4963)] = 46879,
  [SMALL_STATE(4964)] = 46896,
  [SMALL_STATE(4965)] = 46921,
  [SMALL_STATE(4966)] = 46942,
  [SMALL_STATE(4967)] = 46963,
  [SMALL_STATE(4968)] = 46982,
  [SMALL_STATE(4969)] = 46999,
  [SMALL_STATE(4970)] = 47020,
  [SMALL_STATE(4971)] = 47037,
  [SMALL_STATE(4972)] = 47062,
  [SMALL_STATE(4973)] = 47079,
  [SMALL_STATE(4974)] = 47104,
  [SMALL_STATE(4975)] = 47125,
  [SMALL_STATE(4976)] = 47150,
  [SMALL_STATE(4977)] = 47169,
  [SMALL_STATE(4978)] = 47188,
  [SMALL_STATE(4979)] = 47207,
  [SMALL_STATE(4980)] = 47226,
  [SMALL_STATE(4981)] = 47249,
  [SMALL_STATE(4982)] = 47274,
  [SMALL_STATE(4983)] = 47293,
  [SMALL_STATE(4984)] = 47312,
  [SMALL_STATE(4985)] = 47333,
  [SMALL_STATE(4986)] = 47350,
  [SMALL_STATE(4987)] = 47367,
  [SMALL_STATE(4988)] = 47388,
  [SMALL_STATE(4989)] = 47409,
  [SMALL_STATE(4990)] = 47430,
  [SMALL_STATE(4991)] = 47451,
  [SMALL_STATE(4992)] = 47472,
  [SMALL_STATE(4993)] = 47489,
  [SMALL_STATE(4994)] = 47506,
  [SMALL_STATE(4995)] = 47525,
  [SMALL_STATE(4996)] = 47544,
  [SMALL_STATE(4997)] = 47569,
  [SMALL_STATE(4998)] = 47586,
  [SMALL_STATE(4999)] = 47603,
  [SMALL_STATE(5000)] = 47622,
  [SMALL_STATE(5001)] = 47641,
  [SMALL_STATE(5002)] = 47658,
  [SMALL_STATE(5003)] = 47681,
  [SMALL_STATE(5004)] = 47698,
  [SMALL_STATE(5005)] = 47720,
  [SMALL_STATE(5006)] = 47742,
  [SMALL_STATE(5007)] = 47764,
  [SMALL_STATE(5008)] = 47786,
  [SMALL_STATE(5009)] = 47808,
  [SMALL_STATE(5010)] = 47830,
  [SMALL_STATE(5011)] = 47852,
  [SMALL_STATE(5012)] = 47874,
  [SMALL_STATE(5013)] = 47892,
  [SMALL_STATE(5014)] = 47910,
  [SMALL_STATE(5015)] = 47932,
  [SMALL_STATE(5016)] = 47954,
  [SMALL_STATE(5017)] = 47976,
  [SMALL_STATE(5018)] = 47998,
  [SMALL_STATE(5019)] = 48020,
  [SMALL_STATE(5020)] = 48042,
  [SMALL_STATE(5021)] = 48060,
  [SMALL_STATE(5022)] = 48078,
  [SMALL_STATE(5023)] = 48100,
  [SMALL_STATE(5024)] = 48122,
  [SMALL_STATE(5025)] = 48140,
  [SMALL_STATE(5026)] = 48162,
  [SMALL_STATE(5027)] = 48182,
  [SMALL_STATE(5028)] = 48204,
  [SMALL_STATE(5029)] = 48224,
  [SMALL_STATE(5030)] = 48246,
  [SMALL_STATE(5031)] = 48266,
  [SMALL_STATE(5032)] = 48288,
  [SMALL_STATE(5033)] = 48306,
  [SMALL_STATE(5034)] = 48324,
  [SMALL_STATE(5035)] = 48346,
  [SMALL_STATE(5036)] = 48366,
  [SMALL_STATE(5037)] = 48386,
  [SMALL_STATE(5038)] = 48408,
  [SMALL_STATE(5039)] = 48428,
  [SMALL_STATE(5040)] = 48450,
  [SMALL_STATE(5041)] = 48472,
  [SMALL_STATE(5042)] = 48494,
  [SMALL_STATE(5043)] = 48514,
  [SMALL_STATE(5044)] = 48532,
  [SMALL_STATE(5045)] = 48554,
  [SMALL_STATE(5046)] = 48574,
  [SMALL_STATE(5047)] = 48596,
  [SMALL_STATE(5048)] = 48618,
  [SMALL_STATE(5049)] = 48636,
  [SMALL_STATE(5050)] = 48658,
  [SMALL_STATE(5051)] = 48676,
  [SMALL_STATE(5052)] = 48694,
  [SMALL_STATE(5053)] = 48714,
  [SMALL_STATE(5054)] = 48736,
  [SMALL_STATE(5055)] = 48758,
  [SMALL_STATE(5056)] = 48780,
  [SMALL_STATE(5057)] = 48802,
  [SMALL_STATE(5058)] = 48824,
  [SMALL_STATE(5059)] = 48842,
  [SMALL_STATE(5060)] = 48864,
  [SMALL_STATE(5061)] = 48886,
  [SMALL_STATE(5062)] = 48906,
  [SMALL_STATE(5063)] = 48928,
  [SMALL_STATE(5064)] = 48948,
  [SMALL_STATE(5065)] = 48968,
  [SMALL_STATE(5066)] = 48988,
  [SMALL_STATE(5067)] = 49008,
  [SMALL_STATE(5068)] = 49030,
  [SMALL_STATE(5069)] = 49052,
  [SMALL_STATE(5070)] = 49070,
  [SMALL_STATE(5071)] = 49092,
  [SMALL_STATE(5072)] = 49114,
  [SMALL_STATE(5073)] = 49136,
  [SMALL_STATE(5074)] = 49154,
  [SMALL_STATE(5075)] = 49168,
  [SMALL_STATE(5076)] = 49190,
  [SMALL_STATE(5077)] = 49212,
  [SMALL_STATE(5078)] = 49234,
  [SMALL_STATE(5079)] = 49254,
  [SMALL_STATE(5080)] = 49274,
  [SMALL_STATE(5081)] = 49296,
  [SMALL_STATE(5082)] = 49318,
  [SMALL_STATE(5083)] = 49338,
  [SMALL_STATE(5084)] = 49360,
  [SMALL_STATE(5085)] = 49382,
  [SMALL_STATE(5086)] = 49400,
  [SMALL_STATE(5087)] = 49420,
  [SMALL_STATE(5088)] = 49442,
  [SMALL_STATE(5089)] = 49462,
  [SMALL_STATE(5090)] = 49482,
  [SMALL_STATE(5091)] = 49504,
  [SMALL_STATE(5092)] = 49526,
  [SMALL_STATE(5093)] = 49546,
  [SMALL_STATE(5094)] = 49568,
  [SMALL_STATE(5095)] = 49588,
  [SMALL_STATE(5096)] = 49607,
  [SMALL_STATE(5097)] = 49618,
  [SMALL_STATE(5098)] = 49633,
  [SMALL_STATE(5099)] = 49644,
  [SMALL_STATE(5100)] = 49663,
  [SMALL_STATE(5101)] = 49674,
  [SMALL_STATE(5102)] = 49685,
  [SMALL_STATE(5103)] = 49696,
  [SMALL_STATE(5104)] = 49711,
  [SMALL_STATE(5105)] = 49722,
  [SMALL_STATE(5106)] = 49733,
  [SMALL_STATE(5107)] = 49748,
  [SMALL_STATE(5108)] = 49759,
  [SMALL_STATE(5109)] = 49776,
  [SMALL_STATE(5110)] = 49787,
  [SMALL_STATE(5111)] = 49802,
  [SMALL_STATE(5112)] = 49817,
  [SMALL_STATE(5113)] = 49830,
  [SMALL_STATE(5114)] = 49845,
  [SMALL_STATE(5115)] = 49860,
  [SMALL_STATE(5116)] = 49875,
  [SMALL_STATE(5117)] = 49890,
  [SMALL_STATE(5118)] = 49901,
  [SMALL_STATE(5119)] = 49920,
  [SMALL_STATE(5120)] = 49935,
  [SMALL_STATE(5121)] = 49954,
  [SMALL_STATE(5122)] = 49965,
  [SMALL_STATE(5123)] = 49980,
  [SMALL_STATE(5124)] = 49995,
  [SMALL_STATE(5125)] = 50006,
  [SMALL_STATE(5126)] = 50025,
  [SMALL_STATE(5127)] = 50044,
  [SMALL_STATE(5128)] = 50063,
  [SMALL_STATE(5129)] = 50078,
  [SMALL_STATE(5130)] = 50093,
  [SMALL_STATE(5131)] = 50108,
  [SMALL_STATE(5132)] = 50123,
  [SMALL_STATE(5133)] = 50138,
  [SMALL_STATE(5134)] = 50149,
  [SMALL_STATE(5135)] = 50164,
  [SMALL_STATE(5136)] = 50175,
  [SMALL_STATE(5137)] = 50190,
  [SMALL_STATE(5138)] = 50209,
  [SMALL_STATE(5139)] = 50228,
  [SMALL_STATE(5140)] = 50241,
  [SMALL_STATE(5141)] = 50260,
  [SMALL_STATE(5142)] = 50275,
  [SMALL_STATE(5143)] = 50290,
  [SMALL_STATE(5144)] = 50303,
  [SMALL_STATE(5145)] = 50322,
  [SMALL_STATE(5146)] = 50333,
  [SMALL_STATE(5147)] = 50348,
  [SMALL_STATE(5148)] = 50359,
  [SMALL_STATE(5149)] = 50378,
  [SMALL_STATE(5150)] = 50397,
  [SMALL_STATE(5151)] = 50412,
  [SMALL_STATE(5152)] = 50427,
  [SMALL_STATE(5153)] = 50438,
  [SMALL_STATE(5154)] = 50453,
  [SMALL_STATE(5155)] = 50470,
  [SMALL_STATE(5156)] = 50489,
  [SMALL_STATE(5157)] = 50500,
  [SMALL_STATE(5158)] = 50511,
  [SMALL_STATE(5159)] = 50526,
  [SMALL_STATE(5160)] = 50537,
  [SMALL_STATE(5161)] = 50556,
  [SMALL_STATE(5162)] = 50567,
  [SMALL_STATE(5163)] = 50578,
  [SMALL_STATE(5164)] = 50594,
  [SMALL_STATE(5165)] = 50610,
  [SMALL_STATE(5166)] = 50626,
  [SMALL_STATE(5167)] = 50642,
  [SMALL_STATE(5168)] = 50658,
  [SMALL_STATE(5169)] = 50674,
  [SMALL_STATE(5170)] = 50688,
  [SMALL_STATE(5171)] = 50704,
  [SMALL_STATE(5172)] = 50720,
  [SMALL_STATE(5173)] = 50736,
  [SMALL_STATE(5174)] = 50752,
  [SMALL_STATE(5175)] = 50768,
  [SMALL_STATE(5176)] = 50784,
  [SMALL_STATE(5177)] = 50800,
  [SMALL_STATE(5178)] = 50816,
  [SMALL_STATE(5179)] = 50832,
  [SMALL_STATE(5180)] = 50848,
  [SMALL_STATE(5181)] = 50864,
  [SMALL_STATE(5182)] = 50880,
  [SMALL_STATE(5183)] = 50894,
  [SMALL_STATE(5184)] = 50910,
  [SMALL_STATE(5185)] = 50926,
  [SMALL_STATE(5186)] = 50942,
  [SMALL_STATE(5187)] = 50958,
  [SMALL_STATE(5188)] = 50974,
  [SMALL_STATE(5189)] = 50990,
  [SMALL_STATE(5190)] = 51006,
  [SMALL_STATE(5191)] = 51022,
  [SMALL_STATE(5192)] = 51038,
  [SMALL_STATE(5193)] = 51054,
  [SMALL_STATE(5194)] = 51070,
  [SMALL_STATE(5195)] = 51084,
  [SMALL_STATE(5196)] = 51100,
  [SMALL_STATE(5197)] = 51116,
  [SMALL_STATE(5198)] = 51132,
  [SMALL_STATE(5199)] = 51148,
  [SMALL_STATE(5200)] = 51164,
  [SMALL_STATE(5201)] = 51180,
  [SMALL_STATE(5202)] = 51194,
  [SMALL_STATE(5203)] = 51208,
  [SMALL_STATE(5204)] = 51224,
  [SMALL_STATE(5205)] = 51240,
  [SMALL_STATE(5206)] = 51256,
  [SMALL_STATE(5207)] = 51272,
  [SMALL_STATE(5208)] = 51288,
  [SMALL_STATE(5209)] = 51302,
  [SMALL_STATE(5210)] = 51318,
  [SMALL_STATE(5211)] = 51334,
  [SMALL_STATE(5212)] = 51350,
  [SMALL_STATE(5213)] = 51366,
  [SMALL_STATE(5214)] = 51382,
  [SMALL_STATE(5215)] = 51398,
  [SMALL_STATE(5216)] = 51414,
  [SMALL_STATE(5217)] = 51430,
  [SMALL_STATE(5218)] = 51446,
  [SMALL_STATE(5219)] = 51462,
  [SMALL_STATE(5220)] = 51478,
  [SMALL_STATE(5221)] = 51494,
  [SMALL_STATE(5222)] = 51510,
  [SMALL_STATE(5223)] = 51526,
  [SMALL_STATE(5224)] = 51542,
  [SMALL_STATE(5225)] = 51556,
  [SMALL_STATE(5226)] = 51572,
  [SMALL_STATE(5227)] = 51582,
  [SMALL_STATE(5228)] = 51598,
  [SMALL_STATE(5229)] = 51614,
  [SMALL_STATE(5230)] = 51628,
  [SMALL_STATE(5231)] = 51644,
  [SMALL_STATE(5232)] = 51660,
  [SMALL_STATE(5233)] = 51674,
  [SMALL_STATE(5234)] = 51690,
  [SMALL_STATE(5235)] = 51706,
  [SMALL_STATE(5236)] = 51720,
  [SMALL_STATE(5237)] = 51736,
  [SMALL_STATE(5238)] = 51750,
  [SMALL_STATE(5239)] = 51766,
  [SMALL_STATE(5240)] = 51782,
  [SMALL_STATE(5241)] = 51798,
  [SMALL_STATE(5242)] = 51814,
  [SMALL_STATE(5243)] = 51830,
  [SMALL_STATE(5244)] = 51844,
  [SMALL_STATE(5245)] = 51860,
  [SMALL_STATE(5246)] = 51876,
  [SMALL_STATE(5247)] = 51892,
  [SMALL_STATE(5248)] = 51906,
  [SMALL_STATE(5249)] = 51922,
  [SMALL_STATE(5250)] = 51938,
  [SMALL_STATE(5251)] = 51954,
  [SMALL_STATE(5252)] = 51970,
  [SMALL_STATE(5253)] = 51984,
  [SMALL_STATE(5254)] = 52000,
  [SMALL_STATE(5255)] = 52016,
  [SMALL_STATE(5256)] = 52032,
  [SMALL_STATE(5257)] = 52046,
  [SMALL_STATE(5258)] = 52062,
  [SMALL_STATE(5259)] = 52078,
  [SMALL_STATE(5260)] = 52094,
  [SMALL_STATE(5261)] = 52110,
  [SMALL_STATE(5262)] = 52126,
  [SMALL_STATE(5263)] = 52142,
  [SMALL_STATE(5264)] = 52158,
  [SMALL_STATE(5265)] = 52174,
  [SMALL_STATE(5266)] = 52188,
  [SMALL_STATE(5267)] = 52204,
  [SMALL_STATE(5268)] = 52220,
  [SMALL_STATE(5269)] = 52236,
  [SMALL_STATE(5270)] = 52252,
  [SMALL_STATE(5271)] = 52268,
  [SMALL_STATE(5272)] = 52282,
  [SMALL_STATE(5273)] = 52296,
  [SMALL_STATE(5274)] = 52312,
  [SMALL_STATE(5275)] = 52328,
  [SMALL_STATE(5276)] = 52344,
  [SMALL_STATE(5277)] = 52360,
  [SMALL_STATE(5278)] = 52376,
  [SMALL_STATE(5279)] = 52392,
  [SMALL_STATE(5280)] = 52406,
  [SMALL_STATE(5281)] = 52422,
  [SMALL_STATE(5282)] = 52438,
  [SMALL_STATE(5283)] = 52454,
  [SMALL_STATE(5284)] = 52470,
  [SMALL_STATE(5285)] = 52486,
  [SMALL_STATE(5286)] = 52502,
  [SMALL_STATE(5287)] = 52518,
  [SMALL_STATE(5288)] = 52534,
  [SMALL_STATE(5289)] = 52550,
  [SMALL_STATE(5290)] = 52564,
  [SMALL_STATE(5291)] = 52580,
  [SMALL_STATE(5292)] = 52594,
  [SMALL_STATE(5293)] = 52610,
  [SMALL_STATE(5294)] = 52624,
  [SMALL_STATE(5295)] = 52638,
  [SMALL_STATE(5296)] = 52654,
  [SMALL_STATE(5297)] = 52670,
  [SMALL_STATE(5298)] = 52686,
  [SMALL_STATE(5299)] = 52702,
  [SMALL_STATE(5300)] = 52718,
  [SMALL_STATE(5301)] = 52732,
  [SMALL_STATE(5302)] = 52748,
  [SMALL_STATE(5303)] = 52764,
  [SMALL_STATE(5304)] = 52780,
  [SMALL_STATE(5305)] = 52796,
  [SMALL_STATE(5306)] = 52812,
  [SMALL_STATE(5307)] = 52828,
  [SMALL_STATE(5308)] = 52844,
  [SMALL_STATE(5309)] = 52860,
  [SMALL_STATE(5310)] = 52876,
  [SMALL_STATE(5311)] = 52890,
  [SMALL_STATE(5312)] = 52906,
  [SMALL_STATE(5313)] = 52922,
  [SMALL_STATE(5314)] = 52938,
  [SMALL_STATE(5315)] = 52951,
  [SMALL_STATE(5316)] = 52964,
  [SMALL_STATE(5317)] = 52977,
  [SMALL_STATE(5318)] = 52990,
  [SMALL_STATE(5319)] = 53003,
  [SMALL_STATE(5320)] = 53016,
  [SMALL_STATE(5321)] = 53029,
  [SMALL_STATE(5322)] = 53042,
  [SMALL_STATE(5323)] = 53055,
  [SMALL_STATE(5324)] = 53068,
  [SMALL_STATE(5325)] = 53081,
  [SMALL_STATE(5326)] = 53094,
  [SMALL_STATE(5327)] = 53107,
  [SMALL_STATE(5328)] = 53120,
  [SMALL_STATE(5329)] = 53133,
  [SMALL_STATE(5330)] = 53146,
  [SMALL_STATE(5331)] = 53159,
  [SMALL_STATE(5332)] = 53172,
  [SMALL_STATE(5333)] = 53185,
  [SMALL_STATE(5334)] = 53198,
  [SMALL_STATE(5335)] = 53211,
  [SMALL_STATE(5336)] = 53224,
  [SMALL_STATE(5337)] = 53235,
  [SMALL_STATE(5338)] = 53248,
  [SMALL_STATE(5339)] = 53261,
  [SMALL_STATE(5340)] = 53274,
  [SMALL_STATE(5341)] = 53287,
  [SMALL_STATE(5342)] = 53300,
  [SMALL_STATE(5343)] = 53313,
  [SMALL_STATE(5344)] = 53326,
  [SMALL_STATE(5345)] = 53339,
  [SMALL_STATE(5346)] = 53352,
  [SMALL_STATE(5347)] = 53365,
  [SMALL_STATE(5348)] = 53378,
  [SMALL_STATE(5349)] = 53391,
  [SMALL_STATE(5350)] = 53404,
  [SMALL_STATE(5351)] = 53417,
  [SMALL_STATE(5352)] = 53430,
  [SMALL_STATE(5353)] = 53443,
  [SMALL_STATE(5354)] = 53452,
  [SMALL_STATE(5355)] = 53465,
  [SMALL_STATE(5356)] = 53478,
  [SMALL_STATE(5357)] = 53491,
  [SMALL_STATE(5358)] = 53504,
  [SMALL_STATE(5359)] = 53517,
  [SMALL_STATE(5360)] = 53530,
  [SMALL_STATE(5361)] = 53543,
  [SMALL_STATE(5362)] = 53556,
  [SMALL_STATE(5363)] = 53569,
  [SMALL_STATE(5364)] = 53582,
  [SMALL_STATE(5365)] = 53593,
  [SMALL_STATE(5366)] = 53606,
  [SMALL_STATE(5367)] = 53619,
  [SMALL_STATE(5368)] = 53632,
  [SMALL_STATE(5369)] = 53645,
  [SMALL_STATE(5370)] = 53658,
  [SMALL_STATE(5371)] = 53671,
  [SMALL_STATE(5372)] = 53684,
  [SMALL_STATE(5373)] = 53697,
  [SMALL_STATE(5374)] = 53710,
  [SMALL_STATE(5375)] = 53723,
  [SMALL_STATE(5376)] = 53736,
  [SMALL_STATE(5377)] = 53749,
  [SMALL_STATE(5378)] = 53762,
  [SMALL_STATE(5379)] = 53775,
  [SMALL_STATE(5380)] = 53788,
  [SMALL_STATE(5381)] = 53797,
  [SMALL_STATE(5382)] = 53810,
  [SMALL_STATE(5383)] = 53823,
  [SMALL_STATE(5384)] = 53836,
  [SMALL_STATE(5385)] = 53847,
  [SMALL_STATE(5386)] = 53860,
  [SMALL_STATE(5387)] = 53873,
  [SMALL_STATE(5388)] = 53886,
  [SMALL_STATE(5389)] = 53899,
  [SMALL_STATE(5390)] = 53912,
  [SMALL_STATE(5391)] = 53925,
  [SMALL_STATE(5392)] = 53938,
  [SMALL_STATE(5393)] = 53951,
  [SMALL_STATE(5394)] = 53964,
  [SMALL_STATE(5395)] = 53977,
  [SMALL_STATE(5396)] = 53990,
  [SMALL_STATE(5397)] = 54003,
  [SMALL_STATE(5398)] = 54016,
  [SMALL_STATE(5399)] = 54029,
  [SMALL_STATE(5400)] = 54042,
  [SMALL_STATE(5401)] = 54055,
  [SMALL_STATE(5402)] = 54068,
  [SMALL_STATE(5403)] = 54081,
  [SMALL_STATE(5404)] = 54094,
  [SMALL_STATE(5405)] = 54107,
  [SMALL_STATE(5406)] = 54120,
  [SMALL_STATE(5407)] = 54133,
  [SMALL_STATE(5408)] = 54146,
  [SMALL_STATE(5409)] = 54159,
  [SMALL_STATE(5410)] = 54172,
  [SMALL_STATE(5411)] = 54185,
  [SMALL_STATE(5412)] = 54198,
  [SMALL_STATE(5413)] = 54209,
  [SMALL_STATE(5414)] = 54222,
  [SMALL_STATE(5415)] = 54235,
  [SMALL_STATE(5416)] = 54248,
  [SMALL_STATE(5417)] = 54261,
  [SMALL_STATE(5418)] = 54274,
  [SMALL_STATE(5419)] = 54287,
  [SMALL_STATE(5420)] = 54300,
  [SMALL_STATE(5421)] = 54313,
  [SMALL_STATE(5422)] = 54326,
  [SMALL_STATE(5423)] = 54335,
  [SMALL_STATE(5424)] = 54348,
  [SMALL_STATE(5425)] = 54357,
  [SMALL_STATE(5426)] = 54370,
  [SMALL_STATE(5427)] = 54383,
  [SMALL_STATE(5428)] = 54396,
  [SMALL_STATE(5429)] = 54409,
  [SMALL_STATE(5430)] = 54422,
  [SMALL_STATE(5431)] = 54435,
  [SMALL_STATE(5432)] = 54448,
  [SMALL_STATE(5433)] = 54461,
  [SMALL_STATE(5434)] = 54474,
  [SMALL_STATE(5435)] = 54487,
  [SMALL_STATE(5436)] = 54498,
  [SMALL_STATE(5437)] = 54511,
  [SMALL_STATE(5438)] = 54524,
  [SMALL_STATE(5439)] = 54537,
  [SMALL_STATE(5440)] = 54550,
  [SMALL_STATE(5441)] = 54563,
  [SMALL_STATE(5442)] = 54576,
  [SMALL_STATE(5443)] = 54589,
  [SMALL_STATE(5444)] = 54602,
  [SMALL_STATE(5445)] = 54615,
  [SMALL_STATE(5446)] = 54628,
  [SMALL_STATE(5447)] = 54641,
  [SMALL_STATE(5448)] = 54654,
  [SMALL_STATE(5449)] = 54667,
  [SMALL_STATE(5450)] = 54680,
  [SMALL_STATE(5451)] = 54693,
  [SMALL_STATE(5452)] = 54706,
  [SMALL_STATE(5453)] = 54719,
  [SMALL_STATE(5454)] = 54732,
  [SMALL_STATE(5455)] = 54745,
  [SMALL_STATE(5456)] = 54758,
  [SMALL_STATE(5457)] = 54771,
  [SMALL_STATE(5458)] = 54784,
  [SMALL_STATE(5459)] = 54793,
  [SMALL_STATE(5460)] = 54806,
  [SMALL_STATE(5461)] = 54819,
  [SMALL_STATE(5462)] = 54830,
  [SMALL_STATE(5463)] = 54843,
  [SMALL_STATE(5464)] = 54856,
  [SMALL_STATE(5465)] = 54865,
  [SMALL_STATE(5466)] = 54874,
  [SMALL_STATE(5467)] = 54887,
  [SMALL_STATE(5468)] = 54900,
  [SMALL_STATE(5469)] = 54913,
  [SMALL_STATE(5470)] = 54926,
  [SMALL_STATE(5471)] = 54939,
  [SMALL_STATE(5472)] = 54952,
  [SMALL_STATE(5473)] = 54965,
  [SMALL_STATE(5474)] = 54978,
  [SMALL_STATE(5475)] = 54991,
  [SMALL_STATE(5476)] = 55004,
  [SMALL_STATE(5477)] = 55017,
  [SMALL_STATE(5478)] = 55030,
  [SMALL_STATE(5479)] = 55043,
  [SMALL_STATE(5480)] = 55054,
  [SMALL_STATE(5481)] = 55067,
  [SMALL_STATE(5482)] = 55080,
  [SMALL_STATE(5483)] = 55093,
  [SMALL_STATE(5484)] = 55106,
  [SMALL_STATE(5485)] = 55119,
  [SMALL_STATE(5486)] = 55132,
  [SMALL_STATE(5487)] = 55145,
  [SMALL_STATE(5488)] = 55158,
  [SMALL_STATE(5489)] = 55171,
  [SMALL_STATE(5490)] = 55184,
  [SMALL_STATE(5491)] = 55197,
  [SMALL_STATE(5492)] = 55210,
  [SMALL_STATE(5493)] = 55223,
  [SMALL_STATE(5494)] = 55236,
  [SMALL_STATE(5495)] = 55249,
  [SMALL_STATE(5496)] = 55262,
  [SMALL_STATE(5497)] = 55275,
  [SMALL_STATE(5498)] = 55288,
  [SMALL_STATE(5499)] = 55301,
  [SMALL_STATE(5500)] = 55314,
  [SMALL_STATE(5501)] = 55327,
  [SMALL_STATE(5502)] = 55340,
  [SMALL_STATE(5503)] = 55353,
  [SMALL_STATE(5504)] = 55366,
  [SMALL_STATE(5505)] = 55379,
  [SMALL_STATE(5506)] = 55392,
  [SMALL_STATE(5507)] = 55405,
  [SMALL_STATE(5508)] = 55414,
  [SMALL_STATE(5509)] = 55427,
  [SMALL_STATE(5510)] = 55440,
  [SMALL_STATE(5511)] = 55453,
  [SMALL_STATE(5512)] = 55466,
  [SMALL_STATE(5513)] = 55479,
  [SMALL_STATE(5514)] = 55492,
  [SMALL_STATE(5515)] = 55501,
  [SMALL_STATE(5516)] = 55514,
  [SMALL_STATE(5517)] = 55527,
  [SMALL_STATE(5518)] = 55540,
  [SMALL_STATE(5519)] = 55553,
  [SMALL_STATE(5520)] = 55566,
  [SMALL_STATE(5521)] = 55579,
  [SMALL_STATE(5522)] = 55592,
  [SMALL_STATE(5523)] = 55605,
  [SMALL_STATE(5524)] = 55616,
  [SMALL_STATE(5525)] = 55629,
  [SMALL_STATE(5526)] = 55642,
  [SMALL_STATE(5527)] = 55655,
  [SMALL_STATE(5528)] = 55668,
  [SMALL_STATE(5529)] = 55681,
  [SMALL_STATE(5530)] = 55694,
  [SMALL_STATE(5531)] = 55707,
  [SMALL_STATE(5532)] = 55716,
  [SMALL_STATE(5533)] = 55729,
  [SMALL_STATE(5534)] = 55742,
  [SMALL_STATE(5535)] = 55755,
  [SMALL_STATE(5536)] = 55768,
  [SMALL_STATE(5537)] = 55781,
  [SMALL_STATE(5538)] = 55794,
  [SMALL_STATE(5539)] = 55807,
  [SMALL_STATE(5540)] = 55820,
  [SMALL_STATE(5541)] = 55833,
  [SMALL_STATE(5542)] = 55846,
  [SMALL_STATE(5543)] = 55859,
  [SMALL_STATE(5544)] = 55872,
  [SMALL_STATE(5545)] = 55885,
  [SMALL_STATE(5546)] = 55898,
  [SMALL_STATE(5547)] = 55911,
  [SMALL_STATE(5548)] = 55924,
  [SMALL_STATE(5549)] = 55937,
  [SMALL_STATE(5550)] = 55950,
  [SMALL_STATE(5551)] = 55963,
  [SMALL_STATE(5552)] = 55976,
  [SMALL_STATE(5553)] = 55989,
  [SMALL_STATE(5554)] = 56002,
  [SMALL_STATE(5555)] = 56011,
  [SMALL_STATE(5556)] = 56024,
  [SMALL_STATE(5557)] = 56037,
  [SMALL_STATE(5558)] = 56050,
  [SMALL_STATE(5559)] = 56063,
  [SMALL_STATE(5560)] = 56076,
  [SMALL_STATE(5561)] = 56085,
  [SMALL_STATE(5562)] = 56098,
  [SMALL_STATE(5563)] = 56111,
  [SMALL_STATE(5564)] = 56124,
  [SMALL_STATE(5565)] = 56137,
  [SMALL_STATE(5566)] = 56150,
  [SMALL_STATE(5567)] = 56163,
  [SMALL_STATE(5568)] = 56176,
  [SMALL_STATE(5569)] = 56189,
  [SMALL_STATE(5570)] = 56202,
  [SMALL_STATE(5571)] = 56215,
  [SMALL_STATE(5572)] = 56228,
  [SMALL_STATE(5573)] = 56241,
  [SMALL_STATE(5574)] = 56254,
  [SMALL_STATE(5575)] = 56267,
  [SMALL_STATE(5576)] = 56280,
  [SMALL_STATE(5577)] = 56291,
  [SMALL_STATE(5578)] = 56304,
  [SMALL_STATE(5579)] = 56317,
  [SMALL_STATE(5580)] = 56330,
  [SMALL_STATE(5581)] = 56341,
  [SMALL_STATE(5582)] = 56354,
  [SMALL_STATE(5583)] = 56367,
  [SMALL_STATE(5584)] = 56380,
  [SMALL_STATE(5585)] = 56393,
  [SMALL_STATE(5586)] = 56406,
  [SMALL_STATE(5587)] = 56419,
  [SMALL_STATE(5588)] = 56432,
  [SMALL_STATE(5589)] = 56445,
  [SMALL_STATE(5590)] = 56458,
  [SMALL_STATE(5591)] = 56471,
  [SMALL_STATE(5592)] = 56484,
  [SMALL_STATE(5593)] = 56497,
  [SMALL_STATE(5594)] = 56510,
  [SMALL_STATE(5595)] = 56523,
  [SMALL_STATE(5596)] = 56536,
  [SMALL_STATE(5597)] = 56549,
  [SMALL_STATE(5598)] = 56562,
  [SMALL_STATE(5599)] = 56575,
  [SMALL_STATE(5600)] = 56588,
  [SMALL_STATE(5601)] = 56599,
  [SMALL_STATE(5602)] = 56612,
  [SMALL_STATE(5603)] = 56625,
  [SMALL_STATE(5604)] = 56638,
  [SMALL_STATE(5605)] = 56651,
  [SMALL_STATE(5606)] = 56664,
  [SMALL_STATE(5607)] = 56677,
  [SMALL_STATE(5608)] = 56690,
  [SMALL_STATE(5609)] = 56703,
  [SMALL_STATE(5610)] = 56716,
  [SMALL_STATE(5611)] = 56727,
  [SMALL_STATE(5612)] = 56740,
  [SMALL_STATE(5613)] = 56753,
  [SMALL_STATE(5614)] = 56766,
  [SMALL_STATE(5615)] = 56779,
  [SMALL_STATE(5616)] = 56790,
  [SMALL_STATE(5617)] = 56803,
  [SMALL_STATE(5618)] = 56816,
  [SMALL_STATE(5619)] = 56829,
  [SMALL_STATE(5620)] = 56842,
  [SMALL_STATE(5621)] = 56855,
  [SMALL_STATE(5622)] = 56868,
  [SMALL_STATE(5623)] = 56881,
  [SMALL_STATE(5624)] = 56894,
  [SMALL_STATE(5625)] = 56907,
  [SMALL_STATE(5626)] = 56920,
  [SMALL_STATE(5627)] = 56933,
  [SMALL_STATE(5628)] = 56946,
  [SMALL_STATE(5629)] = 56959,
  [SMALL_STATE(5630)] = 56972,
  [SMALL_STATE(5631)] = 56985,
  [SMALL_STATE(5632)] = 56998,
  [SMALL_STATE(5633)] = 57011,
  [SMALL_STATE(5634)] = 57024,
  [SMALL_STATE(5635)] = 57037,
  [SMALL_STATE(5636)] = 57050,
  [SMALL_STATE(5637)] = 57063,
  [SMALL_STATE(5638)] = 57076,
  [SMALL_STATE(5639)] = 57089,
  [SMALL_STATE(5640)] = 57102,
  [SMALL_STATE(5641)] = 57115,
  [SMALL_STATE(5642)] = 57128,
  [SMALL_STATE(5643)] = 57141,
  [SMALL_STATE(5644)] = 57154,
  [SMALL_STATE(5645)] = 57167,
  [SMALL_STATE(5646)] = 57180,
  [SMALL_STATE(5647)] = 57193,
  [SMALL_STATE(5648)] = 57206,
  [SMALL_STATE(5649)] = 57219,
  [SMALL_STATE(5650)] = 57232,
  [SMALL_STATE(5651)] = 57245,
  [SMALL_STATE(5652)] = 57254,
  [SMALL_STATE(5653)] = 57267,
  [SMALL_STATE(5654)] = 57280,
  [SMALL_STATE(5655)] = 57293,
  [SMALL_STATE(5656)] = 57306,
  [SMALL_STATE(5657)] = 57315,
  [SMALL_STATE(5658)] = 57328,
  [SMALL_STATE(5659)] = 57341,
  [SMALL_STATE(5660)] = 57354,
  [SMALL_STATE(5661)] = 57367,
  [SMALL_STATE(5662)] = 57380,
  [SMALL_STATE(5663)] = 57393,
  [SMALL_STATE(5664)] = 57406,
  [SMALL_STATE(5665)] = 57419,
  [SMALL_STATE(5666)] = 57432,
  [SMALL_STATE(5667)] = 57445,
  [SMALL_STATE(5668)] = 57458,
  [SMALL_STATE(5669)] = 57471,
  [SMALL_STATE(5670)] = 57480,
  [SMALL_STATE(5671)] = 57493,
  [SMALL_STATE(5672)] = 57506,
  [SMALL_STATE(5673)] = 57519,
  [SMALL_STATE(5674)] = 57532,
  [SMALL_STATE(5675)] = 57545,
  [SMALL_STATE(5676)] = 57558,
  [SMALL_STATE(5677)] = 57571,
  [SMALL_STATE(5678)] = 57584,
  [SMALL_STATE(5679)] = 57597,
  [SMALL_STATE(5680)] = 57610,
  [SMALL_STATE(5681)] = 57623,
  [SMALL_STATE(5682)] = 57636,
  [SMALL_STATE(5683)] = 57647,
  [SMALL_STATE(5684)] = 57660,
  [SMALL_STATE(5685)] = 57669,
  [SMALL_STATE(5686)] = 57682,
  [SMALL_STATE(5687)] = 57695,
  [SMALL_STATE(5688)] = 57708,
  [SMALL_STATE(5689)] = 57721,
  [SMALL_STATE(5690)] = 57734,
  [SMALL_STATE(5691)] = 57747,
  [SMALL_STATE(5692)] = 57760,
  [SMALL_STATE(5693)] = 57773,
  [SMALL_STATE(5694)] = 57786,
  [SMALL_STATE(5695)] = 57799,
  [SMALL_STATE(5696)] = 57812,
  [SMALL_STATE(5697)] = 57825,
  [SMALL_STATE(5698)] = 57838,
  [SMALL_STATE(5699)] = 57851,
  [SMALL_STATE(5700)] = 57864,
  [SMALL_STATE(5701)] = 57877,
  [SMALL_STATE(5702)] = 57890,
  [SMALL_STATE(5703)] = 57903,
  [SMALL_STATE(5704)] = 57916,
  [SMALL_STATE(5705)] = 57929,
  [SMALL_STATE(5706)] = 57942,
  [SMALL_STATE(5707)] = 57955,
  [SMALL_STATE(5708)] = 57968,
  [SMALL_STATE(5709)] = 57981,
  [SMALL_STATE(5710)] = 57994,
  [SMALL_STATE(5711)] = 58007,
  [SMALL_STATE(5712)] = 58020,
  [SMALL_STATE(5713)] = 58030,
  [SMALL_STATE(5714)] = 58038,
  [SMALL_STATE(5715)] = 58048,
  [SMALL_STATE(5716)] = 58056,
  [SMALL_STATE(5717)] = 58066,
  [SMALL_STATE(5718)] = 58076,
  [SMALL_STATE(5719)] = 58086,
  [SMALL_STATE(5720)] = 58096,
  [SMALL_STATE(5721)] = 58106,
  [SMALL_STATE(5722)] = 58116,
  [SMALL_STATE(5723)] = 58126,
  [SMALL_STATE(5724)] = 58134,
  [SMALL_STATE(5725)] = 58144,
  [SMALL_STATE(5726)] = 58154,
  [SMALL_STATE(5727)] = 58164,
  [SMALL_STATE(5728)] = 58174,
  [SMALL_STATE(5729)] = 58184,
  [SMALL_STATE(5730)] = 58194,
  [SMALL_STATE(5731)] = 58204,
  [SMALL_STATE(5732)] = 58214,
  [SMALL_STATE(5733)] = 58224,
  [SMALL_STATE(5734)] = 58234,
  [SMALL_STATE(5735)] = 58244,
  [SMALL_STATE(5736)] = 58254,
  [SMALL_STATE(5737)] = 58264,
  [SMALL_STATE(5738)] = 58274,
  [SMALL_STATE(5739)] = 58284,
  [SMALL_STATE(5740)] = 58294,
  [SMALL_STATE(5741)] = 58304,
  [SMALL_STATE(5742)] = 58314,
  [SMALL_STATE(5743)] = 58324,
  [SMALL_STATE(5744)] = 58334,
  [SMALL_STATE(5745)] = 58342,
  [SMALL_STATE(5746)] = 58350,
  [SMALL_STATE(5747)] = 58360,
  [SMALL_STATE(5748)] = 58370,
  [SMALL_STATE(5749)] = 58380,
  [SMALL_STATE(5750)] = 58390,
  [SMALL_STATE(5751)] = 58398,
  [SMALL_STATE(5752)] = 58408,
  [SMALL_STATE(5753)] = 58418,
  [SMALL_STATE(5754)] = 58426,
  [SMALL_STATE(5755)] = 58436,
  [SMALL_STATE(5756)] = 58446,
  [SMALL_STATE(5757)] = 58454,
  [SMALL_STATE(5758)] = 58464,
  [SMALL_STATE(5759)] = 58474,
  [SMALL_STATE(5760)] = 58484,
  [SMALL_STATE(5761)] = 58494,
  [SMALL_STATE(5762)] = 58504,
  [SMALL_STATE(5763)] = 58512,
  [SMALL_STATE(5764)] = 58520,
  [SMALL_STATE(5765)] = 58530,
  [SMALL_STATE(5766)] = 58538,
  [SMALL_STATE(5767)] = 58548,
  [SMALL_STATE(5768)] = 58556,
  [SMALL_STATE(5769)] = 58566,
  [SMALL_STATE(5770)] = 58576,
  [SMALL_STATE(5771)] = 58586,
  [SMALL_STATE(5772)] = 58594,
  [SMALL_STATE(5773)] = 58604,
  [SMALL_STATE(5774)] = 58614,
  [SMALL_STATE(5775)] = 58622,
  [SMALL_STATE(5776)] = 58632,
  [SMALL_STATE(5777)] = 58642,
  [SMALL_STATE(5778)] = 58652,
  [SMALL_STATE(5779)] = 58662,
  [SMALL_STATE(5780)] = 58670,
  [SMALL_STATE(5781)] = 58678,
  [SMALL_STATE(5782)] = 58688,
  [SMALL_STATE(5783)] = 58698,
  [SMALL_STATE(5784)] = 58708,
  [SMALL_STATE(5785)] = 58716,
  [SMALL_STATE(5786)] = 58724,
  [SMALL_STATE(5787)] = 58734,
  [SMALL_STATE(5788)] = 58742,
  [SMALL_STATE(5789)] = 58752,
  [SMALL_STATE(5790)] = 58762,
  [SMALL_STATE(5791)] = 58772,
  [SMALL_STATE(5792)] = 58782,
  [SMALL_STATE(5793)] = 58792,
  [SMALL_STATE(5794)] = 58800,
  [SMALL_STATE(5795)] = 58808,
  [SMALL_STATE(5796)] = 58818,
  [SMALL_STATE(5797)] = 58826,
  [SMALL_STATE(5798)] = 58836,
  [SMALL_STATE(5799)] = 58846,
  [SMALL_STATE(5800)] = 58854,
  [SMALL_STATE(5801)] = 58862,
  [SMALL_STATE(5802)] = 58872,
  [SMALL_STATE(5803)] = 58880,
  [SMALL_STATE(5804)] = 58890,
  [SMALL_STATE(5805)] = 58900,
  [SMALL_STATE(5806)] = 58910,
  [SMALL_STATE(5807)] = 58920,
  [SMALL_STATE(5808)] = 58928,
  [SMALL_STATE(5809)] = 58938,
  [SMALL_STATE(5810)] = 58946,
  [SMALL_STATE(5811)] = 58956,
  [SMALL_STATE(5812)] = 58964,
  [SMALL_STATE(5813)] = 58974,
  [SMALL_STATE(5814)] = 58984,
  [SMALL_STATE(5815)] = 58994,
  [SMALL_STATE(5816)] = 59004,
  [SMALL_STATE(5817)] = 59014,
  [SMALL_STATE(5818)] = 59024,
  [SMALL_STATE(5819)] = 59034,
  [SMALL_STATE(5820)] = 59044,
  [SMALL_STATE(5821)] = 59054,
  [SMALL_STATE(5822)] = 59064,
  [SMALL_STATE(5823)] = 59074,
  [SMALL_STATE(5824)] = 59084,
  [SMALL_STATE(5825)] = 59092,
  [SMALL_STATE(5826)] = 59102,
  [SMALL_STATE(5827)] = 59110,
  [SMALL_STATE(5828)] = 59118,
  [SMALL_STATE(5829)] = 59128,
  [SMALL_STATE(5830)] = 59138,
  [SMALL_STATE(5831)] = 59148,
  [SMALL_STATE(5832)] = 59158,
  [SMALL_STATE(5833)] = 59168,
  [SMALL_STATE(5834)] = 59176,
  [SMALL_STATE(5835)] = 59186,
  [SMALL_STATE(5836)] = 59196,
  [SMALL_STATE(5837)] = 59206,
  [SMALL_STATE(5838)] = 59216,
  [SMALL_STATE(5839)] = 59226,
  [SMALL_STATE(5840)] = 59236,
  [SMALL_STATE(5841)] = 59246,
  [SMALL_STATE(5842)] = 59254,
  [SMALL_STATE(5843)] = 59264,
  [SMALL_STATE(5844)] = 59274,
  [SMALL_STATE(5845)] = 59284,
  [SMALL_STATE(5846)] = 59294,
  [SMALL_STATE(5847)] = 59304,
  [SMALL_STATE(5848)] = 59314,
  [SMALL_STATE(5849)] = 59322,
  [SMALL_STATE(5850)] = 59332,
  [SMALL_STATE(5851)] = 59342,
  [SMALL_STATE(5852)] = 59352,
  [SMALL_STATE(5853)] = 59360,
  [SMALL_STATE(5854)] = 59370,
  [SMALL_STATE(5855)] = 59380,
  [SMALL_STATE(5856)] = 59388,
  [SMALL_STATE(5857)] = 59398,
  [SMALL_STATE(5858)] = 59408,
  [SMALL_STATE(5859)] = 59418,
  [SMALL_STATE(5860)] = 59428,
  [SMALL_STATE(5861)] = 59438,
  [SMALL_STATE(5862)] = 59448,
  [SMALL_STATE(5863)] = 59456,
  [SMALL_STATE(5864)] = 59466,
  [SMALL_STATE(5865)] = 59476,
  [SMALL_STATE(5866)] = 59486,
  [SMALL_STATE(5867)] = 59496,
  [SMALL_STATE(5868)] = 59504,
  [SMALL_STATE(5869)] = 59514,
  [SMALL_STATE(5870)] = 59524,
  [SMALL_STATE(5871)] = 59534,
  [SMALL_STATE(5872)] = 59544,
  [SMALL_STATE(5873)] = 59554,
  [SMALL_STATE(5874)] = 59564,
  [SMALL_STATE(5875)] = 59574,
  [SMALL_STATE(5876)] = 59584,
  [SMALL_STATE(5877)] = 59592,
  [SMALL_STATE(5878)] = 59602,
  [SMALL_STATE(5879)] = 59612,
  [SMALL_STATE(5880)] = 59622,
  [SMALL_STATE(5881)] = 59632,
  [SMALL_STATE(5882)] = 59640,
  [SMALL_STATE(5883)] = 59650,
  [SMALL_STATE(5884)] = 59660,
  [SMALL_STATE(5885)] = 59670,
  [SMALL_STATE(5886)] = 59680,
  [SMALL_STATE(5887)] = 59690,
  [SMALL_STATE(5888)] = 59698,
  [SMALL_STATE(5889)] = 59706,
  [SMALL_STATE(5890)] = 59716,
  [SMALL_STATE(5891)] = 59724,
  [SMALL_STATE(5892)] = 59734,
  [SMALL_STATE(5893)] = 59742,
  [SMALL_STATE(5894)] = 59752,
  [SMALL_STATE(5895)] = 59762,
  [SMALL_STATE(5896)] = 59772,
  [SMALL_STATE(5897)] = 59782,
  [SMALL_STATE(5898)] = 59790,
  [SMALL_STATE(5899)] = 59798,
  [SMALL_STATE(5900)] = 59808,
  [SMALL_STATE(5901)] = 59816,
  [SMALL_STATE(5902)] = 59823,
  [SMALL_STATE(5903)] = 59830,
  [SMALL_STATE(5904)] = 59837,
  [SMALL_STATE(5905)] = 59844,
  [SMALL_STATE(5906)] = 59851,
  [SMALL_STATE(5907)] = 59858,
  [SMALL_STATE(5908)] = 59865,
  [SMALL_STATE(5909)] = 59872,
  [SMALL_STATE(5910)] = 59879,
  [SMALL_STATE(5911)] = 59886,
  [SMALL_STATE(5912)] = 59893,
  [SMALL_STATE(5913)] = 59900,
  [SMALL_STATE(5914)] = 59907,
  [SMALL_STATE(5915)] = 59914,
  [SMALL_STATE(5916)] = 59921,
  [SMALL_STATE(5917)] = 59928,
  [SMALL_STATE(5918)] = 59935,
  [SMALL_STATE(5919)] = 59942,
  [SMALL_STATE(5920)] = 59949,
  [SMALL_STATE(5921)] = 59956,
  [SMALL_STATE(5922)] = 59963,
  [SMALL_STATE(5923)] = 59970,
  [SMALL_STATE(5924)] = 59977,
  [SMALL_STATE(5925)] = 59984,
  [SMALL_STATE(5926)] = 59991,
  [SMALL_STATE(5927)] = 59998,
  [SMALL_STATE(5928)] = 60005,
  [SMALL_STATE(5929)] = 60012,
  [SMALL_STATE(5930)] = 60019,
  [SMALL_STATE(5931)] = 60026,
  [SMALL_STATE(5932)] = 60033,
  [SMALL_STATE(5933)] = 60040,
  [SMALL_STATE(5934)] = 60047,
  [SMALL_STATE(5935)] = 60054,
  [SMALL_STATE(5936)] = 60061,
  [SMALL_STATE(5937)] = 60068,
  [SMALL_STATE(5938)] = 60075,
  [SMALL_STATE(5939)] = 60082,
  [SMALL_STATE(5940)] = 60089,
  [SMALL_STATE(5941)] = 60096,
  [SMALL_STATE(5942)] = 60103,
  [SMALL_STATE(5943)] = 60110,
  [SMALL_STATE(5944)] = 60117,
  [SMALL_STATE(5945)] = 60124,
  [SMALL_STATE(5946)] = 60131,
  [SMALL_STATE(5947)] = 60138,
  [SMALL_STATE(5948)] = 60145,
  [SMALL_STATE(5949)] = 60152,
  [SMALL_STATE(5950)] = 60159,
  [SMALL_STATE(5951)] = 60166,
  [SMALL_STATE(5952)] = 60173,
  [SMALL_STATE(5953)] = 60180,
  [SMALL_STATE(5954)] = 60187,
  [SMALL_STATE(5955)] = 60194,
  [SMALL_STATE(5956)] = 60201,
  [SMALL_STATE(5957)] = 60208,
  [SMALL_STATE(5958)] = 60215,
  [SMALL_STATE(5959)] = 60222,
  [SMALL_STATE(5960)] = 60229,
  [SMALL_STATE(5961)] = 60236,
  [SMALL_STATE(5962)] = 60243,
  [SMALL_STATE(5963)] = 60250,
  [SMALL_STATE(5964)] = 60257,
  [SMALL_STATE(5965)] = 60264,
  [SMALL_STATE(5966)] = 60271,
  [SMALL_STATE(5967)] = 60278,
  [SMALL_STATE(5968)] = 60285,
  [SMALL_STATE(5969)] = 60292,
  [SMALL_STATE(5970)] = 60299,
  [SMALL_STATE(5971)] = 60306,
  [SMALL_STATE(5972)] = 60313,
  [SMALL_STATE(5973)] = 60320,
  [SMALL_STATE(5974)] = 60327,
  [SMALL_STATE(5975)] = 60334,
  [SMALL_STATE(5976)] = 60341,
  [SMALL_STATE(5977)] = 60348,
  [SMALL_STATE(5978)] = 60355,
  [SMALL_STATE(5979)] = 60362,
  [SMALL_STATE(5980)] = 60369,
  [SMALL_STATE(5981)] = 60376,
  [SMALL_STATE(5982)] = 60383,
  [SMALL_STATE(5983)] = 60390,
  [SMALL_STATE(5984)] = 60397,
  [SMALL_STATE(5985)] = 60404,
  [SMALL_STATE(5986)] = 60411,
  [SMALL_STATE(5987)] = 60418,
  [SMALL_STATE(5988)] = 60425,
  [SMALL_STATE(5989)] = 60432,
  [SMALL_STATE(5990)] = 60439,
  [SMALL_STATE(5991)] = 60446,
  [SMALL_STATE(5992)] = 60453,
  [SMALL_STATE(5993)] = 60460,
  [SMALL_STATE(5994)] = 60467,
  [SMALL_STATE(5995)] = 60474,
  [SMALL_STATE(5996)] = 60481,
  [SMALL_STATE(5997)] = 60488,
  [SMALL_STATE(5998)] = 60495,
  [SMALL_STATE(5999)] = 60502,
  [SMALL_STATE(6000)] = 60509,
  [SMALL_STATE(6001)] = 60516,
  [SMALL_STATE(6002)] = 60523,
  [SMALL_STATE(6003)] = 60530,
  [SMALL_STATE(6004)] = 60537,
  [SMALL_STATE(6005)] = 60544,
  [SMALL_STATE(6006)] = 60551,
  [SMALL_STATE(6007)] = 60558,
  [SMALL_STATE(6008)] = 60565,
  [SMALL_STATE(6009)] = 60572,
  [SMALL_STATE(6010)] = 60579,
  [SMALL_STATE(6011)] = 60586,
  [SMALL_STATE(6012)] = 60593,
  [SMALL_STATE(6013)] = 60600,
  [SMALL_STATE(6014)] = 60607,
  [SMALL_STATE(6015)] = 60614,
  [SMALL_STATE(6016)] = 60621,
  [SMALL_STATE(6017)] = 60628,
  [SMALL_STATE(6018)] = 60635,
  [SMALL_STATE(6019)] = 60642,
  [SMALL_STATE(6020)] = 60649,
  [SMALL_STATE(6021)] = 60656,
  [SMALL_STATE(6022)] = 60663,
  [SMALL_STATE(6023)] = 60670,
  [SMALL_STATE(6024)] = 60677,
  [SMALL_STATE(6025)] = 60684,
  [SMALL_STATE(6026)] = 60691,
  [SMALL_STATE(6027)] = 60698,
  [SMALL_STATE(6028)] = 60705,
  [SMALL_STATE(6029)] = 60712,
  [SMALL_STATE(6030)] = 60719,
  [SMALL_STATE(6031)] = 60726,
  [SMALL_STATE(6032)] = 60733,
  [SMALL_STATE(6033)] = 60740,
  [SMALL_STATE(6034)] = 60747,
  [SMALL_STATE(6035)] = 60754,
  [SMALL_STATE(6036)] = 60761,
  [SMALL_STATE(6037)] = 60768,
  [SMALL_STATE(6038)] = 60775,
  [SMALL_STATE(6039)] = 60782,
  [SMALL_STATE(6040)] = 60789,
  [SMALL_STATE(6041)] = 60796,
  [SMALL_STATE(6042)] = 60803,
  [SMALL_STATE(6043)] = 60810,
  [SMALL_STATE(6044)] = 60817,
  [SMALL_STATE(6045)] = 60824,
  [SMALL_STATE(6046)] = 60831,
  [SMALL_STATE(6047)] = 60838,
  [SMALL_STATE(6048)] = 60845,
  [SMALL_STATE(6049)] = 60852,
  [SMALL_STATE(6050)] = 60859,
  [SMALL_STATE(6051)] = 60866,
  [SMALL_STATE(6052)] = 60873,
  [SMALL_STATE(6053)] = 60880,
  [SMALL_STATE(6054)] = 60887,
  [SMALL_STATE(6055)] = 60894,
  [SMALL_STATE(6056)] = 60901,
  [SMALL_STATE(6057)] = 60908,
  [SMALL_STATE(6058)] = 60915,
  [SMALL_STATE(6059)] = 60922,
  [SMALL_STATE(6060)] = 60929,
  [SMALL_STATE(6061)] = 60936,
  [SMALL_STATE(6062)] = 60943,
  [SMALL_STATE(6063)] = 60950,
  [SMALL_STATE(6064)] = 60957,
  [SMALL_STATE(6065)] = 60964,
  [SMALL_STATE(6066)] = 60971,
  [SMALL_STATE(6067)] = 60978,
  [SMALL_STATE(6068)] = 60985,
  [SMALL_STATE(6069)] = 60992,
  [SMALL_STATE(6070)] = 60999,
  [SMALL_STATE(6071)] = 61006,
  [SMALL_STATE(6072)] = 61013,
  [SMALL_STATE(6073)] = 61020,
  [SMALL_STATE(6074)] = 61027,
  [SMALL_STATE(6075)] = 61034,
  [SMALL_STATE(6076)] = 61041,
  [SMALL_STATE(6077)] = 61048,
  [SMALL_STATE(6078)] = 61055,
  [SMALL_STATE(6079)] = 61062,
  [SMALL_STATE(6080)] = 61069,
  [SMALL_STATE(6081)] = 61076,
  [SMALL_STATE(6082)] = 61083,
  [SMALL_STATE(6083)] = 61090,
  [SMALL_STATE(6084)] = 61097,
  [SMALL_STATE(6085)] = 61104,
  [SMALL_STATE(6086)] = 61111,
  [SMALL_STATE(6087)] = 61118,
  [SMALL_STATE(6088)] = 61125,
  [SMALL_STATE(6089)] = 61132,
  [SMALL_STATE(6090)] = 61139,
  [SMALL_STATE(6091)] = 61146,
  [SMALL_STATE(6092)] = 61153,
  [SMALL_STATE(6093)] = 61160,
  [SMALL_STATE(6094)] = 61167,
  [SMALL_STATE(6095)] = 61174,
  [SMALL_STATE(6096)] = 61181,
  [SMALL_STATE(6097)] = 61188,
  [SMALL_STATE(6098)] = 61195,
  [SMALL_STATE(6099)] = 61202,
  [SMALL_STATE(6100)] = 61209,
  [SMALL_STATE(6101)] = 61216,
  [SMALL_STATE(6102)] = 61223,
  [SMALL_STATE(6103)] = 61230,
  [SMALL_STATE(6104)] = 61237,
  [SMALL_STATE(6105)] = 61244,
  [SMALL_STATE(6106)] = 61251,
  [SMALL_STATE(6107)] = 61258,
  [SMALL_STATE(6108)] = 61265,
  [SMALL_STATE(6109)] = 61272,
  [SMALL_STATE(6110)] = 61279,
  [SMALL_STATE(6111)] = 61286,
  [SMALL_STATE(6112)] = 61293,
  [SMALL_STATE(6113)] = 61300,
  [SMALL_STATE(6114)] = 61307,
  [SMALL_STATE(6115)] = 61314,
  [SMALL_STATE(6116)] = 61321,
  [SMALL_STATE(6117)] = 61328,
  [SMALL_STATE(6118)] = 61335,
  [SMALL_STATE(6119)] = 61342,
  [SMALL_STATE(6120)] = 61349,
  [SMALL_STATE(6121)] = 61356,
  [SMALL_STATE(6122)] = 61363,
  [SMALL_STATE(6123)] = 61370,
  [SMALL_STATE(6124)] = 61377,
  [SMALL_STATE(6125)] = 61384,
  [SMALL_STATE(6126)] = 61391,
  [SMALL_STATE(6127)] = 61398,
  [SMALL_STATE(6128)] = 61405,
  [SMALL_STATE(6129)] = 61412,
  [SMALL_STATE(6130)] = 61419,
  [SMALL_STATE(6131)] = 61426,
  [SMALL_STATE(6132)] = 61433,
  [SMALL_STATE(6133)] = 61440,
  [SMALL_STATE(6134)] = 61447,
  [SMALL_STATE(6135)] = 61454,
  [SMALL_STATE(6136)] = 61461,
  [SMALL_STATE(6137)] = 61468,
  [SMALL_STATE(6138)] = 61475,
  [SMALL_STATE(6139)] = 61482,
  [SMALL_STATE(6140)] = 61489,
  [SMALL_STATE(6141)] = 61496,
  [SMALL_STATE(6142)] = 61503,
  [SMALL_STATE(6143)] = 61510,
  [SMALL_STATE(6144)] = 61517,
  [SMALL_STATE(6145)] = 61524,
  [SMALL_STATE(6146)] = 61531,
  [SMALL_STATE(6147)] = 61538,
  [SMALL_STATE(6148)] = 61545,
  [SMALL_STATE(6149)] = 61552,
  [SMALL_STATE(6150)] = 61559,
  [SMALL_STATE(6151)] = 61566,
  [SMALL_STATE(6152)] = 61573,
  [SMALL_STATE(6153)] = 61580,
  [SMALL_STATE(6154)] = 61587,
  [SMALL_STATE(6155)] = 61594,
  [SMALL_STATE(6156)] = 61601,
  [SMALL_STATE(6157)] = 61608,
  [SMALL_STATE(6158)] = 61615,
  [SMALL_STATE(6159)] = 61622,
  [SMALL_STATE(6160)] = 61629,
  [SMALL_STATE(6161)] = 61636,
  [SMALL_STATE(6162)] = 61643,
  [SMALL_STATE(6163)] = 61650,
  [SMALL_STATE(6164)] = 61657,
  [SMALL_STATE(6165)] = 61664,
  [SMALL_STATE(6166)] = 61671,
  [SMALL_STATE(6167)] = 61678,
  [SMALL_STATE(6168)] = 61685,
  [SMALL_STATE(6169)] = 61692,
  [SMALL_STATE(6170)] = 61699,
  [SMALL_STATE(6171)] = 61706,
  [SMALL_STATE(6172)] = 61713,
  [SMALL_STATE(6173)] = 61720,
  [SMALL_STATE(6174)] = 61727,
  [SMALL_STATE(6175)] = 61734,
  [SMALL_STATE(6176)] = 61741,
  [SMALL_STATE(6177)] = 61748,
  [SMALL_STATE(6178)] = 61755,
  [SMALL_STATE(6179)] = 61762,
  [SMALL_STATE(6180)] = 61769,
  [SMALL_STATE(6181)] = 61776,
  [SMALL_STATE(6182)] = 61783,
  [SMALL_STATE(6183)] = 61790,
  [SMALL_STATE(6184)] = 61797,
  [SMALL_STATE(6185)] = 61804,
  [SMALL_STATE(6186)] = 61811,
  [SMALL_STATE(6187)] = 61818,
  [SMALL_STATE(6188)] = 61825,
  [SMALL_STATE(6189)] = 61832,
  [SMALL_STATE(6190)] = 61839,
  [SMALL_STATE(6191)] = 61846,
  [SMALL_STATE(6192)] = 61853,
  [SMALL_STATE(6193)] = 61860,
  [SMALL_STATE(6194)] = 61867,
  [SMALL_STATE(6195)] = 61874,
  [SMALL_STATE(6196)] = 61881,
  [SMALL_STATE(6197)] = 61888,
  [SMALL_STATE(6198)] = 61895,
  [SMALL_STATE(6199)] = 61902,
  [SMALL_STATE(6200)] = 61909,
  [SMALL_STATE(6201)] = 61916,
  [SMALL_STATE(6202)] = 61923,
  [SMALL_STATE(6203)] = 61930,
  [SMALL_STATE(6204)] = 61937,
  [SMALL_STATE(6205)] = 61944,
  [SMALL_STATE(6206)] = 61951,
  [SMALL_STATE(6207)] = 61958,
  [SMALL_STATE(6208)] = 61965,
  [SMALL_STATE(6209)] = 61972,
  [SMALL_STATE(6210)] = 61979,
  [SMALL_STATE(6211)] = 61986,
  [SMALL_STATE(6212)] = 61993,
  [SMALL_STATE(6213)] = 62000,
  [SMALL_STATE(6214)] = 62007,
  [SMALL_STATE(6215)] = 62014,
  [SMALL_STATE(6216)] = 62021,
  [SMALL_STATE(6217)] = 62028,
  [SMALL_STATE(6218)] = 62035,
  [SMALL_STATE(6219)] = 62042,
  [SMALL_STATE(6220)] = 62049,
  [SMALL_STATE(6221)] = 62056,
  [SMALL_STATE(6222)] = 62063,
  [SMALL_STATE(6223)] = 62070,
  [SMALL_STATE(6224)] = 62077,
  [SMALL_STATE(6225)] = 62084,
  [SMALL_STATE(6226)] = 62091,
  [SMALL_STATE(6227)] = 62098,
  [SMALL_STATE(6228)] = 62105,
  [SMALL_STATE(6229)] = 62112,
  [SMALL_STATE(6230)] = 62119,
  [SMALL_STATE(6231)] = 62126,
  [SMALL_STATE(6232)] = 62133,
  [SMALL_STATE(6233)] = 62140,
  [SMALL_STATE(6234)] = 62147,
  [SMALL_STATE(6235)] = 62154,
  [SMALL_STATE(6236)] = 62161,
  [SMALL_STATE(6237)] = 62168,
  [SMALL_STATE(6238)] = 62175,
  [SMALL_STATE(6239)] = 62182,
  [SMALL_STATE(6240)] = 62189,
  [SMALL_STATE(6241)] = 62196,
  [SMALL_STATE(6242)] = 62203,
  [SMALL_STATE(6243)] = 62210,
  [SMALL_STATE(6244)] = 62217,
  [SMALL_STATE(6245)] = 62224,
  [SMALL_STATE(6246)] = 62231,
  [SMALL_STATE(6247)] = 62238,
  [SMALL_STATE(6248)] = 62245,
  [SMALL_STATE(6249)] = 62252,
  [SMALL_STATE(6250)] = 62259,
  [SMALL_STATE(6251)] = 62266,
  [SMALL_STATE(6252)] = 62273,
  [SMALL_STATE(6253)] = 62280,
  [SMALL_STATE(6254)] = 62287,
  [SMALL_STATE(6255)] = 62294,
  [SMALL_STATE(6256)] = 62301,
  [SMALL_STATE(6257)] = 62308,
  [SMALL_STATE(6258)] = 62315,
  [SMALL_STATE(6259)] = 62322,
  [SMALL_STATE(6260)] = 62329,
  [SMALL_STATE(6261)] = 62336,
  [SMALL_STATE(6262)] = 62343,
  [SMALL_STATE(6263)] = 62350,
  [SMALL_STATE(6264)] = 62357,
  [SMALL_STATE(6265)] = 62364,
  [SMALL_STATE(6266)] = 62371,
  [SMALL_STATE(6267)] = 62378,
  [SMALL_STATE(6268)] = 62385,
  [SMALL_STATE(6269)] = 62392,
  [SMALL_STATE(6270)] = 62399,
  [SMALL_STATE(6271)] = 62406,
  [SMALL_STATE(6272)] = 62413,
  [SMALL_STATE(6273)] = 62420,
  [SMALL_STATE(6274)] = 62427,
  [SMALL_STATE(6275)] = 62434,
  [SMALL_STATE(6276)] = 62441,
  [SMALL_STATE(6277)] = 62448,
  [SMALL_STATE(6278)] = 62455,
  [SMALL_STATE(6279)] = 62462,
  [SMALL_STATE(6280)] = 62469,
  [SMALL_STATE(6281)] = 62476,
  [SMALL_STATE(6282)] = 62483,
  [SMALL_STATE(6283)] = 62490,
  [SMALL_STATE(6284)] = 62497,
  [SMALL_STATE(6285)] = 62504,
  [SMALL_STATE(6286)] = 62511,
  [SMALL_STATE(6287)] = 62518,
  [SMALL_STATE(6288)] = 62525,
  [SMALL_STATE(6289)] = 62532,
  [SMALL_STATE(6290)] = 62539,
  [SMALL_STATE(6291)] = 62546,
  [SMALL_STATE(6292)] = 62553,
  [SMALL_STATE(6293)] = 62560,
  [SMALL_STATE(6294)] = 62567,
  [SMALL_STATE(6295)] = 62574,
  [SMALL_STATE(6296)] = 62581,
  [SMALL_STATE(6297)] = 62588,
  [SMALL_STATE(6298)] = 62595,
  [SMALL_STATE(6299)] = 62602,
  [SMALL_STATE(6300)] = 62609,
  [SMALL_STATE(6301)] = 62616,
  [SMALL_STATE(6302)] = 62623,
  [SMALL_STATE(6303)] = 62630,
  [SMALL_STATE(6304)] = 62637,
  [SMALL_STATE(6305)] = 62644,
  [SMALL_STATE(6306)] = 62651,
  [SMALL_STATE(6307)] = 62658,
  [SMALL_STATE(6308)] = 62665,
  [SMALL_STATE(6309)] = 62672,
  [SMALL_STATE(6310)] = 62679,
  [SMALL_STATE(6311)] = 62686,
  [SMALL_STATE(6312)] = 62693,
  [SMALL_STATE(6313)] = 62700,
  [SMALL_STATE(6314)] = 62707,
  [SMALL_STATE(6315)] = 62714,
  [SMALL_STATE(6316)] = 62721,
  [SMALL_STATE(6317)] = 62728,
  [SMALL_STATE(6318)] = 62735,
  [SMALL_STATE(6319)] = 62742,
  [SMALL_STATE(6320)] = 62749,
  [SMALL_STATE(6321)] = 62756,
  [SMALL_STATE(6322)] = 62763,
  [SMALL_STATE(6323)] = 62770,
  [SMALL_STATE(6324)] = 62777,
  [SMALL_STATE(6325)] = 62784,
  [SMALL_STATE(6326)] = 62791,
  [SMALL_STATE(6327)] = 62798,
  [SMALL_STATE(6328)] = 62805,
  [SMALL_STATE(6329)] = 62812,
  [SMALL_STATE(6330)] = 62819,
  [SMALL_STATE(6331)] = 62826,
  [SMALL_STATE(6332)] = 62833,
  [SMALL_STATE(6333)] = 62840,
  [SMALL_STATE(6334)] = 62847,
  [SMALL_STATE(6335)] = 62854,
  [SMALL_STATE(6336)] = 62861,
  [SMALL_STATE(6337)] = 62868,
  [SMALL_STATE(6338)] = 62875,
  [SMALL_STATE(6339)] = 62882,
  [SMALL_STATE(6340)] = 62889,
  [SMALL_STATE(6341)] = 62896,
  [SMALL_STATE(6342)] = 62903,
  [SMALL_STATE(6343)] = 62910,
  [SMALL_STATE(6344)] = 62917,
  [SMALL_STATE(6345)] = 62924,
  [SMALL_STATE(6346)] = 62931,
  [SMALL_STATE(6347)] = 62938,
  [SMALL_STATE(6348)] = 62945,
  [SMALL_STATE(6349)] = 62952,
  [SMALL_STATE(6350)] = 62959,
  [SMALL_STATE(6351)] = 62966,
  [SMALL_STATE(6352)] = 62973,
  [SMALL_STATE(6353)] = 62980,
  [SMALL_STATE(6354)] = 62987,
  [SMALL_STATE(6355)] = 62994,
  [SMALL_STATE(6356)] = 63001,
  [SMALL_STATE(6357)] = 63008,
  [SMALL_STATE(6358)] = 63015,
  [SMALL_STATE(6359)] = 63022,
  [SMALL_STATE(6360)] = 63029,
  [SMALL_STATE(6361)] = 63036,
  [SMALL_STATE(6362)] = 63043,
  [SMALL_STATE(6363)] = 63050,
  [SMALL_STATE(6364)] = 63057,
  [SMALL_STATE(6365)] = 63064,
  [SMALL_STATE(6366)] = 63071,
  [SMALL_STATE(6367)] = 63078,
  [SMALL_STATE(6368)] = 63085,
  [SMALL_STATE(6369)] = 63092,
  [SMALL_STATE(6370)] = 63099,
  [SMALL_STATE(6371)] = 63106,
  [SMALL_STATE(6372)] = 63113,
  [SMALL_STATE(6373)] = 63120,
  [SMALL_STATE(6374)] = 63127,
  [SMALL_STATE(6375)] = 63134,
  [SMALL_STATE(6376)] = 63141,
  [SMALL_STATE(6377)] = 63148,
  [SMALL_STATE(6378)] = 63155,
  [SMALL_STATE(6379)] = 63162,
  [SMALL_STATE(6380)] = 63169,
  [SMALL_STATE(6381)] = 63176,
  [SMALL_STATE(6382)] = 63183,
  [SMALL_STATE(6383)] = 63190,
  [SMALL_STATE(6384)] = 63197,
  [SMALL_STATE(6385)] = 63204,
  [SMALL_STATE(6386)] = 63211,
  [SMALL_STATE(6387)] = 63218,
  [SMALL_STATE(6388)] = 63225,
  [SMALL_STATE(6389)] = 63232,
  [SMALL_STATE(6390)] = 63239,
  [SMALL_STATE(6391)] = 63246,
  [SMALL_STATE(6392)] = 63253,
  [SMALL_STATE(6393)] = 63260,
  [SMALL_STATE(6394)] = 63267,
  [SMALL_STATE(6395)] = 63274,
  [SMALL_STATE(6396)] = 63281,
  [SMALL_STATE(6397)] = 63288,
  [SMALL_STATE(6398)] = 63295,
  [SMALL_STATE(6399)] = 63302,
  [SMALL_STATE(6400)] = 63309,
  [SMALL_STATE(6401)] = 63316,
  [SMALL_STATE(6402)] = 63323,
  [SMALL_STATE(6403)] = 63330,
  [SMALL_STATE(6404)] = 63337,
  [SMALL_STATE(6405)] = 63344,
  [SMALL_STATE(6406)] = 63351,
  [SMALL_STATE(6407)] = 63358,
  [SMALL_STATE(6408)] = 63365,
  [SMALL_STATE(6409)] = 63372,
  [SMALL_STATE(6410)] = 63379,
  [SMALL_STATE(6411)] = 63386,
  [SMALL_STATE(6412)] = 63393,
  [SMALL_STATE(6413)] = 63400,
  [SMALL_STATE(6414)] = 63407,
  [SMALL_STATE(6415)] = 63414,
  [SMALL_STATE(6416)] = 63421,
  [SMALL_STATE(6417)] = 63428,
  [SMALL_STATE(6418)] = 63435,
  [SMALL_STATE(6419)] = 63442,
  [SMALL_STATE(6420)] = 63449,
  [SMALL_STATE(6421)] = 63456,
  [SMALL_STATE(6422)] = 63463,
  [SMALL_STATE(6423)] = 63470,
  [SMALL_STATE(6424)] = 63477,
  [SMALL_STATE(6425)] = 63484,
  [SMALL_STATE(6426)] = 63491,
  [SMALL_STATE(6427)] = 63498,
  [SMALL_STATE(6428)] = 63505,
  [SMALL_STATE(6429)] = 63512,
  [SMALL_STATE(6430)] = 63519,
  [SMALL_STATE(6431)] = 63526,
  [SMALL_STATE(6432)] = 63533,
  [SMALL_STATE(6433)] = 63540,
  [SMALL_STATE(6434)] = 63547,
  [SMALL_STATE(6435)] = 63554,
  [SMALL_STATE(6436)] = 63561,
  [SMALL_STATE(6437)] = 63568,
  [SMALL_STATE(6438)] = 63575,
  [SMALL_STATE(6439)] = 63582,
  [SMALL_STATE(6440)] = 63589,
  [SMALL_STATE(6441)] = 63596,
  [SMALL_STATE(6442)] = 63603,
  [SMALL_STATE(6443)] = 63610,
  [SMALL_STATE(6444)] = 63617,
  [SMALL_STATE(6445)] = 63624,
  [SMALL_STATE(6446)] = 63631,
  [SMALL_STATE(6447)] = 63638,
  [SMALL_STATE(6448)] = 63645,
  [SMALL_STATE(6449)] = 63652,
  [SMALL_STATE(6450)] = 63659,
  [SMALL_STATE(6451)] = 63666,
  [SMALL_STATE(6452)] = 63673,
  [SMALL_STATE(6453)] = 63680,
  [SMALL_STATE(6454)] = 63687,
  [SMALL_STATE(6455)] = 63694,
  [SMALL_STATE(6456)] = 63701,
  [SMALL_STATE(6457)] = 63708,
  [SMALL_STATE(6458)] = 63715,
  [SMALL_STATE(6459)] = 63722,
  [SMALL_STATE(6460)] = 63729,
  [SMALL_STATE(6461)] = 63736,
  [SMALL_STATE(6462)] = 63743,
  [SMALL_STATE(6463)] = 63750,
  [SMALL_STATE(6464)] = 63757,
  [SMALL_STATE(6465)] = 63764,
  [SMALL_STATE(6466)] = 63771,
  [SMALL_STATE(6467)] = 63778,
  [SMALL_STATE(6468)] = 63785,
  [SMALL_STATE(6469)] = 63792,
  [SMALL_STATE(6470)] = 63799,
  [SMALL_STATE(6471)] = 63806,
  [SMALL_STATE(6472)] = 63813,
  [SMALL_STATE(6473)] = 63820,
  [SMALL_STATE(6474)] = 63827,
  [SMALL_STATE(6475)] = 63834,
  [SMALL_STATE(6476)] = 63841,
  [SMALL_STATE(6477)] = 63848,
  [SMALL_STATE(6478)] = 63855,
  [SMALL_STATE(6479)] = 63862,
  [SMALL_STATE(6480)] = 63869,
  [SMALL_STATE(6481)] = 63876,
  [SMALL_STATE(6482)] = 63883,
  [SMALL_STATE(6483)] = 63890,
  [SMALL_STATE(6484)] = 63897,
  [SMALL_STATE(6485)] = 63904,
  [SMALL_STATE(6486)] = 63911,
  [SMALL_STATE(6487)] = 63918,
  [SMALL_STATE(6488)] = 63925,
  [SMALL_STATE(6489)] = 63932,
  [SMALL_STATE(6490)] = 63939,
  [SMALL_STATE(6491)] = 63946,
  [SMALL_STATE(6492)] = 63953,
  [SMALL_STATE(6493)] = 63960,
  [SMALL_STATE(6494)] = 63967,
  [SMALL_STATE(6495)] = 63974,
  [SMALL_STATE(6496)] = 63981,
  [SMALL_STATE(6497)] = 63988,
  [SMALL_STATE(6498)] = 63995,
  [SMALL_STATE(6499)] = 64002,
  [SMALL_STATE(6500)] = 64009,
  [SMALL_STATE(6501)] = 64016,
  [SMALL_STATE(6502)] = 64023,
  [SMALL_STATE(6503)] = 64030,
  [SMALL_STATE(6504)] = 64037,
  [SMALL_STATE(6505)] = 64044,
  [SMALL_STATE(6506)] = 64051,
  [SMALL_STATE(6507)] = 64058,
  [SMALL_STATE(6508)] = 64065,
  [SMALL_STATE(6509)] = 64072,
  [SMALL_STATE(6510)] = 64079,
  [SMALL_STATE(6511)] = 64086,
  [SMALL_STATE(6512)] = 64093,
  [SMALL_STATE(6513)] = 64100,
  [SMALL_STATE(6514)] = 64107,
  [SMALL_STATE(6515)] = 64114,
  [SMALL_STATE(6516)] = 64121,
  [SMALL_STATE(6517)] = 64128,
  [SMALL_STATE(6518)] = 64135,
  [SMALL_STATE(6519)] = 64142,
  [SMALL_STATE(6520)] = 64149,
  [SMALL_STATE(6521)] = 64156,
  [SMALL_STATE(6522)] = 64163,
  [SMALL_STATE(6523)] = 64170,
  [SMALL_STATE(6524)] = 64177,
  [SMALL_STATE(6525)] = 64184,
  [SMALL_STATE(6526)] = 64191,
  [SMALL_STATE(6527)] = 64198,
  [SMALL_STATE(6528)] = 64205,
  [SMALL_STATE(6529)] = 64212,
  [SMALL_STATE(6530)] = 64219,
  [SMALL_STATE(6531)] = 64226,
  [SMALL_STATE(6532)] = 64233,
  [SMALL_STATE(6533)] = 64240,
  [SMALL_STATE(6534)] = 64247,
  [SMALL_STATE(6535)] = 64254,
  [SMALL_STATE(6536)] = 64261,
  [SMALL_STATE(6537)] = 64268,
  [SMALL_STATE(6538)] = 64275,
  [SMALL_STATE(6539)] = 64282,
  [SMALL_STATE(6540)] = 64289,
  [SMALL_STATE(6541)] = 64296,
  [SMALL_STATE(6542)] = 64303,
  [SMALL_STATE(6543)] = 64310,
  [SMALL_STATE(6544)] = 64317,
  [SMALL_STATE(6545)] = 64324,
  [SMALL_STATE(6546)] = 64331,
  [SMALL_STATE(6547)] = 64338,
  [SMALL_STATE(6548)] = 64345,
  [SMALL_STATE(6549)] = 64352,
  [SMALL_STATE(6550)] = 64359,
  [SMALL_STATE(6551)] = 64366,
  [SMALL_STATE(6552)] = 64373,
  [SMALL_STATE(6553)] = 64380,
  [SMALL_STATE(6554)] = 64387,
  [SMALL_STATE(6555)] = 64394,
  [SMALL_STATE(6556)] = 64401,
  [SMALL_STATE(6557)] = 64408,
  [SMALL_STATE(6558)] = 64415,
  [SMALL_STATE(6559)] = 64422,
  [SMALL_STATE(6560)] = 64429,
  [SMALL_STATE(6561)] = 64436,
  [SMALL_STATE(6562)] = 64443,
  [SMALL_STATE(6563)] = 64450,
  [SMALL_STATE(6564)] = 64457,
  [SMALL_STATE(6565)] = 64464,
  [SMALL_STATE(6566)] = 64471,
  [SMALL_STATE(6567)] = 64478,
  [SMALL_STATE(6568)] = 64485,
  [SMALL_STATE(6569)] = 64492,
  [SMALL_STATE(6570)] = 64499,
  [SMALL_STATE(6571)] = 64506,
  [SMALL_STATE(6572)] = 64513,
  [SMALL_STATE(6573)] = 64520,
  [SMALL_STATE(6574)] = 64527,
  [SMALL_STATE(6575)] = 64534,
  [SMALL_STATE(6576)] = 64541,
  [SMALL_STATE(6577)] = 64548,
  [SMALL_STATE(6578)] = 64555,
  [SMALL_STATE(6579)] = 64562,
  [SMALL_STATE(6580)] = 64569,
  [SMALL_STATE(6581)] = 64576,
  [SMALL_STATE(6582)] = 64583,
  [SMALL_STATE(6583)] = 64590,
  [SMALL_STATE(6584)] = 64597,
  [SMALL_STATE(6585)] = 64604,
  [SMALL_STATE(6586)] = 64611,
  [SMALL_STATE(6587)] = 64618,
  [SMALL_STATE(6588)] = 64625,
  [SMALL_STATE(6589)] = 64632,
  [SMALL_STATE(6590)] = 64639,
  [SMALL_STATE(6591)] = 64646,
  [SMALL_STATE(6592)] = 64653,
  [SMALL_STATE(6593)] = 64660,
  [SMALL_STATE(6594)] = 64667,
  [SMALL_STATE(6595)] = 64674,
  [SMALL_STATE(6596)] = 64681,
  [SMALL_STATE(6597)] = 64688,
  [SMALL_STATE(6598)] = 64695,
  [SMALL_STATE(6599)] = 64702,
  [SMALL_STATE(6600)] = 64709,
  [SMALL_STATE(6601)] = 64716,
  [SMALL_STATE(6602)] = 64723,
  [SMALL_STATE(6603)] = 64730,
  [SMALL_STATE(6604)] = 64737,
  [SMALL_STATE(6605)] = 64744,
  [SMALL_STATE(6606)] = 64751,
  [SMALL_STATE(6607)] = 64758,
  [SMALL_STATE(6608)] = 64765,
  [SMALL_STATE(6609)] = 64772,
  [SMALL_STATE(6610)] = 64779,
  [SMALL_STATE(6611)] = 64786,
  [SMALL_STATE(6612)] = 64793,
  [SMALL_STATE(6613)] = 64800,
  [SMALL_STATE(6614)] = 64807,
  [SMALL_STATE(6615)] = 64814,
  [SMALL_STATE(6616)] = 64821,
  [SMALL_STATE(6617)] = 64828,
  [SMALL_STATE(6618)] = 64835,
  [SMALL_STATE(6619)] = 64842,
  [SMALL_STATE(6620)] = 64849,
  [SMALL_STATE(6621)] = 64856,
  [SMALL_STATE(6622)] = 64863,
  [SMALL_STATE(6623)] = 64870,
  [SMALL_STATE(6624)] = 64877,
  [SMALL_STATE(6625)] = 64884,
  [SMALL_STATE(6626)] = 64891,
  [SMALL_STATE(6627)] = 64898,
  [SMALL_STATE(6628)] = 64905,
  [SMALL_STATE(6629)] = 64912,
  [SMALL_STATE(6630)] = 64919,
  [SMALL_STATE(6631)] = 64926,
  [SMALL_STATE(6632)] = 64933,
  [SMALL_STATE(6633)] = 64940,
  [SMALL_STATE(6634)] = 64947,
  [SMALL_STATE(6635)] = 64954,
  [SMALL_STATE(6636)] = 64961,
  [SMALL_STATE(6637)] = 64968,
  [SMALL_STATE(6638)] = 64975,
  [SMALL_STATE(6639)] = 64982,
  [SMALL_STATE(6640)] = 64989,
  [SMALL_STATE(6641)] = 64996,
  [SMALL_STATE(6642)] = 65003,
  [SMALL_STATE(6643)] = 65010,
  [SMALL_STATE(6644)] = 65017,
  [SMALL_STATE(6645)] = 65024,
  [SMALL_STATE(6646)] = 65031,
  [SMALL_STATE(6647)] = 65038,
  [SMALL_STATE(6648)] = 65045,
  [SMALL_STATE(6649)] = 65052,
  [SMALL_STATE(6650)] = 65059,
  [SMALL_STATE(6651)] = 65066,
  [SMALL_STATE(6652)] = 65073,
  [SMALL_STATE(6653)] = 65080,
  [SMALL_STATE(6654)] = 65087,
  [SMALL_STATE(6655)] = 65094,
  [SMALL_STATE(6656)] = 65101,
  [SMALL_STATE(6657)] = 65108,
  [SMALL_STATE(6658)] = 65115,
  [SMALL_STATE(6659)] = 65122,
  [SMALL_STATE(6660)] = 65129,
  [SMALL_STATE(6661)] = 65136,
  [SMALL_STATE(6662)] = 65143,
  [SMALL_STATE(6663)] = 65150,
  [SMALL_STATE(6664)] = 65157,
  [SMALL_STATE(6665)] = 65164,
  [SMALL_STATE(6666)] = 65171,
  [SMALL_STATE(6667)] = 65178,
  [SMALL_STATE(6668)] = 65185,
  [SMALL_STATE(6669)] = 65192,
  [SMALL_STATE(6670)] = 65199,
  [SMALL_STATE(6671)] = 65206,
  [SMALL_STATE(6672)] = 65213,
  [SMALL_STATE(6673)] = 65220,
  [SMALL_STATE(6674)] = 65227,
  [SMALL_STATE(6675)] = 65234,
  [SMALL_STATE(6676)] = 65241,
  [SMALL_STATE(6677)] = 65248,
  [SMALL_STATE(6678)] = 65255,
  [SMALL_STATE(6679)] = 65262,
  [SMALL_STATE(6680)] = 65269,
  [SMALL_STATE(6681)] = 65276,
  [SMALL_STATE(6682)] = 65283,
  [SMALL_STATE(6683)] = 65290,
  [SMALL_STATE(6684)] = 65297,
  [SMALL_STATE(6685)] = 65304,
  [SMALL_STATE(6686)] = 65311,
  [SMALL_STATE(6687)] = 65318,
  [SMALL_STATE(6688)] = 65325,
  [SMALL_STATE(6689)] = 65332,
  [SMALL_STATE(6690)] = 65339,
  [SMALL_STATE(6691)] = 65346,
  [SMALL_STATE(6692)] = 65353,
  [SMALL_STATE(6693)] = 65360,
  [SMALL_STATE(6694)] = 65367,
  [SMALL_STATE(6695)] = 65374,
  [SMALL_STATE(6696)] = 65381,
  [SMALL_STATE(6697)] = 65388,
  [SMALL_STATE(6698)] = 65395,
  [SMALL_STATE(6699)] = 65402,
  [SMALL_STATE(6700)] = 65409,
  [SMALL_STATE(6701)] = 65416,
  [SMALL_STATE(6702)] = 65423,
  [SMALL_STATE(6703)] = 65430,
  [SMALL_STATE(6704)] = 65437,
  [SMALL_STATE(6705)] = 65444,
  [SMALL_STATE(6706)] = 65451,
  [SMALL_STATE(6707)] = 65458,
  [SMALL_STATE(6708)] = 65465,
  [SMALL_STATE(6709)] = 65472,
  [SMALL_STATE(6710)] = 65479,
  [SMALL_STATE(6711)] = 65486,
  [SMALL_STATE(6712)] = 65493,
  [SMALL_STATE(6713)] = 65500,
  [SMALL_STATE(6714)] = 65507,
  [SMALL_STATE(6715)] = 65514,
  [SMALL_STATE(6716)] = 65521,
  [SMALL_STATE(6717)] = 65528,
  [SMALL_STATE(6718)] = 65535,
  [SMALL_STATE(6719)] = 65542,
  [SMALL_STATE(6720)] = 65549,
  [SMALL_STATE(6721)] = 65556,
  [SMALL_STATE(6722)] = 65563,
  [SMALL_STATE(6723)] = 65570,
  [SMALL_STATE(6724)] = 65577,
  [SMALL_STATE(6725)] = 65584,
  [SMALL_STATE(6726)] = 65591,
  [SMALL_STATE(6727)] = 65598,
  [SMALL_STATE(6728)] = 65605,
  [SMALL_STATE(6729)] = 65612,
  [SMALL_STATE(6730)] = 65619,
  [SMALL_STATE(6731)] = 65626,
  [SMALL_STATE(6732)] = 65633,
  [SMALL_STATE(6733)] = 65640,
  [SMALL_STATE(6734)] = 65647,
  [SMALL_STATE(6735)] = 65654,
  [SMALL_STATE(6736)] = 65661,
  [SMALL_STATE(6737)] = 65668,
  [SMALL_STATE(6738)] = 65675,
  [SMALL_STATE(6739)] = 65682,
  [SMALL_STATE(6740)] = 65689,
  [SMALL_STATE(6741)] = 65696,
  [SMALL_STATE(6742)] = 65703,
  [SMALL_STATE(6743)] = 65710,
  [SMALL_STATE(6744)] = 65717,
  [SMALL_STATE(6745)] = 65724,
  [SMALL_STATE(6746)] = 65731,
  [SMALL_STATE(6747)] = 65738,
  [SMALL_STATE(6748)] = 65745,
  [SMALL_STATE(6749)] = 65752,
  [SMALL_STATE(6750)] = 65759,
  [SMALL_STATE(6751)] = 65766,
  [SMALL_STATE(6752)] = 65773,
  [SMALL_STATE(6753)] = 65780,
  [SMALL_STATE(6754)] = 65787,
  [SMALL_STATE(6755)] = 65794,
  [SMALL_STATE(6756)] = 65801,
  [SMALL_STATE(6757)] = 65808,
  [SMALL_STATE(6758)] = 65815,
  [SMALL_STATE(6759)] = 65822,
  [SMALL_STATE(6760)] = 65829,
  [SMALL_STATE(6761)] = 65836,
  [SMALL_STATE(6762)] = 65843,
  [SMALL_STATE(6763)] = 65850,
  [SMALL_STATE(6764)] = 65857,
  [SMALL_STATE(6765)] = 65864,
  [SMALL_STATE(6766)] = 65871,
  [SMALL_STATE(6767)] = 65878,
  [SMALL_STATE(6768)] = 65885,
  [SMALL_STATE(6769)] = 65892,
  [SMALL_STATE(6770)] = 65899,
  [SMALL_STATE(6771)] = 65906,
  [SMALL_STATE(6772)] = 65913,
  [SMALL_STATE(6773)] = 65920,
  [SMALL_STATE(6774)] = 65927,
  [SMALL_STATE(6775)] = 65934,
  [SMALL_STATE(6776)] = 65941,
  [SMALL_STATE(6777)] = 65948,
  [SMALL_STATE(6778)] = 65955,
  [SMALL_STATE(6779)] = 65962,
  [SMALL_STATE(6780)] = 65969,
  [SMALL_STATE(6781)] = 65976,
  [SMALL_STATE(6782)] = 65983,
  [SMALL_STATE(6783)] = 65990,
  [SMALL_STATE(6784)] = 65997,
  [SMALL_STATE(6785)] = 66004,
  [SMALL_STATE(6786)] = 66011,
  [SMALL_STATE(6787)] = 66018,
  [SMALL_STATE(6788)] = 66025,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_translation_unit, 0, 0, 0),
  [7] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1616),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4841),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6512),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4614),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6613),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5733),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1531),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3472),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3472),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3557),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3283),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(912),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3954),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2657),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6441),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2354),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3403),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6702),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2907),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2862),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2381),
  [53] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6486),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1805),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(919),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6451),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2973),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4876),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4362),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4932),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5816),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5819),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3709),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6715),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5721),
  [79] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2467),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6100),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4563),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3370),
  [87] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6609),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6227),
  [91] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6024),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3742),
  [95] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3458),
  [97] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6690),
  [99] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6233),
  [101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6242),
  [103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4952),
  [105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4564),
  [107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5610),
  [109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3354),
  [111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5263),
  [113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4564),
  [115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1628),
  [117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6087),
  [119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6144),
  [121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6327),
  [123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6337),
  [125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2372),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6623),
  [129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6110),
  [131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6180),
  [133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6208),
  [135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6225),
  [137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5831),
  [139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6277),
  [141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6277),
  [143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6299),
  [145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6308),
  [147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2909),
  [149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2562),
  [151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1596),
  [153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4837),
  [155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5916),
  [157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4572),
  [159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2129),
  [161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5917),
  [163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(39),
  [165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4588),
  [167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6108),
  [169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5743),
  [171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(914),
  [175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3895),
  [177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2661),
  [179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6375),
  [181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5920),
  [185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1765),
  [187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4361),
  [189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5770),
  [191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5893),
  [193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3545),
  [195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6700),
  [197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5782),
  [199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2458),
  [201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6576),
  [203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1629),
  [205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3384),
  [207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5924),
  [209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6132),
  [211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6591),
  [213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4915),
  [215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1641),
  [217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3385),
  [219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1641),
  [221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6594),
  [223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5928),
  [225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6380),
  [227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6600),
  [229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6601),
  [231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6317),
  [233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5884),
  [235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5436),
  [237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5436),
  [239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3424),
  [241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6736),
  [243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2498),
  [245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1032),
  [247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(851),
  [249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(876),
  [251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2117),
  [253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1044),
  [255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(313),
  [257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(301),
  [259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4057),
  [261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4059),
  [263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(483),
  [265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(526),
  [267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(657),
  [269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(659),
  [271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2116),
  [273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2125),
  [275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2285),
  [277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2294),
  [279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1587),
  [281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4857),
  [283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5986),
  [285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4584),
  [287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(310),
  [289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6045),
  [291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5719),
  [293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(916),
  [297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3911),
  [299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2655),
  [301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6446),
  [303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6090),
  [307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5803),
  [309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5868),
  [311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3549),
  [313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6364),
  [315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5808),
  [317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2425),
  [319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6733),
  [321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3338),
  [323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6366),
  [325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6370),
  [327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5995),
  [329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4924),
  [331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3359),
  [333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5927),
  [335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5997),
  [337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6450),
  [339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5929),
  [341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5930),
  [343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6647),
  [345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5741),
  [347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5478),
  [349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5478),
  [351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3407),
  [353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6754),
  [355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(330),
  [357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif, 3, 0, 47),
  [359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4056),
  [361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elifdef, 3, 0, 16),
  [363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4049),
  [365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif, 4, 0, 47),
  [367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(520),
  [369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(448),
  [371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(648),
  [373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1060),
  [375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(652),
  [377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(869),
  [379] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1596),
  [382] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4837),
  [385] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5916),
  [388] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4572),
  [391] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0),
  [393] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5917),
  [396] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5743),
  [399] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1531),
  [402] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3472),
  [405] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3472),
  [408] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3557),
  [411] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3283),
  [414] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(230),
  [417] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(914),
  [420] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3895),
  [423] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2661),
  [426] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6375),
  [429] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [432] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3403),
  [435] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [438] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2907),
  [441] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(56),
  [444] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2862),
  [447] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [450] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5920),
  [453] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1765),
  [456] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(919),
  [459] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [462] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2973),
  [465] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4876),
  [468] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4361),
  [471] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4932),
  [474] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5770),
  [477] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5893),
  [480] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3545),
  [483] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6700),
  [486] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5782),
  [489] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2458),
  [492] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6576),
  [495] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1629),
  [498] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3384),
  [501] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5924),
  [504] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6132),
  [507] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6591),
  [510] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3742),
  [513] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3458),
  [516] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6690),
  [519] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6233),
  [522] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6242),
  [525] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4952),
  [528] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4915),
  [531] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1641),
  [534] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5610),
  [537] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3385),
  [540] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5263),
  [543] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1641),
  [546] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1628),
  [549] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6087),
  [552] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6594),
  [555] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5928),
  [558] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6380),
  [561] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [564] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [567] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6600),
  [570] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6180),
  [573] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6208),
  [576] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6601),
  [579] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6317),
  [582] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5884),
  [585] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5436),
  [588] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5436),
  [591] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3424),
  [594] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5831),
  [597] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6277),
  [600] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6277),
  [603] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6299),
  [606] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6308),
  [609] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6736),
  [612] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2909),
  [615] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2498),
  [618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(844),
  [620] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elifdef, 2, 0, 16),
  [622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1030),
  [624] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1587),
  [627] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4857),
  [630] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5986),
  [633] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4584),
  [636] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6045),
  [639] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5719),
  [642] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(383),
  [645] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(916),
  [648] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3911),
  [651] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2655),
  [654] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6446),
  [657] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(72),
  [660] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6090),
  [663] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5803),
  [666] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5868),
  [669] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3549),
  [672] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6364),
  [675] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5808),
  [678] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2425),
  [681] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6733),
  [684] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3338),
  [687] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6366),
  [690] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6370),
  [693] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5995),
  [696] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4924),
  [699] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3359),
  [702] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5927),
  [705] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5997),
  [708] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6450),
  [711] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5929),
  [714] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5930),
  [717] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6647),
  [720] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5741),
  [723] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5478),
  [726] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5478),
  [729] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3407),
  [732] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6754),
  [735] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1610),
  [738] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4853),
  [741] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6044),
  [744] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4587),
  [747] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6089),
  [750] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5885),
  [753] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(543),
  [756] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(915),
  [759] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3918),
  [762] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2656),
  [765] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6477),
  [768] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(76),
  [771] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6121),
  [774] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5866),
  [777] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5720),
  [780] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3551),
  [783] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6749),
  [786] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5869),
  [789] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2474),
  [792] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6743),
  [795] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3344),
  [798] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6755),
  [801] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6757),
  [804] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6051),
  [807] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(4926),
  [810] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3387),
  [813] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5996),
  [816] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6052),
  [819] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6480),
  [822] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5998),
  [825] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5999),
  [828] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5949),
  [831] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5834),
  [834] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5494),
  [837] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(5494),
  [840] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(3415),
  [843] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(6761),
  [846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1610),
  [848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4853),
  [850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6044),
  [852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4587),
  [854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else, 1, 0, 0),
  [856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6089),
  [858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5885),
  [860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(915),
  [864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3918),
  [866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2656),
  [868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6477),
  [870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6121),
  [874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5866),
  [876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5720),
  [878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3551),
  [880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6749),
  [882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5869),
  [884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2474),
  [886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6743),
  [888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3344),
  [890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6755),
  [892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6757),
  [894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6051),
  [896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4926),
  [898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3387),
  [900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5996),
  [902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6052),
  [904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6480),
  [906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5998),
  [908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5999),
  [910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5949),
  [912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5834),
  [914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5494),
  [916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5494),
  [918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3415),
  [920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6761),
  [922] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else, 2, 0, 0),
  [924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1606),
  [926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4859),
  [928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6193),
  [930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4589),
  [932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6194),
  [934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5822),
  [936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(911),
  [940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3903),
  [942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2660),
  [944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1080),
  [948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6574),
  [950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5840),
  [952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5775),
  [954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3524),
  [956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5969),
  [958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5863),
  [960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2452),
  [962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6179),
  [964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3357),
  [966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5989),
  [968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5993),
  [970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6210),
  [972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4893),
  [974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3361),
  [976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6216),
  [978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6219),
  [980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6185),
  [982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6220),
  [984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6221),
  [986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5381),
  [988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5381),
  [990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3430),
  [992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6645),
  [994] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1606),
  [997] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4859),
  [1000] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6193),
  [1003] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4589),
  [1006] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6194),
  [1009] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5822),
  [1012] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1531),
  [1015] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3472),
  [1018] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3472),
  [1021] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3557),
  [1024] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3283),
  [1027] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(708),
  [1030] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(911),
  [1033] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3903),
  [1036] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2660),
  [1039] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [1042] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [1045] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3403),
  [1048] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [1051] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2907),
  [1054] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(47),
  [1057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0),
  [1059] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2862),
  [1062] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [1065] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6574),
  [1068] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1765),
  [1071] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(919),
  [1074] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [1077] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2973),
  [1080] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4876),
  [1083] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4361),
  [1086] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4932),
  [1089] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5840),
  [1092] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5775),
  [1095] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3524),
  [1098] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5969),
  [1101] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5863),
  [1104] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2452),
  [1107] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6179),
  [1110] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1629),
  [1113] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3357),
  [1116] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5989),
  [1119] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5993),
  [1122] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6210),
  [1125] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3742),
  [1128] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3458),
  [1131] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6690),
  [1134] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6233),
  [1137] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6242),
  [1140] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4952),
  [1143] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4893),
  [1146] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1641),
  [1149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5610),
  [1152] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3361),
  [1155] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5263),
  [1158] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1641),
  [1161] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1628),
  [1164] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6087),
  [1167] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6216),
  [1170] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6219),
  [1173] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6185),
  [1176] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [1179] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [1182] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6220),
  [1185] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6180),
  [1188] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6208),
  [1191] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6221),
  [1194] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5381),
  [1197] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5381),
  [1200] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3430),
  [1203] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5831),
  [1206] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6277),
  [1209] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6277),
  [1212] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6299),
  [1215] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6308),
  [1218] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6645),
  [1221] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2909),
  [1224] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2498),
  [1227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [1229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1571),
  [1231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(913),
  [1233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [1235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [1237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(909),
  [1239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [1241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [1243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [1245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [1247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [1249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [1251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [1253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [1255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [1257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [1259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1451),
  [1261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1496),
  [1263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [1265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1505),
  [1267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [1269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [1271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [1273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(700),
  [1275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [1277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [1279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [1281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(584),
  [1283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1065),
  [1285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [1287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1568),
  [1289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [1291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [1293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [1295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1104),
  [1297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1569),
  [1299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [1301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2081),
  [1303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2082),
  [1305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2083),
  [1307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [1309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [1311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [1313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(752),
  [1315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [1317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(754),
  [1319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [1321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [1323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [1325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [1327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [1329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [1331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [1333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [1335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [1337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [1339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [1341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [1343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1079),
  [1345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1081),
  [1347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2060),
  [1349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2061),
  [1351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2062),
  [1353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2231),
  [1355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2238),
  [1357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2215),
  [1359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0),
  [1361] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1616),
  [1364] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4841),
  [1367] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6512),
  [1370] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4614),
  [1373] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6613),
  [1376] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(5733),
  [1379] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1531),
  [1382] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3472),
  [1385] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3472),
  [1388] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3557),
  [1391] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3283),
  [1394] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1029),
  [1397] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(912),
  [1400] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3954),
  [1403] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2657),
  [1406] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [1409] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [1412] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3403),
  [1415] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [1418] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2907),
  [1421] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(51),
  [1424] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2862),
  [1427] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [1430] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6486),
  [1433] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1805),
  [1436] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(919),
  [1439] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [1442] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2973),
  [1445] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4876),
  [1448] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4362),
  [1451] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4932),
  [1454] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(5816),
  [1457] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(5819),
  [1460] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3709),
  [1463] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6715),
  [1466] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(5721),
  [1469] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2467),
  [1472] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6100),
  [1475] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4563),
  [1478] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3370),
  [1481] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6609),
  [1484] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6227),
  [1487] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6024),
  [1490] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3742),
  [1493] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3458),
  [1496] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6690),
  [1499] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6233),
  [1502] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6242),
  [1505] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4952),
  [1508] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4564),
  [1511] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(5610),
  [1514] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(3354),
  [1517] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(5263),
  [1520] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4564),
  [1523] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1628),
  [1526] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6087),
  [1529] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6144),
  [1532] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6327),
  [1535] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6337),
  [1538] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [1541] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [1544] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6110),
  [1547] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6180),
  [1550] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6208),
  [1553] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6225),
  [1556] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(5831),
  [1559] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6277),
  [1562] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6277),
  [1565] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6299),
  [1568] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(6308),
  [1571] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2909),
  [1574] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2562),
  [1577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_translation_unit, 1, 0, 0),
  [1579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1608),
  [1581] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 4, 0, 7),
  [1583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1069),
  [1585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3443),
  [1587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1073),
  [1589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3426),
  [1591] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 4, 0, 7),
  [1593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2541),
  [1595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 2, 0, 0),
  [1597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 2, 0, 0),
  [1599] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 3, 0, 7),
  [1601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 3, 0, 7),
  [1603] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 3, 0, 0),
  [1605] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 3, 0, 0),
  [1607] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1608),
  [1610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0),
  [1612] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1531),
  [1615] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3472),
  [1618] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3472),
  [1621] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3557),
  [1624] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3283),
  [1627] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(230),
  [1630] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1069),
  [1633] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3895),
  [1636] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [1639] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [1642] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [1645] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3443),
  [1648] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [1651] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(56),
  [1654] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2862),
  [1657] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5920),
  [1660] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1765),
  [1663] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1073),
  [1666] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [1669] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2973),
  [1672] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(4876),
  [1675] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(4361),
  [1678] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(4932),
  [1681] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5770),
  [1684] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5893),
  [1687] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5782),
  [1690] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2458),
  [1693] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6576),
  [1696] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1629),
  [1699] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3384),
  [1702] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5924),
  [1705] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6132),
  [1708] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6591),
  [1711] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3742),
  [1714] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3458),
  [1717] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6690),
  [1720] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6233),
  [1723] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6242),
  [1726] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(4952),
  [1729] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(4915),
  [1732] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1641),
  [1735] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5610),
  [1738] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3426),
  [1741] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5263),
  [1744] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1641),
  [1747] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1628),
  [1750] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6087),
  [1753] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0),
  [1755] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [1758] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [1761] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5436),
  [1764] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5436),
  [1767] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3424),
  [1770] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5831),
  [1773] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6277),
  [1776] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6277),
  [1779] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6299),
  [1782] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6308),
  [1785] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6736),
  [1788] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2909),
  [1791] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2541),
  [1794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1579),
  [1796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1070),
  [1798] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1579),
  [1801] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(383),
  [1804] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1070),
  [1807] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3911),
  [1810] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(72),
  [1813] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6090),
  [1816] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5803),
  [1819] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5868),
  [1822] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5808),
  [1825] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2425),
  [1828] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6733),
  [1831] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3338),
  [1834] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6366),
  [1837] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6370),
  [1840] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5995),
  [1843] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(4924),
  [1846] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5478),
  [1849] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5478),
  [1852] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3407),
  [1855] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6754),
  [1858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1582),
  [1860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(992),
  [1862] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1582),
  [1865] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(543),
  [1868] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(992),
  [1871] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3918),
  [1874] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(76),
  [1877] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6121),
  [1880] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5866),
  [1883] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5720),
  [1886] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5869),
  [1889] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2474),
  [1892] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6743),
  [1895] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3344),
  [1898] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6755),
  [1901] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6757),
  [1904] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6051),
  [1907] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(4926),
  [1910] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5494),
  [1913] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5494),
  [1916] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3415),
  [1919] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6761),
  [1922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1583),
  [1924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1067),
  [1926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1594),
  [1928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(709),
  [1930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1025),
  [1932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3977),
  [1934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [1936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6046),
  [1938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5886),
  [1940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5805),
  [1942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5896),
  [1944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2469),
  [1946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6668),
  [1948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3371),
  [1950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6713),
  [1952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6716),
  [1954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5925),
  [1956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4922),
  [1958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5463),
  [1960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5463),
  [1962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3399),
  [1964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6746),
  [1966] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1594),
  [1969] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(709),
  [1972] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1025),
  [1975] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3977),
  [1978] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(68),
  [1981] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6046),
  [1984] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5886),
  [1987] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5805),
  [1990] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5896),
  [1993] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2469),
  [1996] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6668),
  [1999] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3371),
  [2002] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6713),
  [2005] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6716),
  [2008] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5925),
  [2011] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(4922),
  [2014] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5463),
  [2017] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5463),
  [2020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3399),
  [2023] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6746),
  [2026] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1583),
  [2029] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(708),
  [2032] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1067),
  [2035] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3903),
  [2038] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(47),
  [2041] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6574),
  [2044] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5840),
  [2047] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5775),
  [2050] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5863),
  [2053] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2452),
  [2056] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6179),
  [2059] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3357),
  [2062] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5989),
  [2065] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5993),
  [2068] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6210),
  [2071] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(4893),
  [2074] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5381),
  [2077] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5381),
  [2080] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3430),
  [2083] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6645),
  [2086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1072),
  [2088] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1586),
  [2091] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1154),
  [2094] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1068),
  [2097] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3924),
  [2100] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(81),
  [2103] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6139),
  [2106] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5716),
  [2109] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5751),
  [2112] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5722),
  [2115] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2476),
  [2118] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6751),
  [2121] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3353),
  [2124] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6161),
  [2127] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6162),
  [2130] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6092),
  [2133] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(4930),
  [2136] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5505),
  [2139] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5505),
  [2142] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3417),
  [2145] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(6764),
  [2148] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1597),
  [2150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [2152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4898),
  [2154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5390),
  [2156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5390),
  [2158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3423),
  [2160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6678),
  [2162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1580),
  [2164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5844),
  [2166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3537),
  [2168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6063),
  [2170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5875),
  [2172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6377),
  [2174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5530),
  [2176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5530),
  [2178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6587),
  [2180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1586),
  [2182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1154),
  [2184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1068),
  [2186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3924),
  [2188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [2190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6139),
  [2192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5716),
  [2194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5751),
  [2196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5722),
  [2198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2476),
  [2200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6751),
  [2202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3353),
  [2204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6161),
  [2206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6162),
  [2208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6092),
  [2210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4930),
  [2212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5505),
  [2214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5505),
  [2216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3417),
  [2218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6764),
  [2220] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sizeof_expression, 4, 0, 43),
  [2222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sizeof_expression, 4, 0, 43),
  [2224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3485),
  [2226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3485),
  [2228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1449),
  [2230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3487),
  [2232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6181),
  [2234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6770),
  [2236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6787),
  [2238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4992),
  [2240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1465),
  [2242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5435),
  [2244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3398),
  [2246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5299),
  [2248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1465),
  [2250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1479),
  [2252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5731),
  [2254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6188),
  [2256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6188),
  [2258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6771),
  [2260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6189),
  [2262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1595),
  [2264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3345),
  [2266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3446),
  [2268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1609),
  [2270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1600),
  [2272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1585),
  [2274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1598),
  [2276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1589),
  [2278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1604),
  [2280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1607),
  [2282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3456),
  [2284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3456),
  [2286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3460),
  [2288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4968),
  [2290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3422),
  [2292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5191),
  [2294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3634),
  [2296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3634),
  [2298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3269),
  [2300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3534),
  [2302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3432),
  [2304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5239),
  [2306] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 3, 0, 0),
  [2308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 3, 0, 0),
  [2310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5223),
  [2312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5223),
  [2314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5521),
  [2316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5521),
  [2318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0),
  [2320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0),
  [2322] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5223),
  [2325] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5223),
  [2328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1667),
  [2330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3735),
  [2332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3735),
  [2334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3584),
  [2336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3444),
  [2338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5165),
  [2340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5165),
  [2342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5326),
  [2344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5326),
  [2346] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 2, 0, 0),
  [2348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 2, 0, 0),
  [2350] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 3, 0, 0),
  [2352] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 3, 0, 0),
  [2354] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5165),
  [2357] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5165),
  [2360] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 4, 0, 0),
  [2362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 4, 0, 0),
  [2364] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clause, 2, 0, 0),
  [2366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clause, 2, 0, 0),
  [2368] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clause, 5, 0, 0),
  [2370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clause, 5, 0, 0),
  [2372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5211),
  [2374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5211),
  [2376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5346),
  [2378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5346),
  [2380] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5211),
  [2383] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5211),
  [2386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5253),
  [2388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5253),
  [2390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5653),
  [2392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5653),
  [2394] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 31),
  [2396] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 3, 0, 31),
  [2398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2457),
  [2400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5230),
  [2402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5230),
  [2404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5314),
  [2406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5314),
  [2408] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 148),
  [2410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 148),
  [2412] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 4, 0, 55),
  [2414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 4, 0, 55),
  [2416] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 38),
  [2418] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 38),
  [2420] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 73),
  [2422] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 73),
  [2424] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 74),
  [2426] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 74),
  [2428] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 5, 0, 91),
  [2430] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 5, 0, 91),
  [2432] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5253),
  [2435] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5253),
  [2438] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 5, 0, 73),
  [2440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 5, 0, 73),
  [2442] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 5, 0, 111),
  [2444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 5, 0, 111),
  [2446] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 5, 0, 74),
  [2448] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 5, 0, 74),
  [2450] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_statement, 1, 0, 0),
  [2452] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0),
  [2454] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement, 1, 0, 0),
  [2456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement, 1, 0, 0),
  [2458] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 2, 0, 0),
  [2460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 2, 0, 0),
  [2462] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_statement, 2, 0, 0),
  [2464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_statement, 2, 0, 0),
  [2466] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_statement, 2, 0, 0),
  [2468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_statement, 2, 0, 0),
  [2470] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attributed_statement, 2, 0, 0),
  [2472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attributed_statement, 2, 0, 0),
  [2474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 3, 0, 38),
  [2476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 3, 0, 38),
  [2478] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_statement, 2, 0, 0),
  [2480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 2, 0, 0),
  [2482] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_statement, 3, 0, 32),
  [2484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_statement, 3, 0, 32),
  [2486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 3, 0, 32),
  [2488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 3, 0, 32),
  [2490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 3, 0, 0),
  [2492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 3, 0, 0),
  [2494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_goto_statement, 3, 0, 33),
  [2496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_goto_statement, 3, 0, 33),
  [2498] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_statement, 3, 0, 35),
  [2500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_statement, 3, 0, 35),
  [2502] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_throw_statement, 3, 0, 0),
  [2504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_throw_statement, 3, 0, 0),
  [2506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 66),
  [2508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 66),
  [2510] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 6, 0, 120),
  [2512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 6, 0, 120),
  [2514] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ms_asm_block, 4, 0, 0),
  [2516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_asm_block, 4, 0, 0),
  [2518] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_finally_clause, 2, 0, 0),
  [2520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_finally_clause, 2, 0, 0),
  [2522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 4, 0, 0),
  [2524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 4, 0, 0),
  [2526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2, 0, 0),
  [2528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2, 0, 0),
  [2530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_statement, 5, 0, 100),
  [2532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_statement, 5, 0, 100),
  [2534] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 5, 0, 104),
  [2536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 5, 0, 104),
  [2538] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_synchronized_statement, 5, 0, 0),
  [2540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_synchronized_statement, 5, 0, 0),
  [2542] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_synchronized_statement, 6, 0, 0),
  [2544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_synchronized_statement, 6, 0, 0),
  [2546] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 6, 0, 111),
  [2548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 6, 0, 111),
  [2550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 0),
  [2552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 0),
  [2554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 7, 0, 138),
  [2556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 7, 0, 138),
  [2558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 8, 0, 147),
  [2560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 8, 0, 147),
  [2562] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5230),
  [2565] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5230),
  [2568] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_throw_statement, 2, 0, 0),
  [2570] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_throw_statement, 2, 0, 0),
  [2572] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 6, 0, 42),
  [2574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 6, 0, 42),
  [2576] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_implementation, 6, 0, 82),
  [2578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_implementation, 6, 0, 82),
  [2580] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_declaration, 7, 0, 0),
  [2582] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_declaration, 7, 0, 0),
  [2584] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 3, 0, 39),
  [2586] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 3, 0, 39),
  [2588] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_def, 3, 0, 16),
  [2590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_def, 3, 0, 16),
  [2592] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_specifier, 7, 0, 0),
  [2594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_specifier, 7, 0, 0),
  [2596] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 4, 0, 77),
  [2598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 4, 0, 77),
  [2600] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_function_def, 5, 0, 83),
  [2602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_function_def, 5, 0, 83),
  [2604] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef, 5, 0, 85),
  [2606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef, 5, 0, 85),
  [2608] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_def, 4, 0, 45),
  [2610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_def, 4, 0, 45),
  [2612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2443),
  [2614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_implementation, 3, 0, 0),
  [2616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_implementation, 3, 0, 0),
  [2618] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_function_def, 4, 0, 46),
  [2620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_function_def, 4, 0, 46),
  [2622] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_implementation, 4, 0, 0),
  [2624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_implementation, 4, 0, 0),
  [2626] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_implementation, 5, 0, 0),
  [2628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_implementation, 5, 0, 0),
  [2630] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_implementation, 6, 0, 0),
  [2632] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_implementation, 6, 0, 0),
  [2634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef, 4, 0, 48),
  [2636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef, 4, 0, 48),
  [2638] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__empty_declaration, 2, 0, 0),
  [2640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__empty_declaration, 2, 0, 0),
  [2642] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 2, 0, 0),
  [2644] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 2, 0, 0),
  [2646] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_implementation, 2, 0, 0),
  [2648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_implementation, 2, 0, 0),
  [2650] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_include, 3, 0, 15),
  [2652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_include, 3, 0, 15),
  [2654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_linkage_specification, 3, 0, 25),
  [2656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_linkage_specification, 3, 0, 25),
  [2658] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__block_item, 1, 0, 3),
  [2660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__block_item, 1, 0, 3),
  [2662] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__block_item, 1, 0, 0), REDUCE(sym_statement, 1, 0, 0),
  [2665] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__block_item, 1, 0, 0), REDUCE(sym_statement, 1, 0, 0),
  [2668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef, 4, 0, 16),
  [2670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef, 4, 0, 16),
  [2672] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_import, 3, 0, 36),
  [2674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_import, 3, 0, 36),
  [2676] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_forward_declaration, 3, 0, 0),
  [2678] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_forward_declaration, 3, 0, 0),
  [2680] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_declaration, 3, 0, 0),
  [2682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_declaration, 3, 0, 0),
  [2684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compatibility_alias_declaration, 3, 0, 37),
  [2686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compatibility_alias_declaration, 3, 0, 37),
  [2688] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_undef, 3, 0, 16),
  [2690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_undef, 3, 0, 16),
  [2692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 3, 0, 42),
  [2694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 3, 0, 42),
  [2696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_implementation, 3, 0, 0),
  [2698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_implementation, 3, 0, 0),
  [2700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_implementation, 3, 0, 42),
  [2702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_implementation, 3, 0, 42),
  [2704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if, 4, 0, 47),
  [2706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if, 4, 0, 47),
  [2708] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_list, 2, 0, 0),
  [2710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_list, 2, 0, 0),
  [2712] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_specifier, 4, 0, 0),
  [2714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_specifier, 4, 0, 0),
  [2716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef, 3, 0, 16),
  [2718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef, 3, 0, 16),
  [2720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 4, 0, 0),
  [2722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 4, 0, 0),
  [2724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_import, 4, 0, 70),
  [2726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_import, 4, 0, 70),
  [2728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_linemarker, 4, 0, 71),
  [2730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_linemarker, 4, 0, 71),
  [2732] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_declaration, 4, 0, 0),
  [2734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_declaration, 4, 0, 0),
  [2736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_forward_declaration, 4, 0, 0),
  [2738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_forward_declaration, 4, 0, 0),
  [2740] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_function_definition, 4, 0, 76),
  [2742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__old_style_function_definition, 4, 0, 76),
  [2744] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 4, 0, 82),
  [2746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 4, 0, 82),
  [2748] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 4, 0, 0),
  [2750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 4, 0, 0),
  [2752] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 4, 0, 42),
  [2754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 4, 0, 42),
  [2756] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_implementation, 4, 0, 82),
  [2758] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_implementation, 4, 0, 82),
  [2760] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_implementation, 4, 0, 0),
  [2762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_implementation, 4, 0, 0),
  [2764] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_implementation, 4, 0, 42),
  [2766] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_implementation, 4, 0, 42),
  [2768] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if, 5, 0, 84),
  [2770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if, 5, 0, 84),
  [2772] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if, 5, 0, 47),
  [2774] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if, 5, 0, 47),
  [2776] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_list, 3, 0, 0),
  [2778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_list, 3, 0, 0),
  [2780] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 0),
  [2782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 0),
  [2784] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_linemarker, 5, 0, 108),
  [2786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_linemarker, 5, 0, 108),
  [2788] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_declaration, 5, 0, 0),
  [2790] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_declaration, 5, 0, 0),
  [2792] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_function_definition, 5, 0, 112),
  [2794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__old_style_function_definition, 5, 0, 112),
  [2796] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 5, 0, 82),
  [2798] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 5, 0, 82),
  [2800] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 5, 0, 0),
  [2802] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 5, 0, 0),
  [2804] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 5, 0, 42),
  [2806] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 5, 0, 42),
  [2808] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_implementation, 5, 0, 82),
  [2810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_implementation, 5, 0, 82),
  [2812] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_implementation, 5, 0, 0),
  [2814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_implementation, 5, 0, 0),
  [2816] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_implementation, 5, 0, 42),
  [2818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_implementation, 5, 0, 42),
  [2820] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_forward_declaration, 5, 0, 0),
  [2822] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_forward_declaration, 5, 0, 0),
  [2824] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if, 6, 0, 115),
  [2826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if, 6, 0, 115),
  [2828] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_specifier, 6, 0, 0),
  [2830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_specifier, 6, 0, 0),
  [2832] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 6, 0, 0),
  [2834] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 6, 0, 0),
  [2836] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_linemarker, 6, 0, 134),
  [2838] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_linemarker, 6, 0, 134),
  [2840] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_declaration, 6, 0, 0),
  [2842] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_declaration, 6, 0, 0),
  [2844] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 6, 0, 82),
  [2846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 6, 0, 82),
  [2848] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 6, 0, 0),
  [2850] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 6, 0, 0),
  [2852] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_call, 3, 0, 17),
  [2854] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_call, 3, 0, 17),
  [2856] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_call, 2, 0, 5),
  [2858] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_call, 2, 0, 5),
  [2860] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 7, 0, 82),
  [2862] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 7, 0, 82),
  [2864] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_interface, 3, 0, 0),
  [2866] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface, 3, 0, 0),
  [2868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5190),
  [2870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5190),
  [2872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5616),
  [2874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5616),
  [2876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2428),
  [2878] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5190),
  [2881] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5190),
  [2884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3630),
  [2886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3630),
  [2888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3272),
  [2890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3632),
  [2892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3431),
  [2894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2450),
  [2896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2464),
  [2898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1825),
  [2900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3711),
  [2902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3711),
  [2904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3552),
  [2906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3434),
  [2908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4220),
  [2910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3686),
  [2912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3948),
  [2914] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0),
  [2916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3411),
  [2918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3451),
  [2920] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_qualifier, 1, 0, 0),
  [2922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3975),
  [2924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3896),
  [2926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3919),
  [2928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3912),
  [2930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2468),
  [2932] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__top_level_item, 1, 0, 0), REDUCE(sym__top_level_statement, 1, 0, 0),
  [2935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__top_level_item, 1, 0, 0), REDUCE(sym__top_level_statement, 1, 0, 0),
  [2938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__top_level_expression_statement, 2, 0, 0),
  [2940] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__top_level_expression_statement, 2, 0, 0),
  [2942] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__top_level_item, 1, 0, 3),
  [2944] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__top_level_item, 1, 0, 3),
  [2946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3905),
  [2948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__top_level_expression_statement, 1, 0, 0),
  [2950] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__top_level_expression_statement, 1, 0, 0),
  [2952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__top_level_expression_statement, 2, 0, 1),
  [2954] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__top_level_expression_statement, 2, 0, 1),
  [2956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3925),
  [2958] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_declaration, 5, 0, 0),
  [2960] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_declaration, 5, 0, 0),
  [2962] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_declaration, 6, 0, 0),
  [2964] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_declaration, 6, 0, 0),
  [2966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5268),
  [2968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5268),
  [2970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5354),
  [2972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5354),
  [2974] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5268),
  [2977] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5268),
  [2980] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__declaration_modifiers, 1, 0, 0), REDUCE(aux_sym_attributed_field_declarator_repeat1, 1, 0, 0),
  [2983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 1, 0, 0),
  [2985] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 1, 0, 0),
  [2987] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_modifiers, 1, 0, 0),
  [2989] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__declaration_modifiers, 1, 0, 0), REDUCE(aux_sym_attributed_field_declarator_repeat1, 1, 0, 0),
  [2992] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_modifiers, 1, 0, 0),
  [2994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2902),
  [2996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6557),
  [2998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4619),
  [3000] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6559),
  [3002] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5891),
  [3004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6269),
  [3006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4828),
  [3008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3294),
  [3010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2851),
  [3012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3875),
  [3014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6741),
  [3016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3271),
  [3018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6413),
  [3020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4896),
  [3022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6595),
  [3024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [3026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6517),
  [3028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5870),
  [3030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2432),
  [3032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [3034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(525),
  [3036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1035),
  [3038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(667),
  [3040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6488),
  [3042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6231),
  [3044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4616),
  [3046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4793),
  [3048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6564),
  [3050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(507),
  [3052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(508),
  [3054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1317),
  [3056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3752),
  [3058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5845),
  [3060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6721),
  [3062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1021),
  [3064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [3066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [3068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(666),
  [3070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(856),
  [3072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [3074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [3076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [3078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1042),
  [3080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1041),
  [3082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(683),
  [3084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [3086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [3088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [3090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1018),
  [3092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1019),
  [3094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [3096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(908),
  [3098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [3100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [3102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [3104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [3106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [3108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6712),
  [3110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(848),
  [3112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [3114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [3116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1026),
  [3118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [3120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5759),
  [3122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [3124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3060),
  [3126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6119),
  [3128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4601),
  [3130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2365),
  [3132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6120),
  [3134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1289),
  [3136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4596),
  [3138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5738),
  [3140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4823),
  [3142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2852),
  [3144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3930),
  [3146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6093),
  [3148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6013),
  [3150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5878),
  [3152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [3154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2239),
  [3156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [3158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif_in_implementation_definition, 4, 0, 47),
  [3160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2236),
  [3162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2389),
  [3164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1023),
  [3166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif_in_implementation_definition, 3, 0, 47),
  [3168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(673),
  [3170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2363),
  [3172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2391),
  [3174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [3176] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(3060),
  [3179] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6119),
  [3182] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4601),
  [3185] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0),
  [3187] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6120),
  [3190] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5738),
  [3193] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4823),
  [3196] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2852),
  [3199] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(3930),
  [3202] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [3205] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [3208] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [3211] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6741),
  [3214] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [3217] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2907),
  [3220] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2862),
  [3223] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [3226] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2973),
  [3229] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4876),
  [3232] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4896),
  [3235] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4932),
  [3238] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6087),
  [3241] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6093),
  [3244] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [3247] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [3250] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6013),
  [3253] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5878),
  [3256] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2909),
  [3259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 1, 0, 0),
  [3261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [3263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(500),
  [3265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(606),
  [3267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(676),
  [3269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [3271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [3273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(833),
  [3275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1054),
  [3277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [3279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6306),
  [3281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4598),
  [3283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2320),
  [3285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1383),
  [3287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4590),
  [3289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4790),
  [3291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3748),
  [3293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2404),
  [3295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [3297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [3299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2269),
  [3301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [3303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1049),
  [3305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [3307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2286),
  [3309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [3311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [3313] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif_in_interface_declaration, 4, 0, 47),
  [3315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(897),
  [3317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [3319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [3321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1024),
  [3323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [3325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(601),
  [3327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [3329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(613),
  [3331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(630),
  [3333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(981),
  [3335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [3337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2304),
  [3339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [3341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [3343] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif_in_interface_declaration, 3, 0, 47),
  [3345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2412),
  [3347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(835),
  [3349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(907),
  [3351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3064),
  [3353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6137),
  [3355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4575),
  [3357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else_in_implementation_definition, 2, 0, 0),
  [3359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6138),
  [3361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5853),
  [3363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4827),
  [3365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2853),
  [3367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3936),
  [3369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6122),
  [3371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6060),
  [3373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5728),
  [3375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [3377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(886),
  [3379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(887),
  [3381] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(3064),
  [3384] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6137),
  [3387] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4575),
  [3390] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6138),
  [3393] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5853),
  [3396] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4827),
  [3399] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2853),
  [3402] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(3936),
  [3405] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6122),
  [3408] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6060),
  [3411] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_implementation_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5728),
  [3414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1056),
  [3416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [3418] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(2902),
  [3421] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6557),
  [3424] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(4619),
  [3427] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6559),
  [3430] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(5891),
  [3433] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(4828),
  [3436] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(2851),
  [3439] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(3875),
  [3442] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [3445] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [3448] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [3451] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6741),
  [3454] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [3457] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(2907),
  [3460] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(2862),
  [3463] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [3466] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(2973),
  [3469] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(4876),
  [3472] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(4896),
  [3475] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(4932),
  [3478] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6087),
  [3481] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6595),
  [3484] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [3487] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [3490] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0),
  [3492] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6517),
  [3495] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(5870),
  [3498] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(2909),
  [3501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [3503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(607),
  [3505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [3507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [3509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(622),
  [3511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(623),
  [3513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [3515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(625),
  [3517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [3519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(636),
  [3521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(501),
  [3523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(505),
  [3525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [3527] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else_in_implementation_definition, 1, 0, 0),
  [3529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1055),
  [3531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [3533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [3535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [3537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1071),
  [3539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(993),
  [3541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [3543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [3545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(995),
  [3547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1057),
  [3549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [3551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [3553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [3555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [3557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [3559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(884),
  [3561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(837),
  [3563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [3565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [3567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [3569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1051),
  [3571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(883),
  [3573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_qualified_protocol_interface_declaration, 1, 0, 0),
  [3575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(602),
  [3577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(890),
  [3579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(603),
  [3581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(873),
  [3583] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2902),
  [3586] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6306),
  [3589] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4598),
  [3592] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0),
  [3594] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6120),
  [3597] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5738),
  [3600] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4790),
  [3603] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2852),
  [3606] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3930),
  [3609] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [3612] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [3615] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [3618] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6741),
  [3621] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [3624] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2907),
  [3627] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2862),
  [3630] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [3633] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2973),
  [3636] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4876),
  [3639] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4896),
  [3642] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4932),
  [3645] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6087),
  [3648] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6093),
  [3651] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [3654] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [3657] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3748),
  [3660] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2909),
  [3663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [3665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [3667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [3669] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_qualified_protocol_interface_declaration, 2, 0, 0),
  [3671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [3673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [3675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [3677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [3679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [3681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1050),
  [3683] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(2902),
  [3686] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(6488),
  [3689] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(4616),
  [3692] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(6559),
  [3695] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(5891),
  [3698] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(4793),
  [3701] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(2851),
  [3704] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(3875),
  [3707] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [3710] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [3713] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [3716] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(6741),
  [3719] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [3722] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(2907),
  [3725] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(2862),
  [3728] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [3731] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(2973),
  [3734] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(4876),
  [3737] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(4896),
  [3740] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(4932),
  [3743] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(6087),
  [3746] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(6595),
  [3749] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [3752] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [3755] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0),
  [3757] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(3752),
  [3760] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0), SHIFT_REPEAT(2909),
  [3763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1006),
  [3765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1007),
  [3767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1052),
  [3769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [3771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [3773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [3775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [3777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [3779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1053),
  [3781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [3783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [3785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [3787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [3789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [3791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6639),
  [3793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4573),
  [3795] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else_in_interface_declaration, 2, 0, 0),
  [3797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4791),
  [3799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3747),
  [3801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [3803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [3805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [3807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1015),
  [3809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [3811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(499),
  [3813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [3815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [3817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [3819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(458),
  [3821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [3823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [3825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [3827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [3829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [3831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [3833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [3835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [3837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(891),
  [3839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(892),
  [3841] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6639),
  [3844] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4573),
  [3847] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6138),
  [3850] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5853),
  [3853] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4791),
  [3856] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2853),
  [3859] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3936),
  [3862] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6122),
  [3865] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3747),
  [3868] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else_in_interface_declaration, 1, 0, 0),
  [3870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [3872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(875),
  [3874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(877),
  [3876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [3878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(881),
  [3880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [3882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(605),
  [3884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [3886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [3888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(617),
  [3890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [3892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(619),
  [3894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(620),
  [3896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(621),
  [3898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [3900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [3902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [3904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(904),
  [3906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [3908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [3910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(633),
  [3912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [3914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(635),
  [3916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [3918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [3920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(836),
  [3922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [3924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [3926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(988),
  [3928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [3930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [3932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(991),
  [3934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1005),
  [3936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [3938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [3940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [3942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1433),
  [3944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concatenated_string, 3, 0, 0),
  [3946] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_concatenated_string, 3, 0, 0),
  [3948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6094),
  [3950] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1433),
  [3953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0),
  [3955] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0),
  [3957] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(6094),
  [3960] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(5299),
  [3963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1432),
  [3965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concatenated_string, 2, 0, 0),
  [3967] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_concatenated_string, 2, 0, 0),
  [3969] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_selector_expression_repeat2, 2, 0, 0),
  [3971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_selector_expression_repeat2, 2, 0, 0),
  [3973] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selector_expression_repeat2, 2, 0, 0), SHIFT_REPEAT(1435),
  [3976] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_expression, 2, 0, 6),
  [3978] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_expression, 2, 0, 6),
  [3980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2076),
  [3982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3653),
  [3984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6280),
  [3986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6280),
  [3988] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cast_expression, 2, 0, 7),
  [3990] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cast_expression, 2, 0, 7),
  [3992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1472),
  [3994] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_selector_expression, 4, 0, 0),
  [3996] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selector_expression, 4, 0, 0),
  [3998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1435),
  [4000] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, 0, 34),
  [4002] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, 0, 34),
  [4004] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__string, 1, 0, 0),
  [4006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__string, 1, 0, 0),
  [4008] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_update_expression, 2, 0, 6),
  [4010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_update_expression, 2, 0, 6),
  [4012] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sizeof_expression, 2, 0, 7),
  [4014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sizeof_expression, 2, 0, 7),
  [4016] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_expression, 2, 0, 6),
  [4018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_expression, 2, 0, 6),
  [4020] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cast_expression, 4, 0, 52),
  [4022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cast_expression, 4, 0, 52),
  [4024] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression, 1, 0, 0),
  [4026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression, 1, 0, 0),
  [4028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3738),
  [4030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3738),
  [4032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6053),
  [4034] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 3, 0, 0),
  [4036] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 3, 0, 0),
  [4038] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_offsetof_expression, 6, 0, 130),
  [4040] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_offsetof_expression, 6, 0, 130),
  [4042] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression, 1, 0, 1),
  [4044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression, 1, 0, 1),
  [4046] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_expression, 8, 0, 0),
  [4048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_expression, 8, 0, 0),
  [4050] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 8, 0, 152),
  [4052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 8, 0, 152),
  [4054] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 6, 0, 132),
  [4056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 6, 0, 132),
  [4058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 6, 0, 133),
  [4060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 6, 0, 133),
  [4062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 8, 0, 153),
  [4064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 8, 0, 153),
  [4066] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_literal, 6, 0, 0),
  [4068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_literal, 6, 0, 0),
  [4070] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dictionary_literal, 6, 0, 0),
  [4072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary_literal, 6, 0, 0),
  [4074] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 2, 0, 0),
  [4076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 2, 0, 0),
  [4078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3011),
  [4080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5811),
  [4082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3775),
  [4084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1503),
  [4086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2219),
  [4088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4121),
  [4090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2528),
  [4092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5872),
  [4094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_literal, 3, 0, 0),
  [4096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_literal, 3, 0, 0),
  [4098] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dictionary_literal, 3, 0, 0),
  [4100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary_literal, 3, 0, 0),
  [4102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 4, 0, 0),
  [4104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 4, 0, 0),
  [4106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_expression, 9, 0, 0),
  [4108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_expression, 9, 0, 0),
  [4110] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 9, 0, 158),
  [4112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 9, 0, 158),
  [4114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 7, 0, 49),
  [4116] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 7, 0, 49),
  [4118] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 3, 0, 0),
  [4120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 3, 0, 0),
  [4122] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_char_literal, 3, 0, 0),
  [4124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_char_literal, 3, 0, 0),
  [4126] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_literal_expression, 4, 0, 52),
  [4128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_literal_expression, 4, 0, 52),
  [4130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_literal, 4, 0, 0),
  [4132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_literal, 4, 0, 0),
  [4134] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_update_expression, 2, 0, 11),
  [4136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_update_expression, 2, 0, 11),
  [4138] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_message_expression, 4, 0, 57),
  [4140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_message_expression, 4, 0, 57),
  [4142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_initializer_list, 3, 0, 0),
  [4144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_list, 3, 0, 0),
  [4146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript_expression, 4, 0, 78),
  [4148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript_expression, 4, 0, 78),
  [4150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_literal, 2, 0, 0),
  [4152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_literal, 2, 0, 0),
  [4154] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_string_literal, 3, 0, 0), REDUCE(sym_string_literal, 4, 0, 0),
  [4157] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_string_literal, 3, 0, 0), REDUCE(sym_string_literal, 4, 0, 0),
  [4160] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_null, 1, 0, 0),
  [4162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_null, 1, 0, 0),
  [4164] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alignof_expression, 4, 0, 43),
  [4166] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alignof_expression, 4, 0, 43),
  [4168] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 4, 0, 68),
  [4170] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 4, 0, 68),
  [4172] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_literal, 4, 0, 0),
  [4174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_literal, 4, 0, 0),
  [4176] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_literal, 6, 0, 0),
  [4178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_literal, 6, 0, 0),
  [4180] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 5, 0, 0),
  [4182] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 5, 0, 0),
  [4184] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_initializer_list, 4, 0, 0),
  [4186] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_list, 4, 0, 0),
  [4188] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_literal, 7, 0, 0),
  [4190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_literal, 7, 0, 0),
  [4192] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_available_expression, 4, 0, 0),
  [4194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_available_expression, 4, 0, 0),
  [4196] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_encode_expression, 4, 0, 0),
  [4198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_encode_expression, 4, 0, 0),
  [4200] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_expression, 3, 0, 40),
  [4202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 3, 0, 40),
  [4204] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dictionary_literal, 5, 0, 0),
  [4206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary_literal, 5, 0, 0),
  [4208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_available_expression, 6, 0, 0),
  [4210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_available_expression, 6, 0, 0),
  [4212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_available_expression, 5, 0, 0),
  [4214] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_available_expression, 5, 0, 0),
  [4216] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_va_arg_expression, 6, 0, 0),
  [4218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_va_arg_expression, 6, 0, 0),
  [4220] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 7, 0, 145),
  [4222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 7, 0, 145),
  [4224] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 7, 0, 146),
  [4226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 7, 0, 146),
  [4228] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 6, 0, 49),
  [4230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 6, 0, 49),
  [4232] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [4234] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [4236] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_literal, 3, 0, 0),
  [4238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_literal, 3, 0, 0),
  [4240] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_initializer_list, 2, 0, 0),
  [4242] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_list, 2, 0, 0),
  [4244] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_literal, 5, 0, 0),
  [4246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_literal, 5, 0, 0),
  [4248] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_initializer_list, 5, 0, 0),
  [4250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_list, 5, 0, 0),
  [4252] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_string_literal, 2, 0, 0), REDUCE(sym_string_literal, 3, 0, 0),
  [4255] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_string_literal, 2, 0, 0), REDUCE(sym_string_literal, 3, 0, 0),
  [4258] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 4, 0, 0),
  [4260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 4, 0, 0),
  [4262] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 2, 0, 0),
  [4264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 2, 0, 0),
  [4266] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 5, 0, 106),
  [4268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 5, 0, 106),
  [4270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call_expression, 2, 0, 12),
  [4272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call_expression, 2, 0, 12),
  [4274] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 5, 0, 107),
  [4276] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 5, 0, 107),
  [4278] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_literal, 5, 0, 0),
  [4280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_literal, 5, 0, 0),
  [4282] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dictionary_literal, 4, 0, 0),
  [4284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary_literal, 4, 0, 0),
  [4286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6088),
  [4288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4608),
  [4290] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif_in_field_declaration_list, 4, 0, 47),
  [4292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6363),
  [4294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1614),
  [4296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4586),
  [4298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6499),
  [4300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5895),
  [4302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4871),
  [4304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6234),
  [4306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2755),
  [4308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2650),
  [4310] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elifdef_in_field_declaration_list, 2, 0, 16),
  [4312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2603),
  [4314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2626),
  [4316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2709),
  [4318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2721),
  [4320] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif_in_field_declaration_list, 3, 0, 47),
  [4322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2752),
  [4324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elifdef_in_field_declaration_list, 3, 0, 16),
  [4326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2686),
  [4328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2763),
  [4330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2711),
  [4332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2689),
  [4334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2611),
  [4336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3329),
  [4338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1551),
  [4340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3678),
  [4342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3678),
  [4344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3679),
  [4346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2099),
  [4348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4165),
  [4350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3759),
  [4352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2085),
  [4354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4872),
  [4356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3681),
  [4358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3682),
  [4360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3420),
  [4362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5801),
  [4364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3477),
  [4366] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2902),
  [4369] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6088),
  [4372] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4608),
  [4375] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0),
  [4377] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6363),
  [4380] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5895),
  [4383] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [4386] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [4389] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [4392] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6741),
  [4395] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [4398] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2862),
  [4401] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [4404] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2973),
  [4407] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4871),
  [4410] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4896),
  [4413] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4932),
  [4416] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6087),
  [4419] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [4422] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [4425] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2909),
  [4428] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6234),
  [4431] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extension_expression, 2, 0, 0),
  [4433] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3475),
  [4436] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0),
  [4438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3476),
  [4440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3478),
  [4442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3480),
  [4444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3488),
  [4446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3489),
  [4448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3492),
  [4450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3494),
  [4452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3543),
  [4454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3543),
  [4456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3554),
  [4458] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_expression, 5, 0, 113),
  [4460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3475),
  [4462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 5, 0, 113),
  [4464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3380),
  [4466] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment_expression, 3, 0, 34),
  [4468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 3, 0, 34),
  [4470] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_expression, 3, 0, 0),
  [4472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_expression, 3, 0, 0),
  [4474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_at_expression, 2, 0, 0),
  [4476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_at_expression, 2, 0, 0),
  [4478] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_expression, 4, 0, 79),
  [4480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 4, 0, 79),
  [4482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2822),
  [4484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2582),
  [4486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4222),
  [4488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1576),
  [4490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2782),
  [4492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4142),
  [4494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 2),
  [4496] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2), REDUCE(sym_expression, 1, 0, 0), SHIFT(4081),
  [4500] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 2), REDUCE(sym_expression, 1, 0, 0),
  [4503] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_expression, 1, 0, 0), SHIFT(3311),
  [4506] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2), REDUCE(sym_expression, 1, 0, 0),
  [4509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2961),
  [4511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3570),
  [4513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [4515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3570),
  [4517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6603),
  [4519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [4521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4617),
  [4523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6381),
  [4525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2932),
  [4527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6032),
  [4529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [4531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [4533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3528),
  [4535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [4537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3541),
  [4539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2945),
  [4541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3841),
  [4543] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6137),
  [4546] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4578),
  [4549] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6679),
  [4552] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5853),
  [4555] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6418),
  [4558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1599),
  [4560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [4562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3575),
  [4564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [4566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [4568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3685),
  [4570] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1599),
  [4573] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(6053),
  [4576] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(5191),
  [4579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3658),
  [4581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1593),
  [4583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3864),
  [4585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3666),
  [4587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3670),
  [4589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3674),
  [4591] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6557),
  [4594] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4617),
  [4597] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6381),
  [4600] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5891),
  [4603] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0),
  [4605] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6032),
  [4608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4578),
  [4610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else_in_field_declaration_list, 2, 0, 0),
  [4612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6679),
  [4614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6418),
  [4616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else_in_field_declaration_list, 1, 0, 0),
  [4618] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(982),
  [4621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1631),
  [4623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1621),
  [4625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5932),
  [4627] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selector_expression_repeat2, 2, 0, 0), SHIFT_REPEAT(1620),
  [4630] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1621),
  [4633] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(5932),
  [4636] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(5239),
  [4639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1620),
  [4641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1618),
  [4643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2962),
  [4645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1997),
  [4647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5031),
  [4649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [4651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6172),
  [4653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3012),
  [4655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1984),
  [4657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5057),
  [4659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [4661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6171),
  [4663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1986),
  [4665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5075),
  [4667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1987),
  [4669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5087),
  [4671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1988),
  [4673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5011),
  [4675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1989),
  [4677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5017),
  [4679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2021),
  [4681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5009),
  [4683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [4685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6151),
  [4687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2036),
  [4689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5039),
  [4691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1999),
  [4693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5007),
  [4695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2026),
  [4697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5056),
  [4699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2001),
  [4701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5044),
  [4703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2054),
  [4705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5049),
  [4707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2003),
  [4709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5060),
  [4711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2004),
  [4713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5072),
  [4715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2005),
  [4717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5016),
  [4719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2006),
  [4721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5004),
  [4723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2047),
  [4725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5062),
  [4727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2022),
  [4729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5083),
  [4731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2018),
  [4733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5054),
  [4735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2046),
  [4737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5081),
  [4739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2025),
  [4741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5018),
  [4743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1627),
  [4745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [4747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [4749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [4751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3014),
  [4753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2035),
  [4755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1985),
  [4757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1998),
  [4759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2023),
  [4761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2049),
  [4763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2002),
  [4765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [4767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2000),
  [4769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2032),
  [4771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [4773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2055),
  [4775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3587),
  [4777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3587),
  [4779] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_init_declarator, 4, 0, 110),
  [4781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3741),
  [4783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_init_declarator, 4, 0, 110),
  [4785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3743),
  [4787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3744),
  [4789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3744),
  [4791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3449),
  [4793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3465),
  [4795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3496),
  [4797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3571),
  [4799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3454),
  [4801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3455),
  [4803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3473),
  [4805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3473),
  [4807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3553),
  [4809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3352),
  [4811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5391),
  [4813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5015),
  [4815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5445),
  [4817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5034),
  [4819] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_init_declarator, 3, 0, 72),
  [4821] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_init_declarator, 3, 0, 72),
  [4823] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3741),
  [4826] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_reference_list, 3, 0, 0),
  [4828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_reference_list, 3, 0, 0),
  [4830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5401),
  [4832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5084),
  [4834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5370),
  [4836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5614),
  [4838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5071),
  [4840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5010),
  [4842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5373),
  [4844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5409),
  [4846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5025),
  [4848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5046),
  [4850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5456),
  [4852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5027),
  [4854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5710),
  [4856] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_reference_list, 4, 0, 0),
  [4858] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_reference_list, 4, 0, 0),
  [4860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5055),
  [4862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2080),
  [4864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5427),
  [4866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5379),
  [4868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5509),
  [4870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5625),
  [4872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5337),
  [4874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5367),
  [4876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5526),
  [4878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5418),
  [4880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5549),
  [4882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1046),
  [4884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5556),
  [4886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5562),
  [4888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5565),
  [4890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5431),
  [4892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5437),
  [4894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5440),
  [4896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2970),
  [4898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2674),
  [4900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2781),
  [4902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2667),
  [4904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3048),
  [4906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3641),
  [4908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3641),
  [4910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5442),
  [4912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5389),
  [4914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5689),
  [4916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5430),
  [4918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5362),
  [4920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5568),
  [4922] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 4, 0, 49),
  [4924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 4, 0, 49),
  [4926] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_type_specifier, 4, -1, 69),
  [4928] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_type_specifier, 4, -1, 69),
  [4930] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 8, 0, 86),
  [4932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 8, 0, 86),
  [4934] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 4, 0, 86),
  [4936] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 4, 0, 86),
  [4938] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 3, 0, 0),
  [4940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 3, 0, 0),
  [4942] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 7, 0, 49),
  [4944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 7, 0, 49),
  [4946] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 7, 0, 86),
  [4948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 7, 0, 86),
  [4950] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2902),
  [4953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0),
  [4955] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [4958] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [4961] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [4964] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6741),
  [4967] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [4970] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2862),
  [4973] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [4976] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2973),
  [4979] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4876),
  [4982] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4896),
  [4985] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4932),
  [4988] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6087),
  [4991] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [4994] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [4997] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2909),
  [5000] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 6, 0, 49),
  [5002] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 6, 0, 49),
  [5004] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 6, 0, 86),
  [5006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 6, 0, 86),
  [5008] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 4, 0, 0),
  [5010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 4, 0, 0),
  [5012] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 5, 0, 49),
  [5014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 5, 0, 49),
  [5016] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameterized_arguments, 5, 0, 86),
  [5018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameterized_arguments, 5, 0, 86),
  [5020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3610),
  [5022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3611),
  [5024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3612),
  [5026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3613),
  [5028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3614),
  [5030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3615),
  [5032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3616),
  [5034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3617),
  [5036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3618),
  [5038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3619),
  [5040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3619),
  [5042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3746),
  [5044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3374),
  [5046] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3610),
  [5049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1915),
  [5051] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1915),
  [5054] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(6603),
  [5057] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(5263),
  [5060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1914),
  [5062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5443),
  [5064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5394),
  [5066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3010),
  [5068] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_implementation_definition, 1, 0, 0),
  [5070] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 0),
  [5072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_implementation_definition, 1, 0, 0),
  [5074] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 0),
  [5076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5573),
  [5078] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_params, 1, 0, 0),
  [5080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4276),
  [5082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1459),
  [5084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3323),
  [5086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4127),
  [5088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5665),
  [5090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3845),
  [5092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4149),
  [5094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3331),
  [5096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6447),
  [5098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6287),
  [5100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5944),
  [5102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1857),
  [5104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1757),
  [5106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1828),
  [5108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1833),
  [5110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1837),
  [5112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1841),
  [5114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3199),
  [5116] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_arguments, 4, 0, 49),
  [5118] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_arguments, 4, 0, 49),
  [5120] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0),
  [5122] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0),
  [5124] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [5127] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [5130] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [5133] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(6741),
  [5136] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [5139] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [5142] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [5145] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [5148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3190),
  [5150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1874),
  [5152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1729),
  [5154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1829),
  [5156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1737),
  [5158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1832),
  [5160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1754),
  [5162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1835),
  [5164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1826),
  [5166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1842),
  [5168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1844),
  [5170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3180),
  [5172] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_implementation_header, 2, 0, 0),
  [5174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_implementation_header, 2, 0, 0),
  [5176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2077),
  [5178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3175),
  [5180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3125),
  [5182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1859),
  [5184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6321),
  [5186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6319),
  [5188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6415),
  [5190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4126),
  [5192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5405),
  [5194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1852),
  [5196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1866),
  [5198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1739),
  [5200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1870),
  [5202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1861),
  [5204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3176),
  [5206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3140),
  [5208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1753),
  [5210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_implementation_header, 3, 0, 0),
  [5212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_implementation_header, 3, 0, 0),
  [5214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2079),
  [5216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1731),
  [5218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1847),
  [5220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3188),
  [5222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6664),
  [5224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4122),
  [5226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5448),
  [5228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5951),
  [5230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4124),
  [5232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5470),
  [5234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6015),
  [5236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4128),
  [5238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5488),
  [5240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4125),
  [5242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5499),
  [5244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4123),
  [5246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5506),
  [5248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1868),
  [5250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1865),
  [5252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1740),
  [5254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6312),
  [5256] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_params, 2, 0, 0),
  [5258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 2, 0, 0),
  [5260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3201),
  [5262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1854),
  [5264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1761),
  [5266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3718),
  [5268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3719),
  [5270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3720),
  [5272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3721),
  [5274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3722),
  [5276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3723),
  [5278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3724),
  [5280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3725),
  [5282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3726),
  [5284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3727),
  [5286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3727),
  [5288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3728),
  [5290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3378),
  [5292] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_interface_header, 3, 0, 0),
  [5294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_interface_header, 3, 0, 0),
  [5296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2162),
  [5298] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_interface_header, 2, 0, 0),
  [5300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_interface_header, 2, 0, 0),
  [5302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2132),
  [5304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2795),
  [5306] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat2, 1, 0, 0),
  [5308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3295),
  [5310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat2, 1, 0, 0),
  [5312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3980),
  [5314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_pointer_declarator, 1, 1, 0),
  [5316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2067),
  [5318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4116),
  [5320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3000),
  [5322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3001),
  [5324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1916),
  [5326] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_implementation_inheritance, 3, 0, 80),
  [5328] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_arguments, 3, 0, 49),
  [5330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_implementation_inheritance, 3, 0, 80),
  [5332] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_implementation_header, 4, 0, 0),
  [5334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_implementation_header, 4, 0, 0),
  [5336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4288),
  [5338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3555),
  [5340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3556),
  [5342] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3718),
  [5345] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 8, 0, 0),
  [5347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 8, 0, 0),
  [5349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2185),
  [5351] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 7, 0, 0),
  [5353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 0),
  [5355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2222),
  [5357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2138),
  [5359] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 8, 0, 0),
  [5361] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 0),
  [5363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2227),
  [5365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 9, 0, 0),
  [5367] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 9, 0, 0),
  [5369] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 9, 0, 0),
  [5371] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 9, 0, 0),
  [5373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2220),
  [5375] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 10, 0, 0),
  [5377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 10, 0, 0),
  [5379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2229),
  [5381] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 11, 0, 0),
  [5383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 11, 0, 0),
  [5385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2205),
  [5387] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 12, 0, 0),
  [5389] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 12, 0, 0),
  [5391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2225),
  [5393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 13, 0, 0),
  [5395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 13, 0, 0),
  [5397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2228),
  [5399] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 14, 0, 0),
  [5401] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 14, 0, 0),
  [5403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2240),
  [5405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 15, 0, 0),
  [5407] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 15, 0, 0),
  [5409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2203),
  [5411] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 7, 0, 0),
  [5413] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 7, 0, 0),
  [5415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3065),
  [5417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1460),
  [5419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2126),
  [5421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4119),
  [5423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 4, 0, 0),
  [5425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 4, 0, 0),
  [5427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2135),
  [5429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4111),
  [5431] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat2, 2, 0, 0),
  [5433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat2, 2, 0, 0),
  [5435] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat2, 2, 0, 0), SHIFT_REPEAT(2138),
  [5438] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 5, 0, 0),
  [5440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 0),
  [5442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2226),
  [5444] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_interface_header, 4, 0, 0),
  [5446] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_interface_header, 4, 0, 0),
  [5448] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 4, 0, 0),
  [5450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 0),
  [5452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2233),
  [5454] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 5, 0, 0),
  [5456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 5, 0, 0),
  [5458] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 3, 0, 0),
  [5460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 3, 0, 0),
  [5462] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 6, 0, 0),
  [5464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 6, 0, 0),
  [5466] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 3, 0, 0),
  [5468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 0),
  [5470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2221),
  [5472] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat2, 2, 0, 0), SHIFT_REPEAT(2185),
  [5475] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 6, 0, 0),
  [5477] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 0),
  [5479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2241),
  [5481] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_interface_inheritance, 3, 0, 80),
  [5483] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_interface_inheritance, 3, 0, 80),
  [5485] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 16, 0, 0),
  [5487] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 16, 0, 0),
  [5489] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instance_variables, 3, 0, 0),
  [5491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variables, 3, 0, 0),
  [5493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2333),
  [5495] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instance_variables, 2, 0, 0),
  [5497] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variables, 2, 0, 0),
  [5499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2307),
  [5501] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_interface_inheritance, 2, 0, 41),
  [5503] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_interface_inheritance, 2, 0, 41),
  [5505] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_implementation_definition, 2, 0, 0),
  [5507] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_implementation_definition, 2, 0, 0),
  [5509] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_implementation_definition, 6, 0, 115),
  [5511] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_implementation_definition, 6, 0, 115),
  [5513] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_implementation_definition, 5, 0, 84),
  [5515] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_implementation_definition, 5, 0, 84),
  [5517] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_implementation_definition, 5, 0, 47),
  [5519] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_implementation_definition, 5, 0, 47),
  [5521] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_implementation_definition, 4, 0, 47),
  [5523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_implementation_definition, 4, 0, 47),
  [5525] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_interface_inheritance, 3, 0, 41),
  [5527] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_interface_inheritance, 3, 0, 41),
  [5529] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_declaration, 3, 0, 0),
  [5531] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 3, 0, 0),
  [5533] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_interface_inheritance, 2, 0, 0),
  [5535] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_interface_inheritance, 2, 0, 0),
  [5537] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_declaration, 5, 0, 0),
  [5539] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 5, 0, 0),
  [5541] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_declaration, 2, 0, 0),
  [5543] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 2, 0, 0),
  [5545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2361),
  [5547] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_interface_repeat1, 2, 0, 0),
  [5549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2376),
  [5551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2347),
  [5553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2388),
  [5555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2390),
  [5557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2366),
  [5559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2394),
  [5561] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_interface_declaration, 4, 0, 47),
  [5563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_interface_declaration, 4, 0, 47),
  [5565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2321),
  [5567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2374),
  [5569] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_atomic_declaration, 6, 0, 154),
  [5571] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_atomic_declaration, 6, 0, 154),
  [5573] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_declaration, 5, 0, 0),
  [5575] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_declaration, 5, 0, 0),
  [5577] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_declaration, 3, 0, 0),
  [5579] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_declaration, 3, 0, 0),
  [5581] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_declaration, 4, 0, 0),
  [5583] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 4, 0, 0),
  [5585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2352),
  [5587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2392),
  [5589] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_interface_declaration, 5, 0, 84),
  [5591] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_interface_declaration, 5, 0, 84),
  [5593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2359),
  [5595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_interface_declaration, 5, 0, 47),
  [5597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_interface_declaration, 5, 0, 47),
  [5599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2367),
  [5601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2350),
  [5603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2396),
  [5605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2370),
  [5607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2385),
  [5609] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_interface_declaration, 6, 0, 115),
  [5611] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_interface_declaration, 6, 0, 115),
  [5613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2377),
  [5615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2360),
  [5617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2378),
  [5619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2384),
  [5621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_interface_declaration_repeat1, 2, 0, 0),
  [5623] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat2, 2, 0, 0), SHIFT_REPEAT(2321),
  [5626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2387),
  [5628] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_declaration, 4, 0, 0),
  [5630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_declaration, 4, 0, 0),
  [5632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2346),
  [5634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instance_variables, 4, 0, 0),
  [5636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variables, 4, 0, 0),
  [5638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2393),
  [5640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_implementation_inheritance, 2, 0, 41),
  [5642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_implementation_inheritance, 2, 0, 41),
  [5644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2395),
  [5646] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ms_declspec_modifier, 4, 0, 0),
  [5648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_declspec_modifier, 4, 0, 0),
  [5650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2399),
  [5652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2369),
  [5654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_availability_attribute_specifier, 8, 0, 0),
  [5656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_availability_attribute_specifier, 8, 0, 0),
  [5658] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_availability_attribute_specifier, 6, 0, 0),
  [5660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_availability_attribute_specifier, 6, 0, 0),
  [5662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_availability_attribute_specifier, 5, 0, 0),
  [5664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_availability_attribute_specifier, 5, 0, 0),
  [5666] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alignas_qualifier, 4, 0, 0),
  [5668] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alignas_qualifier, 4, 0, 0),
  [5670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_availability_attribute_specifier, 4, 0, 0),
  [5672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_availability_attribute_specifier, 4, 0, 0),
  [5674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_availability_attribute_specifier, 7, 0, 0),
  [5676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_availability_attribute_specifier, 7, 0, 0),
  [5678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_availability_attribute_specifier, 1, 0, 0),
  [5680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_availability_attribute_specifier, 1, 0, 0),
  [5682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_storage_class_specifier, 1, 0, 0),
  [5684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_storage_class_specifier, 1, 0, 0),
  [5686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3261),
  [5688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2668),
  [5690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2414),
  [5692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4129),
  [5694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3253),
  [5696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3256),
  [5698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2410),
  [5700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4130),
  [5702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4283),
  [5704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3437),
  [5706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4280),
  [5708] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4280),
  [5711] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(1531),
  [5714] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3472),
  [5717] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3472),
  [5720] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3557),
  [5723] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3283),
  [5726] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(383),
  [5729] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3686),
  [5732] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3437),
  [5735] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(72),
  [5738] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6090),
  [5741] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(1629),
  [5744] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3451),
  [5747] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5803),
  [5750] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5868),
  [5753] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3549),
  [5756] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6364),
  [5759] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5808),
  [5762] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(2425),
  [5765] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6733),
  [5768] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3338),
  [5771] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6366),
  [5774] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6370),
  [5777] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5995),
  [5780] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3742),
  [5783] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3458),
  [5786] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6690),
  [5789] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6233),
  [5792] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6242),
  [5795] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4952),
  [5798] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4924),
  [5801] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(1641),
  [5804] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5610),
  [5807] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3426),
  [5810] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5263),
  [5813] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(1641),
  [5816] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(1628),
  [5819] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5478),
  [5822] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5478),
  [5825] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3407),
  [5828] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5831),
  [5831] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6277),
  [5834] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6277),
  [5837] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6299),
  [5840] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6308),
  [5843] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6754),
  [5846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4279),
  [5848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4275),
  [5850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4277),
  [5852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4278),
  [5854] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4283),
  [5857] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(708),
  [5860] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(51),
  [5863] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6486),
  [5866] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5844),
  [5869] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5819),
  [5872] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3537),
  [5875] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6063),
  [5878] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5875),
  [5881] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(2467),
  [5884] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6377),
  [5887] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3370),
  [5890] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6609),
  [5893] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6227),
  [5896] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6024),
  [5899] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4893),
  [5902] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5530),
  [5905] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5530),
  [5908] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3430),
  [5911] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6587),
  [5914] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4278),
  [5917] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(709),
  [5920] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(68),
  [5923] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6046),
  [5926] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5886),
  [5929] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5805),
  [5932] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3709),
  [5935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6715),
  [5938] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5896),
  [5941] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(2469),
  [5944] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6668),
  [5947] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3371),
  [5950] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6713),
  [5953] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6716),
  [5956] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5925),
  [5959] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4922),
  [5962] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5463),
  [5965] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5463),
  [5968] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3399),
  [5971] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6746),
  [5974] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4279),
  [5977] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(543),
  [5980] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(76),
  [5983] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6121),
  [5986] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5866),
  [5989] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5720),
  [5992] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3551),
  [5995] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6749),
  [5998] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5869),
  [6001] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(2474),
  [6004] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6743),
  [6007] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3344),
  [6010] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6755),
  [6013] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6757),
  [6016] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6051),
  [6019] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4926),
  [6022] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5494),
  [6025] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5494),
  [6028] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3415),
  [6031] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6761),
  [6034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4281),
  [6036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4272),
  [6038] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4281),
  [6041] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(962),
  [6044] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5816),
  [6047] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5721),
  [6050] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6100),
  [6053] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4898),
  [6056] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5390),
  [6059] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5390),
  [6062] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3423),
  [6065] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6678),
  [6068] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4272),
  [6071] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(230),
  [6074] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(56),
  [6077] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5920),
  [6080] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5770),
  [6083] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5893),
  [6086] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3545),
  [6089] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6700),
  [6092] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5782),
  [6095] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(2458),
  [6098] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6576),
  [6101] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3384),
  [6104] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5924),
  [6107] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6132),
  [6110] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6591),
  [6113] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4915),
  [6116] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5436),
  [6119] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5436),
  [6122] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3424),
  [6125] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6736),
  [6128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declarator, 1, 0, 0),
  [6130] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__declarator, 1, 0, 0), REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(4093),
  [6134] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2),
  [6136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3318),
  [6138] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__declarator, 1, 0, 0), REDUCE(sym_type_specifier, 1, 0, 2),
  [6141] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declarator, 1, 0, 0),
  [6143] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__declarator, 1, 0, 0), REDUCE(sym_type_specifier, 1, 0, 2),
  [6146] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(3812),
  [6149] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4277),
  [6152] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(1154),
  [6155] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(81),
  [6158] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6139),
  [6161] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5716),
  [6164] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5751),
  [6167] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5722),
  [6170] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(2476),
  [6173] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6751),
  [6176] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3353),
  [6179] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6161),
  [6182] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6162),
  [6185] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6092),
  [6188] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4930),
  [6191] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5505),
  [6194] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5505),
  [6197] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3417),
  [6200] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6764),
  [6203] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(4275),
  [6206] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(47),
  [6209] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6574),
  [6212] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5840),
  [6215] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5775),
  [6218] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3524),
  [6221] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5969),
  [6224] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5863),
  [6227] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(2452),
  [6230] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6179),
  [6233] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(3357),
  [6236] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5989),
  [6239] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5993),
  [6242] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6210),
  [6245] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5381),
  [6248] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(5381),
  [6251] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT(6645),
  [6254] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_parameter, 2, 0, 81),
  [6256] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_parameter, 2, 0, 81),
  [6258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_parameter, 3, 0, 114),
  [6260] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_parameter, 3, 0, 114),
  [6262] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_parameter, 5, 0, 135),
  [6264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_parameter, 5, 0, 135),
  [6266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3588),
  [6268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3589),
  [6270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3590),
  [6272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3590),
  [6274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3591),
  [6276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3592),
  [6278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3593),
  [6280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3594),
  [6282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3595),
  [6284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3596),
  [6286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3597),
  [6288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3597),
  [6290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3598),
  [6292] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0),
  [6294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0),
  [6296] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [6299] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [6302] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_parameter, 3, 0, 0),
  [6304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_parameter, 3, 0, 0),
  [6306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3377),
  [6308] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typedefed_specifier, 1, 0, 0),
  [6310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typedefed_specifier, 1, 0, 0),
  [6312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6564),
  [6314] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator, 3, 0, 93),
  [6316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator, 3, 0, 93),
  [6318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_parameter, 5, 0, 0),
  [6320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_parameter, 5, 0, 0),
  [6322] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_parameter, 4, 0, 114),
  [6324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_parameter, 4, 0, 114),
  [6326] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_parameter, 4, 0, 0),
  [6328] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_parameter, 4, 0, 0),
  [6330] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3588),
  [6333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_parameter, 4, 0, 135),
  [6335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_parameter, 4, 0, 135),
  [6337] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_parameter, 3, 0, 81),
  [6339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_parameter, 3, 0, 81),
  [6341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3412),
  [6343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2659),
  [6345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3672),
  [6347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4313),
  [6349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2527),
  [6351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3207),
  [6353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2658),
  [6355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6376),
  [6357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_parameter, 2, 0, 0),
  [6359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_parameter, 2, 0, 0),
  [6361] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_pointer_declarator, 2, 1, 0),
  [6363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4946),
  [6365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2803),
  [6367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2519),
  [6369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4193),
  [6371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3435),
  [6373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2779),
  [6375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2529),
  [6377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3406),
  [6379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3986),
  [6381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2531),
  [6383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3442),
  [6385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4904),
  [6387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3325),
  [6389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3418),
  [6391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3392),
  [6393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3264),
  [6395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2544),
  [6397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3427),
  [6399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4838),
  [6401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2540),
  [6403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4172),
  [6405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3429),
  [6407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4928),
  [6409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2518),
  [6411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3395),
  [6413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4842),
  [6415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3389),
  [6417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4836),
  [6419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3441),
  [6421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2780),
  [6423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3419),
  [6425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3056),
  [6427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2538),
  [6429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3414),
  [6431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3992),
  [6433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3408),
  [6435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4847),
  [6437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2522),
  [6439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3404),
  [6441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3071),
  [6443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3498),
  [6445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3498),
  [6447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3638),
  [6449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3656),
  [6451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3504),
  [6453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3400),
  [6455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3409),
  [6457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3251),
  [6459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2545),
  [6461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4226),
  [6463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2778),
  [6465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2830),
  [6467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2586),
  [6469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4219),
  [6471] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declarator, 2, 0, 22),
  [6473] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declarator, 2, 0, 22),
  [6475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2842),
  [6477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2555),
  [6479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4204),
  [6481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2974),
  [6483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3213),
  [6485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3884),
  [6487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2565),
  [6489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4217),
  [6491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3642),
  [6493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3643),
  [6495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3644),
  [6497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3644),
  [6499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3645),
  [6501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3646),
  [6503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3647),
  [6505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3648),
  [6507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3649),
  [6509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3650),
  [6511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3651),
  [6513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3651),
  [6515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3652),
  [6517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3369),
  [6519] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 6, 0, 38),
  [6521] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 6, 0, 38),
  [6523] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 6, 0, 141),
  [6525] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 6, 0, 141),
  [6527] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 6, 0, 74),
  [6529] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 6, 0, 74),
  [6531] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 7, 0, 141),
  [6533] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 7, 0, 141),
  [6535] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_atdef_field, 4, 0, 0),
  [6537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_atdef_field, 4, 0, 0),
  [6539] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 4, 0, 38),
  [6541] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 4, 0, 38),
  [6543] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 4, 0, 74),
  [6545] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 4, 0, 74),
  [6547] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_declarator, 3, 0, 0),
  [6549] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_declarator, 3, 0, 0),
  [6551] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 4, 0, 20),
  [6553] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 4, 0, 20),
  [6555] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_field_declaration_list, 4, 0, 47),
  [6557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_field_declaration_list, 4, 0, 47),
  [6559] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_field_declaration_list, 5, 0, 84),
  [6561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_field_declaration_list, 5, 0, 84),
  [6563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_field_declaration_list, 5, 0, 47),
  [6565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_field_declaration_list, 5, 0, 47),
  [6567] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 5, 0, 85),
  [6569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 5, 0, 85),
  [6571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 5, 0, 38),
  [6573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 5, 0, 38),
  [6575] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 5, 0, 141),
  [6577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 5, 0, 141),
  [6579] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_declarator, 2, 0, 0),
  [6581] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_declarator, 2, 0, 0),
  [6583] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 5, 0, 74),
  [6585] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 5, 0, 74),
  [6587] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 4, 0, 48),
  [6589] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 4, 0, 48),
  [6591] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 4, 0, 16),
  [6593] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 4, 0, 16),
  [6595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 2, 0, 20),
  [6597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 2, 0, 20),
  [6599] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_field_declaration_list, 6, 0, 115),
  [6601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_field_declaration_list, 6, 0, 115),
  [6603] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 3, 0, 20),
  [6605] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 20),
  [6607] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3642),
  [6610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 3, 0, 16),
  [6612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 3, 0, 16),
  [6614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 3, 0, 38),
  [6616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 38),
  [6618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6224),
  [6620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5241),
  [6622] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 4, 0, 0),
  [6624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 4, 0, 0),
  [6626] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_c_method_parameter, 2, 0, 38),
  [6628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_c_method_parameter, 2, 0, 38),
  [6630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1822),
  [6632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2516),
  [6634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3346),
  [6636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3529),
  [6638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3529),
  [6640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3502),
  [6642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3503),
  [6644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3548),
  [6646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3447),
  [6648] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 2, 0, 0),
  [6650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 2, 0, 0),
  [6652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_parameter_list, 3, 0, 0),
  [6654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 3, 0, 0),
  [6656] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_parameter_list, 3, 0, 0), REDUCE(sym__old_style_parameter_list, 3, 0, 0),
  [6659] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_parameter_list, 3, 0, 0), REDUCE(sym__old_style_parameter_list, 3, 0, 0),
  [6662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 3, 0, 0),
  [6664] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_parameter_list, 2, 0, 0),
  [6666] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_parameter_list, 2, 0, 0), REDUCE(sym__old_style_parameter_list, 2, 0, 0),
  [6669] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_parameter_list, 2, 0, 0), REDUCE(sym__old_style_parameter_list, 2, 0, 0),
  [6672] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_method_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5637),
  [6675] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_definition_repeat1, 2, 0, 0),
  [6677] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5363),
  [6680] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_definition_repeat1, 2, 0, 0),
  [6682] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(3012),
  [6685] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declarator, 3, 0, 22),
  [6687] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declarator, 3, 0, 22),
  [6689] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0),
  [6691] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0),
  [6693] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [6696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declarator, 4, 0, 22),
  [6698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declarator, 4, 0, 22),
  [6700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3706),
  [6702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3706),
  [6704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3687),
  [6706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3688),
  [6708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3708),
  [6710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3440),
  [6712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5760),
  [6714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_c_method_parameter_repeat1, 2, 0, 21),
  [6716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_c_method_parameter_repeat1, 2, 0, 21),
  [6718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_declarator, 4, 1, 89),
  [6720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_declarator, 4, 1, 89),
  [6722] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_declarator, 5, 1, 118),
  [6724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_declarator, 5, 1, 118),
  [6726] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keyword_selector_repeat1, 2, 0, 0), SHIFT_REPEAT(6739),
  [6729] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_keyword_selector_repeat1, 2, 0, 0),
  [6731] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_keyword_selector_repeat1, 2, 0, 0), SHIFT_REPEAT(5363),
  [6734] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_keyword_selector_repeat1, 2, 0, 0),
  [6736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_declarator, 2, 1, 21),
  [6738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_declarator, 2, 1, 21),
  [6740] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_pointer_declarator, 2, 1, 21),
  [6742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_pointer_declarator, 2, 1, 21),
  [6744] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_declarator, 3, 1, 54),
  [6746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_declarator, 3, 1, 54),
  [6748] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_pointer_declarator, 3, 1, 54),
  [6750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_pointer_declarator, 3, 1, 54),
  [6752] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_declarator, 4, -10, 0),
  [6754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_declarator, 4, -10, 0),
  [6756] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_declarator, 3, 0, 24),
  [6758] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_declarator, 3, 0, 24),
  [6760] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_declarator, 4, 0, 24),
  [6762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_declarator, 4, 0, 24),
  [6764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2354),
  [6766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2381),
  [6768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6441),
  [6770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2537),
  [6772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6702),
  [6774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6451),
  [6776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2372),
  [6778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6623),
  [6780] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_declarator, 4, 0, 90),
  [6782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_declarator, 4, 0, 90),
  [6784] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_declarator, 3, -10, 0),
  [6786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_declarator, 3, -10, 0),
  [6788] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_declarator, 5, 0, 119),
  [6790] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_declarator, 5, 0, 119),
  [6792] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_c_method_parameter, 3, 0, 74),
  [6794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_c_method_parameter, 3, 0, 74),
  [6796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4215),
  [6798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1820),
  [6800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(686),
  [6802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2543),
  [6804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3382),
  [6806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4952),
  [6808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [6810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keyword_declarator, 2, 0, 0),
  [6812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyword_declarator, 2, 0, 0),
  [6814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(967),
  [6816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [6818] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keyword_declarator, 4, 0, 0),
  [6820] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyword_declarator, 4, 0, 0),
  [6822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(541),
  [6824] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keyword_declarator, 3, 0, 0),
  [6826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyword_declarator, 3, 0, 0),
  [6828] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_definition_repeat1, 3, 0, 0),
  [6830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_definition_repeat1, 3, 0, 0),
  [6832] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_definition_repeat2, 2, 0, 0),
  [6834] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_definition_repeat2, 2, 0, 0), SHIFT_REPEAT(2080),
  [6837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_definition_repeat2, 2, 0, 0),
  [6839] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_definition_repeat1, 4, 0, 0),
  [6841] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_definition_repeat1, 4, 0, 0),
  [6843] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_c_method_parameter_repeat1, 2, 0, 75),
  [6845] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_c_method_parameter_repeat1, 2, 0, 75), SHIFT_REPEAT(4805),
  [6848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_c_method_parameter_repeat1, 2, 0, 75),
  [6850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2823),
  [6852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4160),
  [6854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1824),
  [6856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2264),
  [6858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1922),
  [6860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2158),
  [6862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3880),
  [6864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3931),
  [6866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3937),
  [6868] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_specifiers, 3, 0, 13),
  [6870] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_specifiers, 3, 0, 13),
  [6872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4175),
  [6874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3578),
  [6876] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(2993),
  [6879] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, 0, 0),
  [6881] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(2354),
  [6884] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0),
  [6886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2995),
  [6888] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(6451),
  [6891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2992),
  [6893] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_specifiers, 2, 0, 13),
  [6895] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_specifiers, 2, 0, 13),
  [6897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2569),
  [6899] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_specifiers, 1, 0, 4),
  [6901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_specifiers, 1, 0, 4),
  [6903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3542),
  [6905] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(2969),
  [6908] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 1, 0, 0),
  [6910] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(2354),
  [6913] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0),
  [6915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2914),
  [6917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(6451),
  [6920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2938),
  [6922] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [6925] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(2381),
  [6928] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [6931] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(6209),
  [6934] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(6702),
  [6937] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [6940] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(2372),
  [6943] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(6623),
  [6946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3660),
  [6948] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_specifiers, 2, 0, 4),
  [6950] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_specifiers, 2, 0, 4),
  [6952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3530),
  [6954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3667),
  [6956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3671),
  [6958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3675),
  [6960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3624),
  [6962] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 2, 0, 8),
  [6964] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 2, 0, 8),
  [6966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [6968] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 63),
  [6970] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 63),
  [6972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3229),
  [6974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2037),
  [6976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2066),
  [6978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6209),
  [6980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3008),
  [6982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2053),
  [6984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2027),
  [6986] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 4, 0, 63),
  [6988] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 4, 0, 63),
  [6990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [6992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2007),
  [6994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2070),
  [6996] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_pointer_declarator_repeat1, 2, 0, 0),
  [6998] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pointer_declarator_repeat1, 2, 0, 0),
  [7000] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pointer_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(3000),
  [7003] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pointer_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(3001),
  [7006] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 2, 0, 8),
  [7008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 2, 0, 8),
  [7010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2014),
  [7012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2206),
  [7014] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 1, 0, 0),
  [7016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 1, 0, 0),
  [7018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3751),
  [7020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1991),
  [7022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2013),
  [7024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [7026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3749),
  [7028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1996),
  [7030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2031),
  [7032] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 3, 0, 29),
  [7034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 3, 0, 29),
  [7036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 3, 0, 29),
  [7038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 3, 0, 29),
  [7040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3750),
  [7042] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(4081),
  [7045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3311),
  [7047] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__old_style_parameter_list, 3, 0, 0),
  [7049] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_parameter_list, 4, 0, 0),
  [7051] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__old_style_parameter_list, 4, 0, 0),
  [7053] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_generic_specifier_repeat1, 2, 0, 0),
  [7055] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_specifier_repeat1, 2, 0, 0),
  [7057] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generic_specifier_repeat1, 2, 0, 0), SHIFT_REPEAT(3311),
  [7060] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_function_declarator, 2, 0, 22),
  [7062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__old_style_function_declarator, 2, 0, 22),
  [7064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ms_call_modifier, 1, 0, 0),
  [7066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_call_modifier, 1, 0, 0),
  [7068] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_specifier, 2, 0, 2),
  [7070] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_specifier, 2, 0, 2),
  [7072] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 5, 0, 99),
  [7074] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 5, 0, 99),
  [7076] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 2, 0, 9),
  [7078] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 2, 0, 9),
  [7080] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0),
  [7082] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0),
  [7084] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0), SHIFT_REPEAT(2912),
  [7087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3232),
  [7089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5026),
  [7091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6406),
  [7093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3853),
  [7095] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0),
  [7098] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0),
  [7101] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_generic_specifier_repeat1, 4, 0, 0),
  [7103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_specifier_repeat1, 4, 0, 0),
  [7105] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 3, 0, 28),
  [7107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 3, 0, 28),
  [7109] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 2, 0, 9),
  [7111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 2, 0, 9),
  [7113] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 62),
  [7115] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 62),
  [7117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 64),
  [7119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 64),
  [7121] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 3, 0, 30),
  [7123] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 3, 0, 30),
  [7125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5042),
  [7127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3002),
  [7129] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 3, 0, 30),
  [7131] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 3, 0, 30),
  [7133] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 4, 0, 64),
  [7135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 4, 0, 64),
  [7137] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 3, 0, 28),
  [7139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 3, 0, 28),
  [7141] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 4, 0, 62),
  [7143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 4, 0, 62),
  [7145] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_generic_specifier_repeat1, 3, 0, 0),
  [7147] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_specifier_repeat1, 3, 0, 0),
  [7149] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 5, 0, 99),
  [7151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 5, 0, 99),
  [7153] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_specifier, 5, 0, 0),
  [7155] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_specifier, 5, 0, 0),
  [7157] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 30),
  [7159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 30),
  [7161] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 3, 0, 0),
  [7163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 3, 0, 0),
  [7165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(849),
  [7167] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 28),
  [7169] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 28),
  [7171] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 29),
  [7173] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 29),
  [7175] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_specifier, 4, 0, 0),
  [7177] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_specifier, 4, 0, 0),
  [7179] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator_list, 2, 0, 0),
  [7181] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator_list, 2, 0, 0),
  [7183] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 13),
  [7185] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, 0, 13),
  [7187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2982),
  [7189] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator_list, 5, 0, 0),
  [7191] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator_list, 5, 0, 0),
  [7193] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 8),
  [7195] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 8),
  [7197] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 94),
  [7199] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 94),
  [7201] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 30),
  [7203] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 30),
  [7205] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 9),
  [7207] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 9),
  [7209] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 95),
  [7211] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 95),
  [7213] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 2, 0, 0),
  [7215] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 2, 0, 0),
  [7217] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 4, 0, 28),
  [7219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 4, 0, 28),
  [7221] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 3, 0, 8),
  [7223] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 3, 0, 8),
  [7225] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 4, 0, 29),
  [7227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 4, 0, 29),
  [7229] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator_list, 6, 0, 0),
  [7231] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator_list, 6, 0, 0),
  [7233] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 96),
  [7235] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 96),
  [7237] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 3, 0, 9),
  [7239] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 3, 0, 9),
  [7241] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 3, 0, 8),
  [7243] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 3, 0, 8),
  [7245] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 121),
  [7247] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 7, 0, 121),
  [7249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3533),
  [7251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3533),
  [7253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3536),
  [7255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3538),
  [7257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3540),
  [7259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3421),
  [7261] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 122),
  [7263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 7, 0, 122),
  [7265] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 123),
  [7267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 7, 0, 123),
  [7269] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 3, 0, 9),
  [7271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 3, 0, 9),
  [7273] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 97),
  [7275] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 97),
  [7277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(493),
  [7279] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 124),
  [7281] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 7, 0, 124),
  [7283] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, -1, 10),
  [7285] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, -1, 10),
  [7287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2912),
  [7289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5673),
  [7291] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 125),
  [7293] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 7, 0, 125),
  [7295] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typedefed_specifier, 2, 0, 0),
  [7297] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typedefed_specifier, 2, 0, 0),
  [7299] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 98),
  [7301] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 98),
  [7303] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 4, 0, 43),
  [7305] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 4, 0, 43),
  [7307] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 4, -1, 44),
  [7309] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 4, -1, 44),
  [7311] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 99),
  [7313] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 99),
  [7315] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, -1, 14),
  [7317] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, -1, 14),
  [7319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2985),
  [7321] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(5666),
  [7324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(2903),
  [7327] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 4, 0, 30),
  [7329] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 4, 0, 30),
  [7331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2999),
  [7333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_specifier, 3, 0, 0),
  [7335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_specifier, 3, 0, 0),
  [7337] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator_list, 4, 0, 0),
  [7339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator_list, 4, 0, 0),
  [7341] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 59),
  [7343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 59),
  [7345] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 60),
  [7347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 60),
  [7349] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 6, 0, 99),
  [7351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 6, 0, 99),
  [7353] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 6, 0, 99),
  [7355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 6, 0, 99),
  [7357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 61),
  [7359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 61),
  [7361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 62),
  [7363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 62),
  [7365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 13),
  [7367] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 3, 0, 13),
  [7369] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 63),
  [7371] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 63),
  [7373] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 64),
  [7375] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 64),
  [7377] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, -1, 14),
  [7379] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 3, -1, 14),
  [7381] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 5, 0, 62),
  [7383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 5, 0, 62),
  [7385] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 5, 0, 63),
  [7387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 5, 0, 63),
  [7389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 5, 0, 64),
  [7391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 5, 0, 64),
  [7393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 5, 0, 62),
  [7395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 5, 0, 62),
  [7397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 5, 0, 63),
  [7399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 5, 0, 63),
  [7401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 5, 0, 64),
  [7403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 5, 0, 64),
  [7405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 43),
  [7407] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 3, 0, 43),
  [7409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2966),
  [7411] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, -1, 44),
  [7413] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 3, -1, 44),
  [7415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2967),
  [7417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typeof_specifier, 4, 0, 0),
  [7419] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typeof_specifier, 4, 0, 0),
  [7421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 0),
  [7423] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 3, 0, 0),
  [7425] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 8, 0, 139),
  [7427] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 8, 0, 139),
  [7429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 8, 0, 140),
  [7431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 8, 0, 140),
  [7433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1004),
  [7435] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 4),
  [7437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, 0, 4),
  [7439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ms_pointer_modifier, 1, 0, 0),
  [7441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_pointer_modifier, 1, 0, 0),
  [7443] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ms_unaligned_ptr_modifier, 1, 0, 0),
  [7445] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_unaligned_ptr_modifier, 1, 0, 0),
  [7447] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator_list, 3, 0, 0),
  [7449] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator_list, 3, 0, 0),
  [7451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 28),
  [7453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 28),
  [7455] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 29),
  [7457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 29),
  [7459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [7461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [7463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2511),
  [7465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5135),
  [7467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5098),
  [7469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6503),
  [7471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6680),
  [7473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6177),
  [7475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2916),
  [7477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6478),
  [7479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2500),
  [7481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5105),
  [7483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6186),
  [7485] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2), REDUCE(aux_sym__old_style_parameter_list_repeat1, 2, 0, 0),
  [7488] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__declarator, 1, 0, 0), REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(4081),
  [7492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2481),
  [7494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2084),
  [7496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2270),
  [7498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2167),
  [7500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2183),
  [7502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2290),
  [7504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2187),
  [7506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0),
  [7508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0),
  [7510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2100),
  [7512] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 124),
  [7514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 124),
  [7516] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 124), SHIFT(6441),
  [7519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4980),
  [7521] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 125),
  [7523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 125),
  [7525] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 125), SHIFT(6441),
  [7528] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [7531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2191),
  [7533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2200),
  [7535] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 97),
  [7537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 97),
  [7539] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 97), SHIFT(6441),
  [7542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4787),
  [7544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_declarators, 1, 0, 24),
  [7546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2520),
  [7548] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 98),
  [7550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 98),
  [7552] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 98), SHIFT(6441),
  [7555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2121),
  [7557] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_enumerator_list_repeat1, 1, 0, 0),
  [7559] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enumerator_list_repeat1, 1, 0, 0),
  [7561] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 60),
  [7563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 60),
  [7565] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 60), SHIFT(6441),
  [7568] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 61),
  [7570] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 61),
  [7572] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 61), SHIFT(6441),
  [7575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2314),
  [7577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2296),
  [7579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2102),
  [7581] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 63),
  [7583] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 63),
  [7585] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 63), SHIFT(6441),
  [7588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5778),
  [7590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 2, 0, 38),
  [7592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_field_declarator_repeat1, 2, 0, 0),
  [7594] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_field_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [7597] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_field_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [7600] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_function_field_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [7603] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_function_field_declarator_repeat1, 2, 0, 0),
  [7605] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_function_field_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(2354),
  [7608] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_field_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(6451),
  [7611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2530),
  [7613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2492),
  [7615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2482),
  [7617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2512),
  [7619] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 2, 0, 8),
  [7621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 2, 0, 8),
  [7623] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 2, 0, 8), SHIFT(6441),
  [7626] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 2, 0, 8), SHIFT(5820),
  [7629] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(4097),
  [7632] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 29),
  [7634] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 29),
  [7636] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 29), SHIFT(6441),
  [7639] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 29), SHIFT(5877),
  [7642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5820),
  [7644] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 63), SHIFT(5778),
  [7647] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(4082),
  [7650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5877),
  [7652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2920),
  [7654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4792),
  [7656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2323),
  [7658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2310),
  [7660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [7662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [7664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [7666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [7668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [7670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2172),
  [7672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [7674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [7676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1944),
  [7678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [7680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [7682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [7684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [7686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [7688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(590),
  [7690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [7692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [7694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [7696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [7698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1112),
  [7700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1113),
  [7702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1115),
  [7704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [7706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2159),
  [7708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2165),
  [7710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2169),
  [7712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2178),
  [7714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2182),
  [7716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2195),
  [7718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2145),
  [7720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(733),
  [7722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2265),
  [7724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2273),
  [7726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2278),
  [7728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2284),
  [7730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2287),
  [7732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2305),
  [7734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2325),
  [7736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(742),
  [7738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [7740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [7742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(738),
  [7744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [7746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(963),
  [7748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4051),
  [7750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1945),
  [7752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4050),
  [7754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [7756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1959),
  [7758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1963),
  [7760] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 94),
  [7762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 94),
  [7764] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 94), SHIFT(6441),
  [7767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2133),
  [7769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2275),
  [7771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(793),
  [7773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2313),
  [7775] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 95),
  [7777] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 95),
  [7779] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 95), SHIFT(6441),
  [7782] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 96),
  [7784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 96),
  [7786] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 96), SHIFT(6441),
  [7789] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 64),
  [7791] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 64),
  [7793] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 64), SHIFT(6441),
  [7796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1947),
  [7798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1948),
  [7800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [7802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2105),
  [7804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [7806] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 2, 0, 9),
  [7808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 2, 0, 9),
  [7810] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 2, 0, 9), SHIFT(6441),
  [7813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1923),
  [7815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [7817] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 99),
  [7819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 99),
  [7821] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 99), SHIFT(6441),
  [7824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [7826] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 139),
  [7828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 7, 0, 139),
  [7830] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 139), SHIFT(6441),
  [7833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2344),
  [7835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4190),
  [7837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(918),
  [7839] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_declarators, 2, 0, 56),
  [7841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(687),
  [7843] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 140),
  [7845] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 7, 0, 140),
  [7847] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 140), SHIFT(6441),
  [7850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6523),
  [7852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6523),
  [7854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 28),
  [7856] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 28),
  [7858] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 28), SHIFT(6441),
  [7861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [7863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [7865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [7867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [7869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(970),
  [7871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [7873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [7875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 121),
  [7877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 121),
  [7879] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 121), SHIFT(6441),
  [7882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4198),
  [7884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [7886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2137),
  [7888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [7890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4173),
  [7892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4177),
  [7894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4178),
  [7896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(950),
  [7898] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 30),
  [7900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 30),
  [7902] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 30), SHIFT(6441),
  [7905] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 122),
  [7907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 122),
  [7909] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 122), SHIFT(6441),
  [7912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [7914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [7916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2408),
  [7918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4176),
  [7920] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 123),
  [7922] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 123),
  [7924] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 123), SHIFT(6441),
  [7927] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 59),
  [7929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 59),
  [7931] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 59), SHIFT(6441),
  [7934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2330),
  [7936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2415),
  [7938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(974),
  [7940] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 62),
  [7942] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 62),
  [7944] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 62), SHIFT(6441),
  [7947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4171),
  [7949] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(6523),
  [7952] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(6523),
  [7955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3791),
  [7957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_name, 1, 0, 0),
  [7959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3836),
  [7961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6176),
  [7963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2536),
  [7965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3771),
  [7967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6178),
  [7969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3790),
  [7971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4879),
  [7973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4907),
  [7975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4890),
  [7977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3827),
  [7979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6184),
  [7981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3786),
  [7983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 21),
  [7985] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 21),
  [7987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4884),
  [7989] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3229),
  [7992] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 2, 0, 0),
  [7994] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5363),
  [7997] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat1, 2, 0, 0),
  [7999] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3008),
  [8002] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat1, 1, 0, 0),
  [8004] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 1, 0, 0), SHIFT_REPEAT(3225),
  [8007] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 1, 0, 0),
  [8009] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat1, 1, 0, 0), SHIFT_REPEAT(6441),
  [8012] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 1, 0, 0), SHIFT_REPEAT(3008),
  [8015] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type_declarator, 2, 0, 22),
  [8017] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type_declarator, 2, 0, 22),
  [8019] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type_declarator, 3, 0, 22),
  [8021] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type_declarator, 3, 0, 22),
  [8023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3236),
  [8025] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [8028] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_type, 3, 0, 0),
  [8030] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_type, 3, 0, 0),
  [8032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2251),
  [8034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2318),
  [8036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2411),
  [8038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2362),
  [8040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator, 1, 0, 26),
  [8042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator, 1, 0, 26),
  [8044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3491),
  [8046] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_type, 4, 0, 0),
  [8048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_type, 4, 0, 0),
  [8050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2218),
  [8052] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat1, 3, 0, 0),
  [8054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 3, 0, 0),
  [8056] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat1, 3, 0, 0), SHIFT_REPEAT(6441),
  [8059] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 3, 0, 0), SHIFT_REPEAT(3008),
  [8062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2379),
  [8064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_type, 5, 0, 0),
  [8066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_type, 5, 0, 0),
  [8068] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat1, 4, 0, 0),
  [8070] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 4, 0, 0),
  [8072] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 4, 0, 0), SHIFT_REPEAT(3008),
  [8075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_declarator, 1, 0, 0),
  [8077] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_declarator, 1, 0, 0),
  [8079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3731),
  [8081] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_declarator, 4, 0, 90),
  [8083] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_declarator, 4, 0, 90),
  [8085] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_pointer_type_declarator, 2, 1, 21),
  [8087] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_pointer_type_declarator, 2, 1, 21),
  [8089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2526),
  [8091] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_type_declarator, 4, 1, 89),
  [8093] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_type_declarator, 4, 1, 89),
  [8095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_type_declarator, 2, 1, 21),
  [8097] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_type_declarator, 2, 1, 21),
  [8099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1573),
  [8101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_name_repeat1, 1, 0, 0),
  [8103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [8105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6197),
  [8107] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_declaration_repeat1, 5, 0, 0),
  [8109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_declaration_repeat1, 5, 0, 0),
  [8111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_type_declarator, 5, 1, 118),
  [8113] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_type_declarator, 5, 1, 118),
  [8115] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_declarator, 4, 0, 24),
  [8117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_declarator, 4, 0, 24),
  [8119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_declarator, 5, 0, 119),
  [8121] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_declarator, 5, 0, 119),
  [8123] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_declarator, 1, 0, 23),
  [8125] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_declarator, 1, 0, 23),
  [8127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(62),
  [8129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5990),
  [8131] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_declarator, 1, 0, 0),
  [8133] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_declarator, 1, 0, 0),
  [8135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_type_declarator, 3, -10, 0),
  [8137] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_type_declarator, 3, -10, 0),
  [8139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_type_declarator, 4, -10, 0),
  [8141] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_type_declarator, 4, -10, 0),
  [8143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_type_declarator, 3, 1, 54),
  [8145] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_type_declarator, 3, 1, 54),
  [8147] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_declarator, 1, 0, 2),
  [8149] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_declarator, 1, 0, 2),
  [8151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_definition_declarators_repeat1, 2, 0, 21),
  [8153] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__type_definition_declarators_repeat1, 2, 0, 21),
  [8155] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_pointer_type_declarator, 3, 1, 54),
  [8157] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_pointer_type_declarator, 3, 1, 54),
  [8159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_declarator, 3, 0, 24),
  [8161] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_declarator, 3, 0, 24),
  [8163] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_definition_repeat2, 2, 0, 0), SHIFT_REPEAT(2084),
  [8166] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__type_definition_declarators_repeat1, 2, 0, 75), SHIFT_REPEAT(4787),
  [8169] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_definition_declarators_repeat1, 2, 0, 75),
  [8171] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__type_definition_declarators_repeat1, 2, 0, 75),
  [8173] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 75), SHIFT_REPEAT(4215),
  [8176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 75),
  [8178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 75),
  [8180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4282),
  [8182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6608),
  [8184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3410),
  [8186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1500),
  [8188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6765),
  [8190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4087),
  [8192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4582),
  [8194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5987),
  [8196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6592),
  [8198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2212),
  [8200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4109),
  [8202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4885),
  [8204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4054),
  [8206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6726),
  [8208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1668),
  [8210] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_struct_declaration_repeat2, 2, 0, 0), SHIFT_REPEAT(4792),
  [8213] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_struct_declaration_repeat2, 2, 0, 0),
  [8215] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat2, 2, 0, 0),
  [8217] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(4087),
  [8220] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(4582),
  [8223] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(5987),
  [8226] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(3836),
  [8229] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(6176),
  [8232] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(6592),
  [8235] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0),
  [8237] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(3771),
  [8240] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(4109),
  [8243] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(6178),
  [8246] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(3790),
  [8249] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(4885),
  [8252] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(4907),
  [8255] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(4890),
  [8258] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(3827),
  [8261] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(6184),
  [8264] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(4054),
  [8267] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0), SHIFT_REPEAT(3786),
  [8270] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_c_method_parameter_repeat1, 2, 0, 75), SHIFT_REPEAT(4799),
  [8273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2207),
  [8275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1502),
  [8277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1485),
  [8279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1659),
  [8281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1654),
  [8283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 3, 0, 109),
  [8285] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_declaration_repeat1, 3, 0, 109),
  [8287] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3791),
  [8290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0),
  [8292] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(2822),
  [8295] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(2545),
  [8298] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(4226),
  [8301] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3836),
  [8304] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(6176),
  [8307] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5872),
  [8310] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3771),
  [8313] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(6178),
  [8316] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3790),
  [8319] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(4879),
  [8322] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(4907),
  [8325] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(4890),
  [8328] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3827),
  [8331] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(6184),
  [8334] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3786),
  [8337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5779),
  [8339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5514),
  [8341] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 2, 0, 0), SHIFT_REPEAT(4884),
  [8344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3811),
  [8346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6135),
  [8348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5981),
  [8350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6041),
  [8352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6515),
  [8354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6631),
  [8356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6273),
  [8358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5904),
  [8360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6118),
  [8362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6086),
  [8364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2470),
  [8366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3740),
  [8368] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_array_declarator_repeat1, 2, 0, 0),
  [8370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_declarator_repeat1, 2, 0, 0),
  [8372] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_array_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(3672),
  [8375] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_array_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(3325),
  [8378] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_array_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(6376),
  [8381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2961),
  [8383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3558),
  [8385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3558),
  [8387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1535),
  [8389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3288),
  [8391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3425),
  [8393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3500),
  [8395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3525),
  [8397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1635),
  [8399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [8401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1446),
  [8403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1543),
  [8405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3535),
  [8407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3501),
  [8409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3526),
  [8411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3486),
  [8413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1727),
  [8415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1546),
  [8417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3583),
  [8419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3585),
  [8421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3586),
  [8423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3631),
  [8425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3367),
  [8427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3505),
  [8429] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 2, 0, 67),
  [8431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(546),
  [8433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3379),
  [8435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1445),
  [8437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1541),
  [8439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3544),
  [8441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3546),
  [8443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3459),
  [8445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3574),
  [8447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1458),
  [8449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4267),
  [8451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1544),
  [8453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3499),
  [8455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1461),
  [8457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3539),
  [8459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1549),
  [8461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3707),
  [8463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3495),
  [8465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1155),
  [8467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3393),
  [8469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3336),
  [8471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5195),
  [8473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6202),
  [8475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [8477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3517),
  [8479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [8481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1545),
  [8483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3608),
  [8485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3609),
  [8487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3635),
  [8489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5941),
  [8491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1530),
  [8493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3712),
  [8495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3639),
  [8497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3640),
  [8499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3714),
  [8501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6275),
  [8503] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 3, 0, 103),
  [8505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4255),
  [8507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3703),
  [8509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3343),
  [8511] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 3, 0, 67),
  [8513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3657),
  [8515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(946),
  [8517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(713),
  [8519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1550),
  [8521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3716),
  [8523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3717),
  [8525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3736),
  [8527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3700),
  [8529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3627),
  [8531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1644),
  [8533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1539),
  [8535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3547),
  [8537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3604),
  [8539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3733),
  [8541] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 2, 0, 0),
  [8543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3450),
  [8545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1490),
  [8547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1651),
  [8549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [8551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6630),
  [8553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4265),
  [8555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6007),
  [8557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 4, 0, 129),
  [8559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1634),
  [8561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 3, 0, 47),
  [8563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4864),
  [8565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6543),
  [8567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1511),
  [8569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3349),
  [8571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3355),
  [8573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5282),
  [8575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(720),
  [8577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5258),
  [8579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5897),
  [8581] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0), SHIFT_REPEAT(3672),
  [8584] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0), SHIFT_REPEAT(6376),
  [8587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3059),
  [8589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1457),
  [8591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3987),
  [8593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [8595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4840),
  [8597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3252),
  [8599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4839),
  [8601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1645),
  [8603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3978),
  [8605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [8607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5888),
  [8609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1096),
  [8611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3243),
  [8613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3070),
  [8615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5313),
  [8617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [8619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [8621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4869),
  [8623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4902),
  [8625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(783),
  [8627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1643),
  [8629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2784),
  [8631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3388),
  [8633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6253),
  [8635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2786),
  [8637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4925),
  [8639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3363),
  [8641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3396),
  [8643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1552),
  [8645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1534),
  [8647] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 1, 0, 0), REDUCE(aux_sym_array_declarator_repeat1, 1, 0, 0),
  [8650] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__type_definition_type_repeat1, 1, 0, 0), REDUCE(aux_sym_array_declarator_repeat1, 1, 0, 0),
  [8653] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_declarator_repeat1, 1, 0, 0),
  [8655] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_array_declarator_repeat1, 1, 0, 0),
  [8657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1529),
  [8659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1533),
  [8661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1538),
  [8663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1536),
  [8665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1532),
  [8667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1540),
  [8669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1537),
  [8671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1542),
  [8673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1547),
  [8675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1548),
  [8677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5372),
  [8679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4107),
  [8681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4108),
  [8683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4110),
  [8685] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4087),
  [8688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0),
  [8690] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3836),
  [8693] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0),
  [8695] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3771),
  [8698] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6178),
  [8701] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3790),
  [8704] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4885),
  [8707] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4907),
  [8710] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4890),
  [8713] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3827),
  [8716] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6184),
  [8719] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_struct_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3786),
  [8722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4961),
  [8724] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(3789),
  [8727] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(3836),
  [8730] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(3787),
  [8733] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(6178),
  [8736] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(3788),
  [8739] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(3817),
  [8742] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(3836),
  [8745] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(3818),
  [8748] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(6178),
  [8751] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(3816),
  [8754] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0), SHIFT_REPEAT(3836),
  [8757] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0), SHIFT_REPEAT(6178),
  [8760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1591),
  [8762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6292),
  [8764] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0), SHIFT(3792),
  [8768] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 13), SHIFT(3814),
  [8771] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, -1, 14), SHIFT(3815),
  [8774] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 0), SHIFT(3799),
  [8777] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__declarator, 1, 0, 0), REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(3318),
  [8781] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym__declarator, 1, 0, 0), REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(3812),
  [8785] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0), SHIFT_REPEAT(3792),
  [8788] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 4), SHIFT(3792),
  [8791] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 2, 0, 8), SHIFT(6176),
  [8794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5002),
  [8796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5732),
  [8798] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(6176),
  [8801] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generic_specifier_repeat1, 2, 0, 0), SHIFT_REPEAT(3318),
  [8804] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 4, 0, 43), SHIFT(3792),
  [8807] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 4, -1, 44), SHIFT(3792),
  [8810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3316),
  [8812] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, -1, 10), SHIFT(3792),
  [8815] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 13), SHIFT(3792),
  [8818] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, -1, 14), SHIFT(3792),
  [8821] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 43), SHIFT(3809),
  [8824] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, -1, 44), SHIFT(3810),
  [8827] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 0), SHIFT(3792),
  [8830] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 63), SHIFT(6176),
  [8833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5851),
  [8835] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 29), SHIFT(6176),
  [8838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5798),
  [8840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4200),
  [8842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4185),
  [8844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4260),
  [8846] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_protocol_qualifier, 1, 0, 0),
  [8848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protocol_qualifier, 1, 0, 0),
  [8850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1823),
  [8852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2517),
  [8854] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 97), SHIFT(6176),
  [8857] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 98), SHIFT(6176),
  [8860] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 99), SHIFT(6176),
  [8863] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 139), SHIFT(6176),
  [8866] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 7, 0, 140), SHIFT(6176),
  [8869] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 30), SHIFT(6176),
  [8872] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 124), SHIFT(6176),
  [8875] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 125), SHIFT(6176),
  [8878] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 59), SHIFT(6176),
  [8881] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 121), SHIFT(6176),
  [8884] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 122), SHIFT(6176),
  [8887] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 123), SHIFT(6176),
  [8890] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 62), SHIFT(6176),
  [8893] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 64), SHIFT(6176),
  [8896] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 94), SHIFT(6176),
  [8899] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 95), SHIFT(6176),
  [8902] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_name_repeat1, 1, 0, 0),
  [8904] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_specifier_qualifier, 1, 0, 0),
  [8906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_specifier_qualifier, 1, 0, 0),
  [8908] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 60), SHIFT(6176),
  [8911] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 61), SHIFT(6176),
  [8914] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 96), SHIFT(6176),
  [8917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 2, 0, 9), SHIFT(6176),
  [8920] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 28), SHIFT(6176),
  [8923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2559),
  [8925] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 63), SHIFT(5002),
  [8928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5768),
  [8930] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 2, 0, 8), SHIFT(5002),
  [8933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5830),
  [8935] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 29), SHIFT(5002),
  [8938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5773),
  [8940] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 124), SHIFT(5002),
  [8943] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 125), SHIFT(5002),
  [8946] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 97), SHIFT(5002),
  [8949] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 60), SHIFT(5002),
  [8952] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 98), SHIFT(5002),
  [8955] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 61), SHIFT(5002),
  [8958] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_instance_variables_repeat1, 2, 0, 0),
  [8960] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_visibility_specification, 1, 0, 0),
  [8962] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_visibility_specification, 1, 0, 0),
  [8964] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instance_variable, 1, 0, 0),
  [8966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable, 1, 0, 0),
  [8968] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 2, 0, 8), SHIFT(5777),
  [8971] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 29), SHIFT(5825),
  [8974] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 63), SHIFT(5857),
  [8977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4875),
  [8979] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(4093),
  [8982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4560),
  [8984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4555),
  [8986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4492),
  [8988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4544),
  [8990] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_block_pointer_declarator, 1, 1, 0),
  [8992] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_block_pointer_declarator, 1, 1, 0),
  [8994] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_block_pointer_declarator, 2, 1, 0),
  [8996] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_block_pointer_declarator, 2, 1, 0),
  [8998] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_attributes_declaration, 3, 0, 0),
  [9000] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_attributes_declaration, 3, 0, 0),
  [9002] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_attributes_declaration, 4, 0, 0),
  [9004] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_attributes_declaration, 4, 0, 0),
  [9006] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_attributes_declaration, 2, 0, 0),
  [9008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_attributes_declaration, 2, 0, 0),
  [9010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5890),
  [9012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6437),
  [9014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6490),
  [9016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6254),
  [9018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6009),
  [9020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6621),
  [9022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6323),
  [9024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6657),
  [9026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5718),
  [9028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5850),
  [9030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5789),
  [9032] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 2, 0, 13),
  [9034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2510),
  [9036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2993),
  [9038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2969),
  [9040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4237),
  [9042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 3, 0, 13),
  [9044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 1, 0, 4),
  [9046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 2, 0, 4),
  [9048] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 1, 0, 4),
  [9050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 1, 0, 4),
  [9052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2557),
  [9054] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 2, 0, 13),
  [9056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 2, 0, 13),
  [9058] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(4248),
  [9061] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(4245),
  [9064] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(4251),
  [9067] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__type_definition_type, 1, 0, 4), REDUCE(sym__type_definition_type, 2, 0, 13),
  [9070] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__type_definition_type, 1, 0, 4), REDUCE(sym__type_definition_type, 2, 0, 13),
  [9073] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 3, 0, 43),
  [9075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 3, 0, 43),
  [9077] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(4238),
  [9080] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(4244),
  [9083] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(4241),
  [9086] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 2), SHIFT(4256),
  [9089] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, -1, 44), SHIFT(4261),
  [9092] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 4), SHIFT(2912),
  [9095] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__type_definition_type, 3, 0, 4), REDUCE(sym__type_definition_type, 4, 0, 13),
  [9098] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__type_definition_type, 3, 0, 4), REDUCE(sym__type_definition_type, 4, 0, 13),
  [9101] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 43), SHIFT(4257),
  [9104] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__type_definition_type, 2, 0, 4), REDUCE(sym__type_definition_type, 3, 0, 13),
  [9107] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__type_definition_type, 2, 0, 4), REDUCE(sym__type_definition_type, 3, 0, 13),
  [9110] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 4, 0, 43),
  [9112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 4, 0, 43),
  [9114] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 0), SHIFT(2912),
  [9117] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0), SHIFT(2912),
  [9121] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 3, 0, 13),
  [9123] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 3, 0, 13),
  [9125] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, -1, 14), SHIFT(4263),
  [9128] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 2, 0, 4),
  [9130] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 2, 0, 4),
  [9132] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 13), SHIFT(4262),
  [9135] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 3, 0, 4),
  [9137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 3, 0, 4),
  [9139] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 4, 0, 13),
  [9141] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 4, 0, 13),
  [9143] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 5, 0, 43),
  [9145] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 5, 0, 43),
  [9147] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 1, 0, 26),
  [9149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_expression, 1, 0, 0), SHIFT(3401),
  [9152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6643),
  [9154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3582),
  [9156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3582),
  [9158] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, -1, 10), SHIFT(2912),
  [9161] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 4, 0, 43), SHIFT(2912),
  [9164] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 0), SHIFT(4239),
  [9167] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 4, -1, 44), SHIFT(2912),
  [9170] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 13), SHIFT(2912),
  [9173] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, -1, 14), SHIFT(2912),
  [9176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_function_declarator, 3, 0, 22),
  [9178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_function_declarator, 2, 0, 19),
  [9180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4296),
  [9182] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_function_declarator, 2, 0, 22),
  [9184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_function_declarator, 1, 0, 19),
  [9186] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_field_declarator, 2, 0, 22),
  [9188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_field_declarator, 3, 0, 22),
  [9190] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_field_declarator, 3, 0, 22),
  [9192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3315),
  [9194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3334),
  [9196] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 63), SHIFT(5850),
  [9199] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 29), SHIFT(5789),
  [9202] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 2, 0, 8), SHIFT(5718),
  [9205] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_function_declarator, 3, 0, 22),
  [9207] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_function_declarator, 2, 0, 19),
  [9209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3461),
  [9211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3462),
  [9213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3468),
  [9215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3469),
  [9217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3470),
  [9219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3470),
  [9221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3471),
  [9223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3457),
  [9225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3463),
  [9227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3464),
  [9229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3466),
  [9231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3467),
  [9233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3342),
  [9235] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3457),
  [9238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 2, 0, 0),
  [9240] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 3, 0, 87),
  [9242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3507),
  [9244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3508),
  [9246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3509),
  [9248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3510),
  [9250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3511),
  [9252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3512),
  [9254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3513),
  [9256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3514),
  [9258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3516),
  [9260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3520),
  [9262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3520),
  [9264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3522),
  [9266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3335),
  [9268] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3507),
  [9271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3572),
  [9273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3654),
  [9275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3730),
  [9277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3599),
  [9279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3701),
  [9281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3662),
  [9283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3479),
  [9285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3704),
  [9287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3523),
  [9289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3713),
  [9291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3621),
  [9293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3580),
  [9295] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3689),
  [9298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3690),
  [9300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3691),
  [9302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3692),
  [9304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3693),
  [9306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3694),
  [9308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3695),
  [9310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3696),
  [9312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3697),
  [9314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3698),
  [9316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3698),
  [9318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3699),
  [9320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3689),
  [9322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3365),
  [9324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [9326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2885),
  [9328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3481),
  [9330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3483),
  [9332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3483),
  [9334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3453),
  [9336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3493),
  [9338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3506),
  [9340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3515),
  [9342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3518),
  [9344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3519),
  [9346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3550),
  [9348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3628),
  [9350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3628),
  [9352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3636),
  [9354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3356),
  [9356] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3453),
  [9359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5208),
  [9361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3448),
  [9363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3413),
  [9365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3600),
  [9367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3601),
  [9369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3601),
  [9371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3602),
  [9373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3603),
  [9375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3605),
  [9377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3620),
  [9379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3622),
  [9381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3625),
  [9383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3626),
  [9385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3626),
  [9387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3633),
  [9389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1637),
  [9391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3373),
  [9393] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bitfield_clause, 2, 0, 0),
  [9395] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bitfield_clause, 2, 0, 0),
  [9397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3397),
  [9399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1482),
  [9401] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_message_expression_repeat1, 2, 0, 0),
  [9403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3452),
  [9405] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3448),
  [9408] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_repeat1, 2, 0, 0),
  [9410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3428),
  [9412] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comma_expression, 3, 0, 51),
  [9414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3559),
  [9416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3521),
  [9418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5767),
  [9420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3560),
  [9422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3561),
  [9424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3561),
  [9426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3562),
  [9428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3563),
  [9430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3564),
  [9432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3565),
  [9434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3566),
  [9436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3567),
  [9438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3568),
  [9440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3568),
  [9442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3569),
  [9444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3347),
  [9446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5347),
  [9448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3281),
  [9450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1475),
  [9452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2419),
  [9454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1463),
  [9456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4084),
  [9458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1450),
  [9460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1663),
  [9462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6192),
  [9464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6553),
  [9466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5607),
  [9468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1447),
  [9470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5799),
  [9472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5663),
  [9474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5325),
  [9476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3284),
  [9478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1639),
  [9480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5546),
  [9482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5675),
  [9484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5330),
  [9486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3531),
  [9488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5669),
  [9490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5355),
  [9492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [9494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(956),
  [9496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [9498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [9500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [9502] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_expression_repeat1, 4, 0, 0),
  [9504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary_pair, 3, 0, 0),
  [9506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3629),
  [9508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3341),
  [9510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3606),
  [9512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3394),
  [9514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3362),
  [9516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0),
  [9518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3438),
  [9520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 4, 0, 126),
  [9522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(721),
  [9524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1122),
  [9526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 4, 0, 127),
  [9528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(725),
  [9530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1133),
  [9532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3668),
  [9534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 5, 0, 142),
  [9536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4073),
  [9538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 4, 0, 128),
  [9540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3732),
  [9542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [9544] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0), SHIFT(3559),
  [9547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3576),
  [9549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1498),
  [9551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5458),
  [9553] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 3, 0, 0),
  [9555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [9557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [9559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3581),
  [9561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [9563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [9565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3474),
  [9567] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_availability_repeat1, 4, 0, 0),
  [9569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_pair, 3, 0, 136),
  [9571] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_initializer_list_repeat1, 2, 0, 0),
  [9573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3705),
  [9575] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_pair, 3, 0, 137),
  [9577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3527),
  [9579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3659),
  [9581] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 3, 0, 101),
  [9583] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 3, 0, 102),
  [9585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3391),
  [9587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3734),
  [9589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3623),
  [9591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4086),
  [9593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3490),
  [9595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2356),
  [9597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2936),
  [9599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [9601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4844),
  [9603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4269),
  [9605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(588),
  [9607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [9609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [9611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1476),
  [9613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5422),
  [9615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6545),
  [9617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [9619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3637),
  [9621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2560),
  [9623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2561),
  [9625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5464),
  [9627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6006),
  [9629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2567),
  [9631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2568),
  [9633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2570),
  [9635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2571),
  [9637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2574),
  [9639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2575),
  [9641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2576),
  [9643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2577),
  [9645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5554),
  [9647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [9649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [9651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3445),
  [9653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4077),
  [9655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2563),
  [9657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6283),
  [9659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2564),
  [9661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3482),
  [9663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5560),
  [9665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [9667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3843),
  [9669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3968),
  [9671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3865),
  [9673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2994),
  [9675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [9677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2558),
  [9679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3848),
  [9681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4314),
  [9683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(740),
  [9685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4150),
  [9687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4094),
  [9689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [9691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2930),
  [9693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2546),
  [9695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [9697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5096),
  [9699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2549),
  [9701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [9703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2547),
  [9705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4089),
  [9707] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 1, 0, 20),
  [9709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4568),
  [9711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4663),
  [9713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4602),
  [9715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5734),
  [9717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4580),
  [9719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4580),
  [9721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4583),
  [9723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5461),
  [9725] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__preproc_expression, 1, 0, 0),
  [9727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4569),
  [9729] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__preproc_expression, 1, 0, 0),
  [9731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4625),
  [9733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4606),
  [9735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6285),
  [9737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4579),
  [9739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4651),
  [9741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4613),
  [9743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4571),
  [9745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5859),
  [9747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4610),
  [9749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4610),
  [9751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4671),
  [9753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5384),
  [9755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4678),
  [9757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4652),
  [9759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4682),
  [9761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4653),
  [9763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4654),
  [9765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4685),
  [9767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4612),
  [9769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6281),
  [9771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4591),
  [9773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4592),
  [9775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4592),
  [9777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4593),
  [9779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4594),
  [9781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4597),
  [9783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4599),
  [9785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4600),
  [9787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4605),
  [9789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4607),
  [9791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4607),
  [9793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4609),
  [9795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4640),
  [9797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4664),
  [9799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4666),
  [9801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4684),
  [9803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4658),
  [9805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4660),
  [9807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4647),
  [9809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4680),
  [9811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4661),
  [9813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4657),
  [9815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4675),
  [9817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4632),
  [9819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4633),
  [9821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4636),
  [9823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4637),
  [9825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4646),
  [9827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4679),
  [9829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4641),
  [9831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4681),
  [9833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4642),
  [9835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4643),
  [9837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4655),
  [9839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4669),
  [9841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2751),
  [9843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3497),
  [9845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2760),
  [9847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4644),
  [9849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4635),
  [9851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4645),
  [9853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4659),
  [9855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4624),
  [9857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4686),
  [9859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4683),
  [9861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4630),
  [9863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4567),
  [9865] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [9867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4649),
  [9869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4670),
  [9871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4662),
  [9873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4648),
  [9875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2628),
  [9877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4650),
  [9879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4673),
  [9881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4676),
  [9883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4677),
  [9885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_binary_expression, 3, 0, 34),
  [9887] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_binary_expression, 3, 0, 34),
  [9889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_argument_list, 2, 0, 0),
  [9891] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_argument_list, 2, 0, 0),
  [9893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_call_expression, 2, 0, 12),
  [9895] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_call_expression, 2, 0, 12),
  [9897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_argument_list, 4, 0, 0),
  [9899] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_argument_list, 4, 0, 0),
  [9901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4809),
  [9903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5779),
  [9905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5052),
  [9907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_defined, 4, 0, 0),
  [9909] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_defined, 4, 0, 0),
  [9911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_argument_list_repeat1, 2, 0, 0),
  [9913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_parenthesized_expression, 3, 0, 0),
  [9915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_parenthesized_expression, 3, 0, 0),
  [9917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4802),
  [9919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_argument_list, 3, 0, 0),
  [9921] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_argument_list, 3, 0, 0),
  [9923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4821),
  [9925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_defined, 2, 0, 0),
  [9927] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_defined, 2, 0, 0),
  [9929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_unary_expression, 2, 0, 6),
  [9931] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_unary_expression, 2, 0, 6),
  [9933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4615),
  [9935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4620),
  [9937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4574),
  [9939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4576),
  [9941] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4577),
  [9943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4585),
  [9945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1521),
  [9947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4621),
  [9949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4622),
  [9951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4611),
  [9953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4595),
  [9955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1522),
  [9957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [9959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1185),
  [9961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4667),
  [9963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [9965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [9967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [9969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1517),
  [9971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [9973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1244),
  [9975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4894),
  [9977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [9979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4631),
  [9981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [9983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1251),
  [9985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1254),
  [9987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1193),
  [9989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(29),
  [9991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1217),
  [9993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [9995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1519),
  [9997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_name, 2, 0, 0),
  [9999] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_name, 2, 0, 0),
  [10001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5363),
  [10003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4789),
  [10005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2757),
  [10007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2525),
  [10009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2691),
  [10011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2652),
  [10013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4638),
  [10015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2357),
  [10017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5160),
  [10019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5035),
  [10021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5109),
  [10023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5144),
  [10025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4985),
  [10027] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_clobber_list, 1, 0, 0),
  [10029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_availability_attribute_specifier_repeat1, 2, 0, 0),
  [10031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5118),
  [10033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1756),
  [10035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5725),
  [10037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5117),
  [10039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5273),
  [10041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4581),
  [10043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6108),
  [10045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5815),
  [10047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1648),
  [10049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5514),
  [10051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1760),
  [10053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1755),
  [10055] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 21),
  [10057] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 21),
  [10059] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 21), SHIFT_REPEAT(3497),
  [10062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_output_operand_list, 1, 0, 0),
  [10064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6710),
  [10066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6001),
  [10068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5159),
  [10070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1487),
  [10072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_input_operand_list, 1, 0, 0),
  [10074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6008),
  [10076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_parenthesized_declarator, 4, 0, 0),
  [10078] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_parenthesized_declarator, 4, 0, 0),
  [10080] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 2, 0, 0),
  [10082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5352),
  [10084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6054),
  [10086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5213),
  [10088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6474),
  [10090] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 4, 0, 24),
  [10092] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 4, 0, 24),
  [10094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 3, 0, 87),
  [10096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 4, 0, 90),
  [10098] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 4, 0, 90),
  [10100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5516),
  [10102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6095),
  [10104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 3, 0, 0),
  [10106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 3, 0, 0),
  [10108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4081),
  [10110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_pointer_field_declarator, 3, 1, 54),
  [10112] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_pointer_field_declarator, 3, 1, 54),
  [10114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 3, 0, 24),
  [10116] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 3, 0, 24),
  [10118] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_available_expression_repeat1, 2, 0, 0),
  [10120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_field_declarator, 2, 1, 21),
  [10122] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_field_declarator, 2, 1, 21),
  [10124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5595),
  [10126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6389),
  [10128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_pointer_field_declarator, 2, 1, 21),
  [10130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_pointer_field_declarator, 2, 1, 21),
  [10132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__abstract_declarator, 1, 0, 0),
  [10134] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__abstract_declarator, 1, 0, 0),
  [10136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5451),
  [10138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6758),
  [10140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5493),
  [10142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6043),
  [10144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_parenthesized_declarator, 3, 0, 0),
  [10146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_parenthesized_declarator, 3, 0, 0),
  [10148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0),
  [10150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0),
  [10152] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_field_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(6209),
  [10155] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attributed_field_declarator, 2, 0, 0),
  [10157] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attributed_field_declarator, 2, 0, 0),
  [10159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 4, 0, 116),
  [10161] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 4, 0, 116),
  [10163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_field_declarator, 4, 1, 89),
  [10165] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_field_declarator, 4, 1, 89),
  [10167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_field_declarator, 3, 1, 54),
  [10169] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_field_declarator, 3, 1, 54),
  [10171] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_field_declarator, 5, 1, 118),
  [10173] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_field_declarator, 5, 1, 118),
  [10175] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 5, 0, 119),
  [10177] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 5, 0, 119),
  [10179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3058),
  [10181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6326),
  [10183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4133),
  [10185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6263),
  [10187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1732),
  [10189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1738),
  [10191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4286),
  [10193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3062),
  [10195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1742),
  [10197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3800),
  [10199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6624),
  [10201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3994),
  [10203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4065),
  [10205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5938),
  [10207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5242),
  [10209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5629),
  [10211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3232),
  [10213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5066),
  [10215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3955),
  [10217] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__field_declarator, 1, 0, 0),
  [10219] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__field_declarator, 1, 0, 0),
  [10221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3781),
  [10223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2758),
  [10225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2692),
  [10227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4858),
  [10229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6633),
  [10231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5145),
  [10233] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator, 3, 0, 47),
  [10235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4581),
  [10237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2714),
  [10239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6349),
  [10241] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_field_declarator, 4, -10, 0),
  [10243] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_field_declarator, 4, -10, 0),
  [10245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2715),
  [10247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2717),
  [10249] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_field_declarator, 4, 0, 90),
  [10251] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_field_declarator, 4, 0, 90),
  [10253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2685),
  [10255] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_field_declarator, 4, 0, 24),
  [10257] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_field_declarator, 4, 0, 24),
  [10259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5079),
  [10261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2975),
  [10263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3780),
  [10265] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_ifdef_in_enumerator_repeat1, 2, 0, 0), SHIFT_REPEAT(5725),
  [10268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_ifdef_in_enumerator_repeat1, 2, 0, 0),
  [10270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_preproc_ifdef_in_enumerator_repeat1, 2, 0, 0),
  [10272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2725),
  [10274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2730),
  [10276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6671),
  [10278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2621),
  [10280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator, 4, 0, 47),
  [10282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_field_declarator, 3, -10, 0),
  [10284] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_field_declarator, 3, -10, 0),
  [10286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5036),
  [10288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2939),
  [10290] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3232),
  [10293] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5108),
  [10296] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0), SHIFT_REPEAT(6406),
  [10299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5956),
  [10301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6020),
  [10303] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_field_declarator, 5, 0, 119),
  [10305] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_field_declarator, 5, 0, 119),
  [10307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6066),
  [10309] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_field_declarator, 3, 0, 24),
  [10311] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_field_declarator, 3, 0, 24),
  [10313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2653),
  [10315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6103),
  [10317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2767),
  [10319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2873),
  [10321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2694),
  [10323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2695),
  [10325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2697),
  [10327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2698),
  [10329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2772),
  [10331] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5242),
  [10334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5086),
  [10336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3869),
  [10338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2596),
  [10340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2597),
  [10342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2600),
  [10344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2601),
  [10346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__field_declarator, 1, 0, 65),
  [10348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__field_declarator, 1, 0, 65),
  [10350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2615),
  [10352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6739),
  [10354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5450),
  [10356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5255),
  [10358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5132),
  [10360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5754),
  [10362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5052),
  [10364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3229),
  [10366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1821),
  [10368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4843),
  [10370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4855),
  [10372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_name, 3, 0, 0),
  [10374] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_name, 3, 0, 0),
  [10376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2535),
  [10378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4850),
  [10380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4868),
  [10382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5269),
  [10384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5714),
  [10386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5295),
  [10388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5736),
  [10390] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_initializer_pair_repeat1, 2, 0, 0), SHIFT_REPEAT(3710),
  [10393] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_initializer_pair_repeat1, 2, 0, 0),
  [10395] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_initializer_pair_repeat1, 2, 0, 0), SHIFT_REPEAT(6765),
  [10398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_pointer_declarator, 2, 1, 21),
  [10400] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_block_pointer_declarator, 2, 1, 21),
  [10402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6387),
  [10404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2937),
  [10406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5181),
  [10408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5726),
  [10410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_expression_repeat1, 2, 0, 0),
  [10412] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_gnu_asm_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5145),
  [10415] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_pointer_declarator, 2, 1, 21),
  [10417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_block_pointer_declarator, 2, 1, 21),
  [10419] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_pointer_declarator, 3, 1, 54),
  [10421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_pointer_declarator, 3, 1, 54),
  [10423] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_block_pointer_declarator, 3, 1, 54),
  [10425] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_block_pointer_declarator, 3, 1, 54),
  [10427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4852),
  [10429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3710),
  [10431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3390),
  [10433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6765),
  [10435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4862),
  [10437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4846),
  [10439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6340),
  [10441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3838),
  [10443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2795),
  [10445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2802),
  [10447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3061),
  [10449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6626),
  [10451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4067),
  [10453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5957),
  [10455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4918),
  [10457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5746),
  [10459] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_identifier, 2, 0, 0),
  [10461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5113),
  [10463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2949),
  [10465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3999),
  [10467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6675),
  [10469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1744),
  [10471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4285),
  [10473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6360),
  [10475] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 3, 0, 50),
  [10477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1764),
  [10479] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 3, 0, 53),
  [10481] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 75), SHIFT_REPEAT(4789),
  [10484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 75),
  [10486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 75),
  [10488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_identifier, 1, 0, 0),
  [10490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3069),
  [10492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1730),
  [10494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1742),
  [10496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3820),
  [10498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 2, 0, 18),
  [10500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 4, 0, 88),
  [10502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4899),
  [10504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3852),
  [10506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4139),
  [10508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1732),
  [10510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1738),
  [10512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6556),
  [10514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1038),
  [10516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_protocol_declaration_repeat1, 2, 0, 0),
  [10518] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protocol_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1317),
  [10521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [10523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1016),
  [10525] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enumerator_list_repeat1, 3, 0, 0),
  [10527] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_identifier_repeat1, 2, 0, 0),
  [10529] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(5113),
  [10532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [10534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator, 3, 0, 16),
  [10536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5856),
  [10538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5713),
  [10540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [10542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [10544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [10546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator, 5, 0, 85),
  [10548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2897),
  [10550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5615),
  [10552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1881),
  [10554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3307),
  [10556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1886),
  [10558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3313),
  [10560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [10562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_selector_expression_repeat1, 2, 0, 0),
  [10564] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_selector_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5132),
  [10567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3784),
  [10569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3785),
  [10571] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_field_declaration_repeat1, 3, 0, 21),
  [10573] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_field_declaration_repeat1, 3, 0, 21),
  [10575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2900),
  [10577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_identifier_repeat2, 2, 0, 0),
  [10579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2508),
  [10581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2507),
  [10583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5680),
  [10585] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_qualifier, 1, 0, 0),
  [10587] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_message_expression_repeat1, 3, 0, 0),
  [10589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1899),
  [10591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3310),
  [10593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [10595] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_repeat1, 2, 0, 0), SHIFT_REPEAT(3452),
  [10598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(679),
  [10600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator, 4, 0, 48),
  [10602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator, 4, 0, 16),
  [10604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5385),
  [10606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6115),
  [10608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2397),
  [10610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6127),
  [10612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3298),
  [10614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [10616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6168),
  [10618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2680),
  [10620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5233),
  [10622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5233),
  [10624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2727),
  [10626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4067),
  [10628] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_message_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(3702),
  [10631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6203),
  [10633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3290),
  [10635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1934),
  [10637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2669),
  [10639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2785),
  [10641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3067),
  [10643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2729),
  [10645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1468),
  [10647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(661),
  [10649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5974),
  [10651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1671),
  [10653] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 4, 0, 86), SHIFT(5580),
  [10656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 4, 0, 86),
  [10658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3308),
  [10660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1657),
  [10662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5298),
  [10664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5298),
  [10666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3999),
  [10668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1622),
  [10670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4849),
  [10672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_clobber_list, 2, 0, 144),
  [10674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4860),
  [10676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3257),
  [10678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4069),
  [10680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5973),
  [10682] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 3, 0, 49), SHIFT(5580),
  [10685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 3, 0, 49),
  [10687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3312),
  [10689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6217),
  [10691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5881),
  [10693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5813),
  [10695] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 3, 0, 86), SHIFT_REPEAT(5580),
  [10698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 3, 0, 86),
  [10700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3309),
  [10702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3299),
  [10704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(95),
  [10706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6165),
  [10708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1653),
  [10710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5219),
  [10712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5219),
  [10714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2355),
  [10716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6079),
  [10718] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1669),
  [10720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0),
  [10722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2056),
  [10724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5218),
  [10726] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5218),
  [10728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1506),
  [10730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2776),
  [10732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2643),
  [10734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(503),
  [10736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5682),
  [10738] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6461),
  [10740] = {.entry = {.count = 3, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat1, 3, 0, 86), REDUCE(aux_sym_parameterized_arguments_repeat2, 3, 0, 86), SHIFT(5580),
  [10744] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat1, 3, 0, 86), REDUCE(aux_sym_parameterized_arguments_repeat2, 3, 0, 86),
  [10747] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat1, 3, 0, 86), SHIFT(3309),
  [10750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_message_expression_repeat2, 2, 0, 27),
  [10752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3702),
  [10754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1633),
  [10756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5232),
  [10758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2777),
  [10760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3997),
  [10762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3061),
  [10764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_message_expression_repeat2, 1, 0, 27),
  [10766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2008),
  [10768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2766),
  [10770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3304),
  [10772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [10774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6169),
  [10776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3045),
  [10778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6606),
  [10780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5790),
  [10782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5259),
  [10784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5259),
  [10786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2209),
  [10788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5923),
  [10790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3819),
  [10792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6738),
  [10794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4139),
  [10796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2029),
  [10798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2732),
  [10800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3985),
  [10802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2736),
  [10804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3302),
  [10806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [10808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6166),
  [10810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3401),
  [10812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2740),
  [10814] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 2, 0, 49), SHIFT_REPEAT(5580),
  [10817] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 2, 0, 49),
  [10819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3306),
  [10821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2769),
  [10823] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_identifier, 3, 0, 0),
  [10825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3300),
  [10827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [10829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6163),
  [10831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2605),
  [10833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6286),
  [10835] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_char_literal_repeat1, 2, 0, 0),
  [10837] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_char_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(5232),
  [10840] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_literal_repeat1, 2, 0, 0),
  [10842] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(5233),
  [10845] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(5233),
  [10848] = {.entry = {.count = 3, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat1, 2, 0, 49), REDUCE(aux_sym_parameterized_arguments_repeat2, 2, 0, 49), SHIFT(5580),
  [10852] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat1, 2, 0, 49), REDUCE(aux_sym_parameterized_arguments_repeat2, 2, 0, 49),
  [10855] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat1, 2, 0, 49), SHIFT(3306),
  [10858] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_identifier_repeat2, 2, 0, 0), SHIFT_REPEAT(5746),
  [10861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2702),
  [10863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1469),
  [10865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2771),
  [10867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1673),
  [10869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5260),
  [10871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5260),
  [10873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4285),
  [10875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2747),
  [10877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5285),
  [10879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5285),
  [10881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3301),
  [10883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4882),
  [10885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_input_operand_list, 3, 0, 131),
  [10887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2774),
  [10889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2237),
  [10891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6034),
  [10893] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(5580),
  [10896] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat1, 2, 0, 0),
  [10898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2033),
  [10900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [10902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6027),
  [10904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3069),
  [10906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5196),
  [10908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5196),
  [10910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4674),
  [10912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3297),
  [10914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [10916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6167),
  [10918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1666),
  [10920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4344),
  [10922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5248),
  [10924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5248),
  [10926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1672),
  [10928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5177),
  [10930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5177),
  [10932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5823),
  [10934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1670),
  [10936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4284),
  [10938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6440),
  [10940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3820),
  [10942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2024),
  [10944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5209),
  [10946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5209),
  [10948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5166),
  [10950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5166),
  [10952] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_gnu_asm_input_operand_list_repeat1, 2, 0, 143), SHIFT_REPEAT(4882),
  [10955] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_input_operand_list_repeat1, 2, 0, 143),
  [10957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_else_in_enumerator, 2, 0, 0),
  [10959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2706),
  [10961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3303),
  [10963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(115),
  [10965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6170),
  [10967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1438),
  [10969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2693),
  [10971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_clobber_list, 3, 0, 150),
  [10973] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protocol_forward_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6231),
  [10976] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_protocol_forward_declaration_repeat1, 2, 0, 0),
  [10978] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_else_in_enumerator, 1, 0, 0),
  [10980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2028),
  [10982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6634),
  [10984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4138),
  [10986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3258),
  [10988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4835),
  [10990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1478),
  [10992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4623),
  [10994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2759),
  [10996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [10998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6695),
  [11000] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1504),
  [11002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5278),
  [11004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5278),
  [11006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3063),
  [11008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5843),
  [11010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2749),
  [11012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2724),
  [11014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3982),
  [11016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6304),
  [11018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5882),
  [11020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4881),
  [11022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_output_operand_list, 3, 0, 131),
  [11024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(868),
  [11026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6303),
  [11028] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_output_operand_list, 2, 0, 105),
  [11030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2115),
  [11032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6555),
  [11034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_params_repeat1, 2, 0, 0),
  [11036] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_params_repeat1, 2, 0, 0), SHIFT_REPEAT(5881),
  [11039] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_input_operand_list, 2, 0, 105),
  [11041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4470),
  [11043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5193),
  [11045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5193),
  [11047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1658),
  [11049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1507),
  [11051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5175),
  [11053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5175),
  [11055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_gnu_asm_output_operand_list_repeat1, 2, 0, 143), SHIFT_REPEAT(4881),
  [11058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_output_operand_list_repeat1, 2, 0, 143),
  [11060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2614),
  [11062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2590),
  [11064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2617),
  [11066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2620),
  [11068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [11070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6468),
  [11072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2624),
  [11074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3051),
  [11076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [11078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5942),
  [11080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(502),
  [11082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6772),
  [11084] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_gnu_asm_clobber_list_repeat1, 2, 0, 156), SHIFT_REPEAT(4849),
  [11087] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_clobber_list_repeat1, 2, 0, 156),
  [11089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2646),
  [11091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5284),
  [11093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5284),
  [11095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1664),
  [11097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5254),
  [11099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5254),
  [11101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(510),
  [11103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1456),
  [11105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5880),
  [11107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [11109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4627),
  [11111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1981),
  [11113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4830),
  [11115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2373),
  [11117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2375),
  [11119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2383),
  [11121] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_attribute, 1, 0, 0),
  [11123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6128),
  [11125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5781),
  [11127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4113),
  [11129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4115),
  [11131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5100),
  [11133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3983),
  [11135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4815),
  [11137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3984),
  [11139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4834),
  [11141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [11143] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_message_expression_repeat2, 2, 0, 58), SHIFT_REPEAT(5208),
  [11146] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_message_expression_repeat2, 2, 0, 58),
  [11148] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_specifier_repeat1, 2, 0, 0), SHIFT_REPEAT(5890),
  [11151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_specifier_repeat1, 2, 0, 0),
  [11153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5887),
  [11155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [11157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4570),
  [11159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_output_operand_list_repeat1, 2, 0, 105),
  [11161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1977),
  [11163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6513),
  [11165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2189),
  [11167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2348),
  [11169] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_version_repeat1, 2, 0, 0), SHIFT_REPEAT(5771),
  [11172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_version_repeat1, 2, 0, 0),
  [11174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_availability_repeat1, 2, 0, 0),
  [11176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3291),
  [11178] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_availability_repeat1, 2, 0, 0), SHIFT_REPEAT(6192),
  [11181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [11183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5336),
  [11185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4117),
  [11187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2380),
  [11189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3305),
  [11191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1887),
  [11193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat1, 3, 0, 86),
  [11195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [11197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5412),
  [11199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [11201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5252),
  [11203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2349),
  [11205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3998),
  [11207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5796),
  [11209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3979),
  [11211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6149),
  [11213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3383),
  [11215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1638),
  [11217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [11219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5865),
  [11221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6505),
  [11223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6156),
  [11225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3993),
  [11227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [11229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6724),
  [11231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1942),
  [11233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1898),
  [11235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6622),
  [11237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1949),
  [11239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2074),
  [11241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2876),
  [11243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5227),
  [11245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6501),
  [11247] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 2, 0, 0), SHIFT_REPEAT(5412),
  [11250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 2, 0, 0),
  [11252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2928),
  [11254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3433),
  [11256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6734),
  [11258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5700),
  [11260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1665),
  [11262] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_availability_attribute_specifier_repeat1, 2, 0, 0), SHIFT_REPEAT(4638),
  [11265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_output_operand, 4, 0, 149),
  [11267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3375),
  [11269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_input_operand_list_repeat1, 2, 0, 105),
  [11271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [11273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5948),
  [11275] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_goto_list, 2, 0, 151),
  [11277] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat2, 2, 0, 0), SHIFT_REPEAT(3305),
  [11280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat2, 2, 0, 0),
  [11282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6030),
  [11284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2880),
  [11286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(979),
  [11288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5237),
  [11290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1464),
  [11292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6442),
  [11294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2915),
  [11296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1941),
  [11298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2671),
  [11300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6522),
  [11302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5806),
  [11304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(469),
  [11306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1039),
  [11308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6077),
  [11310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5371),
  [11312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3777),
  [11314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript_range_designator, 5, 0, 155),
  [11316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5771),
  [11318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5774),
  [11320] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generic_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(4084),
  [11323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_expression_repeat1, 2, 0, 0),
  [11325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5279),
  [11327] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_import_repeat1, 2, 0, 0),
  [11329] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_import_repeat1, 2, 0, 0), SHIFT_REPEAT(6724),
  [11332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_input_operand, 4, 0, 149),
  [11334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_clobber_list_repeat1, 2, 0, 144),
  [11336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6697),
  [11338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6157),
  [11340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3981),
  [11342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6255),
  [11344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(488),
  [11346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6310),
  [11348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5566),
  [11350] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2419),
  [11353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_goto_list, 3, 0, 157),
  [11355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [11357] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_property_implementation_repeat1, 2, 0, 0),
  [11359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6548),
  [11361] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameter_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1941),
  [11364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameter_list_repeat1, 2, 0, 0),
  [11366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2382),
  [11368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1484),
  [11370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6584),
  [11372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1928),
  [11374] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6556),
  [11377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2353),
  [11379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6660),
  [11381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1929),
  [11383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6703),
  [11385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5688),
  [11387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [11389] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_property_implementation_repeat1, 2, 0, 0), SHIFT_REPEAT(6115),
  [11392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6539),
  [11394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1775),
  [11396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6018),
  [11398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [11400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6111),
  [11402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5333),
  [11404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2368),
  [11406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6520),
  [11408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5348),
  [11410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5762),
  [11412] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_platform, 1, 0, 0),
  [11414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1037),
  [11416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5356),
  [11418] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__old_style_parameter_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5666),
  [11421] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__old_style_parameter_list_repeat1, 2, 0, 0),
  [11423] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4612),
  [11426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6003),
  [11428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_version, 1, 0, 0),
  [11430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5723),
  [11432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1777),
  [11434] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dictionary_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(3436),
  [11437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_dictionary_literal_repeat1, 2, 0, 0),
  [11439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1778),
  [11441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(521),
  [11443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6019),
  [11445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1001),
  [11447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3784),
  [11449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1660),
  [11451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3785),
  [11453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(662),
  [11455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1762),
  [11457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3296),
  [11459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3238),
  [11461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [11463] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_gnu_asm_goto_list_repeat1, 2, 0, 159), SHIFT_REPEAT(5948),
  [11466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_goto_list_repeat1, 2, 0, 159),
  [11468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [11470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1776),
  [11472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1780),
  [11474] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_repeat1, 2, 0, 0), SHIFT_REPEAT(3521),
  [11477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [11479] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_method_type_repeat1, 2, 0, 0), SHIFT_REPEAT(3296),
  [11482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_type_repeat1, 2, 0, 0),
  [11484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_output_operand, 7, 0, 160),
  [11486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2897),
  [11488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1630),
  [11490] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_input_operand, 7, 0, 160),
  [11492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(663),
  [11494] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generic_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(6513),
  [11497] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_arguments_repeat1, 2, 0, 0),
  [11499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3052),
  [11501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1784),
  [11503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1785),
  [11505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(664),
  [11507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6532),
  [11509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3234),
  [11511] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 3, 0, 92),
  [11513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3416),
  [11515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1473),
  [11517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1474),
  [11519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5656),
  [11521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6596),
  [11523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1800),
  [11525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3782),
  [11527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3783),
  [11529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2454),
  [11531] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_available_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5815),
  [11534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1481),
  [11536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3381),
  [11538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [11540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3946),
  [11542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [11544] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5865),
  [11547] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_declaration_repeat1, 2, 0, 0),
  [11549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(600),
  [11551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2662),
  [11553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3849),
  [11555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1736),
  [11557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_attribute, 3, 0, 0),
  [11559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5855),
  [11561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1646),
  [11563] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_property_attributes_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5781),
  [11566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_property_attributes_declaration_repeat1, 2, 0, 0),
  [11568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1895),
  [11570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(669),
  [11572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5357),
  [11574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2900),
  [11576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5202),
  [11578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1013),
  [11580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5234),
  [11582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6423),
  [11584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3066),
  [11586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1632),
  [11588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1897),
  [11590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [11592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1896),
  [11594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3988),
  [11596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3989),
  [11598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [11600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1992),
  [11602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1508),
  [11604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1510),
  [11606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3405),
  [11608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3348),
  [11610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(862),
  [11612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3976),
  [11614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5157),
  [11616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1492),
  [11618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3830),
  [11620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3226),
  [11622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2398),
  [11624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [11626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat1, 2, 0, 49),
  [11628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(611),
  [11630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(888),
  [11632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4656),
  [11634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1888),
  [11636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5787),
  [11638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1889),
  [11640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript_designator, 3, 0, 0),
  [11642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2234),
  [11644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(626),
  [11646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1453),
  [11648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1455),
  [11650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(628),
  [11652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [11654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5161),
  [11656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1491),
  [11658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5288),
  [11660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5871),
  [11662] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_initializer_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3293),
  [11665] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_designator, 2, 0, 117),
  [11667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2213),
  [11669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3774),
  [11671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2217),
  [11673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3282),
  [11675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1967),
  [11677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [11679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1968),
  [11681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3280),
  [11683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5666),
  [11685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2904),
  [11687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6353),
  [11689] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_repeat1, 2, 0, 0), SHIFT_REPEAT(3577),
  [11692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1494),
  [11694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1495),
  [11696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5162),
  [11698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_availability_attribute_specifier_repeat1, 4, 0, 0),
  [11700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3321),
  [11702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4141),
  [11704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4140),
  [11706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [11708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6398),
  [11710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3319),
  [11712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5459),
  [11714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2044),
  [11716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3607),
  [11718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5324),
  [11720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6071),
  [11722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [11724] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5804),
  [11726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4973),
  [11728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3952),
  [11730] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3951),
  [11732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1036),
  [11734] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6245),
  [11736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4639),
  [11738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6322),
  [11740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5224),
  [11742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2114),
  [11744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6536),
  [11746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2202),
  [11748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6378),
  [11750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5555),
  [11752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6719),
  [11754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(449),
  [11756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5810),
  [11758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [11760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6774),
  [11762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5142),
  [11764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_declaration_repeat1, 3, 0, 0),
  [11766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2045),
  [11768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_property_implementation_repeat1, 4, 0, 0),
  [11770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5906),
  [11772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6320),
  [11774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2246),
  [11776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4951),
  [11778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3368),
  [11780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_version, 5, 0, 0),
  [11782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 5, 0, 49),
  [11784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1983),
  [11786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 5, 0, 86),
  [11788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [11790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6567),
  [11792] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 4, 0, 26),
  [11794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4012),
  [11796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4003),
  [11798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5745),
  [11800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5983),
  [11802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4028),
  [11804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4015),
  [11806] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_version, 4, 0, 0),
  [11808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4068),
  [11810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4062),
  [11812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3026),
  [11814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3025),
  [11816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_available_expression_repeat1, 3, 0, 0),
  [11818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [11820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5832),
  [11822] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 6, 0, 86),
  [11824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2204),
  [11826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6632),
  [11828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4136),
  [11830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4135),
  [11832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2608),
  [11834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6788),
  [11836] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameter_list_repeat1, 2, 0, 0), REDUCE(aux_sym__old_style_parameter_list_repeat1, 2, 0, 0),
  [11839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [11841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6247),
  [11843] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 7, 0, 92),
  [11845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3835),
  [11847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3834),
  [11849] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 6, 0, 92),
  [11851] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 3, 0, 73),
  [11853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4975),
  [11855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(863),
  [11857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6673),
  [11859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(522),
  [11861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6274),
  [11863] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 1, 0, 0),
  [11865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2039),
  [11867] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_params, 4, 0, 0),
  [11869] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_params, 4, 0, 0),
  [11871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4826),
  [11873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5591),
  [11875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4848),
  [11877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5756),
  [11879] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_params, 5, 0, 0),
  [11881] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_params, 5, 0, 0),
  [11883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2041),
  [11885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3040),
  [11887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3039),
  [11889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2020),
  [11891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(901),
  [11893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6104),
  [11895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4064),
  [11897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4046),
  [11899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameterized_arguments_repeat2, 4, 0, 49),
  [11901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4034),
  [11903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4019),
  [11905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4981),
  [11907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [11909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6669),
  [11911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_goto_list_repeat1, 2, 0, 151),
  [11913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5517),
  [11915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5965),
  [11917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [11919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6438),
  [11921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5359),
  [11923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6388),
  [11925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4134),
  [11927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4132),
  [11929] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3872),
  [11931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3871),
  [11933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2208),
  [11935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5918),
  [11937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1040),
  [11939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5858),
  [11941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_attribute, 4, 0, 0),
  [11943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4055),
  [11945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4053),
  [11947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [11949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6143),
  [11951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4672),
  [11953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6773),
  [11955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6521),
  [11957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_arguments_repeat1, 2, 0, 49),
  [11959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4833),
  [11961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5596),
  [11963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5441),
  [11965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6685),
  [11967] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_params, 2, 0, 0),
  [11969] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_params, 2, 0, 0),
  [11971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [11973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5795),
  [11975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2043),
  [11977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_availability_attribute_specifier_repeat1, 5, 0, 0),
  [11979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3034),
  [11981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3032),
  [11983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5648),
  [11985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6026),
  [11987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [11989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6463),
  [11991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5428),
  [11993] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_goto_list, 1, 0, 0),
  [11995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5531),
  [11997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_params, 3, 0, 0),
  [11999] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_params, 3, 0, 0),
  [12001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(977),
  [12003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6294),
  [12005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5613),
  [12007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6412),
  [12009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(641),
  [12011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6779),
  [12013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 5, 0, 26),
  [12015] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 5, 0, 92),
  [12017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5802),
  [12019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1990),
  [12021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6729),
  [12023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [12025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6147),
  [12027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2604),
  [12029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6200),
  [12031] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 3, 0, 26),
  [12033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2292),
  [12035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5953),
  [12037] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_method_type_repeat1, 3, 0, 0),
  [12039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4183),
  [12041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5397),
  [12043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1102),
  [12045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6404),
  [12047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2710),
  [12049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4188),
  [12051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1062),
  [12053] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator, 5, 0, 115),
  [12055] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_field_declaration_list, 5, 0, 115),
  [12057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4197),
  [12059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2588),
  [12061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3715),
  [12063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5249),
  [12065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8),
  [12067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2211),
  [12069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1976),
  [12071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [12073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3376),
  [12075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4170),
  [12077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2216),
  [12079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [12081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5961),
  [12083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2293),
  [12085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5472),
  [12087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5975),
  [12089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1164),
  [12091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5979),
  [12093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3729),
  [12095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5296),
  [12097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2712),
  [12099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2147),
  [12101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5876),
  [12103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [12105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5729),
  [12107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [12109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2589),
  [12111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5120),
  [12113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [12115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6445),
  [12117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1111),
  [12119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3166),
  [12121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5833),
  [12123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5176),
  [12125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [12127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1166),
  [12129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2298),
  [12131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2059),
  [12133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2300),
  [12135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6158),
  [12137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5797),
  [12139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2148),
  [12141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2445),
  [12143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(726),
  [12145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2465),
  [12147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2149),
  [12149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2954),
  [12151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5541),
  [12153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2150),
  [12155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2591),
  [12157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [12159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1114),
  [12161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [12163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1116),
  [12165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5846),
  [12167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5561),
  [12169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [12171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1117),
  [12173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [12175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [12177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(512),
  [12179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2152),
  [12181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5579),
  [12183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2153),
  [12185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6342),
  [12187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2154),
  [12189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4955),
  [12191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5199),
  [12193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [12195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [12197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [12199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [12201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1121),
  [12203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2155),
  [12205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [12207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2156),
  [12209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6420),
  [12211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5605),
  [12213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6429),
  [12215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1169),
  [12217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6433),
  [12219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1940),
  [12221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5264),
  [12223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2157),
  [12225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1467),
  [12227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [12229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2317),
  [12231] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_objc_bridge, 7, 0, 0),
  [12233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5155),
  [12235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6612),
  [12237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5101),
  [12239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3115),
  [12241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3737),
  [12243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2311),
  [12245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5245),
  [12247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1171),
  [12249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [12251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2164),
  [12253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5102),
  [12255] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_availability, 5, 0, 0),
  [12257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6586),
  [12259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2176),
  [12261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [12263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2479),
  [12265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6195),
  [12267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2341),
  [12269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5672),
  [12271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [12273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2243),
  [12275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2343),
  [12277] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_availability, 7, 0, 0),
  [12279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2194),
  [12281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6075),
  [12283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [12285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5685),
  [12287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2199),
  [12289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [12291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2193),
  [12293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2134),
  [12295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4192),
  [12297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [12299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5699),
  [12301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1943),
  [12303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5308),
  [12305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [12307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [12309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2253),
  [12311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2331),
  [12313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2247),
  [12315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2103),
  [12317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6204),
  [12319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6588),
  [12321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5179),
  [12323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5312),
  [12325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2113),
  [12327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2254),
  [12329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2255),
  [12331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3174),
  [12333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2261),
  [12335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5192),
  [12337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4169),
  [12339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [12341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [12343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [12345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6546),
  [12347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5124),
  [12349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2266),
  [12351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2435),
  [12353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2136),
  [12355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5329),
  [12357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2666),
  [12359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2139),
  [12361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1946),
  [12363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6524),
  [12365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2279),
  [12367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [12369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2063),
  [12371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5331),
  [12373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5791),
  [12375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2302),
  [12377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [12379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2306),
  [12381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2316),
  [12383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2324),
  [12385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5335),
  [12387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [12389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5231),
  [12391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(14),
  [12393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [12395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2326),
  [12397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6258),
  [12399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6261),
  [12401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5251),
  [12403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2329),
  [12405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2720),
  [12407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3079),
  [12409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_interface_declaration, 5, 0, 115),
  [12411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [12413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2551),
  [12415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [12417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6431),
  [12419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(895),
  [12421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2722),
  [12423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4070),
  [12425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [12427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [12429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1174),
  [12431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5107),
  [12433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2723),
  [12435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [12437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5133),
  [12439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5479),
  [12441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5349),
  [12443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5292),
  [12445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [12447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [12449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6540),
  [12451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3088),
  [12453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2713),
  [12455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [12457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5482),
  [12459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5600),
  [12461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [12463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1011),
  [12465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2322),
  [12467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [12469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2726),
  [12471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5127),
  [12473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [12475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5214),
  [12477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [12479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [12481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3095),
  [12483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [12485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1002),
  [12487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5612),
  [12489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2756),
  [12491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2319),
  [12493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [12495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4048),
  [12497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [12499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [12501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [12503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3102),
  [12505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [12507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [12509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1497),
  [12511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1146),
  [12513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(755),
  [12515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4194),
  [12517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2728),
  [12519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [12521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1134),
  [12523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [12525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3110),
  [12527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [12529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [12531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [12533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [12535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(113),
  [12537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [12539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(120),
  [12541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [12543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2700),
  [12545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5821),
  [12547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6259),
  [12549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1605),
  [12551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [12553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2065),
  [12555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4096),
  [12557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(760),
  [12559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3332),
  [12561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1584),
  [12563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5003),
  [12565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4628),
  [12567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2716),
  [12569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5864),
  [12571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3317),
  [12573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_implementation_definition, 5, 0, 115),
  [12575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5364),
  [12577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5290),
  [12579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4),
  [12581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [12583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2699),
  [12585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [12587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2580),
  [12589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2573),
  [12591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2606),
  [12593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2701),
  [12595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5095),
  [12597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_availability, 4, 0, 0),
  [12599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [12601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5727),
  [12603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [12605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2009),
  [12607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5737),
  [12609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6235),
  [12611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [12613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [12615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [12617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2718),
  [12619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5547),
  [12621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5817),
  [12623] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_based_modifier, 2, 0, 0),
  [12625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6264),
  [12627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1175),
  [12629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6266),
  [12631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2523),
  [12633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2731),
  [12635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5297),
  [12637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6541),
  [12639] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif, 4, 0, 84),
  [12641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(928),
  [12643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3350),
  [12645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2629),
  [12647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3386),
  [12649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5507),
  [12651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4083),
  [12653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6449),
  [12655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(691),
  [12657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6452),
  [12659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2303),
  [12661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3333),
  [12663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4174),
  [12665] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_field_declaration_list, 3, 0, 48),
  [12667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3739),
  [12669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2196),
  [12671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2197),
  [12673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1047),
  [12675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2417),
  [12677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [12679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4061),
  [12681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [12683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [12685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [12687] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_objc_bridge, 8, 0, 0),
  [12689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [12691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1143),
  [12693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(998),
  [12695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2532),
  [12697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1163),
  [12699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6334),
  [12701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(999),
  [12703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2188),
  [12705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2291),
  [12707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5717),
  [12709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [12711] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator, 4, 0, 84),
  [12713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [12715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1000),
  [12717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3532),
  [12719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5408),
  [12721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2703),
  [12723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2364),
  [12725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2198),
  [12727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5425),
  [12729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(473),
  [12731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5099),
  [12733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [12735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5814),
  [12737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [12739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1489),
  [12741] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_field_declaration_list, 4, 0, 84),
  [12743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_objc_bridge, 9, 0, 0),
  [12745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3286),
  [12747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2607),
  [12749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3661),
  [12751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [12753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1973),
  [12755] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_field_declaration_list, 4, 0, 85),
  [12757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5598),
  [12759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [12761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1058),
  [12763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(921),
  [12765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3171),
  [12767] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_declarators, 3, 0, 56),
  [12769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4665),
  [12771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3680),
  [12773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1059),
  [12775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3218),
  [12777] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef, 4, 0, 85),
  [12779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [12781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [12783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5749),
  [12785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [12787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3314),
  [12789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [12791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [12793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5512),
  [12795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [12797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1978),
  [12799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2424),
  [12801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5305),
  [12803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2422),
  [12805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2064),
  [12807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2336),
  [12809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1168),
  [12811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6464),
  [12813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2075),
  [12815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2742),
  [12817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [12819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5712),
  [12821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6455),
  [12823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [12825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(695),
  [12827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2400),
  [12829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3762),
  [12831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2151),
  [12833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5152),
  [12835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2743),
  [12837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [12839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4970),
  [12841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1795),
  [12843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2733),
  [12845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2421),
  [12847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2418),
  [12849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2734),
  [12851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1801),
  [12853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2735),
  [12855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [12857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [12859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6506),
  [12861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3358),
  [12863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2704),
  [12865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1075),
  [12867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [12869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2459),
  [12871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2542),
  [12873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [12875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3340),
  [12877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5788),
  [12879] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_declarators, 2, 0, 24),
  [12881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1515),
  [12883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [12885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(997),
  [12887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [12889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3945),
  [12891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [12893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [12895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2473),
  [12897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(945),
  [12899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2737),
  [12901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5812),
  [12903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1602),
  [12905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [12907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2210),
  [12909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2705),
  [12911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5001),
  [12913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1514),
  [12915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef, 3, 0, 48),
  [12917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2738),
  [12919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [12921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2739),
  [12923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6106),
  [12925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(924),
  [12927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2741),
  [12929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5748),
  [12931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3858),
  [12933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2143),
  [12935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3991),
  [12937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1488),
  [12939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(504),
  [12941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [12943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3339),
  [12945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1649),
  [12947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(705),
  [12949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2272),
  [12951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4052),
  [12953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4824),
  [12955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [12957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [12959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6707),
  [12961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1647),
  [12963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5655),
  [12965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2338),
  [12967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [12969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2034),
  [12971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2644),
  [12973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1452),
  [12975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6750),
  [12977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [12979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(581),
  [12981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(973),
  [12983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5200),
  [12985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2144),
  [12987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(582),
  [12989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2122),
  [12991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2762),
  [12993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [12995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [12997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [12999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2201),
  [13001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [13003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2146),
  [13005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5156),
  [13007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5074),
  [13009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [13011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(650),
  [13013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1882),
  [13015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5849),
  [13017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5764),
  [13019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2358),
  [13021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2745),
  [13023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(961),
  [13025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5170),
  [13027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5818),
  [13029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1613),
  [13031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2402),
  [13033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6646),
  [13035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4950),
  [13037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1578),
  [13039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6663),
  [13041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2746),
  [13043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1061),
  [13045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [13047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3995),
  [13049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [13051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1151),
  [13053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5839),
  [13055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [13057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(518),
  [13059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [13061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4629),
  [13063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(843),
  [13065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [13067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(845),
  [13069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5693),
  [13071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4058),
  [13073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(364),
  [13075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(823),
  [13077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [13079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2707),
  [13081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(678),
  [13083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [13085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5874),
  [13087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1577),
  [13089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1064),
  [13091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4954),
  [13093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [13095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2764),
  [13097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2690),
  [13099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [13101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [13103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5889),
  [13105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5829),
  [13107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(593),
  [13109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3219),
  [13111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [13113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(651),
  [13115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [13117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(717),
  [13119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5708),
  [13121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2800),
  [13123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2554),
  [13125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [13127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1516),
  [13129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(718),
  [13131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5189),
  [13133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(719),
  [13135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5724),
  [13137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2761),
  [13139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4071),
  [13141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [13143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3684),
  [13145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5735),
  [13147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(830),
  [13149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [13151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [13153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5281),
  [13155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5862),
  [13157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [13159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5377),
  [13161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(597),
  [13163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5163),
  [13165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2645),
  [13167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2651),
  [13169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3220),
  [13171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6005),
  [13173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2401),
  [13175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5752),
  [13177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2775),
  [13179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(730),
  [13181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(598),
  [13183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5755),
  [13185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [13187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [13189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2068),
  [13191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [13193] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [13195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [13197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2118),
  [13199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1936),
  [13201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5147),
  [13203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2258),
  [13205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1020),
  [13207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2592),
  [13209] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_objc_bridge, 6, 0, 0),
  [13211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2593),
  [13213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3439),
  [13215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(531),
  [13217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(732),
  [13219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5753),
  [13221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(734),
  [13223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(735),
  [13225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(737),
  [13227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_availability, 6, 0, 0),
  [13229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1958),
  [13231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2119),
  [13233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5194),
  [13235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5274),
  [13237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [13239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [13241] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_interface_declaration, 4, 0, 84),
  [13243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2688),
  [13245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1960),
  [13247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5599),
  [13249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6410),
  [13251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1961),
  [13253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [13255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(447),
  [13257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2696),
  [13259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif, 5, 0, 115),
  [13261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4063),
  [13263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [13265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2719),
  [13267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [13269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(692),
  [13271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [13273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [13275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2406),
  [13277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2124),
  [13279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [13281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3337),
  [13283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1962),
  [13285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4105),
  [13287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3826),
  [13289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [13291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3573),
  [13293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [13295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1158),
  [13297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6325),
  [13299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5836),
  [13301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [13303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5402),
  [13305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6427),
  [13307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5380),
  [13309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3009),
  [13311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [13313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1176),
  [13315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6476),
  [13317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [13319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5257),
  [13321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2765),
  [13323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(889),
  [13325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5776),
  [13327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(922),
  [13329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2351),
  [13331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(789),
  [13333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4993),
  [13335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7),
  [13337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [13339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(813),
  [13341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(743),
  [13343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(744),
  [13345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(745),
  [13347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(814),
  [13349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2340),
  [13351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3760),
  [13353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4072),
  [13355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4634),
  [13357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5828),
  [13359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5226),
  [13361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5860),
  [13363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [13365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1636),
  [13367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1466),
  [13369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5149),
  [13371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5706),
  [13373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2242),
  [13375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6718),
  [13377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2057),
  [13379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2130),
  [13381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_implementation_definition, 4, 0, 84),
  [13383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3139),
  [13385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5757),
  [13387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1093),
  [13389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1094),
  [13391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1095),
  [13393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5576),
  [13395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2625),
  [13397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3579),
  [13399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2594),
  [13401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5309),
  [13403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2595),
  [13405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [13407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(704),
  [13409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3351),
  [13411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [13413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2599),
  [13415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [13417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2584),
  [13419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(627),
  [13421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2602),
  [13423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2683),
  [13425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4106),
  [13427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1012),
  [13429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1448),
  [13431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1177),
  [13433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2768),
  [13435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2128),
  [13437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [13439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [13441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [13443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(537),
  [13445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6457),
  [13447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3360),
  [13449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(894),
  [13451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5769),
  [13453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3887),
  [13455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(736),
  [13457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3655),
  [13459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1527),
  [13461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5747),
  [13463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1493),
  [13465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(746),
  [13467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(747),
  [13469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [13471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5376),
  [13473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [13475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2466),
  [13477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2282),
  [13479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4092),
  [13481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(567),
  [13483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2647),
  [13485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [13487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1074),
  [13489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [13491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [13493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(818),
  [13495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6784),
  [13497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(637),
  [13499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2224),
  [13501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [13503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1642),
  [13505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2478),
  [13507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3663),
  [13509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5985),
  [13511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1017),
  [13513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6131),
  [13515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [13517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3664),
  [13519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [13521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [13523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2770),
  [13525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(786),
  [13527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5319),
  [13529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3164),
  [13531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2214),
  [13533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2131),
  [13535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3372),
  [13537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5784),
  [13539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(822),
  [13541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2030),
  [13543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(749),
  [13545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [13547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1097),
  [13549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [13551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3763),
  [13553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3899),
  [13555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3665),
  [13557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2773),
  [13559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5740),
  [13561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2610),
  [13563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5835),
  [13565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2612),
  [13567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [13569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1098),
  [13571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3908),
  [13573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3669),
  [13575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2613),
  [13577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [13579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5934),
  [13581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [13583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2654),
  [13585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3915),
  [13587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3673),
  [13589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [13591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2619),
  [13593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(549),
  [13595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2622),
  [13597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3922),
  [13599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3676),
  [13601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3928),
  [13603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3677),
  [13605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5684),
  [13607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3934),
  [13609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1099),
  [13611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3940),
  [13613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [13615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4076),
  [13617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3683),
  [13619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5368),
  [13621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6298),
  [13623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [13625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [13627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2337),
  [13629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [13631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(643),
  [13633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2274),
  [13635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3364),
  [13637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2223),
  [13639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1100),
  [13641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6620),
  [13643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [13645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2858),
  [13647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3745),
  [13649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2618),
};

#ifdef __cplusplus
extern "C" {
#endif
#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_objc(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .supertype_map_slices = ts_supertype_map_slices,
    .supertype_map_entries = ts_supertype_map_entries,
    .supertype_symbols = ts_supertype_symbols,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_identifier,
    .primary_state_ids = ts_primary_state_ids,
    .name = "objc",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 0,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
